package actions;

import gui.CpfViewer;
import gui.CpfConfig;
import gui.ValidateEffectiveDateRange;
import library.FileNamingConventions;
import library.LoadedFile;
import library.LoadedFileSchema;
import library.LoadedFileType;
import gui.ParameterTablePanel;
import workers.CpfExportThread;
import javax.swing.AbstractAction;
import javax.swing.JOptionPane;
import javax.swing.ImageIcon;
import java.awt.event.ActionEvent;
import java.io.File;
import java.util.ArrayList;
import java.util.List;
import org.apache.log4j.Logger;

/**
 * Action to export a file or files to the evaluation database.
 */
public class ExportAction extends AbstractAction {

    private CpfViewer application;
    private ParameterTablePanel parameterTablePanel = null;
    private static final Logger logger = Logger.getLogger(ExportAction.class);

    /**
     * Create an export action for the single file on the given panel --
     * used in the parameter toolbar.
     * @param v   the CPF Viewer application
     * @param p   the panel for the file to be exported
     */
    public ExportAction(CpfViewer v, ParameterTablePanel p) {
        super();
        putValue(NAME, null);
        putValue(SHORT_DESCRIPTION, "Export this file to the EVAL database");
        putValue(SMALL_ICON, new ImageIcon(getClass().getResource(
            CpfConfig.EXPORT_TEST_DB_ICON)));
        application = v;
        parameterTablePanel = p;
    }

    /**
     * Create an export action for all selected files -- used in the File menu.
     * @param v   the CPF Viewer application
     */
    public ExportAction(CpfViewer v) {
        super("Export");
        putValue(SHORT_DESCRIPTION,
            "Export all open selected files to the EVAL database");
        application = v;
        parameterTablePanel = null;
    }

    /**
     * Perform the Export action when the action is invoked.  Basic validation
     * is done, then a thread is created to serially export all designated 
     * files.
     * @param e  the invoking event
     */
    @Override
    public void actionPerformed(ActionEvent e) {
        ArrayList<LoadedFile> inputFiles = new ArrayList<LoadedFile>();
        if (parameterTablePanel == null) {
            // export all selected files
            for (ParameterTablePanel ptp :
                application.parameterViewerPanel.getSelectedPanels()) {
                int lfIndex = ptp.getLoadedFileIndex();
                inputFiles.add(application.loadedFileList.getFile(lfIndex));
            }
        } else {
            // export just the file in the given panel
            int currentFile = parameterTablePanel.getLoadedFileIndex();
            inputFiles.add(application.loadedFileList.getFile(currentFile));
        }

        // Do some basic validation on the files first
        ArrayList<LoadedFile> filesToExport = new ArrayList<LoadedFile>();
        for (LoadedFile lf : inputFiles) {
            // Make sure the file name is "eval_" and make sure the file name
            // (as it's loaded in the GUI) matches the file name contained
            // in the file attributes group
            File file = new File(lf.getFileName());
            String baseName = file.getName();

            if (!ValidateEffectiveDateRange.validate(this.application, lf)) {
                continue;
            }

            String groupName = System.getProperty(
                FileNamingConventions.getFileType(baseName).toString().
                toLowerCase() + ".group.name.file_attributes",
                "FILE_ATTRIBUTES");
            String parameterName = System.getProperty(
                FileNamingConventions.getFileType(baseName).toString().
                toLowerCase() + ".parameter.name.file_attributes.file_name",
                "File_Name");
            List<String> fileNameContainedInFile
                = lf.getParameterValues(groupName, parameterName);

            if (!FileNamingConventions.getSchema(baseName).equals(
                LoadedFileSchema.EVAL)
                || FileNamingConventions.getFileType(baseName)
                == LoadedFileType.OTHER
                || !baseName.equals(fileNameContainedInFile.get(0))) {
                JOptionPane.showMessageDialog(application,
                    "Cannot export " + file.getPath()
                    + ": Please rename as an eval CPF/BPF file and verify file "
                    + "name contained within file",
                    "Error", JOptionPane.ERROR_MESSAGE);
            } else {
                filesToExport.add(lf);
                logger.debug("Exporting "
                    + lf.getFileName() + " to EVAL database");
            }
        }
        if (filesToExport.isEmpty()) {
            return;
        }

        // Export is done in a thread, but there must not be any other
        // background threads running
        if (application.statusBar.progress.isBusy()) {
            JOptionPane.showMessageDialog(application,
                "Cannot export while the GUI is busy.  Please try later.",
                "Busy", JOptionPane.ERROR_MESSAGE);
            return;
        }

        // Next, create a thread to export all the files serially
        try {
            if (!application.databaseLibrary.login(LoadedFileSchema.EVAL)) {
                return; // user cancelled
            }
        } catch (Exception ex) {
            logger.error("Error connecting to EVAL database", ex);
            JOptionPane.showMessageDialog(application,
                "Error connecting to EVAL database: " + ex.getMessage(),
                "Database Error", JOptionPane.ERROR_MESSAGE);
            return;
        }
        application.statusBar.progress.setBusy();
        CpfExportThread exportThread = new CpfExportThread(
            application, filesToExport.toArray(new LoadedFile[0]));
        exportThread.execute();
    }
}
