package datagroup;

import java.net.MalformedURLException;
import java.net.URL;
import library.FileCompare;
import library.LoadedFile;
import org.apache.log4j.Logger;


/**
 * Provides a standalone utility for diffing two CPFs.
 */
public class DiffCpfMain {
    
    private static final int DIFF_NOT_FOUND = 0;
    private static final int DIFF_FOUND = 1;
    private static final int USAGE_ERROR = -1;
    private static final int PARSE_ERROR = -2;
    
    public static void main(String[] args) {
        
        if (args.length != 2) {
            printUsage();
            System.exit(USAGE_ERROR);
        }
        String sourceCpfFilename = args[0];
        String diffCpfFilename = args[1];
        
        // Set logging
        Logger logger = Logger.getLogger(
                DiffCpfMain.class.getPackage().getName());
        
        // Parse the first CPF file
        URL fileUrl = null;
        try {
            fileUrl = new URL("file", "localhost", sourceCpfFilename);
        } catch (MalformedURLException ex) {
            logger.error("Error getting URL for " + sourceCpfFilename, ex);
        }
        LoadedFile sourceCpfFile = null;
        try {
            sourceCpfFile = LoadedFile.create(fileUrl);
        } catch (Exception ex) {
            logger.error("Error parsing " + sourceCpfFilename, ex);
        }
        if (sourceCpfFile == null) {
            System.err.println("Unable to parse CPF file " 
                + fileUrl.getFile());
            System.exit(PARSE_ERROR);
        }
        
        // Parse the second CPF file
        try {
            fileUrl = new URL("file", "localhost", diffCpfFilename);
        } catch (MalformedURLException ex) {
            logger.error("Error getting URL for " + diffCpfFilename, ex);
        }
        LoadedFile diffCpfFile = null;
        try {
            diffCpfFile = LoadedFile.create(fileUrl);
        } catch (Exception ex) {
            logger.error("Error parsing " + diffCpfFilename, ex);
        }
        if (diffCpfFile == null) {
            System.err.println("Unable to parse CPF file " 
                + fileUrl.getFile());
            System.exit(PARSE_ERROR);
        }
        
        FileCompare compare = new FileCompare(sourceCpfFile, diffCpfFile);
        compare.excludeGroup("FILE_ATTRIBUTES");
        compare.compareFiles();
        
        if (compare.differencesFound())
            System.exit(DIFF_FOUND);

        System.exit(DIFF_NOT_FOUND);
    }
    
    /**
     * Print usage information for the diff utility
     */
    private static void printUsage() {
        String caller = System.getProperty("calling.script");
        if (caller == null) {
            System.out.println("java -jar [-Dcalling.script] <classpath> "
                + "datagroup.DiffCpfMain <CPF_File_1> <CPF_File_2>");
        } else {
            System.out.println(caller + " <CPF_File_1> <CPF_File_2>");
        }
    }
}
