package datagroup;

import gov.usgs.ias.constant.Constants;
import gov.usgs.ias.cps.CpsStatusEnum;
import gov.usgs.ias.dao.DataConnection;
import gov.usgs.ias.dao.cpf.CpfDAO;
import java.net.MalformedURLException;
import java.net.URL;
import library.LoadedFile;
import org.apache.commons.cli.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;

/**
 * Provides a standalone utility to insert a CPF into the evaluation database 
 */
public class InsertCpfMain {

    private static final String connStringProperty = "cpf.db.conn";
    private static final String daoAuth = Constants.DAO_AUTH_IAS_EVAL;

    private static final int USAGE_ERROR = 1;
    private static final int PARSE_ERROR = 2;
    private static final int WRITE_ERROR = 3;
   
    /* No-op constructor */
    private InsertCpfMain() {
    }
     
    public static void main(String[] args) {
        
        Options optionList = getCmdOptions();
        
        CommandLineParser parser = new GnuParser();
        CommandLine line = null;
        try {
            line = parser.parse(optionList, args);
        } catch (ParseException ex) {
            printUsage(optionList);   
            System.exit(USAGE_ERROR);
        }

        if (line.hasOption("h") || line.hasOption("help")) {
            printUsage(optionList);
            System.exit(0);
        }

        // Set logging
        Logger logger = Logger.getLogger(
                MergeGroupMain.class.getPackage().getName());
        
        if (line.hasOption("d") || line.hasOption("debug")) {
            logger.setLevel(Level.TRACE);
        } else {
            logger.setLevel(Level.OFF);
        }
        
        String cpfFilename = line.getOptionValue("f");
        
        // Build the CPF URL
        URL fileUrl = null;
        try {
            fileUrl = new URL("file", "localhost", cpfFilename);
        } catch (MalformedURLException ex) {
            logger.error("Error getting URL for " + cpfFilename, ex);
        }
        
        // Load the CPF from file
        LoadedFile cpfFile = null;
        try {
            cpfFile = LoadedFile.create(fileUrl);
        } catch (Exception ex) {
            logger.error("Error parsing " + cpfFilename, ex);
        }
        if (cpfFile == null) {
            System.err.println("Unable to parse CPF ODL file " 
                + fileUrl.getFile());
            System.exit(PARSE_ERROR);
        }
        
        // Insert the CPF into the database
        DataConnection conn = new DataConnection(connStringProperty);
        CpfDAO dbAccess = new CpfDAO(conn, daoAuth);
        try {
            dbAccess.insertFile(cpfFile.getOdlTree(), CpsStatusEnum.ACTIVE);
        } catch (Exception ex) {
            System.err.println("Unable to load CPF: " + ex.getMessage());
            System.exit(WRITE_ERROR);
        }
        
        System.exit(0);
    }
    
    /**
     * Build the list of available command line options
     * @return Populated options list
     */
    private static Options getCmdOptions() {

        Option baseOption = new Option("f", true, "CPF ODL file");
        baseOption.setRequired(true);
        
        Option debugOption = new Option("d", "debug", false, 
            "(optional) Enable debug messages");
        debugOption.setRequired(false);
        
        Option helpOption = new Option("h", "help", false, 
            "(optional) Display usage message");
        helpOption.setRequired(false);

        Options optionList = new Options();
        optionList.addOption(baseOption);
        optionList.addOption(debugOption);
        optionList.addOption(helpOption);
        
        return optionList;
    }
    
    /**
     * Print usage information for the provided option list
     * @param optionList 
     */
    private static void printUsage(Options optionList) {
        
        String cmdMsg = "java -jar <cpf.db.conn> "
            + "<oracle.net.tns.admin> <classpath> "
            + "datagroup.InsertCpfMain <options>";
        
        // If the caller provided an alternate program name using
        // -Dcalling.script, use that instead of the java usage info
        String caller = System.getProperty("calling.script");
        if (caller != null) {
            cmdMsg = caller + " <options>";
        }
        
        HelpFormatter formatter = new HelpFormatter();
        formatter.printHelp(cmdMsg, "Available options:", optionList, "");
    }   
}
