package datagroup;

import java.net.MalformedURLException;
import java.net.URL;
import library.LoadedFile;
import library.LoadedFileType;
import org.apache.commons.cli.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;

/**
 * Provides a standalone utility to merge the group from one ODL
 * file into another.  The resulting merged ODL is printed to
 * standard output.  This will enable automated processes such as
 * autocal to build updated CPF files without having to run the GUI.
 */
public class MergeGroupMain {

    private static final int USAGE_ERROR = 1;
    private static final int PARSE_ERROR = 2;
    private static final int GROUP_ERROR = 3;
    private static final int WRITE_ERROR = 4;

    // No-op contructor
    private MergeGroupMain() {
    }

    /**
     * The main method
     * @param args   Command line arguments provided by caller.
     * The available options:
     * base_ODL_file  - the name of the original input ODL file.
     * merge_ODL_file - the name of the ODL file containing the group to be
     *                  merged in.
     * group_name     - the name of the group in the merge file to merge into
     *                  the base file.
     * "debug" (optional) - specifying this argument provides debugging log
     *                      messages.
     * "help" (optional) - print usage message
     * "output_ODL_file" (optional) - the of the output file for merged ODL
     */
    public static void main(String[] args) {
        
        Options optionList = getCmdOptions();
        CommandLineParser parser = new GnuParser();
        CommandLine line = null;
        
        String baseOdlFilename;
        String mergeOdlFilename;
        String groupToMerge;
        String outputFilename;
        String fileType;
        
        LoadedFile baseOdlFile = null;
        LoadedFile mergeOdlFile = null;
        
        try {
            line = parser.parse(optionList, args);
        } catch (ParseException ex) {
            printUsage(optionList);   
            System.exit(USAGE_ERROR);
        }

        if (line.hasOption("h") || line.hasOption("help")) {
            printUsage(optionList);
            System.exit(0);
        }

        // Set logging
        Logger logger = Logger.getLogger(
                MergeGroupMain.class.getPackage().getName());
        
        if (line.hasOption("d") || line.hasOption("debug")) {
            logger.setLevel(Level.TRACE);
        } else {
            logger.setLevel(Level.OFF);
        }

        // Get the base file
        baseOdlFilename = line.getOptionValue("b");
        
        // Get the merge file
        mergeOdlFilename = line.getOptionValue("m");
        
        // Get the group to merge
        groupToMerge = line.getOptionValue("g");
        
        // Check for an output file
        outputFilename = null;
        if (line.hasOption("o")) {
            outputFilename = line.getOptionValue("o");
        }
        
        // Check for file type
        fileType = null;
        if (line.hasOption("t")) {
            fileType = line.getOptionValue("t");
            if (!fileType.equals("CPF") && 
                    !fileType.equals("BPF") && 
                    !fileType.equals("OTHER")) {
                System.err.println("Invalid file type");
                System.exit(USAGE_ERROR);
            }
        }

        // Parse the merge ODL file, and ensure the named group exists in it
        URL fileUrl = null;
        try {
            fileUrl = new URL("file", "localhost", mergeOdlFilename);
        } catch (MalformedURLException ex) {
            logger.error("Error getting URL for " + mergeOdlFilename, ex);
        }

        try {
            mergeOdlFile = LoadedFile.create(fileUrl);
        } catch (Exception ex) {
            logger.error("Error parsing " + mergeOdlFilename, ex);
        }
        if (mergeOdlFile == null) {
            System.err.println("Unable to parse merge ODL file " 
                + fileUrl.getFile());
            System.exit(PARSE_ERROR);
        }

        String groups[] = groupToMerge.split(",");
        for (String group : groupToMerge.split(",")) {
            // If ALL is part of the merge list get rid of all other groups
            if (group.equals("ALL")) {
                groupToMerge = "ALL";
                break;
            }

            if (mergeOdlFile.getGroup(group) == null) {
                System.err.println("Merge ODL file " + fileUrl.getFile()
                    + " does not contain group " + group);
                System.exit(GROUP_ERROR);
            }
        }

        // Parse the base ODL file
        try {
            fileUrl = new URL("file", "localhost", baseOdlFilename);
        } catch (MalformedURLException ex) {
            logger.error("Error getting URL for " + baseOdlFilename, ex);
        }
        
        try {
            baseOdlFile = LoadedFile.create(fileUrl);
        } catch (Exception ex) {
            logger.error("Error parsing " + baseOdlFilename, ex);
        }
        if (baseOdlFile == null) {
            System.err.println("Unable to parse base ODL file " 
                + fileUrl.getFile());
            System.exit(PARSE_ERROR);
        }

        // Merge the named group into the base ODL file
        if (groupToMerge.equals("ALL")) {
            logger.debug("Merging all groups from " + mergeOdlFilename + " to " 
                    + baseOdlFilename);
            baseOdlFile.mergeFile(mergeOdlFile);
        } else {
            for (String group : groups) {
                logger.debug("Merging group (" + group + ") from " 
                    + mergeOdlFilename + " to " + baseOdlFilename);
                baseOdlFile.mergeGroup(group, mergeOdlFile.getGroup(group));
            }
        }

        // Determine the file type. Default to CPF.
        if (fileType != null) {
            if (fileType.equals("CPF")) {
                baseOdlFile.setFileType(LoadedFileType.CPF);
            } else if (fileType.equals("BPF")) {
                baseOdlFile.setFileType(LoadedFileType.BPF);
            } else {
                baseOdlFile.setFileType(LoadedFileType.OTHER);
            }
        } else {
            baseOdlFile.setFileType(LoadedFileType.CPF);
        }

        // Finally print out the whole thing
        if (outputFilename == null) {
            System.out.println(baseOdlFile.getOdlTree().toOdlString());
        } else {
            baseOdlFile.updateFileName(outputFilename);
            if (!baseOdlFile.saveFile()) {
                System.err.println("Unable to write merged ODL to file: " 
                    + outputFilename);
                System.exit(WRITE_ERROR);
            }
        }
        
        System.exit(0);
    }
    
    /**
     * Build the list of available command line options
     * @return Populated options list
     */
    private static Options getCmdOptions() {

        Option baseOption = new Option("b", true, "Base ODL file");
        baseOption.setRequired(true);
        
        Option mergeOption = new Option("m", true, "Merge ODL file");
        mergeOption.setRequired(true);
        
        Option typeOption = new Option("t", true, "(optional) File type");
        typeOption.setRequired(false);
        
        Option debugOption = new Option("d", "debug", false, 
            "(optional) Enable debug messages");
        debugOption.setRequired(false);
        
        Option outputOption = new Option("o", true, 
            "(optional) Output filename");
        outputOption.setRequired(false);
        
        Option groupOption = new Option("g", true, 
            "Groups to merge as a comma delimited list or ALL to merge " +
            "all groups from the merge file");
        groupOption.setRequired(true);
        
        Option helpOption = new Option("h", "help", false, 
            "Display usage message");
        helpOption.setRequired(false);

        Options optionList = new Options();
        optionList.addOption(baseOption);
        optionList.addOption(mergeOption);
        optionList.addOption(typeOption);
        optionList.addOption(debugOption);
        optionList.addOption(outputOption);
        optionList.addOption(groupOption);
        optionList.addOption(helpOption);
        
        return optionList;
    }
    
    /**
     * Print usage information for the provided option list
     * @param optionList 
     */
    private static void printUsage(Options optionList) {
        String cmdMsg = "java -jar CpfGUI.jar [-Dcalling.script] <classpath> "
            + "datagroup.MergeGroupMain <options>";
        
        // If the caller provided an alternate program name using
        // -Dcalling.script, use that instead of the java usage info
        String caller = System.getProperty("calling.script");
        if (caller != null) {
            cmdMsg = caller + " <options>";
        }
        
        HelpFormatter formatter = new HelpFormatter();
        formatter.printHelp(cmdMsg, "Available options:", optionList, "");
    }
}
