package datagroup;

import gui.CpfViewer;
import library.LoadedFileSchema;
import gov.usgs.ias.gui.common_library.DateRangeProcessor;
import gov.usgs.ias.gui.common_library.CancelException;
import gov.usgs.ias.constant.Constants;
import gov.nasa.pds.tools.label.*;
import java.util.Calendar;
import java.util.ArrayList;
import javax.swing.JOptionPane;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import org.apache.log4j.Logger;
import org.apache.commons.lang.WordUtils;

/**
 * Gather the OLI Sensor Alignment matrix values from the database
 * and put them into a form usable by the CPF GUI.
 * Implements the DateRangeProcessor interface so that it can operate on input
 * from the DateRangeDialog.
 * @author rockvam
 */
public class SensorAlignmentDataGroup implements DateRangeProcessor {

    private CpfViewer application;
    private static Logger logger = Logger.getLogger(
            SensorAlignmentDataGroup.class.getName());

    /**
     * Create an object to generate the Sensor Alignment data group
     * @param application   the CPF Viewer application
     */
    public SensorAlignmentDataGroup(CpfViewer application) {
        this.application = application;
    }

    /**
     * Constructs a new OLI_to_ACS_Ref_Align parameter based on parameters 
     * retrieved from the database, and updates the ATTITUDE_PARAMETERS group of
     * all selected files in the GUI.
     * @param startDate beginning date in the range
     * @param endDate ending date in the range
     */
    @Override
    public void processWithDateRange(Calendar startDate, Calendar endDate) {

        GroupStatement attitudeGroup = null;
        try {
            attitudeGroup = createGroupStatement(startDate, endDate);
        } catch (CancelException ex) {
            logger.debug("Cancel", ex);
            return;
        } catch (DatagroupException dex ) {
            logger.error("Error retrieving attitude data from database", dex);
            JOptionPane.showMessageDialog(application, 
                WordUtils.wrap(
                    "Error retrieving attitude data from database: "
                    + dex.getMessage(), 80));
        }
        if (attitudeGroup != null) {
            // Update each selected file with new Sensor Alignment parameters
            String groupName = System.getProperty("group.name.sensoralign",
                    "ATTITUDE_PARAMETERS");
            application.parameterViewerPanel.parameterTreePanel.
                mergeGroupToFiles(groupName, attitudeGroup);
        }
    }

    /**
     * Create a group statement with the new sensor alignment matrix, like
     * the following:
     * GROUP = ATTITUDE_PARAMETERS
     * OLI_to_ACS_Ref_Align = (0.999998525,-0.00150454392,0.000828929988,
     *   0.0015036322,0.999998265,0.00109939374,-0.000830582636,-0.00109814572,
     *   0.999999052)
     * END_GROUP = ATTITUDE_PARAMETERS
     * @param startDate beginning date in the range
     * @param endDate ending date in the range
     */
    private GroupStatement createGroupStatement(Calendar startDate,
            Calendar endDate) throws CancelException, DatagroupException {

        try {
            // login to database
            if (!application.databaseLibrary.login(LoadedFileSchema.PROD)) {
                throw new CancelException("Cancel Sensor Alignment");
            }
        } catch (Exception ex) {
            throw new DatagroupException("Error logging in to database: "
                + ex.getMessage());
        }

        ArrayList<Numeric> alignmentMatrix = new ArrayList<Numeric>();
        PreparedStatement stmt = null;
        ResultSet rs = null;
        StringBuilder sql = new StringBuilder();

        // do query with start & end dates:
        sql.append("select "
                + "alignment_matrix_0, alignment_matrix_1, alignment_matrix_2, "
                + "alignment_matrix_3, alignment_matrix_4, alignment_matrix_5, "
                + "alignment_matrix_6, alignment_matrix_7, alignment_matrix_8 "
                + "from sensor_alignment "
                + "where alignment_effective_date_start < ? "
                + "and alignment_effective_date_stop > ? "
                + "order by date_processed DESC");

        try {
            stmt = application.databaseLibrary.getDataConnection().
                    getConnection(Constants.DAO_AUTH_IAS_PROD).
                    prepareStatement(sql.toString());
            stmt.setDate(1, new java.sql.Date(endDate.getTimeInMillis()));
            stmt.setDate(2, new java.sql.Date(startDate.getTimeInMillis()));
            rs = stmt.executeQuery();

            if (rs.next()) {
                alignmentMatrix.add(
                    new Numeric("" + rs.getFloat("ALIGNMENT_MATRIX_0")));
                alignmentMatrix.add(
                    new Numeric("" + rs.getFloat("ALIGNMENT_MATRIX_1")));
                alignmentMatrix.add(
                    new Numeric("" + rs.getFloat("ALIGNMENT_MATRIX_2")));
                alignmentMatrix.add(
                    new Numeric("" + rs.getFloat("ALIGNMENT_MATRIX_3")));
                alignmentMatrix.add(
                    new Numeric("" + rs.getFloat("ALIGNMENT_MATRIX_4")));
                alignmentMatrix.add(
                    new Numeric("" + rs.getFloat("ALIGNMENT_MATRIX_5")));
                alignmentMatrix.add(
                    new Numeric("" + rs.getFloat("ALIGNMENT_MATRIX_6")));
                alignmentMatrix.add(
                    new Numeric("" + rs.getFloat("ALIGNMENT_MATRIX_7")));
                alignmentMatrix.add(
                    new Numeric("" + rs.getFloat("ALIGNMENT_MATRIX_8")));
            }
        } catch (SQLException ex) {
            throw new DatagroupException("Database error: " + ex.getMessage());
        } finally {
            try {
                if (rs != null) {
                    rs.close();
                }
                if (stmt != null) {
                    stmt.close();
                }
            } catch (SQLException ex) {
                logger.warn("Database error: " + ex.getMessage());
            }
            application.databaseLibrary.logout();
        }

        if (alignmentMatrix.size() != 9) {
            throw new DatagroupException(
                "Invalid number of matrix points retrieved -- "
                + alignmentMatrix.size());
        }
        
        // Create an ODL group containing the new ATTITUDE_PARAMETERS parameter
        String groupName = System.getProperty("group.name.sensoralign",
                "ATTITUDE_PARAMETERS");
        GroupStatement attitudeGroup = new GroupStatement(1, groupName);

        logger.debug("Creating ODL group " + groupName);
        Sequence value;
        AttributeStatement as;
        value = new Sequence(alignmentMatrix);
        as = new AttributeStatement(System.getProperty(
                "parameter.name.sensoralign.olimatrix", 
                "ACS_to_OLI_Rotation_Matrix"),
                value);
        attitudeGroup.addStatement(as);

        return attitudeGroup;
    }
}
