package datagroup;

import gui.CpfViewer;
import workers.WorkerException;
import gov.usgs.ias.gui.common_library.DateRangeProcessor;
import gov.nasa.pds.tools.label.*;
import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Calendar;
import java.util.ArrayList;
import java.text.DateFormatSymbols;
import java.util.concurrent.ExecutionException;
import javax.swing.SwingWorker;
import javax.swing.JOptionPane;
import org.apache.log4j.Logger;
import org.apache.commons.lang.WordUtils;

/**
 * Gathers the IERS Earth Orientation Parameters from USNO
 * and puts them into a form usable by the CPF GUI.
 * Implements the DateRangeProcessor class so that it can operate on input from
 * the DateRangeDialog.
 * Extends SwingWorker so that the UT1 update can be done in a background
 * thread because it can take a while to go FTP get the file and process it.
 */
public class Ut1DataGroup extends SwingWorker<GroupStatement, Integer>
        implements DateRangeProcessor {

    private CpfViewer application;       // Used when the new UT1 group is to be
                                         // loaded into files in the GUI
    private static final int RANGE = 90; // days before & after the midpoint
    private Calendar ut1DateRangeStart = null;
    private Calendar ut1DateRangeEnd = null;
    private static Logger logger
            = Logger.getLogger(Ut1DataGroup.class.getName());

    /**
     * Constructor when using this class in conjunction with the GUI
     * @param application  the CPF Viewer application
     */
    public Ut1DataGroup(CpfViewer application) {
        this.application = application;
    }

    /**
     * Constructor when using this class in the standalone utility
     */
    public Ut1DataGroup() {
        this.application = null;
    }

    // SwingWorker methods
    /**
     * Create the group statement in the background thread
     * @return                   the new group statement
     * @throws WorkerException   if the group could not be created from UT1 data
     */
    @Override
    public GroupStatement doInBackground() throws WorkerException {
        GroupStatement ut1Group;
        try {
            ut1Group = createGroupStatement(ut1DateRangeStart, ut1DateRangeEnd);
        } catch (Exception ex) {
            // to help tracing the error
            throw new WorkerException("UT1 thread", ex);
        }
        return ut1Group;
    }

    /**
     * After the background processing is done, GUI updates can now be done
     * on the Event Dispatch Thread
     */
    @Override
    protected void done() {
        firePropertyChange("done", null, null);

        // if there is no GUI to update, nothing more need be done
        if (application == null) {
            return;
        }

        GroupStatement ut1Group = null;
        String errorMessage = "Error retrieving UT1 data";
        try {
            ut1Group = get();
        } catch (InterruptedException ex) {
            errorMessage = "Retrieval of UT1 parameters was interrupted";
        } catch (ExecutionException ex) {
            Throwable rootCause = ex;
            while (rootCause.getCause() != null) {
                rootCause = rootCause.getCause();
            }
            errorMessage = WordUtils.wrap(errorMessage + ": "
                + rootCause.getMessage(), 80);
            logger.error("Retrieval of UT1 parameters received an exception",
                ex);
        }
        if (ut1Group == null) {
            JOptionPane.showMessageDialog(application, errorMessage);
        } else {
            // Update each selected file with the new UT1 parameters
            String groupName = System.getProperty("group.name.ut1", 
                    "UT1_TIME_PARAMETERS");
            application.parameterViewerPanel.parameterTreePanel.
                mergeGroupToFiles(groupName, ut1Group);
        }
        application.statusBar.progress.clearBusy();
    }

    // DateRangeProcessor interface methods
    /**
     * Method called by the dateRangeDialog when Ok is selected -- start the
     * background processing here.
     * Retrieves the records from the file that cover RANGE days centered on
     * the input date range, and updates all selected files in the GUI with the
     * new UT1_TIME_PARAMETERS group.
     * @param startDate  beginning date in the range
     * @param endDate    ending date in the range
     */
    @Override
    public void processWithDateRange(Calendar startDate, Calendar endDate) {
        if (application != null) {
            application.statusBar.progress.setBusy();
        }
        // Save the date range, then launch the background processing
        ut1DateRangeStart = startDate;
        ut1DateRangeEnd = endDate;
        execute(); // fire off the doInBackground() method
    }

    GroupStatement createGroupStatement(Calendar startDate,
        Calendar endDate) throws IOException, MalformedURLException {

        ArrayList<Ut1FinalsRecord> records = null;
        Ut1FileRetriever ut1File = new Ut1FileRetriever();
        logger.info("Retrieving file " + ut1File.getUt1FileName());
        records = ut1File.extractData(startDate, endDate, RANGE);
        logger.debug("Retrieved " + records.size() + " records");

        // Create an ODL group for the UT1 parameters
        String groupName = System.getProperty("group.name.ut1",
            "UT1_TIME_PARAMETERS");
        GroupStatement ut1Group = new GroupStatement(1, groupName);
        DateFormatSymbols dfs = new DateFormatSymbols();
        ArrayList<Numeric> yearList = new ArrayList<Numeric>();
        ArrayList<TextString> monthList = new ArrayList<TextString>();
        ArrayList<Numeric> dayList = new ArrayList<Numeric>();
        ArrayList<Numeric> mjdList = new ArrayList<Numeric>();
        ArrayList<Numeric> xList = new ArrayList<Numeric>();
        ArrayList<Numeric> yList = new ArrayList<Numeric>();
        ArrayList<Numeric> utcList = new ArrayList<Numeric>();
        for (Ut1FinalsRecord rec: records) {
            yearList.add(new Numeric("" + rec.getYear()));
            monthList.add(
                    new TextString(dfs.getShortMonths()[rec.getMonth()-1]));
            dayList.add(new Numeric("" + rec.getDay()));
            mjdList.add(new Numeric("" + rec.getMjd()));
            xList.add(new Numeric("" + rec.getPmX()));
            yList.add(new Numeric("" + rec.getPmY()));
            utcList.add(new Numeric("" + rec.getUt1Utc()));
        }
        logger.debug("Creating ODL group");
        Sequence value;
        AttributeStatement as;
        value = new Sequence(yearList);
        as = new AttributeStatement(
            System.getProperty("parameter.name.ut1.year","UT1_Year"),
            value);
        ut1Group.addStatement(as);

        value = new Sequence(monthList);
        as = new AttributeStatement(
            System.getProperty("parameter.name.ut1.month","UT1_Month"),
            value);
        ut1Group.addStatement(as);

        value = new Sequence(dayList);
        as = new AttributeStatement(
            System.getProperty("parameter.name.ut1.day","UT1_Day"),
            value);
        ut1Group.addStatement(as);

        value = new Sequence(mjdList);
        as = new AttributeStatement(
            System.getProperty("parameter.name.ut1.mjd","UT1_Modified_Julian"),
            value);
        ut1Group.addStatement(as);

        value = new Sequence(xList);
        as = new AttributeStatement(
            System.getProperty("parameter.name.ut1.x","UT1_X"),
            value);
        ut1Group.addStatement(as);

        value = new Sequence(yList);
        as = new AttributeStatement(
            System.getProperty("parameter.name.ut1.y","UT1_Y"),
            value);
        ut1Group.addStatement(as);
        
        value = new Sequence(utcList);
        as = new AttributeStatement(
            System.getProperty("parameter.name.ut1.utc","UT1_UTC"),
            value);
        ut1Group.addStatement(as);

        return ut1Group;
    }
}
