package datagroup;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.util.ArrayList;
import java.util.Calendar;
import org.apache.log4j.Logger;

/**
 * Retrieve the UT1 file from the FTP site and extract the desired
 * information from it.
 */
public class Ut1FileRetriever {

    private static final Logger logger
            = Logger.getLogger(Ut1FileRetriever.class);
    private String ut1FileName;       // URL filename of the UT1 file

    final static int MS_PER_HALF_DAY = 43200000;
    final static int MS_PER_FULL_DAY = MS_PER_HALF_DAY * 2;

    /**
     * Constructor using a default Finals file name
     */
    public Ut1FileRetriever() {
        String ut1FileUrl = System.getProperty("ut1.file.url");

        /* Issue a warning if the UT1 system property is null */
        if (ut1FileUrl == null)
        {
            logger.warn("UT1 system property is null");
        }
        this.ut1FileName = ut1FileUrl;
    }

    /**
     * Constructor using a supplied Finals file name (should be a URL)
     * @param ut1FileName name of the file to retrieve
     */
    public Ut1FileRetriever(String ut1FileName) {
        this.ut1FileName = ut1FileName;
    }

    /**
     * Returns an array of records from the IERS Earth Orientation UT1 file
     * that cover a number of days before and after the given date
     * @param middleDate     the center date about which records are retrieved
     * @param range          the number of days before and after the center
     * @return               an array of FinalsRecord objects
     * @throws java.io.IOException if error retrieving the file
     * @throws java.net.MalformedURLException if an unknown protocol is
     *         specified
     */
    public ArrayList<Ut1FinalsRecord> extractData(Calendar middleDate,
            int range)
            throws IOException, MalformedURLException {

        Calendar startDate = (Calendar) middleDate.clone();
        startDate.add(Calendar.DAY_OF_MONTH, -range);
        Calendar endDate = (Calendar) middleDate.clone();
        endDate.add(Calendar.DAY_OF_MONTH, range);
        logger.info("Range is "
                + startDate.getTime().toString() + " - "
                + endDate.getTime().toString());
        
        // Open a connection to the server
        URL ut1Url = new URL(ut1FileName);
        URLConnection conn = ut1Url.openConnection();

        // Get the file
        ArrayList<Ut1FinalsRecord> recList = new ArrayList<Ut1FinalsRecord>();
        BufferedReader rd = new BufferedReader(new InputStreamReader(
                conn.getInputStream()));
        String line;
        Ut1FinalsRecord fr = null;
        try {
            while ((line = rd.readLine()) != null) {
                try {
                    fr = new Ut1FinalsRecord(line);
                } catch (Exception ex) {
                    logger.warn("Error in UT1 file line " + line, ex);
                    throw new IOException("Error in UT1 file: "
                        + ex.getMessage());
                }
                if ((fr.compareTo(startDate) > 0)
                        && (fr.compareTo(endDate) <= 0)) {
                    logger.debug("Adding " + fr.getCal().getTime().toString()
                            + " to list");
                    recList.add(fr);
                }
            }
        } catch (IOException ioe) {
            throw ioe;
        } finally {
            rd.close();
        }

        // Using the date comparison, we may end up with 1 extra record.  If so,
        // remove the first one.
        while (recList.size() > (2 * range)) {
            recList.remove(0);
        }
        // Make sure we got enough records
        if (recList.size() < (2 * range)) {
            throw new IOException("Unable to get " + (2 * range)
                + " valid records from the UT1 file -- found "
                + recList.size());
        }
        return recList;
    }

    /**
     * Returns an array of records from the IERS Earth Orientation file
     * that cover a number of days surrounding the given date range.  The middle
     * date in the range is calculated, then the range is applied before and 
     * after the middle date.
     * @param startDate   beginning date in the range
     * @param endDate     ending date in the range
     * @param range       the number of days before and after the center
     * @return            an array of FinalsRecord objects
     * @throws java.io.IOException if error retrieving the file
     * @throws java.net.MalformedURLException if an unknown protocol is
     *         specified
     */
    public ArrayList<Ut1FinalsRecord> extractData(Calendar startDate,
            Calendar endDate, int range)
            throws IOException, MalformedURLException {

        // Find the middle date -- half the distance between start & end
        // Calculate a day delta.  Calculate the difference in seconds, then
        // divide by 2 to get the midpoint and the number of msec per day.
        // In case the input range is an odd # days,
        // add 1/2 day to move the mid date.  This will ensure a range where
        // there is one more day after the mid date than before.
        Calendar middleDate = (Calendar) startDate.clone();
        int delta = (int) ((endDate.getTimeInMillis()
                - startDate.getTimeInMillis() + MS_PER_HALF_DAY)
                / (2 * MS_PER_FULL_DAY));
        middleDate.add(Calendar.DAY_OF_YEAR, delta);
        logger.info("Middle date for " + startDate.getTime().toString() + " - "
                + endDate.getTime().toString()
                + " is " + middleDate.getTime().toString());
        return extractData(middleDate, range);
    }

    /**
     * Gets the name of the Finals file being used.
     * @return Finals file name
     */
    public String getUt1FileName() {
        return ut1FileName;
    }
}
