package datagroup;

import java.util.Calendar;
import java.text.DateFormat;

/**
 * A record of information from one line of the UT1 finals file.
 * Selected fields are extracted and put into the CPF.  We only
 * need to extract the first 11 fields from each line.
 * See http://maia.usno.navy.mil/ser7/readme.finals
 * for a layout of the file.
 */
public class Ut1FinalsRecord {

    Calendar cal;           // The date of this record, used for comparisons
    int year;               // 4-digit Year
    int month;              // Month number (1-based)
    int day;                // Day of month
    int mjd;                // fractional Modified Julian Date (MJD),
                            // **integer part only**
    char polarMotionPredictionFlag; // IERS (I) or Prediction (P) flag for
                            // polar motion values
    double pmX;             // Bulletin A PM-x (sec. of arc)
    double pmY;             // Bulletin A PM-y (sec. of arc)
    char ut1UtcPredictionFlag; // IERS/Prediction flag for UT1-UTC values
    double ut1Utc;          // Bulletin A UT1-UTC (sec. of time)

    static final int MINIMUM_RECORD_LENGTH = 68;
    static final int MJD_CUTOFF = 51544; // The cutoff date designated by USNO
                                         // for converting 2-digit years to
                                         // 4-digits

    /**
     * Constructor for a record, using a single line from the
     * finals file to set selected fields that are required for the CPF.
     * @param recordLine   String containing a single line from the finals file
     * @throws Ut1DataFormatException  Error in the format of the UT1 record
     */
    public Ut1FinalsRecord(String recordLine)
        throws Ut1DataFormatException {
        if (recordLine.length() < MINIMUM_RECORD_LENGTH) {
            throw new Ut1DataFormatException("Invalid record length ("
                + recordLine.length()
                + "): expecting at least " + MINIMUM_RECORD_LENGTH
                + " characters");
        }

        // Parse out the fields we need for the CPF.  Parsing into integer
        // or double values will ensure numeric fields are indeed numeric.
        // Additional validation done later.
        try {
            this.year = Integer.parseInt(recordLine.substring(0, 2).trim());
            this.month = Integer.parseInt(recordLine.substring(2, 4).trim());
            this.day = Integer.parseInt(recordLine.substring(4, 6).trim());
            // mjd in the file is floating point but our CPF  wants an integer
            this.mjd = Integer.parseInt(recordLine.substring(7, 12).trim());
            // some records at the end of the file only have the first 4 fields
            this.polarMotionPredictionFlag
                = recordLine.substring(16, 17).charAt(0);
            if (this.polarMotionPredictionFlag == ' ') {
                this.pmX = 0;
                this.pmY = 0;
                this.ut1Utc = 0;
            } else {
                this.pmX = Double.parseDouble(
                    recordLine.substring(18, 27).trim());
                this.pmY = Double.parseDouble(
                    recordLine.substring(37, 46).trim());
                this.ut1Utc = Double.parseDouble(
                    recordLine.substring(58, 68).trim());
            }
        } catch (NumberFormatException nfe) {
            throw new Ut1DataFormatException(
                "Invalid numeric field in UT1 record", nfe);
        }

        // Convert the year field value to true calendar year.  Year in the file
        // is 2 digits which must be converted, per the USNO file specification:
        //   "to get true calendar year, add 1900 for MJD<=51543 or
        //    add 2000 for MJD>=51544."
        // (see http://maia.usno.navy.mil/ser7/readme.finals).
        if (this.mjd < MJD_CUTOFF) {
            this.year += 1900;
        } else {
            this.year += 2000;
        }
        
        // Additional validation
        validate();

        // Set the calendar so we can use it in record comparisons
        cal = Calendar.getInstance();
        cal.clear();
        cal.set(this.year, this.month - 1, this.day);
    }

    /**
     * Constructor for a full record, supplying each field value.
     */
    public Ut1FinalsRecord(int year, int month, int day, int mjd,
            char polarMotionPredictionFlag, double pmX, double pmY,
            char ut1UtcPredictionFlag, double ut1Utc)
            throws Ut1DataFormatException {
        this.year = year;
        this.month = month;
        this.day = day;
        this.mjd = mjd;
        this.polarMotionPredictionFlag = polarMotionPredictionFlag;
        this.pmX = pmX;
        this.pmY = pmY;
        this.ut1UtcPredictionFlag = ut1UtcPredictionFlag;
        this.ut1Utc = ut1Utc;
        
        validate();

        cal = Calendar.getInstance();
        cal.clear();
        cal.set(this.year, this.month - 1, this.day);
    }

    /**
     * Special validation of certain fields from the UT1 file
     * @throws Ut1DataFormatException  for validation error on the fields being
     *                                 checked
     */
    private void validate() throws Ut1DataFormatException {
        // Validate the polar motion prediction flag
        if (this.polarMotionPredictionFlag != 'I'
                && this.polarMotionPredictionFlag != 'P'
                && this.polarMotionPredictionFlag != ' ') {
            throw new Ut1DataFormatException(
                "Invalid polar motion prediction flag: "
                + this.polarMotionPredictionFlag);
        }
        
        // Validate the date fields
        String dateString = String.format("%02d/%02d/%04d", // MM/DD/YYYY
                this.month, this.day, this.year);
        try {
            DateFormat df = DateFormat.getDateInstance(DateFormat.SHORT);
            df.setLenient(false); // must be formatted exactly right
            df.parse(dateString);
        } catch (Exception e) {
            throw new Ut1DataFormatException("Invalid date in record: "
                + dateString);
        }
    }

    public Calendar getCal() {
        return cal;
    }

    public void setCal(Calendar cal) {
        this.cal = cal;
    }

    public int compareTo(Ut1FinalsRecord fr) {
        return cal.compareTo(fr.getCal());
    }

    public int compareTo(Calendar inCal) {
        return cal.compareTo(inCal);
    }
    
    public int getDay() {
        return day;
    }

    public void setDay(int day) {
        this.day = day;
    }

    public int getMjd() {
        return mjd;
    }

    public void setMjd(int mjd) {
        this.mjd = mjd;
    }

    public int getMonth() {
        return month;
    }

    public void setMonth(int month) {
        this.month = month;
    }

    public double getPmX() {
        return pmX;
    }

    public void setPmX(float pmX) {
        this.pmX = pmX;
    }

    public double getPmY() {
        return pmY;
    }

    public void setPmY(float pmY) {
        this.pmY = pmY;
    }

    public char getPolarMotionPredictionFlag() {
        return polarMotionPredictionFlag;
    }

    public void setPolarMotionPredictionFlag(char polarMotionPredictionFlag) {
        this.polarMotionPredictionFlag = polarMotionPredictionFlag;
    }

    public double getUt1Utc() {
        return ut1Utc;
    }

    public void setUt1Utc(float ut1Utc) {
        this.ut1Utc = ut1Utc;
    }

    public int getYear() {
        return year;
    }

    public void setYear(int year) {
        this.year = year;
    }
}
