package datagroup;

import gov.nasa.pds.tools.label.*;
import java.text.ParseException;
import java.util.Calendar;
import java.util.Date;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import java.io.IOException;
import java.net.URL;
import java.io.InputStream;
import java.util.Properties;

/**
 * Provides a standalone utility to gather the IERS Earth
 * Orientation Parameters from USNO and put them into ODL format.
 * This is to enable an automated process of updating UT1
 * parameters in a CPF without requiring bringing up the GUI.
 * An updated UT1_TIME_PARAMETERS group is written to standard output.
 */
public class Ut1Main {

    private static final int USAGE_ERROR = 1;
    private static final int ARGUMENT_ERROR = 2;
    private static final int RETRIEVE_ERROR = 3;

    // no-op constructor
    private Ut1Main() {
    }

    /**
     * Main method.  Outputs an ODL group containing UT1 time parameters for the
     * 180-day period that covers the input date range.
     * @param args   2 or 3 input arguments must be provided. They are:
     * startdate   - start of the date range for which to get values
     * enddate     - end of the date range for which to get values
     * "debug" (optional) - specifying this argument provides debugging log
     *                      messages
     */
    public static void main(String[] args) throws IOException {
        //Load properties from a custom properties file
        try {
            initializeProperties();
        } catch (IOException ex) {
            System.err.println("Unable to initialize the system properties");
            System.exit(USAGE_ERROR);
        }

        if (args.length < 2) {
            System.err.println("Usage: java -jar CpfGUI.jar <classpath> "
                    + "datagroup:Ut1Main startdate enddate [debug]\n");
            System.err.println(" where date format is: yyyy-MM-ddTHH:mm:ss");
            System.exit(USAGE_ERROR);
        }

        // Set logging
        Logger logger = Logger.getLogger(Ut1DataGroup.class.getPackage().
            getName());
        if ((args.length >= 3) && args[2].equals("debug")) {
            logger.setLevel(Level.TRACE);
        } else {
            logger.setLevel(Level.OFF);
        }

        // Parse input date arguments
        Calendar startDate = Calendar.getInstance();
        Calendar endDate = Calendar.getInstance();
        DateFormat df = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
        df.setLenient(true);

        Date inputDate = null;
        try {
            inputDate = df.parse(args[0]);
        } catch (ParseException ex) {
            System.err.println("Error parsing date argument " + args[0]);
            System.exit(ARGUMENT_ERROR);
        }
        startDate.setTime(inputDate);

        try {
            inputDate = df.parse(args[1]);
        } catch (ParseException ex) {
            System.err.println("Error parsing date argument " + args[1]);
            System.exit(ARGUMENT_ERROR);
        }
        endDate.setTime(inputDate);
        
        if (!endDate.after(startDate)) {
            System.err.println("Error: end date must be after start date");
            System.exit(ARGUMENT_ERROR);
        }

        // Create the new UT1 group
        final Ut1DataGroup ut1DataGroup = new Ut1DataGroup();
        Label lab = new Label();
        GroupStatement ut1Group = null;
        try {
            ut1Group = ut1DataGroup.createGroupStatement(startDate, endDate);
        } catch (Exception e) {
            logger.error("Error retrieving UT1 parameters", e);
        }
        int status = 0;
        if (ut1Group == null) {
            System.err.println("Unable to retrieve UT1 parameters");
            status = RETRIEVE_ERROR;
        } else {
            lab.addStatement(ut1Group);
            System.out.println(lab.toOdlString());
        }
        System.exit(status);
    }

    /**
     * Set system properties from the properties file.  Will throw an exception
     * if the properties file can't be found or read.
     */
    private static void initializeProperties() throws IOException {

        ClassLoader cl = Ut1Main.class.getClassLoader();
        URL propURL = cl.getResource(".cpfgui.properties");
        InputStream propStream = propURL.openStream();
        Properties appProperties = new Properties();
        appProperties.load(propStream);
        for (String pName : appProperties.stringPropertyNames()) {
            System.setProperty(pName, appProperties.getProperty(pName));
        }
        propStream.close();
    }
}
