package gui;

import library.LoadedFileList;
import library.DiffStatus;
import library.DatabaseLibrary;
import library.LibraryException;
import workers.CpfLoadThread;
import gov.usgs.ias.gui.common_library.CalendarDialog;
import gov.usgs.ias.gui.common_library.DateRangeDialog;
import gov.usgs.ias.gui.common_library.ObservableDate;
import java.awt.BorderLayout;
import java.awt.Image;
import java.awt.Color;
import java.awt.Toolkit;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import javax.swing.UIManager;
import javax.swing.JFrame;
import javax.swing.JMenuBar;
import javax.swing.Box;
import javax.swing.JOptionPane;
import java.net.URL;
import java.io.InputStream;
import java.io.IOException;
import java.io.File;
import java.util.Properties;

/**
 * The main class for the CPF GUI.
 */
public class CpfViewer extends JFrame implements WindowListener {

    // Widgets that make up the layout of the main GUI screen
    public JMenuBar menuBar;          // the main window menu bar
    public FileMenu fileMenu;         //   file menu
    public ToolMenu toolMenu;         //   tools menu
    public ReportMenu reportMenu;     //   report menu
    public ToolBarPanel toolBarPanel; // toolbar of buttons for commonly-
                                      // performed operations
    public ParameterViewerPanel parameterViewerPanel; // contains the parameter 
                                      // tree and a panel for each loaded file
    public StatusBar statusBar;       // status indicator at the bottom

    // Helper objects
    public ObservableDate updateDisplay;     // used to store previous values
                                            // (dates) for use among multiple
                                            // widgets
    public LoadedFileList loadedFileList;   // the list of files currently 
                                            // loaded in the GUI
    public DatabaseLibrary databaseLibrary; // provides access to DB functions

    // Dialogs used for multiple operations in the application
    public DateRangeDialog dateRangeDialog; // for date range selections
    public CalendarDialog calendarDialog;   // the calendar that pops up when
                                            // the user clicks the calendar
                                            // button on a date panel

    /**
     * Constructor for the main GUI window.
     */
    public CpfViewer() throws IOException, LibraryException {
        //Load properties from a custom properties file
        initializeProperties();

        // Make the Enter key activate buttons (default is the Space key)
        UIManager.put("Button.defaultButtonFollowsFocus", Boolean.TRUE);

        // Set up window
        setTitle("Calibration Parameter GUI");
        this.setSize(600, 800);
        setLayout(new BorderLayout());
        setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);

        Image icon = Toolkit.getDefaultToolkit().getImage(getClass().
            getResource(CpfConfig.CPF_ICON));
        setIconImage(icon);

        //Add a MenuBar to the Frame.
        menuBar = new JMenuBar();
        setJMenuBar(menuBar);

        //Create menus to add to the MenuBar
        fileMenu = new FileMenu(this);
        menuBar.add(fileMenu);
        toolMenu = new ToolMenu(this);
        menuBar.add(toolMenu);
        reportMenu = new ReportMenu(this);
        menuBar.add(reportMenu);
        HelpMenu hm = new HelpMenu();
        menuBar.add(Box.createHorizontalGlue());
        menuBar.add(hm);

        // the remaining areas of the GUI window
        toolBarPanel = new ToolBarPanel(this);
        parameterViewerPanel = new ParameterViewerPanel(this);
        statusBar = new StatusBar();
        add(toolBarPanel, BorderLayout.NORTH);
        add(parameterViewerPanel, BorderLayout.CENTER);
        add(statusBar, BorderLayout.SOUTH);

        setVisible(true);

        // Create dialogs that are used multiple times
        calendarDialog = new CalendarDialog((JFrame)this);
        dateRangeDialog = new DateRangeDialog((JFrame)this, calendarDialog);

        // Create helper objects that perform specific functions
        databaseLibrary = new DatabaseLibrary(this);
        
        loadedFileList = new LoadedFileList();
        loadedFileList.addObserver(parameterViewerPanel.parameterTreePanel);
        loadedFileList.addObserver(toolMenu);

        this.addWindowListener(this);
    }

    /**
     * Main method - create the GUI
     */
    public static void main(String args[]) {
        try {
            // Create the GUI
            final CpfViewer cpfViewer = new CpfViewer();

            // Load files specified on the command line
            for (int i = 0; i < args.length; i++) {
                // Get the full path name of the user-specified file
                File userFile = new File(args[i]);
                File fileToLoad = new File(userFile.getCanonicalPath());
                if (fileToLoad.canRead()) {
                    cpfViewer.statusBar.progress.setBusy();
                    CpfLoadThread loadThread = new CpfLoadThread(cpfViewer,
                        fileToLoad);
                    loadThread.execute();
                } else {
                    JOptionPane.showMessageDialog(cpfViewer,
                        "Error loading " + args[i], "Load Error",
                        JOptionPane.ERROR_MESSAGE);
                }
            }
        } catch (Throwable error) {
            // Catch any exceptions or errors in constructing the main gui class
            // Just in case something's not installed where it should be
            System.err.println("Unable to create GUI.  Please check that all "
                + "required jar and properties files are\ninstalled and "
                + "classpath is correct.");
            System.err.println("Details:");
            error.printStackTrace();
            System.exit(1);
        }
    }

    /**
     * Set system properties from the properties file.  Will throw an exception
     * if the properties file can't be found or read.
     */
    private void initializeProperties() throws IOException {

        ClassLoader cl = CpfViewer.class.getClassLoader();
        URL propURL = cl.getResource(".cpfgui.properties");
        InputStream propStream = propURL.openStream();
        Properties appProperties = new Properties();
        appProperties.load(propStream);
        for (String pName : appProperties.stringPropertyNames()) {
            System.setProperty(pName, appProperties.getProperty(pName));
        }
        propStream.close();
    }

    /**
     * Get the highlighting colors used in the GUI.   is returned.
     * @return    an array of colors, one for each possible DiffStatus
     */
    Color[] getHighlightingColors() {
        Color[] hColors = new Color[DiffStatus.values().length];
        // For each color, try to decode the color property.  If it fails,
        // set to a default color.
        try {
            hColors[DiffStatus.MATCH.ordinal()] = Color.decode(
                System.getProperty("color.match"));
        } catch (Exception e) {
            hColors[DiffStatus.MATCH.ordinal()] = Color.WHITE;
        }
        try {
            hColors[DiffStatus.DIFF.ordinal()] = Color.decode(
                System.getProperty("color.different"));
        } catch (Exception e) {
            // light red
            hColors[DiffStatus.DIFF.ordinal()] = new Color(255, 210, 210);  
        }
        try {
            hColors[DiffStatus.MISSING.ordinal()] = Color.decode(
                System.getProperty("color.missing"));
        } catch (Exception e) {
            // light yellow
            hColors[DiffStatus.MISSING.ordinal()] = new Color(255, 255, 164); 
        }
        try {
            hColors[DiffStatus.NONE.ordinal()] = Color.decode(
                System.getProperty("color.none"));
        } catch (Exception e) {
            hColors[DiffStatus.NONE.ordinal()] = Color.BLACK;
        }
        return hColors;
    }

    // Methods for the WindowListener interface
    /**
     * Before closing the GUI, check for any unsaved changes
     * @param e  Window closing event
     */
    @Override
    public void windowClosing(WindowEvent e) {
        // Check if any files have been modified before closing
        int numberModifiedFiles = 0;
        int reallyExit = JOptionPane.YES_OPTION;
        for (ParameterTablePanel ptp :
            parameterViewerPanel.getDisplayedPanels()) {
            if (ptp.parameterToolBar.isModified()) {
                numberModifiedFiles++;
            }
        }
        if (numberModifiedFiles > 0) {
            reallyExit = JOptionPane.showConfirmDialog(this,
                "There are " + numberModifiedFiles
                + " files with unsaved modifications.  "
                + "Do you want to exit anyway?",
                "Unsaved Modifications",
                JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE);
        }
        if (reallyExit == JOptionPane.YES_OPTION) {
            this.dispose();
            System.exit(0);
        }
    }

    // the remaining are methods required by the interface that don't need
    // to do anything
    @Override
    public void windowOpened(WindowEvent e) {
    }
    @Override
    public void windowClosed(WindowEvent e) {
    }
    @Override
    public void windowIconified(WindowEvent e) {
    }
    @Override
    public void windowDeiconified(WindowEvent e) {
    }
    @Override
    public void windowActivated(WindowEvent e) {
    }
    @Override
    public void windowDeactivated(WindowEvent e) {
    }
}
