package gui;

import library.Services;
import library.LoadedFileType;
import library.LoadedFileSchema;
import actions.CloseAction;
import actions.ExportAction;
import actions.SaveAction;
import actions.SaveAsAction;
import actions.FileNameDialogAction;
import java.awt.Point;
import java.awt.event.KeyEvent;
import java.awt.event.ActionEvent;
import java.awt.event.WindowEvent;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.AbstractAction;
import workers.CpfBaselineThread;
import javax.swing.JOptionPane;
import org.apache.log4j.Logger;
import org.apache.commons.lang.WordUtils;

/**
 * Implements the File Menu Items for the Menu Bar.
 */
public class FileMenu extends JMenu {

    private CpfViewer application;
    public SplitCpfDialog splitCpfDialog = null;
    private static final Logger logger
        = Logger.getLogger(FileMenu.class);


    /**
     * Constructor of the FileMenu Class
     * @param app   the GUI application context
     */
    public FileMenu(CpfViewer app) {
        //Call the parent constructor, setting the Jmenu Naming and KeyEvent.
        super("File");
        setMnemonic(KeyEvent.VK_F);
        this.application = app;

        //Create the Open menu item
        AbstractAction openAction = new AbstractAction("Open") {
            @Override
            public void actionPerformed(ActionEvent e) {
                OpenDialog.open(application);
            }
        };
        JMenuItem openMenuItem = new JMenuItem(openAction);
        openMenuItem.setMnemonic(KeyEvent.VK_O);
        openMenuItem.setToolTipText("Open a file from the file system");
        add(openMenuItem);

        //Create the Import submenu item
        JMenu importSubMenuItem = new JMenu("Import");
        importSubMenuItem.setMnemonic(KeyEvent.VK_I);

        //Create the Baseline submenu item
        JMenu importBaselineSubMenuItem = new JMenu("Baselined");
        importBaselineSubMenuItem.setMnemonic(KeyEvent.VK_B);

        //Create the CPF Menu item
        JMenuItem importBaselineCpfSubMenuItem = new JMenuItem(
            new FileNameDialogAction(application,
                "Import Baseline CPF", "Import",
                LoadedFileType.CPF, LoadedFileSchema.PROD, null));
        importBaselineCpfSubMenuItem.setMnemonic(KeyEvent.VK_C);
        importBaselineCpfSubMenuItem.setToolTipText(
            "Import a baselined CPF from the service");
        importBaselineSubMenuItem.add(importBaselineCpfSubMenuItem);

        //Create the BPF Menu item
        JMenuItem importBaselineBpfSubMenuItem = new JMenuItem(
            new FileNameDialogAction(application,
                "Import Baseline BPF", "Import",
                LoadedFileType.BPF, LoadedFileSchema.PROD, null));
        importBaselineBpfSubMenuItem.setMnemonic(KeyEvent.VK_B);
        importBaselineBpfSubMenuItem.setToolTipText(
            "Import a baselined BPF from the service");
        importBaselineSubMenuItem.add(importBaselineBpfSubMenuItem);

        importSubMenuItem.add(importBaselineSubMenuItem);

        //Create the Evaluation menu item
        JMenu importEvalSubMenuItem = new JMenu("Test");
        importEvalSubMenuItem.setMnemonic(KeyEvent.VK_T);

        //Create the CPF Menu item
        JMenuItem importEvalCpfSubMenuItem = new JMenuItem(
            new FileNameDialogAction(application,
                "Import Test CPF", "Import",
                LoadedFileType.CPF, LoadedFileSchema.EVAL, null));
        importEvalCpfSubMenuItem.setMnemonic(KeyEvent.VK_C);
        importEvalCpfSubMenuItem.setToolTipText(
            "Import an evaluation CPF from the service");
        importEvalSubMenuItem.add(importEvalCpfSubMenuItem);

        //Create the BPF Menu item
        JMenuItem importEvalBpfSubMenuItem = new JMenuItem(
            new FileNameDialogAction(application,
                "Import Test BPF", "Import",
                LoadedFileType.BPF, LoadedFileSchema.EVAL, null));
        importEvalBpfSubMenuItem.setMnemonic(KeyEvent.VK_B);
        importEvalBpfSubMenuItem.setToolTipText(
            "Import an evaluation BPF from the service");
        importEvalSubMenuItem.add(importEvalBpfSubMenuItem);

        importSubMenuItem.add(importEvalSubMenuItem);

        add(importSubMenuItem);

        //Create the Close menu item
        AbstractAction closeAllAction = new AbstractAction("Close") {
            @Override
            public void actionPerformed(ActionEvent e) {
                // execute the action for each selected panel
                for (ParameterTablePanel ptp :
                    application.parameterViewerPanel.getSelectedPanels()) {
                    CloseAction ca = new CloseAction(application, ptp);
                    ca.closeFile();
                }
            }
        };
        JMenuItem closeMenuItem = new JMenuItem(closeAllAction);
        closeMenuItem.setMnemonic(KeyEvent.VK_C);
        closeMenuItem.setToolTipText("Close all open selected files");
        add(closeMenuItem);

        addSeparator();

        //Create the Save menu item (saves all selected files)
        AbstractAction saveAllAction = new AbstractAction("Save") {
            @Override
            public void actionPerformed(ActionEvent e) {
                // execute the action for each selected panel
                for (ParameterTablePanel ptp :
                    application.parameterViewerPanel.getSelectedPanels()) {
                    SaveAction sa = new SaveAction(application, ptp);
                    sa.saveFile();
                }
            }
        };
        JMenuItem saveMenuItem = new JMenuItem(saveAllAction);
        saveMenuItem.setToolTipText(
            "Save all open selected files to file system");
        add(saveMenuItem);

        //Create the SaveAs menu item
        AbstractAction saveAllAsAction = new AbstractAction("Save As") {
            @Override
            public void actionPerformed(ActionEvent e) {
                // execute the action for each selected panel
                for (ParameterTablePanel ptp :
                    application.parameterViewerPanel.getSelectedPanels()) {
                    SaveAsAction saa = new SaveAsAction(application, ptp);
                    saa.saveFileAs();
                }
            }
        };
        JMenuItem saveAsMenuItem = new JMenuItem(saveAllAsAction);
        saveAsMenuItem.setToolTipText( "Save all open selected files to file " +
            "system with new file name(s)");
        add(saveAsMenuItem);

        //Create the Export menu item
        JMenuItem exportMenuItem = new JMenuItem(new ExportAction(application));
        exportMenuItem.setMnemonic(KeyEvent.VK_E);
        add(exportMenuItem);

        addSeparator();

        //Create the Baseline submenu item
        JMenu baselineSubMenuItem = new JMenu("Baseline");
        baselineSubMenuItem.setMnemonic(KeyEvent.VK_B);

        //Create the CPF Menu item
        JMenuItem baselineCpfSubMenuItem = new JMenuItem(
            new FileNameDialogAction(application,
                "Select Test CPF to Baseline", "Baseline",
                LoadedFileType.CPF, LoadedFileSchema.EVAL, null));
        baselineCpfSubMenuItem.setMnemonic(KeyEvent.VK_C);
        baselineCpfSubMenuItem.setToolTipText(
            "Move an evaluation CPF into the production database");
        baselineSubMenuItem.add(baselineCpfSubMenuItem);

        //Create the BPF Menu item
        JMenuItem baselineBpfSubMenuItem = new JMenuItem(
            new FileNameDialogAction(application,
            "Select Test BPF to Baseline", "Baseline",
            LoadedFileType.BPF, LoadedFileSchema.EVAL, null));
        baselineBpfSubMenuItem.setMnemonic(KeyEvent.VK_B);
        baselineBpfSubMenuItem.setToolTipText(
            "Move an evaluation BPF into the production database");
        baselineSubMenuItem.add(baselineBpfSubMenuItem);

        add(baselineSubMenuItem);

        //Create the Split submenu item
        JMenu splitSubMenuItem = new JMenu("Split CPFs");
        splitSubMenuItem.setMnemonic(KeyEvent.VK_P);

        // Create the Split dialog which is used by both generate and baseline
        splitCpfDialog = new SplitCpfDialog(application);

        //Create the Split CPF Menu item
        AbstractAction splitAction = new AbstractAction("Generate Splits") {
            @Override
            public void actionPerformed(ActionEvent e) {
                Point loc = application.getLocationOnScreen();
                loc.translate(50, 50);
                splitCpfDialog.setLocation(loc);
                splitCpfDialog.setVisible(true);
            }
        };
        JMenuItem splitCpfSubMenuItem = new JMenuItem(splitAction);
        splitCpfSubMenuItem.setMnemonic(KeyEvent.VK_G);
        splitCpfSubMenuItem.setToolTipText("Generate 2 evaluation CPFs " +
            "from a split of an existing baseline CPF");
        splitSubMenuItem.add(splitCpfSubMenuItem);

        AbstractAction blSplitAction = new AbstractAction("Baseline") {
            @Override
            public void actionPerformed(ActionEvent e) {
                doBaselineSplits();
            }
        };
        JMenuItem blSplitCpfSubMenuItem = new JMenuItem(blSplitAction);
        blSplitCpfSubMenuItem.setMnemonic(KeyEvent.VK_B);
        blSplitCpfSubMenuItem.setToolTipText("Baseline the 2 evaluation CPFs " +
            "generated by the previous split");
        splitSubMenuItem.add(blSplitCpfSubMenuItem);

        add(splitSubMenuItem);

        addSeparator();

        //Create the Exit menu item
        AbstractAction exitAction = new AbstractAction("Exit") {
            @Override
            public void actionPerformed(ActionEvent e) {
                // exit by issuing a window close event
                application.dispatchEvent(new WindowEvent(application,
                    WindowEvent.WINDOW_CLOSING));
            }
        };
        JMenuItem exitMenuItem = new JMenuItem(exitAction);
        exitMenuItem.setMnemonic(KeyEvent.VK_X);
        exitMenuItem.setToolTipText("Exit the GUI");
        add(exitMenuItem);
    }

    // Perform the baselining of previously split CPFs
    private void doBaselineSplits() {
        // Baseline the previously split files.  The user also could
        // use the File->Baseline option.  This action operates just on the
        // pre- and post- split file names
        // so it needs to verify the files have been exported to EVAL first
        if (splitCpfDialog == null
            || splitCpfDialog.getPreSplitFileName() == null
            || splitCpfDialog.getPostSplitFileName() == null) {
            JOptionPane.showMessageDialog(application,
                "Split CPFs have not been generated yet",
                "Error", JOptionPane.ERROR_MESSAGE);
            return;
        }

        // Ensure the 2 split files have been exported to the EVAL db
        String[] filesToBaseline = {splitCpfDialog.getPreSplitFileName(),
            splitCpfDialog.getPostSplitFileName()};
        Services services = new Services();
        Object[][] existingEvalFiles = null;
        try {
            existingEvalFiles = services.getCpfFileNameList(
                    LoadedFileSchema.EVAL, false, null, null, null);
        } catch (Exception ex) {
            JOptionPane.showMessageDialog(application,
                WordUtils.wrap("Unable to get the list of EVAL files: "
                + ex.getMessage(), 80),
                "Service Error", JOptionPane.ERROR_MESSAGE);
            return;
        }
        boolean[] found = {false, false};
        for (int i = 0; i < existingEvalFiles.length; i++) {
            String existingEvalFile = existingEvalFiles[i][0].toString();
            for (int j = 0; j < filesToBaseline.length; j++) {
                if (existingEvalFile.equals(filesToBaseline[j])) {
                    found[j] = true;
                }
            }
        }
        for (int j = 0; j < filesToBaseline.length; j++) {
            if (!found[j]) {
                JOptionPane.showMessageDialog(application,
                    filesToBaseline[j]
                    + " has not been exported to the EVAL database",
                    "Error", JOptionPane.ERROR_MESSAGE);
                return;
            }
        }

        // Make sure the files being baselined are the correct ones.
        int confirm = JOptionPane.YES_OPTION;
        confirm = JOptionPane.showConfirmDialog(this,
            "Baselining split files " + filesToBaseline[0]
            + " and  " + filesToBaseline[1]
            + ".  Are these correct?",
            "Split Files",
            JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
        if (confirm == JOptionPane.NO_OPTION) {
            return;
        }

        // Now, baseline them both
        // Baselining is done in a thread, but there must not be any other
        // background threads running. Since the logic in this method to
        // check the busy state and later set it is done on the Event
        // Dispatch Thread, there should be no possible race condition here.
        if (application.statusBar.progress.isBusy()) {
            JOptionPane.showMessageDialog(application,
                "Cannot baseline while the GUI is busy.  "
                + "Please try later.",
                "Busy", JOptionPane.ERROR_MESSAGE);
            return;
        }
        try {
            if (!application.databaseLibrary.login(LoadedFileSchema.PROD)) {
                return; // user cancelled
            }
        } catch (Exception ex) {
            logger.error("Error connecting to PROD database", ex);
            JOptionPane.showMessageDialog(application,
                "Error connecting to PROD database: " + ex.getMessage(),
                "Database Error", JOptionPane.ERROR_MESSAGE);
            return;
        }
        application.statusBar.progress.setBusy();
        logger.info("Baselining " + filesToBaseline[0] + " and "
            + filesToBaseline[1]);
        CpfBaselineThread baselineThread = new CpfBaselineThread(application,
            filesToBaseline, true);
        baselineThread.execute();
    }
}
