package gui;

import library.FileNamingConventions;
import library.Services;
import library.CpfStatus;
import library.BpfSensor;
import library.LoadedFileType;
import library.LoadedFileSchema;
import gov.usgs.ias.gui.common_library.DateRangeProcessor;
import workers.CpfImportThread;
import workers.CpfBaselineThread;
import workers.CpfUpdateGroupThread;
import gov.usgs.ias.cps.CpsStatusEnum;
import gov.nasa.pds.tools.label.*;
import java.util.ArrayList;
import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.JOptionPane;
import java.awt.Cursor;
import java.util.Calendar;
import org.apache.log4j.Logger;
import org.apache.commons.lang.WordUtils;

/**
 * Dialog to display names of files returned from the service.
 */
public class FileNameDialog extends JDialog implements WindowListener,
        ActionListener, DateRangeProcessor {

    private CpfViewer application;
    private FileNameFilterPanel fileNameFilterPanel;
    private FileNameListPanel fileNameListPanel;
    private JButton okButton;
    private Calendar startDate = null;  // start & end date to use in selecting
    private Calendar endDate = null;    //   files to display
    private CpfStatus status = null;    // status of files to show, default all
    private boolean latestOnly = false; // show latest files only, default no
    private Object userData = null;     // a place to store data specific to
                                        // the operation being performed

    private static final Logger logger = Logger.getLogger(FileNameDialog.class);

    /**
     * Create the file name dialog showing files of a certain type
     * @param application the overall application to which this dialog belongs
     * @param title title of this dialog
     * @param action the action to take -- Import or Baseline
     * @param type type of files to display in the dialog's list
     * @param schema schema source of files to display
     */
    public FileNameDialog(CpfViewer application, String title, String action,
            LoadedFileType type, LoadedFileSchema schema) {
        super(application, title, true);
        this.application = application;
        
        //Make the main panel
        constructPanel(action, type, schema);
    }

    /**
     * Create the file name dialog showing files of a certain type
     * @param application the overall application to which this dialog belongs
     * @param title       title of this dialog
     * @param action      the action to take -- Import or Baseline
     * @param type        type of files to display in the dialog's list
     * @param schema      schema source of files to display
     * @param status      the status of files to display
     * @param latestOnly  flag indicating to display only the latest version
     */
    public FileNameDialog(CpfViewer application, String title, String action,
            LoadedFileType type, LoadedFileSchema schema, CpfStatus status,
            boolean latestOnly) {
        super(application, title, true);
        this.application = application;
         //Set this before building so filter panel built correctly
        this.status = status;
        this.latestOnly = latestOnly;
        
        //Make the main panel
        constructPanel(action, type, schema);
    }
    
    /** Construct the main panel of the dialog.
     * @param dialog    the dialog building the panel within
     * @param action    the action to take -- Import or Baseline
     * @param type      type of files to display in the dialog's list
     * @param schema    schema source of files to display
     */
    private void constructPanel(String action, LoadedFileType type,
        LoadedFileSchema schema) {
        setLayout(new BorderLayout());        

        JPanel mainPanel = new JPanel(new BorderLayout());

        fileNameFilterPanel = new FileNameFilterPanel(application, this, type,
                status);
        mainPanel.add(fileNameFilterPanel, BorderLayout.NORTH);
        
        fileNameListPanel = new FileNameListPanel(type, schema, true);
        mainPanel.add(fileNameListPanel, BorderLayout.CENTER);
        
        JPanel buttonPanel = new JPanel();

        okButton = new JButton("Ok");
        okButton.setActionCommand(action);
        okButton.setMnemonic(KeyEvent.VK_O);
        okButton.setToolTipText(action);
        buttonPanel.add(okButton);

        JButton closeButton = new JButton("Cancel");
        closeButton.setMnemonic(KeyEvent.VK_C);
        buttonPanel.add(closeButton);
        
        mainPanel.add(buttonPanel, BorderLayout.SOUTH);
        add(mainPanel, BorderLayout.CENTER);
        
        pack();

        okButton.addActionListener(this);
        closeButton.addActionListener(this);
        addWindowListener(this);
    }
    
    /**
     * Allows the user to store custom data in the dialog for use by
     * actions that need it
     * @param userData   object data that needs to be made available to an 
     *                   action
     */
    public void setUserData(Object userData) {
        this.userData = userData;
    }
    
    /** Query the service for file list with given filters. This is the
     * communication path between the filter panel and the file name list
     * panel to refresh the file list based on the defined filters.
     * @param filterStartDate   start date for filter
     * @param filterEndDate     end date for filter
     * @param filterSensor      filter BPF files on given sensor
     * @param filterStatus      filter files on given status
     * @param filterLatestOnly  display only the latest version of each file
     */
    public void filterFileList(Calendar filterStartDate, Calendar filterEndDate,
            BpfSensor filterSensor, CpfStatus filterStatus,
            boolean filterLatestOnly) {
        // Make the cursor busy while we contact the service
        disableButtons();
        this.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

        // Load the list of available CPF/BPF files matching given filters
        fileNameListPanel.setFileNameList(filterStartDate, filterEndDate,
                filterSensor, filterStatus, filterLatestOnly);
        if (fileNameListPanel.getFileNames().length > 0) {
            enableButtons();
        }
        this.setCursor(null);
    }
    
    /** Build the file list with no filters.
     */
    public void refreshFileList() {
        // Make the cursor busy while we contact the service
        disableButtons();
        this.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

        // Load the list of all available files
        if (startDate != null & endDate != null) {
            //This case is for when the DateRangeDialog was used to
            //specify a date range for the Update case
            fileNameListPanel.setFileNameList(startDate, endDate,
                    status, latestOnly);
        } else {
            //This case is for all CPF/BPF file listings and can have filters
            fileNameListPanel.setFileNameList(null, null, null, status,
                    latestOnly);
        }
        if (fileNameListPanel.getFileNames().length > 0) {
            enableButtons();
        }
        this.setCursor(null);
    }
    
    // WindowListener interface methods
    /**
     * Load the list of file names when the window opens
     * @param e   Window open event
     */
    @Override
    public void windowOpened(WindowEvent e) {
        refreshFileList();
    }

    // required window event handlers for events that do not need handling
    @Override
    public void windowClosing(WindowEvent e) {
    }
    @Override
    public void windowActivated(WindowEvent e) {
    }
    @Override
    public void windowClosed(WindowEvent e) {
    }
    @Override
    public void windowDeactivated(WindowEvent e) {
    }
    @Override
    public void windowDeiconified(WindowEvent e) {
    }
    @Override
    public void windowIconified(WindowEvent e) {
    }

    // helper methods to disable/enable buttons that depend on data in the list
    private void enableButtons() {
        okButton.setEnabled(true);
    }
    private void disableButtons() {
        okButton.setEnabled(false);
    }

    // ActionListener interface methods
    /**
     * Perform the action specified by the user
     * @param e   the user event (button selection)
     */
    @Override
    public void actionPerformed(ActionEvent e) {
        
        String command = e.getActionCommand();

        if (command.equals("Cancel")
            || fileNameListPanel.getSelectedFileNames().length == 0) {
            // no specific action, just continue on to close the dialog
            ;
        } else if (command.equals("Import")) {
            for (String selectedFileName :
                fileNameListPanel.getSelectedFileNames()) {
                if (selectedFileName == null) {
                    continue;
                }
                logger.info("Starting thread to open file " + selectedFileName);
                application.statusBar.progress.setBusy();
                CpfImportThread importThread = new CpfImportThread(application,
                    selectedFileName);
                importThread.execute();
            }

        } else if (command.equals("Baseline")) {

            // Baselining is done in a thread, but there must not be any other
            // background threads running. Since the logic in this method to
            // check the busy state and later set it is done on the Event
            // Dispatch Thread, there should be no possible race condition here.
            if (application.statusBar.progress.isBusy()) {
                JOptionPane.showMessageDialog(application,
                    "Cannot baseline while the GUI is busy.  " +
                    "Please try later.",
                    "Busy", JOptionPane.ERROR_MESSAGE);
                setVisible(false);
                return;
            }

            // First, make an array of the selected files and make sure all
            // files are EVAL files
            ArrayList<String> filesToBaseline = new ArrayList<String>();
            for (String selectedFileName :
                fileNameListPanel.getSelectedFileNames()) {
                // Verify this was an eval file
                if (selectedFileName == null) {
                    continue;
                }
                if (!FileNamingConventions.getSchema(selectedFileName).equals(
                        LoadedFileSchema.EVAL)) {
                    JOptionPane.showMessageDialog(application, selectedFileName
                        + " is not a valid evaluation file",
                        "Invalid File", JOptionPane.ERROR_MESSAGE);
                    continue;
                }
                filesToBaseline.add(selectedFileName);
            }
            if (!filesToBaseline.isEmpty()) {
                try {
                    if (!application.databaseLibrary.login(
                        LoadedFileSchema.PROD)) {
                        setVisible(false);  // user cancelled
                        return;
                    }
                } catch (Exception ex) {
                    logger.error("Error connecting to PROD database", ex);
                    JOptionPane.showMessageDialog(application,
                        "Error connecting to PROD database: " 
                        + ex.getMessage(),
                        "Database Error", JOptionPane.ERROR_MESSAGE);
                    setVisible(false);
                    return;
                }
                application.statusBar.progress.setBusy();
                CpfBaselineThread baselineThread
                    = new CpfBaselineThread(application,
                    filesToBaseline.toArray(new String[0]));
                baselineThread.execute();
            }

        } else if (command.equals("Terminate")
                || command.equals("Activate")) {

            CpsStatusEnum newStatus
                    = command.equals("Activate") ? CpsStatusEnum.ACTIVE :
                                                   CpsStatusEnum.INACTIVE;
            try {
                if (!application.databaseLibrary.login(LoadedFileSchema.PROD)) {
                    return;  // cancelled
                }
                for (String selectedFileName :
                     fileNameListPanel.getSelectedFileNames()) {
                    if (selectedFileName == null) {
                        continue;
                    }
                    application.databaseLibrary.updateFileStatus(
                            selectedFileName,
                            FileNamingConventions.getFileType(selectedFileName),
                            newStatus);
                }
            } catch (Exception ex) {
                logger.error(command + " Error", ex);
                JOptionPane.showMessageDialog(application, ex.getMessage(),
                    command + " Error", JOptionPane.ERROR_MESSAGE);
            } finally {
                application.databaseLibrary.logout();
            }

        } else if (command.equals("Update")) {

            // Get the latest version of each file in the EVAL schema
            Services services = new Services();
            Object[][] latestEvalFiles = null;
            try {
                latestEvalFiles = services.getCpfFileNameList(
                        LoadedFileSchema.EVAL, true, null, null, null);
            } catch (Exception ex) {
                logger.error("Unable to get list of EVAL CPF files", ex);
                JOptionPane.showMessageDialog(application,
                    WordUtils.wrap(
                    "Unable to get the list of EVAL files: "
                    + ex.getMessage(), 80),
                    "Service Error", JOptionPane.ERROR_MESSAGE);
                setVisible(false);
                return;
            }
            GroupStatement mergeGroup = (GroupStatement)userData;

            for (String updateFileName :
                fileNameListPanel.getSelectedFileNames()) {

                // The file being updated is a production file.  Rename it as
                // an eval file (and figure out what the version number should
                // be).
                int newVersion = 1;
                String evalBaseName;    // Base name of existing eval files
                String updateBaseName;  // Base name of the file being updated
                updateBaseName = updateFileName.split("\\.")[0];
                if (latestEvalFiles != null && latestEvalFiles.length > 0) {
                    // See if any of the known eval files match our file.
                    for (int ii = 0; ii < latestEvalFiles.length; ii++) {
                        evalBaseName = ((String)latestEvalFiles[ii][0]).
                            split("\\.")[0];
                        if (evalBaseName.equals(updateBaseName)) {
                            newVersion = FileNamingConventions.
                                    getVersion((String)latestEvalFiles[ii][0]);
                            newVersion++;
                        }
                    }
               } 

                // Rename the file as eval and update the version
                String origFileName = updateFileName;
                String newFileName = FileNamingConventions.constructFileName(
                    LoadedFileSchema.EVAL,
                    FileNamingConventions.getSatellite(origFileName),
                    FileNamingConventions.getSensor(origFileName),
                    FileNamingConventions.getFileType(origFileName),
                    FileNamingConventions.getStartDate(origFileName),
                    FileNamingConventions.getEndDate(origFileName),
                    newVersion);

                application.statusBar.progress.setBusy();
                CpfUpdateGroupThread ugThread = new CpfUpdateGroupThread(
                        application, updateFileName, newFileName, mergeGroup);
                ugThread.execute();
            }                
            
        } else {           
            JOptionPane.showMessageDialog(application,
                    "Unknown command: " + command,
                    "Unknown", JOptionPane.ERROR_MESSAGE);
        }
        setVisible(false);
        dispose();
    }

    // DateRangeProcessor interface method
    /**
     * Used for activate & terminate a date range.  Sets the start & end
     * date and then makes the window visible, triggering the WindowOpened
     * event.
     * @param startDate   start date provided from the date range dialog
     * @param endDate     end date provided from the date range dialog
     */
    @Override
    public void processWithDateRange(Calendar startDate, Calendar endDate) {
        logger.trace("FileNameDialog processWithDateRange with dates "
            + startDate.getTime().toString() + " - "
            + endDate.getTime().toString());
        this.startDate = startDate;
        this.endDate = endDate; 
        setVisible(true);
    }
}
