package gui;

import gov.usgs.ias.gui.common_library.CalendarDialog;
import gov.usgs.ias.gui.common_library.DateRangePanel;
import gov.usgs.ias.gui.common_library.FilterItemPanel;
import java.awt.BorderLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Calendar;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import library.BpfSensor;
import library.CpfStatus;
import library.LoadedFileType;

/**
 * Panel to allow filtering CPF and BPF file names based on user selectable
 * criteria.
 */
public class FileNameFilterPanel extends JPanel implements ActionListener,
        PropertyChangeListener {
    private FileNameDialog parent;
    private LoadedFileType fileType;
    private FilterItemPanel sensorFilterPanel;
    private FilterItemPanel stateFilterPanel;
    private FilterItemPanel versionFilterPanel;
    private CpfStatus staticState = null;
    private DateRangePanel dateRangePanel;
    private Calendar initStartDate;
    private JButton queryButton;
    
    /**
     * Constructor of the File Name Filter Panel
     * @param application the overall application to which this dialog belongs
     * @param parent      the parent dialog this filter panel belongs to
     * @param type        type of files to display in the dialog's list
     * @param status      status given to display files
     */
    public FileNameFilterPanel(CpfViewer application, FileNameDialog parent,
            LoadedFileType type, CpfStatus status) {
        super(new BorderLayout());
        this.parent = parent;
        this.fileType = type;
        this.staticState = status;

        //Create the main panel to display all of the filter components
        JPanel mainPanel = new JPanel(new GridBagLayout());
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gbc.fill = GridBagConstraints.HORIZONTAL;
        
        //Create a label for this filter panel to guide user usage
        JPanel labelPanel = new JPanel();
        JLabel label = new JLabel("Use Filters to Refine File List");
        labelPanel.add(label);
        mainPanel.add(labelPanel, gbc);
        
        //Create a panel to contain the sensor, state, version filter panels
        JPanel filterOptionsPanel = new JPanel(new GridBagLayout());
        GridBagConstraints gbcItemFilters = new GridBagConstraints();
        gbcItemFilters.fill = GridBagConstraints.HORIZONTAL;
        gbcItemFilters.gridx = 0;
        gbcItemFilters.gridy = 0;
        gbcItemFilters.weightx = 0.5;
        gbcItemFilters.insets = new Insets(0,0,5,0);

        // Create a sensor selection that defaults to "Any" selected.  Only
        // make it visible if it is really needed.
        String[] sensorList = {"Any", "OLI", "TIRS"};
        sensorFilterPanel = new FilterItemPanel("Sensor", sensorList);
        if (type.equals(LoadedFileType.BPF)) {
            filterOptionsPanel.add(sensorFilterPanel, gbcItemFilters);
            sensorFilterPanel.addPropertyChangeListener("filterModified", this);
        }

        gbcItemFilters.gridx = 1;
        String[] versionList = {"Any", "Latest Only"};
        versionFilterPanel = new FilterItemPanel("Version", versionList);
        filterOptionsPanel.add(versionFilterPanel, gbcItemFilters);
        versionFilterPanel.addPropertyChangeListener("filterModified", this);
        
        gbcItemFilters.gridx = 2;
        String[] stateList = {"Any", "Active", "Inactive"};
        stateFilterPanel = new FilterItemPanel("State", stateList);
        if (staticState != null) {
            //When given a static status value to work with (when
            //terminating/activating files), create the filter panel but don't
            //display it. This simplifies later code to retrieve the value.
            stateFilterPanel.setValue(convertStatusValue(staticState));
        } else {
            //Only show the panel if the value can be changed by the user
            stateFilterPanel.addPropertyChangeListener("filterModified", this);
            filterOptionsPanel.add(stateFilterPanel, gbcItemFilters);
        }
        
        mainPanel.add(filterOptionsPanel, gbc);
        
        //Add a date filter using a date range panel
        dateRangePanel = new DateRangePanel(application,
            application.calendarDialog);
        initStartDate = dateRangePanel.getStartDate();
        Calendar today = Calendar.getInstance();
        dateRangePanel.setEndDate(today.get(Calendar.YEAR),
                today.get(Calendar.MONTH), today.get(Calendar.DAY_OF_MONTH),
                23, 59, 59);
        dateRangePanel.addPropertyChangeListener("dateModified", this);
        
        mainPanel.add(dateRangePanel, gbc);

        //Create a button panel
        JPanel buttonPanel = new JPanel();

        queryButton = new JButton("Query");
        queryButton.setActionCommand("Query");
        queryButton.setMnemonic(KeyEvent.VK_Q);
        queryButton.setToolTipText("Query file service using defined filters");
        buttonPanel.add(queryButton);

        JButton resetButton = new JButton("Reset");
        resetButton.setActionCommand("Reset");
        resetButton.setMnemonic(KeyEvent.VK_R);
        resetButton.setToolTipText("Reset file filters");
        buttonPanel.add(resetButton);
        
        mainPanel.add(buttonPanel, gbc);
        
        //Add the main panel to the dialog window
        add(mainPanel, BorderLayout.CENTER);
        
        queryButton.addActionListener(this);
        resetButton.addActionListener(this);
    }
    
    //Helper methods to disable/enable buttons that depend on data in the list
    private void enableQueryButton() {
        queryButton.setEnabled(true);
    }
    private void disableQueryButton() {
        queryButton.setEnabled(false);
    }
    
    /** Helper method to convert the sensor string to BpfSensor enum value.
     * @param filterValue   the string sensor value to convert
     * @return the enumerated sensor type
     */
    private BpfSensor convertSensorValue(String filterValue) {
        if (filterValue.equals("OLI")) {
                return BpfSensor.OLI;
        } else if (filterValue.equals("TIRS")) {
            return BpfSensor.TIRS;
        } else {
            return BpfSensor.ANY;
        }
    }
    
    /** Helper method to convert the version string to boolean.
     * @param filterValue   the string version value to convert
     * @return the version boolean (true = "Latest Only") status type
     */
    private boolean convertVersionValue(String filterValue){
        if (filterValue.equals("Latest Only")) {
            return true;
        } else {
            return false;
        }
    }
    
    /** Helper method to convert the status string to CpfStatus enum value.
     * @param filterValue   the string status value to convert
     * @return the enumerated status type
     */
    private CpfStatus convertStatusValue(String filterValue){
        if (filterValue.equals("Active")) {
            return CpfStatus.ACTIVE;
        } else if (filterValue.equals("Inactive")) {
            return CpfStatus.NOT_ACTIVE;
        } else {
            return CpfStatus.ANY;
        }
    }

    /** Helper method to convert the status CpfStatus enum value to a string.
     * @param status   the status enum value to convert
     * @return the string status type
     */
    private String convertStatusValue(CpfStatus status){
        if (status == CpfStatus.ACTIVE) {
            return "Active";
        } else if (status == CpfStatus.NOT_ACTIVE) {
            return "Inactive";
        } else {
            return "Any";
        }
    }
    
    // ActionListener interface methods
    /**
     * Perform the action specified by the user
     * @param e   the user event (button selection)
     */
    @Override
    public void actionPerformed(ActionEvent e) {
        String command = e.getActionCommand();
        
        if (command.equals("Query")) {
            BpfSensor sensor = convertSensorValue(sensorFilterPanel.getValue());
            boolean latestOnly
                    = convertVersionValue(versionFilterPanel.getValue());
            CpfStatus status = convertStatusValue(stateFilterPanel.getValue());
            Calendar startDate = dateRangePanel.getStartDate();
            Calendar endDate = dateRangePanel.getEndDate();
            
            //Update the file list with the given filter values
            parent.filterFileList(startDate, endDate, sensor, status,
                    latestOnly);            
            this.disableQueryButton();
        } else if (command.equals("Reset")) {
            if (fileType.equals(LoadedFileType.BPF)) {
                sensorFilterPanel.resetValue();
            }
            versionFilterPanel.resetValue();
            if (staticState == null) {
                stateFilterPanel.resetValue();
            }
            Calendar today = Calendar.getInstance();
            dateRangePanel.setStartDate(initStartDate.get(Calendar.YEAR),
                    initStartDate.get(Calendar.MONTH),
                    initStartDate.get(Calendar.DAY_OF_MONTH), 0, 0, 0);
            dateRangePanel.setEndDate(today.get(Calendar.YEAR),
                    today.get(Calendar.MONTH), today.get(Calendar.DAY_OF_MONTH),
                    23, 59, 59);
            parent.refreshFileList();
            this.enableQueryButton();
        }
    }

    /**
     * Handle when a registered property change occurs.
     * @param event  property change event fired
     */
    @Override
    public void propertyChange(PropertyChangeEvent event) {
        this.enableQueryButton();
    }
}
