package gui;

import library.Services;
import library.CpfStatus;
import library.BpfSensor;
import library.LoadedFileType;
import library.LoadedFileSchema;
import library.FileNamingConventions;
import java.util.Calendar;
import java.util.Comparator;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.awt.BorderLayout;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JOptionPane;
import javax.swing.ListSelectionModel;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableRowSorter;
import org.apache.log4j.Logger;
import org.apache.commons.lang.WordUtils;

/**
 * Panel displaying names of files available from the service.  The panel
 * contains a table with name & status of each file.  The type (CPF/BPF)
 * and schema are defined when the FileNameListPanel is created; the
 * particulars of which files to display (date range, status, latest only) are
 * defined when the panel is populated.
 */        
public class FileNameListPanel extends JPanel
{
    private LoadedFileType fileType;  // type of files
    private LoadedFileSchema schema;  // schema the files are from
    private Services services = null; // to interface with the services
    private JTable table;             // widget displaying file info
    private TableModel tableModel;    // model of data displayed in the table
    private static Logger logger = Logger.getLogger(
            FileNameListPanel.class.getName());

    // Maintain the sort state for the CPF/BPF column
    private enum SortState {
        FILENAME_ASCENDING,
        FILENAME_DESCENDING,
        DATE_ASCENDING,
        DATE_DESCENDING;

        // Get the next SortState, looping back to the first when at the end
        public SortState getNext() {
            return this.ordinal() < SortState.values().length - 1
                ? SortState.values()[this.ordinal() + 1]
                : FILENAME_ASCENDING;
        }
    }

    /**
     * Class used to coordinate the display of data in the table.  Uses the
     * AbstractTableModel's default implementations for most of the TableModel
     * interface methods, but overrides those that are specific to the data
     * we are displaying in the table, and adds some custom methods used
     * by the panel.
     */
    class TableModel extends AbstractTableModel {

        private Object[][] data;  // 2 columns of data, file  name and status
        private String[] columnNames = {"File Name", "Status"};

        // Return the name of the requested column
        @Override
        public String getColumnName(int col) {
            return columnNames[col];
        }

        // Return the number of rows in the table
        @Override
        public int getRowCount() {
            if (data != null) {
                return data.length;
            } else {
                return 0;
            }
        }

        // Return the number of columns in the table
        @Override
        public int getColumnCount() {
            return columnNames.length;
        }

        // Return the value in a table cell
        @Override
        public Object getValueAt(int row, int column) {
            if (data == null) {
                return null;
            } else {
                if (row < getRowCount() && column < getColumnCount()) {
                    return data[row][column];

                } else {
                    return null;
                }
            }
        }

        // Custom methods ------------------------------------------------------

        // set the entire list of files that will be displayed in the table
        public void setFileNameList(Object[][] fileNameList) {
            data = fileNameList;

            // notify the table that things have changed
            fireTableDataChanged();

            // if only one row in the table, select it automatically
            if (getRowCount() == 1) {
                table.addRowSelectionInterval(0, 0);
            }
        }

        // get the file name at a given row
        public String getFileName(int row) {
            return (String)data[row][0];
        }
    }

    // Constructor to set up the basic panel layout
    private FileNameListPanel()
    {
        super(new BorderLayout());
        
        tableModel = new TableModel();
        table = new JTable(tableModel);
        table.getColumnModel().getColumn(0).setPreferredWidth(400);
        table.getColumnModel().getColumn(1).setPreferredWidth(100);

        // Set up sorting of the BPF/CPF file name column  
        TableRowSorter<TableModel> sorter =
            new TableRowSorter<TableModel>(tableModel) {

            // Initialize sort state
            SortState state = SortState.FILENAME_ASCENDING;

            // Set up the 2 comparators - date and file
            Comparator fileDateComparator = new FileDateComparator();
            Comparator fileComparator = new FileComparator();

            @Override public void toggleSortOrder(int column) {

                // Update the sort order for the BPF/CPF name column
                if (column == 0 &&
                    column < getModelWrapper().getColumnCount() &&
                    isSortable(column)) {

                    // RowSorter automatically toggles between ascending and
                    // descending order.  Handle toggles between filename
                    // and date sorting, keeping in sync with ASC/DSC toggles
                    if (state == SortState.FILENAME_ASCENDING) {
                        super.setComparator(0, fileComparator);
                    } else if (state == SortState.DATE_ASCENDING) {
                        super.setComparator(0, fileDateComparator);
                    }

                    // Set sort state to the next value
                    state = state.getNext();

                    // Toggle the sort order for the column
                    super.toggleSortOrder(column);
                }
            }
        };
        table.setRowSorter(sorter);

        // The initial SortOrder is SortOrder.UNSORTED, but the filename list
        // actually starts sorted, so the first sort click to get to SortOrder.
        // ASCENDING for filenames has no effect on the order displayed.  Avoid
        // this extra user click by automatically doing the first "sort click".
        sorter.toggleSortOrder(0);

        JScrollPane scrollPane = new JScrollPane(table);
        add(scrollPane, BorderLayout.CENTER);
    }

    /**
     * Create a file name list panel to display and select files as specified
     * @param fileType        type of files to display
     * @param schema          schema of files to display
     * @param allowMultiple   allow selection of multiple files
     */
    public FileNameListPanel(LoadedFileType fileType, LoadedFileSchema schema,
            boolean allowMultiple) {
        this();
        this.fileType = fileType;
        this.schema = schema;
        if (allowMultiple) {
            table.setSelectionMode(
                    ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
        } else {
            table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        }
    }
    
    /**
     * Get the file names of all selected files
     * An array is always returned, even if the panel was created without
     * allowing multiple selection.
     * @return  array of string file names
     */
    public String[] getSelectedFileNames()
    {
        int selectedRows[] = table.getSelectedRows();
        String fileNames[] = new String[selectedRows.length];
        int tableModelRow;    // internal row corresponding to viewed row
        int i = 0;
        for (int row : selectedRows) {
            // Convert the possibly sorted view row to the internal model row
            tableModelRow = table.convertRowIndexToModel(row);

            fileNames[i++] = tableModel.getFileName(tableModelRow);
        }
        return fileNames;
    }
    
    /**
     * Get the file names of all files displayed
     * @return  array of string file names
     */
    public String[] getFileNames()
    {
        String fileNames[] = new String[tableModel.getRowCount()];
        for (int i=0; i < tableModel.getRowCount(); i++) {
            fileNames[i] = tableModel.getFileName(i);
        }
        return fileNames;
    }

    /**
     * Set the list of files to those with the given criteria
     * @param startDate    display files that overlap the startDate - endDate
     *                     date range
     * @param endDate      display files that overlap the startDate - endDate
     *                     date range
     * @param sensor       display BPF files of a given sensor (null for any)
     * @param status       display files of a given status (null for any)
     * @param latestOnly   display only the latest version of each file
     */
    public void setFileNameList(Calendar startDate, Calendar endDate,
            BpfSensor sensor, CpfStatus status, boolean latestOnly)
    {
        setFileNameList(startDate, endDate, sensor, status, latestOnly, true);
    }

    /**
     * Set the list of files to those with the given criteria
     * @param startDate    display files that overlap the startDate - endDate
     *                     date range
     * @param endDate      display files that overlap the startDate - endDate
     *                     date range
     * @param sensor       display BPF files of a given sensor (null for any)
     * @param status       display files of a given status (null for any)
     * @param latestOnly   display only the latest version of each file
     * @param allowOverlap display all date ranges, even if they overlap
     */
    public void setFileNameList(Calendar startDate, Calendar endDate,
            BpfSensor sensor, CpfStatus status, boolean latestOnly,
            boolean allowOverlap)
    {
        if (services == null) {
            services = new Services();
        }
        Object[][] filenames = null;
        try {
            if (fileType.equals(LoadedFileType.CPF)) {
                filenames = services.getCpfFileNameList(schema, latestOnly,
                        status, startDate, endDate);
            } else if (fileType.equals(LoadedFileType.BPF)) {
                filenames = services.getBpfFileNameList(schema, sensor,
                        latestOnly, status, startDate, endDate);
            }
        } catch (Exception ex) {
            logger.error("Error retrieving list of files from service", ex);
            JOptionPane.showMessageDialog(this,
                WordUtils.wrap("Error retrieving list of files from service: "
                + ex.getMessage(), 80),
                "Service Error", JOptionPane.ERROR_MESSAGE);
        }

        Object[][] filesToDisplay = null;
        if (allowOverlap) {
            filesToDisplay = filenames;
        } else {
            // Filter out overlapping date ranges
            ArrayList<Integer> keep = new ArrayList<Integer>();
            for (int i = 0; i < filenames.length; i++) {
                String fileToCheck = (String)filenames[i][0];
                if (FileNamingConventions.hasBeenDivided(fileToCheck,
                    filenames)) {
                        logger.debug("Ignoring " + fileToCheck 
                            + " : date range has been split");
                } else {
                    // this file has no overlaps, so record its index
                    keep.add(i);
                }
            }
            // Create an array of just the ones kept
            filesToDisplay = new Object[keep.size()][2];
            int kept = 0;
            for (Iterator it = keep.iterator(); it.hasNext();) {
                // get the index of the file to keep
                int i = (Integer)it.next();
                filesToDisplay[kept][0] = filenames[i][0];
                filesToDisplay[kept][1] = filenames[i][1];
                kept++;
            }
        }

        tableModel.setFileNameList(filesToDisplay);
    }

    /**
     * Set the list of files to those with the given criteria
     * @param startDate    display files that overlap the startDate - endDate
     *                     date range
     * @param endDate      display files that overlap the startDate - endDate
     *                     date range
     * @param status       display files of a given status (null for any)
     * @param latestOnly   display only the latest version of each file
     */
    public void setFileNameList(Calendar startDate, Calendar endDate,
        CpfStatus status, boolean latestOnly) {
        if (services == null) {
            services = new Services();
        }
        Object[][] filenames = null;
        try {
            if (fileType.equals(LoadedFileType.CPF)) {
                filenames = services.getCpfProdFileNameList(startDate, endDate,
                    status, latestOnly);
            }
            else {
                throw new Exception("Routine does not support BPF file type");
            }
        } catch (Exception ex) {
            logger.error("Error retrieving list of files from service", ex);
            JOptionPane.showMessageDialog(this,
                WordUtils.wrap("Error retrieving list of files from service: "
                + ex.getMessage(), 80),
                "Service Error", JOptionPane.ERROR_MESSAGE);
        }

        tableModel.setFileNameList(filenames);
    }

    // Getters for class attributes
    public LoadedFileType getFileType() {
        return fileType;
    }

    public LoadedFileSchema getSchema() {
        return schema;
    }

}

// Compare the date portions of the BPF/CPF filenames.  These have the format
//     BPF: (eval_)L[OT]8BPFYYYYMMDDhhmmss_YYYYMMDDhhmmss.NN
//     CPF: (eval_)L8CPFYYYYMMDD_YYYYMMDD.NN
// Use a regular expression that extracts the YYYYMMDD[hhmmss]_YYYYMMDD[hhmmss]
// portion of those filename formats, and also the .NN portion in case the 
// dates match.
class FileDateComparator implements Comparator {
    public int compare(Object o1, Object o2) {
        Pattern pattern;     // Regular expression pattern
        String s1;           // String for the first filename
        String s2;           // String for the second filename
        Matcher s1_matcher;  // Matcher for the first filename
        Matcher s2_matcher;  // Matcher for the second filename
        String s1_date;      // Extracted date for the first filename
        String s2_date;      // Extracted date for the second filename

        if (o1 instanceof String && o2 instanceof String) {
            s1 = (String)o1;
            s2 = (String)o2;

            // Defined a regular expression to match date part of the filename
            pattern = Pattern.compile("[0-9]{8,14}_[0-9]{8,14}.[0-9]{2}");

            // Set up matchers for this regular expression for both filenames
            s1_matcher = pattern.matcher(s1);
            s2_matcher = pattern.matcher(s2);

            // Find the filenames' matches to the regular expression, and 
            // extract the part of the filename that matched.  If there is 
            // no match, compare using the original filename string
            s1_date = s1_matcher.find() ? s1_matcher.group() : s1;
            s2_date = s2_matcher.find() ? s2_matcher.group() : s2;
 
            // Compare the 2 dates 
            return s1_date.compareTo(s2_date);
        }
        return 0;
    }
}

// Compare the order of the full BPF/CPF filenames
class FileComparator implements Comparator {
    public int compare(Object o1, Object o2) {
        if (o1 instanceof String && o2 instanceof String) {
            String s1 = (String)o1;
            String s2 = (String)o2;

            // Compare the 2 filenames 
            return s1.compareTo(s2);
        }
        return 0;
    }
}

