package gui;

import library.LoadedFile;
import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JPanel;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import org.apache.log4j.Logger;

/**
 * Dialog to merge a group/parameter from one file to another.
 */
public class MergeDialog extends JDialog {

    private CpfViewer application;
    private JComboBox<String> groupNameComboBox;
    private JComboBox<String> parameterNameComboBox = null;
                // not used in the merge group dialog
    private JComboBox<String> destinationComboBox;
    private JComboBox<String> sourceComboBox;
    private static final Logger logger = Logger.getLogger(MergeDialog.class);

    /**
     * Constructor for the dialog
     * @param application      application context for the dialog
     * @param title            title of the dialog
     */
    public MergeDialog(CpfViewer application, String title) {
        super(application, title, true);
        this.application = application;
        setLayout(new BorderLayout());
        setTitle("Merge " + title);

        JPanel mainPanel = new JPanel();
        mainPanel.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
        GridBagLayout gridbag = new GridBagLayout();
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gbc.weightx = 1;
        gbc.weighty = 1;
        mainPanel.setLayout(gridbag);

        // Create a combo box of group names
        JLabel groupNameLabel = new JLabel("Group Name");
        mainPanel.add(groupNameLabel, gbc);
        List<String> groupNames = application.parameterViewerPanel.
            parameterTreePanel.getGroupNames();
        Collections.sort(groupNames);
        groupNameComboBox = new JComboBox<String>(groupNames.toArray(
            new String[0]));
        mainPanel.add(groupNameComboBox, gbc);
        // Add action to reset the parameter name dropdown contents
        // when a group is selected
        groupNameComboBox.setAction(new AbstractAction("Group Select") {
            @Override
            public void actionPerformed(ActionEvent e) {
                groupSelected();
            }
        });

        // If merging parameters, create a combo box of parameter names
        if (title.equals("Parameter")) {
            JLabel parameterNameLabel = new JLabel("Parameter Name");
            mainPanel.add(parameterNameLabel, gbc);

            parameterNameComboBox = new JComboBox<String>();
            // Make the parameter combo box same size as the group one
            parameterNameComboBox.setPreferredSize(
                groupNameComboBox.getPreferredSize());
            mainPanel.add(parameterNameComboBox, gbc);
            parameterNameComboBox.setAction(new AbstractAction(
                "Parameter Select") {
                @Override
                public void actionPerformed(ActionEvent e) {
                    parameterSelected();
                }
            });
        }

        ArrayList<String> listOfFiles = new ArrayList<String>();
        for (Iterator it = application.loadedFileList.getActiveFiles().
            iterator(); it.hasNext();) {
            listOfFiles.add(((LoadedFile)it.next()).getFileName());
        }

        JLabel sourceLabel = new JLabel("Source");
        mainPanel.add(sourceLabel, gbc);
        sourceComboBox = new JComboBox<String>(listOfFiles.toArray(
            new String[0]));
        mainPanel.add(sourceComboBox, gbc);

        JLabel destinationLabel = new JLabel("Destination");
        mainPanel.add(destinationLabel, gbc);
        destinationComboBox = new JComboBox<String>(listOfFiles.toArray(
            new String[0]));
        mainPanel.add(destinationComboBox, gbc);

        // Create the buttons
        JPanel buttonPanel = new JPanel();

        JButton okButton = new JButton(new AbstractAction(
                "Ok") {
                @Override
                public void actionPerformed(ActionEvent e) {
                    doMerge();
                    setVisible(false);
                    dispose();
                }
            });
        okButton.setMnemonic(KeyEvent.VK_O);
        okButton.setToolTipText(title);
        buttonPanel.add(okButton);

        JButton closeButton = new JButton(new AbstractAction(
                "Cancel") {
                @Override
                public void actionPerformed(ActionEvent e) {
                    setVisible(false);
                    dispose();
                }
            });
        closeButton.setMnemonic(KeyEvent.VK_C);
        buttonPanel.add(closeButton);

        add(mainPanel, BorderLayout.NORTH);
        add(buttonPanel, BorderLayout.SOUTH);
        pack();
        
        // Initialize to the 1st group in the box, then select the group &
        // parameter highlighted in the tree if possible
        if (groupNameComboBox.getItemCount() > 0) {
            groupNameComboBox.setSelectedIndex(0);
            String selectedTreeGroup = application.parameterViewerPanel.
                parameterTreePanel.getSelectedGroup();
            if (selectedTreeGroup != null) {
                groupNameComboBox.setSelectedItem(selectedTreeGroup);
            }
        }
        if (parameterNameComboBox != null) {
            String selectedTreeParameter = application.parameterViewerPanel.
                parameterTreePanel.getSelectedParameter();
            if (selectedTreeParameter != null) {
                parameterNameComboBox.setSelectedItem(selectedTreeParameter);
            }            
        }
    }

    // Update the parameter combo box when a new group is selected
    private void groupSelected() {
            String groupName = (String)groupNameComboBox.getSelectedItem();
            String parameterName = null;

            // Group combo box changed - update the corresponding Parameter
            // combo box with parameters in this group
            if (parameterNameComboBox != null) {
                parameterNameComboBox.removeAllItems();
                List<String> parameterNames = application.parameterViewerPanel.
                    parameterTreePanel.getParameterNames(groupName);
                Collections.sort(parameterNames);
                for (String s : parameterNames) {
                    parameterNameComboBox.addItem(s);
                }
                parameterName = (String)parameterNameComboBox.getSelectedItem();
            }
            // Update the source file combo box
            updateSource(groupName, parameterName);
    }

    // Update the source when a new parameter is selected
    private void parameterSelected() {
        String groupName = (String)groupNameComboBox.getSelectedItem();
        String parameterName = (String)parameterNameComboBox.getSelectedItem();
        updateSource(groupName, parameterName);
    }

    // Update the source file combo box with only files that have the
    // selected group (and parameter, if not null)
    private void updateSource(String groupName, String parameterName) {
        sourceComboBox.removeAllItems();
        for (Iterator it = application.loadedFileList.getActiveFiles().
            iterator(); it.hasNext();) {
            LoadedFile lf = (LoadedFile)it.next();
            if (lf.getGroup(groupName) != null
                && (parameterName == null ? true
                : lf.getParameter(groupName, parameterName) != null)) {
                sourceComboBox.addItem(lf.getFileName());
            }
        }
    }

    // Perform the requested merge
    private void doMerge() {
        String sourceFileName = (String)sourceComboBox.getSelectedItem();
        String destFileName = (String)destinationComboBox.getSelectedItem();
        if (!sourceFileName.equals(destFileName)) {
            // Get access to the contents of the source file
            LoadedFile sourceFile = application.loadedFileList.getFile(
                sourceFileName);
            // Get the panel for the destination file
            int destLfIndex = application.loadedFileList.getFileIndex(
                destFileName);
            ParameterTablePanel destPanel = null;
            for (ParameterTablePanel ptp :
                application.parameterViewerPanel.getDisplayedPanels()) {
                if (ptp.getLoadedFileIndex() == destLfIndex) {
                    destPanel = ptp;
                    break;
                }
            }
            if (destPanel == null) {
                logger.warn("No panel to update for " + destFileName);
                return;
            }
            // Now do the actual group or parameter merge
            String groupName = (String)groupNameComboBox.getSelectedItem();
            if (parameterNameComboBox == null) {
                logger.debug(" Merging "
                    + groupName + " from " + sourceFileName
                    + " into " + destFileName);
                destPanel.getUndoableModel().mergeGroup(groupName,
                    sourceFile.getGroup(groupName));
            } else {
                String parameterName = (String)parameterNameComboBox.
                    getSelectedItem();
                logger.debug(" Merging "
                    + groupName + "/" + parameterName + " from "
                    + sourceFileName + " into " + destFileName);
                destPanel.getUndoableModel().mergeParameter(
                    groupName, parameterName,
                    sourceFile.getParameter(groupName, parameterName));
            }
            // After modifying the destination file, update the tree display
            application.parameterViewerPanel.parameterTreePanel.
                updateDiffStatus(null);
            application.parameterViewerPanel.parameterTreePanel.
                refreshSelected();
        }
    }
}
