package gui;

import library.DiffStatus;
import java.awt.Color;
import java.awt.Component;
import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JTable;
import javax.swing.border.Border;
import javax.swing.border.EtchedBorder;
import javax.swing.table.DefaultTableCellRenderer;

/**
 * Custom table cell renderer that supports displaying a colored background
 * to indicate differences between parameter tables.
 */
class ParameterTableColorTextRenderer extends DefaultTableCellRenderer {

    Color matchingColor;
    Color diffColor;
    Color missingColor;
    Border unselectedBorder;
    Border selectedBorder;

    public ParameterTableColorTextRenderer(CpfViewer application) {
        super();

        // Create selected and unselected borders
        selectedBorder = BorderFactory.createEtchedBorder(EtchedBorder.LOWERED);
        unselectedBorder = BorderFactory.createEmptyBorder();

        // Get the application highlighting colors
        Color[] highlightColors = application.getHighlightingColors();
        matchingColor = highlightColors[DiffStatus.MATCH.ordinal()];
        diffColor = highlightColors[DiffStatus.DIFF.ordinal()];
        missingColor = highlightColors[DiffStatus.MISSING.ordinal()];
    }

    /**
     * Sets the value of the table cell to be displayed, as well as the
     * background color
     * @param value   new object to place in the table cell
     */
    @Override
    public void setValue(Object value) {
        // Note the value will be either a ParameterValueStatus or a String
        //   It's normally a ParameterValueStatus but gets converted to a String
        //   when edited by the user, and converted back to a
        //   ParameterValueStatus after being updated in the LoadedFile and
        //   re-displayed by selecting the parameter in the
        //   tree (ParameterTreePanel.valueChanged).
        if (value instanceof ParameterValueStatus) {
            DiffStatus valueDiffStatus
                = ((ParameterValueStatus)value).getStatus();
            if (valueDiffStatus.equals(DiffStatus.NONE)) {
                setBackground(Color.WHITE);
            } else if (valueDiffStatus.equals(DiffStatus.MATCH)) {
                setBackground(matchingColor);
            } else if (valueDiffStatus.equals(DiffStatus.MISSING)) {
                setBackground(missingColor);
            } else if (valueDiffStatus.equals(DiffStatus.DIFF)) {
                setBackground(diffColor);
            }
        }
        if (value != null) {
            setText(value.toString());
        }
    }

    /**
     * Get a cell renderer component that sets the cell border based on whether 
     * the cell is selected or not.  This is required since background color is
     * usually used to indicate cell selection, but we use background color for
     * difference highlighting.
     */
    @Override
    public Component getTableCellRendererComponent(JTable table, Object value,
        boolean isSelected, boolean hasFocus, int row, int column) {
        Component c = super.getTableCellRendererComponent(table, value,
            isSelected, hasFocus, row, column);
        if (isSelected) {
            ((JLabel)c).setBorder(selectedBorder);
        } else {
            ((JLabel)c).setBorder(unselectedBorder);
        }
        return c;
    }
}
