package gui;

import java.util.ArrayList;
import javax.swing.table.AbstractTableModel;

/**
 * This table model handles access to the underlying data displayed in the
 * JTable.  The table has 2 columns, one for a row label (basically the 
 * value index which is an Integer) and one for the actual data values.  Since
 * there is one column of data in the table, the underlying data
 * is stored in an ArrayList which allows adding, removing, and modifying
 * individual entries.  The individual table entries are objects, and may be
 * either of type String or ParameterStatus.
 * The table is editable so if values are modified, events are fired
 * to the ParameterTableModelListener.
 */
class ParameterTableModel extends AbstractTableModel {

    private String[] columnNames = {" ", "Value(s)"};
    private ArrayList<Object> datalist = new ArrayList<Object>();
    public static int HEADER_COLUMN = 0;
    public static int DATA_COLUMN = 1;

    ParameterTableModel() {
    }

    @Override
    public int getColumnCount() {
        return columnNames.length;
    }

    @Override
    public int getRowCount() {
        return datalist.size();
    }

    @Override
    public String getColumnName(int col) {
        if (col < columnNames.length) {
            return columnNames[col];
        } else {
            return null;
        }
    }

    @Override
    public Object getValueAt(int row, int col) {
        if (row >= 0 && row < datalist.size()) {
            if (col == HEADER_COLUMN) {
                return (row + 1); // the 1-based row number for the row header
            } else {
                return datalist.get(row);
            }
        } else {
            return null;
        }
    }

    /*
     * JTable uses this method to determine the default renderer/
     * editor for each cell.
     */
    @Override
    public Class getColumnClass(int c) {
        return getValueAt(0, c).getClass();
    }

    /*
     * Indicate if table's data is editable.
     */
    @Override
    public boolean isCellEditable(int row, int col) {
        if (col == DATA_COLUMN) { // data column is editable
            return true;
        } else { // label column is not
            return false;
        }
    }

    /*
     * Method to clear out any existing data from the table
     */
    public void clearData() {
        datalist.clear();
        fireTableStructureChanged();
    }

    /*
     * Method to set initial values in the table.  This is for display
     * purposes only, no data change events are fired.
     */
    public void initValueAt(Object value, int row, int col) {
        if (row >= datalist.size()) {
            // add the new element to the end
            datalist.add(row, value);
        } else {
            // reset the specified element
            datalist.set(row, value);
        }
    }

    /**
     * Method to set the value of an individual table cell.  
     * If a value already exists at the specified location, it is replaced.
     * If the value is beyond the end of the existing data, it will be added.
     * If the value is null, the value will be deleted.  A single value is
     *   deleted when undoing operations that added to the end of the list.
     *   So, since a delete removes an item (thus changing the data indices),
     *   groups of deletes should be done in reverse order (from the bottom up).
     * @param value   the new value
     * @param row     the row in the table to be updated
     * @param col     the column in the table to be updated
     */
    @Override
    public void setValueAt(Object value, int row, int col) {
        if (col != DATA_COLUMN) { // only set values in the data column
            return;
        }
        if (row >= datalist.size()) {
            // add the new element to the end
            datalist.add(row, value);
            fireTableRowsInserted(row, row);
        } else {
            if (value == null) {
                // null value indicates it should be removed
                datalist.remove(row);
                fireTableRowsDeleted(row, row);
            } else {
                // reset the specified element
                datalist.set(row, value);
                fireTableCellUpdated(row, col);
            }
        }
    }
}
