package gui;

import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.TableModel;
import org.apache.log4j.Logger;

/**
 * A listener to handle events when the table changed (fired by 
 * UndoableParameterTableModel).
 */
class ParameterTableModelListener implements TableModelListener {

    private ParameterTablePanel parameterTablePanel;
    private CpfViewer application;
    private static final Logger logger =
            Logger.getLogger(ParameterTableModelListener.class);

    ParameterTableModelListener(ParameterTablePanel parameterTablePanel,
            CpfViewer application) {
        this.parameterTablePanel = parameterTablePanel;
        this.application = application;
    }

    /**
     * Method to handle table change events.  Inserts, updates and deletes are
     * handled, and changes are replicated back to the corresponding ODL tree.
     * (Note currently the only way to add a value (INSERT event) is to paste
     * multiple values at a starting row location such that the number of values
     * extends beyond the end of the table.  See 
     * TableValueTransferHandler.importData.
     * Likewise the only way to delete a value is to UNDO such a paste
     * operation.)
     */
    @Override
    public void tableChanged(TableModelEvent e) {
        int row = e.getFirstRow();
        if (row == TableModelEvent.HEADER_ROW) {
            return;
        }
        int type = e.getType();
        int column = e.getColumn();
        TableModel model = (TableModel)e.getSource();
        String groupName = parameterTablePanel.getCurrentGroup();
        String parameterName = parameterTablePanel.getCurrentParameter();

        // If a table value was changed/inserted, update the
        // corresponding file's ODL tree
        if (type == TableModelEvent.INSERT || type == TableModelEvent.UPDATE) {
            for (; row <= e.getLastRow(); row++) {
                Object data = model.getValueAt(row, column);
                logger.trace("Table listener received notice of "
                    + (type == TableModelEvent.INSERT ? "INSERT" : "UPDATE")
                    + " in row " + row + ", new value is " + data);
                logger.trace("Updating corresponding file "
                    + parameterTablePanel.getLoadedFileIndex());
                application.loadedFileList.getFile(parameterTablePanel.
                    getLoadedFileIndex()).updateValue(groupName,
                    parameterName, data.toString(), row);
            }
        // For deleted values, update the ODL tree value to null
        } else if (type == TableModelEvent.DELETE) {
            for (; row <= e.getLastRow(); row++) {
                logger.trace("Table listener received notice of DELETE in row "
                    + row);
                logger.trace("Updating corresponding file "
                    + parameterTablePanel.getLoadedFileIndex());
                application.loadedFileList.getFile(parameterTablePanel.
                    getLoadedFileIndex()).updateValue(groupName,
                    parameterName, null, row);
            }
        }
        // re-fire the tree selection event so that the parameter
        // values get redrawn
        application.parameterViewerPanel.parameterTreePanel.refreshSelected();
    }
}
