package gui;

import library.DiffStatus;
import javax.swing.*;
import java.util.EventObject;
import java.awt.*;
import java.awt.event.*;
import org.apache.log4j.Logger;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.TableColumn;

/**
 * The panel used to display the table parameter values.  Functionality of the
 * table is provided by supporting classes that implement action listeners,
 * table models, undo managers, cell renderers, etc.
 */
public class ParameterTablePanel extends JPanel implements FocusListener {

    public ParameterToolBar parameterToolBar;
    private JTable table;
    private int id;       // The ID for this parameter table
    private int lfIndex;  // The index of the LoadedFile associated with this
                          // table display
    private String currentGroup = null;   // The current group & parameter being
    private String currentParameter = null;    // displayed in the panel's table
    private static int numberOfParameterTables = 0;  // keep track of the number
                                                     // of panels created
    private ParameterTableUndoManager undoManager = null;
    private UndoableParameterTableModel undoableModel = null;
    private JScrollPane tableView;
    private JScrollBar individualScrollBar; // the scroll bar that scrolls this
                                            // panel independently of all others
    private int scrollBarSizeDiff;          // the difference between the size
                                            // of the scrollbar & scrolled pane
    private static final Logger logger
        = Logger.getLogger(ParameterTablePanel.class);

    /**
     * Constructor used to create a new panel whenever a file is opened in the
     * GUI.
     * @param application  the application context
     */
    public ParameterTablePanel(CpfViewer application, 
        String fileName, int fileIndex) {
        
        super(new BorderLayout());
        this.setSize(200, 600);
        id = numberOfParameterTables++;

        parameterToolBar = new ParameterToolBar(application, this);
        this.add(parameterToolBar, BorderLayout.NORTH);

        lfIndex = fileIndex;
        parameterToolBar.setFileNameLabel(fileName);

        // Create the basic table model for managing the parameter values,
        // and an undoable table model for handling undo & redo
        undoableModel = new UndoableParameterTableModel(this, application);
        final ParameterTableModel valueTableModel = new ParameterTableModel();
        undoableModel.setDelegate(valueTableModel);
        undoManager = new ParameterTableUndoManager();
        undoableModel.addUndoableEditListener(undoManager);
        undoableModel.addTableModelListener(
            new ParameterTableModelListener(this, application));
        table = new JTable(undoableModel);
        table.setFillsViewportHeight(true); // allows dropping on an empty table

        // Set the width of the column of row labels and renderer to center them
        table.setAutoCreateColumnsFromModel(false);
        int rowLabelWidth = 30;
        TableColumn column = table.getColumnModel().getColumn(0); // header
        column.setMaxWidth(rowLabelWidth * 4);
        column.setWidth(rowLabelWidth);
        column.setPreferredWidth(rowLabelWidth);
        column = table.getColumnModel().getColumn(1); // data values
        column.setPreferredWidth(rowLabelWidth * 4);
        DefaultTableCellRenderer rowHeaderRenderer
            = new DefaultTableCellRenderer();
        rowHeaderRenderer.setHorizontalAlignment(JLabel.CENTER);
        rowHeaderRenderer.setBackground(table.getTableHeader().getBackground());
        // Column 0 is just Integer labels
        table.setDefaultRenderer(Integer.class, rowHeaderRenderer);

        // Set up a cell renderer to do color difference highlighting
        ParameterTableColorTextRenderer renderer =
            new ParameterTableColorTextRenderer(application);
        //Table cells may contain either strings or ParameterValueStatus objects
        table.setDefaultRenderer(String.class, renderer);
        table.setDefaultRenderer(ParameterValueStatus.class, renderer);

        // Set a default editor to handle table cells that contain 
        // ParameterValueStatus objects (Strings are handled by default)
        // and make a single click show the cursor to begin editing
        ParameterTextField pTextField = new ParameterTextField();
        ParameterTableCellEditor editor
            = new ParameterTableCellEditor(pTextField);
        table.setDefaultEditor(ParameterValueStatus.class, editor);
        table.putClientProperty("terminateEditOnFocusLost", Boolean.TRUE);
        // Bind the tab key to the same as the return key (traverse down)
        // in the table
        InputMap tableInputMap = table.getInputMap(
            JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
        tableInputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_TAB, 0),
            tableInputMap.get(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0)));

        // Bind the return key in the ParameterTextField (the component used for
        // editing) to the same action as the return key in the JTable, to avoid
        // losing focus on the table when editing is done. 
        Object tableEnterAction = tableInputMap.get(
            KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0));
        InputMap textInputMap = pTextField.getInputMap();
        textInputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0),
            tableEnterAction);

        // Disable drag & drop re-ordering of columns
        table.getTableHeader().setReorderingAllowed(false);
        
        // Enable drag & drop in the table values
        table.setDragEnabled(true);
        table.setDropMode(javax.swing.DropMode.ON);
        // Also add the transfer handler for handling copy & paste
        TableValueTransferHandler tvth = new TableValueTransferHandler(table);
        table.setTransferHandler(tvth);

        // Set up Table keystrokes
        // Identifying the Copy & Paste KeyStrokes
        KeyStroke copyKey = KeyStroke.getKeyStroke(KeyEvent.VK_C,
            ActionEvent.CTRL_MASK, false);
        KeyStroke pasteKey = KeyStroke.getKeyStroke(KeyEvent.VK_V,
            ActionEvent.CTRL_MASK, false);
        // Attach the keystrokes to transfer handler actions
        final String copyCommand = "Copy";
        final String pasteCommand = "Paste";
        table.getInputMap().put(copyKey, copyCommand);
        table.getActionMap().put(copyCommand,
            TableValueTransferHandler.getCopyAction());
        table.getInputMap().put(pasteKey, pasteCommand);
        table.getActionMap().put(pasteCommand,
            TableValueTransferHandler.getPasteAction());

        // Tie Control-Z to Undo and Control-Y to Redo
        KeyStroke undoKey = KeyStroke.getKeyStroke(KeyEvent.VK_Z,
            ActionEvent.CTRL_MASK, false);
        AbstractAction undoAction = new AbstractAction("undo") {
            @Override
            public void actionPerformed(ActionEvent e) {
                doUndo();
            }
        };
        table.getInputMap().put(undoKey, undoAction.getValue(Action.NAME));
        table.getActionMap().put(undoAction.getValue(Action.NAME), undoAction);
        
        KeyStroke redoKey = KeyStroke.getKeyStroke(KeyEvent.VK_Y,
            ActionEvent.CTRL_MASK, false);
        AbstractAction redoAction = new AbstractAction("redo") {
            @Override
            public void actionPerformed(ActionEvent e) {
                doRedo();
            }
        };
        table.getInputMap().put(redoKey, redoAction.getValue(Action.NAME));
        table.getActionMap().put(redoAction.getValue(Action.NAME), redoAction);

        // Tie Control-A to Select All
        KeyStroke selectAllKey = KeyStroke.getKeyStroke(KeyEvent.VK_A,
            ActionEvent.CTRL_MASK, false);
        AbstractAction selectAllAction = new AbstractAction("selectAll") {
            @Override
            public void actionPerformed(ActionEvent e) {
                table.selectAll();
            }
        };
        table.getInputMap().put(selectAllKey,
            selectAllAction.getValue(Action.NAME));
        table.getActionMap().put(selectAllAction.getValue(Action.NAME),
            selectAllAction);

        // Tie Control-E to invoking the editor
        KeyStroke invokeEditorKey = KeyStroke.getKeyStroke(KeyEvent.VK_E,
            ActionEvent.CTRL_MASK, false);
        final EventObject invokeEvent = new EventObject("invoke");
        AbstractAction invokeEditorAction = new AbstractAction("invokeEditor") {
            @Override
            public void actionPerformed(ActionEvent e) {
                if (table.editCellAt(table.getSelectedRow(),
                    table.getSelectedColumn(), invokeEvent)) {
                    table.getEditorComponent().requestFocusInWindow();
                }
            }
        };
        table.getInputMap().put(invokeEditorKey,
            invokeEditorAction.getValue(Action.NAME));
        table.getActionMap().put(invokeEditorAction.getValue(Action.NAME),
            invokeEditorAction);

        // Put the table into a scrolled pane and capture the original scrollbar
        tableView = new JScrollPane(table);
        individualScrollBar = tableView.getVerticalScrollBar();
        scrollBarSizeDiff = tableView.getHeight()
            - individualScrollBar.getVisibleAmount();

        this.add(tableView, BorderLayout.CENTER);
        
        // Add a focus listener on the table to clear selection if the table
        // loses focus.  The table cells also have a focus listener (see the
        // ParameterTextField class).  A cell that appears to be selected has
        // the focus on the ParameterTextField (the cell itself) during editing
        // and on the JTable not during editing.  If either loses focus to
        // a different area of the GUI, ensure any editing is completed and
        // table selection is cleared.
        table.addFocusListener(this);
    }

    // Basic getters for class attributes
    public int getId() {
        return id;
    }

    public int getLoadedFileIndex() {
        return lfIndex;
    }

    public String getCurrentGroup() {
        return currentGroup;
    }

    public String getCurrentParameter() {
        return currentParameter;
    }

    public UndoableParameterTableModel getUndoableModel() {
        return undoableModel;
    }

    public JScrollBar getScrollBar() {
        return this.individualScrollBar;
    }

    public void setScrollBar(JScrollBar newScrollBar) {
        if (newScrollBar == null) {
            tableView.setVerticalScrollBar(individualScrollBar);
            tableView.setVerticalScrollBarPolicy(
                ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
            // ensure the scrollbar gets redrawn correctly in case the parameter
            // table has changed or the window size has been adjusted
            individualScrollBar.setValueIsAdjusting(true);
            individualScrollBar.setValueIsAdjusting(false);
            individualScrollBar.setVisibleAmount(
                tableView.getHeight() - scrollBarSizeDiff);
            individualScrollBar.setValue(1); // reset the location
        } else {
            tableView.setVerticalScrollBar(newScrollBar);
            tableView.setVerticalScrollBarPolicy(
                ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);
        }
    }

    public boolean isSelected() {
        return parameterToolBar.isSelected();
    }

    /*
     * This is the method called when the user selects a parameter in the
     * ParameterTreePanel.
     */
    public void displayValues(String groupName, String parameterName,
        Object[] data, Object[] diffs) {

        // If we were previously editing, and user chose a new parameter to
        // display, stop editing to save the previous value before loading the 
        // value of the new parameter
        if (table.isEditing()) {
            table.getCellEditor().stopCellEditing();
        }

        // Hold on to the group/parameter being displayed
        currentGroup = groupName;
        currentParameter = parameterName;

        UndoableParameterTableModel model =
            (UndoableParameterTableModel)table.getModel();

        // place the array of values into the table
        // Use the initValueAt so that no events are generated
        model.clearData();
        for (int i = 0; i < data.length; i++) {
            model.initValueAt(new ParameterValueStatus((String)data[i],
                (DiffStatus)diffs[i]),
                i, 0);
        }
    }

    /*
     * Method to make the undo & redo operations available to the toolbar
     * associated with this panel
     */
    public void doUndo() {
        logger.trace("Calling undo manager's UNDO");
        if (undoManager != null) {
            undoManager.undo();
        }
    }

    public void doRedo() {
        logger.trace("Calling undo manager's REDO");
        if (undoManager != null) {
            undoManager.redo();
        }
    }

    // ----------- JTable FocusListener methods ---------------
    @Override
    public void focusGained(FocusEvent e) {
    }

    @Override
    public void focusLost(FocusEvent e) {
        if (e.getOppositeComponent() != table) {
            // unselect selected cell(s) when losing focus on this table
            table.clearSelection(); 
        }
    }

    /**
     * Extension of the standard JTextField, with a focus listener to ensure
     * loss-of-focus from table cells is handled
     */
    private class ParameterTextField extends JTextField
        implements FocusListener {
        public ParameterTextField() {
            super();
            addFocusListener(this);
        }

        // ----------- Table Cell FocusListener methods (during editing) -------
        @Override
        public void focusGained(FocusEvent e) {
        }

        @Override
        public void focusLost(FocusEvent e) {
            // if the component gaining the focus is not my containing table,
            // finish editing and clear the table selection
            if (e.getOppositeComponent() != table) {
                if (table.isEditing()) {
                    table.getCellEditor().stopCellEditing();
                }
                table.clearSelection();
            }
        }
    }

    /**
     * Custom editor for parameter value cells, which only starts editing
     * after a double-click.  Normally a single click will select the cell
     * and allow appending text to the end.  Changing isCellEditable means
     * the cell is not editable at all until the user has double-clicked.
     * Single click is reserved for cell selection, drag & drop.
     */
    private class ParameterTableCellEditor extends DefaultCellEditor {

        private boolean clear = false; // clear the cell when editing starts

        ParameterTableCellEditor(JTextField textField) {
            super(textField);
        }

        // This method is called when a cell value is edited by the user.
        @Override
        public Component getTableCellEditorComponent(JTable table, 
            Object initialValue, boolean isSelected,
            int rowIndex, int colIndex) {

            // Clear or initialize the cell component, depending on editing mode
            if (clear) {
                ((JTextField)editorComponent).setText("");
            } else {
                ((JTextField)editorComponent).setText(initialValue.toString());
            }

            // Return the configured component
            return editorComponent;
        }

        // Determine when a cell can be edited (how the editor gets invoked)
        @Override
        public boolean isCellEditable(EventObject evt) {
            // Mouse clicks to invoke the editor - set cursor at end & append
            if (evt instanceof MouseEvent) {
                clear = false;
                // Cell is only editable with double click
                return (((MouseEvent)evt).getClickCount() >= 2);
            }
            // Key press in the cell begins editing - clear previous value
            if (evt instanceof KeyEvent) {
                clear = true;
                return true;
            }
            // Manual events to invoke the editor from a ctrl-E keystroke
            if (evt instanceof EventObject) {
                if (evt.getSource().equals("invoke")) {
                    clear = true;
                    return true;
                }
            }
            // Disallow editing for other event types
            return false;
        }
    }
}
