package gui;

import library.FileNamingConventions;
import actions.CloseAction;
import actions.ExportAction;
import actions.SaveAction;
import actions.SaveAsAction;
import actions.RenameAction;
import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import javax.swing.AbstractAction;
import javax.swing.BoxLayout;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JToolBar;
import javax.swing.JPanel;
import javax.swing.JCheckBox;
import java.awt.Color;
import java.io.File;

/**
 * Tool bar for functions that operate on a single file loaded in the GUI
 */
public class ParameterToolBar extends JPanel {

    private JLabel modifiedLabel;
    public JLabel fileName;
    private JCheckBox selectedFile;
    private ParameterTablePanel parameterTablePanel;

    /**
     * Constructor for the toolbar display
     * @param application           overall application context
     * @param ptp                   the panel that the toolbar will be added to
     */
    public ParameterToolBar(CpfViewer application,
        ParameterTablePanel ptp) {
        
        super(new BorderLayout());
        this.parameterTablePanel = ptp;

        // the toolbar has 2 rows, the first contains the
        // file name label & checkbox
        JPanel labelPanel = new JPanel();
        labelPanel.setLayout(new BoxLayout(labelPanel, BoxLayout.X_AXIS));

        // name of file displayed, and check box to select it for Tool menu
        // operations
        AbstractAction checkBoxAction = new AbstractAction(null) {
            @Override
            public void actionPerformed(ActionEvent e) {
                if (selectedFile.isSelected()) {
                    fileName.setBackground(fileName.getBackground().darker());
                } else {
                    fileName.setBackground(null);
                }
            }
        };
        selectedFile = new JCheckBox(checkBoxAction);
        labelPanel.add(selectedFile);
        fileName = new JLabel();
        fileName.setOpaque(true); // so background color is seen
        labelPanel.add(fileName);

        // the 2nd row of the toolbar has buttons for various actions
        JToolBar buttonBar = new JToolBar("");
        buttonBar.setFloatable(false);

        JButton saveButton = new JButton(new SaveAction(application,
            parameterTablePanel));
        saveButton.setBorderPainted(false);
        buttonBar.add(saveButton);

        JButton saveAsButton = new JButton(new SaveAsAction(application,
            parameterTablePanel));
        saveAsButton.setBorderPainted(false);
        buttonBar.add(saveAsButton);

        JButton renameButton = new JButton(new RenameAction(application, ptp));
        renameButton.setBorderPainted(false);
        buttonBar.add(renameButton);

        JButton exportEvaluationButton = new JButton(new ExportAction(
            application, parameterTablePanel));
        exportEvaluationButton.setBorderPainted(false);
        buttonBar.add(exportEvaluationButton);

        AbstractAction undoAction = new AbstractAction(null,
            new ImageIcon(getClass().getResource(CpfConfig.UNDO_ICON))) {
            @Override
            public void actionPerformed(ActionEvent e) {
                parameterTablePanel.doUndo();
            }
        };
        JButton undoButton = new JButton(undoAction);
        undoButton.setToolTipText("Undo");
        undoButton.setBorderPainted(false);
        buttonBar.add(undoButton);

        AbstractAction redoAction = new AbstractAction(null,
            new ImageIcon(getClass().getResource(CpfConfig.REDO_ICON))) {
            @Override
            public void actionPerformed(ActionEvent e) {
                parameterTablePanel.doRedo();
            }
        };
        JButton redoButton = new JButton(redoAction);
        redoButton.setToolTipText("Redo");
        redoButton.setBorderPainted(false);
        buttonBar.add(redoButton);

        JButton closeButton = new JButton(new CloseAction(application,
            parameterTablePanel));
        closeButton.setBorderPainted(false);
        buttonBar.add(closeButton);

        modifiedLabel = new JLabel();
        modifiedLabel.setForeground(Color.RED);
        buttonBar.addSeparator();
        buttonBar.add(modifiedLabel);

        // add the 2 components to the toolbar
        add(labelPanel, BorderLayout.NORTH);
        add(buttonBar, BorderLayout.CENTER);       
    }

    /**
     * Find out if the file is selected (checkbox checked)
     * @return  true if selected
     */
    public boolean isSelected() {
        return selectedFile.isSelected();
    }

    /**
     * Set the selection checkbox
     * @param selected     boolean indicator if file is selected or not
     */
    public void setSelected(boolean selected) {
        selectedFile.setSelected(selected);
    }

    /**
     * Update the file name label with information about the file
     * @param filePath    path and name of the file
     */
    public void setFileNameLabel(String filePath) {
        File file = new File(filePath);
        fileName.setText(file.getName());
        String schema = "NONE";
        // File name with no path comes from service, either PROD or EVAL
        if (file.getPath().equals(file.getName())) {
            schema = FileNamingConventions.getSchema(file.getName()).toString();
        }
        // prefix the string with a space to ensure the html renders correctly
        String fileInfo = "<html>&nbsp;" + file.getPath() + "<br>"
            + "Type: " + FileNamingConventions.getFileType(file.getName())
            + "<br>"
            + "Schema: " + schema
            + "</html>";
        fileName.setToolTipText(fileInfo);
    }

    /**
     * Set the modified status of this file.  When a file is saved to the file
     * system, it is marked as unmodified.  Thereafter, any change that has
     * not been undone causes it to be marked modified.
     * @param modified  boolean indicator if the file has been modified
     */
    public void setModified(boolean modified) {
        if (modified) {
            modifiedLabel.setText("Modified");
        } else {
            modifiedLabel.setText("");
        }
    }

    /**
     * Find out if a file has been modified.
     * @return  boolean indicator if the file has been modified since saving
     */
    public boolean isModified() {
        if (modifiedLabel.getText().equals("Modified")) {
            return true;
        } else {
            return false;
        }
    }

}
