package gui;

import library.DiffStatus;
import library.LoadedFile;
import gov.nasa.pds.tools.label.GroupStatement;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridLayout;
import javax.swing.ImageIcon;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeSelectionModel;
import javax.swing.UIManager;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.TreePath;
import java.util.*;
import org.apache.log4j.Logger;

/**
 * The panel showing a tree of Groups and Parameters contained in the files
 * that are loaded in the GUI.  The parameter tree is a composite of all
 * groups and parameters in all open files.
 */
public class ParameterTreePanel extends JPanel
        implements TreeSelectionListener, Observer {

    private CpfViewer application;
    private DefaultMutableTreeNode top; // the tree displayed on the GUI - a
                                        // combination of all open files' trees
    private DefaultTreeModel treeModel;
    private JTree tree;
    private int max_x = 600;
    private int max_y = 200;
    private final static Logger logger = 
            Logger.getLogger(ParameterTreePanel.class.getName());
    private final static String rootNode = "File Contents";

    /**
     * Constructor for the panel
     * @param application   application context
     */
    public ParameterTreePanel(CpfViewer application) {
        super(new GridLayout(1, 0));

        this.application = application;
        this.setSize(max_x, max_y);

        Dimension maximumSize = new Dimension(max_x, max_y);
        this.setMaximumSize(maximumSize);

        top = new DefaultMutableTreeNode(new TreeNodeStatus(rootNode));
        // Add a tree model for displaying the tree.  To modify the tree, use
        // the treeModel cover methods, or call reload if you modify the tree
        // nodes themselves.
        treeModel = new DefaultTreeModel(top);

        // Set the Tree icons
        UIManager.put("Tree.expandedIcon", new ImageIcon(getClass().
                getResource(CpfConfig.TREE_EXPANDED_ICON)));
        UIManager.put("Tree.collapsedIcon", new ImageIcon(getClass().
                getResource(CpfConfig.TREE_COLLAPSED_ICON)));

        tree = new JTree(treeModel);
        tree.getSelectionModel().setSelectionMode(
                TreeSelectionModel.SINGLE_TREE_SELECTION);
        tree.setEditable(false);
        tree.setShowsRootHandles(true);

        // Remove the default icons in the tree
        GroupTreeCellRenderer renderer = new GroupTreeCellRenderer();
        renderer.setLeafIcon(null);
        renderer.setClosedIcon(null);
        renderer.setOpenIcon(null);
        tree.setCellRenderer(renderer);

        // Create the scroll pane and add the tree to it.
        JScrollPane treeView = new JScrollPane(tree);

        add(treeView);

        // Listen for when the selection changes
        tree.addTreeSelectionListener(this);

    } // End constructor ParameterTreePanel

    /*
     * See if a group exists -- it is an immediate child of the top of the tree
     */
    private DefaultMutableTreeNode groupExists(String groupName) {
        if (groupName == null) {
            return null;
        }
        DefaultMutableTreeNode found = null;
        for (Enumeration e = top.children(); e.hasMoreElements();) {
            Object o = e.nextElement();
            if (groupName.equals(o.toString())) {
                found = (DefaultMutableTreeNode) o;
                break;
            }
        }
        return found;
    }

    /**
     * See if parameter exists in a given group -- it is leaf of the tree
     * @param groupName       name of the group to check
     * @param parameterName   name of the parameter to look for
     * @return                the tree node of the parameter within the group
     */
    private DefaultMutableTreeNode parameterExists(String groupName,
        String parameterName) {
        DefaultMutableTreeNode groupNode = groupExists(groupName);
        if (groupNode == null || parameterName == null) {
            return null;
        }
        DefaultMutableTreeNode found = null;
        for (Enumeration e = groupNode.depthFirstEnumeration(); 
            e.hasMoreElements(); ) {
            DefaultMutableTreeNode n = (DefaultMutableTreeNode)e.nextElement();
            if (n.isLeaf()) {
                if (parameterName.equals(n.toString())) {
                    found = n;
                    break;
                }
            } else {
                break;
            }
         }

        return found;
    }

    /**
     * Get a list of all groups currently displayed in the tree (all children
     * of the root tree node)
     * @return   List of group names
     */
    public List<String> getGroupNames() {
        ArrayList<String> gList = new ArrayList<String>();

         for (Enumeration e = top.children(); e.hasMoreElements();) {
            Object o = e.nextElement();
            gList.add(o.toString());
        }

        return gList;
    }

    /**
     * Get a list of all the parameters contained within a group node
     * @param groupName   name of group for which to get parameter names
     *                    (specify null to get all parameters)
     * @return            List of parameter names
     */
    public List<String> getParameterNames(String groupName) {
        ArrayList<String> aList = new ArrayList<String>();
        if (groupName == null) {
            return aList;
        }
        DefaultMutableTreeNode group = groupExists(groupName);
        if (group == null)
            return aList;
        for (Enumeration e = group.depthFirstEnumeration(); 
            e.hasMoreElements(); ) {
            DefaultMutableTreeNode n = (DefaultMutableTreeNode)e.nextElement();
            if (n.isLeaf() && n != group) {
                aList.add(n.toString());
            } else {
                break;
            }
         }
        return aList;
    }

    /*
     * The required method of the TreeSelectionListener interface - called
     * whenever the value of the tree selection changes (when the user clicks
     * on a node in the tree).
     */
    @Override
    public void valueChanged(TreeSelectionEvent e) {
        DefaultMutableTreeNode node 
                = (DefaultMutableTreeNode) tree.getLastSelectedPathComponent();
        if (node == null) {
            return;
        }

        // The user may click on a group or parameter.  If they select a group,
        // the parameterName is null and the code that follows will result
        // in an empty array of values being displayed in the parameter table,
        // effectively blanking it out.
        String groupName = getSelectedGroup();
        String parameterName = getSelectedParameter();

        // Each table panel has an arraylist of values for this parameter
        // We need to set up a parallel arraylist of statuses for color
        // highlighting
        // Set up a list of value arraylists of values of each panel
        ArrayList<List<String>> allValues = new ArrayList<List<String>>(
                application.parameterViewerPanel.getDisplayedPanels().size());
        int maxSize = 0;
        int index;

        // Get the values from each table panel's loaded file
        for (ParameterTablePanel ptp :
            application.parameterViewerPanel.getDisplayedPanels()) {
            index = ptp.getLoadedFileIndex();
            List<String> values = application.loadedFileList.getFile(index).
                getParameterValues(groupName, parameterName);
            allValues.add(values);
            if (values.size() > maxSize) {
                maxSize = values.size();
            }
        }

        // Create the Diff arraylist of the maximum length
        ArrayList<DiffStatus> diffs =
                new ArrayList<DiffStatus>(maxSize);
        // Find the first file that has any values for this parameter, use it
        // as the comparator (or, use file 0 if none of the files have values)
        int comparator = 0;
        for (int jfile = 0; jfile < allValues.size(); jfile++) {
            if (allValues.get(jfile).size() > 0) {
                comparator = jfile;
                break;
            }
        }
        // Set each diff element based on all the values of that element
        // loop through the max # of elements
        for (int i = 0; i < maxSize; i++) {
            DiffStatus diffStatus = DiffStatus.NONE;
            DiffStatus newStatus = DiffStatus.NONE;
            // check the i'th element of each arraylist in each file against
            // the comparison file
            for (int jfile = 0; jfile < allValues.size(); jfile++) {
                if (i < allValues.get(jfile).size() &&
                    i < allValues.get(comparator).size()) {
                    if (allValues.get(jfile).get(i).equals(
                        allValues.get(comparator).get(i))) {
                        newStatus = DiffStatus.MATCH;
                    } else {
                        newStatus = DiffStatus.DIFF;
                    }
                } else {
                    newStatus = DiffStatus.MISSING;
                }
                if (diffStatus.compareTo(newStatus) < 0) {
                    diffStatus = newStatus;
                }
            }
            diffs.add(diffStatus);
        }

        // Update each displayed parameter table (each active file)
        Iterator<List<String>> valIt = allValues.iterator();
        for (ParameterTablePanel ptp :
            application.parameterViewerPanel.getDisplayedPanels()) {
            List<String> values = valIt.next();
            ptp.displayValues(groupName, parameterName,
                values.toArray(), diffs.toArray());
        }

    } // End valueChanged

    /**
     * Re-select the previously selected node, to force the corresponding
     * parameter table(s) to be redrawn
     */
    public void refreshSelected() {
        TreePath path = tree.getSelectionPath();
        if (path != null) {
            tree.clearSelection();
            tree.setSelectionPath(path);
            if (path.getPathCount() > 1) {
                updateDiffStatus(
                    (DefaultMutableTreeNode)path.getPathComponent(1));
            }
        }
    }

    /**
     * Select the specified group in the tree
     * @param groupName name of the group in the tree to select
     */
    public void selectGroup(String groupName) {
        DefaultMutableTreeNode[] pathItems = new DefaultMutableTreeNode[2];
        pathItems[0] = top;
        pathItems[1] = groupExists(groupName);
        if (pathItems[1] != null) {
            TreePath path = new TreePath(pathItems);
            tree.setSelectionPath(path);
            tree.scrollPathToVisible(path);
        }
    }

    /**
     * Select the specified group/parameter in the tree
     * @param groupName     name of the group in the tree to select
     * @param parameterName name of the parameter in the group to select
     */
    public void selectParameter(String groupName, String parameterName) {
        DefaultMutableTreeNode[] pathItems = new DefaultMutableTreeNode[3];
        pathItems[0] = top;
        pathItems[1] = groupExists(groupName);
        pathItems[2] = parameterExists(groupName, parameterName);
        if (pathItems[1] != null && pathItems[2] != null) {
            TreePath path = new TreePath(pathItems);
            tree.setSelectionPath(path);
            tree.scrollPathToVisible(path);
        } else if (pathItems[1] != null) {
            // if the parameter was null but not the group, select it
            TreePath path = new TreePath(Arrays.copyOfRange(pathItems, 0, 2));
            tree.setSelectionPath(path);
            tree.scrollPathToVisible(path);
        }
    }

    /**
     * Update the diff status of each tree node within the given group, 
     * based on the currently loaded files.
     */
    public void updateGroupDiffStatus(String groupName) {
        DefaultMutableTreeNode groupNode = groupExists(groupName);
        if (groupNode == null) {
             return;           
        }
        updateDiffStatus(groupNode);
    }
    
    /**
     * Update the diff status of each tree node, starting at startNode, based
     * on the currently loaded files.  If startNode is null, the entire tree
     * is updated.
     */
    public void updateDiffStatus(DefaultMutableTreeNode startNode) {
        logger.trace("Updating difference highlighting in tree");
        // First count the # of open files.  If there's only one, we don't need
        // to do all this
        int numOpen = application.loadedFileList.getActiveFiles().size();
        if (numOpen <= 1) {
            return;
        }

        int first = -1;
        if (startNode == null ) {
            startNode = top;
        }
        // Walk tree, groups first
        for (Enumeration e = startNode.preorderEnumeration();
             e.hasMoreElements();) {
            DefaultMutableTreeNode n = (DefaultMutableTreeNode)e.nextElement();
            // Leaf items (parameters)
            if (n.isLeaf()) {
                if (n.getParent() == null) continue;
                TreeNodeStatus parmStatus = (TreeNodeStatus)n.getUserObject();
                parmStatus.initStatus();
                TreeNodeStatus groupStatus
                    = (TreeNodeStatus)(((DefaultMutableTreeNode)n.getParent()).
                    getUserObject());
                // If this is a direct child of the top node, the groupStatus
                // is actually the top node, which we don't want to color.
                // So reset it to be the same as parmStatus so status changes
                // don't get applied to the top node.
                if (n.getParent() == top) {
                    groupStatus = parmStatus;
                }
                String groupName = n.getParent().toString();
                String parameterName = n.toString();

                first = -1;
                ArrayList<LoadedFile> activeFiles =
                        application.loadedFileList.getActiveFiles();

                for (int i = 0; i < activeFiles.size(); i++) {
                    List<String> values = activeFiles.get(i).
                            getParameterValues(groupName, parameterName);
                    if (values.size() < 1) {
                        parmStatus.addStatus(DiffStatus.MISSING);
                        groupStatus.addStatus(DiffStatus.MISSING);
                    } else {
                        if (first == -1)  {
                            // Don't do any comparison until the 2nd file
                            first = i;
                        } else {
                            // Compare this file's values with the ones from
                            // the first file
                            List<String> firstFileValues = activeFiles.
                                get(first).
                                getParameterValues(groupName, parameterName);
                            int numInCommon = Math.min(values.size(), 
                                firstFileValues.size());
                            if (values.equals(firstFileValues)) {
                                // Exact match - set status to MATCH
                                parmStatus.addStatus(DiffStatus.MATCH);
                                groupStatus.addStatus(DiffStatus.MATCH);
                            } else if (values.size() != firstFileValues.size()
                                && values.subList(0, numInCommon).equals(
                                   firstFileValues.subList(0, numInCommon))) {
                                // Different number of values, but the common
                                // ones are the same
                                parmStatus.addStatus(DiffStatus.MISSING);
                                groupStatus.addStatus(DiffStatus.MISSING);
                            } else {
                                // Values differ
                                parmStatus.addStatus(DiffStatus.DIFF);
                                groupStatus.addStatus(DiffStatus.DIFF);
                            }                        
                        }
                    }
                }
            // Groups 
            } else {
                TreeNodeStatus groupStatus
                        = (TreeNodeStatus)n.getUserObject();
                groupStatus.initStatus();
            }
         }

        // redraw the tree with the new diff status colors
        treeModel.nodeChanged(startNode);

    } // End updateDiffStatus

    // TreeCellRenderer class to provide custom rendering of the tree nodes
    // If you want finer control over the node icons or you want to provide
    // tool tips, you do so by creating a subclass of DefaultTreeCellRenderer
    // and overriding the getTreeCellRendererComponent method. Because
    // DefaultTreeCellRenderer is a subclass of JLabel, you can use any JLabel
    // method -- such as setIcon -- to customize the DefaultTreeCellRenderer.
    class GroupTreeCellRenderer extends DefaultTreeCellRenderer {
        // Colors for matching, different, and missing parameters
        Color matchingColor;
        Color diffColor;
        Color missingColor;
        // Darker versions of the above colors, when an item is selected
        Color matchingColorSel;
        Color diffColorSel;
        Color missingColorSel;

        public GroupTreeCellRenderer() {
            // Get the application highlighting colors
            Color[] highlightColors = application.getHighlightingColors();
            matchingColor = highlightColors[DiffStatus.MATCH.ordinal()];
            diffColor     = highlightColors[DiffStatus.DIFF.ordinal()];
            missingColor  = highlightColors[DiffStatus.MISSING.ordinal()];

            // Get darker versions of the highlighters which are used when
            // an item is selected
            matchingColorSel = matchingColor.darker();
            diffColorSel     = diffColor.darker();
            missingColorSel  = missingColor.darker();
        }
        
        @Override
        public Component getTreeCellRendererComponent(JTree tree, Object value,
            boolean sel, boolean expanded, boolean leaf, int row,
            boolean hasFocus) {

            super.getTreeCellRendererComponent(tree, value, sel, expanded,
                leaf, row, hasFocus);
            TreeNodeStatus ps = (TreeNodeStatus)((DefaultMutableTreeNode)value).
                getUserObject();
            if (ps.getStatus() == DiffStatus.DIFF) {
                this.setBackgroundNonSelectionColor(diffColor);
                this.setBackgroundSelectionColor(diffColorSel);
            } else if (ps.getStatus() == DiffStatus.MATCH) {
                this.setBackgroundNonSelectionColor(matchingColor);
                this.setBackgroundSelectionColor(matchingColorSel);
            } else if (ps.getStatus() == DiffStatus.MISSING) {
                this.setBackgroundNonSelectionColor(missingColor);
                this.setBackgroundSelectionColor(missingColorSel);
            } else if (ps.getStatus() == DiffStatus.NONE) {
                this.setBackgroundNonSelectionColor(Color.WHITE);
                this.setBackgroundSelectionColor(Color.LIGHT_GRAY);
            }
            return this;
        }
    }

    /**
     * The observer interface required method
     * This class is an observer of the LoadedFileList who notifies observers
     * when a file is added to the list or removed from the list
     */
     @Override
    public void update(Observable o, Object arg) {
        if (arg == null) { // A null arg means redraw all
            redrawTree();
        } else { // When an arg is provided, it is the new LoadedFile to display
            LoadedFile lf = (LoadedFile) arg;
            logger.debug("Updating tree for " + lf.getFileName());
            updateTree(lf);
        }
    }

    /**
     * Redraw the entire tree using information from all open files
     */
    public void redrawTree() {
        logger.debug("Redrawing Parameter Tree Panel for all files");
        
        // Get any currently selected group, parameter
        LinkedHashSet<String> expandedGroups = getExpandedGroups();
        String selectedGroup = getSelectedGroup();
        String selectedParameter = getSelectedParameter();

        // Reset the tree and redraw it with all currently active files
        clearTree();
        for (Iterator i = application.loadedFileList.getActiveFiles().
            iterator(); i.hasNext();) {
            LoadedFile lf = (LoadedFile)i.next();
            logger.debug("Re-adding " + lf.getFileName()
                + " to tree display");
            updateTree(lf);
        }
        
        // Reselect the previously selected node.
        expandGroups(expandedGroups);
        if (selectedParameter != null && selectedGroup != null) {
            selectParameter(selectedGroup, selectedParameter);
        } else if (selectedGroup != null) {
            selectGroup(selectedGroup);
        }
    }

    /**
     * In the tree, update all the groups contained in this file
     * @param lf   the newly loaded file whose groups will be updated in the
     *             tree
     */
    public void updateTree(LoadedFile lf) {
        updateTree(lf, lf.getGroupNames());
    }

    /**
     * In the tree, update a single group based on its contents in the file
     * @param lf         the newly loaded file containing the named group
     * @param groupName  the group to update in the tree
     */
    public void updateTree(LoadedFile lf, String groupName) {
        updateTree(lf, java.util.Collections.singletonList(groupName));
    }

    /**
     * Update the tree based on the information in a new file that has been
     * loaded.  Nodes may be added to the tree if the new file has groups
     * or parameters that were not previously present.  Rather than use
     * TreeModel methods to insert nodes, the nodes of the tree are directly
     * manipulated, and the entire model reloaded at once after all
     * changes have been made. Doing this causes the root node to expand and all
     * others collapse, so we hold on to the list of expanded groups and
     * re-expand them after reloading.
     * The file's entire set of groups is used to figure out where in the tree
     * things should be placed, but only the groups listed in groupsToUpdate
     * will be updated in the tree.
     * @param lf   the newly loaded file
     */
    private void updateTree(LoadedFile lf, List<String> groupsToUpdate) {
        logger.debug("Updating parameter tree display for " + lf.getFileName());
        
        // Get the currently expanded groups & current selection so we can reset
        // them after redrawing
        LinkedHashSet<String> expandedGroups = getExpandedGroups();
        String selectedGroup = getSelectedGroup();
        String selectedParameter = getSelectedParameter();

        // Identify certain nodes in the tree used when deciding whether &
        // where to add new nodes
        DefaultMutableTreeNode existingGroup = null;
        DefaultMutableTreeNode existingParameter = null;

        DefaultMutableTreeNode firstMatchingGroup = null;
        DefaultMutableTreeNode lastMatchingGroup = null;
        DefaultMutableTreeNode firstMatchingParameter = null;

        // First go through all the nodes and see if anything matches, and 
        // locate the first one
        for (String groupName : lf.getGroupNames()) {
            existingGroup = groupExists(groupName);
            if (existingGroup != null) {
                firstMatchingGroup = existingGroup;
                break;
            }
        }

        // Now place all groups in proper place
        int location = 0;
        for (String groupName : lf.getGroupNames()) {
            // check if this group is already displayed
            existingGroup = groupExists(groupName);
            if (existingGroup == null) {
                logger.trace("  Adding " + groupName + " to tree");
                // The group didn't exist already, so add it in the proper place
                DefaultMutableTreeNode newGrp = new DefaultMutableTreeNode(
                        new TreeNodeStatus(groupName));
                existingGroup = newGrp;
                if (firstMatchingGroup == null) { 
                    // no matches, so just add at the end
                    location = top.getChildCount();
                } else if (lastMatchingGroup == null) { 
                    // there are matches later, so insert in the beginning
                    lastMatchingGroup = newGrp;
                    location = 0;
                } else {
                    // we're in the middle, get the location & update the
                    // last matching group
                    location = top.getIndex(lastMatchingGroup) + 1;
                    lastMatchingGroup = newGrp;
                }
                if (!groupsToUpdate.contains(groupName)) {
                    continue;
                }
                top.insert(newGrp, location);

                // If the group is newly added, add all its parameters now too.
                // This is faster than adding them to an existing group since
                // we don't have to check if they exist or where they should
                // be placed
                for (String parameterName : lf.getParameterNames(groupName)) {
                    DefaultMutableTreeNode newParm = new DefaultMutableTreeNode(
                            new TreeNodeStatus(parameterName));
                    existingGroup.add(newParm);
                } // End parameter loop

            } else {
                lastMatchingGroup = existingGroup;
                if (!groupsToUpdate.contains(groupName)) {
                    continue;
                }
                logger.trace("  Not adding " + groupName
                        + " to tree, already exists");

                // First go through all the parameter nodes and see if anything
                // matches, and locate the first one
                firstMatchingParameter = null;
                for (String parameterName : lf.getParameterNames(groupName)) {
                    existingParameter = parameterExists(groupName,
                            parameterName);
                    if (existingParameter != null) {
                        firstMatchingParameter = existingParameter;
                        break;
                    }
                }

                // For efficiency, create a map of existing parameter names in
                // this tree group for quicker searching
                List<String> parametersInGroup = getParameterNames(groupName);
                int numberParametersInGroup = parametersInGroup.size();
                HashMap<String, Integer> parametersInGroupMap
                        = new HashMap<String, Integer>();
                for (String s: parametersInGroup) {
                    parametersInGroupMap.put(s, 1);
                }
                
                int foundMatches = 0;
                String lastMatchingParameterName = null;
                for (String parameterName : lf.getParameterNames(groupName)) {
                  if (!parametersInGroupMap.containsKey(parameterName)) {
                        // This parameter doesn't exist in the tree, so add it
                        DefaultMutableTreeNode newParm 
                            = new DefaultMutableTreeNode(
                                new TreeNodeStatus(parameterName));
                        if (firstMatchingParameter == null) { 
                            // no matches, so add at the end
                            existingGroup.add(newParm);
                        } else if (lastMatchingParameterName == null) { 
                            //insert at the beginning
                            lastMatchingParameterName = parameterName;
                            existingGroup.insert(newParm, 0);
                        } else {
                            // we're in the middle, insert here
                            existingParameter = parameterExists(groupName,
                                    lastMatchingParameterName);
                            location = existingParameter == null ? 0 :
                                existingGroup.getIndex(existingParameter) + 1;
                            existingGroup.insert(newParm, location);
                            lastMatchingParameterName = parameterName;
                        }
                    } else {
                        lastMatchingParameterName = parameterName;
                        foundMatches++;
                        // Once we've matched everything already in the tree,
                        // reset firstMatchingParameter to null so remaining
                        // parameters get added to the end (this is quicker)
                        if (foundMatches == numberParametersInGroup) {
                            firstMatchingParameter = null;
                        }
                    }
                } // End parameter loop
            } // New group else clause
        } // End group loop

        // Now that all node changes have been made, have the tree redrawn
        treeModel.reload();

        // Reset expanded groups & select the previously selected node.
        expandGroups(expandedGroups);
        if (selectedParameter != null && selectedGroup != null) {
            selectParameter(selectedGroup, selectedParameter);
        } else if (selectedGroup != null) {
            selectGroup(selectedGroup);
        }
        // Refresh the difference highlighting in case nodes were added
        updateDiffStatus(null);
    }

    /**
     * Blank out the tree, usually this is done to redraw it
     */
    private void clearTree() {
        for (String groupName : getGroupNames()) {
            DefaultMutableTreeNode found = groupExists(groupName);
            treeModel.removeNodeFromParent(found);
        }
    }

    /**
     * Get the name of the group of the selected tree node.  If a parameter
     * node is selected, the name of its group is returned.
     * @return   the name of the selected group, null if no group or parameter
     *           is currently selected
     */
    public String getSelectedGroup() {
        String groupName = null;
        TreePath treePath = tree.getSelectionPath();
        if (treePath != null &&
            treePath.getPathCount() > 1) {
            groupName = treePath.getPathComponent(1).toString();
        }
        return groupName;
    }

    /**
     * Get the name of the parameter of the selected tree node.
     * @return   the name of the selected parameter, null if no parameter
     *           is currently selected
     */
    public String getSelectedParameter() {
        String attribName = null;
        TreePath treePath = tree.getSelectionPath();
        if (treePath != null &&
            treePath.getPathCount() > 2) {
            attribName = treePath.getPathComponent(2).toString();
        }
        return attribName;
    }

    private LinkedHashSet<String> getExpandedGroups() {
        // Get the currently expanded groups so we can reset them after
        // redrawing
        LinkedHashSet<String> expandedGroups = new LinkedHashSet<String>();
        DefaultMutableTreeNode root = (DefaultMutableTreeNode)tree.getModel().
            getRoot();
        TreePath rootTreePath = new TreePath(root);
        // Add the root node if it is expanded
        if (tree.isExpanded(0)) {
            logger.trace("Adding " + rootTreePath.getPathComponent(0).toString()
                + " to expanded groups list");
            expandedGroups.add(rootTreePath.getPathComponent(0).toString());
        }
        // Add any expanded groups 
        Enumeration e = tree.getExpandedDescendants(rootTreePath);
        if (e != null) {
            for (; e.hasMoreElements();) {
                TreePath expanded = (TreePath)e.nextElement();
                if (expanded.getPathCount() > 1) {
                    logger.trace("Adding "
                        + expanded.getPathComponent(1).toString()
                        + " to expanded groups list");
                    expandedGroups.add(expanded.getPathComponent(1).toString());
                }
            }
        }
        return expandedGroups;
    }
    private void expandGroups(LinkedHashSet<String> expandedGroups) {
        // Expand the given list of groups
        DefaultMutableTreeNode[] pathItems = new DefaultMutableTreeNode[2];
        pathItems[0] = top;
        for (String expandGroup : expandedGroups.toArray(new String[0])) {
            logger.trace("++EXPANDING " + expandGroup);
            if (expandGroup.equals(rootNode)) {
                // expand the root tree node
                tree.expandRow(0);
            } else {
                // expand this group
                pathItems[1] = groupExists(expandGroup);
                if (pathItems[1] != null) {
                    TreePath path = new TreePath(pathItems);
                    tree.expandPath(path);
                }
            }
        } 
    }
    
    /**
     * Add a group to all selected files AND to the tree, based on its location
     * in all the currently loaded files, relative to the other groups in the
     * tree
     * @param name    Name of the group to add to the tree
     */
    public void addGroupToFiles(String name) {
        // Add the parameter to the tree if not there already
        List<String> groupsInTree = getGroupNames();
        int treeLoc;  // the location in the tree of this group
        if (groupExists(name) == null) {
            // If a group/parameter is selected, add the group after it
            TreePath treePath = tree.getSelectionPath();
            treeLoc = top.getChildCount();
            if (treePath != null
                && treePath.getPathCount() > 1) {
                treeLoc = top.getIndex((DefaultMutableTreeNode)treePath.
                    getPathComponent(1)) + 1;
            }
            treeModel.insertNodeInto(new DefaultMutableTreeNode(
                new TreeNodeStatus(name)), top, treeLoc);
        } else {
            treeLoc = groupsInTree.indexOf(name);
            logger.debug("Not adding " + name
                + " to tree because it already exists");
        }
        selectGroup(name);

        //Figure out the proper location in each file, which may or may not
        //  already contain the selected group
        int fileGroupLoc=0; //location in the tree of each of this file's groups
        int placement=0;    // where to place the new group in the file
        for (ParameterTablePanel ptp :
            application.parameterViewerPanel.getSelectedPanels()) {
            int id = ptp.getLoadedFileIndex();
            LoadedFile lf = application.loadedFileList.getFile(id);
            if (lf.getGroup(name) != null) {
                logger.debug("Not adding " + name + " to "
                    + lf.getFileName() + " because it already exists");
                continue;
            }
            List<String> thisFileGroupNames = lf.getGroupNames();
            // Initialize the placement to be the end
            placement = thisFileGroupNames.size();
            // Check each group in this file to find the one nearest the
            // selected one
            for (int j = 0; j < thisFileGroupNames.size(); j++) {
                // Find the location of this group in the displayed list
                fileGroupLoc = groupsInTree.indexOf(thisFileGroupNames.get(j));
                if (fileGroupLoc >= treeLoc) {
                    placement = j;
                    break;
                }
            }
            logger.trace(" Adding to selected file " + lf.getFileName()
                + " at location " + placement);
            ptp.getUndoableModel().addGroup(name, placement);
        }
        // After each file has been updated, refresh the selection to cause
        // parameter values to be re-displayed
        refreshSelected();
    }

    /**
     * Merge a group into all selected files.
     * @param groupName    Name of the group to merge.
     * @param gs           ODL GroupStatement to merge into selected files
     */
    public void mergeGroupToFiles(String groupName, GroupStatement gs) {
        logger.info("Merging " + groupName + " into selected files...");
        for (ParameterTablePanel ptp :
            application.parameterViewerPanel.getSelectedPanels()) {
            LoadedFile lf = application.loadedFileList.getFile(
                ptp.getLoadedFileIndex());
            logger.info("Merging " + groupName + " group into "
                + lf.getFileName());
            ptp.getUndoableModel().mergeGroup(groupName, gs);
            updateTree(lf, groupName);
        }
    }

    /**
     * Remove a group from all selected files and the tree, if it was removed
     * from all files
     * @param name   Name of group to remove
     */
    public void removeGroupFromFiles(String name) {
        DefaultMutableTreeNode found = groupExists(name);
        if (found == null) {
            logger.warn("Not removing " + name
                + " from tree because it does not exist");
            return;
        }

        int location;
        for (ParameterTablePanel ptp :
            application.parameterViewerPanel.getSelectedPanels()) {
            // Get the current location if this group
            int id = ptp.getLoadedFileIndex();
            LoadedFile lf = application.loadedFileList.getFile(id);
            List<String> thisFileGroupNames = lf.getGroupNames();
            location = thisFileGroupNames == null ? -1
                : thisFileGroupNames.indexOf(name);

            if (location == -1) {
                logger.warn("Not removing " + name + " from "
                    + lf.getFileName()
                    + " because it does not contain the group");
            } else {
                ptp.getUndoableModel().removeGroup(name, location);
            }
        }
        removeGroupFromTree(name);
    }

    /**
     * Check all open files for existence of this group. Remove from tree if
     * no files contain it.
     * @param name  name of the group to check
     */
    public void removeGroupFromTree(String name) {
        DefaultMutableTreeNode found = groupExists(name);
        if (found == null) {
            logger.warn("Not removing " + name
                + " from tree because it does not exist");
            return;
        }

        boolean existsAnywhere = false;
        for (ParameterTablePanel ptp :
            application.parameterViewerPanel.getDisplayedPanels()) {
            // See if this file has the group
            int id = ptp.getLoadedFileIndex();
            LoadedFile lf = application.loadedFileList.getFile(id);
            if (lf.getGroup(name) != null) {
                existsAnywhere = true;
                break;
            }
        }
        if (existsAnywhere) {
            
            // check if all parameter within removed group still exist
            List<String> parms = getParameterNames(name);
            for (String parm : parms) {
                removeParameterFromTree(name, parm);
            }
            
            // if the group still exists in some files, select it
            selectGroup(name);
            
        } else {
            // if the group node is selected, move the selection to the top
            // node before removing the group node from the tree
            if (name.equals(getSelectedGroup())) {
                tree.setSelectionRow(0);
            }
            treeModel.removeNodeFromParent(found);
        } 
        refreshSelected();
    }

    /**
     * Add a parameter to the tree and all selected files.  The parameter may
     * already exist in the tree, in which case it just needs to be added
     * to the files.
     * @param  groupName     group name to add it to
     * @param  parameterName name of the parameter to add
     * @param  values        initial value for the parameter, passed on to each
     *                       file to which it is added
     */
    public void addParameterToFiles(String groupName, String parameterName,
        String [] values) {
        // Add the group to each file if it doesn't have it
        addGroupToFiles(groupName);
        DefaultMutableTreeNode group = groupExists(groupName);
        if (group == null) { // this shouldn't happen now
            logger.warn("Not adding " + parameterName + " to " + groupName +
                    " because the group does not exist in the tree");
            return;
        }
        // default location to add the parameter is the end of the group
        int location = group.getChildCount();

        // Add the parameter to the tree if not there already
        if (parameterExists(groupName, parameterName) == null) {
            // if a parameter in this group is selected, add the new parameter
            // after it
            TreePath treePath = tree.getSelectionPath();
            if (treePath != null
                && treePath.getPathCount() > 2
                && treePath.getPathComponent(1) == group) {
                location = group.getIndex((DefaultMutableTreeNode)treePath.
                    getPathComponent(2)) + 1;
            }
            treeModel.insertNodeInto(new DefaultMutableTreeNode(
                new TreeNodeStatus(parameterName)), group, location);

        } else {
            logger.debug("Not adding " + groupName + "/" + parameterName
                + " to tree because it already exists");
        }
        selectParameter(groupName, parameterName);

        List<String> displayedParameters = getParameterNames(groupName);
        int locInDisplayed=0;
        int placement=0;
        for (ParameterTablePanel ptp :
            application.parameterViewerPanel.getSelectedPanels()) {
            int id = ptp.getLoadedFileIndex();
            LoadedFile lf = application.loadedFileList.getFile(id);
            if (lf.getParameter(groupName, parameterName) != null) {
                logger.debug("Not adding " + parameterName + " to "
                    + lf.getFileName() + " because it already exists");
                continue;
            }
            List<String> thisFileParameterNames
                = lf.getParameterNames(groupName);
            if (thisFileParameterNames == null) {
                continue;
            }
            // Initialize the placement to be the end
            placement = thisFileParameterNames.size();
            // Check each group in this file to find the one nearest the
            // selected one
            for (int j = 0; j < thisFileParameterNames.size(); j++) {
                // Find the location of this parameter in the displayed list
                locInDisplayed = displayedParameters.indexOf(
                    thisFileParameterNames.get(j));
                if (locInDisplayed >= location) {
                    placement = j;
                    break;
                }
            }
            logger.trace(" Adding to selected file " + lf.getFileName()
                + " at location " + placement);
            ptp.getUndoableModel().addParameter(groupName,
                parameterName, placement, values);
        }
        refreshSelected();
    }

    /**
     * Remove a parameter from all selected files, and from the tree if it was
     * removed everywhere
     * @param  groupName   group name to remove from
     * @param  name        name of the parameter to remove
     */
    public void removeParameterFromFiles(String groupName, String name) {
        logger.debug("Removing parameter " + groupName + "/" + name);
        DefaultMutableTreeNode found = parameterExists(groupName, name);
        if (found == null) {
            logger.warn("Not removing " + groupName + "/" + name
                + " from tree because it does not exist");
            return;
        }
        int location;
        for (ParameterTablePanel ptp :
            application.parameterViewerPanel.getSelectedPanels()) {
            // Get the current value & location in this file
            int id = ptp.getLoadedFileIndex();
            LoadedFile lf = application.loadedFileList.getFile(id);
            List<String> thisFileParameterNames
                = lf.getParameterNames(groupName);
            location = thisFileParameterNames == null ? -1 :
                thisFileParameterNames.indexOf(name);
            if (location == -1) {
                logger.warn("Not removing " + name + " from "
                    + lf.getFileName()
                    + " because it does not contain the parameter");
            } else {
                ptp.getUndoableModel().removeParameter(
                    groupName, name, location,
                    lf.getParameterValues(groupName, name).
                    toArray(new String[0]));
            }
        }
        refreshSelected();
        removeParameterFromTree(groupName, name);
    }
    /**
     * Check all open files for existence of this group/parameter. Remove from
     * tree if no files contain it.
     * @param groupName  name of the group to check
     * @param name       name of the parameter to check
     */
    public void removeParameterFromTree(String groupName, String name) {
        DefaultMutableTreeNode found = parameterExists(groupName, name);
        if (found == null) {
            logger.warn("Not removing " + name
                + " from tree because it does not exist");
            return;
        }

        boolean existsAnywhere = false;
        for (ParameterTablePanel ptp :
            application.parameterViewerPanel.getDisplayedPanels()) {
            // See if this file has the group
            int id = ptp.getLoadedFileIndex();
            LoadedFile lf = application.loadedFileList.getFile(id);
            if (lf.getParameter(groupName, name) != null) {
                existsAnywhere = true;
                break;
            }
        }
        if (existsAnywhere) {
            // if the parameter still exists in some files, select it
            selectParameter(groupName, name);
        } else {
            // if this parameter node is selected, move the selection up to the
            // group
            if (name.equals(getSelectedParameter())) {
                selectGroup(groupName);
            }
            treeModel.removeNodeFromParent(found);
        }
        refreshSelected();
    }
} // End class ParameterTreePanel
