package gui;

import java.awt.GridLayout;
import javax.swing.JPanel;
import java.util.*;

/**
 * Container for the set of panels that display the parameter values for each
 * file loaded in the GUI.
 */
public class ParameterViewerPanel extends JPanel
{
    private ArrayList<ParameterTablePanel> parameterTablePanels;
    public ParameterTreePanel parameterTreePanel;
    private CpfViewer application;

    /**
     * Constructor that defines layout of the panel
     * @param application   application context
     */
    public ParameterViewerPanel(CpfViewer application)
    {
        super(new GridLayout(1,0));
        this.application = application;
        this.setSize(200,600);

        // Create one parameter tree and prepare for a list of parameter tables
        parameterTreePanel = new ParameterTreePanel(application);
        this.add(parameterTreePanel);

        parameterTablePanels = new ArrayList<ParameterTablePanel>();
    }

    /**
     * Remove a single panel from the set of displayed panels
     * @param parameterTablePanel   the panel to remove
     */
    public void removePanel(ParameterTablePanel parameterTablePanel) {
        // ensure the panel is not selected before removing it
        parameterTablePanel.parameterToolBar.setSelected(false);
        this.remove(parameterTablePanel);
        this.validate();
    }

    /**
     * Creates a new panel for a given file and returns its id.
     * Returns null if a new panel was not created (the file was already
     * displayed).
     * @param fileName   name of the file associated with this panel
     * @return           the new ParameterTablePanel for this file
     */
    public ParameterTablePanel createNewPanel(String fileName) {
        int lfIndex = application.loadedFileList.getFileIndex(fileName);
        for (ParameterTablePanel ptp : parameterTablePanels) {
            if (ptp.parameterToolBar.isDisplayable()
                && ptp.getLoadedFileIndex() == lfIndex) {
                // the file was already displayed, do not create a new panel
                return null;
            }
        }
        // Add a panel for this file 
        return addPanel(fileName, lfIndex);
    }

    /**
     * Add a panel for the given file
     * @param fileName   name of the file associated with this panel
     * @param fileIndex  the index number of the LoadedFile associated with this
     *                   panel
     */
    public ParameterTablePanel addPanel(String fileName, int fileIndex)
    {
        // create the new sub-panel and add it to this container
        ParameterTablePanel parameterTablePanel
            = new ParameterTablePanel(application, fileName, fileIndex);
        this.add(parameterTablePanel);

        // add the new sub-panel to the list of the ones I'm keeping track of
        parameterTablePanels.add(parameterTablePanel);
        return parameterTablePanel;
    }

    /**
     * Get all the currently displayed panels (open files)
     * @return  List of parameter table panels
     */
    public List<ParameterTablePanel> getDisplayedPanels() {
        ArrayList<ParameterTablePanel> displayedPanels
            = new ArrayList<ParameterTablePanel>();
        for (ParameterTablePanel ptp : parameterTablePanels) {
            if (ptp.parameterToolBar.isDisplayable()) {
                displayedPanels.add(ptp);
            }
        }
        return displayedPanels;
    }

    /**
     * Get all the currently selected panels (open selected files)
     * @return  List of parameter table panels
     */
    public List<ParameterTablePanel> getSelectedPanels() {
        ArrayList<ParameterTablePanel> selectedPanels
            = new ArrayList<ParameterTablePanel>();
        for (ParameterTablePanel ptp : parameterTablePanels) {
            if (ptp.parameterToolBar.isDisplayable() && ptp.isSelected()) {
                selectedPanels.add(ptp);
            }
        }
        return selectedPanels;
    }

}
