package gui;

import library.FileCompare;
import gov.usgs.ias.gui.common_library.SaveDialog;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.Cursor;
import java.io.IOException;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import java.util.ArrayList;
import java.io.File;
import java.net.URI;
import java.awt.Desktop;
import org.apache.log4j.Logger;

/**
 * Implements the Report Menu Items for the Menu Bar.
 */
public class ReportMenu extends JMenu implements ActionListener
{
    private CpfViewer application;
    private static Logger logger = Logger.getLogger(ReportMenu.class.getName());

    /**
     * Constructor of the ReportMenu
     * @param application  the application context
     */
    public ReportMenu(CpfViewer application)
    {
        //Call the parent constructor, setting the Jmenu Naming and KeyEvent.
        super( "Report" );
        this.application = application;
        setMnemonic( KeyEvent.VK_R );

        // Create the Comparison menu items
        JMenuItem saveItem = new JMenuItem("Save Comparison", KeyEvent.VK_S );
        saveItem.setToolTipText("Save a report of a comparison of the two " +
            "open selected files to the file system");
        add(saveItem);

        JMenuItem viewItem = new JMenuItem("View Comparison", KeyEvent.VK_V );
        viewItem.setToolTipText("View a report of a comparison of the two " +
            "open selected files");
        add(viewItem);
        
        // add listeners
        saveItem.addActionListener( this );
        viewItem.addActionListener( this );
    }

    /**
        * Action Performed by selecting the Report menu items
        * @param event   event for the menu item that was selected
    */
    @Override
    public void actionPerformed(ActionEvent event) {
        // For all commands, verify that 2 files were chosen for comparison
        // Get the first 2 selected files from the parameter tables
        ArrayList<Integer> selectedFiles = new ArrayList<Integer>();
        for (ParameterTablePanel ptp :
            application.parameterViewerPanel.getSelectedPanels()) {
            selectedFiles.add(ptp.getLoadedFileIndex());
        }
        if (selectedFiles.size() != 2) {
            JOptionPane.showMessageDialog(application,
                "Please select 2 files to compare",
                "Error", JOptionPane.ERROR_MESSAGE);
            return;
        }

        String command = event.getActionCommand();
        File reportFile;
        if (command.equals("Save Comparison")) {

            // Get the name of the file to save the report as
            String reportFileName = SaveDialog.SaveAs(application);
            if (reportFileName == null) {
                return;
            }
            reportFile = new File(reportFileName);

            // Generate the report on the file system
            try {
                generateReport(reportFile, selectedFiles);
            } catch (Exception ex) {
                logger.error("Error generating report", ex);
                JOptionPane.showMessageDialog(application,
                    "Error generating report: " + ex.getMessage());
            }

        } else if (command.equals("View Comparison")) {

            // Verify we can pop up a browser window to view the report
            Desktop desktop = null;
            if (Desktop.isDesktopSupported()) {
                desktop = Desktop.getDesktop();
            }
            if (desktop == null
                || !desktop.isSupported(Desktop.Action.BROWSE)) {
                JOptionPane.showMessageDialog(application,
                    "Browser interface is not supported on this platform",
                    "Error", JOptionPane.ERROR_MESSAGE);
                return;
            }

            // The report will be written to a temporary file
            try {
                reportFile = File.createTempFile("diff_", ".html");
            } catch (IOException ex) {
                logger.error("Unable to create temporary file", ex);
                JOptionPane.showMessageDialog(application,
                    "Unable to create temporary file: " + ex.getMessage(),
                    "Error", JOptionPane.ERROR_MESSAGE);
                return;
            }
            reportFile.deleteOnExit(); // cleanup the file when the GUI exits

            // Generate the report in the temp file
            try {
                generateReport(reportFile, selectedFiles);
            } catch (Exception ex) {
                logger.error("Error generating report", ex);
                JOptionPane.showMessageDialog(application,
                    "Error generating report: " + ex.getMessage(),
                    "Error", JOptionPane.ERROR_MESSAGE);
            }

            // Display the report temp file in the browser
            application.setCursor(Cursor.getPredefinedCursor(
                Cursor.WAIT_CURSOR));
            URI uri = reportFile.toURI();
            try {
                desktop.browse(uri);
            } catch (Exception ex) {
                logger.error("Unable to open report in browser", ex);
                JOptionPane.showMessageDialog(application,
                    "Unable to open report in browser: " + ex.getMessage(),
                    "Error", JOptionPane.ERROR_MESSAGE);
                return;
            } finally {
                application.setCursor(null);
            }

        } else {
            JOptionPane.showMessageDialog(application,
                    "Unknown command: " + command,
                    "Unknown", JOptionPane.ERROR_MESSAGE);
        }
    }

    private void generateReport(File reportFile,
        ArrayList<Integer> selectedFiles) throws IOException {

            // Set the cursor busy while the report is generated
            application.setCursor(Cursor.getPredefinedCursor(
                    Cursor.WAIT_CURSOR));
            FileCompare fc = new FileCompare(
                    application.loadedFileList.getFile(selectedFiles.get(0)),
                    application.loadedFileList.getFile(selectedFiles.get(1)));
            fc.compareFiles();
            try {
                fc.setReportFileName(reportFile.getCanonicalPath());
                fc.writeHtmlReport();
            } catch (IOException ex) {
                throw ex;
            } finally {
                application.setCursor(null);
            }
    }
}
