package gui;

import library.FileNamingConventions;
import library.Services;
import library.CpfStatus;
import library.LoadedFileType;
import library.LoadedFileSchema;
import workers.CpfSplitThread;
import gov.usgs.ias.gui.common_library.DateRangeProcessor;
import java.awt.BorderLayout;
import java.awt.Point;
import java.awt.Cursor;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.JOptionPane;
import java.util.Calendar;
import java.text.SimpleDateFormat;
import org.apache.log4j.Logger;
import org.apache.commons.lang.WordUtils;

/**
 * Implements a dialog for splitting a CPF.  A FileNameListPanel is displayed
 * with a list of CPF files available to split.  Once a file is selected,
 * a date range dialog allows the user to set the split date, and an optional
 * new end date.  When the split & end dates are defined, the selected file
 * is imported, cloned, and their names, effective dates, and versions are
 * updated to reflect the new split date ranges.
 */
public class SplitCpfDialog extends JDialog implements ComponentListener,
    ActionListener, DateRangeProcessor {

    private CpfViewer application;
    private FileNameListPanel fileNameListPanel;
    private JButton okButton;
    private StatusBar statusBar;
    private Logger logger = Logger.getLogger(SplitCpfDialog.class.getName());
    private String preSplitFileName = null;
    private String postSplitFileName = null;

    /**
     * Create the dialog with a panel for displaying files that can be split.
     * Only PROD CPFs can be split.
     * @param application   application context for this dialog
     */
    public SplitCpfDialog(CpfViewer application) {
        super(application, "Select Baseline CPF to Split", true);
        this.application = application;
        setLayout(new BorderLayout());

        JPanel mainPanel = new JPanel(new BorderLayout());
        mainPanel.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));

        fileNameListPanel = new FileNameListPanel(LoadedFileType.CPF,
            LoadedFileSchema.PROD, false);
        mainPanel.add(fileNameListPanel, BorderLayout.CENTER);

        JPanel buttonPanel = new JPanel();

        okButton = new JButton("Ok");
        okButton.setMnemonic(KeyEvent.VK_O);
        okButton.setToolTipText("Split");
        buttonPanel.add(okButton);

        JButton closeButton = new JButton("Cancel");
        closeButton.setMnemonic(KeyEvent.VK_C);
        buttonPanel.add(closeButton);

        mainPanel.add(buttonPanel, BorderLayout.SOUTH);

        // create the progress indicator in the status bar
        statusBar = new StatusBar();

        add(mainPanel, BorderLayout.CENTER);
        add(statusBar, BorderLayout.SOUTH);
        pack();

        okButton.addActionListener(this);
        closeButton.addActionListener(this);
        addComponentListener(this);
    }

    public String getPostSplitFileName() {
        return postSplitFileName;
    }

    public String getPreSplitFileName() {
        return preSplitFileName;
    }

    // ComponentListener interface methods
    @Override
    public void componentShown(ComponentEvent e) {
        // Make the cursor busy while we contact the service
        disableButtons();
        this.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

        // Get the list of the latest version of active files
        boolean latestOnly = true;
        fileNameListPanel.setFileNameList(null, null, null, CpfStatus.ACTIVE,
                latestOnly, false);
        if (fileNameListPanel.getFileNames().length > 0) {
            enableButtons();
        }
        this.setCursor(null);
    }

    // dummy component event handlers for events that do not need handling
    @Override
    public void componentHidden(ComponentEvent e) {
    }

    @Override
    public void componentMoved(ComponentEvent e) {
    }

    @Override
    public void componentResized(ComponentEvent e) {
    }

    // helper method to enable buttons that depend on data in the list
    //----------------------------------------------------------------
    private void enableButtons() {
        okButton.setEnabled(true);
    }

    // helper method to disable buttons that depend on data in the list
    //-----------------------------------------------------------------
    private void disableButtons() {
        okButton.setEnabled(false);
    }

    // ActionListener interface methods
    @Override
    public void actionPerformed(ActionEvent e) {
        String command = e.getActionCommand();

        Point loc = application.getLocationOnScreen();
        loc.y += 50;
        loc.x += 50;

        if (command.equals("Ok")) {
            String fileNames[] = fileNameListPanel.getSelectedFileNames();
            setVisible(false);
            if (fileNames.length == 0) {
                return;
            }
            String fileName = fileNames[0];
            logger.debug("Splitting " + fileName);
            // Parse the start & end dates out of the filename, use them as the
            //  intial dates on the split date dialog
            Calendar fileStartDate
                = FileNamingConventions.getStartDate(fileName);
            if (fileStartDate == null) {
                fileStartDate = Calendar.getInstance();
            }
            Calendar fileEndDate
                = FileNamingConventions.getEndDate(fileName);
            if (fileEndDate == null) {
                fileEndDate = Calendar.getInstance();
            }
            // See if the end date can be extended.  It can, only if there
            // are no existing files in the time period after this selected
            // file.  So move the end date up one day and see if any active
            // CPFs exist for that date.
            Calendar nextCpfPeriod = (Calendar)fileEndDate.clone();
            nextCpfPeriod.add(Calendar.DAY_OF_YEAR, 1);
            Services services = new Services();
            Object[][] filenames = null;
            try {
                filenames = services.getCpfProdFileNameList(nextCpfPeriod,
                    nextCpfPeriod, CpfStatus.ACTIVE, false);
            } catch (Exception ex) {
                logger.error("Error retrieving list of files from service", ex);
                JOptionPane.showMessageDialog(this, WordUtils.wrap(
                    "Error retrieving list of files from service: "
                    + ex.getMessage(), 80),
                    "Service Error", JOptionPane.ERROR_MESSAGE);
            }
            boolean allowEndDateExtend
                = !(filenames != null && filenames.length > 0);
            if (!allowEndDateExtend) {
                logger.info("Split CPF for " + fileName
                    + " does not allow extending the end date since "
                    + filenames.length + " files exist for "
                    + nextCpfPeriod.getTime().toString());
            }

            logger.trace("Setting date from file name to "
                + fileStartDate.getTime().toString());
            SplitDateDialog splitDateDialog = new SplitDateDialog(application,
                fileStartDate, fileEndDate, allowEndDateExtend, this);
            splitDateDialog.setTitle("Split Date");
            splitDateDialog.setLocation(loc);
            splitDateDialog.setVisible(true);

        } else if (command.equals("Cancel")) {
            setVisible(false);
        }
    }

    // DateRangeProcessor interface method, for hooking up to the
    // SplitDateDialog
    @Override
    public void processWithDateRange(Calendar splitDate, Calendar newEndDate) {
        splitFile(splitDate, newEndDate);
    }

    // Split the currently selected file on the split date, with an optional
    // new end date
    private void splitFile(Calendar splitDate, Calendar newEndDate) {

        if (!fileNameListPanel.getFileType().equals(LoadedFileType.CPF)) {
            JOptionPane.showMessageDialog(application,
                "Splitting can not be done on "
                + fileNameListPanel.getFileType().toString() + " files",
                "Error", JOptionPane.ERROR_MESSAGE);
            return;
        }
        SimpleDateFormat dateFormatter =
            new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
        String origFile = fileNameListPanel.getSelectedFileNames()[0];
        int splitDay = splitDate.get(Calendar.DAY_OF_MONTH);
        int splitYear = splitDate.get(Calendar.YEAR);
        int splitMonth = splitDate.get(Calendar.MONTH);
        int splitHour = splitDate.get(Calendar.HOUR_OF_DAY);
        int splitMinute = splitDate.get(Calendar.MINUTE);
        int splitSecond = splitDate.get(Calendar.SECOND);
        logger.info("Splitting " + origFile + " at date "
            + (splitMonth + 1) + "/" + splitDay + "/" + splitYear
            + " " + splitHour + ":" + splitMinute + ":" + splitSecond);

        // Construct the 2 new file names 
        // If the original file was
        //   originalstart+------------------+originalend then the new files are
        //   originalstart+------+end1                      and
        //                  start2+----------+originalend   or
        //                  start2+-------------------+newend
        Calendar originalStart = FileNamingConventions.getStartDate(origFile);
        Calendar originalEnd = FileNamingConventions.getEndDate(origFile);
        // If a new end date was provided, use it in the filename, otherwise
        // keep the original end date
        Calendar end2;
        if (newEndDate == null) {
            end2 = originalEnd;
        } else {
            // Ensure the new end date is no earlier than the original end date
            if (newEndDate.before(originalEnd)) {
                JOptionPane.showMessageDialog(application,
                    "New end date " + dateFormatter.format(newEndDate.getTime())
                    + " must be after the original end date "
                    + dateFormatter.format(originalEnd.getTime()),
                    "Error", JOptionPane.ERROR_MESSAGE);
                return;
            }
            end2 = newEndDate;
        }

        // Ensure the split date is between the chosen file's begin & end
        Calendar splitStart = Calendar.getInstance();
        splitStart.clear();
        splitStart.set(splitYear, splitMonth, splitDay,
            splitHour, splitMinute, splitSecond);
        if (splitStart.before(originalStart) || splitStart.after(originalEnd)) {
            JOptionPane.showMessageDialog(application,
                "Split date must be between "
                + dateFormatter.format(originalStart.getTime()) + " and "
                + dateFormatter.format(originalEnd.getTime()),
                "Error", JOptionPane.ERROR_MESSAGE);
            return;
        }

        // Get the version for the new eval files.  Default is the version of
        // the original file + 1.
        int newVersion = FileNamingConventions.getVersion(origFile) + 1;
                
        // Split date is the start date of file 2.
        postSplitFileName = FileNamingConventions.constructFileName(
            LoadedFileSchema.EVAL,
            FileNamingConventions.getSatellite(origFile),
            FileNamingConventions.getSensor(origFile),
            FileNamingConventions.getFileType(origFile),
            splitStart, end2, newVersion);

        // Precision of effective dates is to the second. 
        // So, file 2 started at the split date
        // and the new end date of file 1 is (split date - 1 second)
        Calendar end1 = (Calendar)splitStart.clone();
        end1.add(Calendar.SECOND, -1);
        preSplitFileName = FileNamingConventions.constructFileName(
            LoadedFileSchema.EVAL,
            FileNamingConventions.getSatellite(origFile),
            FileNamingConventions.getSensor(origFile),
            FileNamingConventions.getFileType(origFile),
            originalStart, end1, newVersion);

        // Also look in the eval DB to see if this version of eval files already
        // exist.
        // If so, choose the next higher version (use the same for both pre- 
        // and post- split)
        Services services = new Services();
        Object[][] existingFiles = null;
        try {
            // get the newest version of each eval file
            existingFiles = services.getCpfFileNameList(LoadedFileSchema.EVAL,
                    true, null, null, null);
        } catch (Exception ex) {
            JOptionPane.showMessageDialog(this,
                WordUtils.wrap("Error retrieving list of files from service: "
                + ex.getMessage(), 80),
                "Service Error", JOptionPane.ERROR_MESSAGE);
            return;
        }
        String evalPreBaseName = preSplitFileName.split("\\.")[0];
        String evalPostBaseName = postSplitFileName.split("\\.")[0];
        boolean resetVersion = false;
        for (int exist = 0; exist < existingFiles.length; exist++) {
            String existName = (String)existingFiles[exist][0];
            int existVersion = FileNamingConventions.getVersion(existName);
            // If the existing file's root name matches either split file,
            // increment from its version.
            // We want both pre and post to have the same version.
            if (evalPreBaseName.contains(existName.split("\\.")[0])
                || evalPostBaseName.contains(existName.split("\\.")[0])) {
                if (existVersion >= newVersion) {
                    newVersion = existVersion + 1;
                    resetVersion = true;
                    logger.debug("Resetting eval split files' version to "
                        + newVersion + ", incremented from " + existName);
                }
            }
        }
        // Recreate the eval file names if the version changed due to the
        // previous check for existence
        if (resetVersion) {
            postSplitFileName = FileNamingConventions.constructFileName(
                LoadedFileSchema.EVAL,
                FileNamingConventions.getSatellite(origFile),
                FileNamingConventions.getSensor(origFile),
                FileNamingConventions.getFileType(origFile),
                splitStart, end2, newVersion);
            preSplitFileName = FileNamingConventions.constructFileName(
                LoadedFileSchema.EVAL,
                FileNamingConventions.getSatellite(origFile),
                FileNamingConventions.getSensor(origFile),
                FileNamingConventions.getFileType(origFile),
                originalStart, end1, newVersion);
        }

        if (application.loadedFileList.fileIsPresent(preSplitFileName)) {
            JOptionPane.showMessageDialog(application,
                "This split will create a file " + preSplitFileName
                + " which is already loaded in the GUI",
                "Error", JOptionPane.ERROR_MESSAGE);
            return;
        }
        if (application.loadedFileList.fileIsPresent(postSplitFileName)) {
            JOptionPane.showMessageDialog(application,
                "This split will create a file " + postSplitFileName
                + " which is already loaded in the GUI",
                "Error", JOptionPane.ERROR_MESSAGE);
            return;
        }

        // Ready to begin importing the file & splitting into 2 - do this in
        // a thread
        CpfSplitThread splitThread = new CpfSplitThread(application, origFile,
            preSplitFileName, postSplitFileName, end1, splitStart, newEndDate);
        application.statusBar.progress.setBusy();
        splitThread.execute();
    }
}
