package gui;

import gov.usgs.ias.gui.common_library.*;
import java.awt.BorderLayout;
import javax.swing.BoxLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemListener;
import java.awt.event.ItemEvent;
import java.awt.event.KeyEvent;
import java.util.Calendar;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.JCheckBox;
import org.apache.log4j.Logger;

/**
 * Implements a dialog with split date and optional new end date.
 */
public class SplitDateDialog extends JDialog
    implements ActionListener, ItemListener {

    private DatePanel splitDatePanel = null;  // always have a split date
    private DatePanel newEndDatePanel = null; // may have a new end date
    private DateRangeProcessor processor;     // object to process the data once
                                              // dates are selected
    private JCheckBox enableNewEndDate;       // check box to enable setting a
                                              // new end date

    /**
     * Constructor of the Split Date Dialog Box
     * @param application         application context
     * @param startDate           start date of the file being split, used to 
     *                            initialize the split date
     * @param endDate             end date of the file being split
     * @param allowEndDateExtend  flag to allow the user to selecte a new end 
     *                            date. This should be turned off if the
     *                            period after the selected file already has an 
     *                            existing CPF
     * @param processor           the DateRangeProcessor object that will handle
     *                            the processing of the selected date range
     */
    public SplitDateDialog(CpfViewer application, Calendar startDate,
        Calendar endDate, boolean allowEndDateExtend,
        DateRangeProcessor processor) {

        super(application, "", true);
        this.processor = processor;
        setLayout(new BorderLayout());

        // Create a Date Range Panel for the split date
        JPanel dateRangePanel = new JPanel();
        dateRangePanel.setLayout(new BoxLayout(dateRangePanel,
            BoxLayout.Y_AXIS));
        splitDatePanel = new DatePanel(application, "Split Date",
            application.calendarDialog);
        if (startDate != null) {
            splitDatePanel.setDate(startDate.get(Calendar.YEAR),
                startDate.get(Calendar.MONTH),
                startDate.get(Calendar.DAY_OF_MONTH));
        }
        dateRangePanel.add(splitDatePanel);

        // Create a checkbox and 2nd date panel for displaying the
        // "extend the end date" option -- this should be turned off if the
        // selected date range already has an existing CPF
        JPanel hPanel = new JPanel();
        hPanel.setLayout(new BoxLayout(hPanel, BoxLayout.Y_AXIS));
        enableNewEndDate = new JCheckBox("Set New End Date", false);
        enableNewEndDate.setAlignmentX(CENTER_ALIGNMENT);
        hPanel.add(enableNewEndDate);
        newEndDatePanel = new DatePanel(application, "",
            application.calendarDialog);
        if (endDate != null) {
            newEndDatePanel.setDate(endDate.get(Calendar.YEAR),
                endDate.get(Calendar.MONTH),
                endDate.get(Calendar.DAY_OF_MONTH));
        }
        hPanel.add(newEndDatePanel);
        dateRangePanel.add(hPanel);
        if (!allowEndDateExtend) {
            enableNewEndDate.setEnabled(false);
        }

        //Create a button Panel to hold all of the buttons
        JPanel buttonPanel = new JPanel();

        //Create the Ok button
        JButton okButton = new JButton("Ok");
        okButton.setMnemonic(KeyEvent.VK_O);
        okButton.setToolTipText(getTitle());
        buttonPanel.add(okButton);

        //Create the Cancel button
        JButton closeButton = new JButton("Cancel");
        closeButton.setMnemonic(KeyEvent.VK_C);
        buttonPanel.add(closeButton);

        //Add the panels to the dialog window
        add(dateRangePanel, BorderLayout.NORTH);
        add(buttonPanel, BorderLayout.SOUTH);
        pack();
        // After packing, hide the new end date panel
        newEndDatePanel.setVisible(false);
        
        // add listeners
        enableNewEndDate.addItemListener(this);
        okButton.addActionListener(this);
        closeButton.addActionListener(this);
    }

    // ItemListener methods
    /**
     * Show or hide the end date panel when user selects the end date checkbox
     * @param e  checkbox selection event
     */
    @Override
    public void itemStateChanged(ItemEvent e) {
        if (e.getItem() == enableNewEndDate) {
            if (e.getStateChange() == ItemEvent.SELECTED) {
                newEndDatePanel.setVisible(true);
            } else {
                newEndDatePanel.setVisible(false);
            }
        }
    }

    // ActionListener methods
    /**
     * Action performed event handler for Ok/Cancel button
     */
    @Override
    public void actionPerformed(ActionEvent e) {
        
        String command = e.getActionCommand();

        if (command.equals("Ok")) {
            if (processor != null) {
                processor.processWithDateRange(splitDatePanel.getDate(),
                    newEndDatePanel.isVisible() ? newEndDatePanel.getDate()
                    : null);
            }
            setVisible(false);
        } else if (command.equals("Cancel")) {
            setVisible(false);
        } else {
            Logger.getLogger(SplitDateDialog.class.getName()).warn(
                "Unknown command " + command);
        }
    }
}
