package gui;

import java.awt.Toolkit;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.Clipboard;
import java.util.ArrayList;
import javax.swing.JComponent;
import javax.swing.JTable;
import javax.swing.TransferHandler;
import org.apache.log4j.Logger;

/**
 * Transfer handler for drag & drop and cut & paste of data within a table or
 * between tables.
 * Also implements ActionListener to enable other keyboard actions on the table
 * (such as select all).
 */
class TableValueTransferHandler extends TransferHandler {

    private JTable table;
    private Clipboard systemClipboard;
    private static final Logger logger =
        Logger.getLogger(TableValueTransferHandler.class);
    private static final char SPECIAL = '\uFFFF';

    TableValueTransferHandler(JTable table) {
        this.table = table;
        systemClipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
    }

    // ----- TransferHandler methods ------------
    /**
     * Only allow copying from the source, not moving
     */
    @Override
    public int getSourceActions(JComponent c) {
        return COPY;
    }

    /**
     * Create a transferable package of data from a table based on the values
     * in the cells that are currently selected - for drag
     */
    @Override
    public Transferable createTransferable(JComponent c) {
        JTable sourceTable = (JTable)c;
        StringBuilder sb = new StringBuilder();
        int numrows = sourceTable.getSelectedRowCount();
        int[] rowsselected = sourceTable.getSelectedRows();
        // Only transfer from the data column
        int colselected = ParameterTableModel.DATA_COLUMN;
        // Create a transfer string which has all selected values separated by
        // newlines
        for (int currow = 0; currow < numrows; currow++) {
            // translate any newline in the source data into a special character
            // since the rows of the transferable are separated by newlines
            // (this is to maintain compatibility with copying from other
            //  applications as well).  It will get translated back to newline
            // in the paste/drop operation.
            char[] dataChars
                = sourceTable.getValueAt(rowsselected[currow], colselected).
                    toString().toCharArray();
            for (int curchar = 0; curchar < dataChars.length; curchar++) {
                if (dataChars[curchar] == '\n') {
                    dataChars[curchar] = SPECIAL;
                }
            }
            sb.append(new String(dataChars));
            sb.append("\n");
        }
        return new StringSelection(sb.toString());
    }

    /**
     * Decide if a given drop location can accept (import) the data.  We allow
     * drop anywhere in the table as long as it's String data.  You could
     * also only allow drop over existing cells by checking the DropLocation
     * in the info (return false if dl.getRow() == -1).
     */
    @Override
    public boolean canImport(TransferHandler.TransferSupport info) {
        // we only import Strings
        if (!info.isDataFlavorSupported(DataFlavor.stringFlavor)) {
            return false;
        }
        // only allow import if this table has existing values for this
        // parameter
        UndoableParameterTableModel model = (UndoableParameterTableModel)table.
            getModel();
        if (model.getRowCount() < 1) {
            return false;
        }
        return true;
    }

    /**
     * Handle the import (drop or paste) of data into the destination table.
     */
    @Override
    public boolean importData(TransferHandler.TransferSupport info) {
        if (!info.isDataFlavorSupported(DataFlavor.stringFlavor)) {
            return false;
        }
        UndoableParameterTableModel model = (UndoableParameterTableModel)table.
            getModel();
        int startRow = -1;
        if (info.isDrop()) { // Drag & Drop operation
            JTable.DropLocation dl
                = (JTable.DropLocation)info.getDropLocation();
            startRow = dl.getRow();
            if (startRow == -1) {
                logger.trace("Adding to end of table");
                startRow = model.getRowCount();
            }
        } else { // Cut & paste operation
            startRow = (table.getSelectedRows())[0];
        }
        Transferable t = info.getTransferable();
        String data;
        try {
            data = (String)t.getTransferData(DataFlavor.stringFlavor);
        } catch (Exception e) {
            logger.warn("Caught exception on getting drag & drop data");
            return false;
        } 
        // Drop/Paste is for replacing existing values.  So, check if this
        // parameter has any current values.  If not, the parameter needs
        // to be added to the file using Add Parameter before values can
        // be replaced.
        if (model.getRowCount() < 1) {
            logger.warn("Parameter does not exist in this file");
            return false;
        }
        logger.debug("Inserting values into list beginning at " + startRow);
        // Perform the actual import.
        ArrayList<ParameterValueStatus> parameterStatuses
            = new ArrayList<ParameterValueStatus>();
        String[] values = data.split("\n");
        // translate SPECIAL back to newline in each individual value
        for (int i = 0; i < values.length; i++) {
            // translate any newline in the source data into a special char
            char[] dataChars = values[i].toCharArray();
            for (int curchar = 0; curchar < dataChars.length; curchar++) {
                if (dataChars[curchar] == SPECIAL) {
                    dataChars[curchar] = '\n';
                }
            }
            // reset the value string to the new, modified array of chars
            values[i] = new String(dataChars);
        }
        int currRow;
        int currValue;
        String previousValue;
        for (currRow = startRow, currValue = 0;
            currValue < values.length;
            currRow++, currValue++) {
            previousValue =
                model.getValueAt(startRow, 0) == null
                ? "(none)"
                : model.getValueAt(startRow, 0).toString();
            logger.trace("Replacing row " + currRow + " previous value "
                + previousValue + " with " + values[currValue]);
            parameterStatuses.add(new ParameterValueStatus(values[currValue]));
        }
        model.setValuesAt(parameterStatuses.toArray(), startRow, 1);
        return true;
    }

    /**
     * Handle the Copy action
     */
    @Override
    public void exportToClipboard(JComponent comp, Clipboard clip, int action)
        throws IllegalStateException {

        if (action == COPY) {
            StringSelection stsel;
            StringBuffer sbf = new StringBuffer();
            int numrows = table.getSelectedRowCount();
            logger.trace("Copying " + numrows + " rows to Clipboard");
            int[] rowsselected = table.getSelectedRows();
            // Only transfer from the data column
            int colselected = ParameterTableModel.DATA_COLUMN;
            for (int i = 0; i < numrows; i++) {
                // translate any newline in the source data into a special char
                char[] dataChars
                    = table.getValueAt(rowsselected[i], colselected).
                        toString().toCharArray();
                for (int curchar = 0; curchar < dataChars.length; curchar++) {
                    if (dataChars[curchar] == '\n') {
                        dataChars[curchar] = SPECIAL;
                    }
                }
                sbf.append(new String(dataChars));
                sbf.append("\n");
            }
            stsel = new StringSelection(sbf.toString());
            systemClipboard.setContents(stsel, stsel);
        }
    }
}
