package gui;

import actions.FileNameDialogAction;
import library.LoadedFileList;
import library.FileNamingConventions;
import library.CpfStatus;
import library.LoadedFileType;
import library.LoadedFileSchema;
import library.LoadedFile;
import datagroup.Ut1DataGroup;
import datagroup.SensorAlignmentDataGroup;
import javax.swing.AbstractAction;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.awt.Point;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JScrollBar;
import javax.swing.JOptionPane;
import gov.nasa.pds.tools.label.*;
import java.io.File;
import java.util.Calendar;
import java.util.Observer;
import java.util.Observable;

/**
 * Implements the Tool Menu Items for the Menu Bar.
 */
public class ToolMenu extends JMenu implements Observer {

    private JMenuItem groupMenuItem;
    private JMenuItem parameterMenuItem;
    private JMenuItem addGroupMenuItem;
    private JMenuItem retrieveUT1MenuItem;
    private JMenuItem sensorAlignMenuItem;
    private JMenuItem addParameterMenuItem;
    private JMenuItem removeParameterMenuItem;
    private JMenuItem removeGroupMenuItem;
    private JMenuItem terminateCpfMenuItem;
    private JMenuItem terminateBpfMenuItem;
    private JMenuItem activateCpfMenuItem;
    private JMenuItem activateBpfMenuItem;
    private JMenuItem updateGroupMenuItem;
    private JCheckBoxMenuItem scrollLockMenuItem;
    private JMenu mergeSubMenu;
    private JMenu addSubMenu;
    private JMenu removeSubMenu;
    private JMenu terminateSubMenu;
    private JMenu activateSubMenu;
    private CpfViewer application;
    private AddRemoveGroupDialog addRemoveGroupDialog;
    private AddRemoveParameterDialog addRemoveParameterDialog;
    private MergeDialog mergeDialog;

    /**
     * Constructor of the ToolMenu
     * @param v      the CPF Viewer application
     */
    public ToolMenu(CpfViewer v) {
        //Call the parent constructor, setting the Jmenu Naming and KeyEvent.
        super("Tools");

        this.application = v;

        setMnemonic(KeyEvent.VK_T);

        mergeSubMenu = new JMenu("Merge");
        mergeSubMenu.setMnemonic(KeyEvent.VK_M);
        mergeSubMenu.setEnabled(false); // disabled until 2 files are opened

        //Create the Merge menu items
        AbstractAction mergeGroupAction = new AbstractAction("Group") {
            @Override
            public void actionPerformed(ActionEvent e) {
                Point loc = application.getLocationOnScreen();
                loc.translate(50, 50);
                mergeDialog = new MergeDialog(application, "Group");
                mergeDialog.setLocation(loc);
                mergeDialog.setVisible(true);
            }
        };
        groupMenuItem = new JMenuItem(mergeGroupAction);
        groupMenuItem.setMnemonic(KeyEvent.VK_G);
        groupMenuItem.setToolTipText("Merge the contents of a group from " +
            "one open file into another");
        mergeSubMenu.add(groupMenuItem);

        AbstractAction mergeParameterAction = new AbstractAction("Parameter") {
            @Override
            public void actionPerformed(ActionEvent e) {
                Point loc = application.getLocationOnScreen();
                loc.translate(50, 50);
                mergeDialog = new MergeDialog(application, "Parameter");
                mergeDialog.setLocation(loc);
                mergeDialog.setVisible(true);
            }
        };
        parameterMenuItem = new JMenuItem(mergeParameterAction);
        parameterMenuItem.setMnemonic(KeyEvent.VK_P);
        parameterMenuItem.setToolTipText("Merge the contents of a parameter " +
            "from one open file into another");
        mergeSubMenu.add(parameterMenuItem);

        add(mergeSubMenu);

        //Create the Add menu items
        addSubMenu = new JMenu("Add");
        addSubMenu.setMnemonic(KeyEvent.VK_A);

        AbstractAction addGroupAction = new AbstractAction("Group") {
            @Override
            public void actionPerformed(ActionEvent e) {
                if (!atLeastOneSelected()) {
                    return;
                }
                Point loc = application.getLocationOnScreen();
                loc.translate(50, 50);
                addRemoveGroupDialog = new AddRemoveGroupDialog(application,
                    "Add");
                addRemoveGroupDialog.setLocation(loc);
                addRemoveGroupDialog.setVisible(true);
            }
        };
        addGroupMenuItem = new JMenuItem(addGroupAction);
        addGroupMenuItem.setMnemonic(KeyEvent.VK_G);
        addGroupMenuItem.setToolTipText(
            "Add a new group to open selected files");
        addSubMenu.add(addGroupMenuItem);

        AbstractAction addParameterAction = new AbstractAction("Parameter") {
            @Override
            public void actionPerformed(ActionEvent e) {
                if (!atLeastOneSelected()) {
                    return;
                }
                Point loc = application.getLocationOnScreen();
                loc.translate(50, 50);
                addRemoveParameterDialog = new AddRemoveParameterDialog(
                    application, "Add");
                addRemoveParameterDialog.setLocation(loc);
                addRemoveParameterDialog.setVisible(true);
            }
        };
        addParameterMenuItem = new JMenuItem(addParameterAction);
        addParameterMenuItem.setMnemonic(KeyEvent.VK_P);
        addParameterMenuItem.setToolTipText("Add a new parameter to open " +
            "selected files");
        addSubMenu.add(addParameterMenuItem);

        add(addSubMenu);

        //Create the Remove menu items
        removeSubMenu = new JMenu("Remove");
        removeSubMenu.setMnemonic(KeyEvent.VK_R);

        AbstractAction removeGroupAction = new AbstractAction("Group") {
            @Override
            public void actionPerformed(ActionEvent e) {
                if (!atLeastOneSelected()) {
                    return;
                }
                Point loc = application.getLocationOnScreen();
                loc.translate(50, 50);
                addRemoveGroupDialog = new AddRemoveGroupDialog(application,
                    "Remove");
                addRemoveGroupDialog.setLocation(loc);
                addRemoveGroupDialog.setVisible(true);
            }
        };
        removeGroupMenuItem = new JMenuItem(removeGroupAction);
        removeGroupMenuItem.setMnemonic(KeyEvent.VK_G);
        removeGroupMenuItem.setToolTipText("Remove a group from open " +
            "selected files");
        removeSubMenu.add(removeGroupMenuItem);

        AbstractAction removeParameterAction = new AbstractAction("Parameter") {
            @Override
            public void actionPerformed(ActionEvent e) {
                if (!atLeastOneSelected()) {
                    return;
                }
                Point loc = application.getLocationOnScreen();
                loc.translate(50, 50);
                addRemoveParameterDialog = new AddRemoveParameterDialog(
                    application, "Remove");
                addRemoveParameterDialog.setLocation(loc);
                addRemoveParameterDialog.setVisible(true);
            }
        };
        removeParameterMenuItem = new JMenuItem(removeParameterAction);
        removeParameterMenuItem.setMnemonic(KeyEvent.VK_P);
        removeParameterMenuItem.setToolTipText("Remove a parameter from open " +
            "selected files");
        removeSubMenu.add(removeParameterMenuItem);

        add(removeSubMenu);

        addSeparator();

        //Create the Terminate menu item
        terminateSubMenu = new JMenu("Terminate");
        terminateSubMenu.setMnemonic(KeyEvent.VK_T);
        
        //Create the CPF Menu item
        terminateCpfMenuItem = new JMenuItem(
            new FileNameDialogAction(application, "Select CPFs to Terminate",
                "Terminate", LoadedFileType.CPF, LoadedFileSchema.PROD, null,
                CpfStatus.ACTIVE));
        terminateCpfMenuItem.setMnemonic(KeyEvent.VK_C);
        terminateCpfMenuItem.setToolTipText("Select CPFs to terminate " +
                "(set INACTIVE in the database)");
        terminateSubMenu.add(terminateCpfMenuItem);

        //Create the BPF Menu item
        terminateBpfMenuItem = new JMenuItem(
            new FileNameDialogAction(application, "Select BPFs to Terminate",
                "Terminate", LoadedFileType.BPF, LoadedFileSchema.PROD, null,
                CpfStatus.ACTIVE));
        terminateBpfMenuItem.setMnemonic(KeyEvent.VK_B);
        terminateBpfMenuItem.setToolTipText("Select BPFs to terminate " +
                "(set INACTIVE in the database)");
        terminateSubMenu.add(terminateBpfMenuItem);

        add(terminateSubMenu);

        //Create the Re-Activate menu item
        activateSubMenu = new JMenu("Activate");
        activateSubMenu.setMnemonic(KeyEvent.VK_V);

        //Create the CPF Menu item
        activateCpfMenuItem = new JMenuItem(
            new FileNameDialogAction(application, "Select CPFs to Activate",
                "Activate", LoadedFileType.CPF, LoadedFileSchema.PROD, null,
                CpfStatus.NOT_ACTIVE));
        activateCpfMenuItem.setMnemonic(KeyEvent.VK_C);
        activateCpfMenuItem.setToolTipText("Select CPFs to activate " +
                "(set ACTIVE in the database)");
        activateSubMenu.add(activateCpfMenuItem);

        //Create the BPF Menu item
        activateBpfMenuItem = new JMenuItem(
            new FileNameDialogAction(application, "Select BPFs to Activate",
                "Activate", LoadedFileType.BPF, LoadedFileSchema.PROD, null,
                CpfStatus.NOT_ACTIVE));
        activateBpfMenuItem.setMnemonic(KeyEvent.VK_B);
        activateBpfMenuItem.setToolTipText("Select BPFs to activate " +
                "(set ACTIVE in the database)");
        activateSubMenu.add(activateBpfMenuItem);

        add(activateSubMenu);

        addSeparator();

        //Create the Update Group Across Date Range menu item
        AbstractAction updateRangeAction = new AbstractAction(
            "Update Group Across Date Range") {
            @Override
            public void actionPerformed(ActionEvent e) {
                updateGroupAcrossDateRange();
            }
        };
        updateGroupMenuItem = new JMenuItem(updateRangeAction);
        updateGroupMenuItem.setMnemonic(KeyEvent.VK_U);
        updateGroupMenuItem.setToolTipText("Import a set of CPFs and merge " +
            "the selected group into them");
        add(updateGroupMenuItem);

        //Create the Update Data Group menu 
        JMenuItem updateDataGroupMenu = new JMenu("Update Data Group");
        updateDataGroupMenu.setMnemonic(KeyEvent.VK_D);

        //Create the Retrieve UT1 menu item
       AbstractAction ut1Action = new AbstractAction("UT1-UTC") {
            @Override
            public void actionPerformed(ActionEvent e) {
                if (!atLeastOneSelected()) {
                    return;
                }
                Point loc = application.getLocationOnScreen();
                loc.translate(50, 50);
                initializeDateRangeFromSelectedFile();
                application.dateRangeDialog.setTitle("Update UT1");
                application.dateRangeDialog.setLocation(loc);
                Ut1DataGroup ut1 = new Ut1DataGroup(application);
                // Set up the UT1 dialog to process the user's selected dates
                application.dateRangeDialog.setProcessor(ut1);
                application.dateRangeDialog.setVisible(true);
            }
        };
        retrieveUT1MenuItem = new JMenuItem(ut1Action);
        retrieveUT1MenuItem.setMnemonic(KeyEvent.VK_U);
        retrieveUT1MenuItem.setToolTipText("Update the UT1 group in open " +
            "selected files using USNO data");
        updateDataGroupMenu.add(retrieveUT1MenuItem);

        //Create the Sensor Alignment menu item
        AbstractAction sensorAlignAction = new AbstractAction(
            "OLI Sensor Alignment") {
            @Override
            public void actionPerformed(ActionEvent e) {
                if (!atLeastOneSelected()) {
                    return;
                }
                Point loc = application.getLocationOnScreen();
                loc.translate(50, 50);
                initializeDateRangeFromSelectedFile();
                application.dateRangeDialog.setTitle("Update Sensor Alignment");
                application.dateRangeDialog.setLocation(loc);
                SensorAlignmentDataGroup saGroupUpdater
                    = new SensorAlignmentDataGroup(application);
                // Set up the Sensor Alignment dialog to process the 
                // user's selected dates
                application.dateRangeDialog.setProcessor(saGroupUpdater);
                application.dateRangeDialog.setVisible(true);
            }
        };
        sensorAlignMenuItem = new JMenuItem(sensorAlignAction);
        sensorAlignMenuItem.setMnemonic(KeyEvent.VK_S);
        sensorAlignMenuItem.setToolTipText("Update the Sensor Alignment " +
            "group in open selected files using characterization data");
        updateDataGroupMenu.add(sensorAlignMenuItem);

        add(updateDataGroupMenu);

        //Menu item to synchronize the scroll bars of the parameter tables
        AbstractAction lockSrollBarAction = new AbstractAction(
            "Lock Scrollbars") {
            @Override
            public void actionPerformed(ActionEvent e) {
            if (scrollLockMenuItem.getState()) { // TRUE - lock scrollbars
                JScrollBar masterScrollBar = null;
                for (ParameterTablePanel ptp :
                    application.parameterViewerPanel.getDisplayedPanels()) {
                    if (masterScrollBar == null) {
                        masterScrollBar = ptp.getScrollBar();
                    } else {
                        ptp.setScrollBar(masterScrollBar);
                    }
                    masterScrollBar.setValue(1);
                }
            } else {  // FALSE - unlock scrollbars
                for (ParameterTablePanel ptp :
                    application.parameterViewerPanel.getDisplayedPanels()) {
                    ptp.setScrollBar(null);
                }
            }
           }
        };
        scrollLockMenuItem = new JCheckBoxMenuItem(lockSrollBarAction);
        scrollLockMenuItem.setMnemonic(KeyEvent.VK_L);
        scrollLockMenuItem.setToolTipText("Lock the scrollbars of all " +
            "open files together to synchronize scrolling");
        add(scrollLockMenuItem);        
    }

    /**
     * The observer interface required method.
     * This class is an observer of the LoadedFileList who notifies observers
     * when a file is added to the list or removed from the list.
     */
    @Override
    public void update(Observable o, Object arg) {
        LoadedFileList loadedFileList = (LoadedFileList)o;
        if (loadedFileList.getActiveFiles().size() >= 2) {
            mergeSubMenu.setEnabled(true);
        } else {
            mergeSubMenu.setEnabled(false);
        }
    }

    // Set the initial date range in the dialog using the first file's dates.
    // This is a good starting point, and the user may change them if desired.
    private void initializeDateRangeFromSelectedFile() {
        // Get the first selected file's dates
        for (ParameterTablePanel ptp :
            application.parameterViewerPanel.getSelectedPanels()) {
            int id = ptp.getLoadedFileIndex();
            LoadedFile lf = application.loadedFileList.getFile(id);
            File file = new File(lf.getFileName());
            String baseName = file.getName();
            Calendar startDate = FileNamingConventions.getStartDate(baseName);
            Calendar endDate = FileNamingConventions.getEndDate(baseName);
            application.dateRangeDialog.setRange(startDate, endDate);
            break;
        }
    }

    private boolean atLeastOneSelected() {
        if (application.parameterViewerPanel.getSelectedPanels().
            isEmpty()) {
            JOptionPane.showMessageDialog(application,
                "There are no files selected", "Error",
                JOptionPane.ERROR_MESSAGE);
            return false;
        } else {
            return true;
        }
    }

    private boolean oneSelected() {
        if (application.parameterViewerPanel.getSelectedPanels().size() != 1) {
            JOptionPane.showMessageDialog(application,
                "There must be one file selected", "Error",
                JOptionPane.ERROR_MESSAGE);
            return false;
        } else {
            return true;
        }
    }

    private void updateGroupAcrossDateRange() {
        if (!oneSelected()) {
            return;
        }
        String selectedGroup = application.parameterViewerPanel.
            parameterTreePanel.getSelectedGroup();
        String selectedParameter = application.parameterViewerPanel.
            parameterTreePanel.getSelectedParameter();
        if (selectedGroup == null && selectedParameter == null) {
            JOptionPane.showMessageDialog(application, "Please select a "
                + "Group or Parameter from the tree", "Error",
                JOptionPane.ERROR_MESSAGE);
            return;
        }
        // Get the selected group/parameter from the selected file
        ParameterTablePanel ptp = application.parameterViewerPanel.
            getSelectedPanels().get(0);
        int id = ptp.getLoadedFileIndex();
        LoadedFile lf = application.loadedFileList.getFile(id);
        if (lf == null) {
            JOptionPane.showMessageDialog(application, "Please select a "
                + "source file for the selected group/parameter",
                "Error", JOptionPane.ERROR_MESSAGE);
            return;
        }
        GroupStatement gs = null;
        AttributeStatement as = null;
        if (selectedParameter == null) {
            // No parameter was selected, get the entire group
            gs = lf.getGroup(selectedGroup);
        } else {
            // Create a new group with just this parameter
            as = lf.getParameter(selectedGroup, selectedParameter);
            if (as != null) {
                gs = new GroupStatement(1, selectedGroup);
                gs.addStatement(as);
            }
        }
        if (gs == null && as == null) {
            JOptionPane.showMessageDialog(application, "Selected file "
                + lf.getFileName()
                + " does not contain the selected group/parameter",
                "Error", JOptionPane.ERROR_MESSAGE);
            return;
        }
        Point loc = application.getLocationOnScreen();
        loc.translate(50, 50);
        FileNameDialog fnd = new FileNameDialog(application,
            "Update Group Across Archive", "Update",
            LoadedFileType.CPF, LoadedFileSchema.PROD,
            CpfStatus.ACTIVE, true);
        fnd.setUserData(gs);
        application.dateRangeDialog.setTitle(
            "Update Group Across Date Range");
        application.dateRangeDialog.setProcessor(fnd);
        application.dateRangeDialog.setLocation(loc);
        application.dateRangeDialog.setVisible(true);
    }
}
