package gui;

import library.LoadedFile;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.event.UndoableEditListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.undo.AbstractUndoableEdit;
import javax.swing.undo.UndoableEdit;
import javax.swing.undo.UndoableEditSupport;
import gov.nasa.pds.tools.label.*;
import org.apache.log4j.Logger;

/**
 * This Undoable model keeps track of change events to an individual file.
 * Changes may come from 2 "directions" - either from changes to the parameter
 * values displayed in the parameter table (by the user typing, copy & paste, or
 * drag & drop), or by menu options that may change a group, parameter or
 * parameter value(s) regardless if they are currently displayed in the
 * parameter table or not.  In either case, changes effected here are replicated
 * to the ODL tree.
 * ALL CHANGES TO A FILE SHOULD BE BROKERED THROUGH THIS CLASS SO THAT
 * THEY CAN BE UNDONE.
 *
 * Changes to the values in the parameter table are addressed by the fact that
 * this class extends the AbstractTableModel and communicates with the
 * underlying delegate table model.  In other words, this table model sits on
 * top of the underlying table model which actually manages the table data.
 * This portion of the model creates TableChangeEdit objects which are used to
 * keep track of the edits that have happened in the table so they can be
 * undone if requested.
 *
 * Other changes that are initiated by other menu options are recorded in
 * GroupChangeEdit and ParameterChangeEdit objects.
 */
public class UndoableParameterTableModel extends AbstractTableModel
        implements TableModelListener {

    private UndoableEditSupport support = new UndoableEditSupport();
    private ParameterTableModel delegate; // the underlying TableModel
    private ParameterTablePanel parameterTablePanel;
    private CpfViewer application;
    private int numberOfEditsSinceSave = 0;
    private static final Logger logger =
            Logger.getLogger(UndoableParameterTableModel.class);

    UndoableParameterTableModel(ParameterTablePanel parameterTablePanel,
            CpfViewer application) {
        this.parameterTablePanel = parameterTablePanel;
        this.application = application;
    }

    // Methods to support UNDO
    public void addUndoableEditListener(UndoableEditListener l) {
        support.addUndoableEditListener(l);
    }

    public void removeUndoableEditListener(UndoableEditListener l) {
        support.removeUndoableEditListener(l);
    }

    /**
     * Assign the underlying table model.  All operations in this class are
     * forwarded to the underlying model.
     * @param delegate   the underlying table model that will manage the data
     *                   displayed in the table
     */
    void setDelegate(ParameterTableModel delegate) {
        if (this.delegate != null) {
            this.delegate.removeTableModelListener(this);
        }
        this.delegate = delegate;
        if (this.delegate != null) {
            this.delegate.addTableModelListener(this);
        }
        fireTableStructureChanged();
    }

    // Methods to track whether this file has been modified
    private void incrementEdits() {
        numberOfEditsSinceSave++;
        parameterTablePanel.parameterToolBar.setModified(
            numberOfEditsSinceSave != 0);        
    }
    
    private void decrementEdits() {
        numberOfEditsSinceSave--;
        parameterTablePanel.parameterToolBar.setModified(
            numberOfEditsSinceSave != 0);        
    }

    public void resetModified() {
        numberOfEditsSinceSave = 0;
        parameterTablePanel.parameterToolBar.setModified(false);
    }

    // Make sure the item being modified is displayed in the panel.  This
    // may be required when undoing changes after the user has navigated
    // away from the parameter that had changed.
    private void redisplay(String groupName, String parameterName) {
        if (parameterName == null) {
            application.parameterViewerPanel.parameterTreePanel.
                selectGroup(groupName);
        } else {
            application.parameterViewerPanel.parameterTreePanel.
                selectParameter(groupName, parameterName);
        }
    }

    /**
     * This listener method catches events and forwards them on to the
     * ParameterTableModelListener
     * @param e  the model event
     */
    @Override
    public void tableChanged(TableModelEvent e) {
        TableModelEvent newEvent = new TableModelEvent(this, e.getFirstRow(),
                e.getLastRow(), e.getColumn(), e.getType());
        fireTableChanged(newEvent);
    }

    // Methods that provide access to the underlying TableModel
    @Override
    public int getColumnCount() {
        return delegate.getColumnCount();
    }

    @Override
    public int getRowCount() {
        return delegate.getRowCount();
    }

    @Override
    public String getColumnName(int col) {
        return delegate.getColumnName(col);
    }

    @Override
    public Object getValueAt(int row, int col) {
        return delegate.getValueAt(row, col);
    }

    @Override
    public Class getColumnClass(int c) {
        return getValueAt(0, c).getClass();
    }

    @Override
    public boolean isCellEditable(int row, int col) {
        return delegate.isCellEditable(row, col);
    }

    public void clearData() {
        delegate.clearData();
        fireTableStructureChanged();
    }

    // Initialization of data in the table.  Used for displaying when a new
    // parameter is selected in the tree.  No need to keep undo info.
    public void initValueAt(Object value, int row, int col) {
        delegate.initValueAt(value, row, col);
    }

    // When the user changes a value in a single table cell.
    @Override
    public void setValueAt(Object value, int row, int col) {
        String groupName = parameterTablePanel.getCurrentGroup();
        String parameterName = parameterTablePanel.getCurrentParameter();
        Object oldValue = delegate.getValueAt(row, col);
        if ((oldValue == null && value != null) ||
            !oldValue.equals(value)) {
            delegate.setValueAt(value, row, col);
            // Keep track of the previous value 
            fireChangeEdit(groupName, parameterName, row, col, oldValue, value);
            incrementEdits();
        }
    }

    // New method in the undoable table model to handle batches of value edits
    // This method is used for drop or paste of multiple values
    public void setValuesAt(Object[] values, int row, int col) {
        String groupName = parameterTablePanel.getCurrentGroup();
        String parameterName = parameterTablePanel.getCurrentParameter();
        Object[] oldValues = new Object[values.length];
        Object[] newValues = new Object[values.length];
        boolean haveChanges = false;
        for (int i = 0; i < values.length; i++) {
            oldValues[i] = delegate.getValueAt(row + i, col);
            if ((oldValues[i] == null && values[i] != null) ||
                !oldValues[i].equals(values[i])) {
                delegate.setValueAt(values[i], row + i, col);
                newValues[i] = delegate.getValueAt(row + i, col);
                haveChanges = true;
            }
        }
        if (haveChanges) {
            // Fire one change event for all the values at once
            fireChangeEdit(groupName, parameterName, row, col, 
                    oldValues, newValues);
            incrementEdits();
        }
    }

    // Method of the AbstractTableModel, overridden to handle UNDO.
    // This method is called when there's an undoable edit we want to
    // keep track of
    protected void fireChangeEdit(String groupName, String parameterName, 
            int row, int col,
            Object oldValue, Object newValue) {
        // Update the group diff highlighting
        application.parameterViewerPanel.parameterTreePanel.
                updateGroupDiffStatus(groupName);
        // Save an undoable edit
        UndoableEdit edit = new TableChangeEdit(groupName, parameterName,
                row, col, oldValue, newValue);
        support.beginUpdate();
        support.postEdit(edit);
        support.endUpdate();
    }

    /**
     * Keeps track of table changes which may be undone
     */
    private class TableChangeEdit extends AbstractUndoableEdit {

        private String groupName;
        private String parameterName;
        private int columnIndex;
        private int rowIndex;
        private Object oldValue;
        private Object newValue;

        public TableChangeEdit(String groupName, String parameterName, 
                int rowIndex, int columnIndex, Object oldValue,
                Object newValue) {
            this.groupName = groupName;
            this.parameterName = parameterName;
            this.columnIndex = columnIndex;
            this.rowIndex = rowIndex;
            this.oldValue = oldValue;
            this.newValue = newValue;
        }

        @Override
        public void undo() {
            super.undo();
            // redisplay this group/parameter to ensure the undo gets applied
            // to the correct thing
            redisplay(this.groupName, this.parameterName);
            if (!newValue.getClass().isArray()) {
                // It's a single ParameterStatus
                logger.debug("Undoing group/parameter " + groupName + "/" + 
                        parameterName + " back to old value " +
                        oldValue.toString());
                delegate.setValueAt(oldValue, rowIndex, columnIndex);
            } else {
                // Assume it's an array, change them all back (in reverse order)
                Object[] oldValues = (Object[]) oldValue;
                ParameterValueStatus parameterStatus;
                for (int i = oldValues.length - 1; i >= 0; i--) {
                    if (oldValues[i] == null) {
                        logger.debug("Undoing group/parameter " + groupName +
                               "/" + parameterName + " removing value " + i);
                    } else {
                        logger.debug("Undoing group/parameter " + groupName + 
                               "/" + parameterName + " " + i +
                               " back to old value " + oldValues[i].toString());
                    }
                    parameterStatus = (ParameterValueStatus) oldValues[i];
                    delegate.setValueAt(parameterStatus, rowIndex + i,
                            columnIndex);
                }
            }
            decrementEdits();
            application.parameterViewerPanel.parameterTreePanel.
                refreshSelected();
        }

        @Override
        public void redo() {
            super.redo();
            // redisplay this group/parameter to ensure the undo gets applied
            // to the correct thing
            redisplay(this.groupName, this.parameterName);
            if (!newValue.getClass().isArray()) {
                // It's a single ParameterStatus or String
                logger.debug("Redoing group/parameter " + groupName + "/" +
                   parameterName + " back to new value " + newValue.toString());
                delegate.setValueAt(newValue, rowIndex, columnIndex);
            } else {
                // Assume it's an array, change them all back
                Object[] newValues = (Object[]) newValue;
                logger.trace("new values are an array of length "
                    + newValues.length);
                ParameterValueStatus parameterStatus;
                for (int i = 0; i < newValues.length; i++) {
                    if (newValues[i] == null) {
                        logger.debug("Redoing group/parameter " + groupName +
                              "/" + parameterName + " removing value " + i);
                    } else {
                        logger.debug("Redoing group/parameter " + groupName + 
                              "/" + parameterName + " " + i +
                              " back to new value " + newValues[i].toString());
                    }
                    parameterStatus = (ParameterValueStatus) newValues[i];
                    delegate.setValueAt(parameterStatus, rowIndex + i,
                            columnIndex);
                }
            }
            incrementEdits();
            application.parameterViewerPanel.parameterTreePanel.
                refreshSelected();
        }
    }

    /*
     * The following types of changes are initiated by menu options, and
     * overall implementation is in the ParameterTreePanel, since the change
     * may be applied to multiple files at once.  Undo and redo, however, are
     * done on a single file.  So the undo() and redo() methods have logic
     * to refresh the tree, while the original edit (such as addGroup()) does
     * not.
     */
    
    /**
     * Keeps track of other changes that may be undone, but are not
     * reflected in the table of parameter values (i.e., Group
     * and Parameter adds or removes)
     */
    private enum ChangeType {
        ADD,    // The original change was to add the item
        REMOVE  // The original change was to remove the item
    }

    /**
     * A Group Add or Remove.  When a new group is added, it is empty;
     * when a group is removed, it may not be empty.  So we have to hold
     * on to its contents to restore it in case the Remove is undone.
     */
    private class GroupChangeEdit extends AbstractUndoableEdit {

        private String groupName;
        private int groupLocation;
        private GroupStatement groupContents;
        private ChangeType changeType;

        public GroupChangeEdit(String groupName, int groupLocation,
            GroupStatement contents, ChangeType changeType) {
            this.groupName = groupName;
            this.groupLocation = groupLocation;
            this.groupContents = contents;
            this.changeType = changeType;
        }

        @Override
        public void undo() {
            super.undo();
            // Find out if groupName is currently selected
            // in the tree.  Re-display it if not, then apply the undo
            if (changeType == ChangeType.ADD) {
                // Undoing an add, so remove it.  Need to completely redraw the
                // tree when something is removed.
                application.loadedFileList.getFile(parameterTablePanel.
                    getLoadedFileIndex()).removeGroup(groupName);
                application.parameterViewerPanel.parameterTreePanel.
                    removeGroupFromTree(groupName);
            } else {
                // Undoing a delete, so add it back & update the tree
                application.loadedFileList.getFile(parameterTablePanel.
                    getLoadedFileIndex()).addGroup(groupName,
                    groupLocation, groupContents);
                application.parameterViewerPanel.parameterTreePanel.updateTree(
                    application.loadedFileList.getFile(parameterTablePanel.
                    getLoadedFileIndex()), groupName);
                redisplay(this.groupName, null);
            }
            decrementEdits();
        }

        @Override
        public void redo() {
            super.redo();
            if (changeType == ChangeType.ADD) {
                // Redoing an add, so add it again, update the tree, and select
                // it for display
                application.loadedFileList.getFile(parameterTablePanel.
                    getLoadedFileIndex()).addGroup(groupName,
                    groupLocation, groupContents);
                application.parameterViewerPanel.parameterTreePanel.updateTree(
                    application.loadedFileList.getFile(parameterTablePanel.
                    getLoadedFileIndex()), groupName);
                redisplay(this.groupName, null);
            } else {
                // Redoing a delete, so delete it again, select the group in the
                // tree, then redraw the entire tree so the selection is cleared
                application.loadedFileList.getFile(parameterTablePanel.
                    getLoadedFileIndex()).removeGroup(groupName);
                application.parameterViewerPanel.parameterTreePanel.
                    removeGroupFromTree(groupName);
            }
            incrementEdits();
        }
    }

    // add a new, empty group to the file
    protected void addGroup(String groupName, int groupLocation) {

        // Add the group to the file and record it with the undo manager
        logger.trace("Undo Model adding group " + groupName);
        GroupStatement newGroup = application.loadedFileList.getFile(
            parameterTablePanel.getLoadedFileIndex()).addGroup(
            groupName, groupLocation, null);
        incrementEdits();

        // Record the undoable edit
        logger.trace("Recording undoable group ADD for " + groupName);
        UndoableEdit edit = new GroupChangeEdit(groupName, groupLocation,
            newGroup, ChangeType.ADD);
        support.beginUpdate();
        support.postEdit(edit);
        support.endUpdate();
    }

    // Remove an existing group, saving the contents so it can be undone
    protected void removeGroup(String groupName, int groupLocation) {

        // Remove the group from the file and record it with the undo manager
        logger.trace("Undo Model adding group " + groupName);
        GroupStatement previousGroupContents
            = application.loadedFileList.getFile(
            parameterTablePanel.getLoadedFileIndex()).getGroup(groupName);
        application.loadedFileList.getFile(parameterTablePanel.
            getLoadedFileIndex()).removeGroup(groupName);
        incrementEdits();

        // Record the undoable edit
        logger.trace("Recording undoable group REMOVE for " + groupName);
        UndoableEdit edit = new GroupChangeEdit(groupName, groupLocation,
            previousGroupContents, ChangeType.REMOVE);
        support.beginUpdate();
        support.postEdit(edit);
        support.endUpdate();
    }

    /**
     * A Parameter Add or Remove.  Both added and removed parameters have
     * values that must be kept track of for undoing.
     */
    private class ParameterChangeEdit extends AbstractUndoableEdit {

        private String groupName;
        private String parameterName;
        private int parameterLocation;
        private String[] values;
        private ChangeType changeType;

        public ParameterChangeEdit(String groupName, String parameterName,
                int parameterLocation, String[] values, ChangeType changeType) {
            this.groupName = groupName;
            this.parameterName = parameterName;
            this.parameterLocation = parameterLocation;
            this.values = values;
            this.changeType = changeType;
        }

        @Override
        public void undo() {
            super.undo();
            if (changeType == ChangeType.ADD) {
                // Undoing an add, so remove it & update the tree
                application.loadedFileList.getFile(parameterTablePanel.
                    getLoadedFileIndex()).removeParameter(groupName,
                        parameterName);
                application.parameterViewerPanel.parameterTreePanel.
                    removeParameterFromTree(groupName, parameterName);
            } else {
                // Undoing a delete, so add it back & update the tree
                application.loadedFileList.getFile(parameterTablePanel.
                    getLoadedFileIndex()).addParameter(groupName,
                    parameterName, parameterLocation, values);
                application.parameterViewerPanel.parameterTreePanel.updateTree(
                    application.loadedFileList.getFile(parameterTablePanel.
                    getLoadedFileIndex()), groupName);
            }
            decrementEdits();
        }

        @Override
        public void redo() {
            super.redo();
            if (changeType == ChangeType.ADD) {
                // Redoing an add, so add it again & update the tree
                application.loadedFileList.getFile(parameterTablePanel.
                    getLoadedFileIndex()).addParameter(groupName, parameterName,
                    parameterLocation, values);
                application.parameterViewerPanel.parameterTreePanel.updateTree(
                    application.loadedFileList.getFile(parameterTablePanel.
                        getLoadedFileIndex()), groupName);
                redisplay(this.groupName, this.parameterName);
            } else {
                // Redoing a delete, so delete it again & update the tree
                application.loadedFileList.getFile(parameterTablePanel.
                    getLoadedFileIndex()).removeParameter(groupName,
                    parameterName);
                application.parameterViewerPanel.parameterTreePanel.
                    removeParameterFromTree(groupName, parameterName);
            }
            incrementEdits();
        }
    }

    // Add a new parameter with initial value to the file
    protected void addParameter(String groupName, String parameterName, 
            int parameterLocation, String[] values) {
        // Add the parameter to the file
        logger.trace("Undo Model adding parameter " + parameterName +
                " to group " + groupName);
        application.loadedFileList.getFile(parameterTablePanel.
                getLoadedFileIndex()).addParameter(groupName, parameterName,
                parameterLocation, values);
        incrementEdits();

        // Record the undoable edit
        logger.trace("Recording undoable parameter ADD for " + parameterName);
        UndoableEdit pEdit = new ParameterChangeEdit(groupName, parameterName,
                parameterLocation, values, ChangeType.ADD);
        support.beginUpdate();
        support.postEdit(pEdit);
        support.endUpdate();
    }

    // Remove an existing parameter, saving the contents so it can be undone
    protected void removeParameter(String groupName, String parameterName, 
            int parameterLocation, String[] values) {

        // Add the group to the file and record it with the undo manager
        logger.trace("Undo Model removing parameter " + parameterName +
                " from group " + groupName);
        application.loadedFileList.getFile(parameterTablePanel.
                getLoadedFileIndex()).removeParameter(groupName, parameterName);
        incrementEdits();

        logger.trace("Recording undoable parameter REMOVE for "
                + parameterName);
        UndoableEdit edit = new ParameterChangeEdit(groupName, parameterName,
                parameterLocation, values, ChangeType.REMOVE);
        support.beginUpdate();
        support.postEdit(edit);
        support.endUpdate();
    }

    /**
     * A Group merge.  Unlike a Group Add, the new group being merged in is not
     * empty -- and its contents are merged in with any existing group that the
     * file already has.  Undo swaps back in the original group, while redo
     * swaps back in the newly merged one.
     */
    private class GroupMergeEdit extends AbstractUndoableEdit {

        private String groupName;
        private GroupStatement originalGroup;  // original group contents
        private GroupStatement newGroup;  // contents of the new merged group

        public GroupMergeEdit(String groupName,
                GroupStatement originalGroup, GroupStatement newGroup) {
            this.groupName = groupName;
            this.originalGroup = originalGroup;
            this.newGroup = newGroup;
        }

        @Override
        public void undo() {
            super.undo();
            // On undo, set the group back to the original value
            application.loadedFileList.getFile(parameterTablePanel.
                    getLoadedFileIndex()).setGroup(groupName, originalGroup);
            // No tree changes should have been made, just refresh it
            application.parameterViewerPanel.parameterTreePanel.
                updateDiffStatus(null);
            application.parameterViewerPanel.parameterTreePanel.
                refreshSelected();
            decrementEdits();
            redisplay(this.groupName, null);
        }

        @Override
        public void redo() {
            super.redo();
            // On redo, set the group back to the new value
            application.loadedFileList.getFile(parameterTablePanel.
                getLoadedFileIndex()).setGroup(groupName, newGroup);
            // No tree changes should have been made, just refresh it
            application.parameterViewerPanel.parameterTreePanel.
                updateDiffStatus(null);
            application.parameterViewerPanel.parameterTreePanel.
                refreshSelected();
            incrementEdits();
            redisplay(this.groupName, null);
        }
    }

    // Merge the contents of a new group into the existing group in the file
    public void mergeGroup(String groupName, GroupStatement groupToMerge) {

        // Merge the group into the file and record it with the undo manager
        logger.trace("Undo Model merging into group " + groupName);
        LoadedFile lf = application.loadedFileList.getFile(
                parameterTablePanel.getLoadedFileIndex());
        // Get the original group
        GroupStatement originalGroup = lf.getGroup(groupName);
        // Create a new merged group, and set it into the file
        GroupStatement newGroup = lf.createMergedGroup(originalGroup,
                groupToMerge);
        lf.setGroup(groupName, newGroup);
        incrementEdits();

        // Record the merge edit with the before & after merged groups
        logger.trace("Recording undoable group MERGE for " + groupName);
        UndoableEdit edit = new GroupMergeEdit(groupName, originalGroup,
                newGroup);
        support.beginUpdate();
        support.postEdit(edit);
        support.endUpdate();
    }

    /**
     * A Parameter merge.  The value of the new parameter will overwrite the
     * value of the existing parameter.
     */
    private class ParameterMergeEdit extends AbstractUndoableEdit {

        private String groupName;
        private String parameterName;
        private AttributeStatement originalParameter;
        private AttributeStatement newParameter;

        public ParameterMergeEdit(String groupName, String parameterName,
                AttributeStatement originalParameter,
                AttributeStatement newParameter) {
            this.groupName = groupName;
            this.parameterName = parameterName;
            this.originalParameter = originalParameter;
            this.newParameter = newParameter;
        }

        @Override
        public void undo() {
            super.undo();
            // On undo, set the parameter back to the original value
            application.loadedFileList.getFile(parameterTablePanel.
                getLoadedFileIndex()).setParameter(groupName, parameterName,
                    originalParameter);
            // No tree changes should have been made, just refresh it
            application.parameterViewerPanel.parameterTreePanel.
                refreshSelected();
            decrementEdits();
            redisplay(this.groupName, this.parameterName);
        }

        @Override
        public void redo() {
            super.redo();
            // On redo, set the parameter back to the new value
            application.loadedFileList.getFile(parameterTablePanel.
                getLoadedFileIndex()).setParameter(groupName, parameterName,
                newParameter);
            // No tree changes should have been made, just refresh it
            application.parameterViewerPanel.parameterTreePanel.
                refreshSelected();
            incrementEdits();
            redisplay(this.groupName, this.parameterName);
        }
    }

    // Merge the contents of a new parameter into the existing parameter in the
    // file
    public void mergeParameter(String groupName, String parameterName,
            AttributeStatement parameterToMerge) {

        // Merge the group into the file and record it with the undo manager
        logger.trace("Undo Model merging into parameter " + parameterName);
        LoadedFile lf = application.loadedFileList.getFile(
                parameterTablePanel.getLoadedFileIndex());
        // Get the original parameter
        AttributeStatement originalParameter = lf.getParameter(groupName,
                parameterName);
        // Create a new merged parameter, and set it into the file
        AttributeStatement newParameter = lf.createMergedParameter(
            originalParameter, parameterToMerge);
        lf.setParameter(groupName, parameterName, newParameter);
        incrementEdits();

        // Record the merge edit with the before & after merged parameters
        logger.trace("Recording undoable group MERGE for " + parameterName);
        UndoableEdit edit = new ParameterMergeEdit(groupName, parameterName,
                originalParameter, newParameter);
        support.beginUpdate();
        support.postEdit(edit);
        support.endUpdate();
    }
}
