package library;

import gov.usgs.ias.constant.Constants;
import gov.usgs.ias.dao.DataConnection;
import gov.usgs.ias.cps.CpsStatusEnum;
import gov.usgs.ias.dao.cpf.CpfSummaryDAO;
import gov.usgs.ias.dao.bpf.BpfSummaryDAO;
import gov.usgs.ias.dao.cpf.CpfDAO;
import gov.usgs.ias.dao.bpf.BpfDAO;
import gui.CpfViewer;
import gui.LoginDialog;
import java.sql.SQLException;
import java.awt.Point;
import org.apache.log4j.Logger;

/**
 * Provides interface to the various CPF and BPF data access objects, using
 * a common connection.  The schema to be accessed is specified at each login.
 * The only database operations used are those to modify the database --
 * updating status and inserting a file.  Queries are always handled through
 * the service.
 */
public class DatabaseLibrary  {
    private String currentAuth = null; // current authenticator (user)
    private LoadedFileSchema currentSchema = null; // current database
    private static DataConnection dataConn = null; // current data connection
    private static final String connStringProperty = "cpf.db.conn";
    private static final String dbHostProperty = "cpf.db.host";
    private static final String dbPortProperty = "cpf.db.port";
    private static final String dbSidProperty  = "cpf.db.sid";

    private Logger logger = Logger.getLogger(DatabaseLibrary.class.getName());
    private CpfViewer application;  // Parent application context
    
    /**
     * Create the library object and set up the connection.
     * @param application  the GUI application context
     * @throws LibraryException   if the connection property can't be determined
     */
    public DatabaseLibrary(CpfViewer application) throws LibraryException {
        this.application = application;

        // Set up the connection string (DSN) property
        setConnStringProperty();

        // Create the DataConnection that will be used by all DAOs
        //  no actual connection to the database is made yet; but
        //  this single DataConnection will be used by all logins
        dataConn = new DataConnection(connStringProperty);
    }

    /**
     * Set a connection string system property if not alreay set
     * @throws LibraryException   if the connection property can't be set
     */
    private void setConnStringProperty() throws LibraryException {
        
        if (System.getProperty(connStringProperty) == null) {
            String host = System.getProperty(dbHostProperty);
            String port = System.getProperty(dbPortProperty, "1521");

            // Try TWO_TASK, then ORACLE_SID, then property
            String sid = System.getenv("TWO_TASK");
            if (sid == null) sid = System.getenv("ORACLE_SID");
            if (sid == null) sid = System.getProperty(dbSidProperty);

            if (sid == null || host == null) {
                logger.error("Error setting up database connection, please " +
                        "verify database host and SID properties");
                System.clearProperty(connStringProperty);
                throw new LibraryException(
                    "Unable to set up database connection");
            }

            // sample connection string:
            //   jdbc:oracle:thin:authorization@myhost:1521:mysid
            // upon login, the authorization is replaced with the correct
            // username/password, e.g.:
            //   jdbc:oracle:thin:user/password@myhost:1521:mysid
            String connectionString = "jdbc:oracle:thin:authorization@" + host +
                    ":" + port + ":" + sid;
            logger.trace("Setting database connection string to " 
                    + connectionString);
            System.setProperty(connStringProperty, connectionString);
        }
    }

    /**
     * Log in to the specified schema in the database
     * @param  schema     the evaluation or production schema
     * @returns           whether login was attempted (false if user cancelled)
     * @throws LibraryException  if invalid schema or can't log in
     */
    public boolean login(LoadedFileSchema schema) throws LibraryException {
        if (schema == null) {
            currentAuth = null;
            throw new LibraryException("No schema specified");
        } else if (schema.equals(LoadedFileSchema.PROD)) {
            currentAuth = Constants.DAO_AUTH_IAS_PROD;
        } else if (schema.equals(LoadedFileSchema.EVAL)) {
            currentAuth = Constants.DAO_AUTH_IAS_EVAL;
        }

        // Create a data connection if none exists, or 
        //  use the existing one (close any previous connection)
        if (dataConn == null) {
            dataConn = new DataConnection(connStringProperty);
        } else {
            dataConn.close();
        }
        
        // Prompt for login if it's not in the environment or system properties
        if (System.getenv(currentAuth) == null &&
            System.getProperty(currentAuth) == null) {
            Point loc = application.getLocationOnScreen();
            loc.y += 50;
            loc.x += 50;
            LoginDialog loginDialog = new LoginDialog(application,
                    schema.toString(), currentAuth);
            loginDialog.setLocation(loc);
            loginDialog.setVisible(true);
            if (loginDialog.isCancelled()) {
                return false;
            }
        }
        // After prompting, check if it's set
        if (System.getenv(currentAuth) == null &&
            System.getProperty(currentAuth) == null) {
            throw new LibraryException("Could not get login/password from "
                    + currentAuth);
        } else {
            // Attempt to login, to be sure the auth is correct
            logger.trace("Connecting to database with " + currentAuth);
            try {
                dataConn.getConnection(currentAuth);
            } catch (SQLException ex) {
                System.clearProperty(currentAuth);
                // Get the root cause of the exception
                throw new LibraryException(ex.getCause().getMessage());
            }
            currentSchema = schema;
        }
        return true;
    }

    /**
     * Log out of the current database connection.
     */
    public void logout() {
        if (dataConn != null) dataConn.close();
        currentAuth = null;
    }

    /**
     * Update the status in the database of the given file
     * @param fileName           name of the file to update
     * @param type               type of file (CPF or BPF)
     * @param newStatus          the status to set it to
     * @throws LibraryException  if invalid file or type is provided, or can't
     *                           update the status
     */
    public void updateFileStatus(String fileName, LoadedFileType type,
            CpsStatusEnum newStatus) throws LibraryException {
        CpfSummaryDAO dao = null;
        if (type.equals(LoadedFileType.CPF)) {
            dao = new CpfSummaryDAO(dataConn, currentAuth);
        } else if (type.equals(LoadedFileType.BPF)) {
            dao = new BpfSummaryDAO(dataConn, currentAuth);
        }
        if (fileName == null || dao == null) {
            throw new LibraryException("Unknown type of file");
        }
        logger.info("Updating " + fileName + " status to " +
                newStatus.toString());
        try {
            dao.updateCpsStatus(fileName, newStatus.toString());
        } catch (Exception ex) {
            throw new LibraryException("Update failed for " + fileName, ex);
        }
    }

    /**
     * Insert a file into the database
     * @param schema      Which database to load the file into -- EVAL or PROD
     * @param lf          File currently loaded in the GUI
     * @throws LibraryException if the file couldn't be inserted
     */
    public void insertFile(LoadedFileSchema schema, LoadedFile lf)
            throws LibraryException {

        // EVAL files are loaded as ACTIVE; PROD files start out as INACTIVE
        // and must be activated after they are baselined
        CpsStatusEnum initialFileStatus = CpsStatusEnum.ACTIVE;
        if (schema.equals(LoadedFileSchema.PROD)) {
            initialFileStatus = CpsStatusEnum.INACTIVE;
        }

        CpfDAO dao = null;
        if (lf.getFileType().equals(LoadedFileType.CPF)) {
            dao = new CpfDAO(dataConn, currentAuth);
        } else if (lf.getFileType().equals(LoadedFileType.BPF)) {
            dao = new BpfDAO(dataConn, currentAuth);
        } else {
            throw new LibraryException("Unknown file type (not CPF or BPF)");
        }

        // Just go ahead & insert the file, no need to check if it already
        // exists -- the Export and Baseline functions do that already
        logger.info("Inserting " + lf.getFileName() + " into "
                + currentSchema.toString());
        try {
            dao.insertFile(lf.getOdlTree(), initialFileStatus);
        } catch (Exception ex) {
            throw new LibraryException("Insert failed for " + lf.getFileName(),
                ex);
        }
    }

    /**
     * Provide access to the data connection being used
     * @return  the current DataConnection
     */
    public DataConnection getDataConnection() {
        return dataConn;
    }

}
