package library;

/**
 * A record of a difference between two files.  One file is considered the 
 * source file and the other the destination file.  Used in generating a 
 * difference report.
 */
public class Difference {
    private String sourceValue;
    private String destValue;
    private String groupName;
    private String parameterName;
    private Integer location;
    private String comment;

    /**
     * Constructor
     * @param sourceValue   the value for this parameter in the source file
     * @param destValue     the value for this parameter in the destination file
     * @param groupName     the name of the group 
     * @param parameterName the name of the parameter
     * @param location      the location of the difference
     * @param comment       description of the difference, additional 
     *                      information for the user
     */
    public Difference(String sourceValue, String destValue, String groupName,
            String parameterName, Integer location, String comment) {
        this.sourceValue = sourceValue;
        this.destValue = destValue;
        this.groupName = groupName;
        this.parameterName = parameterName;
        this.location = location;
        this.comment = comment;
    }

    /**
     * Get the name of the group for this difference
     * @return  group name
     */
    public String getGroupName() {
        return groupName;
    }

    /**
     * Get the name of the parameter for this difference
     * @return  parameter name
     */
    public String getParameterName() {
        return parameterName;
    }

    /**
     * Provides a way to display a difference as a string
     * @return  string representation of the information in this object
     */
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        if (comment != null) {
            sb.append(comment + " ");
        }
        if (groupName != null) {
            sb.append("Group: " + groupName + " ");
        }
        if (parameterName != null) {
            sb.append("Parameter: " + parameterName + " ");
        }
        if (location != null) {
            sb.append("Value: " + location + " ");
        }
        if (sourceValue != null) {
            sb.append("Source File: " + sourceValue + " ");
        }
        if (destValue != null) {
            sb.append("Destination File: " + destValue + " ");
        }
        return sb.toString();
    }

    /**
     * Provides a way to display a difference as a row in a HTML table
     * @return  HTML representation of the information in this object formatted
     *          into a table row
     */
    public String toHtmlTableRow() {
        StringBuilder sb = new StringBuilder();
        sb.append("<tr>");

        sb.append("<td>");
        if (groupName != null) sb.append(groupName);
        sb.append("</td>");

        sb.append("<td>");
        if (parameterName != null) sb.append(parameterName);
        sb.append("</td>");

        sb.append("<td>");
        if (location != null) sb.append(location);
        sb.append("</td>");

        sb.append(sourceValue == null ?
            ("<td style=\"background-color: rgb(204, 204, 204);\">" +
            "missing"
            + "</td>") :
            ("<td>" + sourceValue + "</td>") );

        sb.append(destValue == null ?
            ("<td style=\"background-color: rgb(204, 204, 204);\">" +
            "missing"
            + "</td>") :
            ("<td>" + destValue + "</td>") );

        sb.append("<td>");
        if (comment != null) sb.append(comment);
        sb.append("</td>");

        sb.append("</tr>");
        return sb.toString();
    }
}
