package library;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.List;
import java.util.TreeMap;
import java.util.ArrayList;
import org.apache.log4j.Logger;

/**
 * Class to provide a detailed report of differences between two files.
 * We don't use the differencing information from the ParameterTreePanel and
 * ParameterTablePanels because (1) there may be more than one file open, the
 * the differences highlighted in the GUI account for all of them and (2) the
 * GUI only tracks the differences of the currently displayed parameter,
 * not all of them.
 */
public class FileCompare {

    LoadedFile sourceFile;
    LoadedFile destFile;
    String reportFileName = null;
    ArrayList<String> excludedGroups = null;
    ArrayList<Difference> differences = null;
    private final static Logger logger
            = Logger.getLogger(FileCompare.class.getName());

    /**
     * Create a comparator of the two specified files
     * @param sourceFile       the loaded file to use as the source for
     *                         comparison
     * @param destinationFile  the loaded file to compare to the source
     */
    public FileCompare(LoadedFile sourceFile, LoadedFile destinationFile) {
        this.sourceFile = sourceFile;
        this.destFile = destinationFile;
        this.excludedGroups = new ArrayList<String>();
    }

    /**
     * Set the output report file name
     * @param reportFileName   name of the report file to create
     */
    public void setReportFileName(String reportFileName) {
        this.reportFileName = reportFileName;
    }
    
    /**
     * Add a group to the list of group exclude from the diff
     * @param groupName 
     */
    public void excludeGroup(String groupName) {
        this.excludedGroups.add(groupName);
    }

    /**
     * Perform the comparison between the two files
     */
    public void compareFiles() {

        differences = new ArrayList<Difference>();

        // File 1 (source) and File 2 (destination) information
        List<String> f1Groups = sourceFile.getGroupNames();
        List<String> f2Groups = destFile.getGroupNames();
        List<String> f1Parms;
        List<String> f2Parms;
        List<String> f1ParmValues;
        List<String> f2ParmValues;

        // Do a comparision on all groups from file 1
        for (String groupName : f1Groups) {
            // skip excluded groups 
            if (this.excludedGroups.contains(groupName)) {
                if (f2Groups.contains(groupName))
                    f2Groups.remove(groupName);
                continue;
            }

            if (f2Groups.contains(groupName)) {
                // Compare the contents of the groups
                f1Parms = sourceFile.getParameterNames(groupName);
                f2Parms = destFile.getParameterNames(groupName);

                for (String parameterName : f1Parms) {
                    f1ParmValues = sourceFile.getParameterValues(groupName,
                            parameterName);

                    if (f2Parms.contains(parameterName)) {
                        f2ParmValues = destFile.getParameterValues(
                                groupName, parameterName);

                        // Compare the individual parameter values
                        int i = 0;
                        for (; i<f1ParmValues.size(); i++) {
                            // this value exists in file 1 but not file 2
                            if (i >= f2ParmValues.size()) {
                                differences.add(new Difference(
                                        f1ParmValues.get(i), null,
                                        groupName, parameterName, i+1,
                                         "Missing value"));

                            // the value exists in both -- check if they match
                            } else if (!f1ParmValues.get(i).toString().equals(
                                        f2ParmValues.get(i).toString())) {
                                // String comparison failed. If both values are
                                // numeric, check if numeric values are equal.
                                Boolean numericallyEqual = false;
                                try {
                                    Double f1Num = Double.valueOf(
                                            f1ParmValues.get(i).toString());
                                    Double f2Num = Double.valueOf(
                                            f2ParmValues.get(i).toString());
                                    numericallyEqual = f1Num.equals(f2Num);
                                } catch (NumberFormatException ex) {
                                    // If either value is not numeric, they
                                    // will not be equal.
                                }
                                if (!numericallyEqual) {
                                    differences.add(new Difference(
                                        f1ParmValues.get(i),
                                        f2ParmValues.get(i),
                                        groupName, parameterName, i+1, 
                                        "Differing values"));
                                }
                            }
                        }
                        // Report remaining values in file 2's parameter as
                        //  missing from file 1
                        for (; i<f2ParmValues.size(); i++) {
                                differences.add(new Difference(
                                        null, f2ParmValues.get(i),
                                        groupName, parameterName, i+1,
                                        "Missing value"));
                        }

                        f2Parms.remove(parameterName);

                    // This parameter is missing from file 2
                    } else {
                        differences.add(new Difference(
                                f1ParmValues.toString(), null,
                                groupName, parameterName,
                                null, "Entire parameter missing"));
                    }
                }
                // After processing all file 1's parameters, any that remain are
                // in f2 but not f1
                for (String parameterName : f2Parms) {
                    f2ParmValues = destFile.getParameterValues(groupName,
                            parameterName);
                    differences.add(new Difference(null,
                            f2ParmValues.toString(), groupName,
                            parameterName, null, "Entire parameter missing"));
                }

                f2Groups.remove(groupName);

            // This group is missing from file 2
            } else {
                differences.add(new Difference("", null, groupName, null,
                        null, "Entire group missing"));
            }
        }

        // After processing all file 1's groups, any that remain are in f2 but
        // not f1
        for (String groupName : f2Groups) {
            differences.add(new Difference(null, "", groupName, null,
                    null, "Entire group missing"));
        }
    }

    /**
     * Get a list of all the groups that had any differences
     * @return  array of group names
     */
    private String[] getGroups() {
        TreeMap<String,String> map = new TreeMap<String,String>();
        for (Difference d : differences) {
            if (d.getGroupName() != null)
                map.put(d.getGroupName(), null);
        }        
        return(map.keySet().toArray(new String[0]));
    }

    /**
     * Get a list of all the parameters that had any differences
     * @return  array of parameter names
     */
    private String[] getParameters() {
        TreeMap<String,String> map = new TreeMap<String,String>();
        for (Difference d : differences) {
            if (d.getParameterName() != null)
                map.put(d.getParameterName(), null);
        }
        return(map.keySet().toArray(new String[0]));
    }
    
    /**
     * Check if differences were found during comparison.
     * @return boolean 
     */
    public boolean differencesFound() {
        if (differences == null || differences.isEmpty()) {
            return false;
        }
        return true;
    }

    /**
     * Write a text report of the differences found.
     * @throws  IOException  if the report cannot be written
     */
    public void writeReport() throws IOException {
        logger.debug("Saving text comparison report " + reportFileName);
        BufferedWriter outputStream = null;
        try {
            outputStream = new BufferedWriter(new FileWriter(reportFileName));
            outputStream.write("Differences between\n " 
                + sourceFile.getFileName() + " and\n "
                + destFile.getFileName());
            outputStream.newLine();
            for (Difference d : differences) {
                outputStream.write(d.toString());
                outputStream.newLine();
            }
        } catch (IOException ex) {
            logger.error("Error writing " + reportFileName);
            throw ex;
        } finally {
            if (outputStream != null) {
                try {
                    outputStream.close();
                } catch (IOException ex) {
                    logger.error("Error completing " + reportFileName);
                    throw ex;
                }
            }
        }
    }

    /**
     * Write an HTML report of the differences found.
     * @throws  IOException  if the report cannot be written
     */
    public void writeHtmlReport() throws IOException {
        logger.debug("Saving html comparison report " + reportFileName);
        BufferedWriter outputStream = null;
        try {
            outputStream = new BufferedWriter(new FileWriter(reportFileName));
            outputStream.write("<html>");
            outputStream.newLine();
            outputStream.write("<head>");
            outputStream.newLine();
            outputStream.write("<title>CPF Difference Report</title>");
            outputStream.newLine();
            outputStream.write("</head>");
            outputStream.newLine();
            outputStream.write("<body>");
            outputStream.newLine();
            outputStream.write("<h1>Difference Report</h1>");
            outputStream.newLine();
            outputStream.write("Differences between the following files:<br>");
            outputStream.write("<ul>");
            outputStream.newLine();
            outputStream.write("<li>" + sourceFile.getFileName() + "</li>");
            outputStream.newLine();
            outputStream.write("<li>" + destFile.getFileName() + "</li>");
            outputStream.newLine();
            outputStream.write("</ul>");
            outputStream.newLine();
            outputStream.write("<h3>Summary</h3>");
            outputStream.newLine();
            outputStream.write("Affected Groups:<br>");
            outputStream.newLine();
            outputStream.write("<ul>");
            outputStream.newLine();
            for (String g : getGroups()) {
                outputStream.write("<li>" + g + "</li>");
                outputStream.newLine();
            }
            outputStream.write("</ul>");
            outputStream.newLine();
            outputStream.write("Affected Parameters:<br>");
            outputStream.newLine();
            outputStream.write("<ul>");
            outputStream.newLine();
            for (String g : getParameters()) {
                outputStream.write("<li>" + g + "</li>");
                outputStream.newLine();
            }
            outputStream.write("</ul>");
            outputStream.newLine();
            outputStream.write("<h3>Details</h3>");
            outputStream.newLine();
            outputStream.write("<table style=\"text-align: left; width: 100%;\""
                + " border=\"1\" cellpadding=\"2\" cellspacing=\"2\">");
            outputStream.newLine();
            outputStream.write("<tbody>");
            outputStream.newLine();
            outputStream.write("<tr>");
            outputStream.newLine();
            outputStream.write("<td style=\"vertical-align: top; font-weight: " 
                    + "bold;\">Group<br>");
            outputStream.newLine();
            outputStream.write("</td>");
            outputStream.newLine();
            outputStream.write("<td style=\"vertical-align: top; font-weight: " 
                    + "bold;\">Parameter<br>");
            outputStream.newLine();
            outputStream.write("</td>");
            outputStream.newLine();
            outputStream.write("<td style=\"vertical-align: top; font-weight: " 
                    + "bold;\">Element</td>");
            outputStream.newLine();
            outputStream.write("<td style=\"vertical-align: top; font-weight: " 
                    + "bold;\">" + sourceFile.getFileName() + "</td>");
            outputStream.newLine();
            outputStream.write("<td style=\"vertical-align: top; font-weight: " 
                    + "bold;\">" + destFile.getFileName() + "</td>");
            outputStream.newLine();
            outputStream.write("<td style=\"vertical-align: top; font-weight: "
                    + "bold;\">Comment</td>");
            outputStream.newLine();
            outputStream.write("</tr>");
            for (Difference d : differences) {
                outputStream.write(d.toHtmlTableRow());
            }
            outputStream.write("</tbody>");
            outputStream.write("</table>");
            outputStream.write("<br>");
            outputStream.write("</body>");
            outputStream.write("</html>");
            outputStream.close();
        } catch (IOException ex) {
            logger.error("Error writing " + reportFileName);
            throw ex;
        } finally {
            if (outputStream != null) {
                try {
                    outputStream.close();
                } catch (IOException ex) {
                    logger.error("Error completing " + reportFileName);
                    throw ex;
                }
            }
        }
    }
}
