package library;

import library.InstanceInfo;
import java.util.Calendar;
import java.util.regex.*;
import java.text.SimpleDateFormat;

/**
 * A class used to infer information from a file name, using the known Landsat
 * naming conventions.  Note this class is used to help the GUI make assumptions
 * about the file; it is not used to validate that the file name is following
 * the proper standards.  All methods are static.
 */
public class FileNamingConventions {
    // Patterns used to check a file name
    //  These patterns allow for inferring information from a file name.
    //  For L8, the validation script will ensure that the file name is actually
    //  correct before storing in the database.
    private static final String EVAL_SCHEMA_PATTERN = "^eval_.*";
    private static final String CPF_TYPE_PATTERN
            = "^(?:eval_)?L(?:\\D)?\\d\\dCPF.*";
    private static final String BPF_TYPE_PATTERN
            = "^(?:eval_)?L(?:\\D)\\dBPF.*";

    // Patterns to pick out pieces of a file name
    private static final String GET_SCHEMA_PATTERN =
            "^(eval_)?.*";
    private static final String GET_SATELLITE_PATTERN =
            "^(?:eval_)?L(?:\\D)?(\\d\\d).*";
    private static final String GET_FILE_TYPE_PATTERN =
            "^(?:eval_)?L(?:\\D)?\\d\\d?([BC]PF).*";
    private static final String GET_START_DATE_PATTERN =
            "^(?:eval_)?L(?:\\D)?\\d\\d?[BC]PF_?(\\d\\d\\d\\d)(\\d\\d)"
            + "(\\d\\d)(\\d\\d)?(\\d\\d)?(\\d\\d)?_.*";
    private static final String GET_END_DATE_PATTERN =
            "^(?:eval_)?L(?:\\D)?\\d\\d?[BC]PF_?\\d\\d\\d\\d\\d\\d\\d\\d"
            + "(?:\\d\\d)?(?:\\d\\d)?(?:\\d\\d)?_(\\d\\d\\d\\d)(\\d\\d)(\\d\\d)"
            + "(\\d\\d)?(\\d\\d)?(\\d\\d)?.*";
    private static final String GET_COLLECTION_PATTERN =
            "[^\\.]*_(\\d\\d)\\.(\\d\\d)$";
    private static final String GET_VERSION_PATTERN =
            "[^\\.]*\\.(\\d+)$";
    private static final String GET_SENSOR_PATTERN =
            "^(?:eval_)?L(\\D)(?:\\d\\d?).*";

    public static LoadedFileSchema getSchema(String fileName) {
        // Files named "eval_*" or not recognized as CPF/BPF are set to EVAL
        if (getFileType(fileName).equals(LoadedFileType.OTHER) ||
            fileName.matches(EVAL_SCHEMA_PATTERN)) {
            return LoadedFileSchema.EVAL;
        }
        return LoadedFileSchema.PROD;
    }

    public static int getSatellite(String fileName) {
        return extractInt(fileName, GET_SATELLITE_PATTERN);
    }

    public static String getSensor(String fileName) {
        Pattern pattern = Pattern.compile(GET_SENSOR_PATTERN);
        Matcher matcher = pattern.matcher(fileName);
        if (matcher.matches() && matcher.groupCount() >= 1) {
            return matcher.group(1);
        } else {
            return "";
        }
    }

    public static int getCollection(String fileName) {
        return extractInt(fileName, GET_COLLECTION_PATTERN);
    }

    public static int getVersion(String fileName) {
        return extractInt(fileName, GET_VERSION_PATTERN);
    }

    private static int extractInt(String fileName, String intPattern) {
        Pattern pattern = Pattern.compile(intPattern);
        Matcher matcher = pattern.matcher(fileName);
        int i=0;
        if (matcher.matches() && matcher.groupCount() >= 1) {
            i  = Integer.parseInt(matcher.group(1));
        }
        return i;
    }

    public static LoadedFileType getFileType(String fileName) {
        if (fileName.matches(CPF_TYPE_PATTERN)) {
            return LoadedFileType.CPF;
        } else if (fileName.matches(BPF_TYPE_PATTERN)) {
            return LoadedFileType.BPF;
        } else {
            return LoadedFileType.OTHER;
        }
    }

    public static Calendar getStartDate(String fileName) {
        return extractDate(fileName, GET_START_DATE_PATTERN);
    }

    public static Calendar getEndDate(String fileName) {
        return extractDate(fileName, GET_END_DATE_PATTERN);
    }

    private static Calendar extractDate(String fileName, String datePattern) {
        Pattern pattern = Pattern.compile(datePattern);
        Matcher matcher = pattern.matcher(fileName);
        int y, m, d;
        int hh=0, mm=0, ss=0;
        if (matcher.matches() && matcher.groupCount() >= 6) {
            y  = Integer.parseInt(matcher.group(1));
            m = Integer.parseInt(matcher.group(2)) - 1;
            d   = Integer.parseInt(matcher.group(3));
            if (matcher.group(4) != null) {
                hh  = Integer.parseInt(matcher.group(4));
            }
            if (matcher.group(5) != null) {
               mm = Integer.parseInt(matcher.group(5));
            }
            if (matcher.group(6) != null) {
                ss   = Integer.parseInt(matcher.group(6));
            }
            Calendar c = Calendar.getInstance();
            c.clear();
            c.set(y, m, d, hh , mm, ss);
            return c;
        } else {
            return null;
        }
    }

    /**
     * Construct a standard file name based on the metadata of the file
     * @param schema         schema of the file
     * @param satellite      satellite number
     * @param senor          sensor name
     * @param fileType       type of file (CPF or BPF)
     * @param startDate      starting effective date
     * @param endDate        ending effective date
     * @param version        file version
     * @return               file name
     */
    public static String constructFileName(LoadedFileSchema schema,
            int satellite, String sensor, LoadedFileType fileType,
            Calendar startDate, Calendar endDate,
            int version) {
        SimpleDateFormat sdf = (SimpleDateFormat)SimpleDateFormat.getInstance();
        if (fileType.equals(LoadedFileType.CPF )) {
            sdf.applyPattern("yyyyMMdd");
            String newFileName =
                (schema.equals(LoadedFileSchema.EVAL) ? "eval_" : "") +
                "L" + sensor +
                String.format("%02d", satellite) +
                fileType.toString() + "_" +
                sdf.format(startDate.getTime()) + "_" +
                sdf.format(endDate.getTime()) +
                String.format("_%02d.%02d", InstanceInfo.getCollection(),
                              version);
            return newFileName;
        } else if (fileType.equals(LoadedFileType.BPF )) {
            sdf.applyPattern("yyyyMMddHHmmss");
            String newFileName =
                (schema.equals(LoadedFileSchema.EVAL) ? "eval_" : "") +
                "L" +
                ((satellite == 8) ? sensor : "") +
                satellite + fileType.toString() +
                sdf.format(startDate.getTime()) + "_" +
                sdf.format(endDate.getTime()) + "." +
                String.format("%02d", version);
            return newFileName;
        } else {
            return null;
        }
    }

    /**
     * Check if the date range for a given filename has been subdivided
     * into smaller date ranges that exist in a list of files.  This is used
     * to check if the date range for a given file has been split, as evidenced
     * by existence of another file where one end of the date range matches and
     * the other doesn't, i.e.
     *      original date range   +-----------------------------------+
     *      divided date ranges   +---------------------++------------+
     * @param fileToCheck  name of file to check
     * @param files        array of existing file names and statuses
     * @return             boolean indicator whether a file exists that matches
     *                     the start or end of the given file
     */
    public static Boolean hasBeenDivided(String fileToCheck,
        Object[][] filesToCompare) {
        Boolean hasDivision = false;
        Calendar fileToCheckStart = getStartDate(fileToCheck);
        Calendar fileToCheckEnd = getEndDate(fileToCheck);

        for (int j = 0; j < filesToCompare.length; j++) {
            String fileCompare = (String)filesToCompare[j][0];
            if (fileToCheck.equals(fileCompare)) {
                continue;
            }
            Calendar fileCompareStart = getStartDate(fileCompare);
            Calendar fileCompareEnd = getEndDate(fileCompare);
            // see if it matches the start of the date range
            if (fileToCheckStart.equals(fileCompareStart)) {
                if (fileToCheckEnd.compareTo(fileCompareEnd) > 0) {
                    hasDivision = true;
                }
            }
            // see if it matches the end of the date range
            if (fileToCheckEnd.equals(fileCompareEnd)) {
                if (fileToCheckStart.compareTo(fileCompareStart) < 0) {
                    hasDivision = true;
                }
            }
        }
        return hasDivision;
    }
}
