package library;

import gov.nasa.pds.tools.label.*;
import gov.nasa.pds.tools.label.parser.LabelParser;
import java.util.*;
import java.io.*;
import java.net.URL;
import java.util.List;
import java.util.Set;
import java.util.ArrayList;
import java.util.HashSet;
import java.text.SimpleDateFormat;
import org.apache.log4j.Logger;
import workers.ThreadedLabelParserFactory;

/**
 * Keeps information about a file that is loaded in memory,
 * including access to the data contained in its ODL tree.
 * This class provides an interface to the functionality in the PDS Tools ODL
 * Parser.
 * <p>
 * <b>Note to the developer:</b>
 * The CPF GUI does not allow nested groups.  This is by convention and also
 * for technical reasons.  The LDCM CPF DFCB states that nested groups will not
 * be allowed.  The Java PDS Tools ODL Parser does not allow groups within
 * groups; nor do the CPF/BPF databases support nested groups.
 * If nested groups were to be used, changes would be required to these
 * components as well as the GUI.  The GUI classes most affected would be
 * LoadedFile and ParameterTreePanel (and others as well).
 */
public class LoadedFile implements Cloneable {
    private String fileName;          // full name of this file
    private LoadedFileType fileType;  // type of file
    private LoadedFileSchema schema;  // schema it came from
    private Label odlTree;            // parsed ODL tree of this file's contents
    private final static Logger logger
        = Logger.getLogger(LoadedFile.class.getName());

    // Max wait time in minutes for StreamReaderThread
    private final static int threadTimeout = 5;

    /**
     * Constructor used when the file has not been parsed yet
     * @param fileName  Name of the file.  Other properties of the file are set,
     *                  if possible, based on information inferred from the
     *                  file name.
     */
    public LoadedFile(String fileName) {
        File file = new File(fileName);
        String baseName = file.getName();
        this.schema = FileNamingConventions.getSchema(baseName);
        this.fileType = FileNamingConventions.getFileType(baseName);
        this.fileName = fileName;
        this.odlTree = null;
    }

    /**
     * Constructor used when the file has been parsed
     * @param fileName  Name of the file.  Other properties of the file are set,
     *                  if possible, based on information inferred from the
     *                  file name.
     */
    public LoadedFile(String fileName, Label label) {
        this(fileName);
        this.odlTree = label;
    }

    /**
     * Creates a LoadedFile object, including a parsed ODL tree, for the given
     * file specified by a URL
     * @param url                URL of the file to parse
     * @return                   a new LoadedFile object for the given file
     * @throws LibraryException  if the file can't be located or parsed         
     */
    static public LoadedFile create(URL url) throws LibraryException {
        String fileName = getFileNameFromUrl(url);
        if (fileName == null) {
            throw new LibraryException("Can't extract file name from URL "
                + url.getFile());
        }

        logger.debug("Parsing " + url.getFile());
        Label aLabel = parse(url);

        // Create a LoadedFile so we can use the DatabaseLibrary insert method.
        LoadedFile instance = new LoadedFile(fileName, aLabel);
        return instance;
    }

    /**
     * Extract the file name out of a URL.  If the URL contains query
     * parameters, get the file name from the "filename" query parameter.
     * Otherwise it will just be the file name portion of the URL.
     * @param url        URL of the REST service's query or local file name
     * @return           the file name (without path)
     */
    public static String getFileNameFromUrl(URL url) {
        if (url.getQuery() == null) {
            return url.getFile();
        } else {
            String[] params = url.getQuery().split("&");
            for (String param : params) {
                String name = param.split("=")[0];
                String value = param.split("=")[1];
                if (name.equals("filename")) {
                    return value;
                }
            }
            return null;
        }
    }

    /**
     * Method that performs ODL parsing and error checking
     * @param url                 URL of the file to parse
     * @return                    a parsed ODL Label
     * @throws LibraryException   if the file could not be parsed or had 
     *                            duplicate groups
     */
    public static Label parse(URL url) throws LibraryException {

        Label aLabel = null;
        try {
            LabelParser lp = ThreadedLabelParserFactory.getInstance().
                newThreadedLabelParser(true);
            aLabel = lp.parse(url);
        } catch (Exception e) {
            throw new LibraryException("Can't parse " + url.getFile(), e);
        }


        // Check the status of the PDS Tools ODL Parser
        if (aLabel == null
            || aLabel.getStatus().equals(
            gov.nasa.pds.tools.label.validate.Status.FAIL)) {
            logger.error("Error parsing " + url.getFile());
            throw new LibraryException("Error parsing " + url.getFile()
                + " - Please see log for details");
        }

        // Check if duplicate groups exist
        List<String> allGroupNames = new ArrayList<String>();
        for (Object o: aLabel.getGroups()) {
            GroupStatement gs = (GroupStatement)o;
            allGroupNames.add(gs.getIdentifier());
        }
        Set<String> uniqueGroupNames = new HashSet<String>(allGroupNames);
        if (allGroupNames.size() != uniqueGroupNames.size()) {
            // Remove 1 of each unique group name from the "all" list,
            // remaining group names are duplicates
            for (String g : uniqueGroupNames.toArray(new String[0])) {
                allGroupNames.remove(g);
            }
            throw new LibraryException("Error parsing: "
                + "file contains duplicate groups " + allGroupNames.toString());
        }

        return aLabel;
    }

    /**
     * Defines how we determine if 2 loaded files are equal -- having the
     * same file name.
     * @param obj      the second LoadedFile to compare to
     * @return         true if the file names are equal; false otherwise
     */
    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        } else if (this.getClass().equals(obj.getClass())) {
            return (this.fileName.equals(((LoadedFile)obj).fileName));
        } else {
            return super.equals(obj);
        }
    }

    /*
     * Returns the hash code for this object, based on the file name.  Required
     * to ensure two equal objects (as defined above) have the same hash code.
     */
    @Override
    public int hashCode() {
        int hash = 5;
        hash = 19 * hash + (this.fileName != null ?
            this.fileName.hashCode() : 0);
        return hash;
    }

    /**
     * Create a clone of this loaded file, including a deep copy of the
     * ODL tree.
     */
    @Override
    public Object clone() {
        LoadedFile newLf = null;
        try {
            newLf = (LoadedFile) super.clone();
        } catch (CloneNotSupportedException e) {
            // This should never happen
            logger.error("Unable to clone LoadedFile " + fileName);
            return null;
        }
        if (odlTree != null) {
            newLf.odlTree = (Label) odlTree.clone();
        }
        return newLf;
    }

// Routines to extract information from the ODL tree ---------------------------
    /**
     * Get the names of the groups contained in this file
     * @return List of String group names
     */
    public List<String> getGroupNames() {
        if (odlTree == null) {
            return null;
        }
        ArrayList<String> groupNameList = new ArrayList<String>();

        List lg = odlTree.getGroups();
        Iterator gi = lg.iterator();
        GroupStatement gs = null;
        while ( gi.hasNext() ) {
            gs = (GroupStatement)gi.next();
            groupNameList.add(gs.getIdentifier());
        }

        return groupNameList;
    }

    /**
     * Get the names of the parameters contained within a given group
     * @param groupName    the name of the group
     * @return             a list of string parameter names
     */
    public List<String> getParameterNames(String groupName) {

        if (odlTree == null) {
            return null;
        }
        List groups = odlTree.getGroups(groupName);
        if (groups == null) {
            return null;
        }

        ArrayList<String> attributeList = new ArrayList<String>();
        Iterator groupIterator = groups.iterator();
        while ( groupIterator.hasNext() ) {
            GroupStatement gs = (GroupStatement)groupIterator.next();
            List attributesInGroup = gs.getAttributes();
            for (Object o : attributesInGroup) {
                AttributeStatement as = (AttributeStatement)o;
                attributeList.add(as.getIdentifier());
            }
        }

        return attributeList;
    }

    /**
     * Get the values of the specified group/parameter
     * @param groupName        the name of the group
     * @param parameterName    the name of the parameter within groupName
     * @return                 a list of string parameter values
     */
    public List<String> getParameterValues(String groupName,
            String parameterName) {

        if (odlTree == null) {
            return null;
        }
        ArrayList<String> attribValueList = new ArrayList<String>();

        List groups = odlTree.getGroups(groupName);
        if (groups == null || groups.size() <= 0) {
            return attribValueList;
        }
        // Find the group that has this parameter
        AttributeStatement as = null;
        for (Object o: groups) {
            GroupStatement gs = (GroupStatement)o;
            as = gs.getAttribute(parameterName);
            if (as != null) {
                break;
            }
        }
        if (as == null) {
            return attribValueList;
        }

        Scalar scalarValue;
        if (as.getValue() instanceof gov.nasa.pds.tools.label.Sequence) {
            Sequence seq = (Sequence) as.getValue();
            Iterator si = seq.iterator();
            while (si.hasNext()) {
                scalarValue = (Scalar) si.next();
                attribValueList.add(scalarValue.getValue());
            }
        } else if (as.getValue() instanceof gov.nasa.pds.tools.label.Set) {
            gov.nasa.pds.tools.label.Set set
                    = (gov.nasa.pds.tools.label.Set) as.getValue();
            Iterator si = set.iterator();
            while (si.hasNext()) {
                scalarValue = (Scalar) si.next();
                attribValueList.add(scalarValue.getValue());
            }
        } else {
            scalarValue = (Scalar) as.getValue();
            attribValueList.add(scalarValue.getValue());
        }

        return attribValueList;
    }

    /**
     * Save the file using the current name.
     * @return   true if successfully written; false if error
     */
    public boolean saveFile() {
        return saveFile(this.fileName);
    }

    /**
     * Save the contents of the ODL tree in a file of the given name
     * @param newFileName  Path & file name to write to
     * @return             true if successfully written; false if error
     */
    private boolean saveFile(String newFileName) {
        logger.debug("Saving file named " + newFileName);
        BufferedWriter outputStream = null;
        try {
            outputStream = new BufferedWriter(new FileWriter(newFileName));
            if (odlTree != null) {
                outputStream.write(odlTree.toOdlString());
            }
            outputStream.close();
        } catch (IOException ex) {
            logger.error("Error writing " + newFileName);
            return false;
        } finally {
            if (outputStream != null) {
                try {
                    outputStream.close();
                } catch (IOException ex) {
                    // if the stream can't be closed, there is not much recourse
                    logger.warn("Error closing output stream for "
                        + newFileName);
                }
            }
        }
        return true;
    }

    /**
     * Validate the contents of the ODL tree using the configured validation
     * script.
     * @throws LibraryException  If the validation fails
     */
    public void validateFile() throws LibraryException {
        logger.debug("Validating " + fileName);
        if (odlTree == null) {
            throw new LibraryException(fileName + " is an empty file");
        }

        // Get the version number & type of file from the file name
        File file = new File(fileName);
        String baseName = file.getName();
        // Get the base filename (sans the dot and version) and set up
        // a pattern for matching other files with the same base filename
        String baseMatchPattern = baseName.split("\\.")[0] + "\\..*";

        // Get the listing of existing files
        Object[][] existingFiles = null;
        boolean thisFileAlreadyExists = false;
        Services services = new Services();
        if (fileType.equals(LoadedFileType.CPF)) {
            existingFiles = services.getCpfFileNameList(
                    FileNamingConventions.getSchema(baseName), false, null,
                    null, null);
        } else if (fileType.equals(LoadedFileType.BPF)) {
            existingFiles = services.getBpfFileNameList(
                    FileNamingConventions.getSchema(baseName),
                    null, false, null, null, null);
        } else {
            throw new LibraryException("Unknown file type (not CPF or BPF) - " +
                    " could not be inferred from file name " + baseName);
        }
        int latestExistingVersion = -1;
        if (existingFiles != null) {            
            // See if any files match this one
            for (int i=0; i < existingFiles.length; i++) {
                String existingFile = existingFiles[i][0].toString();
                if (existingFile.matches(baseMatchPattern)) {
                    int thisMatchVersion = FileNamingConventions.getVersion(
                            existingFile);
                    if (thisMatchVersion > latestExistingVersion) {
                        latestExistingVersion = thisMatchVersion;
                    }
                }
                if (existingFile.equals(baseName)) {
                    thisFileAlreadyExists = true;
                }
            }
        }

        if (thisFileAlreadyExists) {
            throw new LibraryException(baseName
                    + " already exists in the database."
                    + " The latest version on this date range is "
                    + latestExistingVersion + ". To export, update the"
                    + " version on this file.");
        }

        // Locate the validation script
        String validationScript = System.getProperty("validation",
                "cps_validate.pl");
        File validationScriptFile = new File(validationScript);
        if (!validationScriptFile.canExecute()) { 
            // not in the configured location, try in $IAS_BIN
            logger.debug("Cannot execute configured validation script "
                + validationScriptFile.getPath() + ", looking for script "
                + validationScriptFile.getName() + " in $IAS_BIN");
            validationScriptFile = new File(System.getenv("IAS_BIN")
                + File.separator + validationScriptFile.getName());
        }
        logger.debug("Calling validation script " 
            + validationScriptFile.getPath());
        if (!validationScriptFile.canExecute()) {
            throw new LibraryException("Unable to execute validation script "
                + validationScriptFile.getPath());
        }
        try {
            validationScript = validationScriptFile.getCanonicalPath();
        } catch (IOException ex) {
            throw new LibraryException("Unable to locate validation script: " +
                    ex.getMessage());
        }

        // Retrieve the file name from file attributes. 
        List<String> paramValues = getParameterValues(
            System.getProperty("cpf.group.name.file_attributes", 
                               "FILE_ATTRIBUTES"), 
            System.getProperty("cpf.parameter.name.file_attributes.file_name", 
                               "File_Name"));
        
        // Verify that a filename was found
        if (paramValues.size() == 0)
        {
            throw new LibraryException("File_Name value not found");
        }
    
        String cpf_filename = paramValues.get(0);

        // Call the script with the correct command line arguments, e.g.,
        // cps_validate.pl --e --no-warnings --no-format --t type --s filename
        //       --e[rrors] - Enable error messages
        //       --no-warnings - Suppress warning messages
        //       --no-format - Suppress format mismatches
        //       --t[ype] <TYPE> - File type
        //       --s[tandardIn] - Read file from standard in
        // feeding the ODL into stdin
        String fileTypeStr = fileType.toString();
        String sensor = FileNamingConventions.getSensor(baseName);
        ProcessBuilder launchValidation = new ProcessBuilder(validationScript,
            "--e", "--no-warnings", "--no-format", "--t",
            // Don't pass sensor for CPFs 
            (fileTypeStr.compareTo("CPF") == 0 ? "" : sensor ) +
            fileTypeStr, 
            "--s", cpf_filename);
        launchValidation.redirectErrorStream(true); // merge stdout & stderr
        Process validationProcess = null;
        try {
            logger.debug("Running " + launchValidation.command());
            validationProcess = launchValidation.start();
        } catch (IOException ex) {
            throw new LibraryException("Unable to invoke validation script " +
                    validationScript + " : " + ex.getMessage());
        }        
        // Set up streams for writing to the script's stdin
        BufferedOutputStream bufferOut = new BufferedOutputStream(
                validationProcess.getOutputStream());
        PrintWriter commandInput = new PrintWriter(
                (new OutputStreamWriter(bufferOut)), true);
        
        // start a separate thread to read from standard out/err
        StreamReaderThread validationOutputThread = 
            new StreamReaderThread(validationProcess.getInputStream());
        validationOutputThread.start();

        // Send the file's ODL contents to the script & log output
        try {
            commandInput.println(odlTree.toOdlString());
        } catch (Exception e) {
            throw new LibraryException("Error executing validation script: " +
                    e.getMessage());
        }

        // wait for the process to finish
        try {
            validationProcess.waitFor();
        } catch (InterruptedException e) {
            throw new LibraryException("Validation was interrupted: " +
                    e.getMessage());
        }

        // wait to finish reading from standard out/err 
        try {
            validationOutputThread.join(threadTimeout * 60000);
        } catch (InterruptedException ex) {
            throw new LibraryException("Validation output reader was " +
                "interrupted: " + ex.getMessage());
        }

        // check if any errors occurred will reading output from script
        if (validationOutputThread.errorReadingStream()) {
            throw new LibraryException("Error reading output from validation" +
                " script: " + validationOutputThread.getErrorMessage());
        }

        // write standand out/err output to the log
        logger.trace("Validation Script output:");
        if (validationProcess.exitValue() != 0) {
            ArrayList<String> output = validationOutputThread.getOuput();
            for (String s : output) {
                logger.error(s);
            }
        }

        // exit of 0 indicates successful validation
        logger.debug("Validation script exit status: "
            + validationProcess.exitValue());
        if (validationProcess.exitValue() != 0) {
            throw new LibraryException("Validation failed for " + baseName +
                    ". Please check log for validation errors");
        }
        logger.debug(baseName + " passed validation");
    }

    /**
     * Remove the value at a specific location from an attribute (parameter)
     * @param as       parsed attribute statement for this parameter
     * @param index    index of the value to remove
     */
    private void removeValue(AttributeStatement as, int index) {

        if (as.getValue() instanceof gov.nasa.pds.tools.label.Sequence) {
            Sequence seq = (Sequence)as.getValue();
            if (index < seq.size()) {
                logger.debug("Removing value " + index + " of "
                    + as.getIdentifier());
                seq.remove(index);
            } else {
                // nothing at the specified index, so nothing to remove
                logger.warn("There is no " + index + " item to remove from "
                    + as.getIdentifier());
            }
        } else {
            // Since a Set is an unordered collection, removal of an index value
            // doesn't make sense - and we don't use Sets anyway.
            // A Scalar value cannot be removed - this would leave a dangling
            // parameter with no value.  This should not happen with the current
            // GUI capabilities anyway (see
            // ParameterTableModelListener.tableChanged).
            logger.warn("Unable to remove item " + index + " of "
                + as.getIdentifier() + " since it is a "
                + as.getValue().getClass().getName());
        }
    }

/* Routines to modify the ODL tree -------------------------------------------*/

    /**
     * Update one of the values of a parameter.  A parameter may contain a
     * single value or an array of values.  An existing value may be updated
     * or the parameter may be converted from a single value to an array (i.e.,
     * values added to it) by updating the value at an index other than 0.
     * A single value within an array may be removed (i.e., the array is shrunk)
     * by providing null for the newValue.  The value of a scalar parameter will
     * not be removed in this way -- the caller should call removeParameter
     * instead.
     * @param groupName        the group containing the parameter
     * @param parameterName    the name of the parameter to update
     * @param newValue         the new value to add to the parameter
     * @param index            the index location of the new value
     */
    public void updateValue(String groupName, String parameterName, 
        String newValue, int index) {
        if (odlTree == null) {
            return;
        }
        logger.debug(" Updating ODL tree of " + fileName + ", setting " +
                groupName + " " + parameterName + " item " + index +
                " to new value " + newValue);

        List groups = odlTree.getGroups(groupName);
        if (groups == null || groups.size() <= 0) {
            return;
        }

        // Find the group that has this parameter
        AttributeStatement as = null;
        for (Object o: groups) {
            GroupStatement gs = (GroupStatement)o;
            as = gs.getAttribute(parameterName);
            if (as != null) {
                break;
            }
        }
        if (as == null) {
            return;
        }

        // Updating a value to null means removing it
        if (newValue == null) {
            removeValue(as, index);
            return;
        }

        Value newLabelValue = null;
        try { // if the value checks out as a double, set it as numeric
            Double.parseDouble(newValue);
            newLabelValue = new Numeric(newValue);
        } catch (NumberFormatException nfe) {
            // otherwise, text
            newLabelValue = new TextString(newValue);
        }

        if (as.getValue() instanceof gov.nasa.pds.tools.label.Sequence) {
            Sequence seq = (Sequence) as.getValue();
            if (index < seq.size()) {
                seq.set(index, newLabelValue);
            } else {
                seq.add(index, newLabelValue);
            }
        } else if (as.getValue() instanceof gov.nasa.pds.tools.label.Set) {
            // Since a Set is an unordered collection, update of an index value
            // doesn't make sense.  We don't normally use Sets anyway.
            logger.warn("Unable to update item " + index + " of "+ groupName
                    + "/" + parameterName + " since it is a SET");
        } else {
            if (index > 0) { // change the scalar into a Sequence and add the
                             // new value to the existing one
                Scalar existing = (Scalar)as.getValue();
                Sequence seq = new Sequence();
                seq.add(existing);
                seq.add(index, newLabelValue);
                as.setValue(seq);
            } else { // change the existing value
                as.setValue(newLabelValue);
            }
        }
    }

   /**
    * Update the file name and file attributes using information gleaned from
    * the file name (start & end effective dates)
    * Since the file naming convention for CPFs includes date but not time,
    * the time portion of the start & end dates will be taken from the current 
    * start & end effective dates contained in the file.
    * @param newFileName  the new file name to assign
    */
    public void updateFileName(String newFileName) {
        logger.debug("Renaming " + fileName + " to " + newFileName);
        File file = new File(newFileName);
        String baseName = file.getName();
        // intialize dates to the current file's start & end date
        Calendar sdt = FileNamingConventions.getStartDate(baseName);
        Calendar edt = FileNamingConventions.getEndDate(baseName);

        if (getFileType().equals(LoadedFileType.CPF)) {
            // fill in the time portions of the effective dates since we can't
            // get that from the filename
            Calendar originalDateTime = Calendar.getInstance();
            SimpleDateFormat dateFormatter =
                new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
            String groupName = System.getProperty(
                "cpf.group.name.file_attributes", "FILE_ATTRIBUTES");
            String parameterName = System.getProperty(
                "cpf.parameter.name.file_attributes.begin_date",
                "Effective_Date_Begin");
            List<String> originalValues = getParameterValues(groupName,
                parameterName);
            try {
                originalDateTime.setTime(
                    dateFormatter.parse(originalValues.get(0)));
                logger.debug("Setting start to original time "
                    + originalDateTime.getTime());
                sdt.set(Calendar.HOUR_OF_DAY,
                    originalDateTime.get(Calendar.HOUR_OF_DAY));
                sdt.set(Calendar.MINUTE,
                    originalDateTime.get(Calendar.MINUTE));
                sdt.set(Calendar.SECOND,
                    originalDateTime.get(Calendar.SECOND));
            } catch (Exception ex) {
                // if we can't parse the original date, set the time to 00:00:00
                if (sdt != null) {
                    logger.warn("Setting start time to 00:00:00");
                    sdt.set(Calendar.HOUR_OF_DAY, 0);
                    sdt.set(Calendar.MINUTE, 0);
                    sdt.set(Calendar.SECOND, 0);
                }
            }

            parameterName = System.getProperty(
                "cpf.parameter.name.file_attributes.end_date",
                "Effective_Date_End");
            originalValues = getParameterValues(groupName, parameterName);
            try {
                originalDateTime.setTime(
                    dateFormatter.parse(originalValues.get(0)));
                logger.debug("Setting end to original time "
                    + originalDateTime.getTime());
                edt.set(Calendar.HOUR_OF_DAY,
                    originalDateTime.get(Calendar.HOUR_OF_DAY));
                edt.set(Calendar.MINUTE,
                    originalDateTime.get(Calendar.MINUTE));
                edt.set(Calendar.SECOND,
                    originalDateTime.get(Calendar.SECOND));
            } catch (Exception ex) {
                // if we can't parse the original date, set the time to 23:59:59
                if (edt != null) {
                    logger.warn("Setting end time to 23:59:59");
                    edt.set(Calendar.HOUR_OF_DAY, 23);
                    edt.set(Calendar.MINUTE, 59);
                    edt.set(Calendar.SECOND, 59);
                }
            }
        }

        updateFileName(newFileName, sdt, edt);
    }

    /**
     * Update the file name and certain attributes based on new effective dates.
     * The file name is updated to the provided name, and the effective dates
     * are updated if provided.
     * Note that renaming a file, with the corresponding file content changes,
     * is not undoable.
     * @param newFileName new name of the file
     * @param startDate new effective start date/time (null to leave as is)
     * @param endDate new effective end date/time (null to leave as is)
     */
    public void updateFileName(String newFileName,
        Calendar startDate, Calendar endDate) {

        if (newFileName == null) {
            return;
        }
        setFileName(newFileName);

        // In addition to setting the filename, update the corresponding
        //  file type and schema
        File file = new File(newFileName);
        String baseName = file.getName();
        setFileType(FileNamingConventions.getFileType(baseName));
        setFileSchema(FileNamingConventions.getSchema(baseName));

        if (startDate != null) {
            logger.debug("Updating " + fileName + " start date to "
                + startDate.getTime().toString());
        }
        if (endDate != null) {
            logger.debug("Updating " + fileName + " end date to "
                + endDate.getTime().toString());
        }

        // Also update specific ODL fields based on the new name & dates
        String groupName;
        String parameterName;
        if (odlTree != null) {
            if (getFileType().equals(LoadedFileType.CPF)) {

                // Update the file name
                groupName = System.getProperty("cpf.group.name.file_attributes",
                    "FILE_ATTRIBUTES");
                parameterName = System.getProperty(
                    "cpf.parameter.name.file_attributes.file_name",
                    "File_Name");
                updateValue(groupName, parameterName, baseName, 0);

                // Update the version
                parameterName = System.getProperty(
                    "cpf.parameter.name.file_attributes.version", "Version");
                updateValue(groupName, parameterName,
                    "" + FileNamingConventions.getVersion(baseName), 0);

                // Update the collection number
                parameterName = System.getProperty(
                    "cpf.parameter.name.file_attributes.collection",
                     "Collection_Number");
                updateValue(groupName, parameterName,
                    "" + FileNamingConventions.getCollection(baseName), 0);

                // Update the begin effective date
                if (startDate != null) {
                    parameterName = System.getProperty(
                        "cpf.parameter.name.file_attributes.begin_date",
                        "Effective_Date_Begin");
                    updateValue(groupName, parameterName,
                        String.format("%04d-%02d-%02dT%02d:%02d:%02d",
                        startDate.get(Calendar.YEAR),
                        (startDate.get(Calendar.MONTH) + 1),
                        startDate.get(Calendar.DAY_OF_MONTH),
                        startDate.get(Calendar.HOUR_OF_DAY),
                        startDate.get(Calendar.MINUTE),
                        startDate.get(Calendar.SECOND)), 0);
                }

                // Update the end effective date
                if (endDate != null) {
                    parameterName = System.getProperty(
                        "cpf.parameter.name.file_attributes.end_date",
                        "Effective_Date_End");
                    updateValue(groupName, parameterName,
                        String.format("%04d-%02d-%02dT%02d:%02d:%02d",
                        endDate.get(Calendar.YEAR),
                        (endDate.get(Calendar.MONTH) + 1),
                        endDate.get(Calendar.DAY_OF_MONTH),
                        endDate.get(Calendar.HOUR_OF_DAY),
                        endDate.get(Calendar.MINUTE),
                        endDate.get(Calendar.SECOND)), 0);
                }

            } else if (getFileType().equals(LoadedFileType.BPF)) {

                // Update file name
                groupName = System.getProperty("bpf.group.name.file_attributes",
                    "FILE_ATTRIBUTES");
                parameterName = System.getProperty(
                    "bpf.parameter.name.file_attributes.file_name",
                    "File_Name");
                updateValue(groupName, parameterName, baseName, 0);

                // Update version
                parameterName = System.getProperty(
                    "bpf.parameter.name.file_attributes.version", "Version");
                updateValue(groupName, parameterName,
                    "" + FileNamingConventions.getVersion(baseName), 0);

                // Update start date
                if (startDate != null) {
                    parameterName = System.getProperty(
                        "bpf.parameter.name.file_attributes.begin_date",
                        "Effective_Date_Begin");
                    updateValue(groupName, parameterName,
                        String.format("%04d-%02d-%02dT%02d:%02d:%02d",
                        startDate.get(Calendar.YEAR),
                        (startDate.get(Calendar.MONTH) + 1),
                        startDate.get(Calendar.DAY_OF_MONTH),
                        startDate.get(Calendar.HOUR_OF_DAY),
                        startDate.get(Calendar.MINUTE),
                        startDate.get(Calendar.SECOND)), 0);
                }

                // Update end date
                if (endDate != null) {
                    parameterName = System.getProperty(
                        "bpf.parameter.name.file_attributes.end_date",
                        "Effective_Date_End");
                    updateValue(groupName, parameterName,
                        String.format("%04d-%02d-%02dT%02d:%02d:%02d",
                        endDate.get(Calendar.YEAR),
                        (endDate.get(Calendar.MONTH) + 1),
                        endDate.get(Calendar.DAY_OF_MONTH),
                        endDate.get(Calendar.HOUR_OF_DAY),
                        endDate.get(Calendar.MINUTE),
                        endDate.get(Calendar.SECOND)), 0);
                }
            }
        }
    }

    // Methods to add or remove a group
    /**
     * Add a group to the ODL tree at the specified index location
     * @param groupName  name of the new group
     * @param index      0-relative location
     * @return           the newly added group statement
     */
    public GroupStatement addGroup(String groupName, int index,
        GroupStatement contents) {
        if (odlTree == null) {
            return null;
        }
        logger.debug("Updating ODL tree for " + fileName + ", adding group "
            + groupName + " " + " at position " + index);

        List lg = odlTree.getGroups(groupName);
        if (lg != null && lg.size() > 0) {
            logger.warn("Attempting to add " + groupName
                + " : Group " + groupName + " already exists in "
                + fileName);
            return null;
        }
        GroupStatement newGroup;
        if (contents == null) {
            newGroup = new GroupStatement(1, groupName);
        } else {
            newGroup = contents;
        }
        odlTree.addStatement(newGroup, index);
        return newGroup;
    }

    /**
     * Remove a group from the ODL tree
     * @param groupName  name of the group to remove
     */
    public void removeGroup(String groupName) {
        if (odlTree == null) {
            return;
        }
        logger.debug("Updating ODL tree for " + fileName + ", removing group "
            + groupName);
        odlTree.removeStatement(new GroupStatement(1, groupName));
    }

    // Methods to add or remove a parameter
    /**
     * Add a parameter to an existing group at the given index location.
     * Initial value(s) for the parameter must also be supplied.
     * @param groupName      name of the group to which to add the parameter
     * @param parameterName  name of the parameter to add
     * @param index          location within the group to place the parameter
     *                       (0-based)
     * @param initialValues  initial values to assign to the parameter
     */
    public void addParameter(String groupName, String parameterName, int index,
        String[] initialValues) {
        if (odlTree == null) {
            return;
        }
        logger.debug("Updating ODL tree for " + fileName + ", adding parameter "
            + parameterName + " " + " at position " + index);

        List lg = odlTree.getGroups(groupName);
        if (lg == null || lg.size() <= 0) {
            logger.warn("Attempting to add " + groupName + "/"
                + parameterName + " : Group " + groupName + " not found in "
                + fileName);
            return;
        }
        if (getParameter(groupName, parameterName) != null) {
            logger.warn("Attempting to add " + groupName + "/"
                + parameterName + " : Parameter already exists in "
                + fileName);
            return;
        }
        // Add the parameter to the first group with this name
        GroupStatement gs = (GroupStatement)lg.get(0);
        Value attributeValue;
        // Create either a sequence (if >1 value was supplied) or a scalar
        if (initialValues.length > 1) {  // multiple values -- sequence
            ArrayList<Object> valList
                = new ArrayList<Object>(initialValues.length);
            for (String v : initialValues) {
                try {
                    Double.parseDouble(v);
                    valList.add(new Numeric(v));
                } catch (NumberFormatException nfe) {
                    valList.add(new TextString(v));
                }
            }
            attributeValue = new Sequence(valList);
        } else {  // single value -- create a scalar of the correct type
            try {
                Double.parseDouble(initialValues[0]);
                attributeValue = new Numeric(initialValues[0]);
            } catch (NumberFormatException nfe) {
                attributeValue = new TextString(initialValues[0]);
            }
        }
        AttributeStatement as = new AttributeStatement(parameterName,
            attributeValue);
        gs.addStatement(as, index);
    }

    /**
     * Remove a parameter from an existing group
     * @param groupName      name of the group containing the parameter
     * @param parameterName  name of the parameter to remove
     */
    public void removeParameter(String groupName, String parameterName) {
        if (odlTree == null) {
            return;            
        }
        logger.debug("Updating ODL tree for " + fileName 
            + ", removing group " + groupName
            + "'s parameter " + parameterName);
        List groups = odlTree.getGroups(groupName);
        if (groups == null || groups.size() <= 0) {
            logger.warn("Attempting to remove " + groupName + "/"
                + parameterName + " : Group " + groupName + " not found in "
                + fileName);
            return;
        }
        // Find the group that has this parameter
        GroupStatement gs = null;
        for (Object o: groups) {
            GroupStatement checkGroup = (GroupStatement)o;
            if (checkGroup.getAttribute(parameterName) != null) {
                gs = checkGroup;
            }
        }
        if (gs != null) {
            gs.removeStatement(new AttributeStatement(parameterName));
        }
    }

    /* Methods to merge a group or parameter ---------------------------------*/
    /**
     * Merge the contents of a new group into the named group of this file.
     * A Group merge means any Parameter that exists in the new group will be
     * added to the existing group (if it doesn't exist) or will replace it
     * in the existing group (if it already exists)
     * @param groupName      name of the group being merged into
     * @param groupToMerge   new group whose contents will be merged into the
     *                       existing group
     */
    public void mergeGroup(String groupName, GroupStatement groupToMerge) {
        if (groupToMerge == null) {
            logger.warn("Cannot merge a null group for " + groupName);
            return;
        }
        if (odlTree == null) {
            return;
        }
        logger.debug("Merging " + groupName + " into " + fileName);

        // If this file doesn't have the group, add it at the end
        List lg = odlTree.getGroups(groupName);
        GroupStatement gs;
        if (lg == null || lg.size() <= 0) {
            gs = new GroupStatement(1, groupName);
            odlTree.addStatement(gs);
        } else {
            // Merge into the 1st group of this name (normally there is only 1)
            gs = (GroupStatement)lg.get(0);
        }

        // Add each attribute within the group to the new group
        //  Make a clone of each value or attribute so we have our own copy
        AttributeStatement as = null;
        List la = groupToMerge.getAttributes();
        for (Object o : la) {
            as = (AttributeStatement)o;
            if (gs.hasAttribute(as.getIdentifier())) {
                // If the attribute already exists, replace its value
                Value newVal = (Value)groupToMerge.getAttribute(
                    as.getIdentifier()).getValue().clone();
                AttributeStatement existing = gs.getAttribute(
                    as.getIdentifier());
                existing.setValue(newVal);
            } else {
                // If the attribute doesn't exist, add it at the end
                gs.addStatement((AttributeStatement)as.clone());
            }
        }
    }
    
    /**
     * Merge the contents of a new file to an existing file
     * @param mergeFile file to be merge with the current file
     */
    public void mergeFile(LoadedFile mergeFile) {
        
        List<String> mergeGroups = mergeFile.getGroupNames();
        
        // Merge each group from the new file into the current file
        Iterator it = mergeGroups.iterator();
        while (it.hasNext()) {
            String group = it.next().toString();
            this.mergeGroup(group, mergeFile.getGroup(group));
        }
    }

    /**
     * Create a new group that merges the contents of a new group into an
     * existing group.
     * A Group merge means any Parameter that exists in the new group will be
     * added to the existing group (if it doesn't exist) or will replace it
     * in the existing group (if it already exists)
     * @param originalGroup  existing group
     * @param groupToMerge   new group whose contents will be merged into the
     *                       existing group
     */
    public GroupStatement createMergedGroup(GroupStatement originalGroup,
        GroupStatement groupToMerge) {

        // If there is no new group, just return a clone of the original group
        if (groupToMerge == null) {
            return (GroupStatement)originalGroup.clone();
        }        
        // If there is no original group, just return a clone of the new group
        if (originalGroup == null) {
            return (GroupStatement)groupToMerge.clone();
        }
        
        logger.debug("Merging new " + groupToMerge.getIdentifier()
            + " into existing group " + originalGroup.getIdentifier());
        GroupStatement gs = (GroupStatement)originalGroup.clone();
        // Add each attribute within the merge group to the new group
        //  Make a clone of each value or attribute so we have our own copy
        AttributeStatement as = null;
        List la = groupToMerge.getAttributes();
        for (Object o : la) {
            as = (AttributeStatement)o;
            if (gs.hasAttribute(as.getIdentifier())) {
                // If the attribute already exists, replace its value
                Value newValue = (Value)groupToMerge.getAttribute(
                    as.getIdentifier()).getValue().clone();
                AttributeStatement existing
                    = gs.getAttribute(as.getIdentifier());
                existing.setValue(newValue);
            } else {
                // If the attribute doesn't exist, add it at the end
                gs.addStatement((AttributeStatement)as.clone());
            }
        }
        return gs;
    }

    /**
     * Replaces an existing group in the ODL tree with a new one
     * @param groupName          name of the group to replace
     * @param groupStatement     contents of the new group
     */
    public void setGroup(String groupName, GroupStatement groupStatement) {

        if (odlTree == null) {
            return;
        }

        // Find the location of this group in the file, so it can be
        // removed and re-added back in the same place
        List groupList = odlTree.getGroups();
        GroupStatement existingGroupStatement = null;
        int existingIndex = -1;
        for (int i = 0; i < groupList.size(); i++) {
            existingGroupStatement = (GroupStatement)groupList.get(i);
            if (existingGroupStatement.getIdentifier().equals(groupName)) {
                existingIndex = i;
                break;
            }
        }

        if (existingIndex == -1) {
            // add at the end
            if (groupStatement != null) {
                odlTree.addStatement(groupStatement);
            }
        } else {
            // remove the group & replace it
            odlTree.removeStatement(existingGroupStatement);
            if (groupStatement != null) {
                odlTree.addStatement(groupStatement, existingIndex);
            }
        }
    }

    /**
     * Merge the contents of a new parameter into an existing parameter
     * A Parameter merge means the new values (attributeToMerge) will replace
     * any existing values
     * @param originalAttribute  the original parameter value(s)
     * @param attributeToMerge   the new parameter value(s) to merge in
     */
    public AttributeStatement createMergedParameter(
        AttributeStatement originalAttribute,
        AttributeStatement attributeToMerge) {

        // If there is no new parameter, just return a clone of the original
        if (attributeToMerge == null) {
            return (AttributeStatement)originalAttribute.clone();
        }  
        // If there is no original, just return a clone of the new parameter
        if (originalAttribute == null) {
            return (AttributeStatement)attributeToMerge.clone();
        }
        
        // Replace the contents of the original with the new one
        logger.debug("Merging new " + attributeToMerge.getIdentifier()
            + " into existing parameter " + originalAttribute.getIdentifier());
        AttributeStatement as = (AttributeStatement)originalAttribute.clone();
        as.setValue((Value)attributeToMerge.getValue().clone());
        return as;
    }

    /**
     * Set a new ODL attribute as the value for a parameter.  This allows for
     * a wholesale replacement of all values of a parameter at once.
     * @param groupName           the group containing the parameter
     * @param parameterName       the name of the parameter
     * @param attributeStatement  the ODL statement containing the parameter's
     *                            new values
     */
    public void setParameter(String groupName, String parameterName,
        AttributeStatement attributeStatement) {
        if (odlTree == null) {
            return;
        }

        List groups = odlTree.getGroups(groupName);
        if (groups == null || groups.size() <= 0) {
            return;
        }
        GroupStatement gs = null;
        for (Object o: groups) {
            GroupStatement checkGroup = (GroupStatement)o;
            if (checkGroup.getAttribute(parameterName) != null) {
                gs = checkGroup;
            }
        }
        // The parameter doesn't already exist, so add to the first group of
        // this name
        if (gs == null) {
            logger.warn(parameterName + "does not exist in group " + groupName
                + ", adding to first " + groupName);
            gs = (GroupStatement)groups.get(0);
        }

        // Find the location of this parameter in the group, so it can be
        // removed and re-added back in the same place
        AttributeStatement as = null;
        List statementList = gs.getAttributes();
        int existingIndex = -1;
        for (int i = 0; i < statementList.size(); i++) {
            as = (AttributeStatement)statementList.get(i);
            if (as.getIdentifier().equals(parameterName)) {
                existingIndex = i;
                break;
            }
        }

        if (existingIndex == -1) {
            if (attributeStatement != null) {
                gs.addStatement(attributeStatement);
            }
        } else {
            gs.removeStatement(as);
            if (attributeStatement != null) {
                gs.addStatement(attributeStatement, existingIndex);
            }
        }
    }

    /* Routines to get specific ODL objects ----------------------------------*/
    /**
     * Get the first ODL Group Statement for a given group name
     * @param groupName  the name of the group
     * @return           the ODL GroupStatment object which encompasses all the
     *                   contents of the group
     */
    public GroupStatement getGroup(String groupName) {
        if (odlTree == null) {
            return null;
        }
        List groups = odlTree.getGroups(groupName);
        if (groups == null || groups.size() <= 0) {
            return null;
        }
        GroupStatement gs = (GroupStatement)groups.get(0);
        return gs;
    }

    /**
     * Get the ODL Attribute Statement for a given parameter
     * @param groupName      the group that contains the parameter
     * @param parameterName  the name of the parameter
     * @return               the ODL AttributeStatement object which encompasses
     *                       the entire parameter value(s)
     */
    public AttributeStatement getParameter(String groupName,
        String parameterName) {
        if (odlTree == null) {
            return null;
        }
        List groups = odlTree.getGroups(groupName);
        if (groups == null || groups.size() <= 0) {
            return null;
        }
        AttributeStatement as = null;
        for (Object o: groups) {
            GroupStatement gs = (GroupStatement)o;
            as = gs.getAttribute(parameterName);
            if (as != null) {
                break;
            }
        }
        return as;
    }

    /* Basic Getters & Setters -----------------------------------------------*/
    public LoadedFileType getFileType() {
        return fileType;
    }

    public void setFileType(String fileType) {
        this.fileType = LoadedFileType.valueOf(fileType);
    }

    public void setFileType(LoadedFileType fileType) {
        this.fileType = fileType;
    }

    public LoadedFileSchema getFileSchema() {
        return schema;
    }

    public void setFileSchema(String schema) {
        this.schema = LoadedFileSchema.valueOf(schema);
    }

    public void setFileSchema(LoadedFileSchema schema) {
        this.schema = schema;
    }

    public String getFileName() {
        return fileName;
    }

    public void setFileName(String fileName) {
        this.fileName = fileName;
    }

    public Label getOdlTree() {
        return odlTree;
    }

    public void setOdlTree(Label odlTree) {
        this.odlTree = odlTree;
    }

    /**
     * Utility class used to prevent possible deadlock when calling the CPF/BPF
     * validate script.  There are situations where a process's output buffer 
     * may fill with error/warning messages before the entire CPF/BPF has been
     * written to standard in.  If that happens and the standard in buffer fills
     * the parent will block until the child reads from standard in and the
     * child will block waiting for parent to read the standard out/err.
     */
    private class StreamReaderThread extends Thread 
    {
        private BufferedReader reader;
        private ArrayList<String> output;
        private boolean error;
        private String errorMessage;

        public StreamReaderThread(InputStream inStream) 
        {
            reader = new BufferedReader(new InputStreamReader(inStream));
            output = new ArrayList<String>();
            error = false;
        }

        public void run() 
        {
            String line;
            try 
            {
                while ((line = this.reader.readLine()) != null) 
                {
                    output.add(line);
                }
            }
            catch(IOException ex) 
            {
                error = true;
                errorMessage = ex.getMessage();
            }
        }

        public ArrayList<String> getOuput()    
        {
            return output;
        }

        public boolean errorReadingStream()
        {
            return error;
        }
    
        public String getErrorMessage()
        {
            return errorMessage;
        }
    }
}
