package library;
import java.util.ArrayList;
import java.util.List;
import java.util.Observable;
import java.util.Iterator;
import java.util.Collections;
import java.io.*;
import java.net.URL;
import java.net.MalformedURLException;
import org.apache.log4j.Logger;
import gov.nasa.pds.tools.label.Label;

/**
 * Class for keeping track of information about files currently loaded
 * in the GUI
 */
public class LoadedFileList extends Observable {

    final protected List<LoadedFile> loadList; // the list of files currently
                                               // loaded in the GUI

    private final static Logger logger
        = Logger.getLogger(LoadedFileList.class.getName());

    /**
     * Create an empty list of files
     */
    public LoadedFileList() {
        // Make the list of loaded files thread-safe
        //  All accesses are serial, but multiple accesses that need to be
        //  treated as a single unit (e.g., iterating over it, or looking up
        //  then removing an element) need to be manually synchronized on
        //  loadList (not on LoadedFileList.this).
        loadList = Collections.synchronizedList(new ArrayList<LoadedFile>());
    }

    /**
     * Get the files that are active -- those that have a parsed ODL tree
     * @return    list of LoadedFile objects
     */
    public ArrayList<LoadedFile> getActiveFiles() {
        ArrayList<LoadedFile> active = new ArrayList<LoadedFile>();
        LoadedFile thisFile = null;
        synchronized (loadList) {
            for (Iterator i = loadList.iterator(); i.hasNext();) {
                thisFile = (LoadedFile)i.next();
                if (thisFile != null && thisFile.getOdlTree() != null) {
                    active.add(thisFile);
                }
            }
        }
        return active;
    }

    /**
     * Add a new file to the list and notify observers.
     * When a file is added, don't notify right away if it has not been parsed 
     * yet (label is null). If this is the case, refresh() should be called
     * after the file has finished parsing and the LoadedFile now has a valid
     * label, to notify the GUI that the file is ready to be displayed.
     * @param fileName   name of the new file to add
     * @param label      parsed ODL tree for the file, null if not parsed yet
     */
    public void addFile(String fileName, Label label) {
        LoadedFile newFile = new LoadedFile(fileName, label);
        loadList.add(newFile);
        if (label != null) {
            setChanged();
            notifyObservers(newFile);
        }
    }

    /**
     * Indicate that one of the files in the list has changed, and notify
     * observers so they can make any appropriate GUI updates.
     * @param lf    the file that was updated
     */
    public void refresh(LoadedFile lf) {
        setChanged();
        notifyObservers(lf);
    }

    /**
     * Remove the named file from the list.  Removes it from the underlying list
     * and notifies observers that the list has changed.
     * @param fileName
     */
    public void removeFile(String fileName) {
        synchronized (loadList) {
            int existing = loadList.indexOf(new LoadedFile(fileName));
            if (existing > -1) {
                removeFile(existing);
            }
        }
    }

    /**
     * Remove the file at the given index location in the list.
     * @param index   0-based index location
     */
    public void removeFile(int index) {
        synchronized (loadList) {
            if (index < 0 || index > loadList.size()) {
                return;
            }
            // Rather than remove the item, we just null it out so file indexes
            // don't move around
            loadList.set(index, null);
        }
        setChanged();
        notifyObservers();
    }

    /**
     * Get the LoadedFile at the specified index location
     * @param index   0-based index location in the list
     * @return        the LoadedFile object at that location
     */
    public LoadedFile getFile(int index){
        return loadList.get(index);
    }

    /**
     * Get the LoadedFile based on file name
     * @param fileName   Name of the file to retrieve
     * @return           LoadedFile object for the named file, or null if no
     *                   file of that name is in the list
     */
    public LoadedFile getFile(String fileName) {
        synchronized (loadList) {
            int index = loadList.indexOf(new LoadedFile(fileName));
            if (index != -1) {
                return loadList.get(index);
            } else {
                return null;
            }
        }
    }

    /**
     * Get the index in the list for the named file
     * @param fileName   Name of the file to locate
     * @return           Index in the list for the named file, or -1 if no
     *                   file of that name is in the list
     */
    public int getFileIndex(String fileName) {
        return loadList.indexOf(new LoadedFile(fileName));
    }

    /**
     * Find out if the named file is in the list
     * @param fileName   Name of the file to locate
     * @return           true if the list contains the named file; false
     *                   otherwise
     *                   file of that name is i the list
     */
    public boolean fileIsPresent(String fileName) {
        return loadList.contains(new LoadedFile(fileName));
    }

    /**
     * Open a file from the file system & load it into the GUI.
     * @param filePath               Path & file name of the file to open
     * @throws LibraryException      if  error parsing the file
     * @throws MalformedURLException if error creating a URL from the file name
     */
    public void loadFile(File filePath) 
        throws LibraryException, MalformedURLException {
        String fileDirectory = filePath.getPath();
        logger.debug("Opening file using file name " + fileDirectory);

        //check if this file is already open.
        boolean found = fileIsPresent(fileDirectory);
        if (found) {
            throw new LibraryException(fileDirectory + " is already open");
        }

        URL fileUrl = new URL("file", "localhost", fileDirectory);
        parseAndAddFile(fileUrl, fileDirectory);
    }

    /**
     * Open a CPF at the specified URL & load it into the GUI.
     * @param cpfURL       URL for the file to open
     * @throws LibraryException   if unable to open the specified file
     */
    public void loadFile(URL cpfURL) throws LibraryException {
        logger.debug("Opening file using URL " + cpfURL.toString());
        String fileDirectory = LoadedFile.getFileNameFromUrl(cpfURL);
        loadFile(cpfURL, fileDirectory);
    }

    /**
     * Open a CPF at the specified URL, and give it a new name when loading into
     * the GUI.  The file from the URL is downloaded into a temporary file
     * before being parsed.
     * @param cpfURL             the URL of the CPF to open
     * @param newName            the new name to call the file once opened
     * @throws IOException       if unable to create the temporary file
     * @throws LibraryException  if unable to parse the temporary file
     */
    public void loadFile(URL cpfURL, String newName) 
        throws LibraryException {
        logger.debug("Opening file using URL " + cpfURL.toString()
            + ", will be named " + newName);

        //check if this file is already open.
        boolean found = fileIsPresent(newName);
        if (found) {
            throw new LibraryException(newName + " is already open");
        }
        parseAndAddFile(cpfURL, newName);
    }

    /**
     * Call the PDS tools ODL parser using the given URL, and once parsed,
     * add it to the list of files loaded in the GUI with the given name.
     * @param cpfUrl            URL containing the file to parse
     * @param pathAndFileName   the name to give the parsed file
     * @throws LibraryException if error parsing the file
     */
    private void parseAndAddFile(URL cpfUrl, String pathAndFileName)
        throws LibraryException {

        // Make these 2 operations a thread-safe unit
        synchronized (loadList) {

            // Check if this file is already open.
            boolean found = fileIsPresent(pathAndFileName);
            if (found) {
                throw new LibraryException(pathAndFileName
                    + " is already open");
            }

            // Immediately after checking, create an entry in the loaded file
            //   list so all threads know it's being loaded
            addFile(pathAndFileName, null);
        }

        // This may take a while
        logger.debug("Parsing " + pathAndFileName + " from ODL file "
            + cpfUrl.toString());
        Label aLabel = null;
        try {
            aLabel = LoadedFile.parse(cpfUrl);
        } catch (LibraryException le) {
            // If the file wasn't parsed, remove from the list
            removeFile(pathAndFileName);
            throw le;
        }

        // File was parsed successfully, set the parsed ODL tree into the
        //  LoadedFile object.
        LoadedFile lf = getFile(pathAndFileName);
        lf.setOdlTree(aLabel);
        logger.debug("Loaded file " + pathAndFileName + " of type "
            + getFile(pathAndFileName).getFileType()
            + " for schema "
            + getFile(pathAndFileName).
              getFileSchema());

    }
}
