package library;

import java.net.URL;
import java.net.MalformedURLException;
import java.util.List;
import java.util.TreeMap;
import java.util.Calendar;
import javax.xml.namespace.QName;
import javax.xml.datatype.DatatypeFactory;
import org.apache.log4j.Logger;

import gov.usgs.ias.services.clients.CpfService;
import gov.usgs.ias.services.clients.CpfService_Service;
import gov.usgs.ias.services.clients.BpfService;
import gov.usgs.ias.services.clients.BpfService_Service;
import gov.usgs.ias.services.clients.GetAllBpfNamesRequest;
import gov.usgs.ias.services.clients.GetAllCpfRequest;
import gov.usgs.ias.services.clients.GetAllCpfNamesByDateRangeRequest;
import gov.usgs.ias.services.clients.CpsStatusEnum;
import gov.usgs.ias.services.clients.CpsSourceEnum;
import gov.usgs.ias.services.clients.BpfSensorEnum;
import gov.usgs.ias.services.clients.CpsSummary;
import gui.CpfConfig;
import gui.CpfViewer;

/**
 * Provides interfaces to the web services for querying (using SOAP) and
 * retrieving (using REST) CPF and BPF files.
 */
public class Services {

    private static final Logger logger = Logger.getLogger(Services.class);

    /**
     * Create an object to interface with the web services
     */
    public Services() {
    }

    /**
     * Perform a query for filenames from the CPF service
     * @param schema      production or evaluation schema
     * @param latestOnly  flag indicating only the latest version of each file
     * @param status      active or non-active files (null for any)
     * @param startDate   beginning of the date range (null for all dates)
     * @param endDate     end of the date range (null for all dates)
     * @return           table of file names and statuses
     * @throws LibraryException if the service query failed
     */
    public Object[][] getCpfFileNameList(LoadedFileSchema schema,
            boolean latestOnly, CpfStatus status, Calendar startDate,
            Calendar endDate) throws LibraryException {
        // Verify the status is a valid one
        if (status == null) {
            status = CpfStatus.ANY;
        }

        // Call Web Service Operation
        String cpfServiceWsdl = null;
        List<CpsSummary> result = null;
        try {
            // Get the wsdl of the configured service endpoint location
            cpfServiceWsdl = getSoapServiceWsdl(LoadedFileType.CPF);
            CpfService_Service service = new CpfService_Service(
                new URL(gov.usgs.ias.services.clients.CpfService_Service.class.
                getResource("."),
                cpfServiceWsdl),
                new QName(CpfConfig.CPF_SERVICE_NAMESPACE, "CpfService"));
            CpfService port = service.getCpfServicePort();

            // Convert the input schema to the service's source
            GetAllCpfRequest request = new GetAllCpfRequest();
            CpsSourceEnum source;
            if (schema.equals(LoadedFileSchema.PROD)) {
                source = CpsSourceEnum.PROD;
            } else {
                source = CpsSourceEnum.EVAL;
            }
            request.setSource(source);
            result = port.getAllCpfNames(request);
        } catch (Exception ex) {
            throw new LibraryException("Error communicating with service "
                + cpfServiceWsdl + " : " + ex);
        }

        return filterServiceResults(result, status, latestOnly, startDate,
                endDate);
    }

    /**
     * Perform a query for PROD filenames from the CPF service.  This method 
     * returns names of files that match the specified date range.
     * (This method uses the service date range query, which only works for
     *  files in the production schema.  The GUI has no need to query EVAL files
     *  by date range.)
     * @param startDate   beginning of the date range
     * @param endDate     end of the date range
     * @param status      status of files to return, null for all statuses
     * @param latestOnly  flag indicating get only the latest version of each
     * @return            table of file names and statuses
     * @throws LibraryException if the service query failed
     */
    public Object[][] getCpfProdFileNameList(
        Calendar startDate, Calendar endDate,
        CpfStatus status, boolean latestOnly) throws LibraryException {
        // Verify the status is a valid one
        if (status == null) {
            status = CpfStatus.ANY;
        }

        // Call Web Service Operation
        String cpfServiceWsdl = null;
        List<CpsSummary> result = null;
        try {
            // Get the wsdl of the configured service endpoint location
            cpfServiceWsdl = getSoapServiceWsdl(LoadedFileType.CPF);
            CpfService_Service service = new CpfService_Service(
                new URL(gov.usgs.ias.services.clients.CpfService_Service.class.
                getResource("."),
                cpfServiceWsdl),
                new QName(CpfConfig.CPF_SERVICE_NAMESPACE, "CpfService"));
            CpfService port = service.getCpfServicePort();

            // Set up the query parameters
            DatatypeFactory dtf = DatatypeFactory.newInstance();
            GetAllCpfNamesByDateRangeRequest request =
                new GetAllCpfNamesByDateRangeRequest();
            request.setStatus(null);  // query on all status & version; 
            request.setVersion(null); //  results will be filtered later
            request.setStartDate(dtf.newXMLGregorianCalendar(
                startDate.get(Calendar.YEAR),
                startDate.get(Calendar.MONTH) + 1,
                startDate.get(Calendar.DAY_OF_MONTH),
                startDate.get(Calendar.HOUR_OF_DAY),
                startDate.get(Calendar.MINUTE),
                startDate.get(Calendar.SECOND),
                startDate.get(Calendar.MILLISECOND),
                0));
            request.setEndDate(dtf.newXMLGregorianCalendar(
                endDate.get(Calendar.YEAR),
                endDate.get(Calendar.MONTH) + 1,
                endDate.get(Calendar.DAY_OF_MONTH),
                endDate.get(Calendar.HOUR_OF_DAY),
                endDate.get(Calendar.MINUTE),
                endDate.get(Calendar.SECOND),
                endDate.get(Calendar.MILLISECOND),
                0));
            result = port.getAllCpfNamesByDateRange(request);
        } catch (Exception ex) {
            throw new LibraryException("Error communicating with service "
                + cpfServiceWsdl + " : " + ex);
        }

        return filterServiceResults(result, status, latestOnly, null, null);
    }

    /**
     * Perform a query for filenames from the BPF service using the given
     * filtering constraints.
     * @param schema      production or evaluation schema
     * @param sensor      display BPF files of a given sensor (null for any)
     * @param latestOnly  flag indicating only the latest version of each file
     * @param status      active or non-active files (null for any)
     * @param startDate   beginning of the date range (null for all dates)
     * @param endDate     end of the date range (null for all dates)
     * @return            table of file names and statuses
     * @throws LibraryException if the service query failed
     */
    public Object[][] getBpfFileNameList(LoadedFileSchema schema,
            BpfSensor sensor, boolean latestOnly, CpfStatus status,
            Calendar startDate, Calendar endDate) throws LibraryException {
        // Ensure inputs have valid values
        if (status == null) {
            status = CpfStatus.ANY;
        }
        if (sensor == null) {
            sensor = BpfSensor.ANY;
        }

        // Call Web Service Operation
        String bpfServiceWsdl = null;
        List<CpsSummary> result = null;
        try {
            // Get the wsdl of the configured service endpoint location
            bpfServiceWsdl = getSoapServiceWsdl(LoadedFileType.BPF);
            logger.trace("Contacting BPF service at " + bpfServiceWsdl);
            BpfService_Service service = new BpfService_Service(
                new URL(gov.usgs.ias.services.clients.BpfService_Service.class.
                getResource("."),
                bpfServiceWsdl),
                new QName(CpfConfig.BPF_SERVICE_NAMESPACE, "BpfService"));
            BpfService port = service.getBpfServicePort();
            
            // Convert the input params to the service's data types
            GetAllBpfNamesRequest request = new GetAllBpfNamesRequest();
            CpsSourceEnum source;
            BpfSensorEnum bpfSensor = null;
            if (schema.equals(LoadedFileSchema.PROD)) {
                source = CpsSourceEnum.PROD;
            } else {
                source = CpsSourceEnum.EVAL;
            }
            if (sensor.equals(BpfSensor.OLI)) {
                bpfSensor = BpfSensorEnum.OLI;
            } else if (sensor.equals(BpfSensor.TIRS)) {
                bpfSensor = BpfSensorEnum.TIRS;
            }
            request.setSource(source);
            request.setSensor(bpfSensor);
            result = port.getAllBpfNames(request);
            logger.trace("Returned " + result.size() + " files");
        } catch (Exception ex) {
            throw new LibraryException("Error communicating with service "
                + bpfServiceWsdl + " : " + ex);
        }

        return filterServiceResults(result, status, latestOnly, startDate,
                endDate);
    }
    
    /**
     * Filters results from the web service into the set of requested files.
     * and translates them into a 2-d array ([filename][status]) of files for
     * display in the file selection dialog.  We filter the results of the web
     * service query for a couple of reasons: When querying the EVAL schema,
     * the only operation available is to get all files, so we have to
     * do our own filtering of the results.  When querying the PROD schema,
     * query by status is available but not the option to return only the
     * latest version of each file.
     * @param results    the list of summaries returned by the service
     * @param status     the status of files we want returned
     * @param latestOnly flag indicating only the latest version of each file
     *                   should be returned
     * @param startDate  beginning of the date range of files we want returned
     * @param endDate    end of the date range of files we want returned
     * @return 2-d array, which is filename and status for each matching file
     */
    private Object[][] filterServiceResults(List<CpsSummary> results,
        CpfStatus status, boolean latestOnly, Calendar startDate,
        Calendar endDate) {
          
        // Create a hash where key is file name and value is its index in the
        // results
        TreeMap<String, Integer> filesToKeep = new TreeMap<String, Integer>();

        String filename;         // current results filename
        String key;              // key of unique files stored in the hash
        Calendar fileStartDate;  // start date of the results filename
        Calendar fileEndDate;    // end date of the results filename
        int fileVersion;         // version of the results filename
        int foundVersion;        // version of an existing file in the hash
        Object[][] data = null;  // 2-d array to return

        for (int iResult = 0; iResult < results.size(); iResult++) {
            // Check for a null result object in the list.
            // This gets created when the service returns, for example:
            //               <cpf xsd:nil="true">
            if (results.get(iResult) == null ||
                results.get(iResult).getFilename() == null)
            {
                continue;
            }
            filename = results.get(iResult).getFilename();
            fileVersion = FileNamingConventions.getVersion(filename);
            fileStartDate = FileNamingConventions.getStartDate(filename);
            fileEndDate = FileNamingConventions.getEndDate(filename);

            // If keeping the latest version only, hash just the base file
            // name; otherwise use the entire filename (including version)
            // as the hash key
            if (latestOnly) {
                key = filename.split("\\.")[0];
            } else {
                key = filename;
            }

            // Do status filtering to decide if we want to keep this one
            if ((status == CpfStatus.ANY)
                || (status == CpfStatus.NOT_ACTIVE
                    && !results.get(iResult).getStatus().equals(
                        CpsStatusEnum.ACTIVE))
                || (status == CpfStatus.ACTIVE
                && results.get(iResult).getStatus().equals(
                    CpsStatusEnum.ACTIVE))) {
                
                // Include the file if date range filtering is not being done,
                // or if filtering is done and the file date range overlaps
                // the filter date range
                if (((startDate == null) || (endDate == null)) 
                     || ((startDate != null & endDate != null)
                         && fileOverlapsDateFilter(fileStartDate, fileEndDate,
                            startDate, endDate))) {

                    if (filesToKeep.containsKey(key)) {
                        // Find the version of the file that's already there
                        int found = filesToKeep.get(key);
                        foundVersion = FileNamingConventions.getVersion(
                            results.get(found).getFilename());
                        if (fileVersion > foundVersion) {
                            // If the new version is higher, replace the
                            // existing entry with this one
                            filesToKeep.put(key, iResult);
                        }
                    } else {
                        // This file wasn't recorded in the hash yet, so add it
                        filesToKeep.put(key, iResult);
                    }
                }
            }
        }

        // translate the hash into an array
        data = new Object[filesToKeep.size()][2];
        int keep = 0;
        for (int iResult : filesToKeep.values()) {
            data[keep][0] = results.get(iResult).getFilename();
            data[keep][1] = results.get(iResult).getStatus().toString();
            keep++;
        }

        return data;
    }

    /**
     * Retrieve a file from the REST service and parse & load into the GUI under
     * a new file name
     * @param application         application context
     * @param fileType            type of file (CPF or BPF)
     * @param fileToRetrieve      name of the file to retrieve from the service
     * @param newFileName         new name to give the file in the GUI once
     *                            loaded
     * @throws LibraryException   if the file could not be retrieved or loaded
     *                            in the GUI
     */
    public void retrieveFileAndRenameInGui(CpfViewer application,
        LoadedFileType fileType, String fileToRetrieve, String newFileName)
        throws LibraryException, MalformedURLException {
        logger.info("Retrieving " + fileToRetrieve + " from REST service");
        // Construct the REST service URL & open it using the library
        String restServiceURL = getRestServiceUrl(fileType);
        restServiceURL += fileToRetrieve;
        logger.debug("Retrieving " + restServiceURL);
        URL url = new URL(restServiceURL);
        application.loadedFileList.loadFile(url, newFileName);
    }

    /**
     * Retrieve a file from the REST service & parse it, without loading it into
     * the GUI
     * @param fileType      type of file (CPF or BPF)
     * @param fileName      name of the file to retrieve from the service
     * @return              a LoadedFile object that has not been added to the
     *                      application's loaded file list
     */
    public LoadedFile retrieveTempFile(
        LoadedFileType fileType, String fileName) throws LibraryException {
        LoadedFile lf = null;
        logger.info("Retrieving " + fileName + " from REST service");
        try {
            String restServiceUrl = getRestServiceUrl(fileType);
            restServiceUrl += fileName;
            logger.debug(" retrieving " + restServiceUrl);
            URL url = new URL(restServiceUrl);
            lf = LoadedFile.create(url);
        } catch (Exception e) {
            throw new LibraryException(
                "Error retrieving temp file from service: "
                + e.getMessage(), e);
        }
        return lf;
    }

    /**
     * Get the WSDL for the query (SOAP) service.  Uses $IAS_SERVICES if
     * defined.  If not, tries to get the WSDL from properties, and, failing
     * that, uses a default value.
     * @param type   which service to contact (CPF or BPF)
     * @return       string representation of the service's WSDL URL
     * @throws LibraryException if type was null
     */
    private String getSoapServiceWsdl(LoadedFileType type) 
        throws LibraryException {
        if (type == null) {
            throw new LibraryException("Cannot get SOAP service WSDL for null "
                + "file type");
        }
        String wsdl = System.getenv("IAS_SERVICES");
        if (type.equals(LoadedFileType.CPF)) {
            if (wsdl == null) {
                wsdl = System.getProperty("cpf.service.wsdl",
                    CpfConfig.IAS_SERVICES_DEFAULT
                    + CpfConfig.CPF_SERVICE_WSDL);
            } else {
                wsdl += CpfConfig.CPF_SERVICE_WSDL;
            }
        } else if (type.equals(LoadedFileType.BPF)) {
            if (wsdl == null) {
                wsdl = System.getProperty("bpf.service.wsdl",
                    CpfConfig.IAS_SERVICES_DEFAULT
                    + CpfConfig.BPF_SERVICE_WSDL);
            } else {
                wsdl += CpfConfig.BPF_SERVICE_WSDL;
            }
        }
        return wsdl;
    }

    /**
     * Get the base URL for the file retrieval (REST) service
     * @param type   which service to contact (CPF or BPF)
     * @return       string representation of the REST service's URL
     * @throws LibraryException if type was null
     */
    private String getRestServiceUrl(LoadedFileType type) 
        throws LibraryException {
        if (type == null) {
            throw new LibraryException("Cannot get REST service URL for null "
                + "file type");
        }
        String url = System.getenv("IAS_SERVICES");
        if (type.equals(LoadedFileType.CPF)) {
            if (url == null) {
                url = System.getProperty("cpf.service.rest.url",
                    CpfConfig.IAS_SERVICES_DEFAULT
                    + CpfConfig.CPF_REST_URL);
            } else {
                url += CpfConfig.CPF_REST_URL;
            }
        } else if (type.equals(LoadedFileType.BPF)) {
            if (url == null) {
                url = System.getProperty("bpf.service.rest.url",
                    CpfConfig.IAS_SERVICES_DEFAULT
                    + CpfConfig.BPF_REST_URL);
            } else {
                url += CpfConfig.BPF_REST_URL;
            }
        }

        return url;
    }
    
    /**
     * Does the file date range overlap the filter date range?
     * @param fileStartDate
     * @param fileEndDate
     * @param filterStartDate
     * @param filterEndDate
     * @return true if the file falls in the filter date range, or false
     */
    private boolean fileOverlapsDateFilter(
            Calendar fileStartDate, Calendar fileEndDate,
            Calendar filterStartDate, Calendar filterEndDate) {

        if (fileStartDate.equals(filterStartDate)
                || fileStartDate.equals(filterEndDate)
                || fileEndDate.equals(filterStartDate)
                || fileEndDate.equals(filterEndDate)) {
            return true;
        }
        
        if (fileStartDate.before(filterStartDate)
                && fileEndDate.after(filterStartDate)) {
            return true;
        }
        
        if (fileStartDate.before(filterEndDate)
                && fileEndDate.after(filterEndDate)) {
            return true;
        }
        
        if (fileStartDate.after(filterStartDate)
                && fileEndDate.before(filterEndDate)) {
            return true;
        }

        return false;
    }
}
