package workers;

import gui.CpfViewer;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;
import library.CpfStatus;
import library.FileNamingConventions;
import library.LoadedFile;
import library.LoadedFileSchema;
import library.LoadedFileType;
import library.Services;
import org.apache.log4j.Logger;

/**
 * Extends the CpfExportThread class to provide a background thread to baseline
 * files (includes importing an EVAL file from the service, update the version,
 * and export to the PROD database).  This is done in a thread since these
 * operations may take a long time.  It inherits from CpfExportThread and
 * overrides the doInBackground() method, and uses the parent for all others.
 * Note: This thread will baseline multiple files, but it does them one at a
 *       time.  The database operations are not thread-safe.  So the caller must
 *       be sure to check if any other threads are running before starting a
 *       CpfBaselineThread.
 * The 2 required type parameters for this class (&lt;String, String&gt;) are:
 *  String - the result type returned by the doInBackground() and get() methods
 *  String - the type used for carrying out intermediate results by the
 *            publish() and process() methods (if these methods were used)
 */
public class CpfBaselineThread extends CpfExportThread {

    private String[] filesToBaseline = null;
    private CpfViewer application;
    private boolean synchronizeVersion;
    private static final Logger logger
            = Logger.getLogger(CpfBaselineThread.class);

    /**
     * Creates a thread to baseline one or more files.
     * @param application         the CPF Viewer application
     * @param filesToBaseline     the names of EVAL files to baseline
     */
    public CpfBaselineThread(CpfViewer application, String[] filesToBaseline) {
        this(application, filesToBaseline, false);
    }

    /**
     * Creates a thread to baseline one or more files.
     * @param application         the CPF Viewer application
     * @param filesToBaseline     the names of EVAL files to baseline
     * @param synchronizeVersion  synchronize version numbers of
     *                            filesToBaseline
     */
    public CpfBaselineThread(CpfViewer application, String[] filesToBaseline,
            boolean synchronizeVersion) {
        super(application, "Baseline");
        this.filesToBaseline = filesToBaseline;
        this.application = application;
        this.synchronizeVersion = synchronizeVersion;
    }

    /**
     * When a CPF is found to have been split, this dialog prompts the user
     * to continue when baselining.
     */
    class ContinueSplitCpfDialog implements Runnable {
        boolean proceedWithBaseline;
        String cpfName;
        public ContinueSplitCpfDialog(String cpfNameIn)
        {
            this.cpfName = cpfNameIn;
        }
        public void run() {
            Object[] options = {"Continue", "Cancel"};
            int option = JOptionPane.showOptionDialog(application,
                    cpfName + " is merging previously split CPFs. "
                  + "Proceed with baseline?", "Warning",
                    JOptionPane.YES_NO_OPTION,
                    JOptionPane.WARNING_MESSAGE, null, options, options[0]);
            proceedWithBaseline = (option == JOptionPane.YES_OPTION);
       }
    }

    /**
     * For each EVAL file, retrieve it from the service, rename as a PROD file,
     * validate, and insert into the PROD database.
     * @return                  a string indicating the operation is done
     * @throws WorkerException  if an exception occurs retrieving from the 
     * service, parsing the file, validating, or inserting into the database.
     */
    @Override
    public String doInBackground() throws WorkerException {

        Services services = new Services();
        int numFiles = filesToBaseline.length;
        int[] versions = new int[numFiles];
        int maxVersion = 0;

        // Find the next version number for each file in filesToBaseline
        for (int i = 0; i < numFiles; i++) {

            String sourceFile = filesToBaseline[i];
            versions[i] = -1;

            LoadedFileType sourceFileType =
                    FileNamingConventions.getFileType(sourceFile);
            Object[][] existingFiles = null;
            try {
                if (sourceFileType.equals(LoadedFileType.CPF)) {
                    existingFiles = services.getCpfProdFileNameList(
                        FileNamingConventions.getStartDate(sourceFile),
                        FileNamingConventions.getEndDate(sourceFile),
                        CpfStatus.ACTIVE, true);
                } else if (sourceFileType.equals(LoadedFileType.BPF)) {
                    existingFiles = services.getBpfFileNameList(
                            LoadedFileSchema.PROD, null, true, null,
                        FileNamingConventions.getStartDate(sourceFile),
                        FileNamingConventions.getEndDate(sourceFile));
                }
            } catch (Exception ex) {
                publish(new ThreadStatus("Unable to get the next version "
                    + "number for " + sourceFile + " from the PROD database: "
                    + ex.getMessage(), false, sourceFile, false));
                continue;
            }

            int newVersion = 1;
            if (existingFiles != null && existingFiles.length > 0) {
                // Check to make sure this eval file doesn't already have
                //  a split date
                if (FileNamingConventions.hasBeenDivided(sourceFile,
                    existingFiles)) {
                    // Prompt the user to continue the baseline with split
                    // CPFs
                    ContinueSplitCpfDialog contPromptDialog =
                                         new ContinueSplitCpfDialog(sourceFile);
                    try {
                        SwingUtilities.invokeAndWait(contPromptDialog);
                    } catch (Exception ex) {
                        publish(new ThreadStatus("Unable to prompt user to "
                            + "baseline merge of previously split CPFs: "
                            + ex.getMessage() + ". " + sourceFile + " will not "
                            + "be baselined.", false, sourceFile, false));
                        continue;
                    }
                    if (!contPromptDialog.proceedWithBaseline) {
                        // User rejected baseline of this CPF, move to the next
                        continue;
                    }
                }

                // Calculate the new prod version for this eval file
                // which should be one more than the latest version of any
                // file that covers this date range (does not need to match
                // the date range exactly)
                for (int exist = 0; exist < existingFiles.length; exist++) {
                    String existingFileName = (String)existingFiles[exist][0];
                    int existingVersion
                        = FileNamingConventions.getVersion(existingFileName);
                    if (existingVersion >= newVersion) {
                        newVersion = existingVersion + 1;
                        logger.debug("Setting baseline file version of "
                            + sourceFile + " to " + newVersion
                            + ", incremented from " + existingFileName);
                    }
                }
            }

            // Save the next version number for this file.
            // Use the file counter to match index of version to index of
            // the filename.
            versions[i] = newVersion;

            // Track the highest version number.  This will be used if
            // the version numbers need to be synchronized
            if (newVersion > maxVersion) {
                maxVersion = newVersion;
            }
        }

        // If sychronizeVersion set version to max version
        if (synchronizeVersion) {
            logger.debug("Synchronizing all versions for baselined files to "
                + maxVersion);
            for (int i = 0; i < numFiles; i++) {
                // skip files if version could not be determined
                if (versions[i] == -1) {
                    continue;
                }
                versions[i] = maxVersion;
            }
        }

        // Save files
        for (int i = 0; i < numFiles; i++) {

            // skip files if version could not be determined
            if (versions[i] == -1) {
                continue;
            }

            String sourceFile = filesToBaseline[i];

            LoadedFileType sourceFileType =
                    FileNamingConventions.getFileType(sourceFile);

            // Get the new file name: remove the eval_ and set the version #
            String newFileName = FileNamingConventions.constructFileName(
                    LoadedFileSchema.PROD,
                    FileNamingConventions.getSatellite(sourceFile),
                    FileNamingConventions.getSensor(sourceFile),
                    FileNamingConventions.getFileType(sourceFile),
                    FileNamingConventions.getStartDate(sourceFile),
                    FileNamingConventions.getEndDate(sourceFile),
                    versions[i]);

            // Load the eval file into a temp area
            logger.info("Importing " + sourceFile);
            LoadedFile lf = null;
            try {
                lf = services.retrieveTempFile(sourceFileType,
                    sourceFile);
            } catch (Exception e) {
                // If an exception occurred, lf is left as null
            }
            if (lf == null) {
                publish(new ThreadStatus("Error retrieving " + sourceFile
                    + ", please see log for details", false, sourceFile,
                    false));
                continue;
            }

            // Once loaded, update the file name
            logger.info("Baselining " + sourceFile + " as "
                    + newFileName);
            lf.updateFileName(newFileName);
            try {
                // Insert into PROD
                application.databaseLibrary.insertFile(
                    LoadedFileSchema.PROD, lf);
            } catch (Exception ex) {
                Throwable rootCause = ex;
                while (rootCause.getCause() != null) {
                    rootCause = rootCause.getCause();
                }
                publish(new ThreadStatus("Error baselining " + newFileName
                    + " : " + rootCause.getMessage(),
                    false, sourceFile, false));
                continue;
            }
            // Notify the user of successful operation
            publish(new ThreadStatus("Baselined " + sourceFile + " as "
                + lf.getFileName(), true, sourceFile, false));
        }

        return "done";
    }
}
