package workers;

import gui.CpfViewer;
import library.LoadedFile;
import library.LoadedFileSchema;
import gui.ParameterTablePanel;
import java.util.concurrent.ExecutionException;
import javax.swing.SwingWorker;
import javax.swing.JOptionPane;
import org.apache.log4j.Logger;
import org.apache.commons.lang.WordUtils;

/**
 * Extends the SwingWorker class to provide a background thread to export
 * files to the EVAL database.  This also includes running the validation,
 * prior to inserting into the db.  This is done in a thread since these
 * operations may take a long time.
 * Note: This thread will export multiple files, but it does them one at a time.
 *       The database operations are not thread-safe.  So the caller must be
 *       sure to check if any other threads are running before starting a
 *       CpfExportThread.
 * The 2 required type parameters for this class (&lt;String, ThreadStatus&gt;)
 * are:
 * String  - the result type returned by the doInBackground() and get() methods.
 * ThreadStatus - the type used for carrying out intermediate results by the
 *            publish() and process() methods.
 */
public class CpfExportThread extends SwingWorker<String, ThreadStatus> {

    private LoadedFile[] filesToExport = null;
    private String operation = "Export";
    private CpfViewer application;
    private static Logger logger = Logger.getLogger(
        CpfExportThread.class.getName());

    // Constructor used by child classes that perform a
    // different operation
    CpfExportThread(CpfViewer application, String operation) {
        this.application = application;
        this.operation = operation;
    }

   /**
    * Constructor with list of files to export
    * @param application       the CPF Viewer application
    * @param filesToExport     array of LoadedFile objects to export to the
    *                          database
    */
    public CpfExportThread(CpfViewer application, LoadedFile[] filesToExport) {
        this.filesToExport = filesToExport;
        this.application = application;
    }

    /**
     * The main processing method.  Any exceptions thrown by this method
     * will be propagated to the get() method.  
     * As each file is processed, the status of the export is published to 
     * notify the user.
     * @return                   a string indicating the operation is done
     * @throws WorkerException  if background processing fails
     */
    @Override
    public String doInBackground() throws WorkerException {

        for (LoadedFile lf : filesToExport) {
            try {
                // First validate the file
                lf.validateFile();

                // Insert into database (database login is done in the
                // main thread)
                application.databaseLibrary.insertFile(
                    LoadedFileSchema.EVAL, lf);
            } catch (Exception e) {
                logger.error("Export exception", e);
                Throwable rootCause = e;
                while (rootCause.getCause() != null) {
                    rootCause = rootCause.getCause();
                }
                publish(new ThreadStatus("Error exporting " + lf.getFileName() 
                    + ": " + rootCause.getMessage(),
                    false, lf.getFileName(), true));
                continue;
            }
            
            // Notify the user of successful export
            publish(new ThreadStatus("Exported " + lf.getFileName(), true, 
                lf.getFileName(), true));
        }

        return "done";  // this is what get() will return
    }

    /**
     * Method to receive data chunks from the publish() method
     * and act on them in the Event Dispatch Thread.  As each file finishes
     * being exported, a message is published, which is displayed to the
     * user.
     * @param chunks   list of ThreadStatus objects, the export status of each
     *                 file that has been processed
     */
    @Override
    protected void process(java.util.List<ThreadStatus> chunks) {
        // Display each message to the user
        for (ThreadStatus status : chunks) {
            JOptionPane.showMessageDialog(application,
                WordUtils.wrap(status.getMessage(), 80));

            // After exporting, clear the modified indicator if successful
            if (status.isSuccess() && status.doesAffectModified()) {
                // Locate the correct panel for this file
                int lfIndex = application.loadedFileList.getFileIndex(
                    status.getFileName());
                for (ParameterTablePanel ptp :
                    application.parameterViewerPanel.getDisplayedPanels()) {
                    if (ptp.getLoadedFileIndex() == lfIndex) {
                        ptp.getUndoableModel().resetModified();
                        break;
                    }
                }
            }
        }
    }

    /**
     * After the background processing is done, GUI updates can now be done
     * on the Event Dispatch Thread.  In this case, it doesn't really do
     * anything since each file's status was reported by the process() method.
     */
    @Override
    protected void done() {
        firePropertyChange("done", null, null);
        application.databaseLibrary.logout();
        application.statusBar.progress.clearBusy();
        try {
            get();
        } catch (InterruptedException ex) {
            logger.error("Interrupted", ex);
            JOptionPane.showMessageDialog(application,
                operation + " Interrupted");
            return;
        } catch (ExecutionException ex) {
            logger.error("Error in " + operation, ex);
            JOptionPane.showMessageDialog(application,
                "Error in " + operation);
            return;
        }
    }
}
