package workers;

import gui.CpfViewer;
import library.LoadedFile;
import java.util.concurrent.ExecutionException;
import javax.swing.SwingWorker;
import javax.swing.JOptionPane;
import java.io.File;
import org.apache.log4j.Logger;
import org.apache.commons.lang.WordUtils;

/**
 * Extends the SwingWorker class to provide a background thread to load a
 * file from the file system.  This is done in a thread since parsing a large
 * ODL file may take a long time.
 * The 2 required type parameters for this class (&lt;String, Integer&gt;) are:
 *  String - the result type returned by the doInBackground() and get() methods.
 *  Integer - the type used for carrying out intermediate results by the
 *            publish() and process() methods (if these methods were used).
 */
public class CpfLoadThread extends SwingWorker<String, Integer> {

    protected File fileToLoad = null;        // file to load
    protected String fileNameToLoad = null;  // name of file to load
    protected CpfViewer application = null;  // application context
    private static Logger logger = Logger.getLogger(
        CpfLoadThread.class.getName());

    // default no-arg constructor for child classes
    CpfLoadThread() {
    }

    /**
     * Create a thread to load a file from the file system into the GUI
     * @param application  the CPF Viewer application
     * @param fileToLoad   the file to load
     */
    public CpfLoadThread(CpfViewer application, File fileToLoad) {
        this.fileToLoad = fileToLoad;
        this.application = application;
    }

    /**
     * Load the file into the GUI.
     * @return                   the name of the file that was loaded
     * @throws WorkerException    on error accessing or parsing the file
     */
    @Override
    public String doInBackground() throws WorkerException {
        // This provides no intermediate status as it's going
        if (fileToLoad != null && application != null) {
            fileNameToLoad = fileToLoad.getPath();
            try {
                application.loadedFileList.loadFile(fileToLoad);
            } catch (Exception ex) {
                // to help tracing the error
                throw new WorkerException("Load thread", ex);
            }
        }

        return fileNameToLoad;
    }

    /**
     * After the background processing is done, GUI updates can now be done
     * on the Event Dispatch Thread.
     */
    @Override
    protected void done() {
        firePropertyChange("done", null, null);
        if (application == null) {
            return;
        }
        application.statusBar.progress.clearBusy();
        String loadedFileName = null;
        try {
            // Make sure the file was successfully parsed
            loadedFileName = get();
            if (loadedFileName == null) {
                return;
            }
            LoadedFile loadedFile
                    = application.loadedFileList.getFile(loadedFileName);
            if (loadedFile != null && loadedFile.getOdlTree() != null) {
                // add the parameter table panel
                application.parameterViewerPanel.createNewPanel(loadedFileName);
                // update the parameter tree
                application.loadedFileList.refresh(loadedFile);
            }
        } catch (InterruptedException ex) {
            logger.error("Interrupted", ex);
            JOptionPane.showMessageDialog(application,
                "Interrupted load of " + fileNameToLoad);
        } catch (ExecutionException ex) {
            logger.error("Error loading", ex);
            Throwable rootCause = ex;
            while (rootCause.getCause() != null) {
                rootCause = rootCause.getCause();
            }
            JOptionPane.showMessageDialog(application,
                WordUtils.wrap("Error loading " + fileNameToLoad + " : "
                + rootCause.getMessage(), 80));
        }
    }
}
