package workers;

import gui.CpfViewer;
import library.LoadedFile;
import java.util.concurrent.ExecutionException;
import java.util.Calendar;
import javax.swing.JOptionPane;
import org.apache.log4j.Logger;
import org.apache.commons.lang.WordUtils;

/**
 * Extends the CpfImportThread, to do additional processing after the file is
 * imported from the service.  Overrides the done method and uses
 * parent methods otherwise.  The additional processing includes
 * resetting the end date of the pre-split file and the start date of the
 * post-split file.
 * Also optionally, one may extend the end date of the post-split file.
 */
public class CpfSplitThread extends CpfImportThread {

    private final static Logger logger = Logger.getLogger(
            CpfSplitThread.class.getName());
    private String preSplitFileName = null;
    private String postSplitFileName = null;
    private Calendar end1 = null;       // new end date of (pre-split) file 1
    private Calendar split = null;      // the split date is the start
                                        // of (post-split) file 2
    private Calendar newEndDate = null; // optional new end date of (post-split)
                                        // file 2

    /**
     * Create a thread to split a file into two and load both into the GUI.
     * @param application        the CPF Viewer application
     * @param fileNameToLoad     name of the PROD file to import
     * @param preSplitFileName   name to give the pre-split file
     * @param postSplitFileName  name to give the post-split file
     * @param preSplitEndDate    new end date of the pre-split file
     * @param postSplitStartDate new start date of the post-split file
     * @param newEndDate         new end date of the post-split file
     */
    public CpfSplitThread(CpfViewer application, String fileNameToLoad,
            String preSplitFileName, String postSplitFileName,
            Calendar preSplitEndDate,
            Calendar postSplitStartDate, Calendar newEndDate) {
        super(application, fileNameToLoad, preSplitFileName);
        this.fileNameToLoad = fileNameToLoad;
        this.preSplitFileName = preSplitFileName;
        this.postSplitFileName = postSplitFileName;
        this.end1 = preSplitEndDate;
        this.split = postSplitStartDate;
        this.newEndDate = newEndDate;
        this.application = application;
    }

    /**
     * After the background processing (load of the presplit file) is done,
     * additional processing and GUI updates can now be done on the Event
     * Dispatch Thread.  This includes cloning the presplit file into the
     * postsplit file and loading it into the GUI.
     */
    @Override
    protected void done() {
        firePropertyChange("done", null, null);
        application.statusBar.progress.clearBusy();
        String loadedFile = null;
        try {
            // Make sure the file was successfully parsed
            loadedFile = get();
        } catch (InterruptedException ex) {
            logger.error("Interrupted", ex);
            JOptionPane.showMessageDialog(application,
                    "Interrupted load of " + fileNameToLoad);
        } catch (ExecutionException ex) {
            logger.error("Error loading " + fileNameToLoad, ex);
            Throwable rootCause = ex;
            while (rootCause.getCause() != null) {
                rootCause = rootCause.getCause();
            }
            JOptionPane.showMessageDialog(application,
                    WordUtils.wrap("Error loading " + fileNameToLoad + " : "
                    + rootCause.getMessage(), 80));
        }
        if (loadedFile == null) {
            return;
        }

        int preSplitIndex = application.loadedFileList.getFileIndex(loadedFile);
        if (preSplitIndex == -1) {
            JOptionPane.showMessageDialog(application,
                    "Error loading " + loadedFile + " into the GUI");
            return;
        }
        logger.trace("Loaded file " + preSplitFileName + " into list at "
                + preSplitIndex);

        // Clone file 1 as file 2 
        LoadedFile postSplitLf = (LoadedFile)application.loadedFileList.
                getFile(preSplitIndex).clone();

        // Now both split files are loaded with the original effective dates.
        // Update each with the proper begin & end dates (and new file names).

        // Update the filename, version, end date in the 1st file
        logger.trace("Updating presplit file to name "
                + preSplitFileName + " with end date " + end1.getTime());
        LoadedFile preSplitLf
            = application.loadedFileList.getFile(preSplitIndex);
        preSplitLf.updateFileName(preSplitFileName, null, end1);
        // add the parameter table panel
        application.parameterViewerPanel.createNewPanel(preSplitFileName);
        // update the parameter tree
        application.loadedFileList.refresh(preSplitLf);

        // Update the filename, begin (and, optionally, end date), and version
        // in the 2nd file
        // Then add to the list
        logger.trace("Updating postsplit file to name "
                + postSplitFileName + " with start date " + split.getTime()
                + (newEndDate == null
                ? ""
                : " and new end date " + newEndDate.getTime()));
        postSplitLf.updateFileName(postSplitFileName, split, newEndDate);
        // Make sure that the cloned file isn't already there
        if (application.loadedFileList.fileIsPresent(postSplitFileName)) {
            JOptionPane.showMessageDialog(application,
                    WordUtils.wrap("Error: " + postSplitFileName
                    + "is already loaded in the GUI", 80));
        } else {
            application.loadedFileList.addFile(postSplitLf.getFileName(),
                postSplitLf.getOdlTree());
            application.parameterViewerPanel.createNewPanel(postSplitFileName);
            application.loadedFileList.refresh(postSplitLf);
        }
    }
}
