package workers;

import java.util.concurrent.ExecutionException;
import javax.swing.JOptionPane;
import org.apache.log4j.Logger;
import org.apache.commons.lang.WordUtils;
import gov.nasa.pds.tools.label.*;
import gui.CpfViewer;
import library.LoadedFile;
import gui.ParameterTablePanel;

/**
 * Extends the CpfImportThread, to do additional processing after the file is
 * imported from the service.  Overrides the done method and uses
 * parent methods otherwise.  The additional processing includes merging
 * in the new group before displaying the file in the GUI.
 */
public class CpfUpdateGroupThread extends CpfImportThread {

    private final static Logger logger = Logger.getLogger(
            CpfUpdateGroupThread.class.getName());
    private final GroupStatement groupToMerge;

    /**
     * Create a thread to load a file and update a group in it.
     * @param application     the CPF Viewer application
     * @param fileNameToLoad  name of the file to load
     * @param newFileName     name to give the file once loaded
     * @param groupToMerge    group to merge into the loaded file
     */
    public CpfUpdateGroupThread(CpfViewer application, String fileNameToLoad,
            String newFileName, GroupStatement groupToMerge) {
        super(application, fileNameToLoad, newFileName);
        this.fileNameToLoad = fileNameToLoad;
        this.groupToMerge = groupToMerge;
        this.application = application;
    }

    /**
     * After the background processing (load the file into the GUI) is done,
     * additional processing and GUI updates can now be done on the Event
     * Dispatch Thread.  This includes merging the new group into the file
     * that was loaded.
     */
    @Override
    protected void done() {
        firePropertyChange("done", null, null);
        application.statusBar.progress.clearBusy();
        String loadedFileName = null;
        try {
            // Make sure the file was successfully parsed
            loadedFileName = get();
        } catch (InterruptedException ex) {
            logger.error("Interrupted", ex);
            JOptionPane.showMessageDialog(application,
                    "Interrupted load of " + fileNameToLoad);
        } catch (ExecutionException ex) {
            logger.error("Error loading " + fileNameToLoad, ex);
            Throwable rootCause = ex;
            while (rootCause.getCause() != null) {
                rootCause = rootCause.getCause();
            }
            JOptionPane.showMessageDialog(application,
                    WordUtils.wrap("Error loading " + fileNameToLoad + " : "
                    + rootCause.getMessage(), 80));
        }
        if (loadedFileName == null) {
            return;
        }

        int loadedFileIndex = application.loadedFileList.
                getFileIndex(loadedFileName);
        if (loadedFileIndex == -1) {
            JOptionPane.showMessageDialog(application,
                "Error loading " + loadedFileName + " into the GUI");
            return;
        }
        logger.trace("Loaded file " + loadedFileName + " into list at "
                + loadedFileIndex);

        // Update the file attributes & Merge the group/parameter into this file
        LoadedFile loadedFileToUpdate = application.loadedFileList.
                getFile(loadedFileName);
        loadedFileToUpdate.updateFileName(loadedFileName);

        // add the parameter table panel
        ParameterTablePanel panel 
            = application.parameterViewerPanel.createNewPanel(loadedFileName);
        if (panel == null) {
            JOptionPane.showMessageDialog(application,
                "Error loading " + loadedFileName + " into the GUI -- " +
                "this file is already loaded.  Please use the Merge Group " +
                "function instead.");
            return;
        }
        // to make the actual group merge undoable, use the undoable model
        // after the panel has been created
        panel.getUndoableModel().mergeGroup(groupToMerge.getIdentifier(),
                groupToMerge);

        // update the parameter tree
        application.loadedFileList.refresh(loadedFileToUpdate);
    }
}
