package workers;

import gov.nasa.pds.tools.label.Label;
import gov.nasa.pds.tools.label.Statement;
import gov.nasa.pds.tools.label.parser.DefaultLabelParser;
import gov.nasa.pds.tools.label.parser.LabelParser;
import gov.nasa.pds.tools.label.parser.ParseException;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URL;
import java.net.URLConnection;
import java.util.ArrayList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * A threaded parser that breaks a file into chunks and passes each
 * chunk to the DefaultLabelParser in a separate thread.
 */
public class ThreadedLabelParser extends DefaultLabelParser {

    private long splitSize = 1000000;
    /* A value of 12 does not mean it will create 12 threads.  This field may
     * increase the splitSize (if potentially more than 12 threads could be
     * created with the current split size).  The actual number of threads will
     * be less, ~8 (for a normal-sized CPF), ~5 (OLI BPF) and none for a
     * TIRS BPF.  (Note also: the code assumes maxThreads >= 2).
     */
    private int maxThreads = 12;

    public void setsplitSize(long splitSize) {
        this.splitSize = splitSize;
    }
    private boolean useThreadingFlag = true;

    public void setUseThreadingFlag(boolean useThreading) {
        useThreadingFlag = useThreading;
    }

    @Override
    public Label parse(URL url) throws ParseException, IOException {

        if (!useThreadingFlag) {
            return super.parse(url);
        }

        URLConnection conn = url.openConnection();
        int size = conn.getContentLength();

        /* A size of -1 means we have a web resource whose content length
         * is unknown.  One example is using the REST service to obtain a
         * CPF file, such as:
         * http://.../dataservices/rest/cpf?operation=getCalParmFile&filename=
         *
         * The DefaultLabelParser has had problems using a web resource
         * directly, so we must download the file first, before passing it to
         * the parser.
         */
        if ((size > -1 && size < splitSize)) {
            conn.getInputStream().close();
            return super.parse(url);
        }
        /* Limit the number of threads created (by increasing the split size
         * for larger files).
         */
        long adjustedSplitSize = splitSize;
        if (size / maxThreads > splitSize) {
            adjustedSplitSize = size / maxThreads;
        }
        /* Compile the END_GROUP regex for speed. */
        Pattern endGroupPattern = Pattern.compile(
            "\\s*(END_GROUP|end_group)(\\s*=\\s*[A-Z][A-Z_0-9]*\\s*$|\\s*$)");

        ArrayList<URL> urls = new ArrayList<URL>();
        int seq = 0;
        BufferedReader in = 
            new BufferedReader(new InputStreamReader(conn.getInputStream()));
        File temp = File.createTempFile("tmp", "." + seq++);
        temp.deleteOnExit();
        BufferedWriter out = new BufferedWriter(new FileWriter(temp));
        urls.add(temp.toURI().toURL());

        long bytesRead = 0;
        String line = in.readLine();
        while (line != null) {
            bytesRead += line.length();
            if (bytesRead > adjustedSplitSize) {
                Matcher matcher = endGroupPattern.matcher(line);
                if (matcher.find()) {
                    out.write(line);
                    out.newLine();
                    out.write("END");
                    out.newLine();
                    out.close();

                    /* Check to see if this is the last group */
                    line = in.readLine();
                    if (line == null) {
                        break;
                    }
                    if (line.matches("\\s*(END|end)\\s*$")) {
                        /* Read next line to see if this is the truly the end
                         * of the file.
                         */
                        String nextLine = in.readLine();
                        if (nextLine == null) {
                            break;
                        } else {
                            line = line + "\n" + nextLine;
                        }
                    }
                    temp = File.createTempFile("tmp", "." + seq++);
                    temp.deleteOnExit();
                    out = new BufferedWriter(new FileWriter(temp));
                    urls.add(temp.toURI().toURL());
                    bytesRead = 0;
                    if (urls.size() >= maxThreads) {
                        /* If we've reached the maximum number of threads, we
                         * parse the rest of the source in the last thread. This
                         * could happen, if the content length of the resource
                         * is not known (size == -1) to begin with.
                         */
                        adjustedSplitSize = Long.MAX_VALUE;
                    }
                    continue;
                }
            }
            out.write(line);
            out.newLine();
            line = in.readLine();
        }
        in.close();
        out.close();

        /* Start processing the threads */
        ArrayList<LabelParserWorker> threads =
                new ArrayList<LabelParserWorker>();
        for (URL myURL : urls) {
            LabelParserWorker worker = new LabelParserWorker(myURL);
            worker.start();
            threads.add(worker);
        }

        /* Wait for each thread to finish */
        for (LabelParserWorker worker : threads) {
            try {
                worker.join();
            } catch (InterruptedException ignore) {
            }
        }

        /* Clean up temporary files */
        for (URL myURL : urls) {
            new File(myURL.getFile()).delete();
        }

        Label label = new Label();
        label.setFilename(library.LoadedFile.getFileNameFromUrl(url));
        /* Merge the Label results from the threads together */
        for (LabelParserWorker worker : threads) {

            Exception ex = worker.getException();
            if (ex != null) {
                if (ex instanceof ParseException) {
                    throw (ParseException) ex;
                } else {
                    if (ex instanceof IOException) {
                        throw (IOException) ex;
                    } else {
                        label.setStatus(FAIL);
                        label.incrementErrors();
                        break;
                    }
                }
            }
            Label workerLabel = worker.getLabel();
            if (workerLabel == null) {
                label.setStatus(FAIL);
                label.incrementErrors();
                break;
            }
            label.setStatus(workerLabel.getStatus());
            label.incrementErrors(workerLabel.getNumErrors());
            label.incrementWarnings(workerLabel.getNumWarnings());
            for (Object stmt : workerLabel.getStatements()) {
                label.addStatement((Statement) stmt);
            }
        }

        return label;
    }

    private class LabelParserWorker extends Thread {

        private URL url;
        private Label label = null;
        private Exception ex = null;

        public Label getLabel() {
            return label;
        }

        public Exception getException() {
            return ex;
        }

        LabelParserWorker(URL url) {
            this.url = url;
        }

        @Override
        public void run() {
            LabelParser lp = ThreadedLabelParserFactory.getInstance().
                    newThreadedLabelParser(false);
            try {
                label = lp.parse(this.url);
            } catch (Exception e) {
                this.ex = e;
            }
        }
    }
}
