package workers;

import gov.nasa.pds.tools.label.parser.LabelParser;
import gov.nasa.pds.tools.label.validate.CatalogNameValidator;
import gov.nasa.pds.tools.label.validate.DuplicateIdentifierValidator;

/**
 * Factory class used to create ThreadedLabelParsers.  A threaded label
 * label parser has the ability to parse sections of the file (e.g.
 * a set of group statements) in concurrent threads.  Threading itself is
 * optional, as seen in the newThreadedLabelParser method.  If you decline
 * it, the code will revert to using the parse method of the
 * DefaultLabelParser class.
 */
public class ThreadedLabelParserFactory {

    private static ThreadedLabelParserFactory factory = null;

    /**
     * Constructs a parser factory following the Singleton pattern
     */
    private ThreadedLabelParserFactory() {
    }

    /**
     * Retrieves a factory that creates ThreadedLabelParsers.
     * @return parser factory that will generate ThreadedLabelParsers
     */
    public synchronized static ThreadedLabelParserFactory getInstance() {
        if (factory == null) {
            factory = new ThreadedLabelParserFactory();
        }

        return factory;
    }

    /**
     * Retrieves a threaded parser that will read in PDS label files.
     * @param useThreading - whether threading should be used
     * @return The LabelParser interface
     */
    public LabelParser newThreadedLabelParser(Boolean useThreading) {
        LabelParser parser = new ThreadedLabelParser();

        /* The FileCharacteristicValidator was removed.  If left in, the
         * DefaultLabelParser returns a Label object with an incremented
         * numWarnings value.  The warning that the validator is flagging is
         * triggered because the labelType is undefined in DefaultLabelParser.
         */
        parser.addLabelValidator(new DuplicateIdentifierValidator());
        parser.addLabelValidator(new CatalogNameValidator());
        parser.addFragmentValidator(new DuplicateIdentifierValidator());

        ((ThreadedLabelParser) parser).setUseThreadingFlag(useThreading);
        return parser;
    }
}
