import gov.usgs.ias.dao.bpf.*;
import gov.usgs.ias.dao.cpf.*;
import gov.usgs.ias.cps.CpsStatusEnum;
import gov.usgs.ias.dao.ParseODL;
import gov.usgs.ias.dao.DataConnection;
import gov.nasa.pds.tools.label.Label;
import gov.usgs.ias.constant.Constants;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.Properties;
import java.util.regex.*;
import java.time.LocalDateTime;

/**
 * Front end for the insertFile method of the BpfDAO and CpfDAO classes.  This
 * class provides a way to insert all BPF/CPFs in a directory from the command
 * line.
 */
public class CpfLoader {
    private DataConnection conn;
    private String daoAuth = null;
    private CpsStatusEnum defaultCpfStatus = null;
    private boolean filesFound = false;
    private String schema = null;

    /**
     * Creates an instance of the CpfLoader class
     * @param schema Name of the schema to use - PROD or EVAL 
     */
    public CpfLoader(String schemaIn) {
        this.conn = new DataConnection("CPF_DB");
        this.schema = schemaIn;

        // Set up the authentication string for the proper schema
        if (this.schema != null &&
            this.schema.equals("PROD")) {
            System.out.println("Inserting files into PROD database");
            daoAuth = Constants.DAO_AUTH_IAS_PROD;
        }
        else{
            System.out.println("Inserting files into EVAL database");
            daoAuth = Constants.DAO_AUTH_IAS_EVAL;
        }
        defaultCpfStatus = CpsStatusEnum.ACTIVE;
    }

    /**
     * Closes the database connection
     */
    public void closeDb() {
        if (this.conn != null) {
            this.conn.close();
        }
    }

    /**
     * Set system properties from the properties file.  Will throw an exception
     * if the properties file can't be found or read.
     */
    private void initializeProperties() throws IOException {

        ClassLoader cl = CpfLoader.class.getClassLoader();
        URL propURL = cl.getResource(".cpfloader.properties");
        InputStream propStream = propURL.openStream();
        Properties appProperties = new Properties();
        appProperties.load(propStream);
        for (String pName : appProperties.stringPropertyNames()) {
            System.setProperty(pName, appProperties.getProperty(pName));
        }
        propStream.close();
    }

    /**
     * Inserts the specified BPF
     * @param bpfName    Path to the BPF to be inserted
     * @throws Exception if the insert failed
     */
    public void insertBpf(String bpfName) throws Exception{
        System.out.println("Parsing ODL file " + bpfName);
        Label bpfLabel = ParseODL.parseOdlFile(bpfName);
        System.out.println("Done Parsing ODL file");

        BpfDAO instance = new BpfDAO(this.conn, this.daoAuth);
        System.out.println(LocalDateTime.now() + " Inserting file into DB");
        int result = instance.insertFile(bpfLabel, defaultCpfStatus);
        System.out.println(LocalDateTime.now() +
                           " Done Inserting file into DB");
    }

    /**
     * Inserts the specified CPF
     * @param cpfName    Path to the CPF to be inserted
     * @throws Exception if the insert failed
     */
    public void insertCpf(String cpfName) throws Exception{
        
        System.out.println(" Parsing ODL file " + cpfName);
        Label cpfLabel = ParseODL.parseOdlFile(cpfName);
        System.out.println(" Done Parsing ODL file");

        CpfDAO instance = new CpfDAO(this.conn, this.daoAuth);
        System.out.println(LocalDateTime.now() + " Inserting file into DB");
        int result = instance.insertFile(cpfLabel, defaultCpfStatus);
        System.out.println(LocalDateTime.now() +
                           " Done Inserting file into DB");
    }

    /**
     * Load BPFs/CPFs found in specified directory
     * @param dirString    Path to the dir to be searched
     * @throws Exception   If the search or insert failed
     */
    public void findAndLoadCpfs(String dirString) throws Exception{
        File dir = new File(dirString);
        if (dir.isDirectory()){
            loadFilesInDir(dir);
        }
        else if(!dir.exists()){
            throw new Exception(dirString + " does not exist.");
        }
        else{
            throw new Exception(dirString + " is not a directory.");
        }
    }

    /**
     * Internal helper for findAndLoadCpfs. Searchs the passed dir
     * and calls insertBpf/insertCpf as appropriate.
     * @param dir       File object containing a directory
     * @throws Exception   If the search or insert failed
     */
    private void loadFilesInDir(File dir) throws Exception{
        Pattern bpfPattern;
        Pattern cpfPattern;
        if (this.schema.equals("PROD")){
            bpfPattern = Pattern.compile("L.*BPF.*[0-9][0-9]");
            cpfPattern = Pattern.compile("L.*CPF.*[0-9][0-9]");
        }
        // Not PROD, so its EVAL
        else{
            bpfPattern = Pattern.compile("eval_L.*BPF.*[0-9][0-9]");
            cpfPattern = Pattern.compile("eval_L.*CPF.*[0-9][0-9]");
        } 
        if (dir.isDirectory()){
            if (! dir.canRead()){
                throw new Exception (dir + " is not readable");
            }
            for (File entry : dir.listFiles()){
                Matcher bpfMatcher = bpfPattern.matcher(entry.getName());
                Matcher cpfMatcher = cpfPattern.matcher(entry.getName());

                // Load BPFs and CPFs we find
                if (entry.isDirectory()){
                    System.err.println("Warning: Skipping sub-dir: " +
                                       entry.getName());
                }
                else if (bpfMatcher.matches()){
                    insertBpf(entry.getAbsolutePath());
                    filesFound = true;
                }
                else if (cpfMatcher.matches()){
                    insertCpf(entry.getAbsolutePath());
                    filesFound = true;
                }
                else{
                    // Unknown file, do nothing
                }
            }
        }
    }

    /**
     * Main routine called when using the class from the command line. 
     * @param args         List of the command line arguments
     *         These are the expected arguments:
     *         Dir:      Directory containing BPFs and/or CPFs
     *         Database: PROD or EVAL
     */
    public static void main(String args[]) {
        if (args.length != 2) {
            System.err.println("Usage: load_bpf_cpf.sh <dir> <PROD/EVAL>");
            System.exit(1);
        }
        if (!args[1].equalsIgnoreCase("PROD") &&
            !args[1].equalsIgnoreCase("EVAL")){
            System.err.println("Invalid schema. Accepted values are PROD " +
                               "or EVAL");
            System.exit(1);
        }
        CpfLoader cpf = new CpfLoader(args[1].toUpperCase());
        try{
            cpf.initializeProperties();
            cpf.findAndLoadCpfs(args[0]);
        }
        catch (Exception ex){
            System.err.println("Exception: " + ex);
            ex.printStackTrace();
            System.exit(1);
        }
        finally{
            cpf.closeDb();
        }
        if (!cpf.filesFound){
            System.err.println("Warning: No BPFs/CPFs found");
            System.exit(1);
        }
    }
}
