/***************************************************************************
 NAME:     
     create_rlut_loader

 PURPOSE:  
     This application will read a RLUT file and create an Oracle loader
     file to be used by sqlldr to load the RLUT file into the DB.

 RETURNS:  
     Creates an Oracle DB loader file. 
****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>

#include "ias_logging.h"
#include "ias_types.h"
#include "ias_const.h"
#include "ias_rlut.h"
#include "ias_structures.h"
#include "ias_miscellaneous.h"

#include "ias_satellite_attributes.h"

/*----------------------------------------------------------------------------
NAME: verify_file_groups_and_data

PURPOSE:
   Verifies that the groups exist and that the data can be read successfully.

RETURNS:
   This function returns SUCCESS if it can verify the data and groups 
   ERROR if there is a problem while verifying the groups or data.
-----------------------------------------------------------------------------*/

static int verify_file_groups_and_data
(
    IAS_RLUT_IO *rlut       /* I: Used to get data for verification */
)
{
    int band_list_index;    /* Used to iterate through bands */
    int sca_number;         /* Used to iterate through SCAs on a band */
    int status;             /* Used to store group status */
    int linearity_band_numbers[IAS_MAX_TOTAL_BANDS];/* Band numbers for 
                                                       linearity bands */
    int linearity_count;    /* Total number of linearity bands */
    int linearization_band_numbers[IAS_MAX_TOTAL_BANDS];/* Band numbers for 
                                                           linearization */
    int linearization_count;    /* Total number of linearization bands */
    int tirs_band_numbers[IAS_MAX_TOTAL_BANDS]; /* Band numbers for TIRS */
    int tirs_count;     /* Total number of TIRS bands */
    int return_status = SUCCESS;    /* Return status for routine */

    /* Get the bands linearity group*/
    status = ias_sat_attr_get_any_sensor_band_numbers(IAS_OLI, IAS_NORMAL_BAND, 
        0, linearity_band_numbers, IAS_MAX_TOTAL_BANDS, &linearity_count);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("While getting linearity bands");
        return ERROR;
    }
    /* Copy linearity bands into linearization bands */
    memcpy(linearization_band_numbers, linearity_band_numbers,
        linearity_count * sizeof(int));

    /* Get the bands for all tirs */
    status = ias_sat_attr_get_any_sensor_band_numbers(IAS_TIRS,
        IAS_NORMAL_BAND|IAS_SECONDARY_BAND|IAS_BLIND_BAND, 0,
        &linearization_band_numbers[linearity_count], 
        IAS_MAX_TOTAL_BANDS - linearity_count, 
        &linearization_count);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("While getting linearization bands");
        return ERROR;
    }

    /* Add the linearity count to the linearization count */
    linearization_count += linearity_count;

    /* Get the bands for tirs normal and secondary excluding blind */
    status = ias_sat_attr_get_any_sensor_band_numbers(IAS_TIRS,
        IAS_NORMAL_BAND|IAS_SECONDARY_BAND, IAS_BLIND_BAND, tirs_band_numbers, 
        IAS_MAX_TOTAL_BANDS, &tirs_count);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("While getting TIRS bands");
        return ERROR;
    }
    
    /* Loop through the linearity bands */
    for (band_list_index = 0; band_list_index < linearity_count;
         band_list_index++)
    {
        int band_number = linearity_band_numbers[band_list_index];

        /* Get number of SCAs per band */
        int scas_per_band = ias_sat_attr_get_scas_per_band(band_number);
        if (scas_per_band == ERROR)
        {
            IAS_LOG_ERROR("While getting number of SCAs per band for the "
                "linearity bands for Band %d", band_number);
            return ERROR;
        }

        for (sca_number = 1; sca_number <= scas_per_band; sca_number++)
        {               
            /* Checking if group exists and is right data type */
            status = ias_rlut_verify_linearity_group(rlut, band_number,
                sca_number);

            /* If error return otherwise load data on success*/
            if (status == ERROR)
            {
                IAS_LOG_ERROR("While verifying dataset types"
                   " in Linearity lookup for Band %d SCA %d", band_number,
                   sca_number);
                return_status = ERROR;
            }
            else if (status == SUCCESS)
            {
                /* Read in the lookup table */
                IAS_RLUT_LOOKUP_TABLE *table = ias_rlut_read_lookup_table(rlut,
                    IAS_RLUT_LINEARITY_LOOKUP, 
                    band_number, sca_number);

                /* Check to see if table was read in */
                if (table == NULL)
                {
                    IAS_LOG_ERROR("Could not read Linearity lookup table"
                        " for Band %d SCA %d ", band_number, sca_number);
                    return_status = ERROR;
                }

                /* Free table memory and set to null */
                ias_rlut_free_lookup_table(table);
                table = NULL;
            }
        }
    }
    
    /* Loop through the linearization bands */
    for (band_list_index = 0; band_list_index < linearization_count; 
         band_list_index++)
    {
        int band_number = linearization_band_numbers[band_list_index];
        int detectors;

        /* Get the number of SCAs per band */
        int scas_per_band = ias_sat_attr_get_scas_per_band(band_number);
        if (scas_per_band == ERROR)
        {
            IAS_LOG_ERROR("While getting number of SCAs per band for the "
                "linearization bands for Band %d", band_number);
            return ERROR;
        }

        /* Grab number of detectors per SCA by band number */
        detectors = ias_sat_attr_get_detectors_per_sca(band_number);
        if (detectors == ERROR)
        {
            IAS_LOG_ERROR("While getting number of detectors per SCA for the "
                "linearization bands for Band %d", band_number);
            return ERROR;
        }

        for (sca_number = 1; sca_number <= scas_per_band; sca_number++)
        {
            /* Check to see if group exists data type is not checked just */
            status = ias_rlut_verify_linearization_group(rlut, band_number,
                sca_number);

            /* Check if group exists */
            if (status == ERROR)
            {
                IAS_LOG_ERROR("ERROR while verifying dataset types in "
                    "linearization lookup for Band %d SCA %d", band_number,
                    sca_number);
                return_status = ERROR;
            }
            else if (status == SUCCESS) 
            {
                /* Read in the parameters */
                 IAS_RLUT_LINEARIZATION_PARAMS *read_params = 
                    ias_rlut_read_linearization_params(rlut, band_number, 
                    sca_number, detectors);
                /* Check to ensure parameters were read properly */
                if (read_params == NULL)
                {
                    IAS_LOG_ERROR("Could not read linearization parameters "
                        "for Band %d SCA %d ", band_number, sca_number);
                    return_status = ERROR;
                }

                /* Free parameters and set to null */
                free(read_params);
                read_params = NULL;
            }
        }
    }
    
    /* Loop through the tirs bands */
    for (band_list_index = 0; band_list_index < tirs_count; band_list_index++)
    {
        int band_number = tirs_band_numbers[band_list_index];

        /* Get number of SCAs per band */
        int scas_per_band = ias_sat_attr_get_scas_per_band(band_number);
        if (scas_per_band == ERROR)
        {
            IAS_LOG_ERROR("While getting number of SCAs per band for the "
                "TIRS bands for Band %d", band_number);
            return ERROR;
        }

        for (sca_number = 1; sca_number <= scas_per_band; sca_number++)
        {
            /* Checking if group exists and is right data type */
            status = ias_rlut_verify_tirs_secondary_group(rlut, band_number,
                sca_number);

            /*If Error return otherwise load data on success */
            if (status == ERROR)
            {
               IAS_LOG_ERROR("While verifying dataset types in "
                    "TIRS lookup for Band %d SCA %d", band_number, sca_number);
               return_status = ERROR;
            }
            else if (status == SUCCESS)
            {
                /* Read in the lookup table */
                IAS_RLUT_LOOKUP_TABLE *table = ias_rlut_read_lookup_table(rlut,
                    IAS_RLUT_TIRS_SECONDARY_LOOKUP, band_number, sca_number);

                /* Check to ensure table was read properly */
                if (table == NULL)
                {
                    IAS_LOG_ERROR("Could not read TIRS lookup table "
                        "for Band %d SCA %d ", band_number, sca_number);
                    return_status = ERROR;
                }

                /* Free table memory and set to null */
                ias_rlut_free_lookup_table(table);
                table = NULL;
            }
        }
    }

    return return_status;
}

/*----------------------------------------------------------------------------
NAME: verify_file_attributes

PURPOSE:
   Will verify that the start date, end date, and version of filename match
   the attributes in the file contents.

RETURNS:
   This function returns SUCCESS if the attributes match else it will 
   return ERROR.
-----------------------------------------------------------------------------*/

static int verify_file_attributes
(
    const char *rlut_file_name,              /* I: Name of RLUT file */
    IAS_RLUT_FILE_ATTRIBUTES *rlut_file_attr /* I: RLUT file attributes */
)
{
    char * file_name;                   /* Temp file-name to get base */
    char effective_date_begin[9] = {0}; /* Begin date from the RLUT file */
    char effective_date_end[9] = {0};   /* End date from the RLUT file */
    char *source;                       /* Temp pointer for loops */ 
    char *destination;                  /* Temp pointer for loops */
    int status;                         /* Status holder */
    char expected_filename[45] = {0};   /* Buffer for filename fragment */
    int eval_offset = 5;                /* Eval prepend offset */

    /* Check if base name is passed */
    file_name = strrchr(rlut_file_name, '/');
    if (file_name != NULL)
    {
        rlut_file_name = file_name + 1;
    }
    file_name = NULL;
    
    if(strlen(rlut_file_attr->effective_date_begin) < 11 ||
        rlut_file_attr->effective_date_begin[4] != '-' ||
        rlut_file_attr->effective_date_begin[7] != '-' ||
        rlut_file_attr->effective_date_begin[10] != 'T')
    {
        IAS_LOG_ERROR("RLUT file attribute effective_date_begin (%s)"
            " not formatted correctly", rlut_file_attr->effective_date_begin);
        return ERROR;
    }

    if(strlen(rlut_file_attr->effective_date_end) < 11 ||
        rlut_file_attr->effective_date_end[4] != '-' ||
        rlut_file_attr->effective_date_end[7] != '-' ||
        rlut_file_attr->effective_date_end[10] != 'T')
    {
        IAS_LOG_ERROR("RLUT file attribute effective_date_end (%s)"
            " not formatted correctly", rlut_file_attr->effective_date_end);
        return ERROR;
    }

    /* Copy the string while stripping the '-' from the date for begin date */
    for (source = rlut_file_attr->effective_date_begin, 
        destination = effective_date_begin; *source != 'T'; source++)
    {
        if(*source != '-')
        {
            *destination = *source;
            destination++;
        }
    }

    /* Copy the string while stripping the '-' from the date for end date */    
    for (source = rlut_file_attr->effective_date_end, 
        destination = effective_date_end; *source != 'T'; source++)
    {
        if(*source != '-')
        {
            *destination = *source;
            destination++;
        }
    }

    /* Format the version */ 
    status = snprintf(expected_filename, sizeof(expected_filename), 
        "eval_LC08RLUT_%s_%s_%02d_%02d.h5", effective_date_begin,
        effective_date_end, rlut_file_attr->collection,
        rlut_file_attr->version );

    if ((status < 0) || (status >= sizeof(expected_filename)))
    {
        IAS_LOG_ERROR("Expected filename could not be formatted correctly");
        return ERROR;
    }

    /* Check if filename is prepended with 'eval_' */
    if (strncmp(rlut_file_name, "eval_", 5) == 0)
    {
        eval_offset = 0;
    }
    
    /* Check if expected filename matches the actual filename */
    if (strcmp(rlut_file_name , expected_filename + eval_offset) != 0)
    {
        IAS_LOG_ERROR("The file attributes and filename are not consistent. The"
            " expected filename is %s instead of the provided %s", 
             expected_filename + eval_offset, rlut_file_name);
        return ERROR;
    }

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME: read_and_verify_file_attributes

PURPOSE:
   Reads the FILE_ATTRIBUTES block from the RLUT hdf file. It will then
   verify the file attributes and datasets.

RETURNS:
   This function returns SUCCESS if the attributes were read and verified
   successfully, ERROR otherwise.
-----------------------------------------------------------------------------*/
static int read_and_verify_file_attributes
(
    const char *rlut_file_name,               /* I: name of RLUT hdf file */
    IAS_RLUT_FILE_ATTRIBUTES *rlut_file_attr  /* O: RLUT file attributes  */
)
{
    IAS_RLUT_IO *rlut = NULL;
    int status;
    int return_status = SUCCESS;

    /* Open the file */
    rlut = ias_rlut_open_file(rlut_file_name, IAS_READ);
    if (rlut== NULL)
    {
        IAS_LOG_ERROR("Error opening RLUT file %s ", rlut_file_name);
        return ERROR;
    }

    memset(rlut_file_attr, 0, sizeof(IAS_RLUT_FILE_ATTRIBUTES));
    
    /* Read RLUT file attributes */
    status = ias_rlut_read_file_attributes(rlut, rlut_file_attr);
    
    /* Check the return status */
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("File attributes read from RLUT file %s failed", 
              rlut_file_name);
        ias_rlut_close_file(rlut);
        return ERROR;
    }

    /* Verify that the file attributes and filename attributes match */
    status = verify_file_attributes(rlut_file_name, rlut_file_attr);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Could not verify that file attributes matched");
        return_status = ERROR;
    }

    /* Verify that the group data type are valid and data can be read */
    if (verify_file_groups_and_data(rlut) != SUCCESS)
    {
        IAS_LOG_ERROR("Could not verify the file groups and data");
        return_status = ERROR;
    }
 
    /* Close RLUT file */
    ias_rlut_close_file(rlut);

    return return_status;

}   /* read_and_verify_file_attributes */


/*----------------------------------------------------------------------------
NAME: write_loader

PURPOSE:
   Creates the Oracle DB loader file for the RLUT file.

RETURNS:
   This function returns SUCCESS if the ldr was written successfully,
   ERROR otherwise.
-----------------------------------------------------------------------------*/
static int write_loader
(
    char *rlut_file_name,                         /* I: Name of RLUT file */
    const char *loader_file_name,                 /* I: Name of ldr file */
    const IAS_RLUT_FILE_ATTRIBUTES rlut_file_attr /* I: RLUT file attributes */
)
{
    FILE *ldr_file = NULL;          /* Loader file to write to */
    IAS_DATETIME beg_date;          /* Begin date/time from file */
    IAS_DATETIME end_date;          /* End date/time from file */
    IAS_DATETIME baseline_date;     /* baseline date/time from file */
    char *escaped_description = NULL; /* Escaped file description */
    int num_quotes = 0;             /* Number of quotes in string */
    int orig_desc_len = 0;          /* Description length of original */
    int escaped_desc_len = 0;       /* Description length with extra quotes */
    int iteration = 0;              /* Loop variable */
    int escaped_index = 0;          /* Current index of escaped string */
    char *rlut_base_name = NULL;    /* Only base part of RLUT filename */
    int status;                     /* Status holder */
    int overall_status = SUCCESS;   /* overall_status returned at end */

    /* We want to enter the rlut file name without path */
    rlut_base_name = basename(rlut_file_name);

    /* Parse the effectivity begin date string */   
    if (ias_misc_parse_datetime_string(rlut_file_attr.effective_date_begin,
                IAS_DATETIME_CPF_FORMAT, &beg_date) == ERROR) 
    {
        IAS_LOG_ERROR("begin date cannot be parsed");
        return ERROR;
    }

    /* Parse the effectivity end date string */   
    if (ias_misc_parse_datetime_string(rlut_file_attr.effective_date_end,
                IAS_DATETIME_CPF_FORMAT, &end_date) == ERROR)
    {
        IAS_LOG_ERROR("end date cannot be parsed");
        return ERROR;
    }

    /* Parse the baseline date string */   
    if (ias_misc_parse_datetime_string(rlut_file_attr.baseline_date,
                IAS_DATETIME_CPF_FORMAT, &baseline_date) == ERROR)
    {
        IAS_LOG_ERROR("baseline date cannot be parsed");
        return ERROR;
    }

    /* Create a new description string with double quotes escaped by two double
       quotes (i.e. " --> "") so we can delimit the string with double quotes 
       and that allows the commas (which are the field terminators). */

    /* ... count the number of double quotes in the original string */
    orig_desc_len = strlen(rlut_file_attr.description);
    for (iteration = 0; iteration < orig_desc_len; iteration++)
    {
        if (rlut_file_attr.description[iteration] == '"')
        {
            num_quotes++;
        }
    }

    /* Allocate a new string long enough to duplicate each double quote */
    escaped_desc_len = orig_desc_len + num_quotes;

    escaped_description = malloc(escaped_desc_len+1);
    if (escaped_description == NULL)
    {
        IAS_LOG_ERROR("cannot allocate escaped description string");
        return ERROR;
    }

    /* Loop through copying each original character in the string, but 
       inserting an extra double quote for each one found */
    for (iteration = 0; iteration <= orig_desc_len; iteration++)
    {
        if (rlut_file_attr.description[iteration] == '"')
        {
            escaped_description[escaped_index++] = '"';
        }
        escaped_description[escaped_index++] = 
            rlut_file_attr.description[iteration];
    }
    
    /* Open the output loader file */
    ldr_file = fopen(loader_file_name, "w");

    if (ldr_file == NULL)
    {
        IAS_LOG_ERROR("Cannot open %s for writing", loader_file_name);
        return ERROR;
    }

    status = fprintf(ldr_file, "LOAD DATA\n"
                                "INFILE *\n"
                                "INTO TABLE rlut_main\n"
                                "APPEND\n"
                                "FIELDS TERMINATED BY \",\" "
                                "OPTIONALLY ENCLOSED BY '\"'\n"
                                "(\n"
                                "  file_name,\n"
                                "  file_source,\n"
                                "  version_number,\n"
                                "  collection_number,\n"
                                "  status,\n"
                                "  description CHAR(4000),\n"
                                "  effective_date_begin DATE "
                                                  "\"MM-DD-YYYY HH24:MI:SS\",\n"
                                "  effective_date_end DATE "
                                                  "\"MM-DD-YYYY HH24:MI:SS\",\n"
                                "  date_created DATE "
                                "\"MM-DD-YYYY HH24:MI:SS\",\n"
                                "  date_entered EXPRESSION \"SYSTIMESTAMP\",\n"
                                "  rlut_data LOBFILE (file_name) "
                                                  "TERMINATED BY EOF\n"
                                ")\n" 
                                "BEGINDATA\n" 
                                "%s,%s,%d,%d,%s,\"%s\","  
                                "%02d-%02d-%04d %02d:%02d:%02.0f,"  
                                "%02d-%02d-%04d %02d:%02d:%02.0f,"  
                                "%02d-%02d-%04d %02d:%02d:%02.0f\n",
                                rlut_base_name,   
                                rlut_file_attr.file_source,
                                rlut_file_attr.version,
                                rlut_file_attr.collection,
                                rlut_file_attr.status,  
                                escaped_description,  
                                beg_date.month, beg_date.day_of_month, 
                                beg_date.year, beg_date.hour, beg_date.minute, 
                                beg_date.second, end_date.month, 
                                end_date.day_of_month, end_date.year, 
                                end_date.hour, end_date.minute, end_date.second,
                                baseline_date.month, baseline_date.day_of_month,
                                baseline_date.year, baseline_date.hour, 
                                baseline_date.minute, baseline_date.second);
    if (status < 0)
    {
        IAS_LOG_ERROR("Failure to write output file");
        overall_status = ERROR;
    }

    free(escaped_description);
    status = fclose(ldr_file);
    if (status < 0)
    {
        IAS_LOG_ERROR("Failure to close output file");
        overall_status = ERROR;
    }

    return overall_status;
}

/*----------------------------------------------------------------------------
NAME: main

PURPOSE:
   Main routine for creating the .ldr file from a RLUT hdf file.

RETURNS:
   This function returns 0 if the ldr was written successfully,
   ERROR otherwise.
-----------------------------------------------------------------------------*/
int main
(
    int argc,
    char *argv[]
)
{
    IAS_RLUT_FILE_ATTRIBUTES rlut_file_attr;

    char *rlut_file_name = NULL;
    char *ldr_file_name = NULL;
    char *prog_name = NULL;
    int status = SUCCESS;
    
    /* Construct the program name from exe */
    prog_name = strrchr(argv[0], '/');
    if (prog_name != NULL )
    {
        prog_name++;
    }
    else
    {
        prog_name = argv[0];
    }

    /* Initialize the IAS log */
    status = ias_log_initialize(prog_name);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("cannot initialize Log file");
        return EXIT_FAILURE;
    }
     
    /* Initialize the satellite attributes library */
    if (ias_sat_attr_initialize(IAS_L8) != SUCCESS)
    {
        IAS_LOG_ERROR("Could not initialize satellite library");
        return ERROR;
    }

    /* Check if the correct program usage was made */
    if (argc < 3)
    {
        IAS_LOG_ERROR("usage: %s <rlut_hdf_file> <ldr_file>\n", prog_name);
        return EXIT_FAILURE;
    }
    
    /* Assign the program variables from command line arguments */
    rlut_file_name = argv[1];
    ldr_file_name = argv[2];

    /* Read file attributes from file */
    status = read_and_verify_file_attributes(rlut_file_name, &rlut_file_attr);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("cannot read RLUT file");
        return EXIT_FAILURE;
    }

    /* Write the file attributes to ldr file */
    status = write_loader(rlut_file_name, ldr_file_name, rlut_file_attr);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("cannot create Oracle loader file");
        return EXIT_FAILURE;
    }
    
    return EXIT_SUCCESS;
}
