#!/usr/bin/perl -w

################################################################################
#
# Name: cps_validate.pl
#
# Purpose:
#    To validate that the groups and parameters within a given BPF or CPF file
#    are correct, according to a corresponding standard template (BPF or CPF).
#
#    NOTE: This version will not handle nested groups.
#    The script will read a file from STDIN but will require the file name in
#    the argument list in order to determine which template to grab for the
#    validation process.
#
# Parameters: n/a
#
# Syntax: cps_validate.pl --[option] [value] <OLI BPF filename>|
#         <TIRS BPF filename>|<CPF filename>
#
################################################################################

use strict;
use Getopt::Long;
use ias_const;
use ias_date_utils;

# Maximum length of an ODL text line.  The C-based ODL library will have
# issues parsing any lines longer than this.  The constant is defined in the
# 'toolbox.h' header file associated with the NASA Planetary Data System
# (PDS) Label Library "Lite" (which the ODL library provides a wrapper for)
use constant TB_MAXLINE => 2048;

my $infh;               # input file handle
my $standardIn = 0;     # standard in flag
my $help = 0;           # help flag
my $errors = 1;         # error flag default setting.
my $warnings = 1;       # warning flag default setting.
my $formatMismatch = 1; # format error flag default setting
my $cpsType = "na";     # Value is "not applicable" if the type option
                        # has not been selected.
my $path;               # path to cps file to validate
my $suffix;             # file name suffix
my $errorFlag = FALSE;  # flag to ensure we exit with the correct status
my $foundEnd = FALSE;   # flag to determine END tag at eof
my %groupAudit = ();    # track the group occurance to report duplicate groups
my %paramAudit = ();    # track the group/parameter occurance to
                        # report duplicates
my $groupName;          # group hash key
my $parameterName;      # parameter hash key
my $clOptions;          # GetOptions return value
my $cpsFile = '';       # cps file to validate
my $cpsFilePath;        # cps file and path to validate
my $gorder;             # numerical order group appears in template and file
my $endGroup;           # end group name and label
my %groupOrder;         # hash of group labels in order of appearance (template)
my %fileGroupOrder;     # hash of group labels in order of appearance (cps file)
my %pvalue;             # hash of parameter values
my $pnum;               # parameter counter to record order of appearance
my $pname;              # parameter name
my $pRangeStart;        # The first number in parameter names that have a
                        # range that is to be expanded
my $pRangeStop;         # The final number in parameter names that have a
                        # range that is to be expanded
my $pRangeIter;         # An interator to use during parameter expansion
my $pDigitCount;        # The number of digits we will expand (zero-pad)
                        # the range numbers to
my @pExpandedParms;     # An array to store all the values a parameter name
                        # was expanded to
my $pTempParms;         # Temporary storage for elements of the above array
my %parameterOrder;     # hash of parameters in order of appearance (template)
my %fileParamOrder;     # hash of parameters in order of appearance (cps file)
my $predecessorGroupOrder; # previous group to determine sequence
my $predecessorParamOrder; # previous parameter to determine sequence
my $util;               # data place holder to build hashes
my $beginGroup;         # place holder to check if group is correct place
my $key;                # hash loop variable
my $gkey;               # hash loop variable group hash
my $pkey;               # hash loop variable parameter hash
my $pKey;               # key value to locate value in parameter hash
my $pFormat;            # place holder for template format
my $status;             # subroutine return status
my $beginGroupFlag = FALSE; # flag indicating found begin group
my $cpsTemplate;        # name of template to use
my $cpsTemplatePath;    # path and name of template to use

my $ecNumLeapSeconds = undef; # Number of Leap Seconds (used as array size)

# Valid 3 character months (double quotes needed for comparisons)
my @monthArray = ('"Jan"','"Feb"','"Mar"','"Apr"','"May"','"Jun"',
    '"Jul"','"Aug"','"Sep"','"Oct"','"Nov"','"Dec"');
# Valid PREDEFINED lists used for list validation
my %predefinedList = ('MONTHS' => \@monthArray);

my $templateRoot = $ENV{IAS_DATA_DIR};
my $templatePath = $templateRoot . "/cps";

# Get the command line options.
$clOptions = GetOptions("help!"       => \$help,
                        "errors!"      => \$errors,
                        "warnings!"    => \$warnings,
                        "format!"      => \$formatMismatch,
                        "type=s"       => \$cpsType,
                        "standardin"   => \$standardIn
                       );
if ($help eq TRUE)
{
    &cpsHelp();
    exit SUCCESS;
}
if ($clOptions != TRUE)
{
    &cpsHelp();
    exit FAILURE;
}

# Get the CPS filename.
$cpsFilePath = shift;
if (not defined $cpsFilePath)
{
    &cpsHelp();
    print "No filename\n";
    exit FAILURE;
}

chomp($cpsFilePath);

my @cpsPath = split ('\/', $cpsFilePath);
$cpsFile = $cpsPath[$#cpsPath];

# determine template type from cps type option if na then use file name
# to determine template type, if invalid type print help
if ($cpsType =~ /^(c|cp|cpf)$/i)
{
    if ($cpsFile =~ /CPF/)
    {
        $cpsTemplate = "CPF_Template";
    }
    else
    {
        print "Type option has invalid value (must match CPF file).\n";
        &cpsHelp();
        exit FAILURE;
    }
}
elsif ($cpsType =~ /^(o|ob|olib|obpf|olibpf)$/i)
{
    if ($cpsFile =~ /LO8BPF/)
    {
        $cpsTemplate = "OLI_BPF_Template";
    }
    else
    {
        print "Type option has invalid value (must match OLI BPF file).\n";
        &cpsHelp();
        exit FAILURE;
    }
}
elsif ($cpsType =~ /^(t|tb|tbpf|tirsb|tirsbpf)$/i)
{
    if ($cpsFile =~ /LT8BPF/)
    {
        $cpsTemplate = "TIRS_BPF_Template";
    }
    else
    {
        print "Type option has invalid value (must match TIRS BPF file).\n";
        &cpsHelp();
        exit FAILURE;
    }
}
elsif ($cpsType eq "na")
{
    if ($cpsFile =~ /CPF/)
    {
        $cpsTemplate = "CPF_Template";
    }
    elsif ($cpsFile =~ /LO8BPF/)
    {
        $cpsTemplate = "OLI_BPF_Template";
    }
    elsif ($cpsFile =~ /LT8BPF/)
    {
        $cpsTemplate = "TIRS_BPF_Template";
    }
    else
    {
        print "Could not determine template type from file name.\n";
        &cpsHelp();
        exit FAILURE;
    }
}
else
{
    print "Could not determine template type, invalid cps type.\n";
    &cpsHelp();
    exit FAILURE;
}

# determine whether to use a local directory template or the installed one
if (-e $cpsTemplate)
{
    $cpsTemplatePath = $cpsTemplate;
    print "Template file used is from current directory: $cpsTemplate\n";
}
else
{
    $cpsTemplatePath = $templatePath . "/" . $cpsTemplate;
    print "Template file used is: $cpsTemplatePath\n";
}

# read the template into hash
open (TF, $cpsTemplatePath)
    or die("Can't open CPS template file: $cpsTemplatePath!\n");

# initialize some variables.
$gorder = 0;
$pnum = 0;

$groupName = '';

# loop through the file and read in
while (<TF>)
{
    # get rid of the new line
    chomp;

    # Ignore comments and blank lines.
    next if (/^#/ or /^\s*$/);

    # We've found the start of a group.
    if (/^\s*GROUP\s*=\s*(\S+)/)
    {
        # new group set the parameter counter to zero
        $pnum = 0;

        # Build the key
        $groupName = $1;

        # Record the group's position.
        $gorder += 1;
        $groupOrder{$groupName} = $gorder;

        # Initialize the group audit hash
        $groupAudit{$groupName} = 0;
        next;
    }

    # We've found the end of a group.
    if (/^\s*END_GROUP\s*=\s*(\S+)/)
    {
        if ($groupName ne $1)
        {
            print "Template Group End label: $1 does not match "
                . "start name: $groupName\n";
            exit FAILURE;
        }
        undef $groupName;
        next;
    }
    # We're at the end of the template.
    last if (/^\s*END\s*$/);

    # We're looking at a parameter within a group.
    if (not defined $groupName)
    {
        print "Template groupName undefined for parameter: $pname\n";
        exit FAILURE;
    }

    @pExpandedParms = ();
    # Add the parameter to the array. If a parameter needs to be expanded,
    # its expanded values are added to end of the array. When a parameter
    # gets added, it gets removed from the array. The loop continues until
    # the array is empty.
    push(@pExpandedParms, $_);
    while (scalar(@pExpandedParms) >= 1)
    {
        $pTempParms = $pExpandedParms[0];
        # see if this parameter entry needs to be expanded into multiple names
        if ($pTempParms =~ m/(.*?)\[(\d+)-(\d+)\](.*=.+)/)
        {
            $pDigitCount = length($3);
            $pRangeStart = int($2);
            $pRangeStop = int($3);
            if($pRangeStart > $pRangeStop)
            {
                print "Invalid expansion range for $_\n";
                exit FAILURE;
            }

            # Remove the parm with the expansion brackets and add in the
            # range of parm names to the array
            shift(@pExpandedParms);
            for ($pRangeIter = $pRangeStart; $pRangeIter <= $pRangeStop;
                    $pRangeIter++)
            {
                push(@pExpandedParms, sprintf("%s%0".$pDigitCount."i%s",
                            $1, $pRangeIter, $4));
            }
        }
        # fully-expanded parameter, so add it to the parameter hash
        elsif ($pTempParms =~ m/\s*(\S+)\s*=\s*(.+)\s*/)
        {
            # Record its position within the group and its value.
            $pnum += 1;
            $pname = $1;
            $parameterName = $groupName . '/' . $pname;
            $parameterOrder{$parameterName} = $pnum;
            $pvalue{$parameterName} = $2;

            # initialize the parameter audit hash
            $paramAudit{$parameterName} = 0;

            # remove the parm for the expansion array, since we have added
            # it to the hash
            shift(@pExpandedParms);
        }
        else
        {
            print "Unable to get parameter name and value from group: ".
                "$groupName\n";
            exit FAILURE;
        }
    }
}
close TF;

# See if the call is from standard in to set the file handle
if ($standardIn == TRUE)
{
    $infh = *STDIN;
}
else
{
    open ($infh, "<", $cpsFilePath)
        or die "Can't open file: $cpsFilePath  \n";
}

# reset some working variables.
$gorder = 0;
$pnum = 0;

print "Reading input file: $cpsFile...\n" unless $standardIn;

# Look through the CPS file...
while (<$infh>)
{
    # get rid of dos-style carriage-returns and the new line
    s/\r$//;
    chomp;

    # Ignore comments and blank lines.
    next if (/^#/ or /^\s*$/);

    # check to make sure we don't have an "extremely" long line (i.e. more
    # than the ODL limit of TB_MAXLINE).
    my $currStrLength = length($_);
    if ($currStrLength > TB_MAXLINE)
    {
        $errorFlag = TRUE;
        if ($errors == TRUE)
        {
            print "ERROR:  Line $. is $currStrLength characters long:\n"
                . substr($_, 0, 80) . "\n";
        }
    }

    # We've found the start of a group.
    if (/^\s*GROUP\s*=\s*(\S+)/)
    {
        # new group; reset the parameter counter
        $pnum = 0;

        # check to see of the previous group had a matching end group
        if ($beginGroupFlag == TRUE)
        {
            $errorFlag = TRUE;
            if ($errors == TRUE)
            {
                print "New group found ($_) before end of previous "
                    . "group ($beginGroup)\n";
            }
        }
        # Save the new group name and build the key.
        $beginGroup = $groupName = $1;

        # Is this a valid group?
        if ($groupOrder{$groupName})
        {
            # increment the group tracker
            $groupAudit{$groupName} +=1;

            # Check the group is in the right position.
            $gorder += 1;

            # add group to the hash to track group sequence
            $fileGroupOrder{$groupName} = $gorder;

            # Check that the group has not already been found and is in order:
            # groupAudit has the template group order of appearance
            # groupName is the current group name
            if ($groupAudit{$groupName} > 1)
            {
                $errorFlag = TRUE;
                if ($errors == TRUE)
                {
                    print "Group $beginGroup, groupOrder tag: "
                        .   "$groupOrder{$groupName}, groupName: $groupName "
                        .   "appears multiple times in input file.\n";
                }
            }
        } # if groupOrder{groupName}
        else
        {
            $errorFlag = TRUE;
            if ($errors == TRUE)
            {
                print "Group $groupName shouldn't be there!\n";
            }
        }

        # set begin group flag
        $beginGroupFlag = TRUE;
        next;
    }

    # We've found the end of a group.
    if (/^\s*END_GROUP\s*=\s*(\S+)/)
    {
        # reset the flag
        $beginGroupFlag = FALSE;

        $endGroup = $1;

        # Check whether the begin and end tags match
        if ($endGroup ne $beginGroup)
        {
            $errorFlag = TRUE;
            if ($errors == TRUE)
            {
                print "Beginning of group ($beginGroup) and "
                    . "end of group ($endGroup) do not match.\n";
            }
        }

        undef $groupName;
        next;
    }

    # We're at the end of the CPS, set the flag to report errors if missing
    if (/^END$/)
    {
        $foundEnd = TRUE;
        last;
    }

    # We're looking at a parameter within a group.

    # If the current line matches a "parameter =" pattern on a line without
    # anything following it, merge the next line to the current line
    if (/^\s*\S+\s*=\s*$/)
    {
        my $merged_line = $_;
        $merged_line .= <$infh>;
        chomp($merged_line);
        $_ = $merged_line;
    }

    # get the parameter name
    if (/^\s*(\S+)\s*=/)
    {
        $pname = $1;
    }
    else
    {
        print "Invalid parameter: $_\n";
        $errorFlag = TRUE;
        next;
    }

    # Ensure parameter is bracketed by begin and end group labels.
    if ($beginGroupFlag == FALSE)
    {
        print "Parameter: $pname appears outside of a group.\n";
        $errorFlag = TRUE;
        next;
    }

    # For multi-line arrays, check if there is a ) if there is a ( and get
    # the next line if not
    chomp($util = $_);

    # loop to the end of a multi line parameter .i.e detector level params
    while (( $util =~ /\(/ ) and ( $util !~ /\)/ ))
    {
        chomp($util .= <$infh>);
    }

    # build the hash key to find the parameter format info from the template
    $pKey = $groupName.'/'.$pname;
    $pFormat = $pvalue{$pKey};

    # get rid of spaces between nominal values if any defined
    # used to keep from throwing perl error if undef due to parameter
    # in wrong group, that error is reported else where.
    if (defined($pFormat))
    {
        $pFormat =~ s/,\s*/,/g;
    }

    # Is this a valid parameter in the correct order?
    if ($parameterOrder{$pKey})
    {
        $pnum += 1;

        # add the parameter to the fileParamOrder hash to validate sequence
        $fileParamOrder{$pKey} = $pnum;

        # set the audit trail
        $paramAudit{$pKey} +=1;

        # get the parameter values
        if ($util =~ /^\s*\S+\s*=\s*(.+)\s*$/)
        {
            $util = $1;

            # check for multiline strings
            # will start with a "
            if ($util =~ /^"/)
            {
                # append until an ending " is found
                while ($util !~ /"\s*$/)
                {
                    chomp($util .= <$infh>);
                }
            }
        }
        else
        {
            print "ERROR: Failed to get values from $util\n";
            $errorFlag = TRUE;
            next;
        }

        $status = &checkFormat($util, $pKey, $pFormat);
        if ($status != SUCCESS)
        {
            $errorFlag = TRUE;
        }
    }
    else
    {
        if ($warnings == TRUE)
        {
            print "WARNING: Parameter $pname should not be in "
                . "group $groupName!\n";
            $errorFlag = TRUE;
        }
    }
}
close $infh unless $standardIn;

# check the end of file flag
if ($foundEnd != TRUE)
{
    if ($errors == TRUE)
    {
        print "File END tag missing!\n";
        $errorFlag = TRUE;
    }
}

# check the groups to ensure they are in the proper order
$predecessorGroupOrder = 0;
foreach $gkey (sort { $groupOrder{$a} <=> $groupOrder{$b} }
    keys %groupOrder)
{
    # skip groups that were not found
    next unless (defined($fileGroupOrder{$gkey}));

    # Make sure this group's location is after the group that should
    # have been its predecessor
    if ($fileGroupOrder{$gkey} < $predecessorGroupOrder)
    {
        if ($warnings == TRUE)
        {
            print "WARNING: Group $gkey appears to be out of order.\n";
            $errorFlag = TRUE;
        }
    }

    # Now this group should be the predecessor of the next one
    $predecessorGroupOrder = $fileGroupOrder{$gkey};

    # check the prameters within this group.
    $predecessorParamOrder = 0;
    foreach $pkey (sort { $parameterOrder{$a} <=> $parameterOrder{$b} }
        grep {/^$gkey\//} keys %parameterOrder)
    {
        # skip parameters that were not found
        next unless (defined($fileParamOrder{$pkey}));

        # Make sure this parameter's location is after the parameter that
        # should have been its predecessor
        if ($fileParamOrder{$pkey} < $predecessorParamOrder)
        {
            if ($warnings == TRUE)
            {
                $errorFlag = TRUE;
                print "WARNING: Parameter $pkey appears to be out of order.\n";
            }
        }
        # Now this parameter should be the predecessor of the next one
        $predecessorParamOrder = $fileParamOrder{$pkey};
    }
}

# check the group index to make sure we have all the groups
foreach $key (keys %groupOrder)
{
    if ($groupAudit{$key} == 0)
    {
        $errorFlag = TRUE;
        if ($errors == TRUE)
        {
            print "ERROR: Group $key was not found!\n";
        }
    }
}

# check the param Audit hash to make sure we have found all the parameters
# and that we have no duplicates in the file.
foreach $key (keys %paramAudit)
{
    if ($paramAudit{$key} == 0)
    {
        $errorFlag = TRUE;
        if ($errors == TRUE)
        {
            print "ERROR: Parameter $key was not found!\n";
        }
    }
    if ($paramAudit{$key} > 1)
    {
        $errorFlag = TRUE;
        if ($errors == TRUE)
        {
           print "ERROR: Parameter $key is a duplicate!\n";
        }
    }
}

if ($errorFlag != TRUE)
{
    print "INFO: Validation completed successfully\n";
    exit SUCCESS;
}
else
{
    print "ERROR: Validation completed unsuccessfully\n";
    exit FAILURE;
}

###############################################################################
#
# Name: checkFormat
#
# Purpose:
#    Checks the formats of the parameters and validates whether or not
#    the formats used in the specified CPS file are correct.
#
# Returns:
#
# Parameters: actual: value
#             pname:  group/parameter name
#             util:   format strings: datatype array nominal values
#                                     template format
#
###############################################################################
sub checkFormat
{
    my ($actual, $pname, $paramAttribs) = @_;
    my $timeBegin;
    my $dateBegin;
    my $calcDateBegin;
    my $timeEnd;
    my $dateEnd;
    my $calcDateEnd;
    my $middate;
    my $delta;
    my @theArray;
    my @actualArray;
    my @ideal;
    my $type;
    my $form;
    my $utilBegin;          # place holder file name begin date
    my $utilEnd;            # place holder file name end date
    my $startdate;          # calculated start date of ut1/utc parameters
    my $enddate;            # calculated end date of ut1/utc parameters
    my $i;                  # loop counter
    my $fval;               # format value
    my $min;                # place holder, minimum value in range
    my $max;                # place holder, maximum value in range
    my $saveActual;         # place holder to save array for when time to split
    my @nominalArray;       # array of nominal values
    my $nominalVal;         # nominal value
    my @listArray;          # list of valid domain values from template
    my $fileName;           # cps file name
    my $template;           # template value to match
    my $arrValue;           # value from array to work with
    my $format;             # template format to match against parm value
    my $arrayLength = 0;    # array length from template
    my $month;              # contains month digit
    my $day;                # contains day digit
    my $year;               # contains year digit
    my $error_msg;          # function return error message
    my $returnStatus = SUCCESS; # local error status

    $saveActual = $actual;
    # stuff the format string in an array
    @ideal = split(/\s+/, $paramAttribs);

    # Get parameter data type.
    $type = shift @ideal;
    if (not defined $type)
    {
        if ($warnings == TRUE)
        {
            print "WARN: Parameter data type not defined: $pname \n";
            $returnStatus = ERROR;
        }
    }

    # Is this a special type of parameter with hard-coded validation test?
    if ($type eq 'special')
    {
        # Parse filename for filename and date checks.
        $fileName = $cpsFile;

        if ($pname eq 'FILE_ATTRIBUTES/File_Name')
        {
            # put double quotes around name
            $fileName = '"'.$fileName.'"';
            if ($fileName ne $actual)
            {
                $returnStatus = ERROR;
                if ($errors == TRUE)
                { 
                    print "ERROR: Parameter File_Name does not match "
                        . "file name!  ($fileName != $actual)\n";
                }
            }
        }

        if ($pname eq 'FILE_ATTRIBUTES/Version')
        {
            my $version;
            if ($fileName =~ m/CPF.*_\d\d\.(\d\d)/)
            {
                $version = int($1);
            }
            else
            {
                $returnStatus = ERROR;
                if ($errors == TRUE)
                {
                    print "ERROR: Unable to extract version from filename\n";
                }
            }
            if ($actual != $version)
            {
                $returnStatus = ERROR;
                if ($errors == TRUE)
                { 
                    print "ERROR: Version in filename does not match version "
                        . "in file: $version != $actual\n";
                }
            }
        }
        if ($pname eq 'FILE_ATTRIBUTES/Collection_Number' )
        {
            my $collection;
            if ($fileName =~ m/CPF.*_(\d\d)\.\d\d/)
            {
                $collection = int($1);
            }
            else
            {
                $returnStatus = ERROR;
                if ($errors == TRUE)
                {
                    print "ERROR: Unable to extract collection from filename\n";
                }
            }
            if ($actual != $collection)
            {
                $returnStatus = ERROR;
                if ($errors == TRUE)
                { 
                    print "ERROR: Collection in filename does not match "
                        . "collection in file: $collection != $actual\n";
                }
            }
        }

        if (($pname eq 'FILE_ATTRIBUTES/Effective_Date_Begin')
                or ($pname eq 'FILE_ATTRIBUTES/Effective_Date_End'))
        {
            #get rid of hyphens (-)
            $actual =~ s/-//g;

            # get rid of leading and trailing "
            $actual =~ s/\"//g;

            #get the begin and ending dates
            $fileName =~ /(?:LC08CPF_|LO8BPF|LT8BPF)(\d+)_(\d+)/ and do {
                $utilBegin = $1;
                $utilEnd = $2;
            };

            # Match the Effective start date from the filename to the parameter.
            if ($pname eq 'FILE_ATTRIBUTES/Effective_Date_Begin')
            {
               if ($fileName =~ /CPF/)
               {
                  # get rid of the T in time stamp & pick off date
                  ($dateBegin, $timeBegin) = split('T', $actual);
                  if ($utilBegin ne $dateBegin)
                  {
                      $returnStatus = ERROR;
                      if ($errors == TRUE)
                      {
                          print "ERROR: Parameter Effective_Date_Begin does "
                            . "not match filename:\n"
                            .  "  ($dateBegin != $utilBegin)\n";
                      }
                  }
                }
                elsif ($fileName =~ /LO8BPF/ || $fileName =~ /LT8BPF/)
                {
                    # get rid of the T in time stamp
                    $actual =~ s/T//;
                    # get rid of the colons
                    $actual =~ s/://g;
                    if ($utilBegin ne $actual)
                    {
                        $returnStatus = ERROR;
                        if ($errors == TRUE)
                        {
                            print "ERROR: Parameter Effective_Date_Begin "
                                 . "does not match filename:\n"
                                 . "       ($actual != $utilBegin)\n";
                        }
                    }
                }
            }

            # Match the Effective end date from the filename to the parameter.
            if ($pname eq 'FILE_ATTRIBUTES/Effective_Date_End')
            {
               if ($fileName =~ /CPF/)
               {
                    # get rid of the T in time stamp & pick off date
                    ($dateEnd, $timeEnd) = split('T', $actual);
                    if ($utilEnd ne $dateEnd)
                    {
                        $returnStatus = ERROR;
                        if ($errors == TRUE)
                        {
                            print "ERROR: Parameter Effective_Date_End "
                                 . "does not match filename:"
                                 . "       ($dateEnd != $utilEnd)\n";
                        }
                    }
               }
               elsif ($fileName =~ /LO8BPF/ || $fileName =~ /LT8BPF/)
               {
                    #get rid of the T in the time stamp
                    $actual =~ s/T//;
                    # get rid of the colons
                    $actual =~ s/://g;
                    if ($utilEnd ne $actual)
                    {
                        $returnStatus = ERROR;
                        if ($errors == TRUE)
                        {
                            print "ERROR: Parameter Effective_Date_End "
                                . "does not match filename:\n"
                                .  "       ($actual != $utilEnd)\n";
                        }
                    }
               }
            }    
        }
        if ($pname eq 'UT1_TIME_PARAMETERS/UT1_Modified_Julian')
        {
            $utilBegin = '';
            $utilEnd = '';
            @theArray = &splitArray($actual);
            # Extract start and end dates from the file name
            $fileName =~ /(?:LC08CPF_)(\d{8})_(\d{8})_/ and do {
                $utilBegin = $1;
                $utilEnd = $2;
            };
            $year  = substr($utilBegin, 0, 4);
            $month = substr($utilBegin, 4, 2);
            $day   = substr($utilBegin, 6, 2);
            if (not($day =~ /^[\+\-]?\d+$/))
            {
                print "ERROR: Day from filename is not an "
                    . "integer: $fileName.\n";
                $returnStatus = ERROR;
            }

            if (not($month =~ /^[\+\-]?\d+$/))
            {
                print "ERROR: Month from filename is not an "
                    . "integer: $fileName.\n";
                $returnStatus = ERROR;
            }
            if (not($year =~ /^[\+\-]?\d+$/))
            {
                print "ERROR: Year from filename is not an "
                    . "integer: $fileName.\n";
                $returnStatus = ERROR;
            }

            # convert to julian date
            $startdate =
                ias_modified_julian_date($day, $month, $year, \$error_msg);

            # check startdate for returned error
            if ($startdate == ERROR)
            {
                print "ERROR: JulianDate error: $startdate : $error_msg\n";
                $returnStatus = ERROR;
            }

            # now do the end date
            $year  = substr($utilEnd, 0, 4);
            $month = substr($utilEnd, 4, 2);
            $day   = substr($utilEnd, 6, 2);

            # test the validity of the values
            if (not($day =~ /^[\+\-]?\d+$/))
            {
                print "ERROR: Day from filename is not an "
                    . "integer: $fileName.\n";
                $returnStatus = ERROR;
            }
            if (not($month =~ /^[\+\-]?\d+$/))
            {
                print "ERROR: Month from filename is not an "
                    . "integer: $fileName.\n";
                $returnStatus = ERROR;
            }
            if (not($year =~ /^[\+\-]?\d+$/))
            {
                print "ERROR: Year from filename is not an "
                    . "integer: $fileName.\n";
                $returnStatus = ERROR;
            }

            # convert to julian date
            $enddate =
                ias_modified_julian_date($day, $month, $year, \$error_msg);

            # check enddate for returned error
            if ($enddate == ERROR)
            {
                print "ERROR: JulianDate error: $enddate : $error_msg\n";
                $returnStatus = ERROR;
            }

            #Use the final second of the last day in the calculation
            $enddate += (86399/86400);

            #calculate the midpoint between the dates
            $delta = int(($enddate - $startdate) / 2);
            $middate = $startdate + $delta;
            $calcDateBegin = $middate - 89;
            $calcDateEnd = $middate + 90;

            # Make sure it's 180 incrementally increasing integer values.
            if ($#theArray+1 != 180)
            {
                # Not 180 entries, so it is an error
                $returnStatus = ERROR;
                if ($errors == TRUE)
                {
                    print "ERROR: Parameter array $pname does not have "
                        . "correct number of values! ("
                        . "($#theArray+1) != 180)\n";
                }
            }
            else
            {
                # There are 180 entries, so check the modified julian dates
                # against the file name dates, allowing the value to be up
                # to one day off.
                if (($theArray[0] > ($calcDateBegin + 1))
                        or ($theArray[0] < ($calcDateBegin - 1)))
                {
                    if ($warnings == TRUE)
                    {
                        $returnStatus = ERROR;
                        print "WARN: UT1_Modified_Julian dates do not begin "
                            . "at expected day: ($theArray[0] != "
                            . "$calcDateBegin)\n";
                    }

                }

                if (($theArray[179] > ($calcDateEnd + 1))
                        or ($theArray[179] < ($calcDateEnd - 1)))
                {
                    if ($warnings == TRUE)
                    {
                        $returnStatus = ERROR;
                        print "WARN: UT1_Modified_Julian dates do "
                            . "not end at expected day: "
                            . "($theArray[179] != $calcDateEnd)\n";
                    }
                }
            }

            # Make sure the entries are incrementally increasing integer values.
            for ($i = 0; $i <= $#theArray; $i++)
            {
                if ($theArray[$i] != ($theArray[0]+$i))
                {
                    $returnStatus = ERROR;
                    if ($errors == TRUE)
                    {
                        print "ERROR: Parameter array $pname does not have "
                            . "incremental values! "
                            . "($theArray[$i-1], $theArray[$i], "
                            . "$theArray[$i+1])\n";
                    }

                    last;
                }
            }

        }
        elsif($pname eq 'FILE_ATTRIBUTES/Baseline_Date' || 
            $pname eq 'ORBIT_PARAMETERS/Launch_Date')
        {
            if($actual =~ /"(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2})"/)
            {
                my $converteddate;
                my $hour;
                my $minute;
                my $second;

                $day = $3;
                $month = $2;
                $year = $1;

                # convert to julian date (to check the values are reasonable)
                $converteddate =
                    ias_modified_julian_date($day, $month, $year, \$error_msg);

                # check converteddate for returned error
                if ($converteddate == ERROR)
                {
                    print "ERROR: Unable to make $actual a Julian date :"
                        . $error_msg . "\n";
                    $returnStatus = ERROR;
                }

                $hour = $4;
                $minute = $5;
                $second = $6;

                # Verify the time is valid.  Note that 60 seconds is allowed
                # to account for potential leap seconds.
                if (($hour > 23) or ($minute > 59) or ($second > 60))
                {
                    print "ERROR: The time in of $hour:$minute:$second in "
                        . "$actual is not valid \n";
                    $returnStatus = ERROR;

                }
            }
            else
            {
                $returnStatus = ERROR;
                if ($errors == TRUE)
                {
                    print "ERROR: Value of $pname ($actual) is not in the "
                        . "expected format of YYYY-MM-DDThh:mm:ss\n";
                }
            }
        }
        elsif ($pname eq 'EARTH_CONSTANTS/Number_of_Leap_Seconds')
        {
            $ecNumLeapSeconds = $actual;
            $paramAttribs = "int";
        }
        elsif ($pname eq 'EARTH_CONSTANTS/Leap_Years')
        {
            if (not defined $ecNumLeapSeconds)
            {
                print "ERROR: In EARTH_CONSTANTS, Number_of_Leap_Seconds "
                    . "not found before Leap_Years.\n";
                return ERROR;
            }
            $paramAttribs = "int array $ecNumLeapSeconds range 1972-2100";
        }
        elsif ($pname eq 'EARTH_CONSTANTS/Leap_Months')
        {
            if (not defined $ecNumLeapSeconds)
            {
                print "ERROR: In EARTH_CONSTANTS, Number_of_Leap_Seconds "
                    . "not found before Leap_Months.\n";
                return ERROR;
            }


            $paramAttribs = "string array $ecNumLeapSeconds list MONTHS";
        }
        elsif ($pname eq 'EARTH_CONSTANTS/Leap_Days')
        {
            if (not defined $ecNumLeapSeconds)
            {
                print "ERROR: In EARTH_CONSTANTS, Number_of_Leap_Seconds "
                    . "not found before Leap_Days.\n";
                return ERROR;
            }
            $paramAttribs = "int array $ecNumLeapSeconds range 1-31";
        }
        elsif ($pname eq 'EARTH_CONSTANTS/Leap_Seconds')
        {
            if (not defined $ecNumLeapSeconds)
            {
                print "ERROR: In EARTH_CONSTANTS, Number_of_Leap_Seconds "
                    . "not found before Leap_Seconds.\n";
                return ERROR;
            }
            # To date, the IERS has added 1 leap second at a time.  However,
            # it is possible to remove one, as well.  That's why -1 is in here.
            $paramAttribs = "int array $ecNumLeapSeconds list (-1,1)";
        }

        # For Earth constants parameters, we need to modify the parameter type
        # to force further validation
        if ($pname =~ /^EARTH_CONSTANTS\//)
        {
            @ideal = split(/\s+/, $paramAttribs);
            $type = shift @ideal;
        }
        else
        {
            # Return for all other parameters (non-cloud and non-earth)
            return $returnStatus;
        }

    }  # end of if special

    # Get array length (if any) and format.
    my $is_array = 0;
    if ($paramAttribs =~ /array/)
    {
        $is_array = 1;
        # remove word array from attributes list
        shift @ideal;

        # Split the CPF parameter values into an array.
        @theArray = &splitArray($actual);

        # get array length from attributes list, if any, and check that it is
        # a valid number
        $arrayLength = shift @ideal;
        if (defined ($arrayLength) and ($arrayLength =~ /[0-9]+/))
        {
            if ($#theArray+1 != $arrayLength)
            {
                $returnStatus = ERROR;
                if ($errors == TRUE)
                {
                    print "ERROR: Parameter array $pname does not have "
                        . "correct number of values!  ("
                        .  ($#theArray+1) . " != $arrayLength)\n";
                    # skip this group already have error
                    return $returnStatus;
                }
            }
        }
        elsif (!defined ($arrayLength))
        {
            # If the size of the array is not specified, just set the length to
            # the actual size of the array to make sure all the values in the
            # array are checked for the right type.
            $arrayLength = @theArray;
        }
        else
        {
            # This shouldn't be possible, but warn if it happens
            print "WARN: Array $pname has no length\n";
            $arrayLength = 0;
        }
    }
    else
    {
        $arrayLength = 1;
    }

    # get format from template
    $format = shift @ideal;
    if (not defined $format)
    {
        $format = '';
    }

    # get value from template
    $fval = shift @ideal;

    if ($arrayLength > 1 or $is_array == 1)
    {
        # If there is a nominal value, we may need to split them
        # into arrays also.
        if ($format eq 'nominal')
        {
            if ($fval =~ /^\(.*\)$/)
            {
                # It's an array of nominal values from template.
                @nominalArray = &splitArray($fval);
            }
            else
            {
                # It's a single nominal value that applies to the whole array.
                # Create an array of nominal values for it.
                @nominalArray = ('');
                for ($i = 0; $i < $arrayLength; $i++)
                {
                    $nominalArray[$i] = $fval;
                }
            }
        }
    } # array Length > 1
    else
    {
        # It's a scalar.
        @theArray = ($actual);
    }

    # For template values, create a match string beforehand.
    # S  --  Sign (+ or -)
    # E  --  'E', the exponent character.
    # .  --  Decimal point.
    # N  --  Digit or Digits.
    #          A sign is permitted in the first digit if 'S' is not specified.
    if ($format eq 'template')
    {
        $template = $fval;
        $template =~ s/S/[+-]/g;
        $template =~ s/^N/[+-]?N/;
        $template =~ s/N+\./N\./g;
        $template =~ s/N$/\\d+/;
        $template =~ s/N[Ee]/\\d+[Ee]/g;
        $template =~ s/\./\\\./g;
        $template =~ s/N/\\d/g;
    }

    # Check each value in the array.
    # If it's a scalar, we'll only be checking once.
    for ($i = 0; $i < $arrayLength; $i++)
    {
        $arrValue = $theArray[$i];
        if ($format eq 'nominal')
        {
            # If this really is an array, we need to use
            # the nominal value array.
            if ($arrayLength == 1)
            {
                $nominalVal = $fval;
            }
            else
            {
                # get the CPF nominal value
                $nominalVal = $nominalArray[$i];
            }

            # Check to see if we deviate from the nominal value.
            if ($type eq 'string')
            {
                # This test will work for strings.
                if ($arrValue ne $nominalVal)
                {
                    if ($warnings == TRUE)
                    {
                        $returnStatus = ERROR;
                        print "WARN: Parameter $pname differs from "
                            . "nominal value: ($arrValue != $nominalVal)\n";
                    }
                }
            }
            else
            {
                # This test will work for floats or ints.
                if ($arrValue =~ /^[+-]?[0-9]*\.?[0-9]+([Ee][-+]?[0-9]+)?$/)
                {
                    if ($arrValue != $nominalVal)
                    {
                        if ($warnings == TRUE)
                        {
                            $returnStatus = ERROR;
                            print "WARN: Parameter $pname differs from "
                                . "nominal value: ($arrValue != $nominalVal)\n";
                        }
                    }
                }
                else
                {
                    $returnStatus = ERROR;
                    if ($errors == TRUE)
                    {
                        print "ERROR: Parameter $pname is not a "
                            . "number: $arrValue.\n";
                    }
                }
            }
        }  # end of format eq nominal
        elsif ($format eq 'template')
        {
            # Template check for floats.
            # Strings and Ints should never have a template.
            if (not($arrValue =~ /$template/))
            {
                if ($formatMismatch == TRUE)
                {
                    $returnStatus = ERROR;
                    print "FORMAT:  $pname ($arrValue not matching $fval)\n";
                }
            }
        }
        elsif ($format eq 'range')
        {
            # Range check for ints and floats.
            # Strings should never have a range.
            ($min, $max) = split(/-/, $fval);
            if (($arrValue > $max) or ($arrValue < $min))
            {
                $returnStatus = ERROR;
                if ($errors == TRUE)
                {
                    print "ERROR: Parameter $pname is outside allowed "
                        . "range!  ($arrValue not in $fval)\n";
                }
            }
        }
        elsif ($format eq 'list')
        {
            # Sanity check that ensures a list of values (or NAME) was supplied
            if (!defined $fval)
            {
                $returnStatus = ERROR;
                if ($errors)
                {
                    print "ERROR: Missing list values for $pname in ".
                        "template.\n";
                }
                return $returnStatus;
            }
            # See if it's a predefined list (e.g. MONTHS)
            if ((substr($fval, 0, 1) ne '(')
                && (exists $predefinedList{$fval}))
            {
                @listArray = @{$predefinedList{$fval}}
            }
            else
            {
                # Get the array of valid domain values from template.
                @listArray = &splitArray($fval);
            }
            # Mark as invalid, until proven otherwise
            my $validValue = FALSE;
            if ($type eq 'string')
            {
                foreach (@listArray)
                {
                    if ($arrValue eq $_)
                    {
                        $validValue = TRUE;
                        last;
                    }
                }
            }
            else
            {
                # This test will work for floats or ints.
                if ($arrValue =~ /^[+-]?[0-9]*\.?[0-9]+([Ee][-+]?[0-9]+)?$/)
                {
                    foreach (@listArray)
                    {
                        if ($arrValue == $_)
                        {
                            $validValue = TRUE;
                            last;
                        }
                    }
                }
                else
                {
                    $returnStatus = ERROR;
                    if ($errors == TRUE)
                    {
                        print "ERROR: Parameter $pname is not a "
                            . "number: $arrValue.\n";
                    }
                }
            }
            if (!$validValue)
            {
                $returnStatus = ERROR;
                if ($errors == TRUE)
                {
                    print "ERROR: Value ($arrValue) is not in the list of "
                        . "valid domain values for Parameter $pname. "
                        . "($arrValue not in $fval)\n";
                }
            }
        }
        else
        {
            # Free-form value.
        }

        # A couple more checks:
        # Are the ints truly integer and doubles really doubles?
        # BUG: arrValue needs spaces trimmed on both sides.
        if (($type eq 'int') and not($arrValue =~ /^[\+\-]?\d+$/))
        {
            $returnStatus = ERROR;
            if ($errors == TRUE)
            {
                print "ERROR: Parameter $pname is not an integer. "
                    . "Should be integer.  ($arrValue)\n";
            }
        }
        if (($type eq 'double')
                and not
                ($arrValue =~ /^\s*[+-]?[0-9]*\.?[0-9]+([Ee][-+]?[0-9]+)?\s*$/))
        {
            $returnStatus = ERROR;
            if ($errors == TRUE)
            {
                print "ERROR: Parameter $pname is not a double. "
                    . "Should be double.  ($arrValue)\n";
            }
        }
    }  # end of for i to arrayLength
    return $returnStatus;
}

###############################################################################
#
# Name: splitArray
#
# Purpose: To split a comma separated list into an array, white spaces are
#           eliminated, quotes (if any) surrounding each value are preserved.
#
# Returns: Array populated with list
#
# Parameters:
#
###############################################################################
sub splitArray
{
    my ($list) = @_;
    my @theArray;

    # get rid of any spaces between array values
    $list =~ s/\s*//g;

    if ((substr($list,0,1) eq '(') and (substr($list,-1) eq ')'))
    {
        # remove opening parin
        $list =~ s/^.//;
        # remove trailing parin
        $list =~ s/.$//;

        @theArray = split(/,/, $list);
    }
    else
    {
        # Allow arrays without parentheses, but arrays without parens should
        # not contain more than one value
        @theArray = split(/,/, $list);
        if ($#theArray > 0)
        {
            $errorFlag = TRUE;
            if ($errors)
            {
                print "ERROR: $pname is an array without parentheses that "
                    . "contains multiple values\n";
            }
        }
    }

    return @theArray;
}


###############################################################################
#
# Name: cpsHelp
#
# Purpose: To display the help menu
#
# Returns: None
#
# Parameters: None
#
###############################################################################
sub cpsHelp
{
    print <<EOF;

FILE VALIDATION HELP MENU

Syntax:  cps_validate.pl --[option] [value] <OLI BPF filename>|
         <TIRS BPF filename>|<CPF filename>|<NA>

--h[elp]          Help (this message)
--e[rrors]        Enable error messages
--w[arnings]      Enable warnings
--f[ormat]        Report format mismatches
--t[ype] [value]  Specify file type -
                    A value is required with this option.
                    Valid values are:
                      OLIBPF, obpf, o, ob, olib, olibpf - OLI BPF template 
                          will be selected.
                      TIRSBPF, tbpf, t, tb, tirsb, tirsbpf - TIRS BPF 
                          template will be selected.
                      CPF, cpf, c - CPF template will be selected.
                      NA, na  - Template will be determined by input file name.
                    Type values are case insensitive.
                    Example: --type bpf
--s[tandardIn     Read the file from STDIN
                    Example cps_validation.pl --s <filename> <filename>
                       or cat <filename> | cps_validation.pl --s <filename>

Default options are: --t na --errors --warnings --format

Turn off anomaly reporting: --noerrors --nowarnings --noformat
                    or       -noerrors  -nowarnings  -noformat
                    or       -noe       -now         -nof

EOF

}
