#!/usr/bin/perl -w
############################################################################
#   NAME:   makeTraceMatrix 
#
#   PURPOSE: Create traceability matrix mapping requirements to design
#            document sections. 
#
#   PARAMETERS: 1) csv_filename
#               2) reqlist_filename 
#
#               It is assumed that the input files are in the current 
#               working directory.
#
#   INPUT:  Comma-separated file of requirements exported from DOORS
#           application.  The first and second columns are the ones of
#           interest: Requirement ID and Requirement Text.  Requirement
#           Header, which also can appear in the second column, is also
#           of interest for readability reasons.
#
#           .reqlist file in LaTeX table of contents format.  The last
#           section is the one of interest (section number).
#
#   OUTPUT: LaTeX format table of requirement IDs, corresponding section
#           numbers, and requirement text and headers.  This should be 
#           insertable into the Traceability Matrix section of an SDD 
#           document that follows the LDCC SDD template format.
#
############################################################################

use strict;
use Getopt::Long;

sub print_unmatched($$$$$);
sub read_reqlist($$$$);

#############################################################################
# Variable declarations
#############################################################################

my $csv_filename = undef;  # I: File with comma-separated list of requirements.
my $reqlist_filename = undef; # I: LaTeX file in reqlist toc format with 
my $ddreqlist_filename = undef; # sections and corresponding requirements. 
                                # dd is for the Detailed Design version.
my $matrix_filename = "requirements_traceability.tex"; # O: name of LaTeX file
                           # in mainmatter document format that holds the
                           # traceability matrix.
my $ssns = "IAS|PGS|IS|OHS|IP"; # These are the subsystem strings, which
                           # must match those used in the requirement IDs.
my $usage = "usage < makeTraceMatrix.pl -csv_name <string> -reqlist_name <string> [-ddreqlist_name <string>] >\n"; # usage message string

my $csv_row;               # loop counter
my $csv_req_id;            # single requirement ID
my @csv_req_ids = ();      # array of requirement IDs 
my $csv_req_headings;      # single requirement heading 
my @csv_req_headings = (); # array of requirement headings 
my $csv_req_text;          # single requirement text 
my @csv_req_texts = ();    # array of requirement texts 

my $tex_row;               # loop counter
my $num_tex_rows;          # number of valid rows in the tex file 
my $tex_req_id;            # single requirement ID
my @tex_req_ids = ();      # array of requirement IDs 
my $tex_section_num;       # single section number 
my @tex_section_nums = (); # array of section numbers 
my @tex_req_match = ();    # array of flags set if match is found for section
my $unmatched_section;     # flag set if any section is unmatched
my $found;                 # flags set if match is found 

my $ddtex_row;             # loop counter
my $ddnum_tex_rows;        # number of valid rows in the tex file 
my $ddtex_req_id;          # single requirement ID
my @ddtex_req_ids = ();    # array of requirement IDs 
my $ddtex_section_num;     # single section number 
my @ddtex_section_nums = (); # array of section numbers 
my @ddtex_req_match = ();  # array of flags set if match is found for section
my $ddunmatched_section;   # flag set if any section is unmatched
my $ddfound;               # flags set if match is found 

my $dummy1;                # dummy variables as placeholders
my $dummy2;                # dummy variables as placeholders
my $dummy3;                # dummy variables as placeholders

# Get input parameters.
GetOptions("csv_name:s"=>\$csv_filename,
           "reqlist_name:s"=>\$reqlist_filename,
           "ddreqlist_name:s"=>\$ddreqlist_filename);

# If input parameters aren't defined, exit.
(defined $csv_filename) or die "$usage";
(defined $reqlist_filename) or die "$usage";
# The Detailed Design is not currently required, but you may want to 
# enforce this at later software lifecycle phases.
#(defined $ddreqlist_filename) or die "$usage";

#
# Read the .csv file into the requirements list.
#
#
# Here are some assumptions on the acceptable format of the .csv file:
#
# 1. The input CSV file only includes the 2 columns for the Requirement
#    ID and the Requirement Text, in that order.  The general format of 
#    each row is:
#
#    "<SSN><Req#>",<Requirement Text>
#   
#    <SSN> is a string.  Possible values are defined in variable ssns. 
#    <Req#> is a positive integer.
#    Requirement Text can include quotes, commas, and newlines. 
#

# Open the input file.
open (CSV_FPIN, "< $csv_filename") or die "Open csv file failed....";

$csv_row = 0;
# Loop through the data file until eof.
while (<CSV_FPIN>)
{
    my $text;

    chomp;

    # Read the requirement ID and text columns into data structures.
    ($csv_req_ids[$csv_row], $csv_req_headings[$csv_row], 
         $text) = split (',', $_, 3);

    push @{$csv_req_texts[$csv_row]}, $text;

    # Take off any DOS Ctrl-M's.  If they appear they may interfere with
    # later parsing.
    $csv_req_ids[$csv_row] =~ tr/\cM//d;
    $csv_req_headings[$csv_row] =~ tr/\cM//d;
    $csv_req_texts[$csv_row][0] =~ tr/\cM//d;

    # The requirements text and headings could have some special characters 
    # that LaTeX would interpret as a LaTeX instruction.  Prefix them with 
    # a backslash so they are treated as regular text.
    $csv_req_headings[$csv_row] =~ s/([\#\$\%\^\&\_\{\}\~])/\\$&/g;
    $csv_req_texts[$csv_row][0] =~ s/([\#\$\%\^\&\_\{\}\~])/\\$&/g;

    # Take off the beginning and ending quotation marks.
    $_ = $csv_req_ids[$csv_row];
    if (/^"(.*)"$/)
    {
        $csv_req_ids[$csv_row] = $1;
    }

    $_ = $csv_req_headings[$csv_row];
    if (/^"(.*)"$/)
    {
        $csv_req_headings[$csv_row] = $1;
    }

    # The text for the requirement may be single-line or multi-line.  For
    # multi-line comments, the comment will begin with a double-quotes, and
    # will end with a double-quotes at the end of a line.  For single-line
    # comments, all we need to do is remove the beginning and ending quotes,
    # if any.  For multi-line comments, we want to remove the comments and
    # keep track of each line individually because they'll have to be output
    # to tex as separate lines in the table.
    #
    # Drop the leading quotes, if any.  Also, drop the trailing quotes, if
    # any.
    if (($csv_req_texts[$csv_row][0] =~ s/^"//) and
        not ($csv_req_texts[$csv_row][0] =~ s/"$//))
    {
        # Okay, the text began with a quote, but there was no closing quote.
        # We procede by searching for the line with the closing quote, keeping
        # track of the lines as we go.
        while (my $next_line = <CSV_FPIN>)
        {
            chomp $next_line;

            # Take off any DOS Ctrl-M's.  If they appear they may interfere with
            # later parsing.
            $next_line =~ tr/\cM//d;
            $next_line =~ s/([\#\$\%\^\&\_\{\}\~])/\\$&/g;

            push @{$csv_req_texts[$csv_row]}, $next_line;

            # drop the closing quotes, quitting if any found
            last if ($csv_req_texts[$csv_row][-1] =~ s/"$//);
        }
    }

    # Convert any double quotation marks to tex-like
    map { s/""(.+?)""/``$1''/g } @{$csv_req_texts[$csv_row]};

    # If the requirement text is empty, but the requirement heading is
    # not empty, replace the requirement text with the requirement heading.
    # This allows us to show the requirement headings in the output table,
    # similarly to how they appear in DOORS (but not the DOORS exported
    # .CSV file).  While doing this, put in some LaTeX to make the headings
    # show up bold, which they are in DOORS, too.

    if (($csv_req_texts[$csv_row][0] eq '') &&
        ($csv_req_headings[$csv_row] ne ''))
    {
        $csv_req_texts[$csv_row][0] = 
            "\\textbf{" . $csv_req_headings[$csv_row] . "}";
    }

    # Verify that the requirement ID follows the ID format.  Only if it
    # does do we increment the index.
    if ($csv_req_ids[$csv_row] =~ /^($ssns)\d*$/)
    {
        $csv_row++;
    }
    else
    {
        # clear out the existing array, so we don't accidentally use it in the
        # next row
        @{$csv_req_texts[$csv_row]} = ();
    }
} # end of while not eof structure populate statement

close(CSV_FPIN);

if (!read_reqlist ($reqlist_filename, \@tex_req_ids, \@tex_section_nums,
                   \@tex_req_match))
{
    die "Failed to read reqlist file...";
}

if (defined $ddreqlist_filename)
{
    if (!read_reqlist ($ddreqlist_filename, \@ddtex_req_ids, 
                       \@ddtex_section_nums, \@ddtex_req_match))
    {
        die "Failed to read ddreqlist file...";
    }
}

#
# Find the matches in the 2 files, based on the requirement ID string. 
# Output the matches into a LaTeX-format file.  Do this in requirement 
# ID order.  Go through the whole list - it's a matrix, and there can
# be multiple matches in either direction. 
#

# Open output file.
open (TEX_FPOUT, "> $matrix_filename") 
    or die "Could not create output file. ERROR:$!\n";

# Start tabular output here.
print (TEX_FPOUT "\\chapter{Requirements Traceability}\n"); 
print (TEX_FPOUT "\n");
print (TEX_FPOUT "\\begin{longtable}{R}\n");
print (TEX_FPOUT "\\hline\n");
print  TEX_FPOUT "\\textbf{ID} & ",
                 "\\textbf{Requirement Text} & ",
                 "\\textbf{Section}";
print (TEX_FPOUT "\\endhead\n");
print (TEX_FPOUT "\\endfoot\n");
print (TEX_FPOUT "\\caption{Traceability Matrix} \\label{ta:req_matrix}\n");
print (TEX_FPOUT "\\endlastfoot\n");
print (TEX_FPOUT "\\hline \\hline\n");

# Cross-reference the requirements IDs to find what rows need to be printed.
$csv_row = 0;
foreach $csv_req_id (@csv_req_ids)
{
    $tex_row = 0;
    $ddtex_row = 0;
    $found = 0; # false
    $ddfound = 0; # false

    # Skip the requirement if there is no associated requirement text.
    # In such a case it's just a heading pseudo-requirement.
    if ((scalar(@{$csv_req_texts[$csv_row]}) == 0) ||
        (length($csv_req_texts[$csv_row][0]) == 0))
    {
        $csv_row++;
        next;
    }

    # Print the requirement ID and text.
    print (TEX_FPOUT $csv_req_id, " & ", $csv_req_texts[$csv_row][0], " & ");

    foreach $tex_req_id (@tex_req_ids)
    {
        if ($csv_req_id eq $tex_req_id)
        {
            # Print the LaTeX row output.  There is 1 row per requirement. 
            # There may be multiple sections per requirement.

            if ($found)
            {
                # Print the comma and newline separating this section 
                # from the previous one.
                print (TEX_FPOUT ",\\enskip");
            }
            else
            {
                $found = 1; # true
            }

            # Print the matching section number.
            print (TEX_FPOUT $tex_section_nums[$tex_row]);

            # Flag that this [requirement/section number] has a match.
            $tex_req_match[$tex_row] = 1;
        }
        $tex_row++;
    }

    # Print the remaining lines in the text if any remain.  These will be new
    # rows in the table, but they will appear as being in the same row because
    # we will not draw the hline between the rows.
    for (my $text_line = 1; $text_line < scalar(@{$csv_req_texts[$csv_row]});
            $text_line++)
    {
        print (TEX_FPOUT " \\nl\n & $csv_req_texts[$csv_row][$text_line] & ");
    }

    # Complete the row.
    print (TEX_FPOUT " \\nl \\hline\n");

    $csv_row++;
}

#
# Print entries in table for section/requirements entries in the LaTeX
# document that were not matched in DOORS.
#

# Print these rows for unmatched high-level design entries.
if (!print_unmatched (*TEX_FPOUT, \@tex_req_ids, \@tex_section_nums,
                   \@tex_req_match, "HIGH_LEVEL"))
{
    die "Failed to print unmatched requirement/section number pairs ....";
}

# Print these rows for unmatched detailed design entries.
if (!print_unmatched (*TEX_FPOUT, \@ddtex_req_ids, \@ddtex_section_nums,
                   \@ddtex_req_match, "DETAILED"))
{
    die "Failed to print unmatched requirement/section number pairs ....";
}


# End tabular output here.
print (TEX_FPOUT "\\end{longtable}\n");

# Close output file.
close (TEX_FPOUT);

exit 0;

#############################################################################
# Name: print_unmatched
#
# Description: Print the unmatched [requirement/section] pairs in the 
# reqlist file at the bottom of the table.
#
#############################################################################
sub print_unmatched($$$$$)
{
    my $TEX_FPOUT = shift;
    my ($req_ids, $section_nums, $req_match, $type) = @_;

    my $row = 0; # loop counter
    my $req_id;  # single requirement ID

    foreach $req_id (@$req_ids)
    {
        if (! $$req_match[$row])
        {
            # This [requirement/section] pair didn't match any requirement 
            # in the .csv file.  Flag the requirement ID and section number 
            # in the table. 
            print (TEX_FPOUT "$req_id & NO MATCHING REQUIREMENT IN DOORS & ");

            print (TEX_FPOUT $$section_nums[$row]);

            #if ($type eq "HIGH_LEVEL")
            #{
                # Print an additional column AFTER the section numbers.
                #print (TEX_FPOUT " & ");
            #}

            print (TEX_FPOUT " \\nl \\hline\n");
        }
        $row++;
    }

    return 1;
}

#############################################################################
# Name: read_reqlist
#
# Description: Read the reqlist file into the section list. 
#
# Here are some assumptions on the acceptable format of the reqlist file:
#
# 1. The general format of each row is:
#
#    \contentsline {req}{<Requirement ID>}{#}{<section>.<Section Number>}
#
#    We are interested in <Requirement ID> and <Section Number>.
#
#############################################################################
sub read_reqlist($$$$)
{
    my ($req_filename, $req_ids, $section_nums, $req_match) = @_;

    my $row = 0; # loop counter

    # Open the input file.
    open (TEX_FPIN, "< $req_filename") 
        or die "Open reqlist file failed....";

    # Loop through the data file until eof.
    while (<TEX_FPIN>)
    {
        chomp;

        # Read the requirement ID and Section columns into data structures.
        # Using "{" to divide the file, we want the 3rd and 5th columns.
        ($dummy1, $dummy2, $$req_ids[$row], $dummy3, 
             $$section_nums[$row]) = split ('{');

        # Take off the ending "}" on the strings.
        $_ = $$req_ids[$row];
        if (/(.*)}$/)
        {
            $$req_ids[$row] = $1;
        }

        $_ = $$section_nums[$row];
        if (/(.*)}$/)
        {
            $$section_nums[$row] = $1;
        }

        # Take the section type and subsequent period off the section number.
        $_ = $$section_nums[$row];
        ($dummy1, $$section_nums[$row]) = split ('\.', $_, 2);

        # Verify that the requirement ID follows the ID format.  Also verify
        # that the section number follows the section number format.  Only if 
        # it does both do we increment the index.
        if ($$req_ids[$row] =~ /^($ssns)\d*$/)
        {
            if ($$section_nums[$row] =~ /^\d*[\d*\.]*$/)
            {
                $$req_match[$row] = 0; # Initialize flag to false.
                $row++;
            }
        }
    } # end of while not eof structure populate statement

    close(TEX_FPIN);
}

