/* Standard C Includes */
#include <string.h>
#include <stdlib.h>

/* IAS Includes */
#include "ias_l1g.h"
#include "ias_const.h"
#include "ias_types.h"
#include "ias_logging.h"

/* Local Includes */
#include "generate_chips.h"

/* Local Defines */
#define WATER_MASK_WATER 0x0
#define WATER_MASK_LAND  0x1

/*****************************************************************************
 NAME: get_cloud_bounds

 PURPOSE: Determines the bounds for the cloud to ensure they are inside of the
          image.

 RETURN VALUE: SUCCESS / ERROR
*****************************************************************************/
static void get_cloud_bounds
(
    int cloud_border,       /* I: Cloud border expansion */
    int last_cloud_position,/* I: Last position zeroed due to clouds */
    int current_position,   /* I: Current position */
    int image_end,          /* I: End of image */
    int *start_position,    /* O: Starting position for expansion */
    int *end_position       /* O: Ending position for expansion */
)
{
    /* Check to see if last set cloud pixel is within current vicinity */
    *start_position = current_position - cloud_border;
    if (*start_position <= last_cloud_position)
    {
        *start_position = last_cloud_position + 1; 
    }

    /* Set the ending position */
    *end_position = current_position + cloud_border;

    /* Ensure that start position is in the image bounds */
    if (*start_position < 0)
    {
        *start_position = 0;
    }
    
    /* Ensure that end position is in the image_bounds */
    if (*end_position > image_end - 1)
    {   
        *end_position = image_end - 1;
    }
}

/*****************************************************************************
 NAME: build_cloud_mask

 PURPOSE: Build the cloud mask into the exclusion mask.

 RETURN VALUE: SUCCESS / ERROR
*****************************************************************************/
static int build_cloud_mask
(
    const GENERATE_CHIPS_PARAMETERS *parameters,/* I: Parameters structure */
    const IAS_IMAGE *quality_image, /* I: Quality image */             
    IAS_UI2_TYPE quality_cloud_mask,/* I: Value used for cloud detection */
    IAS_UI2_TYPE min_cloud_value,   /* I: Minimum cloud mask level */ 
    IAS_UI2_TYPE mask_value,        /* I: What to set clouds to in mask */
    IAS_BYTE_TYPE *exclusion_mask   /* O: Pointer to allocated mask */
)
{
    int line;                           /* Current line in loop */
    int sample;                         /* Current sample in loop */
    int last_cloud_sample;              /* Last cloud sample information */
    int *last_cloud_line = NULL;        /* Last line of cloud information */
    IAS_UI2_TYPE quality_value;         /* Current quality value */
    int num_lines = quality_image->nl;  /* Total lines */
    int num_samples = quality_image->ns;/* Total samples */
    int expand_cloud_border = 0;        /* Boolean for expanding clouds */
    short *quality_band = ((short *)quality_image->data);/* Data placeholder */
    
    /* Cloud border expansion values */
    int cloud_border_lines = parameters->cloud_border_lines; 
    int cloud_border_samples = parameters->cloud_border_samples;
    
    /* Determine if cloud expansion is needed */
    if (cloud_border_lines != 0 && cloud_border_samples != 0)
    {
        expand_cloud_border = 1;
        
        /* Allocated memory for last cloud line */
        last_cloud_line = (int *) malloc(sizeof(int) * num_samples);
        if (!last_cloud_line)
        {
            IAS_LOG_ERROR("Could not allocated last_cloud_line");    
            return ERROR;
        }

        /* Initialize end line associated with last cloud to be outside
           of area to zero per cloud */
        for (sample = 0; sample < num_samples; sample++) 
        {
            last_cloud_line[sample] = -1;
        }
    }

    /* Determine if cloud bit is set. */
    for (line = 0; line < num_lines; line++)
    {
        last_cloud_sample = -1;

        for (sample = 0; sample < num_samples; sample++)
        {
            int is_cloud_pixel; /* Boolean if current pixel is a cloud pixel */
            quality_value = quality_band[line * num_samples + sample];
            
            /* Determine if pixel is a valid cloud pixel */
            is_cloud_pixel = ((quality_value & quality_cloud_mask) 
                >= min_cloud_value);

            /* If a cloud pixel set appropriate pixels in mask */
            if (is_cloud_pixel)
            {
                if (expand_cloud_border)
                {
                    int start_line;     /* Line start of expansion */         
                    int end_line;       /* Line end of expansion */
                    int mask_line;      /* Loop variable for mask line */
                    int start_sample;   /* Sample start of expansion */
                    int end_sample;     /* Sample end of expansion */
                    int mask_sample;    /* Loop variable for mask sample */

                    /* Find bounds of cloud with respect to previous line
                       and the previous sample */
                    get_cloud_bounds(cloud_border_lines, 
                        last_cloud_line[sample], line, num_lines, &start_line, 
                        &end_line);

                    get_cloud_bounds(cloud_border_samples, last_cloud_sample,
                        sample, num_samples, &start_sample, &end_sample);

                    /* Ensure that start line and sample come before the end */
                    if (start_sample > end_sample || start_line > end_line)
                    {
                        IAS_LOG_ERROR("Starting bounds outside of the "
                            "ending bounds");
                        free(last_cloud_line);
                        return ERROR;
                    }

                    /* Bit OR exclusion mask with the mask value passed in as 
                       parameter in cloud expansion area*/
                    for (mask_line = start_line; mask_line <= end_line; 
                         mask_line++)
                    {
                        for (mask_sample = start_sample; 
                             mask_sample <= end_sample; mask_sample++)
                        {
                            int index = mask_line * num_samples + mask_sample;
                            if (!(quality_band[index] & IAS_QUALITY_BIT_FILL))
                            {
                                exclusion_mask[index] |= mask_value;
                            }
                        }
                    }
                    last_cloud_sample = end_sample;
                    last_cloud_line[sample] = end_line;
                }
                else
                {
                    /* No expansion just single pixel */ 
                    exclusion_mask[line * num_samples + sample] |= mask_value;
                }
            }
        }
    }
   
    /* Free allocated memory */ 
    if (expand_cloud_border)
    {
        free(last_cloud_line);
    }

    return SUCCESS;
}

/*****************************************************************************
 NAME: build_exclusion_mask

 PURPOSE: Creates an exclusion mask from the provided quality band to be used
          by the moravec code and extract_gcps code.

 NOTES:    For the following bit tests:
           If exclude value is low (1) it will exclude if any bit is
           set, for medium (2) it will exclude if 10 confidence bit is
           set or 11 confidence bits are set, for high it will only exclude
           if 11 confidence bits are set.

 RETURN VALUE: SUCCESS / ERROR

*****************************************************************************/
int build_exclusion_mask
(
    const GENERATE_CHIPS_PARAMETERS *parameters, /* I: The input parameters */
    const IAS_IMAGE *quality_image, /* I: The quality image data */
    IAS_BYTE_TYPE *water_mask,      /* I: The images water mask */
    IAS_BYTE_TYPE *exclusion_mask   /* O: The resulting exclusion mask */       
)
{
    int index;                          /* Loop control */
    int data_count;                     /* The total number of data items */
    IAS_UI2_TYPE min_cloud_value;       /* Minimum mask confidence value */
    IAS_UI2_TYPE min_snow_value;        /* Minimum snow confidence level */
    unsigned short *quality_band = (unsigned short *)quality_image->data;

    /* Set the data size and count */
    data_count = quality_image->nl * quality_image->ns;

    /* Initialize the mask to valid */
    for (index = 0; index < data_count; index++)
    { 
        exclusion_mask[index] = EXCLUSION_MASK_VALID;
    }

    /* Use the water mask if it has been built */
    if (water_mask)
    {
        for (index = 0; index < data_count; index++)
        { 
            /* Copy the water from the water mask to the exclusion mask */
            if (water_mask[index] != WATER_MASK_LAND)
            {
                exclusion_mask[index] |= EXCLUSION_MASK_WATER;
            }
        }
    }

    /* Set the minimum cirrus bits and build the cirrus mask */
    min_cloud_value = IAS_QUALITY_BIT_CIRRUS_01 * parameters->exclude_cirrus;
    if (parameters->exclude_cirrus)
    {
        if (build_cloud_mask(parameters, quality_image, 
            IAS_QUALITY_BIT_CIRRUS_11, min_cloud_value, EXCLUSION_MASK_CIRRUS, 
            exclusion_mask) != SUCCESS)
        {
            IAS_LOG_ERROR("Building cirrus cloud mask");
            return ERROR;
        }
    }

    /* Set the minimum cloud bits and build cloud mask */
    min_cloud_value = IAS_QUALITY_BIT_CLOUD_01 * parameters->exclude_cloud;
    if (parameters->exclude_cloud)
    {
        if (build_cloud_mask(parameters, quality_image,
            IAS_QUALITY_BIT_CLOUD_11, min_cloud_value, EXCLUSION_MASK_CLOUD, 
            exclusion_mask) != SUCCESS)
        {
            IAS_LOG_ERROR("Building cloud mask");
            return ERROR;
        }
    }

    /* Determine at what level to exclude snow and ice */
    min_snow_value = IAS_QUALITY_BIT_SNOW_ICE_01 * 
        parameters->exclude_snow_and_ice;
    if (min_snow_value == 0)
    {
        min_snow_value = 0xFFFF;
    }

    /* Process through the data */
    for (index = 0; index < data_count; index++)
    {
        IAS_UI2_TYPE quality_value;         /* Current quality band sample */

        quality_value = quality_band[index];

        /* Always exclude fill data */
        if (quality_value & IAS_QUALITY_BIT_FILL)
        {
            exclusion_mask[index] |= EXCLUSION_MASK_FILL;
            continue;
        }
    
        /* Check snow exclusion */
        if ((quality_value & IAS_QUALITY_BIT_SNOW_ICE_11) >= min_snow_value)
        {
            exclusion_mask[index] |= EXCLUSION_MASK_SNOW_ICE;
        }
    }
    
    return SUCCESS;
}

