
/* Standard C Includes */
#include <stdlib.h>

/* IAS Includes */
#include "ias_const.h"
#include "ias_types.h"
#include "ias_logging.h"

/* Local Includes */
#include "resample_data.h"

/* Because CC weight calculation is from -2 to 2, the range is 4 */
#define CC_RANGE 4.0

/*****************************************************************************
 NAME: calculate_resampling_weights

 PURPOSE: Generates a CC weight table based on the input number of weights and
          the CC step.

 RETURN VALUE: Type = int
    Value      Description
    ---------  --------------------------
    SUCCESS    Successful completion
    ERROR      Operation failed

 NOTE: Returns allocated memory to the calling routine.

*****************************************************************************/
int calculate_resampling_weights
(
    int number_of_weights,  /* I: Number of resampling weights */
    int *kernel_size,       /* O: Calculated kernel size */
    int *number_of_bins,    /* O: Calculated number of bins */
    double **weights        /* O: Resampling weights */
)
{
    int subpixel_line_index; /* Looping control */
    int subpixel_sample_index; /* Looping control */
    int kernel_line_index;   /* Looping control */
    int kernel_sample_index; /* Looping control */
    int weight_index;        /* Looping control */

    int weights_count;       /* Number of weights for all bins */
    double cc_line;          /* "Positions" for scaling weights */
    double cc_sample;        /* "Positions" for scaling weights */
    double cc_line_weight;   /* Resampling weights for line directions */
    double cc_sample_weight; /* Resampling weights for sample directions */
    double bin_weight_sum;   /* Sum for applying weights to input pixels */
    double sub_line_ratio;   /* Fractional line for calculating resampling bin
                                weights */
    double sub_sample_ratio; /* Fractional sample for calculating resampling
                                bin weights */

    int local_number_of_bins; /* The number of samples in a bin line */
    int local_kernel_size;    /* Resampling kernel size */
    double *local_weights;    /* The weights */

    /* Used to initialize the weight values */
    int data_size;

    double cc_kernel_step;    /* Cubic Convolution kernel step for scaling
                                 weights "Positions" */

    /* Determine resampling size and allocate resampling table */
    local_kernel_size = number_of_weights * number_of_weights;

    data_size = local_kernel_size * (SUBPIXEL_STEPS + 1) * (SUBPIXEL_STEPS + 1);

    /* Allocate resampling table */
    local_weights = calloc(data_size, sizeof(double));
    if (local_weights == NULL)
    {
        IAS_LOG_ERROR("Error allocating weights table");
        return ERROR;
    }

    /* Determine the kernel step
       To get the kernel step -> step = (range / (number_of_weights - 1) */
    cc_kernel_step = CC_RANGE / (number_of_weights - 1);

    local_number_of_bins = (SUBPIXEL_STEPS + 1) * local_kernel_size;

    /* Calculate resampling weights for all bins */
    weights_count = 0;
    for (subpixel_line_index = 0;
         subpixel_line_index <= SUBPIXEL_STEPS;
         subpixel_line_index++)
    {
        sub_line_ratio = (double)subpixel_line_index / (double)SUBPIXEL_STEPS;

        for (subpixel_sample_index = 0;
             subpixel_sample_index <= SUBPIXEL_STEPS;
             subpixel_sample_index++)
        {
            sub_sample_ratio =
                (double)subpixel_sample_index / (double)SUBPIXEL_STEPS;
            cc_line = -2.0 - sub_line_ratio;

            /* For each bin calculate the weights */
            bin_weight_sum = 0.0;
            for (kernel_line_index = 0;
                 kernel_line_index < number_of_weights;
                 kernel_line_index++)
            {
                /* Get the line weight */
                cc_line_weight = ias_math_cubic_convolution(-0.5, cc_line);
                cc_sample = -2.0 - sub_sample_ratio;

                for (kernel_sample_index = 0;
                     kernel_sample_index < number_of_weights;
                     kernel_sample_index++)
                {
                    /* Get the sample weight */
                    cc_sample_weight =
                        ias_math_cubic_convolution(-0.5, cc_sample);

                    /* Combine the line and sample weights into one weight */
                    local_weights[weights_count] =
                        cc_line_weight * cc_sample_weight;
                    bin_weight_sum += local_weights[weights_count];
                    weights_count++;
                    cc_sample += cc_kernel_step;
                }
                cc_line += cc_kernel_step;
            }

            /* Normalize the weights within the bin */
            for (weight_index = (weights_count - local_kernel_size);
                 weight_index < weights_count;
                 weight_index++)
            {
                local_weights[weight_index] /= bin_weight_sum;
            }
        }
    }

    /* Assign to the return pointers */ 
    *kernel_size = local_kernel_size;
    *number_of_bins = local_number_of_bins;
    *weights = local_weights;

    return SUCCESS;
}

