
/* Standard C Includes */
#include <stdlib.h>
#include <string.h>

/* IAS Includes */
#include "ias_const.h"
#include "ias_types.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"

/*****************************************************************************
  NAME: convert_to_data_type

  PURPOSE: This code is intended to only be utilized by generate_chips.  It
           converts a source data type to an output data type.

 RETURN VALUE: Type = int
    Value      Description
    ---------  --------------------------
    SUCCESS    Successful completion
    ERROR      Operation failed

  NOTE: No regard to rounding has been implemented, since nothing in
        generate_chips needs to worry about that.

*****************************************************************************/
int convert_to_data_type
(
    int data_count,                      /* I: The number of data elements */
    IAS_DATA_TYPE source_data_type,      /* I: The source data type */
    void *source_data,                   /* I: The source data */
    IAS_DATA_TYPE destination_data_type, /* I: The destination data type */
    void *destination_data               /* O: The data converted */
)
{
    int status; /* Capture function/routine statuses */
    int index;  /* Looping variable */
    int data_type_size; /* Size of the destination data type */

    /* Setup the source pointers */
    IAS_BYTE_TYPE *source_buffer_BYTE = source_data;
    IAS_I2_TYPE   *source_buffer_I2   = source_data;
    IAS_UI2_TYPE  *source_buffer_UI2  = source_data;
    IAS_R4_TYPE   *source_buffer_R4   = source_data;

    /* Setup the destination pointers */
    IAS_BYTE_TYPE *destination_buffer_BYTE = destination_data;
    IAS_I2_TYPE   *destination_buffer_I2   = destination_data;
    IAS_UI2_TYPE  *destination_buffer_UI2  = destination_data;
    IAS_R4_TYPE   *destination_buffer_R4   = destination_data;

    /* If they are the same data type, then all that needs to be performed is
       a memcpy */
    if (source_data_type == destination_data_type)
    {
        status = ias_misc_get_sizeof_data_type(destination_data_type,
            &data_type_size);

        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to convert data type");
            return ERROR;
        }

        memcpy(destination_data, source_data, data_type_size * data_count);

        return SUCCESS;
    }

    /* Convert the data through the data type pointers */
    switch (source_data_type)
    {
        case IAS_BYTE:
            switch (destination_data_type)
            {
                case IAS_I2:
                    for (index = 0; index < data_count; index++)
                    {
                        destination_buffer_I2[index] =
                            source_buffer_BYTE[index];
                    }
                    break;
                case IAS_UI2:
                    for (index = 0; index < data_count; index++)
                    {
                        destination_buffer_UI2[index] =
                            source_buffer_BYTE[index];
                    }
                    break;
                case IAS_R4:
                    for (index = 0; index < data_count; index++)
                    {
                        destination_buffer_R4[index] =
                            source_buffer_BYTE[index];
                    }
                    break;
                default:
                    IAS_LOG_ERROR("Unsupported destination data type");
                    return ERROR;
            }
            break;
        case IAS_I2:
            switch (destination_data_type)
            {
                case IAS_BYTE:
                    for (index = 0; index < data_count; index++)
                    {
                        destination_buffer_BYTE[index] =
                            source_buffer_I2[index];
                    }
                    break;
                case IAS_UI2:
                    for (index = 0; index < data_count; index++)
                    {
                        destination_buffer_UI2[index] = source_buffer_I2[index];
                    }
                break;
                case IAS_R4:
                    for (index = 0; index < data_count; index++)
                    {
                        destination_buffer_R4[index] = source_buffer_I2[index];
                    }
                    break;
                default:
                    IAS_LOG_ERROR("Unsupported destination data type");
                    return ERROR;
            }
            break;
        case IAS_UI2:
            switch (destination_data_type)
            {
                case IAS_BYTE:
                    for (index = 0; index < data_count; index++)
                    {
                        destination_buffer_BYTE[index] =
                            source_buffer_UI2[index];
                    }
                    break;
                case IAS_I2:
                    for (index = 0; index < data_count; index++)
                    {
                        destination_buffer_I2[index] = source_buffer_UI2[index];
                    }
                    break;
                case IAS_R4:
                    for (index = 0; index < data_count; index++)
                    {
                        destination_buffer_R4[index] = source_buffer_UI2[index];
                    }
                    break;
                default:
                    IAS_LOG_ERROR("Unsupported destination data type");
                    return ERROR;
            }
            break;
        case IAS_R4:
            switch (destination_data_type)
            {
                case IAS_BYTE:
                    for (index = 0; index < data_count; index++)
                    {
                        destination_buffer_BYTE[index] =
                            source_buffer_R4[index];
                    }
                    break;
                case IAS_I2:
                    for (index = 0; index < data_count; index++)
                    {
                        destination_buffer_I2[index] = source_buffer_R4[index];
                    }
                    break;
                case IAS_UI2:
                    for (index = 0; index < data_count; index++)
                    {
                        destination_buffer_UI2[index] = source_buffer_R4[index];
                    }
                break;
                default:
                    IAS_LOG_ERROR("Unsupported destination data type");
                    return ERROR;
            }
            break;
        default:
            IAS_LOG_ERROR("Unsupported source data type");
            return ERROR;
    }

    return SUCCESS;
}

