
/* Standard C Includes */
#include <stdlib.h>
#include <string.h>

/* IAS Includes */
#include "ias_gcp.h"
#include "ias_logging.h"

/* Local Includes */
#include "generate_chips.h"

/*****************************************************************************
 NAME: copy_gcp_locations

 PURPOSE: To copy all the requested GCP locations from the source GCPs to the
          destination GCPs.  This will only copy the ones which match the
          provided flag value.

 RETURN VALUE: Type = int
    Value    Description
    -------  --------------------------
    SUCCESS  Successful completion
    ERROR    Operation failed

  NOTE: Memory is allocated for and returned in the destination GCPs

*****************************************************************************/
int copy_gcp_locations
(
    GCP_VALID_FLAG_TYPE flag_value,         /* I: What to copy */
    const GENERATE_CHIPS_GCPS *source_gcps, /* I: Holds the input GCPs */
    GENERATE_CHIPS_GCPS *destination_gcps   /* O: Holds the output GCPs */
)
{
    int source_index;          /* Loop control */
    int destination_index;     /* Loop control */
    int source_number_of_gcps; /* The number of GCPs in the source data */

    /* Local source variable */
    GENERATE_CHIPS_GCP_LOCATION *source_gcp_locations;

    /* Local destination variables */
    GENERATE_CHIPS_GCP_LOCATION *destination_gcp_locations;

    /* Convert GCPs information to local variables */
    source_number_of_gcps = source_gcps->number_of_gcps;
    source_gcp_locations = source_gcps->gcp_locations;

    /* Allocate space for all of the source GCPs */
    destination_gcps->gcp_locations =
        calloc(source_number_of_gcps, sizeof(GENERATE_CHIPS_GCP_LOCATION));

    /* Now assign the local variable */
    destination_gcp_locations = destination_gcps->gcp_locations;

    if (destination_gcp_locations == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for GCP array");
        return ERROR;
    }
 
    /* Iterate through the GCPs and copy the valid ones to the destination */
    destination_index = 0;
    for (source_index = 0;
         source_index < source_number_of_gcps;
         source_index++)
    {
        /* Only copy the requested locations */
        if (source_gcp_locations[source_index].valid_flag == flag_value)
        {
            destination_gcp_locations[destination_index] =
                source_gcp_locations[source_index];

            destination_index++;
        }
    }

    /* Re-purposing the pointer after here */
    destination_gcp_locations = NULL;

    /* Reallocate the memory to the smaller size */
    if (destination_index != source_number_of_gcps && destination_index > 0)
    {
        destination_gcp_locations = realloc(destination_gcps->gcp_locations,
            sizeof(GENERATE_CHIPS_GCP_LOCATION) * destination_index);

        if (destination_gcp_locations == NULL)
        {
            IAS_LOG_ERROR("Reallocating memory for GCP array");
            free(destination_gcps->gcp_locations);
            destination_gcps->gcp_locations = NULL;
            return ERROR;
        }

        destination_gcps->gcp_locations = destination_gcp_locations;
    }

    /* Didn't find any chips so generate a warning */
    if (destination_index == 0)
    {
        IAS_LOG_WARNING("Didn't find any chips to copy");
        free(destination_gcps->gcp_locations);
        destination_gcps->gcp_locations = NULL;
    }

    destination_gcps->number_of_gcps = destination_index;

    return SUCCESS;
}

