
/* Standard C Includes */
#include <stdlib.h>
#include <string.h>

/* IAS Includes */
#include "ias_gcp.h"
#include "ias_logging.h"

/* Local Includes */
#include "generate_chips.h"

/*****************************************************************************
 NAME: copy_gcps

 PURPOSE: To copy all the flagged GCPs from the source records to the
          destination records.  This will only copy the ones which match the
          flag value.

 RETURN VALUE: Type = int
    Value    Description
    -------  --------------------------
    SUCCESS  Successful completion
    ERROR    Operation failed

  NOTE: Memory maybe realloced, and is returned in the destination GCPs

*****************************************************************************/
int copy_gcps
(
    GCP_VALID_FLAG_TYPE flag_value,        /* I: What to copy */
    const GENERATE_CHIPS_GCPS *source_gcps,/* I: Holds the input GCP records */
    GENERATE_CHIPS_GCPS *destination_gcps  /* O: Holds the output GCP records */
)
{
    int status;                /* capture function/routine statuses */
    int destination_index;     /* Loop control */
    int source_number_of_gcps; /* The number of gcps in the source data */

    /* Temp pointers for memory reallocation */
    IAS_GCP_RECORD *temp_gcp_records;
    GENERATE_CHIPS_GCP_LOCATION *temp_gcp_locations;

    /* Convert gcps information to local variables */
    source_number_of_gcps = source_gcps->number_of_gcps;

    /* Allocate the memory for the GCPs */
    status = allocate_gcp_memory(source_number_of_gcps, destination_gcps);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Allocating memory for GCPs");
        return ERROR;
    }

    /* Now assign to local variables after the memory allocation */
    temp_gcp_records = destination_gcps->gcp_records;
    temp_gcp_locations = destination_gcps->gcp_locations;

    destination_index = 0;
    copy_gcps_to_destination_index(flag_value, source_gcps,
        &destination_index, destination_gcps);

    /* Reallocate the memory to the smaller size */
    if (destination_index != source_number_of_gcps && destination_index > 0)
    {
        /* Get new memory */
        temp_gcp_records = realloc(destination_gcps->gcp_records,
            sizeof(IAS_GCP_RECORD) * destination_index);

        /* Make sure memory was allocated */
        if (temp_gcp_records == NULL)
        {
            IAS_LOG_ERROR("Reallocating memory for GCP records");
            free(destination_gcps->gcp_records);
            destination_gcps->gcp_records = NULL;
            free(destination_gcps->gcp_locations);
            destination_gcps->gcp_locations = NULL;
            return ERROR;
        }

        /* Fix the pointer */
        destination_gcps->gcp_records = temp_gcp_records;

        /* Get new memory */
        temp_gcp_locations = realloc(destination_gcps->gcp_locations,
            sizeof(GENERATE_CHIPS_GCP_LOCATION) * destination_index);

        /* Make sure memory was allocated */
        if (temp_gcp_locations == NULL)
        {
            IAS_LOG_ERROR("Reallocating memory for GCP arrays");
            free(destination_gcps->gcp_records);
            destination_gcps->gcp_records = NULL;
            free(destination_gcps->gcp_locations);
            destination_gcps->gcp_locations = NULL;
            return ERROR;
        }

        /* Fix the pointer */
        destination_gcps->gcp_locations = temp_gcp_locations;
    }

    /* Didn't find any chips so error out */
    if (destination_index == 0)
    {
        IAS_LOG_ERROR("No GCPs were found to copy");
        free(destination_gcps->gcp_records);
        destination_gcps->gcp_records = NULL;
        free(destination_gcps->gcp_locations);
        destination_gcps->gcp_locations = NULL;
        return ERROR;
    }

    /* Fix the count */
    destination_gcps->number_of_gcps = destination_index;

    return SUCCESS;
}

