/******************************************************************************
NAME:           create_fake_quality_band

PURPOSE:        Used to create a fake quality image. Set all attributes to the
                same values as source L1G image besides the image data.

RETURN VALUE: 
Type = SUCCESS/ERROR

******************************************************************************/
/* Standard C includes */
#include <stdlib.h>

/* IAS includes */
#include "ias_logging.h"
#include "ias_miscellaneous.h"

/* Local includes */
#include "generate_chips.h"

int create_fake_quality_band 
(
    IAS_IMAGE *image_ptr,               /* O: Populated from the L1G 
                                              image file */
    GENERATE_CHIPS_L1G_DATA *l1g_data,  /* I: L1G Image to populate 
                                              quality image */
    double fill_pixel
)
{
    void *band_data;        /* Buffer for the band_data */
    int count_filled_pixels;/* Counter for # of filled pixels */
    int index;              /* Loop variable */
    int input_data_count;   /* Sample time the number of lines */
    int input_sample_size;  /* L1G image data type */
    IAS_R4_TYPE *R4_input_buffer; /* Float buffer for data */
    int size = sizeof(unsigned short int);/* Always using size of IAS_UI2 */
    int status;             /* Status variable */
    unsigned short int *image_cast_ptr; /* Casted Image ptr to L1G image */

    /* Set pixel size */
    image_ptr->pixel_size_x = l1g_data->band_metadata.projection_distance_x;
    image_ptr->pixel_size_y = l1g_data->band_metadata.projection_distance_y;

    /* Set image sample & lines */
    image_ptr->ns = l1g_data->number_of_image_samples;
    image_ptr->nl = l1g_data->number_of_image_lines;
    input_data_count = image_ptr->ns * image_ptr->nl;
   
    /* Get the IAS_CORNERS values from L1G Image */
    image_ptr->corners.upleft.x = l1g_data->band_metadata.upper_left_x;
    image_ptr->corners.upleft.y = l1g_data->band_metadata.upper_left_y;
    image_ptr->corners.upright.x = l1g_data->band_metadata.upper_right_x;
    image_ptr->corners.upright.y = l1g_data->band_metadata.upper_right_y;
    image_ptr->corners.loleft.x = l1g_data->band_metadata.lower_left_x;
    image_ptr->corners.loleft.y = l1g_data->band_metadata.lower_left_y;
    image_ptr->corners.loright.x = l1g_data->band_metadata.lower_right_x;
    image_ptr->corners.loright.y = l1g_data->band_metadata.lower_right_y;

    /* Allocate space for the image data */
    image_ptr->data = malloc(image_ptr->ns * image_ptr->nl * size);

    /* Check if memory was allocated */
    if (image_ptr->data == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for the zero filled image");
        return ERROR;
    }  

    /* Figure out how many bytes are needed for a sample of the data,
       defaulting to the largest size supported */
    status = ias_misc_get_sizeof_data_type(l1g_data->data_type,
        &input_sample_size);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Determining size of L1G data type");
        return ERROR;
    }
    /* Allocate memory for the band */
    band_data = malloc(input_data_count * input_sample_size);
    if (band_data == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for input band data");
        free(image_ptr->data);
        return ERROR;
    }
 
    /* Allocate memory for converting the band into R4 */
    R4_input_buffer = malloc(input_data_count * sizeof(IAS_R4_TYPE));
    if (R4_input_buffer == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for R*4 converted band data");
        free(band_data);
        free(image_ptr->data);
        return ERROR;
    }

    /* Read the band into the allocated memory */
    if (ias_l1g_read_image(l1g_data->band_io_ptr, 0, 0, 0,
        image_ptr->nl, image_ptr->ns, band_data) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading band data from input image");
        free(band_data);
        free(image_ptr->data);
        free(R4_input_buffer);
        return ERROR;
    }

    /* Convert to the internal data type */
    status = convert_to_data_type(input_data_count, l1g_data->data_type,
        band_data, IAS_R4, R4_input_buffer);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Converting image data type to R*4");
        free(image_ptr->data);
        free(band_data);
        free(R4_input_buffer);
        return ERROR;
    }

    /* Iterate through the L1G Image and determining what is background
       by checking if L1G value is equal to fill_pixel and then setting
       the quality image value to IAS_QUALITY_BIT_FILL */
    count_filled_pixels = 0;
    image_cast_ptr = image_ptr->data;
    for (index = 0; index < input_data_count; index++)
    {
        if (R4_input_buffer[index] == fill_pixel)
        {
            image_cast_ptr[index] = IAS_QUALITY_BIT_FILL;
            count_filled_pixels++;
        }
    }

    /* Print out how many fill pixels there were in L1G image */
    IAS_LOG_DEBUG("L1G image had %d background pixels", count_filled_pixels);

    free(band_data);
    free(R4_input_buffer);

    return SUCCESS;
}
