
/* Standard C Includes */
#include <stdlib.h>
#include <string.h>

/* IAS Includes */
#include "ias_const.h"
#include "ias_types.h"
#include "ias_logging.h"

/* Local Includes */
#include "generate_chips.h"

#define MORAVEC_ENVI_POINT_FORMAT "%lf %lf %lf %lf %lf %d\n"

/*****************************************************************************
 NAME: write_envi_points

 PURPOSE: Used internally to this file.  Used to output the ENVI points based
          on the provided input.  If the line and sample offsets are zero,
          then it is assumed you want a single point to be written.  Otherwise,
          it is assumed that you want a boxed point and also to output the
          moravec values and resolution line and samples.

 RETURN VALUE: None

*****************************************************************************/
void write_envi_points
(
    FILE *output_fd,                 /* I: File descriptor for the output */
    const GENERATE_CHIPS_GCPS *gcps, /* I: The GCPs generated to output */
    double line_offset,  /* I: Line offset for generating the bounding box */
    double sample_offset /* I: Sample offset for generating the bounding box */
)
{
    int point_index; /* Loop control */
    double i_line;   /* input data_line */
    double i_sample; /* input data sample */
    double r_line;   /* resolution line */
    double r_sample; /* resolution sample */

    /* Process through each record */
    for (point_index = 0; point_index < gcps->number_of_gcps; point_index++)
    {
        i_line = gcps->gcp_locations[point_index].input_data_line;
        i_sample = gcps->gcp_locations[point_index].input_data_sample;
        r_line = gcps->gcp_locations[point_index].resolution_line;
        r_sample = gcps->gcp_locations[point_index].resolution_sample;

        if (line_offset == 0 && sample_offset == 0)
        {
            /* Write non-boxed ENVI points */
            fprintf(output_fd, "%lf %lf\n", i_line, i_sample);
        }
        else
        {
            /* Write boxed ENVI points including the moravec resolution
               locations, and other information */
            fprintf(output_fd, MORAVEC_ENVI_POINT_FORMAT,
                i_line - line_offset, i_sample - sample_offset,
                r_line - line_offset, r_sample - sample_offset,
                gcps->gcp_locations[point_index].moravec_value,
                gcps->gcp_locations[point_index].valid_flag);

            fprintf(output_fd, MORAVEC_ENVI_POINT_FORMAT,
                i_line - line_offset, i_sample + sample_offset,
                r_line - line_offset, r_sample + sample_offset,
                gcps->gcp_locations[point_index].moravec_value,
                gcps->gcp_locations[point_index].valid_flag);

            fprintf(output_fd, MORAVEC_ENVI_POINT_FORMAT,
                i_line + line_offset, i_sample + sample_offset,
                r_line + line_offset, r_sample + sample_offset,
                gcps->gcp_locations[point_index].moravec_value,
                gcps->gcp_locations[point_index].valid_flag);

            fprintf(output_fd, MORAVEC_ENVI_POINT_FORMAT,
                i_line + line_offset, i_sample - sample_offset,
                r_line + line_offset, r_sample - sample_offset,
                gcps->gcp_locations[point_index].moravec_value,
                gcps->gcp_locations[point_index].valid_flag);

            fprintf(output_fd, MORAVEC_ENVI_POINT_FORMAT,
                i_line - line_offset, i_sample - sample_offset,
                r_line - line_offset, r_sample - sample_offset,
                gcps->gcp_locations[point_index].moravec_value,
                gcps->gcp_locations[point_index].valid_flag);
        }
    } /* END point loop */
}

/*****************************************************************************
 NAME: open_output_file

 PURPOSE: Called to open/create the output file.

 RETURN VALUE: File descriptor

*****************************************************************************/
FILE * open_output_file
(
    const char *output_directory,
    const char *filename
)
{
    int count;                      /* For testing the snprintf */
    char output_filename[PATH_MAX]; /* The name of the output file to create */
    FILE *output_fd;                /* File descriptor for the output file */

    /* Figure out the filename and path */
    count = snprintf(output_filename, sizeof(output_filename), "%s/%s",
        output_directory, filename);

    if (count < 0 || count >= sizeof(output_filename))
    {
        IAS_LOG_WARNING("Bounds exceeded or internal snprintf error:"
            " count = %d, filename = %s", count, filename);
        return NULL;
    }

    /* Open the file */
    output_fd = fopen(output_filename, "w");
    if (output_fd == NULL)
    {
        IAS_LOG_WARNING("Unable to open %s for writing", output_filename);
        return NULL;
    }

    return output_fd;
}

/*****************************************************************************
 NAME: dump_envi_point_file

 PURPOSE: Called to write single point ENVI output.

 RETURN VALUE: Type = int
    Value    Description
    -------  ---------------------
    SUCCESS  Successful completion
    ERROR    Operation failed

*****************************************************************************/
int dump_envi_point_file
(
    const GENERATE_CHIPS_PARAMETERS *parameters, /* I: The input parameters */
    const char *filename,           /* I: The filename to create */
    const GENERATE_CHIPS_GCPS *gcps /* I: The GCPs generated to output */
)
{
    FILE *output_fd;      /* File descriptor for the output file */

    output_fd = open_output_file(parameters->output_directory, filename);
    if (output_fd == NULL)
    {
        IAS_LOG_ERROR("Unable to open %s", filename);
        return ERROR;
    }

    /* Output the points */
    /* Generate single point output by specifying zero offsets */
    write_envi_points(output_fd, gcps, 0, 0);

    /* Close the file */
    if (fclose(output_fd) == EOF)
    {
        IAS_LOG_ERROR("Closing file %s", filename);
        return ERROR;
    }

    return SUCCESS;
}

/*****************************************************************************
 NAME: dump_intermediate_envi_point_file

 PURPOSE: Called to write boxed ENVI point output, along with other moravec
          information.

 RETURN VALUE: Type = int
    Value    Description
    -------  ---------------------
    SUCCESS  Successful completion
    ERROR    Operation failed

*****************************************************************************/
int dump_intermediate_envi_point_file
(
    const GENERATE_CHIPS_PARAMETERS *parameters, /* I: The input parameters */
    const char *filename,           /* I: The filename to create */
    const GENERATE_CHIPS_GCPS *gcps /* I: The GCPs generated to output */
)
{
    FILE *output_fd;      /* File descriptor for the output file */
    double line_offset;   /* Line offset for generating the bounding box */
    double sample_offset; /* Sample offset for generating the bounding box */

    output_fd = open_output_file(parameters->intermediate_output_directory,
        filename);
    if (output_fd == NULL)
    {
        IAS_LOG_ERROR("Unable to open %s", filename);
        return ERROR;
    }

    /* Generate boxed output by specifying non-zero offsets */
    line_offset = parameters->chip_size_lines / 2.0;
    sample_offset = parameters->chip_size_samples / 2.0;

    /* Output the points */
    write_envi_points(output_fd, gcps, line_offset, sample_offset);

    /* Close the file */
    if (fclose(output_fd) == EOF)
    {
        IAS_LOG_ERROR("Closing file %s", filename);
        return ERROR;
    }

    return SUCCESS;
}

