
/* Standard C Includes */
#include <stdlib.h>

/* IAS Includes */
#include "ias_const.h"
#include "ias_types.h"
#include "ias_logging.h"
#include "ias_l1g.h"
#include "ias_geo.h"
#include "ias_gcp.h"

/* Local Includes */
#include "generate_chips.h"

/* Define the SCA we will extract from */
#define EXTRACT_FROM_SCA 0

/*****************************************************************************
 NAME: extract_gcp_chips

 PURPOSE: To extract the GCP chips from the input image and write them to
          disk.  Validation is performed so that bad chips are not created.

 RETURN VALUE: Type = int
    Value    Description
    -------  --------------------------
    SUCCESS  Successful completion
    ERROR    Operation failed

*****************************************************************************/
int extract_gcp_chips
(
    const GENERATE_CHIPS_PARAMETERS *parameters, /* I: The input parameters */
    const GENERATE_CHIPS_L1G_DATA *l1g_data, /* I: Input L1G data */
    IAS_BYTE_TYPE *exclusion_mask,   /* I: The resulting exclusion mask */
    const IAS_GEO_PROJ_TRANSFORMATION *input_image2geographic,
                                /* I: The projection transformation to use */
    int number_mask_lines,             /* Number of lines in mask */
    int number_mask_samples,           /* Number of samples in mask */
    GENERATE_CHIPS_GCPS *gcps       /* I/O: GCPs to attempt extraction on */
)
{
    int status;        /* Capture function/routine statuses */
    int index;         /* Loop control */
    int line_index;    /* Loop control */
    int sample_index;  /* Loop control */
    int line_position; /* The start of the line in the quality band */

    /* The max size of a line and sample */
    double max_line_size;
    double max_sample_size;

    /* Tolerances to provide a buffer around the image for how close the
       reference line and sample values can be to the edge of the image. */
    double top_line_tolerance;
    double bottom_line_tolerance;
    double left_sample_tolerance;
    double right_sample_tolerance;

    int chip_size[2]; /* Only used for ias_geo_extract_window */

    /* Chip buffer pointers */
    IAS_R4_TYPE *chip_buffer = NULL; /* Chip Buffer */
    void *output_chip_buffer = NULL; /* Output Chip Buffer */

    int chip_buffer_samples; /* The size of the chip buffer in samples */

    /* Variables used to find the GCP chip in the mask */
    int image_start_line;
    int image_start_sample;
    int image_end_line;
    int image_end_sample;
    int mask_start_line;
    int mask_start_sample;
    int mask_end_line;
    int mask_end_sample;
    int mask_chip_size;
    double line_conversion;
    double sample_conversion;

    int fill_count;         /* The amount of fill samples in the chip */
    int water_count;        /* The amount of water samples in the chip */
    int snow_and_ice_count; /* The amount of snow/ice samples in the chip */

    IAS_BYTE_TYPE bit_flags; /* Holds a union of all the bits set in the
                               current chip data */
    IAS_BYTE_TYPE mask_value; /* Holds the value of the current mask sample */

    double out_of_range; /* The percentage of the selected bit value in the
                            mask chip */
    IAS_GCP_IMAGE_CHIP_IO *chip_io; /* Image chip IO structure */

    /* Convert L1G information to local variables */
    int number_of_image_lines = l1g_data->number_of_image_lines;
    int number_of_image_samples = l1g_data->number_of_image_samples;
    IAS_DATA_TYPE data_type = l1g_data->data_type;
    L1G_BAND_IO *l1g_band = l1g_data->band_io_ptr;
    int number_of_gcps = gcps->number_of_gcps;
    GENERATE_CHIPS_GCP_LOCATION *gcp_locations = gcps->gcp_locations;
    IAS_GCP_RECORD *gcp_records = gcps->gcp_records;

    /* Find the max size for a chip */
    max_line_size = 0.0;
    max_sample_size = 0.0;
    for (index = 0; index < number_of_gcps; index++)
    {
        if (gcp_records[index].chip_size_lines > max_line_size)
        {
            max_line_size = gcp_records[index].chip_size_lines;
        }

        if (gcp_records[index].chip_size_samples > max_sample_size)
        {
            max_sample_size = gcp_records[index].chip_size_samples;
        }
    }

    /* Allocate the chip buffer memory */
    chip_buffer_samples = max_line_size * max_sample_size;
    chip_buffer = calloc(chip_buffer_samples, sizeof(*chip_buffer));
    if (chip_buffer == NULL)
    {
        IAS_LOG_ERROR("Allocating the chip buffer");
        return ERROR;
    }
    /* Allocate memory for the max supported output chip buffer size */
    output_chip_buffer = calloc(chip_buffer_samples, sizeof(IAS_R4_TYPE));
    if (output_chip_buffer == NULL)
    {
        IAS_LOG_ERROR("Allocating the output chip buffer");
        free(chip_buffer);
        return ERROR;
    }

    /* Create the image chips container (always a directory for this 
       application) */
    chip_io = ias_gcp_open_image_chip_container(parameters->output_directory,
        IAS_GCP_CONTAINER_DIR, IAS_WRITE);
    if (!chip_io)
    {
        IAS_LOG_ERROR("Opening image chip container at %s",
            parameters->output_directory);
        free(output_chip_buffer);
        free(chip_buffer);
        return ERROR;
    }

    /* Find the conversion values to locate the chip pixels in the mask */
    line_conversion = (double) number_mask_lines / number_of_image_lines;
    sample_conversion = (double) number_mask_samples / number_of_image_samples;

    /* Process through the gcps and generate the chips */
    for (index = 0; index < number_of_gcps; index++)
    {
        /* If we are already false then skip this point */
        if (gcp_locations[index].valid_flag == GCP_VALID_FLAG_DONT_USE)
        {
            continue;
        }
         
        /* Compute the tolerance values for the edge of the image. */
        top_line_tolerance = (gcp_records[index].chip_size_lines / 2) + 2;
        bottom_line_tolerance = number_of_image_lines - top_line_tolerance;

        left_sample_tolerance = (gcp_records[index].chip_size_samples / 2) + 2;
        right_sample_tolerance =
            number_of_image_samples - left_sample_tolerance;

        /* Only extract the chip if it is in the image */
        if ((gcp_locations[index].input_data_line > top_line_tolerance)
            && (gcp_locations[index].input_data_line
                < bottom_line_tolerance)
            && (gcp_locations[index].input_data_sample
                > left_sample_tolerance)
            && (gcp_locations[index].input_data_sample
                < right_sample_tolerance))
        {
            chip_size[0] = gcp_records[index].chip_size_lines;
            chip_size[1] = gcp_records[index].chip_size_samples;

            /* Extract the chip from the image */
            status = ias_geo_extract_window(
                number_of_image_lines, number_of_image_samples,
                EXTRACT_FROM_SCA, data_type,
                gcp_locations[index].input_data_line,
                gcp_locations[index].input_data_sample,
                chip_size, l1g_band, chip_buffer,
                &image_start_line, &image_start_sample);

            if (status != SUCCESS)
            {
                gcp_locations[index].valid_flag = GCP_VALID_FLAG_DONT_USE;
                continue;
            }

            /* Set the buffer size for this GCP */
            chip_buffer_samples = gcp_records[index].chip_size_lines
                * gcp_records[index].chip_size_samples;

            /*** Validate the chip ***/

            /* Determine if the GCP exceeds the valid pixel threshold */
            status = ias_math_check_pixels_in_range(chip_buffer,
                chip_buffer_samples, parameters->valid_pixel_threshold,
                parameters->max_range_value, parameters->min_range_value);

            if (status == FALSE)
            {
                gcp_locations[index].valid_flag = GCP_VALID_FLAG_DONT_USE;
                continue;
            }

            /* The start and end calculations below, do not take into
               consideration for the edges of the data
               This is because we are exculding those situations with a check
               before we get here */

            /* Find the ending line and sample in the input image */ 
            image_end_line = image_start_line
                + gcp_records[index].chip_size_lines;
            image_end_sample = image_start_sample
                + gcp_records[index].chip_size_samples;

            /* Find the starting line and sample in the quality image
               Round to the nearest int */ 
            mask_start_line =
                (int)(image_start_line * line_conversion);
            mask_end_line = (int)(image_end_line * line_conversion);

            /* Find the ending line and sample in the quality image
               Round to the nearest int */ 
            mask_start_sample =
                (int)(image_start_sample * sample_conversion);
            mask_end_sample =
                (int)(image_end_sample * sample_conversion);

            /* Determine how many will be checked */
            mask_chip_size = (mask_end_line - mask_start_line)
                * (mask_end_sample - mask_start_sample);

            /* For the following bit tests......
               If the 10 confidence bit for the quality band type is set, then
               both the 10 and 11 confidence levels will be excluded.  And we
               will keep the 01 confidence level data. */

            /* Reset the flag discriminator variable */
            bit_flags = 0x0000;
            fill_count = 0;
            water_count = 0;
            snow_and_ice_count = 0;
            for (line_index = mask_start_line;
                 line_index < mask_end_line;
                 line_index++)
            {
                line_position = line_index * number_mask_samples;

                for (sample_index = mask_start_sample;
                     sample_index < mask_end_sample;
                     sample_index++)
                {
                    /* Get the value */
                    mask_value = exclusion_mask[line_position + sample_index];

                    /* OR the flags into one variable */
                    bit_flags |= mask_value;

                    /* Get a fill count here for a tolerance check below */
                    if (mask_value & EXCLUSION_MASK_FILL)
                    {
                        fill_count++;
                    }

                    /* Get a water count here for a tolerance check below */
                    if (mask_value & EXCLUSION_MASK_WATER)
                    {
                        water_count++;
                    }

                    /* Get a snow and ice count here for a tolerance check
                       below */
                    if (mask_value & EXCLUSION_MASK_SNOW_ICE)
                    {
                        snow_and_ice_count++;
                    }
                }
            }

            /* If the chip contains any fill pixels we check to see if we have
               too many fill pixels */
            if ((bit_flags & EXCLUSION_MASK_FILL))
            {
                out_of_range = (double)fill_count / (double)mask_chip_size;
                if (out_of_range > parameters->valid_pixel_threshold)
                {
                    gcp_locations[index].valid_flag = GCP_VALID_FLAG_DONT_USE;
                    continue;
                }
            }

            /* If the chip contains any water we check to see if we have too
               much water */
            if ((bit_flags & EXCLUSION_MASK_WATER))
            {
                out_of_range = (double)water_count / (double)mask_chip_size;
                if (out_of_range > parameters->valid_water_threshold)
                {
                    gcp_locations[index].valid_flag = GCP_VALID_FLAG_DONT_USE;
                    continue;
                }
            }

            /* If the chip contains any snow and ice we check to see if we
               have too much snow and ice */
            if ((bit_flags & EXCLUSION_MASK_SNOW_ICE))
            {
                out_of_range =
                    (double)snow_and_ice_count / (double)mask_chip_size;
                if (out_of_range > parameters->valid_snow_and_ice_threshold)
                {
                    gcp_locations[index].valid_flag = GCP_VALID_FLAG_DONT_USE;
                    continue;
                }
            }

            /* If the chip contains any cirrus cloud data we do not want to
               keep it */
            if ((bit_flags & EXCLUSION_MASK_CIRRUS))
            {
                gcp_locations[index].valid_flag = GCP_VALID_FLAG_DONT_USE;
                continue;
            }

            /* If the chip contains any cloud data we do not want to keep it */
            if ((bit_flags & EXCLUSION_MASK_CLOUD))
            {
                gcp_locations[index].valid_flag = GCP_VALID_FLAG_DONT_USE;
                continue;
            }

            /*** Save the chip buffer to a file ***/

            /* Convert the data to the output type, which is the data type of
               the source image */
            status = convert_to_data_type(chip_buffer_samples,
                IAS_R4, chip_buffer, data_type, output_chip_buffer);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Converting chip data type");
                ias_gcp_close_image_chip_container(chip_io);
                free(output_chip_buffer);
                free(chip_buffer);
                return ERROR;
            }

            /* Output the image chip buffer */
            status = ias_gcp_write_image_chip(chip_io,
                    gcp_records[index].chip_name,
                    gcp_records[index].chip_size_lines,
                    gcp_records[index].chip_size_samples, data_type,
                    output_chip_buffer);

            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Creating chip %s", gcp_records[index].chip_name);
                ias_gcp_close_image_chip_container(chip_io);
                free(output_chip_buffer);
                free(chip_buffer);
                return ERROR;
            }

            /*** Update some parts of the GCPLib record ***/

            /* Update the chip reference line and sample to the
               reference point */
            gcp_records[index].reference_line =
                gcp_locations[index].input_data_line - image_start_line;
            gcp_records[index].reference_sample =
                gcp_locations[index].input_data_sample - image_start_sample;

            gcp_locations[index].valid_flag = GCP_VALID_FLAG_USE;
        }
        else
        {
            gcp_locations[index].valid_flag = GCP_VALID_FLAG_DONT_USE;
        }
    }

    /* Free the buffer memory */
    free(output_chip_buffer);
    output_chip_buffer = NULL;
    free(chip_buffer);
    chip_buffer = NULL;

    /* Close the image chip container */
    if (ias_gcp_close_image_chip_container(chip_io) != SUCCESS)
    {
        IAS_LOG_ERROR("Failed closing image chip container %s",
            parameters->output_directory);
        return ERROR;
    }

    return SUCCESS;
}

