
/*****************************************************************************
 NAME: generate_chips.c

 PURPOSE: Create GCPs from a specified input image using one of three
          implemented methods for chip generation.

 RETURN VALUE: Type = int
    Value         Description
    ------------  --------------------------
    EXIT_SUCCESS  Successful completion
    EXIT_FAILURE  Program terminated with errors

*****************************************************************************/

/* Standard C Includes */
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/* IAS Includes */
#include "gctp.h"
#include "ias_const.h"
#include "ias_types.h"
#include "ias_logging.h"
#include "ias_l1g.h"
#include "ias_gcp.h"
#include "ias_geo.h"
#include "ias_miscellaneous.h"

/* Local Includes */
#include "generate_chips.h"

int main(int argc, char *argv[])
{
    int status;      /* Capture function/routine statuses */
    int count;       /* Capture snprintf errors */
    int index;       /* Loop control */
    GENERATE_CHIPS_PARAMETERS parameters; /* Holds the input parameters */
    IAS_CORNERS input_corners; /* Holds the L1G image corners */
    IAS_BYTE_TYPE *exclusion_mask; /* Exclusion mask */
    IAS_BYTE_TYPE *water_mask = NULL; /* Water mask */

    /* Hold the capture date of the input data */
    int capture_year;
    int capture_month;
    int capture_day;
    
    /*** L1G Input Data Variables ***/
    GENERATE_CHIPS_L1G_DATA input_data;
    IAS_L1G_BAND_METADATA *metadata;
    IAS_IMAGE quality_image;
    IAS_IMAGE elevation_image;
    IAS_IMAGE water_image; /* Image data used to build mask */

    /*** GCP Variables ***/
    GENERATE_CHIPS_GCPS evenly_spaced_gcps; /* Used by the Evenly Spaced chip
        generation to store/hold the initial set of GCPs */
    GENERATE_CHIPS_GCPS moravec_gcps; /* Used by the Moravec chip generation
        to store/hold the initial set of GCPs */
    GENERATE_CHIPS_GCPS built_gcps; /* Used by the GCPLib chip to store/hold
        the initial set of GCPs.  This variable is then used by the extraction
        code for extracting and specifying the final set of GCPs.  In the
        Evenly Spaced and Moravec options(including the combined option) all
        of the valid initial set of GCPs are copied to this variable to be
        used by the extraction. */
    GENERATE_CHIPS_GCPS extracted_gcps; /* Post extraction, all of the valid
        GCPs contained in built_gcps are copied to this variable for updating
        before they are written to the GCPLib file. */
    char output_gcplib_path[PATH_MAX];

    /*** Projection Variables ***/
    IAS_PROJECTION geographic_projection;
    IAS_PROJECTION input_image_projection;
    /* Geographic and UTM parameters are the same */
    double geographic_projection_parameters[IAS_PROJ_PARAM_SIZE];
    int projection_units;
    IAS_GEO_PROJ_TRANSFORMATION *input_image2geographic = NULL;

    /* Initialize the logging */
    status = ias_log_initialize("generate_chips");
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the logging library");
        return EXIT_FAILURE;
    }

    /* Verify that the parameter filename was provided */
    if (argc != 2)
    {
        IAS_LOG_ERROR("Usage: generate_chips <parameter filename>");
        return EXIT_FAILURE;
    }

    /* Read the parameters from the parameter file */
    status = read_parameters(argv[1], &parameters);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading input parameters");
        return EXIT_FAILURE;
    }

    /* Notify the user that the application has started */
    IAS_LOG_INFO("Started");

    /* Initialize the output directory */
    if (initialize_directory(parameters.output_directory) 
        != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the output directory %s",
            parameters.output_directory);
        return EXIT_FAILURE;
    }

    /* Check if the intermediate output directory needs to be created */
    if (parameters.output_intermediate_data)
    {
        if (initialize_directory(parameters.intermediate_output_directory) 
            != SUCCESS)
        {
            IAS_LOG_ERROR("Initializing the intermediate output directory %s",
                parameters.intermediate_output_directory);
            return EXIT_FAILURE;
        }
    }

    /* Open the input image and load the metadata */
    status = open_l1g_data(parameters.l1g_filename, parameters.band_number,
        &input_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Opening input image data file %s",
            parameters.l1g_filename);
        return EXIT_FAILURE;
    }

    /* Test for unsupported input data types */
    switch (input_data.data_type)
    {
        case IAS_BYTE:
        case IAS_I2:
        case IAS_UI2:
        case IAS_R4:
            /* Do nothing, these are the supported types */
            break;

        default:
        IAS_LOG_ERROR("The input image contains an unsupported data type");
            close_l1g_data(&input_data);
            return EXIT_FAILURE;
    }

    /* Create pointer to band_metadata */
    metadata = &input_data.band_metadata;

    /* Creating Corners struct from input data */
    input_corners =(struct IAS_CORNERS){
        {metadata->upper_left_x, metadata->upper_left_y},
        {metadata->upper_right_x, metadata->upper_right_y},
        {metadata->lower_left_x, metadata->lower_left_y},
        {metadata->lower_right_x, metadata->lower_right_y}
    };

    /* Check if quality image parameter was passed in */
    if (strlen(parameters.quality_filename) == 0)
    {
        /* Warn use about using zero filled quality image */
        IAS_LOG_WARNING("Quality image absent using zero fill image");

        /* Zero fill the quality image */
        status = create_fake_quality_band(&quality_image, &input_data,
            parameters.background);

        /* Check if zero fill status returned with SUCCESS */
        if (status == ERROR)
        {
            IAS_LOG_ERROR("Could not zero fill quality image");
            close_l1g_data(&input_data);
            return EXIT_FAILURE;    
        }
    }
    else
    {
        /* Load the quality data */
        if (ias_misc_read_coregistered_single_band_l1g(
                parameters.quality_filename, &input_corners,
                input_data.number_of_image_lines, 
                input_data.number_of_image_samples,
                metadata->projection_distance_y,
                &quality_image, IAS_UI2) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading L1G quality image file %s",
                parameters.quality_filename);
            close_l1g_data(&input_data);
            return EXIT_FAILURE;
        }
    }

    /* Maybe open the elevation data and load the metadata */
    if (parameters.elevation_filename_present)
    {
        /* Load the elevation data */
        if (ias_misc_read_coregistered_single_band_l1g(
                parameters.elevation_filename, &input_corners,
                input_data.number_of_image_lines,
                input_data.number_of_image_samples,
                metadata->projection_distance_y, &elevation_image,
                IAS_I2) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading L1G elevation image file %s",
                parameters.elevation_filename);
            free(quality_image.data);
            close_l1g_data(&input_data);
            return EXIT_FAILURE;
        }
    }
    else
    {
        /* Since the elevation data was not read, make sure the data pointer
           is initialized */
        elevation_image.data = NULL;
    }

    /* Read the wrs_path and wrs_row values from the input image metadata.  If
       they are not valid then use the values from the input parameters.
       Verify that values are provided one way or the other.   Provide a
       warning if the input parameters are defined and they do not match. */
    if (input_data.metadata.wrs_path == 0 && parameters.target_wrs_path == 0)
    {
        IAS_LOG_ERROR("WRS_PATH must be defined in the L1G metadata or"
            " TARGET_WRS_PATH must be defined in the input parameter file");
        free(elevation_image.data);
        free(quality_image.data);
        close_l1g_data(&input_data);
        return EXIT_FAILURE;
    }
    if (input_data.metadata.wrs_path != 0 && parameters.target_wrs_path != 0
        && input_data.metadata.wrs_path != parameters.target_wrs_path)
    {
        IAS_LOG_WARNING("Image WRS path and input parameter TARGET_WRS_PATH"
            " do not match (%d, %d), using Image WRS path value",
            input_data.metadata.wrs_path, parameters.target_wrs_path);
    }
    if (input_data.metadata.wrs_path != 0)
    {
        parameters.target_wrs_path = input_data.metadata.wrs_path;
    }
    else
    {
        IAS_LOG_WARNING("Image WRS path not defined, using input parameter"
            " path value %d", parameters.target_wrs_path);
    }
    IAS_LOG_DEBUG("Target WRS Path = %d", parameters.target_wrs_path);

    if (input_data.metadata.wrs_row == 0 && parameters.target_wrs_row == 0)
    {
        IAS_LOG_ERROR("WRS_ROW must be defined in the L1G metadata or"
            " TARGET_WRS_ROW must be defined in the input parameter file");
        free(elevation_image.data);
        free(quality_image.data);
        close_l1g_data(&input_data);
        return EXIT_FAILURE;
    }
    if (input_data.metadata.wrs_row != 0 && parameters.target_wrs_row != 0
        && input_data.metadata.wrs_row != parameters.target_wrs_row)
    {
        IAS_LOG_WARNING("Image WRS row and input parameter TARGET_WRS_ROW"
            " do not match (%d, %d), using Image WRS row value",
            input_data.metadata.wrs_row, parameters.target_wrs_row);
    }
    if (input_data.metadata.wrs_row != 0)
    {
        parameters.target_wrs_row = input_data.metadata.wrs_row;
    }
    else
    {
        IAS_LOG_WARNING("Image WRS row not defined, using input parameter row"
                        " value %d", parameters.target_wrs_row);
    }

    IAS_LOG_DEBUG("Target WRS Row = %d", parameters.target_wrs_row);

    /* Verify that a date was provided in the input data */
    status = sscanf(input_data.metadata.capture_date, "%4d/%2d/%2d",
        &capture_year, &capture_month, &capture_day);
    if (status != 3)
    {
        IAS_LOG_ERROR("Capture date not provided in the input image");
        free(elevation_image.data);
        free(quality_image.data);
        close_l1g_data(&input_data);
        return EXIT_FAILURE;
    }

    /* Verify the projection is supported */
    if (input_data.metadata.projection_code != UTM
        && input_data.metadata.projection_code != PS)
    {
        IAS_LOG_ERROR("Input data projection is not supported");
        free(elevation_image.data);
        free(quality_image.data);
        close_l1g_data(&input_data);
        return EXIT_FAILURE;
    }

    /* Create the geographic projection information */
    for (index = 0; index < IAS_PROJ_PARAM_SIZE; index++)
    {
        geographic_projection_parameters[index] = 0.0;
    }

    ias_geo_set_projection(GEO, NULLZONE, DEGREE, WGS84_SPHEROID,
        geographic_projection_parameters, &geographic_projection);

    /* Create the projection information for the input image */
    if (ias_geo_get_units(input_data.metadata.projection_units,
        &projection_units) != SUCCESS)
    {
        IAS_LOG_ERROR("Converting projection units %s to unit code",
            input_data.metadata.projection_units);
        free(elevation_image.data);
        free(quality_image.data);
        close_l1g_data(&input_data);
        return EXIT_FAILURE;
    }

    ias_geo_set_projection(input_data.metadata.projection_code,
        input_data.metadata.zone_code, projection_units,
        input_data.metadata.spheroid_code,
        input_data.metadata.projection_parameters, &input_image_projection);
    
    /* Allocate memory for the water mask */
    if (parameters.use_water_mask_flag)
    {
        water_mask = calloc(input_data.number_of_image_lines *
            input_data.number_of_image_samples, sizeof(*water_mask));
        if (!water_mask)
        {
            IAS_LOG_ERROR("Error allocating memory for water mask");
            free(elevation_image.data);
            free(quality_image.data);
            close_l1g_data(&input_data);
            return EXIT_FAILURE;
        }
    }
    
    /* Check if water mask file is passed in */
    if (parameters.use_water_mask_flag == USE_POLYGON_WATER_MASK) 
    {
        /* Copy to image data to image struct */
        water_image.data_type = input_data.data_type;
        water_image.corners = (struct IAS_CORNERS){
            {metadata->upper_left_x, metadata->upper_left_y},
            {metadata->upper_right_x, metadata->upper_right_y},
            {metadata->lower_left_x, metadata->lower_left_y},
            {metadata->lower_right_x, metadata->lower_right_y}};
        water_image.pixel_size_y = metadata->projection_distance_y;
        water_image.pixel_size_x = metadata->projection_distance_x;
        water_image.nl = input_data.number_of_image_lines;
        water_image.ns = input_data.number_of_image_samples;
        water_image.band_number = metadata->band_number;

        /* Build mask and polygon list */
        if (ias_geo_shape_mask_projection(parameters.water_mask_filename, 
            &water_image, &input_image_projection, water_mask) != SUCCESS)
        {
            IAS_LOG_ERROR("Creating the water mask");
            free(elevation_image.data);
            free(quality_image.data);
            free(water_mask);
            close_l1g_data(&input_data);
            return ERROR;
        }
    }
    else if (parameters.use_water_mask_flag == USE_BINARY_WATER_MASK) 
    {
        FILE *binary_water_mask; /* Binary water mask file pointer */
        int total_samples;       /* Total number of samples to read */
        
        /* Determine the total number of samples to read */
        total_samples = input_data.number_of_image_lines *
            input_data.number_of_image_samples;

        /* Open the binary water mask */
        binary_water_mask = fopen(parameters.water_mask_filename, "r");
        if (!binary_water_mask)
        {
            IAS_LOG_ERROR("Opening binary water mask %s", 
                parameters.water_mask_filename);
            free(elevation_image.data);
            free(quality_image.data);
            free(water_mask);
            close_l1g_data(&input_data);
            return EXIT_FAILURE;
        }

        /* Read the entire water mask at once */
        if (fread(water_mask, 1, total_samples, binary_water_mask) 
            != total_samples)
        {
            IAS_LOG_ERROR("Reading in binary water mask %s",
                parameters.water_mask_filename);
            free(elevation_image.data);
            free(quality_image.data);
            free(water_mask);
            fclose(binary_water_mask);
            close_l1g_data(&input_data);
            return EXIT_FAILURE;
        }

        /* Close the water mask file */
        if (fclose(binary_water_mask) != 0)
        {
            IAS_LOG_ERROR("Closing binary water mask %s",
                parameters.water_mask_filename);
            free(elevation_image.data);
            free(quality_image.data);
            free(water_mask);
            close_l1g_data(&input_data);
            return EXIT_FAILURE;
        }
    }

    /* Allocate memory for the mask data */
    exclusion_mask = malloc(quality_image.nl * quality_image.ns 
        * sizeof(*exclusion_mask));
    if (!exclusion_mask)
    {
        IAS_LOG_ERROR("Allocating memory for mask data");
        free(elevation_image.data);
        free(quality_image.data);
        free(water_mask);
        close_l1g_data(&input_data);
        return EXIT_FAILURE;
    }

    /* Build the exclusion mask */
    if (build_exclusion_mask(&parameters, &quality_image, water_mask, 
            exclusion_mask) != SUCCESS)
    {
        IAS_LOG_ERROR("Building the exclusion_mask");
        free(elevation_image.data);
        free(quality_image.data);
        free(exclusion_mask);
        free(water_mask);
        close_l1g_data(&input_data);
        return EXIT_FAILURE;
    }

    /* Check if output of mask is needed */    
    if (parameters.output_intermediate_data)
    {
        char output_filename[PATH_MAX];
        int count; 

        /* Check if output of water mask is needed */    
        if (parameters.use_water_mask_flag)
        {
            /* Create the full output file path */                     
            count = snprintf(output_filename, sizeof(output_filename), 
                "%s/water_mask.img", parameters.intermediate_output_directory);
            if (count < 0 || count >= sizeof(output_filename))
            {
                IAS_LOG_ERROR("Bounds exceeded or internal snprintf error:"
                    " count = %d", count);
                free(elevation_image.data);
                free(quality_image.data);
                free(water_mask);
                close_l1g_data(&input_data);
                return EXIT_FAILURE;
            }  

            /* Write the water mask to an image file */ 
            if (ias_misc_write_envi_image(output_filename, 
                &input_image_projection, "Generate Chips Water Mask", 
                water_image.nl, water_image.ns, water_image.corners.upleft.x,
                water_image.corners.upleft.y, water_image.pixel_size_x,
                water_image.pixel_size_y, "Water Mask", IAS_BYTE, water_mask)
                != SUCCESS)
            {
                IAS_LOG_ERROR("Writing the water mask to file");
                free(elevation_image.data);
                free(quality_image.data);
                free(water_mask);
                close_l1g_data(&input_data);
                return EXIT_FAILURE;
            }
        }

        /* Create the full output file path */                     
        count = snprintf(output_filename, sizeof(output_filename), 
            "%s/exclusion_mask.img", parameters.intermediate_output_directory);
        if (count < 0 || count >= sizeof(output_filename))
        {
            IAS_LOG_ERROR("Bounds exceeded or internal snprintf error:"
                " count = %d", count);
            free(elevation_image.data);
            free(quality_image.data);
            free(exclusion_mask);
            free(water_mask);
            close_l1g_data(&input_data);
            return EXIT_FAILURE;
        }  

        /* Dump out the mask to an image files along with a header */
        if (ias_misc_write_envi_image(output_filename, &input_image_projection,
            "Generate Chips Exclusion Mask", quality_image.nl,
            quality_image.ns, quality_image.corners.upleft.x, 
            quality_image.corners.upleft.y, quality_image.pixel_size_x, 
            quality_image.pixel_size_y, "Exclusion Mask", IAS_BYTE, 
            exclusion_mask) != SUCCESS)
        {
            IAS_LOG_ERROR("Writing exclusion mask to file");
            free(elevation_image.data);
            free(quality_image.data);
            free(exclusion_mask);
            free(water_mask);
            close_l1g_data(&input_data);
            return EXIT_FAILURE;
        }
    }

    /* Free the water mask */
    free(water_mask);
    water_mask = NULL;

    /* Initialize the GCP structures */
    memset(&evenly_spaced_gcps, 0, sizeof(GENERATE_CHIPS_GCPS));
    memset(&moravec_gcps, 0, sizeof(GENERATE_CHIPS_GCPS));
    memset(&built_gcps, 0, sizeof(GENERATE_CHIPS_GCPS));
    memset(&extracted_gcps, 0, sizeof(GENERATE_CHIPS_GCPS));

    /* Call the correct generation method */
    switch (parameters.generation_method)
    {
        case GM_GCPLIB:
        {
            /* Generate chips from a GCPLib file */
            status = generate_gcp_lib_chips(&parameters, &input_data,
                &input_image_projection, &built_gcps);

            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Generating GCPs from GCPLib file");
                status = ERROR;
            }

            break;
        } /* END GM_GCPLIB */

        case GM_ENVI_ROI:
        {
            /* Generate chips from a GCPLib file */
            status = generate_envi_roi_chips(&parameters, &input_data,
                &input_image_projection, &geographic_projection, &built_gcps);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Generating GCPs from ENVI ROI file");
                status = ERROR;
            }

            break;
        } /* END GM_ENVI_ROI */

        case GM_MORAVEC:
        case GM_MORAVEC_AND_EVENLY_SPACED:
        {
            /* Generate chips using moravec */
            status = generate_moravec_chips(&parameters, &input_data,
                &quality_image, exclusion_mask, &moravec_gcps);

            if (status == ERROR)
            {
                IAS_LOG_ERROR("Generating moravec GCPs");
                break;
            }

            if (parameters.generation_method == GM_MORAVEC)
            {
                /* We only care about NO_POINTS if only performing moravec
                   chip generation */
                if (status == NO_POINTS)
                {
                    IAS_LOG_ERROR("No moravec GCPs were generated");
                    status = ERROR;
                    break;
                }

                built_gcps = moravec_gcps;

                /* Set the pointers to NULL, because they should not be used
                   after the above copy */
                moravec_gcps.gcp_locations = NULL;
                moravec_gcps.gcp_records = NULL;
                break;
            }
            else /* GM_MORAVEC_AND_EVENLY_SPACED */
            {
                /* If we have enough moravec then skip evenly */
                if (moravec_gcps.number_of_gcps
                    >= parameters.minimum_interest_operator_points)
                {
                    built_gcps = moravec_gcps;

                    /* Set the pointers to NULL, because they should not be
                       used after the above copy */
                    moravec_gcps.gcp_locations = NULL;
                    moravec_gcps.gcp_records = NULL;

                    break;
                }
            }
        } /* END GM_MORAVEC
             GM_MORAVEC_AND_EVENLY_SPACED falls through to GM_EVENLY_SPACED
             if not enough points were found */

        case GM_EVENLY_SPACED:
        {
            /* Generate chips using evenly spaced */
            status = generate_evenly_spaced_chips(&parameters, &input_data,
                &evenly_spaced_gcps);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Generating evenly spaced GCPs");
                status = ERROR;
                break;
            }

            if (parameters.generation_method == GM_MORAVEC_AND_EVENLY_SPACED)
            {
                /* Need to combine both moravec and evenly spaced GCPs into
                   one data structure with renamed chips */
                status = merge_gcps(GCP_VALID_FLAG_USE, &moravec_gcps,
                    &evenly_spaced_gcps, &built_gcps);

                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Merging moravec and evenly spaced GCPs");
                    status = ERROR;
                    break;
                }

                /* Free allocated memory in the moravec and evenly structures
                   because we have new memory in the built structure */
                free(moravec_gcps.gcp_locations);
                free(moravec_gcps.gcp_records);
                free(evenly_spaced_gcps.gcp_locations);
                free(evenly_spaced_gcps.gcp_records);

                /* Set the pointers to NULL, because they should not be used
                   after the above copy */
                moravec_gcps.gcp_locations = NULL;
                moravec_gcps.gcp_records = NULL;
            }
            else
            {
                built_gcps = evenly_spaced_gcps;
            }

            /* Set the pointers to NULL, because they should not be used after
               the above copy */
            evenly_spaced_gcps.gcp_locations = NULL;
            evenly_spaced_gcps.gcp_records = NULL;

            break;
        } /* END GM_EVENLY_SPACED and GM_MORAVEC_AND_EVENLY_SPACED */

        default:
        {
            /* Shouldn't get here due to other checks */
            IAS_LOG_ERROR("Invalid generation method");
            status = ERROR;
            break;
        }
    } /* END switch on the generation method */

    if (status != SUCCESS)
    {
        free(elevation_image.data);
        free(quality_image.data);
        close_l1g_data(&input_data);
        free(exclusion_mask);
        free(evenly_spaced_gcps.gcp_locations);
        free(evenly_spaced_gcps.gcp_records);
        free(moravec_gcps.gcp_locations);
        free(moravec_gcps.gcp_records);
        free(built_gcps.gcp_locations);
        free(built_gcps.gcp_records);
        return EXIT_FAILURE;
    }

    /* If we are not running the GCPLib method, then we need to initialize the
       GCP records so that extract chip has chip names and sizes to work with */
    if (status == SUCCESS && parameters.generation_method != GM_GCPLIB)
    {
        /* First we need to allocate the memory, because until now we have not
           needed to use the gcp_records */
        built_gcps.gcp_records =
            calloc(built_gcps.number_of_gcps, sizeof(IAS_GCP_RECORD));
        if (!built_gcps.gcp_records)
        {
            IAS_LOG_ERROR("Allocating memory for GCP records");
            status = ERROR;
        }
        else
        {
            int start_index = 0; /* Start index for evenly spaced GCPs */

            /* Loop through and initialize each GCP */
            for (index = 0; index < built_gcps.number_of_gcps; index++)
            {
                /* Initialize the GCP record */
                status = initialize_gcp_record(&parameters, index,
                    &built_gcps.gcp_records[index]);

                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Initializing GCP %d", index);
                    break;
                }
            }
            if (parameters.generation_method == GM_ENVI_ROI)
            {
                for (index = 0; index < built_gcps.number_of_gcps; index++)
                {
                    IAS_GCP_RECORD *cur_record = &built_gcps.gcp_records[index];
                    cur_record->moravec_rank = 0;
        
                    count = snprintf(cur_record->chip_select_method,
                        sizeof(cur_record->chip_select_method), "MANUAL"); 
                    if (count < 0 || count >=
                        sizeof(cur_record->chip_select_method))
                    {
                        IAS_LOG_ERROR("Bounds exceeded or internal snprintf "
                            "error: assigning envi roi select method: "
                            "count = %d", count);
                        status = ERROR;
                        break;
                    }
                }
            }

            if (parameters.generation_method == GM_MORAVEC_AND_EVENLY_SPACED ||
                parameters.generation_method == GM_MORAVEC)
            {
                start_index = moravec_gcps.number_of_gcps;
                for (index = 0; index < moravec_gcps.number_of_gcps; index++)
                {
                    IAS_GCP_RECORD *cur_record = &built_gcps.gcp_records[index];
                    cur_record->moravec_rank = index + 1;
        
                    count = snprintf(cur_record->chip_select_method,
                        sizeof(cur_record->chip_select_method), "MORAVEC"); 
                    if (count < 0 || count >=
                        sizeof(cur_record->chip_select_method))
                    {
                        IAS_LOG_ERROR("Bounds exceeded or internal snprintf "
                            "error: assigning moravec select method: "
                            "count = %d", count);
                        status = ERROR;
                        break;
                    }
                }
            }

            if (parameters.generation_method == GM_MORAVEC_AND_EVENLY_SPACED ||
                parameters.generation_method == GM_EVENLY_SPACED)
            {
                for (index = 0; 
                     index < evenly_spaced_gcps.number_of_gcps; index++)
                {
                    IAS_GCP_RECORD *cur_record = 
                        &built_gcps.gcp_records[index + start_index];
                    cur_record->moravec_rank = 0;
        
                    count = snprintf(cur_record->chip_select_method,
                        sizeof(cur_record->chip_select_method), "GRID");
                    if (count < 0 || count >=
                        sizeof(cur_record->chip_select_method))
                    {
                        IAS_LOG_ERROR("Bounds exceeded or internal snprintf "
                            "error: assigning evenly spaced select method: "
                            "count = %d", count);
                        status = ERROR;
                        break;
                    }
                }
            }
        }
    }

    input_image2geographic = ias_geo_create_proj_transformation(
        &input_image_projection, &geographic_projection);
    if (!input_image2geographic)
    {
        IAS_LOG_ERROR("Creating projection transformation");
        status = ERROR;
    }

    /* Extract the GCPs */
    if (status == SUCCESS)
    {
        IAS_LOG_DEBUG("Number of GCPs before extraction = %d",
            built_gcps.number_of_gcps);

        status = extract_gcp_chips(&parameters, &input_data,
            exclusion_mask, input_image2geographic, quality_image.nl, 
            quality_image.ns, &built_gcps);
    }
    
    /* Free the exclusion mask it is no longer needed */
    free(exclusion_mask);
    exclusion_mask = NULL;
   
    /* Copy all the valid input GCPs to the output GCP records */
    if (status == SUCCESS)
    {
        status = copy_gcps(GCP_VALID_FLAG_USE, &built_gcps, &extracted_gcps);

        IAS_LOG_DEBUG("Number of GCPs after extraction = %d",
            extracted_gcps.number_of_gcps);
    }

    /*** Only use extracted_gcps after here ***/

    /* Update the date elements */
    extracted_gcps.year = capture_year;
    extracted_gcps.month = capture_month;
    extracted_gcps.day = capture_day;

    /* Update all the output GCP records */
    if (status == SUCCESS)
    {
        status = update_valid_gcp_records(&parameters, &input_data,
            input_image2geographic, &elevation_image, &extracted_gcps);
    }

    /* Write/Create the output GCPLib file */
    if (status == SUCCESS)
    {
        count = snprintf(output_gcplib_path, sizeof(output_gcplib_path),
            "%s/%s", parameters.output_directory, parameters.output_gcplib);

        if (count < 0 || count >= sizeof(output_gcplib_path))
        {
            IAS_LOG_ERROR("Bounds exceeded or internal snprintf error:"
                " creating output GCPLib filename and path: count = %d", count);
            status = ERROR;
        }
        else
        {
            int overall_gcp_version = 0;

            status = ias_gcp_write_gcplib(output_gcplib_path,
                extracted_gcps.gcp_records, extracted_gcps.number_of_gcps,
                overall_gcp_version);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Creating output GCPLib file");
                status = ERROR;
            }
        }
        
        if (parameters.envi_point_filename_present && status == SUCCESS)
        {
            status = dump_envi_point_file(&parameters,
                parameters.envi_point_filename, &extracted_gcps);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Creating ENVI point file");
                status = ERROR;
            }
        }
    }

    /* Free allocated memory */
    free(built_gcps.gcp_locations);
    free(built_gcps.gcp_records);
    free(extracted_gcps.gcp_locations);
    free(extracted_gcps.gcp_records);

    /* Free the projection transformation */
    ias_geo_destroy_proj_transformation(input_image2geographic);

    /* Close the input data */
    free(elevation_image.data);
    free(quality_image.data);
    close_l1g_data(&input_data);

    /* Notify the user that the application is terminating */
    if (status == SUCCESS)
    {
        IAS_LOG_INFO("Successfully terminated");
        return EXIT_SUCCESS;
    }
    else
    {
        IAS_LOG_INFO("Failed during processing");
        return EXIT_FAILURE;
    }
}

