
#ifndef GENERATE_CHIPS_H
#define GENERATE_CHIPS_H

/* IAS Includes */
#include "ias_l1g.h"
#include "ias_gcp.h"
#include "ias_geo.h"

/*****************************************************************************
 The following are the defines used by generate chips
*****************************************************************************/

/* An additional return value from generate_moravec_chips */
#define NO_POINTS 1

/* Define some values to use when dealing with the exclusion mask */
#define EXCLUSION_MASK_VALID        0x00
#define EXCLUSION_MASK_FILL         0x01
#define EXCLUSION_MASK_CIRRUS       0x02
#define EXCLUSION_MASK_CLOUD        0x04
#define EXCLUSION_MASK_SNOW_ICE     0x08
#define EXCLUSION_MASK_WATER        0x10

/* Define water mask values */
#define USE_POLYGON_WATER_MASK 1 
#define USE_BINARY_WATER_MASK  2 

/*****************************************************************************
 The following are the structures and enumerations used by generate chips
*****************************************************************************/

/* Define the chip generation methods */
typedef enum generation_method
{
    GM_UNKNOWN = 0,
    GM_EVENLY_SPACED,
    GM_GCPLIB,
    GM_MORAVEC,
    GM_MORAVEC_AND_EVENLY_SPACED,
    GM_ENVI_ROI
} GENERATION_METHOD;

/* Define the valid flag indicators */
typedef enum gcp_valid_flag_type
{
   GCP_VALID_FLAG_DONT_USE = 0,
   GCP_VALID_FLAG_USE,
   GCP_VALID_FLAG_MORAVEC_DONT_USE,
   GCP_VALID_FLAG_MORAVEC_GOOD,
   GCP_VALID_FLAG_MORAVEC_GOOD_DONT_USE
} GCP_VALID_FLAG_TYPE;

/* Define the moravec source resolutions */
typedef enum moravec_source_resolution
{ 
    MSR_UPSAMPLED = 0,
    MSR_NOMINAL = 1,
    MSR_DOWNSAMPLED = 2,
    MSR_FEWEST_POINTS = 3,
    MSR_UNKNOWN = -1
} MORAVEC_SOURCE_RESOLUTION;

/* Specify the number of resolutions that will be used for the moravec method
   This must match the resolutions array created in generate_moravec_chips.c */
#define NUMBER_OF_RESOLUTIONS 3
/* Define a structure that can be used to specify a sampling and its sampling
   factor */
typedef struct resolution_sampling
{
    char output_filename[PATH_MAX]; /* Used when intermediate output is
                                       requested */
    int kernel_size;       /* The size of the resampling kernel to use
                              A kernel size of 0 indicates...
                                  "do not resample, make a copy" */
    double scaling_factor; /* The associated scaling to be performed */
} RESOLUTION_SAMPLING;

/* Create a type that holds information about the point locations within the
   image data.  This also maintains whether or not the point is valid. */
typedef struct gcp_point_location
{
    double resolution_line;   /* The line within the resolution for the point */
    double resolution_sample; /* The sample within the resolution for the
                                 point */
    double input_data_line;   /* The line within the image for the point */
    double input_data_sample; /* The sample within the image for the point */
    double moravec_value;     /* Moravec interest operator value at line, sample
                                 location */
    GCP_VALID_FLAG_TYPE valid_flag; /* Identifies if the point is valid */
} GENERATE_CHIPS_GCP_LOCATION;

/* Create a type that holds all the information related to the GCPs */
typedef struct genchips_gcps
{
    GENERATE_CHIPS_GCP_LOCATION *gcp_locations; /* Holds the locations
        within the input image or resolution for the GCPs. */
    IAS_GCP_RECORD *gcp_records; /* Holds all the GCP records. */
    int number_of_gcps;          /* The number of GCPs */
    int year;                    /* The year of the input image */
    int month;                   /* The month of the input image */
    int day;                     /* The day of the input image */
    double scaling_factor; /* The scaling factor needed to convert the
                              resampled line/sample locations back to the
                              input data resolution */
} GENERATE_CHIPS_GCPS;

/* Create a type to hold information about the input L1G file */
typedef struct genchips_l1g_data
{
    L1GIO *io_ptr; /* Pointer to the L1G object */
    IAS_L1G_FILE_METADATA metadata; /* Holds the metadata information for the
                                       L1G data */
    L1G_BAND_IO *band_io_ptr;            /* Pointer to the requested band */
    IAS_L1G_BAND_METADATA band_metadata; /* Holds the metadata for the
                                            specified band */
    IAS_DATA_TYPE data_type;     /* The data type of the band information */
    int number_of_scas;          /* The number of SCAs in the band */
    int number_of_image_lines;   /* The number of lines in the band */
    int number_of_image_samples; /* The number of samples in the band */
} GENERATE_CHIPS_L1G_DATA;

/* Create a type to hold all the possible input parameters */
typedef struct generate_chips_parameters
{
    /* COMMON Parameters */
    GENERATION_METHOD generation_method; /* EVENLY_SPACED, GCPLIB, MORAVEC */
    char gcp_source[IAS_GCP_SOURCE_SIZE];/* GLS, DOQ, TM6 or GRI */
    char l1g_filename[PATH_MAX];         /* Filename of the input L1G */
    char quality_filename[PATH_MAX];     /* Filename of the quality band */
    char water_mask_filename[PATH_MAX];  /* Filename of the water mask */
    int use_water_mask_flag;             /* Flag to use water mask */
    char output_gcplib[PATH_MAX];        /* Filename of the GCPLib file */
    char output_directory[PATH_MAX];     /* Full path to the output directory */
    char absolute_or_relative[IAS_GCP_ABS_REL_SIZE];/* Type of GCPs to create */
    int band_number;              /* Band to use for generating the GCPs */
    double min_range_value;       /* Min pixel value allowed */
    double max_range_value;       /* Max pixel value allowed */
    double valid_pixel_threshold; /* Tolerance for the valid pixel range */
    double valid_water_threshold; /* Tolerance for how much water to allow */
    double valid_snow_and_ice_threshold; /* Tolerance for how much snow and
                                            ice to allow */
    char envi_point_filename[PATH_MAX]; /* Filename for the envi points file */
    int envi_point_filename_present;    /* Is it present or not */
    int starting_id;             /* Defines starting number for chip index */
    char gcp_update_reason[IAS_GCP_UPDATE_REASON_SIZE]; /* Reason for latest 
                                                       update to GCPs */

    /* EVENLY SPACED Parameters */
    int gcp_grid_density_rows;    /* How many rows of GCP points to generate */
    int gcp_grid_density_columns; /* How many columns of GCP points to
                                     generate */
    int target_wrs_path; /* Target WRS-2 path the scene was collected on */
    int target_wrs_row;  /* Target WRS-2 row the scene was collected on */
    char elevation_filename[PATH_MAX]; /* Filename of the elevation data */
    int elevation_filename_present;    /* Is it present or not */

    /* EXISTING GCPLIB Parameters */
    char input_gcplib[PATH_MAX]; /* Filename of the input GCPLib file */

    /* ENVI ROI Parameters */
    char envi_roi_filename[PATH_MAX]; /* Filename of the input roi file */

    /* MORAVEC Parameters */
    int interest_operator_size; /* The size of the interest operator */
    /* The number of sub-divisions/regions for point distribution */
    int number_of_regions_lines;
    int number_of_regions_samples;
    int cloud_border_lines;   /* Cloud border growth lines */
    int cloud_border_samples; /* Cloud border growth samples */
    int exclude_cirrus;       /* Level of cirrus exclusion */
    int exclude_cloud;        /* Level of cloud exclusion */
    MORAVEC_SOURCE_RESOLUTION moravec_source_resolution; /* Source Resolution */
    int exclude_snow_and_ice; /* Level of snow and ice exclusion */
    int minimum_number_of_points_in_region; /* The minimum number of points to
                                               find within a region */
    int minimum_interest_operator_points;   /* The minimum number of points to
                                               find in the image */
    int local_maximum_distance; /* The distance from other maximum values */
    double minimum_interest_operator_threshold; /* The minimum interest
                                                   operator value to accept
                                                   for further processing */
    int minimum_distance_lines; /* Minimum line distance between GCPs */
    int minimum_distance_samps; /* Minimum sample distance between GCPs */
    double background;                  /* Value of the L1G background  */
    int points_before_region_filtering; /* Number of best points kept before
                                           starting to filter by region */
    /* EVENLY SPACED and MORAVEC Parameters */
    int chip_size_lines;   /* GCP chip size in lines */
    int chip_size_samples; /* GCP chip size in samples */

    /* DEBUGGING Parameters */
    int output_intermediate_data; /* Dump all internally generated data */
    char intermediate_output_directory[PATH_MAX]; /* Full path to the
                                                     intermediate output
                                                     directory */
} GENERATE_CHIPS_PARAMETERS;

/*****************************************************************************
 The following are the routines used by generate chips
*****************************************************************************/
int read_parameters
(
    const char *parameter_filename,         /* I: Parameter file name */
    GENERATE_CHIPS_PARAMETERS *parameters   /* O: Generate Chips parameters */
);

int open_l1g_data
(
    const char *filename,
    int band_number,
    GENERATE_CHIPS_L1G_DATA *l1g_data
);

void close_l1g_data
(
    const GENERATE_CHIPS_L1G_DATA *l1g_data
);

int extract_gcp_chips
(
    const GENERATE_CHIPS_PARAMETERS *parameters,
    const GENERATE_CHIPS_L1G_DATA *l1g_data,
    IAS_BYTE_TYPE *exclusion_mask,
    const IAS_GEO_PROJ_TRANSFORMATION *input_image2geographic,
    int number_mask_lines,
    int number_mask_samples,
    GENERATE_CHIPS_GCPS *gcp_points
);

int copy_gcps
(
    GCP_VALID_FLAG_TYPE flag_value,
    const GENERATE_CHIPS_GCPS *source_gcps,
    GENERATE_CHIPS_GCPS *destination_gcps
);

int copy_gcp_locations
(
    GCP_VALID_FLAG_TYPE flag_value,
    const GENERATE_CHIPS_GCPS *source_gcps,
    GENERATE_CHIPS_GCPS *destination_gcps
);

int merge_gcps
(
    GCP_VALID_FLAG_TYPE flag_value,
    const GENERATE_CHIPS_GCPS *source_1_gcps,
    const GENERATE_CHIPS_GCPS *source_2_gcps,
    GENERATE_CHIPS_GCPS *destination_gcps
);

int update_valid_gcp_records
(
    const GENERATE_CHIPS_PARAMETERS *parameters,
    const GENERATE_CHIPS_L1G_DATA *l1g_data,
    const IAS_GEO_PROJ_TRANSFORMATION *input_image2geographic,
    const IAS_IMAGE *elevation_image,
    GENERATE_CHIPS_GCPS *gcps
);

int initialize_gcp_record
(
    const GENERATE_CHIPS_PARAMETERS *parameters,
    int point_id,
    IAS_GCP_RECORD *gcp_records
);

int generate_gcp_lib_chips
(
    const GENERATE_CHIPS_PARAMETERS *parameters,
    const GENERATE_CHIPS_L1G_DATA *l1g_data,
    const IAS_PROJECTION *input_image_projection,
    GENERATE_CHIPS_GCPS *gcps
);

int generate_envi_roi_chips
(
    const GENERATE_CHIPS_PARAMETERS *parameters, /* I: The input parameters */
    const GENERATE_CHIPS_L1G_DATA *l1g_data,     /* I: Input L1G data */
    const IAS_PROJECTION *input_image_projection,/* I: Input data projection */
    const IAS_PROJECTION *geographic_projection, /* I: Input data projection */
    GENERATE_CHIPS_GCPS *gcps /* O: Holds the input and output GCP records */

);

int generate_evenly_spaced_chips
(
    const GENERATE_CHIPS_PARAMETERS *parameters,
    const GENERATE_CHIPS_L1G_DATA *l1g_data,
    GENERATE_CHIPS_GCPS *gcps
);

int generate_moravec_chips
(
    const GENERATE_CHIPS_PARAMETERS *parameters,
    const GENERATE_CHIPS_L1G_DATA *l1g_data,
    const IAS_IMAGE *quality_image,
    IAS_BYTE_TYPE *exclusion_mask,
    GENERATE_CHIPS_GCPS *gcps
);

int convert_to_data_type
(
    int data_count,
    IAS_DATA_TYPE source_data_type,
    void *source_data,
    IAS_DATA_TYPE destination_data_type,
    void *destination_data
);

int build_exclusion_mask
(
    const GENERATE_CHIPS_PARAMETERS *parameters,
    const IAS_IMAGE *quality_image,
    IAS_BYTE_TYPE *water_mask,
    IAS_BYTE_TYPE *exclusion_mask
);

int perform_interest_operator
(
    const GENERATE_CHIPS_PARAMETERS *parameters,
    IAS_R4_TYPE *input_data,
    int number_of_lines,
    int number_of_samples,
    IAS_BYTE_TYPE *exclusion_mask,
    int number_of_mask_lines,
    int number_of_mask_samples,
    GENERATE_CHIPS_GCPS *gcps
);

void moravec_resolution_check
(
    RESOLUTION_SAMPLING *resolutions,
    int picked_resolution,
    GENERATE_CHIPS_GCPS gcps[NUMBER_OF_RESOLUTIONS]
);

void moravec_window_points
(
    const GENERATE_CHIPS_PARAMETERS *parameters,
    GENERATE_CHIPS_GCPS *gcps
);

void moravec_limit_points
(
    const GENERATE_CHIPS_PARAMETERS *parameters,
    GENERATE_CHIPS_GCPS *gcps
);

void moravec_check_regions
(
    const GENERATE_CHIPS_PARAMETERS *parameters,
    const GENERATE_CHIPS_L1G_DATA *l1g_data,
    GENERATE_CHIPS_GCPS *gcps
);

int allocate_gcp_memory(
    int number_of_elements,
    GENERATE_CHIPS_GCPS *destination_gcps
);

void copy_gcps_to_destination_index
(
    GCP_VALID_FLAG_TYPE flag_value,
    const GENERATE_CHIPS_GCPS *source_gcps,
    int *start_index,
    GENERATE_CHIPS_GCPS *destination_gcps
);

int dump_envi_point_file
(
    const GENERATE_CHIPS_PARAMETERS *parameters,
    const char *filename,
    const GENERATE_CHIPS_GCPS *gcps
);

int dump_intermediate_envi_point_file
(
    const GENERATE_CHIPS_PARAMETERS *parameters,
    const char *filename,
    const GENERATE_CHIPS_GCPS *gcps
);

int create_fake_quality_band 
(
    IAS_IMAGE *image_ptr,   
    GENERATE_CHIPS_L1G_DATA *l1g_image,
    double fill_pixel
);

int initialize_directory
(
    const char *directory      /* I: Directory to create/check */
);

#endif /* GENERATE_CHIPS_H */

