/* Standard Library Includes */
#include <stdlib.h>
#include <string.h>

/* IAS Library Includes */
#include "ias_logging.h"
#include "ias_geo.h"

/* Local Includes */
#include "generate_chips.h"

/*****************************************************************************
 NAME: read_envi_roi_file

 PURPOSE: Reads the latitude and longitude from an ENVI ROI file for each point
          present in the file.

 RETURN VALUE: Type = int
    Value    Description
    -------  --------------------------
    SUCCESS  Successful completion
    ERROR    Operation failed
****************************************************************************/
static int read_envi_roi_file
(
    const char *envi_roi_filename,         /* I: ENVI ROI input filename */
    IAS_DBL_LAT_LONG **output_coordinates, /* O: Latitude and Longitude */
    int *num_points                        /* O: Number of ROI points */
)
{
    FILE *envi_roi_file_ptr;    /* Pointer to the input file */
    char parse_format[200];     /* Format used to parse the record lines */
    char *line_ptr;             /* Pointer to line of file from fgets */
    char buffer[120];           /* String read in from file */
    int status;                 /* Routine return variable */
    int count_in_file;          /* Count of chips expected in the file */
    int index;                  /* Generic loop index */
    IAS_DBL_LAT_LONG *coordinates = NULL; /* Local coordinates variable */

    /* Initialize the return values */
    *num_points = 0;
    *output_coordinates = NULL;

    /* Create the format string for parsing the records. 
       Only the latitude and longitude columns are being parsed 
       using this format remove the '*' to parse additional columns */
    status = snprintf(parse_format, sizeof(parse_format),
       "%%*d %%*d %%*d %%*lf %%*lf %%lf %%lf %%*d\n");
    if (status < 0 || status >= sizeof(parse_format))
    {
        IAS_LOG_ERROR("Creating format string to parse ENVI ROI points");
        return ERROR;
    }

    /* Open the input file */
    envi_roi_file_ptr = fopen(envi_roi_filename, "r");
    if (!envi_roi_file_ptr)
    {
        IAS_LOG_ERROR("Opening ENVI ROI file %s", envi_roi_filename);
        return ERROR;
    }

    /* Read the file until "ROI npts" flag to bypass header */
    for (;;)
    {
        line_ptr = fgets(buffer, sizeof(buffer), envi_roi_file_ptr);
        if (!line_ptr) 
        {
            IAS_LOG_ERROR("Unexpected end of ENVI ROI file %s", 
                envi_roi_filename);
            fclose(envi_roi_file_ptr);
            return ERROR;
        }

        /* Find the ROI npts data marker */
        if (strncmp(buffer, "; ROI npts:", 11) == 0)
        {
            /* Get the count of GCPs expected in the file */
            if (sscanf(buffer, "; ROI npts: %d", &count_in_file) != 1)
            {
                IAS_LOG_ERROR("No 'ROI npts:' record found in %s",
                    envi_roi_filename);
                fclose(envi_roi_file_ptr);
                return ERROR;
            }
            
            /* Check that the point count is greater than zero */
            if (count_in_file <= 0)
            {
                IAS_LOG_ERROR("The 'ROI npts:' record must be greater than 0");
                fclose(envi_roi_file_ptr);
                return ERROR;
            }

            /* Read the next line to bypass the envi header and set the line
               pointer at the first line of data */
            line_ptr = fgets(buffer, sizeof(buffer), envi_roi_file_ptr);
            if (!line_ptr)
            {
                IAS_LOG_ERROR("Unexpected end of ENVI ROI file %s", 
                    envi_roi_filename);
                fclose(envi_roi_file_ptr);
                return ERROR;
            }
            break;
        }
    }
    *num_points = count_in_file;

    /* Allocate memory for the number of coordinates */
    coordinates = (IAS_DBL_LAT_LONG *) malloc(count_in_file 
        * sizeof(IAS_DBL_LAT_LONG));
    if (!coordinates)
    {
        IAS_LOG_ERROR("Allocating memory for coordinates");
        fclose(envi_roi_file_ptr);
        return ERROR;
    }

    /* Get the chip information from file */
    for (index = 0; index < count_in_file; index++)
    {
        /* Get the current line */
        line_ptr = fgets(buffer, sizeof(buffer), envi_roi_file_ptr);
        if (!line_ptr) 
        {
            IAS_LOG_ERROR("Unexpected end of ENVI ROI file %s", 
                envi_roi_filename);
            fclose(envi_roi_file_ptr);
            free(coordinates);
            return ERROR;
        }

        /* Read in the coordinates */
        status = sscanf(line_ptr, parse_format, &coordinates[index].lat,
            &coordinates[index].lng);
        if (status == EOF) /* (unexpected) end of file or matching failure */
        {
            IAS_LOG_ERROR("Reading latitude and longitude values in line: %s",
                line_ptr);
            fclose(envi_roi_file_ptr);
            free(coordinates);
            return ERROR;
        }
        
        /* Confirm matched record count */
        if (status != 2)
        {
            IAS_LOG_ERROR("Reading ENVI ROI file %s: too few values",
                envi_roi_filename);
            fclose(envi_roi_file_ptr);
            free(coordinates);
            return ERROR;
        }
    }

    /* Close the ROI input file */
    fclose(envi_roi_file_ptr);

    /* Set the output coordinates pointer */
    *output_coordinates = coordinates;

    return SUCCESS;
}

/*****************************************************************************
 NAME: generate_envi_roi_chips

 PURPOSE: Generates line and sample locations of the GCPs from the input 
          ENVI ROI file latitude and longitude columns.

 RETURN VALUE: Type = int
    Value    Description
    -------  --------------------------
    SUCCESS  Successful completion
    ERROR    Operation failed

 NOTE: Memory is allocated for and returned in the GCPs

*****************************************************************************/
int generate_envi_roi_chips
(
    const GENERATE_CHIPS_PARAMETERS *parameters, /* I: The input parameters */
    const GENERATE_CHIPS_L1G_DATA *l1g_data,     /* I: Input L1G data */
    const IAS_PROJECTION *input_image_projection,/* I: Input data projection */
    const IAS_PROJECTION *geographic_projection, /* I: Input data projection */
    GENERATE_CHIPS_GCPS *gcps /* O: Holds the input and output GCP records */
)
{
    int index;                      /* Generic loop variable */
    IAS_DBL_LAT_LONG *coordinates;  /* Coordinates array */
    IAS_GEO_PROJ_TRANSFORMATION *geographic_transformation = NULL;

    /* Read the entire specified GCPLib file */
    if (read_envi_roi_file(parameters->envi_roi_filename,
            &coordinates, &gcps->number_of_gcps) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading ROI file %s", parameters->envi_roi_filename);
        return ERROR;
    }

    /* Only continue if we have some GCPs */
    if (gcps->number_of_gcps <= 0)
    {
        IAS_LOG_ERROR("No GCP records found");
        free(coordinates);
        return ERROR;
    }

    IAS_LOG_INFO("Number of Input GCPs = %d", gcps->number_of_gcps);

    /* Create the transfomation */
    geographic_transformation = ias_geo_create_proj_transformation( 
        geographic_projection, input_image_projection);
    if (!geographic_transformation)
    {
        IAS_LOG_ERROR("Creating projection transformation");
        free(coordinates);
        return ERROR;
    }

    /* Allocate memory for the locations */
    gcps->gcp_locations = calloc(gcps->number_of_gcps,
        sizeof(GENERATE_CHIPS_GCP_LOCATION));
    if (!gcps->gcp_locations)
    {
        IAS_LOG_ERROR("Allocating memory for GCP locations");
        ias_geo_destroy_proj_transformation(geographic_transformation); 
        free(coordinates);
        return ERROR;
    }

    /* Process through the list of GCPs */
    for (index = 0; index < gcps->number_of_gcps; index++)
    {      
        double map_coordinate_x; /* Map projection x coordinate */
        double map_coordinate_y; /* Map projection y coordinate */

        /* Convert the lat/long into map projection x/y */
        if (ias_geo_transform_coordinate(geographic_transformation,
            coordinates[index].lng, coordinates[index].lat,
            &map_coordinate_x, &map_coordinate_y) != SUCCESS)
        {
            IAS_LOG_ERROR("Transforming latitude/longitude coordinates %lf/%lf "
                "to map projection coordinates", coordinates[index].lat,
                coordinates[index].lng);
            free(coordinates);
            ias_geo_destroy_proj_transformation(geographic_transformation); 
            free(gcps->gcp_locations);
            gcps->gcp_locations = NULL;
            return ERROR; 
        } 

        /* Figure out the line and sample locations for the GCP from the map
           coordinates */
        gcps->gcp_locations[index].input_data_line =
            (l1g_data->band_metadata.upper_left_y - map_coordinate_y)
                / l1g_data->band_metadata.projection_distance_y;

        gcps->gcp_locations[index].input_data_sample = 
            (map_coordinate_x - l1g_data->band_metadata.upper_left_x)
            / l1g_data->band_metadata.projection_distance_x;

        gcps->gcp_locations[index].valid_flag = GCP_VALID_FLAG_USE;
    }

    /* Destroy the transformation */
    ias_geo_destroy_proj_transformation(geographic_transformation); 

    /* Free the coordinates */
    free(coordinates);

    return SUCCESS;
}

