
/* Standard C Includes */
#include <stdlib.h>
#include <string.h>

/* IAS Includes */
#include "gctp.h"
#include "ias_logging.h"
#include "ias_gcp.h"
#include "ias_geo.h"
#include "ias_miscellaneous.h"

/* Local Includes */
#include "generate_chips.h"

/*****************************************************************************
 NAME: generate_evenly_spaced_chips

 PURPOSE: To generate GCPs based on an even distribution.

 RETURN VALUE: Type = int
    Value    Description
    -------  --------------------------
    SUCCESS  Successful completion
    ERROR    Operation failed

 NOTE: Memory is allocated for and returned in the GCPs

*****************************************************************************/
int generate_evenly_spaced_chips
(
    const GENERATE_CHIPS_PARAMETERS *parameters, /* I: The input parameters */
    const GENERATE_CHIPS_L1G_DATA *l1g_data,     /* I: Input L1G data */
    GENERATE_CHIPS_GCPS *gcps /* O: Holds the input and output GCP records */
)
{
    int point_index;   /* Loop control */
    int row_index;     /* Loop control */
    int column_index;  /* Loop control */
    int delta_row;     /* Point offset in the line direction */
    int delta_column;  /* Point offset in the sample direction */
    int line_offset;   /* Offset into the image for the starting line */
    int sample_offset; /* Offset into the image for the starting sample */

    /* Specify how many GCPs to generate */
    gcps->number_of_gcps = parameters->gcp_grid_density_rows
        * parameters->gcp_grid_density_columns;

    /* Allocate memory */
    gcps->gcp_locations =
        calloc(gcps->number_of_gcps, sizeof(GENERATE_CHIPS_GCP_LOCATION));

    /* Verify the memory allocation */
    if (gcps->gcp_locations == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for GCP locations");
        return ERROR;
    }

    /* Generate the point offsets */
    /* The value 4 is just an additional buffer from the edge of the image 2
       samples on both sides */
    delta_row =
        (l1g_data->number_of_image_lines - parameters->chip_size_lines - 4)
        / ((double)(parameters->gcp_grid_density_rows));
    delta_column =
        (l1g_data->number_of_image_samples - parameters->chip_size_samples - 4)
        / ((double)(parameters->gcp_grid_density_columns));

    line_offset = (parameters->chip_size_lines / 2.0) + 2;
    sample_offset = (parameters->chip_size_samples / 2.0) + 2;

    /* Generate the point locations */
    point_index = 0;
    for (row_index = 0;
         row_index < parameters->gcp_grid_density_rows;
         row_index++)
    {
        for (column_index = 0;
             column_index < parameters->gcp_grid_density_columns;
             column_index++)
        {
            /* Calculate the line and sample locations within the input image */
            gcps->gcp_locations[point_index].input_data_line =
                delta_row * row_index + line_offset;
            gcps->gcp_locations[point_index].input_data_sample =
                delta_column * column_index + sample_offset;

            /* The chip is initialized to be good here and later in the
               processing it is validated */
            gcps->gcp_locations[point_index].valid_flag = GCP_VALID_FLAG_USE;

            point_index++;
        } /* END sample index */

    } /* END line index */

    return SUCCESS;
}

