
/* Standard C Includes */
#include <stdlib.h>
#include <string.h>

/* IAS Includes */
#include "gctp.h"
#include "ias_logging.h"
#include "ias_gcp.h"
#include "ias_geo.h"
#include "ias_miscellaneous.h"

/* Local Includes */
#include "generate_chips.h"

/*****************************************************************************
 NAME: generate_gcp_lib_chips

 PURPOSE: To generate GCPs based on an input GCPLib file.

 RETURN VALUE: Type = int
    Value    Description
    -------  --------------------------
    SUCCESS  Successful completion
    ERROR    Operation failed

 NOTE: Memory is allocated for and returned in the GCPs

*****************************************************************************/
int generate_gcp_lib_chips
(
    const GENERATE_CHIPS_PARAMETERS *parameters, /* I: The input parameters */
    const GENERATE_CHIPS_L1G_DATA *l1g_data,     /* I: Input L1G data */
    const IAS_PROJECTION *input_image_projection,/* I: Input data projection */
    GENERATE_CHIPS_GCPS *gcps /* O: Holds the input and output GCP records */
)
{
    int index;               /* Loop control */
    int status;              /* Capture function/routine statuses */
    int overall_gcp_version; /* Overall GCP version read from file */

    /*** Projection Variables ***/
    IAS_PROJECTION input_gcplib_map_projection;
    /* Geographic and UTM parameters are the same */
    double utm_projection_parameters[IAS_PROJ_PARAM_SIZE];
    /* Northern and Southern Polar Stereographic projection parameters */
    double northern_ps_projection_parameters[IAS_PROJ_PARAM_SIZE];
    double southern_ps_projection_parameters[IAS_PROJ_PARAM_SIZE];
    IAS_GEO_PROJ_TRANSFORMATION *input_gcplib2input_image = NULL;
    int perform_point_transform;
    double map_coordinate_x;
    double map_coordinate_y;

    /* Create the projection information */
    for (index = 0; index < IAS_PROJ_PARAM_SIZE; index++)
    {
        utm_projection_parameters[index] = 0.0;
        northern_ps_projection_parameters[index] = 0.0;
        southern_ps_projection_parameters[index] = 0.0;
    }

    /* The following values are taken from get_parms_and_initialize_grid.c in
       create_grid and are used to define the Northern and Southern
       Polar-Stereographic projection information. */
    northern_ps_projection_parameters[0] = 6378137.0;
    northern_ps_projection_parameters[1] = 6356752.314245;
    northern_ps_projection_parameters[5] = 71000000.0;

    southern_ps_projection_parameters[0] = 6378137.0;
    southern_ps_projection_parameters[1] = 6356752.314245;
    southern_ps_projection_parameters[5] = -71000000.0;

    /* Read the entire specified GCPLib file */
    status = ias_gcp_read_gcplib(parameters->input_gcplib,
            &gcps->gcp_records, &gcps->number_of_gcps, &overall_gcp_version);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading input GCPLib file %s", parameters->input_gcplib);
        return ERROR;
    }

    IAS_LOG_INFO("Number of Input GCPs = %d", gcps->number_of_gcps);

    /* Only continue if we have some GCPs */
    if (gcps->number_of_gcps == 0)
    {
        IAS_LOG_ERROR("No GCP records found");
        free(gcps->gcp_records);
        gcps->gcp_records = NULL;
        return ERROR;
    }

    /* Allocate memory for the locations */
    gcps->gcp_locations =
        calloc(gcps->number_of_gcps, sizeof(GENERATE_CHIPS_GCP_LOCATION));

    if (gcps->gcp_locations == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for GCP locations");
        free(gcps->gcp_records);
        gcps->gcp_records = NULL;
        return ERROR;
    }

    /* Process through the list of GCPs */
    for (index = 0; index < gcps->number_of_gcps; index++)
    {
        map_coordinate_x = gcps->gcp_records[index].projection_x;
        map_coordinate_y = gcps->gcp_records[index].projection_y;

        if (strcmp(gcps->gcp_records[index].projection, "UTM") == 0)
        {
            /* Compare zones with the image and reproject the point to
               the image projection if needed */
            if (gcps->gcp_records[index].zone != l1g_data->metadata.zone_code)
            {
                /* Reproject the map coordinates to the image zone */

                /* Set the projection for the point */
                ias_geo_set_projection(UTM, gcps->gcp_records[index].zone,
                    METER, WGS84_SPHEROID, utm_projection_parameters,
                    &input_gcplib_map_projection);

                /* Set the transformation for the point */
                input_gcplib2input_image =
                    ias_geo_create_proj_transformation(
                        &input_gcplib_map_projection,
                        input_image_projection);

                /* Perform the transformation */
                ias_geo_transform_coordinate(input_gcplib2input_image,
                    gcps->gcp_records[index].projection_x,
                    gcps->gcp_records[index].projection_y,
                    &map_coordinate_x, &map_coordinate_y);

                ias_geo_destroy_proj_transformation(input_gcplib2input_image);
            }
            /* Else we are already using the matching map coordinates */
        }
        else if (strcmp(gcps->gcp_records[index].projection, "PS") == 0)
        {
            /* Reproject the map coordinates to the image projection */

            perform_point_transform = 0;
            /* Check and see if we are already in the same projection or
               not */
            if (PS != l1g_data->metadata.projection_code)
            {
                perform_point_transform = 1;
            }

            /* Check to see if they are both Northern or both Southern */
            if (perform_point_transform == 0
                && (
                     (gcps->gcp_records[index].latitude < 0
                       &&
                      l1g_data->metadata.projection_parameters[5] > 0)
                     ||
                     (gcps->gcp_records[index].latitude > 0
                       &&
                      l1g_data->metadata.projection_parameters[5] < 0)
                   )
               )
            {
                perform_point_transform = 1;
            }

            /* If we didn't find matching projections transform the point */
            if (perform_point_transform == 1)
            {
                if (gcps->gcp_records[index].latitude < 0)
                {
                    /* Set the projection for a southern point */
                    ias_geo_set_projection(PS, 0, METER,
                        WGS84_SPHEROID, southern_ps_projection_parameters,
                        &input_gcplib_map_projection);
                }
                else
                {
                    /* Set the projection for a northern point */
                    ias_geo_set_projection(PS, 0, METER,
                        WGS84_SPHEROID, northern_ps_projection_parameters,
                        &input_gcplib_map_projection);
                }

                /* Set the transformation for the point */
                input_gcplib2input_image =
                    ias_geo_create_proj_transformation(
                        &input_gcplib_map_projection,
                        input_image_projection);

                /* Perform the transformation */
                ias_geo_transform_coordinate(input_gcplib2input_image,
                    gcps->gcp_records[index].projection_x,
                    gcps->gcp_records[index].projection_y,
                    &map_coordinate_x, &map_coordinate_y);

                ias_geo_destroy_proj_transformation(input_gcplib2input_image);
            }
        }
        else /* Un-supported projection */
        {
            IAS_LOG_WARNING("Un-supported projection found [%s]",
                gcps->gcp_records[index].projection);
            /* Make the point invalid */
            gcps->gcp_locations[index].valid_flag = GCP_VALID_FLAG_DONT_USE;
            continue;
        }
            
        /* Figure out the line and sample locations for the GCP from the map
           coordinates */
        gcps->gcp_locations[index].input_data_line =
            (l1g_data->band_metadata.upper_left_y - map_coordinate_y)
                / l1g_data->band_metadata.projection_distance_y;

        IAS_LOG_DEBUG("Reference Line = %lf",
            gcps->gcp_locations[index].input_data_line);

        gcps->gcp_locations[index].input_data_sample =
            (map_coordinate_x - l1g_data->band_metadata.upper_left_x)
                / l1g_data->band_metadata.projection_distance_x;

        IAS_LOG_DEBUG("Reference Sample = %lf",
            gcps->gcp_locations[index].input_data_sample);

        gcps->gcp_locations[index].valid_flag = GCP_VALID_FLAG_USE;
    }

    return SUCCESS;
}

