
/* Standard C Includes */
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* IAS Includes */
#include "gctp.h"
#include "ias_const.h"
#include "ias_types.h"
#include "ias_logging.h"
#include "ias_gcp.h"
#include "ias_geo.h"
#include "ias_miscellaneous.h"

/* Local Includes */
#include "generate_chips.h"
#include "resample_data.h"

/* Define the minimum scaling factor, that is then used to get the maximum
   upsample ratio
   The upsample ratio is used later to allocate a maximum amount of memory
   needed for resampling */
#define MIN_SCALING_FACTOR 0.5
#define MAX_UP_SAMPLE_RATIO (1.0 / MIN_SCALING_FACTOR)
/* A kernel size of 0 indicates... "do not resample, make a copy"
   The number of resolutions here must match the defined value
       NUMBER_OF_RESOLUTIONS
   Using a scale value of 1.0 for non-scaled data is required */
RESOLUTION_SAMPLING resolutions[] =
{
    {"UP_SAMPLED_OUTPUT.img", 5, MIN_SCALING_FACTOR}, /* Example: 30m -> 15m */
    {"NONE_SAMPLED_OUTPUT.img", 0, 1.0},              /* Example: 30m -> 30m */
    {"DOWN_SAMPLED_OUTPUT.img", 7, 1.5}               /* Example: 30m -> 45m */
};

/*****************************************************************************
 NAME: compare_points

 PURPOSE: This is used by qsort to compare two moravec points.  It allows the
          qsort to sort from largest to smallest.

 RETURN VALUE: Type = int
    Value  Description
    -----  --------------------------------------------
    0      The values are equal or in the correct order
    1      The order of the values needs to be swapped

*****************************************************************************/
static int compare_points(const void *a, const void *b)
{
    const GENERATE_CHIPS_GCP_LOCATION *a_point = a;
    const GENERATE_CHIPS_GCP_LOCATION *b_point = b;

    if (a_point->moravec_value > b_point->moravec_value)
        return 0;

    return 1;
}

/*****************************************************************************
 NAME: generate_moravec_chips

 PURPOSE: To generate chips based on modified moravec interest operator.

 RETURN VALUE: Type = int
    Value      Description
    ---------  --------------------------
    SUCCESS    Successful completion
    ERROR      Operation failed
    NO_POINTS  Failed to find any points.

 NOTE: Memory is allocated for and returned in the GCPs

*****************************************************************************/
int generate_moravec_chips
(
    const GENERATE_CHIPS_PARAMETERS *parameters,/* I: The input parameters */
    const GENERATE_CHIPS_L1G_DATA *l1g_data,    /* I: Input L1G data */
    const IAS_IMAGE *quality_image,             /* I: The quality image data */
    IAS_BYTE_TYPE *exclusion_mask,              /* I: Exclusion mask */
    GENERATE_CHIPS_GCPS *gcps   /* O: Holds the input and output GCP records */
)
{
    int status;            /* Capture function/routine statuses */
    int count;             /* Capture snprintf errors */
    int resolution_index;  /* Loop control */
    int point_index;       /* Loop control */
    int input_sample_size; /* The data size of a sample */
    int input_data_count;   /* The total number of items in the data */
    int number_of_data_lines;   /* Number of lines in the resampled data */
    int number_of_data_samples; /* Number of samples in the resampled data */
    int picked_resolution; /* The resolution with the lowest count of points */
    int good_point_count;  /* A count of the good points */
    GENERATE_CHIPS_GCPS points[NUMBER_OF_RESOLUTIONS]; /* Hold all the points
                                                          for all the
                                                          resolutions */

    /* The following are buffer pointers */
    void *band_data;
    IAS_R4_TYPE *R4_input_buffer;
    IAS_R4_TYPE *R4_processing_buffer;

    char output_filename[PATH_MAX]; /* The name of the output file to create */

    /* Convert to local variables */
    double threshold = parameters->minimum_interest_operator_threshold;
    /* Convert L1G information */
    int number_of_image_lines = l1g_data->number_of_image_lines;
    int number_of_image_samples = l1g_data->number_of_image_samples;

    /* Zero out the points memory */
    memset(points, 0, sizeof(GENERATE_CHIPS_GCPS) * NUMBER_OF_RESOLUTIONS);

    /* Determine the total number of values */
    input_data_count = number_of_image_lines * number_of_image_samples;

    /* Figure out how many bytes are needed for a sample of the data,
       defaulting to the largest size supported */
    switch (l1g_data->data_type)
    {
        case IAS_BYTE:
            input_sample_size = sizeof(IAS_BYTE_TYPE);
            break;
        case IAS_I2:
            input_sample_size = sizeof(IAS_I2_TYPE);
            break;
        case IAS_UI2:
            input_sample_size = sizeof(IAS_UI2_TYPE);
            break;
        case IAS_R4:
        default:
            input_sample_size = sizeof(IAS_R4_TYPE);
            break;
    }

    /* Allocate memory for the band */
    band_data = calloc(input_data_count, input_sample_size);
    if (band_data == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for input band data");
        return ERROR;
    }

    /* Allocate memory for converting the band into R4 */
    R4_input_buffer = calloc(input_data_count, sizeof(IAS_R4_TYPE));
    if (R4_input_buffer == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for R*4 converted band data");
        free(band_data);
        return ERROR;
    }
 
    /* Read the band into the allocated memory */
    if (ias_l1g_read_image(l1g_data->band_io_ptr, 0, 0, 0,
        number_of_image_lines, number_of_image_samples, band_data) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading band data from input image");
        free(R4_input_buffer);
        free(band_data);
        return ERROR;
    }

    /* Convert to the internal data type */
    status = convert_to_data_type(input_data_count, l1g_data->data_type,
        band_data, IAS_R4, R4_input_buffer);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Converting image data type to R*4");
        free(R4_input_buffer);
        free(band_data);
        return ERROR;
    }

    /* Free the initial band_data and set the pointer to NULL */
    free(band_data);
    band_data = NULL;

    /* Dump intermediate data */
    if (parameters->output_intermediate_data == 1)
    {
        char output_filename[PATH_MAX];
        int count; 
        
        /* Create the full output file path */                     
        count = snprintf(output_filename, sizeof(output_filename), 
            "%s/SAMPLED_INPUT.img", parameters->intermediate_output_directory);
        if (count < 0 || count >= sizeof(output_filename))
        {
            IAS_LOG_ERROR("Bounds exceeded or internal snprintf error:"
                " count = %d", count);
            free(R4_input_buffer);
            free(band_data);
            return ERROR;
        }  

        status = ias_misc_write_envi_image(output_filename, NULL, 
            "Sampled Input Image", number_of_image_lines, 
            number_of_image_samples, 0.0, 0.0, 0.0, 0.0, "Sampled Input", 
            IAS_R4, R4_input_buffer);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to generate ENVI image");
            free(R4_input_buffer);
            free(band_data);
            return ERROR;
        }
    }

    /* Because the MAX_UP_SAMPLE_RATIO could be some weird fractional value,
       we need to make sure we allocate enough memory so we use ceil in the
       math to get it a bit bigger */
    /* Allocate memory for the highest resolution and data type supported */
    R4_processing_buffer = malloc(
        ceil(number_of_image_lines * MAX_UP_SAMPLE_RATIO)
        * ceil(number_of_image_samples * MAX_UP_SAMPLE_RATIO)
        * sizeof(IAS_R4_TYPE));
    if (R4_processing_buffer == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for resampled band data");
        free(R4_input_buffer);
        free(band_data);
        return ERROR;
    }

    /* Check which source resolution to use */
    picked_resolution = 0;
    if (parameters->moravec_source_resolution != MSR_FEWEST_POINTS)
    {
        picked_resolution = parameters->moravec_source_resolution;
    }

    /* Process through each resolution and generate moravec points */
    for (resolution_index = 0;
         resolution_index < NUMBER_OF_RESOLUTIONS;
         resolution_index++)
    {
        /* Scale the data to be used by the interest operator */
        status = resample_data(parameters, resolutions[resolution_index],
            number_of_image_lines, number_of_image_samples, R4_input_buffer,
            &number_of_data_lines, &number_of_data_samples,
            R4_processing_buffer);

        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Resampling input data");
            free(R4_processing_buffer);
            free(R4_input_buffer);
            free(band_data);
            return ERROR;
        }

        /* Specify the scaling value for conversion back to the input data
           resolution (30meters) */
        points[resolution_index].scaling_factor =
            resolutions[resolution_index].scaling_factor;

        /* Perform interest operator on the processing buffer */
        status = perform_interest_operator(parameters, R4_processing_buffer,
            number_of_data_lines, number_of_data_samples, exclusion_mask,
            quality_image->nl, quality_image->ns, &points[resolution_index]);
        /* Every point coming out of perform_interest_operator has the
           following valid_flag
           O: GCP_VALID_FLAG_MORAVEC_GOOD */

        /* In the code that follows, GCP_VALID_FLAG_MORAVEC_* values are
           described using I and O for input and output.  This is meant to
           describe the possible states of the valid_flag for the set of
           points being operated on.
           I: Means that on input to this routine some of the points in the
              set may be in this state.
           O: Means the on output to this routine some of the points may
              transitions from one state to another. from -> to
           ->: Is used to indicate a transition possibility on the output
               lines. */

        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Generating interest operator data");
            free(R4_processing_buffer);
            free(R4_input_buffer);
            free(band_data);
            return ERROR;
        }

        /* Only continue if we have points in all resolutions */
        if (points[resolution_index].number_of_gcps == 0)
        {
            IAS_LOG_WARNING(
                "No points above threshold (%lf) in resolution (%02d)",
                threshold, resolution_index);
            free(R4_processing_buffer);
            free(R4_input_buffer);
            free(band_data);
            /* Return no points found and let the caller decide what to do */
            return NO_POINTS;
        }

        /* Sort points according to interest operator value */
        qsort(points[resolution_index].gcp_locations,
            points[resolution_index].number_of_gcps,
            sizeof(GENERATE_CHIPS_GCP_LOCATION), compare_points);

        /* Dump intermediate data */
        if (parameters->output_intermediate_data == 1)
        {
            count = snprintf(output_filename, sizeof(output_filename),
                "sorted_points_resolution_%02d.envi", resolution_index);

            if (count < 0 || count >= sizeof(output_filename))
            {
                IAS_LOG_ERROR("Bounds exceeded or internal snprintf error:"
                    " creating ENVI point filename: count = %d", count);
                free(R4_processing_buffer);
                free(R4_input_buffer);
                free(band_data);
                return ERROR;
            }
            else
            {
                status = dump_intermediate_envi_point_file(parameters,
                    output_filename, &points[resolution_index]);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Creating intermediate ENVI point file");
                    free(R4_processing_buffer);
                    free(R4_input_buffer);
                    free(band_data);
                    return ERROR;
                }
            }
        }

        /* Check which resolution should be used */
        if (parameters->moravec_source_resolution == MSR_FEWEST_POINTS)
        {
            /* Find correct resolution list to use as baseline */
            if (points[resolution_index].number_of_gcps
                < points[picked_resolution].number_of_gcps)
            {
                picked_resolution = resolution_index;
            }
        }
    } /* END resolution_index loop */

    /* Free the temporary memory and set the pointers to NULL */
    free(R4_processing_buffer);
    free(R4_input_buffer);
    free(band_data);
    R4_processing_buffer = NULL;
    R4_input_buffer = NULL;
    band_data = NULL;

    IAS_LOG_INFO("Using resolution %d as the base", picked_resolution);
    IAS_LOG_INFO("Before resolution check, %d points passed the test",
        points[picked_resolution].number_of_gcps);

    /* Check the resolution. */
    for (resolution_index = 0; resolution_index < NUMBER_OF_RESOLUTIONS;
         resolution_index++)
    {
        moravec_resolution_check(resolutions, resolution_index, points);
    }

    /* The moravec_resolution_check caused the following to happened to the
       valid_flag
       I: GCP_VALID_FLAG_MORAVEC_GOOD
       O: GCP_VALID_FLAG_MORAVEC_GOOD -> GCP_VALID_FLAG_MORAVEC_DONT_USE */

    /* Get a count of the surviving points */
    good_point_count = 0;
    for (point_index = 0;
         point_index < points[picked_resolution].number_of_gcps;
         point_index++)
    {
        if (points[picked_resolution].gcp_locations[point_index].valid_flag
            == GCP_VALID_FLAG_MORAVEC_GOOD)
        {
            good_point_count++;
        }
    }
    IAS_LOG_INFO("After resolution check, %d points passed the test",
        good_point_count);

    /* Filter the remaining points based on the size of chips */
    moravec_window_points(parameters, &points[picked_resolution]);
    /* The moravec_window_points caused the following to happened to the
       valid_flag
       I: GCP_VALID_FLAG_MORAVEC_GOOD
       I: GCP_VALID_FLAG_MORAVEC_DONT_USE
       O: GCP_VALID_FLAG_MORAVEC_GOOD -> GCP_VALID_FLAG_MORAVEC_DONT_USE */

    /* Get a count of the surviving points */
    good_point_count = 0;
    for (point_index = 0;
         point_index < points[picked_resolution].number_of_gcps;
         point_index++)
    {
        if (points[picked_resolution].gcp_locations[point_index].valid_flag
            == GCP_VALID_FLAG_MORAVEC_GOOD)
        {
            good_point_count++;
        }
    }
    IAS_LOG_INFO("After windowing, %d points passed the test",
        good_point_count);

    /* Limit the number of points to the requested number region checking
       performed later will add points back in if needed */
    moravec_limit_points(parameters, &points[picked_resolution]);
    /* The moravec_limit_points caused the following to happened to the
       valid_flag
       I: GCP_VALID_FLAG_MORAVEC_GOOD
       I: GCP_VALID_FLAG_MORAVEC_DONT_USE
       O: GCP_VALID_FLAG_MORAVEC_GOOD -> GCP_VALID_FLAG_MORAVEC_GOOD_DONT_USE */

    /* Dump intermediate data */
    if (parameters->output_intermediate_data == 1)
    {
        /* Print out moravec points */
        for (resolution_index = 0;
             resolution_index < NUMBER_OF_RESOLUTIONS;
             resolution_index++)
        {
            count = snprintf(output_filename, sizeof(output_filename),
                "final_moravec_points_resolution_%02d.envi", resolution_index);

            if (count < 0 || count >= sizeof(output_filename))
            {
                IAS_LOG_ERROR("Bounds exceeded or internal snprintf error:"
                    " creating ENVI point filename: count = %d", count);
                return ERROR;
            }
            else
            {
                status = dump_intermediate_envi_point_file(parameters,
                    output_filename, &points[resolution_index]);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Creating intermediate ENVI point file");
                    return ERROR;
                }
            }
        } /* END resolution_index loop */
    }

    /* Split data set up into sub-divisions, determining if enough
       valid points are within area */ 
    moravec_check_regions(parameters, l1g_data, &points[picked_resolution]);
    /* The moravec_check_regions caused the following to happened to the
       valid_flag
       I: GCP_VALID_FLAG_MORAVEC_GOOD
       I: GCP_VALID_FLAG_MORAVEC_DONT_USE
       I: GCP_VALID_FLAG_MORAVEC_GOOD_DONT_USE
       O: GCP_VALID_FLAG_MORAVEC_GOOD_DONT_USE -> GCP_VALID_FLAG_MORAVEC_GOOD */

    /* Get a count of the surviving points */
    good_point_count = 0;
    for (point_index = 0;
         point_index < points[picked_resolution].number_of_gcps;
         point_index++)
    {
        if (points[picked_resolution].gcp_locations[point_index].valid_flag
            == GCP_VALID_FLAG_MORAVEC_GOOD)
        {
            /* Set the use flag for the rest of the software */
            points[picked_resolution].gcp_locations[point_index].valid_flag =
                GCP_VALID_FLAG_USE;
            good_point_count++;
        }
        else
        {
            points[picked_resolution].gcp_locations[point_index].valid_flag =
                GCP_VALID_FLAG_DONT_USE;
        }
    }
    /* The above for loop caused the following to happened to the valid_flag
       I: GCP_VALID_FLAG_MORAVEC_GOOD
       I: GCP_VALID_FLAG_MORAVEC_DONT_USE
       I: GCP_VALID_FLAG_MORAVEC_GOOD_DONT_USE
       O: GCP_VALID_FLAG_MORAVEC_GOOD -> GCP_VALID_FLAG_USE
       O: GCP_VALID_FLAG_MORAVEC_DONT_USE -> GCP_VALID_FLAG_DONT_USE
       O: GCP_VALID_FLAG_MORAVEC_GOOD_DONT_USE -> GCP_VALID_FLAG_DONT_USE */

    IAS_LOG_INFO("After region checking, %d points passed the test",
        good_point_count);

    /* Verify that we have points */
    if (good_point_count == 0)
    {
        IAS_LOG_WARNING("No valid points after region checking");
        /* Return no points found and let the caller decide what to do */
        return NO_POINTS;
    }

    /* Copy the internal points to the external GCP records */
    status = copy_gcp_locations(GCP_VALID_FLAG_USE,
        &points[picked_resolution], gcps);
    /* Check the status after the following free'ing of points memory */

    /* Free the memory allocated for the GCPs in the location variables since
       it is no-longer needed */
    for (resolution_index = 0;
         resolution_index < NUMBER_OF_RESOLUTIONS;
         resolution_index++)
    {
        if (points[resolution_index].gcp_locations != NULL)
        {
            free(points[resolution_index].gcp_locations);
            points[resolution_index].gcp_locations = NULL;
        }
    }

    /* Now check the status.  Doing this allows the free points above to only
       have to be written once */
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Failed to copy GCPs");
        return ERROR;
    }

    /* Make sure we have some GCPs after the copy */
    if (gcps->number_of_gcps == 0)
    {
        IAS_LOG_WARNING("No GCPs generated");
        /* Return no points found and let the caller decide what to do */
        return NO_POINTS;
    }

    /* Dump intermediate data */
    if (parameters->output_intermediate_data == 1)
    {
        status = dump_intermediate_envi_point_file(parameters,
            "final_good_moravec_points.envi", gcps);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Creating intermediate ENVI point file");
            return ERROR;
        }
    }

    return status;
}

