
/* Standard C Includes */
#include <stdlib.h>
#include <string.h>

/* IAS Includes */
#include "ias_gcp.h"
#include "ias_logging.h"

/* Local Includes */
#include "generate_chips.h"

/*****************************************************************************
 NAME: merge_gcps

 PURPOSE: To merge all the flagged GCPs from the two source datasets records
          to the destination records.  This will only copy the ones which
          match the flag value.

 RETURN VALUE: Type = int
    Value    Description
    -------  --------------------------
    SUCCESS  Successful completion
    ERROR    Operation failed

  NOTE: Memory may be realloced, and is returned in the destination GCPs

*****************************************************************************/
int merge_gcps
(
    GCP_VALID_FLAG_TYPE flag_value,           /* I: What to copy */
    const GENERATE_CHIPS_GCPS *source_1_gcps, /* I: Holds input GCP records */
    const GENERATE_CHIPS_GCPS *source_2_gcps, /* I: Holds input GCP records */
    GENERATE_CHIPS_GCPS *destination_gcps /* O: Holds the output GCP records */
)
{
    int status;                  /* capture function/routine statuses */
    int destination_count;       /* Count of destination GCPs */
    int combined_number_of_gcps; /* The total combined number of source GCPs */

    /* Temp pointers for memory reallocation */
    IAS_GCP_RECORD *temp_gcp_records;
    GENERATE_CHIPS_GCP_LOCATION *temp_gcp_locations;

    /* Setup the number of GCPs */
    combined_number_of_gcps = source_1_gcps->number_of_gcps
        + source_2_gcps->number_of_gcps;

    /* Allocate the memory for the GCPs */
    status = allocate_gcp_memory(combined_number_of_gcps, destination_gcps);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Allocating memory for GCPs");
        return ERROR;
    }

    /* Copy source 1 to the destination */
    destination_count = 0;
    copy_gcps_to_destination_index(flag_value, source_1_gcps,
        &destination_count, destination_gcps);

    /* Copy source 2 to the destination */
    copy_gcps_to_destination_index(flag_value, source_2_gcps,
        &destination_count, destination_gcps);

    /* Reallocate the memory to the smaller size */
    if (destination_count != combined_number_of_gcps && destination_count > 0)
    {
        /* Get new memory */
        temp_gcp_records = realloc(destination_gcps->gcp_records,
            sizeof(IAS_GCP_RECORD) * destination_count);

        /* Make sure memory was allocated */
        if (temp_gcp_records == NULL)
        {
            IAS_LOG_ERROR("Reallocating memory for GCP records");
            free(destination_gcps->gcp_records);
            destination_gcps->gcp_records = NULL;
            free(destination_gcps->gcp_locations);
            destination_gcps->gcp_locations = NULL;
            return ERROR;
        }

        /* Fix the pointer */
        destination_gcps->gcp_records = temp_gcp_records;

        /* Get new memory */
        temp_gcp_locations = realloc(destination_gcps->gcp_locations,
            sizeof(GENERATE_CHIPS_GCP_LOCATION) * destination_count);

        /* Make sure memory was allocated */
        if (temp_gcp_locations == NULL || temp_gcp_records == NULL)
        {
            IAS_LOG_ERROR("Reallocating memory for GCP arrays");
            free(destination_gcps->gcp_records);
            destination_gcps->gcp_records = NULL;
            free(destination_gcps->gcp_locations);
            destination_gcps->gcp_locations = NULL;
            return ERROR;
        }

        /* Fix the pointer */
        destination_gcps->gcp_locations = temp_gcp_locations;
    }

    /* Didn't find any GCPs so error out */
    if (destination_count == 0)
    {
        IAS_LOG_ERROR("No GCPs were found to copy");
        free(destination_gcps->gcp_records);
        destination_gcps->gcp_records = NULL;
        free(destination_gcps->gcp_locations);
        destination_gcps->gcp_locations = NULL;
        return ERROR;
    }

    /* Fix the count */
    destination_gcps->number_of_gcps = destination_count;

    return SUCCESS;
}

