
/* IAS Includes */
#include "ias_logging.h"

/* Local Includes */
#include "generate_chips.h"

/*****************************************************************************
 NAME: check_region

 PURPOSE: Only used in this file.  Determines if a region has enough points
          remaining, and if not, processes through all the points and attempts
          to add good points removed by moravec_limit_points.

 RETURN VALUE: None

*****************************************************************************/
void check_region
(
    const GENERATE_CHIPS_PARAMETERS *parameters, /* I: The input parameters */
    double top_line_boundary,     /* I: Top boundary line */
    double bottom_line_boundary,  /* I: Bottom boundary line */
    double left_sample_boundary,  /* I: Left boundary sample */
    double right_sample_boundary, /* I: Right boundary sample */
    GENERATE_CHIPS_GCPS *gcps     /* I/O: The picked resolution */
)
{
    int point_index;              /* Loop control */
    int valid_region_point_count; /* Initial count of good points found */
    int total_point_count;        /* Total point counts after adding points */

    /* Line and sample location for the current point */
    double line;
    double sample;

    IAS_LOG_DEBUG("Checking area (%lf, %lf) -> (%lf, %lf)",
        top_line_boundary, left_sample_boundary,
        bottom_line_boundary, right_sample_boundary);

    valid_region_point_count = 0;
    /* Process through all the points for the picked resolution and generate a
       count of the number of good points for the region */
    for (point_index = 0;
         point_index < gcps->number_of_gcps;
         point_index++)
    {
        IAS_LOG_DEBUG("Checking point %d, valid_flag %d", point_index,
            gcps->gcp_locations[point_index].valid_flag);

        /* Only look for points marked good */
        if (gcps->gcp_locations[point_index].valid_flag
            == GCP_VALID_FLAG_MORAVEC_GOOD)
        {
            /* Grab the line and sample locations */
            line = gcps->gcp_locations[point_index].input_data_line;
            sample = gcps->gcp_locations[point_index].input_data_sample;

            /* Verify that the point is within the line region */
            if (line >= top_line_boundary && line <= bottom_line_boundary)
            {
                /* Verify that the point is within the sample region */
                if (sample >= left_sample_boundary
                    && sample <= right_sample_boundary)
                {
                    valid_region_point_count++;
                }
            }
        } /* END good GCPs */
    } /* END point_index loop */

    IAS_LOG_DEBUG("Checked area (%lf, %lf) -> (%lf, %lf) found %d good points",
        top_line_boundary, left_sample_boundary,
        bottom_line_boundary, right_sample_boundary, valid_region_point_count);

    /* Check and see if we already have enough points, and if we don't attempt
       to add more points back in to accommodate the region */
    total_point_count = valid_region_point_count;
    if (valid_region_point_count
        < parameters->minimum_number_of_points_in_region)
    {
        /* Process through all the points for the picked resolution */
        for (point_index = 0;
             point_index < gcps->number_of_gcps;
             point_index++)
        {
            /* Only look for points that were limited by moravec_limit_points
               Because the points are sorted by best moravec value first, we
               can add the points back in on a first come basis, as long as
               they are with-in the boundaries */
            if (gcps->gcp_locations[point_index].valid_flag
                == GCP_VALID_FLAG_MORAVEC_GOOD_DONT_USE)
            {
                /* Grab the line and sample locations */
                line = gcps->gcp_locations[point_index].input_data_line;
                sample = gcps->gcp_locations[point_index].input_data_sample;

                /* Verify that the point is within the line region */
                if (line >= top_line_boundary && line <= bottom_line_boundary)
                {
                    /* Verify that the point is within the sample region */
                    if (sample >= left_sample_boundary
                        && sample <= right_sample_boundary)
                    {
                        /* Set to good */
                        gcps->gcp_locations[point_index].valid_flag =
                            GCP_VALID_FLAG_MORAVEC_GOOD;

                        total_point_count++;

                        /* Check if we have enough points yet */
                        if (total_point_count
                            >= parameters->minimum_number_of_points_in_region)
                        {
                            /* We have enough points for the region now, so
                               return back to the calling routine after
                               logging a statement for how many we had to add
                               back in for the region */
                            IAS_LOG_INFO("Area (%lf, %lf) -> (%lf, %lf)"
                                " needed %d points added",
                                top_line_boundary, left_sample_boundary,
                                bottom_line_boundary, right_sample_boundary,
                                parameters->minimum_number_of_points_in_region
                                    - valid_region_point_count);
                            break;
                        } /* END total_point_count check */
                    } /* END sample boundary check */
                } /* END line boundary check */
            } /* END if GCP_VALID_FLAG_MORAVEC_GOOD_DONT_USE */
        } /* END point_index loop */
    } /* END count less than minimum */

    if (total_point_count < parameters->minimum_number_of_points_in_region)
    {
        /* If we get here we did not have enough points to meet the minimum
           for a region so log that bit of information
           It isn't a fatal issue, we just move on with out them,  because it
           could have been caused by water or clouds in the region */
        IAS_LOG_INFO("Area (%lf, %lf) -> (%lf, %lf) will have only %d good"
            " points", top_line_boundary, left_sample_boundary,
            bottom_line_boundary, right_sample_boundary, total_point_count);
    }
};

/*****************************************************************************
 NAME: moravec_check_regions

 PURPOSE: Figure out the region boundaries and then call check_region to
          validate that region.

 RETURN VALUE: None

*****************************************************************************/
void moravec_check_regions
(
    const GENERATE_CHIPS_PARAMETERS *parameters, /* I: The input parameters */
    const GENERATE_CHIPS_L1G_DATA *l1g_data,     /* I: Input L1G data */
    GENERATE_CHIPS_GCPS *gcps                    /* I/O: The GCPs to check */
)
{
    int point_index;         /* Loop control */
    int point_count;         /* The number of points flagged as good after
                                each region check */
    int region_line_index;   /* The line index into the regions */
    int region_sample_index; /* The sample index into the regions */

    double region_line_scale;   /* The line scaling to use for region
                                   boundary */
    double region_sample_scale; /* The sample scaling to use for region
                                   boundary */

    /* Line and sample boundary information for the current region */
    double top_line_boundary;
    double bottom_line_boundary;
    double left_sample_boundary;
    double right_sample_boundary;

    /* Figure out the scaling to use for figuring out the region boundaries */
    region_line_scale = (double)l1g_data->number_of_image_lines
        / parameters->number_of_regions_lines;
    region_sample_scale = (double)l1g_data->number_of_image_samples
        / parameters->number_of_regions_samples;

    /* Process through the regions */
    for (region_line_index = 0;
         region_line_index < parameters->number_of_regions_lines;
         region_line_index++)
    {
        /* Determine bounds on lines for region */
        top_line_boundary = (double)region_line_index * region_line_scale;
        bottom_line_boundary =
            (double)(region_line_index + 1.0) * region_line_scale - 1.0;

        for (region_sample_index = 0;
             region_sample_index < parameters->number_of_regions_samples;
             region_sample_index++)
        {
            /* Determine bounds on samples for region */
            left_sample_boundary =
                (double)region_sample_index * region_sample_scale;
            right_sample_boundary =
                (double)(region_sample_index + 1.0) * region_sample_scale - 1.0;
            
            /* Check regions or sub-divisions for minimum number of 
               GCPs.  Only the picked resolution is checked. */ 
            check_region(parameters, top_line_boundary, bottom_line_boundary,
                    left_sample_boundary, right_sample_boundary, gcps);

            /* Notify how many points we have after each region check */
            point_count = 0;
            for (point_index = 0;
                 point_index < gcps->number_of_gcps;
                 point_index++)
            {
                if (gcps->gcp_locations[point_index].valid_flag
                    == GCP_VALID_FLAG_MORAVEC_GOOD)
                {
                    point_count++;
                }
            }
            IAS_LOG_INFO("Region (%d, %d) using %d good points",
                region_line_index, region_sample_index, point_count);
        } /* END region_sample_index */
    } /* END region_line_index */
}

