
/* IAS Includes */
#include "ias_logging.h"

/* Local Includes */
#include "generate_chips.h"

/* The amount of slop in pixels/samples when looking for matching points
   between the resolutions */
#define SLOP 2

/*****************************************************************************
 NAME: moravec_resolution_check

 PURPOSE: Iterate through the reference locations and if matches are found in
          the other resolutions, mark those points as
          GCP_VALID_FLAG_MORAVEC_GOOD.  Otherwise mark the point as
          GCP_VALID_FLAG_MORAVEC_DONT_USE.

 RETURN VALUE: None

 NOTE: The original cal/val code compared the resolutions in the lowest
       resolution, which caused many extra multiplications and divisions due
       to revisiting the same points over and over for each reference/picked
       resolution point.  Also the SLOP was applied to the lowest resolution
       being compared.

       So in this implementation the bounding box tests are performed in input
       data space, and not in the resolutions.  Thus removing all the
       divisions and most of the multiplications, and the main looping is
       reduced to adding and subtracting and comparisons.  To accommodate the
       SLOP being in the lowest resolution, the SLOP(in input space) is scaled
       based on the resolutions being compared before entering the main
       looping code.

*****************************************************************************/
void moravec_resolution_check
(
    RESOLUTION_SAMPLING *resolutions, /* I: Resolution details */
    int reference_resolution,         /* I: Reference resolution */
    GENERATE_CHIPS_GCPS gcps[NUMBER_OF_RESOLUTIONS] /* I/O: All points for all
                                                            resolutions */
)
{
    int resolution_index;      /* Loop control */
    int reference_point_index; /* Loop control */
    int search_point_index;    /* Loop control */
    int current_valid_flag; /* Validity flag for the current point */
    int reference_line;     /* Line for the reference resolution */
    int reference_sample;   /* Sample for the reference resolution */
    int search_line;        /* Line for the search resolution */
    int search_sample;      /* Sample for the search resolution */

    /* The following define the area to search */
    int top_line_boundary;
    int bottom_line_boundary;
    int left_sample_boundary;
    int right_sample_boundary;

    int slop[NUMBER_OF_RESOLUTIONS]; /* Slop in search area */
    double slop_scale; /* Scaling to use for the reference line/sample */

    /* The following provide local access variables */
    GENERATE_CHIPS_GCP_LOCATION *reference_gcp_locations;
    GENERATE_CHIPS_GCP_LOCATION *search_gcp_locations;

    /* Point to the reference GCP locations */
    reference_gcp_locations = gcps[reference_resolution].gcp_locations;

    /* Figure out the slop needed for each resolution */
    for (resolution_index = 0;
         resolution_index < NUMBER_OF_RESOLUTIONS;
         resolution_index++)
    {
        if (resolution_index != reference_resolution)
        {
            /* Figure out the scaling that needs to be performed on the SLOP
               value so that the search areas are based on the resolutions */
            if (resolutions[reference_resolution].scaling_factor <
                resolutions[resolution_index].scaling_factor)
            {
                slop_scale = resolutions[resolution_index].scaling_factor
                    / resolutions[reference_resolution].scaling_factor;
            }
            else
            {
                slop_scale = resolutions[reference_resolution].scaling_factor
                    / resolutions[resolution_index].scaling_factor;
            }
        } /* END different resolution */
        else
        {
            slop_scale = 1.0;
        }

        slop[resolution_index] = SLOP * slop_scale;

        IAS_LOG_INFO("Reference Resolution(%d)"
            " Search Resolution(%d) Slop Scale(%lf) Slop(%d)",
            reference_resolution, resolution_index, slop_scale,
            slop[resolution_index]);
    } /* END resolution index */

    for (reference_point_index = 0;
         reference_point_index < gcps[reference_resolution].number_of_gcps;
         reference_point_index++)
    {
        /* If the point is already bad, then move on to the next point */
        if (!(reference_gcp_locations[reference_point_index]
              .valid_flag == GCP_VALID_FLAG_MORAVEC_GOOD))
        {
            continue;
        }

        /* Start with it being a valid point */
        current_valid_flag = GCP_VALID_FLAG_MORAVEC_GOOD;

        /* Figure out the matching reference line and sample in input space */
        reference_line =
            reference_gcp_locations[reference_point_index].input_data_line;
        reference_sample =
            reference_gcp_locations[reference_point_index].input_data_sample;

        IAS_LOG_DEBUG("Looking at reference resolution (%d) for point"
            " Resolution(%d, %d) Input(%d, %d)", reference_resolution,
            (int)reference_gcp_locations[reference_point_index]
                .resolution_line,
            (int)reference_gcp_locations[reference_point_index]
                .resolution_sample,
            reference_line, reference_sample);

        for (resolution_index = 0;
             resolution_index < NUMBER_OF_RESOLUTIONS;
             resolution_index++)
        {
            /* Only do work with the other resolutions */
            if (resolution_index != reference_resolution)
            {
                /* Figure out the top and bottom line boundaries */
                top_line_boundary = reference_line - slop[resolution_index];
                bottom_line_boundary = reference_line + slop[resolution_index];

                /* Figure out the left and right sample boundaries */
                left_sample_boundary =
                    reference_sample - slop[resolution_index];
                right_sample_boundary =
                    reference_sample + slop[resolution_index];

                IAS_LOG_DEBUG("Bounding box points From(%d, %d) To(%d, %d)",
                    top_line_boundary, left_sample_boundary,
                    bottom_line_boundary, right_sample_boundary);

                search_gcp_locations = gcps[resolution_index].gcp_locations;

                /* Process through the search points */
                for (search_point_index = 0;
                     search_point_index < gcps[resolution_index].number_of_gcps;
                     search_point_index++)
                {
                    /* Figure out the matching search line and sample in
                       input space */
                    search_line =
                        search_gcp_locations[search_point_index]
                            .input_data_line;
                    search_sample =
                        search_gcp_locations[search_point_index]
                            .input_data_sample;

                    IAS_LOG_DEBUG("Looking at search resolution (%d) for"
                        " point Resolution(%d, %d) Input(%d, %d)",
                        resolution_index,
                        (int)search_gcp_locations[search_point_index]
                            .resolution_line,
                        (int)search_gcp_locations[search_point_index]
                            .resolution_sample,
                        search_line, search_sample);

                    /* Check and see if the reference point falls within range
                       of the search point */
                    if (search_line >= top_line_boundary
                        && search_line <= bottom_line_boundary)
                    {
                        if (search_sample >= left_sample_boundary
                            && search_sample <= right_sample_boundary)
                        {
                            IAS_LOG_DEBUG("Found match at search resolution"
                                " (%d) for point Resolution(%d, %d)"
                                " Input(%d, %d)", resolution_index,
                                (int)reference_gcp_locations[
                                    reference_point_index].resolution_line,
                                (int)reference_gcp_locations[
                                    reference_point_index].resolution_sample,
                                search_line, search_sample);

                            /* Break out of the search loop because we found
                               a hit as there will only be one possible hit
                               due to the windowing */
                            break;
                        }
                    }
                } /* END search point loop */

                if (search_point_index
                    == gcps[resolution_index].number_of_gcps)
                {
                    current_valid_flag = GCP_VALID_FLAG_MORAVEC_DONT_USE;

                    IAS_LOG_DEBUG("Could not find a match for point (%d, %d)"
                        " in resolution (%d)",
                        (int)reference_gcp_locations[reference_point_index]
                            .resolution_line,
                        (int)reference_gcp_locations[reference_point_index]
                            .resolution_sample,
                         reference_resolution);
                }
            } /* END resolution check */
        } /* END resolution loop */

        /* Set the valid bits accordingly */
        reference_gcp_locations[reference_point_index].valid_flag =
                current_valid_flag;
    } /* END reference point loop */
}

