
/* IAS Includes */
#include "ias_logging.h"

/* Local Includes */
#include "generate_chips.h"

/*****************************************************************************
 NAME: moravec_window_points

 PURPOSE: Iterate through the points and remove any points of lower value that
          fall within the size of the chips to be created of higher value
          points.  Since the points are sorted by largest moravec value, the
          points which get removed will be of lower value.

 RETURN VALUE: None

*****************************************************************************/
void moravec_window_points
(
    const GENERATE_CHIPS_PARAMETERS *parameters, /* I: The input parameters */
    GENERATE_CHIPS_GCPS *gcps                    /* I/O: The picked GCPs */
)
{
    int point_index_1;      /* Loop control */
    int point_index_2;      /* Loop control */
    int window_size_line;   /* The size of the chip */
    int window_size_sample; /* The size of the chip */
    int line;               /* The current GCP line to check */
    int sample;             /* The current GCP sample to check */
    int number_of_gcps;     /* The number of GCPs to process */

    /* Used to specify the area covered by a GCP */
    int top_line_boundary;
    int bottom_line_boundary;
    int left_sample_boundary;
    int right_sample_boundary;

    /* Local pointer to the GCP locations */
    GENERATE_CHIPS_GCP_LOCATION *gcp_locations;

    /* Convert to local variables */
    number_of_gcps = gcps->number_of_gcps;
    gcp_locations = gcps->gcp_locations;
    window_size_line = parameters->minimum_distance_lines;
    window_size_sample = parameters->minimum_distance_samps;

    /* Loop through the points and flag the ones we do not want to keep */
    for (point_index_1 = 0; point_index_1 < number_of_gcps - 1; point_index_1++)
    {
        /* Check and see if this point is already bad */
        if (gcp_locations[point_index_1].valid_flag
            == GCP_VALID_FLAG_MORAVEC_DONT_USE)
        {
            continue;
        }

        /* Figure out the boundaries */
        top_line_boundary = gcp_locations[point_index_1].input_data_line
            - window_size_line;
        bottom_line_boundary = gcp_locations[point_index_1].input_data_line
            + window_size_line;
        left_sample_boundary = gcp_locations[point_index_1].input_data_sample
            - window_size_sample;
        right_sample_boundary = gcp_locations[point_index_1].input_data_sample
            + window_size_sample;

        /* Loop through the others */
        for (point_index_2 = point_index_1 + 1;
             point_index_2 < number_of_gcps;
             point_index_2++)
        {
            /* Check and see if this point is already bad */
            if (gcp_locations[point_index_2].valid_flag
                == GCP_VALID_FLAG_MORAVEC_DONT_USE)
            {
                continue;
            }

            line = gcp_locations[point_index_2].input_data_line;
            sample = gcp_locations[point_index_2].input_data_sample;

            /* Check the line first */
            if (((line >= top_line_boundary) && (line <= bottom_line_boundary))
                &&
                ((sample >= left_sample_boundary)
                 && (sample <= right_sample_boundary)))
            {
                /* Set to don't use */
                gcp_locations[point_index_2].valid_flag =
                    GCP_VALID_FLAG_MORAVEC_DONT_USE;
            }
        } /* END inside for loop */
    } /* END outside for loop */
}

