
/* Standard C Includes */
#include <stdlib.h>

/* IAS Includes */
#include "ias_const.h"
#include "ias_l1g.h"
#include "ias_logging.h"

/* Local Includes */
#include "generate_chips.h"

/*****************************************************************************
 NAME: open_l1g_data

 PURPOSE: Opens all the information related to the specified L1G data

 RETURN VALUE: Type = int
    Value    Description
    -------  --------------------------
    SUCCESS  Successful completion
    ERROR    Operation failed

*****************************************************************************/
int open_l1g_data
(
    const char *filename,             /* I: The name of the L1G file to open */
    int band_number,                  /* I: The band number to open */
    GENERATE_CHIPS_L1G_DATA *l1g_data /* O: Contains all the information
                                            needed to access the L1G */
)
{
    int status; /* capture function/routine statuses */

    /* Open the input image */
    l1g_data->io_ptr = ias_l1g_open_image(filename, IAS_READ);
    if (l1g_data->io_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening image %s", filename);
        return ERROR;
    }

    /* Load the metadata */
    status = ias_l1g_get_file_metadata(l1g_data->io_ptr, &l1g_data->metadata);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving file metadata for %s", filename);
        ias_l1g_close_image(l1g_data->io_ptr);
        return ERROR;
    }

    /* Verify the band is present */
    if (!ias_l1g_is_band_present(l1g_data->io_ptr, band_number))
    {
        IAS_LOG_ERROR("Band number %d not present in the input data",
            band_number);
        ias_l1g_close_image(l1g_data->io_ptr);
        return ERROR;
    }

    /* Open the band */
    l1g_data->band_io_ptr = ias_l1g_open_band(l1g_data->io_ptr,
        band_number, &l1g_data->data_type, &l1g_data->number_of_scas,
        &l1g_data->number_of_image_lines, &l1g_data->number_of_image_samples);
    if (l1g_data->band_io_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening L1G band %d of image %s", band_number, filename);
        close_l1g_data(l1g_data);
        return ERROR;
    }

    /* Verify that only 1 SCA is present in the image data */
    if (l1g_data->number_of_scas != 1)
    {
        IAS_LOG_ERROR("Multiple SCAs present in the input image, but only SCA"
            " combined images are supported: Found %d in image %s",
            l1g_data->number_of_scas, filename);
        close_l1g_data(l1g_data);
        return ERROR;
    }

    /* Get the band metadata */
    status = ias_l1g_get_band_metadata(l1g_data->io_ptr,
        band_number, &l1g_data->band_metadata);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading band %d of metadata", band_number);
        close_l1g_data(l1g_data);
        return ERROR;
    }

    return SUCCESS;
}

/* Closes all the open information related to the specified L1G data */
void close_l1g_data
(
    const GENERATE_CHIPS_L1G_DATA *l1g_data /* I: Contains all the information
                                                  needed to access the L1G */
)
{
    if (l1g_data->band_io_ptr != NULL)
        ias_l1g_close_band(l1g_data->band_io_ptr);
    if (l1g_data->io_ptr != NULL)
        ias_l1g_close_image(l1g_data->io_ptr);
}

