
/* Standard C Includes */
#include <stdlib.h>
#include <string.h>
#include <float.h>

/* IAS Includes */
#include "ias_const.h"
#include "ias_types.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"

/* Local Includes */
#include "generate_chips.h"

/* How much to adjust for the new memory allocation */
#define MEMORY_ADJUSTMENT 6

/*****************************************************************************
 NAME: perform_interest_operator

 PURPOSE: To generate GCP locations with computed interest operator values.

 RETURN VALUE: Type = int
    Value      Description
    ---------  --------------------------
    SUCCESS    Successful completion
    ERROR      Operation failed

*****************************************************************************/
int perform_interest_operator
(
    const GENERATE_CHIPS_PARAMETERS *parameters, /* I: The input parameters */
    IAS_R4_TYPE *input_data,       /* I: Input image data */
    int number_of_lines,           /* I: Number of lines in image data */
    int number_of_samples,         /* I: Number of samples in image data */
    IAS_BYTE_TYPE *exclusion_mask, /* I: Mask containing points where interest
                                         operator can be performed */
    int number_of_mask_lines,      /* I: Number of lines in the mask */
    int number_of_mask_samples,    /* I: Number of samples in the mask */
    GENERATE_CHIPS_GCPS *gcps      /* O: The GCPs generated for this data */
)
{
    int line_index;          /* Loop control */
    int sample_index;        /* Loop control */
    int window_line_index;   /* Loop control */
    int window_sample_index; /* Loop control */
    int window_index;        /* Loop control */
    int status;        /* Capture function/routine statuses */
    int count;         /* Capture snprintf errors */
    int filter_size;   /* The size of the filtering window */
    int window_offset; /* Window offset of current line in exclusion mask */
    int size;          /* Distance from interest point for window creation */
    int data_size;     /* Input image size */
    int e_mask_line_location;   /* Line location in the exclusion mask */
    int e_mask_sample_location; /* Sample location in the exclusion mask */
    int e_mask_location;        /* Location in the exclusion mask */
    int data_location;          /* Base location within the data */

    double sum_of_sqrs;         /* Summing variable */
    double scale;               /* Scaling to get back to input resolution */
    double difference;          /* Difference */
    double local_minimum_value; /* Minimum values from interest operator */
    double minimum_value;       /* Minimum values from interest operator */
    double maximum_value;       /* Maximum values from interest operator */
    int invalid;                /* Invalid pixel found/not found */
    int count_of_all_gcps;      /* Counter */

    int current_memory_allocation; /* How much memory is currently allocated */
    GENERATE_CHIPS_GCP_LOCATION *temp_ptr; /* Used by the realloc code */

    IAS_R4_TYPE *moravec_buffer; /* The buffer for holding all the interest
                                    operator values */

    char output_filename[PATH_MAX]; /* The name of the output file to create */
    static int output_file_number = 0; /* Used to number the output files for
                                          each resolution */

    IAS_LOG_INFO("Performing Interest Operator ...");

    /* Convert to local variables */
    size = parameters->interest_operator_size;
    filter_size = parameters->local_maximum_distance;
    scale = gcps->scaling_factor;

    /* Figure out the size of the input data */
    data_size = number_of_lines * number_of_samples;

    /* Allocate memory for the highest resolution needed */
    moravec_buffer = calloc(data_size, sizeof(IAS_R4_TYPE));
    if (moravec_buffer == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for moravec data");
        return ERROR;
    }

    /* Loop on imagery determining interest operator skipping a border of
       size "size"
       "size" is actually about half of the window
       If you have a size of 5, then the window is ssssspsssss, where s
       represents the size specified and p is the point of interest
       So the window "size" in this case is 11 */
    for (line_index = size;
         line_index < (number_of_lines - size);
         line_index++)
    {
        /* Find corresponding line location in mask */
        e_mask_line_location = (double)line_index * scale;
        
        /* Make sure mask line location is valid */
        if (e_mask_line_location < 0
            || e_mask_line_location >= number_of_mask_lines)
        {
            IAS_LOG_WARNING("Mask line location outside mask, continuing with"
                " next loop iteration line_index[%d] e_mask_line_location[%d]",
                line_index, e_mask_line_location);
            continue;
        }

        for (sample_index = size;
             sample_index < (number_of_samples - size);
             sample_index++)
        {
            /* Determine the base data location within the input and output
               data */
            data_location = line_index * number_of_samples + sample_index;

            /* Find corresponding sample location in mask */
            e_mask_sample_location = (double)sample_index * scale;
            
            /* Make sure mask sample location is valid */
            if (e_mask_sample_location < 0
                || e_mask_sample_location >= number_of_mask_samples)
            {
                IAS_LOG_WARNING("Mask sample location outside mask,"
                    " continuing with next loop iteration line_index[%d]"
                    " e_mask_line_location[%d]", sample_index,
                    e_mask_sample_location);
                continue;
            }

            /* Find the location in the exclusion mask */
            e_mask_location = e_mask_line_location * number_of_mask_samples
                + e_mask_sample_location;

            /* Is there valid imagery according to the mask? */
            if (exclusion_mask[e_mask_location] != EXCLUSION_MASK_VALID)
            {
                continue;
            }

            /* Initialize invalid flag.
               Any invalid pixel that falls within the interest operator
               window creates a non-useable point.  To avoid that point
               from being used its value will be left as zero. */
            invalid = FALSE;

            /* get diagonal 1 */
            sum_of_sqrs = 0.0;
            for (window_index = -size; window_index <= size; window_index++)
            {
                window_offset = (double)window_index * scale;

                if (exclusion_mask[(e_mask_line_location + window_offset)
                        * number_of_mask_samples + e_mask_sample_location
                        + window_offset]
                    != EXCLUSION_MASK_VALID)
                {
                    invalid = TRUE;
                    break;
                }

                difference = input_data[data_location]
                    - input_data[(line_index + window_index)
                        * number_of_samples + sample_index + window_index];

                sum_of_sqrs += difference * difference;
            }
            /* Don't need to finish if we already know we have a bad
               location in the calculations */
            if (invalid == TRUE)
                continue;

            /* Initialize the minimum value */
            local_minimum_value = sum_of_sqrs;

            /* get diagonal 2 */
            sum_of_sqrs = 0.0;
            for (window_index = -size; window_index <= size; window_index++)
            {
                window_offset = (double)window_index * scale;

                if (exclusion_mask[(e_mask_line_location + window_offset)
                        * number_of_mask_samples + e_mask_sample_location
                        - window_offset]
                    != EXCLUSION_MASK_VALID)
                {
                    invalid = TRUE;
                    break;
                }

                difference = input_data[data_location]
                    - input_data[(line_index + window_index)
                        * number_of_samples + sample_index - window_index];

                sum_of_sqrs += difference * difference;
            }
            /* Don't need to finish if we already know we have a bad
               location in the calculations */
            if (invalid == TRUE)
                continue;

            /* Check if we have a smaller minimum value and save it */
            if (sum_of_sqrs < local_minimum_value)
            {
                local_minimum_value = sum_of_sqrs;
            }

            /* get horizontal */
            sum_of_sqrs = 0.0;
            for (window_index = -size; window_index <= size; window_index++)
            {
                window_offset = (double)window_index * scale;

                if (exclusion_mask[e_mask_location + window_offset]
                    != EXCLUSION_MASK_VALID)
                {
                    invalid = TRUE;
                    break;
                }

                difference = input_data[data_location]
                    - input_data[data_location + window_index];

                sum_of_sqrs += difference * difference;
            }
            /* Don't need to finish if we already know we have a bad
               location in the calculations */
            if (invalid == TRUE)
                continue;

            /* Check if we have a smaller minimum value and save it */
            if (sum_of_sqrs < local_minimum_value)
            {
                local_minimum_value = sum_of_sqrs;
            }

            /* get vertical */
            sum_of_sqrs = 0.0;
            for (window_index = -size; window_index <= size; window_index++)
            {
                window_offset = (double)window_index * scale;

                if (exclusion_mask[(e_mask_line_location + window_offset)
                        * number_of_mask_samples + e_mask_sample_location]
                    != EXCLUSION_MASK_VALID)
                {
                    invalid = TRUE;
                    break;
                }

                difference = input_data[data_location]
                    - input_data[(line_index + window_index)
                        * number_of_samples + sample_index];

                sum_of_sqrs += difference * difference;
            }

            /* Don't need to finish if we already know we have a bad
               location in the calculations */
            if (invalid == TRUE)
                continue;

            /* Check if we have a smaller minimum value and save it */
            if (sum_of_sqrs < local_minimum_value)
            {
                local_minimum_value = sum_of_sqrs;
            }

            /* Assign the minimum of the four directions calculated */
            moravec_buffer[data_location] = local_minimum_value;
        }
    }

    /* Dump intermediate data */
    if (parameters->output_intermediate_data == 1)
    {
        count = snprintf(output_filename, sizeof(output_filename),
            "%s/INTEREST_OPERATOR_%02d.img", 
            parameters->intermediate_output_directory,output_file_number);

        if (count < 0 || count >= sizeof(output_filename))
        {
            IAS_LOG_WARNING("Bounds exceeded or internal snprintf error:"
                " count = %d", count);
        }
        else
        {
            status = ias_misc_write_envi_image(output_filename, NULL, 
                "Interest Operator Image", number_of_lines, 
                number_of_samples, 0.0, 0.0, 0.0, 0.0, "Interest Operator", 
                IAS_R4, moravec_buffer);

            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Unable to generate ENVI image");
                free(moravec_buffer);
                return ERROR;
            }
        }
    }

    /* Initially allocate enough memory to hold 1/8 of the total data size */
    current_memory_allocation = data_size / 8;
    gcps->gcp_locations =
        calloc(current_memory_allocation, sizeof(GENERATE_CHIPS_GCP_LOCATION));
    if (gcps->gcp_locations == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for point data");
        free(moravec_buffer);
        return ERROR;
    }

    /* Initialize the min and max range values */
    minimum_value = DBL_MAX;
    maximum_value = -DBL_MAX;
    /* Initialize the count of all GCPs that could be kept */
    count_of_all_gcps = 0;
    /* Initialize the count of all GCPs that will be kept above the
       threshold */
    gcps->number_of_gcps = 0;
    /* Loop on imagery determining local maximum of interest operator */
    for (line_index = 0; line_index < number_of_lines; line_index++)
    {
        for (sample_index = 0; sample_index < number_of_samples; sample_index++)
        {
            data_location = line_index * number_of_samples + sample_index;

            if (moravec_buffer[data_location] > 0.0)
            {
                if (moravec_buffer[data_location] < minimum_value)
                    minimum_value = moravec_buffer[data_location];
                if (moravec_buffer[data_location] > maximum_value)
                    maximum_value = moravec_buffer[data_location];

                /* If point is well within image perform local maximum search */
                if (line_index >= filter_size
                    && line_index < (number_of_lines - filter_size)
                    && sample_index >= filter_size
                    && sample_index < (number_of_samples - filter_size))
                {
                    /* Search areas to see if current point is the maximum.
                       On purpose we go from -filter_size to filter_size */
                    for (window_line_index = -filter_size;
                         window_line_index <= filter_size
                         && moravec_buffer[data_location] > 0.0;
                         window_line_index++)
                    {
                        for (window_sample_index = -filter_size;
                             window_sample_index <= filter_size;
                             window_sample_index++)
                        {
                            window_index = (line_index + window_line_index)
                                * number_of_samples
                                + sample_index + window_sample_index;

                            if (moravec_buffer[window_index]
                                > moravec_buffer[data_location])
                            {
                                moravec_buffer[data_location] = 0.0;
                                break;
                            }
                        }
                    }
                }
                else
                {
                    moravec_buffer[data_location] = 0.0;
                }

                /* Save the GCPs above the specified threshold to the
                   output */
                if (moravec_buffer[data_location]
                    >= parameters->minimum_interest_operator_threshold)
                {
                    /* Make sure we have enough allocated memory */
                    if (gcps->number_of_gcps
                        == current_memory_allocation)
                    {
                        /* Change memory allocation to hold more data */
                        current_memory_allocation +=
                            (number_of_samples * MEMORY_ADJUSTMENT);

                        /* Reallocate to the new memory size */
                        temp_ptr = realloc(gcps->gcp_locations,
                            sizeof(GENERATE_CHIPS_GCP_LOCATION)
                                * current_memory_allocation);

                        if (temp_ptr == NULL)
                        {
                            IAS_LOG_ERROR(
                                "Reallocating memory for location data");
                            free(gcps->gcp_locations);
                            free(moravec_buffer);
                            gcps->number_of_gcps = 0;
                            gcps->gcp_locations = NULL;
                            moravec_buffer = NULL;
                            return ERROR;
                        }

                        gcps->gcp_locations = temp_ptr;
                    }

                    /* Save the point location information */
                    gcps->gcp_locations[gcps->number_of_gcps]
                        .resolution_line = line_index;
                    gcps->gcp_locations[gcps->number_of_gcps]
                        .resolution_sample = sample_index;

                    gcps->gcp_locations[gcps->number_of_gcps]
                        .input_data_line = (double)line_index * scale;
                    gcps->gcp_locations[gcps->number_of_gcps]
                        .input_data_sample = (double)sample_index * scale;

                    /* Save the moravec value */
                    gcps->gcp_locations[gcps->number_of_gcps]
                        .moravec_value = moravec_buffer[data_location];

                    /* Initialize the valid flag */
                    gcps->gcp_locations[gcps->number_of_gcps]
                        .valid_flag = GCP_VALID_FLAG_MORAVEC_GOOD;

                    gcps->number_of_gcps++;
                } /* END threshold check */

                count_of_all_gcps++;
            } /* END moravec value was > 0 */
        } /* END sample loop */
    } /* END line loop */

    /* Output some information about what we have */
    IAS_LOG_INFO("Bounds on interest operator %lf, %lf",
        minimum_value, maximum_value);
    IAS_LOG_INFO("Total possible GCPs %d ---- Keeping %d GCPs",
        count_of_all_gcps, gcps->number_of_gcps);
    IAS_LOG_INFO("Interest operator values of zero %d",
        data_size - count_of_all_gcps);

    /* Reduce the memory usage down to the number of GCPs */
    if (gcps->number_of_gcps < current_memory_allocation
        && gcps->number_of_gcps != 0)
    {
        temp_ptr = realloc(gcps->gcp_locations,
            sizeof(GENERATE_CHIPS_GCP_LOCATION) * gcps->number_of_gcps);

        if (temp_ptr == NULL)
        {
            IAS_LOG_ERROR("Reallocating memory for location data");
            free(gcps->gcp_locations);
            free(moravec_buffer);
            gcps->number_of_gcps = 0;
            gcps->gcp_locations = NULL;
            moravec_buffer = NULL;
            return ERROR;
        }

        gcps->gcp_locations = temp_ptr;
    }

    /* Dump intermediate data */
    if (parameters->output_intermediate_data == 1)
    {
        count = snprintf(output_filename, sizeof(output_filename),
            "%s/MAXIMUM_INTEREST_OPERATOR_MASK_%02d.img",
            parameters->intermediate_output_directory, output_file_number);

        if (count < 0 || count >= sizeof(output_filename))
        {
            IAS_LOG_WARNING("Bounds exceeded or internal snprintf error:"
                " count = %d", count);
        }
        else
        {
            if (ias_misc_write_envi_image(output_filename, NULL, 
                "Maximum Interest Operator Mask", number_of_lines,
                number_of_samples, 0.0, 0.0, 0.0, 0.0, 
                "Maximum Interest Operator Mask", IAS_R4, moravec_buffer) 
                != SUCCESS)
            {
                IAS_LOG_ERROR("Writing maximum interest operator mask to file");
                free(gcps->gcp_locations);
                free(moravec_buffer);
                gcps->number_of_gcps = 0;
                gcps->gcp_locations = NULL;
                moravec_buffer = NULL;
                return ERROR;
            }
        }

        /* Increment the output file number for the current iteration */
        output_file_number++;
    }

    free(moravec_buffer);

    return SUCCESS;
}

