
/* Standard C Includes */
#include <stdlib.h>

/* IAS Includes */
#include "ias_const.h"
#include "ias_types.h"
#include "ias_logging.h"

/* Local Includes */
#include "resample_data.h"

/*****************************************************************************
 The following are the defines used by the resampling
*****************************************************************************/

/* This performs a mirroring of the data for the samples on the edge of the
   image data */
#define CALCULATE_RESAMPLE_LOCATION(location, offset, size, result) \
    if (location + offset < 0) \
        result = (-location - offset); \
    else if (location + offset >= size) \
        result = (2 * size - location - offset - 1); \
    else \
        result = (location + offset);


/*****************************************************************************
 NAME: perform_resample

 PURPOSE: Implements a cubic-convolution algorithm for converting data to a
          different resolution.

 RETURN VALUE: Type = int
    Value      Description
    ---------  --------------------------
    SUCCESS    Successful completion
    ERROR      Operation failed

*****************************************************************************/
int perform_resample
(
    double o2i_scaling_factor,    /* I: Output to input scaling factor */
    int number_of_weights,        /* I: Number of resampling weights */
    int number_of_input_lines,    /* I: Number of lines in the input data */
    int number_of_input_samples,  /* I: Number of samples in the input data */
    const IAS_R4_TYPE *input_data,/* I: The input data to scale */
    int *number_of_output_lines,  /* O: Number of lines in the output data */
    int *number_of_output_samples,/* O: Number of samples in the output data */
    IAS_R4_TYPE *output_data      /* O: The results of the scaling */
)
{
    int bin_line_index;      /* Looping control */
    int bin_sample_index;    /* Looping control */
    int kernel_line_index;   /* Looping control */
    int kernel_sample_index; /* Looping control */
    int output_line_index;   /* Looping control */
    int output_sample_index; /* Looping control */

    int weight_location;     /* Location of a weight within a bin */
    double *weights;         /* Resampling weights */
    double *bin_weights;     /* Resampling weights */
    double bin_weight_sum;   /* Sum for applying weights to input pixels */

    int output_data_sample; /* Location of the current sample in the output
                               data */
    IAS_R4_TYPE input_sample_value; /* Input sample value */
    int input_value_line;   /* Line location for the input data value */
    int input_value_sample; /* Sample location for the input data value */

    double input_line_location;   /* Floating point input line location */
    double input_sample_location; /* Floating point input sample location */
    int input_line;               /* Integer input line location */
    int input_sample;             /* Integer input sample location */
    double input_line_delta;      /* Fractional offset in line direction */
    double input_sample_delta;    /* Fractional offset in sampl direction */

    int bin_location;          /* The location of the weights to use */
    int number_of_bin_samples; /* The number of samples in a bin line */
    int kernel_size;           /* Resampling kernel size */
    int size;                  /* Size of resampling kernel */

    /* Used to force a rounding to the closest cell in the kernel */
    double half_kernel_step = 1.0 / (2.0 * SUBPIXEL_STEPS);

    /* Generate the weights table */
    calculate_resampling_weights(number_of_weights, &kernel_size,
        &number_of_bin_samples, &weights);

    /* Determine new image size after upsampling */
    *number_of_output_lines = number_of_input_lines / o2i_scaling_factor;
    *number_of_output_samples = number_of_input_samples / o2i_scaling_factor;

    /* Resample input image */
    size = (number_of_weights - 1) / 2;
    for (output_line_index = 0;
         output_line_index < (*number_of_output_lines);
         output_line_index++)
    {
        /* Determine input line location and fractional offset */
        input_line_location = output_line_index * o2i_scaling_factor;
        input_line = (int)input_line_location;
        input_line_delta = input_line_location - input_line;

        /* The half_kernel_step here forces a rounding to the closest cell in
           the kernel */
        bin_line_index = (input_line_delta + half_kernel_step) * SUBPIXEL_STEPS;

        for (output_sample_index = 0;
             output_sample_index < (*number_of_output_samples);
             output_sample_index++)
        {
            /* Determine input sample location and fractional offset */
            input_sample_location = output_sample_index * o2i_scaling_factor;
            input_sample = (int)input_sample_location;
            input_sample_delta = input_sample_location - input_sample;

            /* The half_kernel_step here forces a rounding to the closest cell
               in the kernel */
            bin_sample_index = (input_sample_delta + half_kernel_step)
                * SUBPIXEL_STEPS;

            /* Determine set of resampling weights to use */
            bin_location = (bin_line_index * number_of_bin_samples)
                + bin_sample_index * kernel_size;
            bin_weights = &(weights[bin_location]);

            /* Apply weights to appropriate input pixels */
            bin_weight_sum = 0.0;
            weight_location = 0;
            for (kernel_line_index = -size;
                 kernel_line_index <= size;
                 kernel_line_index++)
            {
                CALCULATE_RESAMPLE_LOCATION(input_line, kernel_line_index,
                    number_of_input_lines, input_value_line);

                for (kernel_sample_index = -size;
                     kernel_sample_index <= size;
                     kernel_sample_index++)
                {
                    CALCULATE_RESAMPLE_LOCATION(input_sample,
                        kernel_sample_index, number_of_input_samples,
                        input_value_sample);

                    input_sample_value = input_data[input_value_line
                        * number_of_input_samples + input_value_sample];

                    bin_weight_sum += input_sample_value
                        * bin_weights[weight_location];

                    weight_location++;
                }
            }

            /* Determine output pixel location */
            output_data_sample = output_line_index
                * (*number_of_output_samples) + output_sample_index;

            /* We only check for less than 0.0, so that the fill frame around
               the image is kinda maintained.
               Since we are going to throw away the results of this resample,
               we do not cap the data on the top end. */
            if (bin_weight_sum < 0.0)
                output_data[output_data_sample] = 0;
            else
                output_data[output_data_sample] = bin_weight_sum;
        }
    }

    /* Free allocated memory */
    free(weights);

    return SUCCESS;
}

