/* Standard C Includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <float.h>

/* IAS Includes */
#include "ias_const.h"
#include "ias_db.h"
#include "ias_db_query.h"
#include "ias_db_get_connect_info.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"

/* Local Includes */
#include "generate_chips.h"

/* Local Defines */
#define GCPLIB_NAME "GCPLib"
#define PASS_NUMBER 2
#define SOURCE_DB_SIZE 10

/* Local Variable Declarations */
static int local_chip_size[2];                  /* Chip size */
static char local_generation_method[26];        /* Generation Method */
static char local_moravec_source_resolution[14];/* Source resolution */
static int local_cloud_border[2];               /* Cloud Border array */
static int local_number_of_regions[2];          /* Regions array */
static double local_range_values[2];            /* Range array */
static char local_input_gcplib[PATH_MAX];       /* Input GCPLib directory */
static int local_minimum_distance[2];           /* Minimum distance array */

/******************************************************************************
NAME: run_starting_id_query

PURPOSE: Helper for get_starting_id_from_database. It runs the actual query.

RETURN VALUE: Type = int
    Value     Description
    -----     -----------
    SUCCESS   The ID was retrieved succesfully
    ERROR     An error occurred 
******************************************************************************/
int run_starting_id_query
(
    struct ias_db_connection *db, /* I: DB connection */
    int wrs_path,                 /* I: WRS Path get the starting ID for */
    int wrs_row,                  /* I: WRS Row get the starting ID for */
    const char *source_database,  /* I: GCP database source */
    int *starting_id              /* O: The starting ID from the DB */
)
{
    int status;
    struct ias_db_query *query;
    char sql[1024];
    char table_name[1024];
    int local_int;
    const char* processing_instance = getenv("IAS_INSTANCE_NAME");

    /* Check that the processing instance is set */
    if (!processing_instance)
    {
        IAS_LOG_ERROR("Unable to get the IAS_INSTANCE_NAME environment "
            "variable");
        return ERROR;
    }

    /* Retrieve the correct GCP table name to get the starting id */
    status  = snprintf(sql, sizeof(sql),
        "select value from processing_instance_info "
        "where processing_instance = '%s' "
        "and attribute_name = 'GROUND_CONTROL_POINTS_%s'", processing_instance,
        source_database);
    if (status < 0 || status >= sizeof(sql))
    {
        IAS_LOG_ERROR("Unable to create the processing_instance_info query");
        return ERROR;
    }

    /* Prepare the query */
    query = ias_db_prepare_query(db, sql);
    if (query == NULL)
    {
        const char *db_error_message = ias_db_connect_last_error(db);
        IAS_LOG_ERROR("Preparing failed: %s", sql);
        IAS_LOG_ERROR(db_error_message);
        return ERROR;
    }

    /* Execute the query */
    if (ias_db_exec_prepared_query(query) != SUCCESS)
    {
        char db_error_message[300];
        IAS_LOG_ERROR("Failed executing the processing_instance_info query");
        ias_db_query_get_error_message(query, db_error_message,
            sizeof(db_error_message));
        IAS_LOG_ERROR(db_error_message);
        ias_db_query_close(query);
        return ERROR;
    }

    if (!ias_db_query_next(query))
    {
        IAS_LOG_ERROR("No rows returned from the processing_instance_info "
            "query");
        ias_db_query_close(query);
        return ERROR;
    }

    /* Ensure that NULL is not returned as a value */
    ias_db_query_string_value(query, 0, table_name, sizeof(table_name));
    if (strcmp(table_name, "NULL") == 0)
    {
        IAS_LOG_ERROR("Getting the table name from the "
            "processing_instance_info query");
        ias_db_query_close(query);
        return ERROR;
    }

    /* This query gets the largest id (last 4 digits) for the specified path /
       row. If no records are found, coulumn value is -1 */   
    status  = snprintf(sql, sizeof(sql),
                 "SELECT NVL(MAX(SUBSTR(ground_control_points_id, -4)), -1) "
                 "max_id FROM %s WHERE wrs_path = %d AND "
                 "wrs_row = %d", table_name, wrs_path, wrs_row);
    if (status < 0 || status >= sizeof(sql))
    {
        IAS_LOG_ERROR("Ran out of buffer space for sql");
        return ERROR;
    }

    /* Prepare the query */
    query = ias_db_prepare_query(db, sql);
    if (query == NULL)
    {
        const char *db_error_message;   

        db_error_message = ias_db_connect_last_error(db);
        IAS_LOG_ERROR("Preparing failed: %s", sql);
        IAS_LOG_ERROR(db_error_message);
        return ERROR;
    }

    /* Execute the query */
    if (ias_db_exec_prepared_query(query) != SUCCESS)
    {
        char db_error_message[300];   

        IAS_LOG_ERROR("Failed executing prepared ID query");
        ias_db_query_get_error_message(query, db_error_message,
            sizeof(db_error_message));
        IAS_LOG_ERROR(db_error_message);
        ias_db_query_close(query);
        return ERROR;
    }

    if (!ias_db_query_next(query))
    {
        IAS_LOG_ERROR("No rows returned from ID query");
        ias_db_query_close(query);
        return ERROR;
    }
    status = ias_db_query_int_value(query, &local_int, 0);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Obtaining max ID failed");
        ias_db_query_close(query);
        return ERROR;
    }

    /* Only have 4 digits to "work with", so if we found 9999 as the max
       we can't go higher */
    if (local_int >= 9999)
    {
        IAS_LOG_ERROR("Out of IDs for path %d row %d", wrs_path, wrs_row);
        ias_db_query_close(query);
        return ERROR;
    }
    /* Query returns -1 when no max could be found */
    else if (local_int >= 0)
    {
        /* Set the starting ID one higher than the previous max */
        *starting_id = local_int + 1;
    }
    else
    {
        IAS_LOG_WARNING("No records for path: %d row: %d"
                        ". Using 1 as starting ID.", wrs_path, wrs_row);
        *starting_id = 1;
    }

    ias_db_query_close(query);
    return SUCCESS;
}

/******************************************************************************
NAME: get_starting_id_from_database

PURPOSE: This function connects to the database and gets the next value to
         use as the GCP starting ID.

RETURN VALUE: Type = int
    Value     Description
    -----     -----------
    SUCCESS   The ID was retrieved succesfully
    ERROR     An error occurred 
******************************************************************************/
int get_starting_id_from_database
(
    int wrs_path,     /* I: WRS Path get the starting ID for */
    int wrs_row,      /* I: WRS Row get the starting ID for */
    const char *source_database, /* I: GCP database source */
    int *starting_id  /* O: The starting ID from the DB */
)
{
    char *database_name = NULL;
    struct ias_db_connection *db;
    char *user_name = NULL;
    char *password = NULL;
    char *host = NULL;
    int status;

    status = ias_db_initialize_database_lib();
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the database library");
        return ERROR;
    }

    /* Get the database connection information */
    status = ias_db_get_common_connect_info(&database_name, &user_name,
        &password, &host);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting database connection info");
        return ERROR;
    }

    /* Connect to the database */
    db = ias_db_connect_to_database(database_name, user_name, password,
        host);

    /* Free the resources before checking the connection */
    free(database_name);
    database_name = NULL;
    free(user_name);
    user_name = NULL;
    free(password);
    password = NULL;
    free(host);
    host = NULL;

    /* Check the connection */
    if (!db)
    {
        IAS_LOG_ERROR("Connecting to database: %s",
            ias_db_connect_last_error(NULL));
        ias_db_close_database_lib();
        return ERROR;
    }

    /* Get the starting ID */
    status = run_starting_id_query(db, wrs_path, wrs_row, source_database,
        starting_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Failed to get starting ID from database for path: "
                      "%d row: %d\n", wrs_path, wrs_row);
        ias_db_close_connection(db);
        ias_db_close_database_lib();
        return ERROR;
    }

    /* Close the DB connection and DB library */
    ias_db_close_connection(db);
    ias_db_close_database_lib();

    return SUCCESS;
}

/******************************************************************************
NAME: process_parameters

PURPOSE: This function performs special processing on parameters which need
         to be translated after being read from the parameter or OMF files.

RETURN VALUE: Type = int
    Value     Description
    -----     -----------
    SUCCESS   The parameters were successfully processed
    ERROR     An error occurred processing the parameters
******************************************************************************/
int process_parameters
(
    const char *source_database,          /* I: GCP database source */
    GENERATE_CHIPS_PARAMETERS *parameters /* I/O: Generation parameters */
)
{
    int status;                    /* Status variable */
    int issue_min_gcp_msg = FALSE; /* GCP message should be issued flag */

    /* Check for the elevation filename */
    parameters->elevation_filename_present = FALSE;
    if (strlen(parameters->elevation_filename) != 0)
    {
        parameters->elevation_filename_present = TRUE;
    }

    /* Check for the ENVI point filename */
    parameters->envi_point_filename_present = FALSE;
    if (strlen(parameters->envi_point_filename) != 0)
    {
        parameters->envi_point_filename_present = TRUE;
    }

    /* Place the local range values into the parameters structure */
    parameters->min_range_value = local_range_values[0];
    parameters->max_range_value = local_range_values[1];

    /* Place the local number of regions values into the parameters
        structure */
    parameters->number_of_regions_lines = local_number_of_regions[0];
    parameters->number_of_regions_samples = local_number_of_regions[1];

    /* Place the local cloud border lines and samples into the parameters 
       structure */
    parameters->cloud_border_lines = local_cloud_border[0];
    parameters->cloud_border_samples = local_cloud_border[1];

    /* Place the local chip size values into the parameters structure */
    parameters->chip_size_lines = local_chip_size[0];
    parameters->chip_size_samples = local_chip_size[1];
    
    /* Determine the minimum line/sample distances */
    parameters->minimum_distance_lines = parameters->chip_size_lines;
    if (local_minimum_distance[0] != 0)
    {
        parameters->minimum_distance_lines = local_minimum_distance[0];
        issue_min_gcp_msg = TRUE;
    }

    parameters->minimum_distance_samps = parameters->chip_size_samples;
    if (local_minimum_distance[1] != 0)
    {
        parameters->minimum_distance_samps = local_minimum_distance[1];
        issue_min_gcp_msg = TRUE;
    }

    /* Issue the minimum GCP distance message is needed */
    if (issue_min_gcp_msg)
    {
        IAS_LOG_INFO("Minimum GCP distance used (line, sample) = (%d, %d)",
            parameters->minimum_distance_lines, 
            parameters->minimum_distance_samps);
    }

    /* Convert the local resolution string to enumeration */
    if (strcmp(local_moravec_source_resolution, "NOMINAL") == 0)
    {
        parameters->moravec_source_resolution = MSR_NOMINAL;
    }
    else if (strcmp(local_moravec_source_resolution, "UPSAMPLED") == 0)
    {
        parameters->moravec_source_resolution = MSR_UPSAMPLED;
    }
    else if (strcmp(local_moravec_source_resolution, "DOWNSAMPLED") == 0)
    {
        parameters->moravec_source_resolution = MSR_DOWNSAMPLED;
    }
    else
    {
        parameters->moravec_source_resolution = MSR_FEWEST_POINTS;
    }

    /* Convert the generation method from string to enumeration */
    if (strcmp(local_generation_method, "EVENLY_SPACED") == 0)
    {
        parameters->generation_method = GM_EVENLY_SPACED;
    }
    else if (strcmp(local_generation_method, "GCPLIB") == 0)
    {
        parameters->generation_method = GM_GCPLIB;
    }
    else if (strcmp(local_generation_method, "ENVI_ROI") == 0)
    {
        parameters->generation_method = GM_ENVI_ROI;
    }
    else if (strcmp(local_generation_method, "MORAVEC") == 0)
    {
        parameters->generation_method = GM_MORAVEC;
    }
    else if (strcmp(local_generation_method, "MORAVEC_AND_EVENLY_SPACED") == 0)
    {
        parameters->generation_method = GM_MORAVEC_AND_EVENLY_SPACED;
    }
    else
    {
        IAS_LOG_ERROR("Invalid GENERATION_METHOD specified in the input"
            " parameters file.  Valid values are EVENLY_SPACED, GCPLIB,"
            " MORAVEC, and MORAVEC_AND_EVENLY_SPACED.");
        return ERROR;
    }

    /* If -1 is passed as the starting ID, this indicates we should find the
       max ID from the DB */
    if (parameters->starting_id == -1)
    {
        int starting_id;
        status = get_starting_id_from_database(parameters->target_wrs_path,
                                               parameters->target_wrs_row,
                                               source_database, &starting_id);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Failed to get starting ID from database for path: "
                          "%d row: %d\n",parameters->target_wrs_path,
                          parameters->target_wrs_row);
            return ERROR;
        }

        /* Save the determined starting id in the structure */
        parameters->starting_id = starting_id;
    }

    /* Perform verification based on the generation method */
    switch (parameters->generation_method)
    {
        case GM_GCPLIB:
        {
            if (strlen(local_input_gcplib) == 0)
            {
                IAS_LOG_ERROR("Required parameter INPUT_GCPLIB"
                    " not provided");
                return ERROR;
            }

            /* Build GCPLib file location */
            status = snprintf(parameters->input_gcplib, 
                sizeof(parameters->input_gcplib), "%s/%s", local_input_gcplib, 
                GCPLIB_NAME);
            if (status < 0 || status >= sizeof(parameters->input_gcplib))
            {
                IAS_LOG_ERROR("Creating the GCPLib file path");
                return ERROR;
            }
            break;
        }
        case GM_ENVI_ROI:
        {
            if (strlen(parameters->envi_roi_filename) == 0)
            {
                IAS_LOG_ERROR("Required parameter ENVI_ROI_FILENAME "
                    "not provided");
                return ERROR;
            }
            break;
        }
        case GM_MORAVEC:
        case GM_MORAVEC_AND_EVENLY_SPACED:
        {
            /* Verify that the interest operator size is smaller than the
               chip size */
            if ((parameters->interest_operator_size
                 >= parameters->chip_size_lines)
                 || (parameters->interest_operator_size
                 >= parameters->chip_size_samples))
            {
                IAS_LOG_ERROR("INTEREST_OPERATOR_SIZE (%d) must be smaller"
                    " than CHIP_SIZE (%d, %d)",
                    parameters->interest_operator_size,
                    parameters->chip_size_lines,
                    parameters->chip_size_samples);
                    return ERROR;
            }

            if (parameters->generation_method == GM_MORAVEC)
            {
                /* Nothing special to check */
                break;
            }

                /* GM_MORAVEC_AND_EVENLY_SPACED falls through to
                   GM_EVENLY_SPACED */
        }

        case GM_EVENLY_SPACED:
        {
            /* Nothing special to check */
            break;
        }

        default:
            /* Shouldn't get here */
            break;

    } /* END switch */

    IAS_LOG_INFO("Creating chips using the \"%s\" generation method",
        local_generation_method);

    return SUCCESS;
}

/******************************************************************************
NAME: read_omf_parameters

PURPOSE: This function reads all the parameters from the OMF file that 
         generate chips uses.

RETURN VALUE: Type = int
    Value    Description
    -----    -----------
    SUCCESS  The parameters were successfully read
    ERROR    An error occurred reading the parameters
******************************************************************************/
int read_omf_parameters
(
    const char *omf_filename,             /* I: OMF filename */
    GENERATE_CHIPS_PARAMETERS *parameters /* O: Generation parameters */
)
{
    int count;                /* Table parameter count */
    int help_requested;       /* Status placeholder */

    /* Start parameter table definition */
    IAS_PARM_DECLARE_TABLE(omf_table, 6);

    IAS_PARM_TARGET_WRS_PATH(omf_table, IAS_PARM_OPTIONAL,
        &parameters->target_wrs_path, sizeof(parameters->target_wrs_path), 0);

    IAS_PARM_TARGET_WRS_ROW(omf_table, IAS_PARM_OPTIONAL,
        &parameters->target_wrs_row, sizeof(parameters->target_wrs_row), 0);

    IAS_PARM_DEM_FILENAME_FROM_PASS(omf_table, IAS_PARM_OPTIONAL, 
        parameters->elevation_filename, sizeof(parameters->elevation_filename), 
        0, PASS_NUMBER);

    /* Initialize quality filename to empty string */
    strcpy(parameters->quality_filename, "");
    IAS_PARM_QUALITY_BAND_FILENAME(omf_table, IAS_PARM_OPTIONAL,
        parameters->quality_filename, sizeof(parameters->quality_filename));
    
    IAS_PARM_L1G_FILENAME_FROM_PASS(omf_table, IAS_PARM_REQUIRED,
        parameters->l1g_filename, sizeof(parameters->l1g_filename), 0, 
        PASS_NUMBER);

    IAS_PARM_GCP_DIRECTORY(omf_table, IAS_PARM_OPTIONAL,
        local_input_gcplib, sizeof(local_input_gcplib));

    /* Retrieve parameter count for OMF table */
    count = IAS_PARM_GET_TABLE_SIZE(omf_table);

    /* Call the ias_parm_provide_help routine to provide help for the
       parameters if requested; also verifies the OMF file exists */
    help_requested = ias_parm_provide_help(omf_filename, omf_table, count,
            IAS_OMF_PARAMETERS);
    if (help_requested)
    {
        /* Either help was provided or an error occurred*/
        if (help_requested == TRUE)
        {
            /* Successfully provided help, as requested, so just exit */
            exit(EXIT_SUCCESS);
        }
        
        IAS_LOG_ERROR("Error occurred providing OMF help");
        return ERROR;
    }

    /* Read the OMF file */
    if (ias_parm_read(omf_filename, omf_table, count) != SUCCESS)
    {
        IAS_LOG_ERROR("Error reading OMF parameter file %s", omf_filename);

        return ERROR;
    }

    return SUCCESS;
}

/*****************************************************************************
 NAME: read_parameters

 PURPOSE: Read parameters from the specified input ODL file and OMF file.

 RETURN VALUE: Type = int
    Value    Description
    -------  --------------------------
    SUCCESS  Successful completion
    ERROR    Operation failed

*****************************************************************************/
int read_parameters
(
    const char *parameter_filename,         /* I: Parameter file name */
    GENERATE_CHIPS_PARAMETERS *parameters   /* O: Generate Chips parameters */
)
{
    int help_requested;                     /* Help request variable */
    int count;                              /* Table parameter count */
    int status;                             /* Status variable */
    char local_omf_filename[PATH_MAX];      /* OMF filename */
    char local_wo_id[IAS_WORKORDER_ID_SIZE];/* Work order ID */
    char local_source_database[SOURCE_DB_SIZE]; /* Source database for GCPs */

    /* Zero the parameters */
    memset(parameters, 0, sizeof(GENERATE_CHIPS_PARAMETERS));

    /* Parameters table definition */
    IAS_PARM_DECLARE_TABLE(parameter_table, 37);

    /* -------------------------------------------------------------------- */
    /* Common parameters table definition for the parameter file */

    /* Get the work order ID */
    IAS_PARM_WORK_ORDER_ID(parameter_table, local_wo_id, 
        sizeof(local_wo_id), 1);

    const char *valid_generation_methods[] =
        {"EVENLY_SPACED", "GCPLIB", "ENVI_ROI", "MORAVEC",
         "MORAVEC_AND_EVENLY_SPACED"};
    IAS_PARM_ADD_STRING(parameter_table, GENERATION_METHOD,
        "Name of the generation method",
        IAS_PARM_REQUIRED, 5, valid_generation_methods, 0, NULL,
        local_generation_method, sizeof(local_generation_method), 0);

    const int default_band[] = {6};
    IAS_PARM_ADD_INT(parameter_table, BAND_NUMBER,
        "Defines the band to use for generating the chips",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        1, 1, 11, 1, default_band,
        &parameters->band_number, sizeof(parameters->band_number), 0);

    const char *valid_gcp_update_reasons[] =
        {"ORIGINAL", "TRIG_ERROR", "NEW_POINT", "TEMPORAL", "SEASONAL"};
    IAS_PARM_ADD_STRING(parameter_table, GCP_UPDATE_REASON,
        "The reason for the latest update to the GCPs. "
        "Ignored by the GCPLIB generation method", IAS_PARM_REQUIRED, 5,
        valid_gcp_update_reasons, 0, NULL, parameters->gcp_update_reason, 
        sizeof(parameters->gcp_update_reason), 0);

    const char *valid_gcp_sources[] = {"GLS", "DOQ", "GRI", "TM6"};
    const char *default_gcp_source[] = {"GLS"};
    IAS_PARM_ADD_STRING(parameter_table, GCP_SOURCE,
        "The GCP data source.", IAS_PARM_OPTIONAL, 4,
        valid_gcp_sources,  1, default_gcp_source, parameters->gcp_source, 
        sizeof(parameters->gcp_source), 0);

    /* Get the water mask filename.  Note that the name is a bit misleading
       since the polygons in the file are polygons defining the land masses.
       It would be clearer to name it the LAND_MASS_FILENAME, but the precedent
       has been set already by L1-7. */
    IAS_PARM_ADD_STRING(parameter_table, WATER_MASK_FILENAME,
        "Polygon file defining the land masses in the world to override the "
        "default file used to create the water mask. When water mask flag "
        "has been set to binary water mask then used as binary water mask "
        "filename", IAS_PARM_OPTIONAL, 0, NULL, 0, NULL, 
        parameters->water_mask_filename, 
        sizeof(parameters->water_mask_filename), 0);

    const int default_use_water_mask_flag[] = {1};
    IAS_PARM_ADD_INT(parameter_table, USE_WATER_MASK_FLAG, 
        "Flag to use the water mask [0 = don't use water mask; 1 = use water "
        "mask; 2 = use binary water mask file]", 
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 0, 2, 1,
        default_use_water_mask_flag, &parameters->use_water_mask_flag, 
        sizeof(parameters->use_water_mask_flag), 0);

    IAS_PARM_ADD_STRING(parameter_table, OUTPUT_GCPLIB,
        "Name of the output GCPLib file",
        IAS_PARM_REQUIRED,
        0, NULL, 0, NULL,
        parameters->output_gcplib, sizeof(parameters->output_gcplib), 0);

    IAS_PARM_ADD_STRING(parameter_table, OUTPUT_DIRECTORY,
        "Name of the output directory",
        IAS_PARM_REQUIRED,
        0, NULL, 0, NULL,
        parameters->output_directory, sizeof(parameters->output_directory), 0);

    const char *valid_gcp_type[] = {"ABS", "REL"};
    const char *default_gcp_type[] = {"ABS"};
    IAS_PARM_ADD_STRING(parameter_table, ABSOLUTE_OR_RELATIVE,
        "Create absolute or relative GCPs",
        IAS_PARM_OPTIONAL, 2, valid_gcp_type, 1, default_gcp_type,
        parameters->absolute_or_relative,
        sizeof(parameters->absolute_or_relative), 0);

    const double default_range_values[] = {1.0, 65535.0};
    IAS_PARM_ADD_DOUBLE(parameter_table, RANGE,
        "Range of valid image pixels: (min, max)",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY, 1, 0, 65535, 2,
        default_range_values, local_range_values, sizeof(local_range_values),
        2);

    IAS_PARM_ADD_DOUBLE(parameter_table, VALID_PIXEL_THRESHOLD,
        "Minimum fraction of pixels in the image chip that must"
        " fall in the valid RANGE to include the chip",
        IAS_PARM_REQUIRED, IAS_PARM_NOT_ARRAY,
        1, 0.0, 1.0, 0, NULL,
        &parameters->valid_pixel_threshold,
        sizeof(parameters->valid_pixel_threshold), 1);

    IAS_PARM_ADD_DOUBLE(parameter_table, VALID_WATER_THRESHOLD,
        "Maximum fraction of pixels in the image chip identified"
        " as water before excluding the chip",
        IAS_PARM_REQUIRED, IAS_PARM_NOT_ARRAY,
        1, 0.0, 1.0, 0, NULL,
        &parameters->valid_water_threshold,
        sizeof(parameters->valid_water_threshold), 1);

    IAS_PARM_ADD_DOUBLE(parameter_table, VALID_SNOW_AND_ICE_THRESHOLD,
        "Maximum fraction of pixels in the image chip identified"
        " as snow or ice before excluding the chip",
        IAS_PARM_REQUIRED, IAS_PARM_NOT_ARRAY,
        1, 0.0, 1.0, 0, NULL,
        &parameters->valid_snow_and_ice_threshold,
        sizeof(parameters->valid_snow_and_ice_threshold), 1);

    const int default_exclude_cirrus[] = {0};
    IAS_PARM_ADD_INT(parameter_table, EXCLUDE_CIRRUS,
        "Minimum cirrus cloud confidence level to exclude 0 = none, 1 = low, "
        "2 = medium, and 3 = high", IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 0,
        3, 1, default_exclude_cirrus, &parameters->exclude_cirrus,
        sizeof(parameters->exclude_cirrus), 0);

    const int default_exclude_cloud[] = {3};
    IAS_PARM_ADD_INT(parameter_table, EXCLUDE_CLOUD,
        "Minimum cloud confidence level to exclude 0 = none, 1 = low, 2 = "
        "medium, and 3 = high", IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 0,
        3, 1, default_exclude_cloud, &parameters->exclude_cloud, 
        sizeof(parameters->exclude_cloud), 0);

    const int default_exclude_snow_and_ice[] = {0};
    IAS_PARM_ADD_INT(parameter_table, EXCLUDE_SNOW_AND_ICE,
        "Minimum snow and ice confidence level to exclude 0 = none, 1 = low, "
        "2 = medium, and 3 = high", IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 
        0, 3, 1, default_exclude_snow_and_ice, 
        &parameters->exclude_snow_and_ice, 
        sizeof(parameters->exclude_snow_and_ice), 0);

    const int default_cloud_border[] = {0, 0};
    IAS_PARM_ADD_INT(parameter_table, CLOUD_BORDER,
        "Number of pixels to grow the clouds in the (line, sample) dimensions",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY, 1, 0, 1000, 2,
        default_cloud_border, local_cloud_border,
        sizeof(local_cloud_border), 2);

    const int default_starting_id[] = {0};
    IAS_PARM_ADD_INT(parameter_table, STARTING_ID,
        "Defines starting number for chip index. -1 indicates the next "
        "ID should be determined from the database",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        1, -1, 9999, 1, default_starting_id,
        &parameters->starting_id, sizeof(parameters->starting_id), 0);

    const char *default_source_database[] = {"PRIMARY"};
    const char *valid_source_database[] = {"PRIMARY"};
    IAS_PARM_ADD_STRING(parameter_table, GCP_SOURCE_DATABASE,
        "The source database to use for GCPs", IAS_PARM_OPTIONAL, 1,
        valid_source_database, 1, default_source_database,
        local_source_database, sizeof(local_source_database), 0);

    IAS_PARM_ADD_STRING(parameter_table, ENVI_POINT_FILENAME,
        "Name of the output ENVI point file",
        IAS_PARM_OPTIONAL,
        0, NULL, 0, NULL,
        parameters->envi_point_filename,
        sizeof(parameters->envi_point_filename), 0);

    /* ENVI ROI PARAMETERS ------------------------------------------- */

    IAS_PARM_ADD_STRING(parameter_table, ENVI_ROI_FILENAME,
        "Name of the ENVI ROI file", IAS_PARM_OPTIONAL, 0, NULL, 0, NULL,
        parameters->envi_roi_filename, 
        sizeof(parameters->envi_roi_filename), 0);

    /* EVENLY SPACED PARAMETERS ------------------------------------------- */

    const int default_gcp_grid_density_rows[] = {20};
    IAS_PARM_ADD_INT(parameter_table, GCP_GRID_DENSITY_ROWS,
        "Defines how many rows of GCP points to generate "
        "when using the EVENLY_SPACED algorithm",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        1, 1, 2048, 1, default_gcp_grid_density_rows,
        &parameters->gcp_grid_density_rows,
        sizeof(parameters->gcp_grid_density_rows), 0);

    const int default_gcp_grid_density_columns[] = {20};
    IAS_PARM_ADD_INT(parameter_table, GCP_GRID_DENSITY_COLUMNS,
        "Defines how many columns of GCP points to generate "
        "when using the EVENLY_SPACED algorithm",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        1, 1, 2048, 1, default_gcp_grid_density_columns,
        &parameters->gcp_grid_density_columns,
        sizeof(parameters->gcp_grid_density_columns), 0);

    /* MORAVEC PARAMETERS ------------------------------------------------- */

    const char *default_moravec_source_resolution[] = {"NOMINAL"}; 
    const char *valid_source_resolutions[] =
        {"NOMINAL", "UPSAMPLED", "DOWNSAMPLED", "FEWEST_POINTS"};
    IAS_PARM_ADD_STRING(parameter_table, MORAVEC_SOURCE_RESOLUTION,
        "Resolution used to pick points for the moravec generation method",
        IAS_PARM_OPTIONAL, 4, valid_source_resolutions, 1, 
        default_moravec_source_resolution, local_moravec_source_resolution, 
        sizeof(local_moravec_source_resolution), 0);

    const int default_interest_operator_size[] = {5};
    IAS_PARM_ADD_INT(parameter_table, INTEREST_OPERATOR_SIZE,
        "Distance from interest point for window creation",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 2, 64, 1,
        default_interest_operator_size, &parameters->interest_operator_size,
        sizeof(parameters->interest_operator_size), 0);

    const int default_number_of_regions[] = {3, 3};
    IAS_PARM_ADD_INT(parameter_table, NUMBER_OF_REGIONS,
        "Number of regions for GCP extraction, specified by columns and rows",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY, 1, 1, 128, 2,
        default_number_of_regions, local_number_of_regions,
        sizeof(local_number_of_regions), 2);

    const int default_minimum_number_of_points_in_region[] = {10};
    IAS_PARM_ADD_INT(parameter_table, MINIMUM_NUMBER_OF_POINTS_IN_REGION,
        "Minimum number of points in a region for GCP extraction",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 1, 2048, 1,
        default_minimum_number_of_points_in_region,
        &parameters->minimum_number_of_points_in_region,
        sizeof(parameters->minimum_number_of_points_in_region), 0);

    const int default_minimum_interest_operator_points[] = {40};
    IAS_PARM_ADD_INT(parameter_table, MINIMUM_INTEREST_OPERATOR_POINTS,
        "Minimum number of points to keep for GCP extraction",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 1, 2048, 1,
        default_minimum_interest_operator_points,
        &parameters->minimum_interest_operator_points,
        sizeof(parameters->minimum_interest_operator_points), 0);

    const double default_minimum_interest_operator_threshold[] = {262140000.0};
    IAS_PARM_ADD_DOUBLE(parameter_table, MINIMUM_INTEREST_OPERATOR_THRESHOLD,
        "Minimum interest operator threshold, the cut off for not using points"
        " with interest operator values below this threshold",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 0, 0.0, 0.0, 1,
        default_minimum_interest_operator_threshold,
        &parameters->minimum_interest_operator_threshold,
        sizeof(parameters->minimum_interest_operator_threshold), 0);

    IAS_PARM_ADD_INT(parameter_table, MINIMUM_GCP_DISTANCE,
        "The minimum allowed distance (line, sample) between GCPs. "
        "Defaults to chip size.", IAS_PARM_OPTIONAL, IAS_PARM_ARRAY, 1, 0, 1000,
        0, NULL, local_minimum_distance, sizeof(local_minimum_distance), 2);

    const int default_points_before_region_filtering[] = {400};
    IAS_PARM_ADD_INT(parameter_table, POINTS_BEFORE_REGION_FILTERING,
        "The number of the best points kept from the Moravec interest operator"
        " before starting to filter points based on the regions",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 1, 5000, 1,
        default_points_before_region_filtering,
        &parameters->points_before_region_filtering,
        sizeof(parameters->points_before_region_filtering), 0);

    const int default_local_maximum_distance[] = {5};
    IAS_PARM_ADD_INT(parameter_table, LOCAL_MAXIMUM_DISTANCE,
        "Distance from other maximum values",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 2, 128, 1,
        default_local_maximum_distance, &parameters->local_maximum_distance,
        sizeof(parameters->local_maximum_distance), 0);

    const int default_output_intermediate_data[] = {0};
    IAS_PARM_ADD_INT(parameter_table, OUTPUT_INTERMEDIATE_DATA,
        "Dump all internally generated data",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 0, 1, 1,
        default_output_intermediate_data,
        &parameters->output_intermediate_data,
        sizeof(parameters->output_intermediate_data), 0);

    const const char *default_intermediate_output_directory[] =
        {"intermediate_data_dumps"};
    IAS_PARM_ADD_STRING(parameter_table, INTERMEDIATE_OUTPUT_DIRECTORY,
        "Name of the output directory",
        IAS_PARM_OPTIONAL, 0, NULL, 1,
        default_intermediate_output_directory,
        parameters->intermediate_output_directory,
        sizeof(parameters->intermediate_output_directory), 0);

    const double default_fill_pixel[] = {0.0};
    IAS_PARM_BACKGROUND(parameter_table, default_fill_pixel, 
        &parameters->background, sizeof(parameters->background));

    /* EVENLY SPACED and MORAVEC PARAMETERS ------------------------------- */

    const int default_chip_size[] = {64, 64};
    IAS_PARM_ADD_INT(parameter_table, CHIP_SIZE,
        "GCP chip size to generate: (line, sample)",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY, 1, 8, 512, 2,
        default_chip_size, local_chip_size, sizeof(local_chip_size), 2);

    /* -------------------------------------------------------------------- */
    /* Read the parameter file */

    /* Calculate the number of items in the parameter file */
    count = IAS_PARM_GET_TABLE_SIZE(parameter_table);

    /* Call the ias_parm_provide_help routine to provide help for the
       parameters if requested; also verifies the ODL file exists */
    help_requested = ias_parm_provide_help(parameter_filename, parameter_table,
        count, IAS_INPUT_PARAMETERS);
    if (help_requested)
    {
        /* Either help was provided or an error occurred */
        if (help_requested == TRUE)
        {
            /* Provide help for the OMF.  If it succeeds, it will exit.  If
               it returns, an error occurred, so return an error */
            read_omf_parameters(parameter_filename, parameters);
            IAS_LOG_ERROR("Providing application parameter file help");
            return ERROR;
        }
        
        IAS_LOG_ERROR("Error occurred providing ODL help");
        return ERROR;
    }
  
    /* Read the parameter file */
    if (ias_parm_read(parameter_filename, parameter_table, count) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from: %s", parameter_filename);
        return ERROR;
    }

    /* Build the OMF filename */
    status = snprintf(local_omf_filename, sizeof(local_omf_filename),
        "%s.omf", local_wo_id);
    if (status < 0 || status >= sizeof(local_omf_filename))
    {
        IAS_LOG_ERROR("Creating the OMF filename");
        return ERROR;
    }

    /* Read in the OMF parameters */
    if (read_omf_parameters(local_omf_filename, parameters) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from %s", local_omf_filename);
        return ERROR;
    }

    /* If the water mask should be used and the water mask filename was 
       not provided, use the default */
    if (parameters->use_water_mask_flag == USE_POLYGON_WATER_MASK &&
        IAS_PARM_GET_COUNT_READ(parameter_table, WATER_MASK_FILENAME) == 0)
    {
        const char *data_dir = getenv("IAS_DATA_DIR");
        if (!data_dir)
        {
            IAS_LOG_ERROR("IAS_DATA_DIR environment variable not set");
            return ERROR;
        }

        status = snprintf(parameters->water_mask_filename, 
            sizeof(parameters->water_mask_filename), "%s%s", 
            data_dir, "/sysdata/water_mask/land_2k_buf.ply");
        if (status < 0 || status >= sizeof(parameters->water_mask_filename))
        {
            IAS_LOG_ERROR("Creating the water mask filename");
            return ERROR;
        }
    }

    /* Check that the binary water mask filename is present */
    if (parameters->use_water_mask_flag == USE_BINARY_WATER_MASK &&
        strlen(parameters->water_mask_filename) <= 0)
    {
        IAS_LOG_ERROR("Required parameter WATER_MASK_FILENAME not present");
        return ERROR; 
    }

    /* Check if the minimum distance parameter was read */
    if (IAS_PARM_GET_COUNT_READ(parameter_table, MINIMUM_GCP_DISTANCE) < 2)
    {
        local_minimum_distance[0] = 0;
        local_minimum_distance[1] = 0;
    }

    /* Process all of the parameters */
    if (process_parameters(local_source_database, parameters) != SUCCESS)
    {
        IAS_LOG_ERROR("Processing the parameters");
        return ERROR;
    }

    return SUCCESS;
}
