
/* Standard C Includes */
#include <stdlib.h>
#include <string.h>

/* IAS Includes */
#include "ias_const.h"
#include "ias_types.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"

/* Local Includes */
#include "generate_chips.h"
#include "resample_data.h"

/*****************************************************************************
 NAME: resample_data

 PURPOSE: To call the correct resampling algorithm based on the resampling
          kernel.  Also provides a common mechanism to output the resampled
          images to a file if requested by the user.

 RETURN VALUE: Type = int
    Value      Description
    ---------  --------------------------
    SUCCESS    Successful completion
    ERROR      Operation failed

*****************************************************************************/
int resample_data
(
    const GENERATE_CHIPS_PARAMETERS *parameters, /* I: The input parameters */
    RESOLUTION_SAMPLING resolution_info, /* I: The sampling information */
    int number_of_input_lines,     /* I: Number of lines in input data */
    int number_of_input_samples,   /* I: Number of samples in input data */
    const IAS_R4_TYPE *input_data, /* I: The input data to scale */
    int *number_of_output_lines,   /* O: Number of lines in output data */
    int *number_of_output_samples, /* O: Number of samples in output data */
    IAS_R4_TYPE *output_data       /* O: The results of the resampling */
)
{
    int status;                /* Capture function/routine statuses */
    int data_size;             /* The total size of the data */

    /* Initialize the status to ERROR */
    status = ERROR;

    /* Perform the required sampling */
    switch (resolution_info.kernel_size)
    {
        /* A kernel size of 0 indicates "do not resample, make a copy" */
        case 0:
            IAS_LOG_INFO("Not resampling: Creating a Copy ...");

            /* Figure out the sizes for the data */
            data_size = number_of_input_lines * number_of_input_samples;
            *number_of_output_lines = number_of_input_lines;
            *number_of_output_samples = number_of_input_samples;

            /* Only need to copy the memory for this option */
            memcpy(output_data, input_data, sizeof(IAS_R4_TYPE) * data_size);

            /* Need to set this since we didn't call any resampling routine */
            status = SUCCESS;
            break;

        default:
            IAS_LOG_INFO("Resampling using a scaling of %lf",
                resolution_info.scaling_factor);

            /* Call the resampling algorithm */
            status = perform_resample(resolution_info.scaling_factor,
                resolution_info.kernel_size,
                number_of_input_lines, number_of_input_samples, input_data,
                number_of_output_lines, number_of_output_samples, output_data);
            break;
    }

    /* Dump intermediate data */
    if (parameters->output_intermediate_data == 1 && status == SUCCESS)
    {
        char output_filename[PATH_MAX];
        int count; 
        
        IAS_LOG_INFO("Resampled number of (lines, samples) = (%d, %d)",
            *number_of_output_lines, *number_of_output_samples);

        /* Create the full output file path */                     
        count = snprintf(output_filename, sizeof(output_filename), 
            "%s/%s", parameters->intermediate_output_directory,
            resolution_info.output_filename);
        if (count < 0 || count >= sizeof(output_filename))
        {
            IAS_LOG_ERROR("Bounds exceeded or internal snprintf error:"
                " count = %d", count);
            status = ERROR;
        }
        else if (ias_misc_write_envi_image(output_filename, NULL, NULL, 
            *number_of_output_lines, *number_of_output_samples, 0.0, 0.0, 0.0, 
            0.0, NULL, IAS_R4, output_data) != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to generate ENVI image");
            status = ERROR;
        }
    }

    return status;
}

