
/* Standard C Includes */
#include <string.h>
#include <time.h>

/* IAS Includes */
#include "gctp.h"
#include "ias_logging.h"
#include "ias_gcp.h"
#include "ias_geo.h"
#include "ias_miscellaneous.h"
#include "ias_satellite_attributes.h"

/* Local Includes */
#include "generate_chips.h"

/*****************************************************************************
 NAME: update_valid_gcp_records

 PURPOSE: Update the record contents for the valid GCPs.

 RETURN VALUE: Type = int
    Value    Description
    -------  --------------------------
    SUCCESS  Successful completion
    ERROR    Operation failed

*****************************************************************************/
int update_valid_gcp_records
(
    const GENERATE_CHIPS_PARAMETERS *parameters, /* I: The input parameters */
    const GENERATE_CHIPS_L1G_DATA *l1g_data,     /* I: Input L1G data */
    const IAS_GEO_PROJ_TRANSFORMATION *input_image2geographic,
                                /* I: The projection transformation to use */
    const IAS_IMAGE *elevation_image, /* I: The elevation image if specified */
    GENERATE_CHIPS_GCPS *gcps /* I/O: Holds the input and output GCP records */
)
{
    int index;              /* Loop control */
    int status;             /* Return status */
    int count;              /* Capture snprintf errors */
    time_t now;             /* Holds the current time */
    struct tm *info;        /* Time structure */
    char current_date[12];  /* String version of date */
    
    /* Get the current date as a string */
    time(&now);
    info = localtime(&now);
    strftime(current_date, 11, "%m-%d-%Y", info);

    /* Initialize satellite library */
    status = ias_sat_attr_initialize(IAS_L8);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Could not initialize satelite library");
        return ERROR;   
    }

    /* Process through the valid GCPs and update them */
    status = SUCCESS;
    for (index = 0; index < gcps->number_of_gcps; index++)
    {
        /* Update the absolute or relative value specified in the input
           parameters. */
        count = snprintf(gcps->gcp_records[index].absolute_or_relative,
            sizeof(gcps->gcp_records[index].absolute_or_relative),
            parameters->absolute_or_relative);

        if (count < 0
            || count >= sizeof(gcps->gcp_records[index].absolute_or_relative))
        {
            IAS_LOG_ERROR("Bounds exceeded or internal snprintf error:"
                " assigning absolute or relative: count = %d", count);
            status = ERROR;
            break;
        }

        /* Update the date to the date of the data */
        count = snprintf(gcps->gcp_records[index].date,
            sizeof(gcps->gcp_records[index].date),
            IAS_GCP_DATE_FORMAT, gcps->month, gcps->day, gcps->year);

        if (count < 0 || count >= sizeof(gcps->gcp_records[index].date))
        {
            IAS_LOG_ERROR("Bounds exceeded or internal snprintf error:"
                " assigning GCP date: count = %d", count);
            status = ERROR;
            break;
        }

        /* Update the source to the source of the data, if not using GCPLib 
           generation method */
        if (parameters->generation_method != GM_GCPLIB)
        {
            count = snprintf(gcps->gcp_records[index].source,
                sizeof(gcps->gcp_records[index].source),
                parameters->gcp_source);
            if (count < 0 || count >= sizeof(gcps->gcp_records[index].source))
            {
                IAS_LOG_ERROR("Bounds exceeded or internal snprintf error:"
                    " assigning GCP source: count = %d", count);
                status = ERROR;
                break;
            }
        }
        else
        {
            /* Nothing to do. Leave source as-is in the GCP record. */
        }


        /* Update the projection and zone information */
        switch (l1g_data->metadata.projection_code)
        {
            case UTM:
                count = snprintf(gcps->gcp_records[index].projection,
                    sizeof(gcps->gcp_records[index].projection), "UTM");
                break;

            case PS:
                count = snprintf(gcps->gcp_records[index].projection,
                    sizeof(gcps->gcp_records[index].projection), "PS");
                break;

            default:
                count = -1;
        }

        if (count < 0 || count >= sizeof(gcps->gcp_records[index].projection))
        {
            IAS_LOG_ERROR("Bounds exceeded or internal snprintf error:"
                " assigning GCP projection: count = %d", count);
            status = ERROR;
            break;
        }

        gcps->gcp_records[index].zone = l1g_data->metadata.zone_code;

        /* Specify the size of pixels */
        gcps->gcp_records[index].pixel_size_x =
            l1g_data->band_metadata.projection_distance_x;
        gcps->gcp_records[index].pixel_size_y =
            l1g_data->band_metadata.projection_distance_y;

        /* Figure out the map coordinates from the line and sample */
        gcps->gcp_records[index].projection_y =
            l1g_data->band_metadata.upper_left_y
            - (gcps->gcp_locations[index].input_data_line
               * l1g_data->band_metadata.projection_distance_y);

        gcps->gcp_records[index].projection_x =
            (gcps->gcp_locations[index].input_data_sample
             * l1g_data->band_metadata.projection_distance_x)
            + l1g_data->band_metadata.upper_left_x;

        if (parameters->generation_method != GM_GCPLIB)
        {
            /* Figure out the geographic latitude and longitude */
            ias_geo_transform_coordinate(input_image2geographic,
                gcps->gcp_records[index].projection_x,
                gcps->gcp_records[index].projection_y,
                &gcps->gcp_records[index].longitude,
                &gcps->gcp_records[index].latitude);
            
            /* Assign point/chip wrs row/path */
            gcps->gcp_records[index].point_wrs_path = 
                l1g_data->metadata.wrs_path;      
            gcps->gcp_records[index].point_wrs_row = 
                l1g_data->metadata.wrs_row;
            gcps->gcp_records[index].chip_wrs_path = 
                l1g_data->metadata.wrs_path;
            gcps->gcp_records[index].chip_wrs_row = 
                l1g_data->metadata.wrs_row;

            /* Assign the point and chip version to 1 */
            gcps->gcp_records[index].point_version = 1;
            gcps->gcp_records[index].chip_version = 1;

            /* Assign the point_active to "Y" */
            count = snprintf(gcps->gcp_records[index].point_active, 
                sizeof(gcps->gcp_records[index].point_active), "Y");
            if (count < 0 || count >= 
                sizeof(gcps->gcp_records[index].point_active))
            {
                IAS_LOG_ERROR("Bounds exceeded or internal snprintf error:"
                " assigning point active: count = %d", count);
                status = ERROR;
                break;
            }
            
            /* Assign the chip active to "Y" */
            count = snprintf(gcps->gcp_records[index].chip_active, 
                sizeof(gcps->gcp_records[index].chip_active), "Y");
            if (count < 0 || count >= 
                sizeof(gcps->gcp_records[index].chip_active))
            {
                IAS_LOG_ERROR("Bounds exceeded or internal snprintf error:"
                " assigning chip active: count = %d", count);
                status = ERROR;
                break;
            }
            
            /* Assign the point date updated */
            count = snprintf(gcps->gcp_records[index].point_date_updated, 
                sizeof(gcps->gcp_records[index].point_date_updated), 
                current_date);
            if (count < 0 || count >= 
                sizeof(gcps->gcp_records[index].point_date_updated))
            {
                IAS_LOG_ERROR("Bounds exceeded or internal snprintf error:"
                " assigning point date updated: count = %d", count);
                status = ERROR;
                break;
            }

            /* Assign the chip date updated */
            count = snprintf(gcps->gcp_records[index].chip_date_updated, 
                sizeof(gcps->gcp_records[index].chip_date_updated), 
                current_date); 
            if (count < 0 || count >= 
                sizeof(gcps->gcp_records[index].chip_date_updated))
            {
                IAS_LOG_ERROR("Bounds exceeded or internal snprintf error:"
                " assigning chip date updated: count = %d", count);
                status = ERROR;
                break;
            }
            
            /* Assign the update reason */
            count = snprintf(gcps->gcp_records[index].update_reason, 
                sizeof(gcps->gcp_records[index].update_reason), 
                parameters->gcp_update_reason);
            if (count < 0 || count >= 
                sizeof(gcps->gcp_records[index].update_reason))
            {
                IAS_LOG_ERROR("Bounds exceeded or internal snprintf error:"
                " assigning update reason: count = %d", count);
                status = ERROR;
                break;
            }

            /* Assign the Sensor type */
            count = snprintf(gcps->gcp_records[index].chip_sensor,
                sizeof(gcps->gcp_records[index].chip_sensor),
                l1g_data->band_metadata.instrument_source);
            if (count < 0 || count >= 
                sizeof(gcps->gcp_records[index].chip_sensor))
            {
                IAS_LOG_ERROR("Bounds exceeded or internal snprintf error:"
                " assigning chip sensor: count = %d", count);
                status = ERROR;
                break;
            }

            /* Check if Sensor is OLI or TIRS */
            if (strcmp(l1g_data->band_metadata.instrument_source,
                ias_sat_attr_get_sensor_name(IAS_OLI)) != 0 &&
                strcmp(l1g_data->band_metadata.instrument_source,
                ias_sat_attr_get_sensor_name(IAS_TIRS)) != 0)
            {
                IAS_LOG_WARNING("Sensor is not OLI or TIRS - %s",
                    l1g_data->band_metadata.instrument_source);
            }

            /* Assign the Chip name */
            count = snprintf(gcps->gcp_records[index].chip_name,
                sizeof(gcps->gcp_records[index].chip_name),
                "%s_%02d", gcps->gcp_records[index].point_id,
                gcps->gcp_records[index].point_version);         
            if (count < 0 || count >= 
                sizeof(gcps->gcp_records[index].chip_name))
            {
                IAS_LOG_ERROR("Bounds exceeded or internal snprintf error:"
                " assigning chip name: count = %d", count);
                status = ERROR;
                break;
            }
        }

        /* Fix the elevation */
        if (parameters->elevation_filename_present)
        {
            ias_misc_read_elevation_at_line_sample(elevation_image,
                gcps->gcp_locations[index].input_data_line,
                gcps->gcp_locations[index].input_data_sample,
                gcps->gcp_records[index].pixel_size_y,
                gcps->gcp_records[index].pixel_size_x,
                &gcps->gcp_records[index].elevation);
        }

        /* Fix the chip data type */
        gcps->gcp_records[index].chip_data_type = l1g_data->data_type;
    }

    return status;
}

