#ifndef KPR5_H
#define KPR5_H

#include "ias_cpf.h"
#include "kpr_shared.h"

/* #defined constants for the threshold swath widths used in the
   evaluation. */
#define KPR5_REQUIRED_SWATH_WIDTH_KM                  185.0
#define KPR5_FIVE_PERCENT_PENALTY_SWATH_WIDTH_KM      184.0
#define KPR5_TEN_PERCENT_PENALTY_SWATH_WIDTH_KM       183.0
#define KPR5_FIFTEEN_PERCENT_PENALTY_SWATH_WIDTH_KM   182.0
#define KPR5_TWENTY_PERCENT_PENALTY_SWATH_WIDTH_KM    181.0


/* Definition of a data structure containing elements referenced in the KPR 5
   summary report. */
typedef struct kpr5_report_data
{
    double adjusted_max_angle_radians;  /* Smallest maximum crosstrack
                                           angle in radians, adjusted for
                                           inoperable detector effects */
    double unadjusted_max_angle_radians;/* Smallest unadjusted maximum
                                           crosstrack angle in radians */
    double adjusted_min_angle_radians;  /* Largest minimum crosstrack
                                           angle in radians, adjusted for
                                           inoperable detector effects */
    double unadjusted_min_angle_radians;/* Largest unadjusted minimum
                                           crosstrack angle in radians */
    double max_ifov_radians;            /* Maximum IFOV in radians,
                                           over all bands/SCAs*/
    double max_sca1_ifov_radians;       /* Maximum SCA 1 IFOV in radians
                                           across all bands */
    double max_sca14_ifov_radians;      /* Maximum SCA 14 FOV in radians
                                           across all bands */
    double min_ifov_radians;            /* Minimum FOV in radians, between
                                           adjusted swath angles */
    double oli_fov_degrees;             /* Minimum FOV expressed in degrees */
    double starboard_angle_radians;     /* Starboard swath angle in radians */
    double port_angle_radians;          /* Port swath angle in radians */
    double swath_angle_radians;         /* Full swath angle from starboard
                                           to port */
    double earth_radius_km;             /* Radius of earth in km */
    double spacecraft_altitude_km;      /* Nominal altitude of spacecraft */
    double spacecraft_orbit_radius_km;  /* Earth radius + nominal spacecraft
                                           altitude in km */
    double net_swath_width_km;          /* Full swath width in km that
                                           accounts for inoperable edge
                                           detectors */
    int award_penalty_percent;          /* Percentage penalty assessed to
                                           award fee for not meeting
                                           requirement */
    int max_angle_band_number;          /* Band number with largest angle */
    int max_dead_band_number;           /* Band number with largest number of
                                           inoperable detectors */
    int min_angle_band_number;          /* Band number with smallest angle */
    int number_dead_pan_detectors;      /* Number of inoperable pan band
                                           edge detectors */
    int number_dead_ms_detectors;       /* Number of inoperable ms band
                                           detectors */
}   KPR5_REPORT_DATA;



/* KPR5-specific function prototypes */

/* NOTE:  The pointers *across_xxx_angles_sca_band_radians,
   *ifovs_sca_band_radians, and *edge_counts_sca_band

   represent dynamically allocated 1D arrays with (number_of_oli_scas *
   number_of_oli_bands) elements.  The indexing into these arrays is
   calculated as

   array_index = sca_index * number_of_bands + band_list_index

   where band_list_index = 0 to number_of_bands - 1.  The band_list_index
   values correspond to the OLI band number list which is sorted as follows:

   band_list[number_of_bands] = {8, 1, 2, 3, 4, 5, 6, 7, 9}.

   Consequently, the data in these arrays are ordered as follows:

     [0] = SCA 01 band 8,  ...  [7] = SCA 01 band 7,    [8] = SCA 01 band 9,
     [9] = SCA 02 band 8,  ... [16] = SCA 02 band 7,   [17] = SCA 02 band 9,
    [18] = SCA 03 band 8,  ... [25] = SCA 03 band 7,   [26] = SCA 03 band 9,
    [27] = SCA 04 band 8,  ... [34[ = SCA 04 band 7,   [35] = SCA 04 band 9,
     ...                   ...                     ...
   [117] = SCA 14 band 8,  ...[124] = SCA 14 band 7,  [125] = SCA 14 band 9

   To emphasize the point...  THIS DOES NOT FOLLOW THE CONVENTIONAL
   BAND / SCA ORDERING USED IN OTHER IAS PROCESSING.  */

int kpr5_calculate_swath_width
(
    const int *band_number_list,         /* I: List of bands to process */
    int number_of_bands,                 /* I: Number of bands to process */
    int number_of_scas,                  /* I: Number of SCAs */
    const double *across_max_angles_sca_band_radians,
                                         /* I: 1D array of unadjusted maximum
                                            crosstrack angles in radians,
                                            ordered by SCA and band */
    const double *across_min_angles_sca_band_radians,
                                         /* I: 1D array of unadjusted minimum
                                            crosstrack angles in radians,
                                            ordered by SCA and band */
    const double *ifovs_sca_band_radians,/* I: 1D array of crosstrack FOVs in
                                            radians, ordered by SCA and
                                            band */
    const KPR_INOPERABLE_EDGE_DETECTOR_COUNTS *edge_counts_sca_band,
                                         /* I: 1D array of
                                            KPR_INOPERABLE_EDGE_DETECTOR_COUNTS
                                            structures containing inoperable
                                            edge detector counts, ordered by
                                            SCA and band */
    KPR5_REPORT_DATA *report_data        /* O: KPR 5 report data structure
                                            populated with summary assessment
                                            information */
);


int kpr5_write_report
(
    const int *band_number_list,         /* I: List of bands to process */
    int number_of_bands,                 /* I: Number of bands to process */
    int number_of_scas,                  /* I: Number of SCAs */
    const struct IAS_CPF_FOCAL_PLANE *focal_plane,
                                         /* I: Focal plane parameters */
    const double *across_max_angles_sca_band_radians,
                                         /* I: 1D array containing the
                                            unadjusted maximum crosstrack
                                            angles in radians, ordered by
                                            SCA and band */
    const double *across_min_angles_sca_band_radians,
                                         /* I: 1D array containing the
                                            unadjusted minimum crosstrack
                                            angles in radians, ordered by
                                            SCA and band */
    const double *ifovs_sca_band_radians, /* I: 1D array containing the
                                            crosstrack IFOVs in radians,
                                            ordered by SCA and band */
    const char *rpt_filename,            /* I: Name of summary report file */
    const KPR5_REPORT_DATA *rpt_data     /* I: Populated summary report data
                                            structure */
);


#endif

