/***************************************************************************
 NAME:      kpr5_calculate_swath_width

 PURPOSE:   Calculates the net swath width (in km) based on the available
            crosstrack angles and inoperable edge detector information, and
            gathers summary information for the assessment report.

 RETURNS:   Integer status of SUCCESS or ERROR

 NOTES:     The pointers *across_xxx_angles_sca_band_radians,
            *ifovs_sca_band_radians, and *edge_counts_sca_band represent
            dynamically allocated 1D arrays that DO NOT follow the standard
            band/SCA ordering used in other IAS processing.  Refer to
            the NOTE in kpr5.h for additional information.
****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "kpr5.h"


int kpr5_calculate_swath_width
(
    const int *band_number_list,         /* I: List of bands to process */
    int number_of_bands,                 /* I: Number of bands to process */
    int number_of_scas,                  /* I: Number of SCAs */
    const double *across_max_angles_sca_band_radians,
                                         /* I: 1D array of unadjusted maximum
                                            crosstrack angles in radians,
                                            ordered by SCA and band */
    const double *across_min_angles_sca_band_radians,
                                         /* I: 1D array of unadjusted minimum
                                            crosstrack angles in radians,
                                            ordered by SCA and band */
    const double *ifovs_sca_band_radians,/* I: 1D array of crosstrack FOVs in
                                            radians, ordered by SCA and
                                            band */
    const KPR_INOPERABLE_EDGE_DETECTOR_COUNTS *edge_counts_sca_band,
                                         /* I: 1D array of
                                            KPR_INOPERABLE_EDGE_DETECTOR_COUNTS
                                            structures containing inoperable
                                            edge detector counts, ordered by
                                            SCA and band */
    KPR5_REPORT_DATA *report_data        /* O: KPR 5 report data structure
                                            populated with summary assessment
                                            information */
)
{
    /* Dynamically allocated 1D arrays containing the 'adjusted'
       maximum and minimum crosstrack view angles.  The data in these
       arrays are ordered identically to that of the corresponding
       unadjusted arrays as described in kpr5.h. */
    double *adjusted_max_angles_sca_band_radians = NULL;
    double *adjusted_min_angles_sca_band_radians = NULL;
    size_t buffer_size =
        (number_of_scas * number_of_bands) * sizeof(double);

    /* Ratio of spacecraft orbit "radius"
      (i.e. earth radius + nominal sc altitude) to earth radius */
    double ratio = report_data->spacecraft_orbit_radius_km
        / report_data->earth_radius_km;
    double port_swath_width;
    double starboard_swath_width;

    /* Band list index and array offsets */
    int base_sca14_offset = (number_of_scas - 1) * number_of_bands;
    int band_list_index;


    /* Allocate the adjusted angle data buffers. */
    adjusted_max_angles_sca_band_radians = malloc(buffer_size);
    adjusted_min_angles_sca_band_radians = malloc(buffer_size);
    if ((adjusted_max_angles_sca_band_radians == NULL)
        || (adjusted_min_angles_sca_band_radians == NULL))
    {
        IAS_LOG_ERROR("Allocating adjusted angle data buffers");
        free(adjusted_max_angles_sca_band_radians);
        free(adjusted_min_angles_sca_band_radians);
        return ERROR;
    }

    /* Process all the bands. */
    for (band_list_index = 0;
         band_list_index < number_of_bands;
         band_list_index++)
    {
        /* SCA 1 inoperable detector edge correction factors */
        double sca1_leftedge_factor;
        double sca1_rightedge_factor;

        /* SCA 14 inoperable detector edge correction factors */
        double sca14_leftedge_factor;
        double sca14_rightedge_factor;

        int band_number = band_number_list[band_list_index];
        int sca1_offset = band_list_index;
        int sca14_offset = base_sca14_offset + band_list_index;

        /* Since we only need to process the data from SCAs 1 and 14,
           just access those elements directly as needed (without using
           an inner SCA loop).  */

        /*=================================================================
                          SCA 01 PROCESSING
        ===================================================================*/

        /* SCA 1 has positive crosstrack view angles.  This check is
           looking for the innermost (smallest) positive angle present in SCA
           1 across all bands. */
        if (across_max_angles_sca_band_radians[sca1_offset]
            < report_data->unadjusted_max_angle_radians)
        {
            report_data->unadjusted_max_angle_radians =
                across_max_angles_sca_band_radians[sca1_offset];
        }

        /* Calculate edge correction factors to the SCA 1 FOV based on the
           number of inoperable edge detectors that were found. */
        sca1_leftedge_factor =
            edge_counts_sca_band[sca1_offset].left_edge_count
            * ifovs_sca_band_radians[sca1_offset];

        sca1_rightedge_factor =
            edge_counts_sca_band[sca1_offset].right_edge_count
            * ifovs_sca_band_radians[sca1_offset];

        /* Correct the maximum and minimum cross-track angles for SCA 1 to
           account for any inoperable edge detectors.  The subtraction
           for the left edge will make the maximum crosstrack view angle
           smaller, and the addition for the right edge will make the
           minimum crosstrack view angle larger. */
        adjusted_max_angles_sca_band_radians[sca1_offset] =
            across_max_angles_sca_band_radians[sca1_offset]
            - sca1_leftedge_factor;
        adjusted_min_angles_sca_band_radians[sca1_offset] =
            across_min_angles_sca_band_radians[sca1_offset]
            + sca1_rightedge_factor;

        /* Keep track of the smallest adjusted maximum crosstrack angle in
           SCA 1.  Also note the band number where it is found.  If two or
           more bands should have the same smallest adjusted maximum angle
           value, the last band where it is found will be noted.  */
        if (adjusted_max_angles_sca_band_radians[sca1_offset]
            < report_data->adjusted_max_angle_radians)
        {
            report_data->adjusted_max_angle_radians =
                adjusted_max_angles_sca_band_radians[sca1_offset];
            report_data->max_angle_band_number = band_number;
        }

        /* Keep track of the largest crosstrack FOV in SCA 1. */
        if (ifovs_sca_band_radians[sca1_offset]
            > report_data->max_sca1_ifov_radians)
        {
            report_data->max_sca1_ifov_radians =
                ifovs_sca_band_radians[sca1_offset];
        }

        /*=================================================================
                          SCA 14 PROCESSING
        ===================================================================*/

        /* SCA 14 has negative crosstrack view angles.  This check is
           looking for the outermost (least negative) angle present in SCA 14
           across all bands. */
        if (across_min_angles_sca_band_radians[sca14_offset]
            > report_data->unadjusted_min_angle_radians)
        {
            report_data->unadjusted_min_angle_radians =
                across_min_angles_sca_band_radians[sca14_offset];
        }

        /* Calculate edge correction factors to the SCA 14 FOV based on the
           number of inoperable edge detectors that were found. */
        sca14_leftedge_factor =
            edge_counts_sca_band[sca14_offset].left_edge_count
            * ifovs_sca_band_radians[sca14_offset];

        sca14_rightedge_factor =
            edge_counts_sca_band[sca14_offset].right_edge_count
            * ifovs_sca_band_radians[sca14_offset];

        /* Correct the maximum and minimum crosstrack angles for SCA 14 to
           account for any inoperable edge detectors.  The subtraction
           for the left edge will make the maximum crosstrack view angle
           smaller, and the addition for the right edge will make the
           minimum crosstrack view angle larger. */
        adjusted_max_angles_sca_band_radians[sca14_offset] =
            across_max_angles_sca_band_radians[sca14_offset]
            - sca14_leftedge_factor;

        adjusted_min_angles_sca_band_radians[sca14_offset] =
            across_min_angles_sca_band_radians[sca14_offset]
            + sca14_rightedge_factor;

        /* Keep track of the largest adjusted minimum crosstrack angle in SCA
           14.  Again, if two or more bands should have the same largest
           adjusted minimum angle value, the last band where it is found will
           be noted.  */
        if (adjusted_min_angles_sca_band_radians[sca14_offset]
            > report_data->adjusted_min_angle_radians)
        {
            report_data->adjusted_min_angle_radians =
                adjusted_min_angles_sca_band_radians[sca14_offset];
            report_data->min_angle_band_number = band_number;
        }

        /* Keep track of the largest crosstrack FOV in SCA 14. */
        if (ifovs_sca_band_radians[sca14_offset]
            > report_data->max_sca14_ifov_radians)
        {
            report_data->max_sca14_ifov_radians =
                ifovs_sca_band_radians[sca14_offset];
        }
    }

    /* Calculate the minimum FOV in radians from the adjusted crosstrack
       angles. */
    report_data->min_ifov_radians = report_data->adjusted_max_angle_radians
        - report_data->adjusted_min_angle_radians;

    /* Convert the minimum FOV to its degree equivalent. */
    report_data->oli_fov_degrees = report_data->min_ifov_radians
        * ias_math_get_degrees_per_radian();

    /* Calculate the required starboard and port angles in radians.  */
    report_data->starboard_angle_radians =
        asin(ratio * sin(report_data->adjusted_max_angle_radians))
        - report_data->adjusted_max_angle_radians;

    report_data->port_angle_radians =
        asin(ratio * sin(report_data->adjusted_min_angle_radians))
        - report_data->adjusted_min_angle_radians;

    /* Calculate the net swath angle in radians. */
    report_data->swath_angle_radians = report_data->starboard_angle_radians
        - report_data->port_angle_radians;

    /* Calculate the corresponding swath widths in km. */
    starboard_swath_width =
        report_data->earth_radius_km * report_data->starboard_angle_radians;
    port_swath_width =
        report_data->earth_radius_km * report_data->port_angle_radians;

    /* Calculate the net swath width in km. */
    report_data->net_swath_width_km = starboard_swath_width - port_swath_width;

    /* Free the adjusted angle data buffers. */
    free(adjusted_max_angles_sca_band_radians);
    free(adjusted_min_angles_sca_band_radians);
    adjusted_max_angles_sca_band_radians = NULL;
    adjusted_min_angles_sca_band_radians = NULL;

    return SUCCESS;

}   /* END -- kpr5_calculate_swath_width */
