/****************************************************************************
 NAME:       kpr5_evaluate

 PURPOSE:    Performs analysis of current focal plane information to
             determine whether the KPR 5 performance requirement has been
             met.

 RETURNS:    Integer status code of EXIT_SUCCESS (all processing completed
             successfully) or EXIT_FAILURE (error condition detected).
*****************************************************************************/

/* Standard C library includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <float.h>

/* IAS system includes */
#include "ias_const.h"
#include "ias_cpf.h"
#include "ias_logging.h"
#include "ias_satellite_attributes.h"

/* Specific header include */
#include  "kpr5.h"


/* The current KPR 5 requirement (as mentioned in the ADD) applies to
   spacecraft altitudes between 704 km and 728 km.  The assessment should be
   performed at the lowest altitude applicable -- 704 km -- as this will yield
   the smallest, "worst-case" swath width. */
#define KPR5_SPACECRAFT_ALTITUDE_KM    704.00



/****************************************************************************
 NAME:         initialize_report_data

 PURPOSE:      Initializes the KPR5 report data structure members

 RETURNS:      Nothing
*****************************************************************************/
static void initialize_report_data
(
    double earth_radius_km,               /* I: Radius of earth in km from
                                             CPF */
    KPR5_REPORT_DATA *rpt_data            /* O: Report data structure with
                                             initialized values */
)
{
    /* Pre-initialize everything to 0. */
    memset(rpt_data, 0, sizeof(KPR5_REPORT_DATA));

    /* The swath width calculations require use of the innermost (SMALLEST)
       maximum crosstrack angle for SCA 1 (where all crosstrack angles are
       positive), and  the outermost (LARGEST) minimum crosstrack angle for
       SCA 14 (where all crosstrack angles are negative).  */

    /* Initialize the unadjusted and adjusted max angle values to the
       negative of the largest system limit for a floating point value.
       Similarly, initialize the unadjusted and adjusted "max" angle values
       to the largest system limit.  */
    rpt_data->unadjusted_min_angle_radians = -DBL_MAX;
    rpt_data->unadjusted_max_angle_radians =  DBL_MAX;
    rpt_data->adjusted_max_angle_radians   =  DBL_MAX;
    rpt_data->adjusted_min_angle_radians   = -DBL_MAX;

    /* Since we're dealing with negative angles in SCA 14, initialize
       its max IFOV angle value to the negative of the largest system
       limit. */
    rpt_data->max_sca1_ifov_radians        = -DBL_MAX;
    rpt_data->max_sca14_ifov_radians       = -DBL_MAX;
    rpt_data->min_ifov_radians             =  DBL_MAX;

    /* Fill in the earth and orbit-related information. */
    rpt_data->earth_radius_km = earth_radius_km;
    rpt_data->spacecraft_altitude_km = KPR5_SPACECRAFT_ALTITUDE_KM;
    rpt_data->spacecraft_orbit_radius_km = earth_radius_km
        + KPR5_SPACECRAFT_ALTITUDE_KM;

}   /* END -- initialize_report_data */


/****************************************************************************
 NAME:         assess_award_penalty

 PURPOSE:      Assigns a penalty (as a percentage of the contract
               award) based on the total swath width.  A 5% percent
               penalty is assessed for each kilometer below the
               requirement threshold of 185.0 km.  For swath widths
               less than 181.0 km, a 100% penalty is assessed.  For
               swath widths of 185.0 km or more, no penalty is
               assessed.

 RETURNS:      Nothing
*****************************************************************************/
static void assess_award_penalty
(
    double swath_width,          /* I: Net swath width in km */
    int *percent_penalty         /* O: Percentage penalty reduction
                                    of contract award for not meeting the
                                    requirement */
)
{
    if (swath_width >= KPR5_REQUIRED_SWATH_WIDTH_KM)
    {
        *percent_penalty = 0;
    }
    else if (swath_width >= KPR5_FIVE_PERCENT_PENALTY_SWATH_WIDTH_KM)
    {
        *percent_penalty = 5;
    }
    else if (swath_width >= KPR5_TEN_PERCENT_PENALTY_SWATH_WIDTH_KM)
    {
        *percent_penalty = 10;
    }
    else if (swath_width >= KPR5_FIFTEEN_PERCENT_PENALTY_SWATH_WIDTH_KM)
    {
        *percent_penalty = 15;
    }
    else if (swath_width >= KPR5_TWENTY_PERCENT_PENALTY_SWATH_WIDTH_KM)
    {
        *percent_penalty = 20;
    }
    else
    {
        *percent_penalty = 100;
    }

}   /* END -- assess_award_penalty */


/***************************************************************************
 NAME:        free_data_arrays

 PURPOSE:     Frees the allocated data arrays

 RETURNS:     Nothing
****************************************************************************/
static void free_data_arrays
(
    double *across_max_angles_sca_band_radians,
                                        /* I: Maximum crosstrack view angles
                                           in radians, ordered by SCA and
                                           band */
    double *across_min_angles_sca_band_radians,
                                        /* I: Minimum crosstrack view angles
                                           in radians, ordered by SCA and
                                           band */
    double *ifovs_sca_band_radians,     /* I: Instantaneous FOV angles in
                                           radians, ordered by SCA and
                                           band */
    KPR_INOPERABLE_EDGE_DETECTOR_COUNTS *edge_counts_sca_band
                                        /* I: Counts of inoperable detectors,
                                           ordered by SCA and band */
)
{
    free(across_max_angles_sca_band_radians);
    free(across_min_angles_sca_band_radians);
    free(ifovs_sca_band_radians);
    free(edge_counts_sca_band);

}   /* END -- free_data_arrays */


/**************************************************************************
 NAME:     summarize_inoperable_detector_information

 PURPOSE:  Determines the total number of inoperable edge detectors
           for SCAs 1 and 14 in all bands.  For the multispectral bands, also
           keeps track of the band with the most inoperable edge detectors

 RETURNS:  Integer status code of SUCCESS or ERROR

 NOTES:    The total number of inoperable edge detectors for a given band
           is the sum of the number of inoperable edge detectors on the
           'right' side of SCA 1 and the number of inoperable edge detectors
           on the 'left' side of SCA 14.  This assumes the SCAs are aligned
           in 'proper' imaging order (i.e. SCA 14 is on the 'left' side of
           a stitched image, SCA 1 is on the 'right' side of a stitched
           image).
***************************************************************************/
static int summarize_inoperable_detector_information
(
    const int *band_number_list,       /* I: List of bands to process */

    int number_of_bands,               /* I: Number of bands to process */
    int number_of_scas,                /* I: Number of SCAs */
    const KPR_INOPERABLE_EDGE_DETECTOR_COUNTS *edge_counts_sca_band,
                                       /* I: 1D array of inoperable edge
                                          structures containing inoperable
                                          edge detector counts, ordered by
                                          SCA and band */
    KPR5_REPORT_DATA *report_data      /* O: Summary report data */
)
{
    /* Starting location in data buffers for SCA 14 data.   Remember that
       the offsets are based on SCA / band order rather than the normally
       expected band / SCA order.  */
    int base_sca14_offset = (number_of_scas - 1) * number_of_bands;

    int band_list_index;
    int max_num_dead = INT16_MIN;
    int max_dead_band = 0;

    /* Get the band number and band_list_index for the pan band. */
    int pan_band_number =
        ias_sat_attr_get_band_number_from_type(IAS_PAN_BAND);
    if (pan_band_number == ERROR)
    {
        IAS_LOG_ERROR("Retrieving pan band number");
        return ERROR;
    }

    /* Process all of the bands.  */
    for (band_list_index = 0;
         band_list_index < number_of_bands;
         band_list_index++)
    {
        /* Get the band number and determine the required offsets for
           the detector information in SCAs 1 and 14. */
        int band_number = band_number_list[band_list_index];

        /* Get the SCA 1 and SCA 14 offsets for the current band. */
        int sca1_buffer_offset = band_list_index;
        int sca14_buffer_offset = base_sca14_offset + band_list_index;

        /* Calculate the number of dead edge detectors on the right side
           of SCA 1 and the left side of SCA 14. */
        int number_of_dead_edge_detectors =
            edge_counts_sca_band[sca1_buffer_offset].right_edge_count
            + edge_counts_sca_band[sca14_buffer_offset].left_edge_count;

        /* Record the total number of edge detectors.  If this is a
           multispectral band, keep track of the band with the maximum
           number of dead edge detectors. */
        if (band_number == pan_band_number)              /* PAN band */
        {
            report_data->number_dead_pan_detectors =
                number_of_dead_edge_detectors;
        }
        else                         /* Multispectral band */
        {
            if (number_of_dead_edge_detectors > max_num_dead)
            {
                max_num_dead = number_of_dead_edge_detectors;
                max_dead_band = band_number;
            }
        }
    }

    /* Save the final dead edge detector information. */
    report_data->number_dead_ms_detectors = max_num_dead;
    report_data->max_dead_band_number = max_dead_band;

    return SUCCESS;

}   /* END -- summarize_inoperable_detector_information */



/*=========================================================================
                       MAIN PROGRAM
===========================================================================*/
int main(int argc, char *argv[])
{
  /*=====================================================================*/

    /* NOTE:  The following set of pointers are declared for dynamically
       allocated 1D arrays that DO NOT follow the conventional band / SCA
       ordering that other IAS processing uses.  Refer to the NOTE: in
       'kpr5.h' for additional information.  */

    /* Array containing counts for inoperable
       detectors on the edges of each SCA.  The array element is a
       structure containing both left edge and right edge counts.  */
    KPR_INOPERABLE_EDGE_DETECTOR_COUNTS *edge_counts_sca_band = NULL;

    /* Arrays containing view angles derived from the crosstrack LOS Legendre
       coefficients. */
    double *across_max_angles_sca_band_radians = NULL;
    double *across_min_angles_sca_band_radians = NULL;

    /* Array containing the crosstrack IFOV angles. */
    double *ifovs_sca_band_radians = NULL;

  /*=====================================================================*/

    /* CPF parameters. */
    const char *cpf_filename = NULL;
    IAS_CPF *cpf = NULL;
    const struct IAS_CPF_DETECTOR_STATUS *detector_status = NULL;
    const struct IAS_CPF_EARTH_CONSTANTS *earth_constants = NULL;
    const struct IAS_CPF_FOCAL_PLANE *focal_plane = NULL;


    /* Processing parameters */
    double cpf_earth_radius_km;

    int oli_band_number_list[IAS_MAX_NBANDS]; 
                                          /* Band list to process.  As
                                             indicated earlier, the
                                             band list is ordered such
                                             that the pan band is placed
                                             first in the list, with the
                                             remaining bands sorted in
                                             ascending order */
    int number_of_oli_bands;              /* Number of bands in band list */
    int number_of_scas;                   /* Number of SCAs per band */
    int status;                           /* Function return status */

    /* KPR5 report data */
    const char *kpr5_report_filename = NULL;
    KPR5_REPORT_DATA report_data;         /* Summary KPR 5 report
                                             information */

    /* Data buffer sizes */
    size_t buffer_size_angle_data;


    /*--------------------------------------------------------------------
                         INITIALIZATION
    ----------------------------------------------------------------------*/
    /* Initialize logging for this application. */
    ias_log_initialize("kpr5_evaluate");

    /* Make sure there are enough command-line parameters to run the app.
       Required parameters are the name of the CPF file and the name of
       the assessment report file. */
    if (argc != 3)
    {
        printf("USAGE:  kpr5_evaluate <CPF name> "
            "<name of KPR 5 report file>\n");
        exit(EXIT_FAILURE);
    }

    /* Get the CPF and report file names from the command line. */
    cpf_filename = argv[1];
    kpr5_report_filename = argv[2];

    /* Log a message indicating the assessment process has begun.  */
    IAS_LOG_INFO("KPR 5 assessment initiated");

    /* Initialize the satellite attributes library. */
    status = ias_sat_attr_initialize(IAS_L8);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attributes library");
        exit(EXIT_FAILURE);
    }

    /* Get the OLI band list and sort it so that the pan band is the first
       band number in the list. */
    status = ias_sat_attr_get_sensor_band_numbers(IAS_OLI, IAS_NORMAL_BAND,
        0, oli_band_number_list, IAS_MAX_NBANDS, &number_of_oli_bands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving list of 'normal' OLI imaging bands");
        exit(EXIT_FAILURE);
    }
    status = kpr_sort_band_number_list(oli_band_number_list,
        number_of_oli_bands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Sorting OLI band list");
        exit(EXIT_FAILURE);
    }

    /* Get the maximum number of SCAs for OLI. */
    number_of_scas = ias_sat_attr_get_sensor_sca_count(IAS_OLI);
    if (number_of_scas == ERROR)
    {
        IAS_LOG_ERROR("Getting number of SCAs for OLI sensor");
        exit(EXIT_FAILURE);
    }

    /* Allocate the data buffers.  Calculate the buffer sizes required to
       hold the data for all SCAs and bands.  */
    buffer_size_angle_data = number_of_scas * number_of_oli_bands
        * sizeof(double);

    across_max_angles_sca_band_radians = malloc(buffer_size_angle_data);
    across_min_angles_sca_band_radians = malloc(buffer_size_angle_data);
    ifovs_sca_band_radians = malloc(buffer_size_angle_data);

    /* Use calloc for this allocation so that the counts in the array
       structures get initialized to 0. */
    edge_counts_sca_band = calloc(number_of_scas * number_of_oli_bands,
        sizeof(KPR_INOPERABLE_EDGE_DETECTOR_COUNTS));

    if ((across_max_angles_sca_band_radians == NULL)
        || (across_min_angles_sca_band_radians == NULL)
        || (ifovs_sca_band_radians == NULL)
        || (edge_counts_sca_band == NULL))
    {
        IAS_LOG_ERROR("Buffer allocation failure");
        free_data_arrays(across_max_angles_sca_band_radians,
            across_min_angles_sca_band_radians, ifovs_sca_band_radians,
            edge_counts_sca_band);
        exit(EXIT_FAILURE);
    }

    /* Read the CPF. */
    cpf = ias_cpf_read(cpf_filename);
    if (cpf == NULL)
    {
        IAS_LOG_ERROR("Reading CPF %s", cpf_filename);
        free_data_arrays(across_max_angles_sca_band_radians,
            across_min_angles_sca_band_radians, ifovs_sca_band_radians,
            edge_counts_sca_band);
        exit(EXIT_FAILURE);
    }

    /* Get the detector status, focal plane, and earth constant information. */
    detector_status = ias_cpf_get_det_status(cpf);
    focal_plane     = ias_cpf_get_focal_plane(cpf);
    earth_constants = ias_cpf_get_earth_const(cpf);
    if ((detector_status == NULL) || (focal_plane == NULL)
        || (earth_constants == NULL))
    {
        IAS_LOG_ERROR("Retrieving parameter group(s) from CPF %s", argv[1]);
        ias_cpf_free(cpf);
        free_data_arrays(across_max_angles_sca_band_radians,
            across_min_angles_sca_band_radians, ifovs_sca_band_radians,
            edge_counts_sca_band);
        exit(EXIT_FAILURE);
    }

    /* Get the inoperable edge detector counts for all SCAs and bands. */
    status = kpr_get_inoperable_edge_detector_counts(oli_band_number_list,
        number_of_oli_bands, number_of_scas, detector_status, focal_plane,
        edge_counts_sca_band);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Counting inoperable edge detectors");
        ias_cpf_free(cpf);
        free_data_arrays(across_max_angles_sca_band_radians,
            across_min_angles_sca_band_radians, ifovs_sca_band_radians,
            edge_counts_sca_band);
        exit(EXIT_FAILURE);
    }

    /*----------------------------------------------------------------------
                        KPR 5 ASSESSMENT ANALYSIS
    -----------------------------------------------------------------------*/
    /* Get the earth's radius from the CPF earth constant semi-major axis value
       at the equator.  As this value is expressed in meters, multiply it by
       0.001 to convert it to km (which is the distance unit we'll use to do
       subsequent calculations). */
    cpf_earth_radius_km = 0.001 * earth_constants->semi_major_axis;

    /* Begin initialization of the report data structure.  */
    initialize_report_data(cpf_earth_radius_km, &report_data);

    /* Summarize inoperable detector information for SCAs 1 and 14. */
    status = summarize_inoperable_detector_information(oli_band_number_list,
        number_of_oli_bands, number_of_scas, edge_counts_sca_band,
        &report_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Summarizing SCA 1 / SCA 14 inoperable detector "
            "information");
        ias_cpf_free(cpf);
        free_data_arrays(across_max_angles_sca_band_radians,
            across_min_angles_sca_band_radians, ifovs_sca_band_radians,
            edge_counts_sca_band);
        exit(EXIT_FAILURE);
    }

    /* Calculate the crosstrack view angles for all SCAs and bands. */
    status = kpr_calculate_view_angles(focal_plane, oli_band_number_list,
        number_of_oli_bands, number_of_scas, KPR_CROSSTRACK_ANGLES,
        across_max_angles_sca_band_radians,
        across_min_angles_sca_band_radians, ifovs_sca_band_radians,
        &(report_data.unadjusted_max_angle_radians),
        &(report_data.unadjusted_min_angle_radians),
        &(report_data.max_ifov_radians));
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating cross-track view angle data");
        ias_cpf_free(cpf);
        free_data_arrays(across_max_angles_sca_band_radians,
            across_min_angles_sca_band_radians, ifovs_sca_band_radians,
            edge_counts_sca_band);
        exit(EXIT_FAILURE);
    }

    /* Calculate the swath width. */
    status = kpr5_calculate_swath_width(oli_band_number_list,
        number_of_oli_bands,
        number_of_scas, across_max_angles_sca_band_radians,
        across_min_angles_sca_band_radians, ifovs_sca_band_radians,
        edge_counts_sca_band, &report_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating final swath width");
        free_data_arrays(across_max_angles_sca_band_radians,
            across_min_angles_sca_band_radians, ifovs_sca_band_radians,
            edge_counts_sca_band);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }

    /* Free the inoperable detector counts array and set its pointer to
       NULL so that free_data_arrays won't have issues trying to free
       it again. */
    free(edge_counts_sca_band);
    edge_counts_sca_band = NULL;

    /* Check the swath width to see if the requirements were met. */
    assess_award_penalty(report_data.net_swath_width_km,
       &(report_data.award_penalty_percent));

    /*---------------------------------------------------------------------
                    WRITE KPR 5 ASSESSMENT REPORT
    -----------------------------------------------------------------------*/
    IAS_LOG_INFO("KPR 5 assessment analysis completed.  Writing results to "
        "assessment report %s...", kpr5_report_filename);

    status = kpr5_write_report(oli_band_number_list, number_of_oli_bands,
        number_of_scas, focal_plane, across_max_angles_sca_band_radians,
        across_min_angles_sca_band_radians, ifovs_sca_band_radians,
        kpr5_report_filename, &report_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing KPR 5 assessment report %s",
           kpr5_report_filename);
        free_data_arrays(across_max_angles_sca_band_radians,
            across_min_angles_sca_band_radians, ifovs_sca_band_radians,
            edge_counts_sca_band);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }

    /* Free the CPF resources and set the pointers to NULL. */
    ias_cpf_free(cpf);
    cpf = NULL;
    focal_plane = NULL;
    detector_status = NULL;

/* --------------------------------------------------------------------
                              CLEANUP
    -----------------------------------------------------------------------*/
    free_data_arrays(across_max_angles_sca_band_radians,
        across_min_angles_sca_band_radians, ifovs_sca_band_radians,
        edge_counts_sca_band);
    across_max_angles_sca_band_radians = NULL;
    across_min_angles_sca_band_radians = NULL;
    ifovs_sca_band_radians = NULL;

    /* Log successful completion of the assessment. */
    IAS_LOG_INFO("KPR 5 assessment completed successfully.");
    exit(EXIT_SUCCESS);

}   /* END -- kpr5_evaluate MAIN */
