/***************************************************************************
 NAME:            kpr5_write_report

 PURPOSE:         Generates the KPR 5 analysis summary report

 RETURNS:         Integer status code of SUCCESS or ERROR

 NOTES:           The pointers *across_xxx_angles_sca_band_radians
                  and *ifovs_sca_band_radians represent dynamically
                  allocated 1D arrays that DO NOT follow the standard
                  band/SCA ordering used in other IAS processing.  Refer
                  to the NOTE in kpr5.h for additional information.
****************************************************************************/
#include <stdio.h>
#include "ias_cpf.h"
#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "kpr5.h"


int kpr5_write_report
(
    const int *band_number_list,         /* I: List of bands to process */
    int number_of_bands,                 /* I: Number of bands to process */
    int number_of_scas,                  /* I: Number of SCAs */
    const struct IAS_CPF_FOCAL_PLANE *focal_plane,
                                         /* I: Focal plane parameters */
    const double *across_max_angles_sca_band_radians,
                                         /* I: 1D array containing the
                                            unadjusted maximum crosstrack
                                            angles in radians, ordered by
                                            SCA and band */
    const double *across_min_angles_sca_band_radians,
                                         /* I: 1D array containing the
                                            unadjusted minimum crosstrack
                                            angles in radians, ordered by
                                            SCA and band */
    const double *ifovs_sca_band_radians, /* I: 1D array containing the
                                            crosstrack IFOVs in radians,
                                            ordered by SCA and band */
    const char *rpt_filename,            /* I: Name of summary report file */
    const KPR5_REPORT_DATA *rpt_data     /* I: Populated summary report data
                                            structure */
)
{
    FILE *rpt_fptr = NULL;                 /* Report file handle */
    char attribute[512];                   /* Attribute string for Legendre
                                              coefficient */
    int  sca_index;                        /* Current SCA */
    int  status;                           /* Function return status */


    /* Open the report file. */
    rpt_fptr = fopen(rpt_filename, "w");
    if (rpt_fptr == NULL)
    {
        IAS_LOG_ERROR("Opening report file '%s' for writing",
            rpt_filename);
        return ERROR;
    }

    /* Write the header for the angle summary information. */
    status = fprintf(rpt_fptr, "\n                                        "
        "KPR #5 Swath Width Report\n\n        CPF Field"
        "                   c0             c1             c2        "
        "Maximum Angle   Minimum Angle      IFOV\n\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing opening header to report");
        fclose(rpt_fptr);
        return ERROR;
    }

    /* For the purposes of this report, we're interested only in SCAs
       1 and 14.  We need band_index and sca_index to access the
       CPF LOS Legendre parameters; the angle data buffers only need
       an offset representing the particular band/SCA combination.
       For convenience, set up writing the report information in a double
       loop (outer loop=SCA, inner loop=band). */
    for (sca_index = 0;
         sca_index < number_of_scas;
         sca_index += (number_of_scas - 1))
    {
        int band_list_index;            /* Band loop counter */
        int base_offset = (sca_index * number_of_bands);

        for (band_list_index = 0;
             band_list_index < number_of_bands;
             band_list_index++)
        {
            int buffer_offset = base_offset + band_list_index;

            /* Get the current band number in the list. */
            int band_number = band_number_list[band_list_index];

            /* Get the correct band index number for the current band.
               This is needed to correctly access the LOS legendre
               coefficients. */
            int band_index =
                ias_sat_attr_convert_band_number_to_index(band_number);
            if (band_index == ERROR)
            {
                IAS_LOG_ERROR("Converting band number to index value");
                fclose(rpt_fptr);
                return ERROR;
            }

            /* Write the attribute string with the angle information. */
            status = snprintf(attribute, sizeof(attribute),
                "Across_LOS_Legendre_B%02d_SCA%02d", band_number,
                (sca_index + 1));
            if ((status < 0) || (status >= sizeof(attribute)))
            {
                IAS_LOG_ERROR("Constructing attribute string for "
                    "band number %d SCA number %d angle information",
                    band_number, (sca_index + 1));
                fclose(rpt_fptr);
                return ERROR;
            }
            status = fprintf(rpt_fptr, "%s  %13.6e  %13.6e  %13.6e"
                "  %13.6e  %13.6e  %13.6e\n", attribute,
                focal_plane->across_los_legendre[band_index][sca_index][0],
                focal_plane->across_los_legendre[band_index][sca_index][1],
                focal_plane->across_los_legendre[band_index][sca_index][2],
                across_max_angles_sca_band_radians[buffer_offset],
                across_min_angles_sca_band_radians[buffer_offset],
                ifovs_sca_band_radians[buffer_offset]);
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing band number %d SCA number %d angle "
                    "summary information to report", band_number,
                    (sca_index + 1));
                fclose(rpt_fptr);
                return ERROR;
            }
        }
    }

    /* Write the KPR5-specific information here. */
    status = fprintf(rpt_fptr, "\n                                   "
        "Maximum Inboard Angle           %13.6e radians\n"
        "                                   "
        "Mininum Inboard Angle           %13.6e radians\n"
        "                                   "
        "Maximum IFOV for SCA 1          %13.6e radians\n"
        "                                   "
        "Maximum IFOV for SCA 14         %13.6e radians\n",
        rpt_data->unadjusted_max_angle_radians,
        rpt_data->unadjusted_min_angle_radians,
        rpt_data->max_sca1_ifov_radians,
        rpt_data->max_sca14_ifov_radians);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing overall summary angle information to report");
        fclose(rpt_fptr);
        return ERROR;
    }

    /* Write overall detector status informaton. */
    status = fprintf(rpt_fptr, "                                   "
        "Number of dead pan pixels             %d       at edge\n",
        rpt_data->number_dead_pan_detectors);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing pan band inoperable detector summary to "
            "report");
        fclose(rpt_fptr);
        return ERROR;
    }
    if (rpt_data->number_dead_ms_detectors == 0)
    {
        status = fprintf(rpt_fptr, "                                   "
            "Maximum number of dead MS pixels      %d"
            "       at edge\n",
            rpt_data->number_dead_ms_detectors);
    }
    else
    {
        status = fprintf(rpt_fptr, "                                   "
            "Maximum number of dead MS pixels      %d"
            "       at edge of band %d\n",
            rpt_data->number_dead_ms_detectors,
            rpt_data->max_dead_band_number);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing MS band inoperable detector summary to "
            "report");
        fclose(rpt_fptr);
        return ERROR;
    }

    /* Write the maximum and minimum adjusted cross-track angle summary
       information. */
    status = fprintf(rpt_fptr, "                                   "
        "Maximum Adjusted Angle          %13.6e radians on band %d\n"
        "                                   "
        "Minimum Adjusted Angle          %13.6e radians on band %d\n\n"
        "                                   "
        "Minimum Field of View           %13.6e radians\n"
        "                                   "
        "OLI Field of View                       %.2f degrees\n\n",
        rpt_data->adjusted_max_angle_radians,
        rpt_data->max_angle_band_number,
        rpt_data->adjusted_min_angle_radians,
        rpt_data->min_angle_band_number,
        rpt_data->min_ifov_radians,
        rpt_data->oli_fov_degrees);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing adjusted angle summary information to "
            "report");
        fclose(rpt_fptr);
        return ERROR;
    }

    /* Write the Earth summary information.  Although the swath width
       calculations were performed using units of km, report the quantities
       in meters. */
    status = fprintf(rpt_fptr, "                                   "
        "Earth Radius                          %7d meters\n"
        "                                   "
        "Spacecraft Altitude                   %7d meters\n"
        "                                   "
        "Spacecraft Radius                     %7d meters\n\n",
        (int)(rpt_data->earth_radius_km * 1000.0),
        (int)(rpt_data->spacecraft_altitude_km * 1000.0),
        (int)(rpt_data->spacecraft_orbit_radius_km * 1000.0));
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing Earth summary information to report");
        fclose(rpt_fptr);
        return ERROR;
    }

    /* Write out the summary starboard/port angle and swath information. */
    status = fprintf(rpt_fptr, "                                   "
        "Angle Right                     %13.6e radians\n"
        "                                   "
        "Angle Left                      %13.6e radians\n"
        "                                   "
        "Swath Angle                     %13.6e radians\n"
        "                                   "
        "Swath Width in Meters              %10.3f meters\n"
        "                                   "
        "Swath Width in Kilometers          %10.3f kilometers\n",
        rpt_data->starboard_angle_radians,
        rpt_data->port_angle_radians,
        rpt_data->swath_angle_radians,
        rpt_data->net_swath_width_km * 1000.0,
        rpt_data->net_swath_width_km);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing swath angle/width summary information to "
            "report");
        fclose(rpt_fptr);
        return ERROR;
    }

    /* Write out the penalty information. */
    status = fprintf(rpt_fptr, "\nAward Fee Penalty:          %d%%\n\n"
        "   Where:\n"
        "     %.1f <= swath width          is    0%%\n"
        "     %.1f <= swath width < %.1f  is    5%%\n"
        "     %.1f <= swath width < %.1f  is   10%%\n"
        "     %.1f <= swath width < %.1f  is   15%%\n"
        "     %.1f <= swath width < %.1f  is   20%%\n"
        "              swath width < %.1f  is  100%%\n",
        rpt_data->award_penalty_percent,
        KPR5_REQUIRED_SWATH_WIDTH_KM,
        KPR5_FIVE_PERCENT_PENALTY_SWATH_WIDTH_KM,
        KPR5_REQUIRED_SWATH_WIDTH_KM,
        KPR5_TEN_PERCENT_PENALTY_SWATH_WIDTH_KM,
        KPR5_FIVE_PERCENT_PENALTY_SWATH_WIDTH_KM,
        KPR5_FIFTEEN_PERCENT_PENALTY_SWATH_WIDTH_KM,
        KPR5_TEN_PERCENT_PENALTY_SWATH_WIDTH_KM,
        KPR5_TWENTY_PERCENT_PENALTY_SWATH_WIDTH_KM,
        KPR5_FIFTEEN_PERCENT_PENALTY_SWATH_WIDTH_KM,
        KPR5_TWENTY_PERCENT_PENALTY_SWATH_WIDTH_KM);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing award penalty summary information to report");
        fclose(rpt_fptr);
        return ERROR;
    }

    /* Close up the report.  Check the status to make sure it closed
       successfully. */
    if (fclose(rpt_fptr))
    {
        IAS_LOG_ERROR("Closing report file %s", rpt_filename);
        return ERROR;
    }

    return SUCCESS;
}
