#ifndef KPR6_H
#define KPR6_H

#include "kpr_shared.h"



/* Definition of data structure containing elements of KPR 6 summary
   report. */
typedef struct kpr6_report_data
{
    double adjusted_max_angle_radians;
                                     /* Smallest maximum crosstrack
                                        angle in radians, adjusted for
                                        inoperable detector effects */
    double adjusted_min_angle_radians;
                                     /* Largest minimum crosstrack
                                        angle in radians, adjusted for
                                        inoperable detector effects */
    double max_ifov_radians;         /* Maximum FOV in radians across all
                                        SCAs and bands */
    double max_along_angle_radians;  /* Largest alongtrack angle in radians
                                        across all SCAs and bands */
    double min_along_angle_radians;  /* Smallest alongtrack angle in
                                        radians across all SCAs and bands */
    double max_alongtrack_offset_radians;
                                     /* Largest along-track offset in
                                        radians */
    double yaw_steering_accuracy_radians;
                                     /* Yaw direction steering accuracy in
                                        radians */
    double max_crosstrack_steering_error_radians; 
                                     /* Largest cross-track steering error
                                        (yaw) in radians */
    double overlap_with_yaw_steering_error_radians;
                                     /* Overlap angle with yaw error
                                        correction in radians  */
    double min_angular_overlap_radians;
                                     /* Smallest angular overlap in radians
                                        between adjacent SCAs */
    double overlap_threshold_pixels; /* Smallest number of overlapping
                                        detectors that still meets
                                        the requirement */
    double min_detector_overlap_pixels;
                                     /* Smallest estimated detector overlap
                                        in pixels */
    int min_overlap_band_number;     /* Band number with minimum overlap */
    int min_overlap_sca_number;      /* Leftmost SCA number in consecutive
                                        SCA pair with minimum overlap */
    int num_overlap_bad_detectors;   /* Number of inoperable detectors in
                                        the overlapping SCAs */
}   KPR6_REPORT_DATA;



/* KPR6-specific function prototypes */

/* NOTE:  The pointers *across_xxx_angles_sca_band_radians,
   *along_xxx_angles_sca_band_radians, *ifovs_sca_band_radians, and
   *edge_counts_sca_band represent dynamically allocated 1D arrays with
   (number_of_oli_scas * number_of_oli_bands) elements.  The indexing into
   these arrays is calculated as

   array_index = sca_index * number_of_bands + band_list_index

   where band_list_index = 0 to number_of_bands - 1.  The band_list_index
   values correspond to the OLI band number list which is sorted as follows:

   band_list[number_of_bands] = {8, 1, 2, 3, 4, 5, 6, 7, 9}.

   Consequently, the data in these arrays are ordered as follows:

     [0] = SCA 01 band 8,  ...  [7] = SCA 01 band 7,    [8] = SCA 01 band 9,
     [9] = SCA 02 band 8,  ... [16] = SCA 02 band 7,   [17] = SCA 02 band 9,
    [18] = SCA 03 band 8,  ... [25] = SCA 03 band 7,   [26] = SCA 03 band 9,
    [27] = SCA 04 band 8,  ... [34[ = SCA 04 band 7,   [35] = SCA 04 band 9,
     ...                   ...                     ...
   [117] = SCA 14 band 8,  ...[124] = SCA 14 band 7,  [125] = SCA 14 band 9

   To emphasize the point...  THIS DOES NOT FOLLOW THE CONVENTIONAL
   BAND / SCA INDEXING USED IN OTHER IAS PROCESSING.  */

int kpr6_calculate_overlap
(
    const int *band_number_list,        /* I: List of bands to process */
    int number_of_bands,                /* I: Number of bands to process */
    int number_of_scas,                 /* I: Number of SCAs */
    const double *across_max_angles_sca_band_radians,
                                        /* I: 1D array of unadjusted maximum
                                           crosstrack angles in radians,
                                           ordered by SCA and band */
    const double *across_min_angles_sca_band_radians,
                                        /* I: 1D array of unadjusted minimum
                                           crosstrack angles in radians,
                                           ordered by SCA and band */
    const double *ifovs_sca_band_radians,
                                        /* I: 1D array containing the
                                           crosstrack IFOVs in radians,
                                           ordered by SCA and band*/
    const KPR_INOPERABLE_EDGE_DETECTOR_COUNTS *edge_counts_sca_band,
                                        /* I: 1D array of structures
                                           containing the counts of
                                           inoperable edge detectors,
                                           ordered by SCA and band */
    KPR6_REPORT_DATA *rpt_data          /* O: Populated summary report
                                           data structure */
);


int kpr6_write_report
(
    const char *rpt_filename,            /* I: Name of summary report file */
    const KPR6_REPORT_DATA *rpt_data     /* I: Summary report data
                                            structure */
);


#endif

