/**************************************************************************
 NAME:       kpr6_calculate_overlap

 PURPOSE:    Calculates the total pixel overlap between adjacent SCAs
             and performs the KPR 6 assessment of it

 RETURNS:    Integer status code of SUCCESS or ERROR

 NOTES:     The pointers *across_xxx_angles_sca_band_radians.
            *along_xxx_angles_sca_band_radians, *ifovs_sca_band_radians
            and *edge_counts_sca_band represent dynamically allocated 1D
            arrays that DO NOT follow the standard band/SCA ordering that is
            used in other IAS processing.  Refer to the NOTE in 'kpr6.h' for
            additional information.
***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "kpr6.h"


int kpr6_calculate_overlap
(
    const int *band_number_list,        /* I: List of bands to process */
    int number_of_bands,                /* I: Number of bands to process */
    int number_of_scas,                 /* I: Number of SCAs */
    const double *across_max_angles_sca_band_radians,
                                        /* I: 1D array of unadjusted maximum
                                           crosstrack angles in radians,
                                           ordered by SCA and band */
    const double *across_min_angles_sca_band_radians,
                                        /* I: 1D array of unadjusted minimum
                                           crosstrack angles in radians,
                                           ordered by SCA and band */
    const double *ifovs_sca_band_radians,
                                        /* I: 1D array containing the
                                           crosstrack IFOVs in radians,
                                           ordered by SCA and band*/
    const KPR_INOPERABLE_EDGE_DETECTOR_COUNTS *edge_counts_sca_band,
                                        /* I: 1D array of structures
                                           containing the counts of
                                           inoperable edge detectors,
                                           ordered by SCA and band */
    KPR6_REPORT_DATA *rpt_data          /* O: Populated summary report data
                                           structure */
)
{
    /* Dynamically allocated 1D arrays containing the 'adjusted'
       maximum and minimum crosstrack view angles.  The data in these
       arrays are ordered identically to the corresponding unadjusted
       crosstrack view angles as described above in the function prolog
       NOTES.  */
    double *adjusted_max_angles_sca_band_radians = NULL;
    double *adjusted_min_angles_sca_band_radians = NULL;
    size_t buffer_size =
        (number_of_bands * number_of_scas) * sizeof(double);

    /* SCA index. */
    int sca_index;


    /* Allocate the adjusted angle data buffers. */
    adjusted_max_angles_sca_band_radians = malloc(buffer_size);
    adjusted_min_angles_sca_band_radians = malloc(buffer_size);
    if ((adjusted_max_angles_sca_band_radians == NULL)
        || (adjusted_min_angles_sca_band_radians == NULL))
    {
        IAS_LOG_ERROR("Allocating adjusted angle data buffers");
        free(adjusted_max_angles_sca_band_radians);
        free(adjusted_min_angles_sca_band_radians);
        return ERROR;
    }

    /* Calculate the maximum alongtrack offset due to the allowable
       yaw-steering error of 525 microradians. */
    rpt_data->max_alongtrack_offset_radians =
        (tan(rpt_data->max_along_angle_radians)
        - tan(rpt_data->min_along_angle_radians));

    /* Calculate the maximum cross-track steering error. */
    rpt_data->max_crosstrack_steering_error_radians =
        atan(tan(rpt_data->yaw_steering_accuracy_radians)
             * rpt_data->max_alongtrack_offset_radians);

    /* Process all bands/SCAs. */
    for (sca_index = 0; sca_index < number_of_scas; sca_index++)
    {
        int base_current_offset = sca_index * number_of_bands;
        int base_previous_offset = (sca_index - 1) * number_of_bands;
        int band_list_index;

        for (band_list_index = 0;
             band_list_index < number_of_bands;
             band_list_index++)
        {
            double left_edge_correction;
            double right_edge_correction;

            /* Get the current band number. */
            int band_number = band_number_list[band_list_index];

            /* Determine the current and previous offsets for this band. */
            int current_offset = base_current_offset + band_list_index;
            int previous_offset = base_previous_offset + band_list_index;

            /* Calculate left edge and right edge correction factors to
               account for inoperable detectors. */
            left_edge_correction =
                edge_counts_sca_band[current_offset].left_edge_count
                * ifovs_sca_band_radians[current_offset];

            right_edge_correction =
                edge_counts_sca_band[current_offset].right_edge_count
                * ifovs_sca_band_radians[current_offset];

            /* Calculate the adjusted crosstrack angles. */
            adjusted_max_angles_sca_band_radians[current_offset] =
                across_max_angles_sca_band_radians[current_offset]
                - left_edge_correction;

            adjusted_min_angles_sca_band_radians[current_offset] =
                across_min_angles_sca_band_radians[current_offset]
                + right_edge_correction;

            /* Calculate the number of overlap pixels. */
            if (previous_offset >= 0)
            {
                double pixel_overlap =
                    (adjusted_max_angles_sca_band_radians[current_offset]
                    - adjusted_min_angles_sca_band_radians[previous_offset])
                    * 0.5
                    * ((1.0 / ifovs_sca_band_radians[current_offset])
                    + (1.0 / ifovs_sca_band_radians[previous_offset]));

                double angular_overlap_radians = pixel_overlap
                    * 0.5 * (ifovs_sca_band_radians[current_offset]
                    + ifovs_sca_band_radians[previous_offset]);

                /* Keep track of the minimum angular and detector overlaps,
                   the band number, affected SCA pair, and the number of
                   inoperable edge detectors. */
                if (angular_overlap_radians
                    < rpt_data->min_angular_overlap_radians)
                {
                    rpt_data->min_angular_overlap_radians =
                        angular_overlap_radians;
                    rpt_data->min_overlap_band_number = band_number;
                    rpt_data->min_overlap_sca_number  = sca_index + 1;
                    rpt_data->num_overlap_bad_detectors =
                        edge_counts_sca_band[current_offset]
                           .left_edge_count
                        + edge_counts_sca_band[previous_offset]
                           .right_edge_count;
                }
            }
        }
    }

    /* Calculate the net yaw steering error, then determine the minimum
       detector overlap. */
    rpt_data->overlap_with_yaw_steering_error_radians =
        rpt_data->min_angular_overlap_radians
        - rpt_data->max_crosstrack_steering_error_radians;

    rpt_data->min_detector_overlap_pixels =
        rpt_data->overlap_with_yaw_steering_error_radians
        / rpt_data->max_ifov_radians;

    /* Free the adjusted angle data buffers.  */
    free(adjusted_max_angles_sca_band_radians);
    free(adjusted_min_angles_sca_band_radians);
    adjusted_max_angles_sca_band_radians = NULL;
    adjusted_min_angles_sca_band_radians = NULL;

    return SUCCESS;

}   /* END -- kpr6_calculate_overlap */
