/*****************************************************************************
 NAME:           kpr6_evaluate

 PURPOSE:        Performs analysis of current focal plane information to
                 determine whether the KPR 6 performance requirement has
                 been met.

 RETURNS:        Integer status code of EXIT_SUCCESS (all processing
                 completed successfully) or EXIT_FAILURE (error condition
                 detected).
******************************************************************************/

/* Standard C library includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>

/* IAS system includes */
#include "ias_const.h"
#include "ias_cpf.h"
#include "ias_logging.h"
#include "ias_satellite_attributes.h"

/* Specific header include */
#include "kpr6.h"


/* #define for the required yaw steering accuracy (in radians). */
#define KPR6_YAW_STEERING_ACCURACY 525e-6

/* #define for the minimum required overlap between SCAs (in pixels) to
   allow meeting the requirement.  The required overlap corresponds to the
   current size of the Akima interpolation kernel used to perform the
   crosstrack resampling.  */
#define KPR6_REQUIRED_OVERLAP_PIXELS    6.0


/*****************************************************************************
 NAME:         initialize_report_data

 PURPOSE:      Initializes the KPR6 report data structure members

 RETURNS:      Nothing
******************************************************************************/
static void initialize_report_data
(
    KPR6_REPORT_DATA *rpt_data     /* I/O: Populated report summary data
                                      structure */
)
{
    /* Pre-initialize everything to 0. */
    memset(rpt_data, 0, sizeof(KPR6_REPORT_DATA));

    /* As with the swath width calculations for KPR 5, the overlap
       calculations require use of the innermost (SMALLEST)
       maximum crosstrack angle and  the outermost (LARGEST) minimum
       crosstrack angle.  These crosstrack angles have had corrections
       applied to account for inoperable detectors at the edges of each
       SCA.  */

    /* Initialize the compared min/max values to the system limits for a
       floating point value.  Set the adjusted minimum angle value to the
       negative of the largest system limit amd the adjusted maximum angle
       value to the largest system limit.  */
    rpt_data->adjusted_min_angle_radians  = -DBL_MAX;
    rpt_data->max_along_angle_radians     =  DBL_MAX;
    rpt_data->min_along_angle_radians     = -DBL_MAX;
    rpt_data->max_ifov_radians            = -DBL_MAX;
    rpt_data->min_angular_overlap_radians =  DBL_MAX;

    /* Set the yaw steering accuracy. */
    rpt_data->yaw_steering_accuracy_radians = KPR6_YAW_STEERING_ACCURACY;

    /* Set the minium overlap threshold in pixels that is considered
       to meet the requirement. */
    rpt_data->overlap_threshold_pixels = KPR6_REQUIRED_OVERLAP_PIXELS;

}   /* END -- initialize_report_data */


/***************************************************************************
 NAME:        free_data_arrays

 PURPOSE:     Frees the allocated data arrays

 RETURNS:     Nothing
****************************************************************************/
static void free_data_arrays
(
    double *across_max_angles_sca_band_radians,
                                      /* I: 1D array containing the
                                         maximum crosstrack view angles,
                                         ordered by SCA and band */
    double *across_min_angles_sca_band_radians,
                                      /* I: 1D array containing the
                                         minimum crosstrack view angles,
                                         ordered by SCA and band */
    double *along_max_angles_sca_band_radians,
                                      /* I: 1D array containing the
                                         maximum alongtrack view angles,
                                         ordered by SCA and band */
    double *along_min_angles_sca_band_radians,
                                      /* I: 1D array containing the
                                         minimum alongtrack view angles.
                                         ordered by SCA and band */
    double *ifovs_sca_band_radians,   /* I: 1D array containing the
                                         crosstrack IFOV angles, ordered by
                                         SCA and band */
    KPR_INOPERABLE_EDGE_DETECTOR_COUNTS *edge_counts_sca_band
                                      /* I: 1D array of structures
                                         containing counts of inoperable
                                         edge detectors, ordered by SCA and
                                         band */
)
{
    free(across_max_angles_sca_band_radians);
    free(across_min_angles_sca_band_radians);
    free(along_max_angles_sca_band_radians);
    free(along_min_angles_sca_band_radians);
    free(ifovs_sca_band_radians);
    free(edge_counts_sca_band);

}   /* END -- free_data_arrays */



/*=========================================================================
                       MAIN FUNCTION
===========================================================================*/
int main(int argc, char *argv[])
{
  /*=====================================================================*/

    /* NOTE:  The following set of pointers are declared for dynamically
       allocated 1D arrays that DO NOT follow the conventional band / SCA
       ordering that other IAS processing uses.  Refer to the NOTE in
       kpr6.h for additional information.  */

    /* Array containing counts for inoperable
       detectors on the edges of each SCA.  The array element is a
       structure containing both left edge and right edge counts.  */
    KPR_INOPERABLE_EDGE_DETECTOR_COUNTS *edge_counts_sca_band = NULL;

    /* Arrays containing view angles derived from the crosstrack LOS Legendre
       coefficients. */
    double *across_max_angles_sca_band_radians = NULL;
    double *across_min_angles_sca_band_radians = NULL;

    /* Arrays containing view angles derived from the alongtrack LOS Legendre
       coefficients. */
    double *along_max_angles_sca_band_radians = NULL;
    double *along_min_angles_sca_band_radians = NULL;

    /* Array containing the crosstrack IFOV angles. */
    double *ifovs_sca_band_radians = NULL;

  /*=====================================================================*/

    /* CPF parameters. */
    const char *cpf_filename = NULL;
    IAS_CPF *cpf = NULL;
    const struct IAS_CPF_DETECTOR_STATUS *detector_status = NULL;
    const struct IAS_CPF_FOCAL_PLANE *focal_plane = NULL;

    /* Processing parameters */
    int oli_band_number_list[IAS_MAX_NBANDS]; /* Band list to process */
    int number_of_oli_bands;              /* Number of bands in band list */
    int number_of_scas;                   /* Number of SCAs per band */
    int status;                           /* Function return status */

    /* KPR6 report data */
    const char *kpr6_report_filename = NULL;
    KPR6_REPORT_DATA report_data;         /* Summary KPR 6 report
                                             information */

    /* Data buffer sizes */
    size_t buffer_size_angle_data;


    /*--------------------------------------------------------------------
                           INITIALIZATION
    ----------------------------------------------------------------------*/
    /* Initialize logging for this application. */
    ias_log_initialize("kpr6_evaluate");

    /* Make sure we've got enough command-line parameters. */
    if (argc != 3)
    {
        printf("USAGE:  kpr6_evaluate <name of CPF> "
            "<name of KPR 6 report file>\n");
        exit(EXIT_FAILURE);
    }

    /* Get the parameters from the command line.  Make sure the threshold
       parameter converts to a valid number as expected. */
    cpf_filename = argv[1];
    kpr6_report_filename = argv[2];

    /* Log a message indicating the assessment process has begun.  */
    IAS_LOG_INFO("KPR 6 assessment initiated");

    /* Initialize the satellite attributes library. */
    status = ias_sat_attr_initialize(IAS_L8);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attributes library");
        exit(EXIT_FAILURE);
    }

    /* Get the OLI band list and sort it so that the pan band is the first
       band number in the list. */
    status = ias_sat_attr_get_sensor_band_numbers(IAS_OLI, IAS_NORMAL_BAND,
        0, oli_band_number_list, IAS_MAX_NBANDS, &number_of_oli_bands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving list of 'normal' OLI imaging bands");
        exit(EXIT_FAILURE);
    }
    status = kpr_sort_band_number_list(oli_band_number_list,
        number_of_oli_bands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reordering OLI band list");
        exit(EXIT_FAILURE);
    }

    /* Get the maximum number of SCAs for OLI. */
    number_of_scas = ias_sat_attr_get_sensor_sca_count(IAS_OLI);
    if (number_of_scas == ERROR)
    {
        IAS_LOG_ERROR("Getting number of SCAs for OLI sensor");
        exit(EXIT_FAILURE);
    }

    /* Allocate the data buffers.  Calculate the buffer size required to
       hold the data from all SCAs and bands.  These buffers should be
       considered as 1D arrays.  */
    buffer_size_angle_data = number_of_scas * number_of_oli_bands
        * sizeof(double);

    across_max_angles_sca_band_radians = malloc(buffer_size_angle_data);
    across_min_angles_sca_band_radians = malloc(buffer_size_angle_data);
    along_max_angles_sca_band_radians = malloc(buffer_size_angle_data);
    along_min_angles_sca_band_radians = malloc(buffer_size_angle_data);
    ifovs_sca_band_radians = malloc(buffer_size_angle_data);

    /* Use calloc for this allocation so that the counts in the array
       structures get initialized to 0. */
    edge_counts_sca_band = calloc(number_of_scas * number_of_oli_bands,
        sizeof(KPR_INOPERABLE_EDGE_DETECTOR_COUNTS));

    if ((across_max_angles_sca_band_radians == NULL)
        || (across_min_angles_sca_band_radians == NULL)
        || (along_max_angles_sca_band_radians == NULL)
        || (along_min_angles_sca_band_radians == NULL)
        || (ifovs_sca_band_radians == NULL)
        || (edge_counts_sca_band == NULL))
    {
        IAS_LOG_ERROR("Buffer allocation failure");
        free_data_arrays(across_max_angles_sca_band_radians,
            across_min_angles_sca_band_radians,
            along_max_angles_sca_band_radians,
            along_min_angles_sca_band_radians,
            ifovs_sca_band_radians,
            edge_counts_sca_band);
        exit(EXIT_FAILURE);
    }

    /* Read the CPF. */
    IAS_LOG_INFO("Reading detector status and focal plane information "
        "from CPF %s...", cpf_filename);

    cpf = ias_cpf_read(cpf_filename);
    if (cpf == NULL)
    {
        IAS_LOG_ERROR("Reading CPF %s", cpf_filename);
        free_data_arrays(across_max_angles_sca_band_radians,
            across_min_angles_sca_band_radians,
            along_max_angles_sca_band_radians,
            along_min_angles_sca_band_radians,
            ifovs_sca_band_radians,
            edge_counts_sca_band);
        exit(EXIT_FAILURE);
    }

    /* Get the detector status and focal plane information. */
    detector_status = ias_cpf_get_det_status(cpf);
    focal_plane     = ias_cpf_get_focal_plane(cpf);
    if ((detector_status == NULL) || (focal_plane == NULL))
    {
        IAS_LOG_ERROR("Retrieving parameter group(s) from CPF %s",
             cpf_filename);
        ias_cpf_free(cpf);
        free_data_arrays(across_max_angles_sca_band_radians,
            across_min_angles_sca_band_radians,
            along_max_angles_sca_band_radians,
            along_min_angles_sca_band_radians,
            ifovs_sca_band_radians,
            edge_counts_sca_band);
        exit(EXIT_FAILURE);
    }

    /* Get the inoperable edge detector counts , ordered by SCA and band. */
    status = kpr_get_inoperable_edge_detector_counts(oli_band_number_list,
        number_of_oli_bands, number_of_scas, detector_status,
        focal_plane, edge_counts_sca_band);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Counting inoperable edge detectors");
        ias_cpf_free(cpf);
        free_data_arrays(across_max_angles_sca_band_radians,
            across_min_angles_sca_band_radians,
            along_max_angles_sca_band_radians,
            along_min_angles_sca_band_radians,
            ifovs_sca_band_radians,
            edge_counts_sca_band);
        exit(EXIT_FAILURE);
    }

    /*---------------------------------------------------------------------
                      KPR 6 ASSESSMENT ANALYSIS
    -----------------------------------------------------------------------*/
    /* Begin initialization of report data structure. */
    initialize_report_data(&report_data);

    /* Calculate the crosstrack view angles for all SCAs and bands. We don't
       need the smallest maximum crosstrack angle or largest minimum
       crosstrack angle for analysis or reporting, so pass in NULL
       for these parameters.  */
    status = kpr_calculate_view_angles(focal_plane, oli_band_number_list,
        number_of_oli_bands, number_of_scas, KPR_CROSSTRACK_ANGLES,
        across_max_angles_sca_band_radians,
        across_min_angles_sca_band_radians, ifovs_sca_band_radians,
        NULL, NULL, &(report_data.max_ifov_radians));
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating cross-track view angle data");
        ias_cpf_free(cpf);
        free_data_arrays(across_max_angles_sca_band_radians,
            across_min_angles_sca_band_radians,
            along_max_angles_sca_band_radians,
            along_min_angles_sca_band_radians,
            ifovs_sca_band_radians,
            edge_counts_sca_band);
        exit(EXIT_FAILURE);
    }

    /* Calculate the alongtrack view angles for all SCAs and bands.  These
       are needed only to pick up the maximum and minimum alongtrack
       view angles across all SCAs and bands, which are used in calculating an
       alongtrack offset used in the overlap calculation.  We already have
       the IFOV data accounted for from the crosstrack angle calculations, so
       pass in NULL for the IFOV array and the maximum IFOV value. */
    status = kpr_calculate_view_angles(focal_plane, oli_band_number_list,
        number_of_oli_bands, number_of_scas, KPR_ALONGTRACK_ANGLES,
        along_max_angles_sca_band_radians,
        along_min_angles_sca_band_radians, NULL,
        &(report_data.max_along_angle_radians),
        &(report_data.min_along_angle_radians),
        NULL);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating along-track view angle data");
        ias_cpf_free(cpf);
        free_data_arrays(across_max_angles_sca_band_radians,
            across_min_angles_sca_band_radians,
            along_max_angles_sca_band_radians,
            along_min_angles_sca_band_radians,
            ifovs_sca_band_radians,
            edge_counts_sca_band);
        exit(EXIT_FAILURE);
    }

    /* Free the CPF resources.  Set its pointer to NULL. */
    ias_cpf_free(cpf);
    cpf = NULL;

    /* Free the alongtrack angle arrays -- they're not needed anymore.
       Set their pointers to NULL so that free_data_arrays() won't have
       issues trying to free them again. */
    free(along_max_angles_sca_band_radians);
    free(along_min_angles_sca_band_radians);
    along_max_angles_sca_band_radians = NULL;
    along_min_angles_sca_band_radians = NULL;

    /* Calculate the overlap information. */
    status = kpr6_calculate_overlap(oli_band_number_list, number_of_oli_bands,
        number_of_scas, across_max_angles_sca_band_radians,
        across_min_angles_sca_band_radians, ifovs_sca_band_radians,
        edge_counts_sca_band, &report_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating overlap");
        free_data_arrays(across_max_angles_sca_band_radians,
            across_min_angles_sca_band_radians,
            along_max_angles_sca_band_radians,
            along_min_angles_sca_band_radians,
            ifovs_sca_band_radians,
            edge_counts_sca_band);
        exit(EXIT_FAILURE);
    }

    /* Free the data arrays -- we don't need them anymore.  Set their
       pointers back to NULL. */
    free_data_arrays(across_max_angles_sca_band_radians,
        across_min_angles_sca_band_radians,
        along_max_angles_sca_band_radians,
        along_min_angles_sca_band_radians,
        ifovs_sca_band_radians,
        edge_counts_sca_band);
    edge_counts_sca_band = NULL;
    across_min_angles_sca_band_radians = NULL;
    across_max_angles_sca_band_radians = NULL;
    ifovs_sca_band_radians = NULL;

    /*----------------------------------------------------------------------
                      WRITE KPR 6 ASSESSMENT REPORT
    ------------------------------------------------------------------------*/
    IAS_LOG_INFO("KPR 6 assessment analysis completed.  Writing results to "
        "assessment report %s...", kpr6_report_filename);

    status = kpr6_write_report(kpr6_report_filename, &report_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing KPR 6 assessment report %s",
            kpr6_report_filename);
        exit(EXIT_FAILURE);
    }

    /* Log a successful completion message. */
    IAS_LOG_INFO("KPR 6 assessment completed successfully.");
    exit(EXIT_SUCCESS);

}   /* END -- kpr6_evaluate MAIN */
