/***************************************************************************
 NAME:            kpr6_write_report

 PURPOSE:         Generates the KPR 6 analysis summary report

 RETURNS:         Integer status code of SUCCESS or ERROR
****************************************************************************/
#include <stdio.h>

#include "ias_cpf.h"
#include "ias_logging.h"
#include "ias_satellite_attributes.h"

#include "kpr6.h"


int kpr6_write_report
(
    const char *rpt_filename,              /* I: Name of summary report file */
    const KPR6_REPORT_DATA *rpt_data       /* I: Summary report data
                                              structure */
)
{
    FILE *rpt_fptr = NULL;                 /* Summary report file */
    char penalty_string[5];                /* Penalty percentage written as
                                              a string (includes NULL
                                              terminator) */
    int  status;                           /* Function return status */


    /* Open the report file. */
    rpt_fptr = fopen(rpt_filename, "w");
    if (rpt_fptr == NULL)
    {
        IAS_LOG_ERROR("Opening report file '%s' for writing", rpt_filename);
        return ERROR;
    }

    /* Write the KPR6-specific steering accuracy and overlap information to
       the report. */
    status = fprintf(rpt_fptr, "\n                                       "
        "KPR #6 Detector Overlap Report\n\n"
        "                         "
        "Yaw Steering Accuracy:                   %12.6e   radians\n"
        "                         "
        "Maximum Along-Track Offset:              %12.6e   radians\n"
        "                         "
        "Maximum Cross-Track Yaw Steering Error:  %12.6e   radians\n\n"
        "                         "
        "Minimum Angular Overlap:                 %12.6e   radians\n"
        "                         "
        "Overlap with Yaw Steering Error:         %12.6e   radians\n"
        "                         Least Overlap Information:\n"
        "                            Band Number:         %2d\n"
        "                            SCA Pair:          %2d/%d\n"
        "                            Number Dead:          %d\n"
        "                            Overlap:      %6.1f pixels\n",
        rpt_data->yaw_steering_accuracy_radians,
        rpt_data->max_alongtrack_offset_radians,
        rpt_data->max_crosstrack_steering_error_radians,
        rpt_data->min_angular_overlap_radians,
        rpt_data->overlap_with_yaw_steering_error_radians,
        rpt_data->min_overlap_band_number,
        rpt_data->min_overlap_sca_number,
        rpt_data->min_overlap_sca_number + 1,
        rpt_data->num_overlap_bad_detectors,
        rpt_data->min_detector_overlap_pixels);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing overlap summary information to report");
        fclose(rpt_fptr);
        return ERROR;
    }

    /* The penalty is basically all or nothing -- if the estimated
       detector overlap is less than the overlap threshold, the penalty
       is 100%.  If it's greater than the overlap threshold, there is no
       penalty. */
    if (rpt_data->min_detector_overlap_pixels
        < rpt_data->overlap_threshold_pixels)
    {
        sprintf(penalty_string, "100%%");
    }
    else
    {
        sprintf(penalty_string, "0%% (NONE)");
    }

    /* Write out the penalty information. */
    status = fprintf(rpt_fptr, "\n            Award Fee Penalty (based on "
        "%.1f pixel threshold):    %s\n"
        "                  Where:\n"
        "                     %.1f <= overlap         is    0%%\n"
        "                            overlap < %.1f   is  100%%\n",
        rpt_data->overlap_threshold_pixels, penalty_string,
        rpt_data->overlap_threshold_pixels,
        rpt_data->overlap_threshold_pixels);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing penalty summary information to report");
        fclose(rpt_fptr);
        return ERROR;
    }

    /* Close up the report.  Check the status to make sure it closed
       successfully. */
    if (fclose(rpt_fptr))
    {
        IAS_LOG_ERROR("Closing report file %s", rpt_filename);
        return ERROR;
    }

    return SUCCESS;
}
