#ifndef KPR7_H
#define KPR7_H


/* #defines to cover the RMSE thresholds (in meters) used in the assessment. */
#define KPR7_FULL_PENALTY_RMSE_METERS                 8.4
#define KPR7_FIFTY_PERCENT_PENALTY_RMSE_METERS        6.6
#define KPR7_TWENTYFIVE_PERCENT_PENALTY_RMSE_METERS   4.7
#define KPR7_NO_PENALTY_RMSE_METERS                   4.3
#define KPR7_TWENTYFIVE_PERCENT_BONUS_RMSE_METERS     3.8
#define KPR7_THIRTYSEVEN_PERCENT_BONUS_RMSE_METERS    3.7


/* Data structure containing summary line and sample RMSE statistics.
   The 'le90' component in the variable names refers to the average unweighted
   RMSE (line or sample) in meters expressed in terms of 90% linear error
   (LE90).  The LE90 RMSE is just the average unweighted RMSE in meters
   multiplied by 1.6449.  */
typedef struct kpr7_rmse_summary_statistics
{
    double unweighted_line_rmse_pixels;     /* average unweighted line RMSE
                                               in pixels */
    double weighted_line_rmse_pixels;       /* average weighted line RMSE
                                               in pixels */
    double unweighted_sample_rmse_pixels;   /* average unweighted sample
                                               RMSE in pixels */
    double weighted_sample_rmse_pixels;     /* average weighted sample
                                               RMSE in pixels */
    double delta_line_rmse_pixels;          /* difference between average
                                               weighted and unweighted line
                                               RMSEs in pixels */ 
    double delta_sample_rmse_pixels;        /* difference between average
                                               weighted and unweighted
                                               sample RMSEs in pixels */
    double unweighted_line_rmse_meters;     /* average unweighted line RMSE
                                               in meters */
    double unweighted_sample_rmse_meters;   /* average unweighed sample RMSE
                                               in meters */
    double line_le90_rmse_meters;           /* line LE90 RMSE in meters */
    double sample_le90_rmse_meters;         /* sample LE90 RMSE in meters */
    int reference_band_number;              /* reference band number */
    int search_band_number;                 /* search band number */
    int band_code_number;                   /* composite identifier for
                                               a band pair based on the
                                               reference and search band
                                               numbers.  It's equal
                                               to 100*reference_band_number
                                               + search_band_number */
}   KPR7_SUMMARY_STATISTICS;


/* Definition of the data structure containing the information
   for the official and cal/val analysis reports.  */
typedef struct kpr7_report_data
{
    int band_pair_count;             /* Number of band pairs with valid
                                        data */
    KPR7_SUMMARY_STATISTICS *bp_stats;
                                     /* 1D array of KPR7_SUMMARY_STATISTICS
                                        structures for the number of band
                                        pairs that have valid data */
    double ground_sample_distance_meters;
                                     /* Ground sample distance in meters */
    double max_line_le90_rmse_meters;/* Largest line LE90 RMSE in meters */
    double max_sample_le90_rmse_meters;
                                     /* Largest sample LE90 RMSE in meters */
    double max_le90_rmse_meters;     /* Largest LE90 RMSE in meters (line
                                        or sample) rounded to the nearest
                                        0.1 meters. */
    int award_fee_adjustment_percent;/* Amount of adjustment to base
                                        contract award based on meeting
                                        KPR specs */

    const char *acquisition_date_range;    /* Acquisition date range of 
                                        analysed data */
}   KPR7_REPORT_DATA;


/* KPR7-specific function prototypes */


/* A KPR7_RMSE_SUM_OF_SQUARES data structure contains the members for the
   line and sample rmse sum-of-squares calculations for a particular
   (reference band, search band) pair, the number of database records
   retrieved for the band pair, and the number of valid correlation points
   associated with the set of line/sample RMSE values.  */

/* Data structure containing line and sample RMSE sum-of-squares. */
typedef struct kpr7_rmse_sum_of_squares
{
    double unweighted_line_sumsq;
    double unweighted_sample_sumsq;
    double weighted_line_sumsq;
    double weighted_sample_sumsq;
    int number_of_database_records;
    int number_of_valid_correlation_points;
}   KPR7_RMSE_SUM_OF_SQUARES;

/* A dynamically allocated 1D array of these structures containing
   (number_of_bands * number_of_bands) elements is passed in as an input to
   kpr7_calculate_summary_rmse_statistics().  The array structures can then
   be accessed using an index calculated as

   index = reference_band_index * number_of_bands + search_band_index

   where reference_band_index and search_band_index are the indices
   corresponding to the reference and search band numbers returned from the 
   ias_sat_attr_convert_band_number_to_index() function.

   reference_band_number and search_band_number are read from each
   data line of the and-to-band (B2B) post-processing report (generated by
   a tool that queries the database to return band-to-band registration
   line/sample RMSE results for a selected set of scenes).  In general, the
   band pairs where RMSE data can be listed in the current B2B
   post-processing report are given as follows (N=not listed, Y=listed if data
   are present).

    REF  |
    BAND |               SEARCH BAND #
    #    | 1     2     3     4     5     6     7     8     9
    --------------------------------------------------------
     1   | N     Y     Y     Y     Y     Y     Y     N     Y
     2   | N     N     Y     Y     Y     Y     Y     N     Y
     3   | N     N     N     Y     Y     Y     Y     N     Y
     4   | N     N     N     N     Y     Y     Y     N     Y
     5   | N     N     N     N     N     Y     Y     N     Y
     6   | N     N     N     N     N     N     Y     N     Y
     7   | N     N     N     N     N     N     N     N     Y
     8   | Y     Y     Y     Y     Y     Y     Y     N     Y
     9   | N     N     N     N     N     N     N     N     N


     The band_number_list array has been sorted in the following
     order:

     band_number_list[number_of_bands] = {8, 1, 2, 3, 4, 5, 6, 7, 9}

     This order is intended to follow the band list pairs where data
     can be shown, as indicated in the chart above. It also means that the
     pan band band pairs will be listed in the official summary
     report first, followed by the band 1 band pairs, the band 2 band pairs,
     etc.  */

int kpr7_calculate_summary_rmse_statistics
(
    const int *band_number_list,             /* I: List of bands to process */
    int number_of_bands,                     /* I: Total number of bands to
                                                process */
    const KPR7_RMSE_SUM_OF_SQUARES *rmse_sum_of_squares,
                                             /* I: RMSE sum-of-squares data
                                                for all valid band pairs */
    KPR7_REPORT_DATA *report_data            /* O: KPR7 summary report
                                                information */
);


int kpr7_write_official_report
(
    const char *report_filename,         /* I: Name of official report file */
    const KPR7_REPORT_DATA *report_data  /* I: KPR7 summary report data */
);


int kpr7_write_calval_analysis_report
(
    const char *report_filename,         /* I: Name of calval analysis file */
    const KPR7_REPORT_DATA *report_data  /* I: KPR7 summary report data */
);


#endif



