/****************************************************************************
 NAME:      kpr7_calculate_summary_rmse_statistics

 PURPOSE:   Calculates the summary statistical information required for the
            KPR 7 assessment.

 RETURNS:   Integer status code of SUCCESS or ERROR

 NOTES:     A KPR7_RMSE_SUM_OF_SQUARES data structure contains the
            members for the line and sample rmse sum-of-squares
            calculations for a particular (reference band, search band)
            pair, the number of database records retrieved for the
            band pair, and the number of valid correlation points associated
            with the set of line/sample RMSE values.  A dynamically
            allocated 1D array of these structures containing
            (number_of_bands * number_of_bands) elements is passed in as
            an input.  The array structures can then be accessed using an
            index calculated as

            index = reference_band_index * number_of_bands + search_band_index

            where reference_band_index and search_band_index are the
            band indices for the reference and search band numbers returned
            from the 'ias_sat_attr_convert_band_number_to_index()' function.

            reference_band_number and search_band_number are read from
            each data line of the band-to-band (B2B) post-processing
            report (generated by a tool that queries the database to return
            band-to-band registration line/sample RMSE results for a selected
            set of scenes).  In general, the band pairs where RMSE data can be
            listed in the current B2B post-processing report are given as
            follows (N=not listed, Y=listed if data are present).

            REF  |
            BAND |               SEARCH BAND #
            #    | 1     2     3     4     5     6     7     8     9
            --------------------------------------------------------
             1   | N     Y     Y     Y     Y     Y     Y     N     Y
             2   | N     N     Y     Y     Y     Y     Y     N     Y
             3   | N     N     N     Y     Y     Y     Y     N     Y
             4   | N     N     N     N     Y     Y     Y     N     Y
             5   | N     N     N     N     N     Y     Y     N     Y
             6   | N     N     N     N     N     N     Y     N     Y
             7   | N     N     N     N     N     N     N     N     Y
             8   | Y     Y     Y     Y     Y     Y     Y     N     Y
             9   | N     N     N     N     N     N     N     N     N

*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "kpr7.h"


#define SCALE_LE90 1.6449             /* Scale factor used to calculate
                                         average (unweighted) line and
                                         sample RMSEs in meters
                                         corresponding to a 90% linear
                                         error (LE90) */


int kpr7_calculate_summary_rmse_statistics
(
    const int *band_number_list,             /* I: List of bands to process */
    int number_of_bands,                     /* I: Total number of bands to
                                                process */
    const KPR7_RMSE_SUM_OF_SQUARES *rmse_sum_of_squares,
                                             /* I: RMSE sum-of-squares data
                                                for all valid band pairs */
    KPR7_REPORT_DATA *report_data            /* O: KPR7 summary report
                                                information */
)
{
    int band_pair_count = 0;          /* Count of distinct reference/search
                                         band pairs that have data */
    int band_pair_index = 0;          /* Index into arrays containing
                                         summary RMSE statistics */
    int i;                            /* Loop counter */
    int j;                            /* Loop counter */
    int start_band_list_index;        /* Index into the band number list
                                         for a search band */
    int reference_band_number;        /* Current reference band number */
    int reference_band_index;         /* Index corresponding to the current
                                         reference band number*/
    int search_band_number;           /* Current search band number */
    int search_band_index;            /* Index corresponding to the current
                                         search band number */
    int rmse_array_index;             /* Offset into the
                                         KPR7_RMSE_SUM_OF_SQUARES structure
                                         array */
    int pan_band_number;
    int cirrus_band_number;

    /* Get the pan band number. */
    pan_band_number = ias_sat_attr_get_band_number_from_type(IAS_PAN_BAND);
    if (pan_band_number == ERROR)
    {
        IAS_LOG_ERROR("Retrieving pan band number");
        return ERROR;
    }

    /* Get the Cirrus band number. */
    cirrus_band_number =
        ias_sat_attr_get_band_number_from_type(IAS_CIRRUS_BAND);
    if (cirrus_band_number == ERROR)
    {
        IAS_LOG_ERROR("Retrieving Cirrus band number");
        return ERROR;
    }

    /* In a first pass, count the number of reference/search band combinations
       that have data.  In a valid combination, the reference band number does
       not equal the search band number.  As noted in the comment for the
       band_number_list parameter, the list is assumed/expected to be ordered
       such that:
        i.  The pan band is first in the list (if it is specified in the
            list), and the remaining band numbers are sorted in ascending
            order
       ii.  If the pan band is not specified in the list, the band numbers
            are sorted in ascending order.  */
    for (i = 0; i < number_of_bands; i++)
    {
        int reference_band_index;
        int search_band_index;

        /* Set the current band number as the reference band number. */
        reference_band_number = band_number_list[i];

        /* If the reference band number is the pan band, use all other
           bands as the search band. Otherwise, start looking for search
           bands with the next band number in the list.  */
        if (reference_band_number == pan_band_number)
        {
            start_band_list_index = 0;
        }
        else
        {
            start_band_list_index = i + 1;
        }

        /* Get the band index for the current reference band number. */
        reference_band_index =
            ias_sat_attr_convert_band_number_to_index(reference_band_number);
        if (reference_band_index == ERROR)
        {
            IAS_LOG_ERROR("Invalid reference band number %d",
                reference_band_number);
            return ERROR;
        }

        /* Process the remaining bands as a potential search band. */
        for (j = start_band_list_index; j < number_of_bands; j++)
        {
            /* Get the current search band number. */
            search_band_number = band_number_list[j];

            /* If the current search band is the pan band, skip it. */
            if (search_band_number == pan_band_number)
            {
                continue;
            }

            /* Get the band index for the current search band number. */
            search_band_index =
                ias_sat_attr_convert_band_number_to_index(search_band_number);
            if (search_band_index == ERROR)
            {
                IAS_LOG_ERROR("Invalid search band number %d",
                    search_band_number);
                return ERROR;
            }

            /* Calculate the index into the sum-of-squares structure
               array for the current (reference band, search band) pair. */
            rmse_array_index = reference_band_index * number_of_bands
                + search_band_index;

            /* If this band pair has data, count it. */
            if (rmse_sum_of_squares[rmse_array_index]
                .number_of_valid_correlation_points > 0)
            {
                band_pair_count++;
            }
        }
    }

    /* If there's no band pair with data, that's an error. */
    if (band_pair_count == 0)
    {
        IAS_LOG_ERROR("No reference/search band pair with data to analyze");
        return  ERROR;
    }
    report_data->band_pair_count = band_pair_count;

    /* Allocate an array of structures to hold the summary statistics
       information. */
    report_data->bp_stats = calloc(band_pair_count,
        sizeof(KPR7_SUMMARY_STATISTICS));
    if (report_data->bp_stats == NULL)
    {
        IAS_LOG_ERROR("Allocating summary statistics data block ");
        return ERROR;
    }

    /* Process the data in a second pass.  */
    for (i = 0; i < number_of_bands; i++)
    {
        /* Set the current band number as the reference band number. */
        reference_band_number = band_number_list[i];

        /* If the reference band number is the pan band, use all other
           bands as the search band. */
        if (reference_band_number == pan_band_number)
        {
            start_band_list_index = 0;
        }
        else
        {
            start_band_list_index = i + 1;
        }

        /* Get the band index for the current reference band number. */
        reference_band_index =
            ias_sat_attr_convert_band_number_to_index(reference_band_number);
        if (reference_band_index == ERROR)
        {
            IAS_LOG_ERROR("Invalid reference band number %d",
                reference_band_number);
            free(report_data->bp_stats);
            return ERROR;
        }

        /* Process the remaining bands as a potential search band. */
        for (j = start_band_list_index; j < number_of_bands; j++)
        {
            /* Get the current search band number. */
            search_band_number = band_number_list[j];

            /* If the search band number is the pan band, skip it. */
            if (search_band_number == pan_band_number)
            {
                continue;
            }

            /* Get the band index for the current search band. */
            search_band_index =
                ias_sat_attr_convert_band_number_to_index(search_band_number);
            if (search_band_index == ERROR)
            {
                IAS_LOG_ERROR("Invalid search band number %d",
                    search_band_number);
                free(report_data->bp_stats);
                return ERROR;
            }

            /* Calculate the index into the sum-of-squares structure
               array for the current (reference band, search band) pair. */
            rmse_array_index = reference_band_index * number_of_bands
                + search_band_index;

            /* Process all combinations with valid data. */
            if (rmse_sum_of_squares[rmse_array_index]
                .number_of_valid_correlation_points > 0)
            {
                KPR7_SUMMARY_STATISTICS *curr_stats =
                    &(report_data->bp_stats[band_pair_index]);

                /* Save the reference and search band numbers fpr the
                   current (reference band, search band) pair. */
                curr_stats->reference_band_number = reference_band_number;
                curr_stats->search_band_number = search_band_number;

                /* Calculate the band code number.  This is a composite
                   integer value based on the reference and search
                   band numbers.  For example, the band pair (3, 5) would
                   have a corresponding band code number of 305.   This
                   code is currently used just in the Cal/Val analysis
                   report.  */
                curr_stats->band_code_number = reference_band_number * 100
                    + search_band_number;

                /* Calculate the unweighted average line and sample RMSE in
                   pixels for the current (reference band, search band)
                   pair. */
                curr_stats->unweighted_line_rmse_pixels =
                    sqrt((rmse_sum_of_squares[rmse_array_index]
                         .unweighted_line_sumsq)
                    / (rmse_sum_of_squares[rmse_array_index]
                          .number_of_database_records));

                curr_stats->unweighted_sample_rmse_pixels =
                    sqrt((rmse_sum_of_squares[rmse_array_index]
                          .unweighted_sample_sumsq)
                    / (rmse_sum_of_squares[rmse_array_index]
                          .number_of_database_records));

                /* Calculate the weighted average line and sample RMSE in
                   pixels for the current band pair. */
                curr_stats->weighted_line_rmse_pixels =
                    sqrt((rmse_sum_of_squares[rmse_array_index]
                         .weighted_line_sumsq)
                    / (rmse_sum_of_squares[rmse_array_index]
                          .number_of_valid_correlation_points));

                curr_stats->weighted_sample_rmse_pixels =
                    sqrt((rmse_sum_of_squares[rmse_array_index]
                          .weighted_sample_sumsq)
                    / (rmse_sum_of_squares[rmse_array_index]
                          .number_of_valid_correlation_points));

                /* Calculate the difference between the weighted and
                   unweighted average line/sample RMSEs. */
                curr_stats->delta_line_rmse_pixels =
                    curr_stats->weighted_line_rmse_pixels
                    - curr_stats->unweighted_line_rmse_pixels;

                curr_stats->delta_sample_rmse_pixels =
                    curr_stats->weighted_sample_rmse_pixels
                    - curr_stats->unweighted_sample_rmse_pixels;

                /* Calculate the corresponding unweighted average line
                   and sample RMSE in meters. */
                curr_stats->unweighted_line_rmse_meters =
                    report_data->ground_sample_distance_meters
                    * curr_stats->unweighted_line_rmse_pixels;

                curr_stats->unweighted_sample_rmse_meters =
                    report_data->ground_sample_distance_meters
                    * curr_stats->unweighted_sample_rmse_pixels;

                /* Calculate the line and sample LE90 RMSEs in meters. */
                curr_stats->line_le90_rmse_meters =
                    SCALE_LE90 * curr_stats->unweighted_line_rmse_meters;

                curr_stats->sample_le90_rmse_meters =
                    SCALE_LE90 * curr_stats->unweighted_sample_rmse_meters;

                /* Keep track of the maximum line and sample RMSE 90 values.
                   Only if the current search band is not the Cirrus band. */
                if (search_band_number != cirrus_band_number)
                {
                    if (curr_stats->line_le90_rmse_meters
                        > report_data->max_line_le90_rmse_meters)
                    {
                        report_data->max_line_le90_rmse_meters =
                            curr_stats->line_le90_rmse_meters;
                    }

                    if (curr_stats->sample_le90_rmse_meters
                        > report_data->max_sample_le90_rmse_meters)
                    {
                        report_data->max_sample_le90_rmse_meters =
                            curr_stats->sample_le90_rmse_meters;
                    }
                }

                /* Increment the band pair index. */
                band_pair_index++;
            }
        }
    }

    return SUCCESS;
}
