/**************************************************************************
 NAME:         kpr7_evaluate

 PURPOSE:      Performs a band registration accuracy assessment to determine
               whether the KPR 7 specification has been met.  Also
               determines the amount of penalty/bonus for not meeting the
               specification or exceeding the specification, respectively,
               and writes the assessment results to i) an official report
               file; and ii) an analysis report file used by Cal/Val.

 RETURNS:      Integer status code of EXIT_SUCCESS (all processing
               completed successfully) or EXIT_FAILURE (error condition
               detected).
***************************************************************************/

/* Standard C includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>                            /* for isdigit() */
#include <math.h>
#include <float.h>

/* IAS system includes */
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_satellite_attributes.h"

/* General KPR include  */
#include "kpr_shared.h"

/* KPR7-specific include */
#include "kpr7.h"


/* #define for string buffer length */
#define BUFFER_STRLEN 2048

/* #define for the current number of items in a B2B post-processing report
   data line. */
#define NUMBER_OF_REPORT_LINE_ITEMS  26


/****************************************************************************
 NAME:     convert_csv_string_to_space_string

 PURPOSE:  Takes a csv-delimited string and converts it to a space-delimited
           string

 RETURNS:  Nothing
*****************************************************************************/
static void convert_csv_string_to_space_string
(
    const char *csv_buffer,        /* I: CSV string to parse */
    char *space_delimited_buffer   /* O: String where commas are replaced
                                      by spaces, and consecutive commas
                                      are replaced by '0 ' */
)
{
    int csv_length = strlen(csv_buffer);
    int i;                         /* loop counter */
    int j = 0;                     /* index into output buffer */


    for (i = 0; i < csv_length; i++)
    {
        /* If this character isn't a comma, copy it directly to the
           output buffer. */
        if (csv_buffer[i] != ',')
        {
            space_delimited_buffer[j++] = csv_buffer[i];
        }
        else
        {
            /* We have two cases to consider here -- either the first
               character is a comma (signifying that the first field is
               empty), or the previous character was a comma.  Both cases
               require the same action -- put in a '0'.  */
            if ((i == 0) || (csv_buffer[i - 1] == ','))
            {
                /* Add the '0'. */
                space_delimited_buffer[j++] = '0';
            }

            /* Now add the space to deal with the present comma. */
            space_delimited_buffer[j++] = ' ';
        }
    }

    /* Change the newline character to a NULL terminator. */
    space_delimited_buffer[j] = '\0';

}   /* END -- convert_csv_string_to_space_string */


/****************************************************************************
 NAME:      initialize_report_data

 PURPOSE:   Initializes the KPR 7 report data structure

 RETURNS:   Nothing
*****************************************************************************/
static void initialize_report_data
(
    KPR7_REPORT_DATA *rpt_data
)
{
    /* Set the data structure block to a "clean" state. */
    memset(rpt_data, 0, sizeof(KPR7_REPORT_DATA));

    /* Set the compared max values to as small an amount as possible. */
    rpt_data->max_line_le90_rmse_meters   = -DBL_MAX;
    rpt_data->max_sample_le90_rmse_meters = -DBL_MAX;

}   /* END -- initialize_report_data */



/****************************************************************************
 NAME:     evaluate_award_fee_adjustment

 PURPOSE:  Takes the worst case (largest) LE90 RMSE value (line or sample)
           and determines the penalty or bonus adjustment applied to the
           base award fee for not meeting/meeting/exceeding the specification

 RETURNS:  Nothing
*****************************************************************************/
static void evaluate_award_fee_adjustment
(
    KPR7_REPORT_DATA *rpt_data             /* I/O:  Summary report data */
)
{
    double largest_le90_rmse_meters;


    /* Determine the largest LE90 RMSE value. */
    if (rpt_data->max_line_le90_rmse_meters
        > rpt_data->max_sample_le90_rmse_meters)
    {
        largest_le90_rmse_meters = rpt_data->max_line_le90_rmse_meters;
    }
    else
    {
        largest_le90_rmse_meters = rpt_data->max_sample_le90_rmse_meters;
    }

    /* Round it to the nearest tenth of a meter (1 digit of precision).  */
    rpt_data->max_le90_rmse_meters =
        floor(10.0 * (largest_le90_rmse_meters + 0.05)) / 10.0;
    
    /* Compare the largest LE90 RMSE to the various RMSE thresholds. */
    if (rpt_data->max_le90_rmse_meters >= KPR7_FULL_PENALTY_RMSE_METERS)
    {
        /* No award at all. */
        rpt_data->award_fee_adjustment_percent = -100;
    }
    else if (rpt_data->max_le90_rmse_meters
             >= KPR7_FIFTY_PERCENT_PENALTY_RMSE_METERS)
    {
         /* 50% reduction in award. */
        rpt_data->award_fee_adjustment_percent = -50;
    }
    else if (rpt_data->max_le90_rmse_meters
             >= KPR7_TWENTYFIVE_PERCENT_PENALTY_RMSE_METERS)
    {
        /* 25% reduction in award. */
        rpt_data->award_fee_adjustment_percent = -25;
    }
    else if (rpt_data->max_le90_rmse_meters >= KPR7_NO_PENALTY_RMSE_METERS)
    {
        /* No adjustment to award. */
        rpt_data->award_fee_adjustment_percent = 0;
    }
    else if (rpt_data->max_le90_rmse_meters
        >= KPR7_TWENTYFIVE_PERCENT_BONUS_RMSE_METERS)
    {
        /* 25% bonus for exceeding the requirement. */
        rpt_data->award_fee_adjustment_percent = 25;
    }
    else
    {
        /* Maximum 37% bonus for really exceeding the requirement. */
        rpt_data->award_fee_adjustment_percent = 37;
    }

}   /* END -- evaluate_award_fee_adjustment */



/*==========================================================================
                     MAIN PROGRAM
============================================================================*/
int main(int argc, char *argv[])
{
    /* Data structure to contain summary information to be written to
       reports */
    KPR7_REPORT_DATA report_data;

    /* Pointer to a KPR7_RMSE_SUM_OF_SQUARES data structure to contain
       the unweighted and weighted RMSE sum-of-squares for all of the
       band pairs of interest (i.e. those reported in the B2B post-
       processing report).  */
    KPR7_RMSE_SUM_OF_SQUARES *rmse_sum_of_squares = NULL;

    /* Ground sample distance (in meters) read from the command line used to
       invoke this application. */
    double ground_sample_distance_meters;

    /* Report file names */
    const char *b2b_report_filename = NULL;  /* Report file produced from
                                                B2B post-processing
                                                accuracy assessment tool */
    const char *kpr7_report_filename = NULL;
    const char *calval_report_filename = NULL;

    /* B2B post-processing report file handle */ 
    FILE *b2b_fptr = NULL;

    /* Processing band list. */
    int oli_band_number_list[IAS_MAX_NBANDS];

    /* Miscellaneous processing variables */
    int found_valid_line_data = 0;
    int number_of_oli_bands;
    int status;

    /* Acquisition date range variables */
    char date_range_buffer[BUFFER_STRLEN] = "Unknown";
    int acquisition_date_range_found = FALSE;

    /*----------------------------------------------------------------------
                         INITIALIZATION
    ------------------------------------------------------------------------*/
    /* Initialize logging for this application. */
    ias_log_initialize("kpr7_evaluate");

    /* Check the command line for the correct number of parameters. */
    if (argc != 5)
    {
        printf("USAGE:  kpr7_evaluate <B2B post-processing report filename> "
            "<ground sample distance in meters> "
            "<name of KPR7 summary report> <name of Cal/Val report>\n");
        exit(EXIT_FAILURE);
    }

    /* Log the beginning of the assessment process. */
    IAS_LOG_INFO("KPR 7 assessment initiated");

    /* Initialize the satellite attributes library. */
    status = ias_sat_attr_initialize(IAS_L8);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attributes library");
        exit(EXIT_FAILURE);
    }

    /* Get the OLI band list and sort it so that the pan band is the
       first band number in the list.*/
    status = ias_sat_attr_get_sensor_band_numbers(IAS_OLI, IAS_NORMAL_BAND,
        0, oli_band_number_list, IAS_MAX_NBANDS, &number_of_oli_bands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving list of 'normal' OLI imaging bands");
        exit(EXIT_FAILURE);
    }
    status = kpr_sort_band_number_list(oli_band_number_list,
        number_of_oli_bands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reordering OLI band list");
        exit(EXIT_FAILURE);
    }

    /* Get the report file names from the command line. */
    b2b_report_filename = argv[1];
    kpr7_report_filename = argv[3];
    calval_report_filename = argv[4];

    /* Get the ground sample distance value from the command line. */
    status = sscanf(argv[2], "%lf", &ground_sample_distance_meters);
    if (status != 1)
    {
        IAS_LOG_ERROR("Getting ground sample distance value from "
            "command line arguments");
        exit(EXIT_FAILURE);
    }

    /* Allocate the sum-of-squares data block. Use calloc to pre-initialize
       all of the values to 0.  */
    rmse_sum_of_squares = calloc(number_of_oli_bands * number_of_oli_bands,
        sizeof(KPR7_RMSE_SUM_OF_SQUARES));
    if (rmse_sum_of_squares == NULL)
    {
        IAS_LOG_ERROR("Allocating working sum-of-squares data block");
        exit(EXIT_FAILURE);
    }

    /* Open the band-to-band (B2B) post-processing summary report file. */
    b2b_fptr = fopen(b2b_report_filename, "r");
    if (b2b_fptr == NULL)
    {
        IAS_LOG_ERROR("Opening B2B input file %s", b2b_report_filename);
        free(rmse_sum_of_squares);
        exit(EXIT_FAILURE);
    }

    /* Read and process the B2B report file line-by-line.  */
    while (!feof(b2b_fptr))
    {
        char line_buffer[BUFFER_STRLEN];

        /* Get the current line from the file.  fgets() NULL-terminates
           it after it's read. */
        if (fgets(line_buffer, BUFFER_STRLEN, b2b_fptr) == NULL)
        {
            IAS_LOG_ERROR("Reading current line from B2B report file %s",
                b2b_report_filename);
            fclose(b2b_fptr);
            free(rmse_sum_of_squares);
            exit(EXIT_FAILURE);
        }

        /* Check each line for date range */
        if (acquisition_date_range_found == FALSE &&
            strstr(line_buffer, "Acquisition date range:") != NULL)
        {  
            char *date_range_ptr;

            /* Set date range as found */
            acquisition_date_range_found = TRUE;
    
            /* Search string for date */
            date_range_ptr = strchr(line_buffer, ':') + 1;
            while (*date_range_ptr == ' ') 
                date_range_ptr++;
            strcpy(date_range_buffer, date_range_ptr);              
        }

        /* Data lines in the B2B accuracy assessment report file start with
           a reference band / search band pair, separated by commas.  So
           check the first character in the line -- if it contains a numeric
           digit, we'll try to parse it as a data line. */
        if (isdigit(line_buffer[0]))
        {
            char space_delimited_buffer[BUFFER_STRLEN];
            char dummy[128];              /* string variable used for
                                             "throw-away" data */
            double line_rmse;             /* line RMSE  */
            double line_rmse_sq;          /* squared line RMSE */
            double sample_rmse;           /* sample RMSE */
            double sample_rmse_sq;        /* squared sample RMSE */
            int reference_band_number;    /* reference band number */
            int reference_band_index;
            int search_band_number;       /* search band number */
            int search_band_index;
            int number_of_valid_correlation_points;
                                          /* number of points that
                                             successfully correlated */
            int sumsq_band_pair_index;      /* offset into the sum-of-squares
                                             data structure array based on
                                             (reference_band_number,
                                              search_band_number) pair */

            /* Convert the commas to spaces and add 0's to empty fields in
               the current line buffer.  This should make parsing the line
               for the required data much easier. */
            convert_csv_string_to_space_string(line_buffer,
                space_delimited_buffer);

            /* Parse the line.  We're interested in just the reference band
               number, search band number, number of valid correlation points,
               and the RMSEs (in pixels) for the current line and sample.  */
            status = sscanf(space_delimited_buffer,
                "%d %d %s %s %s %s %s %s %s %s %s %s %d %s %s %s %s %s %s %s "
                "%s %s %lf %lf %s %s", &reference_band_number,
                &search_band_number, dummy, dummy, dummy, dummy, dummy, dummy,
                dummy, dummy, dummy, dummy,
                &number_of_valid_correlation_points, dummy,
                dummy, dummy, dummy, dummy, dummy, dummy, dummy, dummy,
                &line_rmse, &sample_rmse, dummy, dummy);
            if (status != NUMBER_OF_REPORT_LINE_ITEMS)
            {
                IAS_LOG_ERROR("Parsing current line from B2B input file "
                    "%s", b2b_report_filename);
                fclose(b2b_fptr);
                free(rmse_sum_of_squares);
                exit(EXIT_FAILURE);
            }

            /* We could successfully parse the current data line, so it can
               be assumed to contain valid data. */
            found_valid_line_data = 1;

            /* Get the indices for the reference and search band numbers. */
            reference_band_index = ias_sat_attr_convert_band_number_to_index(
                reference_band_number);
            search_band_index = ias_sat_attr_convert_band_number_to_index(
                search_band_number);
            if ((reference_band_index == ERROR)
                || (search_band_index == ERROR))
            {
                IAS_LOG_ERROR("Invalid reference and/or search band numbers "
                    "(%d, %d)", reference_band_number, search_band_number);
                fclose(b2b_fptr);
                free(rmse_sum_of_squares);
                exit(EXIT_FAILURE);
            }

            /* Determine the offset into the sum-of-squares data block
               based on the reference and search band indices. */
            sumsq_band_pair_index = reference_band_index * number_of_oli_bands
                + search_band_index;

            /* Calculate the unweighted line and sample RMSE sum-of-squares. */
            line_rmse_sq = line_rmse * line_rmse;
            sample_rmse_sq = sample_rmse * sample_rmse;
            rmse_sum_of_squares[sumsq_band_pair_index].unweighted_line_sumsq
                += line_rmse_sq;
            rmse_sum_of_squares[sumsq_band_pair_index].unweighted_sample_sumsq
                += sample_rmse_sq;

            /* Calculate the weighted line and sample RMSE sum-of-squares.
               Weight by the number of valid correlation points for each
               sample value. */
            rmse_sum_of_squares[sumsq_band_pair_index].weighted_line_sumsq
                +=
                ((double)number_of_valid_correlation_points * line_rmse_sq);
            rmse_sum_of_squares[sumsq_band_pair_index].weighted_sample_sumsq
                +=
                ((double)number_of_valid_correlation_points * sample_rmse_sq);

            /* Count the current record. */
            rmse_sum_of_squares[sumsq_band_pair_index]
                .number_of_database_records++;

            /* Count the total number of valid correlation points. */
            rmse_sum_of_squares[sumsq_band_pair_index]
                .number_of_valid_correlation_points
                += number_of_valid_correlation_points;
        }
        else
        {
            /* If we're at the header line for the summary information
               after the last data record, it should contain the
               sentinel string "NET_LINE_RMSE".  If we've read this
               line, stop. */
            if (strncmp(line_buffer, "NET_LINE_RMSE", 13) == 0)
            {
                break;
            }
        }
    }

    /* Close the report file -- we don't need it anymore. */
    if (fclose(b2b_fptr))
    {
        IAS_LOG_ERROR("Closing B2B report file %s", b2b_report_filename);
        free(rmse_sum_of_squares);
        exit(EXIT_FAILURE);
    }

    /* It's possible that the B2B report file would be completely read through
       but no valid data would be found.  If that's the case, that's an
       error. */
    if (!found_valid_line_data)
    {
        IAS_LOG_ERROR("B2B report file %s contained no valid data useable "
            "for KPR 7 assessment", b2b_report_filename);
        free(rmse_sum_of_squares);
        exit(EXIT_FAILURE);
    }

    /*----------------------------------------------------------------------
                      KPR 7 ASSESSMENT ANALYSIS
    ------------------------------------------------------------------------*/
    
    /* Initialize the report data structure. */
    initialize_report_data(&report_data);
    report_data.ground_sample_distance_meters = ground_sample_distance_meters;
    report_data.acquisition_date_range = date_range_buffer;  
    
    /* Check if date range is present */
    if (!acquisition_date_range_found)
        IAS_LOG_WARNING("Acquisition date not found in B2B report file");
    
    /* Calculate the required summary statistics for the report file. */
    status = kpr7_calculate_summary_rmse_statistics(oli_band_number_list,
        number_of_oli_bands, rmse_sum_of_squares, &report_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating required KPR 7 summary statistics "
            "information");
        free(rmse_sum_of_squares);
        exit(EXIT_FAILURE);
    }

    /* Free the RMSE sum-of-squares data block -- we don't need it anymore. */
    free(rmse_sum_of_squares);
    rmse_sum_of_squares = NULL;

    /* Determine the award fee adjustment based on the results of the
       analysis. */
    evaluate_award_fee_adjustment(&report_data);

    /*--------------------------------------------------------------------
                    WRITE KPR 7 ASSESSMENT REPORTS
    ----------------------------------------------------------------------*/
    IAS_LOG_INFO("KPR 7 assessment analysis completed.  Writing results to "
        "assessment reports %s and %s", kpr7_report_filename,
        calval_report_filename);
 
    /* Write the KPR 7 summary report. */
    status = kpr7_write_official_report(kpr7_report_filename, &report_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing official KPR 7 summary report");
        free(report_data.bp_stats);
        exit(EXIT_FAILURE);
    }

    /* Write the Cal/Val analysis summary report. */
    status = kpr7_write_calval_analysis_report(calval_report_filename,
        &report_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing KPR 7 calval analysis summary report");
        free(report_data.bp_stats);
        exit(EXIT_FAILURE);
    }

    /*---------------------------------------------------------------------
                               CLEANUP
    -----------------------------------------------------------------------*/
    /* Free the allocated report data structure buffer. */
    free(report_data.bp_stats);
    report_data.bp_stats = NULL;

    IAS_LOG_INFO("KPR 7 assessment completed successfully.");
    exit(EXIT_SUCCESS);
}   /* END -- kpr7_evaluate MAIN */
