/****************************************************************************
 NAME:        kpr7_write_reports

 PURPOSE:     Module implementing routines to write the official KPR 7
              assessment report and an associated analysis report for
              Cal/Val use.

 RETURNS:     Integer status code of SUCCESS or ERROR
*****************************************************************************/
#include <stdio.h>

#include "ias_const.h"
#include "ias_logging.h"
#include "ias_satellite_attributes.h"

#include "kpr7.h"


/***************************************************************************
 NAME:     kpr7_write_official_report

 PURPOSE:  Writes KPR 7 summary statistics information to the official
           KPR 7 assessment report

 RETURNS:  Integer status code of SUCCESS or ERROR
****************************************************************************/
int kpr7_write_official_report
(
    const char *report_filename,          /* I: Name of official report file */
    const KPR7_REPORT_DATA *report_data   /* I: KPR7 summary report data */
)
{
    FILE *rpt_fptr;
    int  i;                           /* loop counter */
    int  status;


    /* Get the cirrus band number.  The results for this band are not
       included in the official assessment, but they are useful to
       Cal/Val.  */
    int cirrus_band_number =
        ias_sat_attr_get_band_number_from_type(IAS_CIRRUS_BAND);
    if (cirrus_band_number == ERROR)
    {
        IAS_LOG_ERROR("Retrieving cirrus band number");
        return ERROR;
    }

    /* Open the file for writing. */
    rpt_fptr = fopen(report_filename, "w");
    if (rpt_fptr == NULL)
    {
        IAS_LOG_ERROR("Opening KPR 7 assessment report file %s for writing",
            report_filename);
        return ERROR;
    }

    /* Write the header information. */
    status = fprintf(rpt_fptr, "\n             KPR #7 Band-to-Band "
        "Registration Accuracy Report\n             "
        "Acquisition date range: %s\n"
        "Reference  Search     RMSE (pixels)        RMSE (meters)      "
        "LE90 (meters)\n"
        "  Band      Band    Line      Sample     Line      Sample   Line"
        "      Sample\n\n", report_data->acquisition_date_range);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing header block to KPR 7 assessment report "
            "file %s", report_filename);
        fclose(rpt_fptr);
        return ERROR;
    }

    /* Write the summary information, ignoring results for the cirrus band. */
    for (i = 0; i < report_data->band_pair_count; i++)
    {
        if ((report_data->bp_stats[i].reference_band_number
             == cirrus_band_number)
            || (report_data->bp_stats[i].search_band_number
             == cirrus_band_number))
        {
            continue;
        }

        /* Write the official KPR 7 metrics information to the report. */
        status = fprintf(rpt_fptr, " %4d  %7d  %10.6f  %9.6f  %7.3f  %8.3f  "
            "%7.3f  %8.3f\n",
            report_data->bp_stats[i].reference_band_number,
            report_data->bp_stats[i].search_band_number,
            report_data->bp_stats[i].unweighted_line_rmse_pixels,
            report_data->bp_stats[i].unweighted_sample_rmse_pixels,
            report_data->bp_stats[i].unweighted_line_rmse_meters,
            report_data->bp_stats[i].unweighted_sample_rmse_meters,
            report_data->bp_stats[i].line_le90_rmse_meters,
            report_data->bp_stats[i].sample_le90_rmse_meters);
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing summary metrics information to KPR 7 "
                "assessment report file %s", report_filename);
            fclose(rpt_fptr);
            return ERROR;
        }
    }

    /* Output the summary statistics to the report.  */
    status = fprintf(rpt_fptr,  "\n    Worst Case Band Pair Line:    %6.3f"
        "\n    Worst Case Band Pair Sample:  %6.3f\n\n",
        report_data->max_line_le90_rmse_meters,
        report_data->max_sample_le90_rmse_meters);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing worst-case 90%% RMSE information to KPR 7 "
            "assessment report file %s", report_filename);
        fclose(rpt_fptr);
        return ERROR;
    }

    /* Write the award fee information to the report.  */
    status = fprintf(rpt_fptr, "The award fee adjustment is %d%% for a "
        "worst case pair average of %.1f\n\n",
        report_data->award_fee_adjustment_percent,
        report_data->max_le90_rmse_meters);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing award fee adjustment information to KPR 7 "
            "assessment report file %s", report_filename);
        fclose(rpt_fptr);
        return ERROR;
    }

    status = fprintf(rpt_fptr, "     Where:\n\n"
        "        %.1f <= worst average        is -100%%\n"
        "        %.1f <= worst average <= %.1f is  -50%%\n"
        "        %.1f <= worst average <= %.1f is  -25%%\n"
        "        %.1f <= worst average <= %.1f is    0%%\n"
        "        %.1f <= worst average <= %.1f is   25%%\n"
        "               worst average <= %.1f is   37%%\n",
        KPR7_FULL_PENALTY_RMSE_METERS,
        KPR7_FIFTY_PERCENT_PENALTY_RMSE_METERS,
        KPR7_FULL_PENALTY_RMSE_METERS - 0.1,
        KPR7_TWENTYFIVE_PERCENT_PENALTY_RMSE_METERS,
        KPR7_FIFTY_PERCENT_PENALTY_RMSE_METERS - 0.1,
        KPR7_NO_PENALTY_RMSE_METERS,
        KPR7_TWENTYFIVE_PERCENT_PENALTY_RMSE_METERS - 0.1,
        KPR7_TWENTYFIVE_PERCENT_BONUS_RMSE_METERS,
        KPR7_NO_PENALTY_RMSE_METERS - 0.1,
        KPR7_THIRTYSEVEN_PERCENT_BONUS_RMSE_METERS);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing fee award adjustment schedule to KPR 7 "
            "assessment report file %s", report_filename);
        fclose(rpt_fptr);
        return ERROR;
    }

    /* Close the report file, checking to ensure it closed successfully. */
    if (fclose(rpt_fptr))
    {
        IAS_LOG_ERROR("Closing KPR 7 assessment report file %s",
            report_filename);
        return ERROR;
    }

    return SUCCESS;
}


/***************************************************************************
 NAME:     kpr7_write_calval_analysis_report

 PURPOSE:  Writes KPR 7 summary statistics information to a report used for
           additional Cal/Val analysis

 RETURNS:  Integer status code of SUCCESS or ERROR
****************************************************************************/
int kpr7_write_calval_analysis_report
(
    const char *report_filename,          /* I: Name of Cal/Val report file */
    const KPR7_REPORT_DATA *report_data   /* I: KPR7 summary report data */
)
{
    FILE *rpt_fptr;
    int  i;                           /* loop counter */
    int  status;


    /* Open the file for writing. */
    rpt_fptr = fopen(report_filename, "w");
    if (rpt_fptr == NULL)
    {
        IAS_LOG_ERROR("Opening Cal/Val analysis report file %s for writing",
            report_filename);
        return ERROR;
    }

    /* Write the header information. */
    status = fprintf(rpt_fptr, "                       "
        "Acquisition date range: %s\n"
        "Ref   Srch   BP     Wtd RMSE          "
        "RMSE Diff           Avg RMSE      RMSE (m)       LE90 (m)\n"
        "Band  Band  Code  Line   Samp      Line       Samp      Line    "
        "Samp   Line  Samp     Line  Samp\n\n", 
        report_data->acquisition_date_range);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing header block to calval report");
        fclose(rpt_fptr);
        return ERROR;
    }

    /* Write the summary information, including results for the cirrus band. */
    for (i = 0; i < report_data->band_pair_count; i++)
    {
        /* Write the KPR 7 metrics and additional analysis information to the
           report. */
        status = fprintf(rpt_fptr, " %d  %4d  %6d %6.3f %6.3f %10.6f %10.6f "
            "%7.3f %7.3f %6.2f %5.2f %8.2f %5.2f\n",
            report_data->bp_stats[i].reference_band_number,
            report_data->bp_stats[i].search_band_number,
            report_data->bp_stats[i].band_code_number,
            report_data->bp_stats[i].weighted_line_rmse_pixels,
            report_data->bp_stats[i].weighted_sample_rmse_pixels,
            report_data->bp_stats[i].delta_line_rmse_pixels,
            report_data->bp_stats[i].delta_sample_rmse_pixels,
            report_data->bp_stats[i].unweighted_line_rmse_pixels,
            report_data->bp_stats[i].unweighted_sample_rmse_pixels,
            report_data->bp_stats[i].unweighted_line_rmse_meters,
            report_data->bp_stats[i].unweighted_sample_rmse_meters,
            report_data->bp_stats[i].line_le90_rmse_meters,
            report_data->bp_stats[i].sample_le90_rmse_meters);
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing summary metrics information to Cal/Val "
                "analysis report file %s", report_filename);
            fclose(rpt_fptr);
            return ERROR;
        }
    }

    /* Output the summary statistics to the report.  */
    status = fprintf(rpt_fptr,  "\n    Worst Case Band Pair Line:    %6.3f"
        "\n    Worst Case Band Pair Sample:  %6.3f\n\n",
        report_data->max_line_le90_rmse_meters,
        report_data->max_sample_le90_rmse_meters);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing summary statistics information to "
            "Cal/Val analysis report file %s", report_filename);
        fclose(rpt_fptr);
        return ERROR;
    }

    /* Close the report file, checking to ensure it closed successfully. */
    if (fclose(rpt_fptr))
    {
        IAS_LOG_ERROR("Closing Cal/Val analysis report file %s",
            report_filename);
        return ERROR;
    }

    return SUCCESS;
}
