#ifndef KPR_SHARED_H
#define KPR_SHARED_H

#include "ias_cpf.h"


/* Defines used to specify whether crosstrack or alongtrack view angles
   should be calculated.  Used by the KPR 5/6 assessment tools. */
#define KPR_CROSSTRACK_ANGLES      1
#define KPR_ALONGTRACK_ANGLES      2

/* Definition of data structure to contain left edge and right edge
   inoperable detector counts for an SCA.  Used in the KPR 5/6 assessment
   tools.  */
typedef struct kpr_inoperable_edge_detector_counts
{
    int left_edge_count;
    int right_edge_count;
} KPR_INOPERABLE_EDGE_DETECTOR_COUNTS;


/* Shared KPR function prototypes */

/* NOTE:  The pointers *xxx_view_angles_sca_band_radians
   represent dynamically allocated 1D arrays with (number_of_oli_scas *
   number_of_oli_bands) elements.  The indexing into these arrays is
   calculated as

   array_index = sca_index * number_of_bands + band_list_index

   where band_list_index = 0 to number_of_bands - 1.  The band_list_index
   values correspond to the OLI band number list which is sorted as follows:

   band_list[number_of_bands] = {8, 1, 2, 3, 4, 5, 6, 7, 9}.

   Consequently, the data in these arrays are ordered as follows:

     [0] = SCA 01 band 8,  ...  [7] = SCA 01 band 7,    [8] = SCA 01 band 9,
     [9] = SCA 02 band 8,  ... [16] = SCA 02 band 7,   [17] = SCA 02 band 9,
    [18] = SCA 03 band 8,  ... [25] = SCA 03 band 7,   [26] = SCA 03 band 9,
    [27] = SCA 04 band 8,  ... [34[ = SCA 04 band 7,   [35] = SCA 04 band 9,
     ...                   ...                     ...
   [117] = SCA 14 band 8,  ...[124] = SCA 14 band 7,  [125] = SCA 14 band 9

   To emphasize the point...  THIS DOES NOT FOLLOW THE CONVENTIONAL
   BAND / SCA ORDERING USED IN OTHER IAS PROCESSING.  */

int kpr_calculate_view_angles
(
    const struct IAS_CPF_FOCAL_PLANE *focal_plane,
                                         /* I: CPF focal plane parameters */
    const int *band_number_list,         /* I: 1D array containing sorted
                                            list of bands to process */
    int number_of_bands,                 /* I: Number of bands to process */
    int number_of_scas,                  /* I: Number of SCAs in each band */
    int angle_select_flag,               /* I: Selects which set of angles
                                            to calculate
                                            (KPR_CROSSTRACK_ANGLES,
                                             KPR_ALONGTRACK ANGLES) */
    double *max_view_angles_sca_band_radians,
                                         /* O: 1D array of unadjusted maximum
                                            view angles in radians */
    double *min_view_angles_sca_band_radians,
                                         /* O: 1D array of unadjusted minimum
                                            view angles in radians */
   double *ifovs_sca_band_radians,      /* O: 1D array of instantaneous IFOV
                                            angles in radians; pass in NULL
                                            for this parameter if
                                            calculating alongtrack angles */
    double *max_view_angle_radians,      /* O: Largest view angle in radians
                                            across all SCAs and bands; pass in
                                            NULL for this parameter if the
                                            value is not needed */
    double *min_view_angle_radians,      /* O: Smallest view angle in radians
                                            across all SCAs and bands; pass
                                            NULL for this parameter if the
                                            value is not needed */
    double *max_ifov_radians             /* O: Largest IFOV in radians across
                                            all SCAs and bands; pass in NULL
                                            for this parameter if calculating
                                            alongtrack angles */
);


/* The *edge_counts_sca_band parameter is a dynamically allocated
   1D array with (number_of_scas * number_of_oli_bands) elements.
   It is populated in the same nonstandard order as specified for the
   *xxx_view_angles_sca_band_radians parameters described above. */
int kpr_get_inoperable_edge_detector_counts
(
    const int *band_number_list,         /* I: 1D array containing sorted
                                            list of bands to process */
    int number_of_bands,                 /* I: Number of bands to process */
    int number_of_scas,                  /* I: Number of SCAs per band */
    const struct IAS_CPF_DETECTOR_STATUS *detector_status,
                                         /* I: CPF detector status */
    const struct IAS_CPF_FOCAL_PLANE *focal_plane,
                                         /* I: Focal plane parameters */
    KPR_INOPERABLE_EDGE_DETECTOR_COUNTS *edge_counts_sca_band
                                         /* O: 1D array of structures
                                            containing the final counts of
                                            inoperable edge detctors for
                                            each SCA and band */
);


int kpr_sort_band_number_list
(
    int *band_number_list,      /* I/O: 1D array containing band list to be
                                   sorted.  Sorting occurs in-place. */
    int number_of_bands         /* I: Number of bands in list */
);


#endif
