/*************************************************************************
                           kpr_shared_funcs.c

 PURPOSE:  Module implementing routines that can be used by all
           KPR analysis tools.
**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include "ias_cpf.h"
#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "kpr_shared.h"


/**************************************************************************
 NAME:      qsortcomp

 PURPOSE:   Comparison function used in qsort

 RETURNS:   Difference between input values
***************************************************************************/
static int qsortcomp
(
    const void *p1,             /* I: First number to compare */
    const void *p2              /* I: Second number to compare */
)
{
    /* Cast the pointers to the input values to the correct type. */
    const int *pa = (const int *)p1;
    const int *pb = (const int *)p2;

    /* Return the difference of the two values. */
    return *pa - *pb;
}



/**************************************************************************
 NAME:      kpr_calculate_view_angles

 PURPOSE:   Calculates alongtrack or crosstrack view angles for all SCAs
            and bands.  Which set is calculated is determined by the
            'angle_select_flag' parameter.

 RETURNS:   Integer status code of SUCCESS or ERROR

 NOTES:     The pointers *xxx_view_angles_sca_band_radians and
            *ifovs_sca_band_radians represent dynamically allocated 1D arrays
            that DO NOT follow the standard band/SCA ordering that other
            IAS processing uses.  Refer to the 'NOTE' section in
            kpr_shared.h for additional information.
****************************************************************************/
int kpr_calculate_view_angles
(
    const struct IAS_CPF_FOCAL_PLANE *focal_plane,
                                         /* I: CPF focal plane parameters */
    const int *band_number_list,         /* I: 1D array containing sorted
                                            list of bands to process */
    int number_of_bands,                 /* I: Number of bands to process */
    int number_of_scas,                  /* I: Number of SCAs in each band */
    int angle_select_flag,               /* I: Selects which set of angles
                                            to calculate
                                            (KPR_CROSSTRACK_ANGLES,
                                             KPR_ALONGTRACK ANGLES) */
    double *max_view_angles_sca_band_radians,
                                         /* O: 1D array of unadjusted maximum
                                            view angles in radians */
    double *min_view_angles_sca_band_radians,
                                         /* O: 1D array of unadjusted minimum
                                            view angles in radians */
    double *ifovs_sca_band_radians,      /* O: 1D array of instantaneous IFOV
                                            angles in radians; pass in NULL
                                            for this parameter if
                                            calculating alongtrack angles */
    double *max_view_angle_radians,      /* O: Largest view angle in radians
                                            across all SCAs and bands; pass in
                                            NULL for this parameter if the
                                            value is not needed */
    double *min_view_angle_radians,      /* O: Smallest view angle in radians
                                            across all SCAs and bands; pass
                                            NULL for this parameter if the
                                            value is not needed */
    double *max_ifov_radians             /* O: Largest IFOV in radians across
                                            all SCAs and bands; pass in NULL
                                            for this parameter if calculating
                                            alongtrack angles */
)
{
    int sca_index;


    /* Sanity check -- make sure we've specified a "valid" set of
       view angles to calculate. */
    if (!((angle_select_flag == KPR_CROSSTRACK_ANGLES)
        || (angle_select_flag == KPR_ALONGTRACK_ANGLES)))
    {
        IAS_LOG_ERROR("Invalid KPR view angle selection parameter %d",
            angle_select_flag);
        return ERROR;
    }

    /* Sanity check -- make sure we've allocated data buffers for the
       IFOV data if we're doing crosstrack view angle calculations. */
    if (angle_select_flag == KPR_CROSSTRACK_ANGLES)
    {
        if ((ifovs_sca_band_radians == NULL) || (max_ifov_radians == NULL))
        {
            IAS_LOG_ERROR("Crosstrack angle calculation requested, but "
                "no data buffers provided for crosstrack IFOV results");
            return ERROR;
        }
    }

    /* The arrays for the min/max view angles are in increasing SCA order,
       with the band order in each SCA specified by the input band list. */
    for (sca_index = 0; sca_index < number_of_scas; sca_index++)
    {
        /* Starting offset for first band in each SCA. */
        int base_offset = sca_index * number_of_bands;

        int band_list_index;                  /* Band list loop counter */

        for (band_list_index = 0;
             band_list_index < number_of_bands;
             band_list_index++)
        {
            double coeff_A;      /* constant term in legendre polynomial */
            double coeff_B;      /* linear term in legendre polynomial */
            double coeff_C;      /* quadratic term in legendre polynomial */

            int band_number = band_number_list[band_list_index];

            /* Calculate the buffer offset for the current SCA and band. */
            int buffer_offset = base_offset + band_list_index;

            /* Get the correct band index for the current band. */
            int band_index =
                ias_sat_attr_convert_band_number_to_index(band_number);
            if (band_index == ERROR)
            {
                IAS_LOG_ERROR("Converting band number %d to index value",
                    band_number);
                return ERROR;
            }

            /* Get the number of detectors/SCA for the current band. */
            int number_of_detectors =
                ias_sat_attr_get_detectors_per_sca(band_number);
            if (number_of_detectors == ERROR)
            {
                IAS_LOG_ERROR("Getting number of detectors per SCA for "
                    "band number %d", band_number);
                return ERROR;
            }

            /* Get the correct set of legendre coefficients for the view
               angles we need to calculate. */
            if (angle_select_flag == KPR_CROSSTRACK_ANGLES)
            {
                coeff_A =
                    focal_plane->across_los_legendre[band_index][sca_index][0];
                coeff_B =
                    focal_plane->across_los_legendre[band_index][sca_index][1];
                coeff_C =
                    focal_plane->across_los_legendre[band_index][sca_index][2];
            }
            else
            {
                coeff_A =
                    focal_plane->along_los_legendre[band_index][sca_index][0];
                coeff_B =
                    focal_plane->along_los_legendre[band_index][sca_index][1];
                coeff_C =
                    focal_plane->along_los_legendre[band_index][sca_index][2];
            }

            /* Calculate the view angles. */
            max_view_angles_sca_band_radians[buffer_offset] =
                atan((coeff_A - coeff_B + coeff_C));
            min_view_angles_sca_band_radians[buffer_offset] =
                atan((coeff_A + coeff_B + coeff_C));

            if (angle_select_flag == KPR_CROSSTRACK_ANGLES)
            {
                /* Calculate the crosstrack IFOV. */
                ifovs_sca_band_radians[buffer_offset] =
                    (max_view_angles_sca_band_radians[buffer_offset]
                    - min_view_angles_sca_band_radians[buffer_offset])
                     / (number_of_detectors - 1);

                /* Keep track of the maximum crosstrack IFOV. */
                if (ifovs_sca_band_radians[buffer_offset] > *max_ifov_radians)
                {
                    *max_ifov_radians = ifovs_sca_band_radians[buffer_offset];
                }
            }

            /* Keep track of the maximum and/or minimum view angles if
               needed.  This processing assumes either that
                i.   both are needed; or
               ii.   neither are needed  */
            if (max_view_angle_radians != NULL)
            {
                if (max_view_angles_sca_band_radians[buffer_offset]
                    > *max_view_angle_radians)
                {
                    *max_view_angle_radians =
                        max_view_angles_sca_band_radians[buffer_offset];
                }
            }
            if (min_view_angle_radians != NULL)
            {
                if (min_view_angles_sca_band_radians[buffer_offset]
                    < *min_view_angle_radians)
                {
                    *min_view_angle_radians =
                    min_view_angles_sca_band_radians[buffer_offset];
                }
            }
        }   /* END -- band loop */
    }       /* END -- SCA loop */

    return SUCCESS;
}   /* END -- kpr_calculate_view_angles */



/**************************************************************************
 NAME:      kpr_get_inoperable_edge_detector_counts

 PURPOSE:   Counts the number of inoperable detectors at both edges
            of each SCA.

 RETURNS:   Integer status code of SUCCESS or ERROR

 NOTES:     The pointer *edge_counts_sca_band represents a dynamically
            allocated 1D array of KPR_INOPERABLE_EDGE_DETECTOR_COUNTS
            structures that DOES NOT follow the standard band/SCA ordering
            that other IAS processing uses.  Refer to the NOTE sections
            in kpr5.h, kpr6.h and kpr_shared.h for additional information.
***************************************************************************/
int kpr_get_inoperable_edge_detector_counts
(
    const int *band_number_list,         /* I: 1D array containing sorted
                                            list of bands to process */
    int number_of_bands,                 /* I: Number of bands to process */
    int number_of_scas,                  /* I: Number of SCAs per band */
    const struct IAS_CPF_DETECTOR_STATUS *detector_status,
                                         /* I: CPF detector status */
    const struct IAS_CPF_FOCAL_PLANE *focal_plane,
                                         /* I: Focal plane parameters */
    KPR_INOPERABLE_EDGE_DETECTOR_COUNTS *edge_counts_sca_band
                                         /* O: 1D array of structures
                                            containing the final counts of
                                            inoperable edge detctors for
                                            each SCA and band */
)
{
    int sca_index;


    for (sca_index = 0; sca_index < number_of_scas; sca_index++)
    {
        /* Start location in data buffers for this SCA's data. */
        int base_offset = (sca_index * number_of_bands);

        int band_list_index;        /* Band loop counter */

        for (band_list_index = 0;
             band_list_index < number_of_bands;
             band_list_index++)
        {
            int band_number = band_number_list[band_list_index];
            int band_index;
            int number_of_detectors;
            int current_inoperable_detector_number;
            int local_inoperable_detector_count;
            int detector_delta;
            int previous_inoperable_detector_number = 0;
            int search_detector_index;

            /* Determine the buffer offset for this SCA and band. */
            int buffer_offset = base_offset + band_list_index;

            /* Get the correct band index for the current band. */
            band_index =
                ias_sat_attr_convert_band_number_to_index(band_number);
            if (band_index == ERROR)
            {
                IAS_LOG_ERROR("Converting band number %d to index value",
                    band_number);
                return ERROR;
            }

            /* Get the number of detectors/SCA for the current band. */
            number_of_detectors =
                ias_sat_attr_get_detectors_per_sca(band_number);
            if (number_of_detectors == ERROR)
            {
                IAS_LOG_ERROR("Getting number of detectors per SCA for "
                    "band number %d", band_number);
                return ERROR;
            }

            /* Get the count of inoperable detectors in the current
               SCA.  If there aren't any, go to the next SCA. */
            local_inoperable_detector_count =
                detector_status->inoperable_count[band_index][sca_index];
            if (local_inoperable_detector_count == 0)
            {
                continue;
            }

            /* Search the list to get the detector(s) affecting the left
               edge, starting with detector 1.  If we find a gap between
               inoperable detectors,  quit searching. */
            for (search_detector_index = 0;
                 search_detector_index < local_inoperable_detector_count;
                 search_detector_index++)
            {
                current_inoperable_detector_number =
                    detector_status->inoperable[band_index][sca_index]
                                               [search_detector_index];
                detector_delta = current_inoperable_detector_number
                    - previous_inoperable_detector_number;
                if (detector_delta == 1)
                {
                    edge_counts_sca_band[buffer_offset].left_edge_count++;
                    previous_inoperable_detector_number =
                        current_inoperable_detector_number;
                }
                else
                {
                    break;
                }
            }

            /* Search the list to get the detector(s) affecting the right
               edge.  If we find a gap between inoperable detectors,
               quit searching.  */
            previous_inoperable_detector_number = number_of_detectors + 1;
            for (search_detector_index = local_inoperable_detector_count - 1;
                 search_detector_index >= 0;
                 search_detector_index--)
            {
                current_inoperable_detector_number =
                    detector_status->inoperable[band_index][sca_index]
                                               [search_detector_index];
                detector_delta = previous_inoperable_detector_number
                    - current_inoperable_detector_number;
                if (detector_delta == 1)
                {
                    edge_counts_sca_band[buffer_offset].right_edge_count++;
                    previous_inoperable_detector_number =
                        current_inoperable_detector_number;
                }
                else
                {
                    break;
                }
            }
        }   /* END -- band loop */
    }       /* END -- SCA loop */

    return SUCCESS;

}   /* END -- kpr_get_inoperable_edge_detector_counts */



/***************************************************************************
 NAME:          kpr_sort_band_number_list

 PURPOSE:       Takes a band list and sorts it as follows:
 
                 i.  If the pan band is found in the band list, it is placed
                     first, and the other bands are sorted in ascending order.
                ii.  If the pan band is not found in the band list, the
                     entire band list is sorted in ascending order.

                The sorting is done in-place using the C library
                qsort() function.

 RETURNS:       Integer status code of SUCCESS or ERROR
****************************************************************************/
int kpr_sort_band_number_list
(
    int *band_number_list,      /* I/O: 1D array containing band list to be
                                   sorted */
    int number_of_bands         /* I: Number of bands in list */
)
{
    int band_list_index;        /* Band list loop counter */
    int pan_band_number;        /* Pan band number */
    int pan_band_index;         /* Location of pan band number in band
                                   list */
    int found_pan_band = 0;     /* Flag indicating whether band list
                                   contains the pan band (0=no, 1=yes) */


    /* Get the PAN band number. */
    pan_band_number = ias_sat_attr_get_band_number_from_type(IAS_PAN_BAND);
    if (pan_band_number == ERROR)
    {
        IAS_LOG_ERROR("Retrieving PAN band number");
        return ERROR;
    }

    /* Check for the pan band in the original list.  If it's found,
       record its location in the list.  */
    for (band_list_index = 0;
         band_list_index < number_of_bands;
         band_list_index++)
    {
        if (band_number_list[band_list_index] == pan_band_number)
        {
            pan_band_index = band_list_index;
            found_pan_band = 1;
            break;
        }
    }

    /* If we have found the pan band, put it at the beginning of the list and
       sort the remaining bands.  Otherwise, just sort the entire band list. */
    if (found_pan_band)
    {
       /* Swap the first band number in the list and the pan band number. */
       int temp = band_number_list[pan_band_index];
       band_number_list[pan_band_index] = band_number_list[0];
       band_number_list[0] = temp;

       /* Sort the remaining band list. */
       qsort(&band_number_list[1], number_of_bands - 1, sizeof(int),
           qsortcomp);
    }
    else
    {
        /* Sort the entire band list. */
        qsort(band_number_list, number_of_bands, sizeof(int), qsortcomp);
    }

    return SUCCESS;
}   /* END -- kpr_sort_band_number_list */
