/*************************************************************************
 NAME:       test_kpr_tools

 PURPOSE:    Implements C green unit tests for the KPR 5/6/7 evaluation
             tools.
**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include "cgreen.h"
#include "kpr5.h"
#include "kpr6.h"
#include "kpr7.h"


/* Common variable declarations for all tests. */
static char cpf_name[] = "L8CPF20030401_20030630.01";
static char KPR_dir[] = "ias/gps/kpr_evaluations";
static char *test_data_path = NULL;


/* KPR 5 tool unit test. */
Ensure test_kpr5_evaluate()
{
    char cpf_filename[PATH_MAX];
    char expected_KPR5_report_filename[PATH_MAX];
    char syscommand[PATH_MAX];
    char KPR5_report_name[] = "KPR_5_Report.txt";
    int status;


    /* Build the CPF file name. */
    status = snprintf(cpf_filename, sizeof(cpf_filename), "%s/%s/%s",
        test_data_path, KPR_dir, cpf_name);
    if ((status < 0) || (status >= sizeof(cpf_filename)))
    {
        assert_true_with_message(0, "Error constructing full CPF file name");
        return;
    }

    /* Build the name of the expected report file. */
    status = snprintf(expected_KPR5_report_filename,
        sizeof(expected_KPR5_report_filename), "%s/%s/%s",
        test_data_path, KPR_dir, KPR5_report_name);
    if ((status < 0) ||(status >= sizeof(expected_KPR5_report_filename)))
    {
        assert_true_with_message(0, "Error constructing expected KPR 5 "
            "summary report file name");
        return;
    }

    /* Build the command line required to run the tool. */
    status = snprintf(syscommand, sizeof(syscommand), "../kpr5_evaluate %s %s",
        cpf_filename, KPR5_report_name);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(0, "Error constructing command to run "
            "KPR 5 tool");
        return;
    }

    /* Run the command. */
    status = system(syscommand);
    assert_equal_with_message(status, 0, "Error running KPR 5 evaluation "
        "tool");

    /* Construct the diff command. */
    status = snprintf(syscommand, sizeof(syscommand), "diff --brief %s %s",
        KPR5_report_name, expected_KPR5_report_filename);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(0, "Error constructing 'diff' command to "
            "check output report");
        return;
    }
    status = system(syscommand);
    assert_equal_with_message(status, 0, "Differences found between test "
        "report %s and 'reference' report %s", KPR5_report_name,
        expected_KPR5_report_filename);
}   /* END -- test_kpr5_evaluate */



/* KPR 6 tool unit test. */
Ensure test_kpr6_evaluate()
{
    char cpf_filename[PATH_MAX];
    char expected_KPR6_report_filename[PATH_MAX];
    char syscommand[PATH_MAX];
    char KPR6_report_name[] = "KPR_6_Report.txt";
    int status;


    /* Build the CPF file name. */
    status = snprintf(cpf_filename, sizeof(cpf_filename), "%s/%s/%s",
        test_data_path, KPR_dir, cpf_name);
    if ((status < 0) || (status >= sizeof(cpf_filename)))
    {
        assert_true_with_message(0, "Error constructing full CPF file name");
        return;
    }

    /* Build the name of the expected report file. */
    status = snprintf(expected_KPR6_report_filename,
        sizeof(expected_KPR6_report_filename), "%s/%s/%s",
        test_data_path, KPR_dir, KPR6_report_name);
    if ((status < 0) ||(status >= sizeof(expected_KPR6_report_filename)))
    {
        assert_true_with_message(0, "Error constructing expected KPR 6 "
            "summary report file name");
        return;
    }

    /* Build the command line required to run the tool. */
    status = snprintf(syscommand, sizeof(syscommand),
        "../kpr6_evaluate %s %s", cpf_filename, KPR6_report_name);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(0, "Error constructing command to run "
            "KPR 6 tool");
        return;
    }

    /* Run the command. */
    status = system(syscommand);
    assert_equal_with_message(status, 0, "Error running KPR 6 evaluation "
        "tool");

    /* Construct the diff command. */
    status = snprintf(syscommand, sizeof(syscommand), "diff --brief %s %s",
        KPR6_report_name, expected_KPR6_report_filename);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(0, "Error constructing 'diff' command to "
            "check output report");
        return;
    }
    status = system(syscommand);
    assert_equal_with_message(status, 0, "Differences found between test "
        "report %s and 'reference' report %s", KPR6_report_name,
        expected_KPR6_report_filename);
}   /* END -- test_kpr6_evaluate */



/* KPR 7 tool unit test. */
Ensure test_kpr7_evaluate()
{
    char expected_KPR7_report_filename[PATH_MAX];
    char expected_calval_report_filename[PATH_MAX];
    char b2b_postproc_report_filename[PATH_MAX];
    char syscommand[PATH_MAX];
    char KPR7_report_name[] = "KPR_7_Report.txt";
    char calval_report_name[] = "CalVal_Report.txt";

    /* Name of report file generated from B2B post-processing accuracy
       assessment tool. */
    char b2b_postproc_report_name[] = "b2b_postproc_assessment.txt";
    double gsd = 30.0;
    int status;


    /* Build the name of the expected report file. */
    status = snprintf(b2b_postproc_report_filename,
        sizeof(b2b_postproc_report_filename), "%s/%s/%s",
        test_data_path, KPR_dir, b2b_postproc_report_name);
    if ((status < 0) ||(status >= sizeof(b2b_postproc_report_filename)))
    {
        assert_true_with_message(0, "Error constructing expected KPR 7 "
            "summary report file name");
        return;
    }

    /* Build the name of the expected report file. */
    status = snprintf(expected_KPR7_report_filename,
        sizeof(expected_KPR7_report_filename), "%s/%s/%s",
        test_data_path, KPR_dir, KPR7_report_name);
    if ((status < 0) ||(status >= sizeof(expected_KPR7_report_filename)))
    {
        assert_true_with_message(0, "Error constructing expected KPR 7 "
            "summary report file name");
        return;
    }

    /* Build the name of the expected CalVal analysis report file. */
    status = snprintf(expected_calval_report_filename,
        sizeof(expected_calval_report_filename), "%s/%s/%s",
        test_data_path, KPR_dir, calval_report_name);
    if ((status < 0) ||(status >= sizeof(expected_calval_report_filename)))
    {
        assert_true_with_message(0, "Error constructing expected CalVal "
            "analysis report file name");
        return;
    }

    /* Build the command line required to run the tool. */
    status = snprintf(syscommand, sizeof(syscommand),
        "../kpr7_evaluate %s %lf %s %s", b2b_postproc_report_filename, gsd,
        KPR7_report_name, calval_report_name);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(0, "Error constructing command to run "
            "KPR 7 tool");
        return;
    }

    /* Run the command. */
    status = system(syscommand);
    assert_equal_with_message(status, 0, "Error running KPR 7 evaluation "
        "tool");

    /* Construct the diff command to compare the assessment reports. */
    status = snprintf(syscommand, sizeof(syscommand), "diff --brief %s %s",
        KPR7_report_name, expected_KPR7_report_filename);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(0, "Error constructing 'diff' command to "
            "check output report");
        return;
    }
    status = system(syscommand);
    assert_equal_with_message(status, 0, "Differences found between test "
        "report %s and 'reference' report %s", KPR7_report_name,
        expected_KPR7_report_filename);

    /* Construct the diff command to compare the CalVal analysis reports. */
    status = snprintf(syscommand, sizeof(syscommand), "diff --brief %s %s",
        calval_report_name, expected_calval_report_filename);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(0, "Error constructing 'diff' command to "
            "check CalVal analysis report");
        return;
    }
    status = system(syscommand);
    assert_equal_with_message(status, 0, "Differences found between test "
        "analysis report %s and 'reference' analysis report %s",
        calval_report_name, expected_calval_report_filename);
}   /* END -- test_kpr7_evaluate */



int main(int argc, char *argv[])
{
    TestSuite *suite = create_test_suite();

    /* Get the unit test data directory. */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path == NULL)
    {
        printf("Cannot locate test data directory path\n");
        return 1;
    }

    /* Build the test suite. */
    add_test(suite, test_kpr5_evaluate);
    add_test(suite, test_kpr6_evaluate);
    add_test(suite, test_kpr7_evaluate);

    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    else
    {
        return run_test_suite(suite, create_text_reporter());
    }
}   /* END -- run test(s) MAIN */



