/******************************************************************************
NAME:           get_targets

PURPOSE:
Read the MTF target definition file, populating TARGET structures for each
bridge target in the file.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
ntarg    Number of targets successfully read
0    Operation failed

NOTES:
This routine will open the input target definition file, read the parameters
into TARGET structures, close the parameter file, and return the structures.

******************************************************************************/
#include <string.h>
#include "ias_odl.h"
#include "get_targets.h"

/******************************************************************************
NAME:           read_parameters

PURPOSE:
 Read an entry from the parameter file

 RETURN VALUE
     SUCCESS - parameter was read successfully
     FAILURE - parameter read failed
******************************************************************************/
int read_parameters          
(
    const char *parameter_subname, /* I: Unique portion of the name of the 
                                  parameter to read */
    int target_index,       /* I: Index of the target to read */
    int span_index,         /* I: Index of the span to read; -1 if it's not
                                  a span parameter */
    const char *groupname,  /* I: Target file group name */
    IAS_OBJ_DESC *in_parms, /* I: Pointer to the structure containing all the
                                  ODL parameters */
    int number_to_read,     /* I: Number of values to read */
    double *buffer,         /* O: Buffer for reading input */
    int *count              /* O: Number of values returned from GetODLField */
)
{
    int n;                  /* Number returned from snprintf */
    char paramname[100];    /* Target file parameter name */
    int status;             /* Function return status */

    if (span_index != -1)
    {
        n = snprintf(paramname, sizeof(paramname), "Target%d_%s%d", 
            target_index, parameter_subname, span_index);
    }
    else
    {
        n = snprintf(paramname, sizeof(paramname), "Target%d_%s",
            target_index, parameter_subname);
    }
    if ((n < 0) || (n >= sizeof(paramname)))
    {
        IAS_LOG_ERROR("Parameter name size should be >= 0 and <= %d", n);
        return ERROR;
    }
    status = ias_odl_get_field(buffer, number_to_read * sizeof(double), 
        IAS_ODL_Double, in_parms, groupname, paramname, count);

    return status;
}

int get_targets             /* Return the number of targets loaded */
(
    const char *target_file, /* I: Target parameter file name */
    TARGET target_info[]    /* O: Populated target structure array */
)
{
    IAS_OBJ_DESC *in_parms; /* Pointer to the structure containing all the
                               ODL parameters. */
    char *groupname = NULL; /* Target file group name */
    char paramname[100];    /* Target file parameter name */
    int count;              /* Number of values returned from GetODLField */
    int ntarg;              /* Number of targets in file */
    int nband;              /* Number of bands defined for target(s) */
    double buffer[IAS_MAX_NBANDS]; /* Buffer for reading input arrays */
    int i;              /* Counters */
    int j; 
    int k;
    int n;
    int status;
    const int NUM_TARGET_COORD = sizeof(CENTER_LINE)/sizeof(double);

    if ((target_file == NULL) || (target_info == NULL))
    {
        IAS_LOG_ERROR("target_file and target_info cannot be NULL");
        return ERROR;
    }

    /* Open and read the ODL parameter file. */
    in_parms = ias_odl_read_tree(target_file);
    if (in_parms == NULL)
    {
        IAS_LOG_ERROR("Reading parameters from %s", target_file);
        return 0;
    }

    /* Set the groupname. */
    groupname = "MTF_Target";

    /* Get the number of targets in the target file */
    status = ias_odl_get_field(&ntarg, sizeof(int), IAS_ODL_Int, in_parms,
                               groupname, "Number_of_Targets", &count);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading number of targets in the target definition "
                      "file");
        ias_odl_free_tree(in_parms);
        return 0;
    }
    if ((ntarg < 1) || (ntarg > MAX_TARGETS))
    {
        IAS_LOG_ERROR("Invalid number of targets in file, %d, must be "
                      "between 1 and %d", ntarg, MAX_TARGETS);
        ias_odl_free_tree(in_parms);
        return 0;
    }

    /* Initialize bands for the first target */
    for (i = 0; i < IAS_MAX_NBANDS; i++)
    {
        target_info[0].band_list[i] = -1;
    }

    /* Get the list of bands to process */
    status = ias_odl_get_field(target_info[0].band_list,
                               IAS_MAX_NBANDS * sizeof(int), IAS_ODL_Int,
                               in_parms, groupname, "BAND_LIST", &nband);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading band list from the target definition file");
        ias_odl_free_tree(in_parms);
        return 0;
    }
    if ((nband < 1) || (nband > IAS_MAX_NBANDS))
    {
        IAS_LOG_ERROR("Invalid number of bands: %d", nband);
        ias_odl_free_tree(in_parms);
        return 0;
    }
    for (i = 0; i < nband; i++)
    {
        if ((target_info[0].band_list[i] < 1) ||
                (target_info[0].band_list[i] > IAS_MAX_NBANDS))
        {
            IAS_LOG_ERROR("Invalid band number: %d", 
                          target_info[0].band_list[i]);
            ias_odl_free_tree(in_parms);
            return 0;
        }
    }

    /* Read each target */
    for (i = 0; i < ntarg; i++)
    {
        target_info[i].number_of_bands = nband;

        /* Copy the common band list */
        if (i > 0)
            memcpy(target_info[i].band_list, target_info[0].band_list,
                   IAS_MAX_NBANDS * sizeof(int));

        /* Get the target name */
        n = snprintf(paramname, sizeof(paramname), "Target%d_Name", i);
        if ((n < 0) || (n >= sizeof(paramname)))
        {
            IAS_LOG_ERROR("Parameter name length %lu must be %d", 
                sizeof(paramname), n);
            ias_odl_free_tree(in_parms);
            return 0;
        }
        status = ias_odl_get_field(target_info[i].target_name,
                                   sizeof(target_info[i].target_name),
                                   IAS_ODL_String, in_parms, groupname,
                                   paramname, &count);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("In the target name field for target %d", i);
            ias_odl_free_tree(in_parms);
            return 0;
        }

        /* Get the number of bridge spans */
        n = snprintf(paramname, sizeof(paramname),
                     "Target%d_Number_of_Spans", i);
        if ((n < 0) || (n >= sizeof(paramname)))
        {
            IAS_LOG_ERROR("Parameter name length must be %d", n);
            ias_odl_free_tree(in_parms);
            return 0;
        }

        status = ias_odl_get_field(&target_info[i].number_of_spans, sizeof(int),
                                   IAS_ODL_Int, in_parms, groupname,
                                   paramname, &count);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading the number of spans in target %d", i);
            ias_odl_free_tree(in_parms);
            return 0;
        }

        /* Allocate the span structures */
        target_info[i].span = (SPAN *)malloc(target_info[i].number_of_spans
                                        * sizeof(SPAN));
        if (target_info[i].span == NULL)
        {
            IAS_LOG_ERROR("Allocating span structures for target %d", i);
            ias_odl_free_tree(in_parms);
            for (k = 0; k <= i; k++)
            {
                free(target_info[k].span);
                target_info[k].span = NULL;
            }
            return 0;
        }

        /* Read the span parameters */
        for (j = 0; j < target_info[i].number_of_spans; j++)
        {
            /* Get the span width */
            status = read_parameters("Span", i, j, groupname, in_parms, 1,
                buffer, &count);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Reading the bridge span width from the target "
                              "file");
                ias_odl_free_tree(in_parms);
                for (k = 0; k <= i; k++)
                {
                    free(target_info[k].span);
                    target_info[k].span = NULL;
                }
                return 0;
            }
            target_info[i].span[j].width = buffer[0];

            /* Get the span offset */
            status = read_parameters("Offset", i, j, groupname, in_parms, 1,
                buffer, &count);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Reading the bridge span offset from the target "
                              "target file");
                ias_odl_free_tree(in_parms);
                for (k = 0; k <= i; k++)
                {
                    free(target_info[k].span);
                    target_info[k].span = NULL;
                }
                return 0;
            }
            target_info[i].span[j].offset = buffer[0];

            /* Get the span reflectance parameters */
            status = read_parameters("Reflectance", i, j, groupname, in_parms, 
                IAS_MAX_NBANDS, buffer, &count);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Reading the target reflectance values from "
                              "the target file");
                ias_odl_free_tree(in_parms);
                for (k = 0; k <= i; k++)
                {
                    free(target_info[k].span);
                    target_info[k].span = NULL;
                }
                return 0;
            }
            if (count != nband)
            {
                IAS_LOG_ERROR("Number of target reflectances %d must match "
                              "number of bands %d", count, nband);
                ias_odl_free_tree(in_parms);
                for (k = 0; k <= i; k++)
                {
                    free(target_info[k].span);
                    target_info[k].span = NULL;
                }
                return 0;
            }
            for (k = 0; k < nband; k++)
                target_info[i].span[j].reflectance[k] = buffer[k];
        } /* span loop */

        /* Get the background water reflectance parameters */
        status = read_parameters("Water", i, -1, groupname, in_parms, 
            IAS_MAX_NBANDS, buffer, &count);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading the water background values from the "
                          "target file");
            ias_odl_free_tree(in_parms);
            for (k = 0; k <= i; k++)
            {
                free(target_info[k].span);
                target_info[k].span = NULL;
            }
            return 0;
        }
        if (count != nband)
        {
            IAS_LOG_ERROR("Number of water reflectances %d must match number "
                          "of bands %d", count, nband);
            ias_odl_free_tree(in_parms);
            for (k = 0; k <= i; k++)
            {
                free(target_info[k].span);
                target_info[k].span = NULL;
            }
            return 0;
        }
        for (k = 0; k < IAS_MAX_NBANDS; k++)
        {
            if (k < nband)
            {
                target_info[i].water[k] = buffer[k];
            }
            else
            {
                target_info[i].water[k] = -1.0;
            }
        }

        /* Get the background water asymmetry parameters */
        status = read_parameters("Asymmetry", i, -1, groupname, in_parms, 
            IAS_MAX_NBANDS, buffer, &count);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading the water asymmetry values from the target "
                          "file");
            ias_odl_free_tree(in_parms);
            for (k = 0; k <= i; k++)
            {
                free(target_info[k].span);
                target_info[k].span = NULL;
            }
            return 0;
        }
        if (count != nband)
        {
            IAS_LOG_ERROR("Number of water asymmetry values %d must match "
                          "number of bands %d", count, nband);
            ias_odl_free_tree(in_parms);
            for (k = 0; k <= i; k++)
            {
                free(target_info[k].span);
                target_info[k].span = NULL;
            }
            return 0;
        }

        for (k = 0; k < IAS_MAX_NBANDS; k++)
        {
            if (k < nband)
            {
                target_info[i].asymmetry[k] = buffer[k];
            }
            else
            {
                target_info[i].asymmetry[k] = -1.0;
            }
        }

        /* Get the number of polygons. */
        n = snprintf(paramname, sizeof(paramname),
                     "Target%d_Number_Polygons", i);
        if ((n < 0) || (n >= sizeof(paramname)))
        {
            IAS_LOG_ERROR("Parameter name length must be %d", n);
            ias_odl_free_tree(in_parms);
            for (k = 0; k <= i; k++)
            {
                free(target_info[k].span);
                target_info[k].span = NULL;
            }
            return 0;
        }
        status = ias_odl_get_field(&target_info[i].number_of_polygons, 
                                   sizeof(int), IAS_ODL_Int, in_parms, 
                                   "MTF_ROI", paramname, &count);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading the number of polygons in target %d", i);
            ias_odl_free_tree(in_parms);
            for (k = 0; k <= i; k++)
            {
                free(target_info[k].span);
                target_info[k].span = NULL;
            }
            return 0;
        }

        /* Skip polygon coordinates for now */
        target_info[i].polygon = NULL;

        /* Get target height */
        status = read_parameters("Height", i, -1, groupname, in_parms, 1,
            buffer, &count);
        if (status != SUCCESS)
        {
            IAS_LOG_WARNING("Target height value not found, setting to 0.0");
            target_info[i].height = 0.0;
        }
        if (count > 0)
        {
            target_info[i].height = buffer[0];
        }

        /* Get the target start and stop coordinates */
        status = read_parameters("Coordinates", i, -1, groupname, in_parms,
            NUM_TARGET_COORD, buffer, &count);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading the target coordinates from the target "
                          "file");
            ias_odl_free_tree(in_parms);
            for (k = 0; k <= i; k++)
            {
                free(target_info[k].span);
                target_info[k].span = NULL;
            }
            return 0;
        }
        if (count != NUM_TARGET_COORD)
        {
            IAS_LOG_ERROR("Number of target coordinates %d must be %d",
                          count, NUM_TARGET_COORD);
            ias_odl_free_tree(in_parms);
            for (k = 0; k <= i; k++)
            {
                free(target_info[k].span);
                target_info[k].span = NULL;
            }
            return 0;
        }
        target_info[i].center_line.start_point.x = buffer[0];
        target_info[i].center_line.start_point.y = buffer[1];
        target_info[i].center_line.end_point.x = buffer[2];
        target_info[i].center_line.end_point.y = buffer[3];
    } /* target loop */

    /* Done with the ODL file so close it. */
    ias_odl_free_tree(in_parms);

    return ntarg;
}
