/******************************************************************************
NAME:           model.c

PURPOSE:
A package of routines that construct an analytical model of the across- and
along-track system transfer functions, as well as the specified bridge target.

******************************************************************************/
#include <stdlib.h>         /* Defines malloc */
#include <math.h>           /* Defines several math functions */

#include "ias_math.h"
#include "ias_math_complex.h" /* Defines COMPLEX arithmetic functions */
#include "ias_logging.h"    /* Defines IAS_LOG macros */
#include "model.h"          /* Defines structures and constants */


/******************************************************************************
NAME:           create_frequency_domain_bridge_model

PURPOSE:
Constructs an analytical frequency domain model of the bridge target based
on the input parameters.

RETURN VALUE:
none

******************************************************************************/

void create_frequency_domain_bridge_model
(
    int nspan,                 /* I: Number of spans in current bridge model */
    const BRIDGE_PARMS *bridgeparms,/* I: Input bridge model parameter values */
    int npts,                  /* I: Number of data points to model */
    int osamp,                 /* I: Oversampling factor */
    IAS_COMPLEX *fd_bridge     /* O: Output frequency domain bridge model */
)
{
    double freq;            /* Spatial frequency (radians/meter) */
    double temp;            /* temporary variable */
    IAS_COMPLEX span;       /* Span contribution */
    IAS_COMPLEX bridge;     /* Total bridge contribution */
    IAS_COMPLEX water;      /* Water background contribution */
    IAS_COMPLEX asymmetry;  /* Water asymmetry contribution */
    int ifreq;              /* Integer frequency index */
    int ispan;              /* Integer bridge span index */

    /* Initialize constant */
    const double twopi = 2.0 * ias_math_get_pi();

    /* Loop through the frequency range */
    for (ifreq = 0; ifreq < npts; ifreq++)
    {
        /* Compute the radian frequency */
        freq = osamp * twopi * (ifreq >= npts / 2 ? ifreq - npts : ifreq) /
               npts / bridgeparms->gsd;

        /* Model each bridge span as an offset sinc function */
        bridge.re = 0.0;
        bridge.im = 0.0;
        for (ispan = 0; ispan < nspan; ispan++)
        {
            /* First compute the phase shift argument */
            /* The phase shift is the span offset */
            span.re = 0.0;
            span.im = freq * bridgeparms->span[ispan].offset;

            /* Then compute the phase shift complex exponential */
            ias_math_complex_exponential(&span, &span);

            /* Finally, apply the sinc function */
            temp = (bridgeparms->span[ispan].rad_bridge
                    - bridgeparms->rad_water) * bridgeparms->span[ispan].width
                   * (double)osamp / bridgeparms->gsd
                   * sinc(freq * bridgeparms->span[ispan].width / 2.0);
            span.re *= temp;
            span.im *= temp;

            /* Accumulate the contributions from the bridge spans */
            bridge.re += span.re;
            bridge.im += span.im;
        }

        /* Model the water as a scaled term at DC */
        if (ifreq == 0)
        {
            water.re = (double)npts * bridgeparms->rad_water;
        }
        else
        {
            water.re = 0.0;
        }
        water.im = 0.0;

        /* Model the water asymmetry as another sinc function,
           half the width of the data interval */

        /* first compute the half width */
        temp = (double)npts * bridgeparms->gsd / (double)osamp / 2.0;

        /* then compute the phase shift */
        asymmetry.re = 0.0;
        asymmetry.im = -freq * temp / 2.0;
        ias_math_complex_exponential(&asymmetry, &asymmetry);

        /* now compute the sinc magnitude */
        temp = bridgeparms->rad_asym * npts / 2.0 * sinc(freq * temp / 2.0);
        asymmetry.re *= temp;
        asymmetry.im *= temp;

        /* Add the components for this frequency */
        fd_bridge[ifreq].re = bridge.re + water.re + asymmetry.re;
        fd_bridge[ifreq].im = bridge.im + water.im + asymmetry.im;
    } /* frequency loop */
}

/******************************************************************************
NAME:           create_frequency_domain_stf_model

PURPOSE:
Constructs an analytical frequency domain model of the System Transfer
Function.

RETURN VALUE:
none

******************************************************************************/

void create_frequency_domain_stf_model
(
    const STF_FD_PARMS *stf_fd_parms, /* I: Input STF model parameter values:
                                  optical blur, detector dimension, phase */
    double gsd,             /* I: Model ground sample distance (meters) */
    double range,           /* I: Target range (for scaling to meters) */
    double velocity,        /* I: Target velocity (for scaling to meters) */
    double angle,           /* I: Azimuth angle (for rotating the STF) */
    int npts,               /* I: Number of data points to model */
    int osamp,              /* I: Oversampling factor */
    IAS_COMPLEX *fd_stf     /* O: Output frequency domain STF model */
)
{
    double freq;            /* Spatial frequency (cycles/meter) */
    double gauss;           /* Optical Gaussian blur parameter (meters) */
    double xdetdim;         /* XT Detector dimension parameter (meters) */
    double adetdim;         /* AT Detector dimension parameter (meters) */
    double detint;          /* Detector integration parameter (meters) */
    double diffuse;         /* Charge diffusion parameter (meters) */
    IAS_COMPLEX optics;     /* Optical blur component */
    IAS_COMPLEX xdetector;  /* XT Detector dimension component */
    IAS_COMPLEX adetector;  /* AT Detector dimension component */
    IAS_COMPLEX integration;/* Detector integration component */
    IAS_COMPLEX xdiffusion; /* XT charge diffusion component (for SWIR) */
    IAS_COMPLEX adiffusion; /* AT charge diffusion component (for SWIR) */
    int ifreq;              /* Integer frequency index */

    /* Scale the input parameters based on target range */
    /* milliseconds to meters */
    detint = stf_fd_parms->integ * velocity / 1.0e3;
    gauss = stf_fd_parms->g_optics * range / 1.0e6;      /* microradians to
                                                            meters */
    xdetdim = stf_fd_parms->xt_detector * range / 1.0e6; /* microradians to
                                                            meters */
    adetdim = stf_fd_parms->at_detector * range / 1.0e6; /* microradians to
                                                            meters */
    diffuse = stf_fd_parms->diffusion * range / 1.0e6;  /* microradians to
                                                           meters */

    /* Loop through the frequency range */
    for (ifreq = 0; ifreq < npts; ifreq++)
    {
        /* Compute the frequency (cycles/meter) */
        freq = osamp * (double)(ifreq >= npts / 2 ? ifreq - npts : ifreq) /
               (double)npts / gsd;

        /* Model the optics as a Gaussian */
        compute_frequency_domain_optical_blur(gauss, freq, &optics);

        /* Model charge diffusion effects (mostly for HgCdTe detectors) */
        compute_frequency_domain_charge_diffusion(diffuse, cos(angle)*freq, 
            &xdiffusion);
        compute_frequency_domain_charge_diffusion(diffuse, sin(angle)*freq, 
            &adiffusion);

        /* Model the detector as a sinc function */
        compute_frequency_domain_detector(xdetdim, cos(angle)*freq, &xdetector);
        compute_frequency_domain_detector(adetdim, sin(angle)*freq, &adetector);

        /* Model the integration time as a sinc function */
        compute_frequency_domain_detector(detint, sin(angle)*freq, 
            &integration);

        /* Combine the detector effects */
        ias_math_complex_multiply(&adetector, &integration, &adetector);
        ias_math_complex_multiply(&adetector, &adiffusion, &adetector);
        ias_math_complex_multiply(&adetector, &xdetector, &adetector);
        ias_math_complex_multiply(&adetector, &xdiffusion, &adetector);

        /* Composite optics and detector */
        ias_math_complex_multiply(&optics, &adetector, &fd_stf[ifreq]);
    } /* frequency loop */
}

/******************************************************************************
NAME:           compute_frequency_domain_optical_blur

PURPOSE:
Computes the frequency domain optical blur transfer function value at a
specified frequency using a Gaussian model.

RETURN VALUE: 
none

******************************************************************************/

void compute_frequency_domain_optical_blur
(
    double blur,     /* I: Input STF model optical blur parameter value */
    double freq,     /* I: Frequency (cycles per meter) */
    IAS_COMPLEX *val /* O: Output frequency domain optics model */
)
{
    /* Initialize constant */
    const double twopi = 2.0 * ias_math_get_pi();

    /* Compute the Gaussian model */
    val->re = exp(-twopi * freq * blur * twopi * freq * blur / 2.0);
    val->im = 0.0;
}

/******************************************************************************
NAME:           compute_frequency_domain_charge_diffusion

PURPOSE:
Computes the frequency domain charge diffusion transfer function value at a
specified frequency using an exponential model. The parameter used is 1/f0
to avoid the singularity at f0 = 0.

RETURN VALUE: 
none

******************************************************************************/

void compute_frequency_domain_charge_diffusion
(
    double invf0,    /* I: Input STF model charge diffusion parameter value */
    double freq,     /* I: Frequency (cycles per meter) */
    IAS_COMPLEX *val /* O: Output frequency domain diffusion model */
)
{
    /* Initialize constant */
    const double twopi = 2 * ias_math_get_pi();

    /* Compute the exponential model */
    val->re = exp(-twopi * fabs(freq * invf0));
    val->im = 0.0;
}

/******************************************************************************
NAME:           compute_frequency_domain_detector

PURPOSE:
Computes the frequency domain detector transfer function value at a
specified frequency.

RETURN VALUE: 
none

******************************************************************************/

void compute_frequency_domain_detector
(
    double detdim,   /* I: Input STF model detector dimension parameter */
    double freq,     /* I: Frequency (cycles per meter) */
    IAS_COMPLEX *val /* O: Output frequency domain detector model */
)
{
    /* Initialize constant */
    const double twopi = 2.0 * ias_math_get_pi();

    /* Compute the detector sinc function */
    val->re = sinc(twopi * freq * detdim / 2.0);
    val->im = 0.0;
}

/******************************************************************************
NAME:           sinc

PURPOSE:
Computes the value of the sin(x)/x function at a specified point.

RETURN VALUE:
Type = double

Value           Description
-----           -----------
double          The evaluation of sin(x)/x for input x

******************************************************************************/

double sinc
(
    double x            /* I: Input number to compute sinc for */
)
{
    if (fabs(x) < 1.0e-6)
        return 1.0;
    else
        return (sin(x) / x);
}

/******************************************************************************
NAME:           allocate_spatial_domain_bridge_model

PURPOSE:
Allocate space for an analytical space domain model of the bridge target.

RETURN VALUE:
Type = int

Value   Description
------- --------------------------------------------------------------
SUCCESS Space domain model created
ERROR   Failure to create model

******************************************************************************/
int allocate_spatial_domain_bridge_model
(
    int npts,
    IAS_COMPLEX **fd_bridge_buf, /* O: Frequency domain bridge model */
    IAS_COMPLEX **fd_stf_buf,    /* O: Frequency domain STF model */
    IAS_COMPLEX **model_buf      /* O: Composite target model (FD and SD) */
)
{
    if ((*fd_bridge_buf != NULL) || (*fd_stf_buf != NULL) 
        || (*model_buf != NULL))
    {
        IAS_LOG_ERROR("fd_bridge_buf, fd_stf_buf, and/or "
                      "model_buf already allocated");
        return ERROR;
    }

    *fd_bridge_buf = (IAS_COMPLEX *)malloc(npts * sizeof(IAS_COMPLEX));
    if (*fd_bridge_buf == NULL)
    {
        IAS_LOG_ERROR("Allocating bridge sequence");
        return ERROR;
    }
    *fd_stf_buf = (IAS_COMPLEX *)malloc(npts * sizeof(IAS_COMPLEX));
    if (*fd_stf_buf == NULL)
    {
        IAS_LOG_ERROR("Allocating STF sequence");
        free(*fd_bridge_buf);
        *fd_bridge_buf = NULL;
        return ERROR;
    }
    *model_buf = (IAS_COMPLEX *)malloc(npts * sizeof(IAS_COMPLEX));
    if (*model_buf == NULL)
    {
        IAS_LOG_ERROR("Allocating model sequence");
        free(*fd_bridge_buf);
        *fd_bridge_buf = NULL;
        free(*fd_stf_buf);
        *fd_stf_buf = NULL;
        return ERROR;
    }

    return SUCCESS;
}

/******************************************************************************
NAME:           free_spatial_domain_bridge_model

PURPOSE:
    free memory

RETURN VALUE:
none
******************************************************************************/

void free_spatial_domain_bridge_model
(
    IAS_COMPLEX *fd_bridge_buf, /* I: Frequency domain bridge model */
    IAS_COMPLEX *fd_stf_buf,    /* I: Frequency domain STF model */
    IAS_COMPLEX *model_buf      /* I: Composite target model (FD and SD) */
)
{
    free(fd_bridge_buf);
    fd_bridge_buf = NULL;

    free(fd_stf_buf);
    fd_stf_buf = NULL;

    free(model_buf);
    model_buf = NULL;
}

/******************************************************************************
NAME:           generate_spatial_domain_target_profile

PURPOSE:
Generate the spatial domain target profile

RETURN VALUE:
Type = int

Value   Description
------- --------------------------------------------------------------
SUCCESS Space domain model created
ERROR   Failure to create model

******************************************************************************/
int generate_spatial_domain_target_profile
(
    int npts,               /* I: Number of points in profile */
    int osamp,              /* I: Oversampling factor */
    MTFMODEL mparms,        /* I: Input bridge and STF model parameter values:
                                  water DN, water DN asymmetry, gsd,
                                  {span DN, span width, span offset}
                                  for each span, optical blur, detector
                                  dimensions, phase */
    double range,           /* I: Target range (for scaling to meters) */
    double velocity,        /* I: Target velocity (for scaling to meters) */
    double angle,           /* I: Bridge rotation angle (degrees) */
    double *sd_target       /* O: Output space domain target models */
)
{
    IAS_COMPLEX fd_phase;   /* Scan direction specific FD phase */
    IAS_COMPLEX temp;       /* Temporary variable */
    int ifreq;              /* Frequency index variable */
    int jfreq;              /* IFFT index variable */
    double freq;            /* Radian frequency */
    BRIDGE_PARMS *bridgeparms; /* Pointer to current bridge struct */
    STF_FD_PARMS *stf_fd_parms; /* Pointer to current STF struct */
    IAS_COMPLEX *fd_bridge_buf = NULL; /* Frequency domain bridge model */
    IAS_COMPLEX *fd_stf_buf = NULL; /* Frequency domain STF model */
    IAS_COMPLEX *model_buf = NULL; /* Composite target model (FD and SD) */
    int status;

    /* Initialize constant */
    const double twopi = 2.0 * ias_math_get_pi();

    /* Set the parameter pointers */
    bridgeparms = &mparms.bridge;
    stf_fd_parms = &mparms.stf;

    /* Allocate local working arrays */
    status = allocate_spatial_domain_bridge_model(npts, &fd_bridge_buf, 
        &fd_stf_buf, &model_buf);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Allocating working space for model synthesis");
        return ERROR;
    }

    /* Construct the bridge model */
    create_frequency_domain_bridge_model(mparms.nspan, bridgeparms, npts, 
        osamp, fd_bridge_buf);

    /* Construct the STF model */
    create_frequency_domain_stf_model(stf_fd_parms, bridgeparms->gsd, range, 
        velocity, angle, npts, osamp, fd_stf_buf);

    /* Compute the scan phase and construct the target model */
    for (ifreq = 0; ifreq < npts; ifreq++)
    {
        /* Compute the frequency for this point */
        freq = osamp * twopi * (ifreq >= npts / 2 ? ifreq - npts : ifreq) /
               npts / bridgeparms->gsd;

        /* Compute the phase shift for this point */
        fd_phase.re = 0.0;
        fd_phase.im = -freq * stf_fd_parms->phase;
        ias_math_complex_exponential(&fd_phase, &fd_phase);

        /* Compute the composite FD model for this point */
        ias_math_complex_multiply(&fd_stf_buf[ifreq], &fd_phase, &temp);
        ias_math_complex_multiply(&fd_bridge_buf[ifreq], &temp, 
            &model_buf[ifreq]);

    }

    /* Use the IFFT to convert the model to the space domain */
    ias_math_fft2d(&model_buf[0].re, npts, 1, IAS_MATH_IFFT);

    /* Take the real part and scale to convert to a real output */
    for (ifreq = 0; ifreq < npts; ifreq++)
    {
        jfreq = (ifreq + npts / 2) % npts;
        sd_target[ifreq] = model_buf[jfreq].re / (double)npts;
    }

    free_spatial_domain_bridge_model(fd_bridge_buf, fd_stf_buf, model_buf);
    return SUCCESS;
}
