/******************************************************************************
NAME:           model.h

PURPOSE:
Include file for package of routines that construct an analytical model of the
along- and across-scan system transfer functions, as well as the bridge target.

******************************************************************************/
#ifndef MODEL_H
#define MODEL_H

#include "ias_structures.h"   /* Defines IAS_COMPLEX structure */
#include "ias_satellite_attributes.h"  /* Defines IAS_MAX_NBANDS */

/***************************************************************************
* Constant Declarations
****************************************************************************/
#define MAXNPTS         128 /* Length of test data sequence */

/*
 * Bridge model parameter sizes
 */
#define MINSPAN     1               /* Minimum number of spans supported */
#define MAXSPAN     4               /* Maximum number of spans supported */
#define MINBRIDGE     (MINSPAN*3+3) /* Minimum number of bridge parameters */

/* Bridge model parameter indices */
#define RADWATER        0          /* Background reflectance */
#define RADASYM         1          /* Background reflectance asymmetry */
#define GSD             2          /* Model GSD */
#define RADBRIDGE       3          /* First span reflectance */
#define WIDTH           4          /* First span width */
#define OFFSET          5          /* First span offset from centerline */
#define NBRIDGE       (MAXSPAN*3+3) /* Max number of bridge parameters */


/* STF model parameter indices when in array form*/
#define GOPTICS     0
#define XTDETECTOR  1
#define ATDETECTOR  2
#define INTEG       3
#define DIFFUSION   4
#define PHASE       5       /* Index of phase parameter */
#define NSTF        6       /* Total number of model parameters */



typedef struct bridge_span
{
    double rad_bridge;          /* Span reflectance */
    double width;               /* Span width */
    double offset;              /* Span offset from centerline */
} BRIDGE_SPAN;

typedef struct bridge_parms
{
    double rad_water;           /* Background reflectance */
    double rad_asym;            /* Background reflectance asymmetry */
    double gsd;                 /* Model GSD */
    BRIDGE_SPAN span[MAXSPAN];  /* Span parameters */
} BRIDGE_PARMS;

typedef struct stf_fd_parms
{
    double g_optics;      /* guassian optical blur (urad) */
    double xt_detector;   /* cross-track detector dimension (urad) */
    double at_detector;   /* along-track detector dimension (urad) */
    double integ;         /* detector integration time (msec) */
    double diffusion;     /* charge diffusion distance (not used) */
    double phase;         /* model/data profile registration offset (m) */
} STF_FD_PARMS;

/***************************************************************************
* Structure Declarations
****************************************************************************/
typedef struct mtfmodel
{
    int nstf;            /* Number of STF parameters */
    int  nbridge;        /* Number of bridge parameters (3 + nspan*3) */
    int nspan;           /* Number of bridge spans */
    STF_FD_PARMS stf;    /* Struct of STF parameters */
    BRIDGE_PARMS bridge; /* Struct of bridge target parameters */
} MTFMODEL;

typedef struct stf_fd_parms_mask
{
    int g_optics;      /* guassian optical blur mask */
    int xt_detector;   /* cross-track detector dimension mask */
    int at_detector;   /* along-track detector dimension mask */
    int integ;         /* detector integration time mask */
    int diffusion;     /* charge diffusion distance mask */
    int phase;         /* model/data profile registration offset mask */
} STF_FD_PARMS_MASK;


typedef struct bridge_span_masK
{
    int rad_bridge;          /* Span reflectance mask */
    int width;               /* Span width mask */
    int offset;              /* Span offset from centerline mask */
} BRIDGE_SPAN_MASK;

typedef struct bridge_parms_masK
{
    int rad_water;           /* Background reflectance mask */
    int rad_asym;            /* Background reflectance asymmetry mask */
    int gsd;                 /* Model GSD mask */
    BRIDGE_SPAN_MASK span[MAXSPAN]; /* Span parameter masks */
} BRIDGE_PARMS_MASK;

typedef struct mtfcontrol
{
    STF_FD_PARMS_MASK stfmask;    /* Along-track STF parm flags */
    BRIDGE_PARMS_MASK bridgemask; /* Bridge target parm flags */
} MTFCONTROL;

typedef struct mtfest
{
    int        year_doy[2]; /* Test scene year and day of year */
    int        path_row[2]; /* Test scene path and row */
    int        numbands;    /* Number of bands to process */
    int        band_numbers[IAS_MAX_NBANDS]; /* Band list (user) to process */
    MTFMODEL   model[IAS_MAX_NBANDS]; /* Starting model parameters for each
                                         band */
    MTFMODEL   splex[IAS_MAX_NBANDS]; /* Parameter simplex bounds for each
                                         band */
    int        numiter;     /* Number of fit iterations */
    MTFCONTROL *control;    /* Array of control parameters -
                               one entry per iteration */
} MTFEST;

/***************************************************************************
* Function Declarations
****************************************************************************/
void create_frequency_domain_bridge_model
(
    int nspan,          /* I: Number of spans in current bridge */
    const BRIDGE_PARMS *bridgeparms, /* I: Input bridge model parameter values:
                                  left bridge DN, right bridge DN,
                                  water DN, water DN asymmetry,
                                  span width, span gap */
    int npts,           /* I: Number of data points to model */
    int osamp,          /* I: Oversampling factor */
    IAS_COMPLEX *fd_bridge  /* O: Output frequency domain bridge model */
);

void create_frequency_domain_stf_model
(
    const STF_FD_PARMS *stf_fd_parms, /* I: Input STF model parameter values:
                                  optical blur, detector dimensions,
                                  integration time, charge diffusion,
                                  target phase */
    double  gsd,            /* I: Model ground sample distance (meters) */
    double  range,          /* I: Target range (for scaling to meters) */
    double  velocity,       /* I: Target velocity (for scaling to meters) */
    double  angle,          /* I: Target azimuth (for rotating the STF) */
    int     npts,           /* I: Number of data points to model */
    int     osamp,          /* I: Oversampling factor */
    IAS_COMPLEX *fd_stf);   /* O: Output frequency domain STF model */

void compute_frequency_domain_optical_blur
(
    double blur,      /* I: Input STF model optical blur parameter value */
    double freq,      /* I: Frequency (cycles per meter) */
    IAS_COMPLEX *val  /* O: Output frequency domain optics model */
);

void compute_frequency_domain_detector
(
    double  detdim,   /* I: Input STF model detector dimension parameter */
    double  freq,     /* I: Frequency (cycles per meter) */
    IAS_COMPLEX *val  /* O: Output frequency domain detector model */
);

void compute_frequency_domain_charge_diffusion
(
    double  invf0,    /* I: Input STF model charge diffusion parameter value */
    double  freq,     /* I: Frequency (cycles per meter) */
    IAS_COMPLEX *val  /* O: Output frequency domain diffusion model */
);

double sinc
(
    double x          /* I: Input number to compute sinc for */
);

int generate_spatial_domain_target_profile
(
    int npts,               /* I: Number of points in profile */
    int osamp,              /* I: Oversampling factor */
    MTFMODEL mparms,        /* I: Input bridge and STF parameter values:
                                  water DN, water DN asymmetry, gsd, {span DN,
                                  span width, span offset}, for each span
                                  optical blur, detector dimensions, phase */
    double  range,          /* I: Target range (for scaling to meters) */
    double  velocity,       /* I: Target velocity (for scaling to meters) */
    double  angle,          /* I: Target angle for STF rotation */
    double  *sd_target      /* O: Output space domain target models */
);

int allocate_spatial_domain_bridge_model
(
    int npts,                   /* I: Number of points in profile */
    IAS_COMPLEX **fd_bridge_buf, /* O: Frequency domain bridge model */
    IAS_COMPLEX **fd_stf_buf,   /* O: Frequency domain STF model */
    IAS_COMPLEX **model_buf     /* O: Composite target model (FD and SD) */

);

void free_spatial_domain_bridge_model
(
    IAS_COMPLEX *fd_bridge_buf, /* I: Frequency domain bridge model */
    IAS_COMPLEX *fd_stf_buf,    /* I: Frequency domain STF model */
    IAS_COMPLEX *model_buf      /* I: Composite target model (FD and SD) */
);


#endif /* MODEL_H */
