/******************************************************************************
NAME:           fit_annealing

PURPOSE:
A package of routines that perform the best-fit analysis of analytical models
of the OLI system transfer function and input bridge target(s)
to the observed OLI image data profiles of these target(s).

******************************************************************************/

#include "mtfestimate.h"           /* MTF prototypes */
#include "ias_logging.h"           /* IAS logging library */
#include "model.h"                 /* Defines model and control structures */
#include "profile.h"               /* Defines profile structure */
#include "get_targets.h"           /* Defines target structure */
#include "ias_math.h"              /* IAS math routines */
#include "ias_math_complex.h"      /* IAS complex math routines */
#include "ias_structures.h"        /* For IAS_COMPLEX */


/* The EVAL_PARAMS structure holds parameters needed by the evaluate_model
   routine.  A pointer to an instance of this structure is passed to the 
   library annealing routine. */
typedef struct eval_params
{
    double angle;               /* STF rotation angle */
    int number_bridge_parameters; /* Number of bridge parameters */
    int number_spans;           /* Number of bridge spans */
    double height;              /* Static target range */
    double velocity;            /* Static target velocity */
    int oversample;             /* Oversampling factor for profile */
    double phase_wrap;          /* Maximum absolute phase value */
    double *data_profile;       /* Data profile */
    int *data_weights;          /* Data weight vector */
    double output_bridge_array[NBRIDGE]; /* Bridge parameter values */
    double output_stf_array[NSTF]; /* STF parameter values */
    int bridge_mask_array[NBRIDGE]; /* Bridge parameter hold/vary flags */
    int stf_mask_array[NSTF];    /* STF parameter hold/vary flags */
    IAS_COMPLEX *complex_buffer; /* Complex variable buffer */

} EVAL_PARAMS;

/* Local prototypes */
double evaluate_model
(
    int number_points,             /* I: Number of points */
    double *model_parameters,      /* I/O: Variable model parameters */
    void *params                   /* I: Additional parameters needed for the 
                                         model */
);

double anneal
(
    int number_points,             /* I: Number of points in profile */
    int number_parameters,         /* I: Number of variable parameters */
    EVAL_PARAMS *eval_params,      /* I: Additional parameters needed for the
                                         model */
    const double *simplex_bounds,  /* I: Simplex parameter bounds */
    double *variable_parameters,   /* I/O: Variable parameters */
    int *solution_iterations       /* O: Number of iterations performed */
);

void refine_initial
(
    BRIDGE_PARMS *bridge_parameters, /* I/O: Bridge model parameters */
    int oversample,                  /* I: Oversampling factor for profile */
    const double *data_profile       /* I: Data Profile */
);


/******************************************************************************
NAME:           initialize_fit

PURPOSE:
Sets up the parameters to fit the analytical frequency domain models of the
OLI STF and the bridge target to the input data profile

RETURN VALUE:
Type = int
Value           Description
-----           -----------
SUCCESS         Model setup was successful
ERROR           Model setup failed

******************************************************************************/

int initialize_fit 
(
    int band_index,                /* I: Band index */
    const MTFEST *fit_parameters,  /* I: Initial model parameters */
    const TARGET *target,          /* I: Bridge target parameters */
    const PROFILE *profile,        /* I: Input target profile data */
    MTFMODEL *input_model,         /* O: Bridge/STF model structure */
    MTFMODEL *simplex_model        /* O: Bridge/STF simplex bounds */
)
{
    int band_list_index;           /* Loop counter */
    int target_band_index;         /* Target band index matching STF */
    int span_index;                /* Loop counter */

    /* Load the initial model STF parameters */
    input_model->nstf = fit_parameters->model[band_index].nstf;
    input_model->stf.g_optics = fit_parameters->model[band_index].stf.g_optics;
    input_model->stf.xt_detector 
        = fit_parameters->model[band_index].stf.xt_detector;
    input_model->stf.at_detector 
        = fit_parameters->model[band_index].stf.at_detector;
    input_model->stf.integ = fit_parameters->model[band_index].stf.integ;
    input_model->stf.diffusion 
        = fit_parameters->model[band_index].stf.diffusion;
    input_model->stf.phase = fit_parameters->model[band_index].stf.phase;
        
    /* Load the initial model bridge parameters */
    input_model->nspan = target->number_of_spans;
    input_model->nbridge = 3 * target->number_of_spans + 3;

    /* Match the target band number to the STF band number */
    target_band_index = -1;
    for (band_list_index = 0; band_list_index < target->number_of_bands;
        band_list_index++)
    {
        if (target->band_list[band_list_index] 
            == fit_parameters->band_numbers[band_index])
        {
            target_band_index = band_list_index;
        }
    }

    /* Check for a match */
    if (target_band_index == -1)
    {
        IAS_LOG_ERROR("Unable to find target parameters for requested band "
            "number %d", fit_parameters->band_numbers[band_index]);
        return ERROR;
    }

    /* Load more bridge parameters */
    input_model->bridge.rad_water = target->water[target_band_index];
    input_model->bridge.rad_asym = target->asymmetry[target_band_index];
    input_model->bridge.gsd = profile->gsd; 
    for (span_index = 0; span_index < input_model->nspan; span_index++)
    {
        input_model->bridge.span[span_index].rad_bridge
            = target->span[span_index].reflectance[target_band_index];
        input_model->bridge.span[span_index].width
            = target->span[span_index].width;
        input_model->bridge.span[span_index].offset
            = target->span[span_index].offset;
    }

    /* Load the STF simplex parameters */
    simplex_model->nstf = fit_parameters->splex[band_index].nstf; 
    simplex_model->stf = fit_parameters->splex[band_index].stf;

    /* Load the bridge simplex parameters */
    simplex_model->nspan = input_model->nspan;
    simplex_model->nbridge = input_model->nbridge;
    simplex_model->bridge.rad_water
        = fit_parameters->splex[band_index].bridge.rad_water;
    simplex_model->bridge.rad_asym
        = fit_parameters->splex[band_index].bridge.rad_asym;
    simplex_model->bridge.gsd
        = fit_parameters->splex[band_index].bridge.gsd;
    for (span_index = 0; span_index < simplex_model->nspan; span_index++)
    {
        simplex_model->bridge.span[span_index]
            = fit_parameters->splex[band_index].bridge.span[span_index];
    }

    return SUCCESS;
}

/******************************************************************************
NAME:           fit_annealing

PURPOSE:
Fits an analytical frequency domain model of the OLI STF and bridge target
to the input data profile.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
SUCCESS         Model fit was successful
ERROR           Model fit failed

******************************************************************************/

int fit_annealing
(
    const MTFMODEL *input_model,   /* I: Input bridge/STF model structure */
    const MTFMODEL *simplex_model, /* I: Input bridge/STF simplex bounds */
    int number_iterations,         /* I: Number of fit iterations */
    const MTFCONTROL *control,     /* I: Array of fit control switches */
    const PROFILE *profile,        /* I: Input target profile data */
    double *final_rms,             /* O: Best fit RMS error */
    MTFMODEL *output_model,        /* O: Best-fit output model structure */
    int *solution_iterations       /* O: Number of iterations performed for 
                                         solution */
)
{
    int data_index;                /* Loop counter */
    int iteration;                 /* Loop counter for fit iterations */
    int number_parameters;         /* Number of variable parameters - includes
                                      non-masked bridge and STF parameters */
    int variable_parameter_index;  /* Loop counter into variable parameters */
    int parameter_index;           /* Loop counter into parameters */
    double *variable_parameters;   /* Array of variable parameters */
    double *bounds;                /* Array of variable parameter bounds */
    double rms = 0.0;              /* Model RMS fit */
    int mask_index;                /* Loop counter into bridge and STF masks */
    double simplex_bridge_array[NBRIDGE]; /* Bridge parameter array */
    double simplex_stf_array[NSTF]; /* STF parameter array */
    int number_points;             /* Number of points in profile */
    EVAL_PARAMS eval_params;       /* parameters to pass to evaluate_model */

    /* Extract parameters from profile */
    number_points = profile->number_points;

    /* Initialize the output model */
    *output_model = *input_model;

    /* Initialize the parameter needed for evaluate_model */
    eval_params.angle = profile->angle;
    eval_params.number_bridge_parameters = input_model->nbridge;
    eval_params.number_spans = input_model->nspan;
    eval_params.height = profile->range;
    eval_params.velocity = profile->velocity;
    eval_params.oversample = profile->oversample;
    eval_params.phase_wrap = number_points * output_model->bridge.gsd 
        / eval_params.oversample / 2; 

    /* Allocate the data profile */ 
    eval_params.data_profile = (double *)malloc(number_points * sizeof(double));
    if (eval_params.data_profile == NULL)
    {
        IAS_LOG_ERROR("Allocating temporary profile data array");
        return ERROR;
    }

    /* Initialize the data profile */
    for (data_index = 0; data_index < number_points; data_index++)
    {
        eval_params.data_profile[data_index] = profile->sample_data[data_index];
    }

    /* Allocate the weight vector */ 
    eval_params.data_weights = (int *)malloc(number_points * sizeof(int));
    if (eval_params.data_weights == NULL)
    {
        IAS_LOG_ERROR("Allocating temporary data weight array");
        free(eval_params.data_profile);
        return ERROR;
    }

    /* Initialize the weight vector */
    for (data_index = 0; data_index < number_points; data_index++)
    {
        if (profile->sigma[data_index] > 0.0)
        {
            eval_params.data_weights[data_index] = 1;
        }
        else
        {
            eval_params.data_weights[data_index] = 0;
        }
    }

    /* Allocate the complex variable buffer needed for the annealing 
       procedure */
    eval_params.complex_buffer = (IAS_COMPLEX *)malloc(3 * number_points 
        * sizeof(IAS_COMPLEX));
    if (eval_params.complex_buffer == NULL)
    {
        IAS_LOG_ERROR("Allocating temporary complex data array");
        free(eval_params.data_profile);
        free(eval_params.data_weights);
        return ERROR;
    }

    /* Refine initial estimate of bridge radiances using profile data */
    refine_initial(&output_model->bridge, eval_params.oversample,
        eval_params.data_profile);

    /* Convert the simplex model bridge and STF info to arrays for use in the
       iteration loop */
    convert_bridge_to_array(&simplex_model->bridge, simplex_bridge_array);
    convert_stf_to_array(&simplex_model->stf, simplex_stf_array);
    convert_stf_to_array(&output_model->stf, eval_params.output_stf_array);
    convert_bridge_to_array(&output_model->bridge,
            eval_params.output_bridge_array);

    /* Iteration loop */
    for (iteration = 0; iteration < number_iterations; iteration++)
    {
        /* Count the non-masked variable parameters */
        number_parameters = 0;
        convert_bridge_mask_to_array(&control[iteration].bridgemask, 
            eval_params.bridge_mask_array);
        for (mask_index = 0; mask_index < eval_params.number_bridge_parameters; 
            mask_index++)
        {
            if (eval_params.bridge_mask_array[mask_index] > 0)
            {
                number_parameters++;
            }
        }
        convert_stf_mask_to_array(&control[iteration].stfmask, 
            eval_params.stf_mask_array);
        for (mask_index = 0; mask_index < NSTF; mask_index++)
        {
            if (eval_params.stf_mask_array[mask_index] > 0)
            {
                number_parameters++;
            }
        }

        /* Allocate the variable parameter array */
        variable_parameters = malloc(number_parameters * sizeof(double));
        if (variable_parameters == NULL)
        {
            IAS_LOG_ERROR("Allocating variable parameter array");
            free(eval_params.data_profile);
            free(eval_params.data_weights);
            free(eval_params.complex_buffer);
            return ERROR;
        }

        /* Allocate the variable parameter bounds array */
        bounds = malloc(number_parameters * sizeof(double));
        if (bounds == NULL)
        {
            IAS_LOG_ERROR("Allocating variable parameter bounds array");
            free(eval_params.data_profile);
            free(eval_params.data_weights);
            free(eval_params.complex_buffer);
            free(variable_parameters);
            return ERROR;
        }

        /* Populate variable parameter and variable parameter bounds arrays.
           Don't include masked values.  Include bridge and STF parameters */
        variable_parameter_index = 0;

        for (parameter_index = 0; 
            parameter_index < eval_params.number_bridge_parameters; 
            parameter_index++)
        {
            if (eval_params.bridge_mask_array[parameter_index] > 0)
            {
                variable_parameters[variable_parameter_index]
                    = eval_params.output_bridge_array[parameter_index];
                bounds[variable_parameter_index] 
                    = simplex_bridge_array[parameter_index];
                variable_parameter_index++;
            }
        }

        for (parameter_index = 0; parameter_index < NSTF; parameter_index++)
        {
            if (eval_params.stf_mask_array[parameter_index] > 0)
            {
                variable_parameters[variable_parameter_index]
                    = eval_params.output_stf_array[parameter_index];
                bounds[variable_parameter_index] 
                    = simplex_stf_array[parameter_index];
                variable_parameter_index++;
            }
        }

        /* Evaluate the pre-adjustment fit statistics */
        if (iteration == 0)
        {
            rms = evaluate_model(number_points, variable_parameters,
                &eval_params);
            IAS_LOG_INFO("Iteration #%d  RMS fit %5.3lf", iteration, rms);
        }

        /* Do the simulated annealing fit */
        rms = anneal(number_points, number_parameters, &eval_params, bounds, 
            variable_parameters, &solution_iterations[iteration]);
        if (rms < 0.0)
        {
            IAS_LOG_ERROR("Simulated annealing procedure failed: RMS %f is "
                "less than 0.0", rms);
            free(eval_params.data_profile);
            free(eval_params.data_weights);
            free(eval_params.complex_buffer);
            free(variable_parameters);
            free(bounds);
            return ERROR;
        }

        IAS_LOG_INFO("Iteration #%d  RMS fit %5.3lf", iteration + 1, rms);

        /* Update the model */
        variable_parameter_index = 0;

        for (parameter_index = 0; 
            parameter_index < eval_params.number_bridge_parameters; 
            parameter_index++)
        {
            if (eval_params.bridge_mask_array[parameter_index] > 0)
            {
                eval_params.output_bridge_array[parameter_index]
                    = variable_parameters[variable_parameter_index];
                variable_parameter_index++;
            }
        }

        for (parameter_index = 0; parameter_index < NSTF; parameter_index++)
        {
            if (eval_params.stf_mask_array[parameter_index] > 0)
            {
                eval_params.output_stf_array[parameter_index]
                    = variable_parameters[variable_parameter_index];
                variable_parameter_index++;
            }
        }

        /* Free the parameter arrays */
        free(variable_parameters);
        variable_parameters = NULL;
        free(bounds);
        bounds = NULL;

        /* Make sure the optics and detector parameters are positive (the
           results are symmetric so either solution is correct, but the 
           convention is for them to be positive) */
        eval_params.output_stf_array[GOPTICS]
            = fabs(eval_params.output_stf_array[GOPTICS]);
        eval_params.output_stf_array[XTDETECTOR]
            = fabs(eval_params.output_stf_array[XTDETECTOR]);
        eval_params.output_stf_array[ATDETECTOR]
            = fabs(eval_params.output_stf_array[ATDETECTOR]);
        eval_params.output_stf_array[DIFFUSION]
            = fabs(eval_params.output_stf_array[DIFFUSION]);

        /* Make sure the phase parameters are within range */
        if (eval_params.output_bridge_array[GSD] < 0.0)
        {
            IAS_LOG_WARNING("GSD less than zero %f => Abandoning solution!",
                eval_params.output_bridge_array[GSD]);
            break;
        }

        if (fabs(eval_params.output_stf_array[PHASE]) >= 2.0 
            * eval_params.phase_wrap)
        {
            IAS_LOG_WARNING("Phase has diverged: %lf => Abandoning solution!",
                eval_params.output_stf_array[PHASE]);
            break;
        }
    } /* Iteration loop */

    convert_bridge_to_structure(eval_params.output_bridge_array,
        &output_model->bridge);
    convert_stf_to_structure(eval_params.output_stf_array, &output_model->stf);

    /* Return the final RMS value */
    *final_rms = rms;

    /* Release the allocated buffers */
    free(eval_params.data_profile);
    free(eval_params.data_weights);
    free(eval_params.complex_buffer);

    return SUCCESS;
}

/******************************************************************************
NAME:           evaluate_model

PURPOSE:
Evaluate the model fit objective function.

RETURN VALUE:
Type = double
Value           Description
-----           -----------
double          RMS fit between model and data

******************************************************************************/

double evaluate_model
(
    int number_points,             /* I: Number of points */
    double *model_parameters,      /* I/O: Variable model parameters */
    void *params                   /* I: Additional parameters needed for the 
                                         model */
)
{
    double rms;                    /* RMS model error */
    IAS_COMPLEX *frequency_stf;    /* Frequency domain STF */
    IAS_COMPLEX *frequency_bridge; /* Frequency domain bridge model */
    IAS_COMPLEX frequency_phase;   /* Scan direction specific frequency domain
                                      phase */
    IAS_COMPLEX temp;              /* Temporary variable */
    IAS_COMPLEX *target_model;     /* Composite target model (frequency domain
                                      and spatial domain) */
    int target_model_index;        /* Index into target model structure */
    int variable_parameter_index;  /* Number of variable parameters */
    int number_fit_points;         /* Number of points to fit */
    int point_index;               /* Loop counter */ 
    int scaled_point_index;        /* Point index scaled by number points */
    BRIDGE_PARMS bridge;           /* Local bridge model parameters */
    STF_FD_PARMS stf;              /* Local STF model parameters */
    double frequency;              /* Spatial frequency */
    double model_profile;          /* Model simulated data profile value */
    int mask_index;                /* Loop counter */
    double destination_bridge_array[NBRIDGE]; /* Bridge parameters */
    double destination_stf_array[NSTF]; /* STF parameters */
    double two_pi;                 /* Constant */
    EVAL_PARAMS *eval_params = (EVAL_PARAMS *)params;
    const double *data_profile = eval_params->data_profile;
    const int *data_weights = eval_params->data_weights;
    const double *bridge_array = eval_params->output_bridge_array;
    const double *stf_array = eval_params->output_stf_array;
    const int *bridge_mask_array = eval_params->bridge_mask_array;
    const int *stf_mask_array = eval_params->stf_mask_array;

    /* Initialize constant */
    two_pi = 2.0 * ias_math_get_pi();

    /* Initialize buffers */ 
    frequency_stf = &eval_params->complex_buffer[0];
    frequency_bridge = &eval_params->complex_buffer[number_points];
    target_model = &eval_params->complex_buffer[2 * number_points];

    /* Initialize variables */
    rms = 0.0;
    variable_parameter_index = 0;
    number_fit_points = 0;

    /* Load the variable and static parameters.  First do the bridge 
       parameters */
    for (mask_index = 0; mask_index < eval_params->number_bridge_parameters;
        mask_index++) 
    {
        if (bridge_mask_array[mask_index] > 0)
        {
            destination_bridge_array[mask_index] 
                = model_parameters[variable_parameter_index];
            variable_parameter_index++;
        }
        else
        {
            destination_bridge_array[mask_index] = bridge_array[mask_index];
        }
    }

    /* Next do the STF parameters */
    for (mask_index = 0; mask_index < NSTF; mask_index++) 
    {
        /* Make sure the phase parameter stays in range */
        if (mask_index == PHASE && stf_mask_array[mask_index] > 0)
        {
            double phase_wrap = eval_params->phase_wrap;

            while (model_parameters[variable_parameter_index] < -phase_wrap)
            {
                model_parameters[variable_parameter_index] += 2.0 * phase_wrap;
            }
            while (model_parameters[variable_parameter_index] > phase_wrap)
            {
                model_parameters[variable_parameter_index] -= 2.0 * phase_wrap;
            }
        }

        /* Load the STF parameters */
        if (stf_mask_array[mask_index] > 0)
        {
            destination_stf_array[mask_index] 
                = model_parameters[variable_parameter_index];
            variable_parameter_index++;
        }
        else
        {
            destination_stf_array[mask_index] = stf_array[mask_index];
        }
    }

    /* Put the bridge and STF parameters back in structures */
    convert_bridge_to_structure(destination_bridge_array, &bridge);
    convert_stf_to_structure(destination_stf_array, &stf);

    /* Generate the frequency domain bridge model */
    create_frequency_domain_bridge_model(eval_params->number_spans, &bridge,
        number_points, eval_params->oversample, frequency_bridge);

    /* Generate the frequency domain STF model */
    create_frequency_domain_stf_model(&stf, bridge.gsd, eval_params->height,
        eval_params->velocity, eval_params->angle, number_points, 
        eval_params->oversample, frequency_stf);

    /* Compute the scan phase and construct the target model */
    for (point_index = 0; point_index < number_points; point_index++)
    {
        /* Compute the frequency for this point */
       if (point_index >= number_points / 2)
        {
            scaled_point_index = point_index - number_points;
        }
        else
        {
            scaled_point_index = point_index;
        }

        frequency = eval_params->oversample * two_pi * scaled_point_index  
            / (double)number_points / bridge.gsd;

        /* Compute the phase shift for this point */
        frequency_phase.re = 0.0;
        frequency_phase.im = -frequency * stf.phase;
        ias_math_complex_exponential(&frequency_phase, &frequency_phase);

        /* Compute the composite frequency domain model for this point */
        ias_math_complex_multiply(&frequency_stf[point_index], 
            &frequency_phase, &temp);
        ias_math_complex_multiply(&frequency_bridge[point_index], &temp, 
            &target_model[point_index]);
    }

    /* Use the inverse Fast Fourier Transform to convert the model to the
       space domain */
    ias_math_fft2d((double *)&target_model[0].re, number_points, 1,
        IAS_MATH_IFFT); 

    /* Take the real part and scale to convert to a real output */
    for (point_index = 0; point_index < number_points; point_index++)
    {
        target_model_index = (point_index + number_points / 2) % number_points;
        model_profile = target_model[target_model_index].re 
            / (double)number_points;
        if (point_index > 9 && point_index < number_points - 10)
        {
            rms += (model_profile - data_profile[point_index]) 
                * (model_profile - data_profile[point_index]) 
                * data_weights[point_index];
            number_fit_points += data_weights[point_index];
        }
    }

    /* Compute the final RMS error */
    rms = sqrt(rms / (double)number_fit_points);

    return rms;
}

/******************************************************************************
NAME:           refine_initial

PURPOSE:
Refine the initial estimates of the bridge radiance parameters based on
the observed DN values in the profile data.

RETURN VALUE:
None

******************************************************************************/

void refine_initial
(
    BRIDGE_PARMS *bridge_parameters, /* I/O: Bridge model parameters */
    int oversample,                  /* I: Oversampling factor for profile */
    const double *data_profile       /* I: Data Profile */
)
{
    int data_index;                /* Data loop counter */
    double water_radiance;         /* Estimated water radiance */

    /* Estimate the water radiance using the average of the end of the 
       profiles */
    water_radiance = 0.0;
    for (data_index = 0; data_index < 2 * oversample; data_index++)
    {
        water_radiance += data_profile[data_index];
    }
    water_radiance /= (double)(2 * oversample);

    /* Adjust the input bridge model parameters */
    bridge_parameters->rad_water = water_radiance;
}

/******************************************************************************
NAME:           anneal

PURPOSE:
Find a non-linear best fit by simulated annealing.

RETURN VALUE:
Type = double
Value           Description
-----           -----------
double          RMS fit between model and data

******************************************************************************/

double anneal
(
    int number_points,             /* I: Number of points in profile */
    int number_parameters,         /* I: Number of variable parameters */
    EVAL_PARAMS *eval_params,      /* I: Additional parameters needed for the
                                         model */
    const double *simplex_bounds,  /* I: Simplex parameter bounds */
    double *variable_parameters,   /* I/O: Variable parameters */
    int *solution_iterations       /* O: Number of iterations performed */
)
{
    double rms;                    /* RMS fit */
    double **simplex;              /* Array of parameter vectors comprising
                                      the simplex */
    double *simplex_values;        /* Array of objective function values
                                      evaluated at the simplex points */
    double *best_fit_parameters;   /* Array of best fit parameters */
    double best_value;             /* Objective function value at 
                                      best_fit_parameters */
    double convergence_tolerance;  /* Convergence tolerance */
    double temperature;            /* Annealing temperature */
    double last_value;             /* Previous iteration's best_value */
    unsigned int seed;             /* Seed for random number generator */
    int annealing_iterations;      /* Annealing iteration counter */
    int iteration_limit;           /* Limit on the number of iterations */
    int i;                         /* Loop index */
    int j;                         /* Loop index */
    int status;                    /* Function return status */

    /* Allocate the arrays for the simulated annealing process */
    best_fit_parameters = (double *)malloc(number_parameters * sizeof(double)); 
    if (best_fit_parameters == NULL)
    {
        IAS_LOG_ERROR("Allocating best_fit_parameters");
        return -1.0;
    }

    simplex_values = (double *)malloc((number_parameters + 1) * sizeof(double));
    if (simplex_values == NULL)
    {
        IAS_LOG_ERROR("Allocating simplex_values");
        free(best_fit_parameters);
        return -1.0;
    }

    simplex = (double **)malloc((number_parameters + 1) * sizeof(double *));
    if (simplex == NULL)
    {
        IAS_LOG_ERROR("Allocating simplex");
        free(best_fit_parameters);
        free(simplex_values);
        return -1.0;
    }

    for (i = 0; i < (number_parameters + 1); i++)
    {
        if ((simplex[i] = (double *)malloc(number_parameters * sizeof(double)))
            == NULL)
        {
            IAS_LOG_ERROR("Allocating simplex row %d", i);
            free(best_fit_parameters);
            free(simplex_values);
            for (j = 0; j < i; j++)
            {
                free(simplex[j]);
            }
            free(simplex);
            return -1.0;
        }
    }

    /* Create the simplex */
    memcpy(best_fit_parameters, variable_parameters, number_parameters 
        * sizeof(double));
    best_value = evaluate_model(number_points, best_fit_parameters, 
        eval_params);
    for (i = 0; i < (number_parameters + 1); i++)
    {
        for (j = 0; j < number_parameters; j++)
        {
            simplex[i][j] = variable_parameters[j];
        }
        if (i > 0) 
        {
            simplex[i][i - 1] += simplex_bounds[i - 1];
        }
        simplex_values[i] = evaluate_model(number_points, simplex[i],
            eval_params);
        if (simplex_values[i] < best_value)
        {
            memcpy(best_fit_parameters, simplex[i], number_parameters 
                * sizeof(double));
            best_value = simplex_values[i];
        }
    }

    /* Initialize the number of iterations, last value, and convergence
       tolerance */
    convergence_tolerance = 1.0e-6;
    iteration_limit = 200 * number_parameters;
    last_value = 2 * best_value;

    /* Initialize the random number generator */
    seed = (unsigned int)number_parameters;
    srand(seed);

    /* Annealing Loop */
    for (i = 0; i <= iteration_limit; i++)
    {
        if (best_value < last_value)
        {
            printf("Iter %d (iteration_limit %d) rms %f\n", i, iteration_limit,
                best_value);
            last_value = best_value;
        }

        /* Reset the iteration limit and annealing temperature */
        annealing_iterations = 10 * number_parameters * (number_parameters + 1);
        temperature = 0.2 * best_value * ((double)(iteration_limit - i) 
            / (double)(iteration_limit + 1))
            * ((double)(iteration_limit - i)/(double)(iteration_limit + 1));

        /* Run the simulated annealing procedure */ 
        status = ias_math_simulated_annealing(number_points, simplex, 
            simplex_values, number_parameters, best_fit_parameters, 
            &best_value, convergence_tolerance, &evaluate_model, eval_params,
            &annealing_iterations, temperature);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Simulated annealing failed at iteration %d", i);
            free(best_fit_parameters);
            free(simplex_values);
            for (j = 0; j < (number_parameters + 1); j++)
            {
                free(simplex[j]);
            }
            free(simplex);
            return -1.0;
        }

        /* Check for convergence prior to iteration expiration */
        if (annealing_iterations > 0)
        {
            /* In this case, the best parameters are in simplex index 0 */
            for (j = 0; j < number_parameters; j++)
            {
                best_fit_parameters[j] = simplex[0][j];
            }
            best_value = simplex_values[0];
            break;
        }
    }

    *solution_iterations = annealing_iterations;
    printf("\n");

    /* Move the best parameters to the output array */
    for (i = 0; i < number_parameters; i++)
    {
        variable_parameters[i] = best_fit_parameters[i];
    }

    /* Free the annealing arrays */
    free(best_fit_parameters);
    free(simplex_values);
    for (i = 0; i < (number_parameters + 1); i++)
    {
        free(simplex[i]);
    }
    free(simplex);

    /* Compute the final RMS */
    rms = evaluate_model(number_points, variable_parameters, eval_params);

    return rms;
}

