/******************************************************************************
NAME:           fit_multimin

PURPOSE:
This is an alternative to the fit_annealing.c routines.  Determine solution
for minimization using the multimin simplex method.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
SUCCESS         Model fit was successful
ERROR           Model fit failed

******************************************************************************/

#include <math.h>                  /* Math library */
#include "gsl/gsl_errno.h"         /* Various GSL libraries */
#include "gsl/gsl_math.h"
#include "gsl/gsl_min.h"
#include "gsl/gsl_multimin.h"
#include "gsl/gsl_ieee_utils.h"

#include "ias_logging.h"           /* IAS logging library */
#include "profile.h"               /* Defines profile structure */
#include "mtfestimate.h"           /* MTF prototypes */
#include "model.h"                 /* Defines model and control structures */
#include "minimize_function_multimin.h" /* Defines function to minimize */

int fit_multimin
(
    MTFMODEL *input_model,         /* I/O: Input bridge/STF model structure */
    MTFMODEL *simplex_model,       /* I: Input bridge/STF simplex bounds */
    int number_iterations,         /* I: Number of fit iterations */
    MTFCONTROL *control,           /* I: Array of fit control switches */
    PROFILE *profile,              /* I: Input target profile data */
    double *final_rms,             /* O: Best fit RMS error */
    MTFMODEL *output_model,        /* O: Best-fit output model structure */
    int *solution_iterations       /* O: Number of iterations performed for 
                                         solution */
)
{
    /* Set minimization variables.  Stick with the GSL variable names */
    gsl_multimin_function minimize_function; /* Function to minimize */
    const gsl_multimin_fminimizer_type *T =
        gsl_multimin_fminimizer_nmsimplex2; /* Miminization type T */
    gsl_multimin_fminimizer *s = NULL; /* Minimizer s */
    double epsabs = 1.0e-7;        /* Absolute tolerance */

    size_t data_index;             /* Loop counter */
    int iteration;                 /* Loop counter */
    int iteration_counter;         /* Loop counter */
    double water;                  /* Estimated water radiance */
    double phase_wrap;             /* Maximum absolute phase value */
    struct function_parameters_multimin function_parameters; /* Structure to 
                                      define error function */
    size_t number_parameters;      /* Total number of coefficients */
    int parameter_index;           /* Loop counter */
    gsl_vector *step_size;         /* Step size */
    gsl_vector *solution_vector;   /* Solution vector */
    gsl_vector *best;              /* Best fit parameters */
    double best_value;             /* Objective function value at best */
    int maximum_iterations = get_maximum_iterations(); /* User-selected limit */
    double current_rms;            /* Model RMS fit */
    double size = 0.0;             /* Size of simplex fit */
    int bridge_mask_array[NBRIDGE]; /* Bridge mask */
    int stf_mask_array[NSTF];      /* STF mask */
    double simplex_bridge_array[NBRIDGE]; /* Bridge parameter array */
    double simplex_stf_array[NSTF]; /* STF parameter array */
    double output_bridge_array[NBRIDGE]; /* Bridge parameter array */
    double output_stf_array[NSTF]; /* STF parameter array */
    char message[500];             /* Log message string */
    int message_position = 0;      /* Position in log message string */
    int message_size;              /* Size returned by snprintf */
    int allowed_size;              /* Size remaining in message */
    int status;                    /* Function return status */

    /* Assign the minimize function */
    minimize_function.f = &minimize_function_multimin;

    /* Estimate the water radiance from the end of the profiles */
    water = 0.0;
    for (data_index = 0; data_index < 2 * profile->oversample; data_index++)
    {
        water += profile->sample_data[data_index];
    }
    water /= (double)(2 * profile->oversample);
    input_model->bridge.rad_water = water;

    /* Initialize the output model */
    output_model->nbridge = input_model->nbridge;
    output_model->nspan = input_model->nspan;
    output_model->nstf = input_model->nstf;
    output_model->bridge = input_model->bridge;
    output_model->stf = input_model->stf;

    phase_wrap = (double)profile->number_points * output_model->bridge.gsd 
        / profile->oversample / 2.0;

    /* Allocate complex arrays for MTF function */
    if (initialize_minimize_function_multimin(profile->number_points) 
        != SUCCESS)
    {
        IAS_LOG_ERROR("Setting up simplex");
        return ERROR;
    }

    /* Initialize method and iterate */
    for (iteration = 0; iteration < number_iterations; iteration++)
    {
        /* Set structure parameters */
        function_parameters.input_model = input_model;
        function_parameters.simplex_model = simplex_model;
        function_parameters.control = &control[iteration];
        function_parameters.profile = profile;
        function_parameters.output_model = output_model;
        minimize_function.params = &function_parameters;

        /* Determine number of variables to solve for */
        number_parameters = 0;
        convert_bridge_mask_to_array(&control[iteration].bridgemask,
            bridge_mask_array);
        for (parameter_index = 0; parameter_index < input_model->nbridge;
            parameter_index++)
        {
            if (bridge_mask_array[parameter_index] > 0)
            {
                number_parameters++;
            }
        }
        convert_stf_mask_to_array(&control[iteration].stfmask, stf_mask_array);
        for (parameter_index = 0; parameter_index < NSTF; parameter_index++)
        {
            if (stf_mask_array[parameter_index] > 0)
            {
                number_parameters++;
            }
        }
        minimize_function.n = number_parameters;

        /* Allocate vectors */
        step_size = gsl_vector_alloc(number_parameters);
        solution_vector = gsl_vector_alloc(number_parameters);
        best = gsl_vector_alloc(number_parameters);

        /* Initialize bridge variables */
        number_parameters = 0;
        convert_bridge_to_array(&simplex_model->bridge, simplex_bridge_array);
        convert_bridge_to_array(&output_model->bridge, output_bridge_array);
        for (parameter_index = 0; parameter_index < input_model->nbridge;
            parameter_index++)
        {
            if (bridge_mask_array[parameter_index] > 0)
            {
                gsl_vector_set(step_size, number_parameters, 
                    simplex_bridge_array[parameter_index]);
                gsl_vector_set(solution_vector, number_parameters, 
                    output_bridge_array[parameter_index]);
                number_parameters++;
            }
        }

        /* Initialize spatial variables */
        convert_stf_to_array(&simplex_model->stf, simplex_stf_array);
        convert_stf_to_array(&output_model->stf, output_stf_array);
        for (parameter_index = 0; parameter_index < NSTF; parameter_index++)
        {
            if (stf_mask_array[parameter_index] > 0)
            {
                gsl_vector_set(step_size, number_parameters, 
                    simplex_stf_array[parameter_index]);
                gsl_vector_set(solution_vector, number_parameters, 
                    output_stf_array[parameter_index]);
                number_parameters++;
            }
        }

        /* Get current RMS */
        for (parameter_index = 0; parameter_index < number_parameters; 
            parameter_index++)
        {
            gsl_vector_set(best, parameter_index, gsl_vector_get(
                solution_vector, parameter_index));
        }
        best_value = minimize_function_multimin(solution_vector, 
            &function_parameters);

        iteration_counter = 0;

        /* Print out initial results */
        if (iteration == 0)
        {
            current_rms = minimize_function_multimin(solution_vector, 
                &function_parameters);

            /* Build output message string */
            message_size = snprintf(message, sizeof(message), "Iteration "
                "%5d %5d", iteration, iteration_counter);
            if ((message_size < 0) || (message_size >= sizeof(message)))
            {
                IAS_LOG_ERROR("Cannot build log message");
                return ERROR;
            }

            for (parameter_index = 0; parameter_index < number_parameters; 
                parameter_index++)
            {
                message_position += message_size;
                allowed_size = sizeof(message) - message_position;
                message_size = snprintf(&message[message_position], 
                    allowed_size, " %f", gsl_vector_get(solution_vector, 
                    parameter_index)); 
                if ((message_size < 0) || (message_size >= allowed_size))
                {
                    IAS_LOG_ERROR("Cannot build log message");
                    return ERROR;
                }
            }

            IAS_LOG_INFO("%s RMS = %f", message, current_rms);
            message_position = 0;
        }

        /* Allocate and set up simplex solution */
        s = gsl_multimin_fminimizer_alloc(T, number_parameters);
        gsl_multimin_fminimizer_set(s, &minimize_function, solution_vector, 
            step_size);

        /* Find solution for current starting location */
        do
        {
            iteration_counter++;
            status = gsl_multimin_fminimizer_iterate(s);

            if (status)
            {
                break;
            }

            size = gsl_multimin_fminimizer_size(s);
            status = gsl_multimin_test_size(size, epsabs);

            /* Keep track of the best value */
            if (s->fval < best_value)
            {
                for (parameter_index = 0; parameter_index < number_parameters; 
                    parameter_index++)
                {
                    gsl_vector_set(best, parameter_index, 
                        gsl_vector_get(s->x, parameter_index));
                }
                best_value = minimize_function_multimin(best, 
                    &function_parameters);
            }

            if (status == GSL_SUCCESS)
            {
                IAS_LOG_INFO("converged to minimum at");
            }
        }
        while (status == GSL_CONTINUE && iteration_counter 
            < maximum_iterations);

        /* Store number of iterations used for solution */
        solution_iterations[iteration] = iteration_counter;

        /* Determine whether solution has converged.  If solution has not
           converged used last best value */
        if (iteration_counter < maximum_iterations)
        {
            for (parameter_index = 0; parameter_index < number_parameters; 
                parameter_index++)
            {
                gsl_vector_set(best, parameter_index, 
                    gsl_vector_get(s->x, parameter_index));
            }
            best_value = minimize_function_multimin(best, &function_parameters);
        }

        /* Build output message string */
        message_size = snprintf(message, sizeof(message), "Iteration %5d %5d", 
            iteration + 1, iteration_counter);
        if ((message_size < 0) || (message_size >= sizeof(message)))
        {
            IAS_LOG_ERROR("Cannot build log message");
            return ERROR;
        }

        for (parameter_index = 0; parameter_index < number_parameters; 
            parameter_index++)
        {
            message_position += message_size;
            allowed_size = sizeof(message) - message_position;
            message_size = snprintf(&message[message_position],
                allowed_size, " %f", gsl_vector_get(best, parameter_index)); 
            if ((message_size < 0) || (message_size >= allowed_size))
            {
                IAS_LOG_ERROR("Cannot build log message");
                return ERROR;
            }
        }

        IAS_LOG_INFO("%s RMS = %f size = %.5e", message, best_value, size);
        message_position = 0;

        /* Store current iteration best value in output model structure.  This
           structure will be used for next iteration */
        number_parameters = 0;
        for (parameter_index = 0; parameter_index < input_model->nbridge;
            parameter_index++)
        {
            if (bridge_mask_array[parameter_index] > 0)
            {
                output_bridge_array[parameter_index]
                    = gsl_vector_get(best, number_parameters++);
            }
        }
        convert_bridge_to_structure(output_bridge_array, &output_model->bridge);

        for (parameter_index = 0; parameter_index < NSTF; parameter_index++)
        {
            if (stf_mask_array[parameter_index] > 0)
            {
                output_stf_array[parameter_index]
                    = gsl_vector_get(best, number_parameters++);
            }
        }
        convert_stf_to_structure(output_stf_array, &output_model->stf);

        /* Extract RMS */
        *final_rms = s->fval;

        /* Free vectors and minimizer */
        gsl_multimin_fminimizer_free(s);
        gsl_vector_free(solution_vector);
        gsl_vector_free(step_size);
        gsl_vector_free(best);
        solution_vector = NULL;
        step_size = NULL;
        best = NULL;

        /* Make sure the optics and detector parameters are positive (the
           results are symmetric so either solution is correct, but the
           convention is for them to be positive) */
        output_model->stf.g_optics = fabs(output_model->stf.g_optics);
        output_model->stf.at_detector = fabs(output_model->stf.at_detector);
        output_model->stf.xt_detector = fabs(output_model->stf.xt_detector);
        output_model->stf.diffusion = fabs(output_model->stf.diffusion);

        /* Make sure the phase parameters are within range */
        if (output_model->bridge.gsd < 0.0)
        {
            IAS_LOG_WARNING("GSD less than zero %f => Abandoning solution!",
                output_model->bridge.gsd);
            break;
        }

        if (fabs(output_model->stf.phase) >= 2.0 * phase_wrap)
        {
            IAS_LOG_WARNING("Phase has diverged: %lf => Abandoning solution!", 
                output_model->stf.phase);
            break;
        }
    } /* Iteration loop */

    /* Close minimize function */
    close_minimize_function_multimin();

    return SUCCESS;
}

