/******************************************************************************
NAME:           get_acquisition_date

PURPOSE:
Read target acquisition date from the L0R metadata file.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
SUCCESS         Successful completion
ERROR           Operation failed 

******************************************************************************/
#include "ias_const.h"             /* IAS constants */
#include "ias_structures.h"        /* IAS data structures */
#include "ias_types.h"             /* IAS data types */
#include "ias_miscellaneous.h"     /* IAS miscellaneous functions */
#include "ias_logging.h"           /* IAS logging library */
#include "ias_l0r.h"               /* L0R I/O library */

int get_acquisition_date 
(
    const char *l0r_file_name,     /* I: L0R file name */
    int *year,                     /* O: Acquistion year */
    int *day_of_year               /* O: Acquistion day of year */
)
{
    L0RIO *l0r_image = NULL;       /* L0R image structure */
    IAS_L0R_SCENE_METADATA scene_metadata; /* L0R scene metadata structure */
    IAS_DATETIME date_time;        /* Structure for date and time values */
    int status;                    /* Function status return value */

    /* Open the L0R file */
    l0r_image = ias_l0r_open_from_filename(l0r_file_name, 0);
    if (!l0r_image)
    {
        IAS_LOG_ERROR("Opening the L0R file %s", l0r_file_name);
        return ERROR;
    }

    /* Open the L0R metadata */
    status = ias_l0r_open_metadata(l0r_image, IAS_READ);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Opening the L0R metadata for %s", l0r_file_name);
        ias_l0r_close(l0r_image);
        return ERROR;
    }

    /* Retrieve the acquisition date */
    status = ias_l0r_get_scene_metadata(l0r_image, 0, 1, &scene_metadata);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading the scene metadata for %s", l0r_file_name);
            
        ias_l0r_close_metadata(l0r_image);
        ias_l0r_close(l0r_image);
        return ERROR;
    }

    /* Close the metadata */
    status = ias_l0r_close_metadata(l0r_image);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Closing the L0R metadata for %s", l0r_file_name);
        ias_l0r_close(l0r_image);
        return ERROR;
    }

    /* Close the L0R file */
    status = ias_l0r_close(l0r_image);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Closing the L0R file %s", l0r_file_name);
        return ERROR;
    }
    l0r_image = NULL;

    /* Parse the datetime string provided by the L0R */
    if (ias_misc_parse_datetime_string(scene_metadata.date_acquired,
            IAS_DATETIME_L0R_FORMAT, &date_time) != SUCCESS)
    {
        IAS_LOG_ERROR("Parsing L0R scene metadata date acquired string");
        return ERROR;
    }

    /* Copy the acquisition year and day of year */ 
    *year = date_time.year;
    *day_of_year = date_time.day_of_year;

    return SUCCESS;
}

