/******************************************************************************
NAME:           get_parameters

PURPOSE:
Get the mtfestimate parameters from the input parmaeter file and OMF file.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
SUCCESS         Successful completion
ERROR           Operation failed 

******************************************************************************/
#include <string.h> 
#include <limits.h> 
#include <stdlib.h> 

#include "model.h"
#include "mtf.h"
#include "ias_logging.h"
#include "ias_const.h"
#include "ias_satellite_attributes.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "mtfestimate.h"

static int maximum_iterations;

int get_parameters 
(
    const char *parm_file,         /* I: Name of the parameter file */
    char *target_file,             /* O: Input target description file name */ 
    char *input_directory,         /* O: MTF input data directory */
    char *l0r_file,                /* O: L0R file name */
    char *l1g_file,                /* O: L1G file name */
    char *work_order_id,           /* O: Work order ID string */
    int *stf_fit_method,           /* O: Method to use (annealing or multimin)
                                         for multi-parameter fitting */
    MTFEST *fit_parameters,        /* O: Structure containing fit parameters */
    int *store_in_database_flag    /* O: Flag for storing results in the DB */
)
{
    int count;                     /* Number of elements */
    char omf_file[PATH_MAX];       /* Name of OMF */
    int band_number_list[IAS_MAX_NBANDS]; /* List of band numbers from 
                                      parameter file */
    int band_number_list_count;    /* Count of bands read from parameter file */
    int oli_band_numbers[IAS_MAX_NBANDS]; /* List of OLI band numbers */
    int number_oli_bands;          /* Number of OLI bands */
    int oli_band_index;            /* Loop counter */
    int work_order_band_index;     /* Loop counter */
    IAS_SATELLITE_ID satellite_id; /* Satellite Id */
    int status;                    /* Status returned from functions  */

    /* Local variables for storing ODL values */
    char local_target_file[PATH_MAX];
    char local_input_directory[PATH_MAX];
    char local_l0r_file[PATH_MAX];
    char local_l1g_file[PATH_MAX];
    char local_work_order_id[IAS_WORKORDER_ID_SIZE];
    int local_satellite_number;

    /* Initialize to empty string */
    strcpy(local_work_order_id, "");
    strcpy(local_target_file, "");
    strcpy(local_input_directory, "");
    strcpy(local_l0r_file, "");
    strcpy(local_l1g_file, "");

    /* Start the parameter file table */
    IAS_PARM_DECLARE_TABLE(parms, 5);

    /* Get the work order id */
    IAS_PARM_WORK_ORDER_ID(parms, local_work_order_id, 
        sizeof(local_work_order_id), 0);

    /* Get the parameter fitting method */ 
    int default_stf_fit_method = 1;
    IAS_PARM_ADD_INT(parms, STF_FIT_METHOD, "Method used to perform STF FIT; "
        "0 -> Annealing, 1 -> MultiMin", IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
         1, 0, 1, 1, &default_stf_fit_method, stf_fit_method, 
         sizeof(*stf_fit_method), 1);

    /* Get the maximum number of simulated annealing iterations */ 
    int default_maximum_iterations = 50000;
    IAS_PARM_ADD_INT(parms, MAXIMUM_SOLUTION_ITERATIONS, "Maximum number of "
        "simulated annealing iterations to perform in each solution " 
        "iteration", IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 0, 0, 0, 1, 
        &default_maximum_iterations, &maximum_iterations, 
        sizeof(maximum_iterations), 1);

    /* Get the band numbers to process */
    IAS_PARM_BAND_LIST(parms, IAS_PARM_REQUIRED, band_number_list,
        sizeof(band_number_list), 1);

   /* Get the flag to write (or not write) data to the database */
    IAS_PARM_STORE_CHARACTERIZATION_IN_DB(parms, IAS_PARM_OPTIONAL,
        store_in_database_flag, sizeof(*store_in_database_flag), 0);

    count = IAS_PARM_GET_TABLE_SIZE(parms);

    /* Check if help or template output was requested */
    status = ias_parm_provide_help(parm_file, parms, count,
                                   IAS_INPUT_PARAMETERS);

    if (status == 1)
    {
        /* --help or --template was requested.  Copy the requested option to
           omf_file so the same operation will be performed for the OMF */
        strcpy(omf_file, parm_file);
    }
    else if (status == ERROR) /* Help failed. */
    {
        IAS_LOG_ERROR("Providing help for application parameter file %s",
            parm_file);
        exit(EXIT_FAILURE);
    }
    else
    {
        if (ias_parm_read(parm_file, parms, count) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading parameters from: %s", parm_file);
            return ERROR;
        }

        /* Copy and verify parameters */
        strcpy(target_file, local_target_file);

        if (strcmp(local_work_order_id, "") == 0)
        {
            IAS_LOG_ERROR("Missing work order id in the parameter file.");
            return ERROR;
        }

        /* Build the OMF file name */
        sprintf(omf_file, "%s.omf", local_work_order_id);
    }

    /***** Read additional parameters from OMF *****/

    /* Start the OMF table */
    IAS_PARM_DECLARE_TABLE(omf_parms, 8);

    /* Get the satellite number */
    IAS_PARM_SATELLITE(omf_parms, &local_satellite_number,
        sizeof(local_satellite_number), 1);

    /* Get the MTF Target description file name */
    IAS_PARM_MTF_TARGET_DESCRIPTION_FILENAME(omf_parms, IAS_PARM_REQUIRED, 
        local_target_file, sizeof(local_target_file));

    /* Get the input directory name */
    IAS_PARM_MTF_INPUT_DIRECTORY(omf_parms, local_input_directory, 
        sizeof(local_input_directory));

    /* Get the L0r file name for the standard report header routine */
    IAS_PARM_L0R_FILENAME(omf_parms, IAS_PARM_REQUIRED, local_l0r_file,
        sizeof(local_l0r_file), 1);

    /* Get the L1G file name from pass 1 (MTF is in pass 2 but it does not
       make the pass 2 L1G file).  The L1G is used in the standard report 
       header routine */
    IAS_PARM_L1G_FILENAME_FROM_PASS(omf_parms, IAS_PARM_REQUIRED,
        local_l1g_file, sizeof(local_l1g_file), 0, 1);

    /* Get the target WRS path */
    IAS_PARM_TARGET_WRS_PATH(omf_parms, IAS_PARM_REQUIRED, 
        &fit_parameters->path_row[0], sizeof(fit_parameters->path_row[0]), 0);

    /* Get the target WRS row */
    IAS_PARM_TARGET_WRS_ROW(omf_parms, IAS_PARM_REQUIRED, 
        &fit_parameters->path_row[1], sizeof(fit_parameters->path_row[1]), 0);

    count = IAS_PARM_GET_TABLE_SIZE(omf_parms);

    /* Check if help or template output is being requested */
    status = ias_parm_provide_help(omf_file, omf_parms, count,
        IAS_OMF_PARAMETERS);

    if (status == 1)
    {
        /* the --help or --template was provided. can now exit success */
        exit(EXIT_SUCCESS);
    }
    else if (status == ERROR) /* Help failed */
    {
        IAS_LOG_ERROR("Providing help for OMF file: %s", omf_file);
        return ERROR;
    }

    /* Read the OMF file */
    if (ias_parm_read(omf_file, omf_parms, count) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from OMF file: %s", omf_file);
        return ERROR;
    }

    /* Copy OMF parameters */
    strcpy(input_directory, local_input_directory);
    strcpy(target_file, local_target_file);
    strcpy(l0r_file, local_l0r_file);
    strcpy(l1g_file, local_l1g_file);
    strcpy(work_order_id, local_work_order_id);

    /* Convert the satellite number to the satellite id */
    satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(
        local_satellite_number);
    if (satellite_id == ERROR)
    {
        IAS_LOG_ERROR("Unable to determine satellite ID from satellite "
            "number %d", local_satellite_number);
        return ERROR;
    }

    /* Initialize the satellite attributes */
    if (ias_sat_attr_initialize(satellite_id) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attributes library");
        exit(EXIT_FAILURE);
    }

    /* Make sure the band list only includes OLI bands. First get the list of
       OLI bands */
    status = ias_sat_attr_get_sensor_band_numbers(IAS_OLI, IAS_NORMAL_BAND, 0,
        oli_band_numbers, IAS_MAX_NBANDS, &number_oli_bands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting OLI band numbers");
        return ERROR;
    }

    /* Get the number of bands read from parameter file */
    band_number_list_count = IAS_PARM_GET_COUNT_READ(parms, BAND_LIST);

    /* Compare the list of OLI bands to the work order list of bands, and only
       accept work order bands that are OLI bands into the fit parameter
       structure */
    fit_parameters->numbands = 0;
    for (oli_band_index = 0; oli_band_index < number_oli_bands; 
        oli_band_index++)
    {
        for (work_order_band_index = 0; 
            work_order_band_index < band_number_list_count; 
            work_order_band_index++)
        {
            if (band_number_list[work_order_band_index] == 
                oli_band_numbers[oli_band_index])
            {
                fit_parameters->band_numbers[fit_parameters->numbands] 
                    = band_number_list[work_order_band_index];
                fit_parameters->numbands++;
                break;
            }
        }
    }

    /* Extract the year and day of year from the L0R */
    status = get_acquisition_date(l0r_file, &fit_parameters->year_doy[0], 
        &fit_parameters->year_doy[1]);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting acquisition date from L0R %s", l0r_file);
        return ERROR;
    }

    /* Display a message about the fit method */
    if (*stf_fit_method == 0)
    {
        IAS_LOG_INFO("Using Annealing");        
    }
    else
    {
        IAS_LOG_INFO("Using MultiMin");        
    }

    return SUCCESS;
}

/* Access routine */
int get_maximum_iterations()
{
    return(maximum_iterations);
}
