/******************************************************************************
NAME:           get_system_table

PURPOSE:
Get the MTF Estimate processing parameters from the MTF system table
parameter file.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
SUCCESS         Successful completion
ERROR           Operation failed 

******************************************************************************/
#include <limits.h>                /* PATH_MAX */
#include <stdlib.h>                /* malloc */

#include "mtfestimate.h"           /* MTF Estimate library */
#include "model.h"                 /* MTF model library */
#include "mtf.h"                   /* Shared MTF definitions */
#include "ias_math.h"              /* IAS math library */
#include "ias_logging.h"           /* IAS logging library */
#include "ias_odl.h"               /* IAS ODL library */

#define SPAN_OFFSET 3              /* Span data starts after 3 bridge items */
#define MAX_FIELD_SIZE 256         /* Maximum supported size of ODL fields in
                                      the MTF system table */

int get_system_table 
(
    const char *input_directory,   /* I: Directory for system table */
    MTFEST *fit_parameters         /* O: Structure containing fit parameters */
)
{
    char system_table_file[PATH_MAX]; /* ODL-format MTF system table file */ 
    int band_numbers[IAS_MAX_NBANDS]; /* List of system table band numbers */
    int number_bands;              /* Number of bands in the system table */
    IAS_OBJ_DESC *odl_file;        /* ODL tree structure */
    int count;                     /* Number of elements read */
    int size;                      /* Size of elements to read */
    double *parameter_buffer;      /* Parameters are put here first */
    int *mask_parameter_buffer;    /* Mask parameters are put here first */
    char field_name[MAX_FIELD_SIZE]; /* ODL field to read - assumes no field
                                      name will exceed the ODL line limit */
    int fit_band_index;            /* Loop counter for band numbers in 
                                      fit_parameters structure */
    int table_band_index;          /* Loop counter for band numbers in
                                      system table structure */
    int span_index;                /* Loop counter for spans */
    int iteration_index;           /* Loop counter for fit iterations */
    int band_found;                /* Flag indicating whether the requested
                                      band number was found */
    int status;                    /* Function status return value */

    /* Build the full system table filename */
    size = snprintf(system_table_file, sizeof(system_table_file), "%s/%s", 
        input_directory, MTF_SYSTEM_FILE); 
    if ((size < 0) || (size >= sizeof(system_table_file)))
    {
        IAS_LOG_ERROR("Cannot build system table file name");
        return ERROR;
    }

    /* Read ODL-format system table file */
    odl_file = ias_odl_read_tree(system_table_file);
    if (odl_file == NULL)
    {
        IAS_LOG_ERROR("Reading parameters from %s", system_table_file);
        return ERROR;
    }

    /* Get the band numbers in the system table */
    strcpy(field_name, "BAND_LIST");
    status = ias_odl_get_field(band_numbers, sizeof(band_numbers), IAS_ODL_Int,
        odl_file, NULL, field_name, &count);
    number_bands = count;
    if (status != SUCCESS)
    {
        process_odl_error(status, field_name, system_table_file);
        ias_odl_free_tree(odl_file);
        return ERROR;
    }

    /* Set the size to read based on the number of bands and the larger of the
       sizes of the bridge and STF parameter lists */ 
    size = IAS_MAX(NBRIDGE, NSTF) * number_bands; 

    /* Allocate space for the parameter buffer.  This is sized to fit bridge, 
       STF, bridge simplex, and STF simplex parameters (one at a time) */
    parameter_buffer = (double *)malloc(size * sizeof(double)); 
    if (parameter_buffer == NULL)
    { 
        IAS_LOG_ERROR("Allocating parameter buffer");
        ias_odl_free_tree(odl_file);
        return ERROR;
    } 

    /* Get the default bridge parameters */
    strcpy(field_name, "BRIDGE");
    status = ias_odl_get_field(parameter_buffer, MINBRIDGE * number_bands 
        * sizeof(double), IAS_ODL_Double, odl_file, NULL, field_name, &count);
    if (status != SUCCESS)
    {
        process_odl_error(status, field_name, system_table_file);
        ias_odl_free_tree(odl_file);
        free(parameter_buffer);
        return ERROR;
    }
    if (count < MINBRIDGE * number_bands)
    {
        IAS_LOG_ERROR("Getting %s parameters from %s, %d records were read, "
            "but %d were expected", field_name, system_table_file, count, 
            MINBRIDGE * number_bands);
        ias_odl_free_tree(odl_file);
        free(parameter_buffer);
        return ERROR;
    }

    /* Populate the bridge fit_parameters structure elements */
    for (fit_band_index = 0; fit_band_index < fit_parameters->numbands;
        fit_band_index++)
    {
        /* Check to see if this band number is in the system table. */
        band_found = 0;
        for (table_band_index = 0; table_band_index < number_bands;
            table_band_index++)
        {
            if (fit_parameters->band_numbers[fit_band_index] == 
                band_numbers[table_band_index])
            {
                band_found = 1;

                fit_parameters->model[fit_band_index].bridge.rad_water
                    = parameter_buffer[table_band_index * MINBRIDGE];
                fit_parameters->model[fit_band_index].bridge.rad_asym
                    = parameter_buffer[table_band_index * MINBRIDGE + 1];
                fit_parameters->model[fit_band_index].bridge.gsd
                    = parameter_buffer[table_band_index * MINBRIDGE + 2];

                for (span_index = 0; span_index < MAXSPAN; span_index++)
                {
                    fit_parameters->model[fit_band_index].bridge
                        .span[span_index].rad_bridge
                        = parameter_buffer[table_band_index * MINBRIDGE
                        + SPAN_OFFSET];
                    fit_parameters->model[fit_band_index].bridge
                        .span[span_index].width
                        = parameter_buffer[table_band_index * MINBRIDGE
                        + SPAN_OFFSET + 1];
                    fit_parameters->model[fit_band_index].bridge
                        .span[span_index].offset
                        = parameter_buffer[table_band_index * MINBRIDGE
                        + SPAN_OFFSET + 2];
                }
            }
        }

        if (band_found == 0)
        {
            IAS_LOG_ERROR("Initialization data for band number %d not found "
                "in MTF system table %s", 
                fit_parameters->band_numbers[fit_band_index], 
                system_table_file);
            ias_odl_free_tree(odl_file);
            free(parameter_buffer);
            return ERROR;
        }
    }

    /* Get the bridge simplex bounds */
    strcpy(field_name, "BRIDGE_SIMPLEX");
    status = ias_odl_get_field(parameter_buffer, MINBRIDGE * number_bands
        * sizeof(double), IAS_ODL_Double, odl_file, NULL, field_name, &count);
    if (status != SUCCESS)
    {
        process_odl_error(status, field_name, system_table_file);
        ias_odl_free_tree(odl_file);
        free(parameter_buffer);
        return ERROR;
    }
    if (count < MINBRIDGE * number_bands)
    {
        IAS_LOG_ERROR("Getting %s parameters from %s, %d records were read, "
            "but %d were expected", field_name, system_table_file, count, 
            MINBRIDGE * number_bands);
        ias_odl_free_tree(odl_file);
        free(parameter_buffer);
        return ERROR;
    }

    /* Populate the bridge simplex fit_parameters structure elements */
    for (fit_band_index = 0; fit_band_index < fit_parameters->numbands;
        fit_band_index++)
    {
        /* There is no need to repeat the band check from the otherwise similar
           bridge processing above */ 
        for (table_band_index = 0; table_band_index < number_bands;
            table_band_index++)
        {
            if (fit_parameters->band_numbers[fit_band_index] == 
                band_numbers[table_band_index])
            {
                fit_parameters->splex[fit_band_index].bridge.rad_water
                    = parameter_buffer[table_band_index * MINBRIDGE];
                fit_parameters->splex[fit_band_index].bridge.rad_asym
                    = parameter_buffer[table_band_index * MINBRIDGE + 1];
                fit_parameters->splex[fit_band_index].bridge.gsd
                    = parameter_buffer[table_band_index * MINBRIDGE + 2];

                for (span_index = 0; span_index < MAXSPAN; span_index++)
                {
                    fit_parameters->splex[fit_band_index].bridge
                        .span[span_index].rad_bridge
                        = parameter_buffer[table_band_index * MINBRIDGE
                        + SPAN_OFFSET];
                    fit_parameters->splex[fit_band_index].bridge
                        .span[span_index].width
                        = parameter_buffer[table_band_index * MINBRIDGE
                        + SPAN_OFFSET + 1];
                    fit_parameters->splex[fit_band_index].bridge
                        .span[span_index].offset
                        = parameter_buffer[table_band_index * MINBRIDGE
                        + SPAN_OFFSET + 2];
                }
            }
        }
    }

    /* Get the STF parameters */
    strcpy(field_name, "STF");
    status = ias_odl_get_field(parameter_buffer, NSTF * number_bands 
        * sizeof(double), IAS_ODL_Double, odl_file, NULL, field_name, &count);
    if (status != SUCCESS)
    {
        process_odl_error(status, field_name, system_table_file);
        ias_odl_free_tree(odl_file);
        free(parameter_buffer);
        return ERROR;
    }
    if (count < NSTF * number_bands)
    {
        IAS_LOG_ERROR("Getting %s parameters from %s, %d records were read, "
            "but %d were expected", field_name, system_table_file, count, 
            NSTF * number_bands);
        ias_odl_free_tree(odl_file);
        free(parameter_buffer);
        return ERROR;
    }

    /* Populate the STF fit_parameters structure elements */
    for (fit_band_index = 0; fit_band_index < fit_parameters->numbands;
        fit_band_index++)
    {
        for (table_band_index = 0; table_band_index < number_bands;
            table_band_index++)
        {
            if (fit_parameters->band_numbers[fit_band_index] == 
                band_numbers[table_band_index])
            {
                fit_parameters->model[fit_band_index].stf.g_optics
                    = parameter_buffer[table_band_index * NSTF];
                fit_parameters->model[fit_band_index].stf.xt_detector
                    = parameter_buffer[table_band_index * NSTF + 1];
                fit_parameters->model[fit_band_index].stf.at_detector
                    = parameter_buffer[table_band_index * NSTF + 2];
                fit_parameters->model[fit_band_index].stf.integ
                    = parameter_buffer[table_band_index * NSTF + 3];
                fit_parameters->model[fit_band_index].stf.diffusion
                    = parameter_buffer[table_band_index * NSTF + 4];
                fit_parameters->model[fit_band_index].stf.phase
                    = parameter_buffer[table_band_index * NSTF + 5];

                fit_parameters->model[fit_band_index].nstf = NSTF;
            }
        }
    }

    /* Get the STF simplex parameters */
    strcpy(field_name, "STF_SIMPLEX");
    status = ias_odl_get_field(parameter_buffer, NSTF 
        * number_bands * sizeof(double), IAS_ODL_Double, odl_file, NULL, 
        field_name, &count);
    if (status != SUCCESS)
    {
        process_odl_error(status, field_name, system_table_file);
        ias_odl_free_tree(odl_file);
        free(parameter_buffer);
        return ERROR;
    }
    if (count < NSTF * number_bands)
    {
        IAS_LOG_ERROR("Getting %s parameters from %s, %d records were read, "
            "but %d were expected", field_name, system_table_file, count, 
            NSTF * number_bands);
        ias_odl_free_tree(odl_file);
        free(parameter_buffer);
        return ERROR;
    }

    /* Populate the STF simplex fit_parameters structure elements */
    for (fit_band_index = 0; fit_band_index < fit_parameters->numbands;
        fit_band_index++)
    {
        for (table_band_index = 0; table_band_index < number_bands;
            table_band_index++)
        {
            if (fit_parameters->band_numbers[fit_band_index] == 
                band_numbers[table_band_index])
            {
                fit_parameters->splex[fit_band_index].stf.g_optics
                    = parameter_buffer[table_band_index * NSTF];
                fit_parameters->splex[fit_band_index].stf.xt_detector
                    = parameter_buffer[table_band_index * NSTF + 1];
                fit_parameters->splex[fit_band_index].stf.at_detector
                    = parameter_buffer[table_band_index * NSTF + 2];
                fit_parameters->splex[fit_band_index].stf.integ
                    = parameter_buffer[table_band_index * NSTF + 3];
                fit_parameters->splex[fit_band_index].stf.diffusion
                    = parameter_buffer[table_band_index * NSTF + 4];
                fit_parameters->splex[fit_band_index].stf.phase
                    = parameter_buffer[table_band_index * NSTF + 5];

                fit_parameters->splex[fit_band_index].nstf = NSTF;
            }
        }
    }

    /* Free the memory allocated for the parameter buffer */
    free(parameter_buffer);

    /* Get the number of iterations to perform in the along-track fit */
    strcpy(field_name, "NUMBER_ITERATIONS");
    status = ias_odl_get_field(&fit_parameters->numiter, sizeof(int),
        IAS_ODL_Int, odl_file, NULL, field_name, &count);
    if (status != SUCCESS)
    {
        process_odl_error(status, field_name, system_table_file);
        ias_odl_free_tree(odl_file);
        return ERROR;
    }
    if (count < 1)
    {
        IAS_LOG_ERROR("Getting %s parameter from %s, %d records were read, "
            "but %d were expected", field_name, system_table_file, count, 1);
        ias_odl_free_tree(odl_file);
        return ERROR;
    }

    /* Allocate space for the control structure. */ 
    fit_parameters->control = (MTFCONTROL *)malloc(fit_parameters->numiter 
        * sizeof(MTFCONTROL)); 
    if (fit_parameters->control == NULL)
    { 
        IAS_LOG_ERROR("Allocating fit control parameters");
        ias_odl_free_tree(odl_file);
        return ERROR;
    }

    /* Allocate space for the parameter buffer.  This time it is sized for the 
       fit control parameters */
    mask_parameter_buffer = (int  *)malloc(fit_parameters->numiter * MINBRIDGE 
        * sizeof(int)); 
    if (mask_parameter_buffer == NULL)
    { 
        IAS_LOG_ERROR("Allocating fit control parameter buffer");
        ias_odl_free_tree(odl_file);
        free(fit_parameters->control);         
        return ERROR;
    }

    /* Get the bridge fit control parameters */
    strcpy(field_name, "BRIDGE_MASK");
    status = ias_odl_get_field(mask_parameter_buffer, 
        fit_parameters->numiter * MINBRIDGE * sizeof(int), IAS_ODL_Int,
        odl_file, NULL, field_name, &count);
    if (status != SUCCESS)
    {
        process_odl_error(status, field_name, system_table_file);
        ias_odl_free_tree(odl_file);
        free(fit_parameters->control);         
        free(mask_parameter_buffer);
        return ERROR;
    }
    if (count < fit_parameters->numiter * MINBRIDGE)
    {
        IAS_LOG_ERROR("Getting %s parameter from %s, %d records were read, "
            "but %d were expected", field_name, system_table_file, count, 
            fit_parameters->numiter * MINBRIDGE);
        ias_odl_free_tree(odl_file);
        free(fit_parameters->control);         
        free(mask_parameter_buffer);
        return ERROR;
    }

    /* Populate the fit_parameters bridge mask structure */
    for (iteration_index = 0; iteration_index < fit_parameters->numiter; 
        iteration_index++) 
    {
        fit_parameters->control[iteration_index].bridgemask.rad_water
            = mask_parameter_buffer[iteration_index * MINBRIDGE];
        fit_parameters->control[iteration_index].bridgemask.rad_asym
            = mask_parameter_buffer[iteration_index * MINBRIDGE + 1];
        fit_parameters->control[iteration_index].bridgemask.gsd
            = mask_parameter_buffer[iteration_index * MINBRIDGE + 2];

        /* Replicate the span control parameters for each span */
        for (span_index = 0; span_index < MAXSPAN; span_index++)
        {
            fit_parameters->control[iteration_index].bridgemask
                .span[span_index].rad_bridge
                = mask_parameter_buffer[iteration_index * MINBRIDGE 
                + SPAN_OFFSET]; 
            fit_parameters->control[iteration_index].bridgemask
                .span[span_index].width
                = mask_parameter_buffer[iteration_index * MINBRIDGE 
                + SPAN_OFFSET + 1]; 
            fit_parameters->control[iteration_index].bridgemask
                .span[span_index].offset
                = mask_parameter_buffer[iteration_index * MINBRIDGE 
                + SPAN_OFFSET + 2]; 
        }

        /* The offset for the first span must always be fixed to avoid
           competition with the phase parameter */
        fit_parameters->control[iteration_index].bridgemask.span[0].offset = 0;
    }

    /* Get the STF fit control parameters */
    strcpy(field_name, "STF_MASK");
    status = ias_odl_get_field(mask_parameter_buffer, 
        fit_parameters->numiter * NSTF * sizeof(int), IAS_ODL_Int, odl_file,
        NULL, field_name, &count);
    if (status != SUCCESS)
    {
        process_odl_error(status, field_name, system_table_file);
        ias_odl_free_tree(odl_file);
        free(fit_parameters->control);         
        free(mask_parameter_buffer);
        return ERROR;
    }
    if (count < fit_parameters->numiter * NSTF)
    {
        IAS_LOG_ERROR("Getting %s parameter from %s, %d records were read, "
            "but %d were expected", field_name, system_table_file, count, 
            fit_parameters->numiter * NSTF);
        ias_odl_free_tree(odl_file);
        free(fit_parameters->control);         
        free(mask_parameter_buffer);
        return ERROR;
    }

    /* Populate the appropriate fit_parameters elements */
    for (iteration_index = 0; iteration_index < fit_parameters->numiter; 
        iteration_index++)
    {
        fit_parameters->control[iteration_index].stfmask.g_optics
            = mask_parameter_buffer[iteration_index * NSTF];
        fit_parameters->control[iteration_index].stfmask.xt_detector
            = mask_parameter_buffer[iteration_index * NSTF + 1];
        fit_parameters->control[iteration_index].stfmask.at_detector
            = mask_parameter_buffer[iteration_index * NSTF + 2];
        fit_parameters->control[iteration_index].stfmask.integ
            = mask_parameter_buffer[iteration_index * NSTF + 3];
        fit_parameters->control[iteration_index].stfmask.diffusion
            = mask_parameter_buffer[iteration_index * NSTF + 4];
        fit_parameters->control[iteration_index].stfmask.phase
            = mask_parameter_buffer[iteration_index * NSTF + 5];
    }

    /* Free the parameter buffer */
    free(mask_parameter_buffer);

    /* Free the memory allocated for the ODL object */
    ias_odl_free_tree(odl_file);

    return SUCCESS;
}

