/******************************************************************************
NAME:           minimize_function_multimin

PURPOSE:
Defines the function to minimize and support routines

******************************************************************************/

#include <math.h>                  /* Math library */
#include "gsl/gsl_math.h"          /* various GSL libraries */
#include "gsl/gsl_vector.h"
#include "gsl/gsl_matrix.h"

#include "ias_math.h"              /* IAS math library */
#include "ias_math_complex.h"      /* IAS complex math library */
#include "profile.h"               /* Defines profile structure */
#include "mtfestimate.h"           /* MTF prototypes */
#include "profile.h"               /* Defines profile structure */
#include "model.h"                 /* Defines model and control structures */
#include "minimize_function_multimin.h" /* Defines function to minimize */

static IAS_COMPLEX *complex_buffer; /* Complex variable buffer */

/******************************************************************************
NAME:          initialize_minimize_function_multimin 

PURPOSE:
Initializes the function to minimize

RETURN VALUE:
Type = int
Value           Description
-----           -----------
SUCCESS         Initialization was successful
ERROR           Initialization failed

******************************************************************************/
int initialize_minimize_function_multimin
(
    int number_points              /* Factor in size of buffer to initialize */
)
{
    /* Allocate the complex variable buffer needed for the annealing 
       procedure */
    complex_buffer = (IAS_COMPLEX *)malloc(3 * number_points 
        * sizeof(IAS_COMPLEX));
    if (complex_buffer == NULL)
    {
        IAS_LOG_ERROR("Allocating temporary complex data array");
        return ERROR;
    }
    return SUCCESS;
}

/******************************************************************************
NAME:          close_minimize_function_multimin 

PURPOSE:
Closes the function to minimize

RETURN VALUE:
None
******************************************************************************/
void close_minimize_function_multimin
(
) 
{ 
    free(complex_buffer);
}

/******************************************************************************
NAME:           minimize_function_multimin

PURPOSE:
Defines the function to minimize

RETURN VALUE:
Type = double
Value = result returned by function

******************************************************************************/
double minimize_function_multimin
(
    const gsl_vector *x,           /* I: Vector of x coordinates */
    void *parameters               /* I: Functional data structure */
)
{
    BRIDGE_PARMS bridge;           /* Local bridge model parameters */
    STF_FD_PARMS stf;              /* Local STF model parameters */
    IAS_COMPLEX *frequency_stf;    /* Frequency domain STF model */
    IAS_COMPLEX *frequency_bridge; /* Frequency domain bridge model */
    IAS_COMPLEX frequency_phase;   /* Scan direction specific FD phase */
    IAS_COMPLEX temp;              /* Temporary variable */
    IAS_COMPLEX *target_model;     /* Composite target model (FD and SD) */
    double model_profile;          /* Model simulated data profile value */
    double spatial_frequency;      /* Spatial frequency variable */
    double rms;                    /* RMS model error */
    int number_fit_points;         /* Number of points to fit */
    int point_index;               /* Loop counter */
    int scaled_point_index;        /* Point index scaled by number_points */
    int parameter_index;           /* Loop counter */
    int fft_index;                 /* IFFT index */
    int weight;                    /* Static weight */
    int number_points;             /* Number of points */
    MTFMODEL *input_model;         /* Input bridge/STF model structure */
    MTFCONTROL *control;           /* Array of fit control switches */
    PROFILE *profile;              /* Input target profile data */
    MTFMODEL *output_model;        /* Best-fit output model structure */
    double phase_wrap;             /* Maximum absolute phase value */
    int number_parameters;         /* Number of bridge and STF parameters */
    int bridge_mask_array[NBRIDGE]; /* Bridge mask */
    int stf_mask_array[NSTF];      /* STF mask */
    double bridge_parameter_array[NBRIDGE]; /* Bridge parameters */
    double stf_parameter_array[NSTF]; /* STF parameters */
    double output_bridge_parameter_array[NBRIDGE]; /* Bridge parameters */
    double output_stf_parameter_array[NSTF]; /* STF parameters */
    struct function_parameters_multimin *f_params; /* Function parameters */
    double two_pi;                 /* Static constant */

    two_pi = 2.0 * ias_math_get_pi();

    /* Extract function parameters */
    f_params = (struct function_parameters_multimin *)parameters;
    input_model = (f_params->input_model);
    control = (f_params->control);
    profile = (f_params->profile);
    output_model = (f_params->output_model);

    /* Initialize buffers */
    number_points = profile->number_points;
    frequency_stf = &complex_buffer[0];
    frequency_bridge = &complex_buffer[number_points];
    target_model = &complex_buffer[2 * number_points];

    /* Initialize variables */
    rms = 0.0;
    number_fit_points = 0;

    phase_wrap = (double)number_points * output_model->bridge.gsd 
        / profile->oversample / 2.0;

    /* Load the variable and static parameters.  First load the bridge
       parameters */
    number_parameters = 0;
    convert_bridge_mask_to_array(&control->bridgemask, bridge_mask_array);
    convert_bridge_to_array(&output_model->bridge, 
        output_bridge_parameter_array);
    for (parameter_index = 0; parameter_index < output_model->nbridge; 
        parameter_index++)
    {
        if (bridge_mask_array[parameter_index] > 0)
        {
            bridge_parameter_array[parameter_index] 
                = gsl_vector_get(x, number_parameters++);
        }
        else
        {
            bridge_parameter_array[parameter_index] 
                = output_bridge_parameter_array[parameter_index];
        }
    }
    convert_bridge_to_structure(bridge_parameter_array, &bridge);

    /* Then load the STF parameters */
    convert_stf_mask_to_array(&control->stfmask, stf_mask_array);
    for (parameter_index = 0; parameter_index < NSTF; parameter_index++)
    {
        /* Make sure phase parameter stays in range */
        if (parameter_index == PHASE && stf_mask_array[parameter_index] > 0)
        {
            while (output_model->stf.phase < -phase_wrap) 
            {
                output_model->stf.phase += 2.0 * phase_wrap;
            }
            while (output_model->stf.phase > phase_wrap) 
            {
                output_model->stf.phase -= 2.0 * phase_wrap;
            }
        }
    }

    convert_stf_to_array(&output_model->stf, output_stf_parameter_array);
    for (parameter_index = 0; parameter_index < NSTF; parameter_index++)
    {
        /* Load the STF parameters */
        if (stf_mask_array[parameter_index] > 0)
        {
            stf_parameter_array[parameter_index] 
                = gsl_vector_get(x, number_parameters++);
        }
        else
        {
            stf_parameter_array[parameter_index] 
                = output_stf_parameter_array[parameter_index];
        }
    }
    convert_stf_to_structure(stf_parameter_array, &stf);

    /* Generate the frequency domain bridge model */
    create_frequency_domain_bridge_model(input_model->nspan, &bridge, 
        number_points, profile->oversample, frequency_bridge);

    /* Generate the frequency domain STF model */
    create_frequency_domain_stf_model(&stf, bridge.gsd, profile->range, 
        profile->velocity, profile->angle, number_points, profile->oversample, 
        frequency_stf);

    /* Compute the scan phase and construct the target model */
    for (point_index = 0; point_index < number_points; point_index++)
    {
        /* Compute the frequency for this point */
        if (point_index >= (number_points / 2))
        {
            scaled_point_index = point_index - number_points;
        }
        else
        {
            scaled_point_index = point_index;
        }
        spatial_frequency = profile->oversample * two_pi 
            * (double) (scaled_point_index) / (double)number_points 
            / bridge.gsd;

        /* Compute the phase shift for this point */
        frequency_phase.re = 0.0;
        frequency_phase.im = -spatial_frequency * stf.phase;
        ias_math_complex_exponential(&frequency_phase, &frequency_phase);

        /* Compute the composite frequency domain model for this point */
        ias_math_complex_multiply(&frequency_stf[point_index], &frequency_phase,
            &temp);
        ias_math_complex_multiply(&frequency_bridge[point_index], &temp, 
            &target_model[point_index]);
    }

    /* Use the inverse FFT to convert the model to the space domain */
    ias_math_fft2d((double *)&target_model[0].re, number_points, 1,
        IAS_MATH_IFFT);

    /* Take the real part and scale to convert to a real output */
    for (point_index = 0; point_index < number_points; point_index++)
    {
        fft_index = (point_index + number_points / 2) % number_points;
        model_profile = target_model[fft_index].re / (double)number_points;
        if (point_index > 9 && point_index < number_points - 10)
        {
            if (profile->sigma[point_index] > 0.0)
            {
                weight = 1;
            }
            else
            {
                weight = 0;
            }
            rms = rms + (model_profile - profile->sample_data[point_index])
                * (model_profile - profile->sample_data[point_index]) * weight;
            number_fit_points += weight;
        }
    }

    /* Compute the final RMS error */
    rms = sqrt(rms / (double)number_fit_points);

    return rms;
}

