/******************************************************************************

NAME:           mtfestimate

PURPOSE:
Main procedure for software to estimate the OLI along and across track
STF/MTF model parameters by best-fitting an analytical model of the STF
and bridge target to the measured image profile.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
EXIT_FAILURE    Error estimating STF/MTF 
EXIT_SUCCESS    Successfully generated STF/MTF estimate

******************************************************************************/
#include <stdlib.h>                   /* EXIT_FAILURE/EXIT_SUCCESS */
#include <limits.h>                   /* PATH_MAX */
#include <string.h>                   /* strncpy */

#include "ias_logging.h"              /* IAS logging library */
#include "ias_miscellaneous.h"        /* For report header */
#include "get_targets.h"              /* MTF targets structure and prototype */ 
#include "model.h"                    /* MTF model */
#include "profile.h"                  /* MTF profile structure */
#include "mtfestimate.h"              /* MTF function prototypes */

int main(int argc, char *argv[])
{
    char l0r_file[PATH_MAX];          /* L0R file name */
    char l1g_file[PATH_MAX];          /* L1G file name */
    char target_file[PATH_MAX];       /* Input target description file name */
    char input_directory[PATH_MAX];   /* Input data directory name */
    char work_order_id[IAS_WORKORDER_ID_SIZE]; /* Work order ID */

    int band_number;                  /* Band number to process */
    int band_index;                   /* Index into fit_parameters to find
                                         band_number */
    const char *band_name;            /* Descriptive band name */
    int number_targets;               /* Number of targets in image */
    TARGET targets[MAX_TARGETS];      /* Targets in image */
    int target_id;                    /* Id and index of current target */
    int target_index;                 /* Loop counter */
    int stf_fit_method;               /* Use either Annealing (0) or MultiMin
                                         (1) */
    MTFEST fit_parameters;            /* Structure containing parameters for
                                         this estimatation session */
    int *solution_iterations;         /* Number of iterations performed */
    PROFILE profile;                  /* Measured target profile data */
    MTFMODEL start_model;             /* Initial model values */
    MTFMODEL simplex_model;           /* Simplex values */
    MTFMODEL new_model;               /* Best-fit STF model parameters */
    IAS_REPORT_HEADER *header = NULL; /* Report header structure */
    FILE *report_file_pointer;        /* Pointer to output report file */
    const char *report_file = "mtfestimate.rpt";  /* Output report file name */
    FILE *trend_file_pointer;         /* Pointer to output trending file */
    const char *trend_file = "mtfestimate_trend.dat";  /* Trending output file
                                         name */
    double rms;                       /* Best-fit RMS error */
    int store_in_database_flag;       /* Flag for storing results in database */
    int status;                       /* Function return status */

    /* Initialize the logging library */
    if (ias_log_initialize("mtfestimate") != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing logging library");
        return EXIT_FAILURE;
    }

    /* Verify a parameter file was provided */
    if (argc != 2)
    {
        IAS_LOG_ERROR("USAGE: mtfestimate <parameter file name>");
        return EXIT_FAILURE;
    }

    /* Get work order parameters */
    if (get_parameters(argv[1], target_file, input_directory, l0r_file, 
        l1g_file, work_order_id, &stf_fit_method, &fit_parameters,
        &store_in_database_flag) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from file: %s", argv[1]);
        return EXIT_FAILURE;
    }

    IAS_LOG_INFO("Initiating mtfestimate");

    /* Get system table parameters */
    if (get_system_table(input_directory, &fit_parameters) != SUCCESS) 
    {
        IAS_LOG_ERROR("Reading parameters from system table: %s/%s", 
            input_directory, MTF_SYSTEM_FILE);
        return EXIT_FAILURE;
    }

    /* Read the target definition file */
    number_targets = get_targets(target_file, targets);
    if (number_targets <= 0 || number_targets > MAX_TARGETS)
    {
        IAS_LOG_ERROR("Reading target description file %s; number of targets "
            "is %d but should be between 1 and %d", target_file, number_targets,
             MAX_TARGETS);
        return EXIT_FAILURE;
    }

    /* Allocate memory for the number solved per iteration */
    solution_iterations = (int *)malloc(sizeof(int) * fit_parameters.numiter);
    if (solution_iterations == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for number solved per iteration");
        for (target_id = 0; target_id < number_targets; target_id++)
        {
            free(targets[target_id].span);
        }
        return EXIT_FAILURE;
    }

    /* Open the output report file */
    if ((report_file_pointer = fopen(report_file, "w")) == NULL)
    {
        IAS_LOG_ERROR("Opening output report file %s", report_file);
        free(solution_iterations);
        for (target_id = 0; target_id < number_targets; target_id++)
        {
            free(targets[target_id].span);
        }
        return EXIT_FAILURE;
    }

    /* Initialize the report header and write it to the report file */
    header = ias_misc_initialize_gps_report_header(l1g_file, l0r_file,
            work_order_id, "mtfestimate");
    if (!header)
    {
        IAS_LOG_ERROR("Initializing the GPS report file header");
        fclose(report_file_pointer);
        free(solution_iterations);
        for (target_id = 0; target_id < number_targets; target_id++)
        {
            free(targets[target_id].span);
        }
        return EXIT_FAILURE;
    }
    status = ias_misc_write_gps_report_header(report_file_pointer, header);
    ias_misc_free_report_header(header);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the GPS report file header");
        fclose(report_file_pointer);
        free(solution_iterations);
        for (target_id = 0; target_id < number_targets; target_id++)
        {
            free(targets[target_id].span);
        }
        return EXIT_FAILURE;
    }

    /* Open the output trend file */
    if ((trend_file_pointer = fopen(trend_file, "w")) == NULL)
    {
        IAS_LOG_ERROR("Opening output trending file %s", trend_file);
        fclose(report_file_pointer);
        free(solution_iterations);
        for (target_id = 0; target_id < number_targets; target_id++)
        {
            free(targets[target_id].span);
        }
        return EXIT_FAILURE;
    }

    /* Connect to the database */
    if (store_in_database_flag)
    {
        status = get_database_connection();
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Connecting to the database");
            fclose(report_file_pointer);
            fclose(trend_file_pointer);
            free(solution_iterations);
            for (target_id = 0; target_id < number_targets; target_id++)
            {
                free(targets[target_id].span);
            }
            return EXIT_FAILURE;
        }
    }

    /* Loop through the targets */
    for (target_id = 0; target_id < number_targets; target_id++)
    {
        IAS_LOG_INFO("Begin processing target %s...", 
            targets[target_id].target_name);

        /* Loop through the bands */
        for (band_index = 0; band_index < fit_parameters.numbands;
            band_index++)
        {
            band_number = fit_parameters.band_numbers[band_index];
            band_name = ias_sat_attr_convert_band_number_to_name(band_number);
            if (band_name == NULL)
            {
                IAS_LOG_ERROR("No name was found for band number %d", 
                    band_number);
                fclose(report_file_pointer);
                fclose(trend_file_pointer);
                free(solution_iterations);
                for (target_index = target_id; target_index < number_targets; 
                    target_index++)
                {
                    free(targets[target_index].span);
                }
                disconnect_with_transaction(IAS_DB_ROLLBACK);
                return EXIT_FAILURE;
            }

            IAS_LOG_INFO("Begin processing band %s (%d,%d)...",
                band_name, band_index, band_number);

            /* Process the cross-target profile */
            if (read_profile(target_id, band_number, &profile) != SUCCESS)
            {
                IAS_LOG_ERROR("Reading data profile for target %d and band "
                    "number %d", target_id, band_number);
                fclose(report_file_pointer);
                fclose(trend_file_pointer);
                free(solution_iterations);
                for (target_index = target_id; target_index < number_targets; 
                    target_index++)
                {
                    free(targets[target_index].span);
                }
                disconnect_with_transaction(IAS_DB_ROLLBACK);
                return EXIT_FAILURE;
            }

            /* Set up the fit */
            if (initialize_fit(band_index, &fit_parameters, &targets[target_id],
                &profile, &start_model, &simplex_model) != SUCCESS)
            {
                IAS_LOG_ERROR("Setting up the fit for target %d and band "
                    "number %d", target_id, band_number);
                fclose(report_file_pointer);
                fclose(trend_file_pointer);
                free_profile(&profile);
                free(solution_iterations);
                for (target_index = target_id; target_index < number_targets; 
                    target_index++)
                {
                    free(targets[target_index].span);
                }
                disconnect_with_transaction(IAS_DB_ROLLBACK);
                return EXIT_FAILURE;
            }

            /* Do the fit */
            IAS_LOG_INFO("Begin the STF fit...");
            if (stf_fit_method == 0)
            {
                status = fit_annealing(&start_model, &simplex_model, 
                    fit_parameters.numiter, fit_parameters.control, &profile,
                    &rms, &new_model, solution_iterations);
            }
            else
            {
                status = fit_multimin(&start_model, &simplex_model,
                    fit_parameters.numiter, fit_parameters.control, &profile,
                    &rms, &new_model, solution_iterations);
            }

            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Fitting STF");
                fclose(report_file_pointer);
                fclose(trend_file_pointer);
                free_profile(&profile);
                free(solution_iterations);
                for (target_index = target_id; target_index < number_targets; 
                    target_index++)
                {
                    free(targets[target_index].span);
                }
                disconnect_with_transaction(IAS_DB_ROLLBACK);
                return EXIT_FAILURE;
            }

            /* Report the results for this target and band */
            status = report_fit(report_file_pointer, target_id, 
                targets[target_id], profile, new_model, rms, 
                fit_parameters.year_doy, fit_parameters.path_row, 
                fit_parameters.numiter, solution_iterations);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Generating fit report");
                fclose(report_file_pointer);
                fclose(trend_file_pointer);
                free_profile(&profile);
                free(solution_iterations);
                for (target_index = target_id; target_index < number_targets; 
                    target_index++)
                {
                    free(targets[target_index].span);
                }
                disconnect_with_transaction(IAS_DB_ROLLBACK);
                return EXIT_FAILURE;
            }

            /* Send the trending data to a file */
            status = trend_fit(trend_file_pointer, targets[target_id], profile,
                new_model, rms, fit_parameters.year_doy,
                fit_parameters.path_row);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Sending fit results to file");
                fclose(report_file_pointer);
                fclose(trend_file_pointer);
                free_profile(&profile);
                free(solution_iterations);
                for (target_index = target_id; target_index < number_targets; 
                    target_index++)
                {
                    free(targets[target_index].span);
                }
                disconnect_with_transaction(IAS_DB_ROLLBACK);
                return EXIT_FAILURE;
            }


            /* Send the trending data to the database */
            if (store_in_database_flag)
            {
                if (save_to_database(targets[target_id], profile, new_model,
                    rms, fit_parameters.year_doy, fit_parameters.path_row, 
                    work_order_id) != SUCCESS)
                {
                    IAS_LOG_ERROR("Sending fit results to database");
                    fclose(report_file_pointer);
                    fclose(trend_file_pointer);
                    free_profile(&profile);
                    free(solution_iterations);
                    for (target_index = target_id; target_index < 
                        number_targets; target_index++)
                    {
                        free(targets[target_index].span);
                    }
                    disconnect_with_transaction(IAS_DB_ROLLBACK);
                    return EXIT_FAILURE;
                }
            }

            /* Free the memory for the current profile */
            free_profile(&profile);
        } /* Band loop */

        /* Release this target structure's allocated memory */
        free(targets[target_id].span);
        targets[target_id].span = NULL;
    } /* Target loop */

    /* Commit the database transaction */
    if (store_in_database_flag)
    {
        status = disconnect_with_transaction(IAS_DB_COMMIT);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Committing transaction and disconnecting from "
                "database");
            fclose(trend_file_pointer);
            free(fit_parameters.control);
            free(solution_iterations);
            return EXIT_FAILURE;
        }
    }

    /* Close the report output file */
    status = fclose(report_file_pointer);
    if (status != 0)
    {
        IAS_LOG_ERROR("Closing report file %s", report_file);
        fclose(trend_file_pointer);
        free(fit_parameters.control);
        free(solution_iterations);
        return EXIT_FAILURE;
    }

    /* Close the trending output file */
    status = fclose(trend_file_pointer);
    if (status != 0)
    {
        IAS_LOG_ERROR("Closing trending file %s", trend_file);
        free(fit_parameters.control);
        free(solution_iterations);
        return EXIT_FAILURE;
    }

    /* Free memory */
    free(fit_parameters.control);
    fit_parameters.control = NULL;
    free(solution_iterations);
    solution_iterations = NULL;

    /* Write completion message */
    IAS_LOG_INFO("Successful completion.");

    return EXIT_SUCCESS;
}

