#ifndef _MTFESTIMATE_H
#define _MTFESTIMATE_H

#include "model.h"
#include "profile.h"
#include "get_targets.h"
#include "ias_satellite_attributes.h"
#include "ias_math.h"

#define MTF_SYSTEM_FILE "mtfsystem.table"  /* MTF system table file name */


/* Function prototypes */

int get_parameters
(
    const char *parm_file,         /* I: Name of the parameter file */
    char *target_file,             /* O: Input target description file name */
    char *input_directory,         /* O: MTF input data directory */
    char *l0r_file,                /* O: L0R file name */
    char *l1g_file,                /* O: L1G file name */
    char *work_order_id,           /* O: Work order ID string */
    int *stf_fit_method,           /* O: Method to use (Annealing or MultiMin)
                                         for multi-parameter fitting */
    MTFEST *fit_parameters,        /* O: Structure containing fit parameters */
    int *store_in_database_flag    /* O: Flag for storing results in the DB */
);

int get_maximum_iterations
(
);

int get_system_table
(
    const char *input_directory,   /* I: Directory for system table */
    MTFEST *fit_parameters         /* O: Structure containing fit parameters */
);

int read_profile
(
    int target_id,                 /* I: Target id index */
    int band_number,               /* I: Band number */
    PROFILE *profile               /* O: Populated profile structure */
);

void free_profile
(
    PROFILE *profile              /* I/O: Profile structure to release */
);

int initialize_fit
(
    int band_index,                /* I: Band index */
    const MTFEST *fit_parameters,  /* I: Initial model parameters */
    const TARGET *target,          /* I: Bridge target parameters */
    const PROFILE *profile,        /* I: Input target profile data */
    MTFMODEL *initial_model,       /* O: Bridge/STF model structure */
    MTFMODEL *simplex_model        /* O: Bridge/STF simplex bounds */
);

int fit_annealing
(
    const MTFMODEL *initial_model, /* I: Input bridge/STF model structure */
    const MTFMODEL *simplex_model, /* I: Input bridge/STF simplex bounds */
    int number_iterations,         /* I: Number of fit iterations */
    const MTFCONTROL *control,     /* I: Fit control switches */
    const PROFILE *profile,        /* I: Input target profile data */
    double *final_rms,             /* O: Best fit RMS error */
    MTFMODEL *output_model,        /* O: Best-fit output model structure */
    int *solution_iterations       /* O: Number of iterations performed for
                                         solution */
);

int fit_multimin
(
    MTFMODEL *input_model,         /* I/O: Input bridge/STF model structure */
    MTFMODEL *simplex_model,       /* I: Input bridge/STF simplex bounds */
    int number_iterations,         /* I: Number of fit iterations */
    MTFCONTROL *control,           /* I: Array of fit control switches */
    PROFILE *profile,              /* I: Input target profile data */
    double *final_rms,             /* O: Best fit RMS error */
    MTFMODEL *output_model,        /* O: Best-fit output model structure */
    int *solution_iterations       /* O: Number of iterations performed for
                                         solution */
);

int report_fit
(
    FILE *report_file_pointer,     /* I: Output report file pointer */
    int target_index,              /* I: Current target index */
    TARGET target,                 /* I: Current target structure */
    PROFILE profile,               /* I: Current profile structure */
    MTFMODEL model,                /* I: Best-fit model parameters */
    double rms,                    /* I: Best-fit RMS error */
    int *year_doy,                 /* I: Scene year and day of year */
    int *path_row,                 /* I: Scene path and row */
    int number_iterations,         /* I: Number of times simplex is called */
    int *solution_iterations       /* I: Number of iterations for each call to
                                         simplex */
);

int write_profile
(
    int target_id,                 /* I: Current target index */
    int band_number,               /* I: Data band number */
    int number_points,             /* I: Number of points in profile */
    double *model_profile          /* I: Model profile data */
);

int trend_fit
(
    FILE *trend_file_pointer,      /* I: Output trend file pointer */
    TARGET target,                 /* I: Current target structure */
    PROFILE profile,               /* I: Current profile structure */
    MTFMODEL model,                /* I: Best-fit model parameters */
    double rms,                    /* I: Best-fit RMS error */
    int *year_doy,                 /* I: Scene year and day of year */
    int *path_row                  /* I: Scene path and row */
);

int get_database_connection
(
    void
);

int disconnect_with_transaction
(
   int commit_flag                 /* I: Flag showing if commit should be done
                                         1 = IAS_DB_COMMIT
                                         0 = IAS_DB_ROLLBACK */
);

int save_to_database
(
    TARGET target,                 /* I: Current target structure */
    PROFILE profile,               /* I: Current profile structure */
    MTFMODEL model,                /* I: Best-fit model parameters */
    double rms,                    /* I: Best-fit RMS error */
    int *year_doy,                 /* I: Scene year and day of year */
    int *path_row,                 /* I: Scene path and row */
    const char *work_order_id      /* I: Work order ID */
);

int get_acquisition_date
(
    const char *l0r_file_name,     /* I: L0R file name */
    int *year,                     /* O: Acquistion year */
    int *day_of_year               /* O: Acquistion day of year */
);

void convert_bridge_to_structure
(
    const double* bridge_array,    /* I: Bridge parameters to convert */
    BRIDGE_PARMS *bridge           /* O: Destination for conversion */
);

void convert_stf_to_structure
(
    const double* stf_array,       /* I: STF to convert */
    STF_FD_PARMS *stf              /* O: Destination for conversion */
);

void convert_bridge_to_array
(
    const BRIDGE_PARMS *bridge,    /* I: Bridge parameters to convert */
    double* bridge_array           /* O: Destination for conversion */
);

void convert_stf_to_array
(
    const STF_FD_PARMS *stf,       /* I: STF to convert */
    double* stf_array              /* O: Destination for conversion */
);

void convert_bridge_mask_to_array
(
    const BRIDGE_PARMS_MASK *bridgemask, /* I: Bridge mask to convert */
    int* bridgemask_array          /* O: Destination for conversion */
);

void convert_stf_mask_to_array
(
    const STF_FD_PARMS_MASK *stfmask, /* I: STF mask to convert */
    int* stfmask_array             /* O: Destination for conversion */
);

#endif
