/******************************************************************************
NAME:           read_profile

PURPOSE:
Read the interleaved profiles from an ODL-formatted text file

RETURN VALUE:
Type = int
Value           Description
-----           -----------
SUCCESS         Successfully read profiles 
ERROR           Error reading profiles 

******************************************************************************/
#include <limits.h>                /* PATH_MAX */
#include <stdlib.h>                /* malloc */

#include "profile.h"               /* MTF profile structure */
#include "ias_logging.h"           /* IAS logging library */
#include "ias_odl.h"               /* IAS ODL library */


int read_profile 
(
    int target_id,                 /* I: Target id index */
    int band_number,               /* I: Band number */
    PROFILE *profile               /* O: Populated profile structure */
)
{
    char profile_file[PATH_MAX];   /* Constructed profile file name */
    int size;                      /* Size returned by snprintf */
    IAS_OBJ_DESC *odl_file;        /* ODL tree structure */
    int count;                     /* Number of elements */
    int status;                    /* Function status return value */

    /* Construct profile file name for this target and band number */
    size = snprintf(profile_file, sizeof(profile_file), 
        "Profile_Tar%dBand%d.odl", target_id, band_number);
    if ((size < 0) || (size > sizeof(profile_file)))
    {
        IAS_LOG_ERROR("Profile file name length is %d; it should be between "
            "0 and %lu", size, sizeof(profile_file));
    }

    /* Read ODL-format system table file */
    odl_file = ias_odl_read_tree(profile_file);
    if (odl_file == NULL)
    {
        IAS_LOG_ERROR("Reading parameters from %s", profile_file);
        return ERROR;
    }

    /* Get the target name */
    status = ias_odl_get_field(profile->target_name, NAME_STRLEN * sizeof(char),
        IAS_ODL_String, odl_file, NULL, "TARGET_NAME", &count);
    if (status != SUCCESS)
    {
        process_odl_error(status, "TARGET_NAME", profile_file);
        ias_odl_free_tree(odl_file);
        return ERROR;
    }

    /* Get the target ID */
    status = ias_odl_get_field(&profile->target_id, sizeof(int), IAS_ODL_Int,
        odl_file, NULL, "TARGET_ID", &count);
    if (status != SUCCESS)
    {
        process_odl_error(status, "TARGET_ID", profile_file);
        ias_odl_free_tree(odl_file);
        return ERROR;
    }

    /* Verify the target ID */
    if (profile->target_id != target_id)
    {
        IAS_LOG_ERROR("Target id %d from parameter file %s should match "
            "target id %d", profile->target_id, profile_file, target_id);
        ias_odl_free_tree(odl_file);
        return ERROR;
    }

    /* Read the band number */
    status = ias_odl_get_field(&profile->band_number, sizeof(int), IAS_ODL_Int,
        odl_file, NULL, "BAND", &count);
    if (status != SUCCESS)
    {
        process_odl_error(status, "BAND", profile_file);
        ias_odl_free_tree(odl_file);
        return ERROR;
    }

    /* Verify the band number */
    if (profile->band_number != band_number)
    {
        IAS_LOG_ERROR("Band number %d from parameter file %s should match "
            "band number %d", profile->band_number, profile_file, band_number);
        ias_odl_free_tree(odl_file);
        return ERROR;
    }

    /* Read the number of samples */
    status = ias_odl_get_field(&profile->number_points, sizeof(int), 
        IAS_ODL_Int, odl_file, NULL, "NSAMP", &count);
    if (status != SUCCESS)
    {
        process_odl_error(status, "NSAMP", profile_file);
        ias_odl_free_tree(odl_file);
        return ERROR;
    }

    /* Read the oversample factor */
    status = ias_odl_get_field(&profile->oversample, sizeof(int), 
        IAS_ODL_Int, odl_file, NULL, "OVERSAMP", &count);
    if (status != SUCCESS)
    {
        process_odl_error(status, "OVERSAMP", profile_file);
        ias_odl_free_tree(odl_file);
        return ERROR;
    }

    /* Read the ground sampling distance (GSD) */
    status = ias_odl_get_field(&profile->gsd, sizeof(double), IAS_ODL_Double,
        odl_file, NULL, "GSD", &count);
    if (status != SUCCESS)
    {
        process_odl_error(status, "GSD", profile_file);
        ias_odl_free_tree(odl_file);
        return ERROR;
    }

    /* Read the target azimuth */
    status = ias_odl_get_field(&profile->angle, sizeof(double), IAS_ODL_Double,
        odl_file, NULL, "TARGET_AZIMUTH", &count);
    if (status != SUCCESS)
    {
        process_odl_error(status, "TARGET_AZIMUTH", profile_file);
        ias_odl_free_tree(odl_file);
        return ERROR;
    }

    /* Read the target range */
    status = ias_odl_get_field(&profile->range, sizeof(double), IAS_ODL_Double,
        odl_file, NULL, "TARGET_RANGE", &count);
    if (status != SUCCESS)
    {
        process_odl_error(status, "TARGET_RANGE", profile_file);
        ias_odl_free_tree(odl_file);
        return ERROR;
    }

    /* Read ground velocity */
    status = ias_odl_get_field(&profile->velocity, sizeof(double), 
        IAS_ODL_Double, odl_file, NULL, "GROUND_VELOCITY", &count);
    if (status != SUCCESS)
    {
        process_odl_error(status, "GROUND_VELOCITY", profile_file);
        ias_odl_free_tree(odl_file);
        return ERROR;
    }

    /* Allocate the sample data array */
    profile->sample_data = (double *)malloc(profile->number_points 
        * sizeof(double));
    if (profile->sample_data == NULL)
    {
        IAS_LOG_ERROR("Allocating data array for target %d", 
            profile->target_id);
        ias_odl_free_tree(odl_file);
        return ERROR;
    }

    /* Get the profile data values */
    status = ias_odl_get_field(profile->sample_data, profile->number_points 
        * sizeof(double), IAS_ODL_Double, odl_file, NULL, "PROFILE", &count);
    if (status != SUCCESS)
    {
        process_odl_error(status, "PROFILE", profile_file);
        ias_odl_free_tree(odl_file);
        free(profile->sample_data);
        return ERROR;
    }

    /* Allocate the sigma (standard deviation) array */
    profile->sigma = (double *)malloc(profile->number_points * sizeof(double));
    if (profile->sigma == NULL)
    {
        IAS_LOG_ERROR("Allocating standard deviation array for target %d", 
            profile->target_id);
        ias_odl_free_tree(odl_file);
        free(profile->sample_data);
        return ERROR;
    }

    /* Get the profile data values */
    status = ias_odl_get_field(profile->sigma, profile->number_points 
        * sizeof(double), IAS_ODL_Double, odl_file, NULL, "SIGMAS", &count);
    if (status != SUCCESS)
    {
        process_odl_error(status, "SIGMAS", profile_file);
        ias_odl_free_tree(odl_file);
        free(profile->sample_data);
        free(profile->sigma);
        return ERROR;
    }

    /* Free the memory allocated to the ODL object */
    ias_odl_free_tree(odl_file);

    return SUCCESS;
}

/******************************************************************************
NAME:           free_profile

PURPOSE:
Free the memory allocated for profile sample and standard deviation data

RETURN VALUE:
None
******************************************************************************/
void free_profile
(
    PROFILE *profile
)
{
    free(profile->sample_data);
    profile->sample_data = NULL;
    free(profile->sigma);
    profile->sigma = NULL;
}
