/******************************************************************************
NAME:           report_fit

PURPOSE:
Genarate output report for the MTF estimation results.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

******************************************************************************/

#include "ias_logging.h"           /* IAS logging library */
#include "mtfestimate.h"           /* MTF prototypes */
#include "get_targets.h"           /* Defines target structure */
#include "model.h"                 /* Defines model and control structures */
#include "profile.h"               /* Defines profile structure */

int report_fit 
(
    FILE *report_file_pointer,     /* I: Output report file pointer */
    int target_index,              /* I: Current target index */
    TARGET target,                 /* I: Current target structure */
    PROFILE profile,               /* I: Current profile structure */
    MTFMODEL model,                /* I: Best-fit model parameters */
    double rms,                    /* I: Best-fit RMS error */
    int *year_doy,                 /* I: Scene year and day of year */
    int *path_row,                 /* I: Scene path and row */
    int number_iterations,         /* I: Number of times simplex is called */
    int *solution_iterations       /* I: Number of iterations for each call to
                                         simplex */
)
{
    double *model_profile;         /* Model bridge profile */
    int point_index;               /* Loop counter */
    int span_index;                /* Loop counter */
    int iteration_index;           /* Loop counter */
    int status;                    /* Function return status */

    /* Allocate local buffer */
    model_profile = (double *)malloc(profile.number_points * sizeof(double));
    if (model_profile == NULL)
    {
        IAS_LOG_ERROR("Allocating output model buffer");
        return ERROR;
    }

    /* Initialize buffer */
    for (point_index = 0; point_index < profile.number_points; point_index++)
    {
        model_profile[point_index] = 0.0;
    }

    /* Generate the model data profiles */
    status = generate_spatial_domain_target_profile(profile.number_points, 
        profile.oversample, model, profile.range, profile.velocity, 
        profile.angle, model_profile);
    if (status != SUCCESS)
    {
        free(model_profile);
        IAS_LOG_ERROR("Synthesizing output profile");
        return ERROR;
    }

    /* Write the fitted profile */
    status = write_profile(target_index, profile.band_number, 
        profile.number_points, model_profile);
    if (status != SUCCESS)
    {
        free(model_profile);
        IAS_LOG_ERROR("Writing model profile");
        return ERROR;
    }

    /* Free the model profile structure */
    free(model_profile);
    model_profile = NULL;

    /* Write the across-track results to the text report file */
    if (fprintf(report_file_pointer, "\n\n"
        "Test Scene Acquisition Year-DOY: "
        "%04d-%03d and Path/Row: %03d/%03d\n\n\n"
        "MTF Results for band %d\n\n"
        "Geometric Parameters:\n"
        "    GSD = %6.3lf meters\n"
        "    Target Range = %10.3lf meters\n"
        "    Target Velocity = %8.3lf meters/second\n"
        "    Target Orientation = %9.6lf radians\n\n"
        "Best Fit Target Parameters:\n"
        "    Target Name = %s\n"
        "    Number of Spans = %d\n",
        year_doy[0], year_doy[1], path_row[0], path_row[1], profile.band_number,
        model.bridge.gsd, profile.range, profile.velocity, profile.angle,
        target.target_name, target.number_of_spans) < 0)
    {
        IAS_LOG_ERROR("Writing to report file");
        return ERROR;
    }

    for (span_index = 0; span_index < target.number_of_spans; span_index++)
    {
        if (fprintf(report_file_pointer, 
            "    Span Radiance = %4.1lf\n"
            "    Span Width = %5.2lf meters\n"
            "    Span Offset = %6.2lf meters\n",
            model.bridge.span[span_index].rad_bridge,
            model.bridge.span[span_index].width, 
            model.bridge.span[span_index].offset) < 0)
        {
            IAS_LOG_ERROR("Writing span index %d information to report file",
                span_index);
            return ERROR;
        }
    }

    if (fprintf(report_file_pointer, 
        "    Water Radiance = %4.1lf\n"
        "    Water Asymmetry = %4.1lf\n\n"
        "System Transfer Function Parameters:\n"
        "    Gaussian Optical Blur = %6.3lf\n"
        "    AT Detector Dimension = %6.3lf\n"
        "    XT Detector Dimension = %6.3lf\n"
        "    Integration Time      = %6.3lf\n"
        "    Charge Diffusion      = %6.3lf\n"
        "    Profile Phase = %7.4e\n\n"
        "Fit Statistics:\n"
        "    RMS Fit = %5.3lf\n"
        "    Number Iterations = ",
        model.bridge.rad_water, model.bridge.rad_asym, model.stf.g_optics,
        model.stf.at_detector, model.stf.xt_detector, model.stf.integ, 
        model.stf.diffusion, model.stf.phase, rms) < 0)
    {
        IAS_LOG_ERROR("Writing to report file");
        return ERROR;
    }

    for (iteration_index = 0; iteration_index < number_iterations - 1; 
        iteration_index++)
    {
        if (fprintf(report_file_pointer, "%d ", 
            solution_iterations[iteration_index]) < 0)
        {
            IAS_LOG_ERROR("Writing to report file");
            return ERROR;
        }
    }

    if (fprintf(report_file_pointer, "%d\n", 
        solution_iterations[iteration_index]) < 0)
    {
        IAS_LOG_ERROR("Writing to report file");
        return ERROR;
    }

    fflush(report_file_pointer);

    return SUCCESS;
}
