/******************************************************************************
NAME:  save_to_database.c

PURPOSE:  This sends results to the database for later analysis. It includes
support routines.

RETURNS: SUCCESS or ERROR 
******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "ias_db_insert.h"         /* Database library */
#include "ias_db_insert_transaction_using_table.h" /* Database library */
#include "ias_db_get_connect_info.h" /* Database library */
#include "ias_miscellaneous.h"     /* Used for ias_misc_get_char_id */
#include "ias_logging.h"           /* IAS logging library */
#include "save_to_database.h"      /* Header definitions for this file */
#include "get_targets.h"           /* Target structure */
#include "profile.h"               /* Profile structure */

#define MAX_DELETE_QUERY_LENGTH 85 /* Maximum size of delete query including
                                      ample space for table name and WO ID */

static struct ias_db_connection *db = NULL; /* Database connection */

/******************************************************************************
NAME:  get_database_connection

PURPOSE:  This routine gets a database connection so records can be sent to
the database. 

RETURNS: SUCCESS or ERROR 
******************************************************************************/
int get_database_connection
(
    void
)
{
    char *database_name = NULL;    /* Database name */
    char *user_name = NULL;        /* Database user name */
    char *password = NULL;         /* Database password */
    char *host = NULL;             /* Database host */

    /* Get the database connection info */
    if (ias_db_get_connect_info(&database_name, &user_name, &password, &host)
        != SUCCESS)
    {
        /* Failed getting the database connection info */
        IAS_LOG_ERROR("Getting database connection information");
        return ERROR;
    }

    /* Connect to the database for a transaction */
    db = ias_db_connect_with_transaction(database_name, user_name, password,
        host);

    /* Free the database connection information */
    free(database_name);
    free(user_name);
    free(password);
    free(host);
    database_name = NULL;
    user_name = NULL;
    password = NULL;
    host = NULL;

    if (db == NULL)
    {
        IAS_LOG_ERROR("Connecting to database");
        return ERROR;
    }

    return SUCCESS;
}

/******************************************************************************
NAME:  disconnect_with_transaction

PURPOSE: Disconnect from the database using a transaction (commit or rollback) 
the database. 

RETURNS: SUCCESS or ERROR 
******************************************************************************/
int disconnect_with_transaction
(
   int commit_flag                 /* I: Flag showing if commit should be done
                                         1 = IAS_DB_COMMIT
                                         0 = IAS_DB_ROLLBACK */
)
{
    int status = SUCCESS;

    if (db == NULL)
    {
        return status;
    }

    status = ias_db_disconnect_with_transaction(db, commit_flag);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Disconnecting from database with commit flag %d", 
            commit_flag);
    }

    return status;
}

/******************************************************************************
NAME:  save_to_database

PURPOSE:  This routine sends results to the database for later analysis. 

RETURNS: SUCCESS or ERROR 
******************************************************************************/
int save_to_database
(
    TARGET target,                 /* I: Current target structure */
    PROFILE profile,               /* I: Current profile structure */
    MTFMODEL model,                /* I: Best-fit model parameters */
    double rms,                    /* I: Best-fit RMS error */
    int *year_doy,                 /* I: Scene year and day of year */
    int *path_row,                 /* I: Scene path and row */
    const char *work_order_id      /* I: Work order ID */
)
{
    int status;                    /* Function return status */
    int bridge_table_size;         /* Size of the bridge database table */ 
    int span_table_size;           /* Size of the span database table */ 
    char bridge_database_table_name[35]; /* This table has records that apply
                                      to the whole application run */
    char span_database_table_name[35]; /* This table has per-span records. */
    char geometric_wo_common_id[IAS_CHAR_ID_SIZE];   
                                   /* Primary key in GEOMETRIC_WORK_ORDER_COMMON
                                      to associate processing with the WO */
    char query_string[MAX_DELETE_QUERY_LENGTH]; /* SQL statement string */
    struct ias_db_query *query = NULL; /* SQL query */
    char error_message[IAS_SQL_RTN_MSGLEN]; /* SQL query error message */
    struct bridge_db_info bridge_db; /* Bridge-level data to save */
    struct span_db_info span_db;   /* Span-level data to save */
    int span_index;                /* Loop counter */
    static int first_pass = 1;     /* Flag to show if it's the first pass */

    /* Set up the table names */
    strcpy(bridge_database_table_name, "mtf_bridge");
    strcpy(span_database_table_name, "mtf_spans");

    /* Get the database ID corresponding to the current work order */
    status = ias_misc_get_char_id(db, 
        "GEOMETRIC_WORK_ORDER_COMMON", "geometric_wo_common_id", 
        work_order_id, geometric_wo_common_id);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Retrieving database ID corresponding to current "
                      "geometric work order %s", work_order_id);
        return ERROR;
    }

    /* Put the Geometric WO common ID in the bridge table */
    strncpy(bridge_db.geometric_wo_common_id, geometric_wo_common_id, 
        sizeof(bridge_db.geometric_wo_common_id));
    bridge_db.geometric_wo_common_id[sizeof(
        bridge_db.geometric_wo_common_id) - 1] = '\0';

    /* Put the target name in the bridge table */
    strncpy(bridge_db.target_name, target.target_name, 
        sizeof(bridge_db.target_name));
    bridge_db.target_name[sizeof(bridge_db.target_name) - 1] = '\0';

    /* Put the various numbers in the bridge table */
    bridge_db.band_id = profile.band_number; 
    bridge_db.year = year_doy[0];
    bridge_db.day_of_year = year_doy[1];
    bridge_db.wrs_path = path_row[0]; 
    bridge_db.wrs_row = path_row[1]; 
    bridge_db.angle = profile.angle; 
    bridge_db.g_optics = model.stf.g_optics; 
    bridge_db.at_detector = model.stf.at_detector;
    bridge_db.xt_detector = model.stf.xt_detector; 
    bridge_db.integ = model.stf.integ; 
    bridge_db.diffusion = model.stf.diffusion; 
    bridge_db.phase = model.stf.phase;
    bridge_db.rms = rms; 
    bridge_db.gsd = model.bridge.gsd; 
    bridge_db.range = profile.range;
    bridge_db.velocity = profile.velocity; 
    bridge_db.rad_water = model.bridge.rad_water; 
    bridge_db.rad_asym = model.bridge.rad_asym; 
    bridge_db.nbr_of_spans = target.number_of_spans; 

    /* Build the bridge table of data to save */
    IAS_DB_INSERT_TABLE bridge_table[] =
    {
        IAS_DB_TABLE_STRUCT_STRING(&bridge_db, geometric_wo_common_id),
        IAS_DB_TABLE_STRUCT_STRING(&bridge_db, target_name),
        IAS_DB_TABLE_STRUCT_INT(&bridge_db, band_id),
        IAS_DB_TABLE_STRUCT_INT(&bridge_db, year),
        IAS_DB_TABLE_STRUCT_INT(&bridge_db, day_of_year),
        IAS_DB_TABLE_STRUCT_INT(&bridge_db, wrs_path),
        IAS_DB_TABLE_STRUCT_INT(&bridge_db, wrs_row),
        IAS_DB_TABLE_STRUCT_DOUBLE(&bridge_db, angle),
        IAS_DB_TABLE_STRUCT_DOUBLE(&bridge_db, g_optics),
        IAS_DB_TABLE_STRUCT_DOUBLE(&bridge_db, at_detector),
        IAS_DB_TABLE_STRUCT_DOUBLE(&bridge_db, xt_detector),
        IAS_DB_TABLE_STRUCT_DOUBLE(&bridge_db, integ),
        IAS_DB_TABLE_STRUCT_DOUBLE(&bridge_db, diffusion),
        IAS_DB_TABLE_STRUCT_DOUBLE(&bridge_db, phase),
        IAS_DB_TABLE_STRUCT_DOUBLE(&bridge_db, rms),
        IAS_DB_TABLE_STRUCT_DOUBLE(&bridge_db, gsd),
        IAS_DB_TABLE_STRUCT_DOUBLE(&bridge_db, range),
        IAS_DB_TABLE_STRUCT_DOUBLE(&bridge_db, velocity),
        IAS_DB_TABLE_STRUCT_DOUBLE(&bridge_db, rad_water),
        IAS_DB_TABLE_STRUCT_DOUBLE(&bridge_db, rad_asym),
        IAS_DB_TABLE_STRUCT_INT(&bridge_db, nbr_of_spans),
    };

    bridge_table_size = sizeof(bridge_table) / sizeof(IAS_DB_INSERT_TABLE);

    /* If it's the first pass, delete any previous records */
    if (first_pass)
    {
        status = snprintf(query_string, sizeof(query_string), "DELETE FROM %s "
        "WHERE geometric_wo_common_id=%s", bridge_database_table_name,
        geometric_wo_common_id);
        if (status < 0 || status >= sizeof(query_string))
        {
            IAS_LOG_ERROR("Preparing query string: %s ", query_string);
            return ERROR;
        }

        query = ias_db_perform_query(db, query_string);
        if (!ias_db_query_was_successful(query))
        {
            ias_db_query_get_error_message(query, error_message, sizeof(
                error_message));
            IAS_LOG_ERROR("Query of %s failed: %s", bridge_database_table_name,
                error_message);
            IAS_LOG_ERROR("Deleting records from %s where "
                "geometric_wo_common_id=%s", bridge_database_table_name, 
                geometric_wo_common_id);
            ias_db_query_close(query);
            return ERROR;
        }

        /* Close query */
        ias_db_query_close(query);
        query = NULL;
    }

    /* Insert the data into the bridge database table */
    status = ias_db_insert_using_table(db, bridge_database_table_name, 
        bridge_table, bridge_table_size, 1);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Inserting to %s DB table", bridge_database_table_name);
        return ERROR;
    }

    /* Put the Geometric WO common ID in the span table */
    strncpy(span_db.geometric_wo_common_id, geometric_wo_common_id, 
        sizeof(span_db.geometric_wo_common_id));
    span_db.geometric_wo_common_id[sizeof(
        span_db.geometric_wo_common_id) - 1] = '\0';

    /* Put the target name in the span table */
    strncpy(span_db.target_name, target.target_name, 
        sizeof(span_db.target_name));
    span_db.target_name[sizeof(span_db.target_name) - 1] = '\0';

    /* Put the band number in the span table */
    span_db.band_id = profile.band_number; 

    /* Verify the number of spans is in range */
    if (target.number_of_spans > MAXSPAN)
    {
        IAS_LOG_ERROR("Number of spans %d should be %d or less", 
            target.number_of_spans, MAXSPAN);
        return ERROR;
    }

    /* Assemble the row-specific span data to put in the span table */
    for (span_index = 0; span_index < target.number_of_spans; span_index++)
    {
        span_db.rad_bridge[span_index] 
            = model.bridge.span[span_index].rad_bridge; 
        span_db.width[span_index] = model.bridge.span[span_index].width;
        span_db.offset[span_index] = model.bridge.span[span_index].offset;
    }

    /* Build the span level table of data to save */
    IAS_DB_INSERT_TABLE span_table[] =
    {
        IAS_DB_TABLE_STRUCT_STRING(&span_db, geometric_wo_common_id),
        IAS_DB_TABLE_STRUCT_STRING(&span_db, target_name),
        IAS_DB_TABLE_STRUCT_INT(&span_db, band_id),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(&span_db, rad_bridge),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(&span_db, width),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(&span_db, offset),
    };

    span_table_size = sizeof(span_table) / sizeof(IAS_DB_INSERT_TABLE);

    /* If it's the fist pass, delete any previous records */
    if (first_pass)
    {
        status = snprintf(query_string, sizeof(query_string), 
            "DELETE FROM %s WHERE "
            "geometric_wo_common_id=%s", span_database_table_name,
            geometric_wo_common_id);
        if (status < 0 || status >= sizeof(query_string))
        {
            IAS_LOG_ERROR("Preparing query string: %s ", query_string);
            return ERROR;
        }

        query = ias_db_perform_query(db, query_string);
        if (!ias_db_query_was_successful(query))
        {
            ias_db_query_get_error_message(query, error_message, sizeof(
                error_message));
            IAS_LOG_ERROR("Query of %s failed: %s", span_database_table_name,
                error_message);
            IAS_LOG_ERROR("Deleting records from %s where "
                "geometric_wo_common_id=%s", span_database_table_name,
                geometric_wo_common_id);
            ias_db_query_close(query);
            return ERROR;
        }

        /* Close query */
        ias_db_query_close(query);
        query = NULL;

        first_pass = 0;
    }

    /* Insert the data into the span database table */
    status = ias_db_insert_using_table(db, span_database_table_name, 
        span_table, span_table_size, target.number_of_spans);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Inserting to %s DB table", span_database_table_name);
        return ERROR;
    }

    return status;
}
