/******************************************************************************

NAME:           mtfestimate_test

PURPOSE: Test MTF Estimate algorithm Annealing and MultiMin options 

RETURN VALUE:
Type = int
Value           Description
-----           -----------

******************************************************************************/
#include <stdlib.h>                   /* EXIT_FAILURE/EXIT_SUCCESS */
#include <limits.h>                   /* PATH_MAX */
#include <string.h>                   /* strncpy */

#include "mtfestimate.h"              /* MTF function prototypes */
#include "model.h"                    /* MTF model */
#include "profile.h"                  /* MTF profile structure */
#include "get_targets.h"              /* MTF targets structure and prototype */
#include "ias_logging.h"              /* IAS logging library */
#include "ias_miscellaneous.h"        /* For report header */
#include "cgreen/cgreen.h"

int test_mtfestimate
(
    int fit_method                    /* Annealing or MultiMin*/
)
{
    char parm_file[PATH_MAX] = "MTFESTIMATE.odl"; /* Input parameter file 
                                         name */
    char l0r_file[PATH_MAX];          /* L0R file name */
    char l1g_file[PATH_MAX];          /* L1G file name */
    char target_file[PATH_MAX];       /* Input target description file name */
    char input_directory[PATH_MAX];   /* Input data directory name */
    char work_order_id[IAS_WORKORDER_ID_SIZE]; /* Work order ID */

    int band_number;                  /* Band number to process */
    int band_index;                   /* Index into fit_parameters to find
                                         band_number */
    const char *band_name;            /* Descriptive band name */
    int number_targets;               /* Number of targets in image */
    TARGET targets[MAX_TARGETS];      /* Targets in image */
    int target_id;                    /* Id and index of current target */
    int target_index;                 /* Loop counter */
    int stf_fit_method;               /* Use either Annealing (0) or MultiMin
                                         (1) */
    MTFEST fit_parameters;            /* Structure containing parameters for
                                         this estimatation session */
    int *solution_iterations;         /* Number of iterations performed */
    PROFILE profile;                  /* Measured target profile data */
    MTFMODEL start_model;             /* Initial model values */
    MTFMODEL simplex_model;           /* Simplex values */
    MTFMODEL new_model;               /* Best-fit STF model parameters */
    IAS_REPORT_HEADER *header = NULL; /* Report header structure */
    FILE *report_file_pointer;        /* Pointer to output report file */
    const char *report_file = "mtfestimate.rpt";  /* Output report file name */
    FILE *trend_file_pointer;         /* Pointer to output trending file */
    const char *trend_file = "mtfestimate_trend.dat";  /* Trending output file
                                         name */
    double rms;                       /* Best-fit RMS error */
    int store_in_database_flag;       /* Flag for storing results in database */
    int status;                       /* Function return status */

    char command[PATH_MAX];           /* System command */
    static char *unit_test_directory = NULL; /* IAS unit test data directory */
    char *unit_test_path = "ias/gps/mtf/mtfestimate"; /* Rest of unit test data
                                         directory path */
    const char *profile_file = "ModelProfile_Tar0_Band1.csv";  /* Profile file
                                         for the test's target and band */ 
    char compare_path[PATH_MAX];      /* Path for output file comparisons */
    char *enable_db_tests = NULL;     /* Environment variable flag for DB use */
    int db_setting;                   /* Setting of DB parameter in application
                                         parameter file */

    /* Comparison versions of parameter variables */
    char compare_input_directory[PATH_MAX];
    char compare_l0r_file[PATH_MAX];
    char compare_l1g_file[PATH_MAX];
    char compare_target_file[PATH_MAX];

    /* Initialize the logging library */
    if (fit_method == 0)
    {
        if (ias_log_initialize("mtfestimate_annealing_test") != SUCCESS)
        {
            printf("Initializing logging library");
            return ERROR;
        }
        IAS_LOG_INFO("Initiating mtfestimate_annealing_test");
    }
    else
    {
        if (ias_log_initialize("mtfestimate_multimin_test") != SUCCESS)
        {
            printf("Initializing logging library");
            return ERROR;
        }
        IAS_LOG_INFO("Initiating mtfestimate_multimin_test");
    }

    /* Get DB test environment variable if it exists */
    enable_db_tests = getenv("ENABLE_DB_TESTS");

    /* Get unit test directory */
    unit_test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (unit_test_directory == NULL)
    {
        assert_true_with_message(0, "The IAS_UNIT_TEST_DATA_DIR environment "
            "variable is not set\n");
        return ERROR;
    }

    /* Make application parameter file with IAS_UNIT_TEST_DATA_DIR expanded */
    status = snprintf(command, PATH_MAX,
            "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' MTFESTIMATE_template.odl > "
            "MTFESTIMATE_expanded.odl", unit_test_directory);
    if (status < 0 || PATH_MAX <= status)
    {
        assert_true_with_message(0, "Error preparing application parameter "
            "file conversion system command");
        return ERROR;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Error preparing application parameter "
            "test file for unit test data directory path setup");
        return ERROR;
    }

    /* Make OMF with IAS_UNIT_TEST_DATA_DIR expanded */
    status = snprintf(command, PATH_MAX,
            "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' TSTI9257_template.omf > "
            "TSTI9257.omf", unit_test_directory);
    if (status < 0 || PATH_MAX <= status)
    {
        assert_true_with_message(0, "Error preparing OMF conversion system "
            "command");
        return ERROR;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Error preparing parameter test OMF file "
            "for unit test data directory path setup");
        return ERROR;
    }

    /* Update the application parameter file with the DB test setting */
    if (enable_db_tests == NULL)
    {
        printf("Database tests are disabled.\n");
        db_setting = 0;
    }
    else
    {
        db_setting = 1;
    }

    status = snprintf(command, PATH_MAX,
        "sed 's|STORE_CHARACTERIZATION_IN_DB = Z|"
        "STORE_CHARACTERIZATION_IN_DB = %d|' MTFESTIMATE_expanded.odl > "
        "MTFESTIMATE.odl", db_setting);
    if (status < 0 || PATH_MAX <= status)
    {
        assert_true_with_message(0, "Error preparing application parameter "
            "file DB setting system command");
        return ERROR;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Error preparing application parameter "
            "test file for DB setting");
        return ERROR;
    }

    /* Get work order parameters.  We will ignore stf_fit_method and use
       the passed-in fit_method instead */
    if (get_parameters(parm_file, target_file, input_directory, l0r_file,
        l1g_file, work_order_id, &stf_fit_method, &fit_parameters,
        &store_in_database_flag) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from file: %s", parm_file);
        return ERROR;
    }

    /* Set up comparison strings so the parameters can be checked */
    status = snprintf(compare_input_directory, PATH_MAX, "%s/%s",
        unit_test_directory, unit_test_path);
    if (status < 0 || PATH_MAX <= status)
    {
        assert_true_with_message(0, "Error building compare_input_directory");
        return ERROR;
    }
    status = snprintf(compare_l0r_file, PATH_MAX, "%s/%s/%s",
        unit_test_directory, unit_test_path, "LO80220390392000261EDC00");
    if (status < 0 || PATH_MAX <= status)
    {
        assert_true_with_message(0, "Error building compare_l0r_file");
        return ERROR;
    }
    status = snprintf(compare_l1g_file, PATH_MAX, "%s/%s/%s",
        unit_test_directory, unit_test_path, "sys_l1g.h5");
    if (status < 0 || PATH_MAX <= status)
    {
        assert_true_with_message(0, "Error building compare_l1g_file");
        return ERROR;
    }
    status = snprintf(compare_target_file, PATH_MAX, "%s/%s/%s",
        unit_test_directory, unit_test_path, "MTF_Targets_022039.table");
    if (status < 0 || PATH_MAX <= status)
    {
        assert_true_with_message(0, "Error building compare_target_file");
        return ERROR;
    }

    /* Compare the comparisons strings to the results from get_parameters */
    assert_string_equal(input_directory, compare_input_directory);
    assert_string_equal(l0r_file, compare_l0r_file);
    assert_string_equal(l1g_file, compare_l1g_file);
    assert_string_equal(target_file, compare_target_file);

    /* Compare other get_parameters outputs to expected results */
    assert_string_equal(work_order_id, "TSTI9257");
    assert_equal(stf_fit_method, 0); /* This tests the application parameter,
        not the test input parameter "fit_method" that overrides it */
    assert_equal(fit_parameters.year_doy[0], 2000);
    assert_equal(fit_parameters.year_doy[1], 261);
    assert_equal(fit_parameters.path_row[0], 22);
    assert_equal(fit_parameters.path_row[1], 39);
    assert_equal(fit_parameters.numbands, 1);
    assert_equal(fit_parameters.band_numbers[0], 1);

    /* Get system table parameters */
    if (get_system_table(input_directory, &fit_parameters) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from system table: %s/%s",
            input_directory, MTF_SYSTEM_FILE);
        return ERROR;
    }

    /* Compare some of the MTF system table parameters to expected results */
    assert_equal(fit_parameters.model[0].nstf, 6);
    assert_double_equal(fit_parameters.model[0].bridge.rad_water, 970.0);
    assert_double_equal(fit_parameters.model[0].bridge.rad_asym, 0.0);
    assert_double_equal(fit_parameters.model[0].bridge.gsd, 30.0);
    assert_double_equal(fit_parameters.model[0].bridge.span[0].rad_bridge, 
        1370.0);
    assert_double_equal(fit_parameters.model[0].bridge.span[0].width, 10.0);
    assert_double_equal(fit_parameters.model[0].bridge.span[0].offset, -17.2);
    assert_double_equal(fit_parameters.model[0].stf.g_optics, 20.0);
    assert_double_equal(fit_parameters.model[0].stf.xt_detector, 40.863);
    assert_double_equal(fit_parameters.model[0].stf.at_detector, 42.303);
    assert_double_equal(fit_parameters.model[0].stf.integ, 3.60);
    assert_double_equal(fit_parameters.model[0].stf.diffusion, 0.0);
    assert_double_equal(fit_parameters.model[0].stf.phase, 0.0);
    assert_equal(fit_parameters.splex[0].nstf, 6);
    assert_double_equal(fit_parameters.splex[0].bridge.rad_water, 50.0);
    assert_double_equal(fit_parameters.splex[0].bridge.rad_asym, 15.0);
    assert_double_equal(fit_parameters.splex[0].bridge.gsd, 3.0);
    assert_double_equal(fit_parameters.splex[0].bridge.span[0].rad_bridge, 
        50.0);
    assert_double_equal(fit_parameters.splex[0].bridge.span[0].width, 1.0);
    assert_double_equal(fit_parameters.splex[0].bridge.span[0].offset, 2.0);
    assert_double_equal(fit_parameters.splex[0].stf.g_optics, 5.0);
    assert_double_equal(fit_parameters.splex[0].stf.xt_detector, 2.5);
    assert_double_equal(fit_parameters.splex[0].stf.at_detector, 2.5);
    assert_double_equal(fit_parameters.splex[0].stf.integ, 0.6);
    assert_double_equal(fit_parameters.splex[0].stf.diffusion, 0.2);
    assert_double_equal(fit_parameters.splex[0].stf.phase, 10.0);
    assert_equal(fit_parameters.control[0].stfmask.g_optics, 0);
    assert_equal(fit_parameters.control[0].stfmask.xt_detector, 0);
    assert_equal(fit_parameters.control[0].stfmask.at_detector, 0);
    assert_equal(fit_parameters.control[0].stfmask.integ, 0);
    assert_equal(fit_parameters.control[0].stfmask.diffusion, 0);
    assert_equal(fit_parameters.control[0].stfmask.phase, 1);
    assert_equal(fit_parameters.control[0].bridgemask.rad_water, 1);
    assert_equal(fit_parameters.control[0].bridgemask.rad_asym, 1);
    assert_equal(fit_parameters.control[0].bridgemask.gsd, 0);
    assert_equal(fit_parameters.control[0].bridgemask.span[0].rad_bridge, 1);
    assert_equal(fit_parameters.control[0].bridgemask.span[0].width, 0);
    assert_equal(fit_parameters.control[0].bridgemask.span[0].offset, 0);
    assert_equal(fit_parameters.numiter, 1);

    /* Read the target definition file.  This is tested at the shared level */
    number_targets = get_targets(target_file, targets);
    if (number_targets <= 0 || number_targets > MAX_TARGETS)
    {
        IAS_LOG_ERROR("Reading target description file %s; number of targets "
            "is %d but should be between 1 and %d", target_file, number_targets,
             MAX_TARGETS);
        return ERROR;
    }

    /* Allocate memory for the number solved per iteration */
    solution_iterations = (int *)malloc(sizeof(int) * fit_parameters.numiter);
    if (solution_iterations == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for number solved per iteration");
        for (target_id = 0; target_id < number_targets; target_id++)
        {
            free(targets[target_id].span);
        }
        return ERROR;
    }

    /* Open the output report file */
    if ((report_file_pointer = fopen(report_file, "w")) == NULL)
    {
        IAS_LOG_ERROR("Opening output report file %s", report_file);
        free(solution_iterations);
        for (target_id = 0; target_id < number_targets; target_id++)
        {
            free(targets[target_id].span);
        }
        return ERROR;
    }

    /* Initialize the report header */
    header = ias_misc_initialize_gps_report_header(l1g_file, l0r_file,
            work_order_id, "mtfestimate");
    if (!header)
    {
        IAS_LOG_ERROR("Initializing the GPS report file header");
        fclose(report_file_pointer);
        free(solution_iterations);
        for (target_id = 0; target_id < number_targets; target_id++)
        {
            free(targets[target_id].span);
        }
        return ERROR;
    }

    /* Write the report header to the report file */
    status = ias_misc_write_gps_report_header(report_file_pointer, header);
    ias_misc_free_report_header(header);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the GPS report file header");
        fclose(report_file_pointer);
        free(solution_iterations);
        for (target_id = 0; target_id < number_targets; target_id++)
        {
            free(targets[target_id].span);
        }
        return ERROR;
    }

    /* Open the output trending file */
    if ((trend_file_pointer = fopen(trend_file, "w")) == NULL)
    {
        IAS_LOG_ERROR("Opening output trending file %s", trend_file);
        fclose(report_file_pointer);
        free(solution_iterations);
        for (target_id = 0; target_id < number_targets; target_id++)
        {
            free(targets[target_id].span);
        }
        return ERROR;
    }

    /* Connect to the database */
    if (store_in_database_flag)
    {
        status = get_database_connection();
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Connecting to the database");
            fclose(report_file_pointer);
            fclose(trend_file_pointer);
            free(solution_iterations);
            for (target_id = 0; target_id < number_targets; target_id++)
            {
                free(targets[target_id].span);
            }
            return ERROR;
        }
    }

    /* Loop through the targets (the current test only has 1) */
    for (target_id = 0; target_id < number_targets; target_id++)
    {
        IAS_LOG_INFO("Begin processing target %s...",
            targets[target_id].target_name);

        /* Loop through the bands (the current test only has 1) */
        for (band_index = 0; band_index < fit_parameters.numbands; band_index++)
        {
            band_number = fit_parameters.band_numbers[band_index];
            band_name = ias_sat_attr_convert_band_number_to_name(band_number);
            if (band_name == NULL)
            {
                IAS_LOG_ERROR("No name was found for band number %d",
                    band_number);
                fclose(report_file_pointer);
                fclose(trend_file_pointer);
                free(solution_iterations);
                for (target_index = target_id; target_index < number_targets;
                    target_index++)
                {
                    free(targets[target_index].span);
                }
                return ERROR;
            }

            IAS_LOG_INFO("Begin processing band %s (%d,%d)...",
                band_name, band_index, band_number);

            /* Process the cross-target profile */
            if (read_profile(target_id, band_number, &profile) != SUCCESS)
            {
                IAS_LOG_ERROR("Reading data profile for target %d and band "
                    "number %d", target_id, band_number);
                fclose(report_file_pointer);
                fclose(trend_file_pointer);
                free(solution_iterations);
                for (target_index = target_id; target_index < number_targets;
                    target_index++)
                {
                    free(targets[target_index].span);
                }
                return ERROR;
            }

            /* Compare some of the profile parameters to expected results */
            assert_string_equal(profile.target_name, "Causeway");
            assert_equal(profile.target_id, 0);
            assert_equal(profile.band_number, 1);
            assert_equal(profile.number_points, 32);
            assert_equal(profile.oversample, 8);
            assert_double_equal(profile.gsd, 30.0);
            assert_double_equal(profile.angle, 0.071775242);
            assert_double_equal(profile.range, 705742.540000);
            assert_double_equal(profile.velocity, 6839.067);
            assert_double_equal(profile.sample_data[0], 969.296296);
            assert_double_equal(profile.sample_data[profile.number_points - 1],
                1086.915385);
            assert_double_equal(profile.sigma[0], 69.215333);
            assert_double_equal(profile.sigma[profile.number_points - 1],
                93.132400);

            /* Set up the fit */
            if (initialize_fit(band_index, &fit_parameters, &targets[target_id],
                &profile, &start_model, &simplex_model) != SUCCESS)
            {
                IAS_LOG_ERROR("Setting up the fit for target %d and band "
                    "number %d", target_id, band_number);
                fclose(report_file_pointer);
                fclose(trend_file_pointer);
                free_profile(&profile);
                free(solution_iterations);
                for (target_index = target_id; target_index < number_targets;
                    target_index++)
                {
                    free(targets[target_index].span);
                }
                return ERROR;
            }

            /* Compare the initialization of some start_model values to 
               expected results */
            assert_equal(start_model.nstf, 6);
            assert_double_equal(start_model.stf.g_optics, 20.0);
            assert_double_equal(start_model.stf.xt_detector, 40.863);
            assert_double_equal(start_model.stf.at_detector, 42.303);
            assert_double_equal(start_model.stf.integ, 3.6);
            assert_double_equal(start_model.stf.diffusion, 0.0);
            assert_double_equal(start_model.stf.phase, 0.0);
            assert_equal(start_model.nspan, 2);
            assert_equal(start_model.nbridge, 9);
            assert_double_equal(start_model.bridge.rad_water, 970.0);
            assert_double_equal(start_model.bridge.rad_asym, 0.0);
            assert_double_equal(start_model.bridge.gsd, 30.0);
            assert_double_equal(start_model.bridge.span[0].rad_bridge, 1370.0);
            assert_double_equal(start_model.bridge.span[0].width, 10.0);
            assert_double_equal(start_model.bridge.span[0].offset, -17.2);

            /* Compare the initialization of some simplex model values to 
               expected results */
            assert_equal(simplex_model.nstf, 6);
            assert_double_equal(simplex_model.stf.g_optics, 5.0);
            assert_double_equal(simplex_model.stf.xt_detector, 2.5);
            assert_double_equal(simplex_model.stf.at_detector, 2.5);
            assert_double_equal(simplex_model.stf.integ, 0.6);
            assert_double_equal(simplex_model.stf.diffusion, 0.2);
            assert_double_equal(simplex_model.stf.phase, 10.0);
            assert_equal(simplex_model.nspan, 2);
            assert_equal(simplex_model.nbridge, 9);
            assert_double_equal(simplex_model.bridge.rad_water, 50.0);
            assert_double_equal(simplex_model.bridge.rad_asym, 15.0);
            assert_double_equal(simplex_model.bridge.gsd, 3.0);
            assert_double_equal(simplex_model.bridge.span[0].rad_bridge, 50.0);
            assert_double_equal(simplex_model.bridge.span[0].width, 1.0);
            assert_double_equal(simplex_model.bridge.span[0].offset, 2.0);

            /* Do the fit based on Annealing */
            IAS_LOG_INFO("Begin the STF fit...");
            if (fit_method == 0)
            {
                status = fit_annealing(&start_model, &simplex_model,
                    fit_parameters.numiter, fit_parameters.control, &profile,
                    &rms, &new_model, solution_iterations);

                /* Compare some new_model values to expected results */
                assert_equal(new_model.nstf, 6);
                assert_double_equal(new_model.stf.g_optics, 20.0);
                assert_double_equal(new_model.stf.xt_detector, 40.863);
                assert_double_equal(new_model.stf.at_detector, 42.303);
                assert_double_equal(new_model.stf.integ, 3.6);
                assert_double_equal(new_model.stf.diffusion, 0.0);
                assert_double_equal(new_model.stf.phase, 32.907794);
                assert_equal(new_model.nspan, 2);
                assert_equal(new_model.nbridge, 9);
                assert_double_equal(new_model.bridge.rad_water, 1249.914578); 
                assert_double_equal(new_model.bridge.rad_asym, -326.836826); 
                assert_double_equal(new_model.bridge.gsd, 30.0);
                assert_double_equal(new_model.bridge.span[0].rad_bridge,
                    2539.097457);
                assert_double_equal(new_model.bridge.span[0].width, 10.0);
                assert_double_equal(new_model.bridge.span[0].offset, -17.2);

                /* Compare a solution_iterations value to expected results */
                assert_equal(solution_iterations[0], 298);
            }
            else
            {
                status = fit_multimin(&start_model, &simplex_model,
                    fit_parameters.numiter, fit_parameters.control, &profile,
                    &rms, &new_model, solution_iterations);

                /* Compare some new_model values to expected results */
                assert_equal(new_model.nstf, 6);
                assert_double_equal(new_model.stf.g_optics, 20.0);
                assert_double_equal(new_model.stf.xt_detector, 40.863);
                assert_double_equal(new_model.stf.at_detector, 42.303);
                assert_double_equal(new_model.stf.integ, 3.6);
                assert_double_equal(new_model.stf.diffusion, 0.0);
                assert_double_equal(new_model.stf.phase, -20.479056);
                assert_equal(new_model.nspan, 2);
                assert_equal(new_model.nbridge, 9);
                assert_double_equal(new_model.bridge.rad_water, 998.712023); 
                assert_double_equal(new_model.bridge.rad_asym, 424.991155); 
                assert_double_equal(new_model.bridge.gsd, 30.0);
                assert_double_equal(new_model.bridge.span[0].rad_bridge,
                    681.151073);
                assert_double_equal(new_model.bridge.span[0].width, 10.0);
                assert_double_equal(new_model.bridge.span[0].offset, -17.2);

                /* Compare a solution_iterations value to expected results */
                assert_equal(solution_iterations[0], 621);
            }

            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Fitting STF");
                fclose(report_file_pointer);
                fclose(trend_file_pointer);
                free_profile(&profile);
                free(solution_iterations);
                for (target_index = target_id; target_index < number_targets;
                    target_index++)
                {
                    free(targets[target_index].span);
                }
                return ERROR;
            }

            /* Report the results for this target and band */
            status = report_fit(report_file_pointer, target_id,
                targets[target_id], profile, new_model, rms,
                fit_parameters.year_doy, fit_parameters.path_row,
                fit_parameters.numiter, solution_iterations);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Generating fit report");
                fclose(report_file_pointer);
                fclose(trend_file_pointer);
                free_profile(&profile);
                free(solution_iterations);
                for (target_index = target_id; target_index < number_targets;
                    target_index++)
                {
                    free(targets[target_index].span);
                }
                return ERROR;
            }

            /* Send the trending data to a file */
            trend_fit(trend_file_pointer, targets[target_id], profile,
                new_model, rms, fit_parameters.year_doy,
                fit_parameters.path_row);

            /* Send the trending data to the database */
            if (store_in_database_flag)
            {
                if (save_to_database(targets[target_id], profile, new_model,
                    rms, fit_parameters.year_doy, fit_parameters.path_row,
                    work_order_id) != SUCCESS)
                {
                    IAS_LOG_ERROR("Sending fit results to database");
                    fclose(report_file_pointer);
                    free_profile(&profile);
                    free(solution_iterations);
                    for (target_index = target_id; target_index <
                        number_targets; target_index++)
                    {
                        free(targets[target_index].span);
                    }
                    return ERROR; 
                }
            }

            /* Free the memory for the current profile */
            free_profile(&profile);
        }

        /* Release this target structure's allocated memory */
        free(targets[target_id].span);
        targets[target_id].span = NULL;
    }

    /* Commit the database transaction */
    if (store_in_database_flag)
    {
        status = disconnect_with_transaction(IAS_DB_COMMIT);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Committing transaction and disconnecting from "
                "database");
            free(fit_parameters.control);
            free(solution_iterations);
            return ERROR; 
        }
    }

    /* Close the report output file */
    status = fclose(report_file_pointer);
    if (status != 0)
    {
        IAS_LOG_ERROR("Closing report file %s", report_file);
        fclose(trend_file_pointer);
        free(fit_parameters.control);
        free(solution_iterations);
        return ERROR;
    }

    /* Close the trending output file */
    fclose(trend_file_pointer);
    if (status != 0)
    {
        IAS_LOG_ERROR("Closing trending file %s", trend_file);
        free(fit_parameters.control);
        free(solution_iterations);
        return ERROR;
    }

    /* Free memory */
    free(fit_parameters.control);
    fit_parameters.control = NULL;
    free(solution_iterations);
    solution_iterations = NULL;

    /* Build comparison directory path */
    if (fit_method == 0)
    {
        snprintf(compare_path, sizeof(command), "%s/%s/annealing", 
            unit_test_directory, unit_test_path);
    }
    else
    {
        snprintf(compare_path, sizeof(command), "%s/%s/multimin", 
            unit_test_directory, unit_test_path);
    }
    compare_path[PATH_MAX - 1] = '\0';

    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error building comparison directory "
            "path");
        return ERROR;
    }

    /* Compare trend file to baseline results */
    status = snprintf(command, sizeof(command), "diff -q %s %s/%s", trend_file,
        compare_path, trend_file);
    if (status < 0 || sizeof(command) <= status)
    {
        assert_true_with_message(0, "Error preparing diff system command");
        return ERROR;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found in trend file. Review "
                "%s and %s/%s for details", trend_file, compare_path, 
                trend_file);
    }

    /* Compare report file to baseline results */
    status = snprintf(command, sizeof(command),
        "sed '/L0R Image/,$!'d %s | diff -q - %s/%s", report_file,
        compare_path, report_file);
    if (status < 0 || sizeof(command) <= status)
    {
        assert_true_with_message(0, "Error preparing diff system command");
        return ERROR;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found in report file. Review "
                "%s and %s/%s for details", report_file, compare_path, 
                report_file);
    }

    /* Compare profile csv file to baseline results */
    status = snprintf(command, sizeof(command), "diff -q %s %s/%s", 
        profile_file, compare_path, profile_file);
    if (status < 0 || sizeof(command) <= status)
    {
        assert_true_with_message(0, "Error preparing diff system command");
        return ERROR;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found in profile file. Review "
                "%s and %s/%s for details", profile_file, compare_path, 
                profile_file);
    }

    /* Verify the database results */
    if (enable_db_tests != NULL)
    {
        /* Run the MTF DB retrieval script */
        status = system("../../mtfperform/retrieve_mtf_db.pl "
            "mtf_trendall.dat");
        if (status)
        {
            assert_true_with_message(0, "../../mtfperform/retrieve_mtf_db.pl "
                "returned an error");
        }

        /* The MTF DB retrieval script retrieves all MTF estimate records.
           Look for the row from this test */
        if (fit_method == 0)
        {
            status = system("grep -q \"2000 | 261 | 022 | 039 | Causeway | "
                "1 |  0.071775 | 20.0000 | 42.303 | 40.863 |  3.600 |  "
                "0.000 |  3.2908e+01 |  1.688 | 30.000 | 705742.540 | "
                "6839.067 | 1249.9 | -326.8 | 2 | 2539.1 | 10.00 | -17.20 | "
                "1707.6 | 10.00 |  17.20\" mtf_trendall.dat");
        }
        else
        {
            status = system("grep -q \"2000 | 261 | 022 | 039 | Causeway | "
                "1 |  0.071775 | 20.0000 | 42.303 | 40.863 |  3.600 |  "
                "0.000 | -2.0479e+01 |  1.741 | 30.000 | 705742.540 | "
                "6839.067 |  998.7 | 425.0 | 2 |  681.2 | 10.00 | -17.20 |  "
                "888.9 | 10.00 |  17.20\" mtf_trendall.dat");
        }

        if (status)
        {
            assert_true_with_message(0, "Could not find expected results in "
                "mtf_trendall.dat");
        }
    }

    /* Write completion message */
    IAS_LOG_INFO("Successful completion.");

    return SUCCESS;
}

/* Fit based on Annealing */
Ensure test_mtfestimate_annealing()
{
    int status;

    status = test_mtfestimate(0);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "MTF Estimate with annealing option "
            "failed");
    }
    return;
}

/* Fit based on MultiMin */
Ensure test_mtfestimate_multimin()
{
    int status;

    status = test_mtfestimate(1);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "MTF Estimate with MultiMin option failed");
    }
    return;
}

int main(int argc, char *argv[])
{
    TestSuite *suite = create_test_suite();
    add_test(suite, test_mtfestimate_annealing);
    add_test(suite, test_mtfestimate_multimin);
    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}


