/******************************************************************************
NAME:           write_profile

PURPOSE:
Write the best-fit model profiles to a comma-delimited text file.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successfully wrote profiles 
ERROR    Error writing profiles 

******************************************************************************/
#include <stdio.h>

#include "ias_logging.h"           /* IAS logging library */
#include "mtfestimate.h"           /* MTF prototypes */

#define FILENAME_SIZE 35           /* Size of the known filename string plus
                                      plenty of room for targets and bands */

int write_profile 
(
    int target_id,                 /* I: Current target index */
    int band_number,               /* I: Data band number */
    int number_points,             /* I: Number of points in profile */
   double *model_profile           /* I: Model profile data */
)
{
    FILE *output_file;             /* Output file pointer */
    char filename[FILENAME_SIZE];  /* Output file name */
    int sample_index;              /* Output data sample index */
    int size;                      /* Size returned by snprintf */
    int status;                    /* Function return status */

    /* Construct the output file name */
    size = snprintf(filename, sizeof(filename), 
        "ModelProfile_Tar%d_Band%d.csv", target_id, band_number);
    if ((size < 0) || (size >= sizeof(filename)))
    {
        IAS_LOG_ERROR("Output file name length is %d; it should be between "
            "0 and %lu", size, sizeof(filename));
        return ERROR;
    }

    /* Open the output file */
    output_file = fopen(filename, "w");
    if (output_file == NULL)
    {
        IAS_LOG_ERROR("Opening output file %s", filename);
        return ERROR;
    }

    /* Write the profile data */
    if (fprintf(output_file, "%6.1lf", model_profile[0]) < 0)
    {
        IAS_LOG_ERROR("Writing to output file %s", filename);
        return ERROR;
    }
    for (sample_index = 1; sample_index < number_points; sample_index++)
    {
        if (fprintf(output_file, ", %6.1lf", model_profile[sample_index]) < 0)
        {
            IAS_LOG_ERROR("Writing to output file %s", filename);
            return ERROR;
        }
    }
    if (fprintf(output_file, "\n") < 0)
    {
        IAS_LOG_ERROR("Writing to output file %s", filename);
        return ERROR;
    }

    /* Close the output file */
    status = fclose(output_file);
    if (status != 0)
    {
        IAS_LOG_ERROR("Closing output file %s", filename);
        return ERROR;
    }

    return SUCCESS;
}
