/******************************************************************************
NAME:           generate_report_for_target

PURPOSE:
Define and extract the image windows from the band covering the MTF targets
and its corresponding polygons.  Data is written to an ASCII file.  The pixel
that is written also has its scaled range, ground velocity, and rotation angle
listed.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
SUCCESS         Successfully extracted target 
ERROR           Error extracting target

******************************************************************************/
#include <stdlib.h>                /* malloc */
#include <limits.h>                /* PATH_MAX */
#include <string.h>                /* memcpy */

#include "ias_cpf.h"
#include "ias_grid.h"
#include "ias_grid_io.h"
#include "ias_model_io.h"
#include "ias_structures.h"
#include "ias_logging.h"
#include "mtfextract.h"

/* Get target for the current unique SCA */
int generate_report_for_sca_target
(
    int current_sca,               /* I: SCA of target */
    const char *l1r_file,          /* I: Level 1R data file name */
    int band_number,               /* I: Band number to process */
    int number_scas,               /* I: Number of SCAs for OLI */
    IAS_CPF *cpf,                  /* I: CPF constants */
    int current_target,            /* I: Current target index */
    int current_polygon,           /* I: Current polygon index */
    const MTF_TARGET *target,      /* I: Current target */
    IAS_LOS_MODEL *model,          /* I: Model structure */
    IAS_GRID_TYPE *grid,           /* I: Grid structure */
    int input_number_lines,        /* I: Number of lines in L1R */
    int input_number_samples,      /* I: Number of samples in L1R */
    double elevation,              /* I: Elevation for L1R location */
    int unique_sca_count,          /* I: Unique SCA number */
    IAS_UI2_TYPE *image_data       /* I: Pre-allocated buffer for image data 
                                         to be placed */
)
{
    int band_index;                /* Band index for band number */
    char output_filename[PATH_MAX]; /* Output file name */
    int number_points;             /* Number of polygon points */
    double maximum_line;           /* Maximum L1R line location */
    double minimum_line;           /* Minimum L1R line location */
    double maximum_sample;         /* Maximum L1R sample location */
    double minimum_sample;         /* Minimum L1R sample location */
    int upper_left_line;           /* Upper left coordinate line */
    int upper_left_sample;         /* Upper left coordinate sample */
    int lower_right_line;          /* Lower right coordinate line */
    int lower_right_sample;        /* Lower right coordinate sample */
    int number_lines;              /* Number of lines to read */
    int number_samples;            /* Number of samples to read */
    int current_line;              /* Loop counter */
    int current_sample;            /* Loop counter */
    int current_line_location;     /* Location of the start of current line */
    double input_line;             /* L1R line location */
    double input_sample;           /* L1R sample location */
    double utm_x;                  /* L1R UTM X coordinate */
    double utm_y;                  /* L1R UTM Y coordinate */
    double rotation;               /* Target orientation angle */
    double distance;               /* Distance between satellite and target */
    double ground_velocity;        /* Ground velocity */
    int image_offset;              /* Offset into image buffer */
    FILE *output_file;             /* Output file for pixel information */
    IAS_IMAGE sca_image;           /* Image from L1R file */
    int status;                    /* Function return status */

    /* Set band index */
    band_index = ias_sat_attr_convert_band_number_to_index(band_number);
    if (band_index == ERROR)
    {
        IAS_LOG_ERROR("Invalid band index for band number %d", band_number);
        return ERROR;
    }

    /* Set number of points */
    number_points = target->polygons[current_polygon].number_points;

    /* Initialize the data read */
    if (read_l1r_band_sca(l1r_file, band_number, current_sca, number_scas, cpf,
        &sca_image) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing band number %d SCA %d of L1R file %s",
            band_number, current_sca, l1r_file);
        return ERROR;
    }

    /* Build the output filename */
    sprintf(output_filename, "Tar%dPoly%dBand%dSca%d.dat", current_target,
        current_polygon, band_number, current_sca);

    /* Determine the minimum and maximum line and sample for the polygon */
    if (polygon_sca_bounds(&grid->gridbands[band_index][current_sca], 
        &target->polygons[current_polygon], &minimum_line, &minimum_sample, 
        &maximum_line, &maximum_sample) != SUCCESS)
    { 
        IAS_LOG_ERROR("Determining bounds for target %d polygon %d SCA %d",
            current_target, current_polygon, current_sca);
        cleanup_l1r_image(&sca_image);
        return ERROR;
    } 

    /* Set upper left and lower right line and sample of the polygon
       boundaries */
    upper_left_line = minimum_line;
    upper_left_sample = minimum_sample;
    if (upper_left_line < 0)
    {
        upper_left_line = 0;
    }
    if (upper_left_sample < 0)
    {
        upper_left_sample = 0;
    }
    lower_right_line = maximum_line;
    lower_right_sample = maximum_sample;

    /* Set the number of lines and samples based on the difference between the
       lower right and upper left coordinates */
    number_lines = lower_right_line - upper_left_line;
    number_samples = lower_right_sample - upper_left_sample;

    /* Adjust the number of lines and samples to read if they go out of the
       boundaries defined in the grid */ 
    if (upper_left_line + number_lines > input_number_lines)
    {
        number_lines = input_number_lines - upper_left_line;
    }
    if (upper_left_sample + number_samples > input_number_samples)
    {
        number_samples = input_number_samples - upper_left_sample;
    }

    IAS_LOG_DEBUG("Processing Band %d Target %d Polygon %d SCA %d Unique %d\n",
       band_number, current_target, current_polygon, current_sca,
       unique_sca_count);

    /* Read the desired tile from the image */
    if (get_l1r_window(upper_left_line, upper_left_sample, number_lines, 
        number_samples, &sca_image, image_data) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading tile from %s band number %d SCA %d with upper "
            "left %d,%d number lines %d number samples %d", l1r_file,
            band_number, current_sca, upper_left_line, upper_left_sample,
            number_lines, number_samples);
        cleanup_l1r_image(&sca_image);
        return ERROR;
    }

    /* Open the current output file */
    output_file = fopen(output_filename, "w");
    if (output_file == NULL)
    {
        IAS_LOG_ERROR("Opening target table tile %s", output_filename);
        cleanup_l1r_image(&sca_image);
        return ERROR;
    }

    /* Loop through the lines and samples */
    for (current_line = 0; current_line < number_lines; current_line++)
    {
        /* Calculate the start location of the current line */
        current_line_location = current_line * number_samples;

        input_line = upper_left_line + current_line;

        for (current_sample = 0; current_sample < number_samples; 
            current_sample++)
        {
            input_sample = upper_left_sample + current_sample;

            /* Given the vertices of the polygon, determine if
               point(input_sample,input_line) falls within the polygon. */
            if (ias_math_point_in_polygon(number_points,
                &target->polygons[current_polygon].l1r_sample[
                    number_points * current_sca],  
                &target->polygons[current_polygon].l1r_line[
                    number_points * current_sca], input_sample, input_line)
                == TRUE)
            {
                /* Determine the UTM coordinates, scaled range, ground
                   velocity, and rotation angle for the current pixel */
                if (get_utm(model, band_number, current_sca, elevation,
                    input_line, input_sample, target->angle, &distance,
                    &rotation, &ground_velocity, &utm_y, &utm_x) != SUCCESS)
                {
                    IAS_LOG_ERROR("Finding UTM coordinate for input line and "
                        "sample %f,%f for band number %d SCA %d", input_line,
                        input_sample, band_number, current_sca);
                    cleanup_l1r_image(&sca_image);
                    return ERROR;
                }

                /* Calculate the offset into image data */ 
                image_offset = current_line_location + current_sample;

                /* Write the data to the output file */
                if (fprintf(output_file, "%d %d %f %f %f %f %f %f %f %d\n", 
                    band_number, current_sca, input_line, input_sample, utm_y, 
                    utm_x, ground_velocity, rotation, distance,
                    image_data[image_offset]) < 0)
                {
                    IAS_LOG_ERROR("Writing to output file %s for input line "
                        "and sample %f,%f band number %d SCA %d", 
                        output_filename, input_line, input_sample, 
                        band_number, current_sca);
                    cleanup_l1r_image(&sca_image);
                    return ERROR;
                }
            }
        } /* Sample loop */
    } /* Line loop */

    /* Close the output file */
    status = fclose(output_file);
    if (status != 0)
    {
        IAS_LOG_ERROR("Closing output file %s ", output_filename);
        cleanup_l1r_image(&sca_image);
        return ERROR;
    }

    /* Cleanup allocated data read from the l1r file */
    cleanup_l1r_image(&sca_image);

    return SUCCESS;
}

/* Make a list of the SCAs of a polygon without duplicates */
void get_unique_sca
(
    MTF_POLYGON *polygon,          /* I: Target polygon */
    int *unique_scas,              /* O: SCAs target polygon is in */
    int *number_unique_scas        /* O: Number of unique SCAs */
)
{
    int first_active_sca_index = 0; /* Index of the first active SCA */
    int sca_index;                 /* SCA index */
    int unique_index;              /* Index into unique SCAs */

    *number_unique_scas = 0;

    /* Find the first active SCA */ 
    while (first_active_sca_index < polygon->number_scas)
    {
        /* If the SCA is active */
        if (polygon->sca[first_active_sca_index] != -1)
        {
            /* An active SCA was found, so track which SCA that is, record
               that an active SCA was found, and stop looking for one */
            unique_scas[0] = polygon->sca[first_active_sca_index];
            *number_unique_scas = 1;
            break;
        }

        /* The first active SCA wasn't found yet, so keep looking */
        first_active_sca_index++;
    }

    /* Loop on SCAs looking for one not already found */
    for (sca_index = first_active_sca_index + 1; 
        sca_index < polygon->number_scas; sca_index++)
    {
        for (unique_index = 0; unique_index < *number_unique_scas;
            unique_index++)
        {
            /* If the current SCA is found in the list of unique indexes,
               stop looking for a match */
            if (polygon->sca[sca_index] == unique_scas[unique_index])
                break;
        }

        /* If the current SCA was not found in the list of unique indexes */ 
        if (unique_index == *number_unique_scas)
        {
            /* Add the current SCA to the list of unique SCAs, and increment
               the number of unique SCAs */
            unique_scas[*number_unique_scas] = polygon->sca[sca_index];
            (*number_unique_scas)++;
        }
    } /* SCA loop */
}

int generate_report_for_target 
(
    const char *l1r_file,          /* I: Level 1R data file name */
    IAS_GRID_TYPE *grid,           /* I: Grid structure */
    IAS_LOS_MODEL *model,          /* I: Model structure */
    int band_number,               /* I: Band number to process */
    int number_scas,               /* I: Number of SCAs for OLI */
    IAS_CPF *cpf,                  /* I: CPF constants */
    int number_targets,            /* I: Number of targets */
    MTF_TARGET *targets            /* I: Crossover target list */
)
{
    IAS_GRID_BAND_TYPE *grid_band_pointer; /* Grid band pointer */
    int band_index;                /* Band index for band number */
    int number_grid_lines;         /* Number of grid lines */
    int number_grid_samples;       /* Number of grid samples */
    int input_number_lines;        /* Number of lines in L1R */
    int input_number_samples;      /* Number of samples in L1R */
    IAS_UI2_TYPE *image_data;      /* Image data buffer */
    int unique_scas[IAS_MAX_NSCAS]; /* Unique SCAs */
    int number_unique_scas;        /* Number of unique SCAs */
    int unique_sca_count;          /* Unique SCA loop counter */
    int current_target;            /* Current target */
    int current_polygon;           /* Current polygon */
    double elevation = 0.0;        /* Elevation for L1R location */

    /* Set grid pointer */
    band_index = ias_sat_attr_convert_band_number_to_index(band_number);
    if (band_index == ERROR)
    {
        IAS_LOG_ERROR("Invalid band index for band number %d", band_number);
        return ERROR;
    }

    /* Verify that the grid has the band */
    if (!grid->bands_available[band_index])
    {
        IAS_LOG_ERROR("Band index %d not present in the grid", band_index);
        return ERROR;
    }

    grid_band_pointer = &grid->gridbands[band_index][0];

    /* Extract grid's input number of lines and samples */
    number_grid_lines = grid_band_pointer->ngrid_lines - 1;
    number_grid_samples = grid_band_pointer->ngrid_samps - 1;
    input_number_lines = grid_band_pointer->in_lines[number_grid_lines];
    input_number_samples = grid_band_pointer->in_samps[number_grid_samples];

    /* Allocate buffer for imagery */
    image_data = (IAS_UI2_TYPE *)malloc(sizeof(IAS_UI2_TYPE)
        * input_number_lines * input_number_samples);
    if (image_data == NULL)
    {
        IAS_LOG_ERROR("Allocating image data buffer");
        return ERROR;
    }

    for (current_target = 0; current_target < number_targets; current_target++)
    {
        elevation = targets[current_target].height;
        for (current_polygon = 0; current_polygon < targets[current_target].
            number_polygons; current_polygon++)
        {
            /* Get the SCAs for the polygon */
            get_unique_sca(&targets[current_target].polygons[current_polygon],
                unique_scas, &number_unique_scas);
            
            if (number_unique_scas > 1)
            {
                if (number_unique_scas > 2)
                {
                    IAS_LOG_ERROR("Target %s polygon %d spans %d SCAs (it "
                        "should not span more than two SCAs)", 
                        targets[current_target].name, current_polygon, 
                        number_unique_scas);
                    free(image_data);
                    return ERROR;
                }

                IAS_LOG_INFO("Target %s polygon %d crosses multiple SCAs: "
                    "%d,%d", targets[current_target].name, current_polygon,
                    unique_scas[0], unique_scas[1]);
            }

            /* Loop over all unique SCAs */
            for (unique_sca_count = 0; unique_sca_count < number_unique_scas;
                unique_sca_count++)
            {
    
                if (generate_report_for_sca_target(
                    unique_scas[unique_sca_count], l1r_file, band_number,
                    number_scas, cpf, current_target, current_polygon,
                    &targets[current_target], model, grid, input_number_lines,
                    input_number_samples, elevation, unique_sca_count,
                    image_data) != SUCCESS)
                {
                    IAS_LOG_ERROR("Could not get target for SCA %d",
                        unique_sca_count);
                    free(image_data);
                    return ERROR;
                }
            }
        }
    } /* Target loop */

    /* Free the image data and list of unique SCAs */  
    free(image_data);

    return SUCCESS;
}

