/******************************************************************************
NAME:           get_parameters

PURPOSE:
Get the mtfextract parameters from the input parmaeter file and OMF file.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
SUCCESS         Successful completion
ERROR           Operation failed 

******************************************************************************/
#include <string.h> 
#include <limits.h> 
#include <stdlib.h> 

#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "mtfextract.h"

int get_parameters 
(
    const char parm_file[PATH_MAX], /* I: Name of the parameter file */
    char input_directory[PATH_MAX], /* O: Directory for MTF-specific input
                                         files */
    char omf_file[PATH_MAX],       /* O: Name of of the OMF */
    char grid_file[PATH_MAX],      /* O: Grid file name */
    char model_file[PATH_MAX],     /* O: Model file name */
    char l0r_file[PATH_MAX],       /* O: L0R file name */
    char l1r_file[PATH_MAX],       /* O: L1R file name */
    char cpf_file[PATH_MAX],       /* O: CPF file name */ 
    int *target_wrs_path,          /* O: WRS target path */
    int *target_wrs_row,           /* O: WRS target row */
    int band_numbers[IAS_MAX_NBANDS], /* O: Band numbers to process */
    int *number_bands              /* O: Number of bands to process */
)
{
    int count;                     /* Number of elements */
    int size;                      /* Size of string */
    char work_order_id[IAS_WORKORDER_ID_SIZE]; /* Work order ID */
    int processing_pass = 2;       /* Pass through processing flow algorithms */
    int band_number_list[IAS_MAX_NBANDS]; /* List of band numbers from
                                      parameter file */
    int band_number_list_count;    /* Count of bands read from parameter file */
    int oli_band_numbers[IAS_MAX_NBANDS]; /* List of OLI band numbers */
    int number_oli_bands;          /* Number of OLI bands */
    int oli_band_index;            /* Loop counter */
    int work_order_band_index;     /* Loop counter */
    IAS_SATELLITE_ID satellite_id; /* Satellite Id */
    char *ias_data_dir;            /* IAS_DATA_DIR environment variable value */
    int status;                    /* Status returned from functions  */

    /* Local variables for storing ODL values */
    char local_input_directory[PATH_MAX];
    char local_grid_file[PATH_MAX];
    char local_model_file[PATH_MAX];
    char local_l0r_file[PATH_MAX];
    char local_l1r_file[PATH_MAX];
    char local_cpf_file[PATH_MAX];
    int local_satellite_number;

    /* Initialize to empty string */
    strcpy(work_order_id, "");
    strcpy(local_input_directory, "");
    strcpy(local_grid_file, "");
    strcpy(local_model_file, "");
    strcpy(local_l0r_file, "");
    strcpy(local_l1r_file, "");
    strcpy(local_cpf_file, "");

    /* Start the parameter file table */
    IAS_PARM_DECLARE_TABLE(parms, 4);

    /* Get the work order id */
    IAS_PARM_WORK_ORDER_ID(parms, work_order_id, sizeof(work_order_id), 0);

    /* Get the input directory name */
    IAS_PARM_MTF_INPUT_DIRECTORY(parms, local_input_directory, 
        sizeof(local_input_directory));

    /* Get the processing pass */
    IAS_PARM_PROCESSING_PASS(parms, &processing_pass, sizeof(processing_pass), 
        1);

    /* Get the band numbers to process */
    IAS_PARM_BAND_LIST(parms, IAS_PARM_REQUIRED, band_number_list,
        sizeof(band_number_list), 1);

    count = IAS_PARM_GET_TABLE_SIZE(parms);

    /* Check if help or template output was requested */
    status = ias_parm_provide_help(parm_file, parms, count,
                                   IAS_INPUT_PARAMETERS);

    if (status == 1)
    {
        /* --help or --template was requested.  Copy the requested option to
           omf_file so the same operation will be performed for the OMF */
        strcpy(omf_file, parm_file);
    }
    else if (status == ERROR) /* Help failed. */
    {
        IAS_LOG_ERROR("Providing help for application parameter file %s",
            parm_file);
        exit(EXIT_FAILURE);
    }
    else
    {
        if (ias_parm_read(parm_file, parms, count) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading parameters from: %s", parm_file);
            return ERROR;
        }

        /* Verify parameters */
        if (strcmp(local_input_directory, "") == 0)
        {
            /* Get IAS_DATA_DIR environment variable */
            ias_data_dir = getenv("IAS_DATA_DIR");
            if (ias_data_dir == NULL)
            {
                IAS_LOG_ERROR("Could not get environment variable "
                    "IAS_DATA_DIR.");
                return ERROR;
            }

            /* Build and use the default MTF input directory */
            size = snprintf(local_input_directory, 
                sizeof(local_input_directory), "%s/sysdata/mtf", ias_data_dir);
            if (size < 0 || size >= sizeof(local_input_directory))
            {
                IAS_LOG_ERROR("Building local_input_directory.");
                return ERROR;
            }
        }
        strcpy(input_directory, local_input_directory);

        if (strcmp(work_order_id, "") == 0)
        {
            IAS_LOG_ERROR("Missing work order id in the parameter file.");
            return ERROR;
        }

        /* Build the OMF file name */
        sprintf(omf_file, "%s.omf", work_order_id);
    }


    /***** Read additional parameters from OMF *****/

    /* Start the OMF table */
    IAS_PARM_DECLARE_TABLE(omf_parms, 8);

    /* Get the satellite number */
    IAS_PARM_SATELLITE(omf_parms, &local_satellite_number,
        sizeof(local_satellite_number), 1);

    /* Get the grid file name */
    IAS_PARM_GRID_FILENAME_FROM_PASS(omf_parms, IAS_PARM_REQUIRED,
        local_grid_file, sizeof(local_grid_file), 1, processing_pass);

    /* Get the model file name */
    IAS_PARM_MODEL_FILENAME_FROM_PASS(omf_parms, IAS_PARM_REQUIRED,
        local_model_file, sizeof(local_model_file), 1, processing_pass);

    /* Get the L0R file name */
    IAS_PARM_L0R_FILENAME(omf_parms, IAS_PARM_REQUIRED, local_l0r_file,
            sizeof(local_l0r_file), 1);

    /* Get the L1R file name */
    IAS_PARM_L1R_FILENAME(omf_parms, IAS_PARM_REQUIRED, local_l1r_file,
            sizeof(local_l1r_file), 1);

    /* Get the CPF file name */
    IAS_PARM_CPF(omf_parms, IAS_PARM_REQUIRED, local_cpf_file,
            sizeof(local_cpf_file), 1);

    /* Get the WRS path */
    IAS_PARM_TARGET_WRS_PATH(omf_parms, IAS_PARM_REQUIRED, target_wrs_path,
            sizeof(*target_wrs_path), 1);

    /* Get the WRS row */
    IAS_PARM_TARGET_WRS_ROW(omf_parms, IAS_PARM_REQUIRED, target_wrs_row,
            sizeof(*target_wrs_row), 1);

    count = IAS_PARM_GET_TABLE_SIZE(omf_parms);

    /* Check if help or template output is being requested */
    status = ias_parm_provide_help(omf_file, omf_parms, count,
        IAS_OMF_PARAMETERS);

    if (status == 1)
    {
        /* the --help or --template was provided. can now exit success */
        exit(EXIT_SUCCESS);
    }
    else if (status == ERROR) /* Help failed */
    {
        IAS_LOG_ERROR("Providing help for OMF file: %s", omf_file);
        exit(EXIT_FAILURE);
    }

    /* Read the OMF file */
    if (ias_parm_read(omf_file, omf_parms, count) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from OMF file: %s", omf_file);
        return ERROR;
    }

    /* Copy OMF parameters */
    strcpy(grid_file, local_grid_file);
    strcpy(model_file, local_model_file);
    strcpy(l0r_file, local_l0r_file);
    strcpy(l1r_file, local_l1r_file);
    strcpy(cpf_file, local_cpf_file);

    /* Convert the satellite number to the satellite id */
    satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(
        local_satellite_number);
    if (satellite_id == ERROR)
    {
        IAS_LOG_ERROR("Unable to determine satellite ID from satellite "
            "number %d", local_satellite_number);
        return ERROR;
    }

    /* Initialize the satellite attributes */
    if (ias_sat_attr_initialize(satellite_id) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attributes library");
        return ERROR;
    }

    /* Make sure the band list only includes OLI bands. First get the list of
       OLI bands  */
    status = ias_sat_attr_get_sensor_band_numbers(IAS_OLI, IAS_NORMAL_BAND, 0,
        oli_band_numbers, IAS_MAX_NBANDS, &number_oli_bands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting OLI band numbers");
        return ERROR;
    }

    /* Get the number of bands read from parameter file */
    band_number_list_count = IAS_PARM_GET_COUNT_READ(parms, BAND_LIST);

    /* Compare the list of OLI bands to the work order list of bands, and only
       accept work order bands that are OLI bands into the band number list */ 
    *number_bands = 0;
    for (oli_band_index = 0; oli_band_index < number_oli_bands;
        oli_band_index++)
    {
        for (work_order_band_index = 0;
            work_order_band_index < band_number_list_count;
            work_order_band_index++)
        {
            if (band_number_list[work_order_band_index] ==
                oli_band_numbers[oli_band_index])
            {
                band_numbers[*number_bands]
                    = band_number_list[work_order_band_index];
                (*number_bands)++;
                break;
            }
        }
    }

    return SUCCESS;
}

