/******************************************************************************
NAME:           get_target_descriptions

PURPOSE:
Read target information.  The information is target name(s) and corresponding
polygon information.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
SUCCESS         Successful completion
ERROR           Operation failed 

******************************************************************************/
#include <limits.h>                /* PATH_MAX */
#include <stdio.h>                 /* snprintf */
#include <stdlib.h>                /* malloc */
#include <math.h>                  /* atan2 */

#include "ias_const.h"             /* IAS constants */
#include "ias_logging.h"           /* IAS logging library */
#include "ias_parm.h"              /* IAS parameter file library */
#include "ias_odl.h"               /* IAS ODL library */
#include "mtf.h"                   /* Shared MTF definitions */
#include "mtfextract.h"            /* MTF Extract library */

#define NUMBER_TARGET_COORDINATES 4 /* Number of target centerline 
                                      coordinates */
#define MAX_FIELD_SIZE 256          /* Maximum supported size of ODL fields in 
                                       Target Description File */

int get_target_descriptions 
(
    const char *input_directory,   /* I: Input directory for MTF-specific 
                                         files */
    int target_wrs_path,           /* I: WRS target path number for scene */
    int target_wrs_row,            /* I: WRS target row number for scene */
    int number_scas,               /* I: Number of SCAs for OLI */
    int *number_targets,           /* O: Number of targets in scene */
    MTF_TARGET **return_targets,   /* O: Details about the targets */
    int *utm_zone                  /* O: Target UTM zone */
)
{
    char target_description_file[PATH_MAX]; /* ODL-format target description
                                      file */
    int current_target;            /* Loop counter */
    int current_polygon;           /* Loop counter */
    int current_vertex;            /* Loop counter */
    int current_vertex_value;      /* Loop counter */
    MTF_TARGET *targets;           /* Pointer to MTF targets structure */
    char field_name[MAX_FIELD_SIZE]; /* ODL field to read */
    char prefix[MAX_FIELD_SIZE - 5];   /* Beginning of ODL field to read */
    double target_coordinates[NUMBER_TARGET_COORDINATES]; /* Target centerline
                                      coordinates */
    int count;                     /* Number of elements */
    IAS_OBJ_DESC *odlFile;         /* ODL tree structure */
    double azimuth;                /* Target UTM azimuth */
    int number_polygons;           /* Number of polygons for current target */
    int number_vertices;           /* Number of vertices for current polygon */
    int scas_times_vertices;       /* Number of SCAs times number of vertices */
    double *vertices;              /* Vertex values */
    int maximum_vertices;          /* Maximum number of vertices in all
                                      polygons */
    double height;                 /* Target height */
    int status;                    /* Function status return value */

    status = snprintf(target_description_file, sizeof(target_description_file),
        "%s/MTF_Targets_%03d%03d.table", input_directory, target_wrs_path, 
        target_wrs_row);
    if (status < 0 || status >= sizeof(target_description_file))
    {
        IAS_LOG_ERROR("Constructing the target description file name");
        return ERROR;
    }

    /* Read ODL-format target description file */
    odlFile = ias_odl_read_tree(target_description_file);
    if (odlFile == NULL)
    {
        IAS_LOG_ERROR("Reading parameters from %s", target_description_file);
        return ERROR;
    }

    /* Get the target UTM zone */
    status = ias_odl_get_field(utm_zone, sizeof(int), IAS_ODL_Int, odlFile,
        NULL, "UTM_Zone", &count);
    if (status != SUCCESS)
    {
        process_odl_error(status, "UTM_Zone", target_description_file);
        ias_odl_free_tree(odlFile);
        return ERROR;
    }

    /* Get number of targets */
    status = ias_odl_get_field(number_targets, sizeof(*number_targets), 
        IAS_ODL_Int, odlFile, NULL, "Number_of_Targets", &count);
    if (status != SUCCESS)
    {
        process_odl_error(status, "Number_of_Targets", 
            target_description_file);
        ias_odl_free_tree(odlFile);
        return ERROR;
    }

    /* Allocate memory for the targets and set it to 0 */
    targets = (MTF_TARGET *)calloc(*number_targets, sizeof(MTF_TARGET)); 
    if (targets == NULL)
    {
        IAS_LOG_ERROR("Allocating MTF target structure");
        ias_odl_free_tree(odlFile);
        return ERROR;
    }

    /* Loop over all targets, reading their values */
    maximum_vertices = 0;
    for (current_target = 0; current_target < (*number_targets); 
        current_target++)
    {
        /* Set up the target name prefix used for the field names */
        status = snprintf(prefix, sizeof(prefix), "Target%d", current_target);
        if (status < 0 || status >= sizeof(prefix))
        {
            IAS_LOG_ERROR("Constructing the target %d name prefix", 
                current_target);
            ias_odl_free_tree(odlFile);
            free_polygons(targets, current_target + 1);
            return ERROR;
        }
        
        /* Set up the target name field name */
        status = snprintf(field_name, sizeof(field_name), "%s_Name", prefix);
        if (status < 0 || status >= sizeof(field_name))
        {
            IAS_LOG_ERROR("Constructing the target %d name field name", 
                current_target);
            ias_odl_free_tree(odlFile);
            free_polygons(targets, current_target + 1);
            return ERROR;
        }

        /* Get the target name */
        status = ias_odl_get_field(targets[current_target].name, 
            sizeof(targets[current_target].name), IAS_ODL_String, odlFile,
            NULL, field_name, &count);
        if (status != SUCCESS)
        {
            process_odl_error(status, field_name, target_description_file);
            ias_odl_free_tree(odlFile);
            free_polygons(targets, current_target + 1);
            return ERROR;
        }

        /* Set up the target centerline coordinates field name */
        status = snprintf(field_name, sizeof(field_name), "%s_Coordinates", 
            prefix);
        if (status < 0 || status >= sizeof(field_name))
        {
            IAS_LOG_ERROR("Constructing the target centerline coordinates "
                "field name for target %s", targets[current_target].name);
            ias_odl_free_tree(odlFile);
            free_polygons(targets, current_target + 1);
            return ERROR;
        }

        /* Get the target centerline coordinates */
        status = ias_odl_get_field(target_coordinates, 
            sizeof(target_coordinates), IAS_ODL_Double, odlFile, NULL, 
            field_name, &count);
        if (status != SUCCESS)
        {
            process_odl_error(status, field_name, target_description_file);
            ias_odl_free_tree(odlFile);
            free_polygons(targets, current_target + 1);
            return ERROR;
        }

        if (count != NUMBER_TARGET_COORDINATES)
        {
            IAS_LOG_ERROR("In number of coordinates read %d (should be %d) "
                "from %s for target %s in file %s", count, 
                NUMBER_TARGET_COORDINATES, field_name, 
                targets[current_target].name, target_description_file);
            ias_odl_free_tree(odlFile);
            free_polygons(targets, current_target + 1);
            return ERROR;
        }

        /* Calculate and save the target grid azimuth */
        azimuth = atan2(target_coordinates[0] - target_coordinates[2], 
            target_coordinates[1] - target_coordinates[3]);
        targets[current_target].angle = azimuth;

        /* Set up the target elevation field name */
        status = snprintf(field_name, sizeof(field_name), "%s_Height", prefix);
        if (status < 0 || status >= sizeof(field_name))
        {
            IAS_LOG_ERROR("Constructing the target %s elevation field name",
                targets[current_target].name);
            ias_odl_free_tree(odlFile);
            free_polygons(targets, current_target + 1);
            return ERROR;
        }

        /* Get the target elevation */
        status = ias_odl_get_field(&height, sizeof(double), IAS_ODL_Double, 
            odlFile, NULL, field_name, &count); 
        if (status != SUCCESS)
        {
            IAS_LOG_WARNING("Target height value not found for target %s; "
                "setting it to 0.0", targets[current_target].name);
            targets[current_target].height = 0.0;
        }
        else
        {
            targets[current_target].height = height;
        }

        /* Set up the target number of polygons field name */
        status = snprintf(field_name, sizeof(field_name), "%s_Number_Polygons",
            prefix);
        if (status < 0 || status >= sizeof(field_name))
        {
            IAS_LOG_ERROR("Constructing the target %s number of polygons "
                "field name", targets[current_target].name);
            ias_odl_free_tree(odlFile);
            free_polygons(targets, current_target + 1);
            return ERROR;
        }

        /* Get the target number of polygons */
        status = ias_odl_get_field(&number_polygons, sizeof(number_polygons), 
            IAS_ODL_Int, odlFile, NULL, field_name, &count);
        if (status != SUCCESS)
        {
            process_odl_error(status, field_name, target_description_file);
            ias_odl_free_tree(odlFile);
            free_polygons(targets, current_target + 1);
            return ERROR;
        }

        targets[current_target].number_polygons = number_polygons;

        /* Allocate memory for polygons and set it to 0 */
        targets[current_target].polygons = (MTF_POLYGON *)calloc(
            number_polygons, sizeof(MTF_POLYGON)); 
        if (targets[current_target].polygons == NULL)
        {
            IAS_LOG_ERROR("Allocating MTF target %s polygon structure", 
                targets[current_target].name);
            ias_odl_free_tree(odlFile);
            free_polygons(targets, current_target + 1);
            return ERROR;
        }

        /* Loop over all polygons for the target, reading their values */
        for (current_polygon = 0; current_polygon < number_polygons;
            current_polygon++)
        {
            /* Set up the number of polygon vertices field name */
            status = snprintf(field_name, sizeof(field_name),
                "%s_Polygon%d_Vertices", prefix, current_polygon);
            if (status < 0 || status >= sizeof(field_name))
            {
                IAS_LOG_ERROR("Constructing the number of polygon vertices "
                    "field name for target %s", targets[current_target].name);
                ias_odl_free_tree(odlFile);
                free_polygons(targets, current_target + 1);
                return ERROR;
            }

            /* Get the number of polygon vertices */
            status = ias_odl_get_field(&number_vertices, 
                sizeof(number_vertices), IAS_ODL_Int, odlFile, NULL, 
                field_name, &count);
            if (status != SUCCESS)
            {
                process_odl_error(status, field_name, target_description_file);
                ias_odl_free_tree(odlFile);
                free_polygons(targets, current_target + 1);
                return ERROR;
            }

            /* Keep track of the maximum number of vertices */
            if (number_vertices > maximum_vertices)
            {
                maximum_vertices = number_vertices;
            }

            targets[current_target].polygons[current_polygon].number_points
                = number_vertices;

            /* Allocate memory for for the polygon field utm_x */
            targets[current_target].polygons[current_polygon].utm_x =
                (double *)calloc(number_vertices, sizeof(double)); 
            if (targets[current_target].polygons[current_polygon].utm_x == NULL)
            {
                IAS_LOG_ERROR("Allocating MTF target %d polygon %d UTM X "
                    "structure", current_target, current_polygon);
                ias_odl_free_tree(odlFile);
                free_polygons(targets, current_target + 1);
                return ERROR;
            }

            /* Allocate memory for the polygon field utm_y */
            targets[current_target].polygons[current_polygon].utm_y =
                (double *)calloc(number_vertices, sizeof(double)); 
            if (targets[current_target].polygons[current_polygon].utm_y == NULL)
            {
                IAS_LOG_ERROR("Allocating MTF target %d polygon %d UTM Y "
                    "structure", current_target, current_polygon);
                ias_odl_free_tree(odlFile);
                free_polygons(targets, current_target + 1);
                return ERROR;
            }

            /* Multiply this once in case optimization doesn't do it */
            scas_times_vertices = number_scas * number_vertices;

            /* Allocate memory for the polygon field l1g_sample */
            targets[current_target].polygons[current_polygon].l1g_sample =
                (double *)calloc(number_vertices, sizeof(double)); 
            if (targets[current_target].polygons[current_polygon].l1g_sample
                == NULL)
            {
                IAS_LOG_ERROR("Allocating MTF target %d polygon %d L1G sample "
                    "structure", current_target, current_polygon);
                ias_odl_free_tree(odlFile);
                free_polygons(targets, current_target);
                return ERROR;
            }

            /* Allocate memory for the polygon field l1g_line */
            targets[current_target].polygons[current_polygon].l1g_line =
                (double *)calloc(number_vertices, sizeof(double));  
            if (targets[current_target].polygons[current_polygon].l1g_line 
                == NULL)
            {
                IAS_LOG_ERROR("Allocating MTF target %d polygon %d L1G line "
                    "structure", current_target, current_polygon);
                ias_odl_free_tree(odlFile);
                free_polygons(targets, current_target + 1);
                return ERROR;
            }

            /* Allocate memory for the polygon field l1r_sample */
            targets[current_target].polygons[current_polygon].l1r_sample =
                (double *)calloc(scas_times_vertices, sizeof(double)); 
            if (targets[current_target].polygons[current_polygon].l1r_sample
                == NULL)
            {
                IAS_LOG_ERROR("Allocating MTF target %d polygon %d L1R sample "
                    "structure", current_target, current_polygon);
                ias_odl_free_tree(odlFile);
                free_polygons(targets, current_target + 1);
                return ERROR;
            }

            /* Allocate memory for the polygon field l1r_line */
            targets[current_target].polygons[current_polygon].l1r_line =
                (double *)calloc(scas_times_vertices, sizeof(double)); 
            if (targets[current_target].polygons[current_polygon].l1r_line 
                == NULL)
            {
                IAS_LOG_ERROR("Allocating MTF target %d polygon %d L1R line "
                    "structure", current_target, current_polygon);
                ias_odl_free_tree(odlFile);
                free_polygons(targets, current_target + 1);
                return ERROR;
            }

            /* Allocate memory for the polygon field sca */
            targets[current_target].polygons[current_polygon].sca =
                (int *)calloc(scas_times_vertices, sizeof(int)); 
            if (targets[current_target].polygons[current_polygon].sca == NULL)
            {
                IAS_LOG_ERROR("Allocating MTF target %d polygon %d sca "
                    "structure", current_target, current_polygon);
                ias_odl_free_tree(odlFile);
                free_polygons(targets, current_target + 1);
                return ERROR;
            }
        } /* Polygon loop */
    } /* Target loop */

    /* Get the vertices */
    vertices = (double *)malloc(sizeof(double) * 2 * maximum_vertices);
    if (vertices == NULL)
    {
        IAS_LOG_ERROR("Allocating MTF vertex structure");
        ias_odl_free_tree(odlFile);
        free_polygons(targets, *number_targets);
        return ERROR;
    }

    for (current_target = 0; current_target < (*number_targets); 
        current_target++)
    {
        for (current_polygon = 0; current_polygon < 
            targets[current_target].number_polygons; current_polygon++)
        {
            /* Set up the vertex field name */
            status = snprintf(field_name, sizeof(field_name), 
                "Target%d_Polygon%d_Coordinates", current_target, 
                current_polygon);
            if (status < 0 || status >= sizeof(field_name))
            {
                IAS_LOG_ERROR("Constructing the vertex field name");
                ias_odl_free_tree(odlFile);
                free(vertices);
                free_polygons(targets, *number_targets); 
                return ERROR;
            }

            /* Get the vertices */
            number_vertices = targets[current_target].polygons[
                current_polygon].number_points;
            status = ias_odl_get_field(vertices, 2 * number_vertices *
                sizeof(double), IAS_ODL_Double, odlFile, NULL, field_name,
                &count);
            if (status != SUCCESS)
            {
                process_odl_error(status, field_name, target_description_file);
                ias_odl_free_tree(odlFile);
                free(vertices);
                free_polygons(targets, *number_targets); 
                return ERROR;
            }

            /* Make sure there are 2 point values (x and y) for each vertex */
            if (count != (2 * number_vertices))
            {
                IAS_LOG_ERROR("In number of vertices read %d (should be %d) "
                    "from %s in file %s", count, 2 * number_vertices, 
                    field_name, target_description_file);
                ias_odl_free_tree(odlFile);
                free(vertices);
                free_polygons(targets, *number_targets); 
                return ERROR;
            }

            /* Save the vertex values.  They are read as point0_x, point0_y,
               ... pointn_x, pointn_y.  current_vertex is the counter for the
               vertices (each with 2 values), and current_vertex_value is the
               counter for each value */
            for (current_vertex = 0, current_vertex_value = 0; 
                 current_vertex_value < count; 
                 current_vertex++, current_vertex_value += 2)
            {
                targets[current_target].polygons[current_polygon].
                    utm_x[current_vertex] = vertices[current_vertex_value];
                targets[current_target].polygons[current_polygon].
                    utm_y[current_vertex] = vertices[current_vertex_value + 1];
            }
        } /* Polygon loop */
    } /* Target loop */

    /* Allow the MTF targets data to be returned */
    *return_targets = targets;

    /* Free the memory allocated for the ODL object */
    free(vertices);
    ias_odl_free_tree(odlFile);

    return SUCCESS;
}

