/******************************************************************************
NAME:           get_utm

PURPOSE:
Determine UTM MTF table parameters for a given pixel.  Also determine its
scaled range, ground velocity, and rotation angle.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
SUCCESS         Successful operation
ERROR           Operation failed 

******************************************************************************/
#include <math.h>

#include "ias_grid.h"
#include "ias_grid_io.h"
#include "ias_model_io.h"
#include "ias_geo.h"
#include "ias_math.h"
#include "ias_logging.h"
#include "mtfextract.h"

static IAS_GEO_PROJ_TRANSFORMATION *geo_to_target_projection = NULL;
    /* Projection transformation from latitude/longitude to target */

/* Get the projection transformation */
const IAS_GEO_PROJ_TRANSFORMATION *get_proj_transform_geo2target()
{
    return geo_to_target_projection;
}

/* Initialize the projection transformation */
int set_geo_to_target_projection
(
    IAS_PROJECTION *source_projection, /* I: source projection */
    IAS_PROJECTION *target_projection  /* I: target projection */
)
{
    geo_to_target_projection = ias_geo_create_proj_transformation(
        source_projection, target_projection);
    if (!geo_to_target_projection)
    {
        return ERROR;
    }

    return SUCCESS;
}

/* Free the projection transformation */
void free_geo_to_target_projection()
{
    ias_geo_destroy_proj_transformation(geo_to_target_projection);
    geo_to_target_projection = NULL;
}

int get_utm 
(
    const IAS_LOS_MODEL *model,    /* I: Model structure */
    int band_number,               /* I: Band number to process */
    int current_sca,               /* I: Current SCA */
    double elevation,              /* I: Elevation */
    double l1r_line,               /* I: L1r line location */
    double l1r_sample,             /* I: L1r sample location */
    double target_angle,           /* I: Target azimuth (in radians) */
    double *distance,              /* O: Distance from satellite to target */
    double *rotation,              /* O: Rotation angle */
    double *ground_velocity,       /* O: Ground velocity in meters/second */
    double *utm_y,                 /* O: UTM Y coordinate */
    double *utm_x                  /* O: UTM X coordinate */
)
{
    IAS_SENSOR_DETECTOR_TYPE detector_type = IAS_EXACT_DETECTOR; /* Detector 
                                      type */
    double Time0;                  /* Time for the current pixel */
    double Time1;                  /* Time for current pixel minus one line */
    IAS_VECTOR satellite_position; /* Satellite position */
    double input_latitude;         /* Ground point latitude */
    double input_longitude;        /* Ground point longitude */
    int band_index;                /* Index for band_number */
    double projection_x;           /* Projection X */
    double projection_y;           /* Projection Y */
    double utm_x0;                 /* UTM X coordinate for current point */
    double utm_y0;                 /* UTM Y coordinate for current point */
    double utm_x1;                 /* UTM X coordinate for -1 along track */
    double utm_y1;                 /* UTM Y coordinate for -1 along track */
    double geocentric_latitude;    /* Geocentric latitude */
    double radius;                 /* Radius of coordinate */
    IAS_VECTOR utm_position;       /* UTM position */
    IAS_VECTOR line_of_sight;      /* Vector from satellite to pixel */
    double cos_off_boresight_angle; /* Cosine of the off-boresight angle */
    double cos_zenith_normal;      /* Cosine of zenith normal to target */
    double vector_magnitude;       /* Magnitude of vector */
    IAS_VECTOR satellite_velocity; /* Dummy parameter - not used */
    IAS_VECTOR satellite_attitude; /* Dummy parameter - not used */
    int year;                      /* Dummy parameter - not used */
    int day;                       /* Dummy parameter - not used */
    double utc_time_at_location[3]; /* UTC time (year, doy, and sod) */

    /* Get the band index for band_number */
    band_index = ias_sat_attr_convert_band_number_to_index(band_number);
    if (band_index == ERROR)
    {
        IAS_LOG_ERROR("Invalid band index for band number %d", band_number);
        return ERROR;
    }

    /* Find geographic coordinate for the current point, one meter higher */
    if (ias_los_model_input_line_samp_to_geodetic(l1r_line, l1r_sample, 
        band_index, current_sca, elevation + 1.0, model, detector_type, NULL, 
        &input_latitude, &input_longitude) != SUCCESS)
    {
        IAS_LOG_ERROR("In the ias_los_model_input_line_samp_to_geodetic call "
            "for line %f sample %f band number %d SCA %d elevation %f", 
            l1r_line, l1r_sample, band_number, current_sca, elevation + 1.0); 
        return ERROR;
    }

    /* Project to map coordinate */
    if (ias_geo_transform_coordinate(geo_to_target_projection, input_longitude,
        input_latitude, &projection_x, &projection_y))
    {
        IAS_LOG_ERROR("Projecting location of latitude %f longitude %f for "
            "band number %d SCA %d", input_latitude, input_longitude, 
            band_number, current_sca); 
        return ERROR;
    }

    /* Save projection x and y values */
    utm_x1 = projection_x;
    utm_y1 = projection_y;

    /* Find geographic coordinate for the current point */
    if (ias_los_model_input_line_samp_to_geodetic(l1r_line, l1r_sample, 
        band_index, current_sca, elevation, model, detector_type, NULL, 
        &input_latitude, &input_longitude) != SUCCESS)
    {
        IAS_LOG_ERROR("In the ias_los_model_input_line_samp_to_geodetic call "
            "for line %f sample %f band number %d SCA %d elevation %f", 
            l1r_line, l1r_sample, band_number, current_sca, elevation); 
        return ERROR;
    }

    /* Find satellite position and time for the current point */
    if (ias_los_model_get_satellite_state_vector_at_location(l1r_line, 
        l1r_sample, band_index, current_sca, detector_type, model, 
        &satellite_position, &satellite_velocity, &satellite_attitude,
        &Time0, &year, &day) != SUCCESS)
    {
        IAS_LOG_ERROR("Finding satellite position and time for line %f sample "
            "%f band number %d SCA %d", l1r_line, l1r_sample, band_number,
            current_sca); 
        return ERROR;
    }

    /* Adjust time */
    Time0 -= model->sensor.bands[band_index].utc_epoch_time[2];

    /* Convert pixel geodetic coordinate to geocentric */
    if (ias_geo_convert_geodetic_to_geocentric_height(input_latitude,
        elevation, &model->earth, &geocentric_latitude, &radius) != SUCCESS)
    {
        IAS_LOG_ERROR("Determining geocentric latitude and radius given "
            "geodetic latitude %f and elevation %f", input_latitude, elevation);
        return ERROR;
    }

    /* Convert pixel spherical coordinates to Cartesian coordinates */
    ias_geo_convert_sph2cart(geocentric_latitude, input_longitude, 
        radius, &utm_position);

    /* Compute the sensor line of sight */
    if (ias_sensor_find_los_vector(current_sca, l1r_sample, detector_type,
        &model->sensor.bands[band_index], &line_of_sight) != SUCCESS)
    {
        IAS_LOG_ERROR("Determining line of sight vector for SCA %d sample %f",
            current_sca, l1r_sample);
        return ERROR;
    }

    /* Intialize cosine of off-boresight angle */
    cos_off_boresight_angle = line_of_sight.z;

    /* Determine distance between pixel and satellite */ 
    line_of_sight.x = utm_position.x - satellite_position.x;
    line_of_sight.y = utm_position.y - satellite_position.y;
    line_of_sight.z = utm_position.z - satellite_position.z;
    vector_magnitude = ias_math_compute_3dvec_dot(&line_of_sight, 
        &line_of_sight);
    vector_magnitude = sqrt(vector_magnitude);

    /* Project to map coordinate */
    if (ias_geo_transform_coordinate(geo_to_target_projection, input_longitude,
        input_latitude, &projection_x, &projection_y))
    {
        IAS_LOG_ERROR("Projecting location of latitude %f longitude %f for "
            "band number %d SCA %d", input_latitude, input_longitude, 
            band_number, current_sca); 
        return ERROR;
    }

    /* Save projection x and y values */
    utm_x0 = projection_x;
    utm_y0 = projection_y;

    /* Calculate the cross-target zenith angle */
    utm_x1 -= utm_x0; /* utm_x1 is dX/dh */
    utm_y1 -= utm_y0; /* utm_y1 is dY/dh */
    cos_zenith_normal = cos(atan(utm_x1 * cos(target_angle) 
        - utm_y1 * sin(target_angle)));
    cos_zenith_normal = cos(acos(cos_zenith_normal) 
        - acos(cos_off_boresight_angle));
    *distance = vector_magnitude * cos_off_boresight_angle / cos_zenith_normal;

    /* Find geographic coordinate for the current point minus 1 */
    if (ias_los_model_input_line_samp_to_geodetic(l1r_line - 1.0, l1r_sample, 
        band_index, current_sca, elevation, model, detector_type, NULL, 
        &input_latitude, &input_longitude) != SUCCESS)
    {
        IAS_LOG_ERROR("In the ias_los_model_input_line_samp_to_geodetic call "
            "for line %f sample %f band number %d SCA %d elevation %f", 
            l1r_line - 1.0, l1r_sample, band_number, current_sca, elevation); 
        return ERROR;
    }

    /* Find satellite time for the current point minus 1 */
    if (ias_los_model_get_utc_time_at_location(model, l1r_line - 1.0, 
        l1r_sample, band_index, current_sca, detector_type, 
        utc_time_at_location) != SUCCESS)
    {
        IAS_LOG_ERROR("Finding satellite time for line %f sample %f band "
            "number %d SCA %d", l1r_line - 1.0, l1r_sample, band_number,
            current_sca); 
        return ERROR;
    }

    /* Extract the time */
    Time1 = utc_time_at_location[2];

    /* Adjust time */
    Time1 -= model->sensor.bands[band_index].utc_epoch_time[2];

    /* Project to map coordinate */
    if (ias_geo_transform_coordinate(geo_to_target_projection, input_longitude,
        input_latitude, &projection_x, &projection_y))
    {
        IAS_LOG_ERROR("Projecting location of latitude %f longitude %f for "
            "band number %d SCA %d", input_latitude, input_longitude, 
            band_number, current_sca); 
        return ERROR;
    }

    /* Save projection x and y values */
    utm_x1 = projection_x;
    utm_y1 = projection_y;

    /* Find angle between map north and along track direction */
    *rotation = atan2((utm_x1 - utm_x0), (utm_y1 - utm_y0));

    /* Determine ground velocity */
    vector_magnitude = (sqrt(((utm_x0 - utm_x1) * (utm_x0 - utm_x1)) 
        + (utm_y0 - utm_y1) * (utm_y0 - utm_y1)));
    *ground_velocity = vector_magnitude / (Time0 - Time1);
    *utm_y = utm_y0;
    *utm_x = utm_x0;

    return SUCCESS;
}

