/******************************************************************************

NAME:           mtfextract

PURPOSE:
Main procedure for software to automatically locate and extract MTF target
windows predefined targets.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
EXIT_FAILURE    Error extracting MTF windows
EXIT_SUCCESS    Successfully extracted MTF windows

******************************************************************************/

#include <stdlib.h>         /* EXIT_FAILURE/EXIT_SUCCESS */
#include <limits.h>         /* PATH_MAX */
#include <string.h>         /* strncpy */

#include "gctp.h"
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_cpf.h"
#include "ias_grid.h"
#include "ias_grid_io.h"
#include "ias_model_io.h"
#include "ias_geo.h"
#include "mtf.h"
#include "mtfextract.h"

int main(int argc, char *argv[])
{
    char input_directory[PATH_MAX];   /* Location of MTF-specific input files
                                         like target description file */
    char omf_file[PATH_MAX];          /* Name of the the OMF */
    char grid_file[PATH_MAX];         /* Grid file name */
    char model_file[PATH_MAX];        /* Model file name */
    char l0r_file[PATH_MAX];          /* L0R file name */
    char l1r_file[PATH_MAX];          /* L1R file name */
    char cpf_file[PATH_MAX];          /* CPF file name */
    IAS_GRID_TYPE grid;               /* Geometric resampling grid structure */
    IAS_LOS_MODEL *model;             /* Model structure */
    IAS_CPF *cpf;                     /* CPF structure pointer */
    int number_oli_scas;              /* Number of SCAs in OLI bands */
    int number_bands;                 /* Number of bands to process */
    int band_numbers[IAS_MAX_NBANDS]; /* List of band numbers to process */
    int band_index;                   /* Index into band_numbers */
    int target_wrs_path = 0;          /* Target WRS path */
    int target_wrs_row = 0;           /* Target WRS row */
    int utm_zone = 0;                 /* Target definition UTM zone */
    int number_targets;               /* Number of targets in image */
    MTF_TARGET *targets;              /* Targets in image */
    int out_units;                    /* Code for the output projection units */
    double projection_parameters[IAS_PROJ_PARAM_SIZE]; /* Projection
                                         parameters */
    int parameter_index;              /* Projection parameter loop counter */
    IAS_PROJECTION geo_projection;    /* Source lat/lon projection */
    IAS_PROJECTION target_projection; /* Target projection */
    int status;                       /* Function return status */

    /* Initialize the logging library */
    if (ias_log_initialize("mtfextract") != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing logging library");
        return EXIT_FAILURE;
    }

    /* Verify a parameter file was provided */
    if (argc != 2)
    {
        IAS_LOG_ERROR("USAGE: mtfextract <parameter file name>");
        return EXIT_FAILURE;
    }

    /* Get parameters */
    if (get_parameters(argv[1], input_directory, omf_file, grid_file,
        model_file, l0r_file, l1r_file, cpf_file, &target_wrs_path, 
        &target_wrs_row, band_numbers, &number_bands) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from file: %s", argv[1]);
        return EXIT_FAILURE;
    }

    IAS_LOG_INFO("Initiating mtfextract");

    /* Get the number SCAs for OLI bands */
    number_oli_scas = ias_sat_attr_get_sensor_sca_count(IAS_OLI); 
    if (number_oli_scas == ERROR)
    {
        IAS_LOG_ERROR("Getting number of OLI SCAs");
        return EXIT_FAILURE;
    }

    /* Get the target descriptions */
    if (get_target_descriptions(input_directory, target_wrs_path, 
        target_wrs_row, number_oli_scas, &number_targets, &targets, &utm_zone) 
        != SUCCESS)
    {
        IAS_LOG_ERROR("Getting target descriptions");
        return EXIT_FAILURE;
    }

    /* Read the grid file */
    if (ias_grid_initialize(&grid, 0, 0) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the grid structure");
        free_polygons(targets, number_targets); 
        return EXIT_FAILURE;
    }

    if (ias_grid_header_read(grid_file, &grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading the grid from %s", grid_file);
        free_polygons(targets, number_targets); 
        return EXIT_FAILURE;
    }

    if (ias_grid_band_header_read(grid_file, &grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading the grid band header from %s", grid_file);
        free_polygons(targets, number_targets); 
        ias_grid_free(&grid);
        return EXIT_FAILURE;
    }

    /* Verify the target and grid file UTM zones match */
    if (utm_zone == 0)
    {
        IAS_LOG_WARNING("No target UTM zone provided; assume grid zone (%d) "
            "is correct", grid.zone);
    }
    else if (grid.zone == utm_zone)
    {
        IAS_LOG_INFO("Target and grid UTM zones (%d) match", utm_zone);
    }
    else
    {
        IAS_LOG_ERROR("Target (%d) and grid (%d) UTM zones do not match", 
            utm_zone, grid.zone);
        ias_grid_free(&grid);
        free_polygons(targets, number_targets); 
    }

    /* Set up projection transformations */
    for (parameter_index = 0; parameter_index < IAS_PROJ_PARAM_SIZE;
        parameter_index++)
    {
        projection_parameters[parameter_index] = 0.0;
    }

    /* Convert the grid projection units to a unit code for GCTP. */
    status = ias_geo_get_units(grid.units, &out_units);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Converting projection units %s to unit code",
            grid.units);
        ias_grid_free(&grid);
        free_polygons(targets, number_targets); 
        return EXIT_FAILURE;
    }

    /* Set up projections */
    ias_geo_set_projection(GEO, NULLZONE, RADIAN, grid.spheroid, 
        projection_parameters, &geo_projection);
    ias_geo_set_projection(grid.code, grid.zone, out_units, grid.spheroid,
        grid.projprms, &target_projection); 

    /* Set up projection transformation */
    status = set_geo_to_target_projection(&geo_projection, &target_projection);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Creating geo to target projection transformation");
        ias_grid_free(&grid);
        free_polygons(targets, number_targets); 
        return EXIT_FAILURE;
    }

    /* Read the model file */
    model = ias_model_read(model_file);
    if (model == NULL)
    {
        IAS_LOG_ERROR("Reading model file %s", model_file);
        ias_grid_free(&grid);
        free_geo_to_target_projection();
        free_polygons(targets, number_targets); 
        return EXIT_FAILURE;
    }

    /* Allocate the CPF data structure and read it */
    cpf = ias_cpf_read(cpf_file);
    if (cpf == NULL)
    {
        IAS_LOG_ERROR("Opening CPF %s", cpf_file);
        ias_grid_free(&grid);
        free_geo_to_target_projection();
        free_polygons(targets, number_targets); 
        return EXIT_FAILURE;
    }

    /* Loop on bands, processing all target information */
    for (band_index = 0; band_index < number_bands; band_index++)
    {
        IAS_LOG_DEBUG("Processing band %d %d\n", band_index, 
            band_numbers[band_index]);

        /* Load the grid information for the current band */
        if (ias_grid_band_pointers_read(grid_file, &grid,
            band_numbers[band_index]) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading the grid pointers from %s for band number "
                "%d", grid_file, band_numbers[band_index]);
            ias_cpf_free(cpf);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            model = NULL;
            free_geo_to_target_projection();
            free_polygons(targets, number_targets); 
            ias_grid_free(&grid);
        }

        /* Determine image L1R locations for target */
        if (ols_to_ils(&grid, band_numbers[band_index], number_oli_scas, 
            number_targets, targets) != SUCCESS)
        {
            IAS_LOG_ERROR("Getting target locations for band number %d", 
                band_numbers[band_index]);
            ias_cpf_free(cpf);
            ias_grid_free_band(&grid, band_numbers[band_index]);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            model = NULL;
            free_geo_to_target_projection();
            free_polygons(targets, number_targets); 
            return EXIT_FAILURE;
        }

        /* Extract L1R from imagery and write report */
        if (generate_report_for_target(l1r_file, &grid, model, 
            band_numbers[band_index], number_oli_scas, cpf, number_targets, 
            targets) != SUCCESS)
        { 
            IAS_LOG_ERROR("Extracting imagery for band number %d", 
                band_numbers[band_index]);
            ias_cpf_free(cpf);
            ias_grid_free_band(&grid, band_numbers[band_index]);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            model = NULL;
            free_geo_to_target_projection();
            free_polygons(targets, number_targets); 
            return EXIT_FAILURE;
        }

        /* Free grid data for the current band */
        if (ias_grid_free_band(&grid, band_numbers[band_index]) != SUCCESS)
        {
            IAS_LOG_ERROR("Freeing the grid data for band %d", 
                band_numbers[band_index]);
            ias_cpf_free(cpf);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            model = NULL;
            free_geo_to_target_projection();
            free_polygons(targets, number_targets); 
            return EXIT_FAILURE;
        }
    } /* Band loop */

    /* Free the CPF, grid, model, and projection transformation */
    ias_cpf_free(cpf);
    ias_grid_free(&grid);
    ias_los_model_free(model);
    model = NULL;
    free_geo_to_target_projection();
    free_polygons(targets, number_targets); 

    /* Update the OMF with the MTF input directory */ 
    if (update_omf(omf_file, input_directory) != SUCCESS)
    {
        IAS_LOG_ERROR("Updating OMF file");
        return EXIT_FAILURE;
    }

    /* Write completion message */
    IAS_LOG_INFO("Successful completion.");

    return EXIT_SUCCESS;
}

