#ifndef _MTFEXTRACT_H
#define _MTFEXTRACT_H

#include <limits.h>

#include "ias_satellite_attributes.h"
#include "ias_structures.h"
#include "ias_cpf.h"
#include "ias_grid.h"
#include "ias_model_io.h"
#include "ias_geo.h"
#include "mtf.h"

/* MTF_POLYGON elements are sized dynamically based on factors like the 
   polygon's number of vertices specified in the Target Description File */
typedef struct mtf_polygon
{
    double *utm_x;                 /* Target center line X locations,
                                      number_of_vertices entries */
    double *utm_y;                 /* Target center line Y locations,
                                      number_of_vertices entries */
    int number_scas;               /* Number of SCAs target resides in */
    int *sca;                      /* SCAs target resides in,
                                      number_of_vertices * number OLI SCAS
                                      entries */
    double *l1g_line;              /* Target L1G line locations,
                                      number_of_vertices entries */
    double *l1g_sample;            /* Target L1G sample locations,
                                      number_of_vertices entries */
    double *l1r_line;              /* L1R line location for UTM coordinates,
                                      number_of_vertices * number OLI SCAS
                                      entries */
    double *l1r_sample;            /* L1R sample location for UTM coordinates,
                                      number_of_vertices * number OLI SCAS
                                      entries */
    int number_points;             /* Number of points in polygon */
}  MTF_POLYGON;

typedef struct mtf_target
{
    char name[NAME_STRLEN];        /* Target name */
    double angle;                  /* Target azimuth */
    double height;                 /* Target elevation */
    int number_polygons;           /* Number of polygons */
    MTF_POLYGON *polygons;         /* Region of interest */
} MTF_TARGET;

/* Function prototypes */

int get_parameters
(
    const char parm_file[PATH_MAX], /* I: Name of the parameter File */
    char input_directory[PATH_MAX], /* O: Directory for MTF-specific input
                                         files */
    char omf_file[PATH_MAX],       /* O: Name of of the OMF */
    char grid_file[PATH_MAX],      /* O: Grid file name */
    char model_file[PATH_MAX],     /* O: Model file name */
    char l0r_file[PATH_MAX],       /* O: L0R file name */
    char l1r_file[PATH_MAX],       /* O: L1R file name */
    char cpf_file[PATH_MAX],       /* O: CPF file name */
    int *target_wrs_path,          /* O: WRS target path */
    int *target_wrs_row,           /* O: WRS target row */
    int band_numbers[IAS_MAX_NBANDS], /* O: Band numbers to process */
    int *number_bands              /* O: Number of bands to process */
);

int get_target_descriptions
(
    const char *input_directory,   /* I: Input directory for MTF-specific
                                         files */
    int target_wrs_path,           /* I: WRS target path number for scene */
    int target_wrs_row,            /* I: WRS target row number for scene */
    int number_scas,               /* I: Number of SCAs for OLI */
    int *number_targets,           /* O: Number of targets in scene */
    MTF_TARGET **return_targets,   /* O: Details about the targets */
    int *utm_zone                  /* O: Target definition UTM zone */
);

void free_polygons
(
    MTF_TARGET *targets,           /* I: Pointer to MTF targets structure */
    int current_target             /* I: Number of targets to free */
);

int ols_to_ils
(
    IAS_GRID_TYPE *grid,           /* I: Grid structure */
    int band_number,               /* I: Band number */
    int number_scas,               /* I: Number of SCAs */
    int number_targets,            /* I: Number of MTF targets */
    MTF_TARGET *targets            /* I/O: Structure of target information */
);

int generate_report_for_target
(
    const char *l1r_file,          /* I: Level 1R data file name */
    IAS_GRID_TYPE *grid,           /* I: Grid structure */
    IAS_LOS_MODEL *model,          /* I: Model structure */
    int band_number,               /* I: Band number to process */
    int number_scas,               /* I: Number of SCAs for OLI */
    IAS_CPF *cpf,                  /* I: CPF constants */
    int number_target,             /* I: Number of targets */
    MTF_TARGET *targets            /* I: Crossover target list */
);

void get_unique_sca
(
    MTF_POLYGON *polygon,          /* I: Target polygon */
    int *unique_scas,              /* O: SCAs target polygon is in */
    int *number_unique_scas        /* I/O: Number of unique points */
);

int read_l1r_band_sca
(
    const char *l1r_file,          /* I: Level 1R data file name */
    int input_band_number,         /* I: Band to read */
    int input_sca,                 /* I: SCA to read */
    int number_scas,               /* I: Number of SCAs for OLI */
    IAS_CPF *cpf,                  /* I: CPF constants */
    IAS_IMAGE *sca_image           /* O: L1R image for band/SCA */
);

int get_l1r_window 
(
    int starting_line,             /* I: Starting line within SCA to read */
    int starting_sample,           /* I: Starting sample within SCA to read */
    int number_read_lines,         /* I: Number of lines within SCA to read */
    int number_read_samples,       /* I: Number of samples within SCA to read */
    IAS_IMAGE *sca_image,          /* I: L1R image for band/SCA */
    IAS_UI2_TYPE *output_data      /* O: Image data */
);

void cleanup_l1r_image
(
    IAS_IMAGE *sca_image           /* I: L1R image for band/SCA */
);

int polygon_sca_bounds
(
    IAS_GRID_BAND_TYPE *grid_band_pointer, /* Grid band pointer */
    MTF_POLYGON *polygon,          /* I: Structure of target information */
    double *minimum_line,          /* I: Minimum line of boundary */
    double *minimum_sample,        /* I: Minimum sample of boundary */
    double *maximum_line,          /* I: Maximum line of boundary */
    double *maximum_sample         /* I: Maximum sample of boundary */
);

int get_utm
(
    const IAS_LOS_MODEL *model,    /* I: Model structure */
    int band_number,               /* I: Band number to process */
    int current_sca,               /* I: Current SCA */
    double elevation,              /* I: Elevation */
    double l1r_line,               /* I: L1r line location */
    double l1r_sample,             /* I: L1r sample location */
    double target_angle,           /* I: Target azimuth (in radians) */
    double *distance,              /* O: Distance from satellite to target */
    double *rotation,              /* O: Rotation angle */
    double *ground_velocity,       /* O: Ground velocity in meters/second */
    double *utm_y,                 /* O: UTM Y coordinate */
    double *utm_x                  /* O: UTM X coordinate */
);

int update_omf
(
    const char *omf_filename,      /* I: OMF filename */
    char *mtf_input_directory      /* I: MTF-specific input directory */
);

const IAS_GEO_PROJ_TRANSFORMATION *get_proj_transform_geo2target();

int set_geo_to_target_projection
(
    IAS_PROJECTION *source_projection, /* I: source projection */
    IAS_PROJECTION *target_projection  /* I: target projection */
);

/* Free the projection transformation */
void free_geo_to_target_projection
(
);

#endif
