/******************************************************************************
NAME:           ols_to_ils

PURPOSE:
Map an output space line/sample coordinate to the corresponding input space
line/sample.  Locations are based on MTF targets.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
SUCCESS         Successful completion
ERROR           Operation failed 

******************************************************************************/
#include "ias_grid.h"
#include "ias_logging.h"
#include "mtfextract.h"

/* Map the output space line/sample coordinates to the corresponding input
   space line/sample for a single SCA */
void ols_to_ils_sca
(
    IAS_GRID_BAND_TYPE *grid_band_pointer, /* I: Grid band pointer */
    double output_line,            /* I: Output line location */
    double output_sample,          /* I: Output line sample */
    int band_number,               /* I: Band number */
    int current_polygon,           /* I: Current polygon being processed */
    int current_point,             /* I: Current point being processed */
    int current_sca,               /* I: Current SCA being processed */
    int number_points,             /* I: Number of vertices in a polygon */
    int number_scas,               /* I: Number of SCAs */
    MTF_TARGET *target             /* I/O: Structure of target information */
)
{
    double input_line;             /* Input line location */
    double input_sample;           /* Input sample location */
    int number_grid_lines;         /* Number of grid lines */
    int number_grid_samples;       /* Number of grid samples */
    int input_number_lines;        /* Number of lines in L1R */
    int input_number_samples;      /* Number of samples in L1R */
    int sca_index = 0;             /* Index into MTF target SCAs */

    /* Convert the output point to input space */
    if (ias_grid_3d_ols2ils(grid_band_pointer, output_line, output_sample,
        &input_line, &input_sample, target->height) == SUCCESS)
    {
        /* Extract grid's input number of lines and samples */
        number_grid_lines = grid_band_pointer->ngrid_lines - 1;
        number_grid_samples = grid_band_pointer->ngrid_samps - 1;
        input_number_lines = grid_band_pointer->in_lines[number_grid_lines];
        input_number_samples = grid_band_pointer->in_samps[number_grid_samples];

        /* Check the input line and sample */
        if ((input_line > 1) && (input_line < (input_number_lines - 2)) 
            && (input_sample > 1) 
            && (input_sample < (input_number_samples - 2)))
        {
            IAS_LOG_DEBUG("%s Band %d pt %d SCA %d Output %f,%f Input %f,%f\n",
                target->name, band_number, current_point, current_sca,
                output_line, output_sample, input_line, input_sample); 

            /* Find if the current SCA is already represented in the polygon */
            for (sca_index = 0; sca_index < target->polygons[current_polygon].
                number_scas; sca_index++)
            {
                if (target->polygons[current_polygon].sca[sca_index] 
                    == current_sca)
                {
                    break;
                }
            }

            /* If a match for the current SCA wasn't found */
            if (sca_index == target->polygons[current_polygon].number_scas)
            {
                /* Add the current SCA since a vertex falls inside it */
                target->polygons[current_polygon].sca[sca_index] = current_sca;
                target->polygons[current_polygon].number_scas++;
            }
        }
    }

    /* Save the input line and sample */
    target->polygons[current_polygon].
        l1r_sample[current_sca * number_points + current_point] = input_sample;
    target->polygons[current_polygon].
        l1r_line[current_sca * number_points + current_point] = input_line;
}

/* Map the output space line/sample coordinates to the corresponding input
   space line/sample for all targets, polygons, points, and SCAs */
int ols_to_ils 
(
    IAS_GRID_TYPE *grid,           /* I: Grid structure */
    int band_number,               /* I: Band number */
    int number_scas,               /* I: Number of SCAs */
    int number_targets,            /* I: Number of MTF targets */
    MTF_TARGET *targets            /* I/O: Structure of target information */
)
{
    int current_target;            /* Loop counter */
    int current_polygon;           /* Loop counter */
    int current_point;             /* Loop counter */
    int current_sca;               /* Loop counter */
    int band_index;                /* Band index for band number */
    double pixel_size;             /* Output pixel size */
    int number_points;             /* Number of vertices in a polygon */
    double output_line;            /* Output line location */
    double output_sample;          /* Output line sample */

    /* Initialize SCA to something other than a valid SCA */
    for (current_target = 0; current_target < number_targets; current_target++)
    {
        for (current_polygon = 0; current_polygon < 
            targets[current_target].number_polygons; current_polygon++)
        {
            targets[current_target].polygons[current_polygon].number_scas = 0;
            for (current_sca = 0; current_sca < number_scas; current_sca++)
            {
                targets[current_target].polygons[current_polygon].
                    sca[current_sca] = -1;
            }
        }
    }

    /* Initialize the pixel size */
    band_index = ias_sat_attr_convert_band_number_to_index(band_number);
    if (band_index == ERROR)
    {
        IAS_LOG_ERROR("Invalid band index for band number %d", band_number);
        return ERROR;
    }
    pixel_size = grid->gridbands[band_index][0].pixsize;

    /* Loop over every target, polygon, point, and SCA */
    for (current_target = 0; current_target < number_targets; current_target++)
    {
        for (current_polygon = 0; current_polygon < 
            targets[current_target].number_polygons; current_polygon++)
        {
            number_points = targets[current_target].polygons[current_polygon].
                number_points;
                
            for (current_point = 0; current_point < number_points;
                current_point++)
            {
                output_line = (grid->corners.upleft.y - 
                    targets[current_target].polygons[current_polygon].
                    utm_y[current_point]) / pixel_size;
                output_sample = (targets[current_target].
                    polygons[current_polygon].utm_x[current_point] -
                    grid->corners.upleft.x) / pixel_size;

                targets[current_target].polygons[current_polygon].
                    l1g_line[current_point] = output_line;
                targets[current_target].polygons[current_polygon].
                    l1g_sample[current_point] = output_sample;

                for (current_sca = 0; current_sca < number_scas; current_sca++)
                {
                    /* Map the coordinates for a single SCA */ 
                    ols_to_ils_sca(&grid->gridbands[band_index][current_sca],
                        output_line, output_sample, band_number, 
                        current_polygon, current_point, current_sca,
                        number_points, number_scas, &targets[current_target]);
                } /* SCA loop */
            } /* Point loop */
        } /* Polygon loop */
    } /* Target loop */

    return SUCCESS;
}

