/******************************************************************************
NAME:           polygon_sca_bounds

PURPOSE:
Determine bounds on polygon within one SCA

RETURN VALUE:
Type = int
Value           Description
-----           -----------
SUCCESS         Successful completion
ERROR           Operation failed 

******************************************************************************/
#include "ias_grid.h"
#include "ias_grid_io.h"
#include "ias_logging.h"
#include "mtfextract.h"

int polygon_sca_bounds 
(
    IAS_GRID_BAND_TYPE *grid_band_pointer, /* Grid band pointer */
    MTF_POLYGON *polygon,          /* I: Structure of target information */
    double *minimum_line,          /* O: Minimum line of boundary */
    double *minimum_sample,        /* O: Minimum sample of boundary */
    double *maximum_line,          /* O: Maximum line of boundary */
    double *maximum_sample         /* O: Maximum sample of boundary */
)
{
    int current_point;             /* Vertex or points counter */
    int elevation_plane;           /* Elevation plane */
    double output_line;            /* Output line location */
    double output_sample;          /* Output sample location */
    double input_line;             /* Input line location */
    double input_sample;           /* Input sample location */

    /* Loop over all points in the polygon */
    for (current_point = 0; current_point < polygon->number_points;
        current_point++)
    {
        /* Set up some input parameters and convert the current output point 
           from the polygon to input space */
        elevation_plane = grid_band_pointer->zeroplane;
        output_line = polygon->l1g_line[current_point];
        output_sample = polygon->l1g_sample[current_point];
        if (ias_grid_ols2ils(grid_band_pointer, output_line, output_sample,
            &input_line, &input_sample, elevation_plane) != SUCCESS)
        {
            IAS_LOG_ERROR("Determining input location for point %d output "
                "line %f sample %f", current_point, output_line, output_sample);
            return ERROR;
        }

        /* Keep track of the boundaries, i.e. the minimum and maximum line
           and sample */
        if (current_point == 0)
        {
            /* Initialize the minimum and maximum line and sample */
            *maximum_line = input_line;
            *maximum_sample = input_sample;
            *minimum_line = input_line;
            *minimum_sample = input_sample;
        }
        else
        {
            /* Update the minimum and maximum line and sample if the current
               input line and sample expand the boundaries */
            if (*maximum_line < input_line)
                *maximum_line = input_line;
            if (*maximum_sample < input_sample)
                *maximum_sample = input_sample;
            if (*minimum_line > input_line)
                *minimum_line = input_line;
            if (*minimum_sample > input_sample)
                *minimum_sample = input_sample;
        }
    } /* Point loop */

    IAS_LOG_DEBUG("bounds %f,%f,%f,%f\n", *maximum_line, *maximum_sample,
        *minimum_line, *minimum_sample);

    return SUCCESS;
}

