/******************************************************************************
NAME:           read_image

PURPOSE:
Read a window of data from an L1R image file.

    Includes:
        read_l1r_band_sca()
        get_l1r_window()
        cleanup_l1r_image()

******************************************************************************/
#include <stdlib.h>                /* malloc */
#include <math.h>                  /* floor, ceil */
#include <string.h>                /* memcpy */

#include "ias_cpf.h"
#include "ias_l1r.h"
#include "ias_structures.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "mtfextract.h"

/******************************************************************************
NAME:           apply_scaling_factor

PURPOSE:
Apply the radiance rescaling factors to the L1R float data

RETURN VALUE:
Type = float
Value           Description
-----           -----------
output_value    output value of the L1R data with scaling factors applied

******************************************************************************/
static inline float apply_scaling_factor
(
    float input_value,             /* I: The value to apply scaling to */
    double scaling_multiplier,     /* I: Radiance rescaling factor */
    double scaling_offset          /* I: Radiance rescaling factor */
)
{
    float output_value;

    output_value = (input_value - scaling_offset) * scaling_multiplier;

    /* Make sure this float doesn't go out of the unsigned int range */
    if (output_value < 0.0)
    {
        output_value = 0.0;
    }
    else if (output_value > 65535.0)
    {
        output_value = 65535.0;
    }

    return output_value;
}

/******************************************************************************
NAME:           read_l1r_band_sca

PURPOSE:
Initialize the read for a window of data from an L1R image file.  Acually
do the read and make the data available for calls to read windows of data.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
SUCCESS         Successfully initialized read 
ERROR           Error  initializing read

******************************************************************************/
int read_l1r_band_sca 
(
    const char *l1r_file,          /* I: Level 1R data file name */
    int band_number,               /* I: Band to read */
    int sca,                       /* I: SCA to read */ 
    int number_scas,               /* I: Number of SCAs for OLI */
    IAS_CPF *cpf,                  /* I: CPF constants */
    IAS_IMAGE *sca_image           /* O: L1R image for band/SCA */
)
{
    int band_index;                /* Band index for band_number */
    L1RIO *l1r_pointer;            /* Input image data */
    L1R_BAND_IO *l1r_band;         /* Pointer to L1R band */
    const struct IAS_CPF_RADIANCE_RESCALE *radiance_rescale = NULL; /* Radiance
                                         rescale parameters from CPF */
    int detectors_per_sca;         /* Number of detectors in an SCA */
    long image_size;               /* Size of image */
    float *temporary_buffer;       /* Buffer for image data */
    int image_offset;              /* Offset into image data */
    int line_index;                /* Loop counter */
    int sample_index;              /* Loop counter */
    IAS_DATA_TYPE data_type;       /* Data type for the imagery */
    IAS_UI2_TYPE *data_pointer;    /* Pointer to image data */
    double scaling_multiplier = 1.0; /* Radiance rescaling factor */
    double scaling_offset = 0.0;   /* Radiance rescaling factor */
    

    /* Get the band index for band_number */
    band_index = ias_sat_attr_convert_band_number_to_index(band_number);
    if (band_index == ERROR)
    {
        IAS_LOG_ERROR("Invalid band index for band number %d", band_number);
        return ERROR;
    }

    /* Get the scaling factors from the CPF */
    radiance_rescale = ias_cpf_get_radiance_rescale(cpf);
    if (radiance_rescale == NULL)
    {
        IAS_LOG_ERROR("Reading radiance rescale parameters from the CPF");
        return ERROR;
    }
    scaling_multiplier = radiance_rescale->gain[band_index];
    scaling_offset = radiance_rescale->bias[band_index];

    /* Invert the gain so the repeated division becomes a multiplication */
    scaling_multiplier = 1.0 / scaling_multiplier;

    /* Open the L1R file */
    l1r_pointer = ias_l1r_open_image(l1r_file, IAS_READ);
    if (l1r_pointer == NULL)
    {
        IAS_LOG_ERROR("Opening L1R file %s", l1r_file); 
        return ERROR;
    }

    /* Open the L1R band */
    l1r_band = ias_l1r_open_band(l1r_pointer, band_number, &data_type, 
        &number_scas, &sca_image->nl, &sca_image->ns);
    if (l1r_band == NULL)
    {
        IAS_LOG_ERROR("Opening L1R %s band number %d", l1r_file, band_number); 
        ias_l1r_close_image(l1r_pointer);
        return ERROR;
    }

    /* Verify the band's data type */
    if (data_type != IAS_R4)
    {
        IAS_LOG_ERROR("L1R data type is %d but should be IAS_R4", data_type);
        ias_l1r_close_band(l1r_band);
        ias_l1r_close_image(l1r_pointer);
        return ERROR;
    }

    /* Get the number of detectors for an SCA for the band */
    detectors_per_sca = ias_sat_attr_get_detectors_per_sca(band_number);
    if (detectors_per_sca == ERROR)
    {
        IAS_LOG_ERROR("Could not get number of detectors per SCA for band "
            "number %d", band_number);
        ias_l1r_close_band(l1r_band);
        ias_l1r_close_image(l1r_pointer);
        return ERROR;
    }

    /* Verify the number of detectors in the L1R file matches the CPF */
    if (detectors_per_sca != sca_image->ns)
    {
        IAS_LOG_ERROR("L1R file number of samples %d and expected number %d "
            "for band number %d do not match", detectors_per_sca, 
            sca_image->ns, band_number);
        ias_l1r_close_band(l1r_band);
        ias_l1r_close_image(l1r_pointer);
        return ERROR;
    }

    IAS_LOG_DEBUG("Input size %d %d %d band %d\n", number_scas, sca_image->nl,
        sca_image->ns, band_number);

    /* Allocate data buffer */
    temporary_buffer = (float *)malloc(sca_image->ns * sizeof(float));
    if (temporary_buffer == NULL)
    {
        IAS_LOG_ERROR("Allocating data buffer");
        ias_l1r_close_band(l1r_band);
        ias_l1r_close_image(l1r_pointer);
        return ERROR;
    }

    /* Allocate image buffer */
    image_size = sca_image->ns * (sca_image->nl);
    data_pointer = (IAS_UI2_TYPE *)malloc(image_size * sizeof(IAS_UI2_TYPE));
    if (data_pointer == NULL)
    {
        IAS_LOG_ERROR("Allocating SCA data");
        ias_l1r_close_band(l1r_band);
        ias_l1r_close_image(l1r_pointer);
        free(temporary_buffer);
        return ERROR;
    }

    /* Save the image data pointer in the IAS_IMAGE, typecasting it to agree
       with the short pointer that structure expects */
    sca_image->data = (IAS_I2_TYPE *)data_pointer;

    /* Read L1R imagery into sca_image->data buffer */
    for (line_index = 0; line_index < sca_image->nl; line_index++)
    {
        if (ias_l1r_read_image(l1r_band, sca, line_index, 0, 1, sca_image->ns,
            temporary_buffer) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading L1R line %d for band number %d sca index "
                "%d", line_index, band_number, sca);
            ias_l1r_close_band(l1r_band);
            ias_l1r_close_image(l1r_pointer);
            cleanup_l1r_image(sca_image);
            free(temporary_buffer);
            return ERROR;
        }

        /* Copy the line of data in temporary buffer to sca_image->data, 
           applying the rescaling factor to each data point in the process */
        image_offset = line_index * sca_image->ns;
        for (sample_index = 0; sample_index < sca_image->ns; sample_index++)
        {
            data_pointer[image_offset + sample_index] 
                = apply_scaling_factor(temporary_buffer[sample_index], 
                scaling_multiplier, scaling_offset) + 0.5; 
        }
    } /* Line index loop */

    /* Free structures */
    free(temporary_buffer);

    /* Close the L1r band */
    if (ias_l1r_close_band(l1r_band) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing L1R %s band number %d", l1r_file, band_number); 
        ias_l1r_close_image(l1r_pointer);
        cleanup_l1r_image(sca_image);
    }

    /* Close the L1r file */
    if (ias_l1r_close_image(l1r_pointer) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing L1R %s", l1r_file); 
        cleanup_l1r_image(sca_image);
    }

    return SUCCESS;
}

/******************************************************************************
NAME:           get_l1r_window

PURPOSE:
Read a window of data from an L1R image file.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
SUCCESS         Successfully extracted target 
ERROR           Error extracting target

******************************************************************************/
int get_l1r_window
(
    int starting_line,             /* I: Starting line within SCA to read */
    int starting_sample,           /* I: Starting sample within SCA to read */
    int number_read_lines,         /* I: Number of lines within SCA to read */
    int number_read_samples,       /* I: Number of samples within SCA to read */
    IAS_IMAGE *sca_image,          /* I: L1R image for band/SCA */
    IAS_UI2_TYPE *output_data      /* O: Image data */
)
{
    int line_number;               /* Loop counter */
    int sca_offset;                /* Offset into SCA data for the lines of 
                                      the window */
    int output_data_offset;        /* Offset into the output data to put
                                      SCA data */
    IAS_UI2_TYPE *data_pointer;    /* Pointer to image data */

    /* Check that the image window requested is within range for lines */
    if (starting_line < 0 || starting_line + number_read_lines >= sca_image->nl)
    {
        IAS_LOG_ERROR("Windowing is outside image for the line location; "
            "starting line %d, number of lines to read %d, available lines "
            "%d", starting_line, number_read_lines, sca_image->nl); 
        return ERROR;
    }

    /* Check that the image window requested is within range for samples */
    if (starting_sample < 0 || starting_sample + number_read_samples 
        >= sca_image->ns)
    {
        IAS_LOG_ERROR("Windowing is outside image for the sample location; "
            "starting sample %d, number of samples to read %d, available "
            "samples %d", starting_sample, number_read_samples, sca_image->ns);
        return ERROR;
    }

    /* Point to the image data so we can treat it as unsigned short instead of 
       IAS_IMAGE's short */
    data_pointer = (IAS_UI2_TYPE *)sca_image->data;

    /* "Read" the image window.  It was actually read in the init routine and
       put in a static structure, so just copy the appropriate window from 
       there */
    for (line_number = starting_line; line_number < starting_line 
        + number_read_lines; line_number++)
    {
        sca_offset = line_number * sca_image->ns + starting_sample;
        output_data_offset = (line_number - starting_line) 
            * number_read_samples;
        memcpy(&output_data[output_data_offset], &data_pointer[sca_offset], 
            sizeof(IAS_UI2_TYPE) * number_read_samples);
    }

    return SUCCESS;

}

/******************************************************************************
NAME:           cleanup_l1r_image

PURPOSE:
Cleanup data associated with the read of the l1r image 

RETURN VALUE:
    none

******************************************************************************/
void cleanup_l1r_image
(
    IAS_IMAGE *sca_image           /* I: L1R image for band/SCA */
)
{
    free(sca_image->data);
    sca_image->data = NULL;
}
