/******************************************************************************

NAME:           mtfextract_test

PURPOSE:
Main test procedure for software to automatically locate and extract MTF target
windows predefined targets.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
EXIT_FAILURE    Error extracting MTF windows
EXIT_SUCCESS    Successfully extracted MTF windows

******************************************************************************/

#include <stdlib.h>         /* EXIT_FAILURE/EXIT_SUCCESS */
#include <limits.h>         /* PATH_MAX */
#include <string.h>         /* strncpy */
#include <unistd.h>         /* access */

#include "gctp.h"
#include "cgreen/cgreen.h"
#include "mtf.h"
#include "mtfextract.h"
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_cpf.h"
#include "ias_grid.h"
#include "ias_grid_io.h"
#include "ias_model_io.h"
#include "ias_geo.h"

int compare_files(const char *f1, const char *f2)
{
    int status;
    char command[PATH_MAX];

    status = snprintf(command, sizeof(command), "diff -q %s %s",
        f1, f2);
    if (status < 0 || sizeof(command) <= status)
    {
        IAS_LOG_INFO("Error preparing diff system command");
        return 1;
    }

    status = system(command);
    if (status != 0)
    {
        printf("Differences found in pixel table files. "
            "Review %s and %s for details", f1, f2);
        return 1;
    }

    return 0;
}

Ensure test_mtfextract()
{
    char parm_file[PATH_MAX] = "MTFEXTRACT.odl";  /* Input parameter file 
                                         name */
    char input_directory[PATH_MAX];   /* Location of MTF-specific input 
                                         files like target description file */
    char omf_file[PATH_MAX];          /* Name of the the OMF */
    char grid_file[PATH_MAX];         /* Grid file name */
    char model_file[PATH_MAX];        /* Model file name */
    char l0r_file[PATH_MAX];          /* L0R file name */
    char l1r_file[PATH_MAX];          /* L1R file name */
    char cpf_file[PATH_MAX];          /* CPF file name */
    IAS_GRID_TYPE grid;               /* Geometric resampling grid structure */
    IAS_LOS_MODEL *model;             /* Model structure */
    IAS_CPF *cpf;                     /* CPF structure pointer */
    int number_oli_scas;              /* Number of SCAs in OLI bands */
    int number_bands;                 /* Number of bands to process */
    int band_numbers[IAS_MAX_NBANDS]; /* List of band numbers to process */
    int target_wrs_path = 0;          /* Target WRS path */
    int target_wrs_row = 0;           /* Target WRS row */
    int utm_zone = 0;                 /* Target definition UTM zone */
    int number_targets;               /* Number of targets in image */
    MTF_TARGET *targets;              /* Targets in image */
    int out_units;                    /* Code for the output projection units */
    double projection_parameters[IAS_PROJ_PARAM_SIZE]; /* Projection
                                         parameters */
    int parameter_index;              /* Projection parameter loop counter */
    IAS_PROJECTION geo_projection;    /* Source lat/lon projection */
    IAS_PROJECTION target_projection; /* Target projection */
    int status;                       /* Function return status */

    char command[PATH_MAX];           /* System command */
    static char *unit_test_directory = NULL; /* IAS unit test data directory */
    char *unit_test_path = "ias/gps/mtf/mtfextract"; /* Rest of unit test data
                                         directory path */
    char *pixel_table_basename = "Tar0Poly0Band1Sca7.dat"; /* Base name of the
                                         single output file in the test 
                                         scenario - the test scenario only has
                                         1 target, 1 polygon, and 1 band */

    /* Comparison versions of parameter variables */
    char compare_input_directory[PATH_MAX];
    char compare_grid_file[PATH_MAX];
    char compare_model_file[PATH_MAX];
    char compare_l0r_file[PATH_MAX];
    char compare_l1r_file[PATH_MAX];
    char compare_cpf_file[PATH_MAX];
    char compare_pixel_table_file[PATH_MAX];

    /* Initialize the logging library */
    if (ias_log_initialize("mtfextract_test") != SUCCESS)
    {
        printf("Initializing logging library\n");
        exit(EXIT_FAILURE);
    }

    IAS_LOG_INFO("Initiating mtfextract_test");

    /* Delete output files generated by previous runs of this test */
    unlink(pixel_table_basename);

    /* Get unit test directory */
    unit_test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (unit_test_directory == NULL)
    {
        printf("The IAS_UNIT_TEST_DATA_DIR environment variable is not set\n");
        exit(EXIT_FAILURE);
    }

    /* Make application parameter file with IAS_UNIT_TEST_DATA_DIR expanded */ 
    status = snprintf(command, sizeof(command),
            "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' MTFEXTRACT_template.odl > "
            "MTFEXTRACT.odl", unit_test_directory);
    if (status < 0 || sizeof(command) <= status)
    {
        assert_true_with_message(0, "Error preparing application parameter "
            "file conversion system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Error preparing application parameter "
            "test file for unit test data directory path setup");
        return;
    }

    /* Make OMF with IAS_UNIT_TEST_DATA_DIR expanded */ 
    status = snprintf(command, sizeof(command),
            "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' TSTI89_template.omf > "
            "TSTI89.omf", unit_test_directory);
    if (status < 0 || sizeof(command) <= status)
    {
        assert_true_with_message(0, "Error preparing OMF conversion system "
            "command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Error preparing parameter test OMF file "
            "for unit test data directory path setup");
        return;
    }

    /* Get parameters */
    status = get_parameters(parm_file, input_directory, omf_file, grid_file, 
        model_file, l0r_file, l1r_file, cpf_file, &target_wrs_path, 
        &target_wrs_row, band_numbers, &number_bands);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error reading parameters");
        return;
    }

    /* Set up comparison strings so the parameters can be checked */
    status = snprintf(compare_input_directory, sizeof(compare_input_directory),
        "%s/%s", unit_test_directory, unit_test_path);
    if (status < 0 || sizeof(compare_input_directory) <= status)
    {
        assert_true_with_message(0, "Error building compare_input_directory");
        return;
    }
    status = snprintf(compare_grid_file, sizeof(compare_grid_file), "%s/%s/%s", 
        unit_test_directory, unit_test_path, "grid.h5");
    if (status < 0 || sizeof(compare_grid_file) <= status)
    {
        assert_true_with_message(0, "Error building compare_grid_file");
        return;
    }
    status = snprintf(compare_model_file, sizeof(compare_model_file), 
        "%s/%s/%s", unit_test_directory, unit_test_path, "model.h5");
    if (status < 0 || sizeof(compare_model_file) <= status)
    {
        assert_true_with_message(0, "Error building compare_model_file");
        return;
    }
    status = snprintf(compare_l0r_file, sizeof(compare_l0r_file), "%s/%s/%s", 
        unit_test_directory, unit_test_path, "LO80220390392000261EDC00");
    if (status < 0 || sizeof(compare_l0r_file) <= status)
    {
        assert_true_with_message(0, "Error building compare_l0r_file");
        return;
    }
    status = snprintf(compare_l1r_file, sizeof(compare_l1r_file), "%s/%s/%s", 
        unit_test_directory, unit_test_path, "l1r_image.h5");
    if (status < 0 || sizeof(compare_l1r_file) <= status)
    {
        assert_true_with_message(0, "Error building compare_l1r_file");
        return;
    }
    status = snprintf(compare_cpf_file, sizeof(compare_cpf_file), "%s/%s", 
        unit_test_directory, "ias_lib/cpf/L8CPF20000701_20000930.01");
    if (status < 0 || sizeof(compare_cpf_file) <= status)
    {
        assert_true_with_message(0, "Error building compare_cpf_file");
        return;
    }

    /* Compare the comparison strings to the results from get_parameters */ 
    assert_string_equal(input_directory, compare_input_directory);
    assert_string_equal(omf_file, "TSTI89.omf"); 
    assert_string_equal(grid_file, compare_grid_file); 
    assert_string_equal(model_file, compare_model_file); 
    assert_string_equal(l0r_file, compare_l0r_file); 
    assert_string_equal(l1r_file, compare_l1r_file); 
    assert_string_equal(cpf_file, compare_cpf_file); 

    assert_equal(target_wrs_path, 22);
    assert_equal(target_wrs_row, 39);

    /* Get the number of SCAs for OLI bands */
    number_oli_scas = ias_sat_attr_get_sensor_sca_count(IAS_OLI); 
    assert_equal(number_oli_scas, 14);

    /* Get the target descriptions */
    status = get_target_descriptions(input_directory, target_wrs_path, 
        target_wrs_row, number_oli_scas, &number_targets, &targets, &utm_zone);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error getting target descriptions");
        return;
    }

    assert_equal(number_targets, 1);
    assert_equal(targets[0].number_polygons, 1);
    assert_string_equal(targets[0].name, "Causeway");
    assert_equal(utm_zone, 15);
    assert_equal(targets[0].polygons[0].number_points, 4);
    assert_double_equal(targets[0].height, -25.0);

    /* Read the grid file */
    if (ias_grid_initialize(&grid, 0, 0) != SUCCESS)
    {
        assert_true_with_message(0, "Initializing the grid structure");
        free_polygons(targets, number_targets);
        return;
    }

    if (ias_grid_header_read(grid_file, &grid) != SUCCESS)
    {
        assert_true_with_message(0, "Reading the grid from %s", grid_file);
        free_polygons(targets, number_targets);
        return;
    }

    if (ias_grid_band_header_read(grid_file, &grid) != SUCCESS)
    {
        assert_true_with_message(0, "Reading the grid band header from %s", 
            grid_file);
        free_polygons(targets, number_targets);
        ias_grid_free(&grid);
        return;
    }

    /* Set up projection transformations */
    for (parameter_index = 0; parameter_index < IAS_PROJ_PARAM_SIZE;
        parameter_index++)
    {
        projection_parameters[parameter_index] = 0.0;
    }

    /* Convert the grid projection units to a unit code for GCTP. */
    status = ias_geo_get_units(grid.units, &out_units);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error converting projection units %s to "
            "unit code", grid.units);
        ias_grid_free(&grid);
        free_polygons(targets, number_targets); 
        return;
    }

    /* Set up projections */
    ias_geo_set_projection(GEO, NULLZONE, RADIAN, grid.spheroid, 
        projection_parameters, &geo_projection);
    ias_geo_set_projection(grid.code, grid.zone, out_units, grid.spheroid,
        grid.projprms, &target_projection); 

    /* Set up projection transformation */
    status = set_geo_to_target_projection(&geo_projection, &target_projection);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error creating geo to target projection "
            "transformation");
        ias_grid_free(&grid);
        free_polygons(targets, number_targets); 
        return;
    }

    /* Read the model file */
    model = ias_model_read(model_file);
    if (model == NULL)
    {
        assert_true_with_message(0, "Error reading model file %s", model_file);
        ias_grid_free(&grid);
        free_geo_to_target_projection();
        free_polygons(targets, number_targets); 
        return;
    }

    IAS_LOG_DEBUG("Processing band %d\n", band_numbers[0]);

    /* Load the grid information for the current band */
    if (ias_grid_band_pointers_read(grid_file, &grid, band_numbers[0]) 
        != SUCCESS)
    {
        assert_true_with_message(0, "Reading the grid pointers from %s for "
            "band number %d", grid_file, band_numbers[0]);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        model = NULL;
        free_geo_to_target_projection();
        free_polygons(targets, number_targets);
        ias_grid_free(&grid);
        return;
    }

    /* Determine image L1R locations for target */
    if (ols_to_ils(&grid, band_numbers[0], number_oli_scas, number_targets,
        targets) != SUCCESS)
    {
        assert_true_with_message(0, "Error getting target locations");
        ias_grid_free(&grid);
        ias_los_model_free(model);
        model = NULL;
        free_geo_to_target_projection();
        free_polygons(targets, number_targets);
        return;
    }

    assert_equal(targets[0].polygons[0].number_scas, 1);
    assert_double_equal(targets[0].polygons[0].l1r_line[0], 2937.857456);
    assert_double_equal(targets[0].polygons[0].l1r_sample[0], 3389.983601);
    assert_double_equal(targets[0].polygons[0].l1g_line[0], 3139.794700);
    assert_double_equal(targets[0].polygons[0].l1g_sample[0], 3829.343467);

    /* Allocate the CPF data structure and read it */
    cpf = ias_cpf_read(cpf_file);
    if (cpf == NULL)
    {
        assert_true_with_message(0, "Error opening CPF %s", cpf_file);
        ias_grid_free_band(&grid, band_numbers[0]);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        model = NULL;
        free_geo_to_target_projection();
        free_polygons(targets, number_targets);
        return;
    }

    /* Extract L1R from imagery and write report */
    if (generate_report_for_target(l1r_file, &grid, model, band_numbers[0], 
        number_oli_scas, cpf, number_targets, targets) != SUCCESS)
    { 
        assert_true_with_message(0, "Error extracting imagery");
        ias_cpf_free(cpf);
        ias_grid_free_band(&grid, band_numbers[0]);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        model = NULL;
        free_geo_to_target_projection();
        free_polygons(targets, number_targets);
        return;
    }

    /* Free grid data for the band */
    if (ias_grid_free_band(&grid, band_numbers[0]) != SUCCESS)
    {
        assert_true_with_message(0, "Freeing the grid data for band %d",
            band_numbers[0]);
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        model = NULL;
        free_geo_to_target_projection();
        free_polygons(targets, number_targets);
        return;
    }

    assert_double_equal(targets[0].polygons[0].utm_x[0], 778480.304000);
    assert_double_equal(targets[0].polygons[0].utm_y[0], 3360906.159000);

    /* Free the CPF, grid, and model, and projection transformation */
    ias_cpf_free(cpf);
    ias_grid_free(&grid);
    ias_los_model_free(model);
    model = NULL;
    free_geo_to_target_projection();
    free_polygons(targets, number_targets);

    /* Compare the output file to a saved version that matches CalVal results
       except for small differences */
    status = snprintf(compare_pixel_table_file, 
        sizeof(compare_pixel_table_file), "%s/%s/%s", unit_test_directory, 
        unit_test_path, pixel_table_basename);
    if (status < 0 || sizeof(compare_pixel_table_file) <= status)
    {
        assert_true_with_message(0, "Error building pixel table comparison "
            "filename");
        return;
    }

    if (compare_files(pixel_table_basename, compare_pixel_table_file) != 0)
    {
        assert_true_with_message(0, "Diff failed for files %s and %s",
            pixel_table_basename, compare_pixel_table_file);
        return;
    }

    /* Delete output files generated by previous runs of this test */
    if (unlink(pixel_table_basename) != 0)
    {
        assert_true_with_message(0, "Error removing output files");
        return;
    }

    /* Write completion message */
    IAS_LOG_INFO("Successful completion.");
}

Ensure test_mtfextract_multiple_scas()
{
    char parm_file[PATH_MAX] = "MTFEXTRACT_LC8120.odl"; /* Input parameter 
                                                           file name */
    char input_directory[PATH_MAX];   /* Location of MTF-specific input 
                                         files like target description file */
    char omf_file[PATH_MAX];          /* Name of the the OMF */
    char grid_file[PATH_MAX];         /* Grid file name */
    char model_file[PATH_MAX];        /* Model file name */
    char l0r_file[PATH_MAX];          /* L0R file name */
    char l1r_file[PATH_MAX];          /* L1R file name */
    char cpf_file[PATH_MAX];          /* CPF file name */
    IAS_GRID_TYPE grid;               /* Geometric resampling grid structure */
    IAS_LOS_MODEL *model;             /* Model structure */
    IAS_CPF *cpf;                     /* CPF structure pointer */
    int number_oli_scas;              /* Number of SCAs in OLI bands */
    int number_bands;                 /* Number of bands to process */
    int band_numbers[IAS_MAX_NBANDS]; /* List of band numbers to process */
    int target_wrs_path = 0;          /* Target WRS path */
    int target_wrs_row = 0;           /* Target WRS row */
    int utm_zone = 0;                 /* Target definition UTM zone */
    int number_targets;               /* Number of targets in image */
    MTF_TARGET *targets;              /* Targets in image */
    int out_units;                    /* Code for the output projection units */
    double projection_parameters[IAS_PROJ_PARAM_SIZE]; /* Projection
                                         parameters */
    int parameter_index;              /* Projection parameter loop counter */
    IAS_PROJECTION geo_projection;    /* Source lat/lon projection */
    IAS_PROJECTION target_projection; /* Target projection */
    int status;                       /* Function return status */

    char command[PATH_MAX];           /* System command */
    static char *unit_test_data_directory = NULL; /* IAS unit test data
                                                     directory */
    char *unit_test_path = "ias/gps/mtf/mtfextract/"/* Rest of unit test data */
                            "LC81200352015133LGN00"; /* directory path */
 
    const int num_pixel_tables = 2; /* Size of pixel_table_basenames */
    char *pixel_table_basenames[] = {"Tar0Poly0Band6Sca7.dat",
                                     "Tar0Poly0Band6Sca8.dat"};
    int i = 0; /* Generic loop counter */

    /* Comparison versions of parameter variables */
    char compare_input_directory[PATH_MAX];
    char compare_grid_file[PATH_MAX];
    char compare_model_file[PATH_MAX];
    char compare_l0r_file[PATH_MAX];
    char compare_l1r_file[PATH_MAX];
    char compare_cpf_file[PATH_MAX];
    char compare_pixel_table_file[PATH_MAX];

    /* Initialize the logging library */
    if (ias_log_initialize("mtfextract_test") != SUCCESS)
    {
        printf("Initializing logging library\n");
        exit(EXIT_FAILURE);
    }

    IAS_LOG_INFO("Initiating mtfextract_test");

    /* Delete output files generated by previous runs of this test */
    for (i = 0; i < num_pixel_tables; i++)
    {
        unlink(pixel_table_basenames[i]);
    }

    /* Get unit test directory */
    unit_test_data_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (unit_test_data_directory == NULL)
    {
        printf("The IAS_UNIT_TEST_DATA_DIR environment variable is not set\n");
        exit(EXIT_FAILURE);
    }

    /* Make application parameter file with IAS_UNIT_TEST_DATA_DIR expanded */ 
    status = snprintf(command, sizeof(command),
            "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' MTFEXTRACT_LC8120_template.odl"
            "> MTFEXTRACT_LC8120.odl", unit_test_data_directory);
    if (status < 0 || sizeof(command) <= status)
    {
        assert_true_with_message(0, "Error preparing application parameter "
            "file conversion system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Error preparing application parameter "
            "test file for unit test data directory path setup");
        return;
    }

    /* Make OMF with IAS_UNIT_TEST_DATA_DIR expanded */ 
    status = snprintf(command, sizeof(command),
            "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' TSTI305_template.omf > "
            "TSTI305.omf", unit_test_data_directory);
    if (status < 0 || sizeof(command) <= status)
    {
        assert_true_with_message(0, "Error preparing OMF conversion system "
            "command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Error preparing parameter test OMF file "
            "for unit test data directory path setup");
        return;
    }

    /* Get parameters */
    status = get_parameters(parm_file, input_directory, omf_file, grid_file, 
        model_file, l0r_file, l1r_file, cpf_file, &target_wrs_path, 
        &target_wrs_row, band_numbers, &number_bands);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error reading parameters");
        return;
    }

    /* Set up comparison strings so the parameters can be checked */
    status = snprintf(compare_input_directory, sizeof(compare_input_directory),
        "%s/%s", unit_test_data_directory, unit_test_path);
    if (status < 0 || sizeof(compare_input_directory) <= status)
    {
        assert_true_with_message(0, "Error building compare_input_directory");
        return;
    }
    status = snprintf(compare_grid_file, sizeof(compare_grid_file), "%s/%s/%s", 
        unit_test_data_directory, unit_test_path, "systematic_grid.h5");
    if (status < 0 || sizeof(compare_grid_file) <= status)
    {
        assert_true_with_message(0, "Error building compare_grid_file");
        return;
    }
    status = snprintf(compare_model_file, sizeof(compare_model_file), 
        "%s/%s/%s", unit_test_data_directory, unit_test_path,
        "systematic_model.h5");
    if (status < 0 || sizeof(compare_model_file) <= status)
    {
        assert_true_with_message(0, "Error building compare_model_file");
        return;
    }
    status = snprintf(compare_l0r_file, sizeof(compare_l0r_file), "%s/%s/%s", 
        unit_test_data_directory, unit_test_path, "LC81200352015133LGN00");
    if (status < 0 || sizeof(compare_l0r_file) <= status)
    {
        assert_true_with_message(0, "Error building compare_l0r_file");
        return;
    }
    status = snprintf(compare_l1r_file, sizeof(compare_l1r_file), "%s/%s/%s", 
        unit_test_data_directory, unit_test_path, "l1r_image.h5");
    if (status < 0 || sizeof(compare_l1r_file) <= status)
    {
        assert_true_with_message(0, "Error building compare_l1r_file");
        return;
    }
    status = snprintf(compare_cpf_file, sizeof(compare_cpf_file), "%s/%s", 
        unit_test_data_directory, "ias_lib/cpf/L8CPF20150401_20150630.03");
    if (status < 0 || sizeof(compare_cpf_file) <= status)
    {
        assert_true_with_message(0, "Error building compare_cpf_file");
        return;
    }

    /* Compare the comparison strings to the results from get_parameters */ 
    assert_string_equal(input_directory, compare_input_directory);
    assert_string_equal(omf_file, "TSTI305.omf"); 
    assert_string_equal(grid_file, compare_grid_file); 
    assert_string_equal(model_file, compare_model_file); 
    assert_string_equal(l0r_file, compare_l0r_file); 
    assert_string_equal(l1r_file, compare_l1r_file); 
    assert_string_equal(cpf_file, compare_cpf_file); 

    assert_equal(target_wrs_path, 120);
    assert_equal(target_wrs_row, 35);

    /* Get the number of SCAs for OLI bands */
    number_oli_scas = ias_sat_attr_get_sensor_sca_count(IAS_OLI); 
    assert_equal(number_oli_scas, 14);

    /* Get the target descriptions */
    status = get_target_descriptions(input_directory, target_wrs_path, 
        target_wrs_row, number_oli_scas, &number_targets, &targets, &utm_zone);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error getting target descriptions");
        return;
    }

    assert_equal(number_targets, 1);
    assert_equal(targets[0].number_polygons, 1);
    assert_string_equal(targets[0].name, "Qingdao");
    assert_equal(utm_zone, 51);
    assert_equal(targets[0].polygons[0].number_points, 4);
    assert_double_equal(targets[0].height, 3.0);

    /* Read the grid file */
    if (ias_grid_initialize(&grid, 0, 0) != SUCCESS)
    {
        assert_true_with_message(0, "Initializing the grid structure");
        free_polygons(targets, number_targets);
        return;
    }

    if (ias_grid_header_read(grid_file, &grid) != SUCCESS)
    {
        assert_true_with_message(0, "Reading the grid from %s", grid_file);
        free_polygons(targets, number_targets);
        return;
    }

    if (ias_grid_band_header_read(grid_file, &grid) != SUCCESS)
    {
        assert_true_with_message(0, "Reading the grid band header from %s", 
            grid_file);
        free_polygons(targets, number_targets);
        ias_grid_free(&grid);
        return;
    }

    /* Set up projection transformations */
    for (parameter_index = 0; parameter_index < IAS_PROJ_PARAM_SIZE;
        parameter_index++)
    {
        projection_parameters[parameter_index] = 0.0;
    }

    /* Convert the grid projection units to a unit code for GCTP. */
    status = ias_geo_get_units(grid.units, &out_units);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error converting projection units %s to "
            "unit code", grid.units);
        ias_grid_free(&grid);
        free_polygons(targets, number_targets);
        return;
    }

    /* Set up projections */
    ias_geo_set_projection(GEO, NULLZONE, RADIAN, grid.spheroid, 
        projection_parameters, &geo_projection);
    ias_geo_set_projection(grid.code, grid.zone, out_units, grid.spheroid,
        grid.projprms, &target_projection); 

    /* Set up projection transformation */
    status = set_geo_to_target_projection(&geo_projection, &target_projection);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error creating geo to target projection "
            "transformation");
        ias_grid_free(&grid);
        free_polygons(targets, number_targets);
        return;
    }

    /* Read the model file */
    model = ias_model_read(model_file);
    if (model == NULL)
    {
        assert_true_with_message(0, "Error reading model file %s", model_file);
        ias_grid_free(&grid);
        free_geo_to_target_projection();
        free_polygons(targets, number_targets);
        return;
    }

    IAS_LOG_DEBUG("Processing band %d\n", band_numbers[0]);

    /* Load the grid information for the current band */
    if (ias_grid_band_pointers_read(grid_file, &grid, band_numbers[0]) 
        != SUCCESS)
    {
        assert_true_with_message(0, "Reading the grid pointers from %s for "
            "band number %d", grid_file, band_numbers[0]);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        model = NULL;
        free_geo_to_target_projection();
        free_polygons(targets, number_targets);
        ias_grid_free(&grid);
        return;
    }

    /* Determine image L1R locations for target */
    if (ols_to_ils(&grid, band_numbers[0], number_oli_scas, number_targets,
        targets) != SUCCESS)
    {
        assert_true_with_message(0, "Error getting target locations");
        ias_grid_free(&grid);
        ias_los_model_free(model);
        model = NULL;
        free_geo_to_target_projection();
        free_polygons(targets, number_targets);
        return;
    }

    assert_equal(targets[0].polygons[0].number_scas, 2);
    assert_double_equal(targets[0].polygons[0].l1r_line[0], 3030.259474);
    assert_double_equal(targets[0].polygons[0].l1r_sample[0], 3710.226542);
    assert_double_equal(targets[0].polygons[0].l1g_line[0], 4125.104267);
    assert_double_equal(targets[0].polygons[0].l1g_sample[0], 4662.651833);

    /* Allocate the CPF data structure and read it */
    cpf = ias_cpf_read(cpf_file);
    if (cpf == NULL)
    {
        assert_true_with_message(0, "Error opening CPF %s", cpf_file);
        ias_grid_free_band(&grid, band_numbers[0]);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        model = NULL;
        free_geo_to_target_projection();
        free_polygons(targets, number_targets); 
        return;
    }

    /* Extract L1R from imagery and write report */
    if (generate_report_for_target(l1r_file, &grid, model, band_numbers[0], 
        number_oli_scas, cpf, number_targets, targets) != SUCCESS)
    { 
        assert_true_with_message(0, "Error extracting imagery");
        ias_cpf_free(cpf);
        ias_grid_free_band(&grid, band_numbers[0]);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        model = NULL;
        free_geo_to_target_projection();
        free_polygons(targets, number_targets); 
        return;
    }

    /* Free grid data for the band */
    if (ias_grid_free_band(&grid, band_numbers[0]) != SUCCESS)
    {
        assert_true_with_message(0, "Freeing the grid data for band %d",
            band_numbers[0]);
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        model = NULL;
        free_geo_to_target_projection();
        free_polygons(targets, number_targets);
        return;
    }

    assert_double_equal(targets[0].polygons[0].utm_x[0], 247279.555000);
    assert_double_equal(targets[0].polygons[0].utm_y[0], 4003646.872000);

    /* Free the CPF, grid, and model, and projection transformation */
    ias_cpf_free(cpf);
    ias_grid_free(&grid);
    ias_los_model_free(model);
    model = NULL;
    free_geo_to_target_projection();
    free_polygons(targets, number_targets);

    /* Compare the output file to a saved version that matches CalVal results
       except for small differences */
    for (i = 0; i < num_pixel_tables; i++)
    {
        status = snprintf(compare_pixel_table_file, 
            sizeof(compare_pixel_table_file), "%s/%s/%s", 
            unit_test_data_directory, unit_test_path, pixel_table_basenames[i]);
        if (status < 0 || sizeof(compare_pixel_table_file) <= status)
        {
            assert_true_with_message(0, "Error building pixel table comparison "
                "filename");
            return;
        }

        if (compare_files(pixel_table_basenames[i], 
            compare_pixel_table_file) != 0)
        {
            assert_true_with_message(0, "Diff failed for files %s and %s",
                pixel_table_basenames[i], compare_pixel_table_file);
            return;
        }

        /* Remove generated data file from local test dir. This is to detect
           changes that may hinder the MTF Extract application from detecting
           multiple SCAs. If the utility can no longer detect multiple SCAs, we
           want it to fail at generating the appropriate dat files. If the files
           remained, they would be compared even if the utility didn't
           generate them */
        if (unlink(pixel_table_basenames[i]) != 0)
        {
            assert_true_with_message(0, "Error removing output file %s",
                pixel_table_basenames[i]);
            return;
        }
        IAS_LOG_INFO("Successfully removed file %s", pixel_table_basenames[i]);
    }

    /* Write completion message */
    IAS_LOG_INFO("Successful completion.");
}


int main(int argc, char *argv[])
{
    TestSuite *suite = create_test_suite();
    add_test(suite, test_mtfextract);
    add_test(suite, test_mtfextract_multiple_scas);
    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}


