/******************************************************************************
NAME:           calc_spatial_perf.c

PURPOSE:
Source file for software that invokes the OLI STF model using the fitted
parameters, computes the corresponding PSF and ESF, and extracts the
resulting spatial edge slope and half edge extent performance parameters.

RETURNS:
    SUCCESS or ERROR

******************************************************************************/
#include <stdlib.h>
#include <math.h>

#include "ias_logging.h" /* Defines IAS_LOG macros */
#include "ias_math.h"

#include "model.h"       /* Defines constants and structures */
#include "mtfperform.h"

#define NUM_RESPONSE_POINTS  5

/* Calculate the PSF, ESF, and spatial performance parameters */
int calc_spatial_perf
(                           
    const char *output_dir, /* I: Output directory */
    const STF_FIT_PARMS *stf_fit, /* I: STF fit parameters */
    double gsd,             /* I: Band ground sample distance */
    double range,           /* I: Range to scale STF to ground */
    double gvel,            /* I: Velocity to scale STF to ground */
    int nsamp,              /* I: Length of PSF/ESF to analyze */
    int oversamp,           /* I: PSF/ESF oversample factor */
    double xt_best,         /* I: Low 90% value for xoptics */
    double at_best,         /* I: Low 90% value for aoptics */
    SPATIAL *result         /* O: Parameter results structure */
)
{
    double half_pi;         /* Pi / 2 constant */
    STF_FD_PARMS stf_fd_temp; /* Temporary frequency domain STF parameters */
    IAS_COMPLEX *model_buf = NULL; /* STF model buffer */
    double *psf_buf = NULL; /* PSF model buffer */
    int ifreq, jfreq;       /* Integer frequency indices */
    int status;
    int overall_status = SUCCESS;

    if ((output_dir == NULL) || (result == NULL) || (nsamp <= 1))
    {
        IAS_LOG_ERROR("Invalid inputs");
        return ERROR;
    }

    /* Allocate the modeling buffers */
    model_buf = (IAS_COMPLEX *)malloc(nsamp * sizeof(IAS_COMPLEX));
    if (model_buf == NULL)
    {
        IAS_LOG_ERROR("Allocating STF synthesis buffer, nsamp=%d", nsamp);
        return ERROR;
    }

    psf_buf = (double *)malloc(nsamp * sizeof(double));
    if (psf_buf == NULL)
    {
        free(model_buf);
        IAS_LOG_ERROR("Allocating PSF synthesis buffer");
        return ERROR;
    }

    /* Initialize constant */
    half_pi = ias_math_get_pi() / 2.0;

    /* Calculate cross-track performance */
    stf_fd_temp.g_optics = stf_fit->xt_optics;
    stf_fd_temp.xt_detector = stf_fit->xt_detector;
    stf_fd_temp.at_detector = stf_fit->at_detector;
    stf_fd_temp.integ = 0.0;
    stf_fd_temp.diffusion = stf_fit->diffusion;
    stf_fd_temp.phase = 0.0;

    /* Calculate the XT STF */
    create_frequency_domain_stf_model(&stf_fd_temp, gsd, range, gvel, 0.0, 
        nsamp, oversamp, model_buf);

    /* Use the IFFT to convert STF to cross-track PSF */
    ias_math_fft2d(&model_buf[0].re, nsamp, 1, IAS_MATH_IFFT);

    /* Take the real part and scale to convert to a real output */
    for (ifreq = 0; ifreq < nsamp; ifreq++)
    {
        jfreq = (ifreq + nsamp / 2) % nsamp;
        psf_buf[ifreq] = model_buf[jfreq].re / nsamp;

        /* Integrate the PSF to get the ESF */
        if (ifreq > 0)
        {
            psf_buf[ifreq] += psf_buf[ifreq - 1];
        }
    }

    /* Write out the synthesized ESF */
    status = write_esf(output_dir, "XT", result->band_number, nsamp, psf_buf);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Writing XT ESF for band number %d", result->band_number);
        overall_status = ERROR;
    }

    /* Calculate the XT edge slope and edge extent from the ESF */
    status = calc_spatial_stats(gsd, nsamp, oversamp, psf_buf,
                       &result->xt_slope, &result->xt_extent);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating spatial stats");
        overall_status = ERROR;
    }

    /* Calculate cross-track 90% best case performance */
    stf_fd_temp.g_optics = xt_best; 
    stf_fd_temp.xt_detector = stf_fit->xt_detector;
    stf_fd_temp.at_detector = stf_fit->at_detector;
    stf_fd_temp.integ = 0.0;
    stf_fd_temp.diffusion = stf_fit->diffusion;
    stf_fd_temp.phase = 0.0;

    /* Calculate the XT STF */
    create_frequency_domain_stf_model(&stf_fd_temp, gsd, range, gvel, 0.0, 
        nsamp, oversamp, model_buf);

    /* Use the IFFT to convert STF to cross-track PSF */
    ias_math_fft2d(&model_buf[0].re, nsamp, 1, IAS_MATH_IFFT);

    /* Take the real part and scale to convert to a real output */
    for (ifreq = 0; ifreq < nsamp; ifreq++)
    {
        jfreq = (ifreq + nsamp / 2) % nsamp;
        psf_buf[ifreq] = model_buf[jfreq].re / nsamp;

        /* Integrate the PSF to get the ESF */
        if (ifreq > 0)
        {
            psf_buf[ifreq] += psf_buf[ifreq - 1];
        }
    }

    /* Calculate the XT edge slope and edge extent from the ESF */
    status = calc_spatial_stats(gsd, nsamp, oversamp, psf_buf, 
        &result->xt_slope_best, &result->xt_extent_best);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Error calculating spatial stats");
        overall_status = ERROR;
    }

    /* Calculate along-track performance */
    stf_fd_temp.g_optics = stf_fit->at_optics;
    stf_fd_temp.xt_detector = stf_fit->xt_detector;
    stf_fd_temp.at_detector = stf_fit->at_detector;
    stf_fd_temp.integ = stf_fit->integration_milliseconds;
    stf_fd_temp.diffusion = stf_fit->diffusion;
    stf_fd_temp.phase = 0.0;

    /* Calculate the AT STF */
    create_frequency_domain_stf_model(&stf_fd_temp, gsd, range, gvel, half_pi, 
        nsamp, oversamp, model_buf);

    /* Use the IFFT to convert STF to cross-track PSF */
    ias_math_fft2d(&model_buf[0].re, nsamp, 1, IAS_MATH_IFFT);

    /* Take the real part and scale to convert to a real output */
    for (ifreq = 0; ifreq < nsamp; ifreq++)
    {
        jfreq = (ifreq + nsamp / 2) % nsamp;
        psf_buf[ifreq] = model_buf[jfreq].re / nsamp;

        /* Integrate the PSF to get the ESF */
        if (ifreq > 0)
        {
            psf_buf[ifreq] += psf_buf[ifreq - 1];
        }
    }

    /* Write out the synthesized ESF */
    status = write_esf(output_dir, "AT", result->band_number, nsamp, psf_buf);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Error writing AT ESF for band number %d", 
            result->band_number);
        overall_status = ERROR;
    }

    /* Calculate the AT edge slope and edge extent from the ESF */
    status = calc_spatial_stats(gsd, nsamp, oversamp, psf_buf,
                       &result->at_slope, &result->at_extent);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Error calculating spatial stats");
        overall_status = ERROR;
    }

    /* Calculate along-track 90% best case performance */
    stf_fd_temp.g_optics = at_best;
    stf_fd_temp.xt_detector = stf_fit->xt_detector;
    stf_fd_temp.at_detector = stf_fit->at_detector;
    stf_fd_temp.integ = stf_fit->integration_milliseconds;
    stf_fd_temp.diffusion = stf_fit->diffusion;
    stf_fd_temp.phase = 0.0;

    /* Calculate the AT STF */
    create_frequency_domain_stf_model(&stf_fd_temp, gsd, range, gvel, half_pi, 
        nsamp, oversamp, model_buf);

    /* Use the IFFT to convert STF to cross-track PSF */
    ias_math_fft2d(&model_buf[0].re, nsamp, 1, IAS_MATH_IFFT);

    /* Take the real part and scale to convert to a real output */
    for (ifreq = 0; ifreq < nsamp; ifreq++)
    {
        jfreq = (ifreq + nsamp / 2) % nsamp;
        psf_buf[ifreq] = model_buf[jfreq].re / nsamp;

        /* Integrate the PSF to get the ESF */
        if (ifreq > 0)
        {
            psf_buf[ifreq] += psf_buf[ifreq - 1];
        }
    }

    /* Calculate the AT edge slope and edge extent from the ESF */
    status = calc_spatial_stats(gsd, nsamp, oversamp, psf_buf, 
        &result->at_slope_best, &result->at_extent_best);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Error calculating spatial stats");
        overall_status = ERROR;
    }

    /* Free the buffers */
    free(model_buf);
    free(psf_buf);

    return overall_status;
}

int calc_spatial_stats
(
    double gsd,             /* I: Ground sample distance in meters */
    int number_of_samples,  /* I: Length of ESF sequence */
    int oversamp,           /* I: ESF oversampling factor */
    double *esf_buf,        /* I: ESF data */
    double *edge_slope,     /* O: Edge slope */
    double *edge_extent     /* O: Edge extent */
)
{
    int index;              /* Index to scan the ESF */
    double pos[NUM_RESPONSE_POINTS] = {0.0, 0.0, 0.0, 0.0, 0.0};
                            /* 5, 40, 50, 60, and 95% response points */
    double resp[NUM_RESPONSE_POINTS] = {0.05, 0.40, 0.50, 0.60, 0.95}; 
                            /* Key response values */
    double lhee, uhee;      /* Lower and upper half edge extents */
    int i;                  /* Loop index */

    if ((esf_buf == NULL) || (edge_slope == NULL) || (edge_extent == NULL))
    {
        IAS_LOG_ERROR("esf_buf, edge_slope, and edge_extent cannot be NULL");
        return ERROR;
    }

    /* Find the key response points */
    for (i = 0; i < NUM_RESPONSE_POINTS; i++)
    {
        index = 1;
        while (index < number_of_samples && esf_buf[index] < resp[i])
        {
            index++;
        }

        if (index >= number_of_samples)
        {
            IAS_LOG_ERROR("array esf_buf overrun by index");
            return ERROR;
        }

        pos[i] = (double)(index - 1) + (resp[i] - esf_buf[index - 1])
                 / (esf_buf[index] - esf_buf[index - 1]);
    }

    /* Calculate the edge slope from the 40% and 60% response points */
    *edge_slope = (double)oversamp * (resp[3] - resp[1]) / (pos[3] - pos[1]) 
                  / gsd;

    /* Calculate the edge extent from the 5%, 50%, and 95% response points */
    lhee = gsd / (double)oversamp * (pos[2] - pos[0]);
    uhee = gsd / (double)oversamp * (pos[4] - pos[2]);
    *edge_extent = (lhee > uhee ? lhee : uhee);

    return SUCCESS;
}
