/******************************************************************************
NAME:           fit_stf.c

PURPOSE:
Source code to estimate along- and across-track OLI System Transfer Function
(STF) optics parameters from a series of cross-section solutions at different 
angles.

RETURN VALUE:
SUCCESS     Fit was successful
ERROR       Error in solution

******************************************************************************/
#include <math.h>
#include "ias_math.h"     /* defines IAS math routines */
#include "ias_logging.h"  /* defines IAS_LOG macros */
#include "mtf.h"
#include "fit_stf.h"

#define CE90 2.1460         /* Circular Error of 90% */

/* Estimate the OLI STF parameters */
int fit_stf                 /* Return SUCCESS or ERROR */
(
    int numrec,             /* I: Number of trending records */
    const TRENDREC *trend,  /* I: Array of trending records */
    int band_number,        /* I: Current band number */
    STF_FIT_PARMS *stf_fit, /* O: Best fit STF parameters */
    double *xt_best,        /* O: 90% value of XT optics parameter */
    double *at_best         /* O: 90% value of AT optics parameter */
)
{
    double normeq[2][2];    /* Normal equations matrix */
    double rhs[2];          /* Right hand side vector */
    double atoptics;        /* Along-track optics parameter */
    double xtoptics;        /* Cross-track optics parameter */
    double atdet_m;         /* Mean along-track detector dimension */
    double xtdet_m;         /* Mean cross-track detector dimension */
    double intime_m;        /* Mean detector integration time */
    double diffuse_m;       /* Mean charge diffusion distance */
    double cossq;           /* Cosine squared of target angle */
    double sinsq;           /* Sine squared of target angle */
    double determ;          /* Determinant of normeq */
    double residual;        /* Observation residual */
    double stdev_unit_weight; /* Standard deviation of unit weight */
    double atbound;         /* 90% offset for along-track optics */
    double xtbound;         /* 90% offset for cross-track optics */
    double covariance[2][2]; /* Estimation covariance */
    double correlation;     /* XT/AT estimate correlation */
    int nobs;               /* Number of observations for this band */
    int i;                  /* Loop index */


    /* Initialize variables */
    nobs = 0;
    atdet_m = xtdet_m = intime_m = diffuse_m = 0.0;
    normeq[0][0] = normeq[0][1] = normeq[1][0] = normeq[1][1] = 0.0;
    rhs[0] = rhs[1] = 0.0;

    /* If NULL pointers were passed, return ERROR */
    if ((trend == NULL) || (stf_fit == NULL))
    {
        IAS_LOG_ERROR("trend and stf_fit cannot be NULL");
        return ERROR;
    }

    /* Examine all the incoming trending records */
    for (i = 0; i < numrec; i++)
    {
        if (trend[i].band_number != band_number)
        {
            continue;
        }

        /* Process the records for the current band */
        nobs++;
        cossq = cos(trend[i].angle);
        cossq *= cossq;
        sinsq = sin(trend[i].angle);
        sinsq *= sinsq;
        normeq[0][0] += cossq * cossq;
        normeq[0][1] += cossq * sinsq;
        normeq[1][0] += sinsq * cossq;
        normeq[1][1] += sinsq * sinsq;
        rhs[0] += cossq * trend[i].optics * trend[i].optics;
        rhs[1] += sinsq * trend[i].optics * trend[i].optics;
        atdet_m += trend[i].atdetector;
        xtdet_m += trend[i].xtdetector;
        intime_m += trend[i].inttime;
        diffuse_m += trend[i].diffuse;
    }

    /* If insufficient points were found, return */
    if (nobs < 2)
    {
        IAS_LOG_ERROR("Only %d points found for band number %d - "
                      "not enough for a solution", nobs, band_number);
        return ERROR;
    }

    /* Calculate simple averages */
    atdet_m /= nobs;
    xtdet_m /= nobs;
    intime_m /= nobs;
    diffuse_m /= nobs;

    /* Calculate least squares fit */
    determ = normeq[0][0] * normeq[1][1] - normeq[0][1] * normeq[1][0];
    if (fabs(determ) < 1.0e-6)
    {
        IAS_LOG_ERROR("Ill-conditioned solution - "
                      "need observations at more angles");
        return ERROR;
    }
    xtoptics = (normeq[1][1] * rhs[0] - normeq[0][1] * rhs[1]) / determ;
    atoptics = (normeq[0][0] * rhs[1] - normeq[1][0] * rhs[0]) / determ;

    /* Check the unconstrained solution */
    if (xtoptics < 0.0)
    {
        IAS_LOG_WARNING("Unconstrained solution for across-track optics "
            "parameter: %lf / %lf (across/along) has failed < 0 check - "
            "applying constraints", xtoptics, atoptics);
        xtoptics = 0.0;
        atoptics = rhs[1] / normeq[1][1];
    }
    if (atoptics < 0.0)
    {
        IAS_LOG_WARNING("Unconstrained solution for along-track optics "
            "parameter: %lf / %lf (across/along) has failed < 0 check - "
            "applying constraints", xtoptics, atoptics);
        atoptics = 0.0;
        xtoptics = rhs[0] / normeq[0][0];
    }

    /* Calculate residuals for the relevant records and accumulate standard
       deviation of unit weight */
    stdev_unit_weight = 0.0;
    if (nobs > 2)
    {
        nobs = 0;
        for (i = 0; i < numrec; i++)
        {
            if (trend[i].band_number != band_number)
            {
                continue;
            }

            /* Process the records for the current band */
            nobs++;
            cossq = cos(trend[i].angle);
            cossq *= cossq;
            sinsq = sin(trend[i].angle);
            sinsq *= sinsq;
            residual = trend[i].optics * trend[i].optics - (cossq * xtoptics 
                + sinsq * atoptics);
            stdev_unit_weight += residual * residual;
        }
        stdev_unit_weight = sqrt(stdev_unit_weight / (nobs - 2));
    }
    else
    {
        IAS_LOG_WARNING("Not enough observations (%d) to generate estimates "
            "of precision", nobs);
    }

    /* Calculate covariance */
    covariance[0][0] =  stdev_unit_weight * normeq[1][1] / determ;
    covariance[0][1] = -stdev_unit_weight * normeq[0][1] / determ;
    covariance[1][0] = -stdev_unit_weight * normeq[1][0] / determ;
    covariance[1][1] =  stdev_unit_weight * normeq[0][0] / determ;
    correlation = sqrt((covariance[0][1] * covariance[1][0] )
        / (covariance[0][0] * covariance[1][1]));

    IAS_LOG_INFO("XT variance: %e", covariance[0][0]);
    IAS_LOG_INFO("AT variance: %e", covariance[1][1]);
    IAS_LOG_INFO("XT/AT correlation coefficient: %9.6lf", correlation);

    /* Calculate ranges */
    atbound = CE90 * covariance[1][1];
    atbound = IAS_MAX(0.0, atoptics - atbound);
    xtbound = CE90 * covariance[0][0];
    xtbound = IAS_MAX(0.0, xtoptics - xtbound);

    /* Take square roots */
    atoptics = sqrt(atoptics);
    xtoptics = sqrt(xtoptics);
    atbound = sqrt(atbound);
    xtbound = sqrt(xtbound);

    /* Load the outgoing variables */
    stf_fit->xt_optics = xtoptics;
    stf_fit->at_optics = atoptics;
    stf_fit->xt_detector = xtdet_m;
    stf_fit->at_detector = atdet_m;
    stf_fit->integration_milliseconds = intime_m;
    stf_fit->diffusion = diffuse_m;
    *xt_best = xtbound;
    *at_best = atbound;

    return SUCCESS;
}

