/******************************************************************************
NAME:           get_perform_parms

PURPOSE:
Get the mtfperform processing parameters from the input parameter file.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

******************************************************************************/
#include <libgen.h>
#include <limits.h>  /* defines PATH_MAX */
#include <string.h>
#include <stdlib.h>

#include "ias_satellite_attributes.h"  /* defines IAS_MAX_NBANDS */
#include "mtfperform.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"

int get_perform_parms
(
    const char *param_filename, /* I: Parameter file name */
    char *trend_filename,       /* O: Name of trending data file */
    char *report_filename,      /* O: Name of output report file */
    int *numbands,              /* O: Number of bands to process */
    int band_list[],            /* O: Bands to process */
    int *nsamp,                 /* O: Number of samples in PSF/ESF */
    int *oversamp               /* O: PSF/ESF oversampling factor */
)
{
    int count;                  /* Number of elements */
    int band_index;             /* Index for band list */
    int status;                 /* Status returned from functions */

    /* Local variables for storing parameter values */
    char local_trend_filename[PATH_MAX];   /* file name to read */
    char local_report_filename[PATH_MAX];  /* file name to write */
    int local_band_list[IAS_MAX_NBANDS];
    int local_satellite_number = -1;
    IAS_SATELLITE_ID satellite_id;

    const int MIN_NUM_SAMPLES = 128;
    const int MIN_OVERSAMPLING = 8;
    int i;

    if ((param_filename == NULL) || (trend_filename == NULL) ||
        (report_filename == NULL) || (numbands == NULL) ||
        (band_list == NULL) || (nsamp == NULL) || (oversamp == NULL))
    {
        IAS_LOG_ERROR("NULL pointer passed for one or more input parameters");
        return ERROR;
    }

    *nsamp = -1;
    *oversamp = -1;

    /* default to invalid bands */
    *numbands = IAS_MAX_NBANDS;
    for (i = 0; i < *numbands; i++)
    {
        local_band_list[i] = -1;
    }

    strcpy(local_trend_filename, "");
    strcpy(local_report_filename, "");

    /* Start the parameter file table */
    IAS_PARM_DECLARE_TABLE(parms, 6);

    /* Bands and SCAs are optional parameters.  If they aren't there, the full
       list/range is used. */

    /* Get the list of bands. */
    IAS_PARM_BAND_LIST(parms, IAS_PARM_OPTIONAL, local_band_list,
                       sizeof(local_band_list), 1);

    /* Get the nsamp_pts. */
    int default_nsamp_pts[] = {512};
    IAS_PARM_ADD_INT(parms, PERFORM_NSAMP, "Number of Perform Samples", 
                     IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 1, 10000, 1,
                     default_nsamp_pts, nsamp, sizeof(*nsamp), 0);

    /* Get the oversamp */
    int default_oversamp[] = {30};
    IAS_PARM_ADD_INT(parms, PERFORM_OVERSAMP, "PSF/ESF Oversampling Factor", 
                     IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 
                     MIN_OVERSAMPLING, 32, 1, default_oversamp, oversamp, 
                     sizeof(*oversamp), 0);

    const char *default_trend_filename[] = {"mtfestimate_trend.dat"};
    IAS_PARM_ADD_STRING(parms, TREND_FILENAME,
                        "File containing input trending records",
                        IAS_PARM_OPTIONAL, 0, NULL, 1, default_trend_filename,
                        local_trend_filename, sizeof(local_trend_filename), 0);

    const char *default_report_filename[] = {"mtf_performance.rpt"};
    IAS_PARM_ADD_STRING(parms, REPORT_FILE,
                        "Output summary report file name",
                        IAS_PARM_OPTIONAL, 0, NULL, 1, default_report_filename,
                        local_report_filename, sizeof(local_report_filename),
                        0);

    /* Get the satellite number */
    IAS_PARM_SATELLITE(parms, &local_satellite_number,
                       sizeof(local_satellite_number), 1);

    count = IAS_PARM_GET_TABLE_SIZE(parms);

    /* Check if help or template output was requested */
    status = ias_parm_provide_help(param_filename, parms, count,
                                   IAS_INPUT_PARAMETERS);

    if (status == 1)
    {
        /* the --help or --template was provided. can now exit success */
        exit(EXIT_SUCCESS);
    }
    else if (status == ERROR) /* Help failed. */
    {
        IAS_LOG_ERROR("Providing help for application parameter file %s",
                      param_filename);
        exit(EXIT_FAILURE);
    }

    if (ias_parm_read(param_filename, parms, count) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from: %s", param_filename);
        return ERROR;
    }

   /* Convert the satellite number to the satellite id */
    satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(
                       local_satellite_number);
    if (satellite_id == ERROR)
    {
        IAS_LOG_ERROR("Unable to determine satellite ID from satellite "
                      "number %d", local_satellite_number);
        return ERROR;
    }

    /* Initialize the satellite attributes */
    if (ias_sat_attr_initialize(satellite_id) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attributes library");
        return ERROR;
    }

    /* Copy parameters */
    strcpy(trend_filename, local_trend_filename);
    strcpy(report_filename, local_report_filename);

    *numbands = IAS_PARM_GET_COUNT_READ(parms, BAND_LIST);
    for (count = 0; count < *numbands; count++)
    {
        band_index = ias_sat_attr_convert_band_number_to_index(
            local_band_list[count]);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Invalid band index for band number: %d", 
                local_band_list[count]);
            return ERROR;
        }

        band_list[count] = local_band_list[count];
    }

    if (*nsamp < MIN_NUM_SAMPLES)
    {
        IAS_LOG_ERROR("Invalid number of samples in perform, %d, must be at "
                      "least %d", *nsamp, MIN_NUM_SAMPLES);
        return ERROR;
    }
    else
    {
        IAS_LOG_DEBUG("number of samples in perform = %d", *nsamp);
    }

    if (*oversamp < MIN_OVERSAMPLING)
    {
        IAS_LOG_ERROR("Invalid perform oversample factor, %d, must be at "
                      "least %d", *oversamp, MIN_OVERSAMPLING);
        return ERROR;
    }
    else
    {
        IAS_LOG_DEBUG("perform oversample factor = %d", *oversamp);
    }

    /* Done with the ODL file so close it.  */

    return SUCCESS;
}
