/******************************************************************************
NAME:           mtfperform

PURPOSE:
Main procedure for software to load and analyze OLI STF parameter sets
computed at multiple target angles to determine the best fit along- and
acorss-track STF parameters. These parameters are then used to calculate
spatial edge response performance parameters.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
ERROR           Error estimating spatial performance
SUCCESS         Successfully generated spatial performance estimates

******************************************************************************/
#include <string.h>                    /* defines several string functions */
#include <stdlib.h>                    /* defines free */
#include <libgen.h>                    /* defines dirname */
#include <limits.h>                    /* defines PATH_MAX */
#include "ias_logging.h"               /* defines IAS_LOG macros */
#include "ias_math.h"                  /* defines IAS math routines */
#include "ias_math_complex.h"          /* defines complex math routines */
#include "ias_satellite_attributes.h"  /* defines IAS_MAX_NBANDS */
#include "mtfperform.h"

/* Defined variables */
#define STD_RANGE 705000.0  /* Nominal distance to Earth target (meters) */
#define STD_GVEL 6754.542   /* Nominal ground velocity (meters/second) */

int main(int argc, char *argv[])
{
    char trend_file[PATH_MAX];  /* Input trending data file name */
    char report_file[PATH_MAX]; /* Output report file name */
    int iband;                  /* Integer band index */
    int swir1_index = -1;       /* Index of SWIR1 band */
    int swir2_index = -1;       /* Index of SWIR2 band */
    int cirrus_index = -1;      /* Index of CIRRUS band */
    int populate_cirrus_flag = 0; /* Flag for populating the CIRRUS band based
                                   on the SWIR1 and SWIR2 bands */
    IAS_BAND_TYPE band_type;    /* Type of band */
    int numbands;               /* Number of bands to process */
    int band_list[IAS_MAX_NBANDS]; /* Input band list (user band numbers) */
    int nsamp;                  /* Number of samples in PSF(s) */
    int oversamp;               /* Oversample factor */
    int numrec;                 /* Number of trending records to process */
    TRENDREC *trend;            /* Array of trending data structures */
    STF_FIT_PARMS stf_fit;      /* xoptics, aoptics, xdet, adet, inttime */
    double xt_best;             /* Low 90% value for xoptics */
    double at_best;             /* Low 90% value for aoptics */
    double range = STD_RANGE;   /* Standard distance to target (meters) */
    double gvel = STD_GVEL;     /* Standard ground velocity over target
                                   (meters/sec) */
    SPATIAL results[IAS_MAX_NBANDS]; /* Array of structures for spatial
                                            parameter results */
    char output_dir[PATH_MAX];  /* output directory for */
    int status = ERROR;
    char *parmfile = NULL;
    char *progname = strrchr(argv[0], '/');
    if (progname == NULL)
    {
        progname = argv[0];
    }
    else
    {
        progname++; /* skip slash */
    }

    status = ias_log_initialize(progname);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing logging library");
        return EXIT_FAILURE;
    }

    /* Get input parameters */
    if (argc < 2)
    {
        IAS_LOG_ERROR("usage:  %s <input ODL parameter file>", progname);
        return EXIT_FAILURE;
    }

    parmfile = argv[1];

    status = get_perform_parms(parmfile, trend_file, report_file,
                               &numbands, band_list, &nsamp, &oversamp);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Unable to read input parameter file %s", parmfile);
        return EXIT_FAILURE;
    }

    IAS_LOG_INFO("Initiating %s", progname);

    /* Create and open the output report file */
    status = create_spatial_report(report_file);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Error creating output report file %s", report_file);
        return EXIT_FAILURE;
    }

    strncpy(output_dir, dirname(report_file), sizeof(output_dir));

    /* Loop through the requested bands */
    for (iband = 0; iband < numbands; iband++)
    {
        /* Initialize the spatial results for this band */
        results[iband].band_number = band_list[iband];
        band_type = ias_sat_attr_get_band_type_from_band_number(
            results[iband].band_number);
        if (band_type == IAS_UNKNOWN_BAND_TYPE)
        {
            IAS_LOG_ERROR("Getting band type of band number %d", 
                results[iband].band_number);
            return EXIT_FAILURE;
        }
        if (band_type == IAS_SWIR1_BAND)
        {
            swir1_index = iband;
        }
        if (band_type == IAS_SWIR2_BAND)
        {
            swir2_index = iband;
        }
        if (band_type == IAS_CIRRUS_BAND)
        {
            cirrus_index = iband;
        }
        results[iband].numrec = 0;
        results[iband].xt_slope = 0.0;
        results[iband].xt_slope_best = 0.0;
        results[iband].xt_extent = 0.0;
        results[iband].xt_extent_best = 0.0;
        results[iband].at_slope = 0.0;
        results[iband].at_slope_best = 0.0;
        results[iband].at_extent = 0.0;
        results[iband].at_extent_best = 0.0;

        /* Load the trending information for the current band */
        numrec = get_trend_band(trend_file, band_list[iband], &trend);
        if (numrec == 0)
        {
            IAS_LOG_INFO("No data for band %d found in trending file %s",
                            band_list[iband], trend_file);

            /* Use results from other two SWIR bands for cirrus */
            if (band_type == IAS_CIRRUS_BAND)
            {
                populate_cirrus_flag = 1;
            }
            continue;
        }
        results[iband].numrec = numrec;

        IAS_LOG_INFO("Processing %d records from band number %d.",
                     numrec, band_list[iband]);

        /* Estimate the OLI STF parameters */
        status = fit_stf(numrec, trend, band_list[iband], &stf_fit, &xt_best,
            &at_best);
        if (status != SUCCESS)
        {
            /* Skip this band which will give it fully degraded values in the
               report, issue a warning and continue to process other bands */
            IAS_LOG_WARNING("Unable to estimate STF for band number %d",
                          band_list[iband]);
            free(trend);
            trend = NULL;
            continue;
        }

        /* Add fit results to the output report */
        status = report_stf(band_list[iband], numrec, trend, stf_fit);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Error adding STF parameters for band number %d to "
                          "report", band_list[iband]);
            free(trend);
            trend = NULL;
            close_spatial_report();
            return EXIT_FAILURE;
        }

        /* Calculate the PSF, ESF, and spatial performance parameters */
        status = calc_spatial_perf(output_dir, &stf_fit, trend[0].gsd, range,
                                   gvel, nsamp, oversamp, xt_best, at_best,
                                   &results[iband]);
        if (status == ERROR)
        {
            /* Issue a warning for this band and continue processing other 
               bands */
            IAS_LOG_WARNING("Could not calculate performance parameters for "
                          "band number %d", band_list[iband]);
            free(trend);
            trend = NULL;
            continue;
        }

        /* Release trending data table */
        free(trend);
        trend = NULL;

    }  /* Band Loop */

    if (populate_cirrus_flag)
    {
        /* Use results from other two SWIR bands for cirrus if they are 
           available */
        if (swir1_index >= 0 && swir2_index >= 0)
        {
            IAS_LOG_INFO("Using SWIR1 and SWIR2 results to estimate Cirrus "
                         "band performance");
            results[cirrus_index].xt_slope = IAS_MAX(
                    results[swir1_index].xt_slope,
                    results[swir2_index].xt_slope);
            results[cirrus_index].xt_extent = IAS_MIN(
                    results[swir1_index].xt_extent,
                    results[swir2_index].xt_extent);
            results[cirrus_index].at_slope = IAS_MAX(
                    results[swir1_index].at_slope,
                    results[swir2_index].at_slope);
            results[cirrus_index].at_extent = IAS_MIN(
                    results[swir1_index].at_extent,
                    results[swir2_index].at_extent);
            results[cirrus_index].xt_slope_best = IAS_MAX(
                    results[swir1_index].xt_slope_best,
                    results[swir2_index].xt_slope_best);
            results[cirrus_index].xt_extent_best = IAS_MIN(
                    results[swir1_index].xt_extent_best,
                    results[swir2_index].xt_extent_best);
            results[cirrus_index].at_slope_best = IAS_MAX(
                    results[swir1_index].at_slope_best,
                    results[swir2_index].at_slope_best);
            results[cirrus_index].at_extent_best = IAS_MIN(
                    results[swir1_index].at_extent_best,
                    results[swir2_index].at_extent_best);
        }
    }

    /* Write out the performance summary report */
    status = report_performance(numbands, results);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Error writing spatial performance summary report");
        close_spatial_report();
        return EXIT_FAILURE;
    }

    status = close_spatial_report();
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Error closing spatial performance summary report");
        return EXIT_FAILURE;
    }

    /* Print the completion message.  */
    IAS_LOG_INFO("Successful completion");

    return EXIT_SUCCESS;
}

