/******************************************************************************
NAME:           report.c

PURPOSE:
Source file for software to generate output spatial performance report file.

RETURN VALUE:
type = int

VALUE       DESCRIPTION
----------- ----------------------------------------------------------
SUCCESS     File opened for output
ERROR       Failure to open file for output
******************************************************************************/
#include <math.h>   /* defines floor() */

#include "ias_logging.h"
#include "report.h"

/* Global variable */
static FILE *r_ptr = NULL;  /* Output report file pointer */

/* Functions */
/* Create and open the output report file */
int create_spatial_report   /* Return SUCCESS or ERROR */
(
    char *report_file       /* I: Name of report file to create */
)
{
    if (report_file == NULL)
    {
        IAS_LOG_ERROR("report_file cannot be NULL");
        return ERROR;
    }

    if (r_ptr != NULL)
    {
        IAS_LOG_ERROR("Report file already open");
        return ERROR;
    }

    /* Open the output file */
    r_ptr = fopen(report_file, "w");
    if (r_ptr == NULL)
    {
        IAS_LOG_ERROR("Error opening output report file %s", report_file);
        return ERROR;
    }

    /* Write out introductory/header information */
    if (fprintf(r_ptr, "OLI On-Orbit Spatial Performance Characterization "
            "Report\n\n") < 0)
    {
        IAS_LOG_ERROR("Writing to report file %s", report_file);
        return ERROR;
    }

    return SUCCESS;
}

/* Write fitted STF info to the output report file */
int report_stf              /* Return SUCCESS or ERROR */
(
    int band_number,        /* I: Current band number */
    int numrec,             /* I: Number of records used in fit */
    TRENDREC *trend,        /* I: Array of trending results used */
    STF_FIT_PARMS stf_fit   /* I: Best fit STF parameters */
)
{
    int i;                  /* Loop index */

    if (r_ptr == NULL)
    {
        IAS_LOG_ERROR("Report file not opened");
        return ERROR;
    }

    /* Header for this band and summary of trending input used */
    if (fprintf(r_ptr, "\nSTF Fit Results for Band %d\n\n"
            "    Fit Based on %d Trending Records:\n"
            "    Rec #   Year   DOY   WRS_Path   WRS_Row    Angle"
            "      Optics     Target\n", band_number, numrec) < 0)
    {
        IAS_LOG_ERROR("Writing to report file");
        return ERROR;
    }

    for (i = 0; i < numrec; i++)
    {
        if (fprintf(r_ptr, "     %3d    %04d   %03d     %03d        %03d    "
                "%9.6lf    %6.3lf     %s\n", i + 1, trend[i].year,
                trend[i].doy, trend[i].wrs_path, trend[i].wrs_row,
                trend[i].angle, trend[i].optics, trend[i].target_name) < 0)
        {
            IAS_LOG_ERROR("Writing to report file");
            return ERROR;
        }
    }

    /* The fitted STF parameters */
    if (fprintf(r_ptr, "\nOLI STF Fitted Parameters:\n"
            "    Cross-Track Optics:   %6.3lf urad     Along-Track "
            "Optics:   %6.3lf urad\n"
            "    Cross-Track Detector: %6.3lf urad     Along-Track "
            "Detector: %6.3lf urad\n"
            "    Charge Diffusion:     %6.3lf urad     Integration "
            "Time:     %6.3lf msec\n\n", stf_fit.xt_optics, stf_fit.at_optics,
            stf_fit.xt_detector, stf_fit.at_detector, stf_fit.diffusion, 
            stf_fit.integration_milliseconds) < 0)
    {
        IAS_LOG_ERROR("Writing to report file");
        return ERROR;
    }

    return SUCCESS;
}

/* Write spatial performance parameters to the output report file */
int report_performance  /* Return SUCCESS or ERROR */
(
    int nband,          /* I: Number of bands analyzed */
    SPATIAL *stats      /* I: Array of results structures */
)
{
    int stats_band_index; /* Band index into stats */
    int band_index;     /* Specification band index */
    /* OLI specification values from OLI Requirements Document GSFC 427-05-03
       Rev. E */
    double edge_slope_spec[] =  {0.027, 0.027, 0.027, 0.027, 0.027, 0.027, 
                         0.027, 0.054, 0.027};
    double half_edge_extent_spec[] = { 23.0, 23.0, 23.0, 23.5, 24.0, 28.0, 
                         29.0, 14.0, 27.0};
    /* KPR evaluation variables */
    int standard_perf = 0; /* Standard performance flag */
    int xt_percent;        /* XT ES % of spec value */
    int at_percent;        /* AT ES % of spec value */
    double xt_penalty;     /* XT penalty for missing spec */
    double at_penalty;     /* AT penalty for missing spec */
    double tot_penalty;    /* Total penalty for KPR #4 */

    if (r_ptr == NULL)
    {
        IAS_LOG_ERROR("Report file not opened");
        return ERROR;
    }

    /* Header for the edge response section and header for output table */
    if (fprintf(r_ptr, "\n\nEdge Slope and Half Edge Extent Spatial "
            "Performance by Band\n\n"
            "Band     NObs      XT_ES     XT_HEE      AT_ES     AT_HEE"
            "     ES_Spec     HEE_Spec\n") < 0)
    {
        IAS_LOG_ERROR("Writing to report file");
        return ERROR;
    }

    /* Loop through the bands */
    for (stats_band_index = 0; stats_band_index < nband; stats_band_index++)
    {
        band_index = ias_sat_attr_convert_band_number_to_index(
            stats[stats_band_index].band_number);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Invalid band index for band number: %d",
                stats[stats_band_index].band_number);
            return ERROR;
        }

        /* Write the results for this band */
        if (fprintf(r_ptr, "  %d      %3d     %8.5lf    %6.3lf    %8.5lf    "
                "%6.3lf     >%6.3lf      <%5.1lf\n", 
                stats[stats_band_index].band_number,
                stats[stats_band_index].numrec, 
                stats[stats_band_index].xt_slope,
                stats[stats_band_index].xt_extent, 
                stats[stats_band_index].at_slope,
                stats[stats_band_index].at_extent, 
                edge_slope_spec[band_index], 
                half_edge_extent_spec[band_index]) < 0)
        {
            IAS_LOG_ERROR("Writing to report file");
            return ERROR;
        }
    }

    /* Header for the best case edge response section and header for output
       table */
    if (fprintf(r_ptr, "\n\n\n\nBest Case Edge Slope and Half Edge Extent "
        "Spatial Performance by Band\n\n"
        "Band     NObs      XT_ES     XT_HEE      AT_ES     AT_HEE"
        "     ES_Spec     HEE_Spec\n") < 0)
    {
        IAS_LOG_ERROR("Writing to report file");
        return ERROR;
    }

    /* Loop through the bands */
    for (stats_band_index = 0; stats_band_index < nband; stats_band_index++)
    {
        band_index = ias_sat_attr_convert_band_number_to_index(
            stats[stats_band_index].band_number);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Invalid band index for band number: %d",
                stats[stats_band_index].band_number);
            return ERROR;
        }

        /* Write the results for this band */
        if (fprintf(r_ptr, "  %d      %3d     %8.5lf    %6.3lf    %8.5lf    "
            "%6.3lf     >%6.3lf      <%5.1lf\n",
            stats[stats_band_index].band_number, 
            stats[stats_band_index].numrec, 
            stats[stats_band_index].xt_slope_best,
            stats[stats_band_index].xt_extent_best, 
            stats[stats_band_index].at_slope_best, 
            stats[stats_band_index].at_extent_best, 
            edge_slope_spec[band_index], 
            half_edge_extent_spec[band_index]) < 0)
        {
            IAS_LOG_ERROR("Writing to report file");
            return ERROR;
        }
    }

    /* Write KPR evaluation table */
    tot_penalty = 0.0;
    if (fprintf(r_ptr, "\n\n"
        "KPR #4 - OLI Edge Response Slope\nUsing Best Case Results\n\n"
        "  Band     XT_ES      %% Degraded      Penalty     AT_ES"
        "     %% Degraded     Penalty\n") < 0)
    {
        IAS_LOG_ERROR("Writing to report file");
        return ERROR;
    }

    for (stats_band_index = 0; stats_band_index < nband; stats_band_index++)
    {
        band_index = ias_sat_attr_convert_band_number_to_index(
            stats[stats_band_index].band_number);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Invalid band index for band number: %d",
                stats[stats_band_index].band_number);
            return ERROR;
        }

        /* Cross track degradation and KPR penalty */
        xt_percent = 100 - (int)floor(stats[stats_band_index].xt_slope_best 
            * 100.0 / edge_slope_spec[band_index] + 0.5);
        if (xt_percent < 0)
        {
            xt_percent = 0;
        }
        if (xt_percent > 25)
        {
            xt_penalty = 5.5;
        }
        else
        {
            xt_penalty = 0.0;
        }
        tot_penalty += xt_penalty;

        /* Evaluate standard performance threshold */
        if (xt_percent > 10)
        {
            standard_perf++;
        }

        /* Along track degradation and KPR penalty */
        at_percent = 100 - (int)floor(stats[stats_band_index].at_slope_best 
            * 100.0 / edge_slope_spec[band_index] + 0.5);
                                      
        if (at_percent < 0)
        {
            at_percent = 0;
        }
        if (at_percent > 25)
        {
            at_penalty = 5.5;
        }
        else 
        {
            at_penalty = 0.0;
        }
        tot_penalty += at_penalty;

        /* Evaluate standard performance threshold */
        if (at_percent > 10)
        {
            standard_perf++;
        }

        /* Write out band results */
        if (fprintf(r_ptr, "   %d     %8.5lf          %2d          %4.1lf     "
                "%8.5lf         %2d         %4.1lf\n",
                stats[stats_band_index].band_number, 
                stats[stats_band_index].xt_slope_best, 
                xt_percent, xt_penalty, stats[stats_band_index].at_slope_best, 
                at_percent, at_penalty) < 0)
        {
            IAS_LOG_ERROR("Writing to report file");
            return ERROR;
        }
    }

    if (standard_perf > 0)
    {
        if (fprintf(r_ptr, "\nStandard Performance Level is NOT Met for %d "
            "Bands/Directions.\n\nTotal KPR Penalty: %4.1lf Percent\n\n\n", 
            standard_perf, tot_penalty) < 0)
        {
            IAS_LOG_ERROR("Writing to report file");
            return ERROR;
        }
            
    }
    else
    {
        if (fprintf(r_ptr, "\nStandard Performance Level is Met!\n"
            "\nTotal KPR Penalty: %4.1lf Percent\n\n\n", tot_penalty) < 0)
        {
            IAS_LOG_ERROR("Writing to report file");
            return ERROR;
        }
    }

    return SUCCESS;
}

/******************************************************************************
NAME:           close_spatial_report

PURPOSE:
Close the report file.

RETURN VALUE:
type = int

VALUE       DESCRIPTION
----------- ----------------------------------------------------------
SUCCESS     File closed 
ERROR       Failure to close file 
******************************************************************************/
int close_spatial_report
(
)
{
    int status;            /* Function return status */

    if (r_ptr == NULL)
    {
        IAS_LOG_ERROR("Report file not opened");
        return ERROR;
    }

    /* Close the file */
    status = fclose(r_ptr);
    if (status != 0)
    {
        IAS_LOG_ERROR("Closing spatial report file");
        return ERROR;
    }

    r_ptr = NULL;

    return SUCCESS;
}

