#!/usr/bin/perl -w
###############################################################################
#   NAME:   retrieve_mtf_db.pl
#
#   PURPOSE: Produce a report of MTF records that can be used as input to
#            MTF Profile
#
#   OUTPUT:  Report file
###############################################################################

use strict;                 # enforce the rules
use Getopt::Long;           # for the optional parameters
use Logger;                 # to log errors, warnings, info msgs
use ias_const;              # for things like SUCCESS, FAILURE
use ias_db;                 # for database access
use ias_processing_system;  # ias_init_system and ias_validate_environment

# Define list of environment variables needed by ias_db.pm that we should
# check up-front to bail out sooner if our environment isn't setup as required
my @envVarsList = qw(
        IAS_DB_L8_OLITIRS
        ORACLE_SID
        );
my $envVarsListRef = \@envVarsList;

# Variable declarations
#--------------------------------------------------------------------
my $i;                                  # loop variable
my $j;                                  # loop variable
my $status;                             # return status of sub-routines
my $errStr = "";                        # error string
my $sql;                                # SQL command string built in part
                                        # based on user selections
my %resultsHofA = ();                   # results from DB query, keyed by DB
                                        # field name (Hash of Arrays)
my %spanResultsHofA = ();               # results from DB query, keyed by DB
                                        # field name (Hash of Arrays)
my $report_filename = '';               # name of report file to write;
                                        # default mtfestimate_trend.dat
my $file_handle;                        # report file handle
my $start_date;                         # starting date in mm-dd-yyyy format 
my $end_date;                           # ending date in mm-dd-yyyy
my $opt_min;                         # minimum allowable Gaussian optical range
my $opt_max;                         # maximum allowable Gaussian optical range
my $help;                            # variable for help option
my $query_option = '';               # query option for WHERE clause in SQL
                                     # query depending upon the inputted 
                                     # command line arguments

#--------------------------------------------------------------------
# Initialize the MTF DB retrieval tool 
#--------------------------------------------------------------------

our ($dbh, $errorLog) = &init($envVarsListRef);

# Set up some constraints if requested
if (!GetOptions ( 'help' =>\$help,
                 'opt_min=f'    => \$opt_min,
                 'opt_max=f'    => \$opt_max,
                 'start_date=s' => \$start_date,
                 'end_date=s'   => \$end_date))
{
    usage();
    exit FAILURE;
}

# --help/-h/--h/-help option for the usage of the script
if (defined($help))
{
   usage();
   exit FAILURE;
}

# Validate $opt_min is a floating point value
if (defined($opt_min))
{
   $opt_min = 1.0 * $opt_min;
}

# Validate $opt_max is a floating point value
if (defined($opt_max))
{
    $opt_max = 1.0 * $opt_max;
}

# Validate the format of starting date in mm-dd-yyyy
if (defined($start_date))
{
    if (!($start_date =~ 
            /^(1[0-2]|0[1-9])-(3[01]|[12][0-9]|0[1-9])-(\d{4})$/))
    {
        usage();
        exit FAILURE;
    }
}

# Validate the format of ending date in mm-dd-yyy
if (defined($end_date))
{
    if (!($end_date =~
            /^(1[0-2]|0[1-9])-(3[01]|[12][0-9]|0[1-9])-(\d{4})$/))
    {
        usage();
        exit FAILURE;
    }
}

# Grab the report file, if present
$report_filename = $ARGV[0];

# If the report filename wasn't specified, use the default 
if (!defined($report_filename) || !length($report_filename))
{
    $report_filename = "mtfestimate_trend.dat";
}

# Open the report file for writing
$status = open($file_handle, ">$report_filename");
if (!$status)
{
    ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
    $errorLog->fatal_msg(__FILE__, __LINE__, "Could not open report file " .
        "$report_filename for writing: $!");
    $errorLog->close_log();
    exit FAILURE;
}

#--------------------------------------------------------------------
# Query for the records from the database
#-------------------------------------------------------------------- 

# SQL WHERE clause statement for the command line parameters

if (defined($opt_min))
{
    $query_option = " mtf_bridge.g_optics >= $opt_min and ";
}

if (defined($opt_max))
{
    $query_option = $query_option . " mtf_bridge.g_optics <= $opt_max and ";
}

if (defined($start_date))
{
    $query_option = $query_option . 
                    " geometric_work_order_common.date_acquired >=" .
                    " to_date('$start_date','mm-dd-yyyy') and ";
}

if (defined($end_date))
{
    $query_option = $query_option .
                    " geometric_work_order_common.date_acquired <=" .
                    " to_date('$end_date','mm-dd-yyyy') and ";
}

# Build the SQL command
$sql= qq{ SELECT band_id, target_name, year, day_of_year, 
          wrs_path, wrs_row, angle, g_optics, at_detector, xt_detector, 
          integ, diffusion, phase, rms, gsd, range, velocity, rad_water, 
          rad_asym, nbr_of_spans,
          geometric_work_order_common.geometric_wo_common_id,
          work_order_id
        FROM mtf_bridge, geometric_work_order_common
        WHERE $query_option
          mtf_bridge.geometric_wo_common_id = 
          geometric_work_order_common.geometric_wo_common_id
        ORDER BY work_order_id, target_name, band_id 
        };

# Retrieve the requested records
$status = &retrieveRecords($dbh, $sql, \$errStr, \%resultsHofA);
if ($status != SUCCESS)
{
    ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
    $errorLog->fatal_msg(__FILE__, __LINE__, $errStr);
    $errorLog->close_log();
    exit FAILURE;
}

#--------------------------------------------------------------------
# Print a report in pipe separated values format containing the set
# of records returned by the database query.
#--------------------------------------------------------------------

# Determine the number of records returned by the database query. Use a
# column from the query that should always have a value populated in it
# for every record that was entered to the DB.
my $arrayRef = $resultsHofA{BAND_ID};
if (not defined($arrayRef))
{
    # Unexpectedly found either no rows for the query in the database, or
    # the rows returned didn't have the work order IDs set. Either way,
    # this would be an invalid situation to continue this script under.
    ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
    $errorLog->fatal_msg(__FILE__, __LINE__, 'No records found from query.');
    $errorLog->close_log();
    exit FAILURE;
}
my $numRecords = @$arrayRef;

#--------------------------------------------------------------------
# Print the database query results to the report
#--------------------------------------------------------------------

# Print the report contents
my $spanArrayRef;
for ($i = 0; $i < $numRecords; $i++)
{
    printf $file_handle ("%s | %04d | %03d | %03d | %03d | %s | %d | %9.6lf " .
        "| %7.4lf | %6.3lf | %6.3lf | %6.3lf | %6.3lf | %11.4e | %6.3lf " .
        "| %6.3lf | %10.3lf | %8.3lf | %6.1lf | %4.1lf | %d | ",
        $resultsHofA{WORK_ORDER_ID}[$i], 
        $resultsHofA{YEAR}[$i], 
        $resultsHofA{DAY_OF_YEAR}[$i],
        $resultsHofA{WRS_PATH}[$i], 
        $resultsHofA{WRS_ROW}[$i],
        $resultsHofA{TARGET_NAME}[$i],
        $resultsHofA{BAND_ID}[$i],
        $resultsHofA{ANGLE}[$i],
        $resultsHofA{G_OPTICS}[$i],
        $resultsHofA{AT_DETECTOR}[$i],
        $resultsHofA{XT_DETECTOR}[$i],
        $resultsHofA{INTEG}[$i],
        $resultsHofA{DIFFUSION}[$i],
        $resultsHofA{PHASE}[$i],
        $resultsHofA{RMS}[$i],
        $resultsHofA{GSD}[$i],
        $resultsHofA{RANGE}[$i],
        $resultsHofA{VELOCITY}[$i],
        $resultsHofA{RAD_WATER}[$i],
        $resultsHofA{RAD_ASYM}[$i],
        $resultsHofA{NBR_OF_SPANS}[$i]);

    # Get the span records for this MTF record
    $sql= qq{ SELECT rad_bridge, width, offset
        FROM mtf_spans
            WHERE mtf_spans.geometric_wo_common_id = 
            $resultsHofA{GEOMETRIC_WO_COMMON_ID}[$i]
            AND mtf_spans.band_id = 
            $resultsHofA{BAND_ID}[$i]
            AND mtf_spans.target_name =
            '$resultsHofA{TARGET_NAME}[$i]'
            ORDER BY span_id
    };

    # retrieve the requested records for the spans
    $status = &retrieveRecords($dbh, $sql, \$errStr, \%spanResultsHofA);
    if ($status != SUCCESS)
    {
        ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
        $errorLog->fatal_msg(__FILE__, __LINE__, $errStr);
        $errorLog->close_log();
        exit FAILURE;
    }

    # Determine the number of records returned by the database query. Use a
    # column from the query that should always have a value populated in it
    # for every record that was entered to the DB.
    $spanArrayRef = $spanResultsHofA{RAD_BRIDGE};
    if (not defined($spanArrayRef))
    {
        # Unexpectedly found either no rows for the query in the database, or
        # the rows returned didn't have the work order IDs set. Either way,
        # this would be an invalid situation to continue this script under.
        ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
        $errorLog->fatal_msg(__FILE__, __LINE__, 'No records found from '
            . 'query.');
        $errorLog->close_log();
        exit FAILURE;
    }
    my $numSpanRecords = @$spanArrayRef;

    for ($j = 0; $j < $numSpanRecords - 1; $j++)
    {
        printf $file_handle ("%6.1lf | %5.2lf | %6.2lf | ",
            $spanResultsHofA{RAD_BRIDGE}[$j],
            $spanResultsHofA{WIDTH}[$j],
            $spanResultsHofA{OFFSET}[$j]);
    }

    # The last entry doesn't end in a pipe and it gets a newline
    printf $file_handle ("%6.1lf | %5.2lf | %6.2lf\n",
        $spanResultsHofA{RAD_BRIDGE}[$j],
        $spanResultsHofA{WIDTH}[$j],
        $spanResultsHofA{OFFSET}[$j]);

    # Clear the span results so it can be used again.
    delete $spanResultsHofA{RAD_BRIDGE};
    delete $spanResultsHofA{WIDTH};
    delete $spanResultsHofA{OFFSET};
}

# close the report file
$status = close($file_handle);
if (!$status)
{
    ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
    $errorLog->fatal_msg(__FILE__, __LINE__, "Could not close report file " .
        "$report_filename: $!");
    $errorLog->close_log();
    exit FAILURE;
}

# close the database connection since we don't need anything more from it;
# should have only done inquiries of the DB, so disconnect with a rollback
$status = ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
if (!$status)
{
    $errorLog->fatal_msg(__FILE__, __LINE__, $ias_db_error);
    $errorLog->close_log();
    exit FAILURE;
}

# done
$errorLog->close_log();
exit SUCCESS;

###############################################################################
# Name: init
#
# Description: Initializes the MTF DB retrieval tool
#
# Arguments:
#   Array reference to the list of environment variables
#
# Returns:
#   An array of values:
#       handle to the database connection
#       reference to the error logging object
#   If an error occurs, the routine terminates the application with an error
###############################################################################
sub init
{
    my ($varsListRef) = @_;

    # Initialize the ias_processing_system module.
    if(ias_init_system() == ERROR)
    {
        print STDERR "retrieve_mtf_db.pl: $ias_error_msg\n";
        exit FAILURE;
    }

    if (ias_validate_environment($varsListRef) > 0)
    {
        print STDERR "retrieve_mtf_db.pl: $ias_error_msg\n";
        exit FAILURE;
    }

    my $Logger = Logger->new("STDOUT");
    if (!$Logger)
    {
        print STDERR "retrieve_mtf_db.pl: ERROR: Failed to establish ";
        print STDERR "logger object.\n";
        exit FAILURE;
    }

    set_default_log($Logger);
    $Logger->set_module_name('retrieve_mtf_db.pl');

    # get the username and password for the database
    my ($userId, $password) = ias_db_GetDBConnectionInfo("L8_OLITIRS");

    # connect to the database
    my $dbHandle = ias_db_ConnectToDB($userId, $password);
    if (!$dbHandle)
    {
        $Logger->fatal_msg(__FILE__, __LINE__, $ias_db_error);
        $Logger->close_log();
        exit FAILURE;
    }

    # return the important items
    return $dbHandle, $Logger;
}

###############################################################################
# NAME: retrieveRecords
#
# PURPOSE: retrieve and return selected MTF records from the database 
#
# INPUTS: Pre-formatted SQL SELECT string to send to the database
#
# RETURNS: a hash reference to arrays of values per field, through a parameter
###############################################################################
sub retrieveRecords
{
    my ($dbh, $sql, $errStrRef, $resultsHofARef) = @_;

    # prepare SQL query
    my $sth = $dbh->prepare($sql);
    if (!$sth)
    {
        $$errStrRef = "Build of query returned an error: $DBI::errstr";
        return ERROR;
    }

    # execute SQL query
    if (!$sth->execute())
    {
        $$errStrRef = "Query returned an error: $DBI::errstr";
        return ERROR;
    }

    # fetch and save the results of the query
    while (my $dbRowRef = $sth->fetchrow_hashref())
    {
        foreach (keys %$dbRowRef)
        {
            push @{$resultsHofARef->{$_}}, $dbRowRef->{$_};
        }
    }

    $sth->finish();

    return SUCCESS;
}

###############################################################################
# NAME: usage
#
# PURPOSE: Prints the usage message  
#
# INPUTS: None
#
# RETURNS: None
###############################################################################
sub usage
{
   print "Usage retrieve_mtf_db.pl output_file\n" .
         "       [--opt_min=<minimum Gaussian optical value>]\n" .
         "       [--opt_max=<maximum Gaussian optical value>]\n" .
         "       [--start_date=<MM-DD-YYYY>]\n" .
         "       [--end_date=<MM-DD-YYYY>]\n";
}
