/******************************************************************************

NAME:           mtfperform_test

PURPOSE:
Main test procedure for software to automatically locate and profile MTF target
windows predefined targets.

Expected results values are from CalVal prototype results.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
EXIT_FAILURE    Error profiling MTF windows
EXIT_SUCCESS    Successfully profiled MTF windows

******************************************************************************/

#include <stdlib.h>         /* EXIT_FAILURE/EXIT_SUCCESS */
#include <limits.h>         /* PATH_MAX */
#include <string.h>         /* strncpy */
#include <libgen.h>         /* basename */

#include "cgreen/cgreen.h"
#include "mtfperform.h"
#include "ias_math_complex.h"
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"


static char ut_data_dir[PATH_MAX];

Ensure test_fit_stf(void)
{
    const int NUMREC = 3;

    /* parameters to fit_stf */
    int numrec = -1;        /* Number of trending records */
    TRENDREC trend[NUMREC]; /* Array of trending records */
    int band_number = -1;   /* Current band number */
    STF_FIT_PARMS stf_fit;  /* Best fit STF parameters */
    double xt_best;         /* 90% value of XT optics parameter */
    double at_best;         /* 90% value of AT optics parameter */


    /* return status from fit_stf */
    int status;

    /* record counter */
    int rec;

    /* original logging level, suppressing expected errors/warnings */
    IAS_LOG_MESSAGE_LEVEL old_level;

    memset(trend, 0, sizeof(trend));

    /* suppress error messages */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* test with numrec=-1, band_number=-1, and NULL pointers */
    status = fit_stf(numrec, NULL, band_number, NULL, &xt_best, &at_best);
    assert_equal(status, ERROR);

    /* test with numrec=-1, band_number=-1, and non-NULL pointers */
    status = fit_stf(numrec, trend, band_number, &stf_fit, &xt_best, &at_best);
    assert_equal(status, ERROR);

    /* test with numrec=1, band_number=-1, and non-NULL pointers */
    numrec = 1;
    status = fit_stf(numrec, trend, band_number, &stf_fit, &xt_best, &at_best);
    assert_equal(status, ERROR);

    /* test with numrec=1, band_number=1, and non-NULL pointers. */
    band_number = 1;
    status = fit_stf(numrec, trend, band_number, &stf_fit, &xt_best, &at_best);
    assert_equal(status, ERROR);

    /* test with numrec=1, band_number=1, and non-zero values.  One point is 
       found and that is not enough for a solution, so it should fail */
    for (rec = 0; rec < NUMREC; rec++)
    {
        trend[rec].band_number = band_number;
        trend[rec].angle = (rec - NUMREC / 2) * 3.1415 / 100;
        trend[rec].optics = (rec + 1) * 0.1;
        trend[rec].atdetector = 10 + rec * 0.1;
        trend[rec].xtdetector = 20 + rec * 0.2;
        trend[rec].inttime = 1 + rec * 0.01;
        trend[rec].diffuse = 5 + rec * 0.05;
    }
    status = fit_stf(numrec, trend, band_number, &stf_fit, &xt_best, &at_best);
    assert_equal(status, ERROR);

    /* restore original logging level, not expecting ERRORS/WARNINGS */
    ias_log_set_output_level(old_level);

    /* test with numrec=NUMREC, band_number=1, and non-zero values */
    numrec = NUMREC;
    status = fit_stf(numrec, trend, band_number, &stf_fit, &xt_best, &at_best);
    assert_equal(status, SUCCESS);
    assert_double_equal(stf_fit.xt_optics, 2.0e-1);
    assert_double_equal(stf_fit.at_optics, 3.18999216709594);
    assert_double_equal(stf_fit.xt_detector, 2.02000000e+01);
    assert_double_equal(stf_fit.at_detector, 1.01000000e+01);
    assert_double_equal(stf_fit.integration_milliseconds, 1.01000000e+00);
    assert_double_equal(stf_fit.diffusion, 5.05000000e+00);
    assert_double_equal(xt_best, 0.0);
    assert_double_equal(at_best, 0.0);
}

Ensure test_calc_spatial_stats(void)
{
    const int NSAMP = 4;

    /* input parameters for calc_spatial_stats */
    double gsd = 30.0;      /* I: Ground sample distance in meters */
    int nsamp = -1;         /* I: Length of ESF sequence */
    int oversamp = -1;      /* I: ESF oversampling factor */
    double esf_buf[NSAMP];  /* I: ESF data */
    double edge_slope = -1.0; /* O: Edge slope */
    double edge_extent = -1.0; /* O: Edge extent */

    /* return status from calc_spatial_stats */
    int status;

    /* loop counter */
    int i;

    /* original logging level, suppressing expected errors/warnings */
    IAS_LOG_MESSAGE_LEVEL old_level;

    /* suppress error messages */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    memset(esf_buf, 0, sizeof(esf_buf));

    /* test with NULL pointers, ERROR */
    status = calc_spatial_stats(gsd, nsamp, oversamp, NULL, NULL, NULL);
    assert_equal(status,ERROR);

    /* test with non-NULL pointers and nsamp=-1, not enough points and
       outputs unchanged */
    status = calc_spatial_stats(gsd, nsamp, oversamp, esf_buf, &edge_slope, 
        &edge_extent);
    assert_equal(status,ERROR);

    /* test with nsamp=1 (output unchanged, not enough points) */
    nsamp = 1;
    status = calc_spatial_stats(gsd, nsamp, oversamp, esf_buf, &edge_slope, 
        &edge_extent);
    assert_equal(status,ERROR);

    /* restore original logging level, not expecting ERRORS/WARNINGS */
    ias_log_set_output_level(old_level);

    /* test with input values, slope/extent expected */
    nsamp = NSAMP;
    oversamp = 128;
    for (i = 0; i < NSAMP; i++)
    {
        esf_buf[i] = 0.3 + (1.0 * i) / (i + 1);
    }

    status = calc_spatial_stats(gsd, nsamp, oversamp, esf_buf, &edge_slope, 
        &edge_extent);
    assert_equal(status,SUCCESS);
    assert_double_equal(edge_slope, 2.133333333);
    assert_double_equal(edge_extent, 0.3515625);
}

Ensure test_calc_spatial_perf(void)
{
    /* parameter for calc_spatial_perf */
    char *output_dir = NULL; /* Output directory */
    STF_FIT_PARMS stf_fit;   /* STF fit parameters */
    double gsd = -1.0;       /* Band ground sample distance */
    double range = -1.0;     /* Range to scale STF to ground */
    double gvel = -1.0;      /* Velocity to scale STF to ground */
    int nsamp = -1;          /* Length of PSF/ESF to analyze */
    int oversamp = -1;       /* PSF/ESF oversample factor */
    double xt_best = 0;      /* Low 90% value for xoptics */
    double at_best = 0;      /* Low 90% value for aoptics */
    SPATIAL results;         /* Parameter results structure */

    /* return from calc_spatial_perf */
    int status;

    /* original logging level, suppressing expected errors/warnings */
    IAS_LOG_MESSAGE_LEVEL old_level;

    /* suppress error messages */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    memset(&stf_fit, 0, sizeof(stf_fit));
    memset(&results, 0, sizeof(results));

    /* test with nsamp=-1, NULL pointers */
    status = calc_spatial_perf(output_dir, &stf_fit, gsd, range, gvel,
                               nsamp, oversamp, xt_best, at_best, NULL);
    assert_equal(status, ERROR);

    /* test with nsamp=-1 and non-NULL pointers */
    output_dir = "";
    status = calc_spatial_perf(output_dir, &stf_fit, gsd, range, gvel,
                               nsamp, oversamp, xt_best, at_best, &results);
    assert_equal(status, ERROR);

    /* test with nsamp=1 and results.band_number=0, ERROR */
    output_dir = ".";
    nsamp = 1;
    status = calc_spatial_perf(output_dir, &stf_fit, gsd, range, gvel,
                               nsamp, oversamp, xt_best, at_best, &results);
    assert_equal(status, ERROR);

    /* restore original logging level, not expecting ERRORS/WARNINGS */
    ias_log_set_output_level(old_level);

    /* test with nsamp=16 and results.band_number=1, SUCCESS */
    nsamp = 16;
    results.band_number = 1;
    status = calc_spatial_perf(output_dir, &stf_fit, gsd, range, gvel,
                               nsamp, oversamp, xt_best, at_best, &results);
    assert_equal(status, SUCCESS);
    assert_double_equal(results.xt_slope, 1.0);
    assert_double_equal(results.xt_extent, 0.45);
    assert_double_equal(results.at_slope, 1.0);
    assert_double_equal(results.at_extent, 0.45);
    assert_double_equal(results.xt_slope_best, 1.0);
    assert_double_equal(results.at_slope_best, 1.0);
    assert_double_equal(results.xt_extent_best, 0.45);
    assert_double_equal(results.at_extent_best, 0.45);

    /* test with stf_fit parameters set and results.band_number=2, SUCCESS */
    results.band_number = 2;
    stf_fit.xt_optics = 1.234;
    stf_fit.at_optics = 4.567;
    stf_fit.xt_detector = 2.5;
    stf_fit.at_detector = 3.5;
    stf_fit.integration_milliseconds = 0.3;
    stf_fit.diffusion = 5.7;
    status = calc_spatial_perf(output_dir, &stf_fit, gsd, range, gvel,
                               nsamp, oversamp, xt_best, at_best, &results);
    assert_equal(status, SUCCESS);
    assert_double_equal(results.xt_slope, 0.999991);
    assert_double_equal(results.xt_extent, 0.450004);
    assert_double_equal(results.at_slope, 0.999991);
    assert_double_equal(results.at_extent, 0.450004);
    assert_double_equal(results.xt_slope_best, 0.999991);
    assert_double_equal(results.at_slope_best, 0.999991);
    assert_double_equal(results.xt_extent_best, 0.450004);
    assert_double_equal(results.at_extent_best, 0.450004);

    /* test with remaining parameters set and results.band_number=4, SUCCESS */
    results.band_number = 4;
    gsd = 30.0;
    range = 100.0;
    gvel = 400.0;
    oversamp = 128;
    status = calc_spatial_perf(output_dir, &stf_fit, gsd, range, gvel,
                               nsamp, oversamp, xt_best, at_best, &results);
    assert_equal(status, SUCCESS);
    assert_double_equal(results.xt_slope, 4.25040638);
    assert_double_equal(results.xt_extent, 0.10587223);
    assert_double_equal(results.at_slope, 4.09975434);
    assert_double_equal(results.at_extent, 0.10976267);
    assert_double_equal(results.xt_slope_best, 4.250408315);
    assert_double_equal(results.at_slope_best, 4.099779415);
    assert_double_equal(results.xt_extent_best, 0.10587218);
    assert_double_equal(results.at_extent_best, 0.109762);
}

Ensure test_write_esf(void)
{
    const int NPTS = 5;

    /* input parameters to write_esf */
    char *output_dir = NULL;    /* output directory */
    char *prefix = NULL;        /* "XT" or "AT" */
    int band_number = -1;       /* Data band number */
    int npts = -1;              /* Number of points in ESF */
    double esf_buf[NPTS];       /* Model ESF data */

    /* return status from write_esf */
    int status;

    int i;

    /* original logging level, suppressing expected errors/warnings */
    IAS_LOG_MESSAGE_LEVEL old_level;

    /* suppress error messages */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    memset(esf_buf, 0, sizeof(esf_buf));

    /* test with NULL pointers, should log error and return ERROR */
    status = write_esf(output_dir, prefix, band_number, npts, NULL);
    assert_equal(status, ERROR);

    /* test with non-NULL pointers, but invalid dir, should log error
       and return ERROR */
    output_dir = "bad_dir";
    prefix = "";
    status = write_esf(output_dir, prefix, band_number, npts, esf_buf);
    assert_equal(status, ERROR);

    /* restore original logging level, not expecting ERRORS/WARNINGS */
    ias_log_set_output_level(old_level);

    /* test with current dir and nptrs=-1 should create an empty file and
       return SUCCESS */
    output_dir = ".";
    prefix = "t0";
    status = write_esf(output_dir, prefix, band_number, npts, esf_buf);
    assert_equal(status, SUCCESS);

    /* test with npts=NPTS, should create a file of 0's and return SUCCESS */
    npts = NPTS;
    output_dir = ".";
    prefix = "t1";
    status = write_esf(output_dir, prefix, band_number, npts, esf_buf);
    assert_equal(status, SUCCESS);

    /* test with non-zero esf values, should create a file and return SUCCESS */
    npts = NPTS;
    for (i = 0; i < npts; i++)
    {
        esf_buf[i] = (1.0 * i) / (i + 1);
    }
    prefix = "t2";
    status = write_esf(output_dir, prefix, band_number, npts, esf_buf);
    assert_equal(status, SUCCESS);

}

Ensure test_get_trend_band(void)
{
    /* parameters to get_trend_band */
    char trend_file[PATH_MAX];  /* Trending file name */
    int band_number = -1;       /* Band number to retrieve */
    TRENDREC *trend = NULL;     /* Array of trending records */

    /* return number of records from get_trend_band */
    int nrec = -1;

    /* return number from snprintf */
    int n = -1;

    /* original logging level, suppressing expected errors/warnings */
    IAS_LOG_MESSAGE_LEVEL old_level;

    /* suppress error messages */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* test with NULL pointers */
    nrec = get_trend_band(NULL, band_number, NULL);
    assert_equal(nrec, 0);

    /* test with non-NULL pointers, but non-existent file */
    strcpy(trend_file, "");
    nrec = get_trend_band(trend_file, band_number, &trend);
    assert_equal(nrec, 0);
    assert_equal(trend, NULL);

    /* restore original logging level, not expecting ERRORS/WARNINGS */
    ias_log_set_output_level(old_level);

    /* test with valid file, but band_number = -1 */
    n = snprintf(trend_file, sizeof(trend_file),
                 "%s/input/mtfestimate_trendall.dat", ut_data_dir);
    if ((n < 0) || (n >= sizeof(trend_file)))
    {
        assert_true_with_message(0, "unable to form trend_file name, n=%d", n);
        return;
    }
    nrec = get_trend_band(trend_file, band_number, &trend);
    assert_equal(nrec, 0);
    assert_equal(trend, NULL);

    /* test with valid file, band_number = 1 */
    band_number = 1;
    nrec = get_trend_band(trend_file, band_number, &trend);
    assert_equal(nrec, 6);
    assert_not_equal(trend, NULL);
    if (trend != NULL)
    {
        assert_equal(trend->year, 2000);
        assert_equal(trend->doy, 261);
        assert_equal(trend->wrs_path, 22);
        assert_equal(trend->wrs_row, 39);
        assert_string_equal(trend->target_name, "Causeway");
        assert_equal(trend->band_number, band_number);
        assert_double_equal(trend->angle, 0.071775);
        assert_double_equal(trend->optics, 23.212700);
        assert_double_equal(trend->atdetector, 42.303000);
        assert_double_equal(trend->xtdetector, 40.863000);
        assert_double_equal(trend->inttime, 3.6);
        assert_double_equal(trend->diffuse, 0.0);
        assert_double_equal(trend->phase, 11.817000);
        assert_double_equal(trend->rms, 5.368000);
        assert_double_equal(trend->gsd, 30.0);
        assert_double_equal(trend->range, 705742.562000);
        assert_double_equal(trend->velocity, 6839.067000);
        assert_double_equal(trend->water, 974.500000);
        assert_double_equal(trend->asym, 1.3);
        assert_equal(trend->nspan, 2);
        assert_double_equal(trend->span[0].reflect, 2363.3);
        assert_double_equal(trend->span[0].width, 10.0);
        assert_double_equal(trend->span[0].offset, -17.2);
        assert_double_equal(trend->span[1].reflect, 2108.6);
        assert_double_equal(trend->span[1].width, 10.0);
        assert_double_equal(trend->span[1].offset, 17.2);
    }

    free(trend);
}

Ensure test_report(void)
{
    /* NOTE: only checking error conditions, running mtfperform tests the 
       report file contents */

    const int NBANDS = 3;

    /* parameter to create_spatial_report */
    char *report_file = NULL; /* Name of report file to create */

    /* parameters for report_stf */
    int band_number = -1;     /* Current band number */
    int numrec = -1;          /* Number of records used in fit */
    TRENDREC trend;           /* Array of trending results used */
    STF_FIT_PARMS stf_fit;    /* Best fit STF parameters */

    /* parameters for report_performance */
    int nband = -1;           /* Number of bands analyzed */
    SPATIAL stats[NBANDS];    /* Array of results structures */

    /* return status */
    int status;

    /* original logging level, suppressing expected errors/warnings */
    IAS_LOG_MESSAGE_LEVEL old_level;

    memset(&trend, 0, sizeof(trend));
    memset(&stf_fit, 0, sizeof(stf_fit));
    memset(&stats, 0, sizeof(stats));

    /* suppress error messages */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* test with NULL pointer for filename */
    status = create_spatial_report(report_file);
    assert_equal(status, ERROR);

    /* test with invalid non-NULL string */
    report_file = "";
    status = create_spatial_report(report_file);
    assert_equal(status, ERROR);

    /* restore original logging level, not expecting ERRORS/WARNINGS */
    ias_log_set_output_level(old_level);

    /* test with valid string, expecting SUCCESS */
    report_file = "report1.txt";
    status = create_spatial_report(report_file);
    assert_equal(status, SUCCESS);

    /* suppress error messages */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* test with another valid string, expecting ERROR already open */
    report_file = "report2.txt";
    status = create_spatial_report(report_file);
    assert_equal(status, ERROR);

    /* restore original logging level, not expecting ERRORS/WARNINGS */
    ias_log_set_output_level(old_level);

    /* test writing STF to open file, SUCCESS returned */
    status = report_stf(band_number, numrec, &trend, stf_fit);
    assert_equal(status, SUCCESS);

    /* test writing performance to closed file */
    status = report_performance(nband, stats);
    assert_equal(status, SUCCESS);

    /* test closing, no return */
    close_spatial_report();

    /* suppress error messages */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* test closing again, error logged */
    close_spatial_report();

    /* test writing STF to closed file, error logged, ERROR returned */
    status = report_stf(band_number, numrec, &trend, stf_fit);
    assert_equal(status, ERROR);

    /* test writing performance to closed file */
    status = report_performance(nband, stats);
    assert_equal(status, ERROR);

    /* restore original logging level, not expecting ERRORS/WARNINGS */
    ias_log_set_output_level(old_level);

}

Ensure test_get_perform_parms(void)
{
    /* parameters for get_perform_parms */
    char param_filename[PATH_MAX];  /* Parameter file name */
    char trend_filename[PATH_MAX];  /* Name of trending data file */
    char report_filename[PATH_MAX]; /* Name of output report file */
    int numbands = -1;              /* Number of bands to process */
    int band_list[IAS_MAX_NBANDS];  /* Bands to process */
    int nsamp = -1;                 /* Number of samples in PSF/ESF */
    int oversamp = -1;              /* PSF/ESF oversampling factor */

    /* return status from get_perform_parms */
    int status;

    /* return value from snprintf */
    int n;

    /* expected return strings */
    char expected_trend_filename[PATH_MAX];  /* Name of trending data file */
    char expected_report_filename[PATH_MAX]; /* Name of output report file */

    /* original logging level, suppressing expected errors/warnings */
    IAS_LOG_MESSAGE_LEVEL old_level;

    param_filename[0] = '\0';
    trend_filename[0] = '\0';
    report_filename[0] = '\0';
    memset(band_list, 0, sizeof(band_list));

    /* suppress error messages */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* test with NULL pointers */
    status = get_perform_parms(NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    assert_equal(status, ERROR);

    /* test with non-NULL pointers, but empty strings */
    status = get_perform_parms(param_filename, trend_filename,
                               report_filename, &numbands, band_list,
                               &nsamp, &oversamp);
    assert_equal(status, ERROR);

    /* test with .odl file that has 3 bands */
    n = snprintf(param_filename, sizeof(param_filename),
                 "%s/input/MTFPERFORM_0.odl", ut_data_dir);
    if ((n < 0) || (n >= sizeof(param_filename)))
    {
        assert_true_with_message(0, "Cannot form param_filename, n=%d", n);
        return;
    }
    status = get_perform_parms(param_filename, trend_filename,
                               report_filename, &numbands, band_list,
                               &nsamp, &oversamp);
    assert_string_equal(trend_filename, "");
    assert_string_equal(report_filename, "");
    assert_equal(numbands, 3);
    assert_equal(nsamp, 512);
    assert_equal(oversamp, 30);

    /* restore original logging level, not expecting ERRORS/WARNINGS */
    ias_log_set_output_level(old_level);

    /* test with .odl file that has valid entries */
    n = snprintf(param_filename, sizeof(param_filename),
                 "%s/input/MTFPERFORM_1.odl", ut_data_dir);
    if ((n < 0) || (n >= sizeof(param_filename)))
    {
        assert_true_with_message(0, "Cannot form param_filename, n=%d", n);
        return;
    }
    status = get_perform_parms(param_filename, trend_filename,
                               report_filename, &numbands, band_list,
                               &nsamp, &oversamp);
    assert_equal(status, SUCCESS);
    n = snprintf(expected_trend_filename, sizeof(expected_trend_filename),
                 "mtfestimate_trend.dat");
    if ((n < 0) || (n >= sizeof(expected_trend_filename)))
    {
        assert_true_with_message(0, "Cannot form expected_trend_filename, "
            "n=%d", n);
        return;
    }
    assert_string_equal(trend_filename, expected_trend_filename);
    n = snprintf(expected_report_filename, sizeof(expected_report_filename),
                 "mtf_performance.rpt");
    if ((n < 0) || (n >= sizeof(expected_report_filename)))
    {
        assert_true_with_message(0, "Cannot form expected_report_filename, "
            "n=%d", n);
        return;
    }
    assert_string_equal(report_filename, expected_report_filename);
    assert_equal(numbands, 3);
    assert_equal(nsamp, 512);
    assert_equal(oversamp, 30);

    /* test with .odl file that has valid defaults */
    n = snprintf(param_filename, sizeof(param_filename),
                 "%s/input/MTFPERFORM_2.odl", ut_data_dir);
    if ((n < 0) || (n >= sizeof(param_filename)))
    {
        assert_true_with_message(0, "Cannot form param_filename, n=%d", n);
        return;
    }
    status = get_perform_parms(param_filename, trend_filename,
                               report_filename, &numbands, band_list,
                               &nsamp, &oversamp);
    assert_equal(status, SUCCESS);
    assert_string_equal(trend_filename, expected_trend_filename);
    assert_string_equal(report_filename, expected_report_filename);
    assert_equal(numbands, 9);
    assert_equal(nsamp, 512);
    assert_equal(oversamp, 30);
}


int main(int argc, char *argv[])
{
    char *ias_unit_test_directory = NULL;  /* Top level of IAS unit test
                                              data directory */

    /* subdirectory path to our unit test data */
    const char *ut_subdir = "ias/gps/mtf";

    int status;
    int count;

    char progname[PATH_MAX] = "mtfperform_unit_tests";

    TestSuite *suite = create_test_suite();
    add_test(suite, test_fit_stf);
    add_test(suite, test_calc_spatial_perf);
    add_test(suite, test_calc_spatial_stats);
    add_test(suite, test_write_esf);
    add_test(suite, test_get_trend_band);
    add_test(suite, test_report);
    add_test(suite, test_get_perform_parms);

    status = ias_log_initialize(progname);
    if (status != SUCCESS)
    {
        printf("Failure to initialize logging library\n");
        return EXIT_FAILURE;
    }

    /* Get unit test directory */
    ias_unit_test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (ias_unit_test_directory == NULL)
    {
        printf("The IAS_UNIT_TEST_DATA_DIR environment variable is not set\n");
        return EXIT_FAILURE;
    }

    count = snprintf(ut_data_dir, sizeof(ut_data_dir), "%s/%s/mtfperform",
                     ias_unit_test_directory, ut_subdir);
    if ((count < 0) || (count >= sizeof(ut_data_dir)))
    {
        assert_true_with_message(0, "Cannot set UT data dir, count = %d", 
            count);
        return EXIT_FAILURE;
    }

    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}


