/******************************************************************************
NAME:           trendrec.c

PURPOSE:
Source file for functions for MTF trending record manipulation.

RETURN VALUE:

******************************************************************************/
#include <string.h>
#include <stdlib.h>       /* malloc */

#include "ias_const.h"
#include "ias_logging.h"  /* Defines IAS_LOG macros */
#include "trendrec.h"     /* Definition of TRENDREC structure */

/******************************************************************************
 Load the target parameters from the input file 

RETURN VALUE:

    SUCCESS - successfully read record
    ERROR - error while parsing record 
    END_OF_FILE - encountered end of file

******************************************************************************/
static int get_trendrec   /* Return SUCCESS or ERROR */
(
    FILE *trend_ptr,      /* I: Trending file pointer */
    TRENDREC *trendrec    /* O: Populated trendrec structure */
)
{
    char buffer[RECLEN];  /* Input buffer for trending records */
    char *buf;            /* Buffer pointer */
    int nread;            /* Number of fields read */
    int ispan;            /* Span index */
    int offset;           /* Offset into buffer after sscanfs */

    /* Read the next line in the input file */
    buf = fgets(buffer, RECLEN, trend_ptr);
    if (buf == NULL)
    {
        /* End of file reached */
        return TREND_EOF;
    }

    /* Parse the fields based on the delimiter */
    nread = sscanf(buffer, "%s | %d | %d | %d | %d | %s | %d | %lf | %lf | %lf "
        "| %lf | %lf | %lf | %lf | %lf | %lf | %lf | %lf | %lf | %lf | %d%n", 
        trendrec->work_order_id, 
        &trendrec->year,
        &trendrec->doy,
        &trendrec->wrs_path,
        &trendrec->wrs_row,
        trendrec->target_name,
        &trendrec->band_number,
        &trendrec->angle,
        &trendrec->optics,
        &trendrec->atdetector,
        &trendrec->xtdetector,
        &trendrec->inttime,
        &trendrec->diffuse,
        &trendrec->phase,
        &trendrec->rms,
        &trendrec->gsd,
        &trendrec->range,
        &trendrec->velocity,
        &trendrec->water,
        &trendrec->asym,
        &trendrec->nspan,
        &offset);

    /* Skip past the already-read parts of the string for later sscanfs */
    buf += offset;

    if (nread != 21)
    {
        IAS_LOG_ERROR("Could not scan line of data, expected 21 fields but "
            "scanned %d", nread);
        return ERROR;
    }

    /* Loop through the spans */
    for (ispan = 0; ispan < trendrec->nspan; ispan++)
    {
        nread = sscanf(buf, " | %lf | %lf | %lf%n",
            &trendrec->span[ispan].reflect,
            &trendrec->span[ispan].width,
            &trendrec->span[ispan].offset,
            &offset);

        if (nread != 3)
        {
            IAS_LOG_ERROR("Could not scan data for span %d, expected 3 fields "
                "but scanned %d", ispan, nread);
            return ERROR;
        }

        /* Skip past the already-read parts of the string for later sscanfs */
        buf += offset;

    }
    return SUCCESS;
}

int get_trend_band          /* Return number of trending records */
(
    char *trend_file,       /* I: Trending file name */
    int band_number,        /* I: Band number to retrieve */
    TRENDREC **trend        /* O: Array of trending records */
)
{
    FILE *t_ptr = NULL;     /* Trending file pointer */
    int nrec;               /* Number of trending records found */
    int lrec;               /* Number of trending records loaded */
    int line_number;        /* Line number in file */
    TRENDREC *local = NULL; /* Local pointer to trending array */
    TRENDREC current;       /* Current trending record */
    int status;             /* Function return status */

    /* check for invalid input */
    if ((trend_file == NULL) || (trend == NULL))
    {
        IAS_LOG_ERROR("trend_file (%p) and trend (%p) cannot be NULL",
                      trend_file, trend);
        return 0;
    }

    /* Initialize number of records */
    nrec = 0;
    *trend = NULL;

    /* Open trending file */
    t_ptr = fopen(trend_file, "r");
    if (t_ptr == NULL)
    {
        IAS_LOG_ERROR("Error opening trending file %s", trend_file);
        return nrec;
    }

    /* Scan the file to see how many records there are */
    line_number = 1;
    while (1)
    {
        status = get_trendrec(t_ptr, &current);

        if (status == ERROR)
        { 
            IAS_LOG_ERROR("parsing trending file %s at line %d", trend_file,
                line_number);
            fclose(t_ptr);
            return 0;
        }

        /* Break if status isn't ERROR or SUCCESS (e.g.: end of file) */
        if (status != SUCCESS)
        {
            break;
        }

        if (current.band_number == band_number)
        {
            nrec++;
        }

        line_number++;
    }

    /* If no useful records were found, return */
    if (nrec <= 0)
    {
        fclose(t_ptr);
        return nrec;
    }

    /* Allocate space for the valid records */
    local = (TRENDREC *)malloc((nrec + 1) * sizeof(TRENDREC));
    if (local == NULL)
    {
        IAS_LOG_ERROR("Error allocating space for %d trending records "
                      "for band number %d", nrec, band_number);
        fclose(t_ptr);
        return 0;
    }

    /* Reposition the file to the beginning */
    (void)fseek(t_ptr, 0L, SEEK_SET);

    /* Scan the file to read the records for the desired band */
    lrec = 0;
    while (get_trendrec(t_ptr, &local[lrec]) == SUCCESS)
    {
        if (local[lrec].band_number == band_number)
        {
            lrec++;
        }
        if (lrec > nrec)
        {
            break;
        }
    }
    if (lrec != nrec)
    {
        IAS_LOG_WARNING("Number of records loaded %d does not match number "
                        "expected %d", lrec, nrec);
    }

    /* Pass back loaded array */
    *trend = local;

    /* Close file and exit */
    fclose(t_ptr);
    return nrec;
}

