/******************************************************************************
NAME:           write_esf

PURPOSE:
Write the modeled edge spread function to a comma delimited text file.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
ERROR           Error writing ESF
SUCCESS         Successfully wrote ESF

******************************************************************************/
#include <stdio.h>
#include <limits.h>

#include "ias_logging.h"
#include "report.h"
#include "mtfperform.h"

int write_esf
(
    const char *output_dir, /* I: output directory */
    const char *prefix,     /* I: "XT" or "AT" */
    int band,               /* I: Data band number */
    int npts,               /* I: Number of points in ESF */
    const double *esf_buf   /* I: Model ESF data */
)
{
    FILE *ofile = NULL;     /* Output file pointer */
    char fname[PATH_MAX];   /* Output file name */
    int isamp;              /* Output data sample index */
    int count;
    int status;             /* Function return status */

    if ((output_dir == NULL) || (prefix == NULL) || (esf_buf == NULL))
    {
        IAS_LOG_ERROR("output_dir, prefix, and esf_buf cannot be NULL");
        return ERROR;
    }

    /* Construct output file name */
    count = snprintf(fname, sizeof(fname), "%s/%s_ESF_Band%d.csv",
                     output_dir, prefix, band);
    if ((count < 0) || (count >= sizeof(fname)))
    {
        IAS_LOG_ERROR("Cannot create output filename, count=%d", count);
        return ERROR;
    }

    /* Open the output file */
    ofile = fopen(fname, "w");
    if (ofile == NULL)
    {
        IAS_LOG_ERROR("Error opening ESF output file");
        return ERROR;
    }

    /* Write out the ESF data */
    if (npts > 0)
    {
        if (fprintf(ofile, "%9.6lf", esf_buf[0]) < 0)
        {
            IAS_LOG_ERROR("Writing to output file %s", fname);
            fclose(ofile);
            return ERROR;
        }
    }

    for (isamp = 1; isamp < npts; isamp++)
    {
        if (fprintf(ofile, ", %9.6lf", esf_buf[isamp]) < 0)
        {
            IAS_LOG_ERROR("Writing to output file %s", fname);
            fclose(ofile);
            return ERROR;
        }
    }

    if (fprintf(ofile, "\n") < 0)
    {
        IAS_LOG_ERROR("Writing to output file %s", fname);
        fclose(ofile);
        return ERROR;
    }

    /* Close output file */
    status = fclose(ofile);
    if (status != 0)
    {
        IAS_LOG_ERROR("Closing output file %s", fname);
        return ERROR;
    }

    return SUCCESS;
}
