/******************************************************************************
NAME:           calc_bin

PURPOSE:
Calculate which target bin a particular pixel record belongs in.

RETURN VALUE:   None

NOTES:
This routine calculates the distance from the input pixel to the input target
centerline, computes the associated target bin from the target GSD and
oversample rate, and computes the target orientation angle from the target
azimuth and image azimuth.

******************************************************************************/
#include <math.h>
#include <string.h>
#include "mtfprofile.h"
#include "ias_math.h"

/* Assign pixel samples to profile bins */
void calc_bin
(
    PIXEL_RECORD *ptable, /* I/O: Pixel record */
    TARGET target,      /* I: Input target parameters */
    int nsamp,          /* I: Number of bins in profile */
    int oversamp        /* I: Profile oversampling factor */
)
{
    double length;      /* Length of target centerline */
    double cl_dx;       /* Centerline X and Y deltas */
    double cl_dy;
    double pix_dx;      /* Pixel X and Y deltas */
    double pix_dy;
    double taraz;       /* Grid azimuth of target centerline */
    double rbin;        /* Floating point bin number */
    double pi;          /* Constant */

    pi = ias_math_get_pi();

    /* Process target centerline */
    cl_dx = target.center_line.start_point.x - target.center_line.end_point.x;
    cl_dy = target.center_line.start_point.y - target.center_line.end_point.y;
    length = sqrt(cl_dx * cl_dx + cl_dy * cl_dy);
    if (length > 0.0)
    {
        cl_dx /= length;
        cl_dy /= length;
    }
    else
    {
        IAS_LOG_WARNING("Invalid zero length target centerline");
        ptable->bin = 0;
        return;
    }

    /* Calculate the target grid azimuth */
    taraz = atan2(cl_dx, cl_dy);
    if (taraz > pi / 2.0)
        taraz -= pi;
    if (taraz < -pi / 2.0)
        taraz += pi;
    ptable->image_azimuth -= taraz;

    /* Calculate the pixel distance to the target centerline */
    pix_dx = ptable->coordinates.x - target.center_line.end_point.x;
    pix_dy = ptable->coordinates.y - target.center_line.end_point.y;
    ptable->distance = pix_dx * cl_dy - pix_dy * cl_dx;

    /* Calculate the pixel bin number */
    rbin = ptable->distance * (double)oversamp / ptable->gsd
           + (double)nsamp / 2.0;
    ptable->bin = (int)floor(rbin + 0.5);
    if (ptable->bin < 0) 
        ptable->bin = 0;
    if (ptable->bin > nsamp) 
        ptable->bin = 0;
}
