/******************************************************************************
NAME:           calc_view_geometry

PURPOSE:
Calculate the average range, ground velocity, and target azimuth over all valid
samples.

RETURN VALUE:   Type void

NOTES:
This routine scans the pixel table for valid pixels (i.e., pixels with non-zero
bin numbers) and calculates the mean values of the rotation angle, range, and
ground velocity fields for those records.

******************************************************************************/
#include "mtfprofile.h"

/* Calculate the mean rotation angle, target range, and ground velocity */
void calc_view_geometry   /* No return value */
(
    int numpixel,         /* I: Number of entries in input table  */
    PIXEL_RECORD ptable[],/* I: Input image sample table */
    double *theta_m,      /* O: Mean azimuth angle */
    double *range_m,      /* O: Mean target range */
    double *gvel_m        /* O: Mean ground velocity */
)
{
    double sum_ang;     /* Sum of valid pixel angles */
    double sum_range;   /* Sum of valid pixel ranges */
    double sum_vel;     /* Sum of valid pixel velocities */
    int npts;           /* Number of points in bin */
    int i;              /* Counter */

    /* Initialize sums */
    npts = 0;
    sum_ang = 0.0;
    sum_range = 0.0;
    sum_vel = 0.0;

    /* Scan the pixel table */
    for (i = 0; i < numpixel; i++)
    {
        if (ptable[i].bin > 0)
        {
            npts++;
            sum_ang += ptable[i].image_azimuth;
            sum_range += ptable[i].target_range;
            sum_vel += ptable[i].ground_velocity;
        }
    }

    /* Calculate averages */
    if (npts < 1)
    {
        IAS_LOG_WARNING("No valid points found");
    }
    else
    {
        sum_ang /= npts;
        sum_range /= npts;
        sum_vel /= npts;
    }

    IAS_LOG_DEBUG("Mean azimuth %9.6lf, Mean range %10.3lf, Mean velocity "
                  "%8.3lf", sum_ang, sum_range, sum_vel);

    /* Pass back results */
    *theta_m = sum_ang;
    *range_m = sum_range;
    *gvel_m = sum_vel;
}
