/******************************************************************************
NAME:           filter_outliers.c

PURPOSE:  This routine will filter the outliers from the array of input pixel
values using the input tolerance value for the confidence level.

RETURN VALUE:   Type = void

NOTES:
This routine finds the pixel values that are the furthest from the mean.
The value that is furthest from the mean is the point that is checked against
the T-distribution threshold to see if it is an outlier.  The process continues
until no more outliers are found, or there are less than 2 non-outlier points
left.

******************************************************************************/
#include "mtfprofile.h"
#include "ias_math.h"

void filter_outliers
(
    double tolerance, /* I: Tolerance value for the t-dist */
    int num_pix,      /* I: Number of pixel values in the array */
    double *data,     /* I: Array of pixel values */
    int *non_outlier_flag  /* I/O: Array of non-outlier flags */
)
{
    int npts;         /* Number of non-outlier points */
    int rejected_pt;  /* Flag to specify if any pts were rejected */
    int max_index;    /* Index of pixel with max distance from mean */
    int throw_index;  /* Index of pixel to be thrown */
    int i;            /* Loop variable */
    double sum_v;     /* Sum of the non-outliers */
    double sum_sqr_v; /* Sum of the square of the non-outliers */
    double mean_v;    /* Mean of the non-outliers */
    double stddev_v;  /* Standard deviation of the non-outliers */
    double delta_v;   /* Cutoff value for the points */
    double max_v;     /* Maximum distance from mean */
    double err_v;     /* Error between the max_v and cutoff value */
    double thresh;    /* Threshold value for the t-dist */

    /* Initialize the variables. */
    rejected_pt = TRUE;
    npts = 0;
    sum_v = 0.0;
    sum_sqr_v = 0.0;

    /* Find the number of non-outliers.  Also calculate the mean of these
     * non-outliers. */
    for (i = 0; i < num_pix; i++)
    {
        non_outlier_flag[i] = 1;
        npts++;
        sum_v += data[i];
        sum_sqr_v += data[i] * data[i];
    }

    /* As long as there are at least 2 points and the last iteration produced
       a rejected point, look for another outlier. */
    while (npts >= 2 && rejected_pt)
    {
        /* Clear the rejected flag. */
        rejected_pt = FALSE;

        /* Compute the standard deviation of the non-outliers. */
        mean_v = sum_v / npts;
        stddev_v = sqrt((sum_sqr_v - npts * mean_v * mean_v) / (npts - 1));

        /* Compute the threshold value using the tolerance value and degrees of
           freedom = npts - 1. */
        thresh = ias_math_compute_t_confidence(tolerance, npts - 1);

        /* Compute the cutoff values from the threshold. */
        delta_v = thresh * stddev_v;

        /* Find the points with the largest difference from the mean. */
        max_v = 0.0;
        max_index = 0;
        for (i = 0; i < num_pix; i++)
        {
            /* Only consider the points that are not outliers. */
            if (non_outlier_flag[i])
            {
                if (fabs(data[i] - mean_v) > max_v)
                {
                    max_v = fabs(data[i] - mean_v);
                    max_index = i;
                }
            }
        }

        /* Reject the point that is furthest from its cutoff value. */
        err_v = max_v - delta_v;
        if (stddev_v != 0.0)
        {
            if (err_v > 0.0)
            {
                /* Throw the point with the largest error. */
                throw_index = max_index;

                /* Set the rejected flag and the throwaway's accept flag. */
                rejected_pt = TRUE;
                non_outlier_flag[throw_index] = 0;

                /* Remove the offset values from the non-outlier statistics. */
                npts--;
                sum_v -= data[throw_index];
                sum_sqr_v -= data[throw_index] * data[throw_index];
            }
        }
    }
}
