/******************************************************************************
NAME:           get_profile_parms

PURPOSE:
Get the mtfprofile processing parameters from the input parameter file.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

******************************************************************************/
#include <libgen.h>

#include "mtfprofile.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "ias_miscellaneous.h"

int get_profile_parms
(
    const char *param_filename, /* I: Parameter file name */
    char *target_file,          /* O: Name of target prarameter file */
    int *numbands,              /* O: Number of bands to process */
    int band_list[],            /* O: Bands to process */
    int *numsca,                /* O: Number of SCAs to process */
    int sca_list[],             /* O: SCAs to process */
    int *nsamp,                 /* O: Number of samples in profile */
    int *oversamp,              /* O: Profile oversampling factor */
    double *outthresh,          /* O: T-dist outlier threshold */
    char *omf_file              /* O: Name of the the OMF */
)
{
    int count;                  /* Number of elements */
    int band_index;             /* Index for band list */
    char work_order_id[IAS_WORKORDER_ID_SIZE]; /* Work order ID */
    int status;                 /* Status returned from functions */

    /* Local variables for storing parameter values */
    char local_target_directory[PATH_MAX];
    char local_l0r_name[PATH_MAX];   /* L0R product HDF name */
    int local_satellite_number = -1;
    int local_band_list[IAS_MAX_NBANDS];
    int local_sca_list[IAS_MAX_NSCAS];
    IAS_SATELLITE_ID satellite_id;

    double local_threshold = 0.0;     /* parameter to capture input */
    int target_wrs_path = 0;          /* Target WRS path number */
    int target_wrs_row = 0;           /* Target WRS row number */
    const int MIN_NUM_SAMPLES = 128;
    const int MIN_OVERSAMPLING = 4;
    int i;

    if (param_filename == NULL)
    {
        IAS_LOG_ERROR("NULL passed for input parameter filename");
        return ERROR;
    }

    /* default to invalid bands */
    *numbands = IAS_MAX_NBANDS;
    for (i = 0; i < *numbands; i++)
    {
        local_band_list[i] = -1;
    }

    /* default to invalid SCAs */
    *numsca = IAS_MAX_NSCAS;
    for (i = 0; i < *numsca; i++)
    {
        local_sca_list[i] = -1;
    }

    /* L0R data directory and root file name */
    strcpy(work_order_id, "");
    strcpy(local_target_directory, "");
    strcpy(local_l0r_name, "");

    /* Start the parameter file table */
    IAS_PARM_DECLARE_TABLE(parms, 8);

    /* Get the work order id */
    IAS_PARM_WORK_ORDER_ID(parms, work_order_id, sizeof(work_order_id), 0);

    /* Bands and SCAs are optional parameters.  If they aren't there, the full
       list/range is used. */

    /* Get the list of bands. */
    IAS_PARM_BAND_LIST(parms, IAS_PARM_OPTIONAL, local_band_list,
                       sizeof(local_band_list), 1);

    /* Get the list of SCAs. */
    int default_sca_list_table[IAS_MAX_NSCAS] =
        {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14};
    IAS_PARM_ADD_INT(parms, SCA_LIST, "List of SCAs to process",
                     IAS_PARM_OPTIONAL, IAS_PARM_ARRAY, 1, 1, 14, 14,
                     default_sca_list_table, local_sca_list,
                     sizeof(local_sca_list), 1);

    /* Get the nsamp_pts. */
    int default_nsamp_pts[] = {MIN_NUM_SAMPLES};
    IAS_PARM_ADD_INT(parms, PROFILE_NSAMP, "Number of Profile Samples", 
                     IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 1, 10000, 1,
                     default_nsamp_pts, nsamp, sizeof(*nsamp), 0);

    /* Get the valid pixel percentage threshold value. */
    const double default_outlier_thresh = 0.95;
    IAS_PARM_ADD_DOUBLE(parms, OUTLIER_THRESHOLD, "T-Distribution Test "
                        "Threshold", IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
                        1, 0, 100, 1, &default_outlier_thresh, &local_threshold,
                        sizeof(local_threshold), 0);

    /* Get the oversamp */
    int default_oversamp[] = {8};
    IAS_PARM_ADD_INT(parms, PROFILE_OVERSAMP, "Oversampling Factor", 
                     IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 
                     MIN_OVERSAMPLING, 128, 1, default_oversamp, oversamp, 
                     sizeof(*oversamp), 0);

    count = IAS_PARM_GET_TABLE_SIZE(parms);

    /* Check if help or template output was requested */
    status = ias_parm_provide_help(param_filename, parms, count,
                                   IAS_INPUT_PARAMETERS);

    if (status == 1)
    {
        /* --help or --template was requested.  Copy the requested option to
           omf_file so the same operation will be performed for the OMF */
        strcpy(omf_file, param_filename);
    }
    else if (status == ERROR) /* Help failed. */
    {
        IAS_LOG_ERROR("Providing help for application parameter file %s",
                      param_filename);
        exit(EXIT_FAILURE);
    }
    else
    {
        if (ias_parm_read(param_filename, parms, count) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading parameters from: %s", param_filename);
            return ERROR;
        }

        if (strcmp(work_order_id, "") == 0)
        {
            IAS_LOG_ERROR("Missing work order id in the parameter file.");
            return ERROR;
        }

        /* Build the OMF file name */
        status = snprintf(omf_file, PATH_MAX, "%s.omf", work_order_id);
        if ((status < 0) && (status >= PATH_MAX))
        {
            IAS_LOG_ERROR("Building OMF filename");
            return ERROR;
        }
    }

    /***** Read additional parameters from OMF *****/

    /* Start the OMF table */
    IAS_PARM_DECLARE_TABLE(omf_parms, 5);

    /* Get the satellite number */
    IAS_PARM_SATELLITE(omf_parms, &local_satellite_number,
                       sizeof(local_satellite_number), 1);

    /* Get the L0r file name */
    IAS_PARM_L0R_FILENAME(omf_parms, IAS_PARM_REQUIRED, local_l0r_name,
                          sizeof(local_l0r_name), 1);

    /* Get the target definition directory name */
    IAS_PARM_MTF_INPUT_DIRECTORY(omf_parms, local_target_directory,
                                 sizeof(local_target_directory));

    /* Get the WRS path */
    IAS_PARM_TARGET_WRS_PATH(omf_parms, IAS_PARM_REQUIRED, &target_wrs_path,
            sizeof(target_wrs_path), 1);

    /* Get the WRS row */
    IAS_PARM_TARGET_WRS_ROW(omf_parms, IAS_PARM_REQUIRED, &target_wrs_row,
            sizeof(target_wrs_row), 1);

    count = IAS_PARM_GET_TABLE_SIZE(omf_parms);

    /* Check if help or template output is being requested */
    status = ias_parm_provide_help(omf_file, omf_parms, count,
                                   IAS_OMF_PARAMETERS);

    if (status == 1)
    {
        /* the --help or --template was provided. can now exit success */
        exit(EXIT_SUCCESS);
    }
    else if (status == ERROR) /* Help failed */
    {
        IAS_LOG_ERROR("Providing help for OMF file: %s", omf_file);
        return ERROR;
    }

    /* Read the OMF file */
    if (ias_parm_read(omf_file, omf_parms, count) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from OMF file: %s", omf_file);
        return ERROR;
    }

    /* Convert the satellite number to the satellite id */
    satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(
                       local_satellite_number);
    if (satellite_id == ERROR)
    {
        IAS_LOG_ERROR("Unable to determine satellite ID from satellite "
                      "number %d", local_satellite_number);
        return ERROR;
    }

    /* Initialize the satellite attributes */
    if (ias_sat_attr_initialize(satellite_id) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attributes library");
        return ERROR;
    }

    status = snprintf(target_file, PATH_MAX, "%s/MTF_Targets_%03d%03d.table",
                      local_target_directory, target_wrs_path, target_wrs_row);
    if ((status < 0) && (status >= PATH_MAX))
    {
        IAS_LOG_ERROR("target file requires %d characters", status);
        return ERROR;
    }

    *numbands = IAS_PARM_GET_COUNT_READ(parms, BAND_LIST); 
    for (count = 0; count < *numbands; count++)
    {
       band_index = ias_sat_attr_convert_band_number_to_index(
            local_band_list[count]);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Invalid band index for band number: %d",
                local_band_list[count]);
            return ERROR;
        }

        band_list[count] = local_band_list[count];
    }

    *numsca = IAS_PARM_GET_COUNT_READ(parms, SCA_LIST);
    for (count = 0; count < *numsca; count++)
    {
        if (local_sca_list[count] < 1 || local_sca_list[count] > IAS_MAX_NSCAS)
        {
            IAS_LOG_ERROR("Invalid SCA number: %d", local_sca_list[count]);
            return ERROR;
        }
        sca_list[count] = local_sca_list[count];
    }

    if (*nsamp < MIN_NUM_SAMPLES)
    {
        IAS_LOG_ERROR("Invalid number of samples in profile, %d, must be at "
                      "least %d", *nsamp, MIN_NUM_SAMPLES);
        return ERROR;
    }
    else
    {
        IAS_LOG_DEBUG("number of samples in profile = %d", *nsamp);
    }

    if (*oversamp < MIN_OVERSAMPLING)
    {
        IAS_LOG_ERROR("Invalid profile oversample factor, %d, must be at "
                      "least %d", *oversamp, MIN_OVERSAMPLING);
        return ERROR;
    }
    else
    {
        IAS_LOG_DEBUG("profile oversample factor = %d", *oversamp);
    }

    *outthresh = local_threshold;

    /* Check the allowable range for the outlier threshold */
    if (*outthresh < 0.0 || *outthresh > 1.0)
    {
        IAS_LOG_ERROR("Invalid outlier threshold value %lf: must be 0.0 "
                      "< thresh < 1.0", *outthresh);
        return ERROR;
    }
    else
    {
        IAS_LOG_DEBUG("outlier threshold value = %lf", *outthresh);
    }

    return SUCCESS;
}
