/******************************************************************************
NAME:           load_pixels

PURPOSE:
Read a MTF pixel table file, populating PIXEL_RECORD structures (if requested)
for each entry in the table.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
npix     Number of pixels selected
0        No pixels found 
-1       error 

NOTES:
This routine will open the input pixel table and read each line, selecting
those from the specified target, band, SCA, and polygon. The selected pixels
are counted and the number returned. If the load_pixels routine is invoked
the selected pixel records are loaded into PIXEL_RECORD structures. If the
count_pixels routine is invoked only the number of pixels found is returned.

******************************************************************************/
#include <string.h>
#include <limits.h>

#include "mtfprofile.h"

/* Scan pixel list to count number of samples in given band/SCA */
int count_pixels      /* Return the number of pixels found */
(
    int target_id,    /* I: Index of selected target */
    int npoly,        /* I: Number of polygons */
    int iband,        /* I: Selected band number */
    int isca          /* I: Selected SCA number */
)
{
    int count;
    count = load_pixels(NULL, 0, target_id, npoly, iband, isca);
    return count;
}

/* Scan pixel list to load samples in given band/SCA */
int load_pixels             /* Return the number of pixels loaded */
(
    PIXEL_RECORD ptable[],  /* O: Loaded pixel structures.  This will be NULL
                                  if called in "count pixels" mode */
    int ptable_size,        /* I: Size of the pixel structure */
    int target_id,          /* I: Index of selected target */
    int npoly,              /* I: Number of polygons */
    int iband,              /* I: Selected band number */
    int isca                /* I: Selected SCA number */
)
{
    FILE *pfile = NULL;     /* Pointer to pixel file */
    char file_name[PATH_MAX];/* Pixel table file name (constructed) */
    int ipoly;          /* Polygon counter */
    int npix;           /* Number of selected pixels */
    int cband;          /* Current band number */
    int csca;           /* Current SCA number */
    double cline;       /* Current line number */
    double csamp;       /* Current sample number */
    double cutmy;       /* Current UTM Y coordinate */
    double cutmx;       /* Current UTM X coordinate */
    double cvelg;       /* Current ground velocity */
    double ciazm;       /* Current image azimuth */
    double crange;      /* Current target range */
    double cpixval;     /* Current pixel value */
    int nfield;         /* Number of fields read */
    int n;
    const IAS_BAND_ATTRIBUTES *band_attributes = NULL; /* Band attributes 
                           structure */

    /* Cycle through the polygons */
    npix = 0;

    for (ipoly = 0; ipoly < npoly; ipoly++)
    {
        /* Construct the table file name */
        n = snprintf(file_name, sizeof(file_name),
                     "Tar%dPoly%dBand%dSca%d.dat", target_id, ipoly, iband,
                     isca);
        if ((n < 0) || (n >= sizeof(file_name)))
        {
            IAS_LOG_ERROR("Cannot build table file name for target %d polygon "
                "%d band number %d SCA %d", target_id, ipoly, iband, isca);
            return -1;
        }

        /* Open the file */
        pfile = fopen(file_name, "r");
        if (pfile == NULL)
        {
            /* Not all targets/polygons/bands/SCAs will have target files
               present since the targets are only in particular locations */
            continue;
        }

        /* Read the data */
        while (1)
        {
            /* Read a line */
            nfield = fscanf(pfile, "%d %d %lf %lf %lf %lf %lf %lf %lf %lf",
                                &cband, &csca, &cline, &csamp, &cutmy, &cutmx,
                                &cvelg, &ciazm, &crange, &cpixval);

            /* If EOF was encountered, exit the file read loop */
            if (feof(pfile))
            {
                break;
            }

            /* Verify the number of fields read */
            if (nfield != 10)
            {
                IAS_LOG_ERROR("Number of fields read %d should be 10 for pixel "
                    "file %s", nfield, file_name);
                return -1;
            }

            /* Verify the band and SCA numbers */
            if ((cband != iband) || (csca != isca)) continue;

            /* Load the structure if provided */
            if (ptable != NULL)
            {
                if ((npix + 1) > ptable_size)
                {
                    IAS_LOG_ERROR("Number of pixels read %d is greater than "
                        "the size of the pixel table %d for pixel file %s", 
                        npix + 1, ptable_size, file_name);
                    return -1;
                }

                ptable[npix].band_number = cband;
                ptable[npix].sca_number = csca;
                ptable[npix].bin = -1;
                ptable[npix].line_coordinate = cline;
                ptable[npix].sample_coordinate = csamp;
                ptable[npix].coordinates.x = cutmx;
                ptable[npix].coordinates.y = cutmy;
                ptable[npix].distance = 0.0;
                band_attributes = ias_sat_attr_get_band_attributes(cband);
                if (!band_attributes)
                {
                    /* Operation failed */
                    IAS_LOG_ERROR("Count not get attributes for band number "
                        "%d while processing pixel file %s", cband, file_name);
                    return -1;
                }

                ptable[npix].gsd = band_attributes->pixel_resolution;
                ptable[npix].ground_velocity = cvelg;
                ptable[npix].image_azimuth = ciazm;
                ptable[npix].target_range = crange;
                ptable[npix].sample_value = cpixval;
            }
            npix++;
        } /* data read loop */

        fclose(pfile);
    } /* polygon loop */

    return npix;
}
