/******************************************************************************
NAME:           mtfprofile

PURPOSE:
Main procedure for software to construct oversampled bridge target profiles
from tables of extracted image samples.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
EXIT_FAILURE    Error constructing target profiles
EXIT_SUCCESS    Successfully constructed target profiles

******************************************************************************/
#include <limits.h>

#include "mtfprofile.h"

int main(int argc, char *argv[])
{
    char target_file[PATH_MAX]; /* ODL target definition file name */
    int iband;               /* Integer band index */
    int band_number;         /* Enumerated band number */
    int isca;                /* Integer SCA index */
    int numbands;            /* Number of bands to process */
    int band_list[IAS_MAX_NBANDS]; /* Input band list */
    int numsca;              /* Number of SCAs to process */
    int sca_list[IAS_MAX_NSCAS]; /* List of SCAs to include */
    int numtarg;             /* Number of targets to process */
    int target_id;           /* Index of current target */
    int nsamp;               /* Number of samples in profile(s) */
    int oversamp;            /* Oversample factor */
    int numpixel;            /* Number of pixels to process */
    int pixels_returned;     /* Number of pixels returned or error status */
    int pixel_table_size;    /* Size of the pixel table */
    int ipixel;              /* Integer pixel counter */
    TARGET target[MAX_TARGETS]; /* Target definition data structure */
    PIXEL_RECORD *ptable = NULL;  /* Pointer to table of pixel records */
    double outliertol;       /* T-distribution outlier test tolerance */
    PROFILEREC *bin_prof = NULL; /* Pointer to profile array */
    double theta_m;          /* Mean profile azimuth vs. target (radians) */
    double range_m;          /* Mean distance to target (meters) */
    double gvel_m;           /* Mean ground velocity over target
                                (meters/sec) */
    int lasterror;           /* Flag for empty bin point */
    int status;
    char omf_name[PATH_MAX];
    const char *parm_file = argv[1];
    const char *progname = strrchr(argv[0],'/');

    if (progname == NULL)
        progname = argv[0];
    else 
        progname++;  /* skip slash */

    /* Initialize the logging library */
    if (ias_log_initialize(progname) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing logging library");
        return EXIT_FAILURE;
    }

    /* Get input parameters */
    if (argc != 2)
    {
        IAS_LOG_ERROR("USAGE:  %s <parameter file name>", progname);
        return EXIT_FAILURE;
    }

    status = get_profile_parms(parm_file, target_file, &numbands, band_list, 
        &numsca, sca_list, &nsamp, &oversamp, &outliertol, omf_name);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Reading parameters from file %s", parm_file);
        return EXIT_FAILURE;
    }

    IAS_LOG_INFO("Initiating %s", progname);

    /* Load the target information from the target file */
    numtarg = get_targets(target_file, target);
    if (numtarg == 0)
    {
        IAS_LOG_ERROR("Unable to read target definition file %s", target_file);
        return EXIT_FAILURE;
    }

    /* Allocate the output profile */
    bin_prof = (PROFILEREC *)malloc(nsamp * sizeof(PROFILEREC));
    if (bin_prof == NULL)
    {
        IAS_LOG_ERROR("Error allocating profile array for %d samples", nsamp);
        return EXIT_FAILURE;
    }

    /* Loop through the targets to process */
    for (target_id = 0; target_id < numtarg; target_id++)
    {
        /* Loop through the bands, one profile is generated for each band */
        for (iband = 0; iband < numbands; iband++)
        {
            band_number = band_list[iband];
            IAS_LOG_INFO("Processing target %d band number %d", target_id,
                         band_number);

            /* Count the total number of pixels to process */
            numpixel = 0;
            for (isca = 0; isca < numsca; isca++)
            {
                pixels_returned = count_pixels(target_id,
                                         target[target_id].number_of_polygons,
                                         band_number, sca_list[isca]);
                if (pixels_returned == -1)
                {
                    IAS_LOG_ERROR("Error counting pixels");
                    free(bin_prof);
                    return EXIT_FAILURE;
                }
                else
                {
                    numpixel += pixels_returned;
                }
            }

            /* If there is no data for this band, issue a message and move on */
            if (numpixel == 0)
            {
                IAS_LOG_WARNING("No pixels found for band number %d",
                                band_list[iband]);
                continue;
            }

            /* Allocate pixel table */
            ptable = (PIXEL_RECORD *)malloc(numpixel * sizeof(PIXEL_RECORD));
            if (ptable == NULL)
            {
                IAS_LOG_ERROR("Error allocating pixel table for %d samples",
                              numpixel);
                free(bin_prof);
                return EXIT_FAILURE;
            }
            pixel_table_size = numpixel;

            /* Load pixel table */
            numpixel = 0;
            for (isca = 0; isca < numsca; isca++)
            {
                pixels_returned = load_pixels(&ptable[numpixel], 
                        pixel_table_size, target_id, 
                        target[target_id].number_of_polygons, band_number, 
                        sca_list[isca]);
                if (pixels_returned == -1)
                {
                    IAS_LOG_ERROR("Error loading pixels");
                    free(bin_prof);
                    return EXIT_FAILURE;
                }
                else
                {
                    numpixel += pixels_returned;
                }
            }

            /* Assign pixels to profile bins */
            for (ipixel = 0; ipixel < numpixel; ipixel++)
            {
                calc_bin(&ptable[ipixel], target[target_id], nsamp,
                         oversamp);
            }

            /* Process the bins required for the profile */
            lasterror = 0;
            for (ipixel = 0; ipixel < nsamp; ipixel++)
            {
                IAS_LOG_DEBUG("Processing profile point %d for target %d "
                              "and band number %d", ipixel + 1, target_id,
                              band_list[iband]);
                status = process_bin(numpixel, ptable, ipixel + 1, outliertol,
                                     &bin_prof[ipixel]);
                if (status != SUCCESS)
                {
                    IAS_LOG_WARNING("Error at profile point %d for target %d "
                                    "and band number %d", ipixel + 1, target_id,
                                    band_list[iband]);
                    if (lasterror)
                    {
                        IAS_LOG_WARNING("Consecutive errors detected "
                                        "- no correction attempted");
                        lasterror++;
                    }
                    else
                    {
                        lasterror = 1;
                        if (ipixel >= nsamp - 1)
                        {
                            bin_prof[ipixel].mean = bin_prof[ipixel - 1].mean;
                            bin_prof[ipixel].std_dev = -1.0;
                            lasterror = 0;
                        }
                    }
                }
                else
                {
                    if (lasterror)
                    {
                        if (ipixel == 1)
                        {
                            bin_prof[ipixel - 1].mean = bin_prof[ipixel].mean;
                            bin_prof[ipixel - 1].std_dev = -1.0;
                        }
                        else if (lasterror == 1)
                        {
                            double m1 = bin_prof[ipixel].mean;
                            double m2 = bin_prof[ipixel - 2].mean;
                            double m = (m1 + m2) / 2.0;
                            bin_prof[ipixel - 1].mean = m;
                            bin_prof[ipixel - 1].std_dev = -1.0;
                        }
                        else
                        {
                            IAS_LOG_WARNING("%d consecutive empty bins left "
                                            "unfilled", lasterror);
                        }
                    }
                    lasterror = 0;
                }
            }

            /* Calculate the mean rotation angle, target range, and ground
               velocity */
            calc_view_geometry(numpixel, ptable, &theta_m, &range_m, &gvel_m);

            /* Write out the profile (mean and standard deviation of each
               bin) */
            status = write_profile(target[target_id].target_name, bin_prof,
                                   nsamp, ptable[0].gsd, oversamp, theta_m,
                                   range_m, gvel_m, target_id,
                                   band_list[iband]);
            if (status == ERROR)
            {
                IAS_LOG_ERROR("Error writing profile for band number %d",
                              band_list[iband]);
                free(ptable);
                free(bin_prof);
                return EXIT_FAILURE;
            }

            /* Release pixel table */
            free(ptable);
            ptable = NULL;
        }  /* Band Loop */

        /* Release the current target structure */
        free(target[target_id].span);
        target[target_id].span = NULL;
    }  /* Target Loop */

    /* Release profile array */
    free(bin_prof);

    /* Update the OMF with the MTF input and pixel (for MTF Extract, output)
       directories */
    if (update_omf(omf_name, target_file) != SUCCESS)
    {
        IAS_LOG_ERROR("Updating OMF file");
        return EXIT_FAILURE;
    }

    /* Print the completion message.  */
    IAS_LOG_INFO("Successful completion");

    return EXIT_SUCCESS;
}

