/******************************************************************************
NAME:           mtfprofile.h

PURPOSE:
Main include file for software to construct bridge target profiles from image
samples, extracted by mtfextract, from L1R image of the MTF characterization
bridge scenes.

******************************************************************************/

#ifndef MTFPROFILE_H
#define MTFPROFILE_H

#include "ias_types.h"
#include "ias_const.h"
#include "ias_structures.h"
#include "get_targets.h"

/* Structures */
typedef struct pixel_record
{
    int band_number;          /* Band number for current sample */
    int sca_number;           /* SCA number for current sample */
    int bin;                  /* Profile bin assignment */
    double line_coordinate;   /* L1R line coordinate */
    double sample_coordinate; /* L1R sample coordinate (detector) */
    IAS_DBL_XY coordinates;   /* UTM X/Y coordinates */
    double distance;          /* Distance from target centerline */
    double gsd;               /* Ground sample distance (meters) */
    double ground_velocity;   /* Ground velocity */
    double image_azimuth;     /* Image azimuth (radians) */
    double target_range;      /* Target range */
    double sample_value;      /* Sample value */
} PIXEL_RECORD;

typedef struct profilerec
{
    double mean;              /* Mean sample value */
    double std_dev;           /* Standard deviation */
} PROFILEREC;

/* Function prototypes */
/* Read the input processing parameters */
int get_profile_parms    /* Return SUCCESS or ERROR */
(
    const char *odlfile, /* I: Input ODL parameter file */
    char *target_file,   /* O: File containing target parameters */
    int *numbands,       /* O: Number of bands to process */
    int band_list[],     /* O: List of bands to process */
    int *numsca,         /* O: Number of SCAs to process */
    int sca_list[],      /* O: List of SCAs to process */
    int *nsamp,          /* O: Length of output profile */
    int *oversamp,       /* O: Output profile oversample factor */
    double *outliertol,  /* O: T-dist outlier tolerance */
    char *omf_file       /* O: OMF filename */
);

int get_target_path_row
(
    const char *l0r_file,   /* I: L0R file name */
    int *target_path,       /* O: WRS path read from metadata */
    int *target_row         /* O: WRS row read from metadata */
);

/* Scan pixel list to count number of samples in given band/SCA */
int count_pixels            /* Return the number of pixels found */
(
    int target_id,          /* I: Index of selected target */
    int number_of_polygons, /* I: Number of polygons */
    int band_number,        /* I: Selected band number */
    int sca_number          /* I: Selected SCA number */
);

/* Scan pixel list to load samples in given band/SCA */
int load_pixels             /* Return the number of pixels loaded  */
(
    PIXEL_RECORD pixel_table[], /* O: Loaded pixel structures.  This will be 
                                  NULL if called in "count pixels" mode */
    int pixel_table_size,   /* I: Size of the pixel structure */
    int target_id,          /* I: Index of selected target */
    int number_of_polygons, /* I: Number of polygons */
    int band_number,        /* I: Selected band number */
    int sca_number          /* I: Selected SCA number */
);

/* Assign pixel samples to profile bins */
void calc_bin              /* No return value */
(
    PIXEL_RECORD *point_table, /* I/O: Pixel record */
    TARGET target,         /* I: Input target parameters */
    int number_of_samples, /* I: Number of bins in profile */
    int oversamp           /* I: Profile oversampling factor */
);

/* Process the requested bin to populate the profile */
int process_bin           /* Return SUCCESS or ERROR */
(
    int numpixel,         /* I: Number of entries in input table */
    PIXEL_RECORD ptable[],/* I: Input image sample table */
    int bin,              /* I: Current bin to process */
    double outliertol,    /* I: T-dist outlier threshold */
    PROFILEREC *profile   /* O: Mean sample value and standard deviation for
                                bin */
);

/* Apply T-distribution outlier filter to image samples in bin */
void filter_outliers
(
    double tolerance,     /* I: Tolerance value for the t-dist */
    int num_pix,          /* I: Number of input pixel values */
    double *data,         /* I: Array of pixel values */
    int *non_outlier_flag /* I/O: Array of non-outlier flags */
);

/* Calculate the mean rotation angle, target range, and ground velocity */
void calc_view_geometry   /* No return value */
(
    int numpixel,         /* I: Number of entries in input table */
    PIXEL_RECORD ptable[],/* I: Input image sample table */
    double *theta_m,      /* O: Mean azimuth angle */
    double *range_m,      /* O: Mean target range */
    double *gvel_m        /* O: Mean ground velocity */
);

/* Write out the profile (mean and standard deviation of each bin) */
int write_profile       /* Return SUCCESS or ERROR */
(
    char *target_name,  /* I: Target name */
    PROFILEREC profile[], /* I: Profile data to write */
    int nsamp,          /* I: Number of bins in profile */
    double gsd,         /* I: Image GSD */
    int oversamp,       /* I: Profile oversampling factor */
    double theta_m,     /* I: Mean azimuth angle */
    double range_m,     /* I: Mean target range */
    double gvel_m,      /* I: Mean ground velocity */
    int target_id,      /* I: Target index (used in file name) */
    int bandnum         /* I: Band number (used in file name) */
);

int update_omf
(
    const char *omf_filename,   /* I: OMF filename */
    char *mtf_target_file       /* I: MTF-specific target definition file */
);

#endif /* MTFPROFILE_H */
