/******************************************************************************
NAME:           process_bin

PURPOSE:
Calculate the mean and standard deviation image values for the specified bin.

RETURN VALUE:   Type int

Value           Description
-----           -----------
SUCCESS     Valid bin processed
ERROR       Bin error encountered (e.g., empty bin)

NOTES:
This routine calculates the distance from the input pixel to the input target
centerline, computes the associated target bin from the target GSD and
oversample rate, and computes the target orientation angle from the target
azimuth and image azimuth.

******************************************************************************/
#include "mtfprofile.h"

/* Process the requested bin to populate the profile */
int process_bin            /* Return SUCCESS or ERROR */
(
    int numpixel,          /* I: Number of entries in input table */
    PIXEL_RECORD ptable[], /* I: Input image sample table */
    int bin,               /* I: Current bin to process */
    double outliertol,     /* I: T-dist outlier threshold */
    PROFILEREC *bin_prof   /* O: bin profile mean and standard dev */
)
{
    double *pixval;     /* Pixel values for current bin */
    int *index;         /* Record indices for current bin */
    int *non_outlier_flag; /* Non-outlier flags for current bin */
    double sum_v;       /* Sum of valid pixel values */
    double sum_sqr_v;   /* Sum of the squares */
    double mean_v;      /* Mean of valid pixel values */
    double stddev_v;    /* Standard deviation of valid pixels */
    int npts;           /* Number of points in bin */
    int numval;         /* Number of valid points in bin */
    int i;              /* Counter */

    /* Allocate the data, index, and flag arrays - use full length of pixel
       table */
    pixval = (double *)malloc(numpixel * sizeof(double));
    if (pixval == NULL)
    {
        IAS_LOG_ERROR("Error allocating pixel list for bin %d", bin);
        return ERROR;
    }
    index = (int *)malloc(numpixel * sizeof(int));
    if (index == NULL)
    {
        IAS_LOG_ERROR("Error allocating index list for bin %d", bin);
        free(pixval);
        return ERROR;
    }
    non_outlier_flag = (int *)malloc(numpixel * sizeof(int));
    if (non_outlier_flag == NULL)
    {
        IAS_LOG_ERROR("Error allocating outlier flag array for bin %d", bin);
        free(pixval);
        free(index);
        return ERROR;
    }

    /* Compile the list of pixels in the current bin */
    npts = 0;
    for (i = 0; i < numpixel; i++)
    {
        if (ptable[i].bin == bin)
        {
            pixval[npts] = ptable[i].sample_value;
            index[npts] = i;
            non_outlier_flag[npts] = 1;
            npts++;
        }
    }

    /* Eliminate outliers */
    filter_outliers(outliertol, npts, pixval, non_outlier_flag);

    /* Calculate mean and standard deviation of valid points */
    numval = 0;
    sum_v = 0.0;
    sum_sqr_v = 0.0;
    for (i = 0; i < npts; i++)
    {
        if (non_outlier_flag[i])
        {
            numval++;
            sum_v += pixval[i];
            sum_sqr_v += pixval[i] * pixval[i];
        }
        else
        {
            ptable[index[i]].bin = 0;
        }
    }

    /* Calculate mean and standard deviation if at least 2 valid points */
    if (numval > 1)
    {
        mean_v = sum_v / numval;
        stddev_v = sqrt((sum_sqr_v - numval * mean_v * mean_v)
                        / (numval - 1));

        /* Put a floor under the standard deviation */
        if (stddev_v < 0.5)
        {
            stddev_v = 0.5;   
        }
                                                 
        bin_prof->mean = mean_v;
        bin_prof->std_dev = stddev_v;
    }
    else if (numval > 0)
    {
        bin_prof->mean = sum_v;
        bin_prof->std_dev = -1.0;
        IAS_LOG_WARNING("Only one valid point in bin %d", bin);
    }
    else
    {
        bin_prof->mean = 0.0;
        bin_prof->std_dev = 0.0;
        IAS_LOG_ERROR("No valid points in bin %d.", bin);
        return ERROR;
    }

    /* Release the local arrays */
    free(pixval);
    free(index);
    free(non_outlier_flag);

    return SUCCESS;
}
