/******************************************************************************

NAME:           mtfprofile_test

PURPOSE:
Main test procedure for software to automatically locate and profile MTF target
windows predefined targets.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
EXIT_FAILURE    Error profiling MTF windows
EXIT_SUCCESS    Successfully profiled MTF windows

******************************************************************************/

#include <stdlib.h>         /* EXIT_FAILURE/EXIT_SUCCESS */
#include <limits.h>         /* PATH_MAX */
#include <string.h>         /* strncpy */
#include <libgen.h>         /* basename */

#include "cgreen/cgreen.h"
#include "mtfprofile.h"
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_satellite_attributes.h"


static char ut_data_dir[PATH_MAX];

Ensure test_calc_bin(void)
{
    /* parameters to calc_bin */
    PIXEL_RECORD ptable;
    TARGET target;
    int nsamp = 0;
    int oversamp = 0;
    IAS_LOG_MESSAGE_LEVEL old_level;

    memset(&ptable, 0, sizeof(ptable));
    memset(&target, 0, sizeof(target));

    /* suppress messages generated by failures */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* with zero input, the target centerline has length zero */
    ptable.bin = -9999;
    calc_bin(&ptable,target,nsamp,oversamp);
    assert_equal_with_message(ptable.bin, 0, "calc_bin did not reject 0 "
                              "length center line");

    /* restore messages */
    ias_log_set_output_level(old_level);

    /* define a target center line, everything else remains zero except
       gsd which needs to be non-zero. The pixel bin number should be 0 
       because the nsamp and oversamp used to calculate the bin are both 0 */
    target.center_line.start_point.x = 13;
    target.center_line.start_point.y = 14;
    target.center_line.end_point.x = 10;
    target.center_line.end_point.y = 10;
    ptable.gsd = 1; /* cannot be zero */
    calc_bin(&ptable,target,nsamp,oversamp);
    assert_equal_with_message(ptable.bin, 0, "calc_bin failed target 0 ");

    /* add nominal nsamp and oversamp values */
    nsamp = 128;
    oversamp = 8;
    calc_bin(&ptable,target,nsamp,oversamp);
    assert_equal_with_message(ptable.bin, 48, "calc_bin failed target 1 ");

    /* add ptable coordinates */
    ptable.coordinates.x = 10;
    ptable.coordinates.y = 10;
    calc_bin(&ptable,target,nsamp,oversamp);
    assert_equal_with_message(ptable.bin, 64, "calc_bin failed target 2 ");
}

Ensure test_calc_view_geometry(void)
{
    const int NUMPIXEL = 3;
    /* parameters to calc_view_geometry */
    PIXEL_RECORD ptable[NUMPIXEL];
    int numpixel = 0;
    double theta_m = -1.0;
    double range_m = -1.0;
    double gvel_m = -1.0;

    IAS_LOG_MESSAGE_LEVEL old_level;
    int i;

    memset(&ptable, 0, sizeof(ptable));

    /* suppress warning messages generated */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_ERROR);

    /* test with zero pixels for input */
    calc_view_geometry(numpixel, ptable, &theta_m, &range_m, &gvel_m);
    assert_double_equal(theta_m, 0.0);
    assert_double_equal(range_m, 0.0);
    assert_double_equal(gvel_m, 0.0);

    /* test with all zero bins on 3 pixels for input */
    numpixel = NUMPIXEL;
    calc_view_geometry(numpixel, ptable, &theta_m, &range_m, &gvel_m);
    assert_double_equal(theta_m, 0.0);
    assert_double_equal(range_m, 0.0);
    assert_double_equal(gvel_m, 0.0);

    /* restore logging */
    ias_log_set_output_level(old_level);

    /* test with bins set, but azimuth, range, ground velocity all 0's for 
       input */
    for (i = 0; i < numpixel; i++)
        ptable[i].bin = 1;
    calc_view_geometry(numpixel, ptable, &theta_m, &range_m, &gvel_m);
    assert_double_equal(theta_m, 0.0);
    assert_double_equal(range_m, 0.0);
    assert_double_equal(gvel_m, 0.0);

    /* test with azimuth, range, ground velocity non 0 for input */
    for (i = 0; i < numpixel; i++)
    {
        ptable[i].image_azimuth = 0.5 + 0.5 * i;
        ptable[i].target_range = 1.5 + 0.5 * i;
        ptable[i].ground_velocity = 2.5 + 0.5 * i;
    }
    calc_view_geometry(numpixel, ptable, &theta_m, &range_m, &gvel_m);
    assert_double_equal(theta_m, 1.0);
    assert_double_equal(range_m, 2.0);
    assert_double_equal(gvel_m, 3.0);
}

Ensure test_filter_outliers(void)
{
    const int NUM_PIX = 3;
    /* parameters to filter_outliers */
    double tolerance = 0.0; /* Tolerance value for the t-dist */
    int num_pix = 0;        /* Number of pixel values in the array */
    double data[NUM_PIX];   /* Array of pixel values */
    int out_flag[NUM_PIX];  /* Array of outlier flags */

    int i;

    memset(data, 0, sizeof(data));

    for (i = 0; i < NUM_PIX; i++)
        out_flag[i] = -1;

    /* test with all zero input, with num_pix nothing changes */
    filter_outliers(tolerance, num_pix, data, out_flag);
    for (i = 0; i < NUM_PIX; i++)
        assert_equal(out_flag[i], -1);

    /* test with num_pix set and all zero input */
    num_pix = NUM_PIX;
    filter_outliers(tolerance, num_pix, data, out_flag);
    for (i = 0; i < num_pix; i++)
        assert_equal(out_flag[i], 1);

    /* test with num_pix set and non-zero pixels, but 0 tolerance */
    for (i = 0; i < num_pix; i++)
    {
        out_flag[i] = -1;
        data[i] = 0.5 + 0.5 * i;
    }
    filter_outliers(tolerance, num_pix, data, out_flag);
    assert_equal(out_flag[0], 0);
    assert_equal(out_flag[1], 0);
    assert_equal(out_flag[2], 1);

    /* test with num_pix set and non-zero pixels and a tolerance */
    for (i = 0; i < num_pix; i++)
    {
        out_flag[i] = -1;
    }
    tolerance = 0.5;
    filter_outliers(tolerance, num_pix, data, out_flag);
    assert_equal(out_flag[0], 0);
    assert_equal(out_flag[1], 1);
    assert_equal(out_flag[2], 1);

    /* test with num_pix set and non-zero pixels and a tolerance */
    for (i = 0; i < num_pix; i++)
    {
        out_flag[i] = -1;
    }
    tolerance = 0.95;
    filter_outliers(tolerance, num_pix, data, out_flag);
    assert_equal(out_flag[0], 1);
    assert_equal(out_flag[1], 1);
    assert_equal(out_flag[2], 1);

}

Ensure test_process_bin(void)
{
    const int NUM_PIX = 3;

    /* process_bin parameters */
    int numpixel = 0;             /* Number of entries in input table */
    PIXEL_RECORD ptable[NUM_PIX]; /* Input image sample table */
    int bin = 0;                  /* Current bin to process */
    double outliertol = 0.0;      /* T-dist outlier threshold */
    PROFILEREC bin_prof = { -1.0, -1.0 };

    /* return value from process_bin */
    int status;

    IAS_LOG_MESSAGE_LEVEL old_level;
    int i;                        /* looping variable */

    memset(ptable, 0, sizeof(ptable));

    /* suppress all messages generated */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* test with zero points, should return ERROR */
    status = process_bin(numpixel, ptable, bin, outliertol, &bin_prof);
    assert_equal(status, ERROR);

    /* test with invalid number of points, should return ERROR */
    numpixel = -1;
    status = process_bin(numpixel, ptable, bin, outliertol, &bin_prof);
    assert_equal(status, ERROR);

    /* restore logging */
    ias_log_set_output_level(old_level);

    /* test with 3 points and all zeros, should return SUCCESS */
    numpixel = NUM_PIX;
    status = process_bin(numpixel, ptable, bin, outliertol, &bin_prof);
    assert_equal(status, SUCCESS);
    assert_double_equal(bin_prof.mean,0.0);
    assert_double_equal(bin_prof.std_dev,0.5);

    /* test with non-zeros, should return std dev = -1 */
    for (i = 0; i < numpixel; i++)
    {
        ptable[i].sample_value = 5 + 0.5 * i;
    }

    status = process_bin(numpixel, ptable, bin, outliertol, &bin_prof);
    assert_equal(status, SUCCESS);
    assert_double_equal(bin_prof.mean,6.0);
    assert_double_equal(bin_prof.std_dev,-1.0);

    /* test with non-zero tolerance, should return values */
    for (i = 0; i < numpixel; i++)
    {
        ptable[i].sample_value = 5 + 0.5 * i;
    }
    outliertol = 0.95;
    status = process_bin(numpixel, ptable, bin, outliertol, &bin_prof);
    assert_equal(status, SUCCESS);
    assert_double_equal(bin_prof.mean,5.5);
    assert_double_equal(bin_prof.std_dev,0.5);

}

Ensure test_get_profile_parms(void)
{
    char param_filename[PATH_MAX];     /* Parameter file name */
    char target_file[PATH_MAX];        /* Name of target prarameter file */
    int numbands = -1;                 /* Number of bands to process */
    int band_list[IAS_MAX_NBANDS];     /* Bands to process */
    int numsca = -1;                   /* Number of SCAs to process */
    int sca_list[IAS_MAX_NSCAS];       /* SCAs to process */
    int nsamp = -1;                    /* Number of samples in profile */
    int oversamp = -1;                 /* Profile oversampling factor */
    double outthresh = -1.0;           /* T-dist outlier threshold */
    char omf_file[PATH_MAX];           /* Name of the the OMF */

    char test_target_file[PATH_MAX];   /* Name of target prarameter file */
    char test_omf_file[PATH_MAX];      /* Name of the the OMF */

    int status;
    int count;

    IAS_LOG_MESSAGE_LEVEL old_level;
    char cmd[PATH_MAX];

    /* suppress all messages generated */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* test with NULL file pointer */
    status = get_profile_parms(NULL, target_file, &numbands, band_list, 
        &numsca, sca_list, &nsamp, &oversamp, &outthresh, omf_file);         
    assert_equal(status, ERROR);

    /* test with null string for file */
    strcpy(param_filename, "");
    status = get_profile_parms(param_filename, target_file, &numbands, 
        band_list, &numsca, sca_list, &nsamp, &oversamp, &outthresh, omf_file);
    assert_equal(status, ERROR);

    /* test with non-null string that's not a file */
    strncpy(param_filename, "test0.odl", sizeof(param_filename));
    status = get_profile_parms(param_filename, target_file, &numbands, 
        band_list, &numsca, sca_list, &nsamp, &oversamp, &outthresh, omf_file);
    assert_equal(status, ERROR);

    /* test with valid file that's empty */
    strncpy(param_filename, "test1.odl", sizeof(param_filename));
    /* create empty file */
    count = snprintf(cmd, sizeof(cmd), "/bin/touch %s", param_filename);
    if ((count < 0) || (count >= sizeof(cmd)))
    {
        assert_true_with_message(0, "Cannot set cmd string, count = %d", count);
        return;
    }
    
   
    status = system(cmd);
    assert_true_with_message(status == 0,"Cannot create temp odl");
    status = get_profile_parms(param_filename, target_file, &numbands, 
        band_list, &numsca, sca_list, &nsamp, &oversamp, &outthresh, omf_file);
    assert_equal(status, ERROR);
    count = snprintf(cmd, sizeof(cmd), "/bin/rm %s", param_filename);
    if ((count < 0) || (count >= sizeof(cmd)))
    {
        assert_true_with_message(0, "Cannot set cmd string, count = %d", count);
        return;
    }
    
   
    status = system(cmd);
    assert_true_with_message(status == 0,"Cannot remove temp odl");

    /* test with valid .odl pointing to invalid .omf */
    count = snprintf(param_filename, sizeof(param_filename), 
                     "%s/input/MTFPROFILE_0.odl", ut_data_dir);
    if ((count < 0) || (count >= sizeof(param_filename)))
    {
        assert_true_with_message(0, "Cannot set param_filename string, count "
            "= %d", count);
        return;
    }
    status = get_profile_parms(param_filename, target_file, &numbands, 
        band_list, &numsca, sca_list, &nsamp, &oversamp, &outthresh, omf_file);
    assert_equal(status, ERROR);

    /* restore messages generated */
    ias_log_set_output_level(old_level);

    /* test with valid .odl pointing to valid .omf */
    count = snprintf(param_filename, sizeof(param_filename), 
                     "%s/input/MTFPROFILE_1.odl", ut_data_dir);
    if ((count < 0) || (count >= sizeof(param_filename)))
    {
        assert_true_with_message(0, "Cannot set param_filename string, count "
            "= %d", count);
        return;
    }
    status = get_profile_parms(param_filename, target_file, &numbands, 
        band_list, &numsca, sca_list, &nsamp, &oversamp, &outthresh, omf_file);
    assert_equal(status, SUCCESS);
    count = snprintf(test_target_file, sizeof(test_target_file),
                     "%s/input/MTF_Targets_022039.table", ut_data_dir);
    if ((count < 0) || (count >= sizeof(test_target_file)))
    {
        assert_true_with_message(0, "Cannot set test_target_file, count "
            "= %d", count);
        return;
    }
    assert_string_equal(target_file, test_target_file);
    assert_equal(numbands,3);
    assert_equal(band_list[0],4);
    assert_equal(band_list[1],5);
    assert_equal(band_list[2],6);
    assert_equal(numsca,3);
    assert_equal(sca_list[0],8);
    assert_equal(sca_list[1],9);
    assert_equal(sca_list[2],10);
    assert_equal(nsamp,150);
    assert_equal(oversamp,9);
    assert_double_equal(outthresh,0.5);
    count = snprintf(test_omf_file, sizeof(test_omf_file), "I10.omf");
    if ((count < 0) || (count >= sizeof(test_omf_file)))
    {
        assert_true_with_message(0, "Cannot set test_omf_file, count = %d",
            count);
        return;
    }
    assert_string_equal(omf_file,test_omf_file);

    /* test with valid .odl pointing to valid .omf, and using defaults */
    count = snprintf(param_filename, sizeof(param_filename), 
                     "%s/input/MTFPROFILE_2.odl", ut_data_dir);
    if ((count < 0) || (count >= sizeof(param_filename)))
    {
        assert_true_with_message(0, "Cannot set param_filename, count = %d",
            count);
        return;
    }
    status = get_profile_parms(param_filename, target_file, &numbands, 
        band_list, &numsca, sca_list, &nsamp, &oversamp, &outthresh, omf_file);
    assert_equal(status, SUCCESS);
    count = snprintf(test_target_file, sizeof(test_target_file),
                     "%s/input/MTF_Targets_022039.table", ut_data_dir);
    if ((count < 0) || (count >= sizeof(test_target_file)))
    {
        assert_true_with_message(0, "Cannot set test_target_file, count = %d",
            count);
        return;
    }
    assert_string_equal(target_file, test_target_file);
    assert_equal(numbands,11);
    assert_equal(band_list[0],1);
    assert_equal(band_list[1],2);
    assert_equal(band_list[2],3);
    assert_equal(band_list[3],4);
    assert_equal(band_list[4],5);
    assert_equal(band_list[5],6);
    assert_equal(band_list[6],7);
    assert_equal(band_list[7],8);
    assert_equal(band_list[8],9);
    assert_equal(band_list[9],10);
    assert_equal(band_list[10],11);
    assert_equal(numsca,14);
    assert_equal(sca_list[0],1);
    assert_equal(sca_list[1],2);
    assert_equal(sca_list[2],3);
    assert_equal(sca_list[3],4);
    assert_equal(sca_list[4],5);
    assert_equal(sca_list[5],6);
    assert_equal(sca_list[6],7);
    assert_equal(sca_list[7],8);
    assert_equal(sca_list[8],9);
    assert_equal(sca_list[9],10);
    assert_equal(sca_list[10],11);
    assert_equal(sca_list[11],12);
    assert_equal(sca_list[12],13);
    assert_equal(sca_list[13],14);
    assert_equal(nsamp,128);
    assert_equal(oversamp,8);
    assert_double_equal(outthresh,0.95);
    count = snprintf(test_omf_file, sizeof(test_omf_file), "I10.omf");
    if ((count < 0) || (count >= sizeof(test_omf_file)))
    {
        assert_true_with_message(0, "Cannot set test_omf_file, count = %d",
            count);
        return;
    }
    assert_string_equal(omf_file, test_omf_file);
}

Ensure test_load_pixels(void)
{
    PIXEL_RECORD *ptable = NULL;  /* Loaded pixel structures */
    int target_id = -1;           /* Index of selected target */
    int npoly = -1;               /* Number of polygons */
    int iband = -1;               /* Selected band number */
    int isca = -1;                /* Selected SCA number */

    int num_pix = -1;

    /* test with 2 polygons */
    /* target_id, iband, and isca = -1, so returns 0 */
    npoly = 2;
    num_pix = count_pixels(target_id, npoly, iband, isca);
    assert_equal(num_pix,0);

    /* test with matching target_id, iband, and isca, should return ? */
    target_id = 0;
    iband = 7;
    isca = 9;
    num_pix = count_pixels(target_id, npoly, iband, isca);
    assert_equal(num_pix,12316);

    /* calling count_pixels is essentially calling load_pixels with a NULL
       ptable, so all of the above checks apply to load_pixels and we only
       need to test passing and populating a valid ptable */
    /* Allocate pixel table */
    ptable = (PIXEL_RECORD *)malloc(num_pix * sizeof(PIXEL_RECORD));
    assert_not_equal(ptable,NULL);
    if (ptable == NULL)
    {
        return;
    }
    memset(ptable, 0, sizeof(num_pix * sizeof(PIXEL_RECORD)));
    num_pix = load_pixels(ptable, num_pix, target_id, npoly, iband, isca);
    assert_equal(num_pix,12316);
    assert_equal(ptable[0].band_number, iband);
    assert_equal(ptable[0].sca_number, isca);
    assert_equal(ptable[0].bin, -1);
    assert_equal(ptable[0].line_coordinate, 3549);
    assert_equal(ptable[0].sample_coordinate, 319);
    assert_equal(ptable[0].coordinates.x, 808059);
    assert_equal(ptable[0].coordinates.y, 3346753);
    assert_double_equal(ptable[0].distance, 0.0);
    assert_double_equal(ptable[0].gsd, 30.0);
    assert_double_equal(ptable[0].ground_velocity, 6841.384069 );
    assert_double_equal(ptable[0].image_azimuth, 0.198154 );
    assert_double_equal(ptable[0].target_range, 705873.095945 );
    assert_equal(ptable[0].sample_value, 317);
    assert_equal(ptable[num_pix/2].band_number, iband);
    assert_equal(ptable[num_pix/2].sca_number, isca);
    assert_equal(ptable[num_pix/2].bin, -1);
    assert_equal(ptable[num_pix/2].line_coordinate, 3681);
    assert_equal(ptable[num_pix/2].sample_coordinate, 257);
    assert_equal(ptable[num_pix/2].coordinates.x, 805553);
    assert_equal(ptable[num_pix/2].coordinates.y, 3343375);
    assert_double_equal(ptable[num_pix/2].distance, 0.0);
    assert_double_equal(ptable[num_pix/2].gsd, 30.0);
    assert_double_equal(ptable[num_pix/2].ground_velocity, 6838.434640);
    assert_double_equal(ptable[num_pix/2].image_azimuth, 0.206055);
    assert_double_equal(ptable[num_pix/2].target_range, 705860.144568);
    assert_equal(ptable[num_pix/2].sample_value, 158);
    assert_equal(ptable[num_pix-1].band_number, iband);
    assert_equal(ptable[num_pix-1].sca_number, isca);
    assert_equal(ptable[num_pix-1].bin, -1);
    assert_equal(ptable[num_pix-1].line_coordinate, 3820);
    assert_equal(ptable[num_pix-1].sample_coordinate, 213);
    assert_equal(ptable[num_pix-1].coordinates.x, 803523 );
    assert_equal(ptable[num_pix-1].coordinates.y, 3339689 );
    assert_double_equal(ptable[num_pix-1].distance, 0.0);
    assert_double_equal(ptable[num_pix-1].gsd, 30.0);
    assert_double_equal(ptable[num_pix-1].ground_velocity, 6841.056822);
    assert_double_equal(ptable[num_pix-1].image_azimuth, 0.199189);
    assert_double_equal(ptable[num_pix-1].target_range, 705849.742788);
    assert_equal(ptable[num_pix-1].sample_value, 145);

    free(ptable);
}


Ensure test_update_omf(void)
{
    /* parameters to update_omf */
    char *omf_filename = NULL;    /* OMF filename */
    char *mtf_target_file = NULL; /* MTF-specific target definition file */
    char cmd[PATH_MAX];

    /* return value from update_omf */
    int status;
    int count;

    IAS_LOG_MESSAGE_LEVEL old_level;

    /* suppress all messages generated */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* test for NULL pointers */
    status = update_omf(omf_filename, mtf_target_file);
    assert_equal(status, ERROR);

    /* test for OMF NULL and MTF empty */
    mtf_target_file = "";
    status = update_omf(omf_filename, mtf_target_file);
    assert_equal(status, ERROR);

    /* test for OMF empty and MTF empty */
    omf_filename = "";
    status = update_omf(omf_filename, mtf_target_file);
    assert_equal(status, ERROR);

    /* test for OMF empty and MTF NULL */
    mtf_target_file = NULL;
    status = update_omf(omf_filename, mtf_target_file);
    assert_equal(status, ERROR);

    /* test for OMF empty and MTF empty */
    mtf_target_file = "";
    status = update_omf(omf_filename, mtf_target_file);
    assert_equal(status, ERROR);

    /* restore messages generated */
    ias_log_set_output_level(old_level);

    /* test with valid OMF and a test string for MTF */
    omf_filename = "update_omf_test.omf";
    count = snprintf(cmd, sizeof(cmd), "/bin/cp %s/input/%s .", 
                     ut_data_dir, omf_filename);
    if ((count < 0) || (count >= sizeof(cmd)))
    {
        assert_true_with_message(0, "Cannot set cmd string, count = %d",
            count);
        return;
    }
    
   
    status = system(cmd);
    assert_true_with_message(status == 0,"Cannot copy temp omf");
    mtf_target_file = "UPDATE_OMF_TEST";
    status = update_omf(omf_filename, mtf_target_file);
    assert_equal(status, SUCCESS);

}

Ensure  test_write_profile(void)
{
    const int NSAMP = 5;

    /* parameters to write_profile */
    char *target_name = NULL;   /* Target name */
    PROFILEREC *profile = NULL; /* Profile data to write */
    int nsamp = -1;             /* Number of bins in profile */
    double gsd = 500.0;         /* Image GSD */
    int oversamp = 10;          /* Profile oversampling factor */
    double theta_m = 2.1;       /* Mean azimuth angle */
    double range_m = 3.2;       /* Mean target range */
    double gvel_m = 4.3;        /* Mean ground velocity */
    int target_id = 9900;       /* Target index (used in file name) */
    int bandnum = 99;           /* Band number (used in file name) */

    /* return status from write_profile */
    int status;

    IAS_LOG_MESSAGE_LEVEL old_level;
    int i; /* loop counter */

    /* suppress all messages generated */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* test with NULL pointers */
    status = write_profile(target_name, profile, nsamp, gsd, oversamp,
                           theta_m, range_m, gvel_m, target_id, bandnum);
    assert_equal(status,ERROR);

    /* test with target_name set, but profile is NULL */
    target_name = "TARGET_NAME";
    status = write_profile(target_name, profile, nsamp, gsd, oversamp,
                           theta_m, range_m, gvel_m, target_id, bandnum);
    assert_equal(status,ERROR);

    /* restore messages generated */
    ias_log_set_output_level(old_level);

    /* test with profile set */
    nsamp = NSAMP;
    profile = (PROFILEREC*)malloc(nsamp * sizeof(PROFILEREC));
    assert_not_equal(profile,NULL);
    if (profile == NULL)
    {
        return;
    }
    for (i = 0; i < nsamp; i++)
    {
        profile[i].mean = 3.0 + 0.1 * i;
        profile[i].std_dev = 3.0 + nsamp + 0.1 * i;
    }
    status = write_profile(target_name, profile, nsamp, gsd, oversamp,
                           theta_m, range_m, gvel_m, target_id, bandnum);
    assert_equal(status,SUCCESS);
    free(profile);
}

int main(int argc, char *argv[])
{
    char *ias_unit_test_directory = NULL;  /* Top level of IAS unit test 
                                              data directory */

    /* subdirectory path to our unit test data */
    const char *ut_subdir = "ias/gps/mtf";

    int status;
    int count;

    char progname[PATH_MAX] = "mtfprofile_unit_tests";

    TestSuite *suite = create_test_suite();
    add_test(suite, test_calc_bin);
    add_test(suite, test_calc_view_geometry);
    add_test(suite, test_filter_outliers);
    add_test(suite, test_process_bin);
    add_test(suite, test_get_profile_parms);
    add_test(suite, test_load_pixels);
    add_test(suite, test_update_omf);
    add_test(suite, test_write_profile);

    status = ias_log_initialize(progname);
    if (status != SUCCESS)
    {
        printf("Failure to initialize logging library\n");
        return EXIT_FAILURE;
    }

    /* Get unit test directory */
    ias_unit_test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (ias_unit_test_directory == NULL)
    {
        printf("The IAS_UNIT_TEST_DATA_DIR environment variable is not set\n");
        return EXIT_FAILURE;
    }

    count = snprintf(ut_data_dir, sizeof(ut_data_dir), "%s/%s/mtfprofile", 
                     ias_unit_test_directory, ut_subdir);
    if ((count < 0) || (count >= sizeof(ut_data_dir)))
    {
        assert_true_with_message(0, "Cannot set UT data dir, count = %d",
            count);
        return EXIT_FAILURE;
    }

    /* Initialize the satellite attributes */
    if (ias_sat_attr_initialize(IAS_L8) != SUCCESS)
    {
        printf("Initialize satellite attributes failed.\n");
        return EXIT_FAILURE;
    }

    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}


