/******************************************************************************
NAME:           write_profile

PURPOSE:
Write out the assembled target profile and associated metadata for subsequent
input to and use by mtfestimate.


RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Output profile file successfully written
ERROR    Operation failed

NOTES:
This routine will create and open the output target profile file, write the
number of points, GSD, oversample factor, mean azimuth, mean range, and mean
ground velocity metadata parameters, then write out the profile values, one per
bin, followed by the corresponding standard deviations.

******************************************************************************/
#include <stdio.h>
#include <string.h>
#include <limits.h>

#include "mtfprofile.h"

/* Write out the profile (mean and standard deviation of each bin) */
int write_profile       /* Return SUCCESS or ERROR */
(
    char *target_name,  /* I: Target name */
    PROFILEREC profile[],/* I: Profile data to write */
    int nsamp,          /* I: Number of bins in profile */
    double gsd,         /* I: Image GSD */
    int oversamp,       /* I: Profile oversampling factor */
    double theta_m,     /* I: Mean azimuth angle */
    double range_m,     /* I: Mean target range */
    double gvel_m,      /* I: Mean ground velocity */
    int target_id,      /* I: Target index (used in file name) */
    int bandnum         /* I: Band number (used in file name)  */
)
{
    FILE *ofile;        /* Output file pointer */
    char filename[PATH_MAX];/* Output file name (constructed) */
    int i;              /* Counter */
    int size;           /* Size returned by snprintf */
    int status;         /* Function return status */

    if (target_name == NULL)
    {
        IAS_LOG_ERROR("target_name cannot be NULL");
        return ERROR;
    }

    if (profile == NULL)
    {
        IAS_LOG_ERROR("profile cannot be NULL");
        return ERROR;
    }

    /* Construct the output file name */
    size = snprintf(filename, sizeof(filename), "Profile_Tar%dBand%d.odl",
                 target_id, bandnum);
    if ((size < 0) || (size >= sizeof(filename)))
    {
        IAS_LOG_ERROR("Cannot build profile file name for target id %d band "
            "number %d", target_id, bandnum);
        return ERROR;
    }

    ofile = fopen(filename, "w");
    if (ofile == NULL)
    {
        IAS_LOG_ERROR("Opening output profile file %s", filename);
        return ERROR;
    }

    /* Write out in ODL form */
    if (fprintf(ofile, "OBJECT = TARGET_PROFILE\n"
                   "  TARGET_NAME = \"%s\"\n"
                   "  TARGET_ID = %d\n"
                   "  BAND = %d\n"
                   "  NSAMP = %d\n"
                   "  GSD = %6.3lf\n"
                   "  OVERSAMP = %d\n"
                   "  TARGET_AZIMUTH = %12.9lf\n"
                   "  TARGET_RANGE = %10.3lf\n"
                   "  GROUND_VELOCITY = %8.3lf\n"
                   "  PROFILE = ( %lf", 
                   target_name, target_id, bandnum, nsamp, gsd, 
                   oversamp, theta_m, range_m, gvel_m, profile[0].mean) < 0)
    {
        IAS_LOG_ERROR("Writing to output profile file %s", filename);
        return ERROR;
    }

    for (i = 1; i < nsamp; i++)
    {
        if (i % 8 == 0) 
        {
            if (fprintf(ofile, ",\n              %lf", profile[i].mean) < 0)
            {
                IAS_LOG_ERROR("Writing to output profile file %s", filename);
                return ERROR;
            }
        }
        else 
        {
            if (fprintf(ofile, ", %lf", profile[i].mean) < 0)
            {
                IAS_LOG_ERROR("Writing to output profile file %s", filename);
                return ERROR;
            }
        }
    }
    if (fprintf(ofile, " )\n  SIGMAS = ( %lf", profile[0].std_dev) < 0)
    {
        IAS_LOG_ERROR("Writing to output profile file %s", filename);
        return ERROR;
    }
                   
    for (i = 1; i < nsamp; i++)
    {
        if (i % 8 == 0) 
        {
            if (fprintf(ofile, ",\n             %lf", profile[i].std_dev) < 0)
            {
                IAS_LOG_ERROR("Writing to output profile file %s", filename);
                return ERROR;
            }
        }
        else 
        {
            if (fprintf(ofile, ", %lf", profile[i].std_dev) < 0)
            {
                IAS_LOG_ERROR("Writing to output profile file %s", filename);
                return ERROR;
            }
        }
    }
    if (fprintf(ofile, " )\nEND_OBJECT = TARGET_PROFILE\nEND\n") < 0)
    {
        IAS_LOG_ERROR("Writing to output profile file %s", filename);
        return ERROR;
    }

    /* Close output profile file */
    status = fclose(ofile);
    if (status != 0)
    {
        IAS_LOG_ERROR("Closing output profile file %s", filename);
        return ERROR;
    }

    return SUCCESS;
}
