/******************************************************************************

NAME:           mtf_shared_unit_test

PURPOSE:
Main test procedure for shared software to automatically locate and profile
MTF target windows predefined targets.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
EXIT_FAILURE    Error profiling MTF windows
EXIT_SUCCESS    Successfully profiled MTF windows

******************************************************************************/

#include <stdlib.h>         /* EXIT_FAILURE/EXIT_SUCCESS */
#include <limits.h>         /* PATH_MAX */
#include <string.h>         /* strncpy */
#include <libgen.h>         /* basename */

#include "cgreen/cgreen.h"
#include "model.h"
#include "get_targets.h"
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_satellite_attributes.h"


static char ut_data_dir[PATH_MAX];

Ensure test_create_frequency_domain_bridge_model(void)
{
    /* parameters to create_frequency_domain_bridge_model */
    int nspan = -1;                   /* Number of spans in current bridge */
    BRIDGE_PARMS bridgeparms;         /* Input bridge model parameter values:
                                         left bridge DN, right bridge DN,
                                         water DN, water DN asymmetry, span
                                         width, span gap */
    int npts = -1;                    /* Number of data points to model */
    int osamp = -1;                   /* Oversampling factor */
    IAS_COMPLEX freq_dom_bridge[MAXNPTS]; /* Output frequency domain bridge
                                         model*/

    /* counting variable */
    int i;

    for (i = 0; i < MAXNPTS; i++)
        freq_dom_bridge[i].re = freq_dom_bridge[i].im = 0.0;

    /* test with non-NULL pointers */
    create_frequency_domain_bridge_model(nspan, &bridgeparms, npts, osamp, 
        freq_dom_bridge);
    assert_double_equal(freq_dom_bridge[0].re, 0.0);
    assert_double_equal(freq_dom_bridge[0].im, 0.0);
    assert_double_equal(freq_dom_bridge[1].re, 0.0);
    assert_double_equal(freq_dom_bridge[1].im, 0.0);
    assert_double_equal(freq_dom_bridge[2].re, 0.0);
    assert_double_equal(freq_dom_bridge[2].im, 0.0);
    assert_double_equal(freq_dom_bridge[3].re, 0.0);
    assert_double_equal(freq_dom_bridge[3].im, 0.0);
    assert_double_equal(freq_dom_bridge[4].re, 0.0);
    assert_double_equal(freq_dom_bridge[4].im, 0.0);

    /* test with additional values */
    nspan = 2;
    npts = 5;
    osamp = 8;
    bridgeparms.rad_water = 10.0;
    bridgeparms.rad_asym = 15.0;
    bridgeparms.gsd = 30.0;
    bridgeparms.span[0].rad_bridge = 25.0;
    bridgeparms.span[0].width = 250.0;
    bridgeparms.span[0].offset = 500.0;
    bridgeparms.span[1].rad_bridge = 35.0;
    bridgeparms.span[1].width = 350.0;
    bridgeparms.span[1].offset = 600.0;
    create_frequency_domain_bridge_model(nspan, &bridgeparms, npts, osamp, 
        freq_dom_bridge);
    assert_double_equal(freq_dom_bridge[0].re, 3420.833333);
    assert_double_equal(freq_dom_bridge[0].im, 0.0);
    assert_double_equal(freq_dom_bridge[1].re, 44.795473);
    assert_double_equal(freq_dom_bridge[1].im, -5.96831036594766);
    assert_double_equal(freq_dom_bridge[2].re, 3.61563771e-13);
    assert_double_equal(freq_dom_bridge[2].im, 7.95774715459477);
    assert_double_equal(freq_dom_bridge[3].re, -22.397736);
    assert_double_equal(freq_dom_bridge[3].im, -8.95246554892069);
    assert_double_equal(freq_dom_bridge[4].re, 44.795473);
    assert_double_equal(freq_dom_bridge[4].im, 5.96831036594766);

}

Ensure test_create_frequency_domain_stf_model(void)
{
    const int NPTS = 5;

    /* parameters to fd_stf */
    STF_FD_PARMS stf_fd_parms;   /* Input STF model parameter values:
                                    optical blur, detector dimension, phase */
    double gsd = -1;             /* Model ground sample distance (meters) */
    double range = -1;           /* Target range (for scaling to meters) */
    double velocity = -1;        /* Target velocity (for scaling to meters) */
    double angle = -1;           /* Azimuth angle (for rotating the STF) */
    int npts = -1;               /* Number of data points to model */
    int osamp = -1;              /* Oversampling factor */
    IAS_COMPLEX freq_dom_stf[NPTS];/* Output frequency domain STF model */

    /* loop counter */
    int i;

    stf_fd_parms.g_optics = -1;    /* guassian optical blur (urad) */
    stf_fd_parms.xt_detector = -1; /* cross-track detector dimension (urad) */
    stf_fd_parms.at_detector = -1; /* along-track detector dimension (urad) */
    stf_fd_parms.integ = -1;       /* detector integration time (msec) */
    stf_fd_parms.diffusion = -1;   /* charge diffusion distance (not used) */
    stf_fd_parms.phase = -1;       /* model/data profile registration offset 
                                      (m) */

    for (i = 0; i < NPTS; i++)
    {
        freq_dom_stf[i].re = freq_dom_stf[i].im = 0.0;
    }

    /* test with npts < 1, nothing changes */
    create_frequency_domain_stf_model(&stf_fd_parms, gsd, range, velocity, 
        angle, npts, osamp, freq_dom_stf);
    assert_double_equal(freq_dom_stf[0].re, 0.0);
    assert_double_equal(freq_dom_stf[0].im, 0.0);
    assert_double_equal(freq_dom_stf[1].re, 0.0);
    assert_double_equal(freq_dom_stf[1].im, 0.0);
    assert_double_equal(freq_dom_stf[2].re, 0.0);
    assert_double_equal(freq_dom_stf[2].im, 0.0);
    assert_double_equal(freq_dom_stf[3].re, 0.0);
    assert_double_equal(freq_dom_stf[3].im, 0.0);
    assert_double_equal(freq_dom_stf[4].re, 0.0);
    assert_double_equal(freq_dom_stf[4].im, 0.0);

    /* test with npts set */
    npts = NPTS;
    create_frequency_domain_stf_model(&stf_fd_parms, gsd, range, velocity, 
        angle, npts, osamp, freq_dom_stf);
    assert_double_equal(freq_dom_stf[0].re, 1.0);
    assert_double_equal(freq_dom_stf[0].im, 0.0);
    assert_double_equal(freq_dom_stf[1].re, 0.99999821702391);
    assert_double_equal(freq_dom_stf[1].im, 0.0);
    assert_double_equal(freq_dom_stf[2].re, 0.99999437154125);
    assert_double_equal(freq_dom_stf[2].im, 0.0);
    assert_double_equal(freq_dom_stf[3].re, 0.99999634087083);
    assert_double_equal(freq_dom_stf[3].im, 0.0);
    assert_double_equal(freq_dom_stf[4].re, 0.99999821702391);
    assert_double_equal(freq_dom_stf[4].im, 0.0);

    /* set some additional parameters */
    stf_fd_parms.g_optics = 40.0;     /* guassian optical blur (urad) */
    stf_fd_parms.xt_detector = 100.0; /* cross-track detector dimension
                                         (urad) */
    stf_fd_parms.at_detector = 200.0; /* along-track detector dimension
                                         (urad) */
    stf_fd_parms.integ = 3.3;         /* detector integration time (msec) */
    stf_fd_parms.diffusion = -1;      /* charge diffusion distance (not
                                         used) */
    stf_fd_parms.phase = 1000.0;      /* model/data profile registration
                                         offset (m) */
    gsd = 30.0;                       /* Model ground sample distance
                                         (meters) */
    range = 5.0;                      /* Target range (for scaling to
                                          meters) */
    velocity = 7000.0;                /* Target velocity (for scaling to
                                          meters) */
    angle = 1.0;                      /* Azimuth angle (for rotating the
                                          STF) */
    osamp = 128;                      /* Oversampling factor */

    create_frequency_domain_stf_model(&stf_fd_parms, gsd, range, velocity, 
        angle, npts, osamp, freq_dom_stf);
    assert_double_equal(freq_dom_stf[0].re, 1.0);
    assert_double_equal(freq_dom_stf[0].im, 0.0);
    assert_double_equal(freq_dom_stf[1].re, 0.01847595110480);
    assert_double_equal(freq_dom_stf[1].im, 0.0);
    assert_double_equal(freq_dom_stf[2].re, -0.00436030515984);
    assert_double_equal(freq_dom_stf[2].im, 0.0);
    assert_double_equal(freq_dom_stf[3].re, -0.00499122484265);
    assert_double_equal(freq_dom_stf[3].im, 0.0);
    assert_double_equal(freq_dom_stf[4].re, 0.01847595110480);
    assert_double_equal(freq_dom_stf[4].im, 0.0);
}

Ensure test_compute_frequency_domain_optical_blur(void)
{
    /* parameters to compute_frequency_domain_optical_blur */
    double blur = -1.0; /* Input STF model optical blur parameter value */
    double freq = -1.0; /* Frequency (cycles per meter) */
    IAS_COMPLEX val;    /* Output frequency domain optics model */

    /* test with non-NULL ptr output */
    compute_frequency_domain_optical_blur(blur, freq, &val);
    assert_double_equal(val.re, 2.67528799e-09);
    assert_double_equal(val.im, 0.0);

    /* test with non-neg inputs */
    blur = 10.0;
    freq = 100.0;
    compute_frequency_domain_optical_blur(blur, freq, &val);
    assert_double_equal(val.re, 0.0);
    assert_double_equal(val.im, 0.0);

    /* test with non-neg inputs */
    blur = 0.1;
    freq = 0.3;
    compute_frequency_domain_optical_blur(blur, freq, &val);
    assert_double_equal(val.re, 0.982391584);
    assert_double_equal(val.im, 0.0);
}

Ensure test_compute_frequency_domain_charge_diffusion(void)
{
    /* parameters to compute_frequency_domain_charge_diffusion */
    double  invf0 = -1.0; /* Input STF model charge diffusion parameter
                             value */
    double  freq = -1.0;  /* Frequency (cycles per meter) */
    IAS_COMPLEX val;      /* Output frequency domain diffusion model */

    /* test with non-NULL pointer output */
    compute_frequency_domain_charge_diffusion(invf0, freq, &val);
    assert_double_equal(val.re, 1.86744273e-03);
    assert_double_equal(val.im, 0.0);

    /* test with additional values */
    invf0 = 0.1;
    freq = 0.01;
    compute_frequency_domain_charge_diffusion(invf0, freq, &val);
    assert_double_equal(val.re, 9.93736513e-01);
    assert_double_equal(val.im, 0.0);
}

Ensure test_compute_frequency_domain_detector(void)
{
    /* parameters to compute_frequency_domain_detector */
    double detdim = -1.0; /* Input STF model detector dimension parameter */
    double freq = -1.0;   /* Frequency (cycles per meter) */
    IAS_COMPLEX val = { -1.0, -1.0}; /* Output frequency domain detector
                                        model */

    /* test with non-NULL pointer */
    compute_frequency_domain_detector(detdim, freq, &val);
    assert_double_equal(val.re, 3.89817183e-17);
    assert_double_equal(val.im, 0.0);

    /* test with additional input*/
    detdim = 0.1;
    freq = 0.01;
    compute_frequency_domain_detector(detdim, freq, &val);
    assert_double_equal(val.re, 9.99998355e-01);
    assert_double_equal(val.im, 0.0);
}

Ensure test_sinc(void)
{
    /* input to sinc() */
    double x = 0.0;  /* Input number to compute sinc for */

    /* return value from sinc() */
    double y = 0.0;

    /* test with x < 1e-6 */
    y = sinc(x);
    assert_double_equal(y, 1.0);

    /* test with x == 0.1 */
    x = 0.1;
    y = sinc(x);
    assert_double_equal(y, 9.98334166e-01);

    /* test with x == 0.5 */
    x = 0.5;
    y = sinc(x);
    assert_double_equal(y, 9.58851077e-01);

    /* test with x == 10.0 */
    x = 10.0;
    y = sinc(x);
    assert_double_equal(y, -5.44021111e-02);
}

Ensure test_allocate_spatial_domain_bridge_model(void)
{
    const int NPTS = 10;

    /* input to allocate_spatial_domain_bridge_model */
    int npts = -1;          /* Number of points in profile */
    IAS_COMPLEX *fd_bridge_buf = NULL; /* Frequency domain bridge model */
    IAS_COMPLEX *fd_stf_buf = NULL; /* Frequency domain STF model */
    IAS_COMPLEX *model_buf = NULL; /* Composite target model (FD and SD) */

    /* return status */
    int status;

    /* initial log level */
    IAS_LOG_MESSAGE_LEVEL old_level;

    /* temporarily suppress error messages */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* test allocate_spatial_domain_bridge_model with negative point count */
    status = allocate_spatial_domain_bridge_model(npts, &fd_bridge_buf, 
        &fd_stf_buf, &model_buf);
    assert_equal(status, ERROR);

    /* restore original logging level, not expecting ERRORS/WARNINGS */
    ias_log_set_output_level(old_level);

    /* test allocate_spatial_domain_bridge_model with positive point count */
    npts = NPTS;
    status = allocate_spatial_domain_bridge_model(npts, &fd_bridge_buf, 
        &fd_stf_buf, &model_buf);
    assert_equal(status, SUCCESS);

    /* temporarily suppress error messages */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* test allocate_spatial_domain_bridge_model again with positive point 
       count */
    status = allocate_spatial_domain_bridge_model(npts, &fd_bridge_buf, 
        &fd_stf_buf, &model_buf);
    assert_equal(status, ERROR);

    /* restore original logging level, not expecting ERRORS/WARNINGS */
    ias_log_set_output_level(old_level);

    /* frees allocated arrays, nothing returned */
    free_spatial_domain_bridge_model(fd_bridge_buf, fd_stf_buf, model_buf);
}

Ensure test_generate_spatial_domain_target_profile(void)
{
    const int NPTS = 8; /* Must be a power of 2 for the FFT, else a loop
                           could occur */

    /* input to generate_spatial_domain_target_profile */
    int npts = -1;          /* Number of points in profile */
    int osamp = -1;         /* Oversampling factor */
    MTFMODEL mparms;        /* Input bridge and STF model parameter values:
                                  water DN, water DN asymmetry, gsd,
                                  {span DN, span width, span offset}
                                  for each span, optical blur, detector
                                  dimensions, phase */
    double range = -1;      /* Target range (for scaling to meters) */
    double velocity = -1;   /* Target velocity (for scaling to meters) */
    double angle = -1;      /* Bridge rotation angle (degrees) */
    double sd_targets[NPTS]; /* Output space domain target models */

    /* return status */
    int status;

    /* initial log level */
    IAS_LOG_MESSAGE_LEVEL old_level;

    /* temporarily suppress error messages */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    memset(sd_targets, 0, sizeof(sd_targets));
    memset(&mparms, 0, sizeof(mparms));

    /* test with negative point count and NULL output pointer */
    status = generate_spatial_domain_target_profile(npts, osamp, mparms, range,
                velocity, angle, NULL);
    assert_equal(status, ERROR);

    /* test with non-NULL pointer output */
    status = generate_spatial_domain_target_profile(npts, osamp, mparms, range,
                velocity, angle, sd_targets);
    assert_equal(status, ERROR);

    /* restore original logging level, not expecting ERRORS/WARNINGS */
    ias_log_set_output_level(old_level);

    /* test with non-negative point count */
    npts = NPTS;
    osamp = 1;
    mparms.bridge.gsd = 30.0;
    status = generate_spatial_domain_target_profile(npts, osamp, mparms, range,
                velocity, angle, sd_targets);
    assert_equal(status, SUCCESS);
    assert_double_equal(sd_targets[0], 0.0);
    assert_double_equal(sd_targets[1], 0.0);
    assert_double_equal(sd_targets[2], 0.0);
    assert_double_equal(sd_targets[3], 0.0);
    assert_double_equal(sd_targets[4], 0.0);
    assert_double_equal(sd_targets[5], 0.0);
    assert_double_equal(sd_targets[6], 0.0);
    assert_double_equal(sd_targets[7], 0.0);

    /* test with some additional values */
    osamp = 128;
    mparms.nspan = 2;
    mparms.bridge.rad_water = 10.0;
    mparms.bridge.rad_asym = 15.0;
    mparms.bridge.gsd = 30.0;
    mparms.bridge.span[0].rad_bridge = 45.0;
    mparms.bridge.span[0].width = 4500.0;
    mparms.bridge.span[0].offset = 1500.0;
    mparms.bridge.span[1].rad_bridge = 35.0;
    mparms.bridge.span[1].width = 2500.0;
    mparms.bridge.span[1].offset = 500.0;
    mparms.stf.g_optics = 100.0;
    mparms.stf.xt_detector = 20.0;
    mparms.stf.at_detector = 30.0;
    mparms.stf.integ = 3.3;
    mparms.stf.diffusion = 77.0;
    mparms.stf.phase = 300.0;
    status = generate_spatial_domain_target_profile(npts, osamp, mparms, range,
                velocity, angle, sd_targets);
    assert_equal(status, SUCCESS);
    assert_double_equal(sd_targets[0], 117341.361078);
    assert_double_equal(sd_targets[1], 117338.577016);
    assert_double_equal(sd_targets[2], 117360.702861);
    assert_double_equal(sd_targets[3], 117360.247653);
    assert_double_equal(sd_targets[4], 117355.139337);
    assert_double_equal(sd_targets[5], 117349.441367);
    assert_double_equal(sd_targets[6], 117349.570750);
    assert_double_equal(sd_targets[7], 117351.626604);

}

Ensure test_get_targets(void)
{
    /* parameters to get_targets */
    char target_file[PATH_MAX];      /* Target parameter file name */
    TARGET target_info[MAX_TARGETS]; /* Populated target structure array */

    /* return from get_target */
    int num_targets;

    /* return from snprintf */
    int n;

    /* initial log level */
    IAS_LOG_MESSAGE_LEVEL old_level;

    /* temporarily suppress error messages */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* test with NULL pointers */
    num_targets = get_targets(NULL, NULL);
    assert_equal(num_targets, ERROR);

    n = snprintf(target_file, sizeof(target_file),
                 "%s/mtfprofile/input/MTF_Targets_022039.table",
                 ut_data_dir);
    if (n < 0 || n > sizeof(target_file))
    {
        assert_true_with_message(0, "target_file buffer needs %d bytes", n);
        return;
    }

    num_targets = get_targets(target_file, NULL);
    assert_equal(num_targets, ERROR);

    /* restore original logging level, not expecting ERRORS/WARNINGS */
    ias_log_set_output_level(old_level);

    num_targets = get_targets(target_file, target_info);
    assert_equal(num_targets, 1);

    assert_equal(target_info[0].number_of_bands, 9);
    assert_equal(target_info[0].band_list[0], 1);
    assert_equal(target_info[0].band_list[1], 2);
    assert_equal(target_info[0].band_list[2], 3);
    assert_equal(target_info[0].band_list[3], 4);
    assert_equal(target_info[0].band_list[4], 5);
    assert_equal(target_info[0].band_list[5], 6);
    assert_equal(target_info[0].band_list[6], 7);
    assert_equal(target_info[0].band_list[7], 8);
    assert_equal(target_info[0].band_list[8], 9);
    assert_equal(target_info[0].band_list[9], -1);
    assert_equal(target_info[0].band_list[10], -1);
    assert_string_equal(target_info[0].target_name, "I-10_Old");
    assert_equal(target_info[0].number_of_spans, 2);
    assert_double_equal(target_info[0].span[0].width, 14.2);
    assert_double_equal(target_info[0].span[0].offset, -15.15);
    assert_double_equal(target_info[0].span[0].reflectance[0], 1370.0);
    assert_double_equal(target_info[0].span[0].reflectance[1], 1390.0);
    assert_double_equal(target_info[0].span[0].reflectance[2], 1150.0);
    assert_double_equal(target_info[0].span[0].reflectance[3], 1200.0);
    assert_double_equal(target_info[0].span[0].reflectance[4], 550.0);
    assert_double_equal(target_info[0].span[0].reflectance[5], 850.0);
    assert_double_equal(target_info[0].span[0].reflectance[6], 700.0);
    assert_double_equal(target_info[0].span[0].reflectance[7], 750.0);
    assert_double_equal(target_info[0].span[0].reflectance[8], 850.0);
    assert_double_equal(target_info[0].span[0].reflectance[9], 0.0);
    assert_double_equal(target_info[0].span[0].reflectance[10], 0.0);
    assert_double_equal(target_info[0].water[0], 970.0);
    assert_double_equal(target_info[0].water[1], 990.0);
    assert_double_equal(target_info[0].water[2], 700.0);
    assert_double_equal(target_info[0].water[3], 550.0);
    assert_double_equal(target_info[0].water[4], 200.0);
    assert_double_equal(target_info[0].water[5], 160.0);
    assert_double_equal(target_info[0].water[6], 150.0);
    assert_double_equal(target_info[0].water[7], 300.0);
    assert_double_equal(target_info[0].water[8], 160.0);
    assert_double_equal(target_info[0].water[9], -1.0);
    assert_double_equal(target_info[0].water[10], -1.0);
    assert_double_equal(target_info[0].asymmetry[0], 0.0);
    assert_double_equal(target_info[0].asymmetry[1], 0.0);
    assert_double_equal(target_info[0].asymmetry[2], 0.0);
    assert_double_equal(target_info[0].asymmetry[3], 0.0);
    assert_double_equal(target_info[0].asymmetry[4], 0.0);
    assert_double_equal(target_info[0].asymmetry[5], 0.0);
    assert_double_equal(target_info[0].asymmetry[6], 0.0);
    assert_double_equal(target_info[0].asymmetry[7], 0.0);
    assert_double_equal(target_info[0].asymmetry[8], 0.0);
    assert_double_equal(target_info[0].asymmetry[9], -1.0);
    assert_double_equal(target_info[0].asymmetry[10], -1.0);
    assert_equal(target_info[0].number_of_polygons, 2);
    assert_equal(target_info[0].polygon, NULL);
    assert_double_equal(target_info[0].height, 0.0);
    assert_double_equal(target_info[0].center_line.start_point.x, 
        808503.562500);
    assert_double_equal(target_info[0].center_line.start_point.y, 
        3346462.500000);
    assert_double_equal(target_info[0].center_line.end_point.x, 802998.000000);
    assert_double_equal(target_info[0].center_line.end_point.y, 
        3340122.250000);

}

int main(int argc, char *argv[])
{
    char *ias_unit_test_directory = NULL;  /* Top level of IAS unit test data
                                              directory */

    /* subdirectory path to our unit test data */
    const char *ut_subdir = "ias/gps/mtf";

    int status;
    int count;

    char progname[PATH_MAX] = "mtf_shared_unit_tests";

    TestSuite *suite = create_test_suite();
    add_test(suite, test_create_frequency_domain_bridge_model);
    add_test(suite, test_create_frequency_domain_stf_model);
    add_test(suite, test_compute_frequency_domain_optical_blur);
    add_test(suite, test_compute_frequency_domain_charge_diffusion);
    add_test(suite, test_compute_frequency_domain_detector);
    add_test(suite, test_sinc);
    add_test(suite, test_allocate_spatial_domain_bridge_model);
    add_test(suite, test_generate_spatial_domain_target_profile);
    add_test(suite, test_get_targets);

    status = ias_log_initialize(progname);
    if (status != SUCCESS)
    {
        printf("Failure to initialize logging library\n");
        return EXIT_FAILURE;
    }

    IAS_LOG_INFO("Initiating %s", progname);

    /* Get unit test directory */
    ias_unit_test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (ias_unit_test_directory == NULL)
    {
        printf("The IAS_UNIT_TEST_DATA_DIR environment variable is not set\n");
        return EXIT_FAILURE;
    }

    count = snprintf(ut_data_dir, sizeof(ut_data_dir), "%s/%s",
                     ias_unit_test_directory, ut_subdir);
    if ((count < 0) || (count > sizeof(ut_data_dir)))
    {
        assert_true_with_message(0, "Cannot set UT data dir, count = %d", 
            count);
        return EXIT_FAILURE;
    }

    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
