/* Standard Library Includes */
#include <string.h>
#include <math.h>

/* IAS Library Includes */
#include "ias_const.h" 
#include "ias_logging.h"
#include "ias_math.h"

/* Local Includes */
#include "select_control_gcps.h"

/*****************************************************************************
NAME:     convex_hull.c

PURPOSE:  Will go through each GCP one by one finding the convex hull. The 
          chips will be designated as control chips after the convex hull 
          is found.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed
*****************************************************************************/
void convex_hull_area
(
    const IAS_GCP_RECORD *gcp_hull,/* I: Chip structure */
    const int *hull_points,     /* I: Array of indexes into hull for gcp_hull */
    int num_points,             /* I: Number of points in convex hull */
    double *hull_area           /* O: Area of convex hull */
)   
{
    int iteration;          /* Loop counter */
    double current_x;       /* Current x-coordinate */
    double previous_x;      /* Previous x-coordinate */
    double current_y;       /* Current y-coordinate */
    double previous_y;      /* Previous y-coordinate */
    double origin_y;        /* Y origin */
    double area = 0.0;      /* Current area */
   
    /* Initialize coordinates */
    previous_x = gcp_hull[0].projection_x;
    previous_y = 0.0;
    origin_y = gcp_hull[0].projection_y; 

    /* Determine area */
    for(iteration = num_points - 1; iteration >= 0; iteration--)
    {
        current_x = gcp_hull[hull_points[iteration]].projection_x; 
        current_y = gcp_hull[hull_points[iteration]].projection_y - origin_y;
        area = area + (previous_x - current_x) * (previous_y + current_y);
        previous_x = current_x;
        previous_y = current_y;
    }

    *hull_area = fabs(0.5 * area);
}

int convex_hull
( 
    int num_gcps,               /* I: Number of chips */
    int max_control,            /* I: Maximum number of control points. */
    int min_control,            /* I: Minimum number of control points. */
    IAS_GCP_RECORD *gcp_hull,   /* I/O: Chip structure */
    int *num_control_gcps,      /* O: Number of points designated as control*/
    int *hull_points,           /* O: Array of indexes into hull for gcp_hull */
    double *hull_area           /* O: Area of convex hull */
)        
{
    IAS_VECTOR next_vector;         /* Next point in vector coordinates */
    IAS_VECTOR control_vector;      /* Control point to check for convex hull */
    int next_control;               /* Next control point */ 
    int prev_control;               /* Previous control point */
    int cur_gcp;                    /* Loop counter */
    int iteration;                  /* Loop counter */
    int hull_point_count;           /* Number of control points */
    double next_control_angle;      /* Smallest angle between two vectors */
    double vector_angle;            /* Smallest angle between two vectors */
    double pi = ias_math_get_pi();  /* PI variable */
    double control_magnitude;       /* Magnitude (length) of the control
                                       vector in the loop iteration */
    double dot_next_control;        /* Magnitude between control and next
                                       vectors */
    double next_magnitude;          /* Magnitude (length) of the next vector in
                                       the loop iteration */
    int num_hull_gcps;              /* Number of points in convex hull */
    int leftmost;                   /* Start of convex hull*/

    *hull_area = 0.0;
    
    /* Check if all gcps should be set to control */
    if (num_gcps < min_control)
    {
        for (iteration = 0; iteration < num_gcps; iteration++)
        {
            strcpy(gcp_hull[iteration].chip_type, "CONTROL");
            hull_points[iteration] = iteration;
        }

        *num_control_gcps = num_gcps;

        return SUCCESS;
    }

    /* Determine left most point.  Ties are broken by
       choosing the point with the smallest y coordinate */ 
    for (iteration = 0, cur_gcp = 0; iteration < num_gcps; iteration++)
    {
        if (gcp_hull[iteration].projection_x <=
            gcp_hull[cur_gcp].projection_x)
        {
            if (gcp_hull[iteration].projection_x ==
                gcp_hull[cur_gcp].projection_x)
            {
                if (gcp_hull[iteration].projection_y < 
                    gcp_hull[cur_gcp].projection_y)
                {
                    cur_gcp = iteration;
                }
            }
            else
            {
                cur_gcp = iteration;
            }
        }
    }
    leftmost = cur_gcp;

    /* Set up for next point */
    next_control = leftmost;

    /* Mark left most as a control type */
    strcpy(gcp_hull[next_control].chip_type, "CONTROL");
    next_vector.x = 0.0;
    next_vector.y = 100.0;
    next_vector.z = 0.0;
    hull_points[0] = next_control;
    hull_point_count = 1;
    prev_control = -1;

    /* Determine convex hull */
    while (hull_point_count < max_control)
    {
        vector_angle = 16.0 * pi;

        for (iteration = 0, cur_gcp = -1; iteration < num_gcps; iteration++)
        {
            /* Do not test the point against itself */
            if (iteration == next_control)
                continue;

            /* Do not test the point against the previous control point since
               that will fail calculating the angle */
            if (iteration == prev_control)
                continue;

            control_vector.x = gcp_hull[iteration].projection_x -
                gcp_hull[next_control].projection_x;

            control_vector.y = gcp_hull[iteration].projection_y - 
                gcp_hull[next_control].projection_y;

            control_vector.z = 0.0;

            /* Get the length of the control vector */
            control_magnitude = ias_math_compute_vector_length(&control_vector);
            if (control_magnitude == 0.0)
            {
                IAS_LOG_ERROR("Zero magnitude control vector");
                return ERROR;
            }

            /* Get the length of the next vector */
            next_magnitude = ias_math_compute_vector_length(&next_vector);
            if (next_magnitude == 0.0)
            {
                IAS_LOG_ERROR("Zero magnitude next vector");
                return ERROR;
            }

            /* Compute the dot product for the control and next vector */
            dot_next_control = ias_math_compute_3dvec_dot(&next_vector,
                &control_vector);

            /* Find angle between the 2 vectors, using the following formula:
               A dot B = ||A|| ||B|| Cos angle */
            next_control_angle = 
                acos((dot_next_control) / (control_magnitude * next_magnitude));
            
            if (vector_angle > next_control_angle)
            {
                cur_gcp  = iteration; 
                vector_angle = next_control_angle;
            }
        }

        if (cur_gcp == -1)
        {
            IAS_LOG_ERROR("Error in determining convex hull");
            return ERROR;
        }

        /* Mark point as control */
        if (cur_gcp == leftmost)
        { 
            break;
        }
        else
        {
            hull_points[hull_point_count] = cur_gcp;
        }
        hull_point_count++; 

        next_vector.x = gcp_hull[cur_gcp].projection_x - 
            gcp_hull[next_control].projection_x;

        next_vector.y = gcp_hull[cur_gcp].projection_y - 
            gcp_hull[next_control].projection_y;

        next_vector.z = 0.0;
        prev_control = next_control;
        next_control = cur_gcp;
    }

    num_hull_gcps = hull_point_count;

    for (iteration = 0; iteration < num_hull_gcps; iteration++)
    {
        strcpy(gcp_hull[hull_points[iteration]].chip_type, "CONTROL");
    }

    if (num_hull_gcps == max_control)
    {
        IAS_LOG_ERROR("Maxcontrol not enough to establish convex hull %d,%d",
            num_hull_gcps, max_control);
        return ERROR;
    }
  
    convex_hull_area(gcp_hull, hull_points, num_hull_gcps, hull_area);

    *num_control_gcps = hull_point_count;
    
    return SUCCESS;
}
