/* Standard Library Includes */
#include <string.h>
#include <math.h>

/* IAS Library Includes */
#include "ias_const.h" 
#include "ias_logging.h"
#include "ias_math.h"

/* Local Includes */
#include "select_control_gcps.h"
/*****************************************************************************
NAME:           fill_in

PURPOSE: Given a list of GCPs marked as the hull, continue to mark GCPS as 
         CONTROL until the number of CONTROL GCPS needed has been met. 
         The criteria for the next GCPs to be marked as CONTROL is that 
         has the maximum minimum distance from all GCPs that are marked as 
         CONTROL.

RETURN VALUE:   Type = int
Value       Description
-----       -----------
SUCCESS     Successful completion
ERROR       Operation failed
*****************************************************************************/
int fill_in
(
    int num_gcps,             /* I: Number of chips */
    int num_hull_points,      /* I: Number of points in convex hull */
    int max_control,          /* I: Max number of control points */
    double percentage_control,/* I: Percentage of control points */
    int *hull_points,         /* I/O: Array of indexes into hull for gcp_hull */
    IAS_GCP_RECORD *gcp_hull, /* I/O: Chip structure */
    int *num_control_gcps     /* O: Number of points designated as control */
)     
{
    int next_control;         /* Next control point. */ 
    int local_num_control;    /* Number of control points placeholder */
    
    local_num_control = num_hull_points;
    if (num_gcps > max_control)
    {
        max_control = num_gcps * percentage_control;
    }

    /* Limit the number of control points to the number of GCPs available */
    if (max_control > num_gcps)
        max_control = num_gcps;

    /* If the number of points in the convex hull is less than minimum
       of control points desired find some more. */
    while (local_num_control < max_control)
    {
        int val_point;                  /* Loop var for validation points */
        double point_distance = 0.0;    /* Distance squared from hull */

        /* Checking validation points for possible use as control. */
        for (val_point = 0, next_control = -1 ; val_point < num_gcps;
            val_point++)
        {
            /* Initialize maximum displacement calculated */
            /* Point is validation point check for smallest
               distance between it and the other control
               points. */
            if (strcmp(gcp_hull[val_point].chip_type, "CONTROL") != 0)
            {
                int point;                   /* Point loop variable */
                double max_distance_squared = 1.0e15; /* Max distance squared 
                                                         from hull */

                for (point = 0; point < num_hull_points; point++)
                {
                    double delta_y;       /* Delta latitude */
                    double delta_x;       /* Delta longitude */

                    delta_x = gcp_hull[val_point].projection_x -
                        gcp_hull[hull_points[point]].projection_x;

                    delta_y = gcp_hull[val_point].projection_y - 
                        gcp_hull[hull_points[point]].projection_y;

                    if (max_distance_squared > delta_x * delta_x +
                        delta_y * delta_y)
                    { 
                        max_distance_squared = delta_x * delta_x + 
                            delta_y * delta_y; 
                    }
                }
                /* Check to make sure we found a point. */
                if (max_distance_squared >= 1.0e15 - 1)
                {
                    IAS_LOG_ERROR("Determining controls in convex hull");
                    return ERROR;
                }

                /* Check to see if this validation point is considered a control
                    point up to this point in time. */
                if (point_distance < max_distance_squared)
                {
                    next_control  = val_point;
                    point_distance = max_distance_squared;
                }
            }
        }

        /* Make sure a valid next control point was found */
        if (next_control == -1)
        {
            IAS_LOG_ERROR("No valid control point was found");
            return ERROR;
        }

        /* Mark the validation point with the largest minimum difference between
           it and the other control points as a control point. */
        strcpy(gcp_hull[next_control].chip_type, "CONTROL");

        hull_points[num_hull_points] = next_control;
        num_hull_points++;
        local_num_control++;
    }

    *num_control_gcps = local_num_control;
    return SUCCESS;
}
