/* Standard Library Includes */
#include <stdlib.h>
#include <string.h>

/* IAS Library Includes */         
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "ias_odl.h"

/* Local Includes */
#include "select_control_gcps.h"

/******************************************************************************
NAME: read_omf_parameters

PURPOSE: This function reads all the parameters from the OMF file that 
         select_control_gcps uses.

RETURN VALUE: Type = int
    Value    Description
    -----    -----------
    SUCCESS  The parameters were successfully read
    ERROR    An error occurred reading the parameters
******************************************************************************/
int read_omf_parameters
(
    const char *omf_filename,                   /* I: OMF filename */
    SELECT_CONTROL_GCPS_PARAMETERS *parameters  /* O: Parameters struct */
)
{
    int count;                /* Table parameter count */
    int help_requested;       /* Status placeholder */

    /* Start parameter table definition */
    IAS_PARM_DECLARE_TABLE(omf_table, 2);

    IAS_PARM_TARGET_WRS_PATH(omf_table, IAS_PARM_REQUIRED,
        &parameters->wrs_path, sizeof(parameters->wrs_path), 0);

    IAS_PARM_TARGET_WRS_ROW(omf_table, IAS_PARM_REQUIRED,
        &parameters->wrs_row, sizeof(parameters->wrs_row), 0);

    /* Retrieve parameter count for OMF table */
    count = IAS_PARM_GET_TABLE_SIZE(omf_table);

    /* Call the ias_parm_provide_help routine to provide help for the
       parameters if requested; also verifies the OMF file exists */
    help_requested = ias_parm_provide_help(omf_filename, omf_table, count,
            IAS_OMF_PARAMETERS);
    if (help_requested)
    {
        /* Either help was requested or an error occurred */
        if (help_requested == TRUE)
        {
            /* Successfully provided help, as requested, so just exit */
            exit(EXIT_SUCCESS);
        }
        
        IAS_LOG_ERROR("Error occurred providing OMF help");
        return ERROR;
    }

    /* Read the OMF file */
    if (ias_parm_read(omf_filename, omf_table, count) != SUCCESS)
    {
        IAS_LOG_ERROR("Error reading OMF parameter file %s", omf_filename);

        return ERROR;
    }

    return SUCCESS;
}

/*****************************************************************************
NAME:           get_parameters.c

PURPOSE: Will get the specified parameters from the parameter file and return
         them. It will also provide help and template for the parameters using
         --help and --template.  

RETURN VALUE:   Type = int
Value       Description
-----       -----------
SUCCESS     Successful completion
FAILURE     Operation failed
*****************************************************************************/
int get_parameters
(
    const char *filename,                       /* I: Parameter Filename */
    SELECT_CONTROL_GCPS_PARAMETERS *parameters  /* O: Parameters struct */
)
{
    int help_requested;                     /* Parameter help request */
    int count;                              /* Table parameter count */
    int status;                             /* Status variable */
    char local_omf_filename[PATH_MAX];      /* OMF filename */
    char local_wo_id[IAS_WORKORDER_ID_SIZE];/* Work order ID */

    /* Initialize the parameters structure to zero */
    memset(parameters, 0, sizeof(SELECT_CONTROL_GCPS_PARAMETERS));

    /* Initialize the parameter table */
    IAS_PARM_DECLARE_TABLE(parms, 7);
    
    /* Get the work order ID */
    IAS_PARM_WORK_ORDER_ID(parms, local_wo_id, sizeof(local_wo_id), 1);

    /* Adding the GCPLib file name and path to table */                        
    IAS_PARM_ADD_STRING(parms, GCPLIB_FILENAME, "GCPLib file name and path",   
        IAS_PARM_REQUIRED, IAS_PARM_NOT_ARRAY, NULL, 0, NULL,                  
        parameters->gcplib_filename, sizeof(parameters->gcplib_filename), 0);

    /* Adding the maximum number of controls to the table */
    const int max_default[] = {300};
    IAS_PARM_ADD_INT(parms, MAXIMUM_CONTROL, "Maximum number of points" 
        " that will be selected as control points", IAS_PARM_OPTIONAL, 
        IAS_PARM_NOT_ARRAY, 1, 0, 1000, 1, max_default, 
        &parameters->max_control, sizeof(parameters->max_control), 0);

    /* Adding the minimum number of controls to table */
    const int min_default[] = {200};
    IAS_PARM_ADD_INT(parms, MINIMUM_CONTROL, "Minimum number of points" 
        " that will selected as control points", IAS_PARM_OPTIONAL, 
        IAS_PARM_NOT_ARRAY, 1, 0, 1000, 1, min_default, 
        &parameters->min_control, sizeof(parameters->min_control), 0);

    /* Adding the percentage of control points needed */
    const double per_default[] = {.6};
    IAS_PARM_ADD_DOUBLE(parms, PERCENTAGE_CONTROL, "Percentage of total" 
        " GCPs that will be allowed as controls", IAS_PARM_OPTIONAL,
        IAS_PARM_NOT_ARRAY, 1, 0, 1, 1, per_default, 
        &parameters->percentage_control, sizeof(parameters->percentage_control),
        0);

    /* Adding the GCPLib output file name and path */
    IAS_PARM_ADD_STRING(parms, GCPLIB_OUT_FILENAME, "GCPLib output" 
        " file name and path", IAS_PARM_REQUIRED, IAS_PARM_NOT_ARRAY, NULL, 0,
        NULL, parameters->gcplib_out_filename,
        sizeof(parameters->gcplib_out_filename)*PATH_MAX, 0);

    /* Adding the generate hull file boolean field to table */
    const int default_gen_hull[] = {1};
    IAS_PARM_ADD_INT(parms, GENERATE_HULL_FILE_FLAG, "Flag to control whether " 
        "the hull point file should be generated", IAS_PARM_OPTIONAL, 
        IAS_PARM_NOT_ARRAY, 1, 0, 1, 1, default_gen_hull, 
        &parameters->generate_hull_file, 
        sizeof(parameters->generate_hull_file), 0);

    /* Calculate number of items in the table */
    count = IAS_PARM_GET_TABLE_SIZE(parms);
    
    /* Before read check if file exist along with parameters */
    help_requested = ias_parm_provide_help(filename, parms, count, 
        IAS_INPUT_PARAMETERS); 
    if (help_requested)
    {
        /* Either help was provided or an error occurred */
        if (help_requested == TRUE)
        {
            /* Provide help for the OMF.  If it succeeds, it will exit.  If
               it returns, an error occurred, so return an error */
            read_omf_parameters(filename, parameters);
            IAS_LOG_ERROR("Providing application parameter file help");
            return ERROR;
        }
    }

    /* Reading in the parameters */
    if (ias_parm_read(filename, parms, count) != SUCCESS)
    {
       IAS_LOG_ERROR("Reading parameter file %s.", filename);
       return ERROR;
    }
    
    /* Build the OMF filename */
    status = snprintf(local_omf_filename, sizeof(local_omf_filename),
        "%s.omf", local_wo_id);
    if (status < 0 || status >= sizeof(local_omf_filename))
    {
        IAS_LOG_ERROR("Creating the OMF filename");
        return ERROR;
    }

    /* Read in the OMF parameters */
    if (read_omf_parameters(local_omf_filename, parameters) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from %s", local_omf_filename);
        return ERROR;
    }

    return SUCCESS;
}
