/* Standard Library Includes */
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>

/* IAS Library Includes */
#include "ias_const.h" 
#include "ias_logging.h"
#include "ias_gcp.h"
#include "ias_math.h"

/* Local Includes */
#include "select_control_gcps.h"

/*****************************************************************************
NAME:           SELECT_CONTROL_GCPS

PURPOSE:  Split original GLS GCPs into two sets of image chips.  One set is 
used for control and determining precision corrections that will be applied
to update the geometric mode.  The second set will be used as validation
to determine the validity of the Level-1 Precision/Terrain corrected data 
created from the control chips.

This process works by determining the convex hull associated with a given number
of locations.  The points of the convex hull are the first set of points that
will be designated as control.  Once the convex hull is determined new control
points are added by checking all other points and choosing the point that has 
the maximum minimum distance from all other control points.  Points are added
one at a time in process until the desired number of points at set as control.

The current process involves the following decision making and steps for 
defining control and validation from the GLS chips.

1) The process works on a single WRS set of chips - not from a geographic
   spatial extent, or another way to put this is that chips from adjacent 
   WRS path/rows are not used.

2) There must be a minimum number of control.  If the number of chips for a
   single WRS are below this minimum then no chips are set as validation points
   and all chips are set as control.

3) If the number of GLS chips is above the minimum control choose as control
   a percent of the total number of chips.

4) The percentage of chips that will be set as control, from step 3), cannot
   go above a given maximum number.

5) To ensure that the Moravec points are always used the convex hull is first
   run on the Moravec points.  If there are also evenly spaced points, or
   chips, present then the convex hull is run again on all points.  Note that 
   the convex hull points are always set as control.  

RETURN VALUE:   Type = int

Value           Description
-----           -----------
EXIT_SUCCESS    Successful completion
EXIT_FAILURE    Operation failed
*****************************************************************************/
int main
(
   int   argc,
   char *argv[]
)
{
    SELECT_CONTROL_GCPS_PARAMETERS parameters; /* Parameters Structure */
    IAS_GCP_RECORD *gcp_lib;    /* GCP library info */
    int num_total_gcps;         /* Number of total GCPs */
    int current_num_gcps;       /* Keeps track of current gcps number */
    int num_control_gcps;       /* Number of control points */
    double hull_area;           /* Area of convex hull */
    int num_hull_gcps;          /* Number of gcps in convex hull */
    int iteration;              /* Iteration on convex hull */
    int max_iterations;         /* Maximum iteration on convex hull */
    int status;                 /* Status placeholder */
    int overall_gcp_version;    /* GCPLib overall GCP version */
    int *hull = NULL;           /* Hull pointer */
    int number_even_chips = 0;  /* Number of even chips in GCPLib */
    int number_moravec_chips = 0;/* Number of moravec chips in GCPLib */

    /* Initialize logging library */
    status = ias_log_initialize("select_control_gcps");
    /* Check if logging library was initialized */
    if (status != SUCCESS)
    {
        printf("Initializing logging library \n");
        exit(EXIT_FAILURE);
    }

    /* Check if select_control_gcps was called correctly */
    if (argc < 2)
    {
        IAS_LOG_ERROR("Usage: select_control_gcps <parameter filename>");
        exit(EXIT_FAILURE);
    }

    /* Read in the parameters */    
    status = get_parameters(argv[1], &parameters);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Error reading parameter file %s", argv[1]);
        exit(EXIT_FAILURE);
    }

    /* Read GCPs using the defined criteria */
    status = ias_gcp_read_gcplib(parameters.gcplib_filename, &gcp_lib, 
        &num_total_gcps, &overall_gcp_version);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Error reading GCP library file %s", 
            parameters.gcplib_filename);
        exit(EXIT_FAILURE);
    }

    /* Check to make sure GCPs were found */
    if (num_total_gcps <= 0)
    {
        IAS_LOG_ERROR("No GCPs matched criteria");
        free(gcp_lib);
        exit(EXIT_FAILURE);
    }

    /* Allocate memory for hull */          
    hull = malloc(num_total_gcps * sizeof(int));  
    if (!hull)
    {
        IAS_LOG_ERROR("Allocating hull storage");
        free(gcp_lib);
        exit(EXIT_FAILURE);
    }

    /* Determine how many Moravec and Evenly Spaced Chips are present and
       set all chips to VALIDATION */ 
    for (iteration = 0; iteration < num_total_gcps; iteration++)
    {
        strcpy(gcp_lib[iteration].chip_type, "VALIDATION");
        if (gcp_lib[iteration].moravec_rank > 0)
        {
            number_moravec_chips++;
        }
        else
        {
            number_even_chips++;
        }
    } 

    /* Run convex hull.  This could be done twice, once using only the
       moravec points and once with all the CHIPS. */
    max_iterations = 2;
    if (number_even_chips == 0)
    {   
        max_iterations = 1;
    }
    
    iteration = 0;
    while (iteration < max_iterations)
    {
        /* If running twice so need to set number of control based on
           number of moravec.  Due to how the CHIPS are sorted the
           Moravec points should the the first set of CHIPS in the 
           list. */
        if (iteration == 0)
        {
            current_num_gcps = number_moravec_chips;
        }
        else
        {
            current_num_gcps = num_total_gcps;
        }

        status = convex_hull(current_num_gcps, parameters.max_control,
            parameters.min_control, gcp_lib, &num_hull_gcps, hull, &hull_area); 
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Error determining convex hull for control");
            free(hull);
            free(gcp_lib);
            exit(EXIT_FAILURE);
        }
        
        /* Print out hull area information */
        IAS_LOG_INFO("Hull area = %f for iteration: %d ", hull_area, 
            iteration);

        iteration++;
    }

    /* Print out hull file */
    if (parameters.generate_hull_file)
    {
        FILE *hull_file;                /* HULL file pointer */
        char hull_filename[PATH_MAX];   /* HULL filename */
        
        /* Build the hull filename */
        status = sprintf(hull_filename,"h%03d%03d.hull.dat", 
            gcp_lib[0].point_wrs_path, gcp_lib[0].point_wrs_row);
        if (status < 0 || status > sizeof(hull_filename))
        {
            IAS_LOG_ERROR("Creating hull filename %s", hull_filename);
            free(hull);
            free(gcp_lib);
            exit(EXIT_FAILURE);
        }
        
        hull_file = fopen(hull_filename, "w");
        if (!hull_file)
        {
            IAS_LOG_ERROR("Opening output hull file %s", hull_filename);
            free(hull);
            free(gcp_lib);
            exit(EXIT_FAILURE);
        }

        /* Write out hull to file */
        for (iteration = 0; iteration < num_hull_gcps; iteration++)
        {
            status = fprintf(hull_file,"%s %d %f %f\n",
                gcp_lib[hull[iteration]].chip_name, iteration, 
                gcp_lib[hull[iteration]].projection_y, 
                gcp_lib[hull[iteration]].projection_x);
            if (status < 0)
            {
                IAS_LOG_ERROR("Could not print to hull file %s", hull_filename);
                free(hull);
                free(gcp_lib);
                fclose(hull_file);
                exit(EXIT_FAILURE);
            }
        }

        if (fclose(hull_file) != SUCCESS)
        {
            IAS_LOG_ERROR("Could not close hull file %s", hull_filename);
            free(hull);
            free(gcp_lib);
            exit(EXIT_FAILURE);
        }
    }

    /* Find points inside of convex hull. */
    status = fill_in(current_num_gcps, num_hull_gcps, parameters.max_control, 
        parameters.percentage_control, hull, gcp_lib, &num_control_gcps); 

    /* Free hull which is no longer used */
    free(hull);
    hull = NULL;

    /* Check if fill_in finished without errors */
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Error determining fill in of control GCPs");
        free(gcp_lib);
        exit(EXIT_FAILURE);
    }

    status = ias_gcp_write_gcplib(parameters.gcplib_out_filename, gcp_lib, 
        num_total_gcps, overall_gcp_version);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing out GCPLib to %s", parameters.gcplib_filename);
        free(gcp_lib);
        exit(EXIT_FAILURE);
    }

    /* Free up allocated memory */
    free(gcp_lib);

    exit(EXIT_SUCCESS);
}
