package sensor_alignment;

/**
 * Run Sensor Alignment from the command line.  This replaces the various
 * GUI choices with command line selections.  There is no facility for 
 * selecting and deselecting rows or for selecting the report file name as 
 * there is in the Sensor Alignment GUI
 */
import java.io.IOException;
import java.util.Date;
import java.util.ArrayList;
import java.text.SimpleDateFormat;
import java.text.ParseException;
import java.io.File;

public class CommandLine
{
    /**
     * Constructor
     */ 
    public CommandLine()
    {
    }

    /**
     * Run Sensor Alignment in command-line, rather than GUI, mode.  Perform
     * the same mathematical and database operations that the GUI does, but
     * select the options through command line choices rather than a GUI.
     * This helps with automated test of DB and math routines, scripting,
     * and manual testing.
     */ 
    public static void runApplication
    (
        String queryStartDateStr, // I: Start date of user query in string 
                                  //    format
        String queryEndDateStr,   // I: End date of user query in string format
        String controlTypeStr,    // I: User-selected control type in string 
                                  //    format
        String offNadirAngleStr,  // I: User-selected off-nadir angle in string
                                  //    format
        String offNadirRangeTypeStr, // I: User-selected range type in string 
                                  //       format
        boolean queryStartDateSelected, // I: Flag for query start date choice 
        boolean queryEndDateSelected, // I: Flag for query end date choice
        String cpfStartDateStr,   // I: Start date of user CPF date
        String cpfEndDateStr,     // I: End date of user CPF date
        String saveToDatabaseFlagStr, // I: Flag for saving data to the DB
        String controlTypeWeightStr // I: User-selected weight for DOQ
    ) throws Exception
    {
        ArrayList<PrecisionAlignment> alignRecords; // Alignment records from DB
        Date queryStartDate = null; // Query start date in Date format
        Date queryEndDate = null; // Query end date in Date format
        Date cpfStartDate;        // CPF start date in Date format
        Date cpfEndDate;          // CPF end date in Date format
        double offNadirAngle;     // Off-nadir angle in double format
        boolean saveToDatabaseFlag; // Database save flag
        int controlTypeWeight;    // DOQ weight (out of 100%)
        double[][] alignObs;      // Integrated alignment estimates
        String reportFilename;    // Name of the main sensor alignment report
        File reportFile;          // Report file

        // Set up date format to parse command line date pattern
        SimpleDateFormat cmdSdf = new SimpleDateFormat("dd-MMM-yyyy-HH:mm:ss");

        // Convert command line dates to Date format
        try
        {
            if (queryStartDateSelected)
            {
                queryStartDate = cmdSdf.parse(queryStartDateStr);
            }
            if (queryEndDateSelected)
            {
                queryEndDate = cmdSdf.parse(queryEndDateStr);
            }
            cpfStartDate = cmdSdf.parse(cpfStartDateStr);
            cpfEndDate = cmdSdf.parse(cpfEndDateStr);
        }
        catch(ParseException pe)
        {
            System.out.println("Parse Exception: " + pe);
            throw new IllegalArgumentException("Invalid command line option");
        }

        // Convert numeric command line strings to numeric formats
        try
        {
            offNadirAngle = Double.parseDouble(offNadirAngleStr);
            if (saveToDatabaseFlagStr.equals("1"))
            {
                saveToDatabaseFlag = true; 
            }
            else
            {
                saveToDatabaseFlag = false; 
            }
            controlTypeWeight = Integer.parseInt(controlTypeWeightStr);
        }
        catch(NumberFormatException ne)
        {
            System.out.println("Number Format Exception: " + ne);
            throw new IllegalArgumentException("Invalid command line option");
        }

        // Validate the command line values
        if ((offNadirAngle < SensorAlignment.offNadirAngleMin) || 
            (offNadirAngle > SensorAlignment.offNadirAngleMax))
        {
            throw new IllegalArgumentException("Invalid command line option; "
                + "offNadirAngle \nshould be >= "
                + SensorAlignment.offNadirAngleMin + " and <= "
                + SensorAlignment.offNadirAngleMax);
        }

        // The weight is a percentage out of 100
        if ((controlTypeWeight < 0) || (controlTypeWeight > 100))
        {
            throw new IllegalArgumentException("Invalid command line option; "
                + "controlTypeWeight should be from 0 to 100");
        }

        // The flag for saving to the database should be 0 or 1 
        if ((saveToDatabaseFlag != true) && (saveToDatabaseFlag != false))
        {
            throw new IllegalArgumentException("Invalid command line option; "
                + "saveToDatabaseFlag should be 0 or 1");
        }

        if (!(offNadirRangeTypeStr.equals("INWARD")) &&
            !(offNadirRangeTypeStr.equals("OUTWARD")))
        {
            throw new IllegalArgumentException("Invalid command line option; "
                + "offNadirType should be INWARD or OUTWARD");
        }


        if (!(controlTypeStr.equals("DOQ")) &&
            !(controlTypeStr.equals("GLS")) &&
            !(controlTypeStr.equals("BOTH")))
        {
            throw new IllegalArgumentException("Invalid command line option; "
                + "controlType should be DOQ, GLS, or BOTH");
        }

        // Read the precision alignment records based on the query choices
        try 
        {
            alignRecords = PrecisionAlignmentData.getAlignmentRecords(
                queryStartDate,
                queryEndDate,
                queryStartDateSelected,
                queryEndDateSelected,
                offNadirAngle,
                offNadirRangeTypeStr,
                controlTypeStr);
        } 
        catch (Exception ex) 
        {
            throw new Exception(ex.getMessage());
        }

        // Check that results were returned
        if (alignRecords.size() < 1)
        {
            throw new Exception("No precision alignment records retrieved");
        }

        // Initialize a new result set 
        Results res = new Results(alignRecords.size());

        // Fill in some of the result set elements based on the command line 
        res.doqWeight = controlTypeWeight;
        res.glsWeight = 100 - controlTypeWeight;
        res.gcpSource = controlTypeStr;
        res.acqQueryStartDateSelected = queryStartDateSelected;
        if (queryStartDateSelected)
        {
            res.acqQueryStartDate = queryStartDate;
        }
        res.acqQueryEndDateSelected = queryEndDateSelected;
        if (queryEndDateSelected)
        {
            res.acqQueryEndDate = queryEndDate;
        }
        res.effectiveStartDate = cpfStartDate;
        res.effectiveEndDate = cpfEndDate;

        // Create integrated alignment observations
        alignObs = new double[alignRecords.size()][3];

        // Process the alignment records
        for (int row = 0; row < alignRecords.size(); row++)
        {
            // Compute Roll', Pitch', and Yaw'
            alignObs[row] = alignRecords.get(row).processSingleDataSet();

            // Process a single alignment record, initializing result set,
            // summing weights, and finding the maximum roll angle
            res.processRow(alignRecords.get(row), alignObs[row], row);
        }

        // Compute best-fit alignment, residuals, and RMS statistics
        res.processMeasurements(alignRecords.size());

        // Calculate the updated alignment matrix
        res.calculateAlignmentMatrix();

        // Write the report files
        reportFilename = "sensor_alignment.rpt";
        reportFile = new File(reportFilename);
        try 
        {
            reportFile.createNewFile();
        } 
        catch (IOException ex) 
        {
            throw new Exception("Could not create report file: "
                + ex.getMessage());
        }

        try 
        {
            res.writeSensorAlignmentReports(reportFile);
        } 
        catch (IOException ex) 
        {
            throw new Exception("Could not save results to the report files: "
                + ex.getMessage());
        }

        // Send the results to the database 
        if (saveToDatabaseFlag)
        {
            try 
            {
                res.saveToDatabase();
            } 
            catch (Exception ex) 
            {
                throw new Exception("Could not save results to the database: "
                    + ex.getMessage());
            }
        }
    }
}
