package sensor_alignment;

/**
 * Organize the data read from the Precision Alignment database table.
 * Process a single precision alignment data set from the database.  Combine 
 * the attitude and ephemeris corrections derived by the precision correction
 * process to form a set of integrated alignment estimates roll', pitch', and
 * yaw' for the database
 */
import java.text.ParseException;
import java.util.Date;
import java.text.SimpleDateFormat;
import java.sql.ResultSet;
import java.sql.SQLException;

public class PrecisionAlignment
{
    public Date dateAcquired; // Observation date, including time
    public double[] attitudeBias; // Attitude/alignment corrections: roll, 
                                  // pitch, yaw
    public double[] ephemerisBias; // Ephemeris corrections: X, Y, Z
    public double[][] attitudeAttitudeCovariance; // Attitude-attitude 
                       // covariance matrix:
                       //  roll-roll  pitch-roll  yaw-roll
                       //  roll-pitch pitch-pitch yaw-pitch
                       //  roll-yaw   pitch-yaw   yaw-yaw
    public double[][] attitudeEphemerisCovariance; // attitude-ephemeris 
                       // covariance matrix:
                       //  roll-X pitch-X yaw-X
                       //  roll-Y pitch-Y yaw-Y
                       //  roll-Z pitch-Z yaw-Z
    private double[][] ephemerisAttitudeCovariance; // ephemeris-attitude 
                       // covariance matrix:
                       //  X-roll  Y-roll  Z-roll
                       //  X-pitch Y-pitch Z-pitch
                       //  X-yaw   Y-yaw   Z-yaw
    public double[][] ephemerisEphemerisCovariance; // ephemeris-ephemeris 
                       // covariance matrix:
                       //  X-X  Y-X  Z-X
                       //  X-Y  Y-Y  Z-Y
                       //  X-Z  Y-Z  Z-Z
    public double gcpRmsFit; // Combined along and across-track RMS fit
    public int numGcps; // Number of GCPs in fit
    public double outlierThreshold; // Outlier threshold used for fit
    public String gcpSource; // GCP source/control type (DOQ or GLS)
    public double offNadirAngle; // Off-nadir angle in degrees
    public int wrsPath; // WRS path for scene
    public int wrsRow; // WRS row for scene
    public String workOrderId; // Work order ID
    public String geometricCharId; // Geometric characterization ID

    // Constructor
    public PrecisionAlignment
    (
        ResultSet rs // I: Set of results from the database
    ) throws SQLException, ParseException
    {
        attitudeBias = new double[3];
        ephemerisBias = new double[3];
        attitudeAttitudeCovariance = new double[3][3];
        attitudeEphemerisCovariance = new double[3][3];
        ephemerisAttitudeCovariance = new double[3][3];
        ephemerisEphemerisCovariance = new double[3][3];

        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");

        workOrderId = rs.getString("WORK_ORDER_ID");
        wrsPath = rs.getInt("TARGET_WRS_PATH");
        wrsRow = rs.getInt("TARGET_WRS_ROW");
        geometricCharId = rs.getString("GEOMETRIC_WO_COMMON_ID");
        gcpSource = rs.getString("GCP_SOURCE");
        offNadirAngle = rs.getDouble("OFF_NADIR_ANGLE");
        numGcps = rs.getInt("NBR_GCPS_USED");
        outlierThreshold = rs.getDouble("OUTLIER_THRESHOLD");
        gcpRmsFit = rs.getDouble("RMS_GCP_FIT");
        attitudeBias[0] = rs.getDouble("ALIGN_ANGLE_ROLL");
        attitudeBias[1] = rs.getDouble("ALIGN_ANGLE_PITCH");
        attitudeBias[2] = rs.getDouble("ALIGN_ANGLE_YAW");
        ephemerisBias[0] = rs.getDouble("POS_BIAS_X");
        ephemerisBias[1] = rs.getDouble("POS_BIAS_Y");
        ephemerisBias[2] = rs.getDouble("POS_BIAS_Z");
        attitudeAttitudeCovariance[0][0] = rs.getDouble("ROLL_ROLL_COV");
        attitudeAttitudeCovariance[0][1] = rs.getDouble("ROLL_PITCH_COV");
        attitudeAttitudeCovariance[0][2] = rs.getDouble("ROLL_YAW_COV");
        attitudeEphemerisCovariance[0][0] = rs.getDouble("ROLL_XPOS_COV");
        attitudeEphemerisCovariance[0][1] = rs.getDouble("ROLL_YPOS_COV");
        attitudeEphemerisCovariance[0][2] = rs.getDouble("ROLL_ZPOS_COV");
        attitudeAttitudeCovariance[1][0] = rs.getDouble("PITCH_ROLL_COV");
        attitudeAttitudeCovariance[1][1] = rs.getDouble("PITCH_PITCH_COV");
        attitudeAttitudeCovariance[1][2] = rs.getDouble("PITCH_YAW_COV");
        attitudeEphemerisCovariance[1][0] = rs.getDouble("PITCH_XPOS_COV");
        attitudeEphemerisCovariance[1][1] = rs.getDouble("PITCH_YPOS_COV");
        attitudeEphemerisCovariance[1][2] = rs.getDouble("PITCH_ZPOS_COV");
        attitudeAttitudeCovariance[2][0] = rs.getDouble("YAW_ROLL_COV");
        attitudeAttitudeCovariance[2][1] = rs.getDouble("YAW_PITCH_COV");
        attitudeAttitudeCovariance[2][2] = rs.getDouble("YAW_YAW_COV");
        attitudeEphemerisCovariance[2][0] = rs.getDouble("YAW_XPOS_COV");
        attitudeEphemerisCovariance[2][1] = rs.getDouble("YAW_YPOS_COV");
        attitudeEphemerisCovariance[2][2] = rs.getDouble("YAW_ZPOS_COV");
        ephemerisAttitudeCovariance[0][0] = rs.getDouble("XPOS_ROLL_COV");
        ephemerisAttitudeCovariance[0][1] = rs.getDouble("XPOS_PITCH_COV");
        ephemerisAttitudeCovariance[0][2] = rs.getDouble("XPOS_YAW_COV");
        ephemerisEphemerisCovariance[0][0] = rs.getDouble("XPOS_XPOS_COV");
        ephemerisEphemerisCovariance[0][1] = rs.getDouble("XPOS_YPOS_COV");
        ephemerisEphemerisCovariance[0][2] = rs.getDouble("XPOS_ZPOS_COV");
        ephemerisAttitudeCovariance[1][0] = rs.getDouble("YPOS_ROLL_COV");
        ephemerisAttitudeCovariance[1][1] = rs.getDouble("YPOS_PITCH_COV");
        ephemerisAttitudeCovariance[1][2] = rs.getDouble("YPOS_YAW_COV");
        ephemerisEphemerisCovariance[1][0] = rs.getDouble("YPOS_XPOS_COV");
        ephemerisEphemerisCovariance[1][1] = rs.getDouble("YPOS_YPOS_COV");
        ephemerisEphemerisCovariance[1][2] = rs.getDouble("YPOS_ZPOS_COV");
        ephemerisAttitudeCovariance[2][0] = rs.getDouble("ZPOS_ROLL_COV");
        ephemerisAttitudeCovariance[2][1] = rs.getDouble("ZPOS_PITCH_COV");
        ephemerisAttitudeCovariance[2][2] = rs.getDouble("ZPOS_YAW_COV");
        ephemerisEphemerisCovariance[2][0] = rs.getDouble("ZPOS_XPOS_COV");
        ephemerisEphemerisCovariance[2][1] = rs.getDouble("ZPOS_YPOS_COV");
        ephemerisEphemerisCovariance[2][2] = rs.getDouble("ZPOS_ZPOS_COV");

        dateAcquired = sdf.parse(rs.getString("DATE_ACQUIRED"));
    }

    public double[] processSingleDataSet
    (
    )
    {
        double[][] ephEphCovInverted = new double[3][3]; // Temporary matrix
        double[] y = new double[3]; // Deweighted ephemeris bias 
        double[] yPrime = new double[3]; // Temporary vector
        double[] alignObs = new double[3]; // Integrated alignment estimates

        // Invert the ephemeris covariance
        ephEphCovInverted = invert3x3Matrix(ephemerisEphemerisCovariance);

        // Multiply inverted ephemeris-ephemeris and ephemeris bias matrices
        y = multiply3x3And3x1Matrix(ephEphCovInverted, ephemerisBias);
            
        // Multiply attEph covariance by deweighted ephemeris bias 
        yPrime = multiply3x3And3x1Matrix(attitudeEphemerisCovariance, y);

        // Blend attitude and ephemeris corrections to form integrated 
        // alignment using matrix subtraction 
        for (int rr = 0; rr < 3; rr++)
        {
            alignObs[rr] = attitudeBias[rr] - yPrime[rr];
        }

        return alignObs;
    }

    //       _     _              _     _
    //      | a b c |            | A B C |
    //  X=  | d e f |   (X)^-1 = | D E F | * 1 / Z
    //      | g h i |            | G H I |
    //       -     -              -     -
    //
    // This is derived from ias_math_invert_3x3_matrix.c
    //
    private double[][] invert3x3Matrix
    (
        double inMatrix[][] // I: Input matrix
    )
    {
        double[][] outMatrix = new double[3][3]; // Output inverted matrix

        double A; // Temp variables for calculations - these are straight from
        double B; // ias_math_invert_3x3_matrix.c
        double C;
        double D;
        double E;
        double F;
        double G;
        double H;
        double I;
        double Z;

        A = inMatrix[1][1] * inMatrix[2][2] - inMatrix[1][2] * inMatrix[2][1];
        B = inMatrix[0][2] * inMatrix[2][1] - inMatrix[0][1] * inMatrix[2][2]; 
        C = inMatrix[0][1] * inMatrix[1][2] - inMatrix[0][2] * inMatrix[1][1]; 
        D = inMatrix[1][2] * inMatrix[2][0] - inMatrix[1][0] * inMatrix[2][2]; 
        E = inMatrix[0][0] * inMatrix[2][2] - inMatrix[0][2] * inMatrix[2][0]; 
        F = inMatrix[0][2] * inMatrix[1][0] - inMatrix[0][0] * inMatrix[1][2]; 
        G = inMatrix[1][0] * inMatrix[2][1] - inMatrix[1][1] * inMatrix[2][0]; 
        H = inMatrix[0][1] * inMatrix[2][0] - inMatrix[0][0] * inMatrix[2][1]; 
        I = inMatrix[0][0] * inMatrix[1][1] - inMatrix[0][1] * inMatrix[1][0]; 
        Z = inMatrix[0][0] * A + inMatrix[0][1] * D +inMatrix[0][2] * G; 

        if (Z <= 1e-9)     /* if Z is zero */
        {
            throw new IllegalArgumentException("Error: calculations will " + 
                "result in divide by zero.");
        }

        outMatrix[0][0] = A / Z;
        outMatrix[0][1] = B / Z;
        outMatrix[0][2] = C / Z;

        outMatrix[1][0] = D / Z;
        outMatrix[1][1] = E / Z;
        outMatrix[1][2] = F / Z;

        outMatrix[2][0] = G / Z;
        outMatrix[2][1] = H / Z;
        outMatrix[2][2] = I / Z;

        return outMatrix;
    }

    //
    // This is derived from ias_math_multiply_matrix.c.  However, it is
    // simplified because the matrices are always 3x3 and 3x1
    //
    private double[] multiply3x3And3x1Matrix
    (
        double matrix1[][], // I: Input matrix one (3x3)
        double matrix2[] // I: Input matrix two (3x1)
    )
    {
        double sum; // Summary of multiplied items from old matrices to
                    // form item in the new matrix
        double[] outMatrix = new double[3]; // Output matrix 

        for (int rr = 0; rr < 3; rr++)
        {
            sum = 0.0;
            for (int cc = 0; cc < 3; cc++)
            {
                sum += matrix1[rr][cc] * matrix2[cc];
            }
            outMatrix[rr] = sum;
        }

        return outMatrix;
    }
}
