package sensor_alignment;

/**
 * Get requested precision alignment data from the database
 */
import java.text.*;
import java.util.Date;
import java.util.ArrayList;

// Set up access to the IAS database
import gov.usgs.ias.constant.Constants;
import gov.usgs.ias.dao.DataConnection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.ParseException;

public class PrecisionAlignmentData
{        
    /**
     * Constructor
     */
    public PrecisionAlignmentData()
    {
    }
    
    /**
     * Get the precision alignment records from the database
     * @return  alignment records containing the retrieved data
     */
    public static ArrayList<PrecisionAlignment> getAlignmentRecords
    (
        Date queryStartDate, // I: User-selected query start date
        Date queryEndDate, // I: User-selected query end date
        boolean startDateSelected, // I: Flag for whether start date was picked
        boolean endDateSelected, // I: Flag for whether end date was picked
        double offNadirAngle, // I: User-selected maximum off-nadir angle 
                           // magnitude
        String offNadirRangeType, // I: User-selected range type
        String controlType // I: User-selected control type
    ) throws SQLException, ParseException
    {
        String startDateSelectString; // Part of query for start date
        String endDateSelectString; // Part of query for end date
        String controlTypeSelectString; // Part of query for control type
        String offNadirRangeSelectString; // Part of query for range type

        // Holds precision alignment records retrieved 
        ArrayList<PrecisionAlignment> alignRecords = 
            new ArrayList<PrecisionAlignment>();
        
        // Set up a date format that the database query can use.
        SimpleDateFormat querySdf = 
            new SimpleDateFormat("dd/MMM/yyyy:HH:mm:ss");        
        
        // Set up query strings for optional date filters
        if (startDateSelected)
        {
            startDateSelectString = " and DATE_ACQUIRED >= TO_DATE('" 
                + querySdf.format(queryStartDate) 
                + "', 'DD/MON/YYYY:HH24:MI:SS')";
        }
        else
        {
            startDateSelectString = "";
        }
        
        if (endDateSelected)
        {
            endDateSelectString = " and DATE_ACQUIRED <= TO_DATE('" 
                + querySdf.format(queryEndDate) 
                + "', 'DD/MON/YYYY:HH24:MI:SS')";
        }
        else
        {
            endDateSelectString = "";
        }

        // Set up query strings for control type filter
        if (controlType.equals("BOTH"))
        {
            controlTypeSelectString = "";
        }
        else
        {
            controlTypeSelectString = " and GCP_SOURCE = '" + controlType
                + "' ";
        }

        // Set up query strings for off-nadir angle type filter
        if (offNadirRangeType.equals("INWARD"))
        {
            offNadirRangeSelectString = "and OFF_NADIR_ANGLE <= " 
                + offNadirAngle + " " + "and OFF_NADIR_ANGLE >= " 
                + -offNadirAngle + " ";
        }
        else // offNadirRangeType is "OUTWARD"
        {
            offNadirRangeSelectString = "and (OFF_NADIR_ANGLE >= " 
                + offNadirAngle + " " + "or OFF_NADIR_ANGLE <= " 
                + -offNadirAngle + ") ";
        }

        // Set up the database select query
        String sqlCmd = "select PRECISION_MODEL_CORRECTION."
            + "GEOMETRIC_WO_COMMON_ID, WORK_ORDER_ID, TARGET_WRS_PATH, "
            + "TARGET_WRS_ROW, OFF_NADIR_ANGLE, GCP_SOURCE, NBR_GCPS_USED, "
            + "OUTLIER_THRESHOLD, RMS_GCP_FIT, DATE_ACQUIRED, "
            + "ALIGN_ANGLE_ROLL, ALIGN_ANGLE_PITCH, ALIGN_ANGLE_YAW, "
            + "POS_BIAS_X, POS_BIAS_Y, POS_BIAS_Z, ROLL_ROLL_COV, "
            + "ROLL_PITCH_COV, ROLL_YAW_COV, ROLL_XPOS_COV, ROLL_YPOS_COV, "
            + "ROLL_ZPOS_COV, PITCH_ROLL_COV, PITCH_PITCH_COV, "
            + "PITCH_YAW_COV, PITCH_XPOS_COV, PITCH_YPOS_COV, PITCH_ZPOS_COV, "
            + "YAW_ROLL_COV, YAW_PITCH_COV, YAW_YAW_COV, YAW_XPOS_COV, "
            + "YAW_YPOS_COV, YAW_ZPOS_COV, XPOS_ROLL_COV, XPOS_PITCH_COV, "
            + "XPOS_YAW_COV, XPOS_XPOS_COV, XPOS_YPOS_COV, XPOS_ZPOS_COV, "
            + "YPOS_ROLL_COV, YPOS_PITCH_COV, YPOS_YAW_COV, YPOS_XPOS_COV, "
            + "YPOS_YPOS_COV, YPOS_ZPOS_COV, ZPOS_ROLL_COV, ZPOS_PITCH_COV, "
            + "ZPOS_YAW_COV, ZPOS_XPOS_COV, ZPOS_YPOS_COV, ZPOS_ZPOS_COV from "
            + "PRECISION_MODEL_CORRECTION, GEOMETRIC_WORK_ORDER_COMMON "
            + "where PRECISION_MODEL_CORRECTION.GEOMETRIC_WO_COMMON_ID = "
            + "GEOMETRIC_WORK_ORDER_COMMON.GEOMETRIC_WO_COMMON_ID and "
			+ "GEOMETRIC_WORK_ORDER_COMMON.VALID = 'Y' "
            + offNadirRangeSelectString 
            + startDateSelectString + endDateSelectString
            + controlTypeSelectString + " order by DATE_ACQUIRED";

        String connectionProperty = "sensor_alignment.db.conn";
        String authProperty = Constants.DAO_AUTH_IAS_PROD;

        DataConnection dataConnection = 
            new DataConnection(connectionProperty);
        PreparedStatement stmt = null;
        ResultSet rs = null;
        try {
            stmt = dataConnection.getConnection(authProperty).
                prepareStatement(sqlCmd);
            rs = stmt.executeQuery();

            while (rs.next()) {
                // Initialize new Alignment record with values from DB
                alignRecords.add(new PrecisionAlignment(rs));
            }
        } catch (SQLException sqlException) {
            throw sqlException;
        } catch (ParseException parseException) {
            throw parseException;
        } finally {
            try {
                if (rs != null) {
                    rs.close();
                }
                if (stmt != null) {
                    stmt.close();
                }
            } catch (SQLException ex) {
                System.out.println("Warning: " + ex.getMessage());
            }
            dataConnection.close();
        }

        if (alignRecords.isEmpty())
        {
            throw new SQLException("Warning: no rows retrieved from "
                + "database");
        }
        return alignRecords;
    }
}
