package sensor_alignment;

/**
 * Results contains the data to be sent to the report files and/or the 
 * database.  It performs various computations needed for all database and 
 * report outputs.  These computations are done after precision alignment rows 
 * to process have been selected.  It also writes the results to report files
 * and the database
 */
import java.io.File;
import java.io.PrintWriter;
import java.io.IOException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Formatter;

// Set up access to the IAS database
import gov.usgs.ias.dao.DataConnection;
import java.sql.PreparedStatement;
import java.sql.SQLException;

// Set up access to the IAS version number
import gov.usgs.ias.constant.Constants;

// A Results object should have all of the data needed for results screen
// display, reporting, and database saving when it is filled in and sent
// to the Results Dialog class
public class Results
{
    // Alignment matrix
    public double[][] alignmentMatrix = {{0, 0, 0},
                                         {0, 0, 0}, 
                                         {0, 0, 0}};

    // The following arrays each holds an entry for each selected row
    private String[] woIds; // Work Order IDs 
    private String[] geoCharIds; // Geometric Characterization IDs 
    private int[] paths; // WRS Paths
    private int[] rows; // WRS Rows
    private String[] acqDates; // Acquisition Date
    private String[] acqTimes; // Acquisition Time
    private String[] ctrlTypes; // Control Types (DOQ or GLS)
    private double[] gcpRmsFits; // GCP fit
    public double[] rollPrimes; // Roll'
    public double[] pitchPrimes; // Pitch'
    public double[] yawPrimes; // Yaw'
    private double[] offNadirAngles; // Off-nadir angles

    // Alignment estimates
    public double rollAlignmentEstimate = 0.0;
    public double pitchAlignmentEstimate = 0.0;
    public double yawAlignmentEstimate = 0.0;

    // RMS residuals
    public double rollRms = 0.0;
    public double pitchRms = 0.0;
    public double yawRms = 0.0;

    // User selections 
    public Date acqQueryStartDate; // Acquisition date query start
    public Date acqQueryEndDate; // Acquisition date query end
    public boolean acqQueryStartDateSelected; // Was acquisition start selected?
    public boolean acqQueryEndDateSelected; // Was acq date end selected?
    public Date effectiveStartDate; // Calibration effective date range start
    public Date effectiveEndDate; // Calibration effective date range end 
    public int doqWeight; // DOQ weight used
    public int glsWeight; // GLS weight used
    public String gcpSource; // GCP Source used (DOQ, GLS, or BOTH)
 
    // Miscellaneous
    private String dateOut; // Report file creation date
    private String timeOut; // Report file creation time
    private double maxRoll = 0; // Maximum roll angle within results
        
    // Weighted sums
    private double weightedRollsSum = 0.0; // Roll primes weighted
    private double weightedPitchesSum = 0.0; // Pitch primes weighted
    private double weightedYawsSum = 0.0; // Yaw primes weighted
    private double weightSum = 0.0; // Sum of weights

    /**
     * Constructor for the dialog
     * @param numRows    number of rows of data selected by user to process
     */
    public Results(int numRows)
    {
        // Initialize arrays to be the size of the number of selected rows
        woIds          = new String[numRows];
        geoCharIds     = new String[numRows];
        paths          = new int[numRows];
        rows           = new int[numRows];
        acqDates       = new String[numRows];
        acqTimes       = new String[numRows];
        ctrlTypes      = new String[numRows];
        gcpRmsFits     = new double[numRows];
        rollPrimes     = new double[numRows];
        pitchPrimes    = new double[numRows];
        yawPrimes      = new double[numRows];
        offNadirAngles = new double[numRows];
    }

    /**
     * Process a single row, initializing result set, summing weights, and
     * finding the maximum roll angle 
     */
    public void processRow
    (
        PrecisionAlignment alignRecord, // I: Precision alignment record
        double alignObs[], // I: Alignment observations
        int row // I: Row of results being processed
    )
    {
        double weight; // Weight percent 

        // Set up date formats to extract the date acquired results and
        // format them for the results set
        SimpleDateFormat dateSdf = new SimpleDateFormat("yyyy-DDD");
        SimpleDateFormat timeSdf = new SimpleDateFormat("HH:mm:ss");

        // Fill in various other parts of the result set
        woIds[row] = alignRecord.workOrderId;
        geoCharIds[row] = alignRecord.geometricCharId;
        ctrlTypes[row] = alignRecord.gcpSource;
        paths[row] = alignRecord.wrsPath;
        rows[row] = alignRecord.wrsRow;
        gcpRmsFits[row] = alignRecord.gcpRmsFit;
        offNadirAngles[row] = alignRecord.offNadirAngle;

        // Fill in the date and time for the result set
        acqDates[row] = dateSdf.format(alignRecord.dateAcquired);
        acqTimes[row] = timeSdf.format(alignRecord.dateAcquired);

        // Fill in the Roll', Pitch', and Yaw' parts of the result set
        rollPrimes[row] = alignObs[0];
        pitchPrimes[row] = alignObs[1];
        yawPrimes[row] = alignObs[2];

        // Add the row's contribution to weighted averages
        if (ctrlTypes[row].equals("DOQ"))
        {
            weight = (double)doqWeight / 100.0;
        }
        else
        {
            weight = (double)glsWeight / 100.0;
        }

        weightedRollsSum += weight * rollPrimes[row];
        weightedPitchesSum += weight * pitchPrimes[row];
        weightedYawsSum += weight * yawPrimes[row];
        weightSum += weight;

        // Calculate the maximum roll angle for the result set
        if (Math.abs(offNadirAngles[row]) > maxRoll)
        {
            maxRoll = Math.abs(offNadirAngles[row]);
        }
    }

    /**
     * Compute best-fit alignment, residuals, and RMS statistics.  This method
     * assumes that ProcessRow has been called for each of the selectedCount
     * results 
     */
    public void processMeasurements
    (
        int selectedCount // I: Count of results selected for processing
    )
    {
        // Intermediate results towards RMS residuals
        double rollRmsDiff = 0.0; // Roll RMS difference
        double pitchRmsDiff = 0.0; // Pitch RMS difference
        double yawRmsDiff = 0.0; // Yaw RMS difference
        double rollRmsSum = 0.0; // Roll RMS sum of squares
        double pitchRmsSum = 0.0; // Pitch RMS sum of squares
        double yawRmsSum = 0.0; // Yaw RMS sum of squares

        // Compute best-fit alignment
        rollAlignmentEstimate = weightedRollsSum / weightSum; 
        pitchAlignmentEstimate = weightedPitchesSum / weightSum;  
        yawAlignmentEstimate = weightedYawsSum / weightSum;  

        // Process each selected measurement, computing the summations for 
        // roll, pitch, and yaw
        for (int index = 0; index < selectedCount; index++) 
        {
            rollRmsDiff = rollPrimes[index] - rollAlignmentEstimate; 
            pitchRmsDiff = pitchPrimes[index] - pitchAlignmentEstimate;
            yawRmsDiff = yawPrimes[index] - yawAlignmentEstimate; 

            rollRmsSum += rollRmsDiff * rollRmsDiff;
            pitchRmsSum += pitchRmsDiff * pitchRmsDiff;
            yawRmsSum += yawRmsDiff * yawRmsDiff;
        }

        // Given the summations, find RMS residuals for roll, pitch, and yaw
        rollRms = Math.sqrt (rollRmsSum / selectedCount);
        pitchRms = Math.sqrt (pitchRmsSum / selectedCount);
        yawRms = Math.sqrt (yawRmsSum / selectedCount);
    }

    /**
     * Create updated alignment matrix from the roll, pitch, and yaw
     */
    public void calculateAlignmentMatrix
    (
    )
    {
        // Cache the sine/cosine values of the roll, pitch, and yaw.  Also
        // convert microradians to radians
        double sinRoll = Math.sin(rollAlignmentEstimate * 1.0e-6); 
        double cosRoll = Math.cos(rollAlignmentEstimate * 1.0e-6); 
        double sinPitch = Math.sin(pitchAlignmentEstimate * 1.0e-6); 
        double cosPitch = Math.cos(pitchAlignmentEstimate * 1.0e-6); 
        double sinYaw = Math.sin(yawAlignmentEstimate * 1.0e-6); 
        double cosYaw = Math.cos(yawAlignmentEstimate * 1.0e-6); 

        // Create the alignment matrix from the roll, pitch, and yaw
        alignmentMatrix[0][0] = cosPitch * cosYaw;
        alignmentMatrix[0][1] = -cosPitch * sinYaw;
        alignmentMatrix[0][2] = sinPitch;
        alignmentMatrix[1][0] = (sinRoll * sinPitch * cosYaw)
                                + (cosRoll * sinYaw);
        alignmentMatrix[1][1] = (-sinRoll * sinPitch * sinYaw)
                                + (cosRoll * cosYaw);
        alignmentMatrix[1][2] = -sinRoll * cosPitch;
        alignmentMatrix[2][0] = (-cosRoll * sinPitch * cosYaw)
                                + (sinRoll * sinYaw);
        alignmentMatrix[2][1] = (cosRoll * sinPitch * sinYaw)
                                + (sinRoll * cosYaw);
        alignmentMatrix[2][2] = cosRoll * cosPitch;
    }

    /**
     * Write data to the report files.  This includes the main report and the
     * CPF-formatted alignment matrix group
     */
    public void writeSensorAlignmentReports
    (
        File reportFile // I: Name of the main report file
    ) throws IOException 
    {
        // These are used to generate the report file creation date and time
        Date today; // The date the tool is being run
        DateFormat dateFormatter; // Format the date format of today's date
        DateFormat timeFormatter; // Format the time part of today's date
        SimpleDateFormat calDateFormatter; // Format for CPF date

        // Initialize alignment matrix formatter
        Formatter matrixFmt = new Formatter();

        double secondsInDay; // Number of seconds in the last day of a date 
        String path; // This is for the part of the path before the filename

        // This is for converting the Calibration effective dates to strings
        String calStartDate;
        String calEndDate;

        PrintWriter report = null; // Prints the main report
        PrintWriter attitudeReport = null; // Prints the attitude report

        try
        {
            calDateFormatter = new SimpleDateFormat("yyyy-MM-dd");

            // Get the current date and time here right before the report is 
            // made so the report file creation date and time are accurate
            dateFormatter = new SimpleDateFormat("EEE. MMM. d, yyyy");
            timeFormatter = new SimpleDateFormat("kk:mm");
            today = new Date();
            dateOut = dateFormatter.format(today);
            timeOut = timeFormatter.format(today);

            // Convert the calibration effective dates to presentable format
            calStartDate = calDateFormatter.format(effectiveStartDate); 
            calEndDate = calDateFormatter.format(effectiveEndDate);

            // Write contents to report file
            report = new PrintWriter(reportFile.getAbsolutePath());
            report.println(dateOut + "                   OLI     " 
                + "            Time: " + timeOut);
            report.println("                           EROS Data Center");
            report.println("                  Sensor Alignment Calibration "
                + "Report");
            report.println("------------------------------------------------"
               + "---------------------------");

            // Print IAS version
            report.println("IAS Version:          " 
                + Constants.getIasSoftwareVersion());

            // Print the database schema
            String value;
            try
            {
                value = System.getProperty(Constants.DAO_AUTH_IAS_PROD);
            }
            catch (Exception ex)
            {
                value = null;
            }
            if (value != null)
            {
                String[] parts = value.split("/");
                if (parts.length == 2)
                    report.println("Database schema:      " + parts[0]);
                else
                    report.println("Database schema:");
            }
            else
            {
                report.println("Database schema:");
            }

            // Display various report header values
            report.println("");
            report.println("");
            report.println("Spacecraft:           LDCM");
            report.println("Sensor:               OLI");
            report.println("Acquisition Type:     Earth-viewing");
            report.println("Off-Nadir Angle:      " + maxRoll);
            report.println("Control Type:         " + gcpSource);
            report.println("");
            report.println("");
            report.println("Number of scenes processed:  " + rollPrimes.length);

            // Format and display control type weights
            report.format("DOQ scene weight: %d\n", doqWeight);
            report.format("GLS scene weight: %d\n\n", glsWeight);
                    
            // Set up date formats to extract components of acquisition start 
            // and end dates
            SimpleDateFormat yearSdf = new SimpleDateFormat("yyyy");
            SimpleDateFormat jdaySdf = new SimpleDateFormat("DDD");
            SimpleDateFormat hourSdf = new SimpleDateFormat("HH"); 
            SimpleDateFormat minuteSdf = new SimpleDateFormat("mm"); 
            SimpleDateFormat secondSdf = new SimpleDateFormat("ss.SSS"); 

            if (acqQueryStartDateSelected)
            {
                secondsInDay = 3600 * Double.parseDouble(hourSdf.format(
                    acqQueryStartDate))
                    + 60 * Double.parseDouble(minuteSdf.format(
                    acqQueryStartDate))
                    + Double.parseDouble(secondSdf.format(acqQueryStartDate));
                report.format("Time period start:  year = " +
                    "%4d jday = %3d seconds = %8.6e\n", 
                    Integer.parseInt(yearSdf.format(acqQueryStartDate)), 
                    Integer.parseInt(jdaySdf.format(acqQueryStartDate)),
                    secondsInDay); 
            }
            else
            {
                report.format("Time period start:  Not selected\n"); 
            }

            if (acqQueryEndDateSelected)
            {
                secondsInDay = 3600 * Double.parseDouble(hourSdf.format(
                    acqQueryEndDate))
                    + 60 * Double.parseDouble(minuteSdf.format(acqQueryEndDate))
                    + Double.parseDouble(secondSdf.format(acqQueryEndDate));
                report.format("Time period end:    year = " +
                    "%4d jday = %3d seconds = %8.6e\n\n", 
                    Integer.parseInt(yearSdf.format(acqQueryEndDate)), 
                    Integer.parseInt(jdaySdf.format(acqQueryEndDate)),
                    secondsInDay);
            }
            else
            {
                report.format("Time period end:    Not selected\n\n"); 
            }

            // Set up and display the best fit aligment values
            report.format("Best Fit Roll Alignment:  %9.3f microradians\n",
                rollAlignmentEstimate);
            report.format("Best Fit Pitch Alignment: %9.3f microradians\n",
                pitchAlignmentEstimate);
            report.format("Best Fit Yaw Alignment:   %9.3f microradians\n\n",
                yawAlignmentEstimate);

            // Set up and display the RMS aligment values
            report.format("RMS Roll Alignment:  %6.3f microradians\n", 
                rollRms); 
            report.format("RMS Pitch Alignment: %6.3f microradians\n", 
                pitchRms); 
            report.format("RMS Yaw Alignment:   %6.3f microradians\n\n", 
                yawRms);

            // Display the CPF effective date range 
            report.println("CPF Effective Start Date: " + calStartDate);
            report.println("CPF Effective End Date:   " + calEndDate);
            report.println("");

            // Set up the Alignment Matrix CPF format
            matrixFmt.format("GROUP = ATTITUDE_PARAMETERS\n");
            matrixFmt.format(" ACS_to_OLI_Rotation_Matrix = " +
                "(%11.9f,%11.9f,%11.9f,\n",
                alignmentMatrix[0][0],
                alignmentMatrix[0][1],
                alignmentMatrix[0][2]);
            matrixFmt.format("  %11.9f,%11.9f,%11.9f,\n", 
                alignmentMatrix[1][0],
                alignmentMatrix[1][1],
                alignmentMatrix[1][2]);
            matrixFmt.format("  %11.9f,%11.9f,%11.9f)\n", 
                alignmentMatrix[2][0],
                alignmentMatrix[2][1],
                alignmentMatrix[2][2]);
            matrixFmt.format("END_GROUP = ATTITUDE_PARAMETERS");

            // Print the Alignment Matrix in CPF format
            report.println(matrixFmt);
            report.println("");
    
            // Print the table of scene data
            String heading = String.format(
                    "%12s %10s %8s %7s %17s %8s %11s %11s %11s %11s",
                    "WKORD_ID",
                    "GEO_CHR_ID",
                    "WRS_PATH",
                    "WRS_ROW",
                    "    ACQ_DATE     ",
                    "GCP_TYPE",
                    "GCP_RMS_FIT",
                    " ROLL_ALIGN",
                    "PITCH_ALIGN",
                    "  YAW_ALIGN");
            report.println(heading);

            for (int row = 0; row < rollPrimes.length; row++)
            {
                String f1 = String.format("%6.3f", gcpRmsFits[row]);
                String f2 = String.format("%6.3f", rollPrimes[row]);
                String f3 = String.format("%6.3f", pitchPrimes[row]);
                String f4 = String.format("%6.3f", yawPrimes[row]);
                String dataRow = String.format(
                  "%12s %10s %5s%03d %4s%03d %8sT%8s %8s %11s %11s %11s %11s",
                  woIds[row],
                  geoCharIds[row],
                  " ",
                  paths[row],
                  " ",
                  rows[row],
                  acqDates[row],
                  acqTimes[row],
                  ctrlTypes[row],
                  f1,
                  f2,
                  f3,
                  f4);
                report.println(dataRow);
            }

            // Also, make the attitude parameters CPF-formatted file in the
            // same directory as the report file
            SimpleDateFormat calFiledateFormatter;
            calFiledateFormatter = new SimpleDateFormat("yyyyMMdd");
            String calFileStartDate;
            String calFileEndDate;

            calFileStartDate = calFiledateFormatter.format(effectiveStartDate);
            calFileEndDate = calFiledateFormatter.format(effectiveEndDate); 

            // Build the attitude parameters filename including path
            path = reportFile.getParent();

            // getParent returns null if there is no path
            if (path == null)
                path = "";
            else
                path = path.concat("/");

            String attitudeParamName;
            attitudeParamName = path + "attitude_parameters_" 
                + calFileStartDate + "_" + calFileEndDate + ".odl"; 

            // Make the attitude parameters file
            File attitudeParamFile = new File(attitudeParamName);
            attitudeParamFile.createNewFile();

            // Write contents to attitude parameters file
            
            attitudeReport = new PrintWriter(
                attitudeParamFile.getAbsolutePath());

            // Print the Alignment Matrix in CPF format
            attitudeReport.println(matrixFmt);
        }
        catch (IOException ex)
        {
            throw ex;
        }
        finally 
        {
            if (report != null)
            {
                report.close();
            }
            if (attitudeReport != null)
            {
                attitudeReport.close();
            }
        }
    }

    /**
     * Save results to the IAS database.  Results include the alignment 
     * estimate, alignment matrix, user selections, and other information.  
     * These are sent to the OLI Sensor Alignment database table.
     */
    public void saveToDatabase
    (
    ) throws SQLException
    {    
        Date today; // Today's date and time       
        String processingCenter; // Where Sensor Alignment was processed
        
        today = new Date();
        
        // Look up the processing center environment variable
        processingCenter = System.getenv("PROCESSING_CENTER");
        if (processingCenter == null)
        {
            // Use EROS as the default
            processingCenter = "EROS";
        }
        
        // Set up the database insert command
        String connectionProperty = "sensor_alignment.db.conn";
        String authProperty = Constants.DAO_AUTH_IAS_PROD;
        DataConnection dataConnection = new DataConnection(connectionProperty);
        PreparedStatement stmt = null;
        try 
        {
            stmt = dataConnection.getConnection(authProperty).
                prepareStatement("insert into OLI_SENSOR_ALIGNMENT "
                + "(DATE_PROCESSED, PROCESSING_CENTER, OFF_NADIR_ANGLE, "
                + "GCP_SOURCE, DOQ_WEIGHT, GLS_WEIGHT, NUMBER_OF_SCENES_USED, " 
                + "DATA_START_DATE, DATA_STOP_DATE, ROLL_ALIGNMENT, "
                + "PITCH_ALIGNMENT, YAW_ALIGNMENT, ROLL_RESIDUAL_RMSE, "
                + "PITCH_RESIDUAL_RMSE, YAW_RESIDUAL_RMSE, "
                + "ALIGNMENT_EFFECTIVE_DATE_START, "
                + "ALIGNMENT_EFFECTIVE_DATE_STOP, "
                + "ALIGNMENT_MATRIX_0, ALIGNMENT_MATRIX_1, ALIGNMENT_MATRIX_2, "
                + "ALIGNMENT_MATRIX_3, ALIGNMENT_MATRIX_4, ALIGNMENT_MATRIX_5, "
                + "ALIGNMENT_MATRIX_6, ALIGNMENT_MATRIX_7, ALIGNMENT_MATRIX_8) "
                + "values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, "
                + "?, ?, ?, ?, ?, ?, ?, ?, ?)");      

            // Convert Date format to SQL Timestamp format for CPF and today
            java.sql.Timestamp todaySqlTimestamp = 
                new java.sql.Timestamp(today.getTime()); 
            java.sql.Timestamp startCPFSqlTimestamp = 
                new java.sql.Timestamp(effectiveStartDate.getTime());
            java.sql.Timestamp endCPFSqlTimestamp = 
                new java.sql.Timestamp(effectiveEndDate.getTime());

            // Fill in the insert command variables
            stmt.setTimestamp(1, todaySqlTimestamp);
            stmt.setString(2, processingCenter);
            stmt.setDouble(3, maxRoll);
            stmt.setString(4, gcpSource);
            stmt.setInt(5, doqWeight);
            stmt.setInt(6, glsWeight);
            stmt.setInt(7, rollPrimes.length);
            if (acqQueryStartDateSelected)
            {
                java.sql.Timestamp startAcqSqlTimestamp =
                    new java.sql.Timestamp(acqQueryStartDate.getTime()); 
                stmt.setTimestamp(8, startAcqSqlTimestamp);
            }
            else
            {
                stmt.setTimestamp(8, null);
            }
            if (acqQueryEndDateSelected)
            {
                java.sql.Timestamp endAcqSqlTimestamp =
                    new java.sql.Timestamp(acqQueryEndDate.getTime()); 
                stmt.setTimestamp(9, endAcqSqlTimestamp);
            }
            else
            {
                stmt.setTimestamp(9, null);
            }
            stmt.setDouble(10, rollAlignmentEstimate);
            stmt.setDouble(11, pitchAlignmentEstimate);
            stmt.setDouble(12, yawAlignmentEstimate);
            stmt.setDouble(13, rollRms);
            stmt.setDouble(14, pitchRms);
            stmt.setDouble(15, yawRms);
            stmt.setTimestamp(16, startCPFSqlTimestamp);
            stmt.setTimestamp(17, endCPFSqlTimestamp);
            stmt.setDouble(18, alignmentMatrix[0][0]);
            stmt.setDouble(19, alignmentMatrix[0][1]);
            stmt.setDouble(20, alignmentMatrix[0][2]);
            stmt.setDouble(21, alignmentMatrix[1][0]);
            stmt.setDouble(22, alignmentMatrix[1][1]);
            stmt.setDouble(23, alignmentMatrix[1][2]);
            stmt.setDouble(24, alignmentMatrix[2][0]);
            stmt.setDouble(25, alignmentMatrix[2][1]);
            stmt.setDouble(26, alignmentMatrix[2][2]);

            stmt.executeUpdate();
            dataConnection.commit();
        } 
        catch (SQLException sqlException) 
        {
            dataConnection.rollback();
            throw sqlException;
        } 
        finally 
        {
            try 
            {
                if (stmt != null) {
                    stmt.close();
                }
            } 
            catch (SQLException ex) 
            {
                System.out.println("Warning: " + ex.getMessage());
            }
            dataConnection.close();
        }
    }
}
