package sensor_alignment;

/**
 * Main Sensor Alignment module.  Set up the initial screen that allows the
 * user to make the selections needed to choose what records will be 
 * retrieved from the database
 */
import gov.usgs.ias.gui.common_library.CalendarDialog;
import gov.usgs.ias.gui.common_library.ObservableDate;
import gov.usgs.ias.gui.common_library.DatePanel;
import gov.usgs.ias.constant.Constants;
import javax.swing.JFrame;
import javax.swing.JPanel;
import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import javax.swing.JButton;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import javax.swing.JLabel;
import javax.swing.JRadioButton;
import javax.swing.ButtonGroup;
import javax.swing.JCheckBox;
import javax.swing.UIManager;
import javax.swing.JOptionPane;
import java.awt.Point;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.text.ParseException;
import java.text.Format;
import java.util.Observable;
import java.util.Observer;
import java.util.Calendar;
import java.util.Date;
import javax.swing.JComponent;
import java.awt.Dimension;
import java.net.URL;
import java.io.InputStream;
import java.io.IOException;
import java.util.Properties;

// Set up access to the IAS database for the test login
import gov.usgs.ias.dao.DataConnection;
import java.sql.SQLException;

public class SensorAlignment extends JFrame implements
    WindowListener, ActionListener, ChangeListener, ItemListener, Observer
{
    private ObservableDate updateDisplay; // Used to store previous values
                                     // (dates) for use among multiple widgets
    private CalendarDialog calendarDialog; // The calendar that pops up when
                                     // the user clicks the calendar button on
                                     // a date panel
    private DatePanel startDatePanel; // Panel for the query start date
    private DatePanel endDatePanel;  // Panel for the query end date
    private JCheckBox enableStartDate; // Checkbox to enable start date panel
    private JCheckBox enableEndDate; // Checkbox to enable end date panel
    private boolean startDateSelected = false; // Flag for start date picked
    private boolean endDateSelected = false; // Flag for end date picked
    private double step = 0.1;       // Step size for off nadir angle choices 
    private double offNadirAngleLimit = 0.1; // Default 0.1 degrees
    private String offNadirRangeTypeStr = "INWARD"; // Current range type
    private String controlTypeStr = "DOQ"; // Tracks control type
    private SensorAlignmentParamTableDialog sensorAlignmentParamTableDialog;
    static public final double offNadirAngleMin = 0.1; // Spinner range minimum
    static public final double offNadirAngleMax = 15.0; // Spinner range max

    /**
     * Build the user interface elements of the initial DB query screen
     */
    public SensorAlignment() throws IOException
    {
        JPanel mainPanel;        // Main panel for the window
        JLabel topSpaceLabel;    // Space at the top of the window
        JLabel calendarSpaceLabel; // Space to set apart the calendar
        JButton queryButton;     // Button to issue the query
        JButton exitButton;      // Button to exit the application 
        JPanel buttonPanel;      // Panel for the buttons
        JLabel controlTypeLabel; // Label for the control type widget
        ButtonGroup controlTypeButtonGroup; // Control type button group
        JRadioButton DOQRadioButton; // DOQ selection button
        JRadioButton GLSRadioButton; // GLS selection button
        JRadioButton bothRadioButton; // Both DOQ and GLS selection button
        JLabel offNadirAngleLabel; // Label for the off-nadir angle spinner 
        JSpinner offNadirAngleSpinner; // Spinner for off-nadir angle range
        SpinnerNumberModel offNadirAngleSpinnerModel; // Spinner model for 
                                     // off-nadir angle range
        ButtonGroup offNadirRangeButtonGroup; // Off-nadir range group
        JRadioButton inwardRangeRadioButton; // Off-nadir range from -x to +x
        JRadioButton outwardRangeRadioButton; // Off-nadir range <= -x and
                                     // >= +x
        JLabel offNadirRangeTypeLabel; // Range direction label 
        boolean launchDateParsed = true; // Successful launch date parse flag
        final long serialVersionUID = 0; // Get rid of a compiler warning

        // Allow the Enter key to select certain buttons 
        UIManager.put("Button.defaultButtonFollowsFocus", Boolean.TRUE);

        // Set up window 
        setTitle("Sensor Alignment - Main Window");
        setSize(700,410);
        setLayout(new BorderLayout());
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

        mainPanel = new JPanel();
        GridBagLayout gridbag = new GridBagLayout();
        GridBagConstraints gbc = new GridBagConstraints();
        mainPanel.setLayout(gridbag);

        // Set up checkboxes to enable and disable the date panels.  The user 
        // can select, or refrain from selecting, a start and/or end date
        enableStartDate = new JCheckBox("Select Query Start Date", false);
        enableEndDate = new JCheckBox("Select Query End Date  ", false);
        enableStartDate.setSelected(false);
        enableEndDate.setSelected(false);
        enableStartDate.setMnemonic(KeyEvent.VK_S);
        enableEndDate.setMnemonic(KeyEvent.VK_E);
        enableStartDate.addItemListener(this);
        enableEndDate.addItemListener(this);

        // Create calendar and date dialogs
        calendarDialog = new CalendarDialog((JFrame)this);
        startDatePanel = new DatePanel(this, "Start of Query Date Range", 
            calendarDialog);
        endDatePanel = new DatePanel(this, "End of Query Date Range", 
            calendarDialog);

        // Let the date panels get updates made in the calendar dialogs
        ObservableDate.getInstance().addObserver(this);

        // Initialize start calendar date default to be the launch date, which
        // is defined as a property 
        Calendar startDateCalendar;
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd");
        String launchDateStr = System.getProperty(
            "sensor_alignment.launch_date");

        // In case we can't parse launch date, initialize it to today's date 
        Date launchDate = new Date();
        try 
        {
            launchDate = (Date)sdf.parse(launchDateStr); // Put in Date format
        } 
        catch (ParseException e) 
        {
            // Could not parse launch date property; use today's date instead
            System.out.println("Warning: failed to parse launch date "
                + "property, so using today's date:\n" + e);
        }
        startDateCalendar = Calendar.getInstance();
        startDateCalendar.setTime(launchDate);
        startDatePanel.setDate(startDateCalendar.get(Calendar.YEAR),
            startDateCalendar.get(Calendar.MONTH),
            startDateCalendar.get(Calendar.DAY_OF_MONTH));

        // Initialize end calendar date default to be today's date
        Calendar endDateCalendar;
        endDateCalendar = Calendar.getInstance();  
        endDatePanel.setDate(endDateCalendar.get(Calendar.YEAR),
            endDateCalendar.get(Calendar.MONTH),
            endDateCalendar.get(Calendar.DAY_OF_MONTH));

        // Set up labels to take up vertical space
        topSpaceLabel = new JLabel(" ");
        calendarSpaceLabel = new JLabel(" ");

        // Use this for all components
        gbc.ipadx = 10;
        gbc.ipady = 10;

        // Put some space at the top
        gbc.gridx = 0;
        gbc.gridy = 0;
        gbc.gridwidth = 1;
        gbc.anchor = GridBagConstraints.LINE_START;
        mainPanel.add(topSpaceLabel, gbc);

        // Put some space before the calendars 
        gbc.gridx = 0;
        gbc.gridy = 3;
        gbc.gridwidth = 1;
        gbc.anchor = GridBagConstraints.LINE_START;
        mainPanel.add(calendarSpaceLabel, gbc);

        // Add start date checkbox to the main panel
        gbc.gridx = 0;
        gbc.gridy = 4;
        gbc.gridwidth = 1;
        gbc.anchor = GridBagConstraints.CENTER;
        mainPanel.add(enableStartDate, gbc);

        // Add start date to the main panel
        gbc.gridx = 0;
        gbc.gridy = 5;
        gbc.gridwidth = 4;
        gbc.weighty = 0.3;
        gbc.anchor = GridBagConstraints.CENTER;
        mainPanel.add(startDatePanel, gbc);

        // Add end date checkbox to the main panel
        gbc.gridx = 0;
        gbc.gridy = 6;
        gbc.gridwidth = 1;
        gbc.anchor = GridBagConstraints.CENTER;
        mainPanel.add(enableEndDate, gbc);

        // Add end date to the main panel
        gbc.gridx = 0;
        gbc.gridy = 7;
        gbc.gridwidth = 4;
        gbc.weighty = 0.3;
        gbc.anchor = GridBagConstraints.CENTER;
        mainPanel.add(endDatePanel, gbc);

        // Set up the Control Type selections.
        controlTypeLabel = new JLabel("Control Type");
        gbc.gridx = 1;
        gbc.gridy = 3;
        gbc.gridwidth = 3;
        gbc.anchor = GridBagConstraints.CENTER;
        mainPanel.add(controlTypeLabel, gbc);

        controlTypeButtonGroup = new ButtonGroup();

        DOQRadioButton  = new JRadioButton("DOQ");
        GLSRadioButton  = new JRadioButton("GLS");
        bothRadioButton = new JRadioButton("BOTH");

        // DOQ is the default.  This setting should be in sync with the
        // controlTypeStr
        DOQRadioButton.setSelected(true);

        // Add radio buttons to Control Type button group
        controlTypeButtonGroup.add(DOQRadioButton);
        controlTypeButtonGroup.add(GLSRadioButton);
        controlTypeButtonGroup.add(bothRadioButton);

        // Set keyboard mnemonics for the radio buttons
        DOQRadioButton.setMnemonic(KeyEvent.VK_D);
        GLSRadioButton.setMnemonic(KeyEvent.VK_G);
        bothRadioButton.setMnemonic(KeyEvent.VK_B);

        // Register listeners for the radio buttons
        DOQRadioButton.addActionListener(this);
        GLSRadioButton.addActionListener(this);
        bothRadioButton.addActionListener(this);

        // Set action commands for the radio buttons
        DOQRadioButton.setActionCommand("setDOQ");
        GLSRadioButton.setActionCommand("setGLS");
        bothRadioButton.setActionCommand("setBOTH");

        // Add the radio buttons to the main panel
        gbc.gridx = 1;
        gbc.gridy = 4;
        gbc.gridwidth = 1;
        gbc.weightx = 0.05;
        gbc.anchor = GridBagConstraints.PAGE_START;
        mainPanel.add(DOQRadioButton,gbc);
        gbc.gridx = 2;
        gbc.gridy = 4;
        gbc.gridwidth = 1;
        gbc.weightx = 0.05;
        gbc.anchor = GridBagConstraints.PAGE_START;
        mainPanel.add(GLSRadioButton,gbc);
        gbc.gridx = 3;
        gbc.gridy = 4;
        gbc.gridwidth = 1;
        gbc.weightx = 0.05;
        gbc.anchor = GridBagConstraints.PAGE_START;
        mainPanel.add(bothRadioButton,gbc);

        // Set up the Off-nadir Angle spinner and its label
        offNadirAngleLabel =
            new JLabel("Off-Nadir Angle Limit", JLabel.CENTER);
        offNadirAngleSpinnerModel =
            new SpinnerNumberModel(offNadirAngleLimit, offNadirAngleMin, 
                offNadirAngleMax, step);
        offNadirAngleSpinner = new JSpinner(offNadirAngleSpinnerModel);
        offNadirAngleSpinner.addChangeListener(this);
        offNadirAngleSpinner.setPreferredSize(new Dimension(40, 20));

        // Add the off-nadir angle spinner's label to the main panel
        gbc.gridx = 0;
        gbc.gridy = 1;
        gbc.gridwidth = 1;
        gbc.anchor = GridBagConstraints.CENTER;
        mainPanel.add(offNadirAngleLabel, gbc);

        // Add the off-nadir angle spinner to the main panel
        gbc.gridx = 0;
        gbc.gridy = 2;
        gbc.gridwidth = 1;
        gbc.weightx = 0.7;
        gbc.weighty = 0.4;
        gbc.anchor = GridBagConstraints.CENTER;
        mainPanel.add(offNadirAngleSpinner, gbc);

        // Set up the Off-Nadir angle range type selections
        offNadirRangeTypeLabel = new JLabel("Off-Nadir Angle Range Type");
        gbc.gridx = 1;
        gbc.gridy = 1;
        gbc.gridwidth = 3;
        gbc.anchor = GridBagConstraints.CENTER;
        mainPanel.add(offNadirRangeTypeLabel, gbc);

        offNadirRangeButtonGroup = new ButtonGroup();

        inwardRangeRadioButton = new JRadioButton("Inward");
        outwardRangeRadioButton = new JRadioButton("Outward");
        inwardRangeRadioButton.setSelected(true);

        // Add radio buttons to off-nadir angle range type button group
        offNadirRangeButtonGroup.add(inwardRangeRadioButton);
        offNadirRangeButtonGroup.add(outwardRangeRadioButton);

        // Set keyboard mnemonics for the radio buttons
        inwardRangeRadioButton.setMnemonic(KeyEvent.VK_I);
        outwardRangeRadioButton.setMnemonic(KeyEvent.VK_O);

        // Register listeners for the radio buttons
        inwardRangeRadioButton.addActionListener(this);
        outwardRangeRadioButton.addActionListener(this);

        // Set action commands for the radio buttons
        inwardRangeRadioButton.setActionCommand("setInward");
        outwardRangeRadioButton.setActionCommand("setOutward");

        // Add the off-nadir range type radio buttons to the main panel
        gbc.gridx = 1;
        gbc.gridy = 2;
        gbc.gridwidth = 1;
        gbc.weightx = 0.05;
        gbc.anchor = GridBagConstraints.PAGE_START;
        mainPanel.add(inwardRangeRadioButton, gbc);
        gbc.gridx = 2;
        gbc.gridy = 2;
        gbc.gridwidth = 1;
        gbc.weightx = 0.05;
        gbc.anchor = GridBagConstraints.PAGE_START;
        mainPanel.add(outwardRangeRadioButton, gbc);

        // Set up the Query and Exit buttons
        buttonPanel = new JPanel();

        queryButton = new JButton("Query");
        queryButton.setMnemonic(KeyEvent.VK_Q);
        queryButton.setToolTipText("Query Database for Alignment Profiles");
        queryButton.addActionListener(this);
        buttonPanel.add(queryButton);

        exitButton = new JButton("Exit");
        exitButton.setMnemonic(KeyEvent.VK_X);
        exitButton.setToolTipText("Exit the Application");
        exitButton.addActionListener(this);
        buttonPanel.add(exitButton);

        add(mainPanel,BorderLayout.NORTH);
        add(buttonPanel,BorderLayout.SOUTH);

        // Make sure the date panels aren't visible unless they are selected
        startDatePanel.setVisible(false);
        endDatePanel.setVisible(false);

        setVisible(true);
    }

    // method to handle the windowClosing event
    //-----------------------------------------
    public void windowClosing(WindowEvent e)  {    }

    // dummy window event handlers for events that do not need handling
    //-----------------------------------------------------------------
    public void windowActivated(WindowEvent e) { }
    public void windowClosed(WindowEvent e) { }
    public void windowDeactivated(WindowEvent e) { }
    public void windowDeiconified(WindowEvent e) { }
    public void windowIconified(WindowEvent e) { }
    public void windowOpened(WindowEvent e) { }

    /**
     * Action performed event handler responds to user actions 
     */
    public void actionPerformed(ActionEvent e)
    {
        String command = e.getActionCommand();
        int status; // Status of method

        Calendar startDateCalendar; // Start calendar
        Calendar endDateCalendar; // End calendar
        
        Point loc = this.getLocationOnScreen();
        loc.y += 50;
        loc.x += 50;

        // Query the database, and call the Parameter Table window using the
        // resulting data
        if (command.equals("Query"))
        {
            // Get the start and end Calendars.  If the start date isn't 
            // selected, all data on or before the end date are used.  The 
            // reverse holds for the end date.  If neither date is selected, 
            // all dates are included
            startDateCalendar = startDatePanel.getDate(); 
            endDateCalendar = endDatePanel.getDate(); 

            // Ensure that the start date is not before the end date
            if ((startDateSelected && endDateSelected) &&
                (endDateCalendar.before(startDateCalendar)))
            {
                JOptionPane.showMessageDialog(this, "Query end date is before "
                    + "the start date", "Error", JOptionPane.ERROR_MESSAGE);
                return;
            }

            // Set up the database login
            String authProperty = Constants.DAO_AUTH_IAS_PROD;

            // If the login information isn't already set up, set it up
            if (System.getenv(authProperty) == null &&
                System.getProperty(authProperty) == null)
            {
                // Get the database login information using a login GUI screen
                LoginDialog loginDialog = new LoginDialog(this, authProperty);
                loginDialog.setLocation(loc);
                loginDialog.setVisible(true);
                if (loginDialog.isCancelled()) {
                    return;
                }
            }

            // Ensure the connection information is correct by testing it
            try 
            {
                tryDBConnection(authProperty);
            } 
            catch (Exception ex) 
            {
                JOptionPane.showMessageDialog(this, ex.getMessage(), "Error",
                    JOptionPane.ERROR_MESSAGE);
                return;
            }

            // Start the Parameter Table window
            try 
            {
                sensorAlignmentParamTableDialog =
                    new SensorAlignmentParamTableDialog(this,
                        controlTypeStr,
                        startDateCalendar.getTime(),
                        endDateCalendar.getTime(),
                        startDateSelected,
                        endDateSelected,
                        offNadirAngleLimit,
                        offNadirRangeTypeStr);
            } 
            catch (Exception ex) 
            {
                JOptionPane.showMessageDialog(this, ex.getMessage(),
                    "Error", JOptionPane.ERROR_MESSAGE);
                return;
            }
            sensorAlignmentParamTableDialog.setLocation(loc);
            sensorAlignmentParamTableDialog.setVisible(true);
        }
        else if (command.equals("Exit"))
        {
            System.exit(0);
        }
        else if (command.equals("setDOQ"))
        {
            controlTypeStr = "DOQ";
        }
        else if (command.equals("setGLS"))
        {
            controlTypeStr = "GLS";
        }
        else if (command.equals("setBOTH"))
        {
            controlTypeStr = "BOTH";
        }
        else if (command.equals("setInward"))
        {
            offNadirRangeTypeStr = "INWARD";
        }
        else if (command.equals("setOutward"))
        {
            offNadirRangeTypeStr = "OUTWARD";
        }
    }

    /**
     * Change event handler responds to state changes caused by user actions 
     */
    public void stateChanged(ChangeEvent e)
    {
        JSpinner source = (JSpinner)e.getSource();
        if (source.getValue() instanceof Double)
        {
            offNadirAngleLimit = (Double)source.getValue();
        }
    }

    /**
     * Item performed event handler responds to state changes caused by user
     * actions
     */
    public void itemStateChanged(ItemEvent e)
    {
        Object source = e.getItemSelectable();

        // When the user selects a date enabled checkbox, toggle the enable 
        // flag the visibility of the date panel for that checkbox
        if (source == enableStartDate)
        {
            if (e.getStateChange() == ItemEvent.SELECTED)
            {
                startDateSelected = true;
                startDatePanel.setVisible(true);
            }
            if (e.getStateChange() == ItemEvent.DESELECTED)
            {
                startDateSelected = false;
                startDatePanel.setVisible(false);
            }
        }

        if (source == enableEndDate)
        {
            if (e.getStateChange() == ItemEvent.SELECTED)
            {
                endDateSelected = true;
                endDatePanel.setVisible(true);
            }
            if (e.getStateChange() == ItemEvent.DESELECTED)
            {
                endDateSelected = false;
                endDatePanel.setVisible(false);
            }
        }
    }

    /**
     * Try to establish a database connection
     */
    public static void tryDBConnection
    (
        String authProperty  // DB authorization property
    ) throws SQLException
    {
        String connectionProperty = "sensor_alignment.db.conn";
        String dbHostProperty = "sensor_alignment.db.host";
        String dbPortProperty = "sensor_alignment.db.port";
        String dbSidProperty  = "sensor_alignment.db.sid";

        // Check to see if the connection string is already defined
        if (System.getProperty(connectionProperty) == null) 
        {
            String host = System.getProperty(dbHostProperty);
            String port = System.getProperty(dbPortProperty);

            // Try TWO_TASK, then ORACLE_SID, then property
            String sid = System.getenv("TWO_TASK");
            if (sid == null) sid = System.getenv("ORACLE_SID");
            if (sid == null) sid = System.getProperty(dbSidProperty);
            if (sid == null || host == null ||  port == null) 
            {
                System.clearProperty(connectionProperty);
                throw new SQLException("Error setting up database connection, "
                    + "please verify database host and SID properties: sid = "
                    + sid + ", host = " + host + ", port = " + port);
            }

            // Build the connection string and set the connection property
            String connectionString = "jdbc:oracle:thin:authorization@" + host
                + ":" + port + ":" + sid;
            System.setProperty(connectionProperty, connectionString);
        }

        // Set up the database connection
        DataConnection dataConnection = new DataConnection(connectionProperty);

        // After prompting for the login information, check if it's set
        if (System.getenv(authProperty) == null &&
           System.getProperty(authProperty) == null) {
            throw new SQLException("Error setting up database connection, "
                + "authProperty is not set as a property or environment "
                + "variable");
        } else {
            // Try a DB login to ensure the login information is correct
            try 
            {
                dataConnection.getConnection(authProperty);
            } 
            catch (SQLException ex) 
            {
                System.clearProperty(authProperty);
                // Get the root cause of the exception
                throw ex;
            }
            finally {
                dataConnection.close();
            }
        }
    }

    /**
     * Set system properties from the properties file.  Will throw an exception
     * if the properties file can't be found or read
     */
    private static void initializeProperties() throws IOException {

        ClassLoader cl = SensorAlignment.class.getClassLoader();
        URL propURL = cl.getResource(".sensor_alignment.properties");
        InputStream propStream = null;
        try
        {
            propStream = propURL.openStream();
            Properties appProperties = new Properties();
            appProperties.load(propStream);
            for (String pName : appProperties.stringPropertyNames()) {
                System.setProperty(pName, appProperties.getProperty(pName));
            }
        }
        finally
        {
            propStream.close();
        }
    }

    /**
     * Print command-line usage help.  Since a getops equivalent isn't 
     * available, keep the syntax simple rather than flexible
     */
    private static void printUsage() {
        System.out.println("\nNormally Sensor Alignment is used in a " +
            "graphical user interface environment by ");
        System.out.println("running command sensor_alignment.sh. It can " +
            "also be used in a command-line ");
        System.out.println("mode.  The command-line mode uses the following " +
            "command parameters:\n");
        System.out.println("-c command line (the other options are ignored " +
            "and the GUI is used if not set)");
        System.out.println("-d CPF effective start date");
        System.out.println("-D CPF effective end date");
        System.out.println("-q query start date");
        System.out.println("-Q query end date");
        System.out.println("-o off-nadir angle");
        System.out.println("-r off-nadir angle range type");
        System.out.println("-s save to database");
        System.out.println("-t control type");
        System.out.println("-w control type weight (for DOQ)");
        System.out.println("-h display this help\n");
        System.out.println("These are all followed by a space, the " +
            "selection for the command, and (if there ");
        System.out.println("are more commands) another space.  Here are " +
            "examples of the expected format for ");
        System.out.println("the various command selections, with comments " +
            "that are not part of the actual ");
        System.out.println("format in parentheses:\n");
        System.out.println("-c Y (could also be N)");
        System.out.println("-d 1-Apr-2000-00:00:00 (required if -c is set)");
        System.out.println("-D 30-Jun-2000-00:00:00 (required if -c is set)");
        System.out.println("-q 15-Mar-2000-00:00:00");
        System.out.println("-Q 23-Jul-2000-00:00:00");
        System.out.println("-o 0.5 (non-negative magnitude)");
        System.out.println("-r INWARD (could also be OUTWARD)");
        System.out.println("-s 1 (could also be 0)");
        System.out.println("-t DOQ (could also be GLS, BOTH)");
        System.out.println("-w 80 (integer 0..100)\n");
    }

    /**
     * Sensor Alignment main application.  Decide whether it's a GUI run or
     * a command-line run.  If it's a GUI run, start the GUI.  If it's a 
     * command-line run, parse the command-line options and run in non-GUI
     * mode 
     */
    public static final void main(String[] args) 
    {
        boolean command_line = false; // Flag for command line use; normally
                                      // false
        // Set up some defaults for the command line options
        String queryStartDate = ""; // User-selected query start date
        String queryEndDate = ""; // User-selected query end date
        boolean queryStartDateSelected = false; // Flag for query start date
        boolean queryEndDateSelected = false; // Flag for query end date
        String controlType = "DOQ"; // User-selected control type
        String offNadirAngle = "0.5"; // User-selected off-nadir angle maximum 
                            // magnitude
        String offNadirRangeType = "INWARD"; // User-selected angle type 
        String cpfStartDate = ""; // User-selected CPF start date
        String cpfEndDate = ""; // User-selected CPF start date
        String saveToDatabase = "0"; // User-selected flag for saving to DB 
        String controlTypeWeight = "80"; // User-selected DOQ weight

        // Scan through the command line arguments
        for (int i = 0; i < args.length; i++)
        {
            if (args[i].equals("-c") && i < args.length - 1)
            {
                i++;

                if ((args[i].equals("y")) || (args[i].equals("Y")) 
                    && i < args.length - 1)
                {
                    // Use as a command line rather than GUI application
                    command_line = true;
                }
            }
            else if (args[i].equals("-d") && i < args.length - 1)
            {
                i++;
                cpfStartDate = args[i];
            }
            else if (args[i].equals("-D") && i < args.length - 1)
            {
                i++;
                cpfEndDate = args[i];
            }
            else if (args[i].equals("-q") && i < args.length - 1)
            {
                i++;
                queryStartDateSelected = true;
                queryStartDate = args[i];
            }
            else if (args[i].equals("-Q") && i < args.length - 1)
            {
                i++;
                queryEndDate = args[i];
                queryEndDateSelected = true;
            }
            else if (args[i].equals("-o") && i < args.length - 1)
            {
                i++;
                offNadirAngle = args[i];
            }
            else if (args[i].equals("-r") && i < args.length - 1)
            {
                i++;
                offNadirRangeType = args[i];
            }
            else if (args[i].equals("-s") && i < args.length - 1)
            {
                i++;
                saveToDatabase = args[i];
            }
            else if (args[i].equals("-t") && i < args.length - 1)
            {
                i++;
                controlType = args[i];
            }
            else if (args[i].equals("-w") && i < args.length - 1)
            {
                i++;
                controlTypeWeight = args[i];
            }
            else if (args[i].equals("-h"))
            {
                printUsage();
                return;
            }
            else
            {
                printUsage();
                throw new IllegalArgumentException("\nInvalid command line "
                    + "option: " + args[i]);
            }
        }

        // Load properties from a custom properties file
        try 
        {
            initializeProperties();
        } 
        catch (Throwable error) 
        {
            // Catch exceptions in initializing properties
            System.err.println("Unable to initialize properties.  Please "
                + "check that required properties files are\ninstalled.");
            System.err.println("Details:");
            error.printStackTrace();
            System.exit(1);
        }

        if (command_line == true)
        {
            String authProperty = Constants.DAO_AUTH_IAS_PROD; 
            System.setProperty(authProperty, System.getenv(
                Constants.DAO_AUTH_IAS_PROD));

            // Set up the database login for a command line run
            try 
            {
                tryDBConnection(authProperty);
            } 
            catch (Exception ex) 
            {
                System.out.println(ex.getMessage());
                System.exit(1);
            }
            
            // Run the command line sensor alignment
            try 
            {
                CommandLine.runApplication(
                        queryStartDate,
                        queryEndDate,
                        controlType,
                        offNadirAngle,
                        offNadirRangeType,
                        queryStartDateSelected,
                        queryEndDateSelected,
                        cpfStartDate,
                        cpfEndDate,
                        saveToDatabase,
                        controlTypeWeight);
            } 
            catch (Exception ex) 
            {
                System.out.println("Error: " + ex.getMessage());
                System.exit(1);
            }
        }
        else
        {
            try 
            {
                // Create the GUI
                final SensorAlignment sensorAlignment = new SensorAlignment();
            } 
            catch (Throwable error) 
            {
                // Catch any exceptions or errors in constructing the main GUI 
                // class in case something's not installed where it should be
                System.err.println("Unable to create GUI.  Please check that "
                    + "all required jar and properties files are\ninstalled "
                    + "and classpath is correct.");
                System.err.println("Details:");
                error.printStackTrace();
                System.exit(1);
            }
        }
    }

    /**
     * If a calendar is updated, update corresponding date panels
     */
    @Override
    public void update(Observable o, Object arg) {
        if (calendarDialog != null && calendarDialog.getPanel() != null) {
            ObservableDate observableDate = (ObservableDate)o;
            int year = observableDate.getSelectedYear();
            int month = observableDate.getSelectedMonth();
            int day = observableDate.getSelectedDay();
            calendarDialog.getPanel().setDate(year, month, day);
            calendarDialog.setPanel(null);
        }
    }
}
