package sensor_alignment;

/**
 * Manage the parameter table user interface and user selections for 
 * alignment computations to be run on the parameter table data.  This 
 * includes the query to the database and subsequent calculations needed to
 * display the parameter table, displaying the parameter table, managing the
 * user selections for the parameter table, managing the user selections for 
 * computing the results of the Sensor Alignment calculations, and managing 
 * the Sensor Alignment calculations
 */

import gov.usgs.ias.gui.common_library.CalendarDialog;
import gov.usgs.ias.gui.common_library.ObservableDate;
import gov.usgs.ias.gui.common_library.DatePanel;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import java.awt.event.KeyEvent;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.awt.BorderLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.Point;
import javax.swing.JFrame;
import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.JLabel;
import javax.swing.JCheckBox;
import javax.swing.JSlider;
import javax.swing.JTable;
import javax.swing.table.TableColumn;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.JTableHeader;
import javax.swing.JViewport;
import javax.swing.event.ListSelectionEvent;
import javax.swing.ListSelectionModel;
import javax.swing.JScrollPane;
import javax.swing.JOptionPane;
import java.util.Date;
import java.util.Calendar;
import java.util.ArrayList;
import java.util.Observable;
import java.util.Observer;
import java.text.SimpleDateFormat;
import java.text.NumberFormat;
import java.text.DecimalFormat;

public class SensorAlignmentParamTableDialog extends JFrame implements 
    WindowListener, ActionListener, ItemListener, ChangeListener, Observer
{
    private DatePanel effectiveStartDatePanel; // Panel for CPF start date
    private DatePanel effectiveEndDatePanel; // Panel for CPF end date
    private SensorAlignment application; // This software application
    private int doqWeight = 80;             // default is 80% for DOQ
    private JTable table;           // Scrollable table contents 
    private JTable fixedTable;      // Non-scrollable table contents 
    private String[] columnNames = {"Date", "Time", "Select", "WRS Path",
        "WRS Row", "Roll'", "Pitch'", "Yaw'", "WO ID", "Geo Char ID", 
        "Off-Nadir Angle", "Control Type", "Weight", "Num GCPs", 
        "Outlier Threshold", "RMS Fit", "X", "Y", "Z", "Roll", "Pitch", "Yaw",
        "Roll-Y", "Y-Y", "Pitch-X", "X-X", "Z-Z", 
        "Roll-Roll", "Pitch-Pitch", "Yaw-Yaw"}; // Column headings

    private static final long serialVersionUID = 0; // for a compiler warning

    // Enumeration for the various columns of the tabular display 
    private static final int DATE_COL = 0;
    private static final int TIME_COL = 1; 
    private static final int SELECT_COL = 2; 
    private static final int WRS_PATH_COL = 3; 
    private static final int WRS_ROW_COL = 4; 
    private static final int ROLL_PRIME_COL = 5;
    private static final int PITCH_PRIME_COL = 6; 
    private static final int YAW_PRIME_COL = 7;
    private static final int WO_ID_COL = 8; 
    private static final int GEO_CHAR_ID_COL = 9; 
    private static final int OFF_NADIR_COL = 10; 
    private static final int CONTROL_TYPE_COL = 11; 
    private static final int WEIGHT_COL = 12; 
    private static final int NUM_GCPS_COL = 13; 
    private static final int OUTLIER_THRESHOLD_COL = 14; 
    private static final int RMS_FIT_COL = 15; 
    private static final int X_COL = 16; 
    private static final int Y_COL = 17; 
    private static final int Z_COL = 18; 
    private static final int ROLL_COL = 19; 
    private static final int PITCH_COL = 20; 
    private static final int YAW_COL = 21; 
    private static final int ROLL_Y_COL = 22; 
    private static final int Y_Y_COL = 23; 
    private static final int PITCH_X_COL = 24; 
    private static final int X_X_COL = 25; 
    private static final int Z_Z_COL = 26; 
    private static final int ROLL_ROLL_COL = 27; 
    private static final int PITCH_PITCH_COL = 28; 
    private static final int YAW_YAW_COL = 29;  // The last column 
    private static final int NUM_COLS = YAW_YAW_COL + 1; // Total number of
                                                // columns 
    private static final int NUM_FIXED_COLS = 2; // Date/time columns are fixed

    private ArrayList<PrecisionAlignment> alignRecords; // Precision alignment records
    private double[][] alignObs; // Integrated alignment estimates
    private Object[][] tableData; // Data to display on window's table
    private boolean saveToDatabaseFlag = false; // Flag for saving to DB
    private JCheckBox saveToDatabaseFlagCheckBox; // DB save choice
    private CalendarDialog calendarDialog; // Dialog for CPF dates 

    private SensorAlignmentResultsDialog sensorAlignmentResultsDialog;

    // These are operator selections from the calling Query screen
    private String controlTypeColStr = ""; // Control type in table 
    private boolean startDateSelectedFlag; // Flag for query start date selected
    private boolean endDateSelectedFlag; // Flag for query end date selected
    private String controlTypeChoiceStr = ""; // User-selected control type

    // These are the user-selected query date range
    private Date queryStartDateChoice;
    private Date queryEndDateChoice;

    /**
     * Build the user interface elements for the parameter table screen
     */
    public SensorAlignmentParamTableDialog
    (
        SensorAlignment application, // I: The application itself
        String controlType, // I: User-selected control type 
        Date queryStartDate, // I: User-selected DB query start date 
        Date queryEndDate, // I: User-selected DB query end date
        boolean startDateSelected, // I: Flag for query start date
        boolean endDateSelected, // I: Flag for query end date 
        double offNadirAngle, // I: User-selected maximum off-nadir angle
                     // magnitude 
        String offNadirRangeType // I: User-selected range type 
    ) throws Exception
    {
        super();

        JPanel mainPanel;       // Panel for whole window
        JPanel buttonPanel;     // Panel for buttons
        JButton computeButton;  // Button to compute sensor alignment
        JButton resetButton;    // Button to reset original row selections
        JButton closeButton;    // Button to close the window

        final int doqWeightMin = 0; // These are for percent of control 
        final int doqWeightMax = 100; // type weight assigned to DOQ.  The 
        JLabel controlTypeWeightLabel; // Label for the weight widget
        JLabel doqLabel;        // Label for the widget's DOQ side
        JLabel glsLabel;        // Label for the widget's GLS side
        JSlider controlTypeWeightSlider; // Slider for weight selection
        JLabel topSpaceLabel;   // Spacing at the top of the window
        JLabel calendarSpaceLabel; // Spacing to set calendars apart
        JScrollPane scrollPane; // Allows scrolling along the table
        TableColumn column = null; // Column in the table
        int[] columnWidths = {100, 80, 80, 90, 80, 120, 120, 120, 100, 100, 
            110, 100, 80, 90, 110, 60, 80, 80, 80, 100, 100, 100, 100, 100,
            100, 100, 100, 100, 100, 100}; // Width of table columns

        this.application = application;

        // Make the Query screen choices available to this screen
        controlTypeChoiceStr = controlType;
        startDateSelectedFlag = startDateSelected;
        endDateSelectedFlag = endDateSelected;
        if (startDateSelectedFlag)
        {
            queryStartDateChoice = queryStartDate;
        }
        if (endDateSelectedFlag)
        {
            queryEndDateChoice = queryEndDate;
        }

        // Read the precision alignment records based on the query choices
        try 
        {
            alignRecords = PrecisionAlignmentData.getAlignmentRecords(
                queryStartDate,
                queryEndDate,
                startDateSelected,
                endDateSelected,
                offNadirAngle,
                offNadirRangeType,
                controlType);
        } 
        catch (Exception ex) 
        {
            throw ex; 
        }
            
        // Set up date formats to extract the date acquired results and
        // format them for screen display
        SimpleDateFormat dateSdf = new SimpleDateFormat("yyyy-DDD");
        SimpleDateFormat timeSdf = new SimpleDateFormat("HH:mm:ss");
        
        // Put the values read from the DB into the window table structure
        tableData = new Object[alignRecords.size()][NUM_COLS];
        
        for (int row = 0; row < alignRecords.size(); row++)
        {
            tableData[row][DATE_COL] = 
                dateSdf.format(alignRecords.get(row).dateAcquired);
            tableData[row][TIME_COL] = 
                timeSdf.format(alignRecords.get(row).dateAcquired);           
            tableData[row][WRS_PATH_COL] = alignRecords.get(row).wrsPath; 
            tableData[row][WRS_ROW_COL] = alignRecords.get(row).wrsRow; 
            tableData[row][WO_ID_COL] = alignRecords.get(row).workOrderId; 
            tableData[row][GEO_CHAR_ID_COL] = 
                alignRecords.get(row).geometricCharId; 
            tableData[row][OFF_NADIR_COL] = 
                alignRecords.get(row).offNadirAngle; 
            tableData[row][CONTROL_TYPE_COL] = alignRecords.get(row).gcpSource;
            tableData[row][NUM_GCPS_COL] = alignRecords.get(row).numGcps; 
            tableData[row][OUTLIER_THRESHOLD_COL] = 
                alignRecords.get(row).outlierThreshold; 
            tableData[row][RMS_FIT_COL] = alignRecords.get(row).gcpRmsFit; 
            tableData[row][X_COL] = alignRecords.get(row).ephemerisBias[0]; 
            tableData[row][Y_COL] = alignRecords.get(row).ephemerisBias[1]; 
            tableData[row][Z_COL] = alignRecords.get(row).ephemerisBias[2]; 
            tableData[row][ROLL_COL] = alignRecords.get(row).attitudeBias[0]; 
            tableData[row][PITCH_COL] = alignRecords.get(row).attitudeBias[1]; 
            tableData[row][YAW_COL] = alignRecords.get(row).attitudeBias[2]; 
            tableData[row][ROLL_Y_COL] = 
                alignRecords.get(row).attitudeEphemerisCovariance[0][1];
            tableData[row][Y_Y_COL] = 
                alignRecords.get(row).ephemerisEphemerisCovariance[1][1];
            tableData[row][PITCH_X_COL] = 
                alignRecords.get(row).attitudeEphemerisCovariance[1][0];
            tableData[row][X_X_COL] = 
                alignRecords.get(row).ephemerisEphemerisCovariance[0][0];
            tableData[row][Z_Z_COL] = 
                alignRecords.get(row).ephemerisEphemerisCovariance[2][2];
            tableData[row][ROLL_ROLL_COL] = 
                alignRecords.get(row).attitudeAttitudeCovariance[0][0];
            tableData[row][PITCH_PITCH_COL] = 
                alignRecords.get(row).attitudeAttitudeCovariance[1][1];
            tableData[row][YAW_YAW_COL] = 
                alignRecords.get(row).attitudeAttitudeCovariance[2][2];

            // Initialize the Select column to be true if the row's control 
            // type (or "BOTH") was selected
            controlTypeColStr = (String)tableData[row][CONTROL_TYPE_COL];

            if (controlTypeChoiceStr.equals(controlTypeColStr) ||
                controlTypeChoiceStr.equals("BOTH"))
            {
                tableData[row][SELECT_COL] = true;
            }
            else
            {
                tableData[row][SELECT_COL] = false;
            }

            // Initialize the Weight column
            if (controlTypeColStr.equals("DOQ"))
            {
                tableData[row][WEIGHT_COL] = doqWeight;
            }
            else
            {
                if (controlTypeColStr.equals("GLS"))
                {
                    tableData[row][WEIGHT_COL] = 100 - doqWeight; 
                }
                else
                {
                    // This is an invalid scenario, so put an invalid 
                    // percentage in the column
                    tableData[row][WEIGHT_COL] = -1.0;
                }
            }
        }

        // Create integrated alignment observations
        alignObs = new double[alignRecords.size()][3]; 

        // Compute Roll', Pitch', and Yaw' before displaying them
        for (int row = 0; row < alignRecords.size(); row++)
        {
            alignObs[row] = alignRecords.get(row).processSingleDataSet();
            
            // Put the results in the table so they show in the GUI
            tableData[row][ROLL_PRIME_COL] = alignObs[row][0];
            tableData[row][PITCH_PRIME_COL] = alignObs[row][1];
            tableData[row][YAW_PRIME_COL] = alignObs[row][2];
        }

        // Done getting and calculating table column data 

        // Now set up the window's GUI
        setLayout(new BorderLayout());
        setSize(1200, 725);
        setTitle("Sensor Alignment - Parameter Table and Selections");

        mainPanel = new JPanel();
        GridBagLayout gridbag = new GridBagLayout();
        GridBagConstraints gbc = new GridBagConstraints();
        mainPanel.setLayout(gridbag);

        // Set up the fixed table
        fixedTable = new JTable(new FixedModel()) 
        {
            public void valueChanged(ListSelectionEvent e) 
            {
                super.valueChanged(e);
                checkSelection(true);
            }
        };
        fixedTable.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
        fixedTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

        // Set up the viewport for the fixed table (date and time)
        JViewport viewport = new JViewport();
        viewport.setView(fixedTable);
        viewport.setPreferredSize(fixedTable.getPreferredSize());
        Dimension viewportSize = new Dimension();
        Dimension fixedTableSize = fixedTable.getPreferredSize();
        double width = fixedTableSize.getWidth();
        double height = fixedTableSize.getHeight();
        viewportSize.setSize(width + 30, height);
        viewport.setPreferredSize(viewportSize);

        // Set up the fixed table column widths
        for (int i = 0; i < fixedTable.getColumnCount(); i++)
        {
            column = fixedTable.getColumnModel().getColumn(i);
            column.setMinWidth(columnWidths[i]);
        }

        // Set up fixed table header appearance
        JTableHeader fixedHeader = fixedTable.getTableHeader();
        fixedHeader.setBackground(Color.lightGray);

        // Set up the main (scrollable) table
        table = new JTable(new MyTableModel()) 
        {
            public void valueChanged(ListSelectionEvent e) 
            {
                super.valueChanged(e);
                checkSelection(false);
            }
        };

        // Set the table size
        table.setPreferredScrollableViewportSize(new Dimension(820, 400));

        // Without this the table columns fit in the table screen width (so 
        // they are unusably thin), and the bottom scroll bar isn't displayed
        table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
        table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        for (int i = 0; i < table.getColumnCount(); i++)
        {
            column = table.getColumnModel().getColumn(i);
            column.setPreferredWidth(columnWidths[i + NUM_FIXED_COLS]);
        }

        // Set up main table header appearance
        JTableHeader header = table.getTableHeader();
        header.setBackground(Color.yellow);

        // Create the scroll pane and add the table to it
        scrollPane = new JScrollPane(table);
        scrollPane.setRowHeaderView(viewport);
        scrollPane.setCorner(JScrollPane.UPPER_LEFT_CORNER,
            fixedTable.getTableHeader());

        // Set up the date range selections
        calendarDialog = new CalendarDialog((JFrame)this);
        effectiveStartDatePanel = new DatePanel(this, 
            "Start of Effective Date Range", calendarDialog);
        effectiveEndDatePanel = new DatePanel(this, 
            "End of Effective Date Range", calendarDialog);

        // Let the date panels get updates made in the calendar dialogs
        ObservableDate.getInstance().addObserver(this);

        // Initialize the CPF date range to be the current quarter
        SimpleDateFormat sdfMonth = new SimpleDateFormat("M");
        Calendar quarterCalendar = Calendar.getInstance();
        int lastDay; // Last day of the quarter's last month 
        int firstMonth = 0; // First month of the quarter
        int lastMonth = 0; // Last month of the quarter
        int month;

        // Find the current month.
        month = Integer.parseInt(sdfMonth.format(quarterCalendar.getTime()));

        // Find the first and last month of the current quarter
        if ((month >= 1) && (month <= 3)) // first quarter
        {
            firstMonth = Calendar.JANUARY;
            lastMonth = Calendar.MARCH;
        }
        else
        if ((month >= 4) && (month <= 6)) // second quarter
        {
            firstMonth = Calendar.APRIL;
            lastMonth = Calendar.JUNE;
        }
        else
        if ((month >= 7) && (month <= 9)) // third quarter
        {
            firstMonth = Calendar.JULY;
            lastMonth = Calendar.SEPTEMBER;
        }
        else
        if ((month >= 10) && (month <= 12)) // fourth quarter
        {
            firstMonth = Calendar.OCTOBER;
            lastMonth = Calendar.DECEMBER;
        }

        // Set start date panel to the first day of the quarter's first month
        effectiveStartDatePanel.setDate(quarterCalendar.get(Calendar.YEAR),
            firstMonth, 1);

        // Get the last day of the quarter's last month
        quarterCalendar.set(Calendar.MONTH, lastMonth);
        lastDay = quarterCalendar.getActualMaximum(Calendar.DATE);

        // Set the end date panel to the last day of the quarter's last month
        effectiveEndDatePanel.setDate(quarterCalendar.get(Calendar.YEAR),
            lastMonth, lastDay);

        // Set up the Save to Database Flag with default "unchecked" (no save)
        saveToDatabaseFlagCheckBox = new JCheckBox("Save to Database");
        saveToDatabaseFlagCheckBox.setSelected(false); 
        saveToDatabaseFlagCheckBox.setMnemonic(KeyEvent.VK_A);
        saveToDatabaseFlagCheckBox.addItemListener(this);

        // Set up the Control Type Weight slider and its label
        glsLabel = new JLabel("GLS"); 
        doqLabel = new JLabel("DOQ"); 
        controlTypeWeightLabel = new JLabel("Control Type Weight");
        controlTypeWeightSlider = new JSlider(JSlider.HORIZONTAL, doqWeightMin,
            doqWeightMax, doqWeight);
        controlTypeWeightSlider.addChangeListener(this);
        controlTypeWeightSlider.setMajorTickSpacing(10);
        controlTypeWeightSlider.setPaintTicks(true);
        controlTypeWeightSlider.setPaintLabels(true);
        controlTypeWeightSlider.setSnapToTicks(true);

        // Set up labels to take up vertical space
        topSpaceLabel = new JLabel(" ");
        calendarSpaceLabel = new JLabel(" ");

        // Use this for all components
        gbc.ipadx = 10;
        gbc.ipady = 6;

        // Put some space at the top
        gbc.gridx = 0;
        gbc.gridy = 0;
        gbc.gridwidth = 1;
        gbc.anchor = GridBagConstraints.LINE_START;
        mainPanel.add(topSpaceLabel, gbc);

        // Add the scroll pane to the panel
        gbc.gridx = 0;
        gbc.gridy = 1;
        gbc.gridwidth = 4;
        gbc.anchor = GridBagConstraints.CENTER;
        mainPanel.add(scrollPane, gbc);

        // Put some space before the calendars
        gbc.gridx = 0;
        gbc.gridy = 2;
        gbc.gridwidth = 1;
        gbc.anchor = GridBagConstraints.LINE_START;
        mainPanel.add(calendarSpaceLabel, gbc);

        // Add the effective start date panel to the main panel
        gbc.gridx = 0;
        gbc.gridy = 3;
        gbc.gridwidth = 1;
        gbc.anchor = GridBagConstraints.LINE_END;
        mainPanel.add(effectiveStartDatePanel, gbc);

        // Add the effective end date panel to the main panel
        gbc.gridx = 0;
        gbc.gridy = 4;
        gbc.gridwidth = 1;
        gbc.weightx = 0.5;
        gbc.anchor = GridBagConstraints.LINE_END;
        mainPanel.add(effectiveEndDatePanel, gbc);

        // Add the control type weight slider label
        gbc.gridx = 2;
        gbc.gridy = 3;
        gbc.gridwidth = 1;
        gbc.anchor = GridBagConstraints.PAGE_START;
        mainPanel.add(controlTypeWeightLabel, gbc);

        // Add the control type weight slider "GLS" label
        gbc.gridx = 1;
        gbc.gridy = 3;
        gbc.gridwidth = 1;
        gbc.anchor = GridBagConstraints.LINE_END;
        mainPanel.add(glsLabel, gbc);

        // Add the control type weight slider
        gbc.gridx = 2;
        gbc.gridy = 3;
        gbc.gridwidth = 1;
        gbc.fill = GridBagConstraints.HORIZONTAL;
        gbc.anchor = GridBagConstraints.PAGE_END;
        mainPanel.add(controlTypeWeightSlider, gbc);

        // Add the control type weight slider "DOQ" label
        gbc.gridx = 3;
        gbc.gridy = 3;
        gbc.gridwidth = 1;
        gbc.anchor = GridBagConstraints.LINE_START;
        mainPanel.add(doqLabel, gbc);

        // Add the database save checkbox to the main panel
        gbc.gridx = 2;
        gbc.gridy = 4;
        gbc.gridwidth = 1;
        gbc.anchor = GridBagConstraints.CENTER;
        mainPanel.add(saveToDatabaseFlagCheckBox, gbc);

        // Set up the Compute New Parameters, Reset, and Close buttons 
        buttonPanel = new JPanel();

        computeButton = new JButton("Compute New Parameters");
        computeButton.setMnemonic(KeyEvent.VK_N);
        computeButton.setToolTipText("Compute New Alignment Parameters");
        computeButton.addActionListener(this);
        buttonPanel.add(computeButton);

        resetButton = new JButton("Reset");
        resetButton.setMnemonic(KeyEvent.VK_R);
        resetButton.setToolTipText("Reset Selected Rows to Default");
        resetButton.addActionListener(this);
        buttonPanel.add(resetButton);

        closeButton = new JButton("Close");
        closeButton.setMnemonic(KeyEvent.VK_C);
        closeButton.setToolTipText("Close Dialog Box");
        closeButton.addActionListener(this);
        buttonPanel.add(closeButton);

        add(mainPanel,BorderLayout.NORTH);
        add(buttonPanel,BorderLayout.SOUTH);
    }

    /**
     * Keep the selected rows of the 2 tables (fixed and scrolled) in sync
     */
    private void checkSelection
    (
        boolean isFixedTable // I: Flag for the table being fixed or scrolled
    ) 
    {
        int fixedSelectedIndex = fixedTable.getSelectedRow();
        int selectedIndex = table.getSelectedRow();
        if (fixedSelectedIndex != selectedIndex) 
        {
            if (isFixedTable) 
            {
                table.setRowSelectionInterval(
                    fixedSelectedIndex,fixedSelectedIndex);
            } 
            else 
            {
                fixedTable.setRowSelectionInterval(
                    selectedIndex,selectedIndex);
            }
        }
    }

    // Handle the windowClosing event
    //-------------------------------
    public void windowClosing(WindowEvent e)  {    }

    // Dummy window event handlers for events that do not need handling
    //-----------------------------------------------------------------
    public void windowActivated(WindowEvent e) { }
    public void windowClosed(WindowEvent e) { }
    public void windowDeactivated(WindowEvent e) { }
    public void windowDeiconified(WindowEvent e) { }
    public void windowIconified(WindowEvent e) { }
    public void windowOpened(WindowEvent e) { }

    /**
     * Action performed event handler responds to user actions
     */
    public void actionPerformed(ActionEvent e) 
    {
        String command = e.getActionCommand(); // User-selected command
        boolean selected; // Flag for selection of row based on user choices
        int selectedCount; // Count of selected rows
        Calendar effectiveStartDateCalendar; // Start date calendar
        Calendar effectiveEndDateCalendar; // End date calendar

        Point loc = application.getLocationOnScreen();
        loc.y += 50;
        loc.x += 50;

        if (command.equals("Compute New Parameters"))
        {
            int numRows = table.getRowCount();

            // Find out how many rows are selected
            selectedCount = 0;
            for (int row = 0; row < numRows; row++) 
            {
                selected = 
                   (Boolean)table.getModel().getValueAt(row, SELECT_COL 
                       - NUM_FIXED_COLS);

                if (selected)
                {
                    selectedCount++;
                }
            }

            // Don't compute new parameters if no rows are selected
            if (selectedCount < 1)
            {
                JOptionPane.showMessageDialog(application,
                    "No rows were selected; not computing new parameters",
                    "Warning", JOptionPane.WARNING_MESSAGE); 
                return;
            }

            // Initialize a new result set with those results that are per-row
            // getting selectedCount entries
            Results res = new Results(selectedCount);

            // Fill in some of the user selection options so they can be
            // reported on later

            // The DOQ and GLS results will be presented as percentages
            res.doqWeight = doqWeight;
            res.glsWeight = 100 - doqWeight; 

            res.gcpSource = controlTypeChoiceStr;

            selectedCount = 0;
            for (int row = 0; row < numRows; row++) 
            {
                // Be careful to get/set the value of the model, not the table
                // screen view, since the columns can be moved around on 
                // the screen, but the model columns will not be moved around
                selected = (Boolean)table.getModel().getValueAt(row, SELECT_COL
                    - NUM_FIXED_COLS);

                if (selected)
                {
                    // Process a single alignment record, initializing result
                    // set, summing weights, and finding the maximum roll angle
                    res.processRow(alignRecords.get(row), alignObs[row], 
                        selectedCount);
                    selectedCount++;
                }
            }

            // Get the original query start and end dates in Date format
            res.acqQueryStartDateSelected = startDateSelectedFlag;
            if (startDateSelectedFlag)
            {
                res.acqQueryStartDate = queryStartDateChoice;
            }
            res.acqQueryEndDateSelected = endDateSelectedFlag;
            if (endDateSelectedFlag)
            {
                res.acqQueryEndDate = queryEndDateChoice;
            }
            
            // Compute best-fit alignment, residuals, and RMS statistics
            res.processMeasurements(selectedCount);

            // Calculate the updated alignment matrix
            res.calculateAlignmentMatrix();
            
            // Get the effective start and end date calendars
            effectiveStartDateCalendar = effectiveStartDatePanel.getDate();
            effectiveEndDateCalendar = effectiveEndDatePanel.getDate();

            // Translate the date values into Date type
            res.effectiveStartDate = effectiveStartDateCalendar.getTime();
            res.effectiveEndDate = effectiveEndDateCalendar.getTime();

            // Start the Results window
            sensorAlignmentResultsDialog = 
                new SensorAlignmentResultsDialog(saveToDatabaseFlag, res);
            sensorAlignmentResultsDialog.setLocation(loc);
            sensorAlignmentResultsDialog.setVisible(true);
        }
        else if (command.equals("Close"))
        {
            dispose();
        }
        else if (command.equals("Reset"))
        {
            // Reset Select column entry in all rows to initial value. Only
            // make changes when needed to limit the event processing traffic,
            // even though it makes the logic a bit more complicated
            int numRows = table.getRowCount();

            for (int row = 0; row < numRows; row++) 
            {
                // Be careful to get/set the value of the model, not the table
                // screen view, since the columns can be moved around on 
                // the screen, but the model columns will not be moved around
                selected = (Boolean)table.getModel().getValueAt(row, 
                    SELECT_COL - NUM_FIXED_COLS);
                controlTypeColStr = (String)table.getModel().getValueAt(row,
                    CONTROL_TYPE_COL - NUM_FIXED_COLS); 

                if (controlTypeChoiceStr.equals("BOTH"))
                {
                    if (selected == false)
                    {
                        table.getModel().setValueAt(Boolean.TRUE, row, 
                            SELECT_COL - NUM_FIXED_COLS);
                    }
                }
                else
                if (controlTypeChoiceStr.equals(controlTypeColStr) && !selected)
                {
                    table.getModel().setValueAt(Boolean.TRUE, row, SELECT_COL
                        - NUM_FIXED_COLS);
                }
                else
                if (!controlTypeColStr.equals(controlTypeColStr) && selected)
                {
                    table.getModel().setValueAt(Boolean.FALSE, row, SELECT_COL
                        - NUM_FIXED_COLS);
                }
            }
        }
    }

    /**
     * Item performed event handler responds to state changes caused by user
     * actions
     */
    public void itemStateChanged(ItemEvent e)
    {
        Object source = e.getItemSelectable();

        if (source == saveToDatabaseFlagCheckBox)
        {
            if (e.getStateChange() == ItemEvent.SELECTED)
            {
                saveToDatabaseFlag = true;                
            }
            if (e.getStateChange() == ItemEvent.DESELECTED)
            {
                saveToDatabaseFlag = false;                
            }
        }
    }

    /**
     * Change event handler responds to state changes caused by user actions
     */
    public void stateChanged(ChangeEvent e)
    {
        JSlider source = (JSlider)e.getSource();
        if (!source.getValueIsAdjusting())
        {
            doqWeight = (int)source.getValue();

            // Update the Weight column for the new weight balance. 
            // Be careful to get/set the value of the model, not the table
            // screen view, since the columns can be moved around on 
            // the screen, but the model columns will not be moved around
            int numRows = table.getRowCount();
            for (int row = 0; row < numRows; row++)
            {
                // Get the Control Type of the model row.
                controlTypeColStr = (String)table.getModel().getValueAt(row,
                    CONTROL_TYPE_COL - NUM_FIXED_COLS);

                // Set the updated Weight column of the model row based on the
                // Control Type and the new DOQ/GLS weights
                if (controlTypeColStr.equals("DOQ"))
                {
                    table.getModel().setValueAt(doqWeight, row, 
                        WEIGHT_COL - NUM_FIXED_COLS);
                }
                if (controlTypeColStr.equals("GLS"))
                {
                    table.getModel().setValueAt(100 - doqWeight, row, 
                        WEIGHT_COL - NUM_FIXED_COLS);
                }
            }
        }
    }

    /**
     * Specify a model for the fixed columns at the left side of the table.
     * There will be 2 tables put next to each other to look like 1 table
     */
    class FixedModel extends AbstractTableModel
    {
        /**
         * This is for a compiler warning
         */
        static final long serialVersionUID = 0;

        /**
         * Get the number of columns in the fixed table
         * @return   the number of columns in the table
         */
        public int getColumnCount()
        { 
            return NUM_FIXED_COLS;
        }

        /**
         * Get the number of rows in the table
         * @return   the number of rows in the table
         */
        public int getRowCount()
        { 
            return tableData.length;
        }

        /**
         * Get the specified column name.  The fixed table is first, so there 
         * is no need to account for an offset. 
         * @return   the name of the column 
         */
        public String getColumnName(int col) 
        {
            return (String)columnNames[col];
        }

        /**
         * Gets the value of an element in the table.
         * @return   the value of the specified table element 
         */
        public Object getValueAt(int row, int col) 
        {
            return tableData[row][col];
        }
    }

    /**
     * Specify a model for the table to control cell data types, editing, etc
     */
    class MyTableModel extends AbstractTableModel 
    {
        static final long serialVersionUID = 0; // For a compiler warning

        /**
         * Get the number of columns in the scrollable table
         * @return   the number of columns in the table
         */
        public int getColumnCount() 
        {
            return columnNames.length - NUM_FIXED_COLS;
        }

        /**
         * Get the number of rows in the table
         * @return   the number of rows in the table
         */
        public int getRowCount() 
        {
            return tableData.length;
        }

        /**
         * Get the specified column name, accounting for the offset for the 
         * fixed table
         * @return   the name of the column 
         */
        @Override
        public String getColumnName
        (
            int col // Index into columnNames, which matches table columns
        ) 
        {
            return columnNames[col + NUM_FIXED_COLS];
        }

        /**
         * Gets the value of an element in the table.  The elements have
         * different types depending on the column
         * @return   the value of the specified table element 
         */
        public Object getValueAt
        (
            int row, // I: table cell row
            int col // I: table cell column
        ) 
        {
            NumberFormat formatter = new DecimalFormat("#0.000");

            // Adjust for the fixed columns
            col += NUM_FIXED_COLS;
            switch (col)
            {
                case ROLL_PRIME_COL:
                case PITCH_PRIME_COL:
                case YAW_PRIME_COL:
                case OFF_NADIR_COL:
                case OUTLIER_THRESHOLD_COL:
                case RMS_FIT_COL:
                case X_COL:
                case Y_COL:
                case Z_COL:
                case ROLL_COL:
                case PITCH_COL:
                case YAW_COL:
                case ROLL_Y_COL:
                case Y_Y_COL:
                case PITCH_X_COL:
                case X_X_COL:
                case Z_Z_COL:
                case ROLL_ROLL_COL:
                case PITCH_PITCH_COL:
                case YAW_YAW_COL:
                    return formatter.format(tableData[row][col]);
                default:
                    return tableData[row][col];
            }
            
        }

        /**
         * JTable uses this method to determine the default renderer/editor
         * for each cell.  If we didn't implement this method, then the select
         * column would contain text ("true"/"false"), rather than a check box
         */
        @Override
        public Class getColumnClass
        (
            int c // I: column
        ) 
        {
            return getValueAt(0, c).getClass();
        }

        /**
         * Define what cells are editable.  Note that the data/cell address
         * is constant, no matter where the cell appears onscreen.  Column 
         * SELECT_COL is the Selected column, which is editable
         */
        @Override
        public boolean isCellEditable
        (
            int row, // I: Cell row
            int col // I: Cell column
        ) 
        {
            // Adjust for the fixed columns
            col += NUM_FIXED_COLS;

            if (col == SELECT_COL) 
            {
                return true;
            } 
            else 
            {
                return false;
            }
        }

        /**
         * This method is used to change the data in the table
         */
        @Override
        public void setValueAt
        (
            Object value, // I: Value to set the table cell to
            int row, // I: table row to set
            int col // I: table column to set
        ) 
        {
            tableData[row][col + NUM_FIXED_COLS] = value;
            fireTableCellUpdated(row, col);
        }
    }

    /**
     * If a calendar is updated, update corresponding date panels
     */
    @Override
    public void update(Observable o, Object arg) {
        if (calendarDialog != null && calendarDialog.getPanel() != null) {
            ObservableDate observableDate = (ObservableDate)o;
            int year = observableDate.getSelectedYear();
            int month = observableDate.getSelectedMonth();
            int day = observableDate.getSelectedDay();
            calendarDialog.getPanel().setDate(year, month, day);
            calendarDialog.setPanel(null);
        }
    }
}

