#!/bin/csh -f

#
# NAME: test_sensor_alignment.csh
#
# PURPOSE: Set up the environment for a Sensor Alignment test scenario, and 
# then run that test.  Sensor alignment will be run in a command line mode
# that exercises the math processing, database I/O, and report file creation
# modules of Sensor Alignment, but that does not exercise the GUI modules.
#

#
# Move from the sensor_alignment directory to the test directory.
#

cd test

#
# Set up the environment for the test.
#

set application = "sensor_alignment"
set diff_suffix = "diffs"
set test_dir = "$IAS_UNIT_TEST_DATA_DIR/ias/gps/${application}"
set sensor_alignment = "../dist/SensorAlignment.jar"
set classpath = "${sensor_alignment}:../../../javalib/ias/dist/ias.jar:$JAVA_LIB/ojdbc/ojdbc5.jar:$JAVA_LIB/jboss/log4j.jar"
set report_file = "sensor_alignment.rpt"
set cpf_group_file = "attitude_parameters_20000401_20000630.odl"
set report_file_diffs = ${report_file}.${diff_suffix}
set cpf_group_file_diffs = "${cpf_group_file}.${diff_suffix}" 

#
# Ensure the test environment variable is set.
#

if ( ! $?IAS_UNIT_TEST_DATA_DIR ) then
    echo "IAS_UNIT_TEST_DATA_DIR isn't set"
    exit 1
endif


# Check whether database testing has been enabled.  If it isn't enabled,
# this test can't be run, as it loads data into the database to start with.

#
if ( ! $?ENABLE_DB_TESTS ) then
    echo
    echo "Cannot run sensor alignment testing, ENABLE_DB_TESTS environment" \
        "variable isn't set"
    echo
    exit 0
endif

# Sanity check to make sure the database user ID/password are set in the
# environment (in the IAS_DB_L8_OLITIRS environment variable)
if ( ! $?IAS_DB_L8_OLITIRS ) then
    echo "IAS_DB_L8_OLITIRS is not currently set"
    exit 1
endif

#
# Initialize the test parameters.  Note: Use year 2055 for the acquisition
# dates to avoid conflict with other test data that might be in the database.
set command_line_flag = "Y"
set cpf_date_start = "1-Apr-2000-00:00:00"
set cpf_date_end = "30-Jun-2000-00:00:00"
set acq_date_start = "15-Mar-2055-00:00:00"
set acq_date_end = "24-Jul-2055-00:00:00"
set off_nadir_angle = "1.0"
set off_nadir_range_type = "INWARD"
set control_type = "BOTH"
set control_type_weight = "80"
set save_to_db = "1"
set db_results = "_db_results.txt"
set db_table = "oli_sensor_alignment"
set db_table_file = ${db_table}${db_results}
set db_table_file_diffs = ${db_table_file}.${diff_suffix}
set db_table_sql_script = "db_table_script.sql"
set test_data_sql_script = "populate_sensor_alignment_test_data.sql"

# Set up test data in database.
sqlplus $IAS_DB_L8_OLITIRS @$test_data_sql_script > & /dev/null


#
# Run the Sensor Alignment tool using the selected parameters.  This will
# create a report file, an ASCII CPF group file, and database updates.
#
echo "Testing Sensor Alignment."
java -Doracle.net.tns_admin=$TNS_ADMIN \
    -cp ${classpath} sensor_alignment.SensorAlignment \
    -c ${command_line_flag} -d ${cpf_date_start} \
    -D ${cpf_date_end} -s ${save_to_db} -q ${acq_date_start} \
    -Q ${acq_date_end} -o ${off_nadir_angle} -r ${off_nadir_range_type} \
    -t ${control_type} -w ${control_type_weight}
if ( $status != 0 ) then
    echo "Failed to run $sensor_alignment."
    exit 1
endif

#
# Compare the CPF group file to the unit test version.  Any difference will be
# treated as a failure.
#

echo "Comparing CPF group file $cpf_group_file to $test_dir/$cpf_group_file."
diff ${test_dir}/${cpf_group_file} ${cpf_group_file} > ${cpf_group_file_diffs}
if ( $status != 0 ) then
    echo "$cpf_group_file and $test_dir/$cpf_group_file files do not match."
    exit 1
endif

#
# Prepare to compare the Sensor Alignment report file to the unit test version.
# Exclude the first line of the report header, since that includes a timestamp 
# that we don't want to compare.  Also exclude the next 4 lines, since they
# include a version number that we don't want to have to update repeatedly
# in the test data.
#

sed -e 1,6d ${report_file} > ${report_file}_tmp
cp ${report_file} ${report_file}_sav
mv ${report_file}_tmp ${report_file}

#
# Compare the report file to the unit test version.  Any difference will be
# treated as a failure.
#

echo "Comparing report file $report_file to $test_dir/$report_file."
diff ${test_dir}/${report_file} ${report_file} > ${report_file_diffs}
if ( $status != 0 ) then
    echo "$report_file and $test_dir/$report_file files do not match."
        
    exit 1
endif

#
# Restore the original report file so we don't have an extra file on disk.
#

mv ${report_file}_sav ${report_file}

#
# Check the DB updates.  ------------------------------------------------
#

#
# Create a script to get the database row created by the Sensor Alignment run. 
# This is done here so that a separate script will not have to be maintained.  
# It will be used then removed at the end.  It is assumed that the row with
# the most recent (i.e. highest) DATE_PROCESSED is the correct one, that no
# Sensor Alignment runs happened in the interval, and no Sensor Alignment runs 
# entered a future DATE_PROCESSED. 
#

echo "set term off" > $db_table_sql_script
echo "set echo on" >> $db_table_sql_script
echo "set trimout on" >> $db_table_sql_script
echo "set trimspool on" >> $db_table_sql_script
echo "set heading off" >> $db_table_sql_script
echo "set feedback off" >> $db_table_sql_script
echo "set pause off" >> $db_table_sql_script
echo "set pagesize 0" >> $db_table_sql_script
echo "set linesize 500" >> $db_table_sql_script
echo "set space 0" >> $db_table_sql_script
echo "set colsep |" >> $db_table_sql_script
echo "set tab off" >> $db_table_sql_script
echo "set verify off" >> $db_table_sql_script
echo "spool ./${db_table_file}" >> $db_table_sql_script
echo "select processing_center,off_nadir_angle," >> $db_table_sql_script
echo "gcp_source,doq_weight,gls_weight," >> $db_table_sql_script
echo "number_of_scenes_used,data_start_date," >> $db_table_sql_script
echo "data_stop_date,roll_alignment,pitch_alignment," >> $db_table_sql_script
echo "yaw_alignment,roll_residual_rmse," >> $db_table_sql_script
echo "pitch_residual_rmse,yaw_residual_rmse," >> $db_table_sql_script
echo "alignment_effective_date_start," >> $db_table_sql_script
echo "alignment_effective_date_stop," >> $db_table_sql_script
echo "alignment_matrix_0,alignment_matrix_1," >> $db_table_sql_script
echo "alignment_matrix_2,alignment_matrix_3," >> $db_table_sql_script
echo "alignment_matrix_4,alignment_matrix_5," >> $db_table_sql_script
echo "alignment_matrix_6,alignment_matrix_7," >> $db_table_sql_script
echo "alignment_matrix_8 from oli_sensor_alignment " >> $db_table_sql_script
echo "where date_processed=(select max(date_processed) " >> $db_table_sql_script
echo "from oli_sensor_alignment);" >> $db_table_sql_script
echo "spool off" >> $db_table_sql_script
echo "exit" >> $db_table_sql_script

#
# Query the database.
#

sqlplus $IAS_DB_L8_OLITIRS @$db_table_sql_script > & /dev/null

#
# Compare the results.
#
 
diff ${test_dir}/${db_table_file} . > ${db_table_file_diffs}
if ( $status != 0 ) then
    echo "Database results in file ${db_table_file} differ from" \
        "expected results"
    exit 1
endif

#
# Clean up some files.
#

rm ./${db_table_file} ./${db_table_sql_script}
#
# Exit with success.
#

exit 0
