#!/usr/bin/env python3
""" Master script for the TIRS SSM estimation algorithm.
"""

import argparse
import sys
from ias.log import logger
from ias.db import database as db
from ias.misc import satellite_attributes as sat_attr
from fit_tirs_ssm_switch_event import FitSwitchEvent, FitSwitchEventError
from generate_ssm_position_table import (SsmPosTableGenFrontEnd,
                                         SsmPosTableGenError)
from manage_tirs_ssm_switch_events import ManageSwitchEvents, ManageEventError
from manage_tirs_ssm_models import ManageSsmModels, ManageModelError
from load_estimated_tirs_ssm_positions import LoadEstimatedTirsSsmPositions
from place_lpgs_orders_for_time_period import PlaceOrdersFrontEnd
from ssm_exceptions import SsmException, ParameterError, EventNotFoundError
from ssm_utilities import connect_to_database
from ias.misc.const import EXIT_SUCCESS, EXIT_FAILURE


def parse_parameters(arguments):
    """ Parse the command line arguments.  For each of the sub-commands, add
        the parser definition.  Then call the parse_args routine.

        Parameters:
            arguments: List of command line arguments to parse.  This is passed
                       in as an argument instead of defaulting to sys.argv in
                       the ArgumentParser to allow unit testing to provide
                       lists to test.

        Returns:
            Argument object with the parameters as members.
    """
    # Create the main argument parser
    argparser = argparse.ArgumentParser(
        description='Estimate the TIRS SSM Position')

    # Use sub-parsers for each of the sub-commands
    subparsers = argparser.add_subparsers(
        title='sub-commands', help='available sub-commands')

    # Work around a bug in ArgumentParser that doesn't allow a subcommand to be
    # required via the add_subparsers method
    subparsers.required = True
    subparsers.dest = 'task'

    # Create a subparser for managing switch events
    switch_event_parser = subparsers.add_parser(
        'event', help='Manage switch events')
    switch_event_parser.set_defaults(task='manage_events')
    ManageSwitchEvents.add_command_line_parameters(switch_event_parser)

    # Create a subparser for managing models
    manage_models_parser = subparsers.add_parser(
        'model', help='Manage SSM models')
    manage_models_parser.set_defaults(task='manage_models')
    ManageSsmModels.add_command_line_parameters(manage_models_parser)

    # Create a subparser for fitting switch events
    fit_events_parser = subparsers.add_parser(
        'fit', help='Fit a switch event')
    fit_events_parser.set_defaults(task='fit_event')
    FitSwitchEvent.add_command_line_parameters(fit_events_parser)

    # Create a subparser for loading estimated positions
    load_positions_parser = subparsers.add_parser(
        'load', help='Load estimated SSM positions into the database')
    load_positions_parser.set_defaults(task='load_positions')
    LoadEstimatedTirsSsmPositions.add_command_line_parameters(
        load_positions_parser)

    # Create a subparser for generating estimated positions
    gen_positions_parser = subparsers.add_parser(
        'generate', help='Generate estimated SSM positions from active '
        'models')
    gen_positions_parser.set_defaults(task='gen_positions')
    SsmPosTableGenFrontEnd.add_command_line_parameters(
        gen_positions_parser)

    # Create a subparser for placing LPGS orders
    place_orders_parser = subparsers.add_parser(
        'order', help='Place orders in LPGS for a specified time range')
    place_orders_parser.set_defaults(task='place_orders')
    PlaceOrdersFrontEnd.add_command_line_parameters(
        place_orders_parser)

    # Return the parsed arguments
    return argparser.parse_args(arguments)


def main():
    """ Main script routine
    """
    # Parse any command line arguments, skipping the first one which is the
    # application name
    args = parse_parameters(sys.argv[1:])

    # Initialize the logger
    log = logger.Logger("estimate_tirs_ssm_position.py")
    log.info("Initiated")

    # Set up the satellite attributes for Landsat 8
    sat_attr.set_default(sat_attr.get_satellite_id_from_number(8))

    # Connect to the database
    db_connection = connect_to_database()

    # Assume the operation will complete successfully
    exit_status = EXIT_SUCCESS

    if args.task == 'manage_events':
        try:
            # Direct the manage switch events options to that class
            manage_events = ManageSwitchEvents(db_connection)
            manage_events.run(args)
            log.info("Completed successfully")
        except ManageEventError as ex:
            log.error("Failed to perform the requested operation: {0}".
                      format(ex))
            exit_status = EXIT_FAILURE

    elif args.task == 'manage_models':
        try:
            # Direct the manage models options to that class
            manage_models = ManageSsmModels(db_connection)
            manage_models.run(args)
            log.info("Completed successfully")
        except ManageModelError as ex:
            log.error("Failed to perform the requested operation: {0}".
                      format(ex))
            exit_status = EXIT_FAILURE

    elif args.task == 'fit_event':
        try:
            # Direct the fit events options to that class
            fit_event = FitSwitchEvent(db_connection)
            fit_event.run(args)
            log.info("Completed successfully")
        except FitSwitchEventError as ex:
            log.error("Failed to perform the requested operation: {0}".
                      format(ex))
            exit_status = EXIT_FAILURE

    elif args.task == 'load_positions':
        try:
            # Direct the load positions options to that class
            load_positions = LoadEstimatedTirsSsmPositions(db_connection)
            load_positions.run(args)
            log.info("Completed successfully")
        except (SsmException, FileNotFoundError) as ex:
            log.error("Failed to perform the requested operation: {0}".
                      format(ex))
            exit_status = EXIT_FAILURE

    elif args.task == 'gen_positions':
        try:
            # Direct the generate positions options to that class
            gen_positions = SsmPosTableGenFrontEnd(db_connection)
            gen_positions.run(args)
            log.info("Completed successfully")
        except SsmPosTableGenError as ex:
            log.error("Failed to perform the requested operation: {0}".
                      format(ex))
            exit_status = EXIT_FAILURE

    elif args.task == 'place_orders':
        try:
            # We need to connect as a different user than the other classes
            wo_db = connect_to_database(True)

            # Direct the place order options to that class
            place_orders = PlaceOrdersFrontEnd(wo_db)
            place_orders.run(args)
            log.info("Completed successfully")
        # About the only thing that can go wrong, besides Oracle errors, is
        # bad parameter values
        except (ParameterError, EventNotFoundError) as ex:
            log.error("Failed to perform the requested operation: {0}".
                      format(ex))
            exit_status = EXIT_FAILURE
        finally:
            wo_db.close(db.DISCONNECT_ROLLBACK)

    # Close the database connection.  Any updates should already be committed.
    db_connection.close(db.DISCONNECT_ROLLBACK)

    log.close()

    sys.exit(exit_status)


# Runs main if called from command line
if __name__ == "__main__":
    main()
