#!/usr/bin/env python3
###############################################################################
# Name: estimated_ssm_position.py
#
# Description: Contains classes pertaining to EstimatedSsmPositions.
#   This includes the EstimatedSsmPosition class, a EstimatedSsmPositionFileIO
#   class for creating EstimatedSsmPositions from specially formatted files,
#   and an EstimatedSsmPositionTable class for interfacing with the position
#   estimates table.
###############################################################################

from timestamp import Timestamp
from datetime import datetime
from ssm_pos_model import ModelQuality
from database_objects import DatabaseObjects


class EstimatedSsmPosition:
    """ Represents an EstimatedSsmPosition object, which is composed of a
        timestamp object, position number, quality flag, date_added and
        date_disabled fields.
    """
    def __init__(
        self, timestamp, position, quality_flag, date_added=None,
            date_disabled=None):
        """ Create an estimated SSM position object.

            Parameters:
                timestamp: timestamp for the record
                position: SSM encoder position (int)
                quality_flag: Associated ModelQuality
                date_added: Datetime the record was added to the database
                date_disabled: Datetime the record as disabled in the database

            Notes:
                - date_added and date_disabled usually only have meaning for
                  positions retrieved by the EstimatedSsmPositionTable object.
        """
        self.timestamp = timestamp
        self.position = position
        self.quality_flag = quality_flag
        self.date_added = date_added
        self.date_disabled = date_disabled

    def __str__(self):
        """ Overloaded string conversion. Creates a list of all fields and
            returns that list joined on a space.
        """
        return ("Timestamp: {0}, Position: {1}, Quality: {2}, Date Added: {3},"
                " Date Disabled: {4}".format(
                    self.timestamp, self.position, str(self.quality_flag),
                    self.date_added, self.date_disabled))

    def __repr__(self):
        """ Return the representation of the object """
        return (("EstimatedSsmPosition(Timestamp({0}), {1}, {2}, {3}, " +
                 "{4})").format(self.timestamp, self.position,
                                self.quality_flag, self.date_added,
                                self.date_disabled))

    def __eq__(self, other):
        """ Overloaded equivalence operator. Returns true if all fields in self
            match all corresponding fields in other, with the exception that
            the  date_added/date_disabled fields are not compared.
        """
        return (
            self.timestamp == other.timestamp and
            self.position == other.position and
            self.quality_flag.value == other.quality_flag.value)


class EstimatedSsmPositionFile:
    """ This class reads and writes Estimated SSM position files.
    """

    def __init__(self, filename, quality_flag=None):
        """ Constructor for the EstimatedSsmPositionFile.

            Parameters:
                filename: Name for the file
                quality_flag: Optional parameter that is only needed when
                    reading an older format that does not include the quality
        """
        self.filename = filename
        self.quality_flag = quality_flag
        self.creation_date = None
        self.start_date = None
        self.end_date = None

    def _parse_parameter(self, line, name):
        """ Helper method to parse "parameter = value" lines.

            Parameters:
                line: line of data to parse
                name: parameter name
        """
        split_line = line.partition('=')

        if split_line[0].strip() != name:
            raise IOError("Parameter name {0} not found in '{1}'".format(
                name, line))
        return split_line[2].strip()

    def _get_file_version(self):
        """ The first line of a file with version 1 or higher of the format
            should read "FORMAT_VERSION = #". Version 0 files do not have this

            Returns: Version number
        """
        with open(self.filename, 'r') as f:
            version_line = f.readline()

        # Look for the FORMAT_VERSION parameter.  If not found, assume it is
        # version 0
        try:
            version = self._parse_parameter(version_line, "FORMAT_VERSION")
        except:
            version = 0

        # Try to interpret the version as an integer
        try:
            version = int(version)
        except:
            raise IOError("FORMAT_VERSION parameter is invalid")

        return version

    def write(self, estimated_ssm_positions):
        """ Writes each position in estimated_ssm_positions to a format1 file.
            The timestamps of the first and last occuring estimated ssm
            positions are used as the start and end dates for the file. The
            creation_date field for the file will be set to now.

            The list will be sorted by timestamp before it is used

            Parameters:
                estimated_ssm_positions: List of EstimatedSsmPosition objects
                    to be written to the file

            Returns:
                None
        """
        # Throw an exception if we were sent an empty list
        num_positions = len(estimated_ssm_positions)
        if num_positions == 0:
            raise ValueError("Empty list of EstimatedSsmPosition objects")

        # Sort positions in ascending order by their timestamps
        sorted_positions = sorted(
            estimated_ssm_positions, key=lambda x: x.timestamp)

        # Set start, end, and creation times
        start_time = sorted_positions[0].timestamp
        end_time = sorted_positions[num_positions-1].timestamp
        creation_date = Timestamp.from_datetime(datetime.now())

        # Write file
        with open(self.filename, 'w') as f:
            # Write file header
            f.write('FORMAT_VERSION = 1\n')
            f.write('CREATION_DATE = {0}\n'.format(creation_date))
            f.write('START_DATE = {0}\n'.format(start_time))
            f.write('END_DATE = {0}\n'.format(end_time))
            f.write('NUM_RECORDS = {0}\n'.format(num_positions))
            f.write('# YEAR DOY SECONDS EST_SSM_POS QUALITY\n')

            # Format and write all positions in the list
            for position in sorted_positions:
                f.write('{0:4d} {1:3d} {2:5d} {3:8d} {4}\n'.format(
                    position.timestamp.year, position.timestamp.doy,
                    position.timestamp.seconds, position.position,
                    position.quality_flag.value))

    def read(self, start_time=None, end_time=None):
        """ Reads through the file, creating a list of EstimatedSsmPositions
            from the entries within.

            Parameters:
                start_time: Timestamp indicating when to begin loading the
                    positions. If this is None then all entries from the
                    beginning of the file up to end_time will be read.
                end_time: Timestamp indicating when to stop reading the
                    positions. If set to None then all entries from start_time
                    to the end of file will be read.

            Returns: List of EstimatedSsmPositions
        """
        version = self._get_file_version()
        if version == 0:
            return self._read_format_0(start_time, end_time)
        elif version == 1:
            return self._read_format_1(start_time, end_time)
        else:
            raise IOError("Unsupported format version: {0}".format(version))

    def _read_format_0(self, start_time, end_time):
        """ Helper specifically for reading through format 0 files. Used by
            read. Format 0 files do not have a quality flag on the end of their
            lines. EstimatedSsmPositions created within this method will be
            initialized with self.quality_flag.

            Parameters:
                See read method

            Returns: List of EstimatedSsmPositions
        """
        ssm_pos_list = []
        count = 1
        ending_line = set([-1])
        properly_terminated = False

        if self.quality_flag is None:
            self.quality_flag = ModelQuality.final

        with open(self.filename, 'r') as f:
            for line in f:
                try:
                    int_list = list(map(int, map(float, line.split())))
                except:
                    raise IOError("Line {0} in file {1} is invalid".format(
                        count, self.filename))

                if len(int_list) != 4:
                    raise IOError("Line {0} in file {1} does not have 4 "
                                  "values".format(count, self.filename))

                # A line with all -1's is the ending terminator
                if set(int_list) == ending_line:
                    properly_terminated = True
                    break

                # Create a position record
                record = EstimatedSsmPosition(
                    Timestamp(int_list[0], int_list[1], int_list[2]),
                    int_list[3], ModelQuality(self.quality_flag))

                # Only save the record if it meets the time constraints
                if start_time is not None and record.timestamp < start_time:
                    continue
                if end_time is not None and record.timestamp >= end_time:
                    continue
                ssm_pos_list.append(record)

                # Track the line number in the file
                count += 1

        # If the file wasn't properly terminated, it is an error
        if not properly_terminated:
            raise IOError("File {0} was not properly terminated with a record "
                          "of -1's".format(self.filename))

        return ssm_pos_list

    def _read_format_1(self, start_time, end_time):
        """ Helper specifically for reading through format 1 files. Used by
            read. Lines in a format 1 file have a quality flag on the end.
            This flag is what will be used when creating EstimatedSsmPositions,
            and the self.quality_flag will be completely ignored.

            Parameters:
                See read method

            Returns: List of EstimatedSsmPositions
        """
        ssm_pos_list = []
        count = 0
        with open(self.filename, 'r') as f:

            # version
            count += 1
            line = f.readline()

            try:
                # Creation_date
                count += 1
                line = f.readline()
                date_string = self._parse_parameter(line, "CREATION_DATE")
                self.creation_date = Timestamp.from_string(date_string)

                # Start_date
                count += 1
                line = f.readline()
                date_string = self._parse_parameter(line, "START_DATE")
                self.start_date = Timestamp.from_string(date_string)

                # End_date
                count += 1
                line = f.readline()
                date_string = self._parse_parameter(line, "END_DATE")
                self.end_date = Timestamp.from_string(date_string)

                # Number of records
                count += 1
                line = f.readline()
                num_records = self._parse_parameter(line, "NUM_RECORDS")
                num_records = int(num_records)
            except Exception as error:
                # Add the file name to the error message
                error.args = ("Error reading file {0}: {1}".format(
                    self.filename, error.args[0]),)
                raise error

            # Verify the column header comment is found
            count += 1
            line = f.readline()
            if line[0] != "#":
                raise IOError("Header line in file {0} does not start "
                              "with a # character".format(self.filename))

            for i in range(num_records):
                count += 1
                line = f.readline()

                # Parse the line
                try:
                    int_list = list(map(int, line.split()))
                except:
                    raise IOError(("Line {0} '{1}' in file {2} is " +
                                   "invalid").format(count, line,
                                                     self.filename))

                if len(int_list) != 5:
                    raise IOError("Line {0} '{1}' in file {2} does not "
                                  "have 5 values".format(
                                    count, line, self.filename))

                # Create a new position record
                record = EstimatedSsmPosition(
                    Timestamp(int_list[0], int_list[1], int_list[2]),
                    int_list[3], ModelQuality(int_list[4]))

                # Only save the record if it meets the time constraints
                if start_time is not None and record.timestamp < start_time:
                    continue
                if end_time is not None and record.timestamp >= end_time:
                    continue
                ssm_pos_list.append(record)

        return ssm_pos_list


class EstimatedSsmPositionTable:
    """ This class is for handling interactions with the position estimates
        table.  It contains methods for inserting lists of
        EstimatedSsmPosition objects, disabling entries within the table, and
        querying the table for EstimatedSsmPositions.
    """
    _table_columns = [
            "MODEL_ID", "YEAR", "DAY_OF_YEAR", "SECONDS_OF_DAY",
            "ENCODER_POSITION", "QUALITY_FLAG", "DATE_ADDED", "DATE_DISABLED"]

    def __init__(self, db_connection):
        """ db_conn is of type DatabaseConnection
        """
        self.db = db_connection
        self.table_name = DatabaseObjects.get_name("position_estimates_table")

    def insert_ssm_positions(self, model_id, ssm_positions):
        """ Inserts a list of EstimatedSsmPositions into the position estimates
            table. Each position will be inserted with the same model id. Note
            the date_added and date_disabled fields of ssm_positions will not
            be used in the insert statement. All new EstimatedSsmPositions will
            be inserted with date_added fields of 'now', and date_disabled set
            to NULL.

            Parameters:
                model_id: Model ID of the SSM positions being inserted
                ssm_positions: list containing EstimatedSsmPositions
        """
        bind_columns = list(self._table_columns)

        # The values of these columns will be set automatically
        bind_columns.remove("DATE_ADDED")
        bind_columns.remove("DATE_DISABLED")
        bind_indicies = [":" + str(x) for x in range(1, len(bind_columns) + 1)]

        cursor = self.db.get_cursor()
        insert_statement = (
            "insert into {0} ({1}) values ({2}, sysdate, NULL)").format(
                self.table_name, ", ".join(self._table_columns),
                ", ".join(bind_indicies))
        cursor.prepare(insert_statement)
        for ssm_position in ssm_positions:
            cursor.execute(
                None, [
                    model_id, ssm_position.timestamp.year,
                    ssm_position.timestamp.doy, ssm_position.timestamp.seconds,
                    ssm_position.position, ssm_position.quality_flag.value])
        cursor.close()

    def disable_ssm_positions(self, model_ids):
        """ Sets the date_disabled column to now for all rows with a model id
            in model_ids.

            Parameters:
                model_ids: List of model IDs to disable
        """
        bind_indicies = [":" + str(x) for x in range(1, len(model_ids) + 1)]
        update_statement = """
            update {0}
            set date_disabled=sysdate
            where model_id in ({1}) and date_disabled is null
            """.format(self.table_name, ", ".join(bind_indicies))

        cursor = self.db.get_cursor()
        cursor.prepare(update_statement)
        cursor.execute(None, model_ids)
        cursor.close()

    def get_ssm_positions(self, model_id):
        """ Creates a list of EstimatedSsmPositions from all entries with a
            matching model_id and returns the list.

            Parameters:
                model_id: Model ID of the SSM position records to retrieve

            Returns:
                List of EstimatedSsmPosition objects
        """
        # Build the query statement
        columns = list(self._table_columns)
        columns.remove("MODEL_ID")
        select_statement = ("select {0} from {1} where model_id = {2} order"
                            " by year, day_of_year, seconds_of_day").format(
                            ", ".join(columns), self.table_name, model_id)

        # Fetch all the matching rows
        cursor = self.db.get_cursor()
        cursor.execute(select_statement)
        rows = cursor.fetchall()
        cursor.close()

        # Build the list of estimated positions
        estimated_ssm_positions = []
        for row in rows:
            estimated_ssm_positions.append(EstimatedSsmPosition(
                Timestamp(row[0], row[1], row[2]), row[3],
                ModelQuality(row[4]), row[5], row[6]))

        return estimated_ssm_positions

    def get_all_ssm_positions(self, enabled=None):
        """ Creates a list of EstimatedSsmPositions from all entries with
            the specified status and returns the list.

            Parameters:
                enabled: Optional parameter that if passed will filter the
                         positions returned. The following values are valid:
                            True:  Only enabled records
                            False: Only disabled records
                            Not passed/None: All records

            Returns:
                List of tuples with the format:
                    (model_id, EstimatedSsmPosition)
        """
        # Build the where clause needed
        if enabled is None:
            where_clause = ""
        elif enabled:
            where_clause = "where date_disabled IS NULL"
        else:
            where_clause = "where date_disabled IS NOT NULL"

        # Build the query statement
        select_statement = ("select {0} from {1} {2} order"
                            " by year, day_of_year, seconds_of_day").format(
                            ", ".join(self._table_columns), self.table_name,
                            where_clause)

        # Fetch all the matching rows
        cursor = self.db.get_cursor()
        cursor.execute(select_statement)
        rows = cursor.fetchall()
        cursor.close()

        # Build the list of estimated positions
        estimated_ssm_positions = []
        for row in rows:
            pos = EstimatedSsmPosition(
                Timestamp(row[1], row[2], row[3]), row[4],
                ModelQuality(row[5]), row[6], row[7])
            estimated_ssm_positions.append((row[0], pos))

        return estimated_ssm_positions
