#!/usr/local/bin/python3
from collections import namedtuple
from ssm_exceptions import TimestampError
from ssm_pos_model import SsmPosModel
from timestamp import Timestamp
import matplotlib
import os


class FitResults():
    """ Class to encapsulate the results of the SSM model fit
    """

    def __init__(self, results_filename):
        """ Construct a new FitResults object.

            Parameters:
                results_filename: Results file from the SSM model fit
        """
        self.plot_filename = None
        self.results_filename = results_filename
        self.report_file = []

        # Using named tuples to store information for the results in an
        # attempt (albeit clumsy) to keep everything somewhat organized
        self.Model = namedtuple(
            "Model", ["a0", "a1", "a2", "a3", "tau1", "tau2", "tau3",
                      "slope", "encoder_nadir_pos"])
        self.new_model = None
        self.orig_model = None

        self.alg_version = None
        self.event_id = None
        self.model_id = None
        self.start_time = None
        self.stop_time = None

        self.image_days_from_switch = []
        self.image_measured_position = []
        self.image_modeled_position = []
        self.image_encoder_diff = []

        self.telemetry_days_from_switch = []
        self.telemetry_measured_position = []
        self.telemetry_modeled_position = []
        self.telemetry_encoder_diff = []

        # Used for SSM Estimated Positions
        self.ssm_positions_est_days = []
        self.ssm_positions_est_position = []

        self._retrieve_results()

    def _plot_figure_1(self, plt):
        """ Creates first figure
        """
        # 311 = num rows, num columns, figure num
        fig1 = plt.subplot(311)
        fig1.yaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter(
            '%d'))
        plt.plot(self.telemetry_days_from_switch,
                 self.telemetry_modeled_position,
                 color='black', linewidth=2.5,
                 label="Modeled Telemetry and Image Position")
        plt.plot(self.image_days_from_switch,
                 self.image_modeled_position,
                 # Add a black line to the plot
                 color='black', linewidth=2.5)
        plt.plot(self.telemetry_days_from_switch,
                 self.telemetry_measured_position,
                 # Plot using green x's = gx
                 'gx', label="Observed Telemetry Position")
        plt.plot(self.image_days_from_switch,
                 self.image_measured_position,
                 # Plot using red x's = rx
                 'rx', label="Observed Image Position")
        plt.legend(loc='best', prop={'size': 8}, numpoints=1)
        plt.title("SSM Encoder - Observed and Modeled Positions", y=1.08)
        plt.xlabel("Date Difference (Days)")

    def _plot_figure_2(self, plt):
        """ Creates second figure
        """
        # 312 = num rows, num columns, figure num
        fig2 = plt.subplot(312)
        fig2.yaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter(
            '%d'))
        # Move second figure down so the title doesn't overlap the x
        # axis label
        plt.subplots_adjust(hspace=0.8)
        plt.plot(self.telemetry_days_from_switch,
                 self.telemetry_encoder_diff,
                 # Plot using red o's = ro
                 'ro', label="Telemetry Encoder Difference")
        plt.plot(self.image_days_from_switch,
                 self.image_encoder_diff,
                 # Plot using green o's = go
                 'go', label="Image Encoder Difference")
        plt.legend(loc='best', prop={'size': 8}, numpoints=1)
        plt.title("SSM Encoder Error")
        plt.xlabel("Date Difference (Days)")

    def _plot_figure_3(self, plt):
        """ Creates third figure
        """
        # 313 = num rows, num columns, figure num
        fig3 = plt.subplot(313)
        fig3.yaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter(
            '%d'))
        # Move third figure down so the title doesn't overlap the x
        # axis label
        plt.subplots_adjust(hspace=0.5)
        plt.plot(self.telemetry_days_from_switch,
                 self.telemetry_measured_position,
                 # Plot using green x's = gx
                 'gx', label="Observed Telemetry Position")
        plt.plot(self.ssm_positions_est_days,
                 self.ssm_positions_est_position,
                 'bo', label="Estimated SSM Positions")
        plt.plot(self.image_days_from_switch,
                 self.image_measured_position,
                 # Plot using red x's = rx
                 'rx', label="Observed Image Position")
        # Place legend in the best location on plot, set size, and
        # only display one point (default is 2) ie normally it would
        # show 2 gx 2 bo and 2 rx in legend
        plt.legend(loc='best', prop={'size': 8}, numpoints=1)
        plt.title("SSM Encoder - Telemetry - Calibration - Table")
        plt.xlabel("Date Difference (Days)")

    def _convert_line_to_model(self, line):
        """ Helper method to parse a line into a model
        """
        parm_array = line.split()
        # Expecting 9 elements when line is split
        if len(parm_array) != 9:
            message = ("Model array converted from {0} does not have 9 "
                       "elements.".format(line))
            raise OSError(message)
        a0 = float(parm_array[0])
        a1 = float(parm_array[1])
        a2 = float(parm_array[2])
        a3 = float(parm_array[3])
        tau1 = float(parm_array[4])
        tau2 = float(parm_array[5])
        tau3 = float(parm_array[6])
        slope = float(parm_array[7])
        encoder_nadir_pos = int(parm_array[8])

        return self.Model(a0, a1, a2, a3, tau1, tau2, tau3, slope,
                          encoder_nadir_pos)

    def build_model(self, quality):
        """ Builds model based on results from the C application's
            report file.

            Parameters:
                quality: quality determined by fit script to store
                         in model
        """
        model = SsmPosModel(
            # Create the new model with an invalid model id until we know
            # the new model id (ie when accepting a model)
            -1,
            self.event_id,
            self.alg_version,
            self.new_model.encoder_nadir_pos,
            self.new_model.a0,
            self.new_model.a1,
            self.new_model.a2,
            self.new_model.a3,
            self.new_model.tau1,
            self.new_model.tau2,
            self.new_model.tau3,
            self.new_model.slope,
            quality, True)
        return model

    def _retrieve_results(self):
        """ Parses the SSM Model fit results file and stores information
            in class variables for easy access/use
        """
        alg_version_substring = "Algorithm Version"
        model_id_substring = "Previous Model ID:"
        event_id_substring = "Switch Event ID:"
        start_substring = "Start: "
        stop_substring = "End: "
        previous_substring = "Previous"
        previous_found = False
        new_substring = "New"
        new_found = False

        date_substring = "Date"
        seconds_substring = "Seconds"
        type_substring = "Type"
        measured_substring = "Measured"
        model_substring = "Model"
        residual_substring = "Residual"
        residual_found = False

        # Only want to read file and go through lines once, so store
        # information as each line is read
        try:
            with open(self.results_filename, 'r') as fresults_filename:
                for line in fresults_filename.readlines():
                    # Remove leading and trailing whitespace from line read
                    line = line.strip()
                    # Store all lines in the report_file array so it is easier
                    # to print
                    self.report_file.append(line)
                    # Now store results individually
                    if not line:
                        continue
                    if alg_version_substring in line:
                        self.alg_version = line.split()[-1]
                    elif event_id_substring in line:
                        self.event_id = line.split()[-1]
                    elif model_id_substring in line:
                        self.model_id = line.split()[-1]
                    elif start_substring in line:
                        # Start (End) Year,DOY,SOD line?
                        # Split on whitespace in middle of line, we do not
                        # care about label
                        (label, switch_timestamp) = line.split()
                        try:
                            self.start_time = Timestamp.from_string(
                                switch_timestamp)
                        except TimestampError as error:
                            message = ("Error reading start_time: {0}".
                                       format(line))
                            raise OSError(error)
                    elif stop_substring in line:
                        # Stop (End) Year,DOY,SOD line?
                        # Split on whitespace in middle of line, we do not
                        # care about label
                        (label, switch_timestamp) = line.split()
                        try:
                            self.stop_time = Timestamp.from_string(
                                switch_timestamp)
                        except TimestampError as error:
                            message = ("Error reading stop_time: {0}".
                                       format(line))
                            raise OSError(error)
                    # Check that the model_id substring is not in the line
                    # so we do not accidentally pick up the previous model
                    # ID line ie:
                    # Previous Model ID: 29
                    elif (not previous_found and
                          previous_substring in line and
                          model_id_substring not in line):
                        # In format like:
                        #            a0       a1       a2         a3
                        # Previous 2367.800  317.900  249.300   133.900
                        #   tau1      tau2    tau3   slope     Nadir
                        # 1602.000   0.191   1.949   3.095     16574079

                        # Remove first word from line which is the
                        # "Previous"
                        previous = line.split(' ', 1)[1]
                        previous_found = True
                        self.orig_model = self._convert_line_to_model(previous)
                    elif not new_found and new_substring in line:
                        # In format like:
                        #       a0       a1       a2         a3
                        # New 2367.800  317.900  249.300   133.900
                        #   tau1      tau2    tau3   slope     Nadir
                        # 1602.000   0.191   1.949   3.095     16574079
                        # Remove first word from line which is the "New"
                        new = line.split(' ', 1)[1]
                        print("new: {0}".format(new))
                        new_found = True
                        self.new_model = self._convert_line_to_model(new)
                    elif (not residual_found and
                          (residual_substring in line and
                           date_substring in line and
                           seconds_substring in line and
                           type_substring in line and
                           measured_substring in line and
                           model_substring in line)):
                        # Looking for the header line indicating the
                        # start of the data wanted to store and use: ie
                        # Date              Seconds         Days    Type
                        # Measured        Model           Residual
                        residual_found = True
                    elif residual_found:
                        # Date, Seconds,  Type, Measured, Model, and
                        # Residual ie:
                        # 11/13/2015 13:16:14        384.000    Telem
                        # 16575812        16575808.05     -3.95
                        # Date and time first 2 words
                        date_time = line.split()[:1]
                        # Split rest of line into an array
                        line = line.split()
                        measured = line[3]
                        modeled = line[4]
                        # Parse the individual elements from the array
                        if line[2] == 'Telem':
                            self.telemetry_days_from_switch.append(
                                float(line[1]) / 86400)
                            self.telemetry_measured_position.append(
                                measured)
                            self.telemetry_modeled_position.append(
                                modeled)
                            # Save difference for ease in plotting
                            self.telemetry_encoder_diff.append(
                                float(modeled) - float(measured))
                        else:
                            self.image_days_from_switch.append(
                                float(line[1]) / 86400)
                            self.image_measured_position.append(measured)
                            self.image_modeled_position.append(modeled)
                            # Save difference for ease in plotting
                            self.image_encoder_diff.append(
                                float(modeled) - float(measured))
        except OSError as error:
            message = "Error occurred reading {0}: {1}".format(
                      self.results_filename, error)
            raise OSError(message)

    def _retrieve_ssm_positions(self, estimated_positions_table):
        """ Parses the estimated_positions_table records and calculates
            the days from switch (epoch) to be used later in plots
        """
        estimated_positions_table = estimated_positions_table.position_table
        epoch_timestamp = None
        for estimated_position in estimated_positions_table:
            if not epoch_timestamp:
                # Set the first timestamp as the "epoch"
                epoch_timestamp = estimated_position.timestamp
            else:
                # Get the diff between the current timestamp from the epoch
                epoch_diff = estimated_position.timestamp - epoch_timestamp
                days_from_epoch = epoch_diff.total_seconds() / 86400
                self.ssm_positions_est_days.append(days_from_epoch)
                self.ssm_positions_est_position.append(
                    estimated_position.position)

    def display_results(self):
        """ Displays results file to terminal from array that was
            populated when the results file was initially parsed
        """
        for i in range(len(self.report_file)):
            print(self.report_file[i])

    def plot_results(self, plot_filename, estimated_positions, no_display):
        """ Calls individual plot routines to plot results from the SSM
            Model fit

            Parameters:
                plot_filename: File name to use when storing plot on disk
                estimated_positions: SsmPosTableGenerator class instance with
                                     the estimated positions
        """
        self.plot_filename = plot_filename
        # Retrieve the estimated positions results
        self._retrieve_ssm_positions(estimated_positions)
        # If the DISPLAY is set, change backend so plot will display
        # This has to be done before the import of pyplot or else
        # it will be ignored
        if not no_display and "DISPLAY" in os.environ:
            matplotlib.use('QT4agg')
        import matplotlib.pyplot as plt
        # Plot each figure
        self._plot_figure_1(plt)
        self._plot_figure_2(plt)
        self._plot_figure_3(plt)
        # Get an instance of the figure. If this is not done before
        # the show (and DISPLAY is set so plot is actually shown)
        # the result from savefig will be blank. After show is called
        # the figure is reset and the list of "current" figures and
        # axes is empty
        fig = plt.gcf()
        # Set the size so that the figures on the plot are not overlapping
        # forward = True means to carry it into the actual figure saved on
        # disk
        fig.set_size_inches(10, 10, forward=True)
        # This will display plot if possible
        plt.show()
        # Also save plot to file
        fig.savefig(self.plot_filename)
