#!/usr/local/bin/python3
""" Fit TIRS SSM Model Switch Event

    Collects the inputs required to perform the SSM model fit for a single SSM
    switch event then calls the fit_tirs_ssm_model application to perform
    the fit
"""
from generate_ssm_position_table import (SsmPosTableGenerator,
                                         SsmPosTableGenFrontEnd)
from ias.db import database as db
from ias.log import logger
from ias.misc.const import SUCCESS
from ias.misc import satellite_attributes as sat_attr
from ias.misc.const import EXIT_SUCCESS, EXIT_FAILURE
from ias.io.cpf import retrieve_cpf
from switch_event import SwitchEventsTable
from ssm_exceptions import (
    CPFRetrievalError, EventNotFoundError, InvalidDataError, MissingDataError,
    ParameterError, SsmException)
from ssm_pos_model import SsmPosModel, SsmPosModelTable
from ssm_pos_model import ModelQuality
from ssm_utilities import connect_to_database, valid_timestamp
from subprocess import check_call, CalledProcessError
from timestamp import Timestamp
from tirs_ssm_model_fit_parameters import ModelFitParameters
import argparse
import fit_tirs_ssm_model_results
import os
import subprocess
import sys


# Define the exceptions for the errors that can be encountered
class FitSwitchEventError(SsmException):
    pass


class AcceptModelError(FitSwitchEventError):
    pass


class FitSwitchEvent():
    """ Collects the inputs required to perform the SSM Model fit for a single
        SSM switch event and calls the C application to perform the fit.
        Results are displayed to the user in the form of a report and plot.
        Also allows user to accept resulting model and stored into the
        TIRS_SSM_POSITION_MODEL_PARAMETERS table (along with the estimated
        positions generated to the position table).
    """

    def __init__(self, db_connection, log=None):
        """ Construct a new FitSwitchEvent object.

            Parameters:
                db_connection: Already created database connection
                log: Optional logging object.  If not provided, the default
                     logger will be used.
        """
        self.db_connection = db_connection

        # Use the provided log or the default one
        if log is not None:
            self.log = log
        else:
            # Get the default logger
            self.log = logger.Logger.get_default_logger()

        # Set the variables that we already know
        self.estimated_ssm_positions_filename = \
            "tirs_ssm_generated_positions.txt"
        self.fit_results_filename = "tirs_ssm_model_fit.rpt"
        self.odl_filename = "tirs_ssm_model_fit.odl"
        self.plot_filename = "tirs_ssm_model_fit_plot.png"
        self.ssm_encoder_filename = "ssm_encoder.csv"
        self.working_directory = os.getcwd()
        self.model_table = SsmPosModelTable(self.db_connection)

        # These class variables will be set during processing
        self.coefficient_masks = None
        self.complete_event = False
        self.cpf_name = None
        self.estimated_ssm_positions_table = None
        self.event_id = None
        self.event_directory = None
        self.new_model = None
        self.original_model = None
        self.quality = None
        self.start_time = None
        self.stop_time = None
        self.switch_event = None

    def _parse_parameter_string_to_list(self, parameters):
        """ Helper function to parse a parameter string into a list

            Parameters:
                parameters: string in the format of:
                    (x,x,x,x,x,x,x,x,x,x,x)
        """
        parameter_string = parameters.lstrip('(')
        parameter_string = parameter_string.rstrip(')')
        parameter_string = ''.join(parameter_string.split())
        parameter_list = parameter_string.split(',')
        return parameter_list

    def _valid_model_parameters(self, parameters):
        """ Helper function to validate number of elements in the command
            line parameters for masks and model overrides.

            Parameters:
                parameters: either model_override or masks provided on
                            command line
                Note: Normal model has 9 parameters but there is not an
                      entry for the encoder nadir position in either possible
                      parameter so we check for 8 possible elements
        """
        if parameters is None:
            return

        parameter_list = self._parse_parameter_string_to_list(parameters)

        # Verify 8 parameters were provided
        if (len(parameter_list) != 8):
            message = ("Invalid number of parameters provided ({0}): {1}".
                       format(len(parameter_list), parameters))
            raise ParameterError(message)

    def accept_model(self):
        """ Accepts newly generated model parameters and stores them in the
            TIRS_SSM_POSITION_MODEL_PARAMETERS and TIRS_SSM_POSITION_ESTIMATES
            tables. Disables old model parameters and positions accordingly
        """
        # Start a transaction
        self.db_connection.begin_transaction()
        try:
            # Disable old model, the disable_model_record already throws an
            # exception if it was unable to disable the model
            self.model_table.disable_model_record(self.original_model.model_id)
            self.log.info("Old model disabled, switch id {0} "
                          "model id {1}".format(self.event_id,
                                                self.original_model.model_id))
            # Get next model ID to be inserted
            next_model_id = self.model_table.get_next_model_id()
            # Update the new model with the new model ID
            self.new_model.model_id = next_model_id

            self.log.info("Inserting model {0}".format(self.new_model))

            # Insert the new model into the table, the insert_model_record
            # already throws an exception if errors occur
            self.model_table.insert_model_record(self.new_model)
            self.log.info("New model inserted, model id {0}".
                          format(next_model_id))

            # Store new estimated positions for the new model in the database
            self.estimated_ssm_positions_table.store_results()
            self.log.info("New postions successfully populated")

            # Commit the database changes
            self.db_connection.commit()
        except:
            self.db_connection.rollback()
            raise

    def check_results(self, fit_results):
        """ Verifies the parameters in the fit report matches expected
            values

            Parameters:
                fit_results: Instance of the results class that contains
                             the class members to use in the comparisons.
                             fit_results.event_id: Event ID from fit report
                             fit_results.start_time: Event ID start time from
                                                     fit report
                             fit_results.stop_time: Event ID stop time from
                                                    fit report
                             fit_results.model_id: Original model ID from
                                                   fit report
        """

        # Switch Event ID
        if str(self.event_id).strip() != str(fit_results.event_id).strip():
            self.log.warning("Event ID in fit ({0}) does not match "
                             "event ID ({1}) read from {2}".
                             format(self.event_id,
                                    fit_results.event_id,
                                    self.fit_results_filename))

        # Switch start time
        if self.start_time != fit_results.start_time:
            self.log.warning("Start time in fit ({0}) does not match "
                             "start time ({1}) read from {2}".
                             format(self.start_time,
                                    fit_results.start_time,
                                    self.fit_results_filename))

        # Switch stop time
        if self.stop_time != fit_results.stop_time:
            self.log.warning("Stop time in fit ({0}) does not match "
                             "stop time ({1}) read from {2}".
                             format(self.stop_time,
                                    fit_results.stop_time,
                                    self.fit_results_filename))

        # Original Model ID
        if int(self.original_model.model_id) != int(fit_results.model_id):
            self.log.warning("Previous model id in fit ({0}) does not match "
                             "previous model id ({1}) read from {2}".
                             format(self.original_model.model_id,
                                    fit_results.model_id,
                                    self.fit_results_filename))

    def create_encoder_file(self):
        """ Creates ssm encoder file for the C application
        """
        # Assuming the generate_ssm_encoder_file executable is in the
        # user's path
        try:
            cmd = "generate_ssm_encoder_file " + self.odl_filename
            self.log.info("Creating ssm encoder file: {0}".format(cmd))
            # Change into the event directory so that the ssm encoder file
            # is created in the appropriately named 'event' directory
            os.chdir(self.event_directory)
            check_call([cmd], shell=True)
        except subprocess.CalledProcessError as error:
            message = (("{0} failed with return code {1} - {2}").format(
                cmd, error.returncode, error.output))
            self.log.error(message)
            raise FitSwitchEventError(message)
        finally:
            # Change back into the original working directory, no matter what
            os.chdir(self.working_directory)

    def create_estimated_positions(self):
        """ Creates estimated positions for the newly created model
        """
        # Create the class instance, note that it expects the stop time
        # before the start time, this is not an error
        self.estimated_ssm_positions_table = SsmPosTableGenerator(
            self.db_connection, self.new_model, self.stop_time,
            self.start_time)

        # Generate the ssm pos table
        self.estimated_ssm_positions_table.generate_ssm_pos_table()

        # Create the estimated ssm positions file
        SsmPosTableGenFrontEnd.write_positions_file(
            self.get_event_filename(self.estimated_ssm_positions_filename),
            self.estimated_ssm_positions_table.position_table)

    def create_odl(self):
        """ Creates odl file for input to the fit_tirs_ssm_model
        """
        try:
            # Change into the event directory
            os.chdir(self.event_directory)
            # Write out ODL file
            with open(self.odl_filename, "w") as fodl:
                fodl.write("OBJECT = TIRS_SSM_MODEL_FIT\n")
                fodl.write("  SWITCH_EVENT_ID = {0}\n".format(self.event_id))
                fodl.write("  EVENT_START_TIME = ({0})\n".
                           format(str(self.start_time).replace(':', ', ')))
                fodl.write("  EVENT_END_TIME = ({0})\n" .
                           format(str(self.stop_time).replace(':', ', ')))
                fodl.write("  MODEL_ID = {0}\n".
                           format(self.original_model.model_id))
                fodl.write("  INITIAL_VALUES = ({0}, {1}, {2}, {3}, {4}, "
                           "{5}, {6}, {7})\n".
                           format(self.original_model.a0,
                                  self.original_model.a1,
                                  self.original_model.a2,
                                  self.original_model.a3,
                                  self.original_model.tau1,
                                  self.original_model.tau2,
                                  self.original_model.tau3,
                                  self.original_model.slope))
                fodl.write("  CAL_PARM_FILENAME = \"{0}\"\n".
                           format(self.cpf_name))
                fodl.write("  TIRS_RESULTS_FILE_NAME = \"{0}\"\n".
                           format(self.fit_results_filename))
                fodl.write("  SSM_ENCODER_FILENAME = \"{0}\"\n".
                           format(self.ssm_encoder_filename))
                fodl.write("  ENCODER_VALUE0 = {0}\n".
                           format(self.original_model.encoder_nadir_pos))
                fodl.write("  COEFFICIENT_MASKS = {0}\n" .
                           format(self.coefficient_masks))
                fodl.write("  CAL_OUTLIER_RMSE_THRESHOLD = {0}\n".
                           format(ModelFitParameters.
                                  calibration_outlier_rmse_threshold))
                fodl.write("  TELEMETRY_OBSERVATION_STDEV = {0}\n".
                           format(ModelFitParameters.
                                  telemetry_observation_stdev))
                fodl.write("  IMAGE_OBSERVATION_STDEV = {0}\n".
                           format(ModelFitParameters.image_observation_stdev))
                fodl.write("  APRIORI_WEIGHTS = {0}\n".
                           format(ModelFitParameters.apriori_weights))
                fodl.write("  BREAK_POINT_TIMES = {0}\n".
                           format(ModelFitParameters.model_break_points))
                fodl.write("  OBSERVATION_GAP_WINDOWS = {0}\n".
                           format(ModelFitParameters.observation_gap_windows))
                fodl.write("  CONVERGENCE_THRESHOLD_WEIGHTS = {0}\n".
                           format(ModelFitParameters.
                                  convergence_threshold_weights))
                fodl.write("  MAXIMUM_SOLUTION_ITERATIONS = {0}\n".
                           format(ModelFitParameters.fit_maximum_iterations))
                fodl.write("  TELEMETRY_START_OFFSET = {0}\n".
                           format(ModelFitParameters.telemetry_start_offset))
                fodl.write("  TELEMETRY_SUBSAMPLE_THRESHOLD = {0}\n".
                           format(ModelFitParameters.
                                  telemetry_subsample_threshold))
                fodl.write("  TELEMETRY_SUBSAMPLE_STEP = {0}\n".
                           format(ModelFitParameters.telemetry_subsample_step))
                fodl.write("END_OBJECT = TIRS_SSM_MODEL_FIT\n")
                fodl.write("END\n")
                fodl.close()
            self.log.info("Successfully created odl file: {0}".
                          format(self.get_event_filename(self.odl_filename)))
        except OSError as error:
            message = ("Error writing ODL File - {0}".format(
                self.get_event_filename(self.odl_filename)))
            self.log.error(message)
            raise OSError(message)
        finally:
            # Change back into the original working directory, no matter what
            os.chdir(self.working_directory)

    def get_event_filename(self, orig_filename):
        new_filename = self.event_directory + orig_filename
        return new_filename

    def get_cpf(self):
        """ Retrieves CPF corresponding to the switch event start time
        """
        # Create timestamp
        cpf_date = self.start_time.get_datetime()

        # Replace the blank in the string with the 'T' that the service expects
        cpf_date = str(cpf_date).replace(' ', 'T')
        self.log.info("date for cpf service {0} ".format(cpf_date))

        # Create a list so the retrieve_cpf also returns the CPF name
        outcpf = []

        try:
            # Change into the event directory since the get_cpf only retrieves
            # the cpf file to the current working directory
            os.chdir(self.event_directory)
            status = retrieve_cpf.retrieve_cpf("DATE", cpf_date, None, outcpf)
            if status == SUCCESS:
                self.cpf_name = outcpf[0]
                self.log.info("cpf: {0}".format(self.cpf_name))
            else:
                message = "No CPF found for date: {0}".format(cpf_date)
                self.log.error(message)
                raise CPFRetrievalError(message)
        except CPFRetrievalError as error:
            raise MissingDataError(error)
        finally:
            # Change back into the original working directory, no matter what
            os.chdir(self.working_directory)

    def get_model(self, override=None):
        """ Retrieves current active model for class member event id
        """
        # ModelTable class instance
        temp_model = self.model_table.get_active_model(self.event_id)
        if (override is not None):
            model_list = self._parse_parameter_string_to_list(override)
            temp_model = SsmPosModel(
                # Model override will only have the parameters, use the rest
                # from the active model just obtained
                temp_model.model_id, temp_model.event_id,
                temp_model.alg_version, temp_model.encoder_nadir_pos,
                #   a0              a1            a2             a3
                model_list[0], model_list[1], model_list[2], model_list[3],
                #   tau1              tau2         tau3        slope
                model_list[4], model_list[5], model_list[6], model_list[7],
                #  quality             valid
                self.quality, True)
        self.original_model = temp_model

        self.log.info("Selected Model {0}".format(self.original_model))

    def get_switch_event(self):
        """ Retrieves DB information for the provided switch event ID
            along with start/stop times and quality (if needed)

            Parameters:
                switch_event_id provided to script
        """
        self.log.info("Retrieving information for switch event id {0}".
                      format(self.event_id))
        # Switch Events Table class instance
        switch_table = SwitchEventsTable(self.db_connection)

        # Retrieve event information
        try:
            self.switch_event = switch_table.get_event(self.event_id)
        except EventNotFoundError as ex:
            self.log.error(ex)
            raise MissingDataError(ex)

        if not self.switch_event.active:
            message = "Error: {0} switch_event is not active".format(
                self.event_id)
            self.log.error(message)
            raise InvalidDataError(message)

        # Get start/stop time for the switch event
        (self.start_time, switch_stop_time,
         self.complete_event) = switch_table.get_start_stop_time(
             self.event_id)

        self.log.info("switch event id {0} start time: {1}".
                      format(self.event_id, self.start_time))

        # If a stop time override was not provided on command line,
        # set the stop time to the stop time from the switch event
        if self.stop_time is None:
            self.stop_time = switch_stop_time
            self.log.info("switch event id {0} stop time: {1}".
                          format(self.event_id, self.stop_time))
        else:
            # If a stop time was provided on command line, override the
            # complete event to be false.  Therefore, the quality will
            # automatically be set to preliminary
            self.complete_event = False
            self.log.warn("switch event id {0} stop time overridden on "
                          "line with command: {1}. Overriding complete_event "
                          "to limit quality to preliminary.".
                          format(self.event_id, self.stop_time))

        # Set quality based on complete event
        self.set_quality(self.complete_event)

        # Error check that stop time is greater than start time
        # Probably only happen when a stop time was provided on the
        # command line, but checking every time
        if self.stop_time <= self.start_time:
            message = "Error: stop time is at or before start time "
            "for the specified event: {0}".format(self.event_id)
            self.log.error(message)
            raise argparse.ArgumentTypeError(message)

    def prepare_for_fit(self, event_id, create_directory, masks=None,
                        model_override=None):
        """ Prepares all data needed for performing the fit by calling
            the individual functions for each.  Put in its own function since
            each scenario in the fit script can run with needs all the
            information regardless of the desired results from the run.

            Sets:
                switch event id and all associated information
                report directory with name based on switch event ID provided
                    to script
                ODL file name to store in report directory
                    (tirs_ssm_model_fit.odl)
                Results file name to store in report directory
                    (tirs_ssm_model_fit.rpt)
                Current model corresponding to switch event ID provided to
                    script
                ODL Parameter Coefficient masks

            Parameters:
                switch_event_id
        """
        self.event_id = event_id
        self.get_switch_event()
        self.event_directory = "{:03d}".format(event_id)
        self.event_directory += "_" + str(self.start_time).replace(':', '_')
        self.event_directory += "/"
        self.log.info("event_directory: {0}".format(self.event_directory))
        # Validate the model override is in the proper format and
        # the correct number of parameters were provided
        self._valid_model_parameters(model_override)
        self.get_model(model_override)
        if masks is not None:
            # Validate the coefficient masks are in the proper format and
            # the correct number of parameters were provided
            self._valid_model_parameters(masks)
            self.coefficient_masks = masks
        else:
            self.coefficient_masks = ModelFitParameters.coefficient_masks

        # Create switch event directory if command requested requires it
        if create_directory:
            # If switch event ID subdirectory does not already exist, create it
            # If it does already exist, the files within it will be overwritten
            # if a fit is being performed
            if not os.path.exists(self.event_directory):
                os.mkdir(self.event_directory)
        if not os.path.exists(self.event_directory):
            message = ("Event directory does not exist: {0}".
                       format(self.event_directory))
            self.log.error(message)
            raise OSError(message)

    def retrieve_results(self):
        """ Retrieves and organizes results from the C executable report file
        """
        # Results file should always exist regardless of performing from switch
        # event, using previous files, or only plotting/displaying
        try:
            fit_results = fit_tirs_ssm_model_results.FitResults(
                self.get_event_filename(self.fit_results_filename))
        except OSError as error:
            raise OSError(error)
        # Verify results parameters
        self.check_results(fit_results)
        return fit_results

    def run(self, args):
        """ Run the fit ssm model application using the arguments provided.

            Parameters:
                args: Parsed command line arguments
        """
        # Only required argument
        event_id = args.event_id[0]
        no_display = args.no_display

        # Set up some of the information needed from provided parameters
        model_override = None
        masks = None
        create_directory = False
        # Set override variables for the fit and fit only options, display
        # and accept to not create the ODL file thus do not need these set
        if args.command == "fit" or args.command == "fit_only":
            self.stop_time = args.stop_time
            model_override = args.model_override
            masks = args.masks

            # Set flag indicating directory should be created if it doesn't
            # already exist. Do not want to set this for accept or display
            # since they perform on data already present
            create_directory = True

        # Note: Need to determine quality during accept before the build_model
        # routine is called, or it will be ignored
        if args.command == "accept" and args.quality:
            self.quality = ModelQuality.from_string(args.quality)
        else:
            self.quality = None

        # Set up the event_id required information for every fit type
        self.prepare_for_fit(event_id, create_directory, masks, model_override)

        # If the command is accept, verify that the can_model is set
        # appropriately.  If not, raise an exception.
        if args.command == "accept" and not self.switch_event.can_model:
            message = ("Cannot accept model because the can_model flag is "
                       "set to: {0}".format(
                                      self.switch_event.get_can_model_flag()))
            self.log.error(message)
            raise FitSwitchEventError(message)

        # Perform actions for a 'full' fit or fit only, not needed for the
        # other fit types
        if args.command == "fit" or args.command == "fit_only":
            try:
                # Note that the fit only really doesn't use the CPF because
                # its only used when the ssm encoder file is generated, which
                # only happens during a 'full' fit.  However, this is how the
                # script determines the CPF name to write to the ODL file.
                # The only time this might be a problem is if cal val wanted
                # to run a fit only and the CPF service was down.  Not going
                # to worry about this now, but if needed potential work arounds
                # would be to either look in the directory for the CPF, query
                # the CPF service for just a name, or read the existing ODL
                # file for the CPF name
                self.get_cpf()
                # Generate odl file
                self.create_odl()

                if args.command == "fit":
                    # Create encoder file after odl file, as the C routine to
                    # query the TIRS DB tables to create the encoder file takes
                    # the ODL file as input
                    self.create_encoder_file()

                # Execute the Fit SSM Position Model on the full or fit only
                self.run_ssm_model_fit()
            except FitSwitchEventError as error:
                message = ("Error performing {0} method with event "
                           "{1}-{2}").format(args.command, self.event_id,
                                             error)
                self.log.error(message)
                raise FitSwitchEventError(message)

        # Retrieve and display results needed for every fit type
        try:
            # Retrieve results from file on disk
            fit_results = self.retrieve_results()
            # Build new model with results
            self.new_model = fit_results.build_model(self.quality)
            # Create estimated positions with new model
            self.create_estimated_positions()
            # Display and plot results
            self.show_results(fit_results, no_display)
        except OSError as error:
            message = ("Error retrieving and displaying results for event "
                       "{0}-{1}").format(self.event_id, error)
            raise FitSwitchEventError(message)

        # Peform the accept action if requested
        if args.command == "accept":
            try:
                # Accept model and estimated positions
                self.accept_model()
            except SsmException as error:
                message = ("Error accepting model and estimated positions for "
                           "event {0}-{1}").format(self.event_id, error)
                self.log.error(message)
                raise FitSwitchEventError(message)

    def run_ssm_model_fit(self):
        """ Runs the fit_tirs_ssm_model executable with the ODL input file
        """
        # Create the command to run the fit c executable. Assuming the
        # tirs_fit_ssm_model executable is in the user's path
        try:
            # Change to the event directory so the results file is
            # created in the directory needed
            os.chdir(self.event_directory)
            cmd = "tirs_fit_ssm_model " + self.odl_filename
            self.log.info("Running SSM model fit: {0}".format(cmd))
            check_call([cmd], shell=True)
        except subprocess.CalledProcessError as error:
            message = (("{0} failed with return code {1} - {2}").format(
                cmd, error.returncode, error.output))
            self.log.error(message)
            raise FitSwitchEventError(message)
        finally:
            # Change back into the original working directory, no matter what
            os.chdir(self.working_directory)

    def set_quality(self, complete_event):
        if self.quality is None:
            # If event is complete (ie there are later events)
            # consider the quality to be final
            if complete_event:
                self.quality = ModelQuality.final
            else:
                self.quality = ModelQuality.preliminary

    def show_results(self, fit_results, no_display):
        """ Displays report to screen and shows/saves generated plot
        """
        fit_results.display_results()
        fit_results.plot_results(self.get_event_filename(
            self.plot_filename), self.estimated_ssm_positions_table,
            no_display)

    @staticmethod
    def add_command_line_parameters(argparser):
        """ Add the command line parsing support for fitting switch events to
            ArgumentParser provided

            Parameters:
                argparser: Parent ArgumentParser to add parameters to
        """
        # Use sub-parsers for each of the sub-commands
        subparsers = argparser.add_subparsers(
            title='commands', help='available commands')

        # Work around a bug in ArgumentParser that doesn't allow a subcommand
        # to be required via the add_subparsers method
        subparsers.required = True
        subparsers.dest = 'command'

        # Parent parser for all fit types to share the event id
        shared_parent_parser = argparse.ArgumentParser(add_help=False)
        shared_parent_parser.add_argument(
            'event_id', nargs=1, type=int, help='Event id to use in fit.')
        shared_parent_parser.add_argument(
            '--no_display', action='store_true', default=False,
            help='Option to disable displaying plots.')

        # Parent parser for the fit and fit only command to share their
        # common arguments
        shared_fit_parser = argparse.ArgumentParser(add_help=False)
        shared_fit_parser.add_argument(
            '--model_override', help='Parameter that allows the initial model '
            'to be overridden.  Order is "(a0, a1, a2, a3, tau1, tau2, '
            ' tau3, slope)". All parameters must be present on command line '
            'and encapsulated with double quotes and parentheses as shown.')
        shared_fit_parser.add_argument(
            '--masks', help='Parameter used to suppress selected model '
            'parameters in the model fit by setting the corresponding entry '
            'in the array to zero. Order is "(a0, a1, a2, a3, tau1, tau2, '
            'tau3, slope)". All parameters must be present on command line '
            'and encapsulated with double quotes and parentheses as shown.')
        shared_fit_parser.add_argument(
            '--stop_time', type=valid_timestamp, help='Stop time to '
            'override event stop time - YYYY:DOY:Seconds format. Note that '
            'the stop_time is only relevant when a fit is performed. If '
            'accepting a model when a custom stop time was previously '
            'provided for the fit, the estimated positions will still be '
            'generated over the entire switch event.')

        display_parser = subparsers.add_parser(
            'display', parents=[shared_parent_parser],
            help='Display and plot results from a previous fit.')
        display_parser.set_defaults(command='display')

        fit_only_parser = subparsers.add_parser(
            'fit_only', parents=[shared_parent_parser, shared_fit_parser],
            help='Perform fit from existing data.')
        fit_only_parser.set_defaults(command='fit_only')

        fit_parser = subparsers.add_parser(
            'fit', parents=[shared_parent_parser, shared_fit_parser],
            help='Perform nominal fit')
        fit_parser.set_defaults(command='fit')

        accept_parser = subparsers.add_parser(
            'accept', parents=[shared_parent_parser],
            help='Option to accept the previously created model '
            'and store the parameters in the database.')
        # Note predicted is only allowed through the command line, if no
        # Quality flag is provided, the script will set it to preliminary
        # or final based on whether or not event is complete.  ie, the
        # script does not set the value to predicted unless it is manually
        # overridden from the command line.
        accept_parser.add_argument(
            '--quality', choices=['predicted', 'preliminary', 'final'],
            help='Quality flag to use for the created model. If not provided, '
            'quality will be final for complete events and preliminary for '
            'incomplete events.')
        accept_parser.set_defaults(command='accept')


def parse_parameters(arguments):
    """ Parse the command line arguments.

        Parameters:
            arguments: List of command line arguments to parse.  This is passed
                       in as an argument instead of defaulting to sys.argv in
                       the ArgumentParser to allow unit testing to provide
                       lists to test.

        Returns:
            Arguments object with the parameters as members.
    """
    # Create the main argument parser
    argparser = argparse.ArgumentParser(description='Perform SSM Model Fit')

    # Add the command line parameters for ManageSwitchEvents
    FitSwitchEvent.add_command_line_parameters(argparser)

    # Return the parsed arguments
    return argparser.parse_args(arguments)


def main():
    """Main fit tirs switch events routine
    """
    # Parse command line arguments, skipping first one which is the
    # application name
    args = parse_parameters(sys.argv[1:])

    # Initialize the logger
    log = logger.Logger("Fit TIRS SSM Model Switch Event")
    log.info("Initiated")

    # Set up the satellite attributes for Landsat 8
    sat_attr.set_default(sat_attr.get_satellite_id_from_number(8))

    # Connect to the database
    db_connection = connect_to_database()

    fitSwitch = FitSwitchEvent(db_connection)
    # Assume the operation will complete successfully
    exit_status = EXIT_SUCCESS

    # Run the tirs fit routine
    try:
        fitSwitch.run(args)
        log.info("Completed successfully")
    except SsmException as error:
        log.error("Failed to perform the SSM fit: {0}".format(error))
        exit_status = EXIT_FAILURE
    finally:
        # Close the database connection.  If any inserts were done, they should
        # already be committed.
        db_connection.close(db.DISCONNECT_ROLLBACK)

    log.close()

    sys.exit(exit_status)

# Run the main routine if called from command line
if __name__ == "__main__":
    main()
