#!/usr/local/bin/python3
""" Generate TIRS SSM Position Table

    This class provides the ability to generate and (optionally) store
    estimated SSM positions based on model parameters.
"""
from ias.db import database as db
from ias.log import logger
from ias.misc.const import EXIT_SUCCESS, EXIT_FAILURE
from ias.misc import satellite_attributes as sat_attr
from switch_event import SwitchEventsTable
from ssm_pos_model import SsmPosModelTable, ModelNotFoundError, ModelQuality
from tirs_ssm_model_fit_parameters import ModelFitParameters
from ssm_utilities import connect_to_database, valid_timestamp
from estimated_ssm_position import (EstimatedSsmPosition,
                                    EstimatedSsmPositionTable,
                                    EstimatedSsmPositionFile)
from ssm_exceptions import (SsmException, EventNotFoundError, DateRangeError)
from timestamp import Timestamp
from datetime import timedelta
import argparse
from math import exp
import sys


class SsmPosTableGenError(SsmException):
    pass


class SsmPosTableGenerator():
    """ Implements functionality to generate estimated SSM encoders positions
    """
    def __init__(self,  db_in,  model_in, stop_time_in=None,
                 start_time_in=None):
        """ Construct a new SsmPosTableGenerator object.

            Parameters:
                db_in: Already created database connection
                model_in: The SsmPosModel object to create estimates for
                stop_time_in: Stop creating estimates at this point in time
                start_time_in: Start creating estimates at this point in time
                              This parm is optional, if not passed the time
                              of the switch event will be used
        """
        self.db_connection = db_in
        self.model = model_in

        # Use start time if it is passed, otherwise use the switch event time
        # for the passed model
        if start_time_in is None:
            se_table = SwitchEventsTable(self.db_connection)
            self.start_time = se_table.get_start_stop_time(
                                        self.model.event_id)[0].get_datetime()
        else:
            self.start_time = start_time_in.get_datetime()

        # Use stop time if it is passed, otherwise use 60 seconds before the
        # following switch event
        if stop_time_in is None:
            se_table = SwitchEventsTable(self.db_connection)
            time = se_table.get_start_stop_time(
                                        self.model.event_id)[1].get_datetime()
            time -= timedelta(seconds=60)
            self.stop_time = time
        else:
            self.stop_time = stop_time_in.get_datetime()
        self.position_table = []

    def generate_ssm_pos_table(self):
        """ Generates the position estimates for the model. The generated
            positions are stored to the objects position_table member.

            Parameters:
                None

            Returns:
                None
        """

        # Clear out table, in case of multiple calls to this function
        self.position_table = []

        model_seconds = (self.stop_time - self.start_time).total_seconds()
        for s_time in self._get_sample_times(model_seconds):
            delta_t = timedelta(seconds=s_time)
            self.position_table.append(EstimatedSsmPosition(
                             Timestamp.from_datetime(self.start_time+delta_t),
                             # Round the position to an int
                             int(self._calculate_position(delta_t) + 0.5),
                             self.model.quality_flag))

    def generate_report(self):
        """ Generates a report containing the position estimates.

            Parameters:
                None

            Returns:
               The report as a string
        """

        rep = "Year:Day:Second Position Quality"
        for pos in self.position_table:
            rep += "\n{0:<14} {1} {2}".format(str(pos.timestamp),
                                              pos.position,
                                              str(pos.quality_flag))
        return rep

    def _positions_are_unique(self, position_table):
        """ Checks if all the positions we generated are the same as the
            ones that are already active.

            Parameters:
               position_table: EstimatedSsmPostionTable object to access data
                               in the DB table

            Returns:
                True:  One or more positions are unique
                False: All the positions are the same
        """
        active_pos = position_table.get_ssm_positions(self.model.model_id)

        # First check the lists are of equal size
        if len(active_pos) != len(self.position_table):
            return True

        for old_pos, new_pos in zip(active_pos, self.position_table):
            # Check the old position against the new.
            if old_pos != new_pos:
                return True
        # Everything matched, return False
        return False

    def _disable_old_positions(self, position_table):
        """ Disables the old positions in the database table for the associated
            switch event

            Parameters:
               position_table: EstimatedSsmPostionTable object to access data
                               in the DB table

            Returns:
               None
        """
        # Get all the models for this event ID. In most cases only the active
        # model would need it positions disabled, but odd situations may
        # result in others having active positions. So just disable positions
        # for all models from this switch event.
        mod_table = SsmPosModelTable(self.db_connection)
        models = mod_table.get_model_records(self.model.event_id)
        model_ids = [mod.model_id for mod in models]

        if len(model_ids) > 0:
            position_table.disable_ssm_positions(model_ids)

    def store_results(self):
        """ Stores the positions in the database table

            Parameters:
               None

            Returns:
               True:  If positions were stored
               False: If the positions to be stored were duplicates of those
                      already stored in the database
        """
        pos_table = EstimatedSsmPositionTable(self.db_connection)

        # If we already have all the same active positions, just return
        if not self._positions_are_unique(pos_table):
            return False

        # Disable any old positions, before addding new ones
        self._disable_old_positions(pos_table)

        pos_table.insert_ssm_positions(self.model.model_id,
                                       self.position_table)
        return True

    def _get_sample_times(self, model_seconds):
        """ Helper to get the delta times to create estimates at.

            Parameters:
                model_seconds: The number of seconds to model for.

            Returns:
               List of numbers
        """

        # Get the sampling interval table
        temp_interval = list(
            ModelFitParameters.position_sampling_interval_table)

        # Limit the sampling intervals to only those within our time range
        sample_interval = [rec for rec in temp_interval
                           if rec[0] < model_seconds]

        # End at the model seconds
        sample_interval.append([model_seconds, sample_interval[-1][1]])

        # Start at the initial time
        times = [sample_interval[0][0]]

        # Loop until it has reached the final time requested
        while (times[-1] < model_seconds):

            # Add times until we reach the next interval start
            while times[-1] < sample_interval[1][0]:
                times.append(times[-1] + sample_interval[0][1])

            # Done with the current interval
            sample_interval.pop(0)

            # Make sure the next loop starts at the new interval start
            times[-1] = sample_interval[0][0]

        return times

    def reset_stop_time(self):
        """ Call this function to (possibly) set the stop time to 60 seconds
             before the next switch event. If the model is for the latest
             switch event the time is unchanged.

            Parameters:
                None

            Returns:
               Nothing
        """
        se_table = SwitchEventsTable(self.db_connection)
        event_times = se_table.get_start_stop_time(self.model.event_id)

        # If our event is the latest, the 3rd element will be false
        if not event_times[2]:
            # exit without any change made
            return

        # Per the ADD, set stop time 60 seconds before next event
        stop = event_times[1].get_datetime()
        stop -= timedelta(seconds=60)
        self.stop_time = stop

    def _calculate_position(self, delta_time):
        """ Helper to calculate the position at the specified delta time.

            Parameters:
                delta_time: Time to calculate the position for

            Returns:
                The position (a number)
        """
        secs = delta_time.total_seconds()
        days = delta_time.total_seconds() / 86400

        pos = (self.model.encoder_nadir_pos + self.model.a0 + self.model.a1 +
               self.model.a2 + self.model.a3 + self.model.slope * days)

        if self.model.tau1 != 0.0:
            pos = pos - self.model.a1 * exp(-secs / self.model.tau1)
        if self.model.tau2 != 0.0:
            pos = pos - self.model.a2 * exp(-days / self.model.tau2)
        if self.model.tau3 != 0.0:
            pos = pos - self.model.a3 * exp(-days / self.model.tau3)
        return pos


class SsmPosTableGenFrontEnd():
    """ A front end class that allows the functionality to be called from
        other scripts.
    """
    def __init__(self, db_in, log_in=None):
        """ Construct a new SsmPosTableGenFrontEnd object.

            Parameters:
                db_in:  Already created database connection. Note that commits
                        are done as part of this class. If you want more
                        control of what is done, use the SsmPosTableGenerator
                        class instead of this one.
                log_in: Already created logger
        """
        # Use the provided log or the default one
        if log_in is not None:
            self.log = log_in
        else:
            # Get the default logger
            self.log = logger.Logger.get_default_logger()
        self.db_conn = db_in

    @staticmethod
    def add_command_line_parameters(argparser):
        """ Add the command line parsing support for generating positions to
            ArgumentParser provided

            Parameters:
                argparser: Parent ArgumentParser to add parameters to
        """
        argparser.add_argument(
            '--start_time', type=valid_timestamp,
            help='Estimate positions for switch events starting at this time. '
                 'Cannot be combined with the event_id option.')

        argparser.add_argument(
            '--stop_time', type=valid_timestamp,
            help='Estimate positions for switch events until this time. This '
                 'option is required if using start_time and is optional if '
                 'using a single event_id. It is not allowed for multiple '
                 'event_ids.')

        argparser.add_argument(
            '--event_id', type=int, action='append',
            help='Estimate positions for the specified switch event id. This '
                 'option can be specified multiple times.')

        argparser.add_argument(
            '--accept', action='store_true',  default=False,
            help='Option to accept the estimates and store the new positions '
                 'in the database.')

        argparser.add_argument(
            '--save_positions',
            help='Saves the generated positions to a file. This options can '
                 'only be used with a single event id')

    def parse_parameters(self, arguments):
        """ Parses command line arguments for this script

                Parameters:
                    arguments: List of args to parse

                Returns:
                    Namespace populated with argument values
        """
        argparser = argparse.ArgumentParser(
            description='Generate TIRS SSM Position Table')

        SsmPosTableGenFrontEnd.add_command_line_parameters(argparser)

        return argparser.parse_args(arguments)

    def _verify_parms(self, args):
        """ This helper runs some checks on the passed arguments. These are
            checks that can not be done using the argparse functionality
            alone.

                Parameters:
                    args: Parsed command line arguments

                Returns:
                    Nothing
        """
        # Various restrictions on when the Event ID option can be used with
        # the time options
        if args.event_id and len(args.event_id) > 0:
            if args.start_time:
                msg = ("Start time can not be used in combination with "
                       "the event-id option")
                self.log.error("Invalid parameters: %s", msg)
                raise SsmPosTableGenError(msg)
            if args.stop_time and len(args.event_id) > 1:
                msg = ("Stop time can only be used in combination with "
                       "a single event-id")
                self.log.error("Invalid parameters: %s", msg)
                raise SsmPosTableGenError(msg)
        elif args.start_time and args.stop_time:
            if args.start_time.get_datetime() > args.stop_time.get_datetime():
                msg = "Start time can not come after the stop"
                self.log.error("Invalid parameters: %s", msg)
                raise DateRangeError(msg)
        else:
            msg = "Either start and stop time OR event-id must be passed"
            self.log.error("Invalid parameters: %s", msg)
            raise SsmPosTableGenError(msg)

    @staticmethod
    def write_positions_file(filename, positions):
        """ Stores the passed positions in the specfied file

            Parameters:
                filename:  Name of file to write
                positions: List of positions to write

            Returns:
                None
        """
        pos_file = EstimatedSsmPositionFile(filename)
        pos_file.write(positions)

    def run(self, args):
        """ Run the script with the passed arguments

                Parameters:
                    args: Parsed command line arguments

                Returns:
                    Nothing
        """
        self._verify_parms(args)

        self.log.info("Initiated")

        # A list containing positions for all matching event_ids. This list is
        # used only for saving the positions to a file.
        all_positions = []

        se_table = SwitchEventsTable(self.db_conn)
        if args.event_id is None:
            events = se_table.get_filtered_events(
                                        reverse=False,
                                        can_model=True,
                                        active=True,
                                        start_timestamp=args.start_time,
                                        end_timestamp=args.stop_time)
            if len(events) < 1:
                self.log.warn("No matching switch events found. Exiting")
                return
        else:
            events = []
            for evt in args.event_id:
                try:
                    events.append(se_table.get_event(evt))
                except EventNotFoundError:
                    msg = "Unable to get event for event id {0}".format(evt)
                    self.log.fatal(msg)
                    raise SsmPosTableGenError(msg)

        for ev in events:
            try:
                mod_table = SsmPosModelTable(self.db_conn)
                model = mod_table.get_active_model(ev.event_id)
                # If the active model is not valid throw an exception
                if not model.valid:
                    msg = ("Active model for event id {0} is not "
                           "valid.").format(ev.event_id)
                    self.log.fatal(msg)
                    raise SsmPosTableGenError(msg)
            except ModelNotFoundError:
                msg = "No records for Event ID {0}".format(ev.event_id)
                self.log.fatal(msg)
                raise SsmPosTableGenError(msg)

            if(model.tau1 == 0.0 or model.tau2 == 0.0 or
               model.tau3 == 0.0):
                self.log.warn("One or more tau components of model_id %i"
                              " are 0. These terms will be ignored",
                              model.model_id)
            gen = SsmPosTableGenerator(self.db_conn, model, args.stop_time)
            gen.reset_stop_time()
            gen.generate_ssm_pos_table()
            rep = gen.generate_report()
            print(rep)

            # Store positions in the DB, if user specified the accept option
            if args.accept:
                self.db_conn.begin_transaction()
                try:
                    if not gen.store_results():
                        self.log.warn("The positions to be inserted were the "
                                      "same as those already active.")
                    self.db_conn.commit()
                # If ANYTHING went wrong during storing of the positions, we
                # need to rollback any positions we added to the database
                except Exception as ex:
                    msg = "Unable to store positions: {0}".format(ex)
                    self.log.fatal(msg)
                    self.db_conn.rollback()
                    raise SsmPosTableGenError(msg)

            all_positions += gen.position_table

        # Save the positions to a file, if user specified the save option
        if args.save_positions:
            try:
                self.write_positions_file(args.save_positions, all_positions)
            except IOError:
                msg = "Unable to write file {0}".format(args.save_positions)
                self.log.fatal(msg)
                raise SsmPosTableGenError(msg)

        self.log.info("Completed")


def main():
    """Main generate positions routine
    """
    try:
        # Initialize the logger
        log = logger.Logger("generate_ssm_position_table.py")

        # Init satellite attributes
        sat_attr.set_default(sat_attr.get_satellite_id_from_number(8))
        # Connect to the database
        db_conn = connect_to_database()

        # Create and call run on the Front End object
        fe = SsmPosTableGenFrontEnd(db_conn, log)
        parsed_args = fe.parse_parameters(sys.argv[1:])
        fe.run(parsed_args)
        log.close()
        db_conn.close(db.DISCONNECT_ROLLBACK)
    except SsmException as ex:
        log.error("Exception running the Front End: %s", ex)
        exit(EXIT_FAILURE)
    exit(EXIT_SUCCESS)

# Run the main routine if called from command line
if __name__ == "__main__":
    main()
