#!/usr/bin/env python3
###############################################################################
# NAME: load_estimated_tirs_ssm_positions.py
#
# Description: Accepts a switch_event_id and a filename containing information
#   required to create EstimatedSsmPosition objects.
#   This script will first find model records matching the provided switch
#   event. It will then calculate a date range by looking at the dates
#   associated with the provided switch event ID and the "next" switch event
#   ID in the switch events table. The script will then parse out each entry
#   from the provided file whose date falls within this date range. It will
#   then disable any records in the position estimates table whose model IDs
#   match those found, and will insert the new estimated SSM positions.
#
###############################################################################

import sys
import os
import argparse
from ias.log import logger
from ias.db import database as db
from ias.misc import satellite_attributes as sat_attr
from switch_event import SwitchEventsTable
from estimated_ssm_position import (EstimatedSsmPositionFile,
                                    EstimatedSsmPositionTable)
from ssm_exceptions import (SsmException, InvalidDataError,
                            ParameterError)
from ssm_pos_model import ModelQuality, SsmPosModelTable
from ssm_utilities import connect_to_database


class LoadEstimatedTirsSsmPositions:
    """ This class implements the functionality to load estimated positions
        into the database.
    """

    def __init__(self, db_connection, log=None):
        """ Constructor for the LoadEstimatedTirsSsmPositions class.

            Parameters:
                db_connection: the DatabaseConnection to use
                log: optional logger to use (one is created by default)
        """
        self.db_connection = db_connection
        if log is not None:
            self.log = log
        else:
            self.log = logger.Logger.get_default_logger()

    @staticmethod
    def add_command_line_parameters(argparser):
        """ Add the command line parsing support for loading an estimated
            position file into the ArgumentParser provided

            Parameters:
                argparser: Parent ArgumentParser to add parameters to
        """

        def valid_quality_flag(s):
            """ Validate and return the quality flag parameter. If the
                timestamp fails to parse, throw an ArgumentTypeError exception.
            """
            try:
                return ModelQuality(int(s))
            except Exception as error:
                raise argparse.ArgumentTypeError(error)

        # Use sub-parsers for each of the sub-commands
        subparsers = argparser.add_subparsers(
            title='commands', help='available commands')

        # Work around a bug in ArgumentParser that doesn't allow a subcommand
        # to be required via the add_subparsers method
        subparsers.required = True
        subparsers.dest = 'command'

        load_parser = subparsers.add_parser('load', help='Load SSM positions')
        load_parser.add_argument(
            'switch_id', nargs=1, type=int,
            help='Switch event ID to load positions for')
        load_parser.add_argument(
            'filename', nargs=1, help='Estimated SSM position file to load')
        load_parser.add_argument(
            '--rollback', action='store_true',
            help="Debugging option to prevent committing database changes")
        load_parser.add_argument(
            '--quality_flag', type=valid_quality_flag,
            help="Quality to assign loaded positions. This is only used if " +
                 "we are loading positions from a format0 file. It will not " +
                 "be used if loading from a format1 file, as each entry is " +
                 "paired with its own quality flag.")
        load_parser.set_defaults(command='load')

        # The verify command take no args
        verify_parser = subparsers.add_parser(
            'verify',
            help='Verify SSM positions in the database are consistent')

    def load_positions(self, switch_event_id, ssm_pos_filename,
                       quality_flag=None, rollback=False):
        """ Find model IDs associated with the provided switch event ID,
            calculate date range, read EstimatedSsmPositions from file,
            disable all entries in the position estimates table whose model IDs
            match any of our found model IDs, and insert the new
            EstimatedSsmPositions into the table.

            Parameters:
                switch_event_id: Switch event ID of the positions to load
                ssm_pos_filename: Name of the file with the estimated positions
                quality_flag: optional parameter to apply to the estimates if
                    the file is an older format without the quality available
                rollback: optional debugging parameter to rollback the changes
        """
        # Query the TIRS_SSM_POS_MODEL_PARAMS table for the model records
        # matching the provided switch event.
        self.log.info("Getting Model records")
        model_table = SsmPosModelTable(self.db_connection)
        model_records = model_table.get_model_records(switch_event_id)

        # Create a list of model IDs that will need to have estimated positions
        # disabled
        self.log.info("Extracting model IDs")
        model_ids = [x.model_id for x in model_records]

        # Find our date range for our switch event ID if the switch event is
        # active
        switch_event_table = SwitchEventsTable(self.db_connection)
        switch_event = switch_event_table.get_event(switch_event_id)
        date_range = None
        if switch_event.active:
            self.log.info("Calculating date range")
            date_range = switch_event_table.get_start_stop_time(
                switch_event_id)
        else:
            raise ParameterError("Cannot load estimated ssm positions for "
                                 "inactive switch event ID {0}".format(
                                    switch_event_id))

        # Get the active model record for the switch event
        active_model_record = model_table.get_active_model(switch_event_id)
        active_model_id = active_model_record.model_id

        # Create the file object
        self.log.info("Opening {0}".format(ssm_pos_filename))
        ssm_encoder_data_fh = EstimatedSsmPositionFile(
            ssm_pos_filename, quality_flag)

        # Read all positions which fall into our date range
        self.log.info("Finding entries in ssm_pos_file within date range")
        entries = ssm_encoder_data_fh.read(date_range[0], date_range[1])

        # Start the transaction
        self.db_connection.begin_transaction()

        # Disable previous estimated positions
        est_ssm_position_table = EstimatedSsmPositionTable(self.db_connection)

        self.log.info("Disabling old entries in {0}".format(
            est_ssm_position_table.table_name))
        est_ssm_position_table.disable_ssm_positions(model_ids)

        # Load those positions into the TIRS_SSM_POS_ESTIMATES table
        self.log.info("Loading entries into {0} table".format(
            est_ssm_position_table.table_name))
        est_ssm_position_table.insert_ssm_positions(active_model_id, entries)

        # Commit or rollback (if requested)
        if rollback:
            self.db_connection.rollback()
        else:
            self.db_connection.commit()

    def verify_all_positions(self):
        """ Verifies that all active positions are valid (no overlap
            between positions from different models)
        """
        # Get all the enabled estimated positions in time order
        pos_table = EstimatedSsmPositionTable(self.db_connection)
        positions = pos_table.get_all_ssm_positions(enabled=True)

        # If any model IDs are interleaved, something isn't quite right
        last_seen_model_id = -99999
        seen_ids = []
        for position in positions:
            if position[0] != last_seen_model_id:
                if position[0] in seen_ids:
                    msg = ("Estimated positions for model IDs {0} and {1} "
                           "are interleaved ({2}).  The estimated positions "
                           "for one or both models should be regenerated to "
                           "fix this").format(
                                last_seen_model_id, position[0], position[1])
                    self.log.error(msg)
                    raise InvalidDataError(msg)

                # Remember this model ID has been seen
                seen_ids.append(position[0])

                # Remember the current last seen model ID
                last_seen_model_id = position[0]

        self.log.info("All positions verified")

    def run(self, args):
        """ Either loads SSM positions from the file or verifies no Estimated
            SSM Positions are out of place in the database based on which
            command line arguments were passed in. Load is the default action.
        """
        if args.command == "load":
            ssm_pos_filename = args.filename[0]
            switch_event_id = args.switch_id[0]
            desired_quality_flag = args.quality_flag

            # Make sure the file exists
            self.log.info("Validating command line arguments")
            if not os.path.isfile(ssm_pos_filename):
                message = "SSM Position file '{0}' does not exist".format(
                    ssm_pos_filename)
                self.log.error(message)
                raise FileNotFoundError(message)

            self.load_positions(switch_event_id, ssm_pos_filename,
                                desired_quality_flag, rollback=args.rollback)

        elif args.command == "verify":
            self.verify_all_positions()


def parse_parameters(arguments):
    """ Parse the command line arguments.

        Parameters:
            arguments: List of command line arguments to parse.  This is passed
                       in as an argument instead of defaulting to sys.argv in
                       the ArgumentParser to allow unit testing to provide
                       lists to test.

        Returns:
            Arguments object with the parameters as members.
    """
    # Create the main argument parser
    argparser = argparse.ArgumentParser(
        description='Load estimated SSM positions into database')

    # Add the command line parameters
    LoadEstimatedTirsSsmPositions.add_command_line_parameters(argparser)

    # Return the parsed arguments
    return argparser.parse_args(arguments)


def main():
    # Parse any command line arguments, skipping the first one which is the
    # application name
    args = parse_parameters(sys.argv[1:])

    log = logger.Logger("Load Estimated Tirs SSM Positions")
    log.info("Initiated")

    # Set up the satellite attributes for Landsat 8
    sat_attr.set_default(sat_attr.get_satellite_id_from_number(8))

    # Connect to the database
    db_connection = connect_to_database()

    # Create the loader
    load_estimated_tirs_ssm_positions = LoadEstimatedTirsSsmPositions(
        db_connection, log=log)

    exit_status = 0

    try:
        load_estimated_tirs_ssm_positions.run(args)
        log.info("Successful entry")
    except (SsmException, FileNotFoundError) as error:
        log.error(str(error))
        log.error("Failed to complete the requested operation")
        exit_status = 1
    finally:
        db_connection.close(db.DISCONNECT_ROLLBACK)
        log.close()

    sys.exit(exit_status)

if __name__ == "__main__":
    main()
