#!/usr/bin/env python3
""" Manage TIRS SSM Models

    This implements the ability to list and edit records in the
    TIRS_SSM_POS_MODEL_PARAMS table.
"""

import argparse
import sys
from ssm_pos_model import (SsmPosModel, SsmPosModelTable, ModelQuality,
                           ModelNotFoundError)
from ssm_utilities import connect_to_database
from ssm_exceptions import ParameterError
from ias.log import logger
from ias.db import database as db
from ias.misc import satellite_attributes as sat_attr
from ias.misc.const import EXIT_SUCCESS, EXIT_FAILURE


# Define the exceptions for the errors that can be encountered
class ManageModelError(Exception):
    pass


class ManageSsmModels():
    """ Implements a class to support managing SSM models
    """

    def __init__(self, db_connection, log=None):
        """ Construct a new ManageSsmModels object.

            Parameters:
                db_connection: Already created database connection
                log: Optional logging object.  If not provided, the default
                     logger will be used.
        """
        self.db_connection = db_connection

        # Use the provided log or the default one
        if log is not None:
            self.log = log
        else:
            # Get the default logger
            self.log = logger.Logger.get_default_logger()

    def get_models(self, event_id=None, model_id=None, get_all=False):
        """ Get the models requested by the parameters.  Only one of the
            parameters should be provided.  If multiple ones are provided,
            the first one found will be used and the others ignored.

            Optional parameters:
                event_id: Switch event ID of the models to retrieve
                model_id: Model ID of the model to retrieve
                get_all: True/False flag to get all the models available

            Returns:
                An array of the models retrieved
        """

        # Prepare to query the model table
        model_table = SsmPosModelTable(self.db_connection)

        models = []
        if event_id is not None:
            # Get the models for the requested switch event ID
            models = model_table.get_model_records(event_id)

        elif model_id is not None:
            try:
                # Get the single model requested by model ID
                models = [model_table.get_model(model_id)]
            except ModelNotFoundError:
                # If the model isn't found, just return the empty list
                pass

        elif get_all:
            # Get all the model records
            models = model_table.get_model_records()

        else:
            raise ParameterError("No list option defined")

        return models

    def display_models(self, models):
        """ Display a list of models to the terminal.  The size of the display
            columns are determined from the data.

            Parameters:
                models: List of models to display
        """
        # Define the columns to print.  These are in the same order as the
        # parameters returned by the model to_list method.
        columns = ["Model ID", "Event ID", "Alg Ver", "Enc Nadir Pos", "a0",
                   "a1", "a2", "a3", "tau1", "tau2", "tau3", "Slope",
                   "Quality", "Valid", "Date added", "Date disabled"]

        # Get the width of each header columns
        widths = [len(x) for x in columns]

        # Now find the max column width by looking at each record too
        for model in models:
            field_widths = [len(str(x)) for x in model.to_list()]
            widths = list(map(max, field_widths, widths))

        # Display the column header using the calculated widths
        cols = []
        for value, width in zip(columns, widths):
            cols.append("{0:>{1}s}".format(str(value), width))
        header = " | ".join(cols)
        print()
        print(header)
        print('-' * len(header))

        # Print all the records
        for model in models:
            cols = []
            for value, width in zip(model.to_list(), widths):
                cols.append("{0:>{1}s}".format(str(value), width))
            print(" | ".join(cols))

        print()

    @staticmethod
    def add_command_line_parameters(argparser):
        """ Add the command line parsing support for manage models to the
            ArgumentParser provided

            Parameters:
                argparser: Parent ArgumentParser to add parameters to
        """

        # Use sub-parsers for each of the sub-commands
        subparsers = argparser.add_subparsers(
            title='commands', help='available commands')

        # Work around a bug in ArgumentParser that doesn't allow a subcommand
        # to be required via the add_subparsers method
        subparsers.required = True
        subparsers.dest = 'command'

        # The list command requires either a model or switch event ID
        list_parser = subparsers.add_parser(
            'list', help='List existing models events')
        group = list_parser.add_mutually_exclusive_group(required=True)
        group.add_argument(
            '--event_id', type=int,
            help="List all models associated with this Switch Event ID")
        group.add_argument(
            '--model_id', type=int,
            help="List models associated with this Model ID")
        group.add_argument(
            '--all', action='store_true', help="Display all models")

        # The edit command requires model ID and either a valid/invalid option
        edit_parser = subparsers.add_parser(
            'edit', help="Edit an existing model's validity")
        edit_parser.add_argument(
            'model_id', nargs=1, type=int, help='Model ID to modify')
        # valid/invalid model parameters are mutually exclusive
        group = edit_parser.add_mutually_exclusive_group(required=True)
        group.add_argument(
            '--set_valid', action='store_true',
            help="Option to set the model to valid")
        group.add_argument(
            '--set_invalid', action='store_true',
            help="Option to set the model to invalid")
        edit_parser.set_defaults(command='edit')

    def run(self, args):
        """ Run the manage models application using the arguments provided.

            Parameters:
                args: Parsed command line arguments
        """
        # Create the model table object
        model_table = SsmPosModelTable(self.db_connection)

        if args.command == "list":
            # List requested, so get the models matching the parameters
            models = self.get_models(event_id=args.event_id,
                                     model_id=args.model_id, get_all=args.all)

            if len(models) > 0:
                # Models were found, so display them
                self.display_models(models)
            else:
                # No models found, so report an appropriate message
                if args.event_id:
                    message = "No models found for Switch Event ID {0}".format(
                        args.event_id)
                elif args.model_id:
                    message = "No models found for Model ID {0}".format(
                        args.model_id)
                else:
                    message = "No models found"

                self.log.info(message)

        elif args.command == "edit":
            # Editing requested
            if args.set_valid:
                valid = True
            elif args.set_invalid:
                valid = False
            else:
                # Note: This should not be able to happen since the command
                # line arguments require one of the parameters. Raise a general
                # exception for the problem.
                message = "Missing option"
                self.log.error(message)
                raise ParameterError(message)

            try:
                # Start the transaction to update the DB
                self.db_connection.begin_transaction()

                # Make the validity update
                model_id = args.model_id[0]
                model_table.update_model_validity(model_id, valid)

                # Commit the change to the database
                self.db_connection.commit()

                # Retrieve the model and display the updated model
                model = [model_table.get_model(model_id)]
                self.display_models(model)

            except Exception as error:
                # Report any errors encountered
                self.log.error(error)
                raise ManageModelError(
                    "Failed updating model ID {0} validity".format(model_id))

        else:
            # Note: This should not be able to happen since the command line
            # arguments prevent an unknown command from being identified.
            # Raise a general exception for the problem.
            message = "Unrecognized command: {0}".format(args.command)
            self.log.error(message)
            raise ParameterError(message)


def parse_parameters(arguments):
    """ Parse the command line arguments.

        Parameters:
            arguments: List of command line arguments to parse.  This is passed
                       in as an argument instead of defaulting to sys.argv in
                       the ArgumentParser to allow unit testing to provide
                       lists to test.

        Returns:
            Arguments object with the parameters as members.
    """
    # Create the main argument parser
    argparser = argparse.ArgumentParser(
        description='Manage TIRS SSM models')

    # Add the command line parameters for ManageSsmModels
    ManageSsmModels.add_command_line_parameters(argparser)

    # Return the parsed arguments
    return argparser.parse_args(arguments)


def main():
    """ Main manage models script.  This script allows listing models and
        modifying the model valid flag.
    """
    # Parse any command line arguments, skipping the first one which is the
    # application name
    args = parse_parameters(sys.argv[1:])

    # Initialize the logger
    log = logger.Logger("manage_models")
    log.info("Initiated")

    # Set up the satellite attributes for Landsat 8
    sat_attr.set_default(sat_attr.get_satellite_id_from_number(8))

    # Connect to the database
    db_connection = connect_to_database()

    # Create a manage switch events object
    manage_models = ManageSsmModels(db_connection)

    # Assume the operation will complete successfully
    exit_status = EXIT_SUCCESS

    # Run the manage events processing
    try:
        manage_models.run(args)
        log.info("Completed successfully")
    except ManageModelError:
        log.error("Failed to perform the requested operation")
        exit_status = EXIT_FAILURE
    finally:
        # Close the database connection.  Any updates should already be
        # committed.
        db_connection.close(db.DISCONNECT_ROLLBACK)

    # Close the log
    log.close()

    sys.exit(exit_status)


# Runs main if called from command line
if __name__ == "__main__":
    main()
