#!/usr/bin/env python3
""" Manage TIRS SSM Switch Events

    This implements the ability to add, list, edit, disable and delete entries
    in the TIRS_SSM_MODE_SWITCH_EVENTS table.
"""

import argparse
import sys
from timestamp import Timestamp
from switch_event import SwitchEvent, SwitchEventsTable
from ssm_pos_model import SsmPosModel, SsmPosModelTable, ModelQuality
from ssm_utilities import (connect_to_database, valid_timestamp,
                           valid_timestamp_dual_format)
from ias.log import logger
from ias.db import database as db
from ias.misc import satellite_attributes as sat_attr
from ias.misc.const import EXIT_SUCCESS, EXIT_FAILURE
from ssm_exceptions import (EventNotFoundError, EventTooCloseError,
                            ModelValidityError, ParameterError, SsmException,
                            TimestampError)


# Define the exceptions for the errors that can be encountered
class ManageEventError(SsmException):
    pass


class ManageSwitchEvents():
    """ Implements the function to add, edit, and list TIRS SSM mode 0 switch
        events.
    """

    def __init__(self, db_connection, log=None):
        """ Construct a new ManageSwitchEvents object.

            Parameters:
                db_connection: Already created database connection
                log: Optional logging object.  If not provided, the default
                     logger will be used.
        """
        self.db_connection = db_connection

        # Use the provided log or the default one
        if log is not None:
            self.log = log
        else:
            # Get the default logger
            self.log = logger.Logger.get_default_logger()

    def add_event(self, event_time, can_model=True, active=True):
        """ Adds a new switch event.

            Parameters:
                event_time: Timestamp of the event to add
                can_model: Optional "can model" flag.  If not provided,
                           defaults to True.
                active: Optional active flag.  If not provided, defaults to
                        True.

            Returns:
                The newly created switch event, or None if the event was not
                created.
        """

        # Prepare to query the switch events and position model tables
        switch_table = SwitchEventsTable(self.db_connection)
        model_table = SsmPosModelTable(self.db_connection)

        # Query the database for all the existing query events
        switch_events = switch_table.get_all_events()

        # Find the closest event
        closest_event = None
        closest_seconds_diff = None
        for event in switch_events:
            time_diff = event_time - event.time
            seconds_diff = abs(time_diff.total_seconds())

            if (closest_seconds_diff is None or
                    seconds_diff < closest_seconds_diff):
                closest_seconds_diff = seconds_diff
                closest_event = event

        # If a close event has been found (might not be one if the table is
        # empty), make sure it isn't too close
        if (closest_event):
            self.log.debug("Closest event is {0} seconds away".format(
                closest_seconds_diff))

            # An existing events within 6000 seconds is a fatal error
            if (closest_seconds_diff <= 6000):
                message = (
                    "Switch event ID {0} at {1} is {2} seconds away which is "
                    "not more than 6000 seconds.  New event not added."
                    ).format(
                        closest_event.event_id, closest_event.time,
                        closest_seconds_diff)
                self.log.error(message)
                raise EventTooCloseError(message)

            # An existing events within 86400 seconds (one day) is allowed, but
            # a warning is provided
            if (closest_seconds_diff <= 86400):
                self.log.warning(
                    "Switch event ID {0} at {1} is {2} seconds away which is "
                    "not more than 86400 seconds".format(
                        closest_event.event_id, closest_event.time,
                        closest_seconds_diff))

        # Get the default TIRS SSM model from the CPF and update the ID and
        # valid fields
        model = model_table.get_default_model(
                    event_time, switch_events=switch_events)
        # The model quality is predicted when a switch event is added
        model.quality_flag = ModelQuality.predicted
        model.valid = can_model

        # Start a transaction
        self.db_connection.begin_transaction()

        # Get the next switch event ID from the switch event ID sequence
        switch_event_id = switch_table.get_next_event_id()
        self.log.debug("Got switch event ID: {0}".format(switch_event_id))

        # Get the next model ID from the model ID sequence
        model_id = model_table.get_next_model_id()
        self.log.debug("Got model ID: {0}".format(model_id))

        # Build the switch event to insert.
        new_switch_event = SwitchEvent(switch_event_id, event_time, can_model,
                                       active)

        # Update the model with the correct IDs fields
        model.model_id = model_id
        model.event_id = switch_event_id

        # Insert the switch event into the switch event table
        switch_table.insert_event(new_switch_event)

        # Insert the model into the SSM model table
        model_table.insert_model_record(model)

        # Commit the database changes
        self.db_connection.commit()

        return new_switch_event

    def display_events(self, events):
        """ Display a list of switch events to the terminal.

            Parameters:
                events: List of events to display
        """
        print()

        # Print a header for the events
        print("{0:^12s} | {1:^15s}|{2:^9s}|{3:^6s}".format("ID", "Time",
              "Can Model", "Active"))
        print('-' * 50)

        # Print all the records
        for event in events:
            print("{0:>12d} | {1:<15s}|{2:^9s}|{3:^6s}".format(event.event_id,
                  str(event.time), event.get_can_model_flag(),
                  event.get_active_flag()))
        print()

    def edit_event(self, event_id, event_time=None, can_model=None,
                   active=None):
        """ Edit an existing switch event.

            Parameters:
                event_id: ID of the event to modify
                event_time: Optional new timestamp to set for the event_id
                can_model: Optional flag to update the 'can_model' field.
                active: Optional flag to update the 'active' field

            Returns:
                The updated switch event, or None if no update made.
        """
        # Make sure at least one of the parameters is provided
        if event_time is None and can_model is None and active is None:
            message = "No edit options provided"
            self.log.error(message)
            raise ParameterError(message)

        # Prepare to query the switch event table
        switch_table = SwitchEventsTable(self.db_connection)

        # Query the database for all the existing query events
        switch_events = switch_table.get_all_events()

        # Find the closest event
        closest_event = None
        closest_seconds_diff = None
        found_event = False
        for event in switch_events:
            # Skip the event ID to edit, but make sure it is found
            if event.event_id == event_id:
                found_event = True
                continue

            if event_time is not None:
                time_diff = event_time - event.time
                seconds_diff = abs(time_diff.total_seconds())
                if (closest_seconds_diff is None or
                        seconds_diff < closest_seconds_diff):
                    closest_seconds_diff = seconds_diff
                    closest_event = event

        # If the event wasn't found, it is an error
        if not found_event:
            message = "Switch event ID {0} does not exist".format(event_id)
            self.log.error(message)
            raise EventNotFoundError(message)

        # If a close event has been found (might not be one if the table is
        # empty), make sure it isn't too close
        if closest_event is not None:
            # An existing events within 6000 seconds is a fatal error
            if (closest_seconds_diff <= 6000):
                message = (
                    "Switch event ID {0} at {1} is {2} seconds away which is "
                    "not more than 6000 seconds.  New event not edited."
                    ).format(
                        closest_event.event_id, closest_event.time,
                        closest_seconds_diff)
                self.log.error(message)
                raise EventTooCloseError(message)

        # If the can_model flag is being transitioned to 'N', make sure the
        # active model for the switch event is already invalid
        if can_model is False:
            # Get the active model for the switch event
            model_table = SsmPosModelTable(self.db_connection)
            model = model_table.get_active_model(event_id)

            if model.valid:
                # Can't transition the switch event to invalid if the active
                # model is valid
                message = (
                    "Cannot change switch event ID {0} to disable the model "
                    "since model ID {1} is still valid".format(
                        event_id, model.model_id))
                self.log.error(message)
                raise ModelValidityError(message)

        # If the active flag is being transitioned to False, can_model must be
        # getting changed to False too, or it must already be False
        if active is False and can_model is not False:
            ev = switch_table.get_event(event_id)
            if ev.can_model is True:
                # Can't transition the switch event to invactive if the
                # can_model flag is set to 'Y'
                message = (
                    "Cannot change switch event ID {0} to set the event "
                    "inactive since can_model is set to 'Y'".format(
                        event_id))
                self.log.error(message)
                raise ModelValidityError(message)

        # Start a transaction
        self.db_connection.begin_transaction()

        # If requested, update the timestamp in the database for the switch
        # event
        if event_time is not None:
            switch_table.update_event_time(event_id, event_time)

        # If the can_model field is provided, update its value in the database
        if can_model is not None:
            switch_table.update_event_can_model_flag(event_id, can_model)

        # If the active field is provided, update its value in the database
        if active is not None:
            switch_table.update_event_active_flag(event_id, active)

        # Get the updated switch event
        updated_event = switch_table.get_event(event_id)

        # Commit the database changes
        self.db_connection.commit()

        return updated_event

    @staticmethod
    def add_command_line_parameters(argparser):
        """ Add the command line parsing support for manage switch events to
            ArgumentParser provided

            Parameters:
                argparser: Parent ArgumentParser to add parameters to
        """
        # Use sub-parsers for each of the sub-commands
        subparsers = argparser.add_subparsers(
            title='commands', help='available commands')

        # Work around a bug in ArgumentParser that doesn't allow a subcommand
        # to be required via the add_subparsers method
        subparsers.required = True
        subparsers.dest = 'command'

        # The add command requires a time parameter and supports an optional
        # parameter to create the model with the model disabled
        add_parser = subparsers.add_parser('add', help='Add a switch event')
        # Allow for either the YYYY-DOY-HH:MM:SS[.S] or YYYY:DOY:Seconds format
        # by calling the new valid timestamp format utility
        add_parser.add_argument(
            'time', nargs=1, type=valid_timestamp_dual_format,
            help="Event time in YYYY-DOY-HH:MM:SS[.S] or YYYY:DOY:Seconds"
                 "format")
        add_parser.add_argument(
            '--disable_model', action='store_true', default=False,
            help="Option to set the 'Can Model' field to 'N' (default = 'Y')")
        add_parser.set_defaults(command='add')

        # The list command does not require any additional parameters.  It
        # supports optional can/can't model flags, active/inactive flags, and
        # start/stop timestamp parameters.
        list_parser = subparsers.add_parser(
            'list', help='List existing switch events')
        list_parser.add_argument(
            '--reverse', action='store_true',
            help="Option to sort the list in reverse order")
        # Can/can't model parameters are mutually exclusive
        group = list_parser.add_mutually_exclusive_group()
        group.add_argument(
            '--can_model', action='store_true',
            help="Option to display only switch events that can be modeled")
        group.add_argument(
            '--cannot_model', action='store_true',
            help="Option to display only switch events that cannot be modeled")
        # Active/inactive parameters are mutually exclusive
        group = list_parser.add_mutually_exclusive_group()
        group.add_argument(
            '--active', action='store_true',
            help="Option to display only switch events that are active")
        group.add_argument(
            '--inactive', action='store_true',
            help="Option to display only switch events that are inactive")
        list_parser.add_argument(
            '--start_timestamp', type=valid_timestamp,
            help="Starting event time in YYYY:DOY:Seconds format")
        list_parser.add_argument(
            '--end_timestamp', type=valid_timestamp,
            help="Ending event time in YYYY:DOY:Seconds format")
        list_parser.set_defaults(command='list')

        # The edit command requires event ID and time parameters and supports
        # optional enable/disable model flags
        edit_parser = subparsers.add_parser(
            'edit', help="Edit an existing switch event's time")
        edit_parser.add_argument(
            'event_id', nargs=1, type=int, help='Event ID to modify')
        # Allow for either the YYYY-DOY-HH:MM:SS[.S] or YYYY:DOY:Seconds format
        # by calling the new valid timestamp format utility
        edit_parser.add_argument(
            '--time', type=valid_timestamp_dual_format,
            help="Event time in YYYY-DOY-HH:MM:SS[.S] or YYYY:DOY:Seconds "
                 "format")
        # Enable/disable model parameters are mutually exclusive
        group = edit_parser.add_mutually_exclusive_group()
        group.add_argument(
            '--enable_model', action='store_true',
            help="Option to set the 'Can Model' field to 'Y'")
        group.add_argument(
            '--disable_model', action='store_true',
            help="Option to set the 'Can Model' field to 'N'")

        # Set valid/invalid model parameters are mutually exclusive
        group = edit_parser.add_mutually_exclusive_group()
        group.add_argument(
            '--set_active', action='store_true',
            help="Option to set the 'Active' field to 'Y'")
        group.add_argument(
            '--set_inactive', action='store_true',
            help="Option to set the 'Active' field to 'N'")
        edit_parser.set_defaults(command='edit')

    def run(self, args):
        """ Run the manage events application using the arguments provided.

            Parameters:
                args: Parsed command line arguments
        """
        if args.command == "add":
            # Add a switch event.  Issue a message to indicate what is being
            # added.
            can_model = not args.disable_model
            model_message = "" if can_model else " with model disabled"
            self.log.info("Adding switch event at time {0}{1}".format(
                args.time[0], model_message))

            # Add the event
            try:
                event = self.add_event(args.time[0], can_model=can_model)
                self.display_events([event])
            except EventTooCloseError:
                raise ManageEventError("Error adding new switch event")

        elif args.command == "list":
            # List the switch events, factoring in any filtering options.
            # Convert the mutually exclusive parameters into the correct flags
            if args.can_model:
                can_model = True
            elif args.cannot_model:
                can_model = False
            else:
                can_model = None
            if args.active:
                active = True
            elif args.inactive:
                active = False
            else:
                active = None

            # Get the switch events that match the criteria
            se_table = SwitchEventsTable(self.db_connection)
            events = se_table.get_filtered_events(
                args.reverse, can_model, active,
                start_timestamp=args.start_timestamp,
                end_timestamp=args.end_timestamp)

            # Display the switch events
            self.display_events(events)

        elif args.command == "edit":
            # Edit requires the switch ID to be provided, then at least one
            # additional option to make a change
            if args.disable_model:
                can_model = False
            elif args.enable_model:
                can_model = True
            else:
                can_model = None

            if args.set_active:
                active = True
            elif args.set_inactive:
                active = False
            else:
                active = None

            # Issue message for the updates being made
            if args.time:
                self.log.info(
                    "Changing switch event ID {0} to time {1}".format(
                        args.event_id[0], args.time))
            if can_model is not None:
                change = "Enabling" if can_model else "Disabling"
                self.log.info("{0} model for switch event ID {1}".format(
                    change, args.event_id[0]))
            if active is not None:
                change = "active" if active else "inactive"
                self.log.info("Setting switch event id {1} to {0}".format(
                    change, args.event_id[0]))

            # Edit the event
            try:
                event = self.edit_event(args.event_id[0], event_time=args.time,
                                        can_model=can_model, active=active)
                self.display_events([event])
            except SsmException:
                raise ManageEventError("Error editing switch event")

        else:
            # Note: This should not be able to happen since the command line
            # arguments prevent an unknown command from being identified.
            # Raise a general exception for the problem.
            message = "Unrecognized command: {0}".format(args.command)
            self.log.error(message)
            raise ParameterError(message)


def parse_parameters(arguments):
    """ Parse the command line arguments.

        Parameters:
            arguments: List of command line arguments to parse.  This is passed
                       in as an argument instead of defaulting to sys.argv in
                       the ArgumentParser to allow unit testing to provide
                       lists to test.

        Returns:
            Arguments object with the parameters as members.
    """
    # Create the main argument parser
    argparser = argparse.ArgumentParser(
        description='Manage TIRS SSM mode switch events')

    # Add the command line parameters for ManageSwitchEvents
    ManageSwitchEvents.add_command_line_parameters(argparser)

    # Return the parsed arguments
    return argparser.parse_args(arguments)


def main():
    """ Main manage SSM switch events routine
    """
    # Parse any command line arguments, skipping the first one which is the
    # application name
    args = parse_parameters(sys.argv[1:])

    # Initialize the logger
    log = logger.Logger("manage_tirs_ssm_switch_events")
    log.info("Initiated")

    # Set up the satellite attributes for Landsat 8
    sat_attr.set_default(sat_attr.get_satellite_id_from_number(8))

    # Connect to the database
    db_connection = connect_to_database()

    # Create a manage switch events object
    manage_events = ManageSwitchEvents(db_connection)

    # Assume the operation will complete successfully
    exit_status = EXIT_SUCCESS

    # Run the manage events processing
    try:
        manage_events.run(args)
        log.info("Completed successfully")
    except ManageEventError:
        log.error("Failed to perform the requested operation")
        exit_status = EXIT_FAILURE
    finally:
        # Close the database connection.  Any updates should already be
        # committed.
        db_connection.close(db.DISCONNECT_ROLLBACK)

    # Close the log
    log.close()

    sys.exit(exit_status)


# Runs main if called from command line
if __name__ == "__main__":
    main()
