#!/usr/local/bin/python3
""" Perform initial fit

    Performs an initial fit to the telemetry right after the SSM event has
    occurred to bring the data in line such that SSM Calibration can occur
"""
from datetime import datetime
import argparse
import os
import subprocess
import sys
from database_objects import DatabaseObjects
from ias.db import database as db
from ias.log import logger
from ias.misc.const import SUCCESS
from ias.misc import satellite_attributes as sat_attr
from ias.misc.const import EXIT_SUCCESS, EXIT_FAILURE
from ssm_exceptions import (SsmException, MissingDataError, DateRangeError)
from switch_event import SwitchEventsTable
from ssm_pos_model import SsmPosModelTable
from ssm_pos_model import ModelQuality
from ssm_utilities import connect_to_database
from timestamp import Timestamp


# Define the exceptions for the errors that can be encountered
class InitialFitError(SsmException):
    """ Exception for an issue in the initial fit
    """
    pass


class InitialFit():
    """ Provides the logic to perform the initial fit
    """

    def __init__(self, db_connection, log=None):
        """ Construct a new FitSwitchEvent object.

            Parameters:
                db_connection: Already created database connection
                log: Optional logging object.  If not provided, the default
                     logger will be used.
        """
        self.sec_per_day = 86400.0

        self.db_connection = db_connection
        self.num_telemetry_records_wanted = None
        self.event_id = None
        self.event_time = None

        # SSM Position Model Table class instance
        self.model_table = SsmPosModelTable(self.db_connection)
        # Switch Events Table class instance
        self.switch_table = SwitchEventsTable(self.db_connection)
        # Telemetry table to use in telemetry query
        self.tirs_telemetry_table = DatabaseObjects.get_name(
            "tirs_command_table")

        # Use the provided log or the default one
        if log is not None:
            self.log = log
        else:
            # Get the default logger
            self.log = logger.Logger.get_default_logger()

    @staticmethod
    def add_command_line_parameters(argparser):
        """ Add the command line parsing support for fitting switch events to
            ArgumentParser provided

            Parameters:
                argparser: Parent ArgumentParser to add parameters to
        """
        argparser.add_argument(
            'fit_directory', nargs=1, type=str,
            help='Directory in which to perform the initial fit')

        argparser.add_argument(
            '--num_telemetry_records', type=int, default=300,
            help='Minimum number of telemetry records to consider the fit to '
                 'be preliminary quality (default = 300)')

    def convert_j2000_seconds_to_days_sod(self, seconds):
        """ Converts the seconds from epoch to days and seconds of day from
            epoch.

            Parameters:
                seconds: seconds from epoch

            Returns:
                j2000_day: days from epoch
                j2000_sec: seconds of day from epoch
        """
        j2000_day = int(seconds / self.sec_per_day)
        j2000_sec = int(seconds % self.sec_per_day)
        self.log.debug("seconds to convert to epoch days and epoch seconds "
                       "of day: {0}".format(seconds))
        self.log.debug("j2000_day: {0}".format(j2000_day))
        self.log.debug("j2000_sec: {0}".format(j2000_sec))
        return j2000_day, j2000_sec

    def convert_year_doy_sod_to_j2000_seconds(self, time):
        """ Converts a timestamp value in calendar year, day of year,
            seconds of day format to a value of TAI seconds since J2000.

            Parameters:
                time stamp for most recent switch event

            Returns:
                total_sec: seconds since J2000
        """
        days_in_year = 365
        days_in_leap_year = 366
        j2_epoch_doy = 1
        j2_epoch_sod = 43135.816
        j2_epoch_year = 2000

        calendar_year = time.year
        if calendar_year < j2_epoch_year:
            msg = ("Invalid year provided: {0}.".format(calendar_year))
            raise DateRangeError(msg)

        # Note the lack of leap seconds.  They were intentionally ignored
        # since being a few seconds off isn't important in this instance
        secs_from_begin_calendar_year = (float(time.doy) - 1.0) * \
            self.sec_per_day + float(time.seconds) - \
            (float(j2_epoch_doy - 1) * self.sec_per_day + j2_epoch_sod)

        # Count the number of days in the years from J2000 epoch year to the
        # given year
        days_from_j2000 = 0
        year_iterator = j2_epoch_year
        while year_iterator < calendar_year:
            # Year of reference is J2000 epoch (2000). This case is for
            # conversions of dates after the J2000 epoch (normal).
            if InitialFit.is_leap_year(year_iterator):
                days_from_j2000 += days_in_leap_year
            else:
                days_from_j2000 += days_in_year
            year_iterator = year_iterator + 1

        # Calculate the total seconds from J2000 as the seconds in each day
        # since J2000 plus the seconds calculated previously since beginning
        # of current year
        total_secs = secs_from_begin_calendar_year + \
            float(days_from_j2000 * self.sec_per_day)

        return total_secs

    def populate_event_information(self):
        """ Populates information for the most recent switch event.
        """
        events = self.switch_table.get_filtered_events(
            reverse=True, can_model=True, active=True)

        if len(events) < 1:
            msg = "No switch events found."
            raise MissingDataError(msg)

        # Since we asked for the events to be reversed, the latest one should
        # be the first one in the list
        self.event_id = events[0].event_id
        self.event_time = events[0].time
        self.log.debug("switch event id {0} event_time: {1}".
                       format(self.event_id, self.event_time))

    def get_model_quality(self, models):
        """ Obtains the model quality for the valid model for the most recent
            switch event.

            Parameters:
                models: list of models associated with most recent active
                        switch event

            Returns:
                model.quality_flag: model quality for the valid model for the
                                    most recent switch event
        """
        for model in models:
            if model.date_disabled:
                continue
            self.log.debug("model quality: {0}".format(model.quality_flag))
            return model.quality_flag
        # Should never reach this point, but just in case all of the models
        # are disabled, return predicted for the quality
        return ModelQuality.predicted

    def get_number_actual_telemetry_records(self):
        """ Get the count of actual telemetry records for this switch event

            Returns:
                total_telemetry: total count of telemetry records
        """
        # Convert the seconds from epoch to days from epoch and seconds of day
        # from epoch to use in query
        j2000_seconds = \
            self.convert_year_doy_sod_to_j2000_seconds(self.event_time)
        (j2000_day, j2000_sod) = \
            self.convert_j2000_seconds_to_days_sod(j2000_seconds)
        # Note that the query includes an arbitrary limiting end time which
        # is set to 2 days. Note: This will not impact the records used during
        # the fit.
        query_string = (
            "SELECT COUNT(*) FROM {0} WHERE "
            "((L0R_TIME_DAYS_FROM_J2000=:J2000_DAYS AND "
            "L0R_TIME_SEC_OF_DAY > :J2000_SECS) OR "
            "(L0R_TIME_DAYS_FROM_J2000 BETWEEN :J2000_DAYS + 1 AND "
            " :J2000_DAYS_2)) AND "
            "MC_ENCODER_FLAGS = 9 AND "
            "SSM_MECH_MODE = 0 AND "
            "SSM_ENCODER_POSITION_SAMPLE_2 > 0").format(
                self.tirs_telemetry_table)
        parameters = {'J2000_DAYS': j2000_day, 'J2000_SECS': j2000_sod,
                      'J2000_DAYS_2': j2000_day + 2}
        cursor = self.db_connection.execute(query_string, parameters)
        result = cursor.fetchone()
        cursor.close()
        total_telemetry = result[0]
        self.log.debug("Number of total telemetry records found from query: "
                       "{0}".format(total_telemetry))
        return total_telemetry

    @staticmethod
    def is_leap_year(year):
        """ Determine if given year is a leap year

            Parameters:
                year: year to determine if it is a leap year

            Returns:
                0: Not a leap year
                1: Is a leap year
        """
        if (year % 4 != 0) or ((year % 100 == 0) and (year % 400 != 0)):
            return 0
        else:
            return 1

    def run(self, args):
        """ Run the initial fit using the arguments provided.

            Parameters:
                args: Parsed command line arguments.
        """
        fit_directory = args.fit_directory[0]
        self.num_telemetry_records_wanted = args.num_telemetry_records
        self.log.debug("Number of records wanted in DB {0}".format(
            self.num_telemetry_records_wanted))

        # Get event information for most recent switch event
        self.populate_event_information()

        # If the current time is before the switch event, exit (nothing to do
        # since no telemetry will be available yet).  Note: The switch event
        # time is UTC, so the current time needs to be in UTC.
        now = Timestamp.from_datetime(datetime.utcnow())
        if now < self.event_time:
            self.log.debug("Current time - {0} is before event time - {1}.".
                           format(now, self.event_time))
            return

        # Get the models associated with the most recent event
        # Note using get_model_records instead of get_active_models in order
        # to get a total count of the models associated with the switch
        # event
        models = self.model_table.get_model_records(self.event_id)
        num_models = len(models)
        self.log.debug("num models found: {0} for switch event {1} starting "
                       "at time {2}".format(num_models, self.event_id,
                                            self.event_time))

        # Get model quality
        if num_models > 0:
            model_quality = self.get_model_quality(models)
            if model_quality != ModelQuality.predicted:
                # If the model quality is not "predicted", exit (nothing to do
                # since the model quality is sufficient)
                self.log.debug("Model quality sufficient for switch event: "
                               "{0} starting at time {1}, no fit needed.".
                               format(self.event_id, self.event_time))
                return

        # Query the TIRS_TELEMETRY_COMMAND table like get_db_telemetry_list in
        # the generate_ssm_encoder_file C application to determine how many
        # records of telemetry are available for the switch event
        actual_number_telemetry_records = \
            self.get_number_actual_telemetry_records()
        if actual_number_telemetry_records == 0:
            self.log.info("No matching telemetry records found for switch "
                          "event {0} starting at time {1}, cannot perform "
                          "fit.".format(self.event_id, self.event_time))
            return
        self.log.debug("Number of existing telemetry records in DB: {0} "
                       "for switch event {1} starting at time {2}".format(
                           actual_number_telemetry_records, self.event_id,
                           self.event_time))

        # If the actual number of existing telemetry records is > number_wanted
        # (configurable), set the output model quality to predicted,
        # otherwise set the output model quality to preliminary
        if (self.num_telemetry_records_wanted <=
                actual_number_telemetry_records):
            model_quality = ModelQuality.preliminary
        else:
            model_quality = ModelQuality.predicted
        self.log.debug("model_quality set to {0} for switch event {1} "
                       "starting at time {2}".format(model_quality,
                                                     self.event_id,
                                                     self.event_time))

        # If the output model quality is predicted and the model count for the
        # event is > 1, exit since the fit has already been done once with a
        # small number of telemetry records and now want to wait for at least
        # 300 telemetry records to make a model of preliminary quality.
        if model_quality == ModelQuality.predicted and num_models > 1:
            self.log.info("Fit has already been performed for switch event "
                          "{0} starting at time {1} with a small number of "
                          "telemetry records. Found {2} records this time but "
                          "need at least {3} records to perform the final "
                          "fit - not performing fit.".format(
                              self.event_id, self.event_time,
                              actual_number_telemetry_records,
                              self.num_telemetry_records_wanted))
            return

        # Run fit normally to generate output the accept option needs
        self.log.info("Performing fit for switch event {0} starting at "
                      "time {1} using {2} telemetry_records.".format(
                          self.event_id, self.event_time,
                          actual_number_telemetry_records))
        self.run_fit(fit_directory, None)
        # Now run the TIRS SSM fit with the accept option and the quality
        # set to the output model quality passed to automatically populate
        # estimated positions in the TIRS_SSM_POS_ESTIMATES table
        self.run_fit(fit_directory, model_quality)

    def run_fit(self, fit_directory, model_quality=None):
        """ Runs fit for the most recent active switch event

            Parameters:
                fit_directory: directory in which to perform fit. Fit output
                               will reside in this directory.
                model_quality: quality to use when accepting fit
        """
        if model_quality is None:
            # If model_quality is None, performing normal fit to generate
            # output that the accept option needs
            cmd = "estimate_tirs_ssm_position.py fit fit "
            cmd += str(self.event_id)
        else:
            # If the model_quality is specified, performing the accept
            cmd = "estimate_tirs_ssm_position.py fit accept "
            cmd += str(self.event_id)
            cmd += " --quality "
            cmd += ModelQuality.__str__(model_quality)
        # Set no display since the idea is to have this run as cron
        cmd += " --no_display"

        self.log.debug("cmd: {0}".format(cmd))

        if not os.path.exists(fit_directory):
            os.mkdir(fit_directory)
        curr_dir = os.getcwd()

        try:
            os.chdir(fit_directory)
            self.log.info("Running SSM model fit: {0}".format(cmd))
            subprocess.check_call([cmd], shell=True)
        except subprocess.CalledProcessError as error:
            message = (("{0} failed with return code {1} - {2}").format(
                cmd, error.returncode, error.output))
            self.log.error(message)
            raise InitialFitError(message)
        finally:
            os.chdir(curr_dir)


def parse_parameters(arguments):
    """ Parse the command line arguments.

        Parameters:
            arguments: List of command line arguments to parse.  This is passed
                       in as an argument instead of defaulting to sys.argv in
                       the ArgumentParser to allow unit testing to provide
                       lists to test.

        Returns:
            Arguments object with the parameters as members.
    """
    # Create the main argument parser
    argparser = argparse.ArgumentParser(description='Perform Initial Fit')

    # Add the command line parameters for ManageSwitchEvents
    InitialFit.add_command_line_parameters(argparser)

    # Return the parsed arguments
    return argparser.parse_args(arguments)


def main():
    """Main routine to perform initial fit
    """
    # Parse command line arguments, skipping first one which is the
    # application name
    args = parse_parameters(sys.argv[1:])

    # Initialize the logger
    log = logger.Logger("Perform Initial Fit")
    log.debug("Initiated")

    # Set up the satellite attributes for Landsat 8
    sat_attr.set_default(sat_attr.get_satellite_id_from_number(8))

    # Connect to the database
    db_connection = connect_to_database()

    # Instantiate class
    initial_fit = InitialFit(db_connection)

    # Assume the operation will complete successfully
    exit_status = EXIT_SUCCESS

    # Run the TIRS initial fit routine
    try:
        initial_fit.run(args)
        log.debug("Completed successfully")
    except SsmException as error:
        log.error("Failed to perform the initial SSM fit: {0}".format(error))
        exit_status = EXIT_FAILURE
    finally:
        # Close the database connection.
        db_connection.close(db.DISCONNECT_ROLLBACK)

    log.close()

    sys.exit(exit_status)

# Run the main routine if called from command line
if __name__ == "__main__":
    main()
