#!/usr/local/bin/python3
""" Place LPGS Order For Time Period

Places orders in LPGS for the specified time range
"""
from ias.db import database as db
from ias.log import logger
from ias.misc.const import EXIT_SUCCESS, EXIT_FAILURE
from ias.misc import satellite_attributes as sat_attr
from ssm_utilities import connect_to_database, valid_timestamp
from switch_event import SwitchEventsTable
from ssm_exceptions import (DateRangeError, EventNotFoundError,
                            ParameterError, SsmException)
from datetime import timedelta, datetime
import argparse
import sys


class PlaceLpgsOrders():
    """ Places orders in LPGS for the specified time range
    """
    def __init__(self,  db_in):
        """ Construct a new  PlaceLpgsOrders object.

            Parameters:
                db_in: Already created database connection to the WO common
                       (not L8) schema.
        """
        self.common_db_conn = db_in

    def place_order_for_time(self, start_time, stop_time):
        """ Call the stored procedure to place the orders

            Parameters:
                start_time_in: Order scenes starting at this point in time
                stop_time_in: Order scenes ending at this point in time

            Returns:
                None
        """
        sql = ("BEGIN\nDBPROC_TIRS_AOS_ENQUEUE_MSG(to_date(:START_D,"
               "'YYYY:DDD:SSSSS'),to_date(:STOP_D,'YYYY:DDD:SSSSS'));\nEND;")
        parms = {'START_D': str(start_time),
                 'STOP_D': str(stop_time)}

        # Call the procedure, the proc will rollback the transaction if
        # something goes wrong
        cursor = self.common_db_conn.execute(sql, parms)
        cursor.close()


class PlaceOrdersFrontEnd():
    """ A front end class that allows the functionality to be called from
        other scripts.
    """
    def __init__(self, db_in, log_in=None):
        """ Construct a new PlaceOrdersFrontEnd object.

            Parameters:
                db_in: Already created database connection to the WO common
                       (not L8) schema.
                log_in: Already created logger
        """
        # Use the provided log or the default one
        if log_in is not None:
            self.log = log_in
        else:
            # Get the default logger
            self.log = logger.Logger.get_default_logger()
        self.common_db_conn = db_in

    @staticmethod
    def add_command_line_parameters(argparser):
        """ Add the command line parsing support for placing LPGS orders to the
            ArgumentParser provided

            Parameters:
                argparser: Parent ArgumentParser to add parameters to
        """
        argparser.add_argument(
            '--start_time', type=valid_timestamp, help='Start time '
            '(YYYY:DDD:SSSSSS). Either both start/stop time or an event_id '
            'must be passed')

        argparser.add_argument(
            '--stop_time', type=valid_timestamp, help='Stop time '
            '(YYYY:DDD:SSSSSS). Either both start/stop time or an event_id '
            'must be passed')

        argparser.add_argument(
            '--event_id', type=int, help='Submit orders for this event id. '
            'This option may not be combined with start or stop time')

        # Name of parm is kept generic in case it will force other things
        # in the future
        argparser.add_argument(
            '--force', action='store_true',  default=False,
            help='Disable the age limit')

    def _verify_parms(self, args):
        """ This helper runs some checks on the passed arguments. These are
            checks that can not be done using the argparse functionality
            alone.

                Parameters:
                    args: Parsed command line arguments

                Returns:
                    Nothing
        """
        if (args.event_id is None and args.start_time is not None and
           args.stop_time is not None):
            if args.start_time.get_datetime() > args.stop_time.get_datetime():
                raise DateRangeError("Start time after the stop")
            if (args.start_time.get_datetime() < (datetime.now() -
               timedelta(weeks=6))):
                if not args.force:
                    raise DateRangeError("Start time is to far in the past."
                                         " See the force option")
        elif (args.event_id is not None and args.start_time is None and
              args.stop_time is None):
            try:
                # Connect to the L8 schema, so we can read the switch event
                # times
                l8_db = connect_to_database(use_common=False)
                se_table = SwitchEventsTable(l8_db)
                event = se_table.get_event(args.event_id)

                # Event must be active to order scenes for it
                if not event.active:
                    raise EventNotFoundError("Event ID {0} is not active".
                                             format(args.event_id))
                start_stop = se_table.get_start_stop_time(args.event_id)
                args.start_time = start_stop[0]
                args.stop_time = start_stop[1]
            finally:
                # That's all this connection is needed for, close it
                l8_db.close(db.DISCONNECT_ROLLBACK)
        else:
            raise ParameterError("Time range or event id (but not both) "
                                 "must be passed")

        # Limit the duration between times
        if ((args.stop_time.get_datetime() -
                args.start_time.get_datetime()) > timedelta(days=20)):
            # Passing both event_id and force bypasses this restriction
            if args.event_id is None or not args.force:
                raise DateRangeError("Time range is too large")

    def run(self, args):
        """ Run the script with the passed arguments

                Parameters:
                    args: Parsed command line arguments

                Returns:
                    Nothing
        """
        self._verify_parms(args)

        self.log.info("Initiated")

        orders = PlaceLpgsOrders(self.common_db_conn)
        self.log.info("Placing orders in LPGS for %s until %s",
                      args.start_time, args.stop_time)
        orders.place_order_for_time(args.start_time, args.stop_time)
        self.common_db_conn.commit()


def parse_parameters(arguments):
    """ Parses command line arguments for this script

            Parameters:
                arguments: List of args to parse

            Returns:
                Namespace populated with argument values
    """
    argparser = argparse.ArgumentParser(
        description='Place LPGS Orders For Time Period')

    PlaceOrdersFrontEnd.add_command_line_parameters(argparser)

    return argparser.parse_args(arguments)


def main():
    """ Main place LPGS orders routine
    """
    db_conn = None
    exit_status = EXIT_SUCCESS
    try:
        # Initialize the logger
        log = logger.Logger("Place LPGS Orders For Time Period")

        # Init satellite attributes
        sat_attr.set_default(sat_attr.get_satellite_id_from_number(8))

        parsed_args = parse_parameters(sys.argv[1:])

        # Connect to the database
        db_conn = connect_to_database(use_common=True)

        # Create and call run on the Front End object
        fe = PlaceOrdersFrontEnd(db_conn, log)
        fe.run(parsed_args)
    except SsmException as Ex:
        log.error("Exception running the Front End: %s", Ex)
        exit_status = EXIT_FAILURE
    finally:
        if db_conn is not None:
            db_conn.close(db.DISCONNECT_ROLLBACK)
        log.close()

    exit(exit_status)

# Run the main routine if called from command line
if __name__ == "__main__":
    main()
