#!/usr/bin/env python3
""" TIRS SSM Position Model Class
"""

from switch_event import SwitchEventsTable
from enum import IntEnum, unique
from database_objects import DatabaseObjects
from tirs_ssm_model_fit_parameters import ModelFitParameters
from ssm_exceptions import ModelNotFoundError, ActiveModelExistsError


@unique
class ModelQuality(IntEnum):
    """ Class to represent the model quality

        predicted(0) = generated before the event
        preliminary(1) = generated during the event
        final(2) = generated after the event
    """
    predicted = 0
    preliminary = 1
    final = 2

    def __str__(self):
        """ Provide the enumerated name for the string representation"""
        return self.name

    @classmethod
    def from_string(self, quality):
        """ Convert the quality from a string to the enumerated value"""
        if quality == "predicted":
            return ModelQuality.predicted
        elif quality == "preliminary":
            return ModelQuality.preliminary
        elif quality == "final":
            return ModelQuality.final
        else:
            raise Exception("Invalid quality provided")


class SsmPosModel():
    """ Information for a single TIRS_SSM_POSITION_MODEL_PARAMETER record

        Parameters:
            model_id: Model ID for the new object. Should be obtained from
                      the correct database sequence generator
            event_id: Event ID for the new object
            alg_version: Model algorithm version number.
            encoder_nadir_pos: Encoder Nadir Reference Count
            a0: Initial encoder offset parameter
            a1: First orbit exponential magnitude
            a2: First day exponential magnitude
            a3: First week exponential magnitude
            tau1: First orbit exponential time constant
            tau2: First day exponential time constant
            tau3: First week exponential time constant
            slope: Long term encoder trend slope
            quality_flag: quality flag
            valid: Flag that indicates whether this model can be used to
                   generate estimated SSM positions.
            date_added: Optional Date the record was added to table
            date_disabled: Optional Date the record was superseded by another
                           record.
    """
    def __init__(self, model_id, event_id, alg_version, encoder_nadir_pos,
                 a0, a1, a2, a3, tau1, tau2, tau3, slope, quality_flag, valid,
                 date_added=None, date_disabled=None):
        self.model_id = model_id
        self.event_id = event_id
        self.alg_version = alg_version
        self.encoder_nadir_pos = encoder_nadir_pos
        self.a0 = a0
        self.a1 = a1
        self.a2 = a2
        self.a3 = a3
        self.tau1 = tau1
        self.tau2 = tau2
        self.tau3 = tau3
        self.slope = slope
        self.quality_flag = quality_flag
        self.valid = valid
        self.date_added = date_added
        self.date_disabled = date_disabled

    def get_valid_flag(self):
        """ Return the get_valid_flag member as a Y/N flag """
        return 'Y' if self.valid else 'N'

    def to_list(self):
        """ Return the fields as a list """
        return [self.model_id, self.event_id, self.alg_version,
                self.encoder_nadir_pos, self.a0, self.a1, self.a2, self.a3,
                self.tau1, self.tau2, self.tau3, self.slope, self.quality_flag,
                self.valid, self.date_added, self.date_disabled]

    def __str__(self):
        """ Return a user readable string representing the object """
        return ("model_id:{0}, event_id:{1}, alg_version:{2}, "
                "encoder_nadir_pos:{3}, a0:{4}, a1:{5}, a2:{6}, a3:{7}, "
                "tau1:{8}, tau2:{9}, tau3:{10}, slope:{11}, valid:{12}, "
                "quality_flag:{13}, date_added:{14}, date_disabled:{15}".
                format(self.model_id, self.event_id, self.alg_version,
                       self.encoder_nadir_pos, self.a0, self.a1, self.a2,
                       self.a3, self.tau1, self.tau2, self.tau3, self.slope,
                       self.get_valid_flag(), self.quality_flag,
                       self.date_added, self.date_disabled))


class SsmPosModelTable():
    """ Encapsulates access to the TIRS_SSM_POS_MODEL_PARAMS database table
    """
    # Define the names of the columns in the table
    _table_columns = [
        "MODEL_ID", "SWITCH_EVENT_ID", "ALGORITHM_VERSION",
        "ENCODER_NADIR_POSITION", "A0", "A1", "A2", "A3", "TAU1", "TAU2",
        "TAU3", "SLOPE", "QUALITY_FLAG", "VALID", "DATE_ADDED",
        "DATE_DISABLED"]

    def __init__(self, db_connection):
        """ Construct a new SsmPosModelTable object.

            Parameters:
                db_connection: Connection to the database schema with access
                               to the table
        """
        # Get the table and sequence names
        self.table_name = DatabaseObjects.get_name("position_model_table")
        self.model_id_sequence = DatabaseObjects.get_name("model_id_sequence")

        self.db_connection = db_connection

    def _create_model_from_db_row(self, row):
        """ Helper routine to convert a row returned from the database into
            a model.

            Parameters:
                row: database row returned from a DB query with the fields
                     in the order defined by _table_columns

            Returns:
                The newly created model
        """
        # Convert the Y/N flags to booleans
        valid = True if row[13] == 'Y' else False

        # Really do not have to store these in a variable before returning
        # but I chose to simply because I think it makes it easier to
        # understand exactly what is happening, instead of returning row[n]
        model_id = row[0]
        event_id = row[1]
        alg_version = row[2]
        encoder_nadir_pos = row[3]
        a0 = row[4]
        a1 = row[5]
        a2 = row[6]
        a3 = row[7]
        tau1 = row[8]
        tau2 = row[9]
        tau3 = row[10]
        slope = row[11]
        quality_flag = ModelQuality(row[12])
        valid = True if row[13] == 'Y' else False
        date_added = row[14]
        date_disabled = row[15]

        return SsmPosModel(model_id, event_id, alg_version, encoder_nadir_pos,
                           a0, a1, a2, a3, tau1, tau2, tau3, slope,
                           quality_flag, valid, date_added, date_disabled)

    def disable_model_record(self, model_id):
        """ Disables a model_record by setting the date_disabled

            Parameters:
                model_id: Model ID to disable

            Notes:
                Throws an exception if the corresponding model does not
                exist in the database
        """
        # Verify model exists before trying to update
        model = self.get_model(model_id)

        # Build the update statement
        update_statement = (
            "update {0} set DATE_DISABLED=sysdate where MODEL_ID=:MODEL_ID "
            "and DATE_DISABLED IS NULL").format(self.table_name)

        # Prepare and execute update statement
        parameters = {'MODEL_ID': model_id}
        cursor = self.db_connection.execute(update_statement, parameters)
        row_count = cursor.rowcount
        cursor.close()

    def update_model_validity(self, model_id, valid):
        """ Updates a model's record valid field to the value provided.

            Parameters:
                model_id: Model ID to update
                valid: Boolean flag for the new validity value

            Notes:
                Throws an exception if the wrong number of rows are updated
        """
        # DB expects Y/N, set to N to disable
        valid_flag = "Y" if valid else "N"

        # Build the update statement
        update_statement = ("update {0} set VALID=:VALID where "
                            "MODEL_ID=:MODEL_ID").format(self.table_name)

        # Prepare and execute update statement
        parameters = {'VALID': valid_flag, 'MODEL_ID': model_id}

        cursor = self.db_connection.execute(update_statement, parameters)
        row_count = cursor.rowcount
        cursor.close()

        # Error check that the update was successful
        if row_count != 1:
            raise ModelNotFoundError(
                "Model ID {0} does not exist".format(model_id))

    def get_next_model_id(self):
        """ Get the next available model ID

            Returns:
                The next available model ID
        """
        query = "select {0}.nextval from dual".format(self.model_id_sequence)

        cursor = self.db_connection.execute(query)
        all_rows = cursor.fetchall()
        cursor.close()
        return all_rows[0][0]

    def get_model_records(self, event_id=None, valid=None):
        """ Get all records for the provided event_id

            Parameters:
                event_id: Optional Switch event ID associated with the
                          model records. If not provided, all model records
                          in the table will be retrieved.
                valid: Optional Flag indicating whether the model can be used
                       to generate estimated SSM positions.  If True, only
                       records where valid is 'Y' will be retrieved. If false,
                       only records where valid is 'N' will be retrieved. If
                       not provided, the valid flag will be ignored when
                       retrieving records from the database.
        """
        wanted_valid = None
        if valid is True:
            # Only want records where valid is 'Y'
            wanted_valid = 'Y'
        elif valid is False:
            # Only want records where valid is 'N'
            wanted_valid = 'N'

        if event_id:
            where_clause = " where SWITCH_EVENT_ID=:SWITCH_EVENT_ID"
            parameters = {'SWITCH_EVENT_ID': event_id}
        else:
            where_clause = ""
            parameters = {}

        query = "select {0} from {1}{2} order by MODEL_ID desc".format(
            ", ".join(self._table_columns), self.table_name, where_clause)

        model_params = []

        cursor = self.db_connection.execute(query, parameters)
        for row in cursor:
            # Skip this record if it doesn't match 'wanted_valid'
            if wanted_valid is not None and wanted_valid != row[13]:
                continue
            model_params.append(self._create_model_from_db_row(row))
        cursor.close()

        return model_params

    def get_active_model(self, event_id):
        """ Get the active model for the requested event id.

            Parameters:
                event_id: Event ID to retrieve the active model for

            Returns:
                A single model object or throws a ModelNotFoundError exception
                if a single row isn't retrieved.
        """
        query = ("select {0} from {1} where SWITCH_EVENT_ID=:SWITCH_EVENT_ID "
                 "and DATE_DISABLED IS NULL").format(", ".join(
                    self._table_columns), self.table_name)

        model_params = []

        parameters = {'SWITCH_EVENT_ID': event_id}

        cursor = self.db_connection.execute(query, parameters)

        rows = cursor.fetchall()
        cursor.close()

        if len(rows) != 1:
            raise ModelNotFoundError("No active model found for Event ID {0}".
                                     format(event_id))
        return self._create_model_from_db_row(rows[0])

    def get_model(self, model_id):
        """ Get the model for the requested model id.

            Parameters:
                model_id: Model ID to retrieve

            Returns:
                A single model object or throws a ModelNotFoundError exception
                if a single row isn't retrieved.
        """
        # Query for the model ID requested
        query = "select {0} from {1} where MODEL_ID=:MODEL_ID ".format(
            ", ".join(self._table_columns), self.table_name)

        parameters = {'MODEL_ID': model_id}

        cursor = self.db_connection.execute(query, parameters)
        rows = cursor.fetchall()
        cursor.close()

        # Expect a single row to be returned
        if len(rows) != 1:
            raise ModelNotFoundError(
                "Model ID {0} does not exist".format(model_id))

        # Create a model record
        return self._create_model_from_db_row(rows[0])

    def insert_model_record(self, model_record):
        """Inserts a single model record into TIRS_SSM_POS_MODEL_PARMS
        """
        # If there are other models for the switch event ID that are active,
        # it is an error to try inserting a record that will be made active
        query = ("SELECT count(*) from {0} where SWITCH_EVENT_ID=:EVENT_ID "
                 "and DATE_DISABLED IS NULL").format(self.table_name)
        parameters = {'EVENT_ID': model_record.event_id}

        cursor = self.db_connection.execute(query, parameters)
        rows = cursor.fetchall()
        cursor.close()

        if len(rows) != 1:
            # This should never happen
            raise Exception("Failed counting active rows")
        if rows[0][0] != 0:
            raise ActiveModelExistsError(
                "Attempted to insert a new active model when an active one "
                "already exists")

        # Build the query binding location numbers excluding date_added and
        # date_disabled
        bind_columns = list(self._table_columns)
        bind_columns.remove("DATE_ADDED")
        bind_columns.remove("DATE_DISABLED")
        bind_indices = [":" + str(x) for x in range(1, len(bind_columns) + 1)]

        # Build the insert statement using sysdate for DATE_ADDED,
        # DATE_DISABLED will remain null
        insert_statement = (
            "insert into {0} (DATE_ADDED, {1}) values (sysdate, {2})").format(
                self.table_name, ", ".join(bind_columns),
                ", ".join(bind_indices))

        # Create a cursor and prepare the insert statement
        cursor = self.db_connection.get_cursor()
        cursor.prepare(insert_statement)

        # Insert the new model record
        cursor.execute(
            None, [model_record.model_id, model_record.event_id,
                   model_record.alg_version, model_record.encoder_nadir_pos,
                   model_record.a0, model_record.a1, model_record.a2,
                   model_record.a3, model_record.tau1, model_record.tau2,
                   model_record.tau3, model_record.slope,
                   model_record.quality_flag.value,
                   model_record.get_valid_flag()])
        cursor.close()

    def get_default_model(self, timestamp, switch_events=None):
        """ Get the default model for a given timestamp. The nearest model
            available before the time that is valid and active with the final
            quality is considered the default. If one doesn't exist, get the
            default model from the default parameters.

            Parameters:
                timestamp: timestamp for which to retrieve default model
                switch_events: switch_events for which to retrieve the
                               default model

            Notes: This doesn't follow version 4 of the ADD. It was a
                    modification suggested by cal/val.
        """

        # If an array of switch events wasn't provided, get the full list
        if switch_events is None:
            switch_table = SwitchEventsTable(self.db_connection)
            switch_events = switch_table.get_all_events()

        # Find the closest switch event before or at the request time that
        # can be modeled and is active
        found_index = -1
        for index in range(len(switch_events)):
            event = switch_events[index]
            if event.can_model and event.active:
                if timestamp >= event.time:
                    found_index = index
                else:
                    break

        # If a valid switch event was found, find the nearest previous active
        # model with a final quality
        if found_index >= 0:
            # Look for the closest final model before this event
            for index in range(found_index, -1, -1):
                event = switch_events[index]
                if event.can_model and event.active:
                    try:
                        model = self.get_active_model(event.event_id)
                        if (model.valid and
                                model.quality_flag == ModelQuality.final):
                            # reset the model and event IDs
                            model.model_id = 0
                            model.event_id = 0
                            return model
                    except ModelNotFoundError:
                        # If there is no active model for the switch event
                        # it can be ignored
                        pass

        # No useable model found, so return a default model
        model_id = 0
        event_id = 0
        parms = ModelFitParameters.default_model_parameters
        valid = True
        return SsmPosModel(
            model_id, event_id, ModelFitParameters.default_algorithm_version,
            ModelFitParameters.default_encoder_nadir_position,
            parms[0], parms[1], parms[2], parms[3], parms[4], parms[5],
            parms[6], parms[7], ModelQuality.predicted, valid)
