#!/usr/bin/env python3

from timestamp import Timestamp
from datetime import datetime, timedelta
from database_objects import DatabaseObjects
from ssm_exceptions import EventNotFoundError
from tirs_ssm_model_fit_parameters import ModelFitParameters


class SwitchEvent():
    """ Encapsulates the information about each TIRS SSM mode zero switch event
    """

    def __init__(self, event_id, timestamp, can_model, active):
        """ Construct a new SwitchEvent object.

            Parameters:
                event_id: Event ID for the new object.  Should be obtained from
                          the correct database sequence generator.
                timestamp: The timestamp for when the event occurred.
                can_model: Boolean flag that indicates the switch event can
                           be properly modeled.
                active: Boolean flag that indicates the switch event is active.
        """
        self.event_id = event_id
        self.time = timestamp
        self.can_model = can_model
        self.active = active

    @classmethod
    def from_list(cls, input_list):
        """ Create a switch event from a list

            Parameters:
                input_list: Switch event as [event ID, year, doy, seconds,
                                can_model, active]
        """
        timestamp = Timestamp(input_list[1], input_list[2], input_list[3])
        return cls(input_list[0], timestamp, input_list[4], input_list[5])

    def get_can_model_flag(self):
        """ Return the can_model member as a Y/N flag """
        return 'Y' if self.can_model else 'N'

    def get_active_flag(self):
        """ Return the active member as a Y/N flag """
        return 'Y' if self.active else 'N'

    def __eq__(self, other):
        """ Equal operator for switch events """
        if (self.event_id == other.event_id and self.time == other.time and
                self.can_model == other.can_model and
                self.active == other.active):
            return True
        else:
            return False

    def __str__(self):
        """ Return a user readable string representing the object """
        return ("id:{0}, time:{1}, can_model:{2}, active:{3}".
                format(self.event_id, self.time, self.get_can_model_flag(),
                       self.get_active_flag()))

    def __repr__(self):
        """ Return the representation of the object """
        return ("SwitchEvent({0}, {1}, {2}, {3})".format(
            self.event_id, self.time, self.can_model, self.active))


class SwitchEventsTable():
    """ Encapsulates access to the TIRS_SSM_MODE_SWITCH_EVENTS database table
    """

    # Define the names of the columns in the table
    _table_columns = ["SWITCH_EVENT_ID", "YEAR", "DAY_OF_YEAR",
                      "SECONDS_OF_DAY", "CAN_MODEL", "ACTIVE"]

    def __init__(self, db_connection):
        """ Construct a new SwitchEventsTable object.

            Parameters:
                db_connection: Connection to the database schema with access
                               to the table
        """
        self.db_connection = db_connection

        # Get the table and sequence names
        self.table_name = DatabaseObjects.get_name("mode_switch_events_table")
        self.switch_id_sequence = DatabaseObjects.get_name(
            "switch_event_id_sequence")

    def _create_event_from_db_row(self, row):
        """ Helper routine to convert a row returned from the database into
            a switch event.

            Parameters:
                row: database row returned from a DB query with the fields
                     in the order defined by _table_columns

            Returns:
                The newly created switch event
        """
        # Build the timestamp and convert the Y/N flags to booleans
        timestamp = Timestamp(row[1], row[2], row[3])
        can_model = True if row[4] == 'Y' else False
        active = True if row[5] == 'Y' else False

        # Create the switch event and append it to the list
        return SwitchEvent(row[0], timestamp, can_model, active)

    def get_all_events(self):
        """ Get all the events present in the table

            Returns:
                All the switch events retrieved
        """
        # Query for all the events from the table
        query = ("Select {0} from {1} order by year, day_of_year, "
                 "seconds_of_day").format(", ".join(self._table_columns),
                                          self.table_name)
        cursor = self.db_connection.execute(query)

        # Build a list of all the events returned by the query
        switch_events = []
        for row in cursor:
            switch_events.append(self._create_event_from_db_row(row))

        cursor.close()

        return switch_events

    def insert_event(self, switch_event):
        """ Inserts a single switch event into the switch events table

            Parameters:
                switch_event: Switch event to insert into the table
        """
        # Build the query binding location numbers
        bind_indices = [":" + str(x) for x in range(
            1, len(self._table_columns) + 1)]

        # Build the insert statement
        query = "insert into {0} ({1}) values ({2})".format(
            self.table_name, ", ".join(self._table_columns),
            ", ".join(bind_indices))

        # Create a cursor and prepare the insert statement
        cursor = self.db_connection.get_cursor()
        cursor.prepare(query)

        # Insert the new event
        cursor.execute(None,
                       [switch_event.event_id, switch_event.time.year,
                        switch_event.time.doy, switch_event.time.seconds,
                        switch_event.get_can_model_flag(),
                        switch_event.get_active_flag()])
        cursor.close()

    def update_event_time(self, event_id, timestamp):
        """ Updates a existing switch event's time

            Parameters:
                event_id: Switch event ID to update
                timestamp: Updated timestamp
        """

        # Build the update statement
        update_statement = """update {0} set year=:1, day_of_year=:2,
            seconds_of_day=:3 where switch_event_id=:4""".format(
            self.table_name)

        # Perform the update
        cursor = self.db_connection.get_cursor()
        cursor.prepare(update_statement)
        cursor.execute(None,
                       [timestamp.year, timestamp.doy, timestamp.seconds,
                        event_id])
        row_count = cursor.rowcount
        cursor.close()

        # Throw an exception if a single row wasn't updated
        if row_count != 1:
            raise EventNotFoundError(
                "ERROR: Updated {0} rows when one should have been".format(
                    row_count))

    def update_event_can_model_flag(self, event_id, can_model):
        """ Updates an existing switch event's can_model column

            Parameters:
                event_id: Switch event ID to update
                can_model: New boolean can_model value to set
        """

        # Build the update statement
        update_statement = """update {0} set can_model=:1 where
            switch_event_id=:2""".format(self.table_name)

        # Convert the can_model boolen to a Y/N flag
        can_model_flag = 'Y' if can_model else 'N'

        # Perform the update query
        cursor = self.db_connection.get_cursor()
        cursor.prepare(update_statement)
        cursor.execute(None, [can_model_flag, event_id])
        row_count = cursor.rowcount
        cursor.close()

        # Throw an exception if a single row wasn't updated
        if row_count != 1:
            raise EventNotFoundError(
                "ERROR: Updated {0} rows when one should have been".format(
                    row_count))

    def update_event_active_flag(self, event_id, active):
        """ Updates an existing switch event's active column

            Parameters:
                event_id: Switch event ID to update
                active: New boolean active value to set
        """

        # Build the update statement
        update_statement = """update {0} set active=:1 where
            switch_event_id=:2""".format(self.table_name)

        # Convert the active boolen to a Y/N flag
        active_flag = 'Y' if active else 'N'

        # Perform the update query
        cursor = self.db_connection.get_cursor()
        cursor.prepare(update_statement)
        cursor.execute(None, [active_flag, event_id])
        row_count = cursor.rowcount
        cursor.close()

        # Throw an exception if a single row wasn't updated
        if row_count != 1:
            raise EventNotFoundError(
                "ERROR: Updated {0} rows when one should have been".format(
                    row_count))

    def get_start_stop_time(self, event_id):
        """ Get the start and stop timestamps for switch event ID.  The start
            time is the time of the event.  The stop time is the time of the
            next active event if it exists.  If the next active event does not
            exist, ModelFitParameters.current_switch_event_days is added to
            the start time to get the stop time.

            Parameters:
                event_id: Switch event ID for the start/stop timestamps

            Returns:
                Tuple of (start_timestamp, stop_timestamp, complete_event)
                Note: complete_event is a boolean that indicates whether
                the event was complete and the stop time came from an
                existing subsequent event or the stop time was determined
                using utcnow
        """
        # First set the start timestamp, which is just the time of the
        # current event
        curr_event = self.get_event(event_id)
        start_timestamp = curr_event.time

        # Now find the stop timestamp
        # Query the database for all the existing query events
        events = self.get_all_events()

        # Only keep events which are active and fall after the requested event
        events = [event for event in events if event.active is True and
                  event.time > start_timestamp]

        # Is there a subsequent switch event from which we can garner the
        # stop time
        complete_event = False
        # Were any events found?
        event_count = len(events)
        if event_count > 0:
            # Found the next event, use it as the stop timestamp
            stop_timestamp = events[0].time
            complete_event = True
        else:
            # Start with the starting timestamp
            stop_timestamp = start_timestamp.get_datetime()
            # Add days from the model fit parameters
            stop_time = stop_timestamp + timedelta(
                days=ModelFitParameters.current_switch_event_days)
            stop_timestamp = Timestamp.from_datetime(stop_time)

        return (start_timestamp, stop_timestamp, complete_event)

    def get_next_event_id(self):
        """ Get the next available switch event ID

            Returns:
                The next switch event ID
        """
        query = "SELECT {0}.nextval FROM dual".format(self.switch_id_sequence)

        cursor = self.db_connection.execute(query)
        all_rows = cursor.fetchall()
        cursor.close()

        return all_rows[0][0]

    def get_event(self, event_id):
        """ Get a single switch event

            Parameters:
                event_id: Switch event ID to get

            Returns:
                The switch event retrieved
        """
        # Construct query to retrieve the switch event
        query = (
            "select {0} from {1} where SWITCH_EVENT_ID = :SWITCH_EVENT_ID"
            ).format(", ".join(self._table_columns), self.table_name)
        parameters = {'SWITCH_EVENT_ID': event_id}

        # Query the database
        cursor = self.db_connection.execute(query, parameters)
        rows = cursor.fetchall()
        cursor.close()

        # Make sure only one event was retrieved
        if (len(rows) == 1):
            return self._create_event_from_db_row(rows[0])
        else:
            raise EventNotFoundError(
                "No information found for event_id {0}".format(event_id))

    def get_filtered_events(self, reverse=False, can_model=None,
                            active=None, start_timestamp=None,
                            end_timestamp=None):
        """ Query the switch events table for all the events in time order,
            then filter them based on the requested filtering methods.

            Parameters:
                reverse: Optional flag to reverse the list of events
                can_model: Optional flag to filter the list on the can_model
                           field being True/False
                active: Optional flag to filter the list on the active field
                        being True/False
                start_timestamp: Optional earliest timestamp to return
                end_timestamp: Optional latest timestamp to return

            Returns:
                The list of filtered switch events.
        """
        # Query the database for all the existing query events
        events = self.get_all_events()

        # Keep only events that can/can't be modeled
        if can_model is True:
            events = [event for event in events if event.can_model]
        elif can_model is False:
            events = [event for event in events if not event.can_model]

        # Keep only events that are/aren't active
        if active is True:
            events = [event for event in events if event.active]
        elif active is False:
            events = [event for event in events if not event.active]

        # Eliminate events before the start timestamp
        if start_timestamp:
            events = [event for event in events
                      if event.time >= start_timestamp]

        # Eliminate events after the end timestamp
        if end_timestamp:
            events = [event for event in events if event.time <= end_timestamp]

        # Reverse the events if requested
        if reverse:
            events.reverse()

        return events
