import re


class DatabaseStructure():
    """ Provides functionality to copy the structure of database objects to new
        names.  The primary use for this is for unit testing.
    """

    @staticmethod
    def copy_object(db_connection, source_object, target_name,
                    object_type, extra_changes=None):
        """ Copy the structure of a database object to a new object.

            This uses Oracle specific features to get the definition of the
            object being copied.

            Parameters:
                db_connection: connection to the database
                source_object: Name of object to copy
                target_name: Name to use for the copy
                object_type: Type of object (i.e. TABLE or SEQUENCE)
                extra_changes: A dictionary of additional strings to change
                               in the copied object.  For example, a foreign
                               key column may need to change table names.  This
                               may be a regular expression.
        """

        # Query for the object structure
        query = ("select dbms_metadata.get_ddl('{0}','{1}','{2}') from "
                 "dual".format(object_type.upper(), source_object.upper(),
                               db_connection.username.upper()))

        cursor = db_connection.execute(query)
        all_rows = cursor.fetchall()

        # Should only be one row
        if (len(all_rows) != 1):
            cursor.close()
            raise Exception("Expected 1 row, but got {0}".format(len(
                all_rows)))

        # Convert the resulting Oracle.LOB object to a string and close the
        # cursor
        object_description = all_rows[0][0].read()
        cursor.close()

        # print(object_description)

        # Replace the original object name with the new one
        object_description = object_description.replace(
            source_object.upper(), target_name.upper())

        # If extra changes have been requested, perform those now
        if extra_changes:
            for source, value in extra_changes.items():
                object_description = re.sub(
                    source, value.upper(), object_description,
                    flags=re.IGNORECASE)

        # print(object_description)

        # Create the new object
        cursor = db_connection.execute(object_description)
        cursor.close()

    @staticmethod
    def drop_object(db_connection, name, object_type, ignore_error=False):
        """ Drop an existing object from the database """
        drop_format = "drop {0} {1}"
        if ignore_error:
            # Allow ignoring errors for cases where dropping several objects
            # that might not exist
            drop_format = ("BEGIN EXECUTE IMMEDIATE 'drop {0} {1}'; "
                           "EXCEPTION WHEN OTHERS THEN NULL; END;")
        query = drop_format.format(object_type.upper(), name.upper())
        cursor = db_connection.execute(query)
        cursor.close()
