#!/usr/bin/env python3
""" DatabaseTestTables
    Class to setup and cleanup test database tables
"""

from database_objects import DatabaseObjects
from database_structure import DatabaseStructure
from ias.db import database as db


class DatabaseTestTables():

    def __init__(self, db_connection, switch=True, model=False,
                 position=False, telemetry=False, table_suffix=None):

        self.db_connection = db_connection

        self.test_model_table = "TEST_POS_MODEL_PARAMS"
        self.test_model_id_seq = "DBSEQ_TEST_POS_MODEL_ID_PK"
        # Note that the POS table has a foreign key to the model table, so
        # if the model table is cleaned, this one has to be as well
        self.test_pos_table = "TEST_SSM_POS_ESTIMATES"
        self.test_switch_table = "TEST_MODE_SWITCH_EVENTS"
        self.test_tirs_command_table = "TEST_TIRS_TELEMETRY_COMMAND"
        self.test_switch_id_seq = "DBSEQ_TEST_SWITCH_ID_PK"

        self.model = model
        self.position = position
        self.switch = switch
        self.telemetry = telemetry
        self.table_suffix = ""
        if table_suffix:
            self.table_suffix = table_suffix
            self.test_model_table += table_suffix
            self.test_model_id_seq += table_suffix
            self.test_pos_table += table_suffix
            self.test_switch_table += table_suffix
            self.test_switch_id_seq += table_suffix
            self.test_tirs_command_table += table_suffix

    def cleanup_database(self, ignore_error):
        # Remove the position table first since it has a foreign key to the
        # switch event table
        if self.position:
            DatabaseStructure.drop_object(
                self.db_connection, self.test_pos_table, "TABLE",
                ignore_error)
        # Remove the model table first since it has a foreign key to the switch
        # event table
        if self.model:
            DatabaseStructure.drop_object(
                self.db_connection, self.test_model_table, "TABLE",
                ignore_error)
            DatabaseStructure.drop_object(
                self.db_connection, self.test_model_id_seq, "SEQUENCE",
                ignore_error)
        if self.switch:
            DatabaseStructure.drop_object(
                self.db_connection, self.test_switch_table, "TABLE",
                ignore_error)
            DatabaseStructure.drop_object(
                self.db_connection, self.test_switch_id_seq, "SEQUENCE",
                ignore_error)
        if self.telemetry:
            DatabaseStructure.drop_object(
                self.db_connection, self.test_tirs_command_table, "TABLE",
                ignore_error)

    def setup_database(self):
        if self.switch:
            DatabaseStructure.copy_object(
                self.db_connection, "TIRS_SSM_MODE_SWITCH_EVENTS",
                self.test_switch_table, "TABLE",
                extra_changes={
                    "CONSTRAINT \"TIRS_SSM_MODE_SWITCH":
                        "CONSTRAINT \"TIRS_SSM_MODE" + self.table_suffix})
            DatabaseStructure.copy_object(
                self.db_connection, "DBSEQ_TIRS_SSM_SWITCH_ID_PK",
                self.test_switch_id_seq, "SEQUENCE",
                extra_changes={"START WITH \d+": "START WITH 1"})
            # Configure the system with the test table names
            DatabaseObjects.set_name(
                "mode_switch_events_table", self.test_switch_table)
            DatabaseObjects.set_name(
                "switch_event_id_sequence", self.test_switch_id_seq)
        if self.model:
            DatabaseStructure.copy_object(
                self.db_connection, "TIRS_SSM_POS_MODEL_PARAMS",
                self.test_model_table, "TABLE",
                extra_changes={
                    'TIRS_SSM_MODE_SWITCH_EVENTS': self.test_switch_table,
                    "CONSTRAINT \"TIRS_SSM_POS_MODEL":
                        "CONSTRAINT \"TIRS_SSM_MODEL" + self.table_suffix})
            DatabaseStructure.copy_object(
                self.db_connection, "DBSEQ_TIRS_SSM_POS_MODEL_ID_PK",
                self.test_model_id_seq, "SEQUENCE",
                extra_changes={"START WITH \d+": "START WITH 10"})
            DatabaseObjects.set_name(
                "position_model_table", self.test_model_table)
            DatabaseObjects.set_name(
                "model_id_sequence", self.test_model_id_seq)
        if self.position:
            DatabaseStructure.copy_object(
                self.db_connection, "TIRS_SSM_POS_ESTIMATES",
                self.test_pos_table, "TABLE",
                extra_changes={
                    'TIRS_SSM_POS_MODEL_PARAMS': self.test_model_table,
                    "CONSTRAINT \"TEST_SSM_POS_ESTIMATES":
                        "CONSTRAINT \"TEST_SSM_EST" + self.table_suffix})
            DatabaseObjects.set_name(
                "position_estimates_table", self.test_pos_table)
        if self.telemetry:
            DatabaseStructure.copy_object(
                self.db_connection, "TIRS_TELEMETRY_COMMAND",
                self.test_tirs_command_table, "TABLE",
                extra_changes={
                    "CONSTRAINT \"TIRS_TELEMETRY_CMD":
                        "CONSTRAINT \"TIRS_TELEM_CMD" +
                        self.table_suffix})
            DatabaseObjects.set_name(
                "tirs_command_table", self.test_tirs_command_table)
