#!/usr/bin/env python3
###############################################################################
# Name: test_estimated_ssm_position.py
#
# Description: Contains unit tests for the classes in estimated_ssm_position.py
###############################################################################

import unittest
import os
from timestamp import Timestamp
from estimated_ssm_position import EstimatedSsmPosition
from estimated_ssm_position import EstimatedSsmPositionFile
from estimated_ssm_position import EstimatedSsmPositionTable
from ssm_pos_model import ModelQuality
from ias.db import database as db
from database_objects import DatabaseObjects
from database_test_tables import DatabaseTestTables
from ssm_utilities import connect_to_database
from ias.misc import satellite_attributes as sat_attr
from ssm_pos_model import SsmPosModelTable
from manage_tirs_ssm_switch_events import ManageSwitchEvents
from switch_event import SwitchEventsTable
from datetime import datetime
from ssm_exceptions import TimestampError


class TestEstimatedPosition(unittest.TestCase):

    timestamp = Timestamp(2014, 200, 3000)
    position = 12345678
    quality = ModelQuality(2)
    add_date = datetime(2014, 5, 20)
    dis_date = datetime(2014, 6, 20)

    def test_constructor(self):
        x = EstimatedSsmPosition(
            self.timestamp, self.position, self.quality, self.add_date,
            self.dis_date)
        self.assertEqual(x.timestamp, self.timestamp)
        self.assertEqual(x.position, self.position)
        self.assertEqual(x.quality_flag, self.quality)
        self.assertEqual(x.date_added, self.add_date)
        self.assertEqual(x.date_disabled, self.dis_date)

        y = EstimatedSsmPosition(
            self.timestamp, self.position, self.quality)
        self.assertTrue(y.date_added is None)
        self.assertTrue(y.date_disabled is None)

    def test_str(self):
        x = EstimatedSsmPosition(
            self.timestamp, self.position, self.quality, self.add_date,
            self.dis_date)
        y = EstimatedSsmPosition(self.timestamp, self.position, self.quality)

        expected_string1 = (
            "Timestamp: 2014:200:3000, Position: 12345678, Quality: final, "
            "Date Added: 2014-05-20 00:00:00, "
            "Date Disabled: 2014-06-20 00:00:00")
        expected_string2 = (
            "Timestamp: 2014:200:3000, Position: 12345678, Quality: final, "
            "Date Added: None, Date Disabled: None")

        self.assertEqual(str(x), expected_string1)
        self.assertEqual(str(y), expected_string2)

    def test_eq(self):
        x = EstimatedSsmPosition(
            self.timestamp, self.position, self.quality, self.add_date,
            self.dis_date)
        y = EstimatedSsmPosition(
            self.timestamp, self.position, self.quality, self.add_date,
            self.dis_date)
        self.assertTrue(x == y)

        # Verify the date fields are not involved in the equal comparison
        y = EstimatedSsmPosition(
            self.timestamp, self.position, self.quality)
        self.assertTrue(x == y)

        # Verify unequal objects are reported not equal
        y.timestamp = Timestamp(2014, 201, 5)
        self.assertFalse(x == y)

        y.timestamp = self.timestamp
        y.position = 321
        self.assertFalse(x == y)

        y.position = self.position
        y.quality_flag = ModelQuality(1)
        self.assertFalse(x == y)

        y.quality_flag = self.quality
        self.assertTrue(x == y)


class TestEstimatedSsmPositionFile(unittest.TestCase):
    test_data_dir = "./estimated_ssm_position_test_data/"
    format1 = test_data_dir+"valid_format1.data"
    format0 = test_data_dir+"valid_format0.data"

    def test_constructor(self):
        x = EstimatedSsmPositionFile(self.format1)
        self.assertEqual(x.filename, self.format1)
        self.assertTrue(x.quality_flag is None)
        x = EstimatedSsmPositionFile(self.format1, ModelQuality.preliminary)
        self.assertEqual(x.quality_flag.value, ModelQuality.preliminary)

    def test_get_file_version(self):
        x = EstimatedSsmPositionFile(self.format1)
        self.assertEqual(x._get_file_version(), 1)
        x = EstimatedSsmPositionFile(self.format0)
        self.assertEqual(x._get_file_version(), 0)

    def test_write(self):
        outfile = "write_method_output.txt"
        expected_file = (
            "estimated_ssm_position_test_data/write_method_output.txt")
        x = EstimatedSsmPositionFile(outfile)

        # Write should automatically sort its positions, so create the
        # timestamps out of order
        ts1 = Timestamp(2015, 358, 30650)
        ts2 = Timestamp(2014, 353, 59918)
        ts3 = Timestamp(2014, 350, 70000)
        ts4 = Timestamp(2013, 360, 60600)

        ref_positions = [
            EstimatedSsmPosition(ts1, 16572652, ModelQuality.final),
            EstimatedSsmPosition(ts2, 16573028, ModelQuality.preliminary),
            EstimatedSsmPosition(ts3, 16573117, ModelQuality.final),
            EstimatedSsmPosition(ts4, 16573128, ModelQuality.predicted),
        ]

        x.write(ref_positions)
        self.assertTrue(self.compare_files(outfile, expected_file))
        os.remove(outfile)

        with self.assertRaises(ValueError):
            x.write([])

    def compare_files(self, compare, expected):
        """ We need a separate function to compare files we're writing because
            we have no interest in comparing the creation dates. Of course
            those will be different.
        """
        clines = []
        elines = []
        with open(compare) as c:
            clines = c.readlines()
        with open(expected) as e:
            elines = e.readlines()

        clines = clines[2:]
        elines = elines[2:]

        return (clines == elines)

    def test_read(self):
        ts1 = Timestamp(2014, 353, 59918)
        ts2 = Timestamp(2014, 353, 60518)
        ts3 = Timestamp(2014, 353, 60600)
        ts4 = Timestamp(2014, 353, 60650)

        ref_positions = [
            EstimatedSsmPosition(ts1, 16572652, ModelQuality(0)),
            EstimatedSsmPosition(ts2, 16573028, ModelQuality(2)),
            EstimatedSsmPosition(ts3, 16573117, ModelQuality(1)),
            EstimatedSsmPosition(ts4, 16573128, ModelQuality(2)),
        ]

        # Test reading a valid format1 file
        x = EstimatedSsmPositionFile(self.format1)
        obtained_positions = x.read()

        self.assertTrue(x.creation_date == Timestamp(2016, 21, 57000))
        self.assertTrue(x.start_date == Timestamp(2014, 353, 59918))
        self.assertTrue(x.end_date == Timestamp(2014, 353, 60600))
        self.assertEqual(len(ref_positions), len(obtained_positions))
        for i in range(len(ref_positions)):
            self.assertTrue(ref_positions[i] == obtained_positions[i])

        # We should grab the 2nd and 3rd positions in the file
        # Start time inclusive, end time exclusive.
        obtained_positions = x.read(
            start_time=Timestamp(2014, 353, 60518),
            end_time=Timestamp(2014, 353, 60650))

        self.assertEqual(len(obtained_positions), 2)
        self.assertTrue(obtained_positions[0] == ref_positions[1])
        self.assertTrue(obtained_positions[1] == ref_positions[2])

        # Test reading a valid format0 file
        x = EstimatedSsmPositionFile(self.format0)
        obtained_positions = x.read()

        # Fix the expected results to have final quality
        ref_positions[0].quality_flag = ModelQuality(2)
        ref_positions[2].quality_flag = ModelQuality(2)

        # All values should be equal
        self.assertEqual(len(ref_positions), len(obtained_positions))
        for i in range(len(ref_positions)):
            self.assertTrue(ref_positions[i] == obtained_positions[i])

        # We should grab the 2nd and 3rd positions in the file
        # Start time inclusive, end time exclusive.
        obtained_positions = x.read(
            start_time=Timestamp(2014, 353, 60518),
            end_time=Timestamp(2014, 353, 60650))

        self.assertEqual(len(obtained_positions), 2)
        self.assertTrue(obtained_positions[0] == ref_positions[1])
        self.assertTrue(obtained_positions[1] == ref_positions[2])

        # Begin test read of invalid files
        # Create list of invalid filenames
        invalid_format0_files = [(
            self.test_data_dir + "invalid_format0_{0}.data").format(x) for x
            in range(1, 4)]
        invalid_format1_files = [(
            self.test_data_dir + "invalid_format1_{0}.data").format(x) for x
            in range(1, 11)]

        # Lists of expected exceptions to be caught from reading each invalid
        # file
        expected_format0_exceptions = [
            IOError, IOError, IOError
        ]
        expected_format1_exceptions = [
            IOError, TimestampError, TimestampError, TimestampError,
            ValueError, IOError, ValueError, IOError, IOError, IOError,
        ]

        for invalid_file, exception in zip(
                invalid_format0_files, expected_format0_exceptions):
            x = EstimatedSsmPositionFile(invalid_file)
            with self.assertRaises(exception):
                x.read()

        for invalid_file, exception in zip(
                invalid_format1_files, expected_format1_exceptions):
            x = EstimatedSsmPositionFile(invalid_file)
            with self.assertRaises(exception):
                x.read()


class TestEstimatedSsmPositionTable(unittest.TestCase):
    db_conn = None
    test_tables = None

    switch_events_table = None
    pos_model_table = None
    est_ssm_pos_table = None

    switch_events = None

    @classmethod
    def setUpClass(cls):
        cls.db_conn = connect_to_database()
        cls.test_tables = DatabaseTestTables(
            cls.db_conn, switch=True, model=True, position=True)
        cls.test_tables.cleanup_database(True)
        cls.test_tables.setup_database()

        # The estimated ssm pos table is dependent on data in the switch events
        # table and the ssm pos model table. Insert dummy data into each using
        # the switch events manager
        se_manager = ManageSwitchEvents(cls.db_conn)
        cls.switch_events = [
            se_manager.add_event(Timestamp(2014, 353, 0), True, True),
            se_manager.add_event(Timestamp(2014, 354, 1), True, True),
            se_manager.add_event(Timestamp(2014, 355, 2), True, True),
            se_manager.add_event(Timestamp(2014, 356, 3), True, True),
            se_manager.add_event(Timestamp(2014, 357, 4), True, True),
        ]

        cls.switch_events_table = SwitchEventsTable(cls.db_conn)
        cls.pos_model_table = SsmPosModelTable(cls.db_conn)
        cls.est_ssm_pos_table = EstimatedSsmPositionTable(cls.db_conn)

    @classmethod
    def tearDownClass(cls):
        cls.test_tables.cleanup_database(False)
        cls.db_conn.close(db.DISCONNECT_ROLLBACK)

    def test_constructor(self):
        self.est_ssm_pos_table = EstimatedSsmPositionTable(self.db_conn)
        self.assertEqual(
            self.est_ssm_pos_table.table_name,
            DatabaseObjects.get_name("position_estimates_table"))
        self.assertEqual(self.est_ssm_pos_table._table_columns, [
            "MODEL_ID", "YEAR", "DAY_OF_YEAR", "SECONDS_OF_DAY",
            "ENCODER_POSITION", "QUALITY_FLAG", "DATE_ADDED",
            "DATE_DISABLED"])

    def test_a_insert_and_select(self):
        all_switch_events = self.switch_events_table.get_all_events()
        inserted_positions1 = [
            EstimatedSsmPosition(
                Timestamp(2014, 353, 59001),
                12345678, ModelQuality(2)),
            EstimatedSsmPosition(
                Timestamp(2014, 353, 65001),
                12345679, ModelQuality(2)),
            EstimatedSsmPosition(
                Timestamp(2014, 353, 75002),
                12345689, ModelQuality(2))
        ]

        inserted_positions2 = [
            EstimatedSsmPosition(
                Timestamp(2014, 353, 81003),
                12345789, ModelQuality(2))]

        model_id1 = self.pos_model_table.get_active_model(
            all_switch_events[0].event_id).model_id
        model_id2 = self.pos_model_table.get_active_model(
            all_switch_events[1].event_id).model_id

        self.est_ssm_pos_table.insert_ssm_positions(
            model_id1, inserted_positions1)
        self.est_ssm_pos_table.insert_ssm_positions(
            model_id2, inserted_positions2)

        selected_positions1 = (
            self.est_ssm_pos_table.get_ssm_positions(model_id1))
        self.assertEqual(len(inserted_positions1), len(selected_positions1))
        for ins_pos, sel_pos in zip(inserted_positions1, selected_positions1):
            self.assertTrue(ins_pos == sel_pos)

        selected_positions2 = (
            self.est_ssm_pos_table.get_ssm_positions(model_id2))
        self.assertEqual(len(inserted_positions2), len(selected_positions2))
        for ins_pos, sel_pos in zip(inserted_positions2, selected_positions2):
            self.assertTrue(ins_pos == sel_pos)

        # Verify both sets of positions above are selected by get_all
        all_positions = (
            self.est_ssm_pos_table.get_all_ssm_positions(True))
        for ins_pos, sel_pos in zip(inserted_positions1 + inserted_positions2,
                                    all_positions):
            # [1] because get_all returns a tuple
            self.assertTrue(ins_pos == sel_pos[1])

    def test_b_disable_ssm_positions(self):
        Timestamp(2014, 355, 86200)
        Timestamp(2014, 355, 6001)

        all_switch_events = self.switch_events_table.get_all_events()
        inserted_positions1 = [
            EstimatedSsmPosition(
                Timestamp(2014, 356, 59001),
                12345678, ModelQuality(2)),
            EstimatedSsmPosition(
                Timestamp(2014, 356, 65001),
                12345679, ModelQuality(2)),
        ]
        inserted_positions2 = [
            EstimatedSsmPosition(
                Timestamp(2014, 357, 20000),
                12345679, ModelQuality(2)),
        ]
        model_id1 = self.pos_model_table.get_active_model(
            all_switch_events[3].event_id).model_id
        model_id2 = self.pos_model_table.get_active_model(
            all_switch_events[4].event_id).model_id

        self.est_ssm_pos_table.insert_ssm_positions(
            model_id1, inserted_positions1)
        self.est_ssm_pos_table.insert_ssm_positions(
            model_id2, inserted_positions2)

        self.est_ssm_pos_table.disable_ssm_positions([model_id1])
        selected_positions1 = (
            self.est_ssm_pos_table.get_ssm_positions(model_id1))
        selected_positions2 = (
            self.est_ssm_pos_table.get_ssm_positions(model_id2))
        for pos1 in selected_positions1:
            self.assertTrue(pos1.date_disabled is not None)
        for pos2 in selected_positions2:
            self.assertTrue(pos2.date_disabled is None)


if __name__ == '__main__':
    sat_attr.set_default(sat_attr.get_satellite_id_from_number(8))
    print("-" * 80)
    print("Testing Estimated Position classes")
    unittest.main(verbosity=2)
