#!/usr/bin/env python3

from collections import namedtuple
from database_objects import DatabaseObjects
from database_structure import DatabaseStructure
from database_test_tables import DatabaseTestTables
from datetime import timedelta
from estimated_ssm_position import EstimatedSsmPositionTable
from fit_tirs_ssm_switch_event import (
    FitSwitchEvent, FitSwitchEventError, parse_parameters)
from ias.db import database as db
from ias.log import logger
from ias.misc import satellite_attributes as sat_attr
from io import StringIO
from manage_tirs_ssm_switch_events import (ManageSwitchEvents)
from ssm_exceptions import (
    CPFRetrievalError, EventNotFoundError, MissingDataError, InvalidDataError,
    ParameterError, SsmException)
from ssm_pos_model import SsmPosModel, SsmPosModelTable, ModelQuality
from ssm_utilities import connect_to_database
from subprocess import call, check_call, DEVNULL, CalledProcessError
from switch_event import SwitchEvent, SwitchEventsTable
from timestamp import Timestamp
from tirs_ssm_model_fit_parameters import ModelFitParameters
import argparse
import os
import shutil
import subprocess
import sys
import unittest


class TestFitSwitchEvents(unittest.TestCase):
    db_connection = None
    test_tables = None
    curr_dir = os.getcwd()

    @classmethod
    def setUpClass(cls):
        cls.db_connection = connect_to_database()
        cls.test_tables = DatabaseTestTables(
            cls.db_connection, switch=True, model=True, position=True)

        # Make sure test tables from a previous run have been dropped.  Ignore
        # errors that occur from the table not being there
        cls.test_tables.cleanup_database(True)

        # Create the needed test tables
        cls.test_tables.setup_database()

        # Make sure the current directory is in the user's path
        cls.orig_path = os.environ['PATH']
        os.environ['PATH'] = os.getcwd() + ":" + cls.orig_path

        # create symbolic link pointing at this unit test cases generate_ssm
        if os.path.exists(cls.curr_dir + "/generate_ssm_encoder_file"):
            os.unlink(cls.curr_dir + "/generate_ssm_encoder_file")
        src = cls.curr_dir + \
            "/fit_ssm_model_switch_event_test_data/generate_ssm_encoder_file"
        dest = cls.curr_dir + "/generate_ssm_encoder_file"
        os.symlink(src, dest)

    @classmethod
    def tearDownClass(cls):
        # Remove test tables here
        cls.test_tables.cleanup_database(False)
        cls.db_connection.close(db.DISCONNECT_ROLLBACK)

        # Reset path to the original path
        os.environ['PATH'] = cls.orig_path

        # Remove generate_ssm_encoder_link if it exists
        if os.path.exists(cls.curr_dir + "/generate_ssm_encoder_file"):
            os.unlink(cls.curr_dir + "/generate_ssm_encoder_file")

    def get_events(self):
        """ Event with which to populate the switch event table
        """
        # Note the telemetry encoder csv file used for the tests really only
        # has data from 2015:024 -> 2016:027. However, the C executable
        # does not check the dates, and the data works as expected for the
        # unit tests, so it is considered sufficient for these unit tests.
        # We do not do any "indepth" testing of the C executable
        # results for the most part, just make sure it runs.  We do check
        # the format of the report file and some values in one test, but
        # expect the C executable tests to cover the rest of it.
        events = [
                    SwitchEvent.from_list([1, 2015, 302, 61500, True, True]),
                    SwitchEvent.from_list([2, 2015, 317, 47390, True, False]),
                    SwitchEvent.from_list([3, 2015, 330, 39840, True, True]),
                    SwitchEvent.from_list([4, 2015, 351, 76320, True, True]),
                    SwitchEvent.from_list([5, 2015, 359, 73598, False, True]),
                    SwitchEvent.from_list([6, 2016, 10, 30176, True, True]),
                    SwitchEvent.from_list([7, 2016, 24, 62376, True, True])
                 ]
        return events

    def create_events(self):
        """ Populates events and model table so the unit tests have data to
            test with
        """
        log = logger.Logger.get_default_logger()
        valid_events = self.get_events()

        # Create the test object
        manage = ManageSwitchEvents(self.db_connection)

        # Create the database objects needed
        events_table = SwitchEventsTable(self.db_connection)
        model_table = SsmPosModelTable(self.db_connection)

        # Add events so we have something to work with
        for index in range(len(valid_events)):
            event = valid_events[index]
            added_event = manage.add_event(
                event.time, event.can_model, event.active)

    def is_empty(self, container):
        """ Determines if any python structure is populated or empty
        """
        if container:
            return False
        else:
            return True

    def get_positions_count(self, model_id):
        """ Retrieves/returns the number of positions in the positions table
        """
        posTable = EstimatedSsmPositionTable(self.db_connection)
        pos_from_db = posTable.get_ssm_positions(model_id)
        pos_count = len(pos_from_db)
        return pos_count

    def redirect_c_stdout(self):
        """ Redirects stdout for the C application which is more difficult
            to do than redirecting the python stdout
        """
        # Flush the c level buffer
        sys.stdout.flush()
        # Duplicate stdout (file descriptor 1)
        newstdout = os.dup(1)
        # Set devnull
        devnull = os.open('/dev/null', os.O_WRONLY)
        # Duplicate the dev null file descriptor
        os.dup2(devnull, 1)
        # Close devnull after duplication
        os.close(devnull)
        # Original stdout so python stdout will still work
        sys.stdout = os.fdopen(newstdout, 'w')

    def remove_directory(self, directory_to_remove):
        """ Removes directory named for event id provided, if it exists
        """
        # Use rmtree so directories that are not empty are still removed
        if os.path.exists(directory_to_remove):
            shutil.rmtree(directory_to_remove)

    def test(self):
        """ Setup
            First test run to call the create events function to populate data
        """
        # Clean up possible directories
        self.remove_directory("001_2015_302_61500")
        self.remove_directory("007_2016_24_62376")

        self.create_events()

        # Redirect output for the C application (can't just redirect stdout
        # for python, that does not work)
        self.redirect_c_stdout()

    def test_create_estimated_positions_1(self):
        """ Test create estimated positions for a complete switch event
            Independent of other tests although this is also kind of
            getting tested in the accept routines since the
            create_estimated_positions function as called there as well
        """
        event_id = 1
        fit = FitSwitchEvent(self.db_connection)
        fit.prepare_for_fit(event_id, True)

        # Verify the estimated positions file does not exist
        self.assertFalse(os.path.exists(fit.get_event_filename(
            fit.estimated_ssm_positions_filename)))

        # Make sure the positions are not populated yet
        self.assertIsNone(fit.estimated_ssm_positions_table)

        fit.new_model = fit.original_model
        fit.create_estimated_positions()

        # Positions are now populated
        self.assertNotEqual(fit.estimated_ssm_positions_table, None)

        # Verify the estimated positions file now exists
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.estimated_ssm_positions_filename)))

        # Verify the postions start time is the start of the switch event
        self.assertEqual(fit.start_time.get_datetime(),
                         fit.estimated_ssm_positions_table.start_time)

        # Verify the postions stop time is the stop time of the switch event
        self.assertEqual(fit.stop_time.get_datetime(),
                         fit.estimated_ssm_positions_table.stop_time)

        self.remove_directory(fit.event_directory)

    def test_create_estimated_positions_2(self):
        """ Test create estimated positions for an uncomplete switch event
            ie no ending event date
            Independent of other tests although this is also kind of
            getting tested in the accept routines since the
            create_estimated_positions function gets called there as well
        """
        event_id = 7
        fit = FitSwitchEvent(self.db_connection)
        fit.prepare_for_fit(event_id, True)

        # Verify the estimated positions file does not exist
        self.assertFalse(os.path.exists(fit.get_event_filename(
            fit.estimated_ssm_positions_filename)))

        # Make sure the positions are not populated yet
        self.assertIsNone(fit.estimated_ssm_positions_table)

        fit.new_model = fit.original_model
        fit.create_estimated_positions()

        # Positions are now populated
        self.assertNotEqual(fit.estimated_ssm_positions_table, None)

        # Verify the estimated positions file now exists
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.estimated_ssm_positions_filename)))

        # Verify the postions start time is the start of the switch event 7
        self.assertEqual(fit.start_time.get_datetime(),
                         fit.estimated_ssm_positions_table.start_time)

        # Verify stop time is current stop time + the fudge factor
        stop_timestamp = fit.start_time.get_datetime()
        stop_time = stop_timestamp + timedelta(
            days=ModelFitParameters.current_switch_event_days)
        self.assertEqual(stop_time,
                         fit.estimated_ssm_positions_table.stop_time)

        self.remove_directory(fit.event_directory)

    def test_fit_preparation(self):
        """ Test fit preparation
            Tests the function that sets up the variables used during the fit:
            switch event id and all associated information
            report directory with name based on switch event ID provided to
                 script
            ODL file name to store in report directory
                (tirs_ssm_model_fit.odl)
            Results file name to store in report directory
                (tirs_ssm_model_fit.rpt)
            Current model corresponding to switch event ID provided to script
        """
        event_id = 1
        fit = FitSwitchEvent(self.db_connection)
        # Set up switch event information and file names, etc
        fit.prepare_for_fit(event_id, True)

        # Event directory
        start_timestamp = Timestamp(2015, 302, 61500)
        event_directory = "{:03d}".format(event_id)
        event_directory += "_" + str(start_timestamp).replace(':', '_')
        event_directory += "/"

        # Verify directory name
        self.assertEqual(event_directory, fit.event_directory)
        # Verify directory was created
        # Verify the estimated positions file now exists
        self.assertTrue(os.path.exists(fit.event_directory))
        # Verify start time
        self.assertEqual(start_timestamp, fit.start_time)
        # Verify stop time
        stop_timestamp = Timestamp(2015, 330, 39840)
        self.assertEqual(stop_timestamp, fit.stop_time)
        # Verify complete event
        self.assertTrue(fit.complete_event)

        self.remove_directory(fit.event_directory)

    def test_get_cpf(self):
        """ Test CPF retrieval
            Tests the CPF is retrieved as expected
        """
        log = logger.Logger.get_default_logger()
        event_id = 1

        fit = FitSwitchEvent(self.db_connection)
        fit.prepare_for_fit(event_id, True, None, None)

        fit.get_cpf()
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.cpf_name)))

        # Call with a date where no CPF will be found
        fit.start_time = Timestamp(2050, 13, 61580)
        with self.assertLogs(logger=log, level='ERROR') as catch_log:
            with self.assertRaises(MissingDataError):
                fit.get_cpf()
        self.remove_directory(fit.event_directory)

    def test_masks(self):
        """ Test providing coefficient masks"""
        event_id = 1
        # Fit Class instance
        fit = FitSwitchEvent(self.db_connection)
        coefficient_masks = "(0, 1, 0, 0, 1, 0, 1, 0)"
        # Set up switch event information and file names, etc
        fit.prepare_for_fit(event_id, True, coefficient_masks, None)
        fit.create_odl()
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.odl_filename)))

        # Read the ODL file and verify that the masks in the ODL file
        # matches our override
        initial_values = "COEFFICIENT_MASKS"
        with open(fit.get_event_filename(fit.odl_filename), "r") as fodl:
            for line in fodl.readlines():
                line = line.strip()
                if line:
                    if initial_values in line:
                        masks = str(line).split("=")[-1]
                        masks = masks.strip()
                        self.assertEqual(coefficient_masks, masks)

        # Now remove a parameter from the coefficient_masks and verify
        # that an exception is raised
        masks = "(0, 1, 0, 0, 0, 0)"
        with self.assertRaises(ParameterError):
            fit.prepare_for_fit(event_id, True, masks, None)

        self.remove_directory(fit.event_directory)

    def test_model_override(self):
        """ Test providing model override"""
        log = logger.Logger.get_default_logger()
        event_id = 1

        # Fit Class instance
        fit = FitSwitchEvent(self.db_connection)
        model_override = ("(1787.081, 293.100, 82.748, 100.100, 1272.700, "
                          "0.515, 1.949, 3.508)")
        # Set up switch event information and file names, etc
        fit.prepare_for_fit(event_id, True, None, model_override)
        fit.create_odl()
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.odl_filename)))

        # Read the ODL file and verify that the model in the ODL file
        # matches our override
        initial_values = "INITIAL_VALUES"
        with open(fit.get_event_filename(fit.odl_filename), "r") as fodl:
            for line in fodl.readlines():
                line = line.strip()
                if line:
                    if initial_values in line:
                        model = str(line).split("=")[-1]
                        model = model.strip()
                        self.assertEqual(model_override, model)
        self.remove_directory(fit.event_directory)

        # Now remove a parameter from the model_override and verify
        # that an exception is raised
        model_override = ("(1787.081, 293.100, 100.100, 1272.700, "
                          "0.515, 1.949, 3.508)")
        with self.assertRaises(ParameterError):
            fit.prepare_for_fit(event_id, True, None, model_override)

    def test_odl_creation(self):
        """ Test create_odl/write_odl
            Tests the odl file creation functionality
        """
        log = logger.Logger.get_default_logger()
        event_id = 1

        # Fit Class instance
        fit = FitSwitchEvent(self.db_connection)
        # Set up switch event information and file names, etc
        fit.prepare_for_fit(event_id, True, None, None)

        self.assertFalse(os.path.exists(fit.get_event_filename(
            fit.odl_filename)))
        fit.create_odl()
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.odl_filename)))

        # Read in the lines from the expected file split on a blank space
        # and store in a tuple
        expected_results = "fit_ssm_model_switch_event_test_data/expected.odl"
        with open(expected_results, 'r') as expected_odl:
            expected_lines = set(
                line.strip().split()[0] for line in expected_odl)
            expected_count = len(expected_lines)

        # Verify that the odl file just created matches every line
        # that is in the expected file.  Note this verifies that each
        # parameter exists in each file but does not compare the values
        # since the paths to the CPF and results file will be different
        match = 0
        with open(fit.get_event_filename(fit.odl_filename), "r") as fit_odl:
            for line in filter(lambda x: x.strip().split()[0]
                               in expected_lines, fit_odl):
                match += 1
        self.assertEqual(expected_count, match)

        self.remove_directory(fit.event_directory)

    def test_parameter_parsing(self):
        """ Test parameter parsing
        """
        # Display tests only take fit event id as an argument
        args = parse_parameters(['display', '1'])
        self.assertEqual(args.command, 'display')
        self.assertEqual(args.event_id[0], 1)

        # Fit tests
        # Switch Event ID
        args = parse_parameters(['fit', '1'])
        self.assertEqual(args.command, 'fit')
        self.assertEqual(args.event_id[0], 1)

        # Switch Event ID with masks
        args = parse_parameters(['fit', '1', '--masks',
                                 '(1, 1, 1, 0, 0, 0, 1, 1)'])
        self.assertEqual(args.command, 'fit')
        self.assertEqual(args.event_id[0], 1)
        self.assertEqual(args.masks, '(1, 1, 1, 0, 0, 0, 1, 1)')

        # Switch Event ID with model override
        args = parse_parameters(['fit', '1', '--model_override',
                                 '(1787.081, 293.100, 82.748, 100.100, '
                                 '1272.700, 0.515, 1.949, 3.508)'])
        self.assertEqual(args.command, 'fit')
        self.assertEqual(args.event_id[0], 1)
        self.assertEqual(args.model_override, '(1787.081, 293.100, 82.748, '
                                              '100.100, 1272.700, 0.515, '
                                              '1.949, 3.508)')
        # Switch Event ID with stop time override
        args = parse_parameters(['fit', '1', '--stop_time', '1999:332:76543'])
        self.assertEqual(args.command, 'fit')
        self.assertEqual(args.stop_time, Timestamp(1999, 332, 76543))

        # fit_only tests
        # Switch Event ID
        args = parse_parameters(['fit_only', '1'])
        self.assertEqual(args.command, 'fit_only')
        self.assertEqual(args.event_id[0], 1)

        # Switch Event ID with masks
        args = parse_parameters(['fit_only', '1', '--masks',
                                 '(1, 1, 1, 0, 0, 0, 1, 1)'])
        self.assertEqual(args.command, 'fit_only')
        self.assertEqual(args.event_id[0], 1)
        self.assertEqual(args.masks, '(1, 1, 1, 0, 0, 0, 1, 1)')

        # Switch Event ID with model override
        args = parse_parameters(['fit_only', '1', '--model_override',
                                 '(1787.081, 293.100, 82.748, 100.100, '
                                 '1272.700, 0.515, 1.949, 3.508)'])
        self.assertEqual(args.command, 'fit_only')
        self.assertEqual(args.event_id[0], 1)
        self.assertEqual(args.model_override, '(1787.081, 293.100, 82.748, '
                                              '100.100, 1272.700, 0.515, '
                                              '1.949, 3.508)')
        # Switch Event ID with stop time override
        args = parse_parameters(['fit_only', '1', '--stop_time',
                                 '1999:332:76543'])
        self.assertEqual(args.command, 'fit_only')
        self.assertEqual(args.stop_time, Timestamp(1999, 332, 76543))

        # Accept tests
        # Switch Event ID and preliminary quality
        args = parse_parameters(['accept', '1', '--quality', 'preliminary'])
        self.assertEqual(args.event_id[0], 1)
        self.assertTrue(args.command, 'display')
        self.assertEqual(args.quality, 'preliminary')

        # Switch Event ID and final quality
        args = parse_parameters(['accept', '1', '--quality', 'final'])
        self.assertEqual(args.event_id[0], 1)
        self.assertTrue(args.command, 'display')
        self.assertEqual(args.quality, 'final')

    def test_prepare_for_fit(self):
        """ Test prepare_for_fit
            Tests the routine that retrieves the switch event information from
            the DB
        """
        log = logger.Logger.get_default_logger()

        # Call the get_switch_event with a switch event id not in the table
        fit = FitSwitchEvent(self.db_connection)
        with self.assertLogs(logger=log, level='ERROR') as catch_log:
            with self.assertRaises(MissingDataError):
                fit.prepare_for_fit(15, False)
            self.assertEqual(len(catch_log.output), 1)

        # Call the get_switch_event with a switch event id not active in
        # table
        with self.assertLogs(logger=log, level='ERROR') as catch_log:
            with self.assertRaises(InvalidDataError):
                fit.prepare_for_fit(2, False)
            self.assertEqual(len(catch_log.output), 1)

        # No exception raised when calling with a valid event
        fit.prepare_for_fit(1, True)

        # Superficial tests for start/stop time. These should be tested
        # indepth in the switch event unit tests
        # Valid switch event with an ending stop time
        expected_start_time = "2015:302:61500"
        fit.prepare_for_fit(1, True)
        self.assertEqual(fit.event_id, 1)
        self.assertEqual(fit.start_time,
                         Timestamp.from_string(expected_start_time))
        # Note since switch event 2 ACTIVE is false, the stop time should be
        # from the next switch event
        expected_stop_time = "2015:330:39840"
        self.assertEqual(fit.stop_time,
                         Timestamp.from_string(expected_stop_time))

        # Now test the stop time override. For the override to come into
        # play, the stop time needs to be set before the get_switch_event
        # method is called
        fit.stop_time = Timestamp.from_string("2016:001:39840")
        fit.prepare_for_fit(1, True)
        # Start time remains the same
        expected_start_time = "2015:302:61500"
        self.assertEqual(fit.start_time,
                         Timestamp.from_string(expected_start_time))
        # Stop time override holds and the stop time is not changed
        expected_stop_time = "2016:001:39840"
        self.assertEqual(fit.stop_time,
                         Timestamp.from_string(expected_stop_time))

        # Test providing a stop time that would be before the start time
        fit.stop_time = Timestamp.from_string("2015:275:61500")
        with self.assertLogs(logger=log, level='ERROR') as catch_log:
            with self.assertRaises(argparse.ArgumentTypeError):
                fit.prepare_for_fit(1, True)
            self.assertEqual(len(catch_log.output), 1)

        self.remove_directory(fit.event_directory)

    def test_run_ssm_model_fit(self):
        """ Test run_ssm_model_fit (calling C application)
            Tests that the C executable runs and creates the results file,
            as expected.
        """
        event_id = 1

        fit = FitSwitchEvent(self.db_connection)
        # Set up switch event information and file names, etc
        fit.prepare_for_fit(event_id, True)
        fit.create_odl()
        # Create encoder file after odl file, as the C routine takes
        # the ODL file as input
        fit.create_encoder_file()
        # Note no assert is needed here as the run_ssm_model_fit throws
        # an exception if it is unsuccessful.  If an exception is raised,
        # the unit test framework will automatically flag this as an error
        fit.run_ssm_model_fit()

        # Verify report file was created from running the C executable
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.fit_results_filename)))

    def test_run_ssm_model_retrieve_fit_results(self):
        """ Test the results retrival independent of the other tests
            Has to run after the test_run_ssm_model_fit test above
        """
        event_id = 1

        fit = FitSwitchEvent(self.db_connection)
        fit.prepare_for_fit(event_id, False)

        fit_results = None
        fit_results = fit.retrieve_results()
        self.assertTrue(fit_results is not None)

        self.assertFalse(self.is_empty(fit_results.new_model))
        self.assertFalse(self.is_empty(fit_results.orig_model))
        self.assertFalse(self.is_empty(fit_results.image_days_from_switch))
        self.assertFalse(self.is_empty(fit_results.image_measured_position))
        self.assertFalse(self.is_empty(fit_results.image_modeled_position))
        self.assertFalse(self.is_empty(fit_results.image_encoder_diff))
        self.assertFalse(self.is_empty(fit_results.telemetry_days_from_switch))
        self.assertFalse(self.is_empty(fit_results.
                         telemetry_measured_position))
        self.assertFalse(self.is_empty(fit_results.telemetry_modeled_position))
        self.assertFalse(self.is_empty(fit_results.telemetry_encoder_diff))

        # All Telemetry lists should be same size
        self.assertEqual(len(fit_results.telemetry_days_from_switch),
                         len(fit_results.telemetry_measured_position))

        self.assertEqual(len(fit_results.telemetry_days_from_switch),
                         len(fit_results.telemetry_modeled_position))

        self.assertEqual(len(fit_results.telemetry_days_from_switch),
                         len(fit_results.telemetry_encoder_diff))

        # All Image lists should be same size
        self.assertEqual(len(fit_results.image_days_from_switch),
                         len(fit_results.image_measured_position))

        self.assertEqual(len(fit_results.image_days_from_switch),
                         len(fit_results.image_modeled_position))

        self.assertEqual(len(fit_results.image_days_from_switch),
                         len(fit_results.image_encoder_diff))

        # Keep an eye on these values
        previous = namedtuple("previous", ["a0", "a1", "a2", "a3", "tau1",
                              "tau2", "tau3", "slope", "encoder_nadir_pos"])
        a0 = 1645.200
        a1 = 293.100
        a2 = 286.800
        a3 = 100.100
        tau1 = 1272.700
        tau2 = 0.197
        tau3 = 1.949
        slope = 3.508
        encoder_nadir_pos = 16574079

        previous = previous(a0, a1, a2, a3, tau1, tau2, tau3, slope,
                            encoder_nadir_pos)
        self.assertAlmostEqual(previous.a0, fit_results.orig_model.a0)
        self.assertAlmostEqual(previous.a1, fit_results.orig_model.a1)
        self.assertAlmostEqual(previous.a2, fit_results.orig_model.a2)
        self.assertAlmostEqual(previous.a3, fit_results.orig_model.a3)
        self.assertAlmostEqual(previous.tau1, fit_results.orig_model.tau1)
        self.assertAlmostEqual(previous.tau2, fit_results.orig_model.tau2)
        self.assertAlmostEqual(previous.tau3, fit_results.orig_model.tau3)
        self.assertAlmostEqual(previous.slope, fit_results.orig_model.slope)
        self.assertAlmostEqual(previous.encoder_nadir_pos,
                               fit_results.orig_model.encoder_nadir_pos)

        new = namedtuple("new", ["a0", "a1", "a2", "a3", "tau1",
                         "tau2", "tau3", "slope", "encoder_nadir_pos"])
        a0 = 1368.125
        a1 = 274.701
        a2 = 197.542
        a3 = 100.10
        tau1 = 1496.476
        tau2 = 0.167
        tau3 = 1.949
        slope = 3.508
        encoder_nadir_pos = 16574079
        new = new(a0, a1, a2, a3, tau1, tau2, tau3, slope,
                  encoder_nadir_pos)
        self.assertAlmostEqual(new.a0, fit_results.new_model.a0)
        self.assertAlmostEqual(new.a1, fit_results.new_model.a1)
        self.assertAlmostEqual(new.a2, fit_results.new_model.a2)
        self.assertAlmostEqual(new.a3, fit_results.new_model.a3)
        self.assertAlmostEqual(new.tau1, fit_results.new_model.tau1)
        self.assertAlmostEqual(new.tau2, fit_results.new_model.tau2)
        self.assertAlmostEqual(new.tau3, fit_results.new_model.tau3)
        self.assertAlmostEqual(new.slope, fit_results.new_model.slope)
        self.assertAlmostEqual(new.encoder_nadir_pos,
                               fit_results.new_model.encoder_nadir_pos)

    def test_run_ssm_model_retrieve_position_results(self):
        """ Test the position results retrival independent of the other tests
            Has to run after the test_run_ssm_model_fit test above
        """
        event_id = 1
        fit = FitSwitchEvent(self.db_connection)
        fit.prepare_for_fit(event_id, False)
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.fit_results_filename)))
        self.assertIsNone(fit.estimated_ssm_positions_table)
        fit_results = fit.retrieve_results()
        fit.new_model = fit_results.build_model(ModelQuality.predicted)
        fit.create_estimated_positions()
        self.assertIsNotNone(fit.estimated_ssm_positions_table)
        fit_results._retrieve_ssm_positions(fit.estimated_ssm_positions_table)

        # Superficial tests that everything is populated as expected
        # The unit testing for the generate_ssm_position_table should
        # perform more indepth tests on the actual values
        self.assertFalse(self.is_empty(
            fit_results.ssm_positions_est_days))
        self.assertFalse(self.is_empty(
            fit_results.ssm_positions_est_position))

        # Don't clean environment this time, leave information for switch
        # event ID 1 on disk for the display only tests

    # Forcing display only to be tested after the ssm model tests so that
    # all of the data needed already exists by adding 'subsequent' to the
    # test function name
    def test_subsequent_display_only(self):
        """ Test display only capability
            Verifies the display only capability by capturing the report from
            stdout and verifying some of the data.
        """
        event_id = 1
        no_display = True
        fit = FitSwitchEvent(self.db_connection)
        # Set up switch event information and file names, etc
        fit.prepare_for_fit(event_id, False)
        # Retrieve results
        fit_results = fit.retrieve_results()
        fit.new_model = fit_results.build_model(ModelQuality.predicted)
        fit.create_estimated_positions()
        self.assertIsNotNone(fit.estimated_ssm_positions_table)
        # Keep old stdout
        old_stdout = sys.stdout
        # Redirect stdout
        result = StringIO()
        sys.stdout = result
        # "Show" the report, will be stored in the 'results' variable
        fit.show_results(fit_results, no_display)
        # Restore standard out
        sys.stdout = old_stdout
        # Get the stuff stored in result variable
        result_string = result.getvalue()
        # Verify something was stored in the result string. Stricter tests
        # will be done on the result retrieval in a subsequent test
        self.assertNotEqual(len(result_string), 0)
        # Verify the plot file was created
        fit.show_results(fit_results, no_display)
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.plot_filename)))
        # Verify file size is > 0, which it was when DISPLAY was set and
        # an instance of the figure was not captured before the show
        self.assertTrue(os.path.getsize(fit.get_event_filename(
                        fit.plot_filename)) > 0)
        self.remove_directory(fit.event_directory)

    # Forcing the accept tests to run after most of the other unit tests by
    # adding the word 'the' into the function names since the tests are
    # run alphabetically.  Note that the accept routines should be last or
    # else it is possible that if there are changes to the default model,
    # accepting models will change the model parameters for the tests that
    # are explicitly comparing parameters to a specific value.
    def test_the_accept_model_final_1(self):
        """ Test accept script setting final quality
            Test accept with the fit script determining the quality (its not
            set explicitly).  Quality should be final since it is purposely
            using a switch end that has a follow up switch event.
        """
        # First generate data to accept as the accept method runs on a
        # previously generated model.  Since there are switch events after
        # event_id 1 the quality should be set as final when populating
        # the database
        event_id = 1
        fit = FitSwitchEvent(self.db_connection)
        fit.prepare_for_fit(event_id, True)
        # Create ODL file
        fit.create_odl()
        # Create encoder file after odl file, as the C routine takes
        # the ODL file as input
        fit.create_encoder_file()
        # Get the model for the estimated positions
        fit.get_model()
        original_model_id = fit.original_model.model_id
        # Run the C application
        fit.run_ssm_model_fit()
        # Retrieve the results
        fit_results = fit.retrieve_results()
        # Build
        fit.new_model = fit_results.build_model(fit.quality)
        fit.create_estimated_positions()

        # Have all of our data created, now accept the data
        fit.accept_model()

        # Verify accept
        # This gets the active model associated with this ID, so not only
        # does it show that the new model was successfully populated but the
        # old model was successfully disabled
        fit.get_model()
        new_model_id = fit.original_model.model_id
        self.assertNotEqual(original_model_id, new_model_id)
        self.assertEqual(fit.original_model.quality_flag,
                         fit.quality)

        # Validate files were created as expected
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.fit_results_filename)))
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.odl_filename)))
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.ssm_encoder_filename)))
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.estimated_ssm_positions_filename)))

        # Validate new positions were entered in DB for the new model ID
        new_model_pos_count = self.get_positions_count(new_model_id)
        # Validate positions count. Stricter tests will be done on the
        # positions in a subsequent test
        self.assertNotEqual(new_model_pos_count, 0)

        self.remove_directory(fit.event_directory)

    def test_the_accept_model_final_2(self):
        """ Test accept explicitly setting final quality
            Test accept with explicitly setting the quality to final -
            mimicking the behavior of when it is present as a command line
            parameter.
        """
        # First generate data to accept as the accept method runs on a
        # previously generated model.  Since there are switch events after
        # event_id 1 the quality should be set as final when populating
        # the database
        event_id = 1
        fit = FitSwitchEvent(self.db_connection)
        fit.prepare_for_fit(event_id, True)
        fit.quality = ModelQuality.final
        fit.create_odl()
        # Create encoder file after odl file, as the C routine takes
        # the ODL file as input
        fit.create_encoder_file()
        fit.get_model()
        original_model_id = fit.original_model.model_id
        fit.run_ssm_model_fit()
        fit_results = fit.retrieve_results()
        fit.new_model = fit_results.build_model(fit.quality)
        fit.create_estimated_positions()

        # Have all of our data created, now accept the data
        fit.accept_model()

        # Verify accept
        # This gets the active model associated with this ID, so not only
        # does it show that the new model was successfully populated but the
        # old model was successfully disabled
        fit.get_model()
        new_model_id = fit.original_model.model_id
        self.assertNotEqual(original_model_id, new_model_id)
        self.assertEqual(fit.original_model.quality_flag, ModelQuality.final)

        # Validate files were created as expected
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.fit_results_filename)))
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.odl_filename)))
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.ssm_encoder_filename)))
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.estimated_ssm_positions_filename)))

        # Validate new positions were entered in DB for the new model ID
        new_model_pos_count = self.get_positions_count(new_model_id)
        # Validate positions count. Stricter tests will be done on the
        # positions in a subsequent test
        self.assertNotEqual(new_model_pos_count, 0)

        self.remove_directory(fit.event_directory)

    def test_the_accept_model_preliminary_1(self):
        """ Test accept with script determining/setting quality to preliminary
            Test accept with the fit script determining the quality (its not
            set explicitly).  Quality should be preliminary since it is
            purposely using a switch end that does not have a follow up switch
            event.
        """
        # Set event id to the last one in the table forcing it to make the
        # quality preliminary because we have no following event to get the
        # stop time
        event_id = 7
        fit = FitSwitchEvent(self.db_connection)
        # Set up switch event information and file names, etc
        fit.prepare_for_fit(event_id, True)
        fit.create_odl()
        # Create encoder file after odl file, as the C routine takes
        # the ODL file as input
        fit.create_encoder_file()
        fit.get_model()
        original_model_id = fit.original_model.model_id

        fit.run_ssm_model_fit()
        fit_results = fit.retrieve_results()
        fit.new_model = fit_results.build_model(fit.quality)
        fit.create_estimated_positions()

        # Have all of our data created, now accept the data
        fit.accept_model()

        # Verify accept
        # This gets the active model associated with this ID, so not only
        # does it show that the new model was successfully populated but the
        # old model was successfully disabled
        fit.get_model()
        new_model_id = fit.original_model.model_id
        self.assertNotEqual(original_model_id, new_model_id)
        self.assertEqual(fit.original_model.quality_flag,
                         ModelQuality.preliminary)

        # Validate files were created as expected
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.fit_results_filename)))
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.odl_filename)))
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.ssm_encoder_filename)))
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.estimated_ssm_positions_filename)))

        # Validate new positions were entered in DB for the new model ID
        new_model_pos_count = self.get_positions_count(new_model_id)
        # Validate positions count. Stricter tests will be done on the
        # positions in a subsequent test
        self.assertNotEqual(new_model_pos_count, 0)

        self.remove_directory(fit.event_directory)

    def test_the_accept_model_preliminary_2(self):
        """ Test accept explicitly setting quality to preliminary
            Test accept with explicitly setting the quality to preliminary -
            mimicking the behavior of when it is present as a command line
            parameter.
        """
        # Set event id to the last one in the table forcing it to make the
        # quality preliminary because we have no following event to get the
        # stop time
        event_id = 7
        fit = FitSwitchEvent(self.db_connection)
        # Set up switch event information and file names, etc
        fit.prepare_for_fit(event_id, True)
        fit.quality = ModelQuality.preliminary
        fit.create_odl()
        # Create encoder file after odl file, as the C routine takes
        # the ODL file as input
        fit.create_encoder_file()
        fit.get_model()
        original_model_id = fit.original_model.model_id

        fit.run_ssm_model_fit()
        fit_results = fit.retrieve_results()
        fit.new_model = fit_results.build_model(fit.quality)
        fit.create_estimated_positions()

        # Have all of our data created, now accept the data
        fit.accept_model()

        # Verify accept
        # This gets the active model associated with this ID, so not only
        # does it show that the new model was successfully populated but the
        # old model was successfully disabled
        fit.get_model()
        new_model_id = fit.original_model.model_id
        self.assertNotEqual(original_model_id, new_model_id)
        self.assertEqual(fit.original_model.quality_flag,
                         ModelQuality.preliminary)

        # Validate files were created as expected
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.fit_results_filename)))
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.odl_filename)))
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.ssm_encoder_filename)))
        self.assertTrue(os.path.exists(fit.get_event_filename(
            fit.estimated_ssm_positions_filename)))

        # Validate new positions were entered in DB for the new model ID
        new_model_pos_count = self.get_positions_count(new_model_id)
        # Validate positions count. Stricter tests will be done on the
        # positions in a subsequent test
        self.assertNotEqual(new_model_pos_count, 0)

        self.remove_directory(fit.event_directory)

    def test_whole_application(self):
        """ Test running the fit application with every option
        """
        # Note that this routine is expected to run after all the other tests
        # to get the expected database contents.  The Python unittest class
        # runs the tests in alphabetical order, so if tests are added they
        # might impact the results here. (Hence the use of the 'whole' in the
        # method name.)

        # Set the application name to call
        app = "../fit_tirs_ssm_switch_event.py"

        # Set the output target for stdout from the application runs.  Set
        # to None if the output is wanted on the display for debugging.
        output_target = DEVNULL

        # Save the original PYTHONPATH and add the table_override directory
        # at the start so running the application uses the test tables
        orig_path = os.environ['PYTHONPATH']
        os.environ['PYTHONPATH'] = "./table_override:" + orig_path

        self.remove_directory("001_2015_302_61500")
        status = check_call([app, "fit", "1", "--no_display"],
                            stdout=output_target)
        self.assertEqual(status, 0)

        # Don't remove directory when performing display only
        status = check_call([app, "display", "1", "--no_display"],
                            stdout=output_target)
        self.assertEqual(status, 0)

        status = check_call([app, "fit_only", "1", "--no_display"],
                            stdout=output_target)
        self.assertEqual(status, 0)

        # Don't remove directories when accepting, expects them to exist
        status = check_call([app, "accept", "1", "--no_display"],
                            stdout=output_target)
        self.assertEqual(status, 0)

        status = check_call([app, "accept", "1",
                            "--quality", "preliminary", "--no_display"],
                            stdout=output_target)
        self.assertEqual(status, 0)

        status = check_call([app, "accept", "1",
                            "--quality", "final", "--no_display"],
                            stdout=output_target)
        self.assertEqual(status, 0)

        self.remove_directory("001_2015_302_61500")

        # Fit with stop time override
        status = check_call([app, "fit", "1", "--stop_time",
                            "2015:318:39840", "--no_display"],
                            stdout=output_target)
        self.assertEqual(status, 0)
        self.remove_directory("001_2015_302_61500")

        # Stop time that is just barely before the start time
        with self.assertRaises(CalledProcessError):
            status = check_call([app, "fit", "1", "--stop_time",
                                "2015:302:39840"], stdout=output_target,
                                stderr=output_target)

        # Test with an event-id that we don't have in the DB
        with self.assertRaises(CalledProcessError):
            status = check_call([app, "99"],
                                stdout=output_target, stderr=output_target)

        # Test with an invalid argument
        with self.assertRaises(CalledProcessError):
            status = check_call([app, "--invalid_argument", "ABC"],
                                stdout=output_target, stderr=output_target)

        # Restore the original python path
        os.environ['PYTHONPATH'] = orig_path

if __name__ == '__main__':

    print("-" * 70)
    print("Testing fit switch events...\n")

    # Set up the satellite attributes for Landsat 8
    sat_attr.set_default(sat_attr.get_satellite_id_from_number(8))

    # Buffers stdout and only displays on a unit test error
    # Also want my tests to fail on the first failure so the directory still
    # exists with the data that spawnewd the failure, (since the tests
    # mainly use the same event_id (directory) over and over)
    unittest.main(verbosity=2, buffer=True, failfast=True)
