#!/usr/bin/env python3

import unittest
import os
import re
import cx_Oracle
from codecs import decode
from copy import deepcopy
from tempfile import mkstemp
from subprocess import check_call, check_output, DEVNULL, CalledProcessError
from ias.misc import satellite_attributes as sat_attr
from switch_event import SwitchEvent, SwitchEventsTable
from ssm_utilities import connect_to_database
from database_test_tables import DatabaseTestTables
from ias.db import database as db
from ssm_pos_model import SsmPosModelTable, ModelQuality, SsmPosModel
from estimated_ssm_position import (EstimatedSsmPositionTable,
                                    EstimatedSsmPositionFile)
from generate_ssm_position_table import (SsmPosTableGenerator,
                                         SsmPosTableGenFrontEnd)
from ssm_exceptions import EventNotFoundError
from timestamp import Timestamp


class TestPositionEstTable(unittest.TestCase):
    db_connection = None
    test_tables = None
    test_model = None
    test_positions = None
    stop_time = None
    start_time = None

    @classmethod
    def setUpClass(cls):
        cls.db_connection = connect_to_database()
        cls.test_tables = DatabaseTestTables(
            cls.db_connection, switch=True, model=True, position=True)

        # Make sure test tables from a previous run have been dropped.  Ignore
        # errors that occur from the table not being there
        cls.test_tables.cleanup_database(True)

        # Create the needed test tables
        cls.test_tables.setup_database()

        cls.test_model = SsmPosModel(0, 0, 0, 16574079, 1644.01, 290.05,
                                     274.09, 99.12, 1260.08, 0.184, 1.397,
                                     3.6, ModelQuality.preliminary, True)
        cls.stop_time = Timestamp(2000, 14, 1)
        cls.start_time = Timestamp(2000, 1, 61)

        table_gen = SsmPosTableGenerator(cls.db_connection,  cls.test_model,
                                         cls.stop_time, cls.start_time)
        table_gen.generate_ssm_pos_table()
        cls.test_positions = table_gen.position_table

    @classmethod
    def tearDownClass(cls):
        # Remove test tables here
        cls.test_tables.cleanup_database(False)
        cls.db_connection.close(db.DISCONNECT_ROLLBACK)

    def test_a_gen_position(self):
        """ Test generated positions against verified estimates
        """

        # Note any change to the model parameters (set in constructor) or
        # position_sampling_interval_table (from ModelFitParameters) will
        # make these values incorrect
        self.assertEqual(self.test_positions[1].position, 16575752)
        self.assertEqual(self.test_positions[3].position, 16575778)
        self.assertEqual(self.test_positions[11].position, 16575879)
        self.assertEqual(self.test_positions[13].position, 16575909)
        self.assertEqual(self.test_positions[27].position, 16576093)
        self.assertEqual(self.test_positions[42].position, 16576265)
        self.assertEqual(self.test_positions[43].position, 16576276)
        self.assertEqual(self.test_positions[51].position, 16576336)
        self.assertEqual(self.test_positions[73].position, 16576403)
        self.assertEqual(self.test_positions[76].position, 16576407)
        self.assertEqual(self.test_positions[84].position, 16576422)

        # Note any change to the position_sampling_interval_table will
        # make these values incorrect
        self.assertEqual(self.test_positions[1].timestamp,
                         Timestamp(2000, 1, 181))
        self.assertEqual(self.test_positions[3].timestamp,
                         Timestamp(2000, 1, 301))
        self.assertEqual(self.test_positions[11].timestamp,
                         Timestamp(2000, 1, 901))
        self.assertEqual(self.test_positions[13].timestamp,
                         Timestamp(2000, 1, 1141))
        self.assertEqual(self.test_positions[14].timestamp,
                         Timestamp(2000, 1, 1261))
        self.assertEqual(self.test_positions[15].timestamp,
                         Timestamp(2000, 1, 1441))
        self.assertEqual(self.test_positions[16].timestamp,
                         Timestamp(2000, 1, 1621))
        self.assertEqual(self.test_positions[28].timestamp,
                         Timestamp(2000, 1, 6061))
        self.assertEqual(self.test_positions[30].timestamp,
                         Timestamp(2000, 1, 7261))
        self.assertEqual(self.test_positions[33].timestamp,
                         Timestamp(2000, 1, 9661))
        self.assertEqual(self.test_positions[52].timestamp,
                         Timestamp(2000, 2, 61))
        self.assertEqual(self.test_positions[53].timestamp,
                         Timestamp(2000, 2, 10861))
        self.assertEqual(self.test_positions[55].timestamp,
                         Timestamp(2000, 2, 32461))
        self.assertEqual(self.test_positions[56].timestamp,
                         Timestamp(2000, 2, 43261))
        self.assertEqual(self.test_positions[70].timestamp,
                         Timestamp(2000, 5, 43261))
        self.assertEqual(self.test_positions[74].timestamp,
                         Timestamp(2000, 6, 43261))
        self.assertEqual(self.test_positions[75].timestamp,
                         Timestamp(2000, 6, 64861))
        self.assertEqual(self.test_positions[76].timestamp,
                         Timestamp(2000, 7, 61))
        self.assertEqual(self.test_positions[79].timestamp,
                         Timestamp(2000, 7, 64861))
        self.assertEqual(self.test_positions[85].timestamp,
                         Timestamp(2000, 12, 61))
        self.assertEqual(self.test_positions[87].timestamp,
                         Timestamp(2000, 14, 1))

        # All quality values should be the same as the model
        for pos in self.test_positions:
            self.assertEqual(pos.quality_flag, self.test_model.quality_flag)

    def test_b_zero_tau(self):
        """ Test generated positions are correct when a tau input is 0
        """
        # Use the test model as the starting point
        zero_model = deepcopy(self.test_model)
        # Change tau2 to 0
        zero_model.tau2 = 0.0

        table_gen = SsmPosTableGenerator(self.db_connection,  zero_model,
                                         self.stop_time, self.start_time)
        table_gen.generate_ssm_pos_table()

        self.assertEqual(table_gen.position_table[0].position, 16576011)
        self.assertEqual(table_gen.position_table[1].position, 16576024)
        self.assertEqual(table_gen.position_table[3].position, 16576048)
        self.assertEqual(table_gen.position_table[4].position, 16576059)
        self.assertEqual(table_gen.position_table[9].position, 16576107)
        self.assertEqual(table_gen.position_table[11].position, 16576139)
        self.assertEqual(table_gen.position_table[13].position, 16576165)
        self.assertEqual(table_gen.position_table[14].position, 16576176)
        self.assertEqual(table_gen.position_table[16].position, 16576204)
        self.assertEqual(table_gen.position_table[17].position, 16576216)
        self.assertEqual(table_gen.position_table[18].position, 16576226)
        self.assertEqual(table_gen.position_table[25].position, 16576280)
        self.assertEqual(table_gen.position_table[28].position, 16576290)
        self.assertEqual(table_gen.position_table[52].position, 16576341)
        self.assertEqual(table_gen.position_table[53].position, 16576346)
        self.assertEqual(table_gen.position_table[54].position, 16576350)
        self.assertEqual(table_gen.position_table[55].position, 16576354)
        self.assertEqual(table_gen.position_table[57].position, 16576362)
        self.assertEqual(table_gen.position_table[58].position, 16576366)
        self.assertEqual(table_gen.position_table[59].position, 16576370)
        self.assertEqual(table_gen.position_table[77].position, 16576408)
        self.assertEqual(table_gen.position_table[78].position, 16576409)
        self.assertEqual(table_gen.position_table[79].position, 16576410)
        self.assertEqual(table_gen.position_table[82].position, 16576415)
        self.assertEqual(table_gen.position_table[83].position, 16576419)
        self.assertEqual(table_gen.position_table[85].position, 16576426)
        self.assertEqual(table_gen.position_table[86].position, 16576429)
        self.assertEqual(table_gen.position_table[87].position, 16576433)

    def test_c_empty_database(self):
        """ Test problems caused by empty DB
        """
        table_gen = SsmPosTableGenerator(self.db_connection,  self.test_model,
                                         self.stop_time, self.start_time)
        # Able to generate the results, even with an empty DB since everything
        # is provided as input to the class constructor
        table_gen.generate_ssm_pos_table()
        table_gen.generate_report()

        # Not able to reset the stop time, w/o any switch events
        with self.assertRaises(EventNotFoundError):
            table_gen.reset_stop_time()

        # Not able to add position records to the DB w/o any parent
        with self.assertRaises(cx_Oracle.IntegrityError):
            table_gen.store_results()

    def test_d_store_position(self):
        """ Test storing positions in the database
        """
        # Create a SwitchEvent and Model in the DB (parents for our estimated
        # positions)
        se = SwitchEvent(0, self.start_time, True, True)
        seTable = SwitchEventsTable(self.db_connection)
        seTable.insert_event(se)
        modTable = SsmPosModelTable(self.db_connection)
        modTable.insert_model_record(self.test_model)

        # Generate and store the estimated positions in the DB
        table_gen = SsmPosTableGenerator(self.db_connection,  self.test_model,
                                         self.stop_time, self.start_time)
        table_gen.generate_ssm_pos_table()
        table_gen.store_results()
        self.db_connection.commit()

        posTable = EstimatedSsmPositionTable(self.db_connection)
        pos_from_db = posTable.get_ssm_positions(0)
        for (pos, pos_db) in zip(self.test_positions, pos_from_db):
            self.assertEqual(pos, pos_db)

    def test_e_gen_report(self):
        """ Test generating the estimated positions in report form
        """
        table_gen = SsmPosTableGenerator(self.db_connection,  self.test_model,
                                         self.stop_time, self.start_time)
        table_gen.generate_ssm_pos_table()
        report = table_gen.generate_report().split("\n")
        i = 0
        for line in report:
            # A data record
            expr1 = re.match(r'(\d+:\d+:\d+)\s+(\d+.?\d*)\s+', line)
            # The header
            expr2 = re.match(r'Year:', line)
            if expr1:
                self.assertEqual(Timestamp.from_string(expr1.group(1)),
                                 self.test_positions[i].timestamp)
                self.assertAlmostEqual(float(expr1.group(2)),
                                       self.test_positions[i].position, 1)
                i += 1
            elif(expr2):
                # Do nothing for the header line
                None
            else:
                raise Exception("Unknown line: ", line)

    def test_f_save_pos_file(self):
        """ Test creating a file containing the generated positions
        """
        try:
            # Create temp file to save the positions to
            (temp_desc, temp_file_path) = mkstemp(prefix='gen_unit_test_',
                                                  dir='.')

            # Close the file descriptor, we have no use for it (the functions
            # take a filename).
            os.close(temp_desc)

            table_gen = SsmPosTableGenerator(self.db_connection,
                                             self.test_model,
                                             self.stop_time, self.start_time)
            table_gen.generate_ssm_pos_table()
            SsmPosTableGenFrontEnd.write_positions_file(
                                                   temp_file_path,
                                                   table_gen.position_table)
            # Verify the positions in the file
            est_file = EstimatedSsmPositionFile(temp_file_path)
            pos_from_file = est_file.read()
        finally:
            # Remove the temp file
            os.remove(temp_file_path)

        # Make sure each list has the same number of positions
        self.assertEqual(len(self.test_positions), len(pos_from_file))

        # Make sure the position values are the same before and after reading
        for (before, after) in zip(self.test_positions, pos_from_file):
            self.assertEqual(before, after)

    def test_g_disable_old_position(self):
        """ Test old positions are disabled when inserting new ones
        """
        modTable = SsmPosModelTable(self.db_connection)
        # Make the old model invalid
        modTable.disable_model_record(0)
        # Add the new model
        new_model = SsmPosModel(1, 0, 0, 16574079, 1644.01, 290.05,
                                274.09, 99.12, 1260.08, 0.184, 1.397,
                                3.6, ModelQuality.predicted, True)
        modTable.insert_model_record(new_model)

        # Generate and store positions for the new model
        table_gen = SsmPosTableGenerator(self.db_connection, new_model,
                                         self.stop_time, self.start_time)
        table_gen.generate_ssm_pos_table()
        table_gen.store_results()
        self.db_connection.commit()

        posTable = EstimatedSsmPositionTable(self.db_connection)
        pos_from_db = posTable.get_ssm_positions(0)
        for p in pos_from_db:
            # Verify the positions from the first model (ID 0) are disabled
            self.assertIsNotNone(p.date_disabled)
            # Quality flag should (still) be the same as the test model
            self.assertEqual(p.quality_flag, self.test_model.quality_flag)
        pos_from_db = posTable.get_ssm_positions(1)
        for p in pos_from_db:
            # Verify the positions from the new model (ID 1) are enabled
            self.assertIsNone(p.date_disabled)
            # Quality flag should be the same as the new model
            self.assertEqual(p.quality_flag, new_model.quality_flag)

    def test_h_reset_stop_time(self):
        """ Test the reseting the stop time to 60 seconds before next event
        """
        test_time = Timestamp(2099, 1, 1)
        table_gen = SsmPosTableGenerator(self.db_connection,  self.test_model,
                                         test_time, self.start_time)
        table_gen.reset_stop_time()
        # Time is unchanged because this the latest switch event
        self.assertEqual(test_time.get_datetime(), table_gen.stop_time)

        # Create a 2nd SwitchEvent
        se = SwitchEvent(1, Timestamp(2001, 1, 20060), True, True)
        seTable = SwitchEventsTable(self.db_connection)
        seTable.insert_event(se)

        # Need to add a model to the SwitchEvent
        modTable = SsmPosModelTable(self.db_connection)
        new_model = SsmPosModel(2, 1, 0, 16574079, 1644.01, 290.05,
                                274.09, 99.12, 1260.08, 0.184, 1.397,
                                3.6, ModelQuality(0), True)
        modTable.insert_model_record(new_model)
        self.db_connection.commit()

        # Check the first SwitchEvent after adding a new one
        table_gen2 = SsmPosTableGenerator(self.db_connection,  self.test_model,
                                          test_time, self.start_time)
        # Time is unchanged before the reset
        self.assertEqual(test_time.get_datetime(), table_gen2.stop_time)
        table_gen2.reset_stop_time()
        # Time is expected to be reset to 60 seconds before the next switch
        # event
        expected_time = Timestamp(2001, 1, 20000)
        self.assertEqual(expected_time.get_datetime(), table_gen2.stop_time)

    def test_i_test_gen_intervals(self):
        """ Test the positions are being generated for the proper time range
        """
        # Create a 3rd SwitchEvent. This event occurs shortly after the
        # 2nd (so close that we do not use all of the sampling interval)
        se = SwitchEvent(2, Timestamp(2001, 2, 80060), True, True)
        seTable = SwitchEventsTable(self.db_connection)
        seTable.insert_event(se)

        # Need to add a model to the SwitchEvent
        modTable = SsmPosModelTable(self.db_connection)
        new_model = SsmPosModel(3, 2, 0, 16574079, 1644.01, 290.05,
                                274.09, 99.12, 1260.08, 0.184, 1.397,
                                3.6, ModelQuality(0), True)
        modTable.insert_model_record(new_model)
        self.db_connection.commit()

        # Get the model for the 2nd switch event
        modelTable = SsmPosModelTable(self.db_connection)
        model_2 = modelTable.get_active_model(event_id=1)

        table_gen1 = SsmPosTableGenerator(self.db_connection,  self.test_model)
        table_gen2 = SsmPosTableGenerator(self.db_connection,  model_2)

        table_gen1.generate_ssm_pos_table()
        table_gen2.generate_ssm_pos_table()

        # Verify the first times of both sets of generated positions
        self.assertEqual(Timestamp(2000, 1, 121),
                         table_gen1.position_table[0].timestamp)
        self.assertEqual(Timestamp(2001, 1, 20120),
                         table_gen2.position_table[0].timestamp)

        # Verify the final times of both sets of generated positions
        self.assertEqual(Timestamp(2001, 1, 20000),
                         table_gen1.position_table[-1].timestamp)
        self.assertEqual(Timestamp(2001, 2, 80000),
                         table_gen2.position_table[-1].timestamp)

    def test_j_test_cmd_args(self):
        """ Command line argument parsing test
        """
        front_end = SsmPosTableGenFrontEnd(self.db_connection)

        args1 = ['--start_time=2050:100:1', '--stop_time=2050:105:3000']
        parse = front_end.parse_parameters(args1)
        self.assertFalse(parse.accept)
        self.assertIsNone(parse.event_id)
        self.assertEqual(parse.start_time, Timestamp(2050, 100, 1))
        self.assertEqual(parse.stop_time, Timestamp(2050, 105, 3000))

        args2 = ['--event_id=123', '--event_id=456', '--accept']
        parse = front_end.parse_parameters(args2)
        self.assertTrue(parse.accept)
        self.assertEqual(parse.event_id, [123, 456])

    def test_k_run_application(self):
        """ Run the application
        """
        # Set the application name to call
        app = "../generate_ssm_position_table.py"

        # Set the output target for stdout from the application runs.  Set
        # to None if the output is wanted on the display for debugging.
        output_target = DEVNULL
        # output_target = None

        # Save the original PYTHONPATH and add the table_override directory
        # at the start so running the application uses the test tables
        orig_path = os.environ['PYTHONPATH']
        os.environ['PYTHONPATH'] = "./table_override:" + orig_path

        # Pass the event id for the third switch event
        com_output =\
            decode(check_output([app, '--event_id=2', '--accept'])).split('\n')
        pos_count = 0
        for line in com_output:
            expr = re.match(r'200[01]:', line)
            if expr:
                pos_count += 1
        self.assertEqual(pos_count, 96)

        # Pass the event id for the third switch event again (not a mistake)
        # We should get a warning message indicating the positions are the same
        com_output =\
            decode(check_output([app, '--event_id=2', '--accept'])).split('\n')
        pos_count = 0
        warn_count = 0
        for line in com_output:
            expr1 = re.match(r'200[01]:', line)
            expr2 = re.search(r'positions to be inserted were the same', line)
            if expr1:
                pos_count += 1
            if expr2:
                warn_count += 1
        self.assertEqual(pos_count, 96)
        self.assertEqual(warn_count, 1)

        # Pass the dates that correspond to the first switch
        com_output = decode(check_output([app, '--start_time=2000:1:61',
                            '--stop_time=2001:1:20000'])).split('\n')
        pos_count = 0
        for line in com_output:
            expr = re.match(r'200[01]:', line)
            if expr:
                pos_count += 1
        self.assertEqual(pos_count, 442)

        # Pass start date to the first switch and stop time midway between
        # the first and second. This stop time will reset to 60 seconds
        # before the second (same time as the above check)
        com_output = decode(check_output([app, '--start_time=2000:1:61',
                            '--stop_time=2000:175:20000'])).split('\n')
        pos_count = 0
        for line in com_output:
            expr = re.match(r'200[01]:', line)
            if expr:
                pos_count += 1
        self.assertEqual(pos_count, 442)

        # Pass the start dates that correspond to the third switch and a
        # stop time 3 years away
        com_output = decode(check_output([app, '--start_time=2001:2:80060',
                            '--stop_time=2004:1:20059', '--accept'])
                            ).split('\n')
        pos_count = 0
        for line in com_output:
            expr = re.match(r'200[1234]:', line)
            if expr:
                pos_count += 1
        self.assertEqual(pos_count, 1169)

        # Verify the records were added to the DB, by checking the last
        # records timestamp
        modelTable = SsmPosModelTable(self.db_connection)
        models = modelTable.get_model_records(event_id=2)
        posTable = EstimatedSsmPositionTable(self.db_connection)
        pos_from_db = posTable.get_ssm_positions(models[0].model_id)
        self.assertEqual(pos_from_db[-1].timestamp, Timestamp(2004, 1, 20059))

        # Dates out of the range of our switch events, this will generate
        # a warning that no switch events were found
        com_output = decode(check_output([app, '--start_time=2010:1:1',
                            '--stop_time=2010:2:1'])).split('\n')
        pos_count = 0
        for line in com_output:
            expr = re.search(r'No matching switch events found', line)
            if expr:
                pos_count += 1
        self.assertEqual(pos_count, 1)

        # Use the save_positions arg with a time range, a single id, and
        # multiple ids
        for args in [['--start_time=2000:1:1', '--stop_time=2003:1:1'],
                     ['--event_id=0'], ['--event_id=0', '--event_id=1']]:
            (temp_desc, temp_file_path) = mkstemp(prefix='gen_unit_test_',
                                                  dir='.')
            os.close(temp_desc)
            try:
                status = check_call([app] + args + ['--save_positions={0}'.
                                    format(temp_file_path)],
                                    stdout=output_target, stderr=output_target)
            finally:
                os.remove(temp_file_path)
            self.assertEqual(status, 0)

        # Test with start date coming after the stop
        with self.assertRaises(CalledProcessError):
            status = check_call([app, '--start_time=2000:2:1',
                                 '--stop_time=2000:1:1'],
                                stdout=output_target, stderr=output_target)

        # Test with an invalid start date (not 100K seconds in day)
        with self.assertRaises(CalledProcessError):
            status = check_call([app, '--start_time=2000:2:100000',
                                 '--stop_time=2001:1:1'],
                                stdout=output_target, stderr=output_target)

        # Test with an event_id that we don't have in the DB
        with self.assertRaises(CalledProcessError):
            status = check_call([app, '--event_id=99'],
                                stdout=output_target, stderr=output_target)

        # Test passing both times and event_id
        with self.assertRaises(CalledProcessError):
            status = check_call([app, '--start_time=2000:2:100000',
                                 '--stop_time=2001:1:1', '--event_id=99'],
                                stdout=output_target, stderr=output_target)

        # Test passing multiple event_ids and a stop_time
        with self.assertRaises(CalledProcessError):
            status = check_call([app, '--stop_time=2001:1:1', '--event_id=99',
                                 '--event_id=100'],
                                stdout=output_target, stderr=output_target)

        # Test with a invalid argument
        with self.assertRaises(CalledProcessError):
            status = check_call([app, '--bad-arg=ABC'],
                                stdout=output_target, stderr=output_target)

        # Restore the original python path
        os.environ['PYTHONPATH'] = orig_path

if __name__ == '__main__':
    print("-" * 70)
    print("Testing position table..\n")

    # Set up the satellite attributes for Landsat 8
    sat_attr.set_default(sat_attr.get_satellite_id_from_number(8))

    unittest.main(verbosity=2)
