#!/usr/bin/env python3

import unittest
from subprocess import call, check_call
from ssm_utilities import connect_to_database
from ssm_exceptions import InvalidDataError
from ias.db import database as db
from ias.log import logger
from logging import CRITICAL, ERROR
from ias.misc import satellite_attributes as sat_attr
from timestamp import Timestamp
from ssm_pos_model import ModelQuality, SsmPosModelTable
from switch_event import SwitchEventsTable
from load_estimated_tirs_ssm_positions import LoadEstimatedTirsSsmPositions
from generate_ssm_position_table import SsmPosTableGenFrontEnd
from database_test_tables import DatabaseTestTables
from estimated_ssm_position import EstimatedSsmPosition
from estimated_ssm_position import EstimatedSsmPositionTable
from manage_tirs_ssm_switch_events import ManageSwitchEvents
from os import environ
from os import devnull
import sys


class TestLoadEstimatedPositions(unittest.TestCase):
    db_conn = None
    test_tables = None

    switch_events_table = None
    pos_models_table = None
    est_pos_table = None

    switch_events = []
    pos_models = []

    format0 = "testdata_load_estimated_positions/format0.data"
    format1 = "testdata_load_estimated_positions/format1.data"

    @classmethod
    def setUpClass(cls):
        # Set up the test tables and load dummy switch events and position
        # models into the tables.

        cls.db_conn = connect_to_database()
        cls.test_tables = DatabaseTestTables(
            cls.db_conn, switch=True, model=True, position=True)
        cls.test_tables.cleanup_database(True)
        cls.test_tables.setup_database()

        # Quiet the warnings these events cause since they are too close
        # together
        log = logger.Logger.get_default_logger()
        log.setLevel(ERROR)

        # Load test data
        se_manager = ManageSwitchEvents(cls.db_conn)
        cls.switch_events = [
            se_manager.add_event(Timestamp(2014, 353, 59000), False, True),
            se_manager.add_event(Timestamp(2014, 353, 65001), False, True),
            se_manager.add_event(Timestamp(2014, 353, 71002), False, True),
            se_manager.add_event(Timestamp(2014, 354, 0), False, True),
            se_manager.add_event(Timestamp(2014, 354, 6001), True, False),
            se_manager.add_event(Timestamp(2014, 354, 12002), False, False),
            se_manager.add_event(Timestamp(2014, 354, 18003), True, False),
            se_manager.add_event(Timestamp(2014, 354, 24004), False, True),
            se_manager.add_event(Timestamp(2015, 40, 0), False, True),
            se_manager.add_event(Timestamp(2015, 60, 0), False, True),
            se_manager.add_event(Timestamp(2015, 65, 0), False, True),
            se_manager.add_event(Timestamp(2015, 100, 0), True, True),  # 11
            se_manager.add_event(Timestamp(2015, 150, 0), True, True),  # 12
            se_manager.add_event(Timestamp(2015, 200, 0), True, True)  # 13
        ]

        cls.switch_events_table = SwitchEventsTable(cls.db_conn)
        cls.pos_models_table = SsmPosModelTable(cls.db_conn)
        cls.est_pos_table = EstimatedSsmPositionTable(cls.db_conn)

        cls.orig_python_path = environ['PYTHONPATH']
        cls.temp_python_path = "./table_override:" + cls.orig_python_path

    @classmethod
    def tearDownClass(cls):
        # Clean up the test tables and disconnect from the database

        environ['PYTHONPATH'] = cls.orig_python_path
        cls.test_tables.cleanup_database(False)
        cls.db_conn.close(db.DISCONNECT_ROLLBACK)

    def test_cmd_line_args(self):
        # This test is for the different configurations of command line
        # arguments which may be sent in to the application. We don't care
        # about final results of the different runs, only if those runs were
        # successful or not.

        prog = "../load_estimated_tirs_ssm_positions.py"
        temp_file = "temp_output_file.txt"

        event_id = self.switch_events[7].event_id
        inactive_id = self.switch_events[5].event_id

        environ['PYTHONPATH'] = self.temp_python_path

        # Test help
        output_file = open(temp_file, 'w')
        status = check_call([prog, '-h'], stdout=output_file)
        output_file.close()
        self.assertEqual(status, 0)

        # Test normally
        output_file = open(temp_file, 'w')
        status = check_call([prog, "load", str(event_id), self.format0],
                            stdout=output_file)
        output_file.close()
        self.assertEqual(status, 0)

        # Test with a quality flag parameter
        output_file = open(temp_file, 'w')
        status = check_call([prog, "load", str(event_id), self.format0,
                            '--quality_flag', '1'], stdout=output_file)
        output_file.close()
        self.assertEqual(status, 0)

        # Test the verify command
        output_file = open(temp_file, 'w')
        status = check_call([prog, "verify"], stdout=output_file)
        output_file.close()
        self.assertEqual(status, 0)

        # Test with an invalid switch event id (should fail)
        output_file = open(temp_file, 'w')
        status = call([prog, "load", 'words', self.format0],
                      stdout=output_file, stderr=output_file)
        output_file.close()
        self.assertEqual(status, 2)

        # Test with another invalid event id
        output_file = open(temp_file, 'w')
        status = call([prog, "load", '1.0', self.format0],
                      stdout=output_file, stderr=output_file)
        output_file.close()
        self.assertEqual(status, 2)

        # Test with an invalid filename
        output_file = open(temp_file, 'w')
        status = call([prog, "load", str(event_id), 'not_exist.data'],
                      stdout=output_file)
        output_file.close()
        self.assertEqual(status, 1)

        # Test with invalid quality flag
        output_file = open(temp_file, 'w')
        status = call([prog, "load", str(event_id), self.format0,
                      '--quality_flag', '5'], stdout=output_file,
                      stderr=output_file)
        output_file.close()
        self.assertEqual(status, 2)

        # Test with switch event id pointing to an inactive switch event
        output_file = open(temp_file, 'w')
        status = call([prog, "load", str(inactive_id), self.format1],
                      stdout=output_file, stderr=output_file)
        output_file.close()
        self.assertEqual(status, 1)

        environ['PYTHONPATH'] = self.orig_python_path

    def test_processing_valid_format0_file(self):
        # Ensure the application will correctly load data from a format0
        # file. Certain expected positions should be loaded. We then
        # retrieve all estimated positions matching our expected model_id from
        # the database and compare them with expected values.

        environ['PYTHONPATH'] = self.temp_python_path

        # Our expected model id is associated with our 1st loaded switch event
        switch_event_id = self.switch_events[0].event_id
        model_id = self.pos_models_table.get_active_model(
            switch_event_id).model_id

        # Positions from self.format0 which should be loaded.
        expected_positions = [
            EstimatedSsmPosition(
                Timestamp(2014, 353, 59918), 16572652, ModelQuality.final),
            EstimatedSsmPosition(
                Timestamp(2014, 353, 63000), 16573456, ModelQuality.final),
            EstimatedSsmPosition(
                Timestamp(2014, 353, 63600), 16573499, ModelQuality.final),
        ]

        # Run load script
        prog = "../load_estimated_tirs_ssm_positions.py"
        temp_file = "temp_output_file.txt"
        output_file = open(temp_file, 'w')
        status = check_call([
            prog, "load", str(switch_event_id), self.format0,
            '--quality_flag', '2'], stdout=output_file)
        output_file.close()
        self.assertEqual(status, 0)

        # Get all estimated positions associated with our expected model_id
        selected_positions = self.est_pos_table.get_ssm_positions(model_id)

        # Compare each obtained estimated position with the expected values.
        self.assertEqual(len(selected_positions), len(expected_positions))
        for sel_pos, exp_pos in zip(selected_positions, expected_positions):
            self.assertTrue(sel_pos == exp_pos)

        environ['PYTHONPATH'] = self.orig_python_path

    def test_verify_functionality(self):
        event_id1 = self.switch_events[11].event_id
        event_id2 = self.switch_events[12].event_id
        log = logger.Logger.get_default_logger()
        log.setLevel(CRITICAL)
        genFrontEnd = SsmPosTableGenFrontEnd(self.db_conn, log)
        parsed_args = genFrontEnd.parse_parameters([
            '--accept', '--event_id={0}'.format(event_id1),
            '--event_id={0}'.format(event_id2)])
        # Disable STDOUT for the remainder of this function
        # (function called that dumps a report we don't care about)
        previous_out = sys.stdout
        null = open(devnull, 'w')
        sys.stdout = null
        genFrontEnd.run(parsed_args)

        loadPos = LoadEstimatedTirsSsmPositions(self.db_conn, log)
        # This verify should pass (all position ok so far)
        loadPos.verify_all_positions()

        # Move event_id2 times forward 12 days
        se_manager = ManageSwitchEvents(self.db_conn)
        se_manager.edit_event(event_id2, event_time=Timestamp(2015, 162, 0))

        # Re-generate the position for event_id1 only, this will cause
        # overlap (invalid positions)
        parsed_args = genFrontEnd.parse_parameters([
            '--accept', '--event_id={0}'.format(event_id1)])
        genFrontEnd.run(parsed_args)
        with self.assertRaises(InvalidDataError):
            loadPos.verify_all_positions()

        # Re-generate the position for event_id2, this will fix
        # the overlap
        parsed_args = genFrontEnd.parse_parameters([
            '--accept', '--event_id={0}'.format(event_id2)])
        genFrontEnd.run(parsed_args)

        # This verify should pass w/o Exception once again
        loadPos.verify_all_positions()

        # restore STDOUT
        sys.stdout = previous_out
        null.close()
        log.close()

    def test_processing_valid_format1_file(self):
        # Ensure the application will correctly load data from a format1
        # file. Estimated positions loaded with this test are of a different
        # model id than in the format0 test, so there shouldn't be any
        # overlap.

        environ['PYTHONPATH'] = self.temp_python_path

        # Our expected model id is associated with our 2nd loaded switch event
        switch_event_id = self.switch_events[1].event_id
        model_id = self.pos_models_table.get_active_model(
            switch_event_id).model_id

        # Positions from self.format1 which should be loaded.
        expected_positions = [
            EstimatedSsmPosition(
                Timestamp(2014, 353, 65400), 16573614,
                ModelQuality.predicted),
            EstimatedSsmPosition(
                Timestamp(2014, 353, 66000), 16573649,
                ModelQuality.preliminary),
            EstimatedSsmPosition(
                Timestamp(2014, 353, 66600), 16573682,
                ModelQuality.final),
        ]

        # Run load script
        prog = "../load_estimated_tirs_ssm_positions.py"
        temp_file = "temp_output_file.txt"
        output_file = open(temp_file, 'w')
        status = check_call([
            prog, "load", str(switch_event_id), self.format1],
            stdout=output_file)
        output_file.close()
        self.assertEqual(status, 0)

        # Select estimated positions associated with our expected model_id
        selected_positions = self.est_pos_table.get_ssm_positions(model_id)

        # Compare each obtained estimated position with the expected values.
        self.assertEqual(len(selected_positions), len(expected_positions))
        for sel_pos, exp_pos in zip(selected_positions, expected_positions):
            self.assertTrue(sel_pos == exp_pos)

        environ['PYTHONPATH'] = self.orig_python_path

    def test_should_not_find_end_dates_with_true_can_models_or_actives(self):
        # When determining the end date of the final switch event, we should
        # be skipping any switch event whose can_model or active flag is
        # false.

        environ['PYTHONPATH'] = self.temp_python_path

        # Our model id should be associated with our 4th switch event
        switch_event_id = self.switch_events[3].event_id
        model_id = self.pos_models_table.get_active_model(
            switch_event_id).model_id

        # Run load script
        prog = "../load_estimated_tirs_ssm_positions.py"
        temp_file = "temp_output_file.txt"
        output_file = open(temp_file, 'w')
        status = check_call([
            prog, "load", str(switch_event_id), self.format1],
            stdout=output_file)
        output_file.close()
        self.assertEqual(status, 0)

        # Select all positions with our model id
        selected_positions = self.est_pos_table.get_ssm_positions(model_id)

        # We don't care about each individual switch event for this test, only
        # about the number returned, and the information contained in the first
        # and last entries.
        self.assertEqual(len(selected_positions), 10)
        self.assertTrue(
            selected_positions[0] == EstimatedSsmPosition(
                Timestamp(2014, 354, 0), 16574212, ModelQuality.preliminary))
        self.assertTrue(
            selected_positions[9] == EstimatedSsmPosition(
                Timestamp(2014, 354, 21600), 16574366,
                ModelQuality.preliminary))

        environ['PYTHONPATH'] = self.orig_python_path


if __name__ == '__main__':
    sat_attr.set_default(sat_attr.get_satellite_id_from_number(8))
    unittest.main(verbosity=2)
