#!/usr/bin/env python3

import unittest
import os
import re
from subprocess import check_call, DEVNULL, CalledProcessError
from ias.misc import satellite_attributes as sat_attr
from switch_event import SwitchEvent
from manage_tirs_ssm_switch_events import ManageSwitchEvents
from manage_tirs_ssm_models import (
    ManageSsmModels, parse_parameters, ParameterError)
from ssm_utilities import connect_to_database
from database_test_tables import DatabaseTestTables
from ias.db import database as db
from ssm_pos_model import SsmPosModelTable, ModelQuality


class TestManageEvents(unittest.TestCase):
    db_connection = None
    test_tables = None

    @classmethod
    def setUpClass(cls):
        cls.db_connection = connect_to_database()
        cls.test_tables = DatabaseTestTables(
            cls.db_connection, switch=True, model=True)

        # Make sure test tables from a previous run have been dropped.  Ignore
        # errors that occur from the table not being there
        cls.test_tables.cleanup_database(True)

        # Create the needed test tables
        cls.test_tables.setup_database()

    @classmethod
    def tearDownClass(cls):
        # Remove test tables here
        cls.test_tables.cleanup_database(False)
        cls.db_connection.close(db.DISCONNECT_ROLLBACK)

    def get_valid_events(self):
        events = [
                    SwitchEvent.from_list([1, 2013, 1, 5, False, False]),
                    SwitchEvent.from_list([2, 2013, 2, 6, False, True]),
                    SwitchEvent.from_list([3, 2013, 3, 7, True, True]),
                    SwitchEvent.from_list([4, 2013, 4, 8, True, True])
                 ]
        return events

    def test_get_models(self):
        # Insert a couple events with default models
        events = self.get_valid_events()

        # Create the test object
        manage_events = ManageSwitchEvents(self.db_connection)
        model_table = SsmPosModelTable(self.db_connection)

        # Add switch events (and default models)
        for event in events:
            added_event = manage_events.add_event(
                event.time, event.can_model, event.active)
            self.assertIsNotNone(added_event)
        self.db_connection.commit()

        # Get the model for one of the second switch event, disable the old
        # one and add a second model with modified values
        model = model_table.get_active_model(2)
        model_id = model.model_id
        model.model_id = model_table.get_next_model_id()
        model.quality_flag = ModelQuality.final
        # First try inserting the new record without disabling the old one to
        # make sure it is an error (probably not needed to test here, but it
        # was convenient)
        with self.assertRaises(Exception):
            model_table.insert_model_record(model)
        model_table.disable_model_record(model_id)
        model_table.insert_model_record(model)
        self.db_connection.commit()

        # Create the manage models object to test
        manage_models = ManageSsmModels(self.db_connection)

        # Verify no options results in an exception
        with self.assertRaises(ParameterError):
            models = manage_models.get_models()

        # Test unfound switch ID
        models = manage_models.get_models(event_id=100)
        self.assertEqual(len(models), 0)

        # Test unfound model ID
        models = manage_models.get_models(model_id=100)
        self.assertEqual(len(models), 0)

        # Get a specific switch event ID.  This one should have 2 records
        models = manage_models.get_models(event_id=2)
        self.assertEqual(len(models), 2)
        self.assertEqual(models[0].model_id, 14)
        self.assertEqual(models[0].event_id, 2)
        self.assertEqual(models[1].model_id, 11)
        self.assertEqual(models[1].event_id, 2)

        # Get a specific model ID
        models = manage_models.get_models(model_id=11)
        self.assertEqual(len(models), 1)
        self.assertEqual(models[0].model_id, 11)
        self.assertEqual(models[0].event_id, 2)

        # Get all the models
        models = manage_models.get_models(get_all=True)
        # Reverse the model list to be in descending model ID order
        models.reverse()
        self.assertEqual(len(models), 5)
        # Verify the first 4 models match the first 4 created
        for event_id, model_id, model in zip(
                range(1, 5), range(10, 14), models):
            self.assertEqual(model.model_id, model_id)
            self.assertEqual(model.event_id, event_id)
        # Verify the last model matches the extra one added for event 2
        self.assertEqual(models[4].model_id, 14)
        self.assertEqual(models[4].event_id, 2)

    def test_parameter_parsing(self):
        # Check the list sub-command arguments
        args = parse_parameters(['list', '--all'])
        self.assertEqual(args.command, 'list')
        self.assertIsNone(args.model_id)
        self.assertIsNone(args.event_id)
        self.assertTrue(args.all)

        args = parse_parameters(['list', '--model_id', '11'])
        self.assertEqual(args.command, 'list')
        self.assertEqual(args.model_id, 11)
        self.assertIsNone(args.event_id)
        self.assertFalse(args.all)

        args = parse_parameters(['list', '--event_id', '14'])
        self.assertEqual(args.command, 'list')
        self.assertIsNone(args.model_id)
        self.assertEqual(args.event_id, 14)
        self.assertFalse(args.all)

        # Check the edit options
        args = parse_parameters(['edit', '--set_valid', '5'])
        self.assertEqual(args.command, 'edit')
        self.assertEqual(args.model_id[0], 5)
        self.assertFalse(args.set_invalid)
        self.assertTrue(args.set_valid)

        args = parse_parameters(['edit', '--set_invalid', '6'])
        self.assertEqual(args.command, 'edit')
        self.assertEqual(args.model_id[0], 6)
        self.assertTrue(args.set_invalid)
        self.assertFalse(args.set_valid)

    def compare_files(self, test_file, expected_file):
        """ Routine to compare two files and report any differences
        """

        # Read the test and expected files
        with open(test_file, 'r') as f:
            test_lines = f.readlines()
        with open(expected_file, 'r') as f:
            ref_lines = f.readlines()

        # Drop the first and last line from the list lines since they are
        # messages with timestamps
        test_lines = test_lines[1:-1]

        # Dates in the results will change every time, so replace those with
        # a fixed value
        expression = re.compile("\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}")

        # Compare the lines that were read from the two files
        diff_count = 0
        self.maxDiff = None
        for line, ref in zip(test_lines, ref_lines):
            # Replace the dates that changes every run with a fixed value
            no_date_line = expression.sub("-Date stripped out-", line)
            self.assertEqual(no_date_line, ref)
            if no_date_line != ref:
                diff_count = diff_count + 1
        self.assertEqual(diff_count, 0)

    def test_run_application(self):
        # Note that this routine is expected to run after all the other tests
        # to get the expected database contents.  The Python unittest class
        # runs the tests in alphabetical order, so if tests are added they
        # might impact the results here.
        data_dir = "testdata_manage_tirs_ssm_models/"

        # Get the switch events table
        model_table = SsmPosModelTable(self.db_connection)

        # Set the application name to call
        app = "../manage_tirs_ssm_models.py"
        temp_file = "temp_list_file.txt"

        # Set the output target for stdout from the application runs.  Set
        # to None if the output is wanted on the display for debugging.
        output_target = DEVNULL
        # output_target = None

        # Save the original PYTHONPATH and add the table_override directory
        # at the start so running the application uses the test tables
        orig_path = os.environ['PYTHONPATH']
        os.environ['PYTHONPATH'] = "./table_override:" + orig_path

        # List all the models
        list_file = open(temp_file, 'w')
        status = check_call([app, "list", "--all"], stdout=list_file)
        self.assertEqual(status, 0)
        list_file.close()

        self.compare_files(temp_file, data_dir + "ref_list_all.txt")

        # List model id 13
        list_file = open(temp_file, 'w')
        status = check_call([app, "list", "--model_id", '13'],
                            stdout=list_file)
        self.assertEqual(status, 0)
        list_file.close()

        self.compare_files(temp_file, data_dir + "ref_list_model.txt")

        # List switch event id 2
        list_file = open(temp_file, 'w')
        status = check_call([app, "list", "--event_id", '2'],
                            stdout=list_file)
        self.assertEqual(status, 0)
        list_file.close()

        self.compare_files(temp_file, data_dir + "ref_list_event.txt")

        # Run the application with missing required list options to make sure
        # it reports an error
        with self.assertRaises(CalledProcessError):
            status = check_call([app, "list"], stdout=output_target,
                                stderr=output_target)

        with self.assertRaises(CalledProcessError):
            status = check_call([app, "list", "--event_id"],
                                stdout=output_target, stderr=output_target)

        with self.assertRaises(CalledProcessError):
            status = check_call([app, "list", "--model_id"],
                                stdout=output_target, stderr=output_target)

        # Test with invalid switch and model ids.  It should run successfully.
        status = check_call([app, "list", "--event_id", "500"],
                            stdout=output_target, stderr=output_target)

        status = check_call([app, "list", "--model_id", "5000"],
                            stdout=output_target, stderr=output_target)

        # Test setting a model to invalid - first make sure the model is valid
        # then set it to invalid
        model = model_table.get_model(13)
        self.assertTrue(model.valid)

        list_file = open(temp_file, 'w')
        status = check_call([app, "edit", '13', '--set_invalid'],
                            stdout=list_file)
        self.assertEqual(status, 0)
        list_file.close()

        # Confirm the listing was output as expected and the database has the
        # change
        self.compare_files(temp_file, data_dir + "ref_set_invalid.txt")
        model = model_table.get_model(13)
        self.assertFalse(model.valid)

        # Test setting a model to valid
        list_file = open(temp_file, 'w')
        status = check_call([app, "edit", '13', '--set_valid'],
                            stdout=list_file)
        self.assertEqual(status, 0)
        list_file.close()

        # Confirm the listing was output as expected and the database has the
        # change
        self.compare_files(temp_file, data_dir + "ref_set_valid.txt")
        model = model_table.get_model(13)
        self.assertTrue(model.valid)

        # Run the application with missing required edit options to make sure
        # it reports an error
        with self.assertRaises(CalledProcessError):
            status = check_call([app, "edit", "7"], stdout=output_target,
                                stderr=output_target)

        with self.assertRaises(CalledProcessError):
            status = check_call([app, "edit", "--set_valid"],
                                stdout=output_target, stderr=output_target)

        # Test with a model ID that doesn't exist
        with self.assertRaises(CalledProcessError):
            status = check_call([app, "edit", '5000', "--set_valid"],
                                stdout=output_target, stderr=output_target)

        # Remove the temporary file
        os.remove(temp_file)

        # Restore the original python path
        os.environ['PYTHONPATH'] = orig_path


if __name__ == '__main__':
    print("-" * 70)
    print("Testing manage models...\n")

    # Set up the satellite attributes for Landsat 8
    sat_attr.set_default(sat_attr.get_satellite_id_from_number(8))

    unittest.main(verbosity=2)
