#!/usr/bin/env python3

import unittest
import os
from subprocess import check_call, DEVNULL, CalledProcessError
from ias.misc import satellite_attributes as sat_attr
from switch_event import SwitchEvent, SwitchEventsTable
from manage_tirs_ssm_switch_events import (ManageSwitchEvents,
                                           parse_parameters, ParameterError,
                                           ModelValidityError,
                                           EventTooCloseError)
from ssm_utilities import connect_to_database
from database_test_tables import DatabaseTestTables
from timestamp import Timestamp
from ias.db import database as db
from ssm_pos_model import SsmPosModelTable, ModelQuality
from ias.log import logger


class TestManageEvents(unittest.TestCase):
    db_connection = None
    test_tables = None

    @classmethod
    def setUpClass(cls):
        cls.db_connection = connect_to_database()
        cls.test_tables = DatabaseTestTables(
            cls.db_connection, switch=True, model=True)

        # Make sure test tables from a previous run have been dropped.  Ignore
        # errors that occur from the table not being there
        cls.test_tables.cleanup_database(True)

        # Create the needed test tables
        cls.test_tables.setup_database()

    @classmethod
    def tearDownClass(cls):
        # Remove test tables here
        cls.test_tables.cleanup_database(False)
        cls.db_connection.close(db.DISCONNECT_ROLLBACK)

    def get_valid_events(self):
        events = [
                    SwitchEvent.from_list([1, 2013, 1, 5, False, False]),
                    SwitchEvent.from_list([2, 2013, 2, 6, False, True]),
                    SwitchEvent.from_list([3, 2013, 2, 6007, True, False]),
                    SwitchEvent.from_list([4, 2013, 3, 7, True, True]),
                    SwitchEvent.from_list([5, 2013, 4, 8, True, True])
                 ]
        return events

    def get_expected_warnings(self):
        return [False, False, True, True, False]

    def get_skip_default_parameters(self):
        return [False, False, False, False, True]

    def add_events(self):
        print("Testing adding events...")
        log = logger.Logger.get_default_logger()

        valid_events = self.get_valid_events()
        expect_warnings = self.get_expected_warnings()
        skip_defaults = self.get_skip_default_parameters()

        events_too_close = [
            # one closer than 6000 seconds (less than)
            [2013, 2, 5],
            # one equal
            [2013, 2, 6],
            # one 6000 seconds greater
            [2013, 2, 6006]
            ]

        # Create the test object
        manage = ManageSwitchEvents(self.db_connection)

        # Create the database objects needed
        events_table = SwitchEventsTable(self.db_connection)
        model_table = SsmPosModelTable(self.db_connection)

        # Add valid events using the various options
        log_count = 0
        with self.assertLogs(logger=log, level='WARNING') as catch_log:
            for index in range(len(valid_events)):
                event = valid_events[index]
                if expect_warnings[index]:
                    log_count = log_count + 1
                if skip_defaults[index]:
                    added_event = manage.add_event(event.time)
                else:
                    added_event = manage.add_event(
                        event.time, event.can_model, event.active)
                self.assertIsNotNone(added_event)
        # Verify the correct number of warning messages were found
        self.assertEqual(len(catch_log.output), log_count)

        # Make sure invalid events fail
        for event in events_too_close:
            timestamp = Timestamp(event[0], event[1], event[2])
            catch_log = []
            with self.assertLogs(logger=log, level='ERROR') as catch_log:
                with self.assertRaises(EventTooCloseError):
                    if len(event) == 5:
                        added_event = manage.add_event(
                            timestamp, event[3], event[4])
                    else:
                        added_event = manage.add_event(timestamp)
            # Verify an error message was caught
            self.assertEqual(len(catch_log.output), 1)

        # Get all the events and make sure the valid ones are there
        events = events_table.get_all_events()
        self.assertEqual(len(events), len(valid_events))
        for event, expected_event in zip(events, valid_events):
            # Since the sequence is created starting with 1, the event IDs
            # have predictable values
            self.assertEqual(event.event_id, expected_event.event_id)
            self.assertEqual(event.time, expected_event.time)
            self.assertEqual(event.can_model, expected_event.can_model)
            self.assertEqual(event.active, expected_event.active)

            # Now get the model associated with the event
            models = model_table.get_model_records(event.event_id)

            # There should only be one record with the correct valid flag and
            # the quality_flag of predicted.  The model parameters will not
            # be checked since the default model function parameters are part
            # of the SSM model table class.  The model ID should be 9 larger
            # than the event ID for this test case since the sequence starting
            # values are defined that way.
            self.assertEqual(len(models), 1)
            self.assertEqual(models[0].model_id, event.event_id + 9)
            self.assertEqual(models[0].event_id, event.event_id)
            self.assertEqual(models[0].valid, event.can_model)
            self.assertEqual(models[0].quality_flag, ModelQuality.predicted)

    def edit_events(self):
        print("Testing the editing events...")
        manage = ManageSwitchEvents(self.db_connection)
        events_table = SwitchEventsTable(self.db_connection)
        log = logger.Logger.get_default_logger()

        expected_events = self.get_valid_events()

        # Change the last event's time
        timestamp = Timestamp(2014, 7, 8)
        returned_event = manage.edit_event(5, timestamp)
        self.assertIsNotNone(returned_event)
        event = events_table.get_event(5)
        self.assertEqual(event.event_id, expected_events[4].event_id)
        self.assertEqual(event.time, timestamp)
        self.assertEqual(event.can_model, expected_events[4].can_model)
        self.assertEqual(event.active, expected_events[4].active)

        # Try to change the can_model flag on an event with a valid model
        # to make sure it isn't allowed
        with self.assertLogs(logger=log, level='ERROR') as catch_log:
            with self.assertRaises(ModelValidityError):
                returned_event = manage.edit_event(
                    5, timestamp, can_model=not expected_events[4].can_model)
            self.assertEqual(len(catch_log.output), 1)

        # Try calling edit_event without any optional parameters to make a
        # change and make sure an error is reported
        with self.assertLogs(logger=log, level='ERROR') as catch_log:
            with self.assertRaises(ParameterError):
                returned_event = manage.edit_event(5)
            self.assertEqual(len(catch_log.output), 1)

        # Get the event from the table and make sure it is unchanged
        event = events_table.get_event(5)
        self.assertEqual(event.event_id, expected_events[4].event_id)
        self.assertEqual(event.time, timestamp)
        self.assertEqual(event.can_model, expected_events[4].can_model)
        self.assertEqual(event.active, expected_events[4].active)

        # Now change an event where the model is not valid and make sure it
        # succeeds
        timestamp = Timestamp(2013, 1, 5)
        returned_event = manage.edit_event(
            1, timestamp, can_model=not expected_events[0].can_model)
        self.assertIsNotNone(returned_event)
        event = events_table.get_event(1)
        self.assertEqual(event.event_id, expected_events[0].event_id)
        self.assertEqual(event.time, timestamp)
        self.assertEqual(event.can_model, not expected_events[0].can_model)
        self.assertEqual(event.active, expected_events[0].active)

        # Now flip the can_model flag back to the original value, without
        # providing a timestamp this time since it is optional
        returned_event = manage.edit_event(
            1, can_model=expected_events[0].can_model)
        self.assertIsNotNone(returned_event)
        event = events_table.get_event(1)
        self.assertEqual(event.event_id, expected_events[0].event_id)
        self.assertEqual(event.time, timestamp)
        self.assertEqual(event.can_model, expected_events[0].can_model)
        self.assertEqual(event.active, expected_events[0].active)

        # Try to change the time on an event to make it closer than 6000
        # seconds to another event and make sure it gets an error.
        with self.assertLogs(logger=log, level='ERROR') as catch_log:
            # First, one second before
            timestamp = Timestamp(2013, 2, 5)
            with self.assertRaises(EventTooCloseError):
                returned_event = manage.edit_event(3, timestamp)

            # Now at the same time
            timestamp = Timestamp(2013, 2, 6)
            with self.assertRaises(EventTooCloseError):
                returned_event = manage.edit_event(3, timestamp)

            # Now one second after
            timestamp = Timestamp(2013, 2, 7)
            with self.assertRaises(EventTooCloseError):
                returned_event = manage.edit_event(3, timestamp)

            # Now 6000 seconds after
            timestamp = Timestamp(2013, 2, 6006)
            with self.assertRaises(EventTooCloseError):
                returned_event = manage.edit_event(3, timestamp)

            # Now 6001 seconds after
            timestamp = Timestamp(2013, 2, 6007)
            returned_event = manage.edit_event(3, timestamp)
            self.assertIsNotNone(returned_event)
        self.assertEqual(len(catch_log.output), 4)

        # Set both the can_model and active flags to Y
        timestamp = Timestamp(2013, 1, 5)
        returned_event = manage.edit_event(
            1, timestamp, can_model=not expected_events[0].can_model,
            active=not expected_events[0].active)
        self.assertIsNotNone(returned_event)
        event = events_table.get_event(1)
        self.assertEqual(event.event_id, expected_events[0].event_id)
        self.assertEqual(event.time, timestamp)
        self.assertEqual(event.can_model, not expected_events[0].can_model)
        self.assertEqual(event.active, not expected_events[0].active)

        # Try to change the active flag on an event with the can_model
        # flag set to 'Y' to make sure it isn't allowed
        with self.assertLogs(logger=log, level='ERROR') as catch_log:
            with self.assertRaises(ModelValidityError):
                returned_event = manage.edit_event(
                    1, timestamp, active=expected_events[0].active)
            self.assertEqual(len(catch_log.output), 1)

        # Set the can_model flag to N
        returned_event = manage.edit_event(1, can_model=False)
        self.assertIsNotNone(returned_event)
        event = events_table.get_event(1)
        self.assertEqual(event.event_id, expected_events[0].event_id)
        self.assertEqual(event.can_model, expected_events[0].can_model)

        # Try setting active to N again, after can_model is set to N
        returned_event = manage.edit_event(1, active=False)
        self.assertIsNotNone(returned_event)
        event = events_table.get_event(1)
        self.assertEqual(event.event_id, expected_events[0].event_id)
        self.assertEqual(event.active, expected_events[0].active)

    def test_methods(self):
        print("\nTesting the event manager...")
        self.add_events()
        self.edit_events()

    def test_parameter_parsing(self):
        # Check the add sub-command arguments
        args = parse_parameters(['add', '1999:332:76543'])
        self.assertEqual(args.command, 'add')
        self.assertEqual(args.time[0], Timestamp(1999, 332, 76543))
        self.assertEqual(args.disable_model, False)

        args = parse_parameters(['add', '1999:332:76543', '--disable_model'])
        self.assertEqual(args.command, 'add')
        self.assertEqual(args.time[0], Timestamp(1999, 332, 76543))
        self.assertEqual(args.disable_model, True)

        # Test the alternate time format
        args = parse_parameters(['add', '1999-332-21:15:43',
                                 '--disable_model'])
        self.assertEqual(args.command, 'add')
        self.assertEqual(args.time[0], Timestamp(1999, 332, 76543))
        self.assertEqual(args.disable_model, True)

        # Check the list sub-command arguments
        args = parse_parameters(['list'])
        self.assertEqual(args.command, 'list')
        self.assertEqual(args.reverse, False)
        self.assertEqual(args.can_model, False)
        self.assertEqual(args.cannot_model, False)
        self.assertEqual(args.active, False)
        self.assertEqual(args.inactive, False)
        self.assertIsNone(args.start_timestamp)
        self.assertIsNone(args.end_timestamp)

        args = parse_parameters(
            ['list', '--reverse', '--can_model', '--active',
             '--start_timestamp', '1999:333:1',
             '--end_timestamp', '2000:123:777'])
        self.assertEqual(args.command, 'list')
        self.assertEqual(args.reverse, True)
        self.assertEqual(args.can_model, True)
        self.assertEqual(args.cannot_model, False)
        self.assertEqual(args.active, True)
        self.assertEqual(args.inactive, False)
        self.assertEqual(args.start_timestamp, Timestamp(1999, 333, 1))
        self.assertEqual(args.end_timestamp, Timestamp(2000, 123, 777))

        args = parse_parameters(['list', '--cannot_model', '--inactive'])
        self.assertEqual(args.command, 'list')
        self.assertEqual(args.reverse, False)
        self.assertEqual(args.can_model, False)
        self.assertEqual(args.cannot_model, True)
        self.assertEqual(args.active, False)
        self.assertEqual(args.inactive, True)
        self.assertIsNone(args.start_timestamp)
        self.assertIsNone(args.end_timestamp)

        # Check the edit sub-command arguments
        args = parse_parameters(['edit', '5', '--time', '1999:1:2'])
        self.assertEqual(args.command, 'edit')
        self.assertEqual(args.event_id[0], 5)
        self.assertEqual(args.time, Timestamp(1999, 1, 2))
        self.assertEqual(args.enable_model, False)
        self.assertEqual(args.disable_model, False)
        self.assertEqual(args.set_active, False)
        self.assertEqual(args.set_inactive, False)

        args = parse_parameters(['edit', '5', '--enable_model',
                                 '--set_active'])
        self.assertEqual(args.command, 'edit')
        self.assertEqual(args.event_id[0], 5)
        self.assertIsNone(args.time)
        self.assertEqual(args.enable_model, True)
        self.assertEqual(args.disable_model, False)
        self.assertEqual(args.set_active, True)
        self.assertEqual(args.set_inactive, False)

        args = parse_parameters(
            ['edit', '5', '--time', '1999:1:2', '--disable_model',
             '--set_inactive'])
        self.assertEqual(args.command, 'edit')
        self.assertEqual(args.event_id[0], 5)
        self.assertEqual(args.time, Timestamp(1999, 1, 2))
        self.assertEqual(args.enable_model, False)
        self.assertEqual(args.disable_model, True)
        self.assertEqual(args.set_active, False)
        self.assertEqual(args.set_inactive, True)

        # Test the alternate time format
        args = parse_parameters(
            ['edit', '5', '--time', '1999-1-02:59:59', '--disable_model',
             '--set_inactive'])
        self.assertEqual(args.command, 'edit')
        self.assertEqual(args.event_id[0], 5)
        self.assertEqual(args.time, Timestamp(1999, 1, 10799))
        self.assertEqual(args.enable_model, False)
        self.assertEqual(args.disable_model, True)
        self.assertEqual(args.set_active, False)
        self.assertEqual(args.set_inactive, True)

    def test_run_application(self):
        # Note that this routine is expected to run after all the other tests
        # to get the expected database contents.  The Python unittest class
        # runs the tests in alphabetical order, so if tests are added they
        # might impact the results here.

        # Get the switch events table
        events_table = SwitchEventsTable(self.db_connection)

        # Set the application name to call
        app = "../manage_tirs_ssm_switch_events.py"
        temp_file = "temp_list_file.txt"

        # Set the output target for stdout from the application runs.  Set
        # to None if the output is wanted on the display for debugging.
        output_target = DEVNULL
        # output_target = None

        # Save the original PYTHONPATH and add the table_override directory
        # at the start so running the application uses the test tables
        orig_path = os.environ['PYTHONPATH']
        os.environ['PYTHONPATH'] = "./table_override:" + orig_path

        # Run the application to test adding a record
        status = check_call([app, "add", "2015:5:25"], stdout=output_target)
        self.assertEqual(status, 0)

        # The new event that was added should be switch event ID 6
        timestamp = Timestamp(2015, 5, 25)
        event = events_table.get_event(6)
        self.assertEqual(event.event_id, 6)
        self.assertEqual(event.time, timestamp)
        self.assertTrue(event.can_model)
        self.assertTrue(event.active)

        # Run the application to test adding a record with the model disabled
        status = check_call(
            [app, "add", "2015:6:25", "--disable_model"], stdout=output_target)
        self.assertEqual(status, 0)

        # The new event that was added should be switch event ID 7
        timestamp = Timestamp(2015, 6, 25)
        event = events_table.get_event(7)
        self.assertEqual(event.event_id, 7)
        self.assertEqual(event.time, timestamp)
        self.assertFalse(event.can_model)
        self.assertTrue(event.active)

        # Run the application to edit an entry
        status = check_call(
            [app, "edit", "7", "--time", "2015:6:26", "--enable_model"],
            stdout=output_target)
        self.assertEqual(status, 0)

        # The updated event that was added should be switch event ID 7
        timestamp = Timestamp(2015, 6, 26)
        event = events_table.get_event(7)
        self.assertEqual(event.event_id, 7)
        self.assertEqual(event.time, timestamp)
        self.assertTrue(event.can_model)
        self.assertTrue(event.active)

        # Run the application to edit an entry to just disable the model
        status = check_call(
            [app, "edit", "7", "--disable_model", "--set_inactive"],
            stdout=output_target)
        self.assertEqual(status, 0)

        # The updated event that was added should be switch event ID 7
        timestamp = Timestamp(2015, 6, 26)
        event = events_table.get_event(7)
        self.assertEqual(event.event_id, 7)
        self.assertEqual(event.time, timestamp)
        self.assertFalse(event.can_model)
        self.assertFalse(event.active)

        # Run the application with no edit options to make sure it reports an
        # error
        with self.assertRaises(CalledProcessError):
            status = check_call([app, "edit", "7"], stdout=output_target)

        # Run the application to list the contents
        list_file = open(temp_file, 'w')
        status = check_call([app, "list"], stdout=list_file)
        self.assertEqual(status, 0)
        list_file.close()

        # Read the generated list and the reference file
        with open(temp_file, 'r') as f:
            list_lines = f.readlines()
        with open("ref_switch_event_list_file.txt", 'r') as f:
            ref_lines = f.readlines()

        # Drop the first and last line from the list lines since they are
        # messages with timestamps
        list_lines = list_lines[1:-1]

        # Compare the lines that were read from the two files
        diff_count = 0
        for line, ref in zip(list_lines, ref_lines):
            self.assertEqual(line, ref)
            if line != ref:
                diff_count = diff_count + 1
        self.assertEqual(diff_count, 0)

        # Remove the temporary file
        os.remove(temp_file)

        # Restore the original python path
        os.environ['PYTHONPATH'] = orig_path


if __name__ == '__main__':
    print("-" * 70)
    print("Testing manage events...\n")

    # Set up the satellite attributes for Landsat 8
    sat_attr.set_default(sat_attr.get_satellite_id_from_number(8))

    unittest.main(verbosity=2)
