#!/usr/bin/env python3

from database_objects import DatabaseObjects
from database_structure import DatabaseStructure
from database_test_tables import DatabaseTestTables
from perform_initial_tirs_ssm_fit import (
    InitialFit, InitialFitError, parse_parameters)
from ias.db import database as db
from ias.log import logger
from ias.misc import satellite_attributes as sat_attr
from manage_tirs_ssm_switch_events import (ManageSwitchEvents)
from ssm_exceptions import (SsmException, MissingDataError, DateRangeError)
from ssm_pos_model import SsmPosModel, SsmPosModelTable, ModelQuality
from ssm_utilities import connect_to_database
from switch_event import SwitchEvent, SwitchEventsTable
from timestamp import Timestamp
import argparse
import os
import shutil
import subprocess
import sys
import unittest


class TestInitialFitEvents(unittest.TestCase):
    db_connection = None
    test_tables = None
    initialFit = None
    curr_dir = os.getcwd()
    initial_fit_dir = curr_dir + "/initial_fit"

    @classmethod
    def setUpClass(cls):
        cls.db_connection = connect_to_database()
        cls.test_tables = DatabaseTestTables(
            cls.db_connection, switch=True, model=True, position=True,
            telemetry=True)
        cls.reset_database_tables()

        # Instantiate class to test
        cls.initialFit = InitialFit(cls.db_connection)

        # Make sure the current and parent directories are in the user's path
        cls.orig_path = os.environ['PATH']
        os.environ['PATH'] = (os.getcwd() + ":" + os.getcwd() + "/..:" +
                              cls.orig_path)

        # Save the original PYTHONPATH and add the table_override directory
        # at the start so running the application uses the test tables
        cls.orig_python_path = os.environ['PYTHONPATH']
        os.environ['PYTHONPATH'] = (cls.curr_dir + "/table_override:" +
                                    cls.orig_python_path)

    @classmethod
    def reset_database_tables(cls):
        """ Reset data in database to be able to run the run_fit routine
        """
        # Make sure test tables from a previous run have been dropped.  Ignore
        # errors that occur from the table not being there
        cls.test_tables.cleanup_database(True)
        cls.test_tables.setup_database()

    @classmethod
    def tearDownClass(cls):
        # Remove test tables here
        cls.test_tables.cleanup_database(False)
        cls.db_connection.close(db.DISCONNECT_ROLLBACK)

        if os.path.exists(cls.curr_dir + "/generate_ssm_encoder_file"):
            os.unlink(cls.curr_dir + "/generate_ssm_encoder_file")

        # Reset path to the original path
        os.environ['PATH'] = cls.orig_path

        # Restore the original python path
        os.environ['PYTHONPATH'] = cls.orig_python_path

    def get_events(self):
        """ Events with which to populate the switch event table
        """
        # Note the telemetry encoder csv file used for the tests really only
        # has data from 2015:024 -> 2016:027. However, the C executable
        # does not check the dates, and the data works as expected for the
        # unit tests, so it is considered sufficient for these unit tests.
        # We do not do any "indepth" testing of the C executable
        # results for the most part, just make sure it runs.  We do check
        # the format of the report file and some values in one test, but
        # expect the C executable tests to cover the rest of it.
        events = [
                    SwitchEvent.from_list([1, 2015, 302, 61500, True, True]),
                    SwitchEvent.from_list([2, 2015, 317, 47390, True, False]),
                    SwitchEvent.from_list([3, 2015, 330, 39840, True, True]),
                 ]
        return events

    def create_events(self):
        """ Populates database tables so the unit tests have data with which
            to test
        """
        log = logger.Logger.get_default_logger()
        valid_events = self.get_events()

        # Create the test object
        manage = ManageSwitchEvents(self.db_connection)

        # Create the database objects needed
        events_table = SwitchEventsTable(self.db_connection)
        model_table = SsmPosModelTable(self.db_connection)

        # Add events so we have something to work with
        for event in valid_events:
            added_event = manage.add_event(
                event.time, event.can_model, event.active)

    def populate_telemetry(self):
        """ Populates the TIRS TELEMETRY COMMAND table for the unit tests
        """
        working_dir = os.getcwd()
        test_data_dir = "testdata_perform_initial_tirs_ssm_fit"
        os.chdir(test_data_dir)
        cmd = "./populate_telemetry.csh"
        subprocess.check_call([cmd], shell=True)
        os.chdir(working_dir)

    def redirect_c_stdout(self):
        """ Redirects stdout for the C application which is more difficult
            to do than redirecting the python stdout
        """
        # Flush the c level buffer
        sys.stdout.flush()
        # Duplicate stdout (file descriptor 1)
        newstdout = os.dup(1)
        # Set devnull
        devnull = os.open('/dev/null', os.O_WRONLY)
        # Duplicate the dev null file descriptor
        os.dup2(devnull, 1)
        # Close devnull after duplication
        os.close(devnull)
        # Original stdout so python stdout will still work
        sys.stdout = os.fdopen(newstdout, 'w')

    def setup(self):
        """ Setup
            First test run to call the create events function to populate data
        """
        self.create_events()
        self.populate_telemetry()

    def test(self):
        # Redirect output for the C application (can't just redirect stdout
        # for python, that does not work)
        self.redirect_c_stdout()

        # Set up the encoder  file for this unit test
        if os.path.exists(self.curr_dir + "/generate_ssm_encoder_file"):
            os.unlink(self.curr_dir + "/generate_ssm_encoder_file")
        src = self.curr_dir + \
            "/testdata_perform_initial_tirs_ssm_fit/generate_ssm_encoder_file"
        dest = self.curr_dir + "/generate_ssm_encoder_file"
        os.symlink(src, dest)

        # Clean data if needed
        if os.path.exists(self.initial_fit_dir):
            shutil.rmtree(self.initial_fit_dir)
        self.setup()

    def test_convert_j2000_seconds_to_days_sod(self):
        """ Testing convert_j2000_seconds_to_days_sod
        """
        seconds = 501808003
        (j2000_days, j2000_secs) = \
            self.initialFit.convert_j2000_seconds_to_days_sod(seconds)
        self.assertEqual(j2000_days, 5807)
        self.assertEqual(j2000_secs, 83203)

        seconds = 250904001
        (j2000_days, j2000_secs) = \
            self.initialFit.convert_j2000_seconds_to_days_sod(seconds)
        self.assertEqual(j2000_days, 2903)
        self.assertEqual(j2000_secs, 84801)

    def test_convert_year_doy_sod_to_j2000_seconds(self):
        """ Testing convert_year_doy_sod_to_j2000_seconds
        """
        time = Timestamp(2015, 330, 39840)
        seconds = self.initialFit.convert_year_doy_sod_to_j2000_seconds(time)
        self.assertEqual(seconds, 501807904.184)

        # Year before 2000
        time2 = Timestamp(1999, 330, 39840)
        with self.assertRaises(DateRangeError):
            seconds = \
                self.initialFit.convert_year_doy_sod_to_j2000_seconds(time2)

    def test_get_model_quality(self):
        """ Testing get_model_quality
        """
        models = self.initialFit.model_table.get_model_records(3)
        num_models = len(models)
        self.assertEqual(num_models, 1)

        quality = self.initialFit.get_model_quality(models)
        self.assertEqual(quality, ModelQuality.predicted)

        # Disable the model in order to verify predicted is returned
        # when there are no valid models
        self.initialFit.model_table.disable_model_record(models[0].model_id)
        quality = self.initialFit.get_model_quality(models)
        self.assertEqual(quality, ModelQuality.predicted)

    def test_get_number_telemetry_records(self):
        """ Testing get_number_telemetry_records
        """
        self.initialFit.populate_event_information()
        number_telemetry_records = \
            self.initialFit.get_number_actual_telemetry_records()
        self.assertEqual(number_telemetry_records, 833)

    def test_is_leap_year(self):
        """ Testing is_leap_year
        """
        # Known leap years
        self.assertTrue(self.initialFit.is_leap_year(2000))
        self.assertTrue(self.initialFit.is_leap_year(2004))
        # Known non leap years
        self.assertFalse(self.initialFit.is_leap_year(2001))
        self.assertFalse(self.initialFit.is_leap_year(2015))

    def test_parameter_parsing(self):
        """ Testing parameter parsing
        """
        # Display tests only take fit event id as an argument
        args = parse_parameters([self.curr_dir])
        self.assertEqual(args.fit_directory[0], os.getcwd())
        # Test default is 300 if not specified on command line
        self.assertEqual(args.num_telemetry_records, 300)

        args = parse_parameters([self.curr_dir, '--num_telemetry_records',
                                '10'])
        self.assertEqual(args.fit_directory[0], os.getcwd())
        self.assertEqual(args.num_telemetry_records, 10)

        args = parse_parameters([self.curr_dir, '--num_telemetry_records',
                                '100'])
        self.assertEqual(args.fit_directory[0], os.getcwd())
        self.assertEqual(args.num_telemetry_records, 100)

    def test_populate_event_information(self):
        """ Testing populate_event_information
        """
        time = Timestamp(2015, 330, 39840)
        self.initialFit.populate_event_information()
        self.assertEqual(self.initialFit.event_id, 3)
        self.assertEqual(self.initialFit.event_time, time)

    def test_run(self):
        """ Testing run
        """
        # Reset data in database to be able to run the run_fit routine
        self.reset_database_tables()
        self.setup()

        log = logger.Logger.get_default_logger()

        args = parse_parameters([self.initial_fit_dir,
                                '--num_telemetry_records', '900'])

        # Fit runs but quality is predicted since number of telemetry records
        # existing in tables is less than 900
        self.initialFit.run(args)
        active_model = self.initialFit.model_table.get_active_model(3)
        current_model_string = active_model.__str__()
        self.assertTrue(os.path.exists(self.initial_fit_dir))
        self.assertEqual(active_model.quality_flag, ModelQuality.predicted)
        shutil.rmtree(self.initial_fit_dir)

        with self.assertLogs(logger=log, level='INFO') as catch_log:
            self.initialFit.run(args)
        message = catch_log.output[-1]
        self.assertTrue(message.find("not performing fit.") > 0)

        # Default telemetry records value will be used and fit will be
        # performed
        args = parse_parameters([self.initial_fit_dir])
        self.initialFit.run(args)
        self.assertTrue(os.path.exists(self.initial_fit_dir))

        # Fit does not run but exits gracefully since quality is now sufficient
        args = parse_parameters([self.initial_fit_dir])
        with self.assertLogs(logger=log, level='DEBUG') as catch_log:
            self.initialFit.run(args)
        message = catch_log.output[-1]
        self.assertTrue(message.find(
            "Model quality sufficient, no fit needed."))
        shutil.rmtree(self.initial_fit_dir)

    def test_run_fit(self):
        """ Testing run_fit
        """
        # Reset data in database to be able to run the run_fit routine
        self.reset_database_tables()
        self.setup()

        time = Timestamp(2015, 330, 39840)
        self.initialFit.populate_event_information()

        # Directory that cannot be created
        fit_dir = "/test_initial_fit"
        with self.assertRaises(OSError):
            self.initialFit.run_fit(fit_dir)

        # Get initial model information before performing fit
        initial_active_model = self.initialFit.model_table.get_active_model(3)
        initial_model_string = initial_active_model.__str__()

        # Run fit
        self.initialFit.run_fit(self.initial_fit_dir)
        # Check that the accept was not called and the database was not updated
        active_model = self.initialFit.model_table.get_active_model(3)
        current_model_string = active_model.__str__()
        self.assertTrue(os.path.exists(self.initial_fit_dir))
        self.assertEqual(initial_model_string, current_model_string)
        self.assertEqual(active_model.quality_flag, ModelQuality.predicted)
        # Now test running the fit where accept does get called, keyed off of
        # if the quality is provided
        self.initialFit.run_fit(self.initial_fit_dir, ModelQuality.preliminary)

        # Check fit results
        active_model = self.initialFit.model_table.get_active_model(3)
        current_model_string = active_model.__str__()
        self.assertNotEqual(initial_model_string, current_model_string)
        self.assertEqual(active_model.quality_flag, ModelQuality.preliminary)
        shutil.rmtree(self.initial_fit_dir)

    def test_whole_application(self):
        """ Testing running the initial fit application
        """
        # Note that this routine is expected to run after all the other tests
        # to get the expected database contents.  The Python unittest class
        # runs the tests in alphabetical order, so if tests are added they
        # might impact the results here. (Hence the use of the 'whole' in the
        # method name.)

        # Set the application name to call
        app = "../perform_initial_tirs_ssm_fit.py"

        # Reset data in database to be able to run the run_fit routine
        self.reset_database_tables()
        self.setup()

        # Set the output target for stdout from the application runs.  Set
        # to None if the output is wanted on the display for debugging.
        output_target = subprocess.DEVNULL

        # Normal full run
        status = subprocess.check_call([app, self.initial_fit_dir],
                                       stdout=output_target)
        self.assertEqual(status, 0)
        dir_name = self.initial_fit_dir + "/003_2015_330_39840"
        self.assertTrue(os.path.exists(dir_name))
        shutil.rmtree(self.initial_fit_dir)


if __name__ == '__main__':

    print("-" * 70)
    print("Testing initial fit strategy...\n")

    # Set up the satellite attributes for Landsat 8
    sat_attr.set_default(sat_attr.get_satellite_id_from_number(8))

    # Buffers stdout and only displays on a unit test error
    # Also want my tests to fail on the first failure so the directory still
    # exists with the data that spawnewd the failure, (since the tests
    # mainly use the same event_id (directory) over and over)
    unittest.main(verbosity=2, buffer=True, failfast=True)
