#!/usr/bin/env python3

import unittest
from datetime import timedelta, datetime
import os
from ias.log import logger
from ias.db import database as db
from database_structure import DatabaseStructure
from ssm_utilities import connect_to_database
from place_lpgs_orders_for_time_period import (PlaceOrdersFrontEnd,
                                               PlaceLpgsOrders,
                                               parse_parameters)
from database_test_tables import DatabaseTestTables
from manage_tirs_ssm_switch_events import ManageSwitchEvents
from subprocess import check_call, DEVNULL, CalledProcessError
from ias.misc import satellite_attributes as sat_attr
from timestamp import Timestamp


class TestLoadLpgsOrders(unittest.TestCase):
    db_connection = None
    l8_db = None
    start_time = None
    stop_time = None

    @classmethod
    def setUpClass(cls):
        # Need to make seperate connections to different schemas
        cls.db_connection = connect_to_database(True)
        cls.l8_db = connect_to_database(False)

        cls.start_time = Timestamp(2013, 1, 61)
        cls.stop_time = Timestamp(2013, 20, 1)
        cls.drop_db_objects(True)
        cls.create_dummy_table()
        cls.create_dummy_stored_proc()

        # Create test versions of the tables (needed for test 4 only)
        cls.test_tables = DatabaseTestTables(
            cls.l8_db, switch=True, model=True, position=True)

        # Make sure test tables from a previous run have been dropped.  Ignore
        # errors that occur from the table not being there
        cls.test_tables.cleanup_database(True)

        # Create the needed test tables
        cls.test_tables.setup_database()

    @classmethod
    def tearDownClass(cls):
        cls.drop_db_objects()
        cls.test_tables.cleanup_database(False)
        cls.db_connection.close(db.DISCONNECT_ROLLBACK)
        cls.l8_db.close(db.DISCONNECT_ROLLBACK)

    @classmethod
    def drop_db_objects(self, ignore_errors=False):
        # Drop synonym added in the DB build, so that we can create a local
        # copy of the proc. This will only be present the first run of
        # this script after a fresh DB build.
        DatabaseStructure.drop_object(
            self.db_connection, "DBPROC_TIRS_AOS_ENQUEUE_MSG", "SYNONYM",
            True)
        DatabaseStructure.drop_object(
            self.db_connection, "DBPROC_TIRS_AOS_ENQUEUE_MSG", "PROCEDURE",
            ignore_errors)
        DatabaseStructure.drop_object(
            self.db_connection, "TEST_LPGS_ORDERS_12345", "TABLE",
            ignore_errors)

    @classmethod
    def create_dummy_table(self):
        sql = """CREATE TABLE "TEST_LPGS_ORDERS_12345"
                (
                  START_DATE DATE,
                  STOP_DATE DATE,
                  SUBMIT_DATE TIMESTAMP(7)
                )"""
        cursor = self.db_connection.get_cursor()
        cursor.execute(sql)
        cursor.close()

    @classmethod
    def create_dummy_stored_proc(self):
        sql = """CREATE PROCEDURE "DBPROC_TIRS_AOS_ENQUEUE_MSG"
                (
                  START_D IN DATE
                , STOP_D IN DATE
                ) AS
                BEGIN
                  insert into test_lpgs_orders_12345(start_date,stop_date,
                                                     submit_date)
                     values(start_d,stop_d,current_timestamp);
                END DBPROC_TIRS_AOS_ENQUEUE_MSG;"""
        cursor = self.db_connection.get_cursor()
        cursor.execute(sql)
        cursor.close()

    def test1_parse_cmd_args(self):
        """ Command line argument parsing test
        """
        args1 = ['--start_time={0}'.format(self.start_time),
                 '--stop_time={0}'.format(self.stop_time)]
        parse = parse_parameters(args1)
        self.assertEqual(parse.start_time, self.start_time)
        self.assertEqual(parse.stop_time, self.stop_time)
        self.assertFalse(parse.force)
        self.assertIsNone(parse.event_id)

        args2 = ['--start_time={0}'.format(self.start_time),
                 '--stop_time={0}'.format(self.stop_time),
                 '--force']
        parse = parse_parameters(args2)
        self.assertEqual(parse.start_time, self.start_time)
        self.assertEqual(parse.stop_time, self.stop_time)
        self.assertTrue(parse.force)
        self.assertIsNone(parse.event_id)

        args3 = ['--event_id=999']
        parse = parse_parameters(args3)
        self.assertIsNone(parse.start_time)
        self.assertIsNone(parse.stop_time)
        self.assertFalse(parse.force)
        self.assertEqual(parse.event_id, 999)

        args4 = ['--event_id=999', '--force']
        parse = parse_parameters(args4)
        self.assertIsNone(parse.start_time)
        self.assertIsNone(parse.stop_time)
        self.assertTrue(parse.force)
        self.assertEqual(parse.event_id, 999)

    def test2_place_order(self):
        """ Place orders test
        """
        place_ord = PlaceLpgsOrders(self.db_connection)
        place_ord.place_order_for_time(self.start_time, self.stop_time)
        self.db_connection.commit()

        # Check the date range is in the dummy table
        sql = "SELECT start_date, stop_date FROM test_lpgs_orders_12345"
        cursor = self.db_connection.get_cursor()
        cursor.execute(sql)
        rows = cursor.fetchall()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0][0], self.start_time.get_datetime())
        self.assertEqual(rows[0][1], self.stop_time.get_datetime())

    def test3_run_application(self):
        """ Run the application tests
        """
        # Set the application name to call
        app = "../place_lpgs_orders_for_time_period.py"

        # Set the output target for stdout from the application runs.  Set
        # to None if the output is wanted on the display for debugging.
        output_target = DEVNULL
        # output_target = None

        # Need some current times to use (w/o the force option you
        # can not pass old dates)
        right_now = datetime.today()
        today_start_time = Timestamp.from_datetime(right_now -
                                                   timedelta(days=7))
        today_stop_time = Timestamp.from_datetime(right_now)

        # Run script with current dates
        status = check_call([app, '--start_time={0}'.format(today_start_time),
                             '--stop_time={0}'.format(today_stop_time)],
                            stdout=output_target,
                            stderr=output_target)
        self.assertEqual(status, 0)

        # Run the application with missing required list options to make sure
        # it reports an error
        with self.assertRaises(CalledProcessError):
            status = check_call([app, "--start_time=2000:1:1"],
                                stdout=output_target,
                                stderr=output_target)

        # Run the application with conflicting list options to make sure
        # it reports an error
        with self.assertRaises(CalledProcessError):
            status = check_call([app, "--start_time=2000:1:1",
                                 "--stop_time=2000:5:1", "--event_id=99"],
                                stdout=output_target, stderr=output_target)

        # Run the application with old dates w/o the force option
        with self.assertRaises(CalledProcessError):
            status = check_call([app, "--start_time=2000:1:1",
                                 "--stop_time=2000:5:1"],
                                stdout=output_target, stderr=output_target)

        # Run the application with old dates w/ the force option
        status = check_call([app, "--start_time=2000:1:1",
                             "--stop_time=2000:5:1", "--force"],
                            stdout=output_target, stderr=output_target)
        self.assertEqual(status, 0)

        # Run the application with too large of time range
        with self.assertRaises(CalledProcessError):
            status = check_call([app, "--start_time=2000:1:1",
                                 "--stop_time=2001:5:1", "--force"],
                                stdout=output_target, stderr=output_target)

        # Run the application with both time range and event_id
        with self.assertRaises(CalledProcessError):
            status = check_call([app, "--start_time=2000:1:1",
                                 "--stop_time=2001:5:1", "--event_id=999"],
                                stdout=output_target, stderr=output_target)

    def test4_event_id_conv(self):
        """ Test the event_id is converted into the proper time range
        """
        first_switch_time = Timestamp(2013, 90, 1)
        second_switch_time = Timestamp(2013, 105, 1)

        # Create 2 switch events
        se_manager = ManageSwitchEvents(self.l8_db)
        first_event = se_manager.add_event(first_switch_time, True, True)
        # Don't need to keep the 2nd event, we just need it so we have an end
        # time
        se_manager.add_event(second_switch_time, True, True)

        # Set the application name to call
        app = "../place_lpgs_orders_for_time_period.py"
        output_target = DEVNULL

        # Save the original PYTHONPATH and add the table_override directory
        # at the start so running the application uses the test tables
        orig_path = os.environ['PYTHONPATH']
        os.environ['PYTHONPATH'] = "./table_override:" + orig_path

        status = check_call([app,
                            '--event_id={0}'.format(first_event.event_id)],
                            stdout=output_target, stderr=output_target)
        self.assertEqual(status, 0)

        # Check the switch times are in the dummy table
        sql = ("SELECT start_date, stop_date FROM test_lpgs_orders_12345 "
               "ORDER BY submit_date DESC")
        cursor = self.db_connection.get_cursor()
        cursor.execute(sql)
        rows = cursor.fetchall()
        self.assertEqual(rows[0][0], first_switch_time.get_datetime())
        self.assertEqual(rows[0][1], second_switch_time.get_datetime())

        # Restore the original python path
        os.environ['PYTHONPATH'] = orig_path


if __name__ == '__main__':
    print("-" * 70)
    print("Testing submit LPGS orders..\n")

    # Set up the satellite attributes for Landsat 8
    sat_attr.set_default(sat_attr.get_satellite_id_from_number(8))

    unittest.main(verbosity=2)
