#!/usr/bin/env python3

import unittest
from database_objects import DatabaseObjects
from database_structure import DatabaseStructure
from database_test_tables import DatabaseTestTables
from datetime import datetime
from ias.db import database as db
from ias.misc import satellite_attributes as sat_attr
from manage_tirs_ssm_switch_events import ManageSwitchEvents
from switch_event import SwitchEvent, SwitchEventsTable
from ssm_exceptions import ModelNotFoundError, ActiveModelExistsError
from ssm_pos_model import ModelQuality, SsmPosModel, SsmPosModelTable
from ssm_utilities import connect_to_database
from timestamp import Timestamp


class TestSsmPosModel(unittest.TestCase):
    """ Tests for the SsmPosModel class """
    db_connection = None
    test_tables = None

    @classmethod
    def setUpClass(cls):
        cls.db_connection = connect_to_database()
        cls.test_tables = DatabaseTestTables(
            cls.db_connection, switch=True, model=True)

        # Make sure test tables from a previous run have been dropped.  Ignore
        # errors that occur from the table not being there
        cls.test_tables.cleanup_database(True)

        # Create the needed test tables
        cls.test_tables.setup_database()

    @classmethod
    def tearDownClass(cls):
        # Remove test tables here
        cls.test_tables.cleanup_database(False)
        cls.db_connection.close(db.DISCONNECT_ROLLBACK)

    def models_equal(self, model1, model2, check_ids=None):
        """ Compares equality of models provided
        """
        if check_ids is not None:
            self.assertEqual(model1.model_id, model2.model_id)
            self.assertEqual(model1.event_id, model2.event_id)

        # Model parameters should be equal
        self.assertAlmostEqual(model1.alg_version,
                               model2.alg_version)
        self.assertEqual(model1.encoder_nadir_pos,
                         model2.encoder_nadir_pos)
        self.assertAlmostEqual(model1.a0, model2.a0)
        self.assertAlmostEqual(model1.a1, model2.a1)
        self.assertAlmostEqual(model1.a2, model2.a2)
        self.assertAlmostEqual(model1.a3, model2.a3)
        self.assertAlmostEqual(model1.tau1, model2.tau1)
        self.assertAlmostEqual(model1.tau2, model2.tau2)
        self.assertAlmostEqual(model1.tau3, model2.tau3)
        self.assertAlmostEqual(model1.slope, model2.slope)
        self.assertAlmostEqual(model1.quality_flag, model2.quality_flag)
        self.assertAlmostEqual(model1.valid, model2.valid)
        self.assertAlmostEqual(model1.date_disabled, model2.date_disabled)

    def test_ssm_model_pos(self):
        """ Test the ssm_model_pos class
        """
        # Create a model
        model1 = SsmPosModel(1, 1, 1.0, 16574079, 1646.13, 292.777, 272.844,
                             102.869, 1273.738, 0.188, 1.421, 4.74,
                             ModelQuality.predicted, True)

        # Check the model parameters
        self.assertEqual(model1.model_id, 1)
        self.assertEqual(model1.event_id, 1)
        self.assertAlmostEqual(model1.alg_version, 1.0)
        self.assertEqual(model1.encoder_nadir_pos, 16574079)
        self.assertAlmostEqual(model1.a0, 1646.13)
        self.assertAlmostEqual(model1.a1, 292.777)
        self.assertAlmostEqual(model1.a2, 272.844)
        self.assertAlmostEqual(model1.a3, 102.869)
        self.assertAlmostEqual(model1.tau1, 1273.738)
        self.assertAlmostEqual(model1.tau2, 0.188)
        self.assertAlmostEqual(model1.tau3, 1.421)
        self.assertAlmostEqual(model1.slope, 4.74)
        self.assertEqual(model1.quality_flag, 0)
        # Check the valid flag accessor
        self.assertEqual(model1.get_valid_flag(), 'Y')
        self.assertIsNone(model1.date_added)
        self.assertIsNone(model1.date_disabled)

        # Check the to_list function
        model1_list = model1.to_list()
        self.assertEqual(len(model1_list), 16)

        # Check string representation
        # Set maxDiff so the full string is shown upon error
        self.maxDiff = None
        self.assertEqual(model1.__str__(), "model_id:1, event_id:1, "
                         "alg_version:1.0, encoder_nadir_pos:16574079, "
                         "a0:1646.13, a1:292.777, a2:272.844, a3:102.869, "
                         "tau1:1273.738, tau2:0.188, tau3:1.421, slope:4.74, "
                         "valid:Y, quality_flag:0, date_added:None, "
                         "date_disabled:None")

    def test_model_quality(self):
        """ Test the ModelQuality class
        """
        self.assertEqual(ModelQuality.predicted, 0)
        self.assertEqual(ModelQuality.preliminary, 1)
        self.assertEqual(ModelQuality.final, 2)

        self.assertEqual(str(ModelQuality.predicted), "predicted")
        self.assertEqual(str(ModelQuality.preliminary), "preliminary")
        self.assertEqual(str(ModelQuality.final), "final")

    def test_ssm_pos_model_table(self):
        """ Test the ssm_model_table class
        """
        # Create an SSMPOsModel table object
        model_table = SsmPosModelTable(self.db_connection)

        # Verify number of table columns
        self.assertEqual(len(model_table._table_columns), 16)

        # Verify the table and sequence names are the expected values
        self.assertEqual(model_table.table_name, DatabaseObjects.get_name(
            "position_model_table"))
        self.assertEqual(model_table.model_id_sequence,
                         DatabaseObjects.get_name("model_id_sequence"))

        # Create a switch event to use in model tests
        timestamp1 = Timestamp(2014, 33, 1234)

        # Get the default model, (since one will not exist in the DB, default
        # parameters will be returned)
        default_model = model_table.get_default_model(timestamp1)
        self.assertEqual(default_model.model_id, 0)
        self.assertEqual(default_model.event_id, 0)

        # Now create switch event in DB which will also create a model
        manage = ManageSwitchEvents(self.db_connection)
        added_event = manage.add_event(timestamp1)
        self.db_connection.commit()

        # Verify model was inserted when the switch event was created in DB
        # No assert necessary since an exception will be thrown on failure
        # which the unit test framework will flag as an error
        curr_model = model_table.get_model_records(1)
        self.assertEqual(len(curr_model), 1)

        # Verify the default model and the model inserted when the switch
        # event was added are equal
        for mod in curr_model:
            self.models_equal(default_model, mod)
            # Test get active model
            active_model = model_table.get_active_model(mod.event_id)
            self.models_equal(mod, active_model, True)

            # Test disable record
            model_table.disable_model_record(mod.model_id)
            self.db_connection.commit()
            model = model_table.get_model(mod.model_id)
            self.assertIsNotNone(model.date_disabled)

            # Test update validity functionality
            self.assertTrue(model.valid)
            model_table.update_model_validity(mod.model_id, False)
            self.db_connection.commit()
            model = model_table.get_model(mod.model_id)
            self.assertFalse(model.valid)

        # Create a switch event to use in model tests
        timestamp1 = Timestamp(2014, 54, 1234)
        added_event = manage.add_event(timestamp1)
        self.db_connection.commit()

        # Insert a model
        next_model_id = model_table.get_next_model_id()
        insert_model = SsmPosModel(next_model_id, 1, 1.0, 16574079, 1382.427,
                                   455.287, -13.287, 100.10, 3265.12, 0.193,
                                   1.948, 3.508, ModelQuality.final, True)
        model_table.insert_model_record(insert_model)
        self.db_connection.commit()
        # Get active model, should be the one we just inserted
        active_model = model_table.get_active_model(insert_model.event_id)
        self.assertEqual(next_model_id, active_model.model_id)
        # Get default model, should be the one we just inserted
        default_model = model_table.get_default_model(timestamp1)
        self.models_equal(active_model, default_model)

        # Now test for exceptions that should be raised
        insert_model = SsmPosModel(next_model_id, 1, 1.0, 16574079, 1382.427,
                                   455.287, -13.287, 100.10, 3265.12, 0.193,
                                   1.948, 3.508, ModelQuality.final, True)
        with self.assertRaises(ActiveModelExistsError):
            model_table.insert_model_record(insert_model)
        # Disabling a model record that does not exist
        with self.assertRaises(ModelNotFoundError):
            model_table.disable_model_record(99)
        # Updating the validity of a model record that does not exist
        with self.assertRaises(ModelNotFoundError):
            model_table.update_model_validity(99, False)
        # Retrieving a model record that does not exist
        with self.assertRaises(ModelNotFoundError):
            model_table.get_active_model(99)
        with self.assertRaises(ModelNotFoundError):
            model_table.get_model(99)

if __name__ == '__main__':
    print("-" * 70)
    print("Testing ssm pos model...\n")

    # Set up the satellite attributes for Landsat 8
    sat_attr.set_default(sat_attr.get_satellite_id_from_number(8))

    unittest.main(verbosity=2)
