#!/usr/bin/env python3

import unittest
from datetime import datetime, timedelta
from switch_event import SwitchEvent, SwitchEventsTable
from timestamp import Timestamp
from ssm_utilities import connect_to_database
from database_objects import DatabaseObjects
from database_test_tables import DatabaseTestTables
from ias.db import database as db
from ias.misc import satellite_attributes as sat_attr
from ssm_exceptions import EventNotFoundError
from tirs_ssm_model_fit_parameters import ModelFitParameters


class TestSwitchEvent(unittest.TestCase):
    """ Tests for the SwitchEvent class """
    db_connection = None
    test_tables = None

    @classmethod
    def setUpClass(cls):
        cls.db_connection = connect_to_database()
        cls.test_tables = DatabaseTestTables(
            cls.db_connection, switch=True, model=False)

        # Make sure test tables from a previous run have been dropped.  Ignore
        # errors that occur from the table not being there
        cls.test_tables.cleanup_database(True)

        # Create the needed test tables
        cls.test_tables.setup_database()

    @classmethod
    def tearDownClass(cls):
        # Remove test tables here
        cls.test_tables.cleanup_database(False)
        cls.db_connection.close(db.DISCONNECT_ROLLBACK)

    def test1_switch_event(self):
        """ Test SwitchEvent objects
        """

        # Create some timestamps
        timestamp1 = Timestamp(2013, 44, 5432)
        timestamp2 = Timestamp(2014, 45, 5433)
        timestamp3 = Timestamp(2015, 46, 5434)

        # Create some new switch events
        event1 = SwitchEvent(1, timestamp1, False, False)
        event2 = SwitchEvent(2, timestamp2, True, True)
        event3 = SwitchEvent(3, timestamp3, False, True)

        # Check the event fields
        self.assertEqual(event1.event_id, 1)
        self.assertEqual(event1.time, timestamp1)
        self.assertFalse(event1.can_model)
        self.assertFalse(event1.active)

        self.assertEqual(event2.event_id, 2)
        self.assertEqual(event2.time, timestamp2)
        self.assertTrue(event2.can_model)
        self.assertTrue(event2.active)

        self.assertEqual(event3.event_id, 3)
        self.assertEqual(event3.time, timestamp3)
        self.assertFalse(event3.can_model)
        self.assertTrue(event3.active)

        # Check can_model accessor
        self.assertEqual(event1.get_can_model_flag(), "N")
        self.assertEqual(event2.get_can_model_flag(), "Y")

        # Check active flag accessor
        self.assertEqual(event1.get_active_flag(), "N")
        self.assertEqual(event3.get_active_flag(), "Y")

        # Check string representation
        self.assertEqual(str(event1), "id:1, time:2013:44:5432, can_model:N, "
                         "active:N")
        self.assertEqual(str(event3), "id:3, time:2015:46:5434, can_model:N, "
                         "active:Y")

        # Test creating an event from a list
        event = SwitchEvent.from_list([3, 2015, 46, 5434, False, True])
        self.assertEqual(event3.event_id, event.event_id)
        self.assertEqual(event3.time, event.time)
        self.assertEqual(event3.can_model, event.can_model)
        self.assertEqual(event3.active, event.active)

        # Test the equal operator
        self.assertEqual(event, event3)
        event.event_id = 4
        self.assertFalse(event == event3)
        event.event_id = event3.event_id
        event.can_model = not event3.can_model
        self.assertFalse(event == event3)
        event.can_model = event3.can_model
        event.active = not event3.active
        self.assertFalse(event == event3)
        event.active = event3.active
        event.time = timestamp2
        self.assertFalse(event == event3)

    def test2_switch_event_table(self):
        """ Test accessing SwitchEvents in the database
        """

        # Create a switch events table object
        table = SwitchEventsTable(self.db_connection)

        # Verify the table and sequence names are the expected values
        self.assertEqual(table.table_name, DatabaseObjects.get_name(
            "mode_switch_events_table"))
        self.assertEqual(table.switch_id_sequence, DatabaseObjects.get_name(
            "switch_event_id_sequence"))

        # Verify the sequence is incrementing by one
        id1 = table.get_next_event_id()
        id2 = table.get_next_event_id()
        self.assertEqual(id1, id2 - 1)

        # Create test events
        timestamp1 = Timestamp(2014, 33, 1234)
        timestamp2 = Timestamp(2015, 17, 86399)
        event1 = SwitchEvent(5, timestamp1, True, False)
        event2 = SwitchEvent(7, timestamp2, False, True)

        # Insert both events into the table (in reverse order to confirm the
        # following query returns them in time order)
        table.insert_event(event2)
        table.insert_event(event1)
        self.db_connection.commit()

        # Get all the events
        events = table.get_all_events()
        self.assertEqual(len(events), 2)
        self.assertEqual(events[0], event1)
        self.assertEqual(events[1], event2)

        # Get a specific event
        event = table.get_event(5)
        self.assertEqual(event, event1)
        event = table.get_event(7)
        self.assertEqual(event, event2)

        # Add a second valid timestamp for the start/stop time testing
        timestamp3 = Timestamp(2014, 44, 7890)
        event3 = SwitchEvent(8, timestamp3, True, True)
        table.insert_event(event3)
        self.db_connection.commit()

        # Test two events existing in the table
        times = table.get_start_stop_time(5)
        self.assertEqual(len(times), 3)
        self.assertEqual(times[0], timestamp1)
        self.assertEqual(times[1], timestamp3)
        # Verify the complete_event flag is set
        self.assertEqual(times[2], True)

        # Our ending timestamp should be the starting time + the delta
        stoptime = timestamp2.get_datetime() + timedelta(
            days=ModelFitParameters.current_switch_event_days)
        new_stoptime = Timestamp.from_datetime(stoptime)
        times = table.get_start_stop_time(7)
        self.assertEqual(len(times), 3)
        self.assertEqual(times[0], timestamp2)
        self.assertTrue(times[1] == new_stoptime)
        # Verify the complete_event flag is not set (ending time came from
        # start time + the delta)
        self.assertEqual(times[2], False)

        # Make sure get_start_stop_time can grab timestamps if their can model
        # or active values are false
        timestamp4 = Timestamp(2014, 46, 4456)
        event4 = SwitchEvent(9, timestamp4, False, True)
        timestamp5 = Timestamp(2014, 47, 6000)
        event5 = SwitchEvent(10, timestamp5, True, False)
        table.insert_event(event4)
        table.insert_event(event5)

        times = table.get_start_stop_time(8)
        self.assertEqual(len(times), 3)
        self.assertEqual(times[0], timestamp3)
        self.assertEqual(times[1], timestamp4)
        # Verify the complete_event flag is set
        self.assertEqual(times[2], True)

        # Update the time of an event
        new_timestamp = Timestamp(2014, 34, 1234)
        table.update_event_time(5, new_timestamp)
        self.db_connection.commit()

        new_event1 = SwitchEvent(5, new_timestamp, True, False)
        event = table.get_event(5)
        self.assertEqual(event, new_event1)

        # Update the can model flag for an event
        table.update_event_can_model_flag(5, not event1.can_model)
        self.db_connection.commit()

        # Update the active flag for an event
        table.update_event_active_flag(5, not event1.active)
        self.db_connection.commit()

        event = table.get_event(5)
        new_event1.can_model = not new_event1.can_model
        new_event1.active = not new_event1.active
        self.assertEqual(event, new_event1)

        # Now test for exceptions that should be raised.
        # Test non-existing event IDs
        with self.assertRaises(EventNotFoundError):
            event = table.get_event(20)
        with self.assertRaises(EventNotFoundError):
            table.update_event_time(20, timestamp1)
        with self.assertRaises(EventNotFoundError):
            table.update_event_can_model_flag(20, True)
        with self.assertRaises(EventNotFoundError):
            table.update_event_active_flag(20, True)

    def test3_filter_events(self):
        """ Test filtering SwitchEvents
        """
        # Clear out previous data for testing of filtering
        self.test_tables.cleanup_database(False)
        self.test_tables.setup_database()

        expected_events =\
            [
               SwitchEvent.from_list([1, 2013, 1, 5, False, False]),
               SwitchEvent.from_list([2, 2013, 2, 6, False, True]),
               SwitchEvent.from_list([3, 2013, 2, 6007, True, False]),
               SwitchEvent.from_list([4, 2013, 3, 7, True, True]),
               SwitchEvent.from_list([5, 2013, 4, 8, True, True])
            ]
        # Create a switch events table object
        table = SwitchEventsTable(self.db_connection)

        # Insert the expected events
        for event in expected_events:
            table.insert_event(event)
        self.db_connection.commit()

        # Test getting all the events
        events = table.get_filtered_events()
        for event, expected_event in zip(events, expected_events):
            self.assertEqual(event, expected_event)

        # Get the events in reverse order
        events = table.get_filtered_events(reverse=True)
        reversed_events = expected_events[:]
        reversed_events.reverse()
        for event, expected_event in zip(events, reversed_events):
            self.assertEqual(event, expected_event)

        # Get only events that can be modeled
        events = table.get_filtered_events(can_model=True)
        filtered_events = [x for x in expected_events if x.can_model]
        for event, expected_event in zip(events, filtered_events):
            self.assertEqual(event, expected_event)

        # Get only events that can't be modeled
        events = table.get_filtered_events(can_model=False)
        filtered_events = [x for x in expected_events if not x.can_model]
        for event, expected_event in zip(events, filtered_events):
            self.assertEqual(event, expected_event)

        # Get only events that are active
        events = table.get_filtered_events(active=True)
        filtered_events = [x for x in expected_events if x.active]
        for event, expected_event in zip(events, filtered_events):
            self.assertEqual(event, expected_event)

        # Get only events that are inactive
        events = table.get_filtered_events(active=False)
        filtered_events = [x for x in expected_events if not x.active]
        for event, expected_event in zip(events, filtered_events):
            self.assertEqual(event, expected_event)

        # Get only events that start at the third event's time or later
        events = table.get_filtered_events(
            start_timestamp=expected_events[2].time)
        expected_ids = [3, 4, 5]
        for event, expected_id in zip(events, expected_ids):
            self.assertEqual(event.event_id, expected_id)

        # Get only events that start at the third event's time or before
        events = table.get_filtered_events(
            end_timestamp=expected_events[2].time)
        expected_ids = [1, 2, 3]
        for event, expected_id in zip(events, expected_ids):
            self.assertEqual(event.event_id, expected_id)

        # Get only events that start at the second event's time or after
        # and before or the same as the fourth event
        events = table.get_filtered_events(
            start_timestamp=expected_events[1].time,
            end_timestamp=expected_events[3].time)
        expected_ids = [2, 3, 4]
        for event, expected_id in zip(events, expected_ids):
            self.assertEqual(event.event_id, expected_id)


if __name__ == '__main__':
    print("-" * 70)
    print("Testing switch event...\n")

    # Set up the satellite attributes for Landsat 8
    sat_attr.set_default(sat_attr.get_satellite_id_from_number(8))

    unittest.main(verbosity=2)
