#!/usr/bin/env python3

import unittest
from datetime import datetime
from datetime import timedelta
from timestamp import Timestamp
from ssm_exceptions import TimestampError


class TestTimestamp(unittest.TestCase):
    """ Tests for the Timestamp class """

    def test_construction(self):
        timestamp = Timestamp(2013, 15, 34001)

        self.assertEqual(timestamp.year, 2013)
        self.assertEqual(timestamp.doy, 15)
        self.assertEqual(timestamp.seconds, 34001)

    def test_operators(self):
        timestamp1 = Timestamp(2013, 15, 34001)
        timestamp2 = Timestamp(2014, 66, 1133)
        timestamp1_dup = Timestamp(2013, 15, 34001)

        # Equal operator
        self.assertFalse(timestamp1 == timestamp2)
        self.assertTrue(timestamp1 == timestamp1_dup)

        # Not equal operator
        self.assertTrue(timestamp1 != timestamp2)
        self.assertFalse(timestamp1 != timestamp1_dup)

        # < operator
        self.assertTrue(timestamp1 < timestamp2)
        self.assertFalse(timestamp2 < timestamp1)
        self.assertFalse(timestamp1 < timestamp1_dup)

        # <= operator
        self.assertTrue(timestamp1 <= timestamp2)
        self.assertFalse(timestamp2 <= timestamp1)
        self.assertTrue(timestamp1 <= timestamp1_dup)

        # > operator
        self.assertFalse(timestamp1 > timestamp2)
        self.assertTrue(timestamp2 > timestamp1)
        self.assertFalse(timestamp1 > timestamp1_dup)

        # >= operator
        self.assertFalse(timestamp1 >= timestamp2)
        self.assertTrue(timestamp2 >= timestamp1)
        self.assertTrue(timestamp1 >= timestamp1_dup)

    def test_subtraction(self):
        timestamp1 = Timestamp(2013, 15, 100)
        timestamp2 = Timestamp(2013, 16, 101)

        diff = timestamp2 - timestamp1
        self.assertEqual(diff.total_seconds(), 86401)

        diff = timestamp1 - timestamp2
        self.assertEqual(diff.total_seconds(), -86401)

    def test_creation_from_string(self):
        timestamp1 = Timestamp.from_string("2010:17:7899")
        timestamp2 = Timestamp(2010, 17, 7899)
        self.assertEqual(timestamp1, timestamp2)

        # 7899 is 2:11:39 in HH:MM:SS
        timestamp3 = Timestamp.from_string_dual_format("2010-17-02:11:39")
        self.assertEqual(timestamp1, timestamp3)

        # Test with YYYY-DOY-HH:MM:SS.S
        timestamp4 = Timestamp.from_string_dual_format("2010-17-02:11:39.459")
        self.assertEqual(timestamp1, timestamp4)

        # Make sure invalid strings produce exceptions
        with self.assertRaises(TimestampError):
            timestamp1 = Timestamp.from_string("2010:17:1:2")
        with self.assertRaises(TimestampError):
            timestamp1 = Timestamp.from_string("2010:17")
        with self.assertRaises(TimestampError):
            timestamp1 = Timestamp.from_string("2010")
        with self.assertRaises(TimestampError):
            timestamp1 = Timestamp.from_string("1955:1:1")
        with self.assertRaises(TimestampError):
            timestamp1 = Timestamp.from_string("2012:-1:1")
        with self.assertRaises(TimestampError):
            timestamp1 = Timestamp.from_string("2012:367:1")
        with self.assertRaises(TimestampError):
            timestamp1 = Timestamp.from_string("2012:1:-1")
        with self.assertRaises(TimestampError):
            timestamp1 = Timestamp.from_string("2012:1:1000000")
        with self.assertRaises(TimestampError):
            timestamp1 = Timestamp.from_string("2012:1:1.1")
        with self.assertRaises(TimestampError):
            timestamp1 = Timestamp.from_string("2012:1.1:1")
        with self.assertRaises(TimestampError):
            timestamp1 = Timestamp.from_string("2012.1:1:1")

        with self.assertRaises(TimestampError):
            timestamp1 = Timestamp.from_string_dual_format("2012-1-4:8:-1")
        with self.assertRaises(TimestampError):
            timestamp1 = Timestamp.from_string_dual_format("2012-1-4:8:86400")
        with self.assertRaises(TimestampError):
            timestamp1 = Timestamp.from_string_dual_format("2012-1-1.1")
        with self.assertRaises(TimestampError):
            timestamp1 = Timestamp.from_string_dual_format("2012-1.1:1")
        with self.assertRaises(TimestampError):
            timestamp1 = Timestamp.from_string_dual_format("2012-1-1-1")
        with self.assertRaises(TimestampError):
            timestamp1 = Timestamp.from_string_dual_format("2012-1-25:1:39")
        with self.assertRaises(TimestampError):
            timestamp1 = Timestamp.from_string_dual_format("2012-1-23:61:39")
        with self.assertRaises(TimestampError):
            timestamp1 = Timestamp.from_string_dual_format("2012-1-23:59:139")

    def test_creation_from_list(self):
        timestamp1 = Timestamp.from_list([2014, 17, 8088])
        self.assertEqual(timestamp1.year, 2014)
        self.assertEqual(timestamp1.doy, 17)
        self.assertEqual(timestamp1.seconds, 8088)

    def test_creation_from_datetime(self):
        test_datetime = datetime(2015, 2, 1)
        timestamp1 = Timestamp.from_datetime(test_datetime)
        expected_timestamp = Timestamp(2015, 32, 0)
        self.assertEqual(timestamp1, expected_timestamp)

    def test_leap_year(self):
        # On a leap year, there are 29 days in February
        timestamp1 = Timestamp.from_datetime(datetime(2016, 2, 1))
        timestamp2 = Timestamp.from_datetime(datetime(2016, 3, 1))
        diff = timestamp2 - timestamp1
        self.assertEqual(diff.total_seconds(), 29 * 86400)

        # On non-leap years, there are 28 days in February
        timestamp1 = Timestamp.from_datetime(datetime(2015, 2, 1))
        timestamp2 = Timestamp.from_datetime(datetime(2015, 3, 1))
        diff = timestamp2 - timestamp1
        self.assertEqual(diff.total_seconds(), 28 * 86400)


if __name__ == '__main__':
    print("-" * 70)
    print("Testing timestamp...\n")
    unittest.main(verbosity=2)
