#!/usr/bin/env python3

import datetime
from ssm_exceptions import TimestampError


class Timestamp():
    """ Encapsulates a year, day of year, seconds of day timestamp.

        Notes:
            - This class does not support leap seconds
    """

    def __init__(self, year, doy, seconds):
        """ Construct a new Timestamp object.

            Parameters:
                year: The year of the timestamp (int)
                doy: The day of year (int)
                seconds: The seconds of the day
        """
        self._mydatetime = None
        self.year = year
        self.doy = doy
        self.seconds = seconds

    def __sub__(self, other):
        """ Substract two timestamps and return a timedelta object """
        return self.get_datetime() - other.get_datetime()

    def __str__(self):
        """ Return the timestamp as a string """
        return "{0}:{1}:{2}".format(self.year, self.doy, self.seconds)

    def __repr__(self):
        """ Return the representation of the timestamp """
        return "Timestamp({0}, {1}, {2})".format(
            self.year, self.doy, self.seconds)

    def __eq__(self, other):
        """ Equals operator for timestamps """
        return self.get_datetime() == other.get_datetime()

    def __ne__(self, other):
        """ Not equal operator for timestamps """
        return self.get_datetime() != other.get_datetime()

    def __lt__(self, other):
        """ Less than operator for timestamps """
        return self.get_datetime() < other.get_datetime()

    def __le__(self, other):
        """ Less than or equal operator for timestamps """
        return self.get_datetime() <= other.get_datetime()

    def __gt__(self, other):
        """ Greater than operator for timestamps """
        return self.get_datetime() > other.get_datetime()

    def __ge__(self, other):
        """ Greater than or equal operator for timestamps """
        return self.get_datetime() >= other.get_datetime()

    def get_datetime(self):
        """ Returns the timestamp as a datetime object.  The datetime object
            is cached in the _mydatetime member.
        """
        if self._mydatetime is None:
            self._mydatetime = Timestamp.timestamp_to_datetime(
                self.year, self.doy, self.seconds)
        return self._mydatetime

    def to_list(self):
        """ Convert the timestamp to a [year, day of year, seconds of day]
            array
        """
        return [self.year, self.doy, self.seconds]

    @staticmethod
    def timestamp_to_datetime(year, doy, seconds):
        """ Convert year, day of year, seconds of day to a datetime and return
            the datetime.

            Parameters:
                year: The year of the timestamp (int)
                doy: The day of year (int)
                seconds: The seconds of the day
        """
        m, s = divmod(seconds, 60)
        h, m = divmod(m, 60)
        return datetime.datetime.strptime(
            "{0} {1} {2} {3} {4}".format(year, doy, h, m, s),
            '%Y %j %H %M %S')

    @classmethod
    def from_list(cls, input_list):
        """ Create a timestamp from a list

            Parameters:
                input_list: timestamp as [year, doy, seconds]
        """
        return cls(input_list[0], input_list[1], input_list[2])

    @classmethod
    def from_datetime(cls, datetime):
        """ Create a timestamp from a datetime object """
        time_struct = datetime.timetuple()
        return cls(time_struct.tm_year, time_struct.tm_yday,
                   time_struct.tm_hour * 3600 + time_struct.tm_min * 60 +
                   time_struct.tm_sec)

    @classmethod
    def from_string(cls, timestamp_string):
        """ Create a timestamp from a string in the form YYYY:DOY:SECONDS.
            Some error checking is done on the values in the string.
        """

        # Make sure there are three fields separated by ':'
        fields = timestamp_string.split(':')
        if len(fields) != 3:
            raise TimestampError(
                "Timestamp '{0}' not in YYYY:DOY:SECONDS format".format(
                    timestamp_string))

        # Make sure the fields are integers
        try:
            year, doy, seconds = map(int, fields)
        except:
            raise TimestampError(
                "Timestamp {0} not in YYYY:DOY:SECONDS format".format(
                    timestamp_string))

        # Make sure the values are reasonable.
        # Note these values do not allow for leap seconds, as they are not
        # supported by the datetime classes
        if (year < 1970 or year > 9999 or doy <= 0 or doy > 366 or
                seconds < 0 or seconds >= 86400):
            raise TimestampError("Invalid range in timestamp {0}".format(
                timestamp_string))

        # Create and return a new timestamp
        return cls(year, doy, seconds)

    @classmethod
    def from_string_dual_format(cls, timestamp_string):
        """ Create a timestamp from a string in the form YYYY-DOY-HH:MM:SS[.S]
            or YYYY:DOY:SECONDS.
            Some error checking is done on the values in the string.
        """

        # Is the timestamp string in the YYYY-DOY-HH:MM:SS[.S] format?
        if '-' in timestamp_string:
            fields = timestamp_string.split('-')
            if len(fields) != 3:
                raise TimestampError(
                    "Timestamp '{0}' not in YYYY-DOY-HH:MM:SS[.S] "
                    "format".format(timestamp_string))
            # Year and DOY is correct, but need to convert HH:MM:SS to
            # SOD. Split them
            temp_time = fields[2].split(':')
            if len(temp_time) != 3:
                raise TimestampError(
                    "HH:MM:SS '{0}' not in proper format".format(
                        timestamp_string))
            # Convert to seconds to float first to deal with any values
            # containing decimal points
            temp_time[2] = float(temp_time[2])
            # Now convert floats to integers for all
            hrs, mins, secs = map(int, temp_time)
            # Make sure the values are reasonable, allowing for the unlikely
            # leap second.
            if (hrs < 0 or hrs > 23 or mins < 0 or mins > 59 or
                    secs < 0 or secs > 60):
                raise TimestampError("Invalid range for HH:MM:SS[.S]: {0}".
                                     format(timestamp_string))
            # Create a time object
            time_obj = datetime.timedelta(
                hours=hrs, minutes=mins, seconds=secs)
            # Convert to total seconds
            total_seconds = time_obj.total_seconds()
            # Store seconds in original fields array in the appropriate
            # location
            fields[2] = int(total_seconds)
            # Put into the YYYY:DOY:SOD format which is what the original
            # from string expects
            new_timestamp_string = ':'.join(map(str, fields))
        else:
            new_timestamp_string = timestamp_string

        # Call the original from_string
        return cls.from_string(new_timestamp_string)
