
class ModelFitParameters():
    # The default TIRS SSM model parameters to use. The 8 values represent
    # (a0, a1, a2, a3, tau1, tau2, tau3, slope) as described in the ADD.
    # Units: (count, count, count, count, seconds, days, days, counts per day)
    default_model_parameters = (1645.2, 293.1, 286.8, 100.1, 1272.7, 0.197,
                                1.949, 3.508)

    # Allow the user to suppress selected model parameters in the model fit by
    # setting the corresponding entry in this array to zero.  The order is the
    # same as the model parameters (a0, a1, a2, a3, tau1, tau2, tau3, slope).
    coefficient_masks = (1, 1, 1, 1, 0, 0, 0, 1)

    # The default nadir encoder position.  This value is for the TIRS side B
    # electronics.  If the sensor is switched to side A, it will need to be
    # updated. Units: counts
    default_encoder_nadir_position = 16574079

    # The algorithm version number for the default model
    default_algorithm_version = 0

    # Maximum SSM Calibration algorithm RMSE to include in the fit.
    # Units: micro-radians
    calibration_outlier_rmse_threshold = 35.0

    # Standard deviation of the encoder counts to use to weight the telemetry
    # samples. The application converts this to a weight by calculating
    # 1.0/count^2.  Units: counts
    telemetry_observation_stdev = 10

    # Standard deviation of the encoder counts to use to weight the TIRS SSM
    # Calibration algorithm results derived from the imagery.  The application
    # converts this to a weight by calculating 1.0/count^2.  Units: counts
    image_observation_stdev = 20

    # These are the apriori weights that are added to the least squares
    # solution when no points are available within the break times / model
    # observation gap windows.
    # Units: (count, count, count, count, seconds, days, days, counts per day)
    apriori_weights = (120.0, 30.0, 20.0, 15.0, 150.0, 0.025, 0.3, 0.6)

    # Times which define the number of SSM parameters to be solved for.
    # Units: (seconds, days, days, days)
    model_break_points = (2000.0, 1.0, 7.0, 12.0)

    # Time ranges for which a lack of observations make the solution for a set
    # of SSM model parameters ill defined.
    # Units: (seconds, seconds, seconds, seconds, days, days)
    observation_gap_windows = (2000.0, 6000.0, 6000.0, 18000.0, 1.0, 7.0)

    # Weights applied to coefficients during least squares solution to
    # normalize the parameters such that one single value can be used to test
    # for convergence of the solution.  Unitless
    convergence_threshold_weights = (0.05, 0.05, 0.05, 0.05, 0.05, 0.0005,
                                     0.0005, 0.0005)

    # Defines the position sampling interval when generating estimated SSM
    # positions.  In each row, the first value is the starting time for the
    # interval and the second value is the time period between samples.
    # For example, the (600.0, 120.0) indicates that starting at 600 seconds
    # into the switch event period the time between samples should be 120
    # seconds - sample at 600, 720, 840, 960, 1080, 1200.  At 1200, the next
    # record defines the next block's sampling period.
    # Units: (seconds, delta seconds)
    position_sampling_interval_table = (
            (60.0, 60.0),
            (600.0, 120.0),
            (1200.0, 180.0),
            (2100.0, 300.0),
            (3600.0, 600.0),
            (8400.0, 1200.0),
            (14400.0, 2400.0),
            (21600.0, 3600.0),
            (43200.0, 7200.0),
            (86400.0, 10800.0),
            (129600.0, 14400.0),
            (216000.0, 21600.0),
            (604800.0, 43200.0),
            (691200.0, 86400.0)
        )

    # Maximum number of attempts at converging to a solution during least
    # squares fitting process.
    fit_maximum_iterations = 50000

    # For the current switch event that does not have a switch event to
    # indicate when it ends, make the end switch event this many days after
    # the start of the switch event.
    current_switch_event_days = 21

    # The following parameters control the filtering of TIRS SSM telemetry
    # records in the fit application.
    #   telemetry_start_offset: delete the telemetry records that start
    #       less than this number of seconds after the switch event start time.
    #   telemetry_subsample_threshold: number of TIRS telemetry records to
    #       allow before applying subsampling to the records.
    #   telemetry_subsample_step: when the number of telemetry records exceeds
    #       the telemetry_subsample_threshold parameter, sub-sample them by
    #       keeping the records this number of seconds apart.
    telemetry_start_offset = 60
    telemetry_subsample_threshold = 100
    telemetry_subsample_step = 20
