#!/bin/csh -f
# This script dumps the TIRS_SSM_MODE_SWITCH_EVENTS, TIRS_SSM_POS_MODEL_PARAMS,
# and TIRS_SSM_POS_ESTIMATES from Ops. It then loads them into the
# tables of the user set in IAS_DB_L8_OLITIRS. Ops connection info may either be
# defined in an ENV var named OPS_SSM_CONNECTION_INFO or entered at run time.
# The ORACLE_SID and IAS_DB_L8_OLITIRS ENV vars should be set to the intended
# destination when starting this script.
#
# Arguments =   --quiet         Suppress all output (including sqlldr)
#               --keep-dumps    Leave the dump files in the current directory,
#                               rather than deleting them
#               --no-replace    Do not clear existing table rows before loading
#                               the exported ones

set quiet_flag = 0
set keep_dumps = 0
set no_replace = 0
foreach cla ($*)
    switch ( $cla )
        case "--quiet":
            set quiet_flag = 1
            breaksw
        case "--keep-dumps"
            set keep_dumps = 1
            breaksw
        case "--no-replace"
            set no_replace = 1
            breaksw
        default:
            echo "Unknown argument $cla"
            exit 1
            breaksw
    endsw
end

# Save the original SID and connection info
set ORIG_ORACLE_SID = $ORACLE_SID
set ORIG_LOGIN = $IAS_DB_L8_OLITIRS

# Change the Ops ORACLE_SID
setenv ORACLE_SID CROPS

# See if Ops connection info is in the ENV
if ( $?OPS_SSM_CONNECTION_INFO ) then
    setenv IAS_DB_L8_OLITIRS $OPS_SSM_CONNECTION_INFO
else
    echo "Enter the Ops connection info"
    setenv IAS_DB_L8_OLITIRS $<
endif

# Make a temp sub-dir based on the current Unix timestamp
set TEMP_EXP_DIR = `date +%s`
set TEMP_EXP_DIR = "/tmp/SSM_EXPORT_$TEMP_EXP_DIR"

# Record orginal working dir
set ORIG_CWD = `pwd`

if ($keep_dumps != 1) then
    # Make and change working dir to the temp location
    mkdir $TEMP_EXP_DIR
    cd $TEMP_EXP_DIR
endif

if ($? != 0) then
    echo "Failed to change to temp dir"
    exit 1
endif

create_sqlldr_files.py \
GEOCHAR.TIRS_SSM_MODE_SWITCH_EVENTS \
GEOCHAR.TIRS_SSM_POS_MODEL_PARAMS GEOCHAR.TIRS_SSM_POS_ESTIMATES
if ($? != 0) then
    echo "Call to create_sqlldr_files.py failed"
    exit 1
endif

if ($quiet_flag != 1) then
    echo Successfully dumped SSM tables
endif

# Restore the original SID and connection info
setenv ORACLE_SID $ORIG_ORACLE_SID
setenv IAS_DB_L8_OLITIRS $ORIG_LOGIN

if ($no_replace != 1) then
    cat > clean_tables.sql <<EndSQL
DELETE FROM TIRS_SSM_POS_ESTIMATES;
DELETE FROM TIRS_SSM_POS_MODEL_PARAMS;
DELETE FROM TIRS_SSM_MODE_SWITCH_EVENTS;
COMMIT;
exit
EndSQL
    sqlplus $IAS_DB_L8_OLITIRS @clean_tables > clean_output.txt
    if ($? != 0) then
        echo "Failed to clear existing tables"
        exit 1
    endif
endif

# Load each table
if ($quiet_flag != 1) then
    sqlldr $IAS_DB_L8_OLITIRS control=TIRS_SSM_MODE_SWITCH_EVENTS.ctl
else
    sqlldr $IAS_DB_L8_OLITIRS control=TIRS_SSM_MODE_SWITCH_EVENTS.ctl silent=ALL
endif
if ($? != 0) then
    echo "Load of TIRS_SSM_MODE_SWITCH_EVENTS failed"
    exit 1
endif
if ($quiet_flag != 1) then
    sqlldr $IAS_DB_L8_OLITIRS control=TIRS_SSM_POS_MODEL_PARAMS.ctl
else
    sqlldr $IAS_DB_L8_OLITIRS control=TIRS_SSM_POS_MODEL_PARAMS.ctl silent=ALL
endif
if ($? != 0) then
    echo "Load of TIRS_SSM_POS_MODEL_PARAMS failed"
    exit 1
endif
if ($quiet_flag != 1) then
    sqlldr $IAS_DB_L8_OLITIRS control=TIRS_SSM_POS_ESTIMATES.ctl
else
    sqlldr $IAS_DB_L8_OLITIRS control=TIRS_SSM_POS_ESTIMATES.ctl silent=ALL
endif
if ($? != 0) then
    echo "Load of TIRS_SSM_POS_ESTIMATES failed"
    exit 1
endif

#Change back to the orginal working dir
cd $ORIG_CWD

# Reset our sequence to match the Ops data just loaded
./reset_db_sequence.py TIRS_SSM_MODE_SWITCH_EVENTS SWITCH_EVENT_ID \
DBSEQ_TIRS_SSM_SWITCH_ID_PK
if ($? != 0) then
    echo "Reset of switch event sequence failed"
    exit 1
endif
./reset_db_sequence.py TIRS_SSM_POS_MODEL_PARAMS MODEL_ID \
DBSEQ_TIRS_SSM_POS_MODEL_ID_PK
if ($? != 0) then
    echo "Reset of model sequence failed"
    exit 1
endif

if ($keep_dumps != 1) then
    rm -r $TEMP_EXP_DIR
    if ($? != 0) then
        echo "Failed to cleanup to temp dir"
        exit 1
    endif
endif
