#!/usr/local/bin/python3
""" Resets the value of a sequence to the highest number of a specified
    primary key
"""
from ias.db import database as db
from ias.log import logger
from ias.misc.const import EXIT_SUCCESS, EXIT_FAILURE
from ias.misc import satellite_attributes as sat_attr
import argparse
import numbers


class SequenceError(Exception):
    pass


class ResetDbSequence():
    """ Class to reset sequence values
    """

    def __init__(self, db_in):
        """ Construct a new ResetDbSequence object.

            Parameters:
                db_in: Already created database connection
        """
        self.db = db_in

    def reset(self, table_name, primary_key, sequence_name):
        """ Reset the sequence value

            Parameters:
                table_name: Name of the table with PK
                primary_key: Name of the PK
                sequence_name: Name of the sequence
        """
        # Get the max value of the PK
        sql = "SELECT max({0}) from {1}".format(primary_key, table_name)
        cursor = self.db.get_cursor()
        cursor.execute(sql)
        rows = cursor.fetchall()
        if len(rows) < 1:
            raise SequenceError("Failed to get primary key max value")

        max_pk = rows[0][0]
        if not isinstance(max_pk, numbers.Integral):
            raise SequenceError("primary key max value is not integral")

        # Get the current sequence value
        sql = "SELECT {0}.nextval FROM DUAL".format(sequence_name)
        cursor.execute(sql)
        rows = cursor.fetchall()
        if len(rows) < 1:
            raise SequenceError("Failed to get sequence value")
        current_seq = rows[0][0]

        # Calculate the number of times to increment the sequence generator
        seq_count = max_pk - current_seq

        # Call nextval on the sequence enough times to get it above the largest
        # ID found
        sql = "SELECT {0}.nextval FROM DUAL".format(sequence_name)
        while seq_count > 0:
            seq_count = seq_count - 1
            cursor.execute(sql)

        cursor.close()


def connect_to_database():
    """ Connect to the database using the currently configured satellite/sensor
        schema.  The caller is expected to have properly initialized the
        satellite attributes library prior to calling this routine.
    """
    # Get the satellite attributes
    sa = sat_attr.get_default()

    # Connect to the database sensor schema
    db_connection = db.DatabaseConnection()
    db_connection.set_login_using_satellite_sensor(
        sa.get_satellite_name(), sa.get_satellite_sensor_name())
    db_connection.connect()

    # return the DB connection
    return db_connection


def parse_parameters():
    """ Parses command line arguments for this script

            Parameters:
                arguments: List of args to parse

            Returns:
                Namespace populated with argument values
    """
    argparser = argparse.ArgumentParser(
        description='Reset sequence value to the max PK of a table')

    argparser.add_argument(
        'table', nargs=1,
        help="Table name containg the primary key")

    argparser.add_argument(
        'primary_key', nargs=1,
        help="Primary key to reset sequence to")

    argparser.add_argument(
        'sequence', nargs=1,
        help="Name of sequence to be reset")

    par = argparser.parse_args()

    return par


def main():
    status = EXIT_SUCCESS
    db_conn = None
    try:
        # Initialize the logger
        log = logger.Logger("Reset sequence")

        # Init satellite attributes
        sat_attr.set_default(sat_attr.get_satellite_id_from_number(8))

        parsed_args = parse_parameters()

        # Connect to the database
        db_conn = connect_to_database()

        db_reset = ResetDbSequence(db_conn)
        db_reset.reset(parsed_args.table[0], parsed_args.primary_key[0],
                       parsed_args.sequence[0])
    except SequenceError as ex:
        log.error("Exception running the reset: %s", ex)
        status = EXIT_FAILURE
    finally:
        if db_conn is not None:
            db_conn.close(db.DISCONNECT_ROLLBACK)
        log.close()

    exit(status)

# Run the main routine if called from command line
if __name__ == "__main__":
    main()
