/******************************************************************************
NAME: analyze_time_period

PURPOSE: Parse SSM telemetry information determining whether time brackets are
    filled for sets of coefficients associated with the SSM model.

RETURNS: SUCCESS/ERROR

******************************************************************************/
#include <math.h>                  /* Math library */

#include "ias_const.h"             /* IAS constant library */ 
#include "ias_logging.h"           /* IAS logging library */
#include "tirs_fit_ssm_model.h"    /* Local header */

#define		DEBUG0		0

int analyze_time_period
(
    const ENCODER_RECORD *tirs_data,  /* I: TIRS data records */
    int number_tirs_data,             /* I: Number of TIRS encoder values */
    int *number_samples_per_break     /* O: Number of samples in each 
                                            observation window.  Sized of
                                            SSM_OBSERVATION_GAP_WINDOWS */
)
{
    int time_index;         /* Counter */
    int window_index;       /* Counter */
    double time_difference; /* Time difference between two samples */

    /* Observaton gap window times for defining coefficients in solution */
    double observation_gap_windows[SSM_OBSERVATION_GAP_PARAMS];

    /* Maximum difference between breaks */
    double max_time_diff_per_break[SSM_OBSERVATION_GAP_WINDOWS];  

    get_observation_gap_windows(observation_gap_windows);

    /* Check if the number of TIRS encoder value is zero or not */
    if (number_tirs_data == 0)
    {
        IAS_LOG_ERROR("No TIRS data for fit");
        return ERROR;
    }

    /* Make sure the data times are increasing */
    for (time_index = 0; time_index < number_tirs_data - 1; time_index++)
    {
        time_difference = tirs_data[time_index + 1].secs_diff
            - tirs_data[time_index].secs_diff;
        if (time_difference == 0.0)
            IAS_LOG_INFO("TIRS indexes %d,%d have the same time",
                    time_index, time_index + 1);
        if (time_difference < 0.0)
        {
            IAS_LOG_ERROR("TIRS times are not increasing %d,%d %d,%d", 
                    tirs_data[time_index].secs_diff,
                    tirs_data[time_index+1].secs_diff, time_index,
                    time_index + 1);
            return ERROR;
        }
    }

    /* Initialize returned parameters */
    for (time_index = 0; time_index < SSM_OBSERVATION_GAP_WINDOWS; time_index++)
    {
        number_samples_per_break[time_index] = 0;
        max_time_diff_per_break[time_index] = 0;
    }

    /* Find encoder values that can be found per time break */
    for (time_index = 0; time_index < number_tirs_data; time_index++)
    {
        for (window_index = 0; window_index < SSM_OBSERVATION_GAP_WINDOWS;
             window_index++)
        {
            int gap_index = 2 * window_index;

            /* If the time from the start of the event falls in the current
               gap, count it */
            if (tirs_data[time_index].secs_diff
                    > observation_gap_windows[gap_index]
                && tirs_data[time_index].secs_diff
                    <= observation_gap_windows[gap_index + 1])
            {
                number_samples_per_break[window_index]++;

                if (DEBUG0)
                {
                    printf("time_index %d window_index %d bounds %f "
                            "%f number %d\n", time_index, window_index,
                            observation_gap_windows[gap_index], 
                            observation_gap_windows[gap_index+1], 
                            number_samples_per_break[window_index]);
                }

                /* Keep track of the largest time difference between two
                   consecutive samples in each window */
                if (time_index < number_tirs_data - 1)
                {
                    time_difference = tirs_data[time_index + 1].secs_diff
                        - tirs_data[time_index].secs_diff;

                    if (time_difference > max_time_diff_per_break[window_index])
                        max_time_diff_per_break[window_index] = time_difference;
                }

                /* No need to continue checking this time since the gaps do
                   not overlap */
                break;
            }
        }
    }

    /* Issue a log message for each gap indicating the samples found in the
       window and the largest time between two consecutive samples in the
       window */
    for (window_index = 0; window_index < SSM_OBSERVATION_GAP_WINDOWS; 
         window_index++)
    {
        int gap_index = 2 * window_index;

        IAS_LOG_INFO("Checking encoder fit values for time range %f "
                "%f %d max_time_delta %f",observation_gap_windows[gap_index],
                observation_gap_windows[gap_index+1],
                number_samples_per_break[window_index],
                max_time_diff_per_break[window_index]);
    }

    return SUCCESS;
}
