/******************************************************************************
Name: create_report.c

Purpose: create_report generates a report of the results of the SSM encoder
    model fit calculations.  The report is written to a file.

ALGORITHM REFERENCES:
TIRS SSM Model Fit ADD
******************************************************************************/
#include <stdlib.h>
#include <string.h>               /* String library */
#include <stdio.h>                /* fprintf prototype */

#include "ias_miscellaneous.h"
#include "ias_logging.h"          /* IAS Logging library */
#include "tirs_fit_ssm_model.h"   /* Prototype for this routine */

#define ALGORITHM_VERSION 1       /* TIRS SSM fit model algorithm version */

/******************************************************************************
Name: write_header

Purpose: Routine to write the header to the report.  It is based on the
    miscellaneous library report header routines, but they all complain about
    not having L0R or L1G files available, so this local cut down version
    just deals with the basics.

Returns: SUCCESS/ERROR
******************************************************************************/
static int write_header
(
    FILE *report_file,          /* I: Summary report file */
    const char *algorithm_name  /* I: Algorithm name for the report title */
)
{
    char date_string[40];       /* Date string */
    char time_string[40];       /* Time string */
    const char *processing_center; /* Processing center */
    time_t current_time;        /* Time variable */
    struct tm *timeptr = NULL;  /* Time data structure */
    int num_spaces;
    char spaces[] = "                                 ";
    int status;

    /* Get the current date and put it in the correct structured format
       (i.e. Wed. Dec 10, 2007) */
    current_time = time(NULL);
    timeptr = localtime(&current_time);

    if (!strftime(date_string, sizeof(date_string), "%a. %b. %e, %Y", timeptr))
    {
        IAS_LOG_ERROR("Extracting the current date");
        return ERROR;
    }

    /* Get the current time and put it in the correct structured format
       (i.e. 17:24) */
    if (!strftime(time_string, sizeof(time_string), "%R", timeptr))
    {
        IAS_LOG_ERROR("Extracting the current time");
        return ERROR;
    }

    /* Get the processing center */
    processing_center = getenv("PROCESSING_CENTER");
    if (processing_center == NULL)
        processing_center = "Unknown";

    /* Write out the date, time, and application name */
    num_spaces = (64 - strlen(algorithm_name)) / 2;
    if (num_spaces < 0)
        num_spaces = 0;
    spaces[num_spaces] = '\0';

    status = fprintf(report_file, 
            "%s          Processed at %s          Time: %s\n"
            "%s%s Report\n"
            "-----------------------------------------------------------------"
            "----------\n", date_string, processing_center, time_string,
            spaces, algorithm_name);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing title block to the report file");
        return ERROR;
    }

    /* Write the IAS software version */
    status = fprintf(report_file, "IAS Version:       %s\n",
        ias_misc_get_software_version());
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing IAS software version to the output file");
        return ERROR;
    }

    return SUCCESS;
}

/******************************************************************************
Name: create_report.c

Purpose: create_report generates a report of the results of the SSM encoder
    model fit calculations.  The report is written to a file.

Returns: SUCCESS/ERROR

******************************************************************************/
int create_report
(
    const ENCODER_RECORD *tirs_data, /* I: TIRS data records */
    int number_tirs_data,          /* I: Number of TIRS encoder values */
    double encoder_value0,         /* I: Initial encoder value */
    int switch_event_id,           /* I: Switch event ID */
    int model_id,                  /* I: Model ID */
    const double *new_coefficients,/* I: Array of SSM_FIT_NPARMS new model
                                         coefficients */
    const int *final_num,          /* I: Array of RMSE_COUNT point counts */
    const double *final_rms,       /* I: Array of RMSE_COUNT RMS errors */
    const double *event_start_time,/* I: Event Start time */
    const double *event_end_time   /* I: Event end time */
)
{
    FILE *report_file;                 /* Pointer to the report file */
    int index;                         /* loop index */
    int status;                        /* Status returned by called routines */
    double prev_coefficients[SSM_FIT_NPARMS];/* Initial model coefficients */

    const char *report_file_name = get_report_filename(); /* Report file name */

    get_initial_values(prev_coefficients);

    /* Open the report file for writing */
    report_file = fopen(report_file_name, "w");
    if (report_file == NULL)
    {
        IAS_LOG_ERROR("Opening the report file - %s", report_file_name);
        return ERROR;
    }

    /* Write the report header */
    if (write_header(report_file, "TIRS SSM MODEL FIT") != SUCCESS)
    {
        IAS_LOG_ERROR("Writing report header to report file %s",
                report_file_name);
        fclose(report_file);
        return ERROR;
    }

    status = fprintf(report_file, "Algorithm Version: %d\n", ALGORITHM_VERSION);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    status = fprintf(report_file, "\nSwitch Event ID: %d\n", switch_event_id);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    status = fprintf(report_file, "Previous Model ID: %d\n", model_id);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    /* Write switch start event time */
    status = fprintf(report_file, 
            "\nTIRS SSM Model Switch Event period (Year, DOY, SOD):\n"
            "\tStart: %04d:%03d:%05d\n"
            "\tEnd: %04d:%03d:%05d\n", 
            (int) event_start_time[0], (int) event_start_time[1],
            (int) event_start_time[2],
            (int) event_end_time[0], (int) event_end_time[1],
            (int) event_end_time[2]);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    /* Write the residual statistics information to the report file */
    status = fprintf(report_file,"\n"
            "           a0       a1       a2        a3       tau1        "
            "tau2    tau3     S     Nadir\n");
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    /* Write the previous and new model parameters */
    status = fprintf(report_file, 
            "Previous %.3f  %.3f  %.3f   %.3f   %.3f   %.3f   %.3f   %.3f"
            " %d\n",
            prev_coefficients[SSM_FIT_A0],
            prev_coefficients[SSM_FIT_A1],
            prev_coefficients[SSM_FIT_A2],
            prev_coefficients[SSM_FIT_A3],
            prev_coefficients[SSM_FIT_TAU1],
            prev_coefficients[SSM_FIT_TAU2],
            prev_coefficients[SSM_FIT_TAU3],
            prev_coefficients[SSM_FIT_SLOPE],
            (int)encoder_value0);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    status = fprintf(report_file,
            "New      %.3f  %.3f  %.3f   %.3f   %.3f   %.3f   %.3f   %.3f"
            " %d\n",
            new_coefficients[SSM_FIT_A0],
            new_coefficients[SSM_FIT_A1],
            new_coefficients[SSM_FIT_A2],
            new_coefficients[SSM_FIT_A3],
            new_coefficients[SSM_FIT_TAU1],
            new_coefficients[SSM_FIT_TAU2],
            new_coefficients[SSM_FIT_TAU3],
            new_coefficients[SSM_FIT_SLOPE],
            (int)encoder_value0);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    /* Write the telemetry residual statistics information to the report file */
    status = fprintf(report_file,
            "\nTelemetry RMSE and number of points\t\t%f %d\n", 
            final_rms[TELEMETRY_RMSE], final_num[TELEMETRY_RMSE]);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    /* Write the image residual statistics information to the report file */
    status = fprintf(report_file,
            "Image RMSE and number of points\t\t\t%f %d\n",
            final_rms[IMAGE_RMSE], final_num[IMAGE_RMSE]);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    /* Write the combined residual statistics information to the report file */
    status = fprintf(report_file,
            "Combined RMSE and number of points\t\t%f %d\n",
            final_rms[COMBINED_RMSE], final_num[COMBINED_RMSE]);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    /* Write the residual statistics information to the report file for each
       record used in the solution. */
    status = fprintf(report_file,"\n%19s %9s %5s %8s %11s %8s\n", 
            "Date", "Seconds", "Type", "Measured", "Model", 
            "Residual");
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    for (index = 0; index < number_tirs_data; index++)
    {
        const char *type;      /* Record type */
        double model_profile;  /* Model fit */
        double residual;       /* Difference between model and measured */ 

        /* Calculate the estimated position from the model for the record's
           time */
        model_profile = encoder_model(new_coefficients, encoder_value0, 
                tirs_data[index].secs_diff);

        /* Calculate the residual */
        residual = model_profile - tirs_data[index].encoder_count;

        /* Determine the value to output for the SSM record type */
        if (tirs_data[index].ssm_type == SSM_IMAGE)
            type = "Image";
        else
            type = "Telem";

        if (fprintf(report_file,"%19s %9d %5s %8d %11.2f %8.2f\n", 
                tirs_data[index].date, tirs_data[index].secs_diff, type,
                tirs_data[index].encoder_count, model_profile, residual) < 0)
        {
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            fclose(report_file);
            return ERROR;
        }
    }

    if (fclose(report_file) != 0)
    {
        /* Error returned from fclose */
        IAS_LOG_ERROR("Closing report file %s", report_file_name);
        return ERROR;
    }

    return SUCCESS;
}
