/******************************************************************************
NAME: database_routines.c

PURPOSE: Routines to query the database for input.
            
******************************************************************************/
#include <stdlib.h>
#include <string.h>

#include "ias_logging.h"
#include "ias_const.h"
#include "ias_db.h"
#include "ias_db_query.h"
#include "ias_db_get_connect_info.h"

#include "database_routines.h"

static struct ias_db_connection *db = NULL; /* Database connection */


/******************************************************************************
NAME:   connect_to_database

PURPOSE: Establish a database connection

RETURNS: SUCCESS or ERROR
******************************************************************************/
int connect_to_database(void)
{ 
    char *database_name = NULL;
    char *user_name = NULL;
    char *password = NULL;
    char *host = NULL;
    int status;
    
    status = ias_db_initialize_database_lib();
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the database library");
        return ERROR;
    }
    
    /* Get the database connection information */
    status = ias_db_get_connect_info(&database_name, &user_name, &password,
        &host);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting database connection info");
        return ERROR;
    }
   
    /* Connect to the database */
    db = ias_db_connect_to_database(database_name, user_name, password,
        host);
    
    /* Free the resources before checking the connection */
    free(database_name);
    free(user_name);
    free(password);
    free(host);
    
    /* Check the connection */
    if (!db)
    {
        IAS_LOG_ERROR("Connecting to database");
        ias_db_close_database_lib();
        return ERROR;
    } 
    
    return SUCCESS;   
} /* END connect_to_database */


/******************************************************************************
NAME: disconnect_from_database

PURPOSE: Disconnect from the database

RETURNS: nothing
******************************************************************************/
void disconnect_from_database(void)
{
    if (db == NULL)
        return;
    
    ias_db_close_connection(db);
    db = NULL;
    ias_db_close_database_lib();
} /* END disconnect_from_database */


/******************************************************************************
NAME:       get_db_telemetry_list

PURPOSE: 
Query the TIRS_TELEMETRY_COMMAND and TIRS_TELEMETRY_CIRCUIT ancillary tables
to retrieve records from the switch event time period collected in mode 0
with the encoder on. In this case the query uses the spacecraft J2000 epoch
times.

RETURNS: Pointer to a list of records from database
******************************************************************************/

TELEMETRY_INFORMATION *get_db_telemetry_list
(
    double start_day,         /* I: J2000 Start day */
    double start_sec_of_day,  /* I: J2000 Start Second of day */
    double end_day,           /* I: J2000 End day */
    double end_sec_of_day,    /* I: J2000 End Second of day */
    int *number_of_records    /* O: Number of records */
)
{
    const char *query_string;
    struct ias_db_query *query;
    int status;
    int count;
    int index;
    TELEMETRY_INFORMATION *telemetry_list_ptr;
    TELEMETRY_INFORMATION *temp_ptr;


    /* Allocate initial storage for the records */
    count = 100;
    telemetry_list_ptr = malloc(sizeof(*telemetry_list_ptr) * count);
    if (!telemetry_list_ptr)
    {
        IAS_LOG_ERROR("Allocating memory for the telemetry list");
        return NULL;
    }

    /* Construct and execute a query to get the records
       MC_ENCODER_FLAGS = 9 (power and indexed)
       SSM_MECH_MODE = 0 (Mode 0)
       SSM_ENCODER_POSITION_SAMPLE_2 > 0 (The second sample
       is used to avoid potential problems with high/low
       data word misalignment that can occur in the first
       and last samples in each ancillary data record) */
    query_string = 
        "SELECT "
        "TIRS_TELEMETRY_COMMAND.L0R_TIME_DAYS_FROM_J2000, "
        "TIRS_TELEMETRY_COMMAND.L0R_TIME_SEC_OF_DAY, "
        "TIRS_TELEMETRY_CIRCUIT.ELEC_ENABLED_FLAGS, "
        "TIRS_TELEMETRY_COMMAND.MC_ENCODER_FLAGS, "
        "TIRS_TELEMETRY_COMMAND.SSM_MECH_MODE, "
        "TIRS_TELEMETRY_COMMAND.SSM_ENCODER_POSITION_SAMPLE_2 "
        "FROM "
        "TIRS_TELEMETRY_COMMAND, TIRS_TELEMETRY_CIRCUIT "
        "WHERE "
        "TIRS_TELEMETRY_COMMAND.L0R_TIME_DAYS_FROM_J2000 = "
        "TIRS_TELEMETRY_CIRCUIT.L0R_TIME_DAYS_FROM_J2000 AND "
        "TIRS_TELEMETRY_COMMAND.L0R_TIME_SEC_OF_DAY = "
        "TIRS_TELEMETRY_CIRCUIT.L0R_TIME_SEC_OF_DAY AND "
        "((TIRS_TELEMETRY_COMMAND.L0R_TIME_DAYS_FROM_J2000 = ? AND "
        "TIRS_TELEMETRY_COMMAND.L0R_TIME_SEC_OF_DAY > ?) OR "
        "TIRS_TELEMETRY_COMMAND.L0R_TIME_DAYS_FROM_J2000 > ?) AND "
        "((TIRS_TELEMETRY_COMMAND.L0R_TIME_DAYS_FROM_J2000 = ? AND "
        "TIRS_TELEMETRY_COMMAND.L0R_TIME_SEC_OF_DAY < ?) OR "
        "TIRS_TELEMETRY_COMMAND.L0R_TIME_DAYS_FROM_J2000 < ?) AND "
        "TIRS_TELEMETRY_COMMAND.MC_ENCODER_FLAGS = 9 AND "
        "TIRS_TELEMETRY_COMMAND.SSM_MECH_MODE = 0 AND "
        "TIRS_TELEMETRY_COMMAND.SSM_ENCODER_POSITION_SAMPLE_2 > 0 "
        "ORDER BY "
        "TIRS_TELEMETRY_COMMAND.L0R_TIME_DAYS_FROM_J2000, "
        "TIRS_TELEMETRY_COMMAND.L0R_TIME_SEC_OF_DAY ASC";

    /* Prepare the query */
    query = ias_db_prepare_query(db, query_string);
    if (!ias_db_query_was_successful(query))
    {
        char db_error_message[300];   

        ias_db_query_get_error_message(query, db_error_message,
            sizeof(db_error_message));
        IAS_LOG_ERROR("Preparing failed: %s", query_string);
        IAS_LOG_ERROR(db_error_message);
        free(telemetry_list_ptr);
        ias_db_query_close(query);
        return NULL;
    }

    /* Bind the variables in the where clause */
    if ((ias_db_bind_double_by_index(query, 0, IAS_DB_PARAMETER_MODE_INPUT,
            &start_day, NULL) != SUCCESS)
        || (ias_db_bind_double_by_index(query, 1, IAS_DB_PARAMETER_MODE_INPUT,
                &start_sec_of_day, NULL) != SUCCESS)
        || (ias_db_bind_double_by_index(query, 2, IAS_DB_PARAMETER_MODE_INPUT,
                &start_day, NULL) != SUCCESS)
        || (ias_db_bind_double_by_index(query, 3, IAS_DB_PARAMETER_MODE_INPUT,
                &end_day, NULL) != SUCCESS)
        || (ias_db_bind_double_by_index(query, 4, IAS_DB_PARAMETER_MODE_INPUT,
                &end_sec_of_day, NULL) != SUCCESS)
        || (ias_db_bind_double_by_index(query, 5, IAS_DB_PARAMETER_MODE_INPUT,
                &end_day, NULL) != SUCCESS))
    {
        char db_error_message[300];   

        IAS_LOG_ERROR("Failed binding parameters to the telemetry query");
        ias_db_query_get_error_message(query, db_error_message,
            sizeof(db_error_message));
        IAS_LOG_ERROR(db_error_message);
        free(telemetry_list_ptr);
        ias_db_query_close(query);
        return NULL;
    }

    /* Execute the query */
    if (ias_db_exec_prepared_query(query) != SUCCESS)
    {
        char db_error_message[300];   

        IAS_LOG_ERROR("Failed executing prepared telemetry query");
        ias_db_query_get_error_message(query, db_error_message,
            sizeof(db_error_message));
        IAS_LOG_ERROR(db_error_message);
        free(telemetry_list_ptr);
        ias_db_query_close(query);
        return NULL;
    }

    /* For each telemetry record returned by the query, extract the information
       and store in the structure */
    index = 0;
    while (ias_db_query_next(query))
    {
        int local_int;
        double local_double;

        /* If there is not enough memory allocated for this record, double the
           allocated memory */
        if (index >= count)
        {
            count *= 2;
            temp_ptr = realloc(telemetry_list_ptr,
                sizeof(*telemetry_list_ptr) * count);
            if (!temp_ptr)
            {
                IAS_LOG_ERROR("Failed growing memory for the telemetry list"
                    " to %d records", count);
                free(telemetry_list_ptr);
                ias_db_query_close(query);
                return NULL;
            }
            telemetry_list_ptr = temp_ptr;
        }

        /* Get the l0r_time_days_from_j2000 */
        status = ias_db_query_double_value(query, &local_double, 0);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining l0r_time_days_from_j2000 %d", index + 1);
            free (telemetry_list_ptr);
            ias_db_query_close(query);
            return NULL;
        }
        telemetry_list_ptr[index].l0r_time_days_from_j2000 = local_double;

        /* Get the l0r_time_sec_of_day */
        status = ias_db_query_double_value(query, &local_double, 1);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining l0r_time_sec_of_day %d", index + 1);
            free (telemetry_list_ptr);
            ias_db_query_close(query);
            return NULL;
        }
        telemetry_list_ptr[index].l0r_time_sec_of_day = local_double;

        /* Get the elec_enabled_flags */
        status = ias_db_query_int_value(query, &local_int, 2);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining elec_ebnabled_flags %d", index + 1);
            free (telemetry_list_ptr);
            ias_db_query_close(query);
            return NULL;
        }
        telemetry_list_ptr[index].elec_enabled_flags = local_int;

        /* Get the mc_encoder_flags */
        status = ias_db_query_int_value(query, &local_int, 3);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining mc_encoder_flags %d", index + 1);
            free (telemetry_list_ptr);
            ias_db_query_close(query);
            return NULL;
        }
        telemetry_list_ptr[index].mc_encoder_flags = local_int;

        /* Get the ssm_mech_mode */
        status = ias_db_query_int_value(query, &local_int, 4);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining ssm_mech_mode %d", index + 1);
            free (telemetry_list_ptr);
            ias_db_query_close(query);
            return NULL;
        }
        telemetry_list_ptr[index].ssm_mech_mode = local_int;
        
        /* Get the ssm_encoder_position_sample_2 */
        status = ias_db_query_double_value(query, &local_double, 5);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining ssm_encoder_position_sample_2 %d",
                 index + 1);
            free (telemetry_list_ptr);
            ias_db_query_close(query);
            return NULL;
        }
        telemetry_list_ptr[index].ssm_encoder_position_sample_2 = local_double;

        index++;

    }/* END for index loop */
 
    /* Shrink the allocated memory to just the size used */
    count = index;
    temp_ptr = realloc(telemetry_list_ptr, sizeof(*telemetry_list_ptr) * count);
    if (!temp_ptr && count != 0)
    {
        /* Just issue a warning if this happens since it is very unlikely and
           no great harm is done if using extra memory */
        IAS_LOG_WARNING("Failed to shrink telemetry list allocation");
    }
    else
        telemetry_list_ptr = temp_ptr;
 
    /* Set the number of scenes found and return the pointer to the list */
    *number_of_records = count;   

    return telemetry_list_ptr;

} /* END get_db_telemetry_list */

/******************************************************************************
NAME:       get_db_ssm_calibration_scenes

PURPOSE: 
Query the TIRS_SSM_ESTIMATION and TIRS_SSM_ESTIMATION_SCA tables to retrieve
SSM calibration scene results from the time period between the mode switch
event's start and end times. In this case the query will use the UTC event
times. 

RETURNS: SUCCESS/ERROR

Note: Can return SUCCESS but return a NULL pointer with zero records.
******************************************************************************/
int get_db_ssm_calibration_scenes
(
    double event_start_time[3],  /* I: Event start time in UTC */
    double event_end_time[3],    /* I: Event end time in UTC */
    SSM_CALIBRATION_INFORMATION **records, /* I: Returned records.  Caller
                                                 needs to free this. */
    int *number_of_records       /* O: Number of records returned from db */
)
{
    const char *query_string;
    struct ias_db_query *query;
    int status;
    int count;
    int index;
    SSM_CALIBRATION_INFORMATION *ssm_calibration_scenes_ptr;
    SSM_CALIBRATION_INFORMATION *temp_ptr;
    char start_time_string[15];     /* Start time in YYYYDDDSSSSS format */
    char end_time_string[15];       /* End time in YYYYDDDSSSSS format */
    

    /* Prepare the start and end time strings */
    status = snprintf(start_time_string, sizeof(start_time_string),
        "%04d%03d%05d", (int)event_start_time[0], (int)event_start_time[1], 
        (int)event_start_time[2]);
    if (status < 0 || status >= sizeof(start_time_string))
    {
        IAS_LOG_ERROR("Constructing start time string");
        return ERROR;
    }
    status = snprintf(end_time_string, sizeof(end_time_string),
        "%04d%03d%05d", (int)event_end_time[0], (int)event_end_time[1], 
        (int)event_end_time[2]);
    if (status < 0 || status >= sizeof(end_time_string))
    {
        IAS_LOG_ERROR("Constructing end time string");
        return ERROR;
    }

    /* Allocate initial storage for the records */
    count = 100;
    ssm_calibration_scenes_ptr = malloc(sizeof(SSM_CALIBRATION_INFORMATION) *
     count);
    if (!ssm_calibration_scenes_ptr)
    {
        IAS_LOG_ERROR("Allocating memory for the ssm calibraton scene list");
        return ERROR;
    }

    /* Construct and execute a query to get the records */
    query_string =
        "SELECT "
        "TIRS_SSM_ESTIMATION.SSM_EPOCH_YEAR, "
        "TIRS_SSM_ESTIMATION.SSM_EPOCH_DOY, "
        "TIRS_SSM_ESTIMATION.SSM_EPOCH_SOD, "
        "TIRS_SSM_ESTIMATION.NET_SSM_ANGLE, "
        "TIRS_SSM_ESTIMATION.NUMBER_OF_SCAS, "
        "TIRS_SSM_ESTIMATION_SCA.SCA_NUMBER, "
        "TIRS_SSM_ESTIMATION_SCA.NBR_OF_PNTS_USED, "
        "TIRS_SSM_ESTIMATION_SCA.POST_FIT_ALNG_TRK_RMSE, "
        "TIRS_SSM_ESTIMATION_SCA.POST_FIT_ACRSS_TRK_RMSE "
        "FROM "
        "TIRS_SSM_ESTIMATION, TIRS_SSM_ESTIMATION_SCA, "
        "GEOMETRIC_WORK_ORDER_COMMON "
        "WHERE "
        "GEOMETRIC_WORK_ORDER_COMMON.GEOMETRIC_WO_COMMON_ID = "
        "TIRS_SSM_ESTIMATION.GEOMETRIC_WO_COMMON_ID AND "
        "GEOMETRIC_WORK_ORDER_COMMON.VALID = 'Y' AND "
        "TIRS_SSM_ESTIMATION.GEOMETRIC_WO_COMMON_ID = "
        "TIRS_SSM_ESTIMATION_SCA.GEOMETRIC_WO_COMMON_ID AND "
        "GEOMETRIC_WORK_ORDER_COMMON.DATE_ACQUIRED >= "
        "TO_DATE(?, 'YYYYDDDSSSSS') AND "
        "GEOMETRIC_WORK_ORDER_COMMON.DATE_ACQUIRED <= "
        "TO_DATE(?, 'YYYYDDDSSSSS') "
        "ORDER BY "
        "TIRS_SSM_ESTIMATION.SSM_EPOCH_YEAR, "
        "TIRS_SSM_ESTIMATION.SSM_EPOCH_DOY, "
        "TIRS_SSM_ESTIMATION.SSM_EPOCH_SOD, "
        "TIRS_SSM_ESTIMATION.GEOMETRIC_WO_COMMON_ID, "
        "TIRS_SSM_ESTIMATION_SCA.SCA_NUMBER ASC ";

    query = ias_db_prepare_query(db, query_string);
    if (!ias_db_query_was_successful(query))
    {
        char db_error_message[300];

        ias_db_query_get_error_message(query, db_error_message,
            sizeof(db_error_message));
        IAS_LOG_ERROR("Preparing failed: %s", query_string);
        IAS_LOG_ERROR("%s", db_error_message);
        free(ssm_calibration_scenes_ptr);
        ias_db_query_close(query);
        return ERROR;
    }

    /* Bind the variables in the where clause */
    if ((ias_db_bind_char_by_index(query, 0, IAS_DB_PARAMETER_MODE_INPUT,
            start_time_string, strlen(start_time_string), NULL) != SUCCESS)
        || (ias_db_bind_char_by_index(query, 1, IAS_DB_PARAMETER_MODE_INPUT,
                end_time_string, strlen(end_time_string), NULL) != SUCCESS))
    {
        char db_error_message[300];   

        IAS_LOG_ERROR("Failed binding parameters to the SSM Calibration "
            "query");
        ias_db_query_get_error_message(query, db_error_message,
            sizeof(db_error_message));
        IAS_LOG_ERROR(db_error_message);
        free(ssm_calibration_scenes_ptr);
        ias_db_query_close(query);
        return ERROR;
    }

    /* Execute the query */
    if (ias_db_exec_prepared_query(query) != SUCCESS)
    {
        char db_error_message[300];   

        IAS_LOG_ERROR("Failed executing prepared SSM Calibration query");
        ias_db_query_get_error_message(query, db_error_message,
            sizeof(db_error_message));
        IAS_LOG_ERROR(db_error_message);
        free(ssm_calibration_scenes_ptr);
        ias_db_query_close(query);
        return ERROR;
    }

    /* For each ssm calibration scenes extract the information and store in
       the structure */
    index = 0;
    while (ias_db_query_next(query))
    {
        int local_int;
        double local_double;
        
        /* If there is not enough memory allocated for this record, double the
           allocated memory */
        if (index >= count)
        {
            count *= 2;
            temp_ptr = realloc(ssm_calibration_scenes_ptr,
                sizeof(*ssm_calibration_scenes_ptr) * count);
            if (!temp_ptr)
            {
                IAS_LOG_ERROR("Failed growing memory for the SSM calibration "
                    " scenes list to %d records", count);
                free(ssm_calibration_scenes_ptr);
                ias_db_query_close(query);
                return ERROR;
            }
            ssm_calibration_scenes_ptr = temp_ptr;
        }

        /* Make sure all the fields are initialized */
        memset(&ssm_calibration_scenes_ptr[index], 0,
            sizeof(SSM_CALIBRATION_INFORMATION));

        /* Get the ssm_epoch_year */
        status = ias_db_query_double_value(query, &local_double, 0);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining ssm_epoch_year %d", index + 1);
            free(ssm_calibration_scenes_ptr);
            ias_db_query_close(query);
            return ERROR;
        }
        ssm_calibration_scenes_ptr[index].ssm_epoch_year = local_double;

        /* Get the ssm_epoch_doy */
        status = ias_db_query_double_value(query, &local_double, 1);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining ssm_epoch_doy %d", index + 1);
            free(ssm_calibration_scenes_ptr);
            ias_db_query_close(query);
            return ERROR;
        }
        ssm_calibration_scenes_ptr[index].ssm_epoch_doy = local_double;

        /* Get the ssm_epoch_sod */
        status = ias_db_query_double_value(query, &local_double, 2);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining ssm_epoch_sod %d", index + 1);
            free(ssm_calibration_scenes_ptr);
            ias_db_query_close(query);
            return ERROR;
        }
        ssm_calibration_scenes_ptr[index].ssm_epoch_sod = local_double;
 
        /* Get the net_ssm_angle */
        status = ias_db_query_double_value(query, &local_double, 3);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining ssm_net_angle %d", index + 1);
            free(ssm_calibration_scenes_ptr);
            ias_db_query_close(query);
            return ERROR;
        }
        ssm_calibration_scenes_ptr[index].net_ssm_angle = local_double;

        /* Get the number_of_sca */
        status = ias_db_query_int_value(query, &local_int, 4);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining number_of_sca %d", index + 1);
            free(ssm_calibration_scenes_ptr);
            ias_db_query_close(query);
            return ERROR;
        }
        ssm_calibration_scenes_ptr[index].number_of_scas = local_int;
              
        /* Get the sca_number */
        status = ias_db_query_int_value(query, &local_int, 5);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining sca_number %d", index + 1);
            free(ssm_calibration_scenes_ptr);
            ias_db_query_close(query);
            return ERROR;
        }
        ssm_calibration_scenes_ptr[index].sca_number = local_int;

        /* Get the nbr_of_pnts_used */
        status = ias_db_query_int_value(query, &local_int, 6);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining nbr_of_pnts_used %d", index + 1);
            free(ssm_calibration_scenes_ptr);
            ias_db_query_close(query);
            return ERROR;
        }
        ssm_calibration_scenes_ptr[index].nbr_of_pnts_used = local_int;

        /* Get the post_fit_alng_trk_rsme */
        status = ias_db_query_double_value(query, &local_double, 7);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining post_fit_alng_trk_rmse %d", index + 1);
            free(ssm_calibration_scenes_ptr);
            ias_db_query_close(query);
            return ERROR;
        }
        ssm_calibration_scenes_ptr[index].post_fit_alng_trk_rmse = local_double;
        
        /* Get the post_fit_acrss_trk_rmse */
        status = ias_db_query_double_value(query, &local_double, 8);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining post_fit_acrss_trk_rmse %d", index + 1);
            free(ssm_calibration_scenes_ptr);
            ias_db_query_close(query);
            return ERROR;
        }
        ssm_calibration_scenes_ptr[index].post_fit_acrss_trk_rmse = 
            local_double;

        index++;

    } /* END for index loop */

    /* Shrink the allocated memory to just the size used */
    count = index;
    temp_ptr = realloc(ssm_calibration_scenes_ptr,
        sizeof(*ssm_calibration_scenes_ptr) * count);
    if (!temp_ptr && count != 0)
    {
        /* Just issue a warning if this happens since it is very unlikely and
           no great harm is done if using extra memory */
        IAS_LOG_WARNING("Failed to shrink SSM calibration scenes list "
            "allocation");
    }
    else
        ssm_calibration_scenes_ptr = temp_ptr;
 
    /* Set the number of scenes found and return the pointer */
    *number_of_records = count;
    *records = ssm_calibration_scenes_ptr;

    return SUCCESS;
}
