#include <math.h>                  /* Math library */

#include "ias_const.h"
#include "profile.h"
#include "tirs_fit_ssm_model.h"

/******************************************************************************
NAME: encoder_model 

PURPOSE: Evaulate the encoder model at a given time

RETURNS: encoder value calculated using the model as a double

******************************************************************************/
double encoder_model
(
    const double *coeff,    /* I: Array of SSM_FIT_NPARMS model coefficients */
    double enc0,            /* I: Initial encoder position */
    double secs_diff        /* I: Seconds from start of event */
)
{
    double s = IAS_SEC_PER_DAY;
    double tdays = secs_diff / s; 
    double tsecs = secs_diff;
    double f;
    
    /* Assemble non-time based components */
    f = enc0 + coeff[SSM_FIT_A0] + coeff[SSM_FIT_A1] + coeff[SSM_FIT_A2] 
             + coeff[SSM_FIT_A3] + coeff[SSM_FIT_SLOPE] * tdays;

    /* Add in time dependent components, if the time scaling coefficient
       is not zero. */
    if (coeff[SSM_FIT_TAU1] != 0.0)
        f = f - coeff[SSM_FIT_A1] * exp(-tsecs / coeff[SSM_FIT_TAU1]);
    if (coeff[SSM_FIT_TAU2] != 0.0)
        f = f - coeff[SSM_FIT_A2] * exp(-tdays / coeff[SSM_FIT_TAU2]);
    if (coeff[SSM_FIT_TAU3] != 0.0)
        f = f - coeff[SSM_FIT_A3] * exp(-tdays / coeff[SSM_FIT_TAU3]);

    return f;
}

/******************************************************************************
NAME: encoder_model_partial

PURPOSE: Evaluate the partial derivative of the model at a given time.

RETURNS: nothing

******************************************************************************/
void encoder_model_partial
(
    const double *coeff,    /* I: Array of SSM_FIT_NPARMS model coefficients */
    double secs_diff,       /* I: Seconds from start of event */
    double *partials        /* O: Computed partial with SSM_FIT_NPARMS terms*/
)
{
    double s = IAS_SEC_PER_DAY;
    double tdays = secs_diff / s;
    double tsecs = secs_diff;


    /* Partial of a0 */
    partials[SSM_FIT_A0] = 1.0;

    /* Partial of a1, time component is zero if tau1 = 0.0 */
    if (coeff[SSM_FIT_TAU1] != 0.0)
        partials[SSM_FIT_A1] = (1.0 - exp(-tsecs / coeff[SSM_FIT_TAU1]));
    else 
        partials[SSM_FIT_A1] = 1.0;

    /* Partial of a2, time component is zero if tau2 = 0.0 */
    if (coeff[SSM_FIT_TAU2] != 0.0)
        partials[SSM_FIT_A2] = (1.0 - exp(-tdays / coeff[SSM_FIT_TAU2]));
    else 
        partials[SSM_FIT_A2] = 1.0;

    /* Partial of a3, time component is zero if tau3 = 0.0 */
    if (coeff[SSM_FIT_TAU3] != 0.0)
        partials[SSM_FIT_A3] = (1.0 - exp(-tdays / coeff[SSM_FIT_TAU3]));
    else 
        partials[SSM_FIT_A3] = 1.0;

    /* Partial of tau1, partial is zero if tau1 = 0.0 */
    if (coeff[SSM_FIT_TAU1] != 0.0)
    {
        partials[SSM_FIT_TAU1] = -((coeff[SSM_FIT_A1] * tsecs)
                / (coeff[SSM_FIT_TAU1] * coeff[SSM_FIT_TAU1]))
                * exp(-tsecs / coeff[SSM_FIT_TAU1]);
    }
    else 
        partials[SSM_FIT_TAU1] = 0.0;

    /* Partial of tau2, partial is zero if tau2 = 0.0 */
    if (coeff[SSM_FIT_TAU2] != 0.0)
    {
        partials[SSM_FIT_TAU2] = -((coeff[SSM_FIT_A2] * tdays)
                / (coeff[SSM_FIT_TAU2] * coeff[SSM_FIT_TAU2]))
                * exp(-tdays / coeff[SSM_FIT_TAU2]);
    }
    else 
        partials[SSM_FIT_TAU2] = 0.0;

    /* Partial of tau3, partial is zero if tau3 = 0.0 */
    if (coeff[SSM_FIT_TAU3] != 0.0)
    {
        partials[SSM_FIT_TAU3] = -((coeff[SSM_FIT_A3] * tdays)
                / (coeff[SSM_FIT_TAU3] * coeff[SSM_FIT_TAU3]))
                * exp(-tdays / coeff[SSM_FIT_TAU3]);
    }
    else 
        partials[SSM_FIT_TAU3] = 0.0;

    /* Partial of slope */
    partials[SSM_FIT_SLOPE] = tdays;
}
