/******************************************************************************
NAME: fit_lsq

PURPOSE: Determine solution for SSM encoder model for minimization using a
    weighed least sqaures.  More details can be found in the corresponding ADD.

RETURNS: SUCCESS/ERROR

******************************************************************************/
#include <math.h>                  /* Math library */
#include <gsl/gsl_linalg.h>

#include "ias_const.h"   
#include "ias_logging.h"           /* IAS logging library */
#include "tirs_fit_ssm_model.h" 

#define		TOO_SMALL_COEFF	0.0
#define		DEBUG0		0
#define		DEBUG1		0
#define		DEBUG2		0
#define		DEBUG3		0

/******************************************************************************
Name: print_vector

Purpose: Helper function to issue a log message with the contents of a vector.

Returns: SUCCESS/ERROR

******************************************************************************/
static void print_vector
(
    const char *label,      /* I: label to print out for the vector */
    const double *vector,   /* I: vector to print */
    int vector_size         /* I: entries in the vector */
)
{
    int index;
    int current_position = 0;
    char output[vector_size * 20]; /* temporary output vector large enough
                                      for the entire output */

    for (index = 0; index < vector_size; index++)
    {
        current_position += sprintf(&output[current_position], "%f ",
            vector[index]);
    }

    IAS_LOG_INFO("%s: %s", label, output);
}

/******************************************************************************
Name: setup_apriori_arrays

Purpose: Set up the apriori weight arrays needed to do the fit.

Returns: SUCCESS/ERROR

******************************************************************************/
static int setup_apriori_arrays
(
    const ENCODER_RECORD *tirs_data,/* I: TIRS data records */
    int number_tirs_data,           /* I: Number of TIRS encoder values */
    SSM_SOLUTION_PARMS *out_solution_parms, /* O: SSM solution param pointer */
    double apriori_weights[SSM_FIT_NPARMS], /* O: Apriori weights */
    double apriori_Cvec[SSM_FIT_NPARMS] /* O: Flag for apriori to left hand side
                                              of equation */

)
{
    int number_parameters = SSM_FIT_NPARMS; /* Total number of coefficients */
    int point_index;                /* Counter */
    int vector_index;               /* Counter */

    /* Apriori weights, that are added to the least squares solution when
       no points are available within the break times/model observation
       gap windows */
    double apriori_weights_poor[SSM_FIT_NPARMS];

    /* Masks for model coefficients. Determines of parameter will be included
       in solution.  Overides other constraints for that parameter. */
    int coefficient_masks[SSM_FIT_NPARMS];

    /* Times for when solution can be found for model coefficients 
       and minimum number of samples needed for each break. */
    int break_times_samples[SSM_OBSERVATION_GAP_WINDOWS];

    /* Number of samples between time breaks */
    int number_samples_per_break[SSM_OBSERVATION_GAP_WINDOWS];    
    double max_time;               /* Maximum time from start of event */
    /* Maximum times allowable for inclusion of coefficients in solution */
    double break_point_times[SSM_BREAK_POINT_TIMES];
    SSM_SOLUTION_PARMS solution_parms;

    /* Grab some of the (possibly) user entered information */
    get_break_point_times(break_point_times);
    get_coefficient_masks(coefficient_masks);
    get_break_times_samples(break_times_samples);
    get_apriori_weights(apriori_weights_poor);

    /* Default all the weights to 0.0 */
    for (vector_index = 0; vector_index < number_parameters; vector_index++)
    {
        apriori_weights[vector_index] = 0.0;
        apriori_Cvec[vector_index] = 0.0;
    }

    /* Square the weights that will be used for periods where the data
       is not as trustworthy */
    for (vector_index = 0; vector_index < number_parameters; vector_index++)
    {
        apriori_weights_poor[vector_index] = apriori_weights_poor[vector_index]
            * apriori_weights_poor[vector_index];
    }

    /* Analyze SSM information and time for information to determine whether
       which set of coefficients can be solved for and whether those sets
       will have enough measurements for a good quality solution. */
    if (analyze_time_period(tirs_data, number_tirs_data,
                number_samples_per_break) != SUCCESS)
    {
        IAS_LOG_ERROR("Error analyzing TIRS times");
        return ERROR;
    }

    /* Determine number of parameters that can be solved for based on
       on observations per time breaks. */
    max_time = tirs_data[number_tirs_data - 1].secs_diff;
    IAS_LOG_INFO("Largest time from start of event: %f", max_time); 

    if (max_time < break_point_times[0])
    {
        solution_parms = SSM_SOLUTION_START_UP;
        IAS_LOG_INFO("SSM solution for parm a0");
    }
    else if (max_time < break_point_times[1])
    {
        solution_parms = SSM_SOLUTION_DAY;
        IAS_LOG_INFO("SSM solution for parm a0, a1, tau1");
    }
    else if (max_time < break_point_times[2])
    {
        solution_parms = SSM_SOLUTION_WEEK;
        IAS_LOG_INFO("SSM solution for parm a0, a1, a2, tau1, tau2");
    }
    else if (max_time < break_point_times[3])
    {
        solution_parms = SSM_SOLUTION_TWO_WEEKS;
        IAS_LOG_INFO("SSM solution for parm a0, a1, a2, a3, tau1, tau2, Slope");
    }
    else /* if (max_time > break_point_times[3]) */
    {
        solution_parms = SSM_SOLUTION_FULL;
        IAS_LOG_INFO("SSM solution for all parms");
    }

    /* This probably needs to handled better. Code might be confusing but check
       too see if a there are enough observations per time frame associated
       with parameters of model for a good solution for those coefficients. */

    /* Full, 12 days, 1 week, 1 day solutions all include a1 and tau1 */ 
    if ((solution_parms == SSM_SOLUTION_FULL
         || solution_parms == SSM_SOLUTION_TWO_WEEKS
         || solution_parms == SSM_SOLUTION_WEEK
         || solution_parms == SSM_SOLUTION_DAY)
        && number_samples_per_break[0] <= break_times_samples[0])
    {
        apriori_weights[SSM_FIT_A1] = apriori_weights_poor[SSM_FIT_A1];
        apriori_weights[SSM_FIT_TAU1] = apriori_weights_poor[SSM_FIT_TAU1];
        apriori_Cvec[SSM_FIT_A1] = 1.0;
        apriori_Cvec[SSM_FIT_TAU1] = 1.0;
        if (coefficient_masks[SSM_FIT_A1] == 1)
            coefficient_masks[SSM_FIT_A1] = 0;
        IAS_LOG_INFO("SSM solution for parm a1 and tau1 are defined");
    }

    /* Full, 12 days, 1 week solutions all include a2 and tau2 */ 
    if ((solution_parms == SSM_SOLUTION_FULL
         || solution_parms == SSM_SOLUTION_TWO_WEEKS
         || solution_parms == SSM_SOLUTION_WEEK)
        && number_samples_per_break[1] <= break_times_samples[1])
    {
        apriori_weights[SSM_FIT_A2] = apriori_weights_poor[SSM_FIT_A2];
        apriori_weights[SSM_FIT_TAU2] = apriori_weights_poor[SSM_FIT_TAU2];
        apriori_Cvec[SSM_FIT_A2] = 1.0;
        apriori_Cvec[SSM_FIT_TAU2] = 1.0;
        if (coefficient_masks[SSM_FIT_A2] == 1)
            coefficient_masks[SSM_FIT_A2] = 0;
        if (coefficient_masks[SSM_FIT_TAU2] == 1)
            coefficient_masks[SSM_FIT_TAU2] = 0;
        IAS_LOG_INFO("SSM solution for parm a2 a tau2 are defined");
    }

    if ((solution_parms == SSM_SOLUTION_FULL
         || solution_parms == SSM_SOLUTION_TWO_WEEKS)
        && number_samples_per_break[2] <= break_times_samples[2])
    {
        apriori_weights[SSM_FIT_A3] = apriori_weights_poor[SSM_FIT_A3];
        apriori_Cvec[SSM_FIT_A3] = 1.0;
        if (coefficient_masks[SSM_FIT_A3] == 1)
            coefficient_masks[SSM_FIT_A3] = 0;
        IAS_LOG_INFO("SSM solution for parm a3 is defined");
    }

    /* Check whether coefficients should be masked. This will override any
       previous decisions on what not to run. It does not add a parameter if
       requested and the previous logic had determined that particular
       parameter should not be used or solved for. The mask is for only
       removing a parameter. */
    if (coefficient_masks[SSM_FIT_A0] == 0)
    {
        apriori_weights[SSM_FIT_A0] = apriori_weights_poor[SSM_FIT_A0];
        apriori_Cvec[SSM_FIT_A0] = 1.0;
        IAS_LOG_INFO("Suppressing SSM_FIT_A0 due to mask value");
    }
    if (coefficient_masks[SSM_FIT_A1] == 0)
    {
        apriori_weights[SSM_FIT_A1] = apriori_weights_poor[SSM_FIT_A1];
        apriori_Cvec[SSM_FIT_A1] = 1.0;
        IAS_LOG_INFO("Suppressing SSM_FIT_A1 due to mask value");
    }
    if (coefficient_masks[SSM_FIT_A2] == 0)
    {
        apriori_weights[SSM_FIT_A2] = apriori_weights_poor[SSM_FIT_A2];
        apriori_Cvec[SSM_FIT_A2] = 1.0;
        IAS_LOG_INFO("Suppressing SSM_FIT_A2 due to mask value");
    }
    if (coefficient_masks[SSM_FIT_A3] == 0)
    {
        apriori_weights[SSM_FIT_A3] = apriori_weights_poor[SSM_FIT_A3];
        apriori_Cvec[SSM_FIT_A3] = 1.0;
        IAS_LOG_INFO("Suppressing SSM_FIT_A3 due to mask value");
    }
    if (coefficient_masks[SSM_FIT_TAU1] == 0)
    {
        apriori_weights[SSM_FIT_TAU1] = apriori_weights_poor[SSM_FIT_TAU1];
        apriori_Cvec[SSM_FIT_TAU1] = 1.0;
        IAS_LOG_INFO("Suppressing SSM_FIT_TAU1 due to mask value");
    }
    if (coefficient_masks[SSM_FIT_TAU2] == 0)
    {
        apriori_weights[SSM_FIT_TAU2] = apriori_weights_poor[SSM_FIT_TAU2];
        apriori_Cvec[SSM_FIT_TAU2] = 1.0;
        IAS_LOG_INFO("Suppressing SSM_FIT_TAU2 due to mask value");
    }
    if (coefficient_masks[SSM_FIT_TAU3] == 0)
    {
        apriori_weights[SSM_FIT_TAU3] = apriori_weights_poor[SSM_FIT_TAU3];
        apriori_Cvec[SSM_FIT_TAU3] = 1.0;
        IAS_LOG_INFO("Suppressing SSM_FIT_TAU3 due to mask value");
    }
    if (coefficient_masks[SSM_FIT_SLOPE] == 0)
    {
        apriori_weights[SSM_FIT_SLOPE]
            = apriori_weights_poor[SSM_FIT_SLOPE];
        apriori_Cvec[SSM_FIT_SLOPE] = 1.0;
        IAS_LOG_INFO("Suppressing SSM_FIT_SLOPE due to mask value");
    }

    /* Invert the final weights */
    for (point_index = 0; point_index < number_parameters; point_index++)
    {
        if (apriori_weights[point_index] != 0.0)
            apriori_weights[point_index] = 1.0 / apriori_weights[point_index];
    }
    print_vector("SSM apriori weights to use", apriori_weights,
        number_parameters);

    *out_solution_parms = solution_parms;
    return SUCCESS;
}

/******************************************************************************
Name: perform_fit

Purpose: Perform the fit to determine solution for SSM encoder model for
    minimization using a weighed least sqaures.  More details can be found in
    the corresponding ADD.

Returns: SUCCESS/ERROR

******************************************************************************/
int perform_fit
(
    const ENCODER_RECORD *tirs_data,/* I: TIRS data records */
    int number_tirs_data,           /* I: Number of TIRS encoder values */
    double encoder_value0,          /* I: Initial encoder value */
    SSM_SOLUTION_PARMS solution_parms, /* I: Solution period */
    double apriori_weights[SSM_FIT_NPARMS], /* I: apriori weights */
    double apriori_Cvec[SSM_FIT_NPARMS], /* I: apriori vector */
    double *new_model       /* O: New model coefficients.  Array sized by
                              SSM_FIT_NPARMS. */
)
{
    int iteration = 0;              /* Loop counter */
    int number_parameters = SSM_FIT_NPARMS; /* Total number of coefficients */
    int maximum_iterations;         /* User-selected limit */
    int vector_index;               /* Counter */
    double m;                       /* Partials multiplication */
    double last_coeff[SSM_FIT_NPARMS];   /* Last set of coefficients */
    double coeff[SSM_FIT_NPARMS];  /* Current model parameters */
    double initial_coeff[SSM_FIT_NPARMS];/* Nominal or initial coefficients */

    /* Standard deviation of the encoder counts to use to weight the
       telemetry samples */
    double telemetry_obs_stdev;
    /* Standard deviation of the encoder counts to use to weight the
       image observation samples */
    double image_obs_stdev;
    double telemetry_obs_weight;    /* Telemetry observation weight */
    double image_obs_weight;        /* Image observation weight */

    /* Threshold for determining whether converging to solution */
    double convergence_threshold;

    /* Weights for coefficients for determining whether converging to solution */
    double convergence_threshold_weights[SSM_FIT_NPARMS];

    int return_status = SUCCESS;   /* Final return status */

    gsl_matrix *N;                 /* Total normalized matrix for LSQ */
    gsl_matrix *N00;               /* Single normalized observation matrix
                                      for LSQ */
    gsl_vector *C;                 /* Total total observation vector */
    gsl_vector *C00;               /* Single total observation vector */
    gsl_vector *S;                 /* Goofy vector used for determining
                                      parameters to solve for */
    gsl_vector *v;                 /* fitting vectors */
    gsl_vector *tau;               /* QR vectors */

    /* Grab some of the (possibly) user entered information */
    get_initial_values(initial_coeff);
    telemetry_obs_stdev = get_telemetry_obs_stdev();
    image_obs_stdev = get_image_obs_stdev();
    IAS_LOG_INFO("Telemetry obs stdev = %f", telemetry_obs_stdev);
    IAS_LOG_INFO("Image obs stdev = %f", image_obs_stdev);

    convergence_threshold = get_convergence_threshold();
    get_convergence_threshold_weights(convergence_threshold_weights);
    maximum_iterations = get_maximum_iterations();

    /* Weights are entered as standard deviations, put them in terms
       of variance for LSQ matrices or equations. */
    telemetry_obs_weight = 1.0 / (telemetry_obs_stdev * telemetry_obs_stdev);
    image_obs_weight = 1.0 / (image_obs_stdev * image_obs_stdev);

    for (vector_index=0; vector_index < number_parameters; vector_index++)
    {
        coeff[vector_index]      = initial_coeff[vector_index];
        last_coeff[vector_index] = initial_coeff[vector_index];
    }

    /* Check exponential times constants for getting too small */
    if (coeff[SSM_FIT_TAU1] < TOO_SMALL_COEFF || 
            coeff[SSM_FIT_TAU2] < TOO_SMALL_COEFF ||
            coeff[SSM_FIT_TAU3] < TOO_SMALL_COEFF)
    {
        IAS_LOG_ERROR("A time constant given is too small %f,%f,%f", 
                coeff[SSM_FIT_TAU1], coeff[SSM_FIT_TAU2], coeff[SSM_FIT_TAU3]);
        return ERROR;
    }

    print_vector("Initial parameters", coeff, number_parameters);

    /* Allocate buffer space for least squares routine.  */
    N = gsl_matrix_alloc(number_parameters, number_parameters);
    N00 = gsl_matrix_alloc(number_parameters, number_parameters);
    C = gsl_vector_alloc(number_parameters);
    C00 = gsl_vector_alloc(number_parameters);
    S = gsl_vector_alloc(number_parameters);
    tau = gsl_vector_alloc(number_parameters);
    v = gsl_vector_alloc(number_parameters);
    if (N == NULL || N00 == NULL || C == NULL || C00 == NULL || S == NULL
            || tau == NULL || v == NULL)
    {
        IAS_LOG_ERROR("Error allocating arrays");
        gsl_matrix_free(N);
        gsl_matrix_free(N00);
        gsl_vector_free(C);
        gsl_vector_free(C00);
        gsl_vector_free(S);
        gsl_vector_free(tau);
        gsl_vector_free(v);
        return ERROR;
    }

    while (iteration < maximum_iterations)
    {
        int matrix_index;                   /* Counter */
        int point_index;                    /* Counter */
        int status;                         /* Temporary status */
        double delta_coeff[SSM_FIT_NPARMS]; /* Change in coefficients */
        double tolerance;                   /* Sum of change in coefficients */
        double Svec;  /* Weights for determining parameters to solve for */
        double rms = 0.0;                   /* Sum of RMSE */
        char iteration_string[50];          /* Iteration message buffer */

        /* Initialize the N matrix and C vector */
        gsl_matrix_set_zero(N);
        gsl_vector_set_zero(C);

        for (point_index = 0; point_index < number_tirs_data; point_index++)
        {
            double diff;    /* Difference in model and measured values */
            double weight;  /* Least squares weight */
            double model_partials[SSM_FIT_NPARMS]; /* Initial solution 
                                                      coefficients */
            double model_profile;           /* Model of encoder profile */

            /* Determine current model fit and partials for that fit */
            model_profile = encoder_model(coeff, encoder_value0, 
                    tirs_data[point_index].secs_diff);
            encoder_model_partial(coeff, tirs_data[point_index].secs_diff, 
                    model_partials);

            diff = tirs_data[point_index].encoder_count - model_profile;
            rms = rms + diff * diff;

            /* Determine weight based on observation type */
            if (tirs_data[point_index].ssm_type == SSM_IMAGE)
                weight = image_obs_weight;
            else
            {
                /* ssm_type == SSM_TELEMETRY */
                weight = telemetry_obs_weight;
            }

            /* Build single observation matrice and then add that to the
               current summed matrices */
            for (matrix_index = 0; matrix_index < number_parameters;
                    matrix_index++)
            {
                gsl_vector_set(C00, matrix_index, (diff * weight *
                            model_partials[matrix_index]));
                for (vector_index = 0; vector_index < number_parameters;
                        vector_index++)
                {
                    m = weight * model_partials[matrix_index] *
                        model_partials[vector_index];
                    if (matrix_index == vector_index)
                        m += apriori_weights[matrix_index];

                    gsl_matrix_set(N00, matrix_index, vector_index, m);
                }
            }
            status = gsl_vector_add(C, C00);
            if (status != 0)
            {
                IAS_LOG_ERROR("Failed adding C and C00 vectors: %s", 
                    gsl_strerror(status));
                return_status = ERROR;
                break;
            }
            status = gsl_matrix_add(N, N00);
            if (status != 0)
            {
                IAS_LOG_ERROR("Failed adding N and N00 matrices: %s", 
                    gsl_strerror(status));
                return_status = ERROR;
                break;
            }
        }
        if (return_status == ERROR)
            break;

        rms = sqrt(rms / (double)number_tirs_data);

        /* Debugging step that prints matrices for LSQ */
        if (DEBUG0)
        {
            for (matrix_index = 0; matrix_index < number_parameters;
                    matrix_index++)
            {
                for (vector_index = 0; vector_index < number_parameters;
                        vector_index++)
                {
                    m = gsl_matrix_get(N, matrix_index, vector_index);
                    printf("%d %d %e, ", matrix_index, vector_index, m);
                }
                printf("\n");
            }
            for (vector_index = 0; vector_index < number_parameters;
                    vector_index++)
            {
                m = gsl_vector_get(C, vector_index);
                printf("%d %.2e\n", vector_index, m);
            }
        }

        /* This may not be the best way to handle the situation but I'm going
           to build a one dimensional vector with either 1 (parameter will be
           solved for) or 0 (parameter will not be solved for) and then from
           that vector build a matrix that will have ones and zeros for the
           locations that will be used or solved for.  Mutliplying this matrix
           by the original N matrix will zero out the entries not needed but
           will also zero out the diagonals which is not what we want for 
           determining solution.  I will then go back as a last step and set the
           diagonals to 1 those that need to be only ones. */
        gsl_vector_set_zero(S); 
        if (solution_parms == SSM_SOLUTION_START_UP)
        {
            gsl_vector_set(S, SSM_FIT_A0, 1); /* Solving for a0 */
        }
        else if (solution_parms == SSM_SOLUTION_DAY)
        {
            gsl_vector_set(S, SSM_FIT_A0, 1); /* Solving for a0 */
            gsl_vector_set(S, SSM_FIT_A1, 1); /* Solving for a1 */
            gsl_vector_set(S, SSM_FIT_TAU1, 1); /* Solving for tau1 */
        }
        else if (solution_parms == SSM_SOLUTION_WEEK)
        {
            gsl_vector_set(S, SSM_FIT_A0, 1); /* Solving for a0 */
            gsl_vector_set(S, SSM_FIT_A1, 1); /* Solving for a1 */
            gsl_vector_set(S, SSM_FIT_A2, 1); /* Solving for a2 */
            gsl_vector_set(S, SSM_FIT_TAU1, 1); /* Solving for tau1 */
            gsl_vector_set(S, SSM_FIT_TAU2, 1); /* Solving for tau2 */
        }
        else if (solution_parms == SSM_SOLUTION_TWO_WEEKS)
        {
            gsl_vector_set(S, SSM_FIT_A0, 1); /* Solving for a0 */
            gsl_vector_set(S, SSM_FIT_A1, 1); /* Solving for a1 */
            gsl_vector_set(S, SSM_FIT_A2, 1); /* Solving for a2 */
            gsl_vector_set(S, SSM_FIT_A3, 1); /* Solving for a3 */
            gsl_vector_set(S, SSM_FIT_TAU1, 1); /* Solving for tau1 */
            gsl_vector_set(S, SSM_FIT_TAU2, 1); /* Solving for tau2 */
            gsl_vector_set(S, SSM_FIT_SLOPE, 1); /* Solving for Slope */
        }
        else /* solution_parms == SSM_SOLUTION_FULL */
        {
            gsl_vector_set(S, SSM_FIT_A0, 1); /* Solving for a0 */
            gsl_vector_set(S, SSM_FIT_A1, 1); /* Solving for a1 */
            gsl_vector_set(S, SSM_FIT_A2, 1); /* Solving for a2 */
            gsl_vector_set(S, SSM_FIT_A3, 1); /* Solving for a3 */
            gsl_vector_set(S, SSM_FIT_TAU1, 1); /* Solving for tau1 */
            gsl_vector_set(S, SSM_FIT_TAU2, 1); /* Solving for tau2 */
            gsl_vector_set(S, SSM_FIT_TAU3, 1); /* Solving for tau2 */
            gsl_vector_set(S, SSM_FIT_SLOPE, 1); /* Solving for Slope */
       }

        for (vector_index = 0; vector_index < number_parameters;
                vector_index++)
        {
            /* Get vector that is flag indicating whether solution is to be
               found for this parameter. */
            Svec = gsl_vector_get(S, vector_index);
            m = gsl_vector_get(C, vector_index);

            /* Svec = 0 if parameter is not be be solved for.  apriori_Cvec
               tells whether solution will be ill defined for this
               parameter. */
            m = Svec * (m + apriori_Cvec[vector_index]
              * apriori_weights[vector_index]
              * (initial_coeff[vector_index] - last_coeff[vector_index]));

            /* Reset right hand side matrix or vector according to
               conditions. */
            gsl_vector_set(C, vector_index, m);
            if (DEBUG3)
                printf("S %d %f C %f\n", vector_index, Svec, m);
        }

        /* Mulitply normal matrix by S matrix which consists of either 0 or
           1s depending on whether we are solving for the parameter. */
        for (matrix_index = 0; matrix_index < number_parameters;
                matrix_index++)
        {
            double Smat; /* Weights for determining parameters to
                            solve for */
            Smat = gsl_vector_get(S, matrix_index);
            for (vector_index = 0; vector_index < number_parameters;
                    vector_index++)
            {
                Svec = gsl_vector_get(S, vector_index);
                m = gsl_matrix_get(N, matrix_index, vector_index);
                m = m * Smat * Svec;
                if (DEBUG3)
                    printf("N %d %d %f %f %f\n", matrix_index,
                            vector_index, Smat, Svec, m);
                gsl_matrix_set(N, matrix_index, vector_index, m);
            }
        }

        /* Set the diagonals for parameters that are not to be solved for
           to 1. */
        if (solution_parms == SSM_SOLUTION_START_UP)
        {
            /* Solving only for a0 */
            for (matrix_index = 1, vector_index = 1;
                    matrix_index < number_parameters; matrix_index++, 
                    vector_index++)
            {
                gsl_matrix_set(N, matrix_index, vector_index, 1.0);
            }
        }
        else if (solution_parms == SSM_SOLUTION_DAY)
        {
            /* Solving for only a0, a1 and tau1 */
            for (matrix_index = 2, vector_index = 2; matrix_index < 4;
                    matrix_index++, vector_index++)
            {
                gsl_matrix_set(N, matrix_index, vector_index, 1.0);
            }
            for (matrix_index = 5, vector_index = 5;
                    matrix_index < number_parameters; matrix_index++, 
                    vector_index++)
            {
                gsl_matrix_set(N, matrix_index, vector_index, 1.0);
            }
        }
        else if (solution_parms == SSM_SOLUTION_WEEK)
        {
            /* Solving for only a0, a1, a2 and tau1, tau2 */
            gsl_matrix_set(N, SSM_FIT_A3, SSM_FIT_A3, 1.0);
            gsl_matrix_set(N, SSM_FIT_TAU3, SSM_FIT_TAU3, 1.0);
            gsl_matrix_set(N, SSM_FIT_SLOPE, SSM_FIT_SLOPE, 1.0);
        }
        else if (solution_parms == SSM_SOLUTION_TWO_WEEKS)
        {
            /* Solving for only a0, a1, a2, a3, and tau1, tau2, and Slope */
            gsl_matrix_set(N, SSM_FIT_TAU3, SSM_FIT_TAU3, 1.0);
        }

        /* Debugging step that prints matrices for LSQ */
        if (DEBUG2)
        {
            for (matrix_index = 0; matrix_index < number_parameters;
                    matrix_index++)
            {
                for (vector_index = 0; vector_index < number_parameters;
                        vector_index++)
                {
                    m = gsl_matrix_get(N, matrix_index, vector_index);
                    printf("%d %d %.2e, ", matrix_index, vector_index, m);
                }
                printf("\n");
            }
            for (vector_index = 0; vector_index < number_parameters;
                    vector_index++)
            {
                m = gsl_vector_get(C, vector_index);
                printf("%d %.2e\n", vector_index, m);
            }
        }

        /* Solve the resulting equations */
        status = gsl_linalg_QR_decomp(N, tau);
        if (status != 0)
        {
            IAS_LOG_ERROR("Failed performing QR decomp: %s", 
                gsl_strerror(status));
            return_status = ERROR;
            break;
        }

        status = gsl_linalg_QR_solve(N, tau, C, v);
        if (status != 0)
        {
            IAS_LOG_ERROR("Failed performing QR solve: %s", 
                gsl_strerror(status));
            return_status = ERROR;
            break;
        }

        /* Sum coefficient updates and break from solution if tolerance is
           met */
        tolerance = 0.0;
        for (vector_index = 0; vector_index<number_parameters; vector_index++)
        {
            double scale;  /* Scale for coefficients for normalizing */

            delta_coeff[vector_index] = gsl_vector_get(v, vector_index);

            scale = convergence_threshold_weights[vector_index];

            tolerance = tolerance + (delta_coeff[vector_index] / scale)
                * (delta_coeff[vector_index] / scale);
        }
        tolerance = sqrt(tolerance);

        /* For each iteration, print the current iteration's coefficients to
           help identify where the model is diverging.  The return value of
           snprintf is not being checked since the buffer was made large
           enough that it will ever be a problem (and if it were, the only
           side effect would be a label being cut off). */
        snprintf(iteration_string, sizeof(iteration_string),
            "Iteration %d coeffs", iteration);
        print_vector(iteration_string, coeff, SSM_FIT_NPARMS);
        snprintf(iteration_string, sizeof(iteration_string),
            "Iteration %d delta coeffs", iteration);
        print_vector(iteration_string, delta_coeff, SSM_FIT_NPARMS);

        /* Debugging step that prints new coefficients and  calculated
           adjustments */
        if (DEBUG1)
        {
            printf("iteration %d rms %.2f\n", iteration, rms);
            printf("tolerance %f\n", tolerance);
        }

        for (vector_index = 0; vector_index < number_parameters; vector_index++)
            coeff[vector_index] += delta_coeff[vector_index];
        for (vector_index = 0; vector_index < number_parameters; vector_index++)
            last_coeff[vector_index] = coeff[vector_index];

        /* Check for rms being "nan" since a diverging model can cause that */
        if (isnan(rms))
        {
            IAS_LOG_ERROR("Model is diverging since the RMS value is 'nan'");
            return_status = ERROR;
            break;
        }

        /* If the tolerance has dropped below the convergence threshold, the
           algorithm is done iterating */
        if (tolerance <= convergence_threshold)
            break;
        iteration++;
    }

    if (iteration == maximum_iterations)
        IAS_LOG_INFO("Maximum iterations was reached %d", iteration);

    /* Copy the final model parameters to the new model array */
    for (vector_index = 0; vector_index < number_parameters; vector_index++)
        new_model[vector_index] = coeff[vector_index];

    /* Free memory. */
    gsl_matrix_free(N);
    gsl_matrix_free(N00);
    gsl_vector_free(C);
    gsl_vector_free(C00);
    gsl_vector_free(S);
    gsl_vector_free(tau);
    gsl_vector_free(v);

    return return_status;
}

/******************************************************************************
Name: fit_lsq_ias

Purpose: Determine solution for SSM encoder model for minimization using a
    weighed least sqaures.  More details can be found in the corresponding ADD.

Returns: SUCCESS/ERROR

******************************************************************************/
int fit_lsq_ias
(
    const ENCODER_RECORD *tirs_data,/* I: TIRS data records */
    int number_tirs_data,           /* I: Number of TIRS encoder values */
    double encoder_value0,          /* I: Initial encoder value */
    double *coeff,      /* O: New model coefficients.  Array sized by
                              SSM_FIT_NPARMS. */
    int *final_num,     /* O: Number of points in final rmse.  Array sized by
                              RMSE_COUNT. */
    double *final_rms   /* O: Best fit RMS error.  Array sized by RMSE_COUNT */
)
{
    int vector_index;               /* Counter */

    /* Apriori weights, that are added to the least squares solution when
       no points are available within the break times/model observation
       gap windows */
    double apriori_weights[SSM_FIT_NPARMS];
    double apriori_Cvec[SSM_FIT_NPARMS]; /* Flag for apriori to left hand side
                                            of equation */

    /* Maximum times allowable for inclusion of coefficients in solution */
    SSM_SOLUTION_PARMS solution_parms;
    double telem_fit;              /* Fit of model to telemetry data only */
    int telem_num;                 /* Fit of model to telemetry data only,
                                      number of points */
    double scene_fit;              /* Fit of model to scene calibration
                                      data only */
    int scene_num;                 /* Fit of model to scene calibration data
                                      only, number of points */
    double total_fit;              /* Fit of model to telemtry and scene
                                      calibration data */
    int total_num;                 /* Fit of model to telemtry and scene
                                      calibration data number of points */

    /* Set up the apriori weights controlling the fit */
    if (setup_apriori_arrays(tirs_data, number_tirs_data, &solution_parms, 
            apriori_weights, apriori_Cvec) != SUCCESS)
    {
        IAS_LOG_ERROR("Failed setting up fit coefficients");
        return ERROR;
    }

    if (perform_fit(tirs_data, number_tirs_data, encoder_value0,
            solution_parms, apriori_weights, apriori_Cvec, coeff) != SUCCESS)
    {
        IAS_LOG_ERROR("Failed fitting SSM model");
        return ERROR;
    }

    /* Initialize individual rmses */
    telem_fit = 0.0;
    telem_num = 0;
    scene_fit = 0.0;
    scene_num = 0;
    total_fit = 0.0;

    /* Accumulate the RMSE values */
    for (vector_index = 0; vector_index < number_tirs_data; vector_index++)
    {
        double model_profile;  /* Model of encoder profile */
        double diff;           /* Difference between model and data point */

        model_profile = encoder_model(coeff, encoder_value0, 
                tirs_data[vector_index].secs_diff);
        diff = model_profile - tirs_data[vector_index].encoder_count; 
        if (tirs_data[vector_index].ssm_type == SSM_TELEMETRY)
        {
            telem_fit = telem_fit + diff * diff;
            telem_num++;
        }
        else /* tirs_data[vector_index].ssm_type == SSM_IMAGE) */
        {
            scene_fit = scene_fit + diff * diff;
            scene_num++;
        }
    }

    if (telem_num != 0 && scene_num != 0)
        total_fit = sqrt((scene_fit + telem_fit) / (scene_num + telem_num));
    total_num = scene_num + telem_num;
    if (telem_num != 0)
        telem_fit = sqrt(telem_fit / telem_num);
    if (scene_num != 0)
        scene_fit = sqrt(scene_fit / scene_num);

    final_rms[TELEMETRY_RMSE] = telem_fit;
    final_num[TELEMETRY_RMSE] = telem_num;
    final_rms[IMAGE_RMSE] = scene_fit;
    final_num[IMAGE_RMSE] = scene_num;
    final_rms[COMBINED_RMSE] = total_fit;
    final_num[COMBINED_RMSE] = telem_num + scene_num;

    IAS_LOG_INFO("Scene RMSE statistics %d %f", scene_num, scene_fit);
    IAS_LOG_INFO("Telemetry RMSE statistics %d %f", telem_num, telem_fit);
    IAS_LOG_INFO("Total RMSE statistics %d %f", total_num, total_fit);

    print_vector("New parameters", coeff, SSM_FIT_NPARMS);

    return SUCCESS;
}
