/******************************************************************************
NAME:        generate_encoder_file

PURPOSE:
Main driver that generates the encoder file by querying the database with event
start and end times.  The output file is used by tirs_fit_ssm_model
application.

RETURNS:
    EXIT_SUCCESS    generated the intermediate ssm encoder file successfully
    EXIT_FAILURE    Error occurred generating the intermediate ssm encoder file

*******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_cpf.h"
#include "utilities.h"
#include "database_routines.h"
#include "read_generate_parameters.h"

int main(int argc, char *argv[])
{
    TELEMETRY_INFORMATION *telemetry_list;  /* Pointer for telemetry infomation
                                     from database */
    SSM_CALIBRATION_INFORMATION *ssm_calibration_scenes;
                                  /* Pointer for TIRS SSM calibration scenes
                                     from database */
    ENCODER_RECORD *tirs_data;    /* TIRS data records */
    IAS_CPF *cpf;                 /* CPF pointer */
    const IAS_CPF_EARTH_CONSTANTS *earth_const;  /* Earth constant pointer */
    const char *cpf_filename;     /* CPF filename */
    double event_start_time[3];   /* Date for start of switch event */
    double event_end_time[3];     /* Date for end of switch event */
    double encoder_value0;        /* Encoder value count */
    int number_of_records_tel;    /* Number of records return from the 
                                     database for telemetry tables */
    int number_of_records_est;    /* Number of records return from the
                                     database for estimation tables */
    int num_tirs;                 /* Number of tirs records */
    double start_secs;            /* Start time in seconds from the spacecraft
                                     clock J2000 epoch */
    double end_secs;              /* End time in seconds from the spacecraft
                                     clock J2000 epoch */
    double start_day;             /* J2000 start day from epoch */
    double start_sod;             /* J2000 start - second of day from epoch */
    double end_day;               /* J2000 end day from epoch */
    double end_sod;               /* J2000 end sod from epoch */
    int index;                    /* Loop variable */
    int status;                   /* Return status */


    if (ias_log_initialize("generate_ssm_encoder_file") != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing log");
        exit(EXIT_FAILURE);
    }

    /* Make sure an input file exists */
    if (argc != 2)
    {
        IAS_LOG_ERROR("Usage: <odl_file>");
        exit(EXIT_FAILURE);
    }
    
    /* Read parameters */
    if (read_parameters(argv[1]) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading input parameters from %s", argv[1]);
        exit(EXIT_FAILURE);
    }

    IAS_LOG_INFO("Generating ssm encoder file");   

    /* Initiate the satellite attributes */
    ias_sat_attr_initialize(IAS_L8);

    /* Get the cpf filename from odl */
    cpf_filename = get_cpf_filename();
    cpf = ias_cpf_read(cpf_filename);
    if (!cpf)
    {
        IAS_LOG_ERROR("Unable to open CPF: %s", cpf_filename);
        exit(EXIT_FAILURE);
    }
   
    /* Get the earth constants for the leap second information */
    earth_const = ias_cpf_get_earth_const(cpf);
    if (!earth_const)
    {
        IAS_LOG_ERROR("Getting Earth constants from the CPF");
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }

    /* Get the event start and end time from odl */ 
    get_event_start_time(event_start_time);
    get_event_end_time(event_end_time);

    /* Convert the switch event times, which are in UTC, to seconds from
       J2000 */
    if (convert_utc_to_j2000_seconds(&earth_const->leap_seconds_data,
            event_start_time, &start_secs) != SUCCESS)
    {
        IAS_LOG_ERROR("Converting UTC switch event start time (%d:%d:%d) to "
            "J2000 seconds", (int)event_start_time[0], (int)event_start_time[1],
            (int)event_start_time[2]);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }
    if (convert_utc_to_j2000_seconds(&earth_const->leap_seconds_data,
            event_end_time, &end_secs) != SUCCESS)
    {
        IAS_LOG_ERROR("Converting UTC switch event end time (%d:%d:%d) to "
            "J2000 seconds", (int)event_end_time[0], (int)event_end_time[1],
            (int)event_end_time[2]);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }

    /* Convert the seconds from epoch to days from epoch and seconds of
       day from epoch. This function call will not fail so, no status check */
    convert_j2000_seconds_to_days_sod(start_secs, &start_day, &start_sod);
    convert_j2000_seconds_to_days_sod(end_secs, &end_day, &end_sod);

    IAS_LOG_INFO("Start day and seconds of day: %f:%f", start_day, start_sod);
    IAS_LOG_INFO("End day and seconds of day: %f:%f", end_day, end_sod);
    
    /* Create a database connection */
    status = connect_to_database();
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to connect to database");
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }

    /* TIRS SSM telemetry section */    
    /* Get the telemetry information from the database */
    IAS_LOG_INFO("Retrieving telemetry information from database");
    telemetry_list = get_db_telemetry_list(start_day, start_sod,
        end_day, end_sod, &number_of_records_tel);
    if (telemetry_list == NULL)
    {
        IAS_LOG_ERROR("Unable to get telemetry information from database");
        disconnect_from_database();
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }

    IAS_LOG_INFO("Retrieved %d telemetry records", number_of_records_tel);

    /* For each telemetry record returned by the query, compute time from event
       start time by subtracting the event start time from the record time
       stamp from the database */
    for (index = 0; index < number_of_records_tel; index++)
    {
        telemetry_list[index].diff_secs = convert_j2000_days_sod_to_seconds(
            telemetry_list[index].l0r_time_days_from_j2000,
            telemetry_list[index].l0r_time_sec_of_day) - start_secs;
    }

    /* Refine telemetry list */
    get_refined_telemetry_list(telemetry_list, start_secs,
        &number_of_records_tel);

    IAS_LOG_INFO("Telemetry records (refined) %d", number_of_records_tel);

    /* Check the number of refined telemetry records */
    if (number_of_records_tel == 0)
    {
        IAS_LOG_ERROR("No records found in telemetry list");
        disconnect_from_database();
        free(telemetry_list);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }
        
    /* Get the initial encoder value from the odl file */
    encoder_value0 = get_encoder_value0();

     /* Get the nominal nadir encoder value from the telemetry if it is
        available and use it as the encoder_value0.  Note that if we get
        here, there will be telemetry records since no records is considered
        an error at this point, but just in case that changes, we'll check. */
    if (number_of_records_tel >= 1)
    {
        if (get_nadir_encoder_position(cpf, telemetry_list, 
                number_of_records_tel, &encoder_value0) != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating nominal nadir encoder count");
            disconnect_from_database();
            free(telemetry_list);
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }

        /* Fix any upper 12 bits stuck in the telemetry encoder values */
        fix_stuck_ssm_bits(encoder_value0, telemetry_list,
            number_of_records_tel);
    }

    /* TIRS SSM calibration section */

    /* Get the tirs estimation information from database */
    IAS_LOG_INFO("Retrieving TIRS SSM calibration scenes from database");
    if (get_db_ssm_calibration_scenes(event_start_time,
            event_end_time, &ssm_calibration_scenes, &number_of_records_est)
            != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to get ssm estimation scenes from database");
        disconnect_from_database();
        free(telemetry_list);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }

    IAS_LOG_INFO("Retrieved %d calibration records", number_of_records_est);

    /* Close the database connection */
    disconnect_from_database();
 
    /* Calculate net RMSE and refine scenes with cal_outliers_rmse threshold */
    if (get_refined_ssm_calibration_scenes(ssm_calibration_scenes,
            &number_of_records_est) != SUCCESS)
    {
        IAS_LOG_ERROR("Refining the SSM calibration scene list failed");
        free(telemetry_list);
        free(ssm_calibration_scenes);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }

    IAS_LOG_INFO("Calibration records (refined) %d", number_of_records_est);

    /* If there are no SSM calibration records, it can still run, but warn
       about it */
    if (number_of_records_est == 0)
    {
        IAS_LOG_WARNING("No records found in ssm calibration scenes");
    }

    /* For each valid record, calculate the time from switch event start by
       subtracting the switch event start time from the measurement record
       time. */
    for (index = 0; index < number_of_records_est; index++)
    {
        double timestamp[3];
        double j2000_seconds;

        timestamp[0] = ssm_calibration_scenes[index].ssm_epoch_year;
        timestamp[1] = ssm_calibration_scenes[index].ssm_epoch_doy;
        timestamp[2] = ssm_calibration_scenes[index].ssm_epoch_sod;

        if (convert_utc_to_j2000_seconds(&earth_const->leap_seconds_data,
                timestamp, &j2000_seconds) != SUCCESS)
        {
            IAS_LOG_ERROR("Converting SSM calibration record (year, doy, "
                "seconds) to J2000 seconds");
            free(telemetry_list);
            free(ssm_calibration_scenes);
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }

        ssm_calibration_scenes[index].diff_secs = j2000_seconds - start_secs;
    }
    
    /* Calculate the encoder count for tirs ssm calibration data */
    get_encoder_count_tirs_ssm_estimation(ssm_calibration_scenes,
        number_of_records_est, encoder_value0);

   /* num_tirs = total number of valid observations, telemetry plus
      scene-based */
    num_tirs = number_of_records_tel + number_of_records_est;
    if (num_tirs == 0)
    {
        IAS_LOG_ERROR("No tirs ssm records found");
        free(telemetry_list);
        free(ssm_calibration_scenes);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }

    IAS_LOG_INFO("Total TIRS SSM observations %d", num_tirs);

    /* Combined all the records from the db for tirs ssm (telemetry and
       scene-based ssm postion, all the records are already at the time
       when the data was retrived from db*/
    /* Allocate memory fror the tirs data */
    IAS_LOG_INFO("Combine telemetry and ssm based records");
    tirs_data = malloc(sizeof(*tirs_data) * num_tirs);
    if (tirs_data == NULL)
    {
        IAS_LOG_ERROR("Unable to allocate memory for tirs data");
        free(telemetry_list);
        free(ssm_calibration_scenes);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }

    status = get_combined_telemetry_ssm_base_obs(telemetry_list,
        ssm_calibration_scenes, earth_const, number_of_records_tel, 
        number_of_records_est, tirs_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Combining telemetry and ssm scene based records");
        free(telemetry_list);
        free(ssm_calibration_scenes);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }
 
    /* Write a SSM encoder file */
    if (write_encoder_file(get_ssm_encoder_filename(), tirs_data, num_tirs) 
        != SUCCESS)
    {
        IAS_LOG_ERROR("Writing SSM encoder file");
        free(tirs_data);
        exit(EXIT_FAILURE);
    }; 

    /* Free memory */
    free(tirs_data);
    free(telemetry_list);
    free(ssm_calibration_scenes);
    ias_cpf_free(cpf);

    /* Write completion message */
    IAS_LOG_INFO("Successful completion");
    
    return EXIT_SUCCESS;
}
