/*******************************************************************************
NAME:    read_generate_parameters.c

PURPOSE: Reads the input parameters from the parameter file

RETURNS: SUCCESS or ERROR
         EXIT_SUCCESS if user asked for help

Note: Also provides various simple functions to get the parameter values.

******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "read_generate_parameters.h"

/*--- Local static variables that can be retrieved through getter routines ---*/
static char cpf_filename[PATH_MAX];         /* CPF file name */

/* Name of file containing the telemetry and SSM calibration derived
   information */
static char ssm_encoder_filename[PATH_MAX];

static double event_start_time[3];          /* User requested date for start
                                               of table event */
static double event_end_time[3];            /* User requested date for end
                                               of table event */
static double encoder_value0;               /* Initial encoder value to be
                                               used for model */
static double cal_outlier_rmse_threshold;   /* Maximum SSM Calibration
                                               Algorithm RMSE to include in
                                               the fit */
static int telemetry_start_offset;          /* Offset threshold for telemetry
                                               records */
static int telemetry_subsample_threshold;   /* Subsample threshold for
                                               telemetry records */
static int telemetry_subsample_step;        /* Steps count threshold for
                                               telemetry records */

/*----------- read_parameters ------------*/
int read_parameters
(
    const char *parameter_filename   /* I: Name of the parameter file */
)
{
    int status;                     /* Status of return from function */
    int number_of_parameters;       /* Number of parameters to read or read
                                       from parameter files */

    /* -------------------- Define the parameter table -------------------- */
    IAS_PARM_DECLARE_TABLE(parms, 9);

    /* Add the encoder file name */
    const char *default_tirs_filename[] = {"ssm_encoder.csv"};
    IAS_PARM_ADD_STRING(parms, SSM_ENCODER_FILENAME,
         "Filename for the output telemetry SSM encoder positions and SSM "
         "Calibration derived encoder positions", IAS_PARM_OPTIONAL,
        0, NULL, /* no limits */
        1, default_tirs_filename, /* default */
        ssm_encoder_filename, sizeof(ssm_encoder_filename), 1);

    /* Read CPF path */
    IAS_PARM_CPF(parms, IAS_PARM_REQUIRED, cpf_filename, sizeof(cpf_filename),
        1);

    /* Add descritize tolerance */
    IAS_PARM_ADD_DOUBLE(parms, ENCODER_VALUE0,
        "Initial nadir value for the SSM encoder",
        IAS_PARM_REQUIRED, IAS_PARM_NOT_ARRAY,
        0, 0, 0,
        0, NULL, &encoder_value0, sizeof(encoder_value0), 1);

    /* Get event start time to use */
    const double default_event_start_time[] = {0.0, 0.0, 0.0};
    IAS_PARM_ADD_DOUBLE(parms, EVENT_START_TIME,
        "SSM event start time (Year, Day of Year, Seconds of Day)",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
        1, 0.0, 86400.0, 3,
        default_event_start_time, event_start_time, sizeof(event_start_time),
        3);

    /* Get event end time to use */
    const double default_event_end_time[] = {0.0, 0.0, 0.0};
    IAS_PARM_ADD_DOUBLE(parms, EVENT_END_TIME,
        "SSM event end time (Year, Day of Year, Seconds of Day)",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
        1, 0.0, 86400.0, 3,
        default_event_end_time, event_end_time, sizeof(event_end_time),
        3);

    /* Add cal outlier rmse threshold */
    const double default_cal_outlier_rmse_threshold = 35.0;
    IAS_PARM_ADD_DOUBLE(parms, CAL_OUTLIER_RMSE_THRESHOLD,
        "Maximum SSM Calibration Algorithm RMSE to include in the fit",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        0, 0, 0,
        1, &default_cal_outlier_rmse_threshold, &cal_outlier_rmse_threshold,
        sizeof(cal_outlier_rmse_threshold), 1);

    /* Add telemetry start offset */
    int default_telemetry_start_offset = 60;
    IAS_PARM_ADD_INT(parms, TELEMETRY_START_OFFSET,
        "Offset from event start time for telemetry records in seconds",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 0, 0, 0, 1,
        &default_telemetry_start_offset, &telemetry_start_offset,
        sizeof(telemetry_start_offset), 1);

    /* Add telemetry subsample threshold */
    int default_telemetry_subsample_threshold = 100;
    IAS_PARM_ADD_INT(parms, TELEMETRY_SUBSAMPLE_THRESHOLD,
        "Subsample threshold for telemetry records count",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 0, 0, 0, 1,
        &default_telemetry_subsample_threshold, &telemetry_subsample_threshold,
        sizeof(telemetry_subsample_threshold), 1);

    /* Add telemetry subsample step */
    int default_telemetry_subsample_step = 20;
    IAS_PARM_ADD_INT(parms, TELEMETRY_SUBSAMPLE_STEP,
        "Subsample threshold step size for telemetry records",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 0, 0, 0, 1,
        &default_telemetry_subsample_step, &telemetry_subsample_step,
        sizeof(telemetry_subsample_step), 1);

    /* See if the user wanted help or a template */
    status = ias_parm_provide_help(parameter_filename, parms,
        IAS_PARM_GET_TABLE_SIZE(parms), IAS_INPUT_PARAMETERS);
    if (status == 1)
    {
        /* Help provided, so exit */
        exit(0);
    }
    else if (status == ERROR)
    {
        IAS_LOG_ERROR("Returned from parameter file help");
        return ERROR;
    }

    /* Read the parameter file */
    number_of_parameters = IAS_PARM_GET_TABLE_SIZE(parms);
    status = ias_parm_read(parameter_filename, parms, number_of_parameters);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from: %s", parameter_filename);
        return ERROR;
    }

    /* Check on times given.  A value of (0.0,0.0,0.0) for the event_start_time
       means that the start event should be based on the 1st value given in the
       input file.  A value of (0.0,0.0,0.0) for the event_end_time means that
       the end time should be based on the last value given in the input
       file, */
    if (event_start_time[0] != 0.0 || event_start_time[1] != 0.0
        || event_start_time[2] != 0.0)
    {
        /* Some of these checks are unecessary */
        if ((event_start_time[0] < 2000.0 || event_start_time[0] > 2050.0
            || event_start_time[1] < 0.0 || event_start_time[1] > 366.0
            || event_start_time[2] < 0.0
            || event_start_time[2] > IAS_SEC_PER_DAY))
        {
            IAS_LOG_ERROR("Event start time incorrect %f,%f,%f",
                  event_start_time[0], event_start_time[1],
                  event_start_time[2]);
            return ERROR;
        }
    }

    /* Check end time given by user.  If all zeroes then we can base it off of
       the last time of the TIRS SSM calibration data. */
    if (event_end_time[0] != 0.0 || event_end_time[1] != 0.0
        || event_end_time[2] != 0.0)
    {
        /* Some of these checks are unecessary */
        if ((event_end_time[0] < 2000.0 || event_end_time[0] > 2050.0
            || event_end_time[1] < 0.0 || event_end_time[1] > 366.0
            || event_end_time[2] < 0.0
            || event_end_time[2] > IAS_SEC_PER_DAY))
        {
            IAS_LOG_ERROR("Event end time incorrect %f,%f,%f",
                  event_end_time[0], event_end_time[1], event_end_time[2]);
            return ERROR;
        }
    }

    return SUCCESS;
}

/***********************************************************************
    Functions used to access the ODL parameter file values
***********************************************************************/
const char *get_ssm_encoder_filename(void)
{
    return ssm_encoder_filename;
}

const char *get_cpf_filename(void)
{
    return cpf_filename;
}

void get_event_start_time
(
    double *values /* O: Filled array */
)
{
    values[0] = event_start_time[0];
    values[1] = event_start_time[1];
    values[2] = event_start_time[2];
}

void get_event_end_time
(
    double *values /* O: Filled array */
)
{
    values[0] = event_end_time[0];
    values[1] = event_end_time[1];
    values[2] = event_end_time[2];
}

int get_telemetry_start_offset()
{
    return telemetry_start_offset;
}

int get_telemetry_subsample_threshold()
{
    return telemetry_subsample_threshold;
}

int get_telemetry_subsample_step()
{
    return telemetry_subsample_step;
}

double get_encoder_value0()
{
   return encoder_value0;
}

double get_cal_outlier_rmse_threshold()
{
    return cal_outlier_rmse_threshold;
}
