/*******************************************************************************
NAME:    read_parameters.c

PURPOSE: Reads the input parameters from the parameter file

RETURNS: SUCCESS or ERROR
         EXIT_SUCCESS if user asked for help

Note: Also provides various simple functions to get the parameter values.

******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_parm.h"
#include "tirs_fit_ssm_model.h"

/*--- Local static variables that can be retrieved through getter routines ---*/

/* Name of file containing the telemetry and SSM calibration derived
   information */
static char ssm_encoder_filename[PATH_MAX];

static char report_filename[PATH_MAX];      /* Results file name */
static double initial_values[SSM_FIT_NPARMS]; /* Initial model fit parameters */

/* Times which define the number of SSM parameters to be solved for */
static double break_point_times[SSM_BREAK_POINT_TIMES];

/* Time ranges for which a lack of observations make the solution for a set of
   SSM model parameters ill defined */
static double observation_gap_windows[SSM_OBSERVATION_GAP_PARAMS];
                       
/* SSM times minimum number of samples needed */
static int break_times_samples[SSM_OBSERVATION_GAP_WINDOWS];

/* Apriori weights for a number of points, these are the weights that are added
   to the least squares solution when no points are available within the break
   times/ model observation gap windows */
static double apriori_weights[SSM_FIT_NPARMS];

static int coefficient_masks[SSM_FIT_NPARMS];/* Masks for each coefficient */
static double convergence_threshold;  /* Convergence threshold for solution */

/* Weights for coefficient on convergence */
static double convergence_threshold_weights[SSM_FIT_NPARMS];

static double event_start_time[3];  /* User requested date for start
                                       of table event */
static double event_end_time[3];    /* User requested date for end
                                       of table event */
static double encoder_value0;       /* Initial encoder value for model */
static double telemetry_obs_stdev;  /* Standard deviation of the encoder counts
                                       to use to weight the telemetry samples */
static double image_obs_stdev;      /* Standard deviation of the encoder counts
                                       to use to weight the image observation
                                       samples */

static int switch_event_id;         /* Switch event ID */
static int model_id;                /* Model ID */
static int maximum_iterations;      /* Maximum iterations */

/*----------- read_parameters ------------*/
int read_parameters
(
    const char *parameter_filename   /* I: Name of the parameter file */
)
{
    int status;                     /* Status of return from function */
    int number_of_parameters;       /* Number of parameters to read or read
                                       from parameter files */

    /* -------------------- Define the parameter table -------------------- */
    IAS_PARM_DECLARE_TABLE(parms, 18);

    /* Add the encoder file name */
    const char *default_tirs_filename[] = {"ssm_encoder.csv"};
    IAS_PARM_ADD_STRING(parms, SSM_ENCODER_FILENAME,
         "Input file with telemetry SSM encoder values and SSM Calibration "
         "derived encoder positions", IAS_PARM_OPTIONAL,
        0, NULL, /* no limits */
        1, default_tirs_filename, /* default */
        ssm_encoder_filename, sizeof(ssm_encoder_filename), 1);

    /* Read switch event ID */
    IAS_PARM_ADD_INT(parms, SWITCH_EVENT_ID,
        "Switch event ID",
        IAS_PARM_REQUIRED, IAS_PARM_NOT_ARRAY,
        0, 0, 0, /* no limits */
        0, NULL, &switch_event_id, sizeof(switch_event_id), 1);

    /* Read Model ID */
    IAS_PARM_ADD_INT(parms, MODEL_ID,
        "Model ID",
        IAS_PARM_REQUIRED, IAS_PARM_NOT_ARRAY,
        0, 0, 0, /* no limits */
        0, NULL, &model_id, sizeof(model_id), 1);

    /* Add the output file name */
    const char *default_report_filename[] = {"tirs_ssm_model_fit.rpt"};
    IAS_PARM_ADD_STRING(parms, TIRS_RESULTS_FILE_NAME,
        "TIRS Encoder report output file name",
        IAS_PARM_OPTIONAL,
        0, NULL, /* no limits */
        1, default_report_filename, /* default */
        report_filename, sizeof(report_filename), 1);

    /* Add descritize tolerance */
    IAS_PARM_ADD_DOUBLE(parms, ENCODER_VALUE0, "Initial of bias encoder "
         "value for model",
         IAS_PARM_REQUIRED, IAS_PARM_NOT_ARRAY,
         0, 0, 0,
         0, NULL, &encoder_value0, sizeof(encoder_value0), 1);

    /* Add initial model coefficient parameters */
    IAS_PARM_ADD_DOUBLE(parms, INITIAL_VALUES, "List of initial polynomial "
        "coefficients",
        IAS_PARM_REQUIRED, IAS_PARM_ARRAY,
        0, 0, 0, /* no limits */
        0, NULL, initial_values, sizeof(initial_values), SSM_FIT_NPARMS);

    /* Determine needed maximum times allowable for solving for coefficients. */
    const double default_break_point_times[] = {2000.0, 1.0, 7.0, 12.0};
    IAS_PARM_ADD_DOUBLE(parms, BREAK_POINT_TIMES,
        "List of maximum times for coefficients in solution (secs, days, "
        "days, days)",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
        0, 0, 0, /* no limits */
        SSM_BREAK_POINT_TIMES, default_break_point_times, break_point_times,
        sizeof(break_point_times), SSM_BREAK_POINT_TIMES);

    /* Time ranges for which a lack of observations make the solution for a set
       of SSM model parameters ill defined. */
    const double default_observation_gap_windows[] = 
        {2000.0, 6000.0, 6000.0, 86400.0, 1.0, 7.0};
    IAS_PARM_ADD_DOUBLE(parms, OBSERVATION_GAP_WINDOWS, 
        "List of observation gap windows for solution coefficients (secs, "
        "secs, secs, secs, days, days).  Each pair of parameters defines one "
        "window.", IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
        0, 0, 0, /* no limits */
        SSM_OBSERVATION_GAP_PARAMS, default_observation_gap_windows,
        observation_gap_windows, sizeof(observation_gap_windows), 
        SSM_OBSERVATION_GAP_PARAMS);

    const int default_break_times_samples[] = {0, 0, 0};
    IAS_PARM_ADD_INT(parms, BREAK_TIMES_SAMPLES,
        "List of minimum number of points needed per observation gap window",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
        0, 0, 0, /* no limits */
        SSM_OBSERVATION_GAP_WINDOWS, default_break_times_samples, 
        break_times_samples,
        sizeof(break_times_samples),
        SSM_OBSERVATION_GAP_WINDOWS);

    /* Apriori weights ill conditioned               
       a0    a1    a2     a3   tau1   tau2   tau3    S */
    const double default_apriori_weights[] =
        {0.0, 30.0, 50.0,  30.0, 150.0, 0.025,   0.0,   0.0 };
    IAS_PARM_ADD_DOUBLE(parms, APRIORI_WEIGHTS,
        "Aprior weights ill conditioned "
        "(counts, counts, counts, counts, secs, days, days, counts-per-day",
         IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
         0, 0, 0, /* no limits */
         SSM_FIT_NPARMS, default_apriori_weights, apriori_weights,
         sizeof(apriori_weights), SSM_FIT_NPARMS);

    /* Coefficient masks (a0, a1, a2, a3, tau1, tau2, tau3, Slope */
    const int default_coefficient_masks[] = {1, 1, 1, 1, 1, 1, 1, 1};
    IAS_PARM_ADD_INT(parms, COEFFICIENT_MASKS,
            "Masks for each model coefficient, 0 will remove the coefficient "
            "from consideration for the solution", IAS_PARM_OPTIONAL,
            IAS_PARM_ARRAY, 1, 0, 1, /* only allow 0 or 1 */
            SSM_FIT_NPARMS, default_coefficient_masks, coefficient_masks,
            sizeof(coefficient_masks), SSM_FIT_NPARMS);

    /* Add convergence threshold */
    const double default_convergence_threshold = 1.0;
    IAS_PARM_ADD_DOUBLE(parms, CONVERGENCE_THRESHOLD, "Convergence threshold "
            "to determine whether a solution has been found",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
            1, 0.0001, 1000.0,
            1, &default_convergence_threshold, &convergence_threshold,
            sizeof(convergence_threshold), 1);

    /* Weights to apply to SSM coefficients to when determining convergence */
    const double default_convergence_threshold_weights[] 
        /*   a0      a1      a2      a3    tau1      tau2      tau3     Slope */
        = {0.05,   0.05,   0.05,   0.05,   0.05,   0.0005,   0.0005,   0.0005};
    IAS_PARM_ADD_DOUBLE(parms, CONVERGENCE_THRESHOLD_WEIGHTS,
            "Weights for determining convergence (counts, counts, counts, "
            "counts, secs, days, days, counts-per-day",
            IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
            0, 0, 0, /* no limits */
            SSM_FIT_NPARMS, default_convergence_threshold_weights,
            convergence_threshold_weights,
            sizeof(convergence_threshold_weights), SSM_FIT_NPARMS);

    /* Get event start time to use */
    const double default_event_start_time[] = {0.0, 0.0, 0.0};
    IAS_PARM_ADD_DOUBLE(parms, EVENT_START_TIME,
        "SSM event start time",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
        1, 0.0, 86400.0, 3,
        default_event_start_time, event_start_time, sizeof(event_start_time),
        3);

    /* Get event end time to use */
    const double default_event_end_time[] = {0.0, 0.0, 0.0};
    IAS_PARM_ADD_DOUBLE(parms, EVENT_END_TIME,
        "SSM event end time",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
        1, 0.0, 86400.0, 3,
        default_event_end_time, event_end_time, sizeof(event_end_time),
        3);

    /* Add telemetry observation stdev */
    const double default_telemetry_obs_stdev = 4.0;
    IAS_PARM_ADD_DOUBLE(parms, TELEMETRY_OBSERVATION_STDEV,
        "Standard deviation of the encoder counts to use to weight the"
        "telemetry samples",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        0, 0, 0,
        1, &default_telemetry_obs_stdev, &telemetry_obs_stdev, 
        sizeof(telemetry_obs_stdev), 1);
    
    /* Add image observation stdev */
    const double default_image_obs_stdev = 20.0;
    IAS_PARM_ADD_DOUBLE(parms, IMAGE_OBSERVATION_STDEV,
        "Standard deviation of the encoder counts to use to weight the"
        "image observation samples",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        0, 0, 0,
        1, &default_image_obs_stdev, &image_obs_stdev,
        sizeof(image_obs_stdev), 1);

    /* Get the maximum number of simulated annealing iterations */
    int default_maximum_iterations = 50000;
    IAS_PARM_ADD_INT(parms, MAXIMUM_SOLUTION_ITERATIONS,
        "Maximum number of iterations",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 0, 0, 0, 1,
        &default_maximum_iterations, &maximum_iterations,
        sizeof(maximum_iterations), 1);

    /* See if the user wanted help or a template */
    status = ias_parm_provide_help(parameter_filename, parms,
        IAS_PARM_GET_TABLE_SIZE(parms), IAS_INPUT_PARAMETERS);
    if (status == 1)
    {
        /* Help provided, so exit */
        exit(EXIT_SUCCESS);
    }
    else if (status == ERROR)
    {
        IAS_LOG_ERROR("Returned from parameter file help");
        return ERROR;
    }

    /* Read the parameter file */
    number_of_parameters = IAS_PARM_GET_TABLE_SIZE(parms);
    status = ias_parm_read(parameter_filename, parms, number_of_parameters);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from: %s", parameter_filename);
        return ERROR;
    }

    return SUCCESS;
}

/***********************************************************************
    Functions used to access the ODL parameter file values
***********************************************************************/
char *get_ssm_encoder_filename(void)
{
    return ssm_encoder_filename;
}

char *get_report_filename(void)
{
    return report_filename;
}

void get_initial_values
(
    double *values /* O: Filled array */
)
{
    int index;
    for (index = 0; index < SSM_FIT_NPARMS; index++)
        values[index] = initial_values[index];
}

void get_apriori_weights
(
    double *values /* O: Filled array */
)
{
    int index;
    for (index = 0; index < SSM_FIT_NPARMS; index++)
        values[index] = apriori_weights[index];
}   

void get_break_point_times
(
    double *values /* O: Filled array */
)
{
    int index;
    for(index = 0; index < SSM_BREAK_POINT_TIMES; index++)
    {
        if(index >= 1)
            values[index] = (double)IAS_SEC_PER_DAY * break_point_times[index];
        else
            values[index] = break_point_times[index];
    }
}

void get_observation_gap_windows
(
    double *values /* O: Filled array */
)
{
    int index;
    for(index = 0; index < SSM_OBSERVATION_GAP_PARAMS; index++)
    {
        if(index >= 4)
            values[index] = (double)IAS_SEC_PER_DAY * 
                observation_gap_windows[index];
        else
            values[index] = observation_gap_windows[index];
    }
}

void get_break_times_samples
(
    int *values /* O: Filled array */
)
{
    int index;
    for( index = 0; index < SSM_OBSERVATION_GAP_WINDOWS; index++ )
    {
       values[index] = break_times_samples[index];
    }
}

void get_coefficient_masks
(
    int *values /* O: Filled array */
)
{
    int index;
    for (index = 0; index < SSM_FIT_NPARMS; index++)
    {
        values[index] = coefficient_masks[index];
    }
}

void get_event_start_time
(
    double *values /* O: Filled array */
)
{
    values[0] = event_start_time[0];
    values[1] = event_start_time[1];
    values[2] = event_start_time[2];
}

void get_event_end_time
(
    double *values /* O: Filled array */
)
{
    values[0] = event_end_time[0];
    values[1] = event_end_time[1];
    values[2] = event_end_time[2];
}

int get_maximum_iterations()
{
    return maximum_iterations;
}

double get_encoder_value0()
{
   return encoder_value0;
}

double get_telemetry_obs_stdev()
{
    return telemetry_obs_stdev;
}

double get_image_obs_stdev()
{
    return image_obs_stdev;
}

int get_switch_event_id()
{
    return switch_event_id;
}

int get_model_id()
{
    return model_id;
}

double get_convergence_threshold()
{
    return convergence_threshold;
}

void get_convergence_threshold_weights
(
    double *values /* O: Filled array */
)
{
    int index;
    for (index = 0; index < SSM_FIT_NPARMS; index++)
        values[index] = convergence_threshold_weights[index];
}
