/******************************************************************************
NAME: read_ssm_encoder_file

PURPOSE: Read the file with the SSM encoder positions derived from telemetry
    and SSM calibration data.

RETURNS: SUCCESS/ERROR

******************************************************************************/
#include <stdlib.h>
#include <string.h>
#include "ias_const.h"
#include "ias_logging.h"
#include "tirs_fit_ssm_model.h"

#define RESULTS_SIZE 500   /* Max length of a TIRS results line */

int read_ssm_encoder_file
(
    const char *tirs_data_filename,  /* I: TIRS data file name */
    ENCODER_RECORD **tirs_data,      /* O: TIRS data records */
    int *num_tirs_encoder_values     /* O: Number of TIRS read */
)
{
    ENCODER_RECORD tirs_tmp;         /* Temporary TIRS data record */
    ENCODER_RECORD *tirs_tmp_ptr;    /* Temporary TIRS data pointer */
    FILE *tirs_file_ptr;             /* TIRS data file pointer */
    char dataline[RESULTS_SIZE];     /* Line from TIRS data file */
    char *string_pointer;            /* Pointer for thread-safe strtok_r */
    char *token;                     /* Pointer to token string used
                                        by strtok_r */
    int line_count = 0;              /* Current line being processed */
    int parse_error_count = 0;       /* Parsing error counter */
    int current_tirs;                /* Current TIRS encoder value */
    int num_records;                 /* Number of tirs records */
    int status;                      /* Return status */
    char ssm_type[10];

    /* Initialize the number of TIRS read and the array of structures */
    *num_tirs_encoder_values = 0;
    current_tirs = 0;

    /* Open the file */
    tirs_file_ptr = fopen(tirs_data_filename, "r");
    if (tirs_file_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening TIRS file %s", tirs_data_filename);
        return ERROR;
    }

    num_records = 0;
    while (fgets(dataline, RESULTS_SIZE, tirs_file_ptr) != NULL)
       num_records++;
    num_records--;
    IAS_LOG_INFO("Number of TIRS records %d", num_records);
    *num_tirs_encoder_values = num_records;

    /* Allocate INITIAL_RECORDS records worth of memory to start with */
    *tirs_data = malloc(sizeof(**tirs_data) * num_records);
    if (*tirs_data == NULL)
    {
        IAS_LOG_ERROR("Allocating initial memory");
        return ERROR;
    }

    /* Get the file pointer at the beginning of file */
    if (fseek(tirs_file_ptr, 0L, SEEK_SET) != 0)
    {
        IAS_LOG_ERROR("Cannot determine TIRS file size");
        fclose(tirs_file_ptr);
        return ERROR;
    }

    /* Read header field */
    fgets(dataline, RESULTS_SIZE, tirs_file_ptr);

    /* Read until NULL is returned by fgets */
    while (fgets(dataline, RESULTS_SIZE, tirs_file_ptr) != NULL)
    {
        if (strlen(dataline) == 0)
            break;
        line_count++;
        memset(&tirs_tmp, 0, sizeof(tirs_tmp));

        /* Replace newline character with NULL */
        int dataline_length = strlen(dataline);
        if (dataline[dataline_length - 1] == '\n') 
        {
            dataline[dataline_length - 1] = '\0';
        }

        /* Verify there are no empty fields (two commas together)
           because strtok doesn't handle that as desired */
        token = strstr(dataline, ",,");
        if (token != NULL)
        {
            IAS_LOG_ERROR("A field value is missing");
            parse_error_count++;
        }

        token = strtok_r(dataline, ",", &string_pointer);
        if (token == NULL)
        {
            parse_error_count++;
        }
        else
        {
            status = snprintf(tirs_tmp.date, sizeof(tirs_tmp.date), "%s",
                        token);
            if (status < 0 || status >= sizeof(tirs_tmp.date))
            {
                parse_error_count++;
            }
        }
        token = strtok_r(NULL, ",", &string_pointer);
        if (token == NULL)
        {
            parse_error_count++;
        }
        else
        {
            status = sscanf(token, "%d", &tirs_tmp.secs_diff);
            if (status != 1) /* expected to get a value */
            {
                parse_error_count++;
            }
        }
        token = strtok_r(NULL, ",", &string_pointer);
        if (token == NULL)
        {
            parse_error_count++;
        }
        else
        {
            status = sscanf(token, "%s", ssm_type);
            if( strcmp(ssm_type,"Image") == 0 )
            {
                tirs_tmp.ssm_type = SSM_IMAGE;
            }
            else if( strcmp(ssm_type,"Telemetry") == 0 )
            {
                tirs_tmp.ssm_type = SSM_TELEMETRY;
            }
            else
            {
                IAS_LOG_ERROR("Reading TIRS encoder type %s", ssm_type);
                /* Close file and free data */
                fclose(tirs_file_ptr);
                free(*tirs_data);
                *tirs_data = NULL;
                return ERROR;
            }
            if (status != 1) /* expected to get a value */
            {
                parse_error_count++;
            }
        }
        token = strtok_r(NULL, ",", &string_pointer);
        if (token == NULL)
        {
            parse_error_count++;
        }
        else
        {
            status = sscanf(token, "%d", &tirs_tmp.encoder_count);
            if (status != 1) /* expected to get a value */
            {
                parse_error_count++;
            }
        }

        if (parse_error_count != 0)
        {
            IAS_LOG_ERROR("Reading TIRS data from line %d of %s", line_count,
                    tirs_data_filename);
            /* Close file and free data */
            fclose(tirs_file_ptr);
            free(*tirs_data);
            *tirs_data = NULL;
            return ERROR;
        }

        (*tirs_data)[current_tirs] = tirs_tmp;
        current_tirs++;
    }

    /* No TIRS encoder were found. Free up any memory that was allocated and
       return */
    if (*num_tirs_encoder_values == 0)
    {
        free(*tirs_data);
        *tirs_data = NULL;
        fclose(tirs_file_ptr);
        return SUCCESS;
    }

    /* Realloc memory to correct size */
    tirs_tmp_ptr = realloc(*tirs_data, *num_tirs_encoder_values
            * sizeof(**tirs_data));
    if (tirs_tmp_ptr == NULL)
    {
        IAS_LOG_ERROR("Reallocating memory for %d encoder values",
            *num_tirs_encoder_values);
        fclose(tirs_file_ptr);
        free(*tirs_data);
        *tirs_data = NULL;
        *num_tirs_encoder_values = 0;
        return ERROR;
    }
    *tirs_data = tirs_tmp_ptr;

    /* Close the file */
    status = fclose(tirs_file_ptr);
    if (status != 0)
    {
        IAS_LOG_ERROR("Closing TIRS results file %s", tirs_data_filename);
        free(*tirs_data);
        *tirs_data = NULL;
        *num_tirs_encoder_values = 0;
        return ERROR;
    }

    return SUCCESS;
}
