#!/bin/csh -f

# NAME: test_generate_ssm_encoder.csh
#
# PURPOSE: Set up the environment to generate intermediate ssm encoder file
# then run the test. It will be run in a command line mode that exercise
# database I/O and intermediate ssm ecnoder file creation.
#
# RETURNS: 0 on SUCCESS
#          1 on ERROR

echo "Running test"
# Ensure the test environment variable is set.
if ( ! $?IAS_UNIT_TEST_DATA_DIR ) then
    echo "IAS_UNIT_TEST_DATA_DIR isn't set"
endif

# Set up the environment for the test.
set application = "generate_ssm_encoder_file"
set application_log = test_generate_ssm_encoder_file.log
set test_dir = "$IAS_UNIT_TEST_DATA_DIR/ias/gps/tirs_ssm_model/tirs_fit_ssm_model/${application}"

# Check whether database testing has been enabled. If it isn't enabled,
# this test can't be run, as it loads data into the database at start with.
if ( ! $?ENABLE_DB_TESTS ) then
    echo
    echo "Cannot run testing, ENABLE_DB_TESTS environment variable isn't set"
    echo
    exit 0
endif

# Sanity check to make sure the database user ID/password are set in the
# environment (in the IAS_DB_L8_OLITIRS environment variable)
if ( ! $?IAS_DB_L8_OLITIRS ) then
    echo "IAS_DB_L8_OLITIRS is not currently set"
    exit 1
endif

# Clean up the database
echo "Cleaning up database"
./clean_data.csh  > $application_log
if ( $status != 0 ) then
    echo "Failed to clean data in database"
    exit 1
endif

# Set up test data in database.
echo "Loading test data into database"
./load_data.csh >> $application_log
if ( $status != 0 ) then
    echo "Failed to load data in database"
    exit 1
endif

echo "Running generate_ssm_encoder_file"
../generate_ssm_encoder_file TIRS_SSM_MODEL_FIT_1.odl  >> $application_log
if ( $status != 0 ) then
    echo "Failed to run generate ssm encoder application"
    exit 1
endif

../generate_ssm_encoder_file TIRS_SSM_MODEL_FIT_2.odl  >> $application_log
if ( $status != 0 ) then
    echo "Failed to run generate ssm encoder application"
    exit 1
endif

# Compare the ssm encoder file to the unit test version.  Any difference will be
# treated as a failure.
echo "Comparing results"
echo "Comparing file ssm_encoder_1.csv to $test_dir/ssm_encoder_1.csv."
diff ${test_dir}/ssm_encoder_1.csv ssm_encoder_1.csv > ssm_encoder_1.csv_diffs
if ( $status != 0 ) then
    echo "ssm_encoder_1.csv and $test_dir/ssm_encoder_1.csv files do not match."
    exit 1
endif

echo "Comparing file ssm_encoder_2.csv to $test_dir/ssm_encoder_2.csv."
diff ${test_dir}/ssm_encoder_2.csv ssm_encoder_2.csv > ssm_encoder_2.csv_diffs
if ( $status != 0 ) then
    echo "ssm_encoder_2.csv and $test_dir/ssm_encoder_2.csv files do not match."
    exit 1
endif

# Clean up the database
echo "Cleaning up database"
./clean_data.csh  >> $application_log
if ( $status != 0 ) then
    echo "Failed to clean data in database"
    exit 1
endif

# Clean up some files.
rm -f *.csv *.csv_diffs *_TABLE.log

# Exit with success.
exit 0
