/****************************************************************************
NAME:       tirs_fit_ssm_model

PURPOSE:    
Main driver that retrieves input parameters, calls the fitting function, and
creates the table needed for product generation.

RETURN VALUE:
Type=int
Value           Description
-----           -----------
EXIT_SUCCESS    tirs_fit_ssm_model completed successfully
EXIT_FAILURE    Error occurred while running tirs_fit_ssm_model

*****************************************************************************/
#include <stdlib.h>
#include "ias_const.h"
#include "ias_logging.h"
#include "tirs_fit_ssm_model.h"


int main(int argc, char *argv[])
{
    ENCODER_RECORD *tirs_data;      /* TIRS data records */
    double final_rms[RMSE_COUNT];   /* Final fit rmse */
    int final_num[RMSE_COUNT];      /* Final fit number */
    double model_output[SSM_FIT_NPARMS];/* Output model */
    double encoder_value0;          /* TIRS encoder origin 
                                       value */
    int switch_event_id;            /* Switch event ID */
    int model_id;                   /* Model ID */
    double event_start_time[3];     /* Date for start of switch event */
    double event_end_time[3];       /* Date for end of switch event */
    int number_of_records;          /* Number of records from
                                       ssm encoder file */


    if (ias_log_initialize("tirs_fit_ssm_model") != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing log");
        exit(EXIT_FAILURE);
    }

    /* Make sure an input file exists */
    if (argc != 2)
    {
        IAS_LOG_ERROR("Usage: tirs_fit_ssm_model <odl_file>");
        exit(EXIT_FAILURE);
    }

    if (read_parameters(argv[1]) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading input parameters from %s", argv[1] );
        exit(EXIT_FAILURE);
    }

    /* Read the SSM encoder file */
    if (read_ssm_encoder_file(get_ssm_encoder_filename(), &tirs_data,
            &number_of_records) != SUCCESS)
    {
        IAS_LOG_ERROR("Error reading SSM encoder data file %s",
            get_ssm_encoder_filename());
        exit(EXIT_FAILURE);
    }

    /* Get the initial encoder value from the odl file */
    encoder_value0 = get_encoder_value0();

    /* Perform non_linear lsq fit model fit */
    IAS_LOG_INFO("Performing IAS non-linear lsq for model fit");
    if (fit_lsq_ias(tirs_data, number_of_records, encoder_value0, model_output,
        final_num, final_rms) != SUCCESS )
    {
        IAS_LOG_ERROR("Performing fit model");
        free(tirs_data);
        exit(EXIT_FAILURE);
    }

    /* Get the event start and end time from odl */
    get_event_start_time(event_start_time);
    get_event_end_time(event_end_time);

    /* Get the different values from the odl file to add in the report file */
    switch_event_id = get_switch_event_id();
    model_id = get_model_id();
   
    /* Create a report file */
    if (create_report(tirs_data, number_of_records, encoder_value0,
        switch_event_id, model_id, model_output, final_num, final_rms,
        event_start_time, event_end_time) != SUCCESS)
    {
        IAS_LOG_ERROR("Error writing report");
        free(tirs_data);
        exit(EXIT_FAILURE);
    }

    free(tirs_data);
    tirs_data = NULL;

    /* Write completion message */
    IAS_LOG_INFO("Successful completion");

    return EXIT_SUCCESS;
}
