/******************************************************************************
NAME:       utilities

PURPOSE:
Different function for different purpose.

RETURNS: Various

******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "ias_logging.h"
#include "read_generate_parameters.h"
#include "utilities.h"

#define TIRS_SSM_OVERFLOW 0x01000000

/******************************************************************************
Name: convert_utc_to_j2000_seconds

Purpose: Convert a UTC date/time to J2000 seconds.  The number of leap seconds
    is initialized on each call and cleared to allow for a time period crossing
    a leap second.

Returns: SUCCESS or ERROR
******************************************************************************/
int convert_utc_to_j2000_seconds
(
    const IAS_MATH_LEAP_SECONDS_DATA *leap_seconds_data, /* Leap second info */
    double utc_time[3],     /* I: Array with the UTC time */
    double *j2000_seconds   /* O: output J2000 seconds */
)
{
    if (ias_math_init_leap_seconds_from_UTC_time(utc_time, leap_seconds_data)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing math library leap seconds");
        return ERROR;
    }

    /* Convert the date to J2000 */
    if (ias_math_convert_year_doy_sod_to_j2000_seconds(utc_time,
            j2000_seconds) != SUCCESS)
    {
        IAS_LOG_ERROR("Converting UTC date to J2000 seconds");
        return ERROR;
    }

    /* Clear the leap seconds since the next time a conversion happens it might
       have a different number of leap seconds */
    ias_math_clear_leap_seconds();

    return SUCCESS;
}

/******************************************************************************
NAME:    convert_j2000_seconds_to_days_sod

PURPOSE: Converts the seconds from epoch to days from epoch and seconds of day
         from epoch.

RETURNS: void

******************************************************************************/
void convert_j2000_seconds_to_days_sod
(
    double seconds,     /* I: J2000 time in seconds */
    double *day,        /* O: J2000 start day */
    double *sod         /* O: J2000 start second of day */
)
{
    *day = (int)seconds / IAS_SEC_PER_DAY;
    *sod = (int)seconds % IAS_SEC_PER_DAY;
}

/******************************************************************************
NAME: convert_j2000_days_sod_to_seconds

PURPOSE: Converts the J2000 days and seconds of day to J2000 seconds.

RETURNS: Total second of day
******************************************************************************/
double convert_j2000_days_sod_to_seconds
(
    double day_J2000,  /* I: Time days from J2000*/
    double sod         /* I: Time second of day */
)
{
    return day_J2000 * IAS_SEC_PER_DAY + sod;
}

/******************************************************************************
Name: get_refined_telemetry_list

Purpose: Refine the telemery samples that are less than 60 seconds from the
    switch event start time and if more than 100 records were returned, retain
    only records where the time from the switch event is an integer multiple of
    20 seconds.

RETURNS: Nothing.  The telemetry list is updated in place and the number of
    records is updated.
*******************************************************************************/
void get_refined_telemetry_list
(
    TELEMETRY_INFORMATION *telemetry_list, /* I/O: Telemetry list pointer */
    double start_event_time_sec,           /* I: Event start time in sec */
    int *number_of_records                 /* I/O: Number of records */
)
{
    int index = 0;
    int new_index = 0;
    double telemetry_time_stamp_sec;
    int telemetry_start_offset;        /* Default value: 60 */
    int telemetry_subsample_threshold; /* Default value: 100 */
    int telemetry_subsample_step;      /* Default value: 20 */

    /* Get the values from the odl */
    telemetry_start_offset = get_telemetry_start_offset();
    telemetry_subsample_threshold = get_telemetry_subsample_threshold();
    telemetry_subsample_step = get_telemetry_subsample_step();

    while (index < *number_of_records)
    {
        TELEMETRY_INFORMATION *rec = &telemetry_list[index];

        telemetry_time_stamp_sec = convert_j2000_days_sod_to_seconds(
            rec->l0r_time_days_from_j2000, rec->l0r_time_sec_of_day);

        /* Delete all samples that are less than telemetry_start_offset
           (default 60 sec) from the switch event start time in sec */
        if ((telemetry_time_stamp_sec - start_event_time_sec)
                < telemetry_start_offset)
        {
           /* Do not do anything */
           ;
        }
        else
        {
            /* Move the record to the next index that is being kept.  The
               unwanted records are eliminated by overwriting them. */
            memmove(&telemetry_list[new_index], rec, sizeof(*rec));
            new_index++;
        }
        index++;
    }

    /* Get the total records */
    *number_of_records = new_index;

    /* If more than telemetry_subsample_threshold (default 100) records were
       returned, retain only records where the time from the switch event is an
       integer multiple of telemetry_subsample_step (default 20 secs) */

    if (*number_of_records > telemetry_subsample_threshold)
    {
        index = 0;
        new_index = 0;

        while (index < *number_of_records)
        {
            TELEMETRY_INFORMATION *rec = &telemetry_list[index];
            double j2000_seconds = convert_j2000_days_sod_to_seconds(
                    rec->l0r_time_days_from_j2000, rec->l0r_time_sec_of_day);

            /* If the time from the start is a multiple of the subsampling
               step, keep the record */
            if (((int)(j2000_seconds - start_event_time_sec)
                % telemetry_subsample_step) == 0)
            {
                /* Keeping the record, so move it to the next index of the
                   condensed list. */
                memmove(&telemetry_list[new_index], rec, sizeof(*rec));
                new_index++;
            }
            index++;
        }

        /* Update the number of records */
        *number_of_records = new_index;
    }
}

/******************************************************************************
NAME:      get_nadir_encoder_position

PURPOSE: Use the telemetry records to determine whether side A or B electronics
    are being used and get the nominal nadir encoder value for that side.

RETURNS: SUCCESS if it gets the new encoder value
         ERROR if it fails to get the new encoder value
*******************************************************************************/
int get_nadir_encoder_position
(
    IAS_CPF *cpf,           /* I: Pointer to an open CPF */
    TELEMETRY_INFORMATION *telemetry_list,  /* I: Telemetry list */
    int number_of_records,  /* I: Number of records */
    double *new_encoder     /* O: New encoder value */
)
{
    const struct IAS_CPF_TIRS_PARAMETERS *tirs_parameters; /* Pointer for
                               TIRS CPF PARAMETERS */
    int side_count = 0;     /* If the side count is >= 0, it is side A */
    double nadir_angle;
    double rough_nadir_count;
    int nadir_encoder;
    int index;

    /* Read a tirs parameters from CPF */
    tirs_parameters = ias_cpf_get_tirs_parameters(cpf);
    if (!tirs_parameters)
    {
        IAS_LOG_ERROR("Unable to read tirs parameters from CPF");
        return ERROR;
    }

    /* Count the number of side A and side B indications in the telemetry */
    for (index = 0; index < number_of_records; index++)
    {
        if (telemetry_list[index].elec_enabled_flags & 0x01)
            side_count++;
        else if (telemetry_list[index].elec_enabled_flags & 0x02)
            side_count--;
    }
    if (side_count >= 0)
    {
        IAS_LOG_INFO("Using side A electronics");
        nadir_angle = tirs_parameters->ssm_encoder_origin_side_a;
    }
    else
    {
        IAS_LOG_INFO("Using side B electronics");
        nadir_angle = tirs_parameters->ssm_encoder_origin_side_b;
    }

    /* Convert TIRS SSM nadir angle (in radians) from the CPF to units of
       encoder counts */
    rough_nadir_count = nadir_angle * TIRS_SSM_OVERFLOW
        / (2 * ias_math_get_pi());

    /* Round the encoder count to the nearest integer */
    nadir_encoder = (int)floor(rough_nadir_count + 0.5);
    if (nadir_encoder >= TIRS_SSM_OVERFLOW)
        nadir_encoder -= TIRS_SSM_OVERFLOW;

    IAS_LOG_INFO("Encoder counts %d", nadir_encoder);

    *new_encoder = nadir_encoder;

    return SUCCESS;
}

/******************************************************************************
NAME:      fix_stuck_ssm_bits

PURPOSE: Fix the upper 12 bits of the telemetry SSM encoder values if they
    are stuck.

RETURNS: nothing

*******************************************************************************/
void fix_stuck_ssm_bits
(
    int nadir_encoder,      /* I: Nadir encoder count */
    TELEMETRY_INFORMATION *telemetry_list,  /* I/O: Telemetry list */
    int number_of_records   /* I: Number of records */
)
{
    int index;
    int high_12 = nadir_encoder & 0x00FFF000;
    int stuck_count = 0;

    for (index = 0; index < number_of_records; index++)
    {
        TELEMETRY_INFORMATION *rec = &telemetry_list[index];
        int sample = (int)rec->ssm_encoder_position_sample_2;

        /* If the upper 12 bits of the 24 bit counter are all set, the bits
           are stuck, so replace the upper 12 bits with the nadir value */
        if ((sample & 0x00FFF000) == 0x00FFF000)
        {
            rec->ssm_encoder_position_sample_2 = high_12 | (sample & 0x0FFF);
            stuck_count++;
        }
    }
    if (stuck_count)
    {
        IAS_LOG_WARNING("Fixed %d telemetry encoder values with the upper "
            "12 bits stuck", stuck_count);
    }
}

/******************************************************************************
Name: get_refined_ssm_calibration_scenes

Purpose: Calculate net RMSE for the tirs ssm scenes from the tirs ssm
    estimation tables and remove records where the RSME exceeds the threshold.

Returns: SUCCESS/ERROR
******************************************************************************/
int get_refined_ssm_calibration_scenes
(
    SSM_CALIBRATION_INFORMATION *ssm_calibration_scenes,
                                 /* I/O: SSM calibration scene pointer */
    int *number_of_records_est   /* I/O: Number of records return from db on
                                         input, reduced record count on
                                         return */
)
{
    int index;
    int new_index;

    /* Get cal_outlier from odl */
    double cal_outlier_rmse;
    cal_outlier_rmse = get_cal_outlier_rmse_threshold();

    /* Calculate net rmse for each scene */
    /* Each scene has 3 records */
    index = 0;
    new_index = 0;
    while (index < *number_of_records_est)
    {
        SSM_CALIBRATION_INFORMATION *rec = &ssm_calibration_scenes[index];

        if (rec->number_of_scas == 3)
        {
            int sca_index;
            double net_rmse = 0.0;
            int total_points = 0;

            /* Merge the three separate SCA records into an overall record */
            for (sca_index = 0; sca_index < 3; sca_index++)
            {
                double x = rec[sca_index].post_fit_alng_trk_rmse;
                double y = rec[sca_index].post_fit_acrss_trk_rmse;

                /* Make sure we are in sync with the SCAs in the data */
                if (rec[sca_index].sca_number != (sca_index + 1))
                {
                    IAS_LOG_ERROR("Unexpected SCA number found when merging "
                        "SCA calibration data records at timestamp (%d:%d:%d",
                        (int)rec[sca_index].ssm_epoch_year,
                        (int)rec[sca_index].ssm_epoch_doy,
                        (int)rec[sca_index].ssm_epoch_sod);
                    return ERROR;
                }

                net_rmse += (x * x + y * y) * rec[sca_index].nbr_of_pnts_used;
                total_points += rec[sca_index].nbr_of_pnts_used;
            }
            net_rmse /= (total_points * 2);
            net_rmse = sqrt(net_rmse);

            /* If the RMSE is under the threshold, keep it */
            if (net_rmse < cal_outlier_rmse)
            {
                /* If the timestamp is the same as the previous record, which
                   can happen if the same scene is run multiple times,
                   overwrite the previous record to keep the last run of the
                   data.  This works due to how the results of the database
                   query are ordered. */
                if (new_index > 0 
                    && ssm_calibration_scenes[new_index - 1].ssm_epoch_year
                        == rec[0].ssm_epoch_year
                    && ssm_calibration_scenes[new_index - 1].ssm_epoch_doy
                        == rec[0].ssm_epoch_doy
                    && ssm_calibration_scenes[new_index - 1].ssm_epoch_sod
                        == rec[0].ssm_epoch_sod)
                {
                    /* Decrement the index to overwrite the previous record */
                    new_index--;
                }

                /* Copy the first source record using memmove since the target
                   and source memory might be the same */
                memmove(&ssm_calibration_scenes[new_index], rec, sizeof(*rec));
                ssm_calibration_scenes[new_index].net_rmse = net_rmse;
                new_index++;
            }
        }
        else if (rec->number_of_scas < 1 || rec->number_of_scas > 3)
        {
            /* Invalid data present, so error out */
            IAS_LOG_ERROR("Invalid number of SCAs (%d) retrieved from the "
                "database at time (%d:%d:%d)", rec->number_of_scas,
                (int)rec->ssm_epoch_year, (int)rec->ssm_epoch_doy,
                (int)rec->ssm_epoch_sod);
        }
        else
        {
            IAS_LOG_WARNING("Encountered record that did not provide data for "
                "3 SCAs - skipped");
        }

        /* Advance to the next set of SCA records */
        index += rec->number_of_scas;
    }

    /* Update the number of records */
    *number_of_records_est = new_index;

    return SUCCESS;
}

/******************************************************************************
NAME: compare

PURPOSE: Compare function for calling qsort on an ENCODER_RECORD list.

RETURNS: -1 if the value is less
          1 if the value is greater
          0 if the values are equal

******************************************************************************/
static int compare
(
    const void *element1, /* I: Void pointer */
    const void *element2  /* I: Void pointer */
)
{
    ENCODER_RECORD *f = (ENCODER_RECORD*)element1;
    ENCODER_RECORD *s = (ENCODER_RECORD*)element2;
    if (f->secs_diff < s->secs_diff)
        return -1;
    if (f->secs_diff > s->secs_diff)
        return 1;
    return 0;
}

/******************************************************************************
Name: get_combined_telemetry_ssm_based_obs

Purpose: Converts telemetry l0r_time_days_from_j2000 and l0r_time_sec_of_day to
    J2000 seconds and Combines the valid telemetry and tirs ssm based
    observations.

RETURNS: SUCCESS/ERROR
******************************************************************************/
int get_combined_telemetry_ssm_base_obs
(
    const TELEMETRY_INFORMATION *telemetry_list,
                                  /* I: Telemetry pointer */
    const SSM_CALIBRATION_INFORMATION *ssm_calibration_list,
                                  /* I: SSM calibraton scenes pointer */
    const struct IAS_CPF_EARTH_CONSTANTS *earth_const, /* I: Earth Constant */
    int number_of_records,        /* I:  Telemetry records */
    int number_of_records_est,    /* I:  SSM calibration scene records */
    ENCODER_RECORD *tirs_data     /* O:  Combined telemetry and tirs ssm
                                         base observation pointer */
)
{   int index;
    int status;
    double total_j2000_secs;
    double time[3];

    int numtirs = number_of_records + number_of_records_est;
    for (index = 0; index < number_of_records; index++)
    {

        /* Convert l0r_time_days_from_j2000 and l0r_time_sec_of_day of tirs
           ssm estimation to J2000 seconds */
        total_j2000_secs = convert_j2000_days_sod_to_seconds(
            telemetry_list[index].l0r_time_days_from_j2000,
            telemetry_list[index].l0r_time_sec_of_day);

        status = ias_math_init_leap_seconds(total_j2000_secs,
            &earth_const->leap_seconds_data);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Initializing leap seconds");
            return ERROR;
        }

        /* Convert total J2000 seconds to year, doy and sod */
        status = ias_math_convert_j2000_seconds_to_year_doy_sod(
            total_j2000_secs, time);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Converting J2000 seconds to year, doy and sod");
            return ERROR;
        }

        /* Clear the leap seconds since the next time a conversion happens it
           might have a different number of leap seconds */
        ias_math_clear_leap_seconds();

        tirs_data[index].year = time[0];
        tirs_data[index].doy = time[1];
        tirs_data[index].sod = time[2];
        tirs_data[index].secs_diff = telemetry_list[index].diff_secs;
        tirs_data[index].ssm_type = SSM_TELEMETRY;
        tirs_data[index].encoder_count =
            (int)telemetry_list[index].ssm_encoder_position_sample_2;
    }

    for (index = number_of_records; index < numtirs; index++)
    {
        tirs_data[index].year =
            ssm_calibration_list[index - number_of_records].ssm_epoch_year;
        tirs_data[index].doy =
            ssm_calibration_list[index - number_of_records].ssm_epoch_doy;
        tirs_data[index].sod =
            ssm_calibration_list[index - number_of_records].ssm_epoch_sod;
        tirs_data[index].secs_diff =
            ssm_calibration_list[index - number_of_records].diff_secs;
        tirs_data[index].ssm_type = SSM_IMAGE;
        tirs_data[index].encoder_count =
            (int)ssm_calibration_list[index - number_of_records].encoder_counts;
    }

    /* Sort the combined telemetry and tirs ssm scene based records
       according to date/time (year, doy, sod) */
    qsort(tirs_data, numtirs, sizeof(*tirs_data), compare);

    return SUCCESS;
}

/******************************************************************************
NAME:       get_encoder_count_tirs_estimation

PURPOSE:
Converts net_ssm_angle to the encoder count for tirs ssm estimation.
The side A and B electronics are determined from the telemetry database
table.

RETURNS: encoder count for tirs ssm estimation

******************************************************************************/
void get_encoder_count_tirs_ssm_estimation
(
    SSM_CALIBRATION_INFORMATION *refined_ssm_calibration_scenes,
                              /* I/O: Pointer for ssm calibration information */
    int number_of_records_est,  /* I: Number of records */
    int encoder_origin          /* I: nadir encoder position in counts */
)
{
    double pi;
    double ssm_scale;
    double ssm_encoder_counts_calc;
    double nadir_angle;
    int ssm_encoder_counts;
    int index;

    /* Get the value of pi */
    pi = ias_math_get_pi();

    /* Calculate scaling factor */
    ssm_scale = (double)TIRS_SSM_OVERFLOW / (2.0 * pi);

    /* Convert the nadir counts to an angle */
    nadir_angle = encoder_origin / ssm_scale;

    for (index = 0; index < number_of_records_est; index++)
    {
        ssm_encoder_counts_calc = nadir_angle
            + refined_ssm_calibration_scenes[index].net_ssm_angle;
        while (ssm_encoder_counts_calc < 0.0)
            ssm_encoder_counts_calc += 2.0 * pi;
        while (ssm_encoder_counts_calc >= (2.0 * pi))
            ssm_encoder_counts_calc -= 2.0 * pi;
        ssm_encoder_counts_calc *= ssm_scale;

        /* Round the encoder count to the nearest integer */
        ssm_encoder_counts = (int)floor(ssm_encoder_counts_calc + 0.5);
        if (ssm_encoder_counts >= TIRS_SSM_OVERFLOW)
            ssm_encoder_counts -= TIRS_SSM_OVERFLOW;

        refined_ssm_calibration_scenes[index].encoder_counts =
            ssm_encoder_counts;
    }
}


