/******************************************************************************
NAME:       write_encoder_file

PURPOSE:
Write tirs ssm encoder data to a file.

RETURNS:
SUCCESS if tirs ssm encoder file is created
ERROR if tirs ssm encoder file is not created

*******************************************************************************/
#include <stdio.h>
#include <string.h>

#include "ias_const.h"
#include "ias_logging.h"
#include "ssm_telemetry.h"

int write_encoder_file
(
    const char *ssm_encoder_filename,   /* I/O: TIRS SSM data filename */
    const ENCODER_RECORD *results,      /* I: TIRS SSM data pointer */
    int num_tirs                        /* I: Number of TIRS SSM records */
)
{
    IAS_LOG_INFO("Writing SSM encoder file");
    int index;
    FILE *file;

    /* Open file to write */
    file = fopen(ssm_encoder_filename, "w");
    if (file == NULL)
    {
        IAS_LOG_ERROR("Unable to create file %s", ssm_encoder_filename);
        return ERROR;
    }

    /* Write a header line to the file */
    if (fprintf(file, "Date/Time,Seconds Diff,Type,Abs Count\n") < 0)
    {
        IAS_LOG_ERROR("Failed writing record to SSM encoder file %s",
            ssm_encoder_filename);
        fclose(file);
        return ERROR;
    }

    /* Write all the records to the file */
    for (index = 0; index < num_tirs; index++)
    {
        const char *ssm_encoder_type;
        const ENCODER_RECORD *tirs_data = &results[index];

        /* Convert the record type to Image/Telemetry */
        if (tirs_data->ssm_type == 0)
            ssm_encoder_type = "Image";
        else if (tirs_data->ssm_type == 1)
            ssm_encoder_type =  "Telemetry";
        else
        {
            IAS_LOG_ERROR("SSM encoder type %d not found", tirs_data->ssm_type);
            fclose(file);
            return ERROR;
        }
        if (fprintf(file, "%04d:%03d:%05d,%d,%s,%d\n",
                (int) tirs_data->year,
                (int) tirs_data->doy,
                (int) tirs_data->sod,
                tirs_data->secs_diff,
                ssm_encoder_type,
                tirs_data->encoder_count) < 0)
        {
            IAS_LOG_ERROR("Failed writing record to SSM encoder file %s",
                ssm_encoder_filename);
            fclose(file);
            return ERROR;
        }
    }

    if (fclose(file) != 0)
    {
        IAS_LOG_ERROR("Failed closing the SSM encoder file %s",
            ssm_encoder_filename);
        return ERROR;
    }

    return SUCCESS;
}
