#!/usr/bin/perl 
################################################################################
# NAME: create_gps_work_orders.pl
#
# PURPOSE: This script is to help automate work order creation. While it may be
#   used with any procedure, it was created with TIRS SSM Calibration in mind.
#   The scenes for these work orders may be obtained either from a scene list
#   file, which contains each scene to be loaded on a separate line, or from
#   a path/row file where each line contains a path/row pair. If the user
#   chooses the latter, then he may also choose to specify a date range from
#   which to grab the scenes from the lmd_scene table, and a maximum cloud 
#   cover or roll angle allowed for each scene.
#
#   The user may also choose to only view which scenes are obtained for a given
#   run, write the resulting work order ids to a file, and write the obtained
#   scene ids to a file. These are mostly for debugging purposes.
#
################################################################################

use strict;
use warnings;
use ias_const;
use ias_date_utils;
use ias_db;
use ias_create_work_order;
use Logger;
use File::Basename;
use File::Spec;
use Getopt::Long;
use POSIX;
use Time::Piece;
use Time::Seconds 'ONE_DAY';
use Env;

# Prototypes (needed when use warnings is used)
sub help_output();
sub get_past_date($;$);
sub get_default_date_timestamp($$);
sub get_processing_instance_inventory_info($);
sub read_path_row_file($$);
sub read_scenes_from_db($$$$$$$$);
sub create_work_order($$$$$);
sub write_array_to_file($$);
sub eliminate_previously_ordered_scenes($$$);
sub read_scenes_from_file($$);

use constant MAXIMUM_NUMBER_OF_ORDERS => 2000;

############################# BEGIN MAIN #######################################
# Variable definitions
# Command line vars
my $prog_name = basename($0);
my $help = '';
my $acq_date_range = "";
my $eliminate_scenes;
my $max_cloud_cover = 1.0; 
my $max_roll_angle = "0.1";
my $procedure_id = "TIRS SSM Calibration";
my $path_row_file = "";
my $scene_list_file = "";
my $maximum_orders = 0;
my $skip_work_orders = '';
my $output_woid_filename = "";
my $output_scene_list_filename = "";
my $cal_parm_filename= "";

# Derived vars
my @scenes;

# Set up logging
my $logger = Logger->new("STDOUT");
if (!$logger)
{
    print STDERR "Error: failed to establish logger for STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name($prog_name);
set_default_log($logger);

# Read command line arguments
GetOptions ("help" => \$help,
    "acq_date_range=s" => \$acq_date_range,
    "eliminate_scenes" => \$eliminate_scenes,
    "max_cloud_cover=f" => \$max_cloud_cover,
    "max_roll_angle=f" => \$max_roll_angle,
    "path_row_file=s" => \$path_row_file,
    "scene_list_file=s" => \$scene_list_file,
    "procedure_id=s" => \$procedure_id,
    "skip_work_orders" => \$skip_work_orders,
    "output_woid_filename=s" => \$output_woid_filename,
    "output_scene_list_filename=s" => \$output_scene_list_filename,
    "cal_parm_filename=s" => \$cal_parm_filename,
    "set_maximum_orders=i" => \$maximum_orders,
);

if ($help)
{
    help_output();
    exit SUCCESS;
}

# Check the instance variable
my $processing_instance = $ENV{IAS_INSTANCE_NAME};
if (!$processing_instance)
{
    $logger->fatal_msg(
        "The IAS_INSTANCE_NAME environment variable is not set");
    $logger->close_log();
    exit FAILURE;
}

################################################################################
####################### Verifying command line arguments #######################
################################################################################
# Get the start and end acquisition dates to search for
my ($acq_date_start, $acq_date_end) = split /:/, $acq_date_range;

if ($acq_date_range eq "")
{
    # Set date to yesterday.  Do not want timestamp for this, the default
    # timestamps will be appended later indicating full days
    $acq_date_start = get_past_date(1);
}
elsif ($acq_date_range =~ /^\d+$/)
{
    # Have the number of past days for which to get scenes

    # Flag indicating current timestamp should be appended to date
    my $need_timestamp = 1;
    $acq_date_start = get_past_date($acq_date_range, $need_timestamp);
    $acq_date_end = get_past_date(0, $need_timestamp);
}

# Was end acquitision date provided
if (not defined $acq_date_end)
{
    # If not defined, set the end date to the start date
    $acq_date_end = $acq_date_start;
}
# If time stamp isn't appended to date yet get the default timestamp indicating
# full days
if ($acq_date_start !~ /^\d{2}-\d{2}-\d{4}\s?\d{2}:\d{2}:\d{2}/)
{
    # Append the default timestamps to start and end dates
    ($acq_date_start, $acq_date_end) = 
        get_default_date_timestamp($acq_date_start, $acq_date_end);
}

# Verify the dates are valid
if (!ias_is_valid_date($acq_date_start))
{
    $logger->fatal_msg("Invalid start date format");
    $logger->close_log();
    exit FAILURE;
}
elsif (!ias_is_valid_date($acq_date_end))
{
    $logger->fatal_msg("Invalid end date format");
    $logger->close_log();
    exit FAILURE;
}

# Testing maximum cloud cover
if ($max_cloud_cover < 0.00 or $max_cloud_cover > 100.00)
{
    $logger->fatal_msg("Max cloud cover must be between 0 and 100 inclusive");
    $logger->close_log();
    exit FAILURE;
}

# Testing maximum roll angle
if ($max_roll_angle < 0.00 or $max_roll_angle > 90.00)
{
    $logger->fatal_msg("Max roll angle must be between 0 and 90 inclusive");
    $logger->close_log();
    exit FAILURE;
}

# Test validity of p/r scene file combination
if ($path_row_file eq "" and $scene_list_file eq "")
{
    $logger->fatal_msg("Path/row file or scene list file required");
    $logger->close_log();
    exit FAILURE;
}
elsif ($path_row_file ne "" and $scene_list_file ne "")
{
    $logger->fatal_msg("Provide either a path/row file or a scene list file");
    $logger->close_log();
    exit FAILURE;
}

# If the CPF name provided includes a '/', make sure the path provided is
# an absolute path and that the file exists.  If it doesn't include '/', it is
# either empty to allow the default CPF to be used, or it is a name of a CPF
# in the service.
if ($cal_parm_filename =~ /\//)
{
    $cal_parm_filename = File::Spec->rel2abs($cal_parm_filename);
    if (not (-e $cal_parm_filename))
    {
        $logger->fatal_msg("Provided CPF $cal_parm_filename does not exist");
        $logger->close_log();
        exit FAILURE;
    }
}

if (not $maximum_orders)
{
    $maximum_orders = MAXIMUM_NUMBER_OF_ORDERS;
}

################################################################################
##################### End verifying command line arguments #####################
################################################################################

# Determine what scenes to run
# The user hasn't provided a scene list, so get the scenes from the DB
if ($scene_list_file eq "")
{
    my %pathrow_hash;
    if (read_path_row_file($path_row_file, \%pathrow_hash) != SUCCESS)
    {
        $logger->fatal_msg("Error reading path row file $path_row_file");
        $logger->close_log();
        exit FAILURE;
    }

    # Get the inventory link and schema to use when querying the LMD_SCENE
    # table for scenes
    my ($inv_link, $inv_schema) = get_processing_instance_inventory_info(
        $processing_instance);
    if (!defined($inv_link) || !defined($inv_schema))
    {
        $logger->fatal_msg("Error retrieving inv_link or inv_schema");
        $logger->close_log();
        exit FAILURE;
    }

    if(read_scenes_from_db(\%pathrow_hash, $acq_date_start, $acq_date_end,
        $max_cloud_cover, $max_roll_angle, $inv_link, $inv_schema, \@scenes) !=
        SUCCESS)
    {
        $logger->fatal_msg("Error reading from DB");
        $logger->close_log();
        exit FAILURE;
    }
}
else # A scene list file was provided
{
    if (read_scenes_from_file($scene_list_file, \@scenes) != SUCCESS)
    {
        $logger->fatal_msg("Failure reading scenes from file $scene_list_file");
        $logger->close_log();
        exit FAILURE;
    }
}

# Eliminate previously ordered scenes?
if ($eliminate_scenes)
{
    if (eliminate_previously_ordered_scenes($procedure_id, $processing_instance,
        \@scenes) != SUCCESS)
    {
        $logger->fatal_msg("Failure eliminating previously ordered scenes");
        $logger->close_log();
        exit FAILURE;
    }
}

# Check number of scenes to be ordered against the total maximum that can be
# ordered
my $max_orders_exceeded = 0;
my $scene_count = scalar(@scenes);
$logger->info_msg("Number of orders to be placed: $scene_count");
if ($maximum_orders < $scene_count)
{
    $max_orders_exceeded = 1;
    $skip_work_orders = 1;
}

# Create the work orders
if (not $skip_work_orders)
{
    # Connect to the common database.  Note that this is an optimization to
    # avoid making a database connection for every work order inserted
    my ($user_id, $password) = ias_db_GetDBConnectionInfo();

    # connect to the database
    my $dbh_common = ias_db_ConnectToDB($user_id, $password);
    if (!$dbh_common)
    {
        $logger->fatal_msg("Error connecting to database");
        exit FAILURE;
    }

    my @created_woids;
    foreach my $scene (@scenes)
    {
        my $status = create_work_order($dbh_common, $procedure_id, $scene,
            $cal_parm_filename, \@created_woids);
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Failure creating work orders");
            $logger->close_log();
            exit FAILURE;
        }
    }

    # Disconnect from the database.  Each work order is committed as it is
    # inserted, so the rollback is acceptable here.
    if (!ias_db_DisconnectFromDB($dbh_common, IAS_DB_DISCONNECT_ROLLBACK))
    {
        $logger->fatal_msg("Failure closing the common database connection");
        $logger->close_log();
        exit FAILURE;
    }

    # Write out the work order ids to a file if the user so wishes it
    if ($output_woid_filename ne "")
    {
        if (write_array_to_file($output_woid_filename, \@created_woids)
            != SUCCESS)
        {
            $logger->fatal_msg("Failure write work orders ids to "
                . $output_woid_filename);
            $logger->close_log();
            exit FAILURE;
        }
    }
}
else # We are not running any work orders, just output the scenes
{
    $logger->info_msg("Not running work orders as requested. Writing out list "
        . "of scene ids to std out instead");
    foreach my $scene (@scenes)
    {
        $logger->info_msg("Scene id: $scene");
    }
    if ($max_orders_exceeded)
    {
        $logger->fatal_msg("Number of orders - $scene_count will exceed the "
                 . "maximum - " . "$maximum_orders"); 
        $logger->fatal_msg("Use command line option to increase the default "
                 . "maximum " . "number of orders --set_maximum_orders=N");
        $logger->close_log();
        exit FAILURE;
    }
}

# Output the scenes to a file if output_scene_list_filename is set
if ($output_scene_list_filename ne "")
{
    if (write_array_to_file($output_scene_list_filename, \@scenes) != SUCCESS)
    {
        $logger->fatal_msg(
            "Writing scenes to file $output_scene_list_filename");
        $logger->close_log();
        exit FAILURE;
    }
}

exit SUCCESS;
############################# END MAIN #########################################

################################################################################
# Name: help_output
#
# Description: Outputs usage instructions
#
# Parameters:
#   None
#
# Returns:
#   None
################################################################################
sub help_output()
{
    my $filename = basename($0);

    # Ignore the poor indentation of the following block of text. This is
    # required to obtain the correct output format.
    print 
"Usage of $filename
    Required Params:
        You must have one and only one of the following required parameters.
        --path_row_file expects a text file containing path row combinations.
            Each combination should be on its own line and formatted as
            path/row. If this parameter is provided, then the script will
            search the database for scenes matching any path row in this file.
        --scene_list_file expects a file with scene id's. If you provide this 
            parameter, then the script will skip looking in the database for 
            scenes to load, and will instead load the scenes in this file. Each
            scene should be on its own line.
    Optional params:
        --acq_date_range is either a single date, a date range formatted 
            like mm-dd-yyyy:mm-dd-yyyy, or a single number that indicates how
            many days to look into the past from right now. If not provided,
            the utility will default to yesterday's date.
        --eliminate_scenes is an optional flag.  If present, previously ordered
            scenes from the same time period with the same procedure ID  will
            be eliminated from the final scene list.
        --max_cloud_cover indicates the maximum amount of cloud cover you want
            present in your scenes. Defaulted to 1 percent. Provide a number 
            between 0 and 100 for this parameter.
        --max_roll_angle is the absolute maximum roll angle the satellite could
            have been at when the scene was taken. Defaulted to 0.1 degrees.
        --procedure_id is the type of work order you wish to run. Its default
            value is \"TIRS SSM Calibration\".
        --skip_work_orders is an optional flag. If set, then the script will not
            run any work orders, rather it will just write the scene ids to
            standard out.
        --output_woid_filename is the name of a file where the script will write
            the created woids. If not provided, the file will not be written.
        --output_scene_list_filename is the name of a file where the script will
            output the used scene ids. If not provided, the file will not be
            written.
        --cal_parm_filename is an optional CPF to use in place of the default.
            It can be in the service or a file on disk.;
        --set_maximum_orders optional parameter sets the maximum number of 
            orders possible.  Defaulted to ".MAXIMUM_NUMBER_OF_ORDERS.".\n"
}

################################################################################
# Name: get_past_date
#
# Description: Returns a past date as either mm-dd-yyyy or mm-dd-yyyy HH:MI:SS
#              format
#
# Args
#  I: num_past_days: number of past days to return date/timestamp for
#  I: get_time_stamp: flag indicating if the current timestamp should be
#                     appended to past date (Optional)
#
# Returns:
#   Past date
################################################################################
sub get_past_date($;$)
{
    my $num_past_days = shift;
    my $get_time_stamp = shift;
    my $date_time_string;

    my $past_date = gmtime() - ($num_past_days * ONE_DAY);
    if ($get_time_stamp)
    {
        # Note the use of _year to get the year - 1900 which is needed
        # when using Time::Piece (year returns year based from 0)
        $date_time_string = $past_date->strftime("%m-%d-%Y %H:%M:%S");
    }
    else
    {
        $date_time_string = $past_date->strftime("%m-%d-%Y"); 
    }
    return $date_time_string;
}

################################################################################
# Name: get_default_date_timestamp
#
# Description: Appends default time stamp to start/end dates
#
# Args
#  I: start_date: start date to append default timestamp (midnight)
#  I: end_date: end date to append default timestamp (23:59:59)
#
# Returns:
#   Start/End dates with default timetamps 
################################################################################
sub get_default_date_timestamp($$)
{
    my $start_date = shift;
    my $end_date = shift;
    $start_date .= " 00:00:00";
    $end_date .= " 23:59:59";
    
    return $start_date, $end_date;
}

################################################################################
# Name: read_path_row_file
#
# Description: Reads path/rows from a file, returns as a hash
#
# Args
#   I: path row filename
#   O: Hash of path rows. Key = path/row, Value = "defined". Should be passed
#      by reference.  The hash key is in the form "path/row".
#
# Returns:
#   SUCCESS/ERROR
################################################################################
sub read_path_row_file($$)
{
    my $path_row_file = $_[0];
    my $pathrow_hash = $_[1];

    # Open the path/row file to be read
    if (!defined open (PATH_ROW_FILE, '<', $path_row_file))
    {
        $logger->fatal_msg("Provided path/row file does not exist");
        return ERROR;
    }

    # Read in all the entries from the file
    my @pathrows = <PATH_ROW_FILE>;

    # Close the file
    if (!close(PATH_ROW_FILE))
    {
        $logger->fatal_msg("Closing path row input file failed");
        return ERROR;
    }

    # Validate the path/rows read from the file
    foreach my $pathrow (@pathrows)
    {
        # Make sure the path and row are both defined and are in the legal
        # range
        my @temp = split(/\//, $pathrow);
        if (scalar @temp != 2
            or $temp[0] != $temp[0]+0 or $temp[1] != $temp[1]+0
            or $temp[0] < 1 or $temp[1] < 1
            or $temp[0] > 233 or $temp[1] > 248)
        {
            $logger->fatal_msg("Invalid path/row combination: $pathrow read "
                . "from file $path_row_file");
            return ERROR;
        }
        else
        {
            my $path = $temp[0]+0;
            my $row = $temp[1]+0;
            my $tmp = "$path/$row";
            $pathrow_hash->{$tmp} = "defined";
        }
    }
    return SUCCESS;
}

################################################################################
# Name: read_scenes_from_db
#
# Description: Reads scenes from the database and returns as an array
#
# Args
#   I: Comparison hash. This hash contains path_rows
#   I: Start date
#   I; End date
#   I: Maximum cloud cover
#   I: Maximum roll angle
#   I: DB Link to inventory 
#   I: Inventory DB Schema
#   O: Reference to the array of scene ids
#
# Returns:
#   SUCCESS/ERROR
################################################################################
sub read_scenes_from_db($$$$$$$$)
{
    my %pathrow_hash = %{$_[0]};
    my $acq_date_start = $_[1];
    my $acq_date_end = $_[2];
    my $max_cloud_cover = $_[3];
    my $max_roll_angle = $_[4];
    my $inv_link = $_[5];
    my $inv_schema = $_[6];
    my $scenes = $_[7];

    # Get the database connection info for the sensor schema
    my ($db_username, $db_password) = ias_db_GetDBConnectionInfo("COMMON");

    # Connect to the database
    my $dbh = ias_db_ConnectToDB($db_username, $db_password);
    if (!defined($dbh))
    {
        $logger->fatal_msg("Could not connect to database");
        return ERROR;
    }

    # Create string with schema first, then the table name, then the link
    # to use in the query of the LMD_SCENE table
    my $table_with_inv_schema_and_link = $inv_schema."lmd_scene".$inv_link; 

    # Select the scene id, WRS path and WRS row from the LMD_SCENE table for
    # scenes that match the cloud cover, roll angle, and date acquired criteria
    my $sql =  "SELECT LANDSAT_SCENE_ID,WRS_PATH,WRS_ROW "
        .   "FROM $table_with_inv_schema_and_link "
        .   "WHERE CLOUD_COVER <= $max_cloud_cover "
        .       "AND ABS(ROLL_ANGLE) <= $max_roll_angle "
        .       "AND (DATE_ACQUIRED >= "
        .           "TO_DATE(\'$acq_date_start\', \'MM-DD-YYYY HH24:MI:SS\') "
        .           "AND DATE_ACQUIRED <= "
        .           "TO_DATE(\'$acq_date_end\', \'MM-DD-YYYY HH24:MI:SS\')) "
        .       "AND SCENE_STATUS = \'G\'";

    my $res_handle = ias_db_execute_sql($dbh, $sql);
    if (!defined($res_handle))
    {
        ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
        $logger->fatal_msg("Could not run sql command \"$sql\"");
        return ERROR;
    }

    # Many scenes with path rows not matching our file contents will have been
    # returned. We need to iterate through and only push scenes which have
    # path rows in our hash
    while (my @temp = $res_handle->fetchrow() )
    {
        my $pr = "$temp[1]/$temp[2]";
        
        if (exists $pathrow_hash{$pr})
        {
            push @$scenes, $temp[0];
        }
    }
    $res_handle->finish;

    if (!ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK))
    {
        $logger->fatal_msg("Error disconnecting from database");
        return ERROR;
    }

    return SUCCESS;
}

################################################################################
# Name: read_scenes_from_file
#
# Description: Reads scenes from the provided file
#
# Args
#   I: Filename
#   O: Reference to array of scenes
#
# Returns:
#   SUCCESS/ERROR
################################################################################
sub read_scenes_from_file($$)
{
    my $scene_list_file = $_[0];
    my $scenes = $_[1];

    if (!defined open (SCENES_FILE, '<', $scene_list_file))
    {
        $logger->fatal_msg("Provided scenes file does not exist");
        return ERROR;
    }
    @$scenes = <SCENES_FILE>;
    chomp(@$scenes);

    if (!close(SCENES_FILE))
    {
        $logger->fatal_msg("Closing scene file failed");
        return ERROR;
    }
    return SUCCESS;
}

################################################################################
# Name: get_eliminated_dates
#
# Description: Given the start and end dates, creates a string of yeardoy
#              values for each day within the specified date range
#
# Args
#   I: Start Date
#   I: End Date
#   O: Formatted string to use in SQL query including an entry for each day
#      of year within the range:
#      '2013001','2013002','2013003','2013004','2013005','2013006'
#      Note: This sring cannot be longer than  9999 characters (or 999 days)
#            or the maximum sql query length is surpassed.
#
# Returns:
#   SUCCESS/ERROR
################################################################################
sub get_eliminated_dates($$)
{
    my $start_date = shift;
    my $end_date = shift;
    my @dates_to_eliminate = ();

    # Put strings back to times
    my $start = Time::Piece->strptime($start_date,"%m-%d-%Y %H:%M:%S");
    my $end = Time::Piece->strptime($end_date,"%m-%d-%Y %H:%M:%S");
    # Get the difference between the 2 dates for the loop range
    my $diff = ($end - $start) / ONE_DAY;

    for (my $i = 0; $i <= int($diff); $i++)
    {
        # Get the indivdiual date from the starting date + # of days
        my $ind_date = $start + ($i * ONE_DAY);
        # Parse out the Year doy
        my $year_doy = $ind_date->strftime('%Y%j');
        # Add year doy onto array
        push (@dates_to_eliminate, $year_doy);
    }
    
    # Join the array adding the ' before and the ', after each value until
    # the last value 
    my $date_string = '\'' . join ('\',\'', @dates_to_eliminate) . '\'';

    return $date_string;
}

################################################################################
# Name: eliminate_previously_ordered_scenes
#
# Description: Eliminates previously ordered scenes using the same procedure
#              ID and in the same time period
#
# Args
#   I:   Procedure id
#   I/O: Reference to array of scenes
#
# Returns:
#   SUCCESS/ERROR
################################################################################
sub eliminate_previously_ordered_scenes($$$)
{
    my $procedure_id = $_[0];
    my $instance = $_[1];
    my $scenes = $_[2];

    # Query the IAS work orders table to get the scene IDs of work orders
    # using the same procedure ID in the same time period (how much of a range)

    # Get the database connection info
    my ($db_username, $db_password) = ias_db_GetDBConnectionInfo("COMMON");

    # Connect to the database
    my $dbh = ias_db_ConnectToDB($db_username, $db_password);
    if (!defined($dbh))
    {
        $logger->fatal_msg("Could not connect to database");
        return ERROR;
    }

    my $eliminated_dates = get_eliminated_dates($acq_date_start, $acq_date_end);

    # Query for Scene IDS with the same procedure ID from the same time period 
    my $sql =  "SELECT SCENE_ID FROM WORK_ORDERS "
        .   "WHERE PROCESSING_INSTANCE = '$instance' AND "
        .   "PROCEDURE_ID = '$procedure_id' "
        .   "AND SUBSTR(SCENE_ID, 10, 7) IN ($eliminated_dates)";

    my $res_handle = ias_db_execute_sql($dbh, $sql);
    if (!defined($res_handle))
    {
        ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
        $logger->fatal_msg("Could not run sql command \"$sql\"");
        return ERROR;
    }

    # For each scene returned from the query, eliminate the scene from the list
    # if present
    while (my @temp = $res_handle->fetchrow() )
    {
        @$scenes = grep {$_ ne $temp[0]} @$scenes;
    }
    $res_handle->finish;

    # Disconnect from database
    if (!ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK))
    {
        $logger->fatal_msg("Error disconnecting from database");
        return ERROR;
    }

    return SUCCESS;
}

################################################################################
# Name: create_work_order
#
# Description: Creates a new work order using a provided procedure id, scene id,
#   and cpf filepath. The work order id will be appended to an array if
#   successful.
#
# Args
#   I: Common database connection
#   I: Procedure id
#   I: Scene id
#   I: CPF filename
#   IO: Reference to array of work order ids.
#
# Returns:
#   SUCCESS/ERROR
################################################################################
sub create_work_order($$$$$)
{
    my $dbh_common = shift;
    my $procedure_id = shift;
    my $scene_id = shift;
    my $cal_parm_filename = shift;
    my $created_woids = shift;

    my %args;
    my @parms;

    $args{'dbhcommon'} = $dbh_common;
    $args{'wo_comment'} = "Created using TIRS SSM script";
    if ($cal_parm_filename ne "")
    {
        push(@parms, "DSW:CAL_PARM_FILENAME=$cal_parm_filename");
        $args{'parms'} = \@parms;
    }

    my $woid = ias_create_work_order($procedure_id, $scene_id, %args);
    if (!defined($woid))
    {
        $logger->fatal_msg("Failed to create work order");
        return ERROR;
    }
    else
    {
        $logger->info_msg("Created work order $woid with scene id $scene_id");
        push @$created_woids, "$woid";
        return SUCCESS;
    }
}

#############################################################################
# Name: get_processing_instance_inventory_info
#
# Description: Retrieves inventory link and schema from the PROCESSING_INSTANCE
#              table for the currnet processing instance
#
# Parameters:
#       Current database instance
#
# Returns:
#       inventory_link and inventory_schema for current DB instance on SUCCESS
#       undef on ERROR
#
############################################################################
sub get_processing_instance_inventory_info($)
{
    my $instance = shift;

    my ($user_id, $password) = ias_db_GetDBConnectionInfo("COMMON");

    # connect to the database
    my $dbh = ias_db_ConnectToDB($user_id, $password);

    if (!$dbh)
    {
        $logger->fatal_msg("Error connecting to database");
        return undef;
    }

    # Build the query
    my $sql = "select inventory_link, inventory_schema from "
        . "PROCESSING_INSTANCES where INSTANCE_NAME = '$instance'";

    my $res_handle = ias_db_execute_sql($dbh, $sql);
    if (!defined($res_handle))
    {
        ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
        $logger->fatal_msg("Could not run sql command \"$sql\"");
        return undef;
    }

    # Retrieve the inventory_link and inventory_schema
    my ($inventory_link, $inventory_schema) = $res_handle->fetchrow_array();
    if (!defined($inventory_link))
    {
        $logger->info_msg("No inventory link found");
        $inventory_link = "";
    }
    else
    {
        $inventory_link = "@".$inventory_link;
    }
    if (!defined($inventory_schema))
    {
        $logger->info_msg("No inventory schema found");
        $inventory_schema = "";
    }
    else
    {
        $inventory_schema .= ".";
    }
    $res_handle->finish;

    # Disconnect from database
    if (!ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK))
    {
        $logger->fatal_msg("Error disconnecting from database");
        return undef;
    }

    return $inventory_link, $inventory_schema;
}


################################################################################
# Name: write_array_to_file
#
# Description: Writes each index of the array to the provided file on a separate
#   line.
#
# Args
#   I: Filename
#   I: Reference to array of strings to be written
#
# Returns:
#   SUCCESS/ERROR
################################################################################
sub write_array_to_file($$)
{
    my $filename = $_[0];
    my $lines_ref = $_[1];

    # Open the output file for writing
    if (!defined open (OUT_FILE, '>', $filename))
    {
        $logger->fatal_msg("Error opening $filename for output");
        return ERROR;
    }

    # Write all the lines
    foreach my $line (@$lines_ref)
    {
        chomp $line;
        print OUT_FILE "$line\n";
    }

    # Close the output file
    if (!close(OUT_FILE))
    {
        $logger->fatal_msg("Error closing $filename");
        return ERROR;
    }

    return SUCCESS;
}
