#!/usr/bin/perl
################################################################################
# NAME: load_db.pl
# DESCRIPTION: Responsible for loading the LMD_SCENE table on dev with temporary
#   dummy data for running test2.
#   It can accept one of 2 flags; load or remove. Load will load the table, 
#   remove will delete all rows added by the load option.
#
# USAGE:
#   ./load_db --load, --remove, --help
#
# RETURNS:
#   SUCCESS or FAILURE
################################################################################

use strict;
use warnings;
use ias_const;
use ias_db;
use Logger;
use Getopt::Long;
use Env;

################################################################################
# Name: help_output
#
# Description: Outputs usage instructions
#
# Parameters:
#   None
#
# Returns:
#   None
################################################################################
sub help_output()
{
    print "Usage of $0\n"
        . "    --load: set this flag if you need to load your dummy db\n"
        . "    --remove: set this flag if you need to remove dummy entires\n";
}

sub fd($)
{
    my ($raw_dt) = @_;
    my $date_format = "\'dd-mon-yyyy hh24:mi:ss\'";
    return "TO_DATE(\'$raw_dt\', $date_format)";
}

# Scenes which will be loaded/unloaded by this script
# Were obtained using the following command on IT
# ./create_gps_work_orders.pl --path_row_file=pr.txt \
# --acq_date_range="01-01-2013:01-01-2014" --max_cloud_cover="100" \
# --max_roll_angle=90 --skip_work_orders --write_scenes="expected.txt"
my @rows = (
    # Should be queried when unit test is run
    "\'LO80400332013096LGN00\',040,033,50,45,\'G\',"
        . fd("01-jan-2013 00:00:00"),
    "\'LC80400332013110LGN00\',040,033,50,45,\'G\',"
        . fd("01-jul-2013 23:59:59"),
    "\'LT80400332013126LGN00\',040,033,50,45,\'G\',"
        . fd("07-feb-2013 12:30:30"),
    "\'LO81620382013096LGN00\',162,038,50,-1,\'G\',"
        . fd("01-jan-2013 00:00:00"),
    "\'LC81620382013101LGN00\',162,038,25,45,\'G\',"
        . fd("01-jul-2013 23:59:59"),
    "\'LC81620382013133LGN00\',162,038,25,1,\'G\',"
        . fd("07-feb-2013 13:13:13"),
    "\'LT80440312013071LGN00\',044,031,50,-45,\'G\',"
        . fd("01-jan-2013 00:00:00"),
    "\'LC80440312013089LGN00\',044,031,25,1,\'G\',"
        . fd("01-jul-2013 23:59:59"),
    "\'LC80440312013094LGN00\',044,031,25,1,\'G\',"
        . fd("07-feb-2013 13:13:13"),

    # Should not be queried
    # Non-matching path-row
    "\'LC82280852013115LGN00\',228,085,50,45,\'G\',"
        . fd("01-jan-2013 12:00:00"),
    "\'LC82280852013131LGN00\',228,085,50,45,\'G\',"
        . fd("07-feb-2013 17:16:15"),
    "\'LC82280852013147LGN00\',228,085,25,1,\'G\',"
        . fd("07-feb-2013 17:16:15"),
    "\'LC80400332013110LGN01\',041,033,50,45,\'G\',"
        . fd("01-jan-2013 12:00:00"),
    "\'LC80400332013158LGN00\',040,034,50,45,\'G\',"
        . fd("01-jul-2013 12:59:59"),

    # Date out of range
    "\'LC80400332013078LGN01\',040,033,25,1,\'G\',"
        . fd("31-dec-2012 23:59:59"),
    "\'LO80400332013096LGN01\',040,033,25,1,\'G\',"
        . fd("02-jul-2013 00:00:00"),
    "\'LC80400332013206LGN00\',040,033,25,1,\'G\',"
        . fd("01-jan-2014 00:00:00"),

    # Cloud cover out of range
    "\'LC80400332013254LGN00\',040,033,50.01,1,\'G\',"
        . fd("01-jan-2013 12:30:30"),
    "\'LC80400332013222LGN00\',040,033,100,1,\'G\',"
        . fd("01-jan-2013 12:30:30"),

    # Roll angle out of range
    "\'LC81620382013149LGN00\',162,038,50,45.01,\'G\',"
        . fd("01-jan-2013 12:30:30"),
    "\'LC81620382013133LGN01\',162,038,50,-90,\'G\',"
        . fd("01-jan-2013 12:30:30"),

    # Not "Good" Scene
    "\'LC80440312013294LGN00\',044,031,25,1,\'X\',"
        . fd("07-feb-2013 17:16:15"),
);

# Command line args
my $help = '';
my $load = '';
my $remove = '';

# Derived args
my $db_uname = "";
my $db_password = "";
my $dbh;
my @temp;
my $count;

# Set up logging
my $logger = Logger->new("STDOUT");
if (!$logger)
{
    print STDERR "Error: failed to establish logger for STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name($0);
set_default_log($logger);

# Read command line arguments
GetOptions ("help" => \$help,
    "remove" => \$remove,
    "load" => \$load);

if ($help or not ($load or $remove))
{
    help_output();
    exit SUCCESS;
}

@temp = split(/\//,$IAS_DB_L8_OLITIRS);
$db_uname = $temp[0];
$db_password = $temp[1];

$dbh = ias_db_ConnectToDB($db_uname, $db_password);
if (!defined($dbh))
{
    $logger->fatal_msg("Could not connect to database");
    $logger->close_log();
    exit FAILURE;
}

if ($load)
{
    $logger->info_msg("Loading test scenes into the LMD_SCENE table");
    $count = 0;
    foreach my $row (@rows)
    {
        my $sql = "INSERT INTO LMD_SCENE (LMD_SCENE_ID,LANDSAT_SCENE_ID,"
            . "WRS_PATH,WRS_ROW,CLOUD_COVER,ROLL_ANGLE,SCENE_STATUS,"
            . "DATE_ACQUIRED) VALUES ($count,$row)";
        my $res_handle = ias_db_execute_sql($dbh, $sql);
        if (!defined($res_handle))
        {
            $logger->fatal_msg("Could not run sql command $sql");
            $logger->close_log();
            exit FAILURE;
        }
        else
        {
            $logger->debug_msg("Ran query $sql");
        }
        $count = $count + 1;

        $res_handle->finish;
    }
}
elsif ($remove)
{
    $logger->info_msg("Removing test scenes from the LMD_SCENE table");
    $count = 0;
    foreach my $row (@rows)
    {
        my $sql = "DELETE FROM LMD_SCENE WHERE LMD_SCENE_ID=$count";
        my $res_handle = ias_db_execute_sql($dbh, $sql);
        if (!defined($res_handle))
        {
            $logger->fatal_msg("Could not run sql command $sql");
            $logger->close_log();
            exit FAILURE;
        }
        else
        {
            $logger->debug_msg("Ran query $sql");
        }

        $count = $count + 1;
    }
}

if (!ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_COMMIT))
{
    $logger->fatal_msg("Error disconnecting from database");
    $logger->close_log();
    exit FAILURE;
}

exit SUCCESS;
