#!/usr/bin/perl
################################################################################
# NAME: load_work_orders.pl
# DESCRIPTION: Responsible for loading the WORK_ORDERS table with temporary
#   dummy data for running test2.
#   It can accept one of 2 flags; load or remove. Load will load the table, 
#   remove will delete all rows added by the load option.
#   proc_id is an optional parameter that indicates which procedure ID
#   to use when loading the work orders.
#   wo_file is an optional paramter that is a file name containing specific
#   work order ids to be deleted from the WORK_ORDERS table.
#
# USAGE:
#   ./load_work_orders.pl --load, --remove, --proc_id "PROC_ID" 
#        --wo_file "WORK_ORDER_FILE" --help
#
# RETURNS:
#   SUCCESS or FAILURE
################################################################################

use strict;
use warnings;
use ias_const;
use ias_db;
use Logger;
use Getopt::Long;
use Env;

# Function prototypes
sub close_database($);
sub help_output();

# Default procedure ID to use in insert if none is provided
my $default_procedure = "TIRS SSM Calibration";

# Get the DB login information
my $db_uname = "";
my $db_password = "";
my @temp;
@temp = split(/\//,$IAS_DB_COM);
$db_uname = $temp[0];
$db_password = $temp[1];

# Command line args
my $help = '';
my $load = '';
my $remove = '';
my $proc_id = '';
my $wo_file;

# Derived args
my $dbh;
my $count;

# Set up logging
my $logger = Logger->new("STDOUT");
if (!$logger)
{
    print STDERR "Error: failed to establish logger for STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name($0);
set_default_log($logger);

# Read command line arguments
GetOptions ("help" => \$help,
    "remove" => \$remove,
    "load" => \$load,
    "wo_file=s" => \$wo_file,
    "proc_id=s" => \$proc_id);

if ($help or not ($load or $remove))
{
    help_output();
    exit SUCCESS;
}

# Was the procedure ID provided on command line?
if (not $proc_id)
{
    $proc_id = $default_procedure; 
}

# Scenes which will be loaded/unloaded by this script
my @rows = (
# SCENE_ID, STATE, WO_REPLACED, REQUEST_TYPE, REQUESTER, PRIORITY, DELETED
# ARCHIVED, OUTPUT_DB_SCHEMA, WO_PATH, PROCEDURE_ID, DATE_ENTERED
    "\'LO80400332013096LGN00\',\'N\',0,\'S\',\'$db_uname\',9,0,".
        "0,\'P\',\'xxx\',\'$proc_id\',sysdate",
    "\'LC80400332013110LGN00\',\'E\',0,\'S\',\'$db_uname\',9,0, ".
        "0,\'P\',\'xxx\',\'$proc_id\',sysdate",
    "\'LT80400332013126LGN00\',\'DR\',0,\'S\',\'$db_uname\',9,0, ".
        "0,\'P\',\'xxx\',\'$proc_id\',sysdate",
    "\'LO81620382013096LGN00\',\'E\',0,\'S\',\'$db_uname\',9,0,".
        "0,\'P\',\'xxx\',\'$proc_id\',sysdate",
    "\'LC81620382013101LGN00\',\'R\',0,\'S\',\'$db_uname\',9,0, ".
        "0,\'P\',\'xxx\',\'$proc_id\',sysdate",
    "\'LC81620382013133LGN00\',\'N\',0,\'S\',\'$db_uname\',9,0, ".
        "0,\'P\',\'xxx\',\'$proc_id\',sysdate",
    "\'LT80440312013071LGN00\',\'C\',0,\'S\',\'$db_uname\',9,0, ".
        "0,\'P\',\'xxx\',\'$proc_id\',sysdate",
    "\'LC80440312013089LGN00\',\'C\',0,\'S\',\'$db_uname\',9,0, ".
        "0,\'P\',\'xxx\',\'$proc_id\',sysdate",
    "\'LC80440312013094LGN00\',\'E\',0,\'S\',\'$db_uname\',9,0, ".
        "0,\'P\',\'xxx\',\'$proc_id\',sysdate",
);

$dbh = ias_db_ConnectToDB($db_uname, $db_password);
if (!defined($dbh))
{
    $logger->fatal_msg("Could not connect to database");
    $logger->close_log();
    exit FAILURE;
}

if ($load)
{
    $logger->info_msg("Loading test scenes into the WORK_ORDERS table");
    $count = 0;

    # Get the processing instance
    my $instance = $ENV{IAS_INSTANCE_NAME};
    if (!defined $instance)
    {
        $logger->fatal_msg("The processing instance is not defined");
        $logger->close_log();
        close_database($dbh);
        exit FAILURE;
    }

    foreach my $row (@rows)
    {
        # Make the work_order_id unique
        my $wo = $count."UNIT_TEST";
        my $sql = "INSERT INTO work_orders (work_order_id,processing_instance,".
                      "scene_id, state, wo_replaced, request_type," .
                      "requester, priority, deleted, archived, " .
                      "output_db_schema, wo_path, procedure_id, date_entered) ".
                      "VALUES ('$wo', '$instance', $row)";

        my $res_handle = ias_db_execute_sql($dbh, $sql);
        if (!defined($res_handle))
        {
            $logger->fatal_msg("Could not run sql command $sql");
            $logger->close_log();
            close_database($dbh);
            exit FAILURE;
        }
        else
        {
            $logger->debug_msg("Ran query $sql");
        }
        $count = $count + 1;

        $res_handle->finish;
    }
}
elsif ($remove)
{
    $logger->info_msg("Removing test scenes from the WORK_ORDERS table");
    if (!defined($wo_file))
    {
        # Removing work orders that have UNIT_TEST in the work order ID
        $count = 0;
        foreach my $row (@rows)
        {
            my $wo = $count."UNIT_TEST";
            my $sql = "DELETE FROM WORK_ORDERS WHERE work_order_id='$wo'";
            my $res_handle = ias_db_execute_sql($dbh, $sql);
            if (!defined($res_handle))
            {
                $logger->fatal_msg("Could not run sql command $sql");
                $logger->close_log();
                close_database($dbh);
                exit FAILURE;
            }
            else
            {
                $logger->debug_msg("Ran query $sql");
            }

            $count = $count + 1;
        }
    }
    else
    {
        # File was provided with work order ids to be removed
        if (!open(WO, $wo_file))
        {
            $logger->fatal_msg("Cannot open $wo_file for reading");
            $logger->close_log();
            exit FAILURE;
        }
        my @wo_to_remove = <WO>;
        close(WO);

        foreach my $wo (@wo_to_remove)
        {
            chomp $wo;
            my $sql = "DELETE FROM WORK_ORDERS WHERE work_order_id='$wo'";
            my $res_handle = ias_db_execute_sql($dbh, $sql);
            if (!defined($res_handle))
            {
                $logger->fatal_msg("Could not run sql command $sql");
                $logger->close_log();
                close_database($dbh);
                exit FAILURE;
            }
            else
            {
                $logger->debug_msg("Ran query $sql");
            }        
        }
    }
}

close_database($dbh);
exit SUCCESS;

################################################################################
# Name: close_database
#
# Description: Disconnects from database
#
# Parameters:
#   I: Database connection handle
#
# Returns:
#   None
################################################################################
sub close_database($)
{
    my $dbh = shift;
    if (!ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_COMMIT))
    {
        $logger->fatal_msg("Error disconnecting from database");
        $logger->close_log();
        exit FAILURE;
    }
}


################################################################################
# Name: help_output
#
# Description: Outputs usage instructions
#
# Parameters:
#   None
#
# Returns:
#   None
################################################################################
sub help_output()
{
    print "Usage of $0\n"
        . "    --load: set this flag if you need to load your dummy db\n"
        . "    --remove: set this flag if you need to remove dummy entires\n"
        . "    --proc_id: procedure id to use when inserting work order\n"
        . "    --wo_file: file containing work order ids to be deleted\n"; 
}
