#!/bin/csh -f
################################################################################
# NAME: test1_cmd_args.csh
#
# DESC: Tests the create_gps_work_orders.pl utility. This unit test will not run
# work orders; rather it will test proper input acceptance.
################################################################################

set base_cmd = "../create_gps_work_orders.pl"
set valid_cmd = "$base_cmd --scene_list_file=empty_scene_file.txt"
set log = "test_cmd_args.log"

set valid_dates = ( \
    4 \
    "01-01-2015" \
    "01-31-2015" \
    "01-01-2015:01-02-2015" \
    "12-05-2010:01-02-2015" \
    "02-29-2012" \
)

set invalid_dates = ( \
    "asdf" \
    "asdf:04-20-2014" \
    "04-20-2014:asdf" \
    "13-01-2013" \
    "1-04-2014" \
    "01-1-2014" \
    "05-06-07" \
    "07-nov-2013" \
    "feb-03-2012" \
    "01-32-2013" \
    "02-29-2009" \
)

set valid_max_ccs = ( \
    "0" \
    "0.00" \
    "100" \
    "100.00" \
    "5" \
    "5.58" \
    "20.89" \
    "99.999" \
)

set invalid_max_ccs = ( \
    "-0.01" \
    "-1" \
    "101" \
    "100.01" \
)

set valid_roll_angles = ( \
    "0" \
    "0.00" \
    "90" \
    "90.00" \
    "50.55" \
    "0.01" \
)

set invalid_roll_angles = ( \
    "-1" \
    "-1.00" \
    "-0.01" \
    "91" \
    "91.00" \
    "90.01" \
)
set final_status = 0

echo "Starting test number 1" |& tee $log

# Program should fail if not given an input file
$base_cmd > & /dev/null
if ( $status == 0 ) then
    echo "ERROR: Script should scene list input file" |& tee -a $log
    set final_status = 1
endif

# Program should also fail if given a non-existent scene file
$base_cmd --scene_list_file=idonotexist.txt > & /dev/null
if ( $status == 0 ) then
    echo "ERROR: Script should fail if given non-existent scene file" \
        |& tee -a $log
    set final_status = 1
endif

# Program should succeed if only given a scene list file
touch empty_scene_file.txt
$valid_cmd > & /dev/null
if ( $status != 0 ) then
    echo "ERROR: Script should succeed if given only a scene file" \
        |& tee -a $log
    set final_status = 1
endif

# Program should fail if given non-existant cpf file
$valid_cmd --cal_parm_file="i/do/not/exist" > & /dev/null
if ( $status == 0 ) then
    echo "ERROR: Script should fail if given non-existent cpf" \
        |& tee -a $log
    set final_status = 1
endif

# Testing date validity (if valid range is not provided yesterday will be used)
$valid_cmd  > & /dev/null
if ( $status != 0 ) then
    echo "ERROR: No acq_date_range uses yesterday for the range." |& tee -a $log
    set final_status = 1
endif

foreach valid_date ( $valid_dates )
    $valid_cmd --acq_date_range="$valid_date" > & /dev/null
    if ( $status != 0 ) then
        echo "ERROR: Date $valid_date should be considered valid" |& tee -a $log
        set final_status = 1
    endif
end
foreach invalid_date ( $invalid_dates )
    $valid_cmd --acq_date_range="$invalid_date" > & /dev/null
    if ( $status == 0 ) then
        echo "ERROR: Date $invalid_date should not be considered valid" \
            |& tee -a $log
        set final_status = 1
    endif
end

# Test different cloud cover values
foreach valid_max_cc ( $valid_max_ccs )
    $valid_cmd --max_cloud_cover="$valid_max_cc" > & /dev/null
    if ( $status != 0 ) then
        echo "ERROR: CC value $valid_max_cc should be considered valid" \
             |& tee -a $log
        set final_status = 1
    endif
end
foreach invalid_max_cc ( $invalid_max_ccs )
    $valid_cmd --max_cloud_cover="$invalid_max_cc" > & /dev/null
    if ( $status == 0 ) then
        echo "ERROR: CC value $invalid_max_cc should not be considered valid" \
             |& tee -a $log
        set final_status = 1
    endif
end

# Test different roll angle values
foreach valid_roll_angle( $valid_roll_angles )
    $valid_cmd --max_roll_angle="$valid_roll_angle" > & /dev/null
    if ( $status != 0 ) then
        echo "ERROR: Roll angle of $valid_roll_angle should be considered" \
            "valid" |& tee -a $log
        set final_status = 1
    endif
end
foreach invalid_roll_angle( $invalid_roll_angles )
    $valid_cmd --max_roll_angle="$invalid_roll_angle" > & /dev/null
    if ( $status == 0 ) then
        echo "ERROR: Roll angle of $invalid_roll_angle should not be" \
            "considered valid" |& tee -a $log
        set final_status = 1
    endif
end

if ( $final_status == 1 ) then
    echo "There were failures. Output appended to $log" |& tee -a $log
else
    echo "Test 1 success: Script correctly handles command line arguments" \
        |& tee -a $log
endif

rm empty_scene_file.txt

exit $final_status
