#!/bin/csh -f
################################################################################
#   Name: test2_scene_acquisition.csh
#
#   Desc: Tests the create_gps_work_orders.pl script to ensure its queries
#       will grab the appropriate scenes from the LMD_SCENE table and from
#       any scene file it's sent.
#
#   Usage: ./test_scene_acquisition.csh
################################################################################

set acquired_scenes = "acquired_scenes.tmp"
set log = test_scene_acquisition.log

echo "Starting test number 2" |& tee $log

# Test grabbing scenes from file
../create_gps_work_orders.pl --scene_list_file=scene_file.txt \
    --skip_work_orders \
    --output_scene_list_filename="$acquired_scenes" |& tee -a $log
if ( $status != 0 ) then
    echo "Failed to load scenes from text file" |& tee -a $log
    exit 1
endif

diff $acquired_scenes test2_expected_scenes_file.txt |& tee -a $log
if ( $status != 0 ) then
    echo "Acquired scenes from file are different than expected" |& tee -a $log
    exit 1
endif

# Test grabbing scenes from database
./load_db.pl --remove |& tee -a $log
./load_db.pl --load |& tee -a $log
if ( $status != 0 ) then
    echo "Failed to load data" |& tee -a $log
    exit 1
endif

../create_gps_work_orders.pl --path_row_file=test2_input_path_row.txt \
    --acq_date_range="01-01-2013:07-01-2013" \
    --max_cloud_cover=50 --max_roll_angle=45 --skip_work_orders \
    --output_scene_list_filename="$acquired_scenes" |& tee -a $log
if ( $status != 0 ) then
    echo "Failed to read scenes from lmd_scene table" |& tee -a $log
    exit 1
endif

diff $acquired_scenes test2_expected_scenes_db.txt |& tee -a $log
if ( $status != 0 ) then
    echo "Scenes acquired from database are different than expected" \
        |& tee -a $log
    exit 1
else
    echo "Test 2 success: Script properly acquires scenes" \
        |& tee -a $log
endif

rm $acquired_scenes

#
# Remove any previously entered work orders using acquisition date as date
# entered, but leave the scenes in the LMD_SCENE table
./load_work_orders.pl --remove
./load_work_orders.pl --load

../create_gps_work_orders.pl --path_row_file=test2_input_path_row.txt \
    --acq_date_range="01-01-2013:07-01-2013" --eliminate_scenes \
    --max_cloud_cover=50 --max_roll_angle=45 \
    --output_scene_list_filename="$acquired_scenes" |& tee -a $log

# All scenes were loaded with the date acquired as the date entered
# Therefore, all scenes should have been eliminated and the output file
# size should be 0
if (! -z $acquired_scenes ) then
    echo "Scenes acquired from database are different than expected" \
        |& tee -a $log
    exit 1
else
    echo "Test 2 success: Script properly acquires scenes" \
        |& tee -a $log
endif

./load_work_orders.pl --remove
rm $acquired_scenes

# Now load work orders with a different procedure ID
./load_work_orders.pl --load --proc_id "L1T Product"

# Call create_gps_work_orders
../create_gps_work_orders.pl --path_row_file=test2_input_path_row.txt \
    --acq_date_range="01-01-2013:07-01-2013" \
    --max_cloud_cover=50 --max_roll_angle=45 --skip_work_orders \
    --output_scene_list_filename="$acquired_scenes" |& tee -a $log

# All scenes should have been returned
diff $acquired_scenes test2_expected_scenes_db.txt |& tee -a $log
if ( $status != 0 ) then
    echo "Scenes acquired from database are different than expected" \
        |& tee -a $log
    exit 1
else
    echo "Test 2 success: Script properly acquires scenes" \
        |& tee -a $log
endif

./load_work_orders.pl --remove |& tee -a $log

# Call create_gps_work_orders, setting the maximum work orders lower than the
# number to be inserted
../create_gps_work_orders.pl --path_row_file=test2_input_path_row.txt \
    --acq_date_range="01-01-2013:07-01-2013" \
    --max_cloud_cover=50 --max_roll_angle=45 \
    --output_scene_list_filename="$acquired_scenes" --set_maximum_orders=2 \
    |& tee -a $log
if ( $status == 0 ) then
    echo "Expected to error out when max order number is exceeded" \
        |& tee -a $log
    exit 1
else
    echo "Test 2 success: Errored out as expected when max order number "\
       "is exceeded" |& tee -a $log
endif

# Cleanup
rm $acquired_scenes
./load_db.pl --remove |& tee -a $log
./load_work_orders.pl --remove |& tee -a $log

exit 0
