
/*****************************************************************************
    PROGRAM: convert_l1g_h4_to_h5

    PURPOSE: To convert the older Landsat HDF4 formated L1G data to the
             newer Landsat HDF5 format.

      NOTES: The older IAS system uses/produces data for L1-5 MSS, L1-5 TM,
             and L7 ETM, using HDF4 formated files .  The newer IAS system
             uses/produces data for L8 OLITIRS using HDF5 formated files.

      USAGE:
        convert_l1g_h4_to_h5 <h4_l1g> <h5_l1g> [<h4_sensor>]

           <h4_l1g> - Pathname of the HDF4 L1G input file
           <h5_l1g> - Pathname of the HDF5 L1G output file
        <h4_sensor> - Optional sensor type of the HDF4 l1g, default is ETM
                      Other choices are MSS and TM

*****************************************************************************/

/* HDF4 L1G lib headers */
#include "xxx_Sensor.h"
#include "gxx_hdf_io.h"
#include "gxx_ddr.h"
#include "gxx_utility.h"

/* HDF5 L1G lib header */
#include "ias_l1g.h"

/* IAS logging library */
#include "ias_logging.h"
#include "ias_miscellaneous.h"

#include "date_conversions.h"

/*
 * converts the HDF4 L1G DDR metadata to HDF5 L1G Metadata
 */
int convert_h4_ddr_to_h5
(
    L1GIO *h5_hdfid,                /* I: handle to L1G HDF5 file */
    const char *h4_image_filename,  /* I: Path & name of the input HDF4 1G 
                                          image file*/
    int nbands,                     /* I: number of bands */
    BandNumber *bandno,             /* I: enumerated band numbers */
    BAND_HDR_TYPE *band_hdr         /* I: Band header this satellite */
) 
{
    char h4_ddr_filename[FILENAME_MAX];        /* DDR filename */
    xxx_DDRControl_TYPE *h4_DDRcontrol = NULL; /* DDR control structure from 
                    gxx_ReadDDR.  Space for this is allocated in gxx_ReadDDR*/
    IAS_L1G_BAND_METADATA *h5_band_meta = NULL;
    IAS_L1G_FILE_METADATA h5_file_meta;
    int status = ERROR; 
    int overall_status = ERROR;
    int band_count;
    int correction_type;
    const char correction[] = {'s', 'p', 't'}; /* s = systematic, p = precision,
                                            t = terrain */
    int index;
    int count;

    memset(&h5_file_meta, 0, sizeof(h5_file_meta));

    gxx_create_DDR_name(h4_ddr_filename, h4_image_filename);
    h4_DDRcontrol = gxx_ReadDDR(h4_ddr_filename);
    if (h4_DDRcontrol == NULL)
    {
         IAS_LOG_ERROR("Reading HDF4 DDR");
         goto CLEANUP;
    }

    /* GCTP projection code for the image, copied */
    h5_file_meta.projection_code = h4_DDRcontrol->p_DDR[0].ProjectCode;

    /* zone code for the UTM projection, copied */
    h5_file_meta.zone_code = h4_DDRcontrol->p_DDR[0].ZoneCode;

    /* Copy the datum.  Note that L1-7 uses WGS_84 while L8 uses WGS84, so
       convert those values. */
    if (strcmp(h4_DDRcontrol->p_DDR[0].DatumCode, "WGS_84") == 0)
    {
        /* Convert WGS_84 to WGS84 */
        count = snprintf(h5_file_meta.datum, sizeof(h5_file_meta.datum),
            "WGS84");
    }
    else
    {
        /* Not the expected WGS_84, so just copy what is there */
        count = snprintf(h5_file_meta.datum, sizeof(h5_file_meta.datum),
            h4_DDRcontrol->p_DDR[0].DatumCode);
    }
    if ((count < 0) || (count >= sizeof(h5_file_meta.datum)))
    {
        IAS_LOG_WARNING("Datum string truncated because the target buffer is"
            " too small");
    }

    /* GCTP spheroid code (usually 12 for WGS84), copied */
    h5_file_meta.spheroid_code = h4_DDRcontrol->p_DDR[0].SpheroidCode;

    /* projection units (usually "METERS" or "DEGREES"), copied */
    count = snprintf(h5_file_meta.projection_units,
        sizeof(h5_file_meta.projection_units),
        h4_DDRcontrol->p_DDR[0].ProjectionUnits);
    if ((count < 0) || (count >= sizeof(h5_file_meta.projection_units)))
    {
        IAS_LOG_WARNING("Projection Units truncated because the target buffer "
            "is too small");
    }

    /* make sure the number of projection parameters agrees */
    if (XXX_PROJECTION_COEF_SIZE != IAS_PROJ_PARAM_SIZE)
    {
        /* This should never happen since the number of projection parameters
           is dictated by GCTP which both formats are based on */
        IAS_LOG_ERROR("The formats have a different number of projection "
            "parameters");
        goto CLEANUP;
    }

    /* copy all of the projection parameters */
    for (index = 0; index < IAS_PROJ_PARAM_SIZE; index++)
    {
        h5_file_meta.projection_parameters[index] = 
            h4_DDRcontrol->p_DDR[0].ProjectionCoef[index];
    }

    h5_file_meta.wrs_path = h4_DDRcontrol->p_DDR[0].WRSPath;
    h5_file_meta.wrs_row = (int)(h4_DDRcontrol->p_DDR[0].WRSRows);

    /* roll angle for off-nadir collections. Past Landsat satellites are always
       nadir pointing and the roll angle isn't recorded in the HDF4.  So, we'll
       just use 0 degrees */
    h5_file_meta.roll_angle = 0.0;

    /* spacecraft name */
    count = snprintf(h5_file_meta.spacecraft, sizeof(h5_file_meta.spacecraft),
        h4_DDRcontrol->p_DDR[0].SpaceCraftSource);
    if ((count < 0) || (count >= sizeof(h5_file_meta.spacecraft)))
    {
        IAS_LOG_WARNING("Spacecraft name truncated because the target buffer "
            "is too small");
    }

    /* Past Landsat satellites were always collecting Earth imagery, so we'll 
       set collection_type for HDF5 to EARTH */
    count = snprintf(h5_file_meta.collection_type,
        sizeof(h5_file_meta.collection_type), "EARTH");
    if ((count < 0) || (count >= sizeof(h5_file_meta.collection_type)))
    {
        IAS_LOG_ERROR("Specifying Collection Type");
        goto CLEANUP;
    }

    /* ASCENDING or DESCENDING -- The HDF4 format uses lower case, HDF5 uses 
       upper case */
    char *pDir = NULL;
    if (strcmp(h4_DDRcontrol->p_DDR[0].CaptureDirection, 
                       XXX_DIRECTION_DESCEND) == 0)
    {
        pDir = "DESCENDING";
    }
    else if (strcmp(h4_DDRcontrol->p_DDR[0].CaptureDirection, 
                       XXX_DIRECTION_ASCEND) == 0)
    {
        pDir = "ASCENDING";
    }
    else /* didn't match the above? just copy whatever is there */
        pDir = h4_DDRcontrol->p_DDR[0].CaptureDirection;

    count = snprintf(h5_file_meta.capture_direction,
        sizeof(h5_file_meta.capture_direction), pDir);
    if ((count < 0) || (count >= sizeof(h5_file_meta.capture_direction)))
    {
        IAS_LOG_ERROR("Copying Capture Direction (%s)", pDir);
        goto CLEANUP;
    }

    /* Convert the capture date to the expected format, if possible.
       Otherwise, just copy it unmodified */
    status = convert_date_string(
        DATE_FORMAT_DD_MON_YYYY, h4_DDRcontrol->p_DDR[0].CaptureDate,
        DATE_FORMAT_YYYY_MM_DD, h5_file_meta.capture_date,
        sizeof(h5_file_meta.capture_date));
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Converting Capture Date");
        goto CLEANUP;
    }
    else if (status == INVALID_DATE 
             && strlen(h4_DDRcontrol->p_DDR[0].CaptureDate) > 0)
    {
        
        IAS_LOG_WARNING("Unable to convert input capture date '%s' to expected"
                    " format - copying unmodified",
                    h4_DDRcontrol->p_DDR[0].CaptureDate);
        snprintf(h5_file_meta.capture_date, sizeof(h5_file_meta.capture_date),
                    h4_DDRcontrol->p_DDR[0].CaptureDate);
    }

    /* capture time:  same HH:MM:SS format, just copy */
    count = snprintf(h5_file_meta.capture_time, 
        sizeof(h5_file_meta.capture_time),
        h4_DDRcontrol->p_DDR[0].CaptureTime);
    if ((count < 0) || (count >= sizeof(h5_file_meta.capture_time)))
    {
        IAS_LOG_WARNING("Capture time truncated because the target buffer "
            "is too small");
    }

    /* make sure the source correction type is in the expected range */
    correction_type = h4_DDRcontrol->p_DDR[0].CorrectionType-1;
    if (correction_type >= 0 || correction_type <= 2)
    {
        /* set the correction type "L1G<x>" where <x> = s, p, or t */
        count = snprintf(h5_file_meta.correction_type, 
            sizeof(h5_file_meta.correction_type), "L1G%c", 
            correction[correction_type]);
        if ((count < 0) || (count >= sizeof(h5_file_meta.correction_type)))
        {
            IAS_LOG_ERROR("Determining Correction Type");
            goto CLEANUP;
        }
    }
    else
    {
        IAS_LOG_WARNING("Unexpected correction type in source, leaving blank");
        h5_file_meta.correction_type[0] = '\0';
    }

    /* BI/NN/CC/MTF -- Not in HDF4 format, setting to 'n/a' */
    snprintf(h5_file_meta.resample_type, sizeof(h5_file_meta.resample_type),
        "n/a"); 

    /* software version used to create the image -- not in HDF4 */
    snprintf(h5_file_meta.software_version,
        sizeof(h5_file_meta.software_version), "n/a");

    /* ingest software version -- not in HDF4 */
    snprintf(h5_file_meta.ingest_software_version,
        sizeof(h5_file_meta.ingest_software_version), "n/a");

    /* sun azimuth angle at the scene center -- not in HDF4 */
    h5_file_meta.sun_azimuth = 0.0;         

    /* sun elevation angle at the scene center -- not in HDF4 */
    h5_file_meta.sun_elevation = 0.0;       

    /* 1 for sun data loaded and 0 if not */
    h5_file_meta.sun_angles_valid = 0;       

    IAS_LOG_DEBUG("Writing HDF5 File Metadata");
    status = ias_l1g_set_file_metadata( h5_hdfid, &h5_file_meta);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Cannot set HDF5 File Metadata");
        goto CLEANUP;
    }

    /* allocate memory for band metadata */
    h5_band_meta = malloc(sizeof(IAS_L1G_BAND_METADATA) * nbands);
    if (h5_band_meta == NULL)
    {   
        IAS_LOG_ERROR("Cannot allocate HDF5 Band Metadata for %d bands", 
                        nbands);
        goto CLEANUP;
    }

    /* clear it */
    memset(h5_band_meta, 0, sizeof(IAS_L1G_BAND_METADATA) * nbands);

    /* loop through the bands */ 
    for (band_count = 0; band_count < nbands; band_count++)
    {
        BAND_NODE_TYPE *band_node = NULL;
        int band = bandno[band_count];

        h5_band_meta[band_count].band_number = 
                       h4_DDRcontrol->p_DDR[band_count].BandNumber;

        /* point to the band node of interest */
        band_node = xxx_GetBandNode(band_hdr, band, BAND_ID_TYPE);
        if (band_node == NULL)
        {
            IAS_LOG_ERROR("Cannot find HDF4 band %d", band);
            goto CLEANUP;
        }

        /* copy user readable band name. */
        strncpy(h5_band_meta[band_count].band_name, band_node->band_name,
                 sizeof(h5_band_meta[band_count].band_name));
        h5_band_meta[band_count].band_name[
            sizeof(h5_band_meta[band_count].band_name) - 1] = '\0';

        /* copy corner coordinates */
        h5_band_meta[band_count].upper_left_y = 
                       h4_DDRcontrol->p_DDR[band_count].UpperLeft[0];
        h5_band_meta[band_count].upper_left_x = 
                       h4_DDRcontrol->p_DDR[band_count].UpperLeft[1];

        h5_band_meta[band_count].upper_right_y = 
                       h4_DDRcontrol->p_DDR[band_count].UpperRight[0];
        h5_band_meta[band_count].upper_right_x = 
                       h4_DDRcontrol->p_DDR[band_count].UpperRight[1];

        h5_band_meta[band_count].lower_left_y = 
                       h4_DDRcontrol->p_DDR[band_count].LowerLeft[0];
        h5_band_meta[band_count].lower_left_x = 
                       h4_DDRcontrol->p_DDR[band_count].LowerLeft[1];

        h5_band_meta[band_count].lower_right_y = 
                       h4_DDRcontrol->p_DDR[band_count].LowerRight[0];
        h5_band_meta[band_count].lower_right_x = 
                       h4_DDRcontrol->p_DDR[band_count].LowerRight[1];
       
       /* copy pixel size in projection coordinates */
       h5_band_meta[band_count].projection_distance_y =
                       h4_DDRcontrol->p_DDR[band_count].ProjectionDistY;
       h5_band_meta[band_count].projection_distance_x =
                       h4_DDRcontrol->p_DDR[band_count].ProjectionDistX;
    
       /* copy maximum DN of pixels */
       h5_band_meta[band_count].maximum_pixel_value =
                       h4_DDRcontrol->p_DDR[band_count].MaxPixelValue;

       /* copy minimum DN of pixels */
       h5_band_meta[band_count].minimum_pixel_value =
                       h4_DDRcontrol->p_DDR[band_count].MinPixelValue;

       /* indicates valid min/max pixel values */
       h5_band_meta[band_count].pixel_range_valid = 1;
   
       /* maximum radiance value */
       h5_band_meta[band_count].maximum_radiance =
                       h4_DDRcontrol->p_DDR[band_count].MaxRadianceValue;

       /* minimum radiance value */
       h5_band_meta[band_count].minimum_radiance =
                       h4_DDRcontrol->p_DDR[band_count].MinRadianceValue;

       /* offset to convert to spectral radiance */
       h5_band_meta[band_count].spectral_radiance_scaling_offset = 0.0; 

       /* gain to convert to spectral radiance - set to 1.0 since it isn't
          available in the source file but the same valid flag covers the
          min/max radiance values */
       h5_band_meta[band_count].spectral_radiance_scaling_gain = 1.0;

       /* indicates radiance items are valid */
       h5_band_meta[band_count].radiance_valid = 1; 

       /* offset to convert to reflectance */
       h5_band_meta[band_count].reflectance_scaling_offset = 0.0;

       /* gain to convert to reflectance */
       h5_band_meta[band_count].reflectance_scaling_gain = 0.0;  

       /* indicates reflectance items are valid */
       h5_band_meta[band_count].reflectance_valid = 0;
      
       /* source of the data */
       snprintf(h5_band_meta[band_count].instrument_source,
           sizeof(h5_band_meta[band_count].instrument_source),
           h4_DDRcontrol->p_DDR[band_count].InstrumentSource);

    } 

    IAS_LOG_DEBUG("Writing HDF5 Band Metadata");

    status = ias_l1g_set_band_metadata(h5_hdfid, h5_band_meta, nbands);
    if (status != SUCCESS )
    {
        IAS_LOG_ERROR("Cannot set HDF5 Band Metadata");
        goto CLEANUP;
    }

    overall_status = SUCCESS;

CLEANUP:

    if (h4_DDRcontrol != NULL)
    {
        XXX_FREE_DDR(h4_DDRcontrol);
        h4_DDRcontrol = NULL;
    }

    if (h5_band_meta != NULL)
    {   
        free(h5_band_meta);
        h5_band_meta = NULL;
    }

    if (overall_status == SUCCESS)
    {
        IAS_LOG_DEBUG("Conversion of HDF4 DDR to HDF5 metadata completed "
                       "successfully.");
    }
    else
    {
        IAS_LOG_ERROR("Conversion of HDF4 DDR to HDF5 metadata failed.");
    }

    return overall_status;    
}
 

/* 
 * reads an HDF4 L1G image file and creates an HDF5 L1G equivalent 
 */
int convert_h4_to_h5
(
    const char *h4_image_filename, /* I: Pathname of the input HDF4 1G file */
    const char *h4_sensor,         /* I: Sensor ID of the input HDF4 1G file */
    const char *h5_image_filename  /* I: Pathname of the output HDF5 1G file */
)
{
    HDFIO *h4_hdfid = NULL;        /* Handle to HDF4 file */
    L1GIO *h5_hdfid = NULL;        /* Handle to HDF5 file */
    const IAS_ACCESS_MODE h4_access_mode = IAS_READ;  
    const IAS_ACCESS_MODE h5_access_mode = IAS_WRITE; 
    L1G_BAND_IO *h5_l1g_band = NULL;  /* h5 L1G band handle */
    unsigned int nbands = 0;       /* Number of bands found in HDF4 */
    BandNumber h4_band_number_list[NBANDS]; /* list of HDF4 enumerated band 
                                               numbers */
    IAS_DATA_TYPE band_dtype_list[NBANDS]; /* Data type of each band */
    int  num_lines[NBANDS];        /* Number of lines per band */
    int  num_samples[NBANDS];      /* Number of samples per band */
    BandNumber band_number;        /* (enumerated) band number to read */
    const int sl = 0;              /* Line to start reading/writing (0-rel) */
    const int ss = 0;              /* Sample to start reading/writing (0-rel) */
    int  nl;                       /* Number of lines to read/write */
    int  ns;                       /* Number of samples to read/write */
    const int subsamp = 1;         /* Subsampling factor (none) */
    void *image_buffer = NULL;     /* Image buffer */
    BAND_HDR_TYPE *band_hdr = NULL;   /* HDF4 Band header */
    BAND_NODE_TYPE *band_node = NULL; /* an element of the HDF4 band linked 
                                         list. */
    unsigned int band_index;       /* index for band_number_list */
    const int sca_index = 0; 

    xxx_FuncPtrs *h4_xxx = NULL;   /* HDF4 Sensor specific functions */
    H4_IAS_SENSOR_TYPE enum_sens;  /* enumerated sensor type */

    int h5_band_number;            /* HDF5 band number */

    int status = ERROR;            /* status of function calls */
    int overall_status = ERROR;    /* overall status of this function */

    /* Init the sensor to get the right band list */
    enum_sens = xxx_initialize_sensor_type(h4_sensor);
    if (IAS_SENSOR_UNKNOWN == enum_sens)
    {
        IAS_LOG_ERROR("Unknown HDF4 sensor: %s", h4_sensor);
        goto NO_CLEANUP;
    }

    xxx_initialize_func_ptr();
    h4_xxx = xxx_get_funcs();
    h4_xxx->InitBands();
    band_hdr = xxx_GetBands();

    IAS_LOG_INFO("Reading HDF4 file %s", h4_image_filename);

    /* open the HDF4 image for reading, this will return information about the
       bands present */
    h4_hdfid = gxx_open_1g_image(
        h4_image_filename,   /* Path & name of the 1G image file */
        h4_access_mode,      /* Requested access mode of image */
        &nbands,             /* Number of bands (returned) */
        h4_band_number_list, /* enumerated band numbers (returned) */
        band_dtype_list,     /* Data type of each band (returned) */
        num_lines,           /* Number of lines per band (returned) */
        num_samples);        /* Number of samples per band (returned) */
    if (h4_hdfid == NULL)
    {
        IAS_LOG_ERROR( "Cannot open %s for reading, exit with failure.", 
              h4_image_filename );
        goto CLEANUP;
    }
   
    IAS_LOG_INFO("Writing HDF5 file %s", h5_image_filename);

    h5_hdfid = ias_l1g_open_image(
        h5_image_filename,  /* Path & name of the HDF5 1G image file  */
        h5_access_mode);    /* Requested access mode of the HDF5 image */
    if (h5_hdfid == NULL)
    {
        IAS_LOG_ERROR("Cannot open %s for writing, exit with failure.", 
                  h5_image_filename);
        goto CLEANUP;
    }
  
    /* loop through the bands found in the HDF4 data */ 
    for (band_index = 0; band_index < nbands; band_index++)
    {
        int dt_size = 0;
        int sca_count = 1;             /* only 1 SCA */

        /* determine the enumerated band number */
        band_number = h4_band_number_list[band_index];

        /* point to the band node for this band number */
        band_node = xxx_GetBandNode(band_hdr, band_number, BAND_ID_TYPE);
        if (band_node == NULL)
        {
            IAS_LOG_ERROR("Cannot find HDF4 band %d", band_number);
            goto CLEANUP;
        }

        /* get the image lines (height) and samples (width) */ 
        nl = num_lines[band_index];   /* number of lines in this band */
        ns = num_samples[band_index]; /* number of samples in this band */

        status = ias_misc_get_sizeof_data_type(band_dtype_list[band_index],
                                      &dt_size );
        if (status == ERROR)
        {
            IAS_LOG_ERROR("Cannot determine size of data type");
            goto CLEANUP;
        }

        /* allocate space for lines x samples x sample_size */
        image_buffer = malloc( nl * ns * dt_size );
        if (image_buffer == NULL)
        {
            IAS_LOG_ERROR("Cannot allocate memory for image");
            goto CLEANUP;
        }
 
        IAS_LOG_DEBUG("Reading band HDF4 %s", band_node->band_name );

        status = gxx_read_1g_image(
            h4_hdfid,      /* HDF IO structure */
            band_number,   /* (enumerated) band number to read */
            sl,            /* Line to start reading (0-rel) */
            ss,            /* Sample to start reading (0-rel) */
            nl,            /* Number of lines to read */
            ns,            /* Number of samples to read */
            subsamp,       /* Subsampling factor */
            image_buffer); /* image_buffer buffer */
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR( "Reading HDF4 band %d image", band_number );
            goto CLEANUP;
        }
    
        /* we'll use the long band number to make unique HDF5 band numbers */ 
        h5_band_number = band_node->band_long_number;
        IAS_LOG_INFO("Saving HDF4 band %d (%s) as HDF5 band %d", band_number,
                band_node->band_name, h5_band_number);

        h5_l1g_band = ias_l1g_open_band(h5_hdfid, h5_band_number,
                &band_dtype_list[band_index], &sca_count, &nl, &ns);
        if (h5_l1g_band == NULL)
        {
            IAS_LOG_ERROR("Adding band %d to output image %s", h5_band_number,
                    h5_image_filename );
            goto CLEANUP;
        }

        status = ias_l1g_write_image(
            h5_l1g_band,   /* L1G_BAND_IO structure */
            sca_index,     /* SCA to write to (0-rel) */
            sl,            /* Line to start writing (0-rel) */
            ss,            /* Sample to start writing (0-rel) */
            nl,            /* Number of lines to write */
            ns,            /* Number of samples to write */
            image_buffer); /* image_buffer buffer */
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Writing HDF5 band %d image", h5_band_number);
            goto CLEANUP;
        }

        status = ias_l1g_close_band(h5_l1g_band);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Closing HDF5 band %d", h5_band_number);
            goto CLEANUP;
        }
        
        free(image_buffer);
        image_buffer = NULL;
    }

    IAS_LOG_DEBUG("Converting HDF4 DDR to HDF5");
    status = convert_h4_ddr_to_h5(h5_hdfid, h4_image_filename, nbands, 
                             h4_band_number_list, band_hdr );
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Cannot convert HDF4 DDR");
        goto CLEANUP;
    }

    /* if we've reached this point, everything worked (so far) */
    overall_status = SUCCESS;

CLEANUP:
    /* do we have allocated imagery to release? */
    if (image_buffer != NULL)
    {
        free(image_buffer);
        image_buffer = NULL;
    }

    /* do we need to close the HDF4 image? */
    if (h4_hdfid != NULL)
    {
        status = gxx_close_1g_image(h4_hdfid);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Closing HDF4 file");
            overall_status = ERROR;
        }
        h4_hdfid = NULL;
    }

    /* do we need to close the HDF5 image? */
    if (h5_l1g_band == NULL)
    {
        status = ias_l1g_close_band(h5_l1g_band);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Closing HDF5 band ");
            overall_status = ERROR;
        }
        h5_l1g_band = NULL;
    }
     
    /* do we need to close the HDF5 file? */ 
    if (h5_hdfid != NULL)
    {
        status = ias_l1g_close_image(h5_hdfid);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Closing HDF5 file");
            overall_status = ERROR;
        }
        h5_hdfid = NULL;
    }

    /* do we need to free the band list? */
    if (band_hdr != NULL)
    {
        xxx_FreeBands(band_hdr);
        band_hdr = NULL;
    }

NO_CLEANUP:
    /* are we still successful? */
    if (overall_status == SUCCESS)
    {
        IAS_LOG_DEBUG("Conversion of HDF4 to HDF5 image data"
            " completed successfully.");
    }
    else
    {
        IAS_LOG_ERROR("Conversion of HDF4 to HDF5 image data failed.");
    }

    return overall_status;
}


/* 
 * retrieves the HDF4 sensor, HDF4 l1g input filename, and the HDF5 l1g 
 * filename from the command line and then converts the HDF4 to the HDF5.
 */
int main(int argc, char *argv[])
{
    const char *h4_sensor = "ETM";
    int status = ERROR;
    char *h4_input_filename = NULL;
    char *h5_output_filename = NULL;
    const char *program_name = NULL;

    program_name = strrchr(argv[0], '/');
    if (program_name != NULL)
        program_name++;
    else
        program_name = argv[0];

    if ((argc != 3) && (argc != 4))
    {
        printf("usage:   %s <h4_l1g> <h5_l1g> [<h4_sensor>]\n\n", program_name);
        printf("where:\n");
        printf("   <h4_l1g> is the pathname of the HDF4 l1g input file\n");
        printf("   <h5_l1g> is the pathname of the HDF5 l1g output file\n");
        printf("   <h4_sensor> is an optional sensor type of the HDF4 l1g,"
                    " default is ETM\n");
        return EXIT_FAILURE;
    }
    
    h4_input_filename = argv[1];  /* retrieve input file name */
    h5_output_filename = argv[2]; /* retrieve output file name */

    if (argc == 4)
        h4_sensor = argv[3];      /* retrieve the sensor name */

    if (ias_log_initialize(program_name) != SUCCESS)
    {
        IAS_LOG_ERROR("Error initializing logging library");
        return EXIT_FAILURE;
    }

    status = convert_h4_to_h5(h4_input_filename, h4_sensor, 
                                h5_output_filename);
    if (status == SUCCESS)
    {
        IAS_LOG_INFO("Successfully converted %s HDF4 %s to HDF5 %s ", 
                       h4_sensor, h4_input_filename, h5_output_filename);
    }
    else 
    {
        IAS_LOG_ERROR("Failure to convert %s HDF4 %s to HDF5 %s", 
                        h4_sensor, h4_input_filename, h5_output_filename);
    }

    return EXIT_SUCCESS;
}

