
/*****************************************************************************
    PROGRAM: convert_l1g_h5_to_h4

    PURPOSE: To convert the newer Landsat HDF5 formated L1G data to the
             older Landsat HDF4 format.

      NOTES: The older IAS system uses/produces data for L1-5 MSS, L1-5 TM,
             and L7 ETM, using HDF4 formated files .  The newer IAS system
             uses/produces data for L8 OLITIRS using HDF5 formated files.

      USAGE:
        convert_l1g_h5_to_h4 <h5_l1g> <h4_l1g> [<h4_ddr>]

           <h5_l1g> - Pathname of the HDF5 L1G input file
           <h4_l1g> - Pathname of the HDF4 L1G output file
           <h4_ddr> - Optional name for the HDF4 l1g DDR, default is L8_L1G

*****************************************************************************/

/* HDF4 L1G lib headers */
#include "xxx_Sensor.h"
#include "gxx_hdf_io.h"
#include "gxx_ddr.h"
#include "gxx_utility.h"

/* HDF5 L1G lib header */
#include "ias_l1g.h"

/* IAS logging library */
#include "ias_logging.h"
#include "ias_miscellaneous.h"

#include "date_conversions.h"

/*
 * convert the IAS lib data type to an HDF data type
 */
static int get_hdf_dtype
(
    IAS_DATA_TYPE ias_dtype /* I: IAS data type */
)
{
    int hdf_dtype;

    hdf_dtype = ( (ias_dtype == IAS_BYTE) ? DFNT_UINT8 :
                  (ias_dtype == IAS_CHAR) ? DFNT_INT8 :
                  (ias_dtype == IAS_UCHAR) ? DFNT_UINT8 :
                  (ias_dtype == IAS_I2) ? DFNT_INT16 :
                  (ias_dtype == IAS_UI2) ? DFNT_UINT16 :
                  (ias_dtype == IAS_I4) ? DFNT_INT32 :
                  (ias_dtype == IAS_UI4) ? DFNT_UINT32 :
                  (ias_dtype == IAS_R4) ? DFNT_FLOAT32 : 
                  (ias_dtype == IAS_R8) ? DFNT_FLOAT64 : -1);

    if (hdf_dtype == -1)
    {
        IAS_LOG_ERROR("Unsupported IAS data type %d", ias_dtype);
    }

    return hdf_dtype;
}

/*
 * converts the HDF5 L1G Metadata to HDF4 L1G DDR Metadata
 */
int convert_h5_to_h4_ddr
(
    L1GIO *h5_hdfid,                /* I: handle to L1G HDF5 file */
    const char *h4_image_filename,  /* I: Path & name of the output HDF4 1G 
                                          image file*/
    const char *scene_id,            /* I: landsat scene id */
    int nbands,                      /* I: number of bands */
    int *h5_band_number_list,        /* I: enumerated hdf5 band numbers */
    BandNumber *h4_band_number_list, /* I: enumerated hdf4 band numbers */
    int *num_lines,                  /* I: number of lines per band */
    int *num_samples,                /* I: number of samples per band */
    IAS_DATA_TYPE *band_dtype_list   /* I: Data type of each band */
) 
{
    char h4_ddr_filename[FILENAME_MAX];  /* DDR filename */
    xxx_DDRControl_TYPE h4_DDRcontrol;   /* H4 DDR control structure */ 
    IAS_L1G_FILE_METADATA h5_file_meta;
    int status = ERROR; 
    int overall_status = ERROR;
    int band_index;
    xxx_DDR_TYPE *h4_ddr = NULL;
    int i = 0;
    int count;

    status = ias_l1g_get_file_metadata(h5_hdfid, &h5_file_meta);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Cannot retrieve HDF5 file metadata");
        goto CLEANUP;
    }

    memset(&h4_DDRcontrol, 0, sizeof(h4_DDRcontrol));
    h4_DDRcontrol.Count = nbands;
    strncpy(h4_DDRcontrol.vdataName, scene_id, 
             sizeof(h4_DDRcontrol.vdataName));  
    h4_DDRcontrol.vdataName[sizeof(h4_DDRcontrol.vdataName) - 1] = '\0'; 
    strncpy(h4_DDRcontrol.vdataClass, "IAS_DDR", 
             sizeof(h4_DDRcontrol.vdataClass));
    h4_DDRcontrol.vdataClass[sizeof(h4_DDRcontrol.vdataClass) - 1] = '\0';
    h4_DDRcontrol.Interlace = 0;
    h4_DDRcontrol.vdataSize = 471; 

    h4_DDRcontrol.p_DDR = malloc(sizeof(xxx_DDR_TYPE) * nbands);
    if (h4_DDRcontrol.p_DDR == NULL)
    {
        IAS_LOG_ERROR("Cannot allocate H4 DDR metadata memory for %d bands",
                       nbands);
        goto CLEANUP;
    }

    memset( h4_DDRcontrol.p_DDR, 0, sizeof(xxx_DDR_TYPE) * nbands);

    for (band_index = 0; band_index < nbands; band_index++)
    {
        IAS_L1G_BAND_METADATA h5_band_meta;
        int index;

        status = ias_l1g_get_band_metadata(h5_hdfid, 
                       h5_band_number_list[band_index], &h5_band_meta);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Cannot retrieve hdf5 band %d metadata",
                             h5_band_number_list[band_index]);
            goto CLEANUP;
        }

        /* point to the DDR metadata for this band */
        h4_ddr = &h4_DDRcontrol.p_DDR[band_index];

        /* Band number */
        h4_ddr->BandNumber = h4_band_number_list[band_index];
        /* Number of scans in the scene */
        h4_ddr->NumberScansInScene = num_lines[band_index];     
        /* Number of detectors per scan */
        h4_ddr->NumberdetectorsPerScan = 1;
        /* Number of lines in the scene */
        h4_ddr->NumberLinesInScene = num_lines[band_index];     
        /* Number of samples in the scene */
        h4_ddr->NumberSamplesInScene = num_samples[band_index];   

        /* HDF4 unit8, int16, etc.*/
        h4_ddr->DataType = get_hdf_dtype(band_dtype_list[band_index]);
        if (h4_ddr->DataType == -1)
        {
            goto CLEANUP;
        }

        /* Maximum pixel value */
        h4_ddr->MaxPixelValue = h5_band_meta.maximum_pixel_value;

        /* Minimum pixel value */
        h4_ddr->MinPixelValue = h5_band_meta.minimum_pixel_value;

        /* Maximum radiance value */
        h4_ddr->MaxRadianceValue = h5_band_meta.maximum_radiance;

        /* Minimum radiance value */
        h4_ddr->MinRadianceValue = h5_band_meta.minimum_radiance;

        /* Line relative to original image */
        h4_ddr->MasterLine = 0;

        /* Sample relative to original image */
        h4_ddr->MasterSample = 0;

        /* GCTP project code */
        h4_ddr->ProjectCode = h5_file_meta.projection_code;

        /* UTM zone */
        h4_ddr->ZoneCode = h5_file_meta.zone_code;

        /* GCTP datum code.  Note that L1-7 uses WGS_84 while L8 uses WGS84,
           so convert those values. */
        if (strcmp(h5_file_meta.datum, "WGS84") == 0)
        {
            /* Convert WGS84 to WGS_84 */
            count = snprintf(h4_ddr->DatumCode, sizeof(h4_ddr->DatumCode),
                "WGS_84");
        }
        else
        {
            /* Not the expected WGS84, so copy what is there */
            count = snprintf(h4_ddr->DatumCode, sizeof(h4_ddr->DatumCode),
                h5_file_meta.datum);
        }
        if ((count < 0) || (count >= sizeof(h4_ddr->DatumCode)))
        {
            IAS_LOG_WARNING("Datum string truncated because the target buffer "
                "is too small");
        }

        /* Spheroid code, GPS only */
        h4_ddr->SpheroidCode = h5_file_meta.spheroid_code;

        /* Projection units */
        count = snprintf(h4_ddr->ProjectionUnits,
            sizeof(h4_ddr->ProjectionUnits), h5_file_meta.projection_units);
        if ((count < 0) || (count >= sizeof(h4_ddr->ProjectionUnits)))
        {
            IAS_LOG_WARNING("Projection Units truncated because the target "
                "buffer is too small");
        }

        /* GCTP projection parameters */

        /* make sure the number of projection parameters agrees */
        if (XXX_PROJECTION_COEF_SIZE != IAS_PROJ_PARAM_SIZE)
        {
            /* This should never happen since the number of projection
               parameters is dictated by GCTP which both formats are based on */
            IAS_LOG_ERROR("The formats have a different number of projection "
                "parameters");
            goto CLEANUP;
        }

        /* copy all of the projection parameters */
        for (index = 0; index < IAS_PROJ_PARAM_SIZE; index++)
        {
            h4_ddr->ProjectionCoef[index]
                = h5_file_meta.projection_parameters[index]; 
        }

        /* Upper left Y (lat), x (long) */
        h4_ddr->UpperLeft[0] = h5_band_meta.upper_left_y;  
        h4_ddr->UpperLeft[1] = h5_band_meta.upper_left_x;  

        /* Lower left Y (lat), x (long) */
        h4_ddr->LowerLeft[0] = h5_band_meta.lower_left_y;  
        h4_ddr->LowerLeft[1] = h5_band_meta.lower_left_x;  

        /* Upper right Y (lat), x (long) */
        h4_ddr->UpperRight[0] = h5_band_meta.upper_right_y; 
        h4_ddr->UpperRight[1] = h5_band_meta.upper_right_x; 

        /* Lower right Y (lat), x (long) */
        h4_ddr->LowerRight[0] = h5_band_meta.lower_right_y; 
        h4_ddr->LowerRight[1] = h5_band_meta.lower_right_x; 

        /* Projection distance per pixel (y) */
        h4_ddr->ProjectionDistY = h5_band_meta.projection_distance_y;      

        /* Projection distance per pixel (x) */
        h4_ddr->ProjectionDistX = h5_band_meta.projection_distance_x;      

        /* Line subsampling factor */
        h4_ddr->LineSubsampleInc = 1;     

        /* Sample subsampling factor */
        h4_ddr->SampleSubsampleInc = 1;   

        /* Spacecraft source */
        count = snprintf(h4_ddr->SpaceCraftSource,
            sizeof(h4_ddr->SpaceCraftSource), h5_file_meta.spacecraft);
        if ((count < 0) || (count >= sizeof(h4_ddr->SpaceCraftSource)))
        {
            IAS_LOG_WARNING("Spacecraft name truncated because the target "
                "buffer is too small");
        }

        /* Instrument source */
        count = snprintf(h4_ddr->InstrumentSource, 
            sizeof(h4_ddr->InstrumentSource), h5_band_meta.instrument_source);
        if ((count < 0) || (count >= sizeof(h4_ddr->InstrumentSource)))
        {
            IAS_LOG_WARNING("Instrument source truncated because the target "
                "buffer is too small");
        }

        /* Staring WRS path for the scene, GPS only */
        h4_ddr->WRSPath = h5_file_meta.wrs_path;    

        /* Starting fractional row of the WRS scene, GPS only */
        h4_ddr->WRSRows = h5_file_meta.wrs_row;  

        /* Capture direction */
        if (strcmp(h5_file_meta.capture_direction, "DESCENDING") == 0)
        {
            count = snprintf(h4_ddr->CaptureDirection,
                sizeof(h4_ddr->CaptureDirection), XXX_DIRECTION_DESCEND);
        }
        else if (strcmp(h5_file_meta.capture_direction, "ASCENDING") == 0)
        {
            count = snprintf(h4_ddr->CaptureDirection,
                sizeof(h4_ddr->CaptureDirection), XXX_DIRECTION_ASCEND);
        }
        else
        {
            count = snprintf(h4_ddr->CaptureDirection,
                sizeof(h4_ddr->CaptureDirection),
                h5_file_meta.capture_direction);
        }

        if ((count < 0) || (count >= sizeof(h4_ddr->CaptureDirection)))
        {
            IAS_LOG_WARNING("Capture direction truncated because the target "
                "buffer is too small");
        }

        /* Convert the capture date to the expected format, if possible.
           Otherwise, just copy it unmodified */
        status = convert_date_string(
            DATE_FORMAT_YYYY_MM_DD, h5_file_meta.capture_date,
            DATE_FORMAT_DD_MON_YYYY, h4_ddr->CaptureDate,
            sizeof(h4_ddr->CaptureDate));
        if (status == ERROR)
        {
            IAS_LOG_ERROR("Converting Capture Date");
            goto CLEANUP;
        }
        else if (status == INVALID_DATE 
                 && strlen(h5_file_meta.capture_date) > 0)
        {
            IAS_LOG_WARNING("Unable to convert input capture date '%s' to "
                        "expected format - copying unmodified",
                         h5_file_meta.capture_date);
            snprintf(h4_ddr->CaptureDate, sizeof(h4_ddr->CaptureDate),
                 h5_file_meta.capture_date);
        }

        /* Capture time */
        count = snprintf(h4_ddr->CaptureTime, sizeof(h4_ddr->CaptureTime),
            h5_file_meta.capture_time);
        if ((count < 0) || (count >= sizeof(h4_ddr->CaptureTime)))
        {
            IAS_LOG_WARNING("Capture time truncated because the target buffer "
                "is too small");
        }

        /* Field validity flags, looks like just the last one is set */
        h4_ddr->FieldValidityFlags[7] = 1;     

        /* Indicates image type, systematic, precision, or terrain */
        for (i = 0; i < 3; i++) 
        {
            const char *correction[] = {"L1Gs", "L1Gp", "L1Gt"};
            if (strcmp( h5_file_meta.correction_type, correction[i] ) == 0)
            {
                h4_ddr->CorrectionType = i + 1;          
                break;
            }
        }
    }

    /* write the HDF4 L1G DDR metadata file */
    gxx_create_DDR_name(h4_ddr_filename, h4_image_filename );
    status = gxx_WriteDDR(h4_ddr_filename, &h4_DDRcontrol);
    if (status != SUCCESS)
    {
         IAS_LOG_ERROR( "Writing HDF4 DDR" );
         goto CLEANUP;
    }

    overall_status = SUCCESS;

CLEANUP:

    if (h4_DDRcontrol.p_DDR != NULL)
    {
        free(h4_DDRcontrol.p_DDR);
        h4_DDRcontrol.p_DDR = NULL;
    }

    return overall_status;
}


/*
 * reads an HDF5 L1G image file and creates an HDF4 L1G equivalent
 */
int convert_h5_to_h4
( 
    const char *h5_image_filename, /* I: Pathname of the input HDF5 1G file */
    const char *scene_id,          /* I: landsat scene id */
    const char *h4_image_filename  /* I: Pathname of the output HDF4 1G file */
)
{
    HDFIO *h4_hdfid = NULL;    /* Handle of HDF4 file */
    L1GIO *h5_hdfid = NULL;    /* Handle of HDF5 file */
    const IAS_ACCESS_MODE h4_access_mode = IAS_WRITE;  
    const IAS_ACCESS_MODE h5_access_mode = IAS_READ;  
    int nbands;                /* Number of bands */
    unsigned int unsigned_nbands;  /* Copy of nbands as unsigned int for the 
                                      HDF4 routines */
    int h5_band_number_list[IAS_MAX_NBANDS]; /* enumerated hdf5 band numbers */
    BandNumber h4_band_number_list[NBANDS]; /* enumerated band numbers */
    IAS_DATA_TYPE band_dtype_list[IAS_MAX_NBANDS]; /* Data type of each band */
    int  num_lines[IAS_MAX_NBANDS];    /* Number of lines per band */
    int  num_samples[IAS_MAX_NBANDS];  /* Number of samples per band */
    const int start_line = 0;   /* Line to start reading (0-rel) */
    const int start_sample = 0; /* Sample to start reading (0-rel) */
    int  nl;                   /* Number of lines to read */
    int  ns;                   /* Number of samples to read */
    const int subsamp = 1;     /* Subsampling factor */
    void *image_buffer = NULL; /* Image data buffer */
    BAND_HDR_TYPE *band_hdr = NULL; /* Band header this sensor */
    unsigned int band_index;
    L1G_BAND_IO *h5_l1g_band = NULL; /* h5 L1G band handle */
    int sca_count = 1;         /* expecting only 1 SCA */

    int  status;
    int  overall_status = ERROR;
    xxx_FuncPtrs *xxx = NULL;      /* HDF4 Sensor specific functions */
    H4_IAS_SENSOR_TYPE enum_sens;  /* HDF4 enumerated sensor type */

    /* Init the sensor to get the right band list */

    const char *h4_sensor = "XXX";  /* Generic Sensor ID for the output HDF4 
                                       1G file */
    enum_sens = xxx_initialize_sensor_type(h4_sensor);
    if (IAS_SENSOR_UNKNOWN == enum_sens)
    {
        IAS_LOG_ERROR("Unknown HDF4 sensor: %s", h4_sensor);
        goto CLEANUP;
    }

    xxx_initialize_func_ptr();
    xxx = xxx_get_funcs();
    xxx->InitBands();
    band_hdr = xxx_GetBands();

    IAS_LOG_INFO("Reading HDF5 file %s", h5_image_filename);

    h5_hdfid = ias_l1g_open_image(
        h5_image_filename,  /* Path & name of the HDF5 1G image file */
        h5_access_mode);    /* Requested access mode of the HDF5 image */
    
    if (h5_hdfid == NULL)
    {
        IAS_LOG_ERROR("Cannot open %s for reading", h5_image_filename);
        goto CLEANUP;
    }

    /* retrieve the HDF5 l1g band list */
    status = ias_l1g_get_band_list(h5_hdfid, h5_band_number_list,
        IAS_MAX_NBANDS, &nbands);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Cannot get band number list from %s", 
                       h5_image_filename);
        goto CLEANUP;
    }

    for (band_index = 0; band_index < nbands; band_index++)
    {
        BAND_NODE_TYPE *band_node = NULL; /* element of the band linked list.*/
        int h4_band_id = h5_band_number_list[band_index]-1; 

        /* point to the band node for this h4 band id */
        band_node = xxx_GetBandNode(band_hdr, h4_band_id, BAND_ID_TYPE);
        if (band_node == NULL)
        {
            IAS_LOG_ERROR("Cannot find HDF4 band id %d", h4_band_id);
            goto CLEANUP;
        }

        /* save the corresponding h4 band number */
        h4_band_number_list[band_index] = band_node->band_number;

        /* retrieve the rest of the h5 band information */
        status = ias_l1g_get_band_size(h5_hdfid,
            h5_band_number_list[band_index], &band_dtype_list[band_index],
            &sca_count, &num_lines[band_index], &num_samples[band_index]);

        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Cannot get band size info for HDF5 band number %d", 
                           h5_band_number_list[band_index]);
            goto CLEANUP;
        }
        
        /* we're only converting l1g with 1 SCA, not multiple SCA images */ 
        if (sca_count != 1)
        {
            IAS_LOG_ERROR("HDF5 band number %d has an unsupported number of "
                          "SCAs (%d), only 1 supported", 
                           h5_band_number_list[band_index], sca_count);
            goto CLEANUP;
        }
    }

    IAS_LOG_INFO("Writing HDF4 file %s", h4_image_filename);

    unsigned_nbands = nbands;
    h4_hdfid = gxx_open_1g_image(
        h4_image_filename,   /* Path & name of the 1G image file */
        h4_access_mode,      /* Requested access mode of image */
        &unsigned_nbands,    /* Number of bands */
        h4_band_number_list, /* enumerated band numbers */
        band_dtype_list,     /* Data type of each band */
        num_lines,           /* Number of lines per band */
        num_samples);        /* Number of samples per band */

    if (h4_hdfid == NULL)
    {
        IAS_LOG_ERROR("Cannot open %s for writing, exit with failure.", 
              h4_image_filename);
        goto CLEANUP;
    }

    for (band_index = 0; band_index < nbands; band_index++)
    {
        int dt_size = 0;
        const int sca_index = 0;
        int h5_band_number = h5_band_number_list[band_index];

        nl = num_lines[band_index];
        ns = num_samples[band_index];      

        /* determine the byte size of the samples */
        status = ias_misc_get_sizeof_data_type(band_dtype_list[band_index],
                                      &dt_size );
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Invalid sample size id %d", 
                           band_dtype_list[band_index]);
            goto CLEANUP;
        }

        /* allocate space for this band's image */
        image_buffer = malloc( nl * ns * dt_size );
        if (image_buffer == NULL)
        {
            IAS_LOG_ERROR( "Cannot allocate memory for image" );
            goto CLEANUP;
        }

        /* open the HDF5 band */ 
        h5_l1g_band = ias_l1g_open_band( h5_hdfid, h5_band_number,
                &band_dtype_list[band_index], &sca_count, &nl, &ns );
        if (h5_l1g_band == NULL)
        {
            IAS_LOG_ERROR("Reading band %d in input image %s", h5_band_number,
                    h5_image_filename);
            goto CLEANUP;
        }

        IAS_LOG_DEBUG("Reading HDF5 band %d image", h5_band_number);

        /* retrieve the H5 l1g band image */
        status = ias_l1g_read_image(
            h5_l1g_band,    /* L1G IO band structure */
            sca_index,      /* SCA to read to (0-rel) */
            start_line,     /* Line to start reading (0-rel) */
            start_sample,   /* Sample to start reading (0-rel) */
            nl,             /* Number of lines to read */
            ns,             /* Number of samples to read */
            image_buffer);  /* Image Data buffer */
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading HDF5 band %d image", h5_band_number);
            goto CLEANUP;
        }

        IAS_LOG_INFO("Writing HDF4 band %d image", 
                       h4_band_number_list[band_index]);

        status = gxx_write_1g_image(
            h4_hdfid,      /* HDF IO structure */
            h4_band_number_list[band_index], /* H4 band number to write */
            start_line,    /* Line to start reading (0-rel) */
            start_sample,  /* Sample to start reading (0-rel) */
            nl,            /* Number of lines to read */
            ns,            /* Number of samples to read */
            subsamp,       /* Subsampling factor */
            image_buffer); /* Data buffer */
        if (status != SUCCESS)
        { 
            IAS_LOG_ERROR("Cannot write the HDF4 band %d image", 
                           h4_band_number_list[band_index]);
            goto CLEANUP;
        }

        /* close the H5 band */
        status = ias_l1g_close_band(h5_l1g_band);
        h5_l1g_band = NULL;
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Closing HDF5 band %d", h5_band_number);
            goto CLEANUP;
        }

        free(image_buffer);
        image_buffer = NULL;
    }

    /* create the HDF4 DDR metadata from the HDF5 file/band metadata */
    status = convert_h5_to_h4_ddr(h5_hdfid, h4_image_filename, scene_id, 
                 nbands, h5_band_number_list, h4_band_number_list, num_lines, 
                 num_samples, band_dtype_list); 
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Cannot create HDF4 DDR Metadata");
        goto CLEANUP;
    }

    overall_status = SUCCESS;

CLEANUP:

    /* do we have allocated imagery to release? */
    if (image_buffer != NULL)
    {
        free(image_buffer);
        image_buffer = NULL;
    }

    /* do we need to close the HDF4 image? */
    if (h4_hdfid != NULL)
    {
        status = gxx_close_1g_image(h4_hdfid);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Closing HDF4 file");
            overall_status = ERROR;
        }
        h4_hdfid = NULL;
    }

    /* do we need to close the HDF5 band image? */
    if (h5_l1g_band != NULL)
    {
        status = ias_l1g_close_band(h5_l1g_band);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Closing HDF5 band ");
            overall_status = ERROR;
        }
        h5_l1g_band = NULL;
    }
    
    /* do we need to close the HDF5 file? */
    if (h5_hdfid != NULL)
    {
        status = ias_l1g_close_image(h5_hdfid);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Closing HDF5 file");
            overall_status = ERROR;
        }
        h5_hdfid = NULL;
    }

    /* do we need to free the band list? */
    if (band_hdr != NULL)
    {
        xxx_FreeBands(band_hdr);
        band_hdr = NULL;
    }

    /* are we still successful? */
    if (overall_status == SUCCESS)
    {
        IAS_LOG_INFO("Conversion of HDF5 to HDF4 image data completed"
            " successfully");
    }
    else
    {
        IAS_LOG_ERROR("Conversion of HDF5 to HDF4 image data failed");
    }

    return overall_status;
}


/* 
 * retrieves the HDF4 scene id, HDF5 l1g input filename, and the HDF4 l1g 
 * filename from the command line and then converts the HDF5 to the HDF4.
 */
int main(int argc, char *argv[])
{
    char *h4_ddr_name = "L8_L1G";
    int status = ERROR;
    char *h5_input_filename = NULL;
    char *h4_output_filename = NULL;
    const char *program_name = NULL;

    program_name = strrchr(argv[0], '/');
    if (program_name != NULL)
        program_name++;
    else
        program_name = argv[0];

    if ((argc != 3) && (argc != 4))
    {
        printf("usage:   %s <h5_l1g> <h4_l1g> [<h4_ddr>]\n\n", program_name);
        printf("where:\n");
        printf("   <h5_l1g> is the pathname of the HDF5 l1g input file\n");
        printf("   <h4_l1g> is the pathname of the HDF4 l1g output file\n");
        printf("   <h4_ddr> is an optional name for the HDF4 l1g DDR,"
                    " default is %s\n", h4_ddr_name );
        return EXIT_FAILURE;
    }
    
    h5_input_filename = argv[1];  /* retrieve input file name */
    h4_output_filename = argv[2]; /* retrieve output file name */

    if (argc == 4)
        h4_ddr_name = argv[3];    /* retrieve the h4 ddr name */

    if (ias_log_initialize(program_name) != SUCCESS)
    {
        IAS_LOG_ERROR("Error initializing logging library");
        return EXIT_FAILURE;
    }

    status = convert_h5_to_h4(h5_input_filename, h4_ddr_name, 
                                h4_output_filename);
    if (status == SUCCESS)
    {
        IAS_LOG_INFO("Successfully converted HDF5 %s to HDF4 %s ", 
                       h5_input_filename, h4_output_filename);
    }
    else 
    {
        IAS_LOG_ERROR("Failure to convert HDF5 %s to HDF4 %s", 
                        h5_input_filename, h4_output_filename);
    }

    return EXIT_SUCCESS;
}

