/*****************************************************************************
Name: convert_date_string

Purpose: Converts date strings between DD-MON-YYYY and YYYY/MM/DD formats.

Returns: SUCCESS or ERROR
*****************************************************************************/

#include <stdlib.h>
#include <string.h>

#include "ias_const.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"

#include "date_conversions.h"

static const char *month_names[] =
{
    "jan", "feb", "mar", "apr", "may", "jun",
    "jul", "aug", "sep", "oct", "nov", "dec"
};

#define MAX_CHARACTERS_IN_YEAR 4
#define MAX_CHARACTERS_IN_3D_MONTH 3
#define MAX_CHARACTERS_IN_DAY 2
#define NUMBER_OF_MONTHS 12

int convert_date_string
(
    DATE_FORMAT input_format,   /* I: format of input date string */
    const char *input_string,   /* I: Input date to convert */
    DATE_FORMAT output_format,  /* I: format requested for output string */
    char *output_string,        /* O: output date buffer string */
    int output_size             /* I: size of output date buffer */
)
{
    int status;
    int count;
    int year_number = -9999;
    int month_number = -9999;
    int dom_number = -9999;
    int input_length = strlen(input_string);

    switch (input_format)
    {
        case DATE_FORMAT_DD_MON_YYYY:
        {
            char year_str[MAX_CHARACTERS_IN_YEAR+1];
            char month_str[MAX_CHARACTERS_IN_3D_MONTH+1];
            char dom_str[MAX_CHARACTERS_IN_DAY+1];

            /* Make sure the input string is long enough to be valid */
            if (input_length < 11)
                return INVALID_DATE;

            strncpy(dom_str, input_string, sizeof(dom_str) - 1);
            dom_str[sizeof(dom_str)-1] = '\0'; /* Null terminate it */

            int index = 3;
            strncpy(month_str, &input_string[index], sizeof(month_str) - 1);
            month_str[sizeof(month_str)-1] = '\0'; /* Null terminate it */

            index = 7;
            strncpy(year_str, &input_string[index], sizeof(year_str) - 1);
            year_str[sizeof(year_str)-1] = '\0'; /* Null terminate it */

            /* Make sure the month is lowercase */
            ias_misc_convert_to_lowercase(month_str);

            /* Determine the month number */
            month_number = 0;
            for (index = 0; index < NUMBER_OF_MONTHS; index++)
            {
                if (strcmp(month_str, month_names[index]) == 0)
                {
                    month_number = index + 1;
                    break;
                }
            }

            /* Verify a month was found */
            if (month_number == 0)
            {
                /* Let the caller issue any error/warning messages */
                return INVALID_DATE;
            }

            /* Determine the day of month number */
            dom_number = atoi(dom_str);

            /* Determine the year number */
            year_number = atoi(year_str);

            break;
        }

        case DATE_FORMAT_YYYY_MM_DD:
        {
            /* Make sure the input string is long enough to be valid */
            if (input_length < 10)
                return INVALID_DATE;

            year_number = atoi(input_string);

            int index = 5;
            month_number = atoi(&input_string[index]);

            index = 8;
            dom_number = atoi(&input_string[index]);

            break;
        }

        default:
            IAS_LOG_ERROR("Unknown input date format %d", input_format);
            return ERROR;
    }

    status = ias_misc_check_year_month_day(year_number, month_number,
       dom_number);
    if (status != SUCCESS)
    {
        /* Let the caller issue any error/warning messages */
        return INVALID_DATE;
    }
    
    switch (output_format)
    {
        case DATE_FORMAT_DD_MON_YYYY:
        {
            count = snprintf(output_string, output_size,
                "%02d-%s-%04d", dom_number, month_names[month_number-1],
                year_number);
            if ((count < 0) || (count >= output_size))
            {
                IAS_LOG_ERROR("Creating date");
                return ERROR;
            }

            break;
        }

        case DATE_FORMAT_YYYY_MM_DD:
        {
            count = snprintf(output_string, output_size,
                "%04d/%02d/%02d", year_number, month_number, dom_number);
            if ((count < 0) || (count >= output_size))
            {
                IAS_LOG_ERROR("Creating date");
                return ERROR;
            }

            break;
        }

        default:
            IAS_LOG_ERROR("Unknown output date format %d", output_format);
            return ERROR;
    }

    return SUCCESS;
}

