/*****************************************************************************
NAME:       gxx_create_1g_band

PURPOSE:    Create a SDS for the specified 1G band.

RETURN VALUE: 
Type = int
Value           Description
-----           -----------
SUCCESS     Successfully created band SDS
ERROR       Error occurred; band SDS not created

NOTES:      The 1G image is a VGroup containing a SDS for each band.  The 
        data for each SDS are stored in an external element.

        The external files are named image_filename root + SDS band tag
        (SDS_BAND1, etc.) + image_filename extension. e.g., img_B10.l1g

        The band numbers specified in bandno are assumed to be in
        the range b1..b8.

        The HDFIO structure on input must have the hdf_dir_name
        field filled in.
        The HDFIO structure will have the sds_name and sds_id fields for
        the requested band [bandno] filled in.

******************************************************************************
        Landsat 7 Image Assessment System (IAS) Software
                Property of the U.S. Government
                                NASA/GSFC/EDC
*****************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include "ias_logging.h"
#include "ias_const.h"
#include "gxx_hdf_io.h"
#include "xxx_Imagery.h"

int gxx_create_1g_band
(
    HDFIO *hdfid,       /* I/O: SDS id for the band added to structure */
    BandNumber bandno,  /* I: Enumerated band number */
    IAS_DATA_TYPE dtype,     /* I: Data type of the band  */
    int  nl,            /* I: Number of lines for the band*/
    int  ns             /* I: Number of samples for the band*/
)
{
    intn hdf_status;    /* HDF function return status   */
    int32 ref,          /* vgroup element reference */
        tag,            /* vgroup element tag       */
        data_type,      /* data type for 1G images  */
        rank=2,         /* 2-d SDS (one band)       */
        dim_sizes[MAX_VAR_DIMS];/* dimension sizes (nl, ns) */

    int16   fill_i16=0;     /* integer fill value       */
    uint16  fill_ui16=0;    /* unsigned integer fill value */
    int32   fill_i32=0;     /* long fill value      */
    float32 fill_f32=0.0;   /* float fill value     */
    uint8   fill_ui8=0;     /* byte fill value      */
    void    *fill_val;      /* pointer to fill value    */

    char    SDname[MAX_NC_NAME];    /* SDS name         */
    char    ext_filename[FILENAME_MAX];   /* External file name       */

    
    /*
     * Set up SDS name using name constructed from hdf dir file name and band
     * This will also be the filename of the external element
     */
    gxx_create_band_name(hdfid->hdf_dir_name, "SDS", bandno, SDname);
    strcpy(hdfid->sds_name[(int)bandno],SDname);

    dim_sizes[0] = nl;
    dim_sizes[1] = ns;
    if (dtype == IAS_I2) 
    {
        data_type = DFNT_INT16;
        fill_val = (void *)&fill_i16;
    }
    else if (dtype == IAS_UI2) 
    {
        data_type = DFNT_UINT16;
        fill_val = (void *)&fill_ui16;
    }
    else if (dtype == IAS_I4) 
    {
        data_type = DFNT_INT32;
        fill_val = (void *)&fill_i32;
    }
    else if (dtype == IAS_R4) 
    {
        data_type = DFNT_FLOAT32;
        fill_val = (void *)&fill_f32;
    }
    else if (dtype == IAS_BYTE)
    {
        data_type = DFNT_UINT8;
        fill_val = (void *)&fill_ui8;
    }
    else
    {
        IAS_LOG_ERROR ("Invalid image type %d", dtype);
        return ERROR;
    }

    hdfid->sds_id[(int)bandno] = SDcreate(hdfid->sd_id, SDname, data_type, 
                                          rank, dim_sizes);
    if (hdfid->sds_id[(int)bandno] == FAIL)
    {
        IAS_LOG_ERROR ("Error creating SDS %s", SDname);
        return ERROR;
    }

    /*
     * Set the fill value
     */
    hdf_status = SDsetfillvalue(hdfid->sds_id[(int)bandno], fill_val);
    if (hdf_status == FAIL)
    {
        HEprint(stderr,0);
        IAS_LOG_ERROR ("Error setting fill value");
    }

    /*
     * Set the SDS into an external file
     * Use the SDS name as the exernal file name
     */
    gxx_create_band_name(hdfid->hdf_dir_name, "FILE", bandno, ext_filename);

    /* remove the file in case it already exists */
    unlink(ext_filename);

    hdf_status=SDsetexternalfile(hdfid->sds_id[(int)bandno], ext_filename, 0);
    if (hdf_status == FAIL)
    {
        IAS_LOG_ERROR ("Error creating external SDS %s", SDname);
        return ERROR;
    }

    /*
     * Add the SDS to the group
     */
    ref = SDidtoref(hdfid->sds_id[(int)bandno]);
    tag = DFTAG_SD;
    hdf_status = Vaddtagref(hdfid->group_id, tag, ref);
    if (hdf_status == FAIL)
    {
        IAS_LOG_ERROR("Error creating SDS %s", SDname);
        return ERROR;
    }

    return SUCCESS;
}

