/*****************************************************************************
NAME:           gxx_create_band_name

PURPOSE:
gxx_create_band_name forms a band file name or band SDS name given the 
HDF file name.  The "_HDF" at the end of the root HDF file name is replaced
with the appropriate band tag (e.g., "_B10").  If no "_HDF" is at the end of
the file name, the band tag is appended to the end of the root HDF file name.

******************************************************************************
                Landsat 7 Image Assessment System (IAS) Software
                        Property of the U.S. Government
                                NASA/GSFC/EDC
*****************************************************************************/

#include "gxx_hdf_io.h"  /* HDF i/o function prototypes & SDS band tags */
#include "gxx_utility.h" /* createname and splitname prototypes */

void gxx_create_band_name
(
    const char *HDF_name,/* I: HDF name to form band file name from */
    const char *type,    /* I: SDS or FILE name ("SDS" or "FILE") */
    BandNumber bandno,   /* I: enumerated band number */
    char *band_name      /* O: constructed band file name */
)
{
    char directory[FILENAME_MAX]; /* directory name */
    char root[FILENAME_MAX];      /* root file name */
    char extension[FILENAME_MAX]; /* file extension */
    char *temp_char_ptr;          /* temporary character pointer */
    BAND_NODE_TYPE *band_node;    /* This is an element of the band linked 
                                     list. */
    BAND_HDR_TYPE *band_hdr;


    band_hdr = xxx_GetBands();
    band_node = band_hdr->Head;

    while ((band_node != NULL) && (band_node->band_id != bandno))
    {
        band_node = band_node->Next;
    }

    /* form the band file file name from the HDF name */
    gxx_splitname(HDF_name, directory, root, extension);

    /* find the "_HDF" part of the file name if it exits */
    temp_char_ptr = strrchr(root,'_');
    if (temp_char_ptr == NULL)
    {
        /* no "_HDF" so just append the appropriate band tag */
        /*  for SDS  name it is root.tag */
        /*  for file name it is root_tag.ext */
        if (strcmp(type,"SDS") == 0)
            strcpy(extension, band_node->sds_band_name);
        else
        {
            strcat(root, "_");
            strcat(root, band_node->sds_band_name);
        }
    }
    else
    {
        if (strcmp(temp_char_ptr, "_HDF") == 0)
        {
            /* replace the "_HDF" with the band substring */
            /*  for SDS  name it is root.tag so set a NULL where the _HDF is 
                and set the extension to the tag */
            /*  for file name it is root_tag.ext so replace the HDF with the
                tag */
            if (strcmp(type,"SDS") == 0)
            {
                *temp_char_ptr = '\0';
                strcpy(extension, band_node->sds_band_name);
            }
            else
                strcpy(temp_char_ptr+1, band_node->sds_band_name);
        }
        else
        {
            /* doesn't end with "_HDF" so just append the band tag */
            /*  for SDS  name it is root.tag */
            /*  for file name it is root_tag.ext */
            if (strcmp(type,"SDS") == 0)
                strcpy(extension, band_node->sds_band_name);
            else
            {
                strcat(root, "_");
                strcat(root, band_node->sds_band_name);
            }
        }
    }

    /* create band file name */
    gxx_createname(directory, root, extension, band_name);
}
