/*****************************************************************************
NAME:       gxx_open_1g_band

PURPOSE:    Open a band of the 1G image.

RETURN VALUE: 
Type = int32
Value           Description
-----           -----------
band id     HDF SDS id used to access the band
FAIL        HDF failure code--Band not opened (not found)

NOTES:      THIS ROUTINE SHOULD ONLY BE CALLED BY OTHER GPS 1G I/O ROUTINES
        
        The 1G image is a VGroup containing a SDS for each band.  The 
        data for each SDS are stored in an external element.

        The band number specified in bandno is assumed to be in
        the range b1..b8.

        The HDFIO structure will fill in the sds_name field for the 
        requested band [bandno] and return the sds_id.  

        ***This routine assumes the band SDS is already created.
        The band is simply selected (opened) so that other routines
        may access it.

        ***This routine does not issue any error messages, because
        a non-present band may not necessarily be an error (ie
        if the caller is checking to see which of the bands are
        present).

******************************************************************************
        Landsat 7 Image Assessment System (IAS) Software
                Property of the U.S. Government
                                NASA/GSFC/EDC
*****************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include "gxx_hdf_io.h"
#include "ias_logging.h"


int32 gxx_open_1g_band
(
    HDFIO *hdfid,       /* I: HDF info structure for the image   */
    BandNumber bandno   /* I: enumerated band number     */
)
{
    int32 sds_index,    /* HDF SDS index number     */
        sds_id;         /* HDF SDS id           */

    
    /*
     * Initialize the sds id
     */
    sds_id = -1;

    /*
     * Construct the name of the SDS for this band
     */

    gxx_create_band_name(hdfid->hdf_dir_name, "SDS", bandno, 
                         hdfid->sds_name[(int)bandno]);

    /*
     * Select the user-specified band
     */
    sds_index = SDnametoindex(hdfid->sd_id, hdfid->sds_name[(int)bandno]);

    if (sds_index == FAIL)
    {
        /* xxx_LogStatus("GPSLIB", __FILE__, __LINE__, "Band not found"); */
        return FAIL; /* band not found */
    }

    sds_id = SDselect(hdfid->sd_id, sds_index);
    if (sds_id == FAIL)
    {
        IAS_LOG_ERROR ("Band %s not found",hdfid->sds_name[(int)bandno]);
        return FAIL;
    }

    return sds_id;
}
