/*****************************************************************************
NAME:       gxx_read_1g_image

PURPOSE:    Read the specified 1G image.

RETURN VALUE: 
Type = int
Value           Description
-----           -----------
SUCCESS         Successfully read image
ERROR       Error occurred; image not read

NOTES:      The 1G image is a VGroup containing a SDS for each band.  The 
        data for each SDS are stored in an external element.
        The subsampling factor if set to 0 or 1 will result in reading
        the full resolution data.  Set to n will result in reading every
        n'th data point in both dimensions, starting with sl, ss.
        **This routine assumes the band is a 2-dimensional SDS.  If it 
        is of higher dimension, results will be undefined.

******************************************************************************
        Landsat 7 Image Assessment System (IAS) Software
                Property of the U.S. Government
                                NASA/GSFC/EDC
*****************************************************************************/

#include <stdio.h>
#include "ias_const.h"
#include "ias_logging.h"
#include "gxx_hdf_io.h"


int gxx_read_1g_image
(
    const HDFIO *hdfid, /* I: HDF IO structure           */
    BandNumber band,    /* I: (enumerated) band number to read */
    int  sl,            /* I: Line to start reading (0-rel)  */
    int  ss,            /* I: Sample to start reading (0-rel)    */
    int  nl,            /* I: Number of lines to read        */
    int  ns,            /* I: Number of samples to read      */
    int  subsamp,       /* I: Subsampling factor         */
    void *data          /* O: Data buffer            */
)
{
    intn hdf_status;    /* HDF function return status   */
    int32 start[2],     /* starting data location   */
        edge[2],        /* size of data buffer      */
        stride[2];      /* subsampling factor       */

    
    /*
     * Check the structure
     */
    if (hdfid == NULL)
    {
        IAS_LOG_ERROR ("Error reading from 1G image");
        return ERROR;
    }

    /*
     * Set the search path
     */
    HXsetdir(hdfid->extdir);

    /*
     * Set up specification for block to read
     */
    start[0] = sl; /* start is 0-relative */
    start[1] = ss;
    edge[0] = nl;
    edge[1] = ns;

    if (subsamp == 0 || subsamp == 1)
    {
        /* use NULL for stride */
        hdf_status = SDreaddata(hdfid->sds_id[(int)band], start, NULL, edge, 
                                (VOIDP)data);
    }
    else
    {
        stride[0] = subsamp;
        stride[1] = subsamp;
        hdf_status = SDreaddata(hdfid->sds_id[(int)band], start, stride, edge, 
                                (VOIDP)data);
    }

    if (hdf_status == FAIL)
    {
        IAS_LOG_ERROR ("Error reading from SDS %s", 
                        hdfid->sds_name[(int)band]);
        return ERROR;
    }

    return SUCCESS;
}
