/*****************************************************************************
                Landsat 7 Image Assessment System (IAS) Software
                        Property of the U.S. Government
                            NASA / GSFC / Code 510
******************************************************************************

UNIT NAME: xxx_BandGetInfo_etm.c

PURPOSE: This contains ETM-specific band functions.  These are used to build
an ETM band or band list.
*****************************************************************************/

/* System headers */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Local headers */

#include "ias_logging.h"

#include <xxx_Band.h>
#include <xxx_Types.h>
#include <etm/xxx_Sensor_etm.h>

/************************************************************/
/* Given a band list, add band number band_num to the list. */
/************************************************************/

void xxx_CreateBandNode_etm 
(
    BAND_HDR_TYPE *band_list, /* I: band list to put the band node in */
    BAND_OPTION_TYPE option,  /* I: standard band list, or customized one */
    int bnum                  /* I: number of the band to add         */
)
{
    int band_number;       /* This is the physical band number.             */
    int band_long_number;  /* This is a long version of the band number.    */
    BandNumber band_id;    /* This is an enumerated type (eg: b1, etc)      */
                           /* that is often used in GPS. Can it be merged?  */
    BAND_TYPE band_id_rps; /* This is an enumerated type (eg: B1, etc)      */
                           /* akin to band_id that is often used in RPS.    */
    char *band_short_name; /* This is a short version of the band's name.   */
    char *band_name;       /* This is the band's name for display.          */
    char *sds_band_name;   /* This is the SDS band filename section.        */
    char *band_name_ext;   /* This is the band name with a "_".             */
    char *band_name_ext_81;/* This is the band name with a "_" (and "81").  */
    float band_resolution; /* This is in meters.                            */
    band_mode_type band_mode; /* This is an enumerated type (THERMAL...)   */
    temp_corr_type temp_corr; /* enumerated type (NO_TEMP_CORRECTION...)   */
    int num_detectors;     /* This is the number of detectors for the band. */
    long int maxImgPixels; /* These are from rxx_Constants.h.  They will be */
    long int maxCalPixels; /* sent to structures that look like the         */
    char band_id_msg[5];   /* from rxx_BAND_TYPE_MSG,    */
    char IMGnames[5];      /* from RXX_BAND1, etc...     */
    char CALnames[5];      /* from RXX_CAL_BAND1, etc... */
    char SLOvdata[6];      /* from XXX_SLO_BAND1, etc... */
    int joffset;           /* offset - used in IC Trend write               */
    int active;            /* This indicates whether the band is in use.    */
    int first_therm_band = FALSE;  /* Flag to tell if this is B6H           */
    int second_therm_band = FALSE; /* Flag to tell if this is B6H           */
    int status;            


    if ((bnum < 1) || (bnum > 8))
    {
        if ((bnum != 61) && (bnum != 62))
        {
            IAS_LOG_ERROR ("xxx_CreateBandNode failed: input bnum is invalid");
            return;
        }
    }

    if (bnum == 61)
    {
        first_therm_band = TRUE;
        bnum = 6;  /* Get the regular band 6 information first. */
    }

    else if (bnum == 62)
    {
        second_therm_band = TRUE;
        bnum = 6;  /* Get the regular band 6 information first. */
    }

    bnum--;  /* This translates from band number to array index. */

    band_number = band_number_etm[bnum];
    band_long_number = band_long_number_etm[bnum];
    band_id = band_id_etm[bnum];
    band_id_rps = band_id_rps_etm[bnum];
    band_short_name = malloc (strlen (band_short_name_etm[bnum]) + 1);
    strcpy (band_short_name, band_short_name_etm[bnum]);
    band_name = malloc (strlen (band_name_etm[bnum]) + 2); /* room for "L" or
                                                              "H" */
    strcpy (band_name, band_name_etm[bnum]);
    sds_band_name = malloc (strlen (sds_band_name_etm[bnum]) + 1);
    strcpy (sds_band_name, sds_band_name_etm[bnum]);
    band_name_ext = malloc (strlen (band_name_ext_etm[bnum]) + 1);
    strcpy (band_name_ext, band_name_ext_etm[bnum]);
    band_name_ext_81 = malloc (strlen (band_name_ext_81_etm[bnum]) + 1);
    strcpy (band_name_ext_81, band_name_ext_81_etm[bnum]);
    band_resolution = band_resolution_etm[bnum];
    band_mode = band_mode_etm[bnum];
    temp_corr = temp_corr_etm[bnum];
    num_detectors = num_detectors_etm[bnum];
    maxImgPixels = MaxImgPixels_etm[bnum];
    maxCalPixels = MaxCalPixels_etm[bnum];
    strcpy (band_id_msg, band_id_msg_etm[bnum]);
    strcpy (IMGnames, IMGnames_etm[bnum]);
    strcpy (CALnames, CALnames_etm[bnum]);
    strcpy (SLOvdata, SLOvdata_etm[bnum]);
    joffset = joffset_etm[bnum];
    active = active_etm[bnum];


    /* If the band number is b6l or b6h, a few fields need to have the */
    /* standard b6 values overwritten with b6l or b6h values.  The b6l */
    /* and b6h values are not in the source arrays because the source  */
    /* arrays are shared with programs that operate on the physical    */
    /* bands only and which are better with simple band_number loops.  */
    /* A minor special case saves many other cases and allows sharing. */

    if (first_therm_band)
    {
        band_long_number          = 61;
        band_id                   = b6l;
        band_id_rps               = B6L;
        strcpy (band_name,        "BAND 6L"); /* size as the b6 versions. */
        strcpy (sds_band_name,    "B61"); 
        strcpy (band_name_ext,    "_B61"); 
        strcpy (band_name_ext_81, "_B61"); 
        strcpy (band_id_msg,      "B61");
        strcpy (IMGnames,         "B61");
        strcpy (CALnames,         "C61");
        strcpy (SLOvdata,         XXX_SLO_BAND6L);
    }

    else if (second_therm_band)
    {
        band_long_number          = 62;
        band_id                   = b6h;
        band_id_rps               = B6H;
        strcpy (band_name,        "BAND 6H"); /* size as the b6 versions. */
        strcpy (sds_band_name,    "B62"); 
        strcpy (band_name_ext,    "_B62"); 
        strcpy (band_name_ext_81, "_B62"); 
        strcpy (band_id_msg,      "B62");
        strcpy (IMGnames,         "B62");
        strcpy (CALnames,         "C62");
        strcpy (SLOvdata,         XXX_SLO_BAND6H);
    }

    status = xxx_BuildBandNode (band_list,
                           option,
                           band_number,
                           band_long_number,
                           band_id,
                           band_id_rps,
                           band_short_name,
                           band_name,
                           sds_band_name,
                           band_name_ext,
                           band_name_ext_81,
                           band_resolution,
                           band_mode,
                           temp_corr,
                           num_detectors,
                           maxImgPixels,
                           maxCalPixels,
                           band_id_msg,
                           IMGnames,
                           CALnames,
                           SLOvdata,
                           joffset,
                           active);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR ("xxx_CreateBandNode failed");
    }
    if (band_short_name != NULL) free(band_short_name);
    if (band_name != NULL) free (band_name);
    if (sds_band_name != NULL) free (sds_band_name);
    if (band_name_ext != NULL) free (band_name_ext);
    if (band_name_ext_81 != NULL) free (band_name_ext_81);
}


/***************************************************************/
/* Get the "standard" band information, which is a list of     */
/* band nodes in ascending order of band ID.                   */
/***************************************************************/

void xxx_BandGetInfo_etm(void)
{
    BAND_HDR_TYPE *band_hdr;


    band_hdr = xxx_CreateBandHdr (STANDARD);

    xxx_CreateBandNode_etm (band_hdr, STANDARD, 1);
    xxx_CreateBandNode_etm (band_hdr, STANDARD, 2);
    xxx_CreateBandNode_etm (band_hdr, STANDARD, 3);
    xxx_CreateBandNode_etm (band_hdr, STANDARD, 4);
    xxx_CreateBandNode_etm (band_hdr, STANDARD, 5);
    xxx_CreateBandNode_etm (band_hdr, STANDARD, 61);
    xxx_CreateBandNode_etm (band_hdr, STANDARD, 62);
    xxx_CreateBandNode_etm (band_hdr, STANDARD, 7);
    xxx_CreateBandNode_etm (band_hdr, STANDARD, 8);

    /* Set the reference band pointer. */

    band_hdr = xxx_GetBands();
    band_hdr->ref_band = xxx_GetBandNode (band_hdr, b8, BAND_ID_TYPE);

}
