/*****************************************************************************
                Landsat 7 Image Assessment System (IAS) Software
                        Property of the U.S. Government
                            NASA / GSFC / Code 510
******************************************************************************

UNIT NAME:  xxx_ReadDDR.c

PURPOSE:  Read in HDF DDR file and store it in xxx_DDRControl_TYPE.

RETURN VALUE:
Type = xxx_DDRControl_TYPE *

Value           Description
--------------  --------------------------------------------------------------
Positive	Pointer to xxx_DDRControl_TYPE
NULL		Error Reading in DDRs

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hdf.h"
#include "ias_logging.h"
#include "xxx_DDR.h"

extern int32 error_top;


xxx_DDRControl_TYPE *xxx_ReadDDR
(
    char *p_PathName     /* I: Pathname to DDR file */
)
{
    xxx_DDRControl_TYPE *p_DDRControl = NULL;
    const char *p_HEstring;
    int k;

    int32   file_id;
    int32   vdata_id = 0;
    int32   vdata_ref = -1;
    char    fields[XXX_DDR_FIELD_SIZE];
    uint8   *p_databuff = NULL;
    uint8   *p_Ptr;

    if ((p_DDRControl = calloc(1, sizeof(xxx_DDRControl_TYPE))) == NULL)
    {
        IAS_LOG_ERROR ("Can't calloc xxx_DDRControl_TYPE structure");
        return NULL;
    }

    if ((file_id = Hopen(p_PathName, DFACC_RDONLY, 0)) == FAIL)
    {
        p_HEstring = HEstring((hdf_err_code_t)HEvalue(error_top));
        IAS_LOG_ERROR ("Can't open DDR file %s: %s", p_PathName, p_HEstring);
        goto ERROR_PROC;
    }

    Vstart(file_id);

    vdata_ref = -1;
    if ((vdata_ref = VSgetid(file_id, vdata_ref)) == FAIL)
    {
        p_HEstring = HEstring((hdf_err_code_t)HEvalue(error_top));
        IAS_LOG_ERROR ("Can't find vdata in DDR file: %s", p_HEstring);
        goto ERROR_PROC;
    }

    if ((vdata_id = VSattach(file_id, vdata_ref, "r")) == FAIL)
    {
        p_HEstring = HEstring((hdf_err_code_t)HEvalue(error_top));
        IAS_LOG_ERROR ("Error attaching to DDR file: %s", p_HEstring);
        goto ERROR_PROC;
    }

    VSgetclass(vdata_id, p_DDRControl->vdataClass);

    if (VSinquire(vdata_id, &p_DDRControl->Count, &p_DDRControl->Interlace,
                  fields, &p_DDRControl->vdataSize,
                  p_DDRControl->vdataName) == FAIL)
    {
        p_HEstring = HEstring((hdf_err_code_t)HEvalue(error_top));
        IAS_LOG_ERROR ("Error inquiring DDR file: %s", p_HEstring);
        goto ERROR_PROC;
    }

    if (VSsetfields(vdata_id, fields) == FAIL)
    {
        p_HEstring = HEstring((hdf_err_code_t)HEvalue(error_top));
        IAS_LOG_ERROR ("Error setting DDR fields: %s", p_HEstring);
        goto ERROR_PROC;
    }

    if ((p_databuff = (uint8 *)calloc(p_DDRControl->Count,
                                      p_DDRControl->vdataSize)) == NULL)
    {
        IAS_LOG_ERROR ("Can't calloc unpadded databuff");
        goto ERROR_PROC;
    }

    if ((p_DDRControl->p_DDR =
         (xxx_DDR_TYPE *)calloc(p_DDRControl->Count,
                                sizeof(xxx_DDR_TYPE))) == NULL)
    {
        IAS_LOG_ERROR ("Can't calloc xxx_DDR_TYPE structure");
        goto ERROR_PROC;
    }

    if (VSread(vdata_id, p_databuff, p_DDRControl->Count,
               p_DDRControl->Interlace) == FAIL)
    {
        p_HEstring = HEstring((hdf_err_code_t)HEvalue(error_top));
        IAS_LOG_ERROR ("Can't read DDR file: %s", p_HEstring);
        goto ERROR_PROC;
    }

    for (k=0; k<p_DDRControl->Count; k++)
    {
        p_Ptr = p_databuff+(k*p_DDRControl->vdataSize);

        memcpy(&p_DDRControl->p_DDR[k].BandNumber, p_Ptr, sizeof(int8));
        p_Ptr += sizeof(int8);
        memcpy(&p_DDRControl->p_DDR[k].NumberScansInScene, p_Ptr,
               sizeof(int16));
        p_Ptr += sizeof(int16);
        memcpy(&p_DDRControl->p_DDR[k].NumberdetectorsPerScan, p_Ptr,
               sizeof(int8));
        p_Ptr += sizeof(int8);
        memcpy(&p_DDRControl->p_DDR[k].NumberLinesInScene, p_Ptr,
               sizeof(int32));
        p_Ptr += sizeof(int32);
        memcpy(&p_DDRControl->p_DDR[k].NumberSamplesInScene, p_Ptr,
               sizeof(int32));
        p_Ptr += sizeof(int32);
        memcpy(&p_DDRControl->p_DDR[k].DataType, p_Ptr, sizeof(int16));
        p_Ptr += sizeof(int16);
        memcpy(&p_DDRControl->p_DDR[k].MaxPixelValue, p_Ptr, sizeof(float64));
        p_Ptr += sizeof(float64);
        memcpy(&p_DDRControl->p_DDR[k].MinPixelValue, p_Ptr, sizeof(float64));
        p_Ptr += sizeof(float64);
        memcpy(&p_DDRControl->p_DDR[k].MaxRadianceValue, p_Ptr,
               sizeof(float64));
        p_Ptr += sizeof(float64);
        memcpy(&p_DDRControl->p_DDR[k].MinRadianceValue, p_Ptr,
               sizeof(float64));
        p_Ptr += sizeof(float64);
        memcpy(&p_DDRControl->p_DDR[k].MasterLine, p_Ptr, sizeof(int32));
        p_Ptr += sizeof(int32);
        memcpy(&p_DDRControl->p_DDR[k].MasterSample, p_Ptr, sizeof(int32));
        p_Ptr += sizeof(int32);
        memcpy(&p_DDRControl->p_DDR[k].ProjectCode, p_Ptr, sizeof(int16));
        p_Ptr += sizeof(int16);
        memcpy(&p_DDRControl->p_DDR[k].ZoneCode, p_Ptr, sizeof(int16));
        p_Ptr += sizeof(int16);
        memcpy(p_DDRControl->p_DDR[k].DatumCode, p_Ptr,
               sizeof(char8)*XXX_DATUM_CODE_SIZE);
        p_DDRControl->p_DDR[k].DatumCode[XXX_PROJECTION_UNITS_SIZE] = '\0';
        p_Ptr += sizeof(char8)*XXX_DATUM_CODE_SIZE;
        memcpy(&p_DDRControl->p_DDR[k].SpheroidCode, p_Ptr, sizeof(int32));
        p_Ptr += sizeof(int32);
        memcpy(p_DDRControl->p_DDR[k].ProjectionUnits, p_Ptr,
               sizeof(char8)*XXX_PROJECTION_UNITS_SIZE);
        p_DDRControl->p_DDR[k].ProjectionUnits[XXX_PROJECTION_UNITS_SIZE] =
            '\0';
        p_Ptr += sizeof(char8)*XXX_PROJECTION_UNITS_SIZE;
        memcpy(&p_DDRControl->p_DDR[k].ProjectionCoef, p_Ptr,
               sizeof(float64)*XXX_PROJECTION_COEF_SIZE);
        p_Ptr += sizeof(float64)*XXX_PROJECTION_COEF_SIZE;
        memcpy(&p_DDRControl->p_DDR[k].UpperLeft, p_Ptr,
               sizeof(float64)*XXX_POINT_SIZE);
        p_Ptr += sizeof(float64)*XXX_POINT_SIZE;
        memcpy(&p_DDRControl->p_DDR[k].LowerLeft, p_Ptr,
               sizeof(float64)*XXX_POINT_SIZE);
        p_Ptr += sizeof(float64)*XXX_POINT_SIZE;
        memcpy(&p_DDRControl->p_DDR[k].UpperRight, p_Ptr,
               sizeof(float64)*XXX_POINT_SIZE);
        p_Ptr += sizeof(float64)*XXX_POINT_SIZE;
        memcpy(&p_DDRControl->p_DDR[k].LowerRight, p_Ptr,
               sizeof(float64)*XXX_POINT_SIZE);
        p_Ptr += sizeof(float64)*XXX_POINT_SIZE;
        memcpy(&p_DDRControl->p_DDR[k].ProjectionDistY, p_Ptr,
               sizeof(float64));
        p_Ptr += sizeof(float64);
        memcpy(&p_DDRControl->p_DDR[k].ProjectionDistX, p_Ptr,
               sizeof(float64));
        p_Ptr += sizeof(float64);
        memcpy(&p_DDRControl->p_DDR[k].LineSubsampleInc, p_Ptr,
               sizeof(float64));
        p_Ptr += sizeof(float64);
        memcpy(&p_DDRControl->p_DDR[k].SampleSubsampleInc, p_Ptr,
               sizeof(float64));
        p_Ptr += sizeof(float64);
        memcpy(p_DDRControl->p_DDR[k].SpaceCraftSource, p_Ptr,
               sizeof(char8)*XXX_SOURCE_SIZE);
        p_DDRControl->p_DDR[k].SpaceCraftSource[XXX_SOURCE_SIZE] = '\0';
        p_Ptr += sizeof(char8)*XXX_SOURCE_SIZE;
        memcpy(p_DDRControl->p_DDR[k].InstrumentSource, p_Ptr,
               sizeof(char8)*XXX_INSTRUMENT_SIZE);
        p_DDRControl->p_DDR[k].InstrumentSource[XXX_INSTRUMENT_SIZE] = '\0';
        p_Ptr += sizeof(char8)*XXX_INSTRUMENT_SIZE;
        memcpy(&p_DDRControl->p_DDR[k].WRSPath, p_Ptr, sizeof(uint8));
        p_Ptr += sizeof(uint8);
        memcpy(&p_DDRControl->p_DDR[k].WRSRows, p_Ptr, sizeof(float64));
        p_Ptr += sizeof(float64);
        memcpy(p_DDRControl->p_DDR[k].CaptureDirection, p_Ptr,
               sizeof(char8)*XXX_DIRECTION_SIZE);
        p_DDRControl->p_DDR[k].CaptureDirection[XXX_DIRECTION_SIZE] = '\0';
        p_Ptr += sizeof(char8)*XXX_DIRECTION_SIZE;
        memcpy(p_DDRControl->p_DDR[k].CaptureDate, p_Ptr,
               sizeof(char8)*XXX_CAPTURE_DATE);
        p_DDRControl->p_DDR[k].CaptureDate[XXX_CAPTURE_DATE] = '\0';
        p_Ptr += sizeof(char8)*XXX_CAPTURE_DATE;
        memcpy(p_DDRControl->p_DDR[k].CaptureTime, p_Ptr,
               sizeof(char8)*XXX_CAPTURE_TIME);
        p_DDRControl->p_DDR[k].CaptureTime[XXX_CAPTURE_TIME] = '\0';
        p_Ptr += sizeof(char8)*XXX_CAPTURE_TIME;
        memcpy(&p_DDRControl->p_DDR[k].FieldValidityFlags, p_Ptr,
               sizeof(int8)*XXX_VALID_FLAGS);
        p_Ptr += sizeof(int8)*XXX_VALID_FLAGS;
        memcpy(&p_DDRControl->p_DDR[k].CorrectionType, p_Ptr, sizeof(int8));
        p_Ptr += sizeof(int8);
    }

    VSdetach(vdata_id);
    Vend(file_id);
    if (Hclose(file_id) == FAIL)
    {
        p_HEstring = HEstring((hdf_err_code_t)HEvalue(error_top));
        IAS_LOG_ERROR ("Can't close DDR file: %s", p_HEstring);
        goto ERROR_PROC;
    }
    file_id = 0;
    free(p_databuff);
    p_databuff = NULL;

    return p_DDRControl;

  ERROR_PROC:
    VSdetach(vdata_id);
    Vend(file_id);
    Hclose(file_id);

    free(p_databuff);
    free(p_DDRControl->p_DDR);
    free (p_DDRControl);

    return NULL;
}
