/*****************************************************************************
                Landsat 7 Image Assessment System (IAS) Software
                        Property of the U.S. Government
                            NASA / GSFC / Code 510
******************************************************************************

UNIT NAME:  xxx_WriteDDR.c

PURPOSE:  Create a HDF DDR file and write the contains of xxx_DDRControl_TYPE.
      into it.

RETURN VALUE:
Type = int

Value           Description
--------------  --------------------------------------------------------------
SUCCESS         DDR file created
ERROR       Error creating/write the DDRs

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hdf.h"
#include "xxx_DDR.h"

#include "xxx_Types.h"
#include "xxx_Imagery.h"

#include "ias_logging.h"
#include "ias_const.h"

extern int32 error_top;


int xxx_WriteDDR
(
    char *p_PathName,                  /* I: Pathname to DDR file */
    xxx_DDRControl_TYPE *p_DDRControl /* I: DDR information */
)
{
    unsigned int i;
    const char *p_HEstring;

    int32 file_id;
    int32 vdata_id;

    char FieldName[XXX_DDR_FIELD_SIZE];
    uint8 *p_buffer = NULL;
    uint8 *p_Ptr;
    long BufferLine = 0;

    xxx_DDRInfo_TYPE FieldInfo[] = {
        {"BandNumber", DFNT_INT8, 1},
        {"NumberScansInScene", DFNT_INT16, 1},
        {"NumberdetectorsPerScan", DFNT_INT8, 1},
        {"NumberLinesInScene", DFNT_INT32, 1},
        {"NumberSamplesInScene", DFNT_INT32, 1},
        {"DataType", DFNT_INT16, 1},
        {"MaxPixelValue", DFNT_FLOAT64, 1},
        {"MinPixelValue", DFNT_FLOAT64, 1},
        {"MaxRadianceValue", DFNT_FLOAT64, 1},
        {"MinRadianceValue", DFNT_FLOAT64, 1},
        {"MasterLine", DFNT_INT32, 1},
        {"MasterSample", DFNT_INT32, 1},
        {"ProjectCode", DFNT_INT16, 1},
        {"ZoneCode", DFNT_INT16, 1},
        {"DatumCode", DFNT_CHAR8, XXX_DATUM_CODE_SIZE},
        {"SpheroidCode", DFNT_INT32, 1},
        {"ProjectionUnits", DFNT_CHAR8, XXX_PROJECTION_UNITS_SIZE},
        {"ProjectionCoef", DFNT_FLOAT64, XXX_PROJECTION_COEF_SIZE},
        {"UpperLeft", DFNT_FLOAT64, XXX_POINT_SIZE},
        {"LowerLeft", DFNT_FLOAT64, XXX_POINT_SIZE},
        {"UpperRight", DFNT_FLOAT64, XXX_POINT_SIZE},
        {"LowerRight", DFNT_FLOAT64, XXX_POINT_SIZE},
        {"ProjectionDistY", DFNT_FLOAT64, 1},
        {"ProjectionDistX", DFNT_FLOAT64, 1},
        {"LineSubsampleInc", DFNT_FLOAT64, 1},
        {"SampleSubsampleInc", DFNT_FLOAT64, 1},
        {"SpaceCraftSource", DFNT_CHAR8, XXX_SOURCE_SIZE},
        {"InstrumentSource", DFNT_CHAR8, XXX_INSTRUMENT_SIZE},
        {"WRSPath", DFNT_UINT8, 1},
        {"WRSRows", DFNT_FLOAT64, 1},
        {"CaptureDirection", DFNT_CHAR8, XXX_DIRECTION_SIZE},
        {"CaptureDate", DFNT_CHAR8, XXX_CAPTURE_DATE},
        {"CaptureTime", DFNT_CHAR8, XXX_CAPTURE_TIME},
        {"FieldValidityFlags", DFNT_INT8, XXX_VALID_FLAGS},
        {"CorrectionType", DFNT_INT8, 1},
    };

    if ((file_id = Hopen(p_PathName, DFACC_CREATE, 0)) == FAIL)
    {
        p_HEstring = HEstring((hdf_err_code_t)HEvalue(error_top));
        IAS_LOG_ERROR ("Can't open DDR file %s %s: ", p_PathName, p_HEstring);
        return ERROR;
    }

    Vstart(file_id);

    if ((vdata_id = VSattach(file_id, -1, "w")) == FAIL)
    {
        p_HEstring = HEstring((hdf_err_code_t)HEvalue(error_top));
        IAS_LOG_ERROR ("Error attaching to DDR file: %s", p_HEstring);
        goto ERROR_PROC;
    }

    strcpy(FieldName, "");
    for (i=0; i < XXX_ARRAY_SIZE(FieldInfo); i++)
    {
        if (VSfdefine(vdata_id, FieldInfo[i].FieldName, FieldInfo[i].Datatype,
                      FieldInfo[i].Order) == FAIL)
        {
            p_HEstring = HEstring((hdf_err_code_t)HEvalue(error_top));
            IAS_LOG_ERROR ("Error defining  DDR fields: %s", p_HEstring);
            goto ERROR_PROC;
        }
        strcat(FieldName, FieldInfo[i].FieldName);
        BufferLine +=
            FieldInfo[i].Order*xxx_GetDataTypeSize(FieldInfo[i].Datatype);
        strcat(FieldName, ",");
    }
    FieldName[strlen(FieldName)-1] = '\0';

    if ((p_buffer = calloc(p_DDRControl->Count, BufferLine)) == NULL)
    {
        IAS_LOG_ERROR ("Can't calloc DDR buffer");
        goto ERROR_PROC;
    }

    p_Ptr = p_buffer;
    for (i=0; i < (unsigned int)p_DDRControl->Count; i++)
    {
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].BandNumber, sizeof(int8));
        p_Ptr += sizeof(int8);
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].NumberScansInScene,
               sizeof(int16));
        p_Ptr += sizeof(int16);
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].NumberdetectorsPerScan,
               sizeof(int8));
        p_Ptr += sizeof(int8);
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].NumberLinesInScene,
               sizeof(int32));
        p_Ptr += sizeof(int32);
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].NumberSamplesInScene,
               sizeof(int32));
        p_Ptr += sizeof(int32);
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].DataType, sizeof(int16));
        p_Ptr += sizeof(int16);
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].MaxPixelValue, sizeof(float64));
        p_Ptr += sizeof(float64);
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].MinPixelValue, sizeof(float64));
        p_Ptr += sizeof(float64);
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].MaxRadianceValue,
               sizeof(float64));
        p_Ptr += sizeof(float64);
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].MinRadianceValue,
               sizeof(float64));
        p_Ptr += sizeof(float64);
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].MasterLine, sizeof(int32));
        p_Ptr += sizeof(int32);
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].MasterSample, sizeof(int32));
        p_Ptr += sizeof(int32);
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].ProjectCode, sizeof(int16));
        p_Ptr += sizeof(int16);
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].ZoneCode, sizeof(int16));
        p_Ptr += sizeof(int16);
        strcpy((char *)p_Ptr, (char *)p_DDRControl->p_DDR[i].DatumCode);
        p_Ptr += sizeof(char8)*XXX_DATUM_CODE_SIZE;
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].SpheroidCode, sizeof(int32));
        p_Ptr += sizeof(int32);
        strcpy((char *)p_Ptr, (char *)p_DDRControl->p_DDR[i].ProjectionUnits);
        p_Ptr += sizeof(char8)*XXX_PROJECTION_UNITS_SIZE;
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].ProjectionCoef,
               sizeof(float64)*XXX_PROJECTION_COEF_SIZE);
        p_Ptr += sizeof(float64)*XXX_PROJECTION_COEF_SIZE;
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].UpperLeft,
               sizeof(float64)*XXX_POINT_SIZE);
        p_Ptr += sizeof(float64)*XXX_POINT_SIZE;
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].LowerLeft,
               sizeof(float64)*XXX_POINT_SIZE);
        p_Ptr += sizeof(float64)*XXX_POINT_SIZE;
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].UpperRight,
               sizeof(float64)*XXX_POINT_SIZE);
        p_Ptr += sizeof(float64)*XXX_POINT_SIZE;
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].LowerRight,
               sizeof(float64)*XXX_POINT_SIZE);
        p_Ptr += sizeof(float64)*XXX_POINT_SIZE;
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].ProjectionDistY,
               sizeof(float64));
        p_Ptr += sizeof(float64);
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].ProjectionDistX,
               sizeof(float64));
        p_Ptr += sizeof(float64);
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].LineSubsampleInc,
               sizeof(float64));
        p_Ptr += sizeof(float64);
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].SampleSubsampleInc,
               sizeof(float64));
        p_Ptr += sizeof(float64);
        strcpy((char *)p_Ptr, (char *)p_DDRControl->p_DDR[i].SpaceCraftSource);
        p_Ptr += sizeof(char8)*XXX_SOURCE_SIZE;
        strcpy((char *)p_Ptr, (char *)p_DDRControl->p_DDR[i].InstrumentSource);
        p_Ptr += sizeof(char8)*XXX_INSTRUMENT_SIZE;
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].WRSPath, sizeof(uint8));
        p_Ptr += sizeof(uint8);
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].WRSRows, sizeof(float64));
        p_Ptr += sizeof(float64);
        strcpy((char *)p_Ptr, (char *)p_DDRControl->p_DDR[i].CaptureDirection);
        p_Ptr += sizeof(char8)*XXX_DIRECTION_SIZE;
        strcpy((char *)p_Ptr, (char *)p_DDRControl->p_DDR[i].CaptureDate);
        p_Ptr += sizeof(char8)*XXX_CAPTURE_DATE;
        strcpy((char *)p_Ptr,(char *) p_DDRControl->p_DDR[i].CaptureTime);
        p_Ptr += sizeof(char8)*XXX_CAPTURE_TIME;
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].FieldValidityFlags,
               sizeof(int8)*XXX_VALID_FLAGS);
        p_Ptr += sizeof(int8)*XXX_VALID_FLAGS;
        memcpy(p_Ptr, &p_DDRControl->p_DDR[i].CorrectionType, sizeof(int8));
        p_Ptr += sizeof(int8);
    }

    if (VSsetfields(vdata_id, FieldName) == FAIL)
    {
        p_HEstring = HEstring((hdf_err_code_t)HEvalue(error_top));
        IAS_LOG_ERROR ("Error setting DDR fields: %s", p_HEstring);
        goto ERROR_PROC;
    }

    if (VSwrite(vdata_id, p_buffer, p_DDRControl->Count, FULL_INTERLACE)
        == FAIL)
    {
        p_HEstring = HEstring((hdf_err_code_t)HEvalue(error_top));
        IAS_LOG_ERROR ("Can't write DDR file: %s", p_HEstring);
        goto ERROR_PROC;
    }
    free(p_buffer);

    VSsetname(vdata_id, p_DDRControl->vdataName);
    VSsetclass(vdata_id, p_DDRControl->vdataClass);
    VSdetach(vdata_id);

    Vend(file_id);

    if (Hclose(file_id) == FAIL)
    {
        p_HEstring = HEstring((hdf_err_code_t)HEvalue(error_top));
        IAS_LOG_ERROR ("Can't close DDR file: %s", p_HEstring);
        return ERROR;
    }

    return SUCCESS;

  ERROR_PROC:
    free(p_buffer);

    VSdetach(vdata_id);
    Vend(file_id);
    Hclose(file_id);

    return ERROR;
}
