#!/usr/bin/perl -w
#############################################################################
# Name: add_gcps
#
# Description: Given a directory with a GCPLib file and the matching image
#   chips, add the GCPs and image chips to the master version.
#
# Parameters:
#   Master GCPLib file for the WRS-2 path/row being updated (it is assumed
#       that the image chips are in the same directory)
#   Update GCPLib file with the image chips in the same directory
#   Optional "test only" flag
#   Optional date updated parameter to use (default to today's date)
#   Optional overall GCP version number
#   Optional override select method parameter to indicate the select method
#       and moravec rank in the update file should be overridden by the value
#       from the master file for the same GCP ID. (helpful for adding 60m chips)
#   Optional "chips only" parameter to indicate that only new image chips
#       should be present in the update file.  If a chip is found that doesn't
#       match a GCP in the master file, it is an error. (helpful for adding 
#       60m chips)
#############################################################################

use strict;
use Getopt::Long;           # Command line parsing support
use File::Basename;         # Basename routine
use File::Copy;             # Copy and move routines
use List::Util qw(max);     # Max routine
use Logger;                 # Logging module
use ias_const;              # Define SUCCESS and ERROR
use ias_gcp qw(ias_gcp_write_gcplib); # GCPLib write routine
use manage_gcps_utils;

# Declare program constant
use constant PROG => basename($0);

# Read the optional parameters
my $test_only = 0;          # Flag to only test the additions
my $date_updated;           # Date updated value to use
my $overall_version;        # Overall version number for the output file
my $override_select_method = 0; # Flag to override the select method and moravec
                            # rank in the update file with the value from other
                            # image chips already in the master file
my $chips_only = 0;         # Flag to indicate the updates are adding chips to
                            # existing GCP IDs
my $verbose = 0;            # Flag to provide more messages

if (!GetOptions('test' => \$test_only, 'date_updated=s' => \$date_updated, 
        'version=i' => \$overall_version,
        'override_select_method' => \$override_select_method,
        'verbose' => \$verbose, 'chips_only' => \$chips_only))
{
    # Error reading the options, so issue the usage message and exit
    usage();
    exit FAILURE;
}

# Confirm there are enough parameters remaining for the required parameters
if ($#ARGV != 1)
{
    # Error since no GCPLib filename provided
    usage();
    exit FAILURE;
}

# Set the level of messages from the manage_gcps_utils module
set_verbose_state($verbose);

# Get the master GCPLib filename
my $master_gcplib = $ARGV[0];

# Get the update GCPLib filename
my $update_gcplib = $ARGV[1];

# Set up the logger
my $logger = Logger->new("STDOUT");
if (!$logger)
{
    print STDERR PROG . ": ERROR:  Failed to establish logger object for "
        . "STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name(PROG);
set_default_log($logger);

# Get the date updated to use
my $status;
($status, $date_updated) = get_date_updated($date_updated);
if ($status != SUCCESS)
{
    $logger->close_log();
    usage();
    exit FAILURE;
}

$logger->info_msg("Master GCPLib: $master_gcplib");
$logger->info_msg("Adding GCPLib: $update_gcplib");

if (defined $date_updated)
{
    $logger->info_msg("Using date_updated of $date_updated");
}
else
{
    $logger->info_msg("Using date_updated from update file records");
}

# Read the master file and validate the contents
my $master_file = read_gcplib_file($master_gcplib, MASTER_FILE);
if (!defined $master_file)
{
    $logger->fatal_msg("Error reading master GCPLib file $master_gcplib");
    $logger->close_log();
    exit FAILURE;
}

# Read the update file and validate the contents
my $update_file = read_gcplib_file($update_gcplib, UPDATE_FILE);
if (!defined $update_file)
{
    $logger->fatal_msg("Error reading update GCPLib file $update_gcplib");
    $logger->close_log();
    exit FAILURE;
}

# Set the overall version to the maximum of the master and update files if
# it wasn't specified on the command line
if (!defined $overall_version)
{
    $overall_version = max $update_file->{'version'}, $master_file->{'version'};
}

$logger->info_msg("Assigning new GCPs IDs") if ($verbose);

# Assign the GCP IDs to the update GCPs
if (assign_gcp_ids($master_file->{'lookup'}, $update_file->{'lookup'},
    $chips_only) != SUCCESS)
{
    $logger->fatal_msg("Failed assigning GCP IDs for the new GCPs");
    $logger->close_log();
    exit FAILURE;
}

$logger->info_msg("Assigning new image chip IDs") if ($verbose);

# Get the image chip ID assignments
if (assign_image_chip_ids($master_file->{'lookup'}, $update_file->{'lookup'})
    != SUCCESS)
{
    $logger->fatal_msg("Failed assigning image chip ids");
    $logger->close_log();
    exit FAILURE;
}

$logger->info_msg("Merging GCPs...") if ($verbose);

# Add the update GCPs to the master set, including updating the date updated
# fields as appropriate.
merge_gcps($master_file->{'lookup'}, $master_file->{'records'},
    $update_file->{'lookup'}, $date_updated, $override_select_method);

# Regenerate the master hash
$master_file->{'lookup'} = create_record_index($master_file->{'records'});

$logger->info_msg("Verifying merged GCPs...") if ($verbose);

# Verify the master GCPs are consistent after the merge
if (!verify_gcps_are_consistent($master_file->{'lookup'}, MASTER_FILE))
{
    $logger->fatal_msg("Master GCPs in $master_gcplib are not consistent");
    $logger->close_log();
    exit FAILURE;
}

$logger->info_msg("Verifing image chip files") if ($verbose);

# Verify each chip that needs to be copied
foreach my $rec (@{$master_file->{'records'}})
{
    # If there is an original chip name field, it means the chip needs to
    # be copied to the target directory
    if (defined $rec->{'orig_chip_name'})
    {
        my $source = $update_file->{'dir'} . $rec->{'orig_chip_name'};
        if (verify_image_chip_data($rec, $source) != SUCCESS)
        {
            $logger->fatal_msg("Verifying image chip data");
            exit FAILURE;
        }
    }
}

# By default, replace the original GCPLib file
my $output_filename = $master_gcplib . ".updated";

if (!$test_only)
{
    $logger->info_msg("Adding image chip files to the master copy")
        if ($verbose);
    
    # Copy the new image chips to the master directory with the correct name
    my $chips_update_file_created = 0;
    foreach my $rec (@{$master_file->{'records'}})
    {
        # If there is an original chip name field, it means the chip needs to
        # be copied to the target directory
        if (defined $rec->{'orig_chip_name'})
        {
            # Create chips update file
            if (!$chips_update_file_created)
            {
                if (create_chips_updated_file($master_file->{'dir'}) != SUCCESS)
                {
                    $logger->fatal_msg("Error creating chips updated file");
                    $logger->close_log();
                    exit FAILURE;
                }
                $chips_update_file_created = 1;
            }

            # Copy the chip to the target directory and assigned chip name
            my $source = $update_file->{'dir'} . $rec->{'orig_chip_name'};
            my $target = $master_file->{'dir'} . $rec->{'chip_name'};
            $logger->info_msg("Copying $source to $target") if ($verbose);
            if (!copy($source, $target))
            {
                $logger->fatal_msg("Error copying $source to $target");
                $logger->close_log();
                exit FAILURE;
            }
        }
    }
}
else
{
    # Just testing, override the output GCPLib filename with a test filename
    $output_filename = "GCPLib.test";
    $logger->info_msg("Test mode: Writing test output file");
}

# Scrub the extra fields from the master records to allow the write routine
# to work
foreach my $rec (@{$master_file->{'records'}})
{
    delete $rec->{'orig_point_id'};
    delete $rec->{'orig_chip_name'};
}

# Write the new master GCPLib file
$logger->info_msg("writing updated file $output_filename") if ($verbose);

if (ias_gcp_write_gcplib($output_filename, $overall_version,
    $master_file->{'records'}) != SUCCESS)
{
    $logger->fatal_msg("Error writing test GCPLib file $output_filename");
    $logger->close_log();
    exit FAILURE;
}

$logger->close_log();

exit SUCCESS;

##############################################################################
# Name: usage
# 
# Description: Prints out the usage message for the application
#
##############################################################################
sub usage
{
    print "\nUsage: ". PROG . " master_GCPLib_filename update_GCPLib_filename "
        . "[--test] \\\n"
        . "    [--date_updated mm-dd-yyyy|use_file_date] "
        . "[--version version_number] \\\n"
        . "    [--override_select_method] [--verbose] [--chips_only]\n\n"
        . "where master_GCPLib_filename is the master file to update\n\n"
        . "      update_GCPLib_filename contains the updates to apply\n\n"
        . "      --test indicates the output should be written to a test "
        . "output file\n      instead of updating the master\n\n"
        . "      --date_updated is the date updated to use for the new GCPs "
        . "and image chips\n\n"
        . "      --version is the version number to set in the output GCPLib "
        . "file (defaults\n      to the highest version in the two input "
        . "GCPLib files)\n\n"
        . "      --override_select_method is a flag to override the select "
        . "method and\n      moravec rank from the update file with the "
        . "method present in the master\n      file for the same GCP ID\n\n"
        . "      --verbose requests more informational output\n\n"
        . "      --chips_only indicates that only new chips may be added to "
        . "existing GCP\n      IDs.  If new GCPs are found, it is considered "
        . "an error.\n\n";
}

##############################################################################
# Name: assign_gcp_ids
#
# Description: Assign GCP IDs to the new GCPs.  Match the new GCPs with
#   compatible GCPs in the master list.  If a matching GCP isn't found, a new
#   unique GCP ID is assigned.  The new GCP records are updated with the new
#   GCP ID.
#
# Parameters:
#   - Master GCP ID lookup hash reference
#   - Update GCP ID lookup hash reference (updated with assigned GCP IDs on
#       return)
#   - Flag to indicate that only new chips are being added to existing GCP IDs
#
# Returns:
#   SUCCESS/ERROR
##############################################################################
sub assign_gcp_ids
{
    my $master_hash = shift;
    my $update_hash = shift;
    my $chips_only = shift;

    # Create a hash to map GCP IDs in the update hash to the master hash
    my %update_to_master_hash = ();

    # List of GCP IDs in the update hash that need real GCP IDs assigned
    my @update_ids_needed;

    # For each update GCP, search of a matching GCP in the master list
    foreach my $new_id (sort keys %$update_hash)
    {
        my $id_assigned = 0;
        my $new_gcp = $update_hash->{$new_id}[0];

        # First check whether the update GCP ID matches an existing one
        if (defined $master_hash->{$new_id})
        {
            # The GCP ID exists, check whether it agrees in the rest of the
            # important metadata
            if (is_compatible_gcp($new_gcp, $master_hash->{$new_id}[0]))
            {
                $logger->info_msg("Update GCP ID $new_id agrees with matching "
                    . "master GCP ID") if ($verbose);
                $update_to_master_hash{$new_id} = $new_id;
                $id_assigned = 1;
            }
        }
        elsif ($chips_only)
        {
            $logger->fatal_msg("The update GCP ID $new_id does not match a "
                . "GCP ID in the master file");
            return ERROR;
        }

        # If the chips_only flag is set, it is an error at this point if
        # the GCP ID provided wasn't compatible
        if (!$id_assigned && $chips_only)
        {
            $logger->fatal_msg("The update GCP ID $new_id does not agree with "
                . "the same GCP ID in the master file");
            return ERROR;
        }

        if (!$id_assigned)
        {
            # Look for a compatible GCP with a different GCP ID
            foreach my $master_id (keys %$master_hash)
            {
                my $master_gcp = $master_hash->{$master_id}[0];

                if (is_compatible_gcp($new_gcp, $master_gcp))
                {
                    $logger->warn_msg("Update GCP ID $new_id metadata matches "
                        . "master GCP ID $master_id, so assigning it that ID");
                    $update_to_master_hash{$new_id} = $master_id;
                    $id_assigned = 1;
                    last;
                }
            }
        }

        if (!$id_assigned)
        {
            # A matching GCP wasn't found, so add it to the list that needs
            # a new GCP ID
            push(@update_ids_needed, $new_id);
        }
        else
        {
            # Make sure the matching GCP is not flagged as inactive since
            # adding a chip to an existing GCP that is inactive is suspicious
            my $master_id = $update_to_master_hash{$new_id};
            my $master_gcp = $master_hash->{$master_id}[0];
            if ($master_gcp->{'point_active'} ne "Y")
            {
                $logger->fatal_msg("GCP ID $master_id is getting a new image "
                    . "chip but is not active - that is against the rules");
                return ERROR;
            }
        }
    }

    # Assign GCP IDs if any are needed
    if (scalar @update_ids_needed > 0)
    {
        # Get enough free GCP IDs to make the assignments
        my ($status, $free_gcp_ids) = get_free_gcp_ids($master_hash,
            $update_hash, scalar @update_ids_needed);
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Error getting new GCP IDs");
            return ERROR;
        }

        # Update the lookup map to hold the new assignments
        foreach my $id (@update_ids_needed)
        {
            my $assigned_id = shift(@$free_gcp_ids);
            $update_to_master_hash{$id} = $assigned_id;

            $logger->info_msg("Update GCP ID $id assigned a new GCP ID of "
                . $assigned_id);
        }
    }

    # Now that all GCP IDs have been assigned, update the original records
    foreach my $new_id (keys %$update_hash)
    {
        my $gcp_array = $update_hash->{$new_id};
        foreach my $gcp_rec (@$gcp_array)
        {
            # Remember the original GCP ID used in case it is later useful for
            # error reporting
            $gcp_rec->{'orig_point_id'} = $gcp_rec->{'point_id'};

            # Set the new ID
            $gcp_rec->{'point_id'} = $update_to_master_hash{$new_id};
        }
    }

    return SUCCESS;
}

##############################################################################
# Name: get_free_gcp_ids
#
# Description: Given the master hash of assigned GCP IDs, return a list of the
#   next free GCP IDs.  Since there are some gaps in the current GCP IDs
#   assigned, the search needs to look for gaps in the list.
#
# Parameters:
#   - Master hash reference
#   - Update hash reference - only used if the master hash is empty (which can
#       happen if there are not GCPs in the target path/row)
#   - Count of IDs needed
#
# Returns:
#   SUCCESS/ERROR
#   An array of free GCP IDs
##############################################################################
sub get_free_gcp_ids
{
    my $master_hash = shift;
    my $update_hash = shift;
    my $ids_wanted = shift;

    my @ids;        # list of free IDs

    # Get the WRS path/row from one of the GCPs
    my $gcp_id = (keys %$master_hash)[0];
    my $path;
    my $row;

    # If the GCP ID is defined, get the path/row from that GCP record (it might
    # not be defined if the master file is empty).
    if (defined $gcp_id)
    {
        $path = $master_hash->{$gcp_id}->[0]{'point_path'};
        $row = $master_hash->{$gcp_id}->[0]{'point_row'};
    }
    else
    {
        # The master file was empty, so get the path/row from the update file
        $gcp_id = (keys %$update_hash)[0];
        $path = $update_hash->{$gcp_id}->[0]{'point_path'};
        $row = $update_hash->{$gcp_id}->[0]{'point_row'};
    }

    # Scan the list looking for free ids
    for (my $index = 1; $index < 10000; $index++)
    {
        # Build the next GCP ID
        my $next_id = sprintf("%03d%03d%04d", $path, $row, $index);

        if (not defined $master_hash->{$next_id})
        {
            push(@ids, $next_id);
            return SUCCESS, \@ids if ((scalar @ids) >= $ids_wanted);
        }
    }

    # Couldn't find enough free ids
    $logger->fatal_msg("There are no free GCP IDs");
    return ERROR, undef;
}

##############################################################################
# Name: get_free_chip_ids
#
# Description: Given an array of GCP record references, return a list of the
#   next free image chip IDs.  This routine could fail if the image chips being
#   added goes over 99 chips for a single GCP.
#
# Parameters:
#   - GCP ID (to act as the chip ID basis in case an empty record array is
#       passed in)
#   - GCP record array reference
#   - count of needed image chip IDs
#
# Returns:
#   SUCCESS/ERROR
#   Array with requested image chip IDs
##############################################################################
sub get_free_chip_ids
{
    my $gcp_id = shift;
    my $gcp_records = shift;
    my $needed_count = shift;

    my @existing_ids;
    my @assigned_ids;

    # Get a sorted array with the existing image chip IDs
    foreach my $rec (@$gcp_records)
    {
        push(@existing_ids, $rec->{'chip_name'});
    }
    @existing_ids = sort(@existing_ids);

    # Look for gaps in the existing IDs list.  In theory, there shouldn't be
    # any gaps, but we'll check just in case there are.
    my $last_seq = 1;
    foreach my $id (@existing_ids)
    {
        if ($id =~ /(\d{10})_(\d{2})/)
        {
            my $seq = $2;
            while ($last_seq < $seq)
            {
                # Found a gap in the list of IDs, so it is a free image chip ID
                my $new_id = sprintf("%010d_%02d", $gcp_id, $last_seq);
                push (@assigned_ids, $new_id);

                # Quit if we found enough IDs
                last if (scalar(@assigned_ids) == $needed_count);

                $last_seq++;
            }
            $last_seq = $seq + 1;
        }
        else
        {
            # The pattern didn't match, so the ID isn't legal and something
            # needs to be fixed in the master dataset
            $logger->fatal_msg("Image Chip ID '$id' is invalid");
            return ERROR, undef;
        }
    }

    # If more chips are needed, generate them at the end of the assigned ones
    for ( ; scalar(@assigned_ids) < $needed_count; $last_seq++)
    {
        my $new_id = sprintf("%010d_%02d", $gcp_id, $last_seq);
        push(@assigned_ids, $new_id);
    }

    # Check for overflow of the chip id sequence
    if ($last_seq > 99)
    {
        $logger->fatal_msg("Ran out of image chip ids for GCP ID $gcp_id");
        return ERROR, undef
    }

    return SUCCESS, @assigned_ids;
}

##############################################################################
# Name: assign_image_chip_ids
#
# Description: Assign unique image chip ids to the new image chips.
#
# Parameters:
#   - Master GCP ID lookup hash reference
#   - Update GCP ID lookup hash reference
#
# Returns:
#   SUCCESS/ERROR
##############################################################################
sub assign_image_chip_ids
{
    my $master_hash = shift;
    my $update_hash = shift;

    foreach my $update_id (sort keys %$update_hash)
    {
        # Get the GCP ID from the first entry in the update hash.  By now
        # the records in the update hash have been assigned their final
        # GCP IDs but the update hash still uses the original GCP ID.
        my $master_id = $update_hash->{$update_id}->[0]{'point_id'};

        # Get the reference to the arrays of different image chips for this
        # GCP
        my $master_array = $master_hash->{$master_id};
        my $update_array = $update_hash->{$update_id};

        # Get a count of how many chip ids are needed
        my $needed_chip_names = scalar @$update_array;

        # Get an array of the next free image chip ids
        my ($status, @free_chip_ids) = get_free_chip_ids($master_id,
            $master_array, $needed_chip_names);
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Unable to assign chip ids");
            return ERROR;
        }

        # Assign the image chip ids in the update records while preserving
        # the original name since that will be needed to copy the image
        # chip files later
        foreach my $rec (@$update_array)
        {
            my $chip_id = shift(@free_chip_ids);
            $rec->{'orig_chip_name'} = $rec->{'chip_name'};
            $rec->{'chip_name'} = $chip_id;
        }
    }

    return SUCCESS;
}

##############################################################################
# Name: compare_gcp_records_for_sort
#
# Description: Comparison subroutine for using sort to sort GCP records.
#   The sort primary sort is done using GCP IDs and falling back to image chip
#   IDs if the GCP IDs are the same.
#
# Parameters:
#   References to the two GCP records to compare
#
# Returns:
#   -1 less than
#    0 equal
#    1 greater than
##############################################################################
sub compare_gcp_records_for_sort($$)
{
    my $rec1 = shift;
    my $rec2 = shift;

    # Check the GCP IDs for less than and greater than
    my $id1 = int $rec1->{'point_id'};
    my $id2 = int $rec2->{'point_id'};

    return -1 if ($id1 lt $id2);
    return 1 if ($id1 gt $id2);

    # GCP IDs are equal, so use the image chip IDs
    $id1 = $rec1->{'chip_name'};
    $id2 = $rec2->{'chip_name'};

    return -1 if ($id1 lt $id2);
    return 1 if ($id1 gt $id2);

    # Must be equal (not likely to happen since it would be a violation
    # of the rules)
    return 0
}

##############################################################################
# Name: merge_gcps
#
# Description: Merge the set of update GCPs into the master set of GCPs.
#   It also sets the new date updated and version number in the image chips and
#   GCPs 
#
# Parameters:
#   - Master set GCP ID lookup hash reference
#       - After this routine returns, this lookup needs to be regenerated
#   - Master array of records reference
#       - The update records are added to this and the array is sorted
#   - Update set GCP ID lookup hash reference
#   - Date updated
#   - Flag to override the select method in the update file with values from
#     other chips with the same GCP ID in the master file
#
# Returns: nothing
#
##############################################################################
sub merge_gcps
{
    my $master_hash = shift;
    my $master_records = shift;
    my $update_hash = shift;
    my $date_updated = shift;
    my $override_select_method = shift;

    # Get the list of GCP metadata fields
    my @gcp_members = get_gcp_metadata_members();

    foreach my $update_id (sort keys %$update_hash)
    {
        my $record_array = $update_hash->{$update_id};
        my $master_id = $record_array->[0]{'point_id'};
        my $master_array = $master_hash->{$master_id};

        if (defined $master_array)
        {
            # Master record exists, so copy all the GCP metadata from it to
            # the update record to make sure it is identical (most fields are
            # already identical, but some might not be, like the point_type).
            # The GCP date updated is copied as well since the GCP point hasn't
            # changed.
            my $master_rec = $master_array->[0];

            foreach my $update_rec (@$record_array)
            {
                foreach my $field (@gcp_members)
                {
                    $update_rec->{$field} = $master_rec->{$field};
                }
                # Set the chip date updated to the provided date.  The 
                # point_date_updated is not updated since the metadata
                # related to that has not changed.  Also set the chip version
                # to 1 since it is new.
                $update_rec->{'chip_date_updated'} = $date_updated 
                    if (defined $date_updated);
                $update_rec->{'chip_version'} = 1;

                # If the override_select_method flag is defined, copy the 
                # select method and moravec rank from the master
                if ($override_select_method)
                {
                    $update_rec->{'chip_select_method'}
                        = $master_rec->{'chip_select_method'};
                    $update_rec->{'moravec_rank'}
                        = $master_rec->{'moravec_rank'};
                }

                push(@$master_records, $update_rec);
            }
        }
        else
        {
            # New GCP ID, so add the records to the master records array and
            # make sure the date updated and version fields are correct
            foreach my $update_rec (@$record_array)
            {
                if (defined $date_updated)
                {
                    # The date updated is defined, so override the one in
                    # the record.  If it isn't defined, the date from the
                    # record is used without updating it.
                    $update_rec->{'point_date_updated'} = $date_updated;
                    $update_rec->{'chip_date_updated'} = $date_updated;
                }
                $update_rec->{'point_version'} = 1;
                $update_rec->{'chip_version'} = 1;
                push(@$master_records, $update_rec);
            }
        }
    }

    # Sort the master records after the updates have been made
    my @sorted_gcps = sort compare_gcp_records_for_sort @$master_records;

    # Delete the entries in the master array and replace it with the sorted 
    # array
    @$master_records = @sorted_gcps;
}

