#!/usr/bin/perl -w
##############################################################################
# Name: convert_old_l8_gcplib_to_new.pl
#
# Purpose: Convert the old L8 GCPLib format to the new one
#
# Required parameters:
#   - Old GCPLib filename (including path if needed)
#   - New GCPLib filename (including path if needed)
#
# Optional parameters
#   --gcp_source value
#       Used to override the GCP source metadata field
#   --sensor value
#       Used to set the image chip sensor metadata.  Note that some of the
#       GCPLib files provided by cal/val mistakenly had the sensor in the 
#       GCP source field.  Providing "use_source" for this parameter will get
#       the sensor from the source field and require the --gcp_source parameter
#       to be provided to set it correctly.
#   --rename_chips
#       Flag to set the chip_name field to follow the new naming standard and
#       copy the image chip file to the new name in the same directory as the
#       new GCPLib file.
#   --set_abs
#       Flag to override the abs_rel_flag metadata field to be "ABS" since
#       they should all really be "ABS" and some files have "REL".
#   --update_reason value
#       Value to set the update reason to (defaults to ORIGINAL).  Defined
#       values are ORIGINAL, TRIG_ERROR, NEW_POINT, TEMPORAL, and SEASONAL.
#   --select_method value
#       Value to set the select method to (defaults to UNKNOWN)
##############################################################################
use strict;

use Getopt::Long;           # command line parsing support
use File::Basename;         # basename routine
use File::Copy;             # copy and move routine
use Logger;                 # Logging module
use ias_const;              # define SUCCESS and ERROR
use ias_gcp_old qw(ias_gcp_read_old_gcplib); # old GCPLib reading routine
use ias_gcp qw(ias_gcp_write_gcplib); # GCPLib writing routine
use manage_gcps_utils;

# Declare program constant
use constant PROG => basename($0);

my $help;
my $rename_chips;
my $gcp_source;
my $sensor;
my $set_abs;        # If defined, force the ABS/REL to ABS
my $update_reason;
my $select_method;
my $overall_version = 0;
my $date_updated;
my $chip_date_updated;

if (!GetOptions('help' => \$help, 'rename_chips' => \$rename_chips,
    'gcp_source=s' => \$gcp_source, 'sensor=s' => \$sensor,
    'set_abs' => \$set_abs, 'update_reason=s' => \$update_reason,
    'select_method=s' => \$select_method, 'version=i' => \$overall_version,
    'date_updated=s' => \$date_updated,
    'chip_date_updated=s' => \$chip_date_updated,))
{
    # Error reading the options, so issue the usage message and exit
    usage();
    exit 1;
}

# Provide a usage message if requested or the wrong number of parameters
if ($help || $#ARGV != 1)
{
    usage();
    exit 1;
}

# Get the parameters
my $source_chip_directory;
my $target_chip_directory;
my $old_filename = $ARGV[0];
my $new_filename = $ARGV[1];

# Get directory name if plan to replace or copy
if (defined $rename_chips)
{
    $source_chip_directory = dirname($old_filename);
    $target_chip_directory = dirname($new_filename);
}

# Set up the logger
my $logger = Logger->new("STDOUT");
if (!$logger)
{
    print STDERR PROG . ": ERROR:  Failed to establish logger object for "
        . "STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name(PROG);
set_default_log($logger);

# Determine what to use for the sensor name
if (!defined $sensor)
{
    $sensor = "ETM+";
    $logger->warn_msg("The image chip sensor is being defaulted to 'ETM+' "
        . "- consider providing a more appropriate sensor");
}
elsif ($sensor eq "use_source")
{
    # Some of the cal/val provided files have the sensor in the source column,
    # so this option can be used to indicate that.  When this is used, the
    # gcp_cource parameter is required.
    if (!defined $gcp_source)
    {
        $logger->fatal_msg("When using the '--sensor use_source' option, the "
            . "gcp_source option is required");
        $logger->close_log();
        exit FAILURE;
    }
}

# Check the update reason parameter
if (defined $update_reason)
{
    my $status;
    ($status, $update_reason) = check_parameter("update reason",
        $update_reason, 10,
        qw(ORIGINAL TRIG_ERROR NEW_POINT TEMPORAL SEASONAL));
    exit 1 if ($status != SUCCESS);
}
else
{
    # Parameter not provided, so default to ORIGINAL
    $update_reason = "ORIGINAL";
}

# Check the select method parameter
my $infer_moravec_rank = 0;
if (defined $select_method)
{
    my $status;
    ($status, $select_method) = check_parameter("select method",
        $select_method, 10, qw(MORAVEC GRID MANUAL UNKNOWN));
    exit 1 if ($status != SUCCESS);

    # If the select method is MORAVEC, flag that the moravec rank should be
    # inferred from the order of the entries in the file
    $infer_moravec_rank = 1;
}
else
{
    $select_method = "UNKNOWN";
}

# Get the date updated to use
if (!defined $date_updated)
{
    $date_updated = "11-27-2007";
}
else
{
    my $status;
    ($status, $date_updated) = get_date_updated($date_updated);
    if ($status != SUCCESS)
    {
        $logger->close_log();
        usage();
        exit FAILURE;
    }
}

# Get the date updated to use
if (!defined $chip_date_updated)
{
    # If not provided, use the same as the date updated
    $chip_date_updated = $date_updated;
}
else
{
    my $status;
    ($status, $chip_date_updated) = get_date_updated($chip_date_updated);
    if ($status != SUCCESS)
    {
        $logger->close_log();
        usage();
        exit FAILURE;
    }
}

# Read the old file
my ($status, @gcp_records) = ias_gcp_read_old_gcplib($old_filename);
if ($status == ERROR)
{
    $logger->fatal_msg("Error reading GCPLib file: $old_filename");
    $logger->close_log();
    exit 1;
}

# Fill in the missing fields
my $record_number = 0;
foreach my $gcp_record (@gcp_records)
{
    $record_number++;

    # Make sure the point_id has leading zeroes
    $gcp_record->{'point_id'} = sprintf("%010d", $gcp_record->{'point_id'});
    $gcp_record->{'point_type'} = $gcp_record->{'chip_type'};
    $gcp_record->{'chip_acq_date'} = $gcp_record->{'date'};

    $gcp_record->{'point_active'} = "Y";
    $gcp_record->{'point_date_updated'} = $date_updated;
    $gcp_record->{'update_reason'} = $update_reason;
    $gcp_record->{'point_version'} = 1;
    ($gcp_record->{'point_path'}, $gcp_record->{'point_row'}) 
        = get_wrs_path_row_from_gcp_id($gcp_record->{'point_id'});
    $gcp_record->{'chip_active'} = "Y";
    $gcp_record->{'chip_date_updated'} = $chip_date_updated;
    $gcp_record->{'chip_version'} = 1;
    $gcp_record->{'chip_path'} = $gcp_record->{'point_path'};
    $gcp_record->{'chip_row'} = $gcp_record->{'point_row'};
    # the select method and moravec rank will be filled in later
    $gcp_record->{'chip_select_method'} = $select_method;
    if ($infer_moravec_rank)
    {
        $gcp_record->{'moravec_rank'} = $record_number;
    }
    else
    {
        $gcp_record->{'moravec_rank'} = -1;
    }

    # Override the abs_rel_flag if requested
    $gcp_record->{'abs_rel_flag'} = "ABS" if (defined $set_abs);

    # If the sensor is "use_source", get the sensor from the GCP record 
    # source field and replace the source with the provided value
    if ($sensor eq "use_source")
    {
        $gcp_record->{'chip_sensor'} = $gcp_record->{'source'};
    }
    else
    {
        $gcp_record->{'chip_sensor'} = $sensor;
    }
    $gcp_record->{'source'} = $gcp_source if (defined $gcp_source);

    # For DOQ data, use the acquisition date for the date updated
    if ($gcp_record->{'source'} eq "DOQ")
    {
        $gcp_record->{'point_date_updated'} = $gcp_record->{'date'};
        $gcp_record->{'chip_date_updated'} = $gcp_record->{'date'};
    }

    # Set the image chip name to the new name, assuming it is the first 
    # one in the sequence
    my $chip_name = sprintf("%010d_%02d", $gcp_record->{'point_id'}, 1);

    # Rename or Copy existing chip files
    if (defined $rename_chips) 
    {
        if (-e "$source_chip_directory/$gcp_record->{'chip_name'}")
        {
            if (!copy("$source_chip_directory/$gcp_record->{'chip_name'}",
                 "$target_chip_directory/$chip_name")) 
            {
                $logger->fatal_msg("Unable to copy file "
                    . $source_chip_directory . $gcp_record->{'chip_name'} 
                    . " to $target_chip_directory/$chip_name");
            }
        }
        else 
        {
            $logger->warn_msg("$source_chip_directory/" 
                . $gcp_record->{'chip_name'} . " does not".
                " exist so can not copy chip\n");
        }

        $gcp_record->{'chip_name'} = $chip_name;
    }

    delete $gcp_record->{'chip_type'};
    delete $gcp_record->{'date'};

}

# Write the new format
$status = ias_gcp_write_gcplib($new_filename, $overall_version, \@gcp_records);
if ($status != 0)
{
    $logger->fatal_msg("Unable to write GCPLib file to $new_filename");
    $logger->close_log();
    exit FAILURE;
}

exit SUCCESS;


#############################################################################
sub usage
{
    print "\nUsage: convert_old_l8_gcplib_to_new.pl old_filename new_filename "
        . "[--gcp_source source] [--sensor sensor|use_source] "
        . "[--rename_chips] [--set_abs] [--update_reason reason] "
        . "[--select_method method] [--version #] [--date_updated mm-dd-yyyy "
        . "[--chip_date_updated mm-dd-yyyy]\n\n";
}

#############################################################################
# Temporary routine to parse WRS path and row from a GCP ID
sub get_wrs_path_row_from_gcp_id
{
    my $gcp_id = shift;
    my $wrs_path;
    my $wrs_row;
    my $status = SUCCESS;

    if ($gcp_id =~ /^(\d{3})(\d{3})(\d{4})$/)
    {
        $wrs_path = $1;
        $wrs_row = $2;

        # Check the path/row to make sure they are valid
        if ($wrs_path < 1 || $wrs_path > 233)
        {
            $logger->warn_msg("GCP ID $gcp_id contains an invalid "
                . "WRS path of $wrs_path");
            $status = ERROR;
        }
        if ($wrs_row < 1 || $wrs_row > 248)
        {
            $logger->warn_msg("GCP ID $gcp_id contains an invalid "
                . "WRS row of $wrs_row");
            $status = ERROR;
        }
    }
    else
    {
        # The point id failed the pattern match, so consider it an error
        $logger->warn_msg("Point id $gcp_id in not a legal point id");
        $status = ERROR;
    }

    # Since this is a hack for now, just exit with an error if an error happens
    # since there should never be one
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Invalid GCP ID encountered... exiting");
        exit 1;
    }

    return ($wrs_path, $wrs_row);
}

##############################################################################
# Name: check_parameter
#
# Description: Validates a parameter
##############################################################################
sub check_parameter
{
    my $parm_name = shift;
    my $parm = shift;
    my $len = shift;
    my @valid_values = @_;

    # Make sure it is all uppercase, no longer than 10 characters, and
    # does not contain spaces
    $parm = uc($parm);
    if (length($parm) < 1 || length($parm) > $len)
    {
        $logger->fatal_msg("$parm_name of '$parm' needs to have "
            . "from 1 to $len characters");
        return ERROR, $parm;
    }
    if ($parm =~ /\s+/)
    {
        $logger->fatal_msg("$parm_name '$parm' cannot contain "
            . "spaces");
        return ERROR, $parm;
    }
    # If it isn't one of the defined values, issue a warning
    my $found = 0;
    foreach my $val (@valid_values)
    {
        if ($parm eq $val)
        {
            $found = 1;
            last;
        }
    }
    if (!$found)
    {
        $logger->warn_msg("$parm_name '$parm' is not one of the "
            . "defined values");
    }

    return SUCCESS, $parm;
}
