#!/usr/bin/perl -w
#############################################################################
# Name: load_gcps_into_database
#
# Description: Loads the GCPLib file indicated into the database, including
#   the associated image chips. This script will delete all GCPs and Image Chips
#   associated with the WRS path/row combination and replace them.
#
# Assumptions:
#   - The script will be provided the GCPs for a single WRS path/row
#
# Parameters:
#   Database username/password
#   GCPLib filename
#   Oracle SID
#   Optional parameter: --unit-test
#       When set the Development databases will be used (for development)
#   Optional parameter: -- test
#       Used to test whether the load will work without committing the update.
#
#############################################################################
use strict;
use DBI;
use Getopt::Long;           # command line parsing support
use File::Basename;         # basename routine
use Logger;                 # Logging module
use ias_processing_system;  # define ias_error_msg and ias_validate_environment
use ias_const;              # define SUCCESS and ERROR
use ias_gcp qw(ias_gcp_read_gcplib); # provides the GCPLib reading routine
use ias_db;                 # provides routines for accessing the database
use DBD::Oracle qw(:ora_types); # provices the ORA_BLOB type
use manage_gcps_utils;      # read_image_chip_files routine

# Declare program constant
use constant PROG => basename($0);

# Declare variables for the database table names to allow for testing with
# different table names
my $gcp_table = "GROUND_CONTROL_POINTS";
my $chip_table = "IMAGE_CHIPS";
my $version_table = "GCP_DATA_VERSION";

# Read the optional parameters
my $test_only = 0;          # Flag to only test the load, do not commit it
my $unit_test = 0;          # Flag to use the DEV_ databases instead

if (!GetOptions('test' => \$test_only, 'unit-test' => \$unit_test))
{
    # Error reading the options, so issue the usage message and exit
    usage();
    exit FAILURE;
}

# Confirm there are enough parameters remaining for the required parameters
if ($#ARGV != 2)
{
    # Error since no GCPLib filename provided
    usage();
    exit FAILURE;
}

# Check if development tables should be used
if ($unit_test)
{
    $gcp_table = "DEV_GROUND_CONTROL_POINTS";
    $chip_table = "DEV_IMAGE_CHIPS";
    $version_table = "DEV_GCP_DATA_VERSION";
}

# Get the Database username/password
my $db_username_password = $ARGV[0];

# Get the name of the GCPLib file
my $gcplib_filename = $ARGV[1];

# Get the Oracle SID
my $oracle_sid = uc($ARGV[2]);

# Extract the directory name from the GCPLib filename
my ($temp_name, $gcp_directory, $suffix) = fileparse($gcplib_filename);

# Set up the logger
my $logger = Logger->new("STDOUT");
if (!$logger)
{
    print STDERR PROG . ": ERROR:  Failed to establish logger object for "
        . "STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name(PROG);
set_default_log($logger);

# Check that Oracle SID is valid
if (!($oracle_sid eq "CRDEV" || $oracle_sid eq "CRIT" ||
    $oracle_sid eq "CRST" || $oracle_sid eq "CROPS"))
{
    $logger->fatal_msg("Oracle SID not valid must be (CRDEV, CRIT, CRST, or "
        . "CROPS)");
    $logger->close_log();
    usage();
    exit FAILURE;
}  

# Separate the DB username/password
my ($db_user, $db_password) = split('/', $db_username_password);
if ((!defined $db_user) || (!defined $db_password))
{
    $logger->fatal_msg("Error parsing the database_user/password");
    $logger->close_log();
    usage();
    exit FAILURE;
}

# Read and validate the GCPLib file
my $gcplib_file = read_gcplib_file($gcplib_filename, MASTER_FILE);
if (!defined $gcplib_file)
{
    $logger->fatal_msg("Error reading master GCPLib file $gcplib_filename");
    $logger->close_log();
    exit FAILURE;
}

# Sort the array by GCP ID then by image chip ID to guarantee all 
# GCPs are in the correct order when inserting into the database
my @gcp_array = @{$gcplib_file->{'records'}};
@gcp_array = sort {
    if ($a->{'point_id'} == $b->{'point_id'})
    {
        return 0 if ($a->{'chip_name'} eq $b->{'chip_name'});
        return 1 if ($a->{'chip_name'} gt $b->{'chip_name'});
        return -1;
    }
    return 1 if ($a->{'point_id'} > $b->{'point_id'});
    return -1;
    } @gcp_array;

# Provide a message to indicate how many records were read from the file
$logger->info_msg("Read " . scalar(@gcp_array) . " GCPs from file: "
    . "$gcplib_filename");

# If no GCPs were read, exit with an error
if (scalar(@gcp_array) == 0)
{
    $logger->fatal_msg("No GCPs read from $gcplib_filename\n");
    $logger->close_log();
    exit FAILURE;
}

# Read the matching image chips into memory
$logger->info_msg("Reading the image chip files");

if (read_image_chip_files(\@gcp_array, $gcp_directory) != SUCCESS)
{
    $logger->fatal_msg("Error reading image chips");
    $logger->close_log();
    exit FAILURE;
}

# Connect to the database.  No need to start transaction for all the database
# updates since the routine turns off AutoCommit by default.
my $dbh = DBI->connect("DBI:Oracle:$oracle_sid", $db_user, $db_password,
                       {AutoCommit => 0, PrintError => 0});
if (!defined $dbh)
{
    $logger->fatal_msg("Error connecting to the database: $DBI::errstr");
    $logger->close_log();
    exit FAILURE;
}

# Delete the GCPs and image chips from the database for current path/row
$logger->info_msg("Deleting GCPs for path $gcp_array[0]->{'point_path'} ".
    "row $gcp_array[0]->{'point_row'} from the database");

if(delete_gcps_from_database($dbh, $gcp_array[0]->{'point_path'},
    $gcp_array[0]->{'point_row'}) != SUCCESS)
{
    $logger->fatal_msg("Failed deleting GCPs from the database");
    $logger->close_log();
    ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
    exit FAILURE;
}

# Insert the GCPs and image chips into the database
$logger->info_msg("Inserting the GCPs into the database");

if (insert_gcps_into_database($dbh, \@gcp_array) != SUCCESS)
{
    $logger->fatal_msg("Failed inserting GCPs into the database");
    $logger->close_log();
    ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
    exit FAILURE;
}

# Check if test only flag was set then set to ROLLBACK changes
my $commit_flag = IAS_DB_DISCONNECT_COMMIT;
if ($test_only)
{
    $logger->info_msg("Rolling back database changes due to the --test option");
    $commit_flag = IAS_DB_DISCONNECT_ROLLBACK;
}

# Disconnect database using the correct flag
if (ias_db_DisconnectFromDB($dbh, $commit_flag) != 1)
{
    $logger->fatal_msg("Error disconnecting from the database");
    $logger->close_log();
    exit FAILURE;
}

$logger->close_log();

exit SUCCESS;

##############################################################################
# Name: usage
# 
# Description: Prints out the usage message for the application
#
##############################################################################
sub usage
{
    print "\nUsage: ". PROG . " database_user/password GCPLib_filename "
        . " Database_instance [--test] [--unit_test]"
        . "\n\n";
}

##############################################################################
# Name: delete_gcps_from_database
#
# Description: Delete the GCPs and Image chips from the database that are
#              associated with a given GCP path and row combination. The
#              image chips will be deleted according to the path row that
#              is associated with the GCP ID foreign key.
#
# Parameters:
#   - Database handle
#   - GCP path
#   - GCP row 
#
# Returns:
#   SUCCESS or ERROR
##############################################################################
sub delete_gcps_from_database
{
    my $dbh = shift;    # Database handler
    my $path = shift;   # Current path that needs to be deleted
    my $row = shift;    # Current row that needs to be deleted

    # Prepare the Image Chip delete statement
    my $image_delete_sql = "Delete from $chip_table where ".
        "GROUND_CONTROL_POINTS_ID in (Select GROUND_CONTROL_POINTS_ID ". 
        "from $gcp_table where WRS_PATH = ? AND WRS_ROW = ? )";
    my $image_delete_sth = $dbh->prepare($image_delete_sql);
    if (!$image_delete_sth)
    {
        $logger->fatal_msg("Unable to prepare image chip delete sql".
            ": $DBI::errstr");
        return ERROR;
    }

    # Prepare the GCP delete statement
    my $gcp_delete_sql = "Delete from $gcp_table where WRS_PATH = ? ".
        "AND WRS_ROW = ?";
    my $gcp_delete_sth = $dbh->prepare($gcp_delete_sql);
    if (!$gcp_delete_sth)
    {
        $logger->fatal_msg("Unable to prepare GCP delete sql: $DBI::errstr");
        return ERROR;
    }

    # Execute the image chip deletion first because of dependency on 
    # GCPs in GROUND_CONTROL_POINTS Table
    if (!$image_delete_sth->execute($path, $row))
    {
        $logger->fatal_msg("Error deleting chip information for path $path".
            "row $row: $DBI::errstr");

        $gcp_delete_sth->finish();
        $image_delete_sth->finish();
        return ERROR;
    }

    # Execute the delete of the GCPs now that image chips associated
    # with GCPs for current path/row are deleted
    if (!$gcp_delete_sth->execute($path, $row))
    {
        $logger->fatal_msg("Error deleting GCP information for ".
            "path $path row $row: $DBI::errstr");
        $gcp_delete_sth->finish();
        $image_delete_sth->finish();
        return ERROR;
    }

    # Finish database operations
    $gcp_delete_sth->finish();
    $image_delete_sth->finish();

    return SUCCESS;
}

##############################################################################
# Name: insert_gcps_into_database
#
# Description: Insert the GCPs and associated image chips into the database.
#
# Parameters:
#   - Database handle
#   - Array of GCP references
#
# Returns:
#   SUCCESS or ERROR
##############################################################################
sub insert_gcps_into_database
{
    my $dbh = shift;
    my $gcp_array_ref = shift;

    # Prepare the GCP table insert statement
    my $gcp_sql = "INSERT INTO $gcp_table ". 
        "(ground_control_points_id, active, date_updated, elevation, ".
        "latitude, longitude, source, type, update_reason, version, wrs_path, ".
        "wrs_row) VALUES(?,?,TO_DATE(?,'mm-dd-yyyy'),?,?,?,?,?,?,?,?,?)";

    my $gcp_sth = $dbh->prepare($gcp_sql);
    if (!$gcp_sth)
    {
        $logger->fatal_msg("Unable to prepare GCP table insert: $DBI::errstr");
        return ERROR;
    }

    # Prepare the update query for the spatial index
    my $shape_sql = "UPDATE $gcp_table SET shape = MDSYS.SDO_GEOMETRY("
        . "2001, 8307, MDSYS.SDO_POINT_TYPE(?,?,?),null,null) "
        . "WHERE ground_control_points_id = ?";

    my $shape_sth = $dbh->prepare($shape_sql);
    if (!$shape_sth)
    {
        $logger->fatal_msg("Unable to prepare shape update: $DBI::errstr");
        $gcp_sth->finish();
        return ERROR;
    }

    # Prepare the image chips table insert statement
    my $chips_sql = "INSERT INTO $chip_table ". 
        "(image_chips_id,ground_control_points_id, ".
        "version, active, date_acquired, date_updated, reference_line, ".
        "reference_sample, projection, projection_y, projection_x, zone_code, ".
        "pixel_size_x, pixel_size_y, lines, samples, sensor, abs_rel_flag, ".
        "chip_data, chip_data_type, select_method, moravec_rank, ".
        "source_wrs_path, source_wrs_row) ".
        "VALUES(?,?,?,?,TO_DATE(?,'mm-dd-yyyy'),".
        "TO_DATE(?,'mm-dd-yyyy'),?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)";

    my $chips_sth = $dbh->prepare($chips_sql);
    if (!$chips_sth)
    {
        $logger->fatal_msg("Unable to prepare image chips insert:"
            . " $DBI::errstr");
        $gcp_sth->finish();
        $shape_sth->finish();
        return ERROR;
    }

    my $previous_gcp_id = '-1';
    # Loop through all of the GCP records
    foreach my $gcp_ref (@$gcp_array_ref)
    {
        # Only need to add GCP once for all image chips
        if ($previous_gcp_id != $gcp_ref->{'point_id'})
        {
            $previous_gcp_id = $gcp_ref->{'point_id'};
            # Insert the GCP table info
            if (!$gcp_sth->execute($gcp_ref->{'point_id'}, 
                $gcp_ref->{'point_active'}, $gcp_ref->{'point_date_updated'},
                $gcp_ref->{'elevation'}, $gcp_ref->{'latitude'}, 
                $gcp_ref->{'longitude'}, $gcp_ref->{'source'},
                $gcp_ref->{'point_type'}, $gcp_ref->{'update_reason'},
                $gcp_ref->{'point_version'}, $gcp_ref->{'point_path'}, 
                $gcp_ref->{'point_row'}))
            {
                $logger->fatal_msg("Error inserting information for GCP ID "
                    . $gcp_ref->{'point_id'} . ": $DBI::errstr");

                # Fail immediately with this type of error in case it is a 
                # database Issue.  It could continue instead to check all the 
                # records for possible constraint issues, but that might 
                # produce a lot of output messages that are just due to 
                # a database issue.
                $gcp_sth->finish();
                $shape_sth->finish();
                $chips_sth->finish();
                return ERROR;
            }

            # Add the shape information
            if (!$shape_sth->execute($gcp_ref->{'longitude'}, 
                    $gcp_ref->{'latitude'}, $gcp_ref->{'elevation'},
                    $gcp_ref->{'point_id'}))
            {
                $logger->fatal_msg("Error updating shape information for GCP ID"
                    ." ". $gcp_ref->{'point_id'} . ": $DBI::errstr");

                # Fail immediately with this type of error in case it is a 
                # database Issue.  It could continue instead to check all the 
                # records for possible constraint issues, but that might 
                # produce a lot of output messages that are just due to 
                # a database issue.
                $gcp_sth->finish();
                $shape_sth->finish();
                $chips_sth->finish();
                return ERROR;
            }
        }

        # Manually bind all the parameters for the image chip insert since 
        # that appears to be the only way to get the BLOB insert to work
        my $blob_buffer = $gcp_ref->{'image_chip'};
        if (!$chips_sth->bind_param(1, $gcp_ref->{'chip_name'})
            || !$chips_sth->bind_param(2, $gcp_ref->{'point_id'})
            || !$chips_sth->bind_param(3, $gcp_ref->{'chip_version'})
            || !$chips_sth->bind_param(4, $gcp_ref->{'chip_active'})
            || !$chips_sth->bind_param(5, $gcp_ref->{'chip_acq_date'})
            || !$chips_sth->bind_param(6, $gcp_ref->{'chip_date_updated'})
            || !$chips_sth->bind_param(7, $gcp_ref->{'reference_line'})
            || !$chips_sth->bind_param(8, $gcp_ref->{'reference_sample'})
            || !$chips_sth->bind_param(9, $gcp_ref->{'projection'})
            || !$chips_sth->bind_param(10, $gcp_ref->{'projection_y'})
            || !$chips_sth->bind_param(11, $gcp_ref->{'projection_x'})
            || !$chips_sth->bind_param(12, $gcp_ref->{'zone'})
            || !$chips_sth->bind_param(13, $gcp_ref->{'pixel_size_x'})
            || !$chips_sth->bind_param(14, $gcp_ref->{'pixel_size_y'})
            || !$chips_sth->bind_param(15, $gcp_ref->{'chip_size_lines'})
            || !$chips_sth->bind_param(16, $gcp_ref->{'chip_size_samples'})
            || !$chips_sth->bind_param(17, $gcp_ref->{'chip_sensor'})
            || !$chips_sth->bind_param(18, $gcp_ref->{'abs_rel_flag'})
            || !$chips_sth->bind_param(19, $blob_buffer,
                { ora_type => ORA_BLOB })
            || !$chips_sth->bind_param(20, $gcp_ref->{'chip_data_type'})
            || !$chips_sth->bind_param(21, $gcp_ref->{'chip_select_method'})
            || !$chips_sth->bind_param(22, $gcp_ref->{'moravec_rank'})
            || !$chips_sth->bind_param(23, $gcp_ref->{'chip_path'})
            || !$chips_sth->bind_param(24, $gcp_ref->{'chip_row'}))

         {
            $logger->fatal_msg("Error binding parameters for image chip "
                . "insert for GCP ID " . $gcp_ref->{'point_id'}
                . ": $DBI::errstr");

            # Fail immediately with this type of error in case it is a database
            # Issue.  It could continue instead to check all the records for
            # possible constraint issues, but that might produce a lot of 
            # output messages that are just due to a database issue.
            $gcp_sth->finish();
            $shape_sth->finish();
            $chips_sth->finish();
            return ERROR;
        }

        # Insert the image chip information
        if (!$chips_sth->execute())
        {
            $logger->fatal_msg("Error inserting image chip information for "
                . "Chip " . $gcp_ref->{'chip_name'} . ": $DBI::errstr");

            # Fail immediately with this type of error in case it is a database
            # Issue.  It could continue instead to check all the records for
            # possible constraint issues, but that might produce a lot of 
            # output messages that are just due to a database issue.
            $gcp_sth->finish();
            $shape_sth->finish();
            $chips_sth->finish();
            return ERROR;
        }
    }

    $gcp_sth->finish();
    $shape_sth->finish();
    $chips_sth->finish();

    return SUCCESS;
}
