#!/usr/bin/perl -w
#############################################################################
# Name: load_image_chip_method_and_rank
#
# Description:  Update a GCPLib with SELECT_METHOD and MORAVEC_RANK
#               from the provided file.
#
# Assumptions: 
#   - Input GCPLib is an up to date GCPLib
#
# Parameters:
#   Input GCPLib filename
#   CAL/VAL correction filename
#   Output GCPLib filename
#
###########################################################################
use strict;
use Getopt::Long;           # command line parsing support
use File::Basename;         # basename routine
use Logger;                 # Logging module
use ias_processing_system;  # define ias_error_msg and ias_validate_environment
use ias_const;              # define SUCCESS and ERROR
use ias_gcp qw(ias_gcp_read_gcplib ias_gcp_write_gcplib); # provides the GCPLib 
                                                          # reading/writing
# Declare program constant
use constant PROG => basename($0);

my $gcp_source;

if (!GetOptions('gcp_source=s' => \$gcp_source))
{
    # Error reading the options, so issue the usage message and exit
    usage();
    exit FAILURE;
}

# verify the gcp_source is valid if provided
if (defined $gcp_source && ($gcp_source ne "GLS" && $gcp_source ne "DOQ"))
{
    # Error reading the options, so issue the usage message and exit
    usage();
    exit FAILURE;
}

# Confirm there are enough parameters
if ($#ARGV != 2)
{
    # Error since no input filename provided
    usage();
    exit ERROR;;
}

# Get the name of the GCPLib file
my $input_gcplib = $ARGV[0];

# Get the name of the GCPLib file
my $correction_filename = $ARGV[1];

# Get the name of the GCPLib file
my $output_gcplib = $ARGV[2];

# Set up the logger
my $logger = Logger->new("STDOUT");
if (!$logger)
{
    print STDERR PROG . ": ERROR:  Failed to establish logger object for "
        . "STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name(PROG);
set_default_log($logger);

# Update the GCPs
if (update_gcplib($input_gcplib, $correction_filename,
        $output_gcplib, $gcp_source) != SUCCESS)
{
    $logger->fatal_msg("Failed to update GCPLib file");
    $logger->close_log();
    exit ERROR;;
}

# Close log and exit
$logger->close_log();
exit SUCCESS;

##############################################################################
# Name: usage
# 
# Description: Prints out the usage message for the application
#
##############################################################################
sub usage
{
    print "\nUsage: ". PROG ." Input_GCPLib_filename".
        " Correction_GCPLib_filename Ouput_GCPLib_filename"
        . " [--gcp_source [GLS|DOQ]]\n\n";
}

##############################################################################
# Name: binary_search_by_gcp_id
# 
# Description: Binary search of array passed which matchs the passed in value
#              with the gcp_id hash value.
#
# Parameters:
#   - Array of hashes
#   - Int GCP ID
##############################################################################
sub binary_search_by_gcp_id
{
    my ($aref, $find) = @_; # Set array reference and value to find
    my ($low, $high) = (0, $#$aref); # set low and high values

    while ($low <= $high)
    {
        my $mid = int (($low + $high) / 2); # Set middle of array
        
        # Check if value is higher or lower else its a match
        $low  = $mid + 1, next if (int($find) > int($aref->[$mid]->{'gcp_id'}));
        $high = $mid - 1, next if (int($find) < int($aref->[$mid]->{'gcp_id'}));
        
        # Matches if reaches here
        return $mid;
    }

    return ERROR;
}

##############################################################################
# Name: update_gcplib
#
# Description: Updates the GCPLib file with correct moravec rank and chip
#              method.
#
# Parameters:
#   - Input GCPLib filename
#   - Cal/Val correction filename
#   - Output GCPLib filename
#   - GCP source (GLS or DOQ)
#
# Returns:
#   SUCCESS or ERROR
##############################################################################
sub update_gcplib
{
    # Parameters passed in
    my $input_gcplib = shift;
    my $correction_filename = shift;
    my $output_gcplib = shift;
    my $gcp_source = shift;

    # Fields definded in the correction file
    my @fields = qw(gcp_id latitude longitude projection_y projection_x
        elevation acquisition_date path row moravec even srank irank);
    my @correction_data; 

    # Get the default logger object from the library 
    my $log = get_default_log();
    
    # Read GCPLib input
    my ($status, $overall_gcp_version, @gcp_data) = 
        ias_gcp_read_gcplib($input_gcplib);
    if ($status != SUCCESS)
    {
        $log->fatal_msg("Failed to read GCPLib file");
        $log->close_log();
        return ERROR;
    }
 
    # Open Correction to read line by line
    $status = open COR_FILE, "$correction_filename";
    if (!defined($status))
    {
        $log->fatal_msg("Opening $correction_filename");
        return ERROR;
    }

    # Read in the correction file
    my $line_number = 0;
    while (my $line = <COR_FILE>)
    {
        my $gcp_record = {}; 
        chomp $line;
        $line_number++;

        # Skip comments
        next if $line =~ /^#/;

        my @values = split(/ +/, $line);

        # Verify the correct number of values were found
        if (scalar @fields != scalar @values)
        {
            $log->fatal_msg("Invalid number of fields on line number "
                . "$line_number : $line");
            return ERROR, undef, undef;
        }

        # Assign the values to the corresponding key in the record
        for (my $index = 0; $index <= $#fields; $index++)
        {
            my $field = $fields[$index];
            $gcp_record->{$field} = $values[$index];
        }

        # Add gcp_record hash to gcp_data array
        push @correction_data, $gcp_record;
    }

    # Sort the array by gcp_id so it can be used in binary_search
    @correction_data = sort {int $a->{'gcp_id'} <=> int $b->{'gcp_id'}}
        @correction_data;

    # Iterate through GCPLib file updating moravec rank and 
    # chip select method
    my ($updated_records, $total_records) = (0, $#gcp_data + 1);
    $line_number = 0;
    foreach (@gcp_data)
    {
        # Use the binary search to find index of correct data
        my $match = binary_search_by_gcp_id(\@correction_data, 
           $_->{'point_id'});

        # Binary search returns ERROR if not found
        if ($match == ERROR)
        {
            # if only updating one source, it is OK if there wasn't a match for
            # some other source
            if (defined $gcp_source && ($gcp_source ne $_->{'source'}))
            {
                # not the correct source, so move to the next point
                next;
            }

            $log->fatal_msg("Could not find GCP $_->{'point_id'} in correction".
                "GCPLIB");
            return ERROR;
        }

        # If the rank in correction file is null then set to GRID and -1
        # Else it was selected via moravec        
        if ($correction_data[$match]->{'irank'} eq 'NULL')
        {
            $_->{'chip_select_method'} = 'GRID';
            $_->{'moravec_rank'} = 0;
        }
        else
        {
            $_->{'chip_select_method'} = 'MORAVEC';
            $_->{'moravec_rank'} = $correction_data[$match]->{'irank'};
        }

        # Keep track of how many records were updated 
        $updated_records++;   
    }
    
    # Print out the number of updated records.
    $log->info_msg("Updated $updated_records of $total_records GCPs");
    
    # Write out the GCPLib file
    $status = ias_gcp_write_gcplib($output_gcplib, $overall_gcp_version,
        \@gcp_data);
    if ($status != SUCCESS)
    {
        $log->fatal_msg("Writing out the GCPLib file");
        return ERROR;
    }

    return SUCCESS;
}

