#!/usr/bin/perl -w
#############################################################################
# Name: load_overall_gcp_version
#
# Description: Loads the overall gcp version file and stores it in the 
#              GCP_DATA_VERSION table. It will delete any existing records in
#              the table and load the complete file. If an error occurs any
#              database changes will be rolled back.
#
# Assumptions:
#   
# Parameters:
#   Database username/password
#   Overall GCP Version filename
#   Database instance
#   Optional parameter: -- test
#       Used to test whether the load will work without committing the update.
#
#############################################################################
use strict;
use DBI;                    # Provides database access
use Getopt::Long;           # Command line parsing support
use File::Basename;         # Basename routine
use Logger;                 # Logging module
use ias_processing_system;  # Define ias_error_msg and ias_validate_environment
use ias_const;              # Define SUCCESS and ERROR
use ias_db;                 # Provides routines for accessing the database

# Routine declarations
sub load_version_data($$);

# Declare program constant
use constant PROG => basename($0);

# Declare variable for the database table
my $version_table = "GCP_DATA_VERSION";

# Read the optional parameters
my $test_only = 0;          # Flag to only test the load, do not commit it
my $unit_test = 0;          # Flag to allow for unit testing 

if (!GetOptions('test' => \$test_only, 'unit-test' => \$unit_test))
{
    # Error reading the options, so issue the usage message and exit
    usage();
    exit 1;
}

# Check if unit test flag was set
if ($unit_test)
{
    $version_table = "DEV_GCP_DATA_VERSION";
}

# Confirm there are enough parameters remaining for the required parameters
if ($#ARGV != 2)
{
    usage();
    exit 1;
}

# Get the Database username/password
my $db_username_password = $ARGV[0];

# Get the name of the GCPLib file
my $version_filename = $ARGV[1];

# Get the Oracle SID
my $oracle_sid = uc($ARGV[2]);

# Set up the logger
my $logger = Logger->new("STDOUT");
if (!$logger)
{
    print STDERR PROG . ": ERROR:  Failed to establish logger object for "
        . "STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name(PROG);
set_default_log($logger);

# Check that Oracle SID is valid
if (!($oracle_sid eq "CRDEV" || $oracle_sid eq "CRIT" ||
    $oracle_sid eq "CRST" || $oracle_sid eq "CROPS"))
{
    $logger->fatal_msg("Oracle SID not valid must be (CRDEV, CRIT, CRST, or "
        . "CROPS)");
    $logger->close_log();
    usage();
    exit FAILURE;
}  

# Separate the DB username/password
my ($db_user, $db_password) = split('/', $db_username_password);
if ((!defined $db_user) || (!defined $db_password))
{
    $logger->fatal_msg("Error parsing the database_user/password");
    $logger->close_log();
    usage();
    exit FAILURE;
}

# Connect to the database.  No need to start transaction for all the database
# updates since the routine turns off AutoCommit by default.
my $dbh = DBI->connect("DBI:Oracle:$oracle_sid", $db_user, $db_password,
                       {AutoCommit => 0, PrintError => 0});
if (!defined $dbh)
{
    $logger->fatal_msg("Error connecting to the database: $ias_db_error");
    $logger->close_log();
    exit FAILURE;
}

# Bulk of database work done in load_version_data routine
if (load_version_data($dbh, $version_filename) != SUCCESS)
{
    $logger->fatal_msg("Error loading overall version data");
    $logger->close_log();
    ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
    exit FAILURE;
}

# Check if this a test only run
my $commit_flag = IAS_DB_DISCONNECT_COMMIT;
if ($test_only)
{
    $logger->info_msg("Rolling back database changes due to the --test option");
    $commit_flag = IAS_DB_DISCONNECT_ROLLBACK;
}

# Disconnect from database and close log
if (ias_db_DisconnectFromDB($dbh, $commit_flag) != 1)
{
    $logger->fatal_msg("Error disconnecting from the database");
    $logger->close_log();
    exit FAILURE;
}
$logger->close_log();

exit SUCCESS;

##############################################################################
# Name: usage
# 
# Description: Prints out the usage message for the application
#
##############################################################################
sub usage
{
    print "\nUsage: ". PROG . " database_user/password overall_version_file "
        . "Database_instance [--test]\n\n";
}

##############################################################################
# Name: load_version_data
# 
# Description: Loads the overall GCP version data into the GCP_DATA_VERSION
#              table.
#
# Parameters:
#   - Database handler
#   - Version filename
#
# Returns:
#   - SUCCESS/ERROR
#
##############################################################################
sub load_version_data($$)
{
    my $dbh = shift;                        # Database handler
    my $version_filename = shift;           # Version filename

    my @fields = qw(version date_updated description); # Fields in version file
    my @version_data;                       # Array for version data

    # Get the default logger object from the library 
    my $logger = get_default_log();

    # Open the overall version input file
    my $status = open VERSION_FILE, "$version_filename";
    if (!defined($status))
    {
        $logger->fatal_msg("Opening $version_filename");
        return ERROR;
    }
    
    # Read in the version file
    while (my $line = <VERSION_FILE>)
    {
        next if $line =~ /^#/;
        next if $line =~ /^BEGIN/;
        $line =~ s/^\s+|\s+$//g;    
        next if $line eq ''; 
        my @values = split(/\s*\|\s*/, $line);
        my $record;
        
        # Check that number of values matches number of fields
        if ($#values != $#fields)
        {
            $logger->fatal_msg("Incorrect number of values in $line");
            return ERROR;
        }

        # Assign the values to the corresponding key in the record
        for (my $index = 0; $index <= $#fields; $index++)
        {
            my $field = $fields[$index];
            $record->{$field} = $values[$index];
        }

        # Add record hash to data array
        push @version_data, $record;
    }
    close VERSION_FILE;

    # Sort the version file so newest version is first
    @version_data = sort{$b->{'version'} <=> $a->{'version'}} @version_data;
    
    # Prepare the delete statement
    my $delete_sql = "DELETE FROM $version_table";
    my $version_delete_sth = $dbh->prepare($delete_sql);
    if (!$version_delete_sth)
    {
        $logger->fatal_msg("Unable to prepare version table deletion:".
            " $DBI::errstr");
        return ERROR;
    }

    # Prepare the insert statement
    my $insert_sql = "INSERT INTO $version_table ".
        "(VERSION, DATE_UPDATED, DESCRIPTION) ".
        "VALUES (?, TO_DATE(?, 'MM-DD-YYYY'), ?)";
    my $version_insert_sth = $dbh->prepare($insert_sql);
    if (!$version_insert_sth)
    {
        $logger->fatal_msg("Unable to prepare version table insert:".
            " $DBI::errstr");
        return ERROR;
    }

    # Delete all the records from the table
    if (!$version_delete_sth->execute())
    {
        $logger->fatal_msg("Error deleting information from table "
            .": $DBI::errstr");
        $version_insert_sth->finish();
        return ERROR;
    }

    # Insert each record into the database 
    for my $record (@version_data)
    {
        $logger->info_msg("Loading version: " . $record->{'version'}
            . ", Date Updated : " . $record->{'date_updated'}
            . ", Description : " . $record->{'description'});

        # Insert the version and description
        if (!$version_insert_sth->execute($record->{'version'}, 
            $record->{'date_updated'}, $record->{'description'}))
        {
            $logger->fatal_msg("Error inserting information for version: "
                . $record->{'version'} . ": $DBI::errstr");
            $version_insert_sth->finish();
            return ERROR;
        }
    }

    return SUCCESS;
}

