package manage_gcps_utils;

use strict;
use warnings;
use File::Basename;                       # Fileparse routine
use Logger;                               # Provides logging for module
use ias_const;                            # Defines SUCCESS and ERROR
use ias_date_utils qw(ias_is_valid_date); # Date validator
use Time::localtime;                      # To get today's date
use Archive::Tar;                         # Provides Tarball functionality
use Digest::MD5 qw(md5_hex);              # MD5 to verify image chips
use Time::Local;                          # Getting current date/time
# GCPLib read routine and size of data type routine
use ias_gcp qw(ias_gcp_read_gcplib ias_gcp_get_sizeof_data_type); 
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);
use Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(
                UPDATE_FILE
                MASTER_FILE
                create_chips_updated_file
                create_record_index
                gcps_are_unique
                get_date_updated
                get_gcp_metadata_members
                get_image_chip_metadata_members
                is_compatible_gcp
                is_same_gcp
                is_unique_image_chip
                read_gcplib_file
                read_image_chip_files
                set_verbose_state
                verify_gcps_are_consistent
                verify_image_chip_data
                verify_tar_data
            );

use constant UPDATE_FILE => 0;
use constant MASTER_FILE => 1;

my $verbose = 1;  # Assume verbose mode is wanted

##############################################################################
# Name: set_verbose_state
#
# Description: Allows controlling whether the module produces verbose
#   messages
#
# Parameters: 
#   - State to set: 0 = not verbose, 1 = verbose
##############################################################################
sub set_verbose_state
{
    $verbose = shift;
}

##############################################################################
# Name: get_date_updated
#
# Description: Verify the date_updated passed in is valid.  If the value
#   passed in is undefined, return today's date.  If the value "use_file_date"
#   is found, it returns undef to indicate the date from the update file 
#   should be used.
#
# Parameters:
#   Date updated parameter to the script (may be undefined)
#
# Returns:
#   SUCCESS/ERROR
#   Date updated (undef if the update file value should be used)
##############################################################################
sub get_date_updated
{
    my $date_updated = shift;

    # If a date_updated is provided, confirm it is in the correct format
    if (defined $date_updated)
    {
        if (uc($date_updated) eq "USE_FILE_DATE")
        {
            return SUCCESS, undef;
        }

        # Just in case the date was provided with '/' as dividers instead of
        # '-', replace them
        my $test_date = $date_updated;
        $test_date =~ s/\//-/g;

        if (!ias_is_valid_date($test_date))
        {
            # Invalid date, so issue a message and return an error
            my $logger = get_default_log();
            $logger->fatal_msg("Invalid date_updated value $date_updated");
            return ERROR, undef;
        }
        $date_updated = $test_date;
    }
    else
    {
        # No date provided, so use today's date
        my $time = localtime;
        $date_updated = sprintf("%02d-%02d-%4d", $time->mon + 1, $time->mday,
            $time->year+1900);
    }

    return SUCCESS, $date_updated;
}

##############################################################################
# Name: read_gcplib_file
#
# Description: Reads a GCPLib file and performs consistency checks on the 
#   data.
#
# Parameters:
#   - Filename to read
#   - Type of file being read to know what kind of checks to apply since the
#     update files are not checked as stringently
#
# Returns:
#   - A reference to a hash containing file_type, file directory, overall GCP
#     version read from the file, the records read, and a hash by GCP ID to
#     those records.
##############################################################################
sub read_gcplib_file
{
    my $filename = shift;
    my $file_type = shift;

    my $logger = get_default_log();

    my $is_master = ($file_type == MASTER_FILE);

    my ($temp_name, $dir, $suffix) = fileparse($filename);

    $logger->info_msg("Reading $filename") if ($verbose);

    # Read the GCPLib file
    my ($status, $gcp_version, @gcps) = ias_gcp_read_gcplib($filename);
    if ($status == ERROR)
    {
        $logger->fatal_msg("Error reading GCPLib file: $filename");
        return undef;
    }

    # If no GCPs were read, issue a warning
    if (scalar(@gcps) == 0)
    {
        $logger->warn_msg("No GCPs read from $filename");
    }

    # Provide a message to indicate how many records were read from the file
    $logger->info_msg("Read " . scalar(@gcps) . " GCPs from file: "
        . "$filename") if ($verbose);

    # For each of the records, check whether the projection is UTM with a
    # negative zone.  There shouldn't be any of these, but if there are, the Y
    # coordinate needs to be checked and potentially fixed.
    foreach my $record (@gcps)
    {
        if (($record->{'projection'} eq "UTM") && ($record->{'zone'} < 0))
        {
            # Change the zone to positive
            $record->{'zone'} = -$record->{'zone'};

            # If the Y coordinate is greater than zero, which means it is
            # following the negative zone convention, subtract 10 million from
            # the coordinate to eliminate the false northing the negative UTM
            # zone convention uses.
            if ($record->{'projection_y'} > 0.0)
            {
                $record->{'projection_y'} -= 10000000;
                $logger->warn_msg("Converted image chip ID "
                    . $record->{'chip_name'} . " from the negative UTM zone "
                    . "convention to the positive zone convention and adjusted"
                    . " the Y coordinate");
            }
            else
            {
                $logger->warn_msg("Converted image chip ID "
                    . $record->{'chip_name'} . " from the negative UTM zone "
                    . "convention to the positive zone convention");
            }
        }
    }

    # Create an index hash for the GCP IDs
    my $record_hash = create_record_index(\@gcps);

    # Verify the GCPs are consistent
    if (!verify_gcps_are_consistent($record_hash, $file_type))
    {
        $logger->fatal_msg("GCPs in $filename are not consistent");
        return undef;
    }

    $logger->info_msg("Verifying GCPs in $filename are unique") if ($verbose);

    # Verify the set doesn't have any different GCP IDs that are really
    # the same point
    if (!gcps_are_unique($record_hash))
    {
        if ($is_master)
        {
            $logger->warn_msg("GCPs in $filename are not unique");
        }
        else
        {
            $logger->fatal_msg("GCPs in $filename are not consistent");
            return undef;
        }
    }

    # Build the record to return
    my %file_hash;
    $file_hash{'file_type'} = $file_type;
    $file_hash{'dir'} = $dir;
    $file_hash{'version'} = $gcp_version;
    $file_hash{'records'} = \@gcps;
    $file_hash{'lookup'} = $record_hash;

    return \%file_hash;
}

##############################################################################
# Name: create_record_index
#
# Description: Creates a hash indexed by the GCP ID for the GCP records in an
#   array.
#
# Parameters:
#   - Reference to an array of GCP records
#
# Returns:
#   A reference to a hash indexed by GCP ID.  Each hash entry contains an array
#   of references to the GCP records for that GCP ID.
##############################################################################
sub create_record_index
{
    my $records = shift;

    my %index_hash;

    foreach my $rec (@$records)
    {
        my $gcp_id = $rec->{'point_id'};

        push(@{$index_hash{$gcp_id}}, $rec);
    }

    return \%index_hash;
}

##############################################################################
# Name: verify_gcps_are_consistent
#
# Description: Verify the set of GCPs provided are consistent and valid.
#
# Parameters:
#   - GCP ID lookup hash reference
#   - File type (master/update)
#
# Returns:
#   TRUE if the GCPs are consistent
#   FALSE if they are not consistent
##############################################################################
sub verify_gcps_are_consistent
{
    my $gcp_hash = shift;
    my $file_type = shift;

    my $logger = get_default_log();

    my $wrs_path;
    my $wrs_row;
    my %chip_id_hash;        # hash used to confirm there are no duplicate IDs

    foreach my $gcp_id (keys %$gcp_hash)
    {
        # Get the count of records for this GCP ID
        my $array_ref = $gcp_hash->{$gcp_id};
        my $count = scalar @$array_ref;

        # Cache the reference to the first record for the GCP ID
        my $gcp1 = $array_ref->[0];

        # Initialize the WRS path and row that will be used to make sure all
        # the GCPs are in the same path/row
        if (!defined $wrs_path)
        {
            $wrs_path = $gcp1->{'point_path'};
            $wrs_row = $gcp1->{'point_row'};

            if ($wrs_path < 1 || $wrs_path > 233 
                || $wrs_row < 1 || $wrs_row > 248)
            {
                $logger->fatal_msg("Invalid path/row of $wrs_path/$wrs_row");
                return FALSE;
            }
        }

        # Verify the WRS path/row are the same for all points
        if ($wrs_path != $gcp1->{'point_path'}
            || $wrs_row != $gcp1->{'point_row'})
        {
            $logger->fatal_msg("GCP ID $gcp_id has WRS path/row of "
                . $gcp1->{'point_path'} . "/" . $gcp1->{'point_row'}
                . " instead of expected $wrs_path/$wrs_row");
            return FALSE;
        }

        # Check additional constraints if requested
        if ($file_type == MASTER_FILE)
        {
            # Make sure the path/row fields and the GCP ID are consistent
            if ($gcp_id =~ /^(\d{3})(\d{3})\d{4}$/)
            {
                if (($1 != $gcp1->{'point_path'})
                    || ($2 != $gcp1->{'point_row'}))
                {
                    $logger->fatal_msg("GCP ID $gcp_id path/row do not "
                        . "match metadata values of " . $gcp1->{'point_path'}
                        . "/" .  $gcp1->{'point_row'});
                    return FALSE;
                }
            }
            else
            {
                $logger->fatal_msg("GCP ID $gcp_id is not in the expected "
                    . "format");
                return FALSE;
            }

            # Make sure the image chip IDs are consistent
            foreach my $rec (@$array_ref)
            {
                my $chip_id = $rec->{'chip_name'};

                # Make sure the chip ID is the right format and agrees with
                # the GCP ID
                if ($chip_id =~ /^(\d{10})_\d{2}$/)
                {
                    if ($1 ne $gcp_id)
                    {
                        $logger->fatal_msg("Image chip ID $chip_id does not "
                            . "match the GCP ID $gcp_id");
                        return FALSE;
                    }
                }
                else
                {
                    $logger->fatal_msg("Image chip ID $chip_id is not in the "
                        . "expected format (<gcp id>_##)");
                    return FALSE;
                }

                # Make sure the image chip IDs are unique
                if (defined $chip_id_hash{$chip_id})
                {
                    $logger->fatal_msg("Image chip ID $chip_id appears multiple"
                        . " times in the file");
                    return FALSE;
                }
                $chip_id_hash{$chip_id} = 1;
            }
        }

        # Verify all the chip path/rows are valid
        foreach my $rec (@$array_ref)
        {
            if ($rec->{'chip_path'} < 1 || $rec->{'chip_path'} > 233
                || $rec->{'chip_row'} < 1 || $rec->{'chip_row'} > 248)
            {
                $logger->fatal_msg("Invalid chip path/row of "
                    . $rec->{'chip_path'} . "/" . $rec->{'chip_row'} 
                    . " for chip ID " . $rec->{chip_name});
                return FALSE;
            }
        }
 
        # If there are multiple records, make sure the GCP metadata is
        # consistent in all the records
        if ($count > 1)
        {
            for (my $index = 1; $index < $count; $index++)
            {
                my $gcp2 = $array_ref->[$index];

                # Verify the GCP data is the same for each image chip
                if (!is_same_gcp($gcp1, $gcp2))
                {
                    $logger->fatal_msg("GCP ID $gcp_id has records with "
                        . "inconsistent metadata");
                    return FALSE;
                }
            }

            # Also make sure that every image chip has unique metadata
            for (my $i1 = 0; $i1 < $count - 1; $i1++)
            {
                my $chip1 = $array_ref->[$i1];

                for (my $i2 = $i1 + 1; $i2 < $count; $i2++)
                {
                    my $chip2 = $array_ref->[$i2];
                    if (!is_unique_image_chip($chip1, $chip2))
                    {
                        $logger->fatal_msg("GCP ID $gcp_id image chip metadata"
                            . " for " . $chip1->{'chip_name'} . " and "
                            . $chip2->{'chip_name'} . " are not unique");
                        return FALSE;
                    }
                }
            }
        }
    }

    return TRUE;
}

##############################################################################
# Name: gcps_are_unique
#
# Description: Given a GCP lookup hash, make sure there aren't any different 
#   GCPs that are really the same point
#
# Parameters:
#   - GCP ID lookup hash reference
#
# Returns:
#   TRUE if the points are all unique, FALSE if non-unique points are found
##############################################################################
sub gcps_are_unique
{
    my $gcp_hash = shift;

    my $logger = get_default_log();

    my $status = TRUE;

    # Get the sorted list of GCP IDs from the hash
    my @gcp_ids = sort keys %$gcp_hash;

    # Make a copy of the list and shift the first ID off the list since it
    # doesn't need to be compared
    my @search_ids = (@gcp_ids);
    shift(@search_ids);

    # Loop over the full list of IDs
    foreach my $gcp_id (@gcp_ids)
    {
        # Use the first record for comparisons since all the records in the
        # chip array have the same metadata
        my $gcp_rec = $gcp_hash->{$gcp_id}->[0];

        # Loop over the remaining IDs in the search list
        foreach my $search_id (@search_ids)
        {
            if (is_compatible_gcp($gcp_rec, $gcp_hash->{$search_id}->[0]))
            {
                # The points are the same, so issue a warning and set the
                # return status to indicate the points are not unique
                $logger->warn_msg("GCP IDs $gcp_id and $search_id are the "
                    . "same point");
                $status = FALSE;
            }
        }

        shift(@search_ids);
    }

    return $status;
}

##############################################################################
# Name: get_gcp_metadata_members
#
# Description: Returns an array of field names that are considered GCP metadata
#
# Returns:
#   Array of GCP metadata field names
##############################################################################
sub get_gcp_metadata_members
{
    return qw(point_id point_active point_date_updated
                update_reason point_version point_path point_row
                source latitude longitude elevation point_type);
}

##############################################################################
# Name: get_image_chip_metadata_members
#
# Description: Returns an array of field names that are considered image chip
#   metadata
#
# Returns:
#   Array of image chip metadata field names
##############################################################################
sub get_image_chip_metadata_members
{
    return qw(chip_name chip_active chip_date_updated chip_version chip_path
                chip_row reference_line reference_sample projection_x
                projection_y pixel_size_x pixel_size_y chip_size_lines
                chip_size_samples chip_sensor chip_select_method moravec_rank
                projection zone abs_rel_flag chip_acq_date chip_data_type);
}

##############################################################################
# Name: is_same_gcp
#
# Description: compares two GCP records.  If they are the same, TRUE is
# returned.  Otherwise FALSE is returned.
#
# Parameters: two GCP hash references
#
# Returns: TRUE or FALSE
#
##############################################################################
sub is_same_gcp
{
    my $gcp1 = shift;
    my $gcp2 = shift;

    # GCP metadata that is expected to be the same for the same GCPs
    my @gcp_members = get_gcp_metadata_members();

    # Make a hash that indicates which fields should be compared as numbers
    my @numeric_members = qw(point_version point_path point_row latitude 
        longitude elevation);
    my %numeric_hash;
    @numeric_hash{@numeric_members} = ();

    # Compare each of the fields
    foreach my $key (@gcp_members)
    {
        if (exists $numeric_hash{$key})
        {
            # Use a numeric comparison
            if ($gcp1->{$key} != $gcp2->{$key})
            {
                my $logger = get_default_log();
                $logger->warn_msg("GCP field $key differs");
                return FALSE;
            }
        }
        else
        {
            # Use a string comparison
            if ($gcp1->{$key} ne $gcp2->{$key})
            {
                my $logger = get_default_log();
                $logger->warn_msg("GCP field $key differs");
                return FALSE;
            }
        }
    }

    return TRUE;
}

#############################################################################
# Name: is_compatible_gcp
#
# Description: A modified version of ias_same_gcp that considers fewer fields.
#     This is used to match new GCP records to compatible ones in the master
#     GCPs.
#
# Parameters: two GCP hash references
#
# Returns: TRUE or FALSE
#
##############################################################################
sub is_compatible_gcp
{
    my $gcp1 = shift;
    my $gcp2 = shift;

    # GCP metadata that is expected to be the same for the same GCPs
    my @gcp_members = qw(point_path point_row latitude longitude elevation);

    # Perform the single string comparison
    return FALSE if ($gcp1->{'source'} ne $gcp2->{'source'});

    # Check the numeric fields
    foreach my $key (@gcp_members)
    {
        return FALSE if ($gcp1->{$key} != $gcp2->{$key});
    }

    return TRUE;
}

###############################################################################
# Name: is_unique_image_chip
#
# Description: compares two sets of image chip metadata records.  If they are
#     the unique, TRUE is returned.  Otherwise FALSE is returned.
#
# Parameters: two image chip metadata hash references
#
# Returns: TRUE or FALSE
#
##############################################################################
sub is_unique_image_chip
{
    my $chip1 = shift;
    my $chip2 = shift;

    # Any one of these image chip fields being different make the chip be
    # considered unique (split into two lists to allow the right type of
    # comparison to be done)
    my @string_list = qw(projection chip_sensor abs_rel_flag chip_acq_date
        chip_data_type);
    my @number_list = qw(chip_path chip_row zone reference_line
        reference_sample projection_x projection_y pixel_size_x pixel_size_y
        chip_size_lines chip_size_samples);

    # Check the string fields
    foreach my $key (@string_list)
    {
        return TRUE if ($chip1->{$key} ne $chip2->{$key});
    }

    # Check the numerica fields
    foreach my $key (@number_list)
    {
        return TRUE if ($chip1->{$key} != $chip2->{$key});
    }

    return FALSE;
}

##############################################################################
# Name: verify_image_chip_data
#
# Description: Verify that the image chip is the specified size according to
#              the number of lines, number of samples, and datatype;
#
# Parameters:
#   - Image chip hash reference
#   - Image chip file
#
# Returns:
#   SUCCESS/ERROR
#   Date updated
##############################################################################
sub verify_image_chip_data
{
    my $chip = shift;
    my $chip_filename = shift;

    #Get the default logger object from the library 
    my $logger = get_default_log();
    
    # Check that the file exists first
    if (! -e $chip_filename)
    {
        $logger->fatal_msg("Chip $chip_filename does not exist");
        return ERROR;
    }

    # Calculate the datasize
    my $type_size = ias_gcp_get_sizeof_data_type($chip->{'chip_data_type'});
    if ($type_size == ERROR)
    {
        $logger->fatal_msg("Data type not valid $chip->{'chip_data_type'}");
        return ERROR;
    }

    # Grab samples and lines out of chip
    my $number_of_lines = $chip->{'chip_size_lines'};
    my $number_of_samples = $chip->{'chip_size_samples'};

    # Calculate total number of bytes
    my $total_bytes = $type_size * $number_of_lines * $number_of_samples;

    # Get the chips total filesize
    my $chip_file_size = -s $chip_filename;

    # Check if sizes match up
    if ($total_bytes != $chip_file_size)
    {
        $logger->fatal_msg("Chip size is $chip_file_size but expected chip "
            . "size is $total_bytes");
        return ERROR;
    }   

    return SUCCESS;
}

##############################################################################
# Name: create_chips_updated_file
#
# Description: Creates empty "chips_updated" file as flag that new chips or
#              chips were updated in directory.
#
# Parameters:
#   - Chip directory
#
# Returns:
#   SUCCESS/ERROR
##############################################################################
sub create_chips_updated_file
{
    my $path = shift;

    # Chips updated file name
    my $chips_updated_file =  'chips_updated';

    #Get the default logger object from the library 
    my $logger = get_default_log();

    # Open file
    my $status = open(FH, ">", "$path/$chips_updated_file");
    if (!defined $status)
    {
        $logger->fatal_msg("Unable to create chip updated file");
        return ERROR;
    } 
    close FH;

    return SUCCESS;
}

##############################################################################
# Name: verify_tar_data
#
# Description: Given a GCP record array and tarball file this routine will 
#              verify the data that was put into the tarball matches the 
#              original files
#
# Parameters:
#   - GCP records array
#   - Tarball file
#
# Returns:
#   SUCCESS/ERROR
##############################################################################
sub verify_tar_data
{
    my $gcp_records = shift;        # Records containing MD5 key
    my $tar_file = shift;           # Tarball filename

    my $logger = get_default_log(); # Get the default logger object

    # Use the built in tarball iterator
    my $iteration = Archive::Tar->iter($tar_file, 1, {});
    while (my $file = $iteration->())
    {
        # Grab the filename and calculate the md5 for data stored in tarball
        my $filename = $file->{'name'};
        my $tar_md5 = md5_hex($file->{'data'});

        # Match the tarball file with the file in the records
        foreach my $record (@{$gcp_records})
        {
            if ($filename eq $record->{'chip_name'})
            {
                # Check that the MD5 key is present
                if (!exists $record->{'MD5'})
                {
                    $logger->fatal_msg("MD5 key not present for $filename");
                    return ERROR;
                }
            
                # Compare the MD5Sum keys
                if ($record->{'MD5'} ne $tar_md5)
                {
                    $logger->fatal_msg("MD5Sum keys for $filename do not match".
                        " for file: $record->{'MD5'} and tar: $tar_md5");
                    return ERROR;
                }
                
                # Set this record as VERIFIED and continue to next file
                $record->{'VERIFIED'} = 1;
                last;
            }
        }
    }

    # Verify each record was present in tarball
    foreach my $record (@{$gcp_records})
    {
        if (!defined $record->{'VERIFIED'})
        {
            $logger->fatal_msg("Tarball did not contain ".
                "$record->{'chip_name'}");
            return ERROR;
        }
        delete $record->{'VERIFIED'};
    }

    return SUCCESS;
}

##############################################################################
# Name: read_image_chip_files
#
# Description: Will take a GCP records array reference read in the data 
#              associated with each image chip and attach it to the array
#
# Parameters:
#   - GCP Lookup Hash reference
#       -- Adds a image_chip entry to the hash for each GCP in the array
#   - GCP directory (Base directory where chips are located)
#
# Returns:
#   SUCCESS
#   ERROR
##############################################################################
sub read_image_chip_files
{
    my $gcp_records = shift;        # Reference to GCP lookup Hash
    my $gcp_directory = shift;      # Directory of image chips

    my $logger = get_default_log(); # Get the default logger object 
        
    # Loop through each chip adding it the filelist
    foreach my $record (@$gcp_records)
    {      
        my $file = "$gcp_directory"."$record->{'chip_name'}";

        # Get lines, samples, data size
        my $lines = $record->{'chip_size_lines'};
        my $samples = $record->{'chip_size_samples'};
        my $type = $record->{'chip_data_type'};

        # Get the data size and total bytes
        my $data_size = ias_gcp_get_sizeof_data_type($type);
        if ($data_size == ERROR)
        {
            $logger->fatal_msg("Unable to retrieve data type");
            return ERROR;   
        }

        # Determine the number of bytes to read
        my $bytes_to_read = $lines * $samples * $data_size;

        # Create file handle for current chip
        my $fh;
        if (!defined open($fh, "<", "$file"))
        {
            $logger->fatal_msg("Image chip $file could not be opened");
            return ERROR;
        }

        # Verify the chip file is the expected size.  The read will make sure
        # there are enough bytes in the file, but we want to also make sure
        # the file isn't longer than expected in case there was a problem
        # generating the file.  This is done after opening the file so we know
        # it already exists.
        if ((-s $file) != $bytes_to_read)
        {
            $logger->fatal_msg("Image chip $file does not agree with the "
                . "expected size of $bytes_to_read bytes");
            close($fh);
            return ERROR;
        }

        # Read in the chip 
        my $image_data;
        my $bytes_read = sysread($fh, $image_data, $bytes_to_read); 

        # Check for a serious read error
        if (!defined $bytes_read)
        {
            # Issue a warning for failing to read the file, flag it as an error
            # and skip to the next record
            $logger->fatal_msg("Error reading image chip file '$file': $!");
            close($fh);
            return ERROR;
        }

        # Make sure enough bytes were read
        if ($bytes_read != $bytes_to_read)
        {
            $logger->fatal_msg("Bytes read were $bytes_read when "
                . " $bytes_to_read were supposed to be read image chip file "
                . "'$file'");
            close($fh);
            return ERROR;
        }
        close($fh);

        # Put the chip data into the hash and push onto the array
        $record->{'image_chip'} = $image_data;
    }

    return SUCCESS;
}

