#!/usr/bin/perl -w
#############################################################################
# Name: set_gcp_status 
#
# Description:  Activates or Deactives GCPs and/or Image Chips in a GCPLib file.
#               Using the input file it searchs the passed in root directory for
#               the GCPLib files and updates the status of either the Chip or
#               GCP
#
# Assumptions: 
#       Input file:
#           1. List of GCP IDs.
#           2. List of Image Chip IDs.
#           3. Combination of both GCP IDs and Image Chip IDs.
#
# Parameters:
#   - Input filename
#   - GCPLib Root Directory
#   - Optional parameter: --activate
#       Can be used to activate the list of provided chips
#   - Optional parameter: --deactivate
#       Can be used to deactivate the list of provided chips
#   - Optional parameter: --test
#       Used to test whether the load will work without committing the update.
#
###########################################################################
use strict;
use Getopt::Long;           # command line parsing support
use File::Basename;         # basename routine
use Logger;                 # Logging module
use ias_const;              # define SUCCESS and ERROR
use ias_gcp qw(ias_gcp_read_gcplib ias_gcp_write_gcplib); # provides the GCPLib
                                                          # reading/writing
use manage_gcps_utils;

# Declare program constant
use constant PROG => basename($0);

# Initialize is_active flag to Y and change later if needed
my $is_active = "Y";

# Read the optional parameters
my $test_only = 0;          # Flag to only test the load, do not commit it
my $deactivate = 0;         # Flag to deactivate all chips in the file
my $activate = 0;           # Flag to activate all chips in the file
my $help = 0;               # Flag to display usage
my $date_updated;           # Date updated to use for modified records

if (!GetOptions('test' => \$test_only, 'activate' => \$activate,
       'deactivate' => \$deactivate, 'date_updated=s' => \$date_updated,
       'help' => \$help))
{
    # Error reading the options, so issue the usage message and exit
    usage();
    exit FAILURE;
}

# Enable help option
if ($help)
{
    usage();
    exit FAILURE;
}

# Confirm that activate and deactive are both not flagged
if (($activate && $deactivate) || !($activate || $deactivate))
{
    usage();    
    exit FAILURE;
}
$is_active = "N" if($deactivate);

# Confirm there are enough parameters remaining for the required parameters
if ($#ARGV != 1)
{
    # Error since no input filename provided
    usage();
    exit FAILURE;
}

# Get the name of input file 
my $input_filename = $ARGV[0];

# Get the name of the GCP dataset root directory
my $root_directory = $ARGV[1];

# Set up the logger
my $logger = Logger->new("STDOUT");
if (!$logger)
{
    print STDERR PROG . ": ERROR:  Failed to establish logger object for "
        . "STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name(PROG);
set_default_log($logger);

# Get the date updated to use
my $status;
($status, $date_updated) = get_date_updated($date_updated);
if ($status != SUCCESS)
{
    $logger->close_log();
    usage();
    exit FAILURE;
}

# Update the Active statues in the GCPLib file
if (update_gcplib($input_filename, $root_directory, $is_active, $test_only,
    $date_updated) != SUCCESS)
{
    $logger->fatal_msg("Failed to update GCPLib");
    $logger->close_log();
    exit FAILURE;
}

# Close log and exit
$logger->close_log();
exit SUCCESS;

##############################################################################
# Name: usage
# 
# Description: Prints out the usage message for the application
#
##############################################################################
sub usage
{
    print "\nUsage: ". PROG ." Input filename GCP Root directory \\\n".
        "    --activate|deactivate [--date_updated mm-dd-yyyy] [--test]\n\n";
}

##############################################################################
# Name: binary_search_chip_id
# 
# Description: Binary search of array passed which matchs the passed in value
#              with the chip_name hash value.
#
# Parameters:
#   - Array of hashes
#   - Int GCP ID
#
# Return:
#   - Index to value found
#   - ERROR if not found
##############################################################################
sub binary_search_chip_id
{
    my ($aref, $id) = @_; # Set array reference and value to find

    my ($low, $high) = (0, $#$aref); # set low and high values
    
    while ($low <= $high)
    {
        my $mid = int (($low + $high) / 2); # Set middle of array
        
        # Check if value is higher or lower else its a match
        $low  = $mid + 1, 
            next if ($id gt $aref->[$mid]->{'chip_name'});
        $high = $mid - 1, 
            next if ($id lt $aref->[$mid]->{'chip_name'});

        # Matches if reaches here
        return $mid;
    }

    return ERROR;
}

##############################################################################
# Name: get_path_row
#
# Description: Pulls out the path and row from string GCP ID or Chip ID
#
# Parameters:
#   - GCP ID
#
# Return
#   - SUCCESS, PATH, ROW - on SUCCESS
#   - ERROR, UNDEF, UNDEF - on ERROR
##############################################################################
sub get_path_row
{
    my $id = shift; # GCP ID 

    my $length = length($id);
    my $path;
    my $row;

    # Get the default logger object from the library 
    my $log = get_default_log();

    # Check if not a GCP ID
    if ($length != 10 && $length != 13)
    {
        $log->fatal_msg("ID $id is not formatted correctly"); 
        return ERROR, undef, undef;
    }

    # Parse the path and row out of the id
    if ($id =~ /^(\d{3})(\d{3})\d{4}(_\d{2})?/)
    {
        $path = $1;
        $row = $2;
    }
    else
    {
        $log->fatal_msg("ID $id is not formatted correctly"); 
        return ERROR, undef, undef;
    }

    if ($path < 1 || $path > 233)
    {
        $log->fatal_msg("Path $path incorrect from ID $id "); 
        return ERROR, undef, undef;
    }

    if ($row < 1 || $row > 248)
    {
        $log->fatal_msg("Row $row incorrect from ID $id "); 
        return ERROR, undef, undef;
    }

    return SUCCESS, $path, $row;
}

##############################################################################
# Name: update_gcplib
#
# Description: Updates the GCPLib with the correct status according to the GCP
#              IDs and Chip IDs from the input file.
#
# Parameters:
#   - Input filename
#   - Root Directory
#   - Active flag (Y or N)
#   - Test Flag (1, 0)
#   - Date updated
#
# Returns:
#   SUCCESS or ERROR
##############################################################################
sub update_gcplib
{
    my $input_filename = shift; # Input file with GCP/Chip IDs
    my $root_directory = shift; # GCPLib root directory 
    my $is_active = shift;      # Active flag to set ID's to
    my $is_test = shift;        # Flag to write out GCPLib
    my $date_updated = shift;   # Date to use for updated records
    my @gcp_ids;                # Arrray to read IDs into
    my $cell_count = 0;         # Count of path/row cells updated
    my $total_update_count = 0; # Count of GCPs and specific image chip ids
                                # updated

    # Get the default logger object from the library 
    my $log = get_default_log();

    # Open List of GCP IDs and/or Chip IDs
    my $status = open INPUT_FILE, "$input_filename";
    if (!defined($status))
    {
        $logger->fatal_msg("Opening $input_filename");
        return ERROR, undef;
    }

    # Read in the IDs from the input file
    while (my $line = <INPUT_FILE>)
    {
        # Skip comments, BEGIN, and empty lines
        next if $line =~ /^#/;
        next if $line =~ /^BEGIN/;
        $line =~ s/^\s+|\s+$//g;    
        next if $line eq '';

        # If the line doesn't contain an underscore, it is a GCP ID, so make
        # sure it has the leading zeros
        if (!($line =~ /_/))
        {
            $line = sprintf("%010d", $line);
        }

        # Store the ID into hash
        my $record = {};
        $record->{'ID'} = $line;

        # Pull the path row from the ID 
        ($status, $record->{'PATH'}, $record->{'ROW'}) = get_path_row($line);
        if ($status != SUCCESS)
        {
            $log->fatal_msg("Getting path/row for $line");
            return ERROR;
        }
        
        # Push hash onto array
        push @gcp_ids, $record;
    }
    close INPUT_FILE;

    # Sort the array by path/row
    @gcp_ids = sort 
    {
        if (int $a->{'PATH'} == int $b->{'PATH'})
        {
            return 0 if (int $a->{'ROW'} == int $b->{'ROW'});
            return 1 if (int $a->{'ROW'} > int $b->{'ROW'});
            return -1;
        }
        return 1 if (int $a->{'PATH'} > int $b->{'PATH'});
        return -1;
    } @gcp_ids;

    # Prepare to loop over the list of IDs
    my $previous_path = -1;
    my $previous_row = -1;
    my $gcplib_filename;
    my $overall_gcp_version;
    my @gcp_data;

    # Iterate through the IDs setting the ACTIVE status
    for my $record (@gcp_ids)
    {
        # Assign the record ID, PATH, ROW
        my $id = $record->{'ID'};
        my $path = $record->{'PATH'};
        my $row = $record->{'ROW'};

        # Check if new GCPLib is need and write out previous GCPLib
        if (int $row != int $previous_row || int $path != int $previous_path)
        {
            $cell_count++;

            if (defined $gcplib_filename)
            {
                # If test is enabled don't write out GCPLib file
                if ($is_test)
                {
                    $log->warn_msg("Completed successfully. Not writing out "
                        . "GCPLib because of --test option");
                }
                else
                {
                    $status = ias_gcp_write_gcplib($gcplib_filename, 
                        $overall_gcp_version, \@gcp_data);
                    if ($status != SUCCESS)
                    {
                        $log->fatal_msg("Failed to write GCPLib file");
                        return ERROR;
                    }
                }
            }

            # Set the previous path/row to current path/row
            $previous_path = $path;
            $previous_row = $row;
            
            # Tell user new path/row being used and undefine gcps
            $log->info_msg("Updating path: $path row: $row");
            @gcp_data = undef;

            # Build the new full GCPLib filename
            $gcplib_filename = "$root_directory/" 
                . sprintf("%03d/%03d/GCPLib", $path, $row);

            # Read new GCPLib file
            ($status, $overall_gcp_version, @gcp_data) = 
                ias_gcp_read_gcplib($gcplib_filename);
            if ($status != SUCCESS)
            {
                $log->fatal_msg("Failed to read Master GCPLib file from "
                    . $gcplib_filename);
                return ERROR;
            }

            # Sort the array by GCP ID then by Chip ID
            @gcp_data = sort 
            {
                if (int $a->{'point_id'} == int $b->{'point_id'})
                {
                    return 0 if ($a->{'chip_name'} eq $b->{'chip_name'});
                    return 1 if ($a->{'chip_name'} gt $b->{'chip_name'});    
                    return -1;
                }
                return 1 if (int $a->{'point_id'} > int $b->{'point_id'});
                return -1;
            } @gcp_data;
        }

        # Check if ID is a Chip ID
        if (index($id, '_') != -1)
        {
            # Chip ID is unique use binary search to find it
            my $match = binary_search_chip_id(\@gcp_data, $id);
            if ($match == ERROR)
            {
                $log->fatal_msg("Could not find Chip $id in GCPLib");
                return ERROR;
            }
            
            # Match was found index into array was stored in $match
            # Update the image chip.
            $gcp_data[$match]->{'chip_active'} = $is_active;
            $gcp_data[$match]->{'chip_date_updated'} = $date_updated;
            $total_update_count++;
            next;   
        }

        my $num_updated = 0;
        for my $gcp_record (@gcp_data)
        {
            # Check if IDs match
            if ($gcp_record->{'point_id'} == $id)
            {
                $gcp_record->{'point_active'} = $is_active;
                $gcp_record->{'point_date_updated'} = $date_updated;
                $num_updated++;
            }
        }

        # Check if any records in GCPLib were updated
        if (!$num_updated)
        {
            $log->fatal_msg("Could not find GCP $id in GCPLib");
            return ERROR;
        }
        $total_update_count++;
    }

    # If test is enabled don't write out GCPLib file
    if ($is_test)
    {
        $log->warn_msg("Completed successfully. Not writing out GCPLib".
            " because of --test option");
    }
    else
    {
        # Write out last GCPLib file
        $status = ias_gcp_write_gcplib($gcplib_filename, $overall_gcp_version,
            \@gcp_data);
        if ($status != SUCCESS)
        {
            $log->fatal_msg("Failed to write GCPLib file");
            return ERROR;
        }
    }

    $log->info_msg("Updated $total_update_count GCPs in $cell_count "
        . "directories");

    return SUCCESS;
}
