#!/usr/bin/perl -w
#############################################################################
# Name: tar_image_chips.pl
#
# Description: This script will take a GCPLib file and gather all the data from
#              its image chips and put them all into a single tarball. The 
#              chips_updated file must be present in order for the tarball
#              to be created.
#
# Parameters:
#   - Master GCPLib file
#   - Verify Tar (Enables or disables the tarball verification)
#   - Test       (Save the tarball as test file)
#   - Compression(Determines what type of tar compression will be used)
#############################################################################
use strict;
use Getopt::Long;               # Command line parsing support
use File::Basename;             # Basename routine
use Logger;                     # Logging module
use ias_const;                  # Define SUCCESS and ERROR
use manage_gcps_utils;          # Utility routines for managing GCPs
use ias_gcp qw(ias_gcp_tar_image_chips ias_gcp_create_tar_filename 
               IAS_GCP_BZIP2 IAS_GCP_GZIP IAS_GCP_NO_COMPRESSION
               IAS_GCP_IMAGE_CHIPS_FILENAME IAS_GCP_DO_MD5SUM); # Tar routines 

# Declare program constant
use constant PROG => basename($0);

# Read the optional parameters
my $verify_tar = "YES"; # Flag to verify_chips
my $compression;        # Type of compression to be used
my $test = 0;           # Flag to test tar
my $tar_file;           # Tar filename

if (!GetOptions('verify_tar=s' => \$verify_tar,
        'compression=s' => \$compression, 'test' => \$test))
{
    # Error reading the options, so issue the usage message and exit
    usage();
    exit FAILURE;
}

# Check for valid values for verify_tar
$verify_tar = uc($verify_tar);
if ($verify_tar ne "YES" && $verify_tar ne "NO")
{
    usage();
    exit FAILURE;
}

# Check that the compression option is valid
$compression = uc($compression);
if ($compression ne "BZIP2" && $compression ne "GZIP" && $compression
    ne "NO_COMPRESSION")
{
    usage();
    exit FAILURE;
}
elsif ($compression eq "NO_COMPRESSION")
{
    $compression = IAS_GCP_NO_COMPRESSION;
}
elsif ($compression eq "BZIP2")
{
    $compression = IAS_GCP_BZIP2;
}
else
{
    $compression = IAS_GCP_GZIP;
}

# Confirm there are enough parameters remaining for the required parameters
if ($#ARGV != 0)
{
    # Error since no GCPLib filename provided
    usage();
    exit FAILURE;
}

# Get the master GCPLib filename
my $master_gcplib = $ARGV[0];

# Set up the logger
my $logger = Logger->new("STDOUT");
if (!$logger)
{
    print STDERR PROG . ": ERROR:  Failed to establish logger object for "
        . "STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name(PROG);
set_default_log($logger);

# Read the master file
$logger->info_msg("Master GCPLib: $master_gcplib");

my $master_file = read_gcplib_file($master_gcplib, MASTER_FILE);
if (!defined $master_file)
{
    $logger->fatal_msg("Error reading master GCPLib file $master_gcplib");
    $logger->close_log();
    exit FAILURE;
}

# Read the image chips into memory
$logger->info_msg("Reading image chips");

my $status = read_image_chip_files($master_file->{'records'}, 
    $master_file->{'dir'});
if ($status != SUCCESS)
{
    $logger->fatal_msg("Preparing hash for tarball");
    $logger->close_log();
    exit FAILURE;
}

# Create backup of md5file
my $md5file =  $master_file->{'dir'} . IAS_GCP_IMAGE_CHIPS_FILENAME . ".md5";
if (-e "$md5file" && !$test)
{
    $status = system("mv $md5file $md5file.bak");
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Unable to create backup of md5file");
        $logger->close_log();
        exit FAILURE;
    }
}

# Create tarball using prepared file list of image chips
$logger->info_msg("Creating tarball of image chips");

if ($test)
{
    $tar_file = ias_gcp_create_tar_filename(IAS_GCP_IMAGE_CHIPS_FILENAME 
        . ".test", $compression);
}
else
{
    $tar_file = ias_gcp_create_tar_filename(IAS_GCP_IMAGE_CHIPS_FILENAME, 
        $compression);
}

if (!defined $tar_file)
{   
    $logger->fatal_msg("Creating tarball filename");
    $logger->close_log();
    exit FAILURE;
}

# Create full path to tar file
$tar_file = $master_file->{'dir'} . $tar_file;

# If the tar file does exist create a backup
if (-e $tar_file && !$test)
{
    $status = system("mv $tar_file $tar_file.bak");
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Unable to create backup of tar");
        $logger->close_log();
        exit FAILURE;
    }
}

$status = ias_gcp_tar_image_chips($master_file->{'records'}, 
    $tar_file, $compression, IAS_GCP_DO_MD5SUM);
if ($status != SUCCESS)
{
    $logger->fatal_msg("Creating tarball for image chips");
    $logger->close_log();
    exit FAILURE;
}

# Verification of the tarball using md5 hash
if ($verify_tar eq "YES")
{
    if (defined($tar_file))
    {
        $status = verify_tar_data($master_file->{'records'}, $tar_file);
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Verifying tar file $tar_file");
            $logger->close_log();
            exit FAILURE;
        }
        $logger->info_msg("All md5sum keys were verified");
    }
    else
    {
        $logger->info_msg("Tarball not created, unable to verify it");
    }
}

$logger->close_log();

exit SUCCESS;

##############################################################################
# Name: usage
# 
# Description: Prints out the usage message for the application
#
##############################################################################
sub usage
{
    # Print out the usage message
    print "\nUsage: ". PROG . " master_GCPLib_filename [--verify_tar] [--test]".
        " [--compression] \n\n  --verify_tar\n    - This option enables or".
        " disables the verification of the tar\n    - The default for this".
        " option is true\n    - Examples: --verify_tar=YES --verify_tar=NO".
        " \n\n  --test\n    - This option will not overwrite the current tar".
        " it will create a tarball appended\n      with .test instead \n\n ".
        " --compression\n    - This option is required and determines which".
        " type of tar compression is used\n    - Examples:".
        "--compression=BZIP2 --compression=GZIP --compression=NO_COMPRESSION".
        "\n\n";
}
