#!/usr/bin/perl -w
#############################################################################
# Name: test_add_gcps.pl
#
# Description:  Testing of the perl script add_gcps.pl
#
###########################################################################
use strict;
use File::Basename;         # basename routine
use Logger;                 # Logging module
use ias_const;              # define SUCCESS and ERROR

# Created the formatted date for today
my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime();
my $formatted_date = sprintf("%02d-%02d-%04d", $mon+1, $mday, $year+1900);

# Create arrays for Test, Verification, and Test Descriptions
my @all_tests;
my $logfile = "test_add_gcps.log";

# Where to copy the entire GCP test directory
my $ut_dir = "./data/";
my $gcplib_directory = $ut_dir."add_gcps_directory";

# Remove the log file from a previous run
`rm -f $logfile`;

# Set up the logger
my $logger = Logger->new("STDOUT");
if (!$logger)
{
    print STDERR ": ERROR:  Failed to establish logger object for "
        . "STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name("test_set_gcp_status");
set_default_log($logger);

print "\n--------- Starting " . basename($0) . " test ------------------\n\n";

# Initialize for testing
initialize();

# Setup Description, Test, and Verification for all tests
# Read descriptions for test information
push @all_tests,{desc => "Testing that error thrown for bad date",
    test => "../add_gcps.pl add_gcps_directory/132/050/GCPLib".
    " add_gcps_directory/GCPLib.add --test --date_updated=12132012", 
    success => ERROR, replace => 0};

push @all_tests,{desc => "Testing that error thrown for bad Master GCPLib",
    test => "../add_gcps.pl add_gcps_directory/add_gcps/GCPLib.bad".
    " add_gcps_directory/add_gcps/GCPLib.add --test", 
    success => ERROR, replace => 0};

push @all_tests,{desc => "Testing that error thrown for bad Update GCPLib",
    test => "../add_gcps.pl add_gcps_directory/132/050/GCPLib".
    " add_gcps_directory/add_gcps/GCPLib.bad --test", 
    success => ERROR, replace => 0};

push @all_tests,{desc => "Testing that error thrown for inactive GCP",
    test => "../add_gcps.pl add_gcps_directory/add_gcps/GCPLib.inact.gcp".
    " add_gcps_directory/add_gcps/GCPLib.add --test", 
    success => ERROR, replace => 0};

push @all_tests,{desc => "Testing bad copy of chips",
    test => "../add_gcps.pl add_gcps_directory/132/050/GCPLib".
    " add_gcps_directory/add_gcps/GCPLib.bad_chips", 
    success => ERROR, replace => 0};

push @all_tests,{desc => "Testing that the --test creates GCPLib.test",
    test => "../add_gcps.pl add_gcps_directory/132/050/GCPLib".
    " add_gcps_directory/add_gcps/GCPLib.add --test --version=23", 
    diff =>"GCPLib.test",
    original =>"add_gcps_directory/add_gcps/GCPLib.test.verify",
    success => SUCCESS, replace => 1};

push @all_tests,{desc => "Testing creation of new GCPLib file and Date updated".
    " and that the GCPLib version is used",
    test => "../add_gcps.pl --date_updated 04-23-2014 add_gcps_directory".
    "/132/050/GCPLib add_gcps_directory/add_gcps/GCPLib.add_50", 
    diff =>"add_gcps_directory/132/050/GCPLib.updated",
    original =>"add_gcps_directory/add_gcps/GCPLib.date.verify",
    success => SUCCESS, replace => 0};

push @all_tests,{desc => "Testing that error is throw when wrong".
    " size image chip is provided",
    test => "../add_gcps.pl add_gcps_directory/132/050/GCPLib".
    " add_gcps_directory/add_gcps/GCPLib.add_50", 
    additional_files => "mv add_gcps_directory/add_gcps/1320500012_01.".
    "large_chip add_gcps_directory/add_gcps/1320500012_01",
    success => ERROR, replace => 0};

# Iterate through all of the tests
foreach my $test (@all_tests)
{
    # Print out the test description
    print STDOUT "------ $test->{'desc'}\n";

    # Check if additional files need moved or copied
    if (defined $test->{'additional_files'})
    {
        my $status = system($test->{'additional_files'});
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Getting addtional files for test");
            $logger->close_log();
            exit FAILURE;
        }
    }

    # Run the test
    my $status = system($test->{'test'}." >>$logfile");
    if ($test->{'success'} == ERROR)
    {
        if ($status == SUCCESS)
        {
            $logger->fatal_msg("Failed to run test $test->{'test'}");
        }
    }
    else
    {
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Failed to run test $test->{'test'}");
        }
    }

    if ($test->{'success'} == SUCCESS)
    {
        if ($test->{'replace'})
        {
            # Replace current date with 04-01-2014 to make verifying easier
            my @replace = ("sed", "-i", "s|$formatted_date|04-01-2014|g",
                "$test->{'diff'}");
            $status = system(@replace);
            if ($status != SUCCESS)
            {
                $logger->fatal_msg("Failed to replace date in $test->{'diff'}");
                $logger->close_log();
                exit FAILURE;
            }
        }

        # Run diff to ensure test was successful
        $status = system("diff $test->{'diff'} $test->{'original'} -q");
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Failed to verify test $test->{'diff'} for".
                "$test->{'test'}");
            $logger->close_log();
            exit FAILURE;
        }
    }
}

# Run cleanup after testing
if (-d "add_gcps_directory")
{
    my $cleanup = "rm -rf ./add_gcps_directory";
    my $status = system($cleanup);
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Failed to clean up after test");
        $logger->close_log();
        exit FAILURE;
    }
}

print "\n--------- Completed " . basename($0) . " test ------------------\n\n";

exit SUCCESS;

################################################################################
# Name: initialize
#
# Purpose: Initialize all test directorys and files
# ##############################################################################
sub initialize
{
    my @init_commands; # Init commands array
    
    # Clean up directory first 
    if (-d "add_gcps_directory")
    {
        my $cleanup = "rm -rf ./add_gcps_directory";
        my $status = system($cleanup);
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Failed to cleanup before test");
            $logger->close_log();
            exit FAILURE;
        }
    }

    # Start of all init commands
    push @init_commands,{command => "mkdir add_gcps_directory"};
    push @init_commands,{command => "mkdir add_gcps_directory/132"};
    push @init_commands,{command => "mkdir add_gcps_directory/132/050"};
    push @init_commands,{command => "mkdir add_gcps_directory/add_gcps"};
    push @init_commands,{command => "cp data/add_gcps_directory/add_gcps".
        "/GCP* add_gcps_directory/add_gcps/"};
    push @init_commands,{command => "cp data/add_gcps_directory/add_gcps".
        "/132* add_gcps_directory/add_gcps/"};
    push @init_commands,{command => "cp data/add_gcps_directory/132/050/GCPLib".
        " add_gcps_directory/132/050/GCPLib"};
    push @init_commands,{command => "cp data/add_gcps_directory/132/050".
        "/1320500012_01 add_gcps_directory/132/050/1320500012_01"};
    push @init_commands,{command => "cp data/add_gcps_directory/132/050".
        "/1320500027_01 add_gcps_directory/132/050/1320500027_01"};
    push @init_commands,{command => "cp data/add_gcps_directory/132/050".
        "/1320500034_01 add_gcps_directory/132/050/1320500034_01"};
    push @init_commands,{command => "cp data/add_gcps_directory/132/050".
        "/1320500036_01 add_gcps_directory/132/050/1320500036_01"};
    push @init_commands,{command => "cp data/add_gcps_directory/132/050".
        "/1320500066_01 add_gcps_directory/132/050/1320500066_01"};
    push @init_commands,{command => "cp data/add_gcps_directory/132/050".
        "/1320500139_01 add_gcps_directory/132/050/1320500139_01"};

    # Run each command in order
    foreach my $command (@init_commands)
    {
        # Run Init to copy gcp directory
        my $status = system($command->{'command'});
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Failed to initialize $command->{command}");
            $logger->close_log();
            exit FAILURE;
        }
    }
}
