#!/usr/bin/perl -w
##############################################################################
# Name: test_load_gcps_into_database.pl
#
# Description: Testing the functionality of the load_gcps_into_database.pl 
#              script.
##############################################################################
use strict;
use File::Basename;         # Basename routine
use ias_const;              # Define SUCCESS and ERROR

print "\n--------- Starting " . basename($0) . " test ------------------\n\n";

my $test_result = SUCCESS;
my $ut_dir = "./data/";
my $create_table_sql = $ut_dir."create_dev_version_gcp_image_tables.sql";
my $delete_table_sql = $ut_dir."delete_dev_version_gcp_image_tables.sql";
my $logfile = "test_load_gcps_into_database.log";

# Make a patched version of get_gps.pl to read the fake tables
make_modified_get_gcps_script($ut_dir);

# Remove the log file from a previous run
`rm -f $logfile`;

# Create the fake tables to use for the unit test
if (run_db_script($create_table_sql) != SUCCESS)
{
    print "Error loading the development tables";
    exit FAILURE;
}

# Initialize all the tests
my @all_tests;
initialize_tests(\@all_tests);

#loop through all of the tests
for my $test (@all_tests)
{
    my $status = SUCCESS;
    my $description = shift @$test;
    my $verify_routine = shift @$test;
    my $path = shift @$test;
    my $row = shift @$test;
    my $expected_outcome = shift @$test;
    my $test_path = @$test[2];
    @$test[2] = $ut_dir."load_gcps_test_data/GCPLib_$test_path/GCPLib";
    push @{$test}, ">>";
    push @{$test}, "$logfile";

    # Ouput description of test running
    print "------ $description\n";

    $status = system("@$test");
    if ($expected_outcome != ERROR)
    {
        if ($status != SUCCESS)
        {
            run_db_script($delete_table_sql);
            print "Error running test $test_path\n";
            print "\n------ " . basename($0) . " failed\n\n";
            exit FAILURE;
        }
    }

    # Verify the results
    if ($verify_routine)
    {
        if(verify_db_contents($path, $row, "output_$test_path", 
            $ut_dir."load_gcps_test_data/GCPLib_$test_path") != SUCCESS)
        {
            run_db_script($delete_table_sql);
            print "Error verifying test $test_path \n";
            print "\n------ " . basename($0) . " failed\n\n";
            exit FAILURE;
        }
    }
}

# Delete the fake tables 
if (run_db_script($delete_table_sql) != SUCCESS)
{
    print "Error deleting the development tables";
    exit FAILURE;
}

print "\n--------- Completed " . basename($0) . " test ------------------\n\n";

exit SUCCESS;

##############################################################################
# Name: initialize_tests
#
# Description: This routine adds all of the test parameters to the main test
#              array so it can easily loop through all the tests.
#
# Return: Array of Arrays
##############################################################################
sub initialize_tests
{
    my @all_test = shift;

    push @all_tests, ["Testing error on multiple image ids",
        TRUE, 151, 30, ERROR, "../load_gcps_into_database.pl", $ENV{IAS_DB_COM},
        "multiple_image_ids", "CRDEV", "--unit-test"];

    push @all_tests, ["Testing error on invalid GCP path",
        TRUE, 151, 30, ERROR, "../load_gcps_into_database.pl", $ENV{IAS_DB_COM},
        "bad_gcp_path", "CRDEV", "--unit-test"];

    push @all_tests, ["Testing error on invalid GCP row",
        TRUE, 151, 30, ERROR, "../load_gcps_into_database.pl", $ENV{IAS_DB_COM},
        "bad_gcp_row", "CRDEV", "--unit-test"];

    push @all_tests, ["Testing error on invalid image chip path",
        TRUE, 151, 30, ERROR, "../load_gcps_into_database.pl", $ENV{IAS_DB_COM},
        "bad_chip_path", "CRDEV", "--unit-test"];

    push @all_tests, ["Testing error on invalid image chip row",
        TRUE, 151, 30, ERROR, "../load_gcps_into_database.pl", $ENV{IAS_DB_COM},
        "bad_chip_row", "CRDEV", "--unit-test"];

    push @all_tests, ["Testing error on invalid GCP id",
        TRUE, 151, 30, ERROR, "../load_gcps_into_database.pl", $ENV{IAS_DB_COM},
        "bad_gcp_id", "CRDEV", "--unit-test"];

    push @all_tests, ["Testing error on invalid image chip id",
        TRUE, 151, 30, ERROR, "../load_gcps_into_database.pl", $ENV{IAS_DB_COM},
        "bad_chip_id", "CRDEV", "--unit-test"];

    push @all_tests, ["Testing error on invalid chip size",
        TRUE, 151, 30, ERROR, "../load_gcps_into_database.pl", $ENV{IAS_DB_COM},
        "bad_chip_size", "CRDEV", "--unit-test"];

    push @all_tests, ["Testing error on missing chip",
        TRUE, 151, 30, ERROR, "../load_gcps_into_database.pl", $ENV{IAS_DB_COM},
        "missing_chip", "CRDEV", "--unit-test"];

    push @all_tests, ["Normal Test with test option enabled", TRUE, 151, 30,
        SUCCESS, "../load_gcps_into_database.pl", $ENV{IAS_DB_COM},
        "test_option", "CRDEV", "--test", "--unit-test"]; 

    push @all_tests,  ["Normal Test with test data that is unordered", TRUE,
        151, 30, SUCCESS, "../load_gcps_into_database.pl", $ENV{IAS_DB_COM},
        "unordered_gcplib", "CRDEV", "--unit-test"];

    push @all_tests, ["Normal Test with same path row as previous test", TRUE, 
        151, 30, SUCCESS, "../load_gcps_into_database.pl", $ENV{IAS_DB_COM},
        "delete_same_gcps", "CRDEV", "--unit-test"];

    push @all_tests, ["Normal Test with different path row verify that old".
        " data is still in database ", TRUE, 151, 30, SUCCESS, 
        "../load_gcps_into_database.pl", $ENV{IAS_DB_COM},
        "different_path_row", "CRDEV", "--unit-test"];
}

##############################################################################
# Name: run_db_script
#
# Description: Runs passed in script using sqlplus. Script must contain exit
#              at the end of script.
# 
##############################################################################
sub run_db_script
{
    my $db_script = shift;

    `sqlplus $ENV{IAS_DB_COM} \@$db_script`;
    if ($? != SUCCESS)
    {
        print "Error running script $db_script\n";
        return ERROR;
    }
}

##############################################################################
# Name: make_modified_get_gcps_script
#
# Description: Create a temporary copy of get_gcps.pl that reads the fake
#   database tables
#
##############################################################################
sub make_modified_get_gcps_script
{
    my $ut_dir = shift;

    # get a copy of the script
    `cp $ENV{IAS_BIN}/get_gcps.pl .`;
    if ($? != SUCCESS)
    {
        print "Error copying get_gcps.pl from $ENV{IAS_BIN}\n";
        return ERROR;
    }

    # patch it to use the fake tables

    my $command = "patch -p0 < ".$ut_dir."get_gcps.patch";
    `$command`;
    if ($? != SUCCESS)
    {
        print "Error patching get_gcps.pl\n";
        return ERROR;
    }

    # modify the path to look in the current working directory first
    $ENV{PATH} = ".:$ENV{PATH}";
}

##############################################################################
# Name: verify_db_contents
#
# Description: Verify the DB contents by extracting them from the database
#   with the get_gcps_wrapper.pl script
#
# Parameters:
#   - WRS path
#   - WRS row
#   - The output directory to use
#   - The expected output GCPLib directory (Always uses GCPLib.verify)
#
# Returns:
#   SUCCESS or ERROR
##############################################################################
sub verify_db_contents
{
    my $wrs_path = shift;
    my $wrs_row = shift;
    my $output_dir = shift;
    my $expected_output_dir = shift;

    # Get the GCPs that were just loaded from the database
    my $command = "get_gcps_wrapper.pl --gcp_source GLS --wrs_path $wrs_path "
        . "--wrs_row $wrs_row --output $output_dir";

    my $output = `$command`;
    if ($? != SUCCESS)
    {
        print "Error reading GCPs from the database:\n$output\n";
        return ERROR;
    }

    # Open for reading
    my $status = open GCPLIB, "<", "$output_dir/GCPLib";
    if (!defined($status))
    {
        print "Can not open GCPLib for reading ";
        return ERROR;
    }

    # Open for writing
    $status = open GCPLIBOUT, ">", "$output_dir/GCPLib.verify";
    if (!defined($status))
    {
        print "Can not open GCPLib for writing";
        return ERROR;
    }

    # Read the output file
    while (my $line = <GCPLIB>)
    { 
        chomp $line;
        if ($line =~ /^VERSION = \d+/)
        {
            print GCPLIBOUT "VERSION = 1\n";
            next;
        }
        print GCPLIBOUT "$line\n";
    }
 
    close(GCPLIB);
    close(GCPLIBOUT);

    # Compare the GCPLib file with the expected one
    $command = "diff -q $output_dir/GCPLib.verify ".
        "$expected_output_dir/GCPLib.verify";
    `$command`;
    if ($? != SUCCESS)
    {
        print "Error: Data pulled from the database does not match expected:\n"
            . "    Command $command\n";
        return ERROR;
    } 
    else
    {
        `rm -rf $output_dir`;
        if ($? != SUCCESS)
        {
            print "Error deleting output directory: $output_dir\n";
        }
    }

    return SUCCESS;
}
