#!/usr/bin/perl -w
##############################################################################
# Name: test_load_overall_gcp_version.pl
#
# Description: Tests the functionality of the the load_overall_gcp_version.pl 
#              script.
##############################################################################
use strict;
use DBI;                    # Provides database access  
use Logger;                 # Provides logging functionality
use ias_db;                 # Provides addtional database functionality
use File::Basename;         # Basename routine
use ias_const;              # Define SUCCESS and ERROR

use constant PROG => basename($0);

my $test_result = SUCCESS;
my $db_user_password = $ENV{IAS_DB_COM};
my $ut_dir = "./data/";
my $create_table_sql = $ut_dir."create_dev_version_gcp_image_tables.sql";
my $delete_table_sql = $ut_dir."delete_dev_version_gcp_image_tables.sql";
my $logfile = "test_load_overall_gcp_version.log";

print "\n--------- Starting " . basename($0) . " test ------------------\n\n";

# Remove the log file from a previous run
`rm -f $logfile`;

# Set up the logger
my $logger = Logger->new("STDOUT");
if (!$logger)
{
    print STDERR PROG . ": ERROR:  Failed to establish logger object for "
        . "STDOUT.\n";
    exit ERROR;
}
$logger->set_module_name(PROG);
set_default_log($logger);

# Run error checking tests
# Run error test 1
my $command = "../load_overall_gcp_version.pl $db_user_password ".
    "$ut_dir/gcp_version_master_1.txt CRDEV --unit-test";
print("------ Running test to throw DB table error\n");
my $status = system($command." >>$logfile");
if ($status == SUCCESS)
{
    $logger->fatal_msg("Error running error test 1");
    $logger->close_log();
    run_db_script($delete_table_sql);
    exit ERROR;
}

# Create the fake tables to use for the unit test
if (run_db_script($create_table_sql) != SUCCESS)
{
    $logger->fatal_msg("Error loading the development tables");
    $logger->close_log();
    exit ERROR;
}

# Run error test 2
$command = "../load_overall_gcp_version.pl $db_user_password ".
    "$ut_dir/gcp_version_master_DNE.txt CRDEV --unit-test --test";
print("------ Running test to throw error data file does not exist\n");
$status = system($command." >>$logfile");
if ($status == SUCCESS)
{
    $logger->fatal_msg("Error running error test 2");
    $logger->close_log();
    run_db_script($delete_table_sql);
    exit ERROR;
}

# Run error test 3
$command = "../load_overall_gcp_version.pl $db_user_password ".
    "$ut_dir/gcp_version_master_bad.txt CRDEV --unit-test --test";
print("------ Running test to try to load bad data file \n");
$status = system($command." >>$logfile");
if ($status == SUCCESS)
{
    $logger->fatal_msg("Error running error test 3");
    $logger->close_log();
    run_db_script($delete_table_sql);
    exit ERROR;
}

# Run test 1
$command = "../load_overall_gcp_version.pl $db_user_password ".
    "$ut_dir/gcp_version_master_1.txt CRDEV --unit-test";
print("------ Running test to load unordered overall gcp version file\n");
$status = system($command." >>$logfile");
if ($status != SUCCESS)
{
    $logger->fatal_msg("Error running load_overall_gcp_version.pl test1");
    $logger->close_log();
    run_db_script($delete_table_sql);
    exit ERROR;
}

# Separate the DB username/password
my ($db_user, $db_password) = split('/', $ENV{IAS_DB_COM});
if ((!defined $db_user) || (!defined $db_password))
{
    $logger->fatal_msg("Error parsing the database_user/password");
    $logger->close_log();
    run_db_script($delete_table_sql);
    exit ERROR;
}

# Connect to the database.  No need to start transaction for all the database
# updates since the routine turns off AutoCommit by default.
my $dbh = ias_db_ConnectToDB($db_user, $db_password);
if (!defined $dbh)
{
    $logger->fatal_msg("Error connecting to the database: $ias_db_error");
    $logger->close_log();
    run_db_script($delete_table_sql);
    exit ERROR;
}

# Verify test 1 using verify_test_1 pattern matcher
if (verify_test($dbh, \&verify_test_1) != SUCCESS)
{
    $logger->fatal_msg("Could not verify the data from test 1");
    $logger->close_log();
    ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
    run_db_script($delete_table_sql);
    exit ERROR;
}

# Running test 2 using the --test option to ensure nothing in
# the database is changed
$command = "../load_overall_gcp_version.pl $db_user_password ".
    "$ut_dir/gcp_version_master_2.txt CRDEV --unit-test --test";
print("------ Running test with --test option ".
    "enable to ensure no DB changes\n");
$status = system($command." >>$logfile");
if ($status != SUCCESS)
{
    $logger->fatal_msg("Error running load_overall_gcp_version.pl with --test");
    $logger->close_log();
    ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
    run_db_script($delete_table_sql);
    exit ERROR;
}

# Verify the results are same as test 1
if (verify_test($dbh, \&verify_test_1) != SUCCESS)
{
    $logger->fatal_msg("Could not verify the data from test 1 second check");
    $logger->close_log();
    ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
    run_db_script($delete_table_sql);
    exit ERROR;
}

# Running test two to ensure data is being overwritten
$command = "../load_overall_gcp_version.pl $db_user_password ".
    "$ut_dir/gcp_version_master_2.txt CRDEV --unit-test";
print("------ Running test 2 to overwrite the current data".
    " with newest version\n");
$status = system($command." >>$logfile");
if ($status != SUCCESS)
{
    $logger->fatal_msg("Error running load_overall_gcp_version.pl test2");
    $logger->close_log();
    ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
    run_db_script($delete_table_sql);
    exit ERROR;
}

# Verify results  using the test_2 pattern matcher
if (verify_test($dbh, \&verify_test_2) != SUCCESS)
{
    $logger->fatal_msg("Could not verify the data from test 2");
    $logger->close_log();
    ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
    run_db_script($delete_table_sql);
    exit ERROR;
}

# Disconnect from database and close log
if (ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK) != 1)
{
    $logger->fatal_msg("Error disconnecting from the database");
    $logger->close_log();
    run_db_script($delete_table_sql);
    exit ERROR;
}

# Delete the fake tables to use for the unit test
if (run_db_script($delete_table_sql) != SUCCESS)
{
    $logger->fatal_msg("Error deleting the development tables");
    exit ERROR;
}

print "\n--------- Completed " . basename($0) . " test ------------------\n\n";

exit SUCCESS;

##############################################################################
# Name: run_db_script
#
# Description: Runs passed in script using sqlplus. Script must contain exit
#              at the end of script.
# 
##############################################################################
sub run_db_script
{
    my $db_script = shift;

    `sqlplus $ENV{IAS_DB_COM} \@$db_script`;
    if ($? != 0)
    {
        $logger->fatal_msg("Error running script $db_script\n");
        return ERROR;
    }
}

##############################################################################
# Name: verify_test
#
# Description: Verifys the results of running the load_overall_gcp_version.pl
# 
# Parameters: 
#   - Database handler
#   - Reference to pattern matching routine that takes description and 
#     version as parameters.
# Returns: 
#   SUCCESS/ERROR
##############################################################################
sub verify_test
{
    my $dbh = shift;
    my $pattern_matcher = shift;
    my ($version, $description);
    my $logger = get_default_log();

    # Prepare the statement to be executed
    my $select_sql = "Select version, description from DEV_GCP_DATA_VERSION".
        " order by version desc";
    my $version_sth = $dbh->prepare($select_sql);
    if (!$version_sth)
    {
        $logger->fatal_msg("Could not prepare version select sql");
        return ERROR;
    }

    # Execute the sql and bind the columns
    if (!$version_sth->execute())
    {
        $logger->fatal_msg("Could not execute version select sql");
        return ERROR;
    }
    $version_sth->bind_col(1, \$version);
    $version_sth->bind_col(2, \$description);
    
    # Loop through rows ensure pattern matches correctly 
    while ($version_sth->fetch)
    {
        next if ($pattern_matcher->($description, $version));
        
        $logger->fatal_msg("ERROR: Version $version does not match the".
            " correct description:$description");
        $version_sth->finish();
        return ERROR; 
    }

    $version_sth->finish();
    return SUCCESS;
}

##############################################################################
# Name: verify_test_1
#
# Description: Returns the result of matching pattern for description of version
# 
# Parameters: 
#   - Description 
#   - Version
#
# Returns: 
#   TRUE/FALSE
##############################################################################
sub verify_test_1
{
    my $description = shift;
    my $version = shift;

    return ($description =~ /^This version describes version:$version$/);
}

##############################################################################
# Name: verify_test_2
#
# Description: Returns the result of matching pattern for description of version
# 
# Parameters: 
#   - Description 
#   - Version
#
# Returns: 
#   TRUE/FALSE
##############################################################################
sub verify_test_2
{
    my $description = shift;
    my $version = shift;

    return ($description =~ /$version:This version describes version:$version/);
}
