#!/usr/bin/perl -w
#############################################################################
# Name: test_manage_gcps_utils
#
# Description:  Testing of all the routines contained in the manage_gcps_utils
#               module.
#
###########################################################################
use strict;
use File::Basename;         # Basename routine
use Logger;                 # Logging module
use ias_processing_system;  # Define ias_error_msg and ias_validate_environment
use ias_const;              # Define SUCCESS and ERROR
use Time::localtime;        # To get today's date
use ias_gcp qw(ias_gcp_read_gcplib); # GCPLib read routine
use manage_gcps_utils;      # Manage GCPs module    
use Digest::MD5 qw(md5_hex);# Used to create md5 for image chips
use ias_gcp qw(ias_gcp_tar_image_chips IAS_GCP_BZIP2 
               IAS_GCP_DO_MD5SUM); # Tarball routines

# Declare routine stubs
sub test_get_date_updated();
sub test_read_gcplib_file();
sub test_create_record_index();
sub test_verify_gcps_are_consistent();
sub test_gcps_are_unique();
sub test_get_gcp_metadata_members();
sub test_is_same_gcp();
sub test_is_compatible_gcp();
sub test_is_unique_image_chip();
sub test_verify_image_chip_data();
sub test_read_image_chip_files();
sub test_verify_tar_data();
sub strfy_boolean($$);

# Setup variables
my $ut_dir = "./manage_gcps_data/";
my $cleanup = "rm -rf ./manage_gcps_data";

# Set up the logger
my $logger = Logger->new("STDOUT");
if (!$logger)
{
    print STDERR ": ERROR:  Failed to establish logger object for "
        . "STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name("test_manage_gcps_utils.pl");
set_default_log($logger);

print "\n--------- Starting " . basename($0) . " test ------------------\n\n";

# Initialize the test
initialize();

my $return_status = SUCCESS;

# Turn off verboseness in the manage_gcps_utils
set_verbose_state(0);

print "------ Running get_date_updated test\n";
if (test_get_date_updated() != SUCCESS)
{
    $logger->fatal_msg("test_get_date_updated failed");
    $return_status = ERROR; 
}

print "------ Running read_gcplib_file test\n";
if (test_read_gcplib_file() != SUCCESS)
{
    $logger->fatal_msg("test_read_gcplib_file failed");
    $return_status = ERROR; 
}

print "------ Running create_record_index test\n";
if (test_create_record_index() != SUCCESS)
{
    $logger->fatal_msg("test_create_record_index failed");
    $return_status = ERROR; 
}

print "------ Running test_verify_gcps_are_consistent test\n";
if (test_verify_gcps_are_consistent() != SUCCESS)
{
    $logger->fatal_msg("test_verify_gcps_are_consistent failed");
    $return_status = ERROR; 
}

print "------- Running test_gcps_are_unique test\n";
if (test_gcps_are_unique() != SUCCESS)
{
    $logger->fatal_msg("test_gcps_are_unique failed");
    $return_status = ERROR; 
}

print "------ Running test_is_same_gcp test\n";
if (test_is_same_gcp() != SUCCESS)
{
    $logger->fatal_msg("test_is_same_gcp failed");
    $return_status = ERROR; 
}

print "------ Running test_is_compatible_gcp test\n";
if (test_is_compatible_gcp() != SUCCESS)
{
    $logger->fatal_msg("test_is_compatible_gcp failed");
    $return_status = ERROR; 
}

print "------ Running test_is_unique_image_chip test\n";
if (test_is_unique_image_chip() != SUCCESS)
{
    $logger->fatal_msg("test_is_unique_image_chip failed");
    $return_status = ERROR; 
}

print "------ Running test_verify_image_chip_data test\n";
if (test_verify_image_chip_data() != SUCCESS)
{
    $logger->fatal_msg("test_verify_image_chip_data failed");
    $return_status = ERROR; 
}

print "------ Running test_read_image_chip_files test\n";
if (test_read_image_chip_files() != SUCCESS)
{
    $logger->fatal_msg("test_read_image_chip_files failed");
    $return_status = ERROR; 
}

print "------ Running test_verify_tar_data test\n";
if (test_verify_tar_data() != SUCCESS)
{
    $logger->fatal_msg("test_verify_tar_data failed");
    $return_status = ERROR; 
}

# Cleanup
if (-e $ut_dir)
{
    if (system($cleanup) != SUCCESS)
    {
        $logger->fatal_msg("Running Cleanup");
        exit FAILURE;
    }
}

print "\n--------- Completed " . basename($0) . " test ------------------\n\n";

exit $return_status;

##############################################################################
# Name: initialize
#
# Description: Initializes test
#
# Parameters: N/A
#
# Returns: SUCCESS/ERROR
##############################################################################
sub initialize
{
    my @init_commands; # Init commands array
    
    # Get the default logger object from the library 
    my $logger = get_default_log();

    # Clean up directory first 
    if (-d "manage_gcps_data")
    {
        my $cleanup = "rm -rf ./manage_gcps_data";
        my $status = system($cleanup);
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Failed to cleanup before test");
            $logger->close_log();
            exit FAILURE;
        }
    }

    # Start of all init commands
    push @init_commands,{command => "mkdir manage_gcps_data"};;
    push @init_commands,{command => "cp data/manage_gcps_data".
        "/GCP* manage_gcps_data/"};
    push @init_commands,{command => "cp data/manage_gcps_data".
        "/132* manage_gcps_data/"};
    push @init_commands,{command => "cp data/manage_gcps_data".
        "/chips_* manage_gcps_data/"};
    push @init_commands,{command => "cp data/manage_gcps_data".
        "/image_chips* manage_gcps_data/"};

    # Run each command in order
    foreach my $command (@init_commands)
    {
        # Run Init to copy gcp directory
        my $status = system($command->{'command'});
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Failed to initialize $command->{command}");
            $logger->close_log();
            exit FAILURE;
        }
    }
}

##############################################################################
# Name: test_get_date_updated
#
# Description: Test the get_date_updated routine from the manage_gcps_utils 
#              module.
#
# Parameters: N/A
#
# Returns: SUCCESS/ERROR
##############################################################################
sub test_get_date_updated()
{
    # Return status for entire test routine
    my $return_status = SUCCESS;

    # Setup for current date
    my $time = localtime;
    my $date_today = sprintf("%02d-%02d-%4d", $time->mon + 1, $time->mday, 
        $time->year+1900);

    # Set up the arrays to loop through for the year 2014
    my @months = qw(01 02 03 04 05 06 07 08 09 10 11 12);
    my @days = qw(31 28 31 30 31 30 31 31 30 31 30 31);

    # Set up the correct and error dates to test
    my @correct_dates = qw(02-06-2015 02/06/2015 02-06/2015 02/06-2015);
    my @error_dates = qw(13-11-2012 12-32-2013 12-13-19 1-2-1 2-3-139
        3-4-1230 2/06/2015 02/6/2015 2/6/2015 2/29/2015);

    # Get the default logger object from the library 
    my $logger = get_default_log();

    print "-------- Running test to check passing no date returns todays "
        . "date\n";
    # Test that returns todays date in correct format
    my ($status, $return_date) = get_date_updated(undef);
    if ($status != SUCCESS || $return_date ne $date_today)
    {
        $logger->fatal_msg("Passing no date returned the wrong date:".
            "$return_date");
        $return_status = ERROR;
    }

    print "-------- Running test to check every date in the year 2014\n";
    # Loop through entire year checking every date
    for my $month (@months)
    {
        for my $day (1 .. $days[int $month - 1])
        {
        my $date = sprintf("%02d-%02d-%4d", $month, $day, "2014");
            ($status, $return_date) = get_date_updated($date);
            if ($status != SUCCESS)
            {
                $logger->fatal_msg("Passing $date returned the wrong date:");
                $return_status = ERROR;
            }
        }
    }

    print "-------- Running test to check correct dates\n";
    # Test some simple correct dates
    for my $date (@correct_dates)
    {
        ($status, $return_date) = get_date_updated($date);
        if ($status != SUCCESS || $return_date ne "02-06-2015")
        {
            $return_date = "undef" if (!defined $return_date);
            $logger->fatal_msg("Passing $date returned the wrong date:".
                "$return_date");
            $return_status = ERROR;
        }
    }

    print "-------- Running test to check date errors\n";

    # Test dates that should throw errors
    for my $date (@error_dates)
    {
        $logger->set_log_msg_level(IAS_LOG_LEVEL_DISABLE);

        ($status, $return_date) = get_date_updated($date);

        $logger->set_log_msg_level(IAS_LOG_LEVEL_INFO);

        if ($status == SUCCESS || defined $return_date)
        {
            $logger->fatal_msg("Passing no date returned the wrong date:".
                "$return_date");
            $return_status = ERROR;
        }
    }

    
    if ($return_status != SUCCESS)
    {
        $logger->info_msg("test_get_date_updated failed");
    }

    return $return_status;
}

##############################################################################
# Name: test_read_gcplib_file
#
# Description: Test the read_gcplib_file routine from the manage_gcps_utils 
#              module.
#
# Parameters: N/A
#
# Returns: SUCCESS/ERROR
##############################################################################
sub test_read_gcplib_file()
{
    my $return_status = SUCCESS; # Routine return status holder 
    my $test1 = $ut_dir."GCPLib";# GCPLib that reads successfully
    my $test2 = $ut_dir."GCPLib.empty"; # Should successfully read zero points

    # ID that are in the read GCPLib file
    my @ids = qw(380300012 380300027 380300034 380300036 380300066 380300139 
        380300358 380300361 380300370 380300386);
    
    # All the error test to run and wheter they should be run as master
    my @error_tests = ($ut_dir."GCPLib.inconsistent",
        $ut_dir."GCPLib.inconsistent2", $ut_dir."GCPLib.not.unique",
        $ut_dir."GCPLib.missing");
    my @master_list = (1, 1, 0, 1);

    # Expected version read from GCPLib file
    my $expected_version = 10;

    # Get the default logger object from the library 
    my $logger = get_default_log();

    # Run test 1 should complete successfully
    print "-------- Running test $test1\n";
    $logger->set_log_msg_level(IAS_LOG_LEVEL_DISABLE);
    my $gcps = read_gcplib_file($test1, 1);
    $logger->set_log_msg_level(IAS_LOG_LEVEL_INFO);
    if (!defined $gcps)
    {
        $logger->fatal_msg("Testing read_gcplib_file failed when".
            "it shouldn't have using $test1");
        $return_status = ERROR;
    }
    else
    {
        # Check data in the hash
        for my $record (@{$gcps->{'records'}})
        {
            my $id = $record->{'point_id'};
            if (!defined $gcps->{'lookup'}->{$id}[0]->{'point_id'})
            {
                $logger->fatal_msg("$id should exist in hash");
                $return_status = ERROR;
            }
        }
        # Verify the negative UTM zone handling worked
        my $record = $gcps->{'lookup'}->{'0380300027'}[0];
        if (!defined $record || $record->{'zone'} != 11
            || $record->{'projection_y'} ne '4776891.505427')
        {
            $logger->fatal_msg("Negative UTM zone handling did not work");
            $return_status = ERROR;
        }
        if ($gcps->{'version'} != $expected_version)
        {
            $logger->fatal_msg("Version passed was incorrect");
            $return_status = ERROR;
        }
        if ($gcps->{'file_type'} != 1)
        {
            $logger->fatal_msg("File type passed was incorrect");
            $return_status = ERROR;
        }
        if ($gcps->{'dir'} ne $ut_dir)
        {
            $logger->fatal_msg("Directory passed was incorrect");
            $return_status = ERROR;
        }
    }

    # Run test 2 which should complete successfully by reading zero GCPs
    $logger->set_log_msg_level(IAS_LOG_LEVEL_ERROR);
    $gcps = read_gcplib_file($test2, 1);
    $logger->set_log_msg_level(IAS_LOG_LEVEL_INFO);
    if (!defined $gcps)
    {
        $logger->fatal_msg("Test read_gcplib_file using $test2 failed"
                . " when it should have succeeded");
        $return_status = ERROR;
    }
    else
    {
        # Should have read zero records from an empty file
        my $records = $gcps->{'records'};
        if (scalar(@$records) != 0)
        {
            $logger->fatal_msg("Test read_gcplib_file using $test2 did not"
                    . " read zero records as expected");
            $return_status = ERROR;
        }
    }

    # Run the GCPLib files that throw errors
    my $index = 0;
    for my $test (@error_tests)
    {
        print "-------- Running test $test\n";
        $logger->set_log_msg_level(IAS_LOG_LEVEL_DISABLE);
        my $gcps = read_gcplib_file($test, $master_list[$index]);
        $logger->set_log_msg_level(IAS_LOG_LEVEL_INFO);
        if (defined $gcps)
        {
            $logger->fatal_msg("Test read_gcplib_file using $test succeeded"
                . " when it should have failed");
            $return_status = ERROR;
        }
        $index++;
    }
    
    # Print Success if test routine is a success
    if ($return_status != SUCCESS)
    {
        print "\n";
        $logger->info_msg("test_read_gcplib_file failed");
    }

    return $return_status;
}

##############################################################################
# Name: test_create_record_index
#
# Description: Test the create_record_index routine from the manage_gcps_utils
#              module.
#
# Parameters: N/A
#
# Returns: SUCCESS/ERROR
##############################################################################
sub test_create_record_index()
{
    my $return_status = SUCCESS; # Routine return status
    
    # Only one test need to verify data no errors returned
    my $test1 = $ut_dir."GCPLib.hash";
    my @ids = qw(380300012 380300027 380300034 380300036 380300066 380300139 
        380300358 380300361 380300370 380300386);

    # Get the default logger object from the library 
    my $logger = get_default_log();

    # Read in the GCPLib file
    my ($status, $gcp_version, @gcps) = ias_gcp_read_gcplib($test1);
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Could not read gcplib file $test1");
        return ERROR;
    }

    # Create the Record Hash
    my $record_hash = create_record_index(\@gcps);
    
    # Loop through the hash and verify some of the data
    for my $record (@gcps)
    {
        my $id = $record->{'point_id'};
        if (!defined $record_hash->{$id}[0]->{'point_id'})
        {
            $logger->fatal_msg("$id should exist in hash");
            $return_status = ERROR;
        }
        else
        {
            my $id2 = $record_hash->{$id}[0]->{'point_id'};
            if ($id ne $id2)
            {
                $logger->fatal_msg("$id should match $id2");
                $return_status = ERROR;
            }
        }
    }
    
    # Write out error if routine was not successful
    if ($return_status != SUCCESS)
    {
        print "\n";
        $logger->info_msg("test_create_record_index failed");
    }

    return $return_status;
}

##############################################################################
# Name: test_verify_gcps_are_consistent
#
# Description: Test the verify_gcps_are_consistent routine from the
#              manage_gcps_utils module.
#
# Parameters: N/A
#
# Returns: SUCCESS/ERROR
##############################################################################
sub test_verify_gcps_are_consistent()
{
    my $return_status = SUCCESS;
        # Only one test need to verify data no errors returned
    my @tests;
    
    # Create array of hashes for testing
    push @tests, {gcplib => "GCPLib.consise1", master => 0, success => FALSE,
        desc => "Checking that all paths are the same"};
    push @tests, {gcplib => "GCPLib.consise2", master => 1, success => FALSE,
        desc => "Checking that path same as path in ID"};
    push @tests, {gcplib => "GCPLib.consise3", master => 1, success => FALSE,
        desc => "Checking that row same as row in ID"};
    push @tests, {gcplib => "GCPLib.consise4", master => 1, success => FALSE,
        desc => "Checking that ID is correct format"};
    push @tests, {gcplib => "GCPLib.consise5", master => 1, success => FALSE,
        desc => "Checking that chip name contains GCP ID"};
    push @tests, {gcplib => "GCPLib.consise6", master => 1, success => FALSE,
        desc => "Checking that chip name is correct format"};
    push @tests, {gcplib => "GCPLib.consise7", master => 0, success => FALSE,
        desc => "Checking that gcps metadata is consistent "};
    push @tests, {gcplib => "GCPLib.consise8", master => 0, success => FALSE,
        desc => "Checking that chip metadata is consistent "};
    push @tests, {gcplib => "GCPLib", master => 1, success => TRUE,
        desc => "Checking that all consistency check work"};
    push @tests, {gcplib => "GCPLib.repeated_chip_ids", master => 1,
        success => FALSE,
        desc => "Checking that repeated image chips are found"};

    #Get the default logger object from the library 
    my $logger = get_default_log();

    for my $test (@tests)
    {
        print "-------- $test->{'desc'}\n";

        $logger->set_log_msg_level(IAS_LOG_LEVEL_WARN);

        # Read in the GCPLib file
        my $gcplib_file = $ut_dir.$test->{'gcplib'};
        my ($status, $gcp_version, @gcps) = ias_gcp_read_gcplib($gcplib_file);
        $logger->set_log_msg_level(IAS_LOG_LEVEL_INFO);
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Could not read gcplib file $gcplib_file");
            $return_status = ERROR;
            next;
        }

        $logger->set_log_msg_level(IAS_LOG_LEVEL_DISABLE)
            if ($test->{'success'} == FALSE);

        # Create the Record Hash
        my $record_hash = create_record_index(\@gcps);

        if(verify_gcps_are_consistent($record_hash, $test->{'master'}) 
            != $test->{'success'})
        {
            $logger->set_log_msg_level(IAS_LOG_LEVEL_INFO);
            $logger->fatal_msg("Test failed to produce ".
                "the correct outcome of ".strfy_boolean($test->{'success'}, 1));
            $return_status = ERROR;
        }
        $logger->set_log_msg_level(IAS_LOG_LEVEL_INFO);
    }

    # Write out error if routine was not successful
    if ($return_status != SUCCESS)
    {
        print "\n";
        $logger->info_msg("test_verify_gcps_are_consistent failed");
    }

    return $return_status;
}

##############################################################################
# Name: test_gcps_are_unique
#
# Description: Test the gcps_are_unique routine from the manage_gcps_utils
#              module
#
# Parameters: N/A
#
# Returns: SUCCESS/ERROR
##############################################################################
sub test_gcps_are_unique()
{
    my $return_status = SUCCESS;
        # Only one test need to verify data no errors returned
    my @tests;
    
    # Create array of hashes for testing
    push @tests, {gcplib => "GCPLib.unqe1", master => 0, success => FALSE,
        desc => "Checking that different gcps that are actually the same"};
    push @tests, {gcplib => "GCPLib", master => 1, success => TRUE,
        desc => "Checking that all unique gcp checks work"};

    #Get the default logger object from the library 
    my $logger = get_default_log();

    for my $test (@tests)
    {
        print "-------- $test->{'desc'}\n";

        $logger->set_log_msg_level(IAS_LOG_LEVEL_WARN);

        # Read in the GCPLib file
        my $gcplib_file = $ut_dir.$test->{'gcplib'};
        my ($status, $gcp_version, @gcps) = ias_gcp_read_gcplib($gcplib_file);
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Could not read gcplib file $gcplib_file");
            $return_status = ERROR;
            next;
        }

        # Create the Record Hash
        my $record_hash = create_record_index(\@gcps);

        $logger->set_log_msg_level(IAS_LOG_LEVEL_DISABLE)
            if ($test->{'success'} == FALSE);

        # Test the routine gcps_are_unique
        if(gcps_are_unique($record_hash) != $test->{'success'})
        {
            $logger->set_log_msg_level(IAS_LOG_LEVEL_INFO);
            $logger->fatal_msg("Test failed to produce ".
                "the correct outcome of ".strfy_boolean($test->{'success'}, 1));
            $return_status = ERROR;
        }
        $logger->set_log_msg_level(IAS_LOG_LEVEL_INFO);
    }

    # Write out error if routine was not successful
    if ($return_status != SUCCESS)
    {
        print "\n";
        $logger->info_msg("test_test_gcps_are_unique failed");
    }

    return $return_status;
}

##############################################################################
# Name: test_is_same_gcp
#
# Description: Test the is_same_gcp routine from the manage_gcps_utils module
#
# Parameters: N/A
#
# Returns: SUCCESS/ERROR
##############################################################################
sub test_is_same_gcp()
{
    my $return_status = SUCCESS;
        # Only one test need to verify data no errors returned
    my @tests;
    
    # Create array of hashes for testing
    push @tests, {gcplib => "GCPLib.sme1", gcp1 => 0, gcp2 => 1,
        success => FALSE, desc => "Checking that gcps are different-numeric"};
    push @tests, {gcplib => "GCPLib.sme2", gcp1 => 0, gcp2 => 1,
        success => FALSE, desc => "Checking that gcps are different-string"};
    push @tests, {gcplib => "GCPLib.same", gcp1 => 0, gcp2 => 1,
        success => TRUE, desc => "Checking that gcps are the same"};;

    #Get the default logger object from the library 
    my $logger = get_default_log();

    for my $test (@tests)
    {
        print "-------- $test->{'desc'}\n";

        # Read in the GCPLib file
        my $gcplib_file = $ut_dir.$test->{'gcplib'};
        my ($status, $gcp_version, @gcps) = ias_gcp_read_gcplib($gcplib_file);
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Could not read gcplib file $gcplib_file");
            $return_status = ERROR;
            next;
        }

        # Grab the gcps to test
        my $gcp1 = $gcps[$test->{'gcp1'}];
        my $gcp2 = $gcps[$test->{'gcp2'}];

        # Test the routine gcps_are_unique
        $logger->set_log_msg_level(IAS_LOG_LEVEL_ERROR);
        if(is_same_gcp($gcp1, $gcp2) != $test->{'success'})
        {
            $logger->set_log_msg_level(IAS_LOG_LEVEL_INFO);
            $logger->fatal_msg("Test failed to produce ".
                "the correct outcome of ".strfy_boolean($test->{'success'}, 1));
            $return_status = ERROR;
        }
        $logger->set_log_msg_level(IAS_LOG_LEVEL_INFO);
    }

    # Write out error if routine was not successful
    if ($return_status != SUCCESS)
    {
        print "\n";
        $logger->info_msg("test_is_same_gcp failed");
    }

    return $return_status;
}

##############################################################################
# Name: test_is_compatible_gcp
#
# Description: Test the is_compatible_gcp routine in the manage_gcps_utils
#              module
#
# Parameters: N/A
#
# Returns: SUCCESS/ERROR
##############################################################################
sub test_is_compatible_gcp()
{
    my $return_status = SUCCESS;
        # Only one test need to verify data no errors returned
    my @tests;
    
    # Create array of hashes for testing
    push @tests, {gcplib => "GCPLib.compe1", gcp1 => 0, gcp2 => 1,
        success => FALSE, desc => "Checking that gcps are different-numeric"};
    push @tests, {gcplib => "GCPLib.compe2", gcp1 => 0, gcp2 => 1,
        success => FALSE, desc => "Checking that gcps are different-string"};
    push @tests, {gcplib => "GCPLib.comp", gcp1 => 0, gcp2 => 1,
        success => TRUE, desc => "Checking that gcps are the same"};;

    #Get the default logger object from the library 
    my $logger = get_default_log();

    for my $test (@tests)
    {
        print "-------- $test->{'desc'}\n";

        # Read in the GCPLib file
        my $gcplib_file = $ut_dir.$test->{'gcplib'};
        my ($status, $gcp_version, @gcps) = ias_gcp_read_gcplib($gcplib_file);
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Could not read gcplib file $gcplib_file");
            $return_status = ERROR;
            next;
        }

        # Grab the gcps to test
        my $gcp1 = $gcps[$test->{'gcp1'}];
        my $gcp2 = $gcps[$test->{'gcp2'}];

        # Test the routine gcps_are_unique
        $logger->set_log_msg_level(IAS_LOG_LEVEL_ERROR);
        if(is_same_gcp($gcp1, $gcp2) != $test->{'success'})
        {
            $logger->set_log_msg_level(IAS_LOG_LEVEL_INFO);
            $logger->fatal_msg("Test failed to produce ".
                "the correct outcome of ".strfy_boolean($test->{'success'}, 1));
            $return_status = ERROR;
        }
        $logger->set_log_msg_level(IAS_LOG_LEVEL_INFO);
    }

    # Write out error if routine was not successful
    if ($return_status != SUCCESS)
    {
        print "\n";
        $logger->info_msg("test_is_compatible_gcp failed");
    }

    return $return_status;
}

##############################################################################
# Name: test_is_unique_image_chip
#
# Description: Test the is_unique_image_chip routine in the manage_gcps_utils
#              module. 
#
# Parameters: N/A
#
# Returns: SUCCESS/ERROR
##############################################################################
sub test_is_unique_image_chip()
{
    my $return_status = SUCCESS;
        # Only one test need to verify data no errors returned
    my @tests;
    
    # Create array of hashes for testing
    push @tests, {gcplib => "GCPLib.imgunqe1", gcp1 => 0, gcp2 => 1,
        success => TRUE, desc => "Checking that gcps are different-numeric"};
    push @tests, {gcplib => "GCPLib.imgunqe2", gcp1 => 0, gcp2 => 1,
        success => TRUE, desc => "Checking that gcps are different-string"};
    push @tests, {gcplib => "GCPLib.imgunq", gcp1 => 0, gcp2 => 1,
        success => FALSE, desc => "Checking that gcps are the same"};

    #Get the default logger object from the library 
    my $logger = get_default_log();

    for my $test (@tests)
    {
        print "-------- $test->{'desc'}\n";

        # Read in the GCPLib file
        my $gcplib_file = $ut_dir.$test->{'gcplib'};
        my ($status, $gcp_version, @gcps) = ias_gcp_read_gcplib($gcplib_file);
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Could not read gcplib file $gcplib_file");
            $return_status = ERROR;
            next;
        }
        
        # Grab the gcps to test
        my $gcp1 = $gcps[$test->{'gcp1'}];
        my $gcp2 = $gcps[$test->{'gcp2'}];

        # Test the routine gcps_are_unique
        $logger->set_log_msg_level(IAS_LOG_LEVEL_ERROR);
        if(is_unique_image_chip($gcp1, $gcp2) != $test->{'success'})
        {
            $logger->set_log_msg_level(IAS_LOG_LEVEL_INFO);
            $logger->fatal_msg("Test failed to produce ".
                "the correct outcome of ".strfy_boolean($test->{'success'}, 1));
            $return_status = ERROR;
        }
        $logger->set_log_msg_level(IAS_LOG_LEVEL_INFO);
    }

    # Write out error if routine was not successful
    if ($return_status != SUCCESS)
    {
        print "\n";
        $logger->info_msg("test_is_unique_image_chip failed");
    }

    return $return_status;
}

##############################################################################
# Name: test_verify_image_chip_data
#
# Description: Test that the verify_image_chip_data routine works correctly
#
# Parameters: N/A
#
# Returns: SUCCESS/ERROR
##############################################################################
sub test_verify_image_chip_data()
{
    my $return_status = SUCCESS;
    my $chip_file = $ut_dir."1320500012_01";
    my $chip = {};
    $chip->{'chip_size_lines'} = 64;
    $chip->{'chip_size_samples'} = 64;
    $chip->{'chip_data_type'} = "BYTE";

    #Get the default logger object from the library 
    my $logger = get_default_log();

    # Check the normal case which should pass
    if (verify_image_chip_data($chip, $chip_file) != SUCCESS)
    {
        $return_status = ERROR;
    }
    
    # Change the data type to I2 should throw an error
    $chip->{'chip_data_type'} = "I2";
    $logger->set_log_msg_level(IAS_LOG_LEVEL_DISABLE);
    if (verify_image_chip_data($chip, $chip_file) != ERROR)
    {
        $return_status = ERROR;
    }
    $logger->set_log_msg_level(IAS_LOG_LEVEL_INFO);

    # Write out error if routine was not successful
    if ($return_status != SUCCESS)
    {
        print "\n";
        $logger->info_msg("test_verify_image_chip_data failed");
    }
    
    return $return_status;
}

##############################################################################
# Name: test_read_image_chip_files
#
# Description: Test the read_image_chip_files routine from the 
#              manage_gcps_utils module
#
# Parameters: N/A
#
# Returns: SUCCESS/ERROR
##############################################################################
sub test_read_image_chip_files()
{
    my $return_status = SUCCESS;
    my @tests;
    
    # Create array of hashes for testing
    push @tests, {gcplib => "GCPLib.gcp_hash_e1", success => ERROR,
        desc => "Checking that bad data type does not work"};
    push @tests, {gcplib => "GCPLib.gcp_hash_e2", success => ERROR,
        desc => "Checking that non_exisistant chip throws error"};
    push @tests, {gcplib => "GCPLib.gcp_hash_e3", success => ERROR,
        desc => "Checking that wrong chip size throws error"};
    push @tests, {gcplib => "GCPLib.gcp_hash", master => 1, success => SUCCESS,
        desc => "Checking that read_image_chip_files run successfully"};

    #Get the default logger object from the library 
    my $logger = get_default_log();

    for my $test (@tests)
    {
        print "-------- $test->{'desc'}\n";

        # Read in the GCPLib file
        my $gcplib_file = $ut_dir.$test->{'gcplib'};
        my $master_gcplib = read_gcplib_file($gcplib_file, MASTER_FILE);
        if (!defined $master_gcplib)
        {
            $logger->fatal_msg("Could not read gcplib file $gcplib_file");
            $return_status = ERROR;
            next;
        }

        # Create the Record Hash
        my $records = $master_gcplib->{'records'};

        $logger->set_log_msg_level(IAS_LOG_LEVEL_DISABLE)
            if ($test->{'success'} == ERROR);

        # Test the routine
        my $status = read_image_chip_files($records,
            $ut_dir);
        if ($status != $test->{'success'})
        {
            $logger->fatal_msg("Test failed to produce ".
                "the correct outcome of ".strfy_boolean($test->{'success'}, 1));
            $return_status = ERROR;
        }
        $logger->set_log_msg_level(IAS_LOG_LEVEL_INFO);
       
        # Verify that the image chip data is present
        if ($test->{'gcplib'} eq "GCPLib.gcp_hash")
        {
            foreach my $chip (@{$records})
            {
                my $filename = $chip->{'chip_name'};
                if (!exists $chip->{'image_chip'})
                {
                    $logger->fatal_msg("Data does not exist");
                    return ERROR;   
                }
            }
        }
    }

    # Write out error if routine was not successful
    if ($return_status != SUCCESS)
    {
        print "\n";
        $logger->info_msg("test_read_image_chip_files failed");
    }

    return $return_status;
}

##############################################################################
# Name: test_verify_tar_data
#
# Description: Test the verify_tar_data routine from the 
#              manage_gcps_utils module
#
# Parameters: N/A
#
# Returns: SUCCESS/ERROR
##############################################################################
sub test_verify_tar_data()
{
    my $return_status = SUCCESS;
        # Only one test need to verify data no errors returned
    my @tests;
    
    # Create array of hashes for testing
    push @tests, {gcplib => "GCPLib.vtar", success => SUCCESS,
        desc => "Doing all checks for verify_tar_data in single test"};

    #Get the default logger object from the library 
    my $logger = get_default_log();

    for my $test (@tests)
    {
        print "-------- $test->{'desc'}\n";

        # Read in the GCPLib file
        my $gcplib_file = $ut_dir.$test->{'gcplib'};
        my $master_gcplib = read_gcplib_file($gcplib_file, MASTER_FILE);
        if (!defined $master_gcplib)
        {
            $logger->fatal_msg("Could not read gcplib file $gcplib_file");
            $return_status = ERROR;
            next;
        }

        # Create the Record Hash
        my $records = $master_gcplib->{'records'};

        # Prepare the GCP Hash
        my $status = read_image_chip_files($records, $ut_dir);
        if ($status != $test->{'success'})
        {
            $logger->fatal_msg("Test failed to produce ".
                "the correct outcome of ".strfy_boolean($test->{'success'}, 1));
            $return_status = ERROR;
        }
        
        # Tar up chips to be tested
        my $tar_file = "image_chips.test.tbz";
        $status = ias_gcp_tar_image_chips($records, $ut_dir . $tar_file,
            IAS_GCP_BZIP2, IAS_GCP_DO_MD5SUM);
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Error taring up image_chips");
            return ERROR;
        }

        # Verify normal case works
        $status = verify_tar_data($records, $ut_dir.$tar_file);
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Verifying normal case of verify_tar_data");
            $return_status = ERROR;   
        }

        $logger->set_log_msg_level(IAS_LOG_LEVEL_DISABLE);

        # Verify bad hash key for tar file is caught
        my $old_md5 = @{$records}[0]->{'MD5'};
        @{$records}[0]->{'MD5'} = "0000000000000000";
        $status = verify_tar_data($records,  $ut_dir.$tar_file);
        if ( $status == SUCCESS)
        {
            $logger->set_log_msg_level(IAS_LOG_LEVEL_INFO);
            $logger->fatal_msg("Verifying bad hash key in verify_tar_data");
            $return_status = ERROR;   
        };
        @{$records}[0]->{'MD5'} = $old_md5;
         
        # Verify not all keys were checked in hash
        my $bad_chip = {chip_name => '1320500013_01', MD5 => 000000000000000};
        push @{$records}, $bad_chip;
        $status = verify_tar_data($records,  $ut_dir.$tar_file);
        if ( $status == SUCCESS)
        {
            $logger->set_log_msg_level(IAS_LOG_LEVEL_INFO);
            $logger->fatal_msg("Verifying missing chip in tarball");
            $return_status = ERROR;      
        }
        pop @{$records};

        # Verify missing chip in tar file is caught
        delete @{$records}[0]->{'MD5'};
        $status = verify_tar_data($records,  $ut_dir.$tar_file);
        if ($status == SUCCESS)
        {
            $logger->set_log_msg_level(IAS_LOG_LEVEL_INFO);
            $logger->fatal_msg("Verifying missing key");
            $return_status = ERROR;
        }
        $logger->set_log_msg_level(IAS_LOG_LEVEL_INFO);
    }

    # Write out error if routine was not successful
    if ($return_status != SUCCESS)
    {
        print "\n";
        $logger->info_msg("test_verify_tar_data failed");
    }

    return $return_status;
}

##############################################################################
# Name: strfy_boolean
#
# Description: Will turn success/error and true/false into strings
#
# Parameters:
#   - Value = SUCCESS - ERROR - TRUE - FALSE
#   - Type = 1 - TRUE/FALSE 0- SUCCESS/ERROR
#
# Returns: String version of ERROR, SUCCESS, TRUE, or FALSE
##############################################################################
sub strfy_boolean($$)
{
    my $value = shift;
    my $type = shift;
   
    # If type == 1 then stringify TRUE/FALSE 
    if ($type == 1)
    {
        return "TRUE" if ($value == TRUE);
        return "FALSE" if ($value == FALSE);
    }
    # Else stringify SUCCESS/ERROR
    else
    {
        return "SUCCESS" if ($value == SUCCESS);
        return "ERROR" if ($value == ERROR);
    }

    # Unknown value return undef as string
    return "undef";
}
