#!/usr/bin/perl -w
#############################################################################
# Name: test_set_gcp_status 
#
# Description:  Testing of the perl script set_gcp_status.
#
###########################################################################
use strict;
use File::Basename;         # Basename routine
use Logger;                 # Logging module
use ias_const;              # Define SUCCESS and ERROR

# Created the formatted date for today
my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime();
my $formatted_date = sprintf("%02d-%02d-%04d", $mon+1, $mday, $year+1900);
my $logfile = "test_set_gcp_status.log";

# Create arrays for Test, Verification, and Test Descriptions
my @all_tests;

# Where to copy the entire GCP test directory
my $ut_dir = "./data/";
my $gcplib_directory = $ut_dir."status_directory";

# Set up the Cleanup and Initialize commands
my $cleanup = "rm -rf ./status_directory";

# Remove the log file from a previous run
`rm -f $logfile`;

# Set up the logger
my $logger = Logger->new("STDOUT");
if (!$logger)
{
    print STDERR ": ERROR:  Failed to establish logger object for "
        . "STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name("test_set_gcp_status");
set_default_log($logger);

print "\n--------- Starting " . basename($0) . " test ------------------\n\n";

# Initialize tests
initialize();

# Setup Description, Test, and Verification for all tests
# Read descriptions for test information
push @all_tests, {desc => "Testing that the --test option does not".
    " change librarys",
    test => "../set_gcp_status.pl status_directory/status_ids.txt". 
    " status_directory --deactivate --test",
    diff => "./status_directory/GCPLib.normal",
    success => SUCCESS};

push @all_tests, {desc => "Testing that deactivating of IDs works",
    test => "../set_gcp_status.pl status_directory/status_ids.txt".
    " status_directory --deactivate",
    diff => "./status_directory/GCPLib.deactivated",
    success => SUCCESS};

push @all_tests, {desc => "Testing that activating of IDS works ",
    test => "../set_gcp_status.pl status_directory/status_ids.txt".
    " status_directory --activate",
    diff => "./status_directory/GCPLib.activated",
    success => SUCCESS};

push @all_tests, {desc => "Testing that invalid chip id is caught",
    test =>  "../set_gcp_status.pl status_directory/status_ids_error1.txt".
    " status_directory --deactivate",
    diff => "./status_directory/GCPLib.activated",
    success => ERROR};

push @all_tests, {desc => "Testing that invalid gcp id is caught",
    test =>  "../set_gcp_status.pl status_directory/status_ids_error2.txt". 
    " status_directory --deactivate ",
    diff => "./status_directory/GCPLib.activated",
    success => ERROR};

push @all_tests, {desc => "Testing that missing GCPLib file is caught",
    test =>  "../set_gcp_status.pl status_directory/status_ids_error3.txt".
    " status_directory --deactivate ",
    diff => "./status_directory/GCPLib.activated",
    success => ERROR};

# Iterate through all of the tests
foreach my $test (@all_tests)
{
    # Print out the test description
    print STDOUT "------ $test->{'desc'}\n";

    # Run the test
    my $status = system($test->{'test'}." >>$logfile");
    if ($status == SUCCESS && $test->{'success'} != SUCCESS)
    {
        $logger->fatal_msg("Failed to run test $test->{'test'}");
    }
    elsif ($status != SUCCESS && $test->{'success'} != ERROR)
    {
        $logger->fatal_msg("Failed to run test $test->{'test'}");
    }

    # Replace current date with 03-27-2014 to make verifying easier
    my @replace = ("sed", "-i", "s|$formatted_date|03-27-2014|g", 
        "status_directory/132/050/GCPLib");
    $status = system(@replace);
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Failed to replace date");
        $logger->close_log();
        exit FAILURE;
    }

    # Run diff on provide file run quietly 
    $status = system("diff status_directory/132/050/GCPLib ".$test->{'diff'}.
        " -q");
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Failed to verify test $test->{'diff'}".
            " for $test->{'test'}");
        $logger->close_log();
        exit FAILURE;
    }
}

# Run cleanup after testing
if (-d "status_directory")
{
    my $status = system($cleanup);
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Failed to initialize test");
        $logger->close_log();
        exit FAILURE;
    }
}

print "\n--------- Completed " . basename($0) . " test ------------------\n\n";

exit SUCCESS;

################################################################################
# Name: initialize
#
# Purpose: Initialize all test directories and files
# ##############################################################################
sub initialize
{
    my @init_commands; # Init commands array

    # Clean up directory first 
    if (-d "status_directory")
    {
        my $cleanup = "rm -rf ./status_directory";
        my $status = system($cleanup);
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Failed to cleanup before test");
            $logger->close_log();
            exit FAILURE;
        }
    }

    # Start of all init commands
    push @init_commands,{command => "mkdir status_directory"};
    push @init_commands,{command => "mkdir status_directory/132"};
    push @init_commands,{command => "mkdir status_directory/132/050"};
    push @init_commands,{command => "cp data/status_directory/status_*".
        " status_directory/"};
    push @init_commands,{command => "cp data/status_directory/GCP*".
        " status_directory/"};
    push @init_commands,{command => "cp data/status_directory/132/050/".
        "GCPLib status_directory/132/050/"};

    # Run each command in order
    foreach my $command (@init_commands)
    {
        # Run Init to copy gcp directory
        my $status = system($command->{'command'});
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Failed to initialize $command->{command}");
            $logger->close_log();
            exit FAILURE;
        }
    }
}
