#!/usr/bin/perl -w
#############################################################################
# Name: test_tar_image_chips.pl
#
# Description:  Unit testing for the perl script tar_image_chips.pl
#
###########################################################################
use strict;
use File::Basename;         # Basename routine
use Logger;                 # Logging module
use ias_const;              # Define SUCCESS and ERROR

# Create arrays for Test, Verification, and Test Descriptions
my @all_tests;

# Where to copy the entire GCP test directory
my $ut_dir = "./data/";
my $gcplib_directory = $ut_dir."tar_image_chips_directory";
my $logfile = "test_tar_image_chips.log";

# Remove the log file from a previous run
`rm -f $logfile`;

# Set up the logger
my $logger = Logger->new("STDOUT");
if (!$logger)
{
    print STDERR ": ERROR:  Failed to establish logger object for "
        . "STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name("test_tar_image_chips.pl");
set_default_log($logger);

print "\n--------- Starting " . basename($0) . " test ------------------\n\n";

# Set up the Cleanup and Initialize commands
my $cleanup = "rm -rf ./tar_image_chips_directory";
my $single_init = "cp tar_image_chips_directory/copies/chips_updated ".
    "tar_image_chips_directory/";

# Status holder 
my $status; 

# Initialize test
initialize();

# Setup tests
setup_tests(\@all_tests);

my $test_return = SUCCESS;
# Iterate through all of the tests
foreach my $test (@all_tests)
{
    # Print out the test description
    print "------ $test->{'desc'}\n";
    
    # Run the test
    $status = system("../tar_image_chips.pl ".$test->{'test'}." >>$logfile");
    if ($test->{'success'} == ERROR && $status == SUCCESS)
    {
        $logger->fatal_msg("Failed to run test $test->{'test'}");
        $test_return = ERROR;
    }
    elsif($test->{'success'} == SUCCESS && $status != SUCCESS)
    {        
        $logger->fatal_msg("Failed to run test $test->{'test'}");
        $test_return = ERROR;
    }

    # Ensure that the tarball exists
    if (exists $test->{'tar_file'})
    {
        foreach my $tar_file (@{$test->{'tar_file'}})
        {
            if (! -e "tar_image_chips_directory/".$tar_file)
            {
                $logger->fatal_msg("Tarball $tar_file does not exist");
                $test_return = ERROR;
            }
        }       
    }
    
    # Run init for next test
    $status = system($single_init);
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Could not initialize by doing $single_init");
        exit FAILURE;
    }

    # Run md5sum on provide file run quietly 
    $status = system("md5sum -c ".$test->{'md5sum'}." --quiet");
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Failed to verify test $test->{'md5sum'} for ".
            "$test->{'test'}");
        $test_return = ERROR;
    }
}

if ($test_return != SUCCESS)
{
    $logger->close_log();
    exit FAILURE;
}

# Run cleanup after testing
if (-d "tar_image_chips_directory")
{
    $status = system($cleanup);
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Failed to clean up after test");
        $logger->close_log();
        exit FAILURE;
    }
}

print "\n--------- Completed " . basename($0) . " test ------------------\n\n";

exit SUCCESS;

################################################################################
# Name: setup_tests
#
# Purpose: Pushes all the tests onto all_tests array
#
# Parameters:
#   - Test Array reference
#
# Returns: VOID
# ##############################################################################
sub setup_tests
{
    my $all_tests = shift;

    # Setup Description, Test, and Verification for all tests
    # Read descriptions for test information
    push @{$all_tests}, {
        desc => "Testing that error thrown for no compression value",
        test => "tar_image_chips_directory/GCPLib --test", 
        md5sum => "tar_image_chips_directory/md5/normal.out",
        success => ERROR};
    push @{$all_tests}, {
        desc => "Testing that error thrown for bad compression value",
        test => "tar_image_chips_directory/GCPLib --test".
        " --compression=RZIP", 
        md5sum => "tar_image_chips_directory/md5/normal.out",
        success => ERROR};
    push @{$all_tests}, {
        desc => "Testing that error thrown for bad verify value",
        test => "tar_image_chips_directory/GCPLib --test".
        " --compression=BZIP2 --verify=NOT", 
        md5sum => "tar_image_chips_directory/md5/normal.out",
        success => ERROR};
    push @{$all_tests}, {
        desc => "Testing the option --test using BZIP2",
        test => "tar_image_chips_directory/GCPLib --test ".
        "--compression=BZIP2",
        ,tar_file => ["image_chips.test.tbz"], 
        md5sum => "tar_image_chips_directory/md5/normal.out",
        success => SUCCESS};
    push @{$all_tests}, {
        desc => "Testing the option --test using GZIP",
        test => "tar_image_chips_directory/GCPLib --test ".
        "--compression=GZIP", 
        tar_file => ["image_chips.test.tgz"],
        md5sum => "tar_image_chips_directory/md5/normal.out",
        success => SUCCESS};
   push @{$all_tests}, {
        desc => "Testing using GZIP without --test",
        test => "tar_image_chips_directory/GCPLib --compression=GZIP", 
        tar_file => ["image_chips.tgz"],
        md5sum => "tar_image_chips_directory/md5/normal.out",
        success => SUCCESS};
   push @{$all_tests}, {
        desc => "Testing using BZIP2 without --test",
        test => "tar_image_chips_directory/GCPLib --compression=BZIP2", 
        tar_file => ["image_chips.tbz"],
        md5sum => "tar_image_chips_directory/md5/normal.out",
        success => SUCCESS};
   push @{$all_tests}, {
        desc => "Testing using BZIP2 with --verify_tar=NO",
        test => "tar_image_chips_directory/GCPLib --compression=BZIP2".
        " --verify_tar=NO", 
        tar_file => ["image_chips.tbz", "image_chips.tbz.bak"],
        md5sum => "tar_image_chips_directory/md5/normal.out",
        success => SUCCESS};
   push @{$all_tests}, {
        desc => "Testing using BZIP2 with no compression",
        test => "tar_image_chips_directory/GCPLib --compression=NO_COMPRESSION".
        " --verify_tar=NO", 
        tar_file => ["image_chips.tar"],
        md5sum => "tar_image_chips_directory/md5/normal.out",
        success => SUCCESS};
}

################################################################################
# Name: initialize
#
# Purpose: Initialize all test directorys and files
# ##############################################################################
sub initialize
{
    my @init_commands; # Init commands array
    
    # Clean up directory first 
    if (-d "tar_image_chips_directory")
    {
        my $cleanup = "rm -rf ./tar_image_chips_directory";
        my $status = system($cleanup);
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Failed to cleanup before test");
            $logger->close_log();
            exit FAILURE;
        }
    }

    # Start of all init commands
    push @init_commands,{command => "mkdir tar_image_chips_directory"};
    push @init_commands,{command => "mkdir tar_image_chips_directory/copies"};
    push @init_commands,{command => "mkdir tar_image_chips_directory/md5"};
    push @init_commands,{command => "cp data/tar_image_chips_directory".
        "/GCP* tar_image_chips_directory/"};
    push @init_commands,{command => "cp data/tar_image_chips_directory/".
        "/132* tar_image_chips_directory/"};
    push @init_commands,{command => "cp data/tar_image_chips_directory/".
        "/copies/chips_updated tar_image_chips_directory/copies/"};   
    push @init_commands,{command => "cp data/tar_image_chips_directory/".
        "/copies/image_chips.tbz tar_image_chips_directory/copies/"};    
    push @init_commands,{command => "cp data/tar_image_chips_directory/".
        "/md5/normal.out tar_image_chips_directory/md5/"};

    # Run each command in order
    foreach my $command (@init_commands)
    {
        # Run Init to copy gcp directory
        my $status = system($command->{'command'});
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Failed to initialize $command->{command}");
            $logger->close_log();
            exit FAILURE;
        }
    }
}
