#!/usr/bin/perl -w
#############################################################################
# Name: test_update_gcps.pl
#
# Description:  Unit testing for the perl script update_gcps.pl
#
###########################################################################
use strict;
use File::Basename;         # Basename routine
use Logger;                 # Logging module
use ias_const;              # define SUCCESS and ERROR

# Created the formatted date for today
my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime();
my $formatted_date = sprintf("%02d-%02d-%04d", $mon+1, $mday, $year+1900);
my $logfile = "test_update_gcps.log";

# Create arrays for Test, Verification, and Test Descriptions
my @all_tests;

# Where to copy the entire GCP test directory
my $ut_dir = "./data/";
my $gcplib_directory = $ut_dir."update_gcps_directory";

# Remove the log file from a previous run
`rm -f $logfile`;

# Set up the Cleanup and Initialize commands
my $cleanup = "rm -rf ./update_gcps_directory";

# Set up the logger
my $logger = Logger->new("STDOUT");
if (!$logger)
{
    print STDERR ": ERROR:  Failed to establish logger object for "
        . "STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name("test_update_gcps_status");
set_default_log($logger);

print "\n--------- Starting " . basename($0) . " test ------------------\n\n";

# Initialize test
initialize();

# Setup tests
setup_tests(\@all_tests);

my $test_return = SUCCESS;
# Iterate through all of the tests
foreach my $test (@all_tests)
{
    # Print out the test description
    print "------$test->{'desc'}\n";

    if (defined $test->{'reset_lib'})
    {
        initialize();
    }

    # Run addtional command if needed
    if (defined $test->{'additional_command'})
    {
        my $status = system($test->{'additional_command'});
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Addtional command failed".
                "$test->{'addtional_command'} for $test->{'test'}");
            $test_return = ERROR;
        }
    }

    # Run the test
    my $status = system($test->{'test'}." >>$logfile");
    if ($test->{'success'} == ERROR && $status == SUCCESS)
    {
        $logger->fatal_msg("Failed to run test $test->{'test'}");
        $test_return = ERROR;
    }
    elsif($test->{'success'} == SUCCESS && $status != SUCCESS)
    {        
        $logger->fatal_msg("Failed to run test $test->{'test'}");
        $test_return = ERROR;
    }

    if ($test->{'replace'})
    {
        # Replace current date with 04-01-2014 to make verifying easier
        my @replace = ("sed", "-i", "s|$formatted_date|04-01-2014|g",
            "update_gcps_directory/132/050/GCPLib");
        $status = system(@replace);
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Failed to replace date in GCPLib");
            $test_return = ERROR;            
        }
    }

    # Run diff on provide file run quietly 
    $status = system($test->{'diff'}." -q");
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Failed to verify test $test->{'diff'} for ".
            "$test->{'test'}");
        $test_return = ERROR;
    }
}

if ($test_return != SUCCESS)
{
    $logger->close_log();
    exit FAILURE;
}


# Run cleanup after testing
if (-d "update_gcps_directory")
{
    my $status = system($cleanup);
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Failed to clean up after test");
        $logger->close_log();
        exit FAILURE;
    }
}

print "\n--------- Completed " . basename($0) . " test ------------------\n\n";

exit SUCCESS;

################################################################################
# Name: setup_tests
#
# Purpose: Pushes all the tests onto all_tests array
#
# Parameters:
#   - Test Array reference
#
# Returns: VOID
# ##############################################################################
sub setup_tests
{
    my $all_tests = shift;

    # Setup Description, Test, and Verification for all tests
    # Read descriptions for test information
    push @{$all_tests}, {desc => "Testing that error thrown for bad date",
        test => "../update_gcps.pl update_gcps_directory/132/050/GCPLib".
        " update_gcps_directory/update_gcps/GCPLib.update --test".
        " --date_updated=12132012", 
        diff => "diff update_gcps_directory/update_gcps/GCPLib.normal".
        " update_gcps_directory/132/050/GCPLib",
        success => ERROR, replace => 0};

    push @{$all_tests}, {desc => "Testing error thrown for no list provided",
        test => "../update_gcps.pl update_gcps_directory/132/050/GCPLib".
        " update_gcps_directory/update_gcps/GCPLib.update --test", 
        diff => "diff update_gcps_directory/update_gcps/GCPLib.normal".
        " update_gcps_directory/132/050/GCPLib",
        success => ERROR, replace => 0};

    push @{$all_tests}, {desc => "Testing error thrown for bad Master GCPLib",
        test => "../update_gcps.pl update_gcps_directory/update_gcps/GCPLib.bad"
        ." update_gcps_directory/update_gcps/GCPLib.update --test ".
        "--update_fields=unknown_id", 
        diff => "diff update_gcps_directory/update_gcps/GCPLib.normal".
        " update_gcps_directory/132/050/GCPLib",
        success => ERROR, replace => 0};

    push @{$all_tests}, {desc => "Testing error thrown for bad Update GCPLib",
        test => "../update_gcps.pl update_gcps_directory/132/050/GCPLib".
        " update_gcps_directory/update_gcps/GCPLib.bad --test ".
        "--update_fields=unknown_id", 
        diff => "diff update_gcps_directory/update_gcps/GCPLib.normal".
        " update_gcps_directory/132/050/GCPLib",
        success => ERROR, replace => 0};
  
    push @{$all_tests},{desc => "Testing error thrown for disallowed field",
        test => "../update_gcps.pl update_gcps_directory/132/050/GCPLib".
        " update_gcps_directory/update_gcps/GCPLib.update --test".
        " --update_fields=point_id", 
        diff => "diff update_gcps_directory/update_gcps/GCPLib.normal".
        " update_gcps_directory/132/050/GCPLib",
        success => ERROR, replace => 0};

    push @{$all_tests},{desc => "Testing that error thrown for invalid field",
        test => "../update_gcps.pl update_gcps_directory/132/050/GCPLib ".
        "update_gcps_directory/update_gcps/GCPLib.update --test".
        " --update_fields=unknown_id",
        diff => "diff update_gcps_directory/update_gcps/GCPLib.normal".
        " update_gcps_directory/132/050/GCPLib",
        success => ERROR, replace => 0};

    push @{$all_tests},{desc => "Testing that error thrown for bad GCP ID".
        " to update",
        test => "../update_gcps.pl update_gcps_directory/132/050/GCPLib".
        " update_gcps_directory/update_gcps/GCPLib.update_bad_gcp_id --test ".
        "--update_fields=source", 
        diff => "diff update_gcps_directory/update_gcps/GCPLib.normal".
        " update_gcps_directory/132/050/GCPLib",
        success => ERROR, replace => 0};

    push @{$all_tests},{desc => "Testing that error thrown for bad CHIP ID".
        " to update",
        test => "../update_gcps.pl update_gcps_directory/132/050/GCPLib".
        " update_gcps_directory/update_gcps/GCPLib.update_bad_chip_id --test".
        " --update_fields=chip_sensor", 
        diff => "diff update_gcps_directory/update_gcps/GCPLib.normal".
        " update_gcps_directory/132/050/GCPLib",
        success => ERROR, replace => 0};

    push @{$all_tests},{desc => "Testing that the --test creates GCPLib.test",
        test => "../update_gcps.pl update_gcps_directory/132/050/GCPLib".
        " update_gcps_directory/update_gcps/GCPLib.update_50".
        " --update_fields=update_reason,source,latitude,longitude,elevation,".
        "point_type,chip_path,chip_row,reference_line,reference_sample,".
        "projection_x,projection_y,pixel_size_x,pixel_size_y,chip_sensor,".
        "chip_select_method,moravec_rank,projection,zone,abs_rel_flag,".
        "chip_acq_date,chip_file --date_updated 04-23-2014 --test", 
        diff => "diff ./GCPLib.test update_gcps_directory/update_gcps/".
        "GCPLib.updated.verify",
        success => SUCCESS, replace => 1};

    push @{$all_tests},{desc => "Testing creation of new GCPLib file and ".
        "backup",
        test => "../update_gcps.pl update_gcps_directory/132/050/GCPLib".
        " update_gcps_directory/update_gcps/GCPLib.update_50".
        " --update_fields=update_reason,source,latitude,longitude,elevation,".
        "point_type,chip_path,chip_row,reference_line,reference_sample,".
        "projection_x,projection_y,pixel_size_x,pixel_size_y,chip_sensor,".
        "chip_select_method,moravec_rank,projection,zone,abs_rel_flag,".
        "chip_acq_date,chip_file --date_updated 04-23-2014", 
        diff => "diff update_gcps_directory/update_gcps/GCPLib.updated.verify".
        " update_gcps_directory/132/050/GCPLib.updated",
        success => SUCCESS, replace => 1};

    push @{$all_tests},{desc => "Testing missing image chip to copy",
        test => "../update_gcps.pl update_gcps_directory/132/050/GCPLib".
        " update_gcps_directory/update_gcps/GCPLib.update_50".
        " --update_fields=update_reason,source,latitude,longitude,elevation,".
        "point_type,chip_path,chip_row,reference_line,reference_sample,".
        "projection_x,projection_y,pixel_size_x,pixel_size_y,chip_sensor,".
        "chip_select_method,moravec_rank,projection,zone,abs_rel_flag,".
        "chip_acq_date,chip_file --date_updated 04-23-2014", 
        diff => "diff update_gcps_directory/update_gcps/GCPLib.normal".
        " update_gcps_directory/132/050/GCPLib",
        success => ERROR, replace => 0, reset_lib => 1,
        additional_command => "rm -f update_gcps_directory/update_gcps/".
        "1320500012_01"};

    push @{$all_tests},{desc => "Testing missing original chip to check",
        test => "../update_gcps.pl update_gcps_directory/132/050/GCPLib".
        " update_gcps_directory/update_gcps/GCPLib.update_50".
        " --update_fields=update_reason,source,latitude,longitude,elevation,".
        "point_type,chip_path,chip_row,reference_line,reference_sample,".
        "projection_x,projection_y,pixel_size_x,pixel_size_y,chip_sensor,".
        "chip_select_method,moravec_rank,projection,zone,abs_rel_flag,".
        "chip_acq_date,chip_file --date_updated 04-23-2014", 
        diff => "diff update_gcps_directory/update_gcps/GCPLib.normal".
        " update_gcps_directory/132/050/GCPLib",
        success => ERROR, replace => 0, reset_lib => 1,
        additional_command => "rm -f update_gcps_directory/132/050/".
        "1320500012_01"};

    push @{$all_tests},{desc => "Testing that error is thrown when wrong".
        " size image chip is provided",
        test => "../update_gcps.pl update_gcps_directory/132/050/GCPLib".
        " update_gcps_directory/update_gcps/GCPLib.update_50".
        " --update_fields=update_reason,source,latitude,longitude,elevation,".
        "point_type,chip_path,chip_row,reference_line,reference_sample,".
        "projection_x,projection_y,pixel_size_x,pixel_size_y,chip_sensor,".
        "chip_select_method,moravec_rank,projection,zone,abs_rel_flag,".
        "chip_acq_date,chip_file --date_updated 04-23-2014", 
        diff => "diff update_gcps_directory/update_gcps/GCPLib.normal".
        " update_gcps_directory/132/050/GCPLib",
        success => ERROR, replace => 0, reset_lib => 1,
        additional_command => "cp update_gcps_directory/".
        "update_gcps/1320500012_01.large_chip update_gcps_directory/".
        "update_gcps/1320500012_01"};
}

################################################################################
# Name: initialize
#
# Purpose: Initialize all test directorys and files
# ##############################################################################
sub initialize
{
    my @init_commands; # Init commands array

    # Clean up directory first 
    if (-d "update_gcps_directory")
    {
        my $cleanup = "rm -rf ./update_gcps_directory";
        my $status = system($cleanup);
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Failed to cleanup before test");
            $logger->close_log();
            exit FAILURE;
        }
    }

    # Start of all init commands
    push @init_commands,{command => "mkdir update_gcps_directory"};
    push @init_commands,{command => "mkdir update_gcps_directory/132/"};
    push @init_commands,{command => "mkdir update_gcps_directory/132/050"};
    push @init_commands,{command => "mkdir update_gcps_directory/update_gcps"};
    push @init_commands,{command => "cp data/update_gcps_directory/update_gcps".
        "/GCP* update_gcps_directory/update_gcps/"};
    push @init_commands,{command => "cp data/update_gcps_directory/update_gcps".
        "/132* update_gcps_directory/update_gcps/"};
    push @init_commands,{command => "cp data/update_gcps_directory/132/050/".
        "GCPLib update_gcps_directory/132/050/GCPLib"};
    push @init_commands,{command => "cp data/update_gcps_directory/132/050".
        "/1320500012_01 update_gcps_directory/132/050/1320500012_01"};
    push @init_commands,{command => "cp data/update_gcps_directory/132/050".
        "/1320500027_01 update_gcps_directory/132/050/1320500027_01"};
    push @init_commands,{command => "cp data/update_gcps_directory/132/050".
        "/1320500034_01 update_gcps_directory/132/050/1320500034_01"};
    push @init_commands,{command => "cp data/update_gcps_directory/132/050".
        "/1320500036_01 update_gcps_directory/132/050/1320500036_01"};
    push @init_commands,{command => "cp data/update_gcps_directory/132/050".
        "/1320500066_01 update_gcps_directory/132/050/1320500066_01"};
    push @init_commands,{command => "cp data/update_gcps_directory/132/050".
        "/1320500139_01 update_gcps_directory/132/050/1320500139_01"};

    # Run each command in order
    foreach my $command (@init_commands)
    {
        # Run Init to copy gcp directory
        my $status = system($command->{'command'});
        if ($status != SUCCESS)
        {
            $logger->fatal_msg("Failed to initialize $command->{command}");
            $logger->close_log();
            exit FAILURE;
        }
    }
}
