#!/usr/bin/perl -w
#############################################################################
# Name: update_gcps
#
# Description: Given a directory with a GCPLib file and optional matching image
#   chips, update the existing GCPs and image chips in the master version.
#
# Parameters:
#   Master GCPLib file for the WRS-2 path/row being updated (it is assumed
#       that the image chips are in the same directory)
#   Update GCPLib file with the image chips in the same directory
#   Optional "test only" flag
#   Optional date updated parameter to use (default to today's date)
#   Optional overall GCP version number
#   List of metadata fields to update
#
#############################################################################

use strict;
use Getopt::Long;           # Command line parsing support
use File::Basename;         # Basename routine
use File::Copy;             # Copy and move routines
use List::Util qw(max);     # Max routine
use Logger;                 # Logging module
use ias_const;              # Define SUCCESS and ERROR
use ias_gcp qw(ias_gcp_write_gcplib); # GCPLib write routine
use manage_gcps_utils;      # Utility routines for managing GCPs

# Declare program constant
use constant PROG => basename($0);

# Read the optional parameters
my $test_only = 0;          # Flag to only test the updates
my $date_updated;           # Date updated value to use
my $overall_version;        # Overall version number for the output file
my @update_fields;          # Fields to update
my $update_all_chips = 0;   # Flag to update all the chips from one chip record
                            # for each GCP ID
my $verbose = 0;            # Flag to provide more messages

if (!GetOptions('test' => \$test_only, 'date_updated=s' => \$date_updated, 
        'version=i' => \$overall_version, 'update_fields=s' => \@update_fields,
        'update_all_chips' => \$update_all_chips, 'verbose' => \$verbose))
{
    # Error reading the options, so issue the usage message and exit
    usage();
    exit FAILURE;
}

# Confirm there are enough parameters remaining for the required parameters
if ($#ARGV != 1)
{
    # Error since no GCPLib filename provided
    usage();
    exit FAILURE;
}

# Set the level of messages from the manage_gcps_utils module
set_verbose_state($verbose);

# Get the master GCPLib filename
my $master_gcplib = $ARGV[0];

# Get the update GCPLib filename
my $update_gcplib = $ARGV[1];

# Set up the logger
my $logger = Logger->new("STDOUT");
if (!$logger)
{
    print STDERR PROG . ": ERROR:  Failed to establish logger object for "
        . "STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name(PROG);
set_default_log($logger);

# Get the date updated to use
my $status;
($status, $date_updated) = get_date_updated($date_updated);
if ($status != SUCCESS)
{
    $logger->close_log();
    usage();
    exit FAILURE;
}

# Build the list of fields to update
if (@update_fields)
{
    # Support comma delimited strings or separate parameters for each field
    @update_fields = split(/,/, join(',',@update_fields));
}
else
{
    $logger->fatal_msg("No list of fields to update provided");
    $logger->close_log();
    usage();
    exit FAILURE;
}

$logger->info_msg("Master GCPLib: $master_gcplib");
$logger->info_msg("Update GCPLib: $update_gcplib");

if (defined $date_updated)
{
    $logger->info_msg("Using date_updated of $date_updated");
}
else
{
    $logger->info_msg("Using date_updated from update file records");
}

# Read the master file and validate the contents
my $master_file = read_gcplib_file($master_gcplib, MASTER_FILE);
if (!defined $master_file)
{
    $logger->fatal_msg("Error reading master GCPLib file $master_gcplib");
    $logger->close_log();
    exit FAILURE;
}

# Read the update file and validate the contents
my $update_file = read_gcplib_file($update_gcplib, UPDATE_FILE);
if (!defined $update_file)
{
    $logger->fatal_msg("Error reading update GCPLib file $update_gcplib");
    $logger->close_log();
    exit FAILURE;
}

# Set the overall version to the maximum of the master and update files if
# it wasn't specified on the command line
if (!defined $overall_version)
{
    $overall_version = max $update_file->{'version'}, $master_file->{'version'};
}

# Verify the values in update_fields are present in the GCPLib records and
# determine whether the GCP and/or image chip metadata is being updated
my $gcp_fields_updated;
my $image_chip_fields_updated;
($status, $gcp_fields_updated, $image_chip_fields_updated)
    = validate_fields_to_update(\@update_fields);
if ($status != SUCCESS)
{
    $logger->fatal_msg("Invalid field names provided");
    $logger->close_log();
    exit FAILURE;
}

$logger->info_msg("Applying updates to master...") if ($verbose);

# Make the updates to the master GCPLib file
if (make_updates($update_file->{'lookup'}, $master_file->{'lookup'},
    $gcp_fields_updated, $image_chip_fields_updated, $date_updated,
    $update_file->{'dir'}, $master_file->{'dir'}, $update_all_chips)
    != SUCCESS)
{
    $logger->fatal_msg("Failed making updates to the master file");
    $logger->close_log();
    exit FAILURE;
}

$logger->info_msg("Verifying updated GCPs...") if ($verbose);

# Verify the master GCPs are consistent after the merge
if (!verify_gcps_are_consistent($master_file->{'lookup'}, MASTER_FILE))
{
    $logger->fatal_msg("Master GCPs in $master_gcplib are not consistent");
    $logger->close_log();
    exit FAILURE;
}

$logger->info_msg("Verifing image chip files") if ($verbose);

# Verify each chip that needs to be copied
foreach my $rec (@{$master_file->{'records'}})
{
    # If there is an original chip name field, it means the chip needs to
    # be copied to the target directory
    if (defined $rec->{'copy_chip'})
    {
        my $source = $update_file->{'dir'} . $rec->{'chip_name'};
        if (verify_image_chip_data($rec, $source) != SUCCESS)
        {
            $logger->fatal_msg("Verifying image chip data $rec->{'chip_name'}");
            exit FAILURE;
        }
    }
}

# By default, replace the original GCPLib file
my $output_filename = $master_gcplib . ".updated";

if (!$test_only)
{
    # Copy the image chip updates to the master directory with the correct
    # name
    my $message_issued = 0;
    foreach my $rec (@{$master_file->{'records'}})
    {
        # If there is an original chip name field, it means the chip needs to
        # be copied to the target directory
        if (defined $rec->{'copy_chip'})
        {
            # When the first chip file is copied issue a general message
            # and create the chips update file
            if (!$message_issued)
            {
                if (create_chips_updated_file($master_file->{'dir'}) != SUCCESS)
                {
                    $logger->fatal_msg("Error creating chips updated file");
                    $logger->close_log();
                    exit FAILURE;
                }

                $logger->info_msg("Updating image chip files in the master "
                    . "copy") if ($verbose);
                $message_issued = 1;
            }

            my $source = $update_file->{'dir'} . $rec->{'chip_name'};
            my $target = $master_file->{'dir'} . $rec->{'chip_name'};
            $logger->info_msg("Copying $source to $target") if ($verbose);
            if (!copy($source, $target))
            {
                $logger->fatal_msg("Error copying $source to $target");
                $logger->close_log();
                exit FAILURE;
            }
        }
    }
}
else
{
    # Just testing, override the output GCPLib filename with a test filename
    $output_filename = "GCPLib.test";
    $logger->info_msg("Test mode: Writing test output file");
}

# Scrub the extra fields from the master records to allow the write routine
# to work
foreach my $rec (@{$master_file->{'records'}})
{
    delete $rec->{'copy_chip'};
}

# Write the new master GCPLib file
$logger->info_msg("Writing updated file $output_filename") if ($verbose);

if (ias_gcp_write_gcplib($output_filename, $overall_version,
    $master_file->{'records'}) != SUCCESS)
{
    $logger->fatal_msg("Error writing test GCPLib file $output_filename");
    $logger->close_log();
    exit FAILURE;
}

$logger->close_log();

exit SUCCESS;

##############################################################################
# Name: usage
# 
# Description: Prints out the usage message for the application
#
##############################################################################
sub usage
{
    # Generate a list of fields that can be updated
    my @fields;
    my $disallowed_hash = get_disallowed_fields_hash();
    push (@fields, get_gcp_metadata_members());
    push (@fields, get_image_chip_metadata_members());
    my @filtered_fields;
    foreach my $field (@fields)
    {
        if (!exists $disallowed_hash->{$field})
        {
            push(@filtered_fields, $field);
        }
    }
    # Add chip_file as an additional option
    push(@filtered_fields, 'chip_file (copy updated image chip file)');
    my $list = join("\n          ", @filtered_fields);

    # Print out the usage message
    print "\nUsage: ". PROG . " master_GCPLib_filename update_GCPLib_filename "
        . "[--test] \\\n"
        . "    [--date_updated mm-dd-yyyy|use_file_date] "
        . "[--version version_number] \\\n"
        . "    [--update_all_chips] --update_fields <comma separated field "
        . "name list>\n\n"
        . "where master_GCPLib_filename is the master file to update\n\n"
        . "      update_GCPLib_filename contains the updates to apply\n\n"
        . "      --test indicates the output should be written to a test "
        . "output file\n      instead of updating the master\n\n"
        . "      --date_updated is the date updated to use for the new GCPs "
        . "and image\n      chips (default it today)\n\n"
        . "      --version is the version number to set in the output GCPLib "
        . "file (defaults\n      to the highest version in the two input "
        . "GCPLib files)\n\n"
        . "      --update_all_chips indicates that a single update record for "
        . "a GCP ID\n      should update all the master image chip records\n\n"
        . "      --update_fields is a comma delimited list of field names to "
        . " update from\n      the list:\n          $list\n\n";
}

##############################################################################
# Name: get_disallowed_fields_hash
#
# Description: Returns a hash of field names that are not allowed to be updated
#   by this script.
#
# Returns:
#   Reference to a hash where the keys are fields that are not allowed to be 
#   updated
##############################################################################
sub get_disallowed_fields_hash
{
    my @disallowed_fields = qw(point_id point_active point_date_updated
        point_version point_path point_row chip_name chip_active
        chip_date_updated chip_version chip_data_type chip_size_lines
        chip_size_samples);
    my %disallowed_hash;
    @disallowed_hash{@disallowed_fields} = ();
    return \%disallowed_hash;
}

##############################################################################
# Name: validate_fields_to_update
#
# Description: Given a list of metadata fields to update, verify the fields
#   are valid and split them into GCP and image chip metadata lists.
#
# Parameters:
#   - Reference to a list of fields to update
#
# Returns:
#   - SUCCESS/ERROR flag (Error returned if an invalid field name is found)
#   - Reference to an array of GCP metadata fields to update
#   - Reference to an array of image chip metadata fields to update
##############################################################################
sub validate_fields_to_update
{
    my $fields_array_ref = shift;
    my $status = SUCCESS;

    # List of GCP related metadata fields specified
    my @gcp_fields_updated;

    # List of image chip related metadata fields specified
    my @image_chip_fields_updated;

    # Get the hash of disallowed field names
    my $disallowed_hash = get_disallowed_fields_hash();

    # Convert the list of GCP metadata members to a hash
    my @gcp_fields = get_gcp_metadata_members();
    my %gcp_field_hash;
    @gcp_field_hash{@gcp_fields} = ();

    # Convert the list of image chip metadata members to a hash.  Note that
    # the chip_file is added to the list separately since it isn't metadata.
    my @image_chip_fields = get_image_chip_metadata_members();
    push(@image_chip_fields, 'chip_file');
    my %image_chip_field_hash;
    @image_chip_field_hash{@image_chip_fields} = ();

    # Check each of the fields provided against the GCP and image chip metadata
    foreach my $field (@$fields_array_ref)
    {
        if (exists $disallowed_hash->{$field})
        {
            $logger->fatal_msg("Field name $field is not allowed to be "
                . "modified by this script");
            $status = ERROR;
        }
        elsif (exists $gcp_field_hash{$field})
        {
            # This is a GCP metadata field
            push(@gcp_fields_updated, $field);
        }
        elsif (exists $image_chip_field_hash{$field})
        {
            # This is an image chip metadata field
            push(@image_chip_fields_updated, $field);
        }
        else
        {
            #Invalid field name provided
            $logger->fatal_msg("Field name $field is not a valid field name");
            $status = ERROR;
        }
    }

    return $status, \@gcp_fields_updated, \@image_chip_fields_updated;
}

##############################################################################
# Name: make_updates
#
# Description: Apply the changes in the update GCPLib file to the master 
#   version.
#
# Parameters:
#   - Update GCPLib GCP ID lookup hash reference with the updates to make
#   - Master GCPLib GCP ID lookup hash reference that will be updated
#   - GCP metadata field names to update array reference
#   - Image chip metadata field names to update array reference
#   - Date updated to use
#   - Directory for the update information (used to check whether the image
#       chip files actually changed)
#   - Directory for the master information
#   - Flag to update all the image chips for a GCP from one update record
#
# Returns:
#   SUCCESS/ERROR
##############################################################################
sub make_updates
{
    my $update_hash = shift;
    my $master_hash = shift;
    my $gcp_update_fields = shift;
    my $image_chip_update_fields = shift;
    my $date_updated = shift;
    my $update_dir = shift;
    my $master_dir = shift;
    my $update_all_chips = shift;

    # Go through the list of GCPs to update
    foreach my $update_id (sort keys %$update_hash)
    {
        my $update_array = $update_hash->{$update_id};
        my $master_array = $master_hash->{$update_id};

        # Verify the GCP ID exists in the master file
        if (!defined $master_array)
        {
            $logger->fatal_msg("GCP ID $update_id from the update file does "
                . "not exist in the master file");
            return ERROR;
        }

        # If there is GCP metadata to update, update those fields in every 
        # master record.  Use the first update record as the source since
        # all the update records should have already been checked to confirm
        # they have the same GCP metadata.
        if (scalar @$gcp_update_fields > 0)
        {
            update_gcp_metadata($update_array->[0], $master_array,
                $gcp_update_fields, $date_updated);
        }
        
        # If there is image chip metadata to update, look for the matching 
        # master records to update
        if (scalar $image_chip_update_fields > 0)
        {
            if (update_chip_metadata($update_array, $master_array,
                $image_chip_update_fields, $update_dir, $master_dir,
                $date_updated, $update_all_chips) != SUCCESS)
            {
                $logger->fatal_msg("Failed updating the chip metadata for "
                    . "GCP ID $update_id");
                return ERROR;
            }
        }
    }

    return SUCCESS;
}

##############################################################################
# Name: update_gcp_metadata
#
# Description: Update all the master GCP records for a GCP ID with the values
#   from the update file.
#
# Parameters:
#   - Update metadata record hash reference
#   - Reference to the array of records for the GCP ID from the master file
#   - Reference to the array of field names to update
#   - Date updated for records that receive updates
#
# Returns:
#   Nothing - always succeeds
##############################################################################
sub update_gcp_metadata
{
    my $update_rec = shift;
    my $master_array = shift;
    my $gcp_update_fields = shift;
    my $date_updated = shift;

    my $update_id = $update_rec->{'point_id'};

    foreach my $master_rec (@$master_array)
    {
        # Track whether the master record is updated
        my $changed = 0;

        # Update each of the fields specified
        foreach my $field (@$gcp_update_fields)
        {
            # If the value is different, update it and flag the record
            # changed
            if ($master_rec->{$field} ne $update_rec->{$field})
            {
                $master_rec->{$field} = $update_rec->{$field};
                $changed = 1;
            }
        }

        # Update the version number and date updated fields if any of
        # the fields actually changed
        if ($changed == 1)
        {
            $master_rec->{'point_version'}++;
            if (defined $date_updated)
            {
                $master_rec->{'point_date_updated'} = $date_updated;
            }
            else
            {
                # Use the update record date if $date_updated is not defined
                $master_rec->{'point_date_updated'}
                    = $update_rec->{'point_date_updated'};
            }
        }
        else
        {
            # If the record wasn't updated, it is suspicious, so issue
            # a warning
            $logger->warn_msg("GCP ID $update_id had no GCP metadata "
                . "fields updated for image chip "
                . $master_rec->{'chip_name'});
        }
    }
}

##############################################################################
# Name: update_chip_metadata
#
# Description: Update the image chip metadata for all the records associated
#   with a GCP.
#
# Parameters:
#   - Reference to the array of records for the GCP ID from the update file
#   - Reference to the array of records for the GCP ID from the master file
#   - Reference to the array of field names to update
#   - Directory name containing the updates
#   - Directory name containing the master files
#   - Date updated for records that receive updates
#   - Flag to update all the image chip metadata fields from one update record
#
# Returns:
#   SUCCESS/ERROR
##############################################################################
sub update_chip_metadata
{
    my $update_array = shift;
    my $master_array = shift;
    my $image_chip_update_fields = shift;
    my $update_dir = shift;
    my $master_dir = shift;
    my $date_updated = shift;
    my $update_all_chips = shift;

    # Variables used to determine whether differences in projections should be
    # reported if coordinates are being updated.
    my $projection_coord_updates = 0;
    my $projection_updates = 0;
    my $warn_about_incompatible_projections = 0;

    # Get the GCP ID that is being updated
    my $update_id = $update_array->[0]{'point_id'};

    # If the update all chips flag is set and there is more than one update
    # record, there is a problem.  The update all chips flag is meant to be
    # used when updates are provided that are chip independent (i.e. if all the
    # image chip projection x/y values should be set the same as the update
    # record)
    if ($update_all_chips && (scalar @$update_array > 1))
    {
        $logger->fatal_msg("--update_all_chips specified, but multiple update "
            . "records supposed for GCP ID $update_id");
        return ERROR;
    }

    # Determine whether projection_x/y values are being updated wih
    foreach my $field (@$image_chip_update_fields)
    {
        if ($field =~ /^projection_/)
        {
            $projection_coord_updates = 1;
        }
        elsif ($field =~ /^projection/)
        {
            $projection_updates = 1;
        }
        elsif ($field =~ /^zone$/)
        {
            $projection_updates = 1;
        }
    }
    # If projection coordinates are being updated but the projection is not
    # warn if the projections between the update and master files disagree
    if ($projection_coord_updates && !$projection_updates)
    {
        $warn_about_incompatible_projections = 1;
    }
    

    # Use a hash to track which image chip IDs have been updated
    my %master_chips_updated;

    foreach my $update_rec (@$update_array)
    {
        my $update_chip_id = $update_rec->{'chip_name'};

        # Search for a matching master record with the same image chip id
        my $found = 0;
        foreach my $master_rec (@$master_array)
        {
            if ($master_rec->{'chip_name'} eq $update_chip_id
                || $update_all_chips)
            {
                my $changed = 0;
                my $master_chip_id = $master_rec->{'chip_name'};

                # If updating projection coordinates, warn if the projection
                # or zone are different between the master and update
                if ($warn_about_incompatible_projections)
                {
                    if ($master_rec->{'projection'}
                            ne $update_rec->{'projection'}
                        || $master_rec->{'zone'} ne $update_rec->{'zone'})
                    {
                        $logger->warn_msg("Incompatible projections for image "
                            . "chips for GCP ID $update_id");
                    }
                }

                # Remember which master copy image chips were updated
                $master_chips_updated{$master_chip_id} = 1;

                # Update each of the fields specified
                foreach my $field (@$image_chip_update_fields)
                {
                    if ($field eq "chip_file")
                    {
                        if ($update_all_chips)
                        {
                            # Not legal to update chip files when using the
                            # update_all_chips mode
                            $logger->fatal_msg("--update_all_chips cannot be "
                                . "used when updating chip files");
                            return ERROR;
                        }

                        # The image chip file is being updated, verify the
                        # input file was provided and is different than the
                        # existing chip file
                        my ($status, $chip_changed)
                            = is_image_chip_file_changed($update_dir,
                                $master_dir, $master_chip_id);
                        if ($status != SUCCESS)
                        {
                            $logger->fatal_msg("Failed checking for chip file "
                                . "changes");
                            return ERROR;
                        }

                        if ($chip_changed)
                        {
                            # Chip lines, samples, and datatype must updated
                            $master_rec->{'chip_data_type'} = 
                                $update_rec->{'chip_data_type'};

                            $master_rec->{'chip_size_lines'} = 
                                $update_rec->{'chip_size_lines'};

                            $master_rec->{'chip_size_samples'} = 
                                $update_rec->{'chip_size_samples'};
                            
                            # Add a temporary member to the master record to
                            # indicate the chip needs to be copied
                            $master_rec->{'copy_chip'} = 1;
                            $changed = 1;
                        }
                    }
                    elsif ($master_rec->{$field} ne $update_rec->{$field})
                    {
                        $master_rec->{$field} = $update_rec->{$field};
                        $changed = 1;
                    }
                }

                # If a value really changed, update the version number and date
                # updated fields
                if ($changed == 1)
                {
                    $master_rec->{'chip_version'}++;
                    if (defined $date_updated)
                    {
                        $master_rec->{'chip_date_updated'} = $date_updated;
                    }
                    else
                    {
                        # Use the update record date if $date_updated is not
                        # defined
                        $master_rec->{'chip_date_updated'}
                            = $update_rec->{'chip_date_updated'};
                    }
                }
                else
                {
                    # If the record wasn't updated, it is suspicious, so issue
                    # a warning
                    $logger->warn_msg("Image chip ID $master_chip_id had no "
                        . "chip metadata fields updated");
                }

                $found = 1;
                last if (!$update_all_chips);
            }
        }

        # Make sure a matching record was found
        if ($found == 0)
        {
            $logger->fatal_msg("Could not find image chip id "
                . $update_rec->{'chip_name'} . " in the master file");
            return ERROR;
        }
    }

    # If only a subset of image chips were updated, issue a warning in case it
    # was intended that they all be updated.
    # TBD - This warning may found to be not useful in the future.  But since
    # we currently only have a single image chip per GCP ID, we'll leave it
    # here until it proves to be an annoyance.  It is possible that the
    # warning may need to be made dependent on the fields being updated in
    # the future.
    if (scalar (keys %master_chips_updated) != scalar @$master_array)
    {
        $logger->warn_msg("The update GCPLib file did not update "
            . "all the image chips for GCP ID $update_id");
    }

    return SUCCESS;
}

##############################################################################
# Name: is_image_chip_file_changed
#
# Description: Determine whether the image chip file provided with the update
#   is different than the master chip file.
#
# Parameters:
#   - Directory name containing the update files
#   - Directory name containing the master files
#   - Image chip ID
#
# Returns:
#   SUCCESS/ERROR (error if one of the chip files doesn't exist)
#   TRUE if the chip file changed, otherwise FALSE
##############################################################################
sub is_image_chip_file_changed
{
    my $update_dir = shift;
    my $master_dir = shift;
    my $chip_id = shift;

    # Build the full path to both image chip files
    my $update_file = $update_dir . $chip_id;
    my $master_file = $master_dir . $chip_id;

    # Make sure both files exist
    if (! -e $update_file)
    {
        $logger->fatal_msg("Image chip file $update_file does not exist");
        return ERROR, FALSE;
    }

    if (! -e $master_file)
    {
        $logger->fatal_msg("Image chip file $master_file does not exist");
        return ERROR, FALSE;
    }

    # Compare the two files
    `cmp $update_file $master_file`;
    if ($? == 0)
    {
        # If the status is zero, the files are identical, so no change.  Issue
        # a warning since this is suspicious when it was requested that the
        # image chip files be updated
        $logger->warn_msg("Image chip files for $chip_id are identical - "
            . "that is suspicious");
        return SUCCESS, FALSE;
    }

    # Image chip files are different, so return TRUE
    return SUCCESS, TRUE;
}
