/******************************************************************************
               Image Assessment System (IAS) Software
 ******************************************************************************
 *
 * Name:   BrowseWOParms.cpp
 *
 * PURPOSE:
 * This is the definition of the BrowseWOParms class which is used to
 * view work order information from the work_orders, wo_scripts, and wo_parms
 * tables.
 *
 *****************************************************************************/

// System Includes

// Qt Includes
#include <QtCore>
#include <QtGui>
#include <QtSql>

// IAS Qt Includes
#include <iasQtWoConst.h>

// Local Includes
#include "BrowseWOParms.h"
#include "IASUtilities.h"
#include "IASListView.h"
#include "IASListViewItem.h"
#include "IASSqlQuery.h"
#include "IASSchemaConfig.h"
#include "IASKeyValueComboBox.h"

//***********************************************************************
BrowseWOParms::BrowseWOParms(
    QWidget* parent,    //I: The address to this widget's parent
    Qt::WindowFlags fl) //I: Used by QT
: QWidget(parent, fl)
{
    // This is the constructor for the BrowseWOParms object.  It
    // initializes variables needed by the object, retrieves the
    // necessary values from the database, and displays the values

    QWidget::setAttribute(Qt::WA_DeleteOnClose);

    // Set up the user interface.
    setupUi(this);

    // Initialize listview column information structures
    //--------------------------------------------------
    initStructs();

    // Setup the satellite sensor selection box.
    //------------------------------------------
    QStringList schemas = IASSchemaConfig::instance()->allKeys();
    QString empty("");
    kvcboSatSensor->addKeyValue(empty, empty);
    QStringList::iterator itr;
    for(itr = schemas.begin(); itr != schemas.end(); ++itr)
    {
        QString satSen = (*itr);
        QString schema = IASSchemaConfig::instance()->schemaNameFromKey(satSen);
        satSen.replace("_", " ");
        kvcboSatSensor->addKeyValue(satSen, schema);
    }

    // Populate the instances
    getInstances();

    // Set up the list views and properties
    //-------------------------------------
    setupListViews();

    // Set up the date filter
    //-----------------------
    setupFilters();

    // Connect signals and slots
    //--------------------------
    connect(lvWorkOrders, SIGNAL(itemSelectionChanged()),
            this, SLOT(popWOScripts()));
    connect(lvWOScripts, SIGNAL(itemSelectionChanged()),
            this, SLOT(popWOParms()));
    connect(pbQuery, SIGNAL(clicked()), this, SLOT(queryWO()));
    connect(deFrom, SIGNAL(dateChanged(const QDate &)),
            this, SLOT(changeToDateMin()));
    connect(deTo, SIGNAL(dateChanged(const QDate &)),
            this, SLOT(changeFromDateMax()));
    connect(kvcboSatSensor, SIGNAL(activated(const QString &)),
            this, SLOT(changedSchema(const QString &)));
    connect(pbClose, SIGNAL(clicked()), this, SLOT(close()));

    // Insert the work order listview items
    //-------------------------------------
    popWO();
}

//***********************************************************************
void BrowseWOParms::getInstances()
{
    // This function gets all of the current processing_instances

    IASSqlQuery query;
    QString sqlStmt;
    QString instance_name;
    int index = 0;

    // If the query fails to get the processing instances then
    // the browser will close, the query failure will display a dialog
    // before closing
    sqlStmt = "select instance_name from processing_instances "
        "order by collection_number desc";
    if (!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        close();
    }

    // Get the IAS_INSTANCE_NAME from the environment
    instance_name = getenv("IAS_INSTANCE_NAME");
    if (instance_name.isEmpty())
    {
        QMessageBox::critical(0, "Missing Environment",
            "Could not get IAS_INSTANCE_NAME from environment.\n"
            "This variable is required for initial work order creation.",
            QMessageBox::Ok, 0);
        return;
    }

    // Add each instances to the dropdown and select the current instance
    // as the default
    while (query.next())
    {
        cboInstance->addItem(query.value(0).toString());
        if (query.value(0).toString().compare(instance_name) == 0)
        {
            cboInstance->setCurrentIndex(index);
        }
        index++;
    }

    cboInstance->setEnabled(true);
}

//***********************************************************************
bool BrowseWOParms::getCurrentConfiguration(QString& instance,
    QString& satellite, QString& sensor)
{
    // This function sets the current instance, satellite, and sensor
    QStringList satSenList = kvcboSatSensor->currentText().split(' ');

    // Make sure the satellite and sensor are present
    if (satSenList.size() != 2)
        return false;

    satellite = satSenList[0];
    sensor = satSenList[1];
    instance = cboInstance->currentText();

    // Return false if satellite, sensor, or instance is empty
    if (sensor.isEmpty() || satellite.isEmpty() || instance.isEmpty())
    {
        QMessageBox::warning(this,"Warning","The satellite, sensor, and "
            "instance must be set to retrieve work orders");
        return false;
    }

    return true;
}

//***********************************************************************
void BrowseWOParms::initStructs()
{
    // Initializes the listview column info. structures

    // Initialize temporary arrays and copy to the class column info
    // arrays
    //--------------------------------------------------------------
    lvFields tmpwoFields[] = {
    {"WO Id", "WORK_ORDER_ID", "WORK_ORDERS", Id, 70},
    {"Start Date/Time", "REQ_START_DATE", "WORK_ORDERS", DateTime, 170},
    {"Procedure", "PROCEDURE_ID", "WORK_ORDERS", Alpha, 150},
    {"Scene Id", "SCENE_ID", "WORK_ORDERS", Alpha, 150},
    {"L0R Date/Time", "DATE_L0R_REQUESTED", "WORK_ORDERS", DateTime, -1},
    {"State", "STATE", "WORK_ORDERS", Alpha, 75},
    {"Deleted", "DELETED", "WORK_ORDERS", Alpha, -1},
    {"Archived", "ARCHIVED", "WORK_ORDERS", Alpha, -1},
    {"Requester", "REQUESTER", "WORK_ORDERS", Alpha, -1},
    {"WO Path", "WO_PATH", "WORK_ORDERS", Alpha, -1},
    {"CPF Path", "CPF_OVERRIDE", "WORK_ORDERS", Alpha, 150},
    {"Satellite", "SATELLITE", "WORK_ORDERS", Alpha, -1},
    {"Sensor", "SENSOR", "WORK_ORDERS", Alpha, -1},
    };

    // Set up parameters for the work order scripts table
    //---------------------------------------------------
    lvFields tmpwoScriptsFields[] = {
    {"Script Id", "SCRIPT_ID", "WO_SCRIPTS", Alpha, -1},
    {"Pause", "PAUSE_FLAG", "WO_SCRIPTS", Alpha, -1},
    {"State", "STATE", "WO_SCRIPTS", Alpha, 70},
    {"Compl Status", "COMPL_STATUS", "WO_SCRIPTS", Alpha, -1},
    {"Failure Reason", "FAILURE_REASON", "WO_SCRIPTS", Alpha, -1}
    };

    // Set up parameters for the work order parms table
    //-------------------------------------------------
    lvFields tmpwoParmsFields[] = {
    {"Parameter Name", "PARM_NAME", "WO_PARMS", Alpha, -1},
    {"Element", "PARM_OCCURRENCE", "WO_PARMS", Number, -1},
    {"Value", "PARM_VALUE", "WO_PARMS", Alpha, -1}
    };

    int i;

    // Update the class structures to contain the listview column info
    //----------------------------------------------------------------
    m_numWOFields = ARRAY_SIZE(tmpwoFields);
    for (i = 0; i < m_numWOFields; i++)
        woFields[i] = tmpwoFields[i];

    m_numWOScriptsFields = ARRAY_SIZE(tmpwoScriptsFields);
    for (i = 0; i < m_numWOScriptsFields; i++)
        woScriptsFields[i] = tmpwoScriptsFields[i];

    m_numWOParmsFields = ARRAY_SIZE(tmpwoParmsFields);
    for (i = 0; i < m_numWOParmsFields; i++)
        woParmsFields[i] = tmpwoParmsFields[i];
}

//********************************************************************
void BrowseWOParms::setupListViews()
{
    // Sets up the work order info, script info, and parameters listviews

    int i;
    QHeaderView* hdrWorkOrders = NULL;
    QHeaderView* hdrWOScripts = NULL;
    QHeaderView* hdrWOParms = NULL;
    QStringList labels;

    // Grab the headers.
    hdrWorkOrders = lvWorkOrders->header();
    hdrWOScripts = lvWOScripts->header();
    hdrWOParms = lvWOParms->header();

    lvWorkOrders->setRootIsDecorated(false);
    lvWOScripts->setRootIsDecorated(false);
    lvWOParms->setRootIsDecorated(false);

    // Add columns to the work orders, work order scripts, and work order
    // parms list views, create valuelist for sorting by IASListViewItem
    // class
    //-------------------------------------------------------------------
    labels.clear();
    for (i=0; i < m_numWOFields; i++)
    {
        labels.append(woFields[i].colName);
        m_woTypesList.append(woFields[i].colType);
    }
    lvWorkOrders->setHeaderLabels(labels);

    labels.clear();
    for (i=0; i < m_numWOScriptsFields; i++)
    {
        labels.append(woScriptsFields[i].colName);
    }
    lvWOScripts->setHeaderLabels(labels);

    labels.clear();
    for (i=0; i < m_numWOParmsFields; i++)
    {
        labels.append(woParmsFields[i].colName);
    }
    lvWOParms->setHeaderLabels(labels);

    // Allow sorting of the work orders listview.  Probably don't want to
    // allow sorting on the other two.
    //-------------------------------------------------------------------
    hdrWorkOrders->setSortIndicatorShown(true);
    lvWorkOrders->setSortingEnabled(true);
    lvWOScripts->setSortingEnabled(false);
    lvWOParms->setSortingEnabled(false);

    // Set the resizeMode to 'LastColumn' so the last column
    // will resize itself if a scroll bar is present
    //------------------------------------------------------
    hdrWorkOrders->setStretchLastSection(true);
    hdrWOScripts->setStretchLastSection(true);
    hdrWOParms->setStretchLastSection(true);
}

//***********************************************************************
void BrowseWOParms::setupFilters()
{
    // This function sets up the date, Procedure Id, and Requester filters.
    // Initially the last year of data will be shown, with the other two
    // filters off.

    QDate date = QDate::currentDate();

    // Set the format for how the date will appear: MM/DD/YYYY
    deFrom->setDisplayFormat("MM/dd/yyyy");
    deTo->setDisplayFormat("MM/dd/yyyy");

    // Set the date line edits to filter the previous two weeks' work
    // orders.  We don't need to be able to go back more than 50 years.
    //-------------------------------------------------------------
    deFrom->setDate(date.addDays(-14));
    deTo->setDate(date);

    deFrom->setDateRange(date.addYears(-50), date);
    deTo->setDateRange(deFrom->date(), date);

    // Insert into the work order ID filter combo box
    //-----------------------------------------------
    setWoFilter();

    // Insert into the procedure ID filter combo box, first
    // string is "All" so the filter is initially turned off.
    //-------------------------------------------------------
    changedSchema();

    // Make sure the Requester filter is cleared
    //------------------------------------------
    leRequester->clear();
}

//***********************************************************************
void BrowseWOParms::changedSchema(const QString &)
{
    IASSqlQuery query;
    QString satellite;
    QString sensor;
    QString instance;
    QString sqlStmt;

    setWoFilter();

    cboProcId->clear();

    // Do not retrieve the procedure list if the satellite/sensor/instances
    // are not populated
    if (!getCurrentConfiguration(instance, satellite, sensor))
    {
        return;
    }

    sqlStmt = "select unique module_id from sub_modules where "
        "processing_instance = '" + instance + "' "
        "and satellite = '" + satellite + "' "
        "and sensor = '" + sensor + "' ";
    cboProcId->addItem(ALL_PROCEDURES);

    query.exec(sqlStmt, __FILE__, __LINE__);
    if (query.isActive())
    {
        while (query.next())
        {
            cboProcId->addItem(query.value(0).toString());
        }
    }
}

//***********************************************************************
void BrowseWOParms::setWoFilter()
{
    // This function sets up the work order id filter.  Work order ids
    // displayed in the filter are within the date range specified in
    // the date filter.

    IASSqlQuery query;
    QString sqlStmt;
    QString satellite;
    QString sensor;
    QString instance;

    cboWoId->clear();

    // Do not retrieve the work order list if the satellite/sensor/instances
    // are not populated
    if (!getCurrentConfiguration(instance, satellite, sensor))
    {
        return;
    }

    /* Using 'L%' removes the LPGS orders from the results. */
    sqlStmt = "select work_order_id, procedure_id, date_entered "
              "from work_orders "
              "where work_order_id not like 'L%' and date_entered "
              "between '" + deFrom->date().toString("dd-MMM-yyyy") +
              "' and '" + deTo->date().addDays(1).toString("dd-MMM-yyyy") +
              "' and processing_instance = '" + instance + "' "
              "and satellite = '" + satellite + "' "
              "and sensor = '" + sensor + "' " +
              "order by date_entered desc";

    // Load the combobox with work order IDs for the current date range
    // first choice should be all work orders.  Clear it first in case it
    // has been loaded before.  Current item should be set to "all"
    // initially.

    cboWoId->addItem(ALL_WORK_ORDERS);

    query.exec(sqlStmt, __FILE__, __LINE__);
    if (query.isActive())
    {
        while (query.next())
        {
            cboWoId->addItem(query.value(0).toString() + " - " +
                             query.value(1).toString());
        }
    }
}

//***********************************************************************
void BrowseWOParms::changeToDateMin()
{
    // This slot will reset the To date edit min value when the
    //     From date edit value is changed

    deTo->setMinimumDate(deFrom->date());
}

//***********************************************************************
void BrowseWOParms::changeFromDateMax()
{
    // This slot will reset the From date edit max value when the
    //     To date edit value is changed

    deFrom->setMaximumDate(deTo->date());
}

//***********************************************************************
QString BrowseWOParms::woWhereClause()
{
    // Sets the where clause for the query depending on the current
    // settings of the procedure ID and date filters

    QString filterClause;
    QString satellite;
    QString sensor;
    QString instance;

    // Do not return the where clause if the satellite/sensor/instances
    // are not populated
    if (!getCurrentConfiguration(instance, satellite, sensor))
    {
        return "";
    }

    filterClause += "where processing_instance = '" + instance + "' "
    "and satellite = '" + satellite + "' "
    "and sensor = '" + sensor + "' ";

    // If filtering on work order ID disregard other filters
    // work order ID is first space delimited field
    //---------------------------------------------
    if (cboWoId->currentText() != ALL_WORK_ORDERS)
    {
        filterClause += "and work_order_id= '" +
            cboWoId->currentText().left(cboWoId->currentText().indexOf(" ")) +
            "'";
        return filterClause;
    }

    // Otherwise if not filtering on work order ID, we need to check the
    // procedure, requester, and date filters to determine what to display
    //--------------------------------------------------------------------
    if (cboProcId->currentText() != ALL_PROCEDURES)
    {
        filterClause += " and procedure_id = '"+ cboProcId->currentText() +"'";
    }

    if (leRequester->text().length() > 0)
    {
        filterClause += " and requester = '" + leRequester->text() + "'";
    }

    filterClause += " and date_entered between '" +
                    deFrom->date().toString("dd-MMM-yyyy") +
                    "' and '" +
                    deTo->date().addDays(1).toString("dd-MMM-yyyy") +
                    "'";
    filterClause += " and work_order_id not like 'L%'";

    return filterClause;
}

//***********************************************************************
void BrowseWOParms::popWO()
{
    // Populates the work order information listview using the current
    // filter values

    IASListViewItem *lviWO;
    QString sqlStmt;
    QString whereClause;
    IASSqlQuery query;
    int i;

    QString schema = kvcboSatSensor->currentValue();

    // Change the cursor while waiting for WO from database
    QApplication::setOverrideCursor(Qt::WaitCursor);

    if(!schema.isEmpty())
    {
        // Create the SQL select to get the work_order table rows
        // Date and time fields will have the "/" separator
        sqlStmt = "select " + woFields[0].fieldName;
        for (i = 1; i < m_numWOFields; i++)
        {
            if ((woFields[i].fieldName == "REQ_START_DATE") ||
                (woFields[i].fieldName == "DATE_L0R_REQUESTED"))
            {
                sqlStmt += ",to_char(" +
                           woFields[i].fieldName +
                           ", 'YYYY-MM-DD/HH24:MI:SS')";
            }
            else if(woFields[i].fieldName == "DELETED")
            {
                // If the work order was deleted because it has expired,
                // indicate that in the Deleted field. Expired work orders
                // will have a non-null value in DATE_EXPIRED
                sqlStmt += ",nvl2(DATE_EXPIRED, 'Expired', DELETED)";
            }
            else
            {
                // Is the field from the WO or the L0R DB table?
                sqlStmt += "," + woFields[i].tableName +
                           "." + woFields[i].fieldName;
            }
        }

        sqlStmt += " ,date_entered from work_orders ";

        whereClause = woWhereClause();
        if(!whereClause.isEmpty())
        {
            sqlStmt += whereClause;
        }

        sqlStmt += " order by date_entered desc";

        query.exec(sqlStmt, __FILE__, __LINE__);
        if (query.isActive())
        {
            while (query.next())
            {
                lviWO = new IASListViewItem(lvWorkOrders, m_woTypesList);
                for (i = 0; i < m_numWOFields; i++)
                    lviWO->setText(i, convertText(query.value(i).toString(),
                        woFields[i].fieldName));
            }
        }

        /* Resize the columns to fit the data. */
        for (i = 0; i < m_numWOFields; i++)
        {
            lvWorkOrders->resizeColumnToContents(i);
        }

        // Always select the first row initially (if there is one!)
        //---------------------------------------------------------
        if (lvWorkOrders->topLevelItem(0) != 0)
            lvWorkOrders->setCurrentItem(lvWorkOrders->topLevelItem(0), true);
    }

    // Sort by Work Order Id.
    // ----------------------------------------------------------------------
    lvWorkOrders->sortItems(0, Qt::DescendingOrder);

    // Return the cursor to its original shape
    // ----------------------------------------
    QApplication::restoreOverrideCursor();
}

//***********************************************************************
void BrowseWOParms::popWOScripts()
{
    // Populates the work order scripts listview when a work order is
    // selected from the work order information listview
    IASListViewItem *lviNew = NULL;
    QString sqlStmt;
    IASSqlQuery query;
    int i;
    QTreeWidgetItem* cur = lvWorkOrders->currentItem();

    lvWOScripts->clear();  // Clear the listview so we don't keep adding

    // Create the SQL select to get the work_scripts table rows
    //---------------------------------------------------------
    sqlStmt = "select " + woScriptsFields[0].fieldName;
    for (i = 1; i < m_numWOScriptsFields; i++)
        sqlStmt += "," + woScriptsFields[i].fieldName;

    sqlStmt += ", work_order_id"; // Need WO ID for wo_parms table query

    sqlStmt += " from wo_scripts where work_order_id = '" +
               cur->text(0) +
               "' order by seq_nbr asc";

    query.exec(sqlStmt, __FILE__, __LINE__);
    if (query.isActive())
    while (query.next())
    {
        lviNew = new IASListViewItem(lvWOScripts);
        for (i = 0; i < m_numWOScriptsFields; i++)
            lviNew->setText(i, convertText(query.value(i).toString(),
                woScriptsFields[i].fieldName));
    }

    /* Resize the columns to fit the data. */
    for (i = 0; i < m_numWOScriptsFields; i++)
    {
        lvWOScripts->resizeColumnToContents(i);
    }

    // Always select the first row initially
    //--------------------------------------
    if (lvWOScripts->topLevelItem(0) != 0)
        lvWOScripts->setCurrentItem(lvWOScripts->topLevelItem(0), true);
    else
        lvWOParms->clear(); // Make sure parameters listview shows nothing
}

//***********************************************************************
void BrowseWOParms::popWOParms()
{
    IASListViewItem *lviNew;
    QString sqlStmt;
    IASSqlQuery query;
    int i;
    QTreeWidgetItem* cur = lvWOScripts->currentItem();

    lvWOParms->clear();  // Clear listview so we don't keep adding to it

    QString woId("");
    if(lvWorkOrders->currentItem() != 0)
    {
        woId = lvWorkOrders->currentItem()->text(0);
    }

    if(!woId.isEmpty()) // Verify that we have something to process.
    {
        // Create the SQL select to get the wo_parms table rows
        //-----------------------------------------------------
        sqlStmt = "select " + woParmsFields[0].fieldName;
        for (i = 1; i < m_numWOParmsFields; i++)
            sqlStmt += "," + woParmsFields[i].fieldName;

        sqlStmt += " from wo_parms where work_order_id = '" +
                   woId + "' and script_id = '" +
                   cur->text(0) +
                   "' order by parm_name asc, parm_occurrence asc";

        query.exec(sqlStmt, __FILE__, __LINE__);
        if (query.isActive())
        {
            while (query.next())
            {
                lviNew = new IASListViewItem(lvWOParms);
                for (i = 0; i < m_numWOParmsFields; i++)
                    lviNew->setText(i, convertText(query.value(i).toString(),
                        woParmsFields[i].fieldName));
            }
        }
    }

    /* Resize the columns to fit the data. */
    for (i = 0; i < m_numWOParmsFields; i++)
    {
        lvWOParms->resizeColumnToContents(i);
    }

    // Always select the first row initially
    //--------------------------------------
    if (lvWOParms->topLevelItem(0) != 0)
        lvWOParms->setCurrentItem(lvWOParms->topLevelItem(0), true);
}

//***********************************************************************
void BrowseWOParms::queryWO()
{
    // Slot is called when the query button is clicked

    // Clear all the listviews
    //------------------------
    lvWorkOrders->clear();
    lvWOScripts->clear();
    lvWOParms->clear();

    // Populate the work orders listview
    //----------------------------------
    popWO();

    // Make sure work order filter is for the current date range
    // unless we are currently filtering on one work order
    //----------------------------------------------------------
    if (cboWoId->currentText() == ALL_WORK_ORDERS)
        setWoFilter();
    else
    {
        leRequester->clear();
        cboProcId->setCurrentIndex(0);
    }
}

//***********************************************************************
QString BrowseWOParms::convertText(QString fldText, QString fldName)
{
    // Converts single character values from the database to strings
    // to be displayed in the listviews

    QString uppercaseStr = fldName.toUpper();

    if (uppercaseStr == "STATE")
    {
        return IASUtilities::getStateText(fldText);
    }
    else if (uppercaseStr == "ARCHIVED" ||
             uppercaseStr == "DELETED"  ||
             uppercaseStr == "PAUSE_FLAG")
    {
        if (fldText == "1")
            return "Yes";
        else if (fldText == "0")
            return "No";
        else
            return fldText;
    }
    else if (uppercaseStr == "COMPL_STATUS")
    {
        if (fldText == DB_WO_SCRIPT_STATUS_SUCCESS)
            return UI_WO_SCRIPT_STATUS_SUCCESS_STR;
        else if (fldText == DB_WO_SCRIPT_STATUS_FAILURE)
            return UI_WO_SCRIPT_STATUS_FAILURE_STR;
        else
            return fldText;
    }
    else
        return fldText;
}

