/******************************************************************************
               Image Assessment System (IAS) Software
 ******************************************************************************
 *
 * Name:   ConfigureIAS.cpp
 *
 * PURPOSE:
 * This is the implementation of the ConfigureIAS class which is used to 
 * view and set values in the system parameter table (sys_parms).
 *
 *****************************************************************************/

// System Includes

// Qt Includes 
#include <QtCore>
#include <QtGui>
#include <QtSql>

// Local Includes
#include "ConfigureIAS.h"
#include "IASUtilities.h"
#include "IASSqlQuery.h"

//***********************************************************************
ConfigureIAS::ConfigureIAS(
    QWidget* parent,    // I: The address to this widget's parent 
    Qt::WindowFlags fl) // I: Used by QT
: QWidget(parent, fl)
{
    QWidget::setAttribute(Qt::WA_DeleteOnClose);

    // Set up the user interface.
    setupUi(this);

    m_systemId = getenv("IAS_SYSTEM_ID");
    if(m_systemId.isEmpty())
    {
        QMessageBox::critical(0,"Missing Environment",
            "Could not get IAS_SYSTEM_ID from environment.\n"
            "This variable is required for system startup or shutdown",
            QMessageBox::Ok,0);
        close();
    }
    m_systemId = m_systemId.toUpper();
    /* Using only the first character since the env. var. could be longer. */
    m_systemId = m_systemId.at(0);

    // In the constructor get the initial system configuration values and
    // update the spinbox values.

    IASSqlQuery query;
    
    initParmStructs();
    
    QString strSqlSelect = "Select " + spParms[0].spField;
    int i;

    for(i = 1; i < numSysPrmFields; i++)
    {
        strSqlSelect += "," + spParms[i].spField;
    }
    
    strSqlSelect += " from sys_parms where system_id = '" + m_systemId + "'";

    // Get the current configuration information from table sys_parms
    if(query.exec(strSqlSelect, __FILE__, __LINE__))
    {
        query.next();

        // Assign the values to the appropriate spin boxes
        for(i = 0; i < numSysPrmFields; i++)
        {
            if(!spParms[i].spinObject->suffix().compare("%"))
            {
                spParms[i].spinObject->setValue((int)(query.value(i).toDouble()
                * 100.));
            }
            else
            {
                spParms[i].spinObject->setValue(query.value(i).toInt());
            }
        }
    }

    // Connect all of the signal slots.
    connect(pbCancel, SIGNAL(clicked()), this, SLOT(close()));
    connect(pbOK, SIGNAL(clicked()), this, SLOT(updateDbase()));
}

//***********************************************************************
void ConfigureIAS::initParmStructs()
{
    // Initializes the system parameter information structure
    sysPrmInfo tmpParms[] = {
        {"DISK_FIRST_ADVISORY_LIMIT", spinFirstDALimit},
        {"DISK_SECOND_ADVISORY_LIMIT",spinSecondDALimit},
        {"MAX_CONCURRENT_WO", spinMaxConWO},
        {"WORK_ORDER_SCHEDULER_INTERVAL", spinWOSchedInt},
        {"RESOURCE_MONITOR_INTERVAL", spinResMonInt},
        {"DELETION_INTERVAL", spinDeleteInt},
        {"CHECK_DISK_INTERVAL", spinCheckDiskInt},
        {"PSI_POLLING_INTERVAL", spinPSIPollInt},
        {"DCO_POLLING_INTERVAL", spinDCOPollInt},
        {"DSI_POLLING_INTERVAL", spinDSIPollInt},
        {"WORK_ORDER_EXPIRE_AGE", spinWOExpireAge}};

    numSysPrmFields = ARRAY_SIZE(tmpParms);

    for(int i = 0; i < numSysPrmFields; i++)
    {
        spParms[i] = tmpParms[i];
    }
}

//***********************************************************************
void ConfigureIAS::updateDbase()
{
    // This is a slot that is called when the OK button is clicked
    // It takes the IAS system configuration values from the spin boxes
    // and updates the SYS_PARMS database table accordingly.

    QString strSqlUpdate = "update SYS_PARMS set ";
    IASSqlQuery query;

    // Set up update string
    for(int i = 0; i < numSysPrmFields; i++)
    {
        strSqlUpdate += spParms[i].spField + " = ";
        if(!spParms[i].spinObject->suffix().compare("%"))
        {
            strSqlUpdate += QString::number(spParms[i].spinObject->value()*.01);
        }
        else
        {
            strSqlUpdate += QString::number(spParms[i].spinObject->value());
        }

        if(i != numSysPrmFields - 1)
        {
            strSqlUpdate += ",";
        }
    }
    strSqlUpdate += " where system_id = '" + m_systemId + "'";

    // Update all fields regardless if they have changed since there are
    // few fields to update
    if(query.exec(strSqlUpdate, __FILE__, __LINE__))
    {
        close();
    }
}

