/******************************************************************************
               Image Assessment System (IAS) Software
 ******************************************************************************
 *
 * Name: CreateNewProcedure.cpp
 *
 * PURPOSE:
 * This is the definition of the CreateNewProcedure class which is used to
 * create new IAS test procedures.
 *
 * NOTES: This class uses two qstringlists to temporarily store information
 * about edits made to parameters for each script.  One stringlist 
 * (m_changedParameters) stores information containing all of the contents
 * (every column) of the parameters table (tblParms) for parameters which were
 * modified.  The format of the data in the qstringlist is as follows:
 *
 * TODO - FIXME:
 *    If this code is ever improved upon, it should be modified to use defines
 *    for specifying column instead of using number.
 *
 *-----------------------------------------------------------------------------
 *      m_changedScripts (qstringlist)
 *
 *      <script name> 
 *      <number of changed parameters>
 *      <separated list of modified rows>
 *   ... repeated for each modified script
 * 
 * example contains two scripts, the first RPS_r0c has 3 modified parameters,
 *     they are in rows 0, 3, and 10.  The second script RPS_r0r has one
 *     modified parameter, it is the one in row 5
 * 
 *  --------------
 * | RPS_r0c      | <-- modified script's name (RPS_r0c)
 * | 3            | <-- number of modified parameters in script (3)
 * | 0#3#10       | <-- the rows in the parms table for each parameter
 * | RPS_r0r      | <-- another modified script (RPS_r0r)
 * | 1            | <-- this script (RPS_r0r) has one modified parameter
 * | 5            | <-- and it is row index 5 (6th row 0 based) of the parms
 *  --------------      table
 *
 *----------------------------------------------------------------------------
 *
 * m_changedParameters (qstringlist)
 *      
 *      <script name>
 *      <script name_row#>
 *      <numerical row value>
 *      <value of col 0 of row>
 *      <value of col 1 of row>
 *      <...>
 *      <value of col n of row>
 *      <script name_row#>
 *      <numerical row value>
 *      <value of col 0 of row>
 *      <value of col 1 of row>
 *      <...>
 *      <value of col n of row>
 *      <script name_row#>
 *      <numerical row value>
 *      <value of col 0 of row>
 *      <value of col 1 of row>
 *      <...>
 *      <value of col n of row>
 *      <script name>
 *      <script name_row#>
 *      <numerical row value>
 *      <value of col 0 of row>
 *      <value of col 1 of row>
 *      <...>
 *      <value of col n of row>
 *      
 *----------------------------------------------------------------------------
 *
 *  The class also maintains a record of global parameter values that have been
 *  modified.  This information is stored in four qstringlists.  The global
 *  parameter name, parameter value, the parameter element (occurrence),
 *  required flag, and attribute values are stored for all modified global 
 *  parameters.  The qstringlist names are
 *
 *      m_globalParameterNames
 *      m_globalParameterElements
 *      m_globalParameterValues
 *      m_globalParameterRequired
 *      m_globalParameterAttributes
 *
 *  The information contained in these qstringlists is parallel in order.
 *  The first modified global parameter's name is in the first location in
 *  the m_globalParameterNames qstringlist, its occurrence is in the first
 *  location in the m_globalParameterElements qstringlist, its value in the
 *  first location in m_globalParameterValues, its required value flag
 *  is in the first location of qstringlist m_globalParameterRequired, and
 *  its attribute is in the first location of qstringlist 
 *  m_globalParameterAttributes.
 * 
 *****************************************************************************/
 
// System Includes
#include <iostream>

// Qt Includes 
#include <QtCore>
#include <QtGui>
#include <QtSql>

// Local Includes
#include "CreateNewProcedure.h"
#include "IASSqlQuery.h"
#include "IASErrorDialog.h"
#include "IASSchemaConfig.h"
#include "IASParameterDelegate.h"
#include "IASParameterModel.h"

//****************************************************************************
CreateNewProcedure::CreateNewProcedure(
    QWidget* parent,        // I: The address to this widget's parent
    Qt::WindowFlags fl)     // I: Used by QT
: QWidget(parent, fl)
{
    QWidget::setAttribute(Qt::WA_DeleteOnClose);

    // Set up the user interface.
    setupUi(this);

    // New Procedure name has not been validated
    m_validNewProcedureName = false;

    // Set value to indicate no previous row selected
    m_previousScriptsTblRow = -1;

    // Set up the scripts qtable
    setupScriptsTable();

    // Set number of scripts in scripts table to 0
    m_NumScripts = 0;

    // Set up the parameters qtable
    setupParmsTable();

    // Initiate a database transaction
    QSqlDatabase::database().transaction();

    // Fill Sats and Sensors combobox with list of available sats and sensors 
    fillSatsAndSensorsList();

    // Set flag ignore to false, it is used in the function 
    // verifyScriptSelectionChanged.  When false, it will only refresh the
    // parameters table when a new script has been selected in the scripts
    // qtable.  When true it will refresh the parameters qtable wether or not
    // a new script is selected.
    ignore = false;

    // disable the ok pushbutton
    pbOK->setEnabled(false);

    m_currentParmVal = QString("");

    // Setup Fonts in combo box items here because they keep getting
    // reset by QT designer
    QFont cboSelectProcToCopyFrom_font(cboSelectProcToCopyFrom->font());
    cboSelectProcToCopyFrom_font.setItalic(false);
    cboSelectProcToCopyFrom_font.setBold(false);
    cboSelectProcToCopyFrom->setFont(cboSelectProcToCopyFrom_font);

    QFont cboInsertScript_font(cboInsertScript->font());
    cboInsertScript_font.setItalic(false);
    cboInsertScript_font.setBold(false);
    cboInsertScript->setFont(cboInsertScript_font);

    QFont cboCopyDefaultParmVal_font(cboCopyDefaultParmVal->font());
    cboCopyDefaultParmVal_font.setItalic(false);
    cboCopyDefaultParmVal_font.setBold(false);
    cboCopyDefaultParmVal->setFont(cboCopyDefaultParmVal_font);

    QFont cboSatSensor_font(cboSatSensor->font());
    cboSatSensor_font.setItalic(false);
    cboSatSensor_font.setBold(false);
    cboSatSensor->setFont(cboSatSensor_font);

    // Connect all the signals and slots.
    connect(pbOK, SIGNAL(pressed()),
            this, SLOT(writeNewProceduretoDatabase()));
    connect(pbDelete, SIGNAL(pressed()), this, SLOT(deleteScript()));
    connect(pbCancel, SIGNAL(pressed()), this, SLOT(close()));

    connect(leNewProdId, SIGNAL(returnPressed()),
            this, SLOT(validateProcedureName()));
    connect(cboSelectProcToCopyFrom, SIGNAL(activated(int)),
            this, SLOT(copyFromProcedure()));
    connect(cboCopyDefaultParmVal, SIGNAL(activated(const QString&)),
            this, SLOT(updateParametersWithDefaultValues(const QString&)));
    connect(cboInsertScript, SIGNAL(activated(int)),
            this, SLOT(selectScriptToInsert()));

    connect(leNewProcDesc, SIGNAL(returnPressed()),
            this, SLOT(descReturnPressed()));

    connect(pModel, SIGNAL(itemDataChanged(const QModelIndex&)),
            this, SLOT(modifiedParmCell(const QModelIndex&)));
    connect(tblParms, SIGNAL(entered(const QModelIndex&)),
            this, SLOT(enteredParmCell(const QModelIndex&)));

    connect(tblScripts, SIGNAL(pressed(const QModelIndex&)),
            this, SLOT(verifyScriptSelectionChanged(const QModelIndex&)));
    connect(cboSatSensor, SIGNAL(activated(const QString&)),
            this, SLOT(sensorSelected()));
}

//****************************************************************************
void CreateNewProcedure::setupScriptsTable()
{
    // Creates the scripts qtable, sets all of the initial values of the table

    int widths[] = {55,175,100,-1,-1,-1,-1,-1};
    QFont f = tblScripts->font();
    f.setWeight(QFont::Bold);

    // Display a popup menu when the user presses the right button
    // to allow moving a script up or down within the table
    //------------------------------------------------------------
    tblScripts->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(tblScripts, SIGNAL(customContextMenuRequested(const QPoint&)),
                        SLOT(showPopupMenu(const QPoint&)));

    // set no row currently selected
    m_currentScriptsTblRow = -1;

    // Make headings bold and set focus and selection modes
    tblScripts->horizontalHeader()->setFont(f);
    tblScripts->setSelectionMode(QAbstractItemView::SingleSelection);
    tblScripts->setSelectionBehavior(QAbstractItemView::SelectRows);

    // Remove the vertical header
    tblScripts->verticalHeader()->hide();

    // Create the model to use for the table.
    sModel = new IASParameterModel;
    sModel->setColumnCount(8);

    // Set the header labels.
    QStringList labels;
    labels << "Seq." << "Get Def. Value?" << "Script Id" << "procedure"
           << "parms_mod" << "script desc." << "pause flag"
           << "script_updated";
    sModel->setHorizontalHeaderLabels(labels);

    // Set the model.
    tblScripts->setModel(sModel);

    // Set column widths and whether column is editable or not
    for (int col = 0; col < sModel->columnCount(); col++)
    {
        tblScripts->setColumnWidth(col, widths[col]);
    }

    // Set last column displayed to be stretchable
    tblScripts->horizontalHeader()->setResizeMode(2, QHeaderView::Stretch);

    // Hide columns that shouldn't be displayed
    tblScripts->hideColumn(3);  // procedure
    tblScripts->hideColumn(4);  // parameter modified
    tblScripts->hideColumn(5);  // script description
    tblScripts->hideColumn(6);  // pause flag
    tblScripts->hideColumn(7);  // script updated

    // Set up the delegate for the table.
    tblScripts->setItemDelegate(new IASParameterDelegate);
}

//***********************************************************************
void CreateNewProcedure::setupParmsTable()
{
    // Sets up columns and properties of the script parameters qtable

    int widths[] = {185,60,90,80,60,120,-1,-1,-1,-1,80};
    QFont f = tblParms->font();
    f.setWeight(QFont::Bold);

    // Make headings bold and set focus and selection
    tblParms->horizontalHeader()->setFont(f);
    tblParms->setSelectionMode(QAbstractItemView::SingleSelection);
    tblParms->setSelectionBehavior(QAbstractItemView::SelectRows);

    // Remove the vertical header
    tblParms->verticalHeader()->hide();

    pModel = new IASParameterModel;
    pModel->setColumnCount(11);

    QStringList labels;
    labels << "Parm Name" << "Element" << "Parm Type" << "Common"
           << "Required" << "Parm Value" << "Parm Modified" << "Parm Updated"
           << "Procedure" << "Script" << "Attribute";
    pModel->setHorizontalHeaderLabels(labels);

    // Set the model.
    tblParms->setModel(pModel);

    for (int col=0; col < pModel->columnCount(); col++)
    {
        // Set column widths
        tblParms->setColumnWidth(col, widths[col]);
    }

    // Hide the column that indicates a parameter was modified
    tblParms->hideColumn(6);

    // Hide the column that indicates a default value was entered
    tblParms->hideColumn(7);

    // Hide the column that indicates which procedure's script values
    // are copied from
    tblParms->hideColumn(8);

    // Hide the column that indicates which script values are copied from
    tblParms->hideColumn(9);

    // Set up the delegate for the table.
    tblParms->setItemDelegate(new IASParameterDelegate);

    // Set last column displayed to be stretchable
    tblParms->horizontalHeader()->setStretchLastSection(true);
}

//***********************************************************************
void CreateNewProcedure::insertParmValueItem()
{
    // Check to see if there are a fixed set of parameter values
    // and whether the value read from the database should be mapped to
    // a different string to be shown to the user.  A combobox will be
    // inserted if there is a fixed set of values for the parameter.

    IASSqlQuery query;
    QStringList ouiValues;
    QString strCurrentValue, strModified;
    QString currentParmName, lastParmName;
    QVariant data;

    lastParmName = "";
    // For each row in the parameters qtable
    for(int row = 0; row < pModel->rowCount(); row++)
    {
        IASParameterItem *pItem = pModel->getItem(row, 7);
        strModified = pItem->data().toString();

        pItem = pModel->getItem(row, 0);
        currentParmName = pItem->data().toString();

        if(strModified.compare("1") == 0)
        {
            // Save the current value from table
            pItem = pModel->getItem(row, 5);
            strCurrentValue = pItem->data().toString();

            // Only query the parm_list table if the parameter name is
            // different from the previous row (to cut down on queries)
            if((row == 0) || (currentParmName != lastParmName))
            {
                ouiValues.clear();

                QString sqlStmt = "select oui_value, program_value from "
                    + schema + "parm_list where"
                    + " parm_name = '" + currentParmName + "'";

                // Create a qstringlist with the possible values for the
                // parameter.
                if(query.exec(sqlStmt,__FILE__,__LINE__))
                {
                    while(query.next())
                    {
                        ouiValues += query.value(0).toString();
                    }
                }
                else
                {
                    QMessageBox::warning(this, "Error", 
                                 "Unable to get OUI value for parameter " 
                                 + currentParmName,
                                 QMessageBox::Ok,
                                 QMessageBox::NoButton,
                                 QMessageBox::NoButton);

                    return;
                }
            }
        }

        lastParmName = currentParmName;

        // Copy the present value of the parameter
        if(!ouiValues.isEmpty())
        {
            pItem = pModel->getItem(row, 5);
            data = pItem->data().toString();

            pItem->setEditable(true);
            pItem->setType(IASParameterItem::TextSelection);
            pItem->setOptions(ouiValues);
            pItem->setData(data);
        }
    }
}

//***********************************************************************
void CreateNewProcedure::insertParmAttributeItem()
{
    // Check to see if there are a fixed set of parameter values
    // and whether the value read from the database should be mapped to
    // a different string to be shown to the user.  A combobox will be
    // inserted if there is a fixed set of values for parameter ATTRIBUTE.

    IASSqlQuery query;
    QStringList ouiAttributeValues;
    QString strCurrentValue, strModified;
    QVariant data;

    // Get the possible attributes for an algorithm
    ouiAttributeValues.clear();
    QString sqlStmt = "select oui_value from "
        + schema + "parm_list where parm_name = 'ATTRIBUTE'";

    // Create a qstringlist with the possible values for the parameter
    if(query.exec(sqlStmt,__FILE__,__LINE__))
    {
        while(query.next())
        {
            ouiAttributeValues += query.value(0).toString();
        }
    }
    else
    {
        QMessageBox::warning(this, "Error", 
                     "Unable to get OUI value for parameter ATTRIBUTE",
                     QMessageBox::Ok,
                     QMessageBox::NoButton,
                     QMessageBox::NoButton);

        return;
    }

    // For each row in the parameters qtable
    for(int row = 0; row < pModel->rowCount(); row++)
    {
        IASParameterItem *pItem = pModel->getItem(row, 0);
        if(pItem->data().toString() == "ALGORITHMS")
        {
            pItem = pModel->getItem(row, 7);
            strModified = pItem->data().toString();

            pItem = pModel->getItem(row, 10);

            if(strModified.compare("1") == 0)
            {
                // Save the current value from table
                strCurrentValue = pItem->data().toString();
            }

            // Create a combo box in the cell if there is a set of values for
            // the parameter. Otherwise use the line edit that is already there.
            if(!ouiAttributeValues.isEmpty())
            {
                pItem->setEditable(true);
                pItem->setType(IASParameterItem::TextSelection);
                pItem->setOptions(ouiAttributeValues);
                data = strCurrentValue;
                pItem->setData(data);
            }
        }
        else
        {
            pItem = pModel->getItem(row, 10);
            pItem->setEditable(false);
        }
    }
}

//***********************************************************************
bool CreateNewProcedure::validateParm(
    int row) // I: row from parameter table whose values just changed
             //    and must be validated
{
    // Check the value of the parameter against the max/min in the
    // parm_edits table (if there is one).

    IASSqlQuery query;

    IASParameterItem *pItem = pModel->getItem(row, 0);

    QString sqlStmt = "select min_value, max_value from "
        + schema + "parm_edits where ";
        + "parm_name='" + pItem->data().toString() + "'";

    // Execute the SQL statement, if statement executed successfully, get the
    // returned data (query.next())
    if(query.exec(sqlStmt,__FILE__,__LINE__))
    {
        query.next();
    }
    else
    {
        QMessageBox::warning(this,
                             "Warning", 
                             "Unable to access parm_edits for parm : "
                             + pItem->data().toString(),
                             QMessageBox::Ok,
                             QMessageBox::NoButton,
                             QMessageBox::NoButton);

        return false;
    }

    pItem = pModel->getItem(row, 2);
    QString pType = pItem->data().toString();

    // If there is a max/min associated with the parameter check against
    // it.  Return true if the parameter is not an int or float just in case
    // a max/min is put in the database for a "char" parameter.
    if(!query.isNull(0)) // There is a max/min
    {
        bool validVal = false;
        int valPos = 0;

        pItem = pModel->getItem(row, 3);
        QString value = pItem->data().toString();

        if((pType == "int") || (pType == "long"))
        {
            QIntValidator valInt(query.value(0).toInt(),
                                 query.value(1).toInt(), this);
            validVal = (valInt.validate(value, valPos)
                        == QValidator::Acceptable);
        }
        else if(pType == "float")
        {
            QDoubleValidator valDbl(query.value(0).toDouble(),
                                    query.value(1).toDouble(), 20, this);
            validVal = (valDbl.validate(value, valPos)
                        == QValidator::Acceptable);
        }
        else
        {
            return true;
        }

        // If the new value is invalid, return an error message with
        // the min/max, type information for the parameter and return
        // false to prevent update with this value.
        if(!validVal)
        {
            QMessageBox::warning(this,
                     "Error", 
                     "The Parameter value is invalid, should be:\nType = " +
                     pType + "\n" +
                     "Max = " + query.value(0).toString() + "\n" +
                     "Min = " + query.value(1).toString(),
                     QMessageBox::Ok,
                     QMessageBox::NoButton,
                     QMessageBox::NoButton);

            return false;
        }
    }

    return true;
}

//***********************************************************************
void CreateNewProcedure::commitAndClose()
{
    // Commit all changes for this transaction and close dialogue.

    if (!QSqlDatabase::database().commit())
        QMessageBox::warning(this, 
                             "Error", 
                             "Commit of database changes failed",
                             QMessageBox::Ok,
                             QMessageBox::NoButton,
                             QMessageBox::NoButton);

    close();
}

//***********************************************************************
void CreateNewProcedure::rollbackAndClose()
{
    // Rollback all changes for this transaction and close dialogue

    if (!QSqlDatabase::database().rollback())
        QMessageBox::warning(this,
                             "Error",
                             "Rollback of database changes failed",
                             QMessageBox::Ok,
                             QMessageBox::NoButton,
                             QMessageBox::NoButton);

    close();
}

//***********************************************************************
bool CreateNewProcedure::fillProceduresList()
{
    // Fills the combo box with a list of procedures to be used to
    // copy scripts from for the new procedure

    IASSqlQuery query;

    QString sqlStmt = QString("Select module_id from ")
        + schema
        + "modules where mod_type"
        + "='PROCEDURE' order by module_id";

    // Execute query sqlStmt
    if(query.exec(sqlStmt,__FILE__,__LINE__))
    {
        // While another row is retrieved, insert new row into combobox
        while (query.next())
        {
            cboSelectProcToCopyFrom->addItem(query.value(0).toString());
        }
    }
    else
    {
        QMessageBox::warning(this, 
                         "Error", 
                         "Unable to select module_id's from the modules table",
                         QMessageBox::Ok,
                         QMessageBox::NoButton,
                         QMessageBox::NoButton);

        return false;
    }

    return true;
}

//***********************************************************************
void CreateNewProcedure::fillSatsAndSensorsList()
{
   // Fills the combo box with a list of satellites and sensors

    IASSqlQuery query;

    QString sqlStmtSensor =
        "select satellite, sensor from satellite_sensor_cfg";

    // Issue the query to get the satellites and sensors.  If unsuccessful,
    // warn the user.

    if (query.exec(sqlStmtSensor, __FILE__, __LINE__))
    {
        // While another row is retrieved, insert new row into combobox
        while (query.next())
        {
            cboSatSensor->addItem(query.value(0).toString() + // satellite
                            " " + query.value(1).toString()); // sensor
        }
    }
    else
    {
        QMessageBox::warning(this,
                            "Error",
                            "Error getting satellites and sensors from the "
                            "satellite_sensor_cfg table\n",
                            QMessageBox::Ok,
                            QMessageBox::NoButton,
                            QMessageBox::NoButton);
    }

    // See if the sensor is known by virtue of only 1 sensor being configured
    if (cboSatSensor->count() <= 1)
    {
        // If so select that sensor
        sensorSelected();

        // Enable the new procedure name line edit, so it can be changed
        leNewProdId->setEnabled(true);
        // Select the entire text in the line edit for entering the 
        //procedure name
        leNewProdId->selectAll();
    }
    else
    {
        // Don't have 1 sensor as a default if there are choices (be consistent
        // with other satellite/sensor screens)
        cboSatSensor->addItem("");

        // Otherwise, allow the operator to select the sensor 
        cboSatSensor->setEnabled(true);

        // Disable the new procedure name line edit, so it can't be changed
        leNewProdId->setEnabled(false);
    }
}

//****************************************************************************
void CreateNewProcedure::copyFromProcedure()
{
    // Slot: User selected a procedure from the combobox, and its scripts
    // need to be inserted into the new procedure.

    IASSqlQuery queryScripts;
    IASSqlQuery queryParmCount;
    QString sqlStmt = "";
    sqlStmt = "select seq_nbr, sub_module_id, nvl(sub_mod_desc,''), pause_flag";
    sqlStmt += " from " + schema + "sub_modules";
    sqlStmt += " where module_id = '" +
               cboSelectProcToCopyFrom->currentText() + "'";
    sqlStmt += " order by seq_nbr";

    QString sqlStmt2, strNewScriptId;
    bool foundProcedure = false;
    bool foundScript = false;

    if(m_validNewProcedureName == false)
    {
        QMessageBox::warning(this,
                             "Warning", 
                             "A valid procedure name must be entered",
                             QMessageBox::Ok,
                             QMessageBox::NoButton,
                             QMessageBox::NoButton);

        leNewProdId->setFocus();
        leNewProdId->selectAll();

        return;
    }

    IASParameterItem *sItem = NULL;

    // Determine if the procedure was already used as a template (its scripts
    // were already inserted into the new procedure).
    for(int row = 0; row < sModel->rowCount() && foundProcedure == false; row++)
    {
        sItem = sModel->getItem(row, 3);
        if(sItem->data().toString() == cboSelectProcToCopyFrom->currentText())
        {
            foundProcedure = true;
        }
    }

    // If the procedure was not already inserted, retrieve the scripts
    // in the procedure
    if(foundProcedure == false)
    {
        // Execute query sqlStmt, if query is successful, 
        // insert each row into the scripts table
        if(!queryScripts.exec(sqlStmt,__FILE__,__LINE__))
        {
            QMessageBox::warning(this,
                                 "Warning", 
                                 "Unable to access sub_modules table ",
                                 QMessageBox::Ok,
                                 QMessageBox::NoButton,
                                 QMessageBox::NoButton);

            return;
        }

        while(queryScripts.next())
        {
            foundScript = false;

            // Determine if the script is already in the procedure.
            // It can not be listed more than once.
            strNewScriptId = queryScripts.value(1).toString();

            for(int row = 0;
                row < sModel->rowCount() && foundScript == false;
                row++)
            {
                sItem = sModel->getItem(row, 2);
                if(strNewScriptId == sItem->data().toString())
                {
                    QMessageBox::warning(this, "Warning", 
                            QString("The script : ") + strNewScriptId +
                            ", is already contained in the new procedure, " +
                            leNewProdId->text() +
                            ", it will not be inserted again.",
                            QMessageBox::Ok,
                            QMessageBox::NoButton,
                            QMessageBox::NoButton);

                    foundScript = true;
                }
            }

            // If a new script is to be added, and it is not already present
            if(foundScript == false)
            {
                // Retrieve the number of parameters associated with the script
                sqlStmt2 =  "select count(*) from " + schema + "parameters ";
                sqlStmt2 += "where module_id = '" +
                            queryScripts.value(1).toString() + "'";

                if(queryParmCount.exec(sqlStmt2,__FILE__,__LINE__))
                {
                    queryParmCount.next();

                    IASParameterItemRow* sItemRow = new IASParameterItemRow;
                    sItem = NULL;
                    QVariant data;

                    /* Column = 0 */
                    // Insert the script value for sequence number
                    sItem = new IASParameterItem;
                    data = QString("%1").arg(m_NumScripts+1);
                    sItem->setData(data);
                    sItemRow->append(sItem);

                    /* Column = 1 */
                    // Insert the script value for 'get def. value?'  If
                    // parameters exist, insert NO, otherwise 'N/A'
                    sItem = new IASParameterItem;
                    if(queryParmCount.value(0).toInt() != 0)
                    {
                        data = "NO";
                    }
                    else
                    {
                        data = "N/A";
                    }
                    sItem->setData(data);
                    sItemRow->append(sItem);

                    /* Column = 2 */
                    // Insert the script value for script name
                    sItem = new IASParameterItem;
                    data = queryScripts.value(1).toString();
                    sItem->setData(data);
                    sItemRow->append(sItem);

                    /* Column = 3 */
                    // Insert the name of the procedure the script was copied
                    // from
                    sItem = new IASParameterItem;
                    data = cboSelectProcToCopyFrom->currentText();
                    sItem->setData(data);
                    sItemRow->append(sItem);

                    /* Column = 4 */
                    // Insert a flag indicating the original parameters for the
                    // script were not modified.  It will be changed to a "1"
                    // if the parameters are modified from the original ones
                    sItem = new IASParameterItem;
                    data = "0";
                    sItem->setData(data);
                    sItemRow->append(sItem);

                    /* Column = 5 */
                    // Set value for script description
                    sItem = new IASParameterItem;
                    if(queryScripts.value(2).toString().length() < 1)
                    {
                        data = "";
                    }
                    else
                    {
                        data = queryScripts.value(2).toString();
                    }
                    sItem->setData(data);
                    sItemRow->append(sItem);

                    /* Column = 6 */
                    // Insert the pause flag for the script 
                    sItem = new IASParameterItem;
                    data = queryScripts.value(3).toString();
                    sItem->setData(data);
                    sItemRow->append(sItem);

                    /* Column = 7 */
                    // Insert flag indicating all of the parameters for this
                    // script have been updated
                    sItem = new IASParameterItem;
                    data = "1";
                    sItem->setData(data);
                    sItemRow->append(sItem);

                    m_NumScripts++;

                    sModel->appendRow(sItemRow);
                }
                else
                {
                    QMessageBox::warning(this,
                                         "Warning",
                                         "Unable to access parameters table ",
                                         QMessageBox::Ok,
                                         QMessageBox::NoButton, 
                                         QMessageBox::NoButton);

                    return;
                }
            }
        }

        // Set the first row as selected.
        tblScripts->setCurrentIndex(sModel->index(0,0));
        scriptSelectionChanged();

        // Enable the push buttons.
        pbOK->setEnabled(true);
        pbDelete->setEnabled(true);
    }
}

//****************************************************************************
void CreateNewProcedure::scriptSelectionChanged()
{
    // When a different script in the parms qtable is selected, this procedure
    // is called

    int row, col, scriptsRow;
    IASSqlQuery queryParms;
    IASSqlQuery queryAttributes;
    QString sqlStmtAttr;
    QStringList attributes;        // list of all algorithm attributes
    QString algorithm;             // the algorithm in the parm_value field
    QStringList parmValueSegments; // algorithm name, and potentially
                                   // algorithm attribute as well
    QModelIndex sIndex = tblScripts->currentIndex();
    IASParameterItem *pItem = NULL;
    IASParameterItemRow* pItemRow = NULL;

    if(!sIndex.isValid())
    {
        return;
    }

    scriptsRow = sIndex.row();

    QString sqlStmt =
        QString("Select parm_name, parm_occurrence, parm_type, ") +
                " global, required, parm_value from " +
                schema + "def_parms where module_id = '";

    IASParameterItem *sItem = sModel->getItem(scriptsRow, 3);
    sqlStmt += sItem->data().toString() + "' and sub_module_id = '";

    sItem = sModel->getItem(scriptsRow, 2);
    sqlStmt += sItem->data().toString() +
               "' order by parm_name, parm_occurrence";

    // If the user modified parameters for the script previously selected,
    // those values will have to be saved.  A flag for the row (script) just
    // exited will be a 1 if any parameter had changed

    // Save parameter values that have changed
    saveScriptParmValues();

    // Delete all rows from parameters table
    pModel->deleteData();

    row = 0;
    col = 0;

    // Select parameters from the def_parms table for the newly selected
    // script in the scripts qtable
    if(queryParms.exec(sqlStmt,__FILE__,__LINE__))
    {
        // Here is some preparation for later use to do once before the 
        // loop.  Get a list of all algorithm attributes.

        // Get the list of possible attribute values.
        sqlStmtAttr = QString("select oui_value from ") +
                                    schema +
                                    "parm_list where parm_name = 'ATTRIBUTE'";

        if(queryAttributes.exec(sqlStmtAttr,__FILE__,__LINE__))
        {
            // get next parameter from database cursor
            while (queryAttributes.next())
            {
                attributes << queryAttributes.value(0).toString();
            }
        }
        else
        {
            return;
        }

        // While there are more parameters for script in cursor,
        // insert them into the parameters qtable
        QStringList trueFalseOptions;
        trueFalseOptions << "T" << "F";
        QVariant data;
        while(queryParms.next())
        {
            pItemRow = new IASParameterItemRow;
            for(col = 0; col < pModel->columnCount(); col++)
            {
                pItem = new IASParameterItem;
                // Create qtable items
                switch(col)
                {
                    case 3: // Common column
                    case 4: // Required column
                    {
                        data = "F";
                        if(queryParms.value(col).toInt() == 1)
                        {
                          data = "T";
                        }
                        pItem->setEditable(true);
                        pItem->setType(IASParameterItem::TextSelection);
                        pItem->setOptions(trueFalseOptions);
                        break;
                    }

                    case 5: // Parameter Value column
                    {
                        data = queryParms.value(col).toString();
                        pItem->setEditable(true);
                        break;
                    }

                    case 7: /* (hidden) Indicates a default value was entered */
                    {
                        data = "1";
                        break;
                    }

                    // Default these to empty string.
                    case 6: /* (hidden) Is set if the parameter is modified */
                    case 8: /* (hidden) The procedure the script values are
                                        copied from */
                    case 9: /* (hidden) The script in the procedure where the
                                        values are copied from */
                    {
                        data = "";
                        break;
                    }

                    default: // All other columns
                    {
                        data = queryParms.value(col).toString();
                        break;
                    }
                }

                pItem->setData(data);
                pItemRow->append(pItem);
            }

            pModel->appendRow(pItemRow);

            // If the Parm Name is ALGORITHMS, get the attribute value
            if(queryParms.value(0).toString() == "ALGORITHMS")
            {
                // Since tblParms is initially populated from the def_parms
                // table, the parm_value could have the attribute appended
                // to it.  Take off the attribute if present for this query.

                pItem = pModel->getItem(row, 5);
                QString tmp = pItem->data().toString();

                // find if an attribute is at the end of the parm_value
                parmValueSegments = tmp.split("_");

                // this last element is an attribute if it's in the attribute
                // list
                if(attributes.contains(parmValueSegments.last()))
                {
                    // Remove the last element (the attribute) from the list
                    parmValueSegments.removeLast();

                    // Form the algorithm name from the rest of the parm_value
                    // segments
                    algorithm = parmValueSegments.join ("_");
                }
                else
                {
                    // The algorithm is just the original string
                    algorithm = tmp;
                }

                QVariant data = algorithm;
                pItem->setData(data);

                sqlStmtAttr = QString("select attribute_value from ")
                    + schema + "parm_attributes where parm_attribute = '"
                    + algorithm + "'";

                if (queryAttributes.exec(sqlStmtAttr,__FILE__,__LINE__))
                {
                    // get next parameter from database cursor
                    queryAttributes.next();

                    pItem = pModel->getItem(row, 10);

                    data = queryAttributes.value(0).toString();
                    pItem->setData(data);
                }
            }

            row++;
        }

        // If no rows were found for this script, most likely because it
        // was inserted manually and not copied from a procedure, query
        // the parameters table for parameters for this script
        if(row == 0)
        {
            sItem = sModel->getItem(scriptsRow, 2);

            sqlStmt = QString("select parm_name, parm_occurrence, parm_type ")
                + " from " + schema + "parameters where module_id = '" 
                + sItem->data().toString()
                + "' order by parm_name, parm_occurrence";

            // select parameters from the parameters table for the selected
            // script in the scripts qtable
            if(queryParms.exec(sqlStmt,__FILE__,__LINE__))
            {
                // While there are more parameters for script in cursor,
                // insert them into the parameters qtable
                while(queryParms.next())
                {
                    pItemRow = new IASParameterItemRow;

                    /* 0 */
                    pItem = new IASParameterItem;
                    data = queryParms.value(0).toString();
                    pItem->setData(data);
                    pItemRow->append(pItem);

                    /* 1 */
                    pItem = new IASParameterItem;
                    data = queryParms.value(1).toString();
                    pItem->setData(data);
                    pItemRow->append(pItem);

                    /* 2 */
                    pItem = new IASParameterItem;
                    data = queryParms.value(2).toString();
                    pItem->setData(data);
                    pItemRow->append(pItem);

                    /* 3 */
                    pItem = new IASParameterItem;
                    data = "F";
                    pItem->setEditable(true);
                    pItem->setType(IASParameterItem::TextSelection);
                    pItem->setOptions(trueFalseOptions);
                    pItem->setData(data);
                    pItemRow->append(pItem);

                    /* 4 */
                    pItem = new IASParameterItem;
                    data = "F";
                    pItem->setEditable(true);
                    pItem->setType(IASParameterItem::TextSelection);
                    pItem->setOptions(trueFalseOptions);
                    pItem->setData(data);
                    pItemRow->append(pItem);

                    /* 5 */
                    pItem = new IASParameterItem;
                    data = "";
                    pItem->setData(data);
                    pItemRow->append(pItem);

                    /* 6 */
                    pItem = new IASParameterItem;
                    data = 0;
                    pItem->setData(data);
                    pItemRow->append(pItem);

                    /* 7 */
                    pItem = new IASParameterItem;
                    data = 0;
                    pItem->setData(data);
                    pItemRow->append(pItem);

                    /* 8 */
                    pItem = new IASParameterItem;
                    data = "";
                    pItem->setData(data);
                    pItemRow->append(pItem);

                    /* 9 */
                    pItem = new IASParameterItem;
                    data = "";
                    pItem->setData(data);
                    pItemRow->append(pItem);

                    /* 10 */
                    pItem = new IASParameterItem;
                    data = "";
                    pItem->setData(data);
                    pItemRow->append(pItem);

                    pModel->appendRow(pItemRow);

                    row++;
                }
            }
            else
            {
                QMessageBox::warning(this, 
                                     "Warning", 
                                     "Unable to access the parameters table",
                                     QMessageBox::Ok,
                                     QMessageBox::NoButton,
                                     QMessageBox::NoButton);

                return;
            }
        }
    }
    else
    {
        QMessageBox::warning(this,
                             "Warning",
                             "Unable to access the def_parms table",
                             QMessageBox::Ok,
                             QMessageBox::NoButton,
                             QMessageBox::NoButton);

        return;
    }

    // If script parameters were previously modified, place stored values
    // into table
    sItem = sModel->getItem(scriptsRow, 4);
    if(sItem->data().toString().compare("1") == 0)
    {
        sItem = sModel->getItem(scriptsRow, 2);
        setParmValuesTexttoStoredValues(sItem->data().toString());
    }

    // Update parameters with any global variables that have changed
    setParmGlobalValues();

    insertParmValueItem();
    insertParmAttributeItem();

    // Save script row
    m_previousScriptsTblRow = scriptsRow;

    if(pModel->rowCount() > 0)
    {
        tblParms->setCurrentIndex(pModel->index(0, 0));

        // Set values of Procedure and Script from which parameters were copied
        pItem = pModel->getItem(0, 8);
        leCopyFromProcedure->setText(pItem->data().toString());
        pItem = pModel->getItem(0, 9);
        leCopyFromScript->setText(pItem->data().toString());
    }

    tblScripts->setCurrentIndex(sModel->index(scriptsRow, 0));
}

//****************************************************************************
void CreateNewProcedure::deleteScript()
{
    // Slot: Delete the selected script from the scripts qtable, and
    // parameters from the parameters qtable

    QModelIndex sIndex = tblScripts->currentIndex();

    // If no row is currently selected in the scripts qtable
    if(!sIndex.isValid())
    {
        return;
    }

    // Decrement the number of scripts in the list
    m_NumScripts--;

    // Remove all the rows from the parameter table.
    pModel->deleteData();

    // If there are remaining rows, update the sequence numbers in the
    // scripts qtable.
    if(sModel->rowCount() > 1)
    {
        delScriptAndUpdateSeqNbrs(sIndex);
    }

    // If there are no scripts left, disable the delete script button.
    if(sModel->rowCount() == 0)
    {
        pbDelete->setEnabled(false);
    }
    else 
    {
        // Reselect the previously selected cell
        tblScripts->setCurrentIndex(sIndex);
        scriptSelectionChanged();
    }

    // If the scripts table was disabled and all of the scripts have been
    // updated, enable the scripts table
    if(!tblScripts->isEnabled())
    {
        for(int i = 0; i < sModel->rowCount(); i++)
        {
            IASParameterItem *sItem = sModel->getItem(i, 7);
            if(sItem->data().toString().compare("0"))
            {
                cboCopyDefaultParmVal->setFocus();

                // Set the selection for the script
                tblScripts->setCurrentIndex(sModel->index(i, 0));

                return;
            }
        }

        tblScripts->setEnabled(true);
    }
}

//****************************************************************************
void CreateNewProcedure::delScriptAndUpdateSeqNbrs(const QModelIndex& sIndex)
{
    // Decrements the sequence numbers (execution order) in the scripts qtable
    // after the specified row.

    QString strSeqNbr;
    QVariant data;
    int row = sIndex.row();
    int seqNbr = 0;
    int scriptCount = 0;

    IASParameterItem *sItem = sModel->getItem(row, 0); /* The seqNbr */
    seqNbr = sItem->data().toInt();

    // Set to NULL, since the deleteRow frees the memory.
    sItem = NULL;
    sModel->deleteRow(row); // The script gets deleted here.

    // Grab the remaining number of scripts.
    scriptCount = sModel->rowCount();

    // Update the remaining script's sequence numbers.
    for(int cRow = row; cRow < scriptCount; cRow++)
    {
        sItem = sModel->getItem(cRow, 0); // Grab the sequence item.
        data = seqNbr; // Set to the current value of seqNbr.
        sItem->setData(data); // Set the items value.
        seqNbr++; // Increment to the next value.
    }
}

//****************************************************************************
void CreateNewProcedure::fillScriptsList()
{
    // Fills the add scripts combobox with a list of scripts available to be
    // inserted into the new procedure

    bool  foundScript;
    QString strScript;
    QString sqlStmt = QString("Select distinct module_id from ")
        + schema
        + "modules where mod_type = 'SCRIPT' order by module_id";
    IASSqlQuery query;

    // Execute query sqlStmt
    if(query.exec(sqlStmt,__FILE__,__LINE__))
    {
        while(query.next())
        {
            foundScript = false;

            // Determine if the script is already in the procedure,
            // it can not be listed more than once
            strScript = query.value(0).toString();

            // Search through the list to determine if the script is
            // already present
            for(int row = 0;
                row < sModel->rowCount() && foundScript == false;
                row++)
            {
                IASParameterItem *sItem = sModel->getItem(row, 2);
                if(strScript.compare(sItem->data().toString()) == 0)
                    foundScript = true;
            }

            // If the script is not already in the scripts qtable,
            // add it to the list
            if(!foundScript)
            {
                cboInsertScript->addItem(strScript);
            }
        }
    }
    else
    {
        QMessageBox::warning(this, "Error",
                "Unable to access modules table to retrieve a list of scripts",
                QMessageBox::Ok,
                QMessageBox::NoButton,
                QMessageBox::NoButton);
    }
}

//****************************************************************************
void CreateNewProcedure::selectScriptToInsert()
{
    // Slot:  The operator selected a procedure from the combobox. 
    // Its scripts need to be inserted into the new procedure.
    // This is for when the operator selects a script to insert.

    bool foundScript = false;
    int seqNbr = 0;
    QString strNewScriptId = cboInsertScript->currentText();
    QString sqlStmt2 = "SELECT count(*) from "
        + schema
        + "PARAMETERS WHERE module_id = '"
        + strNewScriptId  + "'";
    IASSqlQuery queryParmCount;

    IASParameterItem *sItem = NULL;
    IASParameterItemRow *sItemRow = NULL;

    QModelIndex sIndex = tblScripts->currentIndex();

    int currentRow = sIndex.row();

    // If all parameters in the current row aren't updated, warn user
    // and do not insert another script
    if(!verifyParmUpdate())
    {
        sItem = sModel->getItem(currentRow, 2);
        QMessageBox::warning(this,
                             "Warning",
                             "You must update the parameter values in script "
                             + sItem->data().toString()
                             + " before inserting more scripts",
                             QMessageBox::Ok,
                             QMessageBox::NoButton,
                             QMessageBox::NoButton);

        if(pModel->rowCount() > 0)
        {
            cboCopyDefaultParmVal->setFocus();
        }

        return;
    }

    if(m_validNewProcedureName == false)
    {
        QMessageBox::warning(this,
                             "Warning", 
                             "A valid procedure name must be entered",
                             QMessageBox::Ok,
                             QMessageBox::NoButton,
                             QMessageBox::NoButton);

        leNewProdId->setFocus();
        leNewProdId->selectAll();

        return;
    }

    // Determine if the script is already in the procedure, it can not be
    // listed more than once
    for(int row = 0;
        row < sModel->rowCount() && foundScript == false;
        row++)
    {
        sItem = sModel->getItem(row, 2);
        if(strNewScriptId.compare(sItem->data().toString()) == 0)
        {
            QMessageBox::warning(this, 
                             "Warning", 
                             QString("The script : ") + strNewScriptId
                             + ", is already contained in the new procedure, "
                             + leNewProdId->text()
                             + ", it will not be inserted again.",
                             QMessageBox::Ok,
                             QMessageBox::NoButton,
                             QMessageBox::NoButton);

            foundScript = true;

            return;
        }
    }

    // If current row = -1, no row is selected,
    // insert new script at the end of procedure
    if(currentRow < 0)
    {
        // If there are no rows in the Qtable, set current row to 0
        if(sModel->rowCount() == 0)
        {
            currentRow = 0;
        }
        // Set current row to last row in table (-1 because 0 based index)
        else
        {
            currentRow = sModel->rowCount() - 1;
        }
    }

    // Get the current sequence number, if there are no rows in procedure
    // already (no rows in scripts qtable) set sequence number to 1
    if(sModel->rowCount() > 0)
    {
        sItem = sModel->getItem(currentRow, 0);
        seqNbr = sItem->data().toInt();
    }
    else
    {
        seqNbr = 1;
    }

    // Determine if the script has any parameters associated with it
    if(queryParmCount.exec(sqlStmt2,__FILE__,__LINE__))
    {
        queryParmCount.next();

        sItemRow = new IASParameterItemRow;

        sItem = new IASParameterItem;
        QVariant data = seqNbr;
        sItem->setData(data);
        sItemRow->append(sItem);

        // Insert the value for get default value, if the script has
        // parameters insert a value of 'YES' otherwise 'N/A'.  Insert YES
        // because there isn't both a procedure and a script, there are no
        // default values to insert
        sItem = new IASParameterItem;
        if(queryParmCount.value(0).toInt() != 0)
        {
            data = "YES";

            cboCopyDefaultParmVal->setFocus();

            // Disable the select script combo box
            cboInsertScript->setEnabled(false);

            // Disable the scripts qtable
            tblScripts->setEnabled(false);

            // Disable the copy from Procedure combo box
            cboSelectProcToCopyFrom->setEnabled(false);

            QMessageBox::information(this, "Information", 
                    QString("You must now select a procedure/script to copy "
                            "values from ") +
                    "\n for script " + strNewScriptId + " before continuing", 
                    QMessageBox::Ok,
                    QMessageBox::NoButton,
                    QMessageBox::NoButton);
        }
        else
        {
            data = "N/A";
        }
        sItem->setData(data);
        sItemRow->append(sItem);

        // Insert the script ID
        sItem = new IASParameterItem;
        data = strNewScriptId;
        sItem->setData(data);
        sItemRow->append(sItem);

        // Insert an empty value for procedure, not copied from another
        // procedure
        sItem = new IASParameterItem;
        data = "";
        sItem->setData(data);
        sItemRow->append(sItem);

        // Insert empty value for script modified, it hasn't been changed yet
        // just added, nothing to save
        sItem = new IASParameterItem;
        data = "0";
        sItem->setData(data);
        sItemRow->append(sItem);

        // Insert empty value for script description,
        // script not copied from another procedure
        sItem = new IASParameterItem;
        data = "";
        sItem->setData(data);
        sItemRow->append(sItem);

        // Insert empty value for pause flag, script not copied
        // from another procedure
        sItem = new IASParameterItem;
        data = "0";
        sItem->setData(data);
        sItemRow->append(sItem);

        // Insert value to indicate the script has not been updated
        // from another procedure
        sItem = new IASParameterItem;
        data = "0";
        sItem->setData(data);
        sItemRow->append(sItem);

        m_NumScripts++;
    }

    // Insert a new row where the currently selected row is
    sModel->insertRow(sIndex, sItemRow);

    // Just go through all the rows and update the sequence numbers.
    // If we ever get so many rows that this doesn't make sense, then there
    // is also a major flaw in the system design.
    for(int row = 0; row < sModel->rowCount(); row++)
    {
        sItem = sModel->getItem(row, 0);
        QVariant data = row + 1;
        sModel->setItemData(row, 0, data);
    }

    // Set the currently selected cell to the new row and ensure it is visible
    tblScripts->setCurrentIndex(sIndex);

    ignore = true;
    scriptSelectionChanged();
    ignore = false;

    pbOK->setEnabled(true); 
    pbDelete->setEnabled(true);

    cboCopyDefaultParmVal->setEnabled(true);

    for(int i = 0; i < cboCopyDefaultParmVal->count(); i++)
    {
        int itemIndex = cboCopyDefaultParmVal->findText(strNewScriptId);
        if(itemIndex != -1)
        {
            cboCopyDefaultParmVal->setCurrentIndex(i);
            return;
        }
    }
}

//****************************************************************************
bool CreateNewProcedure::validateProcedureName()
{
    // Verify that the new procedure name entered by the user does not already
    // exist in the database, and is not too long to insert into the database

    QString strTemp = leNewProdId->text().simplified();
    QString strTemp2;
    QStringList satsAndSensors; // These are the satellites and sensors 
                                // supported in this configuration.

    IASSqlQuery query;
    QString sqlStmt = "select module_id from " + schema + "modules";
    sqlStmt += " where module_id = '" + strTemp + "'";
    sqlStmt += " and mod_type = 'PROCEDURE'";

    // Set line edit back to new value with white space removed
    leNewProdId->setText(strTemp);

    // If the new procedure name is too long, issue a warning
    if(strTemp.length() > 24)
    {
        m_validNewProcedureName = false;
        cboSelectProcToCopyFrom->setEnabled(false);
        cboInsertScript->setEnabled(false);
        QMessageBox::warning(this,
                             "Error",
                             "That Procedure name : "+ strTemp + 
                             ", is too long.\nMaximum length is 25 characters",
                             QMessageBox::Ok,
                             QMessageBox::NoButton,
                             QMessageBox::NoButton);
        leNewProdId->selectAll();
        return false;
    }
    else if(strTemp.length() > 0)
    {
        // Query database for the new procedure name
        if(query.exec(sqlStmt,__FILE__,__LINE__))
        {
            // Retrieve first value from the database
            query.next();

            // If a matching procedure name is found
            if(query.isValid() && (query.value(0).toString() == strTemp))
            {
                m_validNewProcedureName = false;
                cboSelectProcToCopyFrom->setEnabled(false);
                cboInsertScript->setEnabled(false);

                QMessageBox::warning(this, 
                                     "Error", 
                                     "A procedure with that name: "+ strTemp
                                     + ", already exists",
                                     QMessageBox::Ok,
                                     QMessageBox::NoButton,
                                     QMessageBox::NoButton);

                return false;
            }
            else    // No matching procedure name found
            {
                QString strRegExp = QString("[^\\w- ]");

                QRegExp rxp = QRegExp(strRegExp);
                if(strTemp.contains(rxp))
                {
                    QMessageBox::warning(this, "Error", 
                                 "Re-Enter procedure name, name is invalid\n"
                                 "Remove punctuation marks",
                                 QMessageBox::Ok,
                                 QMessageBox::NoButton,
                                 QMessageBox::NoButton);
                }
                else
                {
                    m_validNewProcedureName = true;

                    // Disable the new procedure name line edit, so it can't
                    // be changed
                    leNewProdId->setEnabled(false);

                    // Enable the description line edit
                    leNewProcDesc->setEnabled(true);

                    // Update heading with new procedure name
                    lblFormHeading->setText(lblFormHeading->text()
                                  + " : " + leNewProdId->text());

                    leNewProcDesc->setFocus();
                    leNewProcDesc->selectAll();

                    // Enable widgets so procedures can be inserted as 
                    // templates, scripts can be inserted, and default 
                    // values can be copied
                    cboSelectProcToCopyFrom->setEnabled(true);
                    cboInsertScript->setEnabled(true);
                    cboCopyDefaultParmVal->setEnabled(true);

                    return true;
                }
            }
        }
    }
    else
    {
        m_validNewProcedureName = false;
        cboSelectProcToCopyFrom->setEnabled(false);
        cboInsertScript->setEnabled(false);
        return false;
    }

    return true;
}

//****************************************************************************
void CreateNewProcedure::sensorSelected()
{
    // Slot: The sensor has been selected, or there is only one sensor in
    // the configuration so it by default has effectively been selected.
    // Activate the various later parts of the dialog

    QStringList satAndSensor; // list including satellite and sensor

    if (cboSatSensor->currentText().isEmpty())
    {
        return;
    }

    // Disable the sensor combobox, so it can't be changed
    cboSatSensor->setEnabled(false);

    // Enable the new procedure name line edit, so it can be changed
    leNewProdId->setEnabled(true);
    leNewProdId->setFocus();

    // Select the entire text in the line edit for entering the procedure name
    leNewProdId->selectAll();

    // Algorithm Attributes are not currently supported for L7 ETM.
    // This may change later.
    if(cboSatSensor->currentText() == "L7 ETM")
    {
        tblParms->hideColumn(10); // attributes
    }

    // Extract the satellite and sensor from the combobox choice
    satAndSensor = cboSatSensor->currentText().split(' ');

    // Look up the schema based on the satellite and sensor
    schema = IASSchemaConfig::instance()->schemaName(
        satAndSensor[0],   // satellite
        satAndSensor[1]);  // sensor

    // Fill "copy from procedures" combobox with list of procedures available
    fillProceduresList();

    // Fill scripts combobox with list of scripts available to insert
    fillScriptsList();

    // Fill the "Copy Default Parameters" combobox with list of procedures and
    // scripts available.
    fillCopyDefaultValuesList();
}

//****************************************************************************
void CreateNewProcedure::modifiedParmCell(const QModelIndex& pIndex)
{
    // Slot: The operator modified a value in the GUI parms table
    // connected to signal valueChanged(int,int)

    int row = pIndex.row();
    int column = pIndex.column();

    QVariant data;
    QModelIndex sIndex = tblScripts->currentIndex();

    IASParameterItem *pItem = pModel->getItem(row, column);

    // If the value when leaving the cell does not match the value when
    // the cell was entered, set flag indicating cell changes.
    // Set values of Procedure and Script from which parameters were copied

    // If the stored value is different from the current value.  The stored
    // value is set when a cell is entered.
    if(m_currentParmVal.compare(pItem->data().toString()) != 0)
    {
        // Verify that if the operator modified the value, it is a valid
        // value as listed in the parm_edits table.

        // If item is a table item (not a combo box item), we need to
        // verify it is within limits.  combo box items are filled only
        // with valid values so we do not need to check them.
        if(column == 5)
        {
            if(pItem->type() != IASParameterItem::TextSelection)
            {
                // If value is invalid, notify user and set value back to
                // value when focus entered this cell
                pItem = pModel->getItem(row, 2);
                if(pItem->data().toString().compare("float") == 0)
                {
                    pItem = pModel->getItem(row, column);
                    if(pItem->data().toString().contains(".") == 0)
                    {
                        QString strTemp = pItem->data().toString();
                        strTemp.append(".0");
                        data = strTemp;
                        pItem->setData(data);
                    }
                }

                if(checkParmMaxMin(row) == false)
                {
                    // Set value back to previous value
                    pItem = pModel->getItem(row, column);
                    data = m_currentParmVal;
                    pItem->setData(data);

                    return;
                }
            }
        }

        // Set flag indicating parameter value changed for this script
        IASParameterItem *sItem = sModel->getItem(sIndex.row(), 4);
        data = 1;
        sItem->setData(data);

        // Set flag indicating parameter value changed for this parameter
        pItem = pModel->getItem(row, 6);
        pItem->setData(data); // data is still set to 1 from above.
    }
}

//****************************************************************************
void CreateNewProcedure::enteredParmCell(const QModelIndex& pIndex)
{
    // Slot:Save the value of the parameters qtable cell at row, column into
    // a class variable.  This function is called when a signals has been
    // emitted that the cell was entered (current focus).  This function
    // saves the current value because the textChanged signal is always
    // emitted when a cell is left, even if no change occurred.  Function
    // modifiedParmCell() will check to see if actual entry changed.
    // Connected to signal tlbParms->selectionChanged(int,int)

    int row = pIndex.row();

    IASParameterItem *pItem = pModel->getItem(row, 5);

    QModelIndex sIndex = tblScripts->currentIndex();

    m_currentParmVal = pItem->data().toString();

    // Get the procedure the default values were copied from.
    // If not copied from a procedure, get procedure name from
    // the parms table

    pItem = pModel->getItem(row, 7);
    // If the row has been updated?
    if(pItem->data().toInt() == 1)
    {
        pItem = pModel->getItem(row, 8);
        // If the procedure name is not stored in tblParms
        if(pItem->data().toString().length() != 0)
        {
            leCopyFromProcedure->setText(pItem->data().toString());
            pItem = pModel->getItem(row, 9);
            leCopyFromScript->setText(pItem->data().toString());
        }
        else
        {
            IASParameterItem *sItem = sModel->getItem(sIndex.row(), 3);
            leCopyFromProcedure->setText(sItem->data().toString());

            sItem = sModel->getItem(sIndex.row(), 2);
            leCopyFromScript->setText(sItem->data().toString());
        }
    }
    else
    {
        leCopyFromProcedure->setText("");
        leCopyFromScript->setText("");
    }
}

//****************************************************************************
bool CreateNewProcedure::checkParmMaxMin(
    int row) // I: row whose value we need to check against its limits
{
    // Check the value of the parameter against the max/min in the  parm_edits
    // table (if limits exist).

    IASParameterItem *pItem = pModel->getItem(row, 0);

    IASSqlQuery query;
    QString sqlStmt = "select min_value, max_value";
    sqlStmt += " from " + schema + "parm_edits";
    sqlStmt += " where parm_name='" + pItem->data().toString() + "'";

    // Execute sql statement, retrieve min and max values for the specified row
    if(query.exec(sqlStmt,__FILE__,__LINE__))
    {
        query.next();
    }
    else
    {
        QMessageBox::warning(this, 
                             "Warning", 
                             "Unable to access parm_edits for parm : "
                             + pItem->data().toString()
                             + "\n"
                             + sqlStmt,
                             QMessageBox::Ok,
                             QMessageBox::NoButton,
                             QMessageBox::NoButton);
        return false;
    }

    // If there is a max/min associated with the parameter check against
    // it.  Return true if the parameter is not an int or float just in case
    // a max/min is put in the database for a "char" parameter.
    if(!query.isNull(0))   // There is a max/min
    {
        bool validVal = false;
        int valPos = 0;

        pItem = pModel->getItem(row, 2);
        QString type = pItem->data().toString();
        pItem = pModel->getItem(row, 5);
        QString value = pItem->data().toString();
        if((type == "int") || (type == "long"))
        {
            QIntValidator valInt(query.value(0).toInt(),
                                 query.value(1).toInt(), this);
            validVal = (valInt.validate(value, valPos)
                        == QValidator::Acceptable);
        }
        else if(type == "float")
        {
            QDoubleValidator valDbl(query.value(0).toDouble(),
                                    query.value(1).toDouble(), 20, this);
            validVal = (valDbl.validate(value, valPos) 
                        == QValidator::Acceptable);
        }
        else
        {
            return true;
        }

        if(!validVal)
        {
            QMessageBox::warning(this, "Error", 
                "The Parameter value is invalid, should be:\nType = " + 
                type + "\n" + 
                "Max = " + query.value(0).toString() + "\n" +
                "Min = " + query.value(1).toString() + "\n" + sqlStmt,
                QMessageBox::Ok,
                QMessageBox::NoButton,
                QMessageBox::NoButton);

            return false;
        }
    }

    return true;
}

//****************************************************************************
void CreateNewProcedure::writeNewProceduretoDatabase()
{
    // Slot: Write out new procedure to database

    // Disable the OK pushbutton
    pbOK->setEnabled(false);

    // Save new procedure if it contains at least one script
    if(sModel->rowCount() > 0)
    {
        // Save any values that may have changed but not been saved,
        // modified values are saved when another cell is selected,
        // simulate another cell being selected
        tblParms->setCurrentIndex(pModel->index(0, 0));

        // Save any changes made to the parameters for the current script.
        // Values for a modified script are saved when another script is
        // selected, simulate
        saveScriptParmValues();

        if(writeNewProceduretoModulesTable() != true)
        {
            if(!QSqlDatabase::database().rollback())
                QMessageBox::warning(this,
                                     "Error", 
                                     "Rollback of database changes failed",
                                     QMessageBox::Ok,
                                     QMessageBox::NoButton,
                                     QMessageBox::NoButton);

            // Enable the OK pushbutton
            pbOK->setEnabled(true);

            return;
        }

        if(writeNewProceduretoSubModulesTable() != true)
        {
            if(!QSqlDatabase::database().rollback())
                QMessageBox::warning(this,
                                     "Error", 
                                     "Rollback of database changes failed",
                                     QMessageBox::Ok,
                                     QMessageBox::NoButton,
                                     QMessageBox::NoButton);

            // Enable the OK pushbutton
            pbOK->setEnabled(true);

            return;
        }

        if(writeNewProceduretoDefParmsTable() != true)
        {
            if(!QSqlDatabase::database().rollback())
                QMessageBox::warning(this,
                                     "Error", 
                                     "Rollback of database changes failed",
                                     QMessageBox::Ok,
                                     QMessageBox::NoButton,
                                     QMessageBox::NoButton);

            // Enable the OK pushbutton
            pbOK->setEnabled(true);

            return;
        }

        commitAndClose();
    }
    else
    {
        QMessageBox::warning(this, "Warning", 
                "No scripts to save\n\nInsert scripts or select Cancel to End",
                QMessageBox::Ok,
                QMessageBox::NoButton,
                QMessageBox::NoButton);

        // Enable the OK pushbutton
        pbOK->setEnabled(true); 

        return;
    }

    close();
}

//****************************************************************************
bool CreateNewProcedure::writeNewProceduretoModulesTable()
{
    // Write out new procedure to the modules table in the database

    IASSqlQuery query;

    QString sqlStmt = 
        QString("insert into ")
        + schema
        + "modules (MODULE_ID,MOD_TYPE,MOD_STATUS,MOD_DESC)"
        + " values('" + leNewProdId->text() + "', 'PROCEDURE', 'TEST','"
        + leNewProcDesc->text() + "')";

    // Insert new procedure into the modules table.  If unsuccessful, 
    // warn the user.
    if (!query.exec(sqlStmt,__FILE__,__LINE__))
    {
        QMessageBox::warning(this,
                            "Error", 
                            "Error inserting new procedure into modules table\n"
                            + sqlStmt,
                            QMessageBox::Ok,
                            QMessageBox::NoButton,
                            QMessageBox::NoButton);
        return false;
    }

    return true;
}

//****************************************************************************

bool CreateNewProcedure::writeNewProceduretoSubModulesTable()
{
    // Write out new procedure to the sub_modules table in the database

    IASSqlQuery query;
    QString sqlStmt;

    // For each script in the GUI scripts qtable, 
    // write the script into the submodules table for the new procedure
    for(int row = 0; row < sModel->rowCount(); row++)
    {
        // If all of the parameters for this script have been updated
        // write out the script
        tblScripts->setCurrentIndex(sModel->index(row, 0));

        if(verifyParmUpdate())
        {
            IASParameterItem *sItem = sModel->getItem(row, 2);
            QString subModId = sItem->data().toString();
            sItem = sModel->getItem(row, 0);
            QString seqNbr = sItem->data().toString();
            sItem = sModel->getItem(row, 5);
            QString subModDesc = sItem->data().toString();
            sItem = sModel->getItem(row, 6);
            QString pauseFlag = sItem->data().toString();
            sqlStmt = QString("insert into ")
                + schema + "sub_modules (MODULE_ID,"
                + "SUB_MODULE_ID, SEQ_NBR, SUB_MOD_DESC, PAUSE_FLAG) "
                + "values('" +  leNewProdId->text() + "','"
                + subModId + "', " + seqNbr + ",'" + subModDesc + "'," 
                + pauseFlag + ")";

            // Warn user if insert unsuccessful
            if(!query.exec(sqlStmt,__FILE__,__LINE__))
            {
                QMessageBox::warning(this, "Error", 
                        "Problem inserting new scripts into submodules table\n"
                        + sqlStmt,
                        QMessageBox::Ok,
                        QMessageBox::NoButton,
                        QMessageBox::NoButton);

                return false;
            }
        }
        else
        {
            QMessageBox::warning(this, "Warning", 
                         "Not all parameters for this script have been updated",
                         QMessageBox::Ok,
                         QMessageBox::NoButton,
                         QMessageBox::NoButton);

            tblScripts->setCurrentIndex(sModel->index(row, 0));

            if(!QSqlDatabase::database().rollback())
            {
                QMessageBox::warning(this, "Error", 
                                    "Rollback of database changes failed",
                                     QMessageBox::Ok,
                                     QMessageBox::NoButton,
                                     QMessageBox::NoButton);
            }

            return false;
        }
    }

    return true;
}

//****************************************************************************
bool CreateNewProcedure::writeNewProceduretoDefParmsTable()
{
    // Write out parameters for the new procedure/scripts to the def_parms
    // table
    bool scriptModified;
    int mrow;
    QString sqlStmt;
    QString modId;
    QString scriptId;
    QString Attribute;    // in-database algorithm attribute like "_DUMP"
    QStringList strlParms;
    IASSqlQuery queryParms;
    IASSqlQuery queryExec;

    // For each script in the new procedure, query database and retrieve
    // the parameters 
    for(int row = 0; row < sModel->rowCount(); row ++)
    {
        IASParameterItem *sItem = sModel->getItem(row, 2);
        scriptId = sItem->data().toString();

        sItem = sModel->getItem(row, 4);
        if(sItem->data().toString().compare("1") == 0)
        {
            scriptModified = true;
        }
        else
        {
            scriptModified = false;
        }

        sItem = sModel->getItem(row, 3);
        modId = sItem->data().toString();

        // If there wasn't an original script to copy the values from
        // use the stored values
        if(!scriptModified)
        {
            sqlStmt = QString("Select parm_name, parm_occurrence,")
                + " parm_type, global, required, parm_value from "
                + schema + "def_parms where module_id = '";

            if(modId.length() < 1)
            {
                modId = leCopyFromProcedure->text();
            }
            sqlStmt += modId;

            sqlStmt += "' and sub_module_id = '" + scriptId
                + "' order by parm_name, parm_occurrence";

            if(!queryParms.exec(sqlStmt,__FILE__,__LINE__))
            {
                QMessageBox::warning(this,
                                     "Error",
                                     "Unable to retrieve parameters for module " 
                                     + modId
                                     + " and script "
                                     +  scriptId
                                     + "\n"
                                     + sqlStmt,
                                     QMessageBox::Ok,
                                     QMessageBox::NoButton,
                                     QMessageBox::NoButton);
                return false;
            }

            // The query was successful.  Insert each parameter into the
            // def_parms table for this new module and script
            while(queryParms.next())
            {
                sqlStmt = QString("insert into ")
                    + schema + "def_parms "
                    + " (module_id, sub_module_id,"
                    + " parm_name, parm_occurrence, parm_type, global,";

                if(queryParms.value(5).toString().length() >= 1) // parm_value 
                {
                    sqlStmt += " required, parm_value ) values ('";
                }
                else // no parm_value
                {
                    sqlStmt += " required ) values ('";
                }

                sqlStmt += leNewProdId->text() + "','"      // procedure name
                    + scriptId + "','"                      // script name
                    + queryParms.value(0).toString() + "'," // parm name
                    + queryParms.value(1).toString() + ",'" // parm occ
                    + queryParms.value(2).toString() + "'," // parm type
                    + queryParms.value(3).toString() + ","  // global
                    + queryParms.value(4).toString();       // required

                if(queryParms.value(5).toString().length() >= 1) // parm_value
                {
                    sqlStmt += ",'" + queryParms.value(5).toString() + "')";
                }
                else // no parm_value
                {
                    sqlStmt += ")";
                }
 
                // execute insert, if unsuccessful warn the user, and break
                if(!queryExec.exec(sqlStmt,__FILE__,__LINE__))
                {
                    QMessageBox::warning(this, "Error", 
                                 "Unable to insert parameters for module " +
                                 modId +
                                 " and script " +  scriptId +
                                 "\n" +
                                 sqlStmt,
                                 QMessageBox::Ok,
                                 QMessageBox::NoButton,
                                 QMessageBox::NoButton);

                    return false;
                }
            }

            // If this script was modified, then update the parameters
            // that were modified
            if(scriptModified)
            {
                while(getNextRow(false, scriptId, mrow))
                {
                    if(getParmValues(scriptId, mrow, strlParms))
                    {
                        getAttribute(strlParms[11], Attribute);

                        // Remove row number from the front of qstringlist
                        strlParms.pop_front();

                        sqlStmt = QString("update ")
                            + schema + "def_parms set global = "
                            + QString("%1").arg(
                                  ((strlParms[3].compare("T") == 0) ? 1 : 0))
                            + ", required = " 
                            + QString("%1").arg(
                                  ((strlParms[4].compare("T") == 0) ? 1 : 0));

                        if(strlParms[5].length() >= 1) // parm value
                        {
                            sqlStmt += ",  parm_value = '"
                                + strlParms[5] + Attribute + "' ";
                        }

                        sqlStmt += "where module_id = '" 
                            + leNewProdId->text()
                            + "' and sub_module_id = '" + scriptId 
                            + "' and parm_name = '" + strlParms[0]
                            + "' and parm_occurrence = " + strlParms[1];

                        if(!queryExec.exec(sqlStmt,__FILE__,__LINE__))
                        {
                            QMessageBox::warning(this, "Error", 
                                    "Unable to update parameters for procedure "
                                    + modId
                                    + " and script " 
                                    +  scriptId
                                    + "\n"
                                    + sqlStmt,
                                    QMessageBox::Ok,
                                    QMessageBox::NoButton,
                                    QMessageBox::NoButton);

                            return false;
                        }
                    }
                    else
                    {
                        QMessageBox::warning(this, "Error", 
                                   "Unable to update parameters for procedure "
                                   + modId
                                   + " and script "
                                   +  scriptId
                                   + "\n"
                                   + sqlStmt,
                                   QMessageBox::Ok,
                                   QMessageBox::NoButton,
                                   QMessageBox::NoButton);

                        return false;
                    }
                }
            }
        }
        else 
        {
            while(getNextRow(false, scriptId, mrow))
            {
                if(getParmValues(scriptId, mrow, strlParms))
                {
                    getAttribute(strlParms[11], Attribute);

                    // Remove row number from the front of qstringlist
                    strlParms.pop_front();

                    // Generate insert statement for this row (mrow)
                    sqlStmt = QString("insert into ")
                        + schema + "def_parms (module_id, sub_module_id,"
                        + " parm_name, parm_occurrence, parm_type, global,";

                    if(strlParms[5].length() >= 1) // parm value
                    {
                        sqlStmt += " required, parm_value ) values ('";
                    }
                    else // no parm value
                    {
                        sqlStmt += " required ) values ('";
                    }

                    sqlStmt += leNewProdId->text() + "','" // procedure name
                        + scriptId + "','"            // script name
                        + strlParms[0] + "',"          // parm name
                        + strlParms[1] + ",'"          // parm occ
                        + strlParms[2] + "',"          // parm type
                        + QString("%1").arg(
                              ((strlParms[3].compare("T") == 0) ? 1 : 0))
                        + ","                 // global
                        + QString("%1").arg(
                              ((strlParms[4].compare("T") == 0) ? 1 : 0));
                                              // required

                    if(strlParms[5].length() >= 1)    // parm value
                    {
                        sqlStmt += ",'" + strlParms[5] + Attribute + "')";
                    }
                    else                               // no parm value
                    {
                        sqlStmt += ")";       
                    }

                    if(!queryExec.exec(sqlStmt,__FILE__,__LINE__))
                    {
                        QMessageBox::warning(this, "Error", 
                                   "Unable to insert parameters for procedure "
                                   + modId
                                   + " and script "
                                   +  scriptId
                                   + "\n"
                                   + sqlStmt,
                                   QMessageBox::Ok,
                                   QMessageBox::NoButton,
                                   QMessageBox::NoButton);

                        return false;
                    }
                }
                else
                {
                    QMessageBox::warning(this, "Error", 
                           "Unable to retrieve stored parameters for procedure "
                            + modId
                            + " and script "
                            +  scriptId
                            + "\n"
                            + sqlStmt,
                            QMessageBox::Ok,
                            QMessageBox::NoButton,
                            QMessageBox::NoButton);

                    return false;
                }
            }
        }
    }

    // Update global variables in all scripts if modified global
    // variables exist
    if(m_globalParameterNames.count() > 0)
    {
        QStringList::Iterator itn;
        QStringList::Iterator ite;
        QStringList::Iterator itv;
        QStringList::Iterator itr;
        QString strTemp = "T";

        for(itn = m_globalParameterNames.begin(),
                ite = m_globalParameterElements.begin(),
                itv = m_globalParameterValues.begin(),
                itr = m_globalParameterRequired.begin();
            itn != m_globalParameterNames.end();
            ++itn, ++ite, ++itv, ++itr)
        {
            sqlStmt = QString("update ")
                + schema
                + "def_parms set"
                + " required = " 
                + QString("%1").arg((strTemp.compare(*itr) == 0) ? 1 : 0);
                
            if (!(*itv).isNull()) // parm value
            {
                sqlStmt += ",  parm_value = '" + *itv + "'";
            }

            sqlStmt += " where module_id = '" + leNewProdId->text()
                + "' and parm_name = '" + *itn
                + "' and parm_occurrence = " + *ite
                + " and required = 1";

            // Execute database update for the specified global parameter,
            // warn user and break if unsuccessful
            if (!queryExec.exec(sqlStmt,__FILE__,__LINE__))
            {
                QMessageBox::warning(this, "Error", 
                             "Unable to update global parameters for procedure " 
                             + leNewProdId->text()
                             + "\n"
                             + sqlStmt, 
                             QMessageBox::Ok,
                             QMessageBox::NoButton,
                             QMessageBox::NoButton);

                return false;
            }
        }
    }

    return true;
}

//****************************************************************************
int CreateNewProcedure::getStoredNumParmsChanged( 
    QString strScript) // I: The scipts whose values we are after
{
    // Return the number of parameters that have been changed for the
    // specified script.  If the script is not found a value of 0 will
    // be returned

    int countParmsChanged = 0;
    QString strTemp;

    for(QStringList::Iterator it = m_changedScripts.begin();
        it != m_changedScripts.end();
        ++it)
    {
        if(strScript.compare(*it) == 0)
        {
            // Move to next location
            ++it;
            strTemp = *it;
            countParmsChanged = strTemp.toInt();
            return countParmsChanged;
        }
    }

    return countParmsChanged;
}

//****************************************************************************
int CreateNewProcedure::getCurrentNumParmsChanged( 
    QString &strRows ) // O: contains a # separated list of all of the
                       // rows (parameters) that have been modified for the
                       // script that is currently selected in the GUI scripts
                       // table
{
    // Return the current number of parameters that have been modified from
    // the value read in from the database.  Return a qstring containing the
    // rows in the parms table the have been modified, each value is separated
    // by a #

    int countParmsChanged = 0;

    IASParameterItem *pItem = NULL;

    // Check each row in the GUI parameters qtable to see if the parameter
    // has been changed
    for(int row = 0; row < pModel->rowCount(); row++)
    {
        // Count each parameter that was changed

        pItem = pModel->getItem(row, 6);

        // If the parameter has changed, column 6 of table == "1"
        // append the row number to the list
        if(pItem->data().toString().compare("1") == 0)
        {
            // if not the first changed row, append a separator in front of
            // the row number
            if( countParmsChanged == 0)
            {
                strRows += QString("%1").arg(row);
            }
            else 
            {
                strRows += QString("#%1").arg(row);
            }

            countParmsChanged++;
        }
    }

    return countParmsChanged;
}

//****************************************************************************
void CreateNewProcedure::saveScriptParmValues()
{
    // Save the modified parameters (for the currently selected row in the
    // scripts qtable (tblScipts) from the parameters table into qstringlists.
    // Save all parameters, not just the modified ones. 

    if(m_previousScriptsTblRow == -1)
    {
        return;
    }

    IASParameterItem *pItem = NULL;

    IASParameterItem *sItem = sModel->getItem(m_previousScriptsTblRow, 4);

    QString strVal = sItem->data().toString();

    //  Save all parameters
    // For each parameter in the script (listed in parameters qtable)
    for(int row = 0; row < pModel->rowCount(); row++)
    {
        // Always insert the row as if all parameters were modified so they
        // are always put in the new procedure
        sItem = sModel->getItem(m_previousScriptsTblRow, 2);
        insertReplaceRow(sItem->data().toString(), row);

        // For each parameter, save the table/row contents

        // If this value is a global (common) value, save the value.
        // It will be used for the rest of the scripts
        pItem = pModel->getItem(row, 3);
        if(pItem->data().toString().compare("T") == 0)
        {
            // if the parameter/element is not already listed, add
            pItem = pModel->getItem(row, 0);
            QString pName = pItem->data().toString();
            pItem = pModel->getItem(row, 1);
            QString pElement = pItem->data().toString();
            pItem = pModel->getItem(row, 4);
            QString pRequired = pItem->data().toString();
            pItem = pModel->getItem(row, 5);
            QString pValue = pItem->data().toString();
            pItem = pModel->getItem(row, 10);
            QString pAttrFlag = pItem->data().toString();
            if(!isParamElementPresent(pName, pElement))
            {
                // save parameter name
                m_globalParameterNames += pName;

                // save parameter element number
                m_globalParameterElements += pElement;

                // save parameter value
                m_globalParameterValues += pValue;

                // save parameter required flag
                m_globalParameterRequired += pRequired;

                // save attribute flag
                pItem = pModel->getItem(row, 10);
                m_globalParameterAttributes += pItem->data().toString();
            }
            else    // replace values
            {
                int index = getParamElementIndex(pName, pElement);
                m_globalParameterValues[index] = pValue;
                m_globalParameterRequired[index]= pRequired;
                m_globalParameterAttributes[index]= pAttrFlag;
            }
        }
    }
}

//****************************************************************************
bool CreateNewProcedure::isScriptPresent(
    QString strScript) // I: script to locate in stored values
{
    // Return true if the script is found in the list of modified
    // (stored) scripts

    bool present = false;
    QStringList::Iterator itr = m_changedScripts.begin();

    while(itr != m_changedScripts.end())
    {
        if((*itr) == strScript)
        {
            present = true;
            break;
        }

        itr++;
    }

    return present;
}

//****************************************************************************
bool CreateNewProcedure::isRowPresent(
    QString strScript,  // I: script whose parameter (row) is being checked to
                        // see if it's stored
    int row)            // row in tblParms to check for
{
    // Return true if the specified row (parameter) for the specified script is
    // found in the list of modified rows (parameters)

    QString strRow = QString("%1").arg(row);
    QStringList strlTemp;

    // Get a list of modified rows for the specified script
    strlTemp = getStoredModifiedRows(strScript);

    // Search list for the specified row
    for(QStringList::Iterator it = strlTemp.begin(); it != strlTemp.end(); ++it)
    {
        if(strRow.compare(*it) == 0)
        {
            return true;
        }
    }

    return false;
}

//****************************************************************************
void CreateNewProcedure::insertRow(
    QString strScript, // I: script to insert a new row (parameter) for
    int row)           // I: row in tblParms (parameter)
{
    // Insert a new row into the qstringlist holding the modified parameters.

    int col;
    int count;
    QString strTemp;
    int start;
    QStringList::iterator itr;

    IASParameterItem *pItem = NULL;

    // If script is not present
    if(isScriptPresent(strScript))
    {
        if(isRowPresent(strScript, row))
        {
            replaceRow(strScript, row);

            return;
        }
        else
        {
            // Get insertion point for the specified script
            start = locateScriptEnd(strScript);

            // Save identifier identifying the script and row
            m_changedParameters.insert(start, 
                                       QString(strScript + "_row%1").arg(row));

            start++;
            // Save identifier identifying the row
            m_changedParameters.insert(start, QString("%1").arg(row));

            for(col = 0; col < pModel->columnCount(); col++)
            {
                pItem = pModel->getItem(row, col);
                start++;
                m_changedParameters.insert(start, pItem->data().toString());
            }

            // Update the modified scripts list
            itr = locateScriptBeginS(strScript);

            // Move to the location of the number of modified parameters
            itr++;

            // Save the number of modified parameters to a string
            strTemp = *itr;

            // Calculate the new number of modified parameters
            count = strTemp.toInt() + 1;

            // Update the qstrlinglist with the new number of modified
            // parameters
            *itr = QString("%1").arg(count);

            // Move to next item which is a seperated list of modified
            // rows (parameters) from the parms table tblParms
            itr++;

            // Append the new row to the end of the modified rows list
            if(count != 1)
            {
                strTemp = *itr + "#" + QString("%1").arg(row);
            }
            else 
            {
                strTemp = QString("%1").arg(row);
            }
            *itr = strTemp;
        }
    }
    else // script not present need to add script to list
    {
        m_changedParameters += strScript;
        m_changedParameters += (QString(strScript + "_row%1").arg(row));
        m_changedParameters += (QString("%1").arg(row));

        for(col = 0; col < pModel->columnCount(); col++)
        {
            pItem = pModel->getItem(row, col);
            m_changedParameters += pItem->data().toString();
        }

        // Add script to end of save scripts list
        m_changedScripts += strScript;
        m_changedScripts += QString("1");
        m_changedScripts += QString("%1").arg(row);
    }
}

//****************************************************************************
void CreateNewProcedure::replaceRow(
    QString strScript,   // I: script whose row (parameter) is to be replaced
    int row)             // I: row (parameter) to be replaced
{
    // Replace a row in the qstringlist that is holding the modified
    // parameters

    QStringList::Iterator itr;

    // Locate where script/parameter starts
    itr = locateScriptRow(strScript, row);

    // Move past the row indicator values
    ++itr;
    ++itr;

    for(int col = 0; col < pModel->columnCount(); col++, ++itr)
    {
        IASParameterItem *pItem = pModel->getItem(row, col);
        *itr = pItem->data().toString();
    }
}

//****************************************************************************
void CreateNewProcedure::insertReplaceRow(
    QString strScript,   // I: script whose row (parameter) is to be replaced
    int row)             // I: row (parameter) to be replaced
{
    // Insert a row into the qstringlist that holds modified parameters
    // if it is already present, replace it

    if(isRowPresent(strScript, row))
    {
        replaceRow(strScript, row);
    }
    else
    {
        insertRow(strScript, row);
    }
}

//****************************************************************************
QStringList CreateNewProcedure::getStoredModifiedRows(
    QString strScript)  // I: script requesting a list of modified parameters
                        // They are identified by their row position in
                        // the GUI parms table (tblParms)
{
    // Return a qstringlist of rows that have been modified for the specified
    // script.  Each row number is returned as a QString in a QStringList

    QString strTemp;
    QStringList::Iterator itr;
    QStringList strlRows;

    // Locate where data for this script starts
    itr = locateScriptBeginS(strScript);

    // If no values to read, return an empty script
    if(itr ==  m_changedScripts.end())
    {
        return strlRows;
    }

    // Move to the number of parameters modified and get the number
    itr++;

    // Move to where rows are stored in following format QString("1,2,7,8...")
    itr++;

    // Copy value into a temporary string
    strTemp = (*itr);

    strlRows = strTemp.split(QString("#"));

    return strlRows;
}

//****************************************************************************
int CreateNewProcedure::locateScriptEnd(
    QString strScript) // I: script to locate ending point for
{
    // Locate the insertion point for the specified script in the
    // qstringlist that holds the modified parameters

    int count = m_changedParameters.count();
    int start = m_changedParameters.indexOf(strScript);;

    // end = start + 1 + (number of parameters stored * number of values
    // stored/row) which is the 2 row values and the 7 columns from the table
    int end = start + 1 + (getStoredNumParmsChanged(strScript) *
                           (pModel->columnCount() + 2));;

    if(end <= count)
    {
        return end+1; // +1 because of the insert used after this call.
    }
    else
    {
        return count;
    }
}

//****************************************************************************
QStringList::Iterator CreateNewProcedure::locateScriptBeginP(
    QString strScript)      // I: script to locate in stored values
{
    // Find the QString in the stored parameter list where the entries start
    // for the specified script.  If script not found, return pointer to end
    // of list

    QStringList::Iterator itr = m_changedParameters.begin();

    while(itr != m_changedParameters.end())
    {
        if((*itr) == strScript)
        {
            break;
        }

        itr++;
    }

    return itr;
}

//****************************************************************************
QStringList::Iterator CreateNewProcedure::locateScriptBeginS(
    QString strScript) // I: script to locate in stored values
{
    // Find the QString in the stored script list where entries start for the
    // specified script.  If script not found,  return m_changedScripts.end()

    QStringList::Iterator itr = m_changedScripts.begin();

    while(itr != m_changedScripts.end())
    {
        if((*itr) == strScript)
        {
            break;
        }

        itr++;
    }

    return itr;
}

//****************************************************************************
QStringList::Iterator CreateNewProcedure::locateScriptRow(
    QString strScript,      // I: script to locate in stored values
    int row)                // I: row to locate in stored values
{
    // Find the QString in the stored parameters list where entries start for
    // the specified script and row. If strScript not found, return
    // pointer to end of qstringlist
    QString strTemp = strScript + QString("_row%1").arg(row);

    QStringList::Iterator itr = m_changedParameters.begin();

    while(itr != m_changedParameters.end())
    {
        if((*itr) == strTemp)
        {
            break;
        }

        itr++;
    }

    return itr;
}

//****************************************************************************
bool CreateNewProcedure::getNextRow(
    bool reset,             // I: restart at first row
    const QString strScript,// I: script whose modified rows are to be returned
    int &row)               // O: next row in list
{
    // Used to access a distinct list of modified parameters for the specified
    // script.  The row relates to the parameters row in the parms table
    // (tblParms). This method will return true while there are additional rows
    // available to be retrieved for the specified script.  If the input flag
    // reset is true, the method will restart at the beginning of the list, if
    // the script parameter passed in changes, the method will restart over
    // returning values for the new script

    QString strRow;

    // If reset flag sent or the script we are retrieving parameters from
    // reset
    if(reset == true || !strScript.compare(strCurrentScript) == 0)
    {
        strCurrentScript = strScript;
        m_strRows = getStoredModifiedRows(strCurrentScript);
        m_Rows_it = m_strRows.begin();
    }
    else
        ++m_Rows_it;

    if(m_Rows_it == m_strRows.end())
    {
        row = -1;
        return false;
    }

    strRow = *m_Rows_it;
    row = strRow.toInt();
    return true;
}

//****************************************************************************
bool CreateNewProcedure::getCurrRow(
    int &row)    // O: return the last row that was retrieved
{
    // Return the current value (last value) retrieved for the script opened
    // with method getNextRow()

    QString strRow;

    // Retrieve the current row value;

    if(m_Rows_it == m_strRows.end())
        return false;

    strRow = *m_Rows_it;
    row = strRow.toInt();
    return true;
}

//****************************************************************************
int CreateNewProcedure::getParmValues(
    QString strScript,      // I: script whose values are to be returned
    QStringList &strlParms) // O: qstrlinglist containing the values for
                            //    the specified script
{
    // Return a qstringlist of values for the specified script.  The list will
    // contain information on each parameter for the specified script that was
    // changed.  The first entry for each parameter is the row from the parms
    // table (tblParms) that contained the parameter, followed by a value for
    // each column in the table.  Some columns are hidden in the GUI.  The
    // rows follow each other one after another

    // Empty the passed in value list
    strlParms.clear();

    // Get the number of parameters changed (stored) for the specified script
    int countParmsChanged = getStoredNumParmsChanged(strScript);

    // Move iterator to first parameter for this script
    QStringList::Iterator it = locateScriptBeginP(strScript);

    // If the script was not found, return 0
    if(it == m_changedParameters.end())
    {
        return 0;
    }

    // Move past script name
    it++;

    // For each parameter that was modified, set the current value to the 
    // stored value.
    for(int param = 0; param < countParmsChanged; param++)
    {
        // Move past row value containing the stored row number and script name
        it++;

        // Append the row to the list
        strlParms += *it;
        it++;

        // Append the column values to the parameters list
        for(int col = 0; col < pModel->columnCount(); col ++, it++)
        {
            strlParms += *it;
        }
    }

    return countParmsChanged;
}

//****************************************************************************
bool CreateNewProcedure::getParmValues(
    QString strScript,  // I: Script whose values are to be returned
    int row,            // I: parameter whose value is to be returned
                        //    specified by its row in the parms table
    QStringList &strlParms) // O: qstrlinglist containing the values for
                            //    the specified script and row
{
    // Return true if the tblParms row (parameter) was found in the stored
    // parameter list. Return the values for specified parameter (row) in the
    // qstringlist passed in.  The first value is the row of the parms table
    // (tblParms) followed by the values for each column in the table for the
    // specified row.  If the row is not found, false value will be returned
    // and the qstrlinglist will be empty

    // Get the number of parameters that have changed for this script,
    //      return false if no values are found
    if(getStoredNumParmsChanged(strScript) < 1)
    {
        return false; 
    }

    // Empty the value list
    strlParms.clear();

    // locate the start of values for the specified script and row
    QStringList::Iterator itr = locateScriptRow(strScript, row);

    // If the script was not found
    //      return false
    if(itr == m_changedParameters.end())
    {
        return false;
    }

    // Move past row value containing the stored row number and script name
    itr++;

    // Copy row onto list
    strlParms += *itr;
    itr++;

    // Copy each column value into list
    for(int col = 0; col < pModel->columnCount(); col ++, itr++)
    {
        strlParms += *itr;
    }

    return true;
}

//****************************************************************************
bool CreateNewProcedure::getAttribute(
    QString Attribute,    // I: Qstring containing the attribute value that 
                          //    needs to be translated into a database value. 
    QString &dbAttribute) // O: Qstring containing the translated attribute 
                          //    value.
{
    // Return true if the attribute translation has been specified.
    // parameter list.  Return the translated value.

    if (Attribute.compare("DUMP") == 0)
    {
        dbAttribute = "_DUMP";
    }
    else
    {
        if (Attribute.compare("SKIP") == 0)
        {
            dbAttribute = "_SKIP";
        }
        else
        {
            // All other cases result in no Attribute.
            dbAttribute = "";

            // The Attribute is an empty string if the parameter is 
            // not ALGORITHMS.  "" is a valid ALGORITHM specifier.
            if ((Attribute.compare("RUN") != 0) &&
                (Attribute.compare("") != 0))
            {
                return false;
            }
        }
    }

    return true;
}

//****************************************************************************
void CreateNewProcedure::verifyScriptSelectionChanged(
    const QModelIndex& sIndex)
{
    // Slot: Verifies that the script selected in the script table actually
    // changed.  If it changed, calls the method that handles the change

    int row = sIndex.row();

    // If no rows have been previously selected, or new row is different
    // call method to update script
    if(m_currentScriptsTblRow == -1 && row > 0)
    {
        m_currentScriptsTblRow =  row;
    }
    else if(m_currentScriptsTblRow != row)
    {
        m_currentScriptsTblRow = row;
    }
    else
    {
        return;
    }

    if(!verifyParmUpdate() && !ignore)
    {
        ignore = true;

        tblScripts->setCurrentIndex(sModel->index(m_previousScriptsTblRow, 0));

        QMessageBox::warning(this, "Warning", 
             "You must set all of the parameters in the script before leaving",
             QMessageBox::Ok,
             QMessageBox::NoButton,
             QMessageBox::NoButton);
        ignore = false;
        return;
    }

    // Call method to update parameters table based on latest script selected
    scriptSelectionChanged();
}

//****************************************************************************
bool CreateNewProcedure::fillCopyDefaultValuesList()
{
    // Fills the combo box with a list of procedures to be used to
    // copy scripts from for the new procedure


    QString procedure;
    QString sqlStmt =
        "Select distinct module_id from "
        + schema
        + "def_parms order by module_id";
    IASSqlQuery queryProcedures;
    IASSqlQuery queryScripts;

    // Execute query 
    if(queryProcedures.exec(sqlStmt,__FILE__,__LINE__))
    {
        // If query was successful, for each procedure found, get all of its
        // scripts
        while (queryProcedures.next())
        {
            procedure = queryProcedures.value(0).toString();
            sqlStmt = QString("Select distinct sub_module_id from "
                              + schema
                              + "def_parms where module_id = '" 
                              + procedure 
                              + "' order by sub_module_id");

            if(queryScripts.exec(sqlStmt,__FILE__,__LINE__))
            {
                // If query was successful, insert each row into combo box
                while (queryScripts.next())
                {
                    cboCopyDefaultParmVal->addItem(
                        procedure + ":" + queryScripts.value(0).toString());
                    cboCopyDefaultParmVal->adjustSize();
                }
            }
            else    // error
            {
                QMessageBox::warning(this, "Error", 
                             "Unable to select from the def_parms table\n" 
                             + sqlStmt, 
                             QMessageBox::Ok,
                             QMessageBox::NoButton,
                             QMessageBox::NoButton);
                return false;
            }
        }
    }
    else
    {
        QMessageBox::warning(this,
                             "Error", 
                             "Unable to select from the def_parms table\n"
                             + sqlStmt, 
                             QMessageBox::Ok,
                             QMessageBox::NoButton,
                             QMessageBox::NoButton);
        return false;
    }
    return true;
}

//****************************************************************************
void CreateNewProcedure::updateParametersWithDefaultValues(
    const QString& value)   // I: The procedure/script combination to use to
                            //    set default values for the currently selected
                            //    script.
{
    // Slot: Updates the parameters in the parameters table with values from
    // the selected procedure/script combination selected from the combobox
    // by the user.

    IASSqlQuery query;
    IASSqlQuery queryAttributes;
    QString procedure;
    QString sqlStmt;
    QString sqlStmtAttr;
    QStringList strlValues;
    QStringList attributes;        // list of all algorithm attributes
    QString algorithm;             // the algorithm in the parm_value field
    QStringList parmValueSegments; // algorithm name, and potentially
                                   // algorithm attribute as well

    QVariant data;

    QModelIndex sIndex = tblScripts->currentIndex();

    int currentRow = sIndex.row();
    if(pModel->rowCount() == 0)
    {
        pbOK->setEnabled(true);

        // Enable the scripts table
        tblScripts->setEnabled(true);

        // Enable the select script combo box
        cboInsertScript->setEnabled(true);

        // Enable the copy from Procedure combo box
        cboSelectProcToCopyFrom->setEnabled(true);

        return;
    }

    // Here is some preparation for later use to do once before the 
    // loop.  Get a list of all algorithm attributes.

    // Get the list of possible attribute values.
    sqlStmtAttr = QString("select oui_value from ")
        + schema + "parm_list where parm_name = 'ATTRIBUTE'";

    if(queryAttributes.exec(sqlStmtAttr,__FILE__,__LINE__))
    {
        // get next parameter from database cursor
        while(queryAttributes.next())
        {
            attributes << queryAttributes.value(0).toString();
        }
    }
    else
    {
        return;
    }

    // For each value in the parms table, set the current value to the values
    // found in the def_parms table for that parameter
    strlValues = value.split(':');

    leCopyFromProcedure->setText(strlValues[0]);
    leCopyFromProcedure->adjustSize();
    leCopyFromScript->setText(strlValues[1]);
    leCopyFromScript->adjustSize();

    for(int row = 0; row < pModel->rowCount(); row++)
    {
        IASParameterItem *pItem = pModel->getItem(row, 0);

        QString pName = pItem->data().toString();
        pItem = pModel->getItem(row, 1);
        QString pElement = pItem->data().toString();
        sqlStmt = QString("select parm_name, parm_occurrence, parm_type, ")
            + " global, required, parm_value from "
            + schema + "def_parms where "
            + " parm_name = '" + pName
            + "' and parm_occurrence = " + pElement
            + " and module_id = '"       + leCopyFromProcedure->text()
            + "' and sub_module_id = '"  + leCopyFromScript->text() + "'";

        // select parameters from the select module/submodule
        // if the query executed successfully, copy values into table
        if(query.exec(sqlStmt,__FILE__,__LINE__))
        {
            // get next parameter from database cursor
            query.next();

            for(int col = 3; col < pModel->columnCount(); col++)
            {
                pItem = pModel->getItem(row, col);
                switch(col)
                {
                    case 3: // Common column
                    case 4: // Required column
                        if(query.isValid())
                        {
                            data = "F";
                            if(query.value(col).toInt() == 1)
                            {
                                data = "T";
                            }
                        }
                        break;

                    case 6: /* (hidden) Is set if the parameter is modified */
                    case 7: /* (hidden) Indicates a default value was entered */
                        data = 1;
                        break;

                    case 8: /* (hidden) The procedure the script values are
                                        copied from */
                        data = leCopyFromProcedure->text();
                        break;

                    case 9: /* (hidden) The script in the procedure where the
                                        values are copied from */
                        data = leCopyFromScript->text();
                        break;

                    case 10: /* (hidden) This case is handled below.
                                         Holds any parameter attributes */
                        break;

                    default: // All other columns
                        if(query.isValid())
                        {
                            data = query.value(col).toString();
                        }
                        break;
                }
                pItem->setData(data);
            }

            // If the Parm Attribute is ALGORITHMS, get the attribute value
            if(query.isValid())
            {
                // Since tblParms is initially populated from the def_parms
                // table, the parm_value could have the attribute appended
                // to it.  Take off the attribute if present for this query.

                pItem = pModel->getItem(row, 5);

                // find if an attribute is at the end of the parm_value
                parmValueSegments = pItem->data().toString().split("_");

                // this last element is an attribute if it's in the attribute
                // list
                if(attributes.contains (parmValueSegments.last()))
                {
                    // Remove the last element (the attribute) from the list
                    parmValueSegments.pop_back();

                    // Form the algorithm name from the rest of the parm_value
                    // segments
                    algorithm = parmValueSegments.join ("_");
                }
                else
                {
                    // The algorithm is just the original string
                    algorithm = pItem->data().toString();
                }

                data = algorithm;
                pItem->setData(data);

                if(query.value(0).toString() == "ALGORITHMS")
                {
                    sqlStmtAttr = QString("select attribute_value from ")
                        + schema
                        + "parm_attributes where parm_attribute = '"
                        + algorithm + "'";

                    if(queryAttributes.exec(sqlStmtAttr,__FILE__,__LINE__))
                    {
                        // get next parameter from database cursor
                        queryAttributes.next();

                        pItem = pModel->getItem(row, 10);
                        data = queryAttributes.value(0).toString();
                        pItem->setData(data);
                    }
                }
            }
        }
        else
        {
            QMessageBox::warning(this,
                                 "Error",
                                 "Unable to select from the def_parms table\n"
                                 + sqlStmt,
                                 QMessageBox::Ok,
                                 QMessageBox::NoButton,
                                 QMessageBox::NoButton);

            return;
        }
    }

    // Set parameter values in parameters qtable to global values already saved
    setParmGlobalValues();

    // Update the parameters qtable, for items with a discrete set of values,
    // combo boxes containing these values into the parameters table, and
    // select the default value that was returned above and is stored in the
    // table
    insertParmValueItem();
    insertParmAttributeItem();

    // Set flag indicating parameter value changed for this script so the
    // stored values will be read back when the script is reselected.
    IASParameterItem *sItem = sModel->getItem(currentRow, 4);
    data = 1;
    sItem->setData(data);

    // If all of the parameters have been updated, set ok button to enabled
    if(verifyParmUpdate())
    {
        pbOK->setEnabled(true);

        // Enable the scripts table
        tblScripts->setEnabled(true);

        // Enable the select script combo box
        cboInsertScript->setEnabled(true);

        // Enable the copy from Procedure combo box
        cboSelectProcToCopyFrom->setEnabled(true);

        // Set flag indicating all of the parameters for this row have
        // been updated
        sItem = sModel->getItem(currentRow, 7);
        // Data was set to 1 for column 4 above.  So use it again.
        sItem->setData(data);
    }
}

//****************************************************************************
void CreateNewProcedure::setParmValuesTexttoStoredValues(
    QString strScript)  // I: The currently selected script whose
                        //    parameters are listed in the parms
                        //    table and need to be updated
{
    // Set the values in the parameters qtable to the stored values for the
    // specified script.  Only modified values are stored, and updated here

    int row;
    QStringList strlParms;

    QVariant data;

    // While there are more modified rows for the script (strScript), get the
    // next row (row) and update the displayed values to match the stored
    // parameter values
    while(getNextRow(false,strScript,row))
    {
        if(getParmValues(strScript, row, strlParms))
        {
            // Remove row number from the front of qstringlist
            strlParms.pop_front();

            for(int col = 0; col < strlParms.count(); col++)
            {
                // If an item exists in the qtable, and string is not empty
                if(!strlParms[col].isEmpty())
                {
                    IASParameterItem *pItem = pModel->getItem(row, col);
                    data = strlParms[col];
                    pItem->setData(data);
                }
            }
        }
        else
        {
            return;
        }
    }
}

//****************************************************************************
QStringList CreateNewProcedure::getParamElements(
    QString strParmName)    // I: Parameter name to get the occurrences of
{

    // Return a qstringlist containing the elements (occurrences) of
    // a global parameter that are recorded as being modified

    QStringList strlElements;
    QStringList::Iterator itn;
    QStringList::Iterator ite;

    for(itn = m_globalParameterNames.begin(), 
            ite = m_globalParameterElements.begin(); 
        itn != m_globalParameterNames.end(); ++itn, ++ite)
    {
        if(strParmName.compare(*itn) == 0)
            strlElements += *ite;
    }

    return (strlElements);
}

//****************************************************************************
bool CreateNewProcedure::isParamElementPresent(
    QString strParmName,    // I: Parameter name specified
    QString strElement)     // I: Parameter occurrence (element) specified
{
    // Identify if the specified parameter name and element (occurrence)
    // are included in the qstringlist of modified global parameters

    QStringList strlElements = getParamElements(strParmName);

    // If element not found (index = -1)
    if(strlElements.indexOf(strElement) == -1)
        return false;
    else
        return true;
}

//****************************************************************************
void CreateNewProcedure::setParmGlobalValues()
{
    // Set the parameters in the parameters qtable to the global values that
    // are stored

    int count = m_globalParameterNames.count();
    int index;
    QVariant data;

    if(m_globalParameterNames.count() > 0)
    {
        for(int row = 0; row < pModel->rowCount() && count > 0; row++)
        {
            // find parameter/element is list of globals while we haven't
            // updated all of the globals, or all of the parameters in this
            // script
            IASParameterItem *pItem = pModel->getItem(row, 0);
            QString pName = pItem->data().toString();
            pItem = pModel->getItem(row, 1);
            QString pElement = pItem->data().toString();
            index = getParamElementIndex(pName, pElement);

            // if parameter/element found, update values
            if(index > -1)
            {
                count--;

                // If the user hasn't changed this from a global parameter for
                // this script, update it
                pItem = pModel->getItem(row, 3);
                if(pItem->data().toString().compare("T") == 0)
                {
                    // required
                    pItem = pModel->getItem(row, 4);
                    data = m_globalParameterRequired[index];
                    pItem->setData(data);

                    // value
                    pItem = pModel->getItem(row, 5);
                    data = m_globalParameterValues[index];
                    pItem->setData(data);

                    // modified
                    pItem = pModel->getItem(row, 6);
                    data = 1;
                    pItem->setData(data);
 
                    // attribute
                    if(!m_globalParameterAttributes[index].isNull())
                    {
                        pItem = pModel->getItem(row, 10);
                        data = m_globalParameterAttributes[index];
                        pItem->setData(data);
                    }
                }
            }
        }
    }
}

//****************************************************************************
int CreateNewProcedure::getParamElementIndex(
    QString strParmName,    // I: Parameter Name specified
    QString strElement)     // I: Parameter Occurrence specified
{
    // Return the index in the global qstringlist for the specified parameter
    // name and element combination.  Return -1 if that combination is not
    // found

    QStringList::Iterator itn;
    QStringList::Iterator ite;
    int index;

    for(itn = m_globalParameterNames.begin(), 
        ite = m_globalParameterElements.begin(), index = 0; 
        itn != m_globalParameterNames.end(); ++itn, ++ite, index++)
    {
        if(strParmName.compare(*itn) == 0)
            if(strElement.compare(*ite) == 0)
                return (index);
    }

    return (-1);
}

//****************************************************************************
bool CreateNewProcedure::verifyParmUpdate()
{
    // Return false if any of the parameters for the script have not been
    // updated/set after the current script was manually inserted

    if(sModel->rowCount() > 0)
    {
        for (int row = 0; row < pModel->rowCount(); row++)
        {
            IASParameterItem *pItem = pModel->getItem(row, 7);
            if(pItem->data().toInt() == 0)
            {
                tblParms->setCurrentIndex(pModel->index(row, 5));

                return false;
            }
        }
    }

    return true;
}

//****************************************************************************
void CreateNewProcedure::descReturnPressed()
{
    // SLOT: Called when the return is pressed in the enter new procedure
    // description line edit.  On exit, focus is set to the copy Scipts
    // from drop down combo-box

    cboSelectProcToCopyFrom->setFocus();
}

//****************************************************************************
void CreateNewProcedure::showPopupMenu(const QPoint& pos)
{
    QModelIndex sIndex = tblScripts->currentIndex();

    // Create the popup menu and add the menu items
    //---------------------------------------------
    QMenu* menu = new QMenu(this);

    // Only display the options if they are possible.
    if(sIndex.row() != 0)
    {
        menu->addAction("Move Script UP", this, SLOT(moveScriptUp()));
    }
    if(sIndex.row() != sModel->rowCount()-1)
    {
        menu->addAction("Move Script DOWN", this, SLOT(moveScriptDown()));
    }

    // Execute the menu.
    menu->exec(tblScripts->mapToGlobal(pos));

    delete menu;

    // Setting these allows selections of the moved rows to work.
    sIndex = tblScripts->currentIndex();
    m_previousScriptsTblRow = sIndex.row();
    m_currentScriptsTblRow = sIndex.row();
}

//****************************************************************************
void CreateNewProcedure::moveScriptUp()
{
    QModelIndex sIndex = tblScripts->currentIndex();
    QModelIndex newIndex = sModel->index(sIndex.row()-1, 0);

    // Grab the row and insert into the new location.
    IASParameterItemRow* sItemRow = sModel->takeRow(sIndex);
    sModel->insertRow(newIndex, sItemRow);

    // Setup to process through the rows.
    QVariant data;
    int count = sModel->rowCount();

    // This processes through the entire list and updates the sequence numbers.
    for(int seqNbr = 0; seqNbr < count; seqNbr++)
    {
        // Grab the sequence item.
        IASParameterItem *sItem = sModel->getItem(seqNbr, 0);
        data = seqNbr+1;
        sItem->setData(data); // Set the items value.
    }

    tblScripts->setCurrentIndex(newIndex);
}

//****************************************************************************
void CreateNewProcedure::moveScriptDown()
{
    QModelIndex sIndex = tblScripts->currentIndex();
    QModelIndex newIndex = sModel->index(sIndex.row()+1, 0);

    // Grab the row and insert into the new location.
    IASParameterItemRow* sItemRow = sModel->takeRow(sIndex);
    sModel->insertRow(newIndex, sItemRow);

    // Setup to process through the rows.
    QVariant data;
    int count = sModel->rowCount();

    // This processes through the entire list and updates the sequence numbers.
    for(int seqNbr = 0; seqNbr < count; seqNbr++)
    {
        // Grab the sequence item.
        IASParameterItem *sItem = sModel->getItem(seqNbr, 0);
        data = seqNbr+1;
        sItem->setData(data); // Set the items value.
    }

    tblScripts->setCurrentIndex(newIndex);
}

