/******************************************************************************
               Image Assessment System (IAS) Software
 ******************************************************************************
 *
 * Name:   CreateNewScript.cpp
 *
 * PURPOSE:
 * This is the definition of the CreateNewScript class which is used to 
 * create new test scripts which are stored in the database in the
 * modules and parameters tables
 *
 *****************************************************************************/

// System Includes
#include <iostream>

// Qt Includes 
#include <QtCore>
#include <QtGui>
#include <QtSql>

// Local Includes
#include "CreateNewScript.h"
#include "IASSqlQuery.h"
#include "IASErrorDialog.h"
#include "IASSchemaConfig.h"
#include "IASParameterDelegate.h"
#include "IASParameterModel.h"

//****************************************************************************
CreateNewScript::CreateNewScript(
    QWidget* parent,    //I: The address to this widget's parent 
    Qt::WindowFlags fl) //I: Used by QT
: QWidget(parent, fl)
{
    QWidget::setAttribute(Qt::WA_DeleteOnClose);

    // Set up the user interface.
    setupUi(this);

    // New Script name has not been validated
    m_validNewScriptName = FALSE;

    // Set up the script parameters table
    setupParmsTable();

    // Initiate a transaction
    QSqlDatabase::database().transaction();

    // Fill Sats and Sensors combobox with list of available sats and sensors
    fillSatsAndSensorsList();

    // disable delete pushbutton until parameters are entered
    pbDelete->setEnabled(FALSE);

    // Setup Fonts in combo box items here because they keep getting
    // reset by QT designer
    QFont cboSelectScriptToCopyFrom_font(cboSelectScriptToCopyFrom->font());
    cboSelectScriptToCopyFrom_font.setBold(FALSE);
    cboSelectScriptToCopyFrom_font.setItalic(FALSE);
    cboSelectScriptToCopyFrom->setFont(cboSelectScriptToCopyFrom_font); 

    QFont cboSelectParmtoAdd_font(cboSelectParmtoAdd->font());
    cboSelectParmtoAdd_font.setBold(FALSE);
    cboSelectParmtoAdd_font.setItalic(FALSE);
    cboSelectParmtoAdd->setFont(cboSelectParmtoAdd_font); 

    QFont cboSatSensor_font(cboSatSensor->font());
    cboSatSensor_font.setItalic(FALSE);
    cboSatSensor_font.setBold(FALSE);
    cboSatSensor->setFont(cboSatSensor_font);

    // Connect all the signals and slots.
    connect(cboSelectScriptToCopyFrom, SIGNAL(activated(const QString&)),
            this, SLOT(selectScriptToInsert(const QString&)));
    connect(leNewScriptId, SIGNAL(returnPressed()),
            this, SLOT(validateScriptName()));
    connect(pbDelete, SIGNAL(pressed()), this, SLOT(DeleteParam()));
    connect(cboSelectParmtoAdd, SIGNAL(activated(const QString&)),
            this, SLOT(selectedParameterToInsert(const QString&)));
    connect(pbClearList, SIGNAL(pressed()), this, SLOT(clearList()));
    connect(pbCancel, SIGNAL(pressed()), this, SLOT(close()));
    connect(pbOk, SIGNAL(pressed()),
            this, SLOT(accept()));
    connect(cboSatSensor, SIGNAL(activated(const QString&)),
            this, SLOT(sensorSelected()));
}

//****************************************************************************
void CreateNewScript::setupParmsTable()
{
    QStringList labels;

    // Set up the columns and properties of the GUI parameters qtable.
    QFont f = tblParms->font();
    f.setWeight(QFont::Bold);

    // Create an initial model for the table.
    m_model = new IASParameterModel;

    // Set the header labels.
    m_model->setColumnCount(3);
    labels << tr("Parm Name") << tr("Element") << tr("Parm Type");
    m_model->setHorizontalHeaderLabels(labels);

    // Set up the delegate for the table.
    tblParms->setItemDelegate(new IASParameterDelegate);

    // Set up the edit triggers.
    tblParms->setEditTriggers(QAbstractItemView::AllEditTriggers);

    // Set the rows to alternate colors.
    tblParms->setAlternatingRowColors(true);

    // Make headings bold and set focus and selection.
    tblParms->horizontalHeader()->setFont(f);

    // Only select one row at a time.
    tblParms->setSelectionMode(QAbstractItemView::SingleSelection);
    tblParms->setSelectionBehavior(QAbstractItemView::SelectRows);

    // Get rid of the vertical header.
    tblParms->verticalHeader()->hide();

    // Set the model
    tblParms->setModel(m_model);

    // Adjust the columns to the new data.
    tblParms->resizeColumnsToContents();

    // Set last column to be stretchable
    tblParms->horizontalHeader()->setStretchLastSection(true);
}

//***********************************************************************
void CreateNewScript::commitAndClose()
{
    // Commit all changes for this transaction and close dialogue.

    if(!QSqlDatabase::database().commit())
    {
         QMessageBox::warning(this, 
            "Error", 
            "Commit of database changes failed",
            QMessageBox::Ok,
            QMessageBox::NoButton,
            QMessageBox::NoButton);
    }

    close();
}

//***********************************************************************
void CreateNewScript::rollbackAndClose()
{
    // Rollback all changes for this transaction and close dialogue

    if(!QSqlDatabase::database().rollback())
    {
        QMessageBox::warning(this,
            "Error", 
            "Rollback of database changes failed",
            QMessageBox::Ok,
            QMessageBox::NoButton,
            QMessageBox::NoButton);
    }

    close();
}

//***********************************************************************
void CreateNewScript::DeleteParam()
{
    // Slot: Delete the selected parameter from the GUI parameters qtable

    QModelIndex index = tblParms->currentIndex();

    // If no item is currently selected, issue warning and return
    if(!index.isValid())
    {
        QMessageBox::warning(this,
            "Warning", 
            "You must select a parameter to delete first",
            QMessageBox::Ok,
            QMessageBox::NoButton,
            QMessageBox::NoButton);

        return;
    }

    int row = index.row();

    // Delete currently selected row
    m_model->deleteRow(row);

    // If there are no more rows in the GUI parameters table, disable the
    // delete pushbutton and enable the user to reselect another script to
    // use as a template.
    if(m_model->rowCount() == 0)
    {
        pbDelete->setEnabled(FALSE);
        cboSelectScriptToCopyFrom->setEnabled(TRUE);
    }
}

//***********************************************************************
int CreateNewScript::getNextParmOccur(
    QString strParmName, // I: The parameter which we want the next value for
    int &insertP,        // O: The insertion point (row) where to
                         //    insert the next value
    QString &strType)
{
    // Return the next parameter occurrence value for a named
    // parameter.  Return the point (row) where the parameter should be
    // inserted

    int nxtParmOccur=0;
    insertP =  m_model->rowCount();
    QString tblParmName("");

    for(int row = 0; row < m_model->rowCount(); row++)
    {
        // Grab the item for the parm name.
        IASParameterItem* item = m_model->getItem(row, 0);

        // Grab the parm name.
        tblParmName = item->data().toString();

        // If the parameter name is found.
        if(strParmName.compare(tblParmName) == 0)
        {
            // Set the parameter type to the value found in the list
            IASParameterItem* typeItem = m_model->getItem(row, 2);
            strType = typeItem->data().toString();

            // If the occurrence value of this row is greater 
            // than the previous recorded value, set new value
            IASParameterItem* occItem = m_model->getItem(row, 1);
            int occurence = occItem->data().toInt();

            if(occurence > nxtParmOccur)
            {
                nxtParmOccur = occurence;
                insertP = row + 1;
            }
        }
    }

    return(nxtParmOccur + 1);
}

//***********************************************************************
void CreateNewScript::fillSatsAndSensorsList()
{
   // Fills the combo box with a list of satellites and sensors

    IASSqlQuery query;

    QString sqlStmtSensor =
        "select satellite, sensor from satellite_sensor_cfg";

    // Issue the query to get the satellites and sensors.  If unsuccessful,
    // warn the user.

    if(query.exec(sqlStmtSensor, __FILE__, __LINE__))
    {
        // While another row is retrieved, insert new row into combobox
        while(query.next())
        {
            cboSatSensor->addItem(query.value(0).toString() + // satellite
                            " " + query.value(1).toString()); // sensor
        }
    }
    else
    {
        QMessageBox::warning(this,
                            "Error",
                            "Error getting satellites and sensors from the "
                                "satellite_sensor_cfg table\n",
                            QMessageBox::Ok,
                            QMessageBox::NoButton,
                            QMessageBox::NoButton);
    }

    // See if the sensor is known by virtue of only 1 sensor being configured
    if(cboSatSensor->count() <= 1)
    {
        // If so select that sensor
        sensorSelected();

        // Enable the new procedure name line edit, so it can be changed
        leNewScriptId->setEnabled(TRUE);
        leNewScriptId->selectAll();
    }
    else
    {
        // Don't have 1 sensor as a default if there are choices (be consistent
        // with other satellite/sensor screens) 
        cboSatSensor->addItem("");

        // Otherwise, allow the operator to select the sensor
        cboSatSensor->setEnabled(TRUE);

        // Disable the new procedure name line edit, so it can't be changed
        leNewScriptId->setEnabled(FALSE);
    }
}

//****************************************************************************
void CreateNewScript::fillScriptsList()
{
    // Fill the select scripts combobox with a list of scripts from which 
    // parameters can be copied from 

    IASSqlQuery query;
    IASSqlQuery queryParmCount;
    QString sqlStmt = 
        QString("Select distinct module_id from ")
            + m_schema
            + "modules where mod_type"
            + "= 'SCRIPT' order by module_id";
    QString sqlStmt2;

    // Execute query sqlStmt: get distinct script names from the 
    // modules table

    // If query was successful, insert each row into combobox

    if(query.exec(sqlStmt,__FILE__,__LINE__))
    {
        while (query.next())
        {
            // Retrieve the number of parameters for the 
            // specified script

            sqlStmt2 = " SELECT count(*) from "
                + m_schema
                + "PARAMETERS WHERE module_id = '" 
                + query.value(0).toString() + "'";

            if(queryParmCount.exec(sqlStmt2,__FILE__,__LINE__))
            {
                queryParmCount.next();

                // If the script has parameters, insert 
                // into the select script
                // combobox to be used as a templates

                if(queryParmCount.value(0).toInt() != 0)
                {
                    cboSelectScriptToCopyFrom->addItem(
                        query.value(0).toString());
                }
            }
            else
            {
                QMessageBox::warning(this,
                    "Error", 
                    "Unable to access the parameters table for script: " 
                    + query.value(0).toString()
                    + QString("\n")
                    + sqlStmt,
                    QMessageBox::Ok,
                    QMessageBox::NoButton, 
                    QMessageBox::NoButton);

                return;
            }
        }
    }
    else
    {
        QMessageBox::warning(this, 
            "Error", 
            "Unable to access modules table to retrieve a list of scripts"
            + QString("\n")
            + sqlStmt,
            QMessageBox::Ok,
            QMessageBox::NoButton,
            QMessageBox::NoButton);
    }
}

//****************************************************************************
void CreateNewScript::selectScriptToInsert( 
    const QString& strScript) // I: Script selected by user
{
    // Slot:  Operator selected a script from the combo box. Its parameters
    // will be used as a template for the new script

    IASSqlQuery query;

    if(m_validNewScriptName == FALSE)
    {
        QMessageBox::warning(this, 
            "Warning", 
            "A valid script name must be entered",
            QMessageBox::Ok,
            QMessageBox::NoButton,
            QMessageBox::NoButton);

        // Return focus to the new script name entry line edit widget
        leNewScriptId->setFocus();

        // Select the currently entered name
        leNewScriptId->selectAll();

        return;
    }

    // Delete all the data in the model.
    m_model->deleteData();

    QString sqlStmt=QString("SELECT parm_name, parm_occurrence, parm_type ")
        + " from "
        + m_schema
        + "parameters WHERE module_id = '" 
        + strScript  
        + "' order by parm_name, parm_occurrence";

    // Execute query: retrieve the parameters information for the 
    // selected script
    if(!query.exec(sqlStmt,__FILE__,__LINE__))
    {
        QMessageBox::warning(this, 
            "Error", 
            "Error reading values from parameters table for script" 
            + strScript
            + "\n"
            + sqlStmt,
            QMessageBox::Ok,
            QMessageBox::NoButton,
            QMessageBox::NoButton);

        return;
    }

    // While there are more parameters, fetch their values and store in the
    // parameters qtable
    while(query.next())
    {
        IASParameterItemRow* itemRow = new IASParameterItemRow;

        IASParameterItem* item = new IASParameterItem;
        QVariant data = query.value(0);
        item->setData(data);
        itemRow->append(item);

        item = new IASParameterItem;
        data = query.value(1);
        item->setData(data);
        item->setEditable(true);
        item->setType(IASParameterItem::IntegerRange);
        data = 1; // Set the minimum range value.
        item->setMinimum(data);
        data = 1000; // Set the maximum range value.
        item->setMaximum(data);
        itemRow->append(item);

        item = new IASParameterItem;
        data = query.value(2);
        item->setData(data);
        item->setEditable(true);
        item->setType(IASParameterItem::TextSelection);
        item->setOptions(m_options);
        itemRow->append(item);

        m_model->appendRow(itemRow);
    }

    if(m_model->rowCount() > 0)
    {
        pbDelete->setEnabled(TRUE);
        pbOk->setEnabled(TRUE);
    }

    // Adjust the columns to the new data.
    tblParms->resizeColumnsToContents();

    // Set last column to be stretchable
    tblParms->horizontalHeader()->setStretchLastSection(true);
}

//****************************************************************************
bool  CreateNewScript::validateScriptName()
{
    // Verify that the new script name entered by the user does not already
    // exist in the database

    QString strTemp = leNewScriptId->text().simplified();

    IASSqlQuery query;
    QString sqlStmt= "SELECT module_id from "
        + m_schema
        + "modules WHERE module_id = '" 
        +  strTemp 
        + "' and mod_type = 'SCRIPT'";

    leNewScriptId->setText(strTemp);

    // Check if name is too long to store in the column of the database

    if(strTemp.length() > 24)
    {
        m_validNewScriptName = FALSE;
        cboSelectScriptToCopyFrom->setEnabled(FALSE);
        QMessageBox::warning(this, 
            "Error", 
            "That script name : "+ strTemp + 
            ", is too long.\nMaximum length is 25 characters",
            QMessageBox::Ok,
            QMessageBox::NoButton,
            QMessageBox::NoButton);
        leNewScriptId->selectAll();

        return (FALSE);
    }
    else if(strTemp.length() > 0)
    {
        // Determine if there are any punctuation characters in the
        // script name

        QRegExp rxp = QRegExp("[^\\w- ]");

        if(strTemp.contains(rxp))
        {
            QMessageBox::warning(this, 
                "Error", 
                "Re-Enter script name, name is invalid\n"
                    "Remove punctuation marks",
                QMessageBox::Ok,
                QMessageBox::NoButton,
                QMessageBox::NoButton);
            return (FALSE);
        }

        if(query.exec(sqlStmt,__FILE__,__LINE__))
        {
            // Retrieve the first row

            query.next();

            // If the module name in the database matches the 
            // user entered value, issue a warning and return

            if (query.isValid())
            {
                if(query.value(0).toString() == strTemp)
                {
                    m_validNewScriptName = FALSE;
                    cboSelectScriptToCopyFrom->setEnabled(FALSE);
                    QMessageBox::warning(this, 
                        "Error", 
                        "A script with that name: "
                        + strTemp 
                        + ", already exists",
                        QMessageBox::Ok,
                        QMessageBox::NoButton,
                        QMessageBox::NoButton);

                    return (TRUE);
                }
            }

            m_validNewScriptName = TRUE;

            // Enable the combobox filled with script names 
            // to be used as templates

            cboSelectScriptToCopyFrom->setEnabled(TRUE);

            // Disable the line edit for the script name

            leNewScriptId->setEnabled(FALSE);

            // Add the new script name to heading

            lblFormHeading->setText(lblFormHeading->text() 
                + " : " 
                + leNewScriptId->text());

            // Enable the add parameter combobox

            cboSelectParmtoAdd->setEnabled(TRUE);

            // Enable the line edit widget used for 
            // entering the new script's description

            leNewScriptDesc->setEnabled(TRUE);

            // Set focus on the description entry widget

            leNewScriptDesc->setFocus();
        }
    }
    else
    {
        m_validNewScriptName = FALSE;
        cboSelectScriptToCopyFrom->setEnabled(FALSE);
        return (FALSE);
    }

    return (TRUE);
}

//****************************************************************************
void CreateNewScript::sensorSelected()
{
    // Slot: The sensor has been selected, or there is only one sensor in
    // the configuration so it by default has effectively been selected.
    // Activate the various later parts of the dialog

    QStringList satAndSensor; // list including satellite and sensor

    if (cboSatSensor->currentText().isEmpty())
    {
        return;
    }
 
    // Disable the sensor combobox, so it can't be changed
    cboSatSensor->setEnabled(FALSE);

    // Enable the new procedure name line edit, so it can be changed.
    // Select the entire text in the line edit for entering the procedure name
    leNewScriptId->setEnabled(TRUE);
    leNewScriptId->setFocus();
    leNewScriptId->selectAll();

    // Extract the satellite and sensor from the combobox choice
    satAndSensor = cboSatSensor->currentText().split(' ');

    // Look up the schema based on the satellite and sensor
    m_schema = IASSchemaConfig::instance()->schemaName(
        satAndSensor[0],   // satellite
        satAndSensor[1]);  // sensor

    // Setup the options for the parameter type IASParameterItem's used in the
    // table.
    m_options = parameterTypes();

    // Fill combobox with list of scripts that can be inserted
    fillScriptsList();

    // fill parameters list with a list of parameters to select
    fillAddParamsList();
}

//****************************************************************************
void CreateNewScript::writeNewScripttoDatabase()
{
    // Save the new script to the database

    // If insert of script to modules table was unsuccessful,
    // rollback, issue warning if rollback unsuccessful, and return.
    if(writeNewScripttoModulesTable() != true)
    {
        if (!QSqlDatabase::database().rollback())
        {
            QMessageBox::warning(this, 
                    "Error", 
                    "Rollback of database changes failed",
                    QMessageBox::Ok,
                    QMessageBox::NoButton,
                    QMessageBox::NoButton);
            return;
        }
    }

    if(m_model->rowCount() > 0)
    {
        // If insert of new script parameters into the parameters table
        // was unsuccessful, rollback, issue warning if rollback 
        // unsuccessful and return.

        if(writeNewScripttoParametersTable() != true)
        {
            if(!QSqlDatabase::database().rollback())
            {
                QMessageBox::warning(this, 
                    "Error", 
                    "Rollback of database changes failed",
                    QMessageBox::Ok,
                    QMessageBox::NoButton, 
                    QMessageBox::NoButton);
                return;
            }
        }
    }

    // Successfully updated the database with new script, commit and close 
    // window.
    commitAndClose();
}

//****************************************************************************
bool CreateNewScript::writeNewScripttoModulesTable()
{
    // Write out new script to the modules table in the database

    IASSqlQuery query;

    QString sqlStmt = QString("insert into ")
        + m_schema
        + "modules (MODULE_ID,MOD_TYPE,MOD_STATUS,MOD_DESC)"
        + " values('" + leNewScriptId->text() + "', 'SCRIPT', 'TEST','" 
        + leNewScriptDesc->text() + "')";

    // If query was unsuccessful, issue a warning, and return
    if (!query.exec(sqlStmt,__FILE__,__LINE__))
    {
        QMessageBox::warning(this, 
            "Error", 
            "Error inserting new script into modules table"
            + QString("\n")
            + sqlStmt,
            QMessageBox::Ok,
            QMessageBox::NoButton,
            QMessageBox::NoButton);
        return (FALSE);
    }

    return (TRUE);
}

//****************************************************************************
bool CreateNewScript::writeNewScripttoParametersTable()
{
    // Write out new script's parameters to the parameters table

    IASSqlQuery query;
    QString sqlStmt;

    // For each row (parameter) in the GUI parameters qtable, insert into
    // database parameters table
    for(int row = 0; row < m_model->rowCount(); row++)
    {
        // Generate SQL insert command for the row
        sqlStmt = QString("insert into ")
            + m_schema
            + "parameters" 
            + " (MODULE_ID, PARM_NAME, PARM_OCCURRENCE, PARM_TYPE) values('"
            + leNewScriptId->text() + "'";

        for(int col = 0; col < m_model->columnCount(); col++)
        {
            IASParameterItem* item = m_model->getItem(row, col);

            sqlStmt +=  "," ;
            sqlStmt +=  "'" ;
            sqlStmt += item->data().toString();
            sqlStmt +=  "'";
        }
        sqlStmt +=  ")";
        
        // If query was unsuccessful
        if(!query.exec(sqlStmt,__FILE__,__LINE__))
        {
            QMessageBox::warning(this, 
                "Error", 
                "Error inserting new parameter into the parameters table"
                + QString("\n")
                + sqlStmt,
                QMessageBox::Ok,
                QMessageBox::NoButton,
                QMessageBox::NoButton);
            return false;
        }
    }

    return true;
}

//****************************************************************************
QStringList
CreateNewScript::parameterTypes()
{
    IASSqlQuery query;
    QString sqlStmt;
    QStringList types;

    // Make sure it is an empty list.
    types.clear();

    // Get a list of parameter types
    sqlStmt = QString("Select distinct parm_type from ")
                  + m_schema + "parameters order by parm_type";

    // Execute query, if successful, append each type to list of types
    if(query.exec(sqlStmt,__FILE__,__LINE__))
    {
        while(query.next())
        {
            types += query.value(0).toString();
        }
    }

    return types;
}

//***********************************************************************
bool CreateNewScript::verifyDistinctParameters()
{
    // Return TRUE if all parameters are distinct, if two parameters
    // have the same name and occurrence, return FALSE if not;

    int count;
    QString strTemp;
    QStringList strlParameters;
    QStringList::Iterator itn;
    QStringList::Iterator ite;

    // create a list of parameters along with their occurrences 
    // separated by a colon ":"
    for(int row = 0; row < m_model->rowCount(); row++)
    {
        IASParameterItem* colItem0 = m_model->getItem(row, 0);
        IASParameterItem* colItem1 = m_model->getItem(row, 1);

        strlParameters += QString(colItem0->data().toString() + ":" +
                                  colItem1->data().toString());
    }

    // Take the first parameter and occurrence and check the remainder 
    // of the list for a match
    for(itn = strlParameters.begin(); itn != strlParameters.end(); ++itn)
    {
        count = 0;
        strTemp = *itn;

        // Search the remainder of the list for the parameter and 
        // occurrence value stored in strTemp
        for(ite  = itn; ite != strlParameters.end(); ++ite, count++)
        {
            // If first loop through remainder of list, set 
            // iterator to first value after values searching 
            // for, start searching after the search value
            if(count == 0)
            {
                ite++;
                if(ite == strlParameters.end())
                {
                    return (TRUE);
                }
            }

            // If a duplicate is found, issue warning and return 
            // a FALSE value
            if(strTemp.compare(*ite) == 0)
            {
                QMessageBox::warning(this, 
                    "Error", 
                    "Duplicate parameter name and occurrence: " + *itn,
                    QMessageBox::Ok,
                    QMessageBox::NoButton,
                    QMessageBox::NoButton);

                return (FALSE);
            }
        }
    }

    return TRUE;
}

//****************************************************************************
void CreateNewScript::accept()
{
    // Slot:Write out new script to database

    // If all parameters are not distinct
    // return, and do not write to the database
    if(!verifyDistinctParameters())
    {
        return;
    }

    // disable the OK pushbutton
    pbOk->setEnabled(FALSE);

    // Insert new script into the database
    writeNewScripttoDatabase();
}

//****************************************************************************
void CreateNewScript::fillAddParamsList()
{
    // Fill the add parameter combobox with a list of parameters from the
    // parameters table.

    IASSqlQuery query;
    QString sqlStmt= QString("SELECT distinct parm_name from ")
        + m_schema
        + "parameters order by parm_name";

    // execute query, if unsuccessful issue a warning
    if(!query.exec(sqlStmt,__FILE__,__LINE__))
    {
        QMessageBox::warning(this, 
            "Error", 
            "Error reading values from parameters table"
            + QString("\n")
            + sqlStmt,
            QMessageBox::Ok,
            QMessageBox::NoButton,
            QMessageBox::NoButton);
    }

    // While more parameters in the cursor, insert them into the combobox
    while(query.next())
    {
        cboSelectParmtoAdd->addItem(query.value(0).toString());
    }
}

//****************************************************************************
void CreateNewScript::selectedParameterToInsert(
    const QString &strParmName)  // I: Parameter to insert into script 
                                 // and qtable
{
    // Slot:  Operator selected a parameter from the combo box  

    int currentRow;
    int nxtOccur;
    int insertP;
    QString strType;
    QString sqlStmt = "SELECT distinct parm_type from " +
                      m_schema + "PARAMETERS WHERE parm_name = '" +
                      strParmName + "'";
    IASSqlQuery query;

    // If the script name hasn't been successfully validated, issue 
    // a warning, don't insert the parameter, and return
    if(m_validNewScriptName == false)
    {
        QMessageBox::warning(this, 
            "Warning", 
            "A valid script name must be entered",
            QMessageBox::Ok,
            QMessageBox::NoButton,
            QMessageBox::NoButton);
        leNewScriptId->setFocus();
        leNewScriptId->selectAll();
        return;
    }

    // Get the current selected row if any
    QModelIndex index = tblParms->currentIndex();
    currentRow = index.row();

    if(currentRow < 0)
    {
        currentRow = m_model->rowCount();
    }

    // Get the parameter type
    if(query.exec(sqlStmt,__FILE__,__LINE__))
    {
        query.next();

        // Set parameter type to value from database
        strType = query.value(0).toString();

        // get next occurrence for this parameter, if it already 
        // exits in the list, set the type to the type of the 
        // previous values in the list
        nxtOccur = getNextParmOccur(strParmName, insertP, strType);

        IASParameterItemRow* itemRow = new IASParameterItemRow;
        QVariant data;

        // Insert the parm name
        IASParameterItem* item = new IASParameterItem;
        data = strParmName;
        item->setData(data);
        itemRow->append(item);

        // Insert the parm occurrence
        item = new IASParameterItem;
        QString occurence = QString::number(nxtOccur);
        data = occurence;
        item->setData(data);
        item->setEditable(true);
        item->setType(IASParameterItem::IntegerRange);
        data = 1; // Set the minimum range value.
        item->setMinimum(data);
        data = 1000; // Set the maximum range value.
        item->setMaximum(data);
        itemRow->append(item);

        // Insert the parm type
        item = new IASParameterItem;
        data = strType;
        item->setData(data);
        item->setEditable(true);
        item->setType(IASParameterItem::TextSelection);
        item->setOptions(m_options);
        itemRow->append(item);

        // Add the row to the model.
        m_model->appendRow(itemRow);
    }
    else
    {
        QMessageBox::warning(this, 
            "Warning", 
             QString("Unable to read parameter type from parameters table ")
                + " for parameter: " 
                + strParmName
                +"\n"
                + sqlStmt,
                QMessageBox::Ok,
                QMessageBox::NoButton,
                QMessageBox::NoButton);

        return;
    }

    if(m_model->rowCount() > 0)
    {
        // Set the currently selected cell to the new row.
        tblParms->selectRow(insertP);
        pbDelete->setEnabled(TRUE);
        pbOk->setEnabled(TRUE);
    }
}

//****************************************************************************
void CreateNewScript::clearList()
{
    // Slot:Clear parameters from tblParms qtable

    m_model->deleteData();

    pbDelete->setEnabled(FALSE);
    cboSelectScriptToCopyFrom->setEnabled(TRUE);
}

