/******************************************************************************
               Image Assessment System (IAS) Software
 ******************************************************************************
 *
 * Name:   CreateWO.cpp
 *
 * PURPOSE:
 * This is the definition of the CreateWO class which is used to
 * Create a work order either new or from template.
 *
 *****************************************************************************/

// System Includes
#include <iostream>
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>

// QT Includes
#include <QtCore>
#include <QtGui>
#include <QtSql>

// IAS Qt Includes 
#include <IASQtConst.h>
#include <iasQtWoConst.h>

// Local Includes
#include "CreateWO.h"
#include "IASSqlQuery.h"
#include "ModScriptWOParms.h"
#include "ModGlobalWOParms.h"
#include "IASSchemaConfig.h"
#include "IASKeyValueComboBox.h"
#include "IASParameterDelegate.h"
#include "IASParameterModel.h"

const int defaultPriority = 4;

/* NOTE - leSceneId is used to hold the interval id if the work order happens
          to be one that processes an interval. */
/*
    CreateWO

    Initializes an instance of the create work order interface.  A bunch of
    functionality implemented here is inherited and used in both the ModifyWO
    and TemplateWO interfaces.  All of the required functionality for the
    CreateWO interface is implemented here.
*/
CreateWO::CreateWO(QWidget *parent, 
                   Qt::WindowFlags fl)
: QWidget(parent, fl)
{
    QWidget::setAttribute(Qt::WA_DeleteOnClose);

    // Set up the user interface.
    setupUi(this);

    m_deletedFlag = false; // Initialize to not deleted.

    /* These are utilized by the ModifyWO interface to allow for not changing
       them.  They are initialized here so that the TemplateWO and this
       CreateWO have them set to empty strings. */
    cw_l0rDataPath = "";
    cw_l0rDataStatus = "";

    allowReset = false;
    allowParmChanges = true;
}

/*
    initialize

    This is used to initialize all of the interface objects and variables
    correctly.
*/
bool CreateWO::initialize()
{
    QFont font = this->font();
    QStringList labels;

    m_model = new IASParameterModel;
    m_model->setColumnCount(2);

    labels << "Script ID" << "Pause Flag";
    m_model->setHorizontalHeaderLabels(labels);

    // Set up the delegate for the table.
    tblScriptId->setItemDelegate(new IASParameterDelegate);

    // Set up the edit triggers.
    tblScriptId->setEditTriggers(QAbstractItemView::AllEditTriggers);

    // Set the rows to alternate colors.
    tblScriptId->setAlternatingRowColors(true);

    // Make headings bold and set focus and selection.
    font.setBold(true);
    tblScriptId->horizontalHeader()->setFont(font);

    // Only select one row at a time.
    tblScriptId->setSelectionMode(QAbstractItemView::SingleSelection);
    tblScriptId->setSelectionBehavior(QAbstractItemView::SelectRows);

    // Get rid of the vertical header.
    tblScriptId->verticalHeader()->hide();

    // Set first columns size to
    tblScriptId->setColumnWidth(0, 180);
    tblScriptId->setColumnWidth(1, 80);

    // Resize the table
    tblScriptId->horizontalHeader()->setResizeMode(
                                     QHeaderView::Interactive);

    // Set the model
    tblScriptId->setModel(m_model);

    tblScriptId->resizeColumnsToContents();
    // Capture the initial size of the Script ID header (just header, no data).
    scriptIDHeaderWidth = tblScriptId->horizontalHeader()->sectionSize(0);
    /************************************************************************/
    // Get the IAS_SYSTEM_ID from the environment.
    m_systemId = getenv("IAS_SYSTEM_ID");
    if(m_systemId.isEmpty())
    {
        QMessageBox::critical(0,"Missing Environment",
            "Could not get IAS_SYSTEM_ID from environment.\n"
            "This variable is required for system startup or shutdown",
            QMessageBox::Ok,0);

        close();
    }

    // Get the IAS_INSTANCE_NAME from the environment.
    instance_name = getenv("IAS_INSTANCE_NAME");
    if(instance_name.isEmpty())
    {
        QMessageBox::critical(0, "Missing Environment",
            "Could not get IAS_INSTANCE_NAME from environment.\n"
            "This variable is required for initial work order creation.",
            QMessageBox::Ok, 0);
        return false;
    }

    m_systemId = m_systemId.toUpper();
    /* Using only the first character since the env. var. could be longer. */
    m_systemId = m_systemId.at(0);

    /************************************************************************/
    // Setup the satellite sensor selection box.
    QStringList schemas = IASSchemaConfig::instance()->allKeys();
    QString empty("");
    kvcboSatSensor->addKeyValue(empty, empty);
    QStringList::iterator itr;
    for(itr = schemas.begin(); itr != schemas.end(); ++itr)
    {
        QString satSen = (*itr);
        QString schema =
            IASSchemaConfig::instance()->schemaNameFromKey(satSen);
        satSen.replace("_", " ");
        kvcboSatSensor->addKeyValue(satSen, schema);
    }
    m_oldSatSensor = kvcboSatSensor->currentValue(); // Initialize

    // Setup Fonts in combo box items here because they keep getting
    // reset by QT designer
    QFont kvcboSatSensor_font(kvcboSatSensor->font());
    kvcboSatSensor_font.setItalic(false);
    kvcboSatSensor_font.setBold(false);
    kvcboSatSensor->setFont(kvcboSatSensor_font);

    // Populate the instances
    getInstances();

    if(kvcboSatSensor->count() == 2) // 1 schema, 1 empty row
    {
        // There is only 1 satellite/sensor choice, so select it by default
        kvcboSatSensor->setCurrentIndex(1);
        kvcboSatSensor->setEditable(false);
        kvcboSatSensor->setEnabled(false);

        // Update the ephemeris options.
        updatePcdDefEph();

        // Grab the list of procedures and populate the combo box.
        getProcedureIds();
    }

    // Initialize Priority combobox
    for (int i = 1; i < 10; i++)
        cboPriority->addItem(QString::number(i));

    // Select the default priority level in the combobox
    selectDefaultPriority();

    /************************************************************************/

    setTitle();

    /************************************************************************/

    // Set up the push buttons.
    setupButtons();

    /************************************************************************/

    // Hide Landsat stuff not used by LDCM
    lblCalParmFile->hide();
    leCalParmFile->hide();
    tbCalParmFile->hide();
    lblWorkOrderPath->hide();
    leWorkOrderPath->hide();
    tbWorkOrderPath->hide();

    /************************************************************************/

    // Populate some of the common GUI information with default values.
    populateDefaults();

    /************************************************************************/

    // Populate the database with the default stuff.
    if (!populateInitialDb())
    {
        return false;
    }

    /************************************************************************/

    // Connect the signals and slots.
    setupSignalsSlots();

    // Force a shrink to the visible/displayed contents of the interface.
    resize(0,0);

    return true;
}

/*
    showEvent

    Workaround for the tab order problem with using radio buttons.  This way,
    the user can tab to the checked item (and not to the item with the lowest
    tab order setting).
*/
void CreateWO::showEvent(QShowEvent *event)
{
    Q_UNUSED(event);
    if (rbTest->isChecked())
    {
        QWidget::setTabOrder(rbTest, rbStandard);
    }
    if (rbCustom->isChecked())
    {
        QWidget::setTabOrder(rbCustom, rbTest);
        QWidget::setTabOrder(rbCustom, rbStandard);
    }
    if (rbEvaluation->isChecked())
    {
        QWidget::setTabOrder(rbEvaluation, rbProduction);
    }
    cboPriority->setFocus();
}

bool CreateWO::getCurrentConfiguration(QString& instance, QString& satellite,
    QString& sensor)
{
    QStringList satSenList = kvcboSatSensor->currentText().split(' ');
    satellite = satSenList[0];
    sensor = satSenList[1];
    instance = cboInstance->currentText();

    // Do not update the script parms if sensor/satellite/instance are empty
    if (sensor.isEmpty() || satellite.isEmpty() || instance.isEmpty())
    {
        QMessageBox::warning(this,"Warning","The satellite, sensor, and "
            "instance must be set to create a new work order");
        return false;
    }

    return true;
}

/*
    selectDefaultPriority

    Selects the default priority level (currently 9) in the cboPriority
    combobox.

*/
void CreateWO::selectDefaultPriority()
{
    int findIndex;
    findIndex = cboPriority->findText(QString::number(defaultPriority),
        Qt::MatchExactly);
    cboPriority->setCurrentIndex(findIndex);
}

/*
    setTitle

    This is used to set the titles correctly for the specific interface.
*/
void CreateWO::setTitle()
{
    QString title = "Create New Work Order";

    lblCreateWO->setText(title);
    setWindowTitle(title);
}

void CreateWO::setupSignalsSlots()
{
    // Connect all the signals and slots.
    connect(kvcboSatSensor, SIGNAL(activated(const QString&)), this,
        SLOT(changedSchema(const QString&)));

    connect(cboInstance, SIGNAL(currentIndexChanged(const QString&)), this,
        SLOT(changedInstance(const QString&)));

    connect(tbCalParmFile, SIGNAL(clicked()), this, SLOT(getCalParmFile()));
    connect(tbL0RDataPath, SIGNAL(clicked()), this, SLOT(getL0RDataPath()));
    connect(tbWorkOrderPath, SIGNAL(clicked()), this, SLOT(getWorkOrderPath()));

    connect(cboProcedureId, SIGNAL(activated(const QString&)),
            this, SLOT(populateScriptId(const QString&)));

    connect(pbScript, SIGNAL(clicked()), this, SLOT(updateScriptParms()));
    connect(pbCommon, SIGNAL(clicked()), this, SLOT(updateCommonParms()));
    connect(pbOK, SIGNAL(clicked()), this, SLOT(updateWorkOrders()));
    connect(pbCancel, SIGNAL(clicked()), this, SLOT(cancelWorkOrder()));
}

void CreateWO::setupButtons()
{
}

void CreateWO::deleteWorkOrder()
{
    QString sqlStmt;
    IASSqlQuery query;

    // Set up the SQL statement to delete the wo in the WORK_ORDERS table.
    // There will be a cascade delete for the work order in the wo_script,
    // wo_parms, and wo_common_parms tables.
    sqlStmt = "DELETE FROM work_orders WHERE work_order_id = '";
    sqlStmt += lblWorkOrderID->text() + "'";

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        return;
    }

    m_deletedFlag = true;

    // Close the window
    close();
}

/*
    cancelWorkOrder

    This deletes the work order currently being worked on from the database
    and then calls close to terminate the interface.

    Since this is a new work order, that has not been made official it can be
    deleted from the database.  An official work order is one that has been
    accepted(you selected OK on the CreateWO or TemplateWO interface).
*/
void CreateWO::cancelWorkOrder()
{
    // This needs to delete the work order from the database.
    deleteWorkOrder(); // Only for new and new template.

    // Close the window.
    close();
}

/*
    closeEvent

    This doubles checks to see if the database needs to have the work order
    deleted, and then terminates the interface.

    This method is kicked off either from calling close() or through the "X"
    close button in the title bar.
*/
void CreateWO::closeEvent(QCloseEvent* ce)
{
    if(!m_deletedFlag)
    {
        deleteWorkOrder(); // Only for new and new template.
    }

    ce->accept();
}

void CreateWO::updatePcdDefEph()
{
    // Set the sensor specific default ephemeris type.
    if(kvcboSatSensor->currentText() == "L8 OLITIRS")
    {
        rbPCD->setChecked(true); // Select PCD
        gbEphemerisType->setEnabled(false);
        gbEphemerisType->hide();
    }
    else
    {
        gbEphemerisType->setEnabled(true);
        gbEphemerisType->show();
        if(kvcboSatSensor->currentText() == "L4 TM")
        {
            rbPCD->setChecked(true); // Select PCD
        }
        else
        {
            rbDefinitive->setChecked(true); // Select Definitive
        }
    }
}

void CreateWO::updateOutputDbSchema()
{
    rbProduction->setChecked(true); // Select Production database schema

    // Set the sensor specific default ephemeris type.
    if(kvcboSatSensor->currentText() == "L8 OLITIRS")
    {
        gbOutputDbSchema->setEnabled(true);
        gbOutputDbSchema->show();
    }
    else
    {
        gbOutputDbSchema->setEnabled(false);
        gbOutputDbSchema->hide();
    }
}

void CreateWO::changedSchema(const QString&)
{
    if(m_oldSatSensor != kvcboSatSensor->currentValue())
    {
        // Get list of procedure IDs.
        getProcedureIds();

        // Configure the ephemeris options.
        updatePcdDefEph();

        // Configure the output database schema options.
        updateOutputDbSchema();

        // Shrink to the visible/displayed contents of the interface.
        resize(0, 0);

        // Save the new value.
        m_oldSatSensor = kvcboSatSensor->currentValue();
    }
}

void CreateWO::changedInstance(const QString& test)
{
    getProcedureIds();
}

void CreateWO::populateDefaults()
{
    QString sqlStmt;
    QString date;
    IASSqlQuery query;
    QString woId;

    woId = getWoId();

    // Assign the new work order ID to the label in the interface
    lblWorkOrderID->setText(woId);

    // Get the requester and current time
    sqlStmt = "SELECT to_char(sysdate,'DY MON DD HH24:MI:SS YYYY') "
              "FROM dual";
    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        close();
    }
    query.next();
    lblCurrentTime->setText(query.value(0).toString()); // sysdate

    struct passwd *pws;
    pws = getpwuid(getuid());
    if(pws == NULL)
    {
        QMessageBox::critical(this, "Error",
            "Failed to retrieve user information from the system.");
        close();
    }

    lblRequester->setText(pws->pw_name); // user

    // Set the Request Start Date
    date = changeDateFormat(query.value(0).toString());
    dteRequestStartDate->setDateTime(QDateTime::fromString(date,Qt::TextDate));
    dteRequestStartDate->setMinimumDate(dteRequestStartDate->date());
}

/*
    populateInitialDb

    This is used to insert an initial work order record into the database
    which is populated with some default information.

    This method is used by the child classes to populate the interface
    correctly based on information retrieved from the database.
*/
bool CreateWO::populateInitialDb()
{
    QString sqlStmt;
    IASSqlQuery query;
    int query_size = 0;

    // Check if the processing instance is valid
    sqlStmt = "select instance_name from processing_instances where "
              "instance_name = '" + instance_name + "'";
    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        return false;
    }

    // Loop over the query to determine the size as calling query.size() is
    // not a reliable method to determine the size
    while (query.next())
        query_size++;

    // Check that the instance exists
    if (query_size < 1)
    {
        QMessageBox::critical(this, "Error",
            "Processing instance " + instance_name + " does not exist "
            "in the database. Please check the IAS_INSTANCE_NAME variable.");
        return false;
    }

    // Insert a row into the work_orders table.  This will be updated later.
    sqlStmt = "INSERT INTO work_orders (work_order_id, processing_instance, "
              "product_request_id,"
              "archived, date_entered, "
              "deleted, priority, requester, request_type, state) VALUES ('";
    sqlStmt += lblWorkOrderID->text() + "','";
    sqlStmt += instance_name + "','";
    // Use the work_order_id as the product_request_id
    sqlStmt += lblWorkOrderID->text();
    sqlStmt += "',0,TO_DATE('";
    sqlStmt += lblCurrentTime->text();
    sqlStmt += "','DY MON DD HH24:MI:SS YYYY'),0," +
        QString::number(defaultPriority) + ",'";
    sqlStmt += lblRequester->text();
    sqlStmt += "','S','" + DB_WO_HALTED + "')";

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        return false;
    }

    return true;
}

/*
    getWoId

    This method is used to generate a "New" work order Id from the database
    for the new work order that is to be created.
*/
QString CreateWO::getWoId()
{
    QString sqlStmt;
    IASSqlQuery query;
    QString woId;

    // Get the next sequence number for the work order id
    sqlStmt = "SELECT wo_seq.nextval FROM dual";
    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        // Close window if error occurs getting work order id
        close();
    }
    query.next();

    // Work Order Id equals System Id + next value in sequence
    woId = m_systemId + query.value(0).toString();

    return woId;
}

QString CreateWO::changeDateFormat(const QString& date)
{
    QString dateTemp;
    QString resultDate = date;

    // The date as returned from the database has all caps and QT expects
    // upper and lower.  The following is the process to get the database
    // format to qt format. I could not figure out a way to format Qt so I
    // just used the Qt::TextDate Standard.
    dateTemp = resultDate.mid(1, 2);
    dateTemp = dateTemp.toLower();
    resultDate.replace(1, 2, dateTemp); 
    dateTemp = resultDate.mid(5, 2);
    dateTemp = dateTemp.toLower();
    resultDate.replace(5, 2, dateTemp); 

    return resultDate;
}

void CreateWO::getProcedureIds()
{
    IASSqlQuery query;
    QString sqlStmt;
    QString satellite;
    QString sensor;
    QString instance;

    cboProcedureId->clear();

    // Do not update the procedure list if the satellite/sensor/instances
    // are not populated
    if (!getCurrentConfiguration(instance, satellite, sensor))
    {
        return;
    }

    sqlStmt = "select module_id from "
        "modules WHERE mod_type='PROCEDURE' "
        "and processing_instance = '" + instance + "' "
        "and satellite = '" + satellite + "' "
        "and sensor = '" + sensor + "' "
        "order by module_id asc";

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        close();
    }

    cboProcedureId->addItem("");

    while(query.next())
    {
        cboProcedureId->addItem(query.value(0).toString());
    }
    cboProcedureId->setEnabled(true);
}

void CreateWO::getInstances()
{
    IASSqlQuery query;
    QString sqlStmt;
    int index = 0;

    sqlStmt = "select instance_name from processing_instances "
        "order by collection_number desc";

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        close();
    }

    while(query.next())
    {
        cboInstance->addItem(query.value(0).toString());
        if (query.value(0).toString().compare(instance_name) == 0)
        {
            cboInstance->setCurrentIndex(index);
        }
        index++;
    }

    cboInstance->setEnabled(true);
}

/*
    getScriptSql

    This method configures the SQL required to retrieve the script Id and
    pause flag information from the database.
*/
QString CreateWO::getScriptSql(const QString& id)
{
    QString sqlStmt = "";
    QString satellite;
    QString sensor;
    QString instance;

    // Do not create the sql if sensor/satellite/instance are empty
    if (!getCurrentConfiguration(instance, satellite, sensor))
    {
        return "";
    }

    sqlStmt  = "SELECT sub_module_id, pause_flag ";
    sqlStmt += "FROM sub_modules ";
    sqlStmt += "WHERE module_id = '" + id + "' ";
    sqlStmt += "and processing_instance = '" + instance + "' ";
    sqlStmt += "and satellite = '" + satellite + "' ";
    sqlStmt += "and sensor = '" + sensor + "' ";
    sqlStmt += "ORDER BY seq_nbr";

    return sqlStmt;
}

void CreateWO::populateScriptId(const QString& procId)
{
    // Allow about 4 pixels of empty space on either side of text in a cell.
    const int pixelsAdjustment = 8;
    int rowCnt = 0;
    QStringList scriptList;
    QString sqlStmt;
    IASSqlQuery query;

    IASParameterItemRow* itemRow = NULL;
    IASParameterItem* item = NULL;

    // Clear the model every time a new procedure ID is selected.
    m_model->deleteData();

    // Get the sql statement.
    sqlStmt = getScriptSql(procId);
    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        return;
    }

    // Setup the scripts table.
    QVariant data;
    QStringList selectionOptions;
    selectionOptions << "Yes" << "No";

    int maxWidth = scriptIDHeaderWidth - pixelsAdjustment;
    while(query.next())
    {
        scriptList.append(query.value(0).toString());

        // Because resizeColumnsToContents only considers items that are in
        // view, we figure out the optimal column width for the script ID
        // ourselves.
        int width = tblScriptId->fontMetrics().width(query.value(0).toString());
        if (width > maxWidth)
        {
            maxWidth = width;
        }

        itemRow = new IASParameterItemRow;

        item = new IASParameterItem;
        data = scriptList[rowCnt];
        item->setData(data);
        itemRow->append(item);

        item = new IASParameterItem;
        if(query.value(1).toInt() == 1)
        {
            data = selectionOptions[0]; // Yes
        }
        else
        {
            data = selectionOptions[1]; // No
        }
        item->setData(data);
        item->setEditable(true);
        item->setType(IASParameterItem::TextSelection);
        item->setOptions(selectionOptions);
        itemRow->append(item);

        m_model->appendRow(itemRow);

        rowCnt++;
    }

    // Because an item in a QTreeView needs more space than the text
    // width (~8px), I've added some pixels there. It would be nice to retrieve
    // this size at runtime, but I could live with the current solution.
    tblScriptId->setColumnWidth(0, maxWidth + pixelsAdjustment);

    // Fill in the wo_scripts, and wo_common_parms database tables
    populateNewWorkOrder();
}

void CreateWO::updateCommonParms()
{
    QString satellite;
    QString sensor;
    QString instance;

    // Check if the procedure ID has been set
    if(cboProcedureId->currentText() == "")
    {
        QMessageBox::warning(this,"Warning","The Procedure ID must be set "
                             "before updating common parameters");
        return;
    }
    // Do not update the common parms if sensor/satellite/instance are empty
    else if (!getCurrentConfiguration(instance, satellite, sensor))
    {
        return;
    }

    // Call the common parameters update routine
    ModGlobalWOParms CommonChange(lblWorkOrderID->text(),
                                  cboProcedureId->currentText(),
                                  instance, satellite, sensor);

    CommonChange.exec();
}

void CreateWO::deleteWOParms()
{
    QString sqlStmt;
    IASSqlQuery query;

    // Set up the SQL statement to delete the wo in the WO_SCRIPTS table.
    // The work_order will be deleted from the WO_PARMS table automatically
    sqlStmt = "DELETE FROM wo_scripts WHERE work_order_id = '";
    sqlStmt += lblWorkOrderID->text();
    sqlStmt += "'";

    if(!query.exec(sqlStmt,__FILE__,__LINE__))
    {
        return;
    }

    // Set up the SQL statement to delete the wo in the WO_COMMON_PARMS table
    sqlStmt = "DELETE FROM wo_common_parms WHERE work_order_id = '";
    sqlStmt += lblWorkOrderID->text();
    sqlStmt += "'";

    if(!query.exec(sqlStmt,__FILE__,__LINE__))
    {
        return;
    }
}

/*
    populateNewWorkOrder

    This completes the initial insertion of the work order parameters into the
    work order system.
*/
void CreateWO::populateNewWorkOrder()
{
    QString sqlStmt;
    IASSqlQuery query;
    IASSqlQuery queryOut;
    QString satellite;
    QString sensor;
    QString instance;

    // Start a database transaction.
    QSqlDatabase::database().transaction();

    // Clean out the tables so that they can be filled again
    deleteWOParms();

    // Do not create the sql if sensor/satellite/instance are empty
    if (!getCurrentConfiguration(instance, satellite, sensor))
    {
        return;
    }

    // Fill in the wo_scripts table with the default information
    sqlStmt  = "SELECT sub_module_id, pause_flag, seq_nbr ";
    sqlStmt += "FROM sub_modules ";
    sqlStmt += "WHERE processing_instance = '" + instance + "' ";
    sqlStmt += "and satellite = '" + satellite + "' ";
    sqlStmt += "and sensor = '" + sensor + "' ";
    sqlStmt += "and module_id = '" + cboProcedureId->currentText() + "'";

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        return;
    }

    while(query.next())
    {
        // set up the SQL statement to insert into the wo_scripts table
        sqlStmt = "INSERT INTO wo_scripts ";
        sqlStmt += "(work_order_id, script_id, pause_flag, seq_nbr, state) ";
        sqlStmt += "VALUES ('";
        sqlStmt += lblWorkOrderID->text() + "','";
        sqlStmt += query.value(0).toString() + "',";
        sqlStmt += query.value(1).toString() + ",";
        sqlStmt += query.value(2).toString() + ",'P')";

        if(!queryOut.exec(sqlStmt, __FILE__, __LINE__))
        {
            QSqlDatabase::database().rollback();

            return;
        }

        queryOut.next();
    }
      
    // Fill in the wo_common_parms table
    sqlStmt  = "SELECT DISTINCT parm_name, parm_occurrence, parm_value ";
    sqlStmt += "FROM def_parms ";
    sqlStmt += "WHERE module_id = '" + cboProcedureId->currentText() + "' ";
    sqlStmt += "and processing_instance = '" + instance + "' ";
    sqlStmt += "and satellite = '" + satellite + "' ";
    sqlStmt += "and sensor = '" + sensor + "' ";
    sqlStmt += "AND global = 1";

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        QSqlDatabase::database().rollback();

        return;
    }

    while(query.next())
    {
        // set up the sql statement to insert into the wo_common_parms table
        sqlStmt  = "INSERT INTO wo_common_parms ";
        sqlStmt += "(work_order_id, parm_name, parm_occurrence, parm_value) ";
        sqlStmt += "VALUES ('" + lblWorkOrderID->text() + "','";
        sqlStmt += query.value(0).toString() + "',";
        sqlStmt += query.value(1).toString() + ",'";
        sqlStmt += query.value(2).toString() + "')";

        if(!queryOut.exec(sqlStmt, __FILE__, __LINE__))
        {
            QSqlDatabase::database().rollback();

            return;
        }

        queryOut.next();
    }

    insertWOParms();

    if(!QSqlDatabase::database().commit())
    {
        QMessageBox::warning(this, "Error",
            "Commit of database changes failed",
            QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
    }
}

/*
    getWoParmSql

    This configures the SQL required to retrieve the default parameters for
    the specified procedure. 
*/
QString CreateWO::getWoParmSql()
{
    QString sqlStmt;
    QString satellite;
    QString sensor;
    QString instance;

    // Do not create the sql if sensor/satellite/instance are empty
    if (!getCurrentConfiguration(instance, satellite, sensor))
    {
        return "";
    }

    sqlStmt  = "SELECT sub_module_id, parm_name, parm_occurrence, parm_value ";
    sqlStmt += "FROM def_parms ";
    sqlStmt += "WHERE module_id = '" + cboProcedureId->currentText() + "'";
    sqlStmt += "and processing_instance = '" + instance + "' ";
    sqlStmt += "and satellite = '" + satellite + "' ";
    sqlStmt += "and sensor = '" + sensor + "' ";

    return sqlStmt;
}

void CreateWO::insertWOParms()
{
    QSqlDatabase::database().transaction();
    IASSqlQuery query;
    IASSqlQuery queryOut;
    QString sqlStmt;

    sqlStmt = getWoParmSql();

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        return;
    }

    while(query.next())
    {
        // set up the SQL statement to insert into the wo_parms table
        sqlStmt  = "INSERT INTO wo_parms ";
        sqlStmt += "(work_order_id, script_id, parm_name, parm_occurrence, ";
        sqlStmt += "parm_value) ";
        sqlStmt += "VALUES ('" + lblWorkOrderID->text() + "','";
        sqlStmt += query.value(0).toString() + "','";
        sqlStmt += query.value(1).toString() + "',";
        sqlStmt += query.value(2).toString() + ",'";
        sqlStmt += query.value(3).toString() + "')";

        if(!queryOut.exec(sqlStmt,__FILE__,__LINE__))
        {
            QSqlDatabase::database().rollback();

            return;
        }
    }

    if(!QSqlDatabase::database().commit())
    {
        QMessageBox::warning(this, "Error",
            "Commit of database changes failed",
            QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
    }
}

void CreateWO::updateScriptParms()
{
    QString Script;
    int rowNum;
    QString satellite;
    QString sensor;
    QString instance;

    // Check if the procedure ID has been set
    if(cboProcedureId->currentText() == "")
    {
        QMessageBox::warning(this,"Warning","The Procedure ID must be set "
                             "before updating script parameters");

        return;
    }
    else if (!getCurrentConfiguration(instance, satellite, sensor))
    {
        return;
    }

    QModelIndex index = tblScriptId->currentIndex();

    if(!index.isValid())
    {
        Script = "";
    }
    else
    {
        rowNum = index.row();

        IASParameterItem* item = m_model->getItem(rowNum, 0);

        Script = item->data().toString();
    }

    // Call the script parameter update routine
    ModScriptWOParms scriptChange(lblWorkOrderID->text(),
            cboProcedureId->currentText(), instance, satellite, sensor, Script,
            allowParmChanges, allowReset);

    scriptChange.exec();
}

void CreateWO::updateWorkOrders()
{
    int row = 0;
    QString sqlStmt;
    IASSqlQuery query;
    IASSqlQuery queryTmp;
    QString temp;
    QStringList parmValueSegments; // algorithm name, and potentially
                                   // algorithm attribute as well
    bool found_LifetimeRG = false; // flag indicating LT RG algorithm found
    bool found_HistoRG = false;    // flag indicating Histo RG algorithm found
    IASParameterItem* item = NULL;
    QString satellite;
    QString sensor;
    QString instance;

    // Product ID must be set
    if(cboProcedureId->currentText() == "")
    {
        QMessageBox::warning(this,"Warning","The Procedure ID must be set "
                             "before running the Work Order");

        return;
    }
    // Do not update the work order if sensor/satellite/instance are empty
    else if (!getCurrentConfiguration(instance, satellite, sensor))
    {
        return;
    }

    // Check that only 1 Relative Gain algorithm is run.
    sqlStmt  = "SELECT parm_value ";
    sqlStmt += "FROM wo_parms ";
    sqlStmt += "WHERE parm_name = 'ALGORITHMS'";
    sqlStmt += " AND work_order_id = '" + lblWorkOrderID->text() + "'";

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        QMessageBox::warning(this,"Warning","Could not search database for "
                             "Work Order's ALGORITHMS");

        return;
    }

    while(query.next())
    {
        // Split the parameter value into its components: algorithm name, 
        // processing level, and, if present, modifier.
        parmValueSegments = query.value(0).toString().split("_"); 

        if((parmValueSegments.first() == "RelativeGain") &&
           (parmValueSegments.last() != "SKIP"))
        {
            found_LifetimeRG = true;
        }

        if((parmValueSegments.first() == "HistoRelGainCorr") &&
           (parmValueSegments.last() != "SKIP"))
        {
            found_HistoRG = true;
        }
    }

    if(found_LifetimeRG && found_HistoRG)
    {
        QMessageBox::warning(this,"Warning","Run at most 1 Relative Gain "
            "Correction algorithm (histogram or lifetime model).");

        return;
    }

    // Verify the format of the sceneId/intervalId.
    QRegExp sceneReg("^[A-Z]{2}\\d{14}[A-Z]{3}\\d{2}$");
    // Example L0Rp scene LC80240372000227EDC00
    QRegExp intervalReg("^[A-Z]{2}\\d{17}[A-Z]{3}\\d{2}$");
    // Example L0Ra interval LC82220010042014265LGN00
    QRegExp calibrationReg("^[A-Z]{2}\\d{3}[A-Z]{1}\\d{13}[A-Z]{3}\\d{2}$");
    // Example L0Ra calibration interval LC800U1234562014265LGN00
    QRegExpValidator sceneVal(sceneReg, 0);
    QRegExpValidator intervalVal(intervalReg, 0);
    QRegExpValidator calibrationVal(calibrationReg, 0);
    QString sceneId = leSceneId->text();
    QString l0rDataPath = leL0RDataPath->text().trimmed();
    bool valid = false;
    int pos = 0;

    // Test for sceneId
    if(sceneVal.validate(sceneId, pos) == QValidator::Acceptable)
    {
        valid = true;
    }
    // Test for intervalId
    else if(intervalVal.validate(sceneId, pos) == QValidator::Acceptable)
    {
        valid = true;
    }
    // Test for a calibration intervalId
    else if(calibrationVal.validate(sceneId, pos) == QValidator::Acceptable)
    {
        valid = true;
    }

    // Check if it was valid.
    if(!valid)
    {
        QMessageBox::warning(this,"Warning","The sceneId/intervalId "
                             "did not pass format validation.  Please "
                             "correct it.");

        return;
    }

    /* We validate the path to the L0R data, simply by checking to see if an
       ancillary file for it exists.  We only validate the path if it is not
       empty because subsetter requests will have an empty path. */
    if (!l0rDataPath.isEmpty())
    {
        QFile ancfile(l0rDataPath + "_ANC.h5");
        if (!ancfile.exists())
        {
            QMessageBox::warning(this,"Warning","The L0R Data Path "
                                 "did not pass validation.  Please "
                                 "correct it.");
            return;
        }
    }

    QStringList schemas = IASSchemaConfig::instance()->allKeys();
    QStringList::iterator itr;

    /* Start a transaction before any updates */
    if(!QSqlDatabase::database().transaction())
    {
        QMessageBox::critical(this, "Error",
            "Failed to create a database transaction.");
        return;
    }

    // Begin constructing the update statement for the work_orders table
    sqlStmt = "UPDATE work_orders SET priority = ";
    sqlStmt += cboPriority->currentText();
    sqlStmt += ", request_type = ";

    // Get The letter for the request type
    if(rbStandard->isChecked())
    {
        sqlStmt += "'S'";
    }
    else if(rbTest->isChecked())
    {
        sqlStmt += "'T'";
    }
    else /* Custom */
    {
        sqlStmt += "'C'";
    }
    sqlStmt += ", ephem_type = ";

    // Get the letter for the ephemeris type
    if(rbPCD->isChecked())
    {
        sqlStmt += "'P'";
    }
    else /* Definitive */
    {
        sqlStmt += "'D'";
    }
    sqlStmt += ", output_db_schema = ";

    // Get the output db schema to use
    if(rbProduction->isChecked())
    {
        sqlStmt += "'P'";
    }
    else /* Definitive */
    {
        sqlStmt += "'E'";
    }
    sqlStmt += ", ";

    sqlStmt += "req_start_date = TO_DATE('";
    sqlStmt += dteRequestStartDate->dateTime().toString();
    sqlStmt += "','DY MON DD HH24:MI:SS YYYY'), ";

    sqlStmt += "scene_id = '" + leSceneId->text() + "', ";
    sqlStmt += "procedure_id = '" + cboProcedureId->currentText() + "', ";
    sqlStmt += "cpf_override = '" + leCalParmFile->text() + "', ";
    sqlStmt += "wo_comment = '" + leWorkOrderCmnt->text() + "', ";
    sqlStmt += "wo_path = '" + leWorkOrderPath->text() + "', ";
    sqlStmt += "processing_instance = '" + cboInstance->currentText() + "', ";
    if(cw_l0rDataPath.isEmpty())
    {
        /* We are not in the ModifyWO interface so perform normal actions. */

        sqlStmt += "l0r_data_path = '" + l0rDataPath + "', ";
        if(l0rDataPath.isEmpty())
        {
            /* Set to 0 so that we know the data is requested from the
               subsetter. */
            sqlStmt += "l0r_data_status = 0, ";
            sqlStmt += "state = '" + DB_WO_L0R_NEW + "', ";
        }
        else
        {
            /* Set to 1 so that we know the data is from a local source. */
            sqlStmt += "l0r_data_status = 1, ";
            sqlStmt += "state = '" + DB_WO_PENDING + "', ";
        }
    }
    else
    {
        /* We are in the ModifyWO interface so perform special actions. */
        sqlStmt += "l0r_data_path = '" + cw_l0rDataPath + "', ";
        sqlStmt += "l0r_data_status = " + cw_l0rDataStatus + ", ";
    }

    sqlStmt += "satellite = '" + satellite + "', ";
    sqlStmt += "sensor = '" + sensor + "' ";

    sqlStmt += "WHERE work_order_id = '" + lblWorkOrderID->text() + "'";

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        if(!QSqlDatabase::database().rollback())
        {
            QMessageBox::critical(this, "Error",
                "Failed to rollback the database transaction.");
        }
        return;
    }

    // Set the Pause flags for each script
    sqlStmt  = "SELECT script_id ";
    sqlStmt += "FROM wo_scripts ";
    sqlStmt += "WHERE work_order_id = '" + lblWorkOrderID->text() + "' ";
    sqlStmt += "ORDER by seq_nbr";

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        if(!QSqlDatabase::database().rollback())
        {
            QMessageBox::critical(this, "Error",
                "Failed to rollback the database transaction.");
        }
        return;
    }

    while(query.next())
    {
        // Read pause from window and update pause flag in database

        // Compare script ID from database and form.  This was added as an
        // extra check to make sure that the form table matches the database.
        // They should always match.
        item = m_model->getItem(row, 0);
        if(item->data().toString() != query.value(0).toString())
        {
            QMessageBox::warning(this, "Error",
                                 "Error reading the script id's");

            if(!QSqlDatabase::database().rollback())
            {
                QMessageBox::critical(this, "Error",
                    "Failed to rollback the database transaction.");
            }
            return;
        }

        sqlStmt = "UPDATE wo_scripts SET pause_flag = ";
        item = m_model->getItem(row, 1);
        if(item->data().toString() == "Yes")
        {
            sqlStmt += "1";
        }
        else
        {
            sqlStmt += "0";
        }
        sqlStmt += " ";
        sqlStmt += "WHERE work_order_id = '" +  lblWorkOrderID->text() + "'";
        sqlStmt += " AND script_id = '" + query.value(0).toString() + "'";

        row++;

        if(!queryTmp.exec(sqlStmt, __FILE__, __LINE__))
        {
            if(!QSqlDatabase::database().rollback())
            {
                QMessageBox::critical(this, "Error",
                    "Failed to rollback the database transaction.");
            }
            return;
        }

        queryTmp.next();
    }

    // Check that all required fields are filled.  This must be done after the
    // insert to work_orders table (above) so that all data is accounted for.

    if(leSceneId->text() != "")
    {
        sqlStmt  = "SELECT script_id, wo.parm_name, wo.parm_occurrence ";
        sqlStmt += "FROM wo_parms wo, def_parms dp ";
        sqlStmt += "WHERE wo.work_order_id = '" + lblWorkOrderID->text() + "'";
        sqlStmt += " AND wo.script_id = dp.sub_module_id";
        sqlStmt += " and dp.processing_instance = '" + instance + "' ";
        sqlStmt += " and dp.satellite = '" + satellite + "' ";
        sqlStmt += " and dp.sensor = '" + sensor + "' ";
        sqlStmt += " AND wo.parm_value is NULL";
        sqlStmt += " AND wo.parm_name = dp.parm_name";
        sqlStmt += " AND wo.parm_occurrence = dp.parm_occurrence";
        sqlStmt += " AND dp.module_id = '" +cboProcedureId->currentText()+ "'";
        sqlStmt += " AND dp.required = 1";

        if(!query.exec(sqlStmt, __FILE__, __LINE__))
        {
            if(!QSqlDatabase::database().rollback())
            {
                QMessageBox::critical(this, "Error",
                    "Failed to rollback the database transaction.");
            }
            return;
        }

        /* Gather any results */
        QString missingRequiredParameters = "";
        while (query.next())
        {
            missingRequiredParameters += "\n"
                + query.value("SCRIPT_ID").toString() + "."
                + query.value("PARM_NAME").toString() + "."
                + query.value("PARM_OCCURRENCE").toString();
        }

        if(missingRequiredParameters.length() > 0)
        {
            /* Found missing required parameters so halt the work order */
            sqlStmt  = "UPDATE work_orders ";
            sqlStmt += " SET state = '" + DB_WO_HALTED + "'";
            sqlStmt += " WHERE work_order_id = '" +lblWorkOrderID->text()+ "'";

            if(!query.exec(sqlStmt, __FILE__, __LINE__))
            {
                if(!QSqlDatabase::database().rollback())
                {
                    QMessageBox::critical(this, "Error",
                        "Failed to rollback the database transaction.");
                }
                return;
            }

            /* Report the errors */
            QMessageBox::warning(this, "Error",
                "You must enter values for the Required Parameters\n"
                + missingRequiredParameters);

            if(!QSqlDatabase::database().rollback())
            {
                QMessageBox::critical(this, "Error",
                    "Failed to rollback the database transaction.");
            }
            return;
        }
    }

    // Commit the changes
    if(!QSqlDatabase::database().commit())
    {
        QMessageBox::critical(this, "Error",
                "Commit of database changes failed",
                QMessageBox::Ok, QMessageBox::NoButton, 
                QMessageBox::NoButton);
    }

    // Set this because we don't want the closeEvent to delete the work order.
    m_deletedFlag = true;

    // Close the window
    close();
}

void CreateWO::getCalParmFile()
{
    QString filename = QFileDialog::getOpenFileName(this,
        "Specify - Calibration Parameter Filename", ".", QString(), 0, 
        QFileDialog::DontUseNativeDialog);

    if(!filename.isEmpty())
    {
        leCalParmFile->clear(); // Make sure it is empty.
        leCalParmFile->insert(filename);
    }
}

void CreateWO::getL0RDataPath()
{
    /* Create a pattern matching the naming scheme after the sceneId for
       L0R data.  This pattern is used to decide if the filename is valid
       for a sceneId and also used to remove the contents after the sceneId
       so the strings end with the sceneId. */
    QRegExp search_pattern =
        QRegExp("_(MTA|ANC|MD5|B[1-9][0-7]*).(h5|txt)$");

    // Use the path given in the leL0RDataPath text box, as the directory to
    // to open to, when possible.

    QString dir = ".";
    if (!leL0RDataPath->text().isEmpty())
    {
        QFileInfo file(leL0RDataPath->text());
        QFileInfo path(file.path());
        if (path.isDir())
        {
            dir = file.path();
        }
    }
    QString filename = QFileDialog::getOpenFileName(this,
        "Specify - L0R Data Path and Filename", dir, QString(), 0, 
        QFileDialog::DontUseNativeDialog);

    /* Make sure we think we have a scene filename selected before attempting
       to populate the scene ID lineedit. */
    if (filename.contains(search_pattern))
    {
        filename.remove(search_pattern);

        QStringList parts = filename.split("/");

        /* The sceneId is cleared here because the user may have accidentally
           selected the wrong sceneId and will want to have it displayed
           and/or correctly identified in the database/OUI. */
        leSceneId->clear();
        leSceneId->insert(parts[parts.count()-1]);
    }

    if (!filename.isEmpty())
    {
        leL0RDataPath->clear(); // Make sure it is empty.
        leL0RDataPath->insert(filename);
    }
}

void CreateWO::getWorkOrderPath()
{
    QString filename = QFileDialog::getExistingDirectory(this,
        "Specify - Work Order Directory Path", ".",
        QFileDialog::DontUseNativeDialog | QFileDialog::ShowDirsOnly);

    if(!filename.isEmpty())
    {
        leWorkOrderPath->clear(); // Make sure it is empty.
        leWorkOrderPath->insert(filename);
    }
}

