/******************************************************************************
               Image Assessment System (IAS) Software
 ******************************************************************************
 *
 * Name:   Events.cpp
 *
 * PURPOSE:
 * This is the definition of the Events class which is used to view 
 * messages in the events table and display the status of the tasks.
 *
 *****************************************************************************/

// System Includes
#include <iostream>

// Qt Includes
#include <QtCore>
#include <QtGui>
#include <QtSql>

// IAS Qt Includes
#include <IASListView.h>
#include <IASListViewItem.h>
#include <IASSqlQuery.h>
#include <IASUtilities.h>
#include <IASQtSystemStatus.h>

// Local Includes
#include "Events.h"

#define TASK_WIDTH 165
#define STATUS_WIDTH 60

//Set up parameters for the events table listview
lvFields eventsFields[] = 
{
    {"Date/Time", "EVENT_DATE", "EVENTS", DateTime, -1},
    {"WO Id", "WORK_ORDER_ID", "EVENTS", Id, -1},
    {"Script Id", "SCRIPT_ID", "EVENTS", Alpha, -1},
    {"Task Id", "PROGRAM_ID", "EVENTS", Alpha, -1},
    {"Brief Message", "MESSAGE_ID", "MESSAGES", Alpha, -1}
};

const QString UP_COLOR_STYLE = "background-color : #44ff44; color : black;";
const QString DOWN_COLOR_STYLE = "background-color : #ff4444; color : black;";

//***********************************************************************
Events::Events(QWidget* parent,    //I: The address to this widget's parent
               Qt::WindowFlags fl) //I: Used by QT
: QWidget(parent, fl)
{
    // Constructor sets up and populates listviews with initial values
    // and starts the update timer.

    QWidget::setAttribute(Qt::WA_DeleteOnClose);

    // Set up the user interface.
    setupUi(this);

    lvEvents->setRootIsDecorated(false);

    QFont f = this->font();
    f.setBold(false);
    f.setItalic(false);

    m_systemId = getenv( "IAS_SYSTEM_ID" );
    if (m_systemId.isEmpty())
    {
        QMessageBox::critical(0,"Missing Environment",
            "Could not get IAS_SYSTEM_ID from environment.\n"
            "This variable is required for system startup or shutdown",
            QMessageBox::Ok,0);
        close();
    }
    m_systemId = m_systemId.toUpper();
    /* Using only the first character since the env. var. could be longer. */
    m_systemId = m_systemId.at(0);

    cboWorkOrder->setFont(f);
    cboTaskId->setFont(f);

    m_numEventsFields = ARRAY_SIZE(eventsFields);
    m_lastInterval = -1;  // set to this value so filters will be updated

    for (int i = 0; i <  10; i ++)
    {
        lblTask[i]  = NULL;
        lblStatus[i] = NULL;
    }

    // Set up the events and tasks listviews
    //--------------------------------------
    setupListViews();

    // Set up the filters
    //-------------------
    setupFilters();

    // Start the timer and make it a single shot.  We will restart it
    // again after it expires and the listviews have been populated.
    // This eliminates the possibility of the window becoming locked
    // up if very large queries are executed with a very short refresh rate
    //---------------------------------------------------------------------
    queryTimer = new QTimer(this);
    queryTimer->setSingleShot(true);
    queryTimer->start(spinRefRate->value() * 1000);

    // Connect signals and slots
    //--------------------------
    connect(queryTimer, SIGNAL(timeout()), this, SLOT(timerExpired()));
    connect(spinRefRate, SIGNAL(valueChanged(int)),
            this, SLOT(resetTimer(int)));
    connect(cboWorkOrder, SIGNAL(activated(int)), this, SLOT(changedWO()));
    connect(cboTaskId, SIGNAL(activated(int)), this, SLOT(changedTaskId()));
    connect(lvEvents, SIGNAL(itemSelectionChanged()),
            this, SLOT(dispMessage()));
    connect(pbClose, SIGNAL(clicked()), this, SLOT(close()));
    connect(pbQuery, SIGNAL(clicked()), this, SLOT(issueQuery()));

    // Populate the task status and events listviews
    //----------------------------------------------
    updateTaskStatus();
    updateEvents();
}

//***********************************************************************
void Events::setupListViews()
{
    // Sets up the task and events listviews

    int i;
    QStringList labels;
    QHeaderView* hdr = NULL;

    hdr = lvEvents->header();

    // Add columns to the events listview
    //-----------------------------------
    labels.clear();
    for (i=0; i < m_numEventsFields; i++)
    {
        labels.append(eventsFields[i].colName);
        m_colTypesList.append(eventsFields[i].colType);
    }
    lvEvents->setHeaderLabels(labels);

    QFont f = frmTaskStatus->font();
    f.setBold(true);

    lblTask[0] = new QLabel(frmTaskStatus);
    lblTask[0]->setMinimumSize(QSize(TASK_WIDTH,0));
    lblTask[0]->setFrameShape(QLabel::WinPanel);
    lblTask[0]->setFrameShadow(QLabel::Raised);
    lblTask[0]->setLineWidth(2);
    lblTask[0]->setText(trUtf8("Task"));
    lblTask[0]->setAlignment(Qt::AlignCenter);
    lblTask[0]->setMaximumSize(QSize(TASK_WIDTH,20));
    lblTask[0]->setFont(f);
    layoutTasks->addWidget(lblTask[0]);

    lblStatus[0] = new QLabel(frmTaskStatus);
    lblStatus[0]->setMinimumSize(QSize(STATUS_WIDTH,0));
    lblStatus[0]->setFrameShape(QLabel::WinPanel);
    lblStatus[0]->setFrameShadow(QLabel::Raised);
    lblStatus[0]->setLineWidth(2);
    lblStatus[0]->setText(trUtf8("Status"));
    lblStatus[0]->setAlignment(Qt::AlignCenter);
    lblStatus[0]->setMaximumSize(QSize(STATUS_WIDTH,20));
    lblStatus[0]->setFont(f);
    layoutStatus->addWidget(lblStatus[0]);

    m_TaskCount = 1; //  The header will be displayed

    hdr->setSortIndicatorShown(false);
    lvEvents->setSortingEnabled(false);

    // Resize last column to listview width
    //-------------------------------------
    hdr->setStretchLastSection(true);
}

//***********************************************************************
void Events::updateTaskStatus()
{
    /* This function gets the task IDs and process IDs of each task
       from the tasks table and updates the task statuses. */
    IASSqlQuery query;

    QString sqlStmt = "SELECT task_id, proc_id"
                      " FROM tasks"
                      " WHERE system_id = '" + m_systemId + "'"
                      " ORDER BY task_id";

    query.exec(sqlStmt,__FILE__,__LINE__);
    if(query.isActive())
    {
        m_TaskCount = 1;
    }

    QLabel* taskLbl = NULL;
    QLabel* statusLbl = NULL;
    while(query.next())
    {
        /* Get the values from the query results. */
        QString taskName = query.value("TASK_ID").toString();
        QString processId = query.value("PROC_ID").toString();

        if(lblTask[m_TaskCount] == NULL)
        {
            /* Setup the label that will hold the name of the task. */
            taskLbl = new QLabel(frmTaskStatus);

            taskLbl->setAutoFillBackground(true);
            taskLbl->setFrameShape(QLabel::StyledPanel);
            taskLbl->setFrameShadow(QLabel::Sunken);
            taskLbl->setAlignment(Qt::AlignCenter);
            taskLbl->setMaximumSize(QSize(TASK_WIDTH, 20));

            layoutTasks->addWidget(taskLbl);
            lblTask[m_TaskCount] = taskLbl;
        }
        lblTask[m_TaskCount]->setText(taskName);
            
        if(lblStatus[m_TaskCount] == NULL)
        {
            /* Setup the label that will hold the status of the task. */
            statusLbl = new QLabel(frmTaskStatus);

            statusLbl->setAutoFillBackground(true);
            statusLbl->setFrameShape(QLabel::StyledPanel);
            statusLbl->setFrameShadow(QLabel::Sunken);
            statusLbl->setAlignment(Qt::AlignCenter);
            statusLbl->setMaximumSize(QSize(STATUS_WIDTH, 20));
            layoutStatus->addWidget(statusLbl);
            lblStatus[m_TaskCount] = statusLbl;
        }

        taskLbl = lblTask[m_TaskCount];
        statusLbl = lblStatus[m_TaskCount];

        /* Determine if the task is up or down and color appropriatly. */
        if(SystemStatus::isTaskRunning(taskName, processId))
        {
            statusLbl->setText("UP");
            statusLbl->setStyleSheet(UP_COLOR_STYLE);
            taskLbl->setStyleSheet(UP_COLOR_STYLE);
        }
        else
        {
            statusLbl->setText("DOWN");
            statusLbl->setStyleSheet(DOWN_COLOR_STYLE);
            taskLbl->setStyleSheet(DOWN_COLOR_STYLE);
        }

        m_TaskCount++;
    }
}

//***********************************************************************
void Events::setupFilters()
{
    IASSqlQuery query;
    QString sqlStmt;
    QDate date = QDate::currentDate();
    QMultiMap<double, QString> workOrderIds;

    // Populates the work order ID and task ID filters for the number
    // of days specified in the days filter

    // If the number of days to display has not changed then we do not
    // need to update the filters
    //----------------------------------------------------------------
    if (spinTime->value() == m_lastInterval)
        return;
    else
        m_lastInterval = spinTime->value();
    
    QString endDate =
        date.addDays(-(spinTime->value() - 1)).toString("dd-MMM-yyyy");
    
    // Insert list for the work order id filter combo box, first
    // string is "All" so the filter is initially turned off.
    //----------------------------------------------------------
    sqlStmt = "select unique work_order_id from events where event_date >= '" +
              endDate + "' and system_id = '" + m_systemId + "'";

    cboWorkOrder->clear();
    cboWorkOrder->addItem("All");
    
    query.exec(sqlStmt, __FILE__, __LINE__);
    if (query.isActive())
    {
        while (query.next())
        {
            //Only add to the QMap, if it actually contains text
            if(!query.value(0).toString().isEmpty())
            {
                bool ok = false;
                double idNumber = query.value(0).toString().remove(QRegExp("\\D")).
                    toDouble(&ok); // extract digits from the WO Id 
                if(ok)
                {
                    workOrderIds.insert(idNumber, query.value(0).toString());
                }
            }
        }
        foreach(QString val, workOrderIds)
        {
            cboWorkOrder->addItem(val);
        }
    }

    // Insert list for the task ID filter combo box, first
    // string is "All" so the filter is initially turned off.
    //-------------------------------------------------------
    sqlStmt = "select unique program_id from events where event_date >= '" +
              endDate + "' and system_id = '" + m_systemId + "'";

    cboTaskId->clear();
    cboTaskId->addItem("All");
    
    query.exec(sqlStmt, __FILE__, __LINE__);
    if (query.isActive())
    {
        while (query.next())
        {
            cboTaskId->addItem(query.value(0).toString());
        }
    }
}

//***********************************************************************
void Events::updateEvents()
{
    // This function loads the events listview using values from the
    // current filters.

    int i;

    IASListViewItem *lviNew = NULL;
    QDate date = QDate::currentDate();
    IASSqlQuery query;
    
    // Create the query based on the work order, date, and task ID filters
    // Since the rows are ordered by sequence number, they will also be
    // ordered by date/time
    //--------------------------------------------------------------------
    QString sqlStmt = "Select to_char(e.event_date, 'YYYY-MM-DD/HH24:MI:SS'),"
        " e.work_order_id, e.script_id, e.program_id, m.message_text,"
        " e.event_comment from events e, messages m where m.message_id"
        " = e.message_id and e.system_id = '" + m_systemId + "'";

    sqlStmt += " and e.event_date >= '" +
               date.addDays(-(spinTime->value() - 1)).toString("dd-MMM-yyyy") +
               "'";

    if (cboWorkOrder->currentText() != "All")
        sqlStmt += " and work_order_id = '" + cboWorkOrder->currentText() + "'";

    if (cboTaskId->currentText() != "All")
        sqlStmt += " and program_id = '" + cboTaskId->currentText() + "'";

    sqlStmt += " order by seq_nbr desc";

    query.exec(sqlStmt, __FILE__, __LINE__);

    // Set the cursor to the wait shape in case the query takes several
    // seconds to complete
    //-----------------------------------------------------------------
    QApplication::setOverrideCursor(Qt::WaitCursor);

    // If the query is successful clear items and insert the rows, store
    // the detailed message in the IASListViewItem tag.
    //-------------------------------------------------------------------
    if (query.isActive())
        lvEvents->clear();
    leDetMsg->clear();
    while (query.next())
    {
        lviNew = new IASListViewItem(lvEvents);

        for (i = 0; i < m_numEventsFields; i++)
            lviNew->setText(i, (query.value(i).toString()));

        lviNew->setTag(query.value(m_numEventsFields).toString());
    }

    /* Resize the columns to fit the data. */
    for (i = 0; i < m_numEventsFields; i++)
    {
        lvEvents->resizeColumnToContents(i);
    }

    // Always select the first row initially (if there is one)
    //--------------------------------------------------------
    if (lvEvents->topLevelItem(0) != 0)
        lvEvents->setCurrentItem(lvEvents->topLevelItem(0), true);

    // Check if the filters should be updated and restore the
    // cursor to the shape it initially was
    //-------------------------------------------------------
    setupFilters();
    QApplication::restoreOverrideCursor();
}

//***********************************************************************
void Events::timerExpired()
{
    // The timer has expired so update the task and events information
    // then reset the timer

    updateTaskStatus();
    updateEvents();
    queryTimer->setSingleShot(true);
    queryTimer->start(spinRefRate->value() * 1000);
}

//***********************************************************************
void Events::resetTimer(int timeInterval)
{
    // The refresh rate has been changed so change the timer interval

    queryTimer->setInterval(timeInterval * 1000);
}

//***********************************************************************
void Events::dispMessage()
{
    // Displays the detailed message which corresponds to the selected
    // row in the events listview

    if (lvEvents->currentItem())
    {
        leDetMsg->setText(
            dynamic_cast<IASListViewItem*>(lvEvents->currentItem())->getTag());
        leDetMsg->home(FALSE);  // Make sure beginning of message is displayed
    }
}

//***********************************************************************
void Events::issueQuery()
{
    /* Force the timer to trip by setting the interval to 0. */
    queryTimer->setInterval(0);
}

//***********************************************************************
void Events::changedWO()
{
    // This slot is called when the work order filter is changed.  It
    // changes the task filter to allow all tasks and updates the 
    // events listview

    cboTaskId->setCurrentIndex(0);
    updateEvents();
}

//***********************************************************************
void Events::changedTaskId()
{
    // This slot is called when the task ID filter is changed.  It
    // changes the work order filter to allow all work orders and updates
    // the events listview

    cboWorkOrder->setCurrentIndex(0);
    updateEvents();
}

