/******************************************************************************
               Image Assessment System (IAS) Software
 ******************************************************************************
 *
 * Name:   ModGlobalWOParms.cpp
 *
 * PURPOSE:
 * This is the definition of the ModGlobalWOParms class which is used to 
 * modify parameter values for a given work order and procedure.
 *
 *****************************************************************************/

// System Includes

// Qt Includes 
#include <QtCore>
#include <QtGui>
#include <QtSql>

// Local Includes
#include "ModGlobalWOParms.h"
#include "IASUtilities.h"
#include "IASSqlQuery.h"
#include "IASParameterDelegate.h"
#include "IASParameterModel.h"

//****************************************************************************
ModGlobalWOParms::ModGlobalWOParms(
    const QString workOrderId, //I: Modify parameters for this work order
    const QString procedureId, //I: Modify parameters for this procedure
    const QString instance,  //I: Instance name used throughout this session
    const QString satellite,  //I: Satellite used throughout this session
    const QString sensor,  //I: Sensorused throughout this session
    QWidget* parent,           //I: The address to this widget's parent
    Qt::WindowFlags fl)        //I: Used by QT
: QDialog(parent, fl)
{
    // The constructor sets up the parameter table and labels and
    // begins a transaction

    // Set up the user interface.
    setupUi(this);

    // Set the work order and procedure ID labels
    //-------------------------------------------
    lblWorkOrderIdVal->setText(workOrderId);
    lblProcedureIdVal->setText(procedureId);

    // Define the instance,satellite, and sensor to use throughout the session
    //-------------------------------------------
    m_Instance = instance;
    m_Satellite = satellite;
    m_Sensor = sensor;

    // Set up the global parameter table
    //----------------------------------
    createParmTable();

    // Display the initial parameter description
    //------------------------------------------
    tblCommonParms->setCurrentIndex(m_model->index(0, 0));
    dispParmDesc(m_model->index(0, 0));

    // Initiate a transaction
    //-----------------------
    QSqlDatabase::database().transaction();

    // Connect signals and slots
    //--------------------------

    QItemSelectionModel* selModel = tblCommonParms->selectionModel();
    connect(selModel,
            SIGNAL(currentChanged(const QModelIndex&, const QModelIndex&)),
            this,
            SLOT(dispParmDesc(const QModelIndex&)));

    connect(m_model, SIGNAL(itemDataChanged(const QModelIndex&)),
            this, SLOT(updateDBase(const QModelIndex&)));

    connect(pbOK, SIGNAL(clicked()), this, SLOT(commitAndClose()));
    connect(pbCancel, SIGNAL(clicked()), this, SLOT(rollbackAndClose()));
}

//***************************************************************************
void ModGlobalWOParms::createParmTable()
{
    // Creates the common parameter table and sets all of the
    // initial values of the table.

    QFont f = this->font();
    IASSqlQuery query;

    QStringList labels;

    m_model = new IASParameterModel;
    IASParameterItemRow* itemRow = NULL;
    IASParameterItem* item = NULL;
    m_model->setColumnCount(4);

    labels << "Parameter Name" << "Element" << "Type" << "Parameter Value";
    m_model->setHorizontalHeaderLabels(labels);

    // Set up the delegate for the table.
    tblCommonParms->setItemDelegate(new IASParameterDelegate);

    // Set up the edit triggers.
    tblCommonParms->setEditTriggers(QAbstractItemView::AllEditTriggers);

    // Set the rows to alternate colors.
    tblCommonParms->setAlternatingRowColors(true);

    // Make headings bold and set focus and selection
    f.setBold(true);
    tblCommonParms->horizontalHeader()->setFont(f);

    // Only select one row at a time.
    tblCommonParms->setSelectionMode(QAbstractItemView::SingleSelection);
    tblCommonParms->setSelectionBehavior(QAbstractItemView::SelectRows);

    // Get rid of the vertical header.
    tblCommonParms->verticalHeader()->hide();

    // Add rows to the table.  The parameter value column will
    // contain a lineedit or else a combobox item if there is a
    // defined set of possible values.
    //---------------------------------------------------------
    QString sqlStmt = "Select parm_name, parm_occurrence, parm_value from "
        "wo_common_parms where work_order_id='" + lblWorkOrderIdVal->text()
        + "' order by parm_name, parm_occurrence";

    QVariant data;
    if(query.exec(sqlStmt, __FILE__, __LINE__))
    {
        while(query.next())
        {
            itemRow = new IASParameterItemRow;

            /* 0 - Parameter Name */
            item = new IASParameterItem;
            data = query.value(0).toString();
            item->setData(data);
            itemRow->append(item);

            /* 1 - Element */
            item = new IASParameterItem;
            data = query.value(1).toString();
            item->setData(data);
            itemRow->append(item);

            /* 2 - Type */
            item = new IASParameterItem;
            data = getParmType(query.value(0).toString(),
                               query.value(1).toString());
            item->setData(data);
            itemRow->append(item);

            /* 3 - Parameter Value */
            item = new IASParameterItem;
            // This item has special configuration.
            configParmValueItem(item, query.value(0).toString(),
                                      query.value(2).toString());
            itemRow->append(item);

            m_model->appendRow(itemRow);
        }

        // Set the model
        tblCommonParms->setModel(m_model);

        // Adjust the columns to the new data.
        tblCommonParms->resizeColumnsToContents();

        // Set last column to be stretchable
        tblCommonParms->horizontalHeader()->setStretchLastSection(true);
    }
}

//***********************************************************************
QString ModGlobalWOParms::getParmType(const QString& parmName,
                                      const QString& parmOccur)
{
    // Returns the type of the parameter in the row

    IASSqlQuery query;
    QString result = "";
    QString sqlStmt;

    sqlStmt = "Select parm_type from def_parms where module_id ='";
    sqlStmt += lblProcedureIdVal->text() + "' ";
    sqlStmt += "and parm_name='" + parmName + "' ";
    sqlStmt += "and parm_occurrence='" + parmOccur + "' ";
    sqlStmt += "and processing_instance = '" + m_Instance + "' ";
    sqlStmt += "and satellite = '" + m_Satellite + "' ";
    sqlStmt += "and sensor = '" + m_Sensor + "' ";

    // Select the parameter type in the def_parms table corresponding to the
    // parameter name in the first column of the table
    //----------------------------------------------------------------------
    if(query.exec(sqlStmt, __FILE__, __LINE__))
    {
        query.next();
    }
    else
    {
        return result;
    }

    if(query.isValid())
    {
        if(query.value(0).isValid())
        {
            result = query.value(0).toString();
        }
    }

    return result; // The query or query value was not valid.
}

//***********************************************************************
void ModGlobalWOParms::configParmValueItem(IASParameterItem* item,
                                           const QString& parmName, 
                                           const QString& value)
{
    // Checks to see if there are a fixed set of parameter values
    // and whether the value read from the database should be mapped to
    // a different string to be shown to the user.  A combobox will be
    // inserted if there are a fixed set of values for the parameter.

    IASSqlQuery query;
    QStringList ouiValues;
    QVariant data;

    QString sqlStmt = "Select oui_value, program_value from "
                      "parm_list where parm_name = '" +  parmName + "' "
                      "and processing_instance = '" + m_Instance + "' "
                      "and satellite = '" + m_Satellite + "' "
                      "and sensor = '" + m_Sensor + "' ";

    // Create a stringlist with the possible values for the parameter.
    if(query.exec(sqlStmt, __FILE__, __LINE__))
    {
        while(query.next())
        {
            ouiValues += query.value(0).toString();
        }
    }
    else
    {
        return;
    }

    data = value;
    item->setData(data);
    item->setEditable(true);

    // Create a combo box in the cell if there is a set of values for the
    // parameter.  Otherwise use a line edit.  Set the combo box or line
    // edit to the wo_common_parms table value.
    if (!ouiValues.isEmpty())
    {
        /* Sort the options with a special routine if they are for
           the BAND_LIST */
        if(parmName == "BAND_LIST")
        {
            qSort(ouiValues.begin(), ouiValues.end(),
                  IASUtilities::numericLessThan);
        }

        item->setType(IASParameterItem::TextSelection);
        item->setOptions(ouiValues);
    }
    // Else leave it alone, since the default is a line edit.
}

//***********************************************************************
void ModGlobalWOParms::dispParmDesc(const QModelIndex& index)
{
    // Displays the description of the parameter in the row from the
    // parm_edits table and updates the current parameter value.

    IASSqlQuery query;

    int row = index.row();

    IASParameterItem* item = m_model->getItem(row, 3);

    m_CurrentParmVal = item->data().toString();

    item = m_model->getItem(row, 0);
    QString sqlStmt = "Select parm_desc from "
                      "parm_edits where parm_name='";
    sqlStmt += item->data().toString() + "' "
        "and processing_instance = '" + m_Instance + "' "
        "and satellite = '" + m_Satellite + "' "
        "and sensor = '" + m_Sensor + "' ";

    if(query.exec(sqlStmt, __FILE__, __LINE__))
    {
        query.next();
    }
    else
    {
        return;
    }

    if(query.isValid())
    {
        teParmDesc->setText(query.value(0).toString());
    }
    else
    {
        teParmDesc->setText("");
    }
}

//***********************************************************************
void ModGlobalWOParms::updateDBase(const QModelIndex& index)
{
    // This function checks the parameters updated value against the max
    // and min (if there is one) and updates the wo_parms table with the
    // new value.  We need to verify if the fields value was really updated
    // since the valueUpdated signal seems to be emitted whether the value
    // was actually changed or not.

    int row = index.row();
    int column = index.column();

    IASParameterItem* item = m_model->getItem(row, column);

    if(item->data().toString() == m_CurrentParmVal) // no change
    {
        return;
    }

    // Validate the parameter unless it is from a combobox.  Values in
    // a combobox are all assumed to be valid since they were read from
    // the database and there is only a discrete set of values to choose
    // from.
    //------------------------------------------------------------------
    if(item->type() != IASParameterItem::TextSelection) // Not a combobox value
    {
        if(!checkMaxMin(row)) // Out of range, reset value to original
        {
            QVariant data = m_CurrentParmVal;
            item->setData(data);

            tblCommonParms->setCurrentIndex(m_model->index(row, column));

            return;
        }
    }

    // If the database update is successful reset the current value
    // to the newly updated value.
    if(updateTables(row))
    {
        m_CurrentParmVal = item->data().toString();
    }
}

//***********************************************************************
bool ModGlobalWOParms::updateTables(int row)
{
    // Updates the wo_common_parms and wo_parms tables with the new
    // parameter value

    IASSqlQuery query;

    IASParameterItem* item = m_model->getItem(row, 3);
    QString parmValue = item->data().toString();
    item = m_model->getItem(row, 2);
    QString parmType = item->data().toString();
    item = m_model->getItem(row, 0);
    QString parmName = item->data().toString();
    item = m_model->getItem(row, 1);
    QString parmOccur = item->data().toString();

    // Update the wo_common_parms table first
    QString sqlStmt = "update wo_common_parms set parm_value=";
    if(parmValue.length() == 0)
    {
        sqlStmt += "NULL";
    }
    else if((parmType.toUpper() != "INT") && (parmType.toUpper() != "LONG"))
    {
        sqlStmt += "'" + parmValue + "'";
    }
    else
    {
        sqlStmt += parmValue;
    }

    sqlStmt += " where work_order_id='" + lblWorkOrderIdVal->text() + "'";
    sqlStmt += " and parm_name='" + parmName + "'";
    sqlStmt += " and parm_occurrence=" + parmOccur;

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        return false;  // update was unsuccessfull
    }

    // Update the wo_parms table.  Only update rows for the current
    // procedure and which have the global flag set.
    sqlStmt = "update wo_parms set parm_value=";

    if(parmValue.length() == 0)
    {
        sqlStmt += "NULL";
    }
    else if(parmType.toUpper() == "CHAR")
    {
        sqlStmt += "'" + parmValue + "'";
    }
    else
    {
        sqlStmt += parmValue;
    }

    sqlStmt += " where work_order_id='" + lblWorkOrderIdVal->text() + "'";
    sqlStmt += " and parm_name='" + parmName + "'";
    sqlStmt += " and parm_occurrence=" + parmOccur;
    sqlStmt += " and script_id in ";
    sqlStmt += "(select sub_module_id from def_parms";
    sqlStmt += " where global=1";
    sqlStmt += " and parm_name='" + parmName + "'";
    sqlStmt += " and parm_occurrence=" + parmOccur;
    sqlStmt += " and processing_instance = '" + m_Instance + "' ";
    sqlStmt += " and satellite = '" + m_Satellite + "' ";
    sqlStmt += " and sensor = '" + m_Sensor + "' ";
    sqlStmt += " and module_id='" + lblProcedureIdVal->text() + "')";

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        return false; // update was unsuccessful
    }

    return true;
}

//***********************************************************************
bool ModGlobalWOParms::checkMaxMin(int row)
{
    // Checks the value of the parameter against the max/min in the
    // parm_edits table (if there is one).

    bool ok;

    IASParameterItem* item = m_model->getItem(row, 3);
    QString parmValue = item->data().toString();
    item = m_model->getItem(row, 2);
    QString parmType = item->data().toString();
    item = m_model->getItem(row, 0);
    QString parmName = item->data().toString();
    item = m_model->getItem(row, 1);
    QString parmOccur = item->data().toString();

    // make sure the value is valid and return false if not
    //-----------------------------------------------------
    if((parmType.toUpper() == "INT") || (parmType.toUpper() == "LONG"))
    {
        parmValue.toLong(&ok);
        if (!ok)
        {
            QMessageBox::warning(this, "Error",
                "The Parameter value is not a valid integer number",
                QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);

            return false;
        }
    }
    else if((parmType.toUpper() == "FLOAT") || (parmType.toUpper() == "DOUBLE"))
    {
        parmValue.toDouble(&ok);
        if (!ok)
        {
            QMessageBox::warning(this, "Error",
                "The Parameter value is not a valid floating point number",
                QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);

            return false;
        }
    }
    else 
    {
        return true;
    }
    
    // If there is a max/min associated with the parameter make sure it is
    // within range.  Return false if not so the out of range value will be
    // reset.
    //----------------------------------------------------------------------
    IASSqlQuery query;

    QString sqlStmt = "Select min_value, max_value from parm_edits where ";
    sqlStmt += "parm_name='" + parmName + "' ";
    sqlStmt += "and processing_instance = '" + m_Instance + "' ";
    sqlStmt += "and satellite = '" + m_Satellite + "' ";
    sqlStmt += "and sensor = '" + m_Sensor + "' ";

    if(query.exec(sqlStmt, __FILE__, __LINE__))
    {
        query.next();
    }

    if(!query.isNull(0))   // There is a max/min
    {
        bool validVal = true;

        if((parmType.toUpper() == "INT") || (parmType.toUpper() == "LONG"))
        {
            if((parmValue.toInt(&ok) < query.value(0).toInt(&ok)) ||
               (parmValue.toInt(&ok) > query.value(1).toInt(&ok)))
            {
                validVal = false;
            }
        }
        else if((parmType.toUpper() == "FLOAT") ||
                (parmType.toUpper() == "DOUBLE"))
        {
            if((parmValue.toDouble(&ok) < query.value(0).toDouble(&ok)) ||
               (parmValue.toDouble(&ok) > query.value(1).toDouble(&ok)))
            {
                validVal = false;
            }
        }

        // If the value is not valid report the min/max and type
        // information for the parameter and return false to prevent
        // update with this value
        //-----------------------------------------------------------
        if(!validVal)
        {
            QMessageBox::warning(this, "Error", 
                "The Parameter value is invalid, should be:\nType = " +
                parmType + "\n" +
                "Max = " + query.value(0).toString() + "\n" +
                "Min = " + query.value(1).toString(),
                QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);

            return false;
        }
    }

    return true;
}

//***********************************************************************
void ModGlobalWOParms::commitAndClose()
{
    // Commit all changes for this transaction and close dialogue.
    // Select the previous column in the row to end the edit and emit the
    // value changed signal. This will check for an update of the current
    // cell before committing all changes.

    if(!QSqlDatabase::database().commit())
    {
        QMessageBox::warning(this, "Error", 
            "Commit of database changes failed",
            QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
    }

    close();
}

//***********************************************************************
void ModGlobalWOParms::rollbackAndClose()
{
    // Rollback all changes for this transaction and close dialogue

    if(!QSqlDatabase::database().rollback())
    {
        QMessageBox::warning(this, "Error",
            "Rollback of database changes failed",
            QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
    }

    close();
}

