/******************************************************************************
               Image Assessment System (IAS) Software
 ******************************************************************************
 *
 * Name:   ModScriptWOParms.cpp
 *
 * PURPOSE:
 * This is the definition of the ModScriptWOParms class which is used to 
 * modify parameter values and script status for scripts in a particular
 * work order.
 *
 *****************************************************************************/

// System Includes
#include <iostream>

// Qt Includes 
#include <QtCore>
#include <QtGui>
#include <QtSql>

// IAS Qt Includes 
#include <iasQtWoConst.h>

// Local Includes/
#include "ModScriptWOParms.h"
#include "IASUtilities.h"
#include "IASSqlQuery.h"
#include "IASParameterDelegate.h"
#include "IASParameterModel.h"

//****************************************************************************
ModScriptWOParms::ModScriptWOParms(
    const QString workOrderId, // I: Modify parameters for this work order
    const QString procedureId, // I: Modify parameters for this procedure
    const QString instance,    // I: Instance name used throughout this session
    const QString satellite,   // I: Satellite used throughout this session
    const QString sensor,      // I: Sensorused throughout this session
    const QString scriptId,    // I: Script to be displayed
    bool allowParmChanges,     // I: Allow the user to reset the order or not
    bool allowReset,           // I: Allow the user to reset the order or not
    QWidget* parent)           // I: The address to this widget's parent
: QDialog(parent)
{
    // The constructor sets up the scripts and script parameter tables
    // and sets all of the initial values of the tables

    // Set up the user interface.
    setupUi(this);

    // Set the work order and procedure ID labels
    lblWorkOrderIdVal->setText(workOrderId);
    lblProcedureIdVal->setText(procedureId);

    //* Set up the resetAllowed */
    resetAllowed = allowReset;
    //* Set up the parmChangedtAllowed */
    parmChangesAllowed = allowParmChanges;

    // Define the instance,satellite, and sensor to use throughout the session
    m_Instance = instance;
    m_Satellite = satellite;
    m_Sensor = sensor;

    // Set up the scripts table
    setupScriptsTable();

    // Set up the script parameters table
    setupParmsTable();

    // Initiate a transaction
    QSqlDatabase::database().transaction();

    // Connect signals and slots
    QItemSelectionModel* selModel = tblScripts->selectionModel();
    connect(selModel,
            SIGNAL(currentChanged(const QModelIndex&, const QModelIndex&)),
            this,
            SLOT(scriptChanged(const QModelIndex&, const QModelIndex&)));

    selModel = tblParms->selectionModel();
    connect(selModel,
            SIGNAL(currentChanged(const QModelIndex&, const QModelIndex&)),
            this,
            SLOT(dispParmDesc(const QModelIndex&, const QModelIndex&)));
    selModel = NULL;

    m_initialWoState = getCurrentWOState(workOrderId);

    if (!resetAllowed)
    {
        /* For all other states we hide this because those states can not
           be advanced/changed to the Pending state. */
        grpboxResetWOAndScriptStates->setEnabled(false);
    }
    if (!parmChangesAllowed)
    {
        pbOK->setEnabled(false);
    }

    m_LastScriptsRow = -1;  // To make sure we will populate parms listview

    // Set to the first script and first parameter.
    tblScripts->setCurrentIndex(scriptDataModel->index(0, 0));
    /* Just use the specified index for both parameters. */
    scriptChanged(scriptDataModel->index(0, 0),
                  scriptDataModel->index(0, 0));

    tblParms->setCurrentIndex(parameterDataModel->index(0, 2));
    /* Update the description object.  Just use the specified index for both
       parameters. */
    dispParmDesc(parameterDataModel->index(0, 2),
                 parameterDataModel->index(0, 2));

    // The values have changed in the tables.
    connect(parameterDataModel, SIGNAL(itemDataChanged(const QModelIndex&)),
            this, SLOT(updateParms(const QModelIndex&)));
    connect(scriptDataModel, SIGNAL(itemDataChanged(const QModelIndex&)),
            this, SLOT(checkScriptState(const QModelIndex&)));

    // Connect the interface buttons.
    connect(pbContinueWO, SIGNAL(clicked()), this, SLOT(continueWO()));
    connect(pbRestartWO, SIGNAL(clicked()), this, SLOT(restartWO()));
    connect(pbOK, SIGNAL(clicked()), this, SLOT(commitAndClose()));
    connect(pbCancel, SIGNAL(clicked()), this, SLOT(rollbackAndClose()));
}

//****************************************************************************
QString ModScriptWOParms::getCurrentWOState(QString id)
{
    // Reads and returns the work order state value

    IASSqlQuery query;

    QString sqlStmt = "SELECT state"
                      " FROM work_orders"
                      " WHERE work_order_id = '" + id + "'"; 
    if (query.exec(sqlStmt, __FILE__, __LINE__))
        query.next();

    return query.value("STATE").toString();
}

//****************************************************************************
void ModScriptWOParms::setupScriptsTable()
{
    // Creates the scripts table and sets all of the
    // initial values of the table.

    int rowNumber = 0;
    QStringList scriptStates;
    QFont f = this->font();
    IASSqlQuery query;

    QStringList labels;
    scriptDataModel = new IASParameterModel;
    IASParameterItem *sItem = NULL;
    IASParameterItemRow *sItemRow = NULL;

    labels << "Script Id" << "Script State" << "Exit Code";
    scriptDataModel->setColumnCount(labels.count());
    scriptDataModel->setHorizontalHeaderLabels(labels);

    // Set up the delegate for the table.
    tblScripts->setItemDelegate(new IASParameterDelegate);

    // Set up the edit triggers.
    tblScripts->setEditTriggers(QAbstractItemView::AllEditTriggers);

    // Set the rows to alternate colors.
    tblScripts->setAlternatingRowColors(true);

    // Make headings bold and set focus and selection.
    f.setBold(true);
    tblScripts->horizontalHeader()->setFont(f);

    // Only select one row at a time.
    tblScripts->setSelectionMode(QAbstractItemView::SingleSelection);
    tblScripts->setSelectionBehavior(QAbstractItemView::SelectRows);

    // Get rid of the vertical header.
    tblScripts->verticalHeader()->hide();

    // Add rows to the scripts table.  The state column will have
    // combo boxes to allow the user to choose PENDING or SKIP states
    //---------------------------------------------------------------
    QString sqlStmt = "SELECT script_id, state, compl_status"
              " FROM wo_scripts"
              " WHERE work_order_id = '" + lblWorkOrderIdVal->text() + "'"
              " ORDER BY seq_nbr";

    QVariant data;
    if (query.exec(sqlStmt, __FILE__, __LINE__))
    {
        while (query.next())
        {
            scriptStates.clear();
            scriptStates << UI_WO_SCRIPT_PENDING_STR
                         << UI_WO_SCRIPT_SKIP_STR;

            sItemRow = new IASParameterItemRow;

            sItem = new IASParameterItem;
            data = query.value(0).toString();
            sItem->setData(data);
            sItemRow->append(sItem);

            QString scriptState = query.value(1).toString();
            data = convToState(scriptState);
            if(!scriptStates.contains(data.toString()))
            {
                scriptStates << data.toString();
            }

            sItem = new IASParameterItem;
            sItem->setData(data);
            if (parmChangesAllowed)
            {
                sItem->setEditable(true);
            }
            else
            {
                sItem->setEditable(false);
            }
            sItem->setType(IASParameterItem::TextSelection);
            sItem->setOptions(scriptStates);
            sItemRow->append(sItem);

            sItem = new IASParameterItem;
            if(query.value(2).toString() == DB_WO_SCRIPT_STATUS_SUCCESS)
            {
                data = UI_WO_SCRIPT_STATUS_SUCCESS_STR;
            }
            else if(query.value(2).toString() == DB_WO_SCRIPT_STATUS_FAILURE)
            {
                data = UI_WO_SCRIPT_STATUS_FAILURE_STR;
            }
            else
            {
                data = "";
            }
            sItem->setData(data);
            sItemRow->append(sItem);

            scriptDataModel->appendRow(sItemRow);

            rowNumber++;
        }
    }

    // Set the model.
    tblScripts->setModel(scriptDataModel);

    // Set column widths based on widest record in each column.
    tblScripts->resizeColumnsToContents();

    // Set last column to be stretchable
    tblScripts->horizontalHeader()->setStretchLastSection(true);
}

//***********************************************************************
QString ModScriptWOParms::convToState(QString code)
{
    // Converts wo_scripts table script field codes

    if (code == DB_WO_SCRIPT_COMPLETE)
        return UI_WO_SCRIPT_COMPLETE_STR;
    if (code == DB_WO_SCRIPT_RUNNING)
        return UI_WO_SCRIPT_RUNNING_STR;
    if (code == DB_WO_SCRIPT_PENDING)
        return UI_WO_SCRIPT_PENDING_STR;
    if (code == DB_WO_SCRIPT_SKIP)
        return UI_WO_SCRIPT_SKIP_STR;
    else
        return "";
}

//***********************************************************************
QString ModScriptWOParms::convToCode(QString state)
{
    // Convert a script state name to it's code

    if (state == UI_WO_SCRIPT_COMPLETE_STR)
        return DB_WO_SCRIPT_COMPLETE;
    if (state == UI_WO_SCRIPT_RUNNING_STR)
        return DB_WO_SCRIPT_RUNNING;
    if (state == UI_WO_SCRIPT_PENDING_STR)
        return DB_WO_SCRIPT_PENDING;
    if (state == UI_WO_SCRIPT_SKIP_STR)
        return DB_WO_SCRIPT_SKIP;
    else
        return "";
}

//***********************************************************************
void ModScriptWOParms::setupParmsTable()
{
    QFont f = this->font();
    QStringList labels;

    parameterDataModel = new IASParameterModel;

    labels << "Parm Name" << "Element" << "Type" << "Parm Value"
           << "Required" << "Attribute";
    parameterDataModel->setColumnCount(labels.count());
    parameterDataModel->setHorizontalHeaderLabels(labels);

    // Set up the delegate for the table.
    tblParms->setItemDelegate(new IASParameterDelegate);

    // Set up the edit triggers.
    tblParms->setEditTriggers(QAbstractItemView::AllEditTriggers);

    // Set the rows to alternate colors.
    tblParms->setAlternatingRowColors(true);

    // Make headings bold and set focus and selection
    f.setBold(true);
    tblParms->horizontalHeader()->setFont(f);

    // Only select one row at a time.
    tblParms->setSelectionMode(QAbstractItemView::SingleSelection);
    tblParms->setSelectionBehavior(QAbstractItemView::SelectRows);

    // Get rid of the vertical header.
    tblParms->verticalHeader()->hide();

    // Set the model
    tblParms->setModel(parameterDataModel);

    // Adjust the columns to the new data.
    tblParms->resizeColumnsToContents();

    // Set last column to be stretchable
    tblParms->horizontalHeader()->setStretchLastSection(true);
}

//***********************************************************************
void ModScriptWOParms::scriptChanged(
    const QModelIndex& cIndex, /* I: Current model index */
    const QModelIndex&)        /* I: Previous model index - NOT USED */
{
    // There may have been a value which was being edited in the
    // parameters table when the script selection was changed so we
    // need to check.  Change the current cell to end the edit and emit
    // the value changed signal.

    int scriptsRow = cIndex.row();
    IASParameterItem *sItem = NULL;

    tblParms->setCurrentIndex(parameterDataModel->index(0, 2));

    // If the current scripts table selection has changed but the row
    // has not changed we do not need to repopulate the parameters table
    //------------------------------------------------------------------
    if(m_LastScriptsRow == scriptsRow)
    {
        return;
    }
    else
    {
        m_LastScriptsRow = scriptsRow;
    }

    // Since the value changed signal seems to be always emitted we
    // need to save the state value so we can check it later to see
    // if the value really changed.
    //-------------------------------------------------------------
    sItem = scriptDataModel->getItem(scriptsRow, 1);
    m_LastScriptState = sItem->data().toString();
    sItem = scriptDataModel->getItem(scriptsRow, 2);
    m_LastScriptExitcode = sItem->data().toString();

    // Populate the parameters table with values for the script
    //---------------------------------------------------------
    popParmTable(scriptsRow);
}

//***********************************************************************
void ModScriptWOParms::popParmTable(int scriptsRow)
{
    // Populates the script parameter table with parameter information
    // corresponding to the selected script in the scripts table.  If the
    // current scripts table selection has changed but the row has not
    // changed the parameters table does not need to be repopulated.

    IASSqlQuery query;
    IASParameterItem *pItem = NULL;
    IASParameterItemRow *pItemRow = NULL;
    IASParameterItem *sItem = NULL;

    // Delete any rows currently in the table
    parameterDataModel->deleteData();

    sItem = scriptDataModel->getItem(scriptsRow, 0);

    // Get script parameter information from the wo_parms table.  Since
    // the query size feature is not available as of now, we need
    // to get the number of rows returned by finding the index of the last
    // record in the result.
    //--------------------------------------------------------------------
    QString sqlStmt = "select parm_name, parm_occurrence, parm_value "
                      "from wo_parms where work_order_id = '" +
                      lblWorkOrderIdVal->text() + "' and script_id = '" +
                      sItem->data().toString() +
                      "' order by parm_name, parm_occurrence";

    QVariant data;
    if(query.exec(sqlStmt, __FILE__, __LINE__))
    {
        while(query.next())
        {
            pItemRow = new IASParameterItemRow;

            /* 0 - Parm Name */
            pItem = new IASParameterItem;
            data = query.value(0).toString();
            pItem->setData(data);
            pItemRow->append(pItem);

            /* 1 - Element */
            pItem = new IASParameterItem;
            data = query.value(1).toString();
            pItem->setData(data);
            pItemRow->append(pItem);

            /* 2 - Type */
            pItem = new IASParameterItem;
            data = "";
            pItem->setData(data);
            pItemRow->append(pItem);

            /* 3 - Parm Value */
            pItem = new IASParameterItem;
            data = query.value(2).toString();
            pItem->setData(data);
            pItemRow->append(pItem);

            /* 4 - Required */
            pItem = new IASParameterItem;
            data = "";
            pItem->setData(data);
            pItemRow->append(pItem);

            /* 5 - Attribute */
            pItem = new IASParameterItem;
            data = "";
            pItem->setData(data);
            pItemRow->append(pItem);

            parameterDataModel->appendRow(pItemRow);
        }
    }

    // Update the Type/Required and Parameter Value fields seperately to
    // minimize the number of queries.
    //------------------------------------------------------------------
    insertTypeRequired(scriptsRow);
    insertParmValueItem();

    // Clear the parameter description box of old values
    teParmDesc->clear();

    // Adjust the columns to the new data.
    tblParms->resizeColumnsToContents();

    // Set last column to be stretchable
    tblParms->horizontalHeader()->setStretchLastSection(true);
}

//***********************************************************************
void ModScriptWOParms::insertTypeRequired(int scriptsRow)
{
    // Inserts values into the parameter "type" and "required" fields
    // from the def_parms database table.  Number of queries is cut down
    // by eliminating duplicate queries of parameters with multiple
    // occurrences.

    IASSqlQuery query;
    IASParameterItem *pItem = NULL;
    IASParameterItem *sItem = NULL;

    sItem = scriptDataModel->getItem(scriptsRow, 0);
    QString scriptName = sItem->data().toString();

    QString parmName = ""; // Current and last parameter name.
    QString parmOccur = "";

    for(int row = 0; row < parameterDataModel->rowCount(); row++)
    {
        pItem = parameterDataModel->getItem(row, 0); // Parm Name
        parmName = pItem->data().toString();

        pItem = parameterDataModel->getItem(row, 1); // Parm Occurance
        parmOccur = pItem->data().toString();

        QString sqlStmt = "select parm_type, required from def_parms "
                        "where module_id = '" + lblProcedureIdVal->text()
                        + "' and sub_module_id = '" + scriptName
                        + "' and parm_name = '" + parmName
                        + "' and parm_occurrence = " + parmOccur
                        + " and processing_instance = '" + m_Instance + "' "
                        " and satellite = '" + m_Satellite + "' "
                        " and sensor = '" + m_Sensor + "' ";

        if(query.exec(sqlStmt, __FILE__, __LINE__))
        {
            query.next();
        }
        else
        {
            return;
        }

        // Insert the parameter type and required flag in the parameters table
        //--------------------------------------------------------------------
        if(query.isValid())
        {
            QVariant data;

            /* 2 - Type */
            data = query.value(0).toString();
            parameterDataModel->setItemData(row, 2, data);

            /* 4 - Required */
            if(query.value(1).toInt())
            {
                data = "TRUE";
            }
            else
            {
                data = "FALSE";
            }
            parameterDataModel->setItemData(row, 4, data);
        }
    }
}

//***********************************************************************
void ModScriptWOParms::insertParmValueItem()
{
    // Checks to see if there are a fixed set of parameter values
    // and whether the value read from the database should be mapped to
    // a different string to be shown to the user.  A combobox will be
    // inserted if there are a fixed set of values for the parameter.
    // Also, if the value includes an algorithm attribute, this is 
    // not displayed in the parm_value column, but rather is displayed
    // in a separate combobox.

    IASSqlQuery query;
    QStringList ouiValues;
    QStringList parmValueSegments; // algorithm name, and potentially
                                   // algorithm attribute as well
    IASSqlQuery queryAttributes;   // Query for parm_attributes table
    QString sqlStmtAttributes;     // SQL to query parm_attributes table
    QStringList attributes;        // list of all algorithm attributes 
    QString parmValue;             // parm value (sometimes an algorithm name)
    QString attribute;             // the attribute in the parmValue field 

    IASParameterItem *pItem = NULL;

    // Get the list of possible attribute values.
    sqlStmtAttributes = QString("select oui_value from ")
        + "parm_list where parm_name = 'ATTRIBUTE' "
        + "and processing_instance = '" + m_Instance + "' "
        + "and satellite = '" + m_Satellite + "' "
        + "and sensor = '" + m_Sensor + "' ";

    if(queryAttributes.exec(sqlStmtAttributes,__FILE__,__LINE__))
    {
        // get next parameter from database cursor
        while(queryAttributes.next())
        {
            attributes << queryAttributes.value(0).toString();
        }
    }
    else
    {
        return;
    }

    QString parmName = "";
    QString parmRequired = "";
    QVariant data;

    for(int row = 0; row < parameterDataModel->rowCount(); row++)
    {
        pItem = parameterDataModel->getItem(row, 4); /* 4 - Required */
        parmRequired = pItem->data().toString();
        pItem = parameterDataModel->getItem(row, 0); /* 0 - Parm Name */
        // Only query the parm_list table if the parameter name is different
        // from the previous row (to cut down on queries)
        //------------------------------------------------------------------
        if((row == 0) || (pItem->data().toString() != parmName))
        {
            parmName = pItem->data().toString();

            ouiValues.clear();

            QString sqlStmt = "Select oui_value, program_value from "
                 "parm_list where "
                 "parm_name = '" +  parmName + "' "
                 "and processing_instance = '" + m_Instance + "' "
                 "and satellite = '" + m_Satellite + "' "
                 "and sensor = '" + m_Sensor + "' ";


            // Create a stringlist with the possible values for the parameter
            //---------------------------------------------------------------
            if(query.exec(sqlStmt, __FILE__, __LINE__))
            {
                while(query.next())
                {
                    ouiValues += query.value(0).toString();
                }
            }
            else
            {
                return;
            }
        }

        // Initialize the parmValue value. 
        pItem = parameterDataModel->getItem(row, 3);
        parmValue = pItem->data().toString();

        attribute = ""; // Force to empty string.

        // If the Parm name is ALGORITHMS, get a list of parm attributes
        if(parmName == "ALGORITHMS")
        {
            // Find if an attribute is at the end of the parmValue
            parmValueSegments = parmValue.split(QRegExp("_"));

            // This last element is an attribute if it's in the attribute list
            if(attributes.contains(parmValueSegments.last()))
            {
                // The last element is an attribute
                attribute = parmValueSegments.last();

                // Remove the last element (the attribute) from the list
                parmValueSegments.pop_back();

                // Form the algorithm name from the rest of the parmValue 
                // segments
                parmValue = parmValueSegments.join ("_");
            }
            else
            {
                // The last element is not an attribute
                attribute = "RUN";
            }
        }

        // Create a combo box in the cell if there is a set of values for the
        // parameter. Otherwise use the line edit that is already there.  The
        // present value of the parameter is already in the parm value column
        //-------------------------------------------------------------------
        pItem = parameterDataModel->getItem(row, 3); /* 3 - Parm Value */
        data = parmValue;
        pItem->setData(data);

        if(!ouiValues.isEmpty())
        {
            // Discriminate against the ALGORITHMS.
            // Don't allow any modifications to them.
            if(parmName == "ALGORITHMS")
            {
                pItem->setType(IASParameterItem::Text);
            }
            else
            {
                if(parmRequired == "FALSE")
                {
                    if(!ouiValues.contains("")) // add it if not present
                    {
                        ouiValues.prepend(""); // add an empty selection
                    }
                }
                else
                {
                    if(ouiValues.contains("")) // if present remove it
                    {
                        ouiValues.removeAll("");
                    }
                }

                /* Sort the options with a special routine if they are for
                   the BAND_LIST */
                if(parmName == "BAND_LIST")
                {
                    qSort(ouiValues.begin(), ouiValues.end(),
                          IASUtilities::numericLessThan);
                }

                if (parmChangesAllowed)
                {
                    pItem->setEditable(true);
                }
                else
                {
                    pItem->setEditable(false);
                }
                pItem->setType(IASParameterItem::TextSelection);
                pItem->setOptions(ouiValues);
            }
        }
        else
        {
            if (parmChangesAllowed)
            {
                pItem->setEditable(true);
            }
            else
            {
                pItem->setEditable(false);
            }
        }

        pItem = parameterDataModel->getItem(row, 5); /* 5 - Attribute */
        data = attribute;
        pItem->setData(data);

        // Also do this for the attribute column if appropriate
        if(parmName == "ALGORITHMS")
        {
            if (parmChangesAllowed)
            {
                pItem->setEditable(true);
            }
            else
            {
                pItem->setEditable(false);
            }
            pItem->setType(IASParameterItem::TextSelection);
            pItem->setOptions(attributes);
        }
    }
}

//***********************************************************************
void ModScriptWOParms::dispParmDesc(
    const QModelIndex& cIndex, /* I: Current model index */
    const QModelIndex&)        /* I: Previous model index - NOT USED */
{
    // Displays the description of the parameter in the row from the
    // parm_edits table and updates the current parameter value.

    IASSqlQuery query;
    IASParameterItem *pItem = NULL;
    int row = cIndex.row();

    if(!cIndex.isValid())
    {
        return;
    }

    pItem = parameterDataModel->getItem(row, 3);

    m_CurrentParmVal = pItem->data().toString();

    pItem = parameterDataModel->getItem(row, 0);

    QString sqlStmt = "select parm_desc from "
         "parm_edits where parm_name='"
        + pItem->data().toString() + "' "
         "and processing_instance = '" + m_Instance + "' "
         "and satellite = '" + m_Satellite + "' "
         "and sensor = '" + m_Sensor + "' ";


    if (query.exec(sqlStmt, __FILE__, __LINE__))
    {
        query.next();
    }
    else
    {
        return;
    }

    QString parmDesc("");
    if (query.isValid())
    {
        parmDesc = query.value(0).toString();
        teParmDesc->setText(parmDesc);
    }
    else
    {
        teParmDesc->setText("");
    }
}

//***********************************************************************
void ModScriptWOParms::updateParms(const QModelIndex& pIndex)
{
    // This function checks the parameters updated value against the max
    // and min (if there is one) and updates the wo_parms table with the
    // new value.  We need to verify if the fields value was really updated
    // since the valueUpdated signal is emitted whether the value
    // was actually changed or not.

    if(!pIndex.isValid())
    {
        return;
    }

    IASParameterItem *pItem = NULL;
    int row = pIndex.row();
    int column = pIndex.column();

    pItem = parameterDataModel->getItem(row, column);

    if((pItem->data().toString().length() == 0) && // check for null vs empty
       (m_CurrentParmVal.length() == 0)) 
    {
        return;
    }
    else
    {
        if(pItem->data().toString() == m_CurrentParmVal) // no change
            return;
    }

    // Validate the parameter unless it is from a combobox.  Values in a
    // combobox are all assumed to be valid since they were read from the
    // database and there is only a discrete set of values to choose from.
    //--------------------------------------------------------------------
    if(pItem->type() != IASParameterItem::TextSelection) // Not a combobox value
    {
        if(!validateParm(row)) // Out of range, reset value to original
        {
            QVariant data = m_CurrentParmVal;

            // make sure the correct item is accessed
            pItem = parameterDataModel->getItem(row, 3); /* 3 - Parm Value */
            pItem->setData(data);

            tblParms->setCurrentIndex(parameterDataModel->index(row, column));

            return;
        }
    }

    // If the database update is successful reset the current value
    // to the newly updated value
    //-------------------------------------------------------------
    if(updateParmsTable(row, m_LastScriptsRow))
    {
        m_CurrentParmVal = pItem->data().toString();
    }
}

//***********************************************************************
void ModScriptWOParms::checkScriptState(const QModelIndex& sIndex)
{
    // Check to see whether the state of the script changed and call the
    // update function if so.  We need to do this because the valuechanged
    // signal is emitted whether the value is changed or not.

    if(!sIndex.isValid())
    {
        return;
    }

    int row = sIndex.row();
    IASParameterItem *sItem = NULL;

    sItem = scriptDataModel->getItem(row, 1);
    QString state = sItem->data().toString();

    if(state != m_LastScriptState)
    {
        updateScriptRow(row, state);
    }
}

//***********************************************************************
bool ModScriptWOParms::updateScriptRow(int row,
                                       QString state)
{
    // This function updates the script state in the database when it is
    // changed in the scripts table.

    IASSqlQuery query;
    IASParameterItem *sItem = NULL;

    if(!allowedToUpdate())
        return false;

    QVariant data = state;
    scriptDataModel->setItemData(row, 1, data);

    data = "";
    scriptDataModel->setItemData(row, 2, data);

    sItem = scriptDataModel->getItem(row, 0);
    QString scriptId = sItem->data().toString();

    QString sqlStmt = "UPDATE wo_scripts"
                      " SET state = '" + convToCode(state) + "',"
                          " compl_status = NULL"
                      " WHERE work_order_id = '"
                              + lblWorkOrderIdVal->text() + "'"
                          " AND script_id = '" + scriptId + "'";

    if(query.exec(sqlStmt, __FILE__, __LINE__))
    {
        query.next();
    }
    else
    {
        /* reset state */
        data = m_LastScriptState;
        scriptDataModel->setItemData(row, 1, data);
        /* reset exit code */
        data = m_LastScriptExitcode;
        scriptDataModel->setItemData(row, 2, data);

        return false;
    }

    /* The database update was successful, reset the last state and completion
       status to the new values. */
    sItem = scriptDataModel->getItem(row, 1);
    m_LastScriptState = sItem->data().toString();
    m_LastScriptExitcode = "";

    return true;
}

//***********************************************************************
bool ModScriptWOParms::updateParmsTable(int row, int scriptsRow)
{
    // Updates the value of a parameter in the wo_parms table.

    IASSqlQuery query;
    IASParameterItem *pItem = NULL;
    IASParameterItem *sItem = NULL;

    if(!allowedToUpdate())
        return false;

    sItem = scriptDataModel->getItem(scriptsRow, 0);
    QString scriptId = sItem->data().toString();

    pItem = parameterDataModel->getItem(row, 0); /* 0 - Parm Name */
    QString parmName = pItem->data().toString();
    pItem = parameterDataModel->getItem(row, 1); /* 1 - Parm Occurance */
    QString parmOccur = pItem->data().toString();
    pItem = parameterDataModel->getItem(row, 2); /* 2 - Type */
    QString parmType = pItem->data().toString();
    pItem = parameterDataModel->getItem(row, 3); /* 3 - Parm Value */
    QString parmValue = pItem->data().toString();
    pItem = parameterDataModel->getItem(row, 5); /* 5 - Attribute */
    QString attribute = pItem->data().toString();

    QString sqlStmt = "update wo_parms set parm_value=";
    if(parmValue.length() == 0)
    {
        sqlStmt += "NULL";
    }
    else if((parmType.toUpper() != "INT") && (parmType.toUpper() != "LONG"))
    {
        sqlStmt += "'" + parmValue;

        // include the algorithm attribute if present
        if((attribute != "RUN") && (attribute != ""))
        {
            sqlStmt += "_" + attribute;
        }

        sqlStmt += "'";
    }
    else
    {
        sqlStmt += parmValue;
    }

    sqlStmt += " where work_order_id='" + lblWorkOrderIdVal->text() + "'";
    sqlStmt += " and script_id='" + scriptId + "'";
    sqlStmt += " and parm_name='" + parmName + "'";
    sqlStmt += " and parm_occurrence=" + parmOccur;

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        return false;
    }

    return true;
}

//***********************************************************************
bool ModScriptWOParms::validateParm(int row)
{
    // Checks whether the string can be converted to a valid number
    // and checks the value of the parameter against the max/min in the
    // parm_edits table (if there is one).

    bool ok;
    IASParameterItem *pItem = NULL;

    pItem = parameterDataModel->getItem(row, 0); /* 0 - Parm Name */
    QString parmName = pItem->data().toString();
    pItem = parameterDataModel->getItem(row, 2); /* 2 - Type */
    QString parmType = pItem->data().toString().toUpper();
    pItem = parameterDataModel->getItem(row, 3); /* 3 - Parm Value */
    QString parmValue = pItem->data().toString();
    pItem = parameterDataModel->getItem(row, 4); /* 4 - Required */
    QString parmRequired = pItem->data().toString();

    // allow parameters which are not required to be empty strings.
    if(parmValue == "" && parmRequired == "FALSE")
    {
        return true;
    }

    if(parmValue == "")
    {
        QMessageBox::warning(this, "Error", 
            "The Parameter selected is required",
            QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);

        return false;
    }

    // make sure the value is valid and return false if not
    if((parmType == "INT") || (parmType == "LONG"))
    {
        parmValue.toLong(&ok);
        if(!ok)
        {
            QMessageBox::warning(this, "Error", 
                "The Parameter value is not a valid integer number",
                QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);

            return false;
        }
    }
    else if((parmType == "FLOAT") || (parmType == "DOUBLE"))
    {
        parmValue.toDouble(&ok); 
        if(!ok)
        {
            QMessageBox::warning(this, "Error", 
                "The Parameter value is not a valid floating point number",
                QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);

            return false;
        }
    }
    else 
    {
        return true;
    }

    // If there is a max/min associated with the parameter check if it is
    // in range.  Return false if not so the out of range value will be reset.
    IASSqlQuery query;
    QString sqlStmt = "SELECT min_value, max_value"
                      " FROM parm_edits"
                      " WHERE parm_name = '" + parmName + "' "
                      " and processing_instance = '" + m_Instance + "' "
                      " and satellite = '" + m_Satellite + "' "
                      " and sensor = '" + m_Sensor + "' ";


    if(query.exec(sqlStmt, __FILE__, __LINE__))
    {
        query.next();
    }

    if(!query.isNull(0)) // There is a max/min
    {
        bool validVal = true;

        if((parmType == "INT") || (parmType == "LONG"))
        {
            if((parmValue.toInt(&ok) < query.value(0).toInt(&ok)) ||
               (parmValue.toInt(&ok) > query.value(1).toInt(&ok)))
            {
                validVal = false;
            }
        }
        else if((parmType == "FLOAT") || (parmType == "DOUBLE"))
        {
            if((parmValue.toDouble(&ok) < query.value(0).toDouble(&ok)) ||
               (parmValue.toDouble(&ok) > query.value(1).toDouble(&ok)))
            {
                validVal = false;
            }
        }

        // If the value is not valid report the min/max and type
        // information for the parameter and return false to prevent
        // update with this value
        //-----------------------------------------------------------
        if(!validVal)
        {
            QMessageBox::warning(this, "Warning",
                "The Parameter value is invalid, should be:\nType = " +
                parmType + "\n" + 
                "Min = " + query.value("MIN_VALUE").toString() + "\n" +
                "Max = " + query.value("MAX_VALUE").toString(),
                QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);

            return false;
        }
    }

    return true;
}

/*****************************************************************************
    Process through the scripts for the work order and change any failed
    scripts back to pending.  Then change the work order state to pending.
 *****************************************************************************/
void ModScriptWOParms::continueWO()
{
    IASParameterItem *sItem = NULL;

    // Set failed scripts to pending
    for (int row = 0; row < scriptDataModel->rowCount(); row++)
    {
        sItem = scriptDataModel->getItem(row, 2);
        QString status = sItem->data().toString();
        if (status == UI_WO_SCRIPT_STATUS_FAILURE_STR)
        {
            if (!updateScriptRow(row, UI_WO_SCRIPT_PENDING_STR))
            {
                return;
            }
        }
    }

    // Set the work order state to pending here
    if (setWoStateToPending())
    {
        QMessageBox::information(this, "Modifying States",
            "Work order state and failed scripts have been set to "
                + UI_WO_SCRIPT_PENDING_STR,
            QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
    }
}

/*****************************************************************************
    Process through the scripts for the work order and change all the scripts
    to pending.  Then change the work order state to pending.
 *****************************************************************************/
void ModScriptWOParms::restartWO()
{
    int rowCount = scriptDataModel->rowCount();
    for (int row = 0; row < rowCount; row++)
    {
        if (!updateScriptRow(row, UI_WO_SCRIPT_PENDING_STR))
        {
            return;
        }
    }

    // Set the work order state to pending here
    if (setWoStateToPending())
    {
        QMessageBox::information(this, "Modifying States",
            "Work order state and all scripts have been set to "
                + UI_WO_SCRIPT_PENDING_STR,
            QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
    }
}

/*****************************************************************************
    Update the work order state to pending in the database.
 *****************************************************************************/
bool ModScriptWOParms::setWoStateToPending()
{
    IASSqlQuery query;

    /* Set the work order state */
    QString sqlStmt = "UPDATE work_orders"
                      " SET state = '" + DB_WO_PENDING + "'"
                      " WHERE work_order_id = '"
                          + lblWorkOrderIdVal->text() + "'";

    if (query.exec(sqlStmt, __FILE__, __LINE__))
    {
        /* Reset the initialWoState because we just changed it to pending. */
        m_initialWoState = DB_WO_PENDING;
        return true;
    }

    QMessageBox::warning(this, "Warning",
        "Failed to set work order state to " + UI_WO_PENDING_STR,
        QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);

    return false;
}

//***********************************************************************
void ModScriptWOParms::commitAndClose()
{
    // Commit all changes for this transaction and close dialogue.
    // Since the valueChanged signal is not emitted until another cell is
    // selected, select the previous column in the row to emit this signal.
    // This will check for an update of the current cell before committing
    // all changes.

    if(!allowedToUpdate())
        return;

    if(!QSqlDatabase::database().commit())
    {
        QMessageBox::warning(this, "Warning", 
            "Commit of database changes failed",
            QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
    }

    close();
}

//***********************************************************************
void ModScriptWOParms::rollbackAndClose()
{
    // Rollback all changes for this transaction and close dialogue

    if(!QSqlDatabase::database().rollback())
    {
        QMessageBox::warning(this, "Warning", 
            "Rollback of database changes failed",
            QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
    }

    close();
}

//***********************************************************************
bool ModScriptWOParms::allowedToUpdate()
{
    /* Do some checking to see if we should allow updates or not due to
       state changes.  This is meant to catch changes from Pending to
       Running. */

    /* Check to see if the work order state changed first. */
    QString currentWoState = getCurrentWOState(lblWorkOrderIdVal->text());
    if (currentWoState != m_initialWoState)
    {
        QMessageBox::warning(this, "Warning", 
            "Work Order State has changed.  Unable to perform updates.",
            QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
        return false;
    }

    return true;
}

