/******************************************************************************
               Image Assessment System (IAS) Software
 ******************************************************************************
 *
 * Name: ModifyWO.cpp
 *
 * PURPOSE: This is the definition of the ModifyWO class which is used to
 *          Modify a work order.
 *
 *****************************************************************************/

// System Includes
#include <iostream>

// QT Includes
#include <QtCore>
#include <QtGui>
#include <QtSql>

// IAS Qt Includes 
#include <IASQtConst.h>
#include <iasQtWoConst.h>

// Local Includes
#include "ModifyWO.h"
#include "IASSqlQuery.h"
#include "ModScriptWOParms.h"
#include "ModGlobalWOParms.h"
#include "IASSchemaConfig.h"
#include "IASKeyValueComboBox.h"
#include "IASParameterDelegate.h"
#include "IASParameterModel.h"

/* NOTE - leSceneId is used to hold the interval id if the work order happens
          to be one that processes an interval. */
/*
    ModifyWO

    Initializes an instance of the modify work order interface.  It inherits
    a bunch of functionality from the CreateWO parent class.
*/
ModifyWO::ModifyWO(const QString& woId,
                   QWidget *parent, 
                   Qt::WindowFlags fl)
: CreateWO(parent, fl)
{
    m_modifyWoId = woId;
}

/*
    setTitle

    This is used to set the titles correctly for the specific interface.
*/
void ModifyWO::setTitle()
{
    QString title = "Modify Work Order";

    lblCreateWO->setText(title);
    setWindowTitle(title);
}

/*
    populateInitialDb

    This method overrides the parent class method to populate the interface
    correctly based on information retrieved from the database.
*/
bool ModifyWO::populateInitialDb()
{
    QString date;
    QString sqlStmt;
    IASSqlQuery query;

    // Read the work_orders table
    sqlStmt  = "SELECT requester, priority, ephem_type, request_type,";
    sqlStmt += "to_char(req_start_date,'DY MON DD HH24:MI:SS YYYY'), ";
    sqlStmt += "scene_id, procedure_id, cpf_override, wo_path, wo_comment, ";
    sqlStmt += "output_db_schema, l0r_data_path, state, l0r_data_status, ";
    sqlStmt += "processing_instance ";
    sqlStmt += "FROM work_orders ";
    sqlStmt += "WHERE work_order_id = '" + m_modifyWoId + "'";

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        return false;
    }
    if(!query.next())
    {
        return false;
    }

    // Assign the values to the form
    lblRequester->setText(query.value(0).toString());

    /* Populated the satellite sensor combobox base on the selected
       work order id. */
    getSatSensor();

    // Priority
    if(query.value(1).toString() != "")
    {
        setupPriority(query.value(1).toInt());
    }

    // Ephemeris Type
    if(query.value(2).toString() != "")
    {
        setupDefEphPCD(query.value(2).toString());
    }

    // Request Type
    if(query.value(3).toString() != "")
    {
        setupRequestType(query.value(3).toString());
    }

    // Output Database Schema
    if(query.value(10).toString() != "")
    {
        setupOutputDbSchema(query.value(10).toString());
    }

    // Scene Id
    if(query.value(5).toString() != "")
    {
        leSceneId->setText(query.value(5).toString());
        leSceneId->setReadOnly(true);
    }

    // The procedure Id can not be changed in modify
    if(query.value(6).toString() != "") // Procedure Id
    {
        QString procId = query.value(6).toString();
        cboProcedureId->addItem(procId);
        cboProcedureId->setCurrentIndex(cboProcedureId->findText(procId));
        cboProcedureId->setEditable(false);
        cboProcedureId->setEnabled(false);

        // Assign the script ID and pause to the form
        populateScriptId(m_modifyWoId);
    }

    leCalParmFile->setText(query.value(7).toString());
    leWorkOrderPath->setText(query.value(8).toString());
    leWorkOrderCmnt->setText(query.value(9).toString());

    /* These "cw" variables are from CreateWO.  They allow for not changing
       both the l0r_data_path and l0r_data_status values while in the
       ModifyWO interface. */
    cw_l0rDataPath = query.value(11).toString();
    leL0RDataPath->setText(cw_l0rDataPath);
    cw_l0rDataStatus = query.value(13).toString();

    // Set the dates from the database
    if(query.value(4).toString() != "")
    {
        date = changeDateFormat(query.value(4).toString()); // Req_start_date
        QDateTime theDate = QDateTime::fromString(date,Qt::TextDate);
        dteRequestStartDate->setMinimumDate(theDate.date());
        dteRequestStartDate->setDateTime(theDate);
    }

    m_state = query.value(12).toString(); // work order state

    // The processing instance can not be changed in modify
    if (query.value(14).toString() != "") // Processing Instance
    {
        QString instance = query.value(14).toString();
        cboInstance->addItem(instance);
        cboInstance->setCurrentIndex(cboInstance->findText(instance));
        cboInstance->setEditable(false);
        cboInstance->setEnabled(false);
    }

    if(m_state == DB_WO_PENDING)
    {
        sqlStmt = "UPDATE work_orders"
                  " SET state = '" + DB_WO_HALTED + "',"
                      " date_updated = SYSDATE"
                  " WHERE work_order_id = '" + m_modifyWoId + "'"
                      " AND state = '" + DB_WO_PENDING + "' ";

        if(!query.exec(sqlStmt, __FILE__, __LINE__))
        {
            return false;
        }
        if(query.numRowsAffected() != 1)
        {
            QMessageBox::warning(this, "Error",
                "Failed to update work order state.",
                QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
            return false;
        }
    }

    allowReset = true;
    allowParmChanges = true;

    if(m_state == DB_WO_RUNNING)
    {
        /* This is a hack to prevent modifying running work orders outside of
           changing any scripts to pause. */
        gbPriority->setEnabled(false);
        gbEphemerisType->setEnabled(false);
        gbOrderType->setEnabled(false);
        gbOutputDbSchema->setEnabled(false);
        dteRequestStartDate->setEnabled(false);
        leWorkOrderPath->setEnabled(false);
        tbWorkOrderPath->setEnabled(false);
        leWorkOrderCmnt->setEnabled(false);
        pbCommon->setEnabled(false);
        pbScript->setEnabled(false);

        allowReset = false;
        allowParmChanges = false;
    }

    /* Always disable the L0R Data Path */
    leL0RDataPath->setReadOnly(true);
    tbL0RDataPath->setEnabled(false);

    return true;
}

/*
    getWoId

    This method is used to retrieve the work order Id saved in a previous
    method.  This overrrides the parent class method.
*/
QString ModifyWO::getWoId()
{
    return m_modifyWoId;
}

/*
    getSatSensor

    This sets up tje combobox which holds the information about which
    satellite and sensors are available to chose from.  It only populates one
    entry into the combobox and disables so that it can't be modified.

    There is an identical copy of this in TemplateWO.
*/
void ModifyWO::getSatSensor()
{
    QString satSen;    // Satellite and Sensor string
    QString schema;    // schema with work order Id
    QString schemaCmp; // schema to compare to work order Id's schema

    /* Find the schema and L0R Id associated with the Work Order Id for
       the work order */
    IASSchemaConfig * inst = IASSchemaConfig::instance();
    schema = inst->schemaNameFromWOId(m_modifyWoId);

    QStringList schemas = IASSchemaConfig::instance()->allKeys();
    QStringList::iterator itr;
    for(itr = schemas.begin(); itr != schemas.end(); ++itr)
    {
        satSen = (*itr);
        schemaCmp = IASSchemaConfig::instance()->schemaNameFromKey(satSen);
        if(schema == schemaCmp)
        {
            satSen.replace("_", " ");
            kvcboSatSensor->addKeyValue(satSen, schema);

            // Now that the Sat/Sensor is found, set it in stone.
            kvcboSatSensor->setCurrentIndex(kvcboSatSensor->findText(satSen));

            kvcboSatSensor->setEditable(false);
            kvcboSatSensor->setEnabled(false);
            break;
        }
    }
}

/*
    setupPriority

    Set the interface priority for the work order based on the provided value.
*/
void ModifyWO::setupPriority(const int& priority)
{
    int findIndex = cboPriority->findText(QString::number(priority),
        Qt::MatchExactly);
    if (findIndex > -1)
        cboPriority->setCurrentIndex(findIndex);
    else
    {
        /* This shouldn't happen, but just in case, we override the
           invalid value with the default priority. */
        selectDefaultPriority();
    }
}

/*
    setupDefEphPCD

    Set the interface ephemeris selection based on the provided value.

    There is an identical copy of this in TemplateWO.
*/
void ModifyWO::setupDefEphPCD(const QString& ephemType)
{
    // Ephemeris Type
    if(kvcboSatSensor->currentText() == "L8 OLITIRS")
    {
        rbPCD->setChecked(true); // Force PCD
        gbEphemerisType->setEnabled(false);
        gbEphemerisType->hide();
    }
    else if(ephemType == "P")
    {
        rbPCD->setChecked(true); // Select PCD
    }
    else
    {
        rbDefinitive->setChecked(true); // Select Definitive
    }
}

/*
    setupOutputDbSchema

    Set the interface output database schema selection based on the
    provided value.
*/
void ModifyWO::setupOutputDbSchema(const QString& schema)
{
    if(schema == "P")
    {
        rbProduction->setChecked(true); // Production Schema
    }
    else
    {
        rbEvaluation->setChecked(true); // Evaluation Schema
    }
}

/*
    setupRequestType

    Set the interface request type selection based on the provided value.
*/
void ModifyWO::setupRequestType(const QString& requestType)
{
    if(requestType == "S")
    {
        rbStandard->setChecked(true); // Standard
    }
    else if(requestType == "T") 
    {
        rbTest->setChecked(true); // Test
    }
    else
    {
        rbCustom->setChecked(true); // Custom
    }
}

/*
    getScriptSql

    This method configures the SQL required to retrieve the script Id and
    pause flag information from the database.

    There is an identical copy of this in TemplateWO.
*/
QString ModifyWO::getScriptSql(const QString& id)
{
    QString sqlStmt = "";

    sqlStmt  = "SELECT script_id, pause_flag ";
    sqlStmt += "FROM wo_scripts ";
    sqlStmt += "WHERE work_order_id = '" + id + "' ";
    sqlStmt += "ORDER BY seq_nbr";

    return sqlStmt;
}

/*
    cancelWorkOrder

    This calls close to terminate the interface.

    Since this is a new work order, that has not been made official it can be
    deleted from the database.  An official work order is one that has been
    accepted(you selected OK on the CreateWO or TemplateWO interface).

    This method overrides the parent class method.
*/
void ModifyWO::cancelWorkOrder()
{
    // Close the window.
    close();
}

/*
    closeEvent

    This checks the state the work order was in before the ModifyWO interface
    was started and attempts to reset the pending ones back to pending.

    This method is kicked off either from calling close() or through the "X"
    close button in the title bar.

    This method overrides the parent class method.
*/
void ModifyWO::closeEvent(QCloseEvent* ce)
{
    if(m_state == DB_WO_PENDING)
    {
        /* We changed it to DB_WO_HALTED previously so change it back to
           DB_WO_PENDING now. */
        IASSqlQuery query;
        QString sqlStmt  = "UPDATE work_orders"
                           " SET state = '" + DB_WO_PENDING + "'"
                           " WHERE work_order_id = '" + m_modifyWoId + "'"
                               " AND state = '" + DB_WO_HALTED + "' ";

        /* We don't care if it fails, we will just leave it in a HALTED
           state.  It can be set back to PENDING by another interface.
           But we will provide and error message back to the user. */
        QString errorMsg = "Work order '" + m_modifyWoId +
                "' not set back to pending.\n"
                "Leaving it in the '" + UI_WO_HALTED_STR + "' state.\n"
                "Use 'Script Status' to manually set the work order back to "
                "'" + UI_WO_PENDING_STR + "'.";
        if(!query.exec(sqlStmt, __FILE__, __LINE__))
        {
            QMessageBox::warning(this, "Error", errorMsg,
                QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
        }
        if(query.numRowsAffected() != 1)
        {
            QMessageBox::warning(this, "Error", errorMsg,
                QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
        }
    }

    ce->accept();
}

/*
    populateNewWorkOrder

    We intentionally do nothing here because we already have a work order
    inserted into the database.

    This implementation is needed to override the parent class method.
*/
void ModifyWO::populateNewWorkOrder()
{
    // This intentionally does nothing
}

