/******************************************************************************
               Image Assessment System (IAS) Software
*******************************************************************************
*
* Name:   OUIMainWindow.cpp
*
* PURPOSE:
* This is the main window of the IAS GUI.  It contains the pull-down menu
* structure and tool bar.  All functions of the GUI are accessed by the
* user from here.
*
******************************************************************************/

/* System Includes */
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>

/* Qt Includes */
#include <QtCore>
#include <QtGui>
 
/* Local Includes */
#include "OUIMainWindow.h"
#include "ConfigureIAS.h"
#include "StartStopIas.h"
#include "RestartIas.h"
#include "BrowseWOParms.h"
#include "WOProcessing.h"
#include "Events.h"
#include "WOStatus.h"
#include "CreateWO.h"
#include "CreateNewProcedure.h"
#include "CreateNewScript.h"
#include "IASProcessExecution.h"
#include "SelectModifyWO.h"
#include "SelectTemplateWO.h"
#include "IASConfigureInstances.h"

const int BLACK_MAGIC = 14; // This is the magic number adjustment to allow
                            // for the resizing of the OUI menu window to be
                            // large enough to display all of the top-level
                            // menu items in the current font.

OUIMainWindow::OUIMainWindow(QWidget *parent,
                             Qt::WindowFlags fl)
: QMainWindow(parent, fl)
{
    // Set up the user interface.
    setupUi(this);

    woStatus = NULL;

    // Main window will function as a menu to launch windows
    statusBar()->hide();
    setFixedHeight(menuBar()->height());
    layout()->setSizeConstraint(QLayout::SetNoConstraint);
    resize(width(), minimumHeight());
    signalMapper = new QSignalMapper(this);

    menuFont = mbOUI->font();
    menuFont.setPointSize(12);

    // Connect all the signals and slots.
    connect(actionConfiguration, SIGNAL(activated()),
            this, SLOT(activatedConfiguration()));
    connect(actionInstanceConfiguration, SIGNAL(activated()),
            this, SLOT(activatedInstanceConfiguration()));
    connect(actionStartStop, SIGNAL(activated()),
            this, SLOT(activatedStartStop()));
    connect(actionRestart, SIGNAL(activated()),
            this, SLOT(activatedRestart()));
    connect(actionExit, SIGNAL(activated()), this, SLOT(close()));
    connect(actionBrowseParms, SIGNAL(activated()),
            this, SLOT(activatedBrowseParms()));
    connect(actionModifyWO, SIGNAL(activated()),
            this, SLOT(activatedModifyWO()));
    connect(actionModifyWOState, SIGNAL(activated()),
            this, SLOT(activatedModifyWOState()));
    connect(actionEvents, SIGNAL(activated()),
            this, SLOT(activatedEvents()));
    connect(actionWOStatus, SIGNAL(activated()),
            this, SLOT(activatedWOStatus()));
    connect(actionNewWO, SIGNAL(activated()),
            this, SLOT(activatedNewWO()));
    connect(actionFromTemplate, SIGNAL(activated()),
            this, SLOT(activatedFromTemplate()));
    connect(actionNewScript, SIGNAL(activated()),
            this, SLOT(activatedNewScript()));
    connect(actionNewProcedure, SIGNAL(activated()),
            this, SLOT(activatedNewProcedure()));
    connect(signalMapper, SIGNAL(mapped(const QString &)),
            this, SLOT(executePlugin(const QString &)));

    initialize();
}

void OUIMainWindow::activatedWOStatus()
{
    if (woStatus == NULL)
    {
        woStatus = new WOStatus(0);
        connect(woStatus, SIGNAL(destroyed()),
                this, SLOT(woStatusDestroyed()));
    }

    woStatus->showNormal();
    woStatus->raise();
}

void OUIMainWindow::woStatusDestroyed()
{
    woStatus = NULL;
}

void OUIMainWindow::activatedNewWO()
{
    CreateWO *createWO = new CreateWO(0);

    // Only show the window if the initialization was successful
    if (createWO->initialize())
    {
        createWO->show();
    }
}

void OUIMainWindow::activatedBrowseParms()
{
    BrowseWOParms *browseWOParms = new BrowseWOParms(0);
    browseWOParms->show();
}

void OUIMainWindow::activatedConfiguration()
{
    ConfigureIAS *configureIAS = new ConfigureIAS(0);
    configureIAS->show();
}

void OUIMainWindow::activatedInstanceConfiguration()
{
    IASConfigureInstances *configureInstances = new IASConfigureInstances(0);
    configureInstances->show();
}

void OUIMainWindow::activatedEvents()
{
    Events *events = new Events(0);
    events->show();
}

void OUIMainWindow::activatedRestart()
{
    RestartIas *restartIas = new RestartIas(0);
    restartIas->show();
}

void OUIMainWindow::activatedStartStop()
{
    StartStopIas *startStopIas = new StartStopIas(0);
    startStopIas->show();
}

void OUIMainWindow::activatedModifyWO()
{
    SelectModifyWO *selectModifyWO = new SelectModifyWO(0);
    selectModifyWO->show();
}

void OUIMainWindow::activatedModifyWOState()
{
    WOProcessing *woProcessing = new WOProcessing(0);
    woProcessing->show();
}

void OUIMainWindow::activatedFromTemplate()
{
    SelectTemplateWO *selectTemplateWO = new SelectTemplateWO(0);
    selectTemplateWO->show();
}

void OUIMainWindow::activatedNewScript()
{
    CreateNewScript *createNewScript = new CreateNewScript(0);
    createNewScript->show();
}

void OUIMainWindow::activatedNewProcedure()
{
    CreateNewProcedure *createNewProcedure = new CreateNewProcedure(0);
    createNewProcedure->show();
}

void OUIMainWindow::initialize()
{
    QString menuText;
    QString errorMessage;
    int count = 0;

    /* Disable the start and stop actions because they are reserved for a
       specified user */
    actionStartStop->setEnabled(false);
    actionRestart->setEnabled(false);

    /* Get the system user ID from the environment */
    QString systemUserId = getenv("IAS_SYSTEM_USER_ID");
    if (systemUserId.isEmpty())
    {
        errorMessage = "Failed to get IAS_SYSTEM_USER_ID from environment.\n";
        QMessageBox::critical(this, "System Status - Error", errorMessage);
        return;
    }

    /* Get the OUI executing user ID from the environment */
    struct passwd *pws;
    pws = getpwuid(getuid());
    if (pws == NULL)
    {
        errorMessage = "Failed to retrieve user information from the system.";
        QMessageBox::critical(this, "System Status - Error", errorMessage);
        return;
    }
    QString userId = pws->pw_name;

    /* If we match then turn the start and stop actions back on
       We only want the specified user to be able to perform those actions */
    if (userId == systemUserId)
    {
        actionStartStop->setEnabled(true);
        actionRestart->setEnabled(true);
    }

    // Read configuration file in IAS_SYS_DIR/etc dir.
    // Insert Plugins into the OUI Main Window
    QString ouirc = getenv("IAS_SYS_DIR");
    if(ouirc.isEmpty())
    {
        return;
    }
    ouirc += "/etc/ouirc";

    QFile file(ouirc);
    if(file.open(QIODevice::ReadOnly))
    {
        QTextStream stream(&file);
        QString line;
        while(!stream.atEnd())
        {
            line = stream.readLine(); // line of text excluding '\n'
            QStringList parms = line.split(QRegExp("\\|"));
            // parms[0] is not used, value is currently only PLUGIN
            QStringList path = parms[1].split(QRegExp("\\,"));
            menuText = path[0];
            QString cmd = parms[2];
            insertPlugin(path, cmd, mbOUI);
        }
        file.close();
    }

    // Look in home directory for user defined configuration file
    // Insert Plugins into the OUI Main Window
    ouirc = getenv("HOME");
    if(ouirc.isEmpty())
    {
        return;
    }

    ouirc += "/.ouirc";

    QFile file2(ouirc);
    if(file2.open(QIODevice::ReadOnly))
    {
        QTextStream stream(&file2);
        QString line;
        while(!stream.atEnd())
        {
            line = stream.readLine(); // line of text excluding '\n'
            QStringList parms = line.split(QRegExp("\\|"));
            // parms[0] is not used, value is currently only PLUGIN
            QStringList path = parms[1].split(QRegExp("\\,"));
            menuText = path[0];
            QString cmd = parms[2];
            insertPlugin(path, cmd, mbOUI);
        }
        file2.close();
    }

    //  Get information about font width in pixels for top level
    //  popup menus in Main Window menu bar.
    QObjectList objList = mbOUI->children();
    count = objList.count();
    // Get fontmetrics
    QFontMetrics fm = mbOUI->fontMetrics();
    int pixelWidth = 0;
    for(int idx = 1; idx < count; idx++)
    {
        // Get width of text for idx
        QString text = objList[idx]->objectName();
        text.replace("&", "");
        text.replace("_", " ");
        pixelWidth += fm.width(text);
        pixelWidth += BLACK_MAGIC; // Sprinkle in a little magic.
    }
    // Resize menubar width, and height.
    resize(pixelWidth, mbOUI->height());
}

void OUIMainWindow::insertPlugin(QStringList path,
                                 QString& cmd,
                                 QObject* parent)
{
    if(parent == NULL)return;
    if(path.isEmpty())return;
     
    QStringList newPath = path;

    // Get first element in the path without extra whitespace.
    QString menuText = newPath[0].simplified();
    // Delete it so that the next element is at the beginning.
    newPath.removeFirst();

    // Generate the object name.
    QString objectName = menuText;
    objectName.replace("&", "");
    objectName.replace(" ", "");

    // Specify pointers to the various object types.
    QMenu* menuItem = NULL;
    QAction* actionItem = NULL;
    QMenuBar* mbPtr = NULL;
    QMenu* popup = NULL;

    popup = qobject_cast<QMenu*>(parent);
    if(popup == NULL)
    {
        mbPtr = qobject_cast<QMenuBar*>(parent);
        if(mbPtr != NULL)
        {
            menuItem = mbPtr->findChild<QMenu*>(objectName);
        }
    }
    else
    {
        menuItem = popup->findChild<QMenu*>(objectName);
    }

    if(menuItem != NULL)
    {
        insertPlugin(newPath, cmd, menuItem);
    }
    else
    {
        if(mbPtr == NULL)
        {
            if(newPath.count() >= 1)
            {
                menuItem = new QMenu(menuText, popup);
                menuItem->setObjectName(objectName);
                menuItem->setFont(menuFont);
                popup->addMenu(menuItem);

                insertPlugin(newPath, cmd, menuItem);
            }
            else
            {
                actionItem = new QAction(menuText, popup);
                actionItem->setFont(menuFont);

                signalMapper->setMapping(actionItem, cmd);
                connect(actionItem, SIGNAL(triggered(bool)),
                        signalMapper, SLOT(map()));

                popup->addAction(actionItem);
            }
        }
        else
        {
            if(newPath.count() >= 1)
            {
                menuItem = new QMenu(menuText, mbPtr);
                menuItem->setObjectName(objectName);
                menuFont.setPointSize(14);
                menuItem->setFont(menuFont);
                menuFont.setPointSize(12);
                mbPtr->addMenu(menuItem);

                insertPlugin(newPath, cmd, menuItem);
            }
            else
            {
                actionItem = new QAction(menuText, mbPtr);
                actionItem->setFont(menuFont);

                signalMapper->setMapping(actionItem, cmd);
                connect(actionItem, SIGNAL(triggered(bool)),
                        signalMapper, SLOT(map()));

                mbPtr->addAction(actionItem);
            }
        }
    }
}

void OUIMainWindow::executePlugin(const QString & cmd)
{
    IASProcessExecution pe(this);
    QString command = cmd;

    if(!pe.executeDetached(command))
    {
        QString msg = tr("Failed to execute plugin.\n") + command;
        QMessageBox::critical(this, "OUI - " + tr("Error"), msg);
    }
}

