/******************************************************************************
               Image Assessment System (IAS) Software
 ******************************************************************************
 *
 * Name:   RestartIas.cpp
 *
 * PURPOSE:
 * This is the definition of the RestartIas class which is used to
 * Startup individual IAS background tasks.
 *
 *****************************************************************************/

// System Includes
#include <iostream>

// QT Includes
#include <QtCore>
#include <QtGui>
#include <QtSql>

// Local Includes
#include "IASSqlQuery.h"
#include "RestartIas.h"

RestartIas::RestartIas(
    QWidget *parent, 
    Qt::WindowFlags fl)
: QWidget(parent, fl)
{
    QWidget::setAttribute(Qt::WA_DeleteOnClose);

    // Set up the user interface.
    setupUi(this);

    QString id = getenv( "IAS_SYSTEM_ID" );
    if ( id.isEmpty() )
    {
        QMessageBox::critical(0,"Missing Environment",
            "Could not get IAS_SYSTEM_ID from environment.\n"
            "This variable is required for system startup or shutdown",
            QMessageBox::Ok,0);
        close();
    }
    id = id.toUpper();
    system_id = id.at(0);

    // Connect all the signals and slots.
    connect(pbCancel, SIGNAL(clicked()), this, SLOT(close()));
    connect(pbOk, SIGNAL(clicked()), this, SLOT(changeDatabase()));
}

/**********************************************************************
Name:  changeDatabase

Purpose:  Update the database so that the individual processes will start.

***********************************************************************/
void RestartIas::changeDatabase()
{
    QString sqlStmt;
    IASSqlQuery query;
    int checkedFlag = false;

    // Check if the directives already exist
    // -------------------------------------
    if(cbPws->isChecked())
    {
        checkedFlag = true;
        if(checkDupDirective("restart PWS"))
            return;
    }
    if(cbDco->isChecked())
    {
        checkedFlag = true;
        if(checkDupDirective("restart capture_order"))
            return;
    }
    if(cbDsi->isChecked())
    {
        checkedFlag = true;
        if(checkDupDirective("restart subsetter_interface"))
            return;
    }
    if(cbDrm->isChecked())
    {
        checkedFlag = true;
        if(checkDupDirective("restart DRM"))
            return;
    }

    // Ensure that at least one directive was selected
    // -----------------------------------------------
    if(!checkedFlag)
    {
        QMessageBox::warning(this, "Error",
            "Please select at least one directive");

        return;
    }

    // Find which directives are to be added and add directive in the database
    // -----------------------------------------------------------------------
    QSqlDatabase::database().transaction();

    // Check PWS
    // ---------
    if(cbPws->isChecked())  // Add PWS directive
    {
        sqlStmt="INSERT INTO ipc_directives (RECIPIENT, DIRECTIVE, SYSTEM_ID) ";
        sqlStmt += "VALUES ('PSI', 'restart PWS', '";
        sqlStmt += system_id + "')";

        if(!query.exec(sqlStmt,__FILE__,__LINE__))
        {
            QSqlDatabase::database().rollback();

            return;
        }
    }

    // Check DCO
    // ---------
    if(cbDco->isChecked())  // Add DCO directive
    {
        sqlStmt="INSERT INTO ipc_directives (RECIPIENT, DIRECTIVE, SYSTEM_ID) ";
        sqlStmt += "VALUES ('PSI', 'restart capture_order', '";
        sqlStmt += system_id + "')";

        if(!query.exec(sqlStmt,__FILE__,__LINE__))
        {
            QSqlDatabase::database().rollback();

            return;
        }
    }

    // Check DSI
    // ---------
    if(cbDsi->isChecked())  // Add DSI directive
    {
        sqlStmt="INSERT INTO ipc_directives (RECIPIENT, DIRECTIVE, SYSTEM_ID) ";
        sqlStmt += "VALUES ('PSI', 'restart subsetter_interface', '";
        sqlStmt += system_id + "')";
        if(!query.exec(sqlStmt,__FILE__,__LINE__))
        {
            QSqlDatabase::database().rollback();

            return;
        }
    }

    // Check DRM
    // ---------
    if(cbDrm->isChecked())  // Add DRM directive
    {
        sqlStmt="INSERT INTO ipc_directives (RECIPIENT, DIRECTIVE, SYSTEM_ID) ";
        sqlStmt += "VALUES ('PSI', 'restart DRM', '";
        sqlStmt += system_id + "')";
        if(!query.exec(sqlStmt,__FILE__,__LINE__))
        {
            QSqlDatabase::database().rollback();

            return;
        }
    }

    if(!QSqlDatabase::database().commit())
    {
        QMessageBox::warning(this, "Error",
            "Commit of database changes failed",
            QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
    }

    close();
}

/**********************************************************************
Name:  checkDupDirective

Purpose:  Check that the directive does not already exist in the database.

***********************************************************************/
bool RestartIas::checkDupDirective(QString directive)
{
    QString sqlStmt;
    QString errMsg;
    IASSqlQuery query;

    // Find if any directives are already entered into the database
    // ------------------------------------------------------------
    sqlStmt="SELECT count(*) FROM ipc_directives ";
    sqlStmt += "WHERE recipient = 'PSI' AND directive = '";
    sqlStmt += directive + "' AND SYSTEM_ID = '" + system_id + "'";

    if(!query.exec(sqlStmt,__FILE__,__LINE__))
    {
        return true;
    }

    query.next();
    // Check if the table has any rows
    // -------------------------------
    if(query.value(0).toInt() > 0)
    {
        errMsg = "Invalid selection:  ";
        errMsg += directive;
        errMsg += " directive exists -- No action taken";
        QMessageBox::warning(this, "Error", errMsg);

        return true;
    }

    return false;
}

