/******************************************************************************
               Image Assessment System (IAS) Software
 ******************************************************************************
 *
 * Name:   SelectModifyWO.cpp
 *
 * PURPOSE: This is the declaration file of the SelectModifyWO class which
 *          is used to select a work order id from existing work orders whose
 *          parameters will be modified.  This class then creates a ModifyWO
 *          interface to modify the work order.
 *
 *****************************************************************************/

// System Includes
#include <iostream>
// QT Includes
#include <QtCore>
#include <QtGui>
#include <QtSql>

// Local Includes
#include "IASSqlQuery.h"
#include "ModifyWO.h"
#include "SelectModifyWO.h"
#include "ModifyWOFactory.h"

SelectModifyWO::SelectModifyWO(QWidget* parent,
                               Qt::WindowFlags fl)
: QWidget(parent, fl)
{
    QWidget::setAttribute(Qt::WA_DeleteOnClose);

    // Set up the user interface.
    setupUi(this);

    //Populate the table
    //------------------
    populate();

    // Change the caption to identify the window.
    setWindowTitle("Select Work Order");

    // Connect all the signals and slots.
    connect(pbCancel, SIGNAL(clicked()), this, SLOT(close()));
    connect(pbQueryDb, SIGNAL(pressed()), this, SLOT(populateWorkOrders()));
    connect(deBeginDate, SIGNAL(dateChanged(const QDate&)),
            this, SLOT(changeToDateMin()));
    connect(deEndDate, SIGNAL(dateChanged(const QDate&)),
            this, SLOT(changeFromDateMax()));
    connect(pbOK, SIGNAL(clicked()), this, SLOT(callWOCreate()));
    connect(cboWorkOrder, SIGNAL(activated(int)), this, SLOT(woSelected(int)));
}

/**********************************************************************
Name:  populate
***********************************************************************/

void SelectModifyWO::populate()
{
    // Set the to and from date
    // ------------------------
    QDate date = QDate::currentDate();

    // Set the order in which the date will appear.
    deBeginDate->setDisplayFormat("MM/dd/yyyy");
    deEndDate->setDisplayFormat("MM/dd/yyyy");

    // Set the date line edits to filter the two weeks work orders
    // We don't need to be able to go back more than 50 years
    //-------------------------------------------------------------
    deBeginDate->setDate(date.addMonths(-2));
    deEndDate->setDate(date);

    deBeginDate->setMinimumDate(date.addYears(-50));
    deBeginDate->setMaximumDate(date);
    deEndDate->setMinimumDate(deBeginDate->date());
    deEndDate->setMaximumDate(date);

    // Insert the work orders into the combo box
    // -----------------------------------------
    populateWorkOrders();
}

/**********************************************************************
Name:  callWOCreate

Purpose:  Call the routine CreateWO to change parameters for the wo
***********************************************************************/
void SelectModifyWO::callWOCreate()
{
    QString workOrderId;

    if (cboWorkOrder->currentText() == "")
    {
        QMessageBox::warning(this, "Warning", "Must select work order ID");
        return;
    }

    /* Get the work order ID from the beginning of the string. */
    workOrderId = cboWorkOrder->currentText();
    workOrderId = workOrderId.section(" ", 0, 0);

    /* If we are already modifying this, then maybe the operator chose the
       wrong entry in the selection box.  So let them try again. */
    if (ModifyWOFactory::instance()->isManaged(workOrderId))
    {
        ModifyWOFactory::instance()->showModifyWindow(workOrderId);
        return;
    }

    /* Create a new ModifyWO interface for the selected work order ID. */
    ModifyWO *modifyWO = ModifyWOFactory::instance()->newModifyWO(workOrderId);

    /* If it failed, maybe it was because it is no longer available or in a
       modifiable state, so let the operator choose again or cancel the
       request. */
    if (modifyWO == NULL)
    {
        return;
    }

    /* Close the window */
    close();
}

/***********************************************************************
Name:  changeToDateMin

Purpose:  This slot will reset the To date edit min value when the
From date edit value is changed
************************************************************************/
void SelectModifyWO::changeToDateMin()
{

    deEndDate->setMinimumDate(deBeginDate->date());
}

/***********************************************************************
Name:  changeFromDateMax

Purpose: This slot will reset the From date edit max value when the
To date edit value is changed
************************************************************************/
void SelectModifyWO::changeFromDateMax()
{
    deBeginDate->setMaximumDate(deEndDate->date());
}

/***********************************************************************
Name:  populateWorkOrders

Purpose: This is called at the beginning and when the query button is
selected.  Fill in the combo box with work orders.
************************************************************************/
void SelectModifyWO::populateWorkOrders()
{
    QString sqlStmt;
    QString woAndProc;
    QString woSceneId;
    IASSqlQuery query;

    // Clear the combo box
    // -------------------
    cboWorkOrder->clear();

    // Change the cursor while waiting for wo from db
    // ----------------------------------------------
    QApplication::setOverrideCursor(Qt::WaitCursor);

    // Get the list of work orders from the database
    //----------------------------------------------
    sqlStmt = "SELECT work_order_id, procedure_id, scene_id ";
    sqlStmt += "FROM work_orders ";
    sqlStmt += "WHERE work_order_id not like 'L%' ";
    sqlStmt += "AND date_entered BETWEEN '";
    sqlStmt += deBeginDate->date().toString("dd-MMM-yyyy");
    sqlStmt += "' AND '";
    sqlStmt += deEndDate->date().addDays(1).toString("dd-MMM-yyyy");
    sqlStmt += "' ORDER BY date_entered DESC";

    // execute the SQL statement
    // -------------------------
    if (!query.exec(sqlStmt,__FILE__,__LINE__))
    {
        QApplication::restoreOverrideCursor();
        return;
    }
    cboWorkOrder->addItem("");

    while (query.next())
    {
        woSceneId = query.value(2).toString();
        if (woSceneId.isEmpty())
        {
            continue;
        }

        woAndProc = query.value(0).toString();
        woAndProc += "   ";
        woAndProc += query.value(1).toString();

        cboWorkOrder->addItem(woAndProc);
    }

    // Return the cursor to it's original shape
    // ----------------------------------------
    QApplication::restoreOverrideCursor();
}

/***********************************************************************
Name:  woSelected

Purpose: This slot is called when a work order is selected.
************************************************************************/
void SelectModifyWO::woSelected(int selection)
{
    if(selection > 0)
        pbOK->setEnabled(TRUE);
    else
        pbOK->setEnabled(FALSE);
}

