/******************************************************************************
               Image Assessment System (IAS) Software
 ******************************************************************************
 *
 * Name:   SelectTemplateWO.cpp
 *
 * PURPOSE: This is the declaration file of the SelectTemplateWO class which
 *          is used to select a work order id from existing work orders whose
 *          parameters will be used to create a new work order.  This class
 *          then creates a TemplateWO interface for creating the new work
 *          order.
 *
 *****************************************************************************/

// System Includes
#include <iostream>

// QT Includes
#include <QtCore>
#include <QtGui>
#include <QtSql>

// Local Includes
#include "IASSqlQuery.h"
#include "SelectTemplateWO.h"
#include "CreateWO.h"
#include "TemplateWO.h"

SelectTemplateWO::SelectTemplateWO(QWidget *parent, 
                                   Qt::WindowFlags fl)
: QWidget(parent, fl)
{
    QWidget::setAttribute(Qt::WA_DeleteOnClose);

    // Set up the user interface.
    setupUi(this);

    populate();

    // Change the caption to identify the window.
    setWindowTitle("Select Work Order");

    // Connect all the signals and slots.
    connect(deBeginDate, SIGNAL(dateChanged(const QDate&)),
            this, SLOT(changeToDateMin(const QDate&)));
    connect(deEndDate, SIGNAL(dateChanged(const QDate&)),
            this, SLOT(changeFromDateMax(const QDate&)));

    connect(pbQueryDb, SIGNAL(pressed()), this, SLOT(populateWorkOrders()));
    connect(pbOK, SIGNAL(clicked()), this, SLOT(callWOCreate()));
    connect(pbCancel, SIGNAL(clicked()), this, SLOT(close()));
}

/**********************************************************************
Name:  populate
***********************************************************************/
void SelectTemplateWO::populate()
{
    // Set the to and from date
    // ------------------------
    QDate date = QDate::currentDate();

    // Set the order in which the date will appear.
    deBeginDate->setDisplayFormat("MM/dd/yyyy");
    deEndDate->setDisplayFormat("MM/dd/yyyy");

    // Set the date line edits to filter the two weeks work orders
    // We don't need to be able to go back more than 50 years
    //-------------------------------------------------------------
    deBeginDate->setDate(date.addMonths(-2));
    deEndDate->setDate(date);

    deBeginDate->setMinimumDate(date.addYears(-50));
    deBeginDate->setMaximumDate(date);
    deEndDate->setMinimumDate(deBeginDate->date());
    deEndDate->setMaximumDate(date);

    // Insert the work orders into the combo box
    // -----------------------------------------
    populateWorkOrders();
}

/**********************************************************************
Name:  callWOCreate
***********************************************************************/
void SelectTemplateWO::callWOCreate()
{
    QString tempWO;

    if (cboWorkOrder->currentText() == "")
    {
        QMessageBox::warning(this, "Error", "Must select Work Order Id");
        return;
    }

    // Get the work order ID from the beginning of the string
    // ------------------------------------------------------
    tempWO = cboWorkOrder->currentText();
    tempWO = tempWO.section(" ",0,0);

    TemplateWO *callCreateWO;
    callCreateWO = new TemplateWO(tempWO, 0);

    // Hide the window and clear it when the CreatWO is finished
    // ---------------------------------------------------------
    hide();

    // Call the CreateWO window
    // ------------------------
    callCreateWO->initialize();
    callCreateWO->show();

    // Close the window
    // ----------------
    close();
}

/***********************************************************************
Name:  changeToDateMin

Purpose:  This slot will reset the To date edit min value when the
From date edit value is changed
************************************************************************/
void SelectTemplateWO::changeToDateMin(const QDate& date)
{
    deEndDate->setMinimumDate(date);
}

/***********************************************************************
Name:  changeFromDateMax

Purpose: This slot will reset the From date edit max value when the
To date edit value is changed
************************************************************************/
void SelectTemplateWO::changeFromDateMax(const QDate& date)
{
    deBeginDate->setMaximumDate(date);
}

/***********************************************************************
Name:  populateWorkOrders

Purpose: This is called at the beginning and when the query button is
selected.  Fill in the combo box with work orders.
************************************************************************/
void SelectTemplateWO::populateWorkOrders()
{
    QString sqlStmt;
    QString woAndProc;
    QString woSceneId;
    IASSqlQuery query;

    // Clear the combo box
    // -------------------
    cboWorkOrder->clear();

    // Change the cursor while waiting for wo from db
    // ----------------------------------------------
    QApplication::setOverrideCursor(Qt::WaitCursor);

    // Get the list of work orders from the database
    //----------------------------------------------
    sqlStmt = "SELECT work_order_id, procedure_id, scene_id ";
    sqlStmt += "FROM work_orders ";
    sqlStmt += "WHERE work_order_id not like 'L%' ";
    sqlStmt += "AND date_entered BETWEEN '";
    sqlStmt += deBeginDate->date().toString("dd-MMM-yyyy");
    sqlStmt += "' AND '";
    sqlStmt += deEndDate->date().addDays(1).toString("dd-MMM-yyyy");
    sqlStmt += "' ORDER BY date_entered DESC";

    // execute the SQL statement
    // -------------------------
    if (!query.exec(sqlStmt,__FILE__,__LINE__))
    {
        return;
    }
    cboWorkOrder->addItem("");

    while (query.next())
    {
        woSceneId = query.value(2).toString();
        if (woSceneId.isEmpty())
        {
            continue;
        }

        woAndProc = query.value(0).toString();
        woAndProc += "   ";
        woAndProc += query.value(1).toString();

        cboWorkOrder->addItem(woAndProc);
    }

    // Return the cursor to it's original shape
    // ----------------------------------------
    QApplication::restoreOverrideCursor();
}

