/******************************************************************************
               Image Assessment System (IAS) Software
 ******************************************************************************
 *
 * Name:   StartStopIas.cpp
 *
 * PURPOSE:
 * This is the definition of the StartStopIas class which is used to
 * Startup and Shutdown the IAS processing system.
 *
 *****************************************************************************/

// System Includes
#include <iostream>

// QT Includes
#include <QtCore>
#include <QtGui>
#include <QtSql>

// Local Includes
#include "IASSqlQuery.h"
#include "StartStopIas.h"

StartStopIas::StartStopIas(
    QWidget *parent, 
    Qt::WindowFlags fl)
: QWidget(parent, fl)
{
    QWidget::setAttribute(Qt::WA_DeleteOnClose);

    // Set up the user interface.
    setupUi(this);

    QString sysid = getenv("IAS_SYSTEM_ID");
    if (sysid.isEmpty())
    {
        QMessageBox::critical(0, "Missing Environment",
            "Could not get IAS_SYSTEM_ID from environment.\n"
            "This variable is required for system startup or shutdown",
            QMessageBox::Ok, 0);
        close();
    }

    ias_bin = getenv("IAS_BIN");
    if (ias_bin.isEmpty())
    {
        QMessageBox::critical(0, "Missing Environment",
            "Could not get IAS_BIN from environment.\n"
            "This variable is required for system startup or shutdown",
            QMessageBox::Ok, 0);
        close();
    }

    sysid = sysid.toUpper();
    system_id = sysid.at(0);

    // Connect all the signals and slots.
    connect(pbCancel, SIGNAL(clicked()), this, SLOT(close()));
    connect(pbShutdownIas, SIGNAL(clicked()), this, SLOT(shutdownIas()));
    connect(pbStartIas, SIGNAL(clicked()), this, SLOT(startupIas()));
}

/**********************************************************************
Name:  shutdownIas

Purpose:  This is called when the user selects the "Shutdown' button.
Update the ipc_directives table indicating a shutdown is to occur.

***********************************************************************/
void StartStopIas::shutdownIas()
{
    QString sqlStmt;
    IASSqlQuery query;

    // Check to see if this directive has been duplicated
    // --------------------------------------------------
    if (checkDupDirective())
        return;

    // Set the insert statement to add the shutdown statement
    // ----------------------------------------------------------
    sqlStmt = "INSERT INTO ipc_directives (RECIPIENT, DIRECTIVE, SYSTEM_ID) ";

    if (rbImmediate->isChecked())
        sqlStmt += "VALUES ('PSI', 'shutdown immediately', '";
    else
        sqlStmt += "VALUES ('PSI', 'shutdown gracefully', '";
    sqlStmt += system_id + "')";

    if (!query.exec(sqlStmt,__FILE__,__LINE__))
    {
        return;
    }
    close();
}

/**************************************************************************
Name:  checkDupDirective

Purpose:  Check if the directive has already been entered.  If so, error out.
***************************************************************************/
bool StartStopIas::checkDupDirective()
{
    QString sqlStmt;
    IASSqlQuery query;

    // Find if any directives are already entered into the database
    // ------------------------------------------------------------
    sqlStmt="SELECT count(*) FROM ipc_directives ";
    sqlStmt += "WHERE directive LIKE 'shutdown%' AND ";
    sqlStmt += "SYSTEM_ID = '" + system_id + "'";

    if (!query.exec(sqlStmt,__FILE__,__LINE__))
    {
        return TRUE;
    }
    query.next();

    // Check if the table has any rows
    // -------------------------------
    if (query.value(0).toInt() > 0)
    {
        QMessageBox::warning(this, "Error",
            "shutdown directive exists -- cannot override");
        return TRUE;
    }
    return FALSE;
}

/**********************************************************************
Name:  startupIAS

Purpose:  Run PSI if the user selects the "Start" button.
***********************************************************************/
void StartStopIas::startupIas()
{
    int status;

    /* Verify the PSI executable exists before trying to execute it */
    QFile psiExecutable(ias_bin + "/PSI");
    if (!psiExecutable.exists())
    {
        QMessageBox::critical(0, "Missing PSI",
            "Error Starting PSI.\n"
            "PSI not installed in " + ias_bin);
    }

    status = system("cd $IAS_BIN && PSI &");
    if (status)
    {
        QMessageBox::critical(this, "Error", "Error starting PSI");
        return;
    }

    close();
}

