/******************************************************************************
               Image Assessment System (IAS) Software
 ******************************************************************************
 *
 * Name:   TemplateWO.cpp
 *
 * PURPOSE: This is the definition of the TemplateWO class which is used to
 *          create a work order based off of a previously created work order.
 *
 *****************************************************************************/

// System Includes
#include <iostream>

// QT Includes
#include <QtCore>
#include <QtGui>
#include <QtSql>

// IAS Qt Includes 
#include <IASQtConst.h>
#include <iasQtWoConst.h>

// Local Includes
#include "TemplateWO.h"
#include "IASSqlQuery.h"
#include "ModScriptWOParms.h"
#include "ModGlobalWOParms.h"
#include "IASSchemaConfig.h"
#include "IASKeyValueComboBox.h"
#include "IASParameterDelegate.h"
#include "IASParameterModel.h"

/* NOTE - leSceneId is used to hold the interval id if the work order happens
          to be one that processes an interval. */

/*
    TemplateWO

    Initializes an instance of the create work order from template interface.
    It inherits a bunch of functionality from the CreateWO parent class.
*/
TemplateWO::TemplateWO(const QString& woId,
                       QWidget *parent, 
                       Qt::WindowFlags fl)
: CreateWO(parent, fl)
{
    m_templateWoId = woId;
}

/*
    setTitle

    This is used to set the titles correctly for the specific interface.
*/
void TemplateWO::setTitle()
{
    QString title = "Create Work Order from Template";

    lblCreateWO->setText(title);
    setWindowTitle(title);
}

/*
    populateInitialDb

    This method overrides the parent class method to populate the interface
    correctly based on information retrieved from the database.
*/
bool TemplateWO::populateInitialDb()
{
    QString date;
    QString sqlStmt;
    QString CurrentTime;
    QString whereClause;
    IASSqlQuery query;
    IASSqlQuery queryOut;
    int findIndex;

    // Set up the where clause.
    whereClause = "WHERE work_order_id = '" + m_templateWoId + "'";

    // Read the work_orders table
    sqlStmt  = "SELECT priority, request_type, ephem_type, ";
    sqlStmt += "procedure_id, cpf_override, scene_id, output_db_schema, ";
    sqlStmt += "l0r_data_path, l0r_data_status, wo_comment, ";
    sqlStmt += "processing_instance FROM work_orders " + whereClause;

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        return false;
    }
    if(!query.next()) // If no results are returned exit the method.
    {
        return false;
    }

    // Fill in the template wo info from the work_orders table
    QString request_type("");
    QString ephem_type("");
    QString output_db_schema("");
    QString processing_instance("");

    if(query.value("REQUEST_TYPE").isValid())
    {
        request_type = query.value("REQUEST_TYPE").toString();
    }

    if(query.value("EPHEM_TYPE").isValid())
    {
        ephem_type = query.value("EPHEM_TYPE").toString();
    }

    if(query.value("PROCESSING_INSTANCE").isValid())
    {
        processing_instance = query.value("PROCESSING_INSTANCE").toString();
    }

    // The output databae schema
    if(query.value("OUTPUT_DB_SCHEMA").isValid())
    {
        output_db_schema = query.value("OUTPUT_DB_SCHEMA").toString();
    }

    // Priority combobox
    findIndex = cboPriority->findText(query.value("PRIORITY").toString(),
        Qt::MatchExactly);
    if (findIndex > -1)
    {
        cboPriority->setCurrentIndex(findIndex);
    }
    else
    {
        selectDefaultPriority();
    }

    // Request Type
    if(request_type == "S")
    {
        rbStandard->setChecked(true); // Standard
    }
    else if(request_type == "T") 
    {
        rbTest->setChecked(true); // Test
    }
    else
    {
        rbCustom->setChecked(true); // Custom
    }

    // Output Database Schema
    if(output_db_schema == "P")
    {
        rbProduction->setChecked(true); // Production Schema
    }
    else
    {
        rbEvaluation->setChecked(true); // Evaluation Schema
    }

    // Cal Parm File
    leCalParmFile->setText(query.value("CPF_OVERRIDE").toString());

    // Procedure Id
    cboProcedureId->clear();
    cboProcedureId->addItem(query.value("PROCEDURE_ID").toString());
    cboProcedureId->setCurrentIndex(0);
    cboProcedureId->setEnabled(false);

    // Processing Instance
    cboInstance->clear();
    cboInstance->addItem(processing_instance);
    cboInstance->setCurrentIndex(0);
    cboInstance->setEnabled(false);

    // Scene/Interval Id
    if(query.value("SCENE_ID").toString() != "")
    {
        leSceneId->setText(query.value("SCENE_ID").toString());
    }

    // L0R Data Path --- Only is set to DB_L0R_DATA_STATUS_LOCAL
    if(query.value("L0R_DATA_STATUS").toInt() == DB_L0R_DATA_STATUS_LOCAL)
    {
        // Use the L0R Data Path value
        leL0RDataPath->setText(query.value("L0R_DATA_PATH").toString());
    }

    // Work Order Comment
    if(query.value("WO_COMMENT").toString() != "")
    {
        leWorkOrderCmnt->setText(query.value("WO_COMMENT").toString());
    }
 
    // Insert the information about the template work order into the
    // work_orders table
    CurrentTime = lblCurrentTime->text();

    sqlStmt = "INSERT INTO work_orders (work_order_id, product_request_id, "
              "date_entered, priority,requester, request_type, ephem_type, "
              "state, req_start_date, procedure_id, "
              "cpf_override, processing_instance) ";
    sqlStmt += "VALUES ('" + lblWorkOrderID->text() + "', '"; // New WO Id
    // Use the work_order_id as the product_request_id
    sqlStmt += lblWorkOrderID->text() + "', ";
    sqlStmt += "TO_DATE('" + CurrentTime + "','DY MON DD HH24:MI:SS YYYY'), ";
    sqlStmt += query.value("PRIORITY").toString() + ", ";
    sqlStmt += "'" + lblRequester->text() + "', "; // Current User
    sqlStmt += "'" + request_type + "',"; // Request Type

    if(!ephem_type.isEmpty())
    {
        sqlStmt += "'" + ephem_type + "',"; // Ephemeris Type
    }
    else
    {
        sqlStmt += "NULL,";
    }

    /* Set to halted so that it doesn't start. */
    sqlStmt += "'" + DB_WO_HALTED + "', "; // State
    sqlStmt += " TO_DATE('" + CurrentTime + "','DY MON DD HH24:MI:SS YYYY'), ";
    sqlStmt += "'" + query.value("PROCEDURE_ID").toString() + "', ";
    sqlStmt += "'" + query.value("CPF_OVERRIDE").toString() + "', ";
    sqlStmt += "'" + processing_instance + "')";

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        QSqlDatabase::database().rollback();
        close();
    }

    // Fill in the wo_scripts table with the template information
    sqlStmt  = "SELECT script_id, pause_flag, seq_nbr ";
    sqlStmt += "FROM wo_scripts " + whereClause;

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        QSqlDatabase::database().rollback();

        return false;
    }

    while(query.next())
    {
        // Set up the sql statement to enter into wo_scripts
        sqlStmt  = "INSERT INTO wo_scripts ";
        sqlStmt += "(work_order_id, script_id, pause_flag, seq_nbr, state) ";
        sqlStmt += "VALUES ('" + lblWorkOrderID->text() + "', ";
        sqlStmt += "'" + query.value("SCRIPT_ID").toString() + "', ";
        sqlStmt += query.value("PAUSE_FLAG").toString() + ", ";
        sqlStmt += query.value("SEQ_NBR").toString() + ", ";
        sqlStmt += "'" + DB_WO_SCRIPT_PENDING + "')";

        if(!queryOut.exec(sqlStmt, __FILE__, __LINE__))
        {
            QSqlDatabase::database().rollback();

            return false;
        }
        queryOut.next();
    }

    // Fill in the script ID in the form.
    populateScriptId(lblWorkOrderID->text());

    // Fill in the wo_common_parms table
    sqlStmt = "SELECT parm_name, parm_occurrence, parm_value ";
    sqlStmt += "FROM wo_common_parms " + whereClause;

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        QSqlDatabase::database().rollback();

        return false;
    }

    while(query.next())
    {
        // set up the sql statement to insert into the wo_common_parms table
        sqlStmt  = "INSERT INTO wo_common_parms ";
        sqlStmt += "VALUES ('" + lblWorkOrderID->text() + "', ";
        sqlStmt += "'" + query.value("PARM_NAME").toString() + "', ";
        sqlStmt += query.value("PARM_OCCURRENCE").toString() + ", ";
        sqlStmt += "'" + query.value("PARM_VALUE").toString() + "')";

        if(!queryOut.exec(sqlStmt, __FILE__, __LINE__))
        {
            QSqlDatabase::database().rollback();

            return false;
        }
        queryOut.next();
    }

    // Get the satellite and sensor from the template WO.
    getSatSensor();

    // Set up the definitive ephemeris vs PCD.
    setupDefEphPCD(ephem_type);

    return true;
}

/*
    getSatSensor

    This sets up tje combobox which holds the information about which
    satellite and sensors are available to chose from.  It only populates one
    entry into the combobox and disables so that it can't be modified.

    There is an identical copy of this in ModifyWO.
*/
void TemplateWO::getSatSensor()
{
    QString satSen;    // Satellite and Sensor string
    QString schema;    // schema with work order Id
    QString schemaCmp; // schema to compare to work order Id's schema

    // Find the schema and L0R Id associated with the
    // Work Order Id (or Order Id) for the work order
    IASSchemaConfig *inst = IASSchemaConfig::instance();
    schema = inst->schemaNameFromWOId(m_templateWoId);

    QStringList schemas = inst->allKeys();
    QStringList::iterator itr;
    for(itr = schemas.begin(); itr != schemas.end(); ++itr)
    {
        satSen = (*itr);
        schemaCmp = inst->schemaNameFromKey(satSen);
        if(schema == schemaCmp)
        {
            satSen.replace("_", " ");
            kvcboSatSensor->addKeyValue(satSen, schema);

            // Now that the Sat/Sensor is found, set it in stone.
            kvcboSatSensor->setCurrentIndex(kvcboSatSensor->findText(satSen));

            kvcboSatSensor->setEditable(false);
            kvcboSatSensor->setEnabled(false);
            break;
        }
    }
}

/*
    setupDefEphPCD

    Set the interface ephemeris selection based on the provided value.

    There is an identical copy of this in ModifyWO.
*/
void TemplateWO::setupDefEphPCD(const QString& ephemType)
{
    // Ephemeris Type
    if(kvcboSatSensor->currentText() == "L8 OLITIRS")
    {
        rbPCD->setChecked(true); // Force PCD
        gbEphemerisType->setEnabled(false);
        gbEphemerisType->hide();
    }
    else if(ephemType == "P")
    {
        rbPCD->setChecked(true); // Select PCD
    }
    else
    {
        rbDefinitive->setChecked(true); // Select Definitive
    }
}

/*
    getScriptSql

    This method configures the SQL required to retrieve the script Id and
    pause flag information from the database.

    There is an identical copy of this in ModifyWO.
*/
QString TemplateWO::getScriptSql(const QString& id)
{
    QString sqlStmt = "";

    sqlStmt  = "SELECT script_id, pause_flag ";
    sqlStmt += "FROM wo_scripts ";
    sqlStmt += "WHERE work_order_id = '" + id + "' ";
    sqlStmt += "ORDER BY seq_nbr";

    return sqlStmt;
}

/*
    getWoParmSql

    This configures the SQL required to retrieve the work order parameters for
    the specified work order. 

    This method overrides the parent class method.
*/
QString TemplateWO::getWoParmSql()
{
    QString sqlStmt;

    QString schemaName = kvcboSatSensor->currentValue();
    if(schemaName.isEmpty())
    {
        QMessageBox::warning(this,"Warning","The satellite and sensor must "
                             "be set before inserting WO parameters");

        return "";
    }

    sqlStmt  = "SELECT script_id, parm_name, parm_occurrence, parm_value ";
    sqlStmt += "FROM wo_parms ";
    sqlStmt += "WHERE work_order_id = '" + m_templateWoId + "'";

    return sqlStmt;
}

/*
    populateNewWorkOrder

    This completes the initial insertion of the work order parameters into the
    work order system.  It doesn't need to do as much work as the parent class
    (CreateWO).

    This method overrides the parent class method.
*/
void TemplateWO::populateNewWorkOrder()
{
    insertWOParms();
}

