/******************************************************************************
               Image Assessment System (IAS) Software
 ******************************************************************************
 *
 * Name:   WOProcessing.cpp
 *
 * PURPOSE:
 * This is the definition of the WOProcessing class which is used to
 * modify the state of a work order.
 *
 *****************************************************************************/

// System Includes
#include <iostream>

// QT Includes
#include <QtCore>
#include <QtGui>
#include <QtSql>

// IAS Qt Includes 
#include <IASQtConst.h>
#include <iasQtWoConst.h>

// Local Includes
#include "IASSqlQuery.h"
#include "IASListView.h"
#include "IASListViewItem.h"
#include "IASUtilities.h"
#include "WOProcessing.h"

lvFields woProc[] =
{
    {"WO ID", "WORK_ORDER_ID", "WORK_ORDERS", Id, 100},
    {"State", "STATE", "WORK_ORDERS", Alpha, 20},
    {"Description", NULL, NULL, Alpha, 50},
    {"Requester", "REQUESTER", "WORK_ORDERS", Alpha, 20}
};

/**********************************************************************
Name:  WOProcessing
***********************************************************************/
WOProcessing::WOProcessing(QWidget *parent,
                           Qt::WindowFlags fl)
: QWidget(parent, fl)
{
    QWidget::setAttribute(Qt::WA_DeleteOnClose);
    IASSqlQuery query;
    QStringList requesters; 
 
    // Set up the user interface.
    setupUi(this);

    m_woProcNumCols = ARRAY_SIZE(woProc);

    lvWOIDAndState->setRootIsDecorated(false);
    setupListView();

    /* Code below is dependent on the number of items in the combo box. */
    cboAction->addItem("Complete");
    cboAction->addItem("Abort");
    cboAction->addItem("Retry Ingest Data");


    // Populate the requesters stringlist with all of the requestors

    /* Fill the Requester ID combobox with the Requester ID's, and set the
       default to All Requesters */
    if (query.exec("select distinct requester from work_orders"
                  " order by requester",
                  __FILE__,__LINE__))
    {
        while (query.next())
        {
            requesters.append(query.value(0).toString());
        }
    }

    // Add all the requesters to the Combo Box
    cboRequesterID->addItems(requesters);

    // Add "All Requesters" entry and set as the current selection
    cboRequesterID->insertItem(0, "All Requesters");
    cboRequesterID->setCurrentIndex(0);

    //populate intial entries
    populate();

    // Connect all the signals and slots.
    connect(cboRequesterID, SIGNAL(currentIndexChanged(int)), this,
                                    SLOT(populate())); 
    connect(cboAction, SIGNAL(currentIndexChanged(int)), this,
                                    SLOT(populate()));
    connect(pbOK, SIGNAL(clicked()), this, SLOT(updatedb()));
    connect(pbCancel, SIGNAL(clicked()), this, SLOT(close()));
}

/**********************************************************************
Name:  setupListView 
***********************************************************************/
void WOProcessing::setupListView()
{
    QFont bold = this->font();
    int i;
    QHeaderView* hdr = NULL;
    QStringList labels;

    // Grab the header.
    hdr = lvWOIDAndState->header();

    // Set heading to bold
    bold.setWeight(QFont::Bold);
    hdr->setFont(bold);
    lvWOIDAndState->setSelectionMode(QAbstractItemView::ExtendedSelection);

    // And the column names and widths to lvWOIDAndState list view
    //  and append the column data type to the list
    labels.clear();
    for (i=0; i < m_woProcNumCols; i++)
    {
        labels.append(woProc[i].colName);
        m_listColumnTypes.append(woProc[i].colType);
    }
    lvWOIDAndState->setHeaderLabels(labels);

    // Set the listview to sort on the WO Id column and
    // set to sort in descending
    hdr->setSortIndicatorShown(false);
    hdr->setSortIndicator(0, Qt::DescendingOrder);

    // Set the resizeMode to 'LastColumn' so the last column will
    //  resize itself if a scroll bar is present
    hdr->setStretchLastSection(true);
}

/**********************************************************************
Name:  populate
***********************************************************************/
void WOProcessing::populate()
{
    IASListViewItem *lviTemp;
    QString sqlStmt;
    IASSqlQuery query;
    int currentAction = cboAction->currentIndex();
    int currentRequester = cboRequesterID->currentIndex();

    // Clear the list view
    lvWOIDAndState->clear();

    sqlStmt = "SELECT work_order_id, state,requester FROM work_orders ";
    sqlStmt += "WHERE work_order_id NOT LIKE 'L%' AND state IN ";

    // Get the Desired WO Action from the combo box
    switch (currentAction)
    {
        case 0:  // Complete
        {
            sqlStmt += "('" + DB_WO_EVALUATE + "') ";
            break;
        }
        case 1:  // Abort
        {
            sqlStmt += "('" + DB_WO_REQUEST_FAILED + "', "
                        "'" + DB_WO_SUBSETTER_FAILED + "', "
                        "'" + DB_WO_INGEST_FAILED + "', "
                        "'" + DB_WO_PENDING + "', "
                        "'" + DB_WO_HALTED + "', "
                        "'" + DB_WO_EVALUATE + "') ";
            break;
        }
        case 2:  // Retry Ingest Data
        default:
        {
            sqlStmt += "('" + DB_WO_INGEST_FAILED + "') ";
            break;
        }
    }

    // If Requester is selected append to the Query.
    if (currentRequester > 0)
    {
        sqlStmt += "AND requester = '" + cboRequesterID->currentText() + "' "; 
    }
    sqlStmt += "ORDER BY work_order_id DESC";

    // Execute the SQL statement
    if (!query.exec(sqlStmt, __FILE__, __LINE__))
        return;

    lvWOIDAndState->setSortingEnabled(false);
    while (query.next())
    {
        QString stateValue = query.value("STATE").toString();
        QString stateText = IASUtilities::getStateText(stateValue);

        lviTemp = new IASListViewItem(lvWOIDAndState, m_listColumnTypes);
        lviTemp->setText(0, query.value("WORK_ORDER_ID").toString());
        lviTemp->setText(1, stateValue);
        lviTemp->setText(2, stateText);
    lviTemp->setText(3, query.value("requester").toString());
    }

    /* Resize the columns to fit the data. */
    for (int i = 0; i < m_woProcNumCols; i++)
    {
        lvWOIDAndState->resizeColumnToContents(i);
    }
}

/**********************************************************************
Name:  updatedb
**********************************************************************/
void WOProcessing::updatedb()
{
    int strLength;
    int woCount = 0;
    int msgResult;
    QString message;
    QString strState;
    QString sqlStmt;
    IASSqlQuery query;

    // Set the beginning of the SQL string
    sqlStmt = "UPDATE work_orders SET state = ";

    switch (cboAction->currentIndex())
    {
        case 0:  // Complete
        {
            sqlStmt += "'" + DB_WO_COMPLETE_WO + "'";
            strState = UI_WO_COMPLETE_WO_STR;
            break;
        }
        case 1:  // Abort
        {
            sqlStmt += "'" + DB_WO_ABORT_WO + "'";
            strState = UI_WO_ABORT_WO_STR;
            break;
        }
        case 2:  // Retry Ingest Data
        default:
        {
            sqlStmt += "'" + DB_WO_L0R_READY + "'";
            strState = UI_WO_L0R_READY_STR;
            break;
        }
    }
    sqlStmt += " WHERE work_order_id IN (";

    // Search through the selected items and get the work orders and the count
    QTreeWidgetItemIterator itr(lvWOIDAndState);
    while (*itr)
    {
        if ((*itr)->isSelected())
        {
            woCount++;
            sqlStmt += "'" + (*itr)->text(0) + "',";
        }
        ++itr;
    }
    strLength = sqlStmt.length();
    sqlStmt = sqlStmt.replace(strLength - 1, 1, ")");

    /* Do not change the state if for some reason it is currently running. */
    sqlStmt += " AND state != '" + DB_WO_RUNNING + "'";

    // Bring up a box to ensure that the user does want to change the workorders
    if (woCount > 0)
    {
        message = QString("This will change %1 Work Order State(s) to '%2'")
                  .arg(woCount)
                  .arg(strState);
        msgResult = QMessageBox::information(this, "Change WO State",
                                             message,"OK","Cancel",0,1);
        if (msgResult == 0) // Update the database
        {
            query.exec(sqlStmt,__FILE__,__LINE__);
            close();
        }
    }
    // return to parent if no ID's are selected
    else
        QMessageBox::warning(this, "Error", "Please select a work order");
}

