/******************************************************************************
               Image Assessment System (IAS) Software
 ******************************************************************************
 *
 * Name:   WOStatus.cpp
 *
 * PURPOSE:
 * This is the source file for the WOStatus class.  The purpose of this
 * class is to provide the ability to monitor and modify work orders
 * that are currently in the system.
 *
 *****************************************************************************/

// System Includes
#include <iostream>

// Qt Includes
#include <QtCore>
#include <QtGui>
#include <QtSql>

// IAS Qt Includes 
#include <IASQtConst.h>
#include <iasQtWoConst.h>

// Local Includes
#include "IASListViewItem.h"
#include "IASListView.h"
#include "IASSqlQuery.h"
#include "IASUtilities.h"
#include "WOStatus.h"
#include "ModScriptWOParms.h"
#include "IASSchemaConfig.h"
#include "ModifyWOFactory.h"

/* These columns need to match the field_columns defined below. */
lvFields wosFields[] = {
    {"WO ID", "WORK_ORDER_ID", "WORK_ORDERS", Id, -1},
    {"Instance", "PROCESSING_INSTANCE", "WORK_ORDERS", Alpha, 80},
    {"Requester", "REQUESTER", "WORK_ORDERS", Alpha, 105},
    {"Scene ID", "SCENE_ID", "WORK_ORDERS", Alpha, 165},
    {"Procedure ID", "PROCEDURE_ID", "WORK_ORDERS", Alpha, 175},
    {"Script ID", "SCRIPT_ID", "WO_SCRIPTS", Alpha, 120},
    {"WO State", "STATE", "WORK_ORDERS", Alpha, 80},
    {"Acq. Date", "DATE_ACQUIRED", "SCENES", DateYMD, 135},
    {"Path", "TARGET_WRS_PATH", "SCENES", Number, -1},
    {"Row", "TARGET_WRS_ROW", "SCENES", Number, -1},
    {"Sensor", "SENSOR", "WORK_ORDERS", Alpha, -1},
    {"Satellite", "SATELLITE", "WORK_ORDERS", Alpha, -1},
    {"Db Schema", "OUTPUT_DB_SCHEMA", "WORK_ORDERS", Alpha, 120},
    {"L0R Data Source", "L0R_DATA_STATUS", "WORK_ORDERS", Alpha, -1},
    {"Order#", "CUSTOMER_ORDER_NBR", "WORK_ORDERS", Alpha, 165},
    {"Unit#", "CUSTOMER_UNIT_NBR", "WORK_ORDERS", Alpha, 165}
};

/* These column positions need to match the wosFields defined above. */
enum field_columns
{
    fc_WORK_ORDER_ID = 0,      // 0
    fc_PROCESSING_INSTANCE,    // 1
    fc_REQUESTER,              // 2
    fc_SCENE_ID,               // 3
    fc_PROCEDURE_ID,           // 4
    fc_SCRIPT_ID,              // 5
    fc_STATE,                  // 6
    fc_DATE_ACQUIRED,          // 7
    fc_WRS_PATH,               // 8
    fc_WRS_ROW,                // 9
    fc_SENSOR,                 // 10
    fc_SATELLITE,              // 11
    fc_OUTPUT_DB_SCHEMA,       // 12
    fc_L0R_DATA_STATUS,        // 13
    fc_CUSTOMER_ORDER_NBR,     // 14
    fc_CUSTOMER_UNIT_NBR       // 15
};

//****************************************************************************
WOStatus::WOStatus(QWidget *parent,
                   Qt::WindowFlags fl)
: QWidget(parent, fl)
{
    // This sets up the list view and the defaults for the work order status
    // window

    QWidget::setAttribute(Qt::WA_DeleteOnClose);

    // Set up the user interface.
    setupUi(this);

    m_woNumCols = ARRAY_SIZE(wosFields);

    setupListView();

    setupDefaults();

    populate();

    // Connect all the signals and slots.
    connect(pbStatus, SIGNAL(clicked()), this, SLOT(showScriptStatus()));
    connect(pbModify, SIGNAL(clicked()), this, SLOT(showModWOParms()));
    connect(pbQuery, SIGNAL(clicked()), this, SLOT(populate()));
    connect(pbClose, SIGNAL(clicked()), this, SLOT(close()));
}

//****************************************************************************
void WOStatus::setupListView()
{
    // This sets up the list view columns, sorting and connects the signals

    int i;
    QStringList labels;
    QHeaderView* hdr = NULL;

    hdr = lvWOStatus->header();

    lvWOStatus->setRootIsDecorated(false);

    /* Add the column names and widths to lvWOStatus and append the column
       data type to the list */
    labels.clear();
    for (i=0; i < m_woNumCols; i++)
    {
        labels.append(wosFields[i].colName);
        m_listColumnTypes.append(wosFields[i].colType);
    }
    lvWOStatus->setHeaderLabels(labels);

    /* Set the list view to sort on the WO Id column and to sort in
       descendiing */
    hdr->setSortIndicatorShown(true);
    lvWOStatus->setSortingEnabled(true);

    /* Set the resizeMode to 'LastColumn' so the last column will resize
       itself if a scroll bar is present */
    hdr->setStretchLastSection(true);

    /* Set the itemClicked() signal to the listview */
    connect(lvWOStatus, SIGNAL(itemSelectionChanged()),
            this, SLOT(displayInfo()));
}

//****************************************************************************
void WOStatus::setupDefaults()
{
    // This sets the defaults for the work order status

    IASSqlQuery query;
    QStringList procedures;
    QStringList instances;
    QStringList requesters;
    QStringList schemas = IASSchemaConfig::instance()->allKeys();
    QStringList::iterator itr;

    // Populate the StringList with all procedures from all schemas
    for(itr = schemas.begin(); itr != schemas.end(); ++itr)
    {
        /* Fill the Procedure ID combobox with the Procedure Id's, and set the
           default to All Procedures */
        if(query.exec("select distinct module_id from modules "
                       "where mod_type = 'PROCEDURE'",
                       __FILE__,__LINE__))
        {
            while(query.next())
            {
                procedures.append(query.value(0).toString());
            }
        }
    }
    // Remove duplicate procedure names between the schemas
    procedures.removeDuplicates();
    // Sort the procedure names
    procedures.sort();
    // Add all the procedures to the Combo Box
    cboProcId->addItems(procedures);
    // Add "All Procedures" entry and set as the current selection
    cboProcId->insertItem(0, "All Procedures");
    cboProcId->setCurrentIndex(0);

    // Populate the requesters stringlist with all of the requestors

    /* Fill the Requester ID combobox with the Requester ID's, and set the
       default to All Requesters */
    if(query.exec("select distinct requester from work_orders"
                  " order by requester",
                  __FILE__,__LINE__))
    {
        while(query.next())
        {
            requesters.append(query.value(0).toString());
        }
    }
    // Add all the requesters to the Combo Box
    cboRequesterID->addItems(requesters);
    // Add "All Requesters" entry and set as the current selection
    cboRequesterID->insertItem(0, "All Requesters");
    cboRequesterID->setCurrentIndex(0);

    /* Fill the instances combobox with the instances, and set the
       default to "All Instances" */
    if(query.exec("select distinct processing_instance from work_orders"
                  " order by processing_instance",
                  __FILE__,__LINE__))
    {
        while(query.next())
        {
            instances.append(query.value(0).toString());
        }
    }
    // Add all the instances to the Combo Box
    cboInstance->addItems(instances);

    // Add "All Instances" entry and set as the current selection
    cboInstance->insertItem(0, "All Instances");
    cboInstance->setCurrentIndex(0);

    /* Populates the states filter */
    lbxState->addItem("All States");
    QStringListIterator textIterator(IASUtilities::getListOfStateText());
    while (textIterator.hasNext())
    {
        QString stateText = textIterator.next();
        /* Don't add the LPGS only states */
        if (stateText != UI_WO_INACTIVE_STR &&
            stateText != UI_WO_DIAGNOSTIC_COMPLETE_STR &&
            stateText != UI_WO_READY_TO_SHIP_STR &&
            stateText != UI_WO_TRANSFER_FAILED_STR)
        {
            lbxState->addItem(stateText);
        }
    }
    /* Set the State default to All States */
    lbxState->item(0)->setSelected(true);

    /* Create the timer used for the auto refresh */
    m_timer = new QTimer(this);

    /* Always start with auto refresh turned on
       -- set before connecting the signals */
    chbEnableAutoRefresh->setChecked(true);

    /* Setup the connection for the Auto Refresh */
    connect(chbEnableAutoRefresh, SIGNAL(toggled(bool)), this, 
            SLOT(setAutoRefresh(bool)));

    /* Setup the connection for the Timer object */
    connect(m_timer, SIGNAL(timeout()), this, SLOT(populate()));

    /* Start the auto refresh -- This starts the timer */
    setAutoRefresh(chbEnableAutoRefresh->isChecked());

    /* Set the wo_id and the proc_id variables to empty strings */
    m_woId = "";
}

//****************************************************************************
void WOStatus::populate()
{
    // This queries the database and displays the data in the work order window

    QString startRow, endRow, startNEndRow;
    IASSqlQuery query;
    IASListViewItem *currentListViewItem;
    IASListViewItem *selectedListViewItem = NULL;
    QList<QTreeWidgetItem *> listOfItems;

    bool findWorkOrder;

    if(m_woId != "")
    {
        findWorkOrder = true;
    }
    else
    {
        findWorkOrder = false;
    }

    // This changes the cursor to a wait(processing) cursor
    QApplication::setOverrideCursor(Qt::WaitCursor);

    lvWOStatus->clear();

    /* Build the query to get the work orders to display */
    QString sqlStmt = buildWorkOrderQuery();

    if(query.exec(sqlStmt,__FILE__,__LINE__))
    {
        while(query.next())
        {
            QString woState = query.value("STATE").toString();

            woState = IASUtilities::getStateText(woState);

            currentListViewItem =
                new IASListViewItem(m_listColumnTypes);
            currentListViewItem->setText(fc_WORK_ORDER_ID,
                query.value("WORK_ORDER_ID").toString());
            currentListViewItem->setText(fc_PROCESSING_INSTANCE,
                query.value("PROCESSING_INSTANCE").toString());
            currentListViewItem->setText(fc_REQUESTER,
                query.value("REQUESTER").toString());
            currentListViewItem->setText(fc_SCENE_ID,
                query.value("SCENE_ID").toString());
            currentListViewItem->setText(fc_CUSTOMER_ORDER_NBR,
                query.value("CUSTOMER_ORDER_NBR").toString());
            currentListViewItem->setText(fc_CUSTOMER_UNIT_NBR,
                query.value("CUSTOMER_UNIT_NBR").toString());
            currentListViewItem->setText(fc_L0R_DATA_STATUS,
                query.value("L0R_DATA_STATUS").toString());
            currentListViewItem->setText(fc_STATE, woState);
            currentListViewItem->setText(fc_PROCEDURE_ID,
                query.value("PROCEDURE_ID").toString());
            currentListViewItem->setText(fc_SCRIPT_ID,
                query.value("SCRIPT_ID").toString());
            currentListViewItem->setText(fc_OUTPUT_DB_SCHEMA,
                query.value("OUTPUT_DB_SCHEMA").toString());
            currentListViewItem->setText(fc_SATELLITE,
                query.value("SATELLITE").toString());
            currentListViewItem->setText(fc_SENSOR,
                query.value("SENSOR").toString());
            currentListViewItem->setText(fc_DATE_ACQUIRED,
                query.value("DATE_ACQUIRED").toString());
            currentListViewItem->setText(fc_WRS_PATH,
                query.value("TARGET_WRS_PATH").toString());
            currentListViewItem->setText(fc_WRS_ROW,
                query.value("TARGET_WRS_ROW").toString());

            listOfItems.append(currentListViewItem);

            /* If the user selected a work order and there was
                a refresh, re-select that work order as a common
                courtesy.  Use the m_woId to re-select the correct
                work order.
                If the user selected a work order, findWorkOrder
                will be true.  If that work order is found when
                the new query is performed, set findWorkOrder to
                false so we don't make unneeded string comparisons
            -----------------------------------------------------*/
            if (findWorkOrder)
            {
                if (m_woId == query.value("WORK_ORDER_ID").toString())
                {
                    selectedListViewItem = currentListViewItem;
                    findWorkOrder = false;
                }
            }
        } /* END - while(query.next()) */

        /* Finally add the items to the interface */
        lvWOStatus->addTopLevelItems(listOfItems);

        /* If we are doing Auto Refresh, reset the timer */
        setAutoRefresh(chbEnableAutoRefresh->isChecked());
    }

    /* Resize the columns to fit the data. */
    for (int column = 0; column < m_woNumCols; column++)
    {
        lvWOStatus->resizeColumnToContents(column);
    }

    /* If findWorkOrder is true, we did NOT find the work order when the new
       query was performed.  Set the m_woId variable to "".  If findWorkOrder
       is false, we found the work order.  Call scrollToItem() which makes
       the selected item visible in the screen. */
    if (findWorkOrder)
    {
        m_woId = "";
    }
    else
    {
        if (selectedListViewItem != NULL)
        {
            lvWOStatus->setCurrentItem(selectedListViewItem, true);
        }
        lvWOStatus->scrollToItem(lvWOStatus->currentItem());
    }

    QApplication::restoreOverrideCursor();
}

//****************************************************************************
void WOStatus::displayInfo()
{
    // This displays the directory path and the comments for the work orders

    IASSqlQuery query;
    QString sqlStmt;

    QTreeWidgetItem *lvi = lvWOStatus->currentItem();

    leWOPath->clear();
    leComment->clear();
    leL0RDataPath->clear();

    m_woId = lvi->text(fc_WORK_ORDER_ID);
    sqlStmt = "select wo_path, wo_comment, l0r_data_path from work_orders";
    sqlStmt += " where work_order_id = '" + m_woId + "'";

    if (query.exec(sqlStmt,__FILE__,__LINE__))
    {
        query.next();

        if (query.isValid())
        {
            if (query.value(0).isValid())
            {
                leWOPath->setText(query.value(0).toString());
            }

            if (query.value(1).isValid())
            {
                leComment->setText(query.value(1).toString());
            }

            if (query.value(2).isValid())
            {
                leL0RDataPath->setText(query.value(2).toString());
            }
        }
    }
}

//****************************************************************************
void WOStatus::setAutoRefresh(bool enable)
{
    // This starts and stops the timer that queries the database

    if (enable)
    {
        /* Start/restart the timer */
        m_timer->setSingleShot(true);
        m_timer->start(spinRate->value() * 1000);
    }
    else
    {
        /* Stop it because it may still be running and would cause a refresh
           when it times out */
        m_timer->stop();
    }
}

//****************************************************************************
void WOStatus::showScriptStatus()
{
    // This displays the window that allows the user to modify the Script Status
    QTreeWidgetItem *lvi = lvWOStatus->currentItem();
    QString instance;
    QString satellite;
    QString sensor;

    // If no WO is selected exit out w/o doing anything
    if(lvi == 0)
    {
        return;
    }

    instance = lvi->text(fc_PROCESSING_INSTANCE);
    satellite = lvi->text(fc_SATELLITE);
    sensor = lvi->text(fc_SENSOR);

    // Do not update the script parms if sensor/satellite/instance are empty
    if (sensor.isEmpty() || satellite.isEmpty() || instance.isEmpty())
    {
        QMessageBox::warning(this,"Warning","The satellite, sensor, and "
            "instance must be set before opening the script status window");
        return;
    }

    IASSqlQuery query;

    /* Get the current state */
    QString sqlStmt  = "SELECT state FROM work_orders ";
    sqlStmt += "WHERE work_order_id = '" + lvi->text(fc_WORK_ORDER_ID) + "'";

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        return;
    }
    query.next();

    QString state = query.value("STATE").toString();

    bool allowReset = false;
    bool allowParmChanges = false;

    /* Only the following states can allow modifications to the parameters
       and the scripts.  This is because all the other states are outside
       of running the work order to generate a product.   Also resetting
       of a work order takes it back to the pending state. */
    if(state == DB_WO_PENDING || state == DB_WO_HALTED ||
       state == DB_WO_EVALUATE)
    {
        allowReset = true;
        allowParmChanges = true;
    }

    // Parameters to this routine are:  WOId, Procedure Id, Schema, & Script Id
    ModScriptWOParms *dspModScripts = new ModScriptWOParms(
        lvi->text(fc_WORK_ORDER_ID), lvi->text(fc_PROCEDURE_ID),
        instance, satellite, sensor,"", allowParmChanges, allowReset);

    dspModScripts->exec();
}

//****************************************************************************
void WOStatus::showModWOParms()
{
    // This displays the window that allows the user to Modify the Parameter
    // values

    QTreeWidgetItem *lvi = lvWOStatus->currentItem();
    if(lvi == 0)
    {
        return;
    }

    QString workOrderId = lvi->text(fc_WORK_ORDER_ID);
    ModifyWOFactory::instance()->newModifyWO(workOrderId);
}

//****************************************************************************
QString WOStatus::buildWorkOrderQuery()
{
    /* Define the common columns we need for the inline views */
    QString commonColumns =
        " a.work_order_id,"
        " a.processing_instance, "
        " a.requester,"
        " a.scene_id,"
        " a.customer_order_nbr,"
        " decode (a.customer_unit_nbr, null, '', a.customer_unit_nbr)"
            " as customer_unit_nbr,"
        " decode (a.l0r_data_status, "
            + QString::number(DB_L0R_DATA_STATUS_SUBSETTER) + ", '"
            + UI_L0R_DATA_STATUS_SUBSETTER_STR + "', "
            + QString::number(DB_L0R_DATA_STATUS_LOCAL) + ", '"
            + UI_L0R_DATA_STATUS_LOCAL_STR + "', a.l0r_data_status)"
            " as l0r_data_status,"
        " a.state,"
        " a.procedure_id,"
        " a.satellite,"
        " a.sensor,"
        " decode (a.output_db_schema, 'P', 'Production',"
            " 'E', 'Evaluation', a.output_db_schema)"
            " as output_db_schema,"
        " decode (c.date_acquired, null, '', c.date_acquired)"
            " as date_acquired,"
        " decode (c.target_wrs_path, 0, '', c.target_wrs_path)"
            " as target_wrs_path,"
        " decode (c.target_wrs_row, 0, '', c.target_wrs_row)"
            " as target_wrs_row";

    /* Define the common tables we need for the inline views */
    /* Fix the sensor specific schema name, it already contains the '.' */
    QString commonTables =
        " work_orders a, scenes c";

    /* Figure out the date range portion of the where clause */
    QString daysClause = " a.date_entered >= sysdate - "
        + QString::number(spinInterval->value());

    /* Figure out the requester portion of the where clause */
    QString requesterClause =
        " and a.requester = '" + cboRequesterID->currentText() + "'";

    /* Figure out the requester portion of the where clause */
    QString instanceClause =
        " and a.processing_instance = '" + cboInstance->currentText() + "'";

    /* Figure out the procedure portion of the where clause */
    QString procedureClause =
        " and a.procedure_id = '" + cboProcId->currentText() + "'";

    /* Figure out the state to include in the final results */
    QString stateClause = "";
    if(!lbxState->item(0)->isSelected())
    {
        QString stateStr = "";
        for(int index = 1; index < lbxState->count(); index++)
        {
            if(lbxState->item(index)->isSelected())
            {
                QString stateValue = IASUtilities::
                    getStateValue(lbxState->item(index)->text());
                if(!stateStr.isEmpty())
                {
                    stateStr += ", ";
                }

                stateStr += "'" + stateValue + "'";
            }
        }

        if(!stateStr.isEmpty())
        {
            stateClause = " state in (" + stateStr + ")";
        }
    }

    /* Start building the common where clauses */
    QString commonWhereClauses = " where"
        + daysClause
        + " and a.scene_id is not null"
        + " and a.work_order_id = c.work_order_id(+)";

    /* Only select work orders with a specific requester ID */
    if(cboRequesterID->currentIndex() > 0)
    {
        /* Use the requester clause we built earlier */
        commonWhereClauses += requesterClause;
    }

    /* Only select work orders with a specific instance */
    if(cboInstance->currentIndex() > 0)
    {
        /* Use the requester clause we built earlier */
        commonWhereClauses += instanceClause;
    }

    /* Only select work orders with a specific procedure ID */
    if(cboProcId->currentIndex() > 0)
    {
        /* Use the procedure clause we built earlier */
        commonWhereClauses += procedureClause;
    }

    /* Build the inline view where clauses.  If the order fails before
       the first script runs (PWC problem), convert the null seq_nbr to 1
       to make it look like the first script failed - otherwise the work
       order completely disappears from the display */
    QString inlineViewPart_1_whereClause = commonWhereClauses
        + " and a.state = '" + DB_WO_HALTED + "'"
        + " and a.work_order_id = b.work_order_id"
        + " and b.seq_nbr ="
            " (select nvl(max(c.seq_nbr), 1)"
                " from wo_scripts c"
                " where a.work_order_id = c.work_order_id"
                    " and c.state = '" + DB_WO_SCRIPT_COMPLETE + "')";

    QString inlineViewPart_2_whereClause = commonWhereClauses
        + " and a.state = '" + DB_WO_RUNNING + "'"
        + " and a.work_order_id = b.work_order_id"
        + " and b.state = '" + DB_WO_SCRIPT_RUNNING + "'";

    QString inlineViewPart_3_whereClause = commonWhereClauses
        + " and a.state not in ('" + DB_WO_HALTED + "', '"
            + DB_WO_RUNNING + "')";

    /* Build the inline view parts for inclusion below */
    QString inlineViewPart_1 =
        "(select" + commonColumns + ", b.script_id"
        + " from" + commonTables + ", wo_scripts b"
        + inlineViewPart_1_whereClause + ")";
    QString inlineViewPart_2 =
        "(select" + commonColumns + ", b.script_id"
        + " from" + commonTables + ", wo_scripts b"
        + inlineViewPart_2_whereClause + ")";
    QString inlineViewPart_3 =
        "(select" + commonColumns + ", null script_id"
        + " from" + commonTables
        + inlineViewPart_3_whereClause + ")";

    /* Start building the complete sql */
    QString sql = "select * from ( "
        + inlineViewPart_1
        + " union "
        + inlineViewPart_2
        + " union "
        + inlineViewPart_3
        + " )";

    /* Only select work orders that have a certain state */
    if(!stateClause.isEmpty())
    {
        sql += " where" + stateClause;
    }

    sql += " order by work_order_id desc";

    return sql;
}

