/******************************************************************************
               Image Assessment System (IAS) Software
 ******************************************************************************
 *
 * Name:   main.cpp
 *
 * PURPOSE:
 * This is the main program which instantiates and displays the OUIMainWindow
 * object.  OUIMainWindow is the main window for the IAS Operator User
 * Interface.  The database connection is established before the main window
 * is displayed.
 *
 *****************************************************************************/

// System Includes
#include <unistd.h>

// Qt Includes
#include <QtCore>
#include <QtGui>
#include <QtSql>

/* Local Includes */
#include "OUIMainWindow.h"
#include "IASDatabaseLogin.h"
#include "IASSchemaConfig.h"

void initDBConnections();

int main(int argc, char *argv[])
{
    QApplication app(argc, argv);

    // Connect to the database
    //------------------------
    initDBConnections();

    IASSchemaConfig * inst = IASSchemaConfig::instance();
    if (inst == NULL)
    {
        QMessageBox::critical(0, "Invalid Configuration",
            "An error occurred initializing schema configuration.\n"
            "Terminating the application",
            QMessageBox::Ok,0);
        exit(EXIT_FAILURE);
    }

    // Make sure the IAS_SYSTEM_ID environment is set.  New work orders
    // cannot be created if it is not set.  The check is made here because of
    // an error in the constructor of CreateWO -- see CreateWO:populate().
    QString system_id = getenv("IAS_SYSTEM_ID");
    if (system_id.isEmpty())
    {
        QMessageBox::critical(0, "Missing Environment",
            "Could not get IAS_SYSTEM_ID from environment.\n"
            "This variable is required for work order ID creation.\n"
            "Terminating the application",
            QMessageBox::Ok,0);
        exit(EXIT_FAILURE);
    }

    // Set up and display the main window
    //-----------------------------------
    OUIMainWindow *mainWindow = new OUIMainWindow();
    mainWindow->setWindowTitle("IAS - Operator User Interface");
    mainWindow->show();

    app.setStyle(new QWindowsStyle);

    app.connect(&app, SIGNAL(lastWindowClosed()), &app, SLOT(quit()));

    int status = app.exec();

    return status;
}

//***********************************************************************
void initDBConnections()
{
    // This function attempts to make a default database connection.
    // It gets the connection information from the IASDatabaseLogin dialog.
    // The IASDatabaseLogin dialog attempts the connect and displays any 
    // appropriate error messages.

    QSqlDatabase defDbase;

    // DB_DRIVER is defined in the IASDatabaseLogin.h header file
    defDbase = QSqlDatabase::addDatabase(DB_DRIVER); 
    if (defDbase.isValid()) 
    {
        while (!defDbase.isOpen())
        {
            IASDatabaseLogin loginScreen(NULL, "IAS_OUI_DB");
            loginScreen.setDatabase(defDbase);
            loginScreen.exec();

            // If the user clicked the Cancel button, the dialog returns
            // QDialog::Rejected.  If this happens, then exit the program since
            // there will be no database to access.
            if (loginScreen.result() == QDialog::Rejected)
                exit(EXIT_SUCCESS);
        }
    }
    else
    {
        // An error occurred trying to add a database to QSqlDatabase.
        QMessageBox::critical(0, "Database Connection Error",
                              "Requested database driver does not exist.\n"
                              "Terminating the application",
                              QMessageBox::Ok, 0);
        exit(EXIT_FAILURE);
    }
}

