#!/usr/bin/perl
# Calculate an approximate orbit number for a given date/time,
# using the Landsat 7 3-line orbital elements file S25682.txt
# available from http://www.planet4589.org/space/elements/25600/S25682

use Date::Manip;

if ($ARGV[0] =~ /^(\d\d\d\d):(\d\d\d):(\d\d):(\d\d):(\d\d)/) {
    $year = $1;
    $doy = $2;
    $hour = $3;
    $min = $4;
    $sec = $5;
} elsif ($ARGV[0] =~ /(\d\d\d\d)-(\d\d)-(\d\d)T(\d\d):(\d\d):(\d\d)/) {
    $year = $1;
    $month = $2;
    $day = $3;
    $hour = $4;
    $min = $5;
    $sec = $6;
    $doy = Date_DayOfYear($month,$day,$year);
} else {
    die ("Usage: $0 acq_date\n where acq_date is YYYY:DDD:HH:MM:SS.SSSSSSS\n" .
         "                or YYYY-MM-DDTHH:MM:SS\n");
}

#$mydate = 1999109.56;
$frac_day = ($hour * 60 * 60 + $min * 60 + $sec) / 86400.0;
$mydate = $year * 1000 + $doy + $frac_day;
$myday = $doy + $frac_day;
print "input $ARGV[0] mydate is $mydate\n";

open (ELEMENTS, "S25682.txt") 
    or die ("Unable to open elements file S25682.txt");

$epoch = 0;
while (1) {
    # Save the last epoch & orbit
    $lastepoch = $epoch;
    $lastday = $day;
    $lastorbit = $orbit;

    # Read 3 lines at a time from the elements file
    $line1 = <ELEMENTS> or last;
    $line2 = <ELEMENTS>;
    $line3 = <ELEMENTS>;

    # Get the epoch year and day
    $epoch = substr($line1, 18, 14);
    if ($epoch =~ /^9/) {
        $epoch += 1900000;
    } else {
        $epoch += 2000000;
    }
    #print ("$epoch\n");

    # Get just the day portion
    $day = substr($line1, 20, 12);
    #print "$day\n";

    # Get the orbit 
    $orbit = substr($line2, 63, 5);
    #print "  orbit $orbit\n";

    # If this epoch brackets the input date, interpolate to calculate
    # an approximate orbit number
    if ($lastepoch <= $mydate and $mydate <= $epoch) {
        #print "Found match\n";
        # Calculate the day ranges.  This is an approximation, doesn't take
        # leap year into account.
        $fulldayrange = $day - $lastday;
        $fulldayrange += 365 if ($fulldayrange < 0);
        $mydayrange = $myday - $lastday;
        $mydayrange += 365 if ($mydayrange < 0);

        $myorbit = $lastorbit + ($orbit - $lastorbit) / 
            ($epoch - $lastepoch) * ($mydate - $lastepoch);
        print "Bracketed by $lastepoch $lastorbit  - $epoch $orbit\n";
        #print " day range $fulldayrange\n";
        #print " my day range $mydayrange for $myday to $lastday\n";
        $myorbit = $lastorbit + ($orbit - $lastorbit) / 
            ($fulldayrange) * ($mydayrange);
        print "Orbit for $mydate is $myorbit\n";
        last;
    }
}

close ELEMENTS;

exit 0;

