/* 
 * Changes from the prototype_bpf_data_v3.3 version of this code:
 *   Definitions of even & odd for PAN band are reversed (detectors)
 *   Detectors in even SCAs are reversed
 *   Addition of a1 and C1 to detector_level parameters
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>


#define PAN_ODD        0
#define PAN_EVEN       1
#define NUM_BANDS      9
#define NUM_BLOCKS    12
#define NUM_SCAS      14
#define NUM_DETS     494


int main(int argc, char *argv[])
{
    char a0_filename[PATH_MAX];
    char a1_filename[PATH_MAX];
    char c1_filename[PATH_MAX];
    char preacq_filename[PATH_MAX];
    char pstacq_filename[PATH_MAX];
    char out_odlgroup_filename[PATH_MAX];
    char group_start_line[1000];
    char group_end_line[1000];
    char file_type[20];
    char group_type[20];

    float *preacq = NULL;  /* Pre-acq shutter array */
    float *pstacq = NULL;  /* Post-acq shutter array */
    float *a0 = NULL;      /* Model coeff array */
    float *a1 = NULL;      /* Model coeff array */
    float *c1 = NULL;      /* Model coeff array */

    FILE *a0_fptr = NULL;
    FILE *a1_fptr = NULL;
    FILE *c1_fptr = NULL;
    FILE *preacq_fptr = NULL;
    FILE *pstacq_fptr = NULL;
    FILE *out_odlgroup_fptr = NULL;

    int i;                             /* Internal loop counter */
    int nread;
    int sca_index;
    int num_dets = NUM_DETS;
    int det_index;
    int det_nbr;
    int det_incr;
    int band_number;                    /* 1-based OLI band number */
    int even_odd_sca;                   /* Flag SCAs as even or odd */
    int even_odd_frame;                   /* Flag to distinguish between
                                       odd and even PAN band states -- lines 
                                       within a frame */
    int band_offset,                    /* offset into the array for the band */
        band_offset_even,               /*   for even detectors */
        band_offset_odd;                /*   for odd detectors  */


    /* Create the input and output file names */
    sprintf(preacq_filename, "Sa_example.dat");
    sprintf(pstacq_filename, "Sb_example.dat");
    sprintf(a0_filename, "a0_example.dat");
    sprintf(a1_filename, "a1_example.dat");
    sprintf(c1_filename, "C1_example.dat");

    /* Open the input files */
    preacq_fptr = fopen(preacq_filename, "r");
    pstacq_fptr = fopen(pstacq_filename, "r");
    a0_fptr = fopen(a0_filename, "r");
    a1_fptr = fopen(a1_filename, "r");
    c1_fptr = fopen(c1_filename, "r");

    if (!preacq_fptr || !pstacq_fptr || !a0_fptr || !a1_fptr || !c1_fptr)
    {
        printf("Cannot open one or more input files, aborting\n");
        if (preacq_fptr) fclose(preacq_fptr);
        if (pstacq_fptr) fclose(pstacq_fptr);
        if (a0_fptr) fclose(a0_fptr);
        if (a1_fptr) fclose(a1_fptr);
        if (c1_fptr) fclose(c1_fptr);
        exit(100);
    }

    /* Allocate the data buffers */
    /* Data is layed out in the file in group x SCA x detector order,
     * with the following groups:
     * 1       Pan-odd detectors, odd frames           band 8
     * 2       Pan-even detectors, odd frames          band 8
     * 3       Blue                                    band 2
     * 4       Coastal Aerosol                         band 1
     * 5       NIR                                     band 5
     * 6       Red                                     band 4
     * 7       Green                                   band 3
     * 8       Pan-odd detectors, even frames          band 8
     * 9       Pan-even detectors, even frames         band 8 
     * 10      SWIR2                                   band 7
     * 11      SWIR1                                   band 6
     * 12      Cirrus                                  band 9
     */
    preacq = malloc(NUM_DETS * NUM_SCAS * NUM_BLOCKS * sizeof(float *));
    pstacq = malloc(NUM_DETS * NUM_SCAS * NUM_BLOCKS * sizeof(float *));
    a0     = malloc(NUM_SCAS * NUM_BLOCKS * sizeof(float *));
    a1     = malloc(NUM_DETS * NUM_SCAS * NUM_BLOCKS * sizeof(float *));
    c1     = malloc(NUM_DETS * NUM_SCAS * NUM_BLOCKS * sizeof(float *));
    if (!preacq || !pstacq || !a0 || !a1 || !c1)
    {
        printf("Memory allocation failure\n");
        fclose(preacq_fptr);
        fclose(pstacq_fptr);
        fclose(a0_fptr);
        fclose(a1_fptr);
        fclose(c1_fptr);
        free(preacq);
        free(pstacq);
        free(a0);
        exit(100);
    }

    /* Read the shutter input data */
    nread = fread(preacq, sizeof(*preacq), 
        NUM_DETS * NUM_SCAS * NUM_BLOCKS, preacq_fptr);
    printf("Read %d preacq values\n", nread);
    nread = fread(pstacq, sizeof(*pstacq), 
        NUM_DETS * NUM_SCAS * NUM_BLOCKS, pstacq_fptr);
    printf("Read %d pstacq values\n", nread);

    /* Read the model coefficient a0 */
    printf("Reading %d a0 values size %d\n", NUM_SCAS * NUM_BLOCKS,sizeof(*a0));
    nread = fread(a0, sizeof(*a0), NUM_SCAS * NUM_BLOCKS, a0_fptr);
    printf("Read %d a0 values\n", nread);

    /* Read the model coefficients a1 and C1 */
    nread = fread(a1, sizeof(*a1), 
        NUM_DETS * NUM_SCAS * NUM_BLOCKS, a1_fptr);
    printf("Read %d a1 values\n", nread);
    nread = fread(c1, sizeof(*c1), 
        NUM_DETS * NUM_SCAS * NUM_BLOCKS, c1_fptr);
    printf("Read %d C1 values\n", nread);

    /* a1 and C1 values provided with the ADD relate VRP to L0R image data.
     * The VRP data is 4x the L0R image data, so the a1 part of the solution 
     * includes a .25 multiplier.  We need to get a1 into the full 14-bit space,
     * so multiply by 4. */
    for (i = 0; i < nread; i++)
    {
        a1[i] *= 4.0;
        //c1[i] *= 4.0;
    }

    /* Close the files */
    fclose(preacq_fptr);
    fclose(pstacq_fptr);
    fclose(a0_fptr);
    fclose(a1_fptr);
    fclose(c1_fptr);

    /* Write out the ODL group files */
    for (band_number = 1; band_number <= 9; band_number++)
    {
        if (band_number == 8)
            continue; // do band 8 separately since it's got multiple groups

        switch (band_number)
        {
            case 1:
                band_offset = 3 * NUM_DETS * NUM_SCAS;
                break;
            case 2:
                band_offset = 2 * NUM_DETS * NUM_SCAS;
                break;
            case 3:
                band_offset = 6 * NUM_DETS * NUM_SCAS;
                break;
            case 4:
                band_offset = 5 * NUM_DETS * NUM_SCAS;
                break;
            case 5:
                band_offset = 4 * NUM_DETS * NUM_SCAS;
                break;
            case 6:
                band_offset = 9 * NUM_DETS * NUM_SCAS;
                break;
            case 7:
                band_offset = 10 * NUM_DETS * NUM_SCAS;
                break;
            case 8:
                band_offset  = 0 * NUM_DETS * NUM_SCAS;
                break;
            case 9:
                band_offset = 11 * NUM_DETS * NUM_SCAS;
                break;
            default:
                band_offset = 0;
                break;
        }
        for (sca_index = 0; sca_index < NUM_SCAS; sca_index++)
        {
            printf("Writing ODL group file for band %d SCA %d...\n",
                band_number, (sca_index + 1));

            /* Open the group file */
            sprintf(out_odlgroup_filename,
                "bpf_group_B%02d_SCA%02d.odl",
                band_number, (sca_index + 1));
            out_odlgroup_fptr = fopen(out_odlgroup_filename, "w");
            if (!out_odlgroup_fptr)
            {
                printf("Cannot create ODL group file %s\n",
                    out_odlgroup_filename);
                free(preacq);
                free(pstacq);
                free(a0);
                exit(100);
            }

            /* Create the "GROUP = ..." and "END_GROUP = ..." lines, along
               with the parameter name "base" */
            sprintf(group_start_line,
                "GROUP = BIAS_MODEL_B%02d_SCA%02d\n",
                band_number, (sca_index + 1));
            sprintf(group_end_line,
                "END_GROUP = BIAS_MODEL_B%02d_SCA%02d\n",
                band_number, (sca_index + 1));

            fprintf(out_odlgroup_fptr, "%s", group_start_line);

            /* Write out each detector's data */
            /*  Even SCAs must have their detectors reversed */
            if (sca_index % 2 == 0) // SCA 1, 3, 5, ... odd
            {
                det_nbr = 1;
                det_incr = 1;
            }
            else
            {
                det_nbr = num_dets;
                det_incr = -1;
            }

            for (det_index = 0; det_index < num_dets; det_index++)
            {
                fprintf(out_odlgroup_fptr,
                    "    D%03d = (%8.3lf, %8.3lf, %8.7lf, %9.4lf)\n",
                    det_nbr,
                    preacq[band_offset + sca_index * NUM_DETS + det_index],
                    pstacq[band_offset + sca_index * NUM_DETS + det_index],
                    a1[band_offset + sca_index * NUM_DETS + det_index],
                    c1[band_offset + sca_index * NUM_DETS + det_index]);
                det_nbr += det_incr;
            }
            /* add the a0 to the end of the group */
            fprintf(out_odlgroup_fptr,
                "    A0_Coefficient = %9.7lf\n",
                a0[band_offset/NUM_DETS + sca_index]);

            /* Write out the end line */
            fprintf(out_odlgroup_fptr, "%s", group_end_line);
            //fprintf(out_odlgroup_fptr, "END\n");

            /* Close the output file */
            fclose(out_odlgroup_fptr);
        }
    }

    /* Band 8 odd & even lines within frames */
    band_number = 8;
    for (even_odd_frame = 0; even_odd_frame < 2; even_odd_frame++)
    {
        if (even_odd_frame == PAN_EVEN)
        {
            band_offset_even = 8 * NUM_DETS * NUM_SCAS; /* even detectors */
            band_offset_odd  = 7 * NUM_DETS * NUM_SCAS; /* odd detectors */
            strcpy(group_type, "EVEN");
            strcpy(file_type, "even");
        }
        else  /* PAN_ODD */
        {
            band_offset_even = 1 * NUM_DETS * NUM_SCAS; /* even detectors */
            band_offset_odd  = 0 * NUM_DETS * NUM_SCAS; /* odd detectors */
            strcpy(group_type, "ODD");
            strcpy(file_type, "odd");
        }

        for (sca_index = 0; sca_index < NUM_SCAS; sca_index++)
        {
            printf("Writing ODL odd group file for band %d SCA %d...\n",
                band_number, (sca_index + 1));

            /* Open the group file */
            sprintf(out_odlgroup_filename,
                "bpf_group_%s_B%02d_SCA%02d.odl",
                file_type, band_number, (sca_index + 1));
            out_odlgroup_fptr = fopen(out_odlgroup_filename, "w");
            if (!out_odlgroup_fptr)
            {
                printf("Cannot create ODL group file %s\n",
                    out_odlgroup_filename);
                free(preacq);
                free(pstacq);
                free(a0);
                exit(100);
            }

            /* Create the "GROUP = ..." and "END_GROUP = ..." lines, along
               with the parameter name "base" */
            sprintf(group_start_line,
                "GROUP = BIAS_MODEL_%s_B%02d_SCA%02d\n",
                group_type, band_number, (sca_index + 1));
            sprintf(group_end_line,
                "END_GROUP = BIAS_MODEL_%s_B%02d_SCA%02d\n",
                group_type, band_number, (sca_index + 1));

            fprintf(out_odlgroup_fptr, "%s", group_start_line);

            /*  Even SCAs must have their detectors reversed */
            if (sca_index % 2 == 0) // SCA 1, 3, 5, ... odd
            {
                even_odd_sca = PAN_ODD;
                det_nbr = 1;
                det_incr = 1;
            }
            else
            {
                even_odd_sca = PAN_EVEN;
                /* Set up the detector number to go backwards for even SCAs */
                det_nbr = num_dets * 2;
                det_incr = -1;
            }

            /* Write out each detector's data */
            for (det_index = 0; det_index < num_dets; det_index++)
            {
                fprintf(out_odlgroup_fptr,
                    "    D%03d = (%8.3lf, %8.3lf, %8.7lf, %9.4lf)\n",
                    det_nbr,
                    /* Odd detector */
                    preacq[band_offset_odd+sca_index*NUM_DETS+det_index],
                    pstacq[band_offset_odd+sca_index*NUM_DETS+det_index],
                    a1[band_offset_odd+sca_index*NUM_DETS+det_index],
                    c1[band_offset_odd+sca_index*NUM_DETS+det_index]);
                det_nbr += det_incr;

                fprintf(out_odlgroup_fptr,
                    "    D%03d = (%8.3lf, %8.3lf, %8.7lf, %9.4lf)\n",
                    det_nbr,
                    /* Even detector */
                    preacq[band_offset_even+sca_index*NUM_DETS+det_index],
                    pstacq[band_offset_even+sca_index*NUM_DETS+det_index],
                    a1[band_offset_even+sca_index*NUM_DETS+det_index],
                    c1[band_offset_even+sca_index*NUM_DETS+det_index]);
                det_nbr += det_incr;

            }
            /* add the a0 to the end of the group */
            fprintf(out_odlgroup_fptr,
                "    A0_Coefficient = %9.7lf\n",
                a0[band_offset_odd/NUM_DETS + sca_index]);

            /* Write out the end line */
            fprintf(out_odlgroup_fptr, "%s", group_end_line);
            //fprintf(out_odlgroup_fptr, "END\n");

            /* Close the output file */
            fclose(out_odlgroup_fptr);
        }
    }

    /* Clean up */
    free(preacq);
    free(pstacq);
    free(a0);

    /* Done */
    return 0;
}
