#!/usr/bin/perl -w
use strict;

#  check_tc004_results.pl
#
#  Usage: check_tc004_results.pl <image filename path>
#
#  This tool reads the HDF5 L1G image size and corner coordinates.  Then it
#  performs calculations to make sure image size, pixel size, and corner 
#  coordinates all agree.  h5dump is assumed to be in the path.  It is based on
#  the CalVal scale.pl tool.

my $exec;                   # Command to execute
my $type = "-d \"Band Metadata\" "; # Parameters to pass to tool
my @arr;                    # Array of values
my %metadata;               # Hash of L1G metadata
my $bandname;               # Current band 
my ($nl, $ns);              # Number of lines and samples
my $data;                   # One line of input data
my (%image_nl, %image_ns);  # Arrays of image sizes
my ($pixel_y, $pixel_x);    # Pixel sizes from dump tool output
my ($xsize, $ysize);        # Calculated pixel sizes
my $str;                    # String variable
my ($nlsize, $nssize);      # Image sizes
my ($dy, $dx);              # Difference in corner coordinates

my %bandlut = ( 
   "OLI_Coastal_Aerosol" => "B01",
   "OLI_Blue"            => "B02",
   "OLI_Green"           => "B03",
   "OLI_Red"             => "B04",
   "OLI_NIR"             => "B05",
   "OLI_SWIR1"           => "B06",
   "OLI_SWIR2"           => "B07",
   "OLI_PAN"             => "B08",
   "OLI_CIRRUS"          => "B09",
   "TIRS_THERMAL1"       => "B10",
   "TIRS_THERMAL2"       => "B11");

# Get input file
if (@ARGV != 1) {die ("Usage <h5>\n");}
my $h5file  = $ARGV[0];

$exec = "h5dump ";

# Initialize image size arrays
%image_nl = ();
%image_ns = ();

# Build command to dump image data and extract the number of lines and 
# samples for each band from that dump
my $command = $exec . "-H " . $h5file;
open(META, "$command |") or die "Could not create metadata file";
while ($data = <META>)
{
    chomp($data);

    # Look for DATASET band size listings
    if ($data =~ /DATASET\s\"(B\d{1,2})/)
    {
        $bandname = $1;

        # Skip extra lines
        $data = <META>;
        $data = <META>;
        @arr = split(" ", $data);
        $nl = $arr[5]; # Number of lines in DATASPACE dump output
        $ns = $arr[6]; # Number of samples in DATASPACE dump output

        # Remove any possible commas and store in size arrays
        $nl =~ s/,//g;
        $ns =~ s/,//g;
        $image_nl{$bandname} = $nl;
        $image_ns{$bandname} = $ns;
    }
}
close(META);

# Build command to get L1G band metadata
$command = $exec . $type . $h5file;
open(META, "$command |") or die "Could not create metadata file";

# Find start of band metadata info
while ($data = <META>)
{
   if ($data =~ /DATA\s/) {last;}
}

while ($data = <META>)
{
    chomp($data);
    # End of band metadata
    if ($data =~ /ATTRIBUTE/) {last;}
    # Still in band metadata
    else
    {
        $data =~ s/\s+//g;

        # If bracketed or comma delimited line, append the data
        if ($data =~ /\((\d+)\):/) {$str = $1 . ",";}
        # We are at the end of the data for a band, so process that band's data
        elsif ($data =~ /}/) 
        { 
            # Split the compiled comma-separated string into array elements
            @arr = split(",", $str);

            # Take off the quotes around the band name
            $arr[2] =~ s/"//g;
            unless (exists($arr[2]))
            {
                print "Could not find band $arr[2] in lookup\n";
            }
            # Get the B## name from the lookup table using the full band name
            $bandname = $bandlut{$arr[2]};

            # Merge band metadata and dataset line and sample data, using the 
            # band lookup table results to ensure the right data is matched
            $metadata{$arr[2]} = [@arr[3...$#arr], $image_nl{$bandname}, 
                $image_ns{$bandname}];
        }
        # Concatenate data
        else {$str .= $data;}
        # End of data
        if ($data =~ /ATTRIBUTE/) {last;}
    }
}
close(META);

# Print band metadata
foreach $bandname (keys(%metadata))
{
    print "Band $bandname \n";
    @arr = @{$metadata{$bandname}};

    # Print out corners
    printf "UpLeft     %f, %f\n", $arr[0], $arr[1];
    printf "UpRight    %f, %f\n", $arr[2], $arr[3];
    printf "LowerLeft  %f, %f\n", $arr[4], $arr[5];
    printf "LowerRight %f, %f\n", $arr[6], $arr[7];

    # Verify that the image is north up
    if ($arr[0] != $arr[2] || $arr[4] != $arr[6] ||
        $arr[1] != $arr[5] || $arr[3] != $arr[7])
    {
        print "Image is not north up\n";
    }

    # Extract pixel sizes and force them to float
    $pixel_y = 1.0 * $arr[8];
    $pixel_x = 1.0 * $arr[9];

    # Calculate difference in upper and lower y corner values, and force them
    # to float
    $dy = 1.0 * $arr[0] - 1.0 * $arr[4];

    # Calculate difference in left and right x corner values, and force them
    # to float
    $dx = 1.0 * $arr[3] - 1.0 * $arr[1];

    # Extract number of lines and samples and force them to float
    $nl = 1.0 * $arr[$#arr-1];
    $ns = 1.0 * $arr[$#arr];

    # Calculate the pixel sizes 
    $xsize = $dx / ($ns - 1.0);
    $ysize = $dy / ($nl - 1.0);

    # Round pixel size to nearest 1/100th of a meter
    $xsize = int(($xsize + 0.005) * 100.0) / 100.0;
    $ysize = int(($ysize + 0.005) * 100.0) / 100.0;

    # Print the calculated and retrieved pixel sizes so they can be compared
    printf "Pixel size Y -> Calculated %f File %f\n", $ysize, $pixel_y; 
    printf "Pixel size X -> Calculated %f File %f\n", $xsize, $pixel_x; 

    # Calculate number of lines and samples based on difference between x/y
    # corners and pixels sizes to make sure image size and corners agree
    $nlsize = ($dy / $pixel_y) + 1.0;
    $nssize = ($dx / $pixel_x) + 1.0;

    # Round number of lines and samples
    $nlsize = int(($nlsize + 0.5) * 10.0 / 10.0);
    $nssize = int(($nssize + 0.5) * 10.0 / 10.0);

    # Print the calculated and retrieved number of lines and samples so they
    # can be compared
    printf "Number lines   -> Calculated %f File %f\n", $nlsize, $arr[$#arr-1];
    printf "Number samples -> Calculated %f File %f\n", $nssize, $arr[$#arr];
}

