#!/usr/bin/perl -w
#############################################################################
# This test script runs the TIRS background response determination
# algorithm using a processing parameter file that specifies no
# generation of the optional summary reports.  It is intended to test
# the following requirement as part of test case 007 (TIRS Response
# Determination):
#
# IP2092 -- The IAS shall provide the option to output summary radiometric
#           characterization results in a report.
#
#
# The test script "run_and_check_results.pl" runs the TIRS background
# response algorithm using a processing parameter file that specifies
# generation of the optional summary reports.  The remaining requirements
# for test case 007 can be verified following completion of that script.
#
#############################################################################
use strict;


# Input parameters
my $wo_id;
my $start_date;
my $end_date;


# Global declarations
my $command;
my $rc;
my $template_paramfile_basename = "tc007_br_template";
my $template_filename;
my $dummy_odl = "dummy.odl";
my $dummy2_odl = "dummy2.odl";
my $dummy_omf = "dummy.omf";
my $final_odl = "tc007_br.odl";
my $final_omf = "tc007_br.omf";


# Need three parameters -- a work order ID, a start date for the
# list of deep-space intervals, and an end date for the list
if ($#ARGV != 2)
{
    print "USAGE:  generate_results_noreport.pl <work_order_ID> ",
        "<select_start_date> <select_end_date>\n";
    exit 1;
}


# Get the work order ID and selection start/end dates from the command line
$wo_id = $ARGV[0];
$start_date = $ARGV[1];
$end_date = $ARGV[2];

# Get the needed environmental variables and case subdirectory
my $data_dir = "$ENV{IAS_IT_TEST_DATA_DIR}";
my $exe_dir = "$ENV{IAS_BIN}";
my $tc_dir = "/test_case_007/background_response";
my $wo_dir = "$ENV{IAS_WO}";
my $tc_data_dir = $data_dir . $tc_dir;

# Set the IAS_RPS environment variable to the test directory
$ENV{IAS_RPS} = $data_dir . $tc_dir;


# Copy ODL and OMF template files to the current directory.  We'll need to
# construct the full *.odl template file name before we can copy it.
print "Copying ODL and OMF template files...\n";
$template_filename =
    "$tc_data_dir/$template_paramfile_basename" . "_noreport.odl";
$command = "cp $template_filename $template_paramfile_basename.odl";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}
$command = "cp $tc_data_dir/$template_paramfile_basename.omf .";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}

# Replace the placeholders in the template files with the proper
# locations and names
$command =
    "sed 's|WOID|$wo_id|' $template_paramfile_basename.odl > $dummy_odl";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}
$command = "sed 's|DATE_START|$start_date|' $dummy_odl > $dummy2_odl";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}
$command = "sed 's|DATE_END|$end_date|' $dummy2_odl > $final_odl";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}


# In the get_parameters() routine called by tirs_dark_response_determination,
# the work order ID is used to form the name of the OMF file (we need an
# OMF file to locate a pixel mask file to use in processing).  The template
# OMF file used in this test case does not have this name--we don't know what
# the actual work order ID(s) processing the test scenes are yet.
# One "fix" for this issue is to run a system command to rename the
# OMF file after it has had its true parameter values put into it.
# The idea was to be consistent with the way other RPS apps were
# handling OMF files.
$command =
    "sed 's|IAS_WO|$wo_dir|' $template_paramfile_basename.omf > $dummy_omf";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}
$command = "sed 's|WOID|$wo_id|' $dummy_omf > $final_omf";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}
$command = "mv $final_omf $wo_id.omf";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}


# Run the tirs_background_response_determination algorithm with the test
# case defined in the test_case_007.odl file.  It is expected that the
# standalone application will be built AND INSTALLED in the correct executable
# location on the test server in $IAS_BIN.  If it's not, that's an error...
print "Running: TIRS background response determination...\n";
$command = "$exe_dir/tirs_background_response_determination $final_odl";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}

# Clean up.
$command = "rm *.odl *.omf";
$rc = system($command);
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}
$command = "rm *CPF*";
$rc = system($command);
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}
print "Application completed successfully.\n";
exit 0;
