#!/bin/csh -f

# This script returns a date range for a series of TIRS deep-space intervals
# processed through the RPS.  The starting date in the range is defined as
# the day before the first scene is acquired.  Similarly, the
# ending date in the range is defined as the day after the last scene is acquired.
# This accommodates the current format of the date_acquired field in the
# RADIOMETRIC_WORK_ORDER_COMMON db table, which is 'YYYY-MM-DD', where DD is
# the day of the month, MM is the (2-digit) month--with a leading 0 as
# needed, and YYYY is the (four-digit) year.  There likely will be a time
# component, but for our purposes, it's not needed.
#
# The tester running the background response component of test case 007,
# TIRS Response Determination, will need to run this script prior to
# running the PERL test script 'run_and_check_results.pl', which
# runs the characterization routine and checks the generated output files
# against 'source' files located in
# $IAS_IT_TEST_DATA_DIR/test_case_007/background response.
#


# Ensure there is a database environment variable defined
if (! $?IAS_DB_L8_OLITIRS) then
    echo "IAS_DB_L8_OLITIRS isn't set"
    exit 1
endif


# Create the SQL script on the fly.  For testing purposes, try to limit the
# returned list of intervals to those that were processed 'today' (i.e.
# whenever the RPS procedure generating the histogram statistics was run)
set sql_script = "retrieve_date_range.sql"
echo "set term off" > $sql_script
echo "set echo on" >> $sql_script
echo "set trimout on" >> $sql_script
echo "set trimspool on" >> $sql_script
echo "set heading off" >> $sql_script
echo "set feedback off" >> $sql_script
echo "set pause off" >> $sql_script
echo "set pagesize 0" >> $sql_script
echo "set linesize 500" >> $sql_script
echo "set space 0" >> $sql_script
echo "set colsep |" >> $sql_script
echo "set tab off" >> $sql_script
echo "set verify off" >> $sql_script
echo "spool ./dates.txt" >> $sql_script
echo "select to_char((min(date_acquired)-1), 'YYYY-MM-DD'), " >> $sql_script
echo "to_char((max(date_acquired)+1), 'YYYY-MM-DD')" >> $sql_script
echo "from radiometric_work_order_common" >> $sql_script
echo "where work_order_id" >> $sql_script
echo "in (select work_order_id from radiometric_work_order_common" >> \
    $sql_script
echo "where collection_type='TIRS_DEEPSPACE');" >> $sql_script
echo "spool off" >> $sql_script
echo "exit" >> $sql_script


# Query the database
$ORACLE_HOME/bin/sqlplus $IAS_DB_L8_OLITIRS @$sql_script > & /dev/null

# Remove the SQL file
rm $sql_script

# The query result (two dates in the form DD-MONTH-YY) is on the
# next-to-last line (the SQL prompt with 'spool off' is on the last line).
# Parse the dates then echo them to the screen so the tester can
# record them.  The tester will need to convert the dates returned here
# to the format 'YYYY-MM-DDT12:00:00' in order to pass them into the
# PERL script run_and_check_results.pl.
set date_range = `grep -e '^[[:digit:]][[:digit:]][[:digit:]][[:digit:]]' \
    ./dates.txt | awk -F'|' '{ print $1, $2 }'`
echo $date_range


# Done
exit 0



