#!/usr/bin/perl -w
#############################################################################
# This is the main test script for Test Case 007 - TIRS Response Determination
# (background response determination component).  Requirements are listed
# followed by execution of the script in a manner which
# performs the action specified in the requirement.

# This test case tests compliance to the following requirements:
# IP4108 -- The IAS shall characterize the TIRS deep space data, using the
#           characterization algorithm defined in the Cal/Val Algorithm
#           Description Document (ADD).
#
# IP4211 -- The IAS shall determine the background response for TIRS, as
#           defined in the Cal/Val Algorithm Description Document (ADD).
#
# As background response is part of the overall TIRS detector "bias"
# response, a case could be made that the following requirement is also
# tested:

# IP4112:  The IAS shall derive bias model parameters for each TIRS
#          detector, using the algorithm defined in the Cal/Val Algorithm
#          description document (ADD).
#############################################################################
use strict;


# Input parameters
my $wo_id;
my $start_date;
my $end_date;


# Global declarations
my $command;
my $rc;
my $template_paramfile_basename = "tc007_br_template";
my $dummy_odl = "dummy.odl";
my $dummy2_odl = "dummy2.odl";
my $dummy_omf = "dummy.omf";
my $final_odl = "tc007_br.odl";
my $final_omf = "tc007_br.omf";


# Need three parameters -- a work order ID, a start date for the
# list of deep-space intervals, and an end date for the list
if ($#ARGV != 2)
{
    print "USAGE:  run_and_check_results.pl <work_order_ID> ",
        "<select_start_date> <select_end_date>\n";
    exit 1;
}


# Get the work order ID and selection start/end dates from the command line
$wo_id = $ARGV[0];
$start_date = $ARGV[1];
$end_date = $ARGV[2];

# Get the needed environmental variables and case subdirectory
my $data_dir = "$ENV{IAS_IT_TEST_DATA_DIR}";
my $exe_dir = "$ENV{IAS_BIN}";
my $tc_dir = "/test_case_007/background_response";
my $wo_dir = "$ENV{IAS_WO}";
my $tc_data_dir = $data_dir . $tc_dir;

# Set the IAS_RPS environment variable to the test directory
$ENV{IAS_RPS} = $data_dir . $tc_dir;


# Copy ODL and OMF template files to the current directory
print "Copying ODL and OMF template files...\n";
$command = "cp $tc_data_dir/$template_paramfile_basename.* .";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}


# Replace the placeholders in the template files with the proper
# locations and names
$command =
    "sed 's|WOID|$wo_id|' $template_paramfile_basename.odl > $dummy_odl";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}
$command = "sed 's|DATE_START|$start_date|' $dummy_odl > $dummy2_odl";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}
$command = "sed 's|DATE_END|$end_date|' $dummy2_odl > $final_odl";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}


# In the get_parameters() routine called by tirs_dark_response_determination,
# the work order ID is used to form the name of the OMF file (we need an
# OMF file to locate a pixel mask file to use in processing).  The template
# OMF file used in this test case does not have this name--we don't know what
# the actual work order ID(s) processing the test scenes are yet.
# One "fix" for this issue is to run a system command to rename the
# OMF file after it has had its true parameter values put into it.
# The idea was to be consistent with the way other RPS apps were
# handling OMF files.  On further reflection, this is probably not
# the best way to do this--release 2 should probably fix this issue...
$command =
    "sed 's|IAS_WO|$wo_dir|' $template_paramfile_basename.omf > $dummy_omf";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}
$command = "sed 's|WOID|$wo_id|' $dummy_omf > $final_omf";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}
$command = "mv $final_omf $wo_id.omf";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}



# Run the tirs_background_response_determination algorithm with the test
# case defined in the test_case_007.odl file.  It is expected that the
# standalone application will be built AND INSTALLED in the correct executable
# location on the test server:  $IAS_BIN.  If it's not, that's an error...
print "Running: TIRS background response determination...\n";
$command = "$exe_dir/tirs_background_response_determination $final_odl";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}


# Copy the output files to the source file directory.  This won't be
# necessary after a dry run with the data sets to be processed, and will
# be removed at that point...
#$command = "cp tirs_cpf*.odl background_response_summary*.txt $tc_data_dir";
#$rc = system("$command");
#if ($rc != 0)
#{
#    print "Application failed:  $command\n";
#    exit 1;
#}


# Get the list of output files generated by the characterization with a
# "source" file set containing the expected results
print "Comparing: background response group and summary report files...\n";
my $list_filename = "$tc_data_dir/background_response_comparefiles.txt";
my $listfile;
if (!open($listfile, '<', $list_filename))
{
    print "Cannot open file $list_filename";
    exit 1;
}
my @namelist = <$listfile>;
close $listfile;


# For each base file name, create the full path/name combination in this
# directory and the test data directory, and use the 'diff' command to
# do the actual file comparison.
my $source_filename;
my $target_filename;
my $error_counter = 0;
for my $base_filename (@namelist)
{
    chomp($base_filename);
    $source_filename = "$tc_data_dir/$base_filename";
    $target_filename = "$base_filename";
    $command = "diff -q $source_filename $target_filename";
    $rc = system("$command");
    if ($rc != 0)
    {
        print "Application failed:  $command\n";
        $error_counter++;
    }
}


# If there was an error comparing any file, the whole comparison process
# failed
if ($error_counter)
{
    print "Results comparison failed:\n";
    exit 1;
}


# The comparison completed successfully.  Clean up.
$command = "rm tc007_br* $wo_id.omf $dummy_omf $dummy_odl $dummy2_odl";
$rc = system($command);
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}
$command = "rm *CPF*";
$rc = system($command);
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}
print "Results comparison successfully completed.\n";
exit 0;
