#!/usr/bin/perl -w
#############################################################################
# This is a test script for Test Case 007 - TIRS Response Determination
# (standalone dark response determination component).  Requirements are listed
# followed by execution of the script in a manner which
# performs the action specified in the requirement.

# IP4108:  The IAS shall characterize the TIRS deep space data, using the
#          characterization algorithm defined in the Cal/Val Algorithm
#          Description Document (ADD).
#
# IP4112:  The IAS shall derive bias model parameters for each TIRS
#          detector, using the algorithm defined in the Cal/Val Algorithm
#          description document (ADD).
############################################################################
use strict;


# Input parameters
my $wo_id;
my $cpf_filename;
my $l0r_interval_id;


# Global declarations
my $command;
my $rc;
my $template_paramfile_basename = "tc007_dr_template";
my $dummy_omf = "dummy.omf";
my $dummy2_omf = "dummy2.odl";
my $final_odl = "tc007_dr.odl";
my $final_omf = "tc007_dr.omf";


# Need three command-line parameters, the L0R interval ID to be
# characterized, the work order ID, and the CPF name associated with the
# work order.
if ($#ARGV != 2)
{
    print "USAGE:  run_and_check_results.pl <L0R interval ID>\n";
    print "                                 <work_order_ID> <CPF name>\n";
    exit 1;
}


# Get the L0R interval ID, work order ID and the CPF name from the command
# line
$l0r_interval_id = $ARGV[0];
$wo_id = $ARGV[1];
$cpf_filename = $ARGV[2];

# Get the needed environmental variables and case subdirectory
my $data_dir = "$ENV{IAS_IT_TEST_DATA_DIR}";
my $exe_dir = "$ENV{IAS_BIN}";
my $tc_dir = "/test_case_007/dark_response";
my $wo_dir = "$ENV{IAS_WO}";
my $l0r_dir = "/$l0r_interval_id";
my $tc_parent_dir = $data_dir . $tc_dir;
my $tc_data_dir = $data_dir . $tc_dir . $l0r_dir;

# Set the IAS_RPS environment variable to the test directory
$ENV{IAS_RPS} = $data_dir . $tc_dir;


# Copy ODL and OMF template files to the current directory
print "Copying ODL and OMF template files...\n";
$command = "cp $tc_parent_dir/$template_paramfile_basename.* .";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}


# Replace the placeholders in the template files with the proper
# locations and names
print "Preparing parameter files...\n";
$command =
    "sed 's|WOID|$wo_id|' $template_paramfile_basename.odl > $final_odl";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}


# In the get_parameters() routine called by tirs_dark_response_determination,
# the work order ID is used to form the name of the OMF file (we need an
# OMF file to locate a pixel mask file to use in processing).  The template
# OMF file used in this test case does not have this name--we don't know what
# the actual work order ID(s) processing the test scenes are yet.
# One "fix" for this issue is to run a system command to rename the
# OMF file after it has had its true parameter values put into it.
# The idea was to be consistent with the way other RPS apps were
# handling OMF files.  On further reflection, this is probably not
# the best way to do this--release 2 should probably fix this issue...
$command =
    "sed 's|IAS_WO|$wo_dir|' $template_paramfile_basename.omf > $dummy_omf";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}
$command = "sed 's|WOID|$wo_id|' $dummy_omf > $dummy2_omf";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}
$command = "sed 's|CPFNAME|$cpf_filename|' $dummy2_omf > $final_omf";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}

$command = "mv $final_omf $wo_id.omf";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}



# Run the standalone dark response determination algorithm with the test
# case parameters defined in the tc007_dr.odl file.  It is expected that the
# standalone application will be built AND INSTALLED in the correct executable
# location on the test server:  $IAS_BIN.  If it's not, that's an error...
print "Running: standalone TIRS dark response determination...\n";
$command = "$exe_dir/tirs_dark_response_determination_standalone $final_odl";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}


# Copy the current group files to the correct L0R interval subdirectory.
# Create the subdirectory if it doesn't already exist
$command = "mkdir -p standalone_$l0r_interval_id";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}
$command = "mv cpf_dark_response_group*.odl standalone_$l0r_interval_id";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}


# Clean up
$command = "rm *.omf *.odl";
$rc = system("$command");
if ($rc != 0)
{
    print "Application failed:  $command\n";
    exit 1;
}


# Done!
print "Standalone dark response processing completed.\n";
exit 0;
