#!/usr/bin/perl -w
use strict;

# This is a test script for Test Case 011 - SCA Stitching
# Requirements are listed followed by executing the script in a manner which
# performs the action specified in the requirement.

#### Requirements Tests ####
#
# IP2257 The IAS shall provide a utility that combines the L1R SCA images into 
# a full width L1R image per the L1R SCA Stitching ADD.
#
############################

# Get the work order ID for creating the L1R input file from previous steps
# from command line argument

my $num_args = $#ARGV + 1;
if ( $num_args < 2 )
{
   print "\nPlease provide the work order IDs.\n\n";
   die;
}
my %work_order_ids;
my %cpf_filenames;
my %start_band;
my %sensor;
my $command;
my $rc;
$work_order_ids{'scene'} = $ARGV[0];
$work_order_ids{'interval'} = $ARGV[1];
$cpf_filenames{'scene'} = "L8CPF20000401_20000630.01";
$cpf_filenames{'interval'} = "L8CPF20130101_20130331.02";
$start_band{'scene'} = 1;
$start_band{'interval'} = 10;
$sensor{'scene'} = "OLITIRS";
$sensor{'interval'} = "TIRS";

foreach my $wo_type( keys %work_order_ids )
{
    my $work_order_id = $work_order_ids{$wo_type};

    # Work Order ID directory and name of the input file
    my $workorderdir = "$ENV{IAS_WO}/";
    my $workorderpath = $workorderdir . $work_order_id;
    my $l1r_filename = "l1r_image.h5";
    
    # test data directory and names of output files
    my $data_dir = "$ENV{IAS_IT_TEST_DATA_DIR}";
    my $exe_dir = "$ENV{IAS_BIN}";
    my $tc_dir = "/test_case_011/" . $wo_type;
    
    # define variables for CPF
    my $cpf_filename = $cpf_filenames{$wo_type};
    
    my $tc_data_dir = $data_dir . $tc_dir;
    
    my @files = qw(
        no_overlap_stitch
        left_overlap_stitch
        right_overlap_stitch
        half_overlap_stitch
    );
    my $error_counter = 0;
    my $save_dir = $wo_type;

    # Remove the save directory
    $command = "\\rm -rf $save_dir";
    $rc = system("$command");
    if ($rc != 0)
    {
        print "Application failed: $command\n";
        exit 1;
    }

    # Create a directory to save the files we need to manually examine 
    $command = "\\mkdir -p $save_dir";
    $rc = system("$command");
    if ($rc != 0)
    {
        print "Application failed: $command\n";
        exit 1;
    }

    # Copying ODLs to the current directory
    print "Copying ODLs\n";
    $command = "cp $tc_data_dir/*.odl .";
    $rc = system("$command");
    if ($rc != 0)
    {
        print "Application failed: $command\n";
        exit 1;
    }
    
    # Copying OMFs to the current directory
    print "Copying OMFs\n";
    $command = "cp $tc_data_dir/*.omf .";
    $rc = system("$command");
    if ($rc != 0)
    {
        print "Application failed: $command\n";
        exit 1;
    }
    
    # Copying input L1R file to the current directory
    print "Copying input test $wo_type file\n";
    $command = "cp $workorderpath/$l1r_filename .";
    $rc = system("$command");
    if ($rc != 0)
    {
        print "Application failed: $command\n";
        exit 1;
    }
    
    # Making L0R symoblic link from the work order to current directory
    print "Making L0R symbolic link\n";
    $command = "ln -s $workorderpath/l0r_symlink l0r_symlink";
    $rc = system("$command");
    if ($rc != 0)
    {
        print "Application failed: $command\n";
        exit 1;
    }
    
    # Copying the CPF from the work order directory
    print "Copying CPF from Work Order Directory\n";
    $command = "cp $workorderpath/$cpf_filename .";
    $rc = system("$command");
    if ($rc != 0)
    {
        print "Application failed: $command\n";
        exit 1;
    }
    
    foreach my $file (@files)
    {
        # Running the l1r_sca_stitching algorithm with the test case defined in
        # the test_case_011_no_overlap.odl file
        print "Running: l1r_sca_stitching\n";
        $command = "$exe_dir/l1r_sca_stitching " . "test_case_011_" . $file . 
                   ".odl";
        $rc = system("$command");
        if ($rc != 0)
        {
            print "Application failed: $command\n";
            exit 1;
        }
    
        # Convert to straight binary and generate an ENVI compatible header 
        # file for each band.
    
        # set the binary file root
        my $bin_file_root = $file . "_bin";
        print "Converting the HDF to binary\n";
        $command = "./h52envi.csh " . $bin_file_root . " " . $file . ".h5 "
                    . $sensor{$wo_type};
        $rc = system("$command");
        if ($rc != 0)
        {
            print "Application failed: $command\n";
            exit 1;
        }
    
        # Comparing the the binary image files
        print "Comparing the results\n";
        for (my $band = $start_band{$wo_type}; $band <= 11; $band++)
        {
            my $bin_file;
            if ($band < 10)
            {
                $bin_file = $bin_file_root . "_0" . $band;
            }
            else
            {
                $bin_file = $bin_file_root . "_" . $band;
            }
                
            $command = "diff " . $bin_file . " " . $tc_data_dir; 
            $rc = system("$command");
            if ($rc != 0)
            {
                print "Application failed: $command\n";
                exit 1;
            }
            
            # If this is band 11, then all previous bands passed.  Report pass.
            if ($band == 11)
            {
                print "Binary diffs for $file passed\n";
            }
        }

        # save the files for further examination
        $command = "\\mv $file* $save_dir";
        $rc = system("$command");
        if ($rc != 0)
        {
            print "Application failed: $command\n";
            exit 1;
        }
    }
   
    # Cleaning up
    print "Cleaning up $wo_type dataset\n";
    $command = "rm -rf l1r_image.h5 l0r_symlink";
    $rc = system("$command");
    if ($rc != 0)
    {
        print "Application failed: $command\n";
        exit 1;
    }
}

# Final cleanup of remaining files since finished successfully
unlink $cpf_filenames{'scene'}
    or warn "Could not remove " . $cpf_filenames{'scene'} . ": $!";
unlink $cpf_filenames{'interval'}
    or warn "Could not remove " . $cpf_filenames{'interval'} . ": $!";
$command = "rm -rf *odl *omf";
$rc = system("$command");
if ($rc != 0)
{
    print "Warning: Application cleanup failed: $command\n";
}

# Successfully accomplished
print "Results comparison successfully accomplished\n";
exit 0;
