#!/usr/bin/perl

use warnings;
use strict;

use ias_const;

# constants to define the locations in the array for each dataset.
use constant DATASET_01 => 0;
use constant DATASET_02 => 1;
use constant DATASET_03 => 2;
use constant DATASET_04 => 3;
use constant DATASET_05 => 4;
use constant DATASET_06 => 5;
use constant DATASET_07 => 6;
use constant DATASET_08 => 7;
use constant DATASET_09 => 8;
use constant DATASET_10 => 9;
use constant DATASET_11 => 10;
use constant DATASET_12 => 11;
use constant DATASET_13 => 12;
use constant DATASET_14 => 13;
use constant DATASET_15 => 14;
use constant DATASET_16 => 15;
use constant DATASET_17 => 16;
use constant DATASET_18 => 17;
use constant DATASET_COUNT => 18;

# grab and verify environment variables
my $wo_path = $ENV{IAS_WO};

if (not defined $wo_path)
{
    print "ERROR: Missing environment variable IAS_WO: "
        . "Environment not set up \n";
    exit FAILURE;
}

# only test to see if present
if (not defined $ENV{IAS_DB_COM})
{
    print "ERROR: Missing environment variable IAS_DB_COM: "
        . "Environment not set up \n";
    exit FAILURE;
}

# only test to see if present
if (not defined $ENV{IAS_DB_L8_OLITIRS})
{
    print "ERROR: Missing environment variable IAS_DB_L8_OLITIRS: "
        . "Environment not set up \n";
    exit FAILURE;
}

my $ias_it_test_data_dir = $ENV{IAS_IT_TEST_DATA_DIR};

if (not defined $ias_it_test_data_dir)
{
    print "ERROR: Missing environment variable IAS_IT_TEST_DATA_DIR: "
        . "Environment not set up \n";
    exit FAILURE;
}

# some global variables
my @wo_id = {};
my $test_case_dir = "$ias_it_test_data_dir/test_case_012";
my $test_verify_base = "$test_case_dir/verification.dataset.";

my $status = &_main;
if ($status != SUCCESS)
{
    exit FAILURE;
}

# display the success message and exit with SUCCESS
print "\nSuccessfully executed validation\n";
exit SUCCESS;

##############################################################################
# Begin Subroutines

sub _main()
{
    # grab and verify the input file
    my $input_filename = "";

    if (defined $ARGV[0])
    {
        $input_filename = $ARGV[0];
    }
    else
    {
        print "ERROR: Missing input file on the command line\n";
        return ERROR;
    }

    # grab the IAS work order ID out of the file
    open(INPUT, "<$input_filename") || die "ERROR: Bad file: $!";

    my @wo_id = ();

    my $count = 0;
    while (<INPUT>)
    {
        my $wo = $_;
        chomp($wo);

        if ($wo eq "")
        {
            next;
        }

        $wo =~ / (I\d+), /;

        $wo_id[$count++] = $1;
    }

    close(INPUT);

    if ($count ne DATASET_COUNT)
    {
        print "ERROR: Failed to find the correct number of work orders\n";
        return ERROR;
    }

    # Currently we only validate the results of four of the data sets.
    # The other anomaly data sets should halt at ADP.

    my $status = verify_dataset_02($wo_id[DATASET_02]);
    if ($status != SUCCESS)
    {
        print "ERROR: Failed validation of dataset 02\n";
        return ERROR;
    }

    $status = verify_dataset_10($wo_id[DATASET_10]);
    if ($status != SUCCESS)
    {
        print "ERROR: Failed validation of dataset 10\n";
        return ERROR;
    }

    $status = verify_dataset_12($wo_id[DATASET_12]);
    if ($status != SUCCESS)
    {
        print "ERROR: Failed validation of dataset 12\n";
        return ERROR;
    }

    $status = verify_dataset_13($wo_id[DATASET_13]);
    if ($status != SUCCESS)
    {
        print "ERROR: Failed validation of dataset 13\n";
        return ERROR;
    }

    return SUCCESS;
}

# performs the validation for dataset 02
sub verify_dataset_02($)
{
    my $wo_id = shift;

    print "\nBEGIN: Validation of dataset 02:\n";

    # verify that the work order log exists
    my $wo_log = "$wo_path/$wo_id/save/$wo_id.log";

    if (not -e $wo_log)
    {
        print "ERROR: Work order log file not found: $wo_log\n";
        return ERROR;
    }

    my $verification_dir = "$test_verify_base" . "02";

    # build the verification command to diff the ancillary and model files
    my $verify_cmd = "../tools/verify_test_results.pl"
                   . " $wo_id $verification_dir \$IAS_DB_COM";

    # execute the verification command and return the result
    my $status = system($verify_cmd);

    if ($status)
    {
        print "ERROR: Failed to validate ancillary and model files\n";
        return ERROR;
    }

    foreach my $item ("OLI multispectral", "PAN")
    {
        # build the verification command to find the integration warning
        # messages
        $verify_cmd = "grep 'WARN $item integration time' $wo_log | "
                    . "grep 'exceeds the tolerance' | "
                    . "grep 'nominal integration time'";

        # execute the verification command and return the result
        $status = system($verify_cmd);

        if ($status) # 0 will be returned if the grep was successful.
        {
            print "ERROR: Failed to validate $item warning message in the "
                  . " log file\n";
            return ERROR;
        }
    }

    # build the verification command to find the SSM position warning message
    $verify_cmd = "grep 'WARN SSM position at index' $wo_log | "
                    . "grep 'not' | "
                    . "grep 'nadir position'";

    # execute the verification command and return the result
    $status = system($verify_cmd);

    if ($status) # 0 will be returned if the grep was successful.
    {
        print "ERROR: Failed to validate SSM position warning message in the "
              . " log file\n";
        return ERROR;
    }

    print "\nEND: Validation of dataset 02\n";

    return SUCCESS;
}

# performs the validation for dataset 10
sub verify_dataset_10($)
{
    my $wo_id = shift;

    print "\nBEGIN: Validation of dataset 10:\n";

    # verify that the work order log exists
    my $wo_log = "$wo_path/$wo_id/save/$wo_id.log";

    if (not -e $wo_log)
    {
        print "ERROR: Work order log file not found: $wo_log\n";
        return ERROR;
    }

    my $verification_dir = "$test_verify_base" . "10";

    # build the verification command to diff the ancillary and model files
    my $verify_cmd = "../tools/verify_test_results.pl"
                   . " $wo_id $verification_dir \$IAS_DB_L8_OLITIRS";

    # execute the verification command and return the result
    my $status = system($verify_cmd);

    if ($status)
    {
        print "ERROR: Failed to validate dataset 10\n";
        return ERROR;
    }

    # build the verification command to find the warning message
    $verify_cmd = "grep 'WARN Found' $wo_log | "
                  . "grep 'consecutive invalid IMU gyro samples'";

    # execute the verification command and return the result
    $status = system($verify_cmd);

    if ($status) # 0 will be returned if the grep was successful.
    {
        print "ERROR: Failed to validate warning message in the log file\n";
        return ERROR;
    }

    print "\nEND: Validation of dataset 10\n";

    return SUCCESS;
}

# performs the validation for dataset 12
sub verify_dataset_12($)
{
    my $wo_id = shift;

    print "\nBEGIN: Validation of dataset 12:\n";

    # verify that the work order log exists
    my $wo_log = "$wo_path/$wo_id/save/$wo_id.log";

    if (not -e $wo_log)
    {
        print "ERROR: Work order log file not found: $wo_log\n";
        return ERROR;
    }

    my $verification_dir = "$test_verify_base" . "12";

    # build the verification command to diff the ancillary and model files
    my $verify_cmd = "../tools/verify_test_results.pl"
                   . " $wo_id $verification_dir \$IAS_DB_L8_OLITIRS";

    # execute the verification command and return the result
    my $status = system($verify_cmd);

    if ($status)
    {
        print "ERROR: Failed to validate dataset 12\n";
        return ERROR;
    }

    print "\nEND: Validation of dataset 12\n";

    return SUCCESS;
}

# performs the validation for dataset 13
sub verify_dataset_13($)
{
    my $wo_id = shift;

    print "\nBEGIN: Validation of dataset 13:\n";

    # verify that the work order log exists
    my $wo_log = "$wo_path/$wo_id/save/$wo_id.log";

    if (not -e $wo_log)
    {
        print "ERROR: Work order log file not found: $wo_log\n";
        return ERROR;
    }

    my $verification_dir = "$test_verify_base" . "13";

    # build the verification command to diff the ancillary and model files
    my $verify_cmd = "../tools/verify_test_results.pl"
                   . " $wo_id $verification_dir \$IAS_DB_L8_OLITIRS";

    # execute the verification command and return the result
    my $status = system($verify_cmd);

    if ($status)
    {
        print "ERROR: Failed to validate dataset 13\n";
        return ERROR;
    }

    print "\nEND: Validation of dataset 13\n";

    return SUCCESS;
}
