#!/usr/bin/perl -w
################################################################################
# NAME: verify_ds5.pl
#
# PURPOSE: This script runs I2ICHAR and TDIST on dataset 5.  This is done since
# dataset 5 is run with the L1GT processing flow using a special CPF with
# modified legendre coefficients.  We don't want precision correction done
# on the scene, which would be done if the normal Image Registration processing
# flow was used.  The reference scene in this case is an L1GT scene that was
# processed using the normal CPF with unmodified legendre coefficients.
#
# RETURNS: 0 if data matches, 1 if there was an error
#
# USAGE: ./verify_ds5.pl <WOID>
################################################################################
use strict;

use IO::File;

print "Starting verify_ds5.pl\n"; 

if ($#ARGV != 0)
{
    print "Usage: verify_ds5.pl <WOID> \n";
    print "Exit failure\n";
    exit;
}

my $woid = $ARGV[0];
print "Verifying $woid \n";

# Get the test data dir
my $it_data_dir = $ENV{"IAS_IT_TEST_DATA_DIR"};
if (not defined $it_data_dir)
{
    print "IAS_IT_TEST_DATA_DIR variable is not set\n";
    print "Exit failure";
    exit 1;
}

# Get the WO dir
my $ias_wo = $ENV{"IAS_WO"};
if (not defined $ias_wo)
{
    print "IAS_IT_TEST_DATA_DIR variable is not set\n";
    print "Exit failure";
    exit 1;
}

# Build the directory path
my $wo_dir = "$ias_wo/$woid";
my $data_dir = "$it_data_dir/test_case_014/dataset05";
my $print_wo_dir = "\$IAS_WO/$woid";
my $print_data_dir = "\$IAS_IT_TEST_DATA_DIR/test_case_014/dataset05";

# Generate the I2ICHAR ODL in the WO directory 
my $i2i_fh = IO::File->new("$wo_dir/I2I_CHAR.odl", "w");
if (!$i2i_fh)
{
    print "Error creating I2I_CHAR.odl\n";
    print "Exit failure\n";
    exit 1;
}

my $odl = <<I2I_ODL;
OBJECT = PARAMETERS
    /* Identification string for Work Order */
    /*    Required:  Yes */
    WORK_ORDER_ID = $woid
    /* Defines reference bands to process */
    /*    Required:  No */
    /*    Range: 1 to 11 */
    REFERENCE_BAND_LIST = (8)
    /* Defines search bands to process */
    /*    Required:  No */
    /*    Range: 1 to 11 */
    SEARCH_BAND_LIST = (8)
    /* Reference image that supersedes the default Path/Row DOQ */
    /*    Required:  No */
    L1G_REF_IMAGE = "$it_data_dir/test_case_014/LC80400332013110LGN01.h5"
    /* Search image that supersedes the default L1G image */
    /*    Required:  No */
    L1G_SRCH_IMAGE = "sys_l1g.h5"
    /* Type of tie points to use for correlation (1 = REGULAR_SPACED, */
    /* 2 = retrieve tie points from a file for i2i char or use the grid to */
    /* build tie points for b2b char) */
    /*    Required:  No */
    /*    Range: 1 to 2 */
    TIEPT_TYPE = 1
    /* Number of GCP points in the line direction; these are needed to */
    /* define the number of points if TIEPT_TYPE is REGULAR_SPACED */
    /*    Required:  No */
    /*    Range: 0 to 10000 */
    NLINE_PTS = 15
    /* Number of GCP points in the sample direction; these are needed to */
    /* define the number of points if TIEPT_TYPE is REGULAR_SPACED */
    /*    Required:  No */
    /*    Range: 0 to 10000 */
    NSAMP_PTS = 15
    /* Output data file name */
    /*    Required:  No */
    I2I_CHAR_OUTPUT_FILENAME = "i2i_char.dat"
    /* The current processing pass */
    /*    Required:  Yes */
    /*    Range: 1 to 2 */
    PROCESSING_PASS = 1
END_OBJECT = PARAMETERS
END
I2I_ODL

print $i2i_fh $odl;

$i2i_fh->close();

# Generate the TDIST ODL in the WO directory
my $tdist_fh = IO::File->new("$wo_dir/TDIST.odl", "w");
if (!$tdist_fh)
{
    print "Error creating TDIST.odl \n";
    print "Exit failure\n";
    exit 1;
}

$odl = <<TDIST_ODL;
OBJECT = PARAMETERS
    /* Identification string for Work Order */
    /*    Required:  Yes */
    WORK_ORDER_ID = $woid
    /* Indicates characterization data should be stored in the */
    /* characterization database: 1 to store data, 0 don't store data */
    /*    Required:  No */
    /*    Range: 0 to 1 */
    STORE_CHARACTERIZATION_IN_DB = 1
    /* The current processing pass */
    /*    Required:  Yes */
    /*    Range: 1 to 2 */
    PROCESSING_PASS = 1
    /* Outlier threshold */
    /*    Required:  No */
    /*    Range: 0.000000 to 1.000000 */
    OUTLIER_TOLERANCE = 0.950000
    /* Used to define which parameter to read from the OMF file. (B2B) */
    /* retrieves B2B_CHAR_OUTPUT_FILENAME. (I2I) retrieves */
    /* I2I_CHAR_OUTPUT_FILENAME */
    /*    Required:  Yes */
    /*    Legal Values: B2B, I2I */
    INPUT_GCP_SOURCE = I2I
    /* Output control point residual file name */
    /*    Required:  No */
    RESIDUAL_FILENAME = "i2i_char.res"
    /* Correlation statistics file name */
    /*    Required:  No */
    STATISTICS_FILENAME = "i2i_char.stat"
END_OBJECT = PARAMETERS
END
TDIST_ODL

print $tdist_fh $odl;

$tdist_fh->close();

# Switch to the WO directory
if (!chdir("$wo_dir"))
{
    print "Error changing to directory $wo_dir\n";
    print "Exit failure\n";
    exit 1;
}

# Run I2ICHAR
my $result = system("I2ICHAR", "I2I_CHAR.odl");
if ($result != 0)
{
    print "Error running I2ICHAR\n";
    print "Exit failure\n";
    exit 1;
}

# Run TDIST
$result = system("TDIST", "TDIST.odl");
if ($result != 0)
{
    print "Error running TDIST\n";
    print "Exit failure\n";
    exit 1;
}

# Build the diff command
my $cmd = "sed '/Reference bands/,\$!'d $wo_dir/i2i_char.stat "
        . "| diff -q - $data_dir/i2i_char.stat";
my $print_cmd = "sed '/Reference bands/,\$!'d $print_wo_dir/i2i_char.stat "
        . "| diff - $print_data_dir/i2i_char.stat";

# Run diff
print "Running: $print_cmd \n";

$result = system($cmd);
if ($result != 0)
{
    print "Diff failed: i2i_char.stat\n";
    print "Exit failure\n";
    exit 1;
}

print "Exit success\n";
exit 0;
