#!/bin/csh -f

#***************************************************************************
# NAME:     create_wos.csh
#
#  PURPOSE:  Creates the work orders for test_case_015 using PWG
#  RETURNS:
#        Type:  Integer
#        Value: 0 / 1
#
#  USAGE:    ./create_wos.csh
#****************************************************************************

set mode = "IT"
if ($1 == "ST") then
    set mode = "ST"
endif

set tc = 015
set log = tc${tc}.log
set woid_file = tc${tc}_woids.txt
set tts = "three_tirs_scenes"

# Remove any old log files and work order id files
rm -f $log $woid_file

# Comment fields -- this is the order in which the following lists are
# dependent on
set comments = ( \
    "TC${tc} 01 - Band Calibration OLI SOM projection with cirrus" \
    "TC${tc} 02 - Band Calibration TIRS tiepoint type 1" \
    "TC${tc} 03 - Band Calibration OLI bands 8,2,3,5" \
    "TC${tc} 04 - Band to Band OLI-TIRS" \
    "TC${tc} 05 - Band Calibration OLI off-nadir" \
    )

# L0R data paths in proper order
# $IAS_IT_TEST_DATA_DIR/l0r/ must be pre-pended for them to be complete
set l0r_path_dirs = ( \
    $tts/LC80400332013190LGN00/LC80400332013190LGN00 \
    $tts/LC80400332013190LGN00/LC80400332013190LGN00 \
    $tts/LC80400332013190LGN00/LC80400332013190LGN00 \
    $tts/LC80400332013190LGN00/LC80400332013190LGN00 \
    $tts/LC81750832013112LGN01/LC81750832013112LGN01 \
    )

# Scenes in proper order
set scene_ids = ( \
    LC80400332013190LGN00 \
    LC80400332013190LGN00 \
    LC80400332013190LGN00 \
    LC80400332013190LGN00 \
    LC81750832013112LGN01 \
    )

# Procedures in proper order
set procedures = ( \
    "Band Calibration OLI" \
    "Band Calibration TIRS" \
    "Band Calibration OLI" \
    "Band to Band OLI-TIRS" \
    "Band Calibration OLI" \
    )

# Dataset 1 special parameters using the SOM projection, all bands, and a
# tiepts file
set dataset_1_parms = ( \
    "-parm L8GRID_PREC:TARGET_PROJECTION=22" \
    "-parm B2BCHAR:BAND_LIST='(8,1,2,3,4,5,6,7,9)'" \
    "-parm B2BCHAR:REDUCE_FLAG=1" \
    "-parm B2BCHAR:TIEPT_TYPE=2" \
    )

# Dataset 2 special parameters using a regular grid for tiepts and output the
# CPF group file
set dataset_2_parms = ( \
    "-parm B2BCHAR:TIEPT_TYPE=1" \
    "-parm BANDCAL:CPF_GROUP_FLAG=1" \
    )

# Dataset 3 special parameters to process a subset of bands, output the CPF
# group and output the tiepts for comparison
set dataset_3_parms = ( \
    "-parm B2BCHAR:BAND_LIST='(8,2,3,5)'" \
    "-parm BANDCAL:CPF_GROUP_FLAG=1" \
    "-parm BANDCAL:TIE_POINT_LISTING_FLAG=1" \
    )

set i = 0
set pwg_status = 0
echo "Inserting work orders" |& tee -a $log
while ( $i < ${#comments} )
    @ i++
    set l0r = ""

    if ( $i == 1 ) then
        set additional_parms = `echo $dataset_1_parms`
    else if ( $i == 2 ) then
        set additional_parms = `echo $dataset_2_parms`
    else if ( $i == 3 ) then
        set additional_parms = `echo $dataset_3_parms`
    else
        set additional_parms = ""
    endif

    if ( $mode == "IT" ) then
        set l0r = "-l0r_data_path $IAS_IT_TEST_DATA_DIR/l0r/$l0r_path_dirs[$i]"
    endif

    set output = `PWG -procedure_id "$procedures[$i]" -scene_id $scene_ids[$i] \
        -i $l0r -comment "$comments[$i]" $additional_parms`
    if ( $status ) then
        echo $output >> $log
        echo "Error loading dataset $i" |& tee -a $log
        @ pwg_status++
    else
        echo $output >> $log
        while ( $output[1] != "Work" )
            shift output
        end
        shift output
        if ( $output[1] != "Order" ) then
            echo "Error parsing work order ID from PWG command for dataset $i"\
                |& tee -a $log
        else
            shift output
            set woid = `echo $output[1] | sed 's/,//'`
            /bin/echo -n "$woid "
            echo $woid >> $woid_file
        endif
    endif
end
echo ""

if ( $pwg_status ) then
    echo "ERROR: $pwg_status errors creating work orders with PWG" \
        |& tee -a $log
    exit 1
else
    echo "SUCCESS: created $i work orders" |& tee -a $log
endif

exit 0
