#!/usr/bin/perl -w
use strict;

###
# check_tc017_results.pl
#
# Usage: check_tc017_results.pl <test directory> <db connection info>
#
# The <test directory> contains configuration files listing what will be 
# verified: 
#   - database_tables.txt lists each table to be checked. Each table should be
#     listed in line ('\n') delimited format. Each line should contain the
#     table name. Example: OLI_SENSOR_ALIGNMENT
#   - report_files.txt contains a list of files generated by Sensor Alignment 
#     to verify, followed by a "|", and then a string to mark the end of the
#     report header, which is ignored. If there is no "|" and subsequent text,
#     there is no report header section to ignore.  The report verification 
#     files in the test directory should be edited to remove the header 
#     information that will be skipped in the compared files.  The test 
#     filenames should match the files generated by Sensor Alignment.
#
# The <test directory> should also contain verification files for each table
# in the database_tables.txt list.  The verification files should be named with
# format '<TABLE_NAME>.db.verify'.
#
# <db connection info> should be an Oracle username and password in format
# username/password. Example: value of environment variable $IAS_DB_L8_OLITIRS
#
# This script is derived from verify_test_results.pl.  A separate script is
# needed for test case 017 because verify_test_results.pl is driven by Work
# Orders at several levels, and test case 017 tests Sensor Alignment, which is 
# outside the Work Order environment.  Some new features of 
# verify_test_results.pl like HDF file processing and template file processing
# have not been added to this tool, since they are not needed to test
# Sensor Alignment.
#
###

use Cwd qw(abs_path getcwd);
use File::Basename;
use DBI;

use constant SUCCESS => 0;
use constant ERROR => 1;

# command line parameters
my $test_dir_path;  # <test directory> - Directory containing config and 
                    #                    verfication data
my $db_info;        # <database connection info> 

# configuration 
my $db_list_filename = "database_tables.txt";
my $report_list_filename = "report_files.txt";

# variables
my @database_tables;
my @report_files;
my @cleanup_files;
my @db_result_files;
my $error_cnt = 0;
my $dump_sql_file;
my $script_path = dirname(abs_path($0));
my $dbh;

### Subroutine declarations
sub usage();
sub db_connect($);
sub db_close();
sub dump_table($$);
sub get_columns($);

##### Process parameters ######################################################

# Verify all command line parameters were provided
if ($#ARGV != 1)
{
    usage();
    exit 1;
}

$test_dir_path = $ARGV[0]; # <test directory>
$db_info = $ARGV[1];       # <database info>

# Verify the test directory path exists
if ( !(-d $test_dir_path) )
{
    print "Invalid test directory path \n";
    usage();
    exit 1;
}

##### Verify Test Results #####################################################

# Load the list of database tables to be verified
if ( -f "$test_dir_path/$db_list_filename" )
{
    open(DBFH, "<$test_dir_path/$db_list_filename");
    while(<DBFH>)
    {
        chomp;
        push(@database_tables, $_);
    } 
    close(DBFH);
}

# Load the list of report files to be verified
if ( -f "$test_dir_path/$report_list_filename" )
{
    open(FFH, "<$test_dir_path/$report_list_filename");
    while(<FFH>)
    {
        chomp;
        push(@report_files, $_);
    }
    close(FFH);
}

# If there's nothing to verify, exit the script
my $files_to_verify = scalar(@database_tables) + scalar(@report_files);
if ($files_to_verify == 0)
{
    print "Nothing to verify. Exiting.\n";
    exit;
}

# Initiate the database connection, so we can pull information from the
# database
my $result = db_connect($db_info);
if ($result != SUCCESS)
{
    print "ERROR: could not connect to database\n";
    exit ERROR;
}

# Dump the database table and verify the results
foreach my $line (@database_tables)
{
    my ($table) = split(/\|/, $line);

    # Dump the table data to a pipe-delimited file.  The file will contain
    # all columns except DATE_PROCESSED, which should not be compared to a
    # benchmark since it changes
    my $rc = dump_table($table, "$table.db.results");
    if ($rc != SUCCESS)
    {
        print "ERROR: dumping data for table $table\n";

        # Report this error in the results
        $error_cnt++;

        next;
    }

    # Save the results file name so the file can be removed later
    push(@cleanup_files, "$table.db.results");

    my $cmd = "diff -q $table.db.results " . "$test_dir_path/$table.db.verify";
    
    print "Running: $cmd\n";

    $rc = system($cmd);
    if ($rc != 0)
    {   
        print "Diff failed: $table\n";
        $error_cnt++;
    }

    print "\n";
}

# Done with the database
db_close();

# Verify report files generated by Sensor Alignment 
foreach my $file (@report_files)
{ 
    my ($file, $cutoff_str) = split(/\|/, $file);

    my $cmd;

    if (defined $cutoff_str)
    {
        # Use sed to remove information that changes with each report run (
        # time, etc)
        $cmd = "sed '/$cutoff_str/,\$!'d $file ";
        $cmd .= "| diff -q - $test_dir_path/$file";

        my $print_cmd = "sed '/$cutoff_str/,\$!'d ";
        $print_cmd .= "$file | diff -q - $test_dir_path/$file";
        print "Running: $print_cmd\n";
    }
    else
    {
        # There is no report header section to remove for this file, so just
        # compare it
        $cmd = "diff -q $file $test_dir_path/$file";

        my $print_cmd = "diff -q $file $test_dir_path/$file";
        print "Running: $print_cmd\n";
    }

    my $rc = system($cmd);
    if ($rc != 0)
    {
        print "Diff failed: $file\n";
        $error_cnt++;
    }

    print "\n";
}

if ($error_cnt > 0)
{
    print "Running diffs failed\n";
    exit 1;
}
else
{
    # Test was successful. Clean up database result files
    foreach my $file (@cleanup_files)
    {
        my $rc = system("rm", "-rf", $file);
        if ($rc != 0)
        {   
            print "WARN: could not remove file: $file\n";
        }
    }

    print "Running diffs passed\n";
    exit 0;
}

#### Subroutines ##############################################################

## Usage
sub usage()
{
    print <<USAGE
Usage: check_test_case_017.pl <test directory> <database info>
  <test directory> = Directory containing test data lists and verification data
  <database info> - Database connection information (i.e \$IAS_DB_L8_OLITIRS)
USAGE
}

## db_connect
sub db_connect($)
{
    my ($dbinfo) = @_;
    my ($user, $pass) = split(/\//, $dbinfo);
    my $sid = $ENV{"ORACLE_SID"};

    if (!defined $sid)
    {
        print "ERROR: ORACLE_SID not defined\n";
        return ERROR;
    }

    $dbh = DBI->connect("DBI:Oracle:$sid", $user, $pass,
                         {AutoCommit => 0, PrintError => 0});

    if (!$dbh)
    {
        print "ERROR: Error connecting to the database: ".$DBI::errstr."\n";
        return ERROR;
    }

    return SUCCESS;
}

## db_close
sub db_close()
{
    $dbh->disconnect();
}

## dump_table
# Dump database table information for a single work order to a file.
sub dump_table($$)
{
    my ($tablename, $outfilename) = @_;

    my @columns;

    # Get list of column in the table
    @columns = get_columns($tablename);

    # Verify columns were found for this table
    if (scalar(@columns) == 0)
    {
        print "ERROR: no columns found for table: $tablename\n";
        return ERROR;
    }

    my $column_str = join(",", @columns);

    my $sql = "SELECT $column_str "
            . "FROM $tablename "
            . "WHERE date_processed = "
            . "(SELECT MAX(date_processed) FROM $tablename)";

    my $sth = $dbh->prepare($sql);
    if (!$sth)
    {
        print "ERROR: parsing dump sql ($sql): ";
        return ERROR;
    }

    $sth->execute();

    open(OUTF, ">$outfilename");

    # Write the column header
    print OUTF join("|", @columns) . "\n";

    while (my @data = $sth->fetchrow_array())
    {
        print OUTF join("|", @data) . "\n";
    }
    $sth->finish();

    close(OUTF);

    return SUCCESS;
}

## get_columns
# Return array of columns for a given table name.
# Doesn't return columns contain large binary data or columns with any type of
# unique constraint.
sub get_columns($)
{
    my ($tablename) = @_;
    my @columns;

    $tablename = uc($tablename);

    my $sql = "SELECT column_name "
            . "  FROM all_tab_columns "
            . " WHERE table_name = '$tablename' "
            . "   AND data_type NOT IN ('CLOB', 'BLOB', 'RAW') "
            . "   AND column_name != 'DATE_PROCESSED'"
            . " GROUP BY column_name"
            . " ORDER BY min(column_id)";

    my $sth = $dbh->prepare($sql);

    $sth->execute();

    my @tmp_columns;
    while (my ($column) = $sth->fetchrow_array())
    {
        push(@tmp_columns, $column);
    }
    $sth->finish();

    # Get a list of unique key columns to exclude from the list
    $sql = "SELECT unique(cc.column_name) "
         . "  FROM all_constraints c, "
         . "       all_cons_columns cc "
         . " WHERE c.constraint_name = cc.constraint_name "
         . "   AND c.table_name = '$tablename' "
         . "   AND c.constraint_type IN ('P', 'U')";

    $sth = $dbh->prepare($sql);
    $sth->execute();

    my %exclude_columns;
    while (my ($column) = $sth->fetchrow_array())
    {
        $exclude_columns{$column} = 1;
    }
    $sth->finish();

    # Build the column list
    foreach my $column (@tmp_columns)
    {
        push(@columns, $column) if (not defined $exclude_columns{$column});
    }

    return @columns;
}

