#!/usr/bin/perl -w
############################################################################
#   NAME:  remove_test26_process
#
#   PURPOSE: The remove new test process test will delete all the data
#           in the work order tables of the L8 instance of the database that
#           start with new26.
#
#   PARAMETERS: None
#
#   OUTPUT: The deleted rows of the database
#           Error log messages
#           Process events messages
#
#   NOTE:   All parameters are optional.  If the user enters no parameters,
#           the output files should be identical and this program will
#           return successful.  In order to give this routine flexibility,
#           The user is allowed to enter parameters.  The work order entered
#           will reflect the entered parameters but will likely cause 
#           an error upon return.
#
############################################################################

use strict;
use Getopt::Long;
use Logger;

# include global routines
use ias_create_work_order qw
{
    ias_create_work_order
};

use ias_const;
use ias_processing_system;
use ias_db;

# declare some constants
use constant PROG       => "createwotest";

# debug flag is a method of having select information printed to standard out
# set to 1 to see debug output, set to 0 shut off debug output
use constant DEBUG      =>  0;      # Debug flag

#############################################################################
# Variable declarations
#############################################################################
my $sql;                # SQL statement

my $g_Logger;           # The logger object to use for log messages.

# initialize the ias_ProcSys module
if(ias_init_system() == ERROR)
{
    print STDERR PROG . "$ias_error_msg\n";
    exit FAILURE;
}

# initialize the logger object
$g_Logger = Logger->new($ias_log_filename);
if(!$g_Logger)
{
    print STDERR PROG . ": ERROR: Failed to establish logger object.\n";

    exit FAILURE;
}

set_default_log($g_Logger);
$g_Logger->set_module_name(PROG);

# define list of environment variables needed to run the application
my @vars_list = qw (
    ORACLE_SID
    TWO_TASK
    );

# verify that the environment variables are defined
if (ias_validate_environment(\@vars_list))
{
    $g_Logger->fatal_msg(__FILE__, __LINE__,
                        "Unable to retrieve all environment variables.");
    $g_Logger->fatal_msg(__FILE__, __LINE__,
                        "createwotest has ended");
    $g_Logger->close_log();
    exit FAILURE;
}

# connect to the sensor specific database to add/delete modules
# Note:  This is specifically set up for OLITIRS, adjustments must be made for
# other sensors.
my ($user_id, $password) = ias_db_GetDBConnectionInfo("L8_OLITIRS");

# connect to the database
my $dbhsensor = ias_db_ConnectToDB($user_id, $password);
if (!$dbhsensor)
{
    $g_Logger->fatal_msg(__FILE__, __LINE__,
                            "Error connecting to LDCM database");
    $g_Logger->fatal_msg(__FILE__, __LINE__,
                            "createwotest has ended");
    $g_Logger->close_log();
    exit FAILURE;
}

# delete the test procedure from the database
my $rtn = delete_procedure($dbhsensor);
if ($rtn == ERROR)
{
    $g_Logger->fatal_msg(__FILE__, __LINE__,
                       "Error deleting the test procedure");
    $g_Logger->close_log();
    $dbhsensor->disconnect;
    exit FAILURE;
}

if (!$dbhsensor->commit)
{
    $g_Logger->fatal_msg(__FILE__, __LINE__,
                       "Commit test procedures delete failed");
    $g_Logger->close_log();
    $dbhsensor->disconnect;
    exit FAILURE;
}
$dbhsensor->disconnect;

##############################################################################
#               Subroutines defined here                                     #
##############################################################################


##############################################################################
# Name: delete_procedure
#
# Purpose: Delete the test 26 values in the database module tables
#
##############################################################################
sub delete_procedure
{
    my ($dbhsensor) = @_;
    my $wo_sth;

    # delete the test procedure from the modules table
    my $sql = qq{DELETE FROM modules WHERE module_id like 'New26%'};

    if (! ($wo_sth = $dbhsensor->do($sql)))
    {
        $g_Logger->fatal_msg(__FILE__, __LINE__,
                    "Unable to delete from modules table");
        $dbhsensor->rollback;
        return ERROR;
    }

    # delete the test procedure from sub_modules
    $sql =qq{DELETE FROM sub_modules WHERE module_id like 'New26%'};

    if (! ($wo_sth = $dbhsensor->do($sql)))
    {
        $g_Logger->fatal_msg(__FILE__, __LINE__,
                    "Unable to delete the sub_module table");
        $dbhsensor->rollback;
        return ERROR;
    }

    # delete the test procedure from def_parms
    $sql = qq{DELETE from def_parms WHERE module_id like 'New26%'};

    if (! ($wo_sth = $dbhsensor->do($sql)))
    {
        $g_Logger->fatal_msg(__FILE__, __LINE__,
                    "Unable to delete the def_parms table");
        $dbhsensor->rollback;
        return ERROR;
    }

    # delete the test procedure from parameters
    $sql = qq{DELETE from parameters WHERE module_id like 'New26%'};

    if (! ($wo_sth = $dbhsensor->do($sql)))
    {
        $g_Logger->fatal_msg(__FILE__, __LINE__,
                    "Unable to delete the parameters table");
        $dbhsensor->rollback;
        return ERROR;
    }

    # delete the test procedure from parm_edits
    $sql = qq{DELETE from parm_edits WHERE parm_name like 'New26%'};

    if (! ($wo_sth = $dbhsensor->do($sql)))
    {
        $g_Logger->fatal_msg(__FILE__, __LINE__,
                    "Unable to delete the parm_edits table");
        $dbhsensor->rollback;
        return ERROR;
    }

    # delete the test procedure from parm_list
    $sql = qq{DELETE from parm_list WHERE parm_name like 'New26%'};

    if (! ($wo_sth = $dbhsensor->do($sql)))
    {
        $g_Logger->fatal_msg(__FILE__, __LINE__,
                    "Unable to delete the parm_list table");
        $dbhsensor->rollback;
        return ERROR;
    }

    return SUCCESS;
}
