#!/bin/csh -f
################################################################################
# Name: run_tc032.csh
#
# Purpose: To test that the RLUT loader and RLUT service work correctly
#
# Returns: SUCCESS / ERROR
#
################################################################################
# Set up the log file and SUCCESS and ERROR variables
set ERROR = 1
set SUCCESS = 0
set log = tc032.log
set sym_link = 'L8RLUT20130101_20130131v02.h5'

# Store all of the bad rluts in an array
set bad_rlut = ('L8RLUT20130101_20130131v02.BAD.VERSION.h5' \
    'L8RLUT20130101_20130131v02.BAD.START.DATE.h5' \
    'L8RLUT20130101_20130131v02.BAD.END.DATE.h5' \
    'L8RLUT20130101_20130131v02.BAD.LINEARITY.h5' \
    'L8RLUT20130101_20130131v02.BAD.LINEARIZATION.h5' \
    'L8RLUT20130101_20130131v02.BAD.TIRS.h5')

# Ensure there is no existing sym link or log file
rm -f $sym_link
rm -r $log

# Ensure that the bad RLUTs will not load
foreach rlut ($bad_rlut)
    # Create link with the correct file name
    ln -s $IAS_IT_TEST_DATA_DIR/test_case_032/$rlut $sym_link

    # Load the RLUT
    load_rlut.sh $sym_link $IAS_DB_L8_OLITIRS |& tee -a $log

    # RLUTs should return ERROR if SUCCESS return ERROR
    if ($status == $SUCCESS) then
        printf "FAIL: rlut should not have loaded properly\n\n" |& tee -a $log
        sqlplus -S $IAS_DB_L8_OLITIRS << DEL_SCRIPT
            DELETE FROM RLUT_MAIN WHERE FILE_NAME='$rlut';
            exit 
DEL_SCRIPT

        exit $ERROR
    endif
    printf "SUCCESS: Test $rlut passed\n\n" |& tee -a $log
    
    # Remove link on success
    rm -f $sym_link

end

#------------------------NORMAL RLUT LOAD CHECKS ------------------------------#

set good_rlut = ('L8RLUT20130101_20130131v02.h5' \
                 'eval_L8RLUT20130101_20130131v02.h5')   
set exit_status = $SUCCESS

foreach rlut_filename ($good_rlut)

    # Select the correct database for the RLUT
    set db_instance = $IAS_DB_L8_OLITIRS
    if ($rlut_filename =~ eval_*) then
        set db_instance = $IAS_DB_L8_OLITIRS_EVAL
    endif

    # Ensure the test RLUT is not on the database
    set count = `sqlplus -S $db_instance` << END1
        set head off
        SELECT COUNT(1) FROM RLUT_MAIN WHERE
        FILE_NAME='$rlut_filename';
        exit
END1

    # Remove the entry if it exists
    if ($count != 0) then
        sqlplus -S $db_instance << END2
            DELETE FROM RLUT_MAIN WHERE FILE_NAME='$rlut_filename';
            exit
END2

        if ( $status ) then
            echo "Unable to remove existing RLUT from database: $rlut_filename"
            exit $ERROR
        endif
    endif

    # Ensure the file is not in cache
    rm -f $IAS_DATA_DIR/rlut/$rlut_filename
    if ( $status ) then
        echo "FAIL: unable to remove file $IAS_DATA_DIR/rlut/$rlut_filename" |&
            tee -a $log
        exit $ERROR
    endif

    # Load the RLUT
    load_rlut.sh $IAS_IT_TEST_DATA_DIR/test_case_032/$rlut_filename \
        $db_instance |& tee -a $log
    if ($status) then
        echo "FAIL: unable to load the RLUT" |& tee -a $log
        exit $ERROR
    endif

    # Request that the RLUT be cached
    set service = "$IAS_SERVICES/dataservices/rest/rlut"
    set operation = "operation=cacheRlutParmFile"
    wget -q -O RLUT "${service}?${operation}&filename=$rlut_filename"
    set service_filename = `cat RLUT`

    # The file name returned by the services should match the one loaded and
    # should now be in cache
    if ( $service_filename != $rlut_filename ) then
        echo "FAIL: services returned: $service_filename" |& tee -a $log
        echo "No clean up done" |& tee -a $log
        exit $ERROR
    else
        echo "PASS: service returned correct file name" |& tee -a $log
    endif

    # The files should be identical
    echo "Diffing files ..." |& tee -a $log
    diff $IAS_IT_TEST_DATA_DIR/test_case_032/$rlut_filename \
        $IAS_DATA_DIR/rlut/$rlut_filename |& tee -a $log
    if ($status) then
        echo "FAIL: Differences in the two files noted" |& tee -a $log
        set exit_status = $ERROR
    else
        printf "SUCCESS: Test $rlut_filename passed.\n\n" |& tee -a $log
    endif

    # Clean up if the diff was successful
    if ($exit_status == $SUCCESS) then
        rm -f $IAS_DATA_DIR/rlut/$rlut_filename RLUT ${rlut_filename:r}*
        sqlplus -S $db_instance << END3
            delete from rlut_main where file_name='$rlut_filename';
            exit
END3

        if ($status) then
            echo "Unable to remove RLUT $rlut_filename from database" |& tee -a $log
            set exit_status = $ERROR
        endif
    else
        echo "Due to errors no clean up was done" |& tee -a $log
    endif
end

if ($exit_status == $SUCCESS) then
    echo "PASS: TC_032" |& tee -a $log
else
    echo "FAIL: TC_032" |& tee -a $log
endif
exit $exit_status
