#!/usr/bin/perl -w
use strict;

###
# check_tc040_results.pl
#
# Usage: check_tc040_results.pl <test directory>
#
# The <test directory> contains configuration files listing what will be 
# verified: 
#   - report_files.txt contains a list of files generated by MTF Bridge 
#     to verify, followed by a "|", and then a string to mark the end of the
#     report header, which is ignored. If there is no "|" and subsequent text,
#     there is no report header section to ignore.  The report verification 
#     files in the test directory should be edited to remove the header 
#     information that will be skipped in the compared files.  The test 
#     filenames should match the files generated by MTF Bridge.
#
# This script is derived from verify_test_results.pl.  A separate script is
# needed for test case 040 because verify_test_results.pl is driven by Work
# Orders at several levels, and test case 040 includes some post-Work Order
# steps. Some new features of verify_test_results.pl like HDF file processing,
# template file processing, and database table processing have not been added
# to this tool, since they are not needed to test MTF Bridge.
#
###

use Cwd qw(abs_path getcwd);

# Command line parameters
my $test_dir_path;  # <test directory> - Directory containing config and 
                    #                    verfication data

# Configuration 
my $report_list_filename = "report_files.txt";

# Variables
my @report_files;
my $error_cnt = 0;
my $working_file;   # working version of file to compare
my $sed_replace_string; # substitute command to control sed operation


### Subroutine declarations
sub usage();

##### Process parameters ######################################################

# Verify all command line parameters were provided
if ($#ARGV != 0)
{
    usage();
    exit 1;
}

$test_dir_path = $ARGV[0]; # <test directory>


# Verify the test directory path exists
if ( !(-d $test_dir_path) )
{
    print "Invalid test directory path \n";
    usage();
    exit 1;
}

##### Verify Test Results #####################################################

# Load the list of report files to be verified
if ( -f "$test_dir_path/$report_list_filename" )
{
    open(FFH, "<$test_dir_path/$report_list_filename");
    while(<FFH>)
    {
        chomp;
        push(@report_files, $_);
    }
    close(FFH);
}

# If there's nothing to verify, exit the script
my $files_to_verify = scalar(@report_files);
if ($files_to_verify == 0)
{
    print "Nothing to verify. Exiting.\n";
    exit 0;
}

# Verify report files generated by MTF Bridge 
foreach my $file (@report_files)
{ 
    my ($file, $cutoff_str) = split(/\|/, $file);

    my $cmd;

    # Use sed to remove variations in work order IDs 
    if ($file eq 'mtfestimate_trendall.dat' || 
        $file eq 'output_file_1_trend_date.dat' ||
        $file eq 'output_file_2_trend_date.dat' ||
        $file eq 'output_file_3_trend_date.dat' ||
        $file eq 'output_file_both_trend_g_optics.dat')
    {
        # Make a working copy of the file. The working copy will be modified
        $working_file = "$file.compare";
        $cmd = "cp $file $working_file";
        my $rc = system($cmd);
        if ($rc != 0)
        {
            print "Copy of file $file to $working_file failed. " .
                "Exiting.\n";
            exit 1;
        }

        $sed_replace_string = "s/[[:alnum:]]*[[:space:]]" .
            "/WOID /";
        $rc = system("sed", "-i", $sed_replace_string, $working_file);
        if ($rc != 0)
        {
            print "Filtering of WO IDs from $file to $working_file failed. " .
                "Exiting.\n";
            exit 1;
        }
    }
    else
    {
        $working_file = "$file";
    }

    if (defined $cutoff_str)
    {
        # Use sed to remove information that changes with each report run (
        # time, etc)
        $cmd = "sed '/$cutoff_str/,\$!'d $working_file ";
        $cmd .= "| diff -q - $test_dir_path/$working_file";
    }
    else
    {
        # There is no report header section to remove for this file, so just
        # compare it
        $cmd = "diff -q $working_file $test_dir_path/$working_file";
    }

    print "Running: $cmd\n";
    my $rc = system($cmd);
    if ($rc != 0)
    {
        print "Diff failed: $working_file\n";
        $error_cnt++;
    }

    print "\n";
}

if ($error_cnt > 0)
{
    print "Running diffs failed\n";
    exit 1;
}
else
{
    # Test was successful.
    print "Running diffs passed\n";
    exit 0;
}

#### Subroutines ##############################################################

## Usage
sub usage()
{
    print <<USAGE
Usage: check_test_case_040.pl <test directory>
  <test directory> = Directory containing test data lists and verification data
USAGE
}

