#!/bin/csh -f
#***************************************************************************
# NAME:     create_wos.csh
#
#  PURPOSE:  Creates the work orders for test_case_042 using PWG
#  RETURNS:
#        Type:  Integer
#        Value: SUCCESS(0) / ERROR(1)
#
#  USAGE:    ./create_wos.csh
#****************************************************************************
set SUCCESS = 0
set ERROR = 1
set tc = 042
set log = tc${tc}.log
set woid_file = tc${tc}_woids.txt
set tts = "three_tirs_scenes"

# Remove old log and work order ids file
rm -f $log $woid_file

# Check if we are operating in IT or ST mode
set IS_IT_MODE = 1
if ( $1 == "ST" ) then
    set IS_IT_MODE = 0
endif

# Comment fields -- this is the order in which the following lists are
# dependent on
set comments = ( \
    "TC${tc} 01 Island Scene" \
    "TC${tc} 02 Normal angle band" \
    "TC${tc} 03 Central Meridian Crossing" \
    "TC${tc} 04 Descending 180 Degree Crossing" \
    "TC${tc} 05 North Pole" \
    "TC${tc} 06 Day to Night transition Scene" \
    "TC${tc} 07 Cal/Val Base Scene with Per-Pixel correction" \
    "TC${tc} 08 Cal/Val Base Scene with Scene Center correction" \
    )

# L0R data paths in proper order
# $IAS_IT_TEST_DATA_DIR/l0r/ must be pre-pended for them to be complete
set l0r_path_dirs = ( \
    "$tts/LC80230192013231LGN00/LC80230192013231LGN00" \
    "$tts/LC81620462013341LGN00/LC81620462013341LGN00" \
    "$tts/LC81940522013133LGN01/LC81940522013133LGN01" \
    "$tts/LC80900152013204LGN00/LC80900152013204LGN00" \
    "$tts/LC80282462013234LGN00/LC80282462013234LGN00" \
    "$tts/LC80342472013276LGN00/LC80342472013276LGN00" \
    "$tts/LC81980472013089LGN01/LC81980472013089LGN01" \
    "$tts/LC81980472013089LGN01/LC81980472013089LGN01" \
    )

# Scenes in proper order
set scene_ids = ( \
    "LC80230192013231LGN00" \
    "LC81620462013341LGN00" \
    "LC81940522013133LGN01" \
    "LC80900152013204LGN00" \
    "LC80282462013234LGN00" \
    "LC80342472013276LGN00" \
    "LC81980472013089LGN01" \
    "LC81980472013089LGN01" \
    )

# Procedures in the proper order
set procedure_ids = (\
    "L1T with Sun Correction" \
    "L1T with Sun Correction" \
    "L1T with Sun Correction" \
    "L1T with Sun Correction" \
    "L1T with Sun Correction" \
    "L1T with Sun Correction" \
    "L1T with Sun Correction" \
    "L1T with Sun Correction" \
    )

# Sun corrected output data type
set data_type = (\
    "R*4" \
    "UI*2" \
    "UI*2" \
    "R*4" \
    "UI*2" \
    "UI*2" \
    "R*4" \
    "R*4" \
    )

# Sun angle correction type
set correction_type = (\
    "PER_PIXEL"\
    "PER_PIXEL"\
    "PER_PIXEL"\
    "PER_PIXEL"\
    "PER_PIXEL"\
    "PER_PIXEL"\
    "PER_PIXEL"\
    "SCENE_CENTER"\
    )

# Insert work orders
set i = 0
set pwg_status = 0
echo "Inserting work orders" |& tee -a $log
while ( $i < ${#comments} )
    @ i++

    set additional_parms = ""
    set l0r = ""

    # Only set the l0r path for IT test cases
    if ( $IS_IT_MODE ) then
        set l0r = "-l0r_data_path $IAS_IT_TEST_DATA_DIR/l0r/$l0r_path_dirs[$i]"
    endif

    if ( $i == 2 ) then
        # Override the minimum threshold so no correction is made to the
        # OLI bands. Setting it high at 85 will cause the per-pixel correction
        # to be skipped.
        set additional_parms = ( \
            "-parm SUN_ANGLE_CORRECTION:MINIMUM_SUN_ANGLE_THRESHOLD=85.0" \
        )
    else if ( $i == 3 ) then
        # Override the sun angle correction to use only band 7 for all of the
        # band corrections. Band 8 can not be corrected with band 7 due to the
        # size difference.
        set additional_parms =( \
            '-parm SUN_ANGLE_CORRECTION:SUN_ANGLE_BAND=7' \
            '-parm "COMMON:BAND_LIST=(1,2,3,4,5,6,7,9)"' \
            )
    else if ( $i == 4 ) then
        # Override the scaling factors used to scale to reflectance
        set additional_parms =( \
            '-parm "SUN_ANGLE_CORRECTION:REFLECTANCE_MULTIPLICATION_FACTORS=(.000026,.000027,.000028,.000029,.000021,.000022,.000023,.000024,.000025)"' \
            '-parm "SUN_ANGLE_CORRECTION:REFLECTANCE_ADDITIVE_FACTORS=(.15,.14,.13,.12,.11,.19,.18,.17,.16)"' \
            '-parm SUN_ANGLE_CORRECTION:OVERRIDE_SCALING_FACTORS_FLAG=1' \
             )
    else if ( $i == 6 ) then
        # Override the minimum threshold so the work order doesn't fail
        # to process due to the minimum angle being too low and do not
        # convert the tirs bands to temperature
        set additional_parms = ( \
            "-parm SUN_ANGLE_CORRECTION:MINIMUM_SUN_ANGLE_THRESHOLD=-15.0" \
            "-parm SUN_ANGLE_CORRECTION:CONVERT_TIRS_TO_TEMP_FLAG=0" \
        )
    endif

    # Run PWG to create the work order
    set output = `PWG -procedure_id "$procedure_ids[$i]" -scene_id \
        $scene_ids[$i] -i $l0r -comment "$comments[$i]" $additional_parms \
        -parm SUN_ANGLE_CORRECTION:SUN_ANGLE_CORRECTION="$correction_type[$i]" \
        -parm SUN_ANGLE_CORRECTION:CORRECTION_OUTPUT_DATA_TYPE="$data_type[$i]"`
    if ( $status ) then
        echo $output >> $log
        echo "Error loading dataset $i" |& tee -a $log
        @ pwg_status++
    else
        echo $output >> $log
        while ( $output[1] != "Work" )
            shift output
        end
        shift output
        if ( $output[1] != "Order" ) then
            echo "Error parsing work order ID from PWG command for dataset $i"\
                |& tee -a $log
        else
            shift output
            set woid = `echo $output[1] | sed 's/,//'`
            /bin/echo -n "$woid "
            echo $woid >> $woid_file
        endif
    endif
end
echo ""

# Check the pwg status for ERROR or SUCCESS
if ( $pwg_status ) then
    echo "ERROR: $pwg_status errors creating work orders with PWG" \
        |& tee -a $log
    exit $ERROR
else
    echo "SUCCESS: created $i work orders" |& tee -a $log
endif

exit $SUCCESS
