#!/bin/csh
#
# This shell script is the validation script for IAS Test Case 050 - SCA
# Discontinuity Correction.
#
# Requirements Tested:
# --------------------
#   - IP1757 The IPS shall correct SCA discontinuity in the OLI data by applying
#     the correction algorithm defined in the Cal/Val Algorithm Description
#     Document (ADD).
#
#   - IP3940 The IPS shall correct SCA discontinuity in the TIRS data by
#     applying the correction algorithm defined in the Cal/Val Algorithm
#     Description Document (ADD).
#
# The following benchmark L1R scenes must be in the
# $IAS_IT_TEST_DATA_DIR/test_case_050/
#
#   - Benchmark scene for Antarctica with default CPF dataSet1/l1r_image.h5
#
#   - Benchmark scene for Brookings with default CPF dataSet2/l1r_image.h5
#
#   - Benchmark scene for Antarctica with a special evaluation CPF, that runs
#     the SCA discontinuity algorithm but ignores these results and applies the
#     CPF correction dataSet3/l1r_image.h5
#
#   - Benchmark scene for Brookings with a special evaluation CPF, that runs the
#     SCA discontinuity algorithm but ignores these results and applies the CPF
#     correction dataSet4/l1r_image.h5
#
#   - Benchmark scene for Antarctica with a special evaluation CPF, that does
#     not run the SCA discontinuity algorithm dataSet5/l1r_image.h5.  This image
#     should be identical to the 3rd image.
#
#   - Benchmark scene for Brookings with a special evaluation CPF, that does not
#     run the SCA discontinuity algorithm dataSet6/l1r_image.h5.  This image
#     should be identical to the 4th image.
#
################################################################################

if ($#argv != 6) then
    echo ""
    echo "usage: $0 <WOID1> <WOID2> <WOID3> <WOID4> <WOID5> <WOID6>"
    echo ""
    echo "Test validation script for IAS test case 050"
    echo ""
    exit 1
endif

# Directory where the test case benchmark images are located
set testDir = ${IAS_IT_TEST_DATA_DIR}/test_case_050
set nErrors = 0

# Compare the L1R image files for work orders 1, 2, 3, and 4 against the
# benchmark images in the $testDir directory.  The 5th and 6th work orders
# don't compare the imagery.
../tools/validate_wo_output.csh ${argv[1]} ${argv[2]} ${argv[3]} ${argv[4]} \
    ${argv[5]} ${argv[6]}
if ($status != 0) then
    @ nErrors += 1
endif

# The 5th and 6th work order images aren't compared against benchmark images.
# The 5th work order image data should be identical to what is generated by the
# 3rd work order and the 6th identical to the 4th.  The algorithmic way these
# images were created will result in a difference in the 'File Metadata' so just
# the image bands should be compared.
#
# To reduce overall disk space requirements for benchmark images, we'll manually
# compare the images within these work order directories.
echo "h5diff ${argv[5]}/l1r_image.h5 ${argv[3]}/l1r_image.h5"
set i = 1
while (${i} <= 11)
    set bandName = B`echo ${i} | awk '{printf "%02d",$1}'`
    h5diff ${IAS_WO}/${argv[5]}/l1r_image.h5 ${IAS_WO}/${argv[3]}/l1r_image.h5 ${bandName}
    if ($status != 0) then
	echo "diff failed for ${bandName}"
	@ nErrors += 1
	break
    endif
    @ i += 1
end

echo "h5diff ${argv[6]}/l1r_image.h5 ${argv[4]}/l1r_image.h5"
set i = 1
while (${i} <= 11)
    set bandName = B`echo ${i} | awk '{printf "%02d",$1}'`
    h5diff ${IAS_WO}/${argv[6]}/l1r_image.h5 ${IAS_WO}/${argv[4]}/l1r_image.h5 ${bandName}
    if ($status != 0) then
	echo "diff failed for ${bandName}"
	@ nErrors += 1
	break
    endif
    @ i += 1
end

echo ""
if ($nErrors > 0) then
    echo "FAILED with ${nErrors} errors"
    exit 1
endif

echo "PASSED"
exit 0
