#!/bin/csh -f

#***************************************************************************
# NAME:     create_wos.csh
#
#  PURPOSE:  Creates the work orders for test_case_051 using PWG
#  RETURNS:
#        Type:  Integer
#        Value: 0 / 1
#
#  USAGE:    ./create_wos.csh
#****************************************************************************
set SUCCESS = 0
set ERROR = 1

# Check if its IT or ST mode
set mode = "IT"
if ( $1 == "ST" ) then
    set mode = "ST"
endif

set tc = 051
set log = tc${tc}.log
set woid_file = tc${tc}_woids.txt

# Remove old log and work order id files
rm -f $log $woid_file

# Comment fields -- this is the order in which the following lists are
# dependent on
set comments = ( \
    "TC${tc} 01 Mixture of clouds and snow/ice using CFmask algorithm"\
    "TC${tc} 02 Mixture of clouds and snow/ice using ACCA algorithm"\
    "TC${tc} 03 Mixture of clouds and snow/ice using ATACCA algorithm"\
    "TC${tc} 04 Mixture of clouds and snow/ice using See5 algorithm"\
    "TC${tc} 05 Cloud scene with only OLI bands" \
    "TC${tc} 06 Cloud scene with only OLI band using Non-thermal CFmask algorithm" \
    "TC${tc} 07 Rolling clouds over 50% of scene" \
    "TC${tc} 08 Isolated clouds over 15% of scene with mountains/snow" \
    "TC${tc} 09 Popcorn clouds with most over land" \
    "TC${tc} 10 Clouds over 25% of scene" \
    "TC${tc} 11 Mountainous with terrian occluded pixels (Mt. Everest)" \
    "TC${tc} 12 TIRS only image using CFmask algorithm" \
    "TC${tc} 13 Ascending scene completely cloud covered" \
    "TC${tc} 14 Ascending scene completely cloud covered skip Cloud Cover due to Sun Elevation using CFmask algorithm" \
    "TC${tc} 15 Ascending scene completely cloud covered ship Cloud Cover due to Sun Elevation using ACCA algorithm" \
    "TC${tc} 16 Most of land cloud covered, with water mostly clear" \
    "TC${tc} 17 PS scene 1" \
    "TC${tc} 18 PS scene 2" \
    "TC${tc} 19 PS scene 3" \
    "TC${tc} 20 Off-nadir scene 1" \
    "TC${tc} 21 Off-nadir scene 2" \
    "TC${tc} 22 Scene that has more than 90% of cloud cover" \
    )

# L0R data paths in proper order
# $IAS_IT_TEST_DATA_DIR/l0r/ must be pre-pended for them to be complete
set tts = "three_tirs_scenes"
set ss = "single_scene"
set l0r_path_dirs = ( \
    $tts/LC80100102013124LGN01/LC80100102013124LGN01 \
    $ss/LO81620382014232LGN00/LO81620382014232LGN00 \
    $ss/LO80340342015042LGN00/LO80340342015042LGN00 \
    $tts/LC81380162014256LGN01/LC81380162014256LGN01 \
    $tts/LC81470382013124LGN01/LC81470382013124LGN01 \
    $tts/LC80170392013141LGN01/LC80170392013141LGN01 \
    $tts/LC81790552013124LGN01/LC81790552013124LGN01 \
    $tts/LC81400412013123LGN01/LC81400412013123LGN01 \
    $tts/LT81520362013127LGN01/LT81520362013127LGN01 \
    $tts/LC82121252013339LGN00/LC82121252013339LGN00 \
    $tts/LC80900152013204LGN00/LC80900152013204LGN00 \
    $tts/LC82181062015003LGN00/LC82181062015003LGN00 \
    $tts/LC81411072014069LGN00/LC81411072014069LGN00 \
    $tts/LC80831072015001LGN00/LC80831072015001LGN00 \
    $tts/LC80170372013141LGN01/LC80170372013141LGN01 \
    $tts/LC80440312013138LGN01/LC80440312013138LGN01 \
    $tts/LC81440462014202LGN00/LC81440462014202LGN00 \
    )

# Scenes in proper order
set scene_ids = ( \
    LC80100102013124LGN01 \
    LO81620382014232LGN00 \
    LO80340342015042LGN00 \
    LC81380162014256LGN01 \
    LC81470382013124LGN01 \
    LC80170392013141LGN01 \
    LC81790552013124LGN01 \
    LC81400412013123LGN01 \
    LT81520362013127LGN01 \
    LC82121252013339LGN00 \
    LC80900152013204LGN00 \
    LC82181062015003LGN00 \
    LC81411072014069LGN00 \
    LC80831072015001LGN00 \
    LC80170372013141LGN01 \
    LC80440312013138LGN01 \
    LC81440462014202LGN00 \
    )

set i = 0;
set pwg_status = 0;
set wo_count = 0;

# Insert work orders
echo "Inserting work orders" |& tee -a $log
while ( $i < ${#l0r_path_dirs} )
    @ i++

    set additional_parms = ""
    set procedure_id = "L1T with Quality Band"

    # Most tests will run only 1 iterations with Cfmask algorithm
    set iterations = (0)

    # Add different additional parameters procedure id for different scenes
    if ($i == 1) then
        # For test case 1, set the iteration for 4 times to run with
        # different cloud cover algorithm one at a time
        # (CFMASK, ACCA, ATACCA, SEE5)
        set iterations = (0 1 2 3)
    else if ($i == 10) then
        # For test case 10, loop an additional time to run with a minimum sun
        # elevation in the CPF set to not run cloud cover for the scene
        set iterations = (0 4 5)
    else if ($i == 11) then
        # For test case 11, Dump the land water mask image file
        set additional_parms = ( \
            "-parm ASSESS_CLOUD_COVER:LAND_WATER_MASK_OUTPUT_FILENAME=landwater.img" \
            )
    endif

    # Insert two orders, one with the normal CPF and the other with a CPF that
    # uses the CFmask algorithm
    foreach version ($iterations)
        @ wo_count++

        set cpf = ""
        set write_cfmask = ""
        set algorithm = ""

        # Run CFmask algorithm
        if ($version == 0) then
            set algorithm = "-parm ASSESS_CLOUD_COVER:CLOUD_COVER_ALGORITHM=CFMASK"
            set write_cfmask = "-parm ASSESS_CLOUD_COVER:CFMASK_OUTPUT_FILENAME=cfmask.img"
        # Run ACCA algorithm
        else if ($version == 1) then
            set algorithm = "-parm ASSESS_CLOUD_COVER:CLOUD_COVER_ALGORITHM=ACCA"

        # Run ATACCA algorithm
        else if ($version == 2) then
            set algorithm = "-parm ASSESS_CLOUD_COVER:CLOUD_COVER_ALGORITHM=ATACCA"
        # Run See5 algorithm
        else if ($version == 3) then
            set algorithm = "-parm ASSESS_CLOUD_COVER:CLOUD_COVER_ALGORITHM=SEE5"

        # Run CFmask algorithm with modified cpf (Mininum_Sun_Elevation = 20.0)
        else if ($version == 4) then
            set write_cfmask =  "-parm ASSESS_CLOUD_COVER:CFMASK_OUTPUT_FILENAME=cfmask.img"
            set cpf = "-parm DSW:CAL_PARM_FILENAME=$IAS_IT_TEST_DATA_DIR/test_case_051/dataset10_cfmask_sun_elev/LC08CPF_20131026_20131231_01.01"

        # Run ACCA algorithm with modified cpf (Minimum_Sun_Elevation = 20.0)
        else if ($version == 5) then
            set algorithm = "-parm ASSESS_CLOUD_COVER:CLOUD_COVER_ALGORITHM=ACCA"
            set cpf = "-parm DSW:CAL_PARM_FILENAME=$IAS_IT_TEST_DATA_DIR/test_case_051/dataset10_ACCA_sun_elev/LC08CPF_20131026_20131231_01.01"
        endif

        # Check for the mode or the environment
        if ( $mode == "IT" ) then
            set output = `PWG -procedure_id "$procedure_id" \
                -scene_id $scene_ids[$i] \
                -i -l0r_data_path $IAS_IT_TEST_DATA_DIR/l0r/$l0r_path_dirs[$i] \
                -comment "$comments[$wo_count]" $cpf $algorithm $write_cfmask $additional_parms`
            set saveStatus = $status
        else
            set output = `PWG -procedure_id "$procedure_id" \
                -scene_id $scene_ids[$i] \
                -i -comment "$comments[$wo_count]" $cpf $algorithm $write_cfmask $additional_parms`
            set saveStatus = $status
        endif

        if ( $saveStatus ) then
            echo $output >> $log
            echo "Error loading dataset $i" |& tee -a $log
            @ pwg_status++
        else
            echo $output >> $log
            while ( $output[1] != "Work" )
                shift output
            end
            shift output
            if ( $output[1] != "Order" ) then
                echo "Error parsing work order ID from PWG command for dataset $i"\
                    |& tee -a $log
            else
                shift output
                set woid = `echo $output[1] | sed 's/,//'`
                /bin/echo -n "$woid "
                echo $woid >> $woid_file
            endif
        endif
    end
end
echo ""

# Check the pwg status for ERROR or SUCCESS
if ( $pwg_status ) then
    echo "ERROR: $pwg_status errors creating work orders with PWG" \
        |& tee -a $log
    exit $ERROR
else
    echo "SUCCESS: created $wo_count work orders" |& tee -a $log
endif
exit $SUCCESS
