#!/bin/csh -f

#*************************************************************************
# NAME:    final_cleanup.csh
#
# PURPOSE: Removes all characterization data in the IAS database for
#          the specified work order IDs, and removes any
#          reference data files renamed with the specified acquisition year.

# USAGE:   final_cleanup.csh
#
# RETURNS: Integer status of 0 if script completes successfully;
#          1 if an error occurs
#*************************************************************************

# Initialize the directory containing the reference files with the original
# acquisition year.  These were saved in 'update_benchmark_filenames.csh' in
# the event that the acquisition year had to change to avoid potential
# issues with other data.
set working_directory = `pwd`
set bpf_prefix = "LO8BPF"

# Initialize the array that will contain the work order ID strings.
set idsWO = ( `cat "cleanup.ids" | tr '\n' ' '` )
set idsLM  = ( `cat "cleanup.lmdids" | tr '\n' ' '` )
set idsRWO = ( `cat "cleanup.rwoids" | tr '\n' ' '` )

# Remove any previous instance of the cleanup log file.
rm -f cleanup.log

# Initialize the error status to indicate no errors (since we haven't
# attempted any deletion yet).
set errorStatus = 0

# For each work order, delete its entry from the
# RADIOMETRIC_WORK_ORDER_COMMON table.  As the database is currently
# implemented, this will trigger a "cascade" delete of the characterization
# data in ALL tables for the associated radiometric_wo_common_id value.
# If this database behavior changes, the logic here will need to be
# changed accordingly.
echo "Cleaning up radiometric characterization data..."
foreach id ( ${idsWO} )
    # Execute the delete for the current work order ID.
    sqlplus -s  ${IAS_DB_L8_OLITIRS} >>& cleanup.log << END
        set head off;
        DELETE FROM RADIOMETRIC_WORK_ORDER_COMMON where work_order_id='${id}';
        exit;
END
    # Display the error message if the delete operation fails.
    if ( ${status} ) then
        echo "ERROR: ${sql}"
        set errorStatus = 1
    endif
end

# Clean up the mission data.  This has to be done separately, using the
# mission data IDs generated earlier. On test server, the INVENTORY username
# has to be used and on development server, the normal user name can be used 
echo "Cleaning up mission data..."
set db_user=`echo $IAS_DB_COM | awk -F'/' '{print $1}'`
set test_server_user = 'ias_ops'
if ($db_user == $test_server_user) then
    set DB=`echo $IAS_DB_L8_OLITIRS | sed -e 's/.*\//INVENTORY\//'`
else
    set DB = `echo $IAS_DB_L8_OLITIRS`
endif

foreach lmid ( ${idsLM} )
    # Execute the delete for the current mission ID.
    sqlplus -s ${DB} >>& cleanup.log << END
        set head off;
        DELETE FROM LMD_MISSION where lmd_mission_id=${lmid};
        exit;
END
    # Display the error message if the delete operation fails.
    if ( ${status} ) then
        echo "ERROR: ${sql}"
        set errorStatus = 1
    endif
end

# If an error occurred, indicate to the user to check the final cleanup log
# file for additional information.
if ( ${errorStatus} ) then
    echo ""
    echo "Errors occurred.  See cleanup.log for more information."
    exit 1
endif

# Remove all of the log files and cleanup files--they're no longer needed.
rm -f *.log
rm -f cleanup.*

# Display messages indicating successful deletion of data for each
# work order.
foreach id ( ${idsWO} )
    echo "Data deletion for work order ID ${id} successfully completed."
end

# Done
exit 0
