#!/bin/csh -f
#
#**************************************************************************
# NAME:     load_statistics.csh
#
# PURPOSE:  Populates the database with shutter collect characterization
#           results to be used in Test Case 052 (Bias Model Calibration).
#
# RETURNS:  Integer status code of 0 if script executes successfully,
#           1 if an error occurs.
#
# NOTES:    Test case 052 should be run first, when the database is
#           "clean".
#**************************************************************************

# Establish the location of the data files that will be used to create
# the required loader files for each set of collects.
set datasets = ""
set maxNumDatasets = 5
set templateRoot = $IAS_IT_TEST_DATA_DIR/test_case_052/templates

# Set variables to be used to keep track of the radiometric work order
# common and work order ID strings.
set idsLM  = ""
set idsRWO = ""
set idsWO  = ""

# There are no command line arguments for this script.  All data sets are
# to be processed.
set i = 0
while ( $i < ${maxNumDatasets} )
    @ i++
    set datasets = ( ${datasets} ${i} )
end
echo "Loading data sets:  ${datasets}"
echo ""


# Set the acquisition year.
set acquisition_year = 2001

# Update the list of files that should be cleaned when this script terminates
set cleanFiles = ""
set cleanFiles = (${cleanFiles} tmpFrameAverage.ldr tmpFrameAverage.log)
set cleanFiles = (${cleanFiles} tmpHistoStats.ldr   tmpHistoStats.log)
set cleanFiles = (${cleanFiles} tmpLmdMission.ldr   tmpLmdMission.log)
set cleanFiles = (${cleanFiles} tmpRadiometric.ldr  tmpRadiometric.log)
rm -f ${cleanFiles}

# Copy the SQL loader templates into a temporary local loader file.  Use
# the first argument to get the correct set of template files based on
# the sensor name.
echo "Processing $#datasets collects..."
cp ${templateRoot}/loader_templates/frame_averages.tpl tmpFrameAverage.ldr
cp ${templateRoot}/loader_templates/histo_stats.tpl    tmpHistoStats.ldr
cp ${templateRoot}/loader_templates/lmd_mission.tpl    tmpLmdMission.ldr
cp ${templateRoot}/loader_templates/rad_common.tpl     tmpRadiometric.ldr

# To ensure that we get the same results when this test is run we need to
# ensure that no other data that may be in the user's database, overlaps with
# this test data.  Query the database for the largest YEAR that currently
# exists and then we'll replace the years within our loader files with YEAR+1
# where appropriate.  The loader files use YYYY to indicate which year value
# should be used.
set tableLmd = "LMD_MISSION"
set tableRwoc = "RADIOMETRIC_WORK_ORDER_COMMON"
set YYYY = 0
foreach sql ( \
    "SELECT MAX(EXTRACT(YEAR FROM DATE_ACQUIRED)) AS YEAR FROM ${tableLmd};" \
    "SELECT MAX(EXTRACT(YEAR FROM DATE_ENTERED)) AS YEAR FROM ${tableLmd};" \
    "SELECT MAX(EXTRACT(YEAR FROM DATE_ACQUIRED)) AS YEAR FROM ${tableRwoc};" \
    "SELECT MAX(EXTRACT(YEAR FROM DATE_PROCESSED)) AS YEAR FROM ${tableRwoc};")
    set year = `echo ${sql} | sqlplus ${IAS_DB_L8_OLITIRS} | \
        grep -A 2 YEAR | tail --lines=1`

    # Protect against an empty table.  Only examine the retrieved year as a
    # value if we didn't get an empty string back.
    if ( "${year}" != "" ) then
        if ( ${year} > ${YYYY} ) then
            set YYYY = ${year}
        endif
    endif
end

# If we haven't received a year from the tables, or if the maximum year
# returned is less than our established acquisition year, we'll use our
# acquisition year.  Otherwise, increment the year by 1 so our data does
# not get intermingled with other data that might be in the database.
if ( ${YYYY} < ${acquisition_year} ) then
    set YYYY = ${acquisition_year}
else
    @ YYYY++
endif

# Notify the user of the year value that we are using, and save it.
echo "Populating test data for year ${YYYY}"
echo ""
echo "${YYYY}" >> cleanup.year

# Walk through each of the data sets that we are loading and create a loader
# file that will be sent to sqlldr
set i = 1
set lmdId = 0
while (${i} <= $#datasets)
    set dataSet = `echo $datasets[$i] | awk '{printf "%02d",$1}'`
    echo "Preparing data set ${dataSet}"
    @ i++

    # Get the next radiometric work order common ID
    set sql = "select dbseq_radiometric_wo_common_pk.nextval from dual;"
    set rwoId = `echo ${sql} | sqlplus ${IAS_DB_L8_OLITIRS} | \
        grep -A 2 NEXTVAL | tail --lines=1`
    if ( ${status} ) then
        echo "Error retrieving radiometric work order common ID"
        rm -f ${cleanFiles}
        exit 1
    endif
    set idsRWO = ( ${idsRWO} ${rwoId} )

    # Get the next work order ID
    set sql = "select wo_seq.nextval from dual;"
    set woId = `echo ${sql} | sqlplus ${IAS_DB_COM} | \
        grep -A 2 NEXTVAL | tail --lines=1`
    if (${status}) then
        echo "Error retrieving work order ID string"
        rm -f ${cleanFiles}
        exit 1
    endif

    # The along-track histogram statistics and cross-track frame average
    # statistics are normally calculated during ingest processing of an L0R
    # shutter collect.  Create a simulated ingest work order ID string that
    # will be inserted into the RADIOMETRIC_WORK_ORDER_COMMON table.
    set woId = "ING${woId}"
    set idsWO = ( ${idsWO} ${woId} )

    # Get the next LMD mission ID
    @ lmdId = ${i} + 20
    set idsLM = ( ${idsLM} ${lmdId} )

    # Copy the template files for the current data set, translate the necessary
    # IDs, and append that data to the end of the loader files
    set file = ${templateRoot}/rad_common/com_rwoid${dataSet}.tpl
    sed "s/__RWOID${dataSet}__/${rwoId}/" ${file} >> tmpRadiometric.ldr
    sed -i "s/__WOID${dataSet}__/${woId}/" tmpRadiometric.ldr
    set file = ${templateRoot}/frame_averages/fa_rwoid${dataSet}.tpl
    sed "s/__RWOID${dataSet}__/${rwoId}/" ${file} >> tmpFrameAverage.ldr

    set file = ${templateRoot}/histogram_statistics/hs_rwoid${dataSet}.tpl
    sed "s/__RWOID${dataSet}__/${rwoId}/" ${file} >> tmpHistoStats.ldr
    set file = ${templateRoot}/mission_data/lmd_${dataSet}.tpl
    sed "s/__LMDID${dataSet}__/${lmdId}/" ${file} >> tmpLmdMission.ldr
    sed -i "s/__MISSION_DATA_LOCATION__/RWOID=${rwoId}/" tmpLmdMission.ldr
end
echo ""

# Replace the YYYY designation in the radiometric work order common and LMD
# mission loader files with the year that we determined above
sed -i "s/YYYY/${YYYY}/g" tmpRadiometric.ldr
sed -i "s/YYYY/${YYYY}/g" tmpLmdMission.ldr

# Save a copy of the work order ID strings so that they can be used by
# the final_cleanup.csh script to remove the data that is to be loaded.  Since
# these IDs are obtained from sequences and they can't be used again, we'll
# store these IDs even before the loader scripts are run to ensure that
# partial data can be removed if an error is encountered during loading.
echo ${idsWO} >> cleanup.ids
echo ${idsRWO} >> cleanup.rwoids
echo ${idsLM} >> cleanup.lmdids

# Load all of the data we just generated into the database for testing
echo "Loading RADIOMETRIC_WORK_ORDER_COMMON table"
sqlldr ${IAS_DB_L8_OLITIRS} tmpRadiometric.ldr >& /dev/null
if ($status) then
    echo "ERROR loading RADIOMETRIC_WORK_ORDER_COMMON table"
    ./final_cleanup.csh
    exit 1
endif

echo "Loading BIAS_FRAME_AVERAGES table"
sqlldr ${IAS_DB_L8_OLITIRS} tmpFrameAverage.ldr >& /dev/null
if ($status) then
    echo "ERROR loading BIAS_FRAME_AVERAGES table"
    ./final_cleanup.csh
    exit 1
endif

echo "Loading HISTOGRAM_STATISTICS table"
sqlldr ${IAS_DB_L8_OLITIRS} tmpHistoStats.ldr >& /dev/null
if ($status) then
    echo "ERROR loading HISTOGRAM_STATISTICS table"
    ./final_cleanup.csh
    exit 1
endif

# On test server, the INVENTORY username has to be used and on development 
# server, the normal user name can be used 
echo "Loading LMD_MISSION table"
set db_user=`echo $IAS_DB_COM | awk -F'/' '{print $1}'`
set test_server_user = 'ias_ops'
if ($db_user == $test_server_user) then
    set DB=`echo $IAS_DB_L8_OLITIRS | sed -e 's/.*\//INVENTORY\//'`
else
    set DB = `echo $IAS_DB_L8_OLITIRS`
endif
sqlldr ${DB} tmpLmdMission.ldr >& /dev/null
if ($status) then
    echo "ERROR loading LMD_MMISSION table"
    ./final_cleanup.csh
    exit 1
endif
echo ""

echo "-------------------------------------------"

echo ""
echo "Data successfully loaded."
echo ""
echo "  RWOID = ${idsRWO}"
echo "  WOID  = ${idsWO}"
echo "  LMDID = ${idsLM}"
echo ""

# Final cleanup and exit.
rm -f ${cleanFiles}
exit 0


USAGE:
    echo ""
    echo "load_statistics.csh <sensor_name> [-all] <n1> <n2> ..."
    echo ""
    echo "    This script will populate the database with a set of shutter"
    echo "    collects and histogram statistics for testing the bias model"
    echo "    calibration application.  The user may specify any number of"
    echo "    data set values from 1 to $maxNumDatasets on the command line"
    echo "    or the -all command option to populate all test data sets."
    echo
    echo
    exit 1
