#!/bin/tcsh -f

#*************************************************************************
# NAME:      run_verification_test.csh
#
# PURPOSE:   Creates a directory under $IAS_WO for the BPF and residual
#            model output files, and runs the PERL verification script
#            "verify_test_results.pl".  This script is part of the
#            execution sequence for test case 052 (Bias Model Calibration).
#
# RETURNS:   0 if processing successfully completes; 1 if there is an
#            error.
#
# NOTES:     This script should be run as part of the test case 052
#            processing sequence, from the $IAS_IT_SCRIPTS_DIR/test_case_052
#            directory.  The processing sequence should be set as follows:
#
#            1.  ./load_statistics.csh
#
#            2.  $IAS_BIN/bias_model_calibration -end <acq. date>
#                                                -prev
#                                                <no. of previous
#                                                shutter collects>
#
#            3.  ./run_verification_test.csh
#                               <work_order_ID_string> <benchmark directory>
#*************************************************************************

# Check the number of command line arguments.  If we don't have the right
# number, then print a USAGE message and quit.
if ( $#argv != 2 ) then
    echo "USAGE:  ./run_verification_test.csh <work_order_id_string> "
    echo "                                    <benchmark directory>"
    exit 1
endif

# Save the directory from location from which this script is running.
set script_directory = `pwd`

# Set the common prefix for the BPF file and the residual data files.
set bpf_prefix = "LO8BPF"

# Create the work order directory and put the data files generated from
# the bias model calibration app there.  The -p in the mkdir command ensures
# there won't be an error if the directory already exists.  If any other
# error occurs in creating the directory, or an error occurs in transferring
# the test-generated output files to the directory location, inform the
# user accordingly and quit.
set wo_directory = "$IAS_WO/${1}"
mkdir -p ${wo_directory}
if ( $status ) then
    echo "Error creating work order directory ${1}"
    exit 1
endif
mv -f ${bpf_prefix}* ${wo_directory}
if ( $status ) then
    echo "Error moving test output files to work order directory ${1}"
    exit 1
endif

# Create output file names that match the names of the verification files
# in the test data directory, and use symbolic links to "rename" the output
# files to match the verification file names.  The symbolic link is of the
# form `ln -s target_name link_name`, where the target name is the original
# output file name, and the link name is the name of the corresponding 
# verification file in the test data directory.
cd ${wo_directory}
set real_bpf_name = `ls ${bpf_prefix}*.01`
set real_a0_name  = `ls ${bpf_prefix}*-a0.txt`
set real_a1_494_name = `ls ${bpf_prefix}*-a1-494.txt`
set real_a1_988_name = `ls ${bpf_prefix}*-a1-988.txt`

set match = "expected"
set match_bpf_name = "${match}_bpf"
set match_a0_name  = "${match}_a0"
set match_a1_494_name = "${match}_a1_494"
set match_a1_988_name = "${match}_a1_988"

ln -s ${real_bpf_name} ${match_bpf_name}
ln -s ${real_a0_name} ${match_a0_name}
ln -s ${real_a1_494_name} ${match_a1_494_name}
ln -s ${real_a1_988_name} ${match_a1_988_name}

# Run the verification script.
cd ${script_directory}
set errorStatus = 0
../tools/verify_test_results.pl ${1} ${2} $IAS_DB_L8_OLITIRS
if ( $status ) then
    echo "Error running verification script"
    set errorStatus = 1
endif

# Remove the symbolic links.
rm -f ${wo_directory}/${match}*

# Done with this script.  Check to see whether verification could be
# completed successfully and inform the user accordingly.
if ( $errorStatus == 1 ) then
    echo "Bias model calibration integration test verification failed."
else
    echo "Bias model calibration integration test verification completed."
endif
exit ${errorStatus}
