#!/bin/csh -f
#
# Name:     run_verification_test.csh
#
# Purpose:  Executes the Key Performance Requirements (KPR) 5,6,7 algorithms
#           and verifies output of those algorithms matches expected results.
#           This script is part of the execution sequence for test case 056
#           (Geometric KPR Reporting).
#
# Returns:  0 if processing successfully completes; 1 if there is an error
#
# Notes:    This script should be run as part of the test case 056
#           processing sequence, from the $IAS_IT_SCRIPTS_DIR/test_case_056
#           directory.
#
# Usage:    ./run_verification_test.csh
#---------------------------------------------------------------------------

# Setup arrays to run the tests with a loop
set test_cases = ( \
    "TC056 default KPR #5" \
    "TC056 KPR #5 edge inoperable detectors" \
    "TC056 default KPR #6" \
    "TC056 KPR #6 internal inoperable detectors" \
    "TC056 default KPR #7" \
    "TC056 KPR #7 Cirrus report" \
    "TC056 KPR #7 invalid report" )
set test_scripts = ( \
    "kpr5_evaluate" \
    "kpr5_evaluate" \
    "kpr6_evaluate" \
    "kpr6_evaluate" \
    "kpr7_evaluate" \
    "kpr7_evaluate" \
    "kpr7_evaluate" )
set cpf_files = ( \
    "L8CPF20130401_20130630.06" \
    "$IAS_IT_TEST_DATA_DIR/test_case_056/tc056_L8CPF20130401_20130630.01" \
    "L8CPF20130401_20130630.06" \
    "$IAS_IT_TEST_DATA_DIR/test_case_056/tc056_L8CPF20130401_20130630.02" \
    "" \
    "" \
    "" )
set params = ( \
    "" \
    "" \
    "" \
    "" \
    "$IAS_IT_TEST_DATA_DIR/test_case_056/b2b_accuracy.rpt 30" \
    "$IAS_IT_TEST_DATA_DIR/test_case_056/b2b_accuracy_cirrus.rpt 30" \
    "$IAS_IT_TEST_DATA_DIR/test_case_056/b2b_accuracy_eval.rpt 30" )
set report_files = ( \
    "KPR_5_Default_Report.txt" \
    "KPR_5_Eval_Report.txt" \
    "KPR_6_Default_Report.txt" \
    "KPR_6_Eval_Report.txt" \
    "KPR_7_Default_Formal_Report.txt KPR_7_Default_CalVal_Report.txt" \
    "KPR_7_Cirrus_Formal_Report.txt KPR_7_Cirrus_CalVal_Report.txt" \
    "KPR_7_Eval_Formal_Report.txt KPR_7_Eval_CalVal_Report.txt" )
set kpr7_report_files = ( \
    "KPR_7_Default_Formal_Report.txt" \
    "KPR_7_Default_CalVal_Report.txt" \
    "KPR_7_Cirrus_Formal_Report.txt" \
    "KPR_7_Cirrus_CalVal_Report.txt" )

# Clean up report files left from a previous run
set i = 1
while ( $i < 8 )
    rm -f $report_files[$i]
    @ i++
end

set logfile = kpr_testing.log
rm -f $logfile

# Download the CPF from IAS_SERVICES.
set cpf_file = $cpf_files[1];
echo "Retrieving CPF..."
set wget_command = 'wget -q -O L8CPF20130401_20130630.06 "$IAS_SERVICES/dataservices/rest/cpf?operation=getCalParmFile&filename=L8CPF20130401_20130630.06"'
wget -q -O $cpf_file "$IAS_SERVICES/dataservices/rest/cpf?operation=getCalParmFile&filename=$cpf_file"
if ($status != 0) then
   echo "Unable to download CPF from the IAS_SERVICES : $wget_command"
   echo "Unable to download CPF from the IAS_SERVICES : $wget_command" >> $logfile
   exit 1
endif

# Execute each of the evaluation reporting scripts with defaults and
# evaluative inputs
set i = 1
while ( $i < 8 )
    echo "" >> $logfile

    set cpf_file = ""
    if ( $cpf_files[$i] != "" ) then
        set cpf_file = "$cpf_files[$i]"
    endif

    set cmd = "$test_scripts[$i] $cpf_file $params[$i] $report_files[$i]"

    echo "Executing: $cmd" >> $logfile
    $cmd >>& $logfile
    set ret = $status

    if ( ($i == 7) && ($ret == 1) || ($i != 7) && ($ret == 0)) then
        echo "Successfully executed: $test_cases[$i]"
    else
        echo "Command Line: $cmd"
        echo "FAIL: Executing $test_cases[$i]"
        # No point in continuing if any tests fail
        exit 1
    endif

    @ i++
end

# Variable to count failures
set failures = 0

# Check the reports against saved baseline manually validated previously
# KPR 5 & 6 results can be verified very similarly
set testdir = $IAS_IT_TEST_DATA_DIR/test_case_056
set i = 1
while ( $i < 5 )
    diff -q $report_files[$i] $testdir/$report_files[$i] >& /dev/null
    if ( $status != 0 ) then
        echo ""
        echo "Files differ: $report_files[$i] $testdir/$report_files[$i]"
        echo ""
        @ failures++
        echo "Fail: $test_cases[$i]"
    else
        echo "Pass: $test_cases[$i]"
    endif

    @ i++
end

# The KPR 7 tests generate two reports, which both have to match for the
# test to pass
# Verify the KPR 7 default reports
set pass = 1

diff -q $kpr7_report_files[1] $testdir/$kpr7_report_files[1] >& /dev/null
if ( $status != 0 ) then
    echo "Files differ: $kpr7_report_files[1] $testdir/$kpr7_report_files[1]"
    set pass = 0
endif

diff -q $kpr7_report_files[2] $testdir/$kpr7_report_files[2] >& /dev/null
if ( $status != 0 ) then
    echo "Files differ: $kpr7_report_files[2] $testdir/$kpr7_report_files[2]"
    set pass = 0
endif

if ( $pass == 0 ) then
    @ failures++
    echo "Fail: $test_cases[5]"
else
    echo "Pass: $test_cases[5]"
endif

# Verify the KPR 7 cirrus reports
set pass = 1

diff -q $kpr7_report_files[3] $testdir/$kpr7_report_files[3] >& /dev/null
if ( $status != 0 ) then
    echo "Files differ: $kpr7_report_files[3] $testdir/$kpr7_report_files[3]"
    set pass = 0
endif

diff -q $kpr7_report_files[4] $testdir/$kpr7_report_files[4] >& /dev/null
if ( $status != 0 ) then
    echo "Files differ: $kpr7_report_files[4] $testdir/$kpr7_report_files[4]"
    set pass = 0
endif

if ( $pass == 0 ) then
    @ failures++
    echo "Fail: $test_cases[6]"
else
    echo "Pass: $test_cases[6]"
endif

# If we got here, this one passed
echo "Pass: $test_cases[7]"

set exit_status = 0
echo ""
if ( $failures != 0 ) then
    echo "FAIL: $failures failure(s) occurred"
    set exit_status = 1
else
    echo "PASS: TC056 Geometric KPR Reporting"
    # Clean up
    set i = 1
    while ( $i < 8 )
        rm -f $report_files[$i]
        @ i++
    end
    # Clean up KPR 7
    set i = 1
    while ( $i < 5 )
        rm -f $kpr7_report_files[$i] $kpr7_report_files[$i]
        @ i++
    end
endif

echo ""

exit $exit_status

