#!/bin/csh -f
# ============================================================================
# Program: create_wos.csh
#
# Description: Inserts work orders into the system
#
# Input Parameters: If ST mode needs to be used "ST" can be passed as parameter
#                   to enable ST mode.
#
# Return Values:
#     0 - SUCCESS
#     1 - ERROR
#
# ============================================================================

# Return Values
set SUCCESS = 0
set ERROR = 1

# Set up file variables
set log = tc057.log
set wo_id_file = "tc057_wo_ids.txt"
set current_dir = `pwd`
set ss = "single_scene"

# Remove old ID file
rm -f $wo_id_file $log

# Check if ST mode was passed as argument
set IS_IT_MODE = 1
if ($1 == "ST") then
    set IS_IT_MODE = 0
endif

# Setup the arrays for what to process
# All arrays must contain an equal number of items
# Comments array for PWG
set comments = ( \
    "TC057 1 - Dataset 01 - GCPLIB" \
    "TC057 2 - Dataset 02 - Evenly Spaced" \
    "TC057 3 - Dataset 03 - Moravec" \
    "TC057 4 - Dataset 04 - Moravec and Evenly Spaced" \
    "TC057 5 - Dataset 05 - ENVI ROI" )

# Generation Method used
set method = ( \
    'GCPLIB' \
    'EVENLY_SPACED' \
    'MORAVEC' \
    'MORAVEC_AND_EVENLY_SPACED' \
    'ENVI_ROI' )

# Souce of the GCPs
set gcp_source = ( \
    'GLS' \
    'GLS' \
    'DOQ' \
    'DOQ' \
    'TM6' )

# Exclude Clouds parameters 0 through 3
set exclude_cloud = ( \
    '0' \
    '0' \
    '3' \
    '0' \
    '0' )

# Exclude Cirrus Clouds parameters 0 through 3
set exclude_cirrus = ( \
    '0' \
    '0' \
    '0' \
    '3' \
    '0' )

# Output intermediate data
set intermediate = ( \
    '0' \
    '0' \
    '1' \
    '1' \
    '0' )

# Min number of control GCPs
# Set Min to higher than number of GCPs to set all to control for first WO.
set min_gcps = ( \
    '950' \
    '0' \
    '100' \
    '115' \
    '100' )

# Max number of control GCPs
# Lower number of max GCPs to test that Max holds WO #4 only produces 40 GCPs
set max_gcps = ( \
    '1000' \
    '0' \
    '200' \
    '40' \
    '1000')

# Starting IDs for the GCPs
set starting_id = ( \
    '999' \
    '-1' \
    '0' \
    '999' \
    '888' )

# Scene id array used in PWG
set scene_ids = ( \
    LC80400302013078LGN01 \
    LC81630442013204LGN00 \
    LC81650342013170LGN00 \
    LC80230192013231LGN00 \
    LC80400302013078LGN01)

# L0R relative path
set l0r_path_dirs = ( \
    $ss/LC80400302013078LGN00/LC80400302013078LGN01 \
    $ss/LC81630442013204LGN00/LC81630442013204LGN00 \
    $ss/LC81650342013170LGN00/LC81650342013170LGN00 \
    $ss/LC80230192013231LGN00/LC80230192013231LGN00 \
    $ss/LC80400302013078LGN00/LC80400302013078LGN01)

# Procedure to use
set procedures = ( \
    'Generate Chips' \
    'Generate Chips' \
    'Generate Chips' \
    'Generate Chips' \
    'Generate Chips')

# Procedure to use
set resolution = ( \
    'FEWEST_POINTS' \
    'NOMINAL' \
    'NOMINAL' \
    'NOMINAL' \
    'NOMINAL' )

# Use water mask flags
set water_mask_flag = ( \
    "1" \
    "1" \
    "1" \
    "2" \
    "0" )

# Setup loop variables
set index = 0
set pwg_status = 0

# Loop through arrays creating WOs
echo "Inserting Work Orders" |& tee -a $log
while ( $index < ${#comments} )
    @ index++

    set add_parms = ""
    set min_dist = ""

    # If in IT testing use the L0R path
    if ( $IS_IT_MODE ) then
        set add_parms = "$add_parms -l0r_data_path $IAS_IT_TEST_DATA_DIR/l0r/$l0r_path_dirs[$index]"
    endif

    # Setting additional parameters
    if ( $index == 1 ) then
        # For the first work order, skip creating the quality band to test the
        # generate_chips option to work without the quality band
        set add_parms = \
            "$add_parms -script TERRAIN_OCCLUSION:SKIP -script ASSESS_CLOUD_COVER:SKIP"
    else if ( $index == 2 ) then
        set add_parms = "$add_parms -script SELECT_CONTROL_GCPS:SKIP"
    else if ( $index == 4 ) then
        set min_dist = '-parm "GENERATE_CHIPS:MINIMUM_GCP_DISTANCE=(2,2)"'
        set add_parms = "$add_parms -parm GENERATE_CHIPS:WATER_MASK_FILENAME=$IAS_IT_TEST_DATA_DIR/test_case_057/data/water_mask.img"
    else if ( $index == 5 ) then
        set add_parms = "$add_parms -parm GENERATE_CHIPS:ENVI_ROI_FILENAME=$IAS_IT_TEST_DATA_DIR/test_case_057/data/envi_roi.txt"
    endif

    # Add parameter override code here if needed
    # May also need a parameter override list above to help with things
    set output = `PWG \
        -interactive \
        -procedure_id "$procedures[$index]" \
        -scene_id $scene_ids[$index] \
        $add_parms \
        -parm "DGC:GCP_TYPE=BOTH" \
        -parm "GENERATE_CHIPS:GENERATION_METHOD="$method[$index] \
        -parm "GENERATE_CHIPS:GCP_SOURCE="$gcp_source[$index] \
        -parm "GENERATE_CHIPS:EXCLUDE_CLOUD="$exclude_cloud[$index] \
        -parm "GENERATE_CHIPS:EXCLUDE_CIRRUS="$exclude_cirrus[$index] \
        -parm "GENERATE_CHIPS:OUTPUT_INTERMEDIATE_DATA="$intermediate[$index] \
        -parm "GENERATE_CHIPS:INTERMEDIATE_OUTPUT_DIRECTORY=intermediate" \
        -parm "GENERATE_CHIPS:STARTING_ID="$starting_id[$index] \
        -parm "SELECT_CONTROL_GCPS:MINIMUM_CONTROL="$min_gcps[$index] \
        -parm "SELECT_CONTROL_GCPS:MAXIMUM_CONTROL="$max_gcps[$index] \
        -parm "GENERATE_CHIPS:MORAVEC_SOURCE_RESOLUTION="$resolution[$index] \
        -parm "GENERATE_CHIPS:USE_WATER_MASK_FLAG="$water_mask_flag[$index] \
        $min_dist \
        -comment "$comments[$index]"`

    # Error thrown by PWG
    if ( $status ) then
        echo $output >> $log
        echo "Error Loading Dataset $index" |& tee -a $log
        @ pwg_status++

    # PWG ran successfully
    else
        echo $output >> $log

        # Remove all info before Work
        while ( $output[1] != "Work" )
            shift output
        end

        # Remove Work from output
        shift output

        # Check that next word is Order
        if ( $output[1] != "Order" ) then
            echo "Error parsing work order ID from PWG command for" \
                " dataset $index" |& tee -a $log
            @ pwg_status++
        # Shift off Order and we have the WO ID
        else
            shift output
            set wo_id = `echo $output[1] | sed 's/,//'`
            /bin/echo -n "$wo_id "

            # Put ID into ID file and
            echo $wo_id >> $wo_id_file
        endif
    endif
end
echo ""

# Write out how many errors occured or SUCCESS
if ( $pwg_status ) then
    echo "ERROR: $pwg_status errors creating work orders with PWG" \
        |& tee -a $log
    exit $ERROR
else
    echo "SUCCESS: Created $index Work Orders" |& tee -a $log
endif

exit $SUCCESS
