#!/bin/csh -f
################################################################################
# NAME: create_wos.csh
#
# PURPOSE: Creates the work orders necessary to test the create gps work order
#   creation script
#
# RETURNS:
#        Type:  Integer
#        Value: 0 / 1
#
#  USAGE:    ./create_wos.csh
################################################################################

set tc = 066
set log = tc${tc}.log
set base_cmd = $IAS_BIN/create_gps_work_orders.pl

echo "Creating work orders" |& tee $log

# These work orders should match an expected scene file, each have the TIRS
# Alignment procedure id, and each should have a defined CPF
echo "Creating work orders for group 1" |& tee -a $log
set wo_id_file = "group1_out_woids.tmp"
set cpf_name = "L8CPF20000401_20000630.01"
$base_cmd \
    --scene_list_file="group1_input_scenes.txt" \
    --procedure_id="TIRS Alignment" \
    --cal_parm_filename="$cpf_name"\
    --output_scene_list_filename="group1_out_scenes.tmp" \
    --output_woid_filename="$wo_id_file" |& tee -a $log
if ( $status != 0 ) then
    echo "Error while creating first set of work orders" |& tee -a $log
    exit 1
endif

# These work orders are testing the date range, cloud cover, and roll angle
# parameters for the path rows in group2_input_path_rows.txt. 3 scenes should
# be grabbed for this run
echo "Creating work orders for group 2" |& tee -a $log
set wo_id_file = "group2_out_woids.tmp"
$base_cmd \
    --path_row_file="group2_input_path_rows.txt" \
    --acq_date_range="03-30-2013:04-09-2013" \
    --max_cloud_cover=100 \
    --max_roll_angle=20.0\
    --output_woid_filename="$wo_id_file" |& tee -a $log
if ( $status != 0 ) then
    echo "Error while creating second set of work orders" |& tee -a $log
    exit 1
endif

# The same work orders are used for group 3 as group 2, but we are testing
# the scripts ability to exclude 2 of them.
echo "Creating work orders for group 3" |& tee -a $log
set wo_id_file = "group3_out_woids.tmp"
$base_cmd \
    --path_row_file="group2_input_path_rows.txt" \
    --acq_date_range="03-30-2013:04-09-2013" \
    --max_cloud_cover=20 \
    --max_roll_angle=5.0\
    --output_woid_filename="$wo_id_file" |& tee -a $log
if ( $status != 0 ) then
    echo "Error while creating third set of work orders" |& tee -a $log
    exit 1
endif

# Testing the skip_work_orders flag. No work orders should generate from this
# execution, even though scenes were found.
echo "Creating work orders for group 4" |& tee -a $log
set wo_id_file = "group4_out_woids.tmp"
$base_cmd \
    --path_row_file="group4_input_path_rows.txt" \
    --acq_date_range="01-01-2014:01-01-2015" \
    --skip_work_orders \
    --output_woid_filename=$wo_id_file \
    --output_scene_list_filename="group4_out_scenes.tmp" |& tee -a $log
if ( $status != 0 ) then
    echo "Error while creating fourth set of work orders" |& tee -a $log
    exit 1
endif

# Final
# The single work order created by this execution will be used for validating
# the output of the TIRS SSM Calibration procedure.
echo "Creating work orders for group 5" |& tee -a $log
set wo_id_file = tc${tc}_wo_ids.txt
$base_cmd \
    --scene_list_file="group5_input_scenes.txt" \
    --output_woid_filename=$wo_id_file |& tee -a $log
if ( $status != 0 ) then
    echo "Error while creating fifth set of work orders" |& tee -a $log
    exit 1
endif

# In order for the final work order to be used for output validation, the
# comment field needs to be set to begin with "TC066"
set woids = `cat $wo_id_file`
sqlplus $IAS_DB_COM << EONG
    UPDATE WORK_ORDERS SET WO_COMMENT='TC066 1 - Output Validation' WHERE WORK_ORDER_ID='$woids[1]';
EONG

echo "All work orders created successfully" |& tee -a $log
exit 0
