#!/bin/csh -f
################################################################################
# Name: validate_proper_wo_creation.csh
#
# Description: Test 066 loads a bunch of work orders with unique properties.
#   This script ensures those work orders were loaded properly.
#   The "groups" mentioned in the comments refer to how the work orders are
#   created in create_wos.csh. There are 5 groups. This script only validates
#   the first 4. The last one is used for output validation.
#
# Usage: ./validate_proper_wo_creation.csh
################################################################################

set log = validation.log

# Test work orders loaded from group 1
# Each work order should match a scene in group1_input_scenes.txt, have the cpf
# L8CPF20000401_20000630.01, and have the procedure id TIRS Alignment
set woid_file = "group1_out_woids.tmp"
set input_scene_file = "group1_input_scenes.txt"
set woids = `cat $woid_file`
set scenes = `cat $input_scene_file`
set cpf_name = "L8CPF20000401_20000630.01"
set sql_results_file = "group1_sql_results.tmp"

diff group1_out_scenes.tmp group1_input_scenes.txt
if ( $status != 0 ) then
    echo "ERROR: Group1's collected scenes are different than its input " \
        "scenes" |& tee -a $log
    exit 1
endif

rm -f $sql_results_file
foreach woid ( $woids )
    sqlplus $IAS_DB_COM << EONG
        SPOOL $sql_results_file APPEND;
        SET WRAP OFF;
        SET LINESIZE 120;
        SELECT WO.WORK_ORDER_ID, WO.SCENE_ID, WO.PROCEDURE_ID, WP.PARM_VALUE FROM WORK_ORDERS WO FULL OUTER JOIN WO_PARMS WP ON WO.WORK_ORDER_ID=WP.WORK_ORDER_ID WHERE WO.WORK_ORDER_ID='$woid' AND WP.PARM_NAME='CAL_PARM_FILENAME';
EONG
end

set i = 1
while ( $i <= $#woids )
    grep $woids[$i] $sql_results_file | grep $scenes[$i] \
        | grep "TIRS Alignment" | grep $cpf_name
    if ( $status != 0 ) then
        echo "ERROR: Work order $woid[$i] created with unexpected parameters." \
            "Was expecting scene $scenes[$i], CPF $cpf_name, and procedure " \
            "id TIRS Alignment" |& tee -a $log
        exit 1
    endif
    @ i++
end

echo "Validating group 1 work orders passed" |& tee -a $log

# Test work orders loaded from group 2
# There should only be 3 work orders created, and they should match 3
# hard-coded scene ids. Each should also have the TIRS SSM Calibration procedure
# id.
set woid_file = "group2_out_woids.tmp" 
set woids = `cat $woid_file`
set scenes = ( \
    "LC80440302013094LGN01" \
    "LC80440302013099LGN01" \
    "LC80440302013089LGN03" \
)

set sql_results_file = "group2_sql_results.tmp"

if ( $#woids != $#scenes ) then
    echo "ERROR: Group 2 created incorrect number of work orders. Was " \
        "expecting $#scenes to be created, but $#woids were actually created" \
        |& tee -a $log
    exit 1
endif

rm -f $sql_results_file
foreach woid ( $woids )
    sqlplus $IAS_DB_COM << EONG
        SPOOL $sql_results_file APPEND;
        SELECT WORK_ORDER_ID, SCENE_ID, PROCEDURE_ID FROM WORK_ORDERS WHERE WORK_ORDER_ID='$woid';
EONG
end

set i = 1
while ( $i <= $#woids )
    grep $woids[$i] $sql_results_file | grep $scenes[$i] \
        | grep "TIRS SSM Calibration"
    if ( $status != 0 ) then
        echo "ERROR: Work order $woid[$i] created with unexpected " \
            "parameters. Was expecting scene $scenes[$i] and procedure id " \
            "TIRS SSM Calibration" |& tee -a $log
        exit 1
    endif
    @ i++
end

echo "Validating group 2 work orders passed" |& tee -a $log

# Test the third set of work orders. This test relies on the same 3 work orders
# as group 2, but 2 of the 3 scenes should have been excluded by the query.
set woid_file = "group3_out_woids.tmp" 
set woids = `cat $woid_file`
set scenes = "LC80440302013099LGN01"
set sql_results_file = "group3_sql_results.tmp"

# Only a single work order should have been created
if ( $#woids != $#scenes ) then
    echo "ERROR: Group 3 created incorrect number of work orders. Was " \
        "expecting $#scenes to be created, but $#woids were actually created" \
        |& tee -a $log
    exit 1
endif

rm -f $sql_results_file
sqlplus $IAS_DB_COM << EONG
    SPOOL $sql_results_file APPEND;
    SELECT WORK_ORDER_ID, SCENE_ID, PROCEDURE_ID FROM WORK_ORDERS WHERE WORK_ORDER_ID='$woids[1]';
EONG

grep $woids[1] $sql_results_file | grep $scenes[1] | grep "TIRS SSM Calibration"
if ( $status != 0 ) then
    echo "ERROR: Work order $woids[1] created with unexpected parameters" \
        "Was expecting scene $scenes[1] and procedure id TIRS SSM Calibration" \
        |& tee -a $log
    exit 1
endif

echo "Validating group 3 work orders passed" |& tee -a $log

# Group 4 shouldn't have loaded any work orders, but should have found scenes
set woid_file = "group4_out_woids.tmp"
set scenes_file = "group4_out_scenes.tmp"
set scenes = `cat $scenes_file`
set woids = `cat $woid_file`
if ( $#woids > 0 ) then
    echo "ERROR: Group 4 should not have output any work orders" |& tee -a $log
    exit 1
endif

if ( $#scenes == 0 ) then
    echo "ERROR: Group 4 should have found some scenes" |& tee -a $log
    exit 1
endif

echo "Validating group 4 work orders passed" |& tee -a $log

echo "Successful validation" |& tee -a $log
exit 0
