#!/usr/bin/perl -w
#
##############################################################################
#
# Purpose:
#    The purpose of this script is to test the configurable nearMatch
#    functionality for BPFs is working correctly. 
#
# Usage: 
#    test_bpf_near_matches.pl
#
# Notes:
#    The following BPFs are expected to loaded into the service:
#       LO8BPF20081231230000_20090101000000.01
#       LT8BPF20081231230000_20090101000000.01
##############################################################################
use SOAP::Lite;
use ias_const;

if (!defined $ENV{IAS_SERVICES} )
{
    print "ERROR: Please set IAS_SERVICES ENV var - hostname:port\n";
    exit FAILURE;
}

###################################
# Test 1 - OLI - Slightly beyond effective date with nearMatch on. Expected to
# return expected BPF.
#################################
my $expected_bpf = "LO8BPF20081231230000_20090101000000.01";
my $parms = SOAP::Data->name(
        "getAllBpfByDateRequest" => \SOAP::Data->value(
        SOAP::Data->name("acquisitionDate" => '2009-01-01T13:00:00')->type(""),
        SOAP::Data->name("sensor" => 'OLI')->type(""),
        SOAP::Data->name("status" => 'ACTIVE')->type(""),
        SOAP::Data->name("nearMatch" => 'true')->type("")));
# Print input parms as XML
my $ser = SOAP::Serializer->new();
print $ser->envelope(freeform => $parms) . "\n";
my $soap_response;
eval
{
    $soap_response = SOAP::Lite          
                -> uri("uri:BpfQueryService")                        
                ->ns("uri:BpfQueryService","bpf")
                -> proxy($ENV{IAS_SERVICES}
                        . "/dataservices/BpfService", 
                    timeout => 99)
                ->getAllBpfNamesByDate($parms);
};
if ($@)
{
    print "Error encountered when accessing the CPF ".
        "service: ". $@;
    exit FAILURE;
}
if ($soap_response->fault())
{
    print "SOAP Error: " . $soap_response->fault()->{faultstring};
    exit FAILURE;
}

my $file_returned = $soap_response->result();
if (!defined $file_returned)
{
    print "ERROR: No files returned for test #1\n";
    exit FAILURE;
}
$file_returned = $file_returned->{filename};
if ($file_returned ne $expected_bpf)
{
    print "ERROR: $file_returned was returned. We expected $expected_bpf\n";
    exit FAILURE;
}
print "-------------\nCorrect response for test 1\n-------------\n";

###################################
# Test 2 - OLI - Beyond nearMatch "extension" of end date. Expected to return
# no BPF.
#################################
$parms = SOAP::Data->name(
        "getAllBpfByDateRequest" => \SOAP::Data->value(
        SOAP::Data->name("acquisitionDate" => '2009-01-02T13:00:00')->type(""),
        SOAP::Data->name("sensor" => 'OLI')->type(""),
        SOAP::Data->name("status" => 'ACTIVE')->type(""),
        SOAP::Data->name("nearMatch" => 'true')->type("")));
# Print input parms as XML
$ser = SOAP::Serializer->new();
print $ser->envelope(freeform => $parms) . "\n";
eval
{
    $soap_response = SOAP::Lite          
                -> uri("uri:BpfQueryService")                        
                ->ns("uri:BpfQueryService","bpf")
                -> proxy($ENV{IAS_SERVICES}
                        . "/dataservices/BpfService", 
                    timeout => 99)
                ->getAllBpfNamesByDate($parms);
};
if ($@)
{
    print "Error encountered when accessing the CPF ".
        "service: ". $@;
    exit FAILURE;
}
if ($soap_response->fault())
{
    print "SOAP Error: " . $soap_response->fault()->{faultstring};
    exit FAILURE;
}

$file_returned = $soap_response->result();
if (defined $file_returned)
{
    print "ERROR: Expected no files returned for test #2. We got " .
           $file_return->{filename} . "\n";
    exit FAILURE;
}
print "-------------\nCorrect response for test 2\n-------------\n";

###################################
# Test 3 - TIRS - Slighly beyond effective date with nearMatch on. Expected to
# return expected BPF.
#################################
$expected_bpf = "LT8BPF20081231230000_20090101000000.01";
$parms = SOAP::Data->name(
        "getAllBpfByDateRequest" => \SOAP::Data->value(
        SOAP::Data->name("acquisitionDate" => '2009-01-15T13:00:00')->type(""),
        SOAP::Data->name("sensor" => 'TIRS')->type(""),
        SOAP::Data->name("status" => 'ACTIVE')->type(""),
        SOAP::Data->name("nearMatch" => 'true')->type("")));
# Print input parms as XML
$ser = SOAP::Serializer->new();
print $ser->envelope(freeform => $parms) . "\n";
eval
{
    $soap_response = SOAP::Lite          
                -> uri("uri:BpfQueryService")                        
                ->ns("uri:BpfQueryService","bpf")
                -> proxy($ENV{IAS_SERVICES}
                        . "/dataservices/BpfService", 
                    timeout => 99)
                ->getAllBpfNamesByDate($parms);
};
if ($@)
{
    print "Error encountered when accessing the CPF ".
        "service: ". $@;
    exit FAILURE;
}
if ($soap_response->fault())
{
    print "SOAP Error: " . $soap_response->fault()->{faultstring};
    exit FAILURE;
}

$file_returned = $soap_response->result();
if (!defined $file_returned)
{
    print "ERROR: No files returned for test #3\n";
    exit FAILURE;
}
$file_returned = $file_returned->{filename};
if ($file_returned ne $expected_bpf)
{
    print "ERROR: $file_returned was returned. We expected $expected_bpf\n";
    exit FAILURE;
}
print "-------------\nCorrect response for test 3\n-------------\n";

###################################
# Test 4 - TIRS - Beyond nearMatch "extension" of end date. Expected to return
# no BPF.
#################################
$parms = SOAP::Data->name(
        "getAllBpfByDateRequest" => \SOAP::Data->value(
        SOAP::Data->name("acquisitionDate" => '2009-01-16T13:00:00')->type(""),
        SOAP::Data->name("sensor" => 'TIRS')->type(""),
        SOAP::Data->name("status" => 'ACTIVE')->type(""),
        SOAP::Data->name("nearMatch" => 'true')->type("")));
# Print input parms as XML
$ser = SOAP::Serializer->new();
print $ser->envelope(freeform => $parms) . "\n";
eval
{
    $soap_response = SOAP::Lite          
                -> uri("uri:BpfQueryService")                        
                ->ns("uri:BpfQueryService","bpf")
                -> proxy($ENV{IAS_SERVICES}
                        . "/dataservices/BpfService", 
                    timeout => 99)
                ->getAllBpfNamesByDate($parms);
};
if ($@)
{
    print "Error encountered when accessing the CPF ".
        "service: ". $@;
    exit FAILURE;
}
if ($soap_response->fault())
{
    print "SOAP Error: " . $soap_response->fault()->{faultstring};
    exit FAILURE;
}

$file_returned = $soap_response->result();
if (defined $file_returned)
{
    print "ERROR: Expected no files returned for test #4. We got " .
           $file_return->{filename} . "\n";
    exit FAILURE;
}
print "-------------\nCorrect response for test 4\n-------------\n";

###################################
# Test 5 - OLI - No nearMatch, beyond end date. Expected to return
# no BPF.
#################################
$parms = SOAP::Data->name(
        "getAllBpfByDateRequest" => \SOAP::Data->value(
        SOAP::Data->name("acquisitionDate" => '2009-01-01T13:00:00')->type(""),
        SOAP::Data->name("sensor" => 'OLI')->type(""),
        SOAP::Data->name("status" => 'ACTIVE')->type("")));
# Print input parms as XML
$ser = SOAP::Serializer->new();
print $ser->envelope(freeform => $parms) . "\n";
eval
{
    $soap_response = SOAP::Lite          
                -> uri("uri:BpfQueryService")                        
                ->ns("uri:BpfQueryService","bpf")
                -> proxy($ENV{IAS_SERVICES}
                        . "/dataservices/BpfService", 
                    timeout => 99)
                ->getAllBpfNamesByDate($parms);
};
if ($@)
{
    print "Error encountered when accessing the CPF ".
        "service: ". $@;
    exit FAILURE;
}
if ($soap_response->fault())
{
    print "SOAP Error: " . $soap_response->fault()->{faultstring};
    exit FAILURE;
}

$file_returned = $soap_response->result();
if (defined $file_returned)
{
    print "ERROR: Expected no files returned for test #5. We got " .
           $file_return->{filename} . "\n";
    exit FAILURE;
}
print "-------------\nCorrect response for test 5\n-------------\n";

###################################
# Test 6 - TIRS - No nearMatch, beyond end date. Expected to return
# no BPF.
#################################
$parms = SOAP::Data->name(
        "getAllBpfByDateRequest" => \SOAP::Data->value(
        SOAP::Data->name("acquisitionDate" => '2009-01-01T13:00:00')->type(""),
        SOAP::Data->name("sensor" => 'TIRS')->type(""),
        SOAP::Data->name("status" => 'ACTIVE')->type("")));
# Print input parms as XML
$ser = SOAP::Serializer->new();
print $ser->envelope(freeform => $parms) . "\n";
eval
{
    $soap_response = SOAP::Lite          
                -> uri("uri:BpfQueryService")                        
                ->ns("uri:BpfQueryService","bpf")
                -> proxy($ENV{IAS_SERVICES}
                        . "/dataservices/BpfService", 
                    timeout => 99)
                ->getAllBpfNamesByDate($parms);
};
if ($@)
{
    print "Error encountered when accessing the CPF ".
        "service: ". $@;
    exit FAILURE;
}
if ($soap_response->fault())
{
    print "SOAP Error: " . $soap_response->fault()->{faultstring};
    exit FAILURE;
}

$file_returned = $soap_response->result();
if (defined $file_returned)
{
    print "ERROR: Expected no files returned for test #6. We got " .
           $file_return->{filename} . "\n";
    exit FAILURE;
}
print "-------------\nCorrect response for test 6\n-------------\n";

###################################
# Test 7 - OLI - No nearMatch, within effective date. Expected to
# return expected BPF.
#################################
$expected_bpf = "LO8BPF20081231230000_20090101000000.01";
$parms = SOAP::Data->name(
        "getAllBpfByDateRequest" => \SOAP::Data->value(
        SOAP::Data->name("acquisitionDate" => '2008-12-31T23:30:00')->type(""),
        SOAP::Data->name("sensor" => 'OLI')->type(""),
        SOAP::Data->name("status" => 'ACTIVE')->type("")));
# Print input parms as XML
$ser = SOAP::Serializer->new();
print $ser->envelope(freeform => $parms) . "\n";
eval
{
    $soap_response = SOAP::Lite          
                -> uri("uri:BpfQueryService")                        
                ->ns("uri:BpfQueryService","bpf")
                -> proxy($ENV{IAS_SERVICES}
                        . "/dataservices/BpfService", 
                    timeout => 99)
                ->getAllBpfNamesByDate($parms);
};
if ($@)
{
    print "Error encountered when accessing the CPF ".
        "service: ". $@;
    exit FAILURE;
}
if ($soap_response->fault())
{
    print "SOAP Error: " . $soap_response->fault()->{faultstring};
    exit FAILURE;
}

$file_returned = $soap_response->result();
if (!defined $file_returned)
{
    print "ERROR: No files returned for test #7\n";
    exit FAILURE;
}
$file_returned = $file_returned->{filename};
if ($file_returned ne $expected_bpf)
{
    print "ERROR: $file_returned was returned. We expected $expected_bpf\n";
    exit FAILURE;
}
print "-------------\nCorrect response for test 7\n-------------\n";

###################################
# Test 8 - TIRS - No nearMatch, within effective date. Expected to
# return expected BPF.
#################################
$expected_bpf = "LT8BPF20081231230000_20090101000000.01";
$parms = SOAP::Data->name(
        "getAllBpfByDateRequest" => \SOAP::Data->value(
        SOAP::Data->name("acquisitionDate" => '2008-12-31T23:30:00')->type(""),
        SOAP::Data->name("sensor" => 'TIRS')->type(""),
        SOAP::Data->name("status" => 'ACTIVE')->type("")));
# Print input parms as XML
$ser = SOAP::Serializer->new();
print $ser->envelope(freeform => $parms) . "\n";
eval
{
    $soap_response = SOAP::Lite          
                -> uri("uri:BpfQueryService")                        
                ->ns("uri:BpfQueryService","bpf")
                -> proxy($ENV{IAS_SERVICES}
                        . "/dataservices/BpfService", 
                    timeout => 99)
                ->getAllBpfNamesByDate($parms);
};
if ($@)
{
    print "Error encountered when accessing the CPF ".
        "service: ". $@;
    exit FAILURE;
}
if ($soap_response->fault())
{
    print "SOAP Error: " . $soap_response->fault()->{faultstring};
    exit FAILURE;
}

$file_returned = $soap_response->result();
if (!defined $file_returned)
{
    print "ERROR: No files returned for test #8\n";
    exit FAILURE;
}
$file_returned = $file_returned->{filename};
if ($file_returned ne $expected_bpf)
{
    print "ERROR: $file_returned was returned. We expected $expected_bpf\n";
    exit FAILURE;
}
print "-------------\nCorrect response for test 8\n-------------\n";
exit SUCCESS;