#!/usr/local/bin/python3
""" Script that places orders in LPGS and then verifies correctness of the
    Inventory queue
"""

from ias.db import database as db
from ias.log import logger
from ias.misc.const import EXIT_SUCCESS, EXIT_FAILURE
from timestamp import Timestamp
from time import sleep
from os import getenv
from subprocess import check_call, DEVNULL, CalledProcessError


class MismatchError(Exception):
    pass


def verify_queue(start_time, stop_time):
    db_connection = db.DatabaseConnection()
    db_con_str = getenv("INV_DB_INFO")
    db_connection.username, db_connection.password = db_con_str.split('/')
    db_connection.connect()

    # Look in the queue table for messages enqueued within the last 7 seconds
    sql = ("SELECT aos.user_data.p_landsat_interval_id FROM "
           "aq$tirs_aos_queue_tbl aos WHERE enq_time > sysdate - (7 / 86400)")

    cur = db_connection.execute(sql)
    rows = cur.fetchall()
    print("Landsat Ids found in the queue: {0}".format(len(rows)))
    queue_ids = {row[0] for row in rows}

    sql = ("SELECT landsat_interval_id FROM lmd_interval WHERE date_acquired "
           "BETWEEN to_date(:START_D, 'YYYY:DDD:SSSSS') AND "
           "to_date(:STOP_D,'YYYY:DDD:SSSSS') AND landsat_interval_id IS NOT "
           "NULL")
    parms = {'START_D': str(start_time),
             'STOP_D': str(stop_time)}
    cur = db_connection.execute(sql, parms)
    rows = cur.fetchall()
    print("Landsat Ids found in LMD_MISSION: {0}".format(len(rows)))
    lmd_interval_ids = {row[0] for row in rows}

    diff = lmd_interval_ids.symmetric_difference(queue_ids)

    # Purge all queue entries
    sql = """DECLARE
        po dbms_aqadm.aq$_purge_options_t;

        BEGIN
        DBMS_AQADM.PURGE_QUEUE_TABLE(
        queue_table => 'TIRS_AOS_QUEUE_TBL',
        purge_condition => NULL,
        purge_options => po);
        END;
        """
    db_connection.execute(sql)
    db_connection.close(db.DISCONNECT_ROLLBACK)
    if len(diff) > 0:
        raise MismatchError("Queue does not match LMD_MISSION: {0}".
                            format(diff))


def main():
    status = EXIT_SUCCESS
    try:
        # Initialize the logger
        log = logger.Logger("Test Place LPGS Orders For Time Period")

        # Lists of dates to test
        start_time = [Timestamp(2014, 340, 1), Timestamp(2014, 128, 1)]
        stop_time = [Timestamp(2014, 340, 30000), Timestamp(2014, 128, 30000)]

        # Set the application name to call
        app = "place_lpgs_orders_for_time_period.py"
        output_target = None

        for start, stop in zip(start_time, stop_time):
            check_call([app,
                        '--start_time={0}'.format(start),
                        '--stop_time={0}'.format(stop),
                        '--force'],
                       stdout=output_target, stderr=output_target)
            verify_queue(start, stop)
            # Need to sleep in between calls, so can tell which
            # queue messages are from which call
            sleep(8)

        log.info("Queue functionality verified")
    except (CalledProcessError, MismatchError) as ex:
        log.error("Exception running the test: %s", ex)
        status = EXIT_FAILURE
    finally:
        log.close()
    exit(status)

# Run the main routine if called from command line
if __name__ == "__main__":
    main()
