#!/bin/csh -f
#***************************************************************************
# NAME:     verify_wo_processing_instance.csh
#
#  PURPOSE:  Verifies that the work orders for TC073 are using the correct
#            processing instance. This will be verified that the database
#            contains the correct instance for the precheck and postcheck will
#            check the work order log for the correct processing instance.
#  RETURNS:
#        Type:  Integer
#        Value: 0 / 1
#
#  USAGE:    ./verify_wo_processing_instance.csh {--pre|--pose}
#****************************************************************************
set SUCCESS = 0
set ERROR = 1

if ( -e test_case_id ) then
    set tcnum = `cat test_case_id`
    set log = tc${tcnum}.log
    set woid_file = tc${tcnum}_woids.txt
else
    echo "Error: The test_case_id file is missing"
    exit $SUCCESS
endif

# Must have one argument
if ( ${#argv} < 1 ) then
    echo "" |& tee -a $log
    echo "    Usage: $0 {--pre|--post}}" |& tee -a $log
    echo "" |& tee -a $log
    exit $ERROR
else
    switch ($1)
        case "--pre":
            set is_pre = 1
            breaksw
        case "--post":
            set is_pre = 0
            breaksw
        default:
            echo "" |& tee -a $log
            echo "    Usage: $0 {--pre|--post}}" |& tee -a $log
            echo "" |& tee -a $log
            exit $ERROR
        endsw
endif

# Correct order of work order collections
set collections = (\
    Collection1 \
    "Pre-Collection" \
    Collection1 \
    Collection1 \
    "Pre-Collection" \
    Collection1 \
    )

# Retrieve the work order ids
set woids = `cat $woid_file`

set i = 0
set error = 0


foreach work_order_id ($woids)
    @ i++

    # Verify the work orders in the database have the correct instance
    if ( $is_pre ) then
        set instance = `sqlplus -s ${IAS_DB_L8_OLITIRS}` << END
            set head off;
            SELECT PROCESSING_INSTANCE FROM WORK_ORDERS
                WHERE work_order_id='$work_order_id';
            exit;
END

        # Ensure only 1 record is found (or "no rows selected")
        if ( $#instance > 1 ) then
            echo "$work_order_id is not valid in the WORK_ORDERS table" \
                |& tee -a $log
            set error = 1
        endif

        if ( $instance != $collections[$i] ) then
            echo "$work_order_id was expected to be using processing instance"\
                 "$collections[$i], but was found under $instance" \
                 |& tee -a $log
            set error = 1
        endif
    # Verify that the correct version shows up in the work orders
    # for Collection0 there should be no version information
    else
        set soft_ver = `run_in_instance.csh $collections[$i] ias_version`
        grep -q "version: $soft_ver" $IAS_WO/$work_order_id/save/$work_order_id.log
        set exit_status = $status

        if ( $collections[$i] == "Collection1" ) then
            if ( $exit_status == 1 ) then
                echo "Error: $soft_ver was not found in $work_order_id log." \
                     |& tee -a $log
                set error = 1
            endif
        else
            if ( $exit_status == 0 ) then
                echo "Error: $soft_ver was found in $work_order_id log." \
                     |& tee -a $log
                set error = 1
            endif

        endif
    endif
end


if ( $error != 0 ) then
    echo "There were errors during validation" |& tee -a $log
    exit $ERROR
endif

echo "Successfully verified work order processing instances" |& tee -a $log
exit $SUCCESS
