#!/bin/csh -f

#***************************************************************************
# NAME:     create_wos.csh
#
#  PURPOSE:  Creates the work orders for test_case_074 using PWG
#  RETURNS:
#        Type:  Integer
#        Value: 0 / 1
#
#  USAGE:    ./create_wos.csh
#****************************************************************************

set mode = "IT"
if ( $1 == "ST" ) then
    set mode = "ST"
endif

set tc = 074
set log = tc${tc}.log
set woid_file = tc${tc}_woids.txt
set tts = "three_tirs_scenes"
set ss = "single_scene"
rm -f $log $woid_file

# Comment fields -- this is the order in which the following lists are
# dependent on
set comments = ( \
    "TC${tc} 01 Normal scene -- TIRS Only" \
    "TC${tc} 02 Normal scene -- No Saturation" \
    "TC${tc} 03 Normal scene -- OLI Bands Only" \
    "TC${tc} 04 Minimal elevation change" \
    "TC${tc} 05 Large elevation change #1" \
    "TC${tc} 06 Large elevation change #2" \
    "TC${tc} 07 Off-Nadir scene" \
    )

# L0R data paths in proper order
# $IAS_IT_TEST_DATA_DIR/l0r/ must be pre-pended for them to be complete.  The
# scenes only using a single scene are not using 3 scenes of TIRS data since
# they are not available in ST at the time this test was written, which makes
# difficult to obtain them.  So, the USE_THREE_SCENES_FLAG is used to force
# a single scene to be used so ST results will compare.
set l0r_path_dirs = ( \
        $tts/LT80400312013126LGN01/LT80400312013126LGN01 \
        $ss/LC81420332016178LGN00/LC81420332016178LGN00 \
        $ss/LC81450342016199LGN00/LC81450342016199LGN00 \
        $tts/LC80250872013133LGN01/LC80250872013133LGN01 \
        $tts/LC81720712013347LGN00/LC81720712013347LGN00 \
        $tts/LC82280832016012LGN00/LC82280832016012LGN00 \
        $tts/LC80320372013150LGN00/LC80320372013150LGN00 \
    )

# Scenes in proper order
set scene_ids = ( \
        LT80400312013126LGN01 \
        LC81420332016178LGN00 \
        LC81450342016199LGN00 \
        LC80250872013133LGN01 \
        LC81720712013347LGN00 \
        LC82280832016012LGN00 \
        LC80320372013150LGN00 \
    )

set i = 0
set pwg_status = 0

# Insert work orders
echo "Inserting work orders" |& tee -a $log
while ( $i < ${#comments} )
    @ i++

    set additional_parms = "-parm ASSESS_CLOUD_COVER:OUTPUT_SATURATION_MASK_FLAG=1 "
    set procedure_id = "L1T with Quality Band"
    set l0r = ""

    if ( $i == 3) then
      	set parm_files = ( LT8BPF20160707083823_20160707085159.01\
                           LC08CPF_20160701_20160930_01.01 )

        set additional_parms = ( $additional_parms \
            "-parm DSW:CAL_PARM_FILENAME=$parm_files[2]" \
            "-parm DSW:BIAS_PARM_FILENAME_TIRS=$parm_files[1]" \
            "-parm COMMON:BAND_LIST=(1,2,3,4,5,6,7,9) " \
        )
    else if ( $i == 2 ) then
      	set parm_files = ( LO8BPF20160625083937_20160625102938.01 \
                           LT8BPF20160620173313_20160620195437.01 )

        set additional_parms = ( $additional_parms \
            "-parm DSW:BIAS_PARM_FILENAME_OLI=$parm_files[1]" \
            "-parm DSW:BIAS_PARM_FILENAME_TIRS=$parm_files[2]" \
        )
    endif

    # Force a single scene to be used where the three scene L0Rp was not
    # available.
    if ( $i == 2 || $i == 3) then
        set additional_parms = (\
            $additional_parms \
            "-parm DSW:USE_THREE_TIRS_SCENES_FLAG=0 " \
            )
    endif

    # Check for the mode or the environment
    if ( $mode == "IT" ) then
        set l0r = "-l0r_data_path $IAS_IT_TEST_DATA_DIR/l0r/$l0r_path_dirs[$i]"
    endif

    set output = `PWG -procedure_id "$procedure_id" \
        -scene_id $scene_ids[$i] \
        -i $l0r -comment "$comments[$i]" $additional_parms`
    if ( $status ) then
        echo $output >> $log
        echo "Error loading dataset $i" |& tee -a $log
        @ pwg_status++
    else
        echo $output >> $log
        while ( $output[1] != "Work" )
            shift output
        end
        shift output
        if ( $output[1] != "Order" ) then
            echo "Error parsing work order ID from PWG command for dataset $i"\
                |& tee -a $log
        else
            shift output
            set woid = `echo $output[1] | sed 's/,//'`
            /bin/echo -n "$woid "
            echo $woid >> $woid_file
        endif
    endif
end
echo ""

# Check the pwg status for ERROR or SUCCESS
if ( $pwg_status ) then
    echo "ERROR: $pwg_status errors creating work orders with PWG" \
        |& tee -a $log
    exit 1
else
    echo "SUCCESS: created $i work orders" |& tee -a $log
endif
exit 0
