#!/bin/csh -f
#
#  Usage: validate_output.csh <work order ID> <work order ID> ...
#
#  Required: file: test_case_id, with 3-digit test case ID such as 001
#            file: test_directories, with list of subdirectories where the
#                results are, subdirs must be in:
#                $IAS_IT_TEST_DATA_DIR/test_case_XXX/
#                where XXX is the 3-digit test case ID
#-----------------------------------------------------------------------
# Create a log file
set script_name = `basename $0`
if ( -e test_case_id ) then
    set id = `cat test_case_id`
    set log = tc${id}.log
else
    set log = ${script_name:r}.out
endif

# Check for command line argument and print usage if none supplied
if ( ${#argv} < 1 ) then
    echo
    echo "    Usage: $argv[0] <work order ID> ... " | tee -a $log
    echo
    exit 1
endif

# Get the test case ID and directories
if ( ! -e test_case_id ) then
    echo "File test_case_id doesn't exist" | tee -a $log
    exit 1
endif
set id = `cat test_case_id`
if ( ! -e test_directories ) then
    echo "File test_directories doesn't exit" | tee -a $log
    exit 1
endif
set test_dir = `cat test_directories`

set now = `date +"%F %R"`
echo | tee -a $log
echo "----------------------------------------------------" | tee -a $log
echo "--- Running the validation script at $now" | tee -a $log
echo | tee -a $log

set exit_status = 0
while ( $#argv )

    # Get the work order comment field
    set woid = $argv[1]
    set select = \
        "select wo_comment from work_orders where work_order_id='$woid';"
    set wo_comment = `echo $select | sqlplus $IAS_DB_COM | grep ^TC$id `
    if ( "$wo_comment" == "" ) then
        echo
        /bin/echo -n "  Error in the Work Order $woid comment field: " \
            | tee -a $log
        /bin/echo "doesn't start with TC$id." | tee -a $log
        echo "  It's possible that the Work Order is not from this test case"\
            | tee -a $log
        echo

        # Get the next command line argument and skip the tests
        shift
        continue
    endif

    # Set a path to the work order
    set work_order_path = $IAS_WO/$woid

    # Set a path to the expected results
    set index = $wo_comment[2]
    set results_path = $IAS_IT_TEST_DATA_DIR/test_case_$id/$test_dir[$index]
    if ( ! -e $results_path ) then
        echo "Directory doesn't exist: $results_path" | tee -a $log
        exit 1
    endif

    # If the pixel mask is in the test.config file the work order file needs
    # to be converted to an image.
    set pm = `grep -c pixel_mask.h5 $results_path/test.config`
    if ( $pm == 1 ) then
        write_mask_image $IAS_WO/$woid/pixel_mask.dat \
            $IAS_WO/$woid/pixel_mask.h5 >& /dev/null
        if ( $status ) then
            echo "Error creating pixel mask image" | tee -a $log
        endif
    endif

    # If "listing" is in the test.config file a file called listing needs to
    # be generated in the work order directory.  The listing should skip the
    # omf and itself.  We also need to exclude the pixel mask files that are
    # generated in the above pixel mask check.
    set gen_listing = `grep -c '^dir_listing$' $results_path/test.config`
    if ( $gen_listing == 1 ) then
        ls -1 $IAS_WO/$woid | \
            grep -v 'omf\|dir_listing\|pixel_mask.h5\|pixel_mask_legend.txt' > \
            $IAS_WO/$woid/dir_listing
        if ( $status ) then
            echo "Error creating work order directory listing" | tee -a $log
        endif
    endif

    # Set the environment variable $IAS_DB_L8_OLITIRS to "" if its not defined.
    if ( ! $?IAS_DB_L8_OLITIRS_EVAL ) then
        set IAS_DB_L8_OLITIRS_EVAL = ""
    endif

    # Validate the results
    echo "Validating test ${wo_comment}: " | tee -a $log
    ../tools/verify_test_results.pl $woid $results_path $IAS_DB_L8_OLITIRS \
        $IAS_DB_L8_OLITIRS_EVAL | tee -a $log
    if ( $status != 0 ) then
        set exit_status = 1
        set verify_status = 1
    else
        set verify_status = 0
    endif

    # Output a pass/fail message
    set now = `date +"%F %R"`
    echo | tee -a $log
    if ( $verify_status == 0 ) then
       echo "$now PASS $woid - $test_dir[$index]" | tee -a $log
    else
       echo "****** $now FAIL $woid - $test_dir[$index]" | tee -a $log
    endif
    echo | tee -a $log

    # Get the next command line argument
    shift
end

if ( $exit_status == 1 ) then
    echo "***** FAIL ... There were failures, check the log. *****" \
         | tee -a $log
endif

set now = `date +"%F %R"`
echo "--- Completed the validation script at $now" | tee -a $log
echo | tee -a $log

exit $exit_status
