#!/usr/bin/perl

##############################################################################
#
# Purpose:
#    The purpose of this script is to dump the DATE_ACQUIRED from 
#    RADIOMETRIC_WORK_ORDER_COMMON table in a file.
#
# Usage:
#    ./dump_date_acquired.pl WOID...
#
##############################################################################

use warnings;
use strict;
use DBI;

use ias_const;

# Subroutine Declarations
sub initDb();
sub closeDb();
sub executeQuery($);
sub compareResults($);

# Database Configuration
my $woDbh;

##############################################################################
# Main
##############################################################################

if (not defined $ARGV[0])
{
    print ("Error: Missing work_order_id command line argument.\n");
    exit FAILURE;
}

# Open database connections
if (initDb() == ERROR)
{
    print "ERROR: could not connect to database\n";
    exit FAILURE;
}

while (@ARGV)
{
    my $work_order_id = shift @ARGV;

    # SQL to get the data from the work order table
    my $sql = "SELECT to_char(date_acquired, 'YYYY:DDD:HH24:MI:SS')"
            . " FROM radiometric_work_order_common"
            . " WHERE work_order_id = '$work_order_id'";

    # Call subroutine to dump the DATE_ACQUIRED from the database records
    #  to a file
    executeQuery($sql);

    # Diff the result and the expected value
    compareResults($work_order_id);
}

# Close the database connection
closeDb();
exit SUCCESS;


##############################################################################
# Subroutines
##############################################################################

##############################################################################
# Name: executeQuery
# Purpose: Executes database query and creates DATE_ACQUIRED.db.results
#          file
# Returns: SUCCESS - database query were successfully
#          FAILURE - error occurred
# Parameters: SQL query string
##############################################################################
sub executeQuery($)
{
    my $sql = shift(@_);

    my $dbError = "";

    my $selectSth = $woDbh->prepare($sql);

    if (not defined $selectSth)
    {
        print "ERROR: SQL PREPARE ERROR : $DBI::errstr\n";
        closeDb();
        exit FAILURE;
    }
    elsif (!$selectSth->execute())
    {
        print "ERROR: SQL EXECUTE ERROR : $DBI::errstr\n";
        $selectSth->finish();
        closeDb();
        exit FAILURE;
    }
   
    my @row = ();
    my $r;
    if (!open(OUTF, "> DATE_ACQUIRED.db.results"))
    {
        print "Couldn't open the file DATE_ACQUIRED.db.results for output.\n";
    } 
    print OUTF ("DATE_ACQUIRED\n");

    while(@row = $selectSth->fetchrow_array)
    {
        my $date_acquired = $row[0]; 
        print OUTF "$date_acquired\n";
    }
    $selectSth->finish();
    return SUCCESS;
}

##############################################################################
# Name: CompareResults
# Purpose: Diff the results for DATE_ACQUIRED
# Returns: SUCCESS - diff were successfully
#          ERROR - error occurred
# Parameters: Work order ID
##############################################################################
sub compareResults($)
{
    my $wo_id = shift(@_);
    my @directories;
    my @wo_id_from_file;
    my $i;
    my $index = 0;
    my $fh;
    my $fh1;
    my $fh2;
    my $IAS_IT_TEST_DATA_DIR = $ENV{"IAS_IT_TEST_DATA_DIR"};
    my $test_case_id;
    my $test_case_woids;

    # Get the test case id from test_case_id file.
    if(!open ($fh, '<', "test_case_id"))
    {
        print "Couldn't open test_case_id file to read.\n";
        return ERROR;
    } 
    chomp ($test_case_id = <$fh>);
    $test_case_woids = "tc" . $test_case_id . "_woids.txt";
    
    # Read the work order id for the expected value and store in the array.
    if (!open ($fh1, '<', $test_case_woids))
    {
       print "Couldn't open $test_case_woids file to read.\n";
       return ERROR;
    }
    chomp (@wo_id_from_file = <$fh1>);    

    # Read the directories for the expected value and store in the array.
    if (!open ($fh2, '<', "test_directories"))
    {
       print "Couldn't open test_directories file to read.\n";
       return ERROR;
    }
    chomp (@directories = <$fh2>); 

    close($fh1);
    close($fh2);
    
    $i = 0;
    foreach (@wo_id_from_file)
    {
        if ($wo_id_from_file[$i] eq $wo_id)
        {
           $index = $i;
        }
        $i++;
    }

    my $results_path = "$IAS_IT_TEST_DATA_DIR/test_case_$test_case_id/"
        . "$directories[$index]";
    my $cmd = "diff -q DATE_ACQUIRED.db.results "
             . "$results_path/DATE_ACQUIRED.db.verify";    
    print "Running: $cmd\n";
    my $rc = system($cmd);
    if ($rc != SUCCESS)
    {
        print "Diff failed: DATE_ACQUIRED - $wo_id\n";
        return ERROR;
    }
    else 
    {
        print "Running diffs passed - $wo_id\n";
        # Diff successful, remove the file created for diff.
        $rc = system ("rm", "-rf", "DATE_ACQUIRED.db.results");
        if ($rc != SUCCESS)
        {
             print "WARN: could not remove file: DATE_ACQUIRED.db.results.\n";
             return ERROR;
        }
        return SUCCESS;
    }    
}

##############################################################################
# Name: initDb
# Purpose: Initiate database connections
# Returns: SUCCESS - database connections were successfully opended
#          ERROR - error occurred
# Parameters: None
##############################################################################
sub initDb()
{
    my $woEnv = $ENV{"IAS_DB_L8_OLITIRS"};
    my $sid = $ENV{"ORACLE_SID"};
    my $username;
    my $password;

    # verfiy environment variables were set
    if (not defined $woEnv)
    {
        print "ERROR: IAS_DB_L8_OLITIRS not defined\n";
        return ERROR;
    }

    if (not defined $sid)
    {
        print "ERROR: ORACLE_SID not defined\n";
        return ERROR;
    }

    ## Open work order database connection
    ($username, $password) = split(/\//,$woEnv);

    $woDbh = DBI->connect("DBI:Oracle:$sid", $username, $password,
                           {AutoCommit => 0, PrintError =>0});

    if (!$woDbh)
    {
        print "ERROR: could not connect to database: $DBI::errstr\n";
        return ERROR;
    }
    
    return SUCCESS;
}

##############################################################################
# Name: closeDb
# Purpose: Close database connections to TRAM and WO_COMMON
# Returns: None
# Parameters: None
##############################################################################
sub closeDb()
{
    $woDbh->disconnect();
}

