/*
 * BaseDAO
 *
 * Base class for all Data Access Object (DAO) classes.  This class provides
 * access to the associated DataConnection as well as SQL utility methods.
 */
package gov.usgs.ias.dao;

import gov.usgs.ias.constant.Constants;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Timestamp;
import java.util.Date;
import org.apache.commons.lang.StringEscapeUtils;
import org.apache.log4j.Logger;

public class BaseDAO {
    
    /** Holds the logging component for this class */
    private static final Logger logger
            = Logger.getLogger(Constants.LOGGER_DATA);
    /** error String for use when failed to close a ResultSet */
    protected static final String CLOSE_RS = "cannot close ResultSet";
    /** error String for use when failed to close a Statement */
    protected static final String CLOSE_STMT = "cannot close Statement";
    /** The actual DataConnection to the database */
    private DataConnection conn;
    /** The name of the property containing authorization (username/passwd) */
    protected String daoAuthorizationProperty = null;

    /**
     * Constructor for the BaseDAO class.
     * @param conn   the DataConnection to use for database communication
     */
    public BaseDAO(final DataConnection conn) {
        this.conn = conn;
    }

    /**
     * Constructor for the BaseDAO class, with the name of the property to
     * use to get authorization information for database login.
     * @param conn     The DataConnection to communicate with the database.
     * @param daoAuth  The name of the property containing authorization
     *                 (username/password)
     */
    public BaseDAO(final DataConnection conn, final String daoAuth) {
        this.conn = conn;
        this.daoAuthorizationProperty = daoAuth;
    }
    
    /**
     * Helper method for generating the next sequence number.
     * @param seqName   the Oracle sequence name to use
     * @return          the next sequence value from the database
     * @throws java.sql.SQLException if error communicating with the database
     */
    public long nextSequence(final String seqName) throws SQLException {
        final String sql = "SELECT " + seqName + ".nextval FROM dual";
        PreparedStatement stmt = null;
        ResultSet rs = null;
        long nextVal = -1;
        Connection dbConnection = null;

        try {
            if (daoAuthorizationProperty == null) {
                dbConnection = this.getConn().getConnection();
            } else {
                dbConnection = this.getConn().getConnection(
                    daoAuthorizationProperty);
            }            
            stmt = dbConnection.prepareStatement(sql);
            rs = stmt.executeQuery();

            if (rs.next()) {
                nextVal = rs.getLong(1);
            }
        } finally {
            this.close(rs, stmt);
        }

        return nextVal;
    }

    /**
     * Helper method for querying the current timestamp from the database
     * @return the timestamp as a Java Date object
     * @throws java.sql.SQLException if error communicating with the database
     */
    public Date getSysTimestamp() throws SQLException {
        final String sql = "SELECT SYSTIMESTAMP from DUAL";
        Date result = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
        Connection dbConnection = null;

        try {
            if (daoAuthorizationProperty == null) {
                dbConnection = this.getConn().getConnection();
            } else {
                dbConnection = this.getConn().getConnection(
                    daoAuthorizationProperty);
            }
            stmt = dbConnection.prepareStatement(sql);
            rs = stmt.executeQuery();
            if (rs.next()) {
                result = timestampFromSql(rs.getTimestamp(1));
            }
        } finally {
            this.close(rs, stmt);
        }

        return result;
    }

    /**
     * Converts a database 'Y'/'N' string to java boolean for applications
     * @param s   the database yes/no value
     * @return    true if 'Y' else false
     */
    public static Boolean yesNoFromSql(final String s) {
        return Boolean.valueOf((s != null) && ("Y".equals(s)));
    }

    /**
     * Converts a java boolean to 'Y'/'N' string for the database
     * @param b   the boolean flag to convert
     * @return    'Y' if true else 'N'
     */
    public static String yesNoToSql(final Boolean b) {
        return ((b != null) && (b.booleanValue())) ? "Y" : "N";
    }

    /**
     * Converts a long value 0/1 to java boolean for applications
     * @param v   the 0/1 value to convert
     * @return    true if v not equal to 0 otherwise false
     */
    public static Boolean boolFromSql(final long v) {
        return Boolean.valueOf(v != 0);
    }

    /**
     * Converts a java boolean to a 0/1 long value for the database
     * @param b   the boolean flag to convert
     * @return    1 if true else 0
     */
    public static long boolToSql(final Boolean b) {
        return ((b != null) && (b.booleanValue())) ? 1 : 0;
    }

    /**
     * Helper for converting a java.sql.Timestamp to java.util.Date
     * will use '0' millisecond time as null specifier
     * @param ts   the timestamp from the database
     * @return     null for a null'd out date otherwise a valid java.util.Data
     *             representing the timestamp
     */
    public static Date timestampFromSql(final java.sql.Timestamp ts) {
        Date date = null;
        if (ts != null) {
            date = (ts.getTime() == 0) ? null : new Date(ts.getTime());
        }
        return date;
    }

    /**
     * Helper for converting a java.util.Date to a java.sql.Timestamp
     * If date is null it will use '0' millisecond time as null specifier since
     * the database does not allow nullable columns
     * @param date   the java.util.Date representation of the timestamp
     * @return       a valid java.sql.Timestamp (non-null) that can go to the
     *               database
     */
    public static Timestamp timestampToSql(final Date date) {
        Timestamp ts = null;
        if (date == null) {
            ts = new Timestamp(0);
        } else {
            ts = new Timestamp(date.getTime());
        }
        return ts;
    }

    /**
     * Escapes a string for a Oracle LIKE statement
     * @param text         the LIKE clause text to escape
     * @param escapeChar   the escape character to use
     * @return             a string with special chararacters escaped
     */
    public static String escapeLikeString(String text, char escapeChar) {
        if (text == null || text.trim().length() <= 0) {
            return text;
        }

        // will escape single quotes
        String escaped = StringEscapeUtils.escapeSql(text);

        // find all percent '%' and '_' chars and escape
        char prev = 0;
        StringBuilder sb = new StringBuilder();
        int n = escaped.length();
        for (int i = 0; i < n; i++) {

            char c = escaped.charAt(i);

            if (c == '%' || c == '_') {
                if (prev != escapeChar) {
                    sb.append(escapeChar);
                }
            }
            sb.append(c);
            prev = c;
        }

        return sb.toString();
    }

    /**
     * This method will attempt to close a ResultSet
     * Appropriate logging will be used in case of errors.
     * @param rs   The ResultSet to be closed
     */
    public void close(final ResultSet rs) {
        if (rs != null) {
            try {
                rs.close();
            } catch (SQLException e) {
                logger.debug(BaseDAO.CLOSE_RS + ", " + rs, e);
            }
        }
    }

    /**
     * This method will attempt to close a (Prepared)Statement.
     * Appropriate logging will be used in case of errors.
     * @param stmt   The Statement to be closed
     */
    public void close(final Statement stmt) {
        if (stmt != null) {
            try {
                stmt.close();
            } catch (SQLException e) {
                logger.debug(BaseDAO.CLOSE_STMT + ", " + stmt, e);
            }
        }
    }

    /**
     * This method will attempt to close both a ResultSet and a 
     * (Prepared)Statement.  Attempts will be made to close both,
     * even if one or the other fails.
     * Appropriate logging will be used in case of errors.
     * @param rs     the ResultSet to be closed
     * @param stmt   the Statement to be closed
     */
    public void close(final ResultSet rs, final Statement stmt) {
        this.close(rs);
        this.close(stmt);
    }

    /**
     * Get the current connection being used by this DAO
     * @return   the current DataConnection object
     */
    protected DataConnection getConn() {
        return conn;
    }

    /**
     * Set the connection for this to use.  One may want to switch the
     * connection being used to one that is different than the one used when
     * this DAO was created.
     * @param conn   the new DataConnection to use
     */
    protected void setConn(final DataConnection conn) {
        this.conn = conn;
    }

    /**
     * Get the current authorization string (which login credentials to use)
     * @return   authorization string username/password
     */
    public String getDaoAuth() {
        return daoAuthorizationProperty;
    }

    /**
     * Set the current authorization string (which login credentials to use)
     * @param daoAuth   the authorization string (username/password) to set
     */
    public void setDaoAuth(String daoAuth) {
        this.daoAuthorizationProperty = daoAuth;
    }
}
