/*
 * DataConnection
 *
 * A simple wrapper around a standard Connection object.  Provides some
 * utilities such as commit/rollback as well as managing the connection
 * string
 */
package gov.usgs.ias.dao;

import gov.usgs.ias.constant.Constants;
import gov.usgs.ias.cps.CpsException;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.sql.DataSource;
import org.apache.log4j.Logger;

public class DataConnection {

    /** Holds the logging component for this class */
    private static final Logger logger
            = Logger.getLogger(Constants.LOGGER_DATA);
    /** the actual JDBC connection object */
    private Connection conn = null;
    /** determines if the connection is managed or not (EJB) */
    private boolean managed = false;
    /** the name of the dao system property to use */
    private String daoProperty = null;
    /** the name of the TWO_TASK property/environment variable */
    private final String TWO_TASK = "TWO_TASK";

    /** Constructor for the DataConnection that will allow caller to assign
     * data source via a system property
     * 
     * @param daoProperty    the name of the system property to get the DAO
     *                       connection string from
     */
    public DataConnection(final String daoProperty) {
        this.daoProperty = daoProperty;
    }

    /**
     * Returns the actual Connection object to make calls to the database.  
     * Does a lazy-initialization--if it is the first time called, this method
     * will actually create the connection to the database
     * @return the underlying Connection object
     * @throws java.sql.SQLException if error getting a connection
     */
    public Connection getConnection() throws SQLException {
        if (this.conn == null || this.conn.isClosed()) {
            try {
                // allow a system property to override our connection
                // string (ie. Unit testing will use this to specify a test-db)
                final String connString = System.getProperty(this.daoProperty);
                if (connString == null) {
                    throw new CpsException("Database connection string "
                        + "property " + this.daoProperty + " is not set!");
                }
                this.conn = createConnection(connString);
            } catch (Exception ex) {
                logger.error("Problem getting DB connection", ex);
                throw new SQLException("Problem getting DB connection", ex);
            }
        }

        return this.conn;
    }

    /**
     * Returns the actual Connection object to make calls to the database.
     * This method will get the connection string out of system properties
     * and then replace the "authorization" placeholder with the value of the
     * authorization property/environment variable passed in.  The environment
     * variable will take precedence over the property setting.
     * @param authPropName  name of the environment variable/property to use to
     *                      get the authorization (user name & password)
     * @return              the underlying Connection object
     * @throws java.sql.SQLException if error getting a connection
     */
    public Connection getConnection(String authPropName) throws SQLException {
        if (this.conn == null || this.conn.isClosed()) {
            try {
                String daoUidPwd = System.getenv(authPropName);
                if (daoUidPwd == null) {
                    // environment variable is not set, attempt to get the value
                    // from a system property
                    daoUidPwd = System.getProperty(authPropName);
                }
                if ((daoUidPwd == null) || (daoUidPwd.trim().isEmpty())) {
                    throw new CpsException("Database authorization environment "
                        + " variable/property " + authPropName
                        + " is not set!");
                }

                String connString = System.getProperty(daoProperty);
                if ((connString == null) || (connString.trim().isEmpty())) {
                    throw new CpsException("Database connection string "
                        + "property " + daoProperty + " is not set!");
                }

                // Replace the authorization placeholder with the actual
                // authorization string.
                connString = connString.replace(":authorization@",
                        ":" + daoUidPwd.trim() + "@");

                // Replace the two_task placeholder with the value of the
                // TWO_TASK  environment variable or property
                // Note: to use TWO_TASK in a jdbc:thin connection,
                // you must also define the oracle.net.tns_admin property;
                // it should normally be assigned to the $TNS_ADMIN directory
                if (connString.contains("@two_task")) {
                    String two_task = System.getenv(TWO_TASK);
                    if (two_task == null) {
                        two_task = System.getProperty(TWO_TASK);
                    }
                    if (two_task != null) {
                        connString = connString.replace("@two_task",
                            "@" + two_task);
                    }
                }

                this.conn = createConnection(connString);

            } catch (Exception ex) {
                logger.error("Problem getting DB connection", ex);
                throw new SQLException("Problem getting DB connection", ex);
            }
        }
        return this.conn;
    }

    /**
     * Create a connection using the provided connection string
     * @param connectionString  fully qualified connection string
     * @return                  The Connection object
     * @throws Exception for error accessing a data source, locating the driver,
     *                   getting the connectoin, or setting the automCommit mode
     */
    private Connection createConnection(String connectionString) 
            throws Exception {
        Connection newConnection = null;

        // this is a JNDI name for the connection
        if (connectionString.startsWith("java:")) {
            // get the (pooled) connection from JNDI (EJB context)
            final Context initContext = new InitialContext();
            final DataSource ds
                    = (DataSource)initContext.lookup(connectionString);
            newConnection = ds.getConnection();

        } else {
            // if here then they specified a raw connection string
            Class.forName("oracle.jdbc.driver.OracleDriver");
            newConnection = DriverManager.getConnection(connectionString);
        }

        // Turn off autocommit if it is turned on
        //  When a raw connection is created, it is in the auto-commit mode so
        //    autocommit must be turned off so the user can manage transactions
        //  When a connection is created in EJB context (pooled connection
        //    retrieved from the DataSource), the connection actually has
        //    already been created & is being managed in a connection pool, 
        //    autocommit is already turned off, and transactions are handled
        //    by the EJB container (e.g, in jboss, the CachedConnectionManager).
        if (newConnection.getAutoCommit()) {
            newConnection.setAutoCommit(false);
            managed = false;
        } else {
            managed = true;
        }

        return newConnection;
    }

    /**
     * Returns the managed flag for the connection
     * @return true if the connection is managed (in EJB context) or not
     */
    public boolean isManaged() {
        return this.managed;
    }

    /**
     * Closes the underlying connection 
     */
    public void close() {
        try {
            if (this.conn != null) {
                this.conn.close();
            }
        } catch (SQLException ex) {
            logger.info("Cannot successfully close connection to DB", ex);
        } finally {
            this.conn = null;
        }
    }

    /**
     * Call to commit any pending transactions to the database - this method 
     * will have no effect if using a managed connection.
     * @throws java.sql.SQLException if error communicating with the database
     */
    public void commit() throws SQLException {
        if (!this.isManaged()) {
            this.conn.commit();
        }
    }

    /**
     * Call to rollback any pending transactions to the database - this method 
     * will have no effect if using a managed connection.  No exception is
     * thrown; if the rollback fails, there is nothing you can do about it
     * anyway.
     */
    public void rollback() {
        try {
            if (!this.isManaged()) {
                this.conn.rollback();
            }
        } catch (Exception ex) {
            logger.info("Cannot successfully rollback DB connection.", ex);
        }
    }
}
