/**
 * BpfDAO
 *
 * Data Access Object for BPF content data (groups and attributes)
 * This extends the CpfDAO class and overrides the methods that do the database
 * queries, changing table & field names as appropriate for BPF.
 */
package gov.usgs.ias.dao.bpf;

import gov.usgs.ias.cps.CpsGroup;
import gov.usgs.ias.cps.CpsException;
import gov.usgs.ias.cps.CpsStatusEnum;
import gov.usgs.ias.dao.cpf.*;
import gov.usgs.ias.dao.DataConnection;
import java.sql.*;
import java.util.*;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import gov.nasa.pds.tools.label.*;
import org.apache.log4j.Logger;

public class BpfDAO extends CpfDAO {

    private final Logger logger = Logger.getLogger(this.getClass().getName());

    // map of IDs for BPF parameter names
    private HashMap<String, Long> bpfParameterIds = null;

    // 7 null values, used to fill in the arguments to the stored proc when 
    // fewer than 7 parameter values are supplied
    final private List<String> null7 = Arrays.asList(
                (String)null, (String)null, (String)null, (String)null,
                (String)null, (String)null, (String)null);

    /**
     * Constructor for the BpfDAO class.
     * @param conn      The DataConnection to communicate with the database.
     * @param daoAuth   The name of the authorization property to use when
     *                  making the connection.
     */
    public BpfDAO(final DataConnection conn, final String daoAuth) {
        super(conn, daoAuth);
    }

    /**
     * Creates CpsGroup objects for each group in a BPF.
     * @param bpfName   The name of the BPF.
     * @return          A collection of CpsGroup objects.
     * @throws          java.sql.SQLException for database errors
     */
    @Override
    public Collection<CpsGroup> getCpsGroups(String bpfName)
            throws SQLException {

        Collection<CpsGroup> groups = new ArrayList<CpsGroup>();
        if (bpfName == null) {
            logger.warn("No BPF name supplied");
            return groups;
        }

        // Join the BPF main table with the group table (using the
        // cross-reference bpf_main_group table) to get the names and IDs of
        // all the groups that belong to this BPF.
        // There is no checksum column in the bpf tables, so put a null where it
        // should be for the cpsGroupsFromResultSet(rs) method.
        String sql = "SELECT bpfg.name, bpfg.bpf_group_id, null as checksum, " +
                "bpfmg.group_order " +
                "FROM bpf_main bpfm " +
                "INNER JOIN bpf_main_group bpfmg " +
                "ON bpfm.bpf_main_id = bpfmg.bpf_main_id " +
                "INNER JOIN bpf_group bpfg " +
                "ON bpfmg.bpf_group_id = bpfg.bpf_group_id " +
                "WHERE bpfm.file_name = ? " +
                "ORDER BY bpfmg.group_order";

        PreparedStatement stmt = null;
        ResultSet rs = null;

        try {
            stmt = this.getConn().getConnection(this.daoAuthorizationProperty).
                    prepareStatement(sql);
            stmt.setString(1, bpfName);
            rs = stmt.executeQuery();

            while (rs.next()) {
                groups.add(this.cpsGroupsFromResultSet(rs));
            }
        } catch (SQLException ex) {
            throw ex;
        } finally {
            this.close(rs, stmt);
        }

        return groups;
    }

    /**
     * Generates the PreparedStatement for getting parameters for a BPF group
     * out of the database
     * @return The statement ready for execution
     * @throws java.sql.SQLException for database errors
     */
    @Override
    public PreparedStatement getCpsGroupDataStmt() throws SQLException {

        PreparedStatement stmt = null;
        // Join the parameter table with the parameter values table using a
        // given group ID to get the names, IDs, and values
        // (sorted in order) of all the parameters that
        // belong to the group.
        String sql = "SELECT cv.bpf_parameter_id as parameter_id, bpfp.name, "+
                "cv.parameter_order, cv.value_order, cv.value "+
                "FROM bpf_parameter_values cv "+
                "INNER JOIN bpf_parameter bpfp "+
                "ON cv.bpf_parameter_id = bpfp.bpf_parameter_id "+
                "WHERE cv.bpf_group_id = ? "+
                "ORDER BY cv.parameter_order, cv.value_order ";
        
        stmt = this.getConn().getConnection(this.daoAuthorizationProperty).
                prepareStatement(sql);
        // There are an unkown number of rows that may be returned, but we
        // expect there could be thousands.  So, set the fetch size to an
        // arbitrarily large batch size to reduce the number of trips to the
        // database server.
        stmt.setFetchSize(1000);

        return stmt;
    }

    /**
     * Retreives a CpsGroup from the database.
     * @param bpfGroupId      The bpf_group_id to use as the root for retrieval.
     * @param addParameters   Should the parameters be included in the return
     *                        object.
     * @return                The CpsGroup object with data loaded.
     * @throws                SQLException for database errors
     */
    @Override
    public CpsGroup getCpsGroupById(final long bpfGroupId,
            final boolean addParameters) throws SQLException {
        
        CpsGroup returnGroup = null;

        // there is no checksum column in the bpf tables, so put a null where it
        // should be for the cpsGroupsFromResultSet(rs) method.
        String sql = "SELECT bpfg.name, bpfg.bpf_group_id, null as checksum, " +
                "0 as group_order " +
                "FROM bpf_group bpfg " +
                "WHERE bpfg.bpf_group_id = ? ";

        PreparedStatement stmt = null;
        ResultSet rs = null;

        try {
            stmt = this.getConn().getConnection(this.daoAuthorizationProperty).
                    prepareStatement(sql);
            stmt.setLong(1, bpfGroupId);
            rs = stmt.executeQuery();

            while (rs.next()) {
                returnGroup = this.cpsGroupsFromResultSet(rs);
            }

            if (addParameters) {
                this.getGroupParameters(returnGroup);
            }
        } catch (SQLException ex) {
            throw ex;
        } finally {
            this.close(rs, stmt);
        }

        return returnGroup;
    }

    /**
     * Inserts a BPF into the database
     * @param bpfLabel   The parsed ODL label structure for the BPF.
     * @param bpfStatus  The initital status of the file in the database.
     * @return           the database id of the BPF inserted
     * @throws           java.sql.SQLException for database errors
     */
    @Override
    public int insertFile(Label bpfLabel, CpsStatusEnum bpfStatus)
            throws Exception {

        int bpfId = -1; // ID for this BPF being inserted
        int status;     // general method return status
        Calendar gmtCalendar = Calendar.getInstance(
            TimeZone.getTimeZone("GMT-0:00"));

        // Extract metadata about the BPF from the file itself
        String bpfFileName = getFileAttribute(bpfLabel, "File_Name");
        String bpfVersion = getFileAttribute(bpfLabel, "Version");
        String bpfSource = getFileAttribute(bpfLabel, "File_Source");
        String bpfDescr = getFileAttribute(bpfLabel, "Description");
        String effBegin = getFileAttribute(bpfLabel, "Effective_Date_Begin");
        String effEnd = getFileAttribute(bpfLabel, "Effective_Date_End");
        String sensorName = getFileAttribute(bpfLabel, "Sensor_Name");
        String orbitBegin = getOrbitAttribute(bpfLabel, "Orbit_Number");
        if (bpfFileName == null || bpfVersion == null
                || bpfSource == null || bpfDescr == null
                || effBegin == null || effEnd == null
                || sensorName == null || orbitBegin == null) {
            throw new CpsException(
                    "BPF is missing required FILE_ATTRIBUTES attributes "
                    + "(File_Name, Version, File_Source, Description,"
                    + " Sensor_Name, "
                    + " Effective_Date_Begin, Effective_Date_End) and/or "
                    + "ORBIT_PARAMETERS attribute (Orbit_Number)");
        }

        // Translate sensor name into ID stored in the main table
        String sensorId;
        if (sensorName.matches("^Operational Land Imager$") 
            || sensorName.matches("OLI")) {
            sensorId = "OLI";
        } else if (sensorName.matches("^Thermal Infrared Sensor$") 
            || sensorName.matches("TIRS")) {
            sensorId = "TIRS";
        } else {
            throw new CpsException("Unknown sensor " + sensorName);
        }

        // Get the connection (the DataConnection class turns off autocommit
        //   so we can manage our transaction here)
        Connection conn = this.getConn().getConnection(
                this.daoAuthorizationProperty);
        CallableStatement mainStatement = null;
        CallableStatement groupStatement = null;
        CallableStatement parmStatement = null;

        // Storing a BPF is a 3-step process done in a transaction.
        // First, insert the main record
        // Second, create the group
        // Third, insert the parameters
        try {
            // Set up the stored procedure statements
            mainStatement = conn.prepareCall(
                    "{call PROCESS_BPF_PKG.proc_insert_bpf_main "
                    + "(?,?,?,?,?,?,?,?,?,?)}");
            groupStatement = conn.prepareCall(
                    "{call PROCESS_BPF_PKG.proc_create_bpf_group"
                    + "(?,?,?,?,?)}");
            parmStatement = conn.prepareCall(
                            "{call PROCESS_BPF_PKG.proc_create_bpf_parameter"
                            + "(?,?,?,?,?,?,?,?,?,?,?,?)}");
            
            DateFormat df = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
            df.setTimeZone(TimeZone.getTimeZone("GMT-0:00"));
            mainStatement.setString(1, bpfFileName);
            mainStatement.setString(2, bpfSource);
            mainStatement.setString(3, bpfDescr);
            mainStatement.setInt(4, Integer.parseInt(bpfVersion));
            mainStatement.setInt(5, Integer.parseInt(orbitBegin));
            mainStatement.setString(6, sensorId);
            mainStatement.setTimestamp(7, timestampToSql(df.parse(effBegin)),
                    gmtCalendar);
            mainStatement.setTimestamp(8, timestampToSql(df.parse(effEnd)),
                    gmtCalendar);
            mainStatement.setString(9, bpfStatus.toString());
            mainStatement.setInt(10, -1); // initialize the output BPF ID
            mainStatement.registerOutParameter(10, Types.INTEGER);
            status = mainStatement.executeUpdate();
            if (status != 1) {
                throw new SQLException("Error executing BPF "
                        + "stored procedure to insert main record for "
                        + bpfFileName);
            }
            bpfId = mainStatement.getInt(10);
            if (bpfId == -1) {
                throw new SQLException("Error inserting main record for "
                        + bpfFileName);
            }
            logger.debug("Inserted bpf_main, record_id is " + bpfId);

            // Second, insert each group and the parameters that go with it
            List lg = bpfLabel.getGroups();
            Iterator gi = lg.iterator();
            GroupStatement gs = null;
            int groupId = -1;
            int groupOrder = 1;

            while (gi.hasNext()) {
                gs = (GroupStatement) gi.next();
                logger.trace("Inserting Group " + groupOrder + ": "
                        + gs.getIdentifier());

                // Create the group
                groupStatement.setInt(1, bpfId);
                groupStatement.setString(2, bpfFileName);
                groupStatement.setString(3, gs.getIdentifier());
                groupStatement.setInt(4, groupOrder);
                groupStatement.setInt(5, -1);
                groupStatement.registerOutParameter(5, Types.INTEGER);
                status = groupStatement.executeUpdate();
                if (status != 1) {
                    throw new SQLException("Error executing proc to create  "
                            + gs.getIdentifier());
                }

                groupId = groupStatement.getInt(5);
                logger.trace("Inserted Group " + gs.getIdentifier() + " as ID "
                        + groupId);

                // Create the parameters in the group
                //  If the group is one of the BAND_BIAS_MODEL_* groups, use
                //    straight batch inserts due to the large number of
                //    parameters.
                //  Otherwise, use the stored proc to handle the parameter
                //    creations, etc.
                if (gs.getIdentifier().startsWith("BIAS_MODEL_")) {

                    insertBiasModelParams(conn, gs, groupId);

                } else {

                    int returnId = 0;
                    int parameterOrder = 1;
                    
                    parmStatement.setInt(1, bpfId);
                    parmStatement.setInt(2, groupId);

                    for (String parameterName : getParameterNames(gs)) {
                        // Initialize info for this parameter
                        parmStatement.setString(3, parameterName);
                        parmStatement.setInt(4, parameterOrder++);
                        parmStatement.setInt(12, -1);
                        parmStatement.registerOutParameter(12, Types.INTEGER);

                        // Set the values into the statement
                        //   It is an error if there are more than 7 values
                        List<String> parameterValues = getAttributeValues(
                                gs.getAttribute(parameterName));
                        if (parameterValues.size() > 7) {
                            throw new CpsException("Too many values found for "
                                    + parameterName);
                        }
                        //   Add on nulls to ensure there are 7 values when
                        //     calling the stored proc
                        parameterValues.addAll(null7);
                        int valueOrder = 5;  // parameter values are args 5 - 11
                        for (String parameterValue :
                                parameterValues.subList(0, 7)) {
                            parmStatement.setString(valueOrder++,
                                    parameterValue);
                        }

                        // execute to insert this parameter's values (1 row)
                        status = parmStatement.executeUpdate();
                        if (status != 1) {
                            throw new SQLException("Error executing BPF "
                                    + "stored procedure for " + parameterName);
                        }
                        // expect a 1 returned if the stored proc was successful
                        returnId = parmStatement.getInt(12);
                        if (returnId != 1) {
                            throw new CpsException("BPF stored procedure "
                                    + "returned an error for " + parameterName);
                        }
                    }
                }
                groupOrder++;
            }

            // Finally, commit the transaction
            conn.commit();
            logger.info("Completed BPF, ID is " + bpfId);

            // Propagate any exceptions thrown
        } catch (Exception e) {
            try {
                conn.rollback();
            } catch (SQLException ex) {
                logger.warn("Error rolling back transaction", ex);
            }
            throw e;
        } finally {
            this.close(mainStatement);
            this.close(groupStatement);
            this.close(parmStatement);
        }

        return bpfId;
    }

    /**
     * Gets an attribute from the ORBIT_PARAMETERS group of the parsed ODL tree
     * @param lab             The parsed ODL label structure for the BPF
     * @param attributeName   Name of the attribute to find the value for
     * @return                the value of the named attribute
     */
    private String getOrbitAttribute(Label lab, String attributeName) {

        Scalar sv = null;
        try {
            sv = (Scalar)((GroupStatement)lab.getGroups("ORBIT_PARAMETERS").
                    get(0)).getAttribute(attributeName).getValue();
        } catch (Exception e) {
            // If the group or attribute was not found, an exception is thrown
            // by the getAttribute or getValue, so return null
            return null;
        }
        return sv.getValue();
    }

    /**
     * Inserts parameters for a bias model parameter group.  One insert
     * statement is executed per parameter in the group, and all the statements
     * are executed in one batch.
     * @param conn      The open database connectoin
     * @param gs        The parsed ODL group statement
     * @param groupId   The database ID for this group
     */
    private void insertBiasModelParams(Connection conn, GroupStatement gs,
            int groupId)
            throws Exception {

        // First get the parameter IDs for all the bias model parameter names
        getBpfParameterIds(conn);

        // Set up an insert statement for the parameters in this group
        PreparedStatement insertStmt = null;
        try {
            insertStmt = conn.prepareStatement(
                    "INSERT INTO BPF_PARAMETER_VALUES_DATA (BPF_GROUP_ID, "
                    + "BPF_PARAMETER_ID, PARAMETER_ORDER, "
                    + "VALUE1, VALUE2, VALUE3, VALUE4, VALUE5, VALUE6, VALUE7) "
                    + "values (?,?,?,?,?,?,?,?,?,?)");
            insertStmt.setInt(1, groupId);

            // Insert each parameter in the group
            int parameterOrder = 1;
            for (String parameterName : getParameterNames(gs)) {
                // Initialize info for this parameter
                Long parameterId = bpfParameterIds.get(parameterName);
                if (parameterId == null) {
                    throw new CpsException("BPF Parameter " + parameterName
                            + " does not exist in the database");
                }
                insertStmt.setLong(2, parameterId.longValue());
                insertStmt.setInt(3, parameterOrder++);

                // Set the values into the statement
                //   It is an error if there are more than 7 values
                List<String> parameterValues =
                        getAttributeValues(gs.getAttribute(parameterName));
                if (parameterValues.size() > 7) {
                    throw new CpsException("Too many values found for "
                            + parameterName);
                }
                // Add on nulls to ensure there are 7 values when inserting
                parameterValues.addAll(null7);
                int valueOrder = 4;  // parameter values are args 4 - 10
                for (String parameterValue : parameterValues.subList(0, 7)) {
                    insertStmt.setString(valueOrder++, parameterValue);
                }
                // Add this parameter insert as one of the batch
                insertStmt.addBatch();
            }
            // execute to insert all the parameter values for this group
            int rowsInserted[] = insertStmt.executeBatch();
            insertStmt.close();

            // check all the returns to make sure each parameter was inserted
            for (int i = 0; i < rowsInserted.length; i++) {
                if (rowsInserted[i] == java.sql.Statement.EXECUTE_FAILED) {
                    throw new SQLException("Insert failed for "
                            + getParameterNames(gs).get(i));
                }
            }
        } catch (SQLException se) {
            throw se;
        } finally {
            this.close(null, insertStmt);
        }
    }

    /**
     * Populates the local hash map of IDs for BPF bias parameter names
     * @throws java.sql.SQLException for database errors
     */
    private void getBpfParameterIds(Connection conn)
            throws SQLException {

        if (bpfParameterIds != null) { // IDs have already been retrieved
            return;
        } 

        logger.trace("Loading BPF parameter IDs");
        bpfParameterIds = new HashMap<String, Long>(1 << 17);

        // get all the parameter names & IDs from the table
        String sql = "SELECT name, bpf_parameter_id FROM BPF_PARAMETER";
        PreparedStatement stmt = null;
        ResultSet rs = null;

        try {
            stmt = conn.prepareStatement(sql);
            stmt.setFetchSize(1000); // hint to reduce the db round trips
            rs = stmt.executeQuery();

            while (rs.next()) {
                bpfParameterIds.put(rs.getString(1), rs.getLong(2));
            }
        } catch (SQLException ex) {
            throw ex;
        } finally {
            this.close(rs, stmt);
        }
        logger.trace("Loaded " + bpfParameterIds.size()
                + " parameter IDs into hash map");
    }
}
