/**
 * BpfSummaryDAO
 *
 * Data Access Object for BPF summary data
 * This extends the CpfSummaryDAO class and overrides the methods that do the
 * database queries, changing table & field names as appropriate for BPF.
 * @author rockvam
 */

package gov.usgs.ias.dao.bpf;

import gov.usgs.ias.constant.Constants;
import gov.usgs.ias.cps.CpsSummary;
import gov.usgs.ias.cps.CpsStatusEnum;
import gov.usgs.ias.dao.cpf.*;
import gov.usgs.ias.dao.DataConnection;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.Calendar;
import java.util.TimeZone;
import org.apache.log4j.Logger;

public class BpfSummaryDAO extends CpfSummaryDAO {

    private static final Logger logger = Logger.getLogger(Constants.LOGGER_BPF);
    private final String bpfFrom = "FROM bpf_main bm ";

    /**
     * Constructor for the BpfDAO class.
     * @param conn   The DataConnection to communicate with the database.
     * @param daoAuth  The name of the property containing authorization
     *                 (username/password)
     */
    public BpfSummaryDAO(final DataConnection conn, final String daoAuth) {
        super(conn, daoAuth);
    }

    /* Override the method from the superclass that doesn't specify a sensor */
    @Override
    public List<CpsSummary> searchCpsMain(final Date startDate,
            final Date endDate,
            final Integer version, final CpsStatusEnum status, 
            boolean nearMatch)  throws SQLException {
        return searchCpsMain(startDate, endDate, (BpfSensorEnum)null, 
                version, status, nearMatch);
    }
    
    /**
     * Gets a list of all available BPF's for a given date.
     * @param startDate   The date to search for.
     * @param endDate     The end date for doing a date range search.
     * @param sensor      The sensor to search on.
     * @param version     The version number to return.
     * @param status      The status of BPFs to return, or null for any status
     * @param nearMatch   The flag of whether to search dates near the dates
     *                    given or not.
     * @return            A list of BPF's for the specified date.
     * @throws            java.sql.SQLException for database errors
     */
    public List<CpsSummary> searchCpsMain(final Date startDate,
            final Date endDate, final BpfSensorEnum sensor, 
            final Integer version, final CpsStatusEnum status, 
            boolean nearMatch)  throws SQLException {

        List<CpsSummary> availableBpfs = null;
        if (startDate == null) {
            throw new SQLException("No start date provided to query");
        }

        StringBuilder sb = new StringBuilder();
        sb.append("SELECT bm.file_name, bm.version_number, bm.status ");
        sb.append(this.bpfFrom);

        if (endDate == null) {
            // WHERE startDate BETWEEN bm.effective_date_begin
            //                 AND bm.effective_date_end
            sb.append("WHERE ? BETWEEN bm.effective_date_begin " +
                    "AND bm.effective_date_end ");
        } else {
            // WHERE bm.effective_date_begin <= endDate
            //   AND bm.effective_date_end   >= startDate
            sb.append("WHERE bm.effective_date_begin <= ? " +
                    "AND bm.effective_date_end >= ? ");
        }

        if (version != null) {
            sb.append("AND bm.version_number = ? ");
            //we can't search via nearMatch if there is a version number.
            if (nearMatch) {
                logger.debug("Disabling nearMatch search because version "
                        + "was specified");
                nearMatch = false;
            }
        }

        if (sensor != null) {
            sb.append("AND bm.sensor_id = ? ");
        }

        if (status != null) {
            sb.append("AND bm.status = ? ");
        }

        sb.append("ORDER BY bm.effective_date_begin DESC, " +
                "bm.version_number DESC");
        logger.debug("Date Range Query string is: " + sb.toString());

        PreparedStatement stmt = null;
        ResultSet rs = null;
        // Get a GMT calendar so timestamp conversions won't use local TZ
        Calendar gmtCalendar = Calendar.getInstance(
                TimeZone.getTimeZone("GMT-0:00"));

        try {
            Connection conn = this.getConn().getConnection(this.getDaoAuth());
            stmt = conn.prepareStatement(sb.toString());
            int bindCount = 0;

            if (endDate == null) {
                stmt.setTimestamp(++bindCount, timestampToSql(startDate),
                    gmtCalendar);
                logger.trace(" query on start date: " 
                        + timestampToSql(startDate));
            } else {
                stmt.setTimestamp(++bindCount, timestampToSql(endDate),
                    gmtCalendar);
                stmt.setTimestamp(++bindCount, timestampToSql(startDate),
                    gmtCalendar);
                logger.trace(" query on date range: " 
                        + timestampToSql(startDate)
                        + " - " + timestampToSql(endDate));
            }

            if (version != null) {
                stmt.setInt(++bindCount, version);
                logger.trace(" query on version: " + version);
            }

            if (sensor != null) {
                stmt.setString(++bindCount, sensor.toString());
                logger.trace(" query on sensor: " + sensor.toString());
            }

            if (status != null) {
                stmt.setString(++bindCount, status.toString());
                logger.trace(" query on status: " + status.toString());
            }

            rs = stmt.executeQuery();

            availableBpfs = parseResultSet(rs);
            logger.debug("Results: found " + availableBpfs.size());

            // if the previous query didn't get anything and the customer wants
            // a near match, run the nearMatch query.
            if (((null == availableBpfs) || (availableBpfs.isEmpty()))
                    && (nearMatch)) {
                logger.debug("No exact date range match found, " +
                        "doing near match...");
                this.close(rs, stmt); // close the previous stmt & result set
                rs = nearMatchCpsMain(startDate, endDate, sensor, status, 
                        conn, stmt);
                availableBpfs = parseResultSet(rs);
            }
        } finally {
            this.close(rs, stmt);
        }

        return availableBpfs;
    }

    /**
     * Gets a list of all available BPF's for a given orbit range.
     * If only startOrbit is supplied, only the BPF for that orbit is returned.
     * If a start and end orbit are supplied, all BPFs that overlap that
     * orbit range are returned.
     * @param startOrbit   The orbit to search for.
     * @param endOrbit     The end date for doing an orbit range search.
     * @param sensor       The sensor to search on.
     * @param version      The version number to return.
     * @param status       The status of BPFs to return, or null for any status
     * @param nearMatch    If true and the given orbit does not exist, return
     *                     the latest BPF in the orbit prior to the given orbit
     *                     if one exists.
     * @return             A list of BPF's for the specified date.
     * @throws             java.sql.SQLException for database errors
     */
    public List<CpsSummary> searchCpsMain(final Long startOrbit,
            final Long endOrbit, final BpfSensorEnum sensor, 
            final Integer version, final CpsStatusEnum status, 
            boolean nearMatch) throws SQLException {

        List<CpsSummary> availableBpfs = null;
        if (startOrbit == null) {
            throw new SQLException("No start orbit provided to query");
        }

        StringBuilder sb = new StringBuilder();
        sb.append("SELECT bm.file_name, bm.version_number, "
            + "bm.status, bm.orbit_begin ");
        sb.append(this.bpfFrom);
        // WHERE bm.orbit_begin >= startOrbit AND bm.orbit_begin <= endOrbit
        sb.append("WHERE bm.orbit_begin >= ? AND bm.orbit_begin <= ? ");

        if (version != null) {
            sb.append("AND bm.version_number = ? ");
            //we can't search via nearMatch if there is a version
            if (nearMatch) {
                logger.debug("Disabling nearMatch search because version " +
                        "was specified");
            }
            nearMatch = false;
        }

        if (sensor != null) {
            sb.append("AND bm.sensor_id = ? ");
        }

        if (status != null) {
            sb.append("AND bm.status = ? ");
        }

        sb.append("ORDER BY bm.effective_date_begin DESC, "
                + "bm.version_number DESC");
        logger.debug("Orbit Range Query string is: " + sb.toString());

        PreparedStatement stmt = null;
        ResultSet rs = null;
        try {
            stmt = this.getConn().getConnection(this.getDaoAuth()).
                    prepareStatement(sb.toString());

            int bindCount = 0;
            if (endOrbit == null) {
                stmt.setLong(++bindCount, startOrbit);
                stmt.setLong(++bindCount, startOrbit);
            } else {
                stmt.setLong(++bindCount, startOrbit);
                stmt.setLong(++bindCount, endOrbit);
            }
            if (version != null) {
                stmt.setInt(++bindCount, version);
            }
            if (sensor != null) {
                stmt.setString(++bindCount, sensor.toString());
            }
            if (status != null) {
                stmt.setString(++bindCount, status.toString());
            }

            rs = stmt.executeQuery();
            availableBpfs = parseResultSet(rs);

            // If the availableBpfs is empty and the user wants a near match,
            // run the nearMatch query.
            if (((null == availableBpfs) || (availableBpfs.isEmpty()))
                    && (nearMatch)) {
                logger.debug("No exact orbit match found, doing near match...");
                this.close(rs); // Close the previous result set
                rs = this.nearMatchCpsMain(startOrbit, endOrbit, sensor, status,
                        stmt);
                availableBpfs = parseResultSet(rs);
            }

        } finally {
            this.close(rs, stmt);
        }

        return availableBpfs;
    }

    /**
     * Gets summary information given a BPF ID
     * @param bpfMainId   The ID of the BPF.
     * @return            A CpsSummary object for the given BPF
     * @throws            java.sql.SQLException for database errors
     */
    @Override
    public CpsSummary getCpsSummary(final Long bpfMainId) throws SQLException {

        CpsSummary bpf = null;
        if (bpfMainId == null) {
            logger.warn("No BPF ID supplied");
            return bpf;
        }

        PreparedStatement stmt = null;
        ResultSet rs = null;
        StringBuilder sql = new StringBuilder();
        sql.append("SELECT bm.file_name, bm.file_source, bm.version_number, ");
        sql.append("bm.effective_date_begin, bm.effective_date_end, " +
                "bm.date_created, ");
        sql.append("bm.orbit_begin, bm.date_entered, " +
                "bm.description, bm.status ");
        sql.append(this.bpfFrom);
        sql.append("WHERE bm.bpf_main_id = ?");

        try {
            stmt = this.getConn().getConnection(this.getDaoAuth()).
                    prepareStatement(sql.toString());
            stmt.setLong(1, bpfMainId);

            rs = stmt.executeQuery();

            while (rs.next()) {
                bpf = this.resultSetToCpsSummary(rs);
            }

        } finally {
            this.close(rs, stmt);
        }

        return bpf;
    }

    /**
     * Gets a list of summaries of all available BPF's.
     * @param sensor   Return BPFs for the given sensor.
     * @return A list CpsSummary objects
     * @throws java.sql.SQLException for database errors
     */
    public List<CpsSummary> getAllCpsList(final BpfSensorEnum sensor) 
            throws SQLException {

        List<CpsSummary> bpfs = new LinkedList<CpsSummary>();
        PreparedStatement stmt = null;
        ResultSet rs = null;
        StringBuilder sql = new StringBuilder();
        sql.append("SELECT bm.file_name, bm.file_source, bm.version_number, ");
        sql.append("bm.effective_date_begin, bm.effective_date_end, " +
                "bm.date_created, ");
        sql.append("bm.orbit_begin, bm.date_entered, bm.description, " +
                "bm.status ");
        sql.append(this.bpfFrom);
        if (null != sensor){
            sql.append("WHERE bm.sensor_id = ? ");
        }        
        sql.append("ORDER BY bm.file_name");
        
        try {
            stmt = this.getConn().getConnection(this.getDaoAuth()).
                    prepareStatement(sql.toString());
            if (sensor != null) {
                stmt.setString(1, sensor.toString());
            }
            rs = stmt.executeQuery();

            while (rs.next()) {
                bpfs.add(this.resultSetToCpsSummary(rs));
            }
        } catch (SQLException ex) {
            throw ex;
        } finally {
            this.close(rs, stmt);
        }

        return bpfs;
    }

    /**
     * Gets the BPF ID for a BPF by name.
     * @param filename   Name of the BPF.
     * @return           ID of the given named BPF
     * @throws           java.sql.SQLException for database errors
     */
    @Override
    public Long getCpsMainId(final String filename) throws SQLException {

        Long bpfMainId = null;
        if (filename == null) {
            logger.warn("No BPF file name supplied");
            return bpfMainId;
        }

        PreparedStatement stmt = null;
        ResultSet rs = null;
        String sql = "SELECT bpf_main_id FROM bpf_main WHERE file_name = ?";

        try {
            stmt = this.getConn().getConnection(this.getDaoAuth()).
                    prepareStatement(sql);
            stmt.setString(1, filename);

            rs = stmt.executeQuery();

            while (rs.next()) {
                bpfMainId = rs.getLong("bpf_main_id");
            }
        } catch (SQLException ex) {
            throw ex;
        } finally {
            this.close(rs, stmt);
        }

        return bpfMainId;
    }

    /**
     * Update the status of a BPF in the database.
     * @param bpfName   Name of the BPF.
     * @param status    New status to assign to the BPF.
     * @throws          java.sql.Exception for database errors or invalid status
     *                  provided
     */
    @Override
    public void updateCpsStatus(String bpfName, String status) 
            throws Exception {
        
        if (bpfName == null || status == null) {
            logger.warn("Name and/or status not supplied -"
                    + " no status update done");
            return;
        }
        String sql = "UPDATE bpf_main SET status = ? WHERE file_name = ?";
        PreparedStatement stmt = this.getConn().getConnection(
                this.getDaoAuth()).prepareStatement(sql);
        // make sure it's a valid status
        CpsStatusEnum s = CpsStatusEnum.valueOf(status);
        stmt.setString(1, s.toString());
        stmt.setString(2, bpfName);

        if (stmt.execute()) {
            //this is unexpected; an update statement should not return a result
            logger.warn("Update statement returned a result set");
        } else {
            logger.info("BPF status update changed " +
                    stmt.getUpdateCount() + " BPF records");
        }
        this.close(stmt);
        this.getConn().commit();
    }

    /**
     * Do a "near match" search by date range.  This allows finding a BPF if the
     * requested date range is after (within tolerance) an existing BPF.
     * 
     * For a single date search, the normal search looks for startDate between
     * a BPF's start & end dates:
     * <pre>
     *                     |---------------X---------------|
     *  effective_date_start           startDate           effective_date_end
     *</pre>
     * The near match search will allow finding a BPF whose end date is within
     * NEAR_MATCH_BPF_TOLERANCE_DAYS (T) of startDate:
     * <pre>
     *                     |---------------X---------------------|   X
     *  effective_date_start     startDate-T    effective_date_end   startDate
     *</pre>
     * 
     * For a date range search, the normal search looks for BPFs that overlap
     * the input date range (effective_date_end > startDate and
     * effective_date_start < endDate).  See CpsSummaryDAO.nearMatchCpsMain
     * for more information on overlapping date ranges.
     *
     * The near match search on a startDate-endDate range effectively enlarges 
     * the query range by backing up the startDate by the tolerance.  This
     * enlarged query range is then checked to see if any BPFs overlap it.
     * 
     * @param startDate   The date to search for.
     * @param endDate     The end date for doing a date range search.
     * @param sensor      The sensor to search on
     * @param status      The status of BPFs to return, or null for any status
     * @param conn        The open database connection
     * @return            A list of BPF's for the specified date.
     * @throws            java.sql.SQLException for database errors
     */
    private ResultSet nearMatchCpsMain(final Date startDate, final Date endDate,
            final BpfSensorEnum sensor,
            final CpsStatusEnum status, Connection conn,
            PreparedStatement stmt) throws SQLException {

        ResultSet rs = null;
        Calendar gmtCalendar = Calendar.getInstance(
            TimeZone.getTimeZone("GMT-0:00"));

        // Get the near match days out of the system properties.
        String nearMatchPropertyValue
            = System.getProperty(Constants.NEAR_MATCH_BPF_TOLERANCE_DAYS);
        if (nearMatchPropertyValue == null) {
            throw new SQLException(
                "BPF near match date tolerance property not set");
        }
        Integer nearMatchDays = null;
        try {
            nearMatchDays = Integer.parseInt(nearMatchPropertyValue);
        } catch (Exception e) {
            // Issue warning and leave the value null
            logger.warn("Unable to parse "
                    + Constants.NEAR_MATCH_BPF_TOLERANCE_DAYS 
                    + " property value " + nearMatchPropertyValue + " : "
                    + e.getMessage());
        }
        if (nearMatchDays == null) {
            throw new SQLException("BPF near match date tolerance property " +
                    "is not set correctly");
        }
        logger.debug("Doing near match date query: within " +
            nearMatchDays + " days");

        StringBuilder sql = new StringBuilder();
        sql.append("SELECT bm.file_name, ");
        sql.append("bm.version_number, bm.status ");
        sql.append(this.bpfFrom);

        if (null == endDate) {
            // WHERE bm.effective_date_end BETWEEN (startDate - nearMatchDays)
            //                             AND startDate
            sql.append("WHERE bm.effective_date_end BETWEEN (?-?) AND ? ");
        } else {
            // WHERE bm.effective_date_begin <= endDate
            //   AND bm.effective_date_end   >= (startDate - nearMatchDays)
            sql.append("WHERE bm.effective_date_begin <= ? ");
            sql.append("AND bm.effective_date_end >= (?-?) ");
        }
        if (null != sensor){
            sql.append("AND bm.sensor_id = ? ");
        }
        if (null != status){
            sql.append("AND bm.status = ? ");
        }

        sql.append("ORDER BY bm.effective_date_begin DESC, "
                + "bm.version_number DESC ");
        logger.debug("NearMatch Date Range Query string is: " + sql.toString());
        stmt = conn.prepareStatement(sql.toString());

        int bindCount = 0;
        if (endDate == null) {
            stmt.setTimestamp(++bindCount, timestampToSql(startDate),
                gmtCalendar);
            stmt.setInt(++bindCount, nearMatchDays);
            stmt.setTimestamp(++bindCount, timestampToSql(startDate),
                gmtCalendar);
        } else {
            stmt.setTimestamp(++bindCount, timestampToSql(endDate),
                gmtCalendar);
            stmt.setTimestamp(++bindCount, timestampToSql(startDate),
                gmtCalendar);
            stmt.setInt(++bindCount, nearMatchDays);
        }
        if (sensor != null) {
            stmt.setString(++bindCount, sensor.toString());
        }
        if (status != null) {
            stmt.setString(++bindCount, status.toString());
        }

        rs = stmt.executeQuery();
        return rs;
    }

    /**
     * Do a "near match" search by orbit range.  This allows finding a BPF if
     * the requested orbit range is after (within tolerance) an existing BPF.
     * If the original search was on a single orbit, the near match search
     *  will be on the range [(startOrbit - tolerance + 1), startOrbit]
     * If the original search was on an orbit range, the near match search
     *  will be on the range [(startOrbit - tolerance + 1), endOrbit]
     * @param startOrbit  The orbit to search for (or starting orbit, if
     *                    querying on a range).
     * @param endOrbit    The end orbit for doing a range search.
     * @param sensor      The sensor to search on.
     * @param status      A valid status, or null for any status
     * @param stmt        The prepared statement constructed for the original
     *                    query
     * @return            A list of BPF's for the specified date.
     * @throws            java.sql.SQLException for database errors
     */
    private ResultSet nearMatchCpsMain(final Long startOrbit,
            final Long endOrbit, final BpfSensorEnum sensor,
            final CpsStatusEnum status,
            PreparedStatement stmt) throws SQLException {
        
        String nearMatchPropertyValue
                = System.getProperty(Constants.NEAR_MATCH_BPF_TOLERANCE_ORBIT);
        if (nearMatchPropertyValue == null) {
            throw new SQLException(
                "BPF near match orbit tolerance property not set");
        }
        Integer nearMatchOrbit = null;
        try {
            nearMatchOrbit = Integer.parseInt(nearMatchPropertyValue);
        } catch (Exception e) {
            // Issue warning and leave the value null
            logger.warn("Unable to parse "
                    + Constants.NEAR_MATCH_BPF_TOLERANCE_ORBIT
                    + " property value " + nearMatchPropertyValue + " : "
                    + e.getMessage());
        }
        if (nearMatchOrbit == null){
            throw new SQLException("BPF near match orbit tolerance property " +
                    "is not set correctly");
        }
        logger.debug("Doing near match orbit query: within " +
                nearMatchOrbit + " orbits of " + startOrbit);

        // The query for the orbit near match is the same, just the parameters
        //     are different.
        // We want to use the orbit before the startOrbit.
        ResultSet rs = null;
        int bindCount = 0;
        // Clear the parameters from when this query was run for the
        //  non-nearMatch.
        stmt.clearParameters();

        if (endOrbit == null) {
            stmt.setLong(++bindCount, startOrbit - nearMatchOrbit);
            stmt.setLong(++bindCount, startOrbit);
            logger.debug("Using orbit " + (startOrbit - nearMatchOrbit) +
                    " - " + startOrbit + " in query");
        } else {
            stmt.setLong(++bindCount, startOrbit - nearMatchOrbit);
            stmt.setLong(++bindCount, endOrbit);
            logger.debug("Using orbit " + (startOrbit - nearMatchOrbit) +
                    " - " + endOrbit + " in query");
        }
        if (sensor != null) {
            stmt.setString(++bindCount, sensor.toString());
        }
        if (status != null) {
            stmt.setString(++bindCount, status.toString());
        }

        rs = stmt.executeQuery();
        return rs;
    }
}
