/**
 * CpfDAO
 *
 * Data Access Object for CPF content data (groups and attributes)
 */
package gov.usgs.ias.dao.cpf;

import gov.usgs.ias.cps.CpsGroup;
import gov.usgs.ias.cps.CpsException;
import gov.usgs.ias.cps.CpsStatusEnum;
import gov.usgs.ias.dao.AttributeMap;
import gov.usgs.ias.dao.BaseDAO;
import gov.usgs.ias.dao.DataConnection;
import gov.usgs.ias.constant.Constants;
import org.apache.log4j.Logger;
import java.sql.*;
import oracle.sql.STRUCT;
import oracle.sql.ARRAY;
import oracle.sql.StructDescriptor;
import oracle.sql.ArrayDescriptor;
import java.util.*;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import gov.nasa.pds.tools.label.*;
import java.security.MessageDigest;

public class CpfDAO extends BaseDAO {

    private static final Logger logger = Logger.getLogger(Constants.LOGGER_CPF);
    
    /**
     * Constructor for the CpfDAO class.
     * @param conn     The DataConnection to communicate with the database.
     * @param daoAuth  The name of the property containing authorization
     *                 (username/password)
     */
    public CpfDAO(final DataConnection conn, final String daoAuth) {
        super(conn, daoAuth);
    }

    /**
     * Creates CpsGroup objects for each group in a CPF.
     * @param cpfName   The name of the CPF.
     * @return          A collection of CpsGroup objects.
     * @throws          java.sql.SQLException  for database errors
     */
    public Collection<CpsGroup> getCpsGroups(String cpfName) 
            throws SQLException {
        Collection<CpsGroup> groups = new ArrayList<CpsGroup>();

        // Join the CF main table with the group table (using the
        // cross-reference cpf_main_group table) to get the names and IDs of
        // all the groups that belong to this CPF.
        String sql = "SELECT cpfg.name, cpfg.cpf_group_id, cpfg.checksum, " +
                "cpfmg.group_order " +
                "FROM cpf_main cpfm " +
                "INNER JOIN cpf_main_group cpfmg " +
                "ON cpfm.cpf_main_id = cpfmg.cpf_main_id " +
                "INNER JOIN cpf_group cpfg " +
                "ON cpfmg.cpf_group_id = cpfg.cpf_group_id " +
                "WHERE cpfm.file_name = ? " +
                "ORDER BY cpfmg.group_order";

        PreparedStatement stmt = null;
        ResultSet rs = null;

        try {
            stmt = this.getConn().getConnection(this.getDaoAuth()).
                    prepareStatement(sql);
            stmt.setString(1, cpfName);
            rs = stmt.executeQuery();

            while (rs.next()) {
                groups.add(this.cpsGroupsFromResultSet(rs));
            }
        } catch (SQLException ex) {
            throw ex;
        } finally {
            this.close(rs, stmt);
        }

        return groups;
    }

    /**
     * Loads the CpsGroup class with data from the ResultSet generated from
     * executing the query in getCpsGroupDataStmt
     * @param rs   ResultSet with the data to load into the new CpsGroup object.
     * @return     The new CpsGroup object, or null if the result set was null.
     * @throws     java.sql.SQLException for errors accesing result set fields
     */
    protected CpsGroup cpsGroupsFromResultSet(final ResultSet rs)
        throws SQLException {
        if (rs == null) {
            return null;
        }

        CpsGroup returnGroup = new CpsGroup();
        returnGroup.setName(rs.getString(1));
        returnGroup.setGroupId(rs.getLong(2));
        returnGroup.setChecksum(rs.getString(3));
        returnGroup.setGroupOrder(rs.getInt(4));
        return returnGroup;
    }

    /**
     * Generates the PreparedStatement for getting the contents of a CPF group
     * out of the database
     * @return The statement ready for execution (needs the cpf_group_id bound
     * to 1).
     * @throws java.sql.SQLException  for database errors
     */
    public PreparedStatement getCpsGroupDataStmt() throws SQLException {
        PreparedStatement stmt = null;

        // Join the parameter table with the parameter name and parameter values
        // tables using a given group ID to get the names, IDs, and values
        // (sorted in order) of all the parameters that belong to the group.
        String sql = "SELECT cpfgp.cpf_group_parameter_id as parameter_id, " +
                "cpfp.name, cpfgp.parameter_order, cv.value_order, cv.value " +
                "FROM cpf_group_parameter cpfgp " +
                "INNER JOIN cpf_parameter_values cv " +
                "ON cpfgp.cpf_group_parameter_id = cv.cpf_group_parameter_id " +
                "INNER JOIN cpf_parameter cpfp " +
                "ON cpfgp.cpf_parameter_id = cpfp.cpf_parameter_id " +
                "WHERE cpfgp.cpf_group_id = ? " +
                "ORDER BY cpfgp.parameter_order, cv.value_order";

        stmt = this.getConn().getConnection(this.getDaoAuth()).
                prepareStatement(sql);
        stmt.setFetchSize(1000); // to reduce database round-trips
        return stmt;
    }

    /**
     * Retreives a CpsGroup from the database.
     * @param cpfGroupId      The cpf_group_id to use as the root for retrieval.
     * @param addParameters   Should the parameters be included in the return
     *                        object?
     * @return                The CpsGroup object with data loaded.
     * @throws                SQLException for database errors
     */
    public CpsGroup getCpsGroupById(final long cpfGroupId,
            final boolean addParameters) throws SQLException {
        CpsGroup returnGroup = null;

        // Query the summary information about this group.  Even though there is
        // only one group in question, the query includes name, id, checksum and
        // order fields to make the result set processable by the
        // cpsGroupsFromResultSet method.
        String sql = "SELECT cpfg.name, cpfg.cpf_group_id, cpfg.checksum, " +
                "0 as group_order " +
                "FROM cpf_group cpfg " +
                "WHERE cpfg.cpf_group_id = ? ";

        PreparedStatement stmt = null;
        ResultSet rs = null;
        try {
            stmt = this.getConn().getConnection(this.getDaoAuth()).
                    prepareStatement(sql);
            stmt.setLong(1, cpfGroupId);
            rs = stmt.executeQuery();

            while (rs.next()) {
                returnGroup = this.cpsGroupsFromResultSet(rs);
            }

            if (addParameters) {
                this.getGroupParameters(returnGroup);
            }
        } catch (SQLException ex) {
            throw ex;
        } finally {
            this.close(rs, stmt);
        }

        return returnGroup;
    }

    /**
     * Queries the database to fill in the parameters of a CpsGroup
     * @param group   A CPF Group (only the group ID need be set on input)
     * @throws        SQLException for database errors
     */
    protected void getGroupParameters(final CpsGroup group)
            throws SQLException {
        if (group == null) return;
        AttributeMap parameters = new AttributeMap();
        PreparedStatement stmt = null;
        ResultSet rs = null;

        Long paramId = null;
        Long prevParamId = null;
        String paramName = null;
        String prevParamName = null;
        String paramValue = null;
        int paramValueCnt = 0;
        StringBuilder paramValues = new StringBuilder();

        try {
            stmt = this.getCpsGroupDataStmt();
            stmt.setLong(1, group.getGroupId());
            rs = stmt.executeQuery();

            // The query returns on row for each value of all the
            // parameters in this group, sorted by parameter name and the
            // correct order of the values.  Loop through all the rows,
            // building a string that contains the values for the given
            // parameter. Every time a parameter ID changes, the paramValues
            // string is reset.
            while (rs.next()) {
                paramId = rs.getLong("parameter_id");
                paramName = rs.getString("name");
                paramValue = rs.getString("value");

                // Found a new parameter.  Save the previous parameter
                // name & value to the parameters hash.  Then reset the
                // value string & count in preparation of the new parameter.
                if (!paramId.equals(prevParamId)) {
                    if (paramValueCnt > 0) {
                        parameters.put(prevParamName, paramValues.toString());
                    }
                    paramValues.delete(0, paramValues.length() + 1);
                    paramValueCnt = 0;
                }

                // Parameter with multiple values, concatenate the values
                // with commas
                if (paramValueCnt > 0) {
                    paramValues.append(", ");
                }

                // Append this value onto the value string for this parameter
                paramValues.append(paramValue);
                paramValueCnt++;

                prevParamId = paramId;
                prevParamName = paramName;
            }

            // Put the final parameter into the hash
            parameters.put(prevParamName, paramValues.toString());
            group.setParameters(parameters);
        } catch (SQLException ex) {
            throw ex;
        } finally {
            this.close(rs, stmt);
        }
    }

    /**
     * Inserts a CPF into the database
     * @param cpfLabel   The parsed ODL label structure for the CPF
     * @param cpfStatus  The initital status of the file in the database.
     * @return           the database id of the CPF inserted
     * @throws           java.sql.SQLException for database errors
     */
    public int insertFile(Label cpfLabel, CpsStatusEnum cpfStatus)
            throws Exception {
        String DB_PACKAGE = "PROCESS_CPF_PKG";
        final String procErrString = "Error executing CPF stored procedure";
        int cpfId = -1;  // The ID of the CPF that is being inserted
        int status;      // Stored procedure execution status
        int parm;        // Stored procedure parameter index
        Calendar gmtCalendar = Calendar.getInstance(
            TimeZone.getTimeZone("GMT-0:00"));

        // Extract metadata about the CPF from the file itself
        logger.trace("Checking required fields");
        String cpfSpacecraft = getFileAttribute(cpfLabel, "Spacecraft_Name");
        String cpfSensor; // Will be populated later
        String cpfFileName = getFileAttribute(cpfLabel, "File_Name");
        String cpfCollection = getFileAttribute(cpfLabel, "Collection_Number");
        String cpfVersion = getFileAttribute(cpfLabel, "Version");
        String cpfSource = getFileAttribute(cpfLabel, "File_Source");
        String cpfDescr = getFileAttribute(cpfLabel, "Description");

        // Replace all tabs, newlines, multiple spaces, and carriage returns
        // to prevent irregularities when retrieving the CPFs
        cpfDescr = cpfDescr.replaceAll("[\\t\\n\\r\\s]+", " ");
        if (!setFileAttribute(cpfLabel, "Description", cpfDescr))
        {
            throw new CpsException("Unable to set the CPF description " +
                "in the FileAttributes group");
        }

        String effDateBegin = getFileAttribute(cpfLabel, 
                "Effective_Date_Begin");
        String effDateEnd = getFileAttribute(cpfLabel, "Effective_Date_End");
        if (cpfSpacecraft == null || cpfFileName == null || cpfVersion == null
                || cpfCollection == null || cpfSource == null
                || cpfDescr == null || effDateBegin == null
                || effDateEnd == null ) {
            throw new CpsException("CPF is missing required FILE_ATTRIBUTES "
                    + "attributes (Spacecraft_Name, File_Name, "
                    + "Collection_Number, Version, File_Source, Description, "
                    + "Effective_Date_Begin, Effective_Date_End)");
        }

        // Check the spacecraft is the expected value and then set the sensor
        if (cpfSpacecraft.compareToIgnoreCase("Landsat_8") != 0)
        {
            throw new CpsException("CPF spacecraft value is not the expected "
                         + "value of Landsat_8");
        }
        // Hard-code the sensor (Landsat_8 has only OLI_TIRS CPFs)
        cpfSensor = "OLI_TIRS";
        cpfSpacecraft = cpfSpacecraft.toUpperCase();

        CallableStatement mainStatement = null;
        CallableStatement groupCheckStatement = null;
        CallableStatement buildBodyStatement = null;
        DateFormat df = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
        df.setTimeZone(TimeZone.getTimeZone("GMT-0:00"));

        // Get the DB connection (will throw an exception if connection can't
        //  be made).  Transactions are enabled.
        Connection conn = this.getConn().getConnection(this.getDaoAuth());
        
        // Once the connection is made, catch excpetions so we can rollback if
        // necessary before bailing out
        try {
            // Storing a CPF is a 3-step process.
            //  Before beginning, get descriptors for the complex Oracle object
            //  types required by the stored procedures.
            //  It is slicker to use a typemap (see previous revisions
            //  of this file and CpfGroupUDT.java etc.) but it is faster
            //  to use the Oracle structure descriptor directly.
            ArrayDescriptor parameterValueArrayTypeDescriptor 
                    = ArrayDescriptor.createDescriptor("CPF_VALUES_ARRAY",
                    conn);
            ArrayDescriptor parameterArrayTypeDescriptor 
                    = ArrayDescriptor.createDescriptor("CPF_PARAMETER_ARRAY",
                    conn);
            StructDescriptor groupNameTypeDescriptor 
                    = StructDescriptor.createDescriptor("CPF_GROUP_NAME",
                    conn);
            StructDescriptor parameterValuesTypeDescriptor 
                    = StructDescriptor.createDescriptor("CPF_PARM_VALUES",
                    conn);
            StructDescriptor parameterNameTypeDescriptor 
                    = StructDescriptor.createDescriptor("CPF_PARAMETER_NAME",
                    conn);

            // Set up statements for the stored procedures
            mainStatement = conn.prepareCall("{call " + DB_PACKAGE
                    + ".proc_insert_cpf_main (?,?,?,?,?,?,?,?,?,?,?)}");
            groupCheckStatement = conn.prepareCall("{? = call " + DB_PACKAGE
                    + ".check_if_group_exists(?,?,?)}");
            buildBodyStatement = conn.prepareCall("{call " + DB_PACKAGE
                    + ".proc_build_cpf_body(?,?,?,?)}");

            // First, insert the main record
            logger.trace("Inserting main record");
            parm = 1;
            mainStatement.setString(parm++, cpfFileName);
            mainStatement.setString(parm++, cpfSource);
            mainStatement.setString(parm++, cpfDescr);
            mainStatement.setTimestamp(parm++,timestampToSql(
                df.parse(effDateBegin)), gmtCalendar);
            mainStatement.setTimestamp(parm++,timestampToSql(
                df.parse(effDateEnd)), gmtCalendar);
            mainStatement.setInt(parm++, Integer.parseInt(cpfVersion));
            mainStatement.setString(parm++, cpfStatus.toString());
            mainStatement.setString(parm++, cpfSensor);
            mainStatement.setString(parm++, cpfSpacecraft);
            mainStatement.setInt(parm++, Integer.parseInt(cpfCollection));
            mainStatement.setInt(parm, -1); // initialize the output parameter
            mainStatement.registerOutParameter(parm, Types.INTEGER);
            status = mainStatement.executeUpdate();
            if (status != 1) {
                throw new SQLException(procErrString + 
                        " to insert main record for " + cpfFileName);
            }
            cpfId = mainStatement.getInt(parm);
            if (cpfId == -1) {
                throw new SQLException("Error inserting main record for "
                        + cpfFileName);
            }
            logger.debug("Inserted cpf_main, record_id is " + cpfId);

            // Second, add each group to the CPF
            List groupList = cpfLabel.getGroups();
            Iterator gi = groupList.iterator();
            GroupStatement gs = null;
            int groupId = -1;
            int groupIndex = 1;
            STRUCT[] parms = null;
            while (gi.hasNext()) {
                gs = (GroupStatement) gi.next();
                logger.trace("Inserting Group " + groupIndex + ": "
                        + gs.getIdentifier());

                // Before constructing the group, calculate a checksum
                String cksum = getGroupCheckSum(gs);
                logger.debug(" Inserting group " + gs.getIdentifier()
                        + " with checksum " + cksum);

                // Check if the group exists.  If it does, don't bother to build
                // the group structure as it will be ignored anyway.
                ARRAY pvArr = null;
                parm = 1;
                groupCheckStatement.registerOutParameter(parm++, Types.INTEGER);
                groupCheckStatement.setString(parm++, gs.getIdentifier());
                groupCheckStatement.setString(parm++, cksum);
                groupCheckStatement.setInt(parm++, cpfId);
                status = groupCheckStatement.executeUpdate();
                if (status != 1) {
                    throw new SQLException(procErrString + " to check if group "
                            + gs.getIdentifier() + " exists");
                }
                groupId = groupCheckStatement.getInt(1);
                if (groupId > 0) {
                    logger.trace(" Group already exists: " + groupId);
                } else {
                    logger.trace(" Constructing new group");
                    parms = getParameterNameStructs(gs, conn, 
                            parameterValuesTypeDescriptor,
                            parameterValueArrayTypeDescriptor,
                            parameterNameTypeDescriptor);
                    pvArr = new ARRAY(parameterArrayTypeDescriptor,
                            conn, parms);
                }

                // Set up the elements of the CPF_GROUP_NAME type
                Object[] groupNameAttributes
                        = new Object[groupNameTypeDescriptor.getLength()];
                groupNameAttributes[0] = gs.getIdentifier(); // group name
                groupNameAttributes[1] = groupIndex;         // group index
                groupNameAttributes[2] = cksum;              // checksum
                groupNameAttributes[3] = pvArr;              // parameter array
                STRUCT pnObj = new STRUCT(groupNameTypeDescriptor, conn,
                        groupNameAttributes);

                // Call the proc to insert the group using the STRUCT object
                parm = 1;
                buildBodyStatement.setInt(parm++, cpfId);
                buildBodyStatement.setString(parm++, cpfFileName);
                buildBodyStatement.setObject(parm++, pnObj, Types.STRUCT);
                buildBodyStatement.setInt(parm, -1); // initialize the output
                buildBodyStatement.registerOutParameter(parm, Types.INTEGER);
                status = buildBodyStatement.executeUpdate();
                if (status != 1) {
                    throw new SQLException(procErrString + "to insert group "
                            + gs.getIdentifier());
                }
                // On success, the returned group ID is always 1
                groupId = buildBodyStatement.getInt(parm);
                if (groupId == -1) {
                    throw new SQLException("Error adding group "
                            + gs.getIdentifier());
                }
                logger.trace(" ... group complete");
                groupIndex++;
            }

            // Finally, commit the transaction
            conn.commit();
            logger.info("Completed CPF, ID is " + cpfId);

        // Propagate any exceptions thrown
        } catch (Exception e) {
            try {
                conn.rollback();
            } catch (SQLException ex) {
                logger.warn("Error rolling back transaction", ex);
            }
            throw e;
        } finally {
            this.close(mainStatement);
            this.close(groupCheckStatement);
            this.close(buildBodyStatement);
        }

        return cpfId;
    }

    /**
     * Gets the value of a parameter (attribute) within the
     * FILE_ATTRIBUTES group of a parsed ODL label.
     * @param lab             parsed ODL label
     * @param attributeName   name of the attribute whose value you want
     * @return                the entire attribute value
     */
    public String getFileAttribute(Label lab, String attributeName) {
        Scalar scalarValue = null;
        try {
            scalarValue = (Scalar)((GroupStatement)lab.
                    getGroups("FILE_ATTRIBUTES").get(0)).
                    getAttribute(attributeName).getValue();
        } catch (Exception e) {
            return null;
        }
        return scalarValue.getValue();
    }

    /**
     * Sets the value of a parameter (attribute) within the
     * FILE_ATTRIBUTES group of a parsed ODL label.
     * @param lab             parsed ODL label
     * @param attributeName   name of the attribute whose value you want
     * @param scalarVaue      value of the attribute
     * @return                void
     */
    public boolean setFileAttribute(Label lab, String attributeName,
        String scalarValue) {
        TextString attr = new TextString(scalarValue);
        try {
            ((GroupStatement)lab.getGroups("FILE_ATTRIBUTES").get(0)).
                getAttribute(attributeName).setValue(attr);
        } catch (Exception e) {
            return false;
        }
        return true;
    }

    /**
     * Create an Oracle STRUCT array containing the information for a given ODL
     * group.  Note that the elements of an Oracle ARRAY or STRUCT are called
     * "attributes", and so are the elements of an ODL Group statement.  Try
     * not to be confused.
     * @param gs        The ODL group statement
     * @param conn      The open database connection, required when setting up
     *                  Oracle type objects
     * @param parameterValuesTypeDescriptor       The CPF_PARM_VALUES structure 
     *                                            descriptor
     * @param parameterValueArrayTypeDescriptor   The CPF_VALUES_ARRAY array
     *                                            descriptor
     * @param parameterNameTypeDescriptor         The CPF_PARAMETER_NAME 
     *                                            structure descriptor
     * @return          An array of Oracle STRUCT objects containing the
     *                  information from the ODL group
     */
    protected STRUCT[] getParameterNameStructs(GroupStatement gs,
            Connection conn,
            StructDescriptor parameterValuesTypeDescriptor,
            ArrayDescriptor parameterValueArrayTypeDescriptor,
            StructDescriptor parameterNameTypeDescriptor) throws SQLException {
        // Get all the attributes in this ODL group
        List attributeList = gs.getAttributes();
        STRUCT[] parms = new STRUCT[attributeList.size()];
        int attrIndex = 1;

        for (Object o : attributeList) {

            AttributeStatement as = (AttributeStatement) o;
            ArrayList<String> al = new ArrayList<String>();

            // Get all the values for this attribute and put into an Oracle
            //   array of STRUCTs
            al.addAll(getAttributeValues(as));
            STRUCT[] values = new STRUCT[al.size()];
            // Get the # of attributes in a CPF_PARM_VALUES type (may throw
            //   exception)
            int pvLength = parameterValuesTypeDescriptor.getLength();
            Object[] parameterValuesAttributes = new Object[pvLength];
            // Create each values element as a STRUCT of type
            //   CPF_PARM_VALUES
            for (int i = 0; i < al.size(); i++) {
                parameterValuesAttributes[0] = al.get(i);    // value
                parameterValuesAttributes[1] = i + 1;        // value order
                values[i] = new STRUCT(parameterValuesTypeDescriptor, conn,
                        parameterValuesAttributes);
            }

            // Create a CPF_VALUES_ARRAY array containing the
            //   CPF_PARM_VALUES values
            ARRAY avArr = null;
            avArr = new ARRAY(parameterValueArrayTypeDescriptor, conn, values);
            // Put the CPF_VALUES_ARRAY into a CPF_PARAMETER_NAME
            //   structure
            int pnLength = parameterNameTypeDescriptor.getLength();
            if (pnLength < 3) {
                throw new SQLException("Unexpected number of fields ("
                        + pnLength
                        + ") in CPF_PARAMETER_NAME type, should be 3");
            }
            Object[] parameterNameAttributes = new Object[pnLength];
            parameterNameAttributes[0] = as.getIdentifier(); // name
            parameterNameAttributes[1] = attrIndex;          // name order
            parameterNameAttributes[2] = avArr;    // value (a CPF_VALUES_ARRAY)
            // the following may throw an exception
            parms[attrIndex - 1] = new STRUCT(parameterNameTypeDescriptor, conn,
                    parameterNameAttributes);
            attrIndex++;
        }
        return parms;
    }

    /**
     * Returns the names of the parameters within a group
     * @param gs   A parsed ODL group statement
     * @return     Array of strings containing the parameter names
     */
    protected List<String> getParameterNames(GroupStatement gs) {
        // Get the names of all the attributes in this group
        List la = gs.getAttributes();
        List<String> parameterNames = new ArrayList<String>();
        for (Object o : la) {
           AttributeStatement as = (AttributeStatement)o;
           parameterNames.add(as.getIdentifier());
        }
        return parameterNames;
    }

    /**
     * Get the values from a given ODL attribute
     * @param as   The parsed AttributeStatement
     * @return     a List of (String) attribute values (even if the attribute
     * has a single value)
     */
    protected static List<String> getAttributeValues(AttributeStatement as) {
        ArrayList<String> avList = new ArrayList<String>();

        if (as ==  null) {
            return avList;
        }
        
        Scalar scalarValue;
        // Get all the scalar values out of a sequence or set
        if (as.getValue().getClass().getName().contains("Sequence")) {
            Sequence seq = (Sequence)as.getValue();
            Iterator si = seq.iterator();
            while (si.hasNext()) {
                scalarValue = (Scalar)si.next();
                avList.add(scalarValue.getValue());
            }
        } else if (as.getValue().getClass().getName().contains("Set")) {
            gov.nasa.pds.tools.label.Set set =
                    (gov.nasa.pds.tools.label.Set)as.getValue();
            Iterator si = set.iterator();
            while (si.hasNext()) {
                scalarValue = (Scalar)si.next();
                avList.add(scalarValue.getValue());
            }
        // The attribute just contains a single scalar
        } else {
            scalarValue = (Scalar)as.getValue();
            avList.add(scalarValue.getValue());
        }

        return avList;
    }

    /**
     * Calculate a checksum based on the values contained in the ODL group
     * @param gs   The parsed ODL Group Statement
     * @return     a checksum string
     */
    private String getGroupCheckSum(GroupStatement gs) throws Exception {
        // Get the names & values of all the attributes in this group
        ArrayList<String> allValuesList = new ArrayList<String>();
        List attributeList = gs.getAttributes();
        for (Object o : attributeList) {
            // Get the name and all the values for this attribute
            AttributeStatement as  = (AttributeStatement)o;
            allValuesList.add(as.getIdentifier());
            allValuesList.addAll(getAttributeValues(as));
        }

        // Count the total length of the all values
        Iterator<String> allValuesIterator = allValuesList.iterator();
        int totalLength = 0;
        while (allValuesIterator.hasNext()) {
            totalLength += allValuesIterator.next().length();
        }

        // Copy each value's byte array into one long single byte array
        byte[] allGroupValues = new byte[totalLength];
        allValuesIterator = allValuesList.iterator();
        String oneValue;
        int currentStart = 0;
        while (allValuesIterator.hasNext()) {
            oneValue = allValuesIterator.next();
            System.arraycopy(oneValue.getBytes(), 0, allGroupValues,
                    currentStart, oneValue.length());
            currentStart += oneValue.length();
        }

        // Calculate MD5 hash & convert to a checksum string
        MessageDigest md5 = MessageDigest.getInstance("MD5");
        md5.update(allGroupValues, 0, allGroupValues.length);
        byte[] hashBytes = md5.digest();
        String result = "";
        for (int i = 0; i < hashBytes.length; i++) {
            // Convert each byte to a 2-digit hex character string
            //  In bit operations, bytes are converted to signed ints.
            //  First, convert to a positive int using 0xff.
            //  Then make it 3 hex digits using 0x100
            //    and take the last 2 hex digits (to ensure the result
            //      has a leading zero if its value is < 0x10 )
            result += Integer.toString((hashBytes[i] & 0xff) + 0x100, 16)
                    .substring(1);
        }
        return result;
    }
}
