/**
 * CpfSummaryDAO
 * 
 * Data Access Object for CPF summary information.  This is basically metadata
 * about a CPF including file name, version, status, and optionally
 * effective start & end date, date entered into the db, date created, file
 * source, and description.
 */
package gov.usgs.ias.dao.cpf;

import gov.usgs.ias.constant.Constants;
import gov.usgs.ias.cps.CpsSummary;
import gov.usgs.ias.cps.CpsStatusEnum;
import gov.usgs.ias.dao.BaseDAO;
import gov.usgs.ias.dao.DataConnection;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.Calendar;
import java.util.TimeZone;
import org.apache.log4j.Logger;

public class CpfSummaryDAO extends BaseDAO {

    private static final Logger logger = Logger.getLogger(Constants.LOGGER_CPF);

    /**
     * Constructor for the CpfDAO class.
     * @param conn     The DataConnection to communicate with the database.
     * @param daoAuth  The name of the authorization property containing
     *                 username/password to use when connecting
     */
    public CpfSummaryDAO(final DataConnection conn, final String daoAuth) {
        super(conn, daoAuth);
    }

    /**
     * Gets a list of all available CPF's for a given date.
     * @param startDate   The date to search for.
     * @param endDate     The end date for doing a date range search.
     * @param version     The version number to return.
     * @param status      The status of CPFs to return, or null for any status
     * @param nearMatch   The flag of whether to search dates near the dates
     *                    given or not.
     * @return            A list of CPF's for the specified criteria.
     * @throws            java.sql.SQLException for database errors
     */
    public List<CpsSummary> searchCpsMain(final Date startDate,
            final Date endDate, final Integer version,
            final CpsStatusEnum status, boolean nearMatch) throws SQLException {
        List<CpsSummary> availableCpfs = null;

        StringBuilder sb = new StringBuilder();
        sb.append("SELECT cm.file_name, cm.version_number, cm.status "
                + " FROM cpf_main cm ");

        if (endDate == null) {
            // WHERE startDate BETWEEN cm.effective_date_begin AND
            //       cm.effective_date_end
            sb.append("WHERE ? BETWEEN cm.effective_date_begin AND "
                    + "cm.effective_date_end ");
        } else {
            // WHERE cm.effective_date_begin <= endDate
            //   AND cm.effective_date_end >= startDate
            sb.append("WHERE cm.effective_date_begin <= ? "
                    + "AND cm.effective_date_end >= ? ");
        }

        if (version != null) {
            sb.append("AND cm.version_number = ? ");
            // we can't search via nearMatch if there is a version number.
            if (nearMatch) {
                logger.debug("Disabling nearMatch search because version "
                        + "was specified");
                nearMatch = false;
            }
        }

        if (status != null) {
            sb.append("AND cm.status = ? ");
        }

        sb.append("ORDER BY cm.effective_date_begin DESC, "
                + "cm.version_number DESC");

        logger.debug("Query is: " + sb.toString());
        PreparedStatement stmt = null;
        ResultSet rs = null;
        // Get a GMT calendar so timestamp conversions won't use local TZ
        Calendar gmtCalendar = Calendar.getInstance(
                TimeZone.getTimeZone("GMT-0:00"));

        try {
            Connection conn = this.getConn().getConnection(this.getDaoAuth());
            stmt = conn.prepareStatement(sb.toString());
            int bindCount = 0;

            if (endDate == null) {
                stmt.setTimestamp(++bindCount, timestampToSql(startDate),
                        gmtCalendar);
                logger.debug("Adding start date to query: "
                        + timestampToSql(startDate));
            } else {
                stmt.setTimestamp(++bindCount, timestampToSql(endDate),
                        gmtCalendar);
                stmt.setTimestamp(++bindCount, timestampToSql(startDate),
                        gmtCalendar);
                logger.debug("Adding to query: end date "
                        + timestampToSql(endDate) + " and start date  "
                        + timestampToSql(startDate));
            }

            if (version != null) {
                stmt.setInt(++bindCount, version);
            }

            if (status != null) {
                stmt.setString(++bindCount, status.toString());
            }

            rs = stmt.executeQuery();
            availableCpfs = parseResultSet(rs);

            // if the previous query didn't get anything and the customer wants
            // a near match, run the nearMatch query.
            if (((null == availableCpfs) || (availableCpfs.isEmpty()))
                    && (nearMatch)) {
                this.close(rs); // Close the previous result set
                rs = nearMatchCpsMain(startDate, endDate, status, conn);
                availableCpfs = parseResultSet(rs);
            }
        } finally {
            this.close(rs, stmt);
        }

        return availableCpfs;
    }

    /**
     * Loads a CpfSummary object from the database.
     * @param rs   The data to load.
     * @return     The CPF summary data.
     * @throws     java.sql.SQLException for errors accesing result set fields
     */
    protected CpsSummary resultSetToCpsSummary(final ResultSet rs)
            throws SQLException {
        if (rs == null) {
            return null;
        }

        CpsSummary cpf = new CpsSummary();
        cpf.setFilename(rs.getString("file_name"));
        cpf.setVersion(rs.getString("version_number"));
        cpf.setStatus(CpsStatusEnum.valueOf(rs.getString("status")));

        // Orbit is one of the fields in the CpsSummary, although it is only
        // populated for BPFs.  For completeness we attempt to get it from
        // the query results.
        try {
            cpf.setBeginOrbit(rs.getLong("orbit_begin"));
        } catch (SQLException ignore) {
            logger.trace("orbit_begin was not set.");
        }

        try {
            cpf.setFileSource(rs.getString("file_source"));
        } catch (SQLException ignore) {
            logger.trace("file_source was not set.");
        }

        try {
            cpf.setDescription(rs.getString("description"));
        } catch (SQLException ignore) {
            logger.trace("description was not set.");
        }

        try {
            cpf.setEffectiveBeginDate(
                    rs.getTimestamp("effective_date_begin"));
        } catch (SQLException ignore) {
            logger.trace("effective_date_begin was not set.");
        }

        try {
            cpf.setEffectiveEndDate(rs.getTimestamp("effective_date_end"));
        } catch (SQLException ignore) {
            logger.trace("effective_date_end was not set.");
        }

        try {
            cpf.setDateCreated(rs.getTimestamp("date_created"));
        } catch (SQLException ignore) {
            logger.trace("date_created was not set.");
        }

        try {
            cpf.setDateEntered(rs.getTimestamp("date_entered"));
        } catch (SQLException ignore) {
            logger.trace("date_entered was not set.");
        }

        return cpf;
    }

    /**
     * Gets summary information given a CPF ID
     * @param cpfMainId     The ID of the CPF.
     * @return              A CpfSummary object for the given CPF, null if
     *                      the ID is not found
     * @throws              java.sql.SQLException for database errors
     */
    public CpsSummary getCpsSummary(final Long cpfMainId) throws SQLException {
        if (cpfMainId == null) {
            return null;
        }
        CpsSummary cpf = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
        StringBuilder sql = new StringBuilder();

        sql.append("SELECT cm.file_name, cm.file_source, cm.version_number, "
                + "cm.effective_date_begin, cm.effective_date_end, "
                + "cm.date_created, cm.date_entered, cm.description, cm.status "
                + "FROM cpf_main cm WHERE cm.cpf_main_id = ?");

        try {
            stmt = this.getConn().getConnection(this.getDaoAuth()).
                    prepareStatement(sql.toString());
            stmt.setLong(1, cpfMainId);

            rs = stmt.executeQuery();

            while (rs.next()) {
                cpf = this.resultSetToCpsSummary(rs);
            }

        } catch (SQLException ex) {
            throw ex;
        } finally {
            this.close(rs, stmt);
        }

        return cpf;
    }

    /**
     * Gets a list of summaries of all available CPF's.
     * @return A list CpfSummary objects
     * @throws java.sql.SQLException for database errors
     */
    public List<CpsSummary> getAllCpsList() throws SQLException {
        List<CpsSummary> cpfs = new LinkedList<CpsSummary>();
        PreparedStatement stmt = null;
        ResultSet rs = null;
        StringBuilder sql = new StringBuilder();
        CpsSummary tempCpf = null;

        sql.append("SELECT cm.file_name, cm.file_source, cm.version_number, "
                + "cm.effective_date_begin, cm.effective_date_end, "
                + "cm.date_created, cm.date_entered, cm.description, cm.status "
                + "FROM cpf_main cm ORDER BY cm.file_name");

        try {
            stmt = this.getConn().getConnection(this.getDaoAuth()).
                    prepareStatement(sql.toString());
            rs = stmt.executeQuery();
            while (rs.next()) {
                tempCpf = this.resultSetToCpsSummary(rs);
                cpfs.add(tempCpf);
            }
        } catch (SQLException ex) {
            throw ex;
        } finally {
            this.close(rs, stmt);
        }

        return cpfs;
    }

    /**
     * Gets the CPF ID for a CPF by name.
     * @param filename   Name of the CPF.
     * @return           CPF ID, null if not found
     * @throws           java.sql.SQLException for database errors
     */
    public Long getCpsMainId(final String filename) throws SQLException {
        if (filename == null) {
            return null;
        }
        Long cpfMainId = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;

        String sql = "SELECT cpf_main_id FROM cpf_main WHERE file_name = ?";

        try {
            stmt = this.getConn().getConnection(this.getDaoAuth()).
                    prepareStatement(sql);
            stmt.setString(1, filename);
            rs = stmt.executeQuery();
            while (rs.next()) {
                cpfMainId = rs.getLong("cpf_main_id");
            }
        } catch (SQLException ex) {
            throw ex;
        } finally {
            this.close(rs, stmt);
        }

        return cpfMainId;
    }

    /**
     * Update the status of a CPF in the database.
     * @param cpfName   Name of the CPF.
     * @param status    New status to assign to the CPF.
     * @throws Exception for database errors or if status is not a valid status
     */
    public void updateCpsStatus(String cpfName, String status)
            throws Exception {
        if (cpfName == null || status == null) {
            return;
        }
        String sql = "UPDATE cpf_main SET status = ? WHERE file_name = ?";
        PreparedStatement stmt = this.getConn().
                getConnection(this.getDaoAuth()).prepareStatement(sql);

        // make sure the status passed in is valid
        CpsStatusEnum s = CpsStatusEnum.valueOf(status);
        stmt.setString(1, s.toString());
        stmt.setString(2, cpfName);

        if (stmt.execute()) {
            //this is unexpected; an update statement should not return a result
            logger.warn("Update statement returned a result set");
        } else {
            logger.info("CPF update " + cpfName + " to status "
                + s.toString() + " : updated " + stmt.getUpdateCount()
                + " records");
        }
        this.close(stmt);
        this.getConn().commit();
    }

    /**
     * Do a "near match" search by date range.  This allows finding a CPF if the
     * requested date range is after (within tolerance) an existing CPF.
     *
     * For a single date search, the normal search looks for startDate between
     * a CPF's start & end dates:
     * <PRE>
     *                     |---------------X---------------|
     *  effective_date_start           startDate           effective_date_end
     * </PRE>
     * The near match search will allow finding a CPF whose end date is within
     * NEAR_MATCH_CPF_TOLERANCE (T) of startDate (between startDate-T and
     * startDate):
     * <PRE>
     *                     |---------------X---------------------|   X
     *  effective_date_start     startDate-T    effective_date_end   startDate
     * </PRE>
     *
     * For a date range search, the normal search looks for CPFs that overlap
     * the input date range (effective_date_end > startDate and
     * effective_date_start &lt; endDate).  Thus the CPF shown below will match
     * queries A, B, C but not D or E.
     * <PRE>
     *  effective_date_start |----------------------------| effective_date_end
     * A             X==============X
     * B                               X===========X
     * C                                             X============X
     * D  X==============X
     * E                                                           X===========X
     * </PRE>
     * The near match search on a startDate-endDate range finds a CPF if it
     * overlaps the date range, after backing up the startDate by the tolerance
     * (if eff_date_end > startDate-T and eff_date_start < endDate.
     * <PRE>
     *               |---------------X---------------|   X          X
     *  eff_date_start     startDate-T    eff_date_end   startDate  endDate
     * </PRE>
     * @param startDate   The date to search for.
     * @param endDate     The end date for doing a date range search.
     * @param status      The status of CPFs to return, or null for any status
     * @param conn        The open database connection
     * @return            A list of CPF's for the specified date range.
     * @throws            java.sql.SQLException for database errors
     */
    private ResultSet nearMatchCpsMain(final Date startDate, final Date endDate,
            final CpsStatusEnum status, Connection conn) throws SQLException {
        ResultSet rs = null;
        PreparedStatement stmt = null;
        Calendar gmtCalendar = Calendar.getInstance(
                TimeZone.getTimeZone("GMT-0:00"));

        // Get the near match days out of the system properties.
        String nearMatchPropertyValue
            = System.getProperty(Constants.NEAR_MATCH_CPF_TOLERANCE);
        if (nearMatchPropertyValue == null) {
            throw new SQLException("CPF near match tolerance property not set");
        }
        Integer nearMatchDays = null;
        try {
            nearMatchDays = Integer.parseInt(nearMatchPropertyValue);
        } catch (Exception e) {
            // Issue warning and leave the value null
            logger.warn("Unable to parse "
                    + Constants.NEAR_MATCH_CPF_TOLERANCE
                    + " property value " + nearMatchPropertyValue + " : "
                    + e.getMessage());
        }
        if (nearMatchDays == null) {
            throw new SQLException("CPF near match tolerance property " +
                    "is not set correctly");
        }

        StringBuilder sql = new StringBuilder();

        sql.append("SELECT cm.file_name, cm.version_number, cm.status ");
        sql.append("FROM cpf_main cm ");

        if (null == endDate) {
            // WHERE cm.effective_date_end BETWEEN (startDate - nearMatchDays)
            //                             AND startDate
            sql.append("WHERE cm.effective_date_end BETWEEN (?-?) AND ? ");
        } else {
            // WHERE cm.effective_date_begin <= endDate
            //   AND cm.effective_date_end   >= (startDate - nearMatchDays)
            sql.append("WHERE cm.effective_date_begin <= ? ");
            sql.append("AND cm.effective_date_end >= (?-?) ");
        }

        if (null != status) {
            sql.append("AND cm.status = ? ");
        }

        sql.append("ORDER BY cm.effective_date_begin DESC, " +
                "cm.version_number DESC ");
        logger.trace("Query string is: " + sql.toString());

        stmt = conn.prepareStatement(sql.toString());
        int bindCount = 0;

        if (endDate == null) {
            stmt.setTimestamp(++bindCount, timestampToSql(startDate),
                    gmtCalendar);
            stmt.setInt(++bindCount, nearMatchDays);
            stmt.setTimestamp(++bindCount, timestampToSql(startDate),
                    gmtCalendar);
            logger.trace(" query on start date: " + startDate.toString()
                    + " with near match " + nearMatchDays);
        } else {
            stmt.setTimestamp(++bindCount, timestampToSql(endDate),
                    gmtCalendar);
            stmt.setTimestamp(++bindCount, timestampToSql(startDate),
                    gmtCalendar);
            stmt.setInt(++bindCount, nearMatchDays);
            logger.trace(" query on date range: " + startDate.toString()
                    + " - " + endDate.toString()
                    + " with near match " + nearMatchDays);
        }

        if (status != null) {
            stmt.setString(++bindCount, status.toString());
            logger.trace(" query on status: " + status.toString());
        }

        rs = stmt.executeQuery();

        return rs;
    }

    /**
     * Parse the result set into a list of CPS Summaries
     * @param rs   ResultSet containing the CPF group data
     * @return     A list of CPF Summary objects
     * @throws     java.sql.SQLException for result set errors
     */
    protected List<CpsSummary> parseResultSet(ResultSet rs)
            throws SQLException {

        CpsSummary tempCps = null;
        List<CpsSummary> availableCpfs = new LinkedList<CpsSummary>();

        while (rs.next()) {
            tempCps = this.resultSetToCpsSummary(rs);

            availableCpfs.add(tempCps);
        }
        return availableCpfs;
    }
}
