/**
 * RlutDAO
 *
 * Data Access Object for RLUT related data.
 * This extends the CpfDAO class and overrides the methods that do the database
 * queries, changing table & field names as appropriate for RLUT.
 */
package gov.usgs.ias.dao.rlut;

import gov.usgs.ias.constant.Constants;
import gov.usgs.ias.dao.BaseDAO;
import gov.usgs.ias.dao.DataConnection;
import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import org.apache.log4j.Logger;

public class RlutDAO extends BaseDAO {

    // The default buffer size (1MB) used when caching an RLUT file
    private static final int BUFFER_SIZE = 1048576;
    private static final Logger logger =
            Logger.getLogger(Constants.LOGGER_RLUT);

    /**
     * Constructor for the RlutDAO class.
     * @param conn      The DataConnection to communicate with the database.
     * @param daoAuth  The name of the property containing authorization
     *                 (username/password)
     */
    public RlutDAO(final DataConnection conn, final String daoAuth) {
        super(conn, daoAuth);
    }

    /**
     * Retreives an RLUT from the database and puts it on the disk cache.
     * @param rlutFilename   The name of the RLUT file to retrieve.
     * @param cacheFile      The file object that hold the rlut file on the
     *                       cache.
     * @throws SQLException if there is a problem running the SQL statement.
     * @throws IOException if the rlut file cannot be written to disk cache.
     */
    public void getRlutData(final String rlutFilename, File cacheFile)
            throws SQLException, IOException {
        String sql = "SELECT rm.rlut_data "
                + "FROM rlut_main rm "
                + "WHERE rm.file_name = ? ";

        logger.debug("Query is: " + sql);
        PreparedStatement stmt = null;
        ResultSet rs = null;
        BufferedInputStream bis = null;
        FileOutputStream fos = null;
        int bytes_read;

        try {
            stmt = this.getConn().getConnection(this.daoAuthorizationProperty).
                    prepareStatement(sql);
            stmt.setString(1, rlutFilename);
            rs = stmt.executeQuery();
            //write the data from the query to a file on the disk cache
            if (rs.next()) {
                bis = new BufferedInputStream(rs.getBinaryStream(1));
                fos = new FileOutputStream(cacheFile);
                byte[] bufferedData = new byte[BUFFER_SIZE];
                while ((bytes_read = bis.read(bufferedData)) > -1) {
                    fos.write(bufferedData, 0, bytes_read);
                }
            } else {
                // the file was not found
                throw new SQLException(rlutFilename
                        + " does not exist in the database");
            }
        } catch (SQLException se) {
            // In this case, we didn't even attempt to write the file to cache
            throw se;
        } catch (IOException ioe) {
            // In this case, cleanup anything that was written to cache.
            if (cacheFile.exists()) {
                fos.close();
                fos = null;
                cacheFile.delete();
            }
            throw ioe;
        } finally {
            this.close(rs, stmt);
            try {
                if (bis != null) {
                    bis.close();
                }
                if (fos != null) {
                    fos.close();
                }
            } catch (IOException ex) {
                logger.warn("Could not close the buffered input stream or "
                        + "the file output stream. " + ex.getMessage());
            }

            try {
                if (cacheFile.setReadOnly()) {
                    logger.debug("Successfully set cache file to read-only");
                } else {
                    logger.warn("Could not set cache file to read-only");
                }
            } catch (SecurityException se) {
                logger.warn("Could not set cache file to read-only");
            }
        }
    }
}
