/**
 * RlutSummaryDAO
 *
 * Data Access Object for RLUT summary information.
 * This extends the CpfSummaryDAO class and overrides the methods that do the
 * database queries, changing table & field names as appropriate for RLUT.
 */
package gov.usgs.ias.dao.rlut;

import gov.usgs.ias.constant.Constants;
import gov.usgs.ias.cps.CpsStatusEnum;
import gov.usgs.ias.cps.CpsSummary;
import gov.usgs.ias.dao.DataConnection;
import gov.usgs.ias.dao.cpf.CpfSummaryDAO;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.Calendar;
import java.util.TimeZone;
import org.apache.log4j.Logger;

/**
 *
 * @author bschroeder
 */
public class RlutSummaryDAO extends CpfSummaryDAO {

    private static final Logger logger =
            Logger.getLogger(Constants.LOGGER_RLUT);

    /**
     * Constructor for the RlutDAO class.
     * @param conn      The DataConnection to communicate with the database.
     * @param daoAuth  The name of the property containing authorization
     *                 (username/password)
     */
    public RlutSummaryDAO(final DataConnection conn, final String daoAuth) {
        super(conn, daoAuth);
    }

    /**
     * Gets a list of all available RLUT's for a given date.
     * @param startDate   The date to search for.
     * @param endDate     The end date for doing a date range search.
     * @param version     The version number to return.
     * @param status      Return only RLUTs with this status.
     * @param nearMatch   The flag of whether to search dates near the dates
     *                    given or not.  RLUT does not implement this
     *                    functionality at this time; the parameter is included
     *                    for compatibility with the parent class.
     * @return            A list of RLUT's for the specified date.
     * @throws            java.sql.SQLException for database errors or no start
     *                    date provided to search on
     */
    @Override
    public List<CpsSummary> searchCpsMain(final Date startDate,
            final Date endDate, final Integer version,
            final CpsStatusEnum status, boolean nearMatch) throws SQLException {
        List<CpsSummary> availableRluts = new LinkedList<CpsSummary>();

        if (startDate == null) {
            throw new SQLException("StartDate cannot be null when calling "
                    + "searchCpsMain(startDate, endDate, version, status, "
                    + "nearMatch");
        }

        StringBuilder sb = new StringBuilder();

        sb.append("SELECT rm.file_name, rm.version_number, rm.status "
                + "FROM rlut_main rm ");

        if (endDate == null) {
            sb.append("WHERE ? BETWEEN rm.effective_date_begin AND "
                    + "rm.effective_date_end ");
        } else {
            sb.append("WHERE rm.effective_date_begin <= ? "
                    + "AND rm.effective_date_end >= ? ");
        }

        if (version != null && version > 0) {
            sb.append("AND rm.version_number = ? ");
        }

        if (status != null) {
            sb.append("AND rm.status = ? ");
        }

        sb.append("ORDER BY rm.effective_date_begin DESC, "
                + "rm.version_number DESC");
        logger.debug("searchCpsMain date range query is: " + sb.toString());

        PreparedStatement stmt = null;
        ResultSet rs = null;
        CpsSummary tempRlut = null;
        Calendar gmtCalendar = Calendar.getInstance(
                TimeZone.getTimeZone("GMT-0:00"));

        try {
            stmt = this.getConn().getConnection(this.getDaoAuth()).
                    prepareStatement(sb.toString());
            int bindCount = 0;

            if (endDate == null) {
                stmt.setTimestamp(++bindCount, timestampToSql(startDate),
                        gmtCalendar);
            } else {
                stmt.setTimestamp(++bindCount, timestampToSql(endDate),
                        gmtCalendar);
                stmt.setTimestamp(++bindCount, timestampToSql(startDate),
                        gmtCalendar);
            }

            if (version != null && version > 0) {
                stmt.setInt(++bindCount, version);
            }

            if (status != null) {
                stmt.setString(++bindCount, status.toString());
            }

            rs = stmt.executeQuery();

            while (rs.next()) {
                tempRlut = resultSetToCpsSummary(rs);

                availableRluts.add(tempRlut);
            }
        } finally {
            this.close(rs, stmt);
        }

        return availableRluts;
    }

    /**
     * Gets a list of all available RLUT's for a given date.
     * @param rlutMainId   ID of the RLUT file to retrieve information about.
     * @return             A summary of information about the RLUT
     * @throws             java.sql.SQLException for database errors
     */
    @Override
    public CpsSummary getCpsSummary(final Long rlutMainId) throws SQLException {
        if (rlutMainId == null) {
            return null;
        }
        CpsSummary rlut = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
        StringBuilder sql = new StringBuilder();

        sql.append("SELECT rm.file_name, rm.file_source, rm.version_number, ");
        sql.append("rm.effective_date_begin, rm.effective_date_end, "
                + "rm.date_created, ");
        sql.append("rm.date_entered, rm.description, rm.status ");

        sql.append("FROM rlut_main rm ");
        sql.append("WHERE rm.rlut_main_id = ?");

        logger.debug("getCpsSummary rlutMainId query is: " + sql.toString());
        try {
            stmt = this.getConn().getConnection(this.getDaoAuth()).
                    prepareStatement(sql.toString());
            stmt.setLong(1, rlutMainId);

            rs = stmt.executeQuery();

            while (rs.next()) {
                rlut = this.resultSetToCpsSummary(rs);
            }

        } finally {
            this.close(rs, stmt);
        }

        return rlut;
    }

    /**
     * Gets a list of summaries for all available RLUTs.
     * @return A list of summary of information about the RLUTs
     * @throws java.sql.SQLException for database errors
     */
    @Override
    public List<CpsSummary> getAllCpsList() throws SQLException {
        List<CpsSummary> rluts = new LinkedList<CpsSummary>();
        PreparedStatement stmt = null;
        ResultSet rs = null;
        StringBuilder sql = new StringBuilder();
        CpsSummary tempRlut = null;

        sql.append("SELECT rm.file_name, rm.file_source, rm.version_number, ");
        sql.append("rm.effective_date_begin, rm.effective_date_end, "
                + "rm.date_created, ");
        sql.append("rm.date_entered, rm.description, rm.status ");

        sql.append("FROM rlut_main rm ");
        sql.append("ORDER BY rm.file_name");

        logger.debug("getAllCpsList query is: " + sql.toString());

        try {
            stmt = this.getConn().getConnection(this.getDaoAuth()).
                    prepareStatement(sql.toString());

            rs = stmt.executeQuery();

            while (rs.next()) {
                tempRlut = this.resultSetToCpsSummary(rs);
                rluts.add(tempRlut);
            }
        } finally {
            this.close(rs, stmt);
        }

        return rluts;
    }

    /**
     * Gets the ID for a given RLUT.
     * @param filename   Name of the RLUT file.
     * @return           ID for the given RLUT
     * @throws           java.sql.SQLException for database errors
     */
    @Override
    public Long getCpsMainId(final String filename) throws SQLException {
        if (filename == null) {
            return null;
        }
        Long rlutMainId = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;

        String sql = "SELECT rlut_main_id FROM rlut_main WHERE file_name = ?";
        logger.debug("getCpsMainId(filename) query is: " + sql);

        try {
            stmt = this.getConn().getConnection(this.getDaoAuth()).
                    prepareStatement(sql);
            stmt.setString(1, filename);

            rs = stmt.executeQuery();

            while (rs.next()) {
                rlutMainId = rs.getLong("rlut_main_id");
            }
        } finally {
            this.close(rs, stmt);
        }

        return rlutMainId;
    }
}
