package gov.usgs.ias.gui.common_library;

import java.util.Calendar;
import java.util.Observable;
import java.util.Observer;
import java.text.DateFormatSymbols;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.GridLayout;
import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFormattedTextField;
import javax.swing.JPanel;
import javax.swing.JLabel;
import javax.swing.JFrame;
import javax.swing.JTextField;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * A dialog for a month's calendar showing the month and year at
 * the top, and  a grid of buttons for the days.
 */
public class CalendarDialog extends JDialog
        implements ActionListener, ChangeListener, Observer {

    private JButton[] dayButtons
            = new JButton[42];       // 6 weeks' worth of day buttons
    private WraparoundSpinnerListModel monthSpinnerModel;
    private SpinnerNumberModel yearSpinnerModel;
    private MonthList months;
    private DatePanel datePanel = null; // the date panel that this dialog is
                                     // selecting for
    private int month;               // the currently selected month,
    private int year;                //   year,
    private int day;                 //   and day
    private int savedDayButton;      // index of currently selected day button

    /**
     * Constructor for the dialog
     * @param parentFrame   the frame that this dialog is a child of
     */
    public CalendarDialog(JFrame parentFrame) {
        super(parentFrame, "Calendar", true);
        setLayout(new BorderLayout());
        JPanel mainPanel = new JPanel();
        mainPanel.setLayout(new BorderLayout());

        JPanel monthYearPanel = new JPanel(new GridLayout(1, 0));

        //Create a Month JSpinner
        months = new MonthList();
        monthSpinnerModel = new WraparoundSpinnerListModel(
            months.getMonthStrings());
        JSpinner monthSpinner = new JSpinner(monthSpinnerModel);
        formatSpinnerTextField(monthSpinner);
        monthSpinnerModel.addChangeListener(this);
        monthYearPanel.add(monthSpinner);

        // Create a Year JSpinner, initialized to the configured min/max range
        yearSpinnerModel = new SpinnerNumberModel(DateConfig.STARTYEAR,
                DateConfig.MINYEAR, DateConfig.MAXYEAR, 1);
        JSpinner yearSpinner = new JSpinner(yearSpinnerModel);
        yearSpinner.setEditor(new JSpinner.NumberEditor(yearSpinner, "#"));
        formatSpinnerTextField(yearSpinner);
        yearSpinnerModel.addChangeListener(this);
        monthYearPanel.add(yearSpinner);
        
        mainPanel.add(monthYearPanel, BorderLayout.NORTH);

        JPanel calendarPanel = new JPanel(new GridLayout(1, 0));
        JLabel[] daysOfWeekLabel = new JLabel[7];
        String[] daysOfWeek
                = DateFormatSymbols.getInstance().getShortWeekdays();
        int dow = Calendar.getInstance().getFirstDayOfWeek();
        for (int x = 0; x < daysOfWeekLabel.length; x++, dow++) {
            if (dow >= daysOfWeek.length) {
                dow -= 7;
            }
            daysOfWeekLabel[x] = new JLabel(" ", JLabel.CENTER);
            daysOfWeekLabel[x].setText(daysOfWeek[dow]);
            calendarPanel.add(daysOfWeekLabel[x]);
        }
        mainPanel.add(calendarPanel, BorderLayout.CENTER);

        JPanel dayPanel = new JPanel(new GridLayout(6, 7));
        for (int x = 0; x < dayButtons.length; x++) {
            dayButtons[x] = new JButton("00"); // size the button for 2 digits
            dayButtons[x].addActionListener(this);
            dayButtons[x].setBackground(Color.WHITE);
            dayPanel.add(dayButtons[x]);
        }

        JPanel buttonPanel = new JPanel();
        JButton okButton = new JButton("Ok");
        okButton.setMnemonic(KeyEvent.VK_O);
        okButton.addActionListener(this);
        buttonPanel.add(okButton);

        JButton closeButton = new JButton("Cancel");
        closeButton.setMnemonic(KeyEvent.VK_C);
        closeButton.addActionListener(this);
        buttonPanel.add(closeButton);

        add(mainPanel, BorderLayout.NORTH);
        add(dayPanel, BorderLayout.CENTER);
        add(buttonPanel, BorderLayout.SOUTH);
        pack();

        ObservableDate.getInstance().addObserver(this);
    }

    /**
     * Format the text fields of spinners consistently
     * @param spinner The JSpinner to format its text field
     */
    private void formatSpinnerTextField(JSpinner spinner) {
        JComponent editor = spinner.getEditor();
        JFormattedTextField ftf;
        ftf = ((JSpinner.DefaultEditor)editor).getTextField();
        if (ftf != null) {
            if (editor instanceof JSpinner.ListEditor) {
                //For the month list, want to increase the width also
                ftf.setColumns(6);  //specify more width, default is very narrow
            }
            //In all cases, insert some padding and left-align
            ftf.setBorder(BorderFactory.createCompoundBorder(ftf.getBorder(),
                    BorderFactory.createEmptyBorder(1,5,1,5)));
            ftf.setHorizontalAlignment(JTextField.LEFT);
        }
    }

    /**
     * Set the DatePanel whose date is being set by this dialog
     * @param panel
     */
    public void setPanel(DatePanel panel) {
        datePanel = panel;
    }

    /**
     * Get the DatePanel whose date is being set by this dialog
     * @return  the DatePanel
     */
    public DatePanel getPanel() {
        return datePanel;
    }

    /**
     * Set the labels on the day buttons to reflect the currently selected
     * month and year
     */
    private void setDays() {
        // Clear the previously set text on the buttons
        for (int x = 0; x < dayButtons.length; x++) {
            dayButtons[x].setText("");
        }

        // Get the day of the 1st of the month and the length of the month
        Calendar cal = Calendar.getInstance();
        cal.clear();
        cal.set(year, month, 1);
        int dayOfWeek = cal.get(Calendar.DAY_OF_WEEK);
        int firstDayOfWeek = cal.getFirstDayOfWeek();
        int startIndex = dayOfWeek - firstDayOfWeek;
        int daysInMonth = cal.getActualMaximum(Calendar.DAY_OF_MONTH);
        for (int x = startIndex, days = 1; days <= daysInMonth; x++, days++) {
            dayButtons[x].setText("" + days);
        }

        // Reset the current day if it is beyond the end of the new month
        if (day > daysInMonth) {
            day = daysInMonth;
        }

        findDay();
    }

    /**
     * Highlights the button of the currently selected day
     */
    private void findDay() {
        dayButtons[savedDayButton].setBackground(Color.WHITE);

        for (int x = 0; x < dayButtons.length; x++) {
            String text = dayButtons[x].getText();
            if (!text.equals("")) {
                int buttonNumber = Integer.parseInt(text);
                if (buttonNumber == day) {
                    dayButtons[x].setBackground(Color.YELLOW);
                    savedDayButton = x;
                    break;
                }
            }
        }
    }

    /**
     * Event handler for the ActionListener interface
     * @param e  The event that occurred
     */
    @Override
    public void actionPerformed(ActionEvent e) {
        String command = e.getActionCommand();
        
        if (command.equals("Cancel")) {
            // Ignoring the date selection, so clear association with the
            // DatePanel to be updated
            setPanel(null);
            setVisible(false);
        } else if (command.equals("Ok")) {
            ObservableDate.getInstance().setCurrentDate(year, month, day);
            setVisible(false);
        } else if (!command.matches("")) {
             // Handle clicks in populated day buttons
            try {
                day = Integer.parseInt(command);
                findDay();
            } catch (Exception ex) {
                // If we couldn't parse a day number, don't do anything
            }
        }
    }

    /**
     * The handler for the Observer interface
     * @param o     the observed object
     * @param arg   any argument provided when the observed object sent the
     *              notification
     */
    @Override
    public void update(Observable o, Object arg) {
        ObservableDate observableDate = (ObservableDate)o;
        year = observableDate.getSelectedYear();
        month = observableDate.getSelectedMonth();
        day = observableDate.getSelectedDay();
        yearSpinnerModel.setValue(year);
        monthSpinnerModel.setValue(months.getMonthStrings()[month]);
        setDays();
    }

    @Override
    public void stateChanged(ChangeEvent ce) {
        Object o = ce.getSource();
        if (o instanceof SpinnerNumberModel) {
            year = yearSpinnerModel.getNumber().intValue();
        } else {
            month = months.getMonthIndex(monthSpinnerModel.getValue()
                .toString());
        }
        setDays();
    }
}
