package gov.usgs.ias.gui.common_library;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.util.Calendar;
import java.util.Observable;
import java.util.Observer;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.JFrame;

/**
 * Implements a dialog to select a single date/time
 */
public class DateDialog extends JDialog implements WindowListener,
        ActionListener, Observer {

    private DatePanel datePanel;           // panel to display the date widgets
    private DateRangeProcessor processor;  // the processor that will act on
                                           // the selected date

    /**
     * Constructor of the Date Range Dialog Box.  The calendar dialog is
     * optional; if provided, this dialog will have a calendar button beside
     * each date to allow the user to graphically set the date.
     * @param parentFrame     frame that this dialog is a child of
     * @param calendarDialog  dialog for date selection
     */
    public DateDialog(JFrame parentFrame, CalendarDialog calendarDialog) {
        super(parentFrame, "", true);
        setLayout(new BorderLayout());

        this.processor = null;

        //Create a Date Panel
        datePanel = new DatePanel(parentFrame, null, calendarDialog);
        add(datePanel);

        //Create a button Panel to hold all of the buttons
        JPanel buttonPanel = new JPanel();

        //Create the Ok button
        JButton okButton = new JButton("Ok");
        okButton.setMnemonic(KeyEvent.VK_O);
        okButton.setToolTipText(getTitle());
        okButton.addActionListener(this);
        buttonPanel.add(okButton);

        //Create the Cancel button
        JButton closeButton = new JButton("Cancel");
        closeButton.setMnemonic(KeyEvent.VK_C);
        closeButton.addActionListener(this);
        buttonPanel.add(closeButton);

        //Add the panels to the dialog window
        add(datePanel, BorderLayout.NORTH);
        add(buttonPanel, BorderLayout.SOUTH);
        pack();

        //Add the window events to the dialog window
        addWindowListener(this);

        ObservableDate.getInstance().addObserver(this);
    }

    private void updateDatePanel(Observable o) {
        ObservableDate observableDate = (ObservableDate)o;
        int year = observableDate.getSelectedYear();
        int month = observableDate.getSelectedMonth();
        int day = observableDate.getSelectedDay();
        int hour = observableDate.getSelectedHour();
        int min = observableDate.getSelectedMinute();
        int sec = observableDate.getSelectedSecond();

        //Set the date combo boxes
        datePanel.setDate(year, month, day, hour, min, sec);
    }

    // ActionListener methods
    /**
     * Handle user actions in the dialog.
     * @param e  the action selected by the user
     */
    @Override
    public void actionPerformed(ActionEvent e) {
        String command = e.getActionCommand();

        if (command.equals("Ok")) {
            if (processor != null) {
                Calendar selectedDate = datePanel.getDate();
                // Update the observable date, then call processor
                ObservableDate.getInstance().setCurrentDate(
                    selectedDate.get(Calendar.YEAR),
                    selectedDate.get(Calendar.MONTH),
                    selectedDate.get(Calendar.DAY_OF_MONTH),
                    selectedDate.get(Calendar.HOUR_OF_DAY),
                    selectedDate.get(Calendar.MINUTE),
                    selectedDate.get(Calendar.SECOND));
                processor.processWithDateRange(null, null);
            }
            setVisible(false);
        } else if (command.equals("Cancel")) {
            setVisible(false);
        }
    }

    // WindowListener methods
    /**
     * On window open set the date in the combo box
     * @param e  window open event
     */
    @Override
    public void windowOpened(WindowEvent e) {
        updateDatePanel(ObservableDate.getInstance());
    }

    // dummy window event handlers for unhandled events of the WindowListener
    @Override
    public void windowClosing(WindowEvent e) {
    }
    @Override
    public void windowClosed(WindowEvent e) {
    }
    @Override
    public void windowIconified(WindowEvent e) {
    }
    @Override
    public void windowDeiconified(WindowEvent e) {
    }
    @Override
    public void windowActivated(WindowEvent e) {
    }
    @Override
    public void windowDeactivated(WindowEvent e) {
    }

    // Observer methods
    /**
     * Update the date panel when the observable object is updated.
     * @param o      the observable object
     * @param arg    optional arguments
     */
    @Override
    public void update(Observable o, Object arg) {
        updateDatePanel(o);
    }

    // DateRangeProcessor methods
    /**
     * Set the DateRangeProcessor for this dialog
     * @param processor   the DateRangeProcessor object
     */
    public void setProcessor(DateRangeProcessor processor) {
        this.processor = processor;
    }
}
