package gov.usgs.ias.gui.common_library;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.Point;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.Calendar;
import java.net.URL;
import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JFormattedTextField;
import javax.swing.JPanel;
import javax.swing.JFrame;
import javax.swing.JSpinner;
import javax.swing.JTextField;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * Implements a single date/time selection panel
 */
public class DatePanel extends JPanel implements ActionListener,
        ChangeListener {

    private WraparoundSpinnerListModel monthSpinnerModel;
    private WraparoundSpinnerNumberModel daySpinnerModel;
    private SpinnerNumberModel yearSpinnerModel;
    private WraparoundSpinnerNumberModel hourSpinnerModel;
    private WraparoundSpinnerNumberModel minuteSpinnerModel;
    private WraparoundSpinnerNumberModel secondSpinnerModel;
    private MonthList months;
    private JButton dateCalendarButton;
    private JFrame parentFrame;
    private CalendarDialog calendarDialog; // calendar dialog for date selection
    private final PropertyChangeSupport pcs = new PropertyChangeSupport(this);

    /**
     * Constructor of the Date Range Panel
     * @param parent           the parent frame of this dialog
     * @param dateLabel        string to display above the date/time widgets
     * @param calendarDialog   optional dialog to allow selecting the date from
     *                         a calendar
     */
    public DatePanel(JFrame parent, String dateLabel, 
        CalendarDialog calendarDialog) {
        super(new BorderLayout());

        this.parentFrame = parent;
        this.calendarDialog = calendarDialog;

        //Create the main panel to display all of the components
        JPanel mainPanel = new JPanel();
        GridBagLayout gridbag = new GridBagLayout();
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        mainPanel.setLayout(gridbag);

        //Create a label for the date
        if (dateLabel != null) {
            JLabel label = new JLabel(dateLabel);
            mainPanel.add(label, gbc);
        }

        //Create the date Panel
        JPanel datePanel = new JPanel(new FlowLayout());

        //Create a Month JSpinner
        months = new MonthList();
        monthSpinnerModel = new WraparoundSpinnerListModel(
            months.getMonthStrings());
        JSpinner monthSpinner = new JSpinner(monthSpinnerModel);
        formatSpinnerTextField(monthSpinner);
        monthSpinner.addChangeListener(this);
        datePanel.add(monthSpinner);

        //Create a Day JSpinner
        daySpinnerModel = new WraparoundSpinnerNumberModel(1, 1, 31, 1);
        JSpinner daySpinner = new JSpinner(daySpinnerModel);
        formatSpinnerTextField(daySpinner);
        daySpinner.addChangeListener(this);
        datePanel.add(daySpinner);

        // Create a Year JSpinner, initialized to the configured min/max range
        yearSpinnerModel = new SpinnerNumberModel(DateConfig.STARTYEAR,
                DateConfig.MINYEAR, DateConfig.MAXYEAR, 1);
        JSpinner yearSpinner = new JSpinner(yearSpinnerModel);
        yearSpinner.setEditor(new JSpinner.NumberEditor(yearSpinner, "#"));
        formatSpinnerTextField(yearSpinner);
        yearSpinner.addChangeListener(this);
        datePanel.add(yearSpinner);

        if (calendarDialog != null) {
            // Create a clickable button to open a calendar dialog box
            URL calendarImageUrl = getClass().getResource("Calendar16.gif");
            if (calendarImageUrl == null) {
                dateCalendarButton = new JButton("Select Date");
            } else {
                dateCalendarButton = new JButton(
                    new ImageIcon(calendarImageUrl));
            }
            dateCalendarButton.setToolTipText("Select date using calendar");
            dateCalendarButton.addActionListener(this);
            dateCalendarButton.setActionCommand("displayCalendarDialog");
            datePanel.add(dateCalendarButton);
        }

        // Add JSpinners for the time
        hourSpinnerModel = new WraparoundSpinnerNumberModel(0, 0, 23, 1);
        JSpinner hourSpinner = new JSpinner(hourSpinnerModel);
        hourSpinner.setEditor(new JSpinner.NumberEditor(hourSpinner, "00"));
        formatSpinnerTextField(hourSpinner);
        hourSpinner.addChangeListener(this);
        datePanel.add(hourSpinner);
        datePanel.add(new JLabel(":"));
        minuteSpinnerModel = new WraparoundSpinnerNumberModel(0, 0, 59, 1);
        JSpinner minuteSpinner = new JSpinner(minuteSpinnerModel);
        minuteSpinner.setEditor(new JSpinner.NumberEditor(minuteSpinner, "00"));
        formatSpinnerTextField(minuteSpinner);
        minuteSpinner.addChangeListener(this);
        datePanel.add(minuteSpinner);
        datePanel.add(new JLabel(":"));
        secondSpinnerModel = new WraparoundSpinnerNumberModel(0, 0, 59, 1);
        JSpinner secondSpinner = new JSpinner(secondSpinnerModel);
        secondSpinner.setEditor(new JSpinner.NumberEditor(secondSpinner, "00"));
        formatSpinnerTextField(secondSpinner);
        secondSpinner.addChangeListener(this);
        datePanel.add(secondSpinner);
    
        //Add the date panel to the main panel
        mainPanel.add(datePanel, gbc);

        //Add the main panel to the dialog window
        add(mainPanel, BorderLayout.CENTER);
    }
    
    /**
     * Add a property change listener so parents can be notified when
     * a component has been modified.
     * @param propertyName
     * @param listener
     */
    @Override
    public void addPropertyChangeListener(String propertyName,
            PropertyChangeListener listener) {
         pcs.addPropertyChangeListener(propertyName, listener);
    }
    
    /**
     * Format the text fields of spinners consistently
     * @param spinner The JSpinner to format its text field
     */
    private void formatSpinnerTextField(JSpinner spinner) {
        JComponent editor = spinner.getEditor();
        JFormattedTextField ftf;
        ftf = ((JSpinner.DefaultEditor)editor).getTextField();
        if (ftf != null) {
            if (editor instanceof JSpinner.ListEditor) {
                //For the month list, want to increase the width also
                ftf.setColumns(6);  //specify more width, default is very narrow
            }
            //In all cases, insert some padding and left-align
            ftf.setBorder(BorderFactory.createCompoundBorder(ftf.getBorder(),
                    BorderFactory.createEmptyBorder(1,5,1,5)));
            ftf.setHorizontalAlignment(JTextField.LEFT);
        }
    }

    public void setDate(int year, int month, int day) {
        yearSpinnerModel.setValue(year);
        monthSpinnerModel.setValue(months.getMonthStrings()[month]);
        daySpinnerModel.setValue(day);
    }

    public void setDate(int year, int month, int day,
            int hour, int minute, int second) {
        yearSpinnerModel.setValue(year);
        monthSpinnerModel.setValue(months.getMonthStrings()[month]);
        daySpinnerModel.setValue(day);
        hourSpinnerModel.setValue(hour);
        minuteSpinnerModel.setValue(minute);
        secondSpinnerModel.setValue(minute);
    }

    /**
     * Get a Calendar reflecting the date currently selected in the panel.
     * @return  Calendar object
     */
    public Calendar getDate() {
        Calendar cal = Calendar.getInstance();
        cal.clear();
        cal.set(Calendar.YEAR, yearSpinnerModel.getNumber().intValue());
        cal.set(Calendar.MONTH, months.getMonthIndex(monthSpinnerModel
            .getValue().toString()));
        cal.set(Calendar.DAY_OF_MONTH, daySpinnerModel.getNumber().intValue());
        cal.set(Calendar.HOUR_OF_DAY, hourSpinnerModel.getNumber().intValue());
        cal.set(Calendar.MINUTE, minuteSpinnerModel.getNumber().intValue());
        cal.set(Calendar.SECOND, secondSpinnerModel.getNumber().intValue());
        return cal;
    }

    /**
     * Action performed event handler
     * @param e   The action event
     */
    @Override
    public void actionPerformed(ActionEvent e) {
        String command = e.getActionCommand();

        if (command.equals("displayCalendarDialog")) {
            Point loc = parentFrame.getLocationOnScreen();
            loc.y += 50;
            loc.x += 50;
            
            int year = yearSpinnerModel.getNumber().intValue();
            int day = daySpinnerModel.getNumber().intValue();
            int month = months.getMonthIndex(monthSpinnerModel
                .getValue().toString());

            // Update the observable date to initialize the calendar
            ObservableDate.getInstance().setCurrentDate(year, month, day);
            if (calendarDialog != null) {                
                calendarDialog.setPanel(this);
                calendarDialog.setLocation(loc);
                calendarDialog.setVisible(true);
            }
        }
    }

    /**
     * Change event handler
     * @param e   The change event
     */
    @Override
    public void stateChanged(ChangeEvent e) {
        // Update the days range for the current year/month selection
        // after a change to any of the widgets
        int sy = yearSpinnerModel.getNumber().intValue();
        int sm = months.getMonthIndex(monthSpinnerModel.getValue()
            .toString());
        int sd = daySpinnerModel.getNumber().intValue();

        Calendar cal = Calendar.getInstance();
        cal.clear();
        cal.set(sy, sm, 1);
        int daysInMonth = cal.getActualMaximum(Calendar.DAY_OF_MONTH);

        // If current year/month selection has fewer days in the month than
        // current day selected, back up the day selection to last day valid
        // for current year/month selection and reset the day spinner maximum.
        if (sd > daysInMonth) {
            daySpinnerModel.setValue(daysInMonth);
        }
        
        // This fixes the upper limit no matter if current year/month selection
        // has more or less days than before, gets the upper day range limit
        // correct for the current year/month selection either way.
        daySpinnerModel.setMaximum(daysInMonth);
        
        pcs.firePropertyChange("dateModified", false, true);
    }
}
