package gov.usgs.ias.gui.common_library;

import java.awt.BorderLayout;
import javax.swing.BoxLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.util.Calendar;
import java.util.Observable;
import java.util.Observer;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.JFrame;

/**
 * Implements a date range dialog with a start & end date
 */
public class DateRangeDialog extends JDialog
    implements ActionListener, Observer {

    private CalendarDialog calendarDialog; // calendar dialog for date selection
    private DatePanel startDatePanel;      // panel for the start date
    private DatePanel endDatePanel;        // panel for the end date
    private DateRangeProcessor processor;  // the object that will perform 
                                           // processing when the user has 
                                           // selected dates

    /**
     * Constructor of the Date Range Dialog Box.  The calendar dialog is
     * optional; if provided, this dialog will have a calendar button beside
     * each date to allow the user to graphically set the date.
     * @param parentFrame     frame that this dialog is a child of
     * @param calendarDialog  dialog for date selection
     */
    public DateRangeDialog(JFrame parentFrame, CalendarDialog calendarDialog) {
        super(parentFrame, "", true);
        setLayout(new BorderLayout());

        this.calendarDialog = calendarDialog;
        this.processor = null;

        // Create Date Range Panels for both start & end dates
        JPanel dateRangePanel = new JPanel();
        dateRangePanel.setLayout(new BoxLayout(dateRangePanel,
                BoxLayout.Y_AXIS));
        startDatePanel = new DatePanel(parentFrame, "Start Date",
            calendarDialog);
        dateRangePanel.add(startDatePanel);
        endDatePanel = new DatePanel(parentFrame, "End Date",
            calendarDialog);
        dateRangePanel.add(endDatePanel);
        setToToday();

        //Create a button Panel to hold all of the buttons
        JPanel buttonPanel = new JPanel();

        //Create the Ok button
        JButton okButton = new JButton("Ok");
        okButton.setMnemonic(KeyEvent.VK_O);
        okButton.setToolTipText(getTitle());
        okButton.addActionListener(this);
        buttonPanel.add(okButton);

        //Create the Cancel button
        JButton closeButton = new JButton("Cancel");
        closeButton.setMnemonic(KeyEvent.VK_C);
        closeButton.addActionListener(this);
        buttonPanel.add(closeButton);

        //Add the panels to the dialog window
        add(dateRangePanel, BorderLayout.NORTH);
        add(buttonPanel, BorderLayout.SOUTH);
        pack();

        ObservableDate.getInstance().addObserver(this);
    }

    private void setToToday() {
        //Get the current Date
        Calendar today = Calendar.getInstance();
        setRange(today, today);
    }

    /**
     * Set the start and end dates of the displayed range.
     * @param startDate  Start date
     * @param endDate    End date
     */
    public void setRange(Calendar startDate, Calendar endDate) {
        if (startDate != null) {
            startDatePanel.setDate(startDate.get(Calendar.YEAR),
                    startDate.get(Calendar.MONTH),
                    startDate.get(Calendar.DAY_OF_MONTH));
        }
        if (endDate != null) {
            endDatePanel.setDate(endDate.get(Calendar.YEAR),
                    endDate.get(Calendar.MONTH),
                    endDate.get(Calendar.DAY_OF_MONTH));
        }
    }

    // ActionListener methods
    /**
     * Action performed event handler.
     * @param e  Event for the user action taken
     */
    @Override
    public void actionPerformed(ActionEvent e) {
        String command = e.getActionCommand();

        if (command.equals("Ok")) {
            if (processor != null) {
                processor.processWithDateRange(startDatePanel.getDate(),
                        endDatePanel.getDate());
            }
            setVisible(false);
        } else if (command.equals("Cancel")) {
            setVisible(false);
        }
    }

    // Observer methods
    /**
     * When the observed item notifies this object that a change occurred,
     * get information from it to update the appropriate date panel
     */
    @Override
    public void update(Observable o, Object arg) {

        ObservableDate observableDate = (ObservableDate)o;
        int year = observableDate.getSelectedYear();
        int month = observableDate.getSelectedMonth();
        int day = observableDate.getSelectedDay();

            if (calendarDialog != null && calendarDialog.getPanel() != null) {
                calendarDialog.getPanel().setDate(year, month, day);
                calendarDialog.setPanel(null);
            }
    }

    /**
     * Set the DateRangeProcessor to do processing for the selected dates
     */
    public void setProcessor(DateRangeProcessor processor) {
        this.processor = processor;
    }
}
