package gov.usgs.ias.gui.common_library;

import java.awt.BorderLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Component;
import java.awt.Container;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.Calendar;
import javax.swing.JPanel;
import javax.swing.JFrame;

/**
 * Name: DateRangePanel.java
 * Description: Implements a date range selection panel (begin & end date) by
 * using two DatePanels.
 */
public class DateRangePanel extends JPanel implements PropertyChangeListener {
    private DatePanel startDatePanel;
    private DatePanel endDatePanel;
    private final PropertyChangeSupport pcs = new PropertyChangeSupport(this);

    /**
     * Constructor of the Date Range Panel which will set up all the gui
     * components of the screen.
     * @param application application frame context object
     * @param calendar - The CalendarDialog instance to use in the date panels.
     */
    public DateRangePanel(JFrame application, CalendarDialog calendarDialog){
       super(new BorderLayout());
        
       //Create the main panel to display all of the components
       JPanel mainPanel = new JPanel();
       GridBagLayout gridbag = new GridBagLayout();
       GridBagConstraints gbc = new GridBagConstraints();
       gbc.gridwidth = GridBagConstraints.REMAINDER;
       mainPanel.setLayout(gridbag);

       //Create the start date Panel
       startDatePanel =
               new DatePanel(application, "Start Date", calendarDialog);

       //Add the start date panel to the main panel
       mainPanel.add(startDatePanel,gbc);

       //Create an end date panel
       endDatePanel = new DatePanel(application, "End Date", calendarDialog);

       //Add the end date panel to the main panel
       mainPanel.add(endDatePanel,gbc);

       //Add the main panel to the dialog window
       add(mainPanel,BorderLayout.CENTER);
       
       startDatePanel.addPropertyChangeListener("dateModified", this);
       endDatePanel.addPropertyChangeListener("dateModified", this);
    }

    /**
     * Add a property change listener so parents can be notified when
     * a component has been modified.
     * @param propertyName
     * @param listener
     */
    @Override
    public void addPropertyChangeListener(String propertyName,
            PropertyChangeListener listener) {
         pcs.addPropertyChangeListener(propertyName, listener);
    }
    
    // Set the children components, and their children recursively, to be
    // enabled or disabled
    private void setChildrenEnabled(Container container, boolean enabledFlag){
        for (int i = 0; i < container.getComponentCount(); i++)
        {
            Component datePanelComponent = container.getComponent(i);
            datePanelComponent.setEnabled(enabledFlag);
            if (datePanelComponent instanceof Container)
            {
                // The child component has components, so recursively enable
                // or disable its children
                setChildrenEnabled((Container) datePanelComponent, enabledFlag);
            }
        }
    }

    /**
     * Get the value of the start date in the dialog.
     * @return The value the start date is set to in the dialog.
     */
    public Calendar getStartDate()
    {
        return startDatePanel.getDate();
    }

    /**
     * Get the value of the end date in the dialog.
     * @return The value the end date is set to in the dialog.
     */
    public Calendar getEndDate()
    {
        return endDatePanel.getDate();
    }

    /**
     * Set the Start Date in the range to a year, month, & day.
     * @param year - The year to set the start date to.
     * @param month - The month to set the start date to.
     * @param day - The day to set the start date to.
     */
    public void setStartDate(int year, int month, int day)
    {
        startDatePanel.setDate(year, month, day);
    }

    /**
     * Set the Start Date in the range to a year, month, day, hour, minute, &
     * second.
     * @param year - The year to set the start date to.
     * @param month - The month to set the start date to.
     * @param day - The day to set the start date to.
     * @param hour - The hour to set the start date to.
     * @param min - The minute to set the start date to.
     * @param sec - The second to set the start date to.
     */
    public void setStartDate(int year, int month, int day, int hour, int min,
            int sec)
    {
        startDatePanel.setDate(year, month, day, hour, min, sec);
    }

    /**
     * Set the end date in the range to the year, month, & day passed in.
     * @param year - The year to set the end date to.
     * @param month - The month to set the end date to.
     * @param day - The day to set the end date to.
     */
    public void setEndDate(int year, int month, int day)
    {
        endDatePanel.setDate(year, month, day);
    }

    /**
     * Set the end date in the range to the year, month, day, hour, minute, &
     * second passed in.
     * @param year - The year to set the end date to.
     * @param month - The month to set the end date to.
     * @param day - The day to set the end date to.
     * @param hour - The hour to set the end date to.
     * @param min - The minute to set the end date to.
     * @param sec - The second to set the end date to
     */
    public void setEndDate(int year, int month, int day, int hour, int min,
            int sec)
    {
        endDatePanel.setDate(year, month, day, hour, min, sec);
    }

    /**
     * Set the start date panel to be enabled or disabled.
     * @param enableFlag - Flag specifying to set to enabled or disabled. 
     */
    public void enableStartDate(boolean enableFlag)
    {
        startDatePanel.setEnabled(enableFlag);
        setChildrenEnabled(startDatePanel, enableFlag);
    }

    /**
     * Set the end date panel to be enabled or disabled.
     * @param enableFlag - Flag specifying to set to enabled or disabled. 
     */
    public void enableEndDate(boolean enableFlag)
    {
        endDatePanel.setEnabled(enableFlag);
        setChildrenEnabled(endDatePanel, enableFlag);
    }

    /**
     * Handle when a registered property change occurs. Fire a notice for
     * property change listeners when registered components are modified.
     * @param e  property change event fired
     */
    @Override
    public void propertyChange(PropertyChangeEvent e) {
        pcs.firePropertyChange("dateModified", false, true);
    }
}
