package gov.usgs.ias.gui.common_library;

import java.util.Observable;

/**
 * This class is used to persist the last date entered by the user in one of the
 * date dialogs.  
 * It is a singleton so we only have one copy for the entire application.
 * It is observable, so when one of the dialogs updates the date
 * stored here, other GUI components will be notified of the new date if they
 * have added an observer to this object.
 * The DateDialog, DateRangeDialog and CalendarDialog automatically set
 * themselves up as observers of this object.
 */
public class ObservableDate extends Observable {

    //The singleton variable
    private final static ObservableDate INSTANCE = new ObservableDate();

    private int currentYear;
    private int currentMonth;
    private int currentDay;
    private int currentHour;
    private int currentMinute;
    private int currentSecond;

    // a private constructor prevents direct instantiation, use getInstance()
    // instead
    private ObservableDate() {
    }

    /**
     * Get the instance of this singleton class.
     * @return The instance.
     */
    static public ObservableDate getInstance() {
        return INSTANCE;
    }

    // Set the current date (yy/mm/dd) and notify observers that it has changed
    public void setCurrentDate(int year, int month, int day) {
        currentYear = year;
        currentMonth = month;
        currentDay = day;

        setChanged();
        notifyObservers();
    }

    // Set the current date & time (yy/mm/dd hh:mm:ss) and notify observers
    // that it has changed
    public void setCurrentDate(int year, int month, int day,
        int hour, int min, int sec) {
        currentYear = year;
        currentMonth = month;
        currentDay = day;
        currentHour = hour;
        currentMinute = min;
        currentSecond = sec;

        setChanged();
        notifyObservers();
    }

    public int getSelectedYear() {
        return currentYear;
    }

    public int getSelectedMonth() {
        return currentMonth;
    }

    public int getSelectedDay() {
        return currentDay;
    }

    public int getSelectedHour() {
        return currentHour;
    }

    public int getSelectedMinute() {
        return currentMinute;
    }

    public int getSelectedSecond() {
        return currentSecond;
    }
}
