package gov.usgs.ias.gui.common_library;

import org.apache.log4j.Logger;
import java.util.logging.Handler;
import java.util.logging.LogRecord;

/**
 * Class to reroute java.util.logger messages to log4j.
 * This provides a consistent output location for log messages from IAS
 * components (using log4j) and the PDS Tools (using java.util.logger).
 * Be sure to register this handler in your logging properties files identified
 * by the java.util.logging.config.file system property.
 * This will use the logging configuration in your log4j.properties file
 * instead of the java.util.logging configuration file which is normally in
 * ${JDK}/jre/lib/logging.properties.
 */
public class RouteToLog4jHandler extends Handler {

    /**
     * java.util logger publish method overridden to use a log4j logger.
     * @param record  the java.util logging record
     */
    @Override
    public void publish(LogRecord record) {
        org.apache.log4j.Logger log4jlogger = null;
        if (record.getLoggerName() == null) {
            log4jlogger = getTargetLogger(record.getSourceClassName());
        } else {
            log4jlogger = getTargetLogger(record.getLoggerName());
        }
        org.apache.log4j.Level level = toLog4jLevel(record.getLevel());
        log4jlogger.log(level, toLog4jMessage(record), record.getThrown());
    }

    /**
     * Get the log4j logger by name or class
     * @param loggerName  Name of the logger
     * @return            Log4j logger for the given name
     */
    static Logger getTargetLogger(String loggerName) {
        return Logger.getLogger(loggerName);
    }
    /**
     * Get the log4j logger by class
     * @param aClass      Class for which to retrieve a logger
     * @return            Log4j logger for the given class
     */
    public static Logger getTargetLogger(Class aClass) {
        return getTargetLogger(aClass.getName());
    }

    // 
    /**
     * Convert util log record into a log4j message
     * @param record   java.util log record
     * @return         message string
     */
    private String toLog4jMessage(LogRecord record) {
        String message = record.getMessage();
        // If parameters were included, use a Formatter
        try {
            Object parameters[] = record.getParameters();
            if (parameters != null && parameters.length != 0) {
                StringBuilder sb = new StringBuilder();
                java.util.Formatter formatter = new java.util.Formatter(sb);
                formatter.format(record.getMessage(), record.getParameters());
                message = sb.toString();
            }
        }
        catch (Exception ex) {
            // ignore Exception
        }
        return message;
    }

    /**
     * Convert a java util logger level to a log4j logger level
     * @param utilLevel  java.util logging level
     * @return           log4j logging level
     */
    private org.apache.log4j.Level toLog4jLevel(
            java.util.logging.Level utilLevel) {
        if (utilLevel.intValue()
                == java.util.logging.Level.OFF.intValue()) {
            return org.apache.log4j.Level.OFF;
        } else if (utilLevel.intValue()
                >= java.util.logging.Level.SEVERE.intValue()) {
            return org.apache.log4j.Level.ERROR;
        } else if (utilLevel.intValue()
                >= java.util.logging.Level.WARNING.intValue()) {
            return org.apache.log4j.Level.WARN;
        } else if (utilLevel.intValue()
                >= java.util.logging.Level.INFO.intValue()) {
            return org.apache.log4j.Level.INFO;
        } else if (utilLevel.intValue()
                >= java.util.logging.Level.CONFIG.intValue()) {
            return org.apache.log4j.Level.INFO;
        } else if (utilLevel.intValue()
                >= java.util.logging.Level.FINE.intValue()) {
            return org.apache.log4j.Level.TRACE;
        } else if (utilLevel.intValue()
                >= java.util.logging.Level.FINER.intValue()) {
            return org.apache.log4j.Level.TRACE;
        } else if (utilLevel.intValue()
                >= java.util.logging.Level.FINEST.intValue()) {
            return org.apache.log4j.Level.TRACE;
        }
        return org.apache.log4j.Level.OFF;
    }

    @Override
    public void flush() {
        // do nothing
    }

    @Override
    public void close() {
        // do nothing
    }
}
