
package gov.usgs.ias.gui.common_library;

import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import javax.swing.AbstractCellEditor;
import javax.swing.JButton;
import javax.swing.JTable;
import javax.swing.table.TableCellEditor;
import javax.swing.table.TableCellRenderer;

/**
 * This class is a custom button to embed in a JTable.  This button will get
 * used any time a button is needed in a table on a Java GUI.
 *
 * This button currently does not handle columns, only rows.
 * GetTableCellEditorComponent & getTableCellRendererComponent have columns
 * as parameters, only because those parameters are part of the TableCellEditor
 * interface, they are not used.
 *
 * Columns are currently implemented by placing a button in a specific column
 * in the Table/TableModel.  If columns need to be implemented here, a column
 * will need to be added to onButtonPress and the buttons HashMap logic will
 * need to be modified to handle rows & columns.
 */
public class TableButton extends AbstractCellEditor implements TableCellEditor,
        TableCellRenderer{

    public interface TableButtonPressedHandler {
        /**
         * Gets called when the button is clicked
         * @param row The row of the button in the table.
         */
        void onButtonPress(int row);
    }

    private List<TableButtonPressedHandler> handlers;
    private HashMap<Integer, JButton> buttons;

    /**
     * Construct the TableButton class by creating an ArrayList of
     * TableButtonPressedHandlers, and a HashMap of buttons where the row
     * is the key and the JButton instance is the value.
     */
    public TableButton(){
        handlers = new ArrayList<TableButtonPressedHandler>();
        buttons = new HashMap<Integer, JButton>();
    }

    /**
     *  Add a TableButton handler so that we know what to do when the
     * button is clicked.
     * @param handler The TableButtonHandler to be added.
     */
    public void addHandler(TableButtonPressedHandler handler){
        handlers.add(handler);
    }

    /**
     * Remove a TableButton handler.
     * @param handler The TableButtonHandler to be removed.
     */
    public void removeHandler(TableButtonPressedHandler handler){
        handlers.remove(handler);
    }

    /**
     * Set the text on a specific button
     * @param row The row the button is in.
     * @param text The text to put on the button.
     */
    public void setButtonText(int row, String text){
        JButton button = null;
        if(buttons.containsKey(row)){
            button = buttons.get(row);
            button.setText(text);
        }
    }

    /**
     * Remove a button from a row in the table.
     * @param row The row the button is to be removed from.
     */
    public void removeRow(int row){
        if(buttons.containsKey(row)){
            buttons.remove(row);
        }
    }

    /**
     * Move a button from one row in the table to another row in the table.
     * @param oldRow The row the button is currently in.
     * @param newRow The row the button is to be moved to.
     */
    public void moveRow(int oldRow, int newRow){
        if(buttons.containsKey(oldRow)){
            JButton button = buttons.remove(oldRow);
            buttons.put(newRow, button);
        }
    }
    /**
     * Clean up the object by clearing out the handlers.
     */
    public void dispose(){
        handlers.clear();
    }

    @Override
    //Since this is a button, it has no value in a table cell.
    public Object getCellEditorValue() {
        return null;
    }

    //Return the button as the cell editor for the cell that is selected in
    //the table.
    /**
     * Return the button as the cell editor for the cell that is selected in
     * the table.  Column needs to be here because of the TableCellEditor
     * interface, but it is not used.
     * @param table The table the cell is in.
     * @param value The value of the cell.
     * @param selected If the cell is selected or not.
     * @param row The row the column is in.
     * @param column The column the cell is in.  We are not currently using this
     * but it needs to be here for the interface.  The column is handled by
     * the Table/TableModel itself.
     * @return Return the button for the row in the table given.
     */
    @Override
    public Component getTableCellEditorComponent(JTable table, Object value,
            boolean selected, int row, int column) {
        
        JButton button = null;
        if(buttons.containsKey(row)){
            button = buttons.get(row);
        }else{
            button = new JButton();

            if(value != null && value instanceof String){
                button.setText((String) value);
            }

            buttons.put(row, button);
        }
        
        return button;

    }

    //Return the button as the cell renderer for the cell that is selected in
    // the table.  And set up the onButtonPressHandler
    /**
     * Return the button as the cell renderer for the cell that is selected in
     * the table.  And set up the onButtonPressHandler.
     * @param table The table the cell is in.
     * @param value The value of the cell.
     * @param focus If the cell has focus or not.
     * @param bln1
     * @param row The row of the cell.
     * @param column The column of the cell.  This will not be used.  At this
     * time columns are implemented in the Table/TableModel, not here.
     * @return The component, the Button for this row.
     */
    @Override
    public Component getTableCellRendererComponent(JTable table, Object value,
            boolean focus, boolean bln1, final int row, final int column) {
        JButton button = null;
        if(buttons.containsKey(row)){
            button = buttons.get(row);
        }else{
            button = new JButton();

            if (value != null && value instanceof String){
                button.setText((String)value);
            }

            button.addActionListener(new ActionListener(){
                @Override
               public void actionPerformed(ActionEvent e) {
                   for(TableButtonPressedHandler handler : handlers){
                       handler.onButtonPress(row);
                   }
               }
            });
            buttons.put(row, button);
        }
        return button;
    }

}
