/**
 * BpfDAOTest
 *
 * JUnit tests for BpfDAO class.  All methods with a @Test annoation
 * will be run by the junit 4 test framework.
 * Testing properties are in two locations and prefixed with test-sys-prop
 * Build Automated Testing: test.properties
 * Developer Testing:  nbproject/private/private.properties
 */
package gov.usgs.ias.dao.bpf;

import gov.usgs.ias.dao.ParseODL;
import gov.usgs.ias.dao.TestSupport;
import gov.usgs.ias.cps.CpsGroup;
import gov.usgs.ias.cps.CpsStatusEnum;
import gov.usgs.ias.dao.DataConnection;
import gov.usgs.ias.constant.Constants;
import gov.nasa.pds.tools.label.*;
import java.sql.PreparedStatement;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import static org.junit.Assert.*;

public class BpfDAOTest {

    private DataConnection conn;
    private int nGroups = 0;
    private String daoAuth = null;

    public BpfDAOTest() {
    }

    @BeforeClass
    public static void setUpClass() throws Exception {
    }

    @AfterClass
    public static void tearDownClass() throws Exception {
    }

    /**
     * Setup to be done before each test.
     * N.B.: these unit tests assume the standard set of test BPFs
     * have been loaded into the database.
     */
    @Before
    public void setUp() {
        String bpfDbProperty = "BPF_TEST_DB";
        this.conn = new DataConnection(bpfDbProperty);

        // Set up the authentication string for the proper schema, default PROD
        if (System.getProperty("SCHEMA") != null &&
            System.getProperty("SCHEMA").equals("EVAL")) {
            daoAuth = Constants.DAO_AUTH_IAS_EVAL;
        } else {
            daoAuth = Constants.DAO_AUTH_IAS_PROD;
        }
    }

    @After
    public void tearDown() {
        this.conn.close();
    }

    /**
     * Test of getCpsGroups method, of class BpfDAO.
     */
    @Test
    public void testGetCpsGroups() throws Exception {
        System.out.println("getCpsGroups");
        String bpfName = System.getProperty("BPF_FILE");
        BpfDAO instance = new BpfDAO(this.conn, this.daoAuth);

        // Get the groups from the database
        System.out.println(" getting groups for " + bpfName);
        Collection<CpsGroup> result = instance.getCpsGroups(bpfName);
        nGroups = result.size();
        assertTrue("No groups found for " + bpfName, nGroups > 0);

        // Parse the groups from the file
        Label bpfLabel = ParseODL.parseOdlFile("test/"
                + System.getProperty("BPF_FILE"));
        assertNotNull("Error parsing " + System.getProperty("BPF_FILE"),
                bpfLabel);

        // Compare the group names between db and file
        List odlGroups = bpfLabel.getGroups();
        GroupStatement odlGroup;
        Iterator<CpsGroup> dbGroups = result.iterator();
        String dbGroup;
        int i = 0;
        while (dbGroups.hasNext()) {
            dbGroup = dbGroups.next().getName();
            odlGroup = (GroupStatement)odlGroups.get(i++);
            System.out.println("Comparing " + dbGroup + " to "
                    + odlGroup.getIdentifier());
            assertTrue("Group from database " + dbGroup + " does not match "
                    + " group from ODL file " + odlGroup,
                    dbGroup.equals(odlGroup.getIdentifier()));
        }
    }

    /**
     * Test of getCpsGroupDataStmt method, of class BpfDAO.
     */
    @Test
    public void testGetCpsGroupDataStmt() throws Exception {
        System.out.println("getCpsGroupDataStmt");
        BpfDAO instance = new BpfDAO(this.conn, this.daoAuth);
        PreparedStatement result = instance.getCpsGroupDataStmt();
        assertNotNull("Unable to create prepared statement", result);
    }

    /**
     * Test of getCpsGroupById method, of class BpfDAO.
     */
    @Test
    public void testGetCpsGroupById() throws Exception {
        System.out.println("getCpsGroupById");
        boolean addParameters = false;
        BpfDAO instance = new BpfDAO(this.conn, this.daoAuth);

        // Get all the group IDs for the test file
        Collection<CpsGroup> groups = instance.getCpsGroups(
                System.getProperty("BPF_FILE"));
        assertTrue("No groups found for " + System.getProperty("BPF_FILE"),
                groups.size() > 0);

        // Get each group by ID, and compare to the list we already got
        long bpfGroupId = 0;
        String bpfGroupName;
        Iterator<CpsGroup> cgi = groups.iterator();
        CpsGroup currentGroup;
        while (cgi.hasNext()) {
            currentGroup = cgi.next();
            bpfGroupId = currentGroup.getGroupId();
            bpfGroupName = currentGroup.getName();
            CpsGroup result = instance.getCpsGroupById(bpfGroupId,
                    addParameters);
            assertNotNull("Group not found for ID " + bpfGroupId, result);
            // compare the ID and group name
            assertTrue("Group ID " + result.getGroupId()
                    + " does not match expected ID " + bpfGroupId,
                    result.getGroupId() == bpfGroupId);
            assertTrue("Group name " + result.getName()
                    + " does not match expected name " + bpfGroupName,
                    result.getName().equals(bpfGroupName));
        }
    }

    /**
     * Test of insertBpf method, of class BpfDAO.
     */
    @Test
    public void testInsertBpf() throws Exception {
        System.out.println("insertBpf");
        Label bpfLabel = ParseODL.parseOdlFile(
                System.getProperty("BPF_LOAD_FILE"));
        assertNotNull("Error parsing BPF file "
                + System.getProperty("BPF_LOAD_FILE"), bpfLabel);
        BpfDAO instance = new BpfDAO(this.conn, this.daoAuth);
        String fileName = instance.getFileAttribute(bpfLabel, "File_Name");
        assertNotNull("File_Name attribute not found", fileName);
        int result = -1;
        Exception insertException = new Exception("");
        try {
            result = instance.insertFile(bpfLabel, CpsStatusEnum.ACTIVE);
        } catch (Exception e) {
            // do nothing with the exception until after we clean up the db
            insertException = e;
        }

        // Delete the data before checking the status of the insert to ensure
        //  it gets cleaned up for next time
        PreparedStatement stmt = null;
        String sql = "DELETE from bpf_parameter_values_data a WHERE EXISTS" +
            " (SELECT * FROM bpf_main b, bpf_main_group c " +
            "  WHERE c.bpf_group_id = a.bpf_group_id " +
            "   and b.bpf_main_id = c.bpf_main_id " +
            "   and b.file_name = '" + fileName + "')";
        stmt = this.conn.getConnection(this.daoAuth).prepareStatement(sql);
        assertFalse("Unexpected result set returned from delete statement",
                stmt.execute());  // expecting no result set

        sql = "DELETE from bpf_group a WHERE EXISTS " +
                "(SELECT * FROM bpf_main b, bpf_main_group c  " +
                "WHERE c.bpf_group_id = a.bpf_group_id " +
                "and b.bpf_main_id = c.bpf_main_id   " +
                "and b.file_name = '" + fileName + "')";
        stmt = this.conn.getConnection(this.daoAuth).prepareStatement(sql);
        assertFalse("Unexpected result set returned from delete statement",
                stmt.execute());

        sql = "DELETE from bpf_main where file_name = '" + fileName + "'";
        stmt = this.conn.getConnection(this.daoAuth).prepareStatement(sql);
        assertFalse("Unexpected result set returned from delete statement",
                stmt.execute());
        
        assertTrue(insertException.getMessage(), result != -1);
    }

    /**
     * Test suite, for running the tests with an old version of junit/ant
     */
    public static junit.framework.Test suite() {
        if (System.getenv("ENABLE_DB_TESTS") == null) {
            System.out.println("Database tests are disabled");
            return new junit.framework.JUnit4TestAdapter(TestSupport.class);
        } else {
            return new junit.framework.JUnit4TestAdapter(BpfDAOTest.class);
        }
    }
}
