/**
 * BpfSummaryDAOTest
 *
 * JUnit tests for BpfSummaryDAO class.  All methods with a @Test annoation
 * will be run by the junit 4 test framework.
 * Testing properties are in two locations and prefixed with test-sys-prop
 * Build Automated Testing: test.properties
 * Developer Testing:  nbproject/private/private.properties
 */

package gov.usgs.ias.dao.bpf;

import gov.usgs.ias.dao.TestSupport;
import gov.usgs.ias.constant.Constants;
import gov.usgs.ias.cps.CpsStatusEnum;
import gov.usgs.ias.cps.CpsSummary;
import gov.usgs.ias.dao.DataConnection;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import org.apache.log4j.Logger;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import static org.junit.Assert.*;

public class BpfSummaryDAOTest {

    private DataConnection conn;
    private String daoAuth=null;
    private static final Logger logger = Logger.getLogger(Constants.LOGGER_BPF);

    public BpfSummaryDAOTest() {
    }

    @BeforeClass
    public static void setUpClass() throws Exception {
    }

    @AfterClass
    public static void tearDownClass() throws Exception {
    }

    /**
     * Setup to be done before each test.
     * N.B.: these unit tests assume the standard set of test BPFs
     * have been loaded into the database.
     */
    @Before
    public void setUp() {
        String bpfDbProperty = "BPF_TEST_DB";
        this.conn = new DataConnection(bpfDbProperty);

        // Set up the authentication string for the proper schema, default PROD
        if (System.getProperty("SCHEMA") != null &&
            System.getProperty("SCHEMA").equals("EVAL")) {
            logger.info("Using EVAL database");
            daoAuth = Constants.DAO_AUTH_IAS_EVAL;
        } else {
            logger.info("Using PROD database");
            daoAuth = Constants.DAO_AUTH_IAS_PROD;
        }
    }

    @After
    public void tearDown() {
        this.conn.close();
    }

    /**
     * Test of searchCpsMain method (date range), of class BpfSummaryDAO.
     * Exercises the date search methods of the DAO.
     */
    @Test
    public void testSearchCpsMainDate() throws Exception {
        logger.info("searchCpsMain date");
        SimpleDateFormat formatter =
            new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:sszzz");
        Date startDate = formatter.parse(
                System.getProperty("bpf.summary.daterange1.start"));
        Date endDate = formatter.parse(
                System.getProperty("bpf.summary.daterange1.end"));
        Integer version = 1;
        CpsStatusEnum status = CpsStatusEnum.valueOf("ACTIVE");
        boolean nearMatch = false;
        BpfSummaryDAO instance = new BpfSummaryDAO(this.conn, this.daoAuth);

        // query on start date only
        List<CpsSummary> result = instance.searchCpsMain(startDate, null,
                version, status, nearMatch);
        assertNotNull("Null returned on search for start date "
                + startDate.toString(), result);
        assertTrue("No matches found for start date "
                + startDate.toString(), result.size() > 0);
        logger.info("Single Date search results:");
        for (int i = 0; i < result.size(); i++) {
            logger.info(i + " " + result.get(i).getFilename());
        }

        // query on date range, any sensor
        result = instance.searchCpsMain(startDate, endDate, version, status,
                nearMatch);
        assertNotNull("Null returned on search for date range "
                + startDate.toString() + " - " + endDate.toString(), result);
        assertTrue("No matches found for date range "
                + startDate.toString() + " - " + endDate.toString(),
                result.size() > 0);
        logger.info("Date Range search results:");
        for (int i = 0; i < result.size(); i++) {
            logger.info(i + " " + result.get(i).getFilename());
        }

        // query on date range, OLI only
        result = instance.searchCpsMain(startDate, endDate, BpfSensorEnum.OLI,
                version, status, nearMatch);
        assertNotNull("Null returned on search for date range "
                + startDate.toString() + " - " + endDate.toString(), result);
        for (int i = 0; i < result.size(); i++) {
            logger.info(i + " " + result.get(i).getFilename());
        }
        logger.info("OLI Date Range search results:");
        assertTrue("No matches found for date range "
                + startDate.toString() + " - " + endDate.toString(),
                result.size() > 0);

        // query on date range, TIRS only, expect no matches
        result = instance.searchCpsMain(startDate, endDate, BpfSensorEnum.TIRS,
                version, status, nearMatch);
        assertNotNull("Null returned on search for date range "
                + startDate.toString() + " - " + endDate.toString(), result);
        logger.info("TIRS Date Range search results:");
        for (int i = 0; i < result.size(); i++) {
            logger.info(i + " " + result.get(i).getFilename());
        }
        assertTrue("Unexpected matches found for date range "
                + startDate.toString() + " - " + endDate.toString(),
                result.isEmpty());
        
        // query with a later date (no version), expect not to find one
        nearMatch = false;
        startDate = formatter.parse(
                System.getProperty("bpf.summary.daterange2.start"));
        endDate = formatter.parse(
                System.getProperty("bpf.summary.daterange2.end"));
        result = instance.searchCpsMain(startDate, endDate, BpfSensorEnum.OLI,
                null, status, nearMatch);
        assertNotNull("Null returned on search for date range "
                + startDate.toString() + " - " + endDate.toString(), result);
        logger.info("Found " + result.size() + " matches");
        for (int i = 0; i < result.size(); i++) {
            logger.info(i + " (unexpected) " + result.get(i).getFilename());
        }
        assertTrue("Unexpected matches found for date range "
                + startDate.toString() + " - " + endDate.toString(),
                result.size() == 0);

        // query with near match, no property set, expect an exception
        nearMatch = true;
        result = null;
        System.clearProperty(Constants.NEAR_MATCH_BPF_TOLERANCE_DAYS);
        try {
            result = instance.searchCpsMain(startDate, endDate, 
                BpfSensorEnum.OLI, null, status, nearMatch);
        } catch (Exception e) {
            // ignore exceptions
        }
        assertNull("Found unexpected match using nearMatch date search",
                result); // should have been none found

        // query with near match, now expect to find one
        System.setProperty(Constants.NEAR_MATCH_BPF_TOLERANCE_DAYS,
                System.getProperty("bpf.summary.date.tolerance"));
        result = instance.searchCpsMain(startDate, endDate, null, status,
                    nearMatch);
        assertNotNull("Null returned for nearMatch date query", result);
        assertTrue("No matches found for nearMatch date query",
                result.size() > 0);
        logger.info("Near Match Date Range search results:");
        for (int i = 0; i < result.size(); i++) {
            logger.info(i + " " + result.get(i).getFilename());
        }

        // query with near match, do not expect one
        startDate = formatter.parse(
                System.getProperty("bpf.summary.daterange3.start"));
        endDate = formatter.parse(
                System.getProperty("bpf.summary.daterange3.end"));
        result = instance.searchCpsMain(startDate, endDate, BpfSensorEnum.OLI, 
                null, status, nearMatch);
        assertNotNull("Null returned for nearMatch date range query", result);
        assertTrue("Unexpected matches found for nearMatch date range query",
                result.size() == 0);
    }

    /**
     * Test of searchCpsMain method (orbit range), of class BpfSummaryDAO.
     * Exercises the orbit search methods of the DAO.
     */
    @Test
    public void testSearchCpsMainOrbit() throws Exception {
        logger.info("searchCpsMain orbit");
        Long startOrbit = Long.parseLong(
                System.getProperty("bpf.summary.orbitrange1.start"));
        Long endOrbit = Long.parseLong(
                System.getProperty("bpf.summary.orbitrange1.end"));
        boolean nearMatch = false;
        Integer version = 1;

        CpsStatusEnum status = CpsStatusEnum.valueOf("ACTIVE");
        BpfSummaryDAO instance = new BpfSummaryDAO(this.conn, this.daoAuth);

        // query on start orbit only
        List<CpsSummary> result = instance.searchCpsMain(startOrbit, null,
                BpfSensorEnum.OLI, version, status, nearMatch);
        assertNotNull("Null returned on search for orbit " + startOrbit,
                result);
        assertTrue("No matches returned on search for orbit " + startOrbit,
                result.size() > 0);
        logger.info("Single Orbit search results:");
        for (int i = 0; i < result.size(); i++) {
            logger.info(i + " " + result.get(i).getBeginOrbit() + " "
                + result.get(i).getFilename());
        }

        // query on orbit range
        result = instance.searchCpsMain(startOrbit, endOrbit, BpfSensorEnum.OLI,
                version, status, nearMatch);
        assertNotNull("Null returned on search for orbit range " + startOrbit
                + " - " + endOrbit, result);
        assertTrue("No matches returned on search for orbit range " + startOrbit
                + " - " + endOrbit, result.size() > 0);
        Long foundOrbit = result.get(0).getBeginOrbit();
        assertTrue("Orbit " + foundOrbit
            + " is not within the requested range " + startOrbit
            + " - " + endOrbit,
            (foundOrbit >= startOrbit && foundOrbit <= endOrbit));
        logger.info("Orbit Range search results:");
        for (int i = 0; i < result.size(); i++) {
            logger.info(i + " " + result.get(i).getBeginOrbit() + " "
                + result.get(i).getFilename());
        }

        // query with a later orbit (no version), expect not to find one
        nearMatch = false;
        startOrbit = Long.parseLong(
                System.getProperty("bpf.summary.orbitrange2.start"));
        endOrbit = Long.parseLong(
                System.getProperty("bpf.summary.orbitrange2.end"));
        result = instance.searchCpsMain(startOrbit, endOrbit, BpfSensorEnum.OLI,
                null, status, nearMatch);
        assertNotNull("Null returned on search for orbit range " + startOrbit
                + " - " + endOrbit,result);
        assertTrue("No matches returned on search for orbit range " + startOrbit
                + " - " + endOrbit, result.size() == 0);
        logger.info("Found " + result.size() + " matches");

        // query with near match, no property set, expect an exception
        nearMatch = true;
        result = null;
        System.clearProperty(Constants.NEAR_MATCH_BPF_TOLERANCE_ORBIT);
        try {
            result = instance.searchCpsMain(startOrbit, endOrbit, 
                BpfSensorEnum.OLI, null, status, nearMatch);
        } catch (Exception e) {
            // ignore exceptions
        }
        assertNull("Found unexpected match using nearMatch date search",
                result); // should have been none found

        // query with near match, now expect to find one
        System.setProperty(Constants.NEAR_MATCH_BPF_TOLERANCE_ORBIT,
                System.getProperty("bpf.summary.orbit.tolerance"));
        result = instance.searchCpsMain(startOrbit, endOrbit, BpfSensorEnum.OLI,
                null, status, nearMatch);
        assertNotNull("Null returned for nearMatch orbit query", result);
        assertTrue("No matches found for nearMatch orbit query",
                result.size() > 0);

        result = instance.searchCpsMain(startOrbit, null, BpfSensorEnum.OLI, 
                null, status, nearMatch);
        assertNotNull("Null returned for nearMatch orbit query", result);
        assertTrue("No matches found for nearMatch orbit query",
                result.size() > 0);

        // query with near match, do not expect one
        startOrbit = Long.parseLong(
                System.getProperty("bpf.summary.orbitrange3.start"));
        endOrbit = Long.parseLong(
                System.getProperty("bpf.summary.orbitrange3.end"));
        result = instance.searchCpsMain(startOrbit, endOrbit, BpfSensorEnum.OLI,
                null, status, nearMatch);
        assertNotNull("Null returned for nearMatch orbit range query", result);
        assertTrue("Unexpected matches found for nearMatch orbit range query",
                result.size() == 0);
    }

    /**
     * Test of getCpsSummary method, of class BpfSummaryDAO.
     */
    @Test
    public void testGetCpsSummary() throws Exception {
        logger.info("getCpsSummary");
        BpfSummaryDAO instance = new BpfSummaryDAO(this.conn, this.daoAuth);
        Long bpfMainId = instance.getCpsMainId(System.getProperty("BPF_FILE"));
        CpsSummary result = instance.getCpsSummary(bpfMainId);
        assertEquals("Unexpected file name " + result.getFilename()
                + " for BPF ID " + bpfMainId,
                result.getFilename(), System.getProperty("BPF_FILE"));
    }

    /**
     * Test of getAllCpsList method, of class BpfSummaryDAO.
     */
    @Test
    public void testGetAllCpsList() throws Exception {
        logger.info("getAllCpsList");
        BpfSummaryDAO instance = new BpfSummaryDAO(this.conn, this.daoAuth);
        List<CpsSummary> result = instance.getAllCpsList();
        assertTrue("No files found in getAll", result.size() > 0);
        logger.info("List of all BPFs:");
        for (int i = 0; i < result.size(); i++) {
            logger.info(i + " " + result.get(i).getFilename());
        }
    }

    /**
     * Test of getCpsMainId method, of class BpfSummaryDAO.
     */
    @Test
    public void testGetCpsMainId() throws Exception {
        logger.info("getCpsMainId");
        String filename = System.getProperty("BPF_FILE");
        BpfSummaryDAO instance = new BpfSummaryDAO(this.conn, this.daoAuth);
        Long result = instance.getCpsMainId(filename);
        assertNotNull("ID not found for " + filename, result);
    }

    /**
     * Test of updateCpsStatus method, of class BpfSummaryDAO.
     */
    @Test
    public void testUpdateCpsStatus() throws Exception {
        logger.info("updateCpsStatus");
        String bpfName = System.getProperty("BPF_FILE");
        BpfSummaryDAO instance = new BpfSummaryDAO(this.conn, this.daoAuth);
        
        // Get the original status
        Long bpfMainId = instance.getCpsMainId(bpfName);
        CpsSummary origResult = instance.getCpsSummary(bpfMainId);

        // Set it to a different status, then back to the original value
        String newStatus = origResult.getStatus().equals(CpsStatusEnum.DENIED) ?
            "ACTIVE" : "DENIED";
        instance.updateCpsStatus(bpfName, newStatus);
        assertNotNull("ID not found for " + bpfName, bpfMainId);
        CpsSummary newResult = instance.getCpsSummary(bpfMainId);
        assertNotNull("Summary info not found for ID " + bpfMainId, newResult);
        assertEquals("DB status does not match file status",
                newResult.getStatus().toString(), newStatus);

        // update with an invalid status, should throw an exception
        try {
            instance.updateCpsStatus(bpfName, "XYZ");
            fail("Error - Update should have thrown an exception");
        } catch (Exception e) {
            // ignore the exception
        }

        // Set back to the original status
        instance.updateCpsStatus(bpfName, origResult.getStatus().toString());
    }

    /**
     * Test suite, for running the tests with an old version of junit/ant
     **/
    public static junit.framework.Test suite() {
        if (System.getenv("ENABLE_DB_TESTS") == null) {
            System.out.println("Database tests are disabled");
            return new junit.framework.JUnit4TestAdapter(TestSupport.class);
        } else {
            return new junit.framework.JUnit4TestAdapter(
                    BpfSummaryDAOTest.class);
        }
    }
}
