/**
 * CpfDAOTest
 *
 * JUnit tests for CpfDAO class.  All methods with a @Test annoation
 * will be run by the junit 4 test framework.
 */
package gov.usgs.ias.dao.cpf;

import gov.usgs.ias.dao.ParseODL;
import gov.usgs.ias.dao.TestSupport;
import gov.usgs.ias.cps.CpsGroup;
import gov.usgs.ias.cps.CpsStatusEnum;
import gov.usgs.ias.dao.DataConnection;
import gov.nasa.pds.tools.label.*;
import gov.usgs.ias.constant.Constants;
import java.sql.PreparedStatement;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.io.File;
import org.apache.log4j.Logger;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import static org.junit.Assert.*;

public class CpfDAOTest {

    private DataConnection conn;
    private int nGroups = 0;
    private String daoAuth = null;
    private static final Logger logger = Logger.getLogger(Constants.LOGGER_CPF);

    public CpfDAOTest() {
    }

    @BeforeClass
    public static void setUpClass() throws Exception {
    }

    @AfterClass
    public static void tearDownClass() throws Exception {
    }

    /**
     * Setup to be done before each test.
     * N.B.: these unit tests assume the standard set of test CPFs
     * have been loaded into the database.
     *
     * Put the log4j.properties in the test folder, so it gets copied to the
     * ias/build/test/classes folder when built.  This folder is in the
     * runtime classpath so will find the properties file there.
     */
    @Before
    public void setUp() {
        this.conn = new DataConnection("CPF_TEST_DB");

        // Set up the authentication string for the proper schema, default PROD
        if (System.getProperty("SCHEMA") != null &&
            System.getProperty("SCHEMA").equals("EVAL")) {
            daoAuth = Constants.DAO_AUTH_IAS_EVAL;
        } else {
            daoAuth = Constants.DAO_AUTH_IAS_PROD;
        }
    }

    /**
     * Cleanup to be done after each test.
     */
    @After
    public void tearDown() {
        this.conn.close();
    }

    /**
     * Test of getCpfGroups method, of class CpfDAO.
     */
    @Test
    public void testGetCpfGroups() throws Exception {
        logger.debug("getCpfGroups");
        String cpfName = System.getProperty("CPF_FILE");
        CpfDAO instance = new CpfDAO(this.conn, this.daoAuth);

        // Get the groups from the database
        Collection<CpsGroup> result = instance.getCpsGroups(cpfName);
        assertNotNull("Null returned getting groups for " + cpfName, result);
        nGroups = result.size();
        assertTrue("No groups found for " + cpfName, nGroups > 0);

        // Parse the groups from the file
        String cpfFileToParse = System.getProperty("CPF_FILE_DIR")
                + File.separator
                + System.getProperty("CPF_FILE");
        Label cpfLabel = ParseODL.parseOdlFile(cpfFileToParse);
        assertNotNull("Error parsing " + cpfFileToParse, cpfLabel);

        // Compare the group names between db and file
        List odlGroups = cpfLabel.getGroups();
        GroupStatement odlGroup;
        Iterator<CpsGroup> dbGroups = result.iterator();
        String dbGroup;
        int i = 0;
        while (dbGroups.hasNext()) {
            dbGroup = dbGroups.next().getName();
            odlGroup = (GroupStatement)odlGroups.get(i++);
            assertTrue("Invalid group name: " + dbGroup,
                    dbGroup.equals(odlGroup.getIdentifier()));
        }
    }

    /**
     * Test of getCpfGroupDataStmt method, of class CpfDAO.
     */
    @Test
    public void testGetCpfGroupDataStmt() throws Exception {
        logger.debug("getCpfGroupDataStmt");
        CpfDAO instance = new CpfDAO(this.conn, this.daoAuth);
        PreparedStatement result = instance.getCpsGroupDataStmt();
        assertNotNull("Null returned from getting group data statement",
                result);
    }

    /**
     * Test of getCpfGroupById method, of class CpfDAO.
     */
    @Test
    public void testGetCpfGroupById() throws Exception {
        logger.debug("getCpfGroupById");
        String cpfName = System.getProperty("CPF_FILE");
        CpfDAO instance = new CpfDAO(this.conn, this.daoAuth);

        // Get all the group IDs for the test file
        Collection<CpsGroup> groups = instance.getCpsGroups(cpfName);
        assertTrue("No groups found for " + cpfName, groups.size() > 0);

        // Get each group by ID, and compare to the list we already got
        boolean addParameters = false;
        long cpfGroupId = 0;
        Iterator<CpsGroup> cgi = groups.iterator();
        String cpfGroupName;
        CpsGroup currentGroup;
        while (cgi.hasNext()) {
            currentGroup = cgi.next();
            cpfGroupId = currentGroup.getGroupId();
            cpfGroupName = currentGroup.getName();
            CpsGroup result = instance.getCpsGroupById(cpfGroupId,
                    addParameters);
            assertNotNull("Null returned getting group " + cpfGroupId,
                    result);
            // compare the ID and group name
            assertTrue("Unexpected group ID " + result.getGroupId(),
                    result.getGroupId() == cpfGroupId);
            assertTrue("Unexpected group name " + result.getName(),
                    result.getName().equals(cpfGroupName));
        }
    }

    /**
     * Test of insertCpf method, of class CpfDAO.
     */
    @Test
    public void testInsertCpf() throws Exception {
        logger.debug("Inserting CPF");
        logger.debug(" Parsing ODL file "
                + System.getProperty("CPF_LOAD_FILE"));
        Label cpfLabel = ParseODL.parseOdlFile(
                System.getProperty("CPF_LOAD_FILE"));
        logger.debug(" Done Parsing ODL file");
        assertNotNull("Error parsing " + System.getProperty("CPF_LOAD_FILE"),
                cpfLabel);
        CpfDAO instance = new CpfDAO(this.conn, this.daoAuth);

        String fileName = instance.getFileAttribute(cpfLabel, "File_Name");
        assertNotNull("No File_Name contained in file", fileName);

        logger.debug(" Inserting file into DB");
        int result = -1;
        Exception insertException = new Exception("");
        try {
            result = instance.insertFile(cpfLabel, CpsStatusEnum.UNTESTED);
        } catch (Exception e) {
            // just save the exception & continue to clean up the db
            insertException = e;
        }
        logger.debug(" Done Inserting file into DB");

        // Delete the data before checking the status of the insert to ensure
        //  it gets cleaned up for next time
        logger.debug("Getting group IDs for " + fileName);
        Collection<CpsGroup> groups = null;
        try {
            groups = instance.getCpsGroups(fileName);
            logger.debug(" found " + groups.size() + " groups");
        } catch (Exception e) {
            // do nothing - there are just no groups to clean up
        }

        // Delete from cpf_main_group,
        //             cpf_main,
        //             cpf_group (may fail if other CPFs are using the group)
        logger.debug("Deleting from cpf_main_group");
        PreparedStatement stmt = null;
        String sql = "DELETE from cpf_main_group a WHERE EXISTS" +
            " (SELECT * FROM cpf_main b " +
            "  WHERE a.cpf_main_id = b.cpf_main_id " +
            "   and b.file_name = '" + fileName + "')";
        stmt = this.conn.getConnection(this.daoAuth).prepareStatement(sql);
        assertFalse("Unexpected results returned from delete statement",
                stmt.execute());  // expecting no result set

        logger.debug("Deleting from cpf_main");
        sql = "DELETE from cpf_main where file_name = '" + fileName + "'";
        stmt = this.conn.getConnection(this.daoAuth).prepareStatement(sql);
        assertFalse("Unexpected results returned from delete statement",
                stmt.execute());

        logger.debug("Deleting from cpf_group");
        if (groups != null && groups.size() > 0) {
            // This one may fail, that's ok
            Iterator<CpsGroup> iter = groups.iterator();
            StringBuffer buffer = new StringBuffer(""
                    + iter.next().getGroupId());
            while (iter.hasNext()) {
                buffer.append(",");
                buffer.append("" + iter.next().getGroupId());
            }
            String groupList = buffer.toString();
            sql = "DELETE from cpf_group a WHERE CPF_GROUP_ID in (" + groupList
                    + ")";
            logger.debug(sql);
            stmt = this.conn.getConnection(this.daoAuth).prepareStatement(sql);
            try {
                stmt.execute();
            } catch (Exception e) {
                // ignore any errors
            }
        }

        // final check on the insert
        assertTrue(insertException.getMessage(), result != -1);
    }

    /**
     * Test suite, for running the tests with an old version of junit/ant
     **/
    public static junit.framework.Test suite() {
        if (System.getenv("ENABLE_DB_TESTS") == null) {
            System.out.println("Database tests are disabled");
            return new junit.framework.JUnit4TestAdapter(TestSupport.class);
        } else {
            return new junit.framework.JUnit4TestAdapter(CpfDAOTest.class);
        }
    }
}
