/**
 * CpfDAOTest
 *
 * JUnit tests for CpfDAO class.  All methods with a @Test annotation
 * will be run by the junit 4 test framework.
 */
package gov.usgs.ias.dao.cpf;

import gov.usgs.ias.dao.ParseODL;
import gov.usgs.ias.dao.TestSupport;
import gov.usgs.ias.cps.CpsStatusEnum;
import gov.usgs.ias.dao.DataConnection;
import gov.nasa.pds.tools.label.*;
import gov.usgs.ias.constant.Constants;
import java.sql.PreparedStatement;
import org.apache.log4j.Logger;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import static org.junit.Assert.*;

public class CpfReplaceTest {

    private DataConnection conn;
    private String daoAuth = null;
    private static final Logger logger = Logger.getLogger(Constants.LOGGER_CPF);

    // This test uses a property to define the file to replace.  When
    // run through the test suite, this is normally not defined so this 
    // test is a no-op.  When run from the command line, the file is 
    // provided as an argument.
    private static final String FILE_PROPERTY = "CPF_REPLACE_TEST_FILE";

    public CpfReplaceTest() {
    }

    @BeforeClass
    public static void setUpClass() throws Exception {
    }

    @AfterClass
    public static void tearDownClass() throws Exception {
    }

    /**
     * Setup to be done before each test.
     * N.B.: these unit tests assume the standard set of test CPFs
     * have been loaded into the database.
     *
     * Put the log4j.properties in the test folder, so it gets copied to the
     * ias/build/test/classes folder when built.  This folder is in the
     * runtime classpath so will find the properties file there.
     */
    @Before
    public void setUp() {
        this.conn = new DataConnection("CPF_TEST_DB");

        // Set up the authentication string for the proper schema, default EVAL
        if (System.getProperty("SCHEMA") != null &&
            System.getProperty("SCHEMA").equals("PROD")) {
            daoAuth = Constants.DAO_AUTH_IAS_PROD;
        } else {
            daoAuth = Constants.DAO_AUTH_IAS_EVAL;
        }
    }

    /**
     * Cleanup to be done after each test.
     */
    @After
    public void tearDown() {
        this.conn.close();
    }

    /**
     * Replace a CPF: Delete it from the database, then re-insert it.
     */
    @Test
    public void testReplaceCpf() throws Exception {
        // This test will be a no-op unless a file name is provided
        if (System.getProperty(FILE_PROPERTY) == null) {
            logger.debug("No file specified");
            return;
        }
        logger.debug("Replacing CPF");
        
        // ---------- First make sure we can parse the new file ---------------
        logger.debug("** Parsing ODL file "
                + System.getProperty(FILE_PROPERTY));
        Label cpfLabel = ParseODL.parseOdlFile(
                System.getProperty(FILE_PROPERTY));
        logger.debug("** Done Parsing ODL file");
        assertNotNull("Error parsing " + System.getProperty(FILE_PROPERTY),
                cpfLabel);
        
        CpfDAO instance = new CpfDAO(this.conn, this.daoAuth);
        String fileName = instance.getFileAttribute(cpfLabel, "File_Name");
        assertNotNull("No File_Name contained in file", fileName);

        // ---------- Delete the old file -------------------------------------
        // Delete from cpf_main_group,
        //             cpf_main
        // Leave the rows in cpf_group, as they may be used by another file
        logger.debug("** Deleting existing " + fileName + " from DB");
        logger.debug("Deleting from cpf_main_group");
        PreparedStatement stmt = null;
        String sql = "DELETE from cpf_main_group a WHERE EXISTS" +
            " (SELECT * FROM cpf_main b " +
            "  WHERE a.cpf_main_id = b.cpf_main_id " +
            "   and b.file_name = '" + fileName + "')";
        stmt = this.conn.getConnection(this.daoAuth).prepareStatement(sql);
        assertFalse("Unexpected results returned from delete statement",
            stmt.execute());  // expecting no result set

        logger.debug("Deleting from cpf_main");
        sql = "DELETE from cpf_main where file_name = '" + fileName + "'";
        stmt = this.conn.getConnection(this.daoAuth).prepareStatement(sql);
        assertFalse("Unexpected results returned from delete statement",
            stmt.execute());
        
        this.conn.getConnection(this.daoAuth).commit();
        logger.debug("** Done deleting " + fileName + " from DB");
        
        // ---------- Insert the new file -------------------------------------
        logger.debug("** Inserting file into DB");
        int result = -1;
        Exception insertException = new Exception("");
        try {
            result = instance.insertFile(cpfLabel, CpsStatusEnum.ACTIVE);
        } catch (Exception e) {
            // just save the exception & continue to clean up the db
            insertException = e;
        }
        // final check on the insert
        assertTrue(insertException.getMessage(), result != -1);
        logger.debug("** Done Inserting file into DB");
    }

    /**
     * Test suite, for running the tests with an old version of junit/ant
     **/
    public static junit.framework.Test suite() {
        if (System.getenv("ENABLE_DB_TESTS") == null) {
            System.out.println("Database tests are disabled");
            return new junit.framework.JUnit4TestAdapter(TestSupport.class);
        } else {
            return new junit.framework.JUnit4TestAdapter(CpfReplaceTest.class);
        }
    }
    
    /**
     * Method to allow for running from the command-line, specifying the
     * name of the file to load
     **/
    public static void main(String args[]) {
        if (args.length < 2) {
            System.err.println("Usage: java <classpath> " +
                    "gov.usgs.ias.dao.cpf.CpfReplaceTest <filename> <schema>");
            System.exit(1);
        }
        System.setProperty(FILE_PROPERTY, args[0]);
        System.setProperty("SCHEMA", args[1]);
        org.junit.runner.JUnitCore.main(CpfReplaceTest.class.getName());
    }
    
}
