/***************************************************************************
 NAME:                    angle_prep

 PURPOSE:      Module implementing routines to prepare the solarcal_oli
               application to determine the solar zenith and azimuth angles

 ROUTINES:     convert_datetime_to_year_doy_sod
               get_julian_date_for_jpl_ephem
               process_jpl_ephemeris
****************************************************************************/
#include "ias_math.h"
#include "ias_structures.h"
#include "ias_geo.h"
#include "solarcal_oli.h"


/***********************************************************************
 NAME:     convert_datetime_to_year_doy_sod

 PURPOSE:  Given a populated IAS_DATETIME data structure, extracts the
           year component and determines the corresponding day-of-year
           and seconds-of-day from the remaining DATETIME data
           structure components.

 RETURNS:  Integer status of SUCCESS or ERROR
************************************************************************/
int convert_datetime_to_year_doy_sod
(
    IAS_DATETIME datetime,          /* I: Parsed IAS date/time string */
    double *year,                   /* O: Year */
    double *doy,                    /* O: Day-of-year */
    double *sod                     /* O: Second-of-day */
)
{
    double calc_sod;
    int calc_doy;
    int status;

    /* Extract the year component from the date/time component. */
    *year = datetime.year;

    /* Convert the month and day-of-month for the current year to the
       corresponding day of the year. */
    status = ias_math_convert_month_day_to_doy(datetime.month,
        datetime.day_of_month, datetime.year, &calc_doy);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Converting month / day-of-month information to "
            "day of year");
        return ERROR;
    }
    *doy = calc_doy;

    /* Calculate the total number of seconds corresponding to the current
       time. */
    calc_sod = (double)(datetime.hour * 3600)
        + (double)(datetime.minute * 60) + datetime.second;

    *sod = calc_sod;

    return SUCCESS;
}   /* END -- convert_start_time_year_doy_sod */



/*-----------------------------------------------------------------------
 NAME:            get_julian_date_for_jpl_ephem

 PURPOSE:  Calculates the Julian date for the JPL solar ephemeris
           data

 RETURNS:  Julian date
-------------------------------------------------------------------------*/
double get_julian_date_for_jpl_ephem
(
    const char *date_string,          /* I: String representation of date */
    IAS_DATETIME_FORMAT_TYPE  format  /* I: Date/time string format */
)
{
    IAS_DATETIME datetime;         /* Data structure containing date/time
                                      components */
    int status;

    double secs_in_day;              /* Current number of seconds in the
                                      day based on the current acquisition
                                      date/time */

    double julian_date;            /* Corresponding Julian date */


    /* Clear out the data structure */
    memset(&datetime, 0, sizeof(IAS_DATETIME));

    /* Parse the acquisition date/time string into its components */
    status
        = ias_misc_parse_datetime_string(date_string, format, &datetime);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Parsing acquisition date failed");
        return (double)ERROR;
    }

    /* Calculate the total # of seconds corresponding to the current time */
    secs_in_day
        = (double)(datetime.hour * 3600)
              + (double)(datetime.minute * 60) + datetime.second;

    /* Calculate the full Julian date */
    julian_date = ias_math_compute_full_julian_date(datetime.year,
                                                    datetime.month,
                                                    datetime.day_of_month,
                                                    secs_in_day);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating Julian date for JPL epehemeris "
                      "processing failed");
        return (double)ERROR;
    }

    /* Done */
    return julian_date;
}



/*-----------------------------------------------------------------------
 NAME:               process_jpl_ephemeris

 PURPOSE:   Processes the JPL ephemeris information to obtain the
            earth-sun distance and the instantaneous solar position
            vector for the specified Julian date

 RETURNS:   Integer status code of SUCCESS or ERROR
-------------------------------------------------------------------------*/
int process_jpl_ephemeris
(
    const double jpl_ephem_date,           /* I: Requested Julian date */
    double *earth_sun_distance,            /* O: Derived outputs from
                                              JPL solar ephemeris data */
    IAS_VECTOR *solar_position_vector
)
{
    int status;
    double right_ascension;
    double declination;

    /* Initialize the NOVAS library */
    if (ias_geo_initialize_novas() != SUCCESS)
    {
        IAS_LOG_ERROR("Cannot initialize NOVAS library");
        return ERROR;
    }

    /* Get the earth-sun distance first */
    status = ias_geo_compute_sun_position(jpl_ephem_date, &right_ascension,
                &declination, earth_sun_distance);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Cannot retrieve Earth-sun distance");
        return ERROR;
    }

    /* Determine the solar position vector components */
    status = ias_geo_get_solar_position_vector(jpl_ephem_date,
        solar_position_vector);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Cannot determine instantaneous solar position vector");
        return ERROR;
    }

    /* Clean up by "decommisioning" the NOVAS library / JPL ephemeris */
    if (ias_geo_shutdown_novas() != SUCCESS)
    {
        IAS_LOG_ERROR("Failed when shutting down the NOVAS library");
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}
