/**************************************************************************
 NAME:                      calculate_gains

 PURPOSE:       Derives radiance-based band absolute gains and detector
                relative gains, or reflectance-based band absolute gains,
                based on bias-corrected and linearized 
                histogram statistics obtained from the OLI solar diffuser

 RETURNS:       Status code of SUCCESS or ERROR

***************************************************************************/
#include <math.h>

#include "solarcal_oli.h"


/*----------------------------------------------------------------------
 NAME:            calculate_detector_gain_statistics

 PURPOSE:    calculates a gain statistic for each operable detector
             as follows:

             G[d]   =                Q[d] * (d_es)^2
                                 -----------------------
                                 cos_theta * Diffuser[d]

             G_sigma[d]  =        Q_sigma[d] * (d_es)^2
                                 ------------------------
                                 cos_theta * Diffuser[d] 

             where 

             Q[d]       = histogram statistics mean for the current
                          (operable) detector
             Q_sigma[d] = histogram statistics std deviation
                          for the current (operable) detector

             d_es is the Earth-Sun distance correction factor
             applicable during the diffuser data collect

             cos_theta is the cosine of the estimated solar zenith
             angle.  For radiance-based gains, this value should be
             set to 1.0.

             L_diffuser[d] is the diffuser radiance (reflectance) obtained
             from the CPF for the current (operable) detector

 RETURNS:    on successful completion:  Array (of type double) with detector
             gains (std deviations), and "flag" values for any inoperable
             detectors

             error condition:     NULL pointer
------------------------------------------------------------------------*/
static double *calculate_detector_gain_statistics
(
    double *hist_stats_data,         /* I: Histogram statistics data array
                                        (either means or std deviations) */
    double *diffuser_radiances,      /* I: CPF-based detector radiance array
                                        observed at diffuser panel for the
                                        band */
    double earth_sun_distance_sqr,   /* I: Earth-sun distance factor during
                                        acquisition */
    double cos_theta,                /* I: Cosine of solar zenith angle
                                        (1.0 for radiance-based gains) */
    int num_band_detectors           /* I: Total number of detectors in
                                        the band */
)
{
    int detector_index;

    double *gains = NULL;            /* Returned detector gain array
                                        (or gain std deviations array) */


    /* Allocate return array */
    gains = calloc(num_band_detectors, sizeof(*gains));
    if (gains == NULL)
    {
        IAS_LOG_ERROR("Cannot allocate detector gain array");
        return NULL;
    }

    /* Calculate the absolute detector gains */
    for (detector_index = 0;
         detector_index < num_band_detectors;
         detector_index++)
    {
        double gain_numerator
            = hist_stats_data[detector_index] * earth_sun_distance_sqr;
        double gain_denominator
            = diffuser_radiances[detector_index] * cos_theta;
        gains[detector_index] = gain_numerator / gain_denominator;
    }

    /* Finished. */
    return gains;
}




/*----------------------------------------------------------------------
 NAME:               calculate_gains

 PURPOSE:    "Driver" routine for calculating detector relative gains
             and radiance-based band average absolute gains

 RETURNS:    Integer status code of SUCCESS or ERROR
------------------------------------------------------------------------*/
int calculate_gains
(
    double *hist_stats_mean,       /* I: Histogram statistics means for all
                                      detectors in the current band */
    double *hist_stats_stdev,      /* I: Histogram statistics std dev array
                                      for all detectors the current band */
    double *diffuser_radiances,    /* I: CPF-based diffuser radiances
                                      (reflectances) for all detectors in the
                                      current band */
    const int *statistics_invalid_flags,
                                   /* I: Detector valid data flags for the
                                      current band (0=valid data,
                                      nonzero=invalid data) */
    int num_band_detectors,        /* I: Total number of detectors in the
                                      current band */
    double earth_sun_distance_sqr, /* I: Earth-sun distance
                                      factor (squared) for the current
                                      collection.  Intended to come from
                                      JPL solar ephemeris data */
    double cos_theta,              /* I: Cosine of solar zenith angles,
                                      (1.0 for radiance-based gains) */
    double *relative_gains,        /* O: Radiance-based relative gain array
                                      for all detectors in the current band
                                      (NULL if calculating reflectance-based
                                      band absolute gains) */
    double *relative_gains_stdev,  /* O: Radiance-based relative gain stdev
                                      array for all detectors in the current
                                      band (NULL if calculating
                                      reflectance-based band absolute gains)*/
    double *abs_band_gain,         /* O: Absolute radiance-based gain for
                                       the current band */
    double *abs_band_gain_stdev    /* O: Absolute radiance-based gain
                                      stdev for the current band */
)
{
    int detector_index;
    int status;

    double *abs_detector_gain = NULL;  /* Absolute detector gain array
                                          for the current band*/
    double *abs_detector_gain_stdev = NULL;  /* Absolute detector gain
                                                stdev array for the
                                                current band */


    /* Sanity check -- do we have valid data buffers ? */
    if ((hist_stats_mean == NULL) || (hist_stats_stdev == NULL)
        || (diffuser_radiances == NULL)
        || (statistics_invalid_flags == NULL))
    {
        IAS_LOG_ERROR("Invalid input histogram statistics and/or diffuser "
                      "data buffers");
        return ERROR;
    }

    /* Calculate the absolute detector gains first */
    abs_detector_gain
        = calculate_detector_gain_statistics(hist_stats_mean,
                                             diffuser_radiances,
                                             earth_sun_distance_sqr,
                                             cos_theta,
                                             num_band_detectors);
    if (abs_detector_gain == NULL)
    {
        IAS_LOG_ERROR("Absolute detector gain calculation failed");
        return ERROR;
    }

    abs_detector_gain_stdev
        = calculate_detector_gain_statistics(hist_stats_stdev,
                                             diffuser_radiances,
                                             earth_sun_distance_sqr,
                                             cos_theta,
                                             num_band_detectors);
    if (abs_detector_gain_stdev == NULL)
    {
        IAS_LOG_ERROR("Absolute detector gain std deviation calculation "
                      "failed");
        free(abs_detector_gain);
        return ERROR;
    }

    /* Calculate the band average absolute gain and std deviation */
    status = ias_rps_calculate_filtered_mean(abs_detector_gain,
                                             statistics_invalid_flags,
                                             num_band_detectors,
                                             abs_band_gain, NULL);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Cannot calculate radiance gain mean");
        free(abs_detector_gain);
        free(abs_detector_gain_stdev);
        return ERROR;
    }

    status = ias_rps_calculate_filtered_stdev(abs_detector_gain,
                                              statistics_invalid_flags,
                                              num_band_detectors,
                                              *abs_band_gain,
                                              abs_band_gain_stdev);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Cannot calculate radiance gain std deviations");
        free(abs_detector_gain);
        free(abs_detector_gain_stdev);
        return ERROR;
    }

    /* If we're doing radiance-based gain calculations, also
       calculate relative gains */
    if ((relative_gains != NULL) && (relative_gains_stdev != NULL))
    {
        for (detector_index = 0;
             detector_index < num_band_detectors;
             detector_index++)
        {
            relative_gains[detector_index]
                = abs_detector_gain[detector_index]
                      / *abs_band_gain;
            relative_gains_stdev[detector_index]
                = abs_detector_gain_stdev[detector_index]
                      / *abs_band_gain;
        }

        IAS_LOG_DEBUG("Radiance-based absolute and relative gains "
                      "successfully calculated");
    }
    else
        IAS_LOG_DEBUG("Reflectance-based absolute gains successfully "
                      "calculated");

    /* Clean up */
    free(abs_detector_gain);
    free(abs_detector_gain_stdev);

    /* Done */
    return SUCCESS;
}
