/**************************************************************************
 NAME:      calculate_solar_angles

 PURPOSE:   Routines to read the ancillary data file to obtain the attitude
            and ephemeris information and calculate the observed
            zenith and azimuth angles of the diffuser

 ROUTINES:  calculate_sample_angles
            calculate_solar_angles
**************************************************************************/
#include <math.h>
#include <float.h>
#include "ias_ancillary_io.h"
#include "ias_math.h"
#include "ias_structures.h"
#include "solarcal_oli.h"

#define NOMINAL_VIEW_ANGLE 45.0



/**************************************************************************
 NAME:       calculate_sample_angles

 PURPOSE:    Calculates the 'Incident' and 'View' solar zenith/azimuth angles
             for each sample of ADP-corrected ACS quaternions acquired during
             the solar collect. 

 RETURNS:    Integer status code of SUCCESS or ERROR
***************************************************************************/
int calculate_sample_angles
(
    const IAS_ANC_ATTITUDE_DATA *attitude_data,
                                         /* I: Sensor attitude data records
                                            (ACS quaternions) */
    int start_record_index,              /* I: Index of 1st quaternion
                                            record sample acquired closest
                                            to image start time */
    int number_of_records,               /* I: Number of quaternion
                                            samples during actual solar
                                            collect */
    const IAS_VECTOR* solar_position,    /* I: JPL solar ephemeris data
                                            applicable to the current
                                            acquisition */
    double *incident_solar_zenith_angle, /* O: Populated instantaneous
                                            incident solar zenith angle
                                            array */
    double *view_solar_zenith_angle,     /* O: Populated instantaneous
                                            view solar zenith angle array */
    double *view_solar_azimuth_angle     /* O: Populated instantaneous
                                            view solar azimuth angle array */
)
{
    /* Determine the constants to convert from degrees to radians
       (and vice versa). */
    double degrees_per_radian = ias_math_get_degrees_per_radian();
    double radians_per_degree = ias_math_get_radians_per_degree();

    double earth_to_sc_transform_matrix[3][3];  /* Earth-to-spacecraft
                                                   transformation matrix */

    /* Constant vectors applicable for all records */
    IAS_VECTOR diffuser_normal;       /* Diffuser normal vector */
    IAS_VECTOR view_normal;           /* View normal vector */
    IAS_VECTOR dn_cross_vd;           /* Cross product of diffuser normal
                                         and view normal unit vectors */

    double vector_length; /* Length of a given vector (i.e.
                             ||v|| = sqrt(v_x^2 + v_y^2 + v_z^2)) */
    double cos_vz;        /* Cosine of the view solar zenith angle */
    double sin_vz;        /* Sine of the view solar zenith angle */
    double sin_sq_vz;     /* Squared sine of the view zenith angle */

    int record;           /* 0-based attitude record index */


    /* Set up the view normal unit vector */
    view_normal.x = 0.0;
    view_normal.y = 0.0;
    view_normal.z = 1.0;

    /* Set up the diffuser normal unit vector.  This should be
       constant through a collect.  The panel is currently assumed
       to be oriented at an angle of app. 45.0 degrees relative to the
       sensor.  If there are significant alignment issues with the
       panel and/or variations in sensor view angle wrt to the boresight,
       this logic will need to be modified (possibly including making
       this calculation record-specific).  */
    diffuser_normal.x = 0.0;
    diffuser_normal.y = -cos(radians_per_degree * NOMINAL_VIEW_ANGLE);
    diffuser_normal.z = -sin(radians_per_degree * NOMINAL_VIEW_ANGLE);

    /* Make 'diffuser_normal' a unit vector. */
    vector_length = ias_math_compute_vector_length(&diffuser_normal);
    if (vector_length <= DBL_EPSILON)
    {
        IAS_LOG_ERROR("Zero-length vector magnitude for diffuser "
            "normal vector");
        return ERROR;
    }
    diffuser_normal.x /= vector_length;
    diffuser_normal.y /= vector_length;
    diffuser_normal.z /= vector_length;

    /* Calculate the cross-product of the diffuser normal and
       view normal unit vectors.  Again, in the absence of
       diffuser misalignments and/or boresight variations, this
       calculation stays essentially constant throughout the
       collect, and could (in principle) be pulled out of this loop. */
    ias_math_compute_3dvec_cross(&diffuser_normal, &view_normal,
        &dn_cross_vd);

    /* Make 'dn_cross_vd' a unit vector. */
    vector_length = ias_math_compute_vector_length(&dn_cross_vd);
    if (vector_length <= DBL_EPSILON)
    {
        IAS_LOG_ERROR("Zero-length vector magnitude for cross product "
            "of panel normal and view normal vectors");
        return ERROR;
    }
    dn_cross_vd.x /= vector_length;
    dn_cross_vd.y /= vector_length;
    dn_cross_vd.z /= vector_length;

    /* Calculate the cosine and sine of the view solar zenith angle.  The
       cosine of the view solar zenith angle is just the dot product of the
       view normal unit vector ('vn') and the diffuser normal unit
       vector ('dn').  In this case, however, the spacecraft is
       oriented such that the 'pitch' (Y) axis is aligned towards the
       sun, and the sensor view (Z) axis is aligned in the opposite
       direction with respect to its normal orientation; it must be
       assumed the 'roll' (X) axis is also aligned opposite its normal
       orientation.  Consequently, a minus sign is required.

       cos_vz = -dn_x*vn_x - dn_y*vn_y - dn_z*vn_z

       NOTE:  Unless there are significant misalignment issues wrt to
       the panel(s) and/or variations in sensor view angle along the
       boresight, this can be essentially considered constant through
       the entire collect.  In that case, these calculations could
       be pulled out of this loop and done just once.

       NOTE:  Before proceeding with the calculation of the squared sine,
       sanity check the magnitude of the cosine value to ensure it's not
       effectively 0.  The angle calculations in this routine use the
       'atan' function with the argument being the ratio 'sin/cos'.  'atan'
       returns the corresponding arctangent value of the argument, which
       ranges from -pi/2 to pi/2 radians; this calculation will fail if the
       cosine value is effectively 0.  */
    cos_vz = -ias_math_compute_3dvec_dot(&view_normal, &diffuser_normal);
    if (fabs(cos_vz) <= DBL_EPSILON)
    {
        IAS_LOG_ERROR("Cosine of view solar zenith angle is effectively 0 "
            "-- %lf", cos_vz);
        return ERROR;
    }
    sin_sq_vz = 1.0 - (cos_vz * cos_vz);
    if (sin_sq_vz < 0)
    {
        IAS_LOG_ERROR("squared sine of view zenith angle < 0");
        return ERROR;
    }
    sin_vz = sqrt(sin_sq_vz);

    /* Print the cosine and sine of the view solar zenith angle for
       debugging purposes. */
    IAS_LOG_DEBUG("cosine of view solar zenith angle:  %lf", cos_vz);
    IAS_LOG_DEBUG("sine of view solar zenith angle:    %lf", sin_vz);

    /*  Calculate the instantaneous solar zenith and azimuth angles
        for each attitude quaternion record. */
    for (record = 0; record < number_of_records; record++)
    {
        IAS_VECTOR s_prime;               /* Transformed solar position
                                             vector */
        IAS_VECTOR dn_cross_s_prime;      /* Cross product of the
                                             diffuser normal unit vector
                                             and the transformed solar
                                             position unit vector */

        double cos_iz;      /* Cosine of the incident solar zenith angle */
        double sin_iz;      /* Sine of the incident solar zenith angle */
        double sin_sq_iz;   /* Squared sine of the incident solar zenith
                               angle */
        double cos_va;      /* Cosine of the view solar azimuth angle */
        double sin_va;      /* Sine of the view solar azimuth angle */
        double sin_sq_va;   /* Squared sine of the view solar azimuth
                               angle */

        /* Quaternion-related quantities */
        double q0;          /* 'Scalar' quaternion component */
        double q1;          /* X component of quaternion vector */
        double q2;          /* Y component of quaternion vector */
        double q3;          /* Z component of quaternion vector */
        double q1_sq;       /* Square of X component */
        double q2_sq;       /* Square of Y component */
        double q3_sq;       /* Square of Z component */
        double q0_q1;       /* Product of scalar and X component */
        double q0_q2;       /* Product of scalar and Y component */
        double q0_q3;       /* Product of scalar and Z component */
        double q1_q2;       /* Product of X and Y components */
        double q1_q3;       /* Product of X and Z components */
        double q2_q3;       /* Product of Y and Z components */

        /* Index for current quaternion record.  The base index is offset by
           the index for the first quaternion record timestamped within the
           active imaging interval of the solar collect. */
        int curr_record = record + start_record_index;

        /* Get the quaternion components for the current record, then
           calculate required intermediate quantities from them.
           In the file 'ias_ancillary_convert_imu_to_attitude.c', the
           quaternion components are as follows:
           eci_quaternion[0] -> X component of quaternion vector
           eci_quaternion[1] -> Y component of quaternion vector
           eci_quaternion[2] -> Z component of quaternion vector
           eci_quaternion[3] -> 'Scalar' quaternion component */
        q0 = attitude_data->records[curr_record].eci_quaternion[3];
        q1 = attitude_data->records[curr_record].eci_quaternion[0];
        q2 = attitude_data->records[curr_record].eci_quaternion[1];
        q3 = attitude_data->records[curr_record].eci_quaternion[2];
        q1_sq = q1 * q1;
        q2_sq = q2 * q2;
        q3_sq = q3 * q3;
        q0_q1 = q0 * q1;
        q0_q2 = q0 * q2;
        q0_q3 = q0 * q3;
        q1_q2 = q1 * q2;
        q1_q3 = q1 * q3;
        q2_q3 = q2 * q3;

        /* Compute the earth-to-spacecraft transformation matrix for the
           current quaternion set.  As of the time of implementation, the
           Cal/Val ADD and prototype code access the elements of this matrix
           by column, row (as used in IDL).  The elements here are
           transposed so that they are accessed by row, column (as used in
           C). */
        earth_to_sc_transform_matrix[0][0] = 1.0 - (2.0 * (q2_sq + q3_sq));
        earth_to_sc_transform_matrix[1][1] = 1.0 - (2.0 * (q1_sq + q3_sq));
        earth_to_sc_transform_matrix[2][2] = 1.0 - (2.0 * (q1_sq + q2_sq));
        earth_to_sc_transform_matrix[0][1] = 2.0 * (q1_q2 + q0_q3);
        earth_to_sc_transform_matrix[0][2] = 2.0 * (q1_q3 - q0_q2);
        earth_to_sc_transform_matrix[1][0] = 2.0 * (q1_q2 - q0_q3);
        earth_to_sc_transform_matrix[1][2] = 2.0 * (q0_q1 + q2_q3);
        earth_to_sc_transform_matrix[2][0] = 2.0 * (q0_q2 + q1_q3);
        earth_to_sc_transform_matrix[2][1] = 2.0 * (q2_q3 - q0_q1);

        /* Apply the transformation matrix to the initial solar position
           vector. */
        ias_math_transform_3dvec(solar_position,
            earth_to_sc_transform_matrix, &s_prime);

        /* Calculate the magnitude of the transformed solar position vector,
           so we can determine the corresponding unit vector. */
        vector_length = ias_math_compute_vector_length(&s_prime);
        if (vector_length <= DBL_EPSILON)
        {
            IAS_LOG_ERROR("Zero-length vector magnitude for transformed "
                "solar position vector 's_prime'");
            return ERROR;
        }
        s_prime.x /= vector_length;
        s_prime.y /= vector_length;
        s_prime.z /= vector_length;

        /* Calculate the cosine and sine of the incident solar zenith
           angle.  The cosine of the incident solar zenith angle is
           just the dot product of the diffuser normal unit vector
           ('dn') and the transformed solar position unit vector ('s_prime'),
           i.e.

           cos_iz = dn_x*s_prime_x + dn_y*s_prime_y + dn_z*s_prime_z

           NOTE:  Whenever the cosine of an angle is calculated with
                  a dot product (through a call to
                  ias_math_compute_3dvec_dot), the input vectors should
                  be converted to unit vectors first.       */
        cos_iz = ias_math_compute_3dvec_dot(&diffuser_normal, &s_prime);
        if (fabs(cos_iz) <= DBL_EPSILON)
        {
            IAS_LOG_ERROR("Cosine of incident solar zenith angle is "
                "effectively 0 -- %lf (record %d)", cos_iz, (record + 1));
            return ERROR;
        }
        sin_sq_iz = 1.0 - (cos_iz * cos_iz);
        if (sin_sq_iz < 0)
        {
            IAS_LOG_ERROR("Squared sine of incident solar zenith angle < 0");
            return ERROR;
        }
        sin_iz = sqrt(sin_sq_iz);

        /* Print the cosine and sine of the incident solar zenith angle
           for debugging purposes. */
        IAS_LOG_DEBUG("cosine of incident solar zenith angle:  %lf",
            cos_iz);
        IAS_LOG_DEBUG("sine of incident solar zenith angle:    %lf",
            sin_iz);

        /* Calculate the incident solar zenith angle.  */
        incident_solar_zenith_angle[record] =
            degrees_per_radian * atan(sin_iz / cos_iz);

        /* Calculate the view solar zenith angle.  Even though the sin_vz
           and cos_vz terms were calculated once (before entering this
           processing loop), the record-specific calculation still needs
           to be done here. */
        view_solar_zenith_angle[record] =
            degrees_per_radian * atan(sin_vz / cos_vz);

        /* Determine the cross-product of the diffuser normal unit
           vector and the transformed solar position unit vector. */
        ias_math_compute_3dvec_cross(&diffuser_normal, &s_prime,
            &dn_cross_s_prime);

        /* Make 'dn_cross_s_prime' a unit vector */
        vector_length = ias_math_compute_vector_length(&dn_cross_s_prime);
        if (vector_length <= DBL_EPSILON)
        {
            IAS_LOG_ERROR("Zero-length vector magnitude for cross product "
                "of panel normal and transformed solar position unit "
                "vectors");
            return ERROR;
        }
        dn_cross_s_prime.x /= vector_length;
        dn_cross_s_prime.y /= vector_length;
        dn_cross_s_prime.z /= vector_length;

        /* Determine the cosine and sine of the view solar azimuth angle.
           The cosine is just the dot product of the 'dn_cross_s_prime' and
           'dn_cross_vd' unit vectors. */
        cos_va = ias_math_compute_3dvec_dot(&dn_cross_s_prime, &dn_cross_vd);
        if (fabs(cos_va) <= DBL_EPSILON)
        {
            IAS_LOG_ERROR("Cosine of view solar azimuth angle is effectively "
                "0 -- %lf (record %d)", cos_va, (record + 1));
            return ERROR;
        }
        sin_sq_va = 1.0 - (cos_va * cos_va);
        if (sin_sq_va < 0)
        {
            IAS_LOG_ERROR("Squared sine of view solar azimuth angle < 0");
            return ERROR;
        }
        sin_va = sqrt(sin_sq_va);

        /* Print the cosine and sine of the view solar azimith angle for
           debugging purposes. */
        IAS_LOG_DEBUG("cosine of view solar azimuth angle:  %lf",
            cos_va);
        IAS_LOG_DEBUG("sine of view solar azimuth angle:    %lf",
            sin_va);

        /* Calculate the view solar azimuth angle. */
        view_solar_azimuth_angle[record] =
            degrees_per_radian * atan(sin_va / cos_va);
    }

    return SUCCESS;
}




/***************************************************************************
 NAME:       calculate_solar_angles

 PURPOSE:    Calculates an overall solar zenith and azimuth angle
             (and their corresponding std deviations) based on the
             available ACS quaternion data of the sensor applicable
             during the collect.

 RETURNS:    Integer status code of SUCCESS or ERROR
****************************************************************************/
int calculate_solar_angles
(
    const char *ancillary_data_filename,       /* I: Ancillary data file */
    double *start_time,                        /* I: Year / day-of-year /
                                                  second-of-day
                                                  corresponding to image
                                                  start time */
    double *stop_time,                         /* I: Year / day-of-year /
                                                  second-of-day
                                                  corresponding to image
                                                  stop time */
    const IAS_VECTOR *solar_position,          /* I: JPL solar ephemeris
                                                  data file name */
    RPS_SOLARCAL_SOLAR_ANGLES  *angles         /* O: Solar angle statistics
                                                  for the collect as a
                                                  whole */
)
{
    /* Attitude data structure */
    IAS_ANC_ATTITUDE_DATA  *attitude_data = NULL;

    /* Ephemeris data structure */
    IAS_ANC_EPHEMERIS_DATA *ephemeris_data = NULL;

    /* Buffers to hold sample-specific angle values */
    double *incident_solar_zenith = NULL;     /* Incident solar zenith
                                                 angle for each ACS
                                                 quaternion */
    double *view_solar_zenith = NULL;         /* View solar zenith angle
                                                 for each ACS quaternion */
    double *view_solar_azimuth = NULL;        /* View solar azimuth angle
                                                 for each ACS quaternion */

    /* Time differences between the UTC epoch time and the image start/stop
       times (in seconds). */
    double delta_t_utc_start;
    double delta_t_utc_stop;

    size_t sample_buffer_size;
    int status;                        /* Function return status */
    int record;
    int number_of_records = 0;
    int start_record_index = 0;


    /* Read the attitude and ephemeris data from the ancillary data
       file.  Since we don't need the ephemeris data, free the
       memory block allocated for it.  Set the pointer to the ephemeris
       block to NULL after it's freed. */
    status = ias_ancillary_read(ancillary_data_filename, &attitude_data,
        &ephemeris_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading ancillary data file %s",
            ancillary_data_filename);
        return ERROR;
    }
    ias_ancillary_free_ephemeris(ephemeris_data);
    ephemeris_data = NULL;

    /* Calculate the deltas between the UTC epoch time and the image
       start and stop times.  Both of these times have year/day-of-year/
       seconds-of-day components.  These calculations should account for
       times that cross the year and/or day-of-year boundaries.

       NOTE:  the image start and stop times are later than the UTC epoch
              time.  To obtain positive time deltas, put the image
              start/stop time as the first parameter in the call to
              ias_math_get_time_difference(), and the UTC epoch time as the
              second parameter.  Otherwise, the time deltas will be negative
              (which in this context doesn't make much if any sense).  */
    status = ias_math_get_time_difference(start_time,
        attitude_data->utc_epoch_time, &delta_t_utc_start);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating difference between UTC epoch time "
            "and image start time");
        ias_ancillary_free_attitude(attitude_data);
        return ERROR;
    }
    status = ias_math_get_time_difference(stop_time,
        attitude_data->utc_epoch_time, &delta_t_utc_stop);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating difference between UTC epoch time "
            "and image stop time");
        ias_ancillary_free_attitude(attitude_data);
        return ERROR;
    }

    /* Determine the starting record number and number of quaternion
       records to use given the current record's seconds-from-epoch
       and the time differences between the UTC epoch time and
       the image start/stop times.

       NOTE:  For 'typical' solar collects (those on the order of 511 frames
       or so), the ancillary data output from AGP may not fall within the
       time boundary defined by the image start and stop times.  If that's
       the case, we'll just have to use the available quaternion information
       that was collected prior to the image start time.  */
    for (record = 0; record < attitude_data->number_of_samples;
         record++)
    {
        double seconds_of_day =
            attitude_data->records[record].seconds_from_epoch;
        if ((seconds_of_day >= delta_t_utc_start)
            && (seconds_of_day <= delta_t_utc_stop))
        {
            if (number_of_records == 0)
            {
                start_record_index = record;
            }
            number_of_records++;
        }
    }
    if (number_of_records == 0)
    {
        IAS_LOG_WARNING("No quaternion data located between image start "
            "and stop times, using nearest available quaternion data");
        start_record_index = 0;
        number_of_records = attitude_data->number_of_samples;
    }

    /* Print the time deltas, the starting record index and the total
       number of records for debugging purposes. */
    IAS_LOG_DEBUG("Delta_T UTC-Start:  %lf", delta_t_utc_start);
    IAS_LOG_DEBUG("Delta_T UTC-Stop:   %lf", delta_t_utc_stop);
    IAS_LOG_DEBUG("Start record index:  %d", start_record_index);
    IAS_LOG_DEBUG("Number of records:   %d", number_of_records);

    /* Determine the required buffer size for all of the angle arrays. */
    sample_buffer_size = number_of_records * sizeof(double);

    /* Allocate the record angle arrays */
    incident_solar_zenith = malloc(sample_buffer_size);
    view_solar_zenith = malloc(sample_buffer_size);
    view_solar_azimuth = malloc(sample_buffer_size);
    if ((incident_solar_zenith == NULL) || (view_solar_zenith == NULL)
           || (view_solar_azimuth == NULL))
    {
        IAS_LOG_ERROR("Allocating attitude sample angle arrays");
        free(incident_solar_zenith);
        free(view_solar_zenith);
        free(view_solar_azimuth);
        ias_ancillary_free_attitude(attitude_data);
        return ERROR;
    }

    /* Calculate the angles for all valid records during the collect.  */
    status = calculate_sample_angles(attitude_data, start_record_index,
        number_of_records, solar_position, incident_solar_zenith,
        view_solar_zenith, view_solar_azimuth);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Deriving zenith and azimuth angles for given "
            "ephemeris data");
        free(incident_solar_zenith);
        free(view_solar_zenith);
        free(view_solar_azimuth);
        ias_ancillary_free_attitude(attitude_data);
        return ERROR;
    }

    /* Calculate the mean incident solar zenith angle and its
       standard deviation. Ideally, this will always be around 45 degrees
       with a very small standard deviation (if any).  Use the IAS math
       library functions to calculate the mean and standard deviation (all
       attitude records are expected to be "valid", whether "good"
       to start with or "corrected") */
    status = ias_math_compute_mean(incident_solar_zenith, number_of_records,
        &angles->incident_solar_zenith);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Deriving mean view zenith angle");
        free(incident_solar_zenith);
        free(view_solar_zenith);
        free(view_solar_azimuth);
        ias_ancillary_free_attitude(attitude_data);
        return ERROR;
    }
    status = ias_math_compute_stdev(incident_solar_zenith, number_of_records,
        angles->incident_solar_zenith, &angles->incident_solar_zenith_stdev);
    free(incident_solar_zenith);
    if (status != SUCCESS)
    {
       IAS_LOG_ERROR("Deriving view zenith angle std deviation");
        free(view_solar_zenith);
        free(view_solar_azimuth);
        ias_ancillary_free_attitude(attitude_data);
        return ERROR;
    }

    /* Calculate the mean view solar zenith angle and its associated
       standard deviation.  Assuming no alignment issues, this should
       be at a constant 45 degrees.*/
    status = ias_math_compute_mean(view_solar_zenith, number_of_records,
        &angles->view_solar_zenith);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Deriving mean solar zenith angle");
        free(view_solar_zenith);
        free(view_solar_azimuth);
        ias_ancillary_free_attitude(attitude_data);
        return ERROR;
    }
    status = ias_math_compute_stdev(view_solar_zenith, number_of_records,
        angles->view_solar_zenith, &angles->view_solar_zenith_stdev);
    free(view_solar_zenith);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Deriving solar zenith angle std deviation");
        free(view_solar_azimuth);
        ias_ancillary_free_attitude(attitude_data);
        return ERROR;
    }

    /* Calculate the mean view solar azimuth angle and its associated
       standard deviation.  */
    status = ias_math_compute_mean(view_solar_azimuth, number_of_records,
        &angles->view_solar_azimuth);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Deriving mean solar azimuth angle");
        free(view_solar_azimuth);
        ias_ancillary_free_attitude(attitude_data);
        return ERROR;
    }
    status = ias_math_compute_stdev(view_solar_azimuth, number_of_records,
        angles->view_solar_azimuth, &angles->view_solar_azimuth_stdev);
    free(view_solar_azimuth);
    ias_ancillary_free_attitude(attitude_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Deriving solar azimuth angle std deviation");
        return ERROR;
    }

    return SUCCESS;
}   /* END -- calculate_solar_angles */
