/**************************************************************************
 NAME:          generate_summary_report_data

 PURPOSE:       Module implementing routines used to copy information
                to a summary report data structure for inclusion in the
                optional summary report and/or dump file

 RETURNS:       Nothing

***************************************************************************/
#include <math.h>
#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "solarcal_oli.h"


/*------------------------------------------------------------------------
 NAME:            initialize_solarcal_parameter_data

 PURPOSE:   Copies selected WO and/or OMF parameter information to the
            summary report data structure for inclusion in the summary
            report and/or dump file

 RETURNS:   Nothing
-------------------------------------------------------------------------*/
static void initialize_solarcal_parameter_data
(
    RPS_SOLARCAL_WO_PARAMS wo_params,            /* I: WO parameters */
    RPS_SOLARCAL_OMF_PARAMS omf_params,          /* I: OMF parameters
                                                    (for work order ID and
                                                   directory information) */
    RPS_SOLARCAL_REPORT_SUMMARY *summary         /* O: Summary report data
                                                    structure for current
                                                    band */
)
{
    /* Populate the summary report data structure with WO information */
    summary->store_char_in_db  = wo_params.store_char_in_db;
    summary->report_flag = wo_params.report_flag;
    summary->dump_flag   = wo_params.dump_flag;
}




/*------------------------------------------------------------------------
 NAME:            initialize_solarcal_angle_data

 PURPOSE:   Copies solar angle information to the summary report data
            structure for inclusion in the summary report and/or dump file

 RETURNS:   Nothing
--------------------------------------------------------------------------*/
static void initialize_solarcal_angle_data
(
   
    RPS_SOLARCAL_SOLAR_ANGLES    solar_angles,     /* I: Solar angle data */
    RPS_SOLARCAL_REPORT_SUMMARY  *summary          /* O: Populated summary
                                                      data structure */
)
{
    /* Populate the angle information */
    summary->solar_angles.earth_sun_distance_factor =
        solar_angles.earth_sun_distance_factor;

    summary->solar_angles.incident_solar_zenith =
        solar_angles.incident_solar_zenith;
    summary->solar_angles.incident_solar_zenith_stdev =
        solar_angles.incident_solar_zenith_stdev;

    summary->solar_angles.view_solar_zenith = solar_angles.view_solar_zenith;
    summary->solar_angles.view_solar_zenith_stdev =
        solar_angles.view_solar_zenith_stdev;

    summary->solar_angles.view_solar_azimuth = solar_angles.view_solar_azimuth;
    summary->solar_angles.view_solar_azimuth_stdev =
        solar_angles.view_solar_azimuth_stdev;
}



/*-----------------------------------------------------------------------
 NAME:                  initialize_solarcal_report_data

 PURPOSE:   Populates the summary report data structure with
            band-independent information

 RETURNS:   Integer status code of SUCCESS or ERROR
-------------------------------------------------------------------------*/
int initialize_solarcal_report_data
(
    RPS_SOLARCAL_WO_PARAMS wo_params,            /* I: WO parameters */
    RPS_SOLARCAL_OMF_PARAMS omf_params,          /* I: OMF parameters */
    RPS_SOLARCAL_SOLAR_ANGLES angles,            /* I: Solar angles */
    RPS_SOLARCAL_REPORT_SUMMARY *summary         /* O: Summary report
                                                    data structure */
)
{
    int dummy_band_list[IAS_MAX_NBANDS];
    int num_oli_bands;
    int status;


    /* Get the total number of OLI bands */
    status = ias_sat_attr_get_sensor_band_numbers(IAS_OLI, IAS_NORMAL_BAND,
        0, dummy_band_list, IAS_MAX_NBANDS, &num_oli_bands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Obtaining total number of OLI bands");
        return ERROR;
    }

    /* Allocate absolute gain array for all OLI bands.  We'll populate
       only the bands we've characterized */
    summary->band_gains =
        calloc(num_oli_bands, sizeof(RPS_SOLARCAL_BAND_GAINS));
    if (summary->band_gains == NULL)
    {
        IAS_LOG_ERROR("Summary report data gain array could not be "
                      "allocated");
        return ERROR;
    }

    /* Initialize the "static" (i.e. not band-dependent) members of
       the summary report data structure */
    initialize_solarcal_parameter_data(wo_params, omf_params, summary);
    initialize_solarcal_angle_data(angles, summary);

    /* Done */
    return SUCCESS;
}




/*------------------------------------------------------------------------
 NAME:            initialize_solarcal_summary_band_data

 PURPOSE:   Copies selected bits of the current band's attribute
            information to the summary report data structure

 RETURNS:   Nothing
--------------------------------------------------------------------------*/
void initialize_solarcal_summary_band_data
(
    int band_index,                                /* I: Current 0-based
                                                      band index */
    RPS_SOLARCAL_BAND_GAINS   current_band_gains,  /* I: Band gain data for
                                                      current band */
    RPS_SOLARCAL_REPORT_SUMMARY  *summary          /* I/O: Populated
                                                      summary report data
                                                      structure */
)
{
    /* Pull out the relevant band attributes */
    strncpy(summary->band_gains[band_index].band_name,
            current_band_gains.band_name,
            strlen(current_band_gains.band_name));
    summary->band_gains[band_index].band_number =
        current_band_gains.band_number;

    summary->band_gains[band_index].scas = current_band_gains.scas;
    summary->band_gains[band_index].detectors_per_sca =
        current_band_gains.detectors_per_sca;

    /* Copy the histogram statistics and gain data */
    summary->band_gains[band_index].hist_stats_mean =
        current_band_gains.hist_stats_mean;
    summary->band_gains[band_index].hist_stats_stdev =
        current_band_gains.hist_stats_stdev;

    summary->band_gains[band_index].diffuser_radiance_mean =
        current_band_gains.diffuser_radiance_mean;
    summary->band_gains[band_index].diffuser_radiance_stdev =
        current_band_gains.diffuser_radiance_stdev;

    summary->band_gains[band_index].radiance_gain =
        current_band_gains.radiance_gain;
    summary->band_gains[band_index].radiance_gain_stdev =
        current_band_gains.radiance_gain_stdev;

    summary->band_gains[band_index].diffuser_reflectance_mean =
        current_band_gains.diffuser_reflectance_mean;
    summary->band_gains[band_index].diffuser_reflectance_stdev =
        current_band_gains.diffuser_reflectance_stdev;

    summary->band_gains[band_index].reflectance_gain =
        current_band_gains.reflectance_gain;
    summary->band_gains[band_index].reflectance_gain_stdev =
        current_band_gains.reflectance_gain_stdev;




}
