/**************************************************************************
  NAME:                   get_parameters

  PURPOSE:  Reads the input parameter files and populates the
            required parameter data structures

  RETURNS:  Integer status code of SUCCESS or ERROR

***************************************************************************/
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "ias_radiometry.h"

#include "solarcal_oli.h"

#define MAX_NUM_WO_PARAMS  8
#define MAX_NUM_OMF_PARAMS 5


int get_parameters
(
    const char *wo_params_filename,          /* I: Work order filename
                                                (or parameter help option) */
    RPS_SOLARCAL_WO_PARAMS *wo_param_data,   /* O: Work order parameters */
    RPS_SOLARCAL_OMF_PARAMS *omf_param_data  /* O: OMF parameters */
)
{
    const char *default_rps_proc_level[] =
        {RPS_SOLARCAL_TEMPCORRECT};
    const char *default_report_filename_prefix[] = {"solarcal"};

    int dummy_band_list[IAS_MAX_NBANDS];
    IAS_SATELLITE_ID satellite_id;
    int default_report_flag = RPS_SOLARCAL_DEFAULT_REPORT_FLAG;
    int default_dump_flag   = RPS_SOLARCAL_DEFAULT_DUMP_FLAG;
    int max_oli_bands;
    int status;                      /* Status for each read
                                        operation */
    int read_count;                  /* Count of parameters
                                        successfully read */

    char omf_params_filename[PATH_MAX];    /* Name of OMF file
                                                     (assumed to be in work
                                                     order directory) */

/*-------------------------------------------------------------------------
                      READ WORK ORDER PARAMETER FILE
---------------------------------------------------------------------------*/
    /* Declare the work order parameter table */
    IAS_PARM_DECLARE_TABLE(wo_param_table, MAX_NUM_WO_PARAMS);

    /* Get common parameters first */
    IAS_PARM_BAND_LIST(wo_param_table, IAS_PARM_REQUIRED,
            wo_param_data->bands,
            sizeof(wo_param_data->bands), 1);

    IAS_PARM_WORK_ORDER_ID(wo_param_table, wo_param_data->work_order_id,
            sizeof(wo_param_data->work_order_id), 1);

    IAS_PARM_STORE_CHARACTERIZATION_IN_DB(wo_param_table,
            IAS_PARM_OPTIONAL,
            &(wo_param_data->store_char_in_db),
            sizeof(wo_param_data->store_char_in_db), 1);

    /* Get the RPS prcessing level information. */
    IAS_PARM_OLI_RPS_PROC_LEVEL_DN(wo_param_table, IAS_PARM_OPTIONAL,
        default_rps_proc_level, wo_param_data->rps_proc_level,
        sizeof(wo_param_data->rps_proc_level));

    /* Read the output options */
    IAS_PARM_ADD_STRING(wo_param_table, REPORT_FILENAME_PREFIX, 
        "Output report file name prefix", IAS_PARM_OPTIONAL,
        0, NULL, /* no restrictions */
        1, default_report_filename_prefix, /* Default file name */
        wo_param_data->report_filename_prefix, 
        sizeof(wo_param_data->report_filename_prefix), 0);

    IAS_PARM_ADD_BOOLEAN(wo_param_table, REPORT_FLAG,
            "Write characterization results to summary reports "
            "(1 = Write reports, 0 = Do not write reports)",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
            &default_report_flag,
            &(wo_param_data->report_flag),
            sizeof(wo_param_data->report_flag), 1);

    IAS_PARM_ADD_BOOLEAN(wo_param_table, DUMP_FLAG,
            "Write characterization results to dump files "
            "(1 = Write dump files, 0 = Do not write dump files)",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
            &default_dump_flag,
            &(wo_param_data->dump_flag),
            sizeof(wo_param_data->dump_flag), 1);

    /* Get the total number of items to read from the parameter table */
    read_count = IAS_PARM_GET_TABLE_SIZE(wo_param_table);

    /* If someone enters -- help, --template, etc, on the command line
       when starting the application, the appropriate information will
       be displayed.  If someone enters a file name, its access will be
       checked.  If a help option was specified or the specified file name
       could not be accessed, a 1 will be returned.  In this case, just
       copy the work order file name (or help option) to the OMF
       file name and continue.  Otherwise, read the work order file */
    status = ias_parm_provide_help(wo_params_filename, wo_param_table,
                 read_count, IAS_INPUT_PARAMETERS);

    if (status == ERROR)
    {
        IAS_LOG_ERROR("Work order parameter help failed");
        return ERROR;
    }
    else if (status == 1)
    {
        sprintf(omf_params_filename, "%s", wo_params_filename);
    }
    else
    {
        status = ias_parm_read(wo_params_filename, wo_param_table,
            read_count);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading work order parameters from file %s",
                          wo_params_filename);
            return ERROR;
        }

        /* How many bands (and possibly SCAs) do we need to process? */
        wo_param_data->num_bands_to_process = IAS_PARM_GET_COUNT_READ(
            wo_param_table, BAND_LIST);

        /* Build the name of the OMF file from the work order ID */
        sprintf(omf_params_filename, "%s.omf", wo_param_data->work_order_id);
    }


/*-----------------------------------------------------------------------
                        READ OMF PARAMETER FILE
-------------------------------------------------------------------------*/
    /* Declare the OMF parameter table */
    IAS_PARM_DECLARE_TABLE(omf_param_table, MAX_NUM_OMF_PARAMS);

    /* Construct the parameter table, common parameters first */
    IAS_PARM_SATELLITE(omf_param_table, &omf_param_data->satellite_number,
            sizeof(omf_param_data->satellite_number), 1);

    IAS_PARM_L0R_FILENAME(omf_param_table, IAS_PARM_REQUIRED,
            omf_param_data->l0r_filename, sizeof(omf_param_data->l0r_filename),
            1);

    IAS_PARM_CPF(omf_param_table, IAS_PARM_REQUIRED,
            omf_param_data->cpf_filename,
            sizeof(omf_param_data->cpf_filename), 1);

    /* Get the name of the data file containing just the corrected
       attitude/ephemeris data generated from the ADP application.  It is
       expected that (1) ADP will be run prior to calling this
       characterization, with the name of the adjusted data file written to
       the OMF as the parameter "ANCILLARY_FILENAME"; and (2) the adjusted
       data file name differs in some way from the L0R file name (if they're
       not different, the current ADP implementation will overwrite the L0R
       ancillary data file).  The L0R ancillary data file contains the OLI
       image/frame header information needed for validation.  */
    IAS_PARM_ADD_STRING(omf_param_table, ANCILLARY_FILENAME,
            "Name of ADP-corrected ancillary data file", IAS_PARM_REQUIRED,
            0, NULL, 0, NULL, omf_param_data->adp_ancillary_filename,
            sizeof(omf_param_data->adp_ancillary_filename), 1);

    /* Get the total number of items to read from the OMF parameter
       table*/
    read_count = IAS_PARM_GET_TABLE_SIZE(omf_param_table);

    /* Provide parameter help if requested.  Check the return status
       to distinguish whether help was requested and successfully provided
       (1), no help was requested (0), or there was an error providing help
       (ERROR) */
    status = ias_parm_provide_help(omf_params_filename, omf_param_table,
            read_count, IAS_OMF_PARAMETERS);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("OMF parameter help failed");
        return ERROR;
    }
    else if (status == 1)
    {
        exit(EXIT_SUCCESS);
    }

    /* Read the OMF file */
    status = ias_parm_read(omf_params_filename, omf_param_table, read_count);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading OMF parameter file");
        return ERROR;
    }

    /* Initialize the satellite attributes library */
    satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(
                omf_param_data->satellite_number);
    if (satellite_id == ERROR)
    {
        IAS_LOG_ERROR("Unable to determine satellite ID from satellite "
                      "number %d", omf_param_data->satellite_number);
        return ERROR;
    }
    if (ias_sat_attr_initialize(satellite_id) != SUCCESS)
    {
        IAS_LOG_ERROR("Error initializing the satellite attributes library");
        return ERROR;
    }

    /* OLI has 9 bands.  Use the band number for the cirrus band
       (band 9) as this maximum */
    status = ias_sat_attr_get_sensor_band_numbers(IAS_OLI,
        IAS_NORMAL_BAND, 0, dummy_band_list,  IAS_MAX_NBANDS,
        &max_oli_bands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving maximum number of 'normal' OLI"
                      "imaging bands");
        return ERROR;
    }

    if (wo_param_data->num_bands_to_process > max_oli_bands)
    {
        IAS_LOG_WARNING("Too many bands (%d) to process for OLI solar "
            "calibration, should only be %d",
                wo_param_data->num_bands_to_process,
                max_oli_bands);
        wo_param_data->num_bands_to_process = max_oli_bands;
    }

    /* Done */
    return SUCCESS;
}
