/**************************************************************************
 NAME:              solarcal_db 

 PURPOSE:  Module implementing routines to provide IAS characterization
           database access for the solarcal_oli characterization algorithm.

****************************************************************************/
#include "ias_db.h"
#include "ias_db_get_connect_info.h"
#include "ias_db_insert.h"
#include "ias_db_query.h"
#include "ias_radiometry.h"

#include "solarcal_oli.h"

#define SQL_QUERYLEN 512

/*--------------------------------------------------------------------
 NAME:                  connect_to_db

 PURPOSE:    Retrieves IAS database connection information and attempts
             the connection

 RETURNS:    - Pointer to valid ias_db_struct, indicating connection has
               been established
             - NULL pointer if connection cannot be made
-----------------------------------------------------------------------*/
struct ias_db_connection *connect_to_db()
{
    char *db_name = NULL;            /* Required connection information */
    char *user_id = NULL;
    char *user_password = NULL;
    char *db_host = NULL;

    struct ias_db_connection *db = NULL;

    int status;


    IAS_LOG_DEBUG("Initializing database library...");
    status = ias_db_initialize_database_lib();
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing database library");
        return NULL;
    }

    /* Get the connection information */
    IAS_LOG_DEBUG("Obtaining database connection information...");
    status = ias_db_get_connect_info(&db_name, &user_id, &user_password,
                                  &db_host);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Obtaining database connection information");
        return NULL;
    }

    IAS_LOG_DEBUG("DB Name:    %s\nUser ID:  %s\nUser Pwd:    %s",
                  db_name, user_id, user_password);

    /* Try to connect */
    IAS_LOG_DEBUG("Connecting to database...");
    db = ias_db_connect_to_database(db_name, user_id,
                                    user_password, db_host);

    /* Free the allocated strings with the connection information */
    free(db_name);
    free(user_id);
    free(user_password);
    free(db_host);

    /* Did we connect? */
    if (db == NULL)
        IAS_LOG_ERROR("Connecting to database");
    else
        IAS_LOG_DEBUG("Database connection established");

    /* Return whatever we've got for *db (either NULL or a valid pointer) */
    return db;
}




/*------------------------------------------------------------------------
 NAME:                     clear_solarcal_table

 PURPOSE:       Clears the specified SOLAR_RESPONSE_XXXX table for the
                specified radiometric work order ID key

 RETURNS:       Integer status code of SUCCESS (tables successfully
                cleared) or ERROR (table(s) couldn't be cleared)
--------------------------------------------------------------------------*/
int clear_solarcal_table
(
    struct ias_db_connection *db, /* I: Open database connection */
    char *rad_wo_common_id        /* I: Radiometric work order common ID */,
    const  char *table_name       /* I: Name of solarcal table to clear */
)
{
    int status;                   /* Function return status code */

    /* 1.  Clear out the specified table */
    IAS_LOG_DEBUG("Clearing %s table...", table_name);
    status = ias_db_delete_records(db, table_name,
                             "radiometric_wo_common_id",
                             rad_wo_common_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Clearing %s table", table_name);
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}




/*-------------------------------------------------------------------------
 NAME:        get_histogram_statistics

 PURPOSE:     Queries the HISTOGRAM_STATISTICS table for the
              mean DN values extracted from the diffuser data product
              for the desired work order (indirectly
              represented by the specified RADIOMETRIC_WORK_ORDER_COMMON
              primary key value)

 RETURNS:     Integer status code of SUCCESS or ERROR

 NOTES:       It's the caller's responsibility to ensure the data buffer
              parameters are valid prior to calling this routine.

              This routine populates data buffers sized to all detectors
              in the current band, even though it directly retrieves
              statistics just for the current band/SCA.  It determines
              the index into the band data buffers corresponding to
              detector 1 for the current SCA and populates them
              accordingly.

              As just inoperable detectors are currently considered in this
              characterization, that information comes directly
              from the CPF DETECTOR_STATUS group.  If other classes of
              detectors 'to be defined' are required to be excluded
              from the characterization, this routine should be updated
              to process with a pixel mask.
---------------------------------------------------------------------------*/
int get_histogram_statistics
(
    struct ias_db_connection  *db,     /* I: Open database connection */
    char *rad_wo_common_id,            /* I: RADIOMETRIC_WORK_ORDER_COMMON
                                          primary key value */
    IAS_SPECTRAL_TYPE spectral_type,   /* I: Spectral type for current
                                          band */
    int band_number,                   /* I: Current 1-based band number */
    int sca_number,                    /* I: Current 1-based SCA number */
    int num_detectors_sca,             /* I: Number of detectors for the
                                          current band/SCA */
    const char *rps_proc_level,        /* I: Required RPS processing level
                                          for the histogram statistics input
                                          data */
    double *stats_mean_buffer,         /* O: Histogram statistics means
                                          for all detectors in the current
                                          band */
    double *stats_stdev_buffer,        /* O: Histogram statistics std
                                          deviations for all detectors in
                                          the current band */
    double *stats_num_frames_buffer    /* O: Histogram statistics number
                                          of frames for all detectors in
                                          the current band */
)
{
    char query_string[SQL_QUERYLEN];           /* First SQL query string */
    char db_error_message[IAS_SQL_RTN_MSGLEN]; /* Query error message */

    struct ias_db_query *query = NULL;   /* Query dataset */

    int detector_index;                  /* Detector loop counter */
    int status;
    int start_detector_index;            /* 0-based index into band array
                                            representing detector 1 in the
                                            current SCA */


    /* For the PAN band, write the query so that it returns the average
       scene_pixel_mean and scene_pixel_stdev, and the sums of nbr_of_frames
       for the lineset 0 AND lineset 1 data.  For the MS bands,
       set lineset to 2.  In both cases, the processing level should contain
       bias-correct/linearized statistics */
    if (spectral_type == IAS_SPECTRAL_PAN)
    {
        status = snprintf(query_string, sizeof(query_string),
            "SELECT detector_id, "
            "       AVG(scene_pixel_mean), "
            "       AVG(scene_pixel_stdev), "
            "       SUM(nbr_of_frames) "
            "  FROM histogram_statistics_cal "
            " WHERE band_id = %d "
            "   AND sensor_chip_assembly_id = %d "
            "   AND radiometric_wo_common_id = %s "
            "   AND processing_level = '%s' "
            "   AND lineset IN (0, 1) "
            " GROUP BY detector_id ORDER BY detector_id;",
            band_number, sca_number, rad_wo_common_id, rps_proc_level);
    }
    else
    {
        status = snprintf(query_string, sizeof(query_string),
            "SELECT detector_id, "
            "       scene_pixel_mean, "
            "       scene_pixel_stdev, "
            "       nbr_of_frames "
            "  FROM histogram_statistics_cal "
            " WHERE band_id = %d "
            "   AND sensor_chip_assembly_id = %d "
            "   AND radiometric_wo_common_id = %s "
            "   AND processing_level='%s' "
            "   AND lineset = 2 "
            " ORDER BY detector_id;",
            band_number, sca_number, rad_wo_common_id, rps_proc_level);
    }
    if ((status < 0) || (status >= sizeof(query_string)))
    {
        IAS_LOG_ERROR("Preparing query string");
        return ERROR;
    }

    /* Execute the query and verify it worked */
    query = ias_db_perform_query(db, query_string);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
                                        sizeof(db_error_message));
        IAS_LOG_ERROR("Query of HISTOGRAM STATISTICS failed:  %s",
                      db_error_message);
        return ERROR;
    }

    /* Loop through the detectors, flagging the inoperable (and perhaps, but
       not initially) out-of-spec ones, and copying the returned data for
       the "good" ones */
    start_detector_index = (sca_number - 1) * num_detectors_sca;
    for (detector_index = 0;
         detector_index < num_detectors_sca;
         detector_index++)
    {
        double stats_mean_value;   /* Histogram statistics mean */
        double stats_stdev_value;  /* Histogram statistics std deviation */
        double frames_value;       /* Histogram statistics nbr of frames */
        int dummy_detector;        /* 1-based detector ID value */
        int index;                 /* 0-based index of current detector */
 
        /* Advance to the next record */
        if (!ias_db_query_next(query))
        {
            IAS_LOG_ERROR("Accessing HISTOGRAM_STATISTICS data");
            ias_db_query_close(query);
            return ERROR;
        }

        /* Get the detector ID.  Could need it to properly access the
           data buffers if the query from the database mucked up the
           sorting order.  If things are working as expected, the value
           read into dummy_detector should be identical to
           (detector_index + 1) */
        status = ias_db_query_int_value(query, &dummy_detector, 0);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading histogram statistics "
                          "detector_id value for detector index %d "
                          "in band %d SCA %d\n", detector_index,
                          band_number, sca_number);
            ias_db_query_close(query);
            return ERROR;
        }

        /* Get the mean value */
        status = ias_db_query_double_value(query, &stats_mean_value, 1);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining histogram statistics mean for "
                          "%d in band %d SCA %d", dummy_detector,
                          band_number, sca_number);
            ias_db_query_close(query);
            return ERROR;
        }

        /* Now get the std dev */
        status = ias_db_query_double_value(query, &stats_stdev_value, 2);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining histogram statistics "
                          "stdev for detector %d in band %d SCA %d",
                          dummy_detector, band_number, sca_number);
            ias_db_query_close(query);
            return ERROR;
        }

        /* Finally, get the number of frames */
        status = ias_db_query_double_value(query, &frames_value, 3);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining histogram statistics "
                          "number of frames for detector %d in band %d "
                          "SCA %d", dummy_detector, band_number,
                          sca_number);
            ias_db_query_close(query);
            return ERROR;
        }

         /* Copy the statistics values into the buffers */
        index = dummy_detector - 1;

        stats_mean_buffer[index + start_detector_index] = stats_mean_value;
        stats_stdev_buffer[index + start_detector_index] = stats_stdev_value;
        stats_num_frames_buffer[index + start_detector_index] = frames_value;
    }

    /* Finished with the first query.  Close out the resources */
    ias_db_query_close(query);

    return SUCCESS;
}




/*-----------------------------------------------------------------------
 NAME:            populate_db_sca_data_buffers

 PURPOSE:  Populates SCA-level metrics detector ID and data buffers with
           NULL database flag values for detectors that aren't
           characterized

 RETURNS:  Integer status code of SUCCESS or ERROR

 NOTES:    This function is only called when writing characterization
           results to the database
-------------------------------------------------------------------------*/
void populate_db_sca_data_buffers
(
    int start_detector_index,      /* I: Index corresponding to detector 1
                                      in the current SCA */
    int num_detectors_sca,         /* I: # of detectors in each SCA */
    double *band_mean_data,        /* I: Full data vector (means) for the
                                      current band */
    double *band_stdev_data,       /* I: Full data vector (stdevs) for the
                                      current band */
    const int *statistics_invalid_flags,
                                   /* I: Data invalid flags for all
                                      detectors in in the current band */
    IAS_DB_NULL_TYPE *flagged_detector_ids, /* O: SCA detector IDs with NULLs */
    double *sca_mean_data,         /* O: Data vector (means) for the
                                      current band/SCA */
    double *sca_stdev_data         /* O: Data vector (stdevs) for the
                                      current band/SCA */
)
{

    int detector_index;

    /* Get the NULL data value from the database */
    IAS_DB_NULL_TYPE null_data_flag = ias_db_get_null_data_flag();


    /* Copy data from the band mean and stdev data arrays to the SCA level
       data arrays.  For missing data, populate the corresponding members of
       the detector ID array with NULLs.  In the detector ID
       array, leave the other values at 0, which is what they were
       initialized to  */
    for (detector_index = 0;
         detector_index < num_detectors_sca;
         detector_index++)
    {
        if (statistics_invalid_flags[start_detector_index + detector_index])
        {
            flagged_detector_ids[detector_index] = null_data_flag;
        }

        sca_mean_data[detector_index]
            = band_mean_data[start_detector_index + detector_index];
        sca_stdev_data[detector_index]
            = band_stdev_data[start_detector_index + detector_index];
    }
}




/*-------------------------------------------------------------------------
 NAME:           write_to_detector_table

 PURPOSE:     Inserts one set of relative gains into the
              SOLAR_RESPONSE_DETECTOR table for the current band and SCA,
              using table-based insert functionality 

 RETURNS:     Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int write_to_detector_table
(
    struct ias_db_connection  *db,       /* I: Open database connection */
    char *rad_wo_common_id,              /* I: Radiometric WO ID key to
                                            SOLAR_RESPONSE table */
    int band_number,                     /* I: Current (1-based) band # */
    int sca_number,                      /* I: Current (1-based) SCA # */
    int num_sca_detectors,               /* I: # of detectors for each
                                            SCA in the current band */
    int *detector_ids,                   /* I: Detector IDs for the current
                                            band/SCA (1-based) */
    IAS_DB_NULL_TYPE *nulled_detector_flags, /* I: SCA detector indices
                                                flagged with NULLs */
    double *sca_relgain_data,            /* I: Derived relative gains array
                                            for the current band/SCA */
    double *sca_relgain_stdev_data       /* I: Derived relative gain
                                            stdevs array for the current
                                            band/SCA */
)
{
    int status;                          /* Function return status */
    int table_size;                      /* # of entries to table */

    /* Define the insert table */
    IAS_DB_INSERT_TABLE table[] =
    {
        IAS_DB_TABLE_STRING("radiometric_wo_common_id", rad_wo_common_id),
        IAS_DB_TABLE_INT("band_id", band_number),
        IAS_DB_TABLE_INT("sensor_chip_assembly_id", sca_number),
        IAS_DB_TABLE_INT_ARRAY("detector_id", detector_ids),
        IAS_DB_TABLE_DOUBLE_ARRAY_WITH_NULLS("relative_gain_mean",
                                             sca_relgain_data,
                                             nulled_detector_flags),
        IAS_DB_TABLE_DOUBLE_ARRAY_WITH_NULLS("relative_gain_stdev",
                                             sca_relgain_stdev_data,
                                             nulled_detector_flags)
    };
    table_size = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);
    status = ias_db_insert_using_table(db, "solar_response_detector",
                                       table, table_size, num_sca_detectors);

    if (status != SUCCESS)
        IAS_LOG_ERROR("Writing relative gains and/or std deviations "
                      "to database in band %d SCA %d", band_number,
                      sca_number);
    else
        IAS_LOG_DEBUG("Database write successful for band %d, SCA %d",
                     band_number, sca_number);

    return status;
}




/*-------------------------------------------------------------------------
 NAME:           write_to_band_table

 PURPOSE:     Inserts each band's band average gains (both radiance-based
              and reflectance-based) into the SOLAR_RESPONSE_BAND
              table, using table-based insert functionality 

 RETURNS:     Integer status code of SUCCESS or ERROR 
---------------------------------------------------------------------------*/
int write_to_band_table
(
    struct ias_db_connection *db,        /* I: Open DB connection */
    char *rad_wo_common_id,              /* I: Radiometric WO ID key to
                                            SOLAR_RESPONSE table */
    RPS_SOLARCAL_BAND_GAINS band_gains   /* I: Band-level summary
                                            histogram statistics and
                                            diffuser radiances/
                                            reflectances */
)
{
    int status = SUCCESS;          /* Status return value */
    int table_size = 0;
    int num_records = 1;

    /* Define the insert table */
    IAS_DB_INSERT_TABLE table[] =
    {
        IAS_DB_TABLE_STRING("radiometric_wo_common_id", rad_wo_common_id),
        IAS_DB_TABLE_INT("band_id", band_gains.band_number),
        IAS_DB_TABLE_DOUBLE("histogram_stats_mean",
            band_gains.hist_stats_mean),
        IAS_DB_TABLE_DOUBLE("histogram_stats_stdev",
            band_gains.hist_stats_stdev),
        IAS_DB_TABLE_DOUBLE("diffuser_radiance_mean",
            band_gains.diffuser_radiance_mean),
        IAS_DB_TABLE_DOUBLE("diffuser_radiance_stdev",
            band_gains.diffuser_radiance_stdev),
        IAS_DB_TABLE_DOUBLE("radiance_gain_mean", band_gains.radiance_gain),
        IAS_DB_TABLE_DOUBLE("radiance_gain_stdev",
            band_gains.radiance_gain_stdev),
        IAS_DB_TABLE_DOUBLE("diffuser_reflectance_mean",
            band_gains.diffuser_reflectance_mean),
        IAS_DB_TABLE_DOUBLE("diffuser_reflectance_stdev",
            band_gains.diffuser_reflectance_stdev),
        IAS_DB_TABLE_DOUBLE("reflectance_gain_mean",
            band_gains.reflectance_gain),
        IAS_DB_TABLE_DOUBLE("reflectance_gain_stdev",
            band_gains.reflectance_gain_stdev)
    };

    table_size = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);

    /* Insert the metric data into the SOLAR_RESPONSE_BAND table */
    status = ias_db_insert_using_table(db, "solar_response_band",
                                       table, table_size, num_records);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing band average gains to database in band %d",
            band_gains.band_number);
    }
    else
    {
        IAS_LOG_DEBUG("Band average gains written to database table in band "
            "%d", band_gains.band_number);
    }

    /* Done */
    return status;
}




/*-----------------------------------------------------------------------
 NAME:              write_to_main_table

 PURPOSE:    Writes validation and summary angle information to the
             SOLAR_RESPONSE table, using table-based insert
             functionality

 RETURNS:    Integer status code of SUCCESS or ERROR
-------------------------------------------------------------------------*/
int write_to_main_table
(
    struct ias_db_connection *db,          /* I: Open DB connection */
    char *rad_wo_common_id,                /* I: Radiometric WO ID key to
                                              SOLAR_RESPONSE table */
    RPS_SOLARCAL_REPORT_HEADER header,     /* I: Validation information */
    RPS_SOLARCAL_SOLAR_ANGLES angles       /* I: Solar angle information */
)
{
    char panel_type[IAS_DIFFUSER_TYPE_SIZE];
    double deployment_angle;               /* Deployment angle of
                                              diffuser panel */
    int status;                            /* Status return value */
    int table_size;
    int num_records = 1;

    /* Define the insert table.  Make sure the header information for
       the number of histogram statistics frames and the diffuser resolver
       position do not exceed the allowed maximum values.  They SHOULDN'T
       ever get this large, but in case something happens and they do...*/
    if ((header.num_frames_averaged > INT_MAX)
        || (header.diffuser_resolver_position
            < IAS_RPS_OLI_DIFFUSER_POSITION_PRISTINE_LOW)
        || (header.diffuser_resolver_position
            > IAS_RPS_OLI_DIFFUSER_POSITION_WORKING_HIGH))
    {
        IAS_LOG_ERROR("Number of averaged frames for histogram statistics "
            "and/or diffuser resolver position are beyond their maximum "
            "datatype limits");
        return ERROR;
    }

    /* Write the appropriate string for the panel identification based on
       the ID value given in the report header information */
    if (header.panel_id == IAS_RPS_OLI_DIFFUSER_TYPE_PRISTINE)
    {
        strcpy(panel_type, IAS_RPS_OLI_DIFFUSER_PRISTINE);
    }
    else
    {
        strcpy(panel_type, IAS_RPS_OLI_DIFFUSER_WORKING);
    }

    /* Convert the diffuser resolver position telemetry value to the
       corresponding deployment angle */
    convert_diffuser_resolver_position_to_angle(
        header.diffuser_resolver_position, &deployment_angle);

    IAS_DB_INSERT_TABLE table[] =
    {
        IAS_DB_TABLE_STRING("radiometric_wo_common_id", rad_wo_common_id),
        IAS_DB_TABLE_TIMESTAMP("interval_start_time",
                               header.rps_header.acq_datetime.start_time_oli),
        IAS_DB_TABLE_TIMESTAMP("interval_stop_time",
                               header.rps_header.acq_datetime.stop_time_oli),
        IAS_DB_TABLE_INT("lines_averaged", header.num_frames_averaged),
        IAS_DB_TABLE_STRING("diffuser_collect_type", panel_type),
        IAS_DB_TABLE_DOUBLE("deployment_angle", deployment_angle),

        /* If/when MOC reports are available, put in a field to indicate
           whether angles were taken from the MOC data or were calculated
           internally by the IAS.  To start with, angles will be calculated
           internally by the IAS; this will also be the default
           processing scenario through the mission */
        IAS_DB_TABLE_DOUBLE("incident_solar_zenith_mean",
            angles.incident_solar_zenith),
        IAS_DB_TABLE_DOUBLE("incident_solar_zenith_stdev",
            angles.incident_solar_zenith_stdev),
        IAS_DB_TABLE_DOUBLE("view_solar_zenith_mean",
            angles.view_solar_zenith),
        IAS_DB_TABLE_DOUBLE("view_solar_zenith_stdev",
            angles.view_solar_zenith_stdev),
        IAS_DB_TABLE_DOUBLE("view_solar_azimuth_mean",
            angles.view_solar_azimuth),
        IAS_DB_TABLE_DOUBLE("view_solar_azimuth_stdev",
            angles.view_solar_azimuth_stdev)
    };

    table_size = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);

    /* Insert the metric data into the SOLAR_RESPONSE table */
    status = ias_db_insert_using_table(db, "solar_response", table,
                                       table_size, num_records);
    if (status != SUCCESS)
        IAS_LOG_ERROR("Writing collect summary information to database");
    else
        IAS_LOG_DEBUG("Collect summary information written to database");

    /* Finished */
    return status;
}
