/**************************************************************************
 NAME:          solarcal_oli

 PURPOSE:       Derives band absolute gains and detector relative gains
                based on bias-corrected and linearized histogram statistics
                obtained from the OLI solar diffuser

 RETURNS:       System status code of EXIT_SUCCESS or EXIT_FAILURE

 NOTES:         This application can run standalone outside of the RPS
                Driver once processing of an OLI solar interval has
                completed and the histogram statistics from all bands/SCAs
                have been written to the database.  "Automated"
                characterization can be set up by creating a procedure
                run from the IAS OUI.

                Whether run in standalone mode or as part of an automated
                OUI procedure, the ADP application must be run prior to
                performing this characterization.  ADP reads the L0R
                ancillary data file and creates a reduced ancillary data
                file containing just the "corrected" attitude and ephemeris
                telemetry data.  'solarcal_oli' determines the solar
                geometry applicable during the collect from the
                corrected ACS quaternion data, as part of an effort
                to ensure the collect is a 'valid' solar collect.
                When running ADP, DO NOT give the output ancillary
                data file the same name as the L0R data file--IT WILL
                BE OVERWRITTEN!
***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "ias_cpf.h"
#include "ias_db.h"
#include "ias_l0r.h"
#include "ias_math.h"
#include "ias_radiometry.h"
#include "ias_structures.h"

/* Main application header */
#include "solarcal_oli.h"



/*----------------------------------------------------------------------
              *** MAIN APPLICATION solarcal_oli ***
------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{

    /* Solar angle data */
    RPS_SOLARCAL_SOLAR_ANGLES  angles;

    /* External data/files used to validate current diffuser acquisition */
    IAS_VECTOR solar_position;
    IAS_DATETIME start_datetime;
    IAS_DATETIME stop_datetime;

    /* Gains for a given band */
    RPS_SOLARCAL_BAND_GAINS band_gains;      /* Absolute radiance-based
                                                and reflectance-based
                                                gains for an individual
                                                band */

    /* Working processing buffers/variables */
    char *date_acquired = NULL;             /* Acquisition date from IAS
                                               common header */
    double start_year_doy_sod[3];           /* Image start time expressed
                                               in year/day-of-year/second-
                                               of-day format */
    double stop_year_doy_sod[3];            /* Image stop time expressed
                                               in year/day-of-year/second-
                                               of-day format */
    double cos_theta;                       /* Cosine of mean solar zenith
                                               angle observed during the
                                               collect */
    double earth_sun_dist_sqr;              /* Square of the earth-sun
                                               distance factor */
    double jpl_ephem_date;                  /* Julian date to use when
                                               determining information from
                                               JPL ephemeris */
    double num_frames_averaged = 0;         /* Average number of frames used
                                               in histogram statistics
                                               calculations across all bands */
    double num_frames_averaged_band;        /* Average # of frames used in
                                               histogram statistics
                                               calculations for the current
                                               band */
    double incident_solar_zenith;           /* Solar zenith angle converted
                                               from degrees to radians */

    /* Database-related stuff */
    struct ias_db_connection *open_db = NULL;
                                            /* Open IAS database connection */

    int *dummy_band_list = NULL;            /* "Dummy" band list used in
                                               ias_sat_attr_get_sensor_band
                                               _numbers()--not needed
                                               beyond.  Declaring a pointer
                                               rather than a stack array */
    char rad_wo_common_id[IAS_CHAR_ID_SIZE];  
                                            /* Primary key value in
                                               RADIOMETRIC_WORK_ORDER_COMMON
                                               table (needed to associate
                                               processing with the correct
                                               work order)*/

    /* Optional summary report/dump file stuff */
    RPS_SOLARCAL_REPORT_SUMMARY rpt_summary;

    /* Various processing-related stuff */
    char panel_type_string[IAS_DIFFUSER_TYPE_SIZE];
                                            /* Panel identification
                                               string */
    int num_oli_bands;                      /* Total number of OLI bands */
    int index;                              /* Index into band list */
    int status;                             /* Function return status */
    int final_status = EXIT_SUCCESS;        /* Final application status */

    IAS_SATELLITE_ID satellite_id;          /* Satellite ID number */

    /* CPF stuff */
    IAS_CPF *cpf = NULL;                          /* CPF data structure */
    const struct IAS_CPF_OLI_PARAMETERS *oli_params = NULL;
                                                  /* CPF OLI parameters */
    const struct IAS_CPF_EARTH_CONSTANTS *earth_constants = NULL;
                                                  /* CPF Earth constant
                                                     parameters */
    const struct IAS_CPF_DIFFUSER_RADIANCE *diff_rad = NULL; 
                                                  /* CPF diffuser radiance
                                                     and reflectance data */
    const struct IAS_CPF_DETECTOR_STATUS *det_status = NULL; 
                                                 /* CPF detector status
                                                        information */

    /* Processing parameters information */
    RPS_SOLARCAL_WO_PARAMS wo_params;       /* Work order parameters */
    RPS_SOLARCAL_OMF_PARAMS omf_params;     /* OMF parameters */


    /* Initialize logging for this application */
    ias_log_initialize("solarcal_oli");

    /* Only one command line argument:  the WO parameter file */
    if (argc != 2)
    {
        IAS_LOG_ERROR("USAGE:  %s <parameter filename>\n", argv[0]);
        exit(EXIT_FAILURE);
    }

    IAS_LOG_INFO("Solar calibration processing initiated...");

    /* Initialize parameter buffers */
    memset(&wo_params, 0, sizeof(RPS_SOLARCAL_WO_PARAMS));
    memset(&omf_params, 0, sizeof(RPS_SOLARCAL_OMF_PARAMS));

    /* Read work order parameters */
    status = get_parameters(argv[1], &wo_params, &omf_params);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Parameter processing failed");
        exit(EXIT_FAILURE);
    }

/*--------------------------------------------------------------------------
                         COLLECT VALIDATION
----------------------------------------------------------------------------*/
    /* Initialize the common and RPS-specific sections of the report
       header */
    IAS_LOG_DEBUG("Validating collect information...");
    memset(&rpt_summary, 0, sizeof(RPS_SOLARCAL_REPORT_SUMMARY));

    rpt_summary.header.common_header = ias_misc_initialize_report_header(
        omf_params.l0r_filename, wo_params.work_order_id,
        "OLI Solar Calibration");
    if (rpt_summary.header.common_header == NULL)
    {
        IAS_LOG_ERROR("Initializing common report header information");
        exit(EXIT_FAILURE);
    }

    status = ias_misc_initialize_rps_report_header(
        omf_params.l0r_filename, &(rpt_summary.header.rps_header));
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing RPS-specific report header "
            "information");
        ias_misc_free_report_header(rpt_summary.header.common_header);
        exit(EXIT_FAILURE);
    }

    /* Update the CPF in the header to reflect what is in the OMF file. */
    ias_misc_set_header_cpf_name(rpt_summary.header.common_header,
        omf_params.cpf_filename);

    /* Read the CPF, then get the OLI parameter group data */
    IAS_LOG_DEBUG("Reading CPF parameters...");
    cpf = ias_cpf_read(omf_params.cpf_filename);
    if (!cpf)
    {
        IAS_LOG_ERROR("Reading CPF file %s", omf_params.cpf_filename);
        ias_misc_free_report_header(rpt_summary.header.common_header);
        exit(EXIT_FAILURE);
    }
    oli_params = ias_cpf_get_oli_parameters(cpf);
    if (oli_params == NULL)
    {
        IAS_LOG_ERROR("Obtaining OLI parameter information");
        ias_cpf_free(cpf);
        ias_misc_free_report_header(rpt_summary.header.common_header);
        exit(EXIT_FAILURE);
    }

    /* Get the Earth constants group data--this will be needed to identify
       the diffuser panel deployed during the collect */
    earth_constants = ias_cpf_get_earth_const(cpf);
    if (earth_constants == NULL)
    {
        IAS_LOG_ERROR("Obtaining Earth constants information");
        ias_cpf_free(cpf);
        ias_misc_free_report_header(rpt_summary.header.common_header);
        exit(EXIT_FAILURE);
    }

    /* Identify the diffuser panel used in this solar collect.  We'll need
       the image start time for the collect to do this */
    memset(&start_datetime, 0, sizeof(IAS_DATETIME));
    status = ias_misc_parse_datetime_string(
        rpt_summary.header.rps_header.acq_datetime.start_time_oli,
        IAS_DATETIME_L0R_FORMAT, &start_datetime);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Returning from ias_misc_parse_datetime_string()");
        ias_cpf_free(cpf);
        ias_misc_free_report_header(rpt_summary.header.common_header);
        exit(EXIT_FAILURE);
    }
    status = ias_rps_identify_diffuser_panel(omf_params.l0r_filename,
        earth_constants, start_datetime,
        &rpt_summary.header.diffuser_resolver_position,
        &rpt_summary.header.panel_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Returning from ias_misc_identify_diffuser_panel()");
        ias_cpf_free(cpf);
        ias_misc_free_report_header(rpt_summary.header.common_header);
        exit(EXIT_FAILURE);
    }

    /* Use the report header information to "validate" the collect.  As a
       bonus, if we generate report/dump files, the header information's
       already available  */
    status = validate_collect_data(rpt_summary.header, oli_params);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Cannot validate this is a solar collect, "
            "program terminating");
        ias_misc_free_report_header(rpt_summary.header.common_header);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }

    /* Create the corresponding panel identification string */
    if (rpt_summary.header.panel_id == IAS_RPS_OLI_DIFFUSER_TYPE_PRISTINE)
    {
        strcpy(panel_type_string, IAS_RPS_OLI_DIFFUSER_PRISTINE);
    }
    else
    {
        strcpy(panel_type_string, IAS_RPS_OLI_DIFFUSER_WORKING);
    }

/*--------------------------------------------------------------------------
        CALCULATE SOLAR GEOMETRY FROM ADP-CORRECTED ACS QUATERNION DATA
---------------------------------------------------------------------------*/
    memset(&angles, 0, sizeof(RPS_SOLARCAL_SOLAR_ANGLES));
 
    /* Calculate the Julian date corresponding to the current
       acquisition date */
    date_acquired = ias_misc_get_header_date_acquired(
        rpt_summary.header.common_header);
    if (date_acquired == NULL)
    {
        IAS_LOG_ERROR("No acquisition date information available");
        ias_misc_free_report_header(rpt_summary.header.common_header);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }
    jpl_ephem_date = get_julian_date_for_jpl_ephem(date_acquired,
            IAS_DATETIME_L0R_FORMAT);
    if (jpl_ephem_date == (double)ERROR)
    {
        IAS_LOG_ERROR("Obtaining Julian date corresponding to the " 
            "acquisition date");
        ias_misc_free_report_header(rpt_summary.header.common_header);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }

    /* Get the earth-sun distance factor and the instantaneous
       solar position vector from the JPL ephemeris data */
    status = process_jpl_ephemeris(jpl_ephem_date,
        &angles.earth_sun_distance_factor, &solar_position);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Cannot process JPL ephemeris data, program "
                "terminating");
        ias_misc_free_report_header(rpt_summary.header.common_header);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }
    earth_sun_dist_sqr =
        angles.earth_sun_distance_factor * angles.earth_sun_distance_factor;

    /* Parse the image stop time */
    memset(&stop_datetime, 0, sizeof(IAS_DATETIME));
    status = ias_misc_parse_datetime_string(
        rpt_summary.header.rps_header.acq_datetime.stop_time_oli,
        IAS_DATETIME_L0R_FORMAT, &stop_datetime);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Returning from ias_misc_parse_datetime_string()");
        ias_cpf_free(cpf);
        ias_misc_free_report_header(rpt_summary.header.common_header);
        exit(EXIT_FAILURE);
    }

    /* Convert the image start and stop time information to the
       corresponding year/day-of-year/second-of-day format */
    status = convert_datetime_to_year_doy_sod(start_datetime,
        &start_year_doy_sod[0], &start_year_doy_sod[1],
        &start_year_doy_sod[2]);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Converting image start time to year/day-of-year/"
            "second-of-day format");
        ias_cpf_free(cpf);
        ias_misc_free_report_header(rpt_summary.header.common_header);
        exit(EXIT_FAILURE);
    }
    status = convert_datetime_to_year_doy_sod(stop_datetime,
        &stop_year_doy_sod[0], &stop_year_doy_sod[1], &stop_year_doy_sod[2]);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Converting image stop time to year/day-of-year/"
            "second-of-day format");
        ias_cpf_free(cpf);
        ias_misc_free_report_header(rpt_summary.header.common_header);
        exit(EXIT_FAILURE);
    }

    /* Calculate the solar angle information from the ancillary data */
    IAS_LOG_DEBUG("Obtaining solar angle information...");
    status = calculate_solar_angles(omf_params.adp_ancillary_filename,
        start_year_doy_sod, stop_year_doy_sod, &solar_position, &angles);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Deriving solar geometry from recorded ancillary "
                "data");
        ias_misc_free_report_header(rpt_summary.header.common_header);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }

    /* Determine the cosine of the incident solar zenith angle.  This will be
       needed when reflectance-based gains are calculated */
    incident_solar_zenith =
        (ias_math_get_radians_per_degree() * angles.incident_solar_zenith);
    cos_theta = cos(incident_solar_zenith);

    /****************  COLLECT VALIDATION COMPLETED   *******************/

    /* Initialize the satellite attributes library so that we're processing
       OLI data */
    IAS_LOG_DEBUG("Initializing satellite/sensor attributes...");

    satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(
            omf_params.satellite_number);
    if (satellite_id == ERROR)
    {
        IAS_LOG_ERROR("Invalid local satellite number in OMF parameter file");
        ias_misc_free_report_header(rpt_summary.header.common_header);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }

    status = ias_sat_attr_initialize(satellite_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite/sensor attributes");
        ias_misc_free_report_header(rpt_summary.header.common_header);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }

    /* Get the total number of OLI bands here.  Use heap memory that
       can be freed immediately, rather than a stack-based array */
    dummy_band_list = calloc(IAS_MAX_NBANDS, sizeof(*dummy_band_list));
    if (dummy_band_list == NULL)
    {
        IAS_LOG_ERROR("Allocating 'dummy' band list array");
        ias_misc_free_report_header(rpt_summary.header.common_header);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }
    status = ias_sat_attr_get_sensor_band_numbers(IAS_OLI, IAS_NORMAL_BAND,
        0, dummy_band_list, IAS_MAX_NBANDS, &num_oli_bands);
    free(dummy_band_list);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Obtaining total number of OLI imaging bands");
        ias_misc_free_report_header(rpt_summary.header.common_header);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }

    /* Get the detector status information and diffuser radiance/reflectance
       information from the CPF */
    det_status = ias_cpf_get_det_status(cpf);
    diff_rad   = ias_cpf_get_diffuser_rad(cpf);
    if ((det_status == NULL) || (diff_rad == NULL))
    {
        IAS_LOG_ERROR("Getting required detector status and/or diffuser "
                "radiance data from CPF");
        ias_cpf_free(cpf);
        ias_misc_free_report_header(rpt_summary.header.common_header);
        exit(EXIT_FAILURE);
    }

    /* Further initialize the application based on the selected output
       option(s).

       There are 7 possible "scenarios" where characterization results 
       can be written to.  The settings of the relevant WO parameters
       for each scenario are given below.

       1.  Dump files only (store_char_in_db=0, REPORT_FLAG=0,
       DUMP_FLAG=1)
       2.  Summary report only (store_char_in_db=0, REPORT_FLAG=1,
       DUMP_FLAG=0)
       3.  Dump files + summary report (store_char_in_db=0, REPORT_FLAG=1,
       DUMP_FLAG=1)
       4.  DB only (store_char_in_db=1, REPORT_FLAG=0, DUMP_FLAG=0)
       5.  DB + dump files (store_char_in_db=1, REPORT_FLAG=0,
       DUMP_FLAG=1)
       6.  DB + summary report (store_char_in_db=1, REPORT_FLAG=1,
       DUMP_FLAG=0)
       7.  DB + summary report + dump files (store_char_in_db=1,
       REPORT_FLAG=1,
       DUMP_FLAG=1)
       Scenario 4 will be assumed the default scenario--everything goes
       just to the database, which can be queried off-line to generate
       "custom" reports.

       One other scenario could be selected.  That's where
       store_char_in_db=0, REPORT_FLAG=0, and DUMP_FLAG=0.  This
       combination will be treated as an error (obviously)--
       characterization has been specified, but where are the
       results supposed to go?  */
    if ((!wo_params.store_char_in_db) && (!wo_params.report_flag)
            && (!wo_params.dump_flag))
    {
        IAS_LOG_ERROR("No output destination is specified for "
                "characterization results");
        ias_cpf_free(cpf);
        ias_misc_free_report_header(rpt_summary.header.common_header);
        exit(EXIT_FAILURE);
    }

    if ((wo_params.report_flag) || (wo_params.dump_flag))
    {
        IAS_LOG_DEBUG("Preparing summary report and/or dump file...");
        status = initialize_solarcal_report_data(wo_params, omf_params,
                angles, &rpt_summary);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Initializing report and/or "
                    "dump file information, program is terminating");
            ias_cpf_free(cpf);
            ias_misc_free_report_header(rpt_summary.header.common_header);
            exit(EXIT_FAILURE);
        }
    }

    /* Establish connections to the characterization database (for
       getting histogram statistics data, at minimum) */
    IAS_LOG_DEBUG("Preparing database access...");
    open_db = connect_to_db();
    if (open_db == NULL)
    {
        IAS_LOG_ERROR("Failed to connect to database, program "
                "terminating");
        ias_cpf_free(cpf);
        ias_misc_free_report_header(rpt_summary.header.common_header);
        exit(EXIT_FAILURE);
    }

    /* Get the database ID corresponding to the current work order */
    status = ias_misc_get_char_id(open_db,
            "RADIOMETRIC_WORK_ORDER_COMMON",
            "radiometric_wo_common_id",
            wo_params.work_order_id, rad_wo_common_id);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Retrieving database ID corresponding to current "
                "radiometric work order %s", wo_params.work_order_id);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        ias_cpf_free(cpf);
        ias_misc_free_report_header(rpt_summary.header.common_header);
        exit(EXIT_FAILURE); 
    }

    /* Set up trending */
    if (wo_params.store_char_in_db)
    {
        IAS_LOG_DEBUG("Band absolute and detector relative gains will be "
                "written to the database");

        /* Initiate a transaction for this application */
        ias_db_start_transaction(open_db);

        /* If this work order ID was previously trended, delete all the
           information from the solar_response, solar_response_band and
           solar_response_detector tables */
        status = clear_solarcal_table(open_db, rad_wo_common_id,
                "SOLAR_RESPONSE_DETECTOR");
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Clearing SOLAR_RESPONSE_DETECTOR table");
            ias_db_rollback_transaction(open_db);
            ias_db_close_connection(open_db);
            ias_db_close_database_lib();
            ias_misc_free_report_header(rpt_summary.header.common_header);
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }

        status = clear_solarcal_table(open_db, rad_wo_common_id,
                "SOLAR_RESPONSE_BAND");
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Clearing SOLAR_RESPONSE_BAND table");
            ias_db_rollback_transaction(open_db);
            ias_db_close_connection(open_db);
            ias_db_close_database_lib();
            ias_misc_free_report_header(rpt_summary.header.common_header);
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }

        status = clear_solarcal_table(open_db, rad_wo_common_id,
                "SOLAR_RESPONSE");
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Clearing main SOLAR_RESPONSE table");
            ias_db_rollback_transaction(open_db);
            ias_db_close_connection(open_db);
            ias_db_close_database_lib();
            ias_misc_free_report_header(rpt_summary.header.common_header);
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }
    }


/************************************************************************
                           BAND PROCESSING
************************************************************************/

    /* NOTE:  From this point on,  if there's an error, all buffers will
       be freed.  If they're NULL (i.e. not allocated), nothing is
       done, so we shouldn't crash from an attempt to do something
       to an invalid memory block */
    for (index = 0; index < wo_params.num_bands_to_process; index++)
    {
        const IAS_BAND_ATTRIBUTES *band_attributes = NULL;
        double *hist_stats_mean  = NULL;    /* Histogram statistics means for
                                               the current band/SCA */
        double *hist_stats_stdev = NULL;    /* Histogram statistics std devs
                                               for the current band/SCA */
        double *hist_stats_num_frames = NULL; /* Histogram statistics
                                               number of valid samples for
                                               the current band/SCA */
        double *diffuser_radiances = NULL;  /* CPF-based diffuser radiances
                                               for all detectors in the
                                               current band */
        double *diffuser_reflectances = NULL; /* CPF-based diffuser
                                               reflectances for all
                                               detectors in the current
                                               band */
        double *relative_gains = NULL;      /* Radiance-based relative gains
                                               for all detectors in the
                                               current band */
        double *relative_gains_stdev = NULL;  /* Radiance-based relative gain
                                               std deviations for all
                                               detectors in the current
                                               band */
        int *statistics_invalid_flags = NULL;

        int band_index;
        int sca_index;
        int num_band_detectors;

        /* Get the current band's information */
        IAS_LOG_DEBUG("Getting sensor attributes for band %d...",
                wo_params.bands[index]);

        band_attributes =
            ias_sat_attr_get_band_attributes(wo_params.bands[index]);
        if (band_attributes == NULL)
        {
            IAS_LOG_ERROR("Obtaining sensor attributes for "
                    "band %d", wo_params.bands[index]);
            ias_db_rollback_transaction(open_db);
            ias_db_close_connection(open_db);
            ias_db_close_database_lib();
            ias_misc_free_report_header(rpt_summary.header.common_header);
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }

        /* Get the index for the current band */
        band_index = band_attributes->band_index;

        /* Determine the total number of detectors in the current band */
        num_band_detectors =
            band_attributes->scas * band_attributes->detectors_per_sca;

        IAS_LOG_INFO("Processing band %d...", wo_params.bands[index]);

        /* Check to make sure we're not processing a TIRS band, or
           an OLI blind/VRP band.  If we are, skip it */
        if (band_attributes->spectral_type == IAS_SPECTRAL_THERMAL)
        {
            IAS_LOG_WARNING("TIRS band number %d detected, skipping to "
                "next band in list", band_attributes->band_number);
            continue;
        }
        if (band_attributes->band_classification != IAS_NORMAL_BAND)
        {
            IAS_LOG_WARNING("OLI VRP or blind band number %d detected, "
                "skipping to next band in list",
                band_attributes->band_number);
            continue;
        }

        /* Fill in the band attributes information in band_gains for
           the current band */
        memset(&band_gains, 0, sizeof(RPS_SOLARCAL_BAND_GAINS));
        strcpy(band_gains.band_name, band_attributes->band_name);
        band_gains.band_number = band_attributes->band_number;
        band_gains.scas = band_attributes->scas;
        band_gains.detectors_per_sca = band_attributes->detectors_per_sca;

        /* NOTE:  All buffers will be reallocated for the current band
           being processed.  This should reasonably deal with cases
           where the bands fundamentally change (such as changing
           from/to the PAN band to/from an MS band, etc)  Pointers
           representing these buffers will be reset to NULL when their 
           memory blocks are released. */

        /* Allocate the histogram statistics data buffers */
        IAS_LOG_DEBUG("Allocating data buffers...");

        hist_stats_mean = malloc(num_band_detectors
            * sizeof(*hist_stats_mean));
        hist_stats_stdev = malloc(num_band_detectors
            * sizeof(*hist_stats_stdev));
        hist_stats_num_frames = malloc(num_band_detectors
            * sizeof(*hist_stats_num_frames));
        diffuser_radiances = malloc(num_band_detectors
            * sizeof(*diffuser_radiances));
        diffuser_reflectances = malloc(num_band_detectors
            * sizeof(*diffuser_reflectances));
        relative_gains = malloc(num_band_detectors
            * sizeof(*relative_gains));
        relative_gains_stdev = malloc(num_band_detectors
            * sizeof(*relative_gains_stdev));

        /* Allocate the invalid flag data buffer */
        statistics_invalid_flags = calloc(num_band_detectors,
            sizeof(int));

        if ((hist_stats_mean == NULL) || (hist_stats_stdev == NULL)
            || (hist_stats_num_frames == NULL)
            || (diffuser_radiances == NULL)
            || (diffuser_reflectances == NULL)
            || (relative_gains == NULL)
            || (relative_gains_stdev == NULL)
            || (statistics_invalid_flags == NULL))
        {
            IAS_LOG_ERROR("Allocating data buffers for band %d",
                    band_attributes->band_number);
            cleanup_application(wo_params.store_char_in_db,
                open_db, &rpt_summary, relative_gains,
                relative_gains_stdev, hist_stats_mean,
                hist_stats_stdev, hist_stats_num_frames,
                cpf, diffuser_radiances, diffuser_reflectances,
                statistics_invalid_flags);
            exit(EXIT_FAILURE);
        }

/*--------------------------------------------------------------------------
                            GAIN CALCULATIONS
----------------------------------------------------------------------------*/
        IAS_LOG_DEBUG("Preparing absolute and relative gain calculations "
                "for band %d...", band_attributes->band_number);

        for (sca_index = 0; sca_index < band_attributes->scas; sca_index++)
        {
            int num_inoperable_detectors;
            int sca_number = sca_index + 1;
            int *inoperable_ptr = NULL;

            IAS_LOG_DEBUG("Obtaining inoperable detector info for SCA %d...",
                    sca_number);

            /* Find out how many inoperable detectors are in the list,
               and if there are any, get the list for the current SCA */
            num_inoperable_detectors =
                det_status->inoperable_count[band_index][sca_index];
            if (num_inoperable_detectors != 0)
            {
                inoperable_ptr =
                    det_status->inoperable[band_index][sca_index];
                get_valid_data_flags_from_cpf_inop_list(inoperable_ptr,
                    num_inoperable_detectors, sca_number,
                    band_attributes->detectors_per_sca,
                    statistics_invalid_flags);
            }
            else
            {
                inoperable_ptr = NULL;
            }

            /* Get the histogram statistics for each SCA, and populate the
               data buffers for the entire band.  Inoperable detectors
               will be flagged accordingly */
            IAS_LOG_DEBUG("Getting histogram statistics for SCA %d...",
                    sca_number);

            status = get_histogram_statistics(open_db, rad_wo_common_id,
                band_attributes->spectral_type,
                band_attributes->band_number, sca_number,
                band_attributes->detectors_per_sca,
                wo_params.rps_proc_level,
                hist_stats_mean, hist_stats_stdev,
                hist_stats_num_frames);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Getting histogram statistics for band %d "
                        "SCA %d", band_attributes->band_number,
                        sca_number);
                cleanup_application(wo_params.store_char_in_db,
                    open_db, &rpt_summary, relative_gains,
                    relative_gains_stdev, hist_stats_mean,
                    hist_stats_stdev, hist_stats_num_frames,
                    cpf, diffuser_radiances, diffuser_reflectances,
                    statistics_invalid_flags);
                exit(EXIT_FAILURE);
            }

            /* Get the diffuser radiances and reflectances from the
               CPF */
            status = copy_cpf_diffuser_data(diff_rad, inoperable_ptr,
                band_index, sca_index, band_attributes->detectors_per_sca,
                GET_DIFFUSER_RADIANCES, panel_type_string,
                diffuser_radiances);
            if (status != SUCCESS)
            {
                cleanup_application(wo_params.store_char_in_db,
                    open_db, &rpt_summary, relative_gains,
                    relative_gains_stdev, hist_stats_mean,
                    hist_stats_stdev, hist_stats_num_frames,
                    cpf, diffuser_radiances,
                    diffuser_reflectances,
                    statistics_invalid_flags);
                exit(EXIT_FAILURE);
            }

            status = copy_cpf_diffuser_data(diff_rad, inoperable_ptr,
                band_index, sca_index, band_attributes->detectors_per_sca,
                GET_DIFFUSER_REFLECTANCES, panel_type_string,
                diffuser_reflectances);
            if (status != SUCCESS)
            {
                cleanup_application(wo_params.store_char_in_db,
                    open_db, &rpt_summary, relative_gains,
                    relative_gains_stdev, hist_stats_mean,
                    hist_stats_stdev, hist_stats_num_frames,
                    cpf, diffuser_radiances, diffuser_reflectances,
                    statistics_invalid_flags);
                exit(EXIT_FAILURE);
            }
        }       /*       END SCA PROCESSING LOOP   */

        /* Calculate the band mean and standard deviation.  First,
           make sure inoperable detectors are flagged */
        IAS_LOG_DEBUG("Calculating band average histogram statistics...");

        status = ias_rps_calculate_filtered_mean(hist_stats_mean,
            statistics_invalid_flags, num_band_detectors,
            &(band_gains.hist_stats_mean), NULL);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating histogram statistics mean "
                "for band %d", band_attributes->band_number);
            cleanup_application(wo_params.store_char_in_db,
                open_db, &rpt_summary, relative_gains,
                relative_gains_stdev, hist_stats_mean,
                hist_stats_stdev, hist_stats_num_frames,
                cpf, diffuser_radiances, diffuser_reflectances,
                statistics_invalid_flags);
            exit(EXIT_FAILURE);
        }

        status = ias_rps_calculate_filtered_stdev(hist_stats_mean,
            statistics_invalid_flags, num_band_detectors,
            band_gains.hist_stats_mean, &(band_gains.hist_stats_stdev));
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating histogram statistics std "
                "deviation for band %d", band_attributes->band_number);
            cleanup_application(wo_params.store_char_in_db,
                open_db, &rpt_summary, relative_gains,
                relative_gains_stdev, hist_stats_mean,
                hist_stats_stdev, hist_stats_num_frames, cpf,
                diffuser_radiances, diffuser_reflectances,
                statistics_invalid_flags);
            exit(EXIT_FAILURE);
        }

        /* Calculate the diffuser radiance and reflectance mean and
           std deviation */
        IAS_LOG_DEBUG("Calculating band diffuser radiances and "
            "reflectances...");

        status = ias_rps_calculate_filtered_mean(diffuser_radiances,
            statistics_invalid_flags, num_band_detectors,
            &(band_gains.diffuser_radiance_mean), NULL);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating diffuser radiance mean "
                "for band %d", band_attributes->band_number);
            cleanup_application(wo_params.store_char_in_db,
                open_db, &rpt_summary, relative_gains,
                relative_gains_stdev, hist_stats_mean,
                hist_stats_stdev, hist_stats_num_frames, cpf,
                diffuser_radiances, diffuser_reflectances,
                statistics_invalid_flags);
            exit(EXIT_FAILURE);
        }

        status = ias_rps_calculate_filtered_stdev(diffuser_radiances,
            statistics_invalid_flags, num_band_detectors,
            band_gains.diffuser_radiance_mean,
            &(band_gains.diffuser_radiance_stdev));
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating diffuser radiance std deviation "
                "for band %d", band_attributes->band_number);
            cleanup_application(wo_params.store_char_in_db,
                open_db, &rpt_summary, relative_gains,
                relative_gains_stdev, hist_stats_mean,
                hist_stats_stdev, hist_stats_num_frames,
                cpf, diffuser_radiances, diffuser_reflectances,
                statistics_invalid_flags);
            exit(EXIT_FAILURE);
        }

        status = ias_rps_calculate_filtered_mean(diffuser_reflectances,
            statistics_invalid_flags, num_band_detectors,
            &(band_gains.diffuser_reflectance_mean), NULL);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating diffuser reflectance mean "
                "for band %d", band_attributes->band_number);
            cleanup_application(wo_params.store_char_in_db,
                open_db, &rpt_summary, relative_gains,
                relative_gains_stdev, hist_stats_mean, 
                hist_stats_stdev, hist_stats_num_frames, cpf,
                diffuser_radiances, diffuser_reflectances,
                statistics_invalid_flags);
            exit(EXIT_FAILURE);
        }

        status = ias_rps_calculate_filtered_stdev(diffuser_reflectances,
            statistics_invalid_flags, num_band_detectors,
            band_gains.diffuser_reflectance_mean,
            &(band_gains.diffuser_reflectance_stdev));
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating diffuser reflectance std deviation "
                "for band %d", band_attributes->band_number);
            cleanup_application(wo_params.store_char_in_db,
                open_db, &rpt_summary, relative_gains,
                relative_gains_stdev, hist_stats_mean,
                hist_stats_stdev, hist_stats_num_frames, cpf,
                diffuser_radiances, diffuser_reflectances,
                statistics_invalid_flags);
            exit(EXIT_FAILURE);
        }

        /* Calculate radiance-based absolute gains/detector relative gains
           and reflectance gains.  For radiance-based gains, there's no
           dependence on the solar geometry, so set the cos_theta parameter
           to 1.0.  For reflectance-based gains, we'll need the explicit
           cos_theta value determined earlier; in addition, we'll need to
           set the relative gains data pointers to NULL */
        IAS_LOG_DEBUG("Calculating radiance-based absolute/relative "
                "gains and reflectance-based absolute gains...");

        status = calculate_gains(hist_stats_mean, hist_stats_stdev,
            diffuser_radiances, statistics_invalid_flags,
            num_band_detectors, earth_sun_dist_sqr, 1.0, relative_gains,
            relative_gains_stdev, &(band_gains.radiance_gain),
            &(band_gains.radiance_gain_stdev));
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating band %d radiance-based "
                "absolute and relative gains",
                band_attributes->band_number);
            cleanup_application(wo_params.store_char_in_db,
                open_db, &rpt_summary, relative_gains,
                relative_gains_stdev, hist_stats_mean,
                hist_stats_stdev, hist_stats_num_frames, cpf,
                diffuser_radiances, diffuser_reflectances,
                statistics_invalid_flags);
            exit(EXIT_FAILURE);
        }

        status = calculate_gains(hist_stats_mean, hist_stats_stdev,
            diffuser_reflectances, statistics_invalid_flags,
            num_band_detectors, earth_sun_dist_sqr, cos_theta, NULL,
            NULL, &(band_gains.reflectance_gain),
            &(band_gains.reflectance_gain_stdev));
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating band %d reflectance-based "
                "absolute and gains", band_attributes->band_number);
            cleanup_application(wo_params.store_char_in_db,
                open_db, &rpt_summary, relative_gains,
                relative_gains_stdev, hist_stats_mean,
                hist_stats_stdev, hist_stats_num_frames, cpf,
                diffuser_radiances, diffuser_reflectances,
                statistics_invalid_flags);
            exit(EXIT_FAILURE);
        }

        /* We don't need the diffuser data for this band anymore.  Do,
           however, reset the freed pointers to NULL to prevent issues
           with non-NULL but invalid pointers during shutdown */
        free(diffuser_radiances);
        free(diffuser_reflectances);
        diffuser_radiances = NULL;
        diffuser_reflectances = NULL;

        /* END GAIN CALCULATIONS */

/*------------------------------------------------------------------------
                      WRITE GAINS TO DB
--------------------------------------------------------------------------*/
        /*  Write gain results for each SCA to the database, if the
            store_char_in_db WO parameter is set */
        if (wo_params.store_char_in_db)
        {
            double *sca_relgain_data = NULL;
            double *sca_relgain_std_data = NULL;
            IAS_DB_NULL_TYPE *flagged_detector_ids = NULL;
            int *detector_ids = NULL;
            int detector_index;

            IAS_LOG_DEBUG("Writing band %d results to characterization "
                    "database...", band_attributes->band_number);

            /* Start with the band absolute gains */
            status = write_to_band_table(open_db, rad_wo_common_id,
                band_gains);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Writing band absolute gains to DB in "
                    "band %d", band_attributes->band_number);
                cleanup_application(wo_params.store_char_in_db,
                    open_db, &rpt_summary, relative_gains,
                    relative_gains_stdev, hist_stats_mean,
                    hist_stats_stdev, hist_stats_num_frames, cpf,
                    diffuser_radiances, diffuser_reflectances,
                    statistics_invalid_flags);
                exit(EXIT_FAILURE);
            }

            /* Allocate the (integer-based) NULLed detector ID arrays
               and the relative gain arrays for each SCA */
            detector_ids = malloc(band_attributes->detectors_per_sca
                * sizeof(*detector_ids));
            flagged_detector_ids = malloc(band_attributes->detectors_per_sca
                * sizeof(*flagged_detector_ids));
            sca_relgain_data = malloc(band_attributes->detectors_per_sca
                * sizeof(*sca_relgain_data));
            sca_relgain_std_data = malloc(band_attributes->detectors_per_sca
                * sizeof(*sca_relgain_std_data));
            if ((detector_ids == NULL) || (flagged_detector_ids == NULL)
                || (sca_relgain_data == NULL)
                || (sca_relgain_std_data == NULL))
            {
                IAS_LOG_ERROR("Allocating SCA data buffers for "
                    "band %d", band_attributes->band_number);
                free(detector_ids);
                free_db_sca_data_buffers(flagged_detector_ids,
                    sca_relgain_data,
                    sca_relgain_std_data);
                cleanup_application(wo_params.store_char_in_db,
                    open_db, &rpt_summary, relative_gains,
                    relative_gains_stdev, hist_stats_mean,
                    hist_stats_stdev, hist_stats_num_frames, cpf,
                    diffuser_radiances, diffuser_reflectances,
                    statistics_invalid_flags);
                exit(EXIT_FAILURE);
            }

            /* Populate the detector ID buffer */
            for (detector_index = 0;
                 detector_index < band_attributes->detectors_per_sca;
                 detector_index++)
            {
                detector_ids[detector_index] = (detector_index + 1);
            }

            /* Populate the buffers for each SCA */
            for (sca_index = 0;
                 sca_index < band_attributes->scas;
                 sca_index++)
            {
                int sca_number = sca_index + 1;
                int start_detector_index =
                    sca_index * band_attributes->detectors_per_sca;
                size_t flagged_id_buffer_size =
                    band_attributes->detectors_per_sca
                        * sizeof(IAS_DB_NULL_TYPE);

                /* Initialize the flagged detector data buffer to 0 */
                memset(flagged_detector_ids, 0, flagged_id_buffer_size);

                /* Populate the data buffers */
                populate_db_sca_data_buffers(start_detector_index,
                    band_attributes->detectors_per_sca,
                    relative_gains, relative_gains_stdev,
                    statistics_invalid_flags,
                    flagged_detector_ids, sca_relgain_data,
                    sca_relgain_std_data);

                /* Write the detector relative gains */
                status = write_to_detector_table(open_db, rad_wo_common_id,
                    band_attributes->band_number, sca_number,
                    band_attributes->detectors_per_sca,
                    detector_ids, flagged_detector_ids,
                    sca_relgain_data, sca_relgain_std_data);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Writing detector relative gains to DB in "
                        "band %d SCA %d", band_attributes->band_number,
                        sca_number);
                    free(detector_ids);
                    free_db_sca_data_buffers(flagged_detector_ids,
                        sca_relgain_data, sca_relgain_std_data);
                    cleanup_application(wo_params.store_char_in_db,
                        open_db, &rpt_summary, relative_gains,
                        relative_gains_stdev, hist_stats_mean,
                        hist_stats_stdev, hist_stats_num_frames, cpf,
                        diffuser_radiances, diffuser_reflectances,
                        statistics_invalid_flags);
                    exit(EXIT_FAILURE);
                }
            }

            /* Free the SCA data buffers */
            free(detector_ids);
            free_db_sca_data_buffers(flagged_detector_ids,
                sca_relgain_data, sca_relgain_std_data);
            detector_ids = NULL;
            flagged_detector_ids = NULL;
            sca_relgain_data = NULL;
            sca_relgain_std_data = NULL;

        }   /* END WRITE GAINS TO DB */


/*------------------------------------------------------------------------
          ASSEMBLE SUMMARY REPORT DATA
--------------------------------------------------------------------------*/
        if ((wo_params.report_flag) || (wo_params.dump_flag))
        {
            /* Write the band-specific data to the summary report
               data structure */
            initialize_solarcal_summary_band_data(band_index,
                band_gains, &rpt_summary);

            /* Calculate the average number of frames used in the histogram
               statistics calculations */
            status = ias_rps_calculate_filtered_mean(hist_stats_num_frames,
                statistics_invalid_flags, num_band_detectors,
                &num_frames_averaged_band, NULL);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Calculating band %d histogram "
                    "statstics number-of-frames",
                    band_attributes->band_number);
                cleanup_application(wo_params.store_char_in_db,
                    open_db, &rpt_summary, relative_gains,
                    relative_gains_stdev, hist_stats_mean,
                    hist_stats_stdev, hist_stats_num_frames, cpf,
                    diffuser_radiances, diffuser_reflectances,
                    statistics_invalid_flags);
                exit(EXIT_FAILURE);
            }

            /* Keep track of the average number of frames used to
               calculate the histogram statistics in each band.  */
            num_frames_averaged += num_frames_averaged_band;

            /* Set the band number in the common report header */
            ias_misc_set_header_band_number(rpt_summary.header.common_header,
                band_attributes->band_number);

        }        /* END ASSEMBLE SUMMARY REPORT DATA  */


/*------------------------------------------------------------------------
                        WRITE DATA TO DUMP FILE
--------------------------------------------------------------------------*/
        if (wo_params.dump_flag)
        {
            char dump_filename[PATH_MAX];           /* Name of dump file */
            FILE *dump_fptr = NULL;                 /* Dump file handle*/

            /* Create the dump file for the current band */
            status = snprintf(dump_filename, sizeof(dump_filename),
                "%s_dumpfile_B%02d.txt", wo_params.report_filename_prefix,
                band_attributes->band_number);
            if ((status < 0) || (status >= sizeof(dump_filename)))
            {
                IAS_LOG_ERROR("Constructing band %d dump file name",
                    band_attributes->band_number);
                cleanup_application(wo_params.store_char_in_db,
                   open_db, &rpt_summary, relative_gains,
                   relative_gains_stdev, hist_stats_mean,
                   hist_stats_stdev, hist_stats_num_frames, cpf,
                   diffuser_radiances, diffuser_reflectances,
                   statistics_invalid_flags);
                exit(EXIT_FAILURE);
            }
            dump_fptr = fopen(dump_filename, "w");
            if (dump_fptr == NULL)
            {
                IAS_LOG_ERROR("Cannot open/create dump file %s",
                    dump_filename);
                cleanup_application(wo_params.store_char_in_db,
                    open_db, &rpt_summary, relative_gains,
                    relative_gains_stdev, hist_stats_mean,
                    hist_stats_stdev, hist_stats_num_frames, cpf,
                    diffuser_radiances, diffuser_reflectances,
                    statistics_invalid_flags);
                exit(EXIT_FAILURE);
            }

            /* Write the header and validation information */
            status = write_header_to_dump_file(rpt_summary.header,
                dump_fptr);
            if (status != SUCCESS)
            {
                fclose(dump_fptr);
                IAS_LOG_ERROR("Writing header contents to dump file");
                cleanup_application(wo_params.store_char_in_db,
                    open_db, &rpt_summary, relative_gains,
                    relative_gains_stdev, hist_stats_mean,
                    hist_stats_stdev, hist_stats_num_frames, cpf,
                    diffuser_radiances, diffuser_reflectances,
                    statistics_invalid_flags);
                exit(EXIT_FAILURE);
            }

            /* Write the angle information to the dump file */
            status = write_angle_statistics_to_dump_file(
                rpt_summary.solar_angles, dump_fptr);
            if (status != SUCCESS)
            {
                fclose(dump_fptr);
                IAS_LOG_ERROR("Writing angle summary statistics to "
                    "dump file");
                cleanup_application(wo_params.store_char_in_db,
                   open_db, &rpt_summary, relative_gains,
                   relative_gains_stdev, hist_stats_mean,
                   hist_stats_stdev, hist_stats_num_frames, cpf,
                   diffuser_radiances, diffuser_reflectances,
                   statistics_invalid_flags);
                exit(EXIT_FAILURE);
            }

            /* Write the absolute band summary histogram statistics,
               diffuser panel radiances/reflectances, and the
               radiance-based / reflectance-based absolute gain
               statistics to the dump file */
            status = write_band_gains_to_dump_file(band_gains, dump_fptr);
            if (status != SUCCESS)
            {
                fclose(dump_fptr);
                IAS_LOG_ERROR("Writing band %d relative gains to dump file",
                    band_attributes->band_number);
                cleanup_application(wo_params.store_char_in_db,
                    open_db, &rpt_summary, relative_gains,
                    relative_gains_stdev, hist_stats_mean,
                    hist_stats_stdev, hist_stats_num_frames, cpf,
                    diffuser_radiances, diffuser_reflectances,
                    statistics_invalid_flags);
                exit(EXIT_FAILURE);
            }

            /* Finally, write the relative gains to the dump file */
            status = write_relative_gains_to_dump_file(
                relative_gains, relative_gains_stdev,
                statistics_invalid_flags,
                band_attributes->band_number,
                band_attributes->scas,
                num_band_detectors, dump_fptr);
            if (status != SUCCESS)
            {
                fclose(dump_fptr);
                IAS_LOG_ERROR("Writing band %d relative gains to dump file "
                    "%s", band_attributes->band_number,
                    dump_filename);
                cleanup_application(wo_params.store_char_in_db,
                    open_db, &rpt_summary, relative_gains,
                    relative_gains_stdev, hist_stats_mean,
                    hist_stats_stdev, hist_stats_num_frames, cpf,
                    diffuser_radiances, diffuser_reflectances,
                    statistics_invalid_flags);
                exit(EXIT_FAILURE);
            }

            /* Close the file.  If it doesn't close "properly", it's an
               error */
            if (fclose(dump_fptr) != 0)
            {
                IAS_LOG_ERROR("Closing band %d dump file",
                    band_attributes->band_number);
                cleanup_application(wo_params.store_char_in_db,
                    open_db, &rpt_summary, relative_gains,
                    relative_gains_stdev, hist_stats_mean,
                    hist_stats_stdev, hist_stats_num_frames, cpf,
                    diffuser_radiances, diffuser_reflectances,
                    statistics_invalid_flags);
                exit(EXIT_FAILURE);
            }
        }

        /* Free up resources to allocate for the next band. */
        free(relative_gains);
        free(relative_gains_stdev);
        free(hist_stats_mean);
        free(hist_stats_stdev);
        free(hist_stats_num_frames);
        free(statistics_invalid_flags);
        relative_gains         = NULL;
        relative_gains_stdev   = NULL;
        hist_stats_mean        = NULL;
        hist_stats_stdev       = NULL;
        hist_stats_num_frames  = NULL;
        statistics_invalid_flags = NULL;
    }              /* END BAND PROCESSING LOOP */

    /* Determine the average number of frames used in the histogram
       statistics calculations across all bands */
    num_frames_averaged /= (wo_params.num_bands_to_process);
    rpt_summary.header.num_frames_averaged = (int)num_frames_averaged;

/*-------------------------------------------------------------------------
                   WRITE SOLAR SUMMARY RESULTS TO DB
---------------------------------------------------------------------------*/
    /* Write the summary collect information to the main SOLAR_RESPONSE
       table.  Should be done here because at this point, all the required
       information has been collected */
    if (wo_params.store_char_in_db)
    {
        /* Write the required validation information to the database */
        IAS_LOG_DEBUG("Writing current validation data for this work order "
            "to database...");
        status = write_to_main_table(open_db, rad_wo_common_id,
                rpt_summary.header, angles);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Writing current validation data to "
                    "characterization database");
            ias_db_rollback_transaction(open_db);
            ias_db_close_connection(open_db);
            ias_db_close_database_lib();
            ias_misc_free_report_header(rpt_summary.header.common_header);
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }
    }   /* END WRITE MAIN SOLAR RESPONSE TABLE */

/*------------------------------------------------------------------------
                   WRITE RESULTS TO SUMMARY REPORT
--------------------------------------------------------------------------*/
    if (wo_params.report_flag)
    {
        char report_filename[PATH_MAX];     /* Name of optional summary
                                               report file */
        FILE *rpt_fptr = NULL;              /* Summary report file handle */

        IAS_LOG_DEBUG("Writing band gains to summary report...");

        /* Build the full path and name of the report file */
        status = snprintf(report_filename, sizeof(report_filename),
                          "%s_summary.txt", wo_params.report_filename_prefix);
        if ((status < 0) || (status >= sizeof(report_filename)))
        {
            IAS_LOG_ERROR("Constructing summary report file name");
            ias_cpf_free(cpf);
            free(rpt_summary.band_gains);
            ias_misc_free_report_header(rpt_summary.header.common_header);
            if (wo_params.store_char_in_db)
            {
                ias_db_rollback_transaction(open_db);
            }
            ias_db_close_connection(open_db);
            ias_db_close_database_lib();
            exit(EXIT_FAILURE);
        }

        /* Try to create the summary report file with the header
           information */
        rpt_fptr = fopen(report_filename, "w");
        if (rpt_fptr == NULL)
        {
            IAS_LOG_ERROR("Creating summary report file %s",
                report_filename);
            ias_cpf_free(cpf);
            free(rpt_summary.band_gains);
            ias_misc_free_report_header(rpt_summary.header.common_header);
            if (wo_params.store_char_in_db)
            {
                ias_db_rollback_transaction(open_db);
            }
            ias_db_close_connection(open_db);
            ias_db_close_database_lib();
            exit(EXIT_FAILURE);
        }

        /* Write the header information to the summary report */
        status = write_header_to_summary_report(rpt_summary.header,
            rpt_fptr);
        if (status != SUCCESS)
        {
            fclose(rpt_fptr);
            IAS_LOG_ERROR("Writing header to summary report %s",
                report_filename);
            ias_cpf_free(cpf);
            free(rpt_summary.band_gains);
            ias_misc_free_report_header(rpt_summary.header.common_header);
            if (wo_params.store_char_in_db)
            {
                ias_db_rollback_transaction(open_db);
            }
            ias_db_close_connection(open_db);
            ias_db_close_database_lib();
            exit(EXIT_FAILURE);
        }

        /* Write the solar angle statistics to the summary report */
        status = write_angle_statistics_to_summary_report(
            rpt_summary.solar_angles, rpt_fptr);
        if (status != SUCCESS)
        {
            fclose(rpt_fptr);
            IAS_LOG_ERROR("Writing solar angle statistics to summary "
                "report %s", report_filename);
            ias_cpf_free(cpf);
            free(rpt_summary.band_gains);
            ias_misc_free_report_header(rpt_summary.header.common_header);
            if(wo_params.store_char_in_db);
            {
                ias_db_rollback_transaction(open_db);
            }
            ias_db_close_connection(open_db);
            ias_db_close_database_lib();
            exit(EXIT_FAILURE);
        }

        /* Write the summary report "body" */
        status = write_solarcal_report_data(num_oli_bands, rpt_summary,
            rpt_fptr);
        if (status != SUCCESS)
        {
            fclose(rpt_fptr);
            IAS_LOG_ERROR("Writing results to summary report %s",
                report_filename);
            free(rpt_summary.band_gains);
            ias_cpf_free(cpf);
            ias_misc_free_report_header(rpt_summary.header.common_header);
            if (wo_params.store_char_in_db)
            {
                ias_db_rollback_transaction(open_db);
            }
            ias_db_close_connection(open_db);
            ias_db_close_database_lib();
            exit(EXIT_FAILURE);
        }

        /* We're done with the data buffer containing the band gain
           information--free it */
        free(rpt_summary.band_gains);

        /* Close the report file.  If it doesn't close "properly", it's
           an error */
        if (fclose(rpt_fptr))
        {
            IAS_LOG_ERROR("Closing summary report file %s",
               report_filename);
            final_status = EXIT_FAILURE;
        }
    }   /* END SUMMARY REPORT PROCESSING */


/*-----------------------------------------------------------------------
                         CLEAN-UP
-------------------------------------------------------------------------*/

    /* Make sure we can commit to the transaction */
    if (wo_params.store_char_in_db)
    {
        status = ias_db_commit_transaction(open_db);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Error committing database writes");
            ias_db_rollback_transaction(open_db);
            if (final_status == EXIT_SUCCESS)
            {
                final_status = EXIT_FAILURE;
            }
        }
    }

    /* Shut down the database connection */
    ias_db_close_connection(open_db);
    ias_db_close_database_lib();

    /* Free up CPF resources */
    ias_cpf_free(cpf);

    /* Free the report header information */
    ias_misc_free_report_header(rpt_summary.header.common_header);

    /* Log the final status */
    if (final_status != EXIT_SUCCESS)
    {
        IAS_LOG_INFO("Solar calibration processing failed.");
    }
    else
    {
        IAS_LOG_INFO("Solar calibration processing successfully completed.");
    }

    return final_status;
}
