#ifndef SOLARCAL_OLI_H
#define SOLARCAL_OLI_H

/* "Common" C library headers */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

/* "Common" IAS headers */
#include "ias_ancillary_io.h"
#include "ias_const.h"
#include "ias_cpf.h"
#include "ias_logging.h"
#include "ias_l0r.h"
#include "ias_miscellaneous.h"
#include "ias_radiometry.h"
#include "ias_satellite_attributes.h"
#include "ias_types.h"

                /* Some useful constants...   */
/* Validation constants */
#define RPS_SOLARCAL_METADATA_VALIDATED              0
#define RPS_SOLARCAL_METADATA_INVALID_SPACECRAFT     1
#define RPS_SOLARCAL_METADATA_INVALID_SENSOR         2
#define RPS_SOLARCAL_METADATA_INVALID_WRS_PATHROW    3

#define RPS_SOLARCAL_IMAGEHEADER_VALIDATED                0
#define RPS_SOLARCAL_IMAGEHEADER_INVALID_INTEGRATION_TIME 1

/* Work order option default values */
#define RPS_SOLARCAL_DEFAULT_STORE_CHAR_IN_DB       1
#define RPS_SOLARCAL_DEFAULT_REPORT_FLAG            0
#define RPS_SOLARCAL_DEFAULT_DUMP_FLAG              0

/* CPF diffuser data selection flags */
#define GET_DIFFUSER_RADIANCES    1
#define GET_DIFFUSER_REFLECTANCES 2


/* Default RPS processing level for the OLI histogram statistics input data.
   The currently recommended radiometric processing required for the 
   histogram statistics used in this characterization consists of temperature
   correction of bias, bias correction, and response linearization.  */
#define RPS_SOLARCAL_TEMPCORRECT           "CTBL"

                  /* Internal data structures */
typedef struct rps_solarcal_band_gains
{
    int band_number;
    char band_name[IAS_BAND_NAME_SIZE];
    int scas;
    int detectors_per_sca;
    double hist_stats_mean;
    double hist_stats_stdev;
    double diffuser_radiance_mean;           /* Mean of CPF diffuser
                                                radiances */
    double diffuser_radiance_stdev;          /* Std dev of CPF diffuser
                                                radiances */
    double radiance_gain;
    double radiance_gain_stdev;
    double diffuser_reflectance_mean;        /* Mean of CPF diffuser
                                                reflectances */
    double diffuser_reflectance_stdev;       /* Std dev of reflectance-based
                                                diffuser reflectances */
    double reflectance_gain;
    double reflectance_gain_stdev;
} RPS_SOLARCAL_BAND_GAINS;


typedef struct rps_solarcal_solar_angles
{
    double earth_sun_distance_factor; /* Earth-sun distance (AU) */
    double incident_solar_zenith;     /* Mean incident solar zenith angle for
                                         all samples */
    double incident_solar_zenith_stdev;
                                      /* Std dev of incident solar zenith
                                         angle for all samples */
    double view_solar_zenith;         /* Mean view solar zenith angle for
                                         all samples */
    double view_solar_zenith_stdev;   /* Std dev of view solar zenith angle
                                         for all samples */
    double view_solar_azimuth;        /* Mean view solar azimuth angle for
                                         all samples */
    double view_solar_azimuth_stdev;  /* Std dev of view solar azimuth angle
                                         for all samples */
} RPS_SOLARCAL_SOLAR_ANGLES;


typedef struct rps_solarcal_report_header
{
    IAS_REPORT_HEADER *common_header;
    RPS_REPORT_HEADER rps_header;

    uint16_t diffuser_resolver_position;       /* Diffuser resolver
                                                  position */
    int panel_id;                              /* Diffuser panel ID */
    int num_frames_averaged;                   /* Number of frames averaged
                                                  for histogram statistics */
} RPS_SOLARCAL_REPORT_HEADER;



/* Summary report data structure */

typedef struct rps_solarcal_report_summary
{
    RPS_SOLARCAL_REPORT_HEADER header;   /* RPS standard report header
                                            information */
    int store_char_in_db;                      /* Characterization results
                                            written to database */
    int report_flag;                     /* Characterization results
                                            written to summary files */
    int dump_flag;                       /* Characterization results
                                            written to dump files */

    RPS_SOLARCAL_SOLAR_ANGLES solar_angles;  /* Solar angle
                                                verification data */
    RPS_SOLARCAL_BAND_GAINS *band_gains;     /* Reported absolute gains
                                                for all bands */
} RPS_SOLARCAL_REPORT_SUMMARY;



/*           Data structures required for parameter I/O. */
typedef struct rps_solarcal_wo_params 
{
    /* Work order-related information -- parameter file
       identification/location */
    char work_order_id[IAS_WORKORDER_ID_SIZE];

    /* RPS processing level */
    char rps_proc_level[IAS_RPS_PROCESSING_LEVEL_LENGTH + 1];

    /* Output report filename prefix */
    char report_filename_prefix[PATH_MAX];

    /* Work order-related information -- processing options */
    int num_bands_to_process;
    int bands[IAS_MAX_NBANDS];        /* List of bands to process */

    /* Work order-related information -- output options */
    int store_char_in_db;
    int report_flag;
    int dump_flag;
    char report_name[PATH_MAX];
    char dump_name[PATH_MAX];
} RPS_SOLARCAL_WO_PARAMS;


typedef struct rps_solarcal_omf_params
{
    int satellite_number;                  /* Satellite number */
    char l0r_filename[PATH_MAX];           /* L0R interval filename */
    char cpf_filename[PATH_MAX];           /* CPF filename */
    char adp_ancillary_filename[PATH_MAX]; /* Name of ADP-corrected ancillary
                                              file */
} RPS_SOLARCAL_OMF_PARAMS;



/*-----------------------------------------------------------------------
                     FUNCTION PROTOTYPES
-------------------------------------------------------------------------*/

                         /* Parameter I/O */
int get_parameters
(
    const char *wo_params_filename,            /* I: Name of work order
                                                  parameter file */
    RPS_SOLARCAL_WO_PARAMS *wo_param_data,     /* O: WO parameters */
    RPS_SOLARCAL_OMF_PARAMS *omf_param_data    /* O: OMF parameters */
);



                      /* Collect validation */
void convert_diffuser_resolver_position_to_angle
(
    uint16_t diffuser_resolver_position,  /* I:  Diffuser position
                                             telemetry value */
    double *final_angle                   /* O:  Final deployment angle
                                             between 0 and 360 degrees */
);


int validate_image_header
(
    RPS_SOLARCAL_REPORT_HEADER header,        /* I: Report header data with
                                                 image header information
                                                 to validate */
    double cpf_integration_time_ms,           /* I: OLI nominal integration
                                                 time for MS bands from CPF */
    double cpf_integration_time_pan           /* I: OLI nominal integration
                                                 time for PAN band from CPF */
);


int validate_interval_metadata
(
    RPS_SOLARCAL_REPORT_HEADER header        /* I: Report header data with
                                                interval metadata information
                                                to validate */
);


int validate_collect_data
(
    RPS_SOLARCAL_REPORT_HEADER header,  /* I: Report header information with
                                           required data to validate.  Panel
                                           type is written to the report
                                           header after it is determined */
    const struct IAS_CPF_OLI_PARAMETERS *oli_params
                                        /* I: OLI parameters containing the
                                           nominal integration times in
                                           milliseconds */
);


                    /* Solar angle calculation */
double get_julian_date_for_jpl_ephem
(
    const char *date_string,          /* I: String representation of date */
    IAS_DATETIME_FORMAT_TYPE format   /* I: Date/time string format */
);


int process_jpl_ephemeris
(
    const double jpl_ephem_date,           /* I: Requested Julian date */
    double *earth_sun_distance,            /* O: Derived outputs from
                                              JPL solar ephemeris data */
    IAS_VECTOR *solar_position_vector
);


int get_jpl_earth_sun_distance
(
    const double julian_date,                   /* I: Equivalent Julian
                                                   date of current
                                                   acquisition */
    double *earth_sun_distance                  /* O: Earth-sun distance
                                                   factor applicable during
                                                   collect */
);


int get_jpl_solar_position
(
    const double julian_date,            /* I: Julian date of acqusition*/
    IAS_VECTOR *solar_position           /* O: Solar position vector in km*/
);


int convert_datetime_to_year_doy_sod
(
    IAS_DATETIME datetime,          /* I: Parsed IAS date/time string */
    double *year,                   /* O: Year */
    double *doy,                    /* O: Day-of-year */
    double *sod                     /* O: Second-of-day */
);


int calculate_sample_angles
(
    const IAS_ANC_ATTITUDE_DATA *attitude_data,
                                         /* I: Sensor attitude data records
                                            (ACS quaternions) */
    int start_record_index,              /* I: Index of 1st quaternion
                                            record sample acquired closest
                                            to image start time */
    int number_of_records,               /* I: Number of quaternion
                                            samples during actual solar
                                            collect */
    const IAS_VECTOR *solar_position,    /* I: JPL solar ephemeris data
                                            applicable to the current
                                            acquisition */
    double *incident_solar_zenith_angle, /* O: Populated instantaneous
                                            incident solar zenith angle
                                            array */
    double *view_solar_zenith_angle,     /* O: Populated instantaneous
                                            view solar zenith angle array */
    double *view_solar_azimuth_angle     /* O: Populated instantaneous
                                            view solar azimuth angle array */
);


int calculate_solar_angles
(
    const char *ancillary_data_filename,       /* I: Ancillary data file */
    double *start_time,                        /* I: Year / day-of-year /
                                                  second-of-day
                                                  corresponding to image
                                                  start time */
    double *stop_time,                         /* I: Year / day-of-year /
                                                  second-of-day
                                                  corresponding to image
                                                  stop time */
    const IAS_VECTOR *solar_position,          /* I: JPL solar ephemeris
                                                  data file name */
    RPS_SOLARCAL_SOLAR_ANGLES  *angles         /* O: Solar angle statistics
                                                  for the collect as a
                                                  whole */
);



                       /* Gain calculation */
int calculate_gains
(
    double *hist_stats_mean,          /* I: Histogram statistics means for the
                                         current band */
    double *hist_stats_stdev,         /* I: Histogram statistics std devs
                                         for the current band */
    double *diffuser_radiances,       /* I: CPF-based diffuser radiances for
                                         all detectors in the current band */
    const int *statistics_validity_flags,
                                      /* I: Detector valid data flags for the
                                         current band (0=valid data,
                                         nonzero=invalid data) */
    int num_band_detectors,           /* I: Total number of detectors in
                                         the band */
    double earth_sun_distance_sqr,    /* I: Earth-sun distance
                                         factor (squared) for the current
                                         collection.  Intended to come from
                                         JPL solar ephemeris data, but may
                                         come from CPF as a backup */
    double cos_theta,                 /* I: Cosine of solar zenith angle
                                         (1.0 for radiance-based gains) */

    double *relative_gains,           /* O: Radiance-based relative gain
                                         array for all detectors in the
                                         current band (NULL if calculating
                                         reflectance-based band absolute
                                         gains)*/
    double *relative_gains_stdev,     /* O: Radiance-based relative gain
                                         array for all detectors in the
                                         current band (NULL if calculating
                                         reflectance-based band absolute
                                         gains)
                                         std deviations (NULL if calculating
                                         reflectance-based band absolute
                                         gains)*/

    double *abs_band_gain,            /* O: Absolute radiance-based gain for
                                         the current band */
    double *abs_band_gain_stdev
);



                       /* "Utility" functions */
void free_db_sca_data_buffers
(
    IAS_DB_NULL_TYPE *detector_ids,   /* I: NULLed relative gain detector
                                          indices */
    double *sca_mean_data,            /* O: SCA data vector (means) */
    double *sca_stdev_data            /* O: SCA data vector (stdevs) */
);


void cleanup_application
(
    int wo_store_char_in_db,                    /* I: DB trend flag */
    struct ias_db_connection *db,               /* I: Open DB connection */
    RPS_SOLARCAL_REPORT_SUMMARY *rpt_summary,   /* I: Summary report data */
    double *relative_gains,                     /* I: Relative gain
                                                   arrays */
    double *relative_gains_stdev,
    double *hist_stats_mean,                    /* I: Histogram statistics
                                                   arrays */
    double *hist_stats_stdev,
    double *hist_stats_num_frames,
    IAS_CPF *cpf,                               /* I: CPF information */
    double *diffuser_radiances,                 /* I: CPF diffuser radiances
                                                   for the current band */
    double *diffuser_reflectances,              /* I: CPF diffuser
                                                   reflectances for the
                                                   current band */
    int *statistics_validity_flags    /* I: Data validity flags
                                                   for all detectors in the
                                                   current band */
);


int copy_cpf_diffuser_data
(
    const struct IAS_CPF_DIFFUSER_RADIANCE *diff_rad,
                                  /* I: CPF diffuser data */
    const int *inoperable_list,   /* I: Inoperable detector list for
                                     the current band/SCA */
    int band_index,               /* I: Current 0-based band index */
    int sca_index,                /* I: Current 0-based SCA index */
    int num_detectors_sca,        /* I: Number of detectors in the
                                     current band/SCA */
    const int data_select,        /* I: Flag indicating radiance
                                     data or diffuser data */
    const char *panel_type_string,/* I: Selector for diffuser data */
    double *diffuser_data_buffer  /* O: Diffuser data for the current
                                     band */
);


void get_valid_data_flags_from_cpf_inop_list
(
    const int *start_of_sca_inoperable_list,   /* I: Pointer to start of
                                            inoperable detector list
                                            for the current band/SCA */
    int num_inoperable_detectors,        /* I: Number of inoperable
                                            detectors in the current
                                            band/SCA */
    int sca_number,                      /* I: Current SCA number */
    int number_of_detectors_per_sca,     /* I: Number of detectors in the
                                            current SCA */
    int *statistics_validity_flags       /* O: Populated invalid data flags
                                            for the current band */
);


                       /* Database access */
struct ias_db_connection *connect_to_db();


int clear_solarcal_table
(
    struct ias_db_connection *db, /* I: Open database connection */
    char *rad_wo_common_id,       /* I: Radiometric work order common ID */
    const char *table_name        /* I: Name of solarcal table to clear */
);


int get_histogram_statistics
(
    struct ias_db_connection  *db,     /* I: Open database connection */
    char *rad_wo_common_id,            /* I: RADIOMETRIC_WORK_ORDER_COMMON
                                          primary key value */
    IAS_SPECTRAL_TYPE spectral_type,   /* I: Spectral type for current
                                          band */
    int band_number,                   /* I: Current 1-based band number */
    int sca_number,                    /* I: Current 1-based SCA number */
    int num_detectors_sca,             /* I: Number of detectors for the
                                          current band/SCA */
    const char *rps_proc_level,        /* I: Required RPS processing level
                                          for the histogram statistics input
                                          data */
    double *stats_mean_buffer,         /* O: Histogram statistics means
                                          for all detectors in the current
                                          band */
    double *stats_stdev_buffer,        /* O: Histogram statistics std
                                          deviations for all detectors in
                                          the current band */
    double *stats_num_frames_buffer    /* O: Histogram statistics number
                                          of frames for all detectors in
                                          the current band */
);


void populate_db_sca_data_buffers
(
    int start_detector_index,      /* I: Index corresponding to detector 1
                                      in the current SCA */
    int num_detectors_sca,         /* I: # of detectors in each SCA */
    double *band_mean_data,        /* I: Full data vector (means) for the
                                      current band*/
    double *band_stdev_data,       /* I: Full data vector (stdevs) for the
                                      current band */
    const int *statistics_validity_flags,
                                   /* I: Data validity flags for all
                                      detectors in in the current band */
    IAS_DB_NULL_TYPE *flagged_detector_ids, /* O: SCA detector IDs with
                                               NULLs */
    double *sca_mean_data,         /* O: Data vector (means) for the
                                      current band/SCA */
    double *sca_stdev_data         /* O: Data vector (stdevs) for the
                                      current band/SCA */
);


int write_to_detector_table
(
    struct ias_db_connection  *db,       /* I: Open database connection */
    char *rad_wo_common_id,              /* I: Radiometric WO ID key to
                                            SOLAR_RESPONSE table */
    int band_number,                     /* I: Current (1-based) band # */
    int sca_number,                      /* I: Current (1-based) SCA # */
    int num_sca_detectors,               /* I: # of detectors for each
                                            SCA in the current band */
    int *detector_ids,                   /* I: Detector IDs for the current
                                            band/SCA (1-based) */
    IAS_DB_NULL_TYPE *flagged_detector_ids, /* I: Flag array for detectors
                                               with no valid data */
    double *sca_relgain_data,            /* I: Derived relative gains
                                            for the current band/SCA */
    double *sca_relgain_stdev_data       /* I: Derived relative gain
                                            stdevs for the current
                                            band/SCA */
);


int write_to_band_table
(
    struct ias_db_connection *db,        /* I: Open DB connection */
    char *rad_wo_common_id,              /* I: Radiometric WO ID key to
                                            SOLAR_RESPONSE table */
    RPS_SOLARCAL_BAND_GAINS band_gains   /* I: Band-level summary
                                            histogram statistics and
                                            diffuser radiances/
                                            reflectances */
);


int write_to_main_table
(
    struct ias_db_connection *db,          /* I: Open DB connection */
    char *rad_wo_common_id,                /* I: Radiometric WO ID key to
                                              SOLAR_RESPONSE table */
    RPS_SOLARCAL_REPORT_HEADER header,     /* I: Validation information */
    RPS_SOLARCAL_SOLAR_ANGLES angles       /* I: Solar angle information */
);



                    /* Summary report data generation */
int initialize_solarcal_report_data
(
    RPS_SOLARCAL_WO_PARAMS wo_params,            /* I: WO parameters */
    RPS_SOLARCAL_OMF_PARAMS omf_params,          /* I: OMF parameters */
    RPS_SOLARCAL_SOLAR_ANGLES angles,            /* I: Solar angles */
    RPS_SOLARCAL_REPORT_SUMMARY *summary         /* O: Summary report
                                                    data structure */
);


void initialize_solarcal_summary_band_data
(
    int band_index,                                /* I: Current 0-based
                                                      band index */
    RPS_SOLARCAL_BAND_GAINS current_band_gains,    /* I: Band gain data for
                                                      current band */
    RPS_SOLARCAL_REPORT_SUMMARY *summary           /* I/O: Populated
                                                      summary report data
                                                      structure */
);



                   /* Summary report creation */
int write_header_to_summary_report
(
    RPS_SOLARCAL_REPORT_HEADER report_header,   /* I: IAS/RPS report header
                                                   information */
    FILE *rpt_fptr                              /* I: Report file handle */
);


int write_angle_statistics_to_summary_report
(
    RPS_SOLARCAL_SOLAR_ANGLES angles,            /* I: Summary angle
                                                    statistics */
    FILE *rpt_fptr                               /* I: Report file handle */
);


int write_solarcal_report_data
(
    int num_bands,                        /* I: Number of bands to
                                             report */
    RPS_SOLARCAL_REPORT_SUMMARY summary,  /* I: Summary report data */
    FILE *rpt_fptr                        /* I/O: Report file handle */
);



                  /* Dump file creation */
int write_header_to_dump_file
(
    RPS_SOLARCAL_REPORT_HEADER report_header,   /* I: IAS/RPS report header
                                                   information */
    FILE *dump_fptr                             /* I: Dump file handle */
);


int write_angle_statistics_to_dump_file
(
    RPS_SOLARCAL_SOLAR_ANGLES angles,     /* I: Angle statistics */
    FILE *dump_fptr                       /* I: Dump file handle */
);


int write_relative_gains_to_dump_file
(
    double *relative_gains,           /* I: Relative gains for the current
                                         band */
    double *relative_gains_stdev,     /* I: Relative gain stdevs for the
                                         current band */
    const int *data_validity_flags,
                                      /* I: Data validity flags for all
                                         detectors in the current band
                                         (0=valid data, nonzero=invalid
                                         data) */
    int band_number,                  /* I: Current 1-based band number */
    int num_scas,                     /* I: Number of SCAs in the current
                                         band */
    int num_band_detectors,           /* I: Total number of detectors
                                         in the band */
    FILE *dump_fptr                   /* I: Dump file handle */
);


int write_band_gains_to_dump_file
(
    RPS_SOLARCAL_BAND_GAINS band_gains,     /* I: Summary gain statistics
                                               for the current band */
    FILE *dump_fptr                         /* I: Dump file handle */
);



#endif

