/**************************************************************************
 NAME:          solarcal_utility

 PURPOSE:       Module implementing "utility" functions used by the main
                solarcal_oli application.

***************************************************************************/
#include "solarcal_oli.h"


/*-----------------------------------------------------------------------
 NAME:     free_db_sca_data_buffers

 PURPOSE:  Frees SCA-level metrics detector ID and data buffers

 RETURNS:  Nothing.

 NOTES:    This function is only called when writing characterization
           results to the database
-------------------------------------------------------------------------*/
void free_db_sca_data_buffers
(
    IAS_DB_NULL_TYPE *detector_ids,   /* I: NULLed relative gain detector
                                         indices */
    double *sca_mean_data,            /* I: SCA data vector (means) */
    double *sca_stdev_data            /* I: SCA data vector (stdevs) */
)
{
    free(detector_ids);
    free(sca_mean_data);
    free(sca_stdev_data);
}




/*------------------------------------------------------------------------
 NAME:        cleanup_application

 PURPOSE:     Calls all system/application functions freeing allocated
              resources, depending on the current exit status. On exit due
              to a fatal error, all database transactions are rolled back. 
              On exit due to successful completion of processing, all
              database transactions are committed.  For either exit, the CPF
              and IAS_REPORT_HEADER memory blocks are freed,
              open summary report and/or dump files are closed, the
              database connection is closed, and allocated working
              data buffers are freed.

 RETURNS:     Nothing

 NOTES:       Called only when there's an error during band processing
--------------------------------------------------------------------------*/
void cleanup_application
(
    int wo_store_char_in_db,                    /* I: DB trend flag */
    struct ias_db_connection *db,               /* I: Open DB connection */
    RPS_SOLARCAL_REPORT_SUMMARY *rpt_summary,   /* I: Summary report data */
    double *relative_gains,                     /* I: Relative gain
                                                   arrays */
    double *relative_gains_stdev,
    double *hist_stats_mean,                    /* I: Histogram statistics
                                                   arrays */
    double *hist_stats_stdev,
    double *hist_stats_num_frames,
    IAS_CPF *cpf,                               /* I: CPF information */
    double *diffuser_radiances,                 /* I: CPF diffuser radiances
                                                   and reflectances for
                                                   the current band */
    double *diffuser_reflectances,
    int *statistics_invalid_flags               /* I: Data invalid flags
                                                   for all detectors in the
                                                   current band */
)
{
    /* If trending was specified, rollback all previous writes */
    if (wo_store_char_in_db)
    {
        IAS_LOG_DEBUG("Reversing DB changes...");
        if (ias_db_rollback_transaction(db))
        {
            IAS_LOG_ERROR("Reversing DB changes");
        }
    }

    /* Shut down the database connection */
    ias_db_close_connection(db);
    ias_db_close_database_lib();

    /* Free the band average gains data block in the report summary
       header */
    free(rpt_summary->band_gains);

    /* Free up the memory block used by the common report header */
    IAS_LOG_DEBUG("Freeing common IAS_REPORT_HEADER data block...");
    ias_misc_free_report_header(rpt_summary->header.common_header);

    /* CPF cleanup */
    IAS_LOG_DEBUG("Freeing CPF resources...");
    ias_cpf_free(cpf);

    /* Free the relative gain and std deviation buffers for the current
       band */
    IAS_LOG_DEBUG("Freeing working data buffers...");
    free(relative_gains);
    free(relative_gains_stdev);
    free(hist_stats_mean);
    free(hist_stats_stdev);
    free(hist_stats_num_frames);
    free(diffuser_radiances);
    free(diffuser_reflectances);

    /* Free the data invalid flag data buffer */
    free(statistics_invalid_flags);
}




/*-------------------------------------------------------------------------
 NAME:                copy_cpf_diffuser_data

 PURPOSE:      Copies the specified set of diffuser radiances or
               reflectances (determined by the diffuser type--"primary"
               or "pristine") from the CPF to an array for the current band

 RETURNS:      Integer status code of SUCCESS or ERROR

--------------------------------------------------------------------------*/
int copy_cpf_diffuser_data
(
    const struct IAS_CPF_DIFFUSER_RADIANCE *diff_rad,
                                  /* I: CPF diffuser data */
    const int *inoperable_list,   /* I: Inoperable detector list for
                                     the current band/SCA */
    int band_index,               /* I: Current 0-based band index */
    int sca_index,                /* I: Current 0-based SCA index */
    int num_detectors_sca,        /* I: Number of detectors in the
                                     current band/SCA */
    const int data_select,        /* I: Flag indicating radiance
                                     data or diffuser data */
    const char *panel_type,       /* I: Selector for diffuser data */
    double *diffuser_data_buffer  /* O: Diffuser data for the current
                                     band */
)
{
    /* Calculate starting detector offset for the current SCA */
    int start_detector_index = sca_index * num_detectors_sca;

    if (strcmp(panel_type, IAS_RPS_OLI_DIFFUSER_WORKING) == 0)
    {
        if (data_select == GET_DIFFUSER_RADIANCES)
        {
            memcpy(&diffuser_data_buffer[start_detector_index],
                &(diff_rad->diff_rad_prim[band_index][sca_index][0]),
                num_detectors_sca * sizeof(double));
        }
        else          /* Get the diffuser reflectances */
        {
            memcpy(&diffuser_data_buffer[start_detector_index],
                &(diff_rad->diff_bidir_refl_prim[band_index][sca_index][0]),
                num_detectors_sca * sizeof(double));
        }
    }
    else if (strcmp(panel_type, IAS_RPS_OLI_DIFFUSER_PRISTINE) == 0)
    {
        if (data_select == GET_DIFFUSER_RADIANCES)
        {
            memcpy(&diffuser_data_buffer[start_detector_index],
                &(diff_rad->diff_rad_pris[band_index][sca_index][0]),
                num_detectors_sca * sizeof(double));
        }
        else
        {
            memcpy(&diffuser_data_buffer[start_detector_index],
                &(diff_rad->diff_bidir_refl_pris[band_index][sca_index][0]),
                num_detectors_sca * sizeof(double));
        }
    }
    else
    {
        IAS_LOG_ERROR("Invalid diffuser panel type %s", panel_type);
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}



/*-----------------------------------------------------------------------
 NAME:    get_valid_data_flags_from_cpf_inop_list

 PURPOSE: Walks the inoperable detector list of the current band/SCA and
          flags the corresponding entries into a list over the entire
          band

 RETURNS: Integer status of SUCCESS or ERROR
-------------------------------------------------------------------------*/
void get_valid_data_flags_from_cpf_inop_list
(
    const int *start_of_sca_inoperable_list,   /* I: Pointer to start of
                                            inoperable detector list
                                            for the current band/SCA */
    int num_inoperable_detectors,        /* I: Number of inoperable
                                            detectors in the current
                                            band/SCA */
    int sca_number,                      /* I: Current SCA number */
    int number_of_detectors_per_sca,     /* I: Number of detectors in the
                                            current SCA */
    int *statistics_invalid_flags        /* O: Populated valid data flag
                                            buffer for the current band */
)
{
    int inop_detector_index;
    const int *inop_list_ptr = start_of_sca_inoperable_list;
    int band_detector_start_index =
        (sca_number - 1) * number_of_detectors_per_sca;

    for (inop_detector_index = 0;
         inop_detector_index < num_inoperable_detectors;
         inop_detector_index++)
    {
        int sca_detector_number = *inop_list_ptr;
        int band_detector_index = band_detector_start_index
            + (sca_detector_number - 1);
        statistics_invalid_flags[band_detector_index] =
            IAS_RPS_INOP_DETECTOR;
        inop_list_ptr++;
    }
}
