/***************************************************************************
 NAME:                   solarcal_oli_calculate_angles_test

 PURPOSE:   CGreen unit test of the calculate_solar_angles() routine
            (and its supporting routine calculate_sample_angles())
            to make sure quaternion-based angles are calculated correctly

 RETURNS:   CGreen status

****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include "ias_ancillary_io.h"
#include "ias_logging.h"
#include "ias_math.h"
#include "ias_structures.h"
#include "cgreen.h"
#include "../solarcal_oli.h"
#include "solarcal_oli_tests.h"

/* #define DEBUG_PRINT */

Ensure test_calculate_sample_angles()
{
    char full_test_ancillary_data_filename[PATH_MAX];
    char full_expected_angle_filename[PATH_MAX];
    char test_ancillary_data_filename[100] =
        "quaternion_setB.txt";
    char expected_angle_filename[100] = "expected_angles.txt";
    char *test_data_path = NULL;
    FILE *quaternion_data_fptr = NULL;
    FILE *expected_angle_fptr = NULL;
    IAS_ANC_ATTITUDE_DATA *attitude_data = NULL;
    IAS_VECTOR solar_ephemeris;

    /* Arrays to contain solar angle samples corresponding to each
       quaternion record. */
    double *view_solar_zenith_samples      = NULL;
    double *view_solar_azimuth_samples     = NULL;
    double *incident_solar_zenith_samples  = NULL;
    double *expected_angle_samples[3]      = {NULL};

    int record_index;
    int num_records;
    int num_incident_zenith_diffs = 0;
    int num_view_zenith_diffs = 0;
    int num_view_azimuth_diffs = 0;
    int status;

    size_t record_buffer_size;


    printf("Calculating solar angles from test data...\n");

    /* Get the test_data_path to the test data */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path == NULL)
    {
        assert_true_with_message(FALSE, "Cannot find test data path");
        return;
    }

    /* Build the name of the test ancillary data file. */
    status = snprintf(full_test_ancillary_data_filename,
        sizeof(full_test_ancillary_data_filename),
        "%s/ias/rps/det_response_diffuser/%s",
        test_data_path, test_ancillary_data_filename);
    if ((status < 0)
        || (status >= sizeof(full_test_ancillary_data_filename)))
    {
        assert_true_with_message(FALSE, "Error creating test ancillary "
            "data file name");
        return;
    }

    /* Build the name of the expected angles file. */
    status = snprintf(full_expected_angle_filename,
        sizeof(full_expected_angle_filename),
        "%s/ias/rps/det_response_diffuser/%s",
        test_data_path, expected_angle_filename);
    if ((status < 0) || (status >= sizeof(full_expected_angle_filename)))
    {
        assert_true_with_message(FALSE, "Error creating expected angle "
            "file name");
        return;
    }

    /* Open the quaternion data file. */


    quaternion_data_fptr = fopen(full_test_ancillary_data_filename, "r");
    if (quaternion_data_fptr == NULL)
    {
        assert_true_with_message(FALSE, "Cannot open %s for reading",
            full_test_ancillary_data_filename);
        return;
    }

    /* Read in the values for the instantaneous solar position vector and
       number of records. */
    if (fscanf(quaternion_data_fptr, "%lf%lf%lf\n", &(solar_ephemeris.x),
        &(solar_ephemeris.y), &(solar_ephemeris.z)) != 3)
    {
        assert_true_with_message(FALSE, "Error reading instantaneous solar "
            "position vector components");
        fclose(quaternion_data_fptr);
        return;
    }
    if (fscanf(quaternion_data_fptr, "%d\n", &num_records) != 1)
    {
        assert_true_with_message(FALSE, "Error reading number of "
            "quaternion records");
        fclose(quaternion_data_fptr);
        return;
    }

    /* Allocate a memory block for the quaternion data. */
    attitude_data = ias_ancillary_allocate_attitude(num_records);
    if (attitude_data == NULL)
    {
        assert_true_with_message(FALSE, "Cannot allocate attitude data "
            "structure");
        fclose(quaternion_data_fptr);
        return;
    }

    /* Read in the quaternion data.  The 'scalar' component is read first,
       followed by the 'x' component, the 'y' component, and finally the
       'z' component.  */
    for (record_index = 0; record_index < num_records; record_index++)
    {
        if (fscanf(quaternion_data_fptr, "%lf%lf%lf%lf\n",
            &(attitude_data->records[record_index].eci_quaternion[3]),
            &(attitude_data->records[record_index].eci_quaternion[0]),
            &(attitude_data->records[record_index].eci_quaternion[1]),
            &(attitude_data->records[record_index].eci_quaternion[2])) != 4)
        {
            assert_true_with_message(FALSE, "Error reading quaternion "
                "record %d", (record_index + 1));
            fclose(quaternion_data_fptr);
            ias_ancillary_free_attitude(attitude_data);
            return;
        }
    }

    /* Done with the quaternion data file--close it. */
    fclose(quaternion_data_fptr);

    /* Determine the size for all of the angle samples buffers. */
    record_buffer_size = num_records * sizeof(double);

    /* Allocate buffers for the expected angles. */
    expected_angle_samples[0] = malloc(record_buffer_size);
    expected_angle_samples[1] = malloc(record_buffer_size);
    expected_angle_samples[2] = malloc(record_buffer_size);
    if ((expected_angle_samples[0] == NULL)
        || (expected_angle_samples[1] == NULL)
        || (expected_angle_samples[2] == NULL))
    {
        assert_true_with_message(FALSE, "Error allocating buffers for "
            "expected angle samples");
        free(expected_angle_samples[0]);
        free(expected_angle_samples[1]);
        free(expected_angle_samples[2]);
        ias_ancillary_free_attitude(attitude_data);
        return;
    }

    /* Open the expected angle file. */
    expected_angle_fptr = fopen(full_expected_angle_filename, "r");
    if (expected_angle_fptr == NULL)
    {
        assert_true_with_message(FALSE, "Cannot open angle file '%s' for "
            "reading", full_expected_angle_filename);
        ias_ancillary_free_attitude(attitude_data);
        free(expected_angle_samples[0]);
        free(expected_angle_samples[1]);
        free(expected_angle_samples[2]);
        return;
    }

    /* Read in the expected angles for each record.  The first angle is
       the incident solar zenith angle, the second is the view solar
       zenith angle, and the third is the view solar azimuth angle.
       These angles were calculated from the prototype data using the
       prototype code.  */
    for (record_index = 0; record_index < num_records; record_index++)
    {
        if (fscanf(expected_angle_fptr, "%lf%lf%lf\n",
            &expected_angle_samples[0][record_index],
            &expected_angle_samples[1][record_index],
            &expected_angle_samples[2][record_index]) != 3)
        {
            assert_true_with_message(FALSE, "Error reading expected angle "
                "record %d", (record_index + 1));
            fclose(expected_angle_fptr);
            ias_ancillary_free_attitude(attitude_data);
            free(expected_angle_samples[0]);
            free(expected_angle_samples[1]);
            free(expected_angle_samples[2]);
            return;
        }
    }
    fclose(expected_angle_fptr);

    /* Allocate buffers for the record angles. */
    incident_solar_zenith_samples = malloc(record_buffer_size);
    view_solar_zenith_samples = malloc(record_buffer_size);
    view_solar_azimuth_samples = malloc(record_buffer_size);

    if ((incident_solar_zenith_samples == NULL)
        || (view_solar_zenith_samples == NULL)
        || (view_solar_azimuth_samples == NULL))
    {
        assert_true_with_message(FALSE, "Error allocating sample angle data "
            "buffers");
        free(incident_solar_zenith_samples);
        free(view_solar_zenith_samples);
        free(view_solar_azimuth_samples);
        ias_ancillary_free_attitude(attitude_data);
        free(expected_angle_samples[0]);
        free(expected_angle_samples[1]);
        free(expected_angle_samples[2]);
        return;
    }

    /* Calculate the angles for each quaternion record */
    status = calculate_sample_angles(attitude_data, 0, num_records,
        &solar_ephemeris, incident_solar_zenith_samples,
        view_solar_zenith_samples, view_solar_azimuth_samples);
    ias_ancillary_free_attitude(attitude_data);
    assert_equal_with_message(status, SUCCESS, "Error calculating solar "
        "angle samples");
    if (status != SUCCESS)
    {
        free(incident_solar_zenith_samples);
        free(view_solar_zenith_samples);
        free(view_solar_azimuth_samples);
        free(expected_angle_samples[0]);
        free(expected_angle_samples[1]);
        free(expected_angle_samples[2]);
        return;
    }

    /* Compare the calculated angles with the expected prototype results.
       Since there's a lot of records here, just tally up differences
       greater than a threshold value and test for zero number of
       differences.  */
    for (record_index = 0; record_index < num_records; record_index++)
    {
#ifdef DEBUG_PRINT
        if ((record_index % 10) == 0)
        {
            printf("%lf\t%lf\t%lf\n",
            incident_solar_zenith_samples[record_index],
            view_solar_zenith_samples[record_index],
            view_solar_azimuth_samples[record_index]);
        }
#endif
        double incident_zenith_diff =
            fabs(incident_solar_zenith_samples[record_index]
            - expected_angle_samples[0][record_index]);

        double view_zenith_diff =
            fabs(view_solar_zenith_samples[record_index]
            - expected_angle_samples[1][record_index]);

        double view_azimuth_diff =
            fabs(view_solar_azimuth_samples[record_index]
            - expected_angle_samples[2][record_index]);

        if (incident_zenith_diff > 1e-5)
            num_incident_zenith_diffs++;
        if (view_zenith_diff > 1e-5)
            num_view_zenith_diffs++;
        if (view_azimuth_diff > 1e-5)
            num_view_azimuth_diffs++;
    }
    assert_equal_with_message(num_incident_zenith_diffs, 0, "%d "
        "incident zenith angle differences were > 1e-5");
    assert_equal_with_message(num_view_zenith_diffs, 0, "%d view "
        "zenith angle differences were > 1e-5");
    assert_equal_with_message(num_view_azimuth_diffs, 0, "%d view "
        "azimuth angle differences were > 1e-5");

    /* Clean up */
    free(expected_angle_samples[0]);
    free(expected_angle_samples[1]);
    free(expected_angle_samples[2]);
    free(incident_solar_zenith_samples);
    free(view_solar_zenith_samples);
    free(view_solar_azimuth_samples);
}



void solarcal_oli_calculate_angles_test(TestSuite *suite)
{
    add_test(suite, test_calculate_sample_angles);
}

