/***************************************************************************
 NAME:                   solarcal_oli_calculate_gains_test

 PURPOSE:   CGreen unit test of the calculate_gains() routine
            to make sure both radiance-based and reflectance-based
            gains are calculated correctly

 RETURNS:   CGreen status

****************************************************************************/
#include <math.h>
#include <string.h>

#include "cgreen.h"

#include "ias_math.h"

#include "../solarcal_oli.h"

#include "solarcal_oli_tests.h"

/*#define DEBUG_PRINT */

static int initialize_radiance_statistics_data
(
    const char *rad_data_filename,        /* I: Name of file with histogram
                                             stats and simulated radiances
                                             used in the prototype
                                             implementation */
    int        num_band_detectors,        /* I: # detectors in the band */
    double     *hist_stats_mean,          /* O: Populated histogram
                                             statistics arrays for the
                                             band */
    double     *hist_stats_stdev,
    double     *diffuser_stats,           /* O: Radiance or reflectance
                                             statistics array for the
                                             band */
    int *statistics_invalid_flags         /* O: Data invalid flags for
                                             all detectors in the current
                                             band/SCA */
)
{
    FILE  *data_fptr = NULL;

    int   detector_index = 0;
    int   dummy_int = 0;

    /* Open the file for reading */
    data_fptr = fopen(rad_data_filename, "rt");
    if (data_fptr == NULL)
    {
        printf("Cannot open simulated histogram statistics and radiance "
               "statistics data file %s\n", rad_data_filename);
        return ERROR;
    }
    for (detector_index = 0;
         detector_index < num_band_detectors;
         detector_index++)
    {
        fscanf(data_fptr, "%d%lf%lf%lf", &dummy_int,
               &hist_stats_mean[detector_index],
               &hist_stats_stdev[detector_index],
               &diffuser_stats[detector_index]);
        if ((hist_stats_mean[detector_index] <= 0)
            || (diffuser_stats[detector_index] <= 0))
        {
            statistics_invalid_flags[detector_index] = 1;
        }
    }
    fclose(data_fptr);

    /* Finished */
    return SUCCESS;
}



static void free_buffers(double *hist_stats_mean, double *hist_stats_stdev,
                         double *diffuser_stats, int
                         *statistics_invalid_flags, double *relgains,
                         double *relgain_stdevs)
{
    free(hist_stats_mean);
    free(hist_stats_stdev);
    free(diffuser_stats);
    free(statistics_invalid_flags);
    free(relgains);
    free(relgain_stdevs);
}



Ensure test_calculate_gains()
{
    const IAS_BAND_ATTRIBUTES *ba = NULL;

    const char rad_data_basename[MAX_FILENAME_SIZE]
        = "EK_Prototype_DNRads.txt";
    const char refl_data_basename[MAX_FILENAME_SIZE]
        = "EK_Prototype_DNRefs.txt";


    char *test_data_path = NULL;
    char full_rad_data_filename[PATH_MAX];

    double *hist_stats_mean = NULL;
    double *hist_stats_stdev = NULL;
    double *relgains = NULL;
    double *relgain_stdevs = NULL;
    double *diffuser_stats = NULL;
    double earth_sun_distance = 1.01553;  /* SOCAL prototype distance (AU) */
    double earth_sun_distance_sqr;
    double solar_zenith    = 72.4321;     /* SOCAL prototype solar zenith
                                             angle */
    double cos_solar_zenith;
    double band_radiance_gain = 0.0;
    double band_radiance_gain_stdev = 0.0;
    double test_radiance_gain = 0.9304;      /* SOCAL prototype radiance
                                                gain */
    double band_reflectance_gain = 0.0;
    double band_reflectance_gain_stdev = 0.0;
    double test_reflectance_gain = 1832.146; /* SOCAL prototype reflectance
                                                gain */

    int *statistics_invalid_flags = NULL;


    int num_band_detectors = 0;
    int status = 0;
    int size;


    /* Find the parent test data directory, then construct the name
       of the input test data file */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path == NULL)
    {
        printf("Cannot find test data path\n");
        return;
    }
    size = snprintf(full_rad_data_filename, sizeof(full_rad_data_filename),
            "%s/ias/rps/det_response_diffuser/%s", test_data_path,
            rad_data_basename);
    if ((size < 0) || (size >= sizeof(full_rad_data_filename)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
           "it needs %d bytes", size);
        return;
    }
    /* Get band 1 attributes */
    ba = ias_sat_attr_get_band_attributes(1);
    assert_true(ba != NULL);
    if (ba == NULL)
    {
        printf("Cannot get band 1 attributes\n");
        return;
    }

    /* Allocate buffer space */
    num_band_detectors = ba->scas * ba->detectors_per_sca;

    hist_stats_mean  = calloc(num_band_detectors, sizeof(*hist_stats_mean));
    hist_stats_stdev = calloc(num_band_detectors, sizeof(*hist_stats_stdev));
    diffuser_stats   = calloc(num_band_detectors, sizeof(*diffuser_stats));
    relgains         = calloc(num_band_detectors, sizeof(*relgains));
    relgain_stdevs   = calloc(num_band_detectors, sizeof(*relgain_stdevs));
    statistics_invalid_flags = calloc(num_band_detectors,
        sizeof(int));

    assert_true(((hist_stats_mean != NULL) && (hist_stats_stdev != NULL)
                && (diffuser_stats != NULL) && (relgains != NULL)
                && (relgain_stdevs != NULL)
                && (statistics_invalid_flags != NULL)));
    if ((hist_stats_mean == NULL) || (hist_stats_stdev == NULL)
        || (diffuser_stats == NULL) || (relgains == NULL)
        || (relgain_stdevs == NULL)
        || (statistics_invalid_flags == NULL))
    {
        printf("One or more data buffers could not be allocated\n");
        free_buffers(hist_stats_mean, hist_stats_stdev, diffuser_stats,
                     statistics_invalid_flags, relgains, relgain_stdevs);
        return;
    }

    /* Pretend we've read histogram statistics and copied diffuser 
       radiance data from the CPF.  There are assumed to be no
       inoperable detectors */
    status = initialize_radiance_statistics_data(full_rad_data_filename,
                                                 num_band_detectors,
                                                 hist_stats_mean,
                                                 hist_stats_stdev,
                                                 diffuser_stats,
                                                 statistics_invalid_flags);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Could not retrieve histogram statistics and/or diffuser "
               "radiance statistics from file\n");
        free_buffers(hist_stats_mean, hist_stats_stdev, diffuser_stats,
                     statistics_invalid_flags, relgains, relgain_stdevs);
        return;
    }

    earth_sun_distance_sqr = earth_sun_distance * earth_sun_distance;

    /* Calculate gains, starting with the radiance-based gains.  Use 1.0
       for the cos_solar_zenith parameter */
    status = calculate_gains(hist_stats_mean, hist_stats_stdev,
                             diffuser_stats, statistics_invalid_flags,
                             num_band_detectors,
                             earth_sun_distance_sqr, 1.0, relgains,
                             relgain_stdevs, &band_radiance_gain,
                             &band_radiance_gain_stdev);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot calculate relative gains and/or radiance-based "
               "gains for the current band\n");
        free_buffers(hist_stats_mean, hist_stats_stdev, diffuser_stats,
                     statistics_invalid_flags, relgains, relgain_stdevs);
        return;
    }

    /* Now calculate the reflectance-based gains.  Use NULL for the
       relgains and relgain_stdevs, since we're not interested in these.
       Use the full cosine of the solar zenith angle. */
    cos_solar_zenith = cos(ias_math_get_radians_per_degree() * solar_zenith);

    size = snprintf(full_rad_data_filename, sizeof(full_rad_data_filename),
            "%s/ias/rps/det_response_diffuser/%s", test_data_path,
            refl_data_basename);
    if ((size < 0) || (size >= sizeof(full_rad_data_filename)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
           "it needs %d bytes", size);
        return;
    }

    memset(statistics_invalid_flags, 0, num_band_detectors * sizeof(int));
    status = initialize_radiance_statistics_data(full_rad_data_filename,
                                                 num_band_detectors,
                                                 hist_stats_mean,
                                                 hist_stats_stdev,
                                                 diffuser_stats,
                                                 statistics_invalid_flags);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Could not retrieve histogram statistics and/or diffuser "
               "radiance statistics from file\n");
        free_buffers(hist_stats_mean, hist_stats_stdev, diffuser_stats,
                     statistics_invalid_flags, relgains, relgain_stdevs);
        return;
    }

    status = calculate_gains(hist_stats_mean, hist_stats_stdev,
                             diffuser_stats, statistics_invalid_flags,
                             num_band_detectors, earth_sun_distance_sqr,
                             cos_solar_zenith, NULL, NULL,
                             &band_reflectance_gain,
                             &band_reflectance_gain_stdev);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot calculate reflectance-based gains for the current "
               "band\n");
        free_buffers(hist_stats_mean, hist_stats_stdev, diffuser_stats,
                     statistics_invalid_flags, relgains, relgain_stdevs);
        return;
    }

    /* Test the gains.  The standard deviation can't be checked--
       the RPS library function ias_rps_calculate_filtered_stdev checks
       for an lower bound of DBL_EPSILON on the calculated variance; if it's
       less than that (or the variance is negative), the standard deviation
       is automatically set to 0 (after logging a warning).

       I would also have used the assert_double_equal() routine
       to do the comparison.  Doing that causes the test to fail,
       primarily because the precision of the prototype results is lower.
       So I figured it would be better to test the differences between
       the implementation and the prototype, especially with the large
       value of the prototype reflectance gain... */
    assert_true(fabs(band_radiance_gain - test_radiance_gain) <= 1e-4);
    assert_true(fabs(band_reflectance_gain - test_reflectance_gain) <= 1e-2);

#ifdef DEBUG_PRINT
    double test_radiance_gain_stdev = 8.89e-15;    /* SOCAL prototype
                                                      radiance gain stdev */

    double test_reflectance_gain_stdev = 1.39e-11; /* SOCAL prototype
                                                      reflectance gain
                                                      stdev */

    printf("Radiance-based gain:     %13.10lf (%13.10lf)\n",
           band_radiance_gain, test_radiance_gain);
    printf("Radiance-based gain std dev:     %e (%e)\n",
           band_radiance_gain_stdev, test_radiance_gain_stdev);
    printf("Reflectance-based gain:  %13.10lf (%13.10lf)\n",
           band_reflectance_gain, test_reflectance_gain);
    printf("Reflectance-based gain std dev:  %e (%e)\n",
           band_reflectance_gain_stdev, test_reflectance_gain_stdev);

#endif

    /* Clean up */
    free_buffers(hist_stats_mean, hist_stats_stdev, diffuser_stats,
                 statistics_invalid_flags, relgains, relgain_stdevs);
}



void solarcal_oli_calculate_gains_test(TestSuite *suite)
{
    add_test(suite, test_calculate_gains);
}
