/**********************************************************************
 NAME:         solarcal_oli_dbaccess_test

 PURPOSE:      Use CGREEN to perform unit testing of the various
               component functions making up OLI detector response
               characterization database access

 NOTES:        Even though at the function level they're declared as
               straight pointers-to struct ias_db_connection, they have to
               be passed in as base addresses in order to "save" the
               valid address returned when the database connection is
               established.

***********************************************************************/
#include <math.h>
#include <string.h>

#include "ias_const.h"
#include "ias_db.h"
#include "ias_db_insert.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_radiometry.h"

#include "cgreen.h"

#include "../solarcal_oli.h"

#include "solarcal_oli_tests.h"


/*#define DEBUG_PRINT */

#define COMMAND_STRLEN 300


static char work_order_id[IAS_WORKORDER_ID_SIZE] = "DRDT1002";


static void free_data_buffers(double *hist_stats_mean_buffer,
    double *hist_stats_stdev_buffer, double *hist_stats_frame_buffer,
    int *statistics_invalid_flags, double **hs_dn,
    double **hs_stdev_dn)
{
    free(hist_stats_mean_buffer);
    free(hist_stats_stdev_buffer);
    free(hist_stats_frame_buffer);
    free(statistics_invalid_flags);

    ias_misc_free_2d_array((void **)hs_dn);
    ias_misc_free_2d_array((void **)hs_stdev_dn);
}


Ensure test_get_histogram_statistics()
{
    const IAS_BAND_ATTRIBUTES *ba = NULL;

    char system_command[COMMAND_STRLEN] = "";

    int sca_index;
    int detector_index;
    char rad_wo_common_id[IAS_CHAR_ID_SIZE];  
    int start_detector_index;
    int status;
    int num_band_detectors;
    int offset;


    double num_frames_averaged = 0.0;

    double *hist_stats_mean_buffer = NULL;
    double *hist_stats_stdev_buffer = NULL;
    double *hist_stats_frames_buffer = NULL;

    int *statistics_invalid_flags = NULL;

    /* 2d arrays for database histogram statistics means and std
       deviations */
    double **hs_dn = NULL;
    double **hs_stdev_dn = NULL;

    struct ias_db_connection *open_db = NULL;

    char *path = NULL;
    int size;


    printf("Retrieving histogram statistics from database...\n");

    /* Get the attributes for band 1 */
    ba = ias_sat_attr_get_band_attributes(1);
    if (ba == NULL)
    {
        assert_true_with_message(FALSE, "Cannot obtain band 1 attributes");
        return;
    }

    /* Get the histogram statistics information from the text file */
    path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (path == NULL)
    {
        assert_true_with_message(FALSE, "Cannot find test data directory");
        return;
    }

    /* Allocate 2D arrays for the histogram statistics means data */
    hs_dn = (double **)ias_misc_allocate_2d_array(ba->scas,
        ba->detectors_per_sca, sizeof(double));
    hs_stdev_dn = (double **)ias_misc_allocate_2d_array(ba->scas,
        ba->detectors_per_sca, sizeof(double));

    if ((hs_dn == (double **)NULL) || (hs_stdev_dn == (double **)NULL))
    {
        assert_true_with_message(FALSE, "Cannot allocate 2D arrays to hold "
            "histogram statistics information");
        ias_misc_free_2d_array((void **)hs_dn);
        ias_misc_free_2d_array((void **)hs_stdev_dn);
        return;
    }

    /* Get the histogram statistics for the current band (which is just 1)
       and SCA from the database */
    num_band_detectors = ba->scas * ba->detectors_per_sca;
    hist_stats_mean_buffer  = calloc(num_band_detectors, 
                                     sizeof(*hist_stats_mean_buffer));
    hist_stats_stdev_buffer = calloc(num_band_detectors,
                                     sizeof(*hist_stats_stdev_buffer));
    hist_stats_frames_buffer = calloc(num_band_detectors,
                                      sizeof(*hist_stats_frames_buffer));

    if ((hist_stats_mean_buffer == NULL)
        || (hist_stats_stdev_buffer == NULL)
        || (hist_stats_frames_buffer == NULL))
    {
        assert_true_with_message(FALSE, "Cannot allocate mean and/or "
            "stdev DB statistics buffers");
        free_data_buffers(hist_stats_mean_buffer, hist_stats_stdev_buffer,
                          hist_stats_frames_buffer, statistics_invalid_flags,
                          hs_dn, hs_stdev_dn);
        return;
    }

    /* Allocate the invalid flag data buffer */
    statistics_invalid_flags = calloc(num_band_detectors, sizeof(int));
    if (statistics_invalid_flags == NULL)
    {
        assert_true_with_message(FALSE, "Cannot allocate statistics "
            "invalid flag data buffer");
        free_data_buffers(hist_stats_mean_buffer, hist_stats_stdev_buffer,
            hist_stats_frames_buffer, statistics_invalid_flags,
            hs_dn, hs_stdev_dn);
        return;
    }

    /* Prep the database */
    size = snprintf(system_command, sizeof(system_command),
               "./check_or_insert_woid.csh %s", work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
           "it needs %d bytes", size);
        free_data_buffers(hist_stats_mean_buffer, hist_stats_stdev_buffer,
            hist_stats_frames_buffer, statistics_invalid_flags, hs_dn,
            hs_stdev_dn);
        return;
    }
    status = system(system_command);
    if (status != 0)
    {
        assert_true_with_message(FALSE, "Error prepping database -- cannot "
            "insert histogram statistics");
        free_data_buffers(hist_stats_mean_buffer, hist_stats_stdev_buffer,
            hist_stats_frames_buffer, statistics_invalid_flags, hs_dn,
            hs_stdev_dn);
        return;
    }

    /* Now connect to the database and get the information */
    open_db = connect_to_db();
    assert_true(open_db != NULL);
    if (open_db == NULL)
    {
        assert_true_with_message(FALSE, "Error connecting to database");
        ias_db_close_database_lib();
        free_data_buffers(hist_stats_mean_buffer, hist_stats_stdev_buffer,
                         hist_stats_frames_buffer, statistics_invalid_flags,
                         hs_dn, hs_stdev_dn);
        return;
    }

    /* Get the radiometric work order ID key for the specified work order
       string */
    status = ias_misc_get_char_id(open_db,
        "RADIOMETRIC_WORK_ORDER_COMMON", "radiometric_wo_common_id",
        work_order_id, rad_wo_common_id);
    assert_not_equal(status, ERROR);
    if (status == ERROR)
    {
        assert_true_with_message(FALSE, "Invalid rad_wo_common_id key "
               "value for specified work order ID %s", work_order_id);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        free_data_buffers(hist_stats_mean_buffer, hist_stats_stdev_buffer,
                          hist_stats_frames_buffer, statistics_invalid_flags,
                          hs_dn, hs_stdev_dn);
        return;
    }


    /* Copy the DB statistics into the database array, and read the
       file statistics into the file array */
    for (sca_index = 0; sca_index < ba->scas; sca_index++)
    {
        start_detector_index = sca_index * ba->detectors_per_sca;

        status = get_histogram_statistics(open_db, rad_wo_common_id,
            IAS_SPECTRAL_VNIR, ba->band_number, (sca_index + 1),
            ba->detectors_per_sca,
            RPS_SOLARCAL_TEMPCORRECT, hist_stats_mean_buffer,
            hist_stats_stdev_buffer, hist_stats_frames_buffer);

        assert_true(status == SUCCESS);
        if (status != SUCCESS)
        {
            assert_true_with_message(FALSE, "Cannot retrieve histogram "
                   "statistics for SCA %d", (sca_index + 1));
            free_data_buffers(hist_stats_mean_buffer,
                              hist_stats_stdev_buffer,
                              hist_stats_frames_buffer,
                              statistics_invalid_flags, hs_dn,
                              hs_stdev_dn);
            ias_db_close_connection(open_db);
            ias_db_close_database_lib();
            return;
        }

        /* Copy the statistics for the current SCA into the DB arrays */
        for (detector_index = 0;
             detector_index < ba->detectors_per_sca;
             detector_index++)
        {
            offset = detector_index + start_detector_index;
            hs_dn[sca_index][detector_index]
                = hist_stats_mean_buffer[offset];
            hs_stdev_dn[sca_index][detector_index]
                = hist_stats_stdev_buffer[offset];
        } 
    }

    /* Close the database connection */
    ias_db_close_connection(open_db);
    ias_db_close_database_lib();

    /* Calculate  the average number of frames.  Since all detectors
       "have" 500 frames of data, the average for all detectors should
       be 500 as well */
    status = ias_rps_calculate_filtered_mean(hist_stats_frames_buffer,
                                       statistics_invalid_flags,
                                       num_band_detectors,
                                       &num_frames_averaged, NULL);
    assert_true(status == SUCCESS);
    assert_equal((int)num_frames_averaged,
                 (int)hist_stats_frames_buffer[0]);
    assert_equal((int)num_frames_averaged, 500);
    if ((status != SUCCESS)
         || ((int)num_frames_averaged != (int)hist_stats_frames_buffer[0])
         || ((int)num_frames_averaged != 500))
    {
        assert_true_with_message(FALSE, "Could not retrieve correct "
            "average number of frames used in calculations");
    }

    /* Clean up existing data buffers */
    free_data_buffers(hist_stats_mean_buffer, hist_stats_stdev_buffer,
                      hist_stats_frames_buffer, statistics_invalid_flags,
                      hs_dn, hs_stdev_dn);
}




Ensure test_write_to_detector_table()
{
    int band_number = 1;
    int sca_number  = 1;
    char rad_wo_common_id[IAS_CHAR_ID_SIZE];  
    int detector_index;
    int status;

    int db_null_data_flag = -1; /* Same as ias_db_get_null_flag() returns */
    int *detector_ids = NULL;

    IAS_DB_NULL_TYPE *flagged_detector_ids = NULL;

    double *sca_relgain_data = NULL;
    double *sca_relgain_stdev_data = NULL;

    struct ias_db_connection *open_db = NULL;


    printf("Testing write to detector-level table...\n");

    /* Allocate the test data buffers */
    detector_ids = calloc(100, sizeof(*detector_ids));
    flagged_detector_ids = calloc(100, sizeof(*flagged_detector_ids));
    sca_relgain_data = calloc(100, sizeof(*sca_relgain_data));
    sca_relgain_stdev_data = calloc(100, sizeof(*sca_relgain_stdev_data));

    if ((detector_ids == NULL ) || (flagged_detector_ids == NULL)
        || (sca_relgain_data == NULL) || (sca_relgain_stdev_data == NULL))
    {
        assert_true_with_message(FALSE, "Cannot allocate buffer memory for "
            "this test\n");
        free(detector_ids);
        free(flagged_detector_ids);
        free(sca_relgain_data);
        free(sca_relgain_stdev_data);
        return;
    }

    /* Make up relative gains/stdevs, and throw in some inoperable
       detectors, too (set the element in flagged_detector_ids to -1) */
    for (detector_index = 0; detector_index < 100; detector_index++)
    {
        detector_ids[detector_index] = (detector_index + 1);
        sca_relgain_data[detector_index] = 1.0005925;
        sca_relgain_stdev_data[detector_index] = 0.0005553;
    }

    /* Inoperable detectors 4, 28, and 67 */
    flagged_detector_ids[3]  = db_null_data_flag;
    flagged_detector_ids[27] = db_null_data_flag;
    flagged_detector_ids[66] = db_null_data_flag;

    /* Get a database connection */
    open_db = connect_to_db();
    assert_true(open_db != NULL);
    if (open_db == NULL)
    {
        printf("Cannot connect to database\n");
        free(detector_ids);
        free(flagged_detector_ids);
        free(sca_relgain_data);
        free(sca_relgain_stdev_data);
        return;
    }

    /* Get the radiometric work order ID key for the specified work order
       string */
    status = ias_misc_get_char_id(open_db, "RADIOMETRIC_WORK_ORDER_COMMON",
        "radiometric_wo_common_id", work_order_id, rad_wo_common_id);
    assert_not_equal(status, ERROR);
    if (status == ERROR)
    {
        assert_true_with_message(FALSE, "Invalid rad_wo_common_id key "
            "value for specified work order ID %s", work_order_id);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        free(detector_ids);
        free(flagged_detector_ids);
        free(sca_relgain_data);
        free(sca_relgain_stdev_data);
        return;
    }

    /* Clear the table */
    status = clear_solarcal_table(open_db, rad_wo_common_id,
                                  "SOLAR_RESPONSE_DETECTOR");
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Could not clear "
            "SOLAR_RESPONSE_DETECTOR table\n");
        free(detector_ids);
        free(flagged_detector_ids);
        free(sca_relgain_data);
        free(sca_relgain_stdev_data);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        return;
    }

    /* Write to the table */
    status = write_to_detector_table(open_db, rad_wo_common_id,
                                     band_number, sca_number,
                                     100, detector_ids,
                                     flagged_detector_ids,
                                     sca_relgain_data,
                                     sca_relgain_stdev_data);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Cannot write to "
            "SOLAR_RESPONSE_DETECTOR table\n");
    }

    /* Clean up */
    free(detector_ids);
    free(flagged_detector_ids);
    free(sca_relgain_data);
    free(sca_relgain_stdev_data);
    ias_db_close_connection(open_db);
    ias_db_close_database_lib();
}




Ensure test_write_to_band_table()
{
    int status;
    char rad_wo_common_id[IAS_CHAR_ID_SIZE];  
    int band_number = 1;

    RPS_SOLARCAL_BAND_GAINS band_gains;

    struct ias_db_connection *open_db = NULL;


    printf("Testing write to band-level table...\n");

    /* Connect to the database */
    open_db = connect_to_db();
    assert_true(open_db != NULL);
    if (open_db == NULL)
    {
        printf("Could not connect to database\n");
        return;
    }

    /* Get the radiometric work order ID key for the specified work order
       string */
    status = ias_misc_get_char_id(open_db,
        "RADIOMETRIC_WORK_ORDER_COMMON", "radiometric_wo_common_id",
        work_order_id, rad_wo_common_id);
    assert_not_equal(status, ERROR);
    if (status == ERROR)
    {
        assert_true_with_message(FALSE, "Invalid rad_wo_common_id key "
            "value for specified work order ID %s", work_order_id);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        return;
    }

    /* Clear the SOLAR_RESPONSE_BAND table */
    status = clear_solarcal_table(open_db, rad_wo_common_id,
                                  "SOLAR_RESPONSE_BAND");
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Cannot clear SOLAR_RESPONSE_BAND "
            "table");
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        return;
    }

    /* Initialize the band data structure with prototype test data*/
    memset(&band_gains, 0, sizeof(RPS_SOLARCAL_BAND_GAINS));

    band_gains.band_number = band_number;
    band_gains.hist_stats_mean = 158.50077;
    band_gains.hist_stats_stdev = 0.290000;
    band_gains.diffuser_radiance_mean = 175.684600;
    band_gains.diffuser_radiance_stdev = 0.000000;
    band_gains.radiance_gain = 0.930400;
    band_gains.radiance_gain_stdev = 8.89e-15;
    band_gains.diffuser_reflectance_mean = 0.295600;
    band_gains.diffuser_reflectance_stdev = 0.000000;
    band_gains.reflectance_gain = 1832.146;
    band_gains.reflectance_gain_stdev = 1.39e-11;

    /* Write the information to the table */
    status = write_to_band_table(open_db, rad_wo_common_id,
        band_gains);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Cannot write to "
            "SOLAR_RESPONSE_BAND table");
    }

    /* Close the database connection */
    ias_db_close_connection(open_db);
    ias_db_close_database_lib();
}




Ensure test_write_to_main_table()
{
    struct ias_db_connection *open_db = NULL;

    int status;
    char rad_wo_common_id[IAS_CHAR_ID_SIZE];  

    RPS_SOLARCAL_SOLAR_ANGLES angles;
    RPS_SOLARCAL_REPORT_HEADER header;


    printf("Testing write to main table...\n");

    /* Initialize and populate the data structures */
    memset(&angles, 0, sizeof(RPS_SOLARCAL_SOLAR_ANGLES));
    memset(&header, 0, sizeof(RPS_SOLARCAL_REPORT_HEADER));

    /* Made-up angle/stats values.  The point of this test is to confirm
       that the SOLAR_RESPONSE table can be written to.  */
    angles.incident_solar_zenith       =   45.00002;
    angles.incident_solar_zenith_stdev =    0.00558;
    angles.view_solar_zenith           =   45.00001;
    angles.view_solar_zenith_stdev     =    0.00063;
    angles.view_solar_azimuth          = -150.00078;
    angles.view_solar_azimuth_stdev    =    0.00072;
    angles.earth_sun_distance_factor   =    1.01553;

    /* Elapsed collect time, panel deployment angle from prototype
       SOCAL_REPORT.txt.  Times are not, but difference is reported
       elapsed time.  */
    strcpy(header.rps_header.acq_datetime.start_time_oli,
            "2014:184:17:00:45.00000");
    strcpy(header.rps_header.acq_datetime.stop_time_oli,
            "2014:184:17:01:13.67000");

    header.panel_id = IAS_RPS_OLI_DIFFUSER_TYPE_WORKING;
    header.diffuser_resolver_position = 8752;
    header.rps_header.acq_datetime.total_time_oli = 28.6700;

    /* Get a database connection */
    open_db = connect_to_db();
    assert_true(open_db != NULL);
    if (open_db == NULL)
    {
        assert_true_with_message(FALSE, "Cannot connect to database\n");
        return;
    }

    /* Get the radiometric work order ID key for the specified work order
       string */
    status = ias_misc_get_char_id(open_db,
        "RADIOMETRIC_WORK_ORDER_COMMON", "radiometric_wo_common_id",
        work_order_id, rad_wo_common_id);
    assert_not_equal(status, ERROR);
    if (status == ERROR)
    {
        assert_true_with_message(FALSE, "Invalid rad_wo_common_id key "
            "value for specified work order ID %s", work_order_id);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        return;
    }

    /* Clear the table */
    status = clear_solarcal_table(open_db, rad_wo_common_id, "SOLAR_RESPONSE");
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Cannot clear SOLAR_RESPONSE "
            "table");
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        return;
    }

    /* Write to the table */
    status = write_to_main_table(open_db, rad_wo_common_id, header, angles);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Cannot write to SOLAR_RESPONSE "
            "table");
    }

    /* Clean up */
    ias_db_close_connection(open_db);
    ias_db_close_database_lib();
}



               /* Build the DB access test "sub-suite" */
void solarcal_oli_dbaccess_test(TestSuite *suite)
{
    add_test(suite, test_get_histogram_statistics);
    add_test(suite, test_write_to_detector_table);
    add_test(suite, test_write_to_band_table);
    add_test(suite, test_write_to_main_table);
}
