/***************************************************************************
 NAME:                   solarcal_oli_get_cpf_data_test

 PURPOSE:   CGreen unit test of the routines and write_solarcal_dump_file()
            to make sure the report and dump file are properly created

 RETURNS:   CGreen status

****************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "cgreen.h"
#include "ias_radiometry.h"

#include "../solarcal_oli.h"

#include "solarcal_oli_tests.h"


#define CPF_STRLEN 26


Ensure test_copy_cpf_diffuser_data()
{
    const char cpf_basename[CPF_STRLEN] = "L8CPF20090101_20090331.01";

    char *test_data_path = NULL;
    char base_cpf_datadir[] = "ias_lib/cpf";
    char full_cpf_filename[PATH_MAX] = "";

    IAS_CPF   *cpf = NULL;
    const struct IAS_CPF_DIFFUSER_RADIANCE *diff_rad = NULL;

    const IAS_BAND_ATTRIBUTES *ba = NULL;

    double *diffuser_radiances = NULL;
    double *diffuser_reflectances = NULL;

    double prim_rad_vals[10] = {0.1, 0.2, 0.3, 0.4, 0.05, 0.06, 0.07, 0.08,
                                0.09, 0.10};
    double pris_rad_vals[10] = {0.1, 0.2, 0.3, 0.4, 0.005, 0.006, 0.007,
                                0.008, 0.009, 0.010};
    double prim_reflect_vals[10] = {0.1, 0.2, 0.3, 0.4, 0.0005, 0.0006,
                                    0.0007, 0.0008, 0.0009, 0.0010};
    double pris_reflect_vals[10] = {0.1, 0.2, 0.3, 0.4, 0.00005, 0.00006,
                                    0.00007, 0.00008, 0.00009, 0.00010};

    int sca_index = 0;
    int detector_index = 0;
    int start_detector_index = 0;
    int status = SUCCESS;
    int size;

    printf("Testing retrieval of CPF diffuser radiances/reflectances:\n");

    /* Get the band attributes */
    ba = ias_sat_attr_get_band_attributes(1);
    assert_true(ba != NULL);
    if (ba == NULL)
    {
        printf("Cannot read band 1 attributes\n");
        return;
    }

    /* Get the path to the test data */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(test_data_path != NULL);
    if (test_data_path == NULL)
    {
        printf("Cannot find test data test_data_path\n");
        return;
    }

    /* Build the name of the test CPF file */
    size = snprintf(full_cpf_filename, sizeof(full_cpf_filename), "%s/%s/%s", 
        test_data_path, base_cpf_datadir, cpf_basename);
    if ((size < 0) || (size >= sizeof(full_cpf_filename)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
           "it needs %d bytes", size);
        return;
    }

    /* Read the diffuser radiance data from the test CPF */
    cpf = ias_cpf_read(full_cpf_filename);
    assert_true(cpf != NULL);
    if (cpf == NULL)
    {
        printf("Cannot read test CPF %s\n", full_cpf_filename);
        return;
    }
    diff_rad = ias_cpf_get_diffuser_rad(cpf);
    assert_true(diff_rad != NULL);
    if (diff_rad == NULL)
    {
        printf("Cannot get diffuser radiance data from test CPF\n");
        ias_cpf_free(cpf);
        return;
    }

    /* Allocate the radiance data buffer */
    printf("Reading 'primary' radiances...\n");
    diffuser_radiances = calloc((ba->scas * ba->detectors_per_sca),
        sizeof(*diffuser_radiances));
    assert_not_equal_with_message(diffuser_radiances, NULL,
        "Cannot allocate diffuser radiance data buffer");
    if (diffuser_radiances == NULL)
    {
        ias_cpf_free(cpf);
        return;
    }

    /* The first 10 detectors in each SCA have the values given in the 
       rad_vals[] array declared above.  If we find this pattern in
       each SCA, we'll know the application code is getting the correct
       set of data */
    for (sca_index = 0; sca_index < ba->scas; sca_index++)
    {
        /* Copy the primary diffuser radiances from the CPF data.  Assume
           no inoperable detectors */
        status = copy_cpf_diffuser_data(diff_rad, NULL, 0, sca_index,
            ba->detectors_per_sca, GET_DIFFUSER_RADIANCES,
            IAS_RPS_OLI_DIFFUSER_WORKING, diffuser_radiances);
        assert_equal_with_message(status, SUCCESS, "Cannot read primary "
            "diffuser radiances from CPF");
        if (status != SUCCESS)
        {
            ias_cpf_free(cpf);
            free(diffuser_radiances);
            return;
        }

        start_detector_index = sca_index * ba->detectors_per_sca;

        /* Check */
        for (detector_index = start_detector_index;
             detector_index < start_detector_index + 10;
             detector_index++)
        {
            assert_double_equal_with_message(
                diffuser_radiances[detector_index],
                prim_rad_vals[detector_index - start_detector_index],
                "Detector %d:  CPF primary radiance: %lf, array: %lf",
                (detector_index + 1), diffuser_radiances[detector_index],
                prim_rad_vals[detector_index - start_detector_index]);
        }
    }

     /* Clear the radiance data buffer */
     memset(diffuser_radiances, 0,
        (ba->scas * ba->detectors_per_sca * sizeof(double)));

    printf("Reading 'pristine' radiances...\n");
    for (sca_index = 0; sca_index < ba->scas; sca_index++)
    {
        /* Get the pristine diffuser radiances */
        status = copy_cpf_diffuser_data(diff_rad, NULL, 0, sca_index,
            ba->detectors_per_sca, GET_DIFFUSER_RADIANCES,
            IAS_RPS_OLI_DIFFUSER_PRISTINE, diffuser_radiances);
        assert_equal_with_message(status, SUCCESS, "Cannot read primary "
            "diffuser radiances from CPF");
        if (status != SUCCESS)
        {
            ias_cpf_free(cpf);
            free(diffuser_radiances);
            return;
        }

        start_detector_index = sca_index * ba->detectors_per_sca;

        /* Check */
        for (detector_index = start_detector_index;
             detector_index < start_detector_index + 10;
             detector_index++)
        {
            assert_double_equal_with_message(
                diffuser_radiances[detector_index],
                pris_rad_vals[detector_index - start_detector_index],
                "Detector %d:  CPF primary radiance: %lf, array: %lf",
                (detector_index + 1), diffuser_radiances[detector_index],
                pris_rad_vals[detector_index - start_detector_index]);
        }
    }

    /* We don't need this anymore, get rid of it */
    free(diffuser_radiances);

    /* Allocate the reflectance data buffer */
    printf("Reading 'primary' reflectances...\n");

    diffuser_reflectances = calloc((ba->scas * ba->detectors_per_sca),
        sizeof(*diffuser_reflectances));
    assert_not_equal_with_message(diffuser_reflectances, NULL,
        "Cannot allocate diffuser radiance data buffer");
    if (diffuser_reflectances == NULL)
    {
        ias_cpf_free(cpf);
        return;
    }

    /* The first 10 detectors in each SCA have the values given in the 
       rad_vals[] array declared above.  If we find this pattern in
       each SCA, we'll know the application code is getting the correct
       set of data */
    for (sca_index = 0; sca_index < ba->scas; sca_index++)
    {
        /* Get the primary diffuser reflectances */
        status = copy_cpf_diffuser_data(diff_rad, NULL, 0, sca_index,
            ba->detectors_per_sca, GET_DIFFUSER_REFLECTANCES,
            IAS_RPS_OLI_DIFFUSER_WORKING, diffuser_reflectances);
        assert_equal_with_message(status, SUCCESS, "Cannot read primary "
            "diffuser reflectances from CPF");
        if (status != SUCCESS)
        {
            ias_cpf_free(cpf);
            free(diffuser_reflectances);
            return;
        }

        start_detector_index = sca_index * ba->detectors_per_sca;

        /* Check */
        for (detector_index = start_detector_index;
             detector_index < start_detector_index + 10;
             detector_index++)
        {
            assert_double_equal_with_message(
                diffuser_reflectances[detector_index],
                prim_reflect_vals[detector_index - start_detector_index],
                "Detector %d:  CPF primary reflectance: %lf, array: %lf",
                (detector_index + 1), diffuser_reflectances[detector_index],
                prim_reflect_vals[detector_index - start_detector_index]);
        }
    }

    /* Clear the reflectance data buffer */
    memset(diffuser_reflectances, 0,
        (ba->scas * ba->detectors_per_sca * sizeof(double)));

    printf("Reading 'pristine' reflectances...\n");
    for (sca_index = 0; sca_index < ba->scas; sca_index++)
    {
        /* Get the pristine diffuser reflectances */
        status = copy_cpf_diffuser_data(diff_rad, NULL, 0, sca_index,
            ba->detectors_per_sca, GET_DIFFUSER_REFLECTANCES,
            IAS_RPS_OLI_DIFFUSER_PRISTINE, diffuser_reflectances);
        assert_equal_with_message(status, SUCCESS, "Cannot read primary "
            "diffuser reflectances from CPF");
        if (status != SUCCESS)
        {
            ias_cpf_free(cpf);
            free(diffuser_reflectances);
            return;
        }

        start_detector_index = sca_index * ba->detectors_per_sca;

        /* Check */
        for (detector_index = start_detector_index;
             detector_index < start_detector_index + 10;
             detector_index++)
        {
            assert_double_equal_with_message(
                diffuser_reflectances[detector_index],
                pris_reflect_vals[detector_index - start_detector_index],
                "Detector %d:  CPF primary reflectance: %lf, array: %lf",
                (detector_index + 1), diffuser_reflectances[detector_index],
                pris_reflect_vals[detector_index - start_detector_index]);
        }
    }

    /* Clean up */
    free(diffuser_reflectances);
    ias_cpf_free(cpf);

    /* Done */
    return;
}



void solarcal_oli_get_cpf_data_test(TestSuite *suite)
{
    add_test(suite, test_copy_cpf_diffuser_data);
}
