/***************************************************************************
 NAME:                   solarcal_oli_parameter_test

 PURPOSE:   CGreen unit test of the get_parameters() routine to make sure
            it's reading the WO and OMF parameter files correctly

 RETURNS:   CGreen status

****************************************************************************/
#include <stdio.h>
#include <string.h>
#include <limits.h>

#include "cgreen.h"

#include "../solarcal_oli.h"

#include "solarcal_oli_tests.h"


Ensure test_get_parameters()
{
    char *test_data_path = NULL;
    char full_l0r_name[PATH_MAX] = "";
    char full_cpf_name[PATH_MAX] = "";
    char full_ancillary_name[PATH_MAX] = "";
    char work_order_id[IAS_WORKORDER_ID_SIZE] = "DRDT1002";
    char l0r_data_dir[] = "ias_lib/misc/l0ra_location";
    char l0r_name[] = "test";
    char ancillary_name[] = "test_ANC.h5";
    char cpf_data_dir[] = "ias_lib/cpf";
    char cpf_name[] = "L8CPF20090101_20090331.01";
    char wo_params_filename[PATH_MAX] = "";
    char system_command[1000] = "";

    int status;
    int size;

    RPS_SOLARCAL_WO_PARAMS  wo_params;
    RPS_SOLARCAL_OMF_PARAMS omf_params;


    printf("Testing  parameter read from work order and OMF files...\n");

    /* Get the test data directory path */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(test_data_path != NULL);
    if (test_data_path == NULL)
    {
        assert_true_with_message(FALSE, "Cannot find test data directory\n");
        return;
    }

    /* Replace the $IAS_UNIT_TEST_DATA_DIR placeholder in the OMF file with
       its actual value */
    size = snprintf(system_command, sizeof(system_command), 
        "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' %s_template.omf > %s.omf", 
        test_data_path, work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
           "it needs %d bytes", size);
        return;

    }
    status = system(system_command);
    assert_equal_with_message(status, 0, "Cannot edit OMF template");

    /* Rename the template ODL file as the actual ODL file */
    size = snprintf(system_command, sizeof(system_command), 
        "cp %s_template.odl %s.odl", work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
           "it needs %d bytes", size);
        return;
    }
    status = system(system_command);
    assert_equal_with_message(status, 0, "Cannot copy working template "
        "parameter file");

    /* Read the parameters. */
    memset(&wo_params, 0, sizeof(RPS_SOLARCAL_WO_PARAMS));
    memset(&omf_params, 0, sizeof(RPS_SOLARCAL_OMF_PARAMS));

    size = snprintf(wo_params_filename, sizeof(wo_params_filename), 
           "%s.odl", work_order_id);
    if ((size < 0) || (size >= sizeof(wo_params_filename)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
           "it needs %d bytes", size);
        return;
    }
    status = get_parameters(wo_params_filename, &wo_params,
        &omf_params);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Failed to read parameters\n");
        return;
    }

    /* Initialize parameter data structures */
    memset(&wo_params, 0, sizeof(RPS_SOLARCAL_WO_PARAMS));
    memset(&omf_params, 0, sizeof(RPS_SOLARCAL_OMF_PARAMS));

    /* Read the parameters */
    status = get_parameters(wo_params_filename, &wo_params, &omf_params);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Cannot read processing parameters");
        return;
    }

    /* Check parameters to see if they've been read properly */
    assert_string_equal(wo_params.work_order_id, "DRDT1002");
    assert_equal(wo_params.store_char_in_db, 0);
    assert_equal(wo_params.report_flag, 1);
    assert_equal(wo_params.dump_flag, 1);

    size = snprintf(full_cpf_name, sizeof(full_cpf_name), "%s/%s/%s", 
        test_data_path, cpf_data_dir, cpf_name);
    if ((size < 0) || (size >= sizeof(full_cpf_name)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
           "it needs %d bytes", size);
        return;
    }
    assert_string_equal(omf_params.cpf_filename, full_cpf_name);
    size = snprintf(full_ancillary_name, sizeof(full_ancillary_name), 
        "%s/%s/%s", test_data_path, l0r_data_dir, ancillary_name);
    if ((size < 0) || (size >= sizeof(full_ancillary_name)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
           "it needs %d bytes", size);
        return;
    }
    assert_string_equal(omf_params.adp_ancillary_filename,
        full_ancillary_name);

    size = snprintf(full_l0r_name, sizeof(full_l0r_name), "%s/%s/%s", 
        test_data_path, l0r_data_dir, l0r_name);
    if ((size < 0) || (size >= sizeof(full_l0r_name)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
           "it needs %d bytes", size);
        return;
    }
    assert_string_equal(omf_params.l0r_filename, full_l0r_name);

    size = snprintf(full_cpf_name, sizeof(full_cpf_name), "%s/%s/%s", 
        test_data_path, cpf_data_dir, cpf_name);
    if ((size < 0) || (size >= sizeof(full_cpf_name)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
           "it needs %d bytes", size);
        return;
    }
    assert_string_equal(omf_params.cpf_filename, full_cpf_name);

    /* Clean up */
    size = snprintf(system_command, sizeof(system_command), 
        "rm %s.omf %s.odl", work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
           "it needs %d bytes", size);
        return;
    }
    system(system_command);
}




void solarcal_oli_parameter_test(TestSuite *suite)
{
    add_test(suite, test_get_parameters);
}
