/***************************************************************************
  NAME:              solarcal_oli_validate_data_test

  PURPOSE:  Test routines that acquire a test set of collect data
            (read into a standard RPS report header) and
            checks it with the routines implemented in
            validate_collect_data.c

  NOTES:    Some error messages from the validate_collect_data
            routines will be logged.  This is good--deliberate
            errors are introduced into various pieces of information,
            all of which should be flagged as errors.  If we weren't seeing
            these error messages, that would indicate the routines were not
            working properly.

***************************************************************************/
#include <limits.h>                /* PATH_MAX */
#include "ias_l0r.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_structures.h"
#include "cgreen.h"
#include "../solarcal_oli.h"
#include "solarcal_oli_tests.h"

/* Current OLI image header and CPF values for these parameters.  Define
   the constants here so we don't have to read everything in */
#define CPF_MS_INTEGRATION_TIME_NOMINAL  3.600  /* milliseconds */
#define CPF_PAN_INTEGRATION_TIME_NOMINAL 0.660  /* milliseconds */
#define OLI_L0R_IMAGEHEADER_MS_INTEGRATION_TIME   3600  /* microseconds */
#define OLI_L0R_IMAGEHEADER_PAN_INTEGRATION_TIME   660  /* microseconds */

#define MAX_EXPECTED_WORKING_DIFFUSER_ANGLE     312.6
#define MIN_EXPECTED_WORKING_DIFFUSER_ANGLE     308.2
#define MAX_EXPECTED_PRISTINE_DIFFUSER_ANGLE     20.8
#define MIN_EXPECTED_PRISTINE_DIFFUSER_ANGLE     16.4


/* Check that calculation of diffuser panel deployment angles works as
   expected */
Ensure test_convert_diffuser_resolver_position_to_angle()
{
    double max_working_deploy_angle;
    double max_working_expected_angle = MAX_EXPECTED_WORKING_DIFFUSER_ANGLE;
    double min_working_deploy_angle;
    double min_working_expected_angle = MIN_EXPECTED_WORKING_DIFFUSER_ANGLE;
    double max_pristine_deploy_angle;
    double max_pristine_expected_angle =
        MAX_EXPECTED_PRISTINE_DIFFUSER_ANGLE;
    double min_pristine_deploy_angle;
    double min_pristine_expected_angle =
        MIN_EXPECTED_PRISTINE_DIFFUSER_ANGLE;

    /* Display a status message on screen */
    printf("Checking calculation of deployment angle from "
        "diffuser_resolver_position telemetry values...\n");

    /* Calculate the deployment angles for the minimum and maximum
       diffuser resolver positions */
    convert_diffuser_resolver_position_to_angle(
        IAS_RPS_OLI_DIFFUSER_POSITION_WORKING_HIGH,
        &max_working_deploy_angle);

    convert_diffuser_resolver_position_to_angle(
        IAS_RPS_OLI_DIFFUSER_POSITION_WORKING_LOW,
        &min_working_deploy_angle);

    convert_diffuser_resolver_position_to_angle(
        IAS_RPS_OLI_DIFFUSER_POSITION_PRISTINE_HIGH,
        &max_pristine_deploy_angle);

    convert_diffuser_resolver_position_to_angle(
        IAS_RPS_OLI_DIFFUSER_POSITION_PRISTINE_LOW,
        &min_pristine_deploy_angle);

    /* Check the calculated results against the expected results */
    assert_double_equal_with_message(max_working_deploy_angle,
        max_working_expected_angle, "Estimated max. working panel deployment "
        "angle %lf does not match expected value %lf",
        max_working_deploy_angle, max_working_expected_angle);

    assert_double_equal_with_message(min_working_deploy_angle,
        min_working_expected_angle, "Estimated min. working panel deployment "
        "angle %lf does not match expected value %lf",
        min_working_deploy_angle, min_working_expected_angle);

    assert_double_equal_with_message(max_pristine_deploy_angle,
        max_pristine_expected_angle, "Estimated max. pristine panel "
        "deployment angle %lf does not match expected value %lf",
        max_pristine_deploy_angle, max_pristine_expected_angle);

    assert_double_equal_with_message(min_pristine_deploy_angle,
        min_pristine_expected_angle, "Estimated min. pristine panel "
        "deployment angle %lf does not match expected value %lf",
        min_pristine_deploy_angle, min_pristine_expected_angle);
}   /* END -- test_convert_diffuser_resolver_position_to_angle */



/* Check that logic testing the L0R frame header information in the
   RPS report header works */
Ensure test_validate_image_header()
{
    RPS_SOLARCAL_REPORT_HEADER header;

    const char l0r_basename[PATH_MAX] = "test_solar";
    char *test_data_path = NULL;
    char full_l0r_filename[PATH_MAX];

    int status;
    int validation_status;
    int size;

    printf("Testing validation of image header information...\n");

    /* Disable expected error messages */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* Construct the name of the test header L0R data file */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path == NULL)
    {
        assert_true_with_message(FALSE, "Cannot find test data directory");
        return;
    }
    size = snprintf(full_l0r_filename, sizeof(full_l0r_filename), 
            "%s/ias/rps/det_response_diffuser/nonuniformity/%s", test_data_path,
            l0r_basename);
    if ((size < 0) || (size >= sizeof(full_l0r_filename)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
           "it needs %d bytes", size);
        return;
    }

    /* Read the test data file and populate the header info data structure */
    memset(&header, 0, sizeof(RPS_SOLARCAL_REPORT_HEADER));

    /* Populate the RPS-specific header */
    status = ias_misc_initialize_rps_report_header(full_l0r_filename,
        &(header.rps_header));
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Cannot populate RPS-specific "
            "header information ");
        return;
    }

    /* Case 1:  All RPS-specific image header data correctly read in */
    assert_equal(header.rps_header.ms_integration_time,
        OLI_L0R_IMAGEHEADER_MS_INTEGRATION_TIME);
    assert_equal(header.rps_header.pan_integration_time,
        OLI_L0R_IMAGEHEADER_PAN_INTEGRATION_TIME);
    assert_equal(header.rps_header.current_detector_select_table, 1);
    assert_equal(header.rps_header.image_data_truncation_setting, 1);

    /* b.  The image header has the nominal integration times of 3600
           microsecs and 660 microsecs for the MS and PAN bands,
           respectively.  This should pass */
    validation_status = validate_image_header(header,
        CPF_MS_INTEGRATION_TIME_NOMINAL,
        CPF_PAN_INTEGRATION_TIME_NOMINAL);
    assert_equal(validation_status, RPS_SOLARCAL_IMAGEHEADER_VALIDATED);

    /* Change the header integration times to "incorrect" values, then
       recheck */
    header.rps_header.ms_integration_time =
        OLI_L0R_IMAGEHEADER_MS_INTEGRATION_TIME + 10;
    header.rps_header.pan_integration_time =
        OLI_L0R_IMAGEHEADER_PAN_INTEGRATION_TIME + 150;

    validation_status = validate_image_header(header,
        CPF_MS_INTEGRATION_TIME_NOMINAL,
        CPF_PAN_INTEGRATION_TIME_NOMINAL);
    assert_equal(validation_status,
        RPS_SOLARCAL_IMAGEHEADER_INVALID_INTEGRATION_TIME);

    /* Reset logging level */
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);
}




Ensure test_validate_interval_metadata()
{
    RPS_SOLARCAL_REPORT_HEADER header;

    const char l0r_basename[MAX_FILENAME_SIZE] = "test_solar";
    const char work_order_id[IAS_WORKORDER_ID_SIZE] = "DRDT24000";

    char *test_data_path = NULL;
    char full_l0r_filename[PATH_MAX];

    int status;
    int validation_status;
    int size;

    printf("Testing validation of header metadata information...\n");

    /* Construct the name of the test header L0R data file */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(test_data_path != NULL);
    if (test_data_path == NULL)
    {
        printf("Cannot find test data directory\n");
        return;
    }
    size = snprintf(full_l0r_filename, sizeof(full_l0r_filename), 
                "%s/ias/rps/det_response_diffuser/nonuniformity/%s", 
                test_data_path, l0r_basename);
    if ((size < 0) || (size >= sizeof(full_l0r_filename)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
           "it needs %d bytes", size);
        return;
    }

    /* Read the test data file and populate the header info data structure */
    memset(&header, 0, sizeof(RPS_SOLARCAL_REPORT_HEADER));

    /* Start with the standard header */
    header.common_header = ias_misc_initialize_report_header(
            full_l0r_filename, work_order_id, "test_validate_data");
    if (!header.common_header)
    {
        assert_true_with_message(FALSE, "Cannot populate standard IAS "
                "header information");
        return;
    }

    /* Populate the RPS-specific header */
    status = ias_misc_initialize_rps_report_header(full_l0r_filename,
            &(header.rps_header));
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Cannot populate RPS-specific "
                "header information ");
        ias_misc_free_report_header(header.common_header);
        return;
    }

    /* The SPACECRAFT_ID field should have the current value LANDSAT_8
       (all caps).  If it isn't set to this value in the test data set,
       change it in the header */
    if (strcmp(ias_misc_get_header_spacecraft_id(header.common_header),
        "LANDSAT_8") != 0)
    {
        ias_misc_set_header_spacecraft_id(header.common_header,
            "LANDSAT_8");
    }

    /* Case 1:  Everything should check OK */
    validation_status = validate_interval_metadata(header);
    assert_equal(validation_status, RPS_SOLARCAL_METADATA_VALIDATED);

    /* Case 2:  Change the sensor ID.  Should flag an ERROR */
    ias_misc_set_header_sensor_id(header.common_header, "TIRS");
    validation_status = validate_interval_metadata(header);
    assert_equal(validation_status, RPS_SOLARCAL_METADATA_INVALID_SENSOR);

    /* Case 3:  Change the spacecraft ID.  Should flag an ERROR */
    ias_misc_set_header_sensor_id(header.common_header, "OLI");
    ias_misc_set_header_spacecraft_id(header.common_header, "Landsat_7");
    validation_status = validate_interval_metadata(header);
    assert_equal(validation_status, RPS_SOLARCAL_METADATA_INVALID_SPACECRAFT);

    /* Case 4:  Change the path/row.  Should flag an ERROR */
    int new_startpath = 10;
    int new_startrow = 0;
    int new_endrow = 0;
    ias_misc_set_header_spacecraft_id(header.common_header, "LANDSAT_8");
    ias_misc_set_header_starting_ending_pathrow(header.common_header,
            new_startpath, new_startrow, new_endrow);
    validation_status = validate_interval_metadata(header);
    assert_equal(validation_status, RPS_SOLARCAL_METADATA_INVALID_WRS_PATHROW);

    /* Clean up */
    ias_misc_free_report_header(header.common_header);

    /* Reset logging level */
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);
}




void solarcal_oli_validate_data_test(TestSuite *suite)
{
    add_test(suite, test_convert_diffuser_resolver_position_to_angle);
    add_test(suite, test_validate_image_header);
    add_test(suite, test_validate_interval_metadata);
}


