/***************************************************************************
 NAME:                   solarcal_oli_write_reports_test

 PURPOSE:   CGreen unit tests of the routines
            write_solarcal_summary_report() and write_solarcal_dump_file()
            to make sure the report and dump file are properly
            created

 RETURNS:   CGreen status

****************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <limits.h>                     /* PATH_MAX */
#include "ias_const.h"
#include "ias_l0r.h"
#include "ias_radiometry.h"
#include "ias_miscellaneous.h"
#include "cgreen.h"
#include "../solarcal_oli.h"
#include "solarcal_oli_tests.h"

#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE  1
#endif


static char test_data_dir[1000] = "ias/rps/det_response_diffuser";
static char system_command[1000] = "";
static char wo_params_filename[PATH_MAX] = "";
static char work_order_id[] = "DRDT1002";


Ensure test_write_solarcal_summary_report()
{
    char *test_data_path = NULL;
    char *collection_type = NULL;
    char reportfile_name[PATH_MAX] = "";
    char expected_reportfile_name[PATH_MAX] = "";

    FILE *rpt_fptr = NULL;

    const IAS_BAND_ATTRIBUTES *ba = NULL;

    RPS_SOLARCAL_REPORT_SUMMARY summary;
    RPS_SOLARCAL_WO_PARAMS wo_params;
    RPS_SOLARCAL_OMF_PARAMS omf_params;
    RPS_SOLARCAL_SOLAR_ANGLES angles;
    RPS_SOLARCAL_BAND_GAINS band_gains;

    int status = 0;
    int band_index = 0;
    int size;

    printf("Testing creation of solar calibration summary report...\n");

    /* Get the test data directory path from the environment settings */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path == NULL)
    {
        assert_true_with_message(FALSE, "Cannot find test data directory");
        return;
    }

    /* Replace the $IAS_UNIT_TEST_DATA_DIR placeholder in the OMF file with
       its actual value */
    size = snprintf(system_command, sizeof(system_command), 
        "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' %s_template.omf > %s.omf", 
        test_data_path, work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
           "it needs %d bytes", size);
        return;
    }
    status = system(system_command);
    assert_equal_with_message(status, 0, "Cannot edit OMF template");
    if (status != 0)
    {
        return;
    }

    /* Rename the template ODL file as the actual ODL file */
    size = snprintf(system_command, sizeof(system_command), 
        "cp %s_template.odl %s.odl", work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
           "it needs %d bytes", size);
        return;
    }
    status = system(system_command);
    assert_equal_with_message(status, 0, "Cannot create working "
        "parameter file");
    if (status != 0)
    {
        size = snprintf(system_command, sizeof(system_command), "rm %s.omf", 
                  work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Read the WO and OMF parameters */
    size = snprintf(wo_params_filename, sizeof(wo_params_filename), "%s.odl", 
              work_order_id);
    if ((size < 0) || (size >= sizeof(wo_params_filename)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
           "it needs %d bytes", size);
        return;
    }

    /* Get the processing parameters */
    memset(&wo_params, 0, sizeof(RPS_SOLARCAL_WO_PARAMS));
    memset(&omf_params, 0, sizeof(RPS_SOLARCAL_OMF_PARAMS));
    status = get_parameters(wo_params_filename, &wo_params, &omf_params);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Cannot read work order and/or "
            "OMF parameter files\n");
        size = snprintf(system_command, sizeof(system_command), 
            "rm %s.omf %s.odl", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Initialize the entire report summary data structure */
    memset(&summary, 0, sizeof(RPS_SOLARCAL_REPORT_SUMMARY));

    /* Get the IAS common header information */
    summary.header.common_header = ias_misc_initialize_report_header(
        omf_params.l0r_filename, wo_params.work_order_id,
        "test_write_solarcal_summary_report");
    assert_true(summary.header.common_header != NULL);
    if (!summary.header.common_header)
    {
        assert_true_with_message(FALSE, "Cannot initialize IAS report "
            "header data");
        size = snprintf(system_command, sizeof(system_command), 
            "rm %s.omf %s.odl", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Update the CPF in the header to reflect what is in the OMF file. */
    ias_misc_set_header_cpf_name(summary.header.common_header,
        omf_params.cpf_filename);

    /* Get the RPS_specific common header information */
    status = ias_misc_initialize_rps_report_header(omf_params.l0r_filename,
        &(summary.header.rps_header));
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Cannot initialize RPS-specific "
            "report header information");
        ias_misc_free_report_header(summary.header.common_header);
        size = snprintf(system_command, sizeof(system_command), 
            "rm %s.omf %s.odl", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Initialize the angle information for the summary report */
    memset(&angles, 0, sizeof(RPS_SOLARCAL_SOLAR_ANGLES));

    angles.earth_sun_distance_factor = 1.01553;
    angles.incident_solar_zenith       = 105.00;
    angles.incident_solar_zenith_stdev =  0.003;
    angles.view_solar_zenith           = 45.00;
    angles.view_solar_zenith_stdev     =  0.006;
    angles.view_solar_azimuth          =  24.63;
    angles.view_solar_azimuth_stdev    =  0.0057;

    status = initialize_solarcal_report_data(wo_params, omf_params, angles,
         &summary);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Cannot initialize parameter "
            "and/or angle information for summary report");
        ias_misc_free_report_header(summary.header.common_header);
        size = snprintf(system_command, sizeof(system_command), 
            "rm %s.omf %s.odl", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
             assert_true_with_message(FALSE, "Command string too short, "
                "it needs %d bytes", size);
             return;
        }
        system(system_command);
        return;
    }

    /* Change the test dataset collection type to the L0Ra DFCB
       value "OLI_SOLAR" if it isn't already */
    collection_type = ias_misc_get_header_collection_type(
        summary.header.common_header);
    if ((collection_type == NULL)
        || (strstr(collection_type, IAS_RPS_OLI_SOLAR_COLLECT) != 0))
    {
        ias_misc_set_header_collection_type(summary.header.common_header,
            IAS_RPS_OLI_SOLAR_COLLECT);
    }

    /* Gin up additional information for temperatures, etc */
    summary.header.panel_id = IAS_RPS_OLI_DIFFUSER_TYPE_WORKING;
    summary.header.rps_header.diffuser_resolver_position =
        IAS_RPS_OLI_DIFFUSER_POSITION_WORKING_HIGH;

    summary.header.rps_header.oli_fpm7_start_temp  = 210.0;
    summary.header.rps_header.oli_fpm7_stop_temp   = 211.0;
    summary.header.rps_header.oli_fpm14_start_temp = 212.0;
    summary.header.rps_header.oli_fpm14_stop_temp  = 213.0;
    summary.header.rps_header.oli_fpa_window_start_temp  = 205.0;
    summary.header.rps_header.oli_fpa_window_stop_temp   = 206.0;
    summary.header.rps_header.oli_fpe_chassis_start_temp = 207.0;
    summary.header.rps_header.oli_fpe_chassis_stop_temp  = 208.0;
    summary.header.num_frames_averaged   = 100;

    /* Flesh out the band gain information for the summary report */
    for (band_index = 0; band_index < wo_params.num_bands_to_process;
         band_index++)
    {
        /* Clear out the previous band's information */
        memset(&band_gains, 0, sizeof(RPS_SOLARCAL_BAND_GAINS));

        ba = ias_sat_attr_get_band_attributes((band_index + 1));
        if (ba == NULL)
        {
            assert_true_with_message(FALSE, "Cannot get attributes for "
                "band %d", (band_index + 1));
            ias_misc_free_report_header(summary.header.common_header);
            size = snprintf(system_command, sizeof(system_command), 
                "rm %s.omf %s.odl", work_order_id, work_order_id);
            if ((size < 0) || (size >= sizeof(system_command)))
            {
                assert_true_with_message(FALSE, "Command string too short, "
                   "it needs %d bytes", size);
                return;
            }
            system(system_command);
            return;
        }

        strcpy(band_gains.band_name, ba->band_name);
        band_gains.band_number            = ba->band_number;
        band_gains.scas                   = ba->scas;
        band_gains.detectors_per_sca      = ba->detectors_per_sca;
        band_gains.diffuser_radiance_mean = 175.6846;
        band_gains.diffuser_radiance_stdev = 0.000;
        band_gains.hist_stats_mean        = 158.50077;
        band_gains.hist_stats_stdev       = 0.2900;
        band_gains.radiance_gain          = 0.9304;
        band_gains.radiance_gain_stdev    = 8.89e-15;
        band_gains.diffuser_reflectance_mean = 0.2956;
        band_gains.diffuser_reflectance_stdev = 0.000;
        band_gains.reflectance_gain       = 1832.146;
        band_gains.reflectance_gain_stdev = 1.39e-11;

        /* Copy this band's info to the summary report data structure */
        initialize_solarcal_summary_band_data(band_index, band_gains,
           &summary);
    }

    /* Attempt to create the report, and write the summary header info
       to it */
    status = snprintf(reportfile_name, sizeof(reportfile_name),
        "%s_solarcal_summary.txt", wo_params.work_order_id);
    if ((status < 0) || (status >= sizeof(reportfile_name)))
    {
        assert_true_with_message(FALSE, "Error constructing test summary "
            "report file name");
        free(summary.band_gains);
        ias_misc_free_report_header(summary.header.common_header);
        size = snprintf(system_command, sizeof(system_command), 
            "rm %s.omf %s.odl", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }
    rpt_fptr = fopen(reportfile_name, "w");
    if (rpt_fptr == NULL)
    {
        assert_true_with_message(FALSE, "Cannot create summary report\n");
        free(summary.band_gains);
        ias_misc_free_report_header(summary.header.common_header);
        size = snprintf(system_command, sizeof(system_command), 
            "rm %s.omf %s.odl", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    status = write_header_to_summary_report(summary.header, rpt_fptr);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Cannot write header information "
            "to report");
        free(summary.band_gains);
        ias_misc_free_report_header(summary.header.common_header);
        size = snprintf(system_command, sizeof(system_command), 
            "rm %s.omf %s.odl %s_solarcal_summary.txt",
            work_order_id, work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Write the angle information to the report */
    status = write_angle_statistics_to_summary_report(summary.solar_angles,
        rpt_fptr);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Cannot write solar angle "
            "information to report");
        free(summary.band_gains);
        ias_misc_free_report_header(summary.header.common_header);
        size = snprintf(system_command, sizeof(system_command), 
            "rm %s.omf %s.odl %s_solarcal_summary.txt",
            work_order_id, work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Finally, write the report body */
    status = write_solarcal_report_data(wo_params.num_bands_to_process,
        summary, rpt_fptr);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Cannot write calibration results "
            "to report");
    }

    /* Free up unneeded resources */
    fclose(rpt_fptr);
    free(summary.band_gains);
    ias_misc_free_report_header(summary.header.common_header);

    /* Run a 'diff' on the test report vs the expected report. */
    status = snprintf(expected_reportfile_name,
        sizeof(expected_reportfile_name),
        "%s/%s/%s_expected_solarcal_summary.txt", test_data_path,
        test_data_dir, wo_params.work_order_id);
    if ((status < 0) || (status >= sizeof(expected_reportfile_name)))
    {
        assert_true_with_message(FALSE, "Error constructing test summary "
            "report file name");
        size = snprintf(system_command, sizeof(system_command), 
            "rm %s.omf %s.odl", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    status = snprintf(system_command, sizeof(system_command),
        "sed '/Spacecraft:/,$!'d %s | diff -q - %s",
        reportfile_name, expected_reportfile_name);
    if ((status < 0) || (status >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Error preparing difference "
            "comparison system command");
        size = snprintf(system_command, sizeof(system_command), 
            "rm %s.omf %s.odl", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }
    status = system(system_command);
    assert_equal_with_message(0, status, "Differences found between test "
        "summary report file %s and reference report file %s",
        reportfile_name, expected_reportfile_name);

    /* Final cleanup */
    size = snprintf(system_command, sizeof(system_command), 
        "rm %s.omf %s.odl %s_solarcal_summary.txt",
        work_order_id, work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
           "it needs %d bytes", size);
        return;
    }
    system(system_command);
}




Ensure test_write_solarcal_dump_file()
{
    char *collection_type = NULL;
    char *test_data_path = NULL;
    char dumpfile_name[PATH_MAX];
    char expected_dumpfile_name[PATH_MAX];

    FILE *dump_fptr = NULL;

    const IAS_BAND_ATTRIBUTES *ba = NULL;

    RPS_SOLARCAL_REPORT_SUMMARY summary;
    RPS_SOLARCAL_WO_PARAMS wo_params;
    RPS_SOLARCAL_OMF_PARAMS omf_params;
    RPS_SOLARCAL_SOLAR_ANGLES angles;
    RPS_SOLARCAL_BAND_GAINS band_gains;

    double *relgains       = NULL;
    double *relgains_stdev = NULL;

    int *data_invalid_flags = NULL;

    int  band_list[1] = {1};

    int  status;
    int  band_index;
    int  detector_index;
    int  num_band_detectors;
    int  size;

    printf("Testing creation of solar calibration dump files...\n");

    /* Get the test data directory path from the environment settings */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path == NULL)
    {
        assert_true_with_message(FALSE, "Cannot find test data directory");
        return;
    }

    /* Replace the $IAS_UNIT_TEST_DATA_DIR placeholder in the OMF file with
       its actual value */
    size = snprintf(system_command, sizeof(system_command), 
        "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' %s_template.omf > %s.omf", 
        test_data_path, work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
           "it needs %d bytes", size);
        return;
    }
    status = system(system_command);
    assert_equal_with_message(status, 0, "Cannot edit OMF template");
    if (status != 0)
    {
        return;
    }

    /* Rename the template ODL file as the actual ODL file */
    size = snprintf(system_command, sizeof(system_command), 
        "cp %s_template.odl %s.odl", work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
           "it needs %d bytes", size);
        return;
    }
    status = system(system_command);
    assert_equal_with_message(status, 0, "Cannot create working "
        "parameter file");
    if (status != 0)
    {
        size = snprintf(system_command, sizeof(system_command), 
                 "rm %s.omf", work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Read the WO and OMF parameters */
    size = snprintf(wo_params_filename, sizeof(wo_params_filename), "%s.odl", 
           work_order_id);
    if ((size < 0) || (size >= sizeof(wo_params_filename)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
           "it needs %d bytes", size);
        return;
    }

    memset(&wo_params, 0, sizeof(RPS_SOLARCAL_WO_PARAMS));
    memset(&omf_params, 0, sizeof(RPS_SOLARCAL_OMF_PARAMS));
    status = get_parameters(wo_params_filename, &wo_params, &omf_params);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Cannot read work order and/or "
            "OMF parameter files\n");
        size = snprintf(system_command, sizeof(system_command), 
            "rm %s.omf %s.odl", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Initialize the entire report summary data structure */
    memset(&summary, 0, sizeof(RPS_SOLARCAL_REPORT_SUMMARY));

    /* Get the IAS common header information */
    summary.header.common_header = ias_misc_initialize_report_header(
        omf_params.l0r_filename, wo_params.work_order_id,
        "test_write_solarcal_dump_file");
    assert_true(summary.header.common_header != NULL);
    if (!summary.header.common_header)
    {
        assert_true_with_message(FALSE, "Cannot initialize IAS report "
            "header data");
        size = snprintf(system_command, sizeof(system_command), 
            "rm %s.omf %s.odl", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Update the CPF in the header to reflect what is in the OMF file. */
    ias_misc_set_header_cpf_name(summary.header.common_header,
        omf_params.cpf_filename);

    /* Get the RPS-specific header information */
    status = ias_misc_initialize_rps_report_header(omf_params.l0r_filename,
        &(summary.header.rps_header));
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Cannot initialize RPS-specific "
            "report header information");
        ias_misc_free_report_header(summary.header.common_header);
        size = snprintf(system_command, sizeof(system_command), 
            "rm %s.omf %s.odl", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Initialize the angle information for the dump file */
    memset(&angles, 0, sizeof(RPS_SOLARCAL_SOLAR_ANGLES));

    angles.earth_sun_distance_factor = 1.01553;
    angles.incident_solar_zenith       = 105.00;
    angles.incident_solar_zenith_stdev =  0.003;
    angles.view_solar_zenith           = 45.00;
    angles.view_solar_zenith_stdev     =  0.006;
    angles.view_solar_azimuth          =  24.63;
    angles.view_solar_azimuth_stdev    =  0.0057;

    status = initialize_solarcal_report_data(wo_params, omf_params, angles,
         &summary);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Cannot initialize parameter "
            "and/or angle information for summary report");
        ias_misc_free_report_header(summary.header.common_header);
        size = snprintf(system_command, sizeof(system_command), 
            "rm %s.omf %s.odl", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Change the test dataset collection type to the L0Ra DFCB
       value "OLI_SOLAR" if it isn't already */
    collection_type = ias_misc_get_header_collection_type(
        summary.header.common_header);
    if ((collection_type == NULL)
        || (strcmp(collection_type, IAS_RPS_OLI_SOLAR_COLLECT) != 0))
    {
        ias_misc_set_header_collection_type(summary.header.common_header,
                IAS_RPS_OLI_SOLAR_COLLECT);
    }

    /* Gin up additional information for temperatures, etc */
    summary.header.panel_id = IAS_RPS_OLI_DIFFUSER_TYPE_PRISTINE;
    summary.header.rps_header.diffuser_resolver_position =
        IAS_RPS_OLI_DIFFUSER_POSITION_PRISTINE_LOW;

    summary.header.rps_header.oli_fpm7_start_temp  = 210.0;
    summary.header.rps_header.oli_fpm7_stop_temp   = 211.0;
    summary.header.rps_header.oli_fpm14_start_temp = 212.0;
    summary.header.rps_header.oli_fpm14_stop_temp  = 213.0;
    summary.header.rps_header.oli_fpa_window_start_temp  = 205.0;
    summary.header.rps_header.oli_fpa_window_stop_temp   = 206.0;
    summary.header.rps_header.oli_fpe_chassis_start_temp = 207.0;
    summary.header.rps_header.oli_fpe_chassis_stop_temp  = 208.0;
    summary.header.num_frames_averaged   = 100;

    /* Flesh out the band gain information for the summary report */
    for (band_index = 0; band_index < wo_params.num_bands_to_process;
         band_index++)
    {
        /* Clear out the previous band's information */
        memset(&band_gains, 0, sizeof(RPS_SOLARCAL_BAND_GAINS));

        ba = ias_sat_attr_get_band_attributes((band_index + 1));
        if (ba == NULL)
        {
            assert_true_with_message(FALSE, "Cannot get attributes for "
                "band %d", (band_index + 1));
            ias_misc_free_report_header(summary.header.common_header);
            size = snprintf(system_command, sizeof(system_command), 
                "rm %s.omf %s.odl", work_order_id, work_order_id);
            if ((size < 0) || (size >= sizeof(system_command)))
            {
                assert_true_with_message(FALSE, "Command string too short, "
                   "it needs %d bytes", size);
                return;
            }
            system(system_command);
            return;
        }

        strcpy(band_gains.band_name, ba->band_name);
        band_gains.band_number            = ba->band_number;
        band_gains.scas                   = ba->scas;
        band_gains.detectors_per_sca      = ba->detectors_per_sca;
        band_gains.diffuser_radiance_mean = 175.6846;
        band_gains.diffuser_radiance_stdev = 0.000;
        band_gains.hist_stats_mean        = 158.50077;
        band_gains.hist_stats_stdev       = 0.2900;
        band_gains.radiance_gain          = 0.9304;
        band_gains.radiance_gain_stdev    = 8.89e-15;
        band_gains.diffuser_reflectance_mean = 0.2956;
        band_gains.diffuser_reflectance_stdev = 0.000;
        band_gains.reflectance_gain       = 1832.146;
        band_gains.reflectance_gain_stdev = 1.39e-11;

        /* Copy this band's info to the summary report data structure */
        initialize_solarcal_summary_band_data(band_index, band_gains,
            &summary);
 
        /* Allocate the relative gain and relative gain stdev data buffers */
        num_band_detectors = ba->scas * ba->detectors_per_sca;

        relgains = calloc(num_band_detectors, sizeof(*relgains));
        relgains_stdev = calloc(num_band_detectors,
            sizeof(*relgains_stdev));
        if ((relgains == NULL) || (relgains_stdev == NULL))
        {
            assert_true_with_message(FALSE, "Cannot allocate relative "
                "gain data buffers\n");
            free(relgains);
            free(relgains_stdev);
            free(summary.band_gains);
            ias_misc_free_report_header(summary.header.common_header);
            size = snprintf(system_command, sizeof(system_command), 
                "rm %s.omf %s.odl", work_order_id, work_order_id);
            if ((size < 0) || (size >= sizeof(system_command)))
            {
                assert_true_with_message(FALSE, "Command string too short, "
                   "it needs %d bytes", size);
                return;
            }
            system(system_command);
            return;
        }

        /* Allocate the data invalid flags buffer */
        data_invalid_flags = calloc(num_band_detectors, sizeof(int));
        if (data_invalid_flags == NULL)
        {
            assert_true_with_message(FALSE, "cannot allocate data "
                "invalid buffer");
            free(relgains);
            free(relgains_stdev);
            free(summary.band_gains);
            ias_misc_free_report_header(summary.header.common_header);
            size = snprintf(system_command, sizeof(system_command), 
                "rm %s.omf %s.odl", work_order_id, work_order_id);
            if ((size < 0) || (size >= sizeof(system_command)))
            {
                assert_true_with_message(FALSE, "Command string too short, "
                   "it needs %d bytes", size);
                return;
            }
            system(system_command);
            return;
        }

        /* Populate the relative gain data buffers */
        for (detector_index = 0;
             detector_index < num_band_detectors;
             detector_index++)
        {
            relgains[detector_index]       =
                1.0012345 + 0.01 * detector_index;
            relgains_stdev[detector_index] =
                0.000123456 + 0.001 * detector_index;
        }

        /* Create the dump file for the current band */
        size = snprintf(dumpfile_name, sizeof(dumpfile_name), 
            "%s_solarcal_dumpfile_B%02d.txt", wo_params.work_order_id, 
            ba->band_number);
        if ((size < 0) || (size >= sizeof(dumpfile_name)))
        {
            assert_true_with_message(FALSE, "Filename string too short, "
               "it needs %d bytes", size);
            free(relgains);
            free(relgains_stdev);
            free(data_invalid_flags);
            ias_misc_free_report_header(summary.header.common_header);
            free(summary.band_gains);
            return;
        }

        dump_fptr = fopen(dumpfile_name, "w");
        if (dump_fptr == NULL)
        {
            assert_true_with_message(FALSE, "Cannot create dump file %s",
                dumpfile_name);
            free(relgains);
            free(relgains_stdev);
            free(summary.band_gains);
            ias_misc_free_report_header(summary.header.common_header);
            size = snprintf(system_command, sizeof(system_command), 
                "rm %s.omf %s.odl", work_order_id, work_order_id);
            if ((size < 0) || (size >= sizeof(system_command)))
            {
                assert_true_with_message(FALSE, "Command string too short, "
                   "it needs %d bytes", size);
                return;
            }
            system(system_command);
            return;
        }

        /* Set the band number in the report header */
        ias_misc_set_header_band_number(summary.header.common_header,
            ba->band_number);

        /* Write the header information to the dump file */
        status = write_header_to_dump_file(summary.header,
                dump_fptr);
        assert_true(status == SUCCESS);
        if (status != SUCCESS)
        {
            assert_true_with_message(FALSE, "Cannot create dump file %s",
                dumpfile_name);
            free(relgains);
            free(relgains_stdev);
            free(summary.band_gains);
            ias_misc_free_report_header(summary.header.common_header);
            fclose(dump_fptr);
            size = snprintf(system_command, sizeof(system_command), 
                "rm %s.omf %s.odl %s_solarcal_dumpfile_B*.txt", work_order_id,
                work_order_id, work_order_id);
            if ((size < 0) || (size >= sizeof(system_command)))
            {
                assert_true_with_message(FALSE, "Command string too short, "
                   "it needs %d bytes", size);
                return;
            }
            system(system_command);
            return;
        }

        /* Write the angle information to the dump file */
        status = write_angle_statistics_to_dump_file(summary.solar_angles,
            dump_fptr);
        if (status != SUCCESS)
        {
            assert_true_with_message(FALSE, "Cannot write angle "
                "information to dump file %s", dumpfile_name);
            free(relgains);
            free(relgains_stdev);
            free(summary.band_gains);
            free(data_invalid_flags);
            ias_misc_free_report_header(summary.header.common_header);
            fclose(dump_fptr);
            size = snprintf(system_command, sizeof(system_command), 
                "rm %s.omf %s.odl %s_solarcal_dumpfile_B*.txt", work_order_id,
                work_order_id, work_order_id);
            if ((size < 0) || (size >= sizeof(system_command)))
            {
                assert_true_with_message(FALSE, "Command string too short, "
                   "it needs %d bytes", size);
                return;
            }
            system(system_command);
            return;
        }

        /* Write the relative gains to the dump file */
        status = write_relative_gains_to_dump_file(relgains, relgains_stdev,
            data_invalid_flags, ba->band_number, ba->scas,
            num_band_detectors, dump_fptr);
        assert_true(status == SUCCESS);
        if (status != SUCCESS)
        {
            assert_true_with_message(FALSE, "Cannot write band %d "
                "relative gain information to dump file %s",
                ba->band_number, dumpfile_name);
            free(relgains);
            free(relgains_stdev);
            free(summary.band_gains);
            free(data_invalid_flags);
            ias_misc_free_report_header(summary.header.common_header);
            fclose(dump_fptr);
            size = snprintf(system_command, sizeof(system_command), 
                "rm %s.omf %s.odl %s_solarcal_dumpfile_B*.txt", work_order_id,
                work_order_id, work_order_id);
            if ((size < 0) || (size >= sizeof(system_command)))
            {
                assert_true_with_message(FALSE, "Command string too short, "
                   "it needs %d bytes", size);
                return;
            }
            system(system_command);
            return;
        }

        /* Write the band gains to the dump file */
        status = write_band_gains_to_dump_file(
           summary.band_gains[band_index],
           dump_fptr);
        assert_true(status == SUCCESS);
        if (status != SUCCESS)
        {
            assert_true_with_message(FALSE, "Cannot write band %d "
                "radiance/reflectance gain information to dump file %s",
                ba->band_number, dumpfile_name);
        }

        /* Close the file */
        fclose(dump_fptr);

        /* Clean up */
        free(relgains);
        free(relgains_stdev);
        free(data_invalid_flags);
    }

    /* Free up unneeded resources */
    free(summary.band_gains);
    ias_misc_free_report_header(summary.header.common_header);

    /* Run a diff between the test dump files and expected dump files */
    for (band_index = 0; band_index < wo_params.num_bands_to_process;
         band_index++)
    {
        status = snprintf(dumpfile_name, sizeof(dumpfile_name),
            "%s_solarcal_dumpfile_B%02d.txt", work_order_id,
            band_list[band_index]);
        if ((status < 0) || (status >= sizeof(dumpfile_name)))
        {
            assert_true_with_message(FALSE, "Error creating test dumpfile "
                "name for band number %d", band_list[band_index]);
            return;
        }
        status = snprintf(expected_dumpfile_name,
            sizeof(expected_dumpfile_name),
            "%s/%s/%s_expected_solarcal_dumpfile_B%02d.txt",
            test_data_path, test_data_dir, work_order_id,
            band_list[band_index]);
        if ((status < 0) || (status >= sizeof(expected_dumpfile_name)))
        {
            assert_true_with_message(FALSE, "Error creating expected "
                "dumpfile name for band %d", band_list[band_index]);
            return;
        }

        status = snprintf(system_command, sizeof(system_command),
            "sed '/Spacecraft:/,$!'d %s | diff -q - %s",
            dumpfile_name, expected_dumpfile_name);
        if ((status < 0) || (status >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Error preparing summary "
                "band %d dump file comparison system command",
                band_list[band_index]);
            return;
        }
        status = system(system_command);
        assert_equal_with_message(status, 0, "Error comparing reference and "
            "test dump files for band %d", band_list[band_index]);
    }
}

void solarcal_oli_write_reports_test(TestSuite *suite)
{
    add_test(suite, test_write_solarcal_summary_report);
    add_test(suite, test_write_solarcal_dump_file);
}
