/**************************************************************************
 NAME:                 validate_collect_data

 PURPOSE:       Routines to read information from the L0R metadata,
                L0R ancillary, and L0R frame header data groups,
                and use that information to performs simple "sanity
                checks" to determine whether the collect is "good"

 RETURNS:       Status code of SUCCESS or ERROR

***************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "ias_l0r.h"
#include "ias_const.h"
#include "ias_radiometry.h"
#include "solarcal_oli.h"


/* As defined in LDCM-DFCB-002 L0Ra DFCB V2.0b */
#define RPS_SOLARCAL_VALID_SPACECRAFT_ID      "LANDSAT_8"
#define RPS_SOLARCAL_VALID_SENSOR_ID          "OLI"

#define DEGREES_PER_RESOLVER_COUNT            0.2


/*-----------------------------------------------------------------------
 NAME:     convert_diffuser_resolver_position_to_angle

 PURPOSE:  Converts the diffuser_resolver_position telemetry value for
           the collect to a deployment angle value that is within the range
           [0...360] degrees

 RETURNS:  Nothing

 NOTES:    It is assumed that the input diffuser_resolver_position telemetry
           lies within the expected range of values for the panel
           deployed during the collect.
-------------------------------------------------------------------------*/
void convert_diffuser_resolver_position_to_angle
(
    uint16_t diffuser_resolver_position,  /* I:  Diffuser position
                                             telemetry value */
    double *final_angle                   /* O:  Final deployment angle
                                             between 0 and 360 degrees */
)
{
    /* Calculate the 'raw' angle */
    double raw_angle =
        DEGREES_PER_RESOLVER_COUNT * diffuser_resolver_position;

    /* The computed angle can be greater than 360 degrees.  If it is, adjust
       it so that it is within [0 ... 360] degrees */
    double nearest_integer_raw_angle = floor(raw_angle);
    int    integer_raw_angle = (int)nearest_integer_raw_angle;
    int    number_complete_revs = integer_raw_angle / 360;

    *final_angle = raw_angle - (double)(number_complete_revs * 360);
}   /* END -- convert_diffuser_resolver_position_to_angle */



/*-----------------------------------------------------------------------
 NAME:      validate_image_header

 PURPOSE:   Checks the L0R image header information for values consistent
            with an OLI solar diffuser data collect

 RETURNS:   Integer status code indicating the following conditions:
            0:   image header successfully validated
            1.   invalid integration time (should be nominal integration
                 time for "normal" and "60-second" solar collects)

 NOTES:     So far, the only check is with respect to the collect's
            integration time against the nominal MS/PAN integration
            times given in the OLI_PARAMETERS CPF group.  If other
            information in the image header needs to be considered
            for validation, the appropriate item(s) can be added later
------------------------------------------------------------------------ */
int validate_image_header
(
    RPS_SOLARCAL_REPORT_HEADER header,        /* I: Report header data with
                                                 image header information
                                                 to validate */
    double cpf_integration_time_ms,           /* I: OLI nominal integration
                                                 time for MS bands from CPF */
    double cpf_integration_time_pan           /* I: OLI nominal integration
                                                 time for PAN band from CPF */
)
{
    /* Check the integration times against the nominal values from the
       CPF.  The header times are in integer microseconds, so mutliply
       these by 0.001 to convert them to (floating point) milliseconds */
    if ((0.001 * header.rps_header.ms_integration_time
            != cpf_integration_time_ms))
    {
        IAS_LOG_ERROR("Invalid MS integration time %lf, should be %lf",
            0.001 * header.rps_header.ms_integration_time,
            cpf_integration_time_ms);
        return RPS_SOLARCAL_IMAGEHEADER_INVALID_INTEGRATION_TIME;
    }
    if ((0.001 * header.rps_header.pan_integration_time
                != cpf_integration_time_pan))
    {
        IAS_LOG_ERROR("Invalid PAN integration time(s) %lf, should be %lf",
            0.001 * header.rps_header.pan_integration_time,
            cpf_integration_time_pan);
        return RPS_SOLARCAL_IMAGEHEADER_INVALID_INTEGRATION_TIME;
    }

    /* Add other tests here as appropriate */

    /* Image header data is valid */
    IAS_LOG_DEBUG("L0R image header information validated");
    return RPS_SOLARCAL_IMAGEHEADER_VALIDATED;
}




/*-----------------------------------------------------------------------
 NAME:      validate_interval_metadata

 PURPOSE:   Checks the L0R interval metadata for values consistent with
            an OLI solar diffuser data collect

 RETURNS:   Integer status code indicating the following conditions:
            0:   metadata successfully validated
            1:   invalid spacecraft ID (should be "LANDSAT_8")
            2:   invalid sensor        (should be "OLI")
            3:   invalid WRS path/row  (should be 0 for solar collects)
 ------------------------------------------------------------------------ */
int validate_interval_metadata
(
    RPS_SOLARCAL_REPORT_HEADER header     /* I: Report header data with
                                             interval metadata information
                                             to validate */
)
{
    int starting_path;                    /* First WRS2 path */
    int starting_row;                     /* First WRS2 row */
    int ending_row;                       /* Last WRS2 row */


    /* Get the spacecraft and sensor IDs from the report header */
    char *spacecraft_id =
        ias_misc_get_header_spacecraft_id(header.common_header);
    char *sensor_id = ias_misc_get_header_sensor_id(header.common_header);

    /* Make sure we have valid spacecraft and sensor IDs */
    if (spacecraft_id == NULL)
    {
        IAS_LOG_ERROR("Spacecraft has not been identified in the report "
            "header");
        return ERROR;
    }
    if ((strcmp(spacecraft_id, RPS_SOLARCAL_VALID_SPACECRAFT_ID) != 0))
    {
        IAS_LOG_ERROR("Invalid spacecraft identification: %s",
            spacecraft_id);
        return RPS_SOLARCAL_METADATA_INVALID_SPACECRAFT;
    }

    if (sensor_id == NULL)
    {
        IAS_LOG_ERROR("Sensor has not been identified in the report "
            "header");
        return ERROR;
    }
    if (strcmp(sensor_id, RPS_SOLARCAL_VALID_SENSOR_ID) != 0)
    {
        IAS_LOG_ERROR("Invalid sensor identification: %s",
            sensor_id);
        return RPS_SOLARCAL_METADATA_INVALID_SENSOR;
    }

    /* Check the target path/row information.  For solar collects, these
       should be 0 */
    ias_misc_get_header_starting_ending_pathrow(
        header.common_header, &starting_path, &starting_row,
        &ending_row);
    if ((starting_path != 0) || (starting_row != 0) || (ending_row != 0))
    {
        IAS_LOG_ERROR("Invalid starting path/row identification:  "
            "(%d, %d, %d)", starting_path, starting_row, ending_row);
        return RPS_SOLARCAL_METADATA_INVALID_WRS_PATHROW;
    }

    /* Things aren't obviously out-of-whack, so assume the metadata
       is consistent with what we'd expect for a solar-related
       collect */
    IAS_LOG_DEBUG("L0R interval metadata information validated");
    return RPS_SOLARCAL_METADATA_VALIDATED;
}




/*------------------------------------------------------------------------
  NAME:      validate_collect_data

  PURPOSE:   Reads the L0R interval metadata and image header information
             contained in the report header to ensure the collect is a
             valid OLI solar collect

  RETURNS:   Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int validate_collect_data
(
    RPS_SOLARCAL_REPORT_HEADER header,  /* I: Report header information with
                                           required data to validate */
    const struct IAS_CPF_OLI_PARAMETERS *oli_params
                                        /* I: OLI parameters containing the
                                           nominal integration times in
                                           milliseconds */
)
{
    int validation_status;                 /* Status for each set of
                                              validations */

    char *collection_type = NULL;          /* Type of data collect */

    int is_solar_collect = FALSE;          /* Solar collect flag */


    /* First, make sure we have a "valid" collection type, which has to
       be "OLI SOLAR" or "OLI SOLAR INTEGRATION TIME SWEEP" */
    collection_type = ias_misc_get_header_collection_type(
        header.common_header);
    if (collection_type == NULL)
    {
        IAS_LOG_ERROR("Collection type has not been identified in the "
            "report header");
        return ERROR;
    }
    if ((strcmp(collection_type, IAS_RPS_OLI_SOLAR_COLLECT) == 0)
        || (strcmp(collection_type, IAS_RPS_OLI_SOLAR_TIMESWEEP_COLLECT)
                == 0))
    {
        is_solar_collect = TRUE;
    }
    if (!is_solar_collect)
    {
        IAS_LOG_ERROR("Collection type %s is not a valid solar collect",
            collection_type);
        return ERROR;
    }

    /* The collect type is validated, now validate the L0R image header
       information*/
    validation_status = validate_image_header(header,
        oli_params->nominal_integration_time_ms,
        oli_params->nominal_integration_time_pan);
    if (validation_status != RPS_SOLARCAL_IMAGEHEADER_VALIDATED)
        return ERROR;

    /* Now validate the L0R metadata information */
    validation_status = validate_interval_metadata(header);
    if (validation_status != RPS_SOLARCAL_METADATA_VALIDATED)
        return ERROR;

    /* Check for a 'valid' panel identification by confirming that the
       observed diffuser resolver position is within the expected range
       of values for the panel deployed during the collect.  */
    if (header.panel_id == IAS_RPS_OLI_DIFFUSER_TYPE_PRISTINE)
    {
       if ((header.diffuser_resolver_position
               < IAS_RPS_OLI_DIFFUSER_POSITION_PRISTINE_LOW)
           || (header.diffuser_resolver_position
               > IAS_RPS_OLI_DIFFUSER_POSITION_PRISTINE_HIGH))
       {
           IAS_LOG_ERROR("Panel identified as %s, but given "
               "diffuser_resolver_position %u isn't in the expected range "
               "of %u to %u", IAS_RPS_OLI_DIFFUSER_PRISTINE,
               header.diffuser_resolver_position,
               IAS_RPS_OLI_DIFFUSER_POSITION_PRISTINE_LOW,
               IAS_RPS_OLI_DIFFUSER_POSITION_PRISTINE_HIGH);
           return ERROR;
       }
   }
   else if (header.panel_id == IAS_RPS_OLI_DIFFUSER_TYPE_WORKING)
   {
       if ((header.diffuser_resolver_position
               < IAS_RPS_OLI_DIFFUSER_POSITION_WORKING_LOW)
           || (header.diffuser_resolver_position
               > IAS_RPS_OLI_DIFFUSER_POSITION_WORKING_HIGH))
       {
           IAS_LOG_ERROR("Panel identified as %s, but given "
               "diffuser_resolver_position %u isn't in the expected range "
               "of %u to %u", IAS_RPS_OLI_DIFFUSER_WORKING,
               header.diffuser_resolver_position,
               IAS_RPS_OLI_DIFFUSER_POSITION_WORKING_LOW,
               IAS_RPS_OLI_DIFFUSER_POSITION_WORKING_HIGH);
           return ERROR;
       }
   }
   else  /* Shouldn't ever get here if ias_misc_identify_diffuser_panel()
            is working properly, unless there are memory corruption issues
            trashing the report header data */
   {
       IAS_LOG_ERROR("Unknown/invalid panel identifier %d",
           header.panel_id);
       return ERROR;
   }

    /* All our checks have passed, so we're finished */
    return SUCCESS;
}
