/**************************************************************************
 NAME:              write_solarcal_dump_file

 PURPOSE:           Module implementing routine(s) to create and write one
                    dump data file for detector-level and band-level
                    solarcal_oli results

***************************************************************************/
#include <math.h>
#include "solarcal_oli.h"


/*------------------------------------------------------------------------
 NAME:            write_angle_statistics_to_dump_file

 PURPOSE:  Writes relative gains and stdevs to the dump file

 RETURNS:  Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int write_angle_statistics_to_dump_file
(
    RPS_SOLARCAL_SOLAR_ANGLES angles,     /* I: Angle statistics */
    FILE *dump_fptr                       /* I: Dump file handle */
)
{
    int status;                           /* Function return status */


    /* Write the angle information.  Earth-sun distance factor goes
       first.  */
    status = fprintf(dump_fptr, "Earth-Sun Distance (AU):\t%8.6lf\n",
        angles.earth_sun_distance_factor);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing Earth-Sun distance to dump file");
        return ERROR;
    }

    /* Write the summary angle statistics to the dump file. */
    status = fprintf(dump_fptr, "Incident Solar Zenith Angle (DEG):  "
        "%9.4lf +- %9.4lf\n", angles.incident_solar_zenith,
        angles.incident_solar_zenith_stdev);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing incident solar zenith angle statistics to "
            "dump file");
        return ERROR;
    }

    status = fprintf(dump_fptr, "View Solar Zenith Angle (DEG):      "
        "%9.4lf +- %9.4lf\n", angles.view_solar_zenith,
        angles.view_solar_zenith_stdev);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing solar azimuth angle statistics to dump file");
        return ERROR;
    }

    status = fprintf(dump_fptr, "View Solar Azimuth Angle (DEG):     "
        "%9.4lf +- %9.4lf\n", angles.view_solar_azimuth,
        angles.view_solar_azimuth_stdev);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing solar azimuth angle statistics to dump file");
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}




/*------------------------------------------------------------------------
 NAME:                 write_header_to_dump_file

 PURPOSE:    Writes IAS/RPS report "header" information to a dump file

 RETURNS:    Integer status code of SUCCESS or ERROR

 NOTES:      The intent of this routine is to provide the same header
             information as contained in the summary report, in the same
             order
--------------------------------------------------------------------------*/
int write_header_to_dump_file
(
    RPS_SOLARCAL_REPORT_HEADER report_header,   /* I: IAS/RPS report header
                                                   information */
    FILE *dump_fptr                             /* I: Dump file handle */
)
{
    char timestamp_string[IAS_L0R_DATE_LENGTH]; /* IAS timestamp */

    /* Pointers to strings contained in the "common" IAS report header */
    char *collection_type = NULL;
    char *cpf_name = NULL;
    char *ias_sw_version = NULL;
    char *ingest_sw_version = NULL;
    char *sensor_id = NULL;
    char *spacecraft_id = NULL;
    char *l0r_filename = NULL;
    char *work_order_id = NULL;
    int status;                                 /* Function return status */
    int band_number;                            /* Band number recorded in
                                                   report header */
    int  starting_path;
    int  starting_row;
    int  ending_row;


    /* Write dump file "title" */
    status = fprintf(dump_fptr, "OLI Solar Calibration Dump File\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing title to dump file");
        return ERROR;
    }

    /* Write the timestamp string to the dump file */
    status = ias_misc_get_timestamp("%c", IAS_L0R_DATE_LENGTH,
            timestamp_string);

    if (status != SUCCESS)
    {
        IAS_LOG_WARNING("Cannot obtain current date/time stamp for dump "
            "file");
        status = fprintf(dump_fptr, "Report Date:\t\t\tN/A\n");
    }
    else
    {
        status = fprintf(dump_fptr, "Report Date:\t\t\t%23s\n",
            timestamp_string);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing timestamp information to dump file");
        return ERROR;
    }

    /* Band number goes here */
    band_number = ias_misc_get_header_band_number(
        report_header.common_header);
    if (band_number < 0)
    {
        status = fprintf(dump_fptr, "Band:\t\t\t\tN/A\n");
    }
    else
    {
        status = fprintf(dump_fptr, "Band:\t\t\t\t%d\n", band_number);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing band number to dump file");
        return ERROR;
    }

    /* Software version information goes here */
    work_order_id
        = ias_misc_get_header_work_order_id(report_header.common_header);
    ias_sw_version
        = ias_misc_get_header_ias_sw_version(report_header.common_header);
    ingest_sw_version
        = ias_misc_get_header_ingest_sw_version(report_header.common_header);

    if (work_order_id == NULL)
    {
        status = fprintf(dump_fptr, "Work Order ID:\t\t\tN/A\n");
    }
    else
    {
        status = fprintf(dump_fptr, "Work Order ID:\t\t\t%s\n",
            work_order_id);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing work order ID to dump file");
        return ERROR;
    }

    if (ias_sw_version == NULL)
    {
        status = fprintf(dump_fptr, "IAS Version:\t\t\tN/A\n");
    }
    else
    {
        status = fprintf(dump_fptr, "IAS Version:\t\t\t%s\n",
            ias_sw_version);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing IAS software version to dump file");
        return ERROR;
    }

    if (ingest_sw_version == NULL)
    {
        status = fprintf(dump_fptr, "Ingest Version:\t\t\tN/A\n");
    }
    else
    {
        status = fprintf(dump_fptr, "Ingest Version:\t\t\t%s\n",
            ingest_sw_version);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing INGEST software version to dump file");
        return ERROR;
    }

    /* L0R Reference image information goes here */
    l0r_filename
        = ias_misc_get_header_l0r_filename(report_header.common_header);
    if (l0r_filename == NULL)
    {
        status = fprintf(dump_fptr, "L0R Image:\t\t\tN/A\n");
    }
    else
    {
        status = fprintf(dump_fptr, "L0R Image:\t\t\t%s\n",
                         l0r_filename);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing L0R file name to dump file");
        return ERROR;
    }

    /* WRS starting path/row information */
    ias_misc_get_header_starting_ending_pathrow(report_header.common_header,
        &starting_path, &starting_row, &ending_row);
    status = fprintf(dump_fptr, "Starting WRS Path:\t\t%03d\n"
        "Starting WRS Row:\t\t%03d\nEnding WRS Row:\t\t\t%03d\n",
        starting_path, starting_row, ending_row);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing starting/ending WRS path/row information");
        return ERROR;
    }

    /* Acquisition date */
    status = fprintf(dump_fptr, "Acquisition Date:\t\t%26s\n",
        ias_misc_get_header_date_acquired(report_header.common_header));
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing acquisition date information to dump file");
        return ERROR;
    }

    /* Spacecraft/sensor IDs go here */
    spacecraft_id = ias_misc_get_header_spacecraft_id(
        report_header.common_header);
    sensor_id = ias_misc_get_header_sensor_id(report_header.common_header);

    if (spacecraft_id == NULL)
    {
        status = fprintf(dump_fptr, "Spacecraft:\t\t\tN/A\n");
    }
    else
    {
        status = fprintf(dump_fptr, "Spacecraft:\t\t\t%s\n", spacecraft_id);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing spacecraft ID to dump file");
        return ERROR;
    }

    if (sensor_id == NULL)
    {
        status = fprintf(dump_fptr, "Sensor:\t\t\t\tN/A\n");
    }
    else
    {
        if (strcmp(sensor_id, "OLI") == 0)
        {
            status = fprintf(dump_fptr, "Sensor:\t\t\t\t%s\n", sensor_id);
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing sensor ID to dump file");
                return ERROR;
            }
        }
        else
        {
            IAS_LOG_ERROR("Solar calibration is currently run only for "
                "OLI solar collects");
            return ERROR;
        }
    }

    /* Collection type information next */
    collection_type =
        ias_misc_get_header_collection_type(report_header.common_header);
    if (collection_type == NULL)
    {
        status = fprintf(dump_fptr, "Collection Type:\t\tN/A\n");
    }
    else
    {
        if ((strcmp(collection_type, IAS_RPS_OLI_SOLAR_COLLECT) == 0)
            || (strcmp(collection_type,
                       IAS_RPS_OLI_SOLAR_TIMESWEEP_COLLECT) == 0))
        {
            status = fprintf(dump_fptr, "Collection Type:\t\t%s\n",
                collection_type);
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing collection type to dump file");
                return ERROR;
            }
        }
        else
        {
            IAS_LOG_ERROR("Invalid collection type '%s', should be "
                "'%s' or '%s'", collection_type, IAS_RPS_OLI_SOLAR_COLLECT,
                IAS_RPS_OLI_SOLAR_TIMESWEEP_COLLECT);
            return ERROR;
        }
    }

    /* Processing CPF */
    cpf_name = ias_misc_get_header_cpf_name(report_header.common_header);
    if (cpf_name == NULL)
    {
        status = fprintf(dump_fptr, "Processing CPF:\t\t\tN/A\n");
    }
    else
    {
        status = fprintf(dump_fptr, "Processing CPF:\t\t\t%s\n", cpf_name);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Formatting processing CPF string");
        return ERROR;
    }

    /* Image start/stop times */
    if (strcmp(sensor_id, "OLI") != 0)
    {
        IAS_LOG_ERROR("Solar calibration is currently run only on OLI "
            "solar/solar integration sweep collects");
        return ERROR;
    }
    status = fprintf(dump_fptr, "Image Start Time:\t\t%26s\n"
        "Image Stop Time:\t\t%26s\n",
        report_header.rps_header.acq_datetime.start_time_oli,
        report_header.rps_header.acq_datetime.stop_time_oli);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing image start time and stop times to dump file");
        return ERROR;
    }

    /* Integration times/selected pixel map/image data truncation setting */
    status = fprintf(dump_fptr, "MS Band Integ. Time:\t\t%4d\n"
        "Pan Band Integ. Time:\t\t%-4d\nDetector Select Table ID:\t%d\n"
        "Image Data Truncation Setting:\t%d\n",
        report_header.rps_header.ms_integration_time,
        report_header.rps_header.pan_integration_time,
        report_header.rps_header.current_detector_select_table,
        report_header.rps_header.image_data_truncation_setting);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing integration time/pixel map "
                      "information to dump file");
        return ERROR;
    }

    /* Focal plane temperatures */
    status = fprintf(dump_fptr, "FPM07 Start Temp.:\t\t%7.4lf\n"
        "FPM07 Stop Temp.:\t\t%7.4lf\nFPM14 Start Temp.:\t\t%7.4lf\n"
        "FPM14 Stop Temp:\t\t%7.4lf\nFPA Window Start Temp.:\t\t%7.4lf\n"
        "FPA Window Stop Temp.:\t\t%7.4lf\nFPE Chassis STart Temp.:"
        "\t%7.4lf\nFPE Chassis Stop Temp:\t\t%7.4lf\n",
        report_header.rps_header.oli_fpm7_start_temp,
        report_header.rps_header.oli_fpm7_stop_temp,
        report_header.rps_header.oli_fpm14_start_temp,
        report_header.rps_header.oli_fpm14_stop_temp,
        report_header.rps_header.oli_fpa_window_start_temp,
        report_header.rps_header.oli_fpa_window_stop_temp,
        report_header.rps_header.oli_fpe_chassis_start_temp,
        report_header.rps_header.oli_fpe_chassis_stop_temp);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing temperature information to dump file");
        return ERROR;
    }

    /* Characterization information next */
    if (report_header.panel_id == IAS_RPS_OLI_DIFFUSER_TYPE_PRISTINE)
    {
        status = fprintf(dump_fptr, "Diffuser Panel Deployed:\t%s\n",
            IAS_RPS_OLI_DIFFUSER_PRISTINE);
    }
    else
    {
        status = fprintf(dump_fptr, "Diffuser Panel Deployed:\t%s\n",
            IAS_RPS_OLI_DIFFUSER_WORKING);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing characterization information to dump file");
        return ERROR;
    }
    status = fprintf(dump_fptr, "Diffuser Resolver Position:\t%u\n",
        report_header.diffuser_resolver_position);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing diffuser position information to dump file");
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}




/*------------------------------------------------------------------------
 NAME:            write_relative_gains_to_dump_file

 PURPOSE:  Writes relative gains and stdevs to the dump file

 RETURNS:  Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int write_relative_gains_to_dump_file
(
    double *relative_gains,           /* I: Relative gains for the current
                                         band */
    double *relative_gains_stdev,     /* I: Relative gain stdevs for the
                                         current band */
    const int *data_invalid_flags,    /* I: Data invalid flags for all
                                         detectors in the current band
                                         (0=valid data, nonzero=invalid
                                         data) */
    int  band_number,                 /* I: Current 1-based band number */
    int  num_scas,                    /* I: Number of SCAs in the current
                                         band */
    int  num_band_detectors,          /* I: Total number of detectors
                                         in the band */
    FILE *dump_fptr                   /* I: Dump file handle */
)
{
    int status;
    int sca_index;
    int num_sca_detectors;


    /* Print the header */
    status = fprintf(dump_fptr, "BAND %d RELATIVE GAINS:\nSCA    DET    "
        " RELGAIN   RELGAIN STD\n-------------------------------------\n",
        band_number);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing relative gain header information to dump "
            "file");
        return ERROR;
    }

    /* Determine the number of detectors in an SCA based on the number of
       detectors in the band and the number of SCAs */
    num_sca_detectors = num_band_detectors / num_scas;

    /* Write the relative gain data.  Even though the relative gains are
       calculated per-band, include the current SCA number as a "point of
       "reference" */
    for (sca_index = 0; sca_index < num_scas; sca_index++)
    {
        int sca_number = (sca_index + 1);
        int detector_index;

        for (detector_index = 0; detector_index < num_sca_detectors;
             detector_index++)
        {
            int band_detector_index
                = sca_index * num_sca_detectors + detector_index;
            if (data_invalid_flags[band_detector_index])
            {
                status = fprintf(dump_fptr, "%03d    %03d     %8.6lf**  "
                    "%8.6lf**\n", sca_number, (detector_index + 1),
                    relative_gains[band_detector_index],
                    relative_gains_stdev[band_detector_index]);
            }
            else
            {
                status = fprintf(dump_fptr, "%03d    %03d     %8.6lf    "
                    "%8.6lf\n", sca_number, (detector_index + 1),
                    relative_gains[band_detector_index],
                    relative_gains_stdev[band_detector_index]);
            }
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing band %d relative gain information to "
                    "dump file", band_number);
                return ERROR;
            }
        }
    }

    /* Finished */
    return SUCCESS;
}




/*------------------------------------------------------------------------
 NAME:            write_band_gains_to_dump_file

 PURPOSE:  Writes radiance-based and reflectance-based band average gains
           (and associated stdevs) for the current band to the dump file

 RETURNS:  Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int write_band_gains_to_dump_file
(
    RPS_SOLARCAL_BAND_GAINS band_gains,     /* I: Summary gain statistics
                                               for the current band */
    FILE *dump_fptr                         /* I: Dump file handle */
)
{
    int status;


    /* Write the data headers for the band summary statistics, then write
       the data */
    status = fprintf(dump_fptr, "\nBAND %d SUMMARY STATISTICS:\n\n",
        band_gains.band_number);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing band summary stats header to dump file\n");
        return ERROR;
    }
    status = fprintf(dump_fptr, "   MEAN DN        STD DEV\n-----------"
        "----------------\n%13.6lf   %11.6lf\n\n\n",
        band_gains.hist_stats_mean, band_gains.hist_stats_stdev);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing band summary histogram statistics to dump "
            "file");
        return ERROR;
    }
    status = fprintf(dump_fptr, "   DIFF RAD      STD DEV   DIFF REFL"
        "   STD DEV\n-----------------------------------------------\n"
        "%13.6lf %11.6lf  %8.6lf    %8.6lf\n",
        band_gains.diffuser_radiance_mean,
        band_gains.diffuser_radiance_stdev,
        band_gains.diffuser_reflectance_mean,
        band_gains.diffuser_reflectance_stdev);
    if (status < 0)
    {
       IAS_LOG_ERROR("Writing band diffuser summary statistics to "
           "dump file");
       return ERROR;
    }
    status = fprintf(dump_fptr, "\n\n   RAD GAIN      STD DEV     REFL GAIN"
        "     STD DEV\n----------------------------------------------"
        "-------\n%12.6lf  %11.6lf   %12.6lf  %11.6lf\n\n\n\n",
        band_gains.radiance_gain, band_gains.radiance_gain_stdev,
        band_gains.reflectance_gain, band_gains.reflectance_gain_stdev);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing band %d absolute gains to dump file",
                      band_gains.band_number);
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}
