/**************************************************************************
 NAME:              write_solarcal_summary_report

 PURPOSE:           Module implementing routines to write band-specific
                    optional summary reports for detector response
                    characterization of the OLI solar diffuser

***************************************************************************/
#include <math.h>
#include "ias_satellite_attributes.h"
#include "solarcal_oli.h"


/*------------------------------------------------------------------------
 NAME:            write_header_to_summary_report

 PURPOSE:  Writes "header" information to summary report file(s)

 RETURNS:  Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int write_header_to_summary_report
(
    RPS_SOLARCAL_REPORT_HEADER header,    /* I: IAS/RPS report header
                                             information */
    FILE *rpt_fptr                        /* I: Report file handle */
)
{
    int status;


    char *collection_type = NULL;         /* Pointer to collection type 
                                             string */
    char *cpf_name = NULL;                /* Pointer to CPF name string */
    char *sensor_id = NULL;               /* Pointer to sensor ID string */
    double deployment_angle;              /* Diffuser panel deployment
                                             angle */



    /* Reset the band information here.  The band number gets set when
       this information is written to the dump file */
    ias_misc_set_header_band_number(header.common_header, -1);

    /* Standard IAS report header information here */
    status = ias_misc_write_report_header(rpt_fptr,
        header.common_header);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing common IAS header information to report "
                      "file");
        return ERROR;
    }

    /* Write the RPS-specific stuff here */
    collection_type = ias_misc_get_header_collection_type(
        header.common_header);
    cpf_name = ias_misc_get_header_cpf_name(header.common_header);
    sensor_id = ias_misc_get_header_sensor_id(header.common_header);

    status = ias_misc_write_rps_report_header(collection_type, cpf_name,
        sensor_id, header.rps_header, rpt_fptr);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing RPS-specific header information to report "
            "file");
        return ERROR;
    }

    /* Diffuser panel information goes here */
    status = fprintf(rpt_fptr, "\nDiffuser Panel Summary Information"
        "\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing diffuser panel header string to report");
        return ERROR;
    }
    if (header.panel_id == IAS_RPS_OLI_DIFFUSER_TYPE_PRISTINE)
    {
        status = fprintf(rpt_fptr, "Diffuser Panel:                     "
            "%s\n", IAS_RPS_OLI_DIFFUSER_PRISTINE);
    }
    else
    {
        status = fprintf(rpt_fptr, "Diffuser Panel:                     "
            "%s\n", IAS_RPS_OLI_DIFFUSER_WORKING);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing diffuser panel type string to report");
        return ERROR;
    }

    /* Calculate the deployment angle */
    convert_diffuser_resolver_position_to_angle(
        header.diffuser_resolver_position, &deployment_angle);

    status = fprintf(rpt_fptr, "Diffuser Panel Resolver Position:   "
         "%u\n", header.diffuser_resolver_position);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing diffuser resolver position information "
            "to report");
        return ERROR;
    }
    status = fprintf(rpt_fptr, "Diffuser Panel Deployment Angle:    "
        "%6.2lf DEG\n\n\n", deployment_angle);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing diffuser panel deployment angle to "
            "report");
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}




/*------------------------------------------------------------------------
 NAME:            write_angle_statistics_to_summary_report

 PURPOSE:  Writes solar angle information to summary report file

 RETURNS:  Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int write_angle_statistics_to_summary_report
(
    RPS_SOLARCAL_SOLAR_ANGLES angles,            /* I: Summary angle
                                                    statistics */
    FILE *rpt_fptr                               /* I: Report file handle */
)
{
    int status;


    /* Angle information goes here */
    status = fprintf(rpt_fptr, "Angle Statistics\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing angle statistics header to report");
        return ERROR;
    }

    /* Earth-sun distance goes here */
    status = fprintf(rpt_fptr, "Earth-Sun Distance:   %10.6lf AU\n",
        angles.earth_sun_distance_factor);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing Earth-sun distance factor to report");
        return ERROR;
    }

   /*  Summary view angle statistics go here.  The incident solar zenith
       angle is expected to be just about 45 degrees give-or-take; the
       view solar zenith is expected to be at 45 degrees.  Their displayed
       precision can be set to 2 digits.

       If the standard deviations of the incident and/or view solar zenith
       angles are less than 1.0 degree, extend their displayed precision to
       4 digits, otherwise, set their displayed precision to 2 digits.  */
    status = fprintf(rpt_fptr, "Incident Solar Zenith:     "
        "%9.4lf +- %9.4lf DEG\n", angles.incident_solar_zenith,
        angles.incident_solar_zenith_stdev);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing incident zenith angle results to report");
        return ERROR;
    }

    status = fprintf(rpt_fptr, "View Solar Zenith:         "
        "%9.4lf +- %9.4lf DEG\n", angles.view_solar_zenith,
        angles.view_solar_zenith_stdev);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing view solar zenith angle results to report");
        return ERROR;
    }

    status = fprintf(rpt_fptr, "View Solar Azimuth:        "
        "%9.4lf +- %9.4lf DEG\n", angles.view_solar_azimuth,
        angles.view_solar_azimuth_stdev);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing view solar azimuth angle results to report");
        return ERROR;
    }

    /* Advance two lines */
    status = fprintf(rpt_fptr, "\n\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing line advances to report");
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}




/*------------------------------------------------------------------------
 NAME:            write_solarcal_report_data

 PURPOSE:   Writes each band's summary characterization results to
            an ASCII text file

 RETURNS:   Integer status code of SUCCESS or ERROR
-------------------------------------------------------------------------*/
int write_solarcal_report_data
(
    int  num_oli_bands,                    /* I: Total number of OLI
                                              bands */
    RPS_SOLARCAL_REPORT_SUMMARY summary,   /* I: Summary report data */
    FILE *rpt_fptr                         /* I/O: Report file handle */
)
{
    int index;                             /* Band loop counter */
    int status;


    /* Write the separator string */
    status = fprintf(rpt_fptr, "==================================="
                     "========================================\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing separator string to report file");
        return ERROR;
    }

    /* Write a CHARACTERIZATION RESULTS title */
    status = fprintf(rpt_fptr, "                      "
                     "CHARACTERIZATION RESULTS\n\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing results section title to report");
        return ERROR;
    }

    /* Write the image statistics first */
    status = fprintf(rpt_fptr, "Bias Corrected/Linearized Histogram "
        "Statistics\n\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing histogram statistics header to report");
        return ERROR;
    }

    status = fprintf(rpt_fptr, "Number Frames Averaged:   %d\n",
                     summary.header.num_frames_averaged);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing histogram statistics number of frames to "
            "report");
        return ERROR;
    }

    status = fprintf(rpt_fptr, "Band                          Mean       "
    " Std Dev\n                              (DN)         (DN)\n---------"
    "-----------------------------------------\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing histogram statistics header to report");
        return ERROR;
    }
    for (index = 0; index < num_oli_bands; index++)
    {
        /* If a band hasn't been characterized, skip it */
        if (summary.band_gains[index].band_number == 0)
        {
            continue;
        }

        /* Write the results */
        status = fprintf(rpt_fptr, "%02d (%-19s) %13.6lf %11.6lf\n",
            summary.band_gains[index].band_number,
            summary.band_gains[index].band_name,
            summary.band_gains[index].hist_stats_mean,
            summary.band_gains[index].hist_stats_stdev);
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing band %d image statistics to report",
                          summary.band_gains[index].band_number);
            return ERROR;
        }
    }

    /* Write the band diffuser radiances and reflectances */
    status = fprintf(rpt_fptr, "\n\n\nDiffuser Radiances and Reflectances ("
        "Radiances in units of W / [m^2 sr um])\n\nBand                   "
        "        (Diffuser Radiance)   (Diffuser Reflectance)\n          "
        "                       Mean      Std Dev      Mean    Std Dev\n-"
        "----------------------------------------------------------------"
        "-----\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing diffuser radiance statistics header to "
            "report");
        return ERROR;
    }
    for (index = 0; index < num_oli_bands; index++)
    {
        /* If a band hasn't been characterized, skip it */
        if (summary.band_gains[index].band_number == 0)
        {
            continue;
        }

        status = fprintf(rpt_fptr, "%02d (%-19s)  %13.6lf %11.6lf  %8.6lf "
            "%8.6lf\n",
            summary.band_gains[index].band_number,
            summary.band_gains[index].band_name,
            summary.band_gains[index].diffuser_radiance_mean,
            summary.band_gains[index].diffuser_radiance_stdev,
            summary.band_gains[index].diffuser_reflectance_mean,
            summary.band_gains[index].diffuser_reflectance_stdev);
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing band %d diffuser summary statistics "
                "to report", summary.band_gains[index].band_number);
            return ERROR;
        }
    }

    /* Write the band absolute radiance and reflectance gains */
    status = fprintf(rpt_fptr, "\n\n\nAbsolute Gains (Radiance Gains in "
        "units of DN / [W / (m^2 sr um)])\n\nBand                        Rad"
        ". Gain   Rad. Gain   Refl. Gain   Refl. Gain\n                  "
        "                       Std Dev                  Std Dev\n-----"
        "--------------------------------------------------------------"
        "---------\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing summary band gains header");
        return ERROR;
    }
    for (index = 0; index < num_oli_bands; index++)
    {
        /* If a band hasn't been characterized, skip it */
        if (summary.band_gains[index].band_number == 0)
        {
            continue;
        }

        status = fprintf(rpt_fptr, "%02d (%-19s) %12.6lf %11.6lf  %12.6lf "
            " %11.6lf\n",
            summary.band_gains[index].band_number,
            summary.band_gains[index].band_name,
            summary.band_gains[index].radiance_gain,
            summary.band_gains[index].radiance_gain_stdev,
            summary.band_gains[index].reflectance_gain,
            summary.band_gains[index].reflectance_gain_stdev);
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing summary band %d gain statistics to "
                "report", summary.band_gains[index].band_number);
            return ERROR;
        }
    }

    /* Done */
    return SUCCESS;
}
