/*------------------------------------------------------------------------------
NAME: frame_header_util

PURPOSE:
    This set of functions is used to handle the OLI and TIRS frame
    headers.  The frame header structures will be opened and read in 
    this file.  The structure is different for each 
    sensor thus requiring separate structures.  These routines handle 
    the two separate structures so as not to add confusion to 
    the calling routine.

RETURNS:
    All but the free routine will return a status of SUCCESS and ERROR.
    The free routine has no return value.
------------------------------------------------------------------------------*/

/* System Includes */
#include <stdlib.h>

/* IAS Includes */
#include "ias_logging.h"
#include "ias_db_insert.h"
#include "ias_ancillary.h"

/* Local Includes */
#include "get_parameters.h"
#include "report.h"
#include "long_collect_segment_stats.h"

/* Assign the static parameters for the functions in this file */
static IAS_L0R_OLI_FRAME_HEADER *oliFrameHeaders = NULL;
static IAS_L0R_TIRS_FRAME_HEADER *tirsFrameHeaders = NULL;
static IAS_SENSOR_ID sensorId;

static int tirs_flag = FALSE;
static int oli_flag = FALSE;

/*------------------------------------------------------------------------------
Name: frame_header_util

Purpose:
    Opens the frame header for either the OLI or the TIRS sensor.  
    The sensor will be obtained from the user parameters.
    This routine must be run before all the other routines in order
    for it to work properly.  Otherwise an error will occur in the other
    routines.
------------------------------------------------------------------------------*/

int open_frame_header
(
    int framesPerSegment          /* I:  Number of frames in each segment */
)
{
    if (get_sensor() == IAS_OLI)
    {
        /* Since this will always be the first routine to be called.  Set a 
           flag and id for the sensor here to use throughout the routines. */
        oli_flag = TRUE;
        sensorId = IAS_OLI;

        /* Allocate an array of structures to hold the OLI frame headers for 
           each frame of a segment. */
        oliFrameHeaders = (IAS_L0R_OLI_FRAME_HEADER *)
            malloc(framesPerSegment * sizeof(IAS_L0R_OLI_FRAME_HEADER));
        if (oliFrameHeaders == NULL)
        {
            return ERROR;
        }
    }
    else if (get_sensor() == IAS_TIRS)
    {
        /* Set the sensor flag and id for future use */
        tirs_flag = TRUE;
        sensorId = IAS_TIRS;

        /* Allocate an array of structures to hold the TIRS frame headers
           for each frame of a segment. */
        tirsFrameHeaders = (IAS_L0R_TIRS_FRAME_HEADER *)
            malloc(framesPerSegment * sizeof(IAS_L0R_TIRS_FRAME_HEADER));
        if (tirsFrameHeaders == NULL)
        {
            return ERROR;
        }
    }
    /* Since sensor is specified in the paremter io, it should never reach
       this point */
    else
    {
        IAS_LOG_ERROR("Unknown sensor index %d",get_sensor());
        return ERROR;
    }

    return SUCCESS;
}

/***************************************************************************
Name:  read_frame_header

Purpose:    Get the frame header for either the OLI or the TIRS depending
            on the parameter.  This will be stored in a static variable in
            this file and will only be used in this file.

***************************************************************************/
int read_frame_header
(
    L0RIO * fdL0r,              /* I:  Open L0R image to be read */
    int segmentStartFrame,      /* I:  First record to read */
    int segmentNFrames          /* I:  Number of frames to read */
)
{
    int status;

    if (oli_flag)
    {
        status = ias_l0r_get_oli_frame_headers(fdL0r, segmentStartFrame,
                segmentNFrames, oliFrameHeaders);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to get the OLI frame headers");
            return ERROR;
        }
    }
    else if (tirs_flag)
    {
        status = ias_l0r_get_tirs_frame_headers(fdL0r, segmentStartFrame,
                segmentNFrames, tirsFrameHeaders);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to get the TIRS frame headers");
            return ERROR;
        }
    }

    return SUCCESS;
}

/***************************************************************************
Name:  get_start_stop_frame_time

Purpose:    This routine returns the start and stop frame time indices.  The
            Headers were read previously and stored as static variables.

***************************************************************************/
int get_start_stop_frame_time
(
    int segmentNFrames,         /* I: Number of OLI frame header records */
    double *acquisitionTime,    /* O: start frame time */
    double *stop_frame_time,    /* O: stop frame time */
    int *start_frame_index,     /* O: start frame time */
    int *stop_frame_index       /* O: stop frame time */
)
{
    int status = 0;

    /* Find the first and last valid frame times */
    if (oli_flag)
    {
        status = ias_ancillary_get_start_stop_frame_times(
                oliFrameHeaders,
                NULL, /* There are no TIRS frame headers to pass */
                segmentNFrames, 0, acquisitionTime, stop_frame_time,
                NULL, NULL, /* No TIRS */
                start_frame_index, stop_frame_index,
                NULL, NULL ); /* No TIRS */

    }
    else if (tirs_flag)
    {
        status = ias_ancillary_get_start_stop_frame_times(
                NULL, /* No OLI frame headers to pass */
                tirsFrameHeaders,
                0, /* Frame header size for OLI is 0 */
                segmentNFrames,
                NULL, NULL, /* No OLI */
                acquisitionTime, stop_frame_time,
                NULL, NULL, /* No OLI */
                start_frame_index, stop_frame_index);
    }

    if (status != SUCCESS)
    {
        return ERROR;
    }

    return SUCCESS;
}

/***************************************************************************
Name:  dropped_frame_characterization

Purpose:    Call the rps library routine to retrieve the dropped frame
            characterization.  Use the OLI or TIRS frame header as 
            appropriate.

***************************************************************************/
int dropped_frame_characterization
(
    int  bandNum,                   /* I: 1-based band number */
    int  nDetectors,                /* I: # of detectors per SCA */
    int  segmentNFrames,            /* I: Number of frames of image data */
    IAS_PIXEL_MASK  *pixelMask      /* I: Pixel mask for the current
                                          band/SCA (could be
                                          populated with other artifact
                                          information)
                                       O: Pixel mask populated with
                                          dropped frame locations */
)
{
    int status;

    /* Retrieve the dropped frame characterization by using the RPS
       library routine. */
    if (oli_flag)
    {
        status = ias_rps_dropped_frame_characterization(oliFrameHeaders,
                    sensorId, bandNum, nDetectors, segmentNFrames,
                    pixelMask);
        if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Unable to drop frame characterization for OLI");
                return ERROR;
            }
    }
    else if (tirs_flag)
    {
        status = ias_rps_dropped_frame_characterization(tirsFrameHeaders,
                    sensorId, bandNum, nDetectors, segmentNFrames,
                    pixelMask);
        if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Unable to drop frame characterization for TIRS");
                return ERROR;
            }
    }

    return SUCCESS;
}


/***************************************************************************
Name:  free_frame_header

Purpose:  Free the allocated memory for either the OLI or TIRS depending on
          the assigned parameter 

***************************************************************************/
void free_frame_header()
{
    if (sensorId == IAS_OLI)
    {
        free(oliFrameHeaders);
    }
    else if (sensorId == IAS_TIRS)
    {
        free(tirsFrameHeaders);
    }
    
}
