/*==============================================================================
NAME: Get Parameters

PURPOSE:
    This function retrieve the application's parameters from the specified
    Object Description Language (ODL) file.
==============================================================================*/

/* System Includes */
#include <stdlib.h>
#include <string.h>
#include <limits.h>

/* IAS Includes */
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "ias_satellite_attributes.h"

/* Local Includes */
#include "get_parameters.h"
#include "report.h"

static IAS_SENSOR_ID sensor;
static char rlutFilename[PATH_MAX]; /* O: Path of RLUT */

/*------------------------------------------------------------------------------
NAME: get_parameters

PURPOSE:
    Read this application's arguments from an ODL file.

RETURNS:
    This function will return a status of SUCCESS if the arguments are
    successfully retrieved, ERROR if an error is encountered.
------------------------------------------------------------------------------*/
int get_parameters
(
    const char * arg,       /* I: Command line argument */
    int * satNum,           /* O: Satellite number */
    char workOrderId[IAS_WORKORDER_ID_SIZE],/* O: IAS work order ID */
    char pathCpf[PATH_MAX], /* O: Path of CPF */
    char pathIn[PATH_MAX],  /* O: Path of input L0R image */
    char pathOmf[PATH_MAX], /* O: Path of OMF */
    int * framesPerSegment, /* O: Number of frames per segment */
    int * store_in_database,/* O: Database flag */
    int * reportsRequested  /* O: List of report flags, array of size
                                  NUM_REPORTS */
)
{
    int parmSize;
    int segmentSizeDefault  = 1000;
    int segmentSizeRange[2] = {100, 64000};
    int status;
    char sensor_name[IAS_L0R_SENSOR_ID_LENGTH];

    /*==========================================================================
     * Read this application's parameter file
     *========================================================================*/

    /* Set up the structure needed to read this application's parameters.  These
       macros actually define a hidden struct.  Using a name of xOdl so this
       name stands out. */
    IAS_PARM_DECLARE_TABLE(xOdl, 13);

    IAS_PARM_WORK_ORDER_ID(xOdl, workOrderId,
        IAS_WORKORDER_ID_SIZE * sizeof(char), 1);

    IAS_PARM_ADD_INT_WITH_RANGES(xOdl, FRAMES_PER_SEGMENT,
        "Number of frames per segment", IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        2, segmentSizeRange, 1, &segmentSizeDefault, framesPerSegment,
        sizeof(int), 0);

    /* Report and database options */
    IAS_PARM_STORE_CHARACTERIZATION_IN_DB( xOdl, IAS_PARM_OPTIONAL,
        store_in_database, sizeof(*store_in_database), 0 );

    int report_default = 0;
    int min_report_flag = 0;
    IAS_PARM_ADD_BOOLEAN( xOdl, MINIMUM_REPORT_FLAG,
        "Create minimum report: 1 = create report, 0 = don't create",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &report_default, &min_report_flag, sizeof(min_report_flag), 1 );

    int max_report_flag = 0;
    IAS_PARM_ADD_BOOLEAN( xOdl, MAXIMUM_REPORT_FLAG,
        "Create maximum report: 1 = create report, 0 = don't create",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &report_default, &max_report_flag, sizeof(max_report_flag), 1 );

    int mean_report_flag = 0;
    IAS_PARM_ADD_BOOLEAN( xOdl, MEAN_REPORT_FLAG,
        "Create mean report: 1 = create report, 0 = don't create",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &report_default, &mean_report_flag, sizeof(mean_report_flag), 1 );

    int stdev_report_flag = 0;
    IAS_PARM_ADD_BOOLEAN( xOdl, STANDARD_DEVIATION_REPORT_FLAG,
        "Create standard deviation report: 1 = create report, 0 = don't create",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &report_default, &stdev_report_flag, sizeof(stdev_report_flag), 1 );

    int skewness_report_flag = 0;
    IAS_PARM_ADD_BOOLEAN( xOdl, SKEWNESS_REPORT_FLAG,
        "Create skewness report: 1 = create report, 0 = don't create",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &report_default, &skewness_report_flag, sizeof(skewness_report_flag),
        1 );

    int kurtosis_report_flag = 0;
    IAS_PARM_ADD_BOOLEAN( xOdl, KURTOSIS_REPORT_FLAG,
        "Create kurtosis report: 1 = create report, 0 = don't create",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &report_default, &kurtosis_report_flag, sizeof(kurtosis_report_flag),
        1 );

    int msr_report_flag = 0;
    IAS_PARM_ADD_BOOLEAN( xOdl, MEAN_SQUARED_RESPONSE_REPORT_FLAG,
        "Create mean squared response report: 1 = create report, 0 = don't "
        "create",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &report_default, &msr_report_flag, sizeof(msr_report_flag), 1 );

    int adj_corr_report_flag = 0;
    IAS_PARM_ADD_BOOLEAN( xOdl, ADJACENT_CORRELATION_REPORT_FLAG,
        "Create adjacent correlation report: 1 = create report, 0 = don't "
        "create",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &report_default, &adj_corr_report_flag, sizeof(adj_corr_report_flag),
        1 );

    int num_valid_frames_report_flag = 0;
    IAS_PARM_ADD_BOOLEAN( xOdl, NUM_VALID_FRAMES_REPORT_FLAG,
        "Create number of valid frames report: 1 = create report, 0 = don't "
        "create",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &report_default, &num_valid_frames_report_flag,
        sizeof(num_valid_frames_report_flag),
        1 );

    /* Determine if finding OLI or TIRS statistics */
    const char *valid_sensors[] = {"OLI","TIRS"};
    const char *default_sensor[] = {"OLI"};

    IAS_PARM_ADD_STRING(xOdl, SENSOR_NAME, 
        "Name of the sensor to be processed",
        IAS_PARM_OPTIONAL,
        2, valid_sensors,
        1, default_sensor,
        sensor_name, sizeof(sensor_name), 0);
    
    /* Read this application's parameters */
    parmSize = IAS_PARM_GET_TABLE_SIZE(xOdl);
    status = ias_parm_provide_help(arg, xOdl, parmSize, IAS_INPUT_PARAMETERS);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Providing help for application");
        return ERROR;
    }
    else if (status)
    {
        /* Since the user asked for help, copy the argument they specified into
           the OMF filename so they receive the same help message for that file.
           */
        strcpy(pathOmf, arg);
    }
    else
    {
        if (ias_parm_read(arg, xOdl, parmSize) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading application parameters from file %s", arg);
            return ERROR;
        }

        /* Create the OMF path.  We are assuming that this application is being
           run from within the current work order directory. */
        sprintf(pathOmf, "./%s.omf", workOrderId);
    }

    /* find the enumerated value for the sensor_name.  There are only two
       options from the valid parameter list */
    sensor = (!strcmp(sensor_name,"OLI")) ? IAS_OLI : IAS_TIRS;

    /*==========================================================================
     * Read the Order Metadata File (OMF)
     *========================================================================*/

    /* Set up the structure needed to read this work order's OMF.  These macros
       actually define a hidden struct.  Using a name of xOmf so this name
       stands out. */
    IAS_PARM_DECLARE_TABLE(xOmf, 4);
    IAS_PARM_L0R_FILENAME(xOmf, IAS_PARM_REQUIRED, pathIn, PATH_MAX, 1);
    IAS_PARM_CPF(xOmf, IAS_PARM_REQUIRED, pathCpf, PATH_MAX, 1);
    IAS_PARM_SATELLITE(xOmf, satNum, sizeof(*satNum), 1);
    IAS_PARM_RLUT(xOmf, 1, rlutFilename, sizeof(rlutFilename), 1);

    /* Read this work order's parameters */
    parmSize = IAS_PARM_GET_TABLE_SIZE(xOmf);
    status = ias_parm_provide_help(pathOmf, xOmf, parmSize, IAS_OMF_PARAMETERS);
    if (status == 1)
        exit(EXIT_SUCCESS); /* Provided help to user, still exit successfully */
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Returned from parameter file help");
        return ERROR;
    }
    if (ias_parm_read(pathOmf, xOmf, parmSize) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading OMF parameters from file %s", pathOmf);
        return ERROR;
    }

    /* Assign the report flags to the output parameter */
    reportsRequested[REPORT_MIN] =  min_report_flag;
    reportsRequested[REPORT_MAX] =  max_report_flag;
    reportsRequested[REPORT_MEAN] =  mean_report_flag;
    reportsRequested[REPORT_STDEV] =  stdev_report_flag;
    reportsRequested[REPORT_SKEWNESS] =  skewness_report_flag;
    reportsRequested[REPORT_KURTOSIS] =  kurtosis_report_flag;
    reportsRequested[REPORT_MEAN_SQ_RSP] =  msr_report_flag;
    reportsRequested[REPORT_ADJ_CORRELATION] =  adj_corr_report_flag;
    reportsRequested[REPORT_NUM_VALID_FRAMES] =  num_valid_frames_report_flag;

    return SUCCESS;
}

IAS_SENSOR_ID get_sensor(void)
{
    return sensor;
}

const char *get_rlut_filename(void)
{
    return rlutFilename;
}
