/*==============================================================================
NAME: long_collect_segment_stats

PURPOSE:
    This application is used to process long shutter collects, break them up
    into segments and perform histogram statistics on each segment and store
    them in the database for future analysis.  This routine is set up to
    process either OLI or TIRS but not both at the same time.  Linearization is
    performed for TIRS only before finding the histogram statitistics.

ALGORITHM REFERENCES:
    Long Collect Statistics Characterization ADD
==============================================================================*/

/* System Includes */
#include <stdlib.h>

/* IAS Includes */
#include "ias_logging.h"
#include "ias_db_insert.h"
#include "ias_ancillary.h"
#include "ias_db_get_connect_info.h"

/* Local Includes */
#include "get_parameters.h"
#include "report.h"
#include "long_collect_segment_stats.h"

/*------------------------------------------------------------------------------
NAME: remove_stats

PURPOSE:
    Remove all previous entries within the SEGMENT_STATISTICS table for the
    specified RADIOMETRIC_WO_COMMON_ID.  This will allow a shutter collect to be
    reprocessed without encountering a duplicate primary key error when
    statistics are saved into the table.

RETURNS:
    This function returns a status of SUCCESS if the records are successfully
    removed, ERROR if an error is encountered.
------------------------------------------------------------------------------*/
int remove_stats
(
    struct ias_db_connection * db, /* IO: Database connection */
    char rwoid[IAS_CHAR_ID_SIZE]   /* I: Radiometric work order ID */
)
{
    char                  sql[256];
    int                   status;
    struct ias_db_query * query    = NULL;

    /* Delete any previous entries for this work order from the table */
    status = snprintf(sql, sizeof(sql),
        "DELETE FROM SEGMENT_STATISTICS WHERE RADIOMETRIC_WO_COMMON_ID = %s",
        rwoid);
    if ((status < 0) || (status >= sizeof(sql)))
    {
        IAS_LOG_ERROR("SQL command buffer too small, requires %d characters",
            status);
        return ERROR;
    }

    query = ias_db_perform_query(db, sql);
    if (!ias_db_query_was_successful(query))
    {
        char errMsg[4096];

        ias_db_query_get_error_message(query, errMsg, sizeof(errMsg));
        IAS_LOG_ERROR("%s", errMsg);
        IAS_LOG_ERROR(
            "Deleting previous records from SEGMENT_STATISTICS for "
            "RADIOMETRIC_WO_COMMON_ID=%s ", rwoid);
        ias_db_query_close(query);
        return ERROR;
    }
    ias_db_query_close(query);
    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: save_stats

PURPOSE:
    This function will save a set of histogram statistics in the RADCHAR
    database in the SEGMENT_STATISTICS table for a single segment (set of
    frames) of a single band.

RETURNS:
    This function will return a status of SUCCESS if the statistics are
    successfully stored in the database, ERROR if an error is encountered.
------------------------------------------------------------------------------*/
int save_stats
(
    struct ias_db_connection * db, /* IO: Database connection */
    char rwoid[IAS_CHAR_ID_SIZE],  /* I: Radiometric work order ID */
    int framesPerSegment,          /* I: Number of frames/segment requested */
    int bandNum,    /* I: Band number */
    int isPanBand,  /* I: Is this the PAN band? */
    int scaNum,     /* I: SCA number */
    int nDetectors, /* I: Number of detectors per SCA in the band */
    int nFrames,    /* I: Number of frames in the segment */
    double aTime,   /* I: Acquisition time of first frame in segment */
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS * stats
                    /* I: histogram results */
)
{
    int          lineSet   = 0;
    int          status;
    int          tableSize;

    if ( isPanBand )
        lineSet = 0;
    else
        lineSet = 2;
    do
    {
        int line_index = isPanBand ? lineSet : 0;
        IAS_DB_INSERT_TABLE dbTable[]=
        {
            IAS_DB_TABLE_STRING( "radiometric_wo_common_id", rwoid ),
            IAS_DB_TABLE_INT( "band_id", bandNum ),
            IAS_DB_TABLE_INT( "sensor_chip_assembly_id", scaNum ),
            IAS_DB_TABLE_INT_ARRAY( "detector_id",
                stats->detector[line_index] ),
            IAS_DB_TABLE_INT( "lineset", lineSet ),
            IAS_DB_TABLE_DOUBLE("acquisition_time", aTime ),
            IAS_DB_TABLE_INT( "segment_size", nFrames ),
            IAS_DB_TABLE_FLOAT_ARRAY( "scene_pixel_min",
                stats->min[line_index] ),
            IAS_DB_TABLE_FLOAT_ARRAY( "scene_pixel_max",
                stats->max[line_index] ),
            IAS_DB_TABLE_FLOAT_ARRAY( "scene_pixel_stdev",
                stats->stdev[line_index] ),
            IAS_DB_TABLE_FLOAT_ARRAY( "scene_pixel_mean",
                stats->mean[line_index] ),
            IAS_DB_TABLE_FLOAT_ARRAY( "scene_pixel_skewness",
                stats->skewness[line_index] ),
            IAS_DB_TABLE_FLOAT_ARRAY( "scene_pixel_kurtosis",
                stats->kurtosis[line_index] ),
            IAS_DB_TABLE_INT_ARRAY( "nbr_of_frames",
                stats->num_values[line_index] ),
            IAS_DB_TABLE_FLOAT_ARRAY( "mean_squared_response",
                stats->mean_sq_resp[line_index] ),
            IAS_DB_TABLE_FLOAT_ARRAY( "adjacent_correlation",
                stats->correlation[line_index] )
        };

        tableSize = sizeof(dbTable) / sizeof(IAS_DB_INSERT_TABLE);

        status = ias_db_insert_using_table(db, "SEGMENT_STATISTICS", dbTable,
            tableSize, nDetectors);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR(
                "Storing histogram stats in database for band %d, SCA %d",
                bandNum, scaNum);
            return ERROR;
        }
    } while (isPanBand && (++lineSet < 2));

    return SUCCESS;
}


int main (int argc, char * argv[])
{
    struct ias_db_connection * db = NULL;

    IAS_CPF *        cpf                           = NULL;
    IAS_SATELLITE_ID satId;
    L0RIO *          fdL0r                         = NULL;
    char             pathCpf[PATH_MAX]             = "";
    char             pathIn[PATH_MAX]              = "";
    char             pathOmf[PATH_MAX]             = "";
    char             rwoid[IAS_CHAR_ID_SIZE];
    char             woid[IAS_WORKORDER_ID_SIZE];
    char *           dbHost                        = NULL;
    char *           dbName                        = NULL;
    char *           dbPassword                    = NULL;
    char *           dbUser                        = NULL;
    int              framesPerSegment;
    int              framesSkipBottom;
    int              framesSkipTop;
    int              i;
    int              reportsRequested[NUM_REPORTS];
    int              satNum;
    int              status;
    int store_in_database;
    const struct IAS_CPF_HISTOGRAM_CHAR *hist_char = NULL;

    /* Initialize logging */
    if (ias_log_initialize("long_collect_segment_stats"))
    {
        IAS_LOG_ERROR("Initializing logging capabilities");
        exit(EXIT_FAILURE);
    }

    /* Initialize the array that tracks which reports the user has requested.
       If reportsRequested[REPORT_XXX] is TRUE then the user has requested that
       report.  If the value is FALSE, that report should not be generated. */
    for (i = 0; i < NUM_REPORTS; i++)
        reportsRequested[i] = FALSE;

    /* Process the command line arguments */
    if (argc != 2)
    {
        IAS_LOG_ERROR("usage: long_collect_segment_stats <parameter file>");
        exit(EXIT_FAILURE);
    }
    if (get_parameters(argv[1], &satNum, woid, pathCpf, pathIn, pathOmf,
        &framesPerSegment, &store_in_database, reportsRequested) != SUCCESS)
    {
        IAS_LOG_ERROR("Getting parameters");
        exit(EXIT_FAILURE);
    }
    IAS_LOG_INFO("Processing started");

    /* Using the satellite number read from the OMF file, initialize the
       satellite attributes library */
    satId = ias_sat_attr_get_satellite_id_from_satellite_number(satNum);
    if (satId == ERROR)
    {
        IAS_LOG_ERROR("Unsupported satellite number: %d", satNum);
        exit(EXIT_FAILURE);
    }
    if (ias_sat_attr_initialize(satId) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attribute library");
        exit(EXIT_FAILURE);
    }

    /* Connect to the database if user requested it */
    if ( store_in_database )
    {
        if (ias_db_initialize_database_lib() != SUCCESS)
        {
            IAS_LOG_ERROR("Unable initialize database library");
            exit(EXIT_FAILURE);
        }
        if (ias_db_get_connect_info(&dbName, &dbUser, &dbPassword, &dbHost) !=
            SUCCESS)
        {
            IAS_LOG_ERROR("Unable retrieve database connection information");
            ias_db_close_database_lib();
            exit(EXIT_FAILURE);
        }
        db = ias_db_connect_to_database(dbName, dbUser, dbPassword, dbHost);
        free(dbName);
        free(dbUser);
        free(dbPassword);
        free(dbHost);
        if (db == NULL)
        {
            IAS_LOG_ERROR("Unable to establish database connection");
            ias_db_close_database_lib();
            exit(EXIT_FAILURE);
        }

        /* Obtain the radiometric work order ID for this IAS work order */
        status = ias_misc_get_char_id(db, "RADIOMETRIC_WORK_ORDER_COMMON",
            "RADIOMETRIC_WO_COMMON_ID", woid, rwoid);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR(
                "Unable to obtain RADIOMETRIC_WO_COMMON_ID for work order: %s",
                woid);
            ias_db_close_connection(db);
            ias_db_close_database_lib();
            exit(EXIT_FAILURE);
        }
    }

    /* Display information about what is being processed */
    IAS_LOG_DEBUG("            satNum = %d", satNum);
    IAS_LOG_DEBUG("       workOrderId = %s", woid);
    IAS_LOG_DEBUG("            pathIn = %s", pathIn);
    IAS_LOG_DEBUG("           pathCpf = %s", pathCpf);
    IAS_LOG_DEBUG("           pathOmf = %s", pathOmf);
    IAS_LOG_DEBUG("    frames/segment = %d", framesPerSegment);
    IAS_LOG_DEBUG("             rwoid = %s\n", rwoid);

    /* Check out the specified path, does it point to an L0R? */
    if (!ias_l0r_present(pathIn))
    {
        IAS_LOG_ERROR("Specified path does not point to an L0R image: %s",
            pathIn);
        if ( store_in_database )
        {
            ias_db_close_connection(db);
            ias_db_close_database_lib();
        }
        exit(EXIT_FAILURE);
    }

    /* Open the L0r */
    fdL0r = ias_l0r_open_from_filename(pathIn, IAS_L0R_COMPRESSION_DEFAULT);
    if (fdL0r == NULL)
    {
        IAS_LOG_ERROR("Opening L0R image: %s", pathIn);
        if ( store_in_database )
        {
            ias_db_close_connection(db);
            ias_db_close_database_lib();
        }
        exit(EXIT_FAILURE);
    }

    /* Open the CPF for reading */
    cpf = ias_cpf_read(pathCpf);
    if (cpf == NULL)
    {
        IAS_LOG_ERROR("Reading CPF: %s", pathCpf);
        ias_l0r_close(fdL0r);
        if ( store_in_database )
        {
            ias_db_close_connection(db);
            ias_db_close_database_lib();
        }
        exit(EXIT_FAILURE);
    }

    /* Get the frames to skip top and bottom from the CPF */
    hist_char = ias_cpf_get_histogram_char( cpf );
    if ( !hist_char )
    {
        IAS_LOG_ERROR("Getting the histogram char parameters from the CPF");
        ias_l0r_close(fdL0r);
        if ( store_in_database )
        {
            ias_db_close_connection(db);
            ias_db_close_database_lib();
        }
        exit(EXIT_FAILURE);
    }
    framesSkipTop = hist_char->frames_to_skip_top;
    framesSkipBottom = hist_char->frames_to_skip_bottom;

    /* Process the image specified */
    status = process_image(db, fdL0r, cpf, framesPerSegment, 
            framesSkipTop,
            framesSkipBottom, rwoid, reportsRequested);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to processing image: %s", pathIn);
        ias_cpf_free(cpf);
        ias_l0r_close(fdL0r);
        exit(EXIT_FAILURE);
    }

    /* Close the database connection */
    if ( store_in_database )
    {
        ias_db_close_connection(db);
        ias_db_close_database_lib();
    }
    /* Free the CPF */
    ias_cpf_free(cpf);

    /* Close the L0r image as processing */
    if (ias_l0r_close(fdL0r) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing L0R image");
        exit(EXIT_FAILURE);
    }

    /* Compute the processing time */
    IAS_LOG_INFO("Completed successfully");
    exit(EXIT_SUCCESS);
}
