/*------------------------------------------------------------------------------
NAME: process_image.c

PURPOSE:
    This functions in this routine are used to perform histogram statistics
    for both TIRS and OLI images.  The processing is broken into image segments
    to process long collects. 
------------------------------------------------------------------------------*/

/* System Includes */
#include <stdlib.h>
#include <math.h>

/* IAS Includes */
#include "ias_logging.h"
#include "ias_db_insert.h"
#include "ias_ancillary.h"

/* Local Includes */
#include "get_parameters.h"
#include "report.h"
#include "long_collect_segment_stats.h"

/* Local function prototypes */
void free_local_pointers
(
    L0RIO *fdL0r,                /* I:  L0R image pointer */
    IAS_RLUT_IO *rlut_ptr
);

void free_detector_offsets
(
    int * spatialOffsets[IAS_MAX_NSCAS],/* IO: Array of buffers */
    int number_scas                          /* I: */
);

int get_detector_offsets
(
    IAS_CPF * cpf,  /* I: An opened CPF */
    int bandIndex,  /* I: Band index to retrieve offsets for */
    int nScas,      /* I: Number of SCAs for this band */
    int nDetectors, /* I: Number of detectors per SCA for bandNum */
    int * spatialOffsets[IAS_MAX_NSCAS] /* O: Array of buffers */
);

static IAS_RLUT_LINEARIZATION_PARAMS *lin_params[IAS_TIRS_MAX_NSCAS];
                /* Array of linearization parameter data structures for all
                   detectors in the current band/SCA */
static uint16_t *imageBuffer = NULL;
static float *floatBuffer = NULL;

/*------------------------------------------------------------------------------
NAME: process_image

PURPOSE:
    This function will process an OLI or TIRS long shutter collect L0R image, 
    calculate the histogram statistics and store those values in the database.

RETURNS:
    The function will return a status of SUCCESS if the image is successfully
    processed, ERROR if an error is encountered.
------------------------------------------------------------------------------*/

int process_image
(
    struct ias_db_connection * db, /* IO: Database connection */
    L0RIO * fdL0r,                 /* I:  Open L0R image to be read */
    IAS_CPF * cpf,                 /* I:  CPF opened for reading */
    int framesPerSegment,          /* I:  Number of frames in each segment */
    int framesSkipTop,             /* I:  Frames to skip at top of image */
    int framesSkipBottom,          /* I:  Frames to skip at bottom of image */
    char rwoid[IAS_CHAR_ID_SIZE],  /* I:  Radiometric work order ID */
    int reportsRequested[NUM_REPORTS]/* I: Flags indicating reports to create */
)
{
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS * histoStats = NULL;

    IAS_L0R_INTERVAL_METADATA        metadata;
    IAS_SENSOR_ID                    sensorId;
    IAS_SPECTRAL_TYPE                spectralType;
    const IAS_CPF_DETECTOR_NOISE *   detectorNoise = NULL;
    const IAS_CPF_IMPULSE_NOISE *    impulseNoise = NULL;
    const IAS_CPF_SATURATION_LEVEL * saturation_levels = NULL;

    double     acquisitionTime;
    int        bandIndex;
    int        bandNum;
    int        framesInBand;
    int        i;
    int        isPanBand;
    int        lastLine;
    int        nDetectors;
    int        nLines;
    int        nLinesPerFrame;
    int        nReports = 0;
    int        nScas;
    int        nSegments;
    int        presentFlag;
    int        scaIndex;
    int        segmentLastFrame;
    int        segmentNFrames;
    int        segmentNum;
    int        segmentStartFrame;
    int        startLine;
    int        status;
    int * spatialOffsets[IAS_MAX_NSCAS];
    int band_list[IAS_MAX_TOTAL_BANDS];
    int number_bands;
    int index;
    int temp_nbands;
    int array_size;
    int oliFlag = FALSE;       /* Flag for OLI sensor */
    int tirsFlag = FALSE;      /* Flag for TIRS sensor */
    IAS_RLUT_IO *rlut_ptr = NULL;

    /* Set the linearization pointer to NULL for each SCA.  Although this
       parameter is only used for TIRS, set for all to be on the safe side */
    for (scaIndex = 0; scaIndex < IAS_TIRS_MAX_NSCAS; scaIndex++)
    {
        lin_params[scaIndex] = NULL;
    }

    /* Assign the sensor type in a local variable for ease of use */
    if (get_sensor() == IAS_OLI)
    {
        oliFlag = TRUE;
    }
    else if (get_sensor() == IAS_TIRS)
    {
        tirsFlag = TRUE;
    }

    /* Initialize array of pointers to dynamically allocated memory */
    for (scaIndex = 0; scaIndex < IAS_MAX_NSCAS; scaIndex++)
    {
        spatialOffsets[scaIndex] = NULL;
    }

    /* Determine the number of reports that have been requested */
    for (i = 0; i < NUM_REPORTS; i++)
    {
        if (reportsRequested[i])
            nReports++;
    }

    /* Open and retrieve the L0R metadata for this interval */
    if (ias_l0r_open_metadata(fdL0r, IAS_READ) != SUCCESS)
    {
        IAS_LOG_ERROR("Opening L0R interval metadata");
        return ERROR;
    }
    if (ias_l0r_get_interval_metadata(fdL0r, &metadata) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading L0R interval metadata");
        ias_l0r_close_metadata( fdL0r );
        return ERROR;
    }

    /* Open the L0R header so that we can read the OLI headers for each frame */
    if (ias_l0r_open_header(fdL0r, IAS_READ) != SUCCESS)
    {
        IAS_LOG_ERROR("Opening L0R header");
        ias_l0r_close_metadata( fdL0r );
        return ERROR;
    }

    /* Retrieve the impulse noise from the CPF */
    impulseNoise = ias_cpf_get_impulse_noise(cpf);
    if (impulseNoise == NULL)
    {
        IAS_LOG_ERROR("Retrieving impulse noise from CPF");
        ias_l0r_close_header(fdL0r);
        ias_l0r_close_metadata( fdL0r );
        return ERROR;
    }

    /* Retrieve the detector noise from the CPF */
    detectorNoise = ias_cpf_get_det_noise(cpf);
    if (detectorNoise == NULL)
    {
        IAS_LOG_ERROR("Retrieving detector noise from CPF");
        ias_l0r_close_metadata( fdL0r );
        ias_l0r_close_header(fdL0r);
        return ERROR;
    }

    /* Retrieve the saturation levels from the CPF */
    saturation_levels = ias_cpf_get_saturation_level(cpf);
    if (saturation_levels == NULL)
    {
        IAS_LOG_ERROR("Retrieving saturation levels from CPF");
        ias_l0r_close_header(fdL0r);
        ias_l0r_close_metadata( fdL0r );
        return ERROR;
    }

    /* Delete all prior entries in the SEGMENT_STATISTICS table for this
       radiometric work order */
    if ( db )
    {
        if (remove_stats(db, rwoid) != SUCCESS)
        {
            IAS_LOG_ERROR(
                 "Deleting previous records from SEGMENT_STATISTICS for "
                "RADIOMETRIC_WO_COMMON_ID=%s ", rwoid);
            ias_l0r_close_header(fdL0r);
            ias_l0r_close_metadata( fdL0r );
            return ERROR;
        }
    }

    /* Allocate an array of structure to hold the OLI frame headers for each
       frame of a segment. */
    status = open_frame_header(framesPerSegment);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Allocating memory for %d frame headers",
                framesPerSegment);
        ias_l0r_close_header(fdL0r);
        ias_l0r_close_metadata( fdL0r );
        return ERROR;
    }

    /* Get a list of bands to process and the number of bands for the
       specified sensor */
    sensorId = get_sensor();

    array_size = IAS_MAX_TOTAL_BANDS;
    number_bands = 0;
    status = ias_sat_attr_get_sensor_band_numbers( sensorId,
        IAS_NORMAL_BAND, 0, &band_list[number_bands],
        array_size, &temp_nbands );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Getting the list of bands");
        ias_l0r_close_header(fdL0r);
        free_frame_header();
        ias_l0r_close_metadata( fdL0r );
        return ERROR;
    }
    number_bands += temp_nbands;
    array_size -= temp_nbands;

    if (oliFlag)
    {
        /* Retrieve the available OLI band numbers for the VRP band excluding
           the blind and secondary bands */
        status = ias_sat_attr_get_sensor_band_numbers( sensorId,
            IAS_VRP_BAND, IAS_BLIND_BAND|IAS_SECONDARY_BAND,
            &band_list[number_bands],
            array_size, &temp_nbands );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Getting the list of bands");
            ias_l0r_close_header(fdL0r);
            free_frame_header();
            ias_l0r_close_metadata( fdL0r );
            return ERROR;
        }
        number_bands += temp_nbands;
        array_size -= temp_nbands;

        /* Retrieve the available band numbers for the OLI Blind bands */
        status = ias_sat_attr_get_sensor_band_numbers( sensorId,
            IAS_BLIND_BAND, 0, &band_list[number_bands],
            array_size, &temp_nbands );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Getting the list of bands");
            ias_l0r_close_header(fdL0r);
            free_frame_header();
            ias_l0r_close_metadata( fdL0r );
            return ERROR;
        }
        number_bands += temp_nbands;
    }
    else    /* TIRS */
    {
        /* Retrieve the available band numbers for the TIRS Blind bands 
           excluding the secondary bands */
        status = ias_sat_attr_get_sensor_band_numbers( sensorId,
        IAS_BLIND_BAND, IAS_SECONDARY_BAND, &band_list[number_bands],
        array_size, &temp_nbands );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Getting the list of bands");
            ias_l0r_close_header(fdL0r);
            free_frame_header();
            ias_l0r_close_metadata( fdL0r );
            return ERROR;
        }
        number_bands += temp_nbands;

        /* Open the RLUT file to read for TIRS only */
        rlut_ptr = ias_rlut_open_file(get_rlut_filename(), IAS_READ);
        if (rlut_ptr == NULL)
        {
            IAS_LOG_ERROR("Error opening RLUT file %s to read", 
                get_rlut_filename());
            free_local_pointers(fdL0r, rlut_ptr);
            return ERROR;
        }
    }

    /* Examine ALL bands within the image, not just the normal bands */
    for (index = 0; index < number_bands; index++)
    {
        int is_blind_band;

        /* Convert the band index into a band number */
        bandNum = band_list[index];
        bandIndex = ias_sat_attr_convert_band_number_to_index( bandNum );
        if (bandIndex == ERROR)
        {
            IAS_LOG_ERROR("Converting band number %d to a band index",
                bandNum);
            free_local_pointers(fdL0r, rlut_ptr);
            return ERROR;
        }

        /* Is the band a blind band? */
        is_blind_band = ias_sat_attr_band_classification_matches( bandNum,
            IAS_BLIND_BAND );

        /* Is this band present in the L0R? */
        if (ias_l0r_get_band_present(fdL0r, bandNum, &presentFlag) != SUCCESS)
        {
            IAS_LOG_ERROR("Determining if band %d is present", bandNum);
            free_local_pointers(fdL0r, rlut_ptr);
            return ERROR;
        }
        if (!presentFlag)
            continue;

        /* Get the number of SCAs for this band */
        nScas = ias_sat_attr_get_scas_per_band(bandNum);
        if (nScas == ERROR)
        {
            IAS_LOG_ERROR("Obtaining number of SCAs for band %d", bandNum);
            free_local_pointers(fdL0r, rlut_ptr);
            return ERROR;
        }

        /* Get the number of detectors per SCA */
        nDetectors = ias_sat_attr_get_detectors_per_sca(bandNum);
        if (nDetectors == ERROR)
        {
            IAS_LOG_ERROR("Obtaining number of detectors/SCA for band %d",
                bandNum);
            free_local_pointers(fdL0r, rlut_ptr);
            return ERROR;
        }

        /* Is this the PAN band? */
        spectralType = ias_sat_attr_get_spectral_type_from_band_number(bandNum);
        if (spectralType == ERROR)
        {
            IAS_LOG_ERROR("Unable to determine spectral type for band %d",
                bandNum);
            free_local_pointers(fdL0r, rlut_ptr);
            return ERROR;
        }
        isPanBand = (spectralType == IAS_SPECTRAL_PAN);

        /* Obtain the number of frames in this band based on the sensor ID */
        if (oliFlag)
        {
            if (ias_l0r_get_oli_frame_headers_records_count(fdL0r, 
                &framesInBand) != SUCCESS)
            {
                IAS_LOG_ERROR("Unable to retrieve frame header count");
                free_local_pointers(fdL0r, rlut_ptr);
                return ERROR;
            }
        }
        else if (tirsFlag)
        {
            if (ias_l0r_get_tirs_frame_headers_records_count(fdL0r, 
                &framesInBand) != SUCCESS)
            {
                IAS_LOG_ERROR("Unable to retrieve frame header count");
                free_local_pointers(fdL0r, rlut_ptr);
                return ERROR;
            }
        }
        else
        {
            IAS_LOG_ERROR("Obtaining number of frames for band %d", bandNum);
            free_local_pointers(fdL0r, rlut_ptr);
            return ERROR;
        }

        /* Get the number of lines per frame */
        nLinesPerFrame = ias_sat_attr_get_lines_per_frame(bandNum);
        if (nLinesPerFrame == ERROR)
        {
            IAS_LOG_ERROR("Obtaining number of lines/frame for band %d",
                bandNum);
            free_local_pointers(fdL0r, rlut_ptr);
            return ERROR;
        }

        /* Retrieve the spatial offsets for each detector of each SCA for this
           band */
        status = get_detector_offsets(cpf, bandIndex, nScas, nDetectors,
            spatialOffsets);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining detector offsets for band %d", bandNum);
            free_local_pointers(fdL0r, rlut_ptr);
            return ERROR;
        }

        /* Allocate a buffer for reading image data for this band's SCAs */
        imageBuffer = malloc(framesPerSegment * nLinesPerFrame *
            nDetectors * sizeof(uint16_t));
        if (imageBuffer == NULL)
        {
            IAS_LOG_ERROR("Allocating image buffer for reading: %d frames x %d "
                "lines x %d detectors", framesPerSegment, nLinesPerFrame,
                nDetectors);
            free_local_pointers(fdL0r, rlut_ptr);
            free_detector_offsets(spatialOffsets, nScas);
            return ERROR;
        }

        /* Some TIRS bands must run the linearization routine before
           processing histogram statistics, therefore TIRS data will be
           converted to float.  Allocate a buffer for the conversion. */
        if (tirsFlag)
        {
            /* Allocate a buffer for reading image data for this band's SCAs */
            floatBuffer = malloc(framesPerSegment * nLinesPerFrame *
                nDetectors * sizeof(float));
            if (floatBuffer == NULL)
            {
                IAS_LOG_ERROR("Allocating float image buffer for reading: "
                    "%d frames x %d lines x %d detectors", 
                    framesPerSegment, nLinesPerFrame, nDetectors);
                free_local_pointers(fdL0r, rlut_ptr);
                free_detector_offsets(spatialOffsets, nScas);
                return ERROR;
            }
        }

        /* Allocate a structure to store the results for the histogram
           statistics computations */
        histoStats = ias_rps_db_allocate_histogram_results(nDetectors,
            isPanBand);
        if (histoStats == NULL)
        {
            IAS_LOG_ERROR("Creating histogram statistics buffer");
            free_detector_offsets(spatialOffsets, nScas);
            free_local_pointers(fdL0r, rlut_ptr);
            return ERROR;
        }

        /* Open the band for reading */
        if (ias_l0r_open_band(fdL0r, bandNum, IAS_READ) != SUCCESS)
        {
            IAS_LOG_ERROR("Opening band %d", bandNum);
            free_detector_offsets(spatialOffsets, nScas);
            ias_rps_db_free_histogram_results(histoStats);
            free_local_pointers(fdL0r, rlut_ptr);
            return ERROR;
        }

        /* When processing each band, we'll always start the segments at the top
           of the image after skipping the requested number of frames.  In
           addition, all segments will contain the requested number of frames,
           except for the last segment which may be smaller. */
        segmentStartFrame = framesSkipTop;
        segmentNFrames = framesPerSegment;
        nSegments = ceil(
            (double)(framesInBand - framesSkipTop - framesSkipBottom) /
            (double)framesPerSegment);

        IAS_LOG_INFO("Processing band %02d in %d segment%s", bandNum, nSegments,
            ((nSegments == 1) ? "" : "s"));

        /* Linearization needs to be performed on the TIRS image bands but
           not the TIRS blind band -- there are no linearization parameters
           in the RLUT for the TIRS blind band. */
        if (tirsFlag && !is_blind_band)
        {
            for (scaIndex = 0; scaIndex < nScas; scaIndex++)
            {
                /* Get the correct set of linearization parameters */
                lin_params[scaIndex] = 
                        ias_rlut_read_linearization_params(rlut_ptr,
                        bandNum, scaIndex + 1, nDetectors);
                if (lin_params == NULL)
                {
                    IAS_LOG_ERROR("Getting linearization parameters "
                            "for band %d SCA %d", bandNum, scaIndex + 1);
                    ias_l0r_close_band(fdL0r, bandNum);
                    free_detector_offsets(spatialOffsets, nScas);
                    ias_rps_db_free_histogram_results(histoStats);
                    free_local_pointers(fdL0r, rlut_ptr);
                    return ERROR;
                }
            }
        }

        segmentNum = 0;
        while (segmentStartFrame < (framesInBand - framesSkipBottom))
        {
            double oli_stop_frame_time;
            int oli_start_frame_index;
            int oli_stop_frame_index;

            segmentNum++;
            IAS_LOG_INFO("Processing segment %d", segmentNum );

            /* Determine the extent of this segment within the image */
            segmentLastFrame = segmentStartFrame + framesPerSegment;
            if (segmentLastFrame > (framesInBand - framesSkipBottom))
            {
                segmentLastFrame = framesInBand - framesSkipBottom;
                segmentNFrames = segmentLastFrame - segmentStartFrame;
            }

            /* Read the OLI frame headers for this segment of the image.
               This will be read into memeory in the frame header utility
               and will be used there throughout this routine. */
            status = read_frame_header(fdL0r, segmentStartFrame,
                segmentNFrames);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Reading OLI frame headers for frames %d-%d",
                    segmentStartFrame, segmentLastFrame);
                ias_l0r_close_band(fdL0r, bandNum);
                free_detector_offsets(spatialOffsets, nScas);
                ias_rps_db_free_histogram_results(histoStats);
                free_local_pointers(fdL0r, rlut_ptr);
                return ERROR;
            }

            /* Find the first and last valid frame times */
            status = get_start_stop_frame_time(segmentNFrames,
                        &acquisitionTime, &oli_stop_frame_time,
                        &oli_start_frame_index, &oli_stop_frame_index);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Getting the start and end frame times for "
                    "frames %d-%d", segmentStartFrame, segmentLastFrame);
                ias_l0r_close_band(fdL0r, bandNum);
                free_detector_offsets(spatialOffsets, nScas);
                ias_rps_db_free_histogram_results(histoStats);
                free_local_pointers(fdL0r, rlut_ptr);
                return ERROR;
            }

            /* Convert the frame counts into line numbers */
            startLine = segmentStartFrame * nLinesPerFrame;
            nLines = segmentNFrames * nLinesPerFrame;
            lastLine = startLine + nLines - 1;

            /* Process this segment in each SCA of this band */
            for (scaIndex = 0; scaIndex < nScas; scaIndex++)
            {
                IAS_PIXEL_MASK *pixelMask = ias_pm_create(bandNum, scaIndex+1,
                    nDetectors, nLines);
                if ( !pixelMask )
                {
                    IAS_LOG_ERROR("Creating pixel mask for band %d, sca %d, "
                        "frames %d-%d, lines %d-%d", bandNum, scaIndex+1,
                        segmentStartFrame, segmentLastFrame, startLine,
                        lastLine);
                    ias_l0r_close_band(fdL0r, bandNum);
                    free_detector_offsets(spatialOffsets, nScas);
                    ias_rps_db_free_histogram_results(histoStats);
                    free_local_pointers(fdL0r, rlut_ptr);
                    return ERROR;
                }

                /* Read this segment from the L0R image.  The variable startLine
                   is a 0-based value. */
                status = ias_l0r_get_band_lines_sca(fdL0r, bandNum, scaIndex,
                    startLine, 0, nLines, nDetectors, imageBuffer);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR(
                        "Reading band %d, sca %d, frames %d-%d, lines %d-%d",
                        bandNum, scaIndex+1, segmentStartFrame,
                        segmentLastFrame, startLine, lastLine);
                    ias_pm_destroy( pixelMask );
                    ias_l0r_close_band(fdL0r, bandNum);
                    free_detector_offsets(spatialOffsets, nScas);
                    ias_rps_db_free_histogram_results(histoStats);
                    free_local_pointers(fdL0r, rlut_ptr);
                    return ERROR;
                }

                /* Retrieve the dropped frame characterization by using the RPS
                   library routine. */
                status = dropped_frame_characterization(bandNum,
                    nDetectors, segmentNFrames, pixelMask);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR(
                        "Running ias_rps_dropped_frame_characterization() on "
                        "band %d, sca %d, frames %d-%d (e.g. lines %d-%d)",
                        bandNum, scaIndex+1, segmentStartFrame,
                        segmentLastFrame, startLine, lastLine);
                    ias_pm_destroy( pixelMask );
                    ias_l0r_close_band(fdL0r, bandNum);
                    free_detector_offsets(spatialOffsets, nScas);
                    ias_rps_db_free_histogram_results(histoStats);
                    free_local_pointers(fdL0r, rlut_ptr);
                    return ERROR;
                }

                /* Retrieve the impulse noise characterization for the current
                   band and SCA. */
                status = ias_rps_impulse_noise_characterization(imageBuffer,
                    bandNum, scaIndex+1, nDetectors, nLines,
                    impulseNoise->median_filter_width[bandIndex],
                    detectorNoise->per_detector[bandIndex][scaIndex],
                    impulseNoise->in_limit[bandIndex], pixelMask,
                    NULL, NULL);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR(
                        "Running ias_rps_impulse_noise_characterization() "
                        "on band %d, sca %d, frames %d-%d (e.g. lines "
                        "%d-%d)", bandNum, scaIndex+1, segmentStartFrame,
                        segmentLastFrame, startLine, lastLine);
                    ias_pm_destroy( pixelMask );
                    ias_l0r_close_band(fdL0r, bandNum);
                    free_detector_offsets(spatialOffsets, nScas);
                    ias_rps_db_free_histogram_results(histoStats);
                    free_local_pointers(fdL0r, rlut_ptr);
                    return ERROR;
                }

                /* Retrieve the saturated pixel levels for the current band and
                   SCA. */
                status = ias_rps_saturated_pixel_characterization(
                    imageBuffer,
                    bandNum, scaIndex+1, nDetectors, nLines,
                    saturation_levels->
                        digital_low_saturation_level[bandIndex][scaIndex],
                    saturation_levels->
                        digital_high_saturation_level[bandIndex][scaIndex],
                    saturation_levels->
                        analog_low_saturation_level[bandIndex][scaIndex],
                    saturation_levels->
                        analog_high_saturation_level[bandIndex][scaIndex],
                    pixelMask, NULL, NULL);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR(
                        "Running ias_rps_saturated_pixel_characterization"
                        "() on "
                        "band %d, sca %d, frames %d-%d (e.g. lines %d-%d)",
                        bandNum, scaIndex+1, segmentStartFrame,
                        segmentLastFrame, startLine, lastLine);
                    ias_pm_destroy( pixelMask );
                    ias_l0r_close_band(fdL0r, bandNum);
                    free_detector_offsets(spatialOffsets, nScas);
                    ias_rps_db_free_histogram_results(histoStats);
                    free_local_pointers(fdL0r, rlut_ptr);
                    return ERROR;
                }

                /* Calculate the histogram statistics on a floating point
                   buffer of TIRS data */
                if (tirsFlag)
                {
                    /* Since TIRS only runs linerization and the
                       linearization routine only takes float data, 
                       Copy the unsigned integer data to float.
                       TIRS does not have bit shift issues */
                    for (i = 0; i < 
                        framesPerSegment * nLinesPerFrame * nDetectors; i++)
                    {
                        floatBuffer[i] = (float)imageBuffer[i];
                    }

                    /* Do the linearization correction only for the imaging
                       bands.  There are no linearization parameters for the
                       TIRS blind band. */
                    if ( ! is_blind_band )
                    {
                        status = ias_rps_3q_response_linearization(floatBuffer, 
                            lin_params[scaIndex], nDetectors, nLines);
                        if (status != SUCCESS)
                        {
                            IAS_LOG_ERROR("In call from "
                                "ias_rps_3q_response_linearization for "
                                "band %d SCA %d", bandNum, scaIndex + 1);
                            ias_pm_destroy( pixelMask );
                            ias_l0r_close_band(fdL0r, bandNum);
                            free_detector_offsets(spatialOffsets, nScas);
                            ias_rps_db_free_histogram_results(histoStats);
                            free_local_pointers(fdL0r, rlut_ptr);
                            return ERROR;
                        }
                    }

                    /* Retrieve the histogram statistics with the float
                       data as input.  The number of frames to
                       skip at the top and bottom of the image are hard coded to
                       zero because this application has already skipped the
                       necessary frames. */
                    status = ias_rps_histogram_statistics_characterization(
                        floatBuffer, IAS_R4, bandNum, scaIndex+1, nDetectors,
                        nLines, spatialOffsets[scaIndex], 0, 0,
                        pixelMask, NULL, histoStats, NULL);

                }       /* End if TIRS to get lineraization */
                else    /* OLI */
                {
                    /* Retrieve the OLI histogram statistics which is unsigned
                       integer.  The number of frames to
                       skip at the top and bottom of the image are hard coded to
                       zero because this application has already skipped the
                       necessary frames. */
                    status = ias_rps_histogram_statistics_characterization(
                        imageBuffer, IAS_UI2, bandNum, scaIndex+1, nDetectors,
                        nLines, spatialOffsets[scaIndex], 0, 0,
                        pixelMask, NULL, histoStats, NULL);
                }

                /* Verify that historgram statistics was successful for either
                   OLI or TIRS */
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR(
                        "Running ias_rps_histogram_statistics_characterization("
                        ") on band %d, sca %d, frames %d-%d (e.g. lines %d-%d)",
                        bandNum, scaIndex+1, segmentStartFrame,
                        segmentLastFrame, startLine, lastLine);
                    ias_pm_destroy( pixelMask );
                    ias_l0r_close_band(fdL0r, bandNum);
                    free_detector_offsets(spatialOffsets, nScas);
                    ias_rps_db_free_histogram_results(histoStats);
                    free_local_pointers(fdL0r, rlut_ptr);
                    return ERROR;
                }

                if ( db )
                {
                    status = save_stats(db, rwoid, framesPerSegment, bandNum,
                        isPanBand, scaIndex+1, nDetectors, segmentNFrames,
                        acquisitionTime, histoStats);
                    if (status != SUCCESS)
                    {
                        IAS_LOG_ERROR("Saving histogram statistics to the "
                            "database for band %d, sca %d, frames %d-%d "
                            "(e.g. lines %d-%d)", bandNum,
                            scaIndex+1, segmentStartFrame, segmentLastFrame,
                            startLine, lastLine);
                        ias_pm_destroy( pixelMask );
                        ias_l0r_close_band(fdL0r, bandNum);
                        free_detector_offsets(spatialOffsets, nScas);
                        ias_rps_db_free_histogram_results(histoStats);
                        free_local_pointers(fdL0r, rlut_ptr);
                        return ERROR;
                    }
                }

                /* If the user hasn't requested any report files be generated,
                   we'll skip the rest of this loop */
                if (nReports > 0)
                {

                    /* Save the requested statistics to a report file
                       (e.g. an image of floating point values */
                    status = report_save(reportsRequested, bandNum, isPanBand,
                        scaIndex+1, nScas, nDetectors, segmentNum,
                        framesPerSegment, nSegments, histoStats);
                    if (status != SUCCESS)
                    {
                        IAS_LOG_ERROR(
                            "Saving report file(s) for band %d, sca %d, frames "
                            "%d-%d (e.g. lines %d-%d)", bandNum, scaIndex+1,
                            segmentStartFrame, segmentLastFrame, startLine,
                            lastLine);
                        ias_pm_destroy( pixelMask );
                        ias_l0r_close_band(fdL0r, bandNum);
                        free_detector_offsets(spatialOffsets, nScas);
                        ias_rps_db_free_histogram_results(histoStats);
                        free_local_pointers(fdL0r, rlut_ptr);
                        return ERROR;
                    }
                }
                ias_pm_destroy( pixelMask );

            }   /* SCA for loop */

            /* Update the starting frame number for the next segment that will
               be processed for this band across all SCAs */
            segmentStartFrame += framesPerSegment;

        }       /* while loop for frames */

        /* Close the band */
        if (ias_l0r_close_band(fdL0r, bandNum) != SUCCESS)
        {
            IAS_LOG_ERROR("Closing band %d", bandNum);
            free_detector_offsets(spatialOffsets, nScas);
            ias_rps_db_free_histogram_results(histoStats);
            free_local_pointers(fdL0r, rlut_ptr);
            return ERROR;
        }

        /* Free the dynamically allocated memory for this band */
        free(imageBuffer);
        imageBuffer = NULL;

        if (tirsFlag)
        {
            free(floatBuffer);
            floatBuffer = NULL;

            /* free linearization memory for this band */
            for (scaIndex = 0; scaIndex < IAS_TIRS_MAX_NSCAS; scaIndex++)
            {
                free(lin_params[scaIndex]);
                lin_params[scaIndex] = NULL;
            }
        }

        ias_rps_db_free_histogram_results(histoStats);
        free_detector_offsets(spatialOffsets, nScas);
    }

    /* Free the memory used for the headers and metadata */
    free_local_pointers(fdL0r, rlut_ptr);

    return SUCCESS;
}
 
/*------------------------------------------------------------------------------
NAME: get_detector_offsets

PURPOSE:
    Obtain the detector offsets for each SCA of a specified band.  The array of
    pointers to spatial offset buffers must be all NULL when this function is
    called.  A buffer for the detector offsets for each SCA will be dynamically
    allocated by this function and should be freed by calling
    free_detector_offsets().

RETURNS:
    This function returns a status of SUCCESS if the offsets are successfully
    returned for all SCAs.  It will return a status of ERROR if an error is
    encountered and the pointers in the spatialOffsets array will all be NULL.
------------------------------------------------------------------------------*/
int get_detector_offsets
(
    IAS_CPF * cpf,  /* I: An opened CPF */
    int bandIndex,  /* I: Band index to retrieve offsets for */
    int nScas,      /* I: Number of SCAs for this band */
    int nDetectors, /* I: Number of detectors per SCA for bandNum */
    int * spatialOffsets[IAS_MAX_NSCAS] /* O: Array of buffers */
)
{
    const struct IAS_CPF_DETECTOR_OFFSETS * offsets = NULL;

    int scaIndex;

    /* Retrieve the detector offsets from the CPF */
    offsets = ias_cpf_get_det_offsets(cpf);
    if (offsets == NULL)
    {
        IAS_LOG_ERROR("Retrieving detector offsets from CPF");
        return ERROR;
    }

    /* Allocate a buffer to contain the detector offsets for each SCA */
    for (scaIndex = 0; scaIndex < nScas; scaIndex++)
    {
        spatialOffsets[scaIndex] = malloc(nDetectors * sizeof(int));
        if (spatialOffsets[scaIndex] == NULL)
        {
            IAS_LOG_ERROR("Allocating spatial offsets buffer for %d detectors",
                nDetectors);
            free_detector_offsets(spatialOffsets, nScas);
            return ERROR;
        }

        /* The along_per_detector array is indexed only to IAS_MAX_NBANDS.  Do
           not attempt to make the following ias_rps library call for bands
           with a larger index. */
        if (bandIndex < IAS_MAX_NBANDS)
        {
            ias_rps_convert_detector_delays2spatial_offsets(
                offsets->along_per_detector[bandIndex][scaIndex], nDetectors,
                spatialOffsets[scaIndex], NULL);
        }
        else
        {
            int i;
            for (i = 0; i < nDetectors; i++)
                spatialOffsets[scaIndex][i] = 0;
        }
    }

    return SUCCESS;
}

/*------------------------------------------------------------------------------
NAME: free_detector_offsets

PURPOSE:
    This function will free the dynamically allocated buffers that contain the
    detector offsets for each SCA of a single band.  The array of buffers
    containing the offsets should have been allocated by
    get_detector_offsets().

RETURNS:
    N/A
------------------------------------------------------------------------------*/
void free_detector_offsets
(
    int * spatialOffsets[IAS_MAX_NSCAS],/* IO: Array of buffers */
    int number_scas                          /* I: */
)
{
    int scaIndex;

    for (scaIndex = 0; scaIndex < number_scas; scaIndex++)
    {
        free(spatialOffsets[scaIndex]);
        spatialOffsets[scaIndex] = NULL;
    }
}

/*------------------------------------------------------------------------------
NAME: free_local_pointers

PURPOSE:
    This function will free the dynamically allocated buffers that contain the
    image metadata and headers and the frame headers.  The purpose of this
    function is to save a few lines in the error messages that occur in
    the band loop.  This set of 4 pointers are allocated and freed at
    basically the same time.

RETURNS:
    N/A
------------------------------------------------------------------------------*/
void free_local_pointers
(
    L0RIO *fdL0r,                /* I:  L0R image pointer */
    IAS_RLUT_IO *rlut_ptr
)
{
    int scaIndex;

    if (imageBuffer)
    {
        free(imageBuffer);
        imageBuffer = NULL;
    }

    if (floatBuffer)
    {
        free(floatBuffer);
        floatBuffer = NULL;
    }

    /* if rlut pointer has been allocated, free the pointer */
    if (rlut_ptr)
    {
        ias_rlut_close_file(rlut_ptr);
    }

    /* All pointers should either be allocated or NULL */
    if (lin_params[0])
    {
        for (scaIndex = 0; scaIndex < IAS_TIRS_MAX_NSCAS; scaIndex++)
        {
            free(lin_params[scaIndex]);
            lin_params[scaIndex] = NULL;
        }
    }

    /* Free the memory used for the OLI frame headers */
    free_frame_header();

    /* Close the L0r headers */
    ias_l0r_close_header(fdL0r);
    ias_l0r_close_metadata( fdL0r );

}
