/*--------------------------------------------------------------------------
NAME:    compute_apparent_y_size 

PURPOSE: Compute the apparent y-size of the image.  This is done by finding
         the detector with the longest span of moon pixels.  Note: this is
         not a continuous span -- there may be pixels that are either
         flagged as saturated, dropped, below the radiance threshold etc.
         The length of the span is converted to mrads using the solid angle
         passed into the routine.

RETURNS: SUCCESS or ERROR
---------------------------------------------------------------------------*/
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_pixel_mask.h"
#include "lunar_irradiance_characterization.h"

int compute_apparent_y_size
(
    IAS_PIXEL_MASK *lunar_mask, /* I: lunar mask where 1 indicates a moon
                                      pixel and 0 is not the moon, this mask
                                      has also been modified to include 0
                                      for all image artifacts so they will
                                      be ignored here */
    double solid_angle,         /* I: */
    int *center_detector,       /* O: Detector with the longest span of
                                      moon pixels */
    int *start_frame,           /* O: Frame information for the center
                                      detector span */
    int *stop_frame,            /* O: Frame information for the center
                                      detector span */
    double *apparent_y_size     /* O: Apprarent y-size in mrads */
)
{
    IAS_PIXEL_MASK_ITERATOR *iterator;
    IAS_PIXEL_MASK_SPAN span;

    int detector_index = -1;    /* Keep track of the longest span of moon */
    int longest_span = 0;       /*   pixels.  Note that this isn't a      */
    int longest_first_frame = 0;/*   continuous span -- it's the span     */
    int longest_last_frame = 0; /*   from first moon pixel to last.       */

    int current_det;            /* Keep track of the current detector's */
    int current_length;         /*   span of moon pixels.  Again not a  */
    int current_first_frame;    /*   continuous span.                   */
    int current_last_frame;     /*                                      */

    /* Create an iterator for the mask */
    iterator = ias_pm_get_iterator( lunar_mask, IAS_PM_INCLUDE, 0x1 );
    if ( iterator == NULL )
    {
        IAS_LOG_ERROR("Getting iterator");
        return ERROR;
    }

    /* There should be at least one span.  Get that one and set up the
       starting point for the current detector. */
    if ( !ias_pm_get_next_span( iterator, &span ))
    {
        IAS_LOG_ERROR("No valid spans in mask");
        return ERROR;
    }
    current_det = detector_index = span.detector_index;
    current_length = longest_span = span.length_of_span;
    current_first_frame = span.starting_pixel_index;
    current_last_frame = current_first_frame + span.length_of_span;

    /* Loop through the rest of the spans */
    while ( ias_pm_get_next_span( iterator, &span ))
    {
        /* This means we've moved beyond the current detector.  See if the
           longest needs updating. */
        if ( span.detector_index != current_det )
        {
            /* If the current detector total length is longer then the
               largest span, replace it */
            if ( current_length > longest_span )
            {
                longest_span = current_length;
                detector_index = current_det;
                longest_first_frame = current_first_frame;
                longest_last_frame = current_last_frame;
            }
            /* and start a new current detector */
            current_det = span.detector_index;
            current_length = span.length_of_span;
            current_first_frame = span.starting_pixel_index;
            current_last_frame = current_first_frame + span.length_of_span;
        }
        /* Otherwise update the current detector information with the span */
        else
        {
            current_last_frame = span.starting_pixel_index +
                span.length_of_span;
            current_length = current_last_frame - current_first_frame + 1;
        }
    }
    /* The iterator is no longer needed */
    ias_pm_destroy_iterator( iterator );

    /* Make sure that something got flagged as the longest span.  If this
       ever happens there is some error in the above logic that isn't being
       handled correctly. */
    if ( detector_index == -1 )
    {
        IAS_LOG_ERROR("Didn't find a long span");
        return ERROR;
    }
    *center_detector = detector_index;
    *start_frame = longest_first_frame;
    *stop_frame = longest_last_frame;
    *apparent_y_size = longest_span * solid_angle;

    return SUCCESS;
}  /* END compute_apparent_y_size */
