/*--------------------------------------------------------------------------
NAME:    compute_irradiance 

PURPOSE: Compute the irradiance by summing all valid lunar pixels and
         multiplying by the conversion factor.

RETURNS: SUCCESS or ERROR
---------------------------------------------------------------------------*/
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_pixel_mask.h"
#include "lunar_irradiance_characterization.h"

int compute_irradiance
(
    float *image_ptr,           /* I: pointer to one SCA of image data */
    int number_input_samples,   /* I: */
    IAS_PIXEL_MASK *lunar_mask, /* I: lunar mask where 1 indicates a moon
                                      pixel and 0 is not the moon */
    double conversion_factor,   /* I: Factor to convert to irradiance */
    double *irradiance          /* O: Irradiance */
)
{
    IAS_PIXEL_MASK_ITERATOR *iterator;
    IAS_PIXEL_MASK_SPAN span;
    double sum = 0.0;

    /* Create an iterator for the mask */
    iterator = ias_pm_get_iterator( lunar_mask, IAS_PM_INCLUDE, 0x1 );
    if ( iterator == NULL )
    {
        IAS_LOG_ERROR("Getting iterator");
        return ERROR;
    }

    /* Find all the lunar pixels */
    while ( ias_pm_get_next_span( iterator, &span ))
    {
        int i;
        /* Point to the first pixel of the span */
        float *value_ptr = image_ptr + span.detector_index +
            ( number_input_samples * span.starting_pixel_index );
        /* Loop through each pixel of the span */
        for ( i = 0; i < span.length_of_span; i++ )
        {
            sum += *value_ptr;
            /* Update the pointer to the next pixel in the span */
            value_ptr += number_input_samples;
        }
    }
    ias_pm_destroy_iterator( iterator );

    /* Convert the sum to irradiance */
    *irradiance = sum * conversion_factor;

    return SUCCESS;
}  /* END compute_irradiance */
