/*--------------------------------------------------------------------------
NAME:    create_median_filter_mask 

PURPOSE: Run a median filter in the along-track direction and create a mask
         with value 1 where ending value >= max filtered radiance *
         threshold and 0 elsewhere.  Pixels that are flagged in the artifact
         mask aren't flagged as valid lunar pixels.

RETURNS: SUCCESS or ERROR
---------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>

#include "ias_const.h"
#include "ias_logging.h"
#include "ias_pixel_mask.h"
#include "ias_satellite_attributes.h"
#include "lunar_irradiance_characterization.h"


/*--------------------------------------------------------------------------
NAME:    median

PURPOSE: Floating point median routine.

RETURNS: Median value of the floating point input array
---------------------------------------------------------------------------*/
static float median
(
    float *array,
    int n
)
{
    float local_array[n];
    int half_index = n / 2;
    int i, j;

    /* Copy the data to the local array */
    memcpy( local_array, array, sizeof(float) * n );
    /* Sort half the local array -- the middle value will be the median */
    for ( i = 0; i <= half_index; i++ )
    {
        for ( j = i + 1; j < n; j++ )
        {
            float temp;
            if ( local_array[j] < local_array[i] )
            {
                temp = local_array[j];
                local_array[j] = local_array[i];
                local_array[i] = temp;
            }
        }
    }
    return local_array[half_index];
}  /* END median */


/*--------------------------------------------------------------------------
NAME:    create_median_filter_mask 

PURPOSE: Run a median filter in the along-track direction and create a mask
         with value 1 where ending value >= max filtered radiance *
         threshold and 0 elsewhere.

RETURNS: SUCCESS or ERROR
---------------------------------------------------------------------------*/
int create_median_filter_mask
(
    float *image_ptr,           /* I: Pointer to one SCA of image data */
    int number_input_lines,     /* I: */
    int number_input_samples,   /* I: */
    int filter_width,           /* I: Width of the median filter to apply
                                      this value is expected to be odd.  The
                                      check on this value were done
                                      elsewhere and won't be repeated here. */
    double threshold,           /* I: Threshold value for median filter */
    IAS_PIXEL_MASK *artifact_mask,/* I: Artifacts in the image */
    IAS_PIXEL_MASK *lunar_mask, /* O: The mask a 1 indicates a moon pixel
                                      0 not the moon - or possibly an
                                      artifact */
    int *extents                /* O: */
)
{
    int status;
    float *filtered_image;      /* Pointer to a filtered image */
    float *value_ptr;           /* Value pointer into the filtered image */
    float max_radiance = image_ptr[0];
    double radiance_threshold;
    int line;
    int sample;
    PIXEL_MASK_TYPE *artifact_subset;
    int image_extents[6];       /* Approximate lunar image extents arranged
                                   as top, bottom, left, right, expanded top
                                   and expanded bottom indices */

    /* Need a copy of the image since the threshold is based on the maximum
       filtered radiance value */
    filtered_image = malloc( sizeof(float) * number_input_lines *
        number_input_samples );
    if ( filtered_image ==  NULL )
    {
        IAS_LOG_ERROR("Allocating memory for filtered image");
        return ERROR;
    }

    /* Apply a median filter to the image data.  The filtered data will be
       copied to a locally declared image. */
    for ( sample = 0; sample < number_input_samples; sample++ )
    {
        float filter[filter_width];
        float *input_image_ptr = NULL; /* For compiler warning */
        int oldest_value_index;
        int i;
        int j;
        int half_index = filter_width / 2;

        /* Set up for the first iterations by expanding the filter -- this
           fills all but the last value */
        for ( i = 0, j = half_index; i < filter_width - 1; i++ )
        {
            input_image_ptr =
                image_ptr + sample +          /* This is the first value */
                ( j * number_input_samples ); /* and add j lines in       */
            filter[i] = *input_image_ptr;
            if ( i < half_index )
                j--;
            else
                j++;

            /* Update the max value */
            if ( filter[i] > max_radiance )
                max_radiance = filter[i];
        }
        /* On exit from the loop input_image_ptr is pointing to the last
           value to be added to the filter -- it will be updated first thing
           in the line loop.  Point the following index at the last element
           indicating it's the stalest and that's the position to fill. */
        oldest_value_index = filter_width - 1;

        /* The value pointer is pointing to the filtered image */
        value_ptr = filtered_image + sample;
        for ( line = 0; line < number_input_lines; line++ )
        {
            /* Update the pointer to the data, the filter and index */
            if ( line > number_input_lines - half_index - 1 )
                input_image_ptr -= number_input_samples;
            else
                input_image_ptr += number_input_samples;
            /* Add the next value to the filter overwriting the oldest
               value */
            filter[oldest_value_index] = *input_image_ptr;
            /* Update the maximum radiance value */
            if ( filter[oldest_value_index] > max_radiance )
                max_radiance = filter[oldest_value_index];
            /* Update the pointer to the oldest filter value */
            oldest_value_index++;
            oldest_value_index %= filter_width;

            /* Apply the median filter and update the value pointer
               to the next frame */
            *value_ptr = median( filter, filter_width );
            value_ptr += number_input_samples;
        }
    }  /* END for sample loop */

    /* Set the radiance threshold */
    radiance_threshold = max_radiance * threshold;

    /* Determine the approximate extents of the lunar image using the
       filtered image and the radiance threshold */
    status = determine_lunar_extents( number_input_lines, number_input_samples,
        filtered_image, radiance_threshold, image_extents );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Determining image extents");
        free( filtered_image );
        return ERROR;
    }
// TODO: diagnostic
//printf("+++++ %d %d    %d %d   %d %d\n", image_extents[0], image_extents[1],
//image_extents[2], image_extents[3], image_extents[4], image_extents[5] );
// TODO: Note the image extents 4 and 5 have been expanded to the apparent
// edges of the lunar disc

    /* Use the expanded extents if they are valid */
    if ( image_extents[4] > 0 )
        image_extents[0] = image_extents[4];//0;
    if ( image_extents[5] > 0 )
        image_extents[1] = image_extents[5];//number_input_lines - 1;


    /* Create a subset mask for ease in looking for artifacts */
    artifact_subset = malloc( sizeof(PIXEL_MASK_TYPE) * number_input_lines );
    if ( artifact_subset == NULL )
    {
        IAS_LOG_ERROR("Allocating artifact mask subset");
        free( filtered_image );
        return ERROR;
    }

    /* Loop through the image flagging all pixels that exceed or equal
       the threshold */
    for ( sample = 0; sample < number_input_samples; sample++ )
    {
        int first_lunar_index;
        int last_lunar_index = 0;

        /* Get this sample's artifact mask subset */
        status = ias_pm_get_image_subset( artifact_mask, sample, 0, 1,
            number_input_lines, artifact_subset );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Getting artifact mask for detector %d", sample );
            free( filtered_image );
            free( artifact_subset );
            return ERROR;
        }

        /* Search the image for values greater than the threshold.  At this
           point only interested in the two lines where the lunar image
           appears to start and end. */
        value_ptr = filtered_image + sample +
            ( image_extents[0] * number_input_samples );
        first_lunar_index = -1;
        for ( line = image_extents[0]; line < image_extents[1]; line++ )
        {
            if ( *value_ptr >= radiance_threshold )
            {
                if ( first_lunar_index < 0 )
                    first_lunar_index = line;
                last_lunar_index = line;
            }
            value_ptr += number_input_samples;
        }

        /* If a lunar pixel was found in this detector add the span to the 
           mask */
        if ( first_lunar_index >= 0 )
        {
            int start_span_index;
            int length = last_lunar_index - first_lunar_index + 1;

            /* Add a lunar mask span that covers the full range of pixels
               found */
            status = ias_pm_add_pixels( lunar_mask, sample, first_lunar_index,
                length, 0x1 );
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Adding span for detector %d", sample );
                free( artifact_subset );
                free( filtered_image );
                return ERROR;
            }

            /* Add spans for the artifacts */
            start_span_index = -1;
            length = 0;
            for ( line = first_lunar_index; line <= last_lunar_index; line++ )
            {
                if ( artifact_subset[line] )
                {
                    if ( length > 0 )
                    {
                        length++;
                    }
                    else
                    {
                        start_span_index = line;
                        length = 1;
                    }
                }
                else if ( length > 0 )
                {
                    status = ias_pm_add_pixels( lunar_mask, sample,
                        start_span_index, length, 0x02 );
                    if ( status != SUCCESS )
                    {
                        IAS_LOG_ERROR("Adding span for detector %d", sample );
                        free( artifact_subset );
                        free( filtered_image );
                        return ERROR;
                    }
                    length = 0;
                }
            }

            /* There could be a span that wasn't written if an artifact runs
               to the end of the lunar disk */
            if ( length > 0 )
            {
                status = ias_pm_add_pixels( lunar_mask, sample,
                    start_span_index, length, 0x02 );
                if ( status != SUCCESS )
                {
                    IAS_LOG_ERROR("Adding span for detector %d", sample );
                    free( artifact_subset );
                    free( filtered_image );
                    return ERROR;
                }
            }
        }
    }  /* END for sample loop */

    free( artifact_subset );
    free( filtered_image );
    return SUCCESS;
}  /* END create_median_filter_mask */
