/*----------------------------------------------------------------------------
NAME:    database_util

         public routines:
             connect_to_database
             disconnect_from_database

PURPOSE: Routines to connect and disconnect from the database
-----------------------------------------------------------------------------*/
#include <string.h>
#include <stdlib.h>

#include "ias_logging.h"
#include "ias_const.h"
#include "ias_db.h"
#include "ias_db_get_connect_info.h"
#include "ias_miscellaneous.h"
#include "lunar_irradiance_characterization.h"

    
    
/*---------------------------------------------------------------------------
NAME:    connect_to_database
    
PURPOSE: Establish a database connection

RETURNS: SUCCESS or ERROR
---------------------------------------------------------------------------*/
struct ias_db_connection *connect_to_database(void)
{   
    char *database_name = NULL;
    char *user_name = NULL;
    char *password = NULL;
    char *host = NULL;
    int status; 
    struct ias_db_connection *db;
    
    /* Get the database connection info */
    status = ias_db_get_connect_info( &database_name, &user_name, &password,
            &host );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Getting database connection info");
        return NULL;
    }

    /* Connect to the database */
    db = ias_db_connect_with_transaction( database_name, user_name, password,
            host );

    /* Free the resources before checking the connection */
    free( database_name );
    free( user_name );
    free( password );
    free( host );

    /* Check the connection */
    if ( !db )
    {
        IAS_LOG_ERROR("Connecting to database");
        return NULL;
    }

    return db;
}  /* END connect_to_database */


/*---------------------------------------------------------------------------
NAME:    disconnect_from_database

PURPOSE: Disconnect from the database

RETURNS: SUCCESS or ERROR
---------------------------------------------------------------------------*/
int disconnect_from_database
(
    struct ias_db_connection *db, /* I: database connection information */
    int  commit         /* I: flag to indicate commiting the transaction */
)
{
    int status;

    if ( db == NULL )
        return SUCCESS;

    status = ias_db_disconnect_with_transaction( db, commit );
    if ( status != SUCCESS )
    {
        if ( commit )
        {
            IAS_LOG_ERROR("Commiting and closing database connection");
            return ERROR;
        }
        else
        {
            IAS_LOG_ERROR("Rolling back and closing database connection");
        }
    }

    return SUCCESS;
}  /* END disconnect_from_database */
