/*--------------------------------------------------------------------------
   NAME: db_store_characterization.c

	Public routine: db_store_characterization

	Private routines: store_lunar_irradiance_sca
			  clean_table

PURPOSE: Routine to store data in the database and to clean up tables
         beforehand.

RETURNS: SUCCESS or ERROR
--------------------------------------------------------------------------*/
#include <time.h>

#include "ias_db.h"
#include "ias_db_insert.h"
#include "ias_db_table.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "lunar_irradiance_characterization.h"


/*--------------------------------------------------------------------------
   NAME: store_lunar_irradiance_sca

PURPOSE: Local routine to store data in the lunar irradiance SCA table

RETURNS: SUCCESS or ERROR
--------------------------------------------------------------------------*/
static int store_lunar_irradiance_sca
(
    struct ias_db_connection *db, /* I: Database connection information */
    char *common_id,		  /* I: Rad common WO ID */
    int num_records,		  /* I: Number of records in info struct */
    SCA_INFORMATION *out_info	  /* I: Structure containing information
                                        to store */
)
{
    int status;
    int number_of_columns;
    int i;
    time_t now = time(NULL);
    struct tm *time_struct = gmtime( &now );
    char current_date[50];

    /* Build the curent time */
    status = snprintf( current_date, 50, "%4d:%03d:%02d:%02d:%02d.0",
        time_struct->tm_year + 1900, time_struct->tm_yday,
        time_struct->tm_hour, time_struct->tm_min, time_struct->tm_sec );
    if ( status < 0 || status >= 50 )
    {
        IAS_LOG_ERROR("Building the current date time");
        return ERROR;
    }

    for ( i = 0; i < num_records; i++ )
    {
        int length = out_info[i].stop_frame - out_info[i].start_frame + 1;
        IAS_DB_INSERT_TABLE table[] =
        {
            IAS_DB_TABLE_STRING( "radiometric_wo_common_id", common_id ),
            IAS_DB_TABLE_INT("band_id", out_info[i].band_number ),
            IAS_DB_TABLE_INT("sensor_chip_assembly_id",
                out_info[i].illuminated_sca ),
            IAS_DB_TABLE_DOUBLE("integrated_irradiance",
                out_info[i].irradiance ),
            IAS_DB_TABLE_INT("band_start_x", out_info[i].center_detector ),
            IAS_DB_TABLE_INT("band_start_y", out_info[i].start_frame ),
            IAS_DB_TABLE_INT("band_end_x", out_info[i].center_detector ),
            IAS_DB_TABLE_INT("band_end_y", out_info[i].stop_frame ),
            IAS_DB_TABLE_INT("band_y_size", length ),
            IAS_DB_TABLE_DOUBLE("moon_y_size", out_info[i].apparent_y_size ),
            IAS_DB_TABLE_TIMESTAMP("date_created", current_date )
        };
        number_of_columns = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);

        status = ias_db_insert_using_table( db, "lunar_irradiance_sca",
            table, number_of_columns, 1 );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Storing Lunar irradiance SCA values");
            return ERROR;
        }
    }

    return SUCCESS;
} /* END store_lunar_irradiance_sca */


/*--------------------------------------------------------------------------
   NAME: store_lunar_irradiance

PURPOSE: Local routine to store data in the lunar geometry table

RETURNS: SUCCESS or ERROR
--------------------------------------------------------------------------*/
static int store_lunar_irradiance
(
    struct ias_db_connection *db, /* I: Database connection information */
    char *common_id,		  /* I: Rad common WO ID */
    int num_records,		  /* I: Number of records in info struct */
    SCA_INFORMATION *out_info	  /* I: Structure containing information
                                        to store */
)
{
    int status;
    int number_of_columns;
    int i;

    for ( i = 0; i < num_records; i++ )
    {
        char date_string[50];
        int hour = out_info[i].image_time / 3600.0;
        int min = ( out_info[i].image_time - ( hour * 3600 )) / 60.0;
        float sec = out_info[i].image_time - ( hour * 3600 + min * 60 );
        int length = snprintf( date_string, 50, "%4d:%03d:%02d:%02d:%06.3f",
            out_info[i].year, out_info[i].day, hour, min, sec );
        if ( length < 0 || length >= 50 )
        {
            IAS_LOG_ERROR("Creating date string");
            return ERROR;
        }

        IAS_DB_INSERT_TABLE table[] = 
        {
            IAS_DB_TABLE_STRING( "radiometric_wo_common_id", common_id ),
            IAS_DB_TABLE_INT("band_id", out_info[i].band_number ),
            IAS_DB_TABLE_INT("sensor_chip_assembly_id",
                out_info[i].illuminated_sca ),
            IAS_DB_TABLE_INT("frame_number_image_start",
                out_info[i].start_frame ),
            IAS_DB_TABLE_INT("frame_number_image_stop",
                out_info[i].stop_frame ),
            IAS_DB_TABLE_TIMESTAMP("lunar_image_center_time",
                date_string ),
            IAS_DB_TABLE_DOUBLE("j2000_sc_scene_center_x",
            out_info[i].spacecraft_position.x ),
            IAS_DB_TABLE_DOUBLE("j2000_sc_scene_center_y",
            out_info[i].spacecraft_position.y ),
            IAS_DB_TABLE_DOUBLE("j2000_sc_scene_center_z",
            out_info[i].spacecraft_position.z )
        };
        number_of_columns = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);

        status = ias_db_insert_using_table( db, "lunar_irradiance",
            table, number_of_columns, 1 );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Storing Lunar irradiance SCA values");
            return ERROR;
        }
    }

    return SUCCESS;
}  /* END store_lunar_irradiance */


/*--------------------------------------------------------------------------
   NAME: store_lunar_geometry

PURPOSE: Local routine to store data in the lunar geometry table

RETURNS: SUCCESS or ERROR
--------------------------------------------------------------------------*/
static int store_lunar_geometry
(
    struct ias_db_connection *db, /* I: Database connection information */
    char *common_id,		  /* I: Rad common WO ID */
    int num_records,		  /* I: Number of records in info struct */
    SCA_INFORMATION *out_info	  /* I: Structure containing information
                                        to store */
)
{
    int status;
    int number_of_columns;
    int i;

    for ( i = 0; i < num_records; i++ )
    {
        char date_string[50];
        int hour = out_info[i].image_time / 3600.0;
        int min = ( out_info[i].image_time - ( hour * 3600 )) / 60.0;
        float sec = out_info[i].image_time - ( hour * 3600 + min * 60 );

        status = snprintf( date_string, 50, "%4d%03d%02d%02d%06.3f",
            out_info[i].year, out_info[i].day, hour, min, sec );
        if ( status < 0 || status >= 50 )
        {
            IAS_LOG_ERROR("Creating date string");
            return ERROR;
        }

        IAS_DB_INSERT_TABLE table[] = 
        {
            IAS_DB_TABLE_STRING( "radiometric_wo_common_id", common_id ),
            IAS_DB_TABLE_INT("band_id", out_info[i].band_number ),
            IAS_DB_TABLE_INT("sensor_chip_assembly_id",
                out_info[i].illuminated_sca ),
            IAS_DB_TABLE_STRING("image_time", date_string ),
            IAS_DB_TABLE_DOUBLE("spacecraft_x",
                out_info[i].spacecraft_position.x ),
            IAS_DB_TABLE_DOUBLE("spacecraft_y",
                out_info[i].spacecraft_position.y ),
            IAS_DB_TABLE_DOUBLE("spacecraft_z",
                out_info[i].spacecraft_position.z ),
            IAS_DB_TABLE_DOUBLE("roll_top",
                out_info[i].spacecraft_top_att.x ),
            IAS_DB_TABLE_DOUBLE("roll_rate_top",
                out_info[i].spacecraft_top_rates.x ),
            IAS_DB_TABLE_DOUBLE("roll_middle",
                out_info[i].spacecraft_center_att.x ),
            IAS_DB_TABLE_DOUBLE("roll_rate_middle",
                out_info[i].spacecraft_center_rates.x ),
            IAS_DB_TABLE_DOUBLE("roll_bottom",
                out_info[i].spacecraft_bottom_att.x ),
            IAS_DB_TABLE_DOUBLE("roll_rate_bottom",
                out_info[i].spacecraft_bottom_rates.x ),
            IAS_DB_TABLE_DOUBLE("pitch_top",
                out_info[i].spacecraft_top_att.y ),
            IAS_DB_TABLE_DOUBLE("pitch_rate_top",
                out_info[i].spacecraft_top_rates.y ),
            IAS_DB_TABLE_DOUBLE("pitch_middle",
                out_info[i].spacecraft_center_att.y ),
            IAS_DB_TABLE_DOUBLE("pitch_rate_middle",
                out_info[i].spacecraft_center_rates.y ),
            IAS_DB_TABLE_DOUBLE("pitch_bottom",
                out_info[i].spacecraft_bottom_att.y ),
            IAS_DB_TABLE_DOUBLE("pitch_rate_bottom",
                out_info[i].spacecraft_bottom_rates.y ),
            IAS_DB_TABLE_DOUBLE("yaw_top",
                out_info[i].spacecraft_top_att.z ),
            IAS_DB_TABLE_DOUBLE("yaw_rate_top",
                out_info[i].spacecraft_top_rates.z ),
            IAS_DB_TABLE_DOUBLE("yaw_middle",
                out_info[i].spacecraft_center_att.z ),
            IAS_DB_TABLE_DOUBLE("yaw_rate_middle",
                out_info[i].spacecraft_center_rates.z ),
            IAS_DB_TABLE_DOUBLE("yaw_bottom",
                out_info[i].spacecraft_bottom_att.z ),
            IAS_DB_TABLE_DOUBLE("yaw_rate_bottom",
                out_info[i].spacecraft_bottom_rates.z ),
            IAS_DB_TABLE_DOUBLE("moon_sc_distance",
                out_info[i].spacecraft_moon_distance ),
            IAS_DB_TABLE_DOUBLE("earth_sun_distance",
                out_info[i].earth_sun_distance ),
            IAS_DB_TABLE_DOUBLE("earth_moon_distance",
                out_info[i].earth_moon_distance )
        };
        number_of_columns = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);

        status = ias_db_insert_using_table( db, "lunar_geometry",
            table, number_of_columns, 1 );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Storing Lunar irradiance SCA values");
            return ERROR;
        }
    }

    return SUCCESS;
}  /* END store_lunar_geometry */


/*--------------------------------------------------------------------------
   NAME: clean_table

PURPOSE: Local routine to clean the table passed as parameter

RETURNS: SUCCESS or ERROR
--------------------------------------------------------------------------*/
static int clean_table
(
    struct ias_db_connection *db, /* I: Database connection information */
    char *common_id,
    char *table
)
{
    int status;
    char query_string[256];
    struct ias_db_query *query = NULL;
    char error_message[IAS_SQL_RTN_MSGLEN];

    /* Create a delete string */
    status = snprintf( query_string, sizeof(query_string), "delete from %s "
        "WHERE radiometric_wo_common_id = %s", table, common_id );
    if ( status < 0 || status >= sizeof(query_string) )
    {
        IAS_LOG_ERROR("Creating string to delete records from database");
        return ERROR;
    }

    /* Clean up any database records that may exist for this work order */
    query = ias_db_perform_query( db, query_string );
    if ( !query )
    {
        IAS_LOG_ERROR("Returned from ias_db_perform_query for: %s",
            query_string );
        return ERROR;
    }

    /* Check the result of the query */
    if ( ! ias_db_query_was_successful( query ))
    {
        ias_db_query_get_error_message( query, error_message,
            IAS_SQL_RTN_MSGLEN );
        IAS_LOG_ERROR("Deleting records from database for "
            "radiometric_wo_common_id = %s", common_id );
        IAS_LOG_ERROR("Message is: %s", error_message );
        ias_db_query_close( query );
        return ERROR;
    }
    ias_db_query_close( query );
    return SUCCESS;
}  /* END clean_table */


/*--------------------------------------------------------------------------
   NAME: db_store_characterization

PURPOSE: Public routine to store data in the database and to clean up tables
         beforehand.

RETURNS: SUCCESS or ERROR
--------------------------------------------------------------------------*/
int db_store_characterization
(
    struct ias_db_connection *db, /* I: Database connection information */
    SCA_INFORMATION *out_info     /* I: Information to be stored */
)
{
    int status;
    char common_id[25];
    int num_records = get_num_bands_to_process();
    int i;
    char *table_name[3] =
    {
        "lunar_irradiance_sca",
        "lunar_irradiance",
        "lunar_geometry"
    };

    /* Get or create a radiometric common ID */
    status = ias_misc_get_char_id( db, "radiometric_work_order_common",
        "radiometric_wo_common_id", get_work_order_id(), common_id );

    /* Make sure the tables are clean before attempting storage */
    for ( i = 0; i < 3; i++ )
    {
        status = clean_table( db, common_id, table_name[i] );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Cleaning table: %s", table_name[i] );
            return ERROR;
        }
    }

    /* Store the SCA information */
    status = store_lunar_irradiance_sca( db, common_id, num_records, out_info );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Storing SCA information to database");
        return ERROR;
    }

    /* Store the geometry information */
    status = store_lunar_irradiance( db, common_id, num_records, out_info );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Storing Lunar irradiance information to database");
        return ERROR;
    }

    status = store_lunar_geometry( db, common_id, num_records, out_info );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Storing Lunar geometry information to database");
        return ERROR;
    }

    return SUCCESS;
}  /* END db_store_characterization */
