/*--------------------------------------------------------------------------
NAME:    determine_lunar_extents

PURPOSE: Determine the approximate extents of the lunar image within the 
         SCA

RETURNS: SUCCESS or ERROR
---------------------------------------------------------------------------*/
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "ias_const.h"
#include "ias_logging.h"
#include "lunar_irradiance_characterization.h"

static float median
(
    float *array,
    int n
)
{
    float local_array[n];
    int half_index = n / 2;
    int i, j;

    /* Copy the data to the local array */
    memcpy( local_array, array, sizeof(float) * n );
    /* Sort half the local array -- the middle value will be the median */
    for ( i = 0; i <= half_index; i++ )
    {
        for ( j = i + 1; j < n; j++ )
        {
            float temp;
            if ( local_array[j] < local_array[i] )
            {
                temp = local_array[j];
                local_array[j] = local_array[i];
                local_array[i] = temp;
            }
        }
    }
    return local_array[half_index];
}  /* END median */


int determine_lunar_extents
(
    int number_input_lines,     /* I: */
    int number_input_samples,   /* I: */
    float *filtered_image,      /* I: Pointer to a smoothed image */
    double radiance_threshold,  /* I: Threshold for lunar pixels */
    int extents[6]              /* O: Extents ordered as top, bottom, left
                                      right, expanded top and expanded
                                      bottom indices */
)
{
    int line;
    int i;
    int max_line_index = 0;
    int min_line_index = number_input_lines - 1;
    int max_samp_index = 0;
    int min_samp_index = number_input_samples - 1;
    float minimum;
    float maximum;
    float last_stdev = 0.0;
    float *line_ptr = filtered_image;
    float *diff = malloc(sizeof(float) * number_input_lines );
    if ( diff == NULL )
    {
        IAS_LOG_ERROR("Allocating memory");
        free( diff );
        return ERROR;
    }

    /* Set the expanded indices to -1 - invalid */
    extents[4] = extents[5] = -1;

    /* Loop through the lines and calculate an across-track standard
       deviation and take it's "discrete" 1st derivitive by taking the
       delta x at unit distance */
    for ( line = 0; line < number_input_lines; line++ )
    {
        int sample;
        double mean = 0.0;
        double sum = 0.0;
        double sum_sq = 0.0;
        double stdev;
        /* Loop through the samples of this line accumulating the sums */
        for ( sample = 0; sample < number_input_samples; sample++ )
        {
            sum += line_ptr[sample];
            sum_sq = line_ptr[sample] * line_ptr[sample];
        }
        mean = sum / number_input_samples;
        sum_sq -= number_input_samples * mean * mean;
        stdev = sqrt( fabs( sum_sq ) / ( number_input_samples - 1 ));

        /* Take the "discrete" 1st derivitive */
        diff[line] = (float)( stdev - last_stdev );
        last_stdev = stdev;

        /* Update the line pointer */
        line_ptr += number_input_samples;
    }

    /* Smooth the discrete 1st derivitive and capture the minimum and
       maximum values and their indices.  No point in doing anything fancy
       for the edge cases of the filter since it's unlikely the lunar image
       will be there. */
    minimum = maximum = 0.0;
    for ( line = 2, i = 0; line < number_input_lines - 2; line++, i++ )
    {
        float filt_diff = median( &diff[i], 5 );
        if ( filt_diff > maximum )
        {
            maximum = filt_diff;
            max_line_index = line;
        }
        if ( filt_diff < minimum )
        {
            minimum = filt_diff;
            min_line_index = line;
        }
    }
    free( diff );

    /* Sanity check -- the line index of the minumum value should be greater
       than the maximum.  That is because we should see the standard
       deviation increase (positive 1st derivitive) then decrease (negative
       1st derivitive). */
    if ( min_line_index < max_line_index )
    {
        IAS_LOG_WARNING("Computed a decreasing slope before an increasing "
            "one.  Overriding %d with 0 and %d with %d", max_line_index,
            min_line_index, number_input_lines - 1 );
        max_line_index = 0;
        min_line_index = number_input_lines - 1;
    }
    /* If they are at least in the correct order do a simple check for the
       left and right indices and possibly expand the top and bottom */
    else
    {
        float max_index[5];  /* These are defined as float to use the   */
        float min_index[5];  /*    median routine                       */
        int middle_sample;
        int sample;
        int done;
        line = max_line_index - 2 + ( min_line_index - max_line_index ) / 2;
        line_ptr = filtered_image + line * number_input_samples;
        for ( i = 0; i < 5; i++ )
        {
            /* Search forward for the most likely candidate for the left
               edge index */
            sample = done = 0;
            do
            {
                if ( line_ptr[sample] > radiance_threshold 
                    && line_ptr[sample + 1] > radiance_threshold
                    && line_ptr[sample + 2] > radiance_threshold
                    && line_ptr[sample + 3] > radiance_threshold
                    && line_ptr[sample + 4] > radiance_threshold )
                {
                    max_index[i] = sample;
                    done = 1;
                }
                sample++;
            } while ( !done && sample < number_input_samples );

            /* Search backward for the most likely right edge candidate */
            sample = number_input_samples - 1;
            done = 0;
            do
            {
                if ( line_ptr[sample] > radiance_threshold 
                    && line_ptr[sample - 1] > radiance_threshold
                    && line_ptr[sample - 2] > radiance_threshold
                    && line_ptr[sample - 3] > radiance_threshold
                    && line_ptr[sample - 4] > radiance_threshold )
                {
                    min_index[i] = sample;
                    done = 1;
                }
                sample--;
            } while ( !done && sample >= 0 );
        line_ptr += number_input_samples;
        }
        
        /* Choose the median of these values */
        max_samp_index = median( max_index, 5 );
        min_samp_index = median( min_index, 5 );

        /* Try to expand the top and bottom to account for the even/odd
           offsets.  Point to the middle sample that was determined above. */
        middle_sample = max_samp_index - 2 +
            ( min_samp_index - max_samp_index ) / 2;
        line_ptr = filtered_image + middle_sample + max_line_index *
            number_input_samples;

        /* If for some reason that detector is lower in the image enough to 
           be out of the moon go to the next detector */
        if ( *line_ptr < radiance_threshold )
            line_ptr++;

        /* While either of the two adjacent detectors is above the threshold
           step backwards in the image until both are below the threshold */
        line = max_line_index;
        while ( line_ptr[0] > radiance_threshold
            || line_ptr[1] > radiance_threshold )
        {
            line--;
            line_ptr -= number_input_samples;
        }
        extents[4] = line;

        /* Do the same thing for the bottom of the image just in reverse */
        line_ptr = filtered_image + middle_sample + min_line_index *
            number_input_samples;
        if ( *line_ptr < radiance_threshold )
            line_ptr++;
        line = min_line_index;
        while ( line_ptr[0] > radiance_threshold
            || line_ptr[1] > radiance_threshold )
        {
            line++;
            line_ptr += number_input_samples;
        }
        extents[5] = line;
    }

    /* Assign the return values */
    extents[0] = max_line_index;
    extents[1] = min_line_index;
    extents[2] = max_samp_index;
    extents[3] = min_samp_index;

    return SUCCESS;
}  /* END determine_lunar_extents */
