/*----------------------------------------------------------------------------
NAME:    generate_report

PURPOSE: Output a report containing information about the scene as well as
         the irradiance

RETURNS: SUCCESS or ERROR
----------------------------------------------------------------------------*/
#include <stdio.h>

#include "ias_const.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "lunar_irradiance_characterization.h"

int generate_report
(
    SCA_INFORMATION *out_info   /* I: Information to be reported */
)
{
    char *filename = get_report_filename();
    int status;
    FILE *ouf;
    IAS_REPORT_HEADER *report_header;
    int index;

    /* Open the output file */
    ouf = fopen( filename, "w");
    if ( ouf == NULL )
    {
        IAS_LOG_ERROR("Opening: %s for writing", filename );
        return ERROR;
    }

    /* Get a standard header */
    report_header = ias_misc_initialize_report_header( get_l0r_filename(),
        get_work_order_id(), "Lunar Irradiance Characterization");
    if ( report_header == NULL )
    {
        IAS_LOG_ERROR("Initializing report header");
        fclose( ouf );
        return ERROR;
    }

    /* Write the header and free it before the error check */
    status = ias_misc_write_report_header( ouf, report_header );
    ias_misc_free_report_header( report_header );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Writing the standard header to report file %s",
            filename );
        fclose( ouf );
        return ERROR;
    }

    /* Write out the general information */
    status = fprintf( ouf, "\nIlluminated SCA: %d\n\n",
        out_info[0].illuminated_sca );
    if ( status < 0 )
    {
        IAS_LOG_ERROR("Writing to report file");
        fclose( ouf );
        return ERROR;
    }

    /* Write the SCA data */
    fprintf( ouf, 
        "        SumIrr     Ysize        Center                       "
        "Top/Bottom\n"
        "Band [uW/m^2/nm]  [mrads]     Det  Frame   Time at Center    "
        "  Frame      S/C-Moon       Earth-Sun      Earth-moon       "
        "S/C-X        S/C-Y      S/C-Z\n");
    for ( index = 0; index < get_num_bands_to_process(); index++ )
    {
        int hour = out_info[index].image_time / 3600.0;
        int min = ( out_info[index].image_time - ( hour * 3600 )) / 60.0;
        float sec = out_info[index].image_time - ( hour * 3600 + min * 60 );

        int centerframe = out_info[index].start_frame +
            (( out_info[index].stop_frame - out_info[index].start_frame + 1 )
            / 2 );
        status = fprintf( ouf, "%3d  %10.6f  %f %6d %6d  %02d:%02d:%09.6f "
            "%6d %6d %f %f %f %f %f %f\n",
//            "Band %d\n"
//            "    Irradiance: %f\n"
//            "    Apparent Y-size: %f\n"
//            "    Center detector and frame number: %d, %d\n"
//            "    Time at center: %f\n"
//            "    Top and bottom frame: %d, %d\n"
//            "    S/C-Moon distance: %f\n"
//            "    Earth-Sun distance: %f\n"
//            "    Earth-Moon distance: %f\n"
//            "    S/C Pos: %f, %f, %f\n\n",
            out_info[index].band_number,
            out_info[index].irradiance,
            out_info[index].apparent_y_size * 1000.0,
            out_info[index].center_detector,
            centerframe,
            hour, min, sec,
//            out_info[index].image_time,
            out_info[index].start_frame,
            out_info[index].stop_frame,
            out_info[index].spacecraft_moon_distance,
            out_info[index].earth_sun_distance,
            out_info[index].earth_moon_distance,
            out_info[index].spacecraft_position.x / 1000.0,
            out_info[index].spacecraft_position.y / 1000.0,
            out_info[index].spacecraft_position.z / 1000.0 );
        if ( status < 0 )
        {
            IAS_LOG_ERROR("Writing to report file");
            fclose( ouf );
            return ERROR;
        }
    }

    /* Print a header for the attitude and velocities image top */
    status = fprintf( ouf,
        "\nSatellite Attitudes, Velocities and Rates at Lunar image top\n");
    if ( status < 0 )
    {
        IAS_LOG_ERROR("Writing to report file");
        fclose( ouf );
        return ERROR;
    }
    /* Write the data */
    for ( index = 0; index < get_num_bands_to_process(); index++ )
    {
        status = fprintf( ouf,
            "%2d  Attitude: %13.6e, %13.6e, %13.6e\n    Velocity: %13.6e, "
            "%13.6e, %13.6e\n       Rates: %13.6e, %13.6e, %13.6e\n\n",
            out_info[index].band_number,
            out_info[index].spacecraft_top_att.x,
            out_info[index].spacecraft_top_att.y,
            out_info[index].spacecraft_top_att.z,
            out_info[index].spacecraft_top_vel.x,
            out_info[index].spacecraft_top_vel.y,
            out_info[index].spacecraft_top_vel.z,
            out_info[index].spacecraft_top_rates.x,
            out_info[index].spacecraft_top_rates.y,
            out_info[index].spacecraft_top_rates.z );
        if ( status < 0 )
        {
            IAS_LOG_ERROR("Writing to report file");
            fclose( ouf );
            return ERROR;
        }
    }

    /* Print a header for the attitude and velocities image center */
    status = fprintf( ouf,
        "Satellite Attitudes, Velocities and Rates at Lunar image center\n");
    if ( status < 0 )
    {
        IAS_LOG_ERROR("Writing to report file");
        fclose( ouf );
        return ERROR;
    }
    /* Write the data */
    for ( index = 0; index < get_num_bands_to_process(); index++ )
    {
        status = fprintf( ouf,
            "%2d  Attitude: %13.6e, %13.6e, %13.6e\n    Velocity: %13.6e, "
            "%13.6e, %13.6e\n       Rates: %13.6e, %13.6e, %13.6e\n\n",
            out_info[index].band_number,
            out_info[index].spacecraft_center_att.x,
            out_info[index].spacecraft_center_att.y,
            out_info[index].spacecraft_center_att.z,
            out_info[index].spacecraft_center_vel.x,
            out_info[index].spacecraft_center_vel.y,
            out_info[index].spacecraft_center_vel.z,
            out_info[index].spacecraft_center_rates.x,
            out_info[index].spacecraft_center_rates.y,
            out_info[index].spacecraft_center_rates.z );
        if ( status < 0 )
        {
            IAS_LOG_ERROR("Writing to report file");
            fclose( ouf );
            return ERROR;
        }
    }

    /* Print a header for the attitude and velocities image bottom */
    status = fprintf( ouf,
        "Satellite Attitudes, Velocities and Rates at Lunar image bottom\n");
    if ( status < 0 )
    {
        IAS_LOG_ERROR("Writing to report file");
        fclose( ouf );
        return ERROR;
    }
    /* Write the data */
    for ( index = 0; index < get_num_bands_to_process(); index++ )
    {
        status = fprintf( ouf,
            "%2d  Attitude: %13.6e, %13.6e, %13.6e\n    Velocity: %13.6e, "
            "%13.6e, %13.6e\n       Rates: %13.6e, %13.6e, %13.6e\n\n",
            out_info[index].band_number,
            out_info[index].spacecraft_bottom_att.x,
            out_info[index].spacecraft_bottom_att.y,
            out_info[index].spacecraft_bottom_att.z,
            out_info[index].spacecraft_bottom_vel.x,
            out_info[index].spacecraft_bottom_vel.y,
            out_info[index].spacecraft_bottom_vel.z,
            out_info[index].spacecraft_bottom_rates.x,
            out_info[index].spacecraft_bottom_rates.y,
            out_info[index].spacecraft_bottom_rates.z );
        if ( status < 0 )
        {
            IAS_LOG_ERROR("Writing to report file");
            fclose( ouf );
            return ERROR;
        }
    }

    status = fclose( ouf );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Closing report file: %s", filename );
        return ERROR;
    }

    return SUCCESS;
}  /* END generate_report */
