/*----------------------------------------------------------------------------
NAME:    get_illuminated_sca

PURPOSE: Determine which of the SCAs are illuminated by the moon.  There is
         a check for the possibility of multiple SCAs being illuminated.
         TODO: the latter is not implemented yet.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <string.h>
#include <stdlib.h>

#include "ias_logging.h"
#include "ias_const.h"
#include "ias_satellite_attributes.h"
#include "ias_db_query.h"
#include "ias_miscellaneous.h"
#include "lunar_irradiance_characterization.h"


int get_illuminated_sca
(
    struct ias_db_connection *db, /* I: Database connection information */
    int *illuminated_sca          /* O: SCA numbe / index */
)
{
    int status;
    int count;
    char query_string[300];
    char db_error_message[300];
    struct ias_db_query *query;
    char radiometric_wo_common_id[21];
    char *work_order_id = get_work_order_id();
    struct stats
    {
        int sca_number;
        int max;
        double mean;
        double stdev;
    };
    struct stats *sca_statistics;
    int num_scas = ias_sat_attr_get_sensor_sca_count( IAS_OLI );
    if ( num_scas == ERROR )
    {
        IAS_LOG_ERROR("Getting the number of SCAs for IAS_OLI");
        return ERROR;
    }

    /* The SCA is invalid until set at the end */
    *illuminated_sca = -1;

    /* Get the common ID so we can access the statistics */
    status = ias_misc_get_char_id( db, "radiometric_work_order_common",
        "radiometric_wo_common_id", work_order_id, radiometric_wo_common_id );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Obtaining radiometric work order common ID for %s",
            work_order_id );
        return ERROR;
    }

    /* Construct and execute a query to get the records */
    int largest_band_number = 9;
    status = snprintf( query_string, sizeof(query_string),
        "SELECT sensor_chip_assembly_id, MAX(scene_pixel_max), "
        "   MAX(scene_pixel_mean), MAX(scene_pixel_stdev) "
        "   FROM histogram_statistics_cal WHERE radiometric_wo_common_id = %s "
        "   AND band_id <= %d GROUP BY sensor_chip_assembly_id "
        "   ORDER BY MAX(scene_pixel_mean) DESC;", radiometric_wo_common_id,
        largest_band_number );
    if ( status >= sizeof(query_string) || status < 0 )
    {
        IAS_LOG_ERROR("Building query string");
        return ERROR;
    }

    IAS_LOG_INFO("Querying the database for illuminated SCA");
    query = ias_db_perform_query( db, query_string );
    if ( !ias_db_query_was_successful( query ))
    {
        ias_db_query_get_error_message( query, db_error_message,
            sizeof(db_error_message));
        IAS_LOG_ERROR("Query failed:\n%s\n-- End query", query_string );
        IAS_LOG_ERROR( "Message:\n%s\n-- End message", db_error_message );
        return ERROR;
    }
    IAS_LOG_INFO("The query is done");

    /* Allocate an array of statistics */
    sca_statistics = malloc( num_scas * sizeof(struct stats));
    if ( !sca_statistics )
    {
        IAS_LOG_ERROR("Allocating space for statistics");
        ias_db_query_close( query );
        return ERROR;
    }

    /* Extract the information and store in the structure */
    count = 0;
    while ( ias_db_query_next( query ) != 0 )
    {
        int local_int;
        double local_double;
        int index = 0;

        /* Get the SCA number */
        status = ias_db_query_int_value( query, &local_int, index++ );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Obtaining sensor_chip_assembly_id from record %d",
                index );
            free( sca_statistics );
            ias_db_query_close( query );
            return ERROR;
        }
        sca_statistics[count].sca_number = local_int;

        /* Get the max */
        status = ias_db_query_int_value( query, &local_int, index++ );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Obtaining scene_pixel_max from record %d",
                index );
            free( sca_statistics );
            ias_db_query_close( query );
            return ERROR;
        }
        sca_statistics[count].max = local_int;

        /* Get the mean */
        status = ias_db_query_double_value( query, &local_double, index++ );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Obtaining scene_pixel_mean from record %d",
                index );
            free( sca_statistics );
            ias_db_query_close( query );
            return ERROR;
        }
        sca_statistics[count].mean = local_double;

        /* Get the stdev */
        status = ias_db_query_double_value( query, &local_double, index++ );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Obtaining scene_pixel_stdev from record %d",
                index );
            free( sca_statistics );
            ias_db_query_close( query );
            return ERROR;
        }
        sca_statistics[count].stdev = local_double;

        count++;
    }  /* END for index loop */
    ias_db_query_close( query );
    query = NULL;
    db = NULL;

    /* There should be the correct number of SCAs */
    if ( count != num_scas )
    {
        IAS_LOG_ERROR("Only %d of %d SCAs of information was retrieved "
            "from the database", count, num_scas );
        free( sca_statistics );
        return ERROR;
    }

    /* The first SCA in the list should be the illuminated SCA.  If the
       next SCA in the list is adjacent ensure its standard deviation is
       smaller than the first one in the list.  This is the basic check and
       won't be able to be tweaked until after launch.

       1st check: are the SCAs adjacent? */
    if ( abs( sca_statistics[0].sca_number - sca_statistics[1].sca_number )
        == 1 )
    {
        /* 2nd check: is the standard deviation higher? */
        if ( sca_statistics[0].stdev < sca_statistics[1].stdev )
        {
            IAS_LOG_WARNING("SCA %d has the higest mean however, adjacent "
                "SCA %d has a higher standard deviation",
                sca_statistics[0].sca_number,
                sca_statistics[1].sca_number );
            IAS_LOG_WARNING("  SCA %d: mean %f stdev %f",
                sca_statistics[0].sca_number, sca_statistics[0].mean,
                sca_statistics[0].stdev );
            IAS_LOG_WARNING("  SCA %d: mean %f stdev %f",
                sca_statistics[1].sca_number, sca_statistics[1].mean,
                sca_statistics[1].stdev );
        }
    }

    IAS_LOG_INFO("SCA %d with mean of %f is illuminated",
        sca_statistics[0].sca_number, sca_statistics[0].mean );
    IAS_LOG_INFO("SCA %d with mean of %f is the one with the next lower mean",
        sca_statistics[1].sca_number, sca_statistics[1].mean );

    *illuminated_sca = sca_statistics[0].sca_number;
    return SUCCESS;
}  /* END get_illuminated_sca */
