/*--------------------------------------------------------------------------
NAME:    get_satellite_position

PURPOSE: Get the satellite position in true-of-date inertial cartesian
         coordinates (ECITOD) at the current line/sample of the image

RETURNS: SUCCESS or ERROR
--------------------------------------------------------------------------*/
#include <math.h>
#include "ias_los_model.h"
#include "ias_geo.h"
#include "ias_math.h"
#include "ias_logging.h"
#include "ias_const.h"

#include "lunar_irradiance_characterization.h"

int get_satellite_position
(
    double line,                        /* I: Line of interest (target) */
    double sample,                      /* I: Sample of interest (target) */
    int band_index,                     /* I: Band the line/sample are in */
    int sca_index,                      /* I: SCA the line/sample are in */
    IAS_SENSOR_DETECTOR_TYPE dettype,   /* I: Detector type */
    const IAS_LOS_MODEL *model,         /* I: LOS model information */
    IAS_VECTOR *satvel,                 /* O: */
    IAS_VECTOR *satatt,                 /* O: */
    IAS_VECTOR *satpos,                 /* O: Satellite position (meters) in
                                              inertial true of date
                                              coordinates */
    IAS_VECTOR *satrates,               /* O: Calculated attitude rates
                                              between line of interest and
                                              previous one */
    double *image_time,                 /* O: Time of the target image point
                                              (seconds of day) */
    int  *year,                         /* O: Year of image point */
    int  *day                           /* O: Day of image point */
)
{
    int status;
    double ephem_time[3];
    double ut1_utc;         /* Irregular Earth rotation, in seconds */
    IAS_VECTOR ecij2k_pos;
    double mjd;
    double x, y;
    IAS_CPF *cpf = get_cpf();
    IAS_VECTOR prev_satatt;
    double prev_image_time;
    double time_difference;

    /* Get the state vector for the previous line */
    status = ias_los_model_get_satellite_state_vector_at_location(
        line - 1, sample, band_index, sca_index, dettype, model, &ecij2k_pos,
        satvel, &prev_satatt, &prev_image_time, year, day );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Getting the state vector");
        return ERROR;
    }

    /* Get the state vector for the line of interest */
    status = ias_los_model_get_satellite_state_vector_at_location(
        line, sample, band_index, sca_index, dettype, model, &ecij2k_pos,
        satvel, satatt, image_time, year, day );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Getting the state vector");
        return ERROR;
    }

    ephem_time[0] = *year;
    ephem_time[1] = *day;
    ephem_time[2] = *image_time;

    /* Calculate the apparent attitude rates between the previous line and
       the line of interest */
    time_difference = *image_time - prev_image_time;
    satrates->x = ( satatt->x - prev_satatt.x ) / time_difference;
    satrates->y = ( satatt->y - prev_satatt.y ) / time_difference;
    satrates->z = ( satatt->z - prev_satatt.z ) / time_difference;

    status = ias_geo_compute_getmjdcoords( ephem_time, cpf, &mjd, &x, &y,
        &ut1_utc );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Getting modified julian day coordinates");
        return ERROR;
    }

    if (ias_geo_transform_j2k2tod( ut1_utc, &ecij2k_pos, ephem_time, satpos )
        != SUCCESS)
    {
        IAS_LOG_ERROR("Transforming satellite position to true-of-date");
        return ERROR;
    }
    // TODO hack
    satpos->x = ecij2k_pos.x;
    satpos->y = ecij2k_pos.y;
    satpos->z = ecij2k_pos.z;

    return SUCCESS;
}  /* END get_satellite_position */
