/*----------------------------------------------------------------------------
NAME:    inoperable_detector_replacement

PURPOSE: Replace the inoperable detector's data with the mean of its
         neighbors.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/

#include "ias_logging.h"
#include "ias_const.h"
#include "ias_pixel_mask.h"
#include "ias_radiometry.h"  /* definitions of the artifacts */
#include "lunar_irradiance_characterization.h"

int inoperable_detector_replacement
(
    float *image_ptr,           /* I/O: pointer to the image, on output the
                                        inoperable detector information has
                                        been replaced */
    int *extents,               /* I: */
    IAS_PIXEL_MASK *artifact_mask, /* I: Artifact mask generated by create
                                         L1R, contains inoperable detector
                                         information */
    IAS_PIXEL_MASK *lunar_mask  /* O: the mask a 1 indicates a moon pixel
                                      0 not the moon - or possibly an
                                      artifact */
)
{
    PIXEL_MASK_TYPE artifact = IAS_RPS_INOP_DETECTOR;
    IAS_PIXEL_MASK_ITERATOR *artifact_it;
    IAS_PIXEL_MASK_ITERATOR *lunar_it;
    IAS_PIXEL_MASK_SPAN span;
    // TODO PIXEL_MASK_TYPE *mask_subset;
    int current_detector;
    // TODO int next_detector;
    int left_detector = extents[2];
    int right_detector = extents[3];
    int inop_dets[500];
    int number_inop_dets;
    int num_dets = 494;
    int i;

    /* Get an iterator for the artifacts */
    artifact_it = ias_pm_get_iterator( artifact_mask, IAS_PM_INCLUDE,
        artifact );
    if ( artifact_it == NULL )
    {
        IAS_LOG_ERROR("Obtaining an iterator for the artifacts");
        return ERROR;
    }

    /* Get a list of the inoperable detectors.  An array list is much more
       useful than the mask in this context.  Since there could be multiple
       spans per detector returned, logic is needed to handle that.  We also
       want to remove any detectors that aren't within the extents of the
       lunar image.  However, the detectors that may be used for
       replacement, do have to be taken into account -- that is the +/-2 in
       the loop. */
    number_inop_dets = 0;
    current_detector = -1; /* Illegal detector number */
    while ( ias_pm_get_next_span( artifact_it, &span ))
    {
        if ( span.detector_index != current_detector )
        {
            /* Found the next (or first) inoperable detector.  See if it's
               within the lunar image or needed for replacement. */
            current_detector = span.detector_index;
            if ( current_detector >= left_detector - 2
                || current_detector <= right_detector + 2 )
            {
                inop_dets[number_inop_dets] = current_detector;
                number_inop_dets++;
            }
        }
    }
    ias_pm_destroy_iterator( artifact_it );

    /* If there are no inoperable detectors return */
    if ( number_inop_dets == 0 )
        return SUCCESS;

    /* Get an iterator for the lunar mask */
    lunar_it = ias_pm_get_iterator( lunar_mask, IAS_PM_INCLUDE, 0x1 );
    if ( lunar_it == NULL )
    {
        IAS_LOG_ERROR("Obtaining an iterator for the lunar mask");
    }

    /* Loop over all the detectors in the list looking for detectors to
       replace */
    for ( i = 0; i < number_inop_dets; i++ )
    {
        int inop_detector = inop_dets[i];
        int det_left_of_inop;
        int det_right_of_inop;

        /* Only replace detectors within the lunar image */
        if ( inop_detector < left_detector || inop_detector > right_detector )
            continue;

        /* It's unlikely that the left or right of the image will fall on
           the first or last detector, but the case must be handled */
        if ( inop_detector == 0 )
            det_left_of_inop = det_right_of_inop = 1;
        else if ( inop_detector == num_dets - 1)
            det_left_of_inop = det_right_of_inop = num_dets - 2;
        /* Otherwise the replacement detectors bracket the inoperable one */
        else
        {
            det_left_of_inop = inop_detector - 1;
            det_right_of_inop = inop_detector + 1;
        }

        /* Check the detector to the left of the current one -- if there is
           one */
        if ( i > 0 )
        {
            if ( inop_dets[i - 1] == det_left_of_inop )
                det_left_of_inop--;
        }

        /* May need a check for more than one inop det in a row */

        /* Get the start and end line for the four detectors */

    }

    return SUCCESS;
}  /* END inoperable_detector_replacement */
