/*----------------------------------------------------------------------------
NAME:    lunar_irradiance_characterization.c

PURPOSE: Characterize the lunar irradiance and store the characterizations
         in the database and/or report file.  Information about the spacecraft
         and lunar positions is also reported.

RETURNS: EXIT_SUCCESS or EXIT_FAILURE
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>
#include <libgen.h>

#include "ias_logging.h"
#include "ias_const.h"
#include "ias_db.h"
#include "lunar_irradiance_characterization.h"

/* TODO: define this to do valgrind testing */
//#define VALGRIND

int main( int argc, char *argv[] )
{
    int status;
    char *app_name;
    int illuminated_sca;
    SCA_INFORMATION *out_info;
    int num_bands_to_process;
    struct ias_db_connection *db = NULL;

    /* Strip off the path to the application so that the logging library
       doesn't log it and initialize the library */
    app_name = strdup( basename( argv[0] ));
    if ( !app_name )
    {
        IAS_LOG_ERROR("Allocating memory for application name");
        exit(EXIT_FAILURE);
    }
    ias_log_initialize( app_name );
    free( app_name );
    app_name = NULL;

    /* There must be one command line agrument: ODL parameter file */
    if ( argc != 2 )
    {
        IAS_LOG_ERROR("Missing input command line argument" );
        exit(EXIT_FAILURE);
    }

    /* Get input parameters from input files */ 
    status = read_parameters( argv[1] );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Reading input parameters");
        exit(EXIT_FAILURE);
    }
    IAS_LOG_INFO("Initiated lunar irradiance characterization");
    num_bands_to_process = get_num_bands_to_process();

/* TODO for valgrind reasons don't do the database */
#ifndef VALGRIND
    /* Establish a connection with the database */
    db = connect_to_database();
    if ( db == NULL )
    {
        IAS_LOG_ERROR("No database connection");
        exit(EXIT_FAILURE);
    }

    /* Determine the illuminated SCA */
    status = get_illuminated_sca( db, &illuminated_sca );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Unable to determine the illuminated SCA");
        ias_cpf_free( get_cpf());
        disconnect_from_database( db, IAS_DB_ROLLBACK );
        exit(EXIT_FAILURE);
    }
#else
    /* TODO: Since the database call is bypassed this needs to be the
             correct SCA */
    illuminated_sca = 1;
#endif

    /* Allocate the output information array */
    out_info = malloc( sizeof(SCA_INFORMATION) * num_bands_to_process );
    if ( out_info == NULL )
    {
        IAS_LOG_ERROR("Allocating output information array");
        ias_cpf_free( get_cpf());
        disconnect_from_database( db, IAS_DB_ROLLBACK );
        exit(EXIT_FAILURE);
    }

    /* Process the bands */
    status = process_bands( num_bands_to_process, illuminated_sca, out_info );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Processing the bands");
        free( out_info );
        ias_cpf_free( get_cpf());
        disconnect_from_database( db, IAS_DB_ROLLBACK );
        exit(EXIT_FAILURE);
    }
    ias_cpf_free( get_cpf());

/* TODO for valgrind reasons don't do the database */
#ifndef VALGRIND
    /* Store the output information in the database */
//    status = db_store_characterization( db, out_info );
status = SUCCESS;
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Storing to database");
        free( out_info );
        disconnect_from_database( db, IAS_DB_ROLLBACK );
        exit(EXIT_FAILURE);
    }
    disconnect_from_database( db, IAS_DB_COMMIT );

    /* Generate report */
    status = generate_report( out_info );
    free( out_info );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Generating report");
        exit(EXIT_FAILURE);
    }
#else
    free( out_info );
#endif

    IAS_LOG_INFO("Successful completion");
    return EXIT_SUCCESS;
}  /* END main */
