#ifndef LUNAR_IRRADIANCE_CHARACTERIZATION
#define LUNAR_IRRADIANCE_CHARACTERIZATION

#include "ias_los_model.h"
#include "ias_pixel_mask.h"
#include "ias_structures.h"
#include "ias_db.h"

typedef struct sca_information
{
    int band_number;
    int illuminated_sca;
    double irradiance;
    double apparent_y_size;
    int center_detector;
    int start_frame;
    int stop_frame;
    double image_time;
    int year;
    int day;
    IAS_VECTOR spacecraft_position;
    IAS_VECTOR spacecraft_top_vel;
    IAS_VECTOR spacecraft_top_att;
    IAS_VECTOR spacecraft_top_rates;
    IAS_VECTOR spacecraft_center_vel;
    IAS_VECTOR spacecraft_center_att;
    IAS_VECTOR spacecraft_center_rates;
    IAS_VECTOR spacecraft_bottom_vel;
    IAS_VECTOR spacecraft_bottom_att;
    IAS_VECTOR spacecraft_bottom_rates;
    double spacecraft_moon_distance;
    double earth_sun_distance;
    double earth_moon_distance;
} SCA_INFORMATION;

/* ----- Parameter prototypes ----- */
int read_parameters
(
    char *parameter_file_name   /* I: Input parameter file name */
);


char *get_work_order_id(void);
int *get_process_bandlist(void);
int get_num_bands_to_process(void);
int get_median_filter_size(void);
int get_write_lunar_mask(void);
char *get_l1r_filename(void);
char *get_l0r_filename(void);
char *get_pixel_mask_filename(void);
char *get_model_filename(void);
double *get_threshold_factors(void);
double *get_irradiance_conversion_factors(void);
double *get_along_ifov(void);
IAS_CPF *get_cpf(void);
char *get_report_filename(void);
void do_not_use_set_woid(char *);

/* ----- Other prototypes ----- */
int get_illuminated_sca
(
    struct ias_db_connection *db, /* I: Database connection information */
    int *illuminated_sca          /* O: SCA numbe / index */
);

int process_bands
(
    int num_bands_to_process,   /* I: Number of bands to process */
    int illuminated_sca,        /* I: SCA number illuminated by the moon */
    SCA_INFORMATION *out_info   /* O: */
);

int create_median_filter_mask
(
    float *image_ptr,           /* I: pointer to one SCA of image data */
    int number_input_lines,     /* I: */
    int number_input_samples,   /* I: */
    int filter_width,           /* I: width of the median filter to apply */
    double threshold,           /* I: threshold value for median filter */
    IAS_PIXEL_MASK *artifact_mask,/* I: Artifacts in the image */
    IAS_PIXEL_MASK *lunar_mask, /* O: the mask a 1 indicates a moon pixel
                                      0 not the moon - or possibly an
                                      artifact */
    int *extents                /* O: */
);

int compute_irradiance
(
    float *image_ptr,           /* I: pointer to one SCA of image data */
    int number_input_samples,   /* I: */
    IAS_PIXEL_MASK *lunar_mask, /* I: lunar mask where 1 indicates a moon
                                      pixel and 0 is not the moon */
    double conversion_factor,   /* I: Factor to convert to irradiance */
    double *irradiance          /* O: Irradiance */
);

int compute_apparent_y_size
(
    IAS_PIXEL_MASK *lunar_mask, /* I: lunar mask where 1 indicates a moon
                                      pixel and 0 is not the moon */
    double solid_angle,         /* I: */
    int *center_detector,       /* O: Detector with the longest span of
                                      moon pixels */
    int *start_frame,           /* O: Frame information for the center
                                      detector span */
    int *stop_frame,            /* O: Frame information for the center
                                      detector span */
    double *apparent_y_size     /* O: Apprarent y-size in mrads */
);

int get_satellite_position
(
    double line,                        /* I: Line of interest (target) */
    double sample,                      /* I: Sample of interest (target) */
    int band_index,                     /* I: Band the line/sample are in */
    int sca_index,                      /* I: SCA the line/sample are in */
    IAS_SENSOR_DETECTOR_TYPE dettype,   /* I: Detector type */
    const IAS_LOS_MODEL *model,         /* I: LOS model information */
    IAS_VECTOR *satvel,                 /* O: */
    IAS_VECTOR *satatt,                 /* O: */
    IAS_VECTOR *satpos,                 /* O: Satellite position (meters) in
                                              inertial true of date
                                              coordinates */
    IAS_VECTOR *satrates,               /* O: Calculated rates between line
                                              of interest and previous one */
    double *image_time,                 /* O: Time of the target image point
                                              (seconds of day) */
    int  *year,                         /* O: Year of image point */
    int  *day                           /* O: Day of image point */
);

struct ias_db_connection *connect_to_database(void);

int disconnect_from_database
(
    struct ias_db_connection *db, /* I: database connection information */
    int  commit         /* I: flag to indicate commiting the transaction */
);

int db_store_characterization
(
    struct ias_db_connection *db, /* I: Database connection information */
    SCA_INFORMATION *out_info     /* I: Information to be stored */
);

int generate_report
(
    SCA_INFORMATION *out_info   /* I: Information to be reported */
);

int determine_lunar_extents
(
    int number_input_lines,     /* I: */
    int number_input_samples,   /* I: */
    float *filtered_image,      /* I: Pointer to a smoothed image */
    double radiance_threshold,  /* I: Threshold for lunar pixels */
    int extents[6]              /* O: Extents ordered as top, bottom, left
                                      right, expanded top and expanded
                                      bottom indices */
);

int inoperable_detector_replacement
(
    float *image_ptr,           /* I/O: pointer to the image, on output the
                                        inoperable detector information has
                                        been replaced */
    int *extents,               /* I: */
    IAS_PIXEL_MASK *artifact_mask, /* I: Artifact mask generated by create
                                         L1R, contains inoperable detector
                                         information */
    IAS_PIXEL_MASK *lunar_mask  /* O: the mask a 1 indicates a moon pixel
                                      0 not the moon - or possibly an
                                      artifact */
);
#endif
