/*----------------------------------------------------------------------------
NAME:    process_bands.c

PURPOSE: Loop through the bands of the image and characterize the information
         needed for the reports and database.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>

#include "ias_logging.h"
#include "ias_const.h"
#include "ias_los_model.h"
#include "ias_model_io.h"
#include "ias_geo.h"
#include "ias_pixel_mask.h"
#include "ias_l1r.h"
#include "ias_satellite_attributes.h"
#include "lunar_irradiance_characterization.h"

#define AU_KM 149597870.7

/*----------------------------------------------------------------------------
NAME:    destroy_mask

PURPOSE: Destroy an array of pixel mask pointers

RETURNS: none
-----------------------------------------------------------------------------*/
static void destroy_mask
(
    IAS_PIXEL_MASK **mask,
    int number_of_masks
)
{
    int i;
    for ( i = 0; i < number_of_masks; i++ )
    {
        if ( mask[i] != NULL )
            ias_pm_destroy( mask[i] );
    }
}

/*----------------------------------------------------------------------------
NAME:    read_mask

PURPOSE: Read the pixel mask array from disk and pare down the array to only
         what we need.  We only need a mask for the illuminated SCA for the
         bands in the input list.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int read_mask
(
    int *list_of_bands,
    int number_of_bands,
    int illuminated_sca,
    IAS_PIXEL_MASK **artifact_mask
)
{
    int number_of_masks;
    IAS_PIXEL_MASK **pmask;
    int largest_band_number;
    int mask_index;
    int list_index;

    /* Get the artifact mask - this mask will contain artifacts identified
       by create_l1r and contains all bands and SCAs */
    pmask = ias_pm_read_array_from_file( get_pixel_mask_filename(),
        &number_of_masks );
    if ( !pmask )
    {
        IAS_LOG_ERROR("Obtaining artifact mask");
        return ERROR;
    }
    IAS_LOG_DEBUG("Read %d masks", number_of_masks );

    /* Determine the largest band number in the bands to process.  This is
       so we can easily get rid of masks for all the other bands. */
    largest_band_number = list_of_bands[0];
    for ( list_index = 1; list_index < number_of_bands; list_index++ )
    {
        if ( list_of_bands[list_index] > largest_band_number )
            largest_band_number = list_of_bands[list_index];
    }

    /* We only want to keep the masks that are in the band list with the
       illuminated SCA */
    for ( mask_index = 0; mask_index < number_of_masks; mask_index++ )
    {
        IAS_PIXEL_MASK *pm = pmask[mask_index];
        int band = ias_pm_get_band_number( pm );
        int sca = ias_pm_get_sca_number( pm );
        /* Only look at bands less than or equal to the largest band number
           and the illuminated SCA */
        if ( band <= largest_band_number && sca == illuminated_sca )
        {
            for ( list_index = 0; list_index < number_of_bands; list_index++ )
            {
                if ( band == list_of_bands[list_index] )
                {
                    artifact_mask[list_index] = pmask[mask_index];
                    pmask[mask_index] = NULL;
                    break;
                }
            }
        }

        /* If the mask wasn't identified and moved to the return parameter,
           destroy the mask */
        if ( pmask[mask_index] )
        {
            ias_pm_destroy( pmask[mask_index] );
            pmask[mask_index] = NULL;
        }
    }

    free( pmask );

    return SUCCESS;
}  /* END read_mask */


/*----------------------------------------------------------------------------
NAME:    process_bands

PURPOSE: Loop through the bands of the image and characterize the information
         needed for the reports and database.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int process_bands
(
    int num_bands_to_process,   /* I: Number of bands to process */
    int illuminated_sca,        /* I: SCA number illuminated by the moon */
    SCA_INFORMATION *out_info   /* O: */
)
{
    IAS_PIXEL_MASK *artifact_mask[num_bands_to_process];
    IAS_PIXEL_MASK *lunar_mask[num_bands_to_process];
    int *list_of_bands = get_process_bandlist();
    int status;
    int return_status;
    int index;
    char *input_filename = get_l1r_filename();
    L1RIO *input_image_ptr;     /* Pointer to input image */
    float *image_ptr = NULL;
    double *irradiance_conversion_factors = get_irradiance_conversion_factors();
    double *solid_angle = get_along_ifov();
    IAS_LOS_MODEL *model;
    double julian_date;
    double right_ascension_hrs;
    double declination_deg;
    double local_earth_moon_distance;
    double local_earth_sun_distance;
    double *threshold_factors = get_threshold_factors();
    const struct IAS_CPF_EARTH_CONSTANTS *earth_constants;
    double year_doy_sod[3];

    /* Initialize the time conversion routines */
    earth_constants = ias_cpf_get_earth_const( get_cpf());
    if ( ! earth_constants )
    {
        IAS_LOG_ERROR("Getting Earth constants from CPF");
        return ERROR;
    }
// TODO: This shouldn't be hard-coded -- the acquisition date should be used
    year_doy_sod[0] = 2013.0;
    year_doy_sod[1] = 100.0;
    year_doy_sod[2] = 0.0;
    status = ias_math_init_leap_seconds_from_UTC_time( year_doy_sod,
        &earth_constants->leap_seconds_data );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Initializing math library leap seconds");
        return ERROR;
    }

    /* Read the mask for the bands in the list and the illuminated SCA */
    status = read_mask( list_of_bands, num_bands_to_process, illuminated_sca,
        artifact_mask );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Reading artifact masks for illuminated SCAs");
        return ERROR;
    }

    /* Initialize the lunar_masks to NULL for the destroy routine */
    for ( index = 0; index < num_bands_to_process; index++ )
        lunar_mask[index] = NULL;

    /* Open the imagery */
    input_image_ptr = ias_l1r_open_image( input_filename, IAS_READ );
    if ( input_image_ptr == NULL )
    {
        IAS_LOG_ERROR("Opening %s for reading", input_filename );
        destroy_mask( artifact_mask, num_bands_to_process );
        return ERROR;
    }

    /* Open the model */
    model = ias_model_read( get_model_filename());
    if ( model == NULL )
    {
        IAS_LOG_ERROR("Reading model file %s", get_model_filename());
        destroy_mask( artifact_mask, num_bands_to_process );
        return ERROR;
    }

    /* Initialize NOVAS */
    status = ias_geo_initialize_novas();
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Initializing NOVAS library");
        destroy_mask( artifact_mask, num_bands_to_process );
        ias_los_model_free( model );
        return ERROR;
    }

    /* Process each band in the list */
    for ( index = 0; index < num_bands_to_process; index++ )
    {
        size_t allocated_size = 0;
        size_t allocate_amount;
        int band_number = list_of_bands[index];
        L1R_BAND_IO *input_band_io;
        IAS_DATA_TYPE data_type;
        int number_input_scas;
        int number_input_lines;
        int number_input_samples;
        IAS_PIXEL_MASK *lmask;
        int band_index;
        int center_frame;
        IAS_SENSOR_DETECTOR_TYPE dettype = IAS_NOMINAL_DETECTOR;
        IAS_VECTOR satpos;
        IAS_VECTOR satatt;
        IAS_VECTOR satvel;
        IAS_VECTOR satrates;
        double right_ascension;
        double declination;
        double image_time;
        int year;
        int month;
        int day;
        int extents[10];

        /* Get the band attributes */
        const IAS_BAND_ATTRIBUTES *band_attributes =
            ias_sat_attr_get_band_attributes( band_number );
        if ( band_attributes == NULL )
        {
            IAS_LOG_ERROR("Cannot retrieve band %d attributes",
                band_number );
            ias_l1r_close_image( input_image_ptr );
            destroy_mask( lunar_mask, num_bands_to_process );
            destroy_mask( artifact_mask, num_bands_to_process );
            ias_los_model_free( model );
            ias_geo_shutdown_novas();
            return ERROR;
        }
        band_index = band_attributes->band_index;

        IAS_LOG_INFO("Processing band %d", band_number );
        out_info[index].band_number = band_number;
        out_info[index].illuminated_sca = illuminated_sca;

        /* Open the band */
        input_band_io = ias_l1r_open_band( input_image_ptr,
            band_number, &data_type, &number_input_scas,
            &number_input_lines, &number_input_samples );
        if ( input_band_io ==  NULL )
        {
            IAS_LOG_ERROR("Opening input band %d image data",
                band_number );
            ias_l1r_close_image( input_image_ptr );
            destroy_mask( lunar_mask, num_bands_to_process );
            destroy_mask( artifact_mask, num_bands_to_process );
            ias_los_model_free( model );
            ias_geo_shutdown_novas();
            return ERROR;
        }

        /* Validate that the imgery is as expected */
        if ( data_type != IAS_R4 )
        {
            IAS_LOG_ERROR("Invalid data type for band %d L1R product",
                band_number );
            ias_l1r_close_band( input_band_io );
            ias_l1r_close_image( input_image_ptr );
            destroy_mask( lunar_mask, num_bands_to_process );
            destroy_mask( artifact_mask, num_bands_to_process );
            ias_los_model_free( model );
            ias_geo_shutdown_novas();
            return ERROR;
        }

        /* Allocate one SCA of data */
        allocate_amount = sizeof(float) * number_input_lines *
            number_input_samples;
        if ( allocated_size < allocate_amount )
        {
            float *temp_ptr  = realloc( image_ptr, allocate_amount );
            if ( temp_ptr == NULL )
            {
                IAS_LOG_ERROR("Reallocating image memory for band %d",
                    band_number );
                free( image_ptr );
                ias_l1r_close_band( input_band_io );
                ias_l1r_close_image( input_image_ptr );
                destroy_mask( lunar_mask, num_bands_to_process );
                destroy_mask( artifact_mask, num_bands_to_process );
                ias_los_model_free( model );
                ias_geo_shutdown_novas();
                return ERROR;
            }
            allocated_size = allocate_amount;
            image_ptr = temp_ptr;
        }

        /* Read the illuminated SCA */
        status = ias_l1r_read_image( input_band_io, illuminated_sca - 1, 0, 0,
            number_input_lines, number_input_samples, image_ptr );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Reading band %d SCA %d", band_number,
                illuminated_sca );
            free( image_ptr );
            ias_l1r_close_band( input_band_io );
            ias_l1r_close_image( input_image_ptr );
            destroy_mask( lunar_mask, num_bands_to_process );
            destroy_mask( artifact_mask, num_bands_to_process );
            ias_los_model_free( model );
            ias_geo_shutdown_novas();
            return ERROR;
        }

        /* Create a lunar mask: this involves running a median filter over
           the image; finding the maximum radiance value; setting a
           threshold; and setting the mask to 1 where the pixel radiance is
           greater than or equal to the threshold; finally transferring
           artifacts from the image to the mask effectively ignoring them. */
        lmask = ias_pm_create( band_number, illuminated_sca,
            number_input_samples, number_input_lines );
        if ( lmask == NULL )
        {
            IAS_LOG_ERROR("Creating lunar mask for band %d SCA %d",
                band_number, illuminated_sca );
            free( image_ptr );
            ias_l1r_close_band( input_band_io );
            ias_l1r_close_image( input_image_ptr );
            destroy_mask( lunar_mask, num_bands_to_process );
            destroy_mask( artifact_mask, num_bands_to_process );
            ias_los_model_free( model );
            ias_geo_shutdown_novas();
            return ERROR;
        }
        lunar_mask[index] = lmask;
        status = create_median_filter_mask( image_ptr, number_input_lines,
            number_input_samples, get_median_filter_size(),
            threshold_factors[band_index],
            artifact_mask[index], lmask, extents );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Applying median filter to band %d SCA %d",
                band_number, illuminated_sca );
            free( image_ptr );
            ias_l1r_close_band( input_band_io );
            ias_l1r_close_image( input_image_ptr );
            destroy_mask( lunar_mask, num_bands_to_process );
            destroy_mask( artifact_mask, num_bands_to_process );
            ias_los_model_free( model );
            ias_geo_shutdown_novas();
            return ERROR;
        }

// TODO
#if 0
        /* Replace any detectors that are inoperable with data that is the
           mean of the detectors on either side or the inoperable one */
        status = inoperable_detector_replacement( image_ptr,
            extents, artifact_mask[index], lmask );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Replacing inoperable detector for band %d SCA %d",
                band_number, illuminated_sca );
            free( image_ptr );
            ias_l1r_close_band( input_band_io );
            ias_l1r_close_image( input_image_ptr );
            destroy_mask( lunar_mask, num_bands_to_process );
            destroy_mask( artifact_mask, num_bands_to_process );
            ias_los_model_free( model );
            ias_geo_shutdown_novas();
            return ERROR;
        }
#endif

        /* Calculate the irradiance: sum all pixels marked as 1 in the
           mask; then convert to irradiance in mW/(m^2 nm) */
        status = compute_irradiance( image_ptr, number_input_samples,
            lmask, irradiance_conversion_factors[band_index],
            &out_info[index].irradiance );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Computing irradiance for band %d", band_number );
            free( image_ptr );
            ias_l1r_close_band( input_band_io );
            ias_l1r_close_image( input_image_ptr );
            destroy_mask( lunar_mask, num_bands_to_process );
            destroy_mask( artifact_mask, num_bands_to_process );
            ias_los_model_free( model );
            ias_geo_shutdown_novas();
            return ERROR;
        }


        /* Close the band */
        status = ias_l1r_close_band( input_band_io );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Closing input band %d", band_number );
            free( image_ptr );
            ias_l1r_close_image( input_image_ptr );
            destroy_mask( lunar_mask, num_bands_to_process );
            destroy_mask( artifact_mask, num_bands_to_process );
            ias_los_model_free( model );
            ias_geo_shutdown_novas();
            return ERROR;
        }


        /* Calculate the apparent lunar y-size */
        status = compute_apparent_y_size( lmask, solid_angle[band_index],
            &out_info[index].center_detector, &out_info[index].start_frame,
            &out_info[index].stop_frame, &out_info[index].apparent_y_size );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Computing apparent y-size for band %d",
                band_number );
            free( image_ptr );
            ias_l1r_close_image( input_image_ptr );
            destroy_mask( lunar_mask, num_bands_to_process );
            destroy_mask( artifact_mask, num_bands_to_process );
            ias_los_model_free( model );
            ias_geo_shutdown_novas();
            return ERROR;
        }

        /* Get the attitude and rates at the top of the lunar image */
        status = get_satellite_position(
            out_info[index].start_frame, out_info[index].center_detector,
            band_index, illuminated_sca - 1, dettype, model, &satvel,
            &satatt, &satpos, &satrates, &image_time, &year, &day );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Getting the state vector at image top");
            free( image_ptr );
            ias_l1r_close_image( input_image_ptr );
            destroy_mask( lunar_mask, num_bands_to_process );
            destroy_mask( artifact_mask, num_bands_to_process );
            ias_los_model_free( model );
            ias_geo_shutdown_novas();
            return ERROR;
        }
        out_info[index].spacecraft_top_vel.x = satvel.x;
        out_info[index].spacecraft_top_vel.y = satvel.y;
        out_info[index].spacecraft_top_vel.z = satvel.z;
        out_info[index].spacecraft_top_att.x = satatt.x;
        out_info[index].spacecraft_top_att.y = satatt.y;
        out_info[index].spacecraft_top_att.z = satatt.z;
        out_info[index].spacecraft_top_rates.x = satrates.x;
        out_info[index].spacecraft_top_rates.y = satrates.y;
        out_info[index].spacecraft_top_rates.z = satrates.z;
        

        /* Get the attitude and rates at the bottom of the lunar image */
        status = get_satellite_position(
            out_info[index].stop_frame, out_info[index].center_detector,
            band_index, illuminated_sca - 1, dettype, model, &satvel,
            &satatt, &satpos, &satrates, &image_time, &year, &day );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Getting the state vector at image top");
            free( image_ptr );
            ias_l1r_close_image( input_image_ptr );
            destroy_mask( lunar_mask, num_bands_to_process );
            destroy_mask( artifact_mask, num_bands_to_process );
            ias_los_model_free( model );
            ias_geo_shutdown_novas();
            return ERROR;
        }
        out_info[index].spacecraft_bottom_vel.x = satvel.x;
        out_info[index].spacecraft_bottom_vel.y = satvel.y;
        out_info[index].spacecraft_bottom_vel.z = satvel.z;
        out_info[index].spacecraft_bottom_att.x = satatt.x;
        out_info[index].spacecraft_bottom_att.y = satatt.y;
        out_info[index].spacecraft_bottom_att.z = satatt.z;
        out_info[index].spacecraft_bottom_rates.x = satrates.x;
        out_info[index].spacecraft_bottom_rates.y = satrates.y;
        out_info[index].spacecraft_bottom_rates.z = satrates.z;

        /* Get the model information at the center of the moon's image on
           the SCA */
        center_frame = ( out_info[index].stop_frame +
            out_info[index].start_frame ) / 2;
        status = get_satellite_position(
            center_frame, out_info[index].center_detector, band_index,
            illuminated_sca - 1, dettype, model, &satvel, &satatt, &satpos,
            &satrates, &out_info[index].image_time, &out_info[index].year,
            &out_info[index].day );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Getting the state vector");
            free( image_ptr );
            ias_l1r_close_image( input_image_ptr );
            destroy_mask( lunar_mask, num_bands_to_process );
            destroy_mask( artifact_mask, num_bands_to_process );
            ias_los_model_free( model );
            ias_geo_shutdown_novas();
            return ERROR;
        }
        out_info[index].spacecraft_position.x = satpos.x;
        out_info[index].spacecraft_position.y = satpos.y;
        out_info[index].spacecraft_position.z = satpos.z;
        out_info[index].spacecraft_center_vel.x = satvel.x;
        out_info[index].spacecraft_center_vel.y = satvel.y;
        out_info[index].spacecraft_center_vel.z = satvel.z;
        out_info[index].spacecraft_center_att.x = satatt.x;
        out_info[index].spacecraft_center_att.y = satatt.y;
        out_info[index].spacecraft_center_att.z = satatt.z;
        out_info[index].spacecraft_center_rates.x = satrates.x;
        out_info[index].spacecraft_center_rates.y = satrates.y;
        out_info[index].spacecraft_center_rates.z = satrates.z;


        /* Get the spacecraft to moon distance */
        status = ias_los_model_get_moon_position_at_location( model,
            band_index, illuminated_sca - 1, center_frame,
            out_info[index].center_detector, dettype, &right_ascension,
            &declination, &out_info[index].spacecraft_moon_distance );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Getting the the moon position");
            free( image_ptr );
            ias_l1r_close_image( input_image_ptr );
            destroy_mask( lunar_mask, num_bands_to_process );
            destroy_mask( artifact_mask, num_bands_to_process );
            ias_los_model_free( model );
            ias_geo_shutdown_novas();
            return ERROR;
        }
        out_info[index].spacecraft_moon_distance /= 1000.0;

        /* Calculate the earth-sun and earth-moon distances */
        status = ias_math_convert_doy_to_month_day( out_info[index].day,
            out_info[index].year, &month, &day );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Converting date");
            return_status = ERROR;
        }
        julian_date = ias_math_compute_full_julian_date( out_info[index].year,
            month, day, out_info[index].image_time );

        status = ias_geo_compute_moon_position( julian_date,
            &right_ascension_hrs, &declination_deg,
            &local_earth_moon_distance );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Getting Earth-moon distance");
            return_status = ERROR;
        }
        out_info[index].earth_moon_distance = local_earth_moon_distance;

        status = ias_geo_compute_sun_position( julian_date,
            &right_ascension_hrs, &declination_deg, &local_earth_sun_distance );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Getting Earth-sun distance");
            return_status = ERROR;
        }
        /* Convert to Km */
        out_info[index].earth_sun_distance = local_earth_sun_distance * AU_KM;
    }  /* END for index loop */

    return_status = SUCCESS;
    status = ias_pm_write_array_to_file( lunar_mask, num_bands_to_process,
        "lunar_mask.dat" );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Writing lunar_mask.dat");
        return_status = ERROR;
    }

    /* Clean up */
    free( image_ptr );
    status = ias_l1r_close_image( input_image_ptr );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Closing L1R image");
        return_status = ERROR;
    }
    status = ias_geo_shutdown_novas();
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Shutting down NOVAS library");
        return_status = ERROR;
    }
    destroy_mask( lunar_mask, num_bands_to_process );
    destroy_mask( artifact_mask, num_bands_to_process );
    ias_los_model_free( model );
    return return_status;
}  /* END process_bands */
