/*----------------------------------------------------------------------------
NAME:    read_parameters

PURPOSE: Read the input ODL file and some CPF parameters populating static
         variables that are accessed through getter routines.  The CPF is
         opened and left open on success for a geo lib routine that
         shouldn't, but does, expect to access the CPF instead of an
         individual CPF group.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <string.h>
#include <stdlib.h>

#include "ias_logging.h"
#include "ias_const.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "ias_satellite_attributes.h"
#include "ias_cpf.h"
#include "lunar_irradiance_characterization.h"


/* Static parameter values which are accessed by the getter routines that
   follow the read_parameters routine */
static char work_order_id[50];
static int process_bandlist[IAS_MAX_NBANDS];
static int num_bands_to_process = 0;
static int median_filter_size;
static int write_lunar_mask;
static char l1r_filename[256];
static char l0r_filename[256];
static char pixel_mask_filename[256];
static char model_filename[256];
static double threshold_factors[IAS_MAX_NBANDS];
static double irradiance_conversion_factors[IAS_MAX_NBANDS];
static double along_ifov[IAS_MAX_NBANDS];
static IAS_CPF *cpf = NULL;
static char report_filename[256];


int read_parameters
(
    char *parameter_filename    /* I: Input parameter file name */
)
{
    char omf_filename[256];
    char cpf_filename[256];
    int parameter_bandlist[IAS_MAX_NBANDS];
    int count;
    int status;
    int satellite_number;
    int satellite_id;
    int alternate_median_filter_size;
    int band_list[IAS_MAX_NBANDS];
    int num_bands;
    const IAS_CPF_LUNAR_IRRADIANCE *cpf_factors;
    const IAS_CPF_OLI_PARAMETERS *oli_parameters;
    int band_number;
    int pan_index;
    int i;

    /*-----------------------------------------------------------------*/
    /* This is the table definition for things from the parameter file */
    /*-----------------------------------------------------------------*/
    IAS_PARM_DECLARE_TABLE( parms, 4 );

    IAS_PARM_WORK_ORDER_ID( parms, work_order_id, sizeof(work_order_id), 1 );

    IAS_PARM_BAND_LIST( parms, IAS_PARM_OPTIONAL, parameter_bandlist,
        sizeof(parameter_bandlist), 1 );

    IAS_PARM_ADD_INT( parms, ALTERNATE_MEDIAN_FILTER_SIZE,
        "Alternate median filter width.  Must be >= 3",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        0, 0, 0, /* no limits on input -- limit checks are done later */
        0, NULL, /* no default */
        &alternate_median_filter_size, sizeof(alternate_median_filter_size),
        0 );

    const char *default_report_filename[] = {"lunar_irradiance_char.rpt"};
    IAS_PARM_ADD_STRING( parms, REPORT_FILENAME, "Output report file name",
        IAS_PARM_OPTIONAL, 
        0, NULL, /* no restrictions */
        1, default_report_filename, /* default report file name */
        report_filename, sizeof(report_filename), 0 );

    /*----------------------------------------------------------------*/
    /*----- This is the table definition for things from the OMF -----*/
    /*----------------------------------------------------------------*/
    IAS_PARM_DECLARE_TABLE( omf_parms, 10 );

    IAS_PARM_L0R_FILENAME( omf_parms, IAS_PARM_REQUIRED, l0r_filename,
        sizeof(l0r_filename), 0 );
    IAS_PARM_L1R_FILENAME( omf_parms, IAS_PARM_REQUIRED, l1r_filename,
        sizeof(l1r_filename), 0 );
    IAS_PARM_SATELLITE( omf_parms, &satellite_number,
        sizeof(satellite_number), 1 );
    IAS_PARM_CPF( omf_parms, 1, cpf_filename, sizeof(cpf_filename), 1 );
    IAS_PARM_ADD_STRING( omf_parms, PIXEL_MASK_FILENAME,
        "Pixel mask file name", IAS_PARM_REQUIRED, 
        0, NULL, /* no limits */
        0, NULL, /* no default */
        pixel_mask_filename, sizeof(pixel_mask_filename), 1 );
    IAS_PARM_MODEL_FILENAME_FROM_PASS( omf_parms, IAS_PARM_REQUIRED,
        model_filename, sizeof(model_filename), 0, 1 );


    /* See if the user wanted help or a template */
    status = ias_parm_provide_help( parameter_filename, parms,
        IAS_PARM_GET_TABLE_SIZE( parms ), IAS_INPUT_PARAMETERS);
    if ( status == 1 )
    {
        /* If the status of the provide help is 1 the user asked for help.
           Since the parameter file name contains the help request, copy it
           to the omf file name so the next call will ask for the same help */
        strcpy( omf_filename, parameter_filename );
    }
    else if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Returned from parameter file help");
        return ERROR;
    }
    else
    {
        /* Read the parameter file */
        count = IAS_PARM_GET_TABLE_SIZE(parms);
        status = ias_parm_read( parameter_filename, parms, count );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR( "Reading parameters from: %s", parameter_filename );
            return ERROR;
        }

        /* Build the omf file name */
        sprintf( omf_filename, "%s.omf", work_order_id );
    }

    /* See if the user wanted help or a template this time with the OMF file */
    status = ias_parm_provide_help( omf_filename, omf_parms,
        IAS_PARM_GET_TABLE_SIZE( omf_parms ), IAS_OMF_PARAMETERS);
    if ( status == 1 )
    {
        exit( EXIT_SUCCESS );
    }
    else if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Returned from OMF parameter file help");
        return ERROR;
    }

    /* Read the OMF file */
    count = IAS_PARM_GET_TABLE_SIZE( omf_parms );
    status = ias_parm_read( omf_filename, omf_parms, count );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Reading parameters from OMF file: %s", omf_filename );
        return ERROR;
    }

    /* Get the platform name and initialize the satellite attributes */
    satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(
        satellite_number );
    status = ias_sat_attr_initialize( satellite_id );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Initializing satellite attributes");
        return ERROR;
    }

    /* Get a list of the normal bands */
    status = ias_sat_attr_get_sensor_band_numbers( IAS_OLI, IAS_NORMAL_BAND,
        0, band_list, IAS_MAX_NBANDS, &num_bands );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Retrieving the list of normal bands");
        return ERROR;
    }

    /* If the band list was specified copy the applicable bands to the list
       of bands to process.  Remove duplicates. */
    count = IAS_PARM_GET_COUNT_READ( parms, BAND_LIST );
    if ( count )
    {
        num_bands_to_process = 0;
        for ( i = 0; i < count; i++ )
        {
            int j;
            for ( j = 0; j < num_bands; j++ )
            {
                if ( parameter_bandlist[i] == band_list[j] )
                {
                    process_bandlist[num_bands_to_process++] =
                        parameter_bandlist[i];
                    band_list[j] = -1;
                    continue;
                }
            }
        }
    }
    else
    {
        for ( i = 0; i < num_bands; i++ )
            process_bandlist[i] = band_list[i];
        num_bands_to_process = num_bands;
    }

    /* Read the CPF parameters */
    cpf = ias_cpf_read( cpf_filename );
    if ( !cpf )
    {
        IAS_LOG_ERROR("Opening CPF %s", cpf_filename );
        return ERROR;
    }

    /* Get the lunar irradiance factors */
    cpf_factors = ias_cpf_get_lunar_irradiance( cpf );
    if ( !cpf_factors )
    {
        IAS_LOG_ERROR("Reading the integration threshold factors");
        ias_cpf_free( cpf );
        return ERROR;
    }

    /* Set local copies of the values */
    for ( i = 0; i < num_bands; i++ )
    {
        irradiance_conversion_factors[i] =
            cpf_factors->irradiance_conversion[i];
        threshold_factors[i] = cpf_factors->integration_threshold_factor[i];
    }
    median_filter_size = cpf_factors->median_filter_size;

    /* Find out what the pan band index is so the solid angle can be
       assigned apppropriately */
    band_number = ias_sat_attr_get_band_number_from_type( IAS_PAN_BAND );
    if ( status == ERROR )
    {
        IAS_LOG_ERROR("Getting band number for IAS_PAN_BAND type");
        ias_cpf_free( cpf );
        return ERROR;
    }
    pan_index = ias_sat_attr_convert_band_number_to_index( band_number );
    if ( status == ERROR )
    {
        IAS_LOG_ERROR("Converting band %d to index", band_number );
        ias_cpf_free( cpf );
        return ERROR;
    }

    /* Get the solid angle (as specified in the ADD) a.k.a. IFOV */
    oli_parameters = ias_cpf_get_oli_parameters( cpf );
    if ( oli_parameters == NULL )
    {
        IAS_LOG_ERROR("Reading the OLI parameters");
        ias_cpf_free( cpf );
        return ERROR;
    }

    /* For simplicity save the along track IFOV for MS in all the bands */
    for ( i = 0; i < num_bands; i++ )
        along_ifov[i] = oli_parameters->along_ifov_ms;
    /* and now override the pan band IFOV */
    along_ifov[pan_index] = oli_parameters->along_ifov_pan;

    /* See if the median filter override was provided */
    count = IAS_PARM_GET_COUNT_READ( parms, ALTERNATE_MEDIAN_FILTER_SIZE );
    if ( count == 1 )
    {
        if ( alternate_median_filter_size < 3 )
        {
            IAS_LOG_ERROR("Alternate median filter size %d must be >= 3",
                median_filter_size );
            ias_cpf_free( cpf );
            return ERROR;
        }
        median_filter_size = alternate_median_filter_size;
    }
    
    return SUCCESS;
}  /* END read_parameters */


/* -------------------- Public getter routines -------------------- */
char *get_work_order_id(void)
{
    return work_order_id;
}

int *get_process_bandlist(void)
{
    return process_bandlist;
}

int get_num_bands_to_process(void)
{
    return num_bands_to_process;
}

int get_median_filter_size(void)
{
    return median_filter_size;
}

int get_write_lunar_mask(void)
{
    return write_lunar_mask;
}

char *get_l1r_filename(void)
{
    return l1r_filename;
}

char *get_l0r_filename(void)
{
    return l0r_filename;
}

char *get_pixel_mask_filename(void)
{
    return pixel_mask_filename;
}

char *get_model_filename(void)
{
    return model_filename;
}

double *get_threshold_factors(void)
{
    return threshold_factors;
}

double *get_irradiance_conversion_factors(void)
{
    return irradiance_conversion_factors;
}

double *get_along_ifov(void)
{
    return along_ifov;
}

IAS_CPF *get_cpf(void)
{
    return cpf;
}

void do_not_use_set_woid
(
    char *woid
)
{
    strcpy( work_order_id, woid );
}

char *get_report_filename(void)
{
    return report_filename;
}

