#include <stdio.h>
#include <limits.h>

#include "cgreen/cgreen.h"
#include "ias_logging.h"
#include "../lunar_irradiance_characterization.h"

static char *unit_test_dir = NULL;
static char subdir[] = "ias/rps/lunar_irradiance_characterization";

Ensure test_compute_apparent_y_size(void);
Ensure test_create_median_filter_mask(void);
Ensure test_get_illuminated_sca(void);

int main ( int argc, char *argv[] )
{
    TestSuite *suite = create_test_suite();
    add_test( suite, test_compute_apparent_y_size );
    add_test( suite, test_create_median_filter_mask );

    /* Only do the tests that depend on the database if the database tests
       are enabled */
    if ( getenv("ENABLE_DB_TESTS"))
    {
        add_test( suite, test_get_illuminated_sca );
    }
    else
    {
        printf("*** Some tests skipped since database checks are not "
            "enabled\n");
    }

    unit_test_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    if ( unit_test_dir == NULL )
    {
        printf("Error getting environment variable: IAS_UNIT_TEST_DATA_DIR\n");
        return 1;
    }

    if ( argc > 1 )
        return run_single_test( suite, argv[1], create_text_reporter());

    return run_test_suite( suite, create_text_reporter());
}  /* END main */


/*---------------------------------------------------------------------------
   NAME: read_pixel_mask_file

PURPOSE: Read a pixel mask from the unit test directory.

RETURNS: Pointer to array of masks or NULL
---------------------------------------------------------------------------*/
IAS_PIXEL_MASK **read_pixel_mask_file
(
    char *file_name,    /* I: file name excluding path, path is assumed to
                              be unit_test_dir/subdir */
    int *number_of_masks/* O: Number of masks read from file */
)
{
    char file[PATH_MAX];
    IAS_PIXEL_MASK **mask;

    sprintf( file, "%s/%s/%s", unit_test_dir, subdir, file_name );
    ias_log_set_output_level( IAS_LOG_LEVEL_DISABLE );
    mask = ias_pm_read_array_from_file( file, number_of_masks );
    ias_log_set_output_level( IAS_LOG_LEVEL_INFO );
    return mask;
}


Ensure test_compute_apparent_y_size(void)
{
    char lunar_mask_filename[] = "b5_lunar_mask";
    IAS_PIXEL_MASK **lunar_mask;
    int number_of_masks;
    int status;
    double solid_angle = 0.04255; // TODO: 4.255e-2 != 4.2553e-5;
    int center_detector;
    int start_frame;
    int stop_frame;
    double apparent_y_size;

    lunar_mask = read_pixel_mask_file( lunar_mask_filename, &number_of_masks );
    if ( lunar_mask == NULL )
    {
        assert_true_with_message( 0, "Error reading lunar mask %s",
            lunar_mask_filename );
        return;
    }
    assert_equal( number_of_masks, 1 );

    status = compute_apparent_y_size( lunar_mask[0], solid_angle,
    &center_detector, &start_frame, &stop_frame, &apparent_y_size );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error computing apparent y-size");
        ias_pm_destroy( lunar_mask[0] );
        return;
    }
    assert_equal( 219, center_detector );
    assert_equal( 110, start_frame );
    assert_equal( 1918, stop_frame );
    assert_double_equal( 76.930400, apparent_y_size );

    ias_pm_destroy( lunar_mask[0] );
}  /* END test_compute_apparent_y_size */


/* The main purpose of this test if for the testing the creation of the
   lunar mask.  Secondarily, since it's easy to do here, test computation of
   the irradiance value. */
Ensure test_create_median_filter_mask(void)
{
    char subset[PATH_MAX];
    char file_name[] = "b5_subset";
    FILE *inf;
    int i;
    int status;
    int num_samps = 494;
    int num_lines = 2000;
    int nread;
    short *short_buf;
    float *float_buf;
    double conversion_factor = 1.8108e-9 * 1e4; /* Directly from prototype */
    double irradiance;
    IAS_PIXEL_MASK **artifact_mask;
    IAS_PIXEL_MASK **lunar_mask;
    IAS_PIXEL_MASK *lmask;
    int number_of_masks;
    IAS_PIXEL_MASK_SPAN computed_span;
    IAS_PIXEL_MASK_SPAN expected_span;
    IAS_PIXEL_MASK_ITERATOR *computed_it;
    IAS_PIXEL_MASK_ITERATOR *expected_it;
    int span_count;
    int error_count;
    int extents[10];

    sprintf( subset, "%s/%s/%s", unit_test_dir, subdir, file_name );
    inf = fopen( subset, "r" );
    if ( !inf )
    {
        assert_true_with_message( 0, "Error opening file %s", subset );
        return;
    }

    short_buf = malloc( sizeof(short) * num_samps * num_lines );
    float_buf = malloc( sizeof(float) * num_samps * num_lines );
    if ( !short_buf || !float_buf )
    {
        assert_true_with_message( 0, "Error allocating memory buffers");
        fclose( inf );
        return;
    }

    /* Read in the prototype image and convert it to floating point
       radiance units */
    nread = fread( short_buf, sizeof(short), num_samps * num_lines, inf );
    fclose( inf );
    if ( nread != num_samps * num_lines )
    {
        assert_true_with_message( 0, "Error reading from input file %s",
            subset );
        free( short_buf );
        free( float_buf );
        return;
    }
    for ( i = 0; i < num_samps * num_lines; i++ )
        float_buf[i] = (float)short_buf[i] / 300.0;
    free( short_buf );

    /* Create the lunar mask */
    lmask = ias_pm_create( 5, 1, num_samps, num_lines );
    if ( lmask == NULL )
    {
        assert_true_with_message( 0, "Error creating lunar mask");
        free( float_buf );
        return;
    }

// TODO: this creates a temporary mask
#if 0
    IAS_PIXEL_MASK *dmask1[1];
    dmask1[0] = lmask;
    status = ias_pm_write_array_to_file( dmask1, 1, "b4_artifact_mask" );
    if ( status != SUCCESS )
        printf("error\n");
#endif

    /* Get the artifact mask */
    artifact_mask = read_pixel_mask_file( "b5_artifact_mask",
        &number_of_masks );
    assert_equal( 1, number_of_masks );


    status = create_median_filter_mask( float_buf, num_lines, num_samps,
        5,   /* filter width */
        0.1, /* threshold from the prototype */
        artifact_mask[0], lmask, extents );
    ias_pm_destroy( artifact_mask[0] );
    free( artifact_mask );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error returned from create median "
            "filter mask routine");
        return;
    }

    /* Test the irradiance for this band */
    status = compute_irradiance( float_buf, num_samps, lmask,
        conversion_factor, &irradiance );
    assert_equal( SUCCESS, status );
    assert_double_equal( 20.428848, irradiance );
    free( float_buf );

// TODO: this creates a temporary mask
#if 1
    IAS_PIXEL_MASK *dmask[1];
    dmask[0] = lmask;
    status = ias_pm_write_array_to_file( dmask, 1, "b5_lunar_mask" );
    if ( status != SUCCESS )
        printf("error\n");
#endif

    /* Get the expected lunar mask */
    lunar_mask = read_pixel_mask_file( "b5_lunar_mask", &number_of_masks );
    assert_equal( 1, number_of_masks );

    /* Create iterators */
    expected_it = ias_pm_get_iterator( lunar_mask[0], IAS_PM_INCLUDE, 0x1 );
    computed_it = ias_pm_get_iterator( lmask, IAS_PM_INCLUDE, 0x1 );
    if ( !expected_it || !computed_it )
    {
        assert_true_with_message( 0, "Error getting iterators");
        ias_pm_destroy_iterator( expected_it );
        ias_pm_destroy_iterator( computed_it );
        ias_pm_destroy( lunar_mask[0] );
        free( lunar_mask );
        ias_pm_destroy( lmask );
        return;
    }

    /* Loop through the masks comparing spans */
    span_count = error_count = 0;
    while( ias_pm_get_next_span( computed_it, &computed_span ))
    {
        if ( ias_pm_get_next_span( expected_it, &expected_span ))
        {
            if ( computed_span.detector_index != expected_span.detector_index
                || computed_span.starting_pixel_index !=
                expected_span.starting_pixel_index
                || computed_span.length_of_span !=
                expected_span.length_of_span
                || computed_span.pixel_mask != expected_span.pixel_mask )
            {
                error_count++;
            }
        }
        else
        {
            assert_true_with_message( 0, "Did not get the same number of"
                "expected spans, exhausted spans at %d", span_count );
            break;
        }
        span_count++;
    }

    assert_equal( 0, error_count );
    assert_equal( 222, span_count );

    ias_pm_destroy_iterator( expected_it );
    ias_pm_destroy_iterator( computed_it );
    ias_pm_destroy( lunar_mask[0] );
    ias_pm_destroy( lmask );
    free( lunar_mask );

}  /* END test_create_median_filter_mask */


Ensure test_get_illuminated_sca(void)
{
    int illuminated_sca;
    int status;
    int satellite_id;
    struct ias_db_connection *db;

    satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number( 8 );
    status = ias_sat_attr_initialize( satellite_id );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing satellite attributes");
        return;
    }
    do_not_use_set_woid("LunarUT");

    db = connect_to_database();
    /* Diable logging to suppress the information messages */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = get_illuminated_sca( db, &illuminated_sca );
    assert_equal( SUCCESS, status );
    assert_equal( 1, illuminated_sca );
    disconnect_from_database( db, IAS_DB_ROLLBACK );
}
