/***************************************************************************
 NAME:     calculate_rms_banding_metrics

 PURPOSE:  Calculates per-detector banding nonuniformity metrics based on
           the entire field-of-view statistics within a user-specified
           subset of detectors of the OLI/TIRS focal plane

 RETURNS:  Integer status code of SUCCESS or ERROR

****************************************************************************/
#include "ias_logging.h"
#include "ias_radiometry.h"

#include "char_nonuniformity.h"


int calculate_rms_banding_metrics
(
    const double *radiance_buffer,      /* I: Array of radiances in the
                                           current band */
    const int *statistics_invalid_flags,
                                        /* I: Input data invalid flags for
                                           all detectors in the current band
                                           (0=valid detector data,
                                            nonzero=invalid detector_data) */
    int num_band_detectors,             /* I: Total # of detectors in the
                                           current band */
    int num_contiguous_detectors,       /* I: WO parameter specifying
                                           banding width */
    double radiance_mean,               /* I: Mean radiance for the current
                                           band */
    double *rms_banding_buffer,         /* O: RMS banding metrics for the
                                           current band */
    int *metrics_invalid_flags /* I/O: Characterization results flag for all
                                       detectors in the current band -- it's
                                       expected to have been initialized to
                                       zero by the caller*/
)
{
    int start_detector_index;        /* 0-based detector indices defining
                                        focal plane region to characterize */
    int end_detector_index;
    int detector_index;              /* Current 0-based detector index */
    int status;

    /* Sanity check--data buffers must be valid. */
    if ((radiance_buffer == NULL) || (rms_banding_buffer == NULL)
        || (statistics_invalid_flags == NULL)
        || (metrics_invalid_flags == NULL))
    {
        IAS_LOG_ERROR("Invalid data buffer detected");
        return ERROR;
    }

    /* Determine the starting and ending indices for detectors that can
       be characterized */
    start_detector_index = 0;
    end_detector_index = num_band_detectors - num_contiguous_detectors;

    /* Characterize the operable/in-spec detectors */
    for (detector_index = start_detector_index;
         detector_index <= end_detector_index;
         detector_index++)
    {
        if (statistics_invalid_flags[detector_index])
        {
            IAS_LOG_DEBUG("Skipping detector %d -- currently "
                "flagged as inoperable / out-of-spec",
                (detector_index + 1));
            metrics_invalid_flags[detector_index] =
                RPS_NUCHAR_INVALID_RMS_BANDING;
        }

        /* Calculate the RMS value for the current detector */
        status = ias_rps_calculate_filtered_rms(
            &(radiance_buffer[detector_index]),
            &(statistics_invalid_flags[detector_index]),
            num_contiguous_detectors, radiance_mean,
            &(rms_banding_buffer[detector_index]), NULL);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating RMS value for detector %d",
                          (detector_index + 1));
            return ERROR;
        }
    }

    /* Cannot characterize these last detectors -- flag them */
    for (detector_index = end_detector_index + 1;
         detector_index < num_band_detectors;
         detector_index++)
    {
        metrics_invalid_flags[detector_index] =
            RPS_NUCHAR_INVALID_RMS_BANDING;
    }

    /* Everything appears to be OK. */
    return SUCCESS;
}

