/****************************************************************************
 NAME:              calculate_stdev_banding_metrics

 PURPOSE:   Calculates per-detector banding non-uniformity metrics based on
            the std deviation statistics within a user-specified subset of
            detectors of the OLI focal plane.

 RETURNS:   Integer status code of SUCCESS or ERROR

****************************************************************************/
#include "ias_logging.h"
#include "ias_radiometry.h"

#include "char_nonuniformity.h"


int calculate_stdev_banding_metrics
(
    const double *radiance_buffer,      /* I: Array of radiances in the
                                           current band */
    const int *statistics_invalid_flags,
                                        /* I: Input data invalid flags for
                                           all detectors in the current band
                                           (0=valid detector data,
                                            nonzero=invalid detector_data) */
    int    num_band_detectors,          /* I: Total # of detectors in the
                                           current band */
    int    num_contiguous_detectors,    /* I: WO parameter specifying
                                           banding width */
    double *stdev_banding_buffer,       /* O: stdev banding metrics for the
                                           current band */
    int *metrics_invalid_flags /* I/O: Characterization results flag for all
                                       detectors in the current band -- it's
                                       expected to have been initialized to
                                       zero by the caller*/
)
{
    double contiguous_mean;       /* Mean of radiances for contiguous
                                     detectors */
    double contiguous_stdev;      /* Std deviation of radiances for
                                     contiguous detectors */


    int detector_index;           /* Current 0-based detector index */
    int start_detector_index;     /* 0-based index for the first detector */
    int end_detector_index;       /* 0-based index for the last detector */
    int status;

    /* Sanity check for valid data buffers */
    if ((radiance_buffer == NULL) || (stdev_banding_buffer == NULL)
        || (statistics_invalid_flags == NULL)
        || (metrics_invalid_flags == NULL))
    {
        IAS_LOG_ERROR("Invalid data buffer detected");
        return ERROR;
    }

    /* Determine the starting and ending indices for detectors that can
       be characterized */
    start_detector_index = 0;
    end_detector_index = num_band_detectors - num_contiguous_detectors;

    /* Characterize the operable/in-spec detectors */
    for (detector_index = start_detector_index;
         detector_index <= end_detector_index;
         detector_index++)
    {
        if (statistics_invalid_flags[detector_index])
        {
            IAS_LOG_DEBUG("Skipping detector %d -- currently "
                "flagged as inoperable / out-of-spec",
                (detector_index + 1));
            metrics_invalid_flags[detector_index] =
                RPS_NUCHAR_INVALID_STDEV_BANDING;
            continue;
        }

        /* Calculate the current statistics (the usual mean and
           standard deviation) in the region between
           detector_index and detector_index + num_contigous_detectors. */
        status = ias_rps_calculate_filtered_mean(
                &(radiance_buffer[detector_index]),
                &(statistics_invalid_flags[detector_index]),
                num_contiguous_detectors,
                &contiguous_mean, NULL);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating contiguous mean for strip "
                    "starting at detector %d", (detector_index + 1));
            return ERROR;
        }

        status = ias_rps_calculate_filtered_stdev(
                &(radiance_buffer[detector_index]),
                &(statistics_invalid_flags[detector_index]),
                num_contiguous_detectors,
                contiguous_mean,
                &contiguous_stdev);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating contiguous stdev for strip "
                    "starting at detector %d", (detector_index + 1));
            return ERROR;
        }

        /* This calculates the "raw" metric value.  The final metric
           as a percentage of the mean band radiance will be calculated
           when these are written to the database/reported */
        stdev_banding_buffer[detector_index] = contiguous_stdev;
    }

    /* Can't characterize these detectors -- flag them */
    for (detector_index = end_detector_index + 1;
         detector_index < num_band_detectors;
         detector_index++)
    {
        metrics_invalid_flags[detector_index] =
            RPS_NUCHAR_INVALID_STDEV_BANDING;
    }

    /* Everything appears to be OK. */
    return SUCCESS;
}
