/***************************************************************************
 NAME:               calculate_streaking_metrics

 PURPOSE:    Calculates per-detector streaking metrics based on the
             comparison of the radiance of the current detector relative
             to the average of its immediate neighbors

 RETURNS:    Integer status code of SUCCESS or ERROR

****************************************************************************/
#include <math.h>

#include "ias_logging.h"

#include "char_nonuniformity.h"


int calculate_streaking_metrics
(
    const double *radiance_buffer, /* I: Array of radiances in the current
                                      band */
    const int *statistics_invalid_flags,
                                   /* I: Input data invalid flags for all
                                      detectors in the current band
                                      (0=valid detector data, nonzero=invalid
                                      detector_data) */
    int    num_band_detectors,     /* I: Total # of detectors in the band */
    double *streaking_buffer,      /* O: Streaking metrics for the current
                                      band */
    int *metrics_invalid_flags /* I/O: Characterization results flag for all
                                       detectors in the current band -- it's
                                       expected to have been initialized to
                                       zero by the caller*/
)
{
    double neighbor_mean;          /* Average radiance of nearest neighbor
                                      detectors to current detector */
    double numerator;              /* Intermediate calculation */

    int detector_index;             /* 0-based index for current
                                       detector */


    /* Sanity check -- make sure buffers have been allocated */
    if ((radiance_buffer == NULL) || (streaking_buffer == NULL)
        || (statistics_invalid_flags == NULL)
        || (metrics_invalid_flags == NULL))
    {
        IAS_LOG_ERROR( "Invalid data buffers detected" );
        return ERROR;
    }

    /* Calculate the streaking metric here.  Do not process "boundary"
       detectors (i.e. first and last detectors within the band),
       inoperable detectors, or detectors immediately adjacent to
       inoperable detectors. */
    metrics_invalid_flags[0] = metrics_invalid_flags[num_band_detectors - 1]
        = RPS_NUCHAR_INVALID_STREAKING;
    for (detector_index = 1;
         detector_index < num_band_detectors - 1;
         detector_index++)
    {
        /* Skip detectors flagged as inoperable / out-of-spec */
        if (statistics_invalid_flags[detector_index])
        {
            IAS_LOG_DEBUG("Skipping detector %d -- currently flagged "
                "as inoperable / out-of-spec", (detector_index + 1));
            metrics_invalid_flags[detector_index] =
                RPS_NUCHAR_INVALID_STREAKING;
            continue;
        }

        /* The current detector seems to be OK.  If either of the nearest
           neigbhors is flagged as inoperable/out-of-spec, don't
           characterize */
        if (statistics_invalid_flags[detector_index - 1]
            || statistics_invalid_flags[detector_index + 1])
        {
            IAS_LOG_DEBUG("Skipping detector %d -- nearest neighbor(s) "
                "%d and/or %d currently flagged inoperable / out-of-spec",
                (detector_index + 1), detector_index,
                (detector_index + 2));
            metrics_invalid_flags[detector_index] =
                RPS_NUCHAR_INVALID_STREAKING;
            continue;
        }

        /* The current detector and its immediate nearest neighbors seem
           to have good data.  Calculate the metric */
        neighbor_mean = 0.5 * (radiance_buffer[detector_index - 1]
                + radiance_buffer[detector_index + 1]);

        numerator = radiance_buffer[detector_index] - neighbor_mean;

        streaking_buffer[detector_index] = fabs(numerator)
            / radiance_buffer[detector_index];
    }

    /* Everything looks OK. */
    return SUCCESS;
}
