/**************************************************************************
 NAME:                      char_nonuniformity

 PURPOSE:       Calculates four metrics indicating uniformity of
                response between detectors on the focal plane.
                The metrics are derived from detector radiances
                estimated for an OLI solar diffuser (and/or TIRS blackbody)
                collect acquired at the nominal integration time.

 RETURNS:       Status code of SUCCESS or ERROR

 NOTES:         1.  This application will run following completion of
                data processing through the RPS Driver.  As currently
                designed, it cannot run from within the Driver because
                the statistics from ALL SCAs must be available to process
                a given band.

                2.  The detector radiances used to calculate the
                metrics can come from one of two sources.  The source
                is selectable in the work order.
                    a.  bias-corrected/linearized histogram statistics
                        directly converted to radiance through application
                        of CPF absolute and relative gains.  This is the
                        default source
                    b.  Radiance histogram statistics obtained on
                        a diffuser (blackbody) after the gain application
                        step in RPS Driver processing

***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <float.h>                    /* For DBL_EPSILON constant */

#include "ias_db.h"
#include "ias_cpf.h"
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_pixel_mask.h"
#include "ias_radiometry.h"
#include "ias_satellite_attributes.h"     /* IAS_MAX_BANDS const */


#include "char_nonuniformity.h"       /* Main application header file */


#define RPS_MAX_PROC_LEVEL_SIZE 6     /* 5 characters + NULL terminator */


/*----------------------------------------------------------------------
              *** MAIN APPLICATION char_nonuniformity ***
------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
    RPS_NUCHAR_METRICS metrics;                  /* Non-uniformity metrics
                                                    data structure with
                                                    detector-level results for
                                                    the current band/SCA */

    /* Database-related stuff */
    struct ias_db_connection *open_db = NULL;    /* IAS database connection
                                                    data structure */
    char rad_wo_common_id[IAS_CHAR_ID_SIZE];   /* Database ID number
                                                    corresponding to
                                                    specified work order */

    /* Optional summary report/dump file stuff */
    RPS_NUCHAR_REPORT_SUMMARY rpt_summary;
                                       /* Summary report information for a
                                          given band */

    /* Data / results invalid flags */
    RPS_NUCHAR_INVALID_FLAGS invalid_flags;

    /* Various processing-related stuff */
    char *collection_type = NULL;      /* Collection type string */
    char *sensor_id = NULL;            /* Sensor ID string */
    int is_oli_solar_collect;          /* Is current collect an OLI solar
                                          collect (0=no, 1=yes) */
    int is_tirs_blackbody_collect;     /* Is current collect a TIRS black-
                                          body collect (0=no, 1=yes) */
    int is_earth_collect;              /* Is current collect an Earth
                                          image collect (0=no, 1=yes) */
    int is_oli_present;                /* Did OLI image for this collect
                                          (0=no, 1=yes) */
    int is_tirs_present;               /* Did TIRS image for this collect
                                          (0=no, 1=yes) */
    int index = 0;                     /* 0-based index into band list */
    int status = 0;                    /* Function return status */
    int final_exit_status = EXIT_SUCCESS;

    IAS_SATELLITE_ID satellite_id;  /* Satellite ID used to initialize
                                       the satellite attributes
                                       library */
    /* Pixel mask stuff */
    IAS_PIXEL_MASK **pmasks = NULL;  /* Pixel mask array for all bands/
                                        SCAs processed */
    IAS_PIXEL_MASK *pm_ptr = NULL;   /* Pointer to an individual pixel
                                        mask */
    int num_masks_read = 0;          /* Number of pixel masks successfully
                                        read from file */
    int mask_index = 0;              /* 0-based index into pixel mask
                                        array */

    /* CPF stuff */
    IAS_CPF *cpf = NULL;               /* CPF data structure */

    /* Data structures containing required information from the CPF */
    const struct IAS_CPF_DETECTOR_STATUS *det_status        = NULL;
    const struct IAS_CPF_NONUNIFORMITY *nonuniform          = NULL;
    const struct IAS_CPF_ABSOLUTE_GAINS *abs_gains          = NULL;
    const struct IAS_CPF_RELATIVE_GAINS *relative_gains     = NULL;

    /* Processing parameters information */
    RPS_NUCHAR_WO_PARAMS wo_params;    /* Work order parameter data
                                          structure */
    RPS_NUCHAR_OMF_PARAMS omf_params;  /* OMF parameter data structure */


    /* Initialize logging for this application */
    ias_log_initialize("char_nonuniformity");

    /* Only one command line argument:  the WO parameter file */
    if (argc != 2)
    {
        IAS_LOG_ERROR("USAGE:  %s <name of parameter file>", argv[0]);
        return EXIT_FAILURE;
    }

    /* Read work order parameters */
    IAS_LOG_DEBUG("Reading required ODL and OMF parameters...");

    memset(&wo_params, 0, sizeof(RPS_NUCHAR_WO_PARAMS));
    memset(&omf_params, 0, sizeof(RPS_NUCHAR_OMF_PARAMS));

    status = read_parameters(argv[1], &wo_params, &omf_params);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Parameter processing failed");
        return EXIT_FAILURE;
    }

    /* Initialize the satellite attributes library so that we're processing
       OLI/TIRS data */
    IAS_LOG_DEBUG("Initializing satellite/sensor attributes...");

    satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(
            omf_params.satellite_number);
    if (satellite_id == ERROR)
    {
        IAS_LOG_ERROR("Invalid local satellite number in OMF parameter file");
        return EXIT_FAILURE;
    }

    status = ias_sat_attr_initialize(satellite_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite/sensor attributes");
        return EXIT_FAILURE;
    }

    /* Validate some of the important work order parameters */
    if (wo_params.calculate_radiance == RPS_NUCHAR_CALCULATE_RADIANCE)
    {
        /* Verify a proper relative gain option was selected in the
           work order.  The "invalid option" is 0 */
        if (wo_params.relgain_option == RPS_NUCHAR_INVALID_RELGAINS)
        {
            IAS_LOG_ERROR("Invalid relative gain option selected");
            return EXIT_FAILURE;
        }
    }

    /* Further initialize the application based on the selected output
       option(s).

       There are 7 possible combinations of "destinations" where
       characterization results can be written to.  The settings of
       the relevant WO parameters for each scenario are given below.

       - Dump files only (store_char_in_db=0, REPORT_FLAG=0, DUMP_FLAG=1)
       - Summary report only (store_char_in_db=0, REPORT_FLAG=1,
                              DUMP_FLAG=0)
       - Dump files + summary report (store_char_in_db=0, REPORT_FLAG=1,
                                      DUMP_FLAG=1)
       - DB only (store_char_in_db=1, REPORT_FLAG=0, DUMP_FLAG=0)
       - DB + dump files (store_char_in_db=1, REPORT_FLAG=0, DUMP_FLAG=1)
       - DB + summary report (store_char_in_db=1, REPORT_FLAG=1, DUMP_FLAG=0)
       - DB + summary report + dump files (store_char_in_db=1, REPORT_FLAG=1,
                                           DUMP_FLAG=1)

       Scenario 4 will be assumed the default scenario--everything goes
       just to the database, which can be queried off-line to generate
       "custom" reports.

       One other scenario could be selected.  That's where
       store_char_in_db=0, REPORT_FLAG=0, and DUMP_FLAG=0.  This
       combination will be treated as an error (obviously)--
       characterizations have been specified, but where are the
       results supposed to go?  */
    IAS_LOG_DEBUG("Validating output options...");
    if ((!wo_params.store_char_in_db) && (!wo_params.report_flag)
            && (!wo_params.dump_flag))
    {
        IAS_LOG_ERROR("No output destination is specified for "
                "characterization results");
        return EXIT_FAILURE;
    }

    /* Check to see if a pixel mask file exists.  If it doesn't, issue
       a warning, otherwise, try to read it */
    if (strcmp(omf_params.pmask_filename, "") == 0)
    {
        IAS_LOG_WARNING("Pixel mask file name not found, pixel masks will "
            "be individually created for each band/SCA");
    }
    else
    {
        pmasks = ias_pm_read_array_from_file(omf_params.pmask_filename,
             &num_masks_read);
        if (pmasks == NULL)
        {
            IAS_LOG_ERROR("Loading pixel mask array from pixel mask file %s",
                omf_params.pmask_filename); 
            return EXIT_FAILURE;
        }
    }

    /* Initialize the standard IAS header and RPS-specific header
       information for the summary report and/or dump file(s).
       DO NOT use argv[0] as the algorithm name -- it contains
       path information that shouldn't be in the report */
    IAS_LOG_DEBUG("Initializing report header information...");
    memset(&rpt_summary, 0, sizeof(RPS_NUCHAR_REPORT_SUMMARY));

    rpt_summary.header.common_header = ias_misc_initialize_report_header(
        omf_params.l0r_filename, wo_params.work_order_id,
        "Nonuniformity Char");
    if (rpt_summary.header.common_header == NULL)
    {
        IAS_LOG_ERROR("Initializing standard IAS report header");

        /***************************************************************
           NOTE    NOTE    NOTE    NOTE    NOTE    NOTE    NOTE    NOTE

           The following pixel mask cleanup code is repeated often
           enough in this module to more than justify implementing it as its
           own routine.  It would reduce the code size by
           8*(total number of occurrences) lines in this module alone.
           Not to mention maintenance would be much easier should changes
           need to be made to the core logic.  A similar situation exists
           within the char_rad_stability main driver module, which also
           uses an array of pixel masks read in from a file
        ****************************************************************/
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        return EXIT_FAILURE;
    }

    /* Update the CPF in the header to reflect what is in the OMF file. */
    ias_misc_set_header_cpf_name(rpt_summary.header.common_header,
        omf_params.cpf_filename);

    /* Initialize the RPS report header */
    status = ias_misc_initialize_rps_report_header(omf_params.l0r_filename,
        &(rpt_summary.header.rps_header));
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing RPS-specific report header "
            "information");
        ias_misc_free_report_header(rpt_summary.header.common_header);
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        return EXIT_FAILURE;
    }

    /* Make sure we have the right satellite/sensor */
    sensor_id =
        ias_misc_get_header_sensor_id(rpt_summary.header.common_header);
    if (sensor_id == NULL)
    {
        IAS_LOG_ERROR("Sensor ID information not initialized when IAS "
            "report header was created");
        ias_misc_free_report_header(rpt_summary.header.common_header);
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        return EXIT_FAILURE;
    }

    /* Confirm that the sensor ID references OLI and/or TIRS */
    is_oli_present = (strstr(sensor_id, "OLI") != NULL);
    is_tirs_present = (strstr(sensor_id, "TIRS") != NULL);
    if (!(is_oli_present || is_tirs_present))
    {
        IAS_LOG_ERROR("Invalid sensor ID %s detected", sensor_id);
        ias_misc_free_report_header(rpt_summary.header.common_header);
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        return EXIT_FAILURE;
    }

    /* Check the collection type.  If no information is available,
       quit now */
    collection_type =
        ias_misc_get_header_collection_type(rpt_summary.header.common_header);
    if (collection_type == NULL)
    {
        IAS_LOG_ERROR("No collection type information available");
        ias_misc_free_report_header(rpt_summary.header.common_header);
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        return EXIT_FAILURE;
    }

    /* Make sure we have the "right" collection types on which a
       nonuniformity characterization can be run:
           1.  "Normal" OLI solar calibration collects ("OLI SOLAR")
           2.  OLI solar integration sweep collects, but the
               integration times are at their nominal values
               ("OLI SOLAR INTEGRATION SWEEP")

           3.  "Normal" TIRS blackbody calibration collects
               ("TIRS BLACKBODY")
           4.  Earth imaging collects (in special cases).  If this is
               the collection type, turn off nonuniform source
               correction if it has been turned on in the work order */
    is_oli_solar_collect =
        ((strcmp(collection_type, IAS_RPS_OLI_SOLAR_COLLECT) == 0)
            || (strcmp(collection_type, IAS_RPS_OLI_SOLAR_TIMESWEEP_COLLECT)
                == 0));
    is_tirs_blackbody_collect =
        (strcmp(collection_type, IAS_RPS_TIRS_BLACKBODY_COLLECT) == 0);
    is_earth_collect =
        (strcmp(collection_type, IAS_RPS_EARTH_IMAGING_COLLECT) == 0);

    if (!(is_oli_solar_collect || is_tirs_blackbody_collect
        || is_earth_collect))
    {
        IAS_LOG_ERROR("Invalid collection type '%s' found:  expected "
            "collection types for nonuniformity characterization are "
            "'%s', '%s', '%s', or '%s'", collection_type,
            IAS_RPS_OLI_SOLAR_COLLECT,
            IAS_RPS_OLI_SOLAR_TIMESWEEP_COLLECT,
            IAS_RPS_TIRS_BLACKBODY_COLLECT,
            IAS_RPS_EARTH_IMAGING_COLLECT);
        ias_misc_free_report_header(rpt_summary.header.common_header);
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        return EXIT_FAILURE;
    }

    /* If this is an earth imaging collect, turn off nonuniformity
       source correction if it's been specified in the work order.
       There are currently no parameters defined representing source
       nonuniformity of the Earth */
    if (is_earth_collect)
    {
        if (wo_params.correct_flag)
        {
            IAS_LOG_WARNING("Cannot perform correction of nonuniform "
                "source response on Earth images--no scaling "
                "factors available.  Disabling source correction "
                "processing");
            wo_params.correct_flag = 0;
        }
    }

    /* For earth imaging collects, the OLI diffuser panel shouldn't be
       deployed (one would hope!).  It shouldn't be deployed for TIRS
       blackbody collects, either.  So, set the panel ID and position
       members of the report header data to indicate it's not deployed */
    if (!is_oli_solar_collect)
    {
        rpt_summary.header.panel_id = IAS_RPS_OLI_DIFFUSER_TYPE_NOTDEPLOYED;
        rpt_summary.header.diffuser_resolver_position = 0;
    }

    /* Initialize the metrics data structure for the first time.  If we
       have to bail before we get to the band loop, we aren't causing
       problems with non-NULL but invalid pointers */
    memset(&metrics, 0, sizeof(RPS_NUCHAR_METRICS));

    /* Read the CPF */
    IAS_LOG_DEBUG("Reading CPF parameters...");
    cpf = ias_cpf_read(omf_params.cpf_filename);
    if (cpf == NULL)
    {
        IAS_LOG_ERROR("Reading CPF file %s", omf_params.cpf_filename);
        ias_misc_free_report_header(rpt_summary.header.common_header);
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        return EXIT_FAILURE;
    }

    /* Additional set up is required for OLI solar collects--verification of
       proper integration times and identification of the diffuser panel
       deployed during the collect */
    if (is_oli_solar_collect)
    {
        IAS_DATETIME datetime;

        /* Get the OLI-specific CPF parameters and Earth constants.  The
           OLI parameters group contains integration time constants,
           among other things.  The Earth constants group information will
           be needed for identifying the diffuser panel.  */
        const struct IAS_CPF_OLI_PARAMETERS *oli_params =
            ias_cpf_get_oli_parameters(cpf);
        const struct IAS_CPF_EARTH_CONSTANTS *earth_constants =
            ias_cpf_get_earth_const(cpf);
        if ((oli_params == NULL) || (earth_constants == NULL))
        {
            IAS_LOG_ERROR("Obtaining sensor-specific parameters from CPF");
            ias_cpf_free(cpf);
            ias_misc_free_report_header(rpt_summary.header.common_header);
            if (pmasks != NULL)
            {
                for (mask_index = 0; mask_index < num_masks_read; mask_index++)
                {
                    pm_ptr = pmasks[mask_index];
                    ias_pm_destroy(pm_ptr);
                }
                free(pmasks);
            }
            return EXIT_FAILURE;
        }

        /* Check the integration times against the nominal values in the CPF.
           This is to make sure we're only dealing with "nominal" collects
           (or the "nominal" portion of the collect for a solar integration
           time sweep product).  The multiplication of the header
           integration time information by 0.001 converts those values
           from integer microseconds to (floating point) milliseconds,
           which is what's given in the CPF */
        if (((0.001 * rpt_summary.header.rps_header.ms_integration_time)
                != oli_params->nominal_integration_time_ms)
            || ((0.001 * rpt_summary.header.rps_header.pan_integration_time)
                   != oli_params->nominal_integration_time_pan))
        {
            IAS_LOG_ERROR("Nonuniformity characterization of any OLI solar "
                "collects should only be done for the nominal integration "
                "time subset");
            ias_misc_free_report_header(rpt_summary.header.common_header);
            if (pmasks != NULL)
            {
                for (mask_index = 0; mask_index < num_masks_read; mask_index++)
                {
                    pm_ptr = pmasks[mask_index];
                    ias_pm_destroy(pm_ptr);
                }
                free(pmasks);
            }
            ias_cpf_free(cpf);
            return EXIT_FAILURE;
        }

        /* Identify the diffuser panel deployed for the collect.  Some of
           the EARTH_CONSTANTS group information will be needed to
           obtain a timestamp that can be used to search the OLI telemetry
           for the diffuser_resolver_position information, which will
           be used to identify the panel.  */
        memset(&datetime, 0, sizeof(IAS_DATETIME));
        status = ias_misc_parse_datetime_string(
            rpt_summary.header.rps_header.acq_datetime.start_time_oli,
            IAS_DATETIME_L0R_FORMAT, &datetime);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Returning from ias_misc_parse_datetime_string");
            ias_misc_free_report_header(rpt_summary.header.common_header);
            if (pmasks != NULL)
            {
                for (mask_index = 0; mask_index < num_masks_read; mask_index++)
                {
                    pm_ptr = pmasks[mask_index];
                    ias_pm_destroy(pm_ptr);
                }
                free(pmasks);
            }
            ias_cpf_free(cpf);
            return EXIT_FAILURE;
        }

        status = ias_rps_identify_diffuser_panel(omf_params.l0r_filename,
            earth_constants, datetime,
            &rpt_summary.header.diffuser_resolver_position,
            &rpt_summary.header.panel_id);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Returning from ias_misc_identify_diffuser_panel");
            ias_misc_free_report_header(rpt_summary.header.common_header);
            if (pmasks != NULL)
            {
                for (mask_index = 0; mask_index < num_masks_read; mask_index++)
                {
                    pm_ptr = pmasks[mask_index];
                    ias_pm_destroy(pm_ptr);
                }
                free(pmasks);
            }
            ias_cpf_free(cpf);
            return EXIT_FAILURE;
        }
    }

    /* Finish initializing the summary data structure if we're
       generating a report and/or dump file(s) */
    if ((wo_params.report_flag) || (wo_params.dump_flag))
    {
        IAS_LOG_DEBUG("Preparing data for summary report and/or dump "
            "files...");

        /* Initialize the non-report header "static" members of the summary
           report data structure (those that aren't band-dependent) */
        initialize_nonuniformity_parameter_data(wo_params, &rpt_summary);

        /* Write the "common" header information file */
        IAS_LOG_DEBUG("Writing report header information to file...");
        status = write_report_header_to_file(rpt_summary,
            wo_params.work_order_id);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Writing report header information to file");
            ias_misc_free_report_header(rpt_summary.header.common_header);
            if (pmasks != NULL)
            {
                for (mask_index = 0; mask_index < num_masks_read; mask_index++)
                {
                    pm_ptr = pmasks[mask_index];
                    ias_pm_destroy(pm_ptr);
                }
                free(pmasks);
            }
            return EXIT_FAILURE;
        }
    }

    /* Get detector status information.  This will be used to build
       a band/SCA-specific pixel mask if pixel masks aren't being
       read in from a file.  In either case, out-of-spec detector
       information will need to be used as well */
    det_status = ias_cpf_get_det_status(cpf);
    if (det_status == NULL)
    {
        IAS_LOG_ERROR("Getting det_status group information from CPF");
        ias_cpf_free(cpf);
        ias_misc_free_report_header(rpt_summary.header.common_header);
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        return EXIT_FAILURE;
    }

    /* If we're calculating radiances from bias-corrected/linearized
       histogram statistics, get the CPF gains to be used  */
    if (wo_params.calculate_radiance == RPS_NUCHAR_CALCULATE_RADIANCE)
    {
        abs_gains = ias_cpf_get_abs_gains(cpf);
        switch (wo_params.relgain_option)
        {
            case RPS_NUCHAR_PRELAUNCH_RELGAINS:
                relative_gains = ias_cpf_get_prelaunch_relative_gains(cpf);
                break;
            case RPS_NUCHAR_POSTLAUNCH_RELGAINS:
                relative_gains = ias_cpf_get_postlaunch_relative_gains(cpf);
                break;
            case RPS_NUCHAR_CURRENT_RELGAINS:
                relative_gains = ias_cpf_get_relative_gains(cpf);
                break;
            default:
            {
                IAS_LOG_ERROR("Invalid relative gain option "
                              "specification %d",
                              wo_params.relgain_option);
                if (pmasks != NULL)
                {
                    for (mask_index = 0; mask_index < num_masks_read;
                         mask_index++)
                    {
                        pm_ptr = pmasks[mask_index];
                        ias_pm_destroy(pm_ptr);
                    }
                    free(pmasks);
                }
                ias_cpf_free(cpf);
                ias_misc_free_report_header(rpt_summary.header.common_header);
                return EXIT_FAILURE;
            }
        }
        if ((abs_gains == NULL) || (relative_gains == NULL))
        {
            IAS_LOG_ERROR("Retrieving CPF gain information");
            if (pmasks != NULL)
            {
                for (mask_index = 0; mask_index < num_masks_read; mask_index++)
                {
                    pm_ptr = pmasks[mask_index];
                    ias_pm_destroy(pm_ptr);
                }
                free(pmasks);
            }
            ias_cpf_free(cpf);
            ias_misc_free_report_header(rpt_summary.header.common_header);
            return EXIT_FAILURE;
        }
    }

    /* If we're doing a correction for nonuniform panel response,
       get the correction factors */
    if (wo_params.correct_flag)
    {
        nonuniform = ias_cpf_get_nonuniformity(cpf);
        if (nonuniform == NULL)
        {
            IAS_LOG_ERROR("Getting nonuniformity correction information "
                    " from CPF");
            ias_cpf_free(cpf);
            if (pmasks != NULL)
            {
                for (mask_index = 0; mask_index < num_masks_read; mask_index++)
                {
                    pm_ptr = pmasks[mask_index];
                    ias_pm_destroy(pm_ptr);
                }
                free(pmasks);
            }
            ias_misc_free_report_header(rpt_summary.header.common_header);
            return EXIT_FAILURE;
        }
    }

    /* DATABASE INITIALIZATION */

    /* Establish connections to the characterization database (for
       getting histogram statistics data, at minimum) */
    open_db = connect_to_db();
    if (open_db == NULL)
    {
        IAS_LOG_ERROR("Failed to connect to database, program terminating");
        ias_cpf_free(cpf);
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        ias_misc_free_report_header(rpt_summary.header.common_header);
        return EXIT_FAILURE;
    }

    /* Get the work order common ID number for the given work order ID.
       This will return a new ID if one doesn't exist...*/
    status = ias_misc_get_char_id(open_db, "RADIOMETRIC_WORK_ORDER_COMMON",
        "radiometric_wo_common_id", wo_params.work_order_id,
        rad_wo_common_id);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Obtaining valid radiometric work order common ID");
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        ias_cpf_free(cpf);
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        ias_misc_free_report_header(rpt_summary.header.common_header);
        return EXIT_FAILURE;
    }

    /* Prep for writing this round of characterization data to the database */
    if (wo_params.store_char_in_db)
    {
        IAS_LOG_DEBUG("Nonuniformity metrics will be written to the "
                "database");

        /* Initiate a transaction for this application.  This should
           also be tested to make sure it was successful */
        status = ias_db_start_transaction(open_db);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Initiating transaction");
            ias_db_close_connection(open_db);
            ias_db_close_database_lib();
            ias_cpf_free(cpf);
            ias_misc_free_report_header(rpt_summary.header.common_header);
            if (pmasks != NULL)
            {
                for (mask_index = 0; mask_index < num_masks_read; mask_index++)
                {
                    pm_ptr = pmasks[mask_index];
                    ias_pm_destroy(pm_ptr);
                }
                free(pmasks);
            }
            return EXIT_FAILURE;
        }

        /* Clean out previous characterization results from the tables */
        IAS_LOG_INFO("Clearing information previously saved for this "
                "work order...");
        status = clear_nonuniformity_table(open_db, rad_wo_common_id,
                "non_uniformity_band");
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Clearing NON_UNIFORMITY_BAND table");
            ias_db_rollback_transaction(open_db);
            ias_db_close_connection(open_db);
            ias_db_close_database_lib();
            ias_cpf_free(cpf);
            if (pmasks != NULL)
            {
                for (mask_index = 0; mask_index < num_masks_read; mask_index++)
                {
                    pm_ptr = pmasks[mask_index];
                    ias_pm_destroy(pm_ptr);
                }
                free(pmasks);
            }
            ias_misc_free_report_header(rpt_summary.header.common_header);
            return EXIT_FAILURE;
        }

        status = clear_nonuniformity_table(open_db, rad_wo_common_id,
                "non_uniformity_detector");
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Database update failed, program terminating");
            ias_db_rollback_transaction(open_db);
            ias_db_close_connection(open_db);
            ias_db_close_database_lib();
            ias_cpf_free(cpf);
            if (pmasks != NULL)
            {
                for (mask_index = 0; mask_index < num_masks_read; mask_index++)
                {
                    pm_ptr = pmasks[mask_index];
                    ias_pm_destroy(pm_ptr);
                }
                free(pmasks);
            }
            ias_misc_free_report_header(rpt_summary.header.common_header);
            return EXIT_FAILURE;
        }
    }

/*********************************************************************
                      BAND PROCESSING
*********************************************************************/

    /* NOTE:   if there's an error, all buffers will
       be freed.  This isn't an issue because free() does nothing
       with NULL pointers, so we shouldn't crash from an attempt 
       to do something to an invalid memory block.  When we
       finish processing the current band, however, the pointers
       should be freed and reset to NULL in case an error occurs before
       they get re-allocated */

/*-------------------------------------------------------------------
                    BEGIN BAND PROCESSING
---------------------------------------------------------------------*/
    IAS_LOG_INFO("Number of bands to process:  %d",
        wo_params.num_bands_to_process);
    for (index = 0; index < wo_params.num_bands_to_process; index++)
    {
        const char *band_rps_proc_level = NULL;
        const IAS_BAND_ATTRIBUTES *band_attributes = NULL;
        double *band_rms_buffer = NULL;
        double *band_stdev_buffer = NULL;
        double *band_streaking_buffer = NULL;
        double *radiance_buffer = NULL;
        double *stats_buffer = NULL;
        double radiance_mean = 0.0;    /* Mean radiance for the current
                                          band */
        double radiance_stdev = 0.0;   /* Std deviation of radiances for
                                          the current band */
        double band_ffov_metric = 0.0; /* Band-level FFOV metric */
        int band_index;
        int detector_index;
        int num_mask_pixels = 0;       /* Number of pixels in the mask
                                          for the current band/SCA */
        int num_band_detectors;
        int sca_index;

        /* Initialize the pointers in the invalid flag data structure
           for the current band */
        memset(&invalid_flags, 0, sizeof(RPS_NUCHAR_INVALID_FLAGS));

        /* Get the attributes for the current band.  The bands may or
           may not be listed in order--in case they aren't, get the
           "true" 0-based index for the band.

           If getting the attributes fails for any reason, we can't call
           cleanup_application because it relies on attribute information
           (the number of SCAs), so we'll have to call individual
           components to do the clean up.  We should be OK, however,
           because the band/SCA level data buffers haven't been
           (re)allocated yet... */
        band_attributes =
            ias_sat_attr_get_band_attributes(wo_params.bands[index]);
        if (band_attributes == NULL)
        {
            IAS_LOG_ERROR("Obtaining attribute information for band %d",
                wo_params.bands[index]);
            ias_db_rollback_transaction(open_db);
            ias_db_close_connection(open_db);
            ias_db_close_database_lib();
            ias_cpf_free(cpf);
            if (pmasks != NULL)
            {
                for (mask_index = 0; mask_index < num_masks_read; mask_index++)
                {
                    pm_ptr = pmasks[mask_index];
                    ias_pm_destroy(pm_ptr);
                }
                free(pmasks);
            }
            ias_misc_free_report_header(rpt_summary.header.common_header);
            return EXIT_FAILURE;
        }

        /* Get the "true" band index from the attributes information for
           the current band */
        band_index = band_attributes->band_index;

        /* Determine the total number of detectors for the band */
        num_band_detectors =
            band_attributes->scas * band_attributes->detectors_per_sca;

        /* Allocate the invalid flag data buffers */
        status = allocate_nonuniformity_invalid_flag_buffers(
            num_band_detectors, &invalid_flags);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Allocating invalid flag data buffers");
            ias_db_rollback_transaction(open_db);
            ias_db_close_connection(open_db);
            ias_db_close_database_lib();
            ias_cpf_free(cpf);
            if (pmasks != NULL)
            {
                for (mask_index = 0; mask_index < num_masks_read; mask_index++)
                {
                    pm_ptr = pmasks[mask_index];
                    ias_pm_destroy(pm_ptr);
                }
                free(pmasks);
            }
            ias_misc_free_report_header(rpt_summary.header.common_header);
            return EXIT_FAILURE;
        }

        /* NOTE:  All buffers will be allocated for the current band
           being processed.  This should reasonably deal with cases
           where the bands fundamentally change (such as changing
           from/to the PAN band to/from an MS band, or changing from
           an OLI band to a TIRS band, etc. */

        /* Allocate "local" metrics buffers. These will contain information
           for all detectors in the band--whether "valid" metrics results
           or flag values indicating characterization was not done/couldn't
           be done due to inoperability (and immediate proximity to an
           inoperable detector in the case of streaking)  */
        band_rms_buffer = calloc(num_band_detectors,
            sizeof(*band_rms_buffer));
        band_stdev_buffer = calloc(num_band_detectors,
            sizeof(*band_stdev_buffer));
        band_streaking_buffer = calloc(num_band_detectors,
            sizeof(*band_streaking_buffer));
        radiance_buffer = calloc(num_band_detectors,
            sizeof(*radiance_buffer));
        stats_buffer = calloc(band_attributes->detectors_per_sca,
            sizeof(*stats_buffer));

        if ((band_rms_buffer == NULL)
                || (band_stdev_buffer == NULL)
                || (band_streaking_buffer == NULL)
                || (radiance_buffer == NULL)
                || (stats_buffer == NULL))
        {
            IAS_LOG_ERROR("Allocating required local metrics "
                    "buffers for band %d",
                    band_attributes->band_number);
            cleanup_application(wo_params.store_char_in_db,
                band_attributes->scas, &rpt_summary, &metrics,
                &invalid_flags, open_db, band_rms_buffer,
                band_stdev_buffer, band_streaking_buffer,
                radiance_buffer, stats_buffer, cpf, pmasks, num_masks_read);
            return EXIT_FAILURE;
        }

        /* Set the RPS processing level for the current band based on the
           sensor identifier for the band */
        if (band_attributes->sensor_id == IAS_OLI)
        {
            band_rps_proc_level = wo_params.oli_rps_proc_level;
        }
        else
        {
            band_rps_proc_level = wo_params.tirs_rps_proc_level;
        }

        IAS_LOG_INFO("Processing band %d...", wo_params.bands[index]);

/*--------------------------------------------------------------------------
                           RADIANCE PROCESSING
----------------------------------------------------------------------------*/
        for (sca_index = 0; sca_index < band_attributes->scas; sca_index++)
        {
            int starting_detector_index =
                sca_index * band_attributes->detectors_per_sca;
            int sca_number = sca_index + 1;
            int mask_array_index;

            /* Prepare the pixel mask for the current band/SCA. */
            if (pmasks != NULL)
            {
                /* The mask we need should be in the array read in the
                   from the pixel mask file.  Determine the array index
                   for the correct mask, then set the local pointer to
                   that mask. */
                mask_array_index = ias_pm_get_mask_index(pmasks,
                    num_masks_read, band_attributes->band_number, sca_number);
                if (mask_array_index < 0)
                {
                    IAS_LOG_ERROR("Locating pixel mask for band %d SCA %d",
                        band_attributes->band_number, sca_number);
                    cleanup_application(wo_params.store_char_in_db,
                        band_attributes->scas, &rpt_summary, &metrics,
                        &invalid_flags, open_db, band_rms_buffer,
                        band_stdev_buffer, band_streaking_buffer,
                        radiance_buffer, stats_buffer, cpf, pmasks,
                        num_masks_read);
                    return EXIT_FAILURE;
                }
                pm_ptr = pmasks[mask_array_index];

                /* The dimensions of the source 'image' (number of
                   detectors and the number of pixels per detector)
                   are part of the internal mask information.  Get the
                   number of pixels per detector for the current
                   band/SCA. */
                num_mask_pixels = ias_pm_get_number_of_pixels(pm_ptr);
                if (num_mask_pixels == ERROR)
                {
                    IAS_LOG_ERROR("Obtaining number of pixels in band %d "
                        "SCA %d mask", band_attributes->band_number,
                        sca_number);
                    cleanup_application(wo_params.store_char_in_db,
                        band_attributes->scas, &rpt_summary, &metrics,
                        &invalid_flags, open_db, band_rms_buffer,
                        band_stdev_buffer, band_streaking_buffer,
                        radiance_buffer, stats_buffer, cpf, pmasks,
                        num_masks_read);
                    return EXIT_FAILURE;
                }
            }
            else
            {
                /* We don't have any pixel masks available.  However, we do
                   have available the number of detectors in the current
                   band/SCA (in 'band_attributes->detectors_per_sca').  For
                   the purposes of this characterization, the number of
                   pixels per detector can be any value >= 1, because the
                   inoperable and out-of-spec artifacts considered here
                   are detector-based (i.e. they affect all of a detector's
                   image pixels). */

                /* Set the number of pixels in the mask. */
                num_mask_pixels = 10;

                /* Create the empty mask. */
                pm_ptr = ias_pm_create(band_attributes->band_number,
                    sca_number, band_attributes->detectors_per_sca,
                    num_mask_pixels);
                if (pm_ptr == NULL)
                {
                    IAS_LOG_ERROR("Creating local pixel mask for band %d "
                        "SCA %d", band_attributes->band_number,
                        sca_number);
                    cleanup_application(wo_params.store_char_in_db,
                        band_attributes->scas, &rpt_summary, &metrics,
                        &invalid_flags, open_db, band_rms_buffer,
                        band_stdev_buffer, band_streaking_buffer,
                        radiance_buffer, stats_buffer, cpf, pmasks,
                        num_masks_read);
                    return EXIT_FAILURE;
                }

                /* Add the CPF inoperable detector information to the empty
                   mask. */
                status = ias_rps_inoperable_detector_identification(
                    det_status->inoperable[band_index][sca_index],
                    det_status->inoperable_count[band_index][sca_index],
                    band_attributes->band_number, sca_number,
                    band_attributes->detectors_per_sca, num_mask_pixels,
                    pm_ptr);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Adding inoperable detector "
                        "information to band %d SCA %d pixel mask",
                        band_attributes->band_number,
                        sca_number);
                    cleanup_application(wo_params.store_char_in_db,
                        band_attributes->scas, &rpt_summary, &metrics,
                        &invalid_flags, open_db, band_rms_buffer,
                        band_stdev_buffer, band_streaking_buffer,
                        radiance_buffer, stats_buffer, cpf, pmasks,
                        num_masks_read);
                    return EXIT_FAILURE;
                }
            }

            /* Add the CPF out-of-spec detector information to the mask. */
            status = ias_rps_outofspec_detector_identification(
                det_status->out_of_spec[band_index][sca_index],
                det_status->out_of_spec_count[band_index][sca_index],
                band_attributes->band_number, sca_number,
                band_attributes->detectors_per_sca, num_mask_pixels,
                pm_ptr);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Adding out-of-spec detector "
                    "information to band %d SCA %d pixel mask",
                    band_attributes->band_number, sca_number);
                cleanup_application(wo_params.store_char_in_db,
                    band_attributes->scas, &rpt_summary, &metrics,
                    &invalid_flags, open_db, band_rms_buffer,
                    band_stdev_buffer, band_streaking_buffer,
                    radiance_buffer, stats_buffer, cpf, pmasks,
                    num_masks_read);
                if (pmasks == NULL)
                {
                    ias_pm_destroy(pm_ptr);
                }
                return EXIT_FAILURE;
            }

            /* Populate the input data invalid flag for the current
               SCA */
            status = ias_rps_get_invalid_data_flags(pm_ptr,
                IAS_RPS_INOP_DETECTOR | IAS_RPS_OUTOFSPEC_DETECTOR,
                band_attributes->detectors_per_sca,
                &(invalid_flags.statistics[starting_detector_index]));
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Populating invalid flag information for "
                    "band number %d SCA number %d",
                    band_attributes->band_number, sca_number);
                cleanup_application(wo_params.store_char_in_db,
                    band_attributes->scas, &rpt_summary, &metrics,
                    &invalid_flags, open_db, band_rms_buffer,
                    band_stdev_buffer, band_streaking_buffer,
                    radiance_buffer, stats_buffer, cpf, pmasks,
                    num_masks_read);
                if (pmasks == NULL)
                {
                    ias_pm_destroy(pm_ptr);
                }
                return EXIT_FAILURE;
            }

            /* Grab the set of histogram statistics distinguished by the
               RPS processing level */
            status = get_histogram_statistics(open_db, rad_wo_common_id,
                band_attributes->spectral_type, band_attributes->band_number,
                sca_number, band_attributes->detectors_per_sca,
                band_rps_proc_level, stats_buffer, collection_type);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Obtaining histogram statistics for band %d "
                    "SCA %d", band_attributes->band_number,
                    sca_number);
                if (pmasks == NULL)
                {
                    ias_pm_destroy(pm_ptr);
                }
                cleanup_application(wo_params.store_char_in_db,
                    band_attributes->scas, &rpt_summary, &metrics,
                    &invalid_flags, open_db, band_rms_buffer,
                    band_stdev_buffer, band_streaking_buffer,
                    radiance_buffer, stats_buffer, cpf, pmasks,
                    num_masks_read);
                return EXIT_FAILURE;
            }

            /* Get the detector radiances for the current band/SCA, whether
               directly from the database or through direct calculation with
               bias-corrected/linearized histogram statistics.  If we're
               processing the PAN band, at least for now, compute the
               average of the two sets of histogram statistics and pass
               that in */
            status = determine_radiances(stats_buffer, band_index, sca_index,
                band_attributes->detectors_per_sca, abs_gains, relative_gains,
                nonuniform, wo_params.calculate_radiance,
                wo_params.correct_flag, rpt_summary.header.panel_id,
                invalid_flags.statistics, radiance_buffer);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Determining radiances for band %d SCA %d",
                    band_attributes->band_number, sca_number);
                if (pmasks == NULL)
                {
                    ias_pm_destroy(pm_ptr);
                }
                cleanup_application(wo_params.store_char_in_db,
                    band_attributes->scas, &rpt_summary, &metrics,
                    &invalid_flags, open_db, band_rms_buffer,
                    band_stdev_buffer, band_streaking_buffer,
                    radiance_buffer, stats_buffer, cpf, pmasks,
                    num_masks_read);
                return EXIT_FAILURE;
            }

            /* Free the individual pixel mask here if it was created */
            if (pmasks == NULL)
            {
                ias_pm_destroy(pm_ptr);
                pm_ptr = NULL;
            }

        }   /* END SCA-LEVEL RADIANCE PROCESSING */

/*----------------------------------------------------------------------------
                            METRICS CALCULATIONS
----------------------------------------------------------------------------*/
        /* Determine the band-level summary radiance statistics.  The NULL
           pointer for the last parameter in the function to calculate the
           mean indicates we don't need a return count of the valid samples
           that were used */
        status = ias_rps_calculate_filtered_mean(radiance_buffer,
            invalid_flags.statistics, num_band_detectors,
            &radiance_mean, NULL);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating mean radiance for band "
                    "%d\n", band_attributes->band_number);
            cleanup_application(wo_params.store_char_in_db,
                band_attributes->scas, &rpt_summary, &metrics,
                &invalid_flags, open_db, band_rms_buffer,
                band_stdev_buffer, band_streaking_buffer,
                radiance_buffer, stats_buffer, cpf, pmasks,
                num_masks_read);
            return EXIT_FAILURE;
        }
        status = ias_rps_calculate_filtered_stdev(radiance_buffer,
            invalid_flags.statistics, num_band_detectors,
            radiance_mean, &radiance_stdev);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Cannot calculate radiance stdev for band "
                "%d\n", band_attributes->band_number);
            cleanup_application(wo_params.store_char_in_db,
                band_attributes->scas, &rpt_summary, &metrics,
                &invalid_flags, open_db, band_rms_buffer,
                band_stdev_buffer, band_streaking_buffer,
                radiance_buffer, stats_buffer, cpf, pmasks,
                num_masks_read);
            return EXIT_FAILURE;
        }

        /* The radiance standard deviation is the basis for the FFOV 
           metric for the entire band.  Use this value (actually, a %
           of the mean radiance for the band) for all SCAs in
           the band. Do a final check to make sure radiance is not an
           "invalid" number (in this case, negative) */
        if (radiance_mean < 0)
        {
            IAS_LOG_ERROR("Invalid mean radiance calculated for band %d",
                band_attributes->band_number);
            cleanup_application(wo_params.store_char_in_db,
                band_attributes->scas, &rpt_summary, &metrics,
                &invalid_flags, open_db, band_rms_buffer,
                band_stdev_buffer, band_streaking_buffer,
                radiance_buffer, stats_buffer, cpf, pmasks,
                num_masks_read);
            return EXIT_FAILURE;
        }
        band_ffov_metric = (radiance_stdev / radiance_mean) * 100.0;

        /* Run the selected characterization(s) */
        if (wo_params.rms_banding_flag)
        {
            status = calculate_rms_banding_metrics(radiance_buffer,
                invalid_flags.statistics, num_band_detectors,
                wo_params.num_contiguous_detectors, radiance_mean,
                band_rms_buffer, invalid_flags.rms_banding);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Performing RMS banding characterization "
                        "for band %d", band_attributes->band_number);
                cleanup_application(wo_params.store_char_in_db,
                    band_attributes->scas, &rpt_summary, &metrics,
                    &invalid_flags, open_db, band_rms_buffer,
                    band_stdev_buffer, band_streaking_buffer,
                    radiance_buffer, stats_buffer, cpf, pmasks,
                    num_masks_read);
                return EXIT_FAILURE;
            }
        }

        if (wo_params.stdev_banding_flag)
        {
            status = calculate_stdev_banding_metrics(radiance_buffer,
                invalid_flags.statistics, num_band_detectors,
                wo_params.num_contiguous_detectors, band_stdev_buffer,
                invalid_flags.stdev_banding);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Performing stdev banding characterization "
                        "for band %d", band_attributes->band_number);
                cleanup_application(wo_params.store_char_in_db,
                    band_attributes->scas, &rpt_summary, &metrics,
                    &invalid_flags, open_db, band_rms_buffer,
                    band_stdev_buffer, band_streaking_buffer,
                    radiance_buffer, stats_buffer, cpf, pmasks,
                    num_masks_read);
                return EXIT_FAILURE;
            }
        }

        if (wo_params.streaking_flag)
        {
            status = calculate_streaking_metrics(radiance_buffer,
                invalid_flags.statistics, num_band_detectors,
                band_streaking_buffer, invalid_flags.streaking);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Performing streaking characterization for "
                        "band %d", band_attributes->band_number);
                cleanup_application(wo_params.store_char_in_db,
                    band_attributes->scas, &rpt_summary, &metrics,
                    &invalid_flags, open_db, band_rms_buffer,
                    band_stdev_buffer, band_streaking_buffer,
                    radiance_buffer, stats_buffer, cpf, pmasks,
                    num_masks_read);
                return EXIT_FAILURE;
            }
        }


/*----------------------------------------------------------------------
                      POPULATE METRICS DATA STRUCTURES
------------------------------------------------------------------------*/
        /* Clear out the previous band's data */
        memset(&metrics, 0, sizeof(RPS_NUCHAR_METRICS));

        /* Start populating the metrics data */
        metrics.band_number         = band_attributes->band_number;
        metrics.ffov_metric         = band_ffov_metric;
        metrics.band_radiance_mean  = radiance_mean;
        metrics.band_radiance_stdev = radiance_stdev;

        /* Allocate the metrics data buffers, depending on the
           selected characterization(s) */
        status = allocate_metrics_buffers(wo_params.rms_banding_flag,
            wo_params.stdev_banding_flag,
            wo_params.streaking_flag,
            band_attributes->scas,
            band_attributes->detectors_per_sca,
            &metrics);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Allocating required metrics data buffers");
            cleanup_application(wo_params.store_char_in_db,
                band_attributes->scas, &rpt_summary, &metrics,
                &invalid_flags, open_db, band_rms_buffer,
                band_stdev_buffer, band_streaking_buffer,
                radiance_buffer, stats_buffer, cpf, pmasks,
                num_masks_read);
            return EXIT_FAILURE;
        }

        /* Populate the metrics data buffers for all SCAs in the
           current band */
        for (sca_index = 0; sca_index < band_attributes->scas; sca_index++)
        {
            populate_metrics_buffers(band_rms_buffer,
                band_stdev_buffer,
                band_streaking_buffer,
                wo_params.rms_banding_flag,
                wo_params.stdev_banding_flag,
                wo_params.streaking_flag,
                (sca_index + 1),
                band_attributes->detectors_per_sca,
                &metrics);
        }

/*------------------------------------------------------------------------
                           WRITE METRICS TO DB
--------------------------------------------------------------------------*/
        if (wo_params.store_char_in_db)
        {
            IAS_DB_NULL_TYPE *flagged_rms_detectors = NULL;
            IAS_DB_NULL_TYPE *flagged_stdev_detectors = NULL;
            IAS_DB_NULL_TYPE *flagged_streaking_detectors = NULL;
            int *detector_ids = NULL;
            double *pcnt_rms_metrics = NULL;
            double *pcnt_stdev_metrics = NULL;
            double *db_streaking_metrics = NULL;

            /* Determine the expected size of the SCA-specific flagged
               detector ID and percent metrics data buffers.  The
               explicit type IAS_DB_NULL_TYPE can either be a 64-bit
               unsigned long or a "regular" integer, depending on whether
               64-bit or 32-bit code is built.  To ensure always getting the
               correct size, use the explicit type */
            size_t flagged_detector_buffer_size =
                band_attributes->detectors_per_sca * sizeof(IAS_DB_NULL_TYPE);
            size_t pcnt_detector_buffer_size =
                band_attributes->detectors_per_sca * sizeof(double);

            /* Write the FFOV metric information for the current band */
            IAS_LOG_DEBUG("Writing band %d metric statistics to "
                "NON_UNIFORMITY_BAND table...",
                band_attributes->band_number);
            status = write_to_band_table(open_db, rad_wo_common_id,
                    band_attributes->band_number, metrics.ffov_metric);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Writing FFOV metric information to DB");
                cleanup_application(wo_params.store_char_in_db,
                    band_attributes->scas, &rpt_summary, &metrics,
                    &invalid_flags, open_db, band_rms_buffer,
                    band_stdev_buffer, band_streaking_buffer,
                    radiance_buffer, stats_buffer, cpf, pmasks,
                    num_masks_read);
                return EXIT_FAILURE;
            }

            /* Allocate detector ID data buffers (sized to the number of
               detectors in an SCA) that will contain the database NULL
               value (-1) flagging those detectors that weren't/couldn't
               be characterized, as well as a full detector ID number
               data buffer.  In addition, allocate similarly sized
               data buffers to contain the "raw" metrics converted to
               "%-of-mean-radiance" values*/
            flagged_rms_detectors = calloc(
                band_attributes->detectors_per_sca,
                sizeof(*flagged_rms_detectors));
            flagged_stdev_detectors = calloc(
                band_attributes->detectors_per_sca,
                sizeof(*flagged_stdev_detectors));
            flagged_streaking_detectors = calloc(
                band_attributes->detectors_per_sca,
                sizeof(*flagged_streaking_detectors));
            detector_ids = calloc(band_attributes->detectors_per_sca,
                sizeof(*detector_ids));

            pcnt_rms_metrics = calloc(band_attributes->detectors_per_sca,
                sizeof(*pcnt_rms_metrics));
            pcnt_stdev_metrics = calloc(band_attributes->detectors_per_sca,
                sizeof(*pcnt_stdev_metrics));
            db_streaking_metrics = calloc(
                band_attributes->detectors_per_sca,
                sizeof(*db_streaking_metrics));

            if ((flagged_rms_detectors == NULL)
                || (flagged_stdev_detectors == NULL)
                || (flagged_streaking_detectors == NULL)
                || (detector_ids == NULL)
                || (pcnt_rms_metrics == NULL)
                || (pcnt_stdev_metrics == NULL)
                || (db_streaking_metrics == NULL))
            {
                IAS_LOG_ERROR("Allocating NULL detector flag data buffers "
                    "for DB write in band %d",
                    band_attributes->band_number);

                free(flagged_rms_detectors);
                free(flagged_stdev_detectors);
                free(flagged_streaking_detectors);
                free(detector_ids);
                free(pcnt_rms_metrics);
                free(pcnt_stdev_metrics);
                free(db_streaking_metrics);

                cleanup_application(wo_params.store_char_in_db,
                    band_attributes->scas, &rpt_summary, &metrics,
                    &invalid_flags, open_db, band_rms_buffer,
                    band_stdev_buffer, band_streaking_buffer,
                    radiance_buffer, stats_buffer, cpf, pmasks,
                    num_masks_read);
                return EXIT_FAILURE;
            }

            /* Populate the full detector ID buffer */
            for (detector_index = 0;
                 detector_index < band_attributes->detectors_per_sca;
                 detector_index++)
            {
                detector_ids[detector_index] = (detector_index + 1);
            }

            /* Write each SCA's detector metrics information to the DB */
            for (sca_index = 0; sca_index < band_attributes->scas;
                 sca_index++)
            {
                int sca_number = sca_index + 1;
                int starting_detector_index =
                    sca_index * band_attributes->detectors_per_sca;

                /* Clear out the previous SCA's information for the
                   "NULL" detector flags*/
                memset(flagged_rms_detectors, 0,
                    flagged_detector_buffer_size);
                memset(flagged_stdev_detectors, 0,
                    flagged_detector_buffer_size);
                memset(flagged_streaking_detectors, 0,
                    flagged_detector_buffer_size);

                memset(pcnt_rms_metrics, 0, pcnt_detector_buffer_size);
                memset(pcnt_stdev_metrics, 0, pcnt_detector_buffer_size);
                memset(db_streaking_metrics, 0,
                    pcnt_detector_buffer_size);

                /* Generate the flagged detector ID and % metrics
                   data buffers for the current SCA so they can be written
                   to the database */
                populate_database_detector_write_buffers(
                    &(metrics.rms_metrics[sca_index]),
                    &(metrics.stdev_metrics[sca_index]),
                    &(metrics.streaking_metrics[sca_index]),
                    &(invalid_flags.rms_banding[starting_detector_index]),
                    &(invalid_flags.stdev_banding[starting_detector_index]),
                    &(invalid_flags.streaking[starting_detector_index]),
                    metrics.band_radiance_mean,
                    wo_params.rms_banding_flag,
                    wo_params.stdev_banding_flag,
                    wo_params.streaking_flag,
                    band_attributes->detectors_per_sca,
                    flagged_rms_detectors,
                    pcnt_rms_metrics,
                    flagged_stdev_detectors,
                    pcnt_stdev_metrics,
                    flagged_streaking_detectors,
                    db_streaking_metrics);

                /* Write all detector metrics results (with NULLs) to the
                   database */
                IAS_LOG_DEBUG("Writing band %d SCA %d detector metric "
                    "statistics to NON_UNIFORMITY_DETECTOR table...",
                    band_attributes->band_number, sca_number);
                status = write_to_detector_table(open_db,
                    rad_wo_common_id,
                    band_attributes->band_number,
                    sca_number,
                    band_attributes->detectors_per_sca,
                    wo_params.rms_banding_flag,
                    wo_params.stdev_banding_flag,
                    wo_params.streaking_flag,
                    detector_ids,
                    flagged_rms_detectors,
                    pcnt_rms_metrics,
                    flagged_stdev_detectors,
                    pcnt_stdev_metrics,
                    flagged_streaking_detectors,
                    db_streaking_metrics);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Writing non-uniformity characterization "
                        "results to NON_UNIFORMITY_DETECTOR table "
                        "for band %d SCA %d",
                        band_attributes->band_number,
                        sca_number);
                    free(flagged_rms_detectors);
                    free(flagged_stdev_detectors);
                    free(flagged_streaking_detectors);
                    free(pcnt_rms_metrics);
                    free(pcnt_stdev_metrics);
                    free(db_streaking_metrics);
                    free(detector_ids);
                    cleanup_application(wo_params.store_char_in_db,
                        band_attributes->scas, &rpt_summary, &metrics,
                        &invalid_flags, open_db, band_rms_buffer,
                        band_stdev_buffer, band_streaking_buffer,
                        radiance_buffer, stats_buffer, cpf, pmasks,
                        num_masks_read);
                    return EXIT_FAILURE;
                }
            }

            /* Free the flagged detector index buffers and percent metrics
               data buffers */
            free(flagged_rms_detectors);
            flagged_rms_detectors = NULL;
            free(flagged_stdev_detectors);
            flagged_stdev_detectors = NULL;
            free(flagged_streaking_detectors);
            flagged_streaking_detectors = NULL;
            free(pcnt_rms_metrics);
            pcnt_rms_metrics = NULL;
            free(pcnt_stdev_metrics);
            pcnt_stdev_metrics = NULL;
            free(db_streaking_metrics);
            db_streaking_metrics = NULL;
            free(detector_ids);
            detector_ids = NULL;

        }   /* END DB WRITE BLOCK */


/*------------------------------------------------------------------------
                           SUMMARIZE METRICS DATA
--------------------------------------------------------------------------*/
        if ((wo_params.report_flag) || (wo_params.dump_flag))
        {
            /* Allocate buffers in the report summary data structure */
            status =
                allocate_summary_report_buffers(wo_params.rms_banding_flag,
                     wo_params.stdev_banding_flag,
                     wo_params.streaking_flag,
                band_attributes->scas,
                &rpt_summary);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Allocating report summary data "
                    "buffers for band %d",
                    band_attributes->band_number);
                cleanup_application(wo_params.store_char_in_db,
                    band_attributes->scas, &rpt_summary, &metrics,
                    &invalid_flags, open_db, band_rms_buffer,
                    band_stdev_buffer, band_streaking_buffer,
                    radiance_buffer, stats_buffer, cpf, pmasks,
                    num_masks_read);
                return EXIT_FAILURE;
            }

            /* Write the band-specific data to the summary report
               data structure */
            initialize_summary_report_band_data(wo_params, band_attributes,
                &rpt_summary);

            /* Copy the band metrics information to the summary report
               data */
            rpt_summary.ffov_metric = metrics.ffov_metric;

            /* Calculate the required summary statistics and populate
               the summary report data structure, then copy that info
               to the summary report data structure */
            status = calculate_per_detector_metrics_std(band_rms_buffer,
                band_stdev_buffer, band_streaking_buffer, &invalid_flags,
                num_band_detectors, wo_params.rms_banding_flag,
                wo_params.stdev_banding_flag, wo_params.streaking_flag,
                &rpt_summary);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Calculating per-detector std deviations of "
                    "selected metrics in band %d\n",
                    band_attributes->band_number);
                cleanup_application(wo_params.store_char_in_db,
                    band_attributes->scas, &rpt_summary, &metrics,
                    &invalid_flags, open_db, band_rms_buffer,
                    band_stdev_buffer, band_streaking_buffer,
                    radiance_buffer, stats_buffer, cpf, pmasks,
                    num_masks_read);
                return EXIT_FAILURE;
            }

            /* Determine within- and cross-SCA summary metrics information
               and copy it to the summary report data structure */
            status = initialize_sca_summary_metrics(metrics,
                &invalid_flags,
                wo_params.rms_banding_flag,
                wo_params.stdev_banding_flag,
                wo_params.streaking_flag,
                band_attributes->scas,
                band_attributes->detectors_per_sca,
                &rpt_summary);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Summarizing metrics statistics for band %d",
                    band_attributes->band_number);
                cleanup_application(wo_params.store_char_in_db,
                    band_attributes->scas, &rpt_summary, &metrics,
                    &invalid_flags, open_db, band_rms_buffer,
                    band_stdev_buffer, band_streaking_buffer,
                    radiance_buffer, stats_buffer, cpf, pmasks,
                    num_masks_read);
                return EXIT_FAILURE;
            }
        }   /* END SUMMARIZE METRICS BLOCK */


/*-------------------------------------------------------------------------
                           WRITE SUMMARY REPORTS
---------------------------------------------------------------------------*/
        if (wo_params.report_flag)
        {
            char reportfile_name[PATH_MAX]; /* Name of summary report file */
            FILE *rpt_fptr = NULL;          /* Report file handle */

            /* Build the full path and name of the summary report file
               for the current band */
            status = snprintf(reportfile_name, sizeof(reportfile_name),
                "%s_summary_B%02d.txt", wo_params.report_filename_prefix,
                band_attributes->band_number);
            if ((status < 0) || (status >= sizeof(reportfile_name)))
            {
                IAS_LOG_ERROR("Creating band %d summary report name",
                    band_attributes->band_number);
                cleanup_application(wo_params.store_char_in_db,
                    band_attributes->scas, &rpt_summary, &metrics,
                    &invalid_flags, open_db, band_rms_buffer,
                    band_stdev_buffer, band_streaking_buffer,
                    radiance_buffer, stats_buffer, cpf, pmasks,
                    num_masks_read);
                return EXIT_FAILURE;
            }

            /* Create the report file for the current band */
            rpt_fptr = fopen(reportfile_name, "w");
            if (rpt_fptr == NULL)
            {
                IAS_LOG_ERROR("Creating summary report file %s",
                    reportfile_name);
                cleanup_application(wo_params.store_char_in_db,
                    band_attributes->scas, &rpt_summary, &metrics,
                    &invalid_flags, open_db, band_rms_buffer,
                    band_stdev_buffer, band_streaking_buffer,
                    radiance_buffer, stats_buffer, cpf, pmasks,
                    num_masks_read);
                return EXIT_FAILURE;
            }

            /* Write the summary report "body" */
            status = write_summary_report_data(rpt_fptr, rpt_summary,
                metrics.band_radiance_mean, metrics.band_radiance_stdev,
                metrics.ffov_metric);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Writing summary report data for "
                    "band %d to report file %s",
                    band_attributes->band_number, reportfile_name);
                fclose(rpt_fptr);
                cleanup_application(wo_params.store_char_in_db,
                    band_attributes->scas, &rpt_summary, &metrics,
                    &invalid_flags, open_db, band_rms_buffer,
                    band_stdev_buffer, band_streaking_buffer,
                    radiance_buffer, stats_buffer, cpf, pmasks,
                    num_masks_read);
                return EXIT_FAILURE;
            }

            /* Close the file, but check to make sure the close succeeded */
            if (fclose(rpt_fptr))
            {
                IAS_LOG_ERROR("Closing summary report file %s",
                    reportfile_name);
                cleanup_application(wo_params.store_char_in_db,
                    band_attributes->scas, &rpt_summary, &metrics,
                    &invalid_flags, open_db, band_rms_buffer,
                    band_stdev_buffer, band_streaking_buffer,
                    radiance_buffer, stats_buffer, cpf, pmasks,
                    num_masks_read);
                return EXIT_FAILURE;
            }
            rpt_fptr = NULL;
        }       /* END WRITE SUMMARY REPORT BLOCK */


/*------------------------------------------------------------------------
                                  WRITE DUMP FILES
--------------------------------------------------------------------------*/
        if (wo_params.dump_flag)
        {
            char dumpfile_name[PATH_MAX];      /* Name of dump file */
            FILE *dump_fptr = NULL;            /* Dump file handle*/

            /* Build the full path and name of the dump file.  Due to the
               amount of data, each band will get its own dump file */
            status = snprintf(dumpfile_name, sizeof(dumpfile_name),
                "%s_dumpfile_B%02d.txt", wo_params.report_filename_prefix,
                band_attributes->band_number);
            if ((status < 0) || (status >= sizeof(dumpfile_name)))
            {
                IAS_LOG_ERROR("Creating band %d dump file name",
                    band_attributes->band_number);
                cleanup_application(wo_params.store_char_in_db,
                    band_attributes->scas, &rpt_summary, &metrics,
                    &invalid_flags, open_db, band_rms_buffer,
                    band_stdev_buffer, band_streaking_buffer,
                    radiance_buffer, stats_buffer, cpf, pmasks,
                    num_masks_read);
                return EXIT_FAILURE;
            }

            dump_fptr = fopen(dumpfile_name, "w");
            if (dump_fptr == NULL)
            {
                IAS_LOG_ERROR("Creating dump file %s", dumpfile_name);
                fclose(dump_fptr);
                cleanup_application(wo_params.store_char_in_db,
                    band_attributes->scas, &rpt_summary, &metrics,
                    &invalid_flags, open_db, band_rms_buffer,
                    band_stdev_buffer, band_streaking_buffer,
                    radiance_buffer, stats_buffer, cpf, pmasks,
                    num_masks_read);
                return EXIT_FAILURE;
            }

            /* Set the band number in the header */
            ias_misc_set_header_band_number(rpt_summary.header.common_header,
                band_attributes->band_number);

            /* Write the rest of the data to the dump file */
            status = write_metrics_data_to_dump_file(metrics, rpt_summary,
                &invalid_flags, dump_fptr);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Writing data to dump file %s",
                    dumpfile_name);
                fclose(dump_fptr);
                cleanup_application(wo_params.store_char_in_db,
                    band_attributes->scas, &rpt_summary, &metrics,
                    &invalid_flags, open_db, band_rms_buffer,
                    band_stdev_buffer, band_streaking_buffer,
                    radiance_buffer, stats_buffer, cpf, pmasks,
                    num_masks_read);
                return EXIT_FAILURE;
            }

            /* Close the file, and check to make sure it closed
               successfully */
            if (fclose(dump_fptr))
            {
                IAS_LOG_ERROR("Closing dump file %s", dumpfile_name);
                cleanup_application(wo_params.store_char_in_db,
                    band_attributes->scas, &rpt_summary, &metrics,
                    &invalid_flags, open_db, band_rms_buffer,
                    band_stdev_buffer, band_streaking_buffer,
                    radiance_buffer, stats_buffer, cpf, pmasks,
                    num_masks_read);
                return EXIT_FAILURE;
            }
            dump_fptr = NULL;
        }          /* END WRITE DUMP FILE BLOCK */

        /* Free up resources allocated for this band */
        cleanup_for_next_band(band_attributes->scas, &rpt_summary, &metrics,
            &invalid_flags, band_rms_buffer, band_stdev_buffer,
            band_streaking_buffer, radiance_buffer, stats_buffer);

        /* Set/reset the local data buffer pointers to NULL */
        band_rms_buffer                   = NULL;
        band_stdev_buffer                 = NULL;
        band_streaking_buffer             = NULL;
        radiance_buffer                   = NULL;
        stats_buffer                      = NULL;
        metrics.rms_metrics               = NULL;
        metrics.stdev_metrics             = NULL;
        metrics.streaking_metrics         = NULL;
        rpt_summary.cross_sca_rms         = NULL;
        rpt_summary.cross_sca_stdev       = NULL;
        rpt_summary.sca_summary_rms       = NULL;
        rpt_summary.sca_summary_stdev     = NULL;
        rpt_summary.sca_summary_streaking = NULL;

    }              /* END MAIN BAND PROCESSING LOOP */


/*-------------------------------------------------------------------------
                              CLEAN-UP
--------------------------------------------------------------------------*/
    /* If we're writing characterization results to the database, commit the
       full transaction and close it out*/
    IAS_LOG_DEBUG("Cleaning up application resources...");
    if (wo_params.store_char_in_db)
    {
        status = ias_db_commit_transaction(open_db);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Completing transaction processing");
            final_exit_status = EXIT_FAILURE;
        }
    }

    /* If we're generating report and/or dump files, free up the header */
    ias_misc_free_report_header(rpt_summary.header.common_header);

    /* Close the database connection */
    ias_db_close_connection(open_db);
    ias_db_close_database_lib();

    /* Free up the CPF resources */
    ias_cpf_free(cpf);

    /* Free up the pixel mask resources */
    if (pmasks != NULL)
    {
        for (mask_index = 0; mask_index < num_masks_read; mask_index++)
        {
            pm_ptr = pmasks[mask_index];
            ias_pm_destroy(pm_ptr);
        }
        free(pmasks);
    }

    /* Log a successful completion of the application and
       then quit */
    if (final_exit_status == EXIT_SUCCESS)
    {
        IAS_LOG_INFO("Nonuniformity characterization completed successfully.");
    }
    else
    {
        IAS_LOG_INFO("Nonuniformity characterization failed.");
    }

    return final_exit_status;
}
