#ifndef CHAR_NONUNIFORMITY_H
#define CHAR_NONUNIFORMITY_H

#include <limits.h>

#include "ias_const.h"
#include "ias_cpf.h"
#include "ias_db.h"
#include "ias_miscellaneous.h"
#include "ias_pixel_mask.h"
#include "ias_radiometry.h"
#include "ias_satellite_attributes.h"


/*-------------------------------------------------------------------------
                         USEFUL CONSTANTS
---------------------------------------------------------------------------*/
#define RPS_NUCHAR_VALID_CHARS              0x00
#define RPS_NUCHAR_INVALID_RMS_BANDING      0x08
#define RPS_NUCHAR_INVALID_STDEV_BANDING    0x10
#define RPS_NUCHAR_INVALID_STREAKING        0x20


/* All nonuniformity threshold limits are based on information contained in
   version 3.1 of the nonuniformity ADD.  */
#define RPS_NUCHAR_NONTHERMAL_PAN_FFOV_LIMIT           0.5
#define RPS_NUCHAR_NONTHERMAL_PAN_RMS_LIMIT            1.0
#define RPS_NUCHAR_NONTHERMAL_PAN_STDEV_LIMIT          0.25
#define RPS_NUCHAR_NONTHERMAL_PAN_STREAKING_LIMIT      0.01

#define RPS_NUCHAR_NONTHERMAL_MS_FFOV_LIMIT            0.5
#define RPS_NUCHAR_NONTHERMAL_MS_RMS_LIMIT             1.0
#define RPS_NUCHAR_NONTHERMAL_MS_STDEV_LIMIT           0.25
#define RPS_NUCHAR_NONTHERMAL_MS_STREAKING_LIMIT       0.005

#define RPS_NUCHAR_THERMAL_MS_FFOV_LIMIT               0.5
#define RPS_NUCHAR_THERMAL_MS_RMS_LIMIT                0.5
#define RPS_NUCHAR_THERMAL_MS_STDEV_LIMIT              0.5
#define RPS_NUCHAR_THERMAL_MS_STREAKING_LIMIT          0.005


/* Work order option default values */
#define RPS_NUCHAR_NUM_CONTIGUOUS_DETECTORS     100
#define RPS_NUCHAR_NO_CORRECT                     0
#define RPS_NUCHAR_CORRECT                        1
#define RPS_NUCHAR_USE_DB_RADIANCE                0
#define RPS_NUCHAR_CALCULATE_RADIANCE             1
#define RPS_NUCHAR_NO_RMS_BANDING                 0
#define RPS_NUCHAR_RMS_BANDING                    1
#define RPS_NUCHAR_NO_STDEV_BANDING               0
#define RPS_NUCHAR_STDEV_BANDING                  1
#define RPS_NUCHAR_NO_STREAKING                   0
#define RPS_NUCHAR_STREAKING                      1
#define RPS_NUCHAR_NO_TREND                       0
#define RPS_NUCHAR_TREND                          1
#define RPS_NUCHAR_NO_REPORT                      0
#define RPS_NUCHAR_REPORT                         1
#define RPS_NUCHAR_NO_DUMP                        0
#define RPS_NUCHAR_DUMP                           1


/* Miscellaneous #defined constants */
#define RPS_NUCHAR_NUM_SCA_BOUNDARY_DETECTORS    100


/* Default RPS processing levels for the OLI solar diffuser and TIRS
   histogram statistics input data.  The currently recommended radiometric
   processing required for the OLI histogram statistics used in this
   characterization consists of temperature correction of bias, followed by
   bias correction and response linearization (in that order).  The 'C'
   represents an adjustment in the "raw" DN values by taking the upper 12
   bits of the original 14-bit L0R data and correcting for quantization error.
   For the TIRS histogram statistics, the currently recommended radiometric
   processing consists of response linearization followed by bias
   correction.  */
#define RPS_NUCHAR_OLI_PROC_LEVEL_TEMPCORRECT        "CTBL"
#define RPS_NUCHAR_TIRS_PROC_LEVEL                   "0LB2O"

/*------------------------------------------------------------------------
            USEFUL ENUMERATED TYPES AND DATA STRUCTURES
--------------------------------------------------------------------------*/
    /* Options for the set of CPF-based relative gains to use in
       internal radiance calculations.  These apply to both OLI and
       TIRS bands if both are processed in the same session */
typedef enum rps_nuchar_relgain_options
{
    RPS_NUCHAR_INVALID_RELGAINS = 0,
    RPS_NUCHAR_PRELAUNCH_RELGAINS,
    RPS_NUCHAR_POSTLAUNCH_RELGAINS,
    RPS_NUCHAR_CURRENT_RELGAINS
} RPS_NUCHAR_RELGAIN_OPTIONS;




/*    Base non-uniformity metrics data structures */
typedef struct rps_nuchar_metrics_det
{
    int *detector_number;              /* Detector IDs */
    double *metric;                    /* Detector metric values */
    int sca_number;                    /* Current 1-based SCA number */
} METRICS_DET;

    /* Non-uniformity metrics data structure.  Used more as a
       "central gathering point" for the information in each band.  */
typedef struct rps_nuchar_metrics
{
    int band_number;                   /* 1-based band number */
    double band_radiance_mean;         /* Mean radiance for
                                          the current band */
    double band_radiance_stdev;        /* Std deviation of detector
                                          radiances for the current band */
    double ffov_metric;                /* FFOV metric for the band */
    double rms_band_mean;              /* Mean RMS metric for the band */
    double rms_band_stdev;             /* Std deviation of RMS metric for
                                          the band */
    double stdev_band_mean;            /* Mean "stdev" metric for the
                                          band */
    double stdev_band_stdev;           /* Std deviation of "stdev" metric
                                          for the band */
    double streaking_mean;             /* Mean streaking metric for the
                                          band */
    double streaking_stdev;            /* Std deviation of streaking metric
                                          for the band */
    METRICS_DET *rms_metrics;          /* Valid RMS banding metrics for all
                                          SCAs in the current band */
    METRICS_DET *stdev_metrics;        /* Valid stdev banding metrics for
                                          all SCAs in the current band*/
    METRICS_DET *streaking_metrics;    /* Valid streaking metrics for all
                                          SCAs in the current band */ 
} RPS_NUCHAR_METRICS;

/* Characterization invalid flags data structure */
typedef struct rps_nuchar_invalid_flags
{
    int *statistics;               /* Histogram statistics invalid flags */
    int *rms_banding;              /* RMS banding metric invalid flags */
    int *stdev_banding;            /* STDEV banding metric invalid flags */
    int *streaking;                /* STREAKING metric invalid flags */
}   RPS_NUCHAR_INVALID_FLAGS;

/*    Base data structures containing information written to summary report
      and/or dump files */
typedef struct cross_sca_summary
{
    int sca1_number;                /* SCA on "left" side of boundary */
    int sca2_number;                /* SCA on "right" side of boundary */
    double cross_mean;              /* Mean of detector metrics across
                                       SCA boundary */
    double cross_stdev;             /* Stdev of detector metrics across SCA
                                       boundary */
    double cross_pcnt;              /* Cross-boundary metric as a % of the
                                       mean radiance for the band */ 
    int cross_num_pixels;           /* # of detectors across boundary used
                                       in calculations */
} CROSS_SCA_SUMMARY;


typedef struct sca_summary
{
    int sca_number;                 /* 1-based SCA number  */
    double metric_max;              /* Largest metric in SCA */
    double metric_min;              /* Smallest metric in SCA */
    double metric_mean;             /* Mean metric value across all
                                       operable detectors in SCA */
    double metric_stdev;            /* Stdev of metric over all operable
                                       detectors in SCA */
    double pcnt;                    /* Metric as a % of the mean radiance
                                       for the band */
    int detector_max;               /* Detector with largest metric */
    int detector_min;               /* Detector with smallest metric */
    int num_characterized_detectors;/* Total number of characterized
                                       detectors in SCA */
} SCA_SUMMARY;


/* Report header data structure */
typedef struct rps_nuchar_report_header
{
    IAS_REPORT_HEADER *common_header; /* Standard IAS report header */
    RPS_REPORT_HEADER rps_header;     /* RPS-specific report header */

    int panel_id;                     /* Diffuser panel identifier */
    uint16_t diffuser_resolver_position;
                                      /* Diffuser motor position when
                                         panel is fully deployed and
                                         imaging is taking place */
} RPS_NUCHAR_REPORT_HEADER;


               /* Complete summary report data structure */
typedef struct rps_nuchar_report_summary
{

    RPS_NUCHAR_REPORT_HEADER header;      /* Nonuniformity report header */

    int correct_flag;                     /* Panel/blackbody nonuniformity
                                             correction applied */
    int calculate_radiance;                  /* Radiance data source
                                             0=calculated directly,
                                             1=read from database */
    RPS_NUCHAR_RELGAIN_OPTIONS relgain_option; /* Relative gain set selected
                                                  for internal radiance
                                                  calculations */
    int rms_banding_flag;                 /* RMS banding characterized */
    int stdev_banding_flag;               /* Stdev banding characterized */
    int streaking_flag;                   /* Streaking characterized */
    int num_contiguous_detectors;         /* Number of contiguous detectors
                                             specified for banding
                                             characterizations */
    int store_char_in_db;                 /* Characterization results
                                             written to database */
    int report_flag;                      /* Characterization results
                                             written to summary files */
    int dump_flag;                        /* Characterization results
                                             written to dump files */

    IAS_SENSOR_ID sensor_id;              /* Sensor ID for current band */
    int band_number;                      /* 1-based band number */
    char band_name[IAS_BAND_NAME_SIZE];   /* Band name */
    int scas;                             /* Number of SCAs in the current
                                             band */
    int detectors_per_sca;                /* Number of detectors in each
                                             SCA */
    IAS_SPECTRAL_TYPE spectral_type;      /* Spectral band type */

    double ffov_limit;                    /* Nonuniformity FFOV threshold
                                             limit (in %) */
    double detector_rms_banding_limit;    /* Detector-specific RMS banding
                                             threshold limit (in %) */
    double detector_stdev_banding_limit;  /* Detector-specific stdev banding
                                             threshold limit (in %) */
    double detector_streaking_limit;      /* Detector-specific streaking
                                             threshold limit */

    double rms_band_mean;                 /* RMS banding mean for the current
                                             band */
    double rms_band_stdev;                /* RMS banding std deviation
                                             for the current band */
    double stdev_band_mean;               /* Stdev banding mean for the
                                             current band */
    double stdev_band_stdev;              /* Stdev banding std deviation
                                             for the current band */
    double streaking_mean;                /* Streaking mean for the current
                                             band */
    double streaking_stdev;               /* Streaking std deviation for the
                                             currnt band */

    double ffov_metric;                   /* FFOV metric for the entire
                                             band */

    /* Current metrics summary information to be reported */
    SCA_SUMMARY *sca_summary_rms;         /* RMS banding summary
                                             statistics for all SCAs in
                                             the current band */
    CROSS_SCA_SUMMARY *cross_sca_rms;     /* RMS banding summary
                                             statistics across SCA
                                             boundaries   */
    SCA_SUMMARY *sca_summary_stdev;       /* Stdev banding summary
                                             statistics for all SCAs in
                                             the current band */
    CROSS_SCA_SUMMARY *cross_sca_stdev;   /* Stdev banding summary
                                             statistics across SCA
                                             boundaries   */
    SCA_SUMMARY *sca_summary_streaking;   /* Streaking summary statistics
                                             for all SCAs in the current
                                             band */
} RPS_NUCHAR_REPORT_SUMMARY;


/*           Data structures required for parameter I/O. */
typedef struct nuchar_wo_params 
{
    /* Work order-related information -- parameter file
       identification/location */
    char work_order_id[IAS_WORKORDER_ID_SIZE];


    /* Output report filename prefix */
    char report_filename_prefix[PATH_MAX];

    /* Work order-related information -- processing options */
    char oli_rps_proc_level[IAS_RPS_PROCESSING_LEVEL_LENGTH + 1];
    char tirs_rps_proc_level[IAS_RPS_PROCESSING_LEVEL_LENGTH + 1];
    int num_bands_to_process;
    int bands[IAS_MAX_NBANDS];      /* List of bands to process*/
    int calculate_radiance;            /* Radiance data source
                                       0=characterization database (from
                                       L1R histogram statistics char)
                                       1=calculate internally (default) */
    int relgain_option;             /* Relative gain selection options
                                       1=prelaunch (default)
                                       2=postlaunch
                                       3=best current */
    int rms_banding_flag;           /* Flag indicating whether RMS banding
                                       characterization is performed */
    int stdev_banding_flag;         /* Flag indicating whether stdev banding
                                       characterization is performed */
    int streaking_flag;             /* Flag indicating whether streaking
                                       characterization is performed */
    int correct_flag;               /* Flag indicating whether source
                                       nonunform response correction is
                                       performed (0=no, 1=yes) */
    int num_contiguous_detectors;   /* # of contiguous detectors (needed for
                                       RMS and stdev banding
                                       characterizations)--default is 100 */

    /* Work order-related information -- output options */
    int store_char_in_db;                 /* Write results to DB
                                             0=no, 1=yes */
    int report_flag;                      /* Write summary report
                                             0=no, 1=yes */
    int dump_flag;                        /* Write dump files
                                             0=no, 1=yes */

    /* Work order-related information -- reporting thresholds */
                            /* OLI */
    double nonthermal_ffov_limit_pan;            /* Max. allowed PAN band FFOV
                                                    limit (%) */
    double nonthermal_rms_banding_limit_pan;     /* Max. allowed PAN band
                                                    RMS limit (%) */
    double nonthermal_stdev_banding_limit_pan;   /* Max. allowed PAN band
                                                    stdev limit (%) */
    double nonthermal_streaking_limit_pan;       /* Max. allowed PAN band
                                                    streaking (absolute) */
    double nonthermal_ffov_limit_ms;             /* Max. allowed MS band
                                                    FFOV limit (%) */
    double nonthermal_rms_banding_limit_ms;      /* Max. allowed MS band
                                                    RMS limit (%) */
    double nonthermal_stdev_banding_limit_ms;    /* Max. allowed MS band
                                                    stdev limit (%) */
    double nonthermal_streaking_limit_ms;        /* Max. allowed MS band
                                                    streaking (absolute) */
                            /* TIRS */
    double thermal_ffov_limit_ms;                /* Max. allowed FFOV limit
                                                    (%) */
    double thermal_rms_banding_limit_ms;         /* Max. allowed RMS limit
                                                    (%) */
    double thermal_stdev_banding_limit_ms;       /* Max. allowed stdev limit
                                                    (%) */
    double thermal_streaking_limit_ms;           /* Max. allowed streaking
                                                    limit (absolute) */
} RPS_NUCHAR_WO_PARAMS;


typedef struct nuchar_omf_params
{
    int satellite_number;                  /* Satellite number */
    char l0r_filename[PATH_MAX];           /* L0Ra interval file name */
    char cpf_filename[PATH_MAX];           /* CPF file name */
    char ancillary_filename[PATH_MAX];     /* L0R ancillary file name */
    char pmask_filename[PATH_MAX];         /* Name of pixel mask file */
} RPS_NUCHAR_OMF_PARAMS;



/*************************************************************************
                   FUNCTION PROTOTYPES
**************************************************************************/

                        /* Parameter I/O */
int read_parameters
(
    const char *wo_params_filename,          /* I: Name of work order
                                                parameter file */
    RPS_NUCHAR_WO_PARAMS *wo_param_data,     /* O: WO parameters */
    RPS_NUCHAR_OMF_PARAMS *omf_param_data    /* O: OMF parameters */
);



                   /* Radiance determination */
int determine_radiances
(
    const double *stats_buffer,     /* I: Histogram statistics data buffer
                                       (DN=(0|C)BL, rad=(0|C)BLG).  For PAN 
                                       band, it's the average of the two line
                                       set values; for OLI/TIRS MS bands,
                                       it's the sole statistics value */
    int band_index,                 /* I: Current band index */
    int sca_index,                  /* I: Current SCA index */
    int num_detectors_per_sca,      /* I: Number of detectors in the current
                                       SCA */

    /* I: Absolute gains for all bands/SCAs */
    const struct IAS_CPF_ABSOLUTE_GAINS *cpf_abs_gains,

    /* I: Selected relative gains for all bands/SCAs */
    const struct IAS_CPF_RELATIVE_GAINS *cpf_rel_gains,

    /* I: Nonuniformity correction factors for all bands/SCAs
       (NULL ptr if no correction) */
    const struct IAS_CPF_NONUNIFORMITY *cpf_nonuniform,

    int calculate_radiance,              /* I: Flag indicating use of
                                            database radiance histogram
                                            statistics (0) or direct
                                            calculation from bias-corrected
                                            linearized histogram statistics
                                            (1) */
    int correct_flag,                    /* I: Correct for nonuniform
                                            panel response */
    IAS_RPS_OLI_DIFFUSER_TYPE panel_id,  /* I: Diffuser panel identifier */
    const int *invalid_flags,           /* I: invalid flags for all
                                            detectors in the current band
                                            (derived from pixel mask
                                            information) */
    double *radiance_buffer              /* O:  Estimated radiances for
                                            all detectors in the band,
                                            populated SCA by SCA */
);


int correct_nonuniform_panel_response
(
    double *data_buffer,                   /* I/O: Diffuser statistics to
                                              correct */
    const int *invalid_flags,             /* I: Data invalid flags for
                                              all detectors in the current
                                              band (derived from pixel
                                              mask information) */
    const double *cf_ptr,                  /* I: Pointer to correction
                                              factor data buffer */
    int start_detector_index,              /* I: Starting detector index */
    int num_detectors_sca                  /* I: Number of detectors in
                                              an SCA */
);



                       /* Metrics calculation */
int calculate_rms_banding_metrics
(
    const double *radiance_buffer,      /* I: Array of radiances in the
                                           current band */
    const int *statistics_invalid_flags,
                                        /* I: Input data invalid flags for
                                           all detectors in the current band
                                           (0=valid detector data,
                                            nonzero=invalid detector_data) */
    int num_band_detectors,             /* I: Total # of detectors in the
                                           current band */
    int num_contiguous_detectors,       /* I: WO parameter specifying
                                           banding width */
    double radiance_mean,               /* I: Mean radiance for the current
                                           band */
    double *rms_banding_buffer,         /* O: RMS banding metrics for the
                                           current band */
    int *metrics_invalid_flags         /* O: Characterization results
                                           flag for all detectors in the
                                           current band */
);


int calculate_stdev_banding_metrics
(
    const double *radiance_buffer,      /* I: Array of radiances in the
                                           current band */
    const int *statistics_invalid_flags,
                                        /* I: Input data invalid flags for
                                           all detectors in the current band
                                           (0=valid detector data,
                                            nonzero=invalid detector_data) */
    int    num_band_detectors,          /* I: Total # of detectors in the
                                           current band */
    int    num_contiguous_detectors,    /* I: WO parameter specifying
                                           banding width */
    double *stdev_banding_buffer,       /* O: stdev banding metrics for the
                                           current band */
    int *metrics_invalid_flags         /* O: Characterization results
                                           flag for all detectors in the
                                           current band */
);


int calculate_streaking_metrics
(
    const double *radiance_buffer, /* I: Array of radiances in the current
                                      band */
    const int *statistics_invalid_flags,
                                   /* I: Input ata invalid flag for all
                                      detectors in the current band
                                      (0=valid detector data, nonzero=invalid
                                      detector_data) */
    int    num_band_detectors,     /* I: Total # of detectors in the band */
    double *streaking_buffer,      /* O: Streaking metrics for the
                                      current band */
    int *metrics_invalid_flags    /* O: Streaking invalid flags for
                                      all detectors in the current
                                      band */
);


double calculate_pcnt_metric
(
    double raw_metric,                  /* I: Raw metric to convert to %
                                           metric */
    double band_radiance                /* I: Mean radiance for the current
                                           band */
);



                         /* "Utilities" */
void cleanup_application
(
    int wo_trend_flag,              /* I: DB trend flag */
    int num_scas,                   /* I: Number of SCAs in the
                                       current band */
    RPS_NUCHAR_REPORT_SUMMARY *rpt_summary,/* I: Summary report data */
    RPS_NUCHAR_METRICS *metrics,    /* I: Nonuniformity metrics data */
    RPS_NUCHAR_INVALID_FLAGS *invalid_flags,
                                    /* I: Data structure containing
                                       invalid flag buffers to free */
    struct ias_db_connection *db,   /* I: Open(?) database
                                       connection */
    double *local_rms_buffer,       /* I: Band-level RMS metrics
                                       buffer */
    double *local_stdev_buffer,     /* I: Band-level stdev metrics
                                       buffer */
    double *local_streaking_buffer, /* I: Band-level streaking metrics
                                         buffer */
    double *radiance_buffer,        /* I: Band-level radiance_buffer */
    double *stats_buffer,           /* I: Histogram statistics for the
                                       current band/SCA */
    IAS_CPF *cpf,                   /* I: CPF data structure */
    IAS_PIXEL_MASK **pmasks,        /* I: Pixel mask array */
    int num_masks_read              /* I: Number of pixel masks read from 
                                       the pixel mask file */
);


void cleanup_for_next_band
(
    int num_scas,                          /* I: Number of SCAs in the
                                              current band */
    RPS_NUCHAR_REPORT_SUMMARY *rpt_summary,/* I: Summary report data */
    RPS_NUCHAR_METRICS *metrics,           /* I: Nonuniformity
                                              metrics data */
    RPS_NUCHAR_INVALID_FLAGS *invalid_flags,
                                    /* I: Data structure containing
                                       invalid flag buffers to free */
    double *local_rms_buffer,       /* I: Band-level RMS metrics buffer */
    double *local_stdev_buffer,     /* I: Band-level stdev metrics buffer */
    double *local_streaking_buffer, /* I: Band-level streaking metrics
                                       buffer */
    double *radiance_buffer,        /* I: Band-level radiance_buffer */
    double *stats_buffer            /* I: Histogram statistics for the
                                       current band/SCA (MS bands, lineset=2),
                                       and PAN band 'even' lines (lineset=0) */
);


int allocate_metrics_buffers
(
    int RMS_banding_flag,              /* I: Flag indicating whether RMS
                                          banding characterization is
                                          performed */
    int stdev_banding_flag,            /* I: Flag indicating whether
                                          stdev banding characterization
                                          is performed */
    int streaking_flag,                /* I: Flag indicating whether
                                          streaking characterization is
                                          performed */
    int num_scas,                      /* I: # of SCAs in the current
                                          band */
    int num_detectors_per_sca,         /* I: # of detectors in an SCA */
    RPS_NUCHAR_METRICS *metrics        /* O: Metrics data structure */
);


int allocate_summary_report_buffers
(
    int rms_banding_flag,                    /* I: Flag indicating whether
                                                RMS banding characterization
                                                is performed  */
    int stdev_banding_flag,                  /* I: Flag indicating whether
                                                stdev banding
                                                characterization is
                                                performed */
    int streaking_flag,                      /* I: Flag indicating whether
                                                streaking characterization
                                                is performed */
    int num_scas,                            /* I: # of SCAs in the current
                                                band */
    RPS_NUCHAR_REPORT_SUMMARY *rpt_summary   /* O: Summary info data
                                                structure */
);


void populate_metrics_buffers
(
    const double *band_rms_buffer,  /* I: RMS metrics data buffer for all
                                       detectors in the current band */
    const double *band_stdev_buffer,/* I: Stdev metrics data buffer for
                                       all detectors in the current band */
    const double *band_streaking_buffer,
                                    /* I: Streaking metrics data buffer
                                       for all detectors in the current
                                       band */
    int rms_banding_flag,           /* I: Flag indicating whether RMS
                                       banding characterization is
                                       performed */
    int stdev_banding_flag,         /* I: Flag indicating whether stdev
                                       anding characterization is
                                       performed */
    int streaking_flag,             /* I: Flag indicating whether
                                       streaking characterization is
                                       performed */
    int sca_index,                  /* I: Current 0-based SCA index */
    int num_detectors_per_sca,      /* I: Number of detectors in each
                                       SCA */
    RPS_NUCHAR_METRICS *metrics     /* O: Populated valid metrics data
                                       structure for the current SCA */
);


int write_report_header_to_file
(
    RPS_NUCHAR_REPORT_SUMMARY summary,    /* I: Summary report information */
    const char *work_order_id             /* I: Work order ID currently
                                             processing */
);


int allocate_nonuniformity_invalid_flag_buffers
(
    int num_detectors,                /* I: Number of detectors */
    RPS_NUCHAR_INVALID_FLAGS *flags  /* O: invalid flag data structure */
);


void free_nonuniformity_invalid_flag_buffers
(
    RPS_NUCHAR_INVALID_FLAGS *flags    /* I: invalid flag data structure
                                           with buffers to free */
);



                      /* Database prep/access */
struct ias_db_connection *connect_to_db();


int clear_nonuniformity_table
(
    struct ias_db_connection *db, /* I: Open database connection */
    char *rad_wo_common_id,       /* I: Radiometric work order common ID */
    const char *table_name        /* I: Name of solarcal table to clear */
);


int get_histogram_statistics
(
    struct ias_db_connection *db,  /* I: Open database connection */
    char *rad_wo_common_id,        /* I: RADIOMETRIC_WORK_ORDER_COMMON
                                      primary key value */
    IAS_SPECTRAL_TYPE spectral_type, /* I: Spectral type of current band */
    int band_number,               /* I: Current 1-based band number */
    int sca_number,                /* I: Current 1-based SCA number */
    int num_detectors_sca,         /* I: Number of detectors for the current
                                      band/SCA */
    const char *rps_proc_level,    /* I: RPS processing level */
    double *stats_buffer,          /* O: Histogram statistics for the
                                      current band/SCA */
    const char *collection_type    /* I: Collection type data */
);


int write_to_detector_table
(
    struct ias_db_connection *db,        /* I: Open database connection */
    char *rad_wo_common_id,              /* I: Database ID corresponding
                                            to currently processing
                                            radiometric work order */
    int band_number,                     /* I: Current (1-based) band # */
    int sca_number,                      /* I: Current (1-based) SCA # */
    int num_detectors_sca,               /* I: Total number of detectors in
                                            each SCA */
    int rms_banding_flag,                /* I: Flag indicating whether
                                            RMS banding characterization
                                            is performed */
    int stdev_banding_flag,              /* I: Flag indicating whether
                                            stdev banding
                                            characterization is
                                            performed */
    int streaking_flag,                  /* I: Flag indicating whether
                                            streaking characterization
                                            is performed */
    int *detector_ids,                   /* I: Detector ID numbers for the
                                            current band/SCA */
    IAS_DB_NULL_TYPE *flagged_rms_detectors, /* I: Flagged RMS detector
                                                      indices */
    double *pcnt_rms_metrics,            /* I: Detector RMS banding
                                            metrics converted to %
                                            values */
    IAS_DB_NULL_TYPE *flagged_stdev_detectors, /* I: Flagged stdev
                                                        detector indices */
    double *pcnt_stdev_metrics,          /* I: Detector stdev banding
                                            metrics converted to % values */
    IAS_DB_NULL_TYPE *flagged_streaking_detectors,
                                         /* I: Flagged streaking detector
                                            indices */
    double *db_streaking_metrics         /* I: Detector streaking metrics */
);


int write_to_band_table
(
    struct ias_db_connection *db,        /* I: Open DB connection */
    char *radiometric_wo_common_number,  /* I: Database ID corresponding
                                            to  currently processing
                                            radiometric work order */
    int band_number,                     /* I: Current (1-based) band # */
    double ffov_metric                   /* I: Value to write to DB */
);


void populate_database_detector_write_buffers
(
    const METRICS_DET *rms_metrics,       /* I: Detector RMS metrics for the
                                             current band/SCA */
    const METRICS_DET *stdev_metrics,     /* I: Detector stdev metrics for
                                             the current band/SCA */
    const METRICS_DET *streaking_metrics, /* I: Detector streaking metrics
                                             for the current band/SCA */
    const int *rms_banding_invalid_flags,/* I: RMS banding characterization
                                             result invalid flags for all
                                             detectors in the current
                                             band/SCA */
    const int *stdev_banding_invalid_flags,/* I: STDEV banding
                                            characterization invalid flags
                                            for all detectors in the current
                                            band/SCA */
    const int *streaking_invalid_flags,  /* I: Streaking characterization
                                             invalid flags for all
                                             detectors in the current
                                             band/SCA */
    double band_radiance,                 /* I: Mean radiance for the
                                             current band */
    int rms_banding_flag,                 /* I: Flag indicating whether
                                             RMS banding characterization
                                             is performed */
    int stdev_banding_flag,               /* I: Flag indicating whether
                                             stdev banding
                                             characterization is
                                             performed */
    int streaking_flag,                   /* I: Flag indicating whether
                                             streaking characterization
                                             is performed */
    int num_detectors_per_sca,            /* I: Number of detectors in an
                                             SCA */
    IAS_DB_NULL_TYPE *rms_flagged_detectors, /* O: Flagged RMS detector
                                                indices */
    double *pcnt_rms_metrics,             /* O: Detector RMS banding
                                             metrics converted to %
                                             values */
    IAS_DB_NULL_TYPE *stdev_flagged_detectors, /* O: Flagged stdev detector
                                             indices */
    double *pcnt_stdev_metrics,           /* O: Detector stdev banding
                                             metrics converted to % values */
    IAS_DB_NULL_TYPE *streaking_flagged_detectors,
                                          /* O: Flagged streaking detector
                                             indices */
    double *db_streaking_metrics          /* O: Detector streaking metrics */
);



                 /* Initialization of summary report data */
void initialize_nonuniformity_parameter_data
(
    RPS_NUCHAR_WO_PARAMS wo_params,       /* I: Work order parameters */
    RPS_NUCHAR_REPORT_SUMMARY *summary    /* O: Populated report summary
                                             data structure */
);


void initialize_summary_report_band_data
(
    RPS_NUCHAR_WO_PARAMS wo_params,              /* I: Contains band
                                                    nonuniformity limits */
    const IAS_BAND_ATTRIBUTES *band_attributes,  /* I: Attributes for
                                                    current band */
    RPS_NUCHAR_REPORT_SUMMARY *summary           /* I/O: Populated
                                                    summary report data
                                                    structure */
);


int initialize_sca_summary_metrics
(
    RPS_NUCHAR_METRICS metrics,          /* I: Metrics data structure */
    const RPS_NUCHAR_INVALID_FLAGS *invalid_flags,
                                         /* I: Characterization result
                                            invalid flags for all
                                            detectors in the current band */
    int rms_banding_flag,                /* I: Flag indicating whether
                                            RMS banding characterization is
                                            performed */
    int stdev_banding_flag,              /* I: Flag indicating whether
                                            stdev banding characterization
                                            is performed */
    int streaking_flag,                  /* I: Flag indicating whether
                                            streaking characterization
                                            is performed */
    int num_scas,                        /* I: Total # of SCAs in the
                                            current band */
    int num_detectors_per_sca,           /* I: Total # of detectors per
                                            SCA */
    RPS_NUCHAR_REPORT_SUMMARY *summary   /* O: Populated summary report
                                            data structure (also used for
                                            dump file) */
);



                 /* Summary metrics calculation */
int summarize_within_sca_metrics
(
    const METRICS_DET *detector_metrics_data, /* I: Detector metrics data
                                                 for the current
                                                 characterization */
    const int *metrics_invalid_flags,        /* I: invalid flags for all
                                                 detectors in the current
                                                 SCA for the current
                                                 characterization */
    int num_detectors_per_sca,                /* I: Total # of detectors
                                                 in an SCA */
    double band_radiance,                     /* I: Mean radiance for the
                                                 current band */
    SCA_SUMMARY *sca_summary                  /* O: Populated data structure
                                                 with per-SCA metrics summary
                                                 statistics for the
                                                 current characterization */
);


int summarize_cross_sca_metrics
(
    const METRICS_DET *detector_metrics_data, /* I: Detector metrics data
                                                 for all SCAs in the current
                                                characterization */
    const int *metrics_invalid_flags,        /* I: Metrics invalid flags for
                                                 all detectors in the current
                                                 SCA for the current
                                                 characterization */
    int sca_index,                            /* I: 0-based index for the
                                                 current SCA */
    int num_detectors_per_sca,                /* I: Total # of detectors
                                                 in an SCA */
    int num_boundary_detectors,               /* I: # of detectors on both
                                                 sides of an SCA boundary */
    const char *cross_metric_basis,           /* I: String indicating the
                                                 basis for the cross-metric
                                                 calculation ("MEAN",
                                                "STDEV") */
    double band_radiance,                     /* I: Mean radiance for the
                                                 current band */
    CROSS_SCA_SUMMARY *cross_sca_summary      /* O: Populated data structure
                                                 with cross-SCA summary
                                                 statistics for the
                                                 current characterization */
);


int calculate_per_detector_metrics_std
(
    const double *band_rms_buffer,         /* I: RMS metrics buffer for the
                                              current band */
    const double *band_stdev_buffer,       /* I: Stdev banding metrics
                                              buffer for the current band */
    const double *band_streaking_buffer,   /* I: Streaking metrics buffer
                                              for the current band */
    const RPS_NUCHAR_INVALID_FLAGS *invalid_flags,
                                           /* I: Characterization result
                                              invalid flags for all
                                              detectors in the current
                                              band */
    int num_band_detectors,                /* I: Total # of detectors in the
                                              current band */
    int rms_banding_flag,                  /* I: Flag indicating whether
                                              RMS banding characterization
                                              is performed */
    int stdev_banding_flag,                /* I: Flag indicating whether
                                              stdev banding characterization
                                              is performed */
    int streaking_flag,                    /* I: Flag indicating whether
                                              streaking characterization is
                                              performed */
    RPS_NUCHAR_REPORT_SUMMARY *summary     /* O: Populated metrics data
                                              structure */
);



                       /* Summary report writing */
int write_summary_report_data
(
    FILE *rpt_fptr,                         /* I: Open report file handle */
    RPS_NUCHAR_REPORT_SUMMARY summary,      /* I: SCA-specific metrics
                                               summary information */
    double band_radiance_mean,              /* I: Summary radiance
                                               statistics for the current
                                               band */
    double band_radiance_stdev,
    double band_ffov_metric                 /* I: Band-level FFOV
                                               nonuniformity metric */
);



                     /* Dump file writing */
int write_metrics_data_to_dump_file
(
    RPS_NUCHAR_METRICS metrics,           /* I: Detector-level metrics
                                             array */
    RPS_NUCHAR_REPORT_SUMMARY summary,    /* I: Summary-level metrics
                                             information */
    const RPS_NUCHAR_INVALID_FLAGS *invalid_flags,
                                          /* I: Characterization result
                                             invalid flags for all
                                             detectors in the current
                                             band */
    FILE *dump_fptr                       /* I: Open dump file handle */
);



#endif
