/**************************************************************************
 NAME:              char_nonuniformity_db 

 PURPOSE:  Module implementing routines to provide IAS characterization
           database access for the nonuniformity characterization algorithm.

****************************************************************************/
#include <stdlib.h>
#include <string.h>

#include "ias_db_get_connect_info.h"
#include "ias_db_insert.h"
#include "ias_db_query.h"
#include "ias_logging.h"
#include "ias_radiometry.h"

#include "char_nonuniformity.h"

#define SQL_QUERYLEN 1024

/*--------------------------------------------------------------------
 NAME:                  connect_to_db

 PURPOSE:    Retrieves IAS database connection information and attempts
             the connection

 RETURNS:    - Pointer to valid ias_db_struct, indicating connection has
               been established
             - NULL pointer if connection cannot be made
-----------------------------------------------------------------------*/
struct ias_db_connection *connect_to_db()
{
    char *db_name = NULL;            /* Required connection information */
    char *user_id = NULL;
    char *user_password = NULL;
    char *db_host = NULL;

    struct ias_db_connection *db = NULL;

    int status;


    /* Initialize the library */
    IAS_LOG_INFO("Initializing database library...");
    status = ias_db_initialize_database_lib();
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing database library");
        return NULL;
    }

    /* Get the connection information */
    IAS_LOG_DEBUG("Obtaining database connection information...");
    status = ias_db_get_connect_info(&db_name, &user_id, &user_password,
            &db_host);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Obtaining database connection information");
        return NULL;
    }

    IAS_LOG_DEBUG("DB Name:    %s\nUser ID:  %s\n", db_name, user_id);

    /* Try to connect */
    IAS_LOG_DEBUG("Connecting to database...");

    db = ias_db_connect_to_database(db_name, user_id,
            user_password, db_host);

    /* Free the allocated strings with the connection information */
    free(db_name);
    free(user_id);
    free(user_password);
    free(db_host);

    /* Did we connect? */
    if (db == NULL)
        IAS_LOG_ERROR("Connecting to database");
    else
        IAS_LOG_DEBUG("Database connection established");

    /* Return whatever we've got for *db (either NULL or a valid pointer) */
    return db;
}




/*------------------------------------------------------------------------
 NAME:                     clear_nonuniformity_table

 PURPOSE:       Clears the specified NON_UNIFORMITY_XXXX table for the
                specified radiometric work order ID key

 RETURNS:       Integer status code of SUCCESS (tables successfully
                cleared) or ERROR (table(s) couldn't be cleared)
--------------------------------------------------------------------------*/
int clear_nonuniformity_table
(
    struct ias_db_connection *db, /* I: Open database connection */
    char *rad_wo_common_id        /* I: Radiometric work order common ID */,
    const char *table_name        /* I: Name of solarcal table to clear */
)
{
    int status;                   /* Function return status code */

    /* Clear out the specified table */
    IAS_LOG_DEBUG("Clearing %s table...", table_name);

    status = ias_db_delete_records(db, table_name,
            "radiometric_wo_common_id",
            rad_wo_common_id);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Clearing %s table", table_name);
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}




/*-------------------------------------------------------------------------
 NAME:           get_histogram_statistics

 PURPOSE:     Queries the HISTOGRAM_STATISTICS_EARTH or 
              HISTOGRAM_STATISTICS_CAL table depending upon the collection 
              type for the mean DN values extracted from the diffuser/blackbody
              data product for the desired work order (indirectly
              represented by the specified RADIOMETRIC_WORK_ORDER_COMMON
              primary key value)

 RETURNS:     Integer status code of SUCCESS or ERROR
---------------------------------------------------------------------------*/
int get_histogram_statistics
(
    struct ias_db_connection *db,  /* I: Open database connection */
    char *rad_wo_common_id,        /* I: RADIOMETRIC_WORK_ORDER_COMMON
                                      primary key value */
    IAS_SPECTRAL_TYPE spectral_type, /* I: Spectral type of current band */
    int band_number,               /* I: Current 1-based band number */
    int sca_number,                /* I: Current 1-based SCA number */
    int num_detectors_sca,         /* I: Number of detectors for the current
                                      band/SCA */
    const char *rps_proc_level,    /* I: RPS processing level */
    double *stats_buffer,          /* O: Histogram statistics for the
                                      current band/SCA (MS bands, lineset=2),
                                      and average of PAN band
                                      (AVG(lineset0, linset1))*/
    const char *collection_type    /* I: Collection type data */ 
)
{
    char query_string[SQL_QUERYLEN];            /* SQL query string */
    char db_error_message[IAS_SQL_RTN_MSGLEN];  /* Query error message */
    struct ias_db_query *query = NULL;       /* Query data */
    int ii;                                  /* Loop counter */
    int status;
    const char *hist_stats_table_name;  /* Correct table name for histogram 
                                           statistics */ 


    /* Get the correct histogram statistics table name according to the
        collection type provided. */
    hist_stats_table_name =
        ias_rps_get_histogram_statistics_table_name(collection_type);

    /* Prepare the query string.  It is assumed here that histogram
       statistics values are calculated for every detector, whether the
       data are valid or not.  Use snprintf to write the SQL string so
       we can check for buffer overrun and deal with it appropriately.

       For the PAN band, write the query so that it returns the average
       scene_pixel_mean for the lineset 0 AND lineset 1 data.  For the MS
       bands, set lineset to 2 */
    if (spectral_type == IAS_SPECTRAL_PAN)
    {
        status = snprintf(query_string, sizeof(query_string),
            "SELECT detector_id, AVG(scene_pixel_mean) "
            "  FROM %s "
            " WHERE band_id = %d "
            "   AND sensor_chip_assembly_id = %d "
            "   AND radiometric_wo_common_id = %s "
            "   AND processing_level='%s' "
            "   AND lineset IN (0, 1) "
            " GROUP BY detector_id "
            " ORDER BY detector_id;", 
            hist_stats_table_name, band_number, sca_number, 
            rad_wo_common_id, rps_proc_level);
    }
    else
    {
        status = snprintf(query_string, sizeof(query_string),
            "SELECT detector_id, scene_pixel_mean "
            "  FROM %s "
            " WHERE band_id = %d "
            "   AND sensor_chip_assembly_id = %d "
            "   AND radiometric_wo_common_id = %s "
            "   AND processing_level='%s' "
            "   AND lineset = 2 "
            " ORDER BY detector_id;", 
            hist_stats_table_name, band_number, sca_number, 
            rad_wo_common_id, rps_proc_level);
    }
    if ((status < 0) || (status >= sizeof(query_string)))
    {
        IAS_LOG_ERROR("Building histogram statistics query for band %d "
            "SCA %d", band_number, sca_number);
        return ERROR;
    }

    /* Execute the query and verify it worked */
    query = ias_db_perform_query(db, query_string);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
                                       sizeof(db_error_message));
        IAS_LOG_ERROR("Query of %s failed:  %s",
                      hist_stats_table_name, db_error_message);
        return ERROR;
    }

    /* Start fetching the detector means.  Pick up the detector ids
       as well */
    for (ii = 0; ii < num_detectors_sca; ii++)
    {
        double stats_value;            /* Histogram statisics mean */
        int dummy_detector;            /* 1-based detector ID number */
        int detector_index;            /* 0-based index corresponding
                                          to detector ID */

        /* Advance to the next record.  Have the call go here,
           inside the detector loop.  This way, there's no need to
           check at the end whether we've reached the proper ending
           point (where the next call to ias_db_query_next would
           produce an error because it went past the last valid record).
           Plus, it takes care of the first record */
        if (!ias_db_query_next(query))
        {
            ias_db_query_get_error_message(query, db_error_message,
                sizeof(db_error_message));
            IAS_LOG_ERROR("Query of %s table failed:  "
                "%s", hist_stats_table_name, db_error_message);
            ias_db_query_close(query);
            return ERROR;
        }
        status = ias_db_query_int_value(query, &dummy_detector, 0);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading histogram statistics "
                    "detector_id value for detector "
                    "index %d in band %d SCA %d",
                    ii, band_number, sca_number);
            ias_db_query_close(query);
            return ERROR;
        }
        detector_index = dummy_detector - 1;

        status = ias_db_query_double_value(query, &stats_value, 1);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading histogram statistics mean "
                    "value for detector %d in band %d SCA "
                    "%d", dummy_detector, band_number,
                    sca_number);
            ias_db_query_close(query);
            return ERROR;
        }
        stats_buffer[detector_index] = stats_value;
    }

    /* Release query resources */
    ias_db_query_close(query);

    /* Finished. */
    return SUCCESS;
}




/*-------------------------------------------------------------------------
 NAME:           write_to_detector_table

 PURPOSE:     Inserts one set of non-uniformity metrics into the
              NON_UNIFORMITY_detector table for the current band and SCA,
              using table-based insert functionality 

 RETURNS:     Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int write_to_detector_table
(
    struct ias_db_connection *db,        /* I: Open database connection */
    char *rad_wo_common_id,              /* I: Database ID corresponding
                                            to currently processing
                                            radiometric work order */
    int band_number,                     /* I: Current (1-based) band # */
    int sca_number,                      /* I: Current (1-based) SCA # */
    int num_detectors_sca,               /* I: Total number of detectors in
                                            each SCA */
    int rms_banding_flag,                /* I: Flag indicating whether
                                            RMS banding characterization
                                            is performed */
    int stdev_banding_flag,              /* I: Flag indicating whether
                                            stdev banding
                                            characterization is
                                            performed */
    int streaking_flag,                  /* I: Flag indicating whether
                                            streaking characterization
                                            is performed */
    int *detector_ids,                   /* I: Detector ID numbers for the
                                            current band/SCA */
    IAS_DB_NULL_TYPE *flagged_rms_detectors, /* I: Flagged RMS detector
                                                      indices */
    double *pcnt_rms_metrics,            /* I: Detector RMS banding
                                            metrics converted to %
                                            values */
    IAS_DB_NULL_TYPE *flagged_stdev_detectors, /* I: Flagged stdev
                                                        detector indices */
    double *pcnt_stdev_metrics,          /* I: Detector stdev banding
                                            metrics converted to % values */
    IAS_DB_NULL_TYPE *flagged_streaking_detectors,
                                         /* I: Flagged streaking detector
                                            indices */
    double *db_streaking_metrics         /* I: Detector streaking metrics */
)
{

    /* Misc variables */
    int status;
    int table_size;


    /* Set up the insert table for the current band/SCA
       assuming none of the characterizations have been performed.
       This can be done by setting the metrics entries for all
       detectors to NULL using IAS_DB_TABLE_DOUBLE_ARRAY_NULL.  */
    IAS_DB_INSERT_TABLE table[] =
    {
        IAS_DB_TABLE_STRING("radiometric_wo_common_id",
                rad_wo_common_id),
        IAS_DB_TABLE_INT("band_id", band_number),
        IAS_DB_TABLE_INT("sensor_chip_assembly_id", sca_number),
        IAS_DB_TABLE_INT_ARRAY("detector_id", detector_ids),
        IAS_DB_TABLE_DOUBLE_ARRAY_NULL("banding_metric_rms",
                pcnt_rms_metrics),
        IAS_DB_TABLE_DOUBLE_ARRAY_NULL("banding_metric_stdev",
                pcnt_stdev_metrics),
        IAS_DB_TABLE_DOUBLE_ARRAY_NULL("streaking",
                db_streaking_metrics)
    };

    /* For each characterization that is specified, create a table
       that contains just the resulting metrics values.  It's possible
       that some set of detectors won't have valid metrics values,
       so set the table up to allow NULL entries (using
       IAS_DB_TABLE_DOUBLE_ARRAY_WITH_NULLS).  Assign these tables to
       the corresponding entries in the main insert table defined
       previously.  */
    if (rms_banding_flag)
    {
        IAS_DB_INSERT_TABLE rms_table = 
            IAS_DB_TABLE_DOUBLE_ARRAY_WITH_NULLS("banding_metric_rms",
                    pcnt_rms_metrics,
                    flagged_rms_detectors);

        table[4] = rms_table;
    }

    if (stdev_banding_flag)
    {
        IAS_DB_INSERT_TABLE stdev_table = 
            IAS_DB_TABLE_DOUBLE_ARRAY_WITH_NULLS("banding_metric_stdev",
                    pcnt_stdev_metrics,
                    flagged_stdev_detectors);

        table[5] = stdev_table;
    }

    if (streaking_flag)
    {
        IAS_DB_INSERT_TABLE streaking_table = 
            IAS_DB_TABLE_DOUBLE_ARRAY_WITH_NULLS("streaking",
                    db_streaking_metrics,
                    flagged_streaking_detectors);

        table[6] = streaking_table;
    }

    /* Calculate the main table size, then write to the database */
    table_size = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);

    status = ias_db_insert_using_table(db, "non_uniformity_detector",
            table, table_size, num_detectors_sca);

    /* Done */
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Write to non_uniformity_detector table failed for "
                "band %d, SCA %d", band_number, sca_number);
    }

    return status;
}




/*-------------------------------------------------------------------------
 NAME:           write_to_band_table

 PURPOSE:     Inserts each band's FFOV non-uniformity metric into the
              NON_UNIFORMITY_BAND table, using table-based insert
              functionality 
           
 RETURNS:     Integer status code of SUCCESS or ERROR 
---------------------------------------------------------------------------*/
int write_to_band_table
(
    struct ias_db_connection *db,        /* I: Open DB connection */
    char *rad_wo_common_id,              /* I: Database ID corresponding
                                            to  currently processing
                                            radiometric work order */
    int band_number,                     /* I: Current (1-based) band # */
    double ffov_metric                   /* I: Value to write to DB */
)
{
    int status;                            /* Status return value */
    int num_records = 1;

    /* Define the data set we're writing to the table */
    IAS_DB_INSERT_TABLE table[] =
    {
        IAS_DB_TABLE_STRING("radiometric_wo_common_id", rad_wo_common_id),
        IAS_DB_TABLE_INT("band_id", band_number),
        IAS_DB_TABLE_DOUBLE("fullfov_uniformity_metric", ffov_metric)
    };

    int table_size = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);

    /* Insert the metric data into the NON_UNIFORMITY_BAND table */
    status = ias_db_insert_using_table(db, "non_uniformity_band",
                                       table, table_size, num_records);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Write to non_uniformity_band table failed for "
                "band %d", band_number);
    }

    /* Done */
    return status;
}




/*------------------------------------------------------------------------
 NAME:          populate_database_detector_write_buffers

 PURPOSE:     Searches the metrics data buffers and
              - flags the indices for detectors that could not be/were not
                characterized
              - converts the "raw" metrics values to "%-of-mean-radiance"
                values

 RETURNS:     Nothing
--------------------------------------------------------------------------*/
void populate_database_detector_write_buffers
(
    const METRICS_DET *rms_metrics,       /* I: Detector RMS metrics for the
                                             current band/SCA */
    const METRICS_DET *stdev_metrics,     /* I: Detector stdev metrics for
                                             the current band/SCA */
    const METRICS_DET *streaking_metrics, /* I: Detector streaking metrics
                                             for the current band/SCA */
    const int *rms_banding_invalid_flags, /* I: RMS banding characterization
                                             result invalid flags for all
                                             detectors in the current
                                             band/SCA */
    const int *stdev_banding_invalid_flags,/* I: STDEV banding
                                            characterization invalid flags
                                            for all detectors in the current
                                            band/SCA */
    const int *streaking_invalid_flags,   /* I: Streaking characterization
                                             invalid flags for all
                                             detectors in the current
                                             band/SCA */
    double band_radiance,                 /* I: Mean radiance for the
                                             current band */
    int rms_banding_flag,                 /* I: Flag indicating whether
                                             RMS banding characterization
                                             is performed */
    int stdev_banding_flag,               /* I: Flag indicating whether
                                             stdev banding
                                             characterization is
                                             performed */
    int streaking_flag,                   /* I: Flag indicating whether
                                             streaking characterization
                                             is performed */
    int num_detectors_per_sca,            /* I: Number of detectors in an
                                             SCA */
    IAS_DB_NULL_TYPE *rms_flagged_detectors, /* O: Flagged RMS detector
                                                indices */
    double *pcnt_rms_metrics,             /* O: Detector RMS banding
                                             metrics converted to % values */
    IAS_DB_NULL_TYPE *stdev_flagged_detectors, /* O: Flagged stdev detector
                                             indices */
    double *pcnt_stdev_metrics,           /* O: Detector stdev banding
                                             metrics converted to % values */
    IAS_DB_NULL_TYPE *streaking_flagged_detectors,
                                          /* O: Flagged streaking detector
                                             indices */
    double *db_streaking_metrics          /* O: Detector streaking metrics */
)
{
    int detector_index;              /* Current detector */
    IAS_DB_NULL_TYPE null_data_flag; /* Database NULL value */

    /* Get the NULL value used by the database */
    null_data_flag = ias_db_get_null_data_flag();

    if (rms_banding_flag)
    {
        for (detector_index = 0;
             detector_index < num_detectors_per_sca;
             detector_index++)
        {
            if (rms_banding_invalid_flags[detector_index])
            {
                rms_flagged_detectors[detector_index] = null_data_flag;
                continue;
            }

            pcnt_rms_metrics[detector_index] = calculate_pcnt_metric(
                rms_metrics->metric[detector_index], band_radiance);
        }
    }

    if (stdev_banding_flag)
    {
        for (detector_index = 0;
             detector_index < num_detectors_per_sca;
             detector_index++)
        {
            if (stdev_banding_invalid_flags[detector_index])
            {
                stdev_flagged_detectors[detector_index] = null_data_flag;
                continue;
            }

            pcnt_stdev_metrics[detector_index] = calculate_pcnt_metric(
                stdev_metrics->metric[detector_index], band_radiance);
        }
    }

    if (streaking_flag)
    {
        for (detector_index = 0;
             detector_index < num_detectors_per_sca;
             detector_index++)
        {
            if (streaking_invalid_flags[detector_index])
            {
                streaking_flagged_detectors[detector_index] = null_data_flag;
                continue;
            }

            db_streaking_metrics[detector_index] =
                streaking_metrics->metric[detector_index];
        }
    }

    /* Done */
}
