/***************************************************************************
 NAME:                    determine_radiances

 PURPOSE:       Given radiance histogram statistics for the 
                current band/SCA, copies them directly to the output
                radiance buffer.  Given bias-corrected/linearized histogram
                statistics for the current band/SCA, converts them to
                radiances by application of the proper absolute and relative
                radiometric gains, then writes those to the output radiance
                data buffer.  In both cases, correction to account for
                nonuniform source response can be performed if specified in
                the work order.

 RETURNS:       Integer status code of SUCCESS or ERROR

****************************************************************************/
#include <string.h>
#include "ias_const.h"
#include "ias_cpf.h"
#include "ias_logging.h"
#include "ias_radiometry.h"
#include "ias_satellite_attributes.h"

#include "char_nonuniformity.h"


int determine_radiances
(
    const double *stats_buffer,     /* I: Histogram statistics data buffer
                                       (DN=(0|C)BL, rad=(0|C)BLG).  For PAN 
                                       band, it's the average of the two line
                                       set values; for OLI/TIRS MS bands,
                                       it's the sole statistics value */
    int band_index,                 /* I: Current band index */
    int sca_index,                  /* I: Current SCA index */
    int num_detectors_per_sca,      /* I: Number of detectors in the current
                                       SCA */

    /* I: Absolute gains for all bands/SCAs */
    const struct IAS_CPF_ABSOLUTE_GAINS *cpf_abs_gains,

    /* I: Selected relative gains */
    const struct IAS_CPF_RELATIVE_GAINS *cpf_rel_gains,

    /* I: Nonuniformity correction factors for all bands/SCAs
       (NULL ptr if no correction) */
    const struct IAS_CPF_NONUNIFORMITY *cpf_nonuniform,

    int calculate_radiance,              /* I: Flag to use database radiance
                                            histogram statistics or
                                            calculate them directly */
    int correct_flag,                    /* I: Correct for nonuniform
                                            panel/blackbody response */
    IAS_RPS_OLI_DIFFUSER_TYPE panel_id,  /* I: Diffuser panel identifier */
    const int *invalid_flags,            /* I: Invalid flags for all
                                            detectors in the current band
                                            (derived from pixel mask
                                            information) */
    double *radiance_buffer              /* O: Estimated radiances for
                                            all detectors in the band,
                                            populated SCA by SCA */
)
{
    int status;                 /* Function return status */
    int start_detector_index;   /* Index corresponding to first detector
                                   in an SCA */

    double sca_abs_gain;        /* Absolute gain for the current
                                   band/SCA */
    const double *relgain_ptr = NULL; /* Pointer to returned set of
                                         relative gains */
    const double *cfactor_ptr = NULL; /* Pointer to correction factors for
                                         the current band/SCA */


    /* Calculate the starting detector index for the current SCA */
    start_detector_index = sca_index * num_detectors_per_sca;

    /* If radiance histogram statistics have been obtained from
       processed L1R diffuser image data and the WO specifies these,
       copy them directly to the radiance data buffer */
    if (calculate_radiance == RPS_NUCHAR_USE_DB_RADIANCE)
    {
        memcpy(&(radiance_buffer[start_detector_index]),
           stats_buffer, num_detectors_per_sca * sizeof(*stats_buffer));
    }
    else        /* Calculate radiances directly */
    {
        /* Use the proper absolute gain and relative gains for the
           current band/SCA.  The relative gains are dependent on
           the work order selection */
        sca_abs_gain = cpf_abs_gains->gain[band_index][sca_index];
        relgain_ptr =
            cpf_rel_gains->per_detector[band_index][sca_index];

        /* Calculate the radiances for the current SCA. */
        ias_rps_calculate_radiance(stats_buffer,
            num_detectors_per_sca, sca_abs_gain, relgain_ptr,
            &(radiance_buffer[start_detector_index]));
    }

    /* Adjust the radiances here.  The optional panel correction is
       accounted for, through 'cfactor_ptr' pointing to either the
       proper correction factors for the current band/SCA (if set
       in the work order), or NULL (if no correction is desired).
       Initially, assume it's sufficient to just correct the 
       final radiance values */
    if (correct_flag)
    {
        if (cpf_nonuniform == NULL)
        {
            IAS_LOG_ERROR("Panel correction specified but no "
                    "correction factor data available");
            return ERROR;
        }

        /* Obtain the set of correction factors.  If the panel ID indicates
           the pristine diffuser panel was deployed, get the pristine
           correction factors.  If the panel ID indicates another valid value
           (i.e. either the working diffuser panel was deployed for an OLI
           solar collect OR this was a TIRS collect and the panel didn't
           get deployed at all), get the primary correction factors */
        if (panel_id == IAS_RPS_OLI_DIFFUSER_TYPE_PRISTINE)
        {
            cfactor_ptr = cpf_nonuniform
                ->scale_factor_2[band_index][sca_index];
        }
        else
        {
            cfactor_ptr = cpf_nonuniform
                ->scale_factor_1[band_index][sca_index];
        }

        /* Do the correction */
        status = correct_nonuniform_panel_response(radiance_buffer,
            invalid_flags, cfactor_ptr, start_detector_index,
            num_detectors_per_sca);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Failed to correct nonuniform panel response");
            return ERROR;
        }
    }

    /* Done (at least for this SCA) */
    return SUCCESS;
}




/*------------------------------------------------------------
 NAME:            correct_nonuniform_panel_response

 PURPOSE:     Corrects the current band's diffuser data for observed
              non-uniformities in panel response.  Run only if
              such correction is selected in the work order

 RETURNS:     Integer status code of SUCCESS or ERROR

 NOTES:       If additional processing needs to occur here, pull
              this out and make it a separate routine in a separate
              source file
--------------------------------------------------------------*/
int correct_nonuniform_panel_response
(
    double *radiance_buffer,               /* I/O: Diffuser radiances to
                                              correct for all detectors
                                              in the band */
    const int *invalid_flags,             /* I: Data invalid flags for
                                              all detectors in the current
                                              band (derived from pixel
                                              mask information) */
    const double *cf_ptr,                  /* I: Pointer to correction
                                              factor data buffer */
    int start_detector_index,              /* I: Starting detector index */
    int num_detectors_sca                  /* I: Number of detectors in
                                              an SCA */
)
{
    int detector_index;             /* Detector index */

    /* NOTE:  If "adjustments" to the correction factors have to be done,
              they can be done here.  For now, assume that these factors
              are in their "proper" form */

    /* Sanity check.  We absolutely need a "valid" radiance data buffer.
       The pointer to the CPF correction factor data buffer can be
       NULL if no correction is desired; in that case, just return a
       SUCCESS status code */
    if (radiance_buffer == NULL)
    {
        IAS_LOG_ERROR("Invalid radiance data buffer detected");
        return ERROR;
    }
    else if (cf_ptr == NULL)
    {
        IAS_LOG_DEBUG("No correction for nonuniform response");
        return SUCCESS;
    }

    /* Apply the correction */
    for (detector_index = start_detector_index;
         detector_index < (start_detector_index + num_detectors_sca);
         detector_index++)
    {
        if (!invalid_flags[detector_index])
        {
            radiance_buffer[detector_index] /= (*cf_ptr);
        }
        cf_ptr++;
    }

    /* Done */
    return SUCCESS;
}
