/**************************************************************************
 NAME:          initialize_summary_report_data

 PURPOSE:       Module containing routines used to generate the information
                written to the optional summary report for nonuniformity
                characterization

***************************************************************************/
#include <string.h>

#include "ias_logging.h"
#include "ias_satellite_attributes.h"

#include "char_nonuniformity.h"


/*------------------------------------------------------------------------
 NAME:            initialize_nonuniformity_parameter_data

 PURPOSE:   Initializes WO / OMF parameter information for the summary
            report.  Called only if a summary report is to be
            generated.

 RETURNS:   Nothing
-------------------------------------------------------------------------*/
void initialize_nonuniformity_parameter_data
(
    RPS_NUCHAR_WO_PARAMS wo_params,       /* I: Work order parameters */
    RPS_NUCHAR_REPORT_SUMMARY *summary    /* O: Populated report summary
                                             data structure */
)
{

    /* Populate the data structure with work order parameter information */
    summary->rms_banding_flag         = wo_params.rms_banding_flag;
    summary->stdev_banding_flag       = wo_params.stdev_banding_flag;
    summary->streaking_flag           = wo_params.streaking_flag;
    summary->num_contiguous_detectors = wo_params.num_contiguous_detectors;
    summary->correct_flag             = wo_params.correct_flag;
    summary->calculate_radiance       = wo_params.calculate_radiance;
    summary->relgain_option           = wo_params.relgain_option;
    summary->store_char_in_db         = wo_params.store_char_in_db;
    summary->report_flag              = wo_params.report_flag;
    summary->dump_flag                = wo_params.dump_flag;

    /* Done */
}




/*------------------------------------------------------------------------
 NAME:            initialize_summary_report_band_data

 PURPOSE:   Copies selected bits of the current band's attribute
            information to the summary report data structure

 RETURNS:   Nothing
--------------------------------------------------------------------------*/
void initialize_summary_report_band_data
(
    RPS_NUCHAR_WO_PARAMS wo_params,              /* I: Contains band
                                                    nonuniformity limits */
    const IAS_BAND_ATTRIBUTES *band_attributes,  /* I: Attributes for
                                                    current band */
    RPS_NUCHAR_REPORT_SUMMARY *summary           /* I/O: Populated
                                                    summary report data
                                                    structure */
)
{

    /* Pull out the relevant band attributes */
    strcpy(summary->band_name, band_attributes->band_name);

    summary->sensor_id         = band_attributes->sensor_id;
    summary->band_number       = band_attributes->band_number;
    summary->spectral_type     = band_attributes->spectral_type;
    summary->scas              = band_attributes->scas;
    summary->detectors_per_sca = band_attributes->detectors_per_sca;

    /* Get nonuniformity threshold limits.  Default values are #defined in
       the main header file */
    if (band_attributes->spectral_type
            == IAS_SPECTRAL_PAN)               /* OLI PAN band */
    {
        summary->ffov_limit = RPS_NUCHAR_NONTHERMAL_PAN_FFOV_LIMIT;

        summary->detector_rms_banding_limit =
            RPS_NUCHAR_NONTHERMAL_PAN_RMS_LIMIT;

        summary->detector_stdev_banding_limit =
            RPS_NUCHAR_NONTHERMAL_PAN_STDEV_LIMIT;

        summary->detector_streaking_limit =
            RPS_NUCHAR_NONTHERMAL_PAN_STREAKING_LIMIT;
    }
    else if (band_attributes->spectral_type    /* TIRS bands */
                 == IAS_SPECTRAL_THERMAL)
    {
        summary->ffov_limit = RPS_NUCHAR_THERMAL_MS_FFOV_LIMIT;

        summary->detector_rms_banding_limit =
            RPS_NUCHAR_THERMAL_MS_RMS_LIMIT;

        summary->detector_stdev_banding_limit =
            RPS_NUCHAR_THERMAL_MS_STDEV_LIMIT;

        summary->detector_streaking_limit =
            RPS_NUCHAR_THERMAL_MS_STREAKING_LIMIT;
    }
    else                                       /* OLI VNIR/SWIR bands */
    {
        summary->ffov_limit = RPS_NUCHAR_NONTHERMAL_MS_FFOV_LIMIT;

        summary->detector_rms_banding_limit =
            RPS_NUCHAR_NONTHERMAL_MS_RMS_LIMIT;

        summary->detector_stdev_banding_limit =
            RPS_NUCHAR_NONTHERMAL_MS_STDEV_LIMIT;

        summary->detector_streaking_limit =
            RPS_NUCHAR_NONTHERMAL_MS_STREAKING_LIMIT;
    }

    /* Done */
}




/*------------------------------------------------------------------------
 NAME:                 initialize_sca_summary_metrics

 PURPOSE:      Calls the routines to calculate selected SCA summary
               metrics statistics and populate the summary report
               data structure

 RETURNS:      Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int initialize_sca_summary_metrics
(
    RPS_NUCHAR_METRICS metrics,          /* I: Metrics data structure */
    const RPS_NUCHAR_INVALID_FLAGS *invalid_flags,
                                         /* I: Characterization result
                                            invalid flags for all
                                            detectors in the current band */
    int rms_banding_flag,                /* I: Flag indicating whether
                                            RMS banding characterization is
                                            performed */
    int stdev_banding_flag,              /* I: Flag indicating whether
                                            stdev banding characterization
                                            is performed */
    int streaking_flag,                  /* I: Flag indicating whether
                                            streaking characterization
                                            is performed */
    int num_scas,                        /* I: Total # of SCAs in the
                                            current band */
    int num_detectors_per_sca,           /* I: Total # of detectors per
                                            SCA */
    RPS_NUCHAR_REPORT_SUMMARY *summary   /* O: Populated summary report
                                            data structure (also used for
                                            dump file) */
)
{
    int sca_index;
    int status;

    /* Get the metrics statistics for RMS banding characterization */
    if (rms_banding_flag)
    {
        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            int starting_detector_index =
                sca_index * num_detectors_per_sca;

            summary->sca_summary_rms[sca_index].sca_number = (sca_index + 1);
            status = summarize_within_sca_metrics(
                &(metrics.rms_metrics[sca_index]),
                &(invalid_flags->rms_banding[starting_detector_index]),
                num_detectors_per_sca,
                metrics.band_radiance_mean,
                &(summary->sca_summary_rms[sca_index]));
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Determining SCA summary metrics for "
                              "RMS banding characterization");
                return ERROR;
            }
        }

        /* Now cross-SCA summary metrics statistics */
        for (sca_index = 0; sca_index < num_scas - 1; sca_index++)
        {
            int starting_detector_index =
                sca_index * num_detectors_per_sca;

            summary->cross_sca_rms[sca_index].sca1_number = (sca_index + 1);
            summary->cross_sca_rms[sca_index].sca2_number = (sca_index + 2);
            status = summarize_cross_sca_metrics(metrics.rms_metrics,
                &(invalid_flags->rms_banding[starting_detector_index]),
                sca_index,
                num_detectors_per_sca,
                RPS_NUCHAR_NUM_SCA_BOUNDARY_DETECTORS,
                "STDEV",
                metrics.band_radiance_mean,
                &(summary->cross_sca_rms[sca_index]));
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Determining cross-SCA summary metrics for "
                              "RMS banding characterization");
                return ERROR;
            }
        }
    }

    /* Get the metrics statistics for stdev banding characterization */
    if (stdev_banding_flag)
    {
        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            int starting_detector_index =
                sca_index * num_detectors_per_sca;

            summary->sca_summary_stdev[sca_index].sca_number = 
                (sca_index + 1);
            status = summarize_within_sca_metrics(
                &(metrics.stdev_metrics[sca_index]),
                &(invalid_flags->stdev_banding[starting_detector_index]),
                num_detectors_per_sca,
                metrics.band_radiance_mean,
                &(summary->sca_summary_stdev[sca_index]));
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Determining SCA summary metrics for "
                              "stdev banding characterization");
                return ERROR;
            }
        }

        /* Now cross-SCA summary metrics statistics */
        for (sca_index = 0; sca_index < num_scas - 1; sca_index++)
        {
            int starting_detector_index =
                sca_index * num_detectors_per_sca;

            summary->cross_sca_stdev[sca_index].sca1_number =
                (sca_index + 1);
            summary->cross_sca_stdev[sca_index].sca2_number =
                (sca_index + 2);
            status = summarize_cross_sca_metrics(metrics.stdev_metrics,
                &(invalid_flags->stdev_banding[starting_detector_index]),
                sca_index,
                num_detectors_per_sca,
                RPS_NUCHAR_NUM_SCA_BOUNDARY_DETECTORS,
                "MEAN",
                metrics.band_radiance_mean,
                &(summary->cross_sca_stdev[sca_index]));
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Determining cross-SCA summary metrics for "
                              "stdev banding characterization");
                return ERROR;
            }
        }
    }

    /* Get the metrics statistics for streaking characterization */
    if (streaking_flag)
    {
        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            int starting_detector_index =
                sca_index * num_detectors_per_sca;

            summary->sca_summary_streaking[sca_index].sca_number
                = (sca_index + 1);
            status = summarize_within_sca_metrics(
                &(metrics.streaking_metrics[sca_index]),
                &(invalid_flags->streaking[starting_detector_index]),
                num_detectors_per_sca,
                metrics.band_radiance_mean,
                &(summary->sca_summary_streaking[sca_index]));
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Determining SCA summary metrics for "
                              "streaking characterization");
                return ERROR;
            }
        }
    }

    /* Done */
    return SUCCESS;
}
