/**************************************************************************
 NAME:                      nonuniformity_utility

 PURPOSE:       Module implementing "utility" functions used by the main
                char_nonuniformity application and/or its supporting
                routines

***************************************************************************/
#include <stdlib.h>
#include <string.h>

#include "ias_cpf.h"
#include "ias_db.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_radiometry.h"
#include "ias_pixel_mask.h"

#include "char_nonuniformity.h"

/*------------------------------------------------------------------------
 NAME:     allocate_nonuniformity_invalid_flag_buffers

 PURPOSE:  Allocates data buffers to contain nonuniformity characterization
           data/results invalid flags

 RETURNS:  Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int allocate_nonuniformity_invalid_flag_buffers
(
    int num_detectors,                /* I: Number of detectors */
    RPS_NUCHAR_INVALID_FLAGS *flags   /* O: Invalid flag data structure */
)
{
    /* Allocate each data buffer */
    flags->statistics = calloc(num_detectors, sizeof(int));
    flags->rms_banding = calloc(num_detectors, sizeof(int));
    flags->stdev_banding = calloc(num_detectors, sizeof(int));
    flags->streaking = calloc(num_detectors, sizeof(int));

    if ((flags->statistics == NULL)
        || (flags->rms_banding == NULL)
        || (flags->stdev_banding == NULL)
        || (flags->streaking == NULL))
    {
        IAS_LOG_ERROR("Allocating invalid flag data buffer(s)");
        free(flags->statistics);
        free(flags->rms_banding);
        free(flags->stdev_banding);
        free(flags->streaking);
        return ERROR;
    }

    return SUCCESS;
}


/*-------------------------------------------------------------------------
 NAME:       free_nonuniformity_invalid_flag_buffers

 PURPOSE:    Frees data/results invalid flags allocated by
             allocate_nonuniformity_invalid_flags

 RETURNS:    Nothing
---------------------------------------------------------------------------*/
void free_nonuniformity_invalid_flag_buffers
(
    RPS_NUCHAR_INVALID_FLAGS *flags     /* I: Data structure with buffers
                                           to free */
)
{
    free(flags->statistics);
    free(flags->rms_banding);
    free(flags->stdev_banding);
    free(flags->streaking);
}



/*------------------------------------------------------------------------
 NAME:                 cleanup_for_next_band

 PURPOSE:  Frees allocated data buffers and closes the summary report/dump
           files after processing of the current band is completed

 RETURNS:  Nothing
--------------------------------------------------------------------------*/
void cleanup_for_next_band
(
    int num_scas,                   /* I: Number of SCAs in the
                                       current band */
    RPS_NUCHAR_REPORT_SUMMARY *rpt_summary,/* I: Summary report data */
    RPS_NUCHAR_METRICS *metrics,    /* I: Nonuniformity metrics data */
    RPS_NUCHAR_INVALID_FLAGS *invalid_flags,
                                    /* I: Data structure containing
                                       invalid flag buffers to free */
    double *local_rms_buffer,       /* I: Band-level RMS metrics buffer */
    double *local_stdev_buffer,     /* I: Band-level stdev metrics buffer */
    double *local_streaking_buffer, /* I: Band-level streaking metrics
                                       buffer */
    double *radiance_buffer,        /* I: Band-level radiance_buffer */
    double *stats_buffer            /* I: Histogram statistics for the
                                       current band/SCA (MS bands, lineset=2),
                                       and PAN band 'even' lines (lineset=0) */
)
{
    int sca_index;


    IAS_LOG_DEBUG("Freeing summary report data structures...");
    free(rpt_summary->sca_summary_rms);
    free(rpt_summary->sca_summary_stdev);
    free(rpt_summary->sca_summary_streaking);
    free(rpt_summary->cross_sca_rms);
    free(rpt_summary->cross_sca_stdev);

    /* Free the internal metrics buffers if they haven't been freed yet,
       then free the base detector metrics arrays */
    IAS_LOG_DEBUG("Freeing internal metrics data buffers...");
    if (metrics->rms_metrics != NULL)
    {
        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            free(metrics->rms_metrics[sca_index].detector_number);
            free(metrics->rms_metrics[sca_index].metric);
        }
        free(metrics->rms_metrics);
    }
    if (metrics->stdev_metrics != NULL)
    {
        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            free(metrics->stdev_metrics[sca_index].detector_number);
            free(metrics->stdev_metrics[sca_index].metric);
        }
        free(metrics->stdev_metrics);
    }
    if (metrics->streaking_metrics != NULL)
    {
        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            free(metrics->streaking_metrics[sca_index].detector_number);
            free(metrics->streaking_metrics[sca_index].metric);
        }
        free(metrics->streaking_metrics);
    }


    /* Free the local band-wide metrics working buffers */
    IAS_LOG_DEBUG("Freeing band metrics data buffers...");
    free(local_rms_buffer);
    free(local_stdev_buffer);
    free(local_streaking_buffer);

    /* Free the radiance and histogram statistics buffers */
    IAS_LOG_DEBUG("Freeing radiance and histogram statistics data "
            "buffers...");
    free(radiance_buffer);
    free(stats_buffer);

    /* Free the invalid flag data buffers */
    free_nonuniformity_invalid_flag_buffers(invalid_flags);

    /* Done */
}




/*------------------------------------------------------------------------
 NAME:              cleanup_application

 PURPOSE:     Calls all system/application functions freeing allocated
              resources, depending on the current exit status. On exit due
              to a fatal error, all database transactions are rolled back. 
              On exit due to successful completion of processing, all
              database transactions are committed.  For either exit, the CPF
              memory block is freed, open summary report and/or dump files
              are closed, the database connection is closed, and allocated
              data buffers are freed

 RETURNS:     Nothing
--------------------------------------------------------------------------*/
void cleanup_application
(
    int store_char_in_db,                  /* I: DB trend flag */
    int num_scas,                          /* I: Number of SCAs in the
                                              current band */
    RPS_NUCHAR_REPORT_SUMMARY *rpt_summary,/* I: Summary report data */
    RPS_NUCHAR_METRICS *metrics,           /* I: Nonuniformity metrics data */
    RPS_NUCHAR_INVALID_FLAGS *invalid_flags,
                                           /* I: invalid flag data
                                              structure with flag data
                                              buffers to free */
    struct ias_db_connection *db,   /* I: Open(?) database connection */
    double *local_rms_buffer,       /* I: Band-level RMS metrics buffer */
    double *local_stdev_buffer,     /* I: Band-level stdev metrics buffer */
    double *local_streaking_buffer, /* I: Band-level streaking metrics
                                       buffer */
    double *radiance_buffer,        /* I: Band-level radiance_buffer */
    double *stats_buffer,           /* I: Histogram statistics for the
                                       current band/SCA (MS bands, lineset=2),
                                       and PAN band 'even' lines (lineset=0) */
    IAS_CPF *cpf,                   /* I: CPF data structure */
    IAS_PIXEL_MASK **pmasks,        /* I: Pixel mask array */
    int num_masks_read              /* I: Number of pixel masks read from the
                                       pixel mask file */
)
{
    IAS_PIXEL_MASK *pm_ptr = NULL;         /* Pointer to individual
                                              pixel mask */

    int mask_index;            /* Pixel mask loop counter */
    int sca_index;             /* SCA loop counter */
    int status;                /* Individual function return status */


    /* If trending was specified, rollback all previous writes when
       shutting the application down due to an error, and commit all
       transactions when the application has successfully completed
       all processing. */
    if (db != NULL)
    {
        IAS_LOG_DEBUG("Active database connection...");
        if (store_char_in_db)
        {
            IAS_LOG_DEBUG("Rolling back DB transactions...");
            status = ias_db_rollback_transaction(db);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Rolling back DB transactions");
            }
        }

        /* Close the connection */
        ias_db_close_connection(db);
        ias_db_close_database_lib();
    }
    else
        IAS_LOG_DEBUG("No database connection open...");

    /* Pixel mask cleanup */
    if (pmasks != NULL)
    {
        for (mask_index = 0; mask_index < num_masks_read; mask_index++)
        {
            pm_ptr = pmasks[mask_index];
            ias_pm_destroy(pm_ptr);
        }
        free(pmasks);
    }

    /* CPF cleanup */
    IAS_LOG_DEBUG("Freeing CPF resources...");
    ias_cpf_free(cpf);


    IAS_LOG_DEBUG("Freeing summary report data structures...");
    free(rpt_summary->sca_summary_rms);
    free(rpt_summary->sca_summary_stdev);
    free(rpt_summary->sca_summary_streaking);
    free(rpt_summary->cross_sca_rms);
    free(rpt_summary->cross_sca_stdev);

    /* Free the report header data buffer */
    if (rpt_summary->header.common_header != NULL)
    {
        IAS_LOG_DEBUG("Freeing report header resources...");
        ias_misc_free_report_header(rpt_summary->header.common_header);
    }

    /* Free the internal metrics buffers if they haven't been freed yet,
       then free the base detector metrics arrays */
    IAS_LOG_DEBUG("Freeing internal metrics data buffers...");

    if (metrics->rms_metrics != NULL)
    {
        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            free(metrics->rms_metrics[sca_index].detector_number);
            free(metrics->rms_metrics[sca_index].metric);
        }
        free(metrics->rms_metrics);
    }

    if (metrics->stdev_metrics != NULL)
    {
        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            free(metrics->stdev_metrics[sca_index].detector_number);
            free(metrics->stdev_metrics[sca_index].metric);
        }
        free(metrics->stdev_metrics);
    }

    if (metrics->streaking_metrics != NULL)
    {
        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            free(metrics->streaking_metrics[sca_index].detector_number);
            free(metrics->streaking_metrics[sca_index].metric);
        }
        free(metrics->streaking_metrics);
    }

    /* Free the local band-wide metrics working buffers */
    IAS_LOG_DEBUG("Freeing band metrics data buffers...");
    free(local_rms_buffer);
    free(local_stdev_buffer);
    free(local_streaking_buffer);

    /* Free the radiance and histogram statistics buffers */
    IAS_LOG_DEBUG("Freeing radiance and histogram statistics data "
            "buffers...");
    free(radiance_buffer);
    free(stats_buffer);

    /* Free the invalid flag data buffers */
    free_nonuniformity_invalid_flag_buffers(invalid_flags);

    IAS_LOG_DEBUG("Resources successfully freed.");

    /* Done */

}




/*------------------------------------------------------------------------
 NAME:                 allocate_summary_report_buffers

 PURPOSE:     Allocates the out-of-spec detector and sca summary arrays in
              the report data structure, and initializes the contents to 0.

 RETURNS:     Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int allocate_summary_report_buffers
(
    int rms_banding_flag,                    /* I: RMS banding
                                                characterization flag */
    int stdev_banding_flag,                  /* I: Stdev banding
                                                characterization flag */
    int streaking_flag,                      /* I: Streaking
                                                characterization flag */
    int num_scas,                            /* I: # of SCAs in the current
                                                band */
    RPS_NUCHAR_REPORT_SUMMARY *rpt_summary   /* O: Summary info data
                                                structure */
)
{

    /* Allocate the internal arrays depending on the characterization(s)
       selected */
    if (rms_banding_flag)
    {
        /* Allocate the within-SCA summary array */
        rpt_summary->sca_summary_rms = calloc(num_scas, sizeof(SCA_SUMMARY));

        if (rpt_summary->sca_summary_rms == NULL)
        {
            IAS_LOG_ERROR("Cannot allocate RMS SCA summary data "
                    "array");
            return ERROR;
        }

        /* Allocate the cross-SCA summary array */
        rpt_summary->cross_sca_rms = calloc((num_scas - 1),
                sizeof(CROSS_SCA_SUMMARY));

        if (rpt_summary->cross_sca_rms == NULL)
        {
            IAS_LOG_ERROR("Cannot allocate RMS cross-SCA summary data "
                    "array");
            return ERROR;
        }
    }

    if (stdev_banding_flag)
    {
        /* Allocate the within-SCA summary array */
        rpt_summary->sca_summary_stdev = calloc(num_scas,
                sizeof(SCA_SUMMARY));

        if (rpt_summary->sca_summary_stdev == NULL)
        {
            IAS_LOG_ERROR("Cannot allocate stdev SCA summary data "
                    "array");
            return ERROR;
        }

        /* Allocate the cross-SCA summary array */
        rpt_summary->cross_sca_stdev = calloc((num_scas - 1),
                sizeof(CROSS_SCA_SUMMARY));

        if (rpt_summary->cross_sca_stdev == NULL)
        {
            IAS_LOG_ERROR("Cannot allocate stdev cross-SCA summary data "
                    "array");
            return ERROR;
        }
    }

    if (streaking_flag)
    {
        /* Allocate the within-SCA summary array */
        rpt_summary->sca_summary_streaking = calloc(num_scas,
                sizeof(SCA_SUMMARY));

        if (rpt_summary->sca_summary_streaking == NULL)
        {
            IAS_LOG_ERROR("Cannot allocate streaking SCA summary data "
                    "array");
            return ERROR;
        }
    }

    /* All the requested allocations were successful */
    return SUCCESS;
}




/*------------------------------------------------------------------------
 NAME:                  allocate_metrics_buffers

 PURPOSE:  Allocates the internal metrics data buffers for the specified
           characterization(s).  The buffers are large enough for the data
           in an individual SCA

 RETURNS:  Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int allocate_metrics_buffers
(
    int rms_banding_flag,              /* I: Flag indicating whether RMS
                                          banding characterization is
                                          performed */
    int stdev_banding_flag,            /* I: Flag indicating whether
                                          stdev banding characterization
                                          is performed */
    int streaking_flag,                /* I: Flag indicating whether
                                          streaking characterization is
                                          performed */
    int num_scas,                      /* I: # of SCAs in the current
                                          band */
    int num_detectors_per_sca,         /* I: # of detectors in an SCA */
    RPS_NUCHAR_METRICS *metrics        /* O: Metrics data structure with
                                          newly allocated data buffers*/
)
{
    int sca_index;                     /* # of SCAs in the band */


    /* Allocate the data buffers for the specified characterization(s) */
    if (rms_banding_flag)
    {
        metrics->rms_metrics = calloc(num_scas, sizeof(METRICS_DET));

        if (metrics->rms_metrics == NULL)
        {
            IAS_LOG_ERROR("Cannot allocate RMS metrics data buffers");
            return ERROR;
        }

        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            /* Fill in the SCA number while we're here */
            metrics->rms_metrics[sca_index].sca_number = (sca_index + 1);

            metrics->rms_metrics[sca_index].detector_number = calloc(
                    num_detectors_per_sca,
                    sizeof(*(metrics->rms_metrics[sca_index]
                            .detector_number)));

            metrics->rms_metrics[sca_index].metric = calloc(
                    num_detectors_per_sca,
                    sizeof(*(metrics->rms_metrics[sca_index]
                            .metric)));

            if ((metrics->rms_metrics[sca_index].detector_number == NULL)
                    || (metrics->rms_metrics[sca_index].metric == NULL))
            {
                IAS_LOG_ERROR("Cannot allocate data buffers for SCA %d",
                        (sca_index + 1));

                return ERROR;
            }
        }
    }

    if (stdev_banding_flag)
    {
        metrics->stdev_metrics = calloc(num_scas, sizeof(METRICS_DET));
        if (metrics->stdev_metrics == NULL)
        {
            IAS_LOG_ERROR("Cannot allocate stdev metrics data buffers");
            return ERROR;
        }

        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            metrics->stdev_metrics[sca_index].sca_number = (sca_index + 1);

            metrics->stdev_metrics[sca_index].detector_number = calloc(
                        num_detectors_per_sca,
                        sizeof(*(metrics->stdev_metrics[sca_index]
                                .detector_number)));

            metrics->stdev_metrics[sca_index].metric = calloc(
                        num_detectors_per_sca,
                        sizeof(*(metrics->stdev_metrics[sca_index]
                                .metric)));

            if ((metrics->stdev_metrics[sca_index].detector_number == NULL)
                    || (metrics->stdev_metrics[sca_index].metric == NULL))
            {
                IAS_LOG_ERROR("Cannot allocate data buffers for SCA %d",
                        (sca_index + 1));

                return ERROR;
            }
        }
    }

    if (streaking_flag)
    {
        metrics->streaking_metrics = calloc(num_scas, sizeof(METRICS_DET));

        if (metrics->streaking_metrics == NULL)
        {
            IAS_LOG_ERROR("Cannot allocate RMS metrics data buffers");
            return ERROR;
        }

        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            metrics->streaking_metrics[sca_index]
            .sca_number = (sca_index + 1);

            metrics->streaking_metrics[sca_index].detector_number = calloc(
                  num_detectors_per_sca,
                  sizeof(*(metrics->streaking_metrics[sca_index]
                                  .detector_number)));

            metrics->streaking_metrics[sca_index].metric = calloc(
                  num_detectors_per_sca,
                  sizeof(*(metrics->streaking_metrics[sca_index]
                         .metric)));

            if ((metrics->streaking_metrics[sca_index]
                         .detector_number == NULL)
                || (metrics->streaking_metrics[sca_index].metric == NULL))
            {
                IAS_LOG_ERROR("Cannot allocate data buffers for SCA %d",
                              (sca_index + 1));

                return ERROR;
            }
        }
    }

    /* Done */
    return SUCCESS;
}




/*------------------------------------------------------------------------
 NAME:                  populate_metrics_buffer

 PURPOSE:   Copies the metrics data from the band-wide metrics data buffer
            to the corresponding internal metrics data buffer

 RETURNS:   Integer status code of SUCCESS or ERROR
-------------------------------------------------------------------------*/
void populate_metrics_buffers
(
    const double *band_rms_buffer,  /* I: RMS metrics data buffer for all
                                       detectors in the current band */
    const double *band_stdev_buffer,/* I: Stdev metrics data buffer for
                                       all detectors in the current band */
    const double *band_streaking_buffer, /* I: Streaking metrics data buffer
                                            for all detectors in the current
                                            band */
    int rms_banding_flag,           /* I: Flag indicating whether RMS
                                       banding characterization is
                                       performed */
    int stdev_banding_flag,         /* I: Flag indicating whether stdev
                                       banding characterization is
                                       performed */
    int streaking_flag,             /* I: Flag indicating whether
                                       streaking characterization is
                                       performed */
    int sca_number,                 /* I: Current 1-based SCA number */
    int num_detectors_per_sca,      /* I: Number of detectors in each
                                       SCA */
    RPS_NUCHAR_METRICS *metrics     /* O: Populated valid metrics data
                                       structure for the current SCA */
)
{
    int    sca_index;
    int    start_detector_index;     /* Starting detector index in
                                        band-wide data buffer corresponding
                                        to current SCA */
    int    detector_index;           /* 0-based index for current
                                        detector */


    /* Initialization  */
    sca_index = (sca_number - 1);
    start_detector_index = sca_index * num_detectors_per_sca;

    /* Each specified characterization will have a NON-NULL data buffer.
       Populate those */
    if (rms_banding_flag)
    {
        for (detector_index = 0;
             detector_index < num_detectors_per_sca;
             detector_index++)
        {
            metrics->rms_metrics[sca_index].detector_number[detector_index]
                = (detector_index + 1);
            metrics->rms_metrics[sca_index].metric[detector_index]
                = band_rms_buffer[start_detector_index + detector_index];
        }
    }

    if (stdev_banding_flag)
    {
        for (detector_index = 0;
             detector_index < num_detectors_per_sca;
             detector_index++)
        {
            metrics->stdev_metrics[sca_index].detector_number[detector_index]
               = (detector_index + 1);
            metrics->stdev_metrics[sca_index].metric[detector_index]
               = band_stdev_buffer[(start_detector_index
                                        + detector_index)];
        }
    }

    if (streaking_flag)
    {
        for (detector_index = 0;
             detector_index < num_detectors_per_sca;
             detector_index++)
        {
            metrics->streaking_metrics[sca_index]
                    .detector_number[detector_index] = (detector_index + 1);
            metrics->streaking_metrics[sca_index].metric[detector_index]
                = band_streaking_buffer[(start_detector_index
                                             + detector_index)];
        }
    }
}




/*-----------------------------------------------------------------------
 NAME:              calculate_pcnt_metric

 PURPOSE:  Routine to convert a raw metric value into a % metric value
           relative to the mean radiance of the current band

 RETURNS:  Nothing
-------------------------------------------------------------------------*/
double calculate_pcnt_metric
(
    double raw_metric,              /* I: Raw metric to convert */
    double band_radiance            /* I: Mean radiance for the
                                          current band */
)
{
    return ((raw_metric / band_radiance) * 100.0);
}




/*------------------------------------------------------------------------
 NAME:            write_report_header_to_file

 PURPOSE:  Writes "header" information to a header file

 RETURNS:  Integer status code of SUCCESS or ERROR

--------------------------------------------------------------------------*/
int write_report_header_to_file
(
    RPS_NUCHAR_REPORT_SUMMARY summary,    /* I: Summary report information */
    const char *work_order_id             /* I: Work order ID currently
                                             processing */ 
)
{
    FILE *header_fptr = NULL;             /* Open header file */

    char header_name[PATH_MAX] = "";      /* Name of header file */
    char *collection_type = NULL;         /* OLI/TIRS collection type */
    char *cpf_name = NULL;
    char *sensor_id = NULL;               /* OLI/TIRS/L8_OLITIRS */

    int status = SUCCESS;


    /* Build the name of the header file */
    sprintf(header_name, "nonuniformity_header.txt");

    /* Open the header file */
    header_fptr = fopen(header_name, "w");
    if (header_fptr == NULL)
    {
        IAS_LOG_ERROR("Creating header information file %s",
            header_name);
        return ERROR;
    }

    /* Print the "standard" IAS report header information */
    status = ias_misc_write_report_header(header_fptr,
        summary.header.common_header);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing standard IAS header information to "
            "header file %s", header_name);
        return ERROR;
    }

    /* Print the RPS-specific report header information */
    collection_type = ias_misc_get_header_collection_type(
        summary.header.common_header);
    cpf_name = ias_misc_get_header_cpf_name(summary.header.common_header);
    sensor_id = ias_misc_get_header_sensor_id(summary.header.common_header);

    status = ias_misc_write_rps_report_header(collection_type, cpf_name,
        sensor_id, summary.header.rps_header, header_fptr);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing RPS-specific header information to "
            "header file %s", header_name);
        return ERROR;
    }

    /* If this is a solar collect, print the diffuser type */
    if ((strcmp(collection_type, IAS_RPS_OLI_SOLAR_COLLECT) == 0)
        || (strcmp(collection_type, IAS_RPS_OLI_SOLAR_TIMESWEEP_COLLECT)
            == 0))
    {
        if (summary.header.panel_id == IAS_RPS_OLI_DIFFUSER_TYPE_PRISTINE)
        {
            status = fprintf(header_fptr, "Diffuser Panel:              %s"
                "\n\n\n", IAS_RPS_OLI_DIFFUSER_PRISTINE);
        }
        else
        {
            status = fprintf(header_fptr, "Diffuser Panel:              %s"
                "\n\n\n", IAS_RPS_OLI_DIFFUSER_WORKING);
        }
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing diffuser panel information to header "
                "file %s", header_name);
            return ERROR;
        }
    }

    /* Print sub-information on which set(s) of absolute and relative gains
       were used to calculate radiances, whether radiances were corrected
       for panel nonuniformities, which characterizations were
       selected, etc. */
    if (fprintf(header_fptr, "SELECTED CHARACTERIZATION OPTIONS\n\n") < 0)
    {
        IAS_LOG_ERROR("Writing characterization option header");
        return ERROR;
    }

    /* If radiance source is IAS, print which relative gains were
       used */
    if (summary.calculate_radiance == RPS_NUCHAR_USE_DB_RADIANCE)
    {
        status = fprintf(header_fptr, "Radiance Source:             Radiance "
            "Histogram Statistics\n");
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing radiance info to summary report");
            return ERROR;
        }
    }
    else
    {
        status = fprintf(header_fptr, "Radiance Source:             Bias-"
            "Corrected/Linearized Histogram Statistics\n");
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing radiance info to summary report");
            return ERROR;
        }
        switch (summary.relgain_option)
        {
            case RPS_NUCHAR_PRELAUNCH_RELGAINS:
                status = fprintf(header_fptr, "Relative Gain Source:        "
                "Prelaunch\n");
                break;
            case RPS_NUCHAR_INVALID_RELGAINS:
                status = fprintf(header_fptr, "Relative Gain Source:        "
                "Postlaunch\n");
                break;
            case RPS_NUCHAR_CURRENT_RELGAINS:
                status = fprintf(header_fptr, "Relative Gain Source:        "
                "Current\n");
                break;
            default:
                IAS_LOG_ERROR("Invalid relative gain specification");
                return ERROR;
        }
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing relative gain information to report "
                "file");
            return ERROR;
        }
    }

    /* Print whether panel/blackbody source nonuniform response
       correction was performed */
    if (summary.correct_flag)
    {
        status = fprintf(header_fptr, "Apply Source Correction:     Yes\n");
    }
    else
    {
        status = fprintf(header_fptr, "Apply Source Correction:     No\n");
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing source correction option information");
        return ERROR;
    }

    /* Print the selected characterizations */
    if (summary.rms_banding_flag == RPS_NUCHAR_RMS_BANDING)
    {
        status = fprintf(header_fptr, "Char RMS Banding:            Yes\n");
    }
    else
    {
        status = fprintf(header_fptr, "Char RMS Banding:            No\n");
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing selected characterization information");
        return ERROR;
    }

    if (summary.stdev_banding_flag == RPS_NUCHAR_STDEV_BANDING)
    {
        status = fprintf(header_fptr, "Char STDEV Banding:          Yes\n");
    }
    else
    {
        status = fprintf(header_fptr, "Char STDEV Banding:          No\n");
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing selected characterization information");
        return ERROR;
    }

    if (summary.streaking_flag == RPS_NUCHAR_STREAKING)
    {
        status = fprintf(header_fptr, "Char Streaking:              Yes\n");
    }
    else
    {
        status = fprintf(header_fptr, "Char Streaking:              No\n\n");
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing selected characterization information");
        return ERROR;
    }

    /* Wrap up the file */
    status = fprintf(header_fptr, "END HEADER INFORMATION\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Completing header information file %s",
            header_name);
        return ERROR;
    }

    /* Close the file */
    fclose(header_fptr);

    /* Done */
    return SUCCESS;
}



