/***************************************************************************
 NAME:                    read_parameters

  PURPOSE:  Reads the input parameter files and populates the
            required parameter data structures

  RETURNS:  Integer status code of SUCCESS or ERROR

***************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "ias_miscellaneous.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "ias_radiometry.h"
#include "ias_satellite_attributes.h"

#include "char_nonuniformity.h"

#define MAX_NUM_WO_PARAMS 27
#define MAX_NUM_OMF_PARAMS 4


int read_parameters
(
    const char *wo_params_filename,         /* I: Work order filename */
    RPS_NUCHAR_WO_PARAMS  *wo_param_data,   /* O: Work order parameters */
    RPS_NUCHAR_OMF_PARAMS *omf_param_data   /* O: OMF parameters */
)
{
    char omf_params_filename[PATH_MAX];    /* Local OMF parameter file name */

    /* Default string values for  the filename prefix and the RPS
       processing level. */
    const char *default_report_filename_prefix[] = {"nonuniformity"};
    const char *default_oli_rps_proc_level[] =
        {RPS_NUCHAR_OLI_PROC_LEVEL_TEMPCORRECT};
    const char *default_tirs_rps_proc_level[] =
        {RPS_NUCHAR_TIRS_PROC_LEVEL};

    /* Allow selection of pre-, post-, or current relative gains from the
       work order. */
    int default_relgain_option = RPS_NUCHAR_CURRENT_RELGAINS;

    /* Other processing options.  */
    int default_calculate_radiance = RPS_NUCHAR_CALCULATE_RADIANCE;
    int default_apply_correction   = RPS_NUCHAR_NO_CORRECT;

    int default_num_contiguous_detectors = 
        RPS_NUCHAR_NUM_CONTIGUOUS_DETECTORS;

    int default_char_rms_banding_flag = RPS_NUCHAR_RMS_BANDING;
    int default_char_stdev_banding_flag = RPS_NUCHAR_STDEV_BANDING;
    int default_char_streaking_flag = RPS_NUCHAR_STREAKING;

    /* Output options relating to summary reports and/or dump file
       reports.  */
    int default_report_flag = RPS_NUCHAR_NO_REPORT;
    int default_dump_flag   = RPS_NUCHAR_NO_DUMP;

    /* Internal variables. */
    int status;                      /* Status for each read
                                        operation */
    int read_count;                  /* Count of parameters
                                        successfully read */

/*-------------------------------------------------------------------------
                      READ WORK ORDER PARAMETER FILE
---------------------------------------------------------------------------*/
    /* Declare the work order parameter table */
    IAS_PARM_DECLARE_TABLE(wo_param_table, MAX_NUM_WO_PARAMS);

    /* Set up the WO parameter table for reading */

    /* Common work order parameters first */
    IAS_PARM_BAND_LIST(wo_param_table, IAS_PARM_REQUIRED,
            wo_param_data->bands,
            sizeof(wo_param_data->bands), 1);

    IAS_PARM_WORK_ORDER_ID(wo_param_table, wo_param_data->work_order_id,
            sizeof(wo_param_data->work_order_id), 1);

    IAS_PARM_STORE_CHARACTERIZATION_IN_DB(wo_param_table, IAS_PARM_OPTIONAL,
            &(wo_param_data->store_char_in_db),
            sizeof(wo_param_data->store_char_in_db), 1);

    /* Get the RPS processing levels for OLI and TIRS */
    IAS_PARM_OLI_RPS_PROC_LEVEL_DN(wo_param_table, IAS_PARM_OPTIONAL,
            default_oli_rps_proc_level, wo_param_data->oli_rps_proc_level,
            sizeof(wo_param_data->oli_rps_proc_level));

    IAS_PARM_TIRS_RPS_PROC_LEVEL_DN(wo_param_table, IAS_PARM_OPTIONAL,
            default_tirs_rps_proc_level, wo_param_data->tirs_rps_proc_level,
            sizeof(wo_param_data->tirs_rps_proc_level));

    /* Characterization options */
    IAS_PARM_ADD_BOOLEAN(wo_param_table, RMS_BANDING,
            "Perform RMS banding characterization "
            "(1 = Perform characterization, "
            "0 = Do not perform characterization)",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
            &default_char_rms_banding_flag,
            &(wo_param_data->rms_banding_flag),
            sizeof(wo_param_data->rms_banding_flag), 1);

    IAS_PARM_ADD_BOOLEAN(wo_param_table, STDEV_BANDING,
            "Perform stdev banding characterization (1 = Perform "
            "characterization, 0 = Do not perform characterization)",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
            &default_char_stdev_banding_flag,
            &(wo_param_data->stdev_banding_flag),
            sizeof(wo_param_data->stdev_banding_flag), 1);

    IAS_PARM_ADD_BOOLEAN(wo_param_table, STREAKING,
            "Perform streaking characterization (1 = Perform "
            "characterization, 0 = Do not perform characterization)",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
            &default_char_streaking_flag,
            &(wo_param_data->streaking_flag),
            sizeof(wo_param_data->streaking_flag), 1);

    /* General processing options */
    IAS_PARM_ADD_BOOLEAN(wo_param_table, CALCULATE_RADIANCE,
         "Calculate radiances directly or read from database",
         IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
         &default_calculate_radiance,
         &(wo_param_data->calculate_radiance),
         sizeof(wo_param_data->calculate_radiance), 1);

    IAS_PARM_ADD_INT(wo_param_table, RELGAIN_OPTIONS,
            "Relative gain set options (prelaunch, postlaunch, current)",
            IAS_PARM_REQUIRED, IAS_PARM_NOT_ARRAY,
            1, RPS_NUCHAR_PRELAUNCH_RELGAINS,
            RPS_NUCHAR_CURRENT_RELGAINS, 1,
            &default_relgain_option,
            &(wo_param_data->relgain_option),
            sizeof(wo_param_data->relgain_option), 1);

    IAS_PARM_ADD_BOOLEAN(wo_param_table, CORRECT_SOURCE_NONUNIFORMITY,
            "Correct diffuser panel/blackbody nonuniformities using the CPF "
            "Solar_Diffuser_Scale/Blackbody_Scale parameters",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
            &default_apply_correction,
            &(wo_param_data->correct_flag),
            sizeof(wo_param_data->correct_flag), 1);

    IAS_PARM_ADD_INT(wo_param_table, NUM_CONTIGUOUS_DETECTORS,
            "Number of contiguous detectors allowed in "
            "banding characterizations", IAS_PARM_OPTIONAL,
            IAS_PARM_NOT_ARRAY, 1, 1, 200, 1,
            &default_num_contiguous_detectors,
            &(wo_param_data->num_contiguous_detectors),
            sizeof(wo_param_data->num_contiguous_detectors), 1);

    /* Output options */
    IAS_PARM_ADD_STRING(wo_param_table, REPORT_FILENAME_PREFIX, 
        "Output report file name prefix", IAS_PARM_OPTIONAL,
        0, NULL, /* no restrictions */
        1, default_report_filename_prefix, /* Default file name */
        wo_param_data->report_filename_prefix, 
        sizeof(wo_param_data->report_filename_prefix), 0);

    IAS_PARM_ADD_BOOLEAN(wo_param_table, REPORT_FLAG,
            "Write characterization results to summary reports "
            "(1 = Write summary reports, 0 = Do not write summary reports)",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
            &default_report_flag,
            &(wo_param_data->report_flag),
            sizeof(wo_param_data->report_flag), 1);

    IAS_PARM_ADD_BOOLEAN(wo_param_table, DUMP_FLAG,
            "Write characterization results to dump files "
            "(1 = Write dump files, 0 = Do not write dump files)",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
            &default_dump_flag,
            &(wo_param_data->dump_flag),
            sizeof(wo_param_data->dump_flag), 1);

    /* Get the total number of items to read from the parameter table*/
    read_count = IAS_PARM_GET_TABLE_SIZE(wo_param_table);

    /* If someone enters -- help, --template, etc, on the command line
       when starting the application, the appropriate information will
       be displayed.  If someone enters a file name, its access will be
       checked.  If a help option was specified or the specified file name
       could not be accessed, a 1 will be returned.  In this case, just
       copy the work order file name (or help option) to the OMF
       file name and continue.  Otherwise, read the work order file */
    status = ias_parm_provide_help(wo_params_filename, wo_param_table,
                read_count, IAS_INPUT_PARAMETERS);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Work order parameter help failed");
        return ERROR;
    }
    else if (status == 1)
    {
        sprintf(omf_params_filename, "%s", wo_params_filename);
    }
    else
    {
        status = ias_parm_read(wo_params_filename, wo_param_table,
                read_count);

        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading ODL parameter file");
            return ERROR;
        }

        /* Get the total number of bands to process */
        wo_param_data->num_bands_to_process = IAS_PARM_GET_COUNT_READ(
            wo_param_table, BAND_LIST);

        /* Construct the OMF file name */
        status = snprintf(omf_params_filename, sizeof(omf_params_filename),
            "%s.omf", wo_param_data->work_order_id);
        if ((status < 0) || (status >= sizeof(omf_params_filename)))
        {
            IAS_LOG_ERROR("Creating OMF parameter file name");
            return ERROR;
        }
    }

/*-----------------------------------------------------------------------
                            READ OMF PARAMETER FILE
-------------------------------------------------------------------------*/
    /* Declare the OMF parameter table */
    IAS_PARM_DECLARE_TABLE(omf_param_table, MAX_NUM_OMF_PARAMS);

    /* Construct the parameter table, common parameters first */
    IAS_PARM_SATELLITE(omf_param_table, &omf_param_data->satellite_number,
            sizeof(omf_param_data->satellite_number), 1);

    IAS_PARM_L0R_FILENAME(omf_param_table, IAS_PARM_REQUIRED,
            omf_param_data->l0r_filename, sizeof(omf_param_data->l0r_filename),
            1);

    IAS_PARM_CPF(omf_param_table, IAS_PARM_REQUIRED,
            omf_param_data->cpf_filename,
            sizeof(omf_param_data->cpf_filename), 1);

    /* Get the name of the pixel mask file */
    IAS_PARM_PIXEL_MASK_FILENAME(omf_param_table, IAS_PARM_REQUIRED,
            omf_param_data->pmask_filename,
            sizeof(omf_param_data->pmask_filename), 1);

    /* See if the user wanted help or a template this time with the
       OMF file.  If not, just read the file */
    read_count = IAS_PARM_GET_TABLE_SIZE(omf_param_table);

    /* Provide parameter help if requested.  Check the return status
       to distinguish whether help was requested and successfully provided
       (1), no help was requested (0), or there was an error providing help
       (ERROR) */
    status = ias_parm_provide_help(omf_params_filename, omf_param_table,
                read_count, IAS_OMF_PARAMETERS);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("OMF parameter help failed");
        return ERROR;
    }
    else if (status == 1)
    {
        exit(EXIT_SUCCESS);
    }

    /* Read the OMF file */
    status = ias_parm_read(omf_params_filename, omf_param_table, read_count);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading OMF parameter file");
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}
