/**************************************************************************
 NAME:                     sca_metrics_summary

 PURPOSE:  Module implementing routines to calculate SCA-level summary
           statistics for each nonuniformity characterization, saved for
           display in the summary report and/or dump file

 RETURNS:  Integer status code of SUCCESS or ERROR

***************************************************************************/
#include <float.h>
#include <string.h>

#include "ias_logging.h"
#include "ias_radiometry.h"

#include "char_nonuniformity.h"


/*----------------------------------------------------------------------
 NAME:                    find_minmax_metrics

 PURPOSE:   Internal routine that searches the detector metrics results for
            the minimum and maximum metrics and their corresponding detector
            numbers

 RETURNS:   Nothing
------------------------------------------------------------------------*/
static void find_minmax_metrics
(
    const double *detector_metrics,/* I: Metrics data for all detectors
                                      in the current SCA */
    const int *metrics_invalid_flags,
                                   /* I: Metrics invalid flags for all
                                      detectors in the current SCA for
                                      the current characterization */
    int num_detectors_per_sca,     /* I: Number of detectors in an
                                      SCA for the current band */
    double *min_metric,            /* O: Minimum detector metric value */
    double *max_metric,            /* O: Maximum detector metric value */
    int *detector_min,             /* O: Detector number with minimum
                                      metric */
    int *detector_max              /* O: Detector number with maximum
                                      metric */
)
{
    int detector_index;
    int current_detector_min = 0;
    int current_detector_max = 0;

    double current_max_metric = -DBL_MAX;
    double current_min_metric =  DBL_MAX;

    for (detector_index = 0; detector_index < num_detectors_per_sca;
         detector_index++)
    {
        if (!metrics_invalid_flags[detector_index])
        {
            if (detector_metrics[detector_index] > current_max_metric)
            {
                current_max_metric = detector_metrics[detector_index];
                current_detector_max = (detector_index + 1);
            }
            if (detector_metrics[detector_index] < current_min_metric)
            {
                current_min_metric = detector_metrics[detector_index];
                current_detector_min = (detector_index + 1);
            }
        }
    }

    /* Save the final minimum and maximum metric values */
    *detector_max = current_detector_max;
    *detector_min = current_detector_min;
    *max_metric = current_max_metric;
    *min_metric = current_min_metric;
}



/*------------------------------------------------------------------------
 NAME:                   summarize_within_sca_metrics

 PURPOSE:      Routine to calculate summary metric statistics for all
               operable detectors within an SCA

 RETURNS:      Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int summarize_within_sca_metrics
(
    const METRICS_DET *detector_metrics_data, /* I: Detector metrics data
                                                 for the current
                                                 characterization */
    const int *metrics_invalid_flags,
                                              /* I: Invalid flags for all
                                                 detectors in the current
                                                 SCA for the current
                                                 characterization */
    int num_detectors_per_sca,                /* I: Total # of detectors
                                                 in an SCA */
    double band_radiance,                     /* I: Mean radiance for the
                                                 current band */
    SCA_SUMMARY *sca_summary                  /* O: Populated data structure
                                                 with per-SCA metrics summary
                                                 statistics for the
                                                 current characterization */
)
{
    int status;


    /* Find the min/max metric values and the corresponding detector
       numbers */
    find_minmax_metrics(detector_metrics_data->metric,
            metrics_invalid_flags,
            num_detectors_per_sca,
            &(sca_summary->metric_min),
            &(sca_summary->metric_max),
            &(sca_summary->detector_min),
            &(sca_summary->detector_max));

    /* Calculate the mean and std deviation of the metric for all
       (operable) detectors within an SCA */
    status = ias_rps_calculate_filtered_mean(
            detector_metrics_data->metric,
            metrics_invalid_flags,
            num_detectors_per_sca,
            &(sca_summary->metric_mean),
            &(sca_summary->num_characterized_detectors));
    if (status != SUCCESS)
    {
        return ERROR;
    }

    status = ias_rps_calculate_filtered_stdev(
            detector_metrics_data->metric,
            metrics_invalid_flags,
            num_detectors_per_sca,
            sca_summary->metric_mean,
            &(sca_summary->metric_stdev));
    if (status != SUCCESS)
    {
        return ERROR;
    }

    /* Calculate the mean metric as a % of the current band's mean
       radiance */
    sca_summary->pcnt = calculate_pcnt_metric(sca_summary->metric_mean,
            band_radiance);

    /* Done */
    return SUCCESS;
}




/*------------------------------------------------------------------------
 NAME:                   summarize_cross_sca_metrics

 PURPOSE:      Routine to calculate various cross-SCA summary metric
               statistics

 RETURNS:      Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int summarize_cross_sca_metrics
(
 const METRICS_DET *detector_metrics_data, /* I: Detector metrics data
                                              for all SCAs in the current
                                              characterization */
 const int *metrics_invalid_flags,         /* I: Metrics invalid flags for
                                              all detectors in the current
                                              SCA for the current
                                              characterization */
 int sca_index,                            /* I: 0-based index for the
                                              current SCA */
 int num_detectors_per_sca,                /* I: Total # of detectors
                                              in an SCA */
 int num_boundary_detectors,               /* I: # of detectors on both
                                              sides of an SCA boundary */
 const char *cross_metric_basis,           /* I: String indicating the
                                              basis for the cross-metric
                                              calculation ("MEAN",
                                              "STDEV") */
 double band_radiance,                     /* I: Mean radiance for the
                                              current band */
 CROSS_SCA_SUMMARY *cross_sca_summary      /* O: Populated data structure
                                              with cross-SCA summary
                                              statistics for the
                                              current characterization */
)
{
    int start_detector_index;                 /* Starting detector index */
    int status;

    /* Calculate the mean and standard deviation "across" an SCA boundary
       for RMS and/or stdev banding characterizations.  Ideally, these
       metrics would be calculated in a window centered on each
       detector, and the cross-boundary statistics would represent actually
       "crossing" an SCA boundary (with half the detectors on each side).  In
       practice, however, the left edge of the window used in calculating
       the detector metrics is on the detector itself.  Consequently, the
       calculation window for the cross-boundary statistics has to
       be shifted accordingly such that the window covers the number of
       boundary detectors up to an SCA boundary.  For example,
       the cross-boundary statistics for SCAs 1 and 2, are calculated from
       the last (num_boundary_detectors) group of operable detectors in
       SCA 1, just to the the "left" of SCA 2.

       Not sure that's clearly explained.  It's how the prototype is
       doing the cross-boundary statistics calculations.

       The number of boundary detectors representing the cross-boundary
       "window" is #defined in char_nonuniformity.h  */
    start_detector_index = num_detectors_per_sca
        - num_boundary_detectors;

    status = ias_rps_calculate_filtered_mean(
            &(detector_metrics_data[sca_index].metric[start_detector_index]),
            &(metrics_invalid_flags[start_detector_index]),
            num_boundary_detectors,
            &(cross_sca_summary->cross_mean),
            &(cross_sca_summary->cross_num_pixels));
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating cross-SCA summary metrics mean");
        return ERROR;
    }


    status = ias_rps_calculate_filtered_stdev(
            &(detector_metrics_data[sca_index].metric[start_detector_index]),
            &(metrics_invalid_flags[start_detector_index]),
            num_boundary_detectors,
            cross_sca_summary->cross_mean,
            &(cross_sca_summary->cross_stdev));
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating cross-SCA summary statistics std dev");
        return ERROR;
    }

    /* Do the % cross-metric calculation.  For RMS banding, the
       cross-boundary % metric is based on the cross-boundary std deviation;
       for stdev banding, the cross-boundary % metric is based on the
       cross-boundary mean */
    if (!strcmp(cross_metric_basis, "MEAN"))     /* stdev banding */
    {
        cross_sca_summary->cross_pcnt = calculate_pcnt_metric(
                cross_sca_summary->cross_mean,
                band_radiance);
    }
    else if (!strcmp(cross_metric_basis, "STDEV")) /* RMS banding */
    {
        cross_sca_summary->cross_pcnt = calculate_pcnt_metric(
                cross_sca_summary->cross_stdev,
                band_radiance);
    }
    else
    {
        IAS_LOG_ERROR("Invalid cross-metric (%%) basis identifier %s",
                cross_metric_basis);
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}




/*---------------------------------------------------------------------
 NAME:               calculate_per_detector_metrics_std

 PURPOSE:   Calculates the per-detector std deviation of the
            selected characterization metric(s) for all detectors in the
            current band, and writes the results to the corresponding
            members of the summary report data structure

 RETURNS:   Integer status code of SUCCESS or ERROR
-----------------------------------------------------------------------*/
int calculate_per_detector_metrics_std
(
    const double *band_rms_buffer,         /* I: RMS metrics buffer for the
                                              current band */
    const double *band_stdev_buffer,       /* I: Stdev banding metrics
                                              buffer for the current band */
    const double *band_streaking_buffer,   /* I: Streaking metrics buffer
                                              for the current band */
    const RPS_NUCHAR_INVALID_FLAGS *invalid_flags,
                                           /* I: Characterization result
                                              invalid flags for all
                                              detectors in the current
                                              band */
    int num_band_detectors,                /* I: Total # of detectors in the
                                              current band */
    int rms_banding_flag,                  /* I: Flag indicating whether
                                              RMS banding characterization
                                              is performed */
    int stdev_banding_flag,                /* I: Flag indicating whether
                                              stdev banding characterization
                                              is performed */
    int streaking_flag,                    /* I: Flag indicating whether
                                              streaking characterization is
                                              performed */
    RPS_NUCHAR_REPORT_SUMMARY *summary     /* O: Populated metrics data
                                              structure */
)
{
    double mean_metric = 0.0;
    double stdev_metric = 0.0;

    int status;

    /* Calculate the mean and standard deviation for each selected
       characterization.  We're really only interested in the standard
       deviation, but we need the mean to get it */
    if (rms_banding_flag)
    {
        status = ias_rps_calculate_filtered_mean(band_rms_buffer,
                                                 invalid_flags->rms_banding,
                                                 num_band_detectors,
                                                 &mean_metric, NULL);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining mean RMS banding metric");
            return ERROR;
        }

        status = ias_rps_calculate_filtered_stdev(band_rms_buffer,
                                                  invalid_flags->rms_banding,
                                                  num_band_detectors,
                                                  mean_metric,
                                                  &(stdev_metric));
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining std deviation of RMS banding metric");
            return ERROR;
        }

        summary->rms_band_mean = mean_metric;
        summary->rms_band_stdev = stdev_metric;
    }

    if (stdev_banding_flag)
    {
        status = ias_rps_calculate_filtered_mean(band_stdev_buffer,
                                                 invalid_flags->stdev_banding,
                                                 num_band_detectors,
                                                 &mean_metric, NULL);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining mean stdev banding metric");
            return ERROR;
        }

        status = ias_rps_calculate_filtered_stdev(band_stdev_buffer,
                                                 invalid_flags->stdev_banding,
                                                 num_band_detectors,
                                                 mean_metric,
                                                 &(stdev_metric));
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining std deviation of stdev banding metric");
            return ERROR;
        }

        summary->stdev_band_mean = mean_metric;
        summary->stdev_band_stdev = stdev_metric;
    }

    if (streaking_flag)
    {
        status = ias_rps_calculate_filtered_mean(band_streaking_buffer,
                                                 invalid_flags->streaking,
                                                 num_band_detectors,
                                                 &mean_metric, NULL);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining mean streaking metric");
            return ERROR;
        }

        status = ias_rps_calculate_filtered_stdev(band_streaking_buffer,
                                                  invalid_flags->streaking,
                                                  num_band_detectors,
                                                  mean_metric,
                                                  &(stdev_metric));
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining std deviation of streaking metric");
            return ERROR;
        }

        summary->streaking_mean = mean_metric;
        summary->streaking_stdev = stdev_metric;
    }

    /* Done */
    return SUCCESS;
}
