#!/bin/csh -f
#
# Usage: ./check_or_insert_end2end_woid.csh [Work Order ID] [sensor] 
#

# Ensure there is a database environment variable defined
if ( ! $?IAS_DB_L8_OLITIRS ) then
    echo "IAS_DB_L8_OLITIRS isn't set"
    exit 1
endif

if ( ${#argv} != 2 ) then
    echo "Usage: $0 [Work Order ID] [Sensor (OLITIRS)]"
    exit 1
endif

if (($2 != "OLITIRS")) then
    echo "Invalid sensor ID $2, must be combined 'OLITIRS'"
    exit 
endif

# First check to see if the Work order ID is already in the database
# Get the radiometric common ID
set sql_script = "retrieve_woid.sql"
echo "set term off" > $sql_script
echo "set echo on" >> $sql_script
echo "set trimout on" >> $sql_script
echo "set trimspool on" >> $sql_script
echo "set heading off" >> $sql_script
echo "set feedback off" >> $sql_script
echo "set pause off" >> $sql_script
echo "set pagesize 0" >> $sql_script
echo "set linesize 500" >> $sql_script
echo "set space 0" >> $sql_script
echo "set colsep |" >> $sql_script
echo "set tab off" >> $sql_script
echo "set verify off" >> $sql_script
echo "spool ./woid.txt" >> $sql_script
echo "select work_order_id, radiometric_wo_common_id " >> $sql_script
echo "from radiometric_work_order_common " >> $sql_script
echo "where work_order_id='"$1"';" >> $sql_script
echo "spool off" >> $sql_script
echo "exit" >> $sql_script

# Query the database
echo "Performing initial check..."

$ORACLE_HOME/bin/sqlplus $IAS_DB_L8_OLITIRS @$sql_script > & /dev/null
rm $sql_script
if ( $status ) then
    echo "Error performing initial database check"
    exit 1
endif

# Squeeze out the white space and select the lines beginning with the 
# work order ID 
set common_id = `sed 's/  *//g' ./woid.txt | grep ^$1 | \
    awk -F'|' '{print $2}'`

# How many ID numbers do we have?  In principle, there should be only 1.
# If there are more than 1, that needs to be dealt with...
set id_count = 0
foreach id ($common_id)
    @ id_count += 1
end

# Exactly 0 -- log that it's currently not found
if ( $id_count == 0 ) then
    echo "radiometric_wo_common_id not found for work order '$1', inserting "
    echo "into database..."

# Exactly 1 -- clean out the tables so that they can be reloaded with the
#              current test data set.  Removing the entry in the
#              RADIOMETRIC_WORK_ORDER_COMMON table should 'cascade'
#              delete everything associated with that ID
#echo $common_id, $id_count
else if ( $id_count == 1 ) then
    echo "$1 entered in database, removing previous data..."
    echo "delete from radiometric_work_order_common " > delete_woid.sql
    echo "where work_order_id='$1';" >> delete_woid.sql
    $ORACLE_HOME/bin/sqlplus $IAS_DB_L8_OLITIRS < delete_woid.sql \
        >& /dev/null
    if ( $status != 0 ) then
        echo "Error executing sqlplus work order data deletion."
        /bin/rm delete_woid.sql
        /bin/rm woid.txt
        exit 1
    endif
    echo "Data deletion completed."
    /bin/rm woid.txt
    /bin/rm delete_woid.sql

# Given the current database design/implementation,
# this code should NEVER execute.  In case we get here for some oddball
# reason, however, log an error message and exit out--this will catch
# something needing further investigation...
else
    echo "Should not have reached this code block given the current DB "
    echo "design and implementation.  This needs to be checked..."
    exit 1
endif


# (Re)load the radiometric_work_order_common information.
set controlfile = "rad_woid_for_DB_simearth.ldr_original"
$ORACLE_HOME/bin/sqlldr userid=$IAS_DB_L8_OLITIRS errors=0, silent=ALL, \
    control=$controlfile
if ($status) then
    echo "Error loading radiometric workorder common table information \
          for $2 nonuniformity characterization"
    exit 1
endif


# The radiometric_wo_common_id from the previous load isn't readily
# available here.  So... we'll have to query
# RADIOMETRIC_WORK_ORDER_COMMON to get it.

# Run the query
set sql_script = "get_rad_common_id.sql"
echo "set heading off" > $sql_script
echo "select radiometric_wo_common_id" >> $sql_script
echo "from radiometric_work_order_common" >> $sql_script
echo "where work_order_id='"$1"';" >> $sql_script
echo "exit" >> $sql_script

set common_id = `$ORACLE_HOME/bin/sqlplus -S $IAS_DB_L8_OLITIRS \
    @get_rad_common_id.sql`
echo "radiometric_wo_common_id is $common_id for work order $1..."
/bin/rm get_rad_common_id.sql


# Load the data for bands 1, 7, and 8
echo "Loading band 1 test data for ID $common_id..."
sed -e "s/XXX/$common_id/" \
$IAS_UNIT_TEST_DATA_DIR/ias/rps/nonuniformity/diffuser_stats_DB_B01.ldr_original_earth > diffuser_stats_DB_B01.ldr
$ORACLE_HOME/bin/sqlldr userid=$IAS_DB_L8_OLITIRS errors=0 silent=ALL \
    control = diffuser_stats_DB_B01.ldr
if ( $status ) then
    echo "Error loading band 1 gain test data"
    rm diffuser_stats_DB_B01.ldr
    exit 1
endif
rm diffuser_stats_DB_B01.ldr

echo "Loading band 7 test data for ID $common_id..."
sed -e "s/XXX/$common_id/" \
$IAS_UNIT_TEST_DATA_DIR/ias/rps/nonuniformity/diffuser_stats_DB_B07.ldr_original_earth > diffuser_stats_DB_B07.ldr
$ORACLE_HOME/bin/sqlldr userid=$IAS_DB_L8_OLITIRS errors=0 silent=ALL \
        control = diffuser_stats_DB_B07.ldr
if ( $status ) then
    echo "Error loading band 7 gain test data"
    rm diffuser_stats_DB_B07.ldr
    exit 1
endif
    rm diffuser_stats_DB_B07.ldr

echo "Loading band 8 test data for ID $common_id..."
sed -e "s/XXX/$common_id/" \
$IAS_UNIT_TEST_DATA_DIR/ias/rps/nonuniformity/diffuser_stats_DB_B08.ldr_original_earth > diffuser_stats_DB_B08.ldr
$ORACLE_HOME/bin/sqlldr userid=$IAS_DB_L8_OLITIRS errors=0 silent=ALL \
    control=diffuser_stats_DB_B08.ldr
if ( $status ) then
    echo "Error loading band 8 gain test data"
    rm diffuser_stats_DB_B08.ldr
    exit 1
endif
rm diffuser_stats_DB_B08.ldr


# Load the data for bands 10 and 11
echo "Loading band 10 test data for ID $common_id..."
sed -e "s/XXX/$common_id/" \
$IAS_UNIT_TEST_DATA_DIR/ias/rps/nonuniformity/bb_stats_DB_B10.ldr_original_earth > bb_stats_DB_B10.ldr
$ORACLE_HOME/bin/sqlldr userid=$IAS_DB_L8_OLITIRS errors=0 silent=ALL \
    control=bb_stats_DB_B10.ldr
if ( $status ) then
    echo "Error loading band 10 test data"
    rm bb_stats_DB_B10.ldr
    exit 1
endif
rm bb_stats_DB_B10.ldr

echo "Loading band 11 test data for ID $common_id..."
sed -e "s/XXX/$common_id/" \
$IAS_UNIT_TEST_DATA_DIR/ias/rps/nonuniformity/bb_stats_DB_B11.ldr_original_earth > bb_stats_DB_B11.ldr
$ORACLE_HOME/bin/sqlldr userid=$IAS_DB_L8_OLITIRS errors=0 silent=ALL \
    control=bb_stats_DB_B11.ldr
if ( $status ) then
    echo "Error loading band 11 test data"
    rm bb_stats_DB_B11.ldr
    exit 1
endif
    rm bb_stats_DB_B11.ldr


exit 0
