/**********************************************************************
                     initialize_tests

 PURPOSE:      Module containing routines to initialize data structures
               for nonuniformity characterization unit testing

 NOTES:        Initialization routines are intended to work independently
               of the specific metrics to be calculated

***********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>

#include "ias_radiometry.h"
#include "ias_satellite_attributes.h"

#include "../char_nonuniformity.h"

#include "nonuniformity_tests.h"

/* #define DEBUG_PRINT */


/*-------------------------------------------------------------------
 NAME:              initialize_band_metrics

 PURPOSE:    Calculates the required nonuniformity metrics for all
             operable detectors in the current band

 RETURNS:    Array of calculated nonuniformity metrics for the entire
             band if successful, NULL pointer if there's any error
---------------------------------------------------------------------*/
double *initialize_band_metrics
(
    const char *radiance_filename,         /* I: Name of radiance data
                                              file */
    int num_band_detectors,                /* I: # of detectors in the
                                              band */
    int num_contiguous_detectors,          /* I: Total # of contiugous
                                              detectors (used in RMS
                                              and stdev banding
                                              characterization) */
    const char *metrics_calculation_name,  /* I: Name of function to
                                              call to calculate the
                                              desired nonuniformity
                                              metrics */
    double *band_radiance_mean,            /* O: Mean radiance for all
                                              detectors in the band */
    double *band_radiance_stdev,           /* O: Standard dev of radiances
                                              in the band */
    int *num_samples,                      /* O: Total # of samples used
                                              to calculate the mean */
    int *invalid_results_flags  /* O: Detector list flagged with
                                              invalid characterization
                                              results */
)
{
    double *radiance_buffer = NULL;        /* Array of detector
                                              radiances */
    double *band_metrics_buffer = NULL;    /* Array of metrics for
                                              each operable detector */
    double stats_mean;                     /* Return variable holding
                                              mean radiance */
    double stats_stdev;                    /* Local variable holding
                                              radiance std deviation */

    /* Placeholder variable only used when reading the radiance data
       file */
    double dummy_stdev;

    FILE *rad_fptr = NULL;                /* File handle for
                                              radiance data file */

    int status;                            /* Function return status */
    int detector_index;                    /* 0-based detector index */
    int dummy_int;                         /* Placeholder for unneeded
                                              detector ID when reading
                                              radiance data file */
    int stats_num_samples;                 /* Return variable holding
                                              number of valid samples
                                              used to determine the mean */

    int *flag_buffer = NULL;


    /* Allocate the radiance buffer and the output metrics buffer */
    radiance_buffer = calloc(num_band_detectors, sizeof(*radiance_buffer));
    if (radiance_buffer == NULL)
    {
        printf("Cannot allocate radiance data buffer\n");
        return NULL;
    }
    band_metrics_buffer = calloc(num_band_detectors,
            sizeof(*band_metrics_buffer));
    if (band_metrics_buffer == NULL)
    {
        printf("Cannot allocate output metrics data buffer\n");
        free(radiance_buffer);
        free(band_metrics_buffer);
        return NULL;
    }

    /* Allocate the flag buffers */
    flag_buffer = calloc(num_band_detectors, sizeof(*flag_buffer));
    if ( ! flag_buffer )
    {
        printf("Error allocating flag buffer\n");
        free(radiance_buffer);
        free(band_metrics_buffer);
        return NULL;
    }

    /* Read the radiance data.  The file is expected to be in a
       3-column format:  1st column is detector number (1-based,
       band-relative), 2nd column is the radiance value, and the
       third column is the standard deviation.  There WILL BE issues
       if a radiance data file does not have this format. */
    rad_fptr = fopen(radiance_filename, "r");
    if (rad_fptr == NULL)
    {
        printf("Cannot open radiance data file %s for reading\n",
                radiance_filename);
        free(radiance_buffer);
        free(band_metrics_buffer);
        free(flag_buffer);
        return NULL;
    }

    for (detector_index = 0;
         detector_index < num_band_detectors;
         detector_index++)
    {
        fscanf(rad_fptr, "%d%lf%lf", &dummy_int,
                &radiance_buffer[detector_index], &dummy_stdev);
        if (radiance_buffer[detector_index] <= 0)
        {
            flag_buffer[detector_index] = 1;
        }
    }
    fclose(rad_fptr);

    /* Calculate the summary band radiance statistics */
    status = ias_rps_calculate_filtered_mean(radiance_buffer, flag_buffer,
        num_band_detectors, &stats_mean, &stats_num_samples);
    if (status != SUCCESS)
    {
        printf("Cannot calculate band-level radiance mean\n");
        free(radiance_buffer);
        free(band_metrics_buffer);
        free(flag_buffer);
        return NULL;
    }

    if ((stats_mean < DBL_EPSILON) || (stats_num_samples <= 0))
    {
        printf("Cannot properly retrieve summary statistics\n");
        free(radiance_buffer);
        free(band_metrics_buffer);
        free(flag_buffer);
        return NULL;
    }
    if (num_samples != NULL)
        *num_samples = stats_num_samples;

    status = ias_rps_calculate_filtered_stdev(radiance_buffer, flag_buffer,
        num_band_detectors, stats_mean, &stats_stdev);
    if (status != SUCCESS)
    {
        printf("Cannot calculate std deviation of radiance data\n");
        free(radiance_buffer);
        free(band_metrics_buffer);
        free(flag_buffer);
        return NULL;
    }

    *band_radiance_mean = stats_mean;
    *band_radiance_stdev = stats_stdev;

    /* Calculate the metrics specified by the name of the function to
       be used */
    if (!strcmp(metrics_calculation_name, "calculate_rms_banding_metrics"))
    {
        status = calculate_rms_banding_metrics(radiance_buffer, flag_buffer,
            num_band_detectors, num_contiguous_detectors, *band_radiance_mean,
            band_metrics_buffer, invalid_results_flags);
        if (status != SUCCESS)
        {
            printf("Cannot calculate RMS banding metrics\n");
            free(radiance_buffer);
            free(band_metrics_buffer);
            free(flag_buffer);
            return NULL;
        }
    }

    if (!strcmp(metrics_calculation_name, "calculate_stdev_banding_metrics"))
    {
        status = calculate_stdev_banding_metrics(radiance_buffer, flag_buffer,
            num_band_detectors, num_contiguous_detectors,
            band_metrics_buffer, invalid_results_flags);
        if (status != SUCCESS)
        {
            printf("Cannot calculate stdev banding metrics\n");
            free(radiance_buffer);
            free(band_metrics_buffer);
            free(flag_buffer);
            return NULL;
        }
    }

    if (!strcmp(metrics_calculation_name, "calculate_streaking_metrics"))
    {
        status = calculate_streaking_metrics(radiance_buffer, flag_buffer,
            num_band_detectors, band_metrics_buffer, invalid_results_flags);
        if (status != SUCCESS)
        {
            printf("Cannot calculate streaking metrics\n");
            free(radiance_buffer);
            free(band_metrics_buffer);
            free(flag_buffer);
            return NULL;
        }
    }

    /* We've got a set of metrics calculated.  Cleanup */
    free(radiance_buffer);
    free(flag_buffer);

    return band_metrics_buffer;
}




/*-------------------------------------------------------------------
 NAME:              free_summary_report_buffers

 PURPOSE:    Frees data structure arrays allocated by
             allocate_summary_report_buffers

 RETURNS:    Nothing
---------------------------------------------------------------------*/
void free_summary_report_buffers
(
    RPS_NUCHAR_REPORT_SUMMARY *summary    /* I/O: Report summary
                                             data structure */
)
{
    if (summary->sca_summary_rms != NULL)
        free(summary->sca_summary_rms);

    if (summary->sca_summary_stdev != NULL)
        free(summary->sca_summary_stdev);

    if (summary->sca_summary_streaking != NULL)
        free(summary->sca_summary_streaking);

    if (summary->cross_sca_rms != NULL)
        free(summary->cross_sca_rms);

    if (summary->cross_sca_stdev != NULL)
        free(summary->cross_sca_stdev);
}



/*-------------------------------------------------------------------
 NAME:              free_detector_metrics_buffers

 PURPOSE:    Frees data structure arrays allocated by
             allocate_metrics_buffers

 RETURNS:    Nothing
---------------------------------------------------------------------*/
void free_detector_metrics_buffers
(
    RPS_NUCHAR_METRICS *metrics,   /* I/O:  Metrics data structure */
    int num_scas                   /* I: # of SCAs in the current band */
)
{
    int sca_index;

    if (metrics->rms_metrics != NULL)
    {
        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            free(metrics->rms_metrics[sca_index].detector_number);
            free(metrics->rms_metrics[sca_index].metric);
        }
        free(metrics->rms_metrics);
    }
    if (metrics->stdev_metrics != NULL)
    {
        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            free(metrics->stdev_metrics[sca_index].detector_number);
            free(metrics->stdev_metrics[sca_index].metric);
        }
        free(metrics->stdev_metrics);
    }
    if (metrics->streaking_metrics != NULL)
    {
        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            free(metrics->streaking_metrics[sca_index].detector_number);
            free(metrics->streaking_metrics[sca_index].metric);
        }
        free(metrics->streaking_metrics);
    }
}




/*----------------------------------------------------------------------
 NAME:          initialize_metrics_data

 PURPOSE:  Populates an RPS_NUCHAR_METRICS data structure with
           band level metrics information for the specified
           characterization(s)

 RETURNS:  Integer status code of SUCCESS or ERROR
------------------------------------------------------------------------*/
void initialize_metrics_data
(
    int rms_banding_flag,          /* I: RMS characterization flag */
    int stdev_banding_flag,        /* I: Stdev characterization flag */
    int streaking_flag,            /* I: Streaking characterization flag */
    double *band_rms_metrics,      /* I: RMS metrics data buffer */
    double *band_stdev_metrics,    /* I: Stdev metrics data buffer */
    double *band_streaking_metrics,/* I: Streaking metrics data buffer */
    double radiance_mean,          /* I: Mean radiance for all operable
                                      detectors in the current band */
    double radiance_stdev,         /* I: Stdev of operable detector
                                      radiances in the current band */
    int band_number,               /* I: Current 1-based band number */
    int num_scas,                  /* I: # of SCAs in the current band */
    int num_detectors_per_sca,     /* I: # of detectors in an SCA */
    RPS_NUCHAR_METRICS *metrics    /* O: Metrics data structure to
                                      populate */
)
{
    int sca_index;

    /* Start with band properties */
    metrics->band_number = band_number;

    /* Band summary information */
    metrics->band_radiance_mean = radiance_mean;
    metrics->band_radiance_stdev = radiance_stdev;
    metrics->ffov_metric = (radiance_stdev / radiance_mean) * 100.0;

    /* Now copy metrics from the band-level arrays to the individual
       SCA data buffers */
    for (sca_index = 0; sca_index < num_scas; sca_index++)
    {
        populate_metrics_buffers(band_rms_metrics,
                band_stdev_metrics,
                band_streaking_metrics,
                rms_banding_flag, stdev_banding_flag,
                streaking_flag, (sca_index + 1),
                num_detectors_per_sca,
                metrics);
    }

    /* Done */
}

/*----------------------------------------------------------------------
 NAME:          initialize_summary_report_data_structures

 PURPOSE:   Allocates the report summary data structure buffers and
            populates them with summary metrics information

 RETURNS:   Integer status code of SUCCESS or ERROR
-------------------------------------------------------------------------*/
int initialize_summary_report_data_structures
(
    RPS_NUCHAR_WO_PARAMS wo_params,              /* I: Work order
                                                    parameters */
    int rms_banding_flag,                        /* I: Flag indicating
                                                    whether RMS banding
                                                    characterization is
                                                    performed */
    int stdev_banding_flag,                      /* I: Flag indicating
                                                    whether stdev banding
                                                    characterization is
                                                    performed */
    int streaking_flag,                          /* I: Flag indicating
                                                    whether streaking
                                                    characterization is
                                                    performed */
    RPS_NUCHAR_METRICS metrics,                  /* I: Populated metrics
                                                    data structure */
    const RPS_NUCHAR_INVALID_FLAGS *invalid_flags,
                                                 /* I: Characterization
                                                    invalid flags for all
                                                    detectors in the
                                                    current band */
    RPS_NUCHAR_REPORT_SUMMARY *summary           /* O: Populated summary
                                                    report data
                                                    structures */
)
{
    int status;

    const IAS_BAND_ATTRIBUTES *band_attributes = NULL;

    /* Get the band attributes */
    band_attributes = ias_sat_attr_get_band_attributes(1);
    assert_true(band_attributes != NULL);
    if (band_attributes == NULL)
    {
        printf("Could not retrieve IAS attributes for band 1\n");
        return ERROR;
    }

    /* Allocate the summary report data structures */
    status = allocate_summary_report_buffers(rms_banding_flag,
                                             stdev_banding_flag,
                                             streaking_flag,
                                             band_attributes->scas,
                                             summary);
    if (status != SUCCESS)
    {
        printf("Cannot allocate summary report data structures\n");
        free_detector_metrics_buffers(&metrics, band_attributes->scas);
        return ERROR;
    }

    /* Start populating the summary report data structures */
    initialize_nonuniformity_parameter_data(wo_params, summary);
    initialize_summary_report_band_data(wo_params, band_attributes,
                                        summary);
    status = initialize_sca_summary_metrics(metrics, invalid_flags,
                                         rms_banding_flag,
                                         stdev_banding_flag,
                                         streaking_flag,
                                         band_attributes->scas,
                                         band_attributes->detectors_per_sca,
                                         summary);
    if (status != SUCCESS)
    {
        printf("Cannot populate summary metrics data structures\n");
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}
