/**********************************************************************
 NAME:                non_uniformity_cleanup_test

 PURPOSE:      Use CGREEN to perform unit testing on the cleanup_application
               routine implemented for the char_nonuniformity application

***********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>

#include "ias_cpf.h"
#include "ias_db.h"
#include "ias_logging.h"
#include "ias_pixel_mask.h"

#include "cgreen.h"

#include "../char_nonuniformity.h"

#include "nonuniformity_tests.h"

/*#define DEBUG_PRINT */

#define TEST_CPF_NAME_LENGTH 27   /* Length + NULL terminator */

static char cpf_dir[] = "ias_lib/cpf";
static char cpf_filename[] = "L8CPF20090101_20090331.01";

Ensure test_cleanup_application()
{
    char *path = NULL;
    char full_cpfname[PATH_MAX];

    double *rms_buffer = NULL;
    double *stdev_buffer = NULL;
    double *streaking_buffer = NULL;
    double *hist_buffer = NULL;
    double *radiance_buffer = NULL;

    RPS_NUCHAR_METRICS test_metrics;
    RPS_NUCHAR_REPORT_SUMMARY test_summary;
    RPS_NUCHAR_INVALID_FLAGS invalid_flags;

    struct ias_db_connection *open_db = NULL;

    IAS_PIXEL_MASK **pmasks = NULL;

    int status;
    int rms_banding_flag = 1;
    int stdev_banding_flag = 1;
    int streaking_flag = 1;
    int trend_flag = 0;
    int num_scas = 14;
    int num_detectors_per_sca = 494;
    int num_band_detectors = num_scas * num_detectors_per_sca;
    int num_masks_read = 0;

    IAS_CPF *cpf = NULL;


    printf("Testing application shutdown...\n");

    /* Make sure the test data path exists.  If it does, build the name
       (with path) of the CPF file to read */
    path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (path == NULL)
    {
        assert_true_with_message(FALSE, "Cannot find test data directory path");
        return;
    }
    sprintf(full_cpfname, "%s/%s/%s", path, cpf_dir, cpf_filename);

    /* Initialize the invalid flag data buffers */
    memset(&invalid_flags, 0, sizeof(RPS_NUCHAR_INVALID_FLAGS));
    status = allocate_nonuniformity_invalid_flag_buffers(num_band_detectors,
        &invalid_flags);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        return;
    }

    /* Initialize the metrics and report summary data structures, then
       allocate buffers */
    memset(&test_metrics, 0, sizeof(RPS_NUCHAR_METRICS));
    memset(&test_summary, 0, sizeof(RPS_NUCHAR_REPORT_SUMMARY));

    status = allocate_metrics_buffers(rms_banding_flag, stdev_banding_flag,
            streaking_flag, num_scas,
            num_detectors_per_sca, &test_metrics);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot allocate metrics data structure buffers\n");
        free_detector_metrics_buffers(&test_metrics, num_scas);
        free_nonuniformity_invalid_flag_buffers(&invalid_flags);
        return;
    }

    status = allocate_summary_report_buffers(rms_banding_flag,
            stdev_banding_flag,
            streaking_flag,
            num_scas,
            &test_summary);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot allocate summary report data buffers\n");
        free_detector_metrics_buffers(&test_metrics, num_scas);
        free(test_summary.sca_summary_rms);
        free(test_summary.sca_summary_stdev);
        free(test_summary.sca_summary_streaking);
        free(test_summary.cross_sca_rms);
        free(test_summary.cross_sca_stdev);
        free_nonuniformity_invalid_flag_buffers(&invalid_flags);
        return;
    }

    /* Read the test CPF */
    cpf = ias_cpf_read(full_cpfname);
    assert_true(cpf != NULL);
    if (cpf == NULL)
    {
        printf("Cannot read CPF file %s\n", full_cpfname);
        free_detector_metrics_buffers(&test_metrics, num_scas);
        free(test_summary.sca_summary_rms);
        free(test_summary.sca_summary_stdev);
        free(test_summary.sca_summary_streaking);
        free(test_summary.cross_sca_rms);
        free(test_summary.cross_sca_stdev);
        free_nonuniformity_invalid_flag_buffers(&invalid_flags);
        return;
    }

    /* Finally, allocate the data buffers */
    num_band_detectors = num_scas * num_detectors_per_sca;

    rms_buffer = calloc(num_band_detectors, sizeof(*rms_buffer));
    stdev_buffer = calloc(num_band_detectors, sizeof(*stdev_buffer));
    streaking_buffer = calloc(num_band_detectors,
            sizeof(*streaking_buffer));

    radiance_buffer = calloc(num_band_detectors, sizeof(*radiance_buffer));
    hist_buffer = calloc(num_detectors_per_sca, sizeof(*hist_buffer));

    if ((rms_buffer == NULL)
            || (stdev_buffer == NULL)
            || (streaking_buffer == NULL)
            || (radiance_buffer == NULL)
            || (hist_buffer == NULL))
    {
        assert_true_with_message(FALSE, "Cannot allocate required data "
            "buffers");
        free(rms_buffer);
        free(stdev_buffer);
        free(streaking_buffer);
        free(radiance_buffer);
        free(hist_buffer);
        free_detector_metrics_buffers(&test_metrics, num_scas);
        free(test_summary.sca_summary_rms);
        free(test_summary.sca_summary_stdev);
        free(test_summary.sca_summary_streaking);
        free(test_summary.cross_sca_rms);
        free(test_summary.cross_sca_stdev);
        ias_cpf_free(cpf);
        free_nonuniformity_invalid_flag_buffers(&invalid_flags);
        return;
    }

    /* Call the cleanup application routine, assuming SUCCESS */
    cleanup_application(trend_flag, num_scas, &test_summary, &test_metrics,
            &invalid_flags, open_db, rms_buffer, stdev_buffer,
            streaking_buffer, radiance_buffer, hist_buffer, cpf, pmasks,
            num_masks_read);
    /* Done */
}




/*-----------------------------------------------------------------------
           Define a test suite to add to the "master" test suite
-------------------------------------------------------------------------*/
void nonuniformity_cleanup_test(TestSuite *suite)
{
    add_test(suite, test_cleanup_application);
}
