/**********************************************************************
                     nonuniformity_dbaccess_test

 PURPOSE:      Use CGREEN to perform unit testing of the various
               component functions making up non-uniformity
               characterization database access

 RETURNS:      CGREEN status

***********************************************************************/
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "ias_db.h"
#include "ias_radiometry.h"

#include "cgreen.h"

#include "../char_nonuniformity.h"

#include "nonuniformity_tests.h"

static char work_order_id[IAS_WORKORDER_ID_SIZE] = "NUT1002";
static char rad_filename[] = "JAB_DetRadiances";
static char rad_filedir[] = "ias/rps/nonuniformity";
static char full_rad_filename[PATH_MAX] = "";
static char system_command[1000] = "";


/* Internal function to handle cleanup in test_write_to_detector_table */
static void write_detector_table_cleanup
(
    RPS_NUCHAR_METRICS *metrics,
    int num_scas,
    struct ias_db_connection *db,
    double *pcnt_rms_metrics,
    double *pcnt_stdev_metrics,
    double *db_streaking_metrics,
    IAS_DB_NULL_TYPE *flagged_rms_detectors,
    IAS_DB_NULL_TYPE *flagged_stdev_detectors,
    IAS_DB_NULL_TYPE *flagged_streaking_detectors,
    int *detector_ids
)
{
    /* Disconnect from the DB */
    ias_db_close_connection(db);
    ias_db_close_database_lib();

    /* Free up the metrics data structure data buffers */
    free_detector_metrics_buffers(metrics, num_scas);

    /* Free the other buffers */
    free(pcnt_rms_metrics);
    free(pcnt_stdev_metrics);
    free(db_streaking_metrics);
    free(flagged_rms_detectors);
    free(flagged_stdev_detectors);
    free(flagged_streaking_detectors);
    free(detector_ids);
}



Ensure test_get_histogram_statistics()
{
    char rad_wo_common_id[IAS_CHAR_ID_SIZE];
    int  band_number = 1;
    int  sca_number = 1;
    int  num_detectors_per_sca = 494;
    int  status;
    double *hist_stats_db_buffer      = NULL;
    struct ias_db_connection *open_db = NULL;
    const char *collection_type = IAS_RPS_EARTH_IMAGING_COLLECT;


    printf("Testing histogram statistics retrieval...\n");

    hist_stats_db_buffer = calloc(num_detectors_per_sca,
        sizeof(*hist_stats_db_buffer));
    if (hist_stats_db_buffer == NULL)
    {
        printf("Cannot allocate test data buffer\n");
        return;
    }

    /* Prep the database.  Put the test data in it for this work order ID
       if it isn't already in there */
    sprintf(system_command, "./check_or_insert_woid.csh %s", work_order_id);
    status = system(system_command);
    assert_equal_with_message(status, 0, "Error preparing database, "
        "cannot insert work order ID %s and/or its test data",
        work_order_id);
    if (status != 0)
    {
        free(hist_stats_db_buffer);
        return;
    }

    /* Connect to the database, and get the proper radiometric work order
       ID number */
    open_db = connect_to_db();
    assert_true(open_db != NULL);
    if (open_db == NULL)
    {
        printf("Cannot connect to database\n");
        ias_db_close_database_lib();
        free(hist_stats_db_buffer);
        return;
    }
    status = ias_misc_get_char_id(open_db, "RADIOMETRIC_WORK_ORDER_COMMON",
        "radiometric_wo_common_id", work_order_id, rad_wo_common_id);
    assert_not_equal(status, ERROR);
    if (status == ERROR)
    {
        printf("Cannot obtain rad_wo_common_id\n");
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        free(hist_stats_db_buffer);
        return;
    }

    /* Get the histogram statistics */
        status = get_histogram_statistics(open_db, rad_wo_common_id,
        IAS_SPECTRAL_VNIR, band_number, sca_number, num_detectors_per_sca,
        RPS_NUCHAR_OLI_PROC_LEVEL_TEMPCORRECT, hist_stats_db_buffer,
        collection_type);
    assert_equal_with_message(status, SUCCESS, "Error retrieving band %d "
        "SCA %d histogram statistics", band_number, sca_number);
    if (status != SUCCESS)
    {
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        free(hist_stats_db_buffer);
        return;
    }

    /* Check the first and last detector statistics */
    assert_double_equal(hist_stats_db_buffer[0], 1239.452);
    assert_double_equal(hist_stats_db_buffer[num_detectors_per_sca - 1],
                1238.662);


    /* Clean up */
    ias_db_close_connection(open_db);
    ias_db_close_database_lib();
    free(hist_stats_db_buffer);
}




Ensure test_write_to_detector_table()
{
    char sql[512];
    char *path = NULL;

    RPS_NUCHAR_METRICS test_metrics;

    IAS_DB_NULL_TYPE *flagged_rms_detectors = NULL;
    IAS_DB_NULL_TYPE *flagged_stdev_detectors = NULL;
    IAS_DB_NULL_TYPE *flagged_streaking_detectors = NULL;
    int *detector_ids = NULL;

    int num_scas = 14;
    int num_detectors_per_sca = 494;
    int num_band_detectors;
    int num_contiguous_detectors = 100;
    int num_samples;
    int detector_index;
    int sca_index;
    int status;
    char rad_wo_common_id[IAS_CHAR_ID_SIZE];  
    int rms_banding_flag = 1;
    int stdev_banding_flag = 1;
    int streaking_flag = 1;

    double *band_rms_metrics = NULL;
    double *band_stdev_metrics = NULL;
    double *band_streaking_metrics = NULL;
    double *pcnt_rms_metrics = NULL;
    double *pcnt_stdev_metrics = NULL;
    double *db_streaking_metrics = NULL;
    double radiance_mean = 0.0;
    double radiance_stdev = 0.0;

    int *rms_banding_valid_flags = NULL;
    int *stdev_banding_valid_flags = NULL;
    int *streaking_valid_flags = NULL;

    struct ias_db_connection *open_db = NULL;
    struct ias_db_query *query = NULL;

    size_t flagged_id_buffer_size;
    size_t pcnt_metrics_buffer_size;


    printf("Testing write of detector metrics to database...\n");

    /* Get the test data path */
    path = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(path != NULL);
    if (path == NULL)
    {
        printf("Cannot get test data directory location\n");
        return;
    }

    /* Set up test data */
    memset(&test_metrics, 0, sizeof(RPS_NUCHAR_METRICS));

    num_band_detectors = num_scas * num_detectors_per_sca;

    sprintf(full_rad_filename, "%s/%s/%s.txt", path, rad_filedir,
       rad_filename);

    /* Allocate the metrics flagged data buffers */
    rms_banding_valid_flags = calloc(num_band_detectors,
        sizeof(int));
    stdev_banding_valid_flags = calloc(num_band_detectors,
        sizeof(int));
    streaking_valid_flags = calloc(num_band_detectors, sizeof(int));
    if ((rms_banding_valid_flags == NULL)
        || (stdev_banding_valid_flags == NULL)
        || (streaking_valid_flags == NULL))
    {
        assert_true_with_message(0, "Error allocating metrics invalid "
            "flag buffers");
        free(rms_banding_valid_flags);
        free(stdev_banding_valid_flags);
        free(streaking_valid_flags);
        return;
    }

    band_rms_metrics = initialize_band_metrics(full_rad_filename,
        num_band_detectors, num_contiguous_detectors,
        "calculate_rms_banding_metrics", &radiance_mean,
        &radiance_stdev, &num_samples, rms_banding_valid_flags);

    assert_true(band_rms_metrics != NULL);
    if (band_rms_metrics == NULL)
    {
        printf("Cannot calculate band-level RMS banding metrics\n");
        free(rms_banding_valid_flags);
        free(stdev_banding_valid_flags);
        free(streaking_valid_flags);
        return;
    }

    band_stdev_metrics = initialize_band_metrics(full_rad_filename,
            num_band_detectors,
            num_contiguous_detectors,
            "calculate_stdev_banding_metrics",
            &radiance_mean, &radiance_stdev,
            &num_samples, stdev_banding_valid_flags);
    assert_true(band_rms_metrics != NULL);
    if (band_stdev_metrics == NULL)
    {
        printf("Cannot calculate band-level stdev banding metrics\n");
        free(band_rms_metrics);
        free(rms_banding_valid_flags);
        free(stdev_banding_valid_flags);
        free(streaking_valid_flags);
        return;
    }

    band_streaking_metrics = initialize_band_metrics(full_rad_filename,
            num_band_detectors,
            0,
            "calculate_streaking_metrics",
            &radiance_mean, &radiance_stdev,
            &num_samples, streaking_valid_flags);
    assert_true(band_streaking_metrics != NULL);
    if (band_streaking_metrics == NULL)
    {
        printf("Cannot calculate band-level streaking metrics\n");
        free(band_rms_metrics);
        free(band_stdev_metrics);
        free(rms_banding_valid_flags);
        free(stdev_banding_valid_flags);
        free(streaking_valid_flags);
        return;
    }

    /* Initialize the flag data buffers for the current data. */

    status = allocate_metrics_buffers(rms_banding_flag,
            stdev_banding_flag,
            streaking_flag, num_scas,
            num_detectors_per_sca,
            &test_metrics);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot create detector-level metrics data structure "
                "with RMS information\n");
        free(band_rms_metrics);
        free(band_stdev_metrics);
        free(band_streaking_metrics);
        free(rms_banding_valid_flags);
        free(stdev_banding_valid_flags);
        free(streaking_valid_flags);
        return;
    }

    initialize_metrics_data(rms_banding_flag,
                stdev_banding_flag,
                streaking_flag,
                band_rms_metrics,
                band_stdev_metrics,
                band_streaking_metrics,
                radiance_mean,
                radiance_stdev,
                1,
                num_scas,
                num_detectors_per_sca,
                &test_metrics);

    /* Free the band-level data buffers */
    free(band_rms_metrics);
    free(band_stdev_metrics);
    free(band_streaking_metrics);

    /*free_detector_metrics_buffers(&test_metrics, num_scas);*/

    /* Establish database connection */
    open_db = connect_to_db();
    assert_true(open_db != NULL);
    if (open_db == NULL)
    {
        printf("Cannot establish database connection\n");
        ias_db_close_database_lib();
        free_detector_metrics_buffers(&test_metrics, num_scas);
        free(rms_banding_valid_flags);
        free(stdev_banding_valid_flags);
        free(streaking_valid_flags);
        return;
    }

    /* Get the work order common ID number corresponding to the work
       order */
    status = ias_misc_get_char_id(open_db,
            "RADIOMETRIC_WORK_ORDER_COMMON",
            "radiometric_wo_common_id", work_order_id, rad_wo_common_id);

    assert_not_equal(status, ERROR);
    if (status == ERROR)
    {
        printf("Cannot obtain radiometric_wo_common_id value");
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        free_detector_metrics_buffers(&test_metrics, num_scas);
        free(rms_banding_valid_flags);
        free(stdev_banding_valid_flags);
        free(streaking_valid_flags);

        return;
    }

    /* Clear the detector table */
    status = clear_nonuniformity_table(open_db, rad_wo_common_id,
            "NON_UNIFORMITY_DETECTOR");
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot clear NON_UNIFORMITY_DETECTOR table\n");
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        free_detector_metrics_buffers(&test_metrics, num_scas);
        free(rms_banding_valid_flags);
        free(stdev_banding_valid_flags);
        free(streaking_valid_flags);
        return;
    }

    /* Allocate detector ID buffers to contain flag values for
       detectors that weren't/couldn't be characterized */
    flagged_rms_detectors = calloc(num_detectors_per_sca,
            sizeof(*flagged_rms_detectors));
    flagged_stdev_detectors = calloc(num_detectors_per_sca,
            sizeof(*flagged_stdev_detectors));
    flagged_streaking_detectors = calloc(num_detectors_per_sca,
            sizeof(*flagged_streaking_detectors));

    assert_true((flagged_rms_detectors != NULL)
            && (flagged_stdev_detectors != NULL)
            && (flagged_streaking_detectors != NULL));

    if ((flagged_rms_detectors == NULL)
            || (flagged_stdev_detectors == NULL)
            || (flagged_streaking_detectors == NULL))
    {
        printf("Cannot allocate detector flag buffers\n");

        write_detector_table_cleanup(&test_metrics, num_scas,
               open_db, pcnt_rms_metrics, pcnt_stdev_metrics,
               db_streaking_metrics, flagged_rms_detectors,
               flagged_stdev_detectors, flagged_streaking_detectors,
               detector_ids);
        free(rms_banding_valid_flags);
        free(stdev_banding_valid_flags);
        free(streaking_valid_flags);

        return;
    }

    /* Allocate pcnt_XXX_metrics data buffers for corresponding %-of-mean
       radiance values */
    pcnt_rms_metrics = calloc(num_detectors_per_sca,
            sizeof(*pcnt_rms_metrics));
    pcnt_stdev_metrics = calloc(num_detectors_per_sca,
            sizeof(*pcnt_stdev_metrics));
    db_streaking_metrics = calloc(num_detectors_per_sca,
            sizeof(*db_streaking_metrics));

    assert_true((pcnt_rms_metrics != NULL)
            && (pcnt_stdev_metrics != NULL)
            && (db_streaking_metrics != NULL));

    if ((pcnt_rms_metrics == NULL)
            || (pcnt_stdev_metrics == NULL)
            || (db_streaking_metrics == NULL))
    {
        printf("Cannot allocate detector flag buffers\n");

        write_detector_table_cleanup(&test_metrics, num_scas,
               open_db, pcnt_rms_metrics, pcnt_stdev_metrics,
               db_streaking_metrics, flagged_rms_detectors,
               flagged_stdev_detectors, flagged_streaking_detectors,
               detector_ids);
        free(rms_banding_valid_flags);
        free(stdev_banding_valid_flags);
        free(streaking_valid_flags);

        return;
    }

    flagged_id_buffer_size = num_detectors_per_sca * sizeof(int);
    pcnt_metrics_buffer_size = num_detectors_per_sca * sizeof(double);

    /* Allocate a detector id buffer */
    detector_ids = calloc(num_detectors_per_sca, sizeof(*detector_ids));
    assert_true(detector_ids != NULL);
    if (detector_ids == NULL)
    {
        printf("Cannot allocate standard detector ID number buffer\n");

        write_detector_table_cleanup(&test_metrics, num_scas,
               open_db, pcnt_rms_metrics, pcnt_stdev_metrics,
               db_streaking_metrics, flagged_rms_detectors,
               flagged_stdev_detectors, flagged_streaking_detectors,
               detector_ids);
        free(rms_banding_valid_flags);
        free(stdev_banding_valid_flags);
        free(streaking_valid_flags);

        return;
    }
    for (detector_index = 0;
         detector_index < num_detectors_per_sca;
         detector_index++)
    {
        detector_ids[detector_index] = (detector_index + 1);
    }

    /* Write each SCA's metrics to the database */
    for (sca_index = 0; sca_index < num_scas; sca_index++)
    {
        int starting_detector_index = sca_index * num_detectors_per_sca;

        memset(flagged_rms_detectors, 0, flagged_id_buffer_size);
        memset(flagged_stdev_detectors, 0, flagged_id_buffer_size);
        memset(flagged_streaking_detectors, 0, flagged_id_buffer_size);
        memset(pcnt_rms_metrics, 0, pcnt_metrics_buffer_size);
        memset(pcnt_stdev_metrics, 0, pcnt_metrics_buffer_size);
        memset(db_streaking_metrics, 0, pcnt_metrics_buffer_size);

        populate_database_detector_write_buffers(
                &(test_metrics.rms_metrics[sca_index]),
                &(test_metrics.stdev_metrics[sca_index]),
                &(test_metrics.streaking_metrics[sca_index]),
                &(rms_banding_valid_flags[starting_detector_index]),
                &(stdev_banding_valid_flags[starting_detector_index]),
                &(streaking_valid_flags[starting_detector_index]),
                test_metrics.band_radiance_mean,
                rms_banding_flag,
                stdev_banding_flag,
                streaking_flag,
                num_detectors_per_sca,
                flagged_rms_detectors,
                pcnt_rms_metrics,
                flagged_stdev_detectors,
                pcnt_stdev_metrics,
                flagged_streaking_detectors,
                db_streaking_metrics);

        status = write_to_detector_table(open_db, rad_wo_common_id,
                1,
                (sca_index + 1),
                num_detectors_per_sca,
                rms_banding_flag,
                stdev_banding_flag,
                streaking_flag,
                detector_ids,
                flagged_rms_detectors,
                pcnt_rms_metrics,
                flagged_stdev_detectors,
                pcnt_stdev_metrics,
                flagged_streaking_detectors,
                db_streaking_metrics);

        assert_true(status == SUCCESS);
        if (status != SUCCESS)
        {
            printf("Cannot write to database table\n");
            write_detector_table_cleanup(&test_metrics, num_scas,
                   open_db, pcnt_rms_metrics, pcnt_stdev_metrics,
                   db_streaking_metrics, flagged_rms_detectors,
                   flagged_stdev_detectors, flagged_streaking_detectors,
                   detector_ids);
            free(rms_banding_valid_flags);
            free(stdev_banding_valid_flags);
            free(streaking_valid_flags);
            return;
        }
    }

    /* Check #1:  Ensure no 0's are incorrectly written to the
       'banding_metric_stdev' field in the 'non_uniformity_detector'
       table. */
    if (stdev_banding_flag)
    {
        int num_metric_zeros;

        /* Construct the query string and execute the query */
        status = snprintf(sql, sizeof(sql), "SELECT count(*) FROM "
            "non_uniformity_detector WHERE "
            "band_id=1 AND radiometric_wo_common_id=%s AND "
            "banding_metric_stdev=0;", rad_wo_common_id);
        if ((status < 0) || (status >= sizeof(sql)))
        {
            assert_true_with_message(FALSE, "Error constructing SQL "
                "string");
            write_detector_table_cleanup(&test_metrics, num_scas,
                   open_db, pcnt_rms_metrics, pcnt_stdev_metrics,
                   db_streaking_metrics, flagged_rms_detectors,
                   flagged_stdev_detectors, flagged_streaking_detectors,
                   detector_ids);
            free(rms_banding_valid_flags);
            free(stdev_banding_valid_flags);
            free(streaking_valid_flags);
            return;
        }
        query = ias_db_perform_query(open_db, sql);
        if (!ias_db_query_was_successful(query))
        {
            assert_true_with_message(FALSE, "Error executing query");
            write_detector_table_cleanup(&test_metrics, num_scas,
                open_db, pcnt_rms_metrics, pcnt_stdev_metrics,
                db_streaking_metrics, flagged_rms_detectors,
                flagged_stdev_detectors, flagged_streaking_detectors,
                detector_ids);
            free(rms_banding_valid_flags);
            free(stdev_banding_valid_flags);
            free(streaking_valid_flags);
            return;
        }

        /* Go to the first record. */
        if (!ias_db_query_next(query))
        {
            ias_db_query_close(query);
            assert_true_with_message(FALSE, "Error finding first record");
            write_detector_table_cleanup(&test_metrics, num_scas,
                open_db, pcnt_rms_metrics, pcnt_stdev_metrics,
                db_streaking_metrics, flagged_rms_detectors,
                flagged_stdev_detectors, flagged_streaking_detectors,
                detector_ids);
            free(rms_banding_valid_flags);
            free(stdev_banding_valid_flags);
            free(streaking_valid_flags);
            return;
        }

        /* Read the query value */
        status = ias_db_query_int_value(query, &num_metric_zeros, 0);
        ias_db_query_close(query);
        assert_equal_with_message(status, SUCCESS, "Error retrieving "
            "query value");
        if (status != SUCCESS)
        {
            write_detector_table_cleanup(&test_metrics, num_scas,
                open_db, pcnt_rms_metrics, pcnt_stdev_metrics,
                db_streaking_metrics, flagged_rms_detectors,
                flagged_stdev_detectors, flagged_streaking_detectors,
                detector_ids);
            free(rms_banding_valid_flags);
            free(stdev_banding_valid_flags);
            free(streaking_valid_flags);
            return;
        }

        /* Given the data values used here, the query value should be 0 */
        assert_equal_with_message(num_metric_zeros, 0, "Should not be "
            "any 0's in 'banding_metric_stdev', but query returned %d "
            "0's", num_metric_zeros);
    }

    /* Check #2: Ensure values written to the 'banding_metric_rms' and
       'banding_metric_stdev' columns in the 'non_uniformity_detector'
       table are what we expect. */
    if (rms_banding_flag && stdev_banding_flag)
    {
        int rms_diff_count = 0;
        int stdev_diff_count = 0;

        /* Construct the query string and execute the query */
        status = snprintf(sql, sizeof(sql), "SELECT detector_id, "
            "banding_metric_rms, banding_metric_stdev FROM "
            "non_uniformity_detector WHERE "
            "band_id=1 AND sensor_chip_assembly_id=1 AND "
            "radiometric_wo_common_id=%s ORDER BY detector_id;",
            rad_wo_common_id);
        if ((status < 0) || (status >= sizeof(sql)))
        {
            assert_true_with_message(FALSE, "Error constructing SQL "
                "string");
            write_detector_table_cleanup(&test_metrics, num_scas,
                   open_db, pcnt_rms_metrics, pcnt_stdev_metrics,
                   db_streaking_metrics, flagged_rms_detectors,
                   flagged_stdev_detectors, flagged_streaking_detectors,
                   detector_ids);
            free(rms_banding_valid_flags);
            free(stdev_banding_valid_flags);
            free(streaking_valid_flags);
            return;
        }
        query = ias_db_perform_query(open_db, sql);
        if (!ias_db_query_was_successful(query))
        {
            assert_true_with_message(FALSE, "Error executing query");
            write_detector_table_cleanup(&test_metrics, num_scas,
                open_db, pcnt_rms_metrics, pcnt_stdev_metrics,
                db_streaking_metrics, flagged_rms_detectors,
                flagged_stdev_detectors, flagged_streaking_detectors,
                detector_ids);
            free(rms_banding_valid_flags);
            free(stdev_banding_valid_flags);
            free(streaking_valid_flags);
            return;
        }

        /* Regenerate the data buffers written to the database for band 1,
           SCA 1 */
        memset(flagged_rms_detectors, 0, flagged_id_buffer_size);
        memset(flagged_stdev_detectors, 0, flagged_id_buffer_size);
        memset(flagged_streaking_detectors, 0, flagged_id_buffer_size);
        memset(pcnt_rms_metrics, 0, pcnt_metrics_buffer_size);
        memset(pcnt_stdev_metrics, 0, pcnt_metrics_buffer_size);
        memset(db_streaking_metrics, 0, pcnt_metrics_buffer_size);

        populate_database_detector_write_buffers(
            &(test_metrics.rms_metrics[0]), &(test_metrics.stdev_metrics[0]),
            &(test_metrics.streaking_metrics[0]),
            &(rms_banding_valid_flags[0]), &(stdev_banding_valid_flags[0]),
            &(streaking_valid_flags[0]), test_metrics.band_radiance_mean,
            rms_banding_flag, stdev_banding_flag, streaking_flag,
            num_detectors_per_sca, flagged_rms_detectors,
            pcnt_rms_metrics, flagged_stdev_detectors,
            pcnt_stdev_metrics, flagged_streaking_detectors,
            db_streaking_metrics);

        /* Loop through all detectors, checking the values written
           to the database against the original metrics values */
        for (detector_index = 0; detector_index < num_detectors_per_sca;
             detector_index++)
        {
            double rms_metric;
            double stdev_metric;
            int dummy_detector;

            if (!ias_db_query_next(query))
            {
                ias_db_query_close(query);
                assert_true_with_message(FALSE, "Error finding current "
                    "record");
                write_detector_table_cleanup(&test_metrics, num_scas,
                    open_db, pcnt_rms_metrics, pcnt_stdev_metrics,
                    db_streaking_metrics, flagged_rms_detectors,
                    flagged_stdev_detectors, flagged_streaking_detectors,
                    detector_ids);
                free(rms_banding_valid_flags);
                free(stdev_banding_valid_flags);
                free(streaking_valid_flags);
                return;
            }

            status = ias_db_query_int_value(query, &dummy_detector, 0);
            if (status != SUCCESS)
            {
                ias_db_query_close(query);
                assert_true_with_message(FALSE, "Error reading current "
                    "detector ID %d", (detector_index + 1));
                write_detector_table_cleanup(&test_metrics, num_scas,
                    open_db, pcnt_rms_metrics, pcnt_stdev_metrics,
                    db_streaking_metrics, flagged_rms_detectors,
                    flagged_stdev_detectors, flagged_streaking_detectors,
                    detector_ids);
                free(rms_banding_valid_flags);
                free(stdev_banding_valid_flags);
                free(streaking_valid_flags);
                return;
            }

            status = ias_db_query_double_value(query, &rms_metric, 1);
            if (status != SUCCESS)
            {
                ias_db_query_close(query);
                assert_true_with_message(FALSE, "Error reading current "
                    "RMS banding metric value");
                write_detector_table_cleanup(&test_metrics, num_scas,
                    open_db, pcnt_rms_metrics, pcnt_stdev_metrics,
                    db_streaking_metrics, flagged_rms_detectors,
                    flagged_stdev_detectors, flagged_streaking_detectors,
                    detector_ids);
                free(rms_banding_valid_flags);
                free(stdev_banding_valid_flags);
                free(streaking_valid_flags);
                return;
            }

            status = ias_db_query_double_value(query, &stdev_metric, 2);
            if (status != SUCCESS)
            {
                ias_db_query_close(query);
                assert_true_with_message(FALSE, "Error reading current "
                    "STDEV banding metric value");
                write_detector_table_cleanup(&test_metrics, num_scas,
                    open_db, pcnt_rms_metrics, pcnt_stdev_metrics,
                    db_streaking_metrics, flagged_rms_detectors,
                    flagged_stdev_detectors, flagged_streaking_detectors,
                    detector_ids);
                free(rms_banding_valid_flags);
                free(stdev_banding_valid_flags);
                free(streaking_valid_flags);
                return;
            }

            if (rms_metric != pcnt_rms_metrics[detector_index])
                rms_diff_count++;
            if (stdev_metric != pcnt_stdev_metrics[detector_index])
                stdev_diff_count++;
        }

        ias_db_query_close(query);

        /* Check the difference count.  Both should be 0 */
        assert_equal_with_message(rms_diff_count, 0, "%d "
            "'banding_metric_rms' differences found", rms_diff_count);
        assert_equal_with_message(stdev_diff_count, 0, "%d "
            "'banding_metric_stdev' differences found", stdev_diff_count);
    }

    /* Cleanup */
    write_detector_table_cleanup(&test_metrics, num_scas,
                   open_db, pcnt_rms_metrics, pcnt_stdev_metrics,
                   db_streaking_metrics, flagged_rms_detectors,
                   flagged_stdev_detectors, flagged_streaking_detectors,
                   detector_ids);
    free(rms_banding_valid_flags);
    free(stdev_banding_valid_flags);
    free(streaking_valid_flags);
}




Ensure test_write_to_band_table()
{
    char *path = NULL;

    RPS_NUCHAR_METRICS test_metrics;

    double *band_rms_metrics = NULL;
    double *band_stdev_metrics = NULL;
    double *band_streaking_metrics = NULL;
    double radiance_mean = 0.0;
    double radiance_stdev = 0.0;

    int status;
    char rad_wo_common_id[IAS_CHAR_ID_SIZE];  
    int num_band_detectors;
    int num_contiguous_detectors = 100;
    int num_samples;
    int rms_banding_flag = 1;
    int stdev_banding_flag = 1;
    int streaking_flag = 1;
    int num_scas = 14;
    int num_detectors_per_sca = 494;

    struct ias_db_connection *open_db = NULL;

    int *rms_banding_valid_flags = NULL;
    int *stdev_banding_valid_flags = NULL;
    int *streaking_valid_flags = NULL;


    printf("Testing write of band summary metrics to database...\n");

    /* Get the test data path */
    path = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(path != NULL);
    if (path == NULL)
    {
        printf("Cannot get test data directory location\n");
        return;
    }

    /* Set up test data */
    memset(&test_metrics, 0, sizeof(RPS_NUCHAR_METRICS));

    num_band_detectors = num_scas * num_detectors_per_sca;

    sprintf(full_rad_filename, "%s/%s/%s.txt", path, rad_filedir,
       rad_filename);

    /* Allocate the metrics flagged data buffers */
    rms_banding_valid_flags = calloc(num_band_detectors, sizeof(int));
    stdev_banding_valid_flags = calloc(num_band_detectors, sizeof(int));
    streaking_valid_flags = calloc(num_band_detectors, sizeof(int));
    if ((rms_banding_valid_flags == NULL)
        || (stdev_banding_valid_flags == NULL)
        || (streaking_valid_flags == NULL))
    {
        assert_true_with_message(0, "Error allocating metrics invalid "
            "flag buffers");
        free(rms_banding_valid_flags);
        free(stdev_banding_valid_flags);
        free(streaking_valid_flags);
        return;
    }

    band_rms_metrics = initialize_band_metrics(full_rad_filename,
        num_band_detectors, num_contiguous_detectors,
        "calculate_rms_banding_metrics", &radiance_mean,
        &radiance_stdev, &num_samples, rms_banding_valid_flags);

    assert_true(band_rms_metrics != NULL);
    if (band_rms_metrics == NULL)
    {
        printf("Cannot calculate band-level RMS banding metrics\n");
        free(rms_banding_valid_flags);
        free(stdev_banding_valid_flags);
        free(streaking_valid_flags);
        return;
    }

    band_stdev_metrics = initialize_band_metrics(full_rad_filename,
        num_band_detectors, num_contiguous_detectors,
        "calculate_stdev_banding_metrics", &radiance_mean,
        &radiance_stdev, &num_samples, stdev_banding_valid_flags);

    assert_true(band_stdev_metrics != NULL);
    if (band_stdev_metrics == NULL)
    {
        printf("Cannot calculate band-level stdev banding metrics\n");
        free(band_rms_metrics);
        free(rms_banding_valid_flags);
        free(stdev_banding_valid_flags);
        free(streaking_valid_flags);
        return;
    }

    band_streaking_metrics = initialize_band_metrics(full_rad_filename,
        num_band_detectors, 0, "calculate_streaking_metrics",
        &radiance_mean, &radiance_stdev, &num_samples,
        streaking_valid_flags);

    assert_true(band_streaking_metrics != NULL);
    if (band_streaking_metrics == NULL)
    {
        printf("Cannot calculate band-level streaking metrics\n");
        free(band_rms_metrics);
        free(band_stdev_metrics);
        free(rms_banding_valid_flags);
        free(stdev_banding_valid_flags);
        free(streaking_valid_flags);
        return;
    }

    /* Allocate metrics data buffers */
    status = allocate_metrics_buffers(rms_banding_flag,
            stdev_banding_flag,
            streaking_flag, num_scas,
            num_detectors_per_sca,
            &test_metrics);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot create detector-level metrics data structure "
                "with RMS information\n");
        free(band_rms_metrics);
        free(band_stdev_metrics);
        free(band_streaking_metrics);
        return;
    }
    initialize_metrics_data(rms_banding_flag,
                stdev_banding_flag,
                streaking_flag,
                band_rms_metrics,
                band_stdev_metrics,
                band_streaking_metrics,
                radiance_mean,
                radiance_stdev,
                1,
                num_scas,
                num_detectors_per_sca,
                &test_metrics);

    /* Free the band-level data buffers */
    free(band_rms_metrics);
    free(band_stdev_metrics);
    free(band_streaking_metrics);
    

    /* Establish database connection */
    open_db = connect_to_db();
    assert_true(open_db != NULL);
    if (open_db == NULL)
    {
        printf("Cannot establish database connection\n");
        free_detector_metrics_buffers(&test_metrics, num_scas);
        ias_db_close_database_lib();
        return;
    }

    /* Get the work order common ID number corresponding to the work
       order */
    status = ias_misc_get_char_id(open_db,
            "RADIOMETRIC_WORK_ORDER_COMMON",
            "radiometric_wo_common_id", work_order_id, rad_wo_common_id);
    assert_not_equal(status, ERROR);
    if (status == ERROR)
    {
        printf("Cannot obtain radiometric_wo_common_id value");
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        free_detector_metrics_buffers(&test_metrics, num_scas);
        return;
    }

    /* Clear the band table */
    status = clear_nonuniformity_table(open_db, rad_wo_common_id,
                "NON_UNIFORMITY_BAND");
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot clear NON_UNIFORMITY_BAND table\n");
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        free_detector_metrics_buffers(&test_metrics, num_scas);
        return;
    }

    /* Write the band FFOV metric to the database */
    status = write_to_band_table(open_db, rad_wo_common_id,
            1,
            test_metrics.ffov_metric);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
        printf("Cannot write band FFOV metric to database table\n");

    /* Cleanup */
    ias_db_close_connection(open_db);
    ias_db_close_database_lib();
    free_detector_metrics_buffers(&test_metrics, num_scas);
}




/*-----------------------------------------------------------------------
           Define a test suite to add to the "master" test suite
-------------------------------------------------------------------------*/
void nonuniformity_dbaccess_test(TestSuite *suite)
{
    add_test(suite, test_get_histogram_statistics);
    add_test(suite, test_write_to_detector_table);
    add_test(suite, test_write_to_band_table);
}
