/**********************************************************************
                non_uniformity_determine_radiances_test

 PURPOSE:      Use CGREEN to perform unit testing of the algorithm that
               calculates detector radiances for the current band

***********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "ias_const.h"
#include "ias_radiometry.h"
#include "ias_cpf.h"
#include "ias_satellite_attributes.h"

#include "cgreen.h"

#include "../char_nonuniformity.h"

#include "nonuniformity_tests.h"


static char rad_filename[] = "JAB_DetRadiances";
static char rad_filedir[] = "ias/rps/nonuniformity";
static char full_rad_filename[PATH_MAX];


Ensure test_determine_db_radiances()
{
    char *path = NULL;

    int num_detectors_per_sca = 494;
    int num_scas = 14;
    int num_band_detectors;
    int detector_index;
    int dummy_detector_number;
    int sca_index;
    int sca_number;
    int band_number = 1;
    int calculate_radiance = RPS_NUCHAR_USE_DB_RADIANCE;
    int wo_correct_flag = 0;                  /* No correction */
    int status;

    double *stats_buffer = NULL;              /* Histogram/radiance
                                                 statistic data buffer for
                                                 the current SCA */
    double *radiance_buffer = NULL;           /* Radiance data buffer for
                                                 all detectors in the
                                                 current band */
    double dummy_rad_stdev;
    double test_radiance;
    IAS_RPS_OLI_DIFFUSER_TYPE panel_id = IAS_RPS_OLI_DIFFUSER_TYPE_WORKING;

    FILE *rad_fptr = NULL;

    const struct IAS_CPF_ABSOLUTE_GAINS *cpf_abs_gains = NULL;
    const struct IAS_CPF_RELATIVE_GAINS *cpf_relative_gains = NULL;
    const struct IAS_CPF_NONUNIFORMITY *cpf_nonuniform = NULL;

    int *radiance_flags = NULL;


    printf("Simulating database radiance determination...\n");

    /* Get the test data path */
    path = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(path != NULL);
    if (path == NULL)
    {
        printf("Cannot get test data directory location\n");
        return;
    }

    /* Allocate data buffers */
    num_band_detectors = num_scas * num_detectors_per_sca;

    radiance_buffer = calloc(num_band_detectors, sizeof(*radiance_buffer));
    stats_buffer = malloc(num_detectors_per_sca * sizeof(*stats_buffer));
    radiance_flags = calloc(num_band_detectors, sizeof(*radiance_flags));

    assert_true((stats_buffer != NULL) && (radiance_buffer != NULL)
        && (radiance_flags != NULL));
    if ((stats_buffer == NULL) || (radiance_buffer == NULL)
        || (radiance_flags == NULL))
    {
        printf("Cannot allocate test data buffers\n");
        free(stats_buffer);
        free(radiance_buffer);
        free(radiance_flags);
        return;
    }

    /*  Test radiance data "retrieved" from the characterization
        database (simulating a get_histogram_statistics call)*/
    sprintf(full_rad_filename, "%s/%s/%s.txt", path, rad_filedir,
       rad_filename);

    rad_fptr = fopen(full_rad_filename, "r");
    assert_true(rad_fptr != NULL);
    if (rad_fptr == NULL)
    {
        printf("Cannot open/read radiance data file %s\n", full_rad_filename);
        free(stats_buffer);
        free(radiance_buffer);
        free(radiance_flags);
        return;
    }

    for (sca_index = 0; sca_index < num_scas; sca_index++)
    {
        sca_number = (sca_index + 1);

        /* Populate the statistics buffer with the current SCA's
           radiances */
        for (detector_index = 0;
             detector_index < num_detectors_per_sca;
             detector_index++)
        {
            if (fscanf(rad_fptr, "%d%lf%lf\n", &dummy_detector_number,
                   &stats_buffer[detector_index], &dummy_rad_stdev) < 0)
            {
                printf("Cannot read data for detector %d\n",
                       (detector_index + 1));
                fclose(rad_fptr);
                free(stats_buffer);
                free(radiance_buffer);
                free(radiance_flags);
                return;
            }
        }

        /* "Determine" radiances (copy option from the database); the
           default option to convert bias-corrected/linearized histogram
           statistics is implicitly tested for in the end-to-end test
           (nonuniformity_end2end_test; work order specifies internal
           calculation of radiance).  Flagging of inoperable and/or
           out-of-spec detectors is before the histogram statistics are
           retrieved from the database, using either an available pixel
           mask file or a local pixel mask for the current band/SCA
           created "on-the-fly."  from the CPF lists.  Assume for this
           scenario that there aren't any out-of-spec or inoperable
           detectors (nothing's flagged).  Pass in the band and SCA indices,
           rather than the band and SCA numbers.  Assume we're doing this
           for a solar collect */
        int sca_index = sca_number - 1;
        int band_index =
            ias_sat_attr_convert_band_number_to_index(band_number);
        if (band_index == ERROR)
        {
            assert_true_with_message(FALSE,
                "Error calculating true index for band number %d",
                band_number);
            fclose(rad_fptr);
            free(stats_buffer);
            free(radiance_buffer);
            free(radiance_flags);
            return;
        }
        status = determine_radiances(stats_buffer, band_index,
                                     sca_index, num_detectors_per_sca,
                                     cpf_abs_gains, cpf_relative_gains,
                                     cpf_nonuniform,
                                     calculate_radiance, wo_correct_flag,
                                     panel_id, radiance_flags,
                                     radiance_buffer);
        assert_true(status == SUCCESS);
        if (status != SUCCESS)
        {
            printf("determine_radiances DB radiance option failed\n");
            fclose(rad_fptr);
            free(stats_buffer);
            free(radiance_buffer);
            free(radiance_flags);
            return;
        }
    }

    /* Rewind the file pointer */
    rewind(rad_fptr);

    /* Read the file again, this time comparing the values read off to
       those in radiance_buffer.  We're making sure the mechanics for
       copying the radiances directly from the SCA data buffer to the
       band data buffer are working as expected */
    for (detector_index = 0;
         detector_index < num_band_detectors;
         detector_index++)
    {
        if (fscanf(rad_fptr, "%d%lf%lf\n", &dummy_detector_number,
               &test_radiance, &dummy_rad_stdev) < 0)
        {
            printf("Cannot read detector %d data from file\n",
                   (detector_index + 1));
            fclose(rad_fptr);
            free(stats_buffer);
            free(radiance_buffer);
            free(radiance_flags);
            return;
        }

        assert_double_equal(test_radiance, radiance_buffer[detector_index]);
    }
    fclose(rad_fptr);

    /* Clean up */
    free(stats_buffer);
    free(radiance_buffer);
    free(radiance_flags);
}




Ensure test_correct_nonuniform_panel_response()
{
    char *path = NULL;

    int num_detectors_per_sca = 494;
    int num_scas = 14;
    int num_band_detectors;
    int detector_index;
    int sca_index;
    int start_detector_index;
    int dummy_detector_index;
    int status;

    double *data_buffer_nocorrect = NULL;
    double *data_buffer_correct = NULL;
    double *correction_factor_buffer = NULL;
    double *cf_ptr = NULL;
    double dummy_rad_stdev;
    double test_corrected_radiance;

    FILE *rad_fptr = NULL;

    int *radiance_flags = NULL;

    printf("Testing source correction algorithm...\n");

    /* Get the test data path */
    path = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(path != NULL);
    if (path == NULL)
    {
        printf("Cannot get test data directory location\n");
        return;
    }

    /* Allocate a data buffer and a correction data buffer */
    num_band_detectors = num_scas * num_detectors_per_sca;

    data_buffer_nocorrect = calloc(num_band_detectors,
            sizeof(*data_buffer_nocorrect));

    data_buffer_correct = calloc(num_band_detectors,
            sizeof(*data_buffer_correct));

    correction_factor_buffer = calloc(num_band_detectors,
            sizeof(*correction_factor_buffer));

    radiance_flags = calloc(num_band_detectors, sizeof(*radiance_flags));

    assert_true((data_buffer_nocorrect != NULL)
            && (data_buffer_correct != NULL)
            && (correction_factor_buffer != NULL)
            && (radiance_flags != NULL));

    if ((data_buffer_nocorrect == NULL)
            || (data_buffer_correct == NULL)
            || (correction_factor_buffer == NULL)
            || (radiance_flags == NULL))
    {
        printf("Cannot allocate data and/or correction factor buffers\n");
        free(data_buffer_nocorrect);
        free(data_buffer_correct);
        free(correction_factor_buffer);
        free(radiance_flags);
        return;
    }

    /* Read the test radiance data file */
    sprintf(full_rad_filename, "%s/%s/%s.txt", path, rad_filedir,
       rad_filename);

    rad_fptr = fopen(full_rad_filename, "r");
    assert_true(rad_fptr != NULL);
    if (rad_fptr == NULL)
    {
        printf("Cannot open/read test radiance data file %s\n",
                full_rad_filename);
        free(data_buffer_nocorrect);
        free(data_buffer_correct);
        free(correction_factor_buffer);
        free(radiance_flags);
        return;
    }

    /* Populate these with some values */
    for (detector_index = 0;
         detector_index < num_band_detectors;
         detector_index++)
    {
        if (fscanf(rad_fptr, "%d%lf%lf\n", &dummy_detector_index,
                    &data_buffer_nocorrect[detector_index],
                    &dummy_rad_stdev) < 0)
        {
            printf("Read of data for detector # %d failed\n",
                    (detector_index + 1));
            free(data_buffer_correct);
            free(data_buffer_nocorrect);
            free(correction_factor_buffer);
            free(radiance_flags);
            return;
        }

        data_buffer_correct[detector_index]
            = data_buffer_nocorrect[detector_index];

        if ((detector_index % 2) == 0)
            correction_factor_buffer[detector_index] = 1.5;
        else
            correction_factor_buffer[detector_index] = 1.3;
    }
    fclose(rad_fptr);

    /* Do the correction, 1 SCA at a time */
    for (sca_index = 0; sca_index < num_scas; sca_index++)
    {
        start_detector_index = sca_index * num_detectors_per_sca;

        cf_ptr = &(correction_factor_buffer[start_detector_index]);

        status = correct_nonuniform_panel_response(data_buffer_correct,
            radiance_flags, cf_ptr, start_detector_index,
            num_detectors_per_sca);
        assert_true(status == SUCCESS);
        if (status != SUCCESS)
        {
            printf("Correction algorithm failed for SCA %d\n",
                    (sca_index + 1));
            free(data_buffer_correct);
            free(data_buffer_nocorrect);
            free(correction_factor_buffer);
            free(radiance_flags);
            return;
        }
    }

    /* Validate the corrected data.  The data themselves are made-up
       values--the point of the test is to check the mechanics of
       the algorithm implementation, particularly whether the offsets
       into the band-level data arrays were determined correctly */
    for (detector_index = 0;
         detector_index < num_band_detectors;
         detector_index++)
    {
        test_corrected_radiance = data_buffer_nocorrect[detector_index]
            / correction_factor_buffer[detector_index];

        assert_double_equal(data_buffer_correct[detector_index],
                test_corrected_radiance);
    }

    /* Clean up */
    free(data_buffer_correct);
    free(data_buffer_nocorrect);
    free(correction_factor_buffer);
    free(radiance_flags);
}




/*-----------------------------------------------------------------------
           Define a test suite to add to the "master" test suite
-------------------------------------------------------------------------*/
void nonuniformity_determine_radiances_test(TestSuite *suite)
{
    add_test(suite, test_determine_db_radiances);
    add_test(suite, test_correct_nonuniform_panel_response);
}
