/***************************************************************************
 NAME:      nonuniformity_end2end_test

 PURPOSE:   Performs CGreen unit testing for end-to-end testing of the
            char_nonuniformity application

 ROUTINES:  test_oli_end2end_nonuniformity
            test_tirs_end2end_nonuniformity
            test_earth_end2end_nonuniformity
****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "ias_const.h"
#include "ias_logging.h"
#include "../char_nonuniformity.h"
#include "cgreen.h"


static char test_reportfile_name[PATH_MAX];
static char reference_reportfile_name[PATH_MAX];
static char test_dumpfile_name[PATH_MAX];
static char reference_dumpfile_name[PATH_MAX];
static char syscommand[1000];
static char work_order_id[IAS_WORKORDER_ID_SIZE];
static char *test_data_path = NULL;
static char *enable_db_tests = NULL;


Ensure test_end2end_oli_nonuniformity()
{
    struct stat buff;
    char param_file[PATH_MAX];
    int band_list[3] = {1, 7, 8};
    int status;
    int num_bands_processed = 3;
    int num_dumpfiles = 0;
    int num_reportfiles = 0;
    int index;


    printf("Running end-to-end OLI nonuniformity characterization...\n");

    /* Make sure the test data for the work order is in the database, 
       and if it isn't, load it */
    strcpy(work_order_id, "NUT24000");
    status = snprintf(syscommand, sizeof(syscommand),
        "./check_or_insert_end2end_woid.csh %s OLI", work_order_id);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(FALSE, "Error constructing command line "
            "to check for/insert simulated data into database...");
        return;
    }
    status = system(syscommand);
    if (status != 0)
    {
        assert_true_with_message(FALSE, "Error running script to check for "
            "insert simulated data into database");
        return;
    }

    /* Replace references to $IAS_UNIT_TEST_DATA_DIR with the correct
       test directory location */
    status = snprintf(syscommand, sizeof(syscommand),
        "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' %s_template.omf > %s.omf",
        test_data_path, work_order_id, work_order_id);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(FALSE, "Error creating command to "
           "replace test data directory placeholder");
        return;
    }
    status = system(syscommand);
    if (status != 0)
    {
        assert_true_with_message(FALSE, "Error replacing test data "
            "directory placeholder");
        return;
    }

    /* Set up the command line to run the characterization */
    sprintf(param_file, "%s.odl", work_order_id);
    status = snprintf(syscommand, sizeof(syscommand),
        "../char_nonuniformity %s", param_file);
    if (status < 0)
    {
        assert_true_with_message(FALSE, "Error constructing command line "
            "to run OLI nonuniformity characterization");
        return;
    }

    /* Run the characterization */
    status = system(syscommand);
    if (status != 0)
    {
        assert_true_with_message(FALSE, "End-to-end run of OLI nonuniformity "
            "characterization failed");
        return;
    }

    /* Check the results.  The work order was specified to generate an
       overall summary report, along with dump files for each band
       processed, along with putting results in the database */

    /* Check 1 -- count the number of dump files produced, the number
       of summary report files produced, and the number of "header" files
       produced.  In the work order, 3 bands
       (1, 7, and 8) are specified in the band list.  There should be
       3 dump files and 3 report files (one for each band), and 1
       "header" file */
    for (index = 0; index < num_bands_processed; index++)
    {
        snprintf(test_reportfile_name, sizeof(test_reportfile_name),
            "nonuniformity_summary_B%02d.txt", 
            band_list[index]);

        status = stat(test_reportfile_name, &buff);
        if (status == 0)
            num_reportfiles++;
        snprintf(test_dumpfile_name, sizeof(test_dumpfile_name),
            "nonuniformity_dumpfile_B%02d.txt",
            band_list[index]);
        status = stat(test_dumpfile_name, &buff);
        if (status == 0)
            num_dumpfiles++;
    }
    assert_equal_with_message(num_reportfiles, num_bands_processed,
        "Number of summary report files found %d does not match number of "
        "bands initially processed %d", num_reportfiles, num_bands_processed);
    if (num_reportfiles != num_bands_processed)
        return;
    assert_equal_with_message(num_dumpfiles, num_bands_processed,
        "Number of dump files found %d does not match number of "
        "bands initially processed %d", num_dumpfiles, num_bands_processed);
    if (num_dumpfiles != num_bands_processed)
        return;
    snprintf(test_reportfile_name, sizeof(test_reportfile_name),
        "nonuniformity_header.txt");
    status = stat(test_reportfile_name, &buff);
    if (status != 0)
    {
        assert_true_with_message(FALSE, "Cannot find 'header' file %s",
            test_reportfile_name);
        return;
    }

    /* Check 2 -- check the contents of the header file against a
       "reference".  Do not compare the first lines as those contain
       a timestamp indicating when the file was created */
    snprintf(test_reportfile_name, sizeof(test_reportfile_name),
        "nonuniformity_header.txt");
    snprintf(reference_reportfile_name, sizeof(reference_reportfile_name),
        "%s_ref_nonuniformity_header.txt", work_order_id);
    status = snprintf(syscommand, sizeof(syscommand),
        "sed '/Spacecraft:/,$!'d %s | diff -q - %s",
        test_reportfile_name, reference_reportfile_name);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(FALSE, "Error preparing difference "
            "comparison system command");
        return;
    }
    status = system(syscommand);
    assert_equal_with_message(0, status, "Differences found between test "
        "nonuniformity header file %s and reference header file %s",
        test_reportfile_name, reference_reportfile_name);
 
    /* Check 3 -- compare the contents of the test summary report and
       dump files against the correpsonding "reference" files */
    for (index = 0; index < num_bands_processed; index++)
    {
        snprintf(test_reportfile_name, sizeof(test_reportfile_name),
            "nonuniformity_summary_B%02d.txt", 
            band_list[index]);
        snprintf(reference_reportfile_name, sizeof(reference_reportfile_name),
            "%s_ref_nonuniformity_summary_B%02d.txt", work_order_id,
            band_list[index]);
        status = snprintf(syscommand, sizeof(syscommand),
            "diff %s %s", reference_reportfile_name, test_reportfile_name);
        if ((status < 0) || (status >= sizeof(syscommand)))
        {
            assert_true_with_message(FALSE, "Error preparing summary "
                "band %d report file comparison system command",
                band_list[index]);
            return;
        }
        status = system(syscommand);
        assert_equal_with_message(status, 0, "Error comparing reference and "
            "test report files for band %d", band_list[index]);

        snprintf(test_dumpfile_name, sizeof(test_dumpfile_name),
            "nonuniformity_dumpfile_B%02d.txt",
            band_list[index]);
        snprintf(reference_dumpfile_name, sizeof(reference_dumpfile_name),
            "%s_ref_nonuniformity_dumpfile_B%02d.txt", work_order_id,
            band_list[index]);
        status = snprintf(syscommand, sizeof(syscommand),
            "diff %s %s", reference_dumpfile_name, test_dumpfile_name);
        if ((status < 0) || (status >= sizeof(syscommand)))
        {
            assert_true_with_message(FALSE, "Error preparing summary "
                "band %d dump file comparison system command",
                band_list[index]);
            return;
        }
        status = system(syscommand);
        assert_equal_with_message(status, 0, "Error comparing reference and "
            "test dump files for band %d", band_list[index]);
    }

    /* Cleanup */
    status = snprintf(syscommand, sizeof(syscommand), "rm %s.omf "
        "nonuniformity_summary_B*.txt nonuniformity_dumpfile_B*.txt "
        "nonuniformity_header.txt", work_order_id);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(FALSE, "Error creating command to rm "
            "test data output files");
    }
    else
    {
        status = system(syscommand);
        if (status != 0)
        {
            assert_true_with_message(FALSE, "Error removing test data "
                "output files");
        }
    }
}   /* END -- test_end2end_oli_nonuniformity */

Ensure test_end2end_tirs_nonuniformity()
{
    struct stat buff;
    char param_file[PATH_MAX];
    int status;
    int band_list[2] = {10, 11};
    int num_bands_processed = 2;
    int num_dumpfiles = 0;
    int num_reportfiles = 0;
    int index;


    printf("Running end-to-end TIRS nonuniformity characterization...\n");

    /* Make sure the test data for the work order is in the database, 
       and if it isn't, load it */
    strcpy(work_order_id, "NUT26000");
    status = snprintf(syscommand, sizeof(syscommand),
        "./check_or_insert_end2end_woid.csh %s TIRS", work_order_id);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(FALSE, "Error constructing command line "
            "to check for/insert simulated data into database...");
        return;
    }
    status = system(syscommand);
    if (status != 0)
    {
        assert_true_with_message(FALSE, "Error running script to check for/"
            "insert simulated data into database");
        return;
    }

    /* Replace references to $IAS_UNIT_TEST_DATA_DIR with the correct
       test directory location */
    status = snprintf(syscommand, sizeof(syscommand),
        "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' %s_template.omf > %s.omf",
        test_data_path, work_order_id, work_order_id);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(FALSE, "Error creating command to "
           "replace test data directory placeholder");
        return;
    }
    status = system(syscommand);
    if (status != 0)
    {
        assert_true_with_message(FALSE, "Error replacing test data "
            "directory placeholder");
        return;
    }

    /* Set up the command line to run the characterization */
    sprintf(param_file, "%s.odl", work_order_id);
    status = snprintf(syscommand, sizeof(syscommand),
        "../char_nonuniformity %s", param_file);
    if (status < 0)
    {
        assert_true_with_message(FALSE, "Error constructing command line "
            "to run TIRS nonuniformity characterization");
        return;
    }

    /* Run the characterization */
    status = system(syscommand);
    if (status != 0)
    {
        assert_true_with_message(FALSE, "End-to-end run of TIRS nonuniformity "
            "characterization failed");
        return;
    }

    /* Check the results.  The work order was specified to generate an
       overall summary report, along with dump files for each band
       processed, along with putting results in the database */

    /* Check 1 -- count the number of dump files produced, the number
       of summary report files produced, and the number of "header" files
       produced.  In the work order, 3 bands
       (1, 7, and 8) are specified in the band list.  There should be
       3 dump files and 3 report files (one for each band), and 1
       "header" file */
    for (index = 0; index < num_bands_processed; index++)
    {
        snprintf(test_reportfile_name, sizeof(test_reportfile_name),
            "nonuniformity_summary_B%02d.txt", 
            band_list[index]);
        status = stat(test_reportfile_name, &buff);
        if (status == 0)
            num_reportfiles++;
        snprintf(test_dumpfile_name, sizeof(test_dumpfile_name),
            "nonuniformity_dumpfile_B%02d.txt",
            band_list[index]);
        status = stat(test_dumpfile_name, &buff);
        if (status == 0)
            num_dumpfiles++;
    }
    assert_equal_with_message(num_reportfiles, num_bands_processed,
        "Number of summary report files found %d does not match number of "
        "bands initially processed %d", num_reportfiles, num_bands_processed);
    if (num_reportfiles != num_bands_processed)
        return;
    assert_equal_with_message(num_dumpfiles, num_bands_processed,
        "Number of dump files found %d does not match number of "
        "bands initially processed %d", num_dumpfiles, num_bands_processed);
    if (num_dumpfiles != num_bands_processed)
        return;
    snprintf(test_reportfile_name, sizeof(test_reportfile_name),
        "nonuniformity_header.txt");
    status = stat(test_reportfile_name, &buff);
    if (status != 0)
    {
        assert_true_with_message(FALSE, "Cannot find 'header' file %s",
            test_reportfile_name);
        return;
    }

    /* Check 2 -- check the contents of the header file against a
       "reference".  Do not compare the first lines as those contain
       a timestamp indicating when the file was created */
    snprintf(test_reportfile_name, sizeof(test_reportfile_name),
        "nonuniformity_header.txt");
    snprintf(reference_reportfile_name, sizeof(reference_reportfile_name),
        "%s_ref_nonuniformity_header.txt", work_order_id);
    status = snprintf(syscommand, sizeof(syscommand),
        "sed '/Spacecraft:/,$!'d %s | diff -q - %s",
        test_reportfile_name, reference_reportfile_name);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(FALSE, "Error preparing difference "
            "comparison system command");
        return;
    }
    status = system(syscommand);
    assert_equal_with_message(0, status, "Differences found between test "
        "nonuniformity header file %s and reference header file %s",
        test_reportfile_name, reference_reportfile_name);
 
    /* Check 3 -- compare the contents of the test summary report and
       dump files against the correpsonding "reference" files */
    for (index = 0; index < num_bands_processed; index++)
    {
        snprintf(test_reportfile_name, sizeof(test_reportfile_name),
            "nonuniformity_summary_B%02d.txt", 
            band_list[index]);
        snprintf(reference_reportfile_name, sizeof(reference_reportfile_name),
            "%s_ref_nonuniformity_summary_B%02d.txt", work_order_id,
            band_list[index]);
        status = snprintf(syscommand, sizeof(syscommand),
            "diff %s %s", reference_reportfile_name, test_reportfile_name);
        if ((status < 0) || (status >= sizeof(syscommand)))
        {
            assert_true_with_message(FALSE, "Error preparing summary "
                "band %d report file comparison system command",
                band_list[index]);
            return;
        }
        status = system(syscommand);
        assert_equal_with_message(status, 0, "Error comparing reference and "
            "test report files for band %d", band_list[index]);

        snprintf(test_dumpfile_name, sizeof(test_dumpfile_name),
            "nonuniformity_dumpfile_B%02d.txt",
            band_list[index]);
        snprintf(reference_dumpfile_name, sizeof(reference_dumpfile_name),
            "%s_ref_nonuniformity_dumpfile_B%02d.txt", work_order_id,
            band_list[index]);
        status = snprintf(syscommand, sizeof(syscommand),
            "diff %s %s", reference_dumpfile_name, test_dumpfile_name);
        if ((status < 0) || (status >= sizeof(syscommand)))
        {
            assert_true_with_message(FALSE, "Error preparing summary "
                "band %d dump file comparison system command",
                band_list[index]);
            return;
        }
        status = system(syscommand);
        assert_equal_with_message(status, 0, "Error comparing reference and "
            "test dump files for band %d", band_list[index]);
    }

    /* Cleanup */
    status = snprintf(syscommand, sizeof(syscommand), "rm %s.omf "
        "nonuniformity_summary_B*.txt nonuniformity_dumpfile_B*.txt "
        "nonuniformity_header.txt", work_order_id);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(FALSE, "Error creating command to rm "
            "test data output files");
    }
    else
    {
        status = system(syscommand);
        if (status != 0)
        {
            assert_true_with_message(FALSE, "Error removing test data "
                "output files");
        }
    }
}   /* END -- test_end2end_tirs_nonuniformity */


Ensure test_end2end_earth_nonuniformity()
{
    struct stat buff;
    char param_file[PATH_MAX];
    int status;
    int band_list[5] = {1, 7, 8, 10, 11};
    int num_bands_processed = 5;
    int num_dumpfiles = 0;
    int num_reportfiles = 0;
    int index;


    printf("Running end-to-end 'Earth' image nonuniformity "
        "characterization...\n");

    /* Make sure the test data for the work order is in the database, 
       and if it isn't, load it */
    strcpy(work_order_id, "NUT28000");
    status = snprintf(syscommand, sizeof(syscommand),
        "./check_or_insert_end2end_earthwoid.csh %s OLITIRS", work_order_id);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(FALSE, "Error constructing command line "
            "to check for/insert simulated data into database...");
        return;
    }
    status = system(syscommand);
    if (status != 0)
    {
        assert_true_with_message(FALSE, "Error running script to check for/"
            "insert simulated data into database");
        return;
    }

    /* Replace references to $IAS_UNIT_TEST_DATA_DIR (and
       $IAS_UNIT_TEST_DATA_DIR if present) with the correct
       test directory location */
    status = snprintf(syscommand, sizeof(syscommand),
        "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' %s_template.omf > %s.omf",
        test_data_path, work_order_id, work_order_id);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(FALSE, "Error creating command to "
           "replace test data directory placeholder");
        return;
    }
    status = system(syscommand);
    if (status != 0)
    {
        assert_true_with_message(FALSE, "Error replacing test data "
            "directory placeholder");
        return;
    }

    /* Set up the command line to run the characterization */
    sprintf(param_file, "%s.odl", work_order_id);
    status = snprintf(syscommand, sizeof(syscommand),
        "../char_nonuniformity %s", param_file);
    if (status < 0)
    {
        assert_true_with_message(FALSE, "Error constructing command line "
            "to run Earth nonuniformity characterization");
        return;
    }

    /* Run the characterization */
    status = system(syscommand);
    if (status != 0)
    {
        assert_true_with_message(FALSE, "End-to-end run of Earth "
            "nonuniformity characterization failed");
        return;
    }

    /* Check the results.  The work order was specified to generate an
       overall summary report, along with dump files for each band
       processed, along with putting results in the database */

    /* Check 1 -- count the number of dump files produced, the number
       of summary report files produced, and the number of "header" files
       produced.  In the work order, 3 bands
       (1, 7, and 8) are specified in the band list.  There should be
       3 dump files and 3 report files (one for each band), and 1
       "header" file */
    for (index = 0; index < num_bands_processed; index++)
    {
        snprintf(test_reportfile_name, sizeof(test_reportfile_name),
            "nonuniformity_summary_B%02d.txt", 
            band_list[index]);
        status = stat(test_reportfile_name, &buff);
        if (status == 0)
            num_reportfiles++;
        snprintf(test_dumpfile_name, sizeof(test_dumpfile_name),
            "nonuniformity_dumpfile_B%02d.txt",
            band_list[index]);
        status = stat(test_dumpfile_name, &buff);
        if (status == 0)
            num_dumpfiles++;
    }
    assert_equal_with_message(num_reportfiles, num_bands_processed,
        "Number of summary report files found %d does not match number of "
        "bands initially processed %d", num_reportfiles, num_bands_processed);
    if (num_reportfiles != num_bands_processed)
        return;
    assert_equal_with_message(num_dumpfiles, num_bands_processed,
        "Number of dump files found %d does not match number of "
        "bands initially processed %d", num_dumpfiles, num_bands_processed);
    if (num_dumpfiles != num_bands_processed)
        return;
    snprintf(test_reportfile_name, sizeof(test_reportfile_name),
        "nonuniformity_header.txt");
    status = stat(test_reportfile_name, &buff);
    if (status != 0)
    {
        assert_true_with_message(FALSE, "Cannot find 'header' file %s",
            test_reportfile_name);
        return;
    }

    /* Check 2 -- check the contents of the header file against a
       "reference".  Do not compare the first lines as those contain
       a timestamp indicating when the file was created */
    snprintf(test_reportfile_name, sizeof(test_reportfile_name),
        "nonuniformity_header.txt");
    snprintf(reference_reportfile_name, sizeof(reference_reportfile_name),
        "%s_ref_nonuniformity_header.txt", work_order_id);
    status = snprintf(syscommand, sizeof(syscommand),
        "sed '/Spacecraft:/,$!'d %s | diff -q - %s",
        test_reportfile_name, reference_reportfile_name);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(FALSE, "Error preparing difference "
            "comparison system command");
        return;
    }
    status = system(syscommand);
    assert_equal_with_message(0, status, "Differences found between test "
        "nonuniformity header file %s and reference header file %s",
        test_reportfile_name, reference_reportfile_name);
 
    /* Check 3 -- compare the contents of the test summary report and
       dump files against the correpsonding "reference" files */
    for (index = 0; index < num_bands_processed; index++)
    {
        snprintf(test_reportfile_name, sizeof(test_reportfile_name),
            "nonuniformity_summary_B%02d.txt", 
            band_list[index]);
        snprintf(reference_reportfile_name, sizeof(reference_reportfile_name),
            "%s_ref_nonuniformity_summary_B%02d.txt", work_order_id,
            band_list[index]);
        status = snprintf(syscommand, sizeof(syscommand),
            "diff %s %s", reference_reportfile_name, test_reportfile_name);
        if ((status < 0) || (status >= sizeof(syscommand)))
        {
            assert_true_with_message(FALSE, "Error preparing summary "
                "band %d report file comparison system command",
                band_list[index]);
            return;
        }
        status = system(syscommand);
        assert_equal_with_message(status, 0, "Error comparing reference and "
            "test report files for band %d", band_list[index]);

        snprintf(test_dumpfile_name, sizeof(test_dumpfile_name),
            "nonuniformity_dumpfile_B%02d.txt",
            band_list[index]);
        snprintf(reference_dumpfile_name, sizeof(reference_dumpfile_name),
            "%s_ref_nonuniformity_dumpfile_B%02d.txt", work_order_id,
            band_list[index]);
        status = snprintf(syscommand, sizeof(syscommand),
            "diff %s %s", reference_dumpfile_name, test_dumpfile_name);
        if ((status < 0) || (status >= sizeof(syscommand)))
        {
            assert_true_with_message(FALSE, "Error preparing summary "
                "band %d dump file comparison system command",
                band_list[index]);
            return;
        }
        status = system(syscommand);
        assert_equal_with_message(status, 0, "Error comparing reference and "
            "test dump files for band %d", band_list[index]);
    }

    /* Cleanup */
    status = snprintf(syscommand, sizeof(syscommand), "rm %s.omf "
        "nonuniformity_summary_B*.txt nonuniformity_dumpfile_B*.txt "
        "nonuniformity_header.txt", work_order_id);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(FALSE, "Error creating command to rm "
            "test data output files");
    }
    else
    {
        status = system(syscommand);
        if (status != 0)
        {
            assert_true_with_message(FALSE, "Error removing test data "
                "output files");
        }
    }
}   /* END -- test_end2end_earth_nonuniformity */


int main(int argc, char *argv[])
{
    /* Get the test data path */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path ==  NULL)
    {
        printf("Cannot locate IAS test data location\n");
        exit(1);
    }

    /* Check to see if database support has been set in the environment.
       If it isn't, we can't run the test because we need to access
       histogram statistics from the database.  Use exit(0) so that
       it doesn't come across as a test 'failure'*/
    enable_db_tests = getenv("ENABLE_DB_TESTS");
    if (enable_db_tests == NULL)
    {
        printf("Cannot run end-to-end test of nonuniformity "
            "characterization--database support not enabled\n");
        exit(0);
    }

    /* Create the test suite.  Tests should be implemented to check
       end-to-end processing for OLI AND TIRS */
    TestSuite *suite = create_test_suite();
    add_test(suite, test_end2end_oli_nonuniformity);
    add_test(suite, test_end2end_tirs_nonuniformity);
    add_test(suite, test_end2end_earth_nonuniformity);

    /* Run the test(s) */
    if (argc > 1)
        return run_single_test(suite, argv[1], create_text_reporter());
    return run_test_suite(suite, create_text_reporter());
}   /* END -- main */
