/**********************************************************************
 NAME:                non_uniformity_read_parameters_test

 PURPOSE:      Use CGREEN to perform unit testing on the read_parameters
               routine implemented for the nonuniformity characterization
               algorithm

***********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "ias_const.h"

#include "cgreen.h"

#include "../char_nonuniformity.h"

#include "nonuniformity_tests.h"

/*#define DEBUG_PRINT*/


Ensure test_read_parameters()
{
    char *test_data_path = NULL;
    char l0r_filename[PATH_MAX] = "";
    char cpf_filename[PATH_MAX] = "";
    char work_order_id[IAS_WORKORDER_ID_SIZE] = "NUT1002";
    char l0r_data_dir[] = "ias_lib/misc/l0ra_location";
    char l0r_name[] = "test";
    char cpf_data_dir[] = "ias_lib/cpf";
    char cpf_name[] = "L8CPF20090101_20090331.01";
    char wo_params_filename[PATH_MAX] = "";
    char system_command[1000] = "";
    char pmask_filename[] = "masks.pmask";

    int status = SUCCESS;
    int size;

    RPS_NUCHAR_WO_PARAMS wo_params;
    RPS_NUCHAR_OMF_PARAMS omf_params;

    /* Testing OLI characterization */
    int test_band_list[4] = {1, 2, 5, 8};
    int ii = 0;          /* Local loop counter */


    printf("Testing read of work order and parameter files...\n");

    /* Get the test data path */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(test_data_path != NULL);
    if (test_data_path == NULL)
    {
        printf("Cannot get test data directory location\n");
        return;
    }

    /* Replace the $IAS_UNIT_TEST_DATA_DIR placeholder in the OMF file with
       its actual value */
    size = snprintf(system_command, sizeof(system_command), 
        "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' %s_template.omf > %s.omf", 
        test_data_path, work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return;
    }
    status = system(system_command);
    assert_equal_with_message(status, FALSE, "Cannot edit OMF template");

    /* Rename the template ODL file as the actual ODL file */
    size = snprintf(system_command, sizeof(system_command), 
        "cp %s_template.odl %s.odl", work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return;
    }
    status = system(system_command);
    assert_equal_with_message(status, FALSE, "Cannot rename template "
        "parameter file");
    if (status != 0)
    {
        size = snprintf(system_command, sizeof(system_command), "rm %s.omf", 
                 work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Read the parameters. */
    memset(&wo_params, 0, sizeof(RPS_NUCHAR_WO_PARAMS));
    memset(&omf_params, 0, sizeof(RPS_NUCHAR_OMF_PARAMS));

    size = snprintf(wo_params_filename, sizeof(wo_params_filename), "%s.odl", 
             work_order_id);
    if ((size < 0) || (size >= sizeof(wo_params_filename)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
               "it needs %d bytes", size);
        return;
    }
    status = read_parameters(wo_params_filename, &wo_params, &omf_params);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Failed to read parameters\n");
        size = snprintf(system_command, sizeof(system_command), "rm %s.omf", 
                 work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Check selected WO and OMF parameters out */

    assert_string_equal(wo_params.work_order_id, work_order_id);
    assert_equal(omf_params.satellite_number, 8);
    size = snprintf(l0r_filename, sizeof(l0r_filename), "%s/%s/%s", 
         test_data_path, l0r_data_dir, l0r_name);
    if ((size < 0) || (size >= sizeof(l0r_filename)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
               "it needs %d bytes", size);
        return;
    }
    assert_string_equal(omf_params.l0r_filename, l0r_filename);
    size = snprintf(cpf_filename, sizeof(cpf_filename), "%s/%s/%s", 
        test_data_path, cpf_data_dir, cpf_name);
    if ((size < 0) || (size >= sizeof(cpf_filename)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
               "it needs %d bytes", size);
        return;
    }
    assert_string_equal(omf_params.cpf_filename, cpf_filename);
    assert_string_equal(omf_params.pmask_filename, pmask_filename);

    /* Default RPS processing level strings should be printed out.  */
    assert_string_equal(wo_params.oli_rps_proc_level,
        RPS_NUCHAR_OLI_PROC_LEVEL_TEMPCORRECT);
    assert_string_equal(wo_params.tirs_rps_proc_level,
        RPS_NUCHAR_TIRS_PROC_LEVEL);

    /* This check ensures that the CALCULATE_RADIANCE work order
       parameter has been given its default value of 1 if it is not
       specified in the work order ODL file */
    assert_equal(wo_params.calculate_radiance, 1);

    /* Check the other parameters */
    assert_equal(wo_params.rms_banding_flag, RPS_NUCHAR_RMS_BANDING);
    assert_equal(wo_params.stdev_banding_flag, RPS_NUCHAR_STDEV_BANDING);
    assert_equal(wo_params.streaking_flag, RPS_NUCHAR_STREAKING);
    assert_equal(wo_params.correct_flag, 0);
    assert_equal(wo_params.relgain_option, RPS_NUCHAR_PRELAUNCH_RELGAINS);
    assert_equal(wo_params.num_bands_to_process, 4);

    for (ii = 0; ii < wo_params.num_bands_to_process; ii++)
        assert_equal(test_band_list[ii], wo_params.bands[ii]);

    assert_equal(wo_params.store_char_in_db, 1);
    assert_equal(wo_params.report_flag, RPS_NUCHAR_REPORT);
    assert_equal(wo_params.dump_flag, RPS_NUCHAR_NO_DUMP);

#ifdef DEBUG_PRINT
    print_wo_params_struct(wo_params);
    print_omf_params_struct(omf_params);
#endif

    /* Clean up */
    size = snprintf(system_command, sizeof(system_command), 
        "rm %s.omf %s.odl", work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return;
    }
    system(system_command);

}




/*-----------------------------------------------------------------------
           Define a test suite to add to the "master" test suite
-------------------------------------------------------------------------*/
void nonuniformity_read_parameters_test(TestSuite *suite)
{
    add_test(suite, test_read_parameters);
}
