/**********************************************************************
                     non_uniformity_reporting_test

 PURPOSE:      Use CGREEN to perform unit testing of the various
               component functions making up non-uniformity
               characterization reporting

***********************************************************************/
#include <stdio.h>
#include <limits.h>                      /* PATH_MAX */
#include <string.h>
#include "cgreen.h"
#include "ias_const.h"
#include "../char_nonuniformity.h"

#include "nonuniformity_tests.h"


static char work_order_id[IAS_WORKORDER_ID_SIZE] = "NUT1002";
static char system_command[1000] = "";

/*-----------------------------------------------------------------------
 NAME:               prepare_metrics_data

 PURPOSE:  Populates test RPS_NUCHAR_METRICS and RPS_NUCHAR_REPORT_SUMMARY
           data structures whose information is to be written to a
           summary report and/or dump file

 RETURNS:  Integer status code of SUCCESS or ERROR

------------------------------------------------------------------------*/
static int prepare_metrics_data(int band_number, int num_scas,
    RPS_NUCHAR_METRICS *test_metrics_data,
    RPS_NUCHAR_REPORT_SUMMARY *summary,
    RPS_NUCHAR_INVALID_FLAGS *invalid_flags)
{
    char *test_data_path = NULL;
    char system_command[1000] = "";
    const char rad_filename[] = "JAB_DetRadiances";
    const char rad_filedir[] = "ias/rps/nonuniformity";
    char full_rad_filename[PATH_MAX] = "";
    char wo_params_filename[PATH_MAX] = "";

    RPS_NUCHAR_WO_PARAMS wo_params;
    RPS_NUCHAR_OMF_PARAMS omf_params;

    double *band_rms_metrics = NULL;
    double *band_stdev_metrics = NULL;
    double *band_streaking_metrics = NULL;
    double radiance_mean = 0.0;
    double radiance_stdev = 0.0;

    int status = SUCCESS;
    int num_band_detectors = 0;
    int num_samples = 0;
    int rms_banding_flag = 1;
    int stdev_banding_flag = 1;
    int streaking_flag = 1;
    int num_detectors_per_sca = 494;
    int size;

    /* Make sure the test data path exists */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_not_equal_with_message(test_data_path, NULL, "Cannot find "
        "test data directory");
    if (test_data_path == NULL)
    {
        return ERROR;
    }

    /* Replace the $IAS_UNIT_TEST_DATA_DIR placeholder in the OMF file with
       its actual value */
    size = snprintf(system_command, sizeof(system_command), 
        "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' %s_template.omf > %s.omf", 
        test_data_path, work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return ERROR;
    } 
    status = system(system_command);
    assert_equal_with_message(status, 0, "Cannot edit OMF template");
    if (status != 0)
    {
        return ERROR;
    }

    /* Rename the template ODL file as the actual ODL file */
    size = snprintf(system_command, sizeof(system_command), 
        "cp %s_template.odl %s.odl", work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return ERROR;
    } 

    status = system(system_command);
    assert_equal_with_message(status, 0, "Cannot create working "
        "parameter file");
    if (status != 0)
    {
        size = snprintf(system_command, sizeof(system_command), "rm %s.omf", 
                work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        } 
        return ERROR;
    }

    /* Read the WO and OMF parameters */
    size = snprintf(wo_params_filename, sizeof(wo_params_filename), "%s.odl", 
             work_order_id);
    if ((size < 0) || (size >= sizeof(wo_params_filename)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
               "it needs %d bytes", size);
        return ERROR;
    } 
    status = read_parameters(wo_params_filename, &wo_params,
        &omf_params);
    assert_equal_with_message(status, SUCCESS, "Cannot obtain work order "
        "and/or OMF parameter information");
    if (status != SUCCESS)
    {
        size = snprintf(system_command, sizeof(system_command), 
            "rm %s.omf %s.odl", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return ERROR;
        } 
        system(system_command);
        return ERROR;
    }

    /* Initialize the summary report header data structures */
    summary->header.common_header = ias_misc_initialize_report_header(
            omf_params.l0r_filename, wo_params.work_order_id,
            "char_nonuniformity_test");
    assert_not_equal_with_message(summary->header.common_header, NULL,
        "Cannot initialize standard report header information");
    if (summary->header.common_header == NULL)
    {
        size = snprintf(system_command, sizeof(system_command), 
            "rm %s.odl %s.omf", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return ERROR;
        } 
        system(system_command);
        return ERROR;
    }

    /* Update the CPF in the header to reflect what is in the OMF file. */
    ias_misc_set_header_cpf_name(summary->header.common_header,
        omf_params.cpf_filename);

    ias_misc_set_header_collection_type(summary->header.common_header,
        IAS_RPS_OLI_SOLAR_COLLECT);

    status = ias_misc_initialize_rps_report_header(
        omf_params.l0r_filename, &(summary->header.rps_header));
    assert_equal_with_message(status, SUCCESS, "Cannot initialize RPS "
        "report header information");
    if (status != SUCCESS)
    {
        ias_misc_free_report_header(summary->header.common_header);
        size = snprintf(system_command, sizeof(system_command), 
            "rm %s.odl %s.omf", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return ERROR;
        } 
        system(system_command);
        return ERROR;
    }

    /* Get the band metrics, then populate the metrics data structure */
    num_band_detectors = num_scas * num_detectors_per_sca;

    size = snprintf(full_rad_filename, sizeof(full_rad_filename), 
       "%s/%s/%s.txt", test_data_path, rad_filedir, rad_filename);
    if ((size < 0) || (size >= sizeof(full_rad_filename)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
               "it needs %d bytes", size);
        return ERROR;
    }

    band_rms_metrics = initialize_band_metrics(full_rad_filename,
        num_band_detectors, wo_params.num_contiguous_detectors,
        "calculate_rms_banding_metrics", &radiance_mean,
        &radiance_stdev, &num_samples, invalid_flags->rms_banding);
    assert_not_equal_with_message(band_rms_metrics, NULL, "Cannot derive "
        "band-level RMS metrics");
    if (band_rms_metrics == NULL)
    {
        ias_misc_free_report_header(summary->header.common_header);
        size = snprintf(system_command, sizeof(system_command), 
                 "rm %s.odl %s.omf", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return ERROR;
        } 
        system(system_command);
        return ERROR;
    }

    band_stdev_metrics = initialize_band_metrics(full_rad_filename,
        num_band_detectors, wo_params.num_contiguous_detectors,
        "calculate_stdev_banding_metrics", &radiance_mean,
        &radiance_stdev, &num_samples, invalid_flags->stdev_banding);
    assert_not_equal_with_message(band_stdev_metrics, NULL, "Cannot derive "
        "band-level std deviation metrics");
    if (band_stdev_metrics == NULL)
    {
        free(band_rms_metrics);
        ias_misc_free_report_header(summary->header.common_header);
        size = snprintf(system_command, sizeof(system_command), 
                 "rm %s.odl %s.omf", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return ERROR;
        } 
        system(system_command);
        return ERROR;
    }

    band_streaking_metrics = initialize_band_metrics(full_rad_filename,
        num_band_detectors, 0, "calculate_streaking_metrics",
        &radiance_mean, &radiance_stdev, &num_samples,
        invalid_flags->streaking);
    assert_not_equal_with_message(band_streaking_metrics, NULL, "Cannot "
        "derive band-level streaking metrics");
    if (band_streaking_metrics == NULL)
    {
        free(band_rms_metrics);
        free(band_stdev_metrics);
        ias_misc_free_report_header(summary->header.common_header);
        size = snprintf(system_command, sizeof(system_command), 
                 "rm %s.odl %s.omf", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return ERROR;
        } 
        system(system_command);
        return ERROR;
    }

    status = allocate_metrics_buffers(rms_banding_flag, stdev_banding_flag,
        streaking_flag, num_scas, num_detectors_per_sca,
        test_metrics_data);
    assert_equal_with_message(status, SUCCESS, "Cannot allocate "
        "metrics data structure");
    if (status != SUCCESS)
    {
        free(band_rms_metrics);
        free(band_stdev_metrics);
        free(band_streaking_metrics);
        ias_misc_free_report_header(summary->header.common_header);
        size = snprintf(system_command, sizeof(system_command), 
                 "rm %s.odl %s.omf", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return ERROR;
        } 
        system(system_command);
        return ERROR;
    }

    /* Initialize summary data buffers */
    initialize_metrics_data(rms_banding_flag, stdev_banding_flag,
        streaking_flag, band_rms_metrics, band_stdev_metrics,
        band_streaking_metrics, radiance_mean, radiance_stdev,
        band_number, num_scas, num_detectors_per_sca,
        test_metrics_data);

    /* Calculate per-detector standard deviations */
    status = calculate_per_detector_metrics_std(band_rms_metrics,
        band_stdev_metrics, band_streaking_metrics,
        invalid_flags, num_band_detectors, rms_banding_flag,
        stdev_banding_flag, streaking_flag, summary);
    assert_equal_with_message(status, SUCCESS, "Cannot calculate "
        "per-detector standard deviations");
    if (status != SUCCESS)
    {
        free_summary_report_buffers(summary);
        ias_misc_free_report_header(summary->header.common_header);
        free_detector_metrics_buffers(test_metrics_data, num_scas);
        free(band_rms_metrics);
        free(band_stdev_metrics);
        free(band_streaking_metrics);
        size = snprintf(system_command, sizeof(system_command), 
            "rm %s.odl %s.omf", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return ERROR;
        } 
        system(system_command);
        return ERROR;
    }

    /* Free the band-level data buffers */
    free(band_rms_metrics);
    free(band_stdev_metrics);
    free(band_streaking_metrics);

    /* Initialize the rest of the summary report data structure */
    status = initialize_summary_report_data_structures(wo_params,
        rms_banding_flag, stdev_banding_flag, streaking_flag,
        *test_metrics_data, invalid_flags, summary);
    assert_equal_with_message(status, SUCCESS, "Cannot initialize "
        "summary report data structure");
    if (status != SUCCESS)
    {
        free_summary_report_buffers(summary);
        ias_misc_free_report_header(summary->header.common_header);
        free_detector_metrics_buffers(test_metrics_data, num_scas);
        size = snprintf(system_command, sizeof(system_command), 
            "rm %s.odl %s.omf", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return ERROR;
        } 
        system(system_command);
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}



Ensure test_write_summary_report()
{
    RPS_NUCHAR_METRICS test_metrics_data;
    RPS_NUCHAR_REPORT_SUMMARY summary;
    RPS_NUCHAR_INVALID_FLAGS invalid_flags;

    char test_report_name[PATH_MAX] = "";

    int status;
    int num_scas = 14;
    int band_number = 1;
    int num_detectors_per_sca = 494;
    int num_band_detectors = num_scas * num_detectors_per_sca;

    FILE *rpt_fptr = NULL;
    int size;

    printf("Generating test summary report...\n");

    /* Clear data structures */
    memset(&test_metrics_data, 0, sizeof(RPS_NUCHAR_METRICS));
    memset(&summary, 0, sizeof(RPS_NUCHAR_REPORT_SUMMARY));

    /* Allocate the invalid flag data buffers */
    memset(&invalid_flags, 0, sizeof(RPS_NUCHAR_INVALID_FLAGS));
    status = allocate_nonuniformity_invalid_flag_buffers(num_band_detectors,
        &invalid_flags);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error allocating invalid flag "
            "data buffers");
        return;
    }

    /* Prepare the test data */
    status = prepare_metrics_data(band_number, num_scas,
        &test_metrics_data, &summary, &invalid_flags);
    assert_equal_with_message(status, SUCCESS, "Cannot prepare test "
        "metrics data");
    if (status != SUCCESS)
    {
        free_nonuniformity_invalid_flag_buffers(&invalid_flags);
        return;
    }

    /* Try to create the summary report file */
    ias_misc_set_header_band_number(summary.header.common_header,
        band_number);

    size = snprintf(test_report_name, sizeof(test_report_name), 
        "%s_nonuniformity_summary_B%02d.txt", work_order_id, band_number);
    if ((size < 0) || (size >= sizeof(test_report_name)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
               "it needs %d bytes", size);
        free_nonuniformity_invalid_flag_buffers(&invalid_flags);
        return;
    } 
    rpt_fptr = fopen(test_report_name, "w");
    assert_not_equal_with_message(rpt_fptr, NULL, "Cannot create summary "
        "report file");
    if (rpt_fptr == NULL)
    {
        free_summary_report_buffers(&summary);
        free_detector_metrics_buffers(&test_metrics_data, num_scas);
        ias_misc_free_report_header(summary.header.common_header);
        free_nonuniformity_invalid_flag_buffers(&invalid_flags);
        size = snprintf(system_command, sizeof(system_command), 
              "rm NUT1002.odl NUT1002.omf");
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            free_nonuniformity_invalid_flag_buffers(&invalid_flags);
            return;
        } 
        system(system_command);
        free_nonuniformity_invalid_flag_buffers(&invalid_flags);
        return;
    }

    /* Write the summary report data to the file */
    status = write_summary_report_data(rpt_fptr, summary, 
        test_metrics_data.band_radiance_mean,
        test_metrics_data.band_radiance_stdev,
        test_metrics_data.ffov_metric);
    assert_equal_with_message(status, SUCCESS, "Cannot write summary "
        "metrics to report file");

    /* Cleanup */
    fclose(rpt_fptr);
    free_nonuniformity_invalid_flag_buffers(&invalid_flags);
    free_summary_report_buffers(&summary);
    free_detector_metrics_buffers(&test_metrics_data, num_scas);
    ias_misc_free_report_header(summary.header.common_header);

    /* Clean up output files */
    size = snprintf(system_command, sizeof(system_command), 
        "rm %s_nonuniformity_summary_B01.txt %s.odl %s.omf",work_order_id, 
        work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return;
    } 
    system(system_command);
    /* Done */
}




Ensure test_write_dump_file()
{
    char dumpfile_name[PATH_MAX] = "";

    RPS_NUCHAR_METRICS test_metrics_data;
    RPS_NUCHAR_REPORT_SUMMARY summary;
    RPS_NUCHAR_INVALID_FLAGS invalid_flags;

    int band_number = 1;
    int status = SUCCESS;
    int num_scas = 14;
    int num_detectors_per_sca = 494;
    int num_band_detectors = num_scas * num_detectors_per_sca;

    FILE *dump_fptr = NULL;
    int size;

    printf("Generating test dump file...\n");

    /* Allocate the invalid flag data buffers */
    memset(&invalid_flags, 0, sizeof(RPS_NUCHAR_INVALID_FLAGS));
    status = allocate_nonuniformity_invalid_flag_buffers(num_band_detectors,
        &invalid_flags);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error allocating invalid flag "
            "data buffers");
        return;
    }

    /* Prepare the test data */
    status = prepare_metrics_data(band_number, num_scas,
        &test_metrics_data, &summary, &invalid_flags);
    assert_equal_with_message(status, SUCCESS, "Cannot prepare test "
        "metrics data");
    if (status != SUCCESS)
    {
        free_nonuniformity_invalid_flag_buffers(&invalid_flags);
        return;
    }

    /* Try to create the dump file */
    ias_misc_set_header_band_number(summary.header.common_header,
        band_number);
    size = snprintf(dumpfile_name, sizeof(dumpfile_name), 
        "%s_nonuniformity_dumpfile_B%02d.txt", work_order_id, band_number);
    if ((size < 0) || (size >= sizeof(dumpfile_name)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
               "it needs %d bytes", size);
        free_nonuniformity_invalid_flag_buffers(&invalid_flags);
        return;
    } 

    dump_fptr = fopen(dumpfile_name, "w");
    assert_true(dump_fptr != NULL);
    if (dump_fptr == NULL)
    {
        printf("Cannot create dump file\n");
        free_summary_report_buffers(&summary);
        free_detector_metrics_buffers(&test_metrics_data, num_scas);
        size = snprintf(system_command, sizeof(system_command), 
            "rm %s.odl %s.omf ", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            free_nonuniformity_invalid_flag_buffers(&invalid_flags);
            return;
        } 
        system(system_command);
        free_nonuniformity_invalid_flag_buffers(&invalid_flags);
        return;
    }

    /* Write the metrics data to the dump file */
    status = write_metrics_data_to_dump_file(test_metrics_data,
        summary, &invalid_flags, dump_fptr);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
        printf("Cannot write summary metrics data to report file\n");

    /* Cleanup */
    fclose(dump_fptr);
    free_nonuniformity_invalid_flag_buffers(&invalid_flags);
    free_summary_report_buffers(&summary);
    free_detector_metrics_buffers(&test_metrics_data, num_scas);
    ias_misc_free_report_header(summary.header.common_header);

    /* Clean up output files */
    size = snprintf(system_command, sizeof(system_command), 
        "rm %s_nonuniformity_dumpfile_B01.txt %s_nonuniformity_header.txt "
        "%s.odl %s.omf", work_order_id, work_order_id, work_order_id, 
        work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return;
    } 
    system(system_command);
    /* Done */
}




/*-----------------------------------------------------------------------
           Define a test suite to add to the "master" test suite
-------------------------------------------------------------------------*/
void nonuniformity_reporting_test(TestSuite *suite)
{
    add_test(suite, test_write_summary_report);
    add_test(suite, test_write_dump_file);
}
