/**********************************************************************
                     non_uniformity_sca_summary_test

 PURPOSE:      Use CGREEN to perform unit testing on the algorithm
               used to create the SCA summary statistics data for a
               given nonuniformity characterization.

***********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ias_satellite_attributes.h"

#include "cgreen.h"

#include "../char_nonuniformity.h"

#include "nonuniformity_tests.h"

/*#define DEBUG_PRINT */


static char rad_filename[] = "JAB_DetRadiances";
static char rad_filedir[] = "ias/rps/nonuniformity";
static char full_rad_filename[PATH_MAX];


Ensure test_generate_sca_summary()
{
    char *path = NULL;

    double *band_rms_metrics = NULL;
    double *band_stdev_metrics = NULL;
    double *band_streaking_metrics = NULL;
    double radiance_mean = 0.0;
    double radiance_stdev = 0.0;

    RPS_NUCHAR_METRICS test_metrics_data;
    RPS_NUCHAR_REPORT_SUMMARY summary;
    RPS_NUCHAR_INVALID_FLAGS invalid_flags;

    int status;
    int num_scas = 14;
    int num_detectors_per_sca = 494;
    int sca_index;
    int num_contiguous_detectors = 100;
    int num_band_detectors;
    int num_samples = 0;
    int rms_banding_flag = 1;
    int stdev_banding_flag = 1;
    int streaking_flag = 1;


    printf("Testing generation of SCA-level summary metrics "
           "information...\n");

    /* Get the test data path */
    path = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(path != NULL);
    if (path == NULL)
    {
        printf("Cannot get test data directory location\n");
        return;
    }

    /* Get the band metrics, the populate a metrics data structure */
    num_band_detectors = num_scas * num_detectors_per_sca;

    sprintf(full_rad_filename, "%s/%s/%s.txt", path, rad_filedir,
       rad_filename);

    status = allocate_nonuniformity_invalid_flag_buffers(num_band_detectors,
        &invalid_flags);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error allocating invalid flag "
            "data buffers");
        return;
    }

    band_rms_metrics = initialize_band_metrics(full_rad_filename,
        num_band_detectors, num_contiguous_detectors,
        "calculate_rms_banding_metrics", &radiance_mean,
        &radiance_stdev, &num_samples, invalid_flags.rms_banding);

    assert_true(band_rms_metrics != NULL);
    if (band_rms_metrics == NULL)
    {
        printf("Cannot calculate band-level RMS banding metrics\n");
        free_nonuniformity_invalid_flag_buffers(&invalid_flags);
        return;
    }

    band_stdev_metrics = initialize_band_metrics(full_rad_filename,
        num_band_detectors, num_contiguous_detectors,
        "calculate_stdev_banding_metrics", &radiance_mean,
        &radiance_stdev, &num_samples, invalid_flags.stdev_banding);

    assert_true(band_rms_metrics != NULL);
    if (band_stdev_metrics == NULL)
    {
        printf("Cannot calculate band-level stdev banding metrics\n");
        free(band_rms_metrics);
        free_nonuniformity_invalid_flag_buffers(&invalid_flags);
        return;
    }

    band_streaking_metrics = initialize_band_metrics(full_rad_filename,
        num_band_detectors, 0, "calculate_streaking_metrics",
        &radiance_mean, &radiance_stdev, &num_samples,
        invalid_flags.streaking);

    assert_true(band_streaking_metrics != NULL);
    if (band_streaking_metrics == NULL)
    {
        printf("Cannot calculate band-level streaking metrics\n");
        free(band_rms_metrics);
        free(band_stdev_metrics);
        free_nonuniformity_invalid_flag_buffers(&invalid_flags);
        return;
    }


    /* Set up the detector-level metrics data buffers */
    memset(&test_metrics_data, 0, sizeof(RPS_NUCHAR_METRICS));
    status = allocate_metrics_buffers(rms_banding_flag,
            stdev_banding_flag,
            streaking_flag, num_scas,
            num_detectors_per_sca,
            &test_metrics_data);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot create detector-level metrics data structure "
                "with RMS information\n");
        free(band_rms_metrics);
        free(band_stdev_metrics);
        free(band_streaking_metrics);
        free_nonuniformity_invalid_flag_buffers(&invalid_flags);
        return;
    }
    initialize_metrics_data(rms_banding_flag,
             stdev_banding_flag,
             streaking_flag,
             band_rms_metrics,
             band_stdev_metrics,
             band_streaking_metrics,
             radiance_mean,
             radiance_stdev,
             1,
             num_scas,
             num_detectors_per_sca,
             &test_metrics_data);

    /* Set up the per-SCA and cross-SCA metrics buffers in the
       summary report data structure */
    memset(&summary, 0, sizeof(RPS_NUCHAR_REPORT_SUMMARY));
    status = allocate_summary_report_buffers(rms_banding_flag,
            stdev_banding_flag,
            streaking_flag,
            num_scas,
            &summary);

    assert_true(status == SUCCESS);
    assert_true((summary.sca_summary_rms != NULL)
            && (summary.sca_summary_stdev != NULL)
            && (summary.sca_summary_streaking != NULL)
            && (summary.cross_sca_rms != NULL)
            && (summary.cross_sca_stdev != NULL));

    if ((status != SUCCESS)
            || (summary.sca_summary_rms == NULL)
            || (summary.sca_summary_stdev == NULL)
            || (summary.sca_summary_streaking == NULL)
            || (summary.cross_sca_rms == NULL)
            || (summary.cross_sca_stdev == NULL))
    {
        printf("Cannot allocate report summary data buffers\n");
        free(band_rms_metrics);
        free(band_stdev_metrics);
        free(band_streaking_metrics);
        free_summary_report_buffers(&summary);
        free_detector_metrics_buffers(&test_metrics_data, num_scas);
        free_nonuniformity_invalid_flag_buffers(&invalid_flags);
        return;
    }


    /* Initialize the per-SCA and cross-SCA summary data */
    status = initialize_sca_summary_metrics(test_metrics_data,
            &invalid_flags,
            rms_banding_flag,
            stdev_banding_flag,
            streaking_flag,
            num_scas,
            num_detectors_per_sca,
            &summary);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot populate report summary data buffers\n");
        free(band_rms_metrics);
        free(band_stdev_metrics);
        free(band_streaking_metrics);
        free_summary_report_buffers(&summary);
        free_detector_metrics_buffers(&test_metrics_data, num_scas);
        free_nonuniformity_invalid_flag_buffers(&invalid_flags);
        return;
    }

    /* Check to see if things were allocated/populated properly */
    for (sca_index = 0; sca_index < num_scas; sca_index++)
    {
        /* Right SCA number */
        assert_equal(summary.sca_summary_rms[sca_index].sca_number,
                (sca_index + 1));
        assert_equal(summary.sca_summary_stdev[sca_index].sca_number,
                (sca_index + 1));
        assert_equal(summary.sca_summary_streaking[sca_index].sca_number,
                (sca_index + 1));

        /* Right number of characterized detectors */
        if (sca_index == 0)
        {
            assert_equal(summary.sca_summary_rms[0]
                    .num_characterized_detectors,
                    num_detectors_per_sca);
            assert_equal(summary.sca_summary_stdev[0]
                    .num_characterized_detectors,
                    num_detectors_per_sca);
            assert_equal(summary.sca_summary_streaking[0]
                    .num_characterized_detectors,
                    (num_detectors_per_sca - 1));
        }
        else if (sca_index == num_scas - 1)
        {
            assert_equal(summary.sca_summary_rms[num_scas - 1]
                    .num_characterized_detectors,
                    (num_detectors_per_sca - (num_contiguous_detectors
                                              - 1)));
            assert_equal(summary.sca_summary_stdev[num_scas - 1]
                    .num_characterized_detectors,
                    (num_detectors_per_sca - (num_contiguous_detectors
                                              - 1)));
            assert_equal(summary.sca_summary_streaking[num_scas - 1]
                    .num_characterized_detectors,
                    (num_detectors_per_sca - 1));
        }
        else
        {
            assert_equal(summary.sca_summary_rms[sca_index]
                    .num_characterized_detectors, num_detectors_per_sca);
            assert_equal(summary.sca_summary_stdev[sca_index]
                    .num_characterized_detectors, num_detectors_per_sca);
            assert_equal(summary.sca_summary_streaking[sca_index]
                    .num_characterized_detectors, num_detectors_per_sca);
        }

        /* Nonzero values populated elsewhere */
        assert_true(summary.sca_summary_rms[sca_index].metric_min != 0);
        assert_true(summary.sca_summary_rms[sca_index].metric_max != 0);
        assert_true(summary.sca_summary_rms[sca_index].detector_min != 0);
        assert_true(summary.sca_summary_rms[sca_index].detector_max != 0);
        assert_true(summary.sca_summary_rms[sca_index].metric_mean != 0);
        assert_true(summary.sca_summary_rms[sca_index].metric_stdev != 0);
        assert_true(summary.sca_summary_rms[sca_index].pcnt != 0);

        assert_true(summary.sca_summary_stdev[sca_index].metric_min != 0);
        assert_true(summary.sca_summary_stdev[sca_index].metric_max != 0);
        assert_true(summary.sca_summary_stdev[sca_index].detector_min != 0);
        assert_true(summary.sca_summary_stdev[sca_index].detector_max != 0);
        assert_true(summary.sca_summary_stdev[sca_index].metric_mean != 0);
        assert_true(summary.sca_summary_stdev[sca_index].metric_stdev != 0);
        assert_true(summary.sca_summary_stdev[sca_index].pcnt != 0);

        assert_true(summary.sca_summary_streaking[sca_index].metric_min != 0);
        assert_true(summary.sca_summary_streaking[sca_index].metric_max != 0);
        assert_true(summary.sca_summary_streaking[sca_index]
                .detector_min != 0);
        assert_true(summary.sca_summary_streaking[sca_index]
                .detector_max != 0);
        assert_true(summary.sca_summary_streaking[sca_index]
                .metric_mean != 0);
        assert_true(summary.sca_summary_streaking[sca_index]
                .metric_stdev != 0);
        assert_true(summary.sca_summary_streaking[sca_index].pcnt != 0);
    }


    /* Similar checks for the cross-SCA data */
    for (sca_index = 0; sca_index < (num_scas - 1); sca_index++)
    {
        assert_equal(summary.cross_sca_rms[sca_index].sca1_number,
                (sca_index + 1));
        assert_equal(summary.cross_sca_rms[sca_index].sca2_number,
                (sca_index + 2));
        assert_equal(summary.cross_sca_stdev[sca_index].sca1_number,
                (sca_index + 1));
        assert_equal(summary.cross_sca_stdev[sca_index].sca2_number,
                (sca_index + 2));

        assert_true(summary.cross_sca_rms[sca_index].cross_mean != 0);
        assert_true(summary.cross_sca_rms[sca_index].cross_stdev != 0);
        assert_true(summary.cross_sca_rms[sca_index]
                .cross_num_pixels != 0);
        assert_true(summary.cross_sca_rms[sca_index].cross_pcnt != 0);

        assert_true(summary.cross_sca_stdev[sca_index].cross_mean != 0);
        assert_true(summary.cross_sca_stdev[sca_index].cross_stdev != 0);
        assert_true(summary.cross_sca_stdev[sca_index]
                .cross_num_pixels != 0);
        assert_true(summary.cross_sca_stdev[sca_index].cross_pcnt != 0);
    }

#ifdef DEBUG_PRINT
    for (sca_index = 0; sca_index < num_scas; sca_index++)
    {
        printf("Summary statistics for SCA %d\n", sca_index+1);
        printf("Maximum RMS banding observed:\t%10.6lf (detector %0d)\n",
                summary.sca_summary_rms[sca_index].metric_max,
                summary.sca_summary_rms[sca_index].detector_max);
        printf("Minimum RMS banding observed:\t%10.6lf (detector %0d)\n",
                summary.sca_summary_rms[sca_index].metric_min,
                summary.sca_summary_rms[sca_index].detector_min);
        printf("Mean RMS banding +- std:\t%10.6lf +- %10.6lf\n",
                summary.sca_summary_rms[sca_index].metric_mean,
                summary.sca_summary_rms[sca_index].metric_stdev);
        printf("Number of detectors characterized:\t%d\n",
                summary.sca_summary_rms[sca_index]
                .num_characterized_detectors);
        printf("\n\n");
    }
#endif

    /* Free up allocated memory */
    free(band_rms_metrics);
    free(band_stdev_metrics);
    free(band_streaking_metrics);
    free_summary_report_buffers(&summary);
    free_detector_metrics_buffers(&test_metrics_data, num_scas);
    free_nonuniformity_invalid_flag_buffers(&invalid_flags);
}




/*-----------------------------------------------------------------------
           Define a test suite to add to the "master" test suite
-------------------------------------------------------------------------*/
void nonuniformity_sca_summary_test(TestSuite *suite)
{
    add_test(suite, test_generate_sca_summary);
}
