/**********************************************************************
 NAME:         nonuniformity_write_header_test

 PURPOSE:      Use CGREEN to perform unit testing on the
               write_report_header_to_file routine

***********************************************************************/
#include <stdio.h>
#include <string.h>
#include <limits.h>                      /* PATH_MAX */
#include "cgreen.h"
#include "ias_const.h"
#include "ias_radiometry.h"
#include "../char_nonuniformity.h"

#include "nonuniformity_tests.h"


Ensure test_write_report_header_to_file()
{
    char *test_data_path = NULL;
    const char work_order_id[IAS_WORKORDER_ID_SIZE] = "NUT1002";
    char wo_params_filename[PATH_MAX] = "";
    char system_command[1000] = "";

    int status = SUCCESS;
    int size;

    RPS_NUCHAR_WO_PARAMS wo_params;
    RPS_NUCHAR_OMF_PARAMS omf_params;

    RPS_NUCHAR_METRICS test_metrics_data;
    RPS_NUCHAR_REPORT_SUMMARY summary;


    printf("Generating test header file...\n");

    /* Make sure the test data path exists */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_not_equal_with_message(test_data_path, NULL, "Cannot find "
        "test data directory");
    if (test_data_path == NULL)
    {
        return;
    }

    /* Replace the $IAS_UNIT_TEST_DATA_DIR placeholder in the OMF file with
       its actual value */
    size = snprintf(system_command, sizeof(system_command), 
        "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' %s_template.omf > %s.omf", 
        test_data_path, work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return;
    }
    status = system(system_command);
    assert_equal_with_message(status, FALSE, "Cannot edit OMF template");

    /* Rename the template ODL file as the actual ODL file */
    size = snprintf(system_command, sizeof(system_command), 
        "cp %s_template.odl %s.odl", work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return;
    }
    status = system(system_command);
    assert_equal_with_message(status, FALSE, "Cannot rename template "
        "parameter file");
    if (status != 0)
    {
        size = snprintf(system_command, sizeof(system_command), "rm %s.omf", 
             work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Clear data structures */
    memset(&wo_params, 0, sizeof(RPS_NUCHAR_WO_PARAMS));
    memset(&omf_params, 0, sizeof(RPS_NUCHAR_OMF_PARAMS));
    memset(&test_metrics_data, 0, sizeof(RPS_NUCHAR_METRICS));
    memset(&summary, 0, sizeof(RPS_NUCHAR_REPORT_SUMMARY));

    /* Read the WO and OMF parameters */
    size = snprintf(wo_params_filename, sizeof(wo_params_filename), "%s.odl", 
             work_order_id);
    if ((size < 0) || (size >= sizeof(wo_params_filename)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
               "it needs %d bytes", size);
        return;
    }
    status = read_parameters(wo_params_filename, &wo_params,
        &omf_params);
    assert_equal_with_message(status, SUCCESS, "Cannot obtain work order "
        "and/or OMF parameter information");
    if (status != SUCCESS)
    {
        size = snprintf(system_command, sizeof(system_command), 
                  "rm %s.omf %s.odl", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Initialize the summary report header data structures */
    summary.header.common_header = ias_misc_initialize_report_header(
            omf_params.l0r_filename, wo_params.work_order_id,
            "char_nonuniformity_test");
    assert_not_equal_with_message(summary.header.common_header, NULL,
        "Cannot initialize standard report header information");
    if (summary.header.common_header == NULL)
    {
        size = snprintf(system_command, sizeof(system_command), 
                  "rm %s.omf %s.odl", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Update the CPF in the header to reflect what is in the OMF file. */
    ias_misc_set_header_cpf_name(summary.header.common_header,
        omf_params.cpf_filename);
    ias_misc_set_header_collection_type(summary.header.common_header,
        IAS_RPS_OLI_SOLAR_COLLECT);

    status = ias_misc_initialize_rps_report_header(
        omf_params.l0r_filename, &(summary.header.rps_header));
    assert_equal_with_message(status, SUCCESS, "Cannot initialize "
        "RPS-specific report header information");
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        ias_misc_free_report_header(summary.header.common_header);
        size = snprintf(system_command, sizeof(system_command), 
                  "rm %s.omf rm %s.odl", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Write the report header to a header file */
    status = write_report_header_to_file(summary, work_order_id);
    assert_equal_with_message(status, SUCCESS, "Cannot write report "
        "header information to file");

    /* Clean up */
    ias_misc_free_report_header(summary.header.common_header);

    size = snprintf(system_command, sizeof(system_command), 
              "rm %s.omf %s.odl", work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return;
    }
    system(system_command);

    /* Done */
}




/*-----------------------------------------------------------------------
           Define a test suite to add to the "master" test suite
-------------------------------------------------------------------------*/
void nonuniformity_write_header_test(TestSuite *suite)
{
    add_test(suite, test_write_report_header_to_file);
}
