/**************************************************************************
 NAME:              write_dump_data_file

 PURPOSE:           Module implementing routine(s) to write band-specific
                    dump data files for detector-level nonuniformity
                    characterization results

***************************************************************************/
#include <stdio.h>
#include <string.h>
#include "ias_logging.h"

#include "char_nonuniformity.h"

#define NO_OUTPUT_PCNT_METRIC     0
#define OUTPUT_PCNT_METRIC        1
#define MAX_COL_SIZE             12
#define EMPTY_STRING "           "


/*-----------------------------------------------------------------------
 NAME:        write_detector_metrics_to_dump_file

 PURPOSE:     Writes detector metrics for the current band/SCA to the
              dump file for the current band

 RETURNS:     Integer status code of SUCCESS or ERROR

 NOTES:       The output_pcnt_flags array parameter controls whether the
              percent metrics results are calculated.  If the flag for the
              characterization is set, the results are calculated,
              otherwise, they're not.  If they're not calculated, 'flag'
              values are assigned to them.  This is done in an attempt to
              keep the logic creating the output line buffer as simple as
              possible
-------------------------------------------------------------------------*/
static int write_detector_metrics_to_dump_file
(
    FILE *dump_fptr,                         /* I: Open dump file for current
                                                band */
    RPS_NUCHAR_METRICS metrics,              /* I: Metrics data for all
                                                bands/SCAs */
    const RPS_NUCHAR_INVALID_FLAGS *invalid_flags,
                                             /* I: Characterization result
                                                invalid flags for all
                                                detectors in the current
                                                band (0=valid
                                                characterization result,
                                                nonzero=invalid result)*/
    RPS_NUCHAR_REPORT_SUMMARY summary,       /* I: Report summary data for
                                                the current band */
    int output_pcnt_flags[3]                 /* I: Flags indicating whether
                                                to write pcnt metric
                                                characterization values to
                                                dump file */
)
{
    int det_index;
    int sca_index;
    int status;


    for (sca_index = 0; sca_index < summary.scas; sca_index++)
    {
        int sca_number = sca_index + 1;
        int starting_index = sca_index * summary.detectors_per_sca;

        /* Construct the output line for each detector */
        for (det_index = 0; det_index < summary.detectors_per_sca;
             det_index++)
        {
            char col1[MAX_COL_SIZE];
            char col2[MAX_COL_SIZE];
            char col3[MAX_COL_SIZE];
            char col4[MAX_COL_SIZE];
            char col5[MAX_COL_SIZE];
            char col6[MAX_COL_SIZE];

            int det_number = det_index + 1;

            /* Initialize the columns with an empty string */
            strcpy(col1, EMPTY_STRING);
            strcpy(col2, EMPTY_STRING);
            strcpy(col3, EMPTY_STRING);
            strcpy(col4, EMPTY_STRING);
            strcpy(col5, EMPTY_STRING);
            strcpy(col6, EMPTY_STRING);

            /* Replace the empty strings with data for each
               characterization that was performed (subject to
               whether the detector has valid data to start with) */
            if (summary.rms_banding_flag)
            {
                double raw_rms_metric =
                    metrics.rms_metrics[sca_index].metric[det_index];
                if (!invalid_flags->rms_banding[starting_index + det_index])
                {
                    status = sprintf(col1, " %10.6f", raw_rms_metric);
                    if (status < 0)
                    {
                        IAS_LOG_ERROR("Writing raw RMS metric to output");
                        return ERROR;
                    }

                    /* If we need the pcnt RMS metric, write it out */
                    if (output_pcnt_flags[0])
                    {
                        double pcnt_rms_metric = calculate_pcnt_metric(
                            raw_rms_metric, metrics.band_radiance_mean);
                        status = sprintf(col2, " %10.6f", pcnt_rms_metric);
                        if (status < 0)
                        {
                            IAS_LOG_ERROR("Writing RMS pcnt metric to "
                                "output");
                            return ERROR;
                        }
                    }
                }
            }

            if (summary.stdev_banding_flag)
            {
                double raw_std_metric =
                    metrics.stdev_metrics[sca_index].metric[det_index];
                if (!invalid_flags->stdev_banding[starting_index + det_index])
                {
                    status = sprintf(col3, " %10.6f", raw_std_metric);
                    if (status < 0)
                    {
                        IAS_LOG_ERROR("Writing raw STDEV metric to output");
                        return ERROR;
                    }

                    /* If we need the pcnt STDEV metric, write it out */
                    if (output_pcnt_flags[1])
                    {
                        double pcnt_std_metric = calculate_pcnt_metric(
                            raw_std_metric, metrics.band_radiance_mean);
                        status = sprintf(col4, " %10.6f", pcnt_std_metric);
                        if (status < 0)
                        {
                            IAS_LOG_ERROR("Writing prcnt STDEV metric to "
                                "output");
                            return ERROR;
                        }
                    }
                }
            }

            if (summary.streaking_flag)
            {
                double raw_strk_metric =
                    metrics.streaking_metrics[sca_index].metric[det_index];
                if (!invalid_flags->streaking[starting_index + det_index])
                {
                    status = sprintf(col5, " %10.6f", raw_strk_metric);
                    if (status < 0)
                    {
                        IAS_LOG_ERROR("Writing raw STRK metric to output");
                        return ERROR;
                    }

                    /* If we need the pcnt RMS metric, write it out */
                    if (output_pcnt_flags[2])
                    {
                        double pcnt_strk_metric = calculate_pcnt_metric(
                            raw_strk_metric, metrics.band_radiance_mean);
                        status = sprintf(col6, " %10.6f", pcnt_strk_metric);
                        if (status < 0)
                        {
                            IAS_LOG_ERROR("Writing pcnt STRK metric to "
                                "output");
                            return ERROR;
                        }
                    }
                }
            }

            /* After all that... write the output line buffer to the file */
            status = fprintf(dump_fptr, "%2d  %3d%s%s%s%s%s%s\n", sca_number,
                det_number, col1, col2, col3, col4, col5, col6);
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing final output line buffer to file "
                    "for detector %d", det_number);
                return ERROR;
            }
        }
    }

    return SUCCESS;
}   /* END -- write_detector_metrics_to_dump_files */



/*------------------------------------------------------------------------
 NAME:            write_metrics_data_to_dump_file

 PURPOSE:  Writes detector and SCA-level summary metrics information to
           the dump file

 RETURNS:  Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int write_metrics_data_to_dump_file
(
    RPS_NUCHAR_METRICS metrics,           /* I: Detector-level metrics
                                             array */
    RPS_NUCHAR_REPORT_SUMMARY summary,    /* I: Summary-level metrics
                                             information */
    const RPS_NUCHAR_INVALID_FLAGS *invalid_flags,
                                          /* I: Characterization result
                                             invalid flags for all
                                             detectors in the current
                                             band */
    FILE *dump_fptr                       /* I/O: Open dump file handle */
)
{
    int status;

    /* Flags to specify whether the corresponding pcnt nonuniformity
       metric is written to the dump file.  The values are set based on
       the dump file outputs specified in the current nonuniformity ADD */
    int output_pcnt_flags[3] = {OUTPUT_PCNT_METRIC, OUTPUT_PCNT_METRIC,
                                NO_OUTPUT_PCNT_METRIC};


    /* Write band summary information here, starting with a header */
    status = fprintf(dump_fptr, "BAND %d (%s) SUMMARY RESULTS:\n"
        "NSCAS  NDETS     "
        "RADIANCE    RADIANCE      FFOV%%\n                   "
        "W /        STDDEV\n             "
        "((m^2 sr um))\n--------------------------------------------"
        "---------------------------------\n", summary.band_number,
        summary.band_name);
    if(status < 0)
    {
        IAS_LOG_ERROR("Writing summary band information header to dump "
            "file");
        return ERROR;
    } 
    status = fprintf(dump_fptr, "%2d    %3d    %11.6lf  "
        "%11.6lf  %11.6lf\n", summary.scas, summary.detectors_per_sca,
        metrics.band_radiance_mean, metrics.band_radiance_stdev,
        metrics.ffov_metric);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing band/SCA information to dump file");
        return ERROR;
    }

    /* Write header for detector-specific results */
    status = fprintf(dump_fptr, "\n\n\nBAND %d (%s) DETECTOR RESULTS:\n"
        "SCA DET"
        "      RMS       RMS%%        STD       STD%%       STRK"
        "      STRK%%\n"
        "-------------------------------------------------------------"
        "-------------\n", summary.band_number, summary.band_name);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing detector-specific header information to "
            "dump file");
        return ERROR;
    }

    /* Write the detector metrics */
    status = write_detector_metrics_to_dump_file(dump_fptr, metrics,
        invalid_flags, summary, output_pcnt_flags);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing detector-specific metrics to dump file");
        return ERROR;
    }

    /* Everything appears to have printed OK */
    return SUCCESS;
}
