/**************************************************************************
 NAME:              write_summary_report

 PURPOSE:           Module implementing routines to write band-specific
                    optional summary reports for nonuniformity
                    characterization

***************************************************************************/
#include <stdio.h>

#include "ias_logging.h"

#include "char_nonuniformity.h"

/* Flag values defined for optionally print out the outputs */
#define NO_OUTPUT_PCNT_METRIC   0
#define OUTPUT_PCNT_METRIC      1

/*------------------------------------------------------------------------
 NAME:            write_summary_metrics_data

 PURPOSE:   Writes each band's summary characterization results to
            the report file

 RETURNS:   Integer status code of SUCCESS or ERROR

 NOTES:       The output_pcnt_flags array parameter controls whether the
              percent metrics results are calculated.  If the flag for the
              characterization is set, the results are calculated,
              otherwise, they're not.  If they're not calculated, 'flag'
              values are assigned to them.  This is done in an attempt to
              keep the logic creating the output line buffer as simple as
              possible
-------------------------------------------------------------------------*/
static int write_summary_metrics_data
(
    FILE *rpt_fptr,                         /* I: Open report file handle */
    RPS_NUCHAR_REPORT_SUMMARY summary,      /* I: SCA-specific metrics
                                               summary information */
    double band_radiance_mean,              /* I: Summary radiance
                                               statistics for the current
                                               band */
    int output_pcnt_flags[3]                /* I: Flags indicating whether
                                               to write pcnt metric
                                               characterization values to
                                               dump file */
)
{
    int sca_index;                          /* 0-based SCA index */
    int status;

    /* Print RMS banding results */
    if (summary.rms_banding_flag)
    {
        if ((fprintf(rpt_fptr, "RMS BANDING\n\n") < 0)
            || (fprintf(rpt_fptr, "per-detector stdev:   %15.6f\n",
                summary.rms_band_stdev) < 0)
            || (fprintf(rpt_fptr, "Banding across SCA discontinuity\n") < 0))
        {
            IAS_LOG_ERROR("Writing RMS summary header");
            return ERROR;
        }

        if (output_pcnt_flags[0] != NO_OUTPUT_PCNT_METRIC)
        {
            if ((fprintf(rpt_fptr, "  SCA         BandingRMS"
                     "      BandingRMS      BandingRMS\n") < 0)
                || (fprintf(rpt_fptr, "boundary       mean [rad]"
                     "      std [rad]        std [%%]"
                     "      n_pixels\n") < 0))
            {
                IAS_LOG_ERROR("Writing RMS column titles with percent metric");
                return ERROR;
            }
        }
        else
        {
            if ((fprintf(rpt_fptr, "  SCA         BandingRMS"
                     "      BandingRMS\n") < 0)
                || (fprintf(rpt_fptr, "boundary       mean [rad]"
                     "      std [rad]"
                     "      n_pixels\n") < 0))
            {
                IAS_LOG_ERROR("Writing RMS column titles without "
                              "percent metric");
                return ERROR;
            }
        }

        for (sca_index = 0; sca_index < summary.scas - 1; sca_index++)
        {
            if (output_pcnt_flags[0] != NO_OUTPUT_PCNT_METRIC)
            {
                status = fprintf(rpt_fptr,
                    "%02d/%02d   %15.6f %15.6f %15.6f %11d\n",
                    summary.cross_sca_rms[sca_index].sca1_number,
                    summary.cross_sca_rms[sca_index].sca2_number,
                    summary.cross_sca_rms[sca_index].cross_mean,
                    summary.cross_sca_rms[sca_index].cross_stdev,
                    summary.cross_sca_rms[sca_index].cross_pcnt,
                    summary.cross_sca_rms[sca_index].cross_num_pixels);
                if (status < 0)
                {
                    IAS_LOG_ERROR("Writing RMS banding metrics information "
                                  "with percent metric");
                    return ERROR;
                }
            }
            else
            {
                status = fprintf(rpt_fptr,
                    "%02d/%02d   %15.6f %15.6f %11d\n",
                    summary.cross_sca_rms[sca_index].sca1_number,
                    summary.cross_sca_rms[sca_index].sca2_number,
                    summary.cross_sca_rms[sca_index].cross_mean,
                    summary.cross_sca_rms[sca_index].cross_stdev,
                    summary.cross_sca_rms[sca_index].cross_num_pixels);
                if (status < 0)
                {
                    IAS_LOG_ERROR("Writing RMS banding metrics information "
                                  "without percent metric");
                    return ERROR;
                }
            }
        }

        status = fprintf(rpt_fptr, "\n\n\n");
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing metrics group separation");
            return ERROR;
        }
    }

    /* Print stdev banding metric results next */
    if (summary.stdev_banding_flag)
    {
        if ((fprintf(rpt_fptr, "STDEV BANDING\n\n") < 0)
            || (fprintf(rpt_fptr, "per-detector stdev:   %15.6lf\n",
                summary.stdev_band_stdev) < 0)
            || (fprintf(rpt_fptr, "Banding across SCA discontinuity\n")
                < 0))
        {
            IAS_LOG_ERROR("Writing stdev summary header");
            return ERROR;
        }


        if (output_pcnt_flags[1] != NO_OUTPUT_PCNT_METRIC)
        {
            if((fprintf(rpt_fptr, "  SCA         BandingStdev"
                    "    BandingStdev\n") < 0)
                || (fprintf(rpt_fptr, "boundary       mean [rad]"
                    "      mean [%%]       n_pixels\n") < 0))
            {
                IAS_LOG_ERROR("Writing stdev column titles with percent "
                              "metric");
                return ERROR;
            }
        }
        else
        {
            if ((fprintf(rpt_fptr, "  SCA         BandingStdev\n") < 0)
                || (fprintf(rpt_fptr, "boundary       mean [rad]"
                    "      n_pixels\n") < 0))
            {
                IAS_LOG_ERROR("Writing stdev column titles without "
                              "percent metric");
                return ERROR;
            }
        }

        for (sca_index = 0; sca_index < summary.scas - 1; sca_index++)
        {
            if (output_pcnt_flags[1] != NO_OUTPUT_PCNT_METRIC)
            {
                status = fprintf(rpt_fptr, "%02d/%02d   %15.6f %15.6f %11d\n",
                    summary.cross_sca_stdev[sca_index].sca1_number,
                    summary.cross_sca_stdev[sca_index].sca2_number,
                    summary.cross_sca_stdev[sca_index].cross_mean,
                    summary.cross_sca_stdev[sca_index].cross_pcnt,
                    summary.cross_sca_stdev[sca_index].cross_num_pixels);
                if (status < 0)
                {
                    IAS_LOG_ERROR("Writing stdev banding summary information "
                                  "with percent metric");
                    return ERROR;
                }
            }
            else
            {
                status = fprintf(rpt_fptr, "%02d/%02d   %15.6f %11d\n",
                    summary.cross_sca_stdev[sca_index].sca1_number,
                    summary.cross_sca_stdev[sca_index].sca2_number,
                    summary.cross_sca_stdev[sca_index].cross_mean,
                    summary.cross_sca_stdev[sca_index].cross_num_pixels);
                if (status < 0)
                {
                    IAS_LOG_ERROR("Writing stdev banding summary information "
                                  "without percent metric");
                    return ERROR;
                }
            }
        }

        status = fprintf(rpt_fptr, "\n\n\n");
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing metrics group separation");
            return ERROR;
        }
    }

    /* Now print streaking results */
    if (summary.streaking_flag)
    {
        if ((fprintf(rpt_fptr, "STREAKING\n\n") < 0)
            || (fprintf(rpt_fptr, "per-detector_stdev:   %15.6lf\n",
                summary.streaking_stdev) < 0)
            || (fprintf(rpt_fptr, "Within-SCA streaking\n") < 0))
        {
            IAS_LOG_ERROR("Writing streaking header");
            return ERROR;
        }

        if (output_pcnt_flags[2] != NO_OUTPUT_PCNT_METRIC)
        {
            if ((fprintf(rpt_fptr, "                Streaking"
                    "       Streaking\n") < 0)
                || (fprintf(rpt_fptr, "SCA              mean [rad]      "
                    "mean [%%]       n_pixels\n") < 0))
            {
                IAS_LOG_ERROR("Writing streaking column titles with percent "
                              "metric");
                return ERROR;
            }
        }
        else
        {
            if ((fprintf(rpt_fptr, "                 Streaking\n") < 0)
                || (fprintf(rpt_fptr, "SCA              mean [rad]      "
                    "n_pixels\n") < 0))
            {
                IAS_LOG_ERROR("Writing streaking column titles without percent"
                              " metric");
                return ERROR;
            }
        }

        for (sca_index = 0; sca_index < summary.scas; sca_index++)
        {
            if (output_pcnt_flags[2] != NO_OUTPUT_PCNT_METRIC)
            {
                status = fprintf(rpt_fptr, "%02d        %15.6f %15.6f %11d\n",
                        summary.sca_summary_streaking[sca_index].sca_number,
                        summary.sca_summary_streaking[sca_index].metric_mean,
                        summary.sca_summary_streaking[sca_index].pcnt,
                        summary.sca_summary_streaking[sca_index]
                        .num_characterized_detectors);
                if (status < 0)
                {
                    IAS_LOG_ERROR("Writing streaking summary metrics "
                            "information with percent metric");
                    return ERROR;
                }
            }
            else
            {
                status = fprintf(rpt_fptr, "%02d        %15.6f %11d\n",
                        summary.sca_summary_streaking[sca_index].sca_number,
                        summary.sca_summary_streaking[sca_index].metric_mean,
                        summary.sca_summary_streaking[sca_index]
                        .num_characterized_detectors);
                if (status < 0)
                {
                    IAS_LOG_ERROR("Writing streaking summary metrics "
                            "information without percent metric");
                    return ERROR;
                }
            }
        }
    }

    /* Things seem to have been written OK */
    return SUCCESS;
}

/*------------------------------------------------------------------------
 NAME:            write_summary_report_data

 PURPOSE:  Writes band-level optional summary metrics information to
           the summary report file

 RETURNS:  Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int write_summary_report_data
(
    FILE *rpt_fptr,                         /* I: Open report file handle */
    RPS_NUCHAR_REPORT_SUMMARY summary,      /* I: SCA-specific metrics
                                               summary information */
    double band_radiance_mean,              /* I: Summary radiance
                                               statistics for the current
                                               band */
    double band_radiance_stdev,
    double band_ffov_metric                 /* I: Band-level FFOV
                                               nonuniformity metric */
)
{
    int status;
    int output_pcnt_flags[3] = {OUTPUT_PCNT_METRIC, OUTPUT_PCNT_METRIC,
                                NO_OUTPUT_PCNT_METRIC};

    /* Print the overall results header */
    status = fprintf(rpt_fptr, "RESULTS FOR ALL DETECTORS\n\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing results header to report");
        return ERROR;
    }

    /* Print the summary FFOV metric results for the current band */
    status = fprintf(rpt_fptr, "Full Field of View Metric\n"
        "             FullFieldMean     FullFieldStdev  FullFieldStdev\n"
        "Band           [rad]             [rad]            [%%]\n"
        "%02d       %15.6f %15.6f %15.6f\n\n\n\n",
        summary.band_number, band_radiance_mean,
        band_radiance_stdev, band_ffov_metric);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing estimated FFOV metric to report");
        return ERROR;
    }

    /* Write the summary metrics */
    status = write_summary_metrics_data(rpt_fptr, summary, band_radiance_mean,
             output_pcnt_flags);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing summary metrics data to report file");
        return ERROR;
    }

    /* Everything appears to have printed OK */
    return SUCCESS;
}
