/*==========================================================================
NAME:
    bias_model_calibration

PURPOSE:
    See usage() for a detailed description

ALGORITHM REFERENCES: Bias Model Calibration ADD
============================================================================*/
/* System Includes */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <inttypes.h>
#include <sys/types.h>
#include <sys/stat.h>

/* IAS Includes */
#include "ias_bpf.h"
#include "ias_cpf.h"
#include "ias_db.h"
#include "ias_db_get_connect_info.h"
#include "ias_l0r_constants.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_radiometry.h"
#include "ias_rps_db_bias_model_calibration.h"

/* Local Defines */
#define CPS_LENGTH       256
#define DATETIME_LEN     19               /* YYYY-MM-DDThh:mm:ss */
#define DATETIME_SIZE    (DATETIME_LEN+1)
#define DEFAULT_N0_VALUE 1
#define DEFAULT_N1_VALUE 40


/*---------------------------------------------------------------------------
 NAME:       get_reference_cpf_parameters

 PURPOSE:    Reads the reference CPF and extracts the OLI integration time
             and file name parameters.

 RETURNS:    Integer status code of SUCCESS or ERROR
----------------------------------------------------------------------------*/
int get_reference_cpf_parameters
(
    const char *refCPFName,           /* I: Name of reference CPF */
    double  *integrationTimeMS,       /* O: MS band integration time */
    double  *integrationTimePan,      /* O: PAN band integration time */
    char    *attrCPFName,             /* O: CPF file name from the
                                         FILE_ATTRIBUTES group */
    int     sizeofAttrCPFName         /* I: Size (in bytes) of the
                                         attribute file name buffer */
)
{
    IAS_CPF *cpf;
    const struct IAS_CPF_OLI_PARAMETERS  *oliParms;
    const struct IAS_CPF_FILE_ATTRIBUTES *fileAttrs;
    int status;


    /* Read in the CPF */
    cpf = ias_cpf_read(refCPFName);
    if (cpf == NULL)
    {
        IAS_LOG_ERROR("Reading CPF file '%s'", refCPFName);
        return ERROR;
    }

    /* Get the required integration times and file_name attribute (so we can
       rename it to the "correct" name if we need to).  */
    oliParms = ias_cpf_get_oli_parameters(cpf);
    if (oliParms == NULL)
    {
        IAS_LOG_ERROR("Retrieving OLI parameter information in '%s'",
            refCPFName);
        ias_cpf_free(cpf);
        return ERROR;
    }
    *integrationTimeMS = oliParms->nominal_integration_time_ms;
    *integrationTimePan = oliParms->nominal_integration_time_pan;

    fileAttrs = ias_cpf_get_file_attributes(cpf);
    if (fileAttrs == NULL)
    {
        IAS_LOG_ERROR("Retrieving file attributes information in '%s'",
            refCPFName);
        ias_cpf_free(cpf);
        return ERROR;
    }
    status = snprintf(attrCPFName, sizeofAttrCPFName,
        "%s", fileAttrs->file_name);
    if ((status < 0) || (status >= sizeofAttrCPFName))
    {
        IAS_LOG_ERROR("Copying file name from file attributes parameters "
            "in CPF '%s'", refCPFName);
        ias_cpf_free(cpf);
        return ERROR;
    }

    /* Clean up */
    ias_cpf_free(cpf);

    return SUCCESS;
}


/*---------------------------------------------------------------------------
 NAME:       get_reference_cpf

 PURPOSE:    Retrieves a CPF closest in time to the specified
             date.

 RETURNS:    Integer status code of SUCCESS or ERROR
----------------------------------------------------------------------------*/
int get_reference_cpf
(
    const char *refCPFName,                 /* I: Local disk name for
                                               reference CPF file
                                               retrieved by the service */
    const char *refDate                     /* I: Date used to select
                                               reference CPF */
)
{
    char *iasServicesPath;
    char command[2000];
    char url[1000];
    int  status;


    /* Get the current 'IAS_SERVICES' environment. */
    iasServicesPath = getenv("IAS_SERVICES");
    if (iasServicesPath == NULL)
    {
        IAS_LOG_ERROR("Getting 'IAS_SERVICES' environment setting");
        return ERROR;
    }

    /* Construct the URL where the CPF is located */
    status = snprintf(url, sizeof(url),
        "%s/dataservices/rest/cpf?operation=getCalParmFile&nearMatch=true&"
        "date=%s", iasServicesPath, refDate);
    if ((status < 0) || (status >= sizeof(url)))
    {
        IAS_LOG_ERROR("Constructing URL to retrieve CPF");
        return ERROR;
    }

    /* Display the URL and CPF date for debugging purposes */
    IAS_LOG_DEBUG("IAS_SERVICES URL:  %s", url);
    IAS_LOG_DEBUG("CPF Date:          %s", refDate);

    /* Retrieve the CPF */
    status = snprintf(command, sizeof(command),
        "wget --quiet -O \"%s\" \"%s\"", refCPFName, url);
    if ((status < 0) || (status >= sizeof(command)))
    {
        IAS_LOG_ERROR("Constructing CPF retrieval command");
        return ERROR;
    }
    status = system(command);
    if (status != 0)
    {
        IAS_LOG_ERROR("Retrieving CPF specified by date %s", refDate);
        return ERROR;
    }

    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME:     create_query

PURPOSE:  Create the SQL query to retrieve data from the
          RADIOMETRIC_WORK_ORDER_COMMON table.

RETURNS:
    This function returns a status of SUCCESS if the query is successfully
    generated, ERROR if the specified string is not large enough.
------------------------------------------------------------------------------*/
int create_query
(
    char *dateWhereClause,
                        /* I: Date-specific portion of the where clause */
    double integrationTimeMS,
                        /* I: OLI MS band nominal integration time */
    double integrationTimePan,
                        /* I: OLI PAN band nominal integration time */
    int size,           /* I: Size of the query buffer */
    char * query        /* O: Starting portion of the common query */
)
{
    char commonWhereClause[384];
    int bytes;
    int status;

    /* Generate a common where clause substring that includes conditions
       for the validity flag, collection_type, and integration times.  */
    status = ias_rps_db_create_oli_bias_model_where_clause(
        commonWhereClause, sizeof(commonWhereClause),
        "bias_frame_averages", integrationTimeMS, integrationTimePan);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Constructing query 'where' clause");
        return ERROR;
    }

    /* Start the database query.  All three user scenarios have the same
       initial query with differences in the where clause and the number of
       results that must be read. */
    bytes = snprintf(query, size,
        "SELECT RADIOMETRIC_WO_COMMON_ID, CPF_NAME, LANDSAT_INTERVAL_ID "
        "FROM RADIOMETRIC_WORK_ORDER_COMMON "
        "WHERE %s "
        "AND %s", commonWhereClause, dateWhereClause);
         if ((bytes < 0) || (bytes >= size))
    {
        IAS_LOG_ERROR("String is too small to create SQL query");
        return ERROR;
    }

    IAS_LOG_DEBUG("Query for shutter collects:\n%s\n", query);
    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: db_open

PURPOSE:
    Open a connection to the IAS database

RETURNS:
    This function returns a pointer to an ias_db_connection structure if
    successful, a NULL pointer if the database cannot be opened.
------------------------------------------------------------------------------*/
struct ias_db_connection * db_open ()
{
    struct ias_db_connection * db = NULL;

    char * dbHost     = NULL;
    char * dbName     = NULL;
    char * dbPassword = NULL;
    char * dbUser     = NULL;
    int    status;

    /* Initialize the IAS database library and retrieve the connection info */
    status = ias_db_initialize_database_lib();
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the database library");
        return NULL;
    }
    status = ias_db_get_connect_info(&dbName, &dbUser, &dbPassword, &dbHost);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving database connection info");
        ias_db_close_database_lib();
        return NULL;
    }
    db = ias_db_connect_to_database(dbName, dbUser, dbPassword, dbHost);

    /* Free all of the dynamically allocated memory */
    free(dbHost);
    free(dbName);
    free(dbPassword);
    free(dbUser);

    if (db == NULL)
    {
        IAS_LOG_ERROR("Connecting to the database");
        ias_db_close_database_lib();
    }
    return db;
}


/*------------------------------------------------------------------------------
NAME: db_query_end

PURPOSE:
    Query the desired shutter collects from the RADIOMETRIC_WORK_ORDER_COMMON
    table given a desired ending date/time and the number of previous shutter
    collects to be used in the bias model calibration.

RETURNS:
    This function will return an updated number of previous shutter collects
    that are available if the requested number of collects cannot be located.
    It will also return the following information for the last shutter collect
    in the group (e.g. post shutter collect): the name of the CPF that was used
    during processing, the interval ID of the collect, the work order ID used
    to process the collect, and the acquisition date/time.

    The function will return a status of SUCCESS if a post collect and at least
    one previous shutter collect are located.  It will return a status of ERROR
    required collects cannot be located.
------------------------------------------------------------------------------*/
int db_query_end
(
    struct ias_db_connection * db, /* I: Database connection */
    const char *dateEnd,           /* I: End date/time of interest */
    double integrationTimeMS,      /* I: MS band nominal integration time */
    double integrationTimePan,     /* I: PAN band nominal integration time */
    int * nPrev,                   /* IO: Number of previous collects found*/
    int * rwocid,                  /* O: Radiometric WO Common ID of post
                                      shutter */
    char cpfName[PATH_MAX],        /* O: CPF name for post shutter */
    char intId[IAS_L0R_INTERVAL_ID_LENGTH +1 ]
                                   /* O: Interval ID of post shutter */
)
{
    char                  cpf[PATH_MAX];
    char                  intervalId[IAS_L0R_INTERVAL_ID_LENGTH + 1];
    char                  sql[1024];
    char                  where[256];
    int                   radiometricWorkOrderCommonId;
    int                   foundPostCollect = FALSE;
    int                   row              = 0;
    int                   status;
    struct ias_db_query * query            = NULL;

    /* Create the additional components of the WHERE clause that are specific
       to this function */
    status = snprintf(where, sizeof(where),
        "(DATE_ACQUIRED <= to_date('%s', 'YYYY-MM-DD HH24:MI:SS')) ORDER BY "
        "DATE_ACQUIRED DESC",
        dateEnd);
    if ((status < 0) || (status >= sizeof(where)))
    {
        IAS_LOG_ERROR("String is too small to create SQL WHERE clause");
        return ERROR;
    }

    /* Create the full query for this function */
    if (create_query(where, integrationTimeMS, integrationTimePan,
        sizeof(sql), sql) != SUCCESS)
    {
        IAS_LOG_ERROR("String is too small to create SQL query");
        return ERROR;
    }

    /* Perform the SQL query we developed above */
    query = ias_db_perform_query(db, sql);
    if (!ias_db_query_was_successful(query))
    {
        char dbErrMsg[1024];

        ias_db_query_get_error_message(query, dbErrMsg, sizeof(dbErrMsg));
        IAS_LOG_ERROR("%s",dbErrMsg);
        IAS_LOG_ERROR("SQL query: %s", sql);
        ias_db_query_close(query);
        return ERROR;
    }

    /* Retrieve the results of the query */
    while (ias_db_query_next(query))
    {
        /* This variable resets to 0 in each iteration of the loop.
           It's used as an 'index' to the position of a particular piece
           of data in the current row returned from the query (i.e.
           radiometric work order ID number, CPF name, or Landsat interval
           ID).*/
        int i = 0;

        /* Retrieve the values we asked for. */
        status =
            ias_db_query_int_value(query, &radiometricWorkOrderCommonId, i++);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining radiometric work order common ID from "
                "database query");
            ias_db_query_close(query);
            return ERROR;
        }
        ias_db_query_string_value(query, i++, cpf, PATH_MAX);
        ias_db_query_string_value(query, i++, intervalId,
            IAS_L0R_INTERVAL_ID_LENGTH + 1);

        /* This query is reading the shutter collects in reverse chronological
           order so the first collect read is the post collect, and we want to
           return its CPF name, interval ID, and work order ID. */
        if (!foundPostCollect)
        {
            /* Save the information and display it. */
            *rwocid = radiometricWorkOrderCommonId;
            strcpy(intId, intervalId);
            strcpy(cpfName, cpf);

            IAS_LOG_INFO("Post collect radiometric work order common ID: "
                "%d, Landsat Interval ID: %s, CPF: %s",
                *rwocid, intId, cpfName);

            /* Flag this as the post collect */
            foundPostCollect = TRUE;
        }
        else
        {
            /* Display the information for the current pre collect. */
            IAS_LOG_INFO("Pre collect radiometric work order common ID: "
                "%d, Landsat Interval ID: %s, CPF: %s",
                radiometricWorkOrderCommonId, intervalId, cpf);

            /* Increment the row count. */
            row++;
        }

        /* If we have located the specified number of collects, we're done */
        if (row >= *nPrev)
            break;
    }

    ias_db_query_close(query);

    /* If no data was returned, it's an error */
    if (row == 0)
    {
        IAS_LOG_ERROR("No shutter collects found prior to %s", dateEnd);
        return ERROR;
    }

    /* If we didn't locate the number of previous collects requested, update
       nPrev to reflect the number found */
    if (row < *nPrev)
    {
        IAS_LOG_INFO("Previous shutter collects reset from %d to %d", *nPrev,
            row);
        *nPrev = row;
    }

    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: db_query_range

PURPOSE:
    Query all of the shutter collects from the RADIOMETRIC_WORK_ORDER_COMMON
    table given a desired date range so that they can be used in the bias model
    calibration.

    When given a date range, the last shutter collect in the range will be
    considered the post shutter collect and the rest of the collects in that
    date range will be considered the prior shutter collects.  Therefore the
    bias model parameters that are generated will be for any earth images that
    occur between the last shutter collect and the second to the last collect in
    the range.

RETURNS:
    This function will return the number of previous shutter collects that are
    available within the specified date range.  The number of previous collects
    will exclude the last shutter collect which will be considered the post
    collect.  It will also return the following information for the post shutter
    collect in the group: the name of the CPF that was used during processing,
    the interval ID of the collect, the work order ID used to process that
    collect, and the acquisition date/time.

    The function will return a status of SUCCESS if a post collect and at least
    one previous shutter collect are located in the date range.  It will return
    a status of ERROR required collects cannot be located.
------------------------------------------------------------------------------*/
int db_query_range
(
    struct ias_db_connection * db, /* I: Database connection */
    const char *dateStart,  /* I: Start date/time of interest */
    const char *dateEnd,    /* I: End date/time of interest */
    double integrationTimeMS,  /* I: Nominal MS band integration time */
    double integrationTimePan, /* I: Nominal PAN band integration time */
    int * nPrev,            /* O: Number of previous collects found */
    int * rwocid,           /* O: Radiometric WO Common ID of post shutter */
    char cpfName[PATH_MAX], /* O: CPF name for post shutter */
    char intId[IAS_L0R_INTERVAL_ID_LENGTH + 1]
                            /* O: Interval ID of post shutter*/
)
{
    struct ias_db_query * query = NULL;
    char                  cpf[PATH_MAX];
    char                  intervalId[IAS_L0R_INTERVAL_ID_LENGTH + 1];
    char                  sql[1024];
    char                  where[256];
    int  row                    = 0;
    int  radiometricWorkOrderCommonId;
    int  status;

    *nPrev = 0;

    /* Create the additional components of the WHERE clause that are specific
       to this function */
    status = snprintf(where, sizeof(where),
        "(DATE_ACQUIRED >= to_date('%s', 'YYYY-MM-DD HH24:MI:SS')) AND "
        "(DATE_ACQUIRED <= to_date('%s', 'YYYY-MM-DD HH24:MI:SS')) ORDER BY "
        "DATE_ACQUIRED DESC",
        dateStart, dateEnd);
    if ((status < 0) || (status >= sizeof(where)))
    {
        IAS_LOG_ERROR("String is too small to create SQL WHERE clause");
        return ERROR;
    }

    /* Create the full query for this function */
    if (create_query(where, integrationTimeMS, integrationTimePan,
        sizeof(sql), sql) != SUCCESS)
    {
        IAS_LOG_ERROR("String is too small to create SQL query");
        return ERROR;
    }

    /* Perform the SQL query we developed above */
    query = ias_db_perform_query(db, sql);
    if (!ias_db_query_was_successful(query))
    {
        char dbErrMsg[1024];

        ias_db_query_get_error_message(query, dbErrMsg, sizeof(dbErrMsg));
        IAS_LOG_ERROR("%s",dbErrMsg);
        IAS_LOG_ERROR("SQL query: %s", sql);
        ias_db_query_close(query);
        return ERROR;
    }

    /* Retrieve the results of the query */
    while (ias_db_query_next(query))
    {
        /* This variable resets to 0 in each iteration of the loop.
           It's used as an 'index' to the position of a particular piece
           of data in the current row returned from the query (i.e.
           radiometric work order ID number, CPF name, or Landsat interval
           ID).*/
        int i = 0;

        /* Retrieve the values we asked for */
        status =
            ias_db_query_int_value(query, &radiometricWorkOrderCommonId, i++);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining radiometric work order common ID from "
                "database query");
            ias_db_query_close(query);
            return ERROR;
        }
        ias_db_query_string_value(query, i++, cpf, PATH_MAX);
        ias_db_query_string_value(query, i++, intervalId,
            IAS_L0R_INTERVAL_ID_LENGTH + 1);

        /* This query is reading the shutter collects in reverse chronological
           order so the first collect read is the post collect, and we want to
           return its CPF name, interval ID, and work order ID. */
        if (row == 0)
        {
            /* Save the information and display it. */
            *rwocid = radiometricWorkOrderCommonId;
            strcpy(intId, intervalId);
            strcpy(cpfName, cpf);

            IAS_LOG_INFO("Post collect radiometric work order common ID: "
                "%d, Landsat Interval ID: %s, CPF: %s",
                *rwocid, intId, cpfName);
        }
        else
        {
            /* Display the remaining pre collect information. */
            IAS_LOG_INFO("Pre collect radiometric work order common ID: "
                "%d, Landsat Interval ID: %s, CPF: %s",
                radiometricWorkOrderCommonId, intervalId, cpf);
        }

        /* Increment the row count. */
        row++;
    }

    ias_db_query_close(query);
    if (row == 0)
    {
        IAS_LOG_ERROR("No shutter collects found in date range");
        return ERROR;
    }

    /* This function must locate at least 2 shutter collects.  The last shutter
       collect will be considered the post collect and all others will be
       considered previous collects. */
    if (row < 2)
    {
        IAS_LOG_ERROR("Fewer than 2 shutter collects located in date range");
        return ERROR;
    }

    *nPrev = row - 1;
    IAS_LOG_INFO("Found %d previous shutter collects", *nPrev);
    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: db_query_start

PURPOSE:
    Query the desired shutter collects from the RADIOMETRIC_WORK_ORDER_COMMON
    table given a desired starting date/time and the number of previous shutter
    collects to be used in the bias model calibration.

RETURNS:
    This function will return an updated number of shutter collects that are
    available if the requested number of collects are not available.  It will
    also return the following information for the last shutter collect in the
    group: the name of the CPF that was used during processing, the interval ID
    of the collect, the work order ID used to process the collect, and the
    acquisition date/time.

    The function will return a status of SUCCESS if a post collect and at least
    one previous shutter collect are located.  It will return a status of ERROR
    required collects cannot be located.
------------------------------------------------------------------------------*/
int db_query_start
(
    struct ias_db_connection * db,  /* I: Database connection */
    const char *dateStart,          /* I: Start date/time of interest */
    double integration_time_ms,     /* I: OLI MS band nominal integration
                                       time */
    double integration_time_pan,    /* I: OLI PAN band nominal integration
                                       time */
    int * nPrev,                    /* IO: Number of previous collects */
    int * rwocid,                   /* O: Radiometric WO Common ID of post
                                       shutter */
    char cpfName[PATH_MAX],         /* O: CPF name for post shutter */
    char intId[IAS_L0R_INTERVAL_ID_LENGTH + 1]
                                    /* O: Interval ID of post shutter*/
)
{
    struct ias_db_query * query = NULL;

    char cpf[PATH_MAX];
    char intervalId[IAS_L0R_INTERVAL_ID_LENGTH + 1];
    char sql[1024];
    char where[256];
    int  foundPostCollect                         = FALSE;
    int  radiometricWorkOrderCommonId             = 0;
    int  row                                      = 0;
    int  status;

    /* Create the additional components of the WHERE clause that are specific
       to this function */
    status = snprintf(where, sizeof(where),
        "(DATE_ACQUIRED >= to_date('%s', 'YYYY-MM-DD HH24:MI:SS')) ORDER BY "
        "DATE_ACQUIRED ASC",
        dateStart);
    if ((status < 0) || (status >= sizeof(where)))
    {
        IAS_LOG_ERROR("String is too small to create SQL WHERE clause");
        return ERROR;
    }

    /* Create the full query for this function */
    if (create_query(where, integration_time_ms, integration_time_pan,
        sizeof(sql), sql) != SUCCESS)
    {
        IAS_LOG_ERROR("String is too small to create SQL query");
        return ERROR;
    }

    /* Perform the SQL query we developed above */
    query = ias_db_perform_query(db, sql);
    if (!ias_db_query_was_successful(query))
    {
        char dbErrMsg[1024];

        ias_db_query_get_error_message(query, dbErrMsg, sizeof(dbErrMsg));
        IAS_LOG_ERROR("%s",dbErrMsg);
        IAS_LOG_ERROR("SQL query: %s", sql);
        ias_db_query_close(query);
        return ERROR;
    }

    /* Retrieve the results of the query */
    while (ias_db_query_next(query))
    {
        /* This variable resets to 0 in each iteration of the loop.
           It's used as an 'index' to the position of a particular piece
           of data in the current row returned from the query (i.e.
           radiometric work order ID number, CPF name, or Landsat interval
           ID).*/
        int i = 0;

        /* Retrieve the values we asked for */
        status = ias_db_query_int_value(query, &radiometricWorkOrderCommonId,
            i++);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR(
                "Obtaining radiometric work order ID from database query");
            ias_db_query_close(query);
            return ERROR;
        }
        ias_db_query_string_value(query, i++, cpf, PATH_MAX);
        ias_db_query_string_value(query, i++, intervalId,
            IAS_L0R_INTERVAL_ID_LENGTH + 1);

        /* As with the end date query, use a flag variable
           'foundPostCollect' to indicate the first collect found is the
           post collect (even though it isn't).  The row count is not
           incremented in this first pass through the loop.  It is
           incremented in subsequent passes through the loop (including the
           time the retrieved row is for the "true" post collect).  When the
           loop finally terminates, the number of rows will be at the
           "correct" value for the number of pre collects that were
           actually found.  A neat trick...  */
        if (foundPostCollect)
            row++;
        else
            foundPostCollect = TRUE;

        /* If we've found enough previous collects, we're done */
        if (row >= *nPrev)
            break;

        /* Display the current pre collect information. */
        IAS_LOG_INFO("Pre collect radiometric work order common ID: "
            "%d, Landsat Interval ID: %s, CPF: %s",
            radiometricWorkOrderCommonId, intervalId, cpf);

    }

    ias_db_query_close(query);

    if (row == 0)
    {
        IAS_LOG_ERROR("No shutter collects found after %s", dateStart);
        return ERROR;
    }

    /* All of the shutter collects from the query were read in chronological
       order and the last collect found is the post collect that we are
       interested in. */
    *rwocid = radiometricWorkOrderCommonId;
    strcpy(cpfName, cpf);
    strcpy(intId, intervalId);

    /* Display the post collect information. */
    IAS_LOG_INFO("Post collect radiometric work order common ID: %d, "
        "Landsat interval ID: %s, CPF: %s",
        *rwocid, intId, cpfName);

    /* If we didn't locate the number of collects requested, update nPrev to
       reflect the number of collects found */
    if (row < *nPrev)
    {
        IAS_LOG_INFO("Previous shutter collects reset from %d to %d", *nPrev,
            row);
        *nPrev = row;
    }
    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: fetch_cpf

PURPOSE:
    Retrieve a CPF from the CPS and store it locally on disk.

RETURNS:
    SUCCESS if the CPF is successfully retrieved, ERROR if an error is
    encountered.
------------------------------------------------------------------------------*/
int fetch_cpf
(
    const char cpfName[PATH_MAX]        /* I: Name of a CPF */
)
{
    char   cmd[PATH_MAX * 2 + CPS_LENGTH + 200];
    char   cps[CPS_LENGTH];
    char * pEnv;
    int    nChars;

    /* Obtain the webserver for the CPS */
    if ((pEnv = getenv("IAS_SERVICES")) == NULL)
    {
        IAS_LOG_ERROR("Environment variable IAS_SERVICES not defined");
        return ERROR;
    }
    if (strlen(pEnv) >= sizeof(cps))
    {
        long unsigned my_size = (long unsigned) sizeof(cps);
        IAS_LOG_ERROR("IAS_SERVICES exceeds buffer size: %lu", my_size);
        return ERROR;
    }
    strcpy(cps, pEnv);

    /* Create the wget command to retrieve the CPF from the CPS */
    nChars = snprintf(cmd, sizeof(cmd), "wget --quiet -O %s "
        "'%s/dataservices/rest/cpf?operation=getCalParmFile&filename=%s'",
        cpfName, cps, cpfName);
    if ((nChars < 0) || (nChars >= sizeof(cmd)))
    {
        long unsigned my_size = (long unsigned) sizeof(cmd);
        IAS_LOG_ERROR("Command to retrieve CPF is larger than buffer: %lu",
            my_size);
        return ERROR;
    }

    /* Do the work */
    IAS_LOG_DEBUG("Fetch CPF with: %s", cmd);
    if (system(cmd) != 0)
    {
        IAS_LOG_ERROR("Retrieving CPF: %s", cmd);
        return ERROR;
    }

    return SUCCESS;
}

/*------------------------------------------------------------------------------
NAME: get_date

PURPOSE:
    Parse a date/time value from the list of command line arguments.  It is
    acceptable for the user to specify this value as a single string in the
    format "YYYY/MM/DD hh:mm:ss" or as two separate strings in the argument
    list; "YYYY/MM/DD" and "hh:mm:ss".

RETURNS:
    A status of SUCCESS if a date/time is successfully parsed or ERROR if a
    valid date/time is not found.
------------------------------------------------------------------------------*/
int get_date
(
    int            argc,      /* I: Argument count from main() */
    char *         argv[],    /* I: Argument vector from main() */
    int            position,  /* I: Current position in the argument list */
    int *          nArgs,     /* O: Argument consumed */
    IAS_DATETIME * pDateTime, /* O: User specified date and time */
    double *       pTimeStamp /* O: Integer timestamp YYYYMMDDhhmmss */
)
{
    const int kDateMaxLen = strlen("YYYY-MM-DD"); /* Maximum date length */
    const int kDateMinLen = strlen("YYYY-M-D");   /* Minimum date length */
    const int kTimeMaxLen = strlen("hh:mm:ss");   /* Maximum time length */
    const int kTimeMinLen = strlen("h:m:s");      /* Minimum time length */

    const int kDateTimeMaxLen = (kDateMaxLen + kTimeMaxLen + 1);
    const int kDateTimeMinLen = (kDateMinLen + kTimeMinLen + 1);

    char combined[kDateTimeMaxLen+1];
    int  lengthArg1;
    int  lengthArg2;

    *nArgs = 0;
    *pTimeStamp = 0;

    /* There must be at least one argument */
    if (position >= argc)
    {
        IAS_LOG_ERROR("Too few arguments specified for date/time");
        return ERROR;
    }
    lengthArg1 = strlen(argv[position]);

    if (lengthArg1 >= kDateTimeMinLen)
    {
        /* If the first argument is long enough to hold both the date and time
           then call the IAS routine to parse a date/time. */
        if (ias_misc_parse_datetime_string(argv[position],
            IAS_DATETIME_CPF_FORMAT, pDateTime) != SUCCESS)
        {
            IAS_LOG_ERROR("Parsing date/time");
            return ERROR;
        }
        *nArgs = 1;     /* consumed one argument */
    }
    else if ((position + 1) >= argc)
    {
        /* If the first argument wasn't long enough for both a date and time
           value then there must be another argument available that holds the
           time. */
        IAS_LOG_ERROR("Too few arguments specified for date/time");
        return ERROR;
    }
    else
    {
        lengthArg2 = strlen(argv[position+1]);

        /* The date and time arguments must be within the acceptable lengths */
        if ((lengthArg1 < kDateMinLen) || (lengthArg1 > kDateMaxLen) ||
            (lengthArg2 < kTimeMinLen) || (lengthArg2 > kTimeMaxLen))
        {
            IAS_LOG_ERROR("Date or time argument malformed");
            return ERROR;
        }

        /* Combine the next 2 arguments and see if we can successfully parse a
           date and time.  The IAS method must have a 'T', 't', or ':' between
           the date and time fields.  You cannot use a space. */
        sprintf(combined, "%sT%s", argv[position], argv[position + 1]);
        if (ias_misc_parse_datetime_string(combined, IAS_DATETIME_CPF_FORMAT,
            pDateTime) != SUCCESS)
        {
            IAS_LOG_ERROR("Parsing date/time");
            return ERROR;
        }
        *nArgs = 2;     /* consumed two arguments */
    }

    /* Convert the date/time into a combined numeric value for easy
       comparisons */
    *pTimeStamp = pDateTime->year * 10000000000.0 +
        pDateTime->month * 100000000.0 +
        pDateTime->day_of_month * 1000000.0 +
        pDateTime->hour * 10000.0 +
        pDateTime->minute * 100.0 + pDateTime->second;
    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: set_bpf_file_attributes

PURPOSE:
    Since the library doesn't set all of the file attributes, do it ourselves.

RETURNS:
    This function returns a status of SUCCESS if it successfully updates the
    BPF, ERROR if an error is encountered.

NOTES:  The sensor_name field in the file attributes section is set in
        ias_rps_bias_model_calibration
------------------------------------------------------------------------------*/
int set_bpf_file_attributes
(
    char cmdLine[IAS_BPF_DESCRIPTION_LENGTH], /* I: Command line arguments */
    IAS_BPF * bpf,                            /* I/O: BPF to be updated */
    char bpfPath[PATH_MAX]                    /* I/O: Path of the BPF */
)
{
    const struct IAS_BPF_FILE_ATTRIBUTES * attr;

    IAS_DATETIME dateBegin;
    IAS_DATETIME dateEnd;
    char         bpfName[PATH_MAX];
    char         desc[IAS_BPF_DESCRIPTION_LENGTH];
    char         spacecraft[IAS_BPF_SPACECRAFT_NAME_LENGTH];
    char         tmpPath[PATH_MAX];
    int          satNum;
    int          status;

    /* Obtain the satellite number */
    satNum = ias_sat_attr_get_satellite_number();
    if (satNum == ERROR)
    {
        IAS_LOG_ERROR("Obtaining satellite number from satellite attributes "
           "library");
        return ERROR;
    }

    attr = ias_bpf_get_file_attributes(bpf);

    /* There must be valid effective dates */
    status = ias_misc_parse_datetime_string(attr->effective_date_begin,
        IAS_DATETIME_CPF_FORMAT, &dateBegin);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Invalid effective date begin in BPF: %s",
            attr->effective_date_begin);
        return ERROR;
    }
    status = ias_misc_parse_datetime_string(attr->effective_date_end,
        IAS_DATETIME_CPF_FORMAT, &dateEnd);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Invalid effective date end in BPF: %s",
            attr->effective_date_end);
        return ERROR;
    }

    /* Do not set the baseline date */

    /* Create the BPF's name */
    status = snprintf(bpfName, sizeof(bpfName),
        "LO%dBPF%04d%02d%02d%02d%02d%02d_%04d%02d%02d%02d%02d%02d.%02d",
        satNum, dateBegin.year, dateBegin.month, dateBegin.day_of_month,
        dateBegin.hour, dateBegin.minute, (int)dateBegin.second, dateEnd.year,
        dateEnd.month, dateEnd.day_of_month, dateEnd.hour, dateEnd.minute,
        (int)dateEnd.second, attr->version);
    if ((status < 0) || (status >= sizeof(bpfName)))
    {
        long unsigned my_size = (long unsigned) sizeof(bpfName);
        IAS_LOG_ERROR("BPF name exceeds %lu characters", my_size);
        return ERROR;
    }
    ias_bpf_set_file_attributes_file_name(bpf, bpfName);

    /* Do not set the file source */

    /* Set the spacecraft name */
    status = snprintf(spacecraft, sizeof(spacecraft), "Landsat_%d", satNum);
    if ((status < 0) || (status >= sizeof(spacecraft)))
    {
        long unsigned my_size = (long unsigned) sizeof(spacecraft);
        IAS_LOG_ERROR("Spacecraft name exceeds %lu characters",
            my_size);
        return ERROR;
    }
    ias_bpf_set_file_attributes_spacecraft_name(bpf, spacecraft);

    /* Set the description */
    status = snprintf(desc, sizeof(desc), "Generated via: %s", cmdLine);
    if ((status < 0) || (status >= sizeof(desc)))
    {
        long unsigned my_size = (long unsigned) sizeof(desc);
        IAS_LOG_ERROR("BPF description exceeds %lu characters", my_size);
        return ERROR;
    }
    ias_bpf_set_file_attributes_description(bpf, desc);

    /* Create the full path of the BPF */
    status = snprintf(tmpPath, sizeof(tmpPath), "%s/%s", bpfPath, bpfName);
    if ((status < 0) || (status >= sizeof(tmpPath)))
    {
        long unsigned my_size = (long unsigned) sizeof(tmpPath);
        IAS_LOG_ERROR("BPF path exceeds %lu characters", my_size);
        return ERROR;
    }
    strcpy(bpfPath, tmpPath);
    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: usage

PURPOSE:
    Provide a usage message to the user if they provide invalid command line
    options.

RETURNS:
    Nothing
------------------------------------------------------------------------------*/
void usage ()
{
    printf(
        "usage: bias_model_calibration [+cpf] [-cpf] [-d <dir>] [-end        \n"
        "       <dateTime>] [-n0 <n>] [-prev <n>] [-start <dateTime>]        \n"
        "                                                                    \n"
        "    This application can be used to create OLI Bias Parameter Files \n"
        "    (BPF).  These files are typically created by the Ingest         \n"
        "    subsystem when a shutter collect is processed.  The resulting   \n"
        "    OLI BPF will have an effective time period that spans from the  \n"
        "    acquisition time of a previous shutter collect through the      \n"
        "    acquisition time of the new shutter collect being processed.    \n"
        "                                                                    \n"
        "    The new shutter collect is consider the 'post' shutter collect  \n"
        "    and a certain number of previous shutter collects are used to   \n"
        "    generate the BPF.  When this application is run, all of the     \n"
        "    shutter collects will have been processed and will have an entry\n"
        "    in the RADIOMETRIC_WORK_ORDER_COMMON table.  There are 3 ways   \n"
        "    that a user can identify a set of shutter collects that should  \n"
        "    be used to create a BPF.                                        \n"
        "                                                                    \n"
        "    1) Specify a date range.  This application will identify all of \n"
        "       the shutter collects within the date range  and the last     \n"
        "       collect will be used as the post collect with all of the     \n"
        "       other collects being used as the previous collects.          \n"
        "                                                                    \n"
        "    2) Specify a starting date/time and a number of previous shutter\n"
        "       collects, <n>, to be used in creating a BPF.  In this        \n"
        "       scenario all of the shutter collects to be considered will   \n"
        "       have been acquired on or after the specified date and time.  \n"
        "       If <n>+1 shutter collects are available, the first <n>       \n"
        "       collects will be used as the previous collects and the <n>+1 \n"
        "       collect will be used as the post collect.                    \n"
        "                                                                    \n"
        "       If <n>+1 shutter collects do not exist in the database, the  \n"
        "       value of <n> will be adjusted and the last shutter collect   \n"
        "       available will be used as the post collect while the others  \n"
        "       will be used as the previous collects.  At least 2 shutter   \n"
        "       collects that were acquired on or after the specified time   \n"
        "       must exist so that the algorithm has 1 previous collect and 1\n"
        "       post collect.                                                \n"
        "                                                                    \n"
        "    3) Specify an ending date/time and the number of previous       \n"
        "       shutter collects, denoted by <n>.  In this scenario, the     \n"
        "       first shutter collect acquired on or prior to the specified  \n"
        "       date/time will be used as the post shutter collect.  The     \n"
        "       application will then continue searching backward to locate  \n"
        "       the desired number of previous shutter collects.             \n"
        "                                                                    \n"
        "    Date/time values may specified as two arguments in the form     \n"
        "    'YYYY-MM-DD' and 'hh:mm:ss' or as a single argument in the form \n"
        "    'YYYY-MM-DDThh:mm:ss'.                                          \n"
        "                                                                    \n"
        "    The output from this application will be a BPF file and 3 files \n"
        "    containing the a0 and a1 residual values.  The residual values  \n"
        "    will not be stored in the database.                             \n"
        "                                                                    \n"
        "    +cpf    Leave the CPF that was retrieved in the current         \n"
        "            directory                                               \n"
        "    -cpf    Remove the retrieved CPF when finished, default behavior\n"
        "    -d      Specify an output directory to place the BPF and        \n"
        "            residual files in.  The default is the current working  \n"
        "            directory.                                              \n"
        "    -end    Ending date and time                                    \n"
        "    -start  Starting date and time                                  \n"
        "    -n0     Number of pre-acquisition shutter collects to use for   \n"
        "            pre/post and A0.  See the bias model calibration ADD for\n"
        "            additional details.  The default is %d and must be >= 1.\n"
        "    -prev   Number of previous shutter collects to be used.  Refer  \n"
        "            to the bias model calibration ADD for additional        \n"
        "            details.  This value is the n1 value in the ADD and the \n"
        "            recommended value is %d and must be >= 1.               \n"
        "\n", DEFAULT_N0_VALUE, DEFAULT_N1_VALUE);
    exit(EXIT_FAILURE);
}


/*------------------------------------------------------------------------------
NAME: write_residuals_a0

PURPOSE:
    Write the A0 residual values to a file.

RETURNS:
    This function returns a status of SUCCESS if the residual file is
    successfully created, ERROR if an error is encountered.
------------------------------------------------------------------------------*/
int write_residuals_a0
(
    IAS_RPS_DB_BIAS_MODEL_RESIDUALS * residuals, /* I: Bias model residuals */
    const char * bpfPath /* I: Full path to the BPF file */
)
{
    FILE * fd;
    char   path[PATH_MAX];
    int    bandIndex;
    int    bandNum;
    int    nChars;
    int    nScas          = 0;
    int    printComma;
    int    sca;

    if (residuals == NULL)
    {
        IAS_LOG_ERROR("Residuals structure not allocated");
        return ERROR;
    }

    /* Create the paths for the associated residual values */
    nChars = snprintf(path, PATH_MAX, "%s-a0.txt", bpfPath);
    if ((nChars < 0) || (nChars >= PATH_MAX))
    {
        IAS_LOG_ERROR("Residual file path exceeds buffer size of %d characters",
            PATH_MAX);
        return ERROR;
    }

    /* Open the residual file for writing */
    fd = fopen(path, "w");
    if (fd == NULL)
    {
        IAS_LOG_ERROR("Opening file for writing: %s", path);
        return ERROR;
    }

    IAS_LOG_INFO("Creating A0 residual file: %s", path);

    /* The residual values for each band are written to the file in vertical
       columns to help with analysis in IDL or matlab */
    printComma = FALSE;
    for (bandIndex = 0; bandIndex < IAS_MAX_NBANDS; bandIndex++)
    {
        bandNum = residuals->band_residuals[bandIndex].band_number;
        if (bandNum == 0)
            continue;

        /* Determine the number of SCA's for all bands.  The format of this file
           requires that each band have the same number of SCA's.  If we
           encounter a difference, this report file will have to be revised. */
        sca = ias_sat_attr_get_scas_per_band(bandNum);
        if (sca == ERROR)
        {
            IAS_LOG_ERROR("Obtaining number of SCAs for band %d", bandNum);
            fclose(fd);
            return ERROR;
        }
        else if (nScas == 0)
            nScas = sca;
        else if (sca != nScas)
        {
            IAS_LOG_ERROR(
                "Number of SCAs for band %d does not match other bands",
                bandNum);
            fclose(fd);
            return ERROR;
        }

        /* Do we need to print a leading comma? */
        if (printComma)
            fprintf(fd, ", ");
        else
            printComma = TRUE;

        if (residuals->band_residuals[bandIndex].lineset[0] == 2)
        {
            /* Normal, 30 meter band: only use index 0 of the r_squared_a0 array
               */
            fprintf(fd, "Band%d", bandNum);
        }
        else
        {
            /* 15 meter PAN band: use both indices of the r_squared_a0 array.
               The first is an EVEN line and the second is an ODD line. */
            fprintf(fd, "Band%d-Even, Band%d-Odd", bandNum, bandNum);
        }
    }
    fprintf(fd, "\n");

    /* Print the A0 residual values for each SCA of each band */
    for (sca = 0; sca < nScas; sca++)
    {
        printComma = FALSE;
        for (bandIndex = 0; bandIndex < IAS_MAX_NBANDS; bandIndex++)
        {
            bandNum = residuals->band_residuals[bandIndex].band_number;
            if (bandNum == 0)
                continue;

            if (printComma)
                fprintf(fd, ", ");
            else
                printComma = TRUE;

            /* Always print out index 0 */
            fprintf(fd, "%E",
                residuals->band_residuals[bandIndex].r_squared_a0[0][sca]);

            /* If the lineset is not 2, then print out index 1 as well */
            if (residuals->band_residuals[bandIndex].lineset[0] != 2)
            {
                fprintf(fd, ", %E",
                    residuals->band_residuals[bandIndex].r_squared_a0[1][sca]);
            }

        }
        fprintf(fd, "\n");
    }

    fclose(fd);
    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: write_residuals_a1

PURPOSE:
    Write the A1 residual values for all bands/SCA's that have the specified
    number of detectors to a file.  This function should be called once for each
    unique number of detectors across all of the bands.

RETURNS:
    This function returns a status of SUCCESS if the residual file is
    successfully created, ERROR if an error is encountered.
------------------------------------------------------------------------------*/
int write_residuals_a1
(
    IAS_RPS_DB_BIAS_MODEL_RESIDUALS * residuals, /* I: Bias model residuals */
    const char * bpfPath, /* I: Full path to the BPF file */
    int nDetectors        /* I: Number of detectors being examined */
)
{
    FILE * fd;
    char   path[PATH_MAX];
    int    bandIndex;
    int    bandNum;
    int    detector;
    int    nChars;
    int    nScas;
    int    printComma;
    int    sca;
    int    tmpDetectors;

    /* Create the path for the residual file */
    nChars = snprintf(path, PATH_MAX, "%s-a1-%d.txt", bpfPath, nDetectors);
    if ((nChars < 0) || (nChars >= PATH_MAX))
    {
        IAS_LOG_ERROR("Residual file path exceeds buffer size of %d characters",
            PATH_MAX);
        return ERROR;
    }

    /* Open the residual file for writing */
    fd = fopen(path, "w");
    if (fd == NULL)
    {
        IAS_LOG_ERROR("Opening file for writing: %s", path);
        return ERROR;
    }

    IAS_LOG_INFO("Creating A1 residual file: %s", path);

    /* The residual values are written to the file such that each column
       contains the residual values for a single band and SCA.  Therefore across
       the first line of the file you'll get "band1-SCA01, band1-SCA02, ..." */
    printComma = FALSE;
    for (bandIndex = 0; bandIndex < IAS_MAX_NBANDS; bandIndex++)
    {
        /* Skip this band if the band number is 0 */
        bandNum = residuals->band_residuals[bandIndex].band_number;
        if (bandNum == 0)
            continue;

        /* Skip this band if the detectors in the band doesn't match the number
           that we are working with */
        tmpDetectors = ias_sat_attr_get_detectors_per_sca(bandNum);
        if (tmpDetectors == ERROR)
        {
            IAS_LOG_ERROR("Unable to determine number of detectors for band %d",
                bandNum);
            fclose(fd);
            unlink(path);
            return ERROR;
        }
        if (tmpDetectors != nDetectors)
            continue;

        nScas = ias_sat_attr_get_scas_per_band(bandNum);
        if (nScas == ERROR)
        {
            IAS_LOG_ERROR("Unable to determine number of SCAs for band %d",
                bandNum);
            fclose(fd);
            unlink(path);
            return ERROR;
        }
        for (sca = 0; sca < nScas; sca++)
        {
            /* Do we need to print a leading comma? */
            if (printComma)
                fprintf(fd, ", ");
            else
                printComma = TRUE;

            if (residuals->band_residuals[bandIndex].lineset[0] == 2)
            {
                /* Normal, 30 meter band: only use index 0 of the r_squared_a1
                   array */
                fprintf(fd, "Band%d-SCA%02d", bandNum, sca+1);
            }
            else
            {
                /* 15 meter PAN band: use both indices of the r_squared_a1
                   array.  The first is an EVEN line and the second is an ODD
                   line. */
                fprintf(fd, "Band%d-SCA%02d-Even, Band%d-SCA%02d-Odd", bandNum,
                    sca+1, bandNum, sca+1);
            }
        }
    }
    fprintf(fd, "\n");

    /* Print the A1 residual values in column format */
    for (detector = 0; detector < nDetectors; detector++)
    {
        printComma = FALSE;
        for (bandIndex = 0; bandIndex < IAS_MAX_NBANDS; bandIndex++)
        {
            bandNum = residuals->band_residuals[bandIndex].band_number;
            if (bandNum == 0)
                continue;

            tmpDetectors = ias_sat_attr_get_detectors_per_sca(bandNum);
            if (tmpDetectors == ERROR)
            {
                IAS_LOG_ERROR(
                    "Unable to determine number of detectors for band %d",
                    bandNum);
                fclose(fd);
                unlink(path);
                return ERROR;
            }
            if (tmpDetectors != nDetectors)
                continue;

            nScas = ias_sat_attr_get_scas_per_band(bandNum);
            if (nScas == ERROR)
            {
                IAS_LOG_ERROR("Unable to determine number of SCAs for band %d",
                    bandNum);
                fclose(fd);
                unlink(path);
                return ERROR;
            }
            for (sca = 0; sca < nScas; sca++)
            {
                if (printComma)
                    fprintf(fd, ", ");
                else
                    printComma = TRUE;

                /* Always print out index 0 */
                fprintf(fd, "%E", residuals->band_residuals[bandIndex].
                    r_squared_a1[0][sca][detector]);

                /* If the lineset is not 2, then print out index 1 as well */
                if (residuals->band_residuals[bandIndex].lineset[0] != 2)
                {
                    fprintf(fd, ", %E", residuals->band_residuals[bandIndex].
                        r_squared_a1[1][sca][detector]);
                }
            }
        }
        fprintf(fd, "\n");
    }

    fclose(fd);
    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: write_residuals

PURPOSE:
    Write the Bias Model Residual values to a file for the user to review

RETURNS:
    This function return a status of SUCCESS if the residuals are successfully
    written to a file, ERROR if an error is encountered.
------------------------------------------------------------------------------*/
int write_residuals
(
    IAS_RPS_DB_BIAS_MODEL_RESIDUALS * residuals, /* I: Bias model residuals */
    const char * bpfPath /* I: Full path to the BPF file */
)
{
    int bandIndex;
    int bandNum;
    int nDetectors   = 0;
    int tmpDetectors;

    if (residuals == NULL)
    {
        IAS_LOG_ERROR("Residuals structure not allocated");
        return ERROR;
    }

    if (write_residuals_a0(residuals, bpfPath) != SUCCESS)
    {
        IAS_LOG_ERROR("Creating file for A0 residuals");
        return ERROR;
    }

    /* The A1 residual values are per detector.  To help simplify the analysis
       process, these values are stored in separate files based on the number of
       detectors per SCA.  All bands that have the same number of detectors are
       stored in a single file.  The write_residuals_a1() method should be
       called once for each unique number of detectors per SCA. */
    for (bandIndex = 0; bandIndex < IAS_MAX_NBANDS; bandIndex++)
    {
        bandNum = residuals->band_residuals[bandIndex].band_number;
        if (bandNum == 0)
            continue;

        /* If the number of detectors for the current band is less than or equal
           to the number of detectors that have already been printed, skip it */
        tmpDetectors = ias_sat_attr_get_detectors_per_sca(bandNum);
        if (tmpDetectors == ERROR)
        {
            IAS_LOG_ERROR("Unable to determine number of detectors for band %d",
                bandNum);
            return ERROR;
        }
        if (tmpDetectors <= nDetectors)
            continue;

        /* Remember the current number of detectors being printed and print the
           residual values for all bands that have the same number of detectors
           per SCA. */
        nDetectors = tmpDetectors;
        if (write_residuals_a1(residuals, bpfPath, nDetectors) != SUCCESS)
        {
            IAS_LOG_ERROR("Creating file for A1 residuals");
            return ERROR;
        }
    }

    return SUCCESS;
}


int main (int argc, char * argv[])
{
    const int kFillValue = 0;

    IAS_BPF *                         bpf       = NULL;
    IAS_CPF *                         cpf       = NULL;
    IAS_DATETIME                      iasDate;
    IAS_RPS_DB_BIAS_MODEL_RESIDUALS * residuals = NULL;
    struct ias_db_connection *        db        = NULL;

    const char *dateToUse;
    char bpfPath[PATH_MAX];
    char cmdLine[IAS_BPF_DESCRIPTION_LENGTH];
    char attrCPFName[PATH_MAX];
    char diskCPFName[PATH_MAX] = "refCPF";
    char dbCPFName[PATH_MAX];
    char cpfDateEnd[DATETIME_SIZE];
    char dateEnd[DATETIME_SIZE];
    char cpfDateStart[DATETIME_SIZE];
    char dateStart[DATETIME_SIZE];
    char intervalId[IAS_L0R_INTERVAL_ID_LENGTH+1];
    char strRwocid[256];
    double integrationTimeMS;
    double integrationTimePan;
    int  foundDateEnd                             = FALSE;
    int  foundDateStart                           = FALSE;
    int  foundPrev                                = FALSE;
    int  i;
    int  n0                                       = DEFAULT_N0_VALUE;
    int  nPrev                                    = DEFAULT_N1_VALUE;
    int  nPrevUsed                                = 0;
    int  radiometricWorkOrderCommonId             = 0;
    int  retainCpf                                = FALSE;
    int  status                                   = 0;
    int  tmp;
    double timeStampEnd                           = 0.0;
    double timeStampStart                         = 0.0;

    /* Initialize IAS logging facilities as they are used by the libraries */
    ias_log_initialize("bias_model_calibration");

    /* The default path is the current working directory */
    strcpy(bpfPath, ".");

    for (i = 1; i < argc; i++)
    {
        if (strcmp(argv[i], "+cpf") == 0)
            retainCpf = TRUE;
        else if (strcmp(argv[i], "-cpf") == 0)
            retainCpf = FALSE;
        else if ((strcmp(argv[i], "-d") == 0) && ((i+1) < argc))
        {
            struct stat statData;

            if (stat(argv[i+1], &statData) || !S_ISDIR(statData.st_mode))
            {
                IAS_LOG_ERROR("permission denied: %s\n", argv[i+1]);
                exit(EXIT_FAILURE);
            }
            else if (strlen(argv[i+1]) >= sizeof(bpfPath))
            {
                IAS_LOG_ERROR("Specified path exceed %d characters\n",
                    PATH_MAX);
                exit(EXIT_FAILURE);
            }
            strcpy(bpfPath, argv[++i]);

            /* Remove trailing slashes from user specified paths */
            while (bpfPath[strlen(bpfPath)-1] == '/')
                bpfPath[strlen(bpfPath)-1] = '\0';
        }
        else if (strcmp(argv[i], "-end") == 0)
        {
            int nArgs;
            status = get_date(argc, argv, i+1, &nArgs, &iasDate, &timeStampEnd);
            if (status != SUCCESS)
                usage();

            /* Increment i by the number of arguments that get_date() parsed */
            i += nArgs;
            foundDateEnd = TRUE;

            /* Format the date into a string.  Do not use the IAS standard
               'T' separator between the date and time.  These will be passed
               to the database which uses a space separator. */
            status = snprintf(dateEnd, sizeof(dateEnd),
                "%04d-%02d-%02d %02d:%02d:%02.0lf",
                iasDate.year, iasDate.month, iasDate.day_of_month,
                iasDate.hour, iasDate.minute, iasDate.second);
            if ((status < 0) || (status >= sizeof(dateEnd)))
            {
                IAS_LOG_ERROR("Constructing database date string "
                    "representing the end date");
                exit(EXIT_FAILURE);
            }

            /* Format the end date into a CPF-compatible string.  This date
               will be preferentially used to retrieve a reference CPF
               containing nominal integration times in the OLI_PARAMETERS
               group (and the "correct" file name in the
               FILE_ATTRIBUTES group). */
            status = snprintf(cpfDateEnd, sizeof(cpfDateEnd),
                "%04d-%02d-%02dT%02d:%02d:%02d",
                iasDate.year, iasDate.month, iasDate.day_of_month,
                iasDate.hour, iasDate.minute, (int)iasDate.second);
            if ((status < 0) || (status >= sizeof(cpfDateEnd)))
            {
                IAS_LOG_ERROR("Constructing CPF date string "
                    "representing the end date");
                exit(EXIT_FAILURE);
            }
        }
        else if ((strcmp(argv[i], "-n0") == 0) && ((i+1) < argc) &&
            (sscanf(argv[i+1], "%d", &tmp) == 1) && (tmp > 0))
        {
            /* We aren't tracking the status of n0 because it is an option for
               all 3 of the parameter scenarios.  Simply ensure the value is
               greater than 0. */
            n0 = tmp;
            i++;
        }
        else if ((strcmp(argv[i], "-prev") == 0) && ((i+1) < argc) &&
            (sscanf(argv[i+1], "%d", &tmp) == 1) && (tmp > 0))
        {
            /* The number of previous shutter collects is the same as the n1
               value identified in the bias model calibration ADD */
            nPrev = tmp;
            i++;
            foundPrev = TRUE;
        }
        else if (strcmp(argv[i], "-start") == 0)
        {
            int nArgs;
            status =
                get_date(argc, argv, i+1, &nArgs, &iasDate,&timeStampStart);
            if (status != SUCCESS)
                usage();

            /* Increment i by the number of arguments that get_date() parsed */
            i += nArgs;
            foundDateStart = TRUE;

            /* Format the date into a string.  Do not use the IAS standard 'T'
               separator between the date and time.  These will be passed to
               the database which uses a space separator. */
            status = snprintf(dateStart, sizeof(dateStart),
                "%04d-%02d-%02d %02d:%02d:%02.0lf",
                iasDate.year, iasDate.month, iasDate.day_of_month,
                iasDate.hour, iasDate.minute, iasDate.second);
            if ((status < 0) || (status >= sizeof(dateStart)))
            {
                IAS_LOG_ERROR("Constructing database date string "
                "representing the starting date");
                exit(EXIT_FAILURE);
            }

            /* Format the start date into a CPF-compatiable string.  If -end
               is not defined, this date will be used to retrieve a CPF
               containing integration time and file name information.  */
            snprintf(cpfDateStart, sizeof(cpfDateStart),
                "%04d-%02d-%02dT%02d:%02d:%02d",
                iasDate.year, iasDate.month, iasDate.day_of_month,
                iasDate.hour, iasDate.minute, (int)iasDate.second);
            if ((status < 0) || (status >= sizeof(cpfDateStart)))
            {
                IAS_LOG_ERROR("Constructing CPF date string representing "
                    "the start date");
                exit(EXIT_FAILURE);
            }
        }
        else
            usage();
    }

    /* Create a copy of the command line arguments to be placed in the BPF's
       description to help identify where the BPF came from */
    strcpy(cmdLine, "bias_model_calibration");
    for (i=1; i<argc; i++)
    {
        if ((strlen(cmdLine) + strlen(argv[i]) + 1) >= sizeof(cmdLine))
            continue;
        strcat(cmdLine, " ");
        strcat(cmdLine, argv[i]);
    }

    /* Initialize the satellite attributes library.  Since this isn't coming
       from an ODL parameter file we don't have a way to determine which
       satellite we're working with.  In the future if L9 is launched, this
       application will have to be updated to include an -l8 and -l9 command
       line argument. */
    ias_sat_attr_initialize(IAS_L8);

    /* Retrieve the reference CPF.  */
    if (foundDateEnd)
    {
        dateToUse = cpfDateEnd;
    }
    else
    {
        dateToUse = cpfDateStart;
    }
    status = get_reference_cpf(diskCPFName, dateToUse);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving 'reference' CPF");
        exit(EXIT_FAILURE);
    }

    /* Get the integration time and file name parameters from this CPF */
    status = get_reference_cpf_parameters(diskCPFName, &integrationTimeMS,
        &integrationTimePan, attrCPFName, sizeof(attrCPFName));
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving parameters from reference CPF '%s'",
            diskCPFName);
        unlink(diskCPFName);
        exit(EXIT_FAILURE);
    }

    /* Open the database */
    db = db_open();
    if (db == NULL)
    {
        IAS_LOG_ERROR("Opening database connection");
        unlink(diskCPFName);
        exit(EXIT_FAILURE);
    }

    /* Ensure the user executed this application with one of the 3 valid
       sets of arguments */
    if (foundDateStart && foundPrev && !foundDateEnd)
    {
        status = db_query_start(db, dateStart, integrationTimeMS,
            integrationTimePan, &nPrev, &radiometricWorkOrderCommonId,
            dbCPFName, intervalId);
    }
    else if (foundDateEnd && foundPrev && !foundDateStart)
    {
        status = db_query_end(db, dateEnd, integrationTimeMS,
            integrationTimePan, &nPrev, &radiometricWorkOrderCommonId,
            dbCPFName, intervalId);
    }
    else if (foundDateStart && foundDateEnd && !foundPrev)
    {
        if (timeStampStart >= timeStampEnd)
        {
            IAS_LOG_ERROR("Start date/time must be prior to end date/time");
            ias_db_close_connection(db);
            ias_db_close_database_lib();
            unlink(diskCPFName);
            exit(EXIT_FAILURE);
        }
        status = db_query_range(db, dateStart, dateEnd, integrationTimeMS,
            integrationTimePan, &nPrev, &radiometricWorkOrderCommonId,
            dbCPFName, intervalId);
    }
    else
    {
        ias_db_close_connection(db);
        ias_db_close_database_lib();
        unlink(diskCPFName);
        usage();
    }

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving shutter collects from database");
        ias_db_close_connection(db);
        ias_db_close_database_lib();
        unlink(diskCPFName);
        exit(EXIT_FAILURE);
    }

    /* If the CPF name fetched from the database matches the file attributes
       name of the CPF we retrieved for the integration time information,
       rename the reference CPF to the correct name and go on.  If they don't
       match, we'll need to retrieve 'cpfName' from the service and use
       that.  */
    if (strcmp(dbCPFName, attrCPFName) == 0)
    {
        status = rename(diskCPFName, dbCPFName);
        if (status != 0)
        {
            IAS_LOG_ERROR("Renaming %s to %s", diskCPFName, dbCPFName);
            ias_db_close_connection(db);
            ias_db_close_database_lib();
            unlink(diskCPFName);
            exit(EXIT_FAILURE);
        }
    }
    else
    {
        /* Fetch the CPF used for the last shutter collect in the data
           range */
        IAS_LOG_WARNING("Reference CPF '%s' does not match CPF used for last "
            "shutter collect '%s'; retrieving '%s' to use",
            attrCPFName, dbCPFName, dbCPFName);

        if (fetch_cpf(dbCPFName) != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to retrieve CPF %s", dbCPFName);
            ias_db_close_connection(db);
            ias_db_close_database_lib();
            exit(EXIT_FAILURE);
        }
    }

    /* Read the CPF into the current working directory. */
    cpf = ias_cpf_read(dbCPFName);
    if (cpf == NULL)
    {
        IAS_LOG_ERROR("Reading CPF %s", dbCPFName);
        ias_db_close_connection(db);
        ias_db_close_database_lib();
        exit(EXIT_FAILURE);
    }

    /* Allocate a new BPF structure */
    bpf = ias_bpf_allocate();
    if (bpf == NULL)
    {
        IAS_LOG_ERROR("Initializing BPF structure");
        ias_db_close_connection(db);
        ias_db_close_database_lib();
        unlink(dbCPFName);
        exit(EXIT_FAILURE);
    }

    /* The ias_rps_bias_model_calibration() routine can update the database with
       residual values but this should really only be done during normal DPAS
       processing flows through INGEST.  When run interactively by a user we'll
       allocate a structure and have the routine pass the residuals back to us
       instead of storing them in the database. */
    residuals = ias_rps_bias_residuals_allocate();
    if (residuals == NULL)
    {
        IAS_LOG_ERROR("Allocating memory to store bias model residuals");
        ias_db_close_connection(db);
        ias_db_close_database_lib();
        ias_bpf_free(bpf);
        unlink(dbCPFName);
        exit(EXIT_FAILURE);
    }

    /* ias_rps_bias_model_calibration() expects the RADIOMETRIC_WO_COMMON_ID to
       be a string and not an integer like it is.  Convert the integer to a
       string. */
    IAS_LOG_INFO("Computing bias model");
    sprintf(strRwocid, "%d", radiometricWorkOrderCommonId);
    status = ias_rps_bias_model_calibration(cpf, db, intervalId, strRwocid, n0,
        nPrev, kFillValue, bpf, &nPrevUsed, residuals);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Computing bias model calibration");
        ias_db_close_connection(db);
        ias_db_close_database_lib();
        ias_bpf_free(bpf);
        ias_rps_bias_residuals_free(residuals);
        unlink(dbCPFName);
        exit(EXIT_FAILURE);
    }
    IAS_LOG_INFO("Successfully computed bias model");

    /* Close the database */
    ias_db_close_connection(db);
    ias_db_close_database_lib();

    /* Free the CPF structure */
    ias_cpf_free(cpf);
    cpf = NULL;

    /* Update the file attributes portion of the BPF */
    if (set_bpf_file_attributes(cmdLine, bpf, bpfPath) != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to set file attributes of BPF");
        ias_bpf_free(bpf);
        ias_rps_bias_residuals_free(residuals);
        exit(EXIT_FAILURE);
    }

    /* Write the BPF */
    IAS_LOG_INFO("Creating BPF %s", bpfPath);
    if (ias_bpf_write(bpf, bpfPath) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing BPF to %s", bpfPath);
        ias_bpf_free(bpf);
        ias_rps_bias_residuals_free(residuals);
        exit(EXIT_FAILURE);
    }

    /* Free the BPF structure */
    ias_bpf_free(bpf);

    /* Store the residual information for the user */
    write_residuals(residuals, bpfPath);

    /* Free the residuals memory that was allocated above */
    ias_rps_bias_residuals_free(residuals);

    /* If the CPF isn't supposed to be retained, removed it */
    if (!retainCpf)
        unlink(dbCPFName);
    return EXIT_SUCCESS;
}
