#!/bin/csh -f
#
# Populate the database with shutter collects that can be used for testing
# purposes.  The data files used to load the information can be found in the
# IAS Unit Test directory pointed by $rootDir below.

set cleanFiles = ""
set dataSets   = ""
set maxDataSet = 5
set rootDir    = ${IAS_UNIT_TEST_DATA_DIR}/ias/rps/oli_bias_model_calibration

# Set variables that will be used to keep track of the IDs we generate: LMD
# Mission IDs, Radiometric Work Order IDs, and Work Order IDs
set idsLM  = ""
set idsRWO = ""
set idsWO  = ""

# Process the arguments
if ($#argv == 0) then
    goto USAGE
else if (($#argv == 1) && ($argv[1] == "-all")) then
    set i = 0
    while ($i < ${maxDataSet})
        @ i++
        set dataSets = (${dataSets} ${i})
    end
else
    # All of the arguments must be a numeric value between 1 and $maxDataSet,
    # inclusive.  If the user specifies a non-numeric value we'll get a 0 value
    # for the variable "num".
    while ($#argv > 0)
        set num = `echo $argv[1] | awk '{printf "%d",$1}'`
        if ((${num} < 1) || (${num} > ${maxDataSet})) then
            goto USAGE
        endif
        set dataSets = (${dataSets} ${num})
        shift
    end
endif
echo "Loading data sets:"
echo "  ${dataSets}"
echo ""

# Update the list of files that should be cleaned when this script terminates
set cleanFiles = (${cleanFiles} tmpFrameAverage.ldr tmpFrameAverage.log)
set cleanFiles = (${cleanFiles} tmpHistoStats.ldr   tmpHistoStats.log)
set cleanFiles = (${cleanFiles} tmpLmdMission.ldr   tmpLmdMission.log)
set cleanFiles = (${cleanFiles} tmpRadiometric.ldr  tmpRadiometric.log)
rm -f ${cleanFiles}

# Copy the SQL loader templates into a temporary local loader file
cp ${rootDir}/loader_templates/frame_averages.tpl tmpFrameAverage.ldr
cp ${rootDir}/loader_templates/histo_stats.tpl    tmpHistoStats.ldr
cp ${rootDir}/loader_templates/lmd_mission.tpl    tmpLmdMission.ldr
cp ${rootDir}/loader_templates/rad_common.tpl     tmpRadiometric.ldr

#-------------------------------------------------------------------------------
# To ensure that we get the same results when this test is run we need to
# ensure that no other data that may be in the user's database, overlaps with
# this test data.  It doesn't seem particularly likely that OLI would be
# actively imaging in 2050, so that seems as though that would be far
# enough out to set things up.

# The loader files use YYYY to indicate which year value should be used.
#-------------------------------------------------------------------------------
set tableLmd = "LMD_MISSION"
set tableRwoc = "RADIOMETRIC_WORK_ORDER_COMMON"
set YYYY = 2050

# Notify the user of the year value that we are using.  In addition, we'll store
# this value in a file so that the unit tests can access the value and perform
# an execution of oli_bias_model_calibration with an appropriate date.
echo "Populating test data for year ${YYYY}"
echo ""
echo "${YYYY}" >> cleanup.year

#-------------------------------------------------------------------------------
# Walk through each of the data sets that we are loading and create a loader
# file that will be sent to sqlldr
#-------------------------------------------------------------------------------
set i = 1
while (${i} <= $#dataSets)
    set dataSet = `echo $dataSets[$i] | awk '{printf "%02d",$1}'`
    echo "Preparing data set ${dataSet}"
    @ i++

    # Get the next radiometric work order common ID
    set sql = "select dbseq_radiometric_wo_common_pk.nextval from dual;"
    set rwoId = `echo ${sql} | sqlplus ${IAS_DB_L8_OLITIRS} | \
        grep -A 2 NEXTVAL | tail --lines=1`
    if (${status}) then
        echo "Error retrieving radiometric work order common ID"
        rm -f ${cleanFiles}
        exit 1
    endif
    set idsRWO = (${idsRWO} ${rwoId})

    # Get the next work order ID
    set sql = "select wo_seq.nextval from dual;"
    set woId = `echo ${sql} | sqlplus ${IAS_DB_COM} | \
        grep -A 2 NEXTVAL | tail --lines=1`
    if (${status}) then
        echo "Error retrieving work order ID"
        rm -f ${cleanFiles}
        exit 1
    endif
    set woId = "${woId}IS"
    set idsWO = (${idsWO} ${woId})

    # Get the next LMD mission ID
    set sql = "select dbseq_lmd_mission_pk.nextval from dual;"
    set lmdId = `echo ${sql} | sqlplus ${IAS_DB_L8_OLITIRS} | \
        grep -A 2 NEXTVAL | tail --lines=1`
    if ($status) then
        echo "Error getting LMD Mission ID"
        rm -f ${cleanFiles}
        exit 1
    endif
    set idsLM = (${idsLM} ${lmdId})

    # Copy the template files for the current data set, translate the necessary
    # IDs, and append that data to the end of the loader files
    set file = ${rootDir}/common_data/com_rwoid${dataSet}.tpl
    sed "s/__RWOID${dataSet}__/${rwoId}/" ${file} >> tmpRadiometric.ldr
    sed -i "s/__WOID${dataSet}__/${woId}/" tmpRadiometric.ldr

    set file = ${rootDir}/frame_avg_data/fa_rwoid${dataSet}.tpl
    sed "s/__RWOID${dataSet}__/${rwoId}/" ${file} >> tmpFrameAverage.ldr

    set file = ${rootDir}/histogram_stats_data/hs_rwoid${dataSet}.tpl
    sed "s/__RWOID${dataSet}__/${rwoId}/" ${file} >> tmpHistoStats.ldr

    set file = ${rootDir}/lmd_mission/lmd_${dataSet}.tpl
    sed "s/__LMDID${dataSet}__/${lmdId}/" ${file} >> tmpLmdMission.ldr
    sed -i "s/__MISSION_DATA_LOCATION__/RWOID=${rwoId}/" tmpLmdMission.ldr
end
echo ""

#-------------------------------------------------------------------------------
# Replace the YYYY designation in the radiometric work order common and LMD
# mission loader files with the year that we determined above
#-------------------------------------------------------------------------------
sed -i "s/YYYY/${YYYY}/g" tmpRadiometric.ldr
sed -i "s/YYYY/${YYYY}/g" tmpLmdMission.ldr

#-------------------------------------------------------------------------------
# Save a copy of the radiometric work order ID's so that they can be used by the
# cleanup.csh script to remove the data that is to be loaded.  Since these IDs
# are obtained from sequences and they can't be used again, we'll store these
# IDs even before the loader scripts are run to ensure that partial data can be
# removed if an error is encountered during loading.
#-------------------------------------------------------------------------------
echo ${idsRWO} >> cleanup.ids

#-------------------------------------------------------------------------------
# Load all of the data we just generated into the database for testing
#-------------------------------------------------------------------------------
echo "Loading RADIOMETRIC_WORK_ORDER_COMMON table"
sqlldr ${IAS_DB_L8_OLITIRS} tmpRadiometric.ldr >& /dev/null
if ($status) then
    echo "ERROR loading table"
    exit 1
endif

echo "Loading BIAS_FRAME_AVERAGES table"
sqlldr ${IAS_DB_L8_OLITIRS} tmpFrameAverage.ldr >& /dev/null
if ($status) then
    echo "ERROR loading table"
    exit 1
endif

echo "Loading HISTOGRAM_STATISTICS table"
sqlldr ${IAS_DB_L8_OLITIRS} tmpHistoStats.ldr >& /dev/null
if ($status) then
    echo "ERROR loading table"
    exit 1
endif

echo "Loading LMD_MISSION table"
sqlldr ${IAS_DB_L8_OLITIRS} tmpLmdMission.ldr >& /dev/null
if ($status) then
    echo "ERROR loading table"
    exit 1
endif

echo ""
echo "-------------------------------------------"

echo ""
echo "Data successfully loaded."
echo ""
echo "  RWOID = ${idsRWO}"
echo "  WOID  = ${idsWO}"
echo "  LMDID = ${idsLM}"
echo ""

rm -f ${cleanFiles}
exit 0


USAGE:
    echo ""
    echo "populate.csh [-all] <n1> <n2> ..."
    echo ""
    echo "    This script will populate the database with a set of shutter"
    echo "    collects and histogram statistics for testing the bias model"
    echo "    calibration application.  The user may specify any number of data"
    echo "    set values from 1 to ${maxDataSet} on the command line or the"
    echo "    -all command option to populate all test data sets."
    echo ""
    exit 1
