/*==============================================================================
NAME:
    test_bias

PURPOSE:
    Test the oli_bias_model_calibration application by loading test data into
    the database and then generating a Bias Parameter File (BPF) for that data.
    This will ensure the applicaiton is able to recreate a known BPF.
==============================================================================*/

/* System Includes */
#include <limits.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>

/* IAS Includes */
#include "cgreen/cgreen.h"

/* Local Includes */


/*------------------------------------------------------------------------------
NAME: clean_up

PURPOSE:
    Remove all of the test data that has been loaded into the database
------------------------------------------------------------------------------*/
void clean_up ()
{
    int status;

    status = system("./cleanup.csh -all");
    assert_equal_with_message(status, 0,
        "Error removing test data from database");
}


/*------------------------------------------------------------------------------
NAME: compare_results

PURPOSE:
    Compare the BPF and residual files generated via this unit test against the
    benchmark results.
------------------------------------------------------------------------------*/
void compare_results (
    int year /* I: Year used to load test data into DB */
    )
{
    const char bpfFormat[] = "LO8BPF%04d1003124147_%04d1118132314.01";
    const char egrep[]     = "egrep -v 'Effective_Date|File_Name|Generated '";

    char   benchmarkDir[PATH_MAX];  /* Benchmark directory */
    char   benchmarkPath[PATH_MAX]; /* Benchmark file name */
    char   benchmarkRoot[PATH_MAX]; /* Root BPF file name */
    char   cmd[2*PATH_MAX+256];     /* System command to run */
    char   localPath[PATH_MAX];     /* Local file name */
    char   localRoot[PATH_MAX];     /* Root BPF file name */
    char * ptr;
    int    status;

    /* Create the path to the benchmark directory for this unit test */
    ptr = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (ptr == NULL)
    {
        assert_true_with_message(0, "Error IAS_UNIT_TEST_DATA_DIR not defined");
        return;
    }
    status = snprintf(benchmarkDir, sizeof(benchmarkDir),
        "%s/ias/rps/oli_bias_model_calibration/benchmark", ptr);
    if ((status < 0) || (status >= sizeof(benchmarkDir)))
    {
        assert_true_with_message(0, "Error: Directory path too small");
        return;
    }

    /* Create the root BPF file name for both the benchmark files and the local
       files that have been generated by oli_bias_model_calibration */
    sprintf(benchmarkRoot, bpfFormat, 2001, 2001);
    sprintf(localRoot, bpfFormat, year, year);

    /*--------------------------------------------------------------------------
     * Compare the two BPF files
     *------------------------------------------------------------------------*/

    /* Create the paths to the BPFs */
    status = snprintf(benchmarkPath, sizeof(benchmarkPath), "%s/%s",
        benchmarkDir, benchmarkRoot);
    if ((status < 0) || (status >= sizeof(benchmarkPath)))
    {
        assert_true_with_message(0, "Error: Directory path too small");
        return;
    }
    sprintf(localPath, "./%s", localRoot);

    /* There are three lines that we expect will be different between these
       files.  Use egrep to remove these lines prior to performing a diff
       operation on the files. */
    status = snprintf(cmd, sizeof(cmd), "%s %s > tmpBenchmark.bpf", egrep,
        benchmarkPath);
    if ((status < 0) || (status >= sizeof(cmd)))
    {
        assert_true_with_message(0, "Error: Command string too small");
        return;
    }
    if (system(cmd))
    {
        assert_true_with_message(0, "Error processing benchmark BPF");
        return;
    }

    status = snprintf(cmd, sizeof(cmd), "%s %s > tmpLocal.bpf",egrep,localPath);
    if ((status < 0) || (status >= sizeof(cmd)))
    {
        assert_true_with_message(0, "Error: Command string too small");
        return;
    }
    if (system(cmd))
    {
        assert_true_with_message(0, "Error processing local BPF");
        return;
    }

    /* Run diff on the two temporary BPF files just created */
    status = system("diff tmpBenchmark.bpf tmpLocal.bpf");
    assert_equal_with_message(status, 0, "BPF file differs from benchmark");

    /*--------------------------------------------------------------------------
     * Compare the two A0 residual files
     *------------------------------------------------------------------------*/

    /* Create the path to the A0 files */
    status = snprintf(benchmarkPath, sizeof(benchmarkPath), "%s/%s-a0.txt",
        benchmarkDir, benchmarkRoot);
    if ((status < 0) || (status >= sizeof(benchmarkPath)))
    {
        assert_true_with_message(0, "Error: Directory path too small");
        return;
    }
    sprintf(localPath, "./%s-a0.txt", localRoot);

    status = snprintf(cmd, sizeof(cmd), "diff %s %s", benchmarkPath, localPath);
    if ((status < 0) || (status >= sizeof(cmd)))
    {
        assert_true_with_message(0, "Error: Command string too small");
        return;
    }
    status = system(cmd);
    assert_equal_with_message(status, 0,
        "A0 residual file differs from benchmark");

    /*--------------------------------------------------------------------------
     * Compare the two A1 residual files with 494 detectors per SCA
     *------------------------------------------------------------------------*/

    /* Create the path to the A1 residual files */
    status = snprintf(benchmarkPath, sizeof(benchmarkPath), "%s/%s-a1-494.txt",
        benchmarkDir, benchmarkRoot);
    if ((status < 0) || (status >= sizeof(benchmarkPath)))
    {
        assert_true_with_message(0, "Error: Directory path too small");
        return;
    }
    sprintf(localPath, "./%s-a1-494.txt", localRoot);

    status = snprintf(cmd, sizeof(cmd), "diff %s %s", benchmarkPath, localPath);
    if ((status < 0) || (status >= sizeof(cmd)))
    {
        assert_true_with_message(0, "Error: Command string too small");
        return;
    }
    status = system(cmd);
    assert_equal_with_message(status, 0,
        "A1 residual file with 494 detectors differs from benchmark");

    /*--------------------------------------------------------------------------
     * Compare the two A1 residual files with 988 detectors per SCA
     *------------------------------------------------------------------------*/

    /* Create the path to the A1 residual files */
    status = snprintf(benchmarkPath, sizeof(benchmarkPath), "%s/%s-a1-988.txt",
        benchmarkDir, benchmarkRoot);
    if ((status < 0) || (status >= sizeof(benchmarkPath)))
    {
        assert_true_with_message(0, "Error: Directory path too small");
        return;
    }
    sprintf(localPath, "./%s-a1-988.txt", localRoot);

    status = snprintf(cmd, sizeof(cmd), "diff %s %s", benchmarkPath, localPath);
    if ((status < 0) || (status >= sizeof(cmd)))
    {
        assert_true_with_message(0, "Error: Command string too small");
        return;
    }
    status = system(cmd);
    assert_equal_with_message(status, 0,
        "A1 residual file with 988 detectors differs from benchmark");
    return;
}


/*------------------------------------------------------------------------------
NAME: get_year

PURPOSE:
    Retrieve the year that the test data was populated into the database with.
    To ensure the test data remains distinct from any data that may already be
    in the database, the year is set to 1 larger than any existing dates in the
    LMD_MISSION and RADIOMETRIC_WORK_ORDER_COMMON tables.  The year used will be
    stored as the last line in the ./cleanup.year file.

RETURNS:
    This function will return the year used to populate the test data with or a
    value less than 0 if an error is encountered.
------------------------------------------------------------------------------*/
int get_year ()
{
    FILE * fd;
    int    status;
    int    year   = -1;

    fd = fopen("./cleanup.year", "r");
    if (fd == NULL)
    {
        assert_true_with_message(0, "Error opening ./cleanup.year");
        return year;
    }

    while (fscanf(fd, "%d", &status) == 1)
    {
        if ((status >= 1999) && (status <= 2100))
            year = status;
    }

    fclose(fd);
    if (year < 0)
        assert_true_with_message(0, "Year not found");

    return year;
}


/*------------------------------------------------------------------------------
NAME: populate

PURPOSE:
    Populate the database with a known set of test data.

RETURNS:
    This function returns a status of 0 if the function is successful, non-zero
    if an error is encountered.
------------------------------------------------------------------------------*/
int populate ()
{
    int status;

    status = system("./populate.csh -all");
    assert_equal_with_message(status, 0,
        "Error populating test data into database");
    return status;
}


/*------------------------------------------------------------------------------
NAME: create_bpf

PURPOSE:
    Create a known BPF and compare it to a benchmark BPF
------------------------------------------------------------------------------*/
Ensure create_bpf ()
{
    char   cmd[2048];
    int    isSuccessful = 1;
    int    status;
    int    year;
    time_t start        = time(NULL);
    time_t stop;

    /* Are database tests enabled? */
    if (getenv("ENABLE_DB_TESTS") == NULL)
    {
        printf("DB Tests will not run, setenv ENABLE_DB_TESTS to enable....\n");
        return;
    }

    /* Ensure ../oli_bias_model_calibration is built */
    if (access("../oli_bias_model_calibration", R_OK | X_OK))
    {
        assert_true_with_message(0,
            "Error: ../oli_bias_model_calibration does not exist");
        return;
    }

    printf("\n");
    printf("------------------------------------------------------------\n");
    printf("- Test takes approximately: 3-5 minutes                    -\n");
    printf("- It loads 1,074,090 rows into database                    -\n");
    printf("------------------------------------------------------------\n");

    /* Populate the database with test data */
    if (populate())
    {
        assert_true_with_message(0,
            "Error populating test data into database");
        return;
    }

    /* Obtain the year used to load the test data */
    year = get_year();
    if (year < 0)
    {
        assert_true_with_message(0,
            "Error obtaining year used to load test data");
        isSuccessful = 0;
    }

    /* Run oli_bias_model_calibration */
    if (isSuccessful)
    {
        status = snprintf(cmd, sizeof(cmd),
            "../oli_bias_model_calibration -end %04d-11-19 00:00:00 -prev 4",
            year);
        if ((status < 0) || (status >= sizeof(cmd)))
        {
            assert_true_with_message(0,
                "Error: command buffer is too small, requires %d characters",
                status);
            isSuccessful = 0;
        }
    }

    if (isSuccessful)
    {
        printf("Generating BPF from test data via:\n");
        printf("    %s\n\n", cmd);
        status = system(cmd);
        assert_equal_with_message(status, 0,
            "Error creating BPF from test data");
    }

    /* Cleanup and remove test data from the database */
    clean_up();

    /* Compare the files generated */
    if (isSuccessful)
        compare_results(year);

    stop = time(NULL);
    printf("\n");
    printf("------------------------------------------------------------\n");
    printf("- Test completed in: %3.1f minutes                           -\n",
        (float)((stop - start)/60.0));
    printf("------------------------------------------------------------\n");
    return;
}


int main (int argc, char * argv[])
{
    TestSuite * suite  = create_test_suite();
    int         status;

    add_test(suite, create_bpf);

    if (argc > 1)
        status = run_single_test(suite, argv[1], create_text_reporter());
    else
        status = run_test_suite(suite, create_text_reporter());
    if (status)
        return EXIT_FAILURE;
    return EXIT_SUCCESS;
}
