
/*****************************************************************************
         NAME: convert_from_geographic_to_image

      PURPOSE: Convert the geographic shape vertices to vertices in image line
               and sample
               Also determine the centroid of the shape in both line/sample
               and latitude/longitude

 RETURN VALUE: Type = int
               Value    Description
               -------  --------------------------
               SUCCESS  Successful completion
               ERROR    Operation failed

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "logging_channel.h"
#include "ias_const.h"
#include "ias_types.h"
#include "ias_l1g.h"
#include "ias_shape.h"

#include "pics.h"

/* Width an height values are specified in 30meter pixels for IAS_SHAPE_BOX */
#define PIXEL_RESOLUTION 30.0

int convert_from_geographic_to_image
(
    const IAS_SHAPE *current_shape,     /* I: Current shape to process */
    const IAS_GEO_PROJ_TRANSFORMATION *projtran_geographic_2_l1g,
                                        /* I: Projection transformation */
    const IAS_GEO_PROJ_TRANSFORMATION *projtran_l1g_2_geographic,
                                        /* I: Projection transformation */
    const IAS_L1G_BAND_METADATA *band_metadata, /* I: Band metadata */
    PICS_SHAPE *l1g_shape               /* I/O: Shape points in line/sample */ 
)
{
    int status;      /* Function return values */
    int point_index; /* Loop control */
    double temp_map_x;
    double temp_map_y;
    double centroid_map_line;
    double centroid_map_sample;

     /* Create the l1g shape information */
    switch (l1g_shape->shape_type)
    {
        case IAS_SHAPE_POLYGON:
        {
            for (point_index = 0;
                 point_index < l1g_shape->number_of_vertices;
                 point_index++)
            {
                /* Convert the point from geographic to map */
                status = ias_geo_transform_coordinate(
                    projtran_geographic_2_l1g,
                    current_shape->lon[point_index],
                    current_shape->lat[point_index],
                    &temp_map_x, &temp_map_y);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Transforming to map projection");
                    return ERROR;
                }

                /* Convert the point from map to image */
                l1g_shape->line[point_index] =
                    (band_metadata->upper_left_y - temp_map_y)
                    / band_metadata->projection_distance_y;

                l1g_shape->sample[point_index] =
                    (temp_map_x - band_metadata->upper_left_x)
                    / band_metadata->projection_distance_x;
            }
            break;
        }

        case IAS_SHAPE_BOX:
        {
            /* Variables to allow for different resolutions between the bands */
            double resolution_adjustment_line;
            double resolution_adjustment_sample;
            int height;
            int width;

            /* Convert the point from geometric to map */
            status = ias_geo_transform_coordinate(projtran_geographic_2_l1g,
                current_shape->lon[0], current_shape->lat[0],
                &temp_map_x, &temp_map_y);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Transforming to map projection");
                return ERROR;
            }

            /* Convert the point from map to image */
            l1g_shape->line[0] =
                (band_metadata->upper_left_y - temp_map_y)
                / band_metadata->projection_distance_y;

            l1g_shape->sample[0] =
                (temp_map_x - band_metadata->upper_left_x)
                / band_metadata->projection_distance_x;

            /* Height and width are specified in 30meter pixels so we need to
               convert the height and width values based on the resolution of
               the current band being processed */
            resolution_adjustment_line =
                PIXEL_RESOLUTION / band_metadata->projection_distance_y;
            resolution_adjustment_sample =
                PIXEL_RESOLUTION / band_metadata->projection_distance_x;

            height = (double)current_shape->height * resolution_adjustment_line;
            width = (double)current_shape->width * resolution_adjustment_sample;

            /* Upper Left */
            l1g_shape->line[0] -= 0.5 * (height - 1);
            l1g_shape->sample[0] -= 0.5 * (width - 1);
            /* Upper Right */
            l1g_shape->line[1] = l1g_shape->line[0];
            l1g_shape->sample[1] = l1g_shape->sample[0] + width - 1;
            /* Lower Right */
            l1g_shape->line[2] = l1g_shape->line[0] + height - 1;
            l1g_shape->sample[2] = l1g_shape->sample[1];
            /* Lower left */
            l1g_shape->line[3] = l1g_shape->line[2];
            l1g_shape->sample[3] = l1g_shape->sample[0];
            break;
        }

        default:
        {
            IAS_LOG_ERROR("Unsupported shape type %d", l1g_shape->shape_type);

            return ERROR;
        }
    }

    /*** Determine the centroid of the shape in both lat/long and line/sample
         They will both be used during later processing ***/
    
    /* Calculate the centroid. Centroid coords are stored in the
       shape's centroid_line and centroid_sample */
    status = ias_math_calculate_polygon_centroid(
        l1g_shape->line, l1g_shape->sample, l1g_shape->number_of_vertices,
        &l1g_shape->centroid_line, &l1g_shape->centroid_sample);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating centroid of polygon");
        return ERROR;
    }

    /* Convert the line and sample values back to map */
    centroid_map_line =
        band_metadata->upper_left_y
            - (l1g_shape->centroid_line * band_metadata->projection_distance_y);
    centroid_map_sample =
        (l1g_shape->centroid_sample * band_metadata->projection_distance_x)
            + band_metadata->upper_left_x;

    /* Convert the point from map to geographic */
    status = ias_geo_transform_coordinate(projtran_l1g_2_geographic,
        centroid_map_sample, centroid_map_line,
        &l1g_shape->centroid_longitude, &l1g_shape->centroid_latitude);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Transforming from map to lat/lon");
        return ERROR;
    }

    return SUCCESS;
}
