
/*****************************************************************************
         NAME: create_report.c

      PURPOSE: Provide functionality for creating PICS reports.

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <float.h>
#include <math.h>

#include "logging_channel.h"
#include "pics.h"

/*****************************************************************************
         NAME: create_report_file_with_header

      PURPOSE: Create the report file and populate it with the header
               information

 RETURN VALUE: Type = int
               Value    Description
               -------  --------------------------
               SUCCESS  Successful completion
               ERROR    Operation failed

*****************************************************************************/
int create_report_file_with_header
(
    const PICS_PARAMETERS *parameters, /* I: Application parameters */
    FILE **report_fd                   /* O: File descriptor for the report */
)
{
    int status;                     /* Routine status values */
    int count;                      /* Capture results of snprintf */
    char report_filename[PATH_MAX]; /* Report filename including suffix */

    /* Create the full filename */
    count = snprintf(report_filename, sizeof(report_filename),
        "%s.report", parameters->pics_report_filename_prefix);
    if ((count < 0) || (count >= sizeof(report_filename)))
    {
        IAS_LOG_ERROR("Constructing report filename");

        return ERROR;
    }

    /* Open the report file for writing */
    *report_fd = fopen(report_filename, "w");
    if (*report_fd == NULL)
    {
        IAS_LOG_ERROR("Creating %s", report_filename);

        return ERROR;
    }

    /* Common IAS report header information here */
    status = ias_misc_write_report_header(*report_fd,
        parameters->common_report_header);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing common IAS header information to report file");

        fclose(*report_fd);
        return ERROR;
    }

    return SUCCESS;
}

/*****************************************************************************
         NAME: write_scaling_parameters

      PURPOSE: Write the scaling parameters to the report file

 RETURN VALUE: Type = int
               Value    Description
               -------  --------------------------
               SUCCESS  Successful completion
               ERROR    Operation failed

*****************************************************************************/
int write_scaling_parameters
(
    FILE *report_fd,           /* I: File descriptor for the report */
    const PICS_IMAGE *pics_images, /* I: The pics images to process through */
    int number_of_pics_images  /* I: The number of pics images to process */
)
{
    int status;      /* Used for return values from routines */
    int count;       /* fprintf results */
    int image_index; /* Image looping */
    int band_number; /* Band number of the current band */
    IAS_L1G_BAND_METADATA band_metadata; /* Metadata for the current band */

    count = fprintf(report_fd, "\n\n%4s  %13s  %13s  %13s  %13s\n",
        "", "Reflectance", "Reflectance", "Radiance", "Radiance");
    if (count < 0)
    {
        IAS_LOG_ERROR("Writing metadata to report file");

        return ERROR;
    }

    count = fprintf(report_fd, "%4s  %13s  %13s  %13s  %13s\n",
        "Band", "Multiplier", "Additive", "Multiplier", "Additive");
    if (count < 0)
    {
        IAS_LOG_ERROR("Writing metadata to report file");

        return ERROR;
    }

    count = fprintf(report_fd, "%4s  %13s  %13s  %13s  %13s\n",
        "----",
        "-------------",
        "-------------",
        "-------------",
        "-------------");
    if (count < 0)
    {
        IAS_LOG_ERROR("Writing metadata to report file");

        return ERROR;
    }

    for (image_index = 0; image_index < number_of_pics_images; image_index++)
    {
        /* We only output for the standard bands */
        if (pics_images[image_index].processing_type
            == PICS_STANDARD_PROCESSING)
        {
            band_number = pics_images[image_index].band_number;

            /* Get the band metadata */
            status = ias_l1g_get_band_metadata(
                pics_images[image_index].l1g_image_ptr,
                band_number, &band_metadata);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Reading band %d metadata", band_number);

                return ERROR;
            }

            count = fprintf(report_fd,
                "% 4d  %13.6e  %13.6f  %13.6e  %13.6f\n",
                band_number,
                band_metadata.reflectance_scaling_gain,
                band_metadata.reflectance_scaling_offset,
                band_metadata.spectral_radiance_scaling_gain,
                band_metadata.spectral_radiance_scaling_offset);
            if (count < 0)
            {
                IAS_LOG_ERROR("Writing band %d metadata to report file",
                    band_number);

                return ERROR;
            }
        }
    }

    /* Add a buffer for the columns headers */
    count = fprintf(report_fd, "\n");
    if (count < 0)
    {
        IAS_LOG_ERROR("Writing to report file");

        return ERROR;
    }

    return SUCCESS;
}

/*****************************************************************************
         NAME: write_column_headers

      PURPOSE: Write the column headers to the report file

 RETURN VALUE: Type = int
               Value    Description
               -------  --------------------------
               SUCCESS  Successful completion
               ERROR    Operation failed

*****************************************************************************/
int write_column_headers
(
    FILE *report_fd /* I: File descriptor for the report */
)
{
    int count; /* fprintf results */

    /* Output the record header */
    count = fprintf(report_fd,
        "%4s  %*s"
        "  %13s  %13s"             /* Centroid */
        "  %13s  %13s"             /* View */
        "  %13s  %13s  %13s"       /* Sun */
        "  %13s  %13s  %13s  %13s" /* Mean, StdDev */
        "  %13s"                   /* CoV */
        "  %13s  %13s  %13s  %13s" /* Min, Max */
        "  %7s  %7s  %7s  %7s\n",  /* Total */
        "Band",
        MAX_SHAPE_ID_SIZE, "Region ID",
        "Centroid Lat", "Centroid Lng",
        "View Azimuth",
        "View Zenith",
        "Sun Azimuth",
        "Sun Elevation",
        "Sun Distance",
        "Refl Mean",
        "Refl Std Dev",
        "Rad Mean",
        "Rad Std Dev",
        "Coef of Vari",
        "Refl Minimum",
        "Refl Maximum",
        "Rad Minimum",
        "Rad Maximum",
        "Total",
        "%Valid",
        "\%Fill",
        "%CC");
    if (count < 0)
    {
        IAS_LOG_ERROR("Writing header information to report file");

        return ERROR;
    }

    count = fprintf(report_fd,
        "%4s  %*s"
        "  %13s  %13s"             /* Centroid */
        "  %13s  %13s"             /* View */
        "  %13s  %13s  %13s"       /* Sun */
        "  %13s  %13s  %13s  %13s" /* Mean, StdDev */
        "  %13s"                   /* CoV */
        "  %13s  %13s  %13s  %13s" /* Min, Max */
        "  %7s  %7s  %7s  %7s\n",  /* Total */
        "----",
        MAX_SHAPE_ID_SIZE, "-------------------------------",
        "-------------", /* Centroid */
        "-------------",
        "-------------", /* View */
        "-------------",
        "-------------", /* Sun */
        "-------------",
        "-------------",
        "-------------", /* Mean, StdDev */
        "-------------",
        "-------------",
        "-------------",
        "-------------", /* CoV */
        "-------------", /* Min, Max */
        "-------------",
        "-------------",
        "-------------",
        "-------",       /* Total */
        "-------",
        "-------",
        "-------");
    if (count < 0)
    {
        IAS_LOG_ERROR("Writing header information to report file");

        return ERROR;
    }

    return SUCCESS;
}

/*****************************************************************************
         NAME: append_statistics_to_report

      PURPOSE: Write one line of output to the report file

 RETURN VALUE: Type = int
               Value    Description
               -------  --------------------------
               SUCCESS  Successful completion
               ERROR    Operation failed

*****************************************************************************/
int append_statistics_to_report
(
    FILE *report_fd,                  /* I: File descriptor for the report */
    const PICS_STATISTICS *statistics /* I: The data to write */
)
{
    int count; /* fprintf results */
    double coefficient_of_variation;

    /* Using reflectance of radiance generates the same coefficient value */
    /* Look for zero's and do the correct thing */
    if (fabs(statistics->reflectance_mean) < DBL_EPSILON
        || fabs(statistics->reflectance_stdev) < DBL_EPSILON)
    {
        coefficient_of_variation = 0.0;
    }
    else
    {
        coefficient_of_variation =
            statistics->reflectance_mean / statistics->reflectance_stdev;
    }

    count = fprintf(report_fd,
        "% 4d  %*s"
        "  % 13.8f  % 13.8f"                    /* Centroid */
        "  % 13.8f  % 13.8f"                    /* View */
        "  % 13.8f  % 13.8f  % 13.8f"           /* Sun */
        "  % 13.6f  % 13.6f  % 13.6f  % 13.6f"  /* Mean, StdDev */
        "  % 13.6f"                             /* CoV */
        "  % 13.6f  % 13.6f  % 13.6f  % 13.6f"  /* Min, Max */
        "  % 7d  % 7.2f  % 7.2f  % 7.2f\n",     /* Total */
        statistics->band_number,
        MAX_SHAPE_ID_SIZE, statistics->region_of_interest_id,
        statistics->centroid.lat,             /* Centroid */
        statistics->centroid.lng,
        statistics->view_azimuth_angle,       /* View */
        statistics->view_zenith_angle,
        statistics->sun_azimuth_angle,        /* Sun */
        statistics->sun_elevation_angle,
        statistics->sun_distance,
        statistics->reflectance_mean,         /* Mean, StdDev */
        statistics->reflectance_stdev,
        statistics->radiance_mean,
        statistics->radiance_stdev,
        coefficient_of_variation,             /* CoV */
        statistics->reflectance_minimum,      /* Min, Max */
        statistics->reflectance_maximum,
        statistics->radiance_minimum,
        statistics->radiance_maximum,
        statistics->total_pixel_count,        /* Total */
        statistics->valid_percent,
        statistics->fill_percent,
        statistics->cloud_cover_percent);
    if (count < 0)
    {
        IAS_LOG_ERROR("Writing statistics to report file");

        return ERROR;
    }

    return SUCCESS;
}
