
/*****************************************************************************

         NAME: extract_window.c

      PURPOSE: Extracts a window from the provided band IO pointer and fills
               any buffer data outside the image with the fill value

 RETURN VALUE: Type = int
               Value    Description
               -------  --------------------------
               SUCCESS  Successful completion
               ERROR    Operation failed

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "logging_channel.h"
#include "ias_const.h"
#include "ias_types.h"

#include "pics.h"

int extract_window
(
    const L1G_BAND_IO *l1g_band_ptr, /* I: Pointer to the requested band */
    int number_of_image_lines,    /* I: Number of lines in the band */
    int number_of_image_samples,  /* I: Number of samples in the band */
    int upper_line,               /* I: Upper line to read from */
    int lower_line,               /* I: Lower line to read to */
    int left_sample,              /* I: Left sample to read from */
    int right_sample,             /* I: Right sample to read to */
    int number_of_buffer_lines,   /* I: Number of lines output buffer */
    int number_of_buffer_samples, /* I: Number of samples output buffer */
    IAS_UI2_TYPE *output_buffer   /* O: Populated with image data and fill */
)
{
    int read_upper_line;          /* Upper line to read from */
    int read_lower_line;          /* Lower line to read to */
    int read_left_sample;         /* Left sample to read from */
    int read_right_sample;        /* Right sample to read to */
    int read_number_of_lines;     /* Number of lines in the read buffer */
    int read_number_of_samples;   /* Number of samples in the read buffer */
    int read_line_size_bytes;     /* Size of a read line in bytes */
    int line_index;               /* Loop control */
    int buffer_line_offset;       /* read offset into the output buffer */
    int buffer_sample_offset;     /* read offset into the output buffer */
    int buffer_location;          /* Current line and sample offset location
                                     into the output buffer */
    int buffer_size;              /* Size of the output buffer */
    IAS_UI2_TYPE *read_buffer = NULL; /* Data read from band */

    /* Verify that the data to read fits can fit into the buffer */
    if (number_of_buffer_lines != lower_line - upper_line + 1)
    {
        IAS_LOG_ERROR("Lines to read and output buffer lines do not match");

        return ERROR;
    }
    if (number_of_buffer_samples != right_sample - left_sample + 1)
    {
        IAS_LOG_ERROR("Samples to read and output buffer samples do not match");

        return ERROR;
    }

    /* Adjust the upper line that will be read to fit the image */
    if (upper_line < 0)
    {
        read_upper_line = 0;
    }
    else
    {
        read_upper_line = upper_line;
    }

    /* Adjust the left sample that will be read to fit the image */
    if (left_sample < 0)
    {
        read_left_sample = 0;
    }
    else
    {
        read_left_sample = left_sample;
    }

    /* Adjust the lower line that will be read to fit the image */
    if (lower_line >= number_of_image_lines)
    {
        read_lower_line = number_of_image_lines - 1;
    }
    else
    {
        read_lower_line = lower_line;
    }

    /* Adjust the right sample that will be read to fit the image */
    if (right_sample >= number_of_image_samples)
    {
        read_right_sample = number_of_image_samples - 1;
    }
    else
    {
        read_right_sample = right_sample;
    }

    /* Test to make sure left is left of right */
    if (read_left_sample > read_right_sample)
    {
        IAS_LOG_ERROR("Left sample is right of the right sample");
        IAS_LOG_INFO("left_sample = %d", left_sample);
        IAS_LOG_INFO("right_sample = %d", right_sample);
        IAS_LOG_INFO("read_left_sample = %d", read_left_sample);
        IAS_LOG_INFO("read_right_sample = %d", read_right_sample);

        return ERROR;
    }

    /* Test to make sure upper is upper of lower */
    if (read_upper_line > read_lower_line)
    {
        IAS_LOG_ERROR("Upper line is lower than the lower line");
        IAS_LOG_INFO("upper_line = %d", upper_line);
        IAS_LOG_INFO("lower_line = %d", lower_line);
        IAS_LOG_INFO("read_upper_line = %d", read_upper_line);
        IAS_LOG_INFO("read_lower_line = %d", read_lower_line);

        return ERROR;
    }

    /* Figure out the offsets into the output buffer */
    buffer_line_offset = read_upper_line - upper_line;
    buffer_sample_offset = read_left_sample - left_sample;

    /* Figure out the number of lines and samples to read */
    read_number_of_lines = read_lower_line - read_upper_line + 1;
    read_number_of_samples = read_right_sample - read_left_sample + 1;
    read_line_size_bytes = read_number_of_samples * sizeof(*output_buffer);

    /* Allocate memory for the read_buffer */
    read_buffer = malloc(read_number_of_lines * read_number_of_samples
        * sizeof(*read_buffer));
    if (read_buffer == NULL)
    {
        IAS_LOG_ERROR("Allocating the read buffer");

        return ERROR;
    }

    /* Extract the data from the image (SCA 0) into the read buffer */
    if (ias_l1g_read_image(l1g_band_ptr, 0,
            read_upper_line, read_left_sample,
            read_number_of_lines, read_number_of_samples, read_buffer)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Reading data from band");

        free(read_buffer);
        return ERROR;
    }

    /* Initialize the buffer to the fill value */
    buffer_size = number_of_buffer_lines * number_of_buffer_samples;
    for (buffer_location = 0; buffer_location < buffer_size; buffer_location++)
    {
        output_buffer[buffer_location] = PICS_FILL_VALUE;
    }

    /* Copy the read_buffer into the output_buffer */
    buffer_location = buffer_line_offset * number_of_buffer_samples
        + buffer_sample_offset;
    for (line_index = 0; line_index < read_number_of_lines; line_index++)
    {
        memcpy(&output_buffer[buffer_location],
            &read_buffer[line_index * read_number_of_samples],
            read_line_size_bytes);

        buffer_location += number_of_buffer_samples;
    }

    /* Free the allocated memory */
    free(read_buffer);

    return SUCCESS;
}

