
/*****************************************************************************
         NAME: generate_statistics

      PURPOSE: Generates the statistics for the pixels within the shape

 RETURN VALUE: Type = int
               Value    Description
               -------  --------------------------
               SUCCESS  Successful completion
               ERROR    Operation failed

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>

#include "logging_channel.h"
#include "ias_const.h"
#include "ias_types.h"
#include "ias_angle_gen.h"

#include "pics.h"

int generate_statistics
(
    const IAS_L1G_BAND_METADATA *band_metadata, /* Metadata for the band */
    const IAS_UI2_TYPE *thumbnail_buffer, /* I: Thumbnail buffer */
    int number_of_buffer_lines,   /* I: Number of lines thumbnail buffer */
    int number_of_buffer_samples, /* I: Number of samples thumbnail buffer */
    const PICS_SHAPE *l1g_shape,          /* I: Shape points in line/sample */ 
    PICS_PROCESSING_TYPE pics_processing_type, /* I: Type of processing */
    int number_of_image_lines,    /* I: Number of lines in the band */
    int number_of_image_samples,  /* I: Number of samples in the band */
    int upper_line,               /* I: Upper line to read from */
    int lower_line,               /* I: Lower line to read to */
    int left_sample,              /* I: Left sample to read from */
    int right_sample,             /* I: Right sample to read to */
    int band_index,               /* I: Band index */
    const IAS_IMAGE *dem_image,   /* I: PICS DEM image data */
    const IAS_ANGLE_GEN_METADATA *ang_metadata, /* I: Angle Coeff metadata */
    PICS_STATISTICS *statistics   /* O: Updated statistics information */
)
{
    int status;                  /* Used for return values from routines */
    int line_index;              /* Loop control */
    int sample_index;            /* Loop control */
    int number_of_buffer_values; /* Number of values in the thumbnail buffer */

    int thumbnail_index;         /* Index into the thumbnail data */
    double reflectance_minimum_pixel_value;
    double reflectance_maximum_pixel_value;
    double radiance_minimum_pixel_value;
    double radiance_maximum_pixel_value;
    int number_of_valid_pixels;
    int number_of_fill_pixels;
    int number_of_shape_pixels;
    int number_of_cloud_pixels;

    double reflectance_mean = 0.0;  /* Mean for the valid data in the shape */
    double reflectance_stdev = 0.0; /* Standard deviation for the valid data
                                       in the shape */
    double radiance_mean = 0.0;     /* Mean for the valid data in the shape */
    double radiance_stdev = 0.0;    /* Standard deviation for the valid data
                                       in the shape */

    double *reflectance_buffer = NULL; /* Converted to reflectance */
    double *radiance_buffer = NULL; /* Converted to radiance */

    /* Determine how many values we need */
    number_of_buffer_values = number_of_buffer_lines * number_of_buffer_samples;

    /**** Only performed for standard band processing ****/
    if (pics_processing_type == PICS_STANDARD_PROCESSING)
    {
        /* Allocate memory for the reflectance buffer */
        reflectance_buffer =
            malloc(number_of_buffer_values * sizeof(*reflectance_buffer));
        if (reflectance_buffer == NULL)
        {
            IAS_LOG_ERROR("Allocating the reflectance buffer");

            return ERROR;
        }

        /* Allocate memory for the radiance buffer */
        radiance_buffer =
            malloc(number_of_buffer_values * sizeof(*radiance_buffer));
        if (radiance_buffer == NULL)
        {
            IAS_LOG_ERROR("Allocating the radiance buffer");

            free(reflectance_buffer);
            return ERROR;
        }
    }

    thumbnail_index = -1; /* Start at -1 due to the logic */
    reflectance_minimum_pixel_value = DBL_MAX;
    reflectance_maximum_pixel_value = -DBL_MAX;
    radiance_minimum_pixel_value = DBL_MAX;
    radiance_maximum_pixel_value = -DBL_MAX;
    
    number_of_valid_pixels = 0;
    number_of_fill_pixels = 0;
    number_of_shape_pixels = 0;
    number_of_cloud_pixels = 0;

    /* Process through the bounding box */
    for (line_index = upper_line; line_index <= lower_line; line_index++)
    {
        for (sample_index = left_sample;
             sample_index <= right_sample;
             sample_index++)
        {
            /* We always advance to the next thumbnail value
               This must be done at the begining of the loop */
            thumbnail_index++; /* -1 to 0 during first iteration */

            if (line_index < 0
                || line_index >= number_of_image_lines
                || sample_index < 0
                || sample_index >= number_of_image_samples)
            {
                continue;
            }

            /* If point is not in the shape continue to next point */
            if (l1g_shape->shape_type != IAS_SHAPE_BOX
                && !ias_math_point_in_polygon(l1g_shape->number_of_vertices,
                    l1g_shape->sample, l1g_shape->line,
                    sample_index, line_index))
            {
                continue;
            }

            /**** Only performed for standard band processing ****/
            if (pics_processing_type == PICS_STANDARD_PROCESSING)
            {
                /* Skip the fill values */
                if (thumbnail_buffer[thumbnail_index] != PICS_FILL_VALUE)
                {
                    /* Place converted values into the reflectance buffer */
                    reflectance_buffer[number_of_valid_pixels] =
                        band_metadata->reflectance_scaling_gain
                            * thumbnail_buffer[thumbnail_index]
                        + band_metadata->reflectance_scaling_offset;

                    /* Place converted values into the radiance buffer */
                    radiance_buffer[number_of_valid_pixels] =
                        band_metadata->spectral_radiance_scaling_gain
                            * thumbnail_buffer[thumbnail_index]
                        + band_metadata->spectral_radiance_scaling_offset;

                    /* Figure out the minimum value */
                    if (reflectance_minimum_pixel_value
                        > reflectance_buffer[number_of_valid_pixels])
                    {
                        reflectance_minimum_pixel_value =
                            reflectance_buffer[number_of_valid_pixels];
                    }
                    if (radiance_minimum_pixel_value
                        > radiance_buffer[number_of_valid_pixels])
                    {
                        radiance_minimum_pixel_value =
                            radiance_buffer[number_of_valid_pixels];
                    }
           
                    /* Figure out the maximum value */
                    if (reflectance_maximum_pixel_value
                        < reflectance_buffer[number_of_valid_pixels])
                    {
                        reflectance_maximum_pixel_value =
                            reflectance_buffer[number_of_valid_pixels];
                    }
                    if (radiance_maximum_pixel_value
                        < radiance_buffer[number_of_valid_pixels])
                    {
                        radiance_maximum_pixel_value =
                            radiance_buffer[number_of_valid_pixels];
                    }

                    /* Update the number of valid pixels */
                    number_of_valid_pixels++;
                }
                else
                {
                    /* Update the number of fill pixels */
                    number_of_fill_pixels++;
                }
            }
            /**** Only performed for quality band processing ****/
            else if (pics_processing_type == PICS_QUALITY_PROCESSING)
            {
                /* If the value is cloud increment the cloud count */
                if (thumbnail_buffer[thumbnail_index]
                    & (IAS_QUALITY_BIT_CLOUD_10 | IAS_QUALITY_BIT_CIRRUS_10))
                {
                    number_of_cloud_pixels++;
                }
            }

            /* Increment to the next index (add to the count) */
            number_of_shape_pixels++;
        } /* END - sample_index loop */
    } /* END - line_index loop */

    /**** Only performed for quality band processing ****/
    if (pics_processing_type == PICS_QUALITY_PROCESSING)
    {
        /* Fix things so that we don't get nan's and inf's from using zero */
        if (number_of_cloud_pixels == 0 || number_of_shape_pixels == 0)
        {
            statistics->cloud_cover_percent = 0.0;
        }
        else
        {
            statistics->cloud_cover_percent =
                (double)number_of_cloud_pixels / number_of_shape_pixels * 100.0;
        }
    }
    /**** Only performed for standard band processing ****/
    else if (pics_processing_type == PICS_STANDARD_PROCESSING)
    {
        if (number_of_valid_pixels > 0)
        {
            int outside_image_flag; /* Flag indicating outside image area */
            double radian_angles[2];/* Azimuth and Zenith angles */
            double elevation;       /* Current line/sample elevation */
            double pixel_size;
            double degrees_per_radian = ias_math_get_degrees_per_radian();

            pixel_size = ang_metadata->band_metadata[band_index].pixel_size;

            /* Get elevation for image */
            if (ias_misc_read_elevation_at_line_sample(dem_image, 
                l1g_shape->centroid_line, l1g_shape->centroid_sample, 
                pixel_size, pixel_size,
                &elevation)!= SUCCESS)
            {
                IAS_LOG_ERROR("Error getting elevation");
                free(reflectance_buffer);
                free(radiance_buffer);
                return ERROR;
            }

            /* Always use the scene center earth sun distance */
            statistics->sun_distance = ang_metadata->earth_sun_distance;

            /* Calculate the viewing angles */
            if (ias_angle_gen_calculate_angles_rpc(ang_metadata, 
                l1g_shape->centroid_line, l1g_shape->centroid_sample,
                &elevation, band_index, IAS_ANGLE_GEN_SATELLITE, 
                &outside_image_flag, radian_angles) 
                != SUCCESS)
            {
                IAS_LOG_ERROR("Calculating the viewing angles for band "
                    "number %d", statistics->band_number);
                free(reflectance_buffer);
                free(radiance_buffer);
                return ERROR;
            }
 
            /* Update the viewing angles for this shape */
            if (!outside_image_flag)           
            {
                statistics->view_azimuth_angle = 
                    radian_angles[IAS_ANGLE_GEN_AZIMUTH_INDEX] 
                    * degrees_per_radian;
                statistics->view_zenith_angle = 
                    radian_angles[IAS_ANGLE_GEN_ZENITH_INDEX] 
                    * degrees_per_radian;
            }
            else
            {
                statistics->view_azimuth_angle = PICS_ANGLE_OUTSIDE_IMAGE;
                statistics->view_zenith_angle = PICS_ANGLE_OUTSIDE_IMAGE;
                IAS_LOG_WARNING("Invalidating the satellite viewing angles "
                    "for band %d (%s) since the centroid is ouside the image",
                    statistics->band_number, statistics->region_of_interest_id);
            }

            /* Calculate the solar angles */
            if (ias_angle_gen_calculate_angles_rpc(ang_metadata, 
                l1g_shape->centroid_line, l1g_shape->centroid_sample,
                &elevation, band_index,
                IAS_ANGLE_GEN_SOLAR, &outside_image_flag, radian_angles) 
                != SUCCESS)
            {
                IAS_LOG_ERROR("Calculating the solar angles for band "
                    "number %d", statistics->band_number);
                free(reflectance_buffer);
                free(radiance_buffer);
                return ERROR;
            }
           
            /* Update the solar angles for this shape */
            if (!outside_image_flag)
            {
                statistics->sun_elevation_angle = 0.0;
                statistics->sun_azimuth_angle =
                    radian_angles[IAS_ANGLE_GEN_AZIMUTH_INDEX] 
                    * degrees_per_radian;
            }
            else
            {
                statistics->sun_elevation_angle = PICS_ANGLE_OUTSIDE_IMAGE;
                statistics->sun_azimuth_angle = PICS_ANGLE_OUTSIDE_IMAGE;
                IAS_LOG_WARNING("Invalidating the solar angles for band %d "
                    "(%s) since the centroid is ouside the image",
                    statistics->band_number, statistics->region_of_interest_id);
            }

            /* If line/sample location is outside of the active 
               image then elevation angle stays set to 
               PICS_ANGLE_OUTSIDE_IMAGE  */
            if (!outside_image_flag)
            {
                double half_pi = ias_math_get_pi() / 2.0;
                statistics->sun_elevation_angle = (half_pi 
                    - radian_angles[IAS_ANGLE_GEN_ZENITH_INDEX]) 
                    * degrees_per_radian;
            }
 
            /* Determine the mean */
            status = ias_math_compute_mean(reflectance_buffer,
                number_of_valid_pixels, &reflectance_mean);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Determining reflectance mean for band %d (%s)",
                    statistics->band_number,
                    statistics->region_of_interest_id);

                free(reflectance_buffer);
                free(radiance_buffer);
                return ERROR;
            }
            status = ias_math_compute_mean(radiance_buffer,
                number_of_valid_pixels, &radiance_mean);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Determining radiance mean for band %d (%s)",
                    statistics->band_number,
                    statistics->region_of_interest_id);

                free(reflectance_buffer);
                free(radiance_buffer);
                return ERROR;
            }

            /* Determine the standard deviation */
            status = ias_math_compute_stdev(reflectance_buffer,
            number_of_valid_pixels, reflectance_mean, &reflectance_stdev);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Determining reflectance standard deviation for"
                    " band %d (%s)",
                    statistics->band_number,
                    statistics->region_of_interest_id);

                free(reflectance_buffer);
                free(radiance_buffer);
                return ERROR;
            }
            status = ias_math_compute_stdev(radiance_buffer,
                number_of_valid_pixels, radiance_mean, &radiance_stdev);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Determining radiance standard deviation for"
                    " band %d (%s)",
                    statistics->band_number,
                    statistics->region_of_interest_id);

                free(reflectance_buffer);
                free(radiance_buffer);
                return ERROR;
            }

            /* Assigned them to the statistics */
            statistics->centroid.lat = l1g_shape->centroid_latitude;
            statistics->centroid.lng = l1g_shape->centroid_longitude;
            statistics->reflectance_mean = reflectance_mean;
            statistics->reflectance_stdev = reflectance_stdev;
            statistics->radiance_mean = radiance_mean;
            statistics->radiance_stdev = radiance_stdev;
            statistics->reflectance_minimum = reflectance_minimum_pixel_value;
            statistics->reflectance_maximum = reflectance_maximum_pixel_value;
            statistics->radiance_minimum = radiance_minimum_pixel_value;
            statistics->radiance_maximum = radiance_maximum_pixel_value;
        }
        else
        {
            IAS_LOG_WARNING("No valid pixels found for band %d (%s):"
                " Zeroing out the statistics",
                statistics->band_number, statistics->region_of_interest_id);

            /* Assigned zeros to the statistics */
            statistics->centroid.lat = 0.0;
            statistics->centroid.lng = 0.0;
            statistics->reflectance_mean = 0.0;
            statistics->reflectance_stdev = 0.0;
            statistics->radiance_mean = 0.0;
            statistics->radiance_stdev = 0.0;
            statistics->reflectance_minimum = 0.0;
            statistics->reflectance_maximum = 0.0;
            statistics->radiance_minimum = 0.0;
            statistics->radiance_maximum = 0.0;
            statistics->sun_azimuth_angle = 0.0;
            statistics->sun_elevation_angle = 0.0;
            statistics->view_azimuth_angle = 0.0;
            statistics->view_zenith_angle = 0.0;
            statistics->sun_distance = 0.0;
        }

        /* Gather the pixel count for the shape */
        statistics->total_pixel_count = number_of_shape_pixels;

        if (number_of_shape_pixels == 0)
        {
            IAS_LOG_WARNING("No shape pixels found for band %d (%s)",
                statistics->band_number, statistics->region_of_interest_id);

            statistics->fill_percent = 0.0;
            statistics->valid_percent = 0.0;
        }
        else
        {
            statistics->fill_percent =
                (double)number_of_fill_pixels / number_of_shape_pixels * 100.0;
            statistics->valid_percent =
                (double)number_of_valid_pixels / number_of_shape_pixels * 100.0;
        }
    } /* END - Standard Band Processing */

    /* Free the memory */
    free(reflectance_buffer);
    free(radiance_buffer);

    return SUCCESS;
}

