
/*****************************************************************************
         NAME: ias_read_shapefile

      PURPOSE: Reads a PICS formatted shapefile into memory

 RETURN VALUE: Type = int
               Value    Description
               -------  --------------------------
               SUCCESS  Successful completion
               ERROR    Operation failed

*****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "ias_const.h"
#include "ias_miscellaneous.h"
#include "ias_logging.h"
#include "ias_shape.h"

#define MAX_VERTICES 100 /* Maximum number of vertices allowed */
#define MAX_LINE_LENGTH 256 /* Maximum size of a line to read from the file */

int ias_read_shapefile
(
    const char *filename,  /* I: Shape file name */
    int *number_of_shapes, /* O: Number of shapes */
    IAS_SHAPE **head       /* O: First shape in list */
)
{
    FILE *fp;
    char str[MAX_LINE_LENGTH];
    unsigned int npts;
    double lat[MAX_VERTICES];
    double lon[MAX_VERTICES];
    IAS_SHAPE *curr;
    int count;

    fp = fopen(filename, "r");
    if (fp == NULL)
    {
        IAS_LOG_ERROR("Opening shapefile %s", filename);
        return ERROR;
    }

    if (fgets(str, MAX_LINE_LENGTH, fp) == NULL) /* filename */
    {
        IAS_LOG_ERROR("Reading filename from shapefile %s:", filename);
        return ERROR;
    }

    if (fgets(str, MAX_LINE_LENGTH, fp) == NULL) /* blank line */
    {
        IAS_LOG_ERROR("Reading blank line from shapefile %s:", filename);
        return ERROR;
    }

    /* Read all regions of interest. */
    *head = NULL;
    *number_of_shapes = 0;
    while (1)
    {
        char *newline; /* For finding newline characters */

        /* Read the region ID. */
        if (fgets(str, MAX_LINE_LENGTH, fp) == NULL)
        {
            fclose(fp);

            if (*head != NULL)
                return SUCCESS;

            IAS_LOG_ERROR("No shape ID read");
            return ERROR;
        }

        /* If we have a newline character remove it */
        newline = strrchr(str, '\n');
        if (newline != NULL)
        {
            *newline = '\0';
        }

        if ((curr = calloc(1, sizeof(IAS_SHAPE))) == NULL)
        {
            IAS_LOG_ERROR("Allocating memory");
            fclose(fp);
            ias_shape_free(*head);
            return ERROR;
        }

        /* Set the linked list elements. */
        curr->next = *head;
        *head = curr;

        count = snprintf(curr->region_id, sizeof(curr->region_id), str);
        if ((count < 0) || (count >= sizeof(curr->region_id)))
        {
            IAS_LOG_ERROR("Constructing shape ID (%s)", str);
            fclose(fp);
            ias_shape_free(*head);
            return ERROR;
        }

        /* Read the lat/long coordinates. */
        npts = 0;
        while (npts < MAX_VERTICES)
        {
            if (fscanf(fp, "%lf %lf", &lat[npts], &lon[npts]) != 2)
            {
                break;
            }
            npts++;
        }

        /* Check whether there are more polynomial vertices than allowed. */
        if (npts == MAX_VERTICES)
        {
            IAS_LOG_ERROR("Too many polygon vertices");
            fclose(fp);
            ias_shape_free(*head);
            return ERROR;
        }

        /* Ensure we have enough points. */
        if (npts < 2)
        {
            IAS_LOG_ERROR("Shape must contain at least two points");
            fclose(fp);
            ias_shape_free(*head);
            return ERROR;
        }

        /* Determine the shape type. */
        if (npts > 2)
        {
            curr->npts = npts;
            curr->type = IAS_SHAPE_POLYGON;
            curr->lat = malloc(npts*sizeof(double));
            if (curr->lat == NULL)
            {
                IAS_LOG_ERROR("Allocating memory");
                fclose(fp);
                ias_shape_free(*head);
                return ERROR;
            }
            curr->lon = malloc(npts*sizeof(double));
            if (curr->lon == NULL)
            {
                IAS_LOG_ERROR("Allocating memory");
                fclose(fp);
                ias_shape_free(*head);
                return ERROR;
            }
            memcpy(curr->lat, lat, npts*sizeof(double));
            memcpy(curr->lon, lon, npts*sizeof(double));
        }
        else
        {
            /* The box shape has one set of coordinate values for the center
               of the box/region */
            curr->npts = 1;
            curr->type = IAS_SHAPE_BOX;
            curr->lat = malloc(sizeof(double));
            curr->lon = malloc(sizeof(double));
            if (curr->lat == NULL)
            {
                IAS_LOG_ERROR("Allocating memory");
                fclose(fp);
                ias_shape_free(*head);
                return ERROR;
            }
            if (curr->lon == NULL)
            {
                IAS_LOG_ERROR("Allocating memory");
                fclose(fp);
                ias_shape_free(*head);
                return ERROR;
            }
            curr->lat[0] = lat[0];
            curr->lon[0] = lon[0];

            /* Validate the width and height parameters. */
            if (lat[1] < 1 || lon[1] < 1)
            {
                IAS_LOG_ERROR("Box height and width values must"
                              " be greater than or equal to 1");
                fclose(fp);
                ias_shape_free(*head);
                return ERROR;
            }
            curr->width = (unsigned int)lat[1];
            curr->height = (unsigned int)lon[1];
        }

        (*number_of_shapes)++;
    }

    /* Close the file. */
    fclose(fp);

    return SUCCESS;
}

/*****************************************************************************
         NAME: ias_shape_free

      PURPOSE: Free allocated memory from ias_read_shapefile routine

 RETURN VALUE: None

*****************************************************************************/
void ias_shape_free
(
    IAS_SHAPE *head /* I: First shape in list */
)
{
    IAS_SHAPE *curr = head;

    while (curr)
    {
        free(curr->lat);
        free(curr->lon);
        curr = curr->next;
        free(head);
        head = curr;
    }
}

