
/*****************************************************************************
         NAME: initialize_database.c

      PURPOSE: Create a connection to the database and establish the
               RADIOMETRIC_WO_COMMON_ID

 RETURN VALUE: Type = int
               Value    Description
               -------  --------------------------
               SUCCESS  Successful completion
               ERROR    Operation failed

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "logging_channel.h"
#include "ias_const.h"
#include "ias_types.h"
#include "ias_logging.h"
#include "ias_db_get_connect_info.h"
#include "ias_miscellaneous.h"

#include "pics.h"

int initialize_database
(
    PICS_PARAMETERS *parameters,        /* I/O: Application parameters */
    struct ias_db_connection **database /* O: Database connection */
)
{
    /* Database connection variables */
    char *db_host = NULL;
    char *db_name = NULL;
    char *db_password = NULL;
    char *db_user = NULL;

    if (ias_db_initialize_database_lib() != SUCCESS)
    {
        IAS_LOG_ERROR("Unable initialize database library");

        return ERROR;
    }

    if (ias_db_get_connect_info(&db_name, &db_user, &db_password, &db_host)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Unable retrieve database connection information");

        ias_db_close_database_lib();
        return ERROR;
    }

    *database =
        ias_db_connect_to_database(db_name, db_user, db_password, db_host);

    /* Free them because they are not needed anymore */
    free(db_name);
    db_name = NULL;
    free(db_user);
    db_user = NULL;
    free(db_password);
    db_password = NULL;
    free(db_host);
    db_host = NULL;

    /* Now test for successful connection */
    if (*database == NULL)
    {
        IAS_LOG_ERROR("Unable to establish database connection");

        ias_db_close_database_lib();
        return ERROR;
    }

    /* Obtain the radiometric work order ID for this IAS work order */
    if (ias_misc_get_char_id(*database, "RADIOMETRIC_WORK_ORDER_COMMON",
        "RADIOMETRIC_WO_COMMON_ID", parameters->work_order_id,
        parameters->radiometric_wo_id) != SUCCESS)
    {
        IAS_LOG_ERROR(
            "Unable to obtain RADIOMETRIC_WO_COMMON_ID for work order: %s",
            parameters->work_order_id);

        ias_db_close_connection(*database);
        *database = NULL;
        ias_db_close_database_lib();
        return ERROR;
    }

    /* Delete all prior entries in the PICS_STATISTICS table for this
       radiometric work order common ID */
    if (remove_statistics(*database, parameters->radiometric_wo_id) != SUCCESS)
    {
        IAS_LOG_ERROR(
             "Deleting previous records from PICS_STATISTICS for "
            "RADIOMETRIC_WO_COMMON_ID=%s ", parameters->radiometric_wo_id);

        ias_db_close_connection(*database);
        *database = NULL;
        ias_db_close_database_lib();
        return ERROR;
    }

    return SUCCESS;
}

