
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "logging_channel.h"
#include "pics.h"

/*****************************************************************************

         NAME: open_input_data

      PURPOSE: Opens the input data files and metadata used by the application

 RETURN VALUE: Type = int
               Value    Description
               -------  --------------------------
               SUCCESS  Successful completion
               ERROR    Operation failed

*****************************************************************************/
int open_input_data
(
    PICS_PARAMETERS *parameters, /* I/O: Application parameters */
    PICS_DATA_FILES *input_data  /* O: The opened data */
)
{
    int status;
    int band_index;
    int band_number;
    int pan_band_number = ias_sat_attr_get_band_number_from_type(IAS_PAN_BAND);

    /* Initialize things */
    input_data->l1g_image_ptr = NULL;
    input_data->l1g_quality_ptr = NULL;
    input_data->model_ptr = NULL;

    /* Initialize the dem_image structure to make it easier to clean up */
    memset(&input_data->dem_image, 0, sizeof(input_data->dem_image));

    /* Open the L1G image */
    input_data->l1g_image_ptr =
        ias_l1g_open_image(parameters->l1g_filename, IAS_READ );
    if (input_data->l1g_image_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening L1G image (%s)", parameters->l1g_filename);

        close_input_data(input_data);
        return ERROR;
    }

    /* Load the L1G metadata */
    status = ias_l1g_get_file_metadata(input_data->l1g_image_ptr,
        &input_data->l1g_metadata);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving file metadata for %s",
            parameters->l1g_filename);

        close_input_data(input_data);
        return ERROR;
    }

    /* Open the L1G quality image */
    input_data->l1g_quality_ptr =
        ias_l1g_open_image(parameters->quality_filename, IAS_READ );
    if (input_data->l1g_quality_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening L1G quality image (%s)",
            parameters->quality_filename);

        close_input_data(input_data);
        return ERROR;
    }

    /* Load the L1G metadata */
    status = ias_l1g_get_file_metadata(input_data->l1g_quality_ptr,
        &input_data->l1g_quality_metadata);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving file metadata for %s",
            parameters->quality_filename);

        close_input_data(input_data);
        return ERROR;
    }

    /* Initialize the grid structure */
    if (ias_grid_initialize(&input_data->grid, 0, 0) != SUCCESS)
    {
        IAS_LOG_ERROR("Error initializing the grid structure");

        close_input_data(input_data);
        return ERROR;
    }

    /* Read the grid header */
    if (ias_grid_header_read(parameters->grid_filename, &input_data->grid)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Error reading the grid header from %s",
            parameters->grid_filename);

        close_input_data(input_data);
        return ERROR;
    }

    /* Read the grid band header */
    if (ias_grid_band_header_read(parameters->grid_filename, &input_data->grid)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Error reading the grid band header from %s",
            parameters->grid_filename);

        close_input_data(input_data);
        return ERROR;
    }

    /* Read the model */
    input_data->model_ptr = ias_model_read(parameters->model_filename);
    if (input_data->model_ptr == NULL)
    {
        IAS_LOG_ERROR("Reading model file %s", parameters->model_filename);

        close_input_data(input_data);
        return ERROR;
    }

    /* Loop through the bands present in the grid and use the first
       one found */
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        /* Figure out the band number so we can skip the PAN band since
           we want to use a 30 meter band */
        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        if (band_number == pan_band_number)
        {
            continue;
        }

        if (input_data->grid.bands_available[band_index])
        {
            break;
        }
    }
    
    /* Make sure one was found */
    if (band_index == IAS_MAX_NBANDS)
    {
        IAS_LOG_ERROR("No bands present in GRID file %s",
            parameters->grid_filename);

        close_input_data(input_data);
        return ERROR;
    }

    IAS_LOG_INFO("Using GRID band index %d", band_index);

    /* Figure out the band number for the grid band to read */
    band_number = ias_sat_attr_convert_band_index_to_number(band_index);

    /* Get pointer information for the band being processed */
    if (ias_grid_band_pointers_read(parameters->grid_filename,
            &input_data->grid, band_number)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Error reading the grid pointers from %s",
            parameters->grid_filename);

        close_input_data(input_data);
        return ERROR;
    }

    /* Read the DEM into memory */
    if (ias_misc_read_coregistered_single_band_l1g(parameters->dem_filename,
        &input_data->grid.corners, 
        input_data->grid.gridbands[band_index][0].lines,
        input_data->grid.gridbands[band_index][0].samps, 
        input_data->grid.gridbands[band_index][0].pixsize,
        &input_data->dem_image, IAS_I2) == ERROR)
    {
        IAS_LOG_ERROR("Reading DEM from %s", parameters->dem_filename);

        close_input_data(input_data);
        return ERROR;
    }

    return SUCCESS;
};

/*****************************************************************************

         NAME: close_input_data

      PURPOSE: Closes the input data and metadata used by the application

 RETURN VALUE: Type = int
               Value    Description
               -------  --------------------------
               SUCCESS  Successful completion
               ERROR    Operation failed

*****************************************************************************/
void close_input_data
(
    PICS_DATA_FILES *input_data /* I/O: The data to close */
)
{
    /* Close the L1G image */
    if (input_data->l1g_image_ptr != NULL)
    {
        if (ias_l1g_close_image(input_data->l1g_image_ptr) != SUCCESS)
        {
            IAS_LOG_WARNING("Failed closing L1G image");
        }
        input_data->l1g_image_ptr = NULL;
    }

    /* Close the L1G quality image */
    if (input_data->l1g_quality_ptr != NULL)
    {
        if (ias_l1g_close_image(input_data->l1g_quality_ptr) != SUCCESS)
        {
            IAS_LOG_WARNING("Failed closing L1G quality image");
        }
        input_data->l1g_quality_ptr = NULL;
    }

    ias_grid_free(&input_data->grid);

    if (input_data->model_ptr != NULL)
    {
        ias_los_model_free(input_data->model_ptr);
    }
    input_data->model_ptr = NULL;

    free(input_data->dem_image.data);
    input_data->dem_image.data = NULL;
};

