
/*****************************************************************************
         NAME: pics -- picsiasda (Pseudo-Invariant Calibration Site Image
                                  Assessment System Database Automation)

      PURPOSE: Extract image regions from L1G input, compute statistics on the
               regions, and store the results in the database

 RETURN VALUE: Type = int
               Value         Description
               ------------  --------------------------
               EXIT_SUCCESS  Successful completion
               EXIT_FAILURE  Operation failed

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include <limits.h>
#include <unistd.h>

#include "gctp.h"
#include "logging_channel.h"
#include "ias_const.h"
#include "ias_types.h"
#include "ias_satellite_attributes.h"
#include "ias_db_get_connect_info.h"
#include "ias_miscellaneous.h"
#include "ias_math.h"
#include "ias_shape.h"
#include "ias_angle_gen.h"

#include "pics.h"

/* Proto-type the C99 round routine */
double round(double);

int main(int argc, char *argv[])
{
    int status;                     /* Used for return values from routines */
    int count;                      /* Used for snprintf results */
    int index;                      /* Generic loop control */
    int band_index;                 /* Loop control */
    int processing_status;          /* Stop processing due to error */
    PICS_PARAMETERS parameters;     /* Holds the input parameter information */
    PICS_DATA_FILES input_data;     /* Holds the input data file information */
    IAS_SATELLITE_ID satellite_id;  /* Satellite ID used to initialize the
                                       satellite attributes library */
    IAS_SHAPE *shape_list;          /* List of shapes to process */
    int number_of_shapes;           /* Number of shapes in the shape list */
    int *shape_outside_image_flags; /* Per-shape Flag to indicate the shape is 
                                       outside the image (0 = in image) */

    IAS_UI2_TYPE *thumbnail_buffer = NULL; /* Thumbnail buffer */

    PICS_SHAPE l1g_shape; /* Holds shape points in L1G line/sample */
    IAS_ANGLE_GEN_METADATA ang_metadata; /* Angle Coefficient metadata */

    /* Statistic gathering variables */
    PICS_STATISTICS *statistics;

    /* Variables for image looping */
    int image_index;
    PICS_IMAGE *pics_images;
    int number_of_pics_images;
    int band_present[IAS_MAX_NBANDS];

    /* Projection Variables */
    IAS_PROJECTION geographic_projection;
    IAS_PROJECTION input_image_projection;
    double geographic_projection_parameters[IAS_PROJ_PARAM_SIZE];
    int projection_units;
    IAS_GEO_PROJ_TRANSFORMATION *projtran_geographic_2_l1g = NULL;
    IAS_GEO_PROJ_TRANSFORMATION *projtran_l1g_2_geographic = NULL;

    FILE *report_fd = NULL; /* File descriptor for the report file */

    /* Database connection variables */
    struct ias_db_connection *database = NULL;

    /* Initialize logging for this application */
    ias_log_initialize("pics");

    /* Only one command line argument:  the WO parameter file */
    if (argc != 2)
    {
        IAS_LOG_ERROR("USAGE:  pics <name of parameter file>");

        return EXIT_FAILURE;
    }

    /* Initialize a few of the data structures */
    memset(&parameters, 0, sizeof(PICS_PARAMETERS));
    memset(&input_data, 0, sizeof(PICS_DATA_FILES));
    memset(&l1g_shape, 0, sizeof(PICS_SHAPE));

    /* Read the input ODL and OMF files */
    IAS_LOG_DEBUG("Reading ODL and OMF parameters...");
    status = read_parameters(argv[1], &parameters);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Parameter processing failed");

        return EXIT_FAILURE;
    }

    IAS_LOG_INFO("Started");

    /* Figure out the satellite ID */
    satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(
            parameters.satellite_number);
    if (satellite_id == ERROR)
    {
        IAS_LOG_ERROR("Invalid satellite number %d in OMF parameter file",
            parameters.satellite_number);

        return EXIT_FAILURE;
    }

    /* Initialize the satellite attributes library */
    status = ias_sat_attr_initialize(satellite_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attribute information");

        return EXIT_FAILURE;
    }

    /* Initialize the report header */
    parameters.common_report_header = ias_misc_initialize_report_header(
        parameters.l0r_filename,
        parameters.work_order_id,
        "PICS Characterization");
    if (parameters.common_report_header == NULL)
    {
        IAS_LOG_ERROR("Initializing common report header information");

        return EXIT_FAILURE;
    }

    /* Open the input data files */
    IAS_LOG_DEBUG("Opening the input data files...");
    status = open_input_data(&parameters, &input_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Opening input data files");

        ias_misc_free_report_header(parameters.common_report_header);
        return EXIT_FAILURE;
    }

    /* Verify the bands present in the input image and count how many */
    number_of_pics_images = 0;
    for (band_index = 0;
         band_index < parameters.num_bands_to_process;
         band_index++)
    {
        if (!ias_l1g_is_band_present(input_data.l1g_image_ptr,
            parameters.band_list[band_index]))
        {
            IAS_LOG_WARNING("Band %d not present in L1G image",
                parameters.band_list[band_index]);

            band_present[band_index] = ERROR;
        }
        else
        {
            band_present[band_index] = parameters.band_list[band_index];
            number_of_pics_images++;
        }
    }

    /* Add 1 more for processing the quality band */
    number_of_pics_images += 1;

    /* Determine the shape filename to use if not provided */
    if (parameters.shape_filename[0] == '\0')
    {
        char *ias_data_dir = getenv("IAS_DATA_DIR");
        if (ias_data_dir == NULL)
        {
            IAS_LOG_ERROR("Reading environment variable IAS_DATA_DIR");

            ias_misc_free_report_header(parameters.common_report_header);
            close_input_data(&input_data);
            return EXIT_FAILURE;
        }

        count = snprintf(parameters.shape_filename,
            sizeof(parameters.shape_filename),
            "%s/sysdata/pics/wrs2_%03d_%03d.shp", ias_data_dir,
            input_data.l1g_metadata.wrs_path, input_data.l1g_metadata.wrs_row);
        if (count < 0 || count >= sizeof(parameters.shape_filename))
        {
            IAS_LOG_ERROR("Determining shape filename");

            ias_misc_free_report_header(parameters.common_report_header);
            close_input_data(&input_data);
            return EXIT_FAILURE;
        }

        /* If the 3 digit path row file isn't available/present try the other
           way to specify the filenames */
        if (access(parameters.shape_filename, F_OK) == -1)
        {
            count = snprintf(parameters.shape_filename,
                sizeof(parameters.shape_filename),
                "%s/sysdata/pics/wrs2_%d_%d.shp", ias_data_dir,
                input_data.l1g_metadata.wrs_path,
                input_data.l1g_metadata.wrs_row);
            if (count < 0 || count >= sizeof(parameters.shape_filename))
            {
                IAS_LOG_ERROR("Determining shape filename");

                ias_misc_free_report_header(parameters.common_report_header);
                close_input_data(&input_data);
                return EXIT_FAILURE;
            }
        }
    }

    /* Verify that the file is present and can be read */
    if (access(parameters.shape_filename, F_OK | R_OK) == -1)
    {
        IAS_LOG_ERROR("Unable to access shape file %s",
            parameters.shape_filename);

        ias_misc_free_report_header(parameters.common_report_header);
        close_input_data(&input_data);
        return EXIT_FAILURE;
    }

    /* Read the shape file */
    IAS_LOG_INFO("Using shape file %s", parameters.shape_filename);
    status = ias_read_shapefile(parameters.shape_filename,
        &number_of_shapes, &shape_list);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading shape file %s", parameters.shape_filename);

        ias_misc_free_report_header(parameters.common_report_header);
        close_input_data(&input_data);
        return EXIT_FAILURE;
    }
    if (number_of_shapes == 0)
    {
        IAS_LOG_ERROR("No shapes found in shape file %s",
            parameters.shape_filename);

        ias_misc_free_report_header(parameters.common_report_header);
        close_input_data(&input_data);
        return EXIT_FAILURE;
    }

    /***** Populate the PICS_IMAGE items *****/

    /* Allocate the memory for the PICS images */
    pics_images = calloc(number_of_pics_images, sizeof(*pics_images));
    if (pics_images == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for PICS image looping");

        ias_misc_free_report_header(parameters.common_report_header);
        close_input_data(&input_data);
        return EXIT_FAILURE;
    }

    /* Always place the quality band information into the first index */
    image_index = 0;
    pics_images[image_index].l1g_image_ptr = input_data.l1g_quality_ptr;
    pics_images[image_index].l1g_metadata = input_data.l1g_quality_metadata;
    pics_images[image_index].processing_type = PICS_QUALITY_PROCESSING;
    pics_images[image_index].band_number = 1; /* Only 1 in the quality */
    image_index++;
    
    /* Place the band list information into the remaining indexes for the
       bands that are present */
    for (band_index = 0;
         band_index < parameters.num_bands_to_process;
         band_index++)
    {
        if (band_present[band_index] != ERROR)
        {
            pics_images[image_index].l1g_image_ptr = input_data.l1g_image_ptr;
            pics_images[image_index].l1g_metadata = input_data.l1g_metadata;
            pics_images[image_index].processing_type = PICS_STANDARD_PROCESSING;
            pics_images[image_index].band_number =
                parameters.band_list[band_index];

            image_index++; /* Advance to the next element of the image array */
        }
    }

    IAS_LOG_DEBUG("Configuring projection information...");
    /* Create the geographic projection information */
    for (index = 0; index < IAS_PROJ_PARAM_SIZE; index++)
    {
        geographic_projection_parameters[index] = 0.0;
    }

    ias_geo_set_projection(GEO, NULLZONE, DEGREE, WGS84_SPHEROID,
        geographic_projection_parameters, &geographic_projection);

    /* Create the projection information for the input image */
    if (ias_geo_get_units(input_data.l1g_metadata.projection_units,
        &projection_units) != SUCCESS)
    {
        IAS_LOG_ERROR("Converting projection units %s to unit code",
            input_data.l1g_metadata.projection_units);

        ias_misc_free_report_header(parameters.common_report_header);
        ias_shape_free(shape_list);
        free(pics_images);
        close_input_data(&input_data);
        return EXIT_FAILURE;
    }

    ias_geo_set_projection(input_data.l1g_metadata.projection_code,
        input_data.l1g_metadata.zone_code, projection_units,
        input_data.l1g_metadata.spheroid_code,
        input_data.l1g_metadata.projection_parameters,
        &input_image_projection);

    /* Create the projection transformation information */
    projtran_geographic_2_l1g = ias_geo_create_proj_transformation(
        &geographic_projection, &input_image_projection);
    if (projtran_geographic_2_l1g == NULL)
    {
        IAS_LOG_ERROR("Creating projection transformation");

        ias_misc_free_report_header(parameters.common_report_header);
        ias_shape_free(shape_list);
        free(pics_images);
        close_input_data(&input_data);
        return EXIT_FAILURE;
    }
    /* Create the projection transformation information */
    projtran_l1g_2_geographic = ias_geo_create_proj_transformation(
        &input_image_projection, &geographic_projection);
    if (projtran_l1g_2_geographic == NULL)
    {
        IAS_LOG_ERROR("Creating projection transformation");

        ias_misc_free_report_header(parameters.common_report_header);
        ias_shape_free(shape_list);
        free(pics_images);
        close_input_data(&input_data);
        ias_geo_destroy_proj_transformation(projtran_geographic_2_l1g);
        return EXIT_FAILURE;
    }


    /* Create report file */
    status = create_report_file_with_header(&parameters, &report_fd);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Creating report file");

        ias_misc_free_report_header(parameters.common_report_header);
        ias_shape_free(shape_list);
        free(pics_images);
        close_input_data(&input_data);
        ias_geo_destroy_proj_transformation(projtran_geographic_2_l1g);
        ias_geo_destroy_proj_transformation(projtran_l1g_2_geographic);
        return EXIT_FAILURE;
    }

    /* Free the report header it is no longer needed */
    ias_misc_free_report_header(parameters.common_report_header);
    parameters.common_report_header = NULL;

    /* Write the scaling parameters that will be used to the report file */
    status = write_scaling_parameters(report_fd, pics_images,
        number_of_pics_images);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing scaling parameters");

        ias_shape_free(shape_list);
        free(pics_images);
        close_input_data(&input_data);
        ias_geo_destroy_proj_transformation(projtran_geographic_2_l1g);
        ias_geo_destroy_proj_transformation(projtran_l1g_2_geographic);
        fclose(report_fd);
        return EXIT_FAILURE;
    }

    /* Write the column headers to the report file */
    status = write_column_headers(report_fd);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing scaling parameters");

        ias_shape_free(shape_list);
        free(pics_images);
        close_input_data(&input_data);
        ias_geo_destroy_proj_transformation(projtran_geographic_2_l1g);
        ias_geo_destroy_proj_transformation(projtran_l1g_2_geographic);
        fclose(report_fd);
        return EXIT_FAILURE;
    }

    /* Allocate the memory for the shape statistics */
    statistics = calloc(number_of_shapes, sizeof(*statistics));
    if (statistics == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for statistics");

        ias_shape_free(shape_list);
        free(pics_images);
        close_input_data(&input_data);
        ias_geo_destroy_proj_transformation(projtran_geographic_2_l1g);
        ias_geo_destroy_proj_transformation(projtran_l1g_2_geographic);
        fclose(report_fd);
        return EXIT_FAILURE;
    }

    /* Allocate the memory for the shape_outside_image_flags pointer with
       all the flags set to FALSE */
    shape_outside_image_flags = calloc(number_of_shapes,
            sizeof(*shape_outside_image_flags));
    if (shape_outside_image_flags == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for shape flag");
        ias_shape_free(shape_list);
        free(pics_images);
        close_input_data(&input_data);
        ias_geo_destroy_proj_transformation(projtran_geographic_2_l1g);
        ias_geo_destroy_proj_transformation(projtran_l1g_2_geographic);
        fclose(report_fd);
        free(statistics);
        return EXIT_FAILURE;
    }

    /* Read the angle coefficient file */
    if (ias_angle_gen_read_ang(parameters.angle_coef_filename, &ang_metadata) 
        != SUCCESS)
    {
        IAS_LOG_ERROR("Reading the angle coefficient file %s", 
            parameters.angle_coef_filename);
        ias_shape_free(shape_list);
        free(pics_images);
        close_input_data(&input_data);
        ias_geo_destroy_proj_transformation(projtran_geographic_2_l1g);
        ias_geo_destroy_proj_transformation(projtran_l1g_2_geographic);
        fclose(report_fd);
        free(statistics);
        free(shape_outside_image_flags);
        return EXIT_FAILURE;
    }

    /* Connect to the database if needed */
    if (parameters.store_char_in_db)
    {
        if (initialize_database(&parameters, &database) != SUCCESS)
        {
            IAS_LOG_ERROR("Unable initialize database library");

            ias_shape_free(shape_list);
            free(pics_images);
            close_input_data(&input_data);
            ias_geo_destroy_proj_transformation(projtran_geographic_2_l1g);
            ias_geo_destroy_proj_transformation(projtran_l1g_2_geographic);
            fclose(report_fd);
            free(statistics);
            free(shape_outside_image_flags);
            ias_angle_gen_free(&ang_metadata);
            return EXIT_FAILURE;
        }
    }

    IAS_LOG_DEBUG("Processing %d bands...", parameters.num_bands_to_process);

    processing_status = SUCCESS;

    /* Loop through each of the bands */
    for (image_index = 0;
         image_index < number_of_pics_images && processing_status == SUCCESS;
         image_index++)
    {
        IAS_L1G_BAND_METADATA band_metadata; /* Metadata for the current band */
        int point_index;                /* Loop control */
        char filename_prefix[PATH_MAX]; /* Thumbnail filename prefix */
        IAS_SHAPE *current_shape;       /* Current shape to process */

        /* Band information */
        int band_number;               /* Band number of the current band */
        L1G_BAND_IO *l1g_band_ptr;     /* Pointer to the requested band */
        IAS_DATA_TYPE data_type;       /* Data type of the band information */
        int number_of_scas;            /* Number of SCAs in the band */
        int number_of_image_lines;     /* Number of lines in the band */
        int number_of_image_samples;   /* Number of samples in the band */

        int number_of_buffer_lines;    /* Number of lines in the buffer */
        int number_of_buffer_samples;  /* Number of samples in the buffer */
        int shape_index;               /* Index to shape */
        int outside_shape_count;       /* Counter for shape outside the image */

        band_number = pics_images[image_index].band_number;
        band_index = ias_sat_attr_convert_band_number_to_index(band_number);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Converting band number %d to index", band_number);
           
            processing_status = ERROR;
            break;
        }
 
        IAS_LOG_DEBUG("Processing band number [%d]...", band_number);

        /* Get the band metadata */
        status = ias_l1g_get_band_metadata(
            pics_images[image_index].l1g_image_ptr,
            band_number, &band_metadata);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading band %d of metadata", band_number);

            processing_status = ERROR;
            break;
        }

        /* Open the L1G band */
        l1g_band_ptr = ias_l1g_open_band(
            pics_images[image_index].l1g_image_ptr,
            band_number, &data_type, &number_of_scas,
            &number_of_image_lines, &number_of_image_samples);
        if (l1g_band_ptr == NULL)
        {
            IAS_LOG_ERROR("Opening L1G band %d", band_number);

            processing_status = ERROR;
            break;
        }

        /* Only support IAS_UI2 input data */
        if (data_type != IAS_UI2)
        {
            IAS_LOG_ERROR("The input image contains an unsupported data type:"
                " Only UI*2 is supported");

            processing_status = ERROR;
            break;
        }

        /* Verify that only 1 SCA is present in the image data */
        if (number_of_scas != 1)
        {
            IAS_LOG_ERROR("Multiple SCAs present in the input image, but only"
                " SCA combined images are supported: Found %d SCAs in image",
                number_of_scas);

            processing_status = ERROR;
            break;
        }

        shape_index = 0;
        outside_shape_count = 0;
        current_shape = shape_list;

        while (current_shape && processing_status == SUCCESS)
        {
            double *new_double_ptr = NULL;
            IAS_UI2_TYPE *new_thumbnail_ptr = NULL;
            double upper_line;
            double lower_line;
            double left_sample;
            double right_sample;

            /* These are for the upper left corner of the envi thumbnail */
            double upper_left_corner_map_y;
            double upper_left_corner_map_x;
            IAS_LOG_DEBUG("Processing shape [%s]...", current_shape->region_id);

            /* Check flag if the shape is inside or outside the image */
            /* If the shape is outside the image go to next shape */
            if (shape_outside_image_flags[shape_index] == TRUE)
            {
                /* Advance to the next shape */
                current_shape = current_shape->next;
                shape_index++;
                continue;
            }

            /* Update the band number for this shape */
            statistics[shape_index].band_number = band_number;

            /* Set to the current number of vertices and type */
            l1g_shape.number_of_vertices = current_shape->npts;
            l1g_shape.shape_type = current_shape->type;

            /* If IAS_SHAPE_BOX fix the number of vertices because npts is the
               number of latitude/longitude points read from the shape file */
            if (l1g_shape.shape_type == IAS_SHAPE_BOX)
            {
                l1g_shape.number_of_vertices = 4;
            }

            /* Re-allocate memory for the new number of vertices */
            new_double_ptr = realloc(l1g_shape.line,
                l1g_shape.number_of_vertices * sizeof(double));
            if (new_double_ptr == NULL)
            {
                IAS_LOG_ERROR("Allocating the line vertex buffer");

                processing_status = ERROR;
                break;
            }
            l1g_shape.line = new_double_ptr;

            new_double_ptr = realloc(l1g_shape.sample,
                l1g_shape.number_of_vertices * sizeof(double));
            if (new_double_ptr == NULL)
            {
                IAS_LOG_ERROR("Allocating the sample vertex buffer");

                processing_status = ERROR;
                break;
            }
            l1g_shape.sample = new_double_ptr;

            /* Create the l1g shape information */
            status = convert_from_geographic_to_image(current_shape,
                projtran_geographic_2_l1g, projtran_l1g_2_geographic,
                &band_metadata, &l1g_shape);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Converting to image line and sample");

                processing_status = ERROR;
                break;
            }

            /* Initialize the line/samples */
            upper_line = DBL_MAX;
            lower_line = -DBL_MAX;
            left_sample = DBL_MAX;
            right_sample = -DBL_MAX;

            /* Find the min and max bounding box lines and samples */
            for (point_index = 0;
                 point_index < l1g_shape.number_of_vertices;
                 point_index++)
            {
                /* Line */
                if (l1g_shape.line[point_index] < upper_line)
                    upper_line = l1g_shape.line[point_index];
                if (l1g_shape.line[point_index] > lower_line)
                    lower_line = l1g_shape.line[point_index];

                /* Sample */
                if (l1g_shape.sample[point_index] < left_sample)
                    left_sample = l1g_shape.sample[point_index];
                if (l1g_shape.sample[point_index] > right_sample)
                    right_sample = l1g_shape.sample[point_index];
            }

            /* Round to nearest int */
            upper_line = round(upper_line);
            lower_line = round(lower_line);
            left_sample = round(left_sample);
            right_sample = round(right_sample);
 
            /* Check if the shape file bounding box is outside the 
               image space and set the shape list flag */
            if ((image_index == 0) && 
                (lower_line < 0 || upper_line >= number_of_image_lines ||
                right_sample < 0 || left_sample >= number_of_image_samples))
            { 
                /* If the shape file bounding box is outside the image space
                   then flag it to TRUE */
                shape_outside_image_flags[shape_index] = TRUE;
                IAS_LOG_WARNING(
                    "Shape file bounding box out of image, shape name, %s",
                    current_shape->region_id);
                outside_shape_count++;

                /* Advance to the next shape */
                current_shape = current_shape->next;
                shape_index++;
                
                /* If all the shapes are out of the image then log the
                   the ERROR and exit */
                if (number_of_shapes == outside_shape_count)
                {
                    IAS_LOG_ERROR("All shapes found are invalid %s",
                    parameters.shape_filename);
                    
                    processing_status = ERROR;
                    break;
                }
                continue;
            }
            
            /* Convert the upper left corner point from image to map */
            upper_left_corner_map_y = band_metadata.upper_left_y
                - (upper_line * band_metadata.projection_distance_y);

            upper_left_corner_map_x =
                (left_sample * band_metadata.projection_distance_x)
                + band_metadata.upper_left_x;

            /**** Only performed for quality band processing ****/
            if (pics_images[image_index].processing_type
                == PICS_QUALITY_PROCESSING)
            {
                /* Only need to specify the region ID for the first band
                   processed, which will always be the quality band */
                count = snprintf(
                    statistics[shape_index].region_of_interest_id,
                    sizeof(statistics[shape_index].region_of_interest_id),
                    current_shape->region_id);
                if (count < 0
                    || count >= sizeof(statistics[shape_index]
                                    .region_of_interest_id))
                {
                    IAS_LOG_ERROR("Copying region of interest ID (%s)",
                        current_shape->region_id);

                    processing_status = ERROR;
                    break;
                }
            }

            /* Figure out the buffer lines and samples needed */
            number_of_buffer_lines = lower_line - upper_line + 1;
            number_of_buffer_samples = right_sample - left_sample + 1;

            /* Allocate memory for the thumbnail buffer */
            
            new_thumbnail_ptr = realloc(thumbnail_buffer,
                number_of_buffer_lines * number_of_buffer_samples
                    * sizeof(*thumbnail_buffer));
            if (new_thumbnail_ptr == NULL)
            {
                IAS_LOG_ERROR("Allocating the thumbnail buffer");

                processing_status = ERROR;
                break;
            }
            thumbnail_buffer = new_thumbnail_ptr;

            /* Read the bounding box image data into the thumbnail buffer */
            status = extract_window(l1g_band_ptr,
                number_of_image_lines, number_of_image_samples,
                upper_line, lower_line, left_sample, right_sample,
                number_of_buffer_lines, number_of_buffer_samples,
                thumbnail_buffer);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Creating thumbnail buffer");

                processing_status = ERROR;
                break;
            }

            /* Create the filename prefix for the thumbnail image */
            if (pics_images[image_index].processing_type
                == PICS_QUALITY_PROCESSING)
            {
                count = snprintf(filename_prefix, sizeof(filename_prefix),
                    "q%02d_roi_%s", band_number, current_shape->region_id);
            }
            else
            {
                count = snprintf(filename_prefix, sizeof(filename_prefix),
                    "b%02d_roi_%s", band_number, current_shape->region_id);
            }
            if (count < 0 || count >= sizeof(filename_prefix))
            {
                IAS_LOG_ERROR("Constructing filename prefix for (%s)",
                    current_shape->region_id);

                processing_status = ERROR;
                break;
            }

            /* Write the thumbnail to a file */
            status = save_image(thumbnail_buffer,
                number_of_buffer_lines, number_of_buffer_samples,
                &band_metadata,
                upper_left_corner_map_x, upper_left_corner_map_y,
                &input_image_projection, filename_prefix);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Saving thumbnail for (%s)",
                    filename_prefix);

                processing_status = ERROR;
                break;
            }

            /* Generate statistics for the shape */
            status = generate_statistics(&band_metadata, thumbnail_buffer,
                number_of_buffer_lines, number_of_buffer_samples,
                &l1g_shape, pics_images[image_index].processing_type,
                number_of_image_lines, number_of_image_samples,
                upper_line, lower_line, left_sample, right_sample,
                band_index, &input_data.dem_image, &ang_metadata, 
                &statistics[shape_index]);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Generating statistics for (%s)",
                    filename_prefix);

                processing_status = ERROR;
                break;
            }

            /**** Only performed for standard band processing ****/
            if (pics_images[image_index].processing_type
                == PICS_STANDARD_PROCESSING)
            {
                /* Append stats to report file */
                status = append_statistics_to_report(report_fd,
                    &statistics[shape_index]);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Writing to report file (%s)",
                        filename_prefix);

                    processing_status = ERROR;
                    break;
                }

                if (parameters.store_char_in_db)
                {
                    status = store_statistics(database,
                        parameters.radiometric_wo_id,
                        &statistics[shape_index]);
                    if (status != SUCCESS)
                    {
                        IAS_LOG_ERROR("Storing statistics in database (%s)",
                            filename_prefix);

                        processing_status = ERROR;
                        break;
                    }
                }
            }

            /* Advance to the next shape */
            current_shape = current_shape->next;
            shape_index++;
        } /* END - Shape loop */
 
        /* Close the band */
        ias_l1g_close_band(l1g_band_ptr);
    } /* END - PICS image_index loop */

    /* Free the shape list memory */
    ias_shape_free(shape_list);

    /* Close the database connection */
    if (parameters.store_char_in_db)
    {
        ias_db_close_connection(database);
        ias_db_close_database_lib();
    }

    /* Free the outside image tracking array */
    free(shape_outside_image_flags);

    /* Free the statistics memory */
    free(statistics);

    /* Free the PICS image memory */
    free(pics_images);

    /* Close the input data files */
    close_input_data(&input_data);

    /* Cleanup the projection memory */
    ias_geo_destroy_proj_transformation(projtran_geographic_2_l1g);
    ias_geo_destroy_proj_transformation(projtran_l1g_2_geographic);

    /* Cleanup the shape memory */
    free(l1g_shape.line);
    free(l1g_shape.sample);

    /* Cleanup the band, thumbnail and data buffer memory */
    free(thumbnail_buffer);

    /* Free the angle coefficient metadata */
    ias_angle_gen_free(&ang_metadata);

    /* Close the report file */
    status = fclose(report_fd);
    if (status == EOF)
    {
        IAS_LOG_ERROR("Failed to close the report file");

        processing_status = ERROR; /* To exit with failure below */
    }

    /* Notify the user that the application is terminating */
    if (processing_status != SUCCESS)
    {
        IAS_LOG_ERROR("Failed during processing");

        return EXIT_FAILURE;
    }

    IAS_LOG_INFO("Successfully terminated");

    return EXIT_SUCCESS;
}

