
#ifndef PICS_H
#define PICS_H

#include <limits.h>

#include "ias_miscellaneous.h"
#include "ias_l1r.h"
#include "ias_l1g.h"
#include "ias_grid_io.h"
#include "ias_model_io.h"
#include "ias_shape.h"
#include "ias_angle_gen.h"

#define PICS_FILL_VALUE 0
#define PICS_ANGLE_OUTSIDE_IMAGE 999

typedef struct pics_parameters
{
    /* Structure to hold the ODL and OMF parameters, and the report header */

    IAS_REPORT_HEADER *common_report_header;

    char work_order_id[IAS_WORKORDER_ID_SIZE];  /* Work order ID */
    char pics_report_filename_prefix[PATH_MAX]; /* PICS report prefix */
    char shape_filename[PATH_MAX];              /* Shape file name */
    char radiometric_wo_id[IAS_CHAR_ID_SIZE];
    char l0r_filename[PATH_MAX];        /* L0R filename */
    char l1g_filename[PATH_MAX];        /* L1G filename */
    char dem_filename[PATH_MAX];        /* DEM filename */
    char grid_filename[PATH_MAX];       /* Grid filename */
    char model_filename[PATH_MAX];      /* Model filename */
    char quality_filename[PATH_MAX];    /* Quality filename */
    char angle_coef_filename[PATH_MAX]; /* Angle coefficient filename */

    int num_bands_to_process;      /* Number of bands in the band list */
    int band_list[IAS_MAX_NBANDS]; /* List of bands to process */
    int store_char_in_db;          /* 0 = don't write to database
                                      1 = write to database */
    int report_flag;               /* 0 = no summary report
                                      1 = summary report */
    int satellite_number;            /* Satellite number */
} PICS_PARAMETERS;

typedef struct pics_data_files
{
    /* Structure to hold opened data files and metadata */

    L1GIO *l1g_image_ptr;               /* Pointer to the L1G object */
    IAS_L1G_FILE_METADATA l1g_metadata; /* Holds the metadata information for
                                           the L1G data */
    L1GIO *l1g_quality_ptr;             /* Pointer to the L1G quality object */
    IAS_L1G_FILE_METADATA l1g_quality_metadata; /* Holds the metadata
                                                   information for the L1G
                                                   quality data */
    IAS_GRID_TYPE grid;                 /* Grid data */
    IAS_LOS_MODEL *model_ptr;           /* Pointer to the model data */

    IAS_IMAGE dem_image;                /* DEM data */
} PICS_DATA_FILES;

typedef enum pics_processing_type
{
    /* Used to determine how to process the data */

    PICS_QUALITY_PROCESSING,       /* Quality band processing */
    PICS_STANDARD_PROCESSING       /* Standard processing */
} PICS_PROCESSING_TYPE;

typedef struct pics_image
{
    /* Used to create a list of images/bands to process through */

    L1RIO *l1r_image_ptr;               /* Pointer to the L1R object */
    L1GIO *l1g_image_ptr;               /* Pointer to the L1G object */
    IAS_L1G_FILE_METADATA l1g_metadata; /* Holds the metadata information for
                                           the L1G data */
    PICS_PROCESSING_TYPE processing_type; /* Type of processing */
    int band_number;                    /* Band number to process */
} PICS_IMAGE;

typedef struct pics_shape
{
    /* Structure to hold information for a shape, in l1g line and sample */

    IAS_SHAPE_TYPE shape_type; /* (e.g., polygon, centered box) */

    int number_of_vertices;    /* Number of vertices in the shape*/
    double *line;              /* Shape vertex lines */
    double *sample;            /* Shape vertex samples */
    double centroid_latitude;  /* Centroid latitude for the shape */
    double centroid_longitude; /* Centroid longitude for the shape */
    double centroid_line;      /* Centroid line for the shape */
    double centroid_sample;    /* Centroid sample for the shape */
} PICS_SHAPE;

typedef struct pics_statistics
{
    /* Structure to hold the statistics/database information for the current
       band/shape */

    char region_of_interest_id[MAX_SHAPE_ID_SIZE]; /* Shape identifier */
    IAS_DBL_LAT_LONG centroid;    /* Coordinates of polygon's centroid */
    int band_number;              /* Band number */
    double sun_azimuth_angle;     /* For the center of the bounding box */
    double sun_elevation_angle;   /* For the center of the bounding box */
    double view_azimuth_angle;    /* For the center of the bounding box */
    double view_zenith_angle;     /* For the center of the bounding box */
    double sun_distance;          /* For the center of the bounding box */
    double reflectance_mean;      /* For valid data in the polygon */
    double reflectance_stdev;     /* For valid data in the polygon */
    double reflectance_minimum;   /* For valid data in the polygon */
    double reflectance_maximum;   /* For valid data in the polygon */
    double radiance_mean;         /* For valid data in the polygon */
    double radiance_stdev;        /* For valid data in the polygon */
    double radiance_minimum;      /* For valid data in the polygon */
    double radiance_maximum;      /* For valid data in the polygon */
    int total_pixel_count;        /* Pixels in the polygon */
    double valid_percent;         /* Percent non-fill pixels in the polygon */
    double fill_percent;          /* Percent in the polygon */
    double cloud_cover_percent;   /* Percent in the polygon */
} PICS_STATISTICS;

int read_parameters
(
    /* Reads the input parameter files and populates the required parameter
       data structures */

    const char *odl_filename,   /* I: Input ODL filename */
    PICS_PARAMETERS *parameters /* O: Application parameters */
);

int open_input_data
(
    /* Opens the input data files and metadata used by the application */

    PICS_PARAMETERS *parameters, /* I/O: Application parameters */
    PICS_DATA_FILES *input_data  /* O: The opened data */
);

void close_input_data
(
    /* Closes the input data files and metadata used by the application */

    PICS_DATA_FILES *input_data /* I/O: The data to close */
);

int extract_window
(
    /* Extracts a window from the provided band IO pointer and fills any
       buffer data outside the image with the fill value */

    const L1G_BAND_IO *l1g_band_ptr, /* I: Pointer to the requested band */
    int number_of_image_lines,    /* I: Number of lines in the band */
    int number_of_image_samples,  /* I: Number of samples in the band */
    int upper_line,               /* I: Upper line to read from */
    int lower_line,               /* I: Lower line to read to */
    int left_sample,              /* I: Left sample to read from */
    int right_sample,             /* I: Right sample to read to */
    int number_of_buffer_lines,   /* I: Number of lines output buffer */
    int number_of_buffer_samples, /* I: Number of samples output buffer */
    IAS_UI2_TYPE *output_buffer   /* O: Populated with image data and fill */
);

int save_image
(
    /* Saves the image using the ENVI format in UI*2 */

    const IAS_UI2_TYPE *input_buffer, /* I: Image data to write to the file */
    int number_of_buffer_lines,       /* I: Number of lines input buffer */
    int number_of_buffer_samples,     /* I: Number of samples input buffer */
    const IAS_L1G_BAND_METADATA *band_metadata, /* I: Metadata for the band */
    double upper_left_corner_map_x,   /* I Map x coordinate value */
    double upper_left_corner_map_y,   /* I Map y coordinate value */
    const IAS_PROJECTION *proj_info,  /* I: Projection information */
    const char *filename_prefix       /* I: Prefix of the output filename */
);

int create_report_file_with_header
(
    /* Create the report file and populate it with the header information */

    const PICS_PARAMETERS *parameters, /* I: Application parameters */
    FILE **report_fd                   /* O: File descriptor for the report */
);

int write_scaling_parameters
(
    /* Write the scaling parameters to the report file */

    FILE *report_fd,               /* I: File descriptor for the report */
    const PICS_IMAGE *pics_images, /* I: The pics images to process through */
    int number_of_pics_images      /* I: The number of pics images to process */
);

int write_column_headers
(
    /* Write the column headers to the report file */

    FILE *report_fd /* I: File descriptor for the report */
);

int append_statistics_to_report
(
    /* Write one line of output to the report file */

    FILE *report_fd,                  /* I: File descriptor for the report */
    const PICS_STATISTICS *statistics /* I: The data to write */
);

int generate_statistics
(
    /* Generates the statistics for the valid pixels within the shape */

    const IAS_L1G_BAND_METADATA *band_metadata, /* Metadata for the band */
    const IAS_UI2_TYPE *thumbnail_buffer, /* I: Thumbnail buffer */
    int number_of_buffer_lines,   /* I: Number of lines thumbnail buffer */
    int number_of_buffer_samples, /* I: Number of samples thumbnail buffer */
    const PICS_SHAPE *l1g_shape,  /* I: Shape points in line/sample */ 
    PICS_PROCESSING_TYPE pics_processing_type, /* I: Type of processing */
    int number_of_image_lines,    /* I: Number of lines in the band */
    int number_of_image_samples,  /* I: Number of samples in the band */
    int upper_line,               /* I: Upper line to read from */
    int lower_line,               /* I: Lower line to read to */
    int left_sample,              /* I: Left sample to read from */
    int right_sample,             /* I: Right sample to read to */
    int band_index,               /* I: Band index */
    const IAS_IMAGE *dem_image,   /* I: PICS DEM image data */
    const IAS_ANGLE_GEN_METADATA *ang_metadata, /* I: Angle Coeff metadata */
    PICS_STATISTICS *statistics   /* O: Updated statistics information */
);

int convert_from_geographic_to_image
(
    /* Convert the geographic shape vertices to vertices in image line and
       sample */

    const IAS_SHAPE *current_shape,     /* I: Current shape to process */
    const IAS_GEO_PROJ_TRANSFORMATION *projtran_geographic_2_l1g,
                                        /* I: Projection transformation */
    const IAS_GEO_PROJ_TRANSFORMATION *projtran_l1g_2_geographic,
                                        /* I: Projection transformation */
    const IAS_L1G_BAND_METADATA *band_metadata, /* I: Band metadata */
    PICS_SHAPE *l1g_shape               /* I/O: Shape points in line/sample */ 
);

int remove_statistics
(
    /* Remove all previous entries within the PICS_STATISTICS table for the
       specified RADIOMETRIC_WO_COMMON_ID */

    struct ias_db_connection *database, /* I: Database connection */
    const char *radiometric_wo_id       /* I: Radiometric WO ID */
);

int store_statistics
(
    /* This function will save a set of statistics in the
       RADCHAR.PICS_STATISTICS table for a single shape */

    struct ias_db_connection *database, /* I: Database connection */
    char *radiometric_wo_id,            /* I: Radiometric work order ID */
    PICS_STATISTICS *input_data         /* I: The data to write */
);

int initialize_database
(
    /* Create a connection to the database and establish the
       RADIOMETRIC_WO_COMMON_ID */

    PICS_PARAMETERS *parameters,        /* I/O: Application parameters */
    struct ias_db_connection **database /* O: Database connection */
);

#endif /* PICS_H */

