
#include "logging_channel.h"
#include "ias_const.h"
#include "ias_types.h"
#include "ias_db_insert.h"

#include "pics.h"

/*****************************************************************************

         NAME: remove_statistics

      PURPOSE: Remove all previous entries within the PICS_STATISTICS table
               for the specified RADIOMETRIC_WO_COMMON_ID

 RETURN VALUE: Type = int
               Value    Description
               -------  --------------------------
               SUCCESS  Successful completion
               ERROR    Operation failed

         NOTE: This will allow PICS to be reprocessed without encountering a
               duplicate primary key error when statistics are saved into the
               table

*****************************************************************************/
int remove_statistics
(
    struct ias_db_connection *database, /* I: Database connection */
    const char *radiometric_wo_id       /* I: Radiometric WO ID */
)
{
    char sql[512];
    int status;
    struct ias_db_query *query;

    /* Delete any previous entries for this work order from the table */
    status = snprintf(sql, sizeof(sql),
        "DELETE FROM pics_statistics WHERE radiometric_wo_common_id = %s",
        radiometric_wo_id);
    if ((status < 0) || (status >= sizeof(sql)))
    {
        IAS_LOG_ERROR("SQL command buffer too small, requires %d characters",
            status);

        return ERROR;
    }

    query = ias_db_perform_query(database, sql);
    if (!ias_db_query_was_successful(query))
    {
        char error_message[4096];

        ias_db_query_get_error_message(query, error_message,
            sizeof(error_message));
        IAS_LOG_ERROR("%s", error_message);

        IAS_LOG_ERROR(
            "Deleting previous records from PICS_STATISTICS for "
            "RADIOMETRIC_WO_COMMON_ID=%s ", radiometric_wo_id);

        ias_db_query_close(query);
        return ERROR;
    }

    ias_db_query_close(query);

    return SUCCESS;
}

/*****************************************************************************

         NAME: store_statistics

      PURPOSE: This function will save a set of statistics in the
               RADCHAR.PICS_STATISTICS table for a single shape

 RETURN VALUE: Type = int
               Value    Description
               -------  --------------------------
               SUCCESS  Successful completion
               ERROR    Operation failed

         NOTE: This will allow PICS to be reprocessed without encountering a
               duplicate primary key error when statistics are saved into the
               table.

*****************************************************************************/
int store_statistics
(
    struct ias_db_connection *database, /* I: Database connection */
    char *radiometric_wo_id,            /* I: Radiometric work order ID */
    PICS_STATISTICS *input_data         /* I: The data to write */
)
{
    int status;
    int table_size;

    IAS_DB_INSERT_TABLE database_table[]=
    {
        IAS_DB_TABLE_STRING("radiometric_wo_common_id", radiometric_wo_id),
        IAS_DB_TABLE_STRING("roi_id", input_data->region_of_interest_id),
        IAS_DB_TABLE_INT("band_id", input_data->band_number),
        IAS_DB_TABLE_DOUBLE("sun_azimuth_angle",
            input_data->sun_azimuth_angle),
        IAS_DB_TABLE_DOUBLE("sun_elevation_angle",
            input_data->sun_elevation_angle),
        IAS_DB_TABLE_DOUBLE("view_angle_azimuth", 
            input_data->view_azimuth_angle),
        IAS_DB_TABLE_DOUBLE("view_angle_zenith", input_data->view_zenith_angle),
        IAS_DB_TABLE_DOUBLE("sun_distance", input_data->sun_distance),
        IAS_DB_TABLE_DOUBLE("reflectance_mean", input_data->reflectance_mean),
        IAS_DB_TABLE_DOUBLE("reflectance_stdev",
            input_data->reflectance_stdev),
        IAS_DB_TABLE_DOUBLE("reflectance_minimum",
            input_data->reflectance_minimum),
        IAS_DB_TABLE_DOUBLE("reflectance_maximum",
            input_data->reflectance_maximum),
        IAS_DB_TABLE_DOUBLE("radiance_mean", input_data->radiance_mean),
        IAS_DB_TABLE_DOUBLE("radiance_stdev", input_data->radiance_stdev),
        IAS_DB_TABLE_DOUBLE("radiance_minimum", input_data->radiance_minimum),
        IAS_DB_TABLE_DOUBLE("radiance_maximum", input_data->radiance_maximum),
        IAS_DB_TABLE_INT("total_pixel_count",
            input_data->total_pixel_count),
        IAS_DB_TABLE_DOUBLE("valid_percent", input_data->valid_percent),
        IAS_DB_TABLE_DOUBLE("fill_percent", input_data->fill_percent),
        IAS_DB_TABLE_DOUBLE("cloud_cover_percent",
            input_data->cloud_cover_percent)
    };

    table_size = sizeof(database_table) / sizeof(IAS_DB_INSERT_TABLE);

    status = ias_db_insert_using_table(database, "PICS_STATISTICS",
        database_table, table_size, 1);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Storing PICS statistics in database for"
            " radiometric work order ID %s, region ID %s, band %d",
            radiometric_wo_id, input_data->region_of_interest_id,
            input_data->band_number);

        return ERROR;
    }

    return SUCCESS;
}

