/*****************************************************************************

         NAME: read_parameters

      PURPOSE: Reads the input parameter files and populates the required
               parameter data structures

 RETURN VALUE: Type = int
               Value    Description
               -------  --------------------------
               SUCCESS  Successful completion
               ERROR    Operation failed

        NOTES: If parameter help is successfully provided, application will
               immediately exit

*****************************************************************************/

#include <stdlib.h>
#include <string.h>

#include "logging_channel.h"
#include "ias_const.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"

#include "pics.h"

int read_parameters
(
    const char *odl_filename,   /* I: Input ODL filename */
    PICS_PARAMETERS *parameters /* O: Parameters */
)
{
    int status;          /* Status for each read operation */
    int count;           /* Result of snprintf */
    int parameter_count; /* Count of parameters successfully read */

    /* Local ODL Parameter Variables */
    char omf_filename[PATH_MAX]; /* OMF parameter file name */
    int processing_pass;         /* The processing pass */

    /*------------------------------------------------------------------------
                   READ WORK ORDER PARAMETER FILE
    ------------------------------------------------------------------------*/
    /* Declare the work order parameter table */
    IAS_PARM_DECLARE_TABLE(odl_parameter_table, 7);

    /* Declare the parameters */
    IAS_PARM_WORK_ORDER_ID(odl_parameter_table,
        parameters->work_order_id, sizeof(parameters->work_order_id), 1);

    IAS_PARM_PROCESSING_PASS(odl_parameter_table, &processing_pass, 
        sizeof(processing_pass), 1);

    IAS_PARM_BAND_LIST(odl_parameter_table, IAS_PARM_REQUIRED,
            parameters->band_list, sizeof(parameters->band_list), 1);

    /* Now the custom parameters */
    parameters->shape_filename[0] = '\0'; /* Initialize to empty string */
    IAS_PARM_ADD_STRING(odl_parameter_table, SHAPE_FILENAME,
        "Input shape file name for PICS: overrides the default file"
        " for the path/row", IAS_PARM_OPTIONAL,
        0, NULL, /* no restrictions */
        0, NULL, /* no default */
        parameters->shape_filename, sizeof(parameters->shape_filename), 0);

    const char *default_pics_report_filename_prefix[] = {"pics"};
    IAS_PARM_ADD_STRING(odl_parameter_table, PICS_REPORT_FILENAME_PREFIX,
        "Output report file name prefix for PICS", IAS_PARM_OPTIONAL,
        0, NULL, /* no restrictions */
        1, default_pics_report_filename_prefix, /* Default file name */
        parameters->pics_report_filename_prefix,
        sizeof(parameters->pics_report_filename_prefix), 0);

    IAS_PARM_STORE_CHARACTERIZATION_IN_DB(odl_parameter_table,
            IAS_PARM_OPTIONAL,
            &(parameters->store_char_in_db),
            sizeof(parameters->store_char_in_db), 1);

    int default_report_flag = 0;
    IAS_PARM_ADD_BOOLEAN(odl_parameter_table, REPORT_FLAG,
            "Write characterization results to summary reports "
            "1 = Write summary reports, 0 = Do not write summary reports",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
            &default_report_flag,
            &(parameters->report_flag), sizeof(parameters->report_flag), 1);

    /* Get the total number of items to read from the parameter table */
    parameter_count = IAS_PARM_GET_TABLE_SIZE(odl_parameter_table);

    /* If someone enters -- help, --template, etc, on the command line
       when starting the application, the appropriate information will
       be displayed.  If someone enters a file name, its access will be
       checked.  If a help option was specified or the specified file name
       could not be accessed, a 1 will be returned.  In this case, just
       copy the work order file name (or help option) to the OMF
       file name and continue.  Otherwise, read the work order file */
    status = ias_parm_provide_help(odl_filename, odl_parameter_table,
                parameter_count, IAS_INPUT_PARAMETERS);

    if (status == ERROR)
    {
        IAS_LOG_ERROR("Work order parameter help failed");
        return ERROR;
    }
    else if (status == 1)
    {
        count = snprintf(omf_filename, sizeof(omf_filename), "%s",
            odl_filename);
        if ((count < 0) || (count >= sizeof(omf_filename)))
        {
            IAS_LOG_ERROR("Constructing OMF filename for (%s)", odl_filename);

            return ERROR;
        }
    }
    else
    {
        status = ias_parm_read(odl_filename, odl_parameter_table,
                parameter_count);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading work order parameters from file %s",
                    odl_filename);

            return ERROR;
        }

        /* Construct the full path to the OMF file */
        count = snprintf(omf_filename, sizeof(omf_filename), "%s.omf",
            parameters->work_order_id);
        if ((count < 0) || (count >= sizeof(omf_filename)))
        {
            IAS_LOG_ERROR("Constructing OMF filename for work order (%s)",
                parameters->work_order_id);

            return ERROR;
        }

        /* Get the total number of bands to process */
        parameters->num_bands_to_process =
            IAS_PARM_GET_COUNT_READ(odl_parameter_table, BAND_LIST);

        /* Verify we've got valid values for selected work order parameters */
        IAS_LOG_DEBUG("Checking work order parameters...");

        /* Make sure at least one output source (i.e. database, or summary
           report) has been selected.  If none have, that's an error */
        if ((!parameters->store_char_in_db) && (!parameters->report_flag))
        {
            IAS_LOG_ERROR("No output destination is specified for"
                " characterization results");

            return ERROR;
        }
    }

/*-----------------------------------------------------------------------
                         READ OMF PARAMETER FILE
-------------------------------------------------------------------------*/
    /* Declare the OMF parameter table */
    IAS_PARM_DECLARE_TABLE(omf_param_table, 8);

    /* Declare the parameters */
    IAS_PARM_L0R_FILENAME(omf_param_table, IAS_PARM_REQUIRED,
        parameters->l0r_filename,
        sizeof(parameters->l0r_filename), 1);

    IAS_PARM_L1G_FILENAME_FROM_PASS(omf_param_table, IAS_PARM_REQUIRED,
        parameters->l1g_filename,
        sizeof(parameters->l1g_filename), 1, processing_pass);

    IAS_PARM_DEM_FILENAME_FROM_PASS(omf_param_table, IAS_PARM_REQUIRED,
        parameters->dem_filename,
        sizeof(parameters->dem_filename), 1, processing_pass);

    IAS_PARM_GRID_FILENAME_FROM_PASS(omf_param_table, IAS_PARM_REQUIRED,
        parameters->grid_filename,
        sizeof(parameters->grid_filename), 1, processing_pass);

    IAS_PARM_MODEL_FILENAME_FROM_PASS(omf_param_table, IAS_PARM_REQUIRED,
        parameters->model_filename,
        sizeof(parameters->model_filename), 1, processing_pass);

    IAS_PARM_QUALITY_BAND_FILENAME(omf_param_table, IAS_PARM_REQUIRED,
        parameters->quality_filename,
        sizeof(parameters->quality_filename));

    IAS_PARM_SATELLITE(omf_param_table, &parameters->satellite_number,
        sizeof(parameters->satellite_number), 1);

    IAS_PARM_ANGLE_COEF_FILENAME_FROM_PASS(omf_param_table, IAS_PARM_REQUIRED, 
        parameters->angle_coef_filename, 
        sizeof(parameters->angle_coef_filename), 1, processing_pass);

    /* Get the total number of items to read in the OMF parameter table */
    parameter_count = IAS_PARM_GET_TABLE_SIZE(omf_param_table);

    /* See if the user wanted help or a template this time with the
       OMF file.  If not, just read the file */
    status = ias_parm_provide_help(omf_filename, omf_param_table,
                parameter_count, IAS_OMF_PARAMETERS);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("OMF parameter help failed");

        return ERROR;
    }
    else if (status == 1)
    {
        /* successfully provided help, as requested, so just exit */
        exit(EXIT_SUCCESS);
    }

    /* Read the OMF parameters */
    status = ias_parm_read(omf_filename, omf_param_table, parameter_count);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading OMF parameter file");

        return ERROR;
    }

    /* Done */
    return SUCCESS;
}

