
/*****************************************************************************
         NAME: save_image.c

      PURPOSE: Saves the image using the ENVI format in UI*2.

 RETURN VALUE: Type = int
               Value    Description
               -------  --------------------------
               SUCCESS  Successful completion
               ERROR    Operation failed

         NOTE: Only UTM is supported at this time.  It is assumed.

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "logging_channel.h"
#include "ias_const.h"
#include "ias_types.h"
#include "ias_miscellaneous.h"

#include "pics.h"

int save_image
(
    const IAS_UI2_TYPE *input_buffer, /* I: Image data to write to the file */
    int number_of_buffer_lines,       /* I: Number of lines in the buffer */
    int number_of_buffer_samples,     /* I: Number of samples in the buffer */
    const IAS_L1G_BAND_METADATA *band_metadata, /* I: Metadata for the band */
    double upper_left_corner_map_x,   /* I Map x coordinate value */
    double upper_left_corner_map_y,   /* I Map y coordinate value */
    const IAS_PROJECTION *proj_info,  /* I: Projection information */
    const char *filename_prefix       /* I: Prefix of the output filename */
)
{
    int status;                   /* Capture function/routine statuses */
    int count;                    /* For testing the snprintf */
    int number_of_buffer_values;  /* Amount of data to write */
    char envi_filename[PATH_MAX]; /* Name of the output file to create */
    FILE *output_fd = NULL;       /* File descriptor for the output file */

    /* Create the full filename */
    count = snprintf(envi_filename, sizeof(envi_filename),
        "%s.img", filename_prefix);
    if ((count < 0) || (count >= sizeof(envi_filename)))
    {
        IAS_LOG_ERROR("Constructing ENVI filename for (%s)", filename_prefix);

        return ERROR;
    }

    /* Open the ENVI file for writing */
    output_fd = fopen(envi_filename, "w");
    if (output_fd == NULL)
    {
        IAS_LOG_ERROR("Creating %s", envi_filename);

        return ERROR;
    }

    /* Figure out how many values to write */
    number_of_buffer_values =
        number_of_buffer_lines * number_of_buffer_samples;

    /* Write the buffer data to the file */
    count = fwrite(input_buffer, sizeof(IAS_UI2_TYPE),
        number_of_buffer_values, output_fd);
    if (count != number_of_buffer_values)
    {
        IAS_LOG_ERROR("%s is invalid: Expected output size [%d]:"
            " Wrote [%d] values", envi_filename, number_of_buffer_values,
            count);

        fclose(output_fd);
        return ERROR;
    }

    /* Close the file */
    if (fclose(output_fd) == EOF)
    {
        IAS_LOG_ERROR("Closing file %s", envi_filename);

        return ERROR;
    }

    /* Write the ENVI header */
    status = ias_misc_write_envi_header(envi_filename, proj_info, NULL,
        number_of_buffer_lines, number_of_buffer_samples, 1,
        upper_left_corner_map_x, upper_left_corner_map_y,
        band_metadata->projection_distance_x,
        band_metadata->projection_distance_y, NULL, IAS_UI2);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Creating ENVI header for %s", envi_filename);

        return ERROR;
    }

    return SUCCESS;
}

