#!/usr/bin/perl -w

##############################################################################
#
# This script loads a record file into the database for the specified schema.
#
# Usage:
#     ./dump_records <config file>
#
#     <db connection info> == typically $IAS_DB_COM
#            <record file> == full path to the file
#    [env var replacement] == an environment variable to string replace
#
##############################################################################
#
# Config File Format
#
# Consists of sets of 4 lines, where each line is described below.
#
# line 1 = environment variable for connection information
# line 2 = the name of the table
# line 3 = the sql command to execute
# line 4 = name of the output file to create
#
##############################################################################

use strict;

use ias_const;
use ias_db;

my $dbh;
my $sth;
my $config_filename;

# test for the command line parameters
if (not defined $ARGV[0])
{
    print "Error missing config filename\n";
    exit FAILURE;
}

# test for the existence of the file
if (not -e $ARGV[0])
{
    print "Error config file [$ARGV[0]] does not exist\n";
    exit FAILURE;
}

$config_filename = $ARGV[0] ;

open(CONFIG_FILE, "<$config_filename") or exit FAILURE;

while (<CONFIG_FILE>)
{
    my $db_env_var_name = $_;
    chomp($db_env_var_name);
    my $table_name = <CONFIG_FILE>;
    chomp($table_name);
    my $sql = <CONFIG_FILE>;
    chomp($sql);
    my $output_filename = <CONFIG_FILE>;
    chomp($output_filename);

    # DEBUG printf $db_env_var . "\n";
    # DEBUG printf $sql . "\n";
    # DEBUG printf $output_filename . "\n";

    my $db_env_var_value = $ENV{$db_env_var_name};
    if (not defined $db_env_var_value)
    {
        print "Error environment variable specified but does not exist\n";
        exit FAILURE;
    }

    if ($db_env_var_value eq "")
    {
        print "Error environment variable defined but does not have a value\n";
        exit FAILURE;
    }

    my ($userID, $password) = split('/', $db_env_var_value);

    # connect to the database
    $dbh = ias_db_ConnectToDB($userID, $password);
    if (not $dbh)
    {
        print "Error connecting to database: $ias_db_error\n";
        exit FAILURE;
    }

    # prepare the sql
    $sth = $dbh->prepare($sql);
    if (not $sth)
    {
        print "Error failed to prepare sql [$sql]: $DBI::errstr\n";
        exit FAILURE;
    }

    # execute the sql
    if (not $sth->execute())
    {
        print "Error failed to execute sql [$sql]: $DBI::errstr\n";
        exit FAILURE;
    }

    open(OUTPUT_FILE, ">$output_filename") or exit FAILURE;

    printf OUTPUT_FILE $table_name . "\n";

    # Process the results here
    my $output_columns = 1;
    while (my $db_row_ref = $sth->fetchrow_hashref())
    {
        my @columns = sort { uc($a) cmp uc($b) } keys(%{$db_row_ref});
        my $first_processed = 0;
        if ($output_columns)
        {
            $first_processed = 0;
            foreach my $column (@columns)
            {
                if ($first_processed)
                {
                    printf OUTPUT_FILE ",";
                }

                printf OUTPUT_FILE $column;

                $first_processed = 1;
            }
            printf OUTPUT_FILE "\n";
            $output_columns = 0;
        }
        $first_processed = 0;
        foreach my $column (@columns)
        {
            if ($first_processed)
            {
                printf OUTPUT_FILE "|";
            }

            if (defined $db_row_ref->{$column})
            {
                printf OUTPUT_FILE $db_row_ref->{$column};
            }

            $first_processed = 1;
        }
        printf OUTPUT_FILE "\n";
    }

    close(OUTPUT_FILE);

    $sth->finish();

    # disconnect from the database
    if (not ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_COMMIT))
    {
        print "Warning: Failed disconnect from database: $ias_db_error\n";
    }
}

close(CONFIG_FILE);

# terminate the application
exit SUCCESS;

