#!/bin/csh -f

# If database support is disabled, exit here with the status SUCCESS
if ( ! $?ENABLE_DB_TESTS ) then
    echo
    echo `basename $0` ":  Database tests are disabled."
    echo
    exit 0
endif

# Setup some static values
set pics_test_data_dir = "ias/rps/pics"
set test_data_dir = $IAS_UNIT_TEST_DATA_DIR/$pics_test_data_dir

set testing_dir = "pics_pass_1_test_dir"
set test_id = "Pass 1"

echo $test_data_dir
echo $testing_dir

echo "Testing $test_id processing ..."

##################### Setup a work order directory #####################
# Create the test directory
mkdir -p $testing_dir
if ( $status != 0 ) then
    echo "Failed to create test directory $testing_dir"
    exit 1
endif

# Change to the directory
cd $testing_dir
if ( $status != 0 ) then
    echo "Failed to change directory to $testing_dir"
    exit 1
endif

# Cleanup any old stuff
if ( -e "P1T.omf" ) then
    rm *
    if ( $status != 0 ) then
        echo "Failed to clean directory $testing_dir"
        exit 1
    endif
endif

# symlink the HDF5 files
ln -s $test_data_dir/$testing_dir/*.h5 .
if ( $status != 0 ) then
    echo "Failed linking *.h5 files to $testing_dir"
    exit 1
endif

# symlink the shared testing files
ln -s $test_data_dir/shared/* .
if ( $status != 0 ) then
    echo "Failed linking shared files to $testing_dir"
    exit 1
endif

# symlink the L0R files
ln -s $test_data_dir/l0r_symlink .
if ( $status != 0 ) then
    echo "Failed linking L0R files to $testing_dir"
    exit 1
endif

# symlink the validation directory
ln -s $test_data_dir/pics_pass_1_validation_v2 validation
if ( $status != 0 ) then
    echo "Failed linking validation directory to $testing_dir"
    exit 1
endif

# Copy the OMF file to the test directory
cp ../PICS_test.omf P1T.omf
if ( $status != 0 ) then
    echo "Failed coping OMF file to $testing_dir"
    exit 1
endif

# Copy the Angle coefficient file 
cp ../angle_coefficients_pass1.txt ./angle_coefficients_pass1.txt
if ( $status != 0 ) then
    echo "Failed coping angle coefficient file to $testing_dir"
    exit 1
endif

# Copy the database dump config file to the test directory
cp ../dump_records.P1T.cfg .
if ( $status != 0 ) then
    echo "Failed coping database dump config file to $testing_dir"
    exit 1
endif

##################### Run the PICS software #####################
# Execute the pics application for the pass 1 test
../../pics ../PICS_pass_1_test.odl
if ( $status != 0 ) then
    echo "Failed executing PICS for the $test_id test"
    exit 1
endif

##################### Verify the PICS results #####################
# Verify the report file
sed '/L0R/,$!'d pics.report | diff -q - validation/pics_V3.report
if ( $status != 0 ) then
    echo "Failed executing PICS for the $test_id test"
    exit 1
endif

# Diff all the thumbnail images created
foreach file (*.img)
    diff -q validation/$file $file
    if ( $status != 0 ) then
        echo "Failed diffing $file"
        exit 1
    endif
end

# Diff all the envi headers for the thumbnail images created
foreach file (*.img.hdr)
    diff -q validation/$file $file
    if ( $status != 0 ) then
        echo "Failed diffing $file"
        exit 1
    endif
end

# Compare the results placed into the database
../dump_records.pl dump_records.P1T.cfg
if ( $status != 0 ) then
    echo "Failed dumping PICS records from the database for $test_id test"
    exit 1
endif

diff -q validation/PICS_STATISTICS_V3.records PICS_STATISTICS.records
if ( $status != 0 ) then
    echo "Failed diffing PICS database records for $test_id test"
    exit 1
endif

echo "Successfully tested PICS $test_id"

# Cleanup, and we don't care if it fails
cd ..
rm -rf $testing_dir

