/*********************************************************************
 NAME:             assess_stability

 PURPOSE:   Implements routines to calculate the variability of the base
            signal level and % change in gain (radiance level)

***********************************************************************/
#include <float.h>
#include <math.h>
#include "ias_radiometry.h"
#include "ias_logging.h"
#include "char_rad_stability.h"


/*--------------------------------------------------------------------
 NAME:             assess_gain_stability

 PURPOSE:   Calculate percent variability of the gain (radiance level)

 RETURNS:   Integer status code of SUCCESS or ERROR
----------------------------------------------------------------------*/
int assess_gain_stability
(
    const double *hist_stats_mean,    /* I: Bias-corrected/linearized
                                         histogram statistics or radiances
                                         for all detectors in the current
                                         band/SCA */
    const double *two_sigma_detector, /* I: Two-sigma values for each
                                         detector in the current band/SCA */
    const double two_sigma_sca_mean,  /* I: Mean two-sigma value in the
                                         current band/SCA */
    const double *relative_gains,     /* I: Relative gains data
                                         buffer for all detectors in an SCA
                                         (set to NULL if assessing radiance
                                         stability) */
    const int *statistics_invalid_flags,
                                      /* I: Data invalid flags for all
                                         detectors in the current band/SCA */
    int detectors_per_sca,            /* I: Number of detectors in an SCA */
    double *pcnt_change_detector,     /* O: Percent change in gain(radiance)
                                         for all detectors in the current
                                         band/SCA */
    double *pcnt_change_sca_mean      /* O: Mean percent change in gain
                                         (radiance) for the current
                                         band/SCA */
)
{
    double hist_sca_mean;             /* Histogram statistics mean for
                                         all detectors in the current
                                         band/SCA */
    double *relgain_ptr = NULL;       /* Tracking pointer to relative gains
                                         data buffer */

    int status;
    int detector_index;


    /* Calculate the detector percent change in gain(radiance).  If either
       the two-sigma value or the histogram statistics mean is invalid,
       flag the detector percent change as invalid.  If the histogram
       statistics mean is 0, flag the detector percent change as invalid--
       otherwise, there'll be issues relating to division-by-0 or some
       really tiny number. */

    /* Initialize the tracking pointer to the start of the relative gains
       data buffer */
    relgain_ptr = (double *)relative_gains;

    if (relgain_ptr == NULL)    /* percent change in radiance calculation */
    {
        for (detector_index = 0;
             detector_index < detectors_per_sca;
             detector_index++)
        {
            if (fabs(hist_stats_mean[detector_index]) <= DBL_EPSILON)
            {
                IAS_LOG_WARNING("Detector %d histogram statistics mean is "
                    "effectively 0 -- flagging percent change in radiance "
                    "as invalid", (detector_index + 1));
                pcnt_change_detector[detector_index] = 0.0;
            }
            else
            {
                pcnt_change_detector[detector_index] =
                    (two_sigma_detector[detector_index] * 100.0)
                        / hist_stats_mean[detector_index];
            }
        }
    }
    else                  /* percent change in gain calculation */
    {
        for (detector_index = 0;
             detector_index < detectors_per_sca;
             detector_index++)
        {
            if (fabs(hist_stats_mean[detector_index]) <= DBL_EPSILON)
            {
                IAS_LOG_WARNING("Detector %d histogram statistics mean is "
                    "effectively 0 -- flagging percent change in gain as "
                    "invalid", (detector_index + 1));
                pcnt_change_detector[detector_index] = 0.0;
            }
            else
            {
                pcnt_change_detector[detector_index] =
                    ((*relgain_ptr) * two_sigma_detector[detector_index]
                        * 100.0) / hist_stats_mean[detector_index];
            }
            relgain_ptr++;
        }
    }

    /* Calculate the SCA mean percent change in gain(radiance).  First, get
       the histogram statistics mean */
    status = ias_rps_calculate_filtered_mean(hist_stats_mean,
        statistics_invalid_flags, detectors_per_sca, &hist_sca_mean,
        NULL);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating SCA mean percent change in gain/radiance");
        return ERROR;
    }

    /* Use the same ratio as in the per-detector calculations.  It might
       be a good idea to check the SCA histogram statistics mean here
       as well for 0, just to be safe */
    if (fabs(hist_sca_mean) <= DBL_EPSILON)
    {
        IAS_LOG_WARNING("SCA histogram statistics mean is 0 -- percent "
            "change result is invalid");
        *pcnt_change_sca_mean = 0.0;
    }
    else
        *pcnt_change_sca_mean = (two_sigma_sca_mean / hist_sca_mean) * 100.0;

    /* Done */
    return SUCCESS;
}




/*--------------------------------------------------------------------
 NAME:             assess_signal_stability

 PURPOSE:   Estimates two-sigma variability for each detector based on
            its histogram statistics standard deviation.  Also calculates
            the mean two-sigma variability for each SCA

 RETURNS:   Integer status code of SUCCESS or ERROR
----------------------------------------------------------------------*/
int assess_signal_stability
(
    const double *hist_stats_stdev, /* I: Histogram statistics stdevs for
                                       all detectors in the current 
                                       band/SCA */
    const int *statistics_invalid_flags,
                                    /* I: Data invalid flags for all
                                       detectors in the current band/SCA
                                       (0=valid, nonzero=invalid) */
    int detectors_per_sca,          /* I: Number of detectors in each SCA */
    double *two_sigma_detector,     /* O: Two-sigma values for all detectors
                                       in the current band/SCA */
    double *two_sigma_sca_mean      /* O: Mean two-sigma value for
                                          the current band/SCA */
)
{
    int detector_index;                  /* Detector loop counter */
    int status;


    /* Calculate two-sigma for each detector, then for the current SCA */
    for (detector_index = 0;
         detector_index < detectors_per_sca;
         detector_index++)
    {
        two_sigma_detector[detector_index] =
            2 * hist_stats_stdev[detector_index];
    }


    /* Calculate the two-sigma mean and std deviation for the current SCA */
    status = ias_rps_calculate_filtered_mean(two_sigma_detector,
        statistics_invalid_flags, detectors_per_sca, two_sigma_sca_mean,
        NULL);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating mean two-sigma");
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}
