/*************************************************************************
 NAME:                   char_gain_stability

 PURPOSE:  Implements the processing flow to characterize 60s stability
           of the gains (radiance levels) acquired from "long" solar collects

 RETURNS:  Integer status code of SUCCESS or ERROR

 NOTES:    The scoping of local variables here for the database writes/
           dump files is different than in the other standalone
           characterizations.  Here, the required local variables need to be
           declared at the module level--buffers may be allocated once
           per band (and files may be created once per band, etc), but outputs
           in the data buffers are written to the database and/or dump files
           for each SCA.

***************************************************************************/
#include <unistd.h>                 /* remove() call */
#include <string.h>
#include "ias_miscellaneous.h"
#include "ias_pixel_mask.h"
#include "ias_radiometry.h"
#include "ias_db_insert.h"

#include "char_rad_stability.h"


int char_gain_stability
(
    const RPS_RADSTABILITY_WO_PARAMS wo_params,   /* I: Work order
                                                     parameters */
    const char *l0r_filename,                     /* I: Name of L0R data
                                                     file (includes path) */
    int num_oli_bands,                            /* I: Total number of
                                                     "normal" OLI imaging
                                                     bands */
    int num_oli_scas,                             /* I: Total number of OLI
                                                     SCAs per band */
    IAS_PIXEL_MASK **pmasks,                      /* I/O: Pixel mask array;
                                                     can be NULL if there
                                                     is no pixel mask
                                                     file available--a
                                                     temporary pixel mask
                                                     will be created for
                                                     each band/SCA */
    int num_masks_read,                           /* I: Number of pixel
                                                     masks read from file */
    const struct IAS_CPF_DETECTOR_STATUS *det_status,
                                                  /* I: CPF DETECTOR_STATUS
                                                     group information */
    const struct IAS_CPF_ABSOLUTE_GAINS *absolute_gains,
                                                  /* I: CPF absolute gains */
    const struct IAS_CPF_RELATIVE_GAINS *relative_gains,
                                                  /* I: CPF relative
                                                     gains */
    struct ias_db_connection *db,                 /* I: Characterization
                                                     database connection */
    char *rad_wo_common_id,                       /* I: Database ID
                                                     corresponding to work
                                                     order ID */
    RPS_RADSTABILITY_REPORT_HEADER header_info    /* I: RPS report header
                                                     information */
)
{
    char dumpfile_name[PATH_MAX];           /* Name of dump file--expected
                                               to be written to work order
                                               directory */
    FILE *dump_fptr = NULL;                 /* Dump file handle */

    /* Summary gain/radiance statistics and two-sigmas for all SCAs in
       all bands */
    RPS_RADSTABILITY_DNRAD_BAND_SUMMARY *band_summary = NULL;

    double abs_gain;                       /* Absolute gain for the current
                                              band/SCA (from CPF) */
    double *relgain_ptr = NULL;            /* CPF relative gains for the
                                              current band/SCA */

    int index;                             /* Band loop counter */
    int func_status;                       /* Current function return status */
    int status;                            /* Final application status */

    /* Database stuff */
    IAS_DB_NULL_TYPE *flagged_detector_ids = NULL;
    int *detector_ids = NULL;
    size_t flagged_detector_id_buffer_size = 0;/* Size (in bytes) of the
                                              detector ID buffer flagged
                                              with NULLs */


    /* Sanity check.  If we're calculating radiance internally, we need the
       absolute and the selected set of relative gains from the CPF. */
    if ((wo_params.calculate_radiance == RPS_RADSTABILITY_CALCULATE_RADIANCE)
            && (absolute_gains == NULL || relative_gains == NULL))
    {
        IAS_LOG_ERROR("Work order specifies internal calculation of "
                "radiances, but absolute and/or relative gain "
                "data buffers are invalid");
        return ERROR;
    }

    /* Allocate a band summary data structure to organize summary
       results */
    band_summary = allocate_band_summary_array(num_oli_bands, num_oli_scas);
    if (band_summary == NULL)
    {
        IAS_LOG_ERROR("Allocating summary data array for all bands");
        return ERROR;
    }

    /*  Begin processing */
    status = SUCCESS;            /* Initialize here */

    for (index = 0; index < wo_params.num_bands_to_process; index++)
    {
        DETECTOR_DATA_BUFFERS data_buffers;
        const IAS_BAND_ATTRIBUTES *band_attributes = NULL;
        int *statistics_invalid_flags = NULL;
        int detector_index;
        int sca_index;
        int band_index;
        int num_mask_pixels = 0;

        IAS_LOG_INFO("Processing band %d...", wo_params.bands[index]);

        /* Get the attributes for the current band */
        band_attributes =
            ias_sat_attr_get_band_attributes(wo_params.bands[index]);
        if (band_attributes == NULL)
        {
            IAS_LOG_ERROR("Obtaining attribute information for band %d",
                wo_params.bands[index]);
            free_band_summary_array(band_summary, num_oli_bands);
            return ERROR;
        }

        /* Get the "true" index for the current band */
        band_index = band_attributes->band_index;

        /* Run some sanity checks on the attributes for the current band:
           a.  If it's not a "normal" imaging band, go to the next
           band number
           b.  If it's a TIRS band number (TIRS bands are of spectral
           type IAS_SPECTRAL_THERMAL), don't process it, go to the next
           band */
        if (band_attributes->band_classification != IAS_NORMAL_BAND)
        {
            IAS_LOG_WARNING("Band %d is not a typical imaging band, "
                "continuing to next band in list",
                band_attributes->band_number);
            continue;
        }
        if (band_attributes->spectral_type == IAS_SPECTRAL_THERMAL)
        {
            IAS_LOG_WARNING("Band %d is a TIRS band, continuing to "
                "next band in list", band_attributes->band_number);
            continue;
        }

        /* Fill in the band information in the summary band data
           structure */
        band_summary[band_index].band_number = band_attributes->band_number;
        strcpy(band_summary[band_index].band_name,
               band_attributes->band_name);
        band_summary[band_index].num_scas = band_attributes->scas;

        /* (Re)initialize the detector data buffer pointers to NULL for
           the current band.  This needs to be done so that the cleanup
           routine won't try to free invalid but non-NULL pointers */
        memset(&data_buffers, 0, sizeof(DETECTOR_DATA_BUFFERS));

        /* Allocate the detector data buffers for the current band */
        func_status = allocate_detector_data_buffers(&data_buffers, "GAIN",
            band_attributes->spectral_type,
            band_attributes->detectors_per_sca);
        if (func_status != SUCCESS)
        {
            IAS_LOG_ERROR("Allocating work detector data buffers for "
                    "band %d", band_attributes->band_number);
            free_band_summary_array(band_summary, num_oli_bands);
            free_detector_data_buffers(&data_buffers);
            return ERROR;
        }

        /* Allocate detector ID buffers for the database write (if the
           work order specifies storing characterization results in the
           database).  One buffer is to be populated with all of the
           detector ID numbers for the current band/SCA.  The other is
           to be populated with 0's for non artifact-flagged detectors
           and NULL for artifact-flagged detectors to be excluded from
           further analysis. The populating of these buffers is performed
           just before results are written to the database */
        if (wo_params.store_char_in_db)
        {
            flagged_detector_id_buffer_size =
                band_attributes->detectors_per_sca * sizeof(IAS_DB_NULL_TYPE);

            flagged_detector_ids = malloc(flagged_detector_id_buffer_size);
            detector_ids =
                malloc(band_attributes->detectors_per_sca * sizeof(int));
            if ((detector_ids == NULL) || (flagged_detector_ids == NULL))
            {
                IAS_LOG_ERROR("Allocating required database data "
                    "buffers");
                rad_stability_cleanup(&data_buffers, statistics_invalid_flags,
                    band_summary, num_oli_bands);
                free(flagged_detector_ids);
                free(detector_ids);
                return ERROR;
            }
            for (detector_index = 0;
                 detector_index < band_attributes->detectors_per_sca;
                 detector_index++)
            {
                detector_ids[detector_index] = (detector_index + 1);
            }
        }

        /* Allocate the invalid flag data buffer */
        statistics_invalid_flags = malloc(
            band_attributes->detectors_per_sca * sizeof(int));
        if (statistics_invalid_flags == NULL)
        {
            IAS_LOG_ERROR("Allocating histogram statistics invalid flag "
                "data buffers for band %d", band_attributes->band_number);
            free(flagged_detector_ids);
            free(detector_ids);
            rad_stability_cleanup(&data_buffers, statistics_invalid_flags,
                band_summary, num_oli_bands);
            return ERROR;
        }

        /* Set up the dump file for the current band */
        if (wo_params.dump_flag)
        {
            /* Set the band number in the report header */
            ias_misc_set_header_band_number(header_info.common_header,
                band_attributes->band_number);

            /* Open the dump file */
            func_status = snprintf(dumpfile_name, sizeof(dumpfile_name),
                "%s_dumpfile_B%02d.txt", wo_params.gain_report_filename_prefix,
                band_attributes->band_number);
            if ((func_status < 0) || (func_status >= sizeof(dumpfile_name)))
            {
                IAS_LOG_ERROR("Creating band %d gain stability dump file "
                    "name", band_attributes->band_number);
                rad_stability_cleanup(&data_buffers,
                    statistics_invalid_flags, band_summary,
                    num_oli_bands);
                return ERROR;
            }
            dump_fptr = fopen(dumpfile_name, "w");
            if (dump_fptr == NULL)
            {
                IAS_LOG_ERROR("Cannot open %s", dumpfile_name);
                rad_stability_cleanup(&data_buffers,
                    statistics_invalid_flags, band_summary, num_oli_bands);
                return ERROR;
            }

            func_status = write_header_to_dumpfile(header_info, dump_fptr);
            if (func_status != SUCCESS)
            {
                IAS_LOG_ERROR("Cannot write header information to dump "
                              "file %s", dumpfile_name);
                fclose(dump_fptr);
                remove(dumpfile_name);
                rad_stability_cleanup(&data_buffers,
                    statistics_invalid_flags, band_summary,
                    num_oli_bands);
                return ERROR;
            }
        }

/*-------------------------------------------------------------------------
                            BEGIN SCA PROCESSING
---------------------------------------------------------------------------*/
        /* Process each SCA */
        for (sca_index = 0; sca_index < band_attributes->scas; sca_index++)
        {
            IAS_PIXEL_MASK *pm_ptr = NULL;
            int mask_array_index;
            int sca_number = sca_index + 1;

            /* Fill in the SCA information in the band summary
               data structure */
            band_summary[band_index].summary_stats[sca_index]
                .sca_number = sca_number;
            band_summary[band_index].siglevel_change[sca_index]
                .sca_number = sca_number;

            /* Prepare the pixel mask for the current band/SCA. */
            if (pmasks != NULL)
            {
                /* The mask we need should be in the array read in from
                   the pixel mask file.  Determine the array index for the
                   correct mask, then set the local pointer to that mask */
                mask_array_index = ias_pm_get_mask_index(pmasks,
                    num_masks_read, band_attributes->band_number,
                    sca_number);
                if (mask_array_index == ERROR)
                {
                    IAS_LOG_ERROR("Obtaining index for band %d SCA %d pixel "
                        "mask", band_attributes->band_number, sca_number);
                    if (dump_fptr)
                    {
                        fclose(dump_fptr);
                    }
                    rad_stability_cleanup(&data_buffers,
                        statistics_invalid_flags, band_summary,
                        num_oli_bands);
                    return ERROR;
                }
                pm_ptr = pmasks[mask_array_index];

                /* The dimensions of the source 'image' (number of
                   detectors and the number of pixels per detector)
                   are part of the internal mask information.  Get the
                   number of pixels per detector for the current
                   band/SCA.  */
                num_mask_pixels = ias_pm_get_number_of_pixels(pm_ptr);
                if (num_mask_pixels == ERROR)
                {
                    IAS_LOG_ERROR("Obtaining number of pixels in band %d "
                        "SCA %d mask", band_attributes->band_number,
                        sca_number);
                    if (dump_fptr)
                    {
                        fclose(dump_fptr);
                    }
                    rad_stability_cleanup(&data_buffers,
                        statistics_invalid_flags, band_summary,
                        num_oli_bands);
                    return ERROR;
                }
            }
            else
            {
                /* We don't have any pixel masks available.  However, we do
                   have available the number of detectors in the current
                   band/SCA (in 'band_attributes->detectors_per_sca').  For
                   the purposes of this characterization, the number of
                   pixels per detector can be any value >= 1, because the
                   inoperable and out-of-spec artifacts considered here
                   are detector-based (i.e. they affect all of a detector's
                   image pixels).   */

                /* Set the number of pixels in the mask. */
                num_mask_pixels = 10;

                /* Create the empty mask. */
                pm_ptr = ias_pm_create(band_attributes->band_number,
                    sca_number, band_attributes->detectors_per_sca,
                    num_mask_pixels);
                if (pm_ptr == NULL)
                {
                    IAS_LOG_ERROR("Creating local pixel mask for band %d "
                        "SCA %d", band_attributes->band_number,
                        sca_number);
                    if (dump_fptr)
                    {
                        fclose(dump_fptr);
                    }
                    rad_stability_cleanup(&data_buffers,
                        statistics_invalid_flags, band_summary,
                        num_oli_bands);
                    return ERROR;
                }

                /* Add the CPF inoperable detector information to the empty
                   mask.  */
                status = ias_rps_inoperable_detector_identification(
                    det_status->inoperable[band_index][sca_index],
                    det_status->inoperable_count[band_index][sca_index],
                    band_attributes->band_number, sca_number,
                    band_attributes->detectors_per_sca, num_mask_pixels,
                    pm_ptr);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Adding inoperable detector "
                        "information to band %d SCA %d pixel mask",
                        band_attributes->band_number,
                        sca_number);
                    if (dump_fptr)
                    {
                        fclose(dump_fptr);
                    }
                    rad_stability_cleanup(&data_buffers,
                        statistics_invalid_flags, band_summary,
                        num_oli_bands);
                    return ERROR;
                }
            }

            /* Add the CPF out-of-spec detector information to the mask.  */
            status = ias_rps_outofspec_detector_identification(
                det_status->out_of_spec[band_index][sca_index],
                det_status->out_of_spec_count[band_index][sca_index],
                band_attributes->band_number, sca_number,
                band_attributes->detectors_per_sca, num_mask_pixels,
                pm_ptr);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Adding out-of-spec detector information "
                    "to band %d SCA %d pixel mask",
                    band_attributes->band_number, sca_number);
                if (dump_fptr)
                {
                    fclose(dump_fptr);
                }
                rad_stability_cleanup(&data_buffers,
                    statistics_invalid_flags, band_summary,
                    num_oli_bands);
                if (pmasks == NULL)
                {
                    ias_pm_destroy(pm_ptr);
                }
                return ERROR;
            }

            /* Populate the input data invalid flag for the current
               SCA.  All artifacts need to be considered for this
               characterization. */
            status = ias_rps_get_invalid_data_flags(pm_ptr, ~IAS_RPS_NOVALUE,
                band_attributes->detectors_per_sca, statistics_invalid_flags);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Populating invalid flag information for "
                    "band number %d SCA number %d",
                    band_attributes->band_number, sca_number);
                rad_stability_cleanup(&data_buffers,
                    statistics_invalid_flags, band_summary,
                    num_oli_bands);
                if (pmasks == NULL)
                {
                    ias_pm_destroy(pm_ptr);
                }
                return ERROR;
            }

            /* Get the bias-corrected / linearized histogram statistics
               for each detector in the current band/SCA.  We'll need
               these for the signal stability and change in gain
               metrics independent of the radiance source */
            func_status = get_histogram_statistics(db, rad_wo_common_id,
                    band_attributes->spectral_type,
                    band_attributes->band_number, sca_number,
                    band_attributes->detectors_per_sca,
                    wo_params.rps_solar_proc_level_dn,
                    data_buffers.max_dn,
                    data_buffers.min_dn, data_buffers.mean_dn,
                    data_buffers.stdev_dn);
            if (func_status != SUCCESS)
            {
                IAS_LOG_ERROR("Getting bias-corrected/linearized statistics "
                        "for band %d SCA %d", band_attributes->band_number,
                        sca_number);
                if (dump_fptr)
                {
                    fclose(dump_fptr);
                }
                if (pmasks == NULL)
                {
                    ias_pm_destroy(pm_ptr);
                }
                rad_stability_cleanup(&data_buffers,
                    statistics_invalid_flags, band_summary,
                    num_oli_bands);
                return ERROR;
            }

            func_status = generate_summary_statistics(
                data_buffers.max_dn, data_buffers.min_dn,
                data_buffers.mean_dn,
                statistics_invalid_flags,
                band_attributes->detectors_per_sca,
                "DN", &(band_summary[band_index].summary_stats[sca_index]));
            if (func_status != SUCCESS)
            {
                IAS_LOG_ERROR("Getting summary DN histogram statistics "
                    "in band %d SCA %d", band_attributes->band_number, 
                    sca_number);
                rad_stability_cleanup(&data_buffers,
                    statistics_invalid_flags, band_summary,
                    num_oli_bands);
                if (dump_fptr != NULL)
                {
                    fclose(dump_fptr);
                }
                if (pmasks == NULL)
                {
                     ias_pm_destroy(pm_ptr);
                }
                return ERROR;
            }

            /* Get the radiance histogram statistics from the
               characterization database, or calculate them directly for
               the current band/SCA */
            if (wo_params.calculate_radiance
                == RPS_RADSTABILITY_USE_DB_RADIANCE)
            {
                func_status = get_histogram_statistics(db, rad_wo_common_id,
                    band_attributes->spectral_type,
                    band_attributes->band_number, sca_number,
                    band_attributes->detectors_per_sca,
                    wo_params.rps_solar_proc_level_rad,
                    data_buffers.max_rad, data_buffers.min_rad,
                    data_buffers.mean_rad, data_buffers.stdev_rad);
                if (func_status != SUCCESS)
                {
                    IAS_LOG_ERROR("Getting radiance histogram statistics "
                        "for band %d SCA %d", band_attributes->band_number,
                        sca_number);
                    if (dump_fptr)
                    {
                        fclose(dump_fptr);
                    }
                    if (pmasks == NULL)
                    {
                        ias_pm_destroy(pm_ptr);
                    }
                    rad_stability_cleanup(&data_buffers,
                        statistics_invalid_flags, band_summary,
                        num_oli_bands);
                    return ERROR;
                }

                /* Get the summary radiance statistics. */
                func_status = generate_summary_statistics(
                    data_buffers.max_rad, data_buffers.min_rad,
                    data_buffers.mean_rad,
                    statistics_invalid_flags,
                    band_attributes->detectors_per_sca,
                    "RADIANCE",
                    &(band_summary[band_index].summary_stats[sca_index]));
                if (func_status != SUCCESS)
                {
                    IAS_LOG_ERROR("Getting summary radiance histogram "
                        "statistics in band %d SCA %d",
                        band_attributes->band_number, sca_number);
                    rad_stability_cleanup(&data_buffers,
                        statistics_invalid_flags, band_summary,
                        num_oli_bands);
                    if (dump_fptr != NULL)
                    {
                        fclose(dump_fptr);
                    }
                    if (pmasks == NULL)
                    {
                        ias_pm_destroy(pm_ptr);
                    }
                    return ERROR;
                }
            }
            else
            {
                /* We'll have to generate the SCA-summary level
                   statistics directly in this case.  It's fairly
                   straightforward to do so, but a bit tedious.  The
                   steps are as follows:  */

                /* 1.  Get the proper CPF absolute gain and detector relative
                   gains for the current band/SCA */
                abs_gain = absolute_gains->gain[band_index][sca_index];
                relgain_ptr =
                    relative_gains->per_detector[band_index][sca_index];

                /* 2.  Calculate per-detector radiance means and standard
                   deviations from the bias-corrected / linearized histogram
                   statistics.  Assumes correction for nonuniform panel
                   response is not required; if it is, there should be a
                   call to a function such as
                   correct_nonuniform_panel_response (which is used
                   in nonuniformity characterization, but would need to
                   be incorporated into this algorithm somehow) */
                ias_rps_calculate_radiance(data_buffers.mean_dn,
                    band_attributes->detectors_per_sca,
                    abs_gain, relgain_ptr, data_buffers.mean_rad);

                ias_rps_calculate_radiance(data_buffers.stdev_dn,
                    band_attributes->detectors_per_sca,
                    abs_gain, relgain_ptr, data_buffers.stdev_rad);

                /* 3. Calculate the maximum and minimum radiances from
                   the bias-corrected / linearized histogram statistics.
                   Again, assume no correction for panel nonuniform
                   response is required */
                ias_rps_calculate_radiance(data_buffers.max_dn,
                    band_attributes->detectors_per_sca,
                    abs_gain, relgain_ptr, data_buffers.max_rad);

                ias_rps_calculate_radiance(data_buffers.min_dn,
                    band_attributes->detectors_per_sca, abs_gain,
                    relgain_ptr, data_buffers.min_rad);

                /* Get the summary radiance statistics for the current
                   band/SCA.  */
                func_status = generate_summary_statistics(
                      data_buffers.max_rad,
                      data_buffers.min_rad,
                      data_buffers.mean_rad,
                      statistics_invalid_flags,
                      band_attributes->detectors_per_sca,
                      "RADIANCE",
                      &(band_summary[band_index].summary_stats[sca_index]));
                if (func_status != SUCCESS)
                {
                    IAS_LOG_ERROR("Getting summary radiance histogram "
                                  "statistics in band %d SCA %d",
                                  band_attributes->band_number, 
                                  sca_number);
                    rad_stability_cleanup(&data_buffers,
                        statistics_invalid_flags, band_summary,
                        num_oli_bands);
                    if (dump_fptr != NULL)
                    {
                        fclose(dump_fptr);
                    }
                    if (pmasks == NULL)
                    {
                        ias_pm_destroy(pm_ptr);
                    }
                    return ERROR;
                }
            }

            /* Assess the absolute change in DN(radiance) levels first,
               then assess the percent change in gain/radiance levels */

            /* DN two-sigma */
            func_status = assess_signal_stability(data_buffers.stdev_dn,
                statistics_invalid_flags,
                band_attributes->detectors_per_sca,
                data_buffers.two_sigma_dn,
                &(band_summary[band_index].siglevel_change[sca_index]
                    .sca_two_sigma_dn));
            if (func_status != SUCCESS)
            {
                IAS_LOG_ERROR("Determining DN two-sigma for band %d "
                    "SCA %d", band_attributes->band_number,
                    sca_number);
                if (dump_fptr)
                {
                    fclose(dump_fptr);
                }
                if (pmasks == NULL)
                {
                    ias_pm_destroy(pm_ptr);
                }
                rad_stability_cleanup(&data_buffers,
                    statistics_invalid_flags, band_summary,
                    num_oli_bands);
                    return ERROR;
            }

            /* Radiance two-sigma */
            func_status = assess_signal_stability(data_buffers.stdev_rad,
                statistics_invalid_flags,
                band_attributes->detectors_per_sca,
                data_buffers.two_sigma_rad,
                &(band_summary[band_index].siglevel_change[sca_index]
                    .sca_two_sigma_rad));
            if (func_status != SUCCESS)
            {
                IAS_LOG_ERROR("Determining radiance two-sigma for band "
                    "%d SCA %d", band_attributes->band_number,
                    sca_number);
                if (dump_fptr)
                {
                    fclose(dump_fptr);
                }
                if (pmasks == NULL)
                {
                    ias_pm_destroy(pm_ptr);
                }
                rad_stability_cleanup(&data_buffers,
                    statistics_invalid_flags, band_summary,
                    num_oli_bands);
                return ERROR;
            }

            /* Percent change in gain */
            func_status = assess_gain_stability(data_buffers.mean_dn,
                data_buffers.two_sigma_dn,
                band_summary[band_index].siglevel_change[sca_index]
                    .sca_two_sigma_dn,
                relgain_ptr, statistics_invalid_flags,
                band_attributes->detectors_per_sca,
                data_buffers.pcnt_change_gain,
                &(band_summary[band_index].siglevel_change[sca_index]
                    .sca_pcnt_change_gain));
            if (func_status != SUCCESS)
            {
                IAS_LOG_ERROR("Determining percent change in gains for "
                    "band %d SCA %d", band_attributes->band_number,
                    sca_number);
                if (dump_fptr)
                {
                    fclose(dump_fptr);
                }
                if (pmasks == NULL)
                {
                    ias_pm_destroy(pm_ptr);
                }
                rad_stability_cleanup(&data_buffers,
                    statistics_invalid_flags, band_summary,
                    num_oli_bands);
                return ERROR;
            }

            /* Perecent change in radiance.  Pass in NULL for the
               relative gain pointer parameter */
            func_status = assess_gain_stability(data_buffers.mean_rad,
                data_buffers.two_sigma_rad,
                band_summary[band_index].siglevel_change[sca_index]
                    .sca_two_sigma_rad,
                NULL,  statistics_invalid_flags,
                band_attributes->detectors_per_sca,
                data_buffers.pcnt_change_rad,
                &(band_summary[band_index].siglevel_change[sca_index]
                    .sca_pcnt_change_rad));
            if (func_status != SUCCESS)
            {
                IAS_LOG_ERROR("Determining percent change in radiance "
                    "for band %d SCA %d", band_attributes->band_number,
                    sca_number);
                if (dump_fptr)
                {
                    fclose(dump_fptr);
                }
                if (pmasks == NULL)
                {
                    ias_pm_destroy(pm_ptr);
                }
                rad_stability_cleanup(&data_buffers,
                    statistics_invalid_flags, band_summary,
                    num_oli_bands);
                return ERROR;
            }

            /* Store results in the database, if the work order specifies
               it */
            if (wo_params.store_char_in_db)
            {
                /* Populate the flagged detector ID buffer allocated
                   earlier */
                memset(flagged_detector_ids, 0,
                       flagged_detector_id_buffer_size);

                for (detector_index = 0;
                     detector_index < band_attributes->detectors_per_sca;
                     detector_index++)
                {
                    if (statistics_invalid_flags[detector_index])
                    {
                        flagged_detector_ids[detector_index] =
                            ias_db_get_null_data_flag();
                    }
                }

                /* Write detector-level percent gain change results to DB */
                func_status = write_detector_gain_stability_results_to_db(db,
                    rad_wo_common_id, detector_ids, flagged_detector_ids,
                    data_buffers.pcnt_change_gain,
                    band_attributes->band_number, sca_number,
                    band_attributes->detectors_per_sca);
                if (func_status != SUCCESS)
                {
                    IAS_LOG_ERROR("Writing detector gain stability results "
                            "for band %d SCA %d",
                            band_attributes->band_number,
                            sca_number);
                    if (dump_fptr)
                    {
                        fclose(dump_fptr);
                    }
                    if (pmasks == NULL)
                    {
                        ias_pm_destroy(pm_ptr);
                    }
                    return ERROR;
                }

                /* Write SCA signal change and percent change results to DB */
                func_status = write_sca_gainrad_stability_results_to_db(db,
                    rad_wo_common_id,
                    band_summary[band_index].siglevel_change[sca_index],
                    band_attributes->band_number, 1);
                if (func_status != SUCCESS)
                {
                    IAS_LOG_ERROR("Writing SCA mean stability results to "
                            "characterization database");
                    if (dump_fptr)
                    {
                        fclose(dump_fptr);
                    }
                    if (pmasks == NULL)
                    {
                        ias_pm_destroy(pm_ptr);
                    }
                    rad_stability_cleanup(&data_buffers,
                        statistics_invalid_flags, band_summary,
                        num_oli_bands);
                    return ERROR;
                }
            }

/*-----------------------------------------------------------------------
                        WRITE RESULTS TO DUMP FILE
-------------------------------------------------------------------------*/
            if (wo_params.dump_flag)
            {
                /* Write detector results to dump file.  We'll need to
                   include the (verified) RPS processing parameter level
                   in the header data */
                func_status = write_detector_stability_results_to_dump_file(
                    data_buffers, "GAIN", band_attributes->band_number,
                    band_attributes->band_name, sca_number,
                    band_attributes->detectors_per_sca, dump_fptr);
                if (func_status != SUCCESS)
                {
                    IAS_LOG_ERROR("Writing gain/rad detector results to "
                            "dump file %s in band %d SCA %d",
                            dumpfile_name,
                            band_attributes->band_number,
                            sca_number);
                    if (dump_fptr)
                    {
                        fclose(dump_fptr);
                    }
                    if (pmasks == NULL)
                    {
                        ias_pm_destroy(pm_ptr);
                    }
                    rad_stability_cleanup(&data_buffers,
                        statistics_invalid_flags, band_summary,
                        num_oli_bands);
                    return ERROR;
                }
            }

            /* Destroy the local pixel mask here, and set the pointer to
               NULL */
            if (pmasks == NULL)
            {
                ias_pm_destroy(pm_ptr);
                pm_ptr = NULL;
            }

        }   /* END SCA PROCESSING */

/*-----------------------------------------------------------------------
                        BAND CLEANUP
-------------------------------------------------------------------------*/
        /* Free detector ID data buffers if writing results to the
           database was enabled */
        if (wo_params.store_char_in_db)
        {
            free(detector_ids);
            detector_ids = NULL;

            free(flagged_detector_ids);
            flagged_detector_ids = NULL;
        }

        /* Free up the detector data buffers */
        free_detector_data_buffers(&data_buffers);

        free(statistics_invalid_flags);
        statistics_invalid_flags = NULL;

        /* Write SCA results to the dump file, then close it */
        if (wo_params.dump_flag)
        {
            func_status = write_sca_stability_results_to_dump_file(
                &(band_summary[band_index]), "GAIN", dump_fptr);

            fclose(dump_fptr);
            dump_fptr = NULL;

            if (func_status != SUCCESS)
            {
                IAS_LOG_ERROR("Writing band %d summary results to "
                        "dump file %s", band_attributes->band_number,
                        dumpfile_name);
                free_band_summary_array(band_summary, num_oli_bands);
                return ERROR;
            }
        }
    }   /* END BAND PROCESSING */

/*--------------------------------------------------------------------
                       WRITE SUMMARY REPORT
----------------------------------------------------------------------*/
    /* Write SCA gain summary statistics and two-sigma values to summary
       report */
    if (wo_params.report_flag)
    {
        char reportfile_name[PATH_MAX];     /* Name of summary report file--
                                               expected to be written to
                                               work order directory */
        FILE *rpt_fptr = NULL;              /* Report file handle */

        /* Open/create the summary report file */
        func_status = snprintf(reportfile_name, sizeof(reportfile_name),
            "%s_summary.txt", wo_params.gain_report_filename_prefix);
        if ((func_status < 0) || (func_status >= sizeof(reportfile_name)))
        {
            IAS_LOG_ERROR("Creating summary report file name");
            free_band_summary_array(band_summary, num_oli_bands);
            return ERROR;
        }

        rpt_fptr = fopen(reportfile_name, "w");
        if (rpt_fptr == NULL)
        {
            IAS_LOG_ERROR("Cannot open/create summary report file %s",
                          reportfile_name);
            free_band_summary_array(band_summary, num_oli_bands);
            return ERROR;
        }

        func_status = write_header_to_summary_report(header_info, rpt_fptr);
        if (func_status != SUCCESS)
        {
            IAS_LOG_ERROR("Writing standard RPS report header information "
                          "to summary report file %s", reportfile_name);
            fclose(rpt_fptr);
            remove(reportfile_name);
            free_band_summary_array(band_summary, num_oli_bands);
            return ERROR;
        }

        /* Write the SCA-level gain stability characterization results to the
           summary report file.  Just use the number of bands actually
           in the band list, rather than the total number of OLI bands */
        func_status = write_data_to_summary_report(band_summary, "GAIN",
                num_oli_bands,
                header_info.rps_header.acq_datetime.total_time_oli,
                rpt_fptr);

        /* Close the summary report file */
        fclose(rpt_fptr);
        if (func_status != SUCCESS)
        {
            IAS_LOG_ERROR("Writing gain/rad stability characterization "
                          "results to summary report %s",
                          reportfile_name);
            status = func_status;
        }
    }

    /* Free the band summary data array */
    free_band_summary_array(band_summary, num_oli_bands);

    /* Done */
    return status;
}
