/**************************************************************************
 NAME:                      char_rad_stability

 PURPOSE:       "Driver" that sets up and calls the appropriate
                stability characterization routine depending on the
                type of the calibration data collect (i.e. OLI "SOLAR"
                or "SHUTTER").  Once the characterization program
                ends, also initiates cleanup

 RETURNS:       Status code of EXIT_SUCCESS or EXIT_FAILURE

 NOTES:         This application will run following completion of
                data processing through the RPS Driver.  As currently
                designed, it cannot run from within the Driver because
                the statistics from ALL SCAs must be available to process
                a given band.

***************************************************************************/
#include "ias_const.h"
#include "ias_miscellaneous.h"
#include "ias_radiometry.h"
#include "char_rad_stability.h"


int main(int argc, char *argv[])
{
    /* Database-related stuff */
    struct ias_db_connection *open_db = NULL;   /* IAS database connection
                                                   data structure */

    char rad_wo_common_id[IAS_CHAR_ID_SIZE];   /* Key value in
                                        RADIOMETRIC_WORK_ORDER_COMMON table
                                        (needed to associate processing
                                        with the correct work order)*/

    /* Internal processing */
    char *collection_type = NULL;    /* Collection type string */
    char *sensor_id = NULL;          /* Sensor ID string */
    int status = 0;                  /* Function return status */
    int exit_status;                 /* Final application exit status */
    int is_solar_collect;            /* Current collect is an OLI solar
                                        collect (0=no, 1=yes) */
    int is_shutter_collect;          /* Current collect is an OLI shutter
                                        collect (0=no, 1=yes) */
    int band_list[IAS_MAX_NBANDS];   /* Local band list array */
    int num_oli_bands;               /* Total number of OLI bands */
    int num_oli_scas;                /* Number of SCAs in an OLI band */

    /* Pixel mask stuff */
    IAS_PIXEL_MASK **pmasks = NULL;  /* Pixel mask array for all bands/
                                        SCAs processed */
    IAS_PIXEL_MASK *pm_ptr = NULL;   /* Pointer to an individual pixel
                                        mask */
    int num_masks_read;              /* # of pixel masks successfully read
                                        from file */
    int mask_index;                  /* 0-based index into pixel mask
                                        array */

    /* CPF stuff */
    IAS_CPF *cpf = NULL;              /* CPF data structure */

    /* Detector status information */
    const struct IAS_CPF_DETECTOR_STATUS *det_status = NULL;

    /* OLI-specific parameters */
    const struct IAS_CPF_OLI_PARAMETERS *oli_params = NULL;

    /* CPF relative gains for all bands */
    const struct IAS_CPF_RELATIVE_GAINS *relative_gains = NULL;

    /* CPF absolute gain coefficients for all bands */
    const struct IAS_CPF_ABSOLUTE_GAINS *absolute_gains = NULL;

    /* Processing parameters information */
    RPS_RADSTABILITY_WO_PARAMS wo_params;   /* Work order parameter data
                                               structure */
    RPS_RADSTABILITY_OMF_PARAMS omf_params; /* OMF parameter data structure */

    /* Report header information */
    RPS_RADSTABILITY_REPORT_HEADER header_info;

    /* Satellite ID */
    IAS_SATELLITE_ID satellite_id;    /* Satellite ID used to
                                         initialize the satellite
                                         attributes library */

    /* Initialize logging for this application */
    ias_log_initialize("char_rad_stability");

    /* Only one command line argument:  the WO parameter file */
    if (argc != 2)
    {
        IAS_LOG_ERROR("USAGE:  char_rad_stability <name of parameter file>");
        return EXIT_FAILURE;
    }

    /* Read work order parameters */
    IAS_LOG_DEBUG("Reading required ODL and OMF parameters...");

    memset(&wo_params, 0, sizeof(RPS_RADSTABILITY_WO_PARAMS));
    memset(&omf_params, 0, sizeof(RPS_RADSTABILITY_OMF_PARAMS));

    status = get_parameters(argv[1], &wo_params, &omf_params);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Parameter processing failed");
        return EXIT_FAILURE;
    }


    /* Initialize the satellite attributes library so that we're processing
       OLI/TIRS data */
    IAS_LOG_DEBUG("Initializing satellite/sensor attributes...");

    satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(
            omf_params.satellite_number);
    if (satellite_id == ERROR)
    {
        IAS_LOG_ERROR("Invalid local satellite number %d in OMF parameter "
            "file", omf_params.satellite_number);
        return EXIT_FAILURE;
    }

    status = ias_sat_attr_initialize(satellite_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attribute information");
        return EXIT_FAILURE;
    }

    /* Get the total number of "normal" (OLI) bands and SCAs */
    status = ias_sat_attr_get_sensor_band_numbers(IAS_OLI, IAS_NORMAL_BAND,
             0, band_list, IAS_MAX_NBANDS, &num_oli_bands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Cannot obtain proper number of OLI bands");
        return EXIT_FAILURE;
    }

    num_oli_scas = ias_sat_attr_get_sensor_sca_count(IAS_OLI);
    if (num_oli_scas == ERROR)
    {
        IAS_LOG_ERROR("Cannot obtain proper number of SCAs in OLI bands");
        return EXIT_FAILURE;
    }

    /* If we haven't specified a pixel mask array file in the OMF,
       log a warning letting the user know local pixel masks will be
       created */
    if (strcmp(omf_params.pmask_filename, "") == 0)
    {
        IAS_LOG_WARNING("Pixel mask array file not available, pixel "
            "masks will be created individually for each band/SCA");
    }
    else
    {
        /* Read in the pixel masks from the pixel mask file */
        pmasks = ias_pm_read_array_from_file(omf_params.pmask_filename,
                &num_masks_read);
        if (pmasks == NULL)
        {
            IAS_LOG_ERROR("Loading pixel mask array from pixel mask file %s",
                omf_params.pmask_filename); 
            return EXIT_FAILURE;
        }

        /* At a minimum, there should be num_oli_bands * num_oli_scas pixel
           masks available.  We're assuming the pixel mask array file
           was created either during RPS driver processing of the solar
           collect, or as a "custom" pixel mask; in either case, we're
           assuming all bands/SCAs have a mask in the array (whether masks
           are empty or not) */
        if (num_masks_read < (num_oli_bands * num_oli_scas))
        {
            IAS_LOG_ERROR("Insufficient number of pixel masks %d for all OLI "
                 "bands/SCAs, should be %d", num_masks_read,
                 (num_oli_bands * num_oli_scas));

            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
            return EXIT_FAILURE;
        }
    }

    /* Get the information for the report header.  We'll use some of
       this information to determine what kind of collect we'll be
       processing (OLI solar or OLI shutter).  If we're generating a
       summary report and/or dump files, we'll use the whole thing */
    memset(&header_info, 0, sizeof(RPS_RADSTABILITY_REPORT_HEADER));
    header_info.common_header = ias_misc_initialize_report_header(
            omf_params.l0r_filename, wo_params.work_order_id,
            "60s Radiometric Stability Char");
    if (!header_info.common_header)
    {
        IAS_LOG_ERROR("Retrieving IAS common report header information");

         /**************************************************************
          NOTE     NOTE      NOTE      NOTE     NOTE      NOTE     NOTE

          The following code to clean up the pixel mask array appears
          often enough in this module (including the 6 lines above)
          to justify implementing it as a separate routine, in order to
          a) simplify maintainability of the module; and b) decrease the
          size of the source file
         ***************************************************************/
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        return EXIT_FAILURE;
    }

    /* Update the CPF in the header to reflect what is in the OMF file. */
    ias_misc_set_header_cpf_name(header_info.common_header,
        omf_params.cpf_filename);

    /* Initialize the RPS-specific header information if possible */
    status = ias_misc_initialize_rps_report_header(
        omf_params.l0r_filename, &header_info.rps_header);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving RPS-specific report header information");
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        ias_misc_free_report_header(header_info.common_header);
        return EXIT_FAILURE;
    }

    /* Initialize the remaining header information that's independent
       of the collect type -- the stability_delta_t threshold specified
       in the work order, and the stability_period actually calculated
       for the collect.

       There are two possible estimates for the stability period.  The
       first is based simply on the elapsed time (in seconds) between the
       imaging start and stop times.  The second is based on the average
       OLI frame rate (frames per second) times the total number of frames
       in the interval.  The frame rate is estimated as the total
       number of frames divided by the difference in timestamps for the
       first and last frames.

       One would think that both estimates would come
       out the same, but it appears there are some slight diffrences.
       The algorithm developer has stated that the frame-rate based
       estimate should be used here rather than the elapsed time
       estimate. */
    header_info.stability_delta_t = wo_params.stability_delta_t;
    header_info.stability_period =
        (double)header_info.rps_header.num_oli_frames
        / header_info.rps_header.average_oli_frame_rate;

    /* If the stability period estimated here is less than the
       threshold value specified in the work order, quit. */
    if (header_info.stability_period < header_info.stability_delta_t)
    {
        IAS_LOG_ERROR("Scene is about %lf seconds long--should be at "
            "least %lf seconds", header_info.stability_delta_t,
            header_info.stability_delta_t);
        ias_misc_free_report_header(header_info.common_header);
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        return EXIT_FAILURE;
    }

    /* Read the CPF.  Use the full name + path read from the OMF file */
    IAS_LOG_DEBUG("Reading CPF...");
    cpf = ias_cpf_read(omf_params.cpf_filename);
    if (cpf == NULL)
    {
        IAS_LOG_ERROR("Reading CPF file %s", omf_params.cpf_filename);
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        ias_misc_free_report_header(header_info.common_header);
        return EXIT_FAILURE;
    }

    /* Make sure we're actually processing an OLI scene here */
    sensor_id = ias_misc_get_header_sensor_id(header_info.common_header);
    if (sensor_id == NULL)
    {
        IAS_LOG_ERROR("Sensor ID information not initialized when IAS "
            "report header was created");
        ias_cpf_free(cpf);
        ias_misc_free_report_header(header_info.common_header);
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        return EXIT_FAILURE;
    }
    if (strstr(sensor_id, "OLI") == NULL)
    {
        IAS_LOG_ERROR("Invalid sensor ID string '%s'--should only be 'OLI' "
            "for this characterization", sensor_id);
        ias_cpf_free(cpf);
        ias_misc_free_report_header(header_info.common_header);
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        return EXIT_FAILURE;
    }

    /* Get the OLI parameter group data.  We'll use this to make sure the
       integration time reported in the header is the nominal integration
       time */
    oli_params = ias_cpf_get_oli_parameters(cpf);
    if (oli_params == NULL)
    {
        IAS_LOG_ERROR("Obtaining OLI-specific parameters from CPF");
        ias_cpf_free(cpf);
        ias_misc_free_report_header(header_info.common_header);
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        return EXIT_FAILURE;
    }

    /* Integration times in the image header are given in integer
       microseconds.  The multiplication by 0.001 converts these
       to (floating point) milliseconds values, which are the
       values listed in the CPF */
    if (((0.001 * header_info.rps_header.ms_integration_time)
           != oli_params->nominal_integration_time_ms)
        || ((0.001 * header_info.rps_header.pan_integration_time)
               != oli_params->nominal_integration_time_pan))
    {
        IAS_LOG_ERROR("Radiometric stability chraracterization of any OLI "
            "solar or shutter collects should only be performed for the "
            "nominal integration time subset");
        ias_misc_free_report_header(header_info.common_header);
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        ias_cpf_free(cpf);
        return EXIT_FAILURE;
    }

    /* Get the DETECTOR_STATUS group -- it contains the out-of-spec
       detector information.  We'll need this information in addition
       to the inoperable detector information, which is either a) part of
       the information in the pixel masks written to file; or b) we're
       creating a pixel mask for an individual band/SCA because there
       is no pixel mask file to load the masks from.  The idea is to
       treat out-of-spec detectors as we would inoperable detectors--they
       won't be included in the SCA/band-level summary statistics 
       calculations */
    det_status = ias_cpf_get_det_status(cpf);
    if (det_status == NULL)
    {
        IAS_LOG_ERROR("Getting DETECTOR_STATUS group information");
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        ias_cpf_free(cpf);
        ias_misc_free_report_header(header_info.common_header);
        return EXIT_FAILURE;
    }

    /* Check the collect type information.  Do we have a valid collect
       type for this characterization?  If the string values for these
       change, make sure to update accordingly... */
    collection_type = ias_misc_get_header_collection_type(
        header_info.common_header);
    if (collection_type == NULL)
    {
        IAS_LOG_ERROR("No collection type information available");
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        ias_cpf_free(cpf);
        ias_misc_free_report_header(header_info.common_header);
        return EXIT_FAILURE;
    }

    /* Right now, just check for non-integration-sweep solar or shutter
       collects.  If necessary, update the check to include integration
       sweep products later */
    is_solar_collect =
        (strcmp(collection_type, IAS_RPS_OLI_SOLAR_COLLECT) == 0);
    is_shutter_collect =
        (strcmp(collection_type, IAS_RPS_OLI_SHUTTER_COLLECT) == 0);
    if (!(is_solar_collect || is_shutter_collect))
    {
        IAS_LOG_ERROR("Invalid collect type '%s' for OLI radiometric "
            "stability characterization.  Valid collect types are '%s' "
            "or '%s'", collection_type, IAS_RPS_OLI_SOLAR_COLLECT,
            IAS_RPS_OLI_SHUTTER_COLLECT);
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        ias_cpf_free(cpf);
        ias_misc_free_report_header(header_info.common_header);
        return EXIT_FAILURE;
    }

    /* If this is a solar collect, identify the diffuser panel, and set
       up the gains as appropriate */
    if (is_solar_collect)
    {
        IAS_DATETIME oli_start_datetime;

        /* Get the information in the EARTH_CONSTANTS CPF group.  We'll
           need it for determining the diffuser panel deployed during the
           collect */
        const struct IAS_CPF_EARTH_CONSTANTS *earth_constants =
            ias_cpf_get_earth_const(cpf);
        if (earth_constants == NULL)
        {
            IAS_LOG_ERROR("Obtaining EARTH_CONSTANTS information from "
                "CPF");
            if (pmasks != NULL)
            {
                for (mask_index = 0; mask_index < num_masks_read;
                     mask_index++)
                {
                    pm_ptr = pmasks[mask_index];
                    ias_pm_destroy(pm_ptr);
                }
                free(pmasks);
            }
            ias_cpf_free(cpf);
            ias_misc_free_report_header(header_info.common_header);
            return EXIT_FAILURE;
        }

        /* Parse the OLI start date/time string */
        memset(&oli_start_datetime, 0, sizeof(IAS_DATETIME));
        status = ias_misc_parse_datetime_string(
            header_info.rps_header.acq_datetime.start_time_oli,
            IAS_DATETIME_L0R_FORMAT, &oli_start_datetime);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Returning from ias_misc_parse_datetime_string");
            ias_cpf_free(cpf);
            if (pmasks != NULL)
            {
                for (mask_index = 0; mask_index < num_masks_read; mask_index++)
                {
                    pm_ptr = pmasks[mask_index];
                    ias_pm_destroy(pm_ptr);
                }
                free(pmasks);
            }
            ias_misc_free_report_header(header_info.common_header);
            return EXIT_FAILURE;
        }

        /* Determine the diffuser panel ID given the date/time
           components of the imaging start date/time.  This will be
           done from a scan through the available OLI telemetry
           information */
        status = ias_rps_identify_diffuser_panel(omf_params.l0r_filename,
            earth_constants, oli_start_datetime,
            &header_info.diffuser_resolver_position,
            &header_info.panel_id);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Returning from ias_misc_identify_diffuser_panel");
            ias_cpf_free(cpf);
            if (pmasks != NULL)
            {
                for (mask_index = 0; mask_index < num_masks_read; mask_index++)
                {
                    pm_ptr = pmasks[mask_index];
                    ias_pm_destroy(pm_ptr);
                }
                free(pmasks);
            }
            ias_misc_free_report_header(header_info.common_header);
            return EXIT_FAILURE;
        }

        /* Inform the user which panel was deployed for the collect */
        if (header_info.panel_id == IAS_RPS_OLI_DIFFUSER_TYPE_PRISTINE)
        {
            IAS_LOG_INFO("Pristine diffuser panel was deployed for this "
                "collect");
        }
        else
        {
            IAS_LOG_INFO("Working diffuser panel was deployed for this "
                "collect");
        }

        /* Keep track of the relative gain set used for gain
           stability characterization */
        header_info.relgain_option = wo_params.relgain_option;

        /* If we're calculating radiances from bias-corrected/linearized
           histogram statistics, get the CPF gains to be used */
        if (wo_params.calculate_radiance
            == RPS_RADSTABILITY_CALCULATE_RADIANCE)
        {
            /* Set the RPS processing level for bias-corrected / linearized
               statistics */
            strncpy(header_info.proc_level, wo_params.rps_solar_proc_level_dn,
                sizeof(header_info.proc_level));
            header_info.proc_level[sizeof(header_info.proc_level) - 1] = '\0';

            /* Get the absolute gains */
            absolute_gains = ias_cpf_get_abs_gains(cpf);

            /* Get the correct set of relative gains specified in the
               work order */
            switch (wo_params.relgain_option)
            {
                case RPS_RADSTABILITY_PRELAUNCH_RELGAINS:
                    relative_gains =
                        ias_cpf_get_prelaunch_relative_gains(cpf);
                    break;
                case RPS_RADSTABILITY_POSTLAUNCH_RELGAINS:
                    relative_gains =
                        ias_cpf_get_postlaunch_relative_gains(cpf);
                    break;
                case RPS_RADSTABILITY_CURRENT_RELGAINS:
                    relative_gains = ias_cpf_get_relative_gains(cpf);
                    break;
                default:
                    IAS_LOG_ERROR("Invalid relative gain option "
                                  "specification %d",
                                  wo_params.relgain_option);
                    ias_cpf_free(cpf);
                    if (pmasks != NULL)
                    {
                        for (mask_index = 0; mask_index < num_masks_read;
                             mask_index++)
                        {
                            pm_ptr = pmasks[mask_index];
                            ias_pm_destroy(pm_ptr);
                        }
                        free(pmasks);
                    }
                    ias_misc_free_report_header(header_info.common_header);
                    return EXIT_FAILURE;
            }

            /* Make sure we've got valid pointers to both these data sets.
               Quit if we don't */
            if ((absolute_gains == NULL) || (relative_gains == NULL))
            {
                IAS_LOG_ERROR("Retrieving CPF gain information");
                ias_cpf_free(cpf);
                if (pmasks != NULL)
                {
                    for (mask_index = 0; mask_index < num_masks_read;
                         mask_index++)
                    {
                        pm_ptr = pmasks[mask_index];
                        ias_pm_destroy(pm_ptr);
                    }
                    free(pmasks);
                }
                ias_misc_free_report_header(header_info.common_header);
                return EXIT_FAILURE;
            }
        }
        else
        {
            /* Set the processing level for post-gain application 
               histogram statistics */
            strncpy(header_info.proc_level, wo_params.rps_solar_proc_level_rad,
                sizeof(header_info.proc_level));
            header_info.proc_level[sizeof(header_info.proc_level) - 1] = '\0';
        }
    }
    else
    {
        /* Shutter collect -- the diffuser panel should not be deployed */
        header_info.panel_id = IAS_RPS_OLI_DIFFUSER_TYPE_NOTDEPLOYED;
        header_info.diffuser_resolver_position = 0;
        IAS_LOG_INFO("Shutter acquisition -- diffuser panel not deployed");

        /* Input histogram statistics are "raw" (i.e. directly from the
           L0R shutter product) */
        strncpy(header_info.proc_level, wo_params.rps_shutter_proc_level,
            sizeof(header_info.proc_level));
        header_info.proc_level[sizeof(header_info.proc_level) - 1] = '\0';
    }

    /* Establish connections to the characterization database (for
       getting histogram statistics data, at minimum)  */
    IAS_LOG_DEBUG("Establishing database connection...");

    open_db = connect_to_db();
    if (open_db == NULL)
    {
        IAS_LOG_ERROR("Establishing database connection");
        ias_cpf_free(cpf);
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        ias_misc_free_report_header(header_info.common_header);
        return EXIT_FAILURE;
    }

    /* Get the radiometric work order common ID number for the given
       work order ID  */
    status = ias_misc_get_char_id(open_db,
            "RADIOMETRIC_WORK_ORDER_COMMON",
            "radiometric_wo_common_id",
            wo_params.work_order_id, rad_wo_common_id);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Retrieving RADIOMETRIC_WORK_ORDER_COMMON "
                "ID for current work order %s",
                wo_params.work_order_id);

        disconnect_from_db(open_db, RPS_RADSTABILITY_SHUTDOWN_DB);
        ias_cpf_free(cpf);
        if (pmasks != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                pm_ptr = pmasks[mask_index];
                ias_pm_destroy(pm_ptr);
            }
            free(pmasks);
        }
        ias_misc_free_report_header(header_info.common_header);
        return EXIT_FAILURE; 
    }

    /* If the work order specifies storing characterization results in
       the database, prepare to do so */
    if (wo_params.store_char_in_db)
    {
        /* Initiate a transaction for this application */
        ias_db_start_transaction(open_db);

        /* If this work order was processed with results stored
           in the database, delete the previous set of results from the
           band/SCA and detector tables */
        IAS_LOG_DEBUG("Clearing information previously saved for this "
                "work order...");

        /* For solar collects, clear the gain stability tables */
        if (is_solar_collect)
        {
            int status_gain_sca = clear_radstability_table(open_db,
                rad_wo_common_id, "gain_stability_60s_sca");

            int status_gain_det = clear_radstability_table(open_db,
                rad_wo_common_id, "gain_stability_60s_detector");

            status = clear_radstability_table(open_db,
                rad_wo_common_id, "gain_stability_60s");
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Clearing gain_stability_60s table");
                disconnect_from_db(open_db, RPS_RADSTABILITY_ROLLBACK_DB);
                ias_cpf_free(cpf);
                if (pmasks != NULL)
                {
                    for (mask_index = 0; mask_index < num_masks_read;
                         mask_index++)
                    {
                        pm_ptr = pmasks[mask_index];
                        ias_pm_destroy(pm_ptr);
                    }
                    free(pmasks);
                }
                ias_misc_free_report_header(header_info.common_header);
                return EXIT_FAILURE;
            }

            if ((status_gain_sca != SUCCESS)
                    || (status_gain_det != SUCCESS))
            {
                IAS_LOG_ERROR("Clearing 'gain_stability_60s_sca' and/or "
                        "'gain_stability_60s_detector' tables");
                disconnect_from_db(open_db, RPS_RADSTABILITY_ROLLBACK_DB);
                ias_cpf_free(cpf);
                if (pmasks != NULL)
                {
                    for (mask_index = 0; mask_index < num_masks_read;
                         mask_index++)
                    {
                        pm_ptr = pmasks[mask_index];
                        ias_pm_destroy(pm_ptr);
                    }
                    free(pmasks);
                }
                ias_misc_free_report_header(header_info.common_header);
                return EXIT_FAILURE;
            }

            /* (Re)populate the main gain stability table */
            status = write_gain_stability_main_to_db(open_db,
                rad_wo_common_id, header_info.proc_level,
                wo_params.stability_delta_t,
                header_info.panel_id);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Writing gain_stability_60s table");
                disconnect_from_db(open_db, RPS_RADSTABILITY_ROLLBACK_DB);
                ias_cpf_free(cpf);
                if (pmasks != NULL)
                {
                    for (mask_index = 0; mask_index < num_masks_read;
                         mask_index++)
                    {
                        pm_ptr = pmasks[mask_index];
                        ias_pm_destroy(pm_ptr);
                    }
                    free(pmasks);
                }
                ias_misc_free_report_header(header_info.common_header);
                return EXIT_FAILURE;
            }
        }
        else
        /* For (long) shutter collects, clear the bias stability tables */
        {
            status = clear_radstability_table(open_db,
                rad_wo_common_id, "bias_stability_60s");
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Clearing bias_stability_60s table");
                disconnect_from_db(open_db, RPS_RADSTABILITY_ROLLBACK_DB);
                ias_cpf_free(cpf);
                if (pmasks != NULL)
                {
                    for (mask_index = 0; mask_index < num_masks_read;
                         mask_index++)
                    {
                        pm_ptr = pmasks[mask_index];
                        ias_pm_destroy(pm_ptr);
                    }
                    free(pmasks);
                }
                ias_misc_free_report_header(header_info.common_header);
                return EXIT_FAILURE;
            }

            status = clear_radstability_table(open_db,
                    rad_wo_common_id,
                    "bias_stability_60s_sca");
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Clearing bias_stability_60s_sca table");
                disconnect_from_db(open_db, RPS_RADSTABILITY_ROLLBACK_DB);
                ias_cpf_free(cpf);
                if (pmasks != NULL)
                {
                    for (mask_index = 0; mask_index < num_masks_read;
                         mask_index++)
                    {
                        pm_ptr = pmasks[mask_index];
                        ias_pm_destroy(pm_ptr);
                    }
                    free(pmasks);
                }
                ias_misc_free_report_header(header_info.common_header);
                return EXIT_FAILURE;
            }

            /* (Re)populate the main bias stability table */
            status = write_bias_stability_main_to_db(open_db,
                rad_wo_common_id, header_info.proc_level,
                wo_params.stability_delta_t);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Writing bias_stability_60s table");
                disconnect_from_db(open_db, RPS_RADSTABILITY_ROLLBACK_DB);
                ias_cpf_free(cpf);
                if (pmasks != NULL)
                {
                    for (mask_index = 0; mask_index < num_masks_read;
                         mask_index++)
                    {
                        pm_ptr = pmasks[mask_index];
                        ias_pm_destroy(pm_ptr);
                    }
                    free(pmasks);
                }
                ias_misc_free_report_header(header_info.common_header);
                return EXIT_FAILURE;
            }
        }
    }

    /* Depending on what the input collection type is, run either the
       gain stability characterization flow (solar collect) or the bias
       stability characterization flow (shutter collect) */
    if (is_solar_collect)
    {
        status = char_gain_stability(wo_params, omf_params.l0r_filename,
                num_oli_bands, num_oli_scas, pmasks, num_masks_read,
                det_status, absolute_gains, relative_gains, open_db,
                rad_wo_common_id, header_info);
    }
    else
    {
        status = char_bias_stability(wo_params, omf_params.l0r_filename,
                num_oli_bands, num_oli_scas, pmasks, num_masks_read,
                det_status, open_db, rad_wo_common_id, header_info);
    }

    /* If characterization results were stored in the database, either
       commit the new results if processing was successful, or roll back
       to the previous state if it wasn't */
    if (wo_params.store_char_in_db)
    {
        if (status == SUCCESS)
        {
            int commit_status = disconnect_from_db(open_db,
                    RPS_RADSTABILITY_COMMIT_DB);
            if (commit_status != SUCCESS)
            {
                 exit_status = EXIT_FAILURE;
            }
            else
            {
                 exit_status = EXIT_SUCCESS;
            }
        }
        else
        {
            /* Since we've already failed somewhere upstream due to
               an error condition, we don't really care what the
               return value of disconnect_from_db() is--we're still
               ultimately exiting the program in a failed "state".
               So restore the database to its pre-existing state */
            disconnect_from_db(open_db, RPS_RADSTABILITY_ROLLBACK_DB);
            exit_status = EXIT_FAILURE;
        }
    }
    else
    {
        /* In this case, we're only accessing the database for the
           histogram statistics input data.  All we need to do is just
           close the connection and free up the resources.  The return
           status of disconnect_from_db() will be SUCCESS. */
        disconnect_from_db(open_db, RPS_RADSTABILITY_SHUTDOWN_DB);
        exit_status = EXIT_SUCCESS;
    }

    /* Free up CPF resources */
    ias_cpf_free(cpf);

    /* Free up the pixel mask array if it's been read in from file */
    if (pmasks != NULL)
    {
        for (mask_index = 0; mask_index < num_masks_read; mask_index++)
        {
            pm_ptr = pmasks[mask_index];
            ias_pm_destroy(pm_ptr);
        }
        free(pmasks);
    }

    /* Free up the report header information */
    ias_misc_free_report_header(header_info.common_header);

    /* Log the final status message */
    if (exit_status == EXIT_SUCCESS)
    {
        IAS_LOG_INFO("Radiometric stability characterization "
            "successfully completed.");
    }
    else
    {
        IAS_LOG_ERROR("Radiometric stability characterization failed.");
    }

    /* Done */
    return exit_status;
}
