#ifndef CHAR_RAD_STABILITY_H
#define CHAR_RAD_STABILITY_H

/* Common C #includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

/* Common IAS system #includes */
#include "ias_const.h"
#include "ias_cpf.h"
#include "ias_db.h"
#include "ias_l0r.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_pixel_mask.h"
#include "ias_radiometry.h"
#include "ias_satellite_attributes.h"
#include "ias_types.h"

/* Default RPS processing levels for the OLI histogram statistics used
   in this characterization.  The currently recommended radiometric
   processing for solar diffuser-based statistics consists of temperature
   correction of bias, followed by bias correction and response
   linearization (in that order); if characterization based on radiance
   values derived from an L1R product is desired, gain application should
   be the final processing step.  The 'C' represents an adjustment of the
   "raw" DN values by taking the upper 12 bits of the original 14-bit
   L0R data and correcting for quantization error.

   No radiometric processing is currently recommended for shutter data.
   The "raw" histogram statistics values are derived during Ingest
   processing of a shutter collect.  */
#define RPS_RADSTABILITY_SOLAR_PROC_LEVEL_TEMPCORRECT_DN      "CTBL"
#define RPS_RADSTABILITY_SOLAR_PROC_LEVEL_TEMPCORRECT_RAD     "CTBLG"
#define RPS_RADSTABILITY_SHUTTER_PROC_LEVEL                   "0"

/* Stability characterization interval threshold */
#define RPS_RADSTABILITY_DELTA_T 60.0


/* disconnect_from_db() routine commit status inputs */
#define RPS_RADSTABILITY_COMMIT_DB     1
#define RPS_RADSTABILITY_SHUTDOWN_DB   0
#define RPS_RADSTABILITY_ROLLBACK_DB  -1

/* Radiance data source selection.  Changed in an attempt to improve clarity--
   the default position is to calculate the radiance directly */
#define RPS_RADSTABILITY_CALCULATE_RADIANCE 1
#define RPS_RADSTABILITY_USE_DB_RADIANCE    0

/* String array lengths */
#define PROC_LEVEL_LENGTH     6


/* Relative gain source selection values (in case we can calculate
   radiances through use of CPF gains)*/
typedef enum rps_radstability_relgain_option
{
   RPS_RADSTABILITY_INVALID_RELGAIN_OPTION = 0,
   RPS_RADSTABILITY_PRELAUNCH_RELGAINS,
   RPS_RADSTABILITY_POSTLAUNCH_RELGAINS,
   RPS_RADSTABILITY_CURRENT_RELGAINS
} RPS_RADSTABILITY_RELGAIN_OPTION;



/* This data structure of pointers contains the detector-level working data
   buffers.  The buffers in the array are dynamically sized by the number of
   detectors in the current band/SCA, and contain the following contents.
   The "1" data buffers are to be used for the PAN band lineset=0 or
   MS band lineset=2 data; the "2" buffers are to be used for the PAN band
   lineset=1 data.  In addition, the two_sigma_ and pcnt_change_ data
   buffers for the PAN band are intended to contain the overall
   results (obtained from averaging both sets of statistics) */
typedef struct detector_data_buffers
{
    double *max_dn;                    /* bias-corrected/linear (DN) maxima */
    double *min_dn;                    /* bias-corrected/linear (DN) minima */
    double *mean_dn;                   /* bias-corrected/linear (DN) means */
    double *stdev_dn;                  /* bias-corrected/linuar (DN) 
                                          std deviations */
    double *max_rad;                   /* gain applied (radiance) maxima */
    double *min_rad;                   /* gain applied (radiance) minima */
    double *mean_rad;                  /* gain applied (radiance) means */
    double *stdev_rad;                 /* gain applied (radiance) 
                                          std deviations */
    double *two_sigma_dn;              /* Two-sigma (DN) */
    double *two_sigma_rad;             /* Two-sigma (radiance) */
    double *pcnt_change_gain;          /* Percent change in gain */
    double *pcnt_change_rad;           /* Percent change in radiance */
} DETECTOR_DATA_BUFFERS;


    /* Data structures used to organize summary report information */

/* 1.  SCA summary statistics (bias, diffuser, or radiance) */
typedef struct rps_radstability_dnrad_sca_summary_stats
{
    int sca_number;                /* Current 1-based SCA number */
    double sca_max_dn;             /* Max DN value in the SCA */
    double sca_min_dn;             /* Min DN value in the SCA */
    double sca_mean_dn;            /* Mean DN value in the SCA */
    double sca_stdev_dn;           /* Std dev. of mean DN values in the
                                      SCA */
    double sca_max_rad;            /* Max radiance value in the SCA */
    double sca_min_rad;            /* Min radiance value in the SCA */
    double sca_mean_rad;           /* Mean radiance value in the SCA */
    double sca_stdev_rad;          /* Std dev. of mean radiance values
                                      in the SCA */
} RPS_RADSTABILITY_DNRAD_SCA_SUMMARY_STATS;


/* 2.  SCA two-sigma and percent change in signal level statistics */
typedef struct rps_radstability_dnrad_sca_summary_siglevel_change
{
    int sca_number;
    double sca_two_sigma_dn;       /* Mean two-sigma DN value in the SCA */
    double sca_two_sigma_rad;      /* Mean two-sigma rad value in the SCA */
    double sca_pcnt_change_gain;   /* Mean percent gain change in the SCA */
    double sca_pcnt_change_rad;    /* Mean percent radiance change in the
                                      SCA */
} RPS_RADSTABILITY_DNRAD_SCA_SUMMARY_SIGLEVEL_CHANGE;


typedef struct rps_radstability_band_summary
{
    int band_number;
    char band_name[IAS_BAND_NAME_SIZE];
    int num_scas;
    RPS_RADSTABILITY_DNRAD_SCA_SUMMARY_STATS *summary_stats;
    RPS_RADSTABILITY_DNRAD_SCA_SUMMARY_SIGLEVEL_CHANGE *siglevel_change;
} RPS_RADSTABILITY_DNRAD_BAND_SUMMARY;


typedef struct rps_radstability_report_header
{
    IAS_REPORT_HEADER *common_header;             /* Common header
                                                     information */
    RPS_REPORT_HEADER rps_header;                 /* RPS-specific
                                                     header_information */
    char proc_level[IAS_RPS_PROCESSING_LEVEL_LENGTH + 1];
                                                  /* RPS processing level */
    int relgain_option;                           /* For internally
                                                     calculated radiances,
                                                     the set of relative
                                                     gains to use */
    double stability_delta_t;                     /* Acquisition duration
                                                     (in seconds) threshold
                                                     specified in work
                                                     order */
    double stability_period;                      /* Observed acquisition
                                                     duration (in seconds)
                                                     of the collect */
    int panel_id;                                 /* Diffuser panel type */
    uint16_t diffuser_resolver_position;          /* Position of diffuser
                                                     motor (currently
                                                     optional information
                                                     not reported in this
                                                     application) */
} RPS_RADSTABILITY_REPORT_HEADER;



/*           Data structures required for parameter I/O. */
typedef struct radstability_wo_params
{
    /* Work order-related information -- parameter file
       identification/location */
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char work_order_directory[PATH_MAX];

    /* Bias stability report filename prefix */
    char bias_report_filename_prefix[PATH_MAX];

    /* Gain stability report filename prefix */
    char gain_report_filename_prefix[PATH_MAX];

    /* RPS processing level information. */
    char rps_solar_proc_level_dn[IAS_RPS_PROCESSING_LEVEL_LENGTH + 1];
    char rps_solar_proc_level_rad[IAS_RPS_PROCESSING_LEVEL_LENGTH + 1];
    char rps_shutter_proc_level[IAS_RPS_PROCESSING_LEVEL_LENGTH + 1];

    /* Work order-related information -- processing options */
    int num_bands_to_process;
    int bands[IAS_MAX_NBANDS];      /* List of bands to process */
    int calculate_radiance;         /* Source for radiance statistics
                                       (1=direct calculation, 0=read from
                                       database) */
    int  relgain_option;            /* Selection of pre-launch, post-launch,
                                       or best current relative
                                       gains, respectively (applicable only
                                       when calculate_radiance=1) */
    double stability_delta_t;       /* Radiometric stability acquisition
                                       interval threshold (default is
                                       60.0 seconds) */

    /* Work order-related information -- output options */
    int  store_char_in_db;          /* 0=don't write to database, 1=write
                                       to database */
    int  report_flag;               /* 0=no summary report, 1=summary
                                       report */
    int  dump_flag;                 /* 0=no dump files, 1=dump files */
}   RPS_RADSTABILITY_WO_PARAMS;

typedef struct radstability_omf_params
{
    int satellite_number;                  /* Satellite number */
    char l0r_filename[PATH_MAX];           /* L0R file name */
    char cpf_filename[PATH_MAX];           /* CPF file name */
    char pmask_filename[PATH_MAX];         /* Pixel mask file name */
}   RPS_RADSTABILITY_OMF_PARAMS;



/*************************************************************************
                   FUNCTION PROTOTYPES
**************************************************************************/

                   /* Parameter I/O */
int get_parameters
(
    const char *wo_params_filename,              /* I: Work order filename */
    RPS_RADSTABILITY_WO_PARAMS  *wo_param_data,  /* O: Work order parameters */
    RPS_RADSTABILITY_OMF_PARAMS *omf_param_data  /* O: OMF parameters */
);


                  /* "Sub-driver" (depends on type of collect
                     being processed) */
int char_gain_stability
(
    const RPS_RADSTABILITY_WO_PARAMS wo_params,   /* I: Work order
                                                     parameters */
    const char *l0r_filename,                     /* I: Name of L0R data
                                                     file (includes path) */
    int num_oli_bands,                            /* I: Total number of
                                                     "normal" OLI imaging
                                                     bands */
    int num_oli_scas,                             /* I: Number of SCAs in
                                                     an OLI band */
    IAS_PIXEL_MASK **pmasks,                      /* I/O: Pixel mask array;
                                                     can be NULL if there
                                                     is no pixel mask
                                                     file available--a
                                                     temporary pixel mask
                                                     will be created for
                                                     each band/SCA */
    int num_masks_read,                           /* I: Number of pixel
                                                     masks read from file */
    const struct IAS_CPF_DETECTOR_STATUS *det_status, /* I: CPF DETECTOR_STATUS
                                                     group information */
    const struct IAS_CPF_ABSOLUTE_GAINS *abs_gains,  /* I: CPF absolute gains */
    const struct IAS_CPF_RELATIVE_GAINS *relative_gains,  /* I: CPF relative
                                                     gains */
    struct ias_db_connection *db,                 /* I: Characterization
                                                     database connection */
    char *rad_wo_common_id,                       /* I: Database ID
                                                     corresponding to work
                                                     order ID */
    RPS_RADSTABILITY_REPORT_HEADER header_info    /* I: RPS report header
                                                     information */
);


int char_bias_stability
(
    const RPS_RADSTABILITY_WO_PARAMS  wo_params,  /* I: Work order
                                                     parameters */
    const char *l0r_filename,                     /* I: Name of L0R data
                                                     file (includes path) */
    int num_oli_bands,                            /* I: Total number of
                                                     "normal" OLI imaging
                                                     bands */
    int num_oli_scas,                             /* I: Number of SCAs in
                                                     an OLI band */
    IAS_PIXEL_MASK **pmasks,                      /* I/O: Pixel mask array;
                                                     can be NULL if there
                                                     is no pixel mask
                                                     file available--a
                                                     temporary pixel mask
                                                     will be created for
                                                     each band/SCA */
    int num_masks_read,                           /* I: Number of pixel
                                                     masks read from file */
    const struct IAS_CPF_DETECTOR_STATUS *det_status,/* I: CPF DETECTOR_STATUS
                                                     group information */
    struct ias_db_connection *db,                 /* I: Characterization
                                                     database connection */
    char *rad_wo_common_id,                       /* I: Database ID
                                                     corresponding to work
                                                     order ID */
    RPS_RADSTABILITY_REPORT_HEADER header_info    /* I: RPS report header
                                                     information */
);



                   /* Stability assessment */
int assess_gain_stability
(
    const double *hist_stats_means,   /* I: Bias-corrected/linearized
                                         histogram statistics or radiances
                                         for all detectors in the band */
    const double *two_sigma_detector, /* I: Two-sigma values for each
                                         detector in the current band/SCA */
    const double two_sigma_sca_mean,  /* I: Mean two-sigma value in the
                                         current band/SCA */
    const double *relative_gains,     /* I: Relative gains data
                                         buffer for all detectors in the
                                         current band/SCA */
    const int *statistics_validity_flags,
                                      /* I: Data validity flags for all
                                         detectors in the current band/SCA */
    int num_detectors_sca,            /* I: Number of detectors in an SCA */
    double *pcnt_change_det,          /* O: Percent change in gain(radiance)
                                         for all detectors in the band */
    double *pcnt_change_sca_mean      /* O: Mean percent change in
                                         gain(radiance) for the current
                                         band/SCA */
);


int assess_signal_stability
(
    const double *hist_stats_stdev, /* I: Histogram statistics stdevs for
                                       all detectors in the current
                                       band/SCA */
    const int *statistics_validity_flags,
                                    /* I: Data validity flags for all
                                       detectors in the current band/SCA
                                       (0=valid, nonzero=invalid) */
    int num_dets_sca,               /* I: Number of detectors in each SCA */
    double *two_sigma_det,          /* O: Two-sigma values for all detectors
                                       in the current SCA */
    double *two_sigma_sca_mean      /* O: Mean two-sigma value for the
                                       current band/SCA */
);



                     /* "Utility" routines */
RPS_RADSTABILITY_DNRAD_BAND_SUMMARY *allocate_band_summary_array
(
    int num_oli_bands,                       /* I: Total number of
                                                "normal" OLI imaging
                                                bands */
    int num_oli_scas                         /* I: Number of SCAs in
                                                an OLI band */
);


void free_band_summary_array
(
    RPS_RADSTABILITY_DNRAD_BAND_SUMMARY
        *band_summary,                       /* I: Band summary data
                                                structure to free */
    int num_oli_bands                        /* I: Number of "normal" OLI
                                                imaging bands */
);


int allocate_detector_data_buffers
(
    DETECTOR_DATA_BUFFERS *data_buffers,    /* I/O:  Detector data buffer
                                               structure */
    const char *stability_id,               /* I: Stability characterization
                                               ("BIAS", "GAIN") */
    IAS_SPECTRAL_TYPE spectral_type,        /* I: Spectral type of current
                                               band */
    int num_detectors_sca                   /* I: Number of detectors in an
                                               SCA */
);


void free_detector_data_buffers
(
    DETECTOR_DATA_BUFFERS *data_buffers    /* I/O:  Detector data buffer
                                              data structure */
);


void rad_stability_cleanup
(
    DETECTOR_DATA_BUFFERS *data_buffers,    /* I/O: Detector data buffer
                                               data structure */
    int *statistics_validity_flags,
                                            /* I: Data validity flags for
                                               all detectors in the current
                                               band/SCA */
    RPS_RADSTABILITY_DNRAD_BAND_SUMMARY *band_summary, 
                                            /* I: Data structure with
                                               band-level stability
                                               summary results */

    int num_oli_bands                       /* I: Total number of "normal"
                                               OLI imaging bands */
);


int generate_summary_statistics
(
    const double *det_stats_max_buffer,   /* I: Detector-level histogram
                                             statistics maxima for the
                                             current band/SCA */
    const double *det_stats_min_buffer,   /* I: Detector-level histogram
                                             statistics minima for the
                                             current band/SCA */
    const double *det_stats_mean_buffer,  /* I: Detector-level histogram
                                             statistics means for
                                             the current band/SCA */
    const int *statistics_validity_flags,
                                          /* I: Statistics validity flags
                                             for all detectors in the
                                             current band/SCA (0=valid data,
                                             nonzero=invalid data) */
    int num_detectors_sca,                /* I: Number of detectors in
                                             an SCA */
    const char *buffer_set,              /* I: String indicating which
                                            set of summary statistics data
                                            buffers to populate
                                            ("DN", "RADIANCE") */
    RPS_RADSTABILITY_DNRAD_SCA_SUMMARY_STATS *stats
                                          /* O: Element of SCA summary
                                             histogram statistics array
                                             populated for the current SCA */
);


                     /* Database access */
struct ias_db_connection *connect_to_db();


int disconnect_from_db
(
    struct ias_db_connection *db,  /* I: Open database connection */
    int transaction_flag           /* I: Flag indicating whether
                                      open transactions should be
                                      committed (successful processing)
                                      rolled back (error) */
);


int clear_radstability_table
(
    struct ias_db_connection *db,  /* I: Open database connection */
    char *rad_wo_common_id,        /* I: Radiometric work order common ID */
    const char *table_name         /* I: Name of radiometric stability table
                                      to clear */
);


int get_histogram_statistics
(
    struct ias_db_connection *db,     /* I: Open database connection */
    char *rad_wo_common_id,           /* I: Database ID corresponding to
                                         work order ID */
    IAS_SPECTRAL_TYPE spectral_type,  /* I: Spectral type for current
                                         band */
    int band_number,                  /* I: Current 1-based band number */
    int sca_number,                   /* I: Current 1-based SCA number */
    int num_detectors_sca,            /* I: Number of detectors for the
                                         current band/SCA */
    const char *proc_level,           /* I: RPS processing level string */
    double *det_stats_max_buffer,     /* O: Histogram statistics maxima
                                         for all detectors in the current
                                         band/SCA */
    double *det_stats_min_buffer,     /* O: Histogram statistics minima
                                         for all detectors in the current
                                         band/SCA */
    double *det_stats_mean_buffer,    /* O: Histogram statistics means for
                                         all detectors in the current
                                         band/SCA */
    double *det_stats_stdev_buffer    /* O: Histogram statistics std devs
                                         for all detectors in the current
                                         band/SCA */
);


int write_sca_bias_stability_results_to_db
(
    struct ias_db_connection *db,          /* I: Open DB connection */
    char *rad_wo_common_id,                /* I: Database ID corresponding
                                              to the current work order */
    RPS_RADSTABILITY_DNRAD_SCA_SUMMARY_SIGLEVEL_CHANGE
        siglevel_change,                   /* I: Signal level change data
                                              array for the current SCA */
    int band_number,                       /* I: Current (1-based) band 
                                              number */
    int num_records                        /* I: Number of records to
                                              write to the database */
);


int write_sca_gainrad_stability_results_to_db
(
    struct ias_db_connection *db,          /* I: Open DB connection */
    char *rad_wo_common_id,                /* I: Database ID corresponding
                                              to the current work order */
    RPS_RADSTABILITY_DNRAD_SCA_SUMMARY_SIGLEVEL_CHANGE
        siglevel_change,                   /* I: Signal level change data
                                              array for the current SCA */
    int band_number,                       /* I: Current (1-based) band
                                              number */
    int num_records                        /* I: Number of records to
                                              write to the database */
);


int write_detector_gain_stability_results_to_db
(
    struct ias_db_connection *db,          /* I: Open database
                                              connection */
    char *rad_wo_common_id,                /* I: Database ID corresponding
                                              to processing work order ID */
    int *detector_ids,                     /* I: Detector IDs for the
                                              current band/SCA (1-based) */
    IAS_DB_NULL_TYPE *flagged_detector_ids,/* I: SCA detector indices
                                             flagged with NULLS */
    double *pcnt_change_gain_det_buffer,   /* I: Percent change in gains for
                                              each detector */
    int band_number,                       /* I: Current (1-based) band #*/
    int sca_number,                        /* I: Current (1-based) SCA # */
    int detectors_per_sca                  /* I: Number of detectors in an
                                              SCA */
);


int write_bias_stability_main_to_db
(
    struct ias_db_connection *db,       /* I: Open DB connection */
    char *rad_wo_common_id,             /* I: Radiometric work order common
                                           ID */
    char *proc_level,                   /* I: Highest RPS processing level
                                           for histogram statistics data
                                           source */
    double stability_delta_t            /* I: Work order-specified image
                                           acquisition interval threshold
                                           (seconds ) */
);


int write_gain_stability_main_to_db
(
    struct ias_db_connection *db,       /* I: Open DB connection */
    char *rad_wo_common_id,             /* I: Radiometric work order common
                                           ID */
    char *proc_level,                   /* I: Highest RPS processing level
                                           for histogram statistics data
                                           source */
    double stability_delta_t,           /* I: Work order-specified image
                                           acquisition interval threshold
                                           (seconds ) */
    int panel_id                        /* I: Diffuser panel ID */
);



                /* Summary report and/or dump file */
int write_header_to_summary_report
(
    RPS_RADSTABILITY_REPORT_HEADER report_header,   /* I: Report header
                                                       information to
                                                       write */
    FILE *rpt_fptr                                  /* I: Summary report
                                                       file */
);

int write_data_to_summary_report
(
    const RPS_RADSTABILITY_DNRAD_BAND_SUMMARY
        *band_summary,                       /* I: Summary stability
                                 x               information for all bands */
    const char *stability_id,                /* I: Flag indicating whether
                                                bias stability or gain
                                                stability characterization
                                                has been performed */
    int num_oli_bands,                       /* I: Total number of (OLI)
                                                bands to write to
                                                the report */
    double delta_t,                          /* I: Total elapsed time of
                                                collect in seconds
                                                (contained in RPS report
                                                header information) */
    FILE *rpt_fptr                           /* I: Open summary report
                                                file */
);


int write_header_to_dumpfile
(
    RPS_RADSTABILITY_REPORT_HEADER header_info,/* I: RPS report header
                                                  information */
    FILE *dump_fptr                            /* I: Dump file handle */
);


int write_detector_stability_results_to_dump_file
(
    DETECTOR_DATA_BUFFERS data_buffers,/* I: Data structure containing
                                          detector-level results */
    const char *stability_id,          /* I: Flag indicating whether
                                          bias stability characterization
                                          or gain stability
                                          characterization was performed */
    int band_number,                   /* I: 1-based band number */
    const char *band_name,             /* I: Band name/description*/
    int sca_number,                    /* I: 1-based SCA number */
    int num_detectors_sca,             /* I: # of detectors in an SCA */
    FILE *dump_fptr                    /* O: Open dump file for the current
                                          band */
);


int write_sca_stability_results_to_dump_file
(
    const RPS_RADSTABILITY_DNRAD_BAND_SUMMARY *band_summary,
                                            /* I: Stability results for
                                               the current band */
    const char *stability_id,               /* I: String indicating whether
                                               bias stability or gain
                                               stability was
                                               characterized */
    FILE *dump_fptr                         /* O: Open dump data file */
);




#endif
