/**************************************************************************
 NAME:              char_rad_stability_db

 PURPOSE:  Module implementing routines to provide IAS characterization
           database access for the char_rad_stability algorithm.

****************************************************************************/
#include <float.h>

#include "ias_db.h"
#include "ias_db_get_connect_info.h"
#include "ias_db_insert.h"
#include "ias_db_query.h"
#include "ias_radiometry.h"

#include "char_rad_stability.h"

#define SQL_QUERYLEN 512

/*--------------------------------------------------------------------
 NAME:                  connect_to_db

 PURPOSE:    Retrieves IAS database connection information and attempts
             the connection

 RETURNS:    - Pointer to valid ias_db_struct, indicating connection has
               been established
             - NULL pointer if connection cannot be made
-----------------------------------------------------------------------*/
struct ias_db_connection *connect_to_db()
{
    char *db_name = NULL;            /* Required connection information */
    char *user_id = NULL;
    char *user_password = NULL;
    char *db_host = NULL;

    struct ias_db_connection *db = NULL;

    int status;


    /* Initialize the database library */
    IAS_LOG_DEBUG("Initializing database library...");
    status = ias_db_initialize_database_lib();
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing database library");
        return NULL;
    }

    /* Get the connection information */
    IAS_LOG_DEBUG("Obtaining database connection information...");
    status
        = ias_db_get_connect_info(&db_name, &user_id, &user_password,
                &db_host);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Obtaining database connection information");
        return NULL;
    }

    /* Try to connect */
    IAS_LOG_DEBUG("Connecting to database...");

    db = ias_db_connect_to_database(db_name, user_id, user_password,
               db_host);

    /* Free the allocated strings with the connection information */
    free(db_name);
    free(user_id);
    free(user_password);
    free(db_host);

    /* Did we connect? */
    if (db == NULL)             /* NO -- log an error */
    {
        IAS_LOG_ERROR("Connecting to database");
    }
    else                        /* YES -- log success only in DEBUG mode */
    {
        IAS_LOG_DEBUG("Database connection established");
    }

    /* Return whatever we've got for *db (either NULL or a valid pointer) */
    return db;
}




/*------------------------------------------------------------------------
 NAME:                 disconnect_from_db

 PURPOSE:  Closes an active database connection and frees up resources
           used by the database library.  If writing results to the
           database is specified, also commits/rolls back the transaction
           accordingly

 RETURNS:  Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int disconnect_from_db
(
    struct ias_db_connection *db,       /* I: Open database connection */
    int transaction_flag                /* I: Flag indicating whether
                                           open transactions should be
                                           committed (successful processing)
                                           or rolled back (error) */
)
{
    int status = SUCCESS;               /* Overall return status */
    int transaction_status;             /* Database transaction status */


    switch (transaction_flag)
    {
        case RPS_RADSTABILITY_COMMIT_DB:   /* Successful processing */
            transaction_status = ias_db_commit_transaction(db);
            if (transaction_status != SUCCESS)
            {
                IAS_LOG_ERROR("Database not updated");
            }
            break;
        case RPS_RADSTABILITY_ROLLBACK_DB: /* Processing error */
            transaction_status = ias_db_rollback_transaction(db);
            if (transaction_status != SUCCESS)
            {
                IAS_LOG_ERROR("Database not rolled back");
            }
            break;
        case RPS_RADSTABILITY_SHUTDOWN_DB: /* Before transactions initiated */
            IAS_LOG_DEBUG("Transactions not enabled");
            transaction_status = SUCCESS;
            break;
        default:
            IAS_LOG_ERROR("Invalid transaction identifier %d",
                          transaction_flag);
            transaction_status = ERROR;
            break;
    }

    /* Set the overall status to ERROR if we couldn't process the
       transaction properly, for whatever reason */
    if (transaction_status != SUCCESS)
    {
        IAS_LOG_ERROR("Processing database transactions");
        status = ERROR;
    }

    /* Close the connection and free up the resources used by the
       connection */
    ias_db_close_connection(db);
    ias_db_close_database_lib();

    /* Done */
    return status;
}




/*------------------------------------------------------------------------
 NAME:                     clear_radstability_table

 PURPOSE:       Clears the specified radiometric stability table for the
                specified radiometric work order ID key

 RETURNS:       Integer status code of SUCCESS (table successfully
                cleared) or ERROR (table couldn't be cleared)
--------------------------------------------------------------------------*/
int clear_radstability_table
(
    struct ias_db_connection *db, /* I: Open database connection */
    char *rad_wo_common_id,       /* I: Radiometric work order common ID */
    const char *table_name        /* I: Name of radiometric stability table
                                     to clear */
)
{
    int status;                   /* Function return status code */

    /* 1.  Clear out the specified table */
    IAS_LOG_DEBUG("Clearing %s table...", table_name);

    status = ias_db_delete_records(db, table_name,
                             "radiometric_wo_common_id",
                             rad_wo_common_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Clearing %s table", table_name);
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}




/*-------------------------------------------------------------------------
 NAME:           get_histogram_statistics

 PURPOSE:     Queries the HISTOGRAM_STATISTICS_CAL table for the
              mean values extracted from the shutter data or diffuser data
              product for the desired work order (indirectly
              represented by the specified RADIOMETRIC_WORK_ORDER_COMMON
              primary key value), and determines the summary
              histogram statistics for the current band/SCA

 RETURNS:     Integer status code of SUCCESS or ERROR
---------------------------------------------------------------------------*/
int get_histogram_statistics
(
    struct ias_db_connection *db,     /* I: Open database connection */
    char *rad_wo_common_id,           /* I: Database ID corresponding to
                                         work order ID */
    IAS_SPECTRAL_TYPE spectral_type,  /* I: Spectral type for current
                                         band */
    int band_number,                  /* I: Current 1-based band number */
    int sca_number,                   /* I: Current 1-based SCA number */
    int num_detectors_sca,            /* I: Number of detectors for the
                                         current band/SCA */
    const char *proc_level,           /* I: RPS processing level string */
                                      /* O: Populated detector data
                                         buffers */
    double *det_stats_max_buffer,     /* O: Histogram statistics maxima
                                         for all detectors in the current
                                         band/SCA */
    double *det_stats_min_buffer,     /* O: Histogram statistics minima
                                         for all detectors in the current
                                         band/SCA */
    double *det_stats_mean_buffer,    /* O: Histogram statistics means for
                                         all detectors in the current
                                         band/SCA */
    double *det_stats_stdev_buffer    /* O: Histogram statistics std devs
                                         for all detectors in the current
                                         band/SCA */
)
{
    char query_string[SQL_QUERYLEN];           /* SQL query string */
    char db_error_message[IAS_SQL_RTN_MSGLEN]; /* Query error message */
    struct ias_db_query *query = NULL;    /* Query data */
    int ii;                               /* Detector loop counter */
    int status;                           /* Function return status */


    /* Prepare the query string.  It MUST include the identifier for the
       processing level, either '0' (raw shutter values), 'CTBL' (indicating
       bias-corrected/linearized diffuser DN values) or 'CTBLG' (indicating
       diffuser radiance values).  It is assumed here that histogram
       statistics values are calculated for every detector, whether the
       data are valid or not.  Use snprintf to write the SQL string so
       we can check for buffer overrun and deal with it appropriately

       For the PAN band, write the query so that it returns the average
       scene_pixel_mean for the lineset 0 AND lineset 1 data.  For the MS
       bands, set lineset to 2 */
    if (spectral_type == IAS_SPECTRAL_PAN)
    {
        status = snprintf(query_string, sizeof(query_string),
            "SELECT detector_id, MAX(scene_pixel_max), "
            "MIN(scene_pixel_min), AVG(scene_pixel_mean), "
            "AVG(scene_pixel_stdev) FROM HISTOGRAM_STATISTICS_CAL WHERE "
            "band_id=%d AND sensor_chip_assembly_id=%d AND "
            "radiometric_wo_common_id=%s AND processing_level='%s' "
            "AND lineset IN (0, 1) GROUP BY detector_id ORDER BY "
            "detector_id;", band_number, sca_number, rad_wo_common_id,
            proc_level);
    }
    else
    {
        status = snprintf(query_string, sizeof(query_string),
        "SELECT detector_id, scene_pixel_max, "
        "scene_pixel_min, scene_pixel_mean, scene_pixel_stdev FROM "
        "HISTOGRAM_STATISTICS_CAL WHERE band_id=%d AND "
        "sensor_chip_assembly_id=%d AND "
        "radiometric_wo_common_id=%s AND processing_level='%s' "
        "AND lineset=2 ORDER BY detector_id;", band_number, sca_number,
        rad_wo_common_id, proc_level);
    }
    if ((status < 0) || (status >= sizeof(query_string)))
    {
        IAS_LOG_ERROR("Constructing SQL query for band %d SCA %d",
            band_number, sca_number);
        return ERROR;
    }

    /* Execute the query and verify it worked */
    query = ias_db_perform_query(db, query_string);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
                sizeof(db_error_message));
        IAS_LOG_ERROR("Query of HISTOGRAM STATISTICS CAL failed:  %s",
                db_error_message);
        return ERROR;
    }

    /* Get the max, min, mean and std deviation values for the
       current detector */
    for (ii = 0; ii < num_detectors_sca; ii++)
    {
        double query_max;                 /* Returned query max value */
        double query_min;                 /* Returned query min value */
        double query_mean;                /* Returned query mean value */
        double query_std;                 /* Returned query stdev value */
        int detector_number;              /* 1-based detector ID */
        int detector_index;               /* 0-based index corresponding to
                                             detector ID */

        /* Advance to the next valid record.  It's more efficient to do
           it at this point, inside the loop.  Doing it here should avoid
           an error due to advancing to a non-existent record, plus it
           should be easier to maintain */
        if (!ias_db_query_next(query))
        {
            ias_db_query_get_error_message(query, db_error_message,
                    sizeof(db_error_message));
            IAS_LOG_ERROR("Accessing HISTOGRAM_STATISTICS_CAL data:  %s",
                    db_error_message);
            ias_db_query_close(query);
            return ERROR;
        }

        status = ias_db_query_int_value(query, &detector_number, 0);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Cannot read histogram statistics "
                    "detector_id value at index %d in band %d SCA %d",
                    ii, band_number, sca_number);
            ias_db_query_close(query);
            return ERROR;
        }
        detector_index = (detector_number - 1);

        status = ias_db_query_double_value(query, &query_max, 1);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Cannot obtain histogram statistics "
                    "max value for detector %d in band %d SCA %d",
                    detector_number, band_number, sca_number);
            ias_db_query_close(query);
            return ERROR;
        }

        status = ias_db_query_double_value(query, &query_min, 2);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Cannot obtain histogram statistics "
                    "min value for detector %d in band %d SCA %d",
                    detector_number, band_number, sca_number);
            ias_db_query_close(query);
            return ERROR;
        }

        status = ias_db_query_double_value(query, &query_mean, 3);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Cannot obtain histogram statistics "
                    "mean value for detector %d in band %d SCA %d",
                    detector_number, band_number, sca_number);
            ias_db_query_close(query);
            return ERROR;
        }

        status = ias_db_query_double_value(query, &query_std, 4);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Cannot obtain histogram statistics "
                    "std dev value for detector %d in band %d SCA %d",
                    detector_number, band_number, sca_number);
            ias_db_query_close(query);
            return ERROR;
        }

        /* Copy the data */
        det_stats_max_buffer[detector_index]   = query_max;
        det_stats_min_buffer[detector_index]   = query_min;
        det_stats_mean_buffer[detector_index]  = query_mean;
        det_stats_stdev_buffer[detector_index] = query_std;
    }

    /* Release query resources */
    ias_db_query_close(query);

    /* Finished. */
    return SUCCESS;
}




/*-------------------------------------------------------------------------
 NAME:            write_sca_bias_stability_results_to_db

 PURPOSE:    Writes the 60s bias stability SCA mean results to the DB

 RETURNS:    Integer status code of SUCCESS or ERROR
---------------------------------------------------------------------------*/
int write_sca_bias_stability_results_to_db
(
    struct ias_db_connection *db,          /* I: Open DB connection */
    char *rad_wo_common_id,                /* I: Database ID corresponding
                                              to the current work order */
    RPS_RADSTABILITY_DNRAD_SCA_SUMMARY_SIGLEVEL_CHANGE
        siglevel_change,                   /* I: Signal level change data
                                              array for the current SCA */
    int band_number,                       /* I: Current (1-based) band
                                              number */
    int num_records                        /* I: Number of records to
                                              write to the database */
)
{
    int status;                            /* Function return status */


    /* Define the write table for the database */
    IAS_DB_INSERT_TABLE table[] =
    {
        IAS_DB_TABLE_STRING("radiometric_wo_common_id", rad_wo_common_id),
        IAS_DB_TABLE_INT("band_id", band_number),
        IAS_DB_TABLE_INT("sensor_chip_assembly_id",
            siglevel_change.sca_number),
        IAS_DB_TABLE_DOUBLE("change_in_signal_mean",
            siglevel_change.sca_two_sigma_dn)
    };

    int table_size = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);

    /* Write the data to the appropriate table */
    status = ias_db_insert_using_table(db, "bias_stability_60s_sca",
                                       table, table_size, num_records);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Write to bias_stability_60s_sca table failed in "
               "band %d, SCA %d", band_number,
               siglevel_change.sca_number);
    }

    /* Done */
    return status;
}




/*-------------------------------------------------------------------------
 NAME:            write_sca_gainrad_stability_results_to_db

 PURPOSE:    Writes the 60s gain/radiance stability SCA mean results to the
             DB

 RETURNS:    Integer status code of SUCCESS or ERROR
---------------------------------------------------------------------------*/
int write_sca_gainrad_stability_results_to_db
(
    struct ias_db_connection *db,          /* I: Open DB connection */
    char *rad_wo_common_id,                /* I: Database ID corresponding
                                              to the current work order */
    RPS_RADSTABILITY_DNRAD_SCA_SUMMARY_SIGLEVEL_CHANGE
        siglevel_change,                   /* I: Signal level change data
                                              array for the current SCA */
    int band_number,                       /* I: Current (1-based) band
                                              number */
    int num_records                        /* I: Number of records to
                                              write to the database */
)
{
    int status;


    /* Define the write table for the database */
    IAS_DB_INSERT_TABLE table[] =
    {
        IAS_DB_TABLE_STRING("radiometric_wo_common_id", rad_wo_common_id),
        IAS_DB_TABLE_INT("band_id", band_number),
        IAS_DB_TABLE_INT("sensor_chip_assembly_id",
                siglevel_change.sca_number),
        IAS_DB_TABLE_DOUBLE("two_sigma_change_dn_mean",
                siglevel_change.sca_two_sigma_dn),
        IAS_DB_TABLE_DOUBLE("two_sigma_change_rad_mean",
                siglevel_change.sca_two_sigma_rad),
        IAS_DB_TABLE_DOUBLE("pcnt_change_gain_mean",
                siglevel_change.sca_pcnt_change_gain),
        IAS_DB_TABLE_DOUBLE("pcnt_change_rad_mean",
                siglevel_change.sca_pcnt_change_rad)
    };

    int table_size = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);

    /* Write the data to the appropriate table */
    status = ias_db_insert_using_table(db, "gain_stability_60s_sca",
                                       table, table_size, num_records);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Write to gain_stability_60s_sca table failed in "
               "band %d, SCA %d",
               band_number,
               siglevel_change.sca_number);
    }

    /* Done */
    return status;
}




/*-------------------------------------------------------------------------
 NAME:          write_detector_gain_stability_results_to_DB

 PURPOSE:    Writes detector percent change in gain results to the DB

 RETURNS:    Integer status code of SUCCESS or ERROR
---------------------------------------------------------------------------*/
int write_detector_gain_stability_results_to_db
(
    struct ias_db_connection *db,          /* I: Open database
                                              connection */
    char *rad_wo_common_id,                /* I: Database ID corresponding
                                              to processing work order ID */
    int *detector_ids,                     /* I: Detector IDs for the
                                              current band/SCA (1-based) */
    IAS_DB_NULL_TYPE *flagged_detector_ids,/* I: SCA detector indices
                                              flagged with NULLS */
    double *pcnt_change_gain_det_buffer,   /* I: Percent change in gains for
                                              each detector */
    int band_number,                       /* I: Current (1-based) band #*/
    int sca_number,                        /* I: Current (1-based) SCA # */
    int num_detectors_sca                  /* I: Number of detectors in an
                                              SCA */
)
{
    int status;
    int table_size;


    /* Set up the table structure for writing the data */
    IAS_DB_INSERT_TABLE table[] =
    {
        IAS_DB_TABLE_STRING("radiometric_wo_common_id", rad_wo_common_id),
        IAS_DB_TABLE_INT("band_id", band_number),
        IAS_DB_TABLE_INT("sensor_chip_assembly_id", sca_number),
        IAS_DB_TABLE_INT_ARRAY("detector_id", detector_ids),
        IAS_DB_TABLE_DOUBLE_ARRAY_WITH_NULLS("change_in_gain_mean",
                                     pcnt_change_gain_det_buffer,
                                     flagged_detector_ids)
    };

    table_size = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);
    status = ias_db_insert_using_table(db, "gain_stability_60s_detector",
                                       table, table_size,
                                       num_detectors_sca);

    /* Done */
    return status;
}



/*-------------------------------------------------------------------------
 NAME:        write_bias_stability_main_to_db

 PURPOSE:  Populates the bias_stability_60s main table

 RETURNS:  Integer status code of SUCCESS or ERROR

---------------------------------------------------------------------------*/
int write_bias_stability_main_to_db
(
    struct ias_db_connection *db,       /* I: Open DB connection */
    char *rad_wo_common_id,             /* I: Radiometric work order common
                                           ID */
    char *proc_level,                   /* I: Highest RPS processing level
                                           for histogram statistics data
                                           source */
    double stability_delta_t            /* I: Work order-specified image
                                           acquisition interval threshold
                                           (seconds ) */
)
{
    int num_records = 1;
    int status;


    /* Define the write table for the database */
    IAS_DB_INSERT_TABLE table[] =
    {
        IAS_DB_TABLE_STRING("radiometric_wo_common_id", rad_wo_common_id),
        IAS_DB_TABLE_STRING("processing_level", proc_level),
        IAS_DB_TABLE_DOUBLE("stability_delta_t", stability_delta_t)
    };

    int table_size = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);

    /* Write the data to the appropriate table */
    status = ias_db_insert_using_table(db, "bias_stability_60s",
                                       table, table_size, num_records);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Write to bias_stability_60s table failed ");;
    }

    /* Done */
    return status;
}




/*-------------------------------------------------------------------------
 NAME:        write_gain_stability_main_to_db

 PURPOSE:  Populates the gain_stability_60s main table

 RETURNS:  Integer status code of SUCCESS or ERROR

---------------------------------------------------------------------------*/
int write_gain_stability_main_to_db
(
    struct ias_db_connection *db,       /* I: Open DB connection */
    char *rad_wo_common_id,             /* I: Radiometric work order common
                                           ID */
    char *proc_level,                   /* I: Highest RPS processing level
                                           for histogram statistics data
                                           source */
    double stability_delta_t,           /* I: Work order-specified image
                                           acquisition interval threshold
                                           (seconds ) */
    int panel_id                        /* I: Diffuser identification ID
                                           in report header information */
)
{
    char panel_type[IAS_DIFFUSER_TYPE_SIZE];
    int num_records = 1;
    int status;

    /* Get the appropriate string given the panel type ID */
    if (panel_id == IAS_RPS_OLI_DIFFUSER_TYPE_PRISTINE)
    {
        strcpy(panel_type, IAS_RPS_OLI_DIFFUSER_PRISTINE);
    }
    else
    {
        strcpy(panel_type, IAS_RPS_OLI_DIFFUSER_WORKING);
    }

    /* Define the write table for the database */
    IAS_DB_INSERT_TABLE table[] =
    {
        IAS_DB_TABLE_STRING("radiometric_wo_common_id", rad_wo_common_id),
        IAS_DB_TABLE_STRING("processing_level", proc_level),
        IAS_DB_TABLE_DOUBLE("stability_delta_t", stability_delta_t),
        IAS_DB_TABLE_STRING("diffuser_type", panel_type)
    };

    int table_size = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);

    /* Write the data to the appropriate table */
    status = ias_db_insert_using_table(db, "gain_stability_60s",
                                       table, table_size, num_records);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Write to gain_stability_60s table failed ");
    }

    /* Done */
    return status;
}
