/***************************************************************************
  NAME:                      get_parameters

  PURPOSE:  Reads the input parameter files and populates the
            required parameter data structures

  RETURNS:  Integer status code of SUCCESS or ERROR

  NOTES:    If parameter help is successfully provided, application
            will immediately exit

****************************************************************************/
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"

#include "char_rad_stability.h"

/* Current work order / OMF parameter table sizes */
#define MAX_NUM_WO_PARAMS   13
#define MAX_NUM_OMF_PARAMS  4

/* Work order option default values */
#define RPS_RADSTABILITY_DEFAULT_REPORT_FLAG            0
#define RPS_RADSTABILITY_DEFAULT_DUMP_FLAG              0


int get_parameters
(
    const char *wo_params_filename,              /* I: Work order filename */
    RPS_RADSTABILITY_WO_PARAMS  *wo_param_data,  /* O: Work order parameters */
    RPS_RADSTABILITY_OMF_PARAMS *omf_param_data  /* O: OMF parameters */
)
{
    /* Internal OMF parameter file name */
    char omf_params_filename[PATH_MAX];

    /* Solar and shutter RPS processing level strings. */
    const char *default_solar_rps_proc_level_dn[] =
        {RPS_RADSTABILITY_SOLAR_PROC_LEVEL_TEMPCORRECT_DN};
    const char *default_solar_rps_proc_level_rad[] =
        {RPS_RADSTABILITY_SOLAR_PROC_LEVEL_TEMPCORRECT_RAD};
    const char *default_shutter_rps_proc_level[] =
        {RPS_RADSTABILITY_SHUTTER_PROC_LEVEL};

    /* Default stability delta t interval */
    double default_stability_delta_t = RPS_RADSTABILITY_DELTA_T;

    /* Default radiance "source" is from calculation using bias-corrected/
       linearized histogram statistics */
    int default_calculate_radiance = RPS_RADSTABILITY_CALCULATE_RADIANCE;

    /* Default relative gain set is prelaunch (at least initially).
       Later, this should switch to current */
    int default_relgain_option = RPS_RADSTABILITY_PRELAUNCH_RELGAINS;

    /* Default settings here are no report and no dump files */
    int default_report_flag = RPS_RADSTABILITY_DEFAULT_REPORT_FLAG;
    int default_dump_flag = RPS_RADSTABILITY_DEFAULT_DUMP_FLAG;

    int status;                      /* Status for each read
                                        operation */
    int read_count;                  /* Count of parameters
                                        successfully read */

/*-------------------------------------------------------------------------
                   READ WORK ORDER PARAMETER FILE
---------------------------------------------------------------------------*/
    /* Declare the work order parameter table */
    IAS_PARM_DECLARE_TABLE(wo_param_table, MAX_NUM_WO_PARAMS);

    /* Do common parameters first */
    IAS_PARM_BAND_LIST(wo_param_table, IAS_PARM_REQUIRED,
            wo_param_data->bands,
            sizeof(wo_param_data->bands), 1);

    IAS_PARM_WORK_ORDER_ID(wo_param_table, wo_param_data->work_order_id,
            sizeof(wo_param_data->work_order_id), 1);

    IAS_PARM_STORE_CHARACTERIZATION_IN_DB(wo_param_table,
            IAS_PARM_OPTIONAL,
            &(wo_param_data->store_char_in_db),
            sizeof(wo_param_data->store_char_in_db), 1);

    IAS_PARM_OLI_RPS_PROC_LEVEL_DN(wo_param_table, IAS_PARM_OPTIONAL,
            default_solar_rps_proc_level_dn,
            wo_param_data->rps_solar_proc_level_dn,
            sizeof(wo_param_data->rps_solar_proc_level_dn));

    /* Now the more custom parameters */
    IAS_PARM_ADD_DOUBLE(wo_param_table, STABILITY_DELTA_T,
            "Collect duration threshold needed for radiometric stability "
            "characterization (seconds)", IAS_PARM_OPTIONAL, 
            IAS_PARM_NOT_ARRAY, 0, 0.0, 120.0, 1, 
            &default_stability_delta_t, &(wo_param_data->stability_delta_t),
            sizeof(wo_param_data->stability_delta_t), 1);

    IAS_PARM_ADD_BOOLEAN(wo_param_table, CALCULATE_RADIANCE,
        "Calculate radiance values internally 0=get L1R product radiances "
        "from database, 1=calculate radiances from "
        "bias-corrected/linearized L0R histogram statistics",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &default_calculate_radiance,
        &(wo_param_data->calculate_radiance),
        sizeof(wo_param_data->calculate_radiance), 1);

    IAS_PARM_ADD_INT(wo_param_table, RELGAIN_OPTIONS,
            "CPF relative gain set 1=prelaunch, 2=postlaunch, 3=current",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
            1, RPS_RADSTABILITY_PRELAUNCH_RELGAINS,
            RPS_RADSTABILITY_CURRENT_RELGAINS, 1,
            &default_relgain_option,
            &(wo_param_data->relgain_option),
            sizeof(wo_param_data->relgain_option), 1);

    IAS_PARM_ADD_STRING(wo_param_table, OLI_RPS_PROC_LEVEL_RAD,
        "RPS processing level required for OLI histogram statistics -- input "
        "data have been converted to radiance values", IAS_PARM_OPTIONAL,
        0, NULL, 1, default_solar_rps_proc_level_rad,
        wo_param_data->rps_solar_proc_level_rad,
        sizeof(wo_param_data->rps_solar_proc_level_rad), 0);

    IAS_PARM_ADD_STRING(wo_param_table, OLI_SHUTTER_RPS_PROC_LEVEL_DN,
        "RPS processing level required for OLI shutter histogram statistics "
        "-- input data have not been converted to radiance values",
        IAS_PARM_OPTIONAL, 0, NULL, 1, default_shutter_rps_proc_level,
        wo_param_data->rps_shutter_proc_level,
        sizeof(wo_param_data->rps_shutter_proc_level), 0);

    /* Read the output options */
    const char *default_bias_report_filename_prefix[] = {"bias_stability"};
    IAS_PARM_ADD_STRING(wo_param_table, BIAS_REPORT_FILENAME_PREFIX, 
        "Output report file name prefix for bias stability", IAS_PARM_OPTIONAL,
        0, NULL, /* no restrictions */
        1, default_bias_report_filename_prefix, /* Default file name */
        wo_param_data->bias_report_filename_prefix, 
        sizeof(wo_param_data->bias_report_filename_prefix), 0);

    const char *default_gain_report_filename_prefix[] = {"gain_stability"};
    IAS_PARM_ADD_STRING(wo_param_table, GAIN_REPORT_FILENAME_PREFIX, 
        "Output report file name prefix for gain stability", IAS_PARM_OPTIONAL,
        0, NULL, /* no restrictions */
        1, default_gain_report_filename_prefix, /* Default file name */
        wo_param_data->gain_report_filename_prefix, 
        sizeof(wo_param_data->gain_report_filename_prefix), 0);

    IAS_PARM_ADD_BOOLEAN(wo_param_table, REPORT_FLAG,
            "Write characterization results to summary reports "
            "1=Write summary reports, 0=Do not write summary reports",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
            &default_report_flag,
            &(wo_param_data->report_flag),
            sizeof(wo_param_data->report_flag), 1);

    IAS_PARM_ADD_BOOLEAN(wo_param_table, DUMP_FLAG,
            "Write characterization results to dump files "
            "1=Write dump files, 0=Do not write dump files",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
            &default_dump_flag,
            &(wo_param_data->dump_flag),
            sizeof(wo_param_data->dump_flag), 1);


    /* Get the total number of items to read from the parameter table */
    read_count = IAS_PARM_GET_TABLE_SIZE(wo_param_table);

    /* If someone enters -- help, --template, etc, on the command line
       when starting the application, the appropriate information will
       be displayed.  If someone enters a file name, its access will be
       checked.  If a help option was specified or the specified file name
       could not be accessed, a 1 will be returned.  In this case, just
       copy the work order file name (or help option) to the OMF
       file name and continue.  Otherwise, read the work order file */
    status = ias_parm_provide_help(wo_params_filename, wo_param_table,
                read_count, IAS_INPUT_PARAMETERS);

    if (status == ERROR)
    {
       IAS_LOG_ERROR("Work order parameter help failed");
       return ERROR;
    }
    else if (status == 1)
    {
       sprintf(omf_params_filename, "%s", wo_params_filename);
    }
    else
    {
        status = ias_parm_read(wo_params_filename, wo_param_table,
                read_count);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading work order parameters from file %s",
                    wo_params_filename);
            return ERROR;
        }

        /* Construct the full path to the OMF file */
        sprintf(omf_params_filename, "%s.omf", wo_param_data->work_order_id);

        /* Get the total number of bands to process */
        wo_param_data->num_bands_to_process = IAS_PARM_GET_COUNT_READ(
                wo_param_table,
                BAND_LIST);

        /* Verify we've got valid values for selected work order parameters */
        IAS_LOG_DEBUG("Checking work order parameters...");

        /* Rather than check for a range of values, just check for the two
           possible values of 0 (RPS_RADSTABILITY_USE_DB_RADIANCE) or 1
           (RPS_RADSTABILITY_CALCULATE_RADIANCE) */
        if ((wo_param_data->calculate_radiance
             != RPS_RADSTABILITY_USE_DB_RADIANCE)
            && (wo_param_data->calculate_radiance
                    != RPS_RADSTABILITY_CALCULATE_RADIANCE))
        {
            IAS_LOG_ERROR("Invalid radiance source selection--valid "
                          "options are to read them from the database "
                          "or calculate them directly");
            return ERROR;
        }

        /* Make sure at least one output source (i.e. database, summary
           report, dump file) has been selected.  If none have, that's an
           error */
        if ((!wo_param_data->store_char_in_db)
                && (!wo_param_data->report_flag)
                && (!wo_param_data->dump_flag))
        {
            IAS_LOG_ERROR("No output destination is specified for "
                "characterization results");
            return ERROR;
        }
    }


/*-----------------------------------------------------------------------
                         READ OMF PARAMETER FILE
-------------------------------------------------------------------------*/
    /* Declare the OMF parameter table */
    IAS_PARM_DECLARE_TABLE(omf_param_table, MAX_NUM_OMF_PARAMS);

    /* Common parameters */
    IAS_PARM_SATELLITE(omf_param_table, &omf_param_data->satellite_number,
            sizeof(omf_param_data->satellite_number), 1);

    IAS_PARM_L0R_FILENAME(omf_param_table, IAS_PARM_REQUIRED,
            omf_param_data->l0r_filename, sizeof(omf_param_data->l0r_filename),
            1);

    IAS_PARM_CPF(omf_param_table, IAS_PARM_REQUIRED,
            omf_param_data->cpf_filename,
            sizeof(omf_param_data->cpf_filename), 1);

    /* And the name of the pixel mask file */
    IAS_PARM_PIXEL_MASK_FILENAME(omf_param_table, IAS_PARM_REQUIRED,
            omf_param_data->pmask_filename,
            sizeof(omf_param_data->pmask_filename), 1);

    /* Get the total number of items to read in the OMF parameter table */
    read_count = IAS_PARM_GET_TABLE_SIZE(omf_param_table);

    /* See if the user wanted help or a template this time with the
       OMF file.  If not, just read the file */
    status = ias_parm_provide_help(omf_params_filename, omf_param_table,
                read_count, IAS_OMF_PARAMETERS);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("OMF parameter help failed");
        return ERROR;
    }
    else if (status == 1)
    {
        exit(EXIT_SUCCESS);
    }

    /* Read the OMF parameters */
    status = ias_parm_read(omf_params_filename, omf_param_table, read_count);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading OMF parameter file");
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}
