/**********************************************************************
 NAME:               rad_stability_utility

 PURPOSE:    Module implementing "utility" routines called by the
             char_rad_stability application
************************************************************************/
#include <float.h>
#include <string.h>                 /* For strrchr () */
#include <unistd.h>

#include "ias_l0r.h"
#include "ias_miscellaneous.h"
#include "ias_radiometry.h"

#include "char_rad_stability.h"


/*---------------------------------------------------------------------
 NAME:            allocate_band_summary_array

 PURPOSE:    Allocates an array of RPS_RADSTABILITY_DNRAD_BAND_SUMMARY
             data structures for the required number of bands

 RETURNS:    Pointer to a valid RPS_RADSTABILITY_DNRAD_BAND_SUMMARY
             data structure if successful; NULL pointer if failed
-----------------------------------------------------------------------*/
RPS_RADSTABILITY_DNRAD_BAND_SUMMARY *allocate_band_summary_array
(
    int num_oli_bands,                        /* I: Actual number of
                                                 OLI bands */
    int num_oli_scas                          /* I: Number of SCAs in each
                                                 OLI band */
)
{
    int band_index;                           /* Band loop counter */
    int ii;                                   /* Local loop counter */

    RPS_RADSTABILITY_DNRAD_BAND_SUMMARY *band_summary = NULL;


    /* Allocate an array for all bands */
    band_summary = calloc(num_oli_bands,
          sizeof(RPS_RADSTABILITY_DNRAD_BAND_SUMMARY));
    if (band_summary == NULL)
    {
        IAS_LOG_ERROR("Allocating band array");
        return NULL;
    }

    /* Allocate array space for the summary statistics and
       signal level changes for all SCAs in the band */
    for (band_index = 0; band_index < num_oli_bands; band_index++)
    {
        band_summary[band_index].summary_stats = calloc(num_oli_scas,
              sizeof(RPS_RADSTABILITY_DNRAD_SCA_SUMMARY_STATS));

        band_summary[band_index].siglevel_change = calloc(num_oli_scas,
              sizeof(RPS_RADSTABILITY_DNRAD_SCA_SUMMARY_SIGLEVEL_CHANGE));

        if ((band_summary[band_index].summary_stats == NULL)
                || (band_summary[band_index].siglevel_change == NULL))
        {
            IAS_LOG_ERROR("Allocating internal SCA summary statistics "
                          "and/or signal level change data buffers");
            for (ii = 0; ii < band_index; ii++)
            {
                free(band_summary[ii].summary_stats);
                free(band_summary[ii].siglevel_change);
            }
            free(band_summary[band_index].summary_stats);
            free(band_summary[band_index].siglevel_change);
            free(band_summary);
            return NULL;
        }
    }

    /* Done */
    return band_summary;
}




/*---------------------------------------------------------------------
 NAME:            free_band_summary_array

 PURPOSE:    Frees an allocated array of RPS_RADSTABILITY_DNRAD_BAND_SUMMARY
             data structures

 RETURNS:    Nothing
-----------------------------------------------------------------------*/
void free_band_summary_array
(
    RPS_RADSTABILITY_DNRAD_BAND_SUMMARY
        *band_summary,                   /* I: Band summary data structure
                                            to free */
    int num_oli_bands                    /* I: Number of "normal" OLI
                                            imaging bands */
)
{
    int band_index;

    if (band_summary != NULL)
    {
        for (band_index = 0; band_index < num_oli_bands; band_index++)
        {
            /* Free the internal SCA summary data arrays */
            free(band_summary[band_index].summary_stats);
            free(band_summary[band_index].siglevel_change);
        }

        /* Free the main band data array */
        free(band_summary);
    }

    /* Done */
}




/*--------------------------------------------------------------------
 NAME:             allocate_detector_data_buffers

 PURPOSE:  Allocates all required internal buffers in a
           DETECTOR_DATA_BUFFERS data structure for the specified 
           stability characterization

 RETURNS:  Integer status code of SUCCESS or ERROR
----------------------------------------------------------------------*/
int allocate_detector_data_buffers
(
    DETECTOR_DATA_BUFFERS *data_buffers,  /* I/O: Detector data buffer
                                             structure */
    const char *stability_id,             /* I: Stability characterization
                                             ("BIAS", "GAIN") */
    IAS_SPECTRAL_TYPE spectral_type,      /* I: Spectral type for the
                                             current band */
    int num_detectors_sca                 /* I: Number of detectors in an
                                             SCA */
)
{
    /* Allocate DN-based data buffers for either stability characterization.*/
    data_buffers->max_dn = malloc(
         num_detectors_sca * sizeof(*(data_buffers->max_dn)));
    data_buffers->min_dn = malloc(
         num_detectors_sca * sizeof(*(data_buffers->min_dn)));
    data_buffers->mean_dn = malloc(
         num_detectors_sca * sizeof(*(data_buffers->mean_dn)));
    data_buffers->stdev_dn = malloc(
         num_detectors_sca * sizeof(*(data_buffers->stdev_dn)));

    /* Allocate DN-based two-sigma data buffer for either stability
       characterization */
    data_buffers->two_sigma_dn = malloc(
            num_detectors_sca * sizeof(*(data_buffers->two_sigma_dn)));

    /* Check to make sure these buffers were allocated */
    if ((data_buffers->max_dn == NULL)
            || (data_buffers->min_dn == NULL)
            || (data_buffers->mean_dn == NULL)
            || (data_buffers->stdev_dn == NULL)
            || (data_buffers->two_sigma_dn == NULL))
    {
        IAS_LOG_ERROR("Cannot allocate one or more required detector data "
                     "buffers for DN-based characterizations");
        free_detector_data_buffers(data_buffers);
        return ERROR;
    }

    /* Allocate additional data buffers for gain stability characterization */
    if (strcmp(stability_id, "GAIN") == 0)
    {
        /* Allocate radiance-based data buffers */
        data_buffers->max_rad = malloc(
            num_detectors_sca * sizeof(*(data_buffers->max_rad)));
        data_buffers->min_rad = malloc(
            num_detectors_sca * sizeof(*(data_buffers->min_rad)));
        data_buffers->mean_rad = malloc(
            num_detectors_sca * sizeof(*(data_buffers->mean_rad)));
        data_buffers->stdev_rad = malloc(
            num_detectors_sca * sizeof(*(data_buffers->stdev_rad)));

        data_buffers->two_sigma_rad = malloc(
            num_detectors_sca * sizeof(*(data_buffers->two_sigma_rad)));

        /* Finally allocate the percent change data buffers */
        data_buffers->pcnt_change_gain = malloc(
            num_detectors_sca * sizeof(*(data_buffers->pcnt_change_gain)));
        data_buffers->pcnt_change_rad = malloc(
            num_detectors_sca * sizeof(*(data_buffers->pcnt_change_rad)));

        /* Check for any failed allocations */
        if ((data_buffers->max_rad == NULL)
            || (data_buffers->min_rad == NULL)
            || (data_buffers->mean_rad == NULL)
            || (data_buffers->stdev_rad == NULL)
            || (data_buffers->two_sigma_rad == NULL)
            || (data_buffers->pcnt_change_gain == NULL)
            || (data_buffers->pcnt_change_rad == NULL))
        {
            IAS_LOG_ERROR("Cannot allocate one or more required "
                          "gain/radiance characterization detector data "
                          "buffers");
            free_detector_data_buffers(data_buffers);
            return ERROR;
        }
    }

    /* Done */
    return SUCCESS;
}



/*---------------------------------------------------------------------
 NAME:            free_detector_data_buffers

 PURPOSE:   Frees allocated detector data buffers for the specified
            stability characterization

 RETURNS:   Nothing
-----------------------------------------------------------------------*/
void free_detector_data_buffers
(
    DETECTOR_DATA_BUFFERS *data_buffers    /* I/O:  Detector data buffer
                                              data structure */
)
{
    /* Free the DN-based buffers -- they're allocated for both
       characterizations */
    free(data_buffers->max_dn);
    free(data_buffers->min_dn);
    free(data_buffers->mean_dn);
    free(data_buffers->stdev_dn);
    free(data_buffers->two_sigma_dn);

    /* Free the gain/radiance buffers, too.  It won't matter if these are
       already NULL */
    free(data_buffers->max_rad);
    free(data_buffers->min_rad);
    free(data_buffers->mean_rad);
    free(data_buffers->stdev_rad);
    free(data_buffers->two_sigma_rad);
    free(data_buffers->pcnt_change_gain);
    free(data_buffers->pcnt_change_rad);
}



/*---------------------------------------------------------------------
 NAME:            rad_stability_cleanup

 PURPOSE:   Frees allocated memory for the specified stability
            characterization

 RETURNS:   Nothing
-----------------------------------------------------------------------*/
void rad_stability_cleanup
(
    DETECTOR_DATA_BUFFERS *data_buffers,    /* I/O: Detector data buffer
                                               data structure */
    int *statistics_invalid_flags,
                                            /* I: Data invalid flags for
                                               all detectors in the current
                                               band/SCA */
    RPS_RADSTABILITY_DNRAD_BAND_SUMMARY *band_summary,
                                            /* I: Data structure with
                                               band-level stability
                                               summary results */
    int num_oli_bands                       /* I: Total # of "normal"
                                               OLI imaging bands */
)
{


    /* Free the allocated detector buffers */
    free_detector_data_buffers(data_buffers);

    /* Free the band summary data structure if it's been allocated */
    free_band_summary_array(band_summary, num_oli_bands);

    /* Free the invalid flag data buffer */
    free(statistics_invalid_flags);

    /* Done */
}




/*-----------------------------------------------------------------------
 NAME:              generate_summary_statistics

 PURPOSE:   Determines SCA-level histogram statistics and populates the
            information in an SCA summary statistics data structure

 RETURNS:   Integer status code of SUCCESS or ERROR
-------------------------------------------------------------------------*/
int generate_summary_statistics
(
    const double *det_stats_max_buffer,  /* I: Detector-level histogram
                                            statistics maxima for the
                                            current band/SCA */
    const double *det_stats_min_buffer,  /* I: Detector-level histogram
                                            statistics minima for the
                                            current band/SCA */
    const double *det_stats_mean_buffer, /* I: Detector-level histogram
                                            statistics means for
                                            the current band/SCA */
    const int *statistics_invalid_flags, /* I: Statistics invalid flags
                                            for all detectors in the
                                            current band/SCA (0=valid data,
                                            nonzero=invalid data) */
    int num_detectors_sca,               /* I: Number of detectors in
                                            an SCA */
    const char *buffer_set,              /* I: String indicating which
                                            set of summary statistics data
                                            buffers to populate
                                            ("DN", "RADIANCE") */
    RPS_RADSTABILITY_DNRAD_SCA_SUMMARY_STATS *stats
                                          /* O: Element of SCA summary
                                             histogram statistics array
                                             populated for the current SCA */
)
{
    double sca_max = DBL_MIN;          /* Histogram statistics maximum
                                          for the current band/SCA */
    double sca_min = DBL_MAX;          /* Histogram statistics minimum
                                          for the current band/SCA */
    double sca_mean;                   /* Histogram statistics mean for
                                          all detectors in the current
                                          band/SCA */
    double sca_stdev;                  /* Histogram statistics std dev for
                                          all detectors in the current
                                          band/SCA */

    int status;                        /* Function return status */
    int detector_index;                /* Detector loop counter */


    /* Calculate the SCA mean and standard deviation */
    status = ias_rps_calculate_filtered_mean(det_stats_mean_buffer,
        statistics_invalid_flags, num_detectors_sca, &sca_mean,
        NULL);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating SCA mean");
        return ERROR;
    }

    status = ias_rps_calculate_filtered_stdev(det_stats_mean_buffer,
        statistics_invalid_flags, num_detectors_sca, sca_mean, &sca_stdev);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating SCA std deviation");
        return ERROR;
    }

    /* Keep track of the SCA-level minima and maxima */
    for (detector_index = 0;
         detector_index < num_detectors_sca;
         detector_index++)
    {
        if (!statistics_invalid_flags[detector_index])
        {
            if (det_stats_max_buffer[detector_index] > sca_max)
                sca_max = det_stats_max_buffer[detector_index];

            if (det_stats_min_buffer[detector_index] < sca_min)
                sca_min = det_stats_min_buffer[detector_index];
        }
    }

    /* Set the final summary SCA-level statistics values.  For bias
       statistics and diffuser L0R bias-corrected/linearized statistics,
       populate the DN members of the data structure.  For diffuser
       radiances (L1R or internally calculated), populate the radiance
       members of the data structure. */
    if (strcmp(buffer_set, "DN") == 0)
    {
        stats->sca_max_dn   = sca_max;
        stats->sca_min_dn   = sca_min;
        stats->sca_mean_dn  = sca_mean;
        stats->sca_stdev_dn = sca_stdev;
    }
    else if (strcmp(buffer_set, "RADIANCE") == 0)
    {
        stats->sca_max_rad   = sca_max;
        stats->sca_min_rad   = sca_min;
        stats->sca_mean_rad  = sca_mean;
        stats->sca_stdev_rad = sca_stdev;
    }
    else
    {
        IAS_LOG_ERROR("Invalid buffer set specification '%s'", buffer_set);
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}

