/**************************************************************************
 NAME:              rad_stability_write_dump_data_file

 PURPOSE:           Module implementing routines to write 60s radiometric
                    stability characterization results for a given band
                    to an optional dump file

***************************************************************************/
#include "ias_const.h"
#include "ias_miscellaneous.h"
#include "ias_radiometry.h"
#include "char_rad_stability.h"

/*------------------------------------------------------------------------
 NAME:          write_detector_stability_results_to_dump_file

 PURPOSE:   Writes detector two-sigma and percent change characterization
            results to the dump file for the current band

 RETURNS:   Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int write_detector_stability_results_to_dump_file
(
    DETECTOR_DATA_BUFFERS data_buffers,/* I: Data structure containing
                                          detector-level results */
    const char *stability_id,          /* I: Flag indicating whether
                                          bias stability characterization
                                          or gain stability
                                          characterization was performed */
    int band_number,                   /* I: 1-based band number */
    const char *band_name,             /* I: Band name/description*/
    int sca_number,                    /* I: 1-based SCA number */
    int num_detectors_sca,             /* I: # of detectors in an SCA */
    FILE *dump_fptr                    /* I: Open dump file for the current
                                          band */
)
{
    int detector_index;                /* Detector loop counter */
    int status;                        /* Function return status */


    /* If we're characterizing gain/radiance stability, write the two-sigma
       changes in dn/radiance levels followed by the percent changes in
       gains/radiance levels */
    if (strcmp(stability_id, "GAIN") == 0)
    {
        if ((sca_number - 1) == 0)
        {
            status = fprintf(dump_fptr, "Band %d (%s) DETECTOR GAIN "
                "STABILITY RESULTS\n\n               TWO-SIGMA           "
                "PCNT CHANGE\nSCA DET     DN         RAD       GAIN        "
                "RAD\n---------------------------------------------------\n",
                band_number, band_name);
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing detector gain stability header to "
                              "dump file");
                return ERROR;
            }
        }

        for (detector_index = 0;
             detector_index < num_detectors_sca;
             detector_index++)
        {
            status = fprintf(dump_fptr,
                "%02d  %03d %10.6lf %10.6lf %10.6lf  %10.6lf\n",
                sca_number, (detector_index + 1),
                data_buffers.two_sigma_dn[detector_index],
                data_buffers.two_sigma_rad[detector_index],
                data_buffers.pcnt_change_gain[detector_index],
                data_buffers.pcnt_change_rad[detector_index]);

            if (status < 0)
            {
                IAS_LOG_ERROR("Writing detector gain stability results "
                              "for detector %d in band %d, SCA %d to dump "
                              "file", (detector_index + 1),
                              band_number,
                              sca_number);
                return ERROR;
            }
        }
    }
    else           /* Write detector bias stability results */
    {
        if ((sca_number - 1) == 0)
        {
            status = fprintf(dump_fptr, "Band %d (%s) DETECTOR BIAS "
                "STABILITY RESULTS\n\n         TWO-SIGMA\nSCA DET"
                "     DN\n---------------------------------------------\n",
                band_number, band_name);
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing detector gain stability header to "
                              "dump file");
                return ERROR;
            }
        }

        for (detector_index = 0;
             detector_index < num_detectors_sca;
             detector_index++)
        {
            status = fprintf(dump_fptr, "%02d  %03d %10.6lf\n",
                sca_number, (detector_index + 1), 
                data_buffers.two_sigma_dn[detector_index]);

            if (status  < 0)
            {
                IAS_LOG_ERROR("Writing detector bias stability results "
                              " for detector %d in band %d, SCA %d to dump "
                              "file", (detector_index + 1),
                              band_number,
                              sca_number);
                return ERROR;
            }
        }
    }

    /* Done */
    return SUCCESS;
}



/*------------------------------------------------------------------------
 NAME:            write_sca_stability_results_to_dump_file

 PURPOSE:   Writes each band's SCA stability characterization results
            to its dump file

 RETURNS:   Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int write_sca_stability_results_to_dump_file
(
    const RPS_RADSTABILITY_DNRAD_BAND_SUMMARY *band_summary,
                                            /* I: Stability results for
                                               the current band */
    const char *stability_id,               /* I: String indicating whether
                                               bias stability or gain
                                               stability was
                                               characterized */
    FILE *dump_fptr                         /* I: Open dump data file */
)
{
    int sca_index;                          /* SCA loop counter */
    int status;                             /* Function return status */


    /* Write the DN-based summary histogram statistics*/
    status = fprintf(dump_fptr, "\n\n\nBAND %d (%s) SUMMARY DN HISTOGRAM "
        "STATISTICS\n\n"
        "SCA     MAX         MIN          MEAN         STDEV\n----------"
        "-------------------------------------------\n",
        band_summary->band_number, band_summary->band_name);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing band summary header to dump file");
        return ERROR;
    }

    for (sca_index = 0; sca_index < band_summary->num_scas; sca_index++)
    {
        status = fprintf(dump_fptr, "%02d  %11.6lf  %11.6lf %11.6lf"
            "  %11.6lf\n", (sca_index + 1),
            band_summary->summary_stats[sca_index].sca_max_dn,
            band_summary->summary_stats[sca_index].sca_min_dn,
            band_summary->summary_stats[sca_index].sca_mean_dn,
            band_summary->summary_stats[sca_index].sca_stdev_dn);
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing summary statistics results for band "
                          "%d, SCA %d",
                          band_summary->band_number,
                          (sca_index + 1));
            return ERROR;
        }
    }

    /* If we've done gain stability characterization, write the
       radiance based histogram statistics */
    if (strcmp(stability_id, "GAIN") == 0)
    {
        status = fprintf(dump_fptr, "\n\n\nBAND %d (%s) SUMMARY RAD HISTOGRAM "
            "STATISTICS\n\nSCA      MAX          MIN          MEAN"
            "       STDEV\n-------------------------------------------------"
            "---\n", band_summary->band_number, band_summary->band_name);
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing band summary header to dump file");
            return ERROR;
        }

        for (sca_index = 0; sca_index < band_summary->num_scas; sca_index++)
        {
            status = fprintf(dump_fptr, "%02d  %11.6lf  %11.6lf "
            " %11.6lf %11.6lf\n", (sca_index + 1),
            band_summary->summary_stats[sca_index].sca_max_rad,
            band_summary->summary_stats[sca_index].sca_min_rad,
            band_summary->summary_stats[sca_index].sca_mean_rad,
            band_summary->summary_stats[sca_index].sca_stdev_rad);

            if (status < 0)
            {
                IAS_LOG_ERROR("Writing summary statistics results for band "
                    "%d, SCA %d", band_summary->band_number, (sca_index + 1));
                return ERROR;
            }
        }

        /* Now write the two-sigma and pcnt change in gain/radiance
           information */
        status = fprintf(dump_fptr, "\n\n\nBAND %d (%s) TWO-SIGMA AND PCNT "
            "CHANGE RESULTS\n\n              TWO-SIGMA                  "
            "     PCNT CHANGE\nSCA       DN MEAN    RAD MEAN            "
            "  GAIN         RAD\n--------------------------------------------"
            "-----------------\n", band_summary->band_number,
            band_summary->band_name);
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing stability metrics header to dump file");
            return ERROR;
        }

        for (sca_index = 0; sca_index < band_summary->num_scas; sca_index++)
        {
            status = fprintf(dump_fptr, "%02d     %10.6lf  %10.6lf          "
            "%10.6lf  %10.6lf\n", (sca_index + 1),
            band_summary->siglevel_change[sca_index].sca_two_sigma_dn,
            band_summary->siglevel_change[sca_index].sca_two_sigma_rad,
            band_summary->siglevel_change[sca_index].sca_pcnt_change_gain,
            band_summary->siglevel_change[sca_index].sca_pcnt_change_rad);
        }
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing gain stability metrics results to dump "
                "file");
            return ERROR;
        }
    }
    else
    {
        /* Write the two-sigma results to the dump file */
        status = fprintf(dump_fptr, "\n\n\nBAND %d (%s) TWO-SIGMA "
            "RESULTS\n\n         TWO-SIGMA\nSCA       DN MEAN\n------------"
            "------\n", band_summary->band_number, band_summary->band_name);
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing bias stability metrics header to dump "
                "file");
            return ERROR;
        }
        for (sca_index = 0; sca_index < band_summary->num_scas; sca_index++)
        {
            status = fprintf(dump_fptr, "%02d     %10.6lf\n", (sca_index + 1),
                band_summary->siglevel_change[sca_index].sca_two_sigma_dn);
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing bias stability metrics to dump "
                    "file");
                return ERROR;
            }
        }
    }

    /* Done */
    return SUCCESS;
}



/*------------------------------------------------------------------------
 NAME:                 write_header_to_dumpfile

 PURPOSE:    Writes IAS/RPS report "header" information to a dump file

 RETURNS:    Integer status code of SUCCESS or ERROR

 NOTES:      The intent of this routine is to provide the same header
             information as contained in the summary report, in the same
             order
--------------------------------------------------------------------------*/
int write_header_to_dumpfile
(
    RPS_RADSTABILITY_REPORT_HEADER report_header, /* I: RPS report header
                                                   information */
    FILE *dump_fptr                             /* I: Dump file handle */
)
{
    int status;                                 /* Function return status */
    int band_number;                            /* Band number recorded in
                                                   report header */
    int is_solar_collect = 0;                   /* Is OLI solar collect
                                                   (0=no, 1=yes) */
    int is_shutter_collect = 0;                 /* Is OLI shutter collect
                                                   (0=no, 1=yes) */
    int  starting_path;
    int  starting_row;
    int  ending_row;

    /* Pointers to strings contained in the "common" IAS report header */
    char *algorithm_name = NULL;
    char *collection_type = NULL;
    char *cpf_name = NULL;
    char *ias_sw_version = NULL;
    char *ingest_sw_version = NULL;
    char *sensor_id = NULL;
    char *spacecraft_id = NULL;
    char *l0r_filename = NULL;
    char *work_order_id = NULL;

    char timestamp_string[IAS_L0R_DATE_LENGTH]; /* IAS timestamp */


    /* Write dump file "title" */
    algorithm_name = ias_misc_get_header_algorithm_name(
        report_header.common_header);
    if (algorithm_name == NULL)
    {
        status = fprintf(dump_fptr, "OLI 60s Radiometric Stability "
            "Characterization Dump File\n");
    }
    else
    {
        status = fprintf(dump_fptr, "OLI %s Dump File\n",
            algorithm_name);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing title to dump file");
        return ERROR;
    }

    /* Write the timestamp string */
    status = ias_misc_get_timestamp("%c", IAS_L0R_DATE_LENGTH,
            timestamp_string);

    if (status != SUCCESS)
    {
        IAS_LOG_WARNING("Cannot obtain current date/time stamp for dump "
                      "file");
        status = fprintf(dump_fptr, "Report Date:\t\t\tN/A\n");
    }
    else
    {
        status = fprintf(dump_fptr, "Report Date:\t\t\t%23s\n",
            timestamp_string);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing timestamp information to dump file");
        return ERROR;
    }

    /* Band number goes here */
    band_number = ias_misc_get_header_band_number(
        report_header.common_header);

    if (band_number < 0)
    {
        status = fprintf(dump_fptr, "Band:\t\t\t\tN/A\n");
    }
    else
    {
        status = fprintf(dump_fptr, "Band:\t\t\t\t%d\n", band_number);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing band number to dump file");
        return ERROR;
    }

    /* Write IAS processing software information */
    work_order_id
        = ias_misc_get_header_work_order_id(report_header.common_header);
    ias_sw_version
        = ias_misc_get_header_ias_sw_version(report_header.common_header);
    ingest_sw_version
        = ias_misc_get_header_ingest_sw_version(report_header.common_header);

    if (work_order_id == NULL)
    {
        status = fprintf(dump_fptr, "Work Order ID:\t\t\tN/A\n");
    }
    else
    {
        status = fprintf(dump_fptr, "Work Order ID:\t\t\t%s\n",
            work_order_id);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing work order ID to dump file");
        return ERROR;
    }

    if (ias_sw_version == NULL)
    {
        status = fprintf(dump_fptr, "IAS Version:\t\t\tN/A\n");
    }
    else
    {
        status = fprintf(dump_fptr, "IAS Version:\t\t\t%s\n",
            ias_sw_version);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing IAS software version to dump file");
        return ERROR;
    }

    if (ingest_sw_version == NULL)
    {
        status = fprintf(dump_fptr, "Ingest Version:\t\t\tN/A\n");
    }
    else
    {
        status = fprintf(dump_fptr, "Ingest Version:\t\t\t%s\n",
            ingest_sw_version);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing INGEST software version to dump file");
        return ERROR;
    }

    /* Write L0R Reference image information */
    l0r_filename
        = ias_misc_get_header_l0r_filename(report_header.common_header);
    if (l0r_filename == NULL)
    {
        status = fprintf(dump_fptr, "L0R Image:\t\t\tN/A\n");
    }
    else
    {
        status = fprintf(dump_fptr, "L0R Image:\t\t\t%s\n",
                         l0r_filename);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing L0R file name to dump file");
        return ERROR;
    }

    /* WRS starting path/row information */
    ias_misc_get_header_starting_ending_pathrow(report_header.common_header,
        &starting_path, &starting_row, &ending_row);
    status = fprintf(dump_fptr, "Starting WRS Path:\t\t%03d\n"
        "Starting WRS Row:\t\t%03d\nEnding WRS Row:\t\t\t%03d\n",
        starting_path, starting_row, ending_row);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing starting/ending WRS path/row information");
        return ERROR;
    }

    /* Acquisition date */
    status = fprintf(dump_fptr, "Acquisition Date:\t\t%s\n",
        ias_misc_get_header_date_acquired(report_header.common_header));
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing acquisition date information to dump file");
        return ERROR;
    }

    /* Write spacecraft/sensor IDs */
    spacecraft_id
        = ias_misc_get_header_spacecraft_id(report_header.common_header);
    sensor_id = ias_misc_get_header_sensor_id(report_header.common_header);

    if (spacecraft_id == NULL)
    {
        status = fprintf(dump_fptr, "Spacecraft:\t\t\tN/A\n");
    }
    else
    {
        status = fprintf(dump_fptr, "Spacecraft:\t\t\t%s\n", spacecraft_id);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing spacecraft ID to dump file");
        return ERROR;
    }

    if (sensor_id == NULL)
    {
        status = fprintf(dump_fptr, "Sensor:\t\t\t\tN/A\n");
    }
    else
    {
        if (strcmp(sensor_id, "OLI") == 0)
        {
            status = fprintf(dump_fptr, "Sensor:\t\t\t\t%s\n", sensor_id);
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing sensor ID to dump file");
                return ERROR;
            }
        }
        else
        {
            IAS_LOG_ERROR("Radiometric stability characterization is "
                "currently run only for OLI solar/shutter collects");
            return ERROR;
        }
    }

    /* Write collection type information next */
    collection_type =
        ias_misc_get_header_collection_type(report_header.common_header);
    if (collection_type == NULL)
    {
        status = fprintf(dump_fptr, "Collection Type:\t\t\n");
    }
    else
    {
        is_solar_collect =
            (strcmp(collection_type, IAS_RPS_OLI_SOLAR_COLLECT) == 0);
        is_shutter_collect =
            (strcmp(collection_type, IAS_RPS_OLI_SHUTTER_COLLECT) == 0);
        if (is_solar_collect || is_shutter_collect)
        {
            status = fprintf(dump_fptr, "Collection Type:\t\t%s\n",
                collection_type);
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing collection type to dump file");
                return ERROR;
            }
        }
        else
        {
            IAS_LOG_ERROR("Radiometric stability characterization is "
                "currently run only for OLI solar/shutter collects");
            return ERROR;
        }
    }

    /* Write processing CPF */
    cpf_name = ias_misc_get_header_cpf_name(report_header.common_header);
    if (cpf_name == NULL)
    {
        status = fprintf(dump_fptr, "Processing CPF:\t\t\tN/A\n");
    }
    else
    {
        status = fprintf(dump_fptr, "Processing CPF:\t\t\t%s\n", cpf_name);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Formatting processing CPF string");
        return ERROR;
    }

    /* Write acquisition date and start/stop times */
    status = fprintf(dump_fptr, "Image Start Time:\t\t%s\n"
        "Image Stop Time:\t\t%s\n",
        report_header.rps_header.acq_datetime.start_time_oli,
        report_header.rps_header.acq_datetime.stop_time_oli);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing image start/stop times to dump file");
        return ERROR;
    }

    /* Integration times/selected pixel map/image data truncation setting */
    status = fprintf(dump_fptr, "MS Band Integ. Time:\t\t%d\n"
        "Pan Band Integ. Time:\t\t%-d\nDetector Select Table ID:\t%d\n"
        "Image Data Truncation Setting:\t%d\n",
        report_header.rps_header.ms_integration_time,
        report_header.rps_header.pan_integration_time,
        report_header.rps_header.current_detector_select_table,
        report_header.rps_header.image_data_truncation_setting);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing integration time/pixel map "
                      "information to dump file");
        return ERROR;
    }

    status = fprintf(dump_fptr, "Number of Frames:\t\t%d\nFrame Rate "
        "(frms/s):\t\t%d\n",
        report_header.rps_header.num_oli_frames,
        report_header.rps_header.average_oli_frame_rate);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing number frames and/or frame rate "
            "information to dump file");
        return ERROR;
    }

    /* Focal plane temperatures */
    status = fprintf(dump_fptr, "FPM07 Start Temp.:\t\t%7.4lf\n"
        "FPM07 Stop Temp.:\t\t%7.4lf\nFPM14 Start Temp.:\t\t%7.4lf\n"
        "FPM14 Stop Temp:\t\t%7.4lf\nFPA Window Start Temp.:\t\t%7.4lf\n"
        "FPA Window Stop Temp.:\t\t%7.4lf\nFPE Chassis STart Temp.:"
        "\t%7.4lf\nFPE Chassis Stop Temp:\t\t%7.4lf\n",
        report_header.rps_header.oli_fpm7_start_temp,
        report_header.rps_header.oli_fpm7_stop_temp,
        report_header.rps_header.oli_fpm14_start_temp,
        report_header.rps_header.oli_fpm14_stop_temp,
        report_header.rps_header.oli_fpa_window_start_temp,
        report_header.rps_header.oli_fpa_window_stop_temp,
        report_header.rps_header.oli_fpe_chassis_start_temp,
        report_header.rps_header.oli_fpe_chassis_stop_temp);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing temperature information to dump file");
        return ERROR;
    }

    /* Write diffuser panel summary information */
    if (report_header.panel_id == IAS_RPS_OLI_DIFFUSER_TYPE_PRISTINE)
    {
        status = fprintf(dump_fptr, "Diffuser Panel Deployed:\tPristine\n");
    }
    else if (report_header.panel_id == IAS_RPS_OLI_DIFFUSER_TYPE_WORKING)
    {
        status = fprintf(dump_fptr, "Diffuser Panel Deployed:\tWorking\n");
    }
    else if (report_header.panel_id == IAS_RPS_OLI_DIFFUSER_TYPE_NOTDEPLOYED)
    {
        status = fprintf(dump_fptr, "Diffuser Panel Deployed:\tN/A\n");
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing characterization information to dump file");
        return ERROR;
    }
    status = fprintf(dump_fptr, "Diffuser Resolver Position:\t%u\n",
        report_header.diffuser_resolver_position);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing diffuser position information to dump file");
        return ERROR;
    }

    /* Write stability threshold and observed collect length
       (stability period) */
    status = fprintf(dump_fptr, "Stability delta_t (sec):\t%7.3lf\n"
        "Stability period (sec):\t\t%7.3lf\n",
                     report_header.stability_delta_t,
                     report_header.stability_period);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing stability delta_t and stability period "
                      "to dump file");
        return ERROR;
    }

    /* Write radiance source information here.  Use 'strchr' to look
       for 'G' in the report header RPS processing level string indicating
       that gain application has been applied.  This allows for the
       possibility (however remote) that the processing sequence could change
       from what's currently used.  */
    if (is_solar_collect)
    {
        if (strchr(report_header.proc_level, IAS_RPS_LEVEL_GAIN_APPLICATION))
        {
            status = fprintf(dump_fptr, "Radiance Statistics Source:\tL1R "
                "Product\n");
        }
        else
        {
            status = fprintf(dump_fptr, "Radiance Statistics Source:\tBias-"
                "corrected/linearized L0R product\n");
        }
    }
    else
    {
        /* Look just for the '0' in the report header RPS processing level
           string.  Again, it's possible that the processing sequence
           could change in the future. The important thing here is that
           no correction of the L0R data should have occurred when it
           it was first converted from 12-bit integer to floating point.  */
        if (strchr(report_header.proc_level,
            IAS_RPS_LEVEL_CONVERT_FLOAT_NOSHIFT))
        {
            status = fprintf(dump_fptr, "Radiance Statistics Source:\tRaw "
                "L0R shutter product\n");
        }
        else
        {
            IAS_LOG_ERROR("Invalid RPS processing level '%s' for OLI "
                "shutter collect", report_header.proc_level);
            return ERROR;
        }
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing radiance source information to dump file");
        return ERROR;
    }

    /* Now write the relative gain set used to derive the radiance source
       data, whether bias-corrected L0R statistics or L1R product
       statistics */
    if (is_solar_collect)
    {
        char relgain_option_string[15];
        switch (report_header.relgain_option)
        {
            case RPS_RADSTABILITY_PRELAUNCH_RELGAINS:
                sprintf(relgain_option_string, "PRE-LAUNCH");
                break;
            case RPS_RADSTABILITY_POSTLAUNCH_RELGAINS:
                sprintf(relgain_option_string, "POST-LAUNCH");
                break;
            case RPS_RADSTABILITY_CURRENT_RELGAINS:
                sprintf(relgain_option_string, "CURRENT");
                break;
            default:
                sprintf(relgain_option_string, "N/A");
        }
        status = fprintf(dump_fptr, "Relative Gains:\t\t\t%s\n\n\n\n",
            relgain_option_string);
    }
    else
    {
        status = fprintf(dump_fptr, "Relative Gains:\t\t\tN/A\n\n\n\n");
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing relative gain source information to dump "
            "file");
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}
