/**************************************************************************
 NAME:              rad_stability_write_summary_report

 PURPOSE:           Module implementing routines to write custom report
                    header and SCA-level summary stability characterization
                    results for all bands to an optional summary
                    report file

 RETURNS:  Integer status code of SUCCESS or ERROR

***************************************************************************/
#include "ias_miscellaneous.h"
#include "ias_radiometry.h"
#include "ias_satellite_attributes.h"
#include "char_rad_stability.h"


int write_data_to_summary_report
(
    const RPS_RADSTABILITY_DNRAD_BAND_SUMMARY
        *band_summary,                       /* I: Summary stability
                                                information for all bands */
    const char *stability_id,                /* I: Flag indicating whether
                                                bias stability or gain
                                                stability characterization
                                                has been performed */
    int num_oli_bands,                    /* I: Total number of (OLI)
                                                bands to write to
                                                the report */
    double delta_t,                          /* I: Total elapsed time of
                                                collect in seconds
                                                (contained in RPS report
                                                header information) */
    FILE *rpt_fptr                           /* I: Open report file
                                                file */
)
{
    int index;                               /* Band loop counter */
    int band_number;                         /* Current number */
    int sca_index;                           /* SCA loop counter */
    int status;                              /* Function return status */


    /* Write the algorithm-specific header information */
    status = fprintf(rpt_fptr, "                         "
                     "CHARACTERIZATION RESULTS\n\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing results section header to report file");
        return ERROR;
    }


    status = fprintf(rpt_fptr,
            "HISTOGRAM STATISTICS SUMMARY (DN)\n"
            "BAND                       SCA      MAX        MIN      "
            "MEAN     STD DEV\n--------------------------------------"
            "----------------------------------\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing DN histogram statistics header to summary "
                "report");
        return ERROR;
    }

    /* Write the summary histogram statistics (DN) first.  If the array
       element in the band summary data is not populated (indicated
       by 0 for a band number), go to the next OLI band */
    for (index = 0; index < num_oli_bands; index++)
    {
        band_number = band_summary[index].band_number;
        if (band_number == 0)
            continue;

        for (sca_index = 0;
             sca_index < band_summary[index].num_scas;
             sca_index++)
        {
            status = fprintf(rpt_fptr, "%2d (%-19.19s)    %2d  %10.3lf%10.3lf"
                "%10.3lf%10.3lf\n",
                band_summary[index].band_number,
                band_summary[index].band_name,
                band_summary[index].summary_stats[sca_index]
                    .sca_number,
                band_summary[index].summary_stats[sca_index]
                    .sca_max_dn,
                band_summary[index].summary_stats[sca_index]
                    .sca_min_dn,
                band_summary[index].summary_stats[sca_index]
                    .sca_mean_dn,
                band_summary[index].summary_stats[sca_index]
                    .sca_stdev_dn);
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing band %d SCA %d summary DN stats to "
                    "report", band_summary[index].band_number,
                    (sca_index + 1));
                return ERROR;
            }
        }
    }

    /* If we're doing gain stability characterization, write the
       radiance-based histogram statistics summary information */
    if (strcmp(stability_id, "GAIN") == 0)
    {
        /* Write the radiance histogram statistics header information */
        status = fprintf(rpt_fptr,
            "\n\nHISTOGRAM STATISTICS SUMMARY (RAD)\n"
            "BAND                       SCA      MAX        MIN      "
            "MEAN     STD DEV\n--------------------------------------"
            "----------------------------------\n");
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing radiance statistics header to "
                    "report file");
            return ERROR;
        }

        for (index = 0; index < num_oli_bands; index++)
        {
            band_number = band_summary[index].band_number;
            if (band_number == 0)
                continue;

            for (sca_index = 0;
                 sca_index < band_summary[index].num_scas;
                 sca_index++)
            {
                status = fprintf(rpt_fptr, "%2d (%-19.19s)    %2d  "
                    "%10.3lf%10.3lf%10.3lf%10.3lf\n",
                    band_summary[index].band_number,
                    band_summary[index].band_name,
                    band_summary[index].summary_stats[sca_index]
                        .sca_number,
                    band_summary[index].summary_stats[sca_index]
                        .sca_max_rad,
                    band_summary[index].summary_stats[sca_index]
                        .sca_min_rad,
                    band_summary[index].summary_stats[sca_index]
                        .sca_mean_rad,
                    band_summary[index].summary_stats[sca_index]
                        .sca_stdev_rad);
                if (status < 0)
                {
                    IAS_LOG_ERROR("Writing radiance data in band %d, "
                        "SCA %d to report file",
                        band_summary[index].band_number,
                        (sca_index + 1));
                    return ERROR;
                }
            }
        }

        /* Write out the rest of the gain information */

        /* Two-sigma change in signal levels */
        status = fprintf(rpt_fptr,
            "\n\nTWO-SIGMA CHANGES IN SIGNAL LEVEL\n"
            "BAND                       SCA         DN MEAN"
            "   RAD MEAN\n--------------------------------"
            "-------------------------\n");
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing two-sigma header to summary "
                "report");
            return ERROR;
        }

        for (index = 0; index < num_oli_bands; index++)
        {
            band_number = band_summary[index].band_number;
            if (band_number == 0)
                continue;

            /* Now the actual values */
            for (sca_index = 0;
                 sca_index < band_summary[index].num_scas;
                 sca_index++)
            {
                status = fprintf(rpt_fptr, "%2d (%-19.19s)    %2d     "
                    "%10.6lf  %10.6lf\n",
                    band_summary[index].band_number,
                    band_summary[index].band_name,
                    band_summary[index].siglevel_change[sca_index]
                        .sca_number,
                    band_summary[index].siglevel_change[sca_index]
                        .sca_two_sigma_dn,
                    band_summary[index].siglevel_change[sca_index]
                        .sca_two_sigma_rad);
                if (status < 0)
                {
                    IAS_LOG_ERROR("Writing two-sigma information for "
                            "band %d, SCA %d to report file",
                            band_summary[index].band_number,
                            (sca_index + 1));
                    return ERROR;
                }
            }
        }

        /* Percent changes in gain/radiance */
        status = fprintf(rpt_fptr,
            "\n\nPERCENT CHANGES IN SIGNAL LEVEL\n"
            "BAND                      SCA         DN GAIN"
            "    RAD MEAN\n---------------------------------"
            "------------------------\n");
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing percent change header to report file");
            return ERROR;
        }

        for (index = 0; index < num_oli_bands; index++)
        {
            band_number = band_summary[index].band_number;
            if (band_number == 0)
                continue;

            for (sca_index = 0;
                 sca_index < band_summary[index].num_scas;
                 sca_index++)
            {
                status = fprintf(rpt_fptr,
                    "%2d (%-19.19s)    %2d     %10.6lf  %10.6lf\n",
                    band_summary[index].band_number,
                    band_summary[index].band_name,
                    band_summary[index].siglevel_change[sca_index]
                        .sca_number,
                    band_summary[index].siglevel_change[sca_index]
                        .sca_pcnt_change_gain,
                    band_summary[index].siglevel_change[sca_index]
                        .sca_pcnt_change_rad);
                if (status < 0)
                {
                    IAS_LOG_ERROR("Writing percent change information for "
                        "band %d, SCA %d to report file",
                        band_summary[index].band_number,
                        (sca_index + 1));
                    return ERROR;
                }
            }
        }
    }
    else           /* Finish out the bias report */
    {
        /* Write the change in bias signal level(two-sigma) results */
        status = fprintf(rpt_fptr,
            "\n\nTWO-SIGMA CHANGES IN SIGNAL LEVEL\n"
            "BAND                       SCA         DN MEAN\n"
            "----------------------------------------------\n");
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing bias two-sigma header to report file");
            return ERROR;
        }

        for (index = 0; index < num_oli_bands; index++)
        {
            band_number = band_summary[index].band_number;
            if (band_number == 0)
                continue;

            for (sca_index = 0;
                 sca_index < band_summary[index].num_scas;
                 sca_index++)
            {
                status = fprintf(rpt_fptr, "%2d (%-19.19s)    %2d     "
                    "%10.6lf\n",
                    band_summary[index].band_number,
                    band_summary[index].band_name,
                    band_summary[index].siglevel_change[sca_index]
                        .sca_number,
                    band_summary[index].siglevel_change[sca_index]
                        .sca_two_sigma_dn);
                if (status < 0)
                {
                    IAS_LOG_ERROR("Writing bias two-sigma information for "
                            "band %d, SCA %d to report file",
                            band_summary[index].band_number,
                            (sca_index + 1));
                    return ERROR;
                }
            }
        }
    }

    /* Things seem to have been written OK */
    return SUCCESS;
}




/*--------------------------------------------------------------------------
 NAME:                 write_header_to_summary_report

 PURPOSE:  Writes sensor-specific information to the header section
           of the report file

 RETURNS:  Integer status code of SUCCESS or ERROR
----------------------------------------------------------------------------*/
int write_header_to_summary_report
(
    RPS_RADSTABILITY_REPORT_HEADER report_header,   /* I: Report header
                                                       information to
                                                       write */
    FILE *rpt_fptr                                  /* I: report file
                                                       file */
)
{
    char relgain_option_string[15];                 /* Relative gain
                                                       option string */
    char *collection_type = NULL;                   /* Pointer to collection
                                                       type string */
    char *cpf_name = NULL;                          /* Pointer to CPF
                                                       name string */
    char *sensor_id = NULL;                         /* Pointer to sensor ID
                                                       string */

    int status;                                     /* Return status */


    /* When the header information is written to a band's dump file,
       the band number is set.  When this routine is run (when all bands
       to be characterized have been), the last band processed will be
       the current band number.  The header information in the summary
       report shouldn't have a band number given since all bands
       processed will be reported.

       There are two possible ways to resolve this:
       a.  call this routine prior to the main processing
       b.  keep the routine where it is, but call
       ias_misc_set_header_band_number() to "reset" the band_number to an
       "invalid" value.

       Its position in the processing flow seems to make some sense, so
       for now, we'll follow plan b...  The SCA number in the header does
       not get set after initialization, so we'll reset the band number
       to this value.  It will not be printed  */
    ias_misc_set_header_band_number(report_header.common_header,
        ias_misc_get_header_sca_number(report_header.common_header));

    /* Write the standard IAS report header information */
    status = ias_misc_write_report_header(rpt_fptr,
        report_header.common_header);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing common IAS header information to report "
                      "file");
        return ERROR;
    }

    /* Write the RPS-specific header information */
    collection_type = ias_misc_get_header_collection_type(
        report_header.common_header);
    cpf_name = ias_misc_get_header_cpf_name(report_header.common_header);
    sensor_id = ias_misc_get_header_sensor_id(report_header.common_header);

    status = ias_misc_write_rps_report_header(collection_type, cpf_name,
        sensor_id, report_header.rps_header, rpt_fptr);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing RPS-specific header information to report "
            "file");
        return ERROR;
    }

    /* Processing information goes here */
    if (report_header.panel_id == IAS_RPS_OLI_DIFFUSER_TYPE_PRISTINE)
    {
        status = fprintf(rpt_fptr, "Diffuser Panel:        Pristine\n");
    }
    else if (report_header.panel_id == IAS_RPS_OLI_DIFFUSER_TYPE_WORKING)
    {
        status = fprintf(rpt_fptr, "Diffuser Panel:        Working\n");
    }
    else if (report_header.panel_id == IAS_RPS_OLI_DIFFUSER_TYPE_NOTDEPLOYED)
    {
        status = fprintf(rpt_fptr, "Diffuser Panel:        N/A\n");
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Formatting diffuser panel type string");
        return ERROR;
    }

    /* Print the radiance data "source" information.  For gain stability
       characterization, print out the relative gain source used to
       produce the L1R product or the converted bias-corrected/linearized
       histogram statistics */
    switch (report_header.relgain_option)
    {
        case RPS_RADSTABILITY_PRELAUNCH_RELGAINS:
            sprintf(relgain_option_string, "PRE-LAUNCH");
            break;
        case RPS_RADSTABILITY_POSTLAUNCH_RELGAINS:
            sprintf(relgain_option_string, "POST-LAUNCH");
            break;
        case RPS_RADSTABILITY_CURRENT_RELGAINS:
            sprintf(relgain_option_string, "CURRENT");
            break;
        default:
            sprintf(relgain_option_string, "N/A");
    }

    /* Check the RPS processing level string to identify the
       source product for the histogram statistics used in the
       characterization. */
    if (strcmp(collection_type, IAS_RPS_OLI_SOLAR_COLLECT) == 0)
    {
        /* Use 'strchr' to look for 'G' in the report header RPS processing
           level string indicating that gain application has been applied.
           This allows for the possibility (however remote) that the
           processing sequence could change from what's currently used. */
        if (strchr(report_header.proc_level, IAS_RPS_LEVEL_GAIN_APPLICATION))
        {
            status = fprintf(rpt_fptr, "Characterization data source:  L1R "
                "solar product\n");
        }
        else
        {
            status = fprintf(rpt_fptr, "Characterization data source:  Bias-"
                "corrected/linearized L0R solar product\n");
        }
    }
    else if (strcmp(collection_type, IAS_RPS_OLI_SHUTTER_COLLECT) == 0)
    {
        /* Look just for the '0' in the report header RPS processing level
           string.  Again, it's possible that the processing sequence
           could change in the future.  The important thing here is that
           no correction of the L0R data should have occurred when it
           it was first converted from 12-bit integer to floating point.  */
        if (strchr(report_header.proc_level,
        IAS_RPS_LEVEL_CONVERT_FLOAT_NOSHIFT))
        {
            status = fprintf(rpt_fptr, "Characterization data source:  Raw "
                "L0R shutter product\n");
        }
        else
        {
            IAS_LOG_ERROR("Invalid RPS processing level '%s' for OLI "
                "shutter collect", report_header.proc_level);
            return ERROR;
        }
    }
    else
    {
        IAS_LOG_ERROR("Shouldn't have gotten here, invalid collection type "
            "'%s'", collection_type);
        return ERROR;
    }

    if (status < 0)
    {
        IAS_LOG_ERROR("Writing radiance data source information to "
            "report file");
        return ERROR;
    }
    status = fprintf(rpt_fptr, "CPF relative gain set used to derive "
        "radiances:        %s\n", relgain_option_string);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing relative relative gain information to "
            "summary_report");
        return ERROR;
    }

    status = fprintf(rpt_fptr, "Delta T (s):          %7.3lf\n",
                     report_header.stability_delta_t);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing work order stability delta T info to "
                      "report file");
        return ERROR;
    }
    status = fprintf(rpt_fptr, "Stability period (s): %7.3lf\n",
                     report_header.stability_period);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing observed stability period to report file");
        return ERROR;
    }

    /* Write the separator string */
    status = fprintf(rpt_fptr, "==================================="
                     "========================================\n\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing separator string to report file");
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}
