/***************************************************************************
 NAME:                      rad_stability_assess_stability_test

 PURPOSE:  CGreen unit testing of the assess_bias_stability() and
           assess_gain_stability() routines used by the char_rad_stability
           application

 RETURNS:  CGreen status

***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <float.h>

#include "ias_const.h"
#include "ias_cpf.h"
#include "cgreen.h"

#include "../char_rad_stability.h"


static char test_data_dir[] = "ias/rps/rad_stability";
static char test_data_filename[PATH_MAX] = "";

Ensure test_assess_bias_stability()
{
    char *test_data_path = NULL;
    const char shutter_basename[] = "shutter_stats_gooddets.txt";

    int status;
    int num_sca_detectors = 100;
    int detector_index;
    int dummy_detector;

    double *shutter_stdevs = NULL;
    double *two_sigma_det_buffer = NULL;
    double dummy_shutter_min;
    double dummy_shutter_max;
    double dummy_shutter_mean;
    double two_sigma_sca_mean;           /* Mean two-sigma value for all
                                            operable detectors in the SCA */
    double test_Qbar = 0.637339;         /* ADD value given for mean
                                            two-sigma across all "good"
                                            detectors */
    /* NOTE:  In the prototype code, detectors that had ANY pixel flagged
       in the simulated "pixel mask" were excluded from
       the SCA averaging calculations.  In this case, the "affected"
       detectors are (1-based numbers):

       {4, 7, 15, 21, 22, 23, 24, 25, 31, 46, 47, 61}  */


    int *statistics_invalid_flags = NULL;

    FILE *stats_fptr = NULL;


    printf("Testing bias stability assessment...\n");

    /* Get test data directory */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(test_data_path != NULL);
    if (test_data_path == NULL)
    {
        assert_true_with_message(FALSE, "Cannot locate parent test data "
            "directory");
        return;
    }

    /* Allocate the required output data buffers */
    shutter_stdevs = malloc(num_sca_detectors * sizeof(*shutter_stdevs));

    two_sigma_det_buffer = malloc(num_sca_detectors
         * sizeof(*two_sigma_det_buffer));

    statistics_invalid_flags = calloc(num_sca_detectors, sizeof(int));

    assert_true((two_sigma_det_buffer != NULL) && (shutter_stdevs != NULL)
        && (statistics_invalid_flags != NULL));

    if ((shutter_stdevs == NULL) || (two_sigma_det_buffer == NULL)
        || (statistics_invalid_flags == NULL))
    {
        printf("Cannot allocate required data buffers\n");
        free(two_sigma_det_buffer);
        free(shutter_stdevs);
        free(statistics_invalid_flags);
        return;
    }

    /* Read in the bias statistics */
    sprintf(test_data_filename, "%s/%s/%s", test_data_path, test_data_dir,
        shutter_basename);
    stats_fptr = fopen(test_data_filename, "r");
    assert_true(stats_fptr != NULL);
    if (stats_fptr == NULL)
    {
        printf("Cannot retrieve test statistics from file %s\n",
                test_data_filename);
        free(two_sigma_det_buffer);
        free(shutter_stdevs);
        free(statistics_invalid_flags);
        return;
    }

    for (detector_index = 0;
         detector_index < num_sca_detectors;
         detector_index++)
    {
        fscanf(stats_fptr, "%d%lf%lf%lf%lf\n", &dummy_detector,
                &dummy_shutter_max, &dummy_shutter_min,
                &dummy_shutter_mean, &(shutter_stdevs[detector_index]));
        if (shutter_stdevs[detector_index] < 0)
            statistics_invalid_flags[detector_index] = 1;
    }
    fclose(stats_fptr);

    /* Assess signal stability */
    status = assess_signal_stability(shutter_stdevs, statistics_invalid_flags,
        num_sca_detectors, two_sigma_det_buffer, &two_sigma_sca_mean);
    assert_true(status == SUCCESS);

    /* Compare the computed mean two-sigma value with the ADD-specified
       test value */
    assert_true_with_message(
        (fabs(two_sigma_sca_mean - test_Qbar) <= 1e-6),
        "Computed two-sigma mean %lf not equal to ADD-specified value %lf",
        two_sigma_sca_mean, test_Qbar);


    /* Clean up */
    free(two_sigma_det_buffer);
    free(shutter_stdevs);
    free(statistics_invalid_flags);
}




Ensure test_assess_gain_stability()
{
    char *test_data_path = NULL;

    int status;
    int num_sca_detectors = 100;
    int detector_index;
    int dummy_detector;

    double *solar_dn_means = NULL;
    double *solar_dn_stdevs = NULL;
    double *relgains = NULL;
    double *rg_ptr = NULL;
    double *two_sigma_dn_det_buffer  = NULL;
    double *pcnt_change_gain_det_buffer = NULL;

    double dummy_solar_dn_min;
    double dummy_solar_dn_max;
    double sca_two_sigma_dn_mean;
    double sca_pcnt_change_gain_mean;
    double test_Qbar_gain = 8.39776;     /* ADD values given for mean
                                            two-sigma across all "good"
                                            detectors */
    double test_Qbar_gain_pcnt = 0.4;    /* ADD-specified mean % change in
                                            gain across the SCA */

    /* NOTE:  In the prototype code, detectors that had ANY pixel flagged
       in the simulated "pixel mask" were excluded from
       the SCA averaging calculations.  In this case, the "affected"
       detectors are (1-based numbers):

       {4, 7, 15, 21, 22, 23, 24, 25, 31, 46, 47, 61}  */

    int *statistics_invalid_flags = NULL;
    FILE *stats_fptr = NULL;

    /* Text file with test bias statistics for each operable detector.
       The excluded detectors have been flagged with -150.0, the #defined
       const RPS_RADSTABILITY_INVALID_STATISTICS     */
    const char solar_dn_basename[] = "solar_dn_stats_gooddets.txt";


    printf("Testing gain stability assessment...\n");

   /* Get test data directory */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(test_data_path != NULL);
    if (test_data_path == NULL)
    {
        assert_true_with_message(FALSE, "Cannot locate parent test data "
            "directory");
        return;
    }


    /* Allocate the required input data buffers */
    solar_dn_means = malloc(num_sca_detectors * sizeof(*solar_dn_means));
    solar_dn_stdevs = malloc(num_sca_detectors * sizeof(*solar_dn_stdevs));
    relgains = malloc(num_sca_detectors * sizeof(*relgains));
    statistics_invalid_flags = calloc(num_sca_detectors, sizeof(int));

    assert_true((solar_dn_stdevs != NULL) && (solar_dn_means != NULL)
            && (relgains != NULL) && (statistics_invalid_flags != NULL));

    if ((solar_dn_stdevs == NULL) || (solar_dn_means == NULL)
            || (relgains == NULL) || (statistics_invalid_flags == NULL))
    {
        printf("Cannot allocate required data buffers\n");
        free(solar_dn_stdevs);
        free(solar_dn_means);
        free(relgains);
        free(statistics_invalid_flags);
        return;
    }

    /* Allocate the required output data buffers */
    two_sigma_dn_det_buffer = malloc(num_sca_detectors
             * sizeof(*two_sigma_dn_det_buffer));

    pcnt_change_gain_det_buffer = malloc(num_sca_detectors
            * sizeof(*pcnt_change_gain_det_buffer));

    assert_true((two_sigma_dn_det_buffer != NULL)
            && (pcnt_change_gain_det_buffer != NULL));

    if ((two_sigma_dn_det_buffer == NULL)
            || (pcnt_change_gain_det_buffer == NULL))
    {
        printf("Cannot allocate required data buffers\n");
        free(two_sigma_dn_det_buffer);
        free(pcnt_change_gain_det_buffer);
        free(solar_dn_stdevs);
        free(solar_dn_means);
        free(relgains);
        free(statistics_invalid_flags);
        return;
    }

    /* Read in the solar (DN-based) statistics */
    sprintf(test_data_filename, "%s/%s/%s", test_data_path, test_data_dir,
        solar_dn_basename);
    stats_fptr = fopen(test_data_filename, "r");
    assert_true(stats_fptr != NULL);
    if (stats_fptr == NULL)
    {
        printf("Cannot retrieve test statistics from file %s\n",
            test_data_filename);
        free(two_sigma_dn_det_buffer);
        free(pcnt_change_gain_det_buffer);
        free(solar_dn_stdevs);
        free(solar_dn_means);
        free(relgains);
        free(statistics_invalid_flags);
        return;
    }

    for (detector_index = 0;
         detector_index < num_sca_detectors;
         detector_index++)
    {
        if (fscanf(stats_fptr, "%d%lf%lf%lf%lf%lf\n",
                &dummy_detector, &dummy_solar_dn_max,
                &dummy_solar_dn_min, &solar_dn_means[detector_index],
                &solar_dn_stdevs[detector_index],
                &relgains[detector_index]) < 0)
        {
            printf("Cannot read statistics data for detector %d\n",
                   (detector_index + 1));
            free(two_sigma_dn_det_buffer);
            free(pcnt_change_gain_det_buffer);
            free(solar_dn_stdevs);
            free(solar_dn_means);
            free(relgains);
            free(statistics_invalid_flags);
            return;
        }

        if (solar_dn_stdevs[detector_index] < 0)
            statistics_invalid_flags[detector_index] = 1;
    }

    fclose(stats_fptr);

    /* Assess signal stability */
    status = assess_signal_stability(solar_dn_stdevs,
        statistics_invalid_flags, num_sca_detectors, two_sigma_dn_det_buffer,
        &sca_two_sigma_dn_mean);

    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Gain stability assessment failed\n");
        free(two_sigma_dn_det_buffer);
        free(pcnt_change_gain_det_buffer);
        free(solar_dn_stdevs);
        free(solar_dn_means);
        free(relgains);
        free(statistics_invalid_flags);
        return;
    }

    /* Compare the computed mean two-sigma value with the ADD-specified
       test value */
    assert_true_with_message(
        (fabs(sca_two_sigma_dn_mean - test_Qbar_gain) <= 1e-3),
        "Computed two-sigma mean %lf not equal to ADD-specified value %lf",
        sca_two_sigma_dn_mean, test_Qbar_gain);

    /* Determine the stability in gain, then compare it to the ADD
       estimate */
    rg_ptr = relgains;

    status = assess_gain_stability(solar_dn_means,
            two_sigma_dn_det_buffer,
            sca_two_sigma_dn_mean,
            rg_ptr,
            statistics_invalid_flags,
            num_sca_detectors,
            pcnt_change_gain_det_buffer,
            &sca_pcnt_change_gain_mean);

    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot assess gain stability\n");
        free(two_sigma_dn_det_buffer);
        free(pcnt_change_gain_det_buffer);
        free(solar_dn_stdevs);
        free(solar_dn_means);
        free(relgains);
        free(statistics_invalid_flags);
        return;
    }

    assert_true_with_message(
        (fabs(sca_pcnt_change_gain_mean - test_Qbar_gain_pcnt) <= 1e-4),
        "Computed two-sigma mean %lf not equal to ADD-specified value %lf",
        sca_pcnt_change_gain_mean, test_Qbar_gain_pcnt);

    /* Clean up */
    free(two_sigma_dn_det_buffer);
    free(pcnt_change_gain_det_buffer);
    free(solar_dn_stdevs);
    free(solar_dn_means);
    free(relgains);
    free(statistics_invalid_flags);
}




/*-----------------------------------------------------------------------
           Define a test suite to add to the "master" test suite
-------------------------------------------------------------------------*/
void rad_stability_assess_stability_test(TestSuite *suite)
{
    add_test(suite, test_assess_bias_stability);
    add_test(suite, test_assess_gain_stability);
}
