/**************************************************************************
                        rad_stability_dbaccess_test

 PURPOSE:      Use CGreen to perform unit testing of the various
               component functions making up OLI radiometric stability
               characterization database access

 RETURNS:      CGreen test status

***************************************************************************/
#include <math.h>

#include "ias_const.h"
#include "ias_db.h"
#include "ias_db_insert.h"
#include "ias_db_query.h"
#include "ias_miscellaneous.h"
#include "ias_pixel_mask.h"
#include "ias_radiometry.h"

#include "cgreen.h"

#include "../char_rad_stability.h"

#include "rad_stability_tests.h"

#define SQL_QUERYLEN 256

static char work_order_id[IAS_WORKORDER_ID_SIZE] = "RST1002";
static char proc_level[PROC_LEVEL_LENGTH] = "";
static char system_command[1000] = "";
static char zero_level_str[] = {IAS_RPS_LEVEL_CONVERT_FLOAT_NOSHIFT, '\0'};

/* Useful cleanup functions */
static void free_detector_buffers(double *hist_stats_max_buffer,
                                  double *hist_stats_min_buffer,
                                  double *hist_stats_mean_buffer,
                                  double *hist_stats_stdev_buffer,
                                  int *statistics_invalid_flags
)
{
    free(hist_stats_max_buffer);
    free(hist_stats_min_buffer);
    free(hist_stats_mean_buffer);
    free(hist_stats_stdev_buffer);
    free(statistics_invalid_flags);
}


Ensure test_get_histogram_statistics()
{
    char rad_wo_common_id[IAS_CHAR_ID_SIZE];  
    int inoperable_detectors[2]   = {5, 174};
    int outofspec_detectors[2]    = {31, 210};
    int *statistics_invalid_flags = NULL;
    int num_inoperable_detectors = 2;
    int num_outofspec_detectors  = 2;
    int band_number = 1;
    int sca_number  = 1;
    int invalid_detector_count = 0;
    int max_smaller_count = 0;
    int num_detectors_sca = 494;
    int num_frames = 500;
    int detector_index;
    int status;
    double *hist_stats_max_buffer = NULL;
    double *hist_stats_min_buffer = NULL;
    double *hist_stats_mean_buffer = NULL;
    double *hist_stats_stdev_buffer = NULL;
    IAS_PIXEL_MASK *pm = NULL;
    struct ias_db_connection *open_db = NULL;


    printf("Testing histogram statistics retrieval...\n");

    /* Initialize processing buffers */
    strcpy(proc_level, zero_level_str);

    /* Allocate detector-level data buffers */
    hist_stats_max_buffer = malloc(num_detectors_sca
             * sizeof(*hist_stats_max_buffer));

    hist_stats_min_buffer = malloc(num_detectors_sca
             * sizeof(*hist_stats_min_buffer));

    hist_stats_mean_buffer = malloc(num_detectors_sca
             * sizeof(*hist_stats_mean_buffer));

    hist_stats_stdev_buffer = malloc(num_detectors_sca
             * sizeof(*hist_stats_stdev_buffer));

    statistics_invalid_flags = malloc(num_detectors_sca * sizeof(int));

    if ((hist_stats_max_buffer == NULL)
            || (hist_stats_min_buffer == NULL)
            || (hist_stats_mean_buffer == NULL)
            || (hist_stats_stdev_buffer == NULL)
            || (statistics_invalid_flags == NULL))
    {
        assert_true_with_message(FALSE, "Cannot allocate detector-level "
            "statistics buffers");
        free_detector_buffers(hist_stats_max_buffer,
                hist_stats_min_buffer,
                hist_stats_mean_buffer,
                hist_stats_stdev_buffer,
                statistics_invalid_flags);
        return;
    }

    /* Create the pixel mask, assuming 500 frames */
    pm = ias_pm_create(band_number, sca_number, num_detectors_sca,
        num_frames);
    if (pm == NULL)
    {
        assert_true_with_message(FALSE, "Cannot create pixel mask for "
                "testing band %d SCA %d", band_number, sca_number);
        free_detector_buffers(hist_stats_max_buffer, hist_stats_min_buffer,
            hist_stats_mean_buffer, hist_stats_stdev_buffer,
            statistics_invalid_flags);
        return;
    }

    /* Identify the inoperable detectors */
    status = ias_rps_inoperable_detector_identification(inoperable_detectors,
        num_inoperable_detectors, band_number, sca_number,
        num_detectors_sca, num_frames, pm);
    assert_equal_with_message(status, SUCCESS, "Error adding "
        "inoperable detector information to mask in band 1 SCA 1");
    if (status != SUCCESS)
    {
        free_detector_buffers(hist_stats_max_buffer, hist_stats_min_buffer,
            hist_stats_mean_buffer, hist_stats_stdev_buffer,
            statistics_invalid_flags);
        ias_pm_destroy(pm);
        return;
    }

    /* Identify the out-of-spec detectors */
    status = ias_rps_outofspec_detector_identification(outofspec_detectors,
        num_outofspec_detectors, band_number, sca_number,
        num_detectors_sca, num_frames, pm);
    assert_equal_with_message(status, SUCCESS, "Error adding "
        "out-of-spec detector information to mask in band 1 SCA 1");
    if (status != SUCCESS)
    {
        free_detector_buffers(hist_stats_max_buffer, hist_stats_min_buffer,
            hist_stats_mean_buffer, hist_stats_stdev_buffer,
            statistics_invalid_flags);
        ias_pm_destroy(pm);
        return;
    }

    /* Get the flagged detector information into the invalid data buffer */
    status = ias_rps_get_invalid_data_flags(pm,
        IAS_RPS_INOP_DETECTOR | IAS_RPS_OUTOFSPEC_DETECTOR,
        num_detectors_sca, statistics_invalid_flags);
    ias_pm_destroy(pm);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error populating detector "
           "invalid flag data buffer");
        free_detector_buffers(hist_stats_max_buffer, hist_stats_min_buffer,
            hist_stats_mean_buffer, hist_stats_stdev_buffer,
            statistics_invalid_flags);
        return;
    }

    /* Prep the database in case there's no data in it associated with
       the specified work order string */
    sprintf(system_command, "./check_or_insert_woid.csh %s",
        work_order_id);
    status = system(system_command);
    assert_equal_with_message(status, SUCCESS, "Error preparing "
        "database:  cannot insert work order ID %s and/or its test "
        "data", work_order_id);
    if (status != 0)
    {
        free_detector_buffers(hist_stats_max_buffer, hist_stats_min_buffer,
            hist_stats_mean_buffer, hist_stats_stdev_buffer,
            statistics_invalid_flags);
        return;
     }

     /* Now connect to the database */
    open_db = connect_to_db();
    if (open_db == NULL)
    {
        assert_true_with_message(FALSE, "Error connecting to DB");
        ias_db_close_database_lib();
        free_detector_buffers(hist_stats_max_buffer, hist_stats_min_buffer,
            hist_stats_mean_buffer, hist_stats_stdev_buffer,
            statistics_invalid_flags);
        return;
    }

    /* Get the radiometric work order ID key for the specified work order
       string */
    status = ias_misc_get_char_id(open_db, "RADIOMETRIC_WORK_ORDER_COMMON",
        "radiometric_wo_common_id", work_order_id, rad_wo_common_id);
    if (status == ERROR)
    {
        assert_true_with_message(FALSE, "Invalid rad_wo_common_id for "
            "specified work order ID %s", work_order_id);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        free_detector_buffers(hist_stats_max_buffer, hist_stats_min_buffer,
            hist_stats_mean_buffer, hist_stats_stdev_buffer,
            statistics_invalid_flags);
        return;
    }

    /* Get the histogram statistics */
    status = get_histogram_statistics(open_db, rad_wo_common_id,
        IAS_SPECTRAL_VNIR, band_number, sca_number,
        num_detectors_sca, proc_level, hist_stats_max_buffer,
        hist_stats_min_buffer, hist_stats_mean_buffer,
        hist_stats_stdev_buffer);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Cannot obtain histogram statistics "
            "for band %d SCA %d\n", band_number, sca_number);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        free_detector_buffers(hist_stats_max_buffer, hist_stats_min_buffer,
            hist_stats_mean_buffer, hist_stats_stdev_buffer,
            statistics_invalid_flags);
        return;
    }

    /* Data sanity check #1a -- the inoperable detectors have to be flagged
       with invalid data */
    assert_equal(statistics_invalid_flags[4], 1);
    assert_equal(statistics_invalid_flags[173], 1);

    /* Data sanity check #1b -- the out-of-spec detectors have to be flagged
       with invalid data */
    assert_equal(statistics_invalid_flags[30], 1);
    assert_equal(statistics_invalid_flags[209], 1);

    /* Data sanity check #2 -- the maximum value has to be larger than the
       minimum and the mean values.  Do this check only for detectors
       not flagged as invalid (i.e. those not identified as inoperable
       or out-of-spec) */
    for (detector_index = 0;
         detector_index < num_detectors_sca;
         detector_index++)
    {
        if (statistics_invalid_flags[detector_index])
        {
            invalid_detector_count++;
        }
        else
        {
            if ((hist_stats_max_buffer[detector_index]
                < hist_stats_min_buffer[detector_index])
                || (hist_stats_max_buffer[detector_index]
                   < hist_stats_mean_buffer[detector_index]))
            {
                max_smaller_count++;
            }
        }
    }

    /* There should be four detectors with flagged data (2 inoperable and
       2 out-of-spec detectors).  Check invalid_detector_count to confirm
       this */
    assert_equal_with_message(invalid_detector_count,
        (num_inoperable_detectors + num_outofspec_detectors),
        "Error in flagging inoperable/out-of-spec detectors in "
        "histogram statistics processing");

    assert_equal_with_message(max_smaller_count, 0,
        "Error in get_histogram_statistics : %d instances where the "
        "maximimum is smaller than the minimum or the mean",
        max_smaller_count);

    /* Clean up */
    ias_db_close_connection(open_db);
    ias_db_close_database_lib();
    free_detector_buffers(hist_stats_max_buffer, hist_stats_min_buffer,
        hist_stats_mean_buffer, hist_stats_stdev_buffer,
        statistics_invalid_flags);
}




Ensure test_write_detector_gain_stability_results_to_db()
{
    char query_string[SQL_QUERYLEN];
    char db_error_message[IAS_SQL_RTN_MSGLEN];

    struct ias_db_connection *open_db = NULL;
    struct ias_db_query *query = NULL;

    char rad_wo_common_id[IAS_CHAR_ID_SIZE];  
    int detector_index;
    int sca_index;
    int status;
    int band_number = 1;
    int num_scas = 14;
    int num_detectors_sca = 494;
    int expected_detector_count;
    int db_detector_count;

    int *detector_ids = NULL;
    IAS_DB_NULL_TYPE *flagged_detector_ids = NULL;
    int null_db_flag = -1;

    double *pcnt_change_gain_det_buffer = NULL;

    printf("Testing write of detector gain stability results to "
           "database...\n");

    /* Establish database connection and get the radiometric work order
       common database ID */
    open_db = connect_to_db();
    if (open_db == NULL)
    {
        assert_true_with_message(FALSE, "Error connecting to DB");
        ias_db_close_database_lib();
        return;
    }

    status = ias_misc_get_char_id(open_db,
            "RADIOMETRIC_WORK_ORDER_COMMON",
            "radiometric_wo_common_id",
            work_order_id, rad_wo_common_id);
    if (status == ERROR)
    {
        assert_true_with_message(FALSE, "Invalid rad_wo_common_id "
            "for specified work order ID %s",
            work_order_id);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        return;
    }

    /* Clear the gain stability detector table if necessary */
    status = clear_radstability_table(open_db, rad_wo_common_id,
            "gain_stability_60s_detector");
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error clearing "
            "gain_stability_60s_detector table");
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        return;
    }

    /* Allocate the data buffers */
    detector_ids = malloc(num_detectors_sca * sizeof(*detector_ids));

    flagged_detector_ids = malloc(num_detectors_sca
            * sizeof(*flagged_detector_ids));

    pcnt_change_gain_det_buffer = malloc(num_detectors_sca
            * sizeof(*pcnt_change_gain_det_buffer));

    if ((detector_ids == NULL) || (flagged_detector_ids == NULL)
            || (pcnt_change_gain_det_buffer == NULL))
    {
        assert_true_with_message(FALSE, "Error allocating data buffer");
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        free(detector_ids);
        free(flagged_detector_ids);
        free(pcnt_change_gain_det_buffer);
        return;
    }

    /* Initialize detector IDs */
    for (detector_index = 0;
         detector_index < num_detectors_sca;
         detector_index++)
    {
        detector_ids[detector_index] = (detector_index + 1);
    }

    /* Make up detector data to write to database, then write it */
    for (sca_index = 0; sca_index < num_scas; sca_index++)
    {
        memset(flagged_detector_ids,
               0,
               (num_detectors_sca * sizeof(*flagged_detector_ids)));

        memset(pcnt_change_gain_det_buffer,
               0,
               (num_detectors_sca * sizeof(*pcnt_change_gain_det_buffer)));

        for (detector_index = 0;
             detector_index < num_detectors_sca;
             detector_index++)
        {
            /* Flag every 50 detectors in every odd SCA as NULL */
            if (((sca_index % 2) == 0) && ((detector_index % 50) == 0))
            {
                flagged_detector_ids[detector_index] = null_db_flag;
            }

            /* Fake % gain change */
            pcnt_change_gain_det_buffer[detector_index] = 0.1234567
                + (0.01 * sca_index);
        }

        /* Write detector data to database */
        status = write_detector_gain_stability_results_to_db(open_db,
                    rad_wo_common_id,
                    detector_ids, flagged_detector_ids,
                    pcnt_change_gain_det_buffer,
                    band_number, (sca_index + 1),
                    num_detectors_sca);
        assert_true(status == SUCCESS);
    }

    /* Partial cleanup */
    free(detector_ids);
    free(flagged_detector_ids);
    free(pcnt_change_gain_det_buffer);


    /* Check how many NULLed values were written in a consecutive pair
       of SCAs.  As a representative check, look at SCAs 1 and 2 */

    /* SCA 1  first.  Every 50th detector was flagged
       (e.g. 1, 51, 101, 151, 201, 251, 301, 351, 401, 451) */
    expected_detector_count = num_detectors_sca - 10;

    sprintf(query_string, "SELECT count(*) FROM GAIN_STABILITY_60S_DETECTOR "
            "WHERE radiometric_wo_common_id=%s AND band_id=%d AND "
            "sensor_chip_assembly_id=%d AND "
            "change_in_gain_mean is not null",
            rad_wo_common_id, band_number, 1);

    query = ias_db_perform_query(open_db, query_string);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
                sizeof(db_error_message));
        printf("Query of HISTOGRAM STATISTICS failed:  %s\n",
                db_error_message);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        return;
    }

    /* Move DB "pointer" to the first valid record. */
    if (!ias_db_query_next(query))
    {
        printf("Accessing first data record\n");
        ias_db_query_close(query);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
    }

    status = ias_db_query_int_value(query, &db_detector_count, 0);
    if (status != SUCCESS)
    {
        printf("Accessing database detector count value\n");
        ias_db_query_close(query);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
    }

    /* Compare the counts */
    assert_equal(expected_detector_count, db_detector_count);

    /* Free query resources */
    ias_db_query_close(query);

    /* Now do SCA 2--All detectors should have a value written to them */
    expected_detector_count = num_detectors_sca;

    sprintf(query_string, "SELECT count(*) FROM GAIN_STABILITY_60S_DETECTOR "
            "WHERE radiometric_wo_common_id=%s AND band_id=%d AND "
            "sensor_chip_assembly_id=%d AND "
            "change_in_gain_mean is not null",
            rad_wo_common_id, band_number, 2);

    query = ias_db_perform_query(open_db, query_string);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
                sizeof(db_error_message));
        printf("Query of HISTOGRAM STATISTICS failed:  %s\n",
                db_error_message);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        return;
    }

    /* Move DB "pointer" to the first valid record. */
    if (!ias_db_query_next(query))
    {
        printf("Accessing first data record\n");
        ias_db_query_close(query);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
    }

    status = ias_db_query_int_value(query, &db_detector_count, 0);
    if (status != SUCCESS)
    {
        printf("Accessing database detector count value\n");
        ias_db_query_close(query);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
    }

    /* Compare the counts */
    assert_equal(expected_detector_count, db_detector_count);

    /* Free query resources */
    ias_db_query_close(query);

    /* Clean up */
    ias_db_close_connection(open_db);
    ias_db_close_database_lib();
}




Ensure test_write_sca_bias_stability_results_to_db()
{
    struct ias_db_connection *open_db = NULL;

    int band_number = 1;
    int num_scas = 14;
    int sca_index;
    int status;
    char rad_wo_common_id[IAS_CHAR_ID_SIZE];  

    RPS_RADSTABILITY_DNRAD_BAND_SUMMARY band_summary;


    printf("Testing write of SCA-level bias stability results to "
           "database...\n");

    /* Establish database connection */
    open_db = connect_to_db();
    assert_true(open_db != NULL);
    if (open_db == NULL)
    {
        assert_true_with_message(FALSE, "Error connecting to DB");
        ias_db_close_database_lib();
        return;
    }

    /* Get the radiometric work order ID number from the database */
    status = ias_misc_get_char_id(open_db,
            "RADIOMETRIC_WORK_ORDER_COMMON",
            "radiometric_wo_common_id",
            work_order_id, rad_wo_common_id);
    if (status == ERROR)
    {
        assert_true_with_message(FALSE, "Invalid rad_wo_common_id key "
            "for specified work order ID %s",
            work_order_id);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        return;
    }

    /* Clear bias stability table if necessary */
    status = clear_radstability_table(open_db, rad_wo_common_id,
                                      "bias_stability_60s_sca");
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot clear bias stability table\n");
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        return;
    }

    /* Allocate band summary data structures */
    memset(&band_summary, 0, sizeof(RPS_RADSTABILITY_DNRAD_BAND_SUMMARY));
    band_summary.band_number = band_number;
    band_summary.num_scas    = num_scas;

    band_summary.siglevel_change = calloc(band_summary.num_scas,
                 sizeof(RPS_RADSTABILITY_DNRAD_SCA_SUMMARY_SIGLEVEL_CHANGE));
    if (band_summary.siglevel_change == NULL)
    {
        assert_true_with_message(FALSE, "Error allocating band summary data "
            "structure");
        free(band_summary.siglevel_change);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        return;
    }

    /* Populate "fake" SCA-level bias stability results to write to DB,
       then write them to the DB */
    for (sca_index = 0; sca_index < num_scas; sca_index++)
    {
        band_summary.siglevel_change[sca_index].sca_number
            = (sca_index + 1);
        band_summary.siglevel_change[sca_index].sca_two_sigma_dn
            = 0.857143;

        status = write_sca_bias_stability_results_to_db(open_db,
                rad_wo_common_id,
                band_summary.siglevel_change[sca_index],
                band_number,
                1);

        assert_true(status == SUCCESS);
        if (status != SUCCESS)
        {
            assert_true_with_message(FALSE, "Error writing SCA-level "
                "bias stability results to DB for SCA %d",
                (sca_index + 1));
        }
    }

    /* Clean up */
    free(band_summary.siglevel_change);
    ias_db_close_connection(open_db);
    ias_db_close_database_lib();
}




Ensure test_write_sca_gainrad_stability_results_to_db()
{
    struct ias_db_connection *open_db = NULL;

    int band_number = 1;
    int num_scas = 14;
    int sca_index;
    int status;
    char rad_wo_common_id[IAS_CHAR_ID_SIZE];  

    RPS_RADSTABILITY_DNRAD_BAND_SUMMARY band_summary;


    printf("Testing write of SCA-level gain/radiance stability results to "
           "database...\n");

    /* Establish database connection */
    open_db = connect_to_db();
    if (open_db == NULL)
    {
        assert_true_with_message(FALSE, "Error connecting to DB");
        ias_db_close_database_lib();
        return;
    }

    /* Get the radiometric work order ID number from the database */
    status = ias_misc_get_char_id(open_db,
            "RADIOMETRIC_WORK_ORDER_COMMON",
            "radiometric_wo_common_id",
            work_order_id, rad_wo_common_id);
    if (status == ERROR)
    {
        assert_true_with_message(FALSE, "Invalid rad_wo_common_id key "
             "for specified work order ID %s", work_order_id);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        return;
    }

    /* Clear bias stability table if necessary */
    status = clear_radstability_table(open_db, rad_wo_common_id,
                                      "gain_stability_60s_sca");
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot clear gain stability table\n");
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        return;
    }

    /* Allocate band summary data structures */
    memset(&band_summary, 0, sizeof(RPS_RADSTABILITY_DNRAD_BAND_SUMMARY));

    band_summary.band_number = band_number;
    band_summary.num_scas    = num_scas;

    band_summary.siglevel_change = calloc(band_summary.num_scas,
                 sizeof(RPS_RADSTABILITY_DNRAD_SCA_SUMMARY_SIGLEVEL_CHANGE));

    assert_true(band_summary.siglevel_change != NULL);
    if (band_summary.siglevel_change == NULL)
    {
        printf("Cannot allocate band summary data structure\n");
        free(band_summary.siglevel_change);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        return;
    }

    /* Populate "fake" SCA-level bias stability results to write to DB,
       then write them to the DB */
    for (sca_index = 0; sca_index < num_scas; sca_index++)
    {
        band_summary.siglevel_change[sca_index].sca_number
            = (sca_index + 1);
        band_summary.siglevel_change[sca_index].sca_two_sigma_dn
            = 0.857143;
        band_summary.siglevel_change[sca_index].sca_two_sigma_rad
            = 0.452316;
        band_summary.siglevel_change[sca_index].sca_pcnt_change_gain
            = 0.165814;
        band_summary.siglevel_change[sca_index].sca_pcnt_change_rad
            = 0.089347;

        status = write_sca_gainrad_stability_results_to_db(open_db,
                rad_wo_common_id,
                band_summary.siglevel_change[sca_index],
                band_number, 1);

        assert_true(status == SUCCESS);
        if (status != SUCCESS)
        {
            printf("Cannot write SCA-level bias stability results to "
                    "characterization database for SCA %d\n",
                    (sca_index + 1));
        }
    }

    /* Clean up */
    free(band_summary.siglevel_change);
    ias_db_close_connection(open_db);
    ias_db_close_database_lib();
}



Ensure test_write_to_bias_stability_60s()
{
    struct ias_db_connection *open_db = NULL;

    char rad_wo_common_id[IAS_CHAR_ID_SIZE];  
    int status;

    double stability_delta_t = 60.0;


    printf("Testing write to bias_stability_60s table...\n");

    /* Initialize the database strings */
    strcpy(proc_level, zero_level_str);

    /* Open a connection to the database */
    open_db = connect_to_db();
    if (open_db == NULL)
    {
        assert_true_with_message(FALSE, "Error connecting to DB");
        ias_db_close_database_lib();
        return;
    }

    /* Get the radiometric work order ID number from the database */
    status = ias_misc_get_char_id(open_db,
            "RADIOMETRIC_WORK_ORDER_COMMON",
            "radiometric_wo_common_id",
            work_order_id, rad_wo_common_id);
    if (status == ERROR)
    {
        assert_true_with_message(FALSE, "Invalid rad_wo_common_id key "
             "for specified work order ID %s", work_order_id);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        return;
    }

    /* Clear bias_stability_60s table if necessary */
    status = clear_radstability_table(open_db, rad_wo_common_id,
                                      "bias_stability_60s");
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error clearing "
            "bias_stability_60s table");
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        return;
    }

    /* Write to bias_stability_60s table */
    status = write_bias_stability_main_to_db(open_db, rad_wo_common_id,
        proc_level, stability_delta_t);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error writing to "
            "bias_stability_60s table");
    }

    /* Close down the DB connection */
    ias_db_close_connection(open_db);
    ias_db_close_database_lib();
}




Ensure test_write_to_gain_stability_60s()
{
    struct ias_db_connection *open_db = NULL;

    char rad_wo_common_id[IAS_CHAR_ID_SIZE];  
    int status;

    char diffuser_type[IAS_DIFFUSER_TYPE_SIZE];
    IAS_RPS_OLI_DIFFUSER_TYPE test_panel_id;

    double stability_delta_t = 60.0;


    printf("Testing write to gain_stability_60s table...\n");

    /* Initialize the strings for the database */
    strcpy(proc_level, zero_level_str);
    strcpy(diffuser_type, IAS_RPS_OLI_DIFFUSER_WORKING);

    /* Open a connection to the database */
    open_db = connect_to_db();
    if (open_db == NULL)
    {
        assert_true_with_message(FALSE, "Error connecting to DB");
        ias_db_close_database_lib();
        return;
    }

    /* Get the radiometric work order ID number from the database */
    status = ias_misc_get_char_id(open_db,
            "RADIOMETRIC_WORK_ORDER_COMMON",
            "radiometric_wo_common_id",
            work_order_id, rad_wo_common_id);
    if (status == ERROR)
    {
        assert_true_with_message(FALSE, "Invalid rad_wo_common_id key "
             "for specified work order ID %s", work_order_id);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        return;
    }

    /* Clear bias_stability_60s table if necessary */
    status = clear_radstability_table(open_db, rad_wo_common_id,
                                      "gain_stability_60s");
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error clearing "
            "gain_stability_60s table");
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        return;
    }

    /* Write to gain_stability_60s table */
    test_panel_id = IAS_RPS_OLI_DIFFUSER_TYPE_WORKING;
    status = write_gain_stability_main_to_db(open_db, rad_wo_common_id,
        proc_level, stability_delta_t, test_panel_id);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error writing to "
            "gain_stability_60s table");
    }

    /* Close down the DB connection */
    ias_db_close_connection(open_db);
    ias_db_close_database_lib();
}




       /* Build the DB access test "sub-suite" */
void rad_stability_dbaccess_test(TestSuite *suite)
{
    add_test(suite, test_get_histogram_statistics);
    add_test(suite, test_write_detector_gain_stability_results_to_db);
    add_test(suite, test_write_sca_bias_stability_results_to_db);
    add_test(suite, test_write_sca_gainrad_stability_results_to_db);
    add_test(suite, test_write_to_bias_stability_60s);
    add_test(suite, test_write_to_gain_stability_60s);
}
