/***************************************************************************
 NAME:      rad_stability_end2end_test

 PURPOSE:   Performs CGreen unit testing for end-to-end testing of the
            60 s radiometric stability characterization

 ROUTINES:  strip_timestamp                (internal routine)
            test_end2end_gain_stability
            test_end2end_bias_stability
****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "ias_const.h"
#include "ias_satellite_attributes.h"
#include "cgreen.h"
#include "../char_rad_stability.h"


static char test_reportfile_name[PATH_MAX];
static char reference_reportfile_name[PATH_MAX];
static char test_dumpfile_name[PATH_MAX];
static char reference_dumpfile_name[PATH_MAX];
static char syscommand[1000];
static char work_order_id[IAS_WORKORDER_ID_SIZE];
static char *test_data_path = NULL;
static char *enable_db_tests = NULL;
static int band_list[3] = {1, 7, 8};    /* Test bands */


Ensure test_end2end_gain_stability()
{
    struct stat buff;
    char gain_paramfile_name[PATH_MAX];
    int status;
    int num_bands_processed = 3;
    int num_dumpfiles = 0;
    int index;


    printf("Running end-to-end gain stability characterization...\n");

    /* Make sure the test data for the work order is in the database, 
       and if it isn't, load it */
    strcpy(work_order_id, "RST24000");
    status = snprintf(syscommand, sizeof(syscommand),
        "./check_or_insert_end2end_woid.csh %s GAIN", work_order_id);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(FALSE, "Error constructing command line "
            "to check/insert simulated data into database...");
        return;
    }
    status = system(syscommand);
    if (status != 0)
    {
        assert_true_with_message(FALSE, "Error running script to check/"
            "insert simulated data into database");
        return;
    }

    /* Replace references to $IAS_UNIT_TEST_DATA_DIR  with the correct test
       directory location */
    status = snprintf(syscommand, sizeof(syscommand),
        "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' %s_template.omf > %s.omf",
        test_data_path, work_order_id, work_order_id);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(FALSE, "Error creating command to "
           "replace test data directory placeholder");
        return;
    }
    status = system(syscommand);
    if (status != 0)
    {
        assert_true_with_message(FALSE, "Error replacing test data "
            "directory placeholder");
        return;
    }

    /* Set up the command line to run the characterization */
    sprintf(gain_paramfile_name, "%s.odl", work_order_id);
    status = snprintf(syscommand, sizeof(syscommand),
        "../char_rad_stability %s", gain_paramfile_name);
    if (status < 0)
    {
        assert_true_with_message(FALSE, "Error constructing command line "
            "to run gain stability characterization");
        return;
    }

    /* Run the characterization */
    status = system(syscommand);
    if (status != 0)
    {
        assert_true_with_message(FALSE, "End-to-end run of gain stability "
            "characterization failed");
        return;
    }

    /* Check the results.  The work order was specified to generate an
       overall summary report, along with dump files for each band
       processed, along with putting results in the database */

    /* Check 1 -- count the number of dump files produced and the number
       of summary report files produced.  In the work order, 3 bands
       (1, 7, and 8) are specified in the band list.  There should be
       3 dump files (one for each band), and one summary report */
    snprintf(test_reportfile_name, sizeof(test_reportfile_name),
        "gain_stability_summary.txt");
    status = stat(test_reportfile_name, &buff);
    assert_equal_with_message(status, 0, "Cannot find summary report file "
        "%s", test_reportfile_name);
    if (status != 0)
        return;

    for (index = 0; index < num_bands_processed; index++)
    {
        snprintf(test_dumpfile_name, sizeof(test_dumpfile_name),
            "gain_stability_dumpfile_B%02d.txt",
            band_list[index]);
        status = stat(test_dumpfile_name, &buff);
        if (status == 0)
            num_dumpfiles++;
    }
    assert_equal_with_message(num_dumpfiles, num_bands_processed,
        "Number of dump files found %d does not match number of "
        "bands initially processed %d", num_dumpfiles, num_bands_processed);
    if (num_dumpfiles != num_bands_processed)
        return;

    /* Check 2 -- check the content of the summary report file against a
       "reference".  This can be done simply with a "diff" command. 
       HOWEVER... the header in the summary report contains a
       timestamp indicating when the file was written during processing.
       The reference file will not have this information, so it will have
       to be stripped away from the file just generated */
    snprintf(test_reportfile_name, sizeof(test_reportfile_name),
        "gain_stability_summary.txt");
    snprintf(reference_reportfile_name, sizeof(reference_reportfile_name),
        "%s_ref_gain_stability_summary.txt", work_order_id);
    status = snprintf(syscommand, sizeof(syscommand),
        "sed '/Spacecraft:/,$!'d %s | diff -q - %s",
        test_reportfile_name, reference_reportfile_name);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(FALSE, "Error preparing difference "
            "comparison system command");
        return;
    }
    status = system(syscommand);
    assert_equal_with_message(0, status, "Differences found between test "
        "gain stability summary report file %s and reference report file %s",
        test_reportfile_name, reference_reportfile_name);

    /* Check 3 -- do a similar check on the band dump files.  For the band
       dump files, however, the timestamp when the file was created is on
       the second line */
    for (index = 0; index < num_bands_processed; index++)
    {
        snprintf(test_dumpfile_name, sizeof(test_dumpfile_name),
            "gain_stability_dumpfile_B%02d.txt", band_list[index]);
        snprintf(reference_dumpfile_name, sizeof(reference_dumpfile_name),
            "%s_ref_dumpfile_B%02d.txt", work_order_id, band_list[index]);
        status = snprintf(syscommand, sizeof(syscommand),
            "sed '/Spacecraft:/,$!'d %s | diff -q - %s",
            test_dumpfile_name, reference_dumpfile_name);
        if ((status < 0) || (status >= sizeof(syscommand)))
        {
            assert_true_with_message(FALSE, "Error preparing difference "
                "comparison system command");
            return;
        }
        status = system(syscommand);
        assert_equal_with_message(0, status, "Differences found between "
            "band %d test gain stability dump file %s and reference dump "
            "file %s", band_list[index], test_dumpfile_name,
            reference_dumpfile_name);
    }

    /* Cleanup */
    status = snprintf(syscommand, sizeof(syscommand), "rm RST24000.omf "
        "gain_stability_dumpfile_B*.txt gain_stability_summary.txt");
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(FALSE, "Error creating command to rm "
            "test data output files");
    }
    else
    {
        status = system(syscommand);
        if (status != 0)
        {
            assert_true_with_message(FALSE, "Error removing test data "
                "output files");
        }
    }

}   /* END -- test_end2end_gain_stability */



Ensure test_end2end_bias_stability()
{
    struct stat buff;
    char bias_paramfile_name[PATH_MAX];
    int band_list[3] = {1, 7, 8};
    int status;
    int num_bands_processed = 3;
    int num_dumpfiles = 0;
    int index;


    printf("Running end-to-end bias stability characterization...\n");

    /* Make sure the test data for the work order is in the database, 
       and if it isn't, load it */
    strcpy(work_order_id, "RST25000");
    status = snprintf(syscommand, sizeof(syscommand),
        "./check_or_insert_end2end_woid.csh %s BIAS", work_order_id);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(FALSE, "Error constructing command line "
            "to check/insert simulated data into database...");
        return;
    }
    status = system(syscommand);
    if (status != 0)
    {
        assert_true_with_message(FALSE, "Error running script to check/"
            "insert simulated data into database");
        return;
    }

    /* Replace references to $IAS_UNIT_TEST_DATA_DIR with the correct
       test directory location */
    status = snprintf(syscommand, sizeof(syscommand),
        "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' %s_template.omf > %s.omf",
        test_data_path, work_order_id, work_order_id);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(FALSE, "Error creating command to "
           "replace test data directory placeholder");
        return;
    }
    status = system(syscommand);
    if (status != 0)
    {
        assert_true_with_message(FALSE, "Error replacing test data "
            "directory placeholder");
        return;
    }

    /* Set up the command line to run the characterization */
    sprintf(bias_paramfile_name, "%s.odl", work_order_id);
    status = snprintf(syscommand, sizeof(syscommand),
        "../char_rad_stability %s", bias_paramfile_name);
    if (status < 0)
    {
        assert_true_with_message(FALSE, "Error constructing command line "
            "to run bias stability characterization");
        return;
    }

    /* Run the characterization */
    status = system(syscommand);
    if (status != 0)
    {
        assert_true_with_message(FALSE, "End-to-end run of bias stability "
            "characterization failed");
        return;
    }

    /* Check the results.  The work order was specified to generate an
       overall summary report, along with dump files for each band
       processed, along with putting results in the database */

    /* Check 1 -- count the number of dump files produced and the number
       of summary report files produced.  In the work order, 3 bands
       (1, 7, and 8) are specified in the band list.  There should be
       3 dump files (one for each band), and one summary report */
    snprintf(test_reportfile_name, sizeof(test_reportfile_name),
        "bias_stability_summary.txt");
    status = stat(test_reportfile_name, &buff);
    assert_equal_with_message(status, 0, "Cannot find summary report file "
        "%s", test_reportfile_name);
    if (status != 0)
        return;

    for (index = 0; index < num_bands_processed; index++)
    {
        snprintf(test_dumpfile_name, sizeof(test_dumpfile_name),
            "bias_stability_dumpfile_B%02d.txt",
            band_list[index]);
        status = stat(test_dumpfile_name, &buff);
        if (status == 0)
            num_dumpfiles++;
    }
    assert_equal_with_message(num_dumpfiles, num_bands_processed,
        "Number of dump files found %d does not match number of "
        "bands initially processed %d", num_dumpfiles, num_bands_processed);
    if (num_dumpfiles != num_bands_processed)
        return;

    /* Check 2 -- check the content of the summary report file against a
       "reference".  This can be done simply with a "diff" command. 
       HOWEVER... the header in the summary report contains a
       timestamp indicating when the file was written during processing.
       The reference file will not have this information, so it will have
       to be stripped away from the file just generated */
    snprintf(test_reportfile_name, sizeof(test_reportfile_name),
        "bias_stability_summary.txt");
    snprintf(reference_reportfile_name, sizeof(reference_reportfile_name),
        "%s_ref_bias_stability_summary.txt", work_order_id);
    status = snprintf(syscommand, sizeof(syscommand),
        "sed '/Spacecraft:/,$!'d %s | diff -q - %s",
        test_reportfile_name, reference_reportfile_name);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(FALSE, "Error preparing difference "
            "comparison system command");
        return;
    }
    status = system(syscommand);
    assert_equal_with_message(0, status, "Differences found between test "
        "bias stability summary report file %s and reference report file %s",
        test_reportfile_name, reference_reportfile_name);

    /* Check 3 -- do a similar check on the band dump files.  For the band
       dump files, however, the timestamp when the file was created is on
       the second line */
    for (index = 0; index < num_bands_processed; index++)
    {
        snprintf(test_dumpfile_name, sizeof(test_dumpfile_name),
            "bias_stability_dumpfile_B%02d.txt",
            band_list[index]);
        snprintf(reference_dumpfile_name, sizeof(reference_dumpfile_name),
            "%s_ref_bias_stability_dumpfile_B%02d.txt", work_order_id,
            band_list[index]);
        status = snprintf(syscommand, sizeof(syscommand),
            "sed '/Spacecraft:/,$!'d %s | diff -q - %s",
            test_dumpfile_name, reference_dumpfile_name);
        if ((status < 0) || (status >= sizeof(syscommand)))
        {
            assert_true_with_message(FALSE, "Error preparing difference "
                "comparison system command");
            return;
        }
        status = system(syscommand);
        assert_equal_with_message(0, status, "Differences found between "
            "band %d test gain stability dump file %s and reference dump "
            "file %s", band_list[index], test_dumpfile_name,
            reference_dumpfile_name);
    }

    /* Cleanup */
    status = snprintf(syscommand, sizeof(syscommand), "rm RST25000.omf "
        "bias_stability_dumpfile_B*.txt "
        "bias_stability_summary.txt");
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(FALSE, "Error creating command to rm "
            "test data output files");
    }
    else
    {
        status = system(syscommand);
        if (status != 0)
        {
            assert_true_with_message(FALSE, "Error removing test data "
                "output files");
        }
    }

}  /* END -- test_end2end_bias_stability */



/* At this point, if something in the core environment setup fails, just
   exit.  Don't charge at as a test failure */
int main(int argc, char *argv[])
{
    int status;


    /* Get the test data path */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path ==  NULL)
    {
        printf("Cannot locate IAS test data location\n");
        exit(1);
    }

    /* Initialize the Satellite Attributes Library */
    status = ias_sat_attr_initialize(IAS_L8);
    if (status != SUCCESS)
    {
        printf("Cannot initialize IAS satellite attributes library\n");
        exit(1);
    }

    /* Check to see if database support has been set in the environment.
       If it isn't, we can't run the test because we need to access
       histogram statistics from the database */
    enable_db_tests = getenv("ENABLE_DB_TESTS");
    if (enable_db_tests == NULL)
    {
        printf("Cannot run end-to-end test of radiometric stability "
            "characterization -- database support is not enbabled\n");
        exit(0);
    }

    /* Create the test suite */
    TestSuite *suite = create_test_suite();
    add_test(suite, test_end2end_gain_stability);
    add_test(suite, test_end2end_bias_stability);

    /* Run the test(s) */
    if (argc > 1)
        return run_single_test(suite, argv[1], create_text_reporter());
    return run_test_suite(suite, create_text_reporter());
}   /* END -- main */
