/***************************************************************************
 NAME:            rad_stability_parameter_test

 PURPOSE:  CGreen unit testing of the get_parameters() routine used by the
           char_rad_stability application

 RETURNS:  CGreen status

***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "cgreen.h"

#include "../char_rad_stability.h"

#include "rad_stability_tests.h"


Ensure test_get_parameters()
{
    char *test_data_path = NULL;
    char cpf_data_dir[] = "ias_lib/cpf";
    char cpf_name[] = "L8CPF20090101_20090331.01";
    char full_cpf_name[PATH_MAX];
    char full_l0r_name[PATH_MAX];
    char l0r_data_dir[] = "ias_lib/misc/l0ra_location";
    char l0r_name[] = "test";
    char wo_params_filename[PATH_MAX] = "";
    char work_order_id[IAS_WORKORDER_ID_SIZE] = "RST1002";
    char system_command[1000] = "";
    char pmask_filename[] = "masks.pmask";

    int status;
    int size;

    RPS_RADSTABILITY_WO_PARAMS  wo_params;
    RPS_RADSTABILITY_OMF_PARAMS omf_params;
    RPS_RADSTABILITY_REPORT_HEADER rpt_header;


    printf("Testing parameter read from work order and OMF files...\n");

    /* Initialization */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_not_equal_with_message(test_data_path, NULL, "Cannot find "
        "test data location");
    if (test_data_path == NULL)
        return;

    /* Replace the $IAS_UNIT_TEST_DATA_DIR placeholder in the OMF file with
       its actual value */
    size = snprintf(system_command, sizeof(system_command), 
        "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' %s_template.omf > %s.omf", 
        test_data_path, work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return;
    }
    status = system(system_command);
    assert_equal_with_message(status, 0, "Cannot edit OMF template");
    if (status != 0)
    {
        return;
    }

    /* Rename the template ODL file as the actual ODL file */
    size = snprintf(wo_params_filename, sizeof(wo_params_filename), 
        "%s.odl", work_order_id);
    if ((size < 0) || (size >= sizeof(wo_params_filename)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
               "it needs %d bytes", size);
        return;
    }
    size = snprintf(system_command, sizeof(system_command), 
        "cp %s_template.odl %s", work_order_id, wo_params_filename);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return;
    }
    status = system(system_command);
    assert_equal_with_message(status, 0, "Cannot rename template "
        "parameter file");
    if (status != 0)
    {
        size = snprintf(system_command, sizeof(system_command), 
            "rm %s.omf", work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Read the parameters */
    memset(&wo_params, 0, sizeof(RPS_RADSTABILITY_WO_PARAMS));
    memset(&omf_params, 0, sizeof(RPS_RADSTABILITY_OMF_PARAMS));
    memset(&rpt_header, 0, sizeof(RPS_RADSTABILITY_REPORT_HEADER));
    status = get_parameters(wo_params_filename, &wo_params, &omf_params);
    assert_true(status == SUCCESS);
    if (status == ERROR)
    {
        printf("Cannot read processing parameters\n");
        size = snprintf(system_command, sizeof(system_command), 
            "rm %s.omf %s.odl", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Check parameters to see if they've been read properly */

    /* First the work order parameters */
    assert_string_equal(wo_params.work_order_id, "RST1002");

    /* Since this parameter is not specified in the work order ODL file,
       its default value should be 1.  Make sure that's the case */
    assert_equal(wo_params.calculate_radiance, 1);

    /* Check to see if default RPS processing level strings were written
       correctly. */
    assert_string_equal(wo_params.rps_solar_proc_level_dn,
        RPS_RADSTABILITY_SOLAR_PROC_LEVEL_TEMPCORRECT_DN);
    assert_string_equal(wo_params.rps_solar_proc_level_rad,
        RPS_RADSTABILITY_SOLAR_PROC_LEVEL_TEMPCORRECT_RAD);
    assert_string_equal(wo_params.rps_shutter_proc_level,
        RPS_RADSTABILITY_SHUTTER_PROC_LEVEL);

    /* Check the other parameters */
    assert_equal(wo_params.relgain_option,
            RPS_RADSTABILITY_PRELAUNCH_RELGAINS);
    assert_equal(wo_params.store_char_in_db, 1);
    assert_equal(wo_params.report_flag, 1);
    assert_equal(wo_params.dump_flag, 1);
    assert_double_equal(wo_params.stability_delta_t, 60.0);

    /* Now the OMF parameters */
    size = snprintf(full_l0r_name, sizeof(full_l0r_name), "%s/%s/%s", 
        test_data_path, l0r_data_dir, l0r_name);
    if ((size < 0) || (size >= sizeof(full_l0r_name)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
               "it needs %d bytes", size);
        return;
    }
    assert_string_equal(omf_params.l0r_filename, full_l0r_name);
    size = snprintf(full_cpf_name, sizeof(full_cpf_name), "%s/%s/%s", 
        test_data_path, cpf_data_dir, cpf_name);
    if ((size < 0) || (size >= sizeof(full_cpf_name)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
               "it needs %d bytes", size);
        return;
    }
    assert_string_equal(omf_params.cpf_filename, full_cpf_name);
    assert_string_equal(omf_params.pmask_filename, pmask_filename);

    /* Clean up */
    size = snprintf(system_command, sizeof(system_command), 
        "rm %s.odl %s.omf", work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return;
    }
    status = system(system_command);
    assert_equal_with_message(status, 0, "Cannot delete local parameter "
        "files");
}




/*-----------------------------------------------------------------------
           Define a test suite to add to the "master" test suite
-------------------------------------------------------------------------*/
void rad_stability_parameter_test(TestSuite *suite)
{
    add_test(suite, test_get_parameters);
}

