/*************************************************************************
 NAME:      background_response_cpf

 PURPOSE:   Modules implemented to process CPF information for TIRS
            background response determination

 ROUTINES:  get_cpf_scene_count
            get_cpf_file
            get_cpf_list
            create_local_pixel_mask

 NOTES:     It's quite possible that multiple scenes will use the same
            CPF.  We only want to retrieve one copy of the CPF in that
            case
**************************************************************************/
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_radiometry.h"
#include "ias_structures.h"
#include "tirs_background_response_determination.h"


/*   Service-related #defines */
#define COMMAND_LENGTH          1024
#define URL_LENGTH              500
#define SERVICE_PATH_STRING     "/dataservices/rest/"
#define CPF_SERVICE_STRING      "cpf?operation=getCalParmFile"
#define SERVICE_FILENAME_STRING "&filename="
#define SERVICE_DATE_STRING     "&nearMatch=true&date="


/*************************************************************************
 NAME:     get_cpf_scene_count

 PURPOSE:  Allocates an array to hold the counts of scenes associated with
           each CPF

 RETURNS:  Allocated and populated array if successful; NULL pointer if
           there's an error

 NOTES:    The routine get_cpf_scene_list MUST be called before this
           routine.  get_cpf_scene_list a) counts up the number of
           CPFs in the set of work order information returned from the
           database; and b) sets the particular index position for the
           current scene's CPF in the CPF list 
**************************************************************************/
int *get_cpf_scene_count
(
    const WOID_DATA *woid_data,      /* I: Scene/work order data */
    int num_rad_woids,               /* I: Number of scenes */
    int total_cpf_count              /* I: Number of distinct CPFs */
)
{
    int *cpf_scene_count = NULL;     /* Array to allocate/populate and
                                        return */
    int woid_index;
    int cpf_index;


    /* Allocate the array */
    cpf_scene_count = calloc(total_cpf_count, sizeof(*cpf_scene_count));
    if (cpf_scene_count == NULL)
    {
        IAS_LOG_ERROR("Allocating CPF scene count array");
        return NULL;
    }

    /* Count the number of scenes having the current CPF index */
    for (cpf_index = 0; cpf_index < total_cpf_count; cpf_index++)
    {
        for (woid_index = 0; woid_index < num_rad_woids; woid_index++)
        {
            if (woid_data[woid_index].cpf_index == cpf_index)
                (cpf_scene_count[cpf_index])++;
        }
    }

    return cpf_scene_count;
}   /* END get_cpf_scene_count */



/**************************************************************************
 NAME:     get_cpf_file

 PURPOSE:  Uses the CPF service to retrieve the specified CPF file

 RETURNS:  Integer status code of SUCCESS or ERROR
***************************************************************************/
int get_cpf_file
(
    char *cpf_filename,                   /* I/O: Name of current CPF file */
    const char *date_acquired             /* I: Acquisition date for
                                             interval of interest */
)
{
    char dummy_cpf_name[MAX_DB_CPFNAME_LENGTH + 1];
    struct stat buff;                           /* File information data
                                                   buffer (used to check
                                                   for existence) */
    char *services_location = NULL;             /* Location for service
                                                   code */
    char url[URL_LENGTH] = "";                  /* CPF file URL */
    char wget_command[COMMAND_LENGTH] = "";     /* Buffer for 'wget'
                                                   command */
    int status;


    /* Get the service location via the IAS_SERVICES environment
       variable */
    services_location = getenv("IAS_SERVICES");
    if (services_location == NULL)
    {
        IAS_LOG_ERROR("Cannot locate IAS services directory--make sure "
           "IAS_SERVICES environment variable has been set");
        return ERROR;
    }

    /* Check to see if the CPF file name string is empty.  If it is,
       we'll have to attempt to get a CPF through the acquisition date */
    if (strcmp(cpf_filename, "") == 0)
    {
        /* Build the URL */
        memset(url, 0, sizeof(url));
        status = snprintf(url, sizeof(url), "%s%s%s%s%s",
            services_location, SERVICE_PATH_STRING, CPF_SERVICE_STRING,
            SERVICE_DATE_STRING, date_acquired);
        if (status < 0)
        {
            IAS_LOG_ERROR("Building URL string:  service location %s/%s, "
                "CPF service %s", services_location, SERVICE_PATH_STRING,
                CPF_SERVICE_STRING);
            return ERROR;
        }
        if (status >= sizeof(url))
        {
            IAS_LOG_ERROR("Truncated URL string %s", url);
            return ERROR;
        }

        /* Create a CPF file name */
        memset(dummy_cpf_name, 0, sizeof(dummy_cpf_name));
        status = snprintf(dummy_cpf_name, sizeof(dummy_cpf_name),
            "CPF%s", date_acquired);
        if ((status < 0) || (status >= sizeof(dummy_cpf_name)))
        {
            IAS_LOG_ERROR("Forming CPF file name");
            return ERROR;
        }

        /* Build the service command */
        memset(wget_command, 0, sizeof(wget_command));
        status = snprintf(wget_command, sizeof(wget_command),
            "wget --quiet -O \"%s\" \"%s\"", dummy_cpf_name, url);
        if (status < 0)
        {
            IAS_LOG_ERROR("Building retrieval command string");
            return ERROR;
        }
        if (status >= sizeof(wget_command))
        {
            IAS_LOG_ERROR("Truncated wget command");
            return ERROR;
        }

        /* Save the CPF name */
        strncpy(cpf_filename, dummy_cpf_name, MAX_DB_CPFNAME_LENGTH);
        cpf_filename[MAX_DB_CPFNAME_LENGTH] = '\0';
    }
    else
    {
        /* Check to see if the named CPF has already been downloaded
           for another scene.  If it has, just return a SUCCESS status */
        status = stat(cpf_filename, &buff);
        if (status == 0)
        {
            IAS_LOG_WARNING("CPF %s has already been downloaded",
                cpf_filename);
            return SUCCESS;
        }

        /* Build the URL to retrieve */
        memset(url, 0, sizeof(url));
        status = snprintf(url, sizeof(url), "%s%s%s%s%s", services_location,
            SERVICE_PATH_STRING, CPF_SERVICE_STRING, SERVICE_FILENAME_STRING,
            cpf_filename);
        if (status < 0)
        {
            IAS_LOG_ERROR("Building URL string:  service location %s/%s, "
                "CPF service %s, CPF filename %s", services_location,
                SERVICE_PATH_STRING, CPF_SERVICE_STRING, cpf_filename);
            return ERROR;
        }
        if (status >= sizeof(url))
        {
            IAS_LOG_ERROR("Truncated URL string %s", url);
            return ERROR;
        }

        /* Build the service command */
        memset(wget_command, 0, sizeof(wget_command));
        status = snprintf(wget_command, sizeof(wget_command),
            "wget --quiet -O \"%s\" \"%s\"", cpf_filename, url);
        if (status < 0)
        {
            IAS_LOG_ERROR("Building retrieval command string");
            return ERROR;
        }
        if (status >= sizeof(wget_command))
        {
            IAS_LOG_ERROR("Truncated wget command");
            return ERROR;
        }
    }

    /* Get the CPF */
    status = system(wget_command);
    if (status != 0)
    {
        IAS_LOG_ERROR("CPF %s retrieval failed", cpf_filename);
        return ERROR;
    }

    return SUCCESS;
}   /* END get_cpf_file */



/**************************************************************************
 NAME:          get_cpf_list

 PURPOSE:       Counts the number of distinct CPFs for the set of specified
                scenes and creates the list

 RETURNS:       Pointer to string array containing distinct CPF names if
                successful; NULL pointer if there's an error

 SIDE EFFECTS:  a) Returns the CPF count for the given set of scenes
                b) Records the index position of the CPF for each scene
***************************************************************************/
char **get_cpf_list
(
    WOID_DATA *woid_data,                /* I/O: Array of work order data
                                            for each scene */
    int num_rad_woids,                   /* I: Number of scenes */
    int *distinct_cpf_count              /* O: Total number of distinct
                                            CPFs */
)
{
    char **cpf_list = NULL;
    int cpf_index_list[num_rad_woids];   /* CPF index list */
    int woid_index;                      /* Scene loop counter */
    int cpf_count = 0;                   /* Current CPF count */
    int cpf_index;                       /* CPF loop counter */
    int found_cpf;                       /* Flag indicating we've found
                                            an existing CPF filename */


    /* Initialize */
    memset(cpf_index_list, 0, num_rad_woids * sizeof(int));

    /* Use two loops to count the number of distinct CPFs */

    /* The outer loop checks from the first scene to scene num_rad_woids */
    for (woid_index = 0; woid_index < num_rad_woids; woid_index++)
    {
        /* Flag indicating we haven't found the CPF name yet */
        found_cpf = 0;

        /* The inner loop checks to the current CPF count */
        for (cpf_index = 0; cpf_index < cpf_count; cpf_index++)
        {
            if (strcmp(woid_data[woid_index].cpf_filename,
                woid_data[cpf_index_list[cpf_index]].cpf_filename) == 0)
            {
                /* We've matched a CPF name.  Flag that we've found a
                   match; and record the current CPF index for the current
                   scene */
                found_cpf = 1;
                woid_data[woid_index].cpf_index = cpf_index;
                break;
            }
        }
        if (!found_cpf) /* No match--this is a new CPF */
        {
            if (strcmp(woid_data[woid_index].cpf_filename, "") != 0)
            {
                /* Save the scene index where the different CPF name was
                   found, and increment the CPF count */
                woid_data[woid_index].cpf_index = cpf_count;
                cpf_index_list[cpf_count] = woid_index;
                cpf_count++;
            }
        }
    }

    /* Save the total CPF count */
    *distinct_cpf_count = cpf_count;

    /* Allocate the list buffer */
    cpf_list = (char **)ias_misc_allocate_2d_array(cpf_count,
        (MAX_DB_CPFNAME_LENGTH + 1), sizeof(char));
    if (cpf_list == NULL)
    {
        IAS_LOG_ERROR("Allocating CPF list data buffer -- insufficient "
            "memory available for %d element list", cpf_count);
        free(cpf_index_list);
        return NULL;
    }

    /* Populate the list buffer */
    for (cpf_index = 0; cpf_index < cpf_count; cpf_index++)
    {
        strncpy(cpf_list[cpf_index],
            woid_data[cpf_index_list[cpf_index]].cpf_filename,
            MAX_DB_CPFNAME_LENGTH);
        cpf_list[cpf_index][MAX_DB_CPFNAME_LENGTH] = '\0';
    }

    return cpf_list;
}   /* END get_cpf_list */



/**************************************************************************
 NAME:     create_local_pixel_mask

 PURPOSE:  Creates a local pixel mask for the current band/SCA based just
           on the inoperable detector information from the CPF

 RETURNS:  Pointer to a pixel mask for the current band/SCA if
           successful; a NULL pointer if there's an error
***************************************************************************/
IAS_PIXEL_MASK *create_local_pixel_mask
(
    const int *inop_list,                        /* I: Inoperable detector
                                                    list from CPF */
    int inop_count,                              /* I: Inoperable detector
                                                    list from CPF */
    int band_number,                             /* I: Band number */
    int sca_number,                              /* I: SCA number */
    int num_detectors,                           /* I: Number of detectors
                                                    for the current
                                                    band/SCA */
    int num_lines                                /* I: Number of lines of
                                                    data in the pixel
                                                    mask */
)
{
    IAS_PIXEL_MASK *local_pmask = NULL;          /* Pixel mask to return */
    int status;


    /* Create an empty pixel mask.  We only need a single line, since
       we're dealing only with inoperable detectors */
    local_pmask = ias_pm_create(band_number, sca_number, num_detectors,
        num_lines);
    if (local_pmask == NULL)
    {
        IAS_LOG_ERROR("Creating empty pixel mask for band %d SCA %d",
            band_number, sca_number);
        return NULL;
    }

    /* Identify the inoperable detectors */
    status = ias_rps_inoperable_detector_identification((int *)inop_list,
        inop_count, band_number, sca_number, num_detectors, num_lines,
        local_pmask);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Identifying inoperable detectors in band %d SCA %d",
            band_number, sca_number);
        ias_pm_destroy(local_pmask);
        return NULL;
    }

    return local_pmask;
}   /* END create_local_pixel_mask */

