/**************************************************************************
 NAME:      background_response_determination_db

 PURPOSE:   Module implementing routines to provide IAS characterization
            database access for the TIRS background response determination
            algorithm.

 ROUTINES:  cleanup_db
            connect_to_db
            get_rad_woid_data_in_date_range
            get_histogram_statistics
            get_selection_dates_from_interval (no db access)
**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <time.h>
#include "ias_const.h"
#include "ias_db.h"
#include "ias_db_get_connect_info.h"
#include "ias_db_insert.h"
#include "ias_db_query.h"
#include "ias_l0r_constants.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_structures.h"
#include "ias_types.h"
#include "tirs_background_response_determination.h"


#define QUERY_STRLEN 1640
#define SQL_RTN_STRLEN 512

/***********************************************************************
 NAME:     cleanup_db

 PURPOSE:  Closes an active database connection and frees the
             associated resources

 RETURNS:  Nothing
************************************************************************/
void cleanup_db
(
    struct ias_db_connection *open_db    /* I: Open database connection */
){
    if (open_db != NULL)
    {
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
    }
}   /* END cleanup_db */



/***********************************************************************
 NAME:     connect_to_db

 PURPOSE:  Retrieves IAS database connection information and attempts
           to make the connection

 RETURNS:  Pointer to valid ias_db_struct, indicating connection has
           been established, NULL pointer if connection cannot be made
************************************************************************/
struct ias_db_connection *connect_to_db()
{
    char *db_name = NULL;            /* Required connection information */
    char *user_id = NULL;
    char *user_password = NULL;
    char *db_host = NULL;
    struct ias_db_connection *db = NULL;   /* Open database connection */
    int status = SUCCESS;


    /* Initialize the database library */
    IAS_LOG_DEBUG("Initializing database library...");
    status = ias_db_initialize_database_lib();
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing database library");
        return NULL;
    }

    /* Get the connection information */
    IAS_LOG_DEBUG("Obtaining database connection information...");
    status = ias_db_get_connect_info(&db_name, &user_id, &user_password,
        &db_host);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Obtaining database connection information");
        return NULL;
    }

    /* Try to connect */
    IAS_LOG_DEBUG("Connecting to database...");

    db = ias_db_connect_to_database(db_name, user_id, user_password,
               db_host);

    /* Free the allocated strings with the connection information */
    free(db_name);
    free(user_id);
    free(user_password);
    free(db_host);

    /* Did we connect? */
    if (db == NULL)             /* NO -- log an error */
    {
        IAS_LOG_ERROR("Connecting to database");
    }
    else                        /* YES -- log success only in DEBUG mode */
    {
        IAS_LOG_DEBUG("Database connection established");
    }

    /* Return whatever we've got for *db (either NULL or a valid pointer) */
    return db;
}   /* END connect_to_db */



/***************************************************************************
 NAME:     get_rad_woid_data_in_date_range

 PURPOSE:  Queries the characterization database to obtain the
           radiometric work order information of all TIRS deep space
           intervals acquired during the specified date range
           having histogram stats of the correct processing level

 RETURNS:  Integer status code of SUCCESS or ERROR
****************************************************************************/
WOID_DATA *get_rad_woid_data_in_date_range
(
    struct ias_db_connection *db,     /* I: Database connection */
    const char *start_date,           /* I: Start date for date range */
    const char *end_date,             /* I: End date for date range */
    int *num_rad_woids                /* O: Total number of radiometric
                                         work order IDs */
)
{
    char query_string[QUERY_STRLEN];
    char from_clause[QUERY_STRLEN];
    char db_error_message[SQL_RTN_STRLEN + 1] = "";
    char start_date_without_T[IAS_L0R_DATE_LENGTH + 1];
    char end_date_without_T[IAS_L0R_DATE_LENGTH + 1];
    WOID_DATA *woid_buffer = NULL;
    struct ias_db_query *query = NULL;
    int status;
    int count = 0;


    /* Construct the query to obtain the array of ID numbers and
       acquisition dates.  This query selects the latest processed work
       order for the highest version of each interval where histogram
       statistics of the proper processing level exist (for deep space data,
       the highest level would be '0L' for the imaging bands).  The query
       does not retrieve the actual statistics data--just the information
       for the work order.*/
    strcpy(start_date_without_T, start_date);
    strcpy(end_date_without_T, end_date);
    start_date_without_T[10] = ' ';
    end_date_without_T[10] = ' ';
    status = snprintf(from_clause, sizeof(from_clause), "FROM (    "
      "SELECT radiometric_wo_common_id, "
      "   landsat_interval_id, "
      "   MAX(landsat_interval_id) OVER (PARTITION BY "
      "       SUBSTR(landsat_interval_id, 1, 22)) latest_interval_version, "
      "   date_acquired, "
      "   cpf_name, "
      "   work_order_id, "
      "   date_processed, "
      "   MAX(date_processed) OVER (PARTITION BY "
      "       landsat_interval_id) latest_processed "
      " FROM RADIOMETRIC_WORK_ORDER_COMMON rwoc "
      " WHERE valid = 'Y' AND collection_type = 'TIRS_DEEPSPACE' "
      "   AND date_acquired>=to_date('%s', 'YYYY-MM-DD HH24:MI:SS') "
      "   AND date_acquired<=to_date('%s', 'YYYY-MM-DD HH24:MI:SS') "
      "   AND EXISTS (SELECT 1 FROM  HISTOGRAM_STATISTICS_CAL hs "
      "               WHERE  hs.radiometric_wo_common_id = "
      "                      rwoc.radiometric_wo_common_id "
      "                      AND hs.processing_level IN ('0L')) "
      ")  "
      " WHERE date_processed = latest_processed "
      "   AND latest_interval_version = landsat_interval_id",
        start_date_without_T, end_date_without_T);
    if ((status < 0) || (status >= sizeof(from_clause)))
    {
        IAS_LOG_ERROR("Building query string");
        return NULL;
    }

    /* First get the number of work orders */
    status = snprintf(query_string, sizeof(query_string),
        "SELECT count(*) %s", from_clause);
    if ((status < 0) || (status >= sizeof(query_string)))
    {
        IAS_LOG_ERROR("Building query string");
        return NULL;
    }

    /* Run the query */
    query = ias_db_perform_query(db, query_string);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
            sizeof(db_error_message));
        IAS_LOG_ERROR("Query of RADIOMETRIC_WORK_ORDER_COMMON failed:  %s",
            db_error_message);
        return NULL;
    }

    /* Move DB "pointer" to the first valid record. */
    if (!ias_db_query_next(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
            sizeof(db_error_message));
        IAS_LOG_ERROR("Accessing RADIOMETRIC_WORK_ORDER_COMMON data:  %s",
            db_error_message);
        ias_db_query_close(query);
        return NULL;
    }

    /* Get the value and save it */
    status = ias_db_query_int_value(query, num_rad_woids, 0);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving scene count based on number of "
            "rad_wo_common_ids in characterization database");
        return NULL;
    }

    /* Cleanup these query resources */
    ias_db_query_close(query);

    /* Allocate the data buffer after checking to make sure we've got
       at least one scene acquired within the specified date range */
    if (*num_rad_woids == 0)
    {
        IAS_LOG_ERROR("No scenes were acquired between dates %s and %s",
            start_date, end_date);
        return NULL;
    }
    woid_buffer = calloc(*num_rad_woids, sizeof(WOID_DATA));
    if (woid_buffer == NULL)
    {
        IAS_LOG_ERROR("Allocating %d-element work order information data "
            "buffer", *num_rad_woids);
        return NULL;
    }

    /* Now construct the query to actually obtain the required work order
       information for the set of scenes.  Order the data by (1) cpf_name;
       and (2) acquisition dates */
    status = snprintf(query_string, sizeof(query_string),
        "SELECT radiometric_wo_common_id, landsat_interval_id, "
        "date_acquired, cpf_name, work_order_id %s "
        "order by cpf_name, date_acquired, radiometric_wo_common_id",
        from_clause);
    if ((status < 0) || (status >= sizeof(query_string)))
    {
        IAS_LOG_ERROR("Building work order data query string");
        free(woid_buffer);
        return NULL;
    }

    /* Run the query */
    query = ias_db_perform_query(db, query_string);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
            sizeof(db_error_message));
        IAS_LOG_ERROR("Query of RADIOMETRIC_WORK_ORDER_COMMON failed:  %s",
            db_error_message);
        free(woid_buffer);
        return NULL;
    }

    for (count = 0; count < *num_rad_woids; count++)
    {
        /* Move DB "pointer" to the next valid record. */
        if (!ias_db_query_next(query))
        {
            IAS_LOG_ERROR("Accessing RADIOMETRIC_WORK_ORDER_COMMON data");
            ias_db_query_close(query);
            free(woid_buffer);
            return NULL;
        }

        /* Get the id numbers */
        ias_db_query_string_value(query, 0,
            woid_buffer[count].rad_wo_common_id, 
            sizeof(woid_buffer[count].rad_wo_common_id));
        if (strcmp(woid_buffer[count].rad_wo_common_id, "") == 0)
        {
            IAS_LOG_ERROR("Retrieving radiometric work order ID numbers "
               "for scene %d", (count + 1));
            ias_db_query_close(query);
            free(woid_buffer);
            return NULL;
        }

        /* Get the interval name */
        ias_db_query_string_value(query, 1,
            woid_buffer[count].landsat_interval_id,
            IAS_L0R_INTERVAL_ID_LENGTH);

        /* Get the acquisition date in the format YYYY-MM-DD HH:MM:SS */
        ias_db_query_datetime_value(query, 2,
            woid_buffer[count].date_acquired, IAS_L0R_DATE_LENGTH);

        /* Get the CPF file name */
        ias_db_query_string_value(query, 3,
            woid_buffer[count].cpf_filename,
                sizeof(woid_buffer[count].cpf_filename));

        /* Get the work order ID */
        ias_db_query_string_value(query, 4,
            woid_buffer[count].work_order_id, 
            sizeof(woid_buffer[count].work_order_id));
    }

    /* Cleanup */
    ias_db_query_close(query);

    return woid_buffer;
}   /* END get_rad_woid_data_in_date_range */



/************************************************************************
 NAME:     get_histogram_statistics

 PURPOSE:  Queries the HISTOGRAM_STATISTICS_CAL table for the
           deep-space mean statistics values extracted from the
           product for the work orders corresponding to a single CPF

 RETURNS:  Integer status code of SUCCESS or ERROR
*************************************************************************/
int get_histogram_statistics
(
    struct ias_db_connection *db,     /* I: Open database connection */
    WOID_DATA *woid_data,       /* I: Work order list */
    int num_rad_woids,          /* I: Number of woid_data records */
    const char *cpf_name,             /* I: Name of processing CPF */
    int cpf_scene_count,              /* I: Number of scenes associated
                                         with the current CPF */
    int band_number,                  /* I: Current 1-based band number */
    int sca_number,                   /* I: Current 1-based SCA number */
    int detectors_per_sca,            /* I: Number of detectors for the
                                         current band/SCA */
    const char *proc_level_string,    /* I: RPS processing level for current
                                         band */
    double **scene_statistics_buffer  /* O: Scene statistics for all
                                         detectors within the current
                                         band/SCA */
)
{
    char query_string[QUERY_STRLEN];      /* SQL query string */
    char db_error_message[SQL_RTN_STRLEN] = "";
                                          /* Query error message */
    struct ias_db_query *query = NULL;    /* Query data */
    int scene_index;                      /* Scene loop counter */
    int ii;                               /* Detector loop counter */
    int work_order_index = 0;             /* Index of work orders for this CPF*/
    int status;
    int record_count = 0;                 /* Record counter */


    /* Process the list of scenes.  First, determine the work order(s)
       processed using the specified CPF */
    for (scene_index = 0; scene_index < cpf_scene_count; scene_index++)
    {
        /* Find the next scene/WO/char ID for this CPF. If the current
           work order's CPF name matches the specified name, check to
           make sure work_order_index isn't more than the number of
           available work order IDs.
           
           NOTE:  DO NOT reset work_order_index to 0 when processing
                  the next scene--the index needs to progress through
                  the number of work order IDs */
        for (; work_order_index < num_rad_woids; work_order_index++)
        {
            if (strcmp(woid_data[work_order_index].cpf_filename, cpf_name)
                == 0)
            {
                break;
            }
        }
        if (work_order_index >= num_rad_woids)
        {
            /* We ran out of work orders for this CPF */
            /*  this shouldn't happen if woid_data is correct */
            IAS_LOG_ERROR("Unable to find scene #%d for CPF %s",
                (scene_index + 1), cpf_name);
            return ERROR;
        }

        /* Build the full query string using previously identified char IDs */
        status = snprintf(query_string, sizeof(query_string),
            "select hs.detector_id, hs.scene_pixel_mean from "
            "histogram_statistics_cal hs where hs.band_id=%d and "
            "hs.sensor_chip_assembly_id=%d and hs.processing_level='%s' "
            "and hs.radiometric_wo_common_id = %s order by hs.detector_id;",
            band_number, sca_number, proc_level_string,
            woid_data[work_order_index].rad_wo_common_id);
        IAS_LOG_DEBUG("%s", query_string);
        if ((status < 0) || (status >= sizeof(query_string)))
        {
            IAS_LOG_ERROR("Constructing full SQL string to get band %d SCA %d "
                "histogram statistics", band_number, sca_number);
            return ERROR;
        }

        /* Execute the query and verify it succeeded */
        query = ias_db_perform_query(db, query_string);
        if (!ias_db_query_was_successful(query))
        {
            ias_db_query_get_error_message(query, db_error_message,
                sizeof(db_error_message));
            IAS_LOG_ERROR("Query of HISTOGRAM_STATISTICS_CAL table failed:  %s",
                db_error_message);
            return ERROR;
        }

        for (ii = 0; ii < detectors_per_sca; ii++)
        {
            int detector_number;         /* 1-based detector ID */
            int detector_index;          /* 0-based index corresponding to
                                            detector ID */

            /* Move the DB "pointer" to the next valid record */
            if (!ias_db_query_next(query))
            {
                ias_db_query_close(query);
                IAS_LOG_ERROR("Accessing HISTOGRAM_STATISTICS_CAL records");
                return ERROR;
            }
 
            /* Increment the record counter */
            record_count++;

            /* Pick off the detector number */
            status = ias_db_query_int_value(query, &detector_number, 0);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Reading detector_id value %d in scene %d, "
                    "band %d SCA %d", detector_number, (scene_index + 1),
                    band_number, sca_number);
                ias_db_query_close(query);
                return ERROR;
            }
            detector_index = (detector_number - 1);

            /* Finally, read in the mean statistics value */
            status = ias_db_query_double_value(query,
                &(scene_statistics_buffer[scene_index][detector_index]), 1);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Reading histogram statistics mean for "
                    "detector %d, scene %d band %d, SCA %d",
                    detector_number, (scene_index + 1), band_number,
                    sca_number);
                ias_db_query_close(query);
                return ERROR;
            }
        }

        /* Check the next set of work order information */
        work_order_index++;

    }   /* END  -- SCENE PROCESSING LOOP */

    /* Free up query resources */
    ias_db_query_close(query);

    return SUCCESS;
}   /* END get_histogram_statistics */



/*************************************************************************
 NAME:     get_selection_date_from_interval

 PURPOSE:  Given the starting or ending date and the number of days defining
           a date range interval, determines a corresponding
           ending(starting) date

 RETURNS:  Integer status code of SUCCESS or ERROR

 NOTES:    If the first date is not specified (i.e. the string is empty),
           the current processing date is obtained and used as the
           base date
**************************************************************************/
int get_selection_date_from_interval
(
    char *base_date,                   /* I/O: Base starting or ending date
                                          (if empty, populated with the
                                          current date time) */
    int num_days_in_interval,          /* I: Specified interval of days
                                          before current date (can be
                                          positive or negative) */
    char *base_plus_interval_date      /* O: Date accounting for
                                          the specified interval and
                                          interval direction */
)
{
    IAS_DATETIME base_datetime;        /* Data structure to hold specified
                                          date/time components */
    time_t epoch_now;                  /* Current seconds since 01/01/1970 */
    time_t epoch_base_plus_interval;   /* Seconds since 01/01/1970 for
                                          the interval date */
    struct tm *struct_tm_base_plus_interval = NULL;
                                          /* Broken-down time information
                                             for the interval-adjusted
                                             date/time */
    struct tm *struct_tm_existing = NULL; /* Broken-down time information for
                                             an existing base date */
    struct tm *struct_tm_now = NULL;      /* Broken-down time information for
                                             the current date */
    struct tm *struct_tm_current = NULL;  /* "Floating" pointer to
                                             tm_struct_now or
                                             tm_struct_existing data
                                             buffers.  If there's no
                                             specified base date string,
                                             it points to tm_struct_now,
                                             otherwise it points to
                                             tm_struct_existing
                                             data buffer */
    int status;


    /* If the base date string is empty, get the current elapsed seconds
       since the 01/01/1970 epoch when this part of the application is
       running */
    if (strcmp(base_date, "") == 0)
    {
        epoch_now = time(NULL);
        if (epoch_now == -1)
        {
            IAS_LOG_ERROR("Obtaining current number of seconds since "
                "01/01/1970 epoch");
            return ERROR;
        }

        /* Break this elapsed seconds since 01/01/1970 into its
           constituent date/time components */
        struct_tm_now = localtime(&epoch_now);
        if (struct_tm_now == NULL)
        {
            IAS_LOG_ERROR("Obtaining date/time components from elapsed "
                "time since 01/01/1970 epoch");
            return ERROR;
        }
        struct_tm_current = struct_tm_now;

        /* Create the base date/time string */
        status = sprintf(base_date, "%4d-%02d-%02dT%02d:%02d:%02d",
            struct_tm_current->tm_year + 1900,
            struct_tm_current->tm_mon + 1,
            struct_tm_current->tm_mday,
            struct_tm_current->tm_hour,
            struct_tm_current->tm_min,
            struct_tm_current->tm_sec);
        if (status < 0)
        {
            IAS_LOG_ERROR("Converting base date information to the desired "
                "ASCII string format 'YYYY-MM-DDTHH:MI:SS");
            return ERROR;
        }
    }
    else
    {
        /* Parse the date/time string */
        memset(&base_datetime, 0, sizeof(IAS_DATETIME));
        status = ias_misc_parse_datetime_string(base_date,
            IAS_DATETIME_CPF_FORMAT, &base_datetime);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Parsing specified base date/time string %s",
                base_date);
            return ERROR;
        }

        /* Allocate a struct_tm data structure for the existing date/time
           components */
        struct_tm_existing = calloc(1, sizeof(struct tm));
        if (struct_tm_existing == NULL)
        {
            IAS_LOG_ERROR("Allocating data buffer for existing date/time "
                "components");
            return ERROR;
        }

        /* Copy the component information.  The tm_year field in a
           'struct tm' data structure is populated with the number of
           years since 1900, so 1900 needs to be subtracted off to
           get the correct value.  Similarly, the tm_mon field
           is populated with the number of months since January,
           so 1 has to be subtracted off to get the correct value.
           The seconds field in an IAS_DATETIME data structure is
           a double value; since we just want the "core" seconds
           (no fraction), just cast it to an int.

           It may not be necessary, but just to be sure we get an
           accurate calculation, populate the tm->yday field with the
           corresponding value */
        struct_tm_existing->tm_year = base_datetime.year - 1900;
        struct_tm_existing->tm_yday = base_datetime.day_of_year;
        struct_tm_existing->tm_mon = base_datetime.month - 1;
        struct_tm_existing->tm_mday = base_datetime.day_of_month;
        struct_tm_existing->tm_hour = base_datetime.hour;
        struct_tm_existing->tm_min = base_datetime.minute;
        struct_tm_existing->tm_sec = (int)base_datetime.second;

        /* Set the current date/time data pointer to the existing data */
        struct_tm_current = struct_tm_existing;
    }

    /* Add the number of days in the interval to the current day of the
       month */
    struct_tm_current->tm_mday += num_days_in_interval;

    /* Convert this information into an elapsed number of seconds since
       01/01/1970 */
    epoch_base_plus_interval = mktime(struct_tm_current);
    if (epoch_base_plus_interval == -1)
    {
        IAS_LOG_ERROR("Determining seconds since 01/01/1970 epoch with "
            "interval adjustment to base date");
        free(struct_tm_existing);
        return ERROR;
    }

    /* Convert the adjusted seconds since 01/01/1970 to its constituent
       components.  Date/time components outside of their expected ranges
       will be re-determined correctly*/
    struct_tm_base_plus_interval = localtime(&epoch_base_plus_interval);
    if (struct_tm_base_plus_interval == NULL)
    {
        IAS_LOG_ERROR("Decomposing interval adjusted base date/time to "
            "its components");
        free(struct_tm_existing);
        return ERROR;
    }

    /* Write the interval adjusted date string with the desired format:
       'YYYY-MM-DD HH:MI:SS'.  Add 1900 to the year component and 1 to the
       month component to get the "true" 4-digit year and month */
    status = sprintf(base_plus_interval_date, "%4d-%02d-%02dT%02d:%02d:%02d",
        struct_tm_base_plus_interval->tm_year + 1900,
        struct_tm_base_plus_interval->tm_mon + 1,
        struct_tm_base_plus_interval->tm_mday,
        struct_tm_base_plus_interval->tm_hour,
        struct_tm_base_plus_interval->tm_min,
        struct_tm_base_plus_interval->tm_sec);

    /* Free the struct tm buffer for the existing date in case it was
       allocated */
    free(struct_tm_existing);

    /* Check the status from writing the start date string */
    if (status < 0)
    {
        IAS_LOG_ERROR("Converting base + interval date information to the "
            "desired ASCII string format 'YYYY-MM-DDTHH24:MI:SS");
        return ERROR;
    }

    return SUCCESS;
}   /* END get_select_dates_from_interval */
