/************************************************************************
 NAME:        background_response_mem

 PURPOSE:     Routines to allocate/free memory blocs required for the
              TIRS background response determination algorithm

 ROUTINES:  free_running_sums_buffer
            allocate_running_sums_buffer
            free_background_response_buffer
            allocate_background_response_buffer
            final_cleanup
*************************************************************************/
#include <stdlib.h>
#include "ias_const.h"
#include "ias_db.h"
#include "ias_cpf.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_pixel_mask.h"
#include "tirs_background_response_determination.h"


/*************************************************************************
 NAME:     free_running_sums_buffer

 PURPOSE:  Frees a multi-dimensional array of RUNNING_SUMS_DATA
           structures

 RETURNS:  Nothing
**************************************************************************/
void free_running_sums_buffer
(
    RUNNING_SUMS_DATA ****buffer,  /* I: Data buffer to free */
    int num_cpfs,                  /* I: Number of CPFs */
    int num_bands,                 /* I: Number of bands */
    int num_scas                   /* I: Number of SCAs per band */
)
{
    int cc;                        /* CPF loop counter */
    int bb;                        /* Band loop counter */
    int ss;                        /* SCA loop counter */


    /* Free only if the base pointer has been allocated */
    if (buffer != NULL)
    {
        for (cc = 0; cc < num_cpfs; cc++)
        {
            if (buffer[cc] != NULL)
            {
                for (bb = 0; bb < num_bands; bb++)
                {
                    if (buffer[cc][bb] != NULL)
                    {
                        for (ss = 0; ss < num_scas; ss++)
                        {
                            if (buffer[cc][bb][ss] != NULL)
                                free(buffer[cc][bb][ss]);
                        }
                        free(buffer[cc][bb]);
                    }
                }
                free(buffer[cc]);
            }
        }
        free(buffer);
    }
}   /* END free_running_sums_buffer */



/*************************************************************************
 NAME:     allocate_running_sums_buffer

 PURPOSE:  Allocates a four-dimensional array of RUNNING_SUMS_DATA
           structures.  The final array dimensions are
           a[NUM_CPFS][NUM_BANDS][NUM_SCAS][NUM_DETECTORS], and the
           size of each element is sizeof(RUNNING_SUMS_DATA)

 RETURNS:  Pointer to allocated array if successful; NULL pointer if there
           is an error
**************************************************************************/
RUNNING_SUMS_DATA ****allocate_running_sums_buffer
(
    int num_cpfs,                  /* I: Number of CPFs */
    int num_bands,                 /* I: Number of bands */
    int num_scas,                  /* I: Number of SCAs per band */
    int num_detectors              /* I: Number of detectors */
)
{
    RUNNING_SUMS_DATA ****buffer = NULL;
    int cpf_index;                 /* CPF loop counter */
    int band_index;                /* Band loop counter */
    int sca_index;                 /* SCA loop counter */


    /* Allocate the base data buffer */
    buffer = calloc(num_cpfs, sizeof(RUNNING_SUMS_DATA ***));
    if (buffer == NULL)
    {
        IAS_LOG_ERROR("Allocating base running sums data buffer for %d "
            "CPFs worth of data", num_cpfs);
        return NULL;
    }

    /* Now allocate for the number of bands */
    for (cpf_index = 0; cpf_index < num_cpfs; cpf_index++)
    {
        buffer[cpf_index] = calloc(num_bands, sizeof(RUNNING_SUMS_DATA **));
        if (buffer[cpf_index] == NULL)
        {
            IAS_LOG_ERROR("Allocating running sums buffer space for CPF %d",
                (cpf_index + 1));
            free_running_sums_buffer(buffer, num_cpfs, num_bands, num_scas);
            return NULL;
        }
    }

    /* Allocate for the number of SCAs */
    for (cpf_index = 0; cpf_index < num_cpfs; cpf_index++)
    {
        for (band_index = 0; band_index < num_bands; band_index++)
        {
            buffer[cpf_index][band_index] =
                calloc(num_scas, sizeof(RUNNING_SUMS_DATA *));
            if (buffer[cpf_index][band_index] == NULL)
            {
                IAS_LOG_ERROR("Allocating running sums buffer space for "
                    "%d SCAs for CPF %d band index %d",
                    num_scas, (cpf_index + 1), band_index);
                free_running_sums_buffer(buffer, num_cpfs, num_bands,
                    num_scas);
                return NULL;
            }
        }
    }

    /* Finally, allocate for the number of detectors */
    for (cpf_index = 0; cpf_index < num_cpfs; cpf_index++)
    {
        for (band_index = 0; band_index < num_bands; band_index++)
        {
            for (sca_index = 0; sca_index < num_scas; sca_index++)
            {
                buffer[cpf_index][band_index][sca_index] =
                    calloc(num_detectors, sizeof(RUNNING_SUMS_DATA));
                if (buffer[cpf_index][band_index][sca_index] == NULL)
                {
                    IAS_LOG_ERROR("Allocating running sums buffer space "
                        "for %d detectors in SCA %d for CPF %d band index "
                        "%d", num_detectors, num_scas, (cpf_index + 1),
                        band_index);
                    free_running_sums_buffer(buffer, num_cpfs, num_bands,
                        num_scas);
                    return NULL;
                }
            }
        }
    }

    return buffer;
}   /* END allocate_running_sums_buffer */



/*************************************************************************
 NAME:     free_background_response_buffer

 PURPOSE:  Frees a background response data block (means or standard
           deviations)

 RETURNS:  Nothing
**************************************************************************/
void free_background_response_buffer
(
    BACKGROUND_RESPONSE_DATA ***response,
    int num_bands,                    /* I: Number of bands */
    int num_scas                      /* I: Number of SCAs */
)
{
    int band_index;                   /* Band loop counter */
    int sca_index;                    /* SCA loop counter */


    if (response != NULL)
    {
        for (band_index = 0; band_index < num_bands; band_index++)
        {
            if (response[band_index] != NULL)
            {
                for (sca_index = 0; sca_index < num_scas; sca_index++)
                {
                    if (response[band_index][sca_index] != NULL)
                    {
                        free(response[band_index][sca_index]);
                    }
                }
                free(response[band_index]);
            }
        }
        free(response);
    }
}   /* END free_background_response_buffer */



/*************************************************************************
 NAME:     allocate_background_response_buffer

 PURPOSE:  Allocates the data block to contain the overall background
           response statistics for all bands/SCAs/detectors

 RETURNS:  Pointer to allocated block if successful; NULL pointer on
           failure
**************************************************************************/
BACKGROUND_RESPONSE_DATA ***allocate_background_response_buffer
(
    int num_bands,                /* I: Total number of bands */
    int num_scas,                 /* I: Total number of SCAs */
    int num_detectors             /* I: Total number of detectors */
)
{
    BACKGROUND_RESPONSE_DATA ***buffer = NULL;
                                  /* Data buffer to allocate and return */
    int band_index;               /* Band loop counter */
    int sca_index;                /* SCA loop counter */


    /* Allocate the base buffer */
    buffer = calloc(num_bands, sizeof(BACKGROUND_RESPONSE_DATA **));
    if (buffer == NULL)
    {
        IAS_LOG_ERROR("Allocating base background response data buffer "
            "for given number of bands");
        return NULL;
    }

    /* Allocate for the total number of SCAs in each band */
    for (band_index = 0; band_index < num_bands; band_index++)
    {
        buffer[band_index] =  calloc(num_scas,
            sizeof(BACKGROUND_RESPONSE_DATA *));
        if (buffer[band_index] == NULL)
        {
            IAS_LOG_ERROR("Allocating background response data buffer "
                "for %d SCAs in band index %d", num_scas, band_index);
            free_background_response_buffer(buffer, num_bands, num_scas);
            return NULL;
        }
    }

    /* Allocate for the total number of detectors in each SCA */
    for (band_index = 0; band_index < num_bands; band_index++)
    {
        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            buffer[band_index][sca_index] = calloc(num_detectors,
                sizeof(BACKGROUND_RESPONSE_DATA));
            if (buffer[band_index][sca_index] == NULL)
            {
                IAS_LOG_ERROR("Allocating background response data buffer "
                    "for %d detectors per SCA, band index %d",
                    num_detectors, band_index);
                free_background_response_buffer(buffer, num_bands, num_scas);
                return NULL;
            }
        }
    }

    return buffer;
}   /* END allocate_background_response_buffer */



/**************************************************************************
 NAME:     final_cleanup

 PURPOSE:  Takes all allocated buffers and frees their resources

 RETURNS:  Nothing
***************************************************************************/
void final_cleanup
(
    IAS_CPF *cpf,                           /* I: Populated CPF structure */
    struct ias_db_connection *open_db,      /* I: Open DB connection */
    int use_file_pmasks,                    /* I: Array of pixel masks read
                                               from a file (can be NULL if
                                               local pixel masks are used) */
    IAS_PIXEL_MASK **pmasks_from_file,      /* I: Array of pixel masks read
                                               from a file (can be NULL if
                                               local pixel masks are used) */
    int num_masks_read,                     /* I: Number of masks read from
                                               the pixel file */
    IAS_PIXEL_MASK *local_bandsca_pmask,    /* I: Local pixel mask created
                                               for the current band/SCA */
    WOID_DATA *woid_data,                   /* I: Work order data */
    int *cpf_scene_count,                   /* I: Number of scenes / CPF */
    char **cpf_list,                        /* I: List of CPF names */
    RUNNING_SUMS_DATA ****running_sums,     /* I: Running sums data */
    BACKGROUND_RESPONSE_DATA ***response,   /* I: Background response
                                               statistics */
    double **scene_statistics_buffer,       /* I: Scene statistics for the
                                               current band/SCA */
    int num_cpfs,                           /* I: Total number of CPFs */
    int num_bands,                          /* I: Number of bands */
    int num_scas                            /* I: Number of SCAs */
)
{
    IAS_PIXEL_MASK *curr_pmask_in_array = NULL;

    int mask_index;


    /* Free up the current CPF resources */
    if (cpf != NULL)
        ias_cpf_free(cpf);

    /* Clean up database resources */
    cleanup_db(open_db);

    /* Free the pixel mask(s) if they exist */
    if (use_file_pmasks)
    {
        if (pmasks_from_file != NULL)
        {
            for (mask_index = 0; mask_index < num_masks_read; mask_index++)
            {
                curr_pmask_in_array = pmasks_from_file[mask_index];
                ias_pm_destroy(curr_pmask_in_array);
            }
            free(pmasks_from_file);
        }
    }
    else
    {
        if (local_bandsca_pmask != NULL)
        {
            ias_pm_destroy(local_bandsca_pmask);
        }
    }

    /* Free up the running sums data buffer */
    free_running_sums_buffer(running_sums, num_cpfs, num_bands, num_scas);

    /* Free up the background response data buffers */
    free_background_response_buffer(response, num_bands, num_scas);

    /* Free up the scene statistics data buffer */
    ias_misc_free_2d_array((void **)scene_statistics_buffer);

    /* Free up the CPF string list */
    ias_misc_free_2d_array((void **)cpf_list);

    /* Free up the work order data buffer */
    free(woid_data);

    /* Free up the CPF scene count data buffer */
    free(cpf_scene_count);
}   /* END final_cleanup */
