/************************************************************************
 NAME:      calculate_background_response_statistics

 PURPOSE:   Routines to perform the basic steps of calculating the
            background response from TIRS deep-space histogram statistics

 ROUTINES:  subtract_dark_response
            calculate_background_response_sums_for_cpf
            calculate_final_background_response
*************************************************************************/
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include "ias_logging.h"
#include "ias_pixel_mask.h"
#include "ias_radiometry.h"
#include "tirs_background_response_determination.h"


/************************************************************************
 NAME:     subtract_dark_response

 PURPOSE:  Subtracts the dark response for each scene associated with
           a given CPF

 RETURNS:  Nothing
*************************************************************************/
void subtract_dark_response
(
    double **scene_statistics_buffer,      /* I/O: Dark response corrected
                                              histogram statistics for all
                                              scenes in the current CPF,
                                              all bands/SCAs */
    const double *dark_response,           /* I: Baseline dark response for
                                              the currently processing
                                              band/SCA from the currently
                                              processing CPF */
    int scene_count,                       /* I: Number of scenes for the
                                              currently processing CPF */
    int num_detectors                      /* I: Number of detectors in the
                                               current band/SCA */
)
{
    int scene_index;                       /* Scene loop counter */
    int det_index;                         /* Detector loop counter */


    for (scene_index = 0; scene_index < scene_count; scene_index++)
    {
        for (det_index = 0; det_index < num_detectors; det_index++)
        {
            scene_statistics_buffer[scene_index][det_index]
                -= dark_response[det_index];
        }
    }
}   /* END subtract_dark_response */



/************************************************************************
 NAME:     calculate_background_response_sums_for_cpf

 PURPOSE:  Calculates the running sums/counts for the current band/SCA
           for all scenes associated with the current CPF

 RETURNS:  Integer status code of SUCCESS or ERROR
*************************************************************************/
int calculate_background_response_sums_for_cpf
(
    const double **scene_statistics_buffer, /* I:  Detector-specific
                                               scene statistics for all
                                               scenes in the current
                                               band/SCA */
    const IAS_PIXEL_MASK *pmask,            /* I: Pixel mask for the
                                               current band/SCA */
    int scene_count,                        /* I: Number of scenes
                                               associated with the
                                               current CPF */
    int cpf_index,                          /* I: Current CPF */
    int band_number,                        /* I: Current band number */
    int band_index,                         /* I: Array index corresponding
                                               to the current band number */
    int sca_number,                         /* I: Current SCA number */
    int num_detectors,                      /* I: Number of detectors in
                                               the current band/SCA */
    RUNNING_SUMS_DATA ****rsums             /* O: Populated running sums
                                               array (num_cpfs x num_bands
                                               x num_scas x num_dets) */
)
{

    PIXEL_MASK_TYPE mask = IAS_RPS_INOP_DETECTOR;
                                            /* Mask value for the
                                               current detector */
    double curr_val;
    int det_index;                          /* Detector loop counter */
    int sca_index = sca_number - 1;         /* Index into running sums
                                               array */


    /* Calculate the statistics for each operable detector */
    for (det_index = 0; det_index < num_detectors; det_index++)
    {
        /* If a detector is inoperable (or flagged with a specified
           artifact) flag the data as invalid */
        if (ias_pm_is_detector_flagged((IAS_PIXEL_MASK *)pmask, det_index,
            mask))
        {
            rsums[cpf_index][band_index][sca_index][det_index].sum =
                INVALID_DATA;
            rsums[cpf_index][band_index][sca_index][det_index].sum_sqr =
                INVALID_DATA;
            rsums[cpf_index][band_index][sca_index][det_index].count =
                (int)INVALID_DATA;
        }
        else
        {
            int scene_index;

            /* Get the sum and the valid data count */
            for (scene_index = 0; scene_index < scene_count; scene_index++)
            {
                curr_val = scene_statistics_buffer[scene_index][det_index];
                rsums[cpf_index][band_index][sca_index][det_index].sum
                    += curr_val;
                rsums[cpf_index][band_index][sca_index][det_index].sum_sqr
                    += (curr_val * curr_val);
                rsums[cpf_index][band_index][sca_index][det_index].count++;
            }
        }
    }

    return SUCCESS;
}   /* END calculate_background_response_sums_for_cpf */



/************************************************************************
 NAME:     calculate_overall_background_response

 PURPOSE:  Calculates the overall background response estimate and
           standard deviation for each detector from the running sums
           data calculated for the CPFs associated with the selected
           scenes

 RETURNS:  Integer status code of SUCCESS or ERROR
************************************************************************/
int calculate_overall_background_response
(
    const RUNNING_SUMS_DATA ****rsums,       /* I: Running sums data */
    int total_cpf_count,                     /* I: Number of CPFs */
    int band_number,                         /* I: Current band number */
    int band_index,                          /* I: Array index to access
                                                current band */
    int sca_number,                          /* I: Array index to access
                                                current SCA */
    int num_detectors,                       /* I: Number of detectors in
                                                the current band/SCA */
    BACKGROUND_RESPONSE_DATA ***response_data /* O: Populated background
                                                 response data buffer for
                                                 all bands/SCAs/detectors */
)
{
    double curr_mean;                      /* Background response mean
                                              for the current detector */
    double var;                            /* Calculated variance */
    int det_index;                         /* Detector loop counter */
    int sca_index = sca_number - 1;        /* Array index to access current
                                              SCA */


    /* Determine the overall background response and standard deviation */
    for (det_index = 0; det_index < num_detectors; det_index++)
    {
        double det_sum = 0.0;
        double det_sum_sqr = 0.0;
        int det_scene_count = 0;
        int cpf_index;

        for (cpf_index = 0; cpf_index < total_cpf_count; cpf_index++)
        {
            if (rsums[cpf_index][band_index][sca_index]
                             [det_index].count == (int)INVALID_DATA)
            {
                continue;
            }

            det_sum
                += rsums[cpf_index][band_index][sca_index][det_index].sum;
            det_sum_sqr
                += rsums[cpf_index][band_index][sca_index][det_index].sum_sqr;
            det_scene_count
                += rsums[cpf_index][band_index][sca_index][det_index].count;
        }

        if (det_scene_count == 0)
        {
            IAS_LOG_WARNING("None of the selected scenes contains valid "
                "data for detector %d, flagging the background response "
                "statistics accordingly", (det_index + 1));
            response_data[band_index][sca_index][det_index].valid_count = 0;
            response_data[band_index][sca_index][det_index].mean =
                INVALID_DATA;
            response_data[band_index][sca_index][det_index].stdev =
                INVALID_DATA;
            continue;
        }
        else
        {
            curr_mean = det_sum / det_scene_count;
            response_data[band_index][sca_index][det_index].valid_count =
                det_scene_count;
            response_data[band_index][sca_index][det_index].mean =
                curr_mean;

            var = (det_sum_sqr - (det_scene_count * (curr_mean * curr_mean)))
                / (det_scene_count - 1);
            if (var <= 0.0)
            {
                IAS_LOG_WARNING("0 or negative variance calculated, setting "
                    "standard deviation to 0.0");
                response_data[band_index][sca_index][det_index].stdev = 0.0;
            }
            else
            {
                response_data[band_index][sca_index][det_index].stdev =
                     sqrt(var);
            }
        }
    }

    return SUCCESS;
}   /* END calculate_overall_background_response */
