/***************************************************************************
  NAME:     get_parameters

  PURPOSE:  Reads the input parameter files and populates the
            required parameter data structures

  RETURNS:  Integer status code of SUCCESS or ERROR

  NOTES:    If parameter help is successfully provided, application
            will immediately exit
****************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "ias_miscellaneous.h"
#include "ias_radiometry.h"
#include "ias_structures.h"
#include "tirs_background_response_determination.h"


#define MAX_NUM_PROC_PARAMS  7
#define MAX_NUM_OMF_PARAMS   2


int get_parameters
(
    const char *proc_params_filename,    /* I: Parameter filename */
    PROC_PARAMS *proc_params,            /* O: Processing parameters */
    PROC_PARAMS_OMF *proc_params_omf     /* O: OMF processing parameters */
)
{
    char omf_params_filename[PATH_MAX];

    /* RPS processing level */
    const char *default_deepspace_rps_proc_level[] =
        {RPS_BKRESPONSE_PROC_LEVEL_IMAGE};
    const char *default_deepspace_rps_proc_level_blind[] =
        {RPS_BKRESPONSE_PROC_LEVEL_BLIND};

    int status;                      /* Status for each read operation */
    int read_count;                  /* Count of parameters
                                        successfully read */
    int default_report_flag = TRUE;  /* Write a summary report by default */
    int default_days_in_interval = DEFAULT_NUM_DAYS;


/******************** READ ODL PARAMETER FILE **************************/
    /* Declare the work order parameter table */
    IAS_PARM_DECLARE_TABLE(proc_param_table, MAX_NUM_PROC_PARAMS);

    /* Do the common parameters first -- the work order ID and the
       needed RPS processing levels for the histogram statistics */
    IAS_PARM_WORK_ORDER_ID(proc_param_table, proc_params->work_order_id,
        sizeof(proc_params->work_order_id), 1);

    IAS_PARM_TIRS_RPS_PROC_LEVEL_DN(proc_param_table, IAS_PARM_OPTIONAL,
        default_deepspace_rps_proc_level, proc_params->rps_proc_level,
        sizeof(proc_params->rps_proc_level));

    IAS_PARM_TIRS_RPS_PROC_LEVEL_DN_BLIND(proc_param_table,
        IAS_PARM_OPTIONAL, default_deepspace_rps_proc_level_blind,
        proc_params->rps_proc_level_blind,
        sizeof(proc_params->rps_proc_level_blind));

    /* Then do the specific parameters.  Start with whether to generate
       a report */
    IAS_PARM_ADD_BOOLEAN(proc_param_table, REPORT_FLAG,
            "Write characterization results to summary report "
            "(1 = Write summary reports, 0 = Do not write summary reports)",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, &default_report_flag,
            &(proc_params->report_flag), sizeof(proc_params->report_flag), 1);

    /* Get the start and end of the date range used to select the scenes
       for analysis */
    IAS_PARM_ADD_STRING(proc_param_table, SCENE_SELECT_START_DATE,
        "Starting date range for scene selection in format:  "
        "YYYY[-]mm[-]dd[T]HH24[:]MI[:]SS",
        IAS_PARM_OPTIONAL, 0, NULL, 0, NULL,
        proc_params->start_date, sizeof(proc_params->start_date), 1);

    IAS_PARM_ADD_STRING(proc_param_table, SCENE_SELECT_END_DATE,
        "Ending date range for scene selection in format:  "
        "YYYY[-]mm[-]dd[T]HH24[:]MI[:]SS",
        IAS_PARM_OPTIONAL, 0, NULL, 0, NULL,
        proc_params->end_date, sizeof(proc_params->end_date), 1);

    /* In case the start and/or end dates are not specified, get the
       number of days defining an interval and the direction in which
       to calculate the interval */
    IAS_PARM_ADD_INT(proc_param_table, NUM_DAYS_IN_INTERVAL,
        "Number of days in scene selection interval", IAS_PARM_OPTIONAL,
        IAS_PARM_NOT_ARRAY, 0, 0, 0, 1, &default_days_in_interval,
        &(proc_params->num_days_in_interval),
        sizeof(proc_params->num_days_in_interval), 1);

    /* Get the total number of items to read from the parameter table */
    read_count = IAS_PARM_GET_TABLE_SIZE(proc_param_table);

    /* If someone enters -- help, --template, etc, on the command line
       when starting the application, the appropriate information will
       be displayed.  If someone enters a file name, its access will be
       checked.  If a help option was specified or the specified file name
       could not be accessed, a 1 will be returned.  In this case, just
       copy the work order file name (or help option) to the OMF
       file name and continue.  Otherwise, read the work order file */
    status = ias_parm_provide_help(proc_params_filename, proc_param_table,
        read_count, IAS_INPUT_PARAMETERS);

    if (status == ERROR)
    {
        IAS_LOG_ERROR("Work order parameter help failed");
        return ERROR;
    }
    else if (status == 1)
    {
        /* Copy the name of the processing parameter file to the name of the
           OMF file */
        sprintf(omf_params_filename, "%s", proc_params_filename);
    }
    else
    {
        IAS_DATETIME start_datetime;
        IAS_DATETIME end_datetime;
        int start_date_given = (strcmp(proc_params->start_date, "") != 0);
        int end_date_given = (strcmp(proc_params->end_date, "") != 0);

        status = ias_parm_read(proc_params_filename, proc_param_table,
            read_count);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading ODL processing parameters from file %s",
                proc_params_filename);
            return ERROR;
        }

        /* Validate scene selection date parameters */
        if (start_date_given)
        {
            memset(&start_datetime, 0, sizeof(IAS_DATETIME));
            status = ias_misc_parse_datetime_string(proc_params->start_date,
                IAS_DATETIME_CPF_FORMAT, &start_datetime);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Invalid start date string %s",
                    proc_params->start_date);
                return ERROR;
            }
        }

        if (end_date_given)
        {
            memset(&end_datetime, 0, sizeof(IAS_DATETIME));
            status = ias_misc_parse_datetime_string(proc_params->end_date,
                IAS_DATETIME_CPF_FORMAT, &end_datetime);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Invalid end date string %s",
                    proc_params->end_date);
                return ERROR;
            }
        }

        if (start_date_given && end_date_given)
        {
            /* If we're given the start and end dates, ensure the start date
               is before the end date.  Check date/time components for
               consistency.  The validation tests implmented here assume
               (hope) that the user will not enter a date range consisting
               of a single day within a time range.  */
            int equal_year = (start_datetime.year == end_datetime.year);
            int greater_year = (start_datetime.year > end_datetime.year);
            int equal_month = (start_datetime.month == end_datetime.month);
            int greater_month = (start_datetime.month > end_datetime.month);
            int equal_day =
                (start_datetime.day_of_month == end_datetime.day_of_month);
            int greater_day =
                (start_datetime.day_of_month > end_datetime.day_of_month);

            if (greater_year)
            {
                IAS_LOG_ERROR("Start year %d is after end year %d",
                    start_datetime.year, end_datetime.year);
                return ERROR;
            }
            else if (equal_year && greater_month)
            {
                IAS_LOG_ERROR("Same year %d, but start month %d is after "
                    "end month %d", start_datetime.year, start_datetime.month,
                    end_datetime.month);
                return ERROR;
            }
            else if (equal_year && equal_month && greater_day)
            {
                IAS_LOG_ERROR("Same year %d and month %d, but start day %d "
                    "is after end day %d", start_datetime.year,
                    start_datetime.month, start_datetime.day_of_month,
                    end_datetime.day_of_month);
                return ERROR;
            }
            else if (equal_year && equal_month && equal_day)
            {
                IAS_LOG_WARNING("Validation of time components will not be"
                    "performed.  It is up to the user of this application "
                    "to ensure that the specified start and end times "
                    "are consistent (i.e. start time before end time)...");
            }
        }
        /* If we're given just a start date(or end date), ensure that the
           number of days in the interval is appropriately signed to 
           calculate the corresponding end date(or start date) */
        else if ((start_date_given) && (proc_params->num_days_in_interval < 0))
        {
            IAS_LOG_ERROR("End date of scene selection interval to be "
                "calculated from start date %s, but number of days in "
                "interval is negative--estimated end date would fall "
                "before start date", proc_params->start_date);
            return ERROR;
        }
        else if ((end_date_given) && (proc_params->num_days_in_interval > 0))
        {
            IAS_LOG_ERROR("Start date of scene selection interval to be "
                "calculated from end date %s, but number of days in "
                "interval is positive--estimated start date would fall "
                "after end date", proc_params->end_date);
            return ERROR;
        }

        /* Use the work order ID to derive the name of the OMF parameter
           file.  Add 4 for the file extension. */
        snprintf(omf_params_filename,
            (sizeof(proc_params->work_order_id) + 4),
            "%s.omf", proc_params->work_order_id);
    }

/*********************** READ OMF PARAMETER FILE ************************/
    /* Declare the OMF parameter table */
    IAS_PARM_DECLARE_TABLE(omf_param_table, MAX_NUM_OMF_PARAMS);

    /* Common parameters */
    IAS_PARM_SATELLITE(omf_param_table, &proc_params_omf->satellite_number,
        sizeof(proc_params_omf->satellite_number), 1);

    /* Name of pixel mask file to use (in the context of a "custom"
       pixel mask file with artifacts to consider other than
       inoperable detectors) */
    IAS_PARM_PIXEL_MASK_FILENAME(omf_param_table, IAS_PARM_OPTIONAL,
        proc_params_omf->pmask_filename,
        sizeof(proc_params_omf->pmask_filename), 1);

    /* Get the total number of items to read in the OMF parameter table */
    read_count = IAS_PARM_GET_TABLE_SIZE(omf_param_table);

    /* See if the user wanted help or a template this time with the
       OMF file.  If not, just read the file */
    status = ias_parm_provide_help(omf_params_filename, omf_param_table,
        read_count, IAS_OMF_PARAMETERS);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("OMF parameter help failed");
        return ERROR;
    }
    else if (status == 1)
    {
        exit(EXIT_SUCCESS);
    }

    /* Read the OMF parameters */
    status = ias_parm_read(omf_params_filename, omf_param_table,
        read_count);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading OMF parameter file");
        return ERROR;
    }

    return SUCCESS;
}   /* END get_parameters */
