/*************************************************************************
 NAME:      background_response_cpf_tests

 PURPOSE:   CGreen unit tests for CPF-related processing routines used in
            the tirs_background_response_determination algorithm

 ROUTINES:  initialize_woid_data           (internal)
            test_get_cpf_scene_count       (unit test)
            test_get_cpf_list              (unit test)
            test_get_cpf_file              (unit test)
**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include "ias_logging.h"
#include "ias_l0r_constants.h"
#include "ias_miscellaneous.h"
#include "cgreen.h"
#include "../tirs_background_response_determination.h"


/**************************************************************************
 NAME:     initialize_woid_data

 PURPOSE:  Takes an empty WOID_DATA array and populates it with test
           values

 RETURNS:  Nothing
***************************************************************************/
static void initialize_woid_data
(
    WOID_DATA *woid_data,                  /* I/O:  Array of work order ID
                                              data to populated */
    int num_rad_woids,                     /* I: Number of scenes */
    int num_cpfs                           /* I: Number of CPFs */
)
{
    int ii;                                /* Loop counter */
    int scene_increment;                   /* Number of scenes per CPF */
    int cpf_index;                         /* Current CPF */

    scene_increment = num_rad_woids / num_cpfs;

    for (ii = 0; ii < num_rad_woids; ii++)
    {
        /* Calculate the CPF index */
        cpf_index = (ii / scene_increment);

        woid_data[ii].cpf_index = cpf_index;
        sprintf(woid_data[ii].cpf_filename, "test_CPF%d",
            (cpf_index + 1));
        sprintf(woid_data[ii].landsat_interval_id, "calval_prototype%d",
            (ii + 1));
        sprintf(woid_data[ii].rad_wo_common_id, "%d", ii);
    }
}



Ensure test_get_cpf_list()
{
    WOID_DATA woid_data[25];             /* Radiometric work order data */
    char test_cpf_name[MAX_DB_CPFNAME_LENGTH + 1];
    char **cpf_list = NULL;
    int num_rad_woids = 25;              /* Number of scenes */
    int num_cpfs = 5;                    /* Expected number of CPFs */
    int test_num_cpfs = 0;               /* CPF number returned from
                                            get_cpf_list() */
    int ii;                              /* General loop counter */


    printf("Testing CPF list generation...\n");

    /* Initialize */
    memset(woid_data, 0, num_rad_woids * sizeof(WOID_DATA));
    initialize_woid_data(woid_data, num_rad_woids, num_cpfs);

    /* Get the list of CPF names */
    cpf_list = get_cpf_list(woid_data, num_rad_woids, &test_num_cpfs);
    assert_not_equal_with_message(cpf_list, NULL, "get_cpf_list() call "
        "failed");
    if (cpf_list == NULL)
        return;

    /* Check the returned number of CPFs against the expected number */
    assert_equal_with_message(num_cpfs, test_num_cpfs, "Number of CPFs %d "
        "does not match number of expected CPFs %d", test_num_cpfs,
        num_cpfs);
    if (test_num_cpfs != num_cpfs)
    {
        ias_misc_free_2d_array((void **)cpf_list);
        return;
    }

    /* Check the names in the list */
    for (ii = 0; ii < num_cpfs; ii++)
    {
        sprintf(test_cpf_name, "test_CPF%d", (ii + 1));
        assert_string_equal_with_message(test_cpf_name, cpf_list[ii],
           "Test CPF name %s does not match expected CPF name %s",
           cpf_list[ii], test_cpf_name);
    }

    /* Done */
    ias_misc_free_2d_array((void **)cpf_list);
}



Ensure test_get_cpf_scene_count()
{
    WOID_DATA woid_data[25];            /* Radiometric work order data */
    int *test_cpf_scene_count = NULL;   /* Test CPF count array */
    int num_rad_woids = 25;             /* Number of scenes */
    int num_cpfs = 5;                   /* 5 CPFs */
    int cpf_index;                      /* CPF loop counter */
    int expected_cpf_scene_count = 5;   /* 5 scenes per CPF */


    printf("Testing CPF scene count...\n");

    /* Initialize */
    memset(woid_data, 0, num_rad_woids * sizeof(WOID_DATA));
    initialize_woid_data(woid_data, num_rad_woids, num_cpfs);

    /* Get the test scene count array */
    test_cpf_scene_count = get_cpf_scene_count(woid_data, num_rad_woids,
        num_cpfs);
    assert_not_equal_with_message(test_cpf_scene_count, NULL,
        "get_cpf_scene_count() call failed");
    if (test_cpf_scene_count == NULL)
        return;

    /* Check the results */
    for (cpf_index = 0; cpf_index < num_cpfs; cpf_index++)
    {
        assert_equal_with_message(test_cpf_scene_count[cpf_index],
            expected_cpf_scene_count, "Test scene count %d "
            "does not match expected scene count %d",
            test_cpf_scene_count[cpf_index], expected_cpf_scene_count);
    }

    /* Done */
    free(test_cpf_scene_count);
}



Ensure test_get_cpf_file()
{
    char test_cpf_name1[MAX_DB_CPFNAME_LENGTH + 1];
    char test_cpf_name2[MAX_DB_CPFNAME_LENGTH + 1];
    char syscommand[512];
    char test_date_acquired[IAS_L0R_DATE_LENGTH + 1] = "";
    struct stat buff;

    int status;


    printf("Testing CPF retrieval by name...\n");

    /* Check to make sure we don't have this CPF already loaded */
    memset(test_cpf_name1, 0, sizeof(test_cpf_name1));
    memset(test_cpf_name2, 0, sizeof(test_cpf_name2));

    /* Check to make sure we don't have this CPF already loaded */
    strcpy(test_cpf_name1, "L8CPF20030101_20030331.01");
    status = stat(test_cpf_name1, &buff);
    if (status == 0)
    {
        printf("CPF %s found, removing for test...\n", test_cpf_name1);
        remove(test_cpf_name1);
    }

    /* Try retrieving it by name.  We don't care what the acquisition
       date is in this case */
    status = get_cpf_file(test_cpf_name1, test_date_acquired);
    assert_equal_with_message(status, SUCCESS, "get_cpf_file() call using "
        "CPF name failed");
    if (status != SUCCESS)
        return;

    /* Make sure it's now in this directory */
    status = stat(test_cpf_name1, &buff);
    assert_equal_with_message(status, 0, "CPF file %s should now be here",
        test_cpf_name1);

    /* Pull the CPF using an acquisition date within the interval, and check
       to make sure it was retrieved */
    printf("Testing CPF retrieval by acquisition date...\n");

    status = snprintf(test_date_acquired, sizeof(test_date_acquired),
        "2003-02-26T17:00:00");
    if ((status < 0) || (status >= sizeof(test_date_acquired)))
    {
        assert_true_with_message(FALSE, "Error creating acquisition date "
            "string");
        return;
    }
    status = get_cpf_file(test_cpf_name2, test_date_acquired);
    assert_equal_with_message(status, SUCCESS, "get_cpf_file() call using "
        "scene/interval acquisition date failed");
    if (status != SUCCESS)
        return;

    status = stat(test_cpf_name2, &buff);
    assert_equal_with_message(status, 0, "CPF file %s should now be here",
        test_cpf_name2);

    /* Run a diff to make sure they're the "same" CPF */
    status = snprintf(syscommand, sizeof(syscommand), "diff %s %s",
        test_cpf_name1, test_cpf_name2);
    if ((status < 0) || (status >= sizeof(syscommand)))
    {
        assert_true_with_message(FALSE, "Error creating 'diff' command");
        remove(test_cpf_name1);
        remove(test_cpf_name2);
        return;
    }
    status = system(syscommand);
    assert_equal_with_message(status, 0, "CPFs are not the same");
    if (status)
        printf("Diff return status:  %d\n", status);

    /* Clean up */
    remove(test_cpf_name1);
    remove(test_cpf_name2);
}



int main(int argc, char *argv[])
{
    /* Suppress IAS_LOG_DEBUG messages */
    ias_log_initialize("tirs_background_cpf_test");

    /* Check the command line */
    if (argc > 2)
    {
       printf("USAGE:  background_response_cpf_tests "
           "<name of single test>\n");
       return 1;
    }

    /* Set up the test suite */
    TestSuite *suite = create_test_suite();

    add_test(suite, test_get_cpf_scene_count);
    add_test(suite, test_get_cpf_list);
    add_test(suite, test_get_cpf_file);

    /* Run the test(s) */
    if (argc > 1)
        return run_single_test(suite, argv[1], create_text_reporter());
    else
        return run_test_suite(suite, create_text_reporter());
}
