/*************************************************************************
 NAME:     background_response_db_tests

 PURPOSE:  CGreen unit test suite for database functionality in the TIRS
           background response determination

 RETURNS:  CGreen status
**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "ias_const.h"
#include "ias_logging.h"
#include "ias_l0r_constants.h"
#include "ias_miscellaneous.h"
#include "ias_db.h"

#include "cgreen.h"

#include "../tirs_background_response_determination.h"


#define COMMAND_STRLEN 256

static char start_date[IAS_L0R_DATE_LENGTH + 1];
static char end_date[IAS_L0R_DATE_LENGTH + 1];


/**************************************************************************
 NAME:     test_get_selection_date_from_interval

 PURPOSE:  get_selection_date_from_interval unit tests

 RETURNS:  Nothing
***************************************************************************/
Ensure test_get_selection_date_from_interval()
{
    int num_days_in_interval = -25;       /* 25 days before end date */
    int status;


    printf("Testing default creation of start date from specified interval "
        "and end date...\n");

    /* "Gin up" a fake end date generated from a system 'date' command */
    strcpy(start_date, "");
    snprintf(end_date, sizeof(end_date), "2011-01-28T17:00:00");

    /* Call get_selection_dates_from_interval */
    status = get_selection_date_from_interval(end_date, num_days_in_interval,
        start_date);
    assert_equal_with_message(status, SUCCESS, "Error obtaining starting "
        "date from %d-day interval and specified end date %s",
        num_days_in_interval, end_date);
    if (status != SUCCESS)
    {
        return;
    }
    printf("Start date:  %s\n", start_date);
    printf("End date:    %s\n", end_date);

    /* Check the returned start date string */
    assert_true_with_message((strcmp(start_date, "2011-01-03T17:00:00")
        == 0), "Incorrect start date %s derived from %d-day interval and "
        "end date %s", start_date, num_days_in_interval, end_date);

    /* Another test--set the start date as the previous end date, and
       add 25 days.  end date should be sometime torwards the end of
       February */
    printf("Testing default creation of end date from specified interval "
        "and start date...\n");
 
    strcpy(end_date, "");
    snprintf(start_date, sizeof(start_date), "2011-01-28T17:00:00");
    num_days_in_interval = 25;
    status = get_selection_date_from_interval(start_date,
        num_days_in_interval, end_date);
    assert_equal_with_message(status, SUCCESS, "Error obtaining ending "
        "date from %d-day interval and specified start date %s",
        num_days_in_interval, start_date);
    if (status != SUCCESS)
        return;
    printf("Start date:  %s\n", start_date);
    printf("End date:    %s\n", end_date);

    /* Check the returned end date */
    assert_true_with_message((strcmp(end_date, "2011-02-22T17:00:00")
        == 0), "Incorrect end date %s derived from %d-day interval and "
        "start date %s", end_date, num_days_in_interval, start_date);

    /* Another test -- pass in an empty start date, determine the date
       range based on the current date as the start date, and print the
       start date and end date */
    printf("Testing default creation of date range using current date as "
        "start date...\n");
    strcpy(start_date, "");
    if (num_days_in_interval < 0)
       num_days_in_interval = -num_days_in_interval;
    status = get_selection_date_from_interval(start_date,
        num_days_in_interval, end_date);
    assert_equal_with_message(status, SUCCESS, "Error obtaining date range "
       "using current date as start date");
    if (status != SUCCESS)
        return;
    printf("Start date:  %s\n", start_date);
    printf("End date:    %s\n", end_date);

    /* Final test -- pass in an empty end date, determine the date range
       based on the current date as the end date, and print the start and
       end dates */
    printf("Testing default creation of date range using current date as "
        "end date...\n");
    strcpy(end_date, "");
    if (num_days_in_interval > 0)
        num_days_in_interval = -num_days_in_interval;
    status = get_selection_date_from_interval(end_date,
        num_days_in_interval, start_date);
    assert_equal_with_message(status, SUCCESS, "Error obtaining date range "
       "using current date as start date");
    printf("Start date:  %s\n", start_date);
    printf("End date:    %s\n", end_date);


}   /* END test_get_selection_date_from_interval */



/***************************************************************************
 NAME:     test_get_rad_woid_data_in_date_range

 PURPOSE:  Unit test for get_rad_woid_data_in_date_range

 RETURNS:  Nothing
****************************************************************************/
Ensure test_get_rad_woid_data_in_date_range()
{
    struct ias_db_connection *open_db = NULL;

    WOID_DATA *woid_data = NULL;

    int test_num_rad_woids;
    int num_rad_woids = 5;
    int ii;
    int num_cpfs;

    char **cpf_list = NULL;

    char expected_cpf_name[MAX_DB_CPFNAME_LENGTH + 1];
    char expected_interval_id[IAS_L0R_INTERVAL_ID_LENGTH + 1] = "";
    char expected_work_order_id[IAS_WORKORDER_ID_SIZE] = "";


    printf("Testing retrieval of scene work order ID data within specified "
        "date range...\n");

    /* Open database connection */
    open_db = connect_to_db();
    assert_not_equal_with_message(open_db, NULL, "Error connecting to "
        "database");
    if (open_db == NULL)
        return;

    /* Get the work order ID information for all the scenes */
    snprintf(start_date, sizeof(start_date), "2009-01-01T00:00:00");
    snprintf(end_date, sizeof(end_date), "2009-03-31T23:59:59");
    woid_data = get_rad_woid_data_in_date_range(open_db, start_date,
        end_date, &test_num_rad_woids);
    assert_not_equal_with_message(woid_data, NULL,
        "get_rad_woid_data_in_date_range() call failed");
    if (woid_data == NULL)
    {
        cleanup_db(open_db);
        return;
    }
    assert_equal_with_message(test_num_rad_woids, num_rad_woids,
        "Expected number of scenes %d not equal to number of scenes %d "
        "found in database", num_rad_woids, test_num_rad_woids);
    if (num_rad_woids != test_num_rad_woids)
    {
        cleanup_db(open_db);
        free(woid_data);
        return;
    }

    /* Get the CPF list */
    cpf_list = get_cpf_list(woid_data, num_rad_woids, &num_cpfs);
    assert_not_equal_with_message(cpf_list, NULL, "Error obtaining CPF "
        "list");
    assert_equal_with_message(num_cpfs, 2, "Incorrect number of CPFs %d "
        "found from database");
    if ((cpf_list == NULL) || (num_cpfs != 2))
    {
        cleanup_db(open_db);
        free(woid_data);
        ias_misc_free_2d_array((void **)cpf_list);
        return;
    }

    for (ii = 0; ii < test_num_rad_woids; ii++)
    {
        if (ii == (test_num_rad_woids - 1))
        {
            sprintf(expected_cpf_name, "test_CPF2");
            assert_equal(woid_data[ii].cpf_index, (num_cpfs - 1));
        }
        else
        {
            sprintf(expected_cpf_name, "test_CPF1");
            assert_equal(woid_data[ii].cpf_index, (num_cpfs - 2));
        }

        sprintf(expected_work_order_id, "ITDSC%03d", (ii + 1));
        sprintf(expected_interval_id, "calval_prototype%d", (ii + 1));
        assert_string_equal(woid_data[ii].landsat_interval_id,
            expected_interval_id);
        assert_string_equal(woid_data[ii].work_order_id,
            expected_work_order_id);
        assert_string_equal(woid_data[ii].cpf_filename,
            expected_cpf_name);
    }

    cleanup_db(open_db);
    free(woid_data);
    ias_misc_free_2d_array((void **)cpf_list);
}   /* END test_get_rad_woid_data_in_date_range */



/**************************************************************************
 NAME:     test_get_histogram_statistics

 PURPOSE:  Unit test for get_histogram_statistics routine

 RETURNS:  Nothing
***************************************************************************/
Ensure test_get_histogram_statistics()
{
    struct ias_db_connection *open_db = NULL;
    WOID_DATA *woid_data = NULL;
    char **cpf_list = NULL;
    double **scene_stats_buffer = NULL;
    int *cpf_scene_count = NULL;
    int num_cpfs;
    int cpf_index = 0;                   /* data from test_CPF1 */
    int num_rad_woids;
    int status;
    int band_number;
    int sca_number;
    int num_detectors = 640;


    printf("Testing histogram statistics retrieval...\n");

    /* Establish the database connection */
    open_db = connect_to_db();
    assert_not_equal_with_message(open_db, NULL, "Error establishing "
        "database connection");
    if (open_db == NULL)
        return;

    /* Populate the WOID data buffer */
    sprintf(start_date, "2009-01-01T00:00:00");
    sprintf(end_date, "2009-03-31T23:59:59");
    woid_data = get_rad_woid_data_in_date_range(open_db, start_date,
        end_date, &num_rad_woids);
    assert_not_equal_with_message(woid_data, NULL, "Error retrieving work "
        "order ID data");
    if (woid_data == NULL)
    {
        cleanup_db(open_db);
        return;
    }

    /* Get the CPF list */
    cpf_list = get_cpf_list(woid_data, num_rad_woids, &num_cpfs);
    assert_not_equal_with_message(cpf_list, NULL, "Error creating CPF "
        "list");
    assert_equal_with_message(num_cpfs, 2, "Invalid number of CPFs %d "
        "found", num_cpfs);
    if ((cpf_list == NULL) || (num_cpfs != 2))
    {
        if (cpf_list != NULL)
            ias_misc_free_2d_array((void **)cpf_list);
        cleanup_db(open_db);
        free(woid_data);
        return;
    }
    assert_string_equal(cpf_list[0], "test_CPF1");
    assert_string_equal(cpf_list[1], "test_CPF2");

    /* Get the scene count */
    cpf_scene_count = get_cpf_scene_count(woid_data, num_rad_woids,
        num_cpfs);
    assert_not_equal_with_message(cpf_scene_count, NULL, "Error retrieving "
        "scene count associated with CPFs");
    if (cpf_scene_count == NULL)
    {
        ias_misc_free_2d_array((void **)cpf_list);
        cleanup_db(open_db);
        free(woid_data);
        return;
    }

    /* Allocate the scene stats data buffer */
    scene_stats_buffer = (double **)ias_misc_allocate_2d_array(
        cpf_scene_count[cpf_index], num_detectors, sizeof(double));
    assert_not_equal_with_message(scene_stats_buffer, NULL,
        "Error allocating scene statistics buffer");
    if (scene_stats_buffer == NULL)
    {
        ias_misc_free_2d_array((void **)cpf_list);
        cleanup_db(open_db);
        free(woid_data);
        free(cpf_scene_count);
        return;
    }

    /* Get the histogram statistics for the specified band/SCA */
    band_number = 10;
    sca_number = 2;
    status = get_histogram_statistics(open_db, woid_data, num_rad_woids,
        cpf_list[cpf_index], cpf_scene_count[cpf_index], band_number,
        sca_number, num_detectors, "0L", scene_stats_buffer);
    assert_equal_with_message(status, SUCCESS, "Error retrieving band %d "
        "SCA %d histogram statistics", band_number, sca_number);
    if (status != SUCCESS)
    {
        free(woid_data);
        ias_misc_free_2d_array((void **)cpf_list);
        ias_misc_free_2d_array((void **)scene_stats_buffer);
        free(cpf_scene_count);
        cleanup_db(open_db);
        return;
    }

    /* Test the read with values for detectors 1, 320, and 640 */
    assert_double_equal(scene_stats_buffer[0][0], 219.216583);
    assert_double_equal(scene_stats_buffer[0][319], 220.189987); 
    assert_double_equal(scene_stats_buffer[0][639], 219.736145);
    assert_double_equal(scene_stats_buffer[1][0], 219.544586);
    assert_double_equal(scene_stats_buffer[1][319], 219.977142); 
    assert_double_equal(scene_stats_buffer[1][639], 219.769547);
    assert_double_equal(scene_stats_buffer[2][0], 221.396576);
    assert_double_equal(scene_stats_buffer[2][319], 219.559631); 
    assert_double_equal(scene_stats_buffer[2][639], 221.027039);
    assert_double_equal(scene_stats_buffer[3][0], 221.242172);
    assert_double_equal(scene_stats_buffer[3][319], 223.359146); 
    assert_double_equal(scene_stats_buffer[3][639], 219.709579);

    /* Cleanup */
    free(woid_data);
    ias_misc_free_2d_array((void **)cpf_list);
    ias_misc_free_2d_array((void **)scene_stats_buffer);
    free(cpf_scene_count);
    cleanup_db(open_db);
}   /* END test_get_histogram_statistics */



int main(int argc, char *argv[])
{
    char *enable_db_tests = NULL;
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char syscommand[COMMAND_STRLEN];
    int num_scenes = 8; /* one is invalid, and one has ING1 stats, and one is a 
                           duplicate interval ID, leaving 5 to work with */
    int scene_index;
    int status;


    /* Make sure database tests have been enabled */
    enable_db_tests = getenv("ENABLE_DB_TESTS");
    if (enable_db_tests == NULL)
    {
        printf("Cannot run database tests--environment not set to do so\n");
        exit(0);
    }

    /* Make sure test data has been preloaded into the database, and
       load it if it hasn't been */
    for (scene_index = 0; scene_index < num_scenes; scene_index++)
    {
        sprintf(work_order_id, "ITDSC%03d", (scene_index + 1));
        sprintf(syscommand, "./check_or_insert_woid.csh %s",
            work_order_id);
        status = system(syscommand);
        if (status != 0)
        {
            printf("Error finding and/or inserting scene %d with work "
                "order ID %s\n", (scene_index + 1),
                work_order_id);
            exit(200);
        }
    }

    /* Suppress IAS_LOG_DEBUG messages */
    ias_log_initialize("tirs_background_response_db_tests");

    /* Initialize the satellite attributes library */
    ias_sat_attr_initialize(IAS_L8);

    /* Check the command line */
    if (argc > 2)
    {
       printf("USAGE:  background_response_db_tests "
           "<name of single test>\n");
       return 1;
    }

    /* Set up the test suite */
    TestSuite *suite = create_test_suite();

    add_test(suite, test_get_selection_date_from_interval);
    add_test(suite, test_get_rad_woid_data_in_date_range);
    add_test(suite, test_get_histogram_statistics);

    /* Run the test(s) */
    if (argc > 1)
        return run_single_test(suite, argv[1], create_text_reporter());
    else
        return run_test_suite(suite, create_text_reporter());
}
