/************************************************************************
 NAME:      background_response_end2end_test

 PURPOSE:   Unit tests of the tirs_background_response_application as
            an end-to-end process

 ROUTINES:  check_group_files              (internal routine)
            test_end2end_process
            test_get_parameters
**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <math.h>
#include "cgreen.h"
#include "ias_db.h"
#include "ias_l0r_constants.h"
#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "ias_const.h"
#include "../tirs_background_response_determination.h"


#define STRING_BUFFER_LENGTH         1000
#define DATA_BUFFER_LENGTH              5
#define NAME_STRLEN                   100
#define EXIT_FAIL_GET_DB_TEST_DATA    200
#define EXIT_FAIL_NO_TEST_DATA_DIR    300
#define EXIT_FAIL_NO_SAT_ID           400
#define EXIT_FAIL_NO_SAT_LIB          500


static char *test_data_path = NULL;
static char test_data_dir[] =
    "ias/rps/tirs_background_response_determination";
static char system_command[STRING_BUFFER_LENGTH + 1] = "";
static char paramfile_name[NAME_STRLEN + 1] = "br_params.odl";


/**************************************************************************
 NAME:     check_group_files

 PURPOSE:  Internal routine used to compare an adjusted background
           response output file produced by the TIRS background response
           determination algorithm to a "reference" output file

 RETURNS:  Integer status code of SUCCESS or ERROR.  ERROR is only
           returned if the test and/or reference file cannot be read
***************************************************************************/
static int check_group_files
(
    const char *test_group_filename,       /* I: Name of group file
                                              generated by the algorithm to
                                              be tested */
    const char *reference_group_filename,  /* I: Name of group file used
                                              as reference */
    double pass_threshold                  /* I: Maximum allowed between
                                              a value in the reference
                                              file and the corresponding
                                              value in the test file */
)
{
    char test_buff[STRING_BUFFER_LENGTH + 1];
                                           /* String buffer used for
                                              "test" data file read */
    char reference_buff[STRING_BUFFER_LENGTH + 1];  /* String buffer used for
                                              "source" data file read */
    FILE *test_fptr = NULL;                /* "Test" data file */
    FILE *reference_fptr = NULL;              /* "Source data file for
                                              comparison */
    double test_response[DATA_BUFFER_LENGTH + 1];/* Array of test background
                                              responses from a line in
                                              the test output file */
    double reference_response[DATA_BUFFER_LENGTH + 1];
                                           /* Array of source background
                                              responses from a line in the
                                              source output file */
    double diff = 0;
    int ii = 0;                            /* Loop counter */
    int jj = 0;                            /* Loop counter */
    int response_diffs = 0;                /* # of differences in background
                                              response (for a given detector
                                              between the "test" data and the
                                              "source" data) exceeding
                                              the specified threshold
                                              value*/
    int status = SUCCESS;


    /* Open the group files */
    test_fptr = fopen(test_group_filename, "r");
    reference_fptr = fopen(reference_group_filename, "r");

    assert_not_equal_with_message(test_fptr, NULL, "Cannot open test "
        "group file %s", test_group_filename);
    assert_not_equal_with_message(reference_fptr, NULL, "Cannot open "
        "reference group file %s", reference_group_filename);
    if ((test_fptr == NULL) || (reference_fptr == NULL))
    {
        if (test_fptr)
            fclose(test_fptr);
        if (reference_fptr)
            fclose(reference_fptr);
        return ERROR;
    }

    /* Check the GROUP header and parameter names to make sure the
       test one is correctly written */
    fgets(test_buff, STRING_BUFFER_LENGTH, test_fptr);
    fgets(reference_buff, STRING_BUFFER_LENGTH, reference_fptr);
    assert_string_equal_with_message(test_buff, reference_buff,
        "GROUP header statement difference");

    fgets(test_buff, STRING_BUFFER_LENGTH, test_fptr);
    fgets(reference_buff, STRING_BUFFER_LENGTH, reference_fptr);
    assert_string_equal_with_message(test_buff, reference_buff,
        "Parameter name difference");

    /* Check each line of data.  As currently implemented,
       there are 5 dark response values per line.  For 640
       detectors, that translates into 128 lines of dark response
       values */
    for (ii = 0; ii < 128; ii++)
    {
        status = fscanf(test_fptr, "%lf,%lf,%lf,%lf,%lf%*[ ,)]\n",
            &test_response[0], &test_response[1], &test_response[2],
            &test_response[3], &test_response[4]);
        if (status < 0)
        {
            assert_true_with_message(FALSE, "Cannot read test line %d",
                (ii + 1));
            fclose(test_fptr);
            fclose(reference_fptr);
            return ERROR;
        }

        status = fscanf(reference_fptr, "%lf,%lf,%lf,%lf,%lf%*[ ,)]\n",
            &reference_response[0], &reference_response[1],
            &reference_response[2], &reference_response[3],
            &reference_response[4]);
        if (status < 0)
        {
            assert_true_with_message(FALSE, "Cannot read reference line %d",
                (ii + 1));
            fclose(test_fptr);
            fclose(reference_fptr);
            return ERROR;
        }

        /* Check the absolute difference for each set of dark responses.
           Count those that are greater than the passing threshold */
        for (jj = 0; jj < DATA_BUFFER_LENGTH; jj++)
        {
            diff = fabs(test_response[jj] - reference_response[jj]);
            if (diff > pass_threshold)
            {
                response_diffs++;
            }
        }
    }

    assert_equal_with_message(response_diffs, 0,
        "%d differences of %lf or greater detected", response_diffs,
        pass_threshold);

    /* Check the END_GROUP and END statements */
    fgets(test_buff, STRING_BUFFER_LENGTH, test_fptr);
    fgets(reference_buff, STRING_BUFFER_LENGTH, reference_fptr);
    assert_string_equal_with_message(test_buff, reference_buff,
        "END_GROUP header statement difference");

    fgets(test_buff, STRING_BUFFER_LENGTH, test_fptr);
    fgets(reference_buff, STRING_BUFFER_LENGTH, reference_fptr);
    assert_string_equal_with_message(test_buff, reference_buff,
        "END closing statement difference");

    /* Close the files */
    fclose(test_fptr);
    fclose(reference_fptr);

    return SUCCESS;
}   /* END check_group_files */



Ensure test_end2end_process()
{
    char reference_odl_group_file[PATH_MAX] = "";
    char test_odl_group_file[PATH_MAX] = "";
    char cpf_filename1[MAX_DB_CPFNAME_LENGTH + 1] = "test_CPF1";
    char cpf_filename2[MAX_DB_CPFNAME_LENGTH + 1] = "test_CPF2";
    double pass_threshold = 1e-5;
    int band_number;
    int band_index;
    int sca_index;
    int num_bands = 3;
    int num_scas = 3;
    int sca_number;
    int status;


    printf("Testing end-to-end processing...\n");

    /* Get a local copy of the 2 test CPF files */
    status = snprintf(system_command, sizeof(system_command), 
        "cp %s/%s/%s %s", test_data_path, test_data_dir, cpf_filename1, 
        cpf_filename1);
    if ((status < 0) || (status >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Error creating command to copy "
            "CPF to current working directory");
        return;
    }
    status = system(system_command);
    if (status != 0)
    {
        assert_true_with_message(FALSE, "Error copying CPF %s to current "
        "working directory", cpf_filename1);
        return;
    }

    status = snprintf(system_command, sizeof(system_command), 
        "cp %s/%s/%s %s", test_data_path, test_data_dir, cpf_filename2, 
        cpf_filename2);
    if ((status < 0) || (status >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Error creating command to copy "
            "CPF to current working directory");
        remove(cpf_filename1);
        return;
    }
    status = system(system_command);
    if (status != 0)
    {
        assert_true_with_message(FALSE, "Error copying CPF %s to current "
           "current working directory", cpf_filename2);
        remove(cpf_filename1);
        return;
    }

    /* Run the application */
    status = snprintf(system_command, sizeof(system_command), 
        "../tirs_background_response_determination %s", paramfile_name);
    if ((status < 0) || (status >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Error creating command to run "
            "background response determination application");
        remove(cpf_filename1);
        remove(cpf_filename2);
        return;
    }
    status = system(system_command);
    assert_equal_with_message(status, 0, "Error running background "
        "response determination application");

    /* Check the ODL files for bands 10 and 11--there was no test data
       for the blind band.  Band 15 is a duplicate of band 11, but check
       it as well */
    for (band_index = 0; band_index < num_bands; band_index++)
    {
        if (band_index == num_bands - 1)
            band_number = 15;
        else
            band_number = band_index + 10;

        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            sca_number = sca_index + 1;

            /*Get the names of the test and reference ODL files */
            status = snprintf(reference_odl_group_file, 
                sizeof(reference_odl_group_file),
                "%s/%s/odl_group_br_response_B%02d_SCA%02d.odl",
                test_data_path, test_data_dir, band_number, sca_number);
            if ((status < 0) || (status >= sizeof(reference_odl_group_file)))
            {
                assert_true_with_message(FALSE, "Filename string too short, "
                    "it needs %d bytes", status);
                return;
            }
            
            status = snprintf(test_odl_group_file, sizeof(test_odl_group_file),
                "tirs_cpf_background_response_B%02d_SCA%02d.odl",
                band_number, sca_number);
            if ((status < 0) || (status >= sizeof(test_odl_group_file)))
            {
                assert_true_with_message(FALSE, "Filename string too short, "
                    "it needs %d bytes", status);
                return;
            }

            /*Compare the files */
            status = check_group_files(test_odl_group_file,
                reference_odl_group_file, pass_threshold);
            if (status != SUCCESS)
            {
                assert_true_with_message(FALSE, "Error comparing test ODL "
                    "group file %s to reference %s", test_odl_group_file,
                    reference_odl_group_file);
            }
        }
    }

    /* Clean up */
    remove(cpf_filename1);
    remove(cpf_filename2);
    status = snprintf(system_command, sizeof(system_command), 
               "rm tirs_cpf*.odl background_response*.txt");
    if ((status < 0) || (status >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", status);
        return;
    }
    system(system_command);
}   /* END test_end2end_process */



Ensure test_get_parameters()
{
    PROC_PARAMS proc_params;
    PROC_PARAMS_OMF proc_params_omf;
    int status = SUCCESS;


    printf("Testing read of work order and parameter files...\n");

    /* Try to read the files */
    memset(&proc_params, 0, sizeof(PROC_PARAMS));
    memset(&proc_params_omf, 0, sizeof(PROC_PARAMS_OMF));
    status= get_parameters(paramfile_name, &proc_params, &proc_params_omf);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "get_parameters() call failed");
        return;
    }

    /* Check the ODL parameters */
    int exp_num_days_in_interval = 30;
    assert_string_equal(proc_params.start_date, "2009-01-01T00:00:00");
    assert_string_equal(proc_params.end_date, "2009-03-31T23:59:59");
    assert_string_equal(proc_params.work_order_id, "br_params");
    assert_equal(exp_num_days_in_interval, proc_params.num_days_in_interval);
    assert_equal(proc_params.report_flag, 1);

    /* Check the OMF parameters */
    assert_equal(proc_params_omf.satellite_number, 8);
    assert_string_equal(proc_params_omf.pmask_filename, "");

    /* Disable ERROR messages */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* Now check the scene selection parameter validation */
    memset(&proc_params, 0, sizeof(PROC_PARAMS));
    snprintf(paramfile_name, sizeof(paramfile_name), 
        "params_date_error_year.odl");
    status = get_parameters(paramfile_name, &proc_params, &proc_params_omf);
    assert_equal_with_message(status, ERROR, "Start year after end year--"
        "should have returned ERROR");

    memset(&proc_params, 0, sizeof(PROC_PARAMS));
    snprintf(paramfile_name, sizeof(paramfile_name), 
        "params_date_error_yearmonth.odl");
    status = get_parameters(paramfile_name, &proc_params, &proc_params_omf);
    assert_equal_with_message(status, ERROR, "Same year, but start month "
        "after end month--should have returned ERROR");

    /* Restore normal logging */
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

}   /* END test_get_parameters */



int main(int argc, char *argv[])
{
    char *enable_db_tests = NULL;
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    IAS_SATELLITE_ID satellite_id = 0;
    int status = SUCCESS;
    int num_scenes = 8; /* one is invalid, one has ING1 stats, and one is a 
                           duplicate, leaving 5 to work with */
    int scene_index;


    /* Check how the driver routine is called.  If there aren't any
       other command line arguments, the whole test suite is run.
       Suppress any IAS_LOG_DEBUG messages */
    ias_log_initialize("tirs_background_response_end2end_test");

    /* Check to make sure database testing is enabled.  If it is,
       setup the database library and add the test.  If it isn't,
       skip it */
    enable_db_tests = getenv("ENABLE_DB_TESTS");
    if (enable_db_tests == NULL)
    {
        printf("Cannot run end-to-end testing--database support not "
            "available\n");
        exit(0);
    }

    /* Make sure test data has been preloaded into the database, and
       load it if it hasn't been */
    for (scene_index = 0; scene_index < num_scenes; scene_index++)
    {
        sprintf(work_order_id, "ITDSC%03d", (scene_index + 1));
        status = snprintf(system_command, sizeof(system_command), 
            "./check_or_insert_woid.csh %s", work_order_id);
        if ((status < 0) || (status >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", status);
            exit(EXIT_FAILURE);
        }
        status = system(system_command);
        if (status != 0)
        {
            printf("Error finding and/or inserting scene %d with work "
                "order ID %s\n", (scene_index + 1),
                work_order_id);
            exit(EXIT_FAIL_GET_DB_TEST_DATA);
        }
    }

    /* Get the test data path */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path == NULL)
    {
        printf("Cannot obtain test data directory location\n");
        exit(EXIT_FAIL_NO_TEST_DATA_DIR);
    }

    /* Initialize the Satellite Attributes Library */
    satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(8);
    if (satellite_id != IAS_L8)
    {
        printf("Cannot obtain satellite ID number\n");
        exit(EXIT_FAIL_NO_SAT_ID);
    }

    status = ias_sat_attr_initialize(satellite_id);
    if (status != SUCCESS)
    {
        printf("Failure initializing satellite attributes\n");
        exit(EXIT_FAIL_NO_SAT_LIB);
    }

    /* Set up the master test suite */
    TestSuite *suite = create_test_suite();

    add_test(suite, test_end2end_process);
    add_test(suite, test_get_parameters);

    /* Run the tests */
    if (argc > 1)
        return run_single_test(suite, argv[1], create_text_reporter());
    else
        return run_test_suite(suite, create_text_reporter());
}
