#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "ias_const.h"
#include "ias_l0r_constants.h"
#include "ias_logging.h"

#include "cgreen.h"

#include "../tirs_background_response_determination.h"


#define COMMAND_STRLEN 1024

static char *test_data_path = NULL;
static char test_data_dir[] =
    "ias/rps/tirs_background_response_determination";
static char syscommand[COMMAND_STRLEN] = "";


/************************************************************************
 NAME:     test_write_response_means_to_cpf_group

 PURPOSE:  Unit test for write_response_means_to_cpf_group

 RETURNS:  Nothing
*************************************************************************/
Ensure test_write_response_means_to_cpf_group()
{
    BACKGROUND_RESPONSE_DATA ***response_data = NULL;

    int band_number = 10;
    int band_index = 0;
    int sca_number = 1;
    int num_detectors = 640;
    int num_bands = 3;
    int num_scas = 3;
    int det_index;
    int sca_index;
    int status;
    int size;

    char ref_filename[PATH_MAX] = "";
    char test_filename[PATH_MAX] = "";


    printf("Testing CPF group file creation...\n");

    /* Allocate data buffer */
    response_data = allocate_background_response_buffer(
        num_bands, num_scas, num_detectors);
    assert_not_equal_with_message(response_data, NULL,
        "Error allocating test background response data buffer");
    if (response_data == NULL)
        return;

    /* Populate the test data buffer with some data */
    sca_index = sca_number - 1;
    for (det_index = 0; det_index < num_detectors; det_index++)
    {
        response_data[band_index][sca_index][det_index].mean =
            1000 * band_number + det_index + ((double)det_index / 1000);
    }

    /* Call the routine to write the buffer to a CPF group */
    status = write_background_response_means_to_cpf_group(
        (const BACKGROUND_RESPONSE_DATA ***)response_data,
        band_number, band_index, sca_number, num_detectors);
    assert_equal_with_message(status, SUCCESS, "Error writing test CPF "
        "group");
    if (status != SUCCESS)
    {
        free_background_response_buffer(response_data, num_bands, num_scas);
        return;
    }

    /* Compare it to the reference file */
    size = snprintf(test_filename, sizeof(test_filename),
        "tirs_cpf_background_response_B%02d_SCA%02d.odl",
        band_number, sca_number);
    if ((size < 0) || (size >= sizeof(test_filename)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
               "it needs %d bytes", size);
        return;
    }
    size = snprintf(ref_filename, sizeof(ref_filename),
        "%s/%s/ref_tirs_cpf_background_response_B%02d_SCA%02d.odl",
        test_data_path, test_data_dir, band_number, sca_number);
    if ((size < 0) || (size >= sizeof(ref_filename)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
               "it needs %d bytes", size);
        return;
    }

    size = snprintf(syscommand, sizeof(syscommand), "diff %s %s", 
        test_filename, ref_filename);
    if ((size < 0) || (size >= sizeof(syscommand)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return;
    }
    status = system(syscommand);
    assert_equal_with_message(status, SUCCESS, "File differences "
        "detected");

    /* Cleanup */
    free_background_response_buffer(response_data, num_bands, num_scas);
    remove(test_filename);
}   /* END test_write_response_means_to_cpf_group */



/************************************************************************
 NAME:     test_write_response_stdevs_to_report

 PURPOSE:  Unit test for write_response_stdevs_to_report

 RETURNS:  Nothing
*************************************************************************/
Ensure test_write_response_stdevs_to_report()
{
    char test_report_filename[PATH_MAX] = "";

    int band_number = 10;
    int band_index = 0;
    int sca_number = 1;
    int sca_index;
    int num_detectors = 640;
    int num_bands = 2;
    int num_scas = 3;
    int status;
    int det_index;
    int size;

    BACKGROUND_RESPONSE_DATA ***response_data = NULL;

    struct stat buff;

    printf("Testing summary report creation...\n");

    /* Allocate data buffers */
    response_data = allocate_background_response_buffer(num_bands,
        num_scas, num_detectors);
    assert_not_equal_with_message(response_data, NULL,
        "Error allocating background response buffer");
    if (response_data== NULL)
    {
        return;
    }

    /* Populate the data buffers with some data */
    sca_index = sca_number - 1;
    for (det_index = 0; det_index < num_detectors; det_index++)
    {
        response_data[band_index][sca_index][det_index].valid_count = 10;
        response_data[band_index][sca_index][det_index].mean =
            1000 * band_number + det_index + ((double)det_index / 1000);
        response_data[band_index][sca_index][det_index].stdev =
            (double)det_index / 1000;
    }

    /* Write the report */
    status = write_background_response_stdevs_to_report(
        (const BACKGROUND_RESPONSE_DATA ***)response_data,
        band_number, band_index, sca_number, num_detectors);

    assert_equal_with_message(status, SUCCESS, "Error creating summary "
        "report");
    if (status != SUCCESS)
    {
        free_background_response_buffer(response_data, num_bands, num_scas);
        return;
    }

    /* Make sure the report file exists.  The format could change, so
       it doesn't make any sense to compare it to a reference */
    size = snprintf(test_report_filename, sizeof(test_report_filename),
        "background_response_summary_B%02d_SCA%02d.txt",
        band_number, sca_number);
    if ((size < 0) || (size >= sizeof(test_report_filename)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
               "it needs %d bytes", size);
        return;
    }
    status = stat(test_report_filename, &buff);
    assert_equal_with_message(status, 0, "Test report file %s not found",
        test_report_filename);

    /* Cleanup */
    free_background_response_buffer(response_data, num_bands, num_scas);
    remove(test_report_filename);
}



Ensure test_write_response_report_header()
{
    struct stat buff;

    int scene_month[5] = {1, 1, 1, 2, 2};
    int scene_day[5] = {1, 15, 19, 12, 26};
    int status;
    int num_scenes = 5;
    int num_cpfs = 1;
    int scene_index;
    int cpf_index;
    int scene_increment;

    WOID_DATA woid_data[5];

    const char start_date[IAS_L0R_DATE_LENGTH] = "2009-01-01 00:00:00";
    const char end_date[IAS_L0R_DATE_LENGTH] = "2009-03-31 23:59:59";


    printf("Testing report header file creation...\n");

    /* Populate the work order header info data structure */
    memset(woid_data, 0, num_scenes * sizeof(WOID_DATA));

    scene_increment = num_scenes / num_cpfs;
    for (scene_index = 0; scene_index < num_scenes; scene_index++)
    {
        /* Determine the correct CPF index */
        cpf_index = scene_index / scene_increment;

        woid_data[scene_index].cpf_index = cpf_index;
        sprintf(woid_data[scene_index].rad_wo_common_id, "%d", 
            (scene_index + 1));
        sprintf(woid_data[scene_index].cpf_filename, "test_CPF%d",
            cpf_index);
        sprintf(woid_data[scene_index].landsat_interval_id,
            "calval_prototype%d", (scene_index + 1));
        sprintf(woid_data[scene_index].date_acquired,
            "2009-%02d-%02d 00:00:00",
            scene_month[scene_index], scene_day[scene_index]);
    }

    /* Write the header information to the file */
    status = write_woid_data_header(woid_data, num_scenes,
       start_date, end_date);
    assert_equal_with_message(status, SUCCESS, "Error creating "
        "background_response_summary_header.txt header file");

    /* If creation was successful, check to see if the file was written
       to the current working directory */
    if (status == SUCCESS)
    {
        status = stat("background_response_summary_header.txt", &buff);
        assert_equal_with_message(status, 0,
            "background_response_summary_header.txt file not found in "
            "current working directory");
    }

    /* Clean up */
    remove("background_response_summary_header.txt");
}








int main(int argc, char *argv[])
{
    IAS_SATELLITE_ID satellite_id;

    int status;


    /* Suppress IAS_DEBUG messages */
    ias_log_initialize("tirs_background_output_test");
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* Check the command line */
    if (argc > 2)
    {
        printf("USAGE:  background_response_output_tests "
            "<name of output test>\n");
        return 1;
    }

    /* Make sure we've defined a test data directory */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path == NULL)
    {
        printf("Cannot locate test data directory\n");
        return 1;
    }

    /* Initialize the Satellite Attributes Library */
    satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(8);
    if (satellite_id == ERROR)
    {
        printf("Cannot obtain valid Landsat 8 satellite ID number\n");
        return 1;
    }

    status = ias_sat_attr_initialize(satellite_id);
    if (status != SUCCESS)
    {
        printf("Failure initializing satellite attributes\n");
        return 1;
    }

    /* Create the test suite */
    TestSuite *suite = create_test_suite();

    add_test(suite, test_write_response_means_to_cpf_group);
    add_test(suite, test_write_response_stdevs_to_report);
    add_test(suite, test_write_response_report_header);

    /* Run test(s) */
    if (argc > 1)
        return run_single_test(suite, argv[1], create_text_reporter());
    else
        return run_test_suite(suite, create_text_reporter());
}

