/************************************************************************
 NAME:      calculate_background_response_tests

 PURPOSE:   CGreen unit testing for the
            calculate_background_response_sums_for_cpf routine

 RETURNS:   CGreen unit status

 ROUTINES:  test_calculate_background_response_sums_for_cpf
            test_calculate_overall_background_response
*************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_pixel_mask.h"
#include "ias_radiometry.h"
#include "ias_const.h"
#include "cgreen.h"
#include "../tirs_background_response_determination.h"


/***********************************************************************
 NAME:     test_calculate_background_response_sums_for_cpf

 PURPOSE:  Unit test for calculate_background_response_sums_for_cpf
           routine

 RETURNS:  Nothing
************************************************************************/
Ensure test_calculate_background_response_sums_for_cpf()
{
    RUNNING_SUMS_DATA ****test_rsd = NULL;
    IAS_PIXEL_MASK *pmask = NULL;
    double **test_scene_stats = NULL;
    double test_sum;
    double test_sum_sqr;
    int num_cpfs = 2;
    int num_bands = 3;
    int num_scas = 3;
    int sca_number = 1;
    int band_number = 10;
    int band_index = 0;
    int cpf_index = 0;
    int sca_index = 0;
    int scene_index = 0;
    int det_index;
    int num_detectors = 640;
    int num_scenes = 2;
    int status;
    int num_image_pixels = 1;


    printf("Testing calculation of running sums/sums-of-squares and "
        "counters...\n");

    /* Allocate running sums data buffer */
    test_rsd = allocate_running_sums_buffer(num_cpfs, num_bands, num_scas,
        num_detectors);
    assert_not_equal_with_message(test_rsd, NULL, "Error allocating "
        "running sums data buffer");
    if (test_rsd == NULL)
        return;

    /* Allocate test scenes data buffer */
    test_scene_stats = (double **)ias_misc_allocate_2d_array(
        num_scenes, num_detectors, sizeof(double));
    assert_not_equal_with_message(test_scene_stats, NULL,
        "Scene statistics data buffer should have been allocated");
    if (test_scene_stats == NULL)
    {
        free(test_rsd);
        return;
    }

    /* Populate the scene statistics data buffer */
    for (scene_index = 0; scene_index < num_scenes; scene_index++)
    {
        for (det_index = 0; det_index < num_detectors; det_index++)
        {
            test_scene_stats[scene_index][det_index] =
                200000*(scene_index + 1) + 10000*band_number + 1000*sca_index
                + det_index;
        }
    }

    /* Create a pixel mask without any inoperable detectors */
    pmask = ias_pm_create(band_number, sca_number, num_detectors,
       num_image_pixels);
    if (pmask == NULL)
    {
        assert_true_with_message(FALSE, "Pixel mask not created");
        free_running_sums_buffer(test_rsd, num_cpfs, num_bands, num_scas);
        ias_misc_free_2d_array((void **)test_scene_stats);
        return;
    }

    /* Calculate the running sums for band 10, SCA 1 */
    status = calculate_background_response_sums_for_cpf(
        (const double **)test_scene_stats,
        (const IAS_PIXEL_MASK *)pmask, num_scenes, cpf_index,
        band_number, band_index, sca_number, num_detectors,
        test_rsd);
    assert_equal_with_message(status, SUCCESS,
        "calculate_background_response_sums_for_cpf() call failed");
    if (status != SUCCESS)
    {
        free_running_sums_buffer(test_rsd, num_cpfs, num_bands, num_scas);
        ias_misc_free_2d_array((void **)test_scene_stats);
        ias_pm_destroy(pmask);
        return;
    }

    /* Destroy the pixel mask */
    ias_pm_destroy(pmask);

    /* Check the running sums data buffer */
    for (det_index = 0; det_index < num_detectors; det_index++)
    {
        /* Check the scene count.  Should be 2 */
        if (test_rsd[cpf_index][band_index][sca_index][det_index].count
            != num_scenes)
        {
            assert_true_with_message(FALSE, "Scene count for detector %d "
                "should be %d in band index %d SCA %d",
                (det_index + 1),
                test_rsd[cpf_index][band_index][sca_index][det_index].count,
                band_index, (sca_index + 1));
        }

        /* Check the sum and sum_sqr values */
        test_sum = 0.0;
        test_sum_sqr = 0.0;
        for (scene_index = 0; scene_index < num_scenes; scene_index++)
        {
            double curr_val = test_scene_stats[scene_index][det_index];
            test_sum += curr_val;
            test_sum_sqr += (curr_val * curr_val);
        }
        if (test_rsd[cpf_index][band_index][sca_index][det_index].sum
            != test_sum)
        {
            assert_true_with_message(FALSE, "Incorrect sum value %lf "
                "for detector %d in band index %d SCA %d",
                test_rsd[cpf_index][band_index][sca_index][det_index].sum,
                (det_index + 1), band_index, (sca_index + 1));
        }
        if (test_rsd[cpf_index][band_index][sca_index][det_index].sum_sqr
            != test_sum_sqr)
        {
            assert_true_with_message(FALSE, "Incorrect sum_sqr value %lf "
                "for detector %d in band index %d SCA %d",
                test_rsd[cpf_index][band_index][sca_index][det_index].sum_sqr,
                (det_index + 1), band_index, (sca_index + 1));
        }
    }

    /* Free the running sums data buffer */
    free_running_sums_buffer(test_rsd, num_cpfs, num_bands, num_scas);

    /* Free the scene statistics buffer */
    ias_misc_free_2d_array((void **)test_scene_stats);
}   /* END test_calculate_background_response_sums_for_cpf */



/*************************************************************************
 NAME:     test_calculate_overall_background_response

 PURPOSE:  CGreen unit test for the calculate_overall_background_response
           routine

 RETURNS:  Nothing
**************************************************************************/
Ensure test_calculate_overall_background_response()
{
    RUNNING_SUMS_DATA ****test_rsd = NULL;
    BACKGROUND_RESPONSE_DATA ***test_response = NULL;
    BACKGROUND_RESPONSE_DATA ***ref_response = NULL;
    IAS_PIXEL_MASK *pmask = NULL;
    double **test_scene_stats = NULL;
    double curr_mean;
    double det_sum;
    double det_sum_sqr;
    double var;
    int cpf_index;
    int num_cpfs = 2;
    int num_bands = 3;
    int num_scas = 3;
    int sca_index = 0;
    int scene_index;
    int det_index;
    int det_scene_count; 
    int num_detectors = 640;
    int num_scenes = 2;
    int band_number = 10;
    int sca_number = 1;
    int status;
    int num_image_pixels = 1;


    printf("Testing calculation of overall background response "
        "statistics...\n");

    /* Allocate running sums data buffer */
    test_rsd = allocate_running_sums_buffer(num_cpfs, num_bands, num_scas,
        num_detectors);
    assert_not_equal_with_message(test_rsd, NULL, "Error allocating "
        "running sums data buffer");
    if (test_rsd == NULL)
    {
        assert_true_with_message(FALSE, "Error allocating running sums data "
            "buffer");
        return;
    }

    /* Allocate the background response data buffers */
    test_response = allocate_background_response_buffer(num_bands, num_scas,
        num_detectors);
    if (test_response == NULL)
    {
        assert_true_with_message(FALSE, "Error allocating reference "
           "background response data buffer");
        free_running_sums_buffer(test_rsd, num_cpfs, num_bands, num_scas);
        return;
    }
    ref_response = allocate_background_response_buffer(num_bands, num_scas,
        num_detectors);
    if (ref_response == NULL)
    {
        assert_true_with_message(FALSE, "Error allocating reference "
           "background response data buffer");
        free_running_sums_buffer(test_rsd, num_cpfs, num_bands, num_scas);
        free_background_response_buffer(test_response, num_bands, num_scas);
        return;
    }

    /* Allocate test scenes data buffer */
    test_scene_stats = (double **)ias_misc_allocate_2d_array(
        num_scenes, num_detectors, sizeof(double));
    assert_not_equal_with_message(test_scene_stats, NULL,
        "Scene statistics data buffer should have been allocated");
    if (test_scene_stats == NULL)
    {
        free_running_sums_buffer(test_rsd, num_cpfs, num_bands, num_scas);
        free_background_response_buffer(test_response, num_bands, num_scas);
        free_background_response_buffer(ref_response, num_bands, num_scas);
        return;
    }

    /* Populate the scene statistics data buffer for CPF */
    for (scene_index = 0; scene_index < num_scenes; scene_index++)
    {
        for (det_index = 0; det_index < num_detectors; det_index++)
        {
            test_scene_stats[scene_index][det_index] =
                200000 * (scene_index + 1) + 10000 * band_number
                + 1000 * sca_index + det_index;
        }
    }

    /* Create a pixel mask without any inoperable detectors */
    pmask = ias_pm_create(band_number, sca_number, num_detectors,
        num_image_pixels);
    if (pmask == NULL)
    {
        assert_true_with_message(FALSE, "Pixel mask not created");
        free_running_sums_buffer(test_rsd, num_cpfs, num_bands, num_scas);
        free_background_response_buffer(test_response, num_bands, num_scas);
        free_background_response_buffer(ref_response, num_bands, num_scas);
        ias_misc_free_2d_array((void **)test_scene_stats);
        return;
    }

    /* Calculate the running sums for CPF 1, band 10, SCA 1 */
    int band_index1 = 0;
    int cpf_index1 = 0;
    status = calculate_background_response_sums_for_cpf(
        (const double **)test_scene_stats,
        (const IAS_PIXEL_MASK *)pmask, num_scenes, cpf_index1,
        band_number, band_index1, sca_number, num_detectors,
        test_rsd);
    assert_equal_with_message(status, SUCCESS, "calculate_background_"
        "response_sums_for_cpf() call failed for CPF 1");
    if (status != SUCCESS)
    {
        free_running_sums_buffer(test_rsd, num_cpfs, num_bands, num_scas);
        free_background_response_buffer(test_response, num_bands, num_scas);
        free_background_response_buffer(ref_response, num_bands, num_scas);
        ias_misc_free_2d_array((void **)test_scene_stats);
        ias_pm_destroy(pmask);
        return;
    }

    /* Change the scene statistics for CPF 2 */
    for (scene_index = 0; scene_index < num_scenes; scene_index++)
    {
        for (det_index = 0; det_index < num_detectors; det_index++)
        {
            test_scene_stats[scene_index][det_index] =
                400000 * (scene_index + 1) + 10000 * band_number
                + 1000 * sca_index + det_index;
        }
    }

    /* Calculate the running sums for CPF 2, band 10, SCA 1 */
    int band_index2 = 0;
    int cpf_index2 = 1;
    status = calculate_background_response_sums_for_cpf(
        (const double **)test_scene_stats,
        (const IAS_PIXEL_MASK *)pmask, num_scenes, cpf_index2,
        band_number, band_index2, sca_number, num_detectors,
        test_rsd);
    assert_equal_with_message(status, SUCCESS, "calculate_background_"
        "response_sums_for_cpf() call failed for CPF 2");
    if (status != SUCCESS)
    {
        free_running_sums_buffer(test_rsd, num_cpfs, num_bands, num_scas);
        free_background_response_buffer(test_response, num_bands, num_scas);
        free_background_response_buffer(ref_response, num_bands, num_scas);
        ias_misc_free_2d_array((void **)test_scene_stats);
        ias_pm_destroy(pmask);
        return;
    }

    /* Destroy the pixel mask */
    ias_pm_destroy(pmask);

    /* Calculate the overall background response */
    status = calculate_overall_background_response(
        (const RUNNING_SUMS_DATA ****)test_rsd, num_cpfs,
        band_number, band_index1, sca_number, num_detectors,
        test_response);
    assert_equal_with_message(status, SUCCESS, "calculate_overall_background_"
        "response() call failed");
    if (status != SUCCESS)
    {
        free_running_sums_buffer(test_rsd, num_cpfs, num_bands, num_scas);
        free_background_response_buffer(test_response, num_bands, num_scas);
        free_background_response_buffer(ref_response, num_bands, num_scas);
        ias_misc_free_2d_array((void **)test_scene_stats);
        return;
    }

    /* Calculate the reference background response.  These calculations are
       as implemented in calculate_overall_background_response_statistics.
       If those calculations change, these should change accordingly...*/
    for (det_index = 0; det_index < num_detectors; det_index++)
    {
        det_sum = 0.0;
        det_sum_sqr = 0.0;
        det_scene_count = 0;

        for (cpf_index = 0; cpf_index < num_cpfs; cpf_index++)
        {
            if (test_rsd[cpf_index][band_index1][sca_index]
                             [det_index].count == (int)INVALID_DATA)
            {
                continue;
            }

            det_sum
                += test_rsd[cpf_index][band_index1][sca_index][det_index].sum;
            det_sum_sqr
                += test_rsd[cpf_index][band_index1][sca_index][det_index]
                   .sum_sqr;
            det_scene_count
                += test_rsd[cpf_index][band_index1][sca_index][det_index]
                   .count;
        }

        if (det_scene_count == 0)
        {
            IAS_LOG_WARNING("None of the selected scenes contains valid "
                "data for detector %d, flagging the background response "
                "statistics accordingly", (det_index + 1));
            ref_response[band_index1][sca_index][det_index].valid_count = 0;
            ref_response[band_index1][sca_index][det_index].mean =
                INVALID_DATA;
            ref_response[band_index1][sca_index][det_index].stdev =
                INVALID_DATA;
            continue;
        }
        else
        {
            curr_mean = det_sum / det_scene_count;
            ref_response[band_index1][sca_index][det_index].valid_count =
                det_scene_count;
            ref_response[band_index1][sca_index][det_index].mean =
                curr_mean;
            var = (det_sum_sqr - (det_scene_count * (curr_mean * curr_mean)))
                / (det_scene_count - 1);
            if (var <= 0.0)
            {
                IAS_LOG_WARNING("0 or negative variance calculated, setting "
                    "standard deviation to 0.0");
                ref_response[band_index1][sca_index][det_index].stdev = 0.0;
            }
            else
            {
                ref_response[band_index1][sca_index][det_index].stdev =
                     sqrt(var);
            }
        }

        /* Compare results */
        status = memcmp(&test_response[band_index1][sca_index][det_index],
            &ref_response[band_index1][sca_index][det_index],
            sizeof(BACKGROUND_RESPONSE_DATA));
        if (status != 0)
        {
            assert_true_with_message(FALSE, "Detector %d test data "
                "does not equal reference data ", (det_index + 1));
        }
    }

    /* Free the running sums data buffer */
    free_running_sums_buffer(test_rsd, num_cpfs, num_bands, num_scas);

    /* Free the test and reference background response buffers */
    free_background_response_buffer(test_response, num_bands, num_scas);
    free_background_response_buffer(ref_response, num_bands, num_scas);

    /* Free the scene statistics buffer */
    ias_misc_free_2d_array((void **)test_scene_stats);
}   /* END test_calculate_overall_background_response */



int main(int argc, char *argv[])
{
    /* Initialize */
    ias_log_initialize("tirs_calculate_background_response_test");

    /* Check the command line */
    if (argc > 2)
    {
       printf("USAGE:  calculate_backround_response_tests "
           "<name of single test>\n");
       return 1;
    }

    /* Set up the test suite */
    TestSuite *suite = create_test_suite();

    add_test(suite, test_calculate_background_response_sums_for_cpf);
    add_test(suite, test_calculate_overall_background_response);

    /* Run the test(s) */
    if (argc > 1)
        return run_single_test(suite, argv[1], create_text_reporter());
    else
        return run_test_suite(suite, create_text_reporter());
}
