#!/bin/sh
# /************************************************************************
#  NAME:        generate_background_response_test_cpf.sh
#
#  PURPOSE:     Calls the (standalone) TIRS background response determination
#               determination algorithm with the option to calculate the
#               estimated background for a specified date range (generating
#               CPF group data files), then calls the merge script used by the
#               CPFgui to merge the individual group files into the main CPF.
#
#  RETURNS:     Integer status code of 0 (successful completion) or 1
#               (error condition)

#  INVOCATION:  generate_background_response_test_cpf.sh
#               <processing parameter file name>
#               <baseline CPF file to modify>
#               <test CPF file name>

#  NOTES:       The processing parameter file is assumed to be in the same
#               working directory the script is running from.  The individual
#               group data files produced by
#               tirs_background_response_determination will be written to
#               the working directory, as will the output "test" CPF

# *************************************************************************/
#
# Look for the shell script cpfmerge.sh in $IAS_BIN.  If it's not in
# $IAS_BIN, quit here.
find $IAS_BIN/cpfmerge.sh
if [ $? -ne 0 ]
then
    echo "cpfmerge.sh not installed in " $IAS_BIN ".  Build/rebuild CPF "
    echo "GUI application, which will install this script into the system."
    echo "Otherwise, this script will not run..."
    exit 1
fi


# Make sure the processing parameter file exists
proc_param_file="$1"
if [ ! -f $proc_param_file ]
then
    echo "Processing parameter file " $proc_param_file " does not exist"
    exit 1
fi


# Call the characterization algorithm for which a test CPF is to be
# generated.  In a production context, this would be prefaced with $IAS_BIN.
# Since this is a test context, however, it's assumed the current directory
# is the test directory; the app is in the parent directory.
../tirs_background_response_determination $proc_param_file
if [ $? -ne 0 ]
then
    echo "CPF group file creation failed"
    exit 1
fi


# Set the names of the CPF group to be merged.
group_name="TIRS_DETECTOR_RESPONSE"
group_name_blind="TIRS_DETECTOR_RESPONSE_BLIND"


# Get the list of names for each file
groupfile_names_b10=`ls tirs_cpf_background_response_B10*.odl`
groupfile_names_b11=`ls tirs_cpf_background_response_B11*.odl`
groupfile_names_blind=`ls tirs_cpf_background_response_B15*.odl`


# Look for the "base" CPF file to convert to a test CPF.  If it's not in
# the current working directory, get a copy
if [ ! -f $2 ]
then
    echo "CPF to modify $2 not found, getting copy"
    test_data_path="/edclxs110/data1/test_data"
    test_data_dir="ias/rps/tirs_background_response_determination"
    cp $test_data_path/$test_data_dir/$2 .
    if [ $? -ne 0 ]
    then
        echo "Cannot obtain base CPF to modify $2"
        exit 1
    fi
fi


# Make a "working" copy of the current basline CPF
tempfile="tempfile.cpf"
cp $2 $tempfile


# Create a "merged" TIRS_DETECTOR_RESPONSE CPF group file
merge_group_file="merge.odl"
echo "GROUP = TIRS_DETECTOR_RESPONSE" > $merge_group_file
echo "END_GROUP = TIRS_DETECTOR_RESPONSE" >> $merge_group_file
echo "END" >> $merge_group_file

merge_group_file_blind="merge_blind.odl"
echo "GROUP = TIRS_DETECTOR_RESPONSE_BLIND" > $merge_group_file
echo "END_GROUP = TIRS_DETECTOR_RESPONSE_BLIND" >> $merge_group_file
echo "END" >> $merge_group_file

echo "Merging individual band 10 background responses..."
temp2file="temp2.cpf"
temp3file="temp3.cpf"
cp $merge_group_file $temp2file
for file_name in $groupfile_names_b10
do
    $IAS_BIN/cpfmerge.sh $temp2file $file_name $group_name > $temp3file
    if [ $? -ne 0 ]
    then
        echo "TIRS band 10 group file merge failed:  " $file_name
        rm $tempfile $temp2file $merge_group_file $merge_group_file_blind
        exit 1
    fi
    mv $temp3file $temp2file
done

echo "Merging individual band 11 background responses..."
for file_name in $groupfile_names_b11
do
    $IAS_BIN/cpfmerge.sh $temp2file $file_name $group_name > $temp3file
    if [ $? -ne 0 ]
    then
        echo "TIRS band 11 group file merge failed:  " $file_name
        rm $tempfile $temp2file $merge_group_file $merge_group_file_blind
        exit 1
    fi
    mv $temp3file $temp2file
done
mv $temp2file $merge_group_file


echo "Merging individual blind band background responses..."
temp2file_blind="temp2_blind.cpf"
temp3file_blind="temp3_blind.cpf"
cp $merge_group_file_blind $temp2file_blind
for file_name in $groupfile_names_b15
do
    $IAS_BIN/cpfmerge.sh $temp2file_blind $file_name $group_name_blind > \
        $temp3file_blind
    if [ $? -ne 0 ]
    then
        echo "TIRS band 10 group file merge failed:  " $file_name
        rm $tempfile $temp2file
        exit 1
    fi
    mv $temp3file_blind $temp2file_blind
done
mv $temp2file_blind $merge_group_file_blind


# Merge the background responses in the group files into the main
# CPF
echo "Merging TIRS_DETECTOR_RESPONSE group into CPF..."
$IAS_BIN/cpfmerge.sh $tempfile $merge_group_file $group_name > $temp2file
if [ $? -ne 0 ]
then
    echo "TIRS_DETECTOR_RESPONSE group merge failed"
    rm $tempfile $merge_group_file $merge_group_file_blind
    exit 1
fi


# Merge the blind band group data
echo "Merging TIRS_DETECTOR_RESPONSE_BLIND group into CPF..."
$IAS_BIN/cpfmerge.sh $temp2file $merge_group_file_blind $group_name_blind > \
    $3
if [ $? -ne 0 ]
then
    echo "TIRS_DETECTOR_RESPONSE_BLIND group merge failed"
    rm $temp2file $merge_group_file $merge_group_file_blind
    exit 1
fi


# Done
echo "Creation of background response test CPF " $3 " completed."
rm $temp2file
rm $merge_group_file $merge_group_file_blind

exit 0
