/***************************************************************************
 NAME:     tirs_background_response_determination

 PURPOSE:  Determine the TIRS background response level based on analysis of
           a specified set of deep-space histogram statistics

 RETURNS:  Integer status code of SUCCESS or ERROR
****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "ias_cpf.h"
#include "ias_db.h"
#include "ias_l0r_constants.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_pixel_mask.h"
#include "ias_radiometry.h"
#include "ias_satellite_attributes.h"
#include "tirs_background_response_determination.h"


int main(int argc, char *argv[])
{
    const char *band_rps_proc_level = NULL;
    struct ias_db_connection *open_db = NULL;  /* Active DB connection */
    RUNNING_SUMS_DATA ****running_sums = NULL; /* Data buffer to hold running
                                                  sums for all CPFs in all
                                                  bands/SCAs/detectors */
    BACKGROUND_RESPONSE_DATA ***response_data = NULL;
                                               /* Data buffer to hold overall
                                                  background response
                                                  statistics for all
                                                  bands/SCAs/detectors */
    double *dark_response = NULL;              /* CPF baseline dark response
                                                  array for the current
                                                  band/SCA */
    IAS_PIXEL_MASK **pmasks_from_file = NULL;  /* Pixel mask array populated
                                                  from a pixel mask file*/
    IAS_PIXEL_MASK *curr_pmask = NULL;         /* Pixel mask for the
                                                  current band/SCA */
    IAS_SATELLITE_ID  satellite_id;            /* Satellite ID number
                                                  used to initialize
                                                  Satellite Attributes
                                                  Library */
    char **cpf_list = NULL;                    /* List of CPF names for
                                                  current scenes */
    PROC_PARAMS proc_params;                   /* Processing params from
                                                  ODL file */
    PROC_PARAMS_OMF proc_params_omf;           /* Processing params from
                                                  OMF file */
    WOID_DATA *woid_data = NULL;               /* Array of work order data
                                                  for each scene (wo ID,
                                                  acquisition date, CPF
                                                  file name) */
    int band_list[IAS_MAX_NBANDS];             /* List of bands to process */
    int *cpf_scene_count = NULL;               /* Array containing the
                                                  number of scenes acquired
                                                  with each distinct CPF */
    int num_masks_read = 0;                    /* Total number of masks read
                                                  from the pixel mask
                                                  file */
    int num_lines = 1;                         /* Number of "lines" of image
                                                  data in local pixel
                                                  mask */
    int num_scas;                              /* Number of SCAs */
    int num_detectors;                         /* Number of detectors */
    int num_image_bands;                       /* Number of imaging bands */
    int num_blind_bands;                       /* Number of blind bands */
    int num_bands;                             /* Total number of bands */
    int num_rad_woids;                         /* Number of radiometric
                                                  work order IDs found in
                                                  the specified date
                                                  range */
    int use_file_pmasks;                       /* Flag to indicate whether
                                                  to use pixel masks from
                                                  file (1) or use local
                                                  pixel masks (0) */
    int num_cpfs = 0;                          /* Total number of distinct
                                                  CPFs for the selected set
                                                  of scenes */
    int status;
    int band_index;                            /* Band list loop counter
                                                  (starts at 0) */
    int cpf_index;                             /* CPF list loop counter */
    int sca_index;                             /* SCA loop counter */
    int mask_index;                            /* Pixel mask loop counter */
    int ii;                                    /* General loop counter */


    /* Initialize logging */
    ias_log_initialize(argv[0]);

    /* Make sure we have the proper number of command line arguments.
       Should be just 1, the processing parameter file name */
    if (argc != 2)
    {
        IAS_LOG_ERROR("USAGE:  %s <name of parameter file>", argv[0]);
        exit(EXIT_FAILURE);
    }

    /* Get the required processing parameters */
    IAS_LOG_DEBUG("Reading processing parameter file %s...",
        argv[1]);
    memset(&proc_params, 0, sizeof(PROC_PARAMS));
    memset(&proc_params_omf, 0, sizeof(PROC_PARAMS_OMF));
    status = get_parameters(argv[1], &proc_params, &proc_params_omf);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Parameter processing failed");
        exit(EXIT_FAILURE);
    }

    /* Initialize the satellite attributes library.  The routine to obtain
       the satellite ID from the satellite number does not depend on
       the library already being initialized.  The initialization call
       essentially loads the attribute information for all bands of
       each sensor into memory... */
    IAS_LOG_DEBUG("Initializing IAS satellite attributes library...");
    satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(
        proc_params_omf.satellite_number);
    if (satellite_id == ERROR)
    {
        IAS_LOG_ERROR("Retrieving proper satellite ID");
        exit(EXIT_FAILURE);
    }
    status = ias_sat_attr_initialize(satellite_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing Satellite Attributes Library");
        exit(EXIT_FAILURE);
    }

    /* Get the total number of imaging bands.  As currently designed,
       TIRS has a set of "primary" imaging bands and a set of
       redundant or "secondary" imaging bands.  We're interested only
       in the primary set, which is selected by the IAS_NORMAL_BAND
       type value */
    IAS_LOG_DEBUG("Constructing TIRS band list...");
    memset(band_list, 0, IAS_MAX_NBANDS * sizeof(int));
    status = ias_sat_attr_get_sensor_band_numbers(IAS_TIRS,
        IAS_NORMAL_BAND, 0, band_list, IAS_MAX_NBANDS, &num_image_bands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Obtaining TIRS imaging band information");
        exit(EXIT_FAILURE);
    }

    /* Get the total number of blind bands.  As currently designed,
       there should be only 1 */
    status = ias_sat_attr_get_sensor_band_numbers(IAS_TIRS,
        IAS_BLIND_BAND, IAS_SECONDARY_BAND, &band_list[num_image_bands],
        (IAS_MAX_NBANDS - num_image_bands), &num_blind_bands);
    if ((status != SUCCESS) || (num_blind_bands != 1))
    {
        IAS_LOG_ERROR("Obtaining TIRS blind band information");
        exit(EXIT_FAILURE);
    }

    /* Calculate the total number of bands */
    num_bands = num_image_bands + num_blind_bands;

    /* Get the number of SCAs and number of detectors per SCA for
       the first imaging band.  As currently designed, these are
       the same for all TIRS bands */
    IAS_LOG_DEBUG("Obtaining number of TIRS SCAs and number of "
        "detectors per SCA...");
    num_scas = ias_sat_attr_get_sensor_sca_count(IAS_TIRS);
    num_detectors = ias_sat_attr_get_detectors_per_sca(band_list[0]);
    if ((num_scas == ERROR) || (num_detectors == ERROR))
    {
        IAS_LOG_ERROR("Obtaining TIRS SCA and/or detector information");
        exit(EXIT_FAILURE);
    }

    /* Determine whether we'll create local pixel masks or try to read
       them in from a file */
    IAS_LOG_DEBUG("Determining pixel mask processing capability...");
    if (strcmp(proc_params_omf.pmask_filename, "") != 0)
    {
        IAS_LOG_INFO("Pixel masks for each band/SCA will be read into an "
            "array from the file %s", proc_params_omf.pmask_filename);
        pmasks_from_file = ias_pm_read_array_from_file(
            proc_params_omf.pmask_filename, &num_masks_read);
        if (pmasks_from_file == NULL)
        {
            IAS_LOG_WARNING("Cannot read pixel file %s, switching to "
                "using locally created pixel masks instead",
                proc_params_omf.pmask_filename);
            use_file_pmasks = 0;
        }
        else
        {
            use_file_pmasks = 1;
        }
    }
    else
    {
        IAS_LOG_INFO("No pixel mask file specified, local pixel masks "
            "with inoperable detector information only will be created "
            "for each band/SCA" );
        use_file_pmasks = 0;
    }

    /* Determine the scene selection date range */
    IAS_LOG_DEBUG("Determining scene selection date range...");
    if ((strcmp(proc_params.start_date, "") != 0)
        && (strcmp(proc_params.end_date, "") != 0))
    {
        IAS_LOG_DEBUG("Scene selection date range completely specified");
    }
    else
    {
        /* Get the corresponding ending(starting) date to establish
           the scene selection date range */
        if (proc_params.num_days_in_interval < 0)
        {
            status = get_selection_date_from_interval(
                proc_params.end_date,
                proc_params.num_days_in_interval,
                proc_params.start_date);
        }
        else
        {
            status = get_selection_date_from_interval(
                proc_params.start_date,
                proc_params.num_days_in_interval,
                proc_params.end_date);
        }
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Determining scene selection date range from "
                "available information");
            if (pmasks_from_file)
            {
                for (mask_index = 0; mask_index < num_masks_read;
                     mask_index++)
                {
                    curr_pmask = pmasks_from_file[mask_index];
                    ias_pm_destroy(curr_pmask);
                }
                free(pmasks_from_file);
            }
            exit(EXIT_FAILURE);
        }
    }

   /***************** SCENE LIST CONSTRUCTION *************************/
    /* Establish a connection to the database */
    IAS_LOG_DEBUG("Establishing database connection...");
    open_db = connect_to_db();
    if (open_db == NULL)
    {
        IAS_LOG_ERROR("Establishing database connection");
        if (pmasks_from_file)
        {
            for (mask_index = 0; mask_index < num_masks_read;
                 mask_index++)
            {
                curr_pmask = pmasks_from_file[mask_index];
                ias_pm_destroy(curr_pmask);
            }
            free(pmasks_from_file);
        }
        exit(EXIT_FAILURE);
    }

    /* Get the radiometric work order data for the scenes */
    IAS_LOG_DEBUG("Obtaining work order information for set of selected "
        "scenes...");
    woid_data = get_rad_woid_data_in_date_range(open_db,
        proc_params.start_date, proc_params.end_date, &num_rad_woids);
    if (woid_data == NULL)
    {
        IAS_LOG_ERROR("Retrieving radiometric work order data for the "
            "selected scenes");
        cleanup_db(open_db);
        if (pmasks_from_file)
        {
            for (mask_index = 0; mask_index < num_masks_read;
                 mask_index++)
            {
                curr_pmask = pmasks_from_file[mask_index];
                ias_pm_destroy(curr_pmask);
            }
            free(pmasks_from_file);
        }
        exit(EXIT_FAILURE);
    }

    /* Logging message:  Total number of scenes acquired within date range */
    IAS_LOG_DEBUG("Number of scenes acquired within the specified date range "
        "%s to %s:  %d", proc_params.start_date, proc_params.end_date,
        num_rad_woids);

    /* Display message:  Scene information summary */
    printf("INTERVAL ID                  WORK ORDER ID"
        "           DATE ACQUIRED\n=============================="
        "========================================\n");
    for (ii = 0; ii < num_rad_woids; ii++)
    {
        printf("%-26s%-26s%-26s\n", woid_data[ii].landsat_interval_id,
           woid_data[ii].work_order_id, woid_data[ii].date_acquired);
    }
    printf("\n\n");

    /* Take the woid data and generate a common report header to be
       associated with the summary reports */
    if (proc_params.report_flag)
    {
        status = write_woid_data_header(woid_data, num_rad_woids,
            proc_params.start_date, proc_params.end_date);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Writing woid data header information to file");
            cleanup_db(open_db);
            free(woid_data);
            if (pmasks_from_file)
            {
                for (mask_index = 0; mask_index < num_masks_read;
                     mask_index++)
                {
                    curr_pmask = pmasks_from_file[mask_index];
                    ias_pm_destroy(curr_pmask);
                }
                free(pmasks_from_file);
            }
            exit(EXIT_FAILURE);
        }
    }

    /* Generate the list of CPFs and also get their total number */
    IAS_LOG_DEBUG("Constructing CPF list...");
    cpf_list = get_cpf_list(woid_data, num_rad_woids, &num_cpfs);
    if (cpf_list == NULL)
    {
        IAS_LOG_ERROR("Compiling CPF list for selected scenes");
        cleanup_db(open_db);
        free(woid_data);
        if (pmasks_from_file)
        {
            for (mask_index = 0; mask_index < num_masks_read;
                 mask_index++)
            {
                curr_pmask = pmasks_from_file[mask_index];
                ias_pm_destroy(curr_pmask);
            }
            free(pmasks_from_file);
        }
        exit(EXIT_FAILURE);
    }

    /* Count the number of scenes associated with each CPF */
    IAS_LOG_DEBUG("Counting number of scenes associated with each CPF in "
        "list...");
    cpf_scene_count = get_cpf_scene_count(woid_data, num_rad_woids,
        num_cpfs);
    if (cpf_scene_count == NULL)
    {
        IAS_LOG_ERROR("Getting scene counts for each CPF");
        cleanup_db(open_db);
        ias_misc_free_2d_array((void **)cpf_list);
        free(woid_data);
        if (pmasks_from_file)
        {
            for (mask_index = 0; mask_index < num_masks_read;
                 mask_index++)
            {
                curr_pmask = pmasks_from_file[mask_index];
                ias_pm_destroy(curr_pmask);
            }
            free(pmasks_from_file);
        }
        exit(EXIT_FAILURE);
    }

    /* Allocate the running sums data buffer.  Its size should be
       num_cpfs x num_bands x num_scas x num_detectors
       x sizeof(RUNNING_SUMS_DATA) */
    IAS_LOG_DEBUG("Allocating working data buffers...");
    running_sums = allocate_running_sums_buffer(num_cpfs, num_bands, num_scas,
        num_detectors);
    if (running_sums == NULL)
    {
        IAS_LOG_ERROR("Allocating running sums data buffer to "
            "accommodate %d CPFs x %d bands x %d SCAs x %d detectors, "
            "element size %d bytes", num_cpfs, num_bands, num_scas,
            num_detectors, (int)sizeof(RUNNING_SUMS_DATA));
        cleanup_db(open_db);
        ias_misc_free_2d_array((void **)cpf_list);
        free(cpf_scene_count);
        free(woid_data);
        if (pmasks_from_file)
        {
            for (mask_index = 0; mask_index < num_masks_read;
                 mask_index++)
            {
                curr_pmask = pmasks_from_file[mask_index];
                ias_pm_destroy(curr_pmask);
            }
            free(pmasks_from_file);
        }
        exit(EXIT_FAILURE);
    }

    /* Allocate the background response statistics data buffer.  Its
       size should be num_bands x num_scas x num_detectors x
       sizeof(BACKGROUND_RESPONSE_DATA) */
    response_data = allocate_background_response_buffer(num_bands, num_scas,
        num_detectors);
    if (response_data == NULL)
    {
        IAS_LOG_ERROR("Allocating background response data buffer to "
            "accommodate %d bands x %d SCAs x %d detectors, element "
            "size %d bytes", num_bands, num_scas, num_detectors,
            (int)sizeof(BACKGROUND_RESPONSE_DATA));
        cleanup_db(open_db);
        free_running_sums_buffer(running_sums, num_cpfs, num_bands, num_scas);
        ias_misc_free_2d_array((void **)cpf_list);
        free(cpf_scene_count);
        free(woid_data);
        if (pmasks_from_file)
        {
            for (mask_index = 0; mask_index < num_masks_read;
                 mask_index++)
            {
                curr_pmask = pmasks_from_file[mask_index];
                ias_pm_destroy(curr_pmask);
            }
            free(pmasks_from_file);
        }
        exit(EXIT_FAILURE);
    }

   /******************** CPF PROCESSING LOOP ****************************/
    /* Retrieve the current CPF in the list */
    for (cpf_index = 0; cpf_index < num_cpfs; cpf_index++)
    {
        IAS_CPF *cpf = NULL;
        const struct IAS_CPF_TIRS_DETECTOR_RESPONSE *det_response = NULL;
        const struct IAS_CPF_TIRS_DETECTOR_RESPONSE *det_response_blind = NULL;
        const struct IAS_CPF_DETECTOR_STATUS *det_status = NULL;
        const struct IAS_CPF_DETECTOR_STATUS *det_status_blind = NULL;
        double **scene_statistics_buffer = NULL;

        /* If we have an error in this section of code, we'll have to
           pass in NULL for the local_bandsca_pmask pointer until it
           comes into scope */
        IAS_LOG_DEBUG("Processing scenes associated with CPF %s...",
            cpf_list[cpf_index]);
        status = get_cpf_file(cpf_list[cpf_index],
            woid_data[cpf_index].date_acquired);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Retrieving CPF %s in list", cpf_list[cpf_index]);
            final_cleanup(cpf, open_db, use_file_pmasks, pmasks_from_file,
                num_masks_read, NULL, woid_data,
                cpf_scene_count, cpf_list, running_sums,
                response_data, scene_statistics_buffer, num_cpfs, num_bands,
                num_scas);
            exit(EXIT_FAILURE);
        }

        /* Read the current CPF's data */
        IAS_LOG_DEBUG("Obtaining processing parameter values from CPF "
            "%s...", cpf_list[cpf_index]);
        cpf = ias_cpf_read(cpf_list[cpf_index]);
        if (cpf == NULL)
        {
            IAS_LOG_ERROR("Reading CPF %s in CPF list", cpf_list[cpf_index]);
            final_cleanup(cpf, open_db, use_file_pmasks, pmasks_from_file,
                num_masks_read, NULL, woid_data,
                cpf_scene_count, cpf_list, running_sums,
                response_data, scene_statistics_buffer, num_cpfs, num_bands,
                num_scas);
            exit(EXIT_FAILURE);
        }

        /* Get the detector response information (which contains the base
           dark response) */
        det_response = ias_cpf_get_tirs_det_response(cpf);
        det_response_blind = ias_cpf_get_tirs_det_response_blind(cpf);
        if ((det_response == NULL) || (det_response_blind == NULL))
        {
            IAS_LOG_ERROR("Extracting TIRS_DETECTOR_RESPONSE(_BLIND) "
                "group data from CPF %s", cpf_list[cpf_index]);
            final_cleanup(cpf, open_db, use_file_pmasks, pmasks_from_file,
                num_masks_read, NULL, woid_data,
                cpf_scene_count, cpf_list, running_sums,
                response_data, scene_statistics_buffer, num_cpfs, num_bands,
                num_scas);
            exit(EXIT_FAILURE);
        }

        /* Read in the detector status information */
        det_status = ias_cpf_get_det_status(cpf);
        det_status_blind = ias_cpf_get_tirs_det_status_blind(cpf);
        if ((det_status == NULL) || (det_status_blind == NULL))
        {
            IAS_LOG_ERROR("Getting DETECTOR_STATUS group data from CPF %s",
                cpf_list[cpf_index]);
            final_cleanup(cpf, open_db, use_file_pmasks, pmasks_from_file,
                num_masks_read, NULL, woid_data,
                cpf_scene_count, cpf_list, running_sums,
                response_data, scene_statistics_buffer, num_cpfs, num_bands,
                num_scas);
            exit(EXIT_FAILURE);
        }

        /* Allocate the scene statistics data buffer for the current
           band/SCA */
        IAS_LOG_DEBUG("Allocating scene statistics data buffer for scenes "
            "from CPF %s", cpf_list[cpf_index]);
        scene_statistics_buffer = (double **)ias_misc_allocate_2d_array(
            cpf_scene_count[cpf_index], num_detectors, sizeof(double));
        if (scene_statistics_buffer == NULL)
        {
            IAS_LOG_ERROR("Allocating scene statistics data buffer");
            final_cleanup(cpf, open_db, use_file_pmasks, pmasks_from_file,
                num_masks_read, NULL, woid_data, 
                cpf_scene_count, cpf_list, running_sums,
                response_data, scene_statistics_buffer, num_cpfs, num_bands,
                num_scas);
            exit(EXIT_FAILURE);
        }

        /* Logging message:  List of scenes (actually work order IDs)
           to characterize using information from the current CPF */
        printf("Scenes currently being processed...\n");
        printf("CPF NAME                 WORK ORDER ID\n"
            "=====================================\n");
        for (ii = 0; ii < num_rad_woids; ii++)
        {
            if (woid_data[ii].cpf_index == cpf_index)
            {
                printf("%-26s%-26s\n", cpf_list[cpf_index],
                    woid_data[ii].work_order_id);
            }
        }        printf("\n");

    /***************** BAND/SCA-LEVEL PROCESSING LOOP *********************/
        for (band_index = 0; band_index < num_bands; band_index++)
        {
            int status_band_index;  /* Index into detector status data */
            int dark_band_index;    /* Index into baseline dark response
                                       data */
            /* Pull the current band number from the band list */
            int band_number = band_list[band_index];

            /* Get the current band's attributes */
            const IAS_BAND_ATTRIBUTES *band_attributes =
                ias_sat_attr_get_band_attributes(band_number);
            if (band_attributes == NULL)
            {
                IAS_LOG_ERROR("Obtaining band attribute information for "
                    "band %d", band_number);
                final_cleanup(cpf, open_db, use_file_pmasks, pmasks_from_file,
                    num_masks_read, NULL, woid_data,
                    cpf_scene_count, cpf_list, running_sums,
                    response_data, scene_statistics_buffer, num_cpfs,
                    num_bands, num_scas);
                exit(EXIT_FAILURE);
            }

            /* Detector status information for the TIRS imaging bands
               (currently, 10 and 11) is obtained from
               ias_cpf_get_detector_status().  To access this information,
               the detector status array needs to be indexed relative
               to the band number.  The blind band detector status and
               the dark response arrays need to be indexed starting
               from 0.

               Set the RPS processing level appropriate for the current
               band.  */
            if (band_attributes->band_classification == IAS_NORMAL_BAND)
            {
                status_band_index = band_attributes->band_index;
                dark_band_index = band_index;
                band_rps_proc_level = proc_params.rps_proc_level;
            }
            else
            {
                status_band_index = band_index - num_image_bands;
                dark_band_index = band_index - num_image_bands;
                band_rps_proc_level = proc_params.rps_proc_level_blind;
            }

            for (sca_index = 0; sca_index < num_scas; sca_index++)
            {
                IAS_PIXEL_MASK *local_bandsca_pmask = NULL;
                int *inop_list = NULL;    /* Inoperable detector list for
                                             the current band/SCA */
                int inop_count = 0;       /* Number of inoperable detectors
                                             in the current band/SCA */

                /* Convert the index to the current SCA number */
                int sca_number = (sca_index + 1);

                /*  Get the histogram statistics for all scenes in
                    the current band/SCA */
                IAS_LOG_DEBUG("Retrieving histogram statistics for band %d "
                    "SCA %d... for scenes processed with CPF %s",
                    band_number, sca_number, cpf_list[cpf_index]);
                status = get_histogram_statistics(open_db,
                    woid_data, num_rad_woids,
                    (const char *)cpf_list[cpf_index],
                    cpf_scene_count[cpf_index], 
                    band_number, sca_number, num_detectors,
                    band_rps_proc_level,
                    scene_statistics_buffer);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Obtaining histogram statistics for "
                        "band %d SCA %d for scenes acquired within CPF %s "
                        "effective date range...",
                        band_number, sca_number, cpf_list[cpf_index]);
                    final_cleanup(cpf, open_db, use_file_pmasks,
                        pmasks_from_file, num_masks_read,
                        local_bandsca_pmask, woid_data,
                        cpf_scene_count, cpf_list, running_sums,
                        response_data, scene_statistics_buffer, num_cpfs,
                        num_bands, num_scas);
                    exit(EXIT_FAILURE);
                }

                /* Set the correct baseline dark response */
                if (band_index < num_image_bands)
                {
                    dark_response = det_response
                        ->baseline_dark_response[dark_band_index][sca_index];
                }
                else
                {
                    dark_response = det_response_blind
                        ->baseline_dark_response[dark_band_index][sca_index];
                }

                /* Subtract the baseline dark response */
                subtract_dark_response(scene_statistics_buffer,
                    (const double *)dark_response,
                    cpf_scene_count[cpf_index], num_detectors);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Subtracting dark response for "
                        "band %d SCA %d scene data associated with "
                        "CPF %s", band_number, sca_number,
                        cpf_list[cpf_index]);
                    final_cleanup(cpf, open_db, use_file_pmasks,
                        pmasks_from_file, num_masks_read,
                        local_bandsca_pmask, woid_data,
                        cpf_scene_count, cpf_list, running_sums,
                        response_data, scene_statistics_buffer, num_cpfs,
                        num_bands, num_scas);
                    exit(EXIT_FAILURE);
                }

                /* Establish the pixel mask for the current band/SCA */
                if (use_file_pmasks) /* Get a mask from the array */
                {
                    mask_index = ias_pm_get_mask_index(pmasks_from_file,
                        num_masks_read, band_number, sca_number);
                    if (mask_index == ERROR)
                    {
                        IAS_LOG_ERROR("Obtaining pixel mask from array "
                            "in band %d SCA %d", band_number,
                            sca_number);
                        final_cleanup(cpf, open_db, use_file_pmasks,
                            pmasks_from_file, num_masks_read,
                            local_bandsca_pmask, woid_data,
                            cpf_scene_count, cpf_list, running_sums,
                            response_data, scene_statistics_buffer, num_cpfs,
                            num_bands, num_scas);
                        exit(EXIT_FAILURE);
                    }
                    curr_pmask = pmasks_from_file[mask_index];
                }
                else  /* Just use the inoperable detector information
                         directly from the CPF */
                {
                    if (band_attributes->band_classification
                        == IAS_NORMAL_BAND)
                    {
                        inop_count = det_status
                            ->inoperable_count[status_band_index][sca_index];
                        if (inop_count != 0)
                        {
                            inop_list = det_status
                                ->inoperable[status_band_index][sca_index];
                        }
                        else
                        {
                            inop_list = NULL;
                        }
                    }
                    else
                    {
                        inop_count = det_status_blind
                            ->inoperable_count[status_band_index][sca_index];
                        if (inop_count != 0)
                        {
                            inop_list = det_status_blind
                                ->inoperable[status_band_index][sca_index];
                        }
                        else
                        {
                            inop_list = NULL;
                        }
                    }

                    local_bandsca_pmask = create_local_pixel_mask(inop_list,
                        inop_count, band_number, sca_number, num_detectors,
                        num_lines);
                    if (local_bandsca_pmask == NULL)
                    {
                        IAS_LOG_ERROR("Creating inoperable detector "
                            "pixel mask for band %d SCA %d",
                            band_number, sca_number);
                        final_cleanup(cpf, open_db, use_file_pmasks,
                            pmasks_from_file, num_masks_read,
                            local_bandsca_pmask, woid_data,
                            cpf_scene_count, cpf_list, running_sums,
                            response_data, scene_statistics_buffer, num_cpfs,
                            num_bands, num_scas);
                        exit(EXIT_FAILURE);
                    }
                    curr_pmask = local_bandsca_pmask;
                }

                /* Calculate the detector statistics running sums/counts for
                   the current band/SCA in the current CPF */
                status = calculate_background_response_sums_for_cpf(
                    (const double **)scene_statistics_buffer,
                    (const IAS_PIXEL_MASK *)curr_pmask,
                    cpf_scene_count[cpf_index], cpf_index, band_number,
                    band_index, sca_number, num_detectors, running_sums);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Calculating background response "
                        "CPF-level statistics sums/counts for scenes from "
                        "CPF %s, band %d SCA %d", cpf_list[cpf_index],
                        band_number, sca_number);
                    final_cleanup(cpf, open_db, use_file_pmasks,
                        pmasks_from_file, num_masks_read,
                        local_bandsca_pmask, woid_data,
                        cpf_scene_count, cpf_list, running_sums,
                        response_data, scene_statistics_buffer, num_cpfs,
                        num_bands, num_scas);
                    exit(EXIT_FAILURE);
                }

                /* Destroy the (locally created) pixel mask for the current
                   band/SCA */
                if ((!use_file_pmasks) && (local_bandsca_pmask != NULL))
                {
                    ias_pm_destroy(local_bandsca_pmask);
                    local_bandsca_pmask = NULL;
                }
            }   /* END SCA LEVEL PROCESSING LOOP */
        }   /* END BAND LEVEL PROCESSING LOOP */

        /* Free up the scene statistics data buffer.  Reset the base
           pointer to NULL to prevent issues during cleanup of an
           error condition before it's reallocated.  Pass in NULL for
           the local_bandsca_pmask pointer as it's gone out of scope */
        status = ias_misc_free_2d_array((void **)scene_statistics_buffer);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Freeing the scene statistics data buffer");
            final_cleanup(cpf, open_db, use_file_pmasks,
                    pmasks_from_file, num_masks_read,
                    NULL, woid_data,
                    cpf_scene_count, cpf_list, running_sums,
                    response_data, scene_statistics_buffer, num_cpfs,
                    num_bands, num_scas);
            exit(EXIT_FAILURE);
        }
        scene_statistics_buffer = NULL;
    }   /* END CPF PROCESSING LOOP */

/******************** FINAL AGGREGATION TO BACKGROUND RESPONSE **********/

    /* We have the required running sums.  Now we need to aggregate these to
       the final background response estimate (and associated standard
       deviation) for each detector.

       If we get an error from now on, we'll have to pass in NULL for the cpf,
       local_bandsca_pmask, and scene_statistics_buffer pointers -- they're
       now out of scope */
    IAS_LOG_DEBUG("Aggregating scene statistics to overall background "
        "response statistics...");
    for (band_index = 0; band_index < num_bands; band_index++)
    {
        /* Get the current band number from the list */
        int band_number = band_list[band_index];

        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            int sca_number = (sca_index + 1);

            /* Calculate the overall background response for the current
               band/SCA */
            status = calculate_overall_background_response(
                (const RUNNING_SUMS_DATA ****)running_sums, num_cpfs,
                band_number, band_index, sca_number, num_detectors,
                response_data);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Determining final background response "
                    "for band %d SCA %d", band_number, sca_number);
                final_cleanup(NULL, open_db, use_file_pmasks,
                    pmasks_from_file, num_masks_read, NULL,
                    woid_data, cpf_scene_count, cpf_list,
                    running_sums, response_data, NULL,
                    num_cpfs, num_bands, num_scas);
            }

            /* Write the statistics to a CPF group */
            status = write_background_response_means_to_cpf_group(
                (const BACKGROUND_RESPONSE_DATA ***)response_data,
                band_number, band_index, sca_number, num_detectors);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Writing background response statistics "
                    "to CPF group for blind band SCA %d", sca_number);
                if (pmasks_from_file == NULL)
                    ias_pm_destroy(curr_pmask);
                final_cleanup(NULL, open_db, use_file_pmasks,
                    pmasks_from_file, num_masks_read, NULL,
                    woid_data, cpf_scene_count, cpf_list,
                    running_sums, response_data, NULL,
                    num_cpfs, num_bands, num_scas);
                exit(EXIT_FAILURE);
            }

            /* Write the detector standard deviations to the summary
               report if specified in the processing parameters */
            if (proc_params.report_flag)
            {
                status = write_background_response_stdevs_to_report(
                    (const BACKGROUND_RESPONSE_DATA ***)response_data,
                    band_number, band_index, sca_number, num_detectors);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Writing background response stdev "
                        "to summary report for band %d SCA %d",
                        band_number, sca_number);
                    final_cleanup(NULL, open_db, use_file_pmasks,
                        pmasks_from_file, num_masks_read,
                        NULL, woid_data, cpf_scene_count,
                        cpf_list, running_sums, response_data,
                        NULL, num_cpfs, num_bands,
                        num_scas);
                    exit(EXIT_FAILURE);
                }
            }
        }   /* END SCA-LEVEL BACKGROUND RESPONSE STATISTICS LOOP */
    }   /* END BAND LEVEL BACKGROUND RESPONSE STATISTICS LOOP */

    /* Final cleanup */
    final_cleanup(NULL, open_db, use_file_pmasks, pmasks_from_file,
        num_masks_read, NULL, woid_data,
        cpf_scene_count, cpf_list,running_sums, response_data,
        NULL, num_cpfs, num_bands, num_scas);

    IAS_LOG_INFO("TIRS background response determination successfully "
        "completed.");
    return EXIT_SUCCESS;
}   /* END tirs_background_response_determination */
