#ifndef TIRS_BACKGROUND_RESPONSE_DETERMINATION_H
#define TIRS_BACKGROUND_RESPONSE_DETERMINATION_H

/* Common C #includes */
#include <stdio.h>
#include <limits.h>

/* IAS #includes */
#include "ias_const.h"
#include "ias_cpf.h"
#include "ias_db.h"
#include "ias_l0r_constants.h"
#include "ias_linked_list.h"
#include "ias_pixel_mask.h"
#include "ias_radiometry.h"
#include "ias_types.h"

/* Constants indicating which "direction" to calculate a date to define
   a date range */
#define DATE_AFTER_START_DATE 0
#define DATE_BEFORE_END_DATE  1


/* RPS processing levels */
#define RPS_BKRESPONSE_PROC_LEVEL_IMAGE    "0L"
#define RPS_BKRESPONSE_PROC_LEVEL_BLIND    "0."


/* Misc. constants */
#define MAX_WOID_LENGTH           20
#define DEFAULT_NUM_DAYS          30
#define MAX_DB_CPFNAME_LENGTH    256
#define INVALID_DATA       -100000.0


/*           Data structures required for parameter I/O. */
typedef struct proc_params
{
    /* Start and end dates defining a date range from which to select
       scenes to analyze */
    char start_date[IAS_L0R_DATE_LENGTH + 1];
    char end_date[IAS_L0R_DATE_LENGTH + 1];

    /* Work order-related information -- RPS processing level */
    char rps_proc_level[IAS_RPS_PROCESSING_LEVEL_LENGTH + 1];
    char rps_proc_level_blind[IAS_RPS_PROCESSING_LEVEL_LENGTH + 1];

    /* Work order-related information -- processing options */
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    int num_days_in_interval;       /* Number of days prior to the
                                       current date--used to select
                                       scene(s) to characterize.  This
                                       value can be positive, negative,
                                       or 0...*/
    int report_flag;                /* Flag indicating whether to
                                       generate a summary background
                                       response report (0=no, 1=yes) */

} PROC_PARAMS;


typedef struct omf_params
{
    /* Satellite number */
    int satellite_number;

    /* Name of pixel mask file to use in processing */
    char pmask_filename[PATH_MAX];
} PROC_PARAMS_OMF;



/*                  "Data" data structures               */
typedef struct woid_data
{
    char landsat_interval_id[IAS_L0R_INTERVAL_ID_LENGTH + 1];
    char date_acquired[IAS_L0R_DATE_LENGTH + 1];
    char cpf_filename[MAX_DB_CPFNAME_LENGTH + 1];
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char rad_wo_common_id[IAS_CHAR_ID_SIZE];
    int cpf_index;
} WOID_DATA;


typedef struct running_sums_data
{
    double sum;
    double sum_sqr;
    int count;

} RUNNING_SUMS_DATA;


typedef struct background_response_data
{
    double mean;
    double stdev;
    int valid_count;
} BACKGROUND_RESPONSE_DATA;




/*************************************************************************
                   FUNCTION PROTOTYPES
**************************************************************************/
                      /* Parameter I/O */
int get_parameters
(
    const char *proc_params_filename,    /* I: Name of work order
                                            parameter file */
    PROC_PARAMS *proc_params,            /* O: ODL processing parameters */
    PROC_PARAMS_OMF *proc_params_omf     /* O: OMF processing parameters */
);



                      /* Memory management */
void free_running_sums_buffer
(
    RUNNING_SUMS_DATA ****buffer,  /* I: Data buffer to free */
    int num_cpfs,                  /* I: Number of CPFs */
    int num_bands,                 /* I: Number of bands */
    int num_scas                   /* I: Number of SCAs per band */
);


RUNNING_SUMS_DATA ****allocate_running_sums_buffer
(
    int num_cpfs,                  /* I: Number of CPFs */
    int num_bands,                 /* I: Number of bands */
    int num_scas,                  /* I: Number of SCAs per band */
    int num_detectors              /* I: Number of detectors */
);


void free_background_response_buffer
(
    BACKGROUND_RESPONSE_DATA ***response, /* I: Data buffer to free */
    int num_bands,                        /* I: Number of bands */
    int num_scas                          /* I: Number of SCAs */
);


BACKGROUND_RESPONSE_DATA ***allocate_background_response_buffer
(
    int num_bands,                /* I: Total number of bands */
    int num_scas,                 /* I: Total number of SCAs */
    int num_detectors             /* I: Total number of detectors */
);


void final_cleanup
(
    IAS_CPF *cpf,                           /* I: Populated CPF structure */
    struct ias_db_connection *open_db,      /* I: Open DB connection */
    int use_file_pmasks,                    /* I: Array of pixel masks read
                                               from a file (can be NULL if
                                               local pixel masks are used) */
    IAS_PIXEL_MASK **pmasks_from_file,      /* I: Array of pixel masks read
                                               from a file (can be NULL if
                                               local pixel masks are used) */
    int num_masks_read,                     /* I: Number of masks read from
                                               the pixel file */
    IAS_PIXEL_MASK *local_bandsca_pmask,    /* I: Local pixel mask created
                                               for the current band/SCA */
    WOID_DATA *woid_data,                   /* I: Work order data */
    int *cpf_scene_count,                   /* I: Number of scenes / CPF */
    char **cpf_list,                        /* I: List of CPF names */
    RUNNING_SUMS_DATA ****running_sums,     /* I: Running sums data */
    BACKGROUND_RESPONSE_DATA ***response,   /* I: Background response
                                               statistics */
    double **scene_statistics_buffer,       /* I: Scene statistics for the
                                               current band/SCA */
    int num_cpfs,                           /* I: Total number of CPFs */
    int num_bands,                          /* I: Number of bands */
    int num_scas                            /* I: Number of SCAs */
);



                      /* CPF processing */
int *get_cpf_scene_count
(
    const WOID_DATA *woid_data,      /* I: Scene/work order data */
    int num_rad_woids,               /* I: Number of scenes */
    int total_cpf_count              /* I: Number of distinct CPFs */
);


int get_cpf_file
(
    char cpf_filename[],                 /* I/O: Name of current CPF file */
    const char *date_acquired            /* I: Acquisition date for interval
                                            of interest */
);


char **get_cpf_list
(
    WOID_DATA *woid_data,                /* I/O: Array of work order data
                                            for each scene */
    int num_rad_woids,                   /* I: Number of scenes */
    int *distinct_cpf_count              /* O: Total number of distinct
                                            CPFs */
);


IAS_PIXEL_MASK *create_local_pixel_mask
(
    const int *inop_list,                        /* I: Inoperable detector
                                                    list from CPF */
    int inop_count,                              /* I: Inoperable detector
                                                    list from CPF */
    int band_number,                             /* I: Band number */
    int sca_number,                              /* I: SCA number */
    int num_detectors,                           /* I: Number of detectors
                                                    for the current
                                                    band/SCA */
    int num_lines                                /* I: Number of lines of
                                                    data in the pixel
                                                    mask */
);



                      /* Calculate background response */
void subtract_dark_response
(
    double **scene_statistics_buffer,      /* I/O: Dark response corrected
                                              histogram statistics for all
                                              scenes in the current CPF,
                                              for the current band/SCA */
    const double *dark_response,           /* I: Baseline dark response for
                                              the currently processing
                                              band/SCA from the currently
                                              processing CPF */
    int scene_count,                       /* I: Number of scenes for the
                                              currently processing CPF */
    int num_detectors                      /* I: Number of detectors in the
                                               current band/SCA */
);


int calculate_background_response_sums_for_cpf
(
    const double **scene_statistics_buffer, /* I:  Detector-specific
                                               scene statistics for all
                                               scenes in the current
                                               band/SCA */
    const IAS_PIXEL_MASK *pmask,            /* I: Pixel mask for the
                                               current band/SCA */
    int scene_count,                        /* I: Number of scenes
                                               associated with the
                                               current CPF */
    int cpf_index,                          /* I: Current CPF */
    int band_number,                        /* I: Current band number */
    int band_index,                         /* I: Array index corresponding
                                               to the current band number */
    int sca_number,                         /* I: Current SCA number */
    int num_detectors,                      /* I: Number of detectors in
                                               the current band/SCA */
    RUNNING_SUMS_DATA ****rsums             /* O: Populated running sums
                                               array (num_cpfs x num_bands
                                               x num_scas x num_dets) */
);


int calculate_overall_background_response
(
    const RUNNING_SUMS_DATA ****rsums,       /* I: Running sums data */
    int total_cpf_count,                     /* I: Number of CPFs */
    int band_number,                         /* I: Current band number */
    int band_index,                          /* I: Array index to access
                                                current band */
    int sca_number,                          /* I: Array index to access
                                                current SCA */
    int num_detectors,                       /* I: Number of detectors in
                                                the current band/SCA */
    BACKGROUND_RESPONSE_DATA ***response_data
                                             /* O: Populated background
                                                response data buffer for
                                                all bands/SCAs/detectors */
);



                     /* Database access */
void cleanup_db
(
    struct ias_db_connection *open_db      /* I: Open database connection */
);


struct ias_db_connection *connect_to_db();


int get_selection_date_from_interval
(
    char *base_date,                   /* I/O: Base starting or ending date
                                          (if empty, will be populated with
                                          current date/time string) */
    int num_days_in_interval,          /* I: Specified interval of days
                                          before current date */
    char *base_plus_interval_date      /* O: Date accounting for
                                          the specified interval and
                                          interval direction */
);


WOID_DATA *get_rad_woid_data_in_date_range
(
    struct ias_db_connection *db,     /* I: Database connection */
    const char *start_date,           /* I: Start date for date range */
    const char *end_date,             /* I: End date for date range */
    int *num_rad_woids                /* I: Total number of radiometric
                                         work order IDs */
);


int get_histogram_statistics
(
    struct ias_db_connection *db,     /* I: Open database connection */
    WOID_DATA *woid_data,       /* I: Scene/work order data */
    int num_rad_woids,          /* I: Number of woid_data */
    const char *cpf_name,             /* I: Name of processing CPF */
    int cpf_scene_count,              /* I: Number of scenes associated
                                         with the current CPF */
    int band_number,                  /* I: Current 1-based band number */
    int sca_number,                   /* I: Current 1-based SCA number */
    int detectors_per_sca,            /* I: Number of detectors for the
                                         current band/SCA */
    const char *proc_level_string,    /* I: RPS processing level for current
                                         band */
    double **scene_statistics_buffer  /* O: Scene statistics for all
                                         detectors within the current
                                         band/SCA */
);



                   /* Output processing */
int write_woid_data_header
(
    const WOID_DATA *woid_data,       /* I: Work order ID data to write
                                         to the header file */
    int num_rad_woids,                /* I: Number of scenes */
    const char *start_date,           /* I: Selected starting date */
    const char *end_date              /* I: Selected ending date */
);


int write_background_response_means_to_cpf_group
(
    const BACKGROUND_RESPONSE_DATA ***response_data,
                                      /* I: Overall background response
                                         statistics */
    int band_number,                  /* I: 1-based band number */
    int band_index,                   /* I: Array access for band dimension
                                         of data buffer */
    int sca_number,                   /* I: 1-based SCA number */
    int num_detectors                 /* I: Total number of detectors in
                                         the current band/SCA */
);


int write_background_response_stdevs_to_report
(
    const BACKGROUND_RESPONSE_DATA ***response_data,
                                         /* I: Overall background response
                                            statistics */
    int band_number,                     /* I: 1-based band number */
    int band_index,                      /* I: Access for current band
                                            into background response
                                            arrays */
    int sca_number,                      /* I: 1-based SCA number */
    int num_detectors                    /* I: Total number of detectors
                                            in the current band/SCA */
);



#endif
