/****************************************************************************
 NAME:     write_background_response_means_to_cpf_group

 PURPOSE:  Writes the detector-specific background response to a CPF
           group file

 RETURNS:  Integer status code of SUCCESS or ERROR

 NOTES:    band_index = {0, 1, 2, ..., etc} relative to the band list
           generated in the main application
****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "tirs_background_response_determination.h"


#define GROUP_NAME_STRLEN 100


int write_background_response_means_to_cpf_group
(
    const BACKGROUND_RESPONSE_DATA ***response_data,
                                      /* I: Overall background response
                                         statistics */
    int band_number,                  /* I: 1-based band number */
    int band_index,                   /* I: Array access for band dimension
                                         of data buffer */
    int sca_number,                   /* I: 1-based SCA number */
    int num_detectors                 /* I: Total number of detectors in
                                         the current band/SCA */
)
{
    char blind_band_name[IAS_BAND_NAME_SIZE];
    char group_name[GROUP_NAME_STRLEN] = "";   /* Group "opening" statement */
    char endgroup_name[GROUP_NAME_STRLEN] = "";/* Group "closing" statement */
    char groupfile_name[PATH_MAX] = "";  /* Group file name */
    FILE *group_fptr = NULL;             /* File pointer */
    int det_index = 0;                   /* Detector loop counter */
    int sca_index = sca_number - 1;      /* SCA index */
    int blind_band_number;               /* Blind band number */
    int status = SUCCESS;


    /* Get the blind band name and number.  We'll need this as a reference */
    status = ias_sat_attr_get_band_name_from_type_and_class(
       IAS_THERMAL1_BAND, IAS_BLIND_BAND, blind_band_name);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving TIRS blind band name");
        return ERROR;
    }
    blind_band_number = ias_sat_attr_convert_band_name_to_number(
        blind_band_name);
    if (blind_band_number == ERROR)
    {
        IAS_LOG_ERROR("Retrieving TIRS blind band number");
        return ERROR;
    }

    /* Construct the group file name */
    status = snprintf(groupfile_name, sizeof(groupfile_name),
        "tirs_cpf_background_response_B%02d_SCA%02d.odl",
        band_number, sca_number);
    if ((status < 0) || (status >= sizeof(groupfile_name)))
    {
        IAS_LOG_ERROR("Constructing file name for band %d SCA %d "
            "CPF background response group", band_number, sca_number);
        return ERROR;
    }

    /* Try to create the file */
    group_fptr = fopen(groupfile_name, "w");
    if (group_fptr == NULL)
    {
        IAS_LOG_ERROR("Creating group file %s", groupfile_name);
        return ERROR;
    }

    /* Create the GROUP and END_GROUP strings */
    if (band_number == blind_band_number)
    {
        strcpy(group_name, "GROUP = TIRS_DETECTOR_RESPONSE_BLIND");
        strcpy(endgroup_name, "END_GROUP = TIRS_DETECTOR_RESPONSE_BLIND");
    }
    else
    {
        strcpy(group_name, "GROUP = TIRS_DETECTOR_RESPONSE");
        strcpy(endgroup_name, "END_GROUP = TIRS_DETECTOR_RESPONSE");
    }

    /* Start writing to the file with the group name */
    status = fprintf(group_fptr, "%s\n", group_name);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing starting group header line %s to file %s",
            group_name, groupfile_name);
        fclose(group_fptr);
        return ERROR;
    }

    /* Write the parameter name to the file */
    status = fprintf(group_fptr, "  Background_Response_B%02d_SCA%02d = ( ",
        band_number, sca_number);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing band %d SCA %d parameter name to group file "
            "%s", band_number, sca_number, groupfile_name);
        fclose(group_fptr);
        return ERROR;
    }

    /* Write the background response for each detector to the file */
    for (det_index = 0; det_index < num_detectors; det_index++)
    {
        if ((det_index) % 5 == 0)
        {
            status = fprintf(group_fptr, "\n  %11.6lf,",
                response_data[band_index][sca_index][det_index].mean);
        }
        else if (det_index == (num_detectors - 1))
        {
            status = fprintf(group_fptr, "%11.6lf )\n",
                response_data[band_index][sca_index][det_index].mean);
        }
        else
        {
            status = fprintf(group_fptr, "%11.6lf,",
                response_data[band_index][sca_index][det_index].mean);
        }
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing detector %d background response "
                "to group file %s\nparameter value=%11.6lf",
                (det_index + 1), groupfile_name,
                response_data[band_index][sca_index][det_index].mean);
            fclose(group_fptr);
            return ERROR;
        }
    }

    /* Write the "END_GROUP" and "END" statements to the file */
    status = fprintf(group_fptr, "%s\nEND\n", endgroup_name);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing ending group closing line %s to file %s",
            endgroup_name, groupfile_name);
        fclose(group_fptr);
        return ERROR;
    }

    /* Cleanup */
    fclose(group_fptr);

    return SUCCESS;
}   /* END write_background_response_means_to_cpf_group */
