/*************************************************************************
 NAME:     write_background_response_stdevs_to_report

 PURPOSE:  Writes the background response std deviations to a summary
           report.  This is an option that can be specified in the
           work order

 RETURNS:  Integer status code of SUCCESS or ERROR
**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <float.h>
#include <math.h>
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_radiometry.h"
#include "tirs_background_response_determination.h"


int write_background_response_stdevs_to_report
(
    const BACKGROUND_RESPONSE_DATA ***response_data,
                                          /* I: Overall background response
                                             statistics */
    int band_number,                      /* I: 1-based band number */
    int band_index,                       /* I: Access for current band
                                             into background response
                                             arrays */
    int sca_number,                       /* I: 1-based SCA number */
    int num_detectors                     /* I: Total number of detectors
                                             in the current band/SCA */
)
{
    char reportfile_name[PATH_MAX] = "";
    FILE *rpt_fptr = NULL;                /* Open report file handle */
    double sum_response = 0.0;            /* Background response sum */
    double sum_stdev = 0.0;               /* Sum of std devss */
    double sum_sqr_response = 0.0;        /* Response squared sum */
    double sum_sqr_stdev = 0.0;           /* Sum of squared std devs */
    double sca_max_response = -DBL_MIN;   /* Max. background response */
    double sca_min_response = DBL_MAX;    /* Min. background response */
    double sca_mean_response;             /* SCA mean response */
    double sca_stdev_response;            /* SCA std dev of responses
                                             about the mean */
    double sca_max_stdev = -DBL_MIN;      /* Max. std deviation */
    double sca_min_stdev = DBL_MAX;       /* Min. std devation */
    double sca_mean_stdev;                /* SCA mean of std devs */
    double sca_stdev_stdev;               /* SCA std dev of std devs 
                                             about the mean */
    double var_response;                  /* Response variance */
    double var_stdev;                     /* Variance of std deviations */
    int det_index;                        /* Detector loop counter */
    int sca_index = sca_number - 1;       /* Current SCA's index */
    int status = SUCCESS;
    int valid_count = 0;


    /* Construct the file name */
    status = snprintf(reportfile_name, sizeof(reportfile_name),
        "background_response_summary_B%02d_SCA%02d.txt",
        band_number, sca_number);
    if ((status < 0) || (status >= sizeof(reportfile_name)))
    {
        IAS_LOG_ERROR("Constructing summary report file name");
        return ERROR;
    }

    /* Try to create the file */
    rpt_fptr = fopen(reportfile_name, "w");
    if (rpt_fptr == NULL)
    {
        IAS_LOG_ERROR("Opening file %s", reportfile_name);
        return ERROR;
    }

    /* Write the detector data to the file */
    status = fprintf(rpt_fptr, "BAND %02d SCA%02d\nSUMMARY BACKGROUND "
        "RESPONSE AND 1-SIGMA, BY DETECTOR (DN):\n\n", band_number,
        sca_number);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing summary statistics information to report");
        fclose(rpt_fptr);
        return ERROR;
    }

    /* Calculate the SCA-level summary mean and standard deviation */
    for (det_index = 0; det_index < num_detectors; det_index++)
    {
        double current_det_mean =
            response_data[band_index][sca_index][det_index].mean;
        double current_det_stdev =
            response_data[band_index][sca_index][det_index].stdev;

        if (current_det_stdev == INVALID_DATA)
        {
            status = fprintf(rpt_fptr, "Det %03d:  INOPERABLE -- N/A\n",
                (det_index + 1));
        }
        else
        {

            /* Find the min/max background response values */
            if (current_det_mean > sca_max_response)
                sca_max_response = current_det_mean;
            if (current_det_mean < sca_min_response)
                sca_min_response = current_det_mean;

            /* Find the min/max std deviation values */
            if (current_det_stdev > sca_max_stdev)
                sca_max_stdev = current_det_stdev;
            if (current_det_stdev < sca_min_stdev)
                sca_min_stdev = current_det_stdev;

            /* Compute the running sums for the mean response, mean
               std deviation, and the squared sums for the response
               std deviation and "std deviation std deviation" */
            valid_count++;
            sum_response += current_det_mean;
            sum_sqr_response += (current_det_mean * current_det_mean);
            sum_stdev += current_det_stdev;
            sum_sqr_stdev += (current_det_stdev * current_det_stdev);

            /* Print out the detector information */
            status = fprintf(rpt_fptr, "Det %03d:   %8.3lf +- %9.6lf\n",
                (det_index + 1), current_det_mean, current_det_stdev);
        }
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing detector %d background response "
                "statistics to band %d SCA %d report file", (det_index + 1),
                band_number, sca_number);
            fclose(rpt_fptr);
            return ERROR;
        }
    }
    if (valid_count == 0)
    {
        IAS_LOG_ERROR("No valid background response data in band %d SCA %d",
            band_number, sca_number);
        fclose(rpt_fptr);
        return ERROR;
    }

    /* Finish calculating the SCA level statistics.  If the variance is
       less than 0 (shouldn't happen ordinarily), set the calculated
       standard deviation to 0.0 */
    sca_mean_response = sum_response / valid_count;
    sca_mean_stdev = sum_stdev / valid_count;

    var_response = (sum_sqr_response
        - (valid_count * (sca_mean_response * sca_mean_response)))
        / (valid_count - 1);
    if (var_response < 0.0)
        sca_stdev_response = 0.0;
    else
        sca_stdev_response = sqrt(var_response);

    var_stdev = (sum_sqr_stdev
        - (valid_count * (sca_mean_stdev * sca_mean_stdev)))
        / (valid_count - 1);
    if (var_stdev < 0.0)
        sca_stdev_stdev = 0.0;
    else
        sca_stdev_stdev = sqrt(var_stdev);

    /* Write the SCA summary statistics to the file */
    status = fprintf(rpt_fptr,
        "\n\nBAND %d SCA %d SUMMARY STATISTICS:\n\n", band_number,
        sca_number);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing background response summary statistics "
            "header string to band %d SCA %d report file", band_number,
            sca_number);
        fclose(rpt_fptr);
        return ERROR;
    }

    status = fprintf(rpt_fptr, "MAX. BACKGROUND RESPONSE (DN):  %8.3lf\n",
        sca_max_response);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing SCA max. background response "
            "to band %d SCA %d report file", band_number, sca_number);
        fclose(rpt_fptr);
        return ERROR;
    }
    status = fprintf(rpt_fptr, "MIN. BACKGROUND RESPONSE (DN):  %8.3lf\n",
        sca_min_response);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing SCA min. background response "
            "to band %d SCA %d report file", band_number, sca_number);
        fclose(rpt_fptr);
        return ERROR;
    }
    status = fprintf(rpt_fptr, "MEAN BACKGROUND RESPONSE (DN):  %8.3lf\n",
        sca_mean_response);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing SCA mean background response "
            "to band %d SCA %d report file", band_number, sca_number);
        fclose(rpt_fptr);
        return ERROR;
    }
    status = fprintf(rpt_fptr, "STD DEV BACKGROUND RESPONSE (DN):  "
        "%9.6lf\n\n",
        sca_stdev_response);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing SCA background response std deviation "
            "to band %d SCA %d report file", band_number, sca_number);
        fclose(rpt_fptr);
        return ERROR;
    }

    status = fprintf(rpt_fptr, "MAX. RESPONSE STDEV (DN):  %9.6lf\n",
        sca_max_stdev);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing SCA max. response std deviation "
            "to band %d SCA %d report file", band_number, sca_number);
        fclose(rpt_fptr);
        return ERROR;
    }
    status = fprintf(rpt_fptr, "MIN. RESPONSE STDEV (DN):  %9.6lf\n",
        sca_min_stdev);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing SCA min. response std deviation "
            "to band %d SCA %d report file", band_number, sca_number);
        fclose(rpt_fptr);
        return ERROR;
    }
    status = fprintf(rpt_fptr, "MEAN RESPONSE STDEV (DN):  %9.6lf\n",
        sca_mean_stdev);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing SCA mean response std deviation "
            "to band %d SCA %d report file", band_number, sca_number);
        fclose(rpt_fptr);
        return ERROR;
    }
    status = fprintf(rpt_fptr, "STD DEV RESPONSE STD DEV (DN):  "
        "%9.6lf\n", sca_stdev_stdev);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing std deviation of response std deviation "
            "to band %d SCA %d report file", band_number, sca_number);
        fclose(rpt_fptr);
        return ERROR;
    }

    /* If less than 90% of detectors are found to have valid data
       (i.e. 10% or more of the detectors are inoperable), write a final
       summary warning to the report.
       If desired, this could be a user-selected processing parameter
       threshold value*/
    if (valid_count < (int)(0.9 * num_detectors))
    {
        status = fprintf(rpt_fptr, "\n\nWARNING: >= 10%% of "
            "detectors are classed as inoperable--further analysis should "
            "be conducted to verify/refute issues with the data and/or"
            "the background response determination algorithm.  Statistics "
            "reported here should be considered 'suspect' until then...");
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing detector status warning message to "
                "band %d SCA %d report file", band_number, sca_number);
            fclose(rpt_fptr);
            return ERROR;
        }
    }

    /* Close the completed report */
    fclose(rpt_fptr);

    return SUCCESS;
}   /* END write_background_response_stdevs_to_report */
