/*************************************************************************
 NAME:     write_woid_data_header

 PURPOSE:  Routine to write interval ID, acquisition date, and associated
           CPF name for a selected set of TIRS deep space collects to a
           header file associated with the optional background response
           characterization summary reports

 RETURNS:  Integer status code of SUCCESS or ERROR
**************************************************************************/
#include <stdio.h>
#include <limits.h>
#include "ias_logging.h"
#include "tirs_background_response_determination.h"


int write_woid_data_header
(
    const WOID_DATA *woid_data,       /* I: Work order ID data to write
                                         to the header file */
    int num_rad_woids,                /* I: Number of scenes */
    const char *start_date,           /* I: Selected starting date */
    const char *end_date              /* I: Selected ending date */
)
{
    char woid_header_filename[PATH_MAX];
    FILE *woid_fptr = NULL;
    int woid_index;                   /* Work order loop counter */
    int status;


    /* Create the file name and the file itself */
    status = snprintf(woid_header_filename, sizeof(woid_header_filename),
        "background_response_summary_header.txt");
    if ((status < 0) || (status >= sizeof(woid_header_filename)))
    {
        IAS_LOG_ERROR("Constructing work order data header file name");
        return ERROR;
    }
    woid_fptr = fopen(woid_header_filename, "w");
    if (woid_fptr == NULL)
    {
        IAS_LOG_ERROR("Creating header file");
        return ERROR;
    }

    /* Write the scene selection starting and ending dates first */
    status = fprintf(woid_fptr, "SCENE SELECTION START DATE:  %s\n",
        start_date);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing scene selection start date to header file");
        fclose(woid_fptr);
        return ERROR;
    }
    status = fprintf(woid_fptr, "SCENE SELECTION END DATE:    %s\n\n\n",
        end_date);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing scene selection end date to header file");
        fclose(woid_fptr);
        return ERROR;
    }

    /* Now write the scene information */
    status = fprintf(woid_fptr, "INTERVAL ID               WO ID"
        "               DATE ACQUIRED\n"
        "==============================================================\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing data header string to header file");
        fclose(woid_fptr);
        return ERROR;
    }

    for (woid_index = 0; woid_index < num_rad_woids; woid_index++)
    {
        status = fprintf(woid_fptr, "%-26s%-20s%-26s\n",
            woid_data[woid_index].landsat_interval_id,
            woid_data[woid_index].work_order_id,
            woid_data[woid_index].date_acquired);
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing interval ID/work order ID/acquisition "
                "date information for scene %d to header file\n"
                "interval ID=%s\nwork order ID=%s\nacquisition date=%s",
                (woid_index + 1), woid_data[woid_index].landsat_interval_id,
                woid_data[woid_index].work_order_id,
                woid_data[woid_index].date_acquired);
            fclose(woid_fptr);
            return ERROR;
        }
    }

    status = fprintf(woid_fptr, "\n\nINTERVAL ID               CPF NAME"
        "\n=====================================================\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing interval_id/CPF name data header string to "
            "header file");
        fclose(woid_fptr);
        return ERROR;
    }
    for (woid_index = 0; woid_index < num_rad_woids; woid_index++)
    {
        status = fprintf(woid_fptr, "%-26s%-26s\n",
            woid_data[woid_index].landsat_interval_id,
            woid_data[woid_index].cpf_filename);
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing interval ID/CPF name information for "
                "scene %d to header file\ninterval ID=%s\nCPF name=%s",
                (woid_index + 1), woid_data[woid_index].landsat_interval_id,
                woid_data[woid_index].cpf_filename);
            fclose(woid_fptr);
            return ERROR;
        }
    }

    /* Close the header file */
    fclose(woid_fptr);

    return SUCCESS;
}   /* END write_woid_data_header */
