/**************************************************************************
 NAME:              dark_response_determination_db

 PURPOSE:  Module implementing routines to provide IAS characterization
           database access for the standalone TIRS dark response
           determination algorithm.

**************************************************************************/
#include <stdlib.h>
#include <float.h>

#include "ias_db.h"
#include "ias_db_get_connect_info.h"
#include "ias_db_insert.h"
#include "ias_db_query.h"
#include "ias_logging.h"

#include "tirs_dark_response_determination_standalone.h"

#define SQL_QUERYLEN 256

/*--------------------------------------------------------------------
 NAME:                  connect_to_db

 PURPOSE:    Retrieves IAS database connection information and attempts
             to make the connection

 RETURNS:    - Pointer to valid ias_db_struct, indicating connection has
               been established
             - NULL pointer if connection cannot be made
-----------------------------------------------------------------------*/
struct ias_db_connection *connect_to_db()
{
    char *db_name = NULL;            /* Required connection information */
    char *user_id = NULL;
    char *user_password = NULL;
    char *db_host = NULL;

    struct ias_db_connection *db = NULL;   /* Open database connection */

    int status = SUCCESS;


    /* Initialize the database library */
    IAS_LOG_DEBUG("Initializing database library...");
    status = ias_db_initialize_database_lib();
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing database library");
        return NULL;
    }

    /* Get the connection information */
    IAS_LOG_DEBUG("Obtaining database connection information...");
    status = ias_db_get_connect_info(&db_name, &user_id, &user_password,
        &db_host);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Obtaining database connection information");
        return NULL;
    }

    /* Try to connect */
    IAS_LOG_DEBUG("Connecting to database...");

    db = ias_db_connect_to_database(db_name, user_id, user_password,
               db_host);

    /* Free the allocated strings with the connection information */
    free(db_name);
    free(user_id);
    free(user_password);
    free(db_host);

    /* Did we connect? */
    if (db == NULL)             /* NO -- log an error */
    {
        IAS_LOG_ERROR("Connecting to database");
    }
    else                        /* YES -- log success only in DEBUG mode */
    {
        IAS_LOG_DEBUG("Database connection established");
    }

    /* Return whatever we've got for *db (either NULL or a valid pointer) */
    return db;
}




/*-------------------------------------------------------------------------
 NAME:           get_histogram_statistics

 PURPOSE:     Queries the HISTOGRAM_STATISTICS_CAL table for the
              mean values extracted from the TIRS blind band 
              product for the desired work order

 RETURNS:     Integer status code of SUCCESS or ERROR

---------------------------------------------------------------------------*/
int get_blind_band_histogram_statistics
(
    struct ias_db_connection *db,     /* I: Open database connection */
    char *rad_wo_common_id,           /* I: Database ID corresponding to
                                         work order ID */
    int ref_blind_band_number,        /* I: Current 1-based band number */
    int sca_number,                   /* I: Current 1-based SCA number */
    int detectors_per_sca,            /* I: Number of detectors for the
                                         current band/SCA */
    const char *rps_proc_level,       /* I: RPS processing level
                                         identifier */
    IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *blind_stats
                                      /* O: Blind band mean histogram
                                         statistics for the current
                                         SCA */
)
{
    char query_string[SQL_QUERYLEN] = "";
                                      /* SQL query string */

    char db_error_message[IAS_SQL_RTN_MSGLEN] = "";
                                      /* Query error message */

    struct ias_db_query *query = NULL;    /* Query data */

    int ii = 0;                           /* Detector loop counter */
    int status = SUCCESS;


    /* Prepare the query string.  It MUST include the identifier for the
       processing level '0L'.  It is assumed here that histogram
       statistics values are calculated for every detector, whether the
       data are valid or not*/
    sprintf(query_string, "SELECT detector_id, scene_pixel_mean "
            "FROM HISTOGRAM_STATISTICS_CAL WHERE band_id=%d AND "
            "sensor_chip_assembly_id=%d AND "
            "radiometric_wo_common_id=%s AND processing_level='%s' "
            "ORDER BY detector_id;", ref_blind_band_number, sca_number,
            rad_wo_common_id, rps_proc_level);

    /* Execute the query and verify it worked */
    query = ias_db_perform_query(db, query_string);

    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
                sizeof(db_error_message));
        IAS_LOG_ERROR("Query of HISTOGRAM STATISTICS CAL failed:  %s",
                db_error_message);
        return ERROR;
    }

    /* Move DB "pointer" to the first valid record. */
    if (!ias_db_query_next(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
                sizeof(db_error_message));
        IAS_LOG_ERROR("Accessing HISTOGRAM_STATISTICS_CAL data:  %s",
                db_error_message);
        ias_db_query_close(query);
        return ERROR;
    }

    /* Populate the blind band mean data buffer.  Only the mean
       histogram statistics value for each detector is populated;
       everything else is zeroed out */
    for (ii = 0; ii < detectors_per_sca; ii++)
    {
        int detector_number;           /* 1-based detector ID */
        int detector_index;            /* 0-based index corresponding to
                                          detector ID */

        status = ias_db_query_int_value(query, &detector_number, 0);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Cannot read blind band histogram statistics "
                    "detector_id value at index %d in SCA %d",
                    ii, sca_number);
            ias_db_query_close(query);
            return ERROR;
        }
        detector_index = (detector_number - 1);

        status = ias_db_query_double_value(query,
            &(blind_stats[detector_index].mean[0]), 1);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Cannot obtain histogram statistics "
                    "mean value for detector %d in SCA %d",
                    detector_number, sca_number);
            ias_db_query_close(query);
            return ERROR;
        }

        /* Advance DB "pointer" to next record as long as we're
           not at the last detector */
        if (ii < (detectors_per_sca - 1))
        {
            if (!ias_db_query_next(query))
            {
                IAS_LOG_ERROR("Cannot move to next detector %d's records",
                              (ii + 1));
                ias_db_query_close(query);
                return ERROR;
            }
        }
    }

    /* Release query resources */
    ias_db_query_close(query);

    /* Finished. */
    return SUCCESS;
}
