/***************************************************************************
  NAME:                      get_parameters

  PURPOSE:  Reads the input parameter files and populates the
            required parameter data structures

  RETURNS:  Integer status code of SUCCESS or ERROR

  NOTES:    If parameter help is successfully provided, application
            will immediately exit

****************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "ias_radiometry.h"

#include "tirs_dark_response_determination_standalone.h"


#define MAX_NUM_WO_PARAMS 5
#define MAX_NUM_OMF_PARAMS 3


int get_parameters
(
    const char *wo_params_filename,      /* I: Work order filename */
    WO_PARAMS  *wo_param_data,           /* O: Work order parameters */
    OMF_PARAMS *omf_param_data           /* O: OMF parameters */
)
{
    /* Internal OMF parameter file name */
    char omf_params_filename[PATH_MAX] = "";

    /* RPS processing level string */
    const char *default_rps_proc_level[] =
        {RPS_DARKRESPONSE_PROC_LEVEL_BLIND};

    /* Default relative gain set is current.  It's assumed that for the
       initial CPFs, pre-launch, post-launch, and current relative gains
       are the same */
    int default_relgain_option = CURRENT_RELGAINS;
    int default_pmask_option   = NO_USE_PMASKS_FROM_FILE;
    int default_dark_response_option = NO_USE_CPF_DARK_RESPONSE;
                                     /* Calculate adjusted dark response */

    int status = SUCCESS;            /* Status for each read
                                        operation */
    int read_count = 0;              /* Count of parameters
                                        successfully read */

/*-------------------------------------------------------------------------
                   READ WORK ORDER PARAMETER FILE
---------------------------------------------------------------------------*/
    /* Declare the work order parameter table */
    IAS_PARM_DECLARE_TABLE(wo_param_table, MAX_NUM_WO_PARAMS);

    /* Do common parameters first -- the work order ID and the RPS
       processing levels needed for the blind band histogram statistics */
    IAS_PARM_WORK_ORDER_ID(wo_param_table, wo_param_data->work_order_id,
        sizeof(wo_param_data->work_order_id), 1);

    IAS_PARM_TIRS_RPS_PROC_LEVEL_DN_BLIND(wo_param_table,
        IAS_PARM_OPTIONAL, default_rps_proc_level,
        wo_param_data->rps_proc_level,
        sizeof(wo_param_data->rps_proc_level));

    /* Now the more custom parameters */
    IAS_PARM_ADD_INT(wo_param_table, RELGAIN_OPTIONS,
        "Relative gain source selection (pre, post, current)",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, PRELAUNCH_RELGAINS,
        CURRENT_RELGAINS, 1, &default_relgain_option,
        &(wo_param_data->rg_option),
        sizeof(wo_param_data->rg_option), 1);

    /* Allow for the possibility of reading pixel mask information from
       a pixel mask file.  By default, the pixel mask is created locally
       for each band/SCA, and just loaded with inoperable detector
       information */
    IAS_PARM_ADD_BOOLEAN(wo_param_table, USE_PMASKS_FROM_FILE,
        "Use pixel masks from pixel mask file", IAS_PARM_OPTIONAL,
        IAS_PARM_NOT_ARRAY, &default_pmask_option,
        &(wo_param_data->pmask_option),
        sizeof(wo_param_data->pmask_option), 1);

    /* Allow for the possibilities of using CPF baseline dark response
       and writing dark response data to an ODL-formatted group data
       file */
    IAS_PARM_ADD_BOOLEAN(wo_param_table, USE_CPF_DARK_RESPONSE,
        "Use CPF baseline dark response", IAS_PARM_OPTIONAL,
        IAS_PARM_NOT_ARRAY, &default_dark_response_option,
        &wo_param_data->dark_response_option,
        sizeof(wo_param_data->dark_response_option),
        1);

    /* Get the total number of items to read from the parameter table */
    read_count = IAS_PARM_GET_TABLE_SIZE(wo_param_table);

    /* If someone enters -- help, --template, etc, on the command line
       when starting the application, the appropriate information will
       be displayed.  If someone enters a file name, its access will be
       checked.  If a help option was specified or the specified file name
       could not be accessed, a 1 will be returned.  In this case, just
       copy the work order file name (or help option) to the OMF
       file name and continue.  Otherwise, read the work order file */
    status = ias_parm_provide_help(wo_params_filename, wo_param_table,
        read_count, IAS_INPUT_PARAMETERS);

    if (status == ERROR)
    {
        IAS_LOG_ERROR("Work order parameter help failed");
        return ERROR;
    }
    else if (status == 1)
    {
        sprintf(omf_params_filename, "%s", wo_params_filename);
    }
    else
    {
        status = ias_parm_read(wo_params_filename, wo_param_table,
            read_count);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading work order parameters from file %s",
                wo_params_filename);
            return ERROR;
        }

        /* Construct the full path to the OMF file */
        sprintf(omf_params_filename, "%s.omf", wo_param_data->work_order_id);
    }


/*-----------------------------------------------------------------------
                         READ OMF PARAMETER FILE
-------------------------------------------------------------------------*/
    /* Declare the OMF parameter table */
    IAS_PARM_DECLARE_TABLE(omf_param_table, MAX_NUM_OMF_PARAMS);

    /* Common parameters */
    IAS_PARM_SATELLITE(omf_param_table, &omf_param_data->satellite_number,
        sizeof(omf_param_data->satellite_number), 1);

    IAS_PARM_CPF(omf_param_table, IAS_PARM_REQUIRED,
        omf_param_data->cpf_filename,
        sizeof(omf_param_data->cpf_filename), 1);

    IAS_PARM_PIXEL_MASK_FILENAME(omf_param_table, IAS_PARM_OPTIONAL,
        omf_param_data->pmask_filename,
        sizeof(omf_param_data->pmask_filename), 1);

    /* Get the total number of items to read in the OMF parameter table */
    read_count = IAS_PARM_GET_TABLE_SIZE(omf_param_table);

    /* See if the user wanted help or a template this time with the
       OMF file.  If not, just read the file */
    status = ias_parm_provide_help(omf_params_filename, omf_param_table,
        read_count, IAS_OMF_PARAMETERS);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("OMF parameter help failed");
        return ERROR;
    }
    else if (status == 1)
    {
        exit(EXIT_SUCCESS);
    }

    /* Read the OMF parameters */
    status = ias_parm_read(omf_params_filename, omf_param_table,
        read_count);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading OMF parameter file");
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}
