#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <math.h>

#include "cgreen.h"

#include "ias_db.h"
#include "ias_logging.h"
#include "ias_radiometry.h"
#include "ias_satellite_attributes.h"

#include "../tirs_dark_response_determination_standalone.h"

#ifndef FALSE
#define FALSE  0
#endif

#define STRING_BUFFER_SIZE 1000
#define DATA_BUFFER_SIZE 5

/* Failure codes */
#define EXIT_FAIL_NO_DB_SUPPORT    100
#define EXIT_FAIL_NO_TEST_DATA_DIR 200
#define EXIT_FAIL_NO_SAT_ID        300
#define EXIT_FAIL_NO_SAT_LIB       400

static char work_order_id[IAS_WORKORDER_ID_SIZE] = "I30000";
static char *test_data_path = NULL;
static char test_data_dir[] =
    "ias/rps/tirs_dark_response_determination_standalone";
static char system_command[STRING_BUFFER_SIZE] = "";
static char cpf_basename[] = "L8CPF20090101_20090331.01";
static char full_cpf_name[PATH_MAX];


/*-------------------------------------------------------------------------
 NAME:                 check_group_files

 PURPOSE:  Internal routine used to compare an adjusted dark response
           output file produced by the TIRS dark response determination
           algorithm to a "reference" output file

 RETURNS:  Integer status code of SUCCESS or ERROR.  ERROR is only
           returned if the test and/or reference file cannot be read

--------------------------------------------------------------------------*/
static int check_group_files(const char *test_group_filename,
    const char *reference_group_filename, double pass_threshold)
{
    char test_buff[STRING_BUFFER_SIZE];    /* String buffer used for
                                              "test" data file read */
    char reference_buff[STRING_BUFFER_SIZE];  /* String buffer used for
                                              "source" data file read */

    FILE *test_fptr = NULL;                /* "Test" data file */
    FILE *reference_fptr = NULL;              /* "Source data file for
                                              comparison */

    double test_response[DATA_BUFFER_SIZE];/* Array of test dark responses
                                              from a line in the test
                                              output file */
    double reference_response[DATA_BUFFER_SIZE];
                                           /* Array of source dark responses
                                              from a line in the source
                                              output file */
    double diff = 0;

    int ii = 0;                            /* Loop counter */
    int jj = 0;                            /* Loop counter */
    int response_diffs = 0;                /* # of differences in dark
                                              response (for a given detector
                                              between the "test" data and the
                                              "source" data) exceeding
                                              the specified threshold
                                              value*/
    int status = SUCCESS;


    /* Open the group files */
    test_fptr = fopen(test_group_filename, "r");
    reference_fptr = fopen(reference_group_filename, "r");

    assert_not_equal_with_message(test_fptr, NULL, "Cannot open test "
        "group file %s", test_group_filename);
    assert_not_equal_with_message(reference_fptr, NULL, "Cannot open "
        "reference group file %s", reference_group_filename);
    if ((test_fptr == NULL) || (reference_fptr == NULL))
    {
        if (test_fptr)
            fclose(test_fptr);
        if (reference_fptr)
            fclose(reference_fptr);
        return ERROR;
    }

    /* Check the GROUP header and parameter names to make sure the
       test one is correctly written */
    fgets(test_buff, STRING_BUFFER_SIZE, test_fptr);
    fgets(reference_buff, STRING_BUFFER_SIZE, reference_fptr);
    assert_string_equal_with_message(test_buff, reference_buff,
        "GROUP header statement difference");

    fgets(test_buff, STRING_BUFFER_SIZE, test_fptr);
    fgets(reference_buff, STRING_BUFFER_SIZE, reference_fptr);
    assert_string_equal_with_message(test_buff, reference_buff,
        "Parameter name difference");

    /* Check each line of data.  As currently implemented,
       there are 5 dark response values per line.  For 640
       detectors, that translates into 128 lines of dark response
       values */
    for (ii = 0; ii < 128; ii++)
    {
        status = fscanf(test_fptr,
            "%lf,%lf,%lf,%lf,%lf%*[ ,)]\n",
            &test_response[0], &test_response[1],
            &test_response[2], &test_response[3],
            &test_response[4]);
        if (status < 0)
        {
            assert_true_with_message(FALSE, "Cannot read test line %d",
                (ii + 1));
            fclose(test_fptr);
            fclose(reference_fptr);
            return ERROR;
        }

        status = fscanf(reference_fptr,
            "%lf,%lf,%lf,%lf,%lf%*[ ,)]\n",
            &reference_response[0], &reference_response[1],
            &reference_response[2], &reference_response[3],
            &reference_response[4]);
        if (status < 0)
        {
            assert_true_with_message(FALSE, "Cannot read reference line %d",
                (ii + 1));
            fclose(test_fptr);
            fclose(reference_fptr);
            return ERROR;
        }

        /* Check the absolute difference for each set of dark responses.
           Count those that are greater than the passing threshold */
        for (jj = 0; jj < DATA_BUFFER_SIZE; jj++)
        {
            diff = fabs(test_response[jj] - reference_response[jj]);
            if (diff > pass_threshold)
            {
                response_diffs++;
            }
        }
    }

    assert_equal_with_message(response_diffs, 0,
        "%d differences of 1e-3 or greater detected");

    /* Check the END_GROUP and END statements */
    fgets(test_buff, STRING_BUFFER_SIZE, test_fptr);
    fgets(reference_buff, STRING_BUFFER_SIZE, reference_fptr);
    assert_string_equal_with_message(test_buff, reference_buff,
        "END_GROUP header statement difference");

    fgets(test_buff, STRING_BUFFER_SIZE, test_fptr);
    fgets(reference_buff, STRING_BUFFER_SIZE, reference_fptr);
    assert_string_equal_with_message(test_buff, reference_buff,
        "END closing statement difference");

    /* Close the (blind band) files */
    fclose(test_fptr);
    fclose(reference_fptr);

    /* Done */
    return SUCCESS;
}



/* Test application parameter help processing */
Ensure test_end2end_parameter_help()
{
    int status = SUCCESS;
    int size;

    /* Construct the system command */
    size = snprintf(system_command,sizeof(system_command),
        "../tirs_dark_response_determination_standalone --help ");
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return;
    }

    /* Run the system command */
    status = system(system_command);
    assert_equal_with_message(status, EXIT_SUCCESS, "Parameter help should "
        "have caused non-fatal exit");

    /* Done */

}



/* Major end-to-end system test */
Ensure test_end2end_process()
{
    char test_group_filename[PATH_MAX] = "";
    char reference_group_filename[PATH_MAX] = "";

    int num_scas = 3;                   /* Total number of TIRS SCAs */
    int sca_index = 0;                  /* SCA loop counter */
    int sca_number = 0;                 /* Current (1-based) SCA number */
    int status = SUCCESS;               /* Function return status */

    double pass_threshold = 1e-3;       /* Maximum difference between
                                           implementation results and
                                           prototype results */
    int size;

    /* Prep the database */
    printf("Checking for test blind band histogram statistics in "
       "database...\n");

    size = snprintf(system_command, sizeof(system_command),
        "./check_or_insert_woid.csh %s", work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return;
    }
    status = system(system_command);
    if (status != 0)
    {
        assert_true_with_message(FALSE, "Cannot prep database--insert "
            "failed");
        return;
    }

    /* Get the parameter files ready */
    /* Replace the $IAS_UNIT_TEST_DATA_DIR placeholder in the OMF file with
       its actual value */
    size = snprintf(system_command, sizeof(system_command),
        "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' %s_template.omf > %s.omf", 
        test_data_path, work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return;
    }
    status = system(system_command);
    assert_equal_with_message(status, FALSE, "Cannot edit OMF template");
    if (status != 0)
    {
        return;
    }

    /* Rename the template ODL file as the actual ODL file */
    size = snprintf(system_command, sizeof(system_command),
        "cp %s_template.odl %s.odl", work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return;
    }
    status = system(system_command);
    assert_equal_with_message(status, FALSE, "Cannot create working "
        "parameter file");
    if (status != 0)
    {
        size = snprintf(system_command, sizeof(system_command),
            "rm %s.omf", work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Run the app (stand back and cross the fingers!) */
    printf("Running tirs_dark_response_determination_standalone app...\n");
    size = snprintf(system_command,sizeof(system_command),
        "../tirs_dark_response_determination_standalone %s.odl",
        work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return;
    }
    status = system(system_command);
    if (status != EXIT_SUCCESS)
    {
        assert_true_with_message(FALSE, "Cannot complete end-to-end "
            "processing");
        size = snprintf(system_command, sizeof(system_command),
            "rm %s.omf %s.odl", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Check the blind band output files produced */
    for (sca_index = 0; sca_index < num_scas; sca_index++)
    {
        sca_number = (sca_index + 1);

        size = snprintf(test_group_filename,sizeof(test_group_filename),
            "cpf_dark_response_group_B15_SCA%02d.odl",
            sca_number);
        if ((size < 0) || (size >= sizeof(test_group_filename)))
        {
            assert_true_with_message(FALSE, "Filename string too short, "
               "it needs %d bytes", size);
            return;
        }
        size = snprintf(reference_group_filename,
            sizeof(reference_group_filename),
            "%s/%s/cpf_adrgroup_B15_SCA%02d.odl",
            test_data_path, test_data_dir, sca_number);
        if ((size < 0) || (size >= sizeof(reference_group_filename)))
        {
            assert_true_with_message(FALSE, "Filename string too short, "
               "it needs %d bytes", size);
            return;
        }

        status = check_group_files(test_group_filename,
            reference_group_filename, pass_threshold);
        if (status != SUCCESS)
        {
            assert_true_with_message(FALSE, "Test output file does not "
               "compare with reference output file in SCA %d",
               sca_number);
        }
    }

    /* Done */
    size = snprintf(system_command, sizeof(system_command),
         "rm cpf_dark_response_group*.odl ");
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return;
    }
    system(system_command);
}



Ensure test_get_blind_band_scene_means()
{
    char source_stats_filename[PATH_MAX] = "";
    char band_name[IAS_BAND_NAME_SIZE];

    FILE *source_stats_fptr = NULL;

    struct ias_db_connection *open_db = NULL;

    IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *db_stats = NULL;

    double *source_stats = NULL;
    double diff = 0.0;

    int status = SUCCESS;
    int det_index = 0;
    int blind_band_number = 0;
    int sca_number = 1;
    int number_detectors = 640;
    char rad_wo_common_id[IAS_CHAR_ID_SIZE];
    int size;

    IAS_BAND_TYPE band_type = IAS_THERMAL1_BAND;
    IAS_BAND_CLASSIFICATION band_class = IAS_BLIND_BAND;


    printf("Testing retrieval of blind band SCA 1 scene means...\n");

    /* Get the test data directory path from the environment */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_not_equal_with_message(test_data_path, NULL, "Cannot find "
        "test data directory");
    if (test_data_path == NULL)
    {
        return;
    }

    /* Get the TIRS blind band number */
    status = ias_sat_attr_get_band_name_from_type_and_class(band_type,
        band_class, band_name);
    assert_equal_with_message(status, SUCCESS, "Cannot obtain current "
       "blind band name");
    if (status != SUCCESS)
        return;

    blind_band_number = ias_sat_attr_convert_band_name_to_number(band_name);
    assert_not_equal_with_message(blind_band_number, ERROR,
        "Cannot obtain TIRS blind band number");
    if (blind_band_number == ERROR)
    {
        return;
    }

    /* Allocate the scene means data buffers */
    db_stats = calloc(number_detectors,
        sizeof(IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS));
    assert_not_equal_with_message(db_stats, NULL, "Cannot allocate "
        "database scene means data buffer");
    if (db_stats == NULL)
    {
        return;
    }

    source_stats = calloc(number_detectors,
        sizeof(*source_stats));
    assert_not_equal_with_message(source_stats, NULL,
        "Cannot allocate source scene means data buffer");
    if (source_stats == NULL)
    {
        free(db_stats);
        return;
    }

    /* Read in the test scene means data */
    size = snprintf(source_stats_filename, sizeof(source_stats_filename), 
        "%s/%s/sc_B%02d_SCA%02d.txt", test_data_path, test_data_dir, 
        blind_band_number, sca_number);
    if ((size < 0) || (size >= sizeof(source_stats_filename)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
               "it needs %d bytes", size);
        return;
    }
    source_stats_fptr = fopen(source_stats_filename, "r");
    assert_not_equal_with_message(source_stats_fptr, NULL,
        "Cannot open file %s", source_stats_filename);
    if (source_stats_fptr == NULL)
    {
        free(db_stats);
        free(source_stats);
        return;
    }
    for (det_index = 0; det_index < number_detectors; det_index++)
    {
        status = fscanf(source_stats_fptr, "%lf\n",
            &source_stats[det_index]);
        assert_true_with_message((status >= 0), "Error reading file %s",
            source_stats_filename);
        if (status < 0)
        {
            free(db_stats);
            free(source_stats);
            fclose(source_stats_fptr);
            return;
        }
    }
    fclose(source_stats_fptr);

    /* Make sure the test data has already been loaded, or load if it
       it hasn't */
    size = snprintf(system_command, sizeof(system_command),
        "./check_or_insert_woid.csh %s", work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return;
    }
    status = system(system_command);
    assert_equal_with_message(status, 0, "Error preparing database, ",
        "cannot insert work order ID %s and/or its test data",
        work_order_id);
    if (status != 0)
    {
        free(db_stats);
        free(source_stats);
        return;
    }

    /* Connect to the test database */
    open_db = connect_to_db();
    assert_true(open_db != NULL);
    if (open_db == NULL)
    {
        free(db_stats);
        free(source_stats);
        return;
    }

    /* Get the radiometric work order ID key for the specified work order
       string */
    status = ias_misc_get_char_id(open_db,
        "RADIOMETRIC_WORK_ORDER_COMMON",
        "radiometric_wo_common_id",
        work_order_id, rad_wo_common_id);
    assert_not_equal_with_message(status, ERROR, "Cannot get "
        "radiometric_wo_common_id corresponding to %s",
        work_order_id);
    if (status == ERROR)
    {
        free(db_stats);
        free(source_stats);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        return;
    }

    /* Get the blind band scene means */
    status = get_blind_band_histogram_statistics(open_db, rad_wo_common_id,
        blind_band_number, sca_number, number_detectors,
        RPS_DARKRESPONSE_PROC_LEVEL_BLIND, db_stats);

    /* Check the results */
    assert_equal_with_message(status, SUCCESS, "Cannot get blind histogram "
        "statistics for SCA %d", sca_number);
    if (status != SUCCESS)
    {
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        free(db_stats);
        free(source_stats);
        return;
    }
    for (det_index = 0; det_index < number_detectors; det_index++)
    {
        diff = db_stats[det_index].mean[0] - source_stats[det_index];
        assert_true_with_message((fabs(diff) < 1e-4), "Difference greater "
            "than 1e-4 found for detector %d:  db:  %lf, file: %lf",
            (det_index + 1), db_stats[det_index].mean[0],
            source_stats[det_index]);
    }

    /* Clean up */
    free(db_stats);
    free(source_stats);
    ias_db_close_connection(open_db);
    ias_db_close_database_lib();
}




Ensure test_get_parameters()
{
    char wo_params_filename[PATH_MAX] = "";
    char pmask_filename[] = "TIRS_test_pmask.pmask";

    int status = SUCCESS;
    int size;
    WO_PARAMS wo_params;
    OMF_PARAMS omf_params;


    printf("Testing read of work order and parameter files...\n");

    /* Make sure the test data path has been set */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_not_equal_with_message(test_data_path, NULL, "Test data "
       "directory not found");
    if (test_data_path == NULL)
        return;

    /* Replace the $IAS_UNIT_TEST_DATA_DIR placeholder in the OMF file with
       its actual value */
    size = snprintf(system_command, sizeof(system_command),
        "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' %s_template.omf > %s.omf", 
        test_data_path, work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return;
    }
    status = system(system_command);
    assert_equal_with_message(status, 0, "Cannot edit OMF template");
    if (status != 0)
    {
        return;
    }

    /* Rename the template ODL file as the actual ODL file */
    size = snprintf(wo_params_filename, sizeof(wo_params_filename), 
        "%s.odl", work_order_id);
    if ((size < 0) || (size >= sizeof(wo_params_filename)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
               "it needs %d bytes", size);
        return;
    }
    size = snprintf(system_command, sizeof(system_command),
        "cp %s_template.odl %s", work_order_id, wo_params_filename);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return;
    }
    status = system(system_command);
    assert_equal_with_message(status, 0, "Cannot rename template "
        "parameter file");
    if (status != 0)
    {
        size = snprintf(system_command, sizeof(system_command),
            "rm %s.omf", work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Try to read the files.  This time, it should succeed */
    memset(&wo_params, 0, sizeof(WO_PARAMS));
    memset(&omf_params, 0, sizeof(OMF_PARAMS));
    status= get_parameters(wo_params_filename, &wo_params, &omf_params);
    assert_equal_with_message(status, SUCCESS, "get_parameters failed");
    if (status != SUCCESS)
    {
        size = snprintf(system_command, sizeof(system_command),
            "rm %s.omf %s.odl", work_order_id, work_order_id);
        if ((size < 0) || (size >= sizeof(system_command)))
        {
            assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
            return;
        }
        system(system_command);
        return;
    }

    /* Check selected WO parameters out */
    assert_string_equal(wo_params.rps_proc_level,
        RPS_DARKRESPONSE_PROC_LEVEL_BLIND);
    assert_string_equal(wo_params.work_order_id, work_order_id);
    assert_equal(wo_params.rg_option, CURRENT_RELGAINS);
    assert_equal(wo_params.pmask_option, USE_PMASKS_FROM_FILE);
    assert_equal(wo_params.dark_response_option, NO_USE_CPF_DARK_RESPONSE);

    /* Check the OMF parameters out */
    assert_equal(omf_params.satellite_number, 8);
    size = snprintf(full_cpf_name, sizeof(full_cpf_name), "%s/%s/prep_%s", 
        test_data_path, test_data_dir, cpf_basename);
    if ((size < 0) || (size >= sizeof(full_cpf_name)))
    {
        assert_true_with_message(FALSE, "Filename string too short, "
               "it needs %d bytes", size);
        return;
    }
    assert_string_equal(omf_params.cpf_filename, full_cpf_name);
    assert_string_equal(omf_params.pmask_filename, pmask_filename);

    /* Done */
    size = snprintf(system_command, sizeof(system_command),
        "rm %s.omf %s.odl", work_order_id, work_order_id);
    if ((size < 0) || (size >= sizeof(system_command)))
    {
        assert_true_with_message(FALSE, "Command string too short, "
               "it needs %d bytes", size);
        return;
    }
    system(system_command);

}




int main(int argc, char *argv[])
{
    char *enable_db_tests = NULL;

    IAS_SATELLITE_ID satellite_id = 0;

    int status = SUCCESS;


    /* Check how the driver routine is called.  If there aren't any
       other command line arguments, the whole test suite is run.
       Suppress any IAS_LOG_DEBUG messages */
    ias_log_initialize("tirs_standalone_end2end_test");

    /* Set up the master test suite */
    TestSuite *suite = create_test_suite();

    /* Check to make sure database testing is enabled.  If it is not
       enabled, quit */
    enable_db_tests = getenv("ENABLE_DB_TESTS");
    if (enable_db_tests == NULL)
    {
        printf("Cannot run end-to-end testing--database support not "
            "available\n");
        exit(0);
    }

    /* Get the test data path */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path == NULL)
    {
        printf("Cannot obtain test data directory location\n");
        exit(EXIT_FAIL_NO_TEST_DATA_DIR);
    }

    /* Initialize the Satellite Attributes Library */
    satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(8);
    if (satellite_id != IAS_L8)
    {
        printf("Cannot obtain satellite ID number\n");
        exit(EXIT_FAIL_NO_SAT_ID);
    }

    status = ias_sat_attr_initialize(satellite_id);
    if (status != SUCCESS)
    {
        printf("Failure initializing satellite attributes\n");
        exit(EXIT_FAIL_NO_SAT_LIB);
    }

    add_test(suite, test_get_parameters);
    add_test(suite, test_get_blind_band_scene_means);
    add_test(suite, test_end2end_parameter_help);
    add_test(suite, test_end2end_process);

    /* Run the tests */
    if (argc > 1)
        return run_single_test(suite, argv[1], create_text_reporter());
    else
        return run_test_suite(suite, create_text_reporter());

}
