/*************************************************************************
 NAME:         tirs_dark_response_determination_standalone

 PURPOSE:  Standalone implementation of the TIRS dark response
           determination algorithm, to be run outside of the
           RPS Driver.  It is almost exclusively concerned with gathering
           the required data.  The actual work of determining the
           correct dark response to use is performed by the
           RPS Library function ias_rps_tirs_dark_response_determination()

 RETURNS:  Integer status code of EXIT_SUCCESS or EXIT_FAILURE

**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "ias_cpf.h"
#include "ias_db.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_pixel_mask.h"
#include "ias_radiometry.h"
#include "ias_rps_histogram_statistics_characterization.h"
#include "ias_rps_tirs_dark_response_determination.h"
#include "ias_satellite_attributes.h"

#include "tirs_dark_response_determination_standalone.h"


/*---------------------------------------------------------------------
 NAME:                     cleanup

 PURPOSE:   Frees allocated memory for various objects

 RETURNS:   Nothing

-----------------------------------------------------------------------*/
static void cleanup(
    IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *stats_buffer,
    IAS_PIXEL_MASK **pmask_array,
    int number_masks,
    double *sca_response_difference_buffer
)
{
    IAS_PIXEL_MASK *pm = NULL;             /* "Local" pointer to
                                              pixel mask for the current
                                              band/SCA */

    int mask_index = 0;                    /* Pixel mask loop counter */

    /* Free the blind band statistics data buffer */
    free(stats_buffer);

    /* Free the pixel mask array (if pixel masks were read in from a
       file) */
    if (pmask_array != NULL)
    {
        for (mask_index = 0; mask_index < number_masks; mask_index++)
        {
            pm = (IAS_PIXEL_MASK *)pmask_array[mask_index];
            ias_pm_destroy(pm);
        }
        free(pmask_array);
    }

    /* Free the dark response and SCA dark response difference buffers */
    free(sca_response_difference_buffer);

    /* Done */
}




int main(int argc, char *argv[])
{
    /* Histogram intermediate results data structure */
    IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *blind_stats = NULL;

    /* Database-related stuff */
    struct ias_db_connection *open_db = NULL;
                                           /* IAS DB connection */

    char rad_wo_common_id[IAS_CHAR_ID_SIZE];
                                           /* Database ID corresponding
                                              to specified work order */

    /* CPF-related stuff */
    IAS_CPF *cpf = NULL;                    /* Complete CPF data set */

    const struct IAS_CPF_DETECTOR_STATUS *det_status = NULL;
                                           /* Detector status for TIRS
                                              imaging bands */
    const struct IAS_CPF_DETECTOR_STATUS *tirs_det_status_blind = NULL;
                                           /* Detector status for TIRS
                                              blind band */

    const struct IAS_CPF_ABSOLUTE_GAINS *absolute_gains = NULL;
                                           /* Absolute gains */
    const struct IAS_CPF_ABSOLUTE_GAINS *absolute_gains_blind = NULL;

    const struct IAS_CPF_RELATIVE_GAINS *relative_gains = NULL;
                                           /* Relative gains  */
    const struct IAS_CPF_RELATIVE_GAINS *relative_gains_blind = NULL;

    const struct IAS_CPF_TIRS_DETECTOR_RESPONSE *det_response = NULL;
                                           /* TIRS parameter group that
                                              currently includes
                                              baseline dark response */
    const struct IAS_CPF_TIRS_DETECTOR_RESPONSE *det_response_blind = NULL;

    /* Processing parameters */
    WO_PARAMS wo_params;                   /* Work order parameter
                                              data structure */
    OMF_PARAMS omf_params;                 /* OMF parameter data
                                              structure */
    double *dark_response = NULL;          /* Baseline dark response */
    double *relgain_ptr = NULL;            /* Pointer to correct relative
                                              gains set for the current
                                              band/SCA */
    double *sca_dark_response_difference = NULL;
                                           /* Mean dark response differences
                                              array */
    double abs_gain = 0.0;                 /* Absolute gain for the current
                                              band/SCA */

    char blind_band_name[IAS_BAND_NAME_SIZE];
                                           /* Current "name" of TIRS
                                              blind band */
    int blind_band_number = 0;             /* TIRS blind band number */
    int band_list[IAS_MAX_NBANDS];         /* List of primary TIRS
                                              band numbers */
    int *inop_list = NULL;                 /* Pointer to inoperable detector
                                              list for the current
                                              band/SCA */
    int inop_count = 0;                    /* Number of inoperable detectors
                                              for the current band/SCA */
    int num_image_bands = 0;               /* Number of "normal" TIRS
                                              imaging bands */
    int num_scas = 0;                      /* Number of TIRS SCAs */
    int num_detectors = 0;                 /* Number of detectors per
                                              TIRS SCA */
    int sca_index = 0;                     /* SCA loop counter */
    int index = 0;                         /* General loop counter */
    int status = SUCCESS;

    /* Pixel mask */
    int num_pixel_mask_lines = 2;          /* Number of lines in local pixel
                                              mask */
    int num_file_masks = 0;                /* Number of masks in pixel
                                              mask file */
    int mask_index;                        /* Current pixel mask in
                                              array */
    int pmask_option = NO_USE_PMASKS_FROM_FILE;
                                           /* Local pixel mask option flag
                                              (copied from work order
                                              parameters) */

    IAS_PIXEL_MASK *curr_pmask = NULL;     /* Pointer to the pixel mask
                                              for the current band/SCA */
    IAS_PIXEL_MASK **pmasks_from_file = NULL;
                                           /* If work order option is set,
                                              pixel mask information will
                                              be read into this array
                                              from a file */

    /* Satellite Attributes Library stuff */
    IAS_SATELLITE_ID satellite_id = 0;     /* Satellite ID used to
                                              initialize the Satellite
                                              Attributes Library */


    /* Initialize logging */
    ias_log_initialize(argv[0]);

    /* Make sure we have the proper number of command line arguments.
       Should be just 1, the WO parameter file name */
    if (argc != 2)
    {
        IAS_LOG_ERROR("USAGE:  %s <name of parameter file>", argv[0]);
        return EXIT_FAILURE;
    }

    /* Get the required processing parameters */
    memset(&wo_params, 0, sizeof(WO_PARAMS));
    memset(&omf_params, 0, sizeof(OMF_PARAMS));

    status = get_parameters(argv[1], &wo_params, &omf_params);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Parameter processing failed");
        return EXIT_FAILURE;
    }

    /* Initialize the Satellite Attributes Library */
    IAS_LOG_DEBUG("Initializing Satellite Attributes Library...");

    satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(
        omf_params.satellite_number);
    if (satellite_id == ERROR)
    {
        IAS_LOG_ERROR("Retrieving proper satellite ID");
        return EXIT_FAILURE;
    }

    status = ias_sat_attr_initialize(satellite_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing Satellite Attributes Library");
        return EXIT_FAILURE;
    }

    /* Get the total number of imaging bands.  As currently designed,
       TIRS has a set of "primary" imaging bands and a set of
       redundant or "secondary" imaging bands.  We're interested only
       in the primary set, which is selected by the IAS_NORMAL_BAND
       type value */
    status = ias_sat_attr_get_sensor_band_numbers(IAS_TIRS,
        IAS_NORMAL_BAND, 0, band_list, IAS_MAX_NBANDS, &num_image_bands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Obtaining TIRS imaging band information");
        return EXIT_FAILURE;
    }

    /* Get the current name of the blind band.  As currently designed, TIRS
       has a "primary" blind band and a redundant or "secondary" blind
       band.  Find the set of band attributes by matching the
       IAS_THERMAL1_BAND type to identify the TIRS "primary" bands,
       AND also use the IAS_BLIND_BAND classification to identify the
       blind band itself.

       I did it this way to allow for the possibility of the band names
       and/or numbers changing in the future -- I didn't want to use a
       hard-coded value */
    status = ias_sat_attr_get_band_name_from_type_and_class(
        IAS_THERMAL1_BAND, IAS_BLIND_BAND, blind_band_name);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Finding band name for specified classification and "
            "type");
        return EXIT_FAILURE;
    }

    /* Get the blind band number from its current name */
    blind_band_number = ias_sat_attr_convert_band_name_to_number(
        blind_band_name);
    if (blind_band_number == ERROR)
    {
        IAS_LOG_ERROR("Obtaining TIRS blind band number");
        return EXIT_FAILURE;
    }

    /* Get the number of SCAs and number of detectors per SCA for
       the first imaging band.  As currently designed, these are
       the same for all TIRS bands */
    num_scas = ias_sat_attr_get_sensor_sca_count(IAS_TIRS);
    num_detectors = ias_sat_attr_get_detectors_per_sca(band_list[0]);
    if ((num_scas == ERROR) || (num_detectors == ERROR))
    {
        IAS_LOG_ERROR("Obtaining TIRS SCA and/or detector information");
        return EXIT_FAILURE;
    }

    /* Copy the work order pixel mask option to a local flag.  We can use
       this flag to test for the selected pixel mask option, and change it
       if circumstances require.  The default work order option is to use
       locally created pixel masks (NO_USE_PMASKS_FROM_FILE) */
    pmask_option = wo_params.pmask_option;

    /* If the work order specifies using file-based pixel masks, try to
       read them in from the file specified in the OMF.  If we can't do
       this, reset the local option flag to indicate we're using locally
       created pixel masks populated only with inoperable detector
       information.   */
    if (pmask_option == USE_PMASKS_FROM_FILE)
    {
        pmasks_from_file = ias_pm_read_array_from_file(
            omf_params.pmask_filename, &num_file_masks);
        if (pmasks_from_file == NULL)
        {
            IAS_LOG_WARNING("Cannot read pixel mask file %s, pixel "
                "masks will be created locally and populated only with "
                "inoperable detector information",
                omf_params.pmask_filename);
            pmask_option = NO_USE_PMASKS_FROM_FILE;
        }
    }

    /* Allocate the intermediate histogram statistics data buffer */
    blind_stats = calloc(num_detectors,
        sizeof(IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS));
    if (blind_stats == NULL)
    {
        IAS_LOG_ERROR("Cannot allocate blind band statistics data "
            "structure");
        free(dark_response);
        if (pmask_option == USE_PMASKS_FROM_FILE)
        {
            for (mask_index = 0; mask_index < num_file_masks; mask_index++)
            {
                curr_pmask =
                    (IAS_PIXEL_MASK *)pmasks_from_file[mask_index];
                ias_pm_destroy(curr_pmask);
            }
            free(pmasks_from_file);
        }
        return EXIT_FAILURE;
    }

    /* Allocate the SCA response difference array */
    sca_dark_response_difference = calloc(num_scas, 
        sizeof(*sca_dark_response_difference));
    if (sca_dark_response_difference == NULL)
    {
        IAS_LOG_ERROR("Allocating SCA dark response difference array");
        cleanup(blind_stats, pmasks_from_file, num_file_masks,
            sca_dark_response_difference);
        return EXIT_FAILURE;
    }

    /* Read the CPF contents */
    cpf = ias_cpf_read(omf_params.cpf_filename);
    if (cpf == NULL)
    {
        IAS_LOG_ERROR("Reading CPF file %s", omf_params.cpf_filename);
        cleanup(blind_stats, pmasks_from_file, num_file_masks,
            sca_dark_response_difference);
        return EXIT_FAILURE;
    }

    /* Get the TIRS_DETECTOR_RESPONSE data */
    det_response = ias_cpf_get_tirs_det_response(cpf);
    det_response_blind = ias_cpf_get_tirs_det_response_blind(cpf);

    /* Get the TIRS absolute gains */
    absolute_gains = ias_cpf_get_abs_gains(cpf);
    absolute_gains_blind = ias_cpf_get_tirs_abs_gains_blind(cpf);

    /* Get the TIRS relative gains */
    switch (wo_params.rg_option)
    {
        case PRELAUNCH_RELGAINS:
            relative_gains = ias_cpf_get_prelaunch_relative_gains(cpf);
            relative_gains_blind =
                ias_cpf_get_tirs_pre_rel_gains_blind(cpf);
            break;
        case POSTLAUNCH_RELGAINS:
            relative_gains = ias_cpf_get_postlaunch_relative_gains(cpf);
            relative_gains_blind =
                ias_cpf_get_tirs_post_rel_gains_blind(cpf);
            break;
        case CURRENT_RELGAINS:
            relative_gains = ias_cpf_get_relative_gains(cpf);
            relative_gains_blind = ias_cpf_get_tirs_rel_gains_blind(cpf);
            break;
        default:
            IAS_LOG_ERROR("Invalid relative gain selection option");
            cleanup(blind_stats, pmasks_from_file, num_file_masks,
                sca_dark_response_difference);
            ias_cpf_free(cpf);
            return EXIT_FAILURE;
    }

    if ((det_response == NULL) || (det_response_blind == NULL)
            || (absolute_gains == NULL) || (absolute_gains_blind == NULL)
            || (relative_gains == NULL) || (relative_gains_blind == NULL))
    {
        IAS_LOG_ERROR("Retrieving detector response and gains");
        ias_cpf_free(cpf);
        cleanup(blind_stats, pmasks_from_file, num_file_masks,
            sca_dark_response_difference);
        return EXIT_FAILURE;
    }

    /* Get the detector status information */
    det_status = ias_cpf_get_det_status(cpf);
    tirs_det_status_blind = ias_cpf_get_tirs_det_status_blind(cpf);
    if ((det_status == NULL) || (tirs_det_status_blind == NULL))
    {
        IAS_LOG_ERROR("Retrieving detector status");
        ias_cpf_free(cpf);
        cleanup(blind_stats, pmasks_from_file, num_file_masks,
            sca_dark_response_difference);
        return EXIT_FAILURE;
    }

    /* Establish a connection to the database */
    IAS_LOG_DEBUG("Establishing database connection...");

    open_db = connect_to_db();
    if (open_db == NULL)
    {
        IAS_LOG_ERROR("Establishing database connection");
        ias_cpf_free(cpf);
        cleanup(blind_stats, pmasks_from_file, num_file_masks,
            sca_dark_response_difference);
        return EXIT_FAILURE;
    }

    /* Get the radiometric work order common ID number for the given
       work order ID  */
    status = ias_misc_get_char_id(open_db,
        "RADIOMETRIC_WORK_ORDER_COMMON",
        "radiometric_wo_common_id",
        wo_params.work_order_id, rad_wo_common_id);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Retrieving RADIOMETRIC_WORK_ORDER_COMMON "
            "ID for current work order %s",
            wo_params.work_order_id);
        ias_db_close_connection(open_db);
        ias_db_close_database_lib();
        ias_cpf_free(cpf);
        cleanup(blind_stats, pmasks_from_file, num_file_masks,
            sca_dark_response_difference);
        return EXIT_FAILURE;
    }

    /* Begin SCA processing */
    for (sca_index = 0; sca_index < num_scas; sca_index++)
    {
        /* Convert the loop counter to its corresponding SCA number */
        int sca_number = (sca_index + 1);

/*********************** BLIND BAND PROCESSING ***************************/
        /* Get the blind band baseline dark response */
        dark_response = &(det_response_blind
            ->baseline_dark_response[0][sca_index][0]);

        /* Get the pixel mask for the current SCA from the array if it
           exists, otherwise create a local one */
        if (pmask_option == USE_PMASKS_FROM_FILE)
        {
            mask_index = ias_pm_get_mask_index(pmasks_from_file,
                num_file_masks, blind_band_number, sca_number);
            if (mask_index == ERROR)
            {
                IAS_LOG_ERROR("Finding pixel mask for band %d SCA %d "
                    "in array", blind_band_number, sca_number);
                ias_db_close_connection(open_db);
                ias_db_close_database_lib();
                ias_cpf_free(cpf);
                cleanup(blind_stats, pmasks_from_file, num_file_masks,
                    sca_dark_response_difference);
                return EXIT_FAILURE;
            }
            curr_pmask = (IAS_PIXEL_MASK *)pmasks_from_file[mask_index];
        }
        else
        {
            /* Create a local pixel mask */
            curr_pmask = ias_pm_create(blind_band_number, sca_number,
                num_detectors, num_pixel_mask_lines);
            if (curr_pmask == NULL)
            {
                IAS_LOG_ERROR("Obtaining pixel mask for blind band SCA %d",
                    sca_number);
                ias_db_close_connection(open_db);
                ias_db_close_database_lib();
                ias_cpf_free(cpf);
                cleanup(blind_stats, pmasks_from_file, num_file_masks,
                    sca_dark_response_difference);
                return EXIT_FAILURE;
            }

            /* Get the inoperable detector count and the pointer to the
               start of the inoperable detector list */
            inop_count = tirs_det_status_blind->inoperable_count[0][sca_index];
            inop_list = &(tirs_det_status_blind->inoperable[0][sca_index][0]);

            /* Populate the local pixel mask with the inoperable detector
               information */
            status = ias_rps_inoperable_detector_identification(inop_list,
                inop_count, blind_band_number, sca_number, num_detectors,
                num_pixel_mask_lines, curr_pmask);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Loading inoperable detector information "
                    "into pixel mask for blind band SCA %d", sca_number);
                ias_db_close_connection(open_db);
                ias_db_close_database_lib();
                ias_cpf_free(cpf);
                cleanup(blind_stats, pmasks_from_file, num_file_masks,
                    sca_dark_response_difference);
                ias_pm_destroy(curr_pmask);
                return EXIT_FAILURE;
            }
        }

        /* Get the blind band histogram statistics */
        status = get_blind_band_histogram_statistics(open_db,
            rad_wo_common_id, blind_band_number, sca_number,
            num_detectors, wo_params.rps_proc_level, blind_stats);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Retrieving blind band statistics for SCA %d",
            sca_number);
            ias_db_close_connection(open_db);
            ias_db_close_database_lib();
            ias_cpf_free(cpf);
            cleanup(blind_stats, pmasks_from_file, num_file_masks,
                sca_dark_response_difference);
            if (pmask_option == NO_USE_PMASKS_FROM_FILE)
                ias_pm_destroy(curr_pmask);
            return EXIT_FAILURE;
        }

        /* Get the absolute gain and relative detector gains */
        relgain_ptr = &(relative_gains_blind->per_detector[0][sca_index][0]);
        abs_gain = absolute_gains_blind->gain[0][sca_index];

        /* Call the algorithm for the blind band.  The is_blind_band
           parameter after the relative gain pointer parameter is a flag
           indicating whether we're processing the blind band or not;
           since we are processing the blind band now, set it to TRUE */
        if (wo_params.dark_response_option == NO_USE_CPF_DARK_RESPONSE)
        {
            status = ias_rps_tirs_dark_response_determination(blind_stats,
                dark_response, &sca_dark_response_difference[sca_index],
                abs_gain, relgain_ptr, TRUE, num_detectors, curr_pmask);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Determining adjusted TIRS blind band "
                    "dark response for SCA %d", sca_number);
                ias_db_close_connection(open_db);
                ias_db_close_database_lib();
                ias_cpf_free(cpf);
                cleanup(blind_stats, pmasks_from_file, num_file_masks,
                    sca_dark_response_difference);
                if (pmask_option == NO_USE_PMASKS_FROM_FILE)
                    ias_pm_destroy(curr_pmask);
                return EXIT_FAILURE;
            }
        }

        /* Write the blind band adjusted dark response for each
           SCA to an ODL-formatted group file.  Since we "know"
           we're processing the blind band, set the 'is_blind_band'
           parameter in the call to
           ias_rps_tirs_dark_response_write_to_group_file() to 1 */
        status = ias_rps_tirs_dark_response_write_to_group_file(
            dark_response, 1, blind_band_number, sca_number, num_detectors);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Writing SCA %d blind band dark response "
                "to output file", sca_number);
            ias_cpf_free(cpf);
            ias_db_close_connection(open_db);
            ias_db_close_database_lib();
            cleanup(blind_stats, pmasks_from_file, num_file_masks,
                sca_dark_response_difference);
            return EXIT_FAILURE;
        }

        /* If we have a locally created pixel mask, free its resources */
        if (pmask_option == NO_USE_PMASKS_FROM_FILE)
        {
            ias_pm_destroy(curr_pmask);
        }

/********************** IMAGE BAND PROCESSING ***************************/
        /* Now do the imaging bands */
        for (index = 0; index < num_image_bands; index++)
        {
            int band_number = band_list[index];
            int band_index = 
                ias_sat_attr_convert_band_number_to_index(band_number);
            if ((band_number == ERROR) || (band_index == ERROR))
            {
                IAS_LOG_ERROR("Determining valid TIRS band number and/or "
                    "corresponding band index");
                ias_db_close_connection(open_db);
                ias_db_close_database_lib();
                ias_cpf_free(cpf);
                cleanup(blind_stats, pmasks_from_file, num_file_masks,
                    sca_dark_response_difference);
                return EXIT_FAILURE;
            }

            /* Get a pixel mask for the current band/SCA */
            if (pmask_option == USE_PMASKS_FROM_FILE)
            {
                mask_index = ias_pm_get_mask_index(pmasks_from_file,
                    num_file_masks, band_number, sca_number);
                if (mask_index == ERROR)
                {
                    IAS_LOG_ERROR("Finding pixel mask for band %d SCA %d "
                        "in array", band_number, sca_number);
                    ias_cpf_free(cpf);
                    cleanup(blind_stats, pmasks_from_file, num_file_masks,
                        sca_dark_response_difference);
                    return EXIT_FAILURE;
                }
                curr_pmask = 
                    (IAS_PIXEL_MASK *)pmasks_from_file[mask_index];
            }
            else
            {
                /* Create a local pixel mask and populate it with the
                   inoperable detector information */
                curr_pmask = ias_pm_create(band_number, sca_number,
                    num_detectors, num_pixel_mask_lines);
                if (curr_pmask == NULL)
                {
                    IAS_LOG_ERROR("Creating pixel mask for blind band SCA "
                        "%d", sca_number);
                    ias_cpf_free(cpf);
                    cleanup(blind_stats, pmasks_from_file, num_file_masks,
                        sca_dark_response_difference);
                    return EXIT_FAILURE;
                }

                /* Get the inoperable detector count and the pointer to
                   the start of the list for the current band/SCA */
                inop_count = 
                    det_status->inoperable_count[band_index][sca_index];
                inop_list = 
                    &(det_status->inoperable[band_index][sca_index][0]);

                /* Populate the pixel mask with the current inoperable
                   detector information */
                status = ias_rps_inoperable_detector_identification(
                    inop_list, inop_count, band_number, sca_number,
                    num_detectors, num_pixel_mask_lines, curr_pmask);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Loading inoperable detector information "
                        "into pixel mask for band %d, SCA %d",
                        band_number, sca_number);
                    ias_cpf_free(cpf);
                    cleanup(blind_stats, pmasks_from_file, num_file_masks,
                        sca_dark_response_difference);
                    if (pmask_option == NO_USE_PMASKS_FROM_FILE)
                        ias_pm_destroy(curr_pmask);
                    return EXIT_FAILURE;
                }
            }

            /* Get the baseline dark response */
            dark_response = &(det_response
                ->baseline_dark_response[index][sca_index][0]);

            /* Retrieve detector gains */
            abs_gain = absolute_gains->gain[band_index][sca_index];
            relgain_ptr =
                &(relative_gains->per_detector[band_index][sca_index][0]);

            /* Call the algorithm.  Set the is_blind_band parameter to
               FALSE, since we're not processing a blind band now */
            if (wo_params.dark_response_option == NO_USE_CPF_DARK_RESPONSE)
            {
                status = ias_rps_tirs_dark_response_determination(
                    blind_stats, dark_response,
                    &sca_dark_response_difference[sca_index],
                    abs_gain, relgain_ptr, FALSE, num_detectors,
                    curr_pmask);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Determining band %d, SCA %d adjusted "
                        "TIRS dark response", band_list[index],
                        sca_number);
                    ias_cpf_free(cpf);
                    cleanup(blind_stats, pmasks_from_file, num_file_masks,
                        sca_dark_response_difference);
                    if (pmask_option == NO_USE_PMASKS_FROM_FILE)
                        ias_pm_destroy(curr_pmask);
                    return EXIT_FAILURE;
                }
            }

            /* Write the dark response to a file.  Typically, it will
               be the adjusted dark response (if
               wo_params.use_cpf_baseline_response is set, it will be
               the CPF dark response).  We're not processing the blind
               band, so set the 'is_blind_band' parameter to 0 */
            status = ias_rps_tirs_dark_response_write_to_group_file(
                dark_response, 0, band_number, sca_number, num_detectors);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Writing SCA %d blind band dark response "
                    "to output file", sca_number);
                ias_cpf_free(cpf);
                cleanup(blind_stats, pmasks_from_file, num_file_masks,
                    sca_dark_response_difference);
                if (pmask_option == NO_USE_PMASKS_FROM_FILE)
                    ias_pm_destroy(curr_pmask);
                return EXIT_FAILURE;
            }

            /* If we have a local pixel mask, free up its resources */
            if (pmask_option == NO_USE_PMASKS_FROM_FILE)
            {
                ias_pm_destroy(curr_pmask);
            }
        }
    }

    /* Close DB connection */
    ias_db_close_connection(open_db);
    ias_db_close_database_lib();

    /* Final cleanup */
    ias_cpf_free(cpf);
    cleanup(blind_stats, pmasks_from_file, num_file_masks,
        sca_dark_response_difference);

    /* Done */
    IAS_LOG_INFO("TIRS standalone dark response determination successfully "
        "completed.");
    return EXIT_SUCCESS;
}
