#ifndef TIRS_DARK_RESPONSE_DETERMINATION_STANDALONE_H
#define TIRS_DARK_RESPONSE_DETERMINATION_STANDALONE_H

/* Common C #includes */
#include <stdio.h>
#include <limits.h>

/* IAS #includes */
#include "ias_db.h"
#include "ias_const.h"
#include "ias_radiometry.h"


/* Work order options controlling source of pixel mask information and
   selection of dark response */
#define NO_USE_PMASKS_FROM_FILE       0
#define NO_USE_CPF_DARK_RESPONSE      0
#define USE_PMASKS_FROM_FILE          1
#define USE_CPF_DARK_RESPONSE         1


/* RPS processing level information */
#define RPS_DARKRESPONSE_PROC_LEVEL_BLIND   "0."


/* Relative gain source selection values (in case we can calculate
   radiances through use of CPF gains)*/
typedef enum relgain_option
{
    INVALID_RELGAIN_OPTION = 0,
    PRELAUNCH_RELGAINS,
    POSTLAUNCH_RELGAINS,
    CURRENT_RELGAINS
} RELGAIN_OPTION;


/*           Data structures required for parameter I/O. */
typedef struct wo_params
{
    /* Work order-related information -- RPS processing level */
    char rps_proc_level[IAS_RPS_PROCESSING_LEVEL_LENGTH + 1];

    /* Work order-related information -- parameter file
       identification/location */
    char work_order_id[IAS_WORKORDER_ID_SIZE];

    /* Work order-related information -- processing options */
    int rg_option;                  /* Selection of pre-launch, post-launch,
                                       or best current relative
                                       gains, respectively */
    int pmask_option;               /* Generate pixel masks locally
                                       or read them in from file
                                       (0=create local, 1=read from
                                       file) */
    int dark_response_option;       /* Option to use CPF baseline
                                       dark response (0=calculate adjusted
                                       dark response, 1=use cpf dark
                                       response) */
 } WO_PARAMS;

typedef struct omf_params
{
    int satellite_number;                  /* Satellite number */
    char cpf_filename[PATH_MAX];           /* CPF file name */
    char pmask_filename[PATH_MAX];         /* Name of pixel mask file */
} OMF_PARAMS;



/*************************************************************************
                   FUNCTION PROTOTYPES
**************************************************************************/

                     /* Parameter I/O */
int get_parameters
(
    const char *wo_params_filename,        /* I: Name of work order
                                              parameter file */
    WO_PARAMS *wo_param_data,              /* O: WO parameters */
    OMF_PARAMS *omf_param_data             /* O: OMF parameters */
);



                     /* Database access */
struct ias_db_connection *connect_to_db();


int get_blind_band_histogram_statistics
(
    struct ias_db_connection *db,     /* I: Open database connection */
    char *rad_wo_common_id,           /* I: Database ID corresponding to
                                         work order ID */
    int ref_blind_band_number,        /* I: Current 1-based band number */
    int sca_number,                   /* I: Current 1-based SCA number */
    int detectors_per_sca,            /* I: Number of detectors for the
                                         current band/SCA */
    const char *rps_proc_level,       /* I: RPS processing level */
    IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *blind_stats
                                      /* O: Blind band mean histogram
                                         statistics for the current
                                         SCA */
);


                   /* Miscellaneous */
int get_mask_index_for_band_sca
(
    const IAS_PIXEL_MASK **pmask_array,  /* I: Array to search */
    int band_number,                     /* I: 1-based band number */
    int sca_number,                      /* I: 1-based SCA number */
    int num_scas                         /* I: Number of SCAs */
);


#endif
