#!/usr/local/bin/python3
""" Calculate OLI SNR Trend

    Calculates the OLI signal-to-noise ratio (SNR) trend for the input
    date range
"""
import sys
import argparse
from ias.log import logger as lg
from ias.db import database as db
from ias.io.cpf import cpf as ias_cpf
import re
import numpy as np
from ias.misc import satellite_attributes as sa
from ias.misc import band_specs as bs
import snr_trend
import matplotlib.pyplot as plt
from math import sqrt, ceil, fabs, floor
import snr_char
import os
import datetime as dt
import dateutil.parser
from dateutil import relativedelta as ddelta
from scipy.stats import t as t_dist


def main():
    # Parse in input arguments
    parser = argparse.ArgumentParser(description='Calculate OLI SNR Trends')
    parser.add_argument('--start_date', dest='start_date', required=True,
                        help="Date to start the SNR trending (MM-DD-YYYY)",
                        type=valid_date)
    parser.add_argument('--end_date', dest='end_date', required=True,
                        help="Date to end the SNR trending (MM-DD-YYYY)",
                        type=valid_date)
    parser.add_argument('--output_directory', dest='output_dir', required=True,
                        help="Output directory")
    parser.add_argument('--store_in_db', dest='store_in_db',
                        action='store_true',
                        help="Commit results to the database")
    parser.add_argument('--unit_test', dest='unit_test', action='store_true',
                        help="Unit test development flag")
    parser.add_argument('--satellite_id', dest='satellite_number', default=8,
                        help="Satellite Id example: 8", type=int)
    args = parser.parse_args()

    # Initialize the logger
    logger = lg.Logger("Calculate OLI SNR Trend")

    # Convert to the string dates to datetime
    start_date = args.start_date.strftime("%d-%b-%Y")
    end_date = args.end_date.strftime("%d-%b-%Y")

    # Validate the input date range
    if args.start_date > args.end_date:
        raise Exception("Invalid input date range {0} to {1}".format(
            args.start_date, args.end_date))

    # Create the output directory if it doesn't exist
    args.output_dir = os.path.realpath(args.output_dir)
    if not os.path.exists(args.output_dir):
        os.makedirs(args.output_dir)

    # Load the band specs
    sat_attr = sa.set_default(sa.get_satellite_id_from_number(
                              args.satellite_number))
    band_specs = bs.BandSpecs(args.satellite_number).band_specs

    # Load the CPF
    cpf = ias_cpf.CPF(date=end_date)

    # Get the trending period
    period = get_period_since_launch(cpf, args.end_date)

    # Out of spec factore
    kpr_out_of_spec_factor = 0.80

    # Create the database connection
    mydb = db.DatabaseConnection()
    mydb.set_login_using_satellite_sensor(sat_attr.get_satellite_name(),
                                          sat_attr.get_satellite_sensor_name())
    mydb.connect()

    # Set the correct radiometric work order common table
    rad_com_tbl = "RADIOMETRIC_WORK_ORDER_COMMON"
    lamp_response_table = "LAMP_RESPONSE"
    if args.unit_test:
        rad_com_tbl = "DEV_RAD_WORK_ORDER_COMMON"
        lamp_response_table = "DEV_LAMP_RESPONSE"

    # Get all of the valid solar collects for the time frame requested
    cursor = mydb.execute(
        "Select RADIOMETRIC_WO_COMMON_ID, DATE_ACQUIRED from "
        "{0} where date_acquired between "
        "'{1}' and '{2}' and COLLECTION_TYPE = 'OLI_SOLAR' "
        "and VALID = 'Y'".format(rad_com_tbl, start_date, end_date))
    solar_rows = cursor.fetchall()
    cursor.close()

    # Get all of the valid lamp collects for the time frame requested
    cursor = mydb.execute(
        "Select common.RADIOMETRIC_WO_COMMON_ID, common.DATE_ACQUIRED from "
        "{0} common, {1} lamp where "
        "common.radiometric_wo_common_id = lamp.radiometric_wo_common_id "
        "and common.COLLECTION_TYPE = 'OLI_LAMP' "
        "and common.IMAGE_DATA_TRUNCATION_SETTING = 0 "
        "and lamp.lamp_warmup_time = 150 and common.date_acquired between "
        "'{2}' and '{3}' order by common.date_acquired".format(
            rad_com_tbl, lamp_response_table, start_date, end_date))
    lamp_rows = cursor.fetchall()
    cursor.close()

    # KPR #3 graph values
    ltyp_all_bands = []
    error_all_bands = []

    # Format the report filename with the output directory
    spec_ver_file = "{0}/SNR_Spec_Verification.dat".format(args.output_dir)
    out_spec_file = "{0}/SNR_OutOfSpec.dat".format(args.output_dir)
    spec_sum_file = "{0}/SNR_SpecSummary.dat".format(args.output_dir)
    model_fit_file = "{0}/SNR_LtypModelFit.dat".format(args.output_dir)

    # Get the absolute and relative gains
    abs_gains = cpf.get_oli_absolute_gains(sat_attr)
    rel_gains = cpf.get_oli_relative_gains(sat_attr, combined=True)

    # Loop through all OLI bands, combining the data per detector
    for band_number in sat_attr.band_numbers(sa.SensorId.OLI):
        band_index = sat_attr.convert_band_number_to_index(band_number)
        number_of_scas = sat_attr.get_scas_per_band(band_number)
        number_of_detectors = sat_attr.get_detectors_per_sca(band_number)

        total_det = number_of_scas * number_of_detectors
        signal = [[] for i in range(total_det)]
        noise = [[] for i in range(total_det)]
        band_snr = BandSnr(total_det)
        used_collects = 0
        pi_function = np.zeros((2, total_det))

        # Build a per-detector array of absolute gains from the per-sca
        # values from the CPF
        band_abs_gains = np.repeat(abs_gains[band_index], number_of_detectors)
        band_rel_gains = np.array(rel_gains[band_index])

        # Collect all of the data per band, per detector
        for type in ['SOLAR', 'LAMP']:
            type_used_collects = 0
            if type == 'SOLAR':
                rows = solar_rows
            else:
                rows = lamp_rows
            logger.info("Found {0} {1} collects for band {2}".format(
                len(rows), type, band_number))

            # Combine all of the data into single arrays
            for row in rows:
                data = snr_char.get_snr_char_stats(
                    mydb, row[0], row[1], band_number, type, args.unit_test)

                # Check if this collect will be used
                if (data.size > 0):
                    type_used_collects += 1

                for det_index in range(data.size):
                    signal[det_index].append(data.signal[det_index])

                    # Add the dark signal of '0' for each collect to stabilize
                    # the results better
                    signal[det_index].append(0)
                    noise[det_index].append(data.noise[det_index])

                    # Quantify the shutter collect noise
                    dn = data.dark_noise[det_index]
                    quantization_dn = sqrt((dn * dn) + (1.15 * 1.15))
                    noise[det_index].append(quantization_dn)

            if type_used_collects > 0:
                logger.info("Found {0} {1} collects with SNR characterization "
                            "data for band {2}".format(type_used_collects,
                                                       type, band_number))
                used_collects += type_used_collects

        if used_collects < 2:
            logger.error("Only {0} collections found which is insufficient"
                         " for trending".format(used_collects))
            sys.exit(1)

        logger.info("Used a total of {0} collects for band {1}".format(
            used_collects, band_number))

        # Get the LTypical band specs and LHigh band specs
        band_typ_spec = band_specs[band_index].ltypical
        band_high_spec = band_specs[band_index].lhigh

        # Calculate the DN_Typ and DN_High
        dn_typ = band_abs_gains * band_rel_gains * band_typ_spec
        dn_high = band_abs_gains * band_rel_gains * band_high_spec

        # Trend the data per detector
        logger.info("Trending data for band {0}".format(band_number))
        for detector in range(total_det):
            data_len = used_collects
            sig_matr = np.ones((2, len(signal[detector])))
            dn_matr = np.ones((2, 2))

            # Fit the signal and noise
            sig_matr[1, :] = signal[detector]
            noise_matr = np.power(noise[detector], 2)
            cfit = np.dot(np.linalg.inv(np.inner(sig_matr, sig_matr)),
                          np.inner(sig_matr, noise_matr))
            a = cfit.item(0)
            b = cfit.item(1)

            # Calculate the uncertainty in noise for the two spec points
            # using prediction interval of noise at spec points as uncertainty
            dn_matr[1, :] = (dn_typ[detector], dn_high[detector])
            e = noise_matr - np.dot(cfit, sig_matr)
            s2 = np.dot(np.transpose(e), e) / (len(signal[detector]) - 2)
            exp_hat = np.dot(np.transpose(dn_matr), np.dot(np.linalg.inv(
                np.inner(sig_matr, sig_matr)), dn_matr))
            pi_function[:, detector] = t_dist.ppf(
                1 - 0.025, len(signal[detector]) - 2) * np.sqrt(
                s2.item(0) * (1 + np.diag(exp_hat)))

            # Calculate the uncertainties
            noise_mean = np.mean(noise[detector])
            unc_a = 2 * a / sqrt(2 * (data_len - 1))
            rel_unc_signal = (noise_mean / sqrt(data_len) /
                              np.mean(signal[detector]))
            unc_noise = 2 * noise_mean / sqrt(2 * (data_len - 1))
            t = (sqrt((unc_a * unc_a) + (unc_noise * unc_noise)) /
                 (noise_mean * noise_mean - a))
            rel_unc_b = sqrt((rel_unc_signal * rel_unc_signal) + (t * t))
            band_snr.unc_b[detector] = rel_unc_b * b

            # Calculate the Ltypical and Lhigh SNR
            band_snr.snr_at_ltyp[detector] = dn_typ[detector] / sqrt(fabs(
                dn_typ[detector] * b + a))
            band_snr.snr_at_lhigh[detector] = dn_high[detector] / sqrt(
                fabs(dn_high[detector] * b + a))

            # Save all the results
            band_snr.unc_a[detector] = unc_a
            band_snr.a[detector] = a
            band_snr.b[detector] = b

        # Calculate the SNR uncertainties
        unc_snr_at_ltyp = np.multiply(np.divide(np.sqrt(
            pi_function[0, :]), dn_typ), band_snr.snr_at_ltyp)

        unc_snr_at_lhigh = np.multiply(np.divide(np.sqrt(
            pi_function[1, :]), dn_high), band_snr.snr_at_lhigh)

        # Calculate the median index for ltypical and lhigh
        ltyp_med = (np.abs(band_snr.snr_at_ltyp - np.median(
            band_snr.snr_at_ltyp))).argmin()
        lhigh_med = (np.abs(band_snr.snr_at_lhigh - np.median(
            band_snr.snr_at_lhigh))).argmin()

        # Only insert the rows if requested
        if args.store_in_db:
            # Store the SNR Trend for the current band
            snr_trend.insert_row(
                mydb, band_number, band_snr.snr_at_ltyp[ltyp_med],
                band_snr.snr_at_lhigh[lhigh_med], unc_snr_at_ltyp[ltyp_med],
                unc_snr_at_lhigh[lhigh_med], band_snr.a[ltyp_med],
                band_snr.b[ltyp_med], band_snr.unc_a[ltyp_med],
                band_snr.unc_b[ltyp_med], used_collects, start_date,
                end_date, np.std(band_snr.snr_at_ltyp),
                np.std(band_snr.snr_at_lhigh), args.unit_test)

        # Calculate the Out of Spec limits
        band_oos_typ_spec = band_specs[band_index].snr_at_ltyp
        band_oos_high_spec = band_specs[band_index].snr_at_lhigh

        # Save the band data for later to be used in graphing the KPR #3
        ltyp_all_bands.append(band_snr.snr_at_ltyp[ltyp_med])
        error_all_bands.append(np.std(band_snr.snr_at_ltyp))

        # Create the file headers if this is the first band
        if band_index == 0:
            write_report_line(
                model_fit_file, 'w', "{:>2s} {:>8s} {:>5s} "
                "{:>12s} {:>12s} {:>12s} {:>12s} {:>12s} {:>12s}\n".format(
                    "B#", "Band", "Pd", "PdRange", "N_Collects", "OLILtyp",
                    "FPMGain", "A", "B"))

            write_report_line(
                spec_sum_file, 'w', "{:>2s} {:>8s} {:>20s} "
                "{:>15s} {:>20s} {:>15s} {:>20s} {:>20s} {:>20s}\n".format(
                    "B#", "Band", "Median SNR@Ltyp", "Requirement",
                    "Median SNR@Lhigh", "Requirement", "MedianLtypOOS",
                    "NDetOOS@Ltyp", "NDetsOOS@Lhigh"))

            write_report_line(
                out_spec_file, 'w', "{:>2s} {:>10s} {:>10s} "
                "{:>10s} {:>10s} {:>10s} {:>10s} {:>15s} {:>15s} "
                "{:>10s}\n".format(
                    "B#", "Band", "Level", "Period", "SCA", "Detector",
                    "SNR", "Uncertainty", "Requirement", "Fails"))

            write_report_line(
                spec_ver_file, 'w', "{:>2s} {:>10s} {:>10s} "
                "{:>12s} {:>10s} {:>10s} {:>10s} {:>10s} {:>10s} {:>11s} | "
                "{:>10s} {:>10s} {:>10s} {:>10s} {:>12s}\n".format(
                    "B#", "Band", "Pd", "PdRange", "N_Collects", "LtypSpec",
                    "SNRmedian", "stdev(SNR)", "unc(SNR)", "%ofLtypSpec",
                    "LhighSpec", "SNRmedian", "stdev(SNR)", "unc(SNR)",
                    "%ofLhighSpec"))

        # Reset the detector count for this band
        oos_typ = 0
        oos_high = 0
        oos_typ_list = list()

        # Output the data for all out of spec detectors
        with open(out_spec_file, 'a') as f:
            typ_out_of_spec = kpr_out_of_spec_factor * band_oos_typ_spec
            high_out_of_spec = kpr_out_of_spec_factor * band_oos_high_spec
            oos_format = ("{:>2d} {:>10s} {:>10s} {:>10d} {:>10d} {:>10d} "
                          "{:>10.3f} {:>15.3f} {:>15.3f} {:>10s}\n")
            for i in range(total_det):
                sca = floor(i / number_of_detectors) + 1
                det = i % number_of_detectors + 1
                if (band_snr.snr_at_ltyp[i] + (2 * unc_snr_at_ltyp[i]) <
                        typ_out_of_spec):
                    oos_typ += 1
                    oos_typ_list.append(band_snr.snr_at_ltyp[i])
                    f.write(oos_format.format(
                        band_number, band_specs[band_index].band_color,
                        "Ltypical", period, sca, det, band_snr.snr_at_ltyp[i],
                        unc_snr_at_ltyp[i], band_oos_typ_spec, "Y"))

                if (band_snr.snr_at_lhigh[i] + (2 * unc_snr_at_lhigh[i]) <
                        high_out_of_spec):
                    oos_high += 1
                    f.write(oos_format.format(
                        band_number, band_specs[band_index].band_color,
                        "Lhigh", period, sca, det, band_snr.snr_at_lhigh[i],
                        unc_snr_at_lhigh[i], band_oos_high_spec, "Y"))

        # Set the OOS Ltypical median
        oos_med = np.median(oos_typ_list) if oos_typ_list else 0.0

        # Output the data for the Fit model and Spec summary
        write_report_line(
            model_fit_file, 'a', "{:>2d} {:>8s} {:>5d} {:>12s} "
            "{:>12d} {:>12.3f} {:>12.6f} {:>12.6f} {:>12.6f}\n".format(
                band_number, band_specs[band_index].band_color, period,
                end_date, used_collects, band_typ_spec,
                band_abs_gains[ltyp_med], band_snr.a[ltyp_med],
                band_snr.b[ltyp_med]))

        write_report_line(
            spec_sum_file, 'a', "{:>2d} {:>8s} {:>20.6f} "
            "{:>15f} {:>20.6f} {:>15f} {:>20.6f} {:>20d} {:>20d}\n".format(
                band_number, band_specs[band_index].band_color,
                band_snr.snr_at_ltyp[ltyp_med], band_oos_typ_spec,
                band_snr.snr_at_lhigh[lhigh_med], band_oos_high_spec,
                oos_med, oos_typ, oos_high))

        write_report_line(
            spec_ver_file, 'a', "{:>2d} {:>10s} {:>10d} {:>12s} "
            "{:>10d} {:>10.2f} {:>10.3f} {:>10.3f} {:>10.3f} {:>11.2f} | "
            "{:>10.3f} {:>10.3f} {:>10.3f} {:>10.3f} {:>12.3f}\n".format(
                band_number, band_specs[band_index].band_color, period,
                end_date, used_collects, band_oos_typ_spec,
                band_snr.snr_at_ltyp[ltyp_med], np.std(band_snr.snr_at_ltyp),
                unc_snr_at_ltyp[ltyp_med], band_snr.snr_at_ltyp[ltyp_med] /
                band_oos_typ_spec * 100, band_oos_high_spec,
                band_snr.snr_at_lhigh[lhigh_med],
                np.std(band_snr.snr_at_lhigh), unc_snr_at_lhigh[lhigh_med],
                band_snr.snr_at_lhigh[lhigh_med]/band_oos_high_spec*100))

    # Disconnect from the database
    if args.store_in_db:
        mydb.close(db.DISCONNECT_COMMIT)
    else:
        mydb.close(db.DISCONNECT_ROLLBACK)

    # Plot the KPR graphs
    plot_graphs(ltyp_all_bands, error_all_bands, band_specs, args.output_dir)


class BandSnr():
    """ Band SNR object that stores SNR data per detector"""

    def __init__(self, num_detectors):
        members = ("unc_b", "snr_at_ltyp", "snr_at_lhigh", "a", "b",
                   "unc_a")
        for member in members:
            setattr(self, member, np.empty(num_detectors, dtype=float))


def write_report_line(filename, file_access, line):
    """ Writes a single line to the report file """
    with open(filename, file_access) as f:
        f.write(line)


def plot_graphs(ltyp_all_bands, error_all_bands, band_specs, output_dir):

    # Start KPR 3 graph
    index = np.arange(9)
    bar_width = .25
    opacity = 0.4
    error_config = {'ecolor': '0.3'}

    # Set up the specification arrays
    req_bands = [x.snr_at_ltyp for x in band_specs[0:9]]
    req_2_bands = [x.snr_at_ltyp * 1.5 for x in band_specs[0:9]]

    # Plot each bar onto the graph
    ax = plt.subplot(111)

    # Plot the first bar which is the requirement for each of the bands
    bar1 = ax.bar(index, req_bands, bar_width, alpha=opacity, color='b',
                  label="OLI SNR Requirement [median at Ltyp]")

    # Plot the second bar which is the actual value for Signal-to-Noise Ration
    # at Ltypical, along with the standard error plotting
    bar2 = ax.bar(index + bar_width, ltyp_all_bands, bar_width, alpha=opacity,
                  color='g', yerr=error_all_bands, error_kw=error_config,
                  label="OLI On Orbit [median at Ltyp]")

    # Plot the third bar, which is the requirement for each band * 1.5
    bar3 = ax.bar(index + (bar_width * 2), req_2_bands, bar_width,
                  alpha=opacity, color='m',
                  label="OLI SNR Requirement * 1.5 [median at Ltyp]")

    # Shrink current axis's height by 10% on the bottom to allow for putting
    # the legend below the graph
    box = ax.get_position()
    ax.set_position([box.x0, box.y0 + box.height * 0.1,
                     box.width, box.height * 0.9])

    # Put a legend below current axis
    ax.legend(loc='upper center', bbox_to_anchor=(0.5, -0.1),
              fancybox=True, prop={'size': 6}, shadow=True, ncol=3)

    # Set the plot labels
    plt.xlabel("Band")
    plt.ylabel("SNR")
    plt.title("OLI Signal-to-Noise Performance")
    band_names = [x.band_color for x in band_specs]
    plt.xticks(index + ((3 / 2) * bar_width), band_names)
    plt.savefig("{0}/kpr3.png".format(output_dir))


def get_period_since_launch(cpf, datetime):
    """ Returns the number of months since the launch of L8 """
    launch_date = dateutil.parser.parse(
        cpf.get_orbit_parameters().launch_date)
    rd = ddelta.relativedelta(datetime, launch_date)
    months = (rd.years * 12) + rd.months
    return months


def valid_date(str_date):
    """ Valid date matcher """
    if re.match("[0-9]{2}-[0-9]{2}-[0-9]{4}", str_date):
        return dt.datetime.strptime(str_date, "%m-%d-%Y")
    raise Exception("Invalid date passed ({0}) use MM-DD-YYY "
                    "(02-12-15)".format(s))

# Runs main if called from command line
if __name__ == "__main__":
    main()
