#!/usr/local/bin/python3
import unittest
from subprocess import call
from ias.db import database as db
import os
import filecmp
import imp
import snr_trend
import snr_char


class TestCalculateTrend(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """ Setup method ran before all unit tests"""

        # Get the database credentials
        db_cred = os.getenv("IAS_DB_L8_OLITIRS")
        if not db_cred:
            raise Exception("IAS_DB_L8_OLITIRS environment variable not set")

        # Drop and Create the need tables
        status = call(["sqlplus", db_cred, "@create_tables.sql"])
        if status != 0:
            raise Exception("Creating db tables")

        # Load the data
        status = call(["./load_data.csh"])
        if status != 0:
            raise Exception("Loading data into db tables")

    def test_calculate_trend_no_db(self):
        """ Run thes trending application without storing the statistics
            to the SNR_TREND database table"""

        # Run the application
        status = call(
            ["../calculate_oli_snr_trend.py", "--start_date",
             "03-04-2015", "--end_date", "03-05-2015", "--output_directory",
             "test1", "--unit_test"])
        if status != 0:
            raise Exception("Running calculate_oli_snr_trend.py")

        # Verfiy the output files
        verifylist = ("kpr3.png", "SNR_LtypModelFit.dat", "SNR_OutOfSpec.dat",
                      "SNR_SpecSummary.dat", "SNR_Spec_Verification.dat")
        for file in verifylist:
            files_equal = filecmp.cmp("test1/" + file, "verify_no_db/" + file)
            self.assertEqual(files_equal, True, "Unable to verify {0}".format(
                file))

        # Verify there are no records in the Trend database table
        with db.DatabaseConnection('L8_OLITIRS') as conn:
            cursor = conn.execute("Select count(*) from DEV_SNR_TREND")
            self.assertEqual(0, cursor.fetchall()[0][0])

    def test_calculate_trend_with_db(self):
        """ Run thes trending application without storing the statistics
            to the SNR_TREND database table"""

        # Run the application
        status = call(
            ["../calculate_oli_snr_trend.py", "--start_date",
             "02-24-2015", "--end_date", "03-23-2015", "--output_directory",
             "test2", "--unit_test", "--store_in_db"])
        if status != 0:
            raise Exception("Loading data into db tables")

        # Verfiy the output files
        verifylist = ("kpr3.png", "SNR_LtypModelFit.dat", "SNR_OutOfSpec.dat",
                      "SNR_SpecSummary.dat", "SNR_Spec_Verification.dat")
        for file in verifylist:
            files_equal = filecmp.cmp("test2/" + file, "verify/" + file)
            self.assertEqual(files_equal, True, "Unable to verify {0}".format(
                file))

        # Verify the number of records in the table
        with db.DatabaseConnection('L8_OLITIRS') as conn:
            cursor = conn.execute("Select count(*) from DEV_SNR_TREND")
            self.assertEqual(cursor.fetchall()[0][0], 9)

        # Get the L8 database credentials
        db_cred = os.getenv("IAS_DB_L8_OLITIRS")
        if not db_cred:
            raise Exception("IAS_DB_L8_OLITIRS environment variable not set")

        # Dump the trend records to file
        status = call(["sqlplus", db_cred, "@trend_dump.sql"])
        if status != 0:
            raise Exception("Creating db tables")

        # Check the dump file
        self.assertEqual(filecmp.cmp("trend_dump.db",
                                     "verify/trend_dump.verify"), True)

    def test_snr_trend(self):
        """ Test the SNR trend module """
        # Create the database connection
        mydb = db.DatabaseConnection("L8_OLITIRS")
        mydb.connect()

        datasets = ((1, 250.0, 1200.0, 2.5, 120.0, 1.1, 2.2, .1, .2, 100,
                     '03-MAY-2014', '04-MAY-2014', .11, .22),
                    (2, 251.0, 1201.0, 2.3, 1120.0, 1.15, 21.2, .21, .32,
                     1100, '04-MAY-2014', '05-MAY-2014', .81, .2),
                    (3, 252.0, 1203.0, 2.1, 1210.0, 1.16, 2.21, .12, .2,
                     1040, '05-MAY-2014', '06-MAY-2014', .81, .21)
                    )

        # Clean up the trending table
        del_q = "Delete from DEV_SNR_TREND"
        cur = mydb.execute(del_q)
        cur.close()

        # load each data set individually verify all data in the database
        # everytime a insert occurs
        for set in datasets:
            # Insert the trend data
            snr_trend.insert_row(mydb, *set, unit_test=True)

            query = """Select
                BAND_ID, SNR_AT_LTYPICAL, SNR_AT_LHIGH,
                UNCERTAINTY_LTYPICAL, UNCERTAINTY_LHIGH, A, B,
                UNCERTAINTY_A, UNCERTAINTY_B, NUMBER_OF_COLLECTS,
                START_DATE, END_DATE, STDEV_LTYPICAL,
                STDEV_LHIGH, VALID from DEV_SNR_TREND order by BAND_ID"""
            cur = mydb.execute(query)
            rows = cur.fetchall()
            cur.close()

            # Verify that the correct data was inserted
            for row in range(len(rows)):
                for index in range(len(set)):
                    if index == 10 or index == 11:
                        continue
                    self.assertAlmostEqual(
                        datasets[row][index], rows[row][index], places=9,
                        msg="Error comparing row {0} at index {1}".format(
                            row, index))

        mydb.close(db.DISCONNECT_ROLLBACK)

    @classmethod
    def tearDownClass(cls):
        """ Tear down method ran after all unit tests"""
        db_cred = os.getenv("IAS_DB_L8_OLITIRS")
        if not db_cred:
            raise Exception("IAS_DB_L8_OLITIRS environment variable not set")

        status = call(["sqlplus", db_cred, "@cleanup_tables.sql"])
        if status != 0:
            raise Exception("Cleaning up db tables")

if __name__ == '__main__':
    unittest.main()
