#!/usr/local/bin/python3
""" Characterize the OLI Signal to Noise ratio (SNR) of a single Lamp or Solar
    collect. If requested the data is written to the SNR Characterization
    table.

    Additional Files produced:
       {Band}_SNR_SpecWarnings.dat
       SNR_SpecSummary.dat
"""
import argparse
from sys import exit, argv
import os
from ias.log import logger as lg
from ias.db import database as db
from ias.io.cpf import cpf as ias_cpf
from ias.io.parameter_file import parameter_file as pf
from ias.misc.const import ERROR
import numpy as np
from ias.misc import satellite_attributes as sa
from ias.misc import band_specs as bs
from band_snr import BandSnr
import matplotlib.pyplot as plt
import matplotlib.animation as animation
import histogram_stats as hs
from math import floor


@lg.log_exceptions
def main():
    """ Characterize OLI SNR main routine """
    # Initialize the logger
    logger = lg.Logger("Characterize OLI SNR")

    # Parse in input arguments
    parms = read_parameters()

    # Create the output directory if it doesn't exist
    parms.output_dir = os.path.realpath(parms.output_dir)
    if not os.path.exists(parms.output_dir):
        os.makedirs(parms.output_dir)

    # Set up the satellite attributes
    sat_attr = sa.set_default(sa.get_satellite_id_from_number(
                              parms.satellite_number))

    # Create the database connection
    mydb = db.DatabaseConnection()
    mydb.set_login_using_satellite_sensor(sat_attr.get_satellite_name(),
                                          sat_attr.get_satellite_sensor_name())
    mydb.connect()

    # Determine the correct table to query
    rad_com_tbl = "RADIOMETRIC_WORK_ORDER_COMMON"
    if parms.unit_test:
        rad_com_tbl = "DEV_RAD_WORK_ORDER_COMMON"

    # Get the CPF associated with this id
    cursor = mydb.execute(
        """Select RADIOMETRIC_WO_COMMON_ID, to_char(date_acquired,
        'DD-MON-YY '), COLLECTION_TYPE from {0} where
        WORK_ORDER_ID = '{1}'""".format(rad_com_tbl, parms.wo_id))
    wo_com_row = cursor.fetchall()

    # Check for a valid radiometric work order id
    if (not wo_com_row):
        raise Exception("Radiometric WO ID {0} does not exist".format(
                        parms.wo_id))

    rad_id = str(wo_com_row[0][0])
    collection_type = str(wo_com_row[0][2])

    # Percentage below the specification to be considered out of specification
    kpr_spec = 0.8

    # Load the cpf
    cpf = ias_cpf.CPF(filename=parms.cpf_name)
    abs_gains = cpf.get_oli_absolute_gains(sat_attr)
    rel_gains = cpf.get_oli_relative_gains(sat_attr, combined=True)

    # Load the band specs
    band_specs = bs.BandSpecs(parms.satellite_number).band_specs

    # Delete any existing statistics after starting a transaction
    if parms.store_in_db:
        mydb.begin_transaction()
        hs.delete_existing_snr_characterization(mydb, rad_id, parms.unit_test)

    # Loop over all OLI Bands
    for band_number in sat_attr.band_numbers(sa.SensorId.OLI):
        band_index = sat_attr.convert_band_number_to_index(band_number)
        number_of_scas = sat_attr.get_scas_per_band(band_number)
        number_of_detectors = sat_attr.get_detectors_per_sca(band_number)

        # Only process requested OLI Bands
        if band_number not in parms.band_list:
            continue

        logger.info("Processing band number {0}".format(band_number))

        # Get the total detectors for the entire band
        total_det = number_of_scas * number_of_detectors

        # Get the absolute gains for the current band (Repeated for every
        # detector in the band) and relative gains per SCA (combined per band)
        gains = (np.repeat(abs_gains[band_index], number_of_detectors) *
                     np.array(rel_gains[band_index]))

        # Get the histogram stats data as a histogram stats object
        data = hs.get_histogram_stats(mydb, rad_id, band_number,
                                      collection_type, parms.unit_test)

        # If no data was found, it is likely due to no dark collect being
        # available close enough to the collection.  Issue a warning and exit
        # successfully.
        if (data.size <= 0):
            logger.warn("No histogram statistics found for rad id {0}".format(
                    rad_id))
            logger.warn("Exiting without performing the characterization")
            exit(0)

        # Create the SNR object
        snr = BandSnr(band_number, data, rad_id)

        # All the histogram stats are stored as numpy arrays so instead
        # of looping over all detectors use numpy operations on arrays
        squared_stdev = np.power(data.stdev, 2)
        snr.a = np.power(data.dark_stdev, 2)
        snr.unc_a = 2 * snr.a / np.sqrt(2 * (data.n_dark_frames - 1))
        snr.rel_unc_a = snr.unc_a / snr.a
        snr.b = (squared_stdev - snr.a) / data.mean

        # Create a copy of snr.b that changes negative values to zero.  This
        # is to prevent trying to do sqrt operations on negative numbers.  The
        # values with negative snr.b are later reset anyway.
        neg_indices = np.where(snr.b < 0.0)
        filtered_snr_b = np.copy(snr.b)
        filtered_snr_b[neg_indices] = 0.0

        rel_unc_signal = data.stdev / np.sqrt(data.n_frames) / data.mean
        unc_noise = 2 * squared_stdev / np.sqrt(2 * (data.n_frames - 1))
        t = (np.sqrt(np.power(snr.unc_a, 2) + np.power(unc_noise, 2)) /
             (squared_stdev - snr.a))
        snr.rel_unc_b = np.sqrt(np.power(rel_unc_signal, 2) + np.power(t, 2))
        snr.unc_b = snr.rel_unc_b * snr.b
        dn_typ = gains * band_specs[band_index].ltypical
        dn_high = gains * band_specs[band_index].lhigh
        snr.snr_at_ltyp = dn_typ / np.sqrt(snr.a + filtered_snr_b * dn_typ)
        snr.snr_at_lhigh = dn_high / np.sqrt(snr.a + filtered_snr_b * dn_high)

        # Calculate the SNR uncertainties
        correlation_r = 1
        dsnr_da = -1 * snr.snr_at_lhigh / (2 * (snr.a + snr.b * dn_high))
        dsnr_db = -1 * dn_high * snr.snr_at_lhigh / (
            2 * (snr.a + snr.b * dn_high))
        unc_snr_at_lhigh = np.sqrt(
            np.power(snr.unc_a * dsnr_da, 2) +
            np.power(snr.unc_b * dsnr_db, 2) + 2 * correlation_r * dsnr_da *
            dsnr_db * snr.unc_a * snr.unc_b)

        dsnr_da = -1 * snr.snr_at_ltyp / (2 * (snr.a + snr.b * dn_typ))
        dsnr_db = -1 * dn_typ * snr.snr_at_ltyp / (2 * (
            snr.a + snr.b * dn_typ))
        unc_snr_at_ltyp = np.sqrt(
            np.power(snr.unc_a * dsnr_da, 2) +
            np.power(snr.unc_b * dsnr_db, 2) + 2 * correlation_r * dsnr_da *
            dsnr_db * snr.unc_a * snr.unc_b)
        snr.unc_snr_at_ltyp = unc_snr_at_ltyp
        snr.unc_snr_at_lhigh = unc_snr_at_lhigh

        # Calculate the median values
        median_at_lhigh = np.median(snr.snr_at_lhigh)
        median_at_ltyp = np.median(snr.snr_at_ltyp)

        oos_typ = 0
        oos_high = 0
        hdr_created = False
        oos_list = list()

        # Loop over all detectors
        for det in range(total_det):
            # Determine the SCA and Detector numbers
            sca_num = floor(det / number_of_detectors) + 1
            det_num = det % number_of_detectors + 1
            oos_flag = 0
            write_flag = 0

            # Check SNR at ltypical for below and out of spec detectors
            adjusted_snr = snr.snr_at_ltyp[det] + (2 * unc_snr_at_ltyp[det])
            if adjusted_snr < band_specs[band_index].snr_at_ltyp:
                write_flag = 1
            if adjusted_snr < band_specs[band_index].snr_at_ltyp * kpr_spec:
                oos_list.append(snr.snr_at_ltyp[det])
                oos_flag = 1
                oos_typ += 1

            # Check SNR at Lhigh for below and out of spec detectors
            adjusted_snr = snr.snr_at_lhigh[det] + (2 * unc_snr_at_lhigh[det])
            if adjusted_snr < band_specs[band_index].snr_at_lhigh:
                write_flag = 1
            if adjusted_snr < band_specs[band_index].snr_at_lhigh * kpr_spec:
                oos_flag = 1
                oos_high += 1

            if write_flag or oos_flag:
                # Create the file header if not yet created
                if not hdr_created:
                    hdr_created = True
                    write_spec_warning_header(
                        parms.output_dir, band_specs[band_index].band_color)

                # Output the below-spec/out-of-spec detector record
                write_spec_warning_record(
                    parms.output_dir, band_specs[band_index].band_color,
                    sca_num, det_num, snr.snr_at_ltyp[det],
                    unc_snr_at_ltyp[det], band_specs[band_index].snr_at_ltyp,
                    snr.snr_at_lhigh[det], unc_snr_at_lhigh[det],
                    band_specs[band_index].snr_at_lhigh, oos_flag)

            # Set this detector to invalid by setting data to -999
            if oos_flag:
                snr.a[det] = -999
                snr.b[det] = -999
                snr.snr_at_ltyp[det] = -999
                snr.snr_at_lhigh[det] = -999
                unc_snr_at_ltyp[det] = -999
                unc_snr_at_lhigh[det] = -999
                snr.unc_a[det] = -999
                snr.unc_b[det] = -999

            # If the b coefficient is negative then invalidate the model and
            # uncertainty
            if snr.b[det] <= 0.0:
                snr.snr_at_ltyp[det] = -999
                snr.snr_at_lhigh[det] = -999
                unc_snr_at_ltyp[det] = -999
                unc_snr_at_lhigh[det] = -999

        # Set the median to zero if no detectors were out of spec
        oos_med = np.median(oos_list) if oos_list else 0

        # Write the band specification report
        if band_index == 0:
            with open("{0}/SNR_SpecSummary.dat".format(
                      parms.output_dir), 'w') as f:
                f.write("{:>2s} {:>8s} {:>20s} {:>15s} {:>20s} {:>15s} "
                        "{:>20s} {:>20s} {:>20s}\n".format(
                            "B#", "Band", "Median SNR@Ltyp", "Requirement",
                            "Median SNR@Lhigh", "Requirement", "MedianLtypOOS",
                            "NDetOOS@Ltyp", "NDetsOOS@Lhigh"))

        with open("{0}/SNR_SpecSummary.dat".format(
                  parms.output_dir), 'a') as f:
            f.write("{:>2d} {:>8s} {:>20.6f} {:>15f} {:>20.6f} {:>15f} "
                    "{:>20.6f} {:>20d} {:>20d}\n".format(
                        band_number, band_specs[band_index].band_color,
                        median_at_ltyp, band_specs[band_index].snr_at_ltyp,
                        median_at_lhigh, band_specs[band_index].snr_at_lhigh,
                        oos_med, oos_typ, oos_high))

        # Store in the characterization table if requested
        if parms.store_in_db:
            snr.insert_snr_rows(mydb, parms.unit_test)

    # Disconnect from the database
    if parms.store_in_db:
        mydb.close(db.DISCONNECT_COMMIT)
    else:
        mydb.close(db.DISCONNECT_ROLLBACK)


def write_spec_warning_header(dir, band):
    """ Writes the specification warning header for a single band """
    header = ("{:>5s} {:>10s} {:>10s} {:>15s} {:>15s} {:>10s} {:>15s} "
              "{:>15s} {:>15s}\n")
    with open("{0}/{1}_SNR_SpecWarnings.dat".format(dir, band), 'w') as f:
        f.write(header.format(
            "SCA", "Detector", "SNR@Ltyp", "Uncertainty", "Requirement",
            "SNR@Lhigh", "Uncertainty", "Requirement", "Out-of-Spec"))


def write_spec_warning_record(dir, band, sca, detector, snr_ltyp, unc_ltyp,
                              req_ltyp, snr_lhigh, unc_lhigh, req_lhigh,
                              oos_flag):
    """ Writes out a single row of data for the specification warning
        file
    """
    # Most warnings will be below spec and not out of spec
    if not oos_flag:
        is_out_of_spec = "N"
    else:
        is_out_of_spec = "Y"

    # Output this detector to the warnings file
    with open("{0}/{1}_SNR_SpecWarnings.dat".format(dir, band), 'a') as f:
        row = ("{:>5d} {:>10d} {:>10.4f} {:>15.4f} {:>15.4f} {:>10.4f} "
               "{:>15.4f} {:>15.4f} {:>15s}\n")
        f.write(row.format(sca, detector, snr_ltyp, unc_ltyp, req_ltyp,
                snr_lhigh, unc_lhigh, req_lhigh, is_out_of_spec))


def read_parameters():
    """ Reads the Parameter and OMF files using the requested parameters """
    parser = pf.ParameterParser()
    parser.add_standard_parameter('WORK_ORDER_ID', 'wo_id')
    parser.add_standard_parameter('OUTPUT_DIRECTORY', 'output_dir',
                                  default="SNR")
    parser.add_standard_parameter('STORE_CHARACTERIZATION_IN_DB',
                                  'store_in_db')
    parser.add_standard_parameter('BAND_LIST', 'band_list', is_list=True,
                                  required=False,
                                  default=(1, 2, 3, 4, 5, 6, 7, 8, 9))
    parser.add_parameter('UNIT_TEST', 'unit_test', required=False,
                         default=False, help='Unit testing flag')

    omf = pf.ParameterParser()
    omf.add_standard_parameter('CAL_PARM_FILENAME', 'cpf_name')
    omf.add_standard_parameter('SATELLITE', 'satellite_number', required=True)

    # Parse the parameter file if it exist
    if len(argv) >= 2:
        # Provide help if requested
        if parser.provide_help(argv[1], pf.PARAMETERS):
            omf.provide_help(argv[1], pf.OMF)
            exit(0)

        parms = parser.parse_parameters(argv[1])
        parms = omf.parse_parameters("{0}.omf".format(parms.wo_id), 
                                     target = parms)
    else:
        lg.Logger.get_default_logger().error("No parameter file provided")
        exit(ERROR)

    return parms

# Runs main if called from command line
if __name__ == "__main__":
    main()
