#!/usr/local/bin/python3
""" Histogram Stats

    Histogram Statisitcs helper Module
"""
import numpy as np
from ias.log import logger as lg


class HistogramStatsList():
    """ Histogram Statistics Object """
    def __init__(self, size):
        self.band_id = np.empty(size, dtype=int)
        self.date_acquired = np.empty(size, dtype=str)
        self.image_start_time = np.empty(size, dtype=int)
        self.collection_type = np.empty(size, dtype=str)
        # The collection source is the diffuser type for solar collections
        # and the bulb pair for lamp collections
        self.collection_source = np.empty(size, dtype=int)
        self.sca = np.empty(size, dtype=int)
        self.det = np.empty(size, dtype=int)
        self.n_frames = np.empty(size, dtype=int)
        self.n_dark_frames = np.empty(size, dtype=int)
        self.mean = np.empty(size, dtype=float)
        self.stdev = np.empty(size, dtype=float)
        self.dark_mean = np.empty(size, dtype=float)
        self.dark_stdev = np.empty(size, dtype=float)
        self.truncation = np.empty(size, dtype=int)
        self.size = size

    def add_data(self, rows):
        """ Add a row of queried data to the lists """

        # Check the size of the rows
        if len(rows) > self.size:
            raise ValueError("Size of input rows is too large")

        for row in range(len(rows)):
            self.band_id[row] = rows[row][0]
            self.date_acquired[row] = rows[row][1]
            self.image_start_time[row] = rows[row][2]
            self.collection_type[row] = rows[row][3]
            self.collection_source[row] = rows[row][4]
            self.sca[row] = rows[row][5]
            self.det[row] = rows[row][6]
            self.n_frames[row] = rows[row][7]
            self.n_dark_frames[row] = rows[row][8]
            self.mean[row] = rows[row][9]
            self.stdev[row] = rows[row][10]
            self.dark_mean[row] = rows[row][11]
            self.dark_stdev[row] = rows[row][12]
            self.truncation[row] = rows[row][13]
        return self


def get_histogram_stats(db, rad_id, band_number, collection_type,
                        unit_test=None):
    """ Get the histogram statistics for the input radiometric id """

    # Check for a shutter collect before retrieving the statistics
    dark_rad_id = get_closest_dark_shutter_id(db, rad_id, unit_test)
    if (not dark_rad_id):
        logger = lg.Logger.get_default_logger()
        logger.info("No shutter collect found close enough to "
                    "use for dark histogram statistics")
        return HistogramStatsList(0)

    # Retrieve the histogram statisics
    if collection_type == 'OLI_SOLAR':
        rows = get_histogram_stats_solar(db, rad_id, dark_rad_id, band_number,
                                         unit_test)
    else:
        rows = get_histogram_stats_lamp(db, rad_id, dark_rad_id, band_number,
                                        unit_test)

    # Store the rows into a histogram statistics object and return it
    return HistogramStatsList(len(rows)).add_data(rows)


def get_histogram_stats_solar(db, rad_id, dark_id, band_number, unit_test=None):
    """ Retrieve the Solar histogram statisitcs for a given radiometric
        work order common id
    """
    hist_cal_tbl = "HISTOGRAM_STATISTICS_CAL"
    rad_com_tbl = "RADIOMETRIC_WORK_ORDER_COMMON"
    solar_tbl = "SOLAR_RESPONSE"

    # Check if unit testing is being used
    if unit_test:
        hist_cal_tbl = "DEV_HISTOGRAM_STATISTICS_CAL"
        rad_com_tbl = "DEV_RAD_WORK_ORDER_COMMON"
        solar_tbl = "DEV_SOLAR_RESPONSE"

    query = """ with dhs as
        (
        select
            dark_histogram_statistics.band_id,
            dark_histogram_statistics.processing_level,
            dark_histogram_statistics.sensor_chip_assembly_id,
            dark_histogram_statistics.detector_id,
            avg(dark_histogram_statistics.scene_pixel_mean) as dark_mean,
            avg(dark_histogram_statistics.scene_pixel_stdev) as dark_stdev,
            avg(dark_histogram_statistics.nbr_of_frames) as n_frames
        from
            {0} dark_histogram_statistics,
            {1} radiometric_work_order_common
        where
            radiometric_work_order_common.radiometric_wo_common_id
            = dark_histogram_statistics.radiometric_wo_common_id
        AND
            dark_histogram_statistics.band_id = :BAND
        AND
            dark_histogram_statistics.processing_level = '0'
        AND
            radiometric_work_order_common.radiometric_wo_common_id = :DARKID
        group by
            dark_histogram_statistics.band_id,
            dark_histogram_statistics.processing_level,
            dark_histogram_statistics.sensor_chip_assembly_id,
            dark_histogram_statistics.detector_id
        )

        select
            histogram_statistics.band_id,
            radiometric_work_order_common.date_acquired,
            radiometric_work_order_common.image_start_time,
            radiometric_work_order_common.collection_type,
            decode(solar_response.diffuser_collect_type,
                'PRISTINE','1','WORKING','0','-1') as diffuser,
            histogram_statistics.sensor_chip_assembly_id as sca,
            histogram_statistics.detector_id as det,
            histogram_statistics.nbr_of_frames as n_frames,
            dhs.n_frames as n_dark_frames,
            avg(histogram_statistics.scene_pixel_mean) as mean,
            avg(histogram_statistics.scene_pixel_stdev) as stdev,
            dhs.dark_mean as dark_mean,
            dhs.dark_stdev as dark_stdev,
            radiometric_work_order_common.image_data_truncation_setting
                AS truncation
        from
            {1} radiometric_work_order_common,
            {2} solar_response,
            {0} histogram_statistics,
            dhs
        where
            radiometric_work_order_common.radiometric_wo_common_id
                = solar_response.radiometric_wo_common_id
        AND
            solar_response.radiometric_wo_common_id
                = histogram_statistics.radiometric_wo_common_id
        AND
            histogram_statistics.band_id = dhs.band_id
        AND
            histogram_statistics.sensor_chip_assembly_id
            = dhs.sensor_chip_assembly_id
        AND
            histogram_statistics.detector_id = dhs.detector_id
        AND
            radiometric_work_order_common.radiometric_wo_common_id = :RADID
        --AND
        --    radiometric_work_order_common.valid = 'Y'
        AND
            histogram_statistics.processing_level = 'CTBL'
        group by
            histogram_statistics.band_id,
            radiometric_work_order_common.date_acquired,
            radiometric_work_order_common.image_start_time,
            radiometric_work_order_common.collection_type,
            solar_response.diffuser_collect_type,
            histogram_statistics.sensor_chip_assembly_id,
            histogram_statistics.detector_id,
            histogram_statistics.nbr_of_frames, dhs.n_frames,
            dhs.dark_mean, dhs.dark_stdev,
            radiometric_work_order_common.image_data_truncation_setting
        order by
            histogram_statistics.sensor_chip_assembly_id,
            histogram_statistics.detector_id""".format(
                hist_cal_tbl, rad_com_tbl, solar_tbl)

    parameters = {'BAND': str(band_number), 'RADID':
                  str(rad_id), 'DARKID': dark_id}
    cursor = db.execute(query, parameters)
    all_rows = cursor.fetchall()
    cursor.close()
    return all_rows


def get_histogram_stats_lamp(db, rad_id, dark_id, band_number, unit_test=None):
    """ Retrieve the Lamp histogram statisitcs for a given radiometric
        work order common id
    """
    hist_cal_tbl = "HISTOGRAM_STATISTICS_CAL"
    rad_com_tbl = "RADIOMETRIC_WORK_ORDER_COMMON"
    lamp_tbl = "LAMP_RESPONSE"
    lamp_det_tbl = "LAMP_RESPONSE_DETECTOR"

    # Check if unit testing is being used
    if unit_test:
        hist_cal_tbl = "DEV_HISTOGRAM_STATISTICS_CAL"
        rad_com_tbl = "DEV_RAD_WORK_ORDER_COMMON"
        lamp_tbl = "DEV_LAMP_RESPONSE"
        lamp_det_tbl = "DEV_LAMP_RESPONSE_DETECTOR"

    # Note that the dark frames grouping was removed because it was
    #   creating duplicated rows which created errors in processing
    # Query for the dark collection stats and Lamp collections stats
    query = """ with dhs as (
        select
            radiometric_work_order_common.image_start_time,
            to_char(radiometric_work_order_common.date_acquired,
                'yyyy-mm-yy hh24:mi:ss'),
            dark_histogram_statistics.band_id,
            dark_histogram_statistics.processing_level,
            dark_histogram_statistics.sensor_chip_assembly_id,
            dark_histogram_statistics.detector_id,
            avg(dark_histogram_statistics.scene_pixel_mean)
                as scene_pixel_mean,
            avg(dark_histogram_statistics.scene_pixel_stdev) as
            scene_pixel_stdev,
            avg(dark_histogram_statistics.nbr_of_frames) as n_frames
        from
            {0} dark_histogram_statistics,
            {1} radiometric_work_order_common
        where
            radiometric_work_order_common.radiometric_wo_common_id
                = dark_histogram_statistics.radiometric_wo_common_id
        AND
            dark_histogram_statistics.band_id = :BAND
        AND
            dark_histogram_statistics.processing_level = '0'
        AND
            radiometric_work_order_common.radiometric_wo_common_id = :DARKID
        group by
            radiometric_work_order_common.image_start_time,
            radiometric_work_order_common.date_acquired,
            dark_histogram_statistics.band_id,
            dark_histogram_statistics.processing_level,
            dark_histogram_statistics.sensor_chip_assembly_id,
            dark_histogram_statistics.detector_id
        )
        SELECT
            lamp_response_detector.band_id,
            radiometric_work_order_common.date_acquired,
            radiometric_work_order_common.image_start_time,
            radiometric_work_order_common.collection_type,
            lamp_response.lamp_pair_id as bulb_pair,
            lamp_response_detector.sensor_chip_assembly_id as sca,
            lamp_response_detector.detector_id as det,
            lamp_response.interval_stop_line
                - lamp_response.interval_start_line as n_frames,
            dhs.n_frames as n_dark_frames,
            lamp_response_detector.mean,
            lamp_response_detector.stdev,
            dhs.scene_pixel_mean as dark_mean,
            dhs.scene_pixel_stdev as dark_stdev,
            radiometric_work_order_common.image_data_truncation_setting
            AS truncation
        from
            {1} radiometric_work_order_common,
            {2} lamp_response,
            {3} lamp_response_detector,
            dhs
        where
            radiometric_work_order_common.radiometric_wo_common_id
                = lamp_response.radiometric_wo_common_id
        AND
            lamp_response.radiometric_wo_common_id
                = lamp_response_detector.radiometric_wo_common_id
        AND
            lamp_response_detector.band_id = :BAND
        AND
            lamp_response_detector.band_id = dhs.band_id
        AND
            lamp_response_detector.sensor_chip_assembly_id
                = dhs.sensor_chip_assembly_id
        AND
            lamp_response_detector.detector_id = dhs.detector_id
        AND
            radiometric_work_order_common.radiometric_wo_common_id = :RADID
        AND
            (lamp_response_detector.processing_level = 'CTBL'
            OR
            lamp_response_detector.processing_level = '0TBL' )
        --AND
            --radiometric_work_order_common.valid = 'Y'
        order by
            lamp_response_detector.sensor_chip_assembly_id,
            lamp_response_detector.detector_id""".format(
                hist_cal_tbl, rad_com_tbl, lamp_tbl, lamp_det_tbl)

    parameters = {'BAND': str(band_number), 'RADID':
                  str(rad_id), 'DARKID': dark_id}
    cursor = db.execute(query, parameters)
    all_rows = cursor.fetchall()
    cursor.close()
    return all_rows


def get_closest_dark_shutter_id(db, rad_id, unit_test=None):
    """ Get the shutter collect radiometric work order id that can be paired
        with a single Lamp or Solar collect.

        Note: The shutter collect must be within 5 minutes of Lamp or Solar
              collect
    """
    hist_cal_tbl = "HISTOGRAM_STATISTICS_CAL"
    rad_com_tbl = "RADIOMETRIC_WORK_ORDER_COMMON"

    # Check if unit testing is being used
    if unit_test:
        hist_cal_tbl = "DEV_HISTOGRAM_STATISTICS_CAL"
        rad_com_tbl = "DEV_RAD_WORK_ORDER_COMMON"

    # Create the querys
    hist_query = """ Select Count(*) from {0} where RADIOMETRIC_WO_COMMON_ID
                     = :RADID""".format(hist_cal_tbl)
    query = """Select
        radiometric_wo_common_id from {0} radiometric_work_order_common where
        (SELECT max(date_acquired)
            FROM
                {1}
            WHERE
                radiometric_wo_common_id = :RADID)
        - date_acquired between 0 and .0030
        AND
            collection_type LIKE 'OLI_SHUTTER%'
        AND
            radiometric_work_order_common.work_order_id LIKE '%IS'
        AND
            radiometric_work_order_common.integration_time_oli_ms ='3.6'
        AND
            radiometric_work_order_common.valid = 'Y'
        order by DATE_ACQUIRED desc""".format(rad_com_tbl, rad_com_tbl)
    parameters = {'RADID': rad_id}
    id = db.execute(query, parameters)
    all_ids = id.fetchall()
    id.close()
    if not all_ids:
        return False

    # Check all collects and return the first one with histograms statistics
    for dark_id in all_ids:
        parameters = {'RADID': dark_id[0]}
        cursor = db.execute(hist_query, parameters)
        count = cursor.fetchall()
        cursor.close()
        if count[0][0] > 0:
            return dark_id[0]

    return False


def delete_existing_snr_characterization(db, rad_id, unit_test=None):
    """ Delete the SNR Characterization records for the current id """
    # Check if unit testing is being used
    snr_tbl = "SNR_CHARACTERIZATION"
    if unit_test:
        snr_tbl = "DEV_SNR_CHARACTERIZATION"

    # Create the delete statement
    query = """Delete from {0} where radiometric_wo_common_id = :RADID
            """.format(snr_tbl)
    parameters = {'RADID': rad_id}

    # Delete the records
    delete_cursor = db.execute(query, parameters)
    delete_cursor.close()
