#!/usr/local/bin/python3
import unittest
from subprocess import call
import os
import filecmp
import imp
from ias.db import database as db
import numpy as np
import histogram_stats as hs
import band_snr as bsnr


class TestSNRCharacterization(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """ Setup method ran before all unit tests"""
        db_cred = os.getenv("IAS_DB_L8_OLITIRS")
        if not db_cred:
            raise Exception("IAS_DB_L8_OLITIRS environment variable not set")

        status = call(["sqlplus", db_cred, "@create_tables.sql"])
        if status != 0:
            raise Exception("Creating db tables")

        # Load the data
        status = call(["./load_data.csh"])
        if status != 0:
            raise Exception("Loading data into db tables")

    @staticmethod
    def create_sql_dump_file(filename, results_file, rad_id):
        with open(filename, "w") as f:
            f.write(
                ("""SPOOL "{0}"\nSET LINESIZE 10000\nSET TRIMSPOOL ON\n"""
                 """SET TRIMOUT ON\nSET PAGESIZE 50000\nSET TERMOUT OFF\n"""
                 """Select * from DEV_SNR_CHARACTERIZATION where """
                 """radiometric_wo_common_id = '{1}' order by band_id,\n"""
                 """sensor_chip_assembly_id,detector_id;\n"""
                 """exit;""").format(results_file, rad_id))

    def test_snr_solar(self):
        snr_file = "SNR_SOLAR/SNR_SpecSummary.dat"
        snr_verify = "verify/SOLAR_SNR_SpecSummary.verify"

        ut_dir = os.getenv("IAS_UNIT_TEST_DATA_DIR")
        if not ut_dir:
            raise Exception("IAS_UNIT_TEST_DATA_DIR environment variable not "
                            "set")

        out_file = open("I870.omf", "w")
        # Create the OMF file from the template
        status = call(["sed", "-e",
                       "s|$IAS_UNIT_TEST_DATA_DIR|{0}|".format(ut_dir),
                       "I870.template"], stdout=out_file)
        out_file.close()
        if status != 0:
            raise Exception("Running SED to create the SOLAR OMF")

        status = call(["../characterize_oli_snr.py", "solar.odl"])
        if status != 0:
            raise Exception("Running Solar SNR")

        # Check the Spec file
        self.assertEqual(filecmp.cmp(snr_file, snr_verify), True)

        TestSNRCharacterization.create_sql_dump_file(
            "solar_dump.sql", "solar_results.db", "3716")

        # Remove the previous results file if it exists
        try:
            os.remove("solar_results.db")
        except:
            pass

        db_cred = os.getenv("IAS_DB_L8_OLITIRS")
        if not db_cred:
            raise Exception("IAS_DB_L8_OLITIRS environment variable not set")

        status = call(["sqlplus", db_cred, "@solar_dump.sql"])
        if status != 0:
            raise Exception("Creating db tables")

        # Check the Spec file
        verify_file = (ut_dir + "/ias/rps/toolkit/characterize_oli_snr/"
                       "solar/solar_results.verify")
        self.assertEqual(filecmp.cmp("solar_results.db", verify_file), True)


    def test_snr_lamp(self):
        snr_file = "SNR_LAMP/SNR_SpecSummary.dat"
        snr_verify = "verify/LAMP_SNR_SpecSummary.verify"
        warn_file = "SNR_LAMP/CA_SNR_SpecWarnings.dat"
        warn_verify = "verify/LAMP_CA_SNR_SpecWarnings.verify"

        ut_dir = os.getenv("IAS_UNIT_TEST_DATA_DIR")
        if not ut_dir:
            raise Exception("IAS_UNIT_TEST_DATA_DIR environment variable not "
                            "set")

        # Create the OMF file from the template
        out_file = open("I52851.omf", "w")
        status = call(["sed", "-e",
                       "s|$IAS_UNIT_TEST_DATA_DIR|{0}|".format(ut_dir),
                       "I52851.template"], stdout=out_file)
        out_file.close()
        status = call(["../characterize_oli_snr.py", "lamp.odl"])
        if status != 0:
            raise Exception("Running SED to create the LAMP OMF")

        # Check the Spec file
        self.assertEqual(filecmp.cmp(snr_file, snr_verify), True)

        # Check the CA band spec warnings file
        self.assertEqual(filecmp.cmp(warn_file, warn_verify), True)

        TestSNRCharacterization.create_sql_dump_file(
            "lamp_dump.sql", "lamp_results.db", "824129")

        # Remove the previous results file if it exists
        try:
            os.remove("lamp_results.db")
        except:
            pass

        db_cred = os.getenv("IAS_DB_L8_OLITIRS")
        if not db_cred:
            raise Exception("IAS_DB_L8_OLITIRS environment variable not set")

        status = call(["sqlplus", db_cred, "@lamp_dump.sql"])
        if status != 0:
            raise Exception("Creating db tables")

        # Check the Spec file
        verify_file = (ut_dir + "/ias/rps/toolkit/characterize_oli_snr/"
                       "lamp/lamp_results.verify")
        self.assertEqual(filecmp.cmp("lamp_results.db", verify_file), True)


    def test_get_shutter_collect(self):
        """ Test the shutter collect retrieval function """
        # Create the database connection
        mydb = db.DatabaseConnection("L8_OLITIRS")
        mydb.connect()

        # Setup the ids to retrieve
        ids = ((3716, 3584),
               (824129, 823104),
               (823104, 823104),
               (3584, 3584),
               (1, False),
               (2, False),
               (3583, False),
               (3717, False)
               )

        # Retrieve the shutter collect id for each id
        for id in ids:
            ret_id = hs.get_closest_dark_shutter_id(mydb, id[0], True)
            self.assertEqual(ret_id, id[1], str(id))

        mydb.close(db.DISCONNECT_ROLLBACK)

    def test_get_histogram_statistics(self):
        """ Test the Solar and Lamp hist stats retrieval functions"""
        # Create the database connection
        mydb = db.DatabaseConnection("L8_OLITIRS")
        mydb.connect()

        # Setup the ids to retrieve
        ids = ((3716, 'OLI_SOLAR', 6916),
               (824129, 'OLI_LAMP',  6916),
               (823104, 'OLI_LAMP', 0),
               (3584, 'OLI_LAMP', 0),
               (1, 'OLI_LAMP', 0),
               (2, 'OLI_SOLAR', 0),
               (3583, 'OLI_LAMP', 0),
               (3717, 'OLI_LAMP', 0)
               )

        # Retrieve the shutter collect id for each id
        for id in ids:
            for band in range(1, 10):
                ret_id = hs.get_histogram_stats(mydb, id[0], band, id[1],
                                                 True)
                if band == 8:
                    self.assertEqual(ret_id.size, id[2] * 2, (str(id) +
                                     " failed for band number " + str(band)))
                else:
                    self.assertEqual(ret_id.size, id[2], (str(id) +
                                     " failed for band number " + str(band)))

        mydb.close(db.DISCONNECT_ROLLBACK)

    def sstest_insert_snr_rows(self):
        """ Test the insert SNR rows functions"""

        # Create the database connection
        mydb = db.DatabaseConnection("L8_OLITIRS")
        mydb.connect()
        delete_q = "Delete from DEV_SNR_CHARACTERIZATION"
        cur = mydb.execute(delete_q)
        cur.close()

        # Create the data
        total_rows = 10
        rad_id = np.arange(4000, 4010)
        band = np.arange(1, 11)
        sca = np.arange(2, 12)
        detector = np.arange(1, 11)
        signal = np.arange(1000.0, 1100.0, 10.0)
        frames = np.arange(500.0, 501.0, .1)
        dark_frames = np.arange(502.0, 503.0, .1)
        a = np.arange(.1, 1.1, .1)
        b = np.arange(2.1, 3.1, .1)
        unc_a = np.arange(4.0, 14.0, 1.0)
        unc_b = np.arange(10.0, 20.0, 1.0)
        snr_ltyp = np.arange(250.0, 260.0, 1.0)
        snr_lhigh = np.arange(1200.0, 1300.0, 10.0)
        unc_snr_ltyp = np.arange(3.5, 4.5, .1)
        unc_snr_lhigh = np.arange(20.0, 70.0, 5.0)
        noise = np.arange(30.0, 40.0, 1.0)
        dark_noise = np.arange(2.0, 3.0, .1)

        # Insert the rows
        bsnr.insert_snr_rows(
            mydb, total_rows, rad_id, band, sca, detector,
            signal, frames, dark_frames, a, b, unc_a, unc_b, snr_ltyp,
            snr_lhigh, unc_snr_ltyp, unc_snr_lhigh, noise, dark_noise, True)

        # Get the inserted rows
        query = """Select
                     RADIOMETRIC_WO_COMMON_ID,
                     BAND_ID,
                     SENSOR_CHIP_ASSEMBLY_ID,
                     DETECTOR_ID,
                     SIGNAL_LEVEL,
                     N_FRAMES,
                     N_DARK_FRAMES,
                     A,
                     B,
                     UNCERTAINTY_A,
                     UNCERTAINTY_B,
                     MODELED_SNR_LTYP,
                     MODELED_SNR_LHIGH,
                     SNR_UNCERTAINTY_LTYP,
                     SNR_UNCERTAINTY_LHIGH,
                     NOISE,
                     DARK_NOISE
                   from DEV_SNR_CHARACTERIZATION order by BAND_ID,
                   SENSOR_CHIP_ASSEMBLY_ID, DETECTOR_ID"""
        cur = mydb.execute(query)
        rows = cur.fetchall()
        cur.close()

        # Verify the rows
        self.assertEqual(cur.rowcount, 10)
        for row in range(len(rows)):
            self.assertEqual(rad_id[row], rows[row][0])
            self.assertEqual(band[row], rows[row][1])
            self.assertEqual(sca[row], rows[row][2])
            self.assertEqual(detector[row], rows[row][3])
            self.assertAlmostEqual(signal[row], rows[row][4])
            self.assertAlmostEqual(frames[row], rows[row][5])
            self.assertAlmostEqual(dark_frames[row], rows[row][6])
            self.assertAlmostEqual(a[row], rows[row][7])
            self.assertAlmostEqual(b[row], rows[row][8])
            self.assertAlmostEqual(unc_a[row], rows[row][9])
            self.assertAlmostEqual(unc_b[row], rows[row][10])
            self.assertAlmostEqual(snr_ltyp[row], rows[row][11])
            self.assertAlmostEqual(snr_lhigh[row], rows[row][12])
            self.assertAlmostEqual(unc_snr_ltyp[row], rows[row][13])
            self.assertAlmostEqual(unc_snr_lhigh[row], rows[row][14])
            self.assertAlmostEqual(noise[row], rows[row][15])
            self.assertAlmostEqual(dark_noise[row], rows[row][16])

        mydb.close(db.DISCONNECT_ROLLBACK)

    @classmethod
    def tearDownClass(cls):
        """ Tear down method ran after all unit tests"""
        db_cred = os.getenv("IAS_DB_L8_OLITIRS")
        if not db_cred:
            raise Exception("IAS_DB_L8_OLITIRS environment variable not set")

        status = call(["sqlplus", db_cred, "@cleanup_tables.sql"])
        if status != 0:
            raise Exception("Cleaning up db tables")

if __name__ == '__main__':
    unittest.main()
