#!/usr/local/bin/python3
""" Creates the Trending plots for the input date range """

import argparse
from ias.log import logger as logger
from ias.db import database as db
import re
import numpy as np
from ias.misc import satellite_attributes as sa
from ias.misc import band_specs as bs
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from math import sqrt, ceil, fabs, floor
import os
import datetime as dt


def main():
    # Plot colors for each band
    color = ['#87CEFA', '#0000FF', '#008000', '#FF0000', '#FF6347', '#FF8C00',
             '#BA55D3', '#9ACD32', '#800080']

    # Parse in input arguments
    parser = argparse.ArgumentParser(description='Plot OLI SNR Trend')
    parser.add_argument('--start_date', dest='start_date', required=True,
                        help="Date to start the plots (MM-YYYY)",
                        type=valid_date)
    parser.add_argument('--end_date', dest='end_date', required=True,
                        help="Date to end the plots (MM-YYYY)",
                        type=valid_date)
    parser.add_argument('--output_directory', dest='output_dir', required=True,
                        help="Output directory")
    parser.add_argument('--satellite_id', dest='satellite_number', default=8,
                        help="Satellite Id example: 8", type = int)

    # The unit test argument denotes that the application is being used for the
    # purpose of unit testing so it will point at the development tables
    parser.add_argument('--unit_test', dest='unit_test', action='store_true',
                        help="Unit test development flag")
    args = parser.parse_args()

    # Initialize the logger
    error_logger = logger.Logger("Plot OLI SNR Trend")

    # Validate the start and end date
    if args.start_date > args.end_date:
        raise ValueError("Start date ({0}) is after the end date ({1})".format(
                         args.start_date, args.end_date))

    # Create the output directory if it doesn't exist
    args.output_dir = os.path.realpath(args.output_dir)
    if not os.path.exists(args.output_dir):
        os.makedirs(args.output_dir)

    # Set the correct database table
    trend_db = "SNR_TREND"
    if args.unit_test:
        # Unit test is being run so point at the development instance
        # of the SNR_TREND table
        trend_db = "DEV_SNR_TREND"

    # Create the query to be used for all bands, years, and month combos
    query = """SELECT SNR_AT_LTYPICAL, UNCERTAINTY_LTYPICAL FROM {0}
               where BAND_ID = :band_id and VALID = 'Y'
               and EXTRACT(month from END_DATE) = :month and
               EXTRACT(year from END_DATE) = :year""".format(trend_db)

    # Create the ticks locators
    months = mdates.MonthLocator(interval=3)
    minor_months = mdates.MonthLocator()
    # Format the months.  The single quote is to indicate it is a 2-digit year
    monthsFmt = mdates.DateFormatter("%b '%y")

    # Initialize the satellite attributes
    sat_attr = sa.set_default(sa.get_satellite_id_from_number(
                              args.satellite_number))
    band_specs = bs.BandSpecs(args.satellite_number).band_specs

    # Create the database connection
    mydb = db.DatabaseConnection()
    mydb.set_login_using_satellite_sensor(sat_attr.get_satellite_name(),
                                          sat_attr.get_satellite_sensor_name())
    mydb.connect()

    # Loop through all bands combining the data per detector
    for band_number in sat_attr.band_numbers(sa.SensorId.OLI):
        band_index = sat_attr.convert_band_number_to_index(band_number)

        # Set the looping date
        cur_date = args.start_date

        # Create the trending object
        trend = PlotTrend()
        error_logger.info("Trending band number {0}".format(band_number))

        # Loop over every month
        while True:
            # Set the parameters for this query
            params = {"band_id": band_number, "year": cur_date.year,
                      "month": cur_date.month}
            cursor = mydb.execute(query, params)
            if (cursor.rowcount > 1):
                raise Exception("More that 1 valid trend row found for "
                                "{0}-{1}".format(cur_date.month, cur_date.year))
            rows = cursor.fetchall()

            # Add to the existing plot data if trend data is valid
            if not rows:
                error_logger.warn("No trend date for date {0}-{1}".format(
                        cur_date.month, cur_date.year))
            else:
                trend.add(rows[0][0], rows[0][1], cur_date)

            # Check if this is the stop date
            if (cur_date.year == args.end_date.year and
                    cur_date.month == args.end_date.month):
                break
            cur_date = add_months(cur_date, 1)

        # Check if any months were found to be valid
        if not trend.snr:
            raise Exception("No records were found for band number",
                            band_number)

        # Create subplots for the figure and axis
        figure, axis = plt.subplots()

        # Plot the data and error bars(2 * uncertainty(SNR))
        # A marker of 'D' uses the diamond marker, ls=none dictates
        # that there is no 'LineStyle'
        axis.plot(trend.dates, trend.snr, marker='D',
                markerfacecolor=color[band_index], ls='none')

        # Plotting the error bars. No line style set by 'ls=none'
        # Color 'k' is equivalent to black ref:
        #   matplotlib.org/api/colors_api.html
        axis.errorbar(trend.dates, trend.snr, yerr=trend.uncert, ecolor='k',
                    ls='none')

        # Format the ticks
        axis.xaxis.set_minor_locator(minor_months)
        axis.xaxis.set_major_locator(months)
        axis.xaxis.set_major_formatter(monthsFmt)

        # Convert to numpy array to use min and max methods
        trend.dates = np.array(trend.dates)

        # Setup the ticks and grid
        min_date = add_months(trend.dates.min(), -1)
        max_date = add_months(trend.dates.max(), 1)
        axis.set_xlim(min_date, max_date)
        max_y = np.amax(np.maximum(trend.uncert, trend.snr))
        max_uncert = np.maximum(
            max_y - np.amax(trend.uncert), np.amax(trend.uncert) * 1.25)
        axis.set_ylim(0, max_y + max_uncert)
        axis.yaxis.grid(True)

        # Set the plot labels
        plt.xlabel("Evaluation Period")
        plt.ylabel("SNR Median +/- 2 * UNC(SNR)")
        plt.title("{0} SNR median @Ltypical".format(
            band_specs[band_index].band_color))
        figure.autofmt_xdate()

        # Save the plot to the output directory
        plt.savefig("{0}/{1}.png".format(args.output_dir,
                                         band_specs[band_index].band_color))

    # Close the database connection
    mydb.close(db.DISCONNECT_ROLLBACK)


def valid_date(in_date):
    """ Validates the Input dates

        Example input: "04-2015"
                        MM-YYYY
    """
    # Try and parse the date
    return dt.date(int(in_date[3:]), int(in_date[0:2]), 1)


def add_months(sourcedate, months):
    """ Adds a number of months to the provided date

        Returns: Date
    """
    # Subtraction of months
    if (months < 0):
        # Determine the number of years to subtract
        m1 = -1 * int((np.abs(months) - sourcedate.month) / 12)
        if (months <= -12):
            m1-=1

        # Calculate the month and year
        month = (sourcedate.month + months) % 12
        year = sourcedate.year + m1

    # Addition of months
    else:
        # Calculate the month and year
        month = sourcedate.month - 1 + months
        year = sourcedate.year + int(month / 12)
        month = month % 12 + 1

    return dt.date(year, month, 1)


class PlotTrend:
    """ Per band trending data for plotting """

    def __init__(self):
        self.snr = list()
        self.uncert = list()
        self.dates = list()

    def add(self, snr, uncert, input_date):
        self.snr.append(snr)
        self.uncert.append(2 * uncert)
        self.dates.append(input_date)

# Run the main routine if called from command line
if __name__ == "__main__":
    main()
