#!/usr/local/bin/python3
import unittest
from subprocess import call
from ias.db import database as db
import os
import filecmp
from ias.log import logger as lg
import imp
from datetime import date
import plot_oli_snr_trend as plt


class TestPlotTrend(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """ Setup method ran before all unit tests"""

        # Get the database credentials
        db_cred = os.getenv("IAS_DB_L8_OLITIRS")
        if not db_cred:
            raise Exception("IAS_DB_L8_OLITIRS environment variable not set")

        # Drop and Create the needed tables
        status = call(["sqlplus", db_cred, "@create_tables.sql"])
        if status != 0:
            raise Exception("Creating db tables")

        # Load the data
        status = call(["./load_data.csh"])
        if status != 0:
            raise Exception("Loading data into db tables")

    def test_plot_trend(self):
        """ Run thes plot trend application using the statistics
            in the DEV_SNR_TREND database table"""

        # Disable logging
        os.putenv("IAS_LOG_LEVEL", "DISABLE")

        # Set up all tests in the list
        # START DATE- END DATE- OUTPUT DIR - THROWS EXCEPTION
        tests = (
            ("03-2013", "03-2015", "two_years", False),
            ("03-2013", "03-2013", "one_month", False),
            ("01-2013", "02-2013", "no_data_before", True),
            ("04-2015", "03-2015", "bad_date", True),
            ("03-2000", "03-2400", "400_years", False),
            ("05-2016", "03-2017", "no_data_after", True),
            )

        # Loop over all tests
        for test in tests:
            start = test[0]
            end = test[1]
            output = test[2]
            throws_exception = test[3]
            parms = ["../plot_oli_snr_trend.py", "--start_date", start,
                     "--end_date", end, "--output_directory", output,
                     "--unit_test"]

            if throws_exception:
                print("\nException expected here")

            # Run the application
            status = call(parms)
            if throws_exception:
                if status == 0:
                    raise Exception("No Exception was thrown", parms)
                else:
                    print("End where exception expected")
            else:
                if status != 0:
                    raise Exception("Exception was thrown", parms)

                # Verify the output files
                files = os.listdir("verify/" + output)
                for file in files:
                    if not os.path.isdir("verify/" + output + "/" + file):
                        self.assertEqual(True, filecmp.cmp(
                            "verify/" + output + "/" + file,
                            output + "/" + file))

    def test_valid_dates(self):
        """ Tests the valid date function """
        tests = (("A0-2014", True), ("0B-2014", True), ("04-C014", True),
                 ("03-2D14", True), ("08-20U4", True), ("08-201H", True),
                 ("41-2015", True), ("29-2015", True), ("00-2015", True),
                 ("13-2015", True), ("01-2015", False), ("12-2015", False),
                 )

        # Run each test through the function
        for test in tests:
            # If it throws exception ensure that it does
            if test[1]:
                with self.assertRaises(ValueError):
                    plt.valid_date(test[0])
            else:
                # If exception is thrown the test will fail
                plt.valid_date(test[0])

    def test_add_months(self):
        """ Tests the add months function """
        tests = ((date(2014, 1, 1), 12, date(2015, 1, 1)),
                 (date(2014, 1, 1), 5, date(2014, 6, 1)),
                 (date(2014, 1, 1), 6, date(2014, 7, 1)),
                 (date(2014, 1, 1), 7, date(2014, 8, 1)),
                 (date(2014, 1, 1), 8, date(2014, 9, 1)),
                 (date(2014, 1, 1), 2, date(2014, 3, 1)),
                 (date(2014, 1, 1), 3, date(2014, 4, 1)),
                 (date(2014, 1, 1), 4, date(2014, 5, 1)),
                 (date(2014, 1, 1), 1, date(2014, 2, 1)),
                 (date(2014, 1, 1), 0, date(2014, 1, 1)),
                 (date(2014, 1, 1), 9, date(2014, 10, 1)),
                 (date(2014, 1, 1), 10, date(2014, 11, 1)),
                 (date(2014, 1, 1), 11, date(2014, 12, 1)),
                 (date(2014, 1, 1), 13, date(2015, 2, 1)),
                 (date(2014, 2, 1), 14, date(2015, 4, 1)),
                 (date(2014, 3, 1), 15, date(2015, 6, 1)),
                 (date(2014, 4, 1), 16, date(2015, 8, 1)),
                 (date(2014, 5, 1), 17, date(2015, 10, 1)),
                 (date(2014, 6, 1), 18, date(2015, 12, 1)),
                 (date(2014, 7, 1), 19, date(2016, 2, 1)),
                 (date(2014, 1, 1), 23, date(2015, 12, 1)),
                 (date(2014, 1, 1), 24, date(2016, 1, 1)),
                 (date(2014, 1, 1), 25, date(2016, 2, 1)),
                 )

        # Run through all the tests
        for test in tests:
            # Use the first test to add the months
            return_date = plt.add_months(test[0], test[1])
            self.assertEqual(return_date, test[2], str(test))

            # Use the second to subtract the months
            return_date = plt.add_months(test[2], -1 * test[1])
            self.assertEqual(return_date, test[0], str(test))

    @classmethod
    def tearDownClass(cls):
        """ Tear down method ran after all unit tests"""

        # Get the database credentials
        db_cred = os.getenv("IAS_DB_L8_OLITIRS")
        if not db_cred:
            raise Exception("IAS_DB_L8_OLITIRS environment variable not set")

        # Call the cleanup script
        status = call(["sqlplus", db_cred, "@cleanup_tables.sql"])
        if status != 0:
            raise Exception("Cleaning up db tables")

if __name__ == '__main__':
    unittest.main()
