
/*==============================================================================
NAME: batc_data_segment_stats

PURPOSE:
    This application is used to process BATC data collects, break them up
    into segments and perform histogram statistics on each segment and store
    them in the database for future analysis.

Note: Since the BATC data file is not large, normally only one segment is
      used (that is to do hostogram statistics for the whole collection file).

==============================================================================*/

/* System Includes */
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <math.h>

/* IAS Includes */
#include "ias_cpf.h"
#include "ias_l0r.h"
#include "ias_logging.h"
#include "ias_radiometry.h"
#include "ias_satellite_attributes.h"
#include "ias_db.h"
#include "ias_db_insert.h"
#include "ias_db_get_connect_info.h"
#include "ias_ancillary.h"

/* Local Includes */
#include "get_parameters.h"
#include "report.h"
#include "batc_file_utilities.h"

/* To get debugging output of the raw converted files define the following:
#define DEBUG_DUMP
*/


/*------------------------------------------------------------------------------
NAME: free_detector_offsets

PURPOSE:
    This function will free the dynamically allocated buffers that contain the
    detector offsets for each SCA of a single band.  The array of buffers
    containing the offsets should have been allocated by
    get_detector_offsets().

RETURNS:
    N/A
------------------------------------------------------------------------------*/
void free_detector_offsets
(
    int * spatialOffsets[IAS_MAX_NSCAS],/* IO: Array of buffers */
    int number_scas                          /* I: */
)
{
    int scaIndex;
    
    for (scaIndex = 0; scaIndex < number_scas; scaIndex++)
    {
        free(spatialOffsets[scaIndex]);
        spatialOffsets[scaIndex] = NULL;
    }
}


/*------------------------------------------------------------------------------
NAME: get_detector_offsets

PURPOSE:
    Obtain the detector offsets for each SCA of a specified band.  The array of
    pointers to spatial offset buffers must be all NULL when this function is
    called.  A buffer for the detector offsets for each SCA will be dynamically
    allocated by this function and should be freed by calling
    free_detector_offsets().

RETURNS:
    This function returns a status of SUCCESS if the offsets are successfully
    returned for all SCAs.  It will return a status of ERROR if an error is
    encountered and the pointers in the spatialOffsets array will all be NULL.
------------------------------------------------------------------------------*/
int get_detector_offsets
(
    IAS_CPF * cpf,  /* I: An opened CPF */
    int bandIndex,  /* I: Band index to retrieve offsets for */
    int nScas,      /* I: Number of SCAs for this band */
    int nDetectors, /* I: Number of detectors per SCA for bandNum */
    int * spatialOffsets[IAS_MAX_NSCAS] /* O: Array of buffers */
)
{
    const struct IAS_CPF_DETECTOR_OFFSETS * offsets = NULL;

    int scaIndex;

    /* Retrieve the detector offsets from the CPF */
    offsets = ias_cpf_get_det_offsets(cpf);
    if (offsets == NULL)
    {
        IAS_LOG_ERROR("Retrieving detector offsets from CPF");
        return ERROR;
    }

    /* Allocate a buffer to contain the detector offsets for each SCA */
    for (scaIndex = 0; scaIndex < nScas; scaIndex++)
    {
        spatialOffsets[scaIndex] = malloc(nDetectors * sizeof(int));
        if (spatialOffsets[scaIndex] == NULL)
        {
            IAS_LOG_ERROR("Allocating spatial offsets buffer for %d detectors",
                nDetectors);
            free_detector_offsets(spatialOffsets, nScas);
            return ERROR;
        }

        /* The along_per_detector array is indexed only to IAS_MAX_NBANDS.  Do
           not attempt to make the following ias_rps library call for bands with
           a larger index. */
        if (bandIndex < IAS_MAX_NBANDS)
        {
            ias_rps_convert_detector_delays2spatial_offsets(
                offsets->along_per_detector[bandIndex][scaIndex], nDetectors,
                spatialOffsets[scaIndex], NULL);
        }
        else
        {
            int i;
            for (i = 0; i < nDetectors; i++)
                spatialOffsets[scaIndex][i] = 0;
        }
    }

    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: remove_stats

PURPOSE:
    Remove all previous entries within the SEGMENT_STATISTICS table for the
    specified RADIOMETRIC_WO_COMMON_ID.  This will allow a BATC data collect to
    be reprocessed without encountering a duplicate primary key error when
    statistics are saved into the table.

RETURNS:
    This function returns a status of SUCCESS if the records are successfully
    removed, ERROR if an error is encountered.
------------------------------------------------------------------------------*/
int remove_stats
(
    struct ias_db_connection * db, /* IO: Database connection */
    char rwoid[IAS_CHAR_ID_SIZE]   /* I: Radiometric work order ID */
)
{
    char                  sql[256];
    int                   status;
    struct ias_db_query * query    = NULL;

    /* Delete any previous entries for this work order from the table */
    status = snprintf(sql, sizeof(sql),
        "DELETE FROM SEGMENT_STATISTICS WHERE RADIOMETRIC_WO_COMMON_ID = %s",
        rwoid);
    if ((status < 0) || (status >= sizeof(sql)))
    {
        IAS_LOG_ERROR("SQL command buffer too small, requires %d characters",
            status);
        return ERROR;
    }

    IAS_LOG_INFO("Removing old entries in SEGMENT_STATISTICS for radiometric "
        "work order %s", rwoid);
    query = ias_db_perform_query(db, sql);
    if (!ias_db_query_was_successful(query))
    {
        char errMsg[4096];

        ias_db_query_get_error_message(query, errMsg, sizeof(errMsg));
        IAS_LOG_ERROR("%s", errMsg);
        IAS_LOG_ERROR(
            "Deleting previous records from SEGMENT_STATISTICS for "
            "RADIOMETRIC_WO_COMMON_ID=%s ", rwoid);
        ias_db_query_close(query);
        return ERROR;
    }
    ias_db_query_close(query);
    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: save_stats

PURPOSE:
    This function will save a set of histogram statistics in the RADCHAR
    database in the SEGMENT_STATISTICS table for a single segment (set of
    frames) of a single band.

RETURNS:
    This function will return a status of SUCCESS if the statistics are
    successfully stored in the database, ERROR if an error is encountered.
------------------------------------------------------------------------------*/
int save_stats
(
    struct ias_db_connection * db, /* IO: Database connection */
    char rwoid[IAS_CHAR_ID_SIZE],  /* I: Radiometric work order ID */
    int framesPerSegment,          /* I: Number of frames/segment requested */
    int bandNum,    /* I: Band number */
    int isPanBand,  /* I: Is this the PAN band? */
    int scaNum,     /* I: SCA number */
    int nDetectors, /* I: Number of detectors per SCA in the band */
    int nFrames,    /* I: Number of frames in the segment */
    double aTime,   /* I: Acquisition time of first frame in segment */
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS * stats
                    /* I: histogram results */
)
{
    int          lineSet   = 0;
    int          status;
    int          tableSize;

    if ( isPanBand )
        lineSet = 0;
    else
        lineSet = 2;
    do
    {
        int line_index = isPanBand ? lineSet : 0;
        IAS_DB_INSERT_TABLE dbTable[]=
        {
            IAS_DB_TABLE_STRING( "radiometric_wo_common_id", rwoid ),
            IAS_DB_TABLE_INT( "band_id", bandNum ),
            IAS_DB_TABLE_INT( "sensor_chip_assembly_id", scaNum ),
            IAS_DB_TABLE_INT_ARRAY( "detector_id",
                stats->detector[line_index] ),
            IAS_DB_TABLE_INT( "lineset", lineSet ),
            IAS_DB_TABLE_DOUBLE("acquisition_time", aTime ),
            IAS_DB_TABLE_INT( "segment_size", nFrames ),
            IAS_DB_TABLE_FLOAT_ARRAY( "scene_pixel_min",
                stats->min[line_index] ),
            IAS_DB_TABLE_FLOAT_ARRAY( "scene_pixel_max",
                stats->max[line_index] ),
            IAS_DB_TABLE_FLOAT_ARRAY( "scene_pixel_stdev",
                stats->stdev[line_index] ),
            IAS_DB_TABLE_FLOAT_ARRAY( "scene_pixel_mean",
                stats->mean[line_index] ),
            IAS_DB_TABLE_FLOAT_ARRAY( "scene_pixel_skewness",
                stats->skewness[line_index] ),
            IAS_DB_TABLE_FLOAT_ARRAY( "scene_pixel_kurtosis",
                stats->kurtosis[line_index] ),
            IAS_DB_TABLE_INT_ARRAY( "nbr_of_frames",
                stats->num_values[line_index] ),
            IAS_DB_TABLE_FLOAT_ARRAY( "mean_squared_response",
                stats->mean_sq_resp[line_index] ),
            IAS_DB_TABLE_FLOAT_ARRAY( "adjacent_correlation",
                stats->correlation[line_index] )
        };

        tableSize = sizeof(dbTable) / sizeof(IAS_DB_INSERT_TABLE);

        status = ias_db_insert_using_table(db, "SEGMENT_STATISTICS", dbTable,
            tableSize, nDetectors);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR(
                "Storing histogram stats in database for band %d, SCA %d",
                bandNum, scaNum);
            return ERROR;
        }
    } while (isPanBand && (++lineSet < 2));

    return SUCCESS;
}

/*------------------------------------------------------------------------------
NAME: process_image

PURPOSE:
    This function will process a BATC data file, calculate the
    histogram statistics and store those values in the database.

RETURNS:
    The function will return a status of SUCCESS if the image is successfully
    processed, ERROR if an error is encountered.
------------------------------------------------------------------------------*/
static int process_image
(
    struct ias_db_connection * db, /* IO: Database connection */
    IAS_CPF * cpf,                 /* I:  CPF opened for reading */
    char pathIn[PATH_MAX],         /* I:  Input BATC filename */ 
    int framesPerSegment,          /* I:  Number of frames in each segment */
    int framesSkipTop,             /* I:  Frames to skip at top of image */
    int framesSkipBottom,          /* I:  Frames to skip at bottom of image */
    char rwoid[IAS_CHAR_ID_SIZE],  /* I:  Radiometric work order ID */
    int reportsRequested[NUM_REPORTS],/* I: Flags indicating reports to 
                                            create */
    int truncation_override        /* I: Flag indicating whether to override
                                         the truncation of lower 2 bits of
                                         data.  0 = truncate as normal
                                         1 = do not truncate */
)
{
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS * histoStats = NULL;

    IAS_SPECTRAL_TYPE                spectralType;
    const IAS_CPF_DETECTOR_NOISE *   detectorNoise             = NULL;
    const IAS_CPF_IMPULSE_NOISE *    impulseNoise              = NULL;
    const IAS_CPF_SATURATION_LEVEL * saturation_levels         = NULL;

    double     acquisitionTime;
    int        bandIndex;
    int        bandNum;
    unsigned int framesInBand;
    int        i;
    int        isPanBand;
    int        lastLine;
    int        nDetectors;
    int        nLines;
    int        nLinesPerFrame; 
    int        nReports                      = 0;
    int        nScas;
    int        nSegments;
    int        scaIndex;
    int        segmentLastFrame;
    int        segmentNFrames;
    int        segmentNum;
    int        segmentStartFrame;
    int        startLine;
    int        status;
    uint16_t * imageBuffer = NULL;
    float *flt_image_buffer = NULL;
    int * spatialOffsets[IAS_MAX_NSCAS];
    int band_list[IAS_MAX_TOTAL_BANDS];
    int number_bands;
    int index;
    int temp_nbands;
    int array_size;
    int truncation_setting;

    status = initialize_batc_read();
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Initializing the BATC read routine");
        return ERROR;
    }

    /* Get the truncation setting and the number of frames from the frame
       header */
    status = read_frame_header( pathIn, &framesInBand, &truncation_setting );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Reading the frame header");
        cleanup_batc_read();
        return ERROR;
    }
printf("frames: %u   truncation setting: %d\n", framesInBand,
truncation_setting );

    /* Initialize array of pointers to dynamically allocated memory */
    for (scaIndex = 0; scaIndex < IAS_MAX_NSCAS; scaIndex++)
    {
        spatialOffsets[scaIndex] = NULL;
    }

    /* Determine the number of reports that have been requested */
    for (i = 0; i < NUM_REPORTS; i++)
    {
        if (reportsRequested[i])
            nReports++;
    }

    /* Retrieve the impulse noise from the CPF */
    impulseNoise = ias_cpf_get_impulse_noise(cpf);
    if (impulseNoise == NULL)
    {
        IAS_LOG_ERROR("Retrieving impulse noise from CPF");
        cleanup_batc_read();
        return ERROR;
    }

    /* Retrieve the detector noise from the CPF */
    detectorNoise = ias_cpf_get_det_noise(cpf);
    if (detectorNoise == NULL)
    {
        IAS_LOG_ERROR("Retrieving detector noise from CPF");
        cleanup_batc_read();
        return ERROR;
    }

    /* Retrieve the saturation levels from the CPF */
    saturation_levels = ias_cpf_get_saturation_level(cpf);
    if (saturation_levels == NULL)
    {
        IAS_LOG_ERROR("Retrieving saturation levels from CPF");
        cleanup_batc_read();
        return ERROR;
    }

    /* Delete all prior entries in the SEGMENT_STATISTICS table for this
       radiometric work order */
    if ( db )
    {
        if (remove_stats(db, rwoid) != SUCCESS)
        {
            IAS_LOG_ERROR(
                "Deleting previous records from SEGMENT_STATISTICS for "
                "RADIOMETRIC_WO_COMMON_ID=%s ", rwoid);
            cleanup_batc_read();
            return ERROR;
        }
    }

    /* Get a list of bands to process and the number of bands */
    array_size = IAS_MAX_TOTAL_BANDS;
    number_bands = 0;
    status = ias_sat_attr_get_sensor_band_numbers( IAS_OLI,
        IAS_NORMAL_BAND, 0, &band_list[number_bands],
        array_size, &temp_nbands );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Getting the list of bands");
        cleanup_batc_read();
        return ERROR;
    }
    number_bands += temp_nbands;

    /* Examine ALL bands within the image, not just the normal bands */
    for (index = 0; index < number_bands; index++)
    {
        /* Convert the band index into a band number */
        bandNum = band_list[index];
        bandIndex = ias_sat_attr_convert_band_number_to_index( bandNum );
        if (bandIndex == ERROR)
        {
            IAS_LOG_ERROR("Converting band number %d to a band index",
                bandNum);
            cleanup_batc_read();
            return ERROR;
        }

        /* Get the number of SCAs for this band */
        nScas = ias_sat_attr_get_scas_per_band(bandNum);
        if (nScas == ERROR)
        {
            IAS_LOG_ERROR("Obtaining number of SCAs for band %d", bandNum);
            cleanup_batc_read();
            return ERROR;
        }

        /* Get the number of detectors per SCA */
        nDetectors = ias_sat_attr_get_detectors_per_sca(bandNum);
        if (nDetectors == ERROR)
        {
            IAS_LOG_ERROR("Obtaining number of detectors/SCA for band %d",
                bandNum);
            cleanup_batc_read();
            return ERROR;
        }

        /* Is this the PAN band? */
        spectralType = ias_sat_attr_get_spectral_type_from_band_number(bandNum);
        if (spectralType == ERROR)
        {
            IAS_LOG_ERROR("Unable to determine spectral type for band %d",
                bandNum);
            cleanup_batc_read();
            return ERROR;
        }
        isPanBand = (spectralType == IAS_SPECTRAL_PAN);

        /* Get the number of lines per frame */
        nLinesPerFrame = ias_sat_attr_get_lines_per_frame(bandNum);
        if (nLinesPerFrame == ERROR)
        {
            IAS_LOG_ERROR("Obtaining number of lines/frame for band %d",
                bandNum);
            cleanup_batc_read();
            return ERROR;
        }

        /* Retrieve the spatial offsets for each detector of each SCA for this
           band */
        status = get_detector_offsets(cpf, bandIndex, nScas, nDetectors,
            spatialOffsets);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining detector offsets for band %d", bandNum);
            cleanup_batc_read();
            return ERROR;
        }

        /* Allocate a buffer for reading image data for this band's SCAs */
        imageBuffer = malloc(framesPerSegment * nLinesPerFrame * nDetectors *
            sizeof(uint16_t));
        if (imageBuffer == NULL)
        {
            IAS_LOG_ERROR("Allocating image buffer for reading: %d frames x %d "
                "lines x %d detectors", framesPerSegment, nLinesPerFrame,
                nDetectors);
            free_detector_offsets(spatialOffsets, nScas);
            cleanup_batc_read();
            return ERROR;
        }
        flt_image_buffer = malloc(framesPerSegment * nLinesPerFrame *
            nDetectors * sizeof(float));
        if (flt_image_buffer == NULL)
        {
            IAS_LOG_ERROR("Allocating image buffer for reading: %d frames x %d "
                "lines x %d detectors", framesPerSegment, nLinesPerFrame,
                nDetectors);
            free_detector_offsets(spatialOffsets, nScas);
            free(imageBuffer);
            cleanup_batc_read();
            return ERROR;
        }

        /* Allocate a structure to store the results for the histogram
           statistics computations */
        histoStats = ias_rps_db_allocate_histogram_results(nDetectors,
            isPanBand);
        if (histoStats == NULL)
        {
            IAS_LOG_ERROR("Creating histogram statistics buffer");
            free_detector_offsets(spatialOffsets, nScas);
            free(imageBuffer);
            free( flt_image_buffer );
            cleanup_batc_read();
            return ERROR;
        }

        /* When processing each band, we'll always start the segments at the top
           of the image after skipping the requested number of frames.  In
           addition, all segments will contain the requested number of frames,
           except for the last segment which may be smaller. */
        segmentStartFrame = framesSkipTop;
        segmentNFrames = framesPerSegment;
        nSegments = ceil(
            (double)(framesInBand - framesSkipTop - framesSkipBottom) /
            (double)framesPerSegment);

        IAS_LOG_INFO("Processing band %02d in %d segment%s", bandNum, nSegments,
            ((nSegments == 1) ? "" : "s"));

        segmentNum = 0;
        while (segmentStartFrame < (framesInBand - framesSkipBottom))
        {
            segmentNum++;
            IAS_LOG_INFO("Processing segment %d", segmentNum );

            /* Determine the extent of this segment within the image */
            segmentLastFrame = segmentStartFrame + framesPerSegment;
            if (segmentLastFrame > (framesInBand - framesSkipBottom))
            {
                segmentLastFrame = framesInBand - framesSkipBottom;
                segmentNFrames = segmentLastFrame - segmentStartFrame;
            }

            /* Convert the frame counts into line numbers */
            startLine = segmentStartFrame * nLinesPerFrame;
            nLines = segmentNFrames * nLinesPerFrame;
            lastLine = startLine + nLines - 1;

            /* Process this segment in each SCA of this band */
            for (scaIndex = 0; scaIndex < nScas; scaIndex++)
            {
                IAS_PIXEL_MASK *pixelMask = ias_pm_create(bandNum, scaIndex+1,
                    nDetectors, nLines);
                if ( !pixelMask )
                {
                    IAS_LOG_ERROR("Creating pixel mask for band %d, sca %d, "
                        "frames %d-%d, lines %d-%d", bandNum, scaIndex+1,
                        segmentStartFrame, segmentLastFrame, startLine,
                        lastLine);
                    free_detector_offsets(spatialOffsets, nScas);
                    free(imageBuffer);
                    free( flt_image_buffer );
                    ias_rps_db_free_histogram_results(histoStats);
                    cleanup_batc_read();
                    return ERROR;
                }

                /* Read this segment from the BATC data.  */
                status = read_batc_sca(pathIn, bandNum, scaIndex, 
                         segmentStartFrame, segmentNFrames, 
                         (short *)imageBuffer, &acquisitionTime);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR(
                        "Reading band %d, sca %d, frames %d-%d",
                        bandNum, scaIndex+1, segmentStartFrame,
                        segmentLastFrame);
                    ias_pm_destroy( pixelMask );
                    free_detector_offsets(spatialOffsets, nScas);
                    free(imageBuffer);
                    free( flt_image_buffer );
                    ias_rps_db_free_histogram_results(histoStats);
                    cleanup_batc_read();
                    return ERROR;
                }

                /* If the truncation setting is 1, the data has been
                   barrel-shifted to the left.  This data *must* be shifted
                   back to the lower bits. */
                if ( truncation_setting )
                {
                    for ( i = 0; i < nDetectors * nLines; i++ )
                        imageBuffer[i] >>= 2;
                }

#ifdef DEBUG_DUMP
                if ( 1 )
                {
                    char fn[20];
                    sprintf( fn, "raw2_B%02dS%02d", bandNum, scaIndex + 1 );
                    FILE *ouf = fopen( fn, "w" );
                    if ( !ouf )
                    {
                        IAS_LOG_WARNING("Opening debug output file %s, it "
                            "won't be written", fn );
                    }
                    else
                    {
                        int size = nLines * nDetectors;
                        IAS_LOG_INFO("+++ band %d sca %d nline %d ndets %d "
                            "size %d", bandNum, scaIndex + 1, nLines,
                            nDetectors, size );
                        fwrite( imageBuffer, sizeof(short), size, ouf );
                        fclose( ouf );
                    }
                }
#endif

                /* Retrieve the impulse noise characterization for the current
                   band and SCA. */
                status = ias_rps_impulse_noise_characterization(
                    imageBuffer, bandNum, scaIndex+1, nDetectors, nLines,
                    impulseNoise->median_filter_width[bandIndex],
                    detectorNoise->per_detector[bandIndex][scaIndex],
                    impulseNoise->in_limit[bandIndex], pixelMask,
                    NULL, NULL);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR(
                        "Running ias_rps_impulse_noise_characterization() "
                        "on band %d, sca %d, frames %d-%d (e.g. lines "
                        "%d-%d)", bandNum, scaIndex+1, segmentStartFrame,
                        segmentLastFrame, startLine, lastLine);
                    ias_pm_destroy( pixelMask );
                    free_detector_offsets(spatialOffsets, nScas);
                    free(imageBuffer);
                    free( flt_image_buffer );
                    ias_rps_db_free_histogram_results(histoStats);
                    cleanup_batc_read();
                    return ERROR;
                }

                /* Retrieve the saturated pixel levels for the current band and
                   SCA. */
                status = ias_rps_saturated_pixel_characterization(
                    imageBuffer, bandNum, scaIndex+1, nDetectors, nLines,
                    saturation_levels->
                        digital_low_saturation_level[bandIndex][scaIndex],
                    saturation_levels->
                        digital_high_saturation_level[bandIndex][scaIndex],
                    saturation_levels->
                        analog_low_saturation_level[bandIndex][scaIndex],
                    saturation_levels->
                        analog_high_saturation_level[bandIndex][scaIndex],
                    pixelMask, NULL, NULL);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR(
                        "Running ias_rps_saturated_pixel_characterization"
                        "() on "
                        "band %d, sca %d, frames %d-%d (e.g. lines %d-%d)",
                        bandNum, scaIndex+1, segmentStartFrame,
                        segmentLastFrame, startLine, lastLine);
                    ias_pm_destroy( pixelMask );
                    free_detector_offsets(spatialOffsets, nScas);
                    free(imageBuffer);
                    free( flt_image_buffer );
                    ias_rps_db_free_histogram_results(histoStats);
                    cleanup_batc_read();
                    return ERROR;
                }

                /* Convert the image data to floating point.  There are
                   three possible situations with two reactions:

                   1) truncation setting is 1
                   2) truncation setting is 0 and override flag is 1
                   3) truncation setting is 0 and override flag is 0

                   For 1) and 2) no action other than copy to float.
                   For 3) the lower two bits are set to zero; it's copied
                     to floating point and 1.5 is added

                   Steps 1) and 3) follow the standard create_l1r processing
                   flow's reaction to the two data types.  Step 2) gives
                   Cal/Val the option to look at all 14 bits of data. */
                if ( truncation_setting  == 0 && truncation_override == 0 )
                {
                    for ( i = 0; i < nDetectors * nLines; i++ )
                    {
                        flt_image_buffer[i] =
                            (float)( imageBuffer[i] & 0xfffc ) + 1.5;
                    }
                }
                else
                {
                    for ( i = 0; i < nDetectors * nLines; i++ )
                        flt_image_buffer[i] = (float)imageBuffer[i];
                }

                /* Retrieve the histogram statistics.  The number of frames to
                   skip at the top and bottom of the image are hard coded to
                   zero because this application has already skipped the
                   necessary frames. */
                status = ias_rps_histogram_statistics_characterization(
                    flt_image_buffer, IAS_R4, bandNum, scaIndex+1, nDetectors,
                    nLines, spatialOffsets[scaIndex], 0, 0,
                    pixelMask, NULL, histoStats, NULL);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR(
                        "Running ias_rps_histogram_statistics_characterization("
                        ") on band %d, sca %d, frames %d-%d (e.g. lines %d-%d)",
                        bandNum, scaIndex+1, segmentStartFrame,
                        segmentLastFrame, startLine, lastLine);
                    ias_pm_destroy( pixelMask );
                    free_detector_offsets(spatialOffsets, nScas);
                    free(imageBuffer);
                    free( flt_image_buffer );
                    ias_rps_db_free_histogram_results(histoStats);
                    cleanup_batc_read();
                    return ERROR;
                }

                /* Save the histogram statistics to the database */
                if ( db )
                {
                    status = save_stats(db, rwoid, framesPerSegment, bandNum,
                        isPanBand, scaIndex+1, nDetectors, segmentNFrames,
                        acquisitionTime, histoStats);
                    if (status != SUCCESS)
                    {
                        IAS_LOG_ERROR("Saving histogram statistics to the "
                            "database for band %d, sca %d, frames %d-%d "
                            "(e.g. lines %d-%d)", bandNum,
                            scaIndex+1, segmentStartFrame, segmentLastFrame,
                            startLine, lastLine);
                        ias_pm_destroy( pixelMask );
                        free_detector_offsets(spatialOffsets, nScas);
                        free(imageBuffer);
                        free( flt_image_buffer );
                        ias_rps_db_free_histogram_results(histoStats);
                        cleanup_batc_read();
                        return ERROR;
                    }
                }

                /* If the user hasn't requested any report files be generated,
                   we'll skip the rest of this loop */
                if (nReports > 0)
                {

                    /* Save the requested statistics to a report file
                       (e.g. an image of floating point values */
                    status = report_save(reportsRequested, bandNum, isPanBand,
                        scaIndex+1, nScas, nDetectors, segmentNum,
                        framesPerSegment, nSegments, histoStats);
                    if (status != SUCCESS)
                    {
                        IAS_LOG_ERROR(
                            "Saving report file(s) for band %d, sca %d, frames "
                            "%d-%d (e.g. lines %d-%d)", bandNum, scaIndex+1,
                            segmentStartFrame, segmentLastFrame, startLine,
                            lastLine);
                        ias_pm_destroy( pixelMask );
                        free_detector_offsets(spatialOffsets, nScas);
                        free(imageBuffer);
                        free( flt_image_buffer );
                        ias_rps_db_free_histogram_results(histoStats);
                        cleanup_batc_read();
                        return ERROR;
                    }
                }
                ias_pm_destroy( pixelMask );
            }

            /* Update the starting frame number for the next segment that will
               be processed for this band across all SCAs */
            segmentStartFrame += framesPerSegment;
        }

        /* Free the dynamically allocated memory for this band */
        free(imageBuffer);
        free( flt_image_buffer );
        imageBuffer = NULL;
        flt_image_buffer = NULL;

        ias_rps_db_free_histogram_results(histoStats);
        free_detector_offsets(spatialOffsets, nScas);
    }

    cleanup_batc_read();
    return SUCCESS;
}


int main (int argc, char * argv[])
{
    struct ias_db_connection * db = NULL;

    IAS_CPF *        cpf                           = NULL;
    IAS_SATELLITE_ID satId;
    char             pathCpf[PATH_MAX]             = "";
    char             pathIn[PATH_MAX]              = "";
    char             rwoid[IAS_CHAR_ID_SIZE];
    char             woid[IAS_WORKORDER_ID_SIZE];
    char *           dbHost                        = NULL;
    char *           dbName                        = NULL;
    char *           dbPassword                    = NULL;
    char *           dbUser                        = NULL;
    int              framesPerSegment;
    int              framesSkipBottom;
    int              framesSkipTop;
    int              i;
    int              reportsRequested[NUM_REPORTS];
    int              satNum = 8;
    int              status;
    int store_in_database;
    const struct IAS_CPF_HISTOGRAM_CHAR *hist_char = NULL;
    int truncation_override;

    /* Initialize logging */
    if (ias_log_initialize("batc_data_segment_stats"))
    {
        IAS_LOG_ERROR("Initializing logging capabilities");
        exit(EXIT_FAILURE);
    }

    /* Initialize the array that tracks which reports the user has requested.
       If reportsRequested[REPORT_XXX] is TRUE then the user has requested that
       report.  If the value is FALSE, that report should not be generated. */
    for (i = 0; i < NUM_REPORTS; i++)
        reportsRequested[i] = FALSE;

    /* Process the command line arguments */
    if (argc != 2)
    {
        IAS_LOG_ERROR("usage: batc_data_segment_stats <parameter file>");
        exit(EXIT_FAILURE);
    }
    if (get_parameters(argv[1], woid, pathCpf, pathIn, 
        &framesPerSegment, &store_in_database, &truncation_override,
        reportsRequested) != SUCCESS)
    {
        IAS_LOG_ERROR("Getting parameters");
        exit(EXIT_FAILURE);
    }
    IAS_LOG_INFO("Processing started");

    /* Using the satellite number read from the OMF file, initialize the
       satellite attributes library */
    satId = ias_sat_attr_get_satellite_id_from_satellite_number(satNum);
    if (satId == ERROR)
    {
        IAS_LOG_ERROR("Unsupported satellite number: %d", satNum);
        exit(EXIT_FAILURE);
    }
    if (ias_sat_attr_initialize(satId) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attribute library");
        exit(EXIT_FAILURE);
    }

    /* Connect to the database if user requested it */
    if ( store_in_database )
    {
        if (ias_db_initialize_database_lib() != SUCCESS)
        {
            IAS_LOG_ERROR("Unable initialize database library");
            exit(EXIT_FAILURE);
        }
        if (ias_db_get_connect_info(&dbName, &dbUser, &dbPassword, &dbHost) !=
            SUCCESS)
        {
            IAS_LOG_ERROR("Unable retrieve database connection information");
            ias_db_close_database_lib();
            exit(EXIT_FAILURE);
        }
        db = ias_db_connect_to_database(dbName, dbUser, dbPassword, dbHost);
        free(dbName);
        free(dbUser);
        free(dbPassword);
        free(dbHost);
        if (db == NULL)
        {
            IAS_LOG_ERROR("Unable to establish database connection");
            ias_db_close_database_lib();
            exit(EXIT_FAILURE);
        }

        /* Obtain the radiometric work order ID for this IAS work order */
        status = ias_misc_get_char_id(db, "RADIOMETRIC_WORK_ORDER_COMMON",
            "RADIOMETRIC_WO_COMMON_ID", woid, rwoid);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR(
                "Unable to obtain RADIOMETRIC_WO_COMMON_ID for work order: %s",
                woid);
            ias_db_close_connection(db);
            ias_db_close_database_lib();
            exit(EXIT_FAILURE);
        }
    }

    /* Display information about what is being processed */
    IAS_LOG_DEBUG("            satNum = %d", satNum);
    IAS_LOG_DEBUG("       workOrderId = %s", woid);
    IAS_LOG_DEBUG("            pathIn = %s", pathIn);
    IAS_LOG_DEBUG("           pathCpf = %s", pathCpf);
    IAS_LOG_DEBUG("    frames/segment = %d", framesPerSegment);
    IAS_LOG_DEBUG("             rwoid = %s\n", rwoid);

    /* Open the CPF for reading */
    cpf = ias_cpf_read(pathCpf);
    if (cpf == NULL)
    {
        IAS_LOG_ERROR("Reading CPF: %s", pathCpf);
        if ( store_in_database )
        {
            ias_db_close_connection(db);
            ias_db_close_database_lib();
        }
        exit(EXIT_FAILURE);
    }

    /* Get the frames to skip top and bottom from the CPF */
    hist_char = ias_cpf_get_histogram_char( cpf );
    if ( !hist_char )
    {
        IAS_LOG_ERROR("Getting the histogram char parameters from the CPF");
        if ( store_in_database )
        {
            ias_db_close_connection(db);
            ias_db_close_database_lib();
        }
        exit(EXIT_FAILURE);
    }
    framesSkipTop = hist_char->frames_to_skip_top;
    framesSkipBottom = hist_char->frames_to_skip_bottom;

    /* Process the image specified */
    if (process_image(db, cpf, pathIn, framesPerSegment, framesSkipTop,
            framesSkipBottom, rwoid, reportsRequested, truncation_override)
            != SUCCESS)
    {
        IAS_LOG_ERROR("Processing image: %s", pathIn);
        ias_cpf_free(cpf);
        if ( store_in_database )
        {
            ias_db_close_connection(db);
            ias_db_close_database_lib();
        }
        exit(EXIT_FAILURE);
    }

    /* Close the database connection */
    if ( store_in_database )
    {
        ias_db_close_connection(db);
        ias_db_close_database_lib();
    }

    /* Free the CPF */
    ias_cpf_free(cpf);

    /* Compute the processing time */
    IAS_LOG_INFO("Completed successfully");
    exit(EXIT_SUCCESS);
}
