/*-----------------------------------------------------------------------------
NAME:    batc_file_utilities.c

PURPOSE: Routines to access the BATC data files includes:
             initialize_batc_read
             cleanup_batc_read
             read_frame_header
             read_batc_sca

             local - get_line_header_acquisition_time

NOTES:   There are no "good" documents that explain the layout of the data
         in the file.  The best for general layout, including the locations
         of flags in the data headers, is "INTERFACE CONTROL DRAWING,
         OLI INSTRUMENT ELECTRICAL" which is document 2288002_B_EICD.doc.
         However it doesn't explain how the data is transferred from the
         instrument, and ultimately, to the file.
         
         The main thing to be aware of is that the all data are transferred
         over the same bus and is therefore the upper 12-bits of data
         registers.  That means counters for line number, time, etc as well
         as some data (indicated by truncation setting) are barrel-shifted.
         This 12-bit information is then stored, big-endian style, in
         16-bits.  So a 32-bit integer gets stored in 4, 16-bit locations;
         16-bit short in 2, 16-bit locations, and anything <= 12-bits into a
         single 16-bit location.

RETURNS: various
-----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include "ias_const.h"
#include "ias_logging.h"

#define  PSEUDO_BANDS      13       /* Number of pseudo bands.  Pseudo band
                                       is a Cal/Val convention: there are 
                                       eight standard bands and a blind
                                       band; and four "bands" of pan band
                                       data -- hence pseudo bands */
#define  NUMBER_SCA        14       /* Number of SCAs */
#define  MS_NDET           494      /* Multispectral detectors per SCA */
#define  MS_NVRP           12       /* Multispectral VRPs per SCA (6 on 
                                             each side) */
#define  PAN_NDET          988      /* Pan detectors per SCA */
#define  PAN_NVRP          24       /* Pan VRPs per SCA (12 on each side) */
#define  LINE_HEADER_SIZE  16       /* Size of line header */
#define  TRAILER_SIZE      4        /* Size of trailer (CRC) */
#define  FRAME_HEADER_SIZE 48       /* Size of frame header, this is
                                       contained in the first frame and
                                       includes the line header size */
#define  SEC_PER_DAY       86400 

/* The block size is the size of the data for each pseudo band */
#define  BLOCK_SIZE (LINE_HEADER_SIZE + (MS_NDET + MS_NVRP) * NUMBER_SCA + \
                    TRAILER_SIZE)
#define  FRAME_SIZE (BLOCK_SIZE * PSEUDO_BANDS)


/* One frame of data: each frame of BATC data contains a line header and 13
   pseudo bands of data and a CRC */
static short *frame_buffer = NULL;


/*-----------------------------------------------------------------------------
NAME:    initialize_batc_read

PURPOSE: Allocate memory for the frame buffer

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int initialize_batc_read(void)
{
    frame_buffer = malloc( FRAME_SIZE * sizeof(short) );
    if( frame_buffer == NULL )
    {
        IAS_LOG_ERROR("Error allocating frame_buffer\n");
        return ERROR;
    }
    return SUCCESS;
}


/*-----------------------------------------------------------------------------
NAME:    cleanup_batc_read

PURPOSE: Allocate memory for the frame buffer

RETURNS: none
-----------------------------------------------------------------------------*/
void cleanup_batc_read(void)
{
    free( frame_buffer );
    frame_buffer = NULL;
}


/*-----------------------------------------------------------------------------
NAME: get_line_header_acquisition_time

PURPOSE:
    This function will read in the acquisition time from the line 
    header data.

RETURNS: none
-----------------------------------------------------------------------------*/
static void get_line_header_acquisition_time
(
    short *rawbuf,              /* I: raw buffer */
    double *acquisition_time    /* O: Acquisition time from the line header */
)
{
    int day = 0;                /* Day of line */
    int milli = 0;              /* Milliseconds of line */
    int micro = 0;              /* Microseconds of line */
    double total_secs;          /* Total seconds of line time */

    /* The data goes through some gyrations: first it's barrel shifted and
       upper 12-bits read in then stored big-endian style in 16-bit values.
       To extract the data, we need to  shift the values we are interested
       in to the right twice. */
    rawbuf[4]  >>= 2;
    rawbuf[5]  >>= 2;
    rawbuf[6]  >>= 2;
    rawbuf[7]  >>= 2;
    rawbuf[8]  >>= 2;
    rawbuf[9]  >>= 2;
    rawbuf[10] >>= 2;
    rawbuf[11] >>= 2;

    /* Can't just read out the values as a long or int since there are only
       12 bits of information in each of the chunks.  Convert the big-endian
       value to appropriate little-endian int or short */
    day    = rawbuf[4] << 8;
    day   |= rawbuf[5];
    milli  = rawbuf[6] << 24;
    milli |= rawbuf[7] << 16;
    milli |= rawbuf[8] << 8;
    milli |= rawbuf[9];
    micro  = rawbuf[10] << 8;
    micro |= rawbuf[11];

    total_secs = day * SEC_PER_DAY + milli / 1.0e3 + micro / 1.0e6;

    *acquisition_time = total_secs;
}  /* END get_line_header_acquisition_time */


/*-----------------------------------------------------------------------------
NAME:    read_frame_header

PURPOSE: Read the dat file frame header and determine the number of frames
         in the file and the truncation setting.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int read_frame_header
(
    char *filename,                /* I: Name of the data file */
    unsigned int *number_of_frames,/* O: Number of frames listed in the
                                         header */
    int *truncation_setting        /* O: The value of the truncation setting */
)
{
    int status;
    FILE *inf;
    short buffer[FRAME_HEADER_SIZE];
    unsigned int num_frames;
    int n;

    inf = fopen( filename, "r" );
    if ( inf == NULL )
    {
        IAS_LOG_ERROR("Opening file: %s", filename );
        return ERROR;
    }

    /* Read in the frame header */
    n = fread( buffer, sizeof(short), FRAME_HEADER_SIZE, inf );
    if ( n != FRAME_HEADER_SIZE )
    {
        IAS_LOG_ERROR("Reading the frame header: read %d of %d", n,
            FRAME_HEADER_SIZE );
        fclose( inf );
        return ERROR;
    }
    status = fclose( inf );
    if ( status )
    {
        IAS_LOG_ERROR("Closing file: %s", filename );
        return ERROR;
    }

    /* The data goes through some gyrations: first it's barrel shifted and
       upper 12-bits read in then stored big-endian style in 16-bit values.
       To extract the data, we need to  shift the values we are interested
       in to the right twice. */
    buffer[16] >>= 2;  /* first four are the number of lines */
    buffer[17] >>= 2;
    buffer[18] >>= 2;
    buffer[19] >>= 2;
    buffer[47] >>= 2;  /* this is the truncation setting */

    /* Can't just read out the value as a long since there are only 12 bits
       of information in each of the four chunks.  Convert the big-endian
       value to an unsigned integer */
    num_frames  = buffer[16] << 24;
    num_frames += buffer[17] << 16;
    num_frames += buffer[18] << 8;
    num_frames += buffer[19];
    *number_of_frames = num_frames;

    /* The truncation setting is only a single value at bit 0 so no
       conversion is needed */
    *truncation_setting = (int)buffer[47];

    return SUCCESS;
}  /* END read_frame_header */


/*-----------------------------------------------------------------------------
NAME: read_batc_sca

PURPOSE:
    This function will read in one SCA BATC data.  

RETURNS:
    This function returns a status of SUCCESS if the one SCA data is read, 
    ERROR if an error is encountered.
-----------------------------------------------------------------------------*/
int read_batc_sca
(
    char *dat_name,            /* I: EDU data file */
    int rband,                 /* I: Band to retrieve */
    int tsca,                  /* I: SCA with profile */
    int start_line,            /* I: start_line for one segment to read */
    int num_lines,             /* I: num_lines for one segment to read */
    short *outbuf,             /* O: Buffer containing image data */
    double *acquisition_time   /* O: Acquisition time for the first line of
                                     the data buffer */
)
{
    short *imgbuf;              /* One set of data, includes several items */
    FILE *ifp;                  /* Image file pointer */
    long line, det, sca, chunk; /* Counters */
    int sca_size;               /* Size of a SCA's worth of data */
    int band_offset;            /* Offset into misison frmae */
    short *iptr;                /* Pointer to input data */
    short *optr;                /* Pointer to output data */
    long image_size;            /* Size of one line of imagery */
    int nchunks = 1;            /* Number of "sets" of band data within a 
                                   frame */
    long sca_offset;            /* Offset do to SCA knife edge resides in */
    int index;
    int not_have_time = 1;
    /* This band look up table gives the block in the frame where the data
       for that band begins not counting the line header.  These are the
       blocks:
           Block Band Name
             0     8  pan odd
             1     8  pan even
             2     2  blue
             3     1  coastal aerosol
             4     5  NIR
             5     4  red
             6     3  green
             7     8  pan odd
             8     8  pan even
             9     7  SWIR2
            10     6  SWIR1
            11     9  cirrus
            12    10  dark

       Indices into the LUT are rband - 1.  The pan band (index 7) is set to
       zero in the table since it isn't used. */
    int band_lut[] = {3,2,6,5,4,10,9,0,11}; 

    /* Open file */
    ifp = fopen(dat_name, "r");
    if(ifp == NULL)
    {
        IAS_LOG_ERROR("Error opening input %s\n", dat_name);
        return ERROR;
    }

    /* Set number of detectors according to band number */
    if(rband == 8)
        sca_size = PAN_NDET;
    else
        sca_size = MS_NDET;
    image_size = sca_size * NUMBER_SCA;

    /* Pan band is twice the size of normal bands */
    /* TODO there is no need for these intermediate buffers, data could be
       TODO copied directly into the output buffer, however, in the interest
       TODO of time, they are being left as is for now */
        if(rband == 8)
        imgbuf = malloc( 2 * image_size * sizeof(short) );
    else
        imgbuf = malloc( image_size * sizeof(short) );
    if( imgbuf == NULL )
    {
        IAS_LOG_ERROR("Error allocating imgbuf");
        return ERROR;
    }

    /* Skip the frame header and lines before start_line */
    fseek(ifp, (( 1 + start_line ) * FRAME_SIZE * sizeof(short)), SEEK_SET);

    int line_number = start_line;
    for(line = 0; line < num_lines; line++, line_number++)
    {
        not_have_time = 1;
        /* Read in one mission level frames worth of data */
        if (fread(frame_buffer, sizeof(short), FRAME_SIZE, ifp )
            != FRAME_SIZE)
        {
            IAS_LOG_ERROR("Error reading line %d from file %s",
                line_number, dat_name);
            return ERROR;
        }

        /* Get the first line acquisition time */
        if ( not_have_time )
        {
            get_line_header_acquisition_time( frame_buffer,
                acquisition_time );
            not_have_time = 0;
        }

        if ( rband == 8 )
        {
            /* Output is 2 times number of frames */
            band_lut[0] = 0;
            band_lut[1] = 1;
            band_lut[2] = 7;
            band_lut[3] = 8;
            nchunks = 4;

            /* One frame contains two lines of PAN data.  These two lines are 
               spit into four chunks of data.
               1st chunk (1st in frame) PAN odd  line N
               2nd chunk (2nd in frame) PAN even line N
               3rd chunk (8th in frame) PAN odd  line N+1
               4th chunk (9th in frame) PAN even line N+1
               This data needs to be reassembled into image order. */
            for(chunk=0; chunk < nchunks; chunk++)
            {
                /* Calculate band or chunk offset */
                band_offset = band_lut[chunk] * BLOCK_SIZE + LINE_HEADER_SIZE; 
                iptr = frame_buffer + band_offset;
                iptr = iptr + NUMBER_SCA * PAN_NVRP / 4;

                /* Reorder data */
                for(det = 0; det < PAN_NDET; det = det + 2)
                {
                    if (chunk == 0)
                        optr = imgbuf + det;
                    else if (chunk == 1)
                        optr = imgbuf + det + 1;
                    else if (chunk == 2)
                        optr = imgbuf + det + NUMBER_SCA * PAN_NDET;
                    else 
                        optr = imgbuf + det + NUMBER_SCA * PAN_NDET + 1;
                    for(sca = 0; sca < NUMBER_SCA; sca++, optr = optr + 
                            PAN_NDET, iptr++)
                    {
                        *optr = *iptr;
                    }
                }
            }

            /* Write out two lines of PAN data */
            sca_offset = tsca * sca_size;
            if (tsca % 2)
            {
                for (index = 0; index < sca_size; index++) 
                {
                    outbuf[2 * line * sca_size + index] = 
                        imgbuf[sca_offset + sca_size - index - 1]; 
                }
            }
            else
            {
                memcpy(&outbuf[2 * line * sca_size], &imgbuf[sca_offset],
                        sizeof(short) * sca_size);
            }
            sca_offset = sca_offset + NUMBER_SCA * sca_size;
            if (tsca % 2)
            {
                for (index = 0; index < sca_size; index++) 
                {
                    outbuf[(2 * line + 1) * sca_size + index] = 
                        imgbuf[sca_offset + sca_size - index - 1]; 
                }
            }
            else
            {
                memcpy(&outbuf[(2 * line + 1) * sca_size], &imgbuf[sca_offset],
                        sizeof(short) * sca_size);
            }
        }  /* END pan band processing */
        else
        {
            /* Calculate the band offset index into the look up table is 
               band number - 1 */
            band_offset = band_lut[rband - 1] * BLOCK_SIZE + LINE_HEADER_SIZE;

            /* Reorganize image data.*
               DAT data is stored:
               detector 1 all 14 SCAs
               detector 2 all 14 SCAs
               ....
               detector N all 14 SCAs
               where N is the number of detectors within a band. */
            iptr = frame_buffer + band_offset;
            iptr = iptr + NUMBER_SCA * MS_NVRP / 2;
            for(det=0 ; det < MS_NDET ; det++)
            {
                for(sca = 0, optr = imgbuf+det ; sca < NUMBER_SCA ; sca++, 
                        optr = optr + MS_NDET, iptr++)
                    *optr = *iptr;
            }
            sca_offset = tsca * sca_size;
            if (tsca % 2)
            {
                for (index = 0; index < sca_size; index++) 
                {
                    outbuf[line * sca_size + index] = 
                        imgbuf[sca_offset + sca_size - index - 1]; 
                }
            }
            else
            {
                memcpy(&outbuf[line * sca_size], &imgbuf[sca_offset],
                        sizeof(short) * sca_size);
            }
        }  /* END MS band processing */

    }  /* END for line loop */

    fclose(ifp);
    free(imgbuf);

    return SUCCESS;
}  /* END read_batc_sca */
