/*==============================================================================
NAME: report.c

PURPOSE:
    This file contains the functions used to create the optional report files
    that a user may request.
==============================================================================*/

/* System Includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>
#include <sys/stat.h>

/* IAS Includes */
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_radiometry.h"
#include "config.h" /* This is for the endian define */

/* Local Includes */
#include "report.h"

#define REPORT_DIR "./report_files"


/*------------------------------------------------------------------------------
NAME: util_create_envi_header

PURPOSE:
    This function will create an ENVI header file to make it easier for the user
    to view the report files using ENVI.

RETURNS:
    This function returns a status of SUCCESS if the header is successfully
    created, ERROR if an error is encountered.
------------------------------------------------------------------------------*/
static int util_create_envi_header (
    const char path[PATH_MAX], /* I: Path to the report file */
    int nLines,                /* I: Number of lines in the image */
    int nSamples               /* I: Number of samples in the image */
    )
{
    FILE * fd;
    char   hdrPath[PATH_MAX];
    int    status;
    int byte_order;

    status = snprintf(hdrPath, sizeof(hdrPath), "%s.hdr", path);
    if ((status < 0) || (status >= sizeof(hdrPath)))
    {
        IAS_LOG_ERROR("Buffer for ENVI header requires %d characters", status);
        return ERROR;
    }

    if ( HAVE_LITTLE_ENDIAN )
        byte_order = 0;
    else
        byte_order = 1;

    fd = fopen(hdrPath, "w");
    if (fd == NULL)
    {
        IAS_LOG_ERROR("Creating ENVI header: %s", hdrPath);
        return ERROR;
    }

    status = fprintf(fd,
        "ENVI\n"
        "description = {BATC data collect segment statistics}\n"
        "samples = %d\n"
        "lines   = %d\n"
        "bands   = 1\n"
        "header offset = 0\n"
        "file type = ENVI Standard\n"
        "data type = 4\n"
        "interleave = bsq\n"
        "sensor type = Unknown\n"
        "byte order = %d\n"
        "wavelength units = Unknown\n",
        nSamples, nLines, byte_order );
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing ENVI header: %s", hdrPath);
        return ERROR;
    }
    status = fclose(fd);
    if ( status )
    {
        IAS_LOG_ERROR("Closing %s", hdrPath);
        return ERROR;
    }
    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: util_setup

PURPOSE:
    Given the report number being generated, band information, the number of
    frames in the segment, and the histogram statistics generated by
    ias_rps_histogram_statistics_characterization(), this function will create a
    report file name and return a pointer to the floating point data that should
    be written to the report file.

    The pointer returned, references an array of floating point values that
    contains a value for each detector within an SCA of the band.

RETURNS:
    If successful, this function will return a pointer to the floating point
    information that should be written to the report file, if an error is
    encountered it will return NULL.
------------------------------------------------------------------------------*/
static float * util_setup (
    int reportNum,        /* I: Report number */
    int bandNum,          /* I: Band number to create report for */
    int isPanBand,        /* I: Is this the PAN band? */
    int lineSet,          /* I: Line set, ignored for all bands except PAN */
    int framesPerSegment, /* I: Number of frames per segment */
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS * stats,
                          /* I: histogram results */
    char path[PATH_MAX]   /* O: Report file name */
    )
{
    char        reportName[64];
    float *     data;
    int         nChars;
    int data_index;

    /* Validate the line set specified */
    if (isPanBand && ((lineSet < 0) || (lineSet > 1)))
    {
        IAS_LOG_ERROR("Specified line set value %d is out of range 0-1",
            lineSet);
        return NULL;
    }

    /* Set the data index.  The data for non-pan band are placed in index 0
       and for pan band either index 0 or 1. */
    if ( isPanBand )
        data_index = lineSet;
    else
        data_index = 0;

    switch (reportNum)
    {
        case REPORT_MIN:
            strcpy(reportName, "min");
            data = stats->min[data_index];
            break;

        case REPORT_MAX:
            strcpy(reportName, "max");
            data = stats->max[data_index];
            break;

        case REPORT_MEAN:
            strcpy(reportName, "mean");
            data = stats->mean[data_index];
            break;

        case REPORT_STDEV:
            strcpy(reportName, "stdev");
            data = stats->stdev[data_index];
            break;

        case REPORT_SKEWNESS:
            strcpy(reportName, "skewness");
            data = stats->skewness[data_index];
            break;

        case REPORT_KURTOSIS:
            strcpy(reportName, "kurtosis");
            data = stats->kurtosis[data_index];
            break;

        case REPORT_MEAN_SQ_RSP:
            strcpy(reportName, "meanSqRsp");
            data = stats->mean_sq_resp[data_index];
            break;

        case REPORT_ADJ_CORRELATION:
            strcpy(reportName, "adjCorrelation");
            data = stats->correlation[data_index];
            break;

        case REPORT_NUM_VALID_FRAMES:
            strcpy(reportName, "numValidValues");
            data = (float *)stats->num_values[data_index];
            break;

        default:
            IAS_LOG_ERROR("Implemention error.  Unknown report type: %d",
                reportNum);
            return NULL;
    }

    if (isPanBand)
    {
        nChars = snprintf(path, PATH_MAX, "%s/band%02d-%s-%s-frames-%d.dat",
            REPORT_DIR, bandNum, ((lineSet == 0) ? "even" : "odd"), reportName,
            framesPerSegment);
    }
    else
    {
        nChars = snprintf(path, PATH_MAX, "%s/band%02d-%s-frames-%d.dat",
            REPORT_DIR, bandNum, reportName, framesPerSegment);
    }
    if ((nChars < 0) || (nChars >= PATH_MAX))
    {
        IAS_LOG_ERROR("Report path too small, requires %d characters", nChars);
        return NULL;
    }

    if (data == NULL)
        IAS_LOG_ERROR("Expected data for report %s is NULL", reportName);
    return data;
}


/*------------------------------------------------------------------------------
NAME: report_save

PURPOSE:
    This function will save the user requested reports (e.g. image files).  The
    resulting reports will be binary files that contain floating point values
    where the number of detectors in the band are along the x-axis (e.g. the
    samples in the image) and the number of segments in the shutter collect are
    along the y-axis (e.g. the lines within the image.

RETURNS:
    This function will return a status of SUCCESS if the reports are
    successfully updated, ERROR if an error is encountered.
------------------------------------------------------------------------------*/
int report_save
(
    int reportsRequested[NUM_REPORTS], /* I: Array of TRUE/FALSE values */
    int bandNum,          /* I: Band number to create report for */
    int isPanBand,        /* I: Is this the PAN band? */
    int scaNum,           /* I: SCA number */
    int nScas,            /* I: Number of SCAs in this band */
    int nDetectors,       /* I: Number of detectors per SCA in the band */
    int segmentNum,       /* I: Segment number being written (1-based) */
    int framesPerSegment, /* I: Number of frames per segment */
    int totalSegments,    /* I: Total number of segment to process */
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS * stats
                          /* I: histogram results */
)
{
    /* Declare a static variable that will control the validation of the report
       files directory.  This will only be done the first time this function is
       run.  Subsequent runs will assume the directory exists. */
    static int firstRun = TRUE;

    FILE *  fd;
    char    path[PATH_MAX];
    float * values;
    int     lineSet;
    int     report;
    int     status;
    float *num_valid_values = NULL;

    if (firstRun)
    {
        /* Ensure the report files directory exists.  If it does, ensure it is a
           directory and not a file. */
        if (access(REPORT_DIR, F_OK) == 0)
        {
            struct stat fileStats;

            status = stat(REPORT_DIR, &fileStats);
            if ( status )
            {
                IAS_LOG_ERROR("Getting status of %s", REPORT_DIR );
                return ERROR;
            }
            if (!S_ISDIR(fileStats.st_mode))
            {
                IAS_LOG_ERROR("%s exists but is not a directory", REPORT_DIR);
                return ERROR;
            }
        }
        else if (mkdir(REPORT_DIR, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH))
        {
            IAS_LOG_ERROR("Creating %s directory", REPORT_DIR);
            return ERROR;
        }

        /* Set the static variable so we don't check the directory again */
        firstRun = FALSE;
    }

    for (report = 0; report < NUM_REPORTS; report++)
    {
        if (!reportsRequested[report])
            continue;

        lineSet = 0;
        do
        {
            /* Retrieve the report file name and a pointer to the floating point
               values that should be written to the file. */
            values = util_setup(report, bandNum, isPanBand, lineSet,
                framesPerSegment, stats, path);
            if (values == NULL)
            {
                IAS_LOG_ERROR("Setting up report information");
                return ERROR;
            }

            /* The values for the valid frames are int instead of float */
            if ( report == REPORT_NUM_VALID_FRAMES )
            {
                int det;
                int *int_ptr = (int *)values;
                num_valid_values = malloc( nDetectors * sizeof(float));
                if ( !num_valid_values )
                {
                    IAS_LOG_ERROR("Allocating memory for valid frames buffer");
                    return ERROR;
                }
                for ( det = 0; det < nDetectors; det++ )
                    num_valid_values[det] = (float)int_ptr[det];
                values = num_valid_values;
            }

            /* Open the file for writing.  If this is the first SCA of the first
               segment for the band, we'll open with truncation, otherwise we'll
               open for appending. */
            if ((segmentNum == 1) && (scaNum == 1))
            {
                /* Create an ENVI header file for the image.  The number of
                   lines in the image will be equal to the total number of
                   segments that are processed for the band.  The number of
                   samples in the image is equal to the number of SCA's x number
                   of detectors per SCA */
                status = util_create_envi_header(path, totalSegments,
                    nScas * nDetectors);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Creating ENVI header for image");
                    free( num_valid_values );
                    return ERROR;
                }

                fd = fopen(path, "w");  /* Open for writing & truncate */
            }
            else
                fd = fopen(path, "a");  /* Open with append */
            if (fd == NULL)
            {
                IAS_LOG_ERROR("Opening report file: %s", path);
                free( num_valid_values );
                return ERROR;
            }

            /* Write these values to the file */
            status = fwrite(values, sizeof(float), nDetectors, fd);
            if (status != nDetectors)
            {
                IAS_LOG_ERROR("Writing to %s", path);
                fclose(fd);
                return ERROR;
            }

            if ( report == REPORT_NUM_VALID_FRAMES )
            {
                free( num_valid_values );
                num_valid_values = NULL;
            }

            /* Close the file */
            status = fclose(fd);
            if ( status )
            {
                IAS_LOG_ERROR("Closing file %s", path );
                return ERROR;
            }
        } while (isPanBand && (++lineSet < 2));
    }
    return SUCCESS;
}
