#!/bin/csh
# 
# This script will run the batc_data_segment_stats application on the chosen
# BATC data file, dump those statistics to a file, and then compare these
# values to the benchmark results.

# Set up constants for this test using environment variables.  By using
# environment variables, these values will be available to the other scripts of
# this test.
set app          = "batc_data_segment_stats"
set colRwoid     = "RADIOMETRIC_WO_COMMON_ID"
set colWoid      = "WORK_ORDER_ID"
set tableR       = "RADIOMETRIC_WORK_ORDER_COMMON"
set tableS       = "SEGMENT_STATISTICS"
set tmpResults   = "tmpResults.txt"
set tmpSqlScript = tmpSqlScript.sql

# Which tests should be run.  Due to the length of running the long shutter
# collect, we won't run that all the time.  The user can specify which tests (1,
# 2 or both) they want to run.
if ($#argv == 0) then
    set tests = (0001 0002)
else
    set tests = ""
    while ($#argv > 0)
	set tmp = `echo $1 | awk '{printf "%04d",$1}'`
	if ("${tmp}" == "") then
	    echo "usage: runTest.csh [<testNum>]"
	    echo ""
	    echo "    Run one of the tests for this application where <testNum>"
	    echo "    is a number of 1 or 2."
	    exit 1
	endif
	set tests = (${tests} ${tmp})
	shift
    end
endif

# Ensure the require environment is set
if (! -x ../${app}) then
    echo "Permission denied: ../${app}"
    exit 1;
endif
if (! $?IAS_DB_L8_OLITIRS) then
    echo "Environment variable IAS_DB_L8_OLITIRS not defined"
    exit 1;
endif

foreach num (${tests})
    echo "Tesing ${num} test case"

    # Ensure the template ODL file exists for the specified test
    if (! -e I${num}.txt) then
	echo "ERROR: test ${num} does not exist"
	exit 1
    endif

    # Create an OMF file using the UTxxxx.txt template
    set odl = "I${num}.odl"
    cat I${num}.txt | \
	sed s%__IAS_UNIT_TEST_DATA_DIR__%$IAS_UNIT_TEST_DATA_DIR% > ${odl}

    # Run ${app} on the shutter collect
    echo "Running: ../${app} ${odl}"
    ../${app} ${odl} >& /dev/null
    if ($status != 0) then
	echo "ERROR running ${app}"
	exit 1
    endif

    # Retrieve the RADIOMETRIC_WO_COMMON_ID from the
    # RADIOMETRIC_WORK_ORDER_COMMON table that was just created by running the
    # application.
    set sql = "SELECT ${colRwoid} FROM ${tableR} WHERE ${colWoid}='I"${num}"';"
    set rwoid = `echo ${sql} | sqlplus ${IAS_DB_L8_OLITIRS} | \
	    grep -A 2 ${colRwoid} | tail --lines=1`
    if ("${rwoid}" == "") then
	echo "ERROR retrieving ${colRwoid} from ${tableR}"
	exit 1
    endif

    # Extract the resulting statistics from the database.  The output file name
    # is actually specified in the sqlScript.sql file.  We have a variable set
    # above to reference the same name: $tmpResults.
    echo "Extracting statistics from database"
    cat sqlScript.sql | sed s%__RWOID__%${rwoid}% > ${tmpSqlScript}
    sqlplus ${IAS_DB_L8_OLITIRS} @${tmpSqlScript} >& /dev/null
    if (! -e ${tmpResults}) then
	echo "ERROR retrieving statistics from the database"
	exit 1
    endif

    # Compare the results against the benchmark results in
    # ${IAS_UNIT_TEST_DATA_DIR}/ias/rps/tools/batc_data_segment_stats
    echo "Comparing current results with benchmark statistics"
    set dir = ${IAS_UNIT_TEST_DATA_DIR}/ias/rps/tools/${app}
    diff -q ${dir}/benchmarkResults-${num}.txt ${tmpResults}
    if ($status != 0) then
	echo "ERROR differences found"
	exit 1
    endif
    echo "SUCCESS: statistics match benchmark stats for test ${num}"

    # Compare one of the reports
    if ( $num == 0001 || $num == 1 ) then
        set report_file = band02-mean-frames-640.dat
        diff -q ${dir}/$report_file ./report_files/$report_file
        if ( $status != 0 ) then
            echo "ERROR differences found in $report_file"
            exit 1
        endif
        set report_file = band02-mean-frames-640.dat.hdr
        diff -q ${dir}/$report_file ./report_files/$report_file
        if ( $status != 0 ) then
            echo "ERROR differences found in $report_file"
            exit 1
        endif
    echo "SUCCESS: report file match benchmark file for test ${num}"
    endif

    if ( $num == 0002 || $num == 2 ) then
        set report_file = band08-even-max-frames-6400.dat
        diff -q ${dir}/$report_file ./report_files/$report_file
        if ( $status != 0 ) then
            echo "ERROR differences found in $report_file"
            exit 1
        endif
        set report_file = band08-even-max-frames-6400.dat.hdr
        diff -q ${dir}/$report_file ./report_files/$report_file
        if ( $status != 0 ) then
            echo "ERROR differences found in $report_file"
            exit 1
        endif
    echo "SUCCESS: report file match benchmark file for test ${num}"
    endif

    # Remove the results from the database
    if ("${rwoid}" != "") then
	echo "Removing test results from the database"
	set sql = "DELETE FROM ${tableS} WHERE ${colRwoid}=${rwoid};"
	echo ${sql} | sqlplus ${IAS_DB_L8_OLITIRS} >& /dev/null
	set sql = "DELETE FROM ${tableR} WHERE ${colRwoid}=${rwoid};"
	echo ${sql} | sqlplus ${IAS_DB_L8_OLITIRS} >& /dev/null
    endif

    # Remove the temporary OMF file that was generated from the template
    # and the report directory and contents
    rm -f ${tmpResults} ${tmpSqlScript}
    rm -rf ./report_files

end
exit 0
