#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include "ias_const.h"
#include "ias_logging.h"
#include "cgreen.h"
#include "../batc_file_utilities.h"

#include "b4_expected_results.h"
#include "b8_expected_results.h"

static char *test_data_path = NULL;

#define NDETS 494

Ensure test_batc_read(void)
{
    char datafile[PATH_MAX];
    char filename[] =
        "ias/rps/tools/batc_data_segment_stats/20100718170419ESP00118.dat";
    int status;
    int band_number = 4;
    int sca_index = 5;
    int start_line = 15;
    int number_lines = 2;
    size_t buf_size = NDETS * 2 * number_lines * 2 * sizeof(short);
    short *buffer;
    double acquisition_time;
    unsigned int number_of_frames;
    int truncation_setting;
    int i;
    int bad_count;

    /* Point to the data file */
    sprintf( datafile, "%s/%s", test_data_path, filename );
    /* Initialize the BATC read routine */
    status = initialize_batc_read();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing batc read");
        return;
    }

    /* Get information from the frame header */
    status = read_frame_header( datafile, &number_of_frames,
        &truncation_setting );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error reading frame header");
        cleanup_batc_read();
        return;
    }

    /* Ensure the number of frames and truncation settings are what we
       expect */
    assert_equal( 127, number_of_frames );
    assert_equal( 0, truncation_setting );

    /* Allocate a buffer.  This will be allocated large enough to handle the
       pan band test as well. */
    buffer = malloc( buf_size );
    if ( buffer == NULL )
    {
        assert_true_with_message( 0, "Error allocating memory");
        cleanup_batc_read();
        return;
    }

    /* Read number of lines for band number */
    status = read_batc_sca( datafile, band_number, sca_index, start_line,
        number_lines, buffer, &acquisition_time );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error returned from read_batc_sca "
            "checks for band %d skipped", band_number );
    }
    else
    {
        /* Check the results */
        assert_double_equal( 332784262.54105, acquisition_time );
        bad_count = 0;
        for ( i = 0; i < NDETS * number_lines; i++ )
        {
            if ( b4_expected_results[i] != buffer[i] )
                bad_count++;
        }
        assert_equal_with_message( 0, bad_count,
            "Error band %d has %d differences", band_number, bad_count );
    }

    /* If the number of frames is less that what we'll be looking for in
       the band 8 test fail and exit now */
    band_number = 8;
    sca_index = 9;
    start_line = 120;
    if ( number_of_frames < start_line + 2 )
    {
        assert_true_with_message( 0, "Not enough frames to complete band "
            "8 test.  Need %d and have %d.", start_line + 2,
            number_of_frames );
        cleanup_batc_read();
        free( buffer );
        return;
    }

    /* Read number of lines for band number */
    status = read_batc_sca( datafile, band_number, sca_index, start_line,
        number_lines, buffer, &acquisition_time );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error returned from read_batc_sca "
            "checks for band %d skipped", band_number );
    }
    else
    {
        /* Check the results */
        assert_double_equal( 332784262.985830, acquisition_time );
        bad_count = 0;
        for ( i = 0; i < NDETS * 2 * number_lines * 2; i++ )
        {
            if ( b8_expected_results[i] != buffer[i] )
                bad_count++;
        }
        assert_equal_with_message( 0, bad_count,
            "Error band %d has %d differences", band_number, bad_count );
    }

    cleanup_batc_read();
    free( buffer );
}

int main(int argc, char *argv[])
{
    TestSuite *suite = create_test_suite();
    add_test( suite, test_batc_read );

    /* Get the test data directory */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if ( test_data_path == NULL )
    {
        printf("Cannot find test data directory path\n");
        return 1;
    }

    if ( argc > 1 )
        return run_single_test( suite, argv[1], create_text_reporter());
    else
        return run_test_suite( suite, create_text_reporter());
}
