/*----------------------------------------------------------------------------
NAME:    create_3q_rlut.c

PURPOSE: Create the Attributes and 3Q RLUT tables.  This was written because
         using HDF tools to import won't work on the 3Q tables.  It's
         designed primarily for Cal/Val use to allow them to create and test
         new RLUT parameters.

ASSUMPTIONS:
    1) Comma delimited ASCII input file of the following form:
        - one line per detector
        - first four fields are: integer band, sca, detector and running
          number of detectors
        - fields 4 and 5: floating point low and high cutoff points
        - fields 6, 7 and 8: floating point low coefficients - C0, C1, C2
        - fields 9, 10 and 11: floating point mid coefficients - C0, C1, C2
        - fields 12, 13 and 14: floating point high coefficients - C0, C1, C2

    2) File should be numerically sorted by band, sca and detector

    3) All bands are present

USAGE: create_3q_rlut <input file> [--pat]

    --pat will specify the optional band list defined in static int band_list
          and is for testing purposes only using the original file from Pat S.

--------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ias_const.h"
#include "ias_rlut.h"
#include "ias_satellite_attributes.h"

static char evaluation_rlut[] = "evaluation_rlut.h5";
static int satellite_number = 8;
static int band_list[] = {6, 7, 2, 1, 9, 3, 5, 8, 4};
static int tirs_band_list[4] = {15,16,17,18};

int main( int argc, char *argv[] )
{
    const IAS_SATELLITE_ATTRIBUTES *sat_attributes;
    IAS_SATELLITE_ID satellite_id;
    FILE *inf;
    IAS_RLUT_IO *rlut;
    int band_indices[IAS_MAX_TOTAL_BANDS + 1];
    int i;
    int max_dets;
    IAS_RLUT_FILE_ATTRIBUTES file_attr;
    IAS_RLUT_LINEARIZATION_PARAMS *parameters;
    int status;
    int alternate = 0;
    int tirs_ndets;

    /* Check for command line parameter */
    if ( argc == 3 )
    {
        if ( strcmp( argv[2], "--pat") == 0 )
            alternate = 1;
    }
    else if ( argc != 2 )
    {
        printf("\n\n\tUsage: %s <parameter file name>\n\n", argv[0] );
        return 1;
    }

    inf = fopen( argv[1], "r" );
    if ( ! inf )
    {
        printf("ERROR: unable to open file: %s\n", argv[1] );
        return 1;
    }

    /* Open the evaluation RLUT */
    rlut = ias_rlut_open_file( evaluation_rlut, IAS_WRITE );
    if ( ! rlut )
    {
        printf("ERROR: unable to open file: %s\n", evaluation_rlut );
        fclose( inf );
        return 1;
    }

    /* Initialize the satellite attributes */
    satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(
        satellite_number );
    if ( ias_sat_attr_initialize( satellite_id ))
    {
        
        printf("ERROR: initializing satellite attributes library\n");
        fclose( inf );
        ias_rlut_close_file( rlut );
        return 1;
    }
    sat_attributes = ias_sat_attr_get_attributes();
    max_dets = 0;
    for ( i = 0; i < sat_attributes->total_bands; i++ )
    {
        int band_number = sat_attributes->band_attributes[i].band_number;
        band_indices[band_number] = i;
        if ( sat_attributes->band_attributes[i].detectors_per_sca > max_dets )
            max_dets = sat_attributes->band_attributes[i].detectors_per_sca;
    }

    /* Set the attributes -- at this time only dummy attributes are added */
    strcpy( file_attr.file_source, argv[1] );
    strcpy( file_attr.effective_date_begin, "no date" );
    strcpy( file_attr.effective_date_end, "no date" );
    strcpy( file_attr.status, "UNTESTED" );
    strcpy( file_attr.baseline_date, "no date" );
    strcpy( file_attr.description, "Created by create_3q_tables using file "
        "listed in file_source" );
    file_attr.version = 0;
    status = ias_rlut_write_file_attributes( rlut, &file_attr );
    if ( status != SUCCESS )
    {
        printf("ERROR: writing file attributes\n");
        fclose( inf );
        ias_rlut_close_file( rlut );
        return 1;
    }

    /* Allocate the memory */
    parameters = malloc( max_dets * sizeof(IAS_RLUT_LINEARIZATION_PARAMS));
    if ( ! parameters )
    {
        printf("ERROR: allocating parameter memory\n");
        fclose( inf );
        ias_rlut_close_file( rlut );
        return 1;
    }

    /* Loop through the bands */
    int nbands = sat_attributes->bands;
    if ( alternate )
        nbands = 9;
    for ( i = 0; i < nbands; i++ )
    {
        int band_index;
        int band_number;
        if ( alternate )
        {
            band_number = band_list[i];
            band_index = band_indices[band_number];
        }
        else
        {
            band_index = i;
            band_number = sat_attributes->band_attributes[band_index].
                normal_band_number;
        }
        int nscas = sat_attributes->band_attributes[band_index].scas;
        int ndets = sat_attributes->band_attributes[band_index].
            detectors_per_sca;
        int sca_index;


        printf("Processing band %d(%d), ndets %d\n", band_number,
            band_index, ndets );
        for ( sca_index = 0; sca_index < nscas; sca_index++ )
        {
            int detector;
            int b, s, d;
            float low, high, lc0, lc1, lc2, mc0, mc1, mc2, hc0, hc1, hc2;
            for ( detector = 0; detector < ndets; detector++ )
            {
                int n = fscanf( inf, "%d,%d,%d,%*d,%f,%f,%f,%f,%f,%f,%f,%f,"
                    "%f,%f,%f", &b, &s, &d, &low, &high, &lc0, &lc1,
                    &lc2, &mc0, &mc1, &mc2, &hc0, &hc1, &hc2 );
                if ( n != 14 )
                {
                    printf("ERROR: reading from input file, only read %d of "
                        "14\n", n );
                    fclose( inf );
                    ias_rlut_close_file( rlut );
                    free( parameters );
                    return 1;
                }
                if ( b != band_number || s != sca_index + 1
                    || d != detector + 1 )
                {
                    printf("ERROR: in input file order: band %d, %d; "
                        "SCA %d, %d; det %d, %d\n", band_number, b,
                        sca_index + 1, s, detector+ 1 , d );
                    fclose( inf );
                    ias_rlut_close_file( rlut );
                    free( parameters );
                    return 1;
                }

                parameters[detector].cutoff_threshold_low = low;
                parameters[detector].cutoff_threshold_high = high;
                parameters[detector].remap_coeff0_low = lc0;
                parameters[detector].remap_coeff1_low = lc1;
                parameters[detector].remap_coeff2_low = lc2;
                parameters[detector].remap_coeff0_mid = mc0;
                parameters[detector].remap_coeff1_mid = mc1;
                parameters[detector].remap_coeff2_mid = mc2;
                parameters[detector].remap_coeff0_high = hc0;
                parameters[detector].remap_coeff1_high = hc1;
                parameters[detector].remap_coeff2_high = hc2;
            }

            /* Write this SCAs parameters */
            status = ias_rlut_write_linearization_params( rlut, parameters,
                band_number, sca_index + 1, ndets );
            if ( status != SUCCESS )
            {
                printf("ERROR: writing parameters for band %d SCA %d\n",
                    band_number, sca_index + 1 );
                fclose( inf );
                ias_rlut_close_file( rlut );
                free( parameters );
                return 1;
            }
        }
    }

    /* For some reason Ingest checks to make sure the TIRS redundant bands
       are present as well as the blind so add unity for these. */
    i = tirs_band_list[0];
    tirs_ndets = sat_attributes->band_attributes[band_indices[i]].
                detectors_per_sca;
    for ( i = 0; i < tirs_ndets; i++ )
    {
        parameters[i].cutoff_threshold_low = 1000;
        parameters[i].cutoff_threshold_high = 3000;
        parameters[i].remap_coeff0_low = 0.0;
        parameters[i].remap_coeff1_low = 1.0;
        parameters[i].remap_coeff2_low = 0.0;
        parameters[i].remap_coeff0_mid = 0.0;
        parameters[i].remap_coeff1_mid = 1.0;
        parameters[i].remap_coeff2_mid = 0.0;
        parameters[i].remap_coeff0_high = 0.0;
        parameters[i].remap_coeff1_high = 1.0;
        parameters[i].remap_coeff2_high = 0.0;
    }
    for ( i = 0; i < 4; i++ )
    {
        int band_number = tirs_band_list[i];
        int sca_index;
        for ( sca_index = 0; sca_index < 3; sca_index++ )
        {
            status = ias_rlut_write_linearization_params( rlut, parameters,
                band_number, sca_index + 1, tirs_ndets );
            if ( status != SUCCESS )
            {
                printf("ERROR: writing parameters for band %d SCA %d\n",
                    band_number, sca_index + 1 );
                fclose( inf );
                ias_rlut_close_file( rlut );
                free( parameters );
                return 1;
            }
        }
    }

    /* Clean up */
    status = fclose( inf );
    if ( status < 0 )
        printf("ERROR closing input file\n");
    status = ias_rlut_close_file( rlut );
    if ( status != SUCCESS )
        printf("ERROR: Closing the rlut\n");
    free( parameters );

    return 0;
}  /* END main */
