#!/usr/bin/perl
use strict;
use warnings;

#############################################################################
#  Name : check_orphaned_wo_dirs.pl
#
#  Description: Occasionally CalVal users will create directories in a work 
#   order directory.  Becauses the new directories are not owned by the IAS 
#   user, DRM cannot remove the work order directory.  This script will output
#   all work order directories in IAS_WO that have been marked as removed in
#   the database.
#
#  Parameters: None. Configuration is pulled from IAS environment variables.
#
#############################################################################

use DBI;

use ias_const;
use ias_db;

use constant DRM_DELETED  => 1;

if (not exists $ENV{'IAS_WO'})
{
    print "ERROR: IAS_WO environment variable is not set\n";
    exit FAILURE;
}
my $ias_wo = $ENV{'IAS_WO'};

if (! -d $ias_wo)
{
    print "ERROR: IAS_WO directory does not exist\n";
    exit FAILURE;
}

my @wo_dirs;

# Find all directories in IAS_WO
opendir my ($DH), $ias_wo;
while (my $item = readdir($DH))
{
    if ($item ne "." && $item ne "..")
    {
        if ( -d "$ias_wo/$item" )
        {
            push(@wo_dirs, $item);
        }
    }
}
closedir($DH);

my ($username, $password) = ias_db_GetDBConnectionInfo();
my $dbh = ias_db_ConnectToDB($username, $password);
if (!$dbh)
{
    print "ERROR: Unable to connect to the database as $username:\n "
        . "     $ias_db_error\n";
    exit FAILURE;
}

my $sql = "SELECT deleted FROM work_orders WHERE work_order_id = ?";
my $sth = $dbh->prepare($sql);
if (!$sth)
{
    print "ERROR: Unable to prepare database query: $DBI::errstr\n";
    $dbh->disconnect();
    exit FAILURE;
}

my @orphaned_wos;
foreach my $wodir (@wo_dirs)
{
    $sth->bind_param(1, $wodir);
    if (not defined $sth->execute())
    {
        print "ERROR: Unable to execute database query: $DBI::errstr\n";
        $sth->finish();
        $dbh->disconnect();
        exit FAILURE;
    }

    my $row = $sth->fetchrow_hashref();
    if (defined $row)
    {
        if ($row->{"DELETED"} == DRM_DELETED)
        {
            push(@orphaned_wos, $wodir);
        }
    }
    else
    {
        print "WARN: Work order ID '$wodir' doesn't exist in the database.\n";
    }
}

$sth->finish();
$dbh->disconnect();

my $orphaned_found = FALSE;
my $output_message;
if (@orphaned_wos > 0)
{
    $output_message = "Orphaned directories found:\n";
    foreach my $wo_id (@orphaned_wos)
    {
        # If the directory still exists. This check will catch cases when
        # DRM cleans up a directory while the script is running. 
        if (-d "$ias_wo/$wo_id")
        {
            $output_message .= "    $wo_id\n";
            $orphaned_found = TRUE;
        }
    }
}

if (!$orphaned_found)
{
    $output_message = "No orphaned directories found\n";
}

print "----\n";
print "$output_message";

exit SUCCESS;
