#!/usr/bin/perl -w
#------------------------------------------------------------------------------
# NAME: test_tram_interface.pl
#
# PURPOSE: Test interface between DCO and TRAM. This includes loading orders
#          from TRAM to the work order database and synchronizing statuses
#          between the work order database and TRAM.
#
#          Since DCO could potentially be used for manual orders not
#          originating with the TRAM, it is worth noting here that the
#          processing done by the DCO for a non-TRAM order is a sub-set of
#          what is done for a TRAM order, so there is no significant need
#          to make a separate regression test flow for these types of orders.
#
# PARAMETERS: None
#
# OUTPUT: SUCCESS - Test completed successfully
#         FAILURE
#------------------------------------------------------------------------------

use strict;
use DBI;
use POSIX ":sys_wait_h";

use ias_const;
use ias_wo_const;
use ias_processing_system;

## Database configuration
my $tramDbh;
my $woDbh;

my $USER = $ENV{USER};

# Wait time when waiting for DCO to run
use constant WAIT_TIME => 30;
use constant MAX_DCO_CHECKS => 5;
my $dcoWaitCounter;
my $dcoPid = undef;

my $prodCode = 'L095';
my $badProdCode = 'L119';

# Test unit information.
# NOTE: must contain at least two units for completion status testing.
#  first unit will test 'C'ompleted, second unit will test 'A'borted
# ProdCode will be set later, after createProdCodeModule
my @units = (
    {OrderNbr => "0121102120002", UnitNbr => 1, ProdCode => "$prodCode",
     LandsatSceneId => "LO80010022011023EDC01", Priority => 1},
    {OrderNbr => "0121102120001", UnitNbr => 1, ProdCode => "$prodCode",
     LandsatSceneId => "LO80010022011023EDC01", Priority => 3},
    {OrderNbr => "0121102120005", UnitNbr => 1, ProdCode => "$prodCode",
     LandsatSceneId => "LO80010022011023EDC01", Priority => 7},
    {OrderNbr => "0121102120003", UnitNbr => 1, ProdCode => "$prodCode",
     LandsatSceneId => "LO80010022011023EDC01", Priority => 9},
    {OrderNbr => "0121102120004", UnitNbr => 1, ProdCode => "$prodCode",
     LandsatSceneId => "LO80010022011023EDC01", Priority => 9}
);
# Units that should be rejected due to an invalid product code
my @rejectedUnits = (
    {OrderNbr => "0121102120006", UnitNbr => 1, ProdCode => "$badProdCode",
     LandsatSceneId => "LO80010022011023EDC01", Priority => 6}
);

# Combined units
my @combinedUnits = (@units, @rejectedUnits);

## Sub Declarations
sub initDb();
sub closeDb();
sub startDco();
sub stopDco($);
sub testCleanup();
sub createProdCodeModule();
sub createTramUnit($$$$$);
sub inIas($$);
sub checkOrder();
sub checkPriority($$);
sub getTramStatus($$);
sub getWoStatus($$);
sub getAnomalyStatus($$);
sub updateWoStatus($$$);
sub createAnomalies($$$$);
sub cleanupTramOrder($);
sub cleanupWorkOrder($);
sub cleanupProdCodeModule();
sub cleanupAnomalies($$$$);
sub checkInstanceEnabledFlag($);

##############################################################################
## Main
##############################################################################

# If database tests are disabled, skip this test
if (!defined $ENV{ENABLE_DB_TESTS})
{
    print "INFO: database tests disabled. Skipping test_tram_interface.pl\n";
    exit SUCCESS;
}

my $processing_instance = $ENV{"IAS_INSTANCE_NAME"};
if (!defined $processing_instance)
{
    print "ERROR: No processing instance defined";
    exit FAILURE;
}

if (ias_init_system() != SUCCESS)
{
    print "ERROR: could not initialize ias system: $ias_error_msg\n";
    exit FAILURE;
}

# Open database connections
if (initDb() == ERROR)
{
    print "ERROR: could not connect to database\n";
    exit FAILURE;
}

if (!checkInstanceEnabledFlag($processing_instance))
{
    testCleanup();
    print "ERROR: $processing_instance is not enabled. Enable it to complete " .
        "testing.\n";
    exit FAILURE;
}

# Clean up TRAM tables
for my $unit (@combinedUnits)
{
    if (cleanupTramOrder($unit->{OrderNbr}) != SUCCESS)
    {
        testCleanup();
        print "ERROR: could not clean up TRAM database\n";
        exit FAILURE;
    }
}

# Clean up work order tables
for my $unit (@units)
{
    if (cleanupWorkOrder($unit->{OrderNbr}) != SUCCESS)
    {
        testCleanup();
        print "ERROR: could not clean up WO database\n";
        exit FAILURE;
    }
    if (cleanupProdCodeModule() != SUCCESS)
    {
        testCleanup();
        print "ERROR: could not clean up Product Code in database\n";
        exit FAILURE;
    }
    if ($isLPGS)
    {
        if (cleanupAnomalies($units[0]->{OrderNbr}, $units[0]->{UnitNbr},
                    $units[1]->{OrderNbr}, $units[1]->{UnitNbr}) != SUCCESS)
        {
            testCleanup();
            print "ERROR: could not clean up Anomalies in database\n";
            exit FAILURE;
        }
    }
}

# Create a PRODUCT_CODE_MODULES table entry for testing
if (createProdCodeModule() != SUCCESS)
{
    testCleanup();
    print "ERROR: could not create product code module in WO database\n";
    exit FAILURE;
}

# Create test TRAM units
for my $unit (@combinedUnits)
{
    if(createTramUnit($unit->{OrderNbr},$unit->{UnitNbr},
        $unit->{LandsatSceneId},$unit->{ProdCode},
        $unit->{Priority}) != SUCCESS)
    {
        testCleanup();
        print "ERROR: could not create test orders in TRAM database\n";
        exit FAILURE;
    }
}

# Start DCO
$dcoPid = startDco();

waitpid($dcoPid, WNOHANG);

# Give DCO time to run
sleep(3);

# Check that all the orders in the TRAM queue have been picked up by DCO
# and a work order created for each
my $woInIas = FALSE;
$dcoWaitCounter = 0;
while ($woInIas == FALSE)
{
    my $iasCnt = 0;
    for my $unit (@units)
    {
        if (inIas($unit->{OrderNbr}, $unit->{UnitNbr}) == SUCCESS) 
        {
            $iasCnt++;
        }    
    }

    if (scalar(@units) == $iasCnt)
    {
        $woInIas = TRUE;
        last;
    }

    if ($dcoWaitCounter > MAX_DCO_CHECKS)
    {
        $woInIas = FALSE;
        last;
    }

    print "INFO: waiting for DCO\n";
    sleep(WAIT_TIME);

    $dcoWaitCounter++;
}

if ($woInIas == TRUE)
{
    print "INFO: work orders successfully entered into IAS\n";
}
else
{
    testCleanup();
    print "ERROR: work order not in IAS\n";
    exit FAILURE;
}

# Verify the orders were created in the same order as the TRAM orders.
if (checkOrder() == TRUE)
{
    print "INFO: work orders successfully created in TRAM order sequence\n";
}
else
{
    testCleanup();
    print "ERROR: Orders were not created in TRAM order of creation.\n";
    exit FAILURE;
}

# Verify the TRAM unit priorities carried over to the work order tables
my $priorityMatch = TRUE;
for my $unit (@units)
{
    if (checkPriority($unit->{OrderNbr}, $unit->{Priority}) != TRUE)
    {
        print "ERROR: TRAM unit priority does not match WO unit priority ";
        print "Order Number: ". $unit->{OrderNbr} . "\n";
        $priorityMatch = FALSE;
    }
}

if ($priorityMatch)
{
    print "INFO: TRAM unit priorities successfully brought over to WO table\n";
}
else
{
    testCleanup();
    print "ERROR: TRAM unit priorities not brought over properly\n";
    exit FAILURE;
}

# Verify TRAM units statused to 'I' and 'R'
my $tramStatusInProcess = FALSE;
my $tramStatusRejected = FALSE;
$dcoWaitCounter = 0;
while ($tramStatusInProcess == FALSE)
{
    my $inProcessCnt = 0;
    my $rejectedCnt = 0;

    for my $unit (@units)
    {
        my ($status) = getTramStatus($unit->{OrderNbr}, $unit->{UnitNbr});
        if ($status eq "I")
        {
            $inProcessCnt++;
        }
    }

    for my $unit (@rejectedUnits)
    {
        my ($status) = getTramStatus($unit->{OrderNbr}, $unit->{UnitNbr});
        if ($status eq "R")
        {
            $rejectedCnt++;
        }
    }

    if (scalar(@units) == $inProcessCnt)
    {
        $tramStatusInProcess = TRUE;
    }

    if (scalar(@rejectedUnits) == $rejectedCnt)
    {
        $tramStatusRejected = TRUE;
    }

    last if ($tramStatusInProcess && $tramStatusRejected);

    if ($dcoWaitCounter > MAX_DCO_CHECKS)
    {
        $tramStatusInProcess = FALSE;
        last;
    }

    print "INFO: waiting for DCO\n";
    sleep(WAIT_TIME);

    $dcoWaitCounter++;
}

if ($tramStatusInProcess == TRUE)
{
    print "INFO: units successfully statused to 'I' in TRAM\n";
}
else
{   
    testCleanup();
    print "ERROR: units not statused to (I)n process in TRAM\n";
    exit FAILURE;
}

if ($tramStatusRejected == TRUE)
{
    print "INFO: rejected units successfully statused to 'R' in TRAM\n";
}
else
{   
    testCleanup();
    print "ERROR: rejected units not statused to rejected in TRAM\n";
    exit FAILURE;
}

# Kill currently running DCO. Will be restarted later to avoid wait time.
stopDco($dcoPid);

# Update 1st work order status to DB_WO_COMPLETE_WO ('CW')
if (updateWoStatus($units[0]->{OrderNbr}, $units[0]->{UnitNbr},
            DB_WO_COMPLETE_WO) != SUCCESS)
{
    testCleanup();
    print "ERROR: error statusing first work order to DB_WO_COMPLETE_WO\n";
    exit FAILURE;
}
else
{
    print "INFO: first work order statused to DB_WO_COMPLETE_WO\n";
}

# Update 2nd work order status to DB_WO_ABORT_WO ('AW')
if (updateWoStatus($units[1]->{OrderNbr}, $units[1]->{UnitNbr}, DB_WO_ABORT_WO)
        != SUCCESS)
{
    testCleanup();
    print "ERROR: error statusing second work order to DB_WO_ABORT_WO\n";
    exit FAILURE;
}
else
{
    print "INFO: second work order statused to DB_WO_ABORT_WO\n";
}

# Create anomaly entries for the work orders moved to complete and abort
# in order to test DCO aborting anomaly work orders associated with a work
# order being completed or aborted. Only do when running in "LPGS mode".
if ($isLPGS)
{
    if (createAnomalies($units[0]->{OrderNbr}, $units[0]->{UnitNbr},
                $units[1]->{OrderNbr}, $units[1]->{UnitNbr}) != SUCCESS)
    {
        testCleanup();
        print "ERROR: error creating mock anomalies for completed and " .
            "aborted work orders\n";
        exit FAILURE;
    }
    else
    {
        print "INFO: anomalies mocked-up for completed and aborted work " .
            "orders\n";
    }
}

# Restart DCO 
$dcoPid = startDco();

# Give DCO time to update statuses
sleep(3);

# Verify units statused to 'C' and 'R' in TRAM
$dcoWaitCounter = 0;
my $tramStatusComplete = FALSE;
while ($tramStatusComplete == FALSE)
{
    if (getTramStatus($units[0]->{OrderNbr}, $units[0]->{UnitNbr}) eq 'C')
    {
        $tramStatusComplete = TRUE;
        last;
    }
    
    if ($dcoWaitCounter > MAX_DCO_CHECKS)
    {
        $tramStatusComplete = FALSE;
        last;
    }

    print "INFO: waiting for DCO...\n";
    sleep(WAIT_TIME);

    $dcoWaitCounter++;
}

if ($tramStatusComplete == TRUE)
{
    print "INFO: TRAM unit successfully statused to 'C'\n";
}
else
{
    testCleanup();
    print "ERROR: failed to status TRAM unit to 'C'\n";
    exit FAILURE;
}

$dcoWaitCounter = 0;
$tramStatusRejected = FALSE;
while ($tramStatusRejected == FALSE)
{
    if (getTramStatus($units[1]->{OrderNbr}, $units[1]->{UnitNbr}) eq 'R')
    {
        $tramStatusRejected = TRUE;
        last;
    }
    
    if ($dcoWaitCounter > MAX_DCO_CHECKS)
    {
        $tramStatusRejected = FALSE;
        last;
    }

    print "INFO: waiting for DCO...\n";
    sleep(WAIT_TIME);
    
    $dcoWaitCounter++;
}

if ($tramStatusRejected == TRUE)
{
    print "INFO: TRAM unit successfully statused to 'R'\n";
}
else
{
    testCleanup();
    print "ERROR: failed to status TRAM unit to 'R'\n";
    exit FAILURE;
}

# Verify units statused to DB_WO_RELEASE_COMPLETE and DB_WO_RELEASE_ABORT
# in WORK_ORDERS
$dcoWaitCounter = 0;
my $woStatusComplete = FALSE;
while ($woStatusComplete == FALSE)
{
    if (getWoStatus($units[0]->{OrderNbr}, $units[0]->{UnitNbr})
            eq DB_WO_RELEASE_COMPLETE)
    {
        $woStatusComplete = TRUE;
        last;
    }
    
    if ($dcoWaitCounter > MAX_DCO_CHECKS)
    {
        last;
    }

    print "INFO: waiting for DCO...\n";
    sleep(WAIT_TIME);
    
    $dcoWaitCounter++;
}
if ($woStatusComplete == TRUE)
{
    print "INFO: Work Order successfully statused to DB_WO_RELEASE_COMPLETE\n";
}
else
{
    testCleanup();
    print "ERROR: failed to status Work Order to DB_WO_RELEASE_COMPLETE\n";
    exit FAILURE;
}

$dcoWaitCounter = 0;
my $woStatusRejected = FALSE;
while ($woStatusRejected == FALSE)
{
    if (getWoStatus($units[1]->{OrderNbr}, $units[1]->{UnitNbr})
            eq DB_WO_RELEASE_ABORT)
    {
        $woStatusRejected = TRUE;
        last;
    }
    
    if ($dcoWaitCounter > MAX_DCO_CHECKS)
    {
        last;
    }

    print "INFO: waiting for DCO...\n";
    sleep(WAIT_TIME);
    
    $dcoWaitCounter++;
}
if ($woStatusRejected == TRUE)
{
    print "INFO: Work Order successfully statused to DB_WO_RELEASE_ABORT\n";
}
else
{
    testCleanup();
    print "ERROR: failed to status Work Order to DB_WO_RELEASE_ABORT\n";
    exit FAILURE;
}

# Verify associated anomaly work orders were statused to DB_WO_RELEASE_ABORT
# when running in "LPGS mode"
if ($isLPGS)
{
    $dcoWaitCounter = 0;
    my $anomalyWosCanceled = FALSE;
    while ($anomalyWosCanceled == FALSE)
    {
        if (getAnomalyStatus($units[0]->{OrderNbr}, $units[0]->{UnitNbr})
                eq DB_WO_RELEASE_ABORT)
        {
            $anomalyWosCanceled = TRUE;
            last;
        }

        if ($dcoWaitCounter > MAX_DCO_CHECKS)
        {
            last;
        }

        print "INFO: waiting for DCO...\n";
        sleep(WAIT_TIME);

        $dcoWaitCounter++;
    }
    if ($anomalyWosCanceled == TRUE)
    {
        print "INFO: First anomaly successfully statused to "
              . "DB_WO_RELEASE_ABORT\n";
    }
    else
    {
        testCleanup();
        print "ERROR: failed to status first anomaly to DB_WO_RELEASE_ABORT\n";
        exit FAILURE;
    }

    $dcoWaitCounter = 0;
    $anomalyWosCanceled = FALSE;  # reset to check second anomaly
    while ($anomalyWosCanceled == FALSE)
    {
        if (getAnomalyStatus($units[1]->{OrderNbr}, $units[1]->{UnitNbr})
                eq DB_WO_RELEASE_ABORT)
        {
            $anomalyWosCanceled = TRUE;
            last;
        }

        if ($dcoWaitCounter > MAX_DCO_CHECKS)
        {
            last;
        }

        print "INFO: waiting for DCO...\n";
        sleep(WAIT_TIME);

        $dcoWaitCounter++;
    }
    if ($anomalyWosCanceled == TRUE)
    {
        print "INFO: Second anomaly successfully statused to "
              . "DB_WO_RELEASE_ABORT\n";
    }
    else
    {
        testCleanup();
        print "ERROR: failed to status second anomaly to DB_WO_RELEASE_ABORT\n";
        exit FAILURE;
    }
}

# Clean up TRAM, WORK_ORDERS, and PRODUCT_CODE_MODULES tables
for my $unit (@units)
{
    if (cleanupTramOrder($unit->{OrderNbr}) != SUCCESS)
    {
        testCleanup();
        print "ERROR: could not clean up TRAM database\n";
        exit FAILURE;
    }
    if (cleanupWorkOrder($unit->{OrderNbr}) != SUCCESS)
    {
        testCleanup();
        print "ERROR: could not clean up WO database\n";
        exit FAILURE;
    }
    if ($isLPGS)
    {
        if (cleanupAnomalies($units[0]->{OrderNbr}, $units[0]->{UnitNbr},
                    $units[1]->{OrderNbr}, $units[1]->{UnitNbr}) != SUCCESS)
        {
            testCleanup();
            print "ERROR: could not clean up Anomalies in database\n";
            exit FAILURE;
        }
    }
}
for my $unit (@rejectedUnits)
{
    if (cleanupTramOrder($unit->{OrderNbr}) != SUCCESS)
    {
        testCleanup();
        print "ERROR: could not clean up TRAM database for rejected units\n";
        exit FAILURE;
    }
}

if (cleanupProdCodeModule() != SUCCESS)
{
    testCleanup();
    print "ERROR: could not clean up Product Code in database\n";
    exit FAILURE;
}


# kill the capture_order process and close database connections
testCleanup();

print "Exiting test_tram_interface.\n";

exit SUCCESS;

##############################################################################
## Subroutines
##############################################################################

##############################################################################
# Name: initDb
# Purpose: Initiate database connections to TRAM and WO_COMMON
# Returns: SUCCESS - database connections were successfully opended
#          ERROR - error occurred
# Parameters: None
##############################################################################
sub initDb()
{
    my $tramEnv = $ENV{"IAS_DB_TRAM"};
    my $woEnv = $ENV{"IAS_DB_COM"};
    my $sid = $ENV{"ORACLE_SID"};
    my $username;
    my $password;

    # verfiy environment variables were set
    if (!defined $tramEnv)
    {
        print "ERROR: IAS_DB_TRAM not defined\n";
        return ERROR;
    }

    if (!defined $woEnv)
    {
        print "ERROR: IAS_DB_COM not defined\n";
        return ERROR;
    }

    if (!defined $sid)
    {
        print "ERROR: ORACLE_SID not defined\n";
        return ERROR;
    }

    ## Open TRAM database connection
    ($username, $password) = split(/\//,$tramEnv);

    $tramDbh = DBI->connect("DBI:Oracle:$sid", $username, $password,
                           {AutoCommit => 0, PrintError =>0});

    if (!$tramDbh)
    {
        print "ERROR: could not connect to TRAM db: $DBI::errstr\n";
        return ERROR;
    }

    ## Open work order database connection
    ($username, $password) = split(/\//,$woEnv);

    $woDbh = DBI->connect("DBI:Oracle:$sid", $username, $password,
                           {AutoCommit => 0, PrintError =>0});

    if (!$woDbh)
    {
        print "ERROR: could not connect to WO_COMMON: $DBI::errstr\n";
        return ERROR;
    }
    
    return SUCCESS;
}

##############################################################################
# Name: closeDb
# Purpose: Close database connections to TRAM and WO_COMMON
# Returns: None
# Parameters: None
##############################################################################
sub closeDb()
{
    $tramDbh->disconnect();
    $woDbh->disconnect();
}


##############################################################################
# Name: startDco
# Purpose: Start DCO in a child process.
# Returns: DCO PID
# Parameters: None
##############################################################################
sub startDco()
{
    my $pid = fork();
    if (!$pid)
    {
        # Disable warnings otherwise the error handling after exec() causes
        # warning messages.
        no warnings;
        print "starting DCO...\n";
        exec('../capture_order');
        print "ERROR: could not exec capture_order\n";
        exit FAILURE;
    }
 
    return $pid;
}

##############################################################################
# Name: stopDco
# Purpose: Stop the DCO process.
# Parameters: DCO PID
# Returns: None
##############################################################################
sub stopDco($)
{
    my ($pid) = @_;
    system("kill $pid");
}

##############################################################################
# Name: testCleanup
# Purpose: Close the database connections and kill the running DCO child.
# Returns: None
# Parameters: None
##############################################################################
sub testCleanup()
{
    closeDb();
    if (defined $dcoPid)
    {
        system("kill $dcoPid");
    }
}

##############################################################################
# Name: createTramUnit
# Purpose: Create an order in the TRAM tables
# Returns: SUCCESS - Order was successfully created
#          ERROR - Order was not created
# Parameters: orderNbr - TRAM order number
#             unitNbr - TRAM unit number
#             landsatSceneId -
#             prodCode - 
#             priority - 
# Notes: Right now the unitkey and lpgsondkey are hard-coded to '1'.  In TRAM
#   these are primary key and should be unique, but replicated TRAM in IAS
#   doesn't have these constraints.  DCO does not validate the acquistion date
#   against the scene Id, so the acquisition_date in the tram_lpgs_orders table
#   can be set to sysdate. 
##############################################################################
sub createTramUnit($$$$$)
{
    my ($orderNbr,$unitNbr,$landsatSceneId,$prodCode,$priority) = @_;
    my $insertCnt;

    my $orderSql = "INSERT INTO tram_orders "
                 . "            (order_nbr, "
                 . "             order_status, "
                 . "             date_entered, "
                 . "             priority) "
                 . "     VALUES ('$orderNbr', "
                 . "             'Q', "
                 . "             sysdate, "
                 . "             '$priority')";

    my $sth = $tramDbh->prepare($orderSql);

    $insertCnt = $sth->execute();
    $sth->finish();
    
    if ($insertCnt != 1)
    {
        $tramDbh->rollback();
        return ERROR;
    } 

    my $unitSql = "INSERT INTO tram_units "
                . "            (unitkey, "
                . "             order_nbr, "
                . "             unit_nbr, "
                . "             unit_status, "
                . "             ordering_id, "
                . "             prod_code, "
                . "             date_queued) " 
                . "     VALUES (1, "
                . "             '$orderNbr', "
                . "             $unitNbr, "
                . "             'Q', "
                . "             '$landsatSceneId', "
                . "             '$prodCode', "
                . "             sysdate)";
    
    $sth = $tramDbh->prepare($unitSql);

    $insertCnt = $sth->execute();
    $sth->finish();
    if ($insertCnt != 1)
    {
        $tramDbh->rollback();
        return ERROR;
    } 

    # Parse the path/row and acquisition date
    my $wrsPath = substr($landsatSceneId, 3, 3);
    my $wrsRow = substr($landsatSceneId, 6, 3);

    my $lpgsSql = "INSERT INTO tram_lpgs_orders "
                . "            (lpgsondkey, "
                . "             order_nbr, "
                . "             unit_nbr, "
                . "             satellite_id, "
                . "             wrs_path, "
                . "             wrs_row, "
                . "             landsat_scene_id, "
                . "             acquisition_date, "
                . "             priority, "
                . "             prod_code) "
                . "     VALUES (1, "
                . "             '$orderNbr', "
                . "             $unitNbr, "
                . "             'L8ORDER', "
                . "             $wrsPath, "
                . "             $wrsRow, "
                . "             '$landsatSceneId', "
                . "             sysdate, "
                . "             $priority, "
                . "             '$prodCode')";

    $sth = $tramDbh->prepare($lpgsSql);

    $insertCnt = $sth->execute();
    $sth->finish();
    if ($insertCnt != 1)
    {
        $tramDbh->rollback();
        return ERROR;
    } 

    if (!$tramDbh->commit())
    {
        print "ERROR: Create Tram unit commit failed: $DBI::errstr\n";
        return ERROR;
    }

    return SUCCESS;
}

##############################################################################
# Name: inIas
# Purpose: Check in an order is present in the work order database
# Returns: SUCCESS - Order is in work order database
#          ERROR - Order not found in work order database
# Parameters: orderNbr - TRAM order number
#             unitNbr - TRAM unit number
##############################################################################
sub inIas($$)
{
    my ($orderNbr, $unitNbr) = @_;

    my $sql = "SELECT count(*) cnt "
            . "  FROM work_orders "
            . " WHERE customer_order_nbr = '$orderNbr' "
            . "   AND customer_unit_nbr = $unitNbr "
            . "   AND wo_replaced = 0";

    my $sth = $woDbh->prepare($sql) or return ERROR;
    $sth->execute();

    my $found = ERROR;
    while (my ($cnt) = $sth->fetchrow_array())
    {
        if ($cnt > 0)
        {
            $found = SUCCESS;
        }
    }
    $sth->finish();

    return $found;
}

##############################################################################
# Name: getPriorityAdjustment
# Purpose: Returns the priority adjustment for the current processing instance
# Returns: Priority adjusment or -999 as Error
##############################################################################
sub getPriorityAdjustment()
{
    my $sql = "SELECT priority_adjustment FROM processing_instances " .
        " WHERE instance_name = '$processing_instance'";

    my $sth = $woDbh->prepare($sql) or return -999;
    $sth->execute();

    my ($adjustment) = $sth->fetchrow_array();
    $sth->finish();

    return -999 if (!defined $adjustment);

    return $adjustment;
}

##############################################################################
# Name: getTramStatus
# Purpose: Get status of unit in TRAM.
# Returns: TRAM unit status or empty string if unit not found.
# Parameters: orderNbr - TRAM order number
#             unitNbr - TRAM unit number
##############################################################################
sub getTramStatus($$)
{
    my ($orderNbr, $unitNbr) = @_;

    my $sql = "SELECT unit_status "
            . "  FROM tram_units "
            . " WHERE order_nbr = '$orderNbr' "
            . "   AND unit_nbr = $unitNbr";

    my $sth = $tramDbh->prepare($sql);
    $sth->execute();

    my $status = '';
    if (my ($dbStatus) = $sth->fetchrow_array())
    {
        $status = $dbStatus;
    }
    $sth->finish();

    return $status;
}

##############################################################################
# Name: getWoStatus
# Purpose: Get status of Work Order in WORK_ORDERS.
# Returns: Work Order status or empty string if WO not found.
# Parameters: orderNbr - TRAM order number
#             unitNbr - TRAM unit number
##############################################################################
sub getWoStatus($$)
{
    my ($orderNbr, $unitNbr) = @_;
    
    my $sql = "SELECT state "
            . "  FROM work_orders "
            . " WHERE customer_order_nbr = '$orderNbr' "
            . "   AND customer_unit_nbr = $unitNbr "
            . "   AND wo_replaced = 0";

    my $sth = $woDbh->prepare($sql);
    $sth->execute();

    my $status = '';
    if (my ($dbStatus) = $sth->fetchrow_array())
    {
        $status = $dbStatus;
    }
    $sth->finish();

    return $status;
}

##############################################################################
# Name: getAnomalyStatus
# Purpose: Get status of the anomaly Work Order in WORK_ORDERS created earlier
# Returns: Anomaly Work Order status or empty string if WO not found.
# Parameters: orderNbr - TRAM order number
#             unitNbr - TRAM unit number
##############################################################################
sub getAnomalyStatus($$)
{
    my ($orderNbr, $unitNbr) = @_;

    my $sql = "SELECT state "
            . "  FROM work_orders "
            . " WHERE customer_order_nbr = '$orderNbr' "
            . "   AND customer_unit_nbr = $unitNbr "
            . "   AND wo_replaced = 1";

    my $sth = $woDbh->prepare($sql);
    $sth->execute();

    my $state = '';
    if (my $dbState = $sth->fetchrow_array())
    {
        $state = $dbState;
    }
    $sth->finish();

    return $state;
}

##############################################################################
# Name: updateWoStatus
# Purpose: Updates status of order in work order database
# Returns: SUCCESS - update successful
#          ERROR - update failed
# Parameteres: orderNbr - TRAM order number
#              unitNbr - TRAM unit number
#              status - New order status
##############################################################################
sub updateWoStatus($$$)
{
    my ($orderNbr, $unitNbr, $status) = @_;

    my $sql = "UPDATE work_orders "
            . "   SET state = '$status' "
            . " WHERE customer_order_nbr = '$orderNbr' "
            . "   AND customer_unit_nbr = $unitNbr "
            . "   AND wo_replaced = 0";

    $woDbh->do($sql) or return ERROR;
    if (!$woDbh->commit())
    {
        print "ERROR: Update WO status commit failed: $DBI::errstr\n";
        return ERROR;
    }
    return SUCCESS;
}

##############################################################################
# Name: createAnomalies
# Purpose: Create duplicate work orders for the work orders given (by order
#          and unit number) - quick and dirty manual entry to WORK_ORDERS. A
#          work order that had an anomaly will be in DB_WO_ABORT_WO state and
#          WO_REPLACED will be '1'. So mock this up so DCO can be tested to
#          update these to DB_WO_RELEASE_ABORT.
# Returns: SUCCESS
#          ERROR
# Parameters: orderNbr1 - TRAM order number of 1st work order to duplicate
#             unitNbr1 - TRAM unit number of 1st work order to duplicate
#             orderNbr2 - TRAM order number of 2nd work order to duplicate
#             unitNbr2 - TRAM unit number of 2nd work order to duplicate
##############################################################################
sub createAnomalies($$$$)
{
    my ($orderNbr1, $unitNbr1, $orderNbr2, $unitNbr2) = @_;
    my $productRequestId1;
    my $productRequestId2;
    my $insertCnt;
    my $procInstance1;
    my $procInstance2;

    # Get the product request id for the two orders that will have fake
    # anomalies created for them, as the anomaly finding in DCO checks for
    # a duplicate product request id and wo_replaced of '1'
    my $sql = "SELECT product_request_id, processing_instance "
            . "  FROM work_orders "
            . " WHERE customer_order_nbr = '$orderNbr1' "
            . "   AND customer_unit_nbr = $unitNbr1 "
            . "   AND wo_replaced = 0";

    my $sth = $woDbh->prepare($sql);
    $sth->execute();

    if (my $dbHashRef = $sth->fetchrow_hashref())
    {
        $productRequestId1 = $dbHashRef->{PRODUCT_REQUEST_ID};
        $procInstance1 = $dbHashRef->{PROCESSING_INSTANCE};
    }

    $sth->finish();

    $sql = "SELECT product_request_id "
         . "  FROM work_orders "
         . " WHERE customer_order_nbr = '$orderNbr2' "
         . "   AND customer_unit_nbr = $unitNbr2 "
         . "   AND wo_replaced = 0";

    $sth = $woDbh->prepare($sql);
    $sth->execute();

    if (my $dbHashRef = $sth->fetchrow_hashref())
    {
        $productRequestId2 = $dbHashRef->{PRODUCT_REQUEST_ID};
        $procInstance2 = $dbHashRef->{PROCESSING_INSTANCE};
    }

    $sth->finish();

    # Create two work orders that look like duplicates to the ones provided as
    # parameters, but flag that they have been replaced (anomalies happened).
    $sql = "INSERT INTO work_orders "
         . "            (work_order_id, processing_instance, "
         . "             customer_order_nbr, "
         . "             customer_unit_nbr, "
         . "             product_request_id, "
         . "             state, "
         . "             wo_replaced, "
         . "             priority, "
         . "             requester, "
         . "             request_type) "
         . "     VALUES ('X1234', " . "'$procInstance1', "
         . "             '$orderNbr1', "
         . "             $unitNbr1, "
         . "             '$productRequestId1', "
         . "             '" . DB_WO_ABORT_WO . "', "
         . "             1, "
         . "             9, "
         . "             'DCO_REGRESSION_TEST', "
         . "             'S')";

    $sth = $woDbh->prepare($sql);
    if (!$sth)
    {
        $woDbh->rollback();
        print "ERROR: Insert anomaly work order prepare failed: " .
            "$DBI::errstr\n";
        return ERROR;
    }

    $insertCnt = $sth->execute();
    $sth->finish();

    if ($insertCnt != 1)
    {
        $woDbh->rollback();
        print "ERROR: Insert anomaly work order commit failed: $DBI::errstr\n";
        return ERROR;
    }

    if (!$woDbh->commit())
    {
        print "ERROR: Create anomaly work order commit failed: $DBI::errstr\n";
        return ERROR;
    }

    $sql = "INSERT INTO work_orders "
         . "            (work_order_id, processing_instance, "
         . "             customer_order_nbr, "
         . "             customer_unit_nbr, "
         . "             product_request_id, "
         . "             state, "
         . "             wo_replaced, "
         . "             priority, "
         . "             requester, "
         . "             request_type) "
         . "     VALUES ('X2345', " . "'$procInstance1', "
         . "             '$orderNbr2', "
         . "             $unitNbr2, "
         . "             '$productRequestId2', "
         . "             '" . DB_WO_ABORT_WO . "', "
         . "             1, "
         . "             9, "
         . "             'DCO_REGRESSION_TEST', "
         . "             'S')";

    $sth = $woDbh->prepare($sql);
    if (!$sth)
    {
        $woDbh->rollback();
        print "ERROR: Insert anomaly work order prepare failed: " .
            "$DBI::errstr\n";
        return ERROR;
    }

    $insertCnt = $sth->execute();
    $sth->finish();

    if ($insertCnt != 1)
    {
        $woDbh->rollback();
        print "ERROR: Insert anomaly work order commit failed: $DBI::errstr\n";
        return ERROR;
    }

    if (!$woDbh->commit())
    {
        print "ERROR: Create anomaly work order commit failed: $DBI::errstr\n";
        return ERROR;
    }
    
    return SUCCESS;
}

##############################################################################
# Name: createProdCodeModule
# Purpose: Create an entry in the PRODUCT_CODE_MODULES table for testing
# Returns: SUCCESS
#          ERROR
# Parameters: None
##############################################################################
sub createProdCodeModule()
{
    # create a new product code module entry to use for testing
    # can't use what is in the table already, since IAS/LPGS vary

    my $sql = "INSERT INTO product_code_modules "
            . "            (prod_code, module_id, processing_instance) "
            . "    VALUES ('$prodCode', 'L1T Product', '$processing_instance')";

    my $sth = $woDbh->prepare($sql);
    if (!$sth)
    {
        $woDbh->rollback();
        print "ERROR: Insert product code module prepare failed: " .
            "$DBI::errstr\n";
        return ERROR;
    }

    my $insertCnt = $sth->execute();
    $sth->finish();

    if ($insertCnt != 1)
    {
        $woDbh->rollback();
        print "ERROR: Insert product code module commit failed: $DBI::errstr\n";
        return ERROR;
    }

    if (!$woDbh->commit())
    {
        print "ERROR: Create product code commit failed: $DBI::errstr\n";
        return ERROR;
    }

    $sql = "INSERT INTO tram_products "
         . "            (prod_code, "
         . "             prod_description, "
         . "             prod_type, "
         . "             prod_group, "
         . "             prod_status, "
         . "             date_entered) "
         . "     VALUES ('$prodCode', "
         . "             'DCO TEST - L1R Product', "
         . "             'L', "
         . "             'OLITIRS', "
         . "             'A', "
         . "             sysdate)";

    $sth = $tramDbh->prepare($sql);
    if (!$sth)
    {
        $tramDbh->rollback();
        print "ERROR: Insert TRAM product code prepare failed: $DBI::errstr\n";
        return ERROR;
    }

    $insertCnt = $sth->execute();
    $sth->finish();
    
    if ($insertCnt != 1)
    {
        $tramDbh->rollback();
        print "ERROR: Insert TRAM product code commit failed: $DBI::errstr\n";
        return ERROR;
    } 

    # Insert a bad product code that isn't present in our tables
    $sql = "INSERT INTO tram_products "
         . "            (prod_code, "
         . "             prod_description, "
         . "             prod_type, "
         . "             prod_group, "
         . "             prod_status, "
         . "             date_entered) "
         . "     VALUES ('$badProdCode', "
         . "             'DCO TEST - L1X Product', "
         . "             'L', "
         . "             'OLITIRS', "
         . "             'A', "
         . "             sysdate)";

    $sth = $tramDbh->prepare($sql);
    if (!$sth)
    {
        $tramDbh->rollback();
        print "ERROR: Insert TRAM product code prepare failed: $DBI::errstr\n";
        return ERROR;
    }

    $insertCnt = $sth->execute();
    $sth->finish();
    
    if ($insertCnt != 1)
    {
        $tramDbh->rollback();
        print "ERROR: Insert TRAM product code commit failed: $DBI::errstr\n";
        return ERROR;
    } 

    if (!$tramDbh->commit())
    {
        print "ERROR: Create product code commit failed: $DBI::errstr\n";
        return ERROR;
    }

    return SUCCESS;
}

##############################################################################
# Name: cleanupTramOrder
# Purpose: Remove order from TRAM database
# Returns: SUCCESS
#          ERROR
# Parameters: orderNbr - TRAM order number
##############################################################################
sub cleanupTramOrder($)
{
    my ($orderNbr) = @_;

    $tramDbh->do("DELETE FROM tram_lpgs_orders "
               . "      WHERE order_nbr = '$orderNbr'");

    $tramDbh->do("DELETE FROM tram_units "
               . "      WHERE order_nbr = '$orderNbr'");

    $tramDbh->do("DELETE FROM tram_orders "
               . "      WHERE order_nbr = '$orderNbr'");

    if(!$tramDbh->commit())
    {
        print "ERROR: Cleanup Tram Order commit failed: $DBI::errstr\n";
        return ERROR;
    }
    return SUCCESS;
}

##############################################################################
# Name: cleanupWorkOrder
# Purpose: Remove order from work order database
# Returns: SUCCESS
#          ERROR
# Parameters: orderNbr - TRAM order number
##############################################################################
sub cleanupWorkOrder($)
{
    my ($orderNbr) = @_;

    # Only need to clean up the work orders tables
    # Cascading deletes will clean up secondary tables
    $woDbh->do("DELETE FROM work_orders "
             . "      WHERE customer_order_nbr = '$orderNbr' "
             . "        AND wo_replaced = 0");

    if (!$woDbh->commit())
    {
        print "ERROR: Cleanup Work Order commit failed: $DBI::errstr\n";
        return ERROR;
    }

    return SUCCESS;
}

##############################################################################
# Name: cleanupProdCodeModule
# Purpose: Remove fake product code entry from database that was inserted
#          for testing purposes only.
# Returns: SUCCESS
#          ERROR
# Parameters: None
##############################################################################
sub cleanupProdCodeModule()
{
    $woDbh->do("DELETE FROM product_code_modules "
             . "      WHERE prod_code = '$prodCode'");

    if (!$woDbh->commit())
    {
        print "ERROR: Cleanup Product Code commit failed: $DBI::errstr\n";
        return ERROR;
    }

    $tramDbh->do("DELETE FROM tram_products "
               . "      WHERE prod_code = '$prodCode'");
    $tramDbh->do("DELETE FROM tram_products "
               . "      WHERE prod_code = '$badProdCode'");
    if (!$tramDbh->commit())
    {
        print "ERROR: Cleanup TRAM Product Code commit failed: $DBI::errstr\n";
        return ERROR;
    }

    return SUCCESS;
}

##############################################################################
# Name: cleanupAnomalies
# Purpose: Remove duplicate "anomaly" work orders that were inserted for test.
# Returns: SUCCESS
#          ERROR
# Parameters: None
##############################################################################
sub cleanupAnomalies($$$$)
{
    my ($orderNbr1, $unitNbr1, $orderNbr2, $unitNbr2) = @_;

    $woDbh->do("DELETE FROM work_orders "
             . "      WHERE customer_order_nbr = '$orderNbr1' "
             . "        AND customer_unit_nbr = $unitNbr1 "
             . "        AND wo_replaced = 1");

    $woDbh->do("DELETE FROM work_orders "
             . "      WHERE customer_order_nbr = '$orderNbr2' "
             . "        AND customer_unit_nbr = $unitNbr2 "
             . "        AND wo_replaced = 1");

    if (!$woDbh->commit())
    {
        print "ERROR: Cleanup Anomaly Work Orders commit failed: " .
            $DBI::errstr . "\n";
        return ERROR;
    }

    return SUCCESS;
}
##############################################################################
# Name: checkOrder
# Purpose: Verify the orders created in WORK_ORDERS table are in the same
#          sequence that the tram orders were created in.
# Returns: TRUE - Work orders are created in same sequence as tram orders
#          FALSE - Work orders were not created in same sequence as tram orders
#          ERROR - Database issue prevented order validation
# Parameters: 
#            
##############################################################################
sub checkOrder()
{
    my $prevId; 
    my $isInorder = TRUE;
    my $isFirstTime = TRUE;

    my $sql = "   SELECT substr(work_order_id, 2, length(work_order_id)) "
            . "     FROM work_orders "
            . "    WHERE customer_order_nbr in "
            . "          ('0121102120002', "
            . "           '0121102120001', "
            . "           '0121102120005', "
            . "           '0121102120003', "
            . "           '0121102120004') "
            . "      AND wo_replaced = 0 "
            . " ORDER BY priority, substr(customer_order_nbr, 4), "
            . "          customer_unit_nbr";

    my $sth = $woDbh->prepare($sql) or return ERROR;
    $sth->execute();

    while (my ($currentId) = $sth->fetchrow_array())
    {
        if ($isFirstTime == TRUE)
        {
            $isFirstTime = FALSE;
            $prevId = $currentId;
        }
        else
        {
            if ( $prevId > $currentId )
            {
                $isInorder = FALSE;
                return $isInorder;
            }
            else
            {
                $prevId = $currentId;
            }
        }
    }
    return $isInorder;
}

##############################################################################
# Name: checkPriority
# Purpose: Check to see if the correct priority was brought over
# Returns: TRUE - Work order priorities match the TRAM record priorities.
#          FALSE - match failure
##############################################################################
sub checkPriority($$)
{
    my ($orderNbr, $priority) = @_;

    my $match = FALSE;
    my $sql = "SELECT priority "
            . "  FROM work_orders "
            . " WHERE customer_order_nbr = '$orderNbr' "
            . "   AND wo_replaced = 0";
    
    my $sth = $woDbh->prepare($sql) or return ERROR;
    $sth->execute();

    if (my ($dbpriority) = $sth->fetchrow_array())
    {
        my $priorityAdjustment = getPriorityAdjustment();
        if ($priorityAdjustment == -999)
        {
            print "ERROR: Unable to get the priority adjustment for " .
            "processing instance: $processing_instance\n";
            return FALSE;
        }

        if ($dbpriority == $priority + $priorityAdjustment)
        {
           $match = TRUE;
        }
    }
    $sth->finish();
    return $match;
}


##############################################################################
# Name: checkInstanceEnabledFlag
# Purpose: Check to see if the instance is enabled
# Returns: TRUE/FALSE
###############################################################################
sub checkInstanceEnabledFlag($)
{
    my ($processing_instance) = @_;

    my $is_enabled = FALSE;
    my $sql = "SELECT enabled "
            . "  FROM processing_instances "
            . " Where instance_name = '" . $processing_instance . "' ";

    my $sth = $woDbh->prepare($sql) or return FALSE;
    $sth->execute();

    if (my ($enabled) = $sth->fetchrow_array())
    {
        if ($enabled eq 'Y')
        {
            $is_enabled = TRUE;
        }
    }
    $sth->finish();

    return $is_enabled;
}
