#!/usr/bin/perl -w
#------------------------------------------------------------------------------
# Name: determine_projection.pl
#
# Description: Determines which output projection to use, retrieve projection
#    parameters, and updates the OMF with the projection parameters and
#    TARGET_PROJECTION_PASS_#
#
#------------------------------------------------------------------------------
use strict;
# Global routines
use DBI;
use File::Copy qw(move);
use Logger; # Log routines
use ias_processing_system;
use ias_const;
use ias_db; # Database routines

# Declare program constant
use constant PROG => 'determine_projection.pl';
use constant OVERRIDE_PROJECTION => 'OVERRIDE_PROJECTION_PASS';

my $status;

# Begin main module
# Set up logger
my $logger = Logger->new('STDOUT');
if (!$logger)
{
    print PROG . ": Unable to create logger object for STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name(PROG);
set_default_log($logger);

if (ias_init_system() == ERROR)
{
    $ias_error_msg = 'Undefined error' if !defined $ias_error_msg;
    $logger->fatal_msg("$ias_error_msg");
    $logger->close_log();
    exit FAILURE; # Exits with status = 1.
}


# Check the instance variable
my $processing_instance = $ENV{IAS_INSTANCE_NAME};
if (!$processing_instance)
{
    $logger->fatal_msg("The IAS_INSTANCE_NAME enironment variable is not set");
    $logger->close_log();
    exit FAILURE;
}

# Grab the parameter file that was passed in
if ($#ARGV != 0) # 1 parameter passed in
{
    $logger->fatal_msg("Requires a parameter file argument.");
    $logger->close_log();
    exit FAILURE;
}
# The main parameter file that is passed in on the command-line
my $parameter_filename = $ARGV[0];

# Validate the parameter file's existence
if (!-e $parameter_filename)
{
    $logger->fatal_msg("Parameter file '$parameter_filename' does not exist.");
    $logger->close_log();
    exit FAILURE;
}

# Get the work order id and work order directory from the parameter file
my ($wo_id, $wo_directory, $processing_pass, $project) = 
    parse_parameters($parameter_filename);
if (!defined $wo_id || !defined $wo_directory || !defined $processing_pass ||
    !defined $project)
{
    $logger->fatal_msg("Unable to parse parameter file: '$parameter_filename'");
    $logger->close_log();
    exit FAILURE;
}

$logger->info_msg("Starting WORK_ORDER_ID: $wo_id");

# Get the wrs path and row from the OMF
my $omf_name = $wo_directory . "/" . $wo_id . ".omf";
my ($path, $row) = parse_omf($omf_name);
if (!defined $path || !defined $row)
{
    $logger->fatal_msg("parse_omf failed");
    $logger->close_log();
    exit FAILURE;
}

# Get the username and password for the common database
my ($user_id, $password) = ias_db_GetDBConnectionInfo("COMMON");
my $dbh = ias_db_ConnectToDB($user_id, $password);
if (!defined $dbh)
{
    $logger->fatal_msg("Unable to connect to database: $ias_db_error");
    $logger->close_log();
    exit FAILURE;
}

my %attributes = (
    # Schema to use when retrieving the projection information
    "DETERMINE_PROJECTION_SCHEMA" => "",
    # Link to use when retrieving the projection information
    "DETERMINE_PROJECTION_LINK" => "",
    );

# Grab the synonym names
$status = ias_db_GetProcessingInstanceAttributes(\%attributes,
    $processing_instance, $dbh, $logger);
if ($status != SUCCESS)
{
    $logger->fatal_msg("Unable to retrieve processing instance attributes");
    $logger->close_log();
    $dbh->disconnect();
    exit FAILURE;
}
# Get the schema and link to use for the DETERMINE_PROJECTION script
my $dp_schema = $attributes{"DETERMINE_PROJECTION_SCHEMA"}; 
# If the current instance is used, link will be "NA"
my $dp_link = ""; 
if (defined($attributes{"DETERMINE_PROJECTION_LINK"}) && 
    $attributes{"DETERMINE_PROJECTION_LINK"} ne "NA")
{
    $dp_link = "@" . $attributes{"DETERMINE_PROJECTION_LINK"}; 
}

# Determine projection to use
my ($projection) = get_projection_name($dbh, $project, $path, $row, $dp_schema,
                                       $dp_link);
if (!defined $projection)
{
    $logger->fatal_msg("get_projection_name failed");
    $dbh->disconnect();
    $logger->close_log();
    exit FAILURE;
}

# Get the projection parameters for the projection to use
my ($projection_code, $projection_parameters_ref) = 
    get_projection_parameters($dbh, $projection, $dp_schema, $dp_link);
if (!defined $projection_code || !defined $projection_parameters_ref)
{
    $logger->fatal_msg("get_projection_parameters failed");
    $dbh->disconnect();
    $logger->close_log();
    exit FAILURE;
}
# No longer need a database connection
$dbh->disconnect();

# Update the OMF with target projection pass, projection code, and parameters
$status = update_omf($dbh, $omf_name, $processing_pass, $projection_code, 
                     $projection_parameters_ref);
if ($status != SUCCESS)
{
    $logger->fatal_msg("update_omf failed");
    $logger->close_log();
    exit FAILURE;
}

# Log completion message.
$logger->info_msg("Successful completion");
$logger->close_log();

# Success!
exit SUCCESS;

#------------------------------------------------------------------------------
# Subroutines below
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
#
# Name: get_projection_name
#
# Purpose: Queries PATH_ROW_TO_PROJECTION to get projection to use based
#    on wrs path and row 
#
# Inputs:
# I: Database connection handle
# I: Project for which the projection is supported
# I: WRS Path
# I: WRS Row
# I: DETERMINE_PROJECTION_SCHEMA to use when retrieving projection name
# I: DETERMINE_PROJECTION_LINK to use when retrieving projection name
#
# Returns:
# Projection on SUCCESS
# undef on ERROR
#
#------------------------------------------------------------------------------
sub get_projection_name
{
    my $dbh = shift;
    my $project = shift;
    my $path = shift;
    my $row = shift;
    my $dp_schema = shift;
    my $db_link = shift;

    my $query = "select PROJECTION_NAME from ".
        "$dp_schema.PATH_ROW_TO_PROJECTION$db_link where ".
        "PROJECT = :PROJECT and WRS_PATH = :WRS_PATH and WRS_ROW = :WRS_ROW";
    
    my $sth = $dbh->prepare($query);
    if (!$sth)
    {
        $logger->fatal_msg("Error preparing SQL $query");
        return undef;
    }
    $sth->bind_param(':PROJECT', $project);
    $sth->bind_param(':WRS_PATH', $path);
    $sth->bind_param(':WRS_ROW', $row);
    if (!$sth->execute())
    {
        $logger->fatal_msg("Error executing SQL $query");
        $sth->finish();
        return undef;
    }

    my ($projection) = $sth->fetchrow_array();
    if (!defined $projection)
    {
        $logger->fatal_msg("Projection not found for Path $path Row $row");
        $sth->finish();
        return undef;
    }
    $sth->finish();
    return $projection;
}

#------------------------------------------------------------------------------
#
# Name: get_projection_parameters
#
# Purpose: Queries PROJECTION_PARAMETERS to get parameters to use based
#    on projection
#
# Inputs:
# I: Database connection handle
# I: Projection
# I: DETERMINE_PROJECTION_SCHEMA to use when retrieving projection name
# I: DETERMINE_PROJECTION_LINK to use when retrieving projection name
#
# Returns:
# Projection code and Projection parameters on SUCCESS
# undef, undef on ERROR
#
#------------------------------------------------------------------------------
sub get_projection_parameters
{
    my $dbh = shift;
    my $projection = shift;
    my $dp_schema = shift;
    my $db_link = shift;

    my $query = "select ".
        "PROJECTION_CODE, ".
        "PARAMETER1, ".
        "PARAMETER2, ".
        "PARAMETER3, ".
        "PARAMETER4, ".
        "PARAMETER5, ".
        "PARAMETER6, ".
        "PARAMETER7, ".
        "PARAMETER8, ".
        "PARAMETER9, ".
        "PARAMETER10, ".
        "PARAMETER11, ".
        "PARAMETER12, ".
        "PARAMETER13, ".
        "PARAMETER14, ".
        "PARAMETER15 from $dp_schema.PROJECTION_PARAMETERS$db_link ".
        "where PROJECTION_NAME = :PROJECTION";
    
    my $sth = $dbh->prepare($query);
    if (!$sth)
    {
        $logger->fatal_msg("Error preparing SQL $query");
        return undef, undef;
    }
    $sth->bind_param(':PROJECTION', $projection);
    if (!$sth->execute())
    {
        $logger->fatal_msg("Error executing SQL $query");
        $sth->finish();
        return undef, undef;
    }

    my @parameters = ();
    @parameters = $sth->fetchrow_array();
    if (!@parameters)
    {
        $logger->fatal_msg("Projection parameters not found for $projection");
        $sth->finish();
        return undef, undef;
    }
    $sth->finish();

    # Projection Code is the first element in array
    my $projection_code = shift @parameters;
    # Projection parameters are left in the array after the shift
    return $projection_code, \@parameters;
}

#------------------------------------------------------------------------------
#
# Name: parse_omf
#
# Purpose: Parse an OMF for WRS Path and Row
#
# Inputs:
# I: OMF filename
#
# Returns:
# WRS Path and Row, Project on SUCCESS
# undef, undef, undef on ERROR
#
#------------------------------------------------------------------------------
sub parse_omf
{
    my $omf = shift;

    my $input_line;
    my $status;
    my $wrs_path;
    my $wrs_row;

    # Open the OMF for reading
    $status = open (INPUT, "< $omf");
    if (!defined($status))
    {
        $logger->fatal_msg("Can't open input parameter file: $omf");
        return undef, undef, undef;
    }

    # Process each line in the file
    while ($input_line = <INPUT>)
    {
        # WRS path
        if ($input_line =~ /^\s*TARGET_WRS_PATH\s*=\s*(.*)/)
        {
            if ($1 =~ /(\d+)/)
            {
                $wrs_path = $1;
            }
            else
            {
                $logger->fatal_msg("Invalid WRS path format: $input_line");
                return undef, undef, undef;
            }
        }

        # WRS row
        if ($input_line =~ /^\s*TARGET_WRS_ROW\s*=\s*(.*)/)
        {
            if ($1 =~ /(\d+)/)
            {
                $wrs_row = $1;
            }
            else
            {
                $logger->fatal_msg("Invalid WRS row format: $input_line");
                return undef, undef, undef;
            }
        }
    }
    # Check whether the WRS path and/or row have been specified.  If the
    # path has been specified but the row hasn't, or vice versa, treat it
    # as an error -- both must be specified
    if ($wrs_path && $wrs_row)
    {
        if (($wrs_path < 1) || ($wrs_path > 233))
        {
            $logger->fatal_msg("WRS path must be between 1 and 233");
            return undef, undef, undef;
        }

        if (($wrs_row < 1) || ($wrs_row > 248) &&
            ($wrs_row < 880 || $wrs_row > 889) &&
            ($wrs_row < 990 || $wrs_row > 999))
        {
            $logger->fatal_msg("WRS row must be between 1 and 248, ".
            "between 880 and 889, or between 990 and 999");
            return undef, undef, undef;
        }
    }
    else
    {
        $logger->fatal_msg("WRS path and WRS row should both be ".
            "specified");
    }

    return $wrs_path, $wrs_row, $project;
}

#------------------------------------------------------------------------------
#
# Name: parse_parameters
#
# Purpose: Parse a parameter file for the work order ID and directory and
#          processing pass
#
# Inputs:
# I: parameter filename
#
# Returns:
# work order ID, work order directory, and processing pass on SUCCESS
# undef, undef, undef, undef on ERROR
#
#------------------------------------------------------------------------------
sub parse_parameters
{
    # Get the parameter file name
    my $parm_filename = shift;
    my $input_line;
    my $processing_pass;
    my $project;
    my $status;
    my $wo_id;
    my $wo_directory;

    # Open the file to read
    $status = open (INPUT, "< $parm_filename");
    if (!defined($status))
    {
        $logger->fatal_msg("Can't open input parameter file: $parm_filename");
        return undef, undef, undef, undef;
    }

    # Process each line in the file
    while ($input_line = <INPUT>)
    {
        # Work order directory
        if ($input_line =~ /^\s*WO_DIRECTORY\s*=\s*"?(.*)"?/)
        {
            if ($1 =~ /([\w\.\/]*)/)
            {
                $wo_directory = $1;
            }
            else
            {
                $logger->fatal_msg("Invalid work order directory string: ".
                    "$input_line");
                return undef, undef, undef, undef;
            }
        }

        # Work order ID
        if ($input_line =~ /^\s*WORK_ORDER_ID\s*=\s*"?(.+)"?/)
        {
            if ($1 =~ /(\w+)/)
            {
                $wo_id = $1;
            }
            else
            {
                $logger->fatal_msg("Invalid work order ID string: $input_line");
                return undef, undef, undef, undef;
            }
        }

        # Processing Pass #
        if ($input_line =~ /^\s*PROCESSING_PASS\s*=\s*(\d+)/)
        {
            if ($1 =~ /(\d+)/)
            {
                $processing_pass = $1;
            }
            else
            {
                $logger->fatal_msg("Invalid processing pass: $input_line");
                return undef, undef, undef, undef;
            }
        }

        # Project 
        if ($input_line =~ /^\s*PROJECT\s*=\s*\"?(.*)\"?/)
        {
            if ($1 =~ /(\w+)/)
            {
                $project = $1;
            }
            else
            {
                $logger->fatal_msg("Invalid project project format: $input_line");
                return undef, undef, undef, undef;
            }
        }
    }
    # Close the input file
    close (INPUT);

    return $wo_id, $wo_directory, $processing_pass, $project;
}

#------------------------------------------------------------------------------
#
# Name: update_omf
#
# Purpose: Adds target projection pass # and projection parameters to omf
#
# Inputs:
# I: Database handle
# I: Original OMF name
# I: Projection Code
# I: Projection parameters
#
# Returns:
# SUCCESS or ERROR
#
#------------------------------------------------------------------------------
sub update_omf
{
    my $dbh = shift;
    my $omf = shift;
    my $processing_pass = shift;
    my $projection_code = shift;
    my $parameters_ref = shift;
    my $new_omf = $omf."_new";
    # Dereference array reference
    my @parameters = @{$parameters_ref};

    # Open old omf for reading
    my $status = open (OLD, "< $omf");
    if (!defined $status)
    {
        $logger->fatal_msg("Can't open omf $omf");
        return ERROR;
    }
    # Open new omf for writing
    $status = open (NEW, "> $new_omf");
    if (!defined $status)
    {
        $logger->fatal_msg("Can't open new omf $new_omf");
        close(OLD);
        return ERROR;
    }

    while(<OLD>)
    {
        if (/END_OBJECT/)
        {
            # Create a string with the parameters, separated by commas
            my $parms = join(",", @parameters);
            my $print_string = "  OVERRIDE_PROJECTION_PASS_$processing_pass = ".
                "$projection_code\n";
            $print_string .= "  OVERRIDE_PROJ_PARMS_PASS_$processing_pass = (".
                "$parms)\n";
            print NEW $print_string;
        }
        print NEW $_;
    }
    if (!close(NEW) || !close(OLD))
    {
        $logger->fatal_msg("Problems occurred closing open files.");
        return ERROR;
    }

    # Rename to original name
    if (!move($new_omf, $omf))
    {
        $logger->fatal_msg("Can't move $new_omf to $omf");
        return ERROR;
    }

    return SUCCESS;
}
