#!/usr/bin/perl 
#-----------------------------------------------------------------------------
# Name: test_determine_projection.pl
#
# Description: Test the determine_projection.pl script
#-----------------------------------------------------------------------------

use strict;
use warnings;
use ias_const;
use ias_db;

# If database tests are disabled, skip this test
if (!defined $ENV{ENABLE_DB_TESTS})
{
    print "Database tests are disabled... Skipping $0\n";
    exit SUCCESS;
}

# Get the username and password for the common database
my ($user_id, $password) = ias_db_GetDBConnectionInfo('COMMON');

# Connect to the database
my $dbh = ias_db_ConnectToDB($user_id, $password);
if (!defined $dbh)
{
    print "Unable to connect to database for user $user_id\n";
    exit FAILURE;
}

# Set target path/row in OMF to 40/30, this should return albers_US as the
# projection name
my $path = 40;
my $row = 30;
my $status = run_test_case($dbh, $path, $row, "albers_US");
if ($status != SUCCESS)
{
    print "Error running $path $row albers_US test case\n";
    ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
    exit FAILURE;
}

# Set target path/row in OMF to 61/18, this should return albers_Alaska as the
# projection name
$path = 61;
$row = 18;
$status = run_test_case($dbh, $path, $row, "albers_Alaska");
if ($status != SUCCESS)
{
    print "Error running $path $row albers_Alaska test case\n";
    ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
    exit FAILURE;
}

# Set target path/row in OMF to 66/45, this should return albers_Hawaii as the
# projection name
$path = 66;
$row = 45;
$status = run_test_case($dbh, $path, $row, "albers_Hawaii");
if ($status != SUCCESS)
{
    print "Error running $path $row albers_Hawaii test case\n";
    ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
    exit FAILURE;
}

# Disconnect from DB
if (!ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK))
{
    print "Unable to disconnect from database: $ias_db_error";
    exit FAILURE;
}

exit SUCCESS;

#-----------------------------------------------------------------------------
# Name: run_test_case
#
# Description: Updates the OMF, performs one run of the test case,
#              and verifies results
#
# Parameters:
#   Database connection
#   Path for omf
#   Row for omf
#   Projection Name
#
# Returns
#   SUCCESS/ERROR
#-----------------------------------------------------------------------------
sub run_test_case
{
    my $dbh = shift;
    my $path = shift;
    my $row  = shift;
    my $projection = shift;

    # Update OMF for test with new path, row
    $status = update_omf($path, $row);
    if ($status != SUCCESS)
    {
        print "Error updating OMF for Path: $path Row $row test case\n";
        return ERROR;
    }
    # Run the application
    $status = system("../determine_projection.pl DETERMINE_PROJECTION.odl");
    if ($status != SUCCESS)
    {
        print "Error returned from determine_projection.pl for ".
            "Path: $path Row $row test case\n";
        return ERROR;
    }
    # Verify OMF was updated as expected
    $status = verify_omf($dbh, $projection);
    if ($status != SUCCESS)
    {
        print "Error verifying OMF for Path $path Row $row ".
            "Projection $projection test case\n";
        return ERROR;
    }
    # Remove OMF after test has successfully completed
    unlink ("I1234.omf") if (-e "I1234.omf");

    return SUCCESS;
}

#-----------------------------------------------------------------------------
# Name: update_omf
#
# Description: Updates target path/row in OMF with provided values
#
# Parameters:
#   Path
#   Row 
#
# Returns
#   SUCCESS/ERROR
#-----------------------------------------------------------------------------
sub update_omf
{
    my $omf_path = shift;
    my $omf_row = shift;

    # Create a new OMF from template (-e), changing path to the path provided
    my $sed_cmd = "sed -e \"s/__PATH__/$path/\" -e \"s/__ROW__/$row/\"";
    my $status = system("$sed_cmd I1234.omf.template " . " > I1234.omf");
    if ($status)
    {
        # Error checking in the upper level will provide path row
        print "Error creating OMF file\n";
        return ERROR;
    }
    return SUCCESS;
}

#-----------------------------------------------------------------------------
# Name: verify_omf
#
# Description: Verifies override projection and projection parameters
#              were appended to the omf by comparing newly created OMF with
#              an existing expected OMF
#
# Parameters:
#   Database connection
#   Projection Name
#
# Returns
#   SUCCESS/ERROR
#-----------------------------------------------------------------------------
sub verify_omf
{
    my $dbh = shift;
    my $proj_name = shift;
    my $diff;

    my $omf = "I1234.omf";
    my $expected_omf = $omf.".".$proj_name;

    my $diff_command = "diff $omf $expected_omf";
    my $status = system($diff_command);
    if ($status)
    {
        print "Issues encountered when performing diff on $omf and ".
            "$expected_omf\n";
        return ERROR;
    }
    return SUCCESS;
}
