/*----------------------------------------------------------------------------
NAME:    convert_band

PURPOSE: Conversion wrapper for create_gtiff to convert a band of data from
         L1G to GeoTIFF. Keeps the filename setup separate so create_gtiff
         is kept as in-tact as possible to its reuse source from D1R (L7 IAS).

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "ias_logging.h"
#include "ias_l1g.h"
#include "ias_satellite_attributes.h"
#include "format_and_package.h"
#include "dfp_metadata.h"

int convert_band
(
    L1GIO *l1g_ptr,                         /* I: L1G image pointer */
    const IAS_L1G_FILE_METADATA *l1g_meta,  /* I: L1G file metadata */
    int band_number,                        /* I: L1G band number to convert
                                                  or 0 if quality band */
    IAS_L1G_BAND_METADATA *band_meta,       /* I: L1G metadata for band */
    char filename[MAX_FILE_LENGTH],         /* O: Filename saved for package */
    DFP_METADATA *dfp_meta                  /* O: Filename added for L1G meta */
)
{
    int status;
    int quality_band = 0;
    int band_index;

    /* Set a quality band indicator flag for easier handling of that case */
    if (band_number == 0)
    {
        /* Set the quality band flag and reset the band number to '1' since
           the quality band file has only a single band in it, to simplify
           the create_gtiff() code handling of the band to work with */
        quality_band = 1;
        band_number = 1;
    }

    /* Determine if converting quality band to create the GeoTIFF file name
       properly, since all regular band geotiff filenames are based on the
       band number, but the quality band file will have a special geotiff
       filename that doesn't follow the band-based naming scheme. */
    if (!quality_band)
    {
        /* Setup the GeoTIFF file name for this band */
        status = snprintf(filename, MAX_FILE_LENGTH, "%s_B%d.TIF",
                determine_l1_basename(dfp_meta), band_number);
        if (status < 0 || status >= MAX_FILE_LENGTH)
        {
            IAS_LOG_ERROR("Creating geotiff filename for band %d", band_number);
            return ERROR;
        }

        /* Save the filename for inclusion in the L1G metadata file */
        band_index = ias_sat_attr_convert_band_number_to_index(band_number);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Converting band number %d to band index",
                    band_number);
            return ERROR;
        }
        strcpy(dfp_meta->geotiff_filenames[band_index], filename);
    }
    else
    {
        /* The quality band needs its GeoTIFF file name created
           a bit differently */

        /* Setup GeoTIFF file name for this band */
        status = snprintf(filename, MAX_FILE_LENGTH, "%s_BQA.TIF",
                determine_l1_basename(dfp_meta));
        if (status < 0 || status >= MAX_FILE_LENGTH)
        {
            IAS_LOG_ERROR("Creating geotiff quality band filename");
            return ERROR;
        }

        /* Save the filename for inclusion in the L1G metadata file */
        /* The quality band data is being saved in the DFP metadata geotiff
           filenames array at index IAS_MAX_NBANDS */
        strcpy(dfp_meta->geotiff_filenames[IAS_MAX_NBANDS], filename);
    }

    /* Record a status message */
    if (!quality_band)
    {
        IAS_LOG_INFO("Converting band %d", band_number);
    }
    else
        IAS_LOG_INFO("Converting quality band");

    /* Create the GeoTIFF file */
    status = create_gtiff(l1g_ptr, l1g_meta, filename, band_number, band_meta);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Creating GeoTIFF file %s.", filename);
        return ERROR;
    }

    return SUCCESS;
}
