/*----------------------------------------------------------------------------
NAME:    create_gtiff

PURPOSE: Create the GeoTIFF by copying the data from the L1G input file

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "geotiffio.h"
#include "xtiffio.h"
#include "ias_types.h"
#include "ias_logging.h"
#include "ias_l1g.h"
#include "gctp.h"
#include "ias_geo.h"
#include "format_and_package.h"

#define MAX_BUF_SIZE 10485760  /* 10Mb */

/* Structure used for storing information about a partially created GeoTIFF
 * file. */
typedef struct gtiff_generation_data
{
    TIFF *tif;          /* Pointers necessary for interfacing */
    GTIF *gtif;         /* libtiff and libgeotiff */
    int nlines;         /* number of lines in GeoTIFF image */
    int nsamples;       /* number of samples in GeoTIFF image */
} GTIFF_GENERATION_DATA;

/* Used to convert the zone code from the L1G metadata to a Geo Key macro
   for easy use in setting the ProjectedCSTypeGeoKey */
/* With LDCM, all zones are positive, essentially treating them all like they
   are north. By not using negative zones, we no longer make a distinction
   between N/S (for the zone code geotiff tags and keys). The only thing the
   N/S zone was used for in the past was to mess with the false northing
   definition to force the coordinates used in the southern hemisphere to be
   positive. For LDCM we will instead have negative coordinates in the southern
   hemisphere. */
static int zone_code_geo_key_lut[60] =
{
    PCS_WGS84_UTM_zone_1N,
    PCS_WGS84_UTM_zone_2N,
    PCS_WGS84_UTM_zone_3N,
    PCS_WGS84_UTM_zone_4N,
    PCS_WGS84_UTM_zone_5N,
    PCS_WGS84_UTM_zone_6N,
    PCS_WGS84_UTM_zone_7N,
    PCS_WGS84_UTM_zone_8N,
    PCS_WGS84_UTM_zone_9N,
    PCS_WGS84_UTM_zone_10N,
    PCS_WGS84_UTM_zone_11N,
    PCS_WGS84_UTM_zone_12N,
    PCS_WGS84_UTM_zone_13N,
    PCS_WGS84_UTM_zone_14N,
    PCS_WGS84_UTM_zone_15N,
    PCS_WGS84_UTM_zone_16N,
    PCS_WGS84_UTM_zone_17N,
    PCS_WGS84_UTM_zone_18N,
    PCS_WGS84_UTM_zone_19N,
    PCS_WGS84_UTM_zone_20N,
    PCS_WGS84_UTM_zone_21N,
    PCS_WGS84_UTM_zone_22N,
    PCS_WGS84_UTM_zone_23N,
    PCS_WGS84_UTM_zone_24N,
    PCS_WGS84_UTM_zone_25N,
    PCS_WGS84_UTM_zone_26N,
    PCS_WGS84_UTM_zone_27N,
    PCS_WGS84_UTM_zone_28N,
    PCS_WGS84_UTM_zone_29N,
    PCS_WGS84_UTM_zone_30N,
    PCS_WGS84_UTM_zone_31N,
    PCS_WGS84_UTM_zone_32N,
    PCS_WGS84_UTM_zone_33N,
    PCS_WGS84_UTM_zone_34N,
    PCS_WGS84_UTM_zone_35N,
    PCS_WGS84_UTM_zone_36N,
    PCS_WGS84_UTM_zone_37N,
    PCS_WGS84_UTM_zone_38N,
    PCS_WGS84_UTM_zone_39N,
    PCS_WGS84_UTM_zone_40N,
    PCS_WGS84_UTM_zone_41N,
    PCS_WGS84_UTM_zone_42N,
    PCS_WGS84_UTM_zone_43N,
    PCS_WGS84_UTM_zone_44N,
    PCS_WGS84_UTM_zone_45N,
    PCS_WGS84_UTM_zone_46N,
    PCS_WGS84_UTM_zone_47N,
    PCS_WGS84_UTM_zone_48N,
    PCS_WGS84_UTM_zone_49N,
    PCS_WGS84_UTM_zone_50N,
    PCS_WGS84_UTM_zone_51N,
    PCS_WGS84_UTM_zone_52N,
    PCS_WGS84_UTM_zone_53N,
    PCS_WGS84_UTM_zone_54N,
    PCS_WGS84_UTM_zone_55N,
    PCS_WGS84_UTM_zone_56N,
    PCS_WGS84_UTM_zone_57N,
    PCS_WGS84_UTM_zone_58N,
    PCS_WGS84_UTM_zone_59N,
    PCS_WGS84_UTM_zone_60N
};

/* To help prepare tiff tag and geo key values */
static int get_bit_depth(IAS_DATA_TYPE data_type);
static int get_gtiff_data_type(IAS_DATA_TYPE data_type);

/* To set the tiff tags */
static int set_tifftag_uint16(
        GTIFF_GENERATION_DATA *gtiff_vars,
        int itag,
        uint16 value
        );
static int set_tifftag_uint32(
        GTIFF_GENERATION_DATA *gtiff_vars,
        int itag,
        uint32 value
        );
static int set_tifftag_double_array(
        GTIFF_GENERATION_DATA *gtiff_vars,
        int itag,
        double *values_array,
        int num_values
        );

/* To set the geo keys */
static int set_geo_key_short(
        GTIFF_GENERATION_DATA *gtiff_vars,
        int geo_key,
        int value
        );
static int set_geo_key_double(
        GTIFF_GENERATION_DATA *gtiff_vars,
        int geo_key,
        double value
        );
static int set_geo_key_ascii(
        GTIFF_GENERATION_DATA *gtiff_vars,
        int geo_key,
        const char *value
        );

int create_gtiff
(
    L1GIO *l1g_ptr,                         /* I: L1G image pointer */
    const IAS_L1G_FILE_METADATA *l1g_meta,  /* I: L1G file metadata */
    const char *input_filename,             /* I: L1G band filename to make */
    int band_number,                        /* I: L1G band number converting */
    IAS_L1G_BAND_METADATA *band_meta        /* I: L1G metadata for GeoTIFF */
)
{
    GTIFF_GENERATION_DATA gtiff_vars; /* Gtiff file create values */
    int geo_key_val;            /* GeoTIFF geokey value number */
    double geo_key_val_double;  /* GeoTIFF geokey value number */
    char geo_key_str[50];       /* GeoTIFF geokey string value */
    int num_params;             /* Number parameters in TIFF tag double array */
    double geo_UL_tie_point[6]; /* The GeoTIFF UL corner */
    double geo_pix_scale_xyz[3];/* Projection distance X/Y/Z */
    uint8 *buffer = NULL;       /* Buffer to store data being copied */
    int bit_depth;              /* Return value from get_bit_depth to check */
    int bytes_wide;             /* Number of bytes per line of data */
    int lines_to_copy;          /* Number of lines to copy at a time */
    int lines_remaining;        /* Number of lines left to copy */
    int error_flag;             /* Used to help during copying of data */
    int status;
    L1G_BAND_IO *l1g_band;      /* File handle for band to convert */
    IAS_DATA_TYPE data_type;    /* IAS enum value for L1G data type */
    int num_scas;
    int num_lines;
    int num_samples;

    /* Open the band */
    l1g_band = ias_l1g_open_band(l1g_ptr, band_number, &data_type,
            &num_scas, &num_lines, &num_samples);
    if (l1g_band == NULL)
    {
        IAS_LOG_ERROR("Opening L1G band %d of image %s", band_number,
                get_l1g_input_filename());
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }

    /* Initialize the gtiff_vars structure */
    /* Clear the size of image values */
    gtiff_vars.nlines = 0;
    gtiff_vars.nsamples = 0;

    /* Open TIFF file */
    gtiff_vars.tif = XTIFFOpen(input_filename, "w");
    if (gtiff_vars.tif == NULL)
    {
        IAS_LOG_ERROR("Opening TIFF file %s.", input_filename);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }

    /* Copy TIFF file to GeoTIFF file */
    gtiff_vars.gtif = GTIFNew(gtiff_vars.tif);
    if (gtiff_vars.gtif == NULL)
    {
        IAS_LOG_ERROR("Copying TIFF to GeoTIFF file %s.", input_filename);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }

    /* Set the size of the image */
    if (set_tifftag_uint32(&gtiff_vars, TIFFTAG_IMAGEWIDTH, num_samples)
            != SUCCESS)
    {
        IAS_LOG_ERROR("Setting GeoTIFF width (%d)", num_samples);
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }
    /* Save this value in our structure */
    gtiff_vars.nsamples = num_samples;

    /* Set the size of the image */
    if (set_tifftag_uint32(&gtiff_vars, TIFFTAG_IMAGELENGTH, num_lines)
            != SUCCESS)
    {
        IAS_LOG_ERROR("Setting GeoTIFF length (%d)", num_lines);
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }
    /* Save this value in our structure */
    gtiff_vars.nlines = num_lines;

    /* Set the number of bands */
    /* (Only a single band is put in each GeoTIFF) */
    if (set_tifftag_uint16(&gtiff_vars, TIFFTAG_SAMPLESPERPIXEL, 1)
            != SUCCESS)
    {
        IAS_LOG_ERROR("Setting GeoTIFF number of bands (1)");
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }

    /* Set the bit depth */
    bit_depth = get_bit_depth(data_type);
    if (bit_depth == ERROR)
    {
        IAS_LOG_ERROR("Unexpected data type (%d)", data_type);
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }
    if (set_tifftag_uint16(&gtiff_vars, TIFFTAG_BITSPERSAMPLE, bit_depth)
            != SUCCESS)
    {
        IAS_LOG_ERROR("Setting GeoTIFF bit depth (%u)", bit_depth);
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }

    /* Set the data type */
    /* Setting TIFFTAG_SAMPLEFORMAT has a dependency on TIFFTAG_BITSPERSAMPLE
       having already been set so get_bit_depth() has already been executed,
       which error checked data_type. */
    if (set_tifftag_uint16(&gtiff_vars, TIFFTAG_SAMPLEFORMAT,
                get_gtiff_data_type(data_type)) != SUCCESS)
    {
        IAS_LOG_ERROR("Setting GeoTIFF data type (%u)",
                get_gtiff_data_type(data_type));
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }

    /* Set the compression */
    if (set_tifftag_uint16(&gtiff_vars, TIFFTAG_COMPRESSION, COMPRESSION_NONE)
            != SUCCESS)
    {
        IAS_LOG_ERROR("Setting GeoTIFF compression format (%u)",
                COMPRESSION_NONE);
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }

    /* Set the photometric type */
    if (set_tifftag_uint16(&gtiff_vars, TIFFTAG_PHOTOMETRIC,
                PHOTOMETRIC_MINISBLACK) != SUCCESS)
    {
        IAS_LOG_ERROR("Setting GeoTIFF photometric format (%u)",
                PHOTOMETRIC_MINISBLACK);
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }

    /* Set the planar configuration */
    if (set_tifftag_uint16(&gtiff_vars, TIFFTAG_PLANARCONFIG,
                PLANARCONFIG_CONTIG) != SUCCESS)
    {
        IAS_LOG_ERROR("Setting GeoTIFF planar configuration (%u)",
                PLANARCONFIG_CONTIG);
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }

    /* Set the rows per strip */
    /* (There is only ever one row per strip) */
    if (set_tifftag_uint32(&gtiff_vars, TIFFTAG_ROWSPERSTRIP, 1) != SUCCESS)
    {
        IAS_LOG_ERROR("Setting GeoTIFF rows per strip (1)");
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }

    /* Set the upper left corner as a tie point so we know the origin */
    /* Raster-to-model tie point pair (I, J, K, X, Y, Z) where (I, J, K) is
       the point at location (I, J) in raster space with pixel value K and
       (X, Y, Z) is a vector in model space. Requires that K and Z are zero. */
    /* Set this array to 0.0, 0.0, 0.0, ULX, ULY, 0.0 */
    geo_UL_tie_point[0] = 0.0;
    geo_UL_tie_point[1] = 0.0;
    geo_UL_tie_point[2] = 0.0;
    geo_UL_tie_point[3] = band_meta->upper_left_x;
    geo_UL_tie_point[4] = band_meta->upper_left_y;
    geo_UL_tie_point[5] = 0.0;
    num_params = 6;
    if (set_tifftag_double_array(&gtiff_vars, TIFFTAG_GEOTIEPOINTS,
                geo_UL_tie_point, num_params) != SUCCESS)
    {
        IAS_LOG_ERROR("Setting GeoTIFF upper left corner");
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }

    /* Set the pixel scale X/Y */
    /* The size of raster pixel spacing in the model space units when the
       raster space can be embedded in the model space coordinate system
       without rotation. */
    geo_pix_scale_xyz[0] = band_meta->projection_distance_x;
    geo_pix_scale_xyz[1] = band_meta->projection_distance_y;
    /* The third (z) array item is the elevation component. It's 0 since
       the image doesn't contain elevation info. */
    geo_pix_scale_xyz[2] = 0;
    num_params = 3;
    if (set_tifftag_double_array(&gtiff_vars, TIFFTAG_GEOPIXELSCALE,
                geo_pix_scale_xyz, num_params) != SUCCESS)
    {
        IAS_LOG_ERROR("Setting GeoTIFF pixel scale");
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }

    /* Set the Geo keys - some are the same no matter which projection and
       some are different based on the projection */

    /* Make sure we have an expected projection */
    if ((l1g_meta->projection_code != UTM) && 
        (l1g_meta->projection_code != PS) &&
        (l1g_meta->projection_code != ALBERS))
    {
        IAS_LOG_ERROR("Unsupported projection code (%d)",
                l1g_meta->projection_code);
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }

    /* Set model type Geo Key */
    if (set_geo_key_short(&gtiff_vars, GTModelTypeGeoKey, ModelTypeProjected)
            != SUCCESS)
    {
        IAS_LOG_ERROR("Setting GeoTIFF model type (%d)", ModelTypeProjected);
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }

    /* Set raster pixel type Geo Key */
    if (set_geo_key_short(&gtiff_vars, GTRasterTypeGeoKey, RasterPixelIsPoint)
            != SUCCESS)
    {
        IAS_LOG_ERROR("Setting GeoTIFF raster pixel format (%d)",
                RasterPixelIsPoint);
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }

    /* Set the projection citation Geo Key, depending on projection */
    if (l1g_meta->projection_code == UTM)
    {
        /* Prepare the projection citation Geo Key string for UTM */
        /* See the note at top of file on the zone_code_geo_key_lut[]
           definition about why all zones are positive ("N"). */
        status = snprintf(geo_key_str, sizeof(geo_key_str),
                "UTM Zone %d N with WGS84", l1g_meta->zone_code);
        if (status < 0 || status >= sizeof(geo_key_str))
        {
            IAS_LOG_ERROR("Creating projection citation UTM zone Geo Key");
            GTIFFree(gtiff_vars.gtif);
            XTIFFClose(gtiff_vars.tif);
            ias_l1g_close_band(l1g_band);
            return ERROR;
        }
    }
    else if (l1g_meta->projection_code == PS)
    {
        /* Prepare the projection citation Geo Key string for PS */
        strcpy(geo_key_str, "PS         WGS84");
    }
    else if (l1g_meta->projection_code == ALBERS)
    {
        /* Prepare the projection citation Geo Key string for ALBERS */
        strcpy(geo_key_str, "AEA        WGS84");
    }
    else
    {
        /* Should not get here, as this was previously error checked */
        IAS_LOG_ERROR("Unexpected projection code found (%d)",
                l1g_meta->projection_code);
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }
    if (set_geo_key_ascii(&gtiff_vars, GTCitationGeoKey, geo_key_str)
            != SUCCESS)
    {
        IAS_LOG_ERROR("Setting GeoTIFF projection citation (%s)", geo_key_str);
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }

    /* Set linear units Geo Key; it should always be meters, but still verify */
    if (strcmp(l1g_meta->projection_units, "METERS") != 0)
    {
        IAS_LOG_ERROR("Setting GeoTIFF linear units; not meters");
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }
    if (set_geo_key_short(&gtiff_vars, GeogLinearUnitsGeoKey, Linear_Meter)
            != SUCCESS)
    {
        IAS_LOG_ERROR("Setting GeoTIFF linear units (%d)", Linear_Meter);
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }

    /* Set angular units Geo Key */
    if (set_geo_key_short(&gtiff_vars, GeogAngularUnitsGeoKey, Angular_Degree)
            != SUCCESS)
    {
        IAS_LOG_ERROR("Setting GeoTIFF angular units (%d)", Angular_Degree);
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }

    /* Set the projected coordinate system Geo Key, depending on projection */
    if (l1g_meta->projection_code == UTM)
    {
        if ((l1g_meta->zone_code < 1) || (l1g_meta->zone_code > 60))
        {
            IAS_LOG_ERROR("Unexpected UTM zone code (%d)", l1g_meta->zone_code);
            GTIFFree(gtiff_vars.gtif);
            XTIFFClose(gtiff_vars.tif);
            ias_l1g_close_band(l1g_band);
            return ERROR;
        }
        geo_key_val = zone_code_geo_key_lut[l1g_meta->zone_code - 1];
    }
    else if (l1g_meta->projection_code == PS ||
        l1g_meta->projection_code == ALBERS)
    {
        geo_key_val = KvUserDefined;
    }
    else
        geo_key_val = KvUndefined;
    if (set_geo_key_short(&gtiff_vars, ProjectedCSTypeGeoKey, geo_key_val)
            != SUCCESS)
    {
        IAS_LOG_ERROR("Setting GeoTIFF projected coordinate system (%d)",
                geo_key_val);
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }

    /* More Geo Keys that are specific to Polar Stereographic (PS) or ALBERS;
       mostly projection parameters */
    if (l1g_meta->projection_code == PS ||
        l1g_meta->projection_code == ALBERS)
    {
        /* Set datum Geo Key */
        if (set_geo_key_short(&gtiff_vars, GeographicTypeGeoKey, GCS_WGS_84)
                != SUCCESS)
        {
            IAS_LOG_ERROR("Setting GeoTIFF datum (%d)", GCS_WGS_84);
            GTIFFree(gtiff_vars.gtif);
            XTIFFClose(gtiff_vars.tif);
            ias_l1g_close_band(l1g_band);
            return ERROR;
        }

        /* Set projection Geo Key */
        if (set_geo_key_short(&gtiff_vars, ProjectionGeoKey, KvUserDefined)
                != SUCCESS)
        {
            IAS_LOG_ERROR("Setting GeoTIFF projection (%d)", KvUserDefined);
            GTIFFree(gtiff_vars.gtif);
            XTIFFClose(gtiff_vars.tif);
            ias_l1g_close_band(l1g_band);
            return ERROR;
        }

        /* Set projection linear units Geo Key */
        /* (The units were validated previously) */
        if (set_geo_key_short(&gtiff_vars, ProjLinearUnitsGeoKey,
                    Linear_Meter) != SUCCESS)
        {
            IAS_LOG_ERROR("Setting GeoTIFF projection linear units (%d)",
                    Linear_Meter);
            GTIFFree(gtiff_vars.gtif);
            XTIFFClose(gtiff_vars.tif);
            ias_l1g_close_band(l1g_band);
            return ERROR;
        }

        /* Set latitude origin Geo Key */
        if (ias_geo_convert_dms2deg(l1g_meta->projection_parameters[5],
                    &geo_key_val_double, "LAT") != SUCCESS)
        {
            IAS_LOG_ERROR("Setting GeoTIFF latitude origin (%lf)",
                    geo_key_val_double);
            GTIFFree(gtiff_vars.gtif);
            XTIFFClose(gtiff_vars.tif);
            ias_l1g_close_band(l1g_band);
            return ERROR;
        }
        if (set_geo_key_double(&gtiff_vars, ProjNatOriginLatGeoKey,
                    geo_key_val_double) != SUCCESS)
        {
            IAS_LOG_ERROR("Setting GeoTIFF latitude origin (%lf)",
                    geo_key_val_double);
            GTIFFree(gtiff_vars.gtif);
            XTIFFClose(gtiff_vars.tif);
            ias_l1g_close_band(l1g_band);
            return ERROR;
        }

        /* Set false easting Geo Key */
        geo_key_val_double = l1g_meta->projection_parameters[6];
        if (set_geo_key_double(&gtiff_vars, ProjFalseEastingGeoKey,
                    geo_key_val_double) != SUCCESS)
        {
            IAS_LOG_ERROR("Setting GeoTIFF false easting (%lf)",
                    geo_key_val_double);
            GTIFFree(gtiff_vars.gtif);
            XTIFFClose(gtiff_vars.tif);
            ias_l1g_close_band(l1g_band);
            return ERROR;
        }

        /* Set false northing Geo Key */
        geo_key_val_double = l1g_meta->projection_parameters[7];
        if (set_geo_key_double(&gtiff_vars, ProjFalseNorthingGeoKey,
                    geo_key_val_double) != SUCCESS)
        {
            IAS_LOG_ERROR("Setting GeoTIFF false northing (%lf)",
                    geo_key_val_double);
            GTIFFree(gtiff_vars.gtif);
            XTIFFClose(gtiff_vars.tif);
            ias_l1g_close_band(l1g_band);
            return ERROR;
        }
    }

    /* PS specific parameters */
    if (l1g_meta->projection_code == PS)
    {
        if (set_geo_key_short(&gtiff_vars, ProjCoordTransGeoKey,
                        CT_PolarStereographic) != SUCCESS)
        {
            IAS_LOG_ERROR("Setting GeoTIFF coordinate transformation for "
                "projection %d", l1g_meta->projection_code);
            GTIFFree(gtiff_vars.gtif);
            XTIFFClose(gtiff_vars.tif);
            ias_l1g_close_band(l1g_band);
            return ERROR;
        }

        /* Set longitude straight vertical pole Geo Key */
        if (ias_geo_convert_dms2deg(l1g_meta->projection_parameters[4],
                    &geo_key_val_double, "LON") != SUCCESS)
        {
            IAS_LOG_ERROR("Setting GeoTIFF meridian projection pole (%lf)",
                    geo_key_val_double);
            GTIFFree(gtiff_vars.gtif);
            XTIFFClose(gtiff_vars.tif);
            ias_l1g_close_band(l1g_band);
            return ERROR;
        }
        if (set_geo_key_double(&gtiff_vars, ProjStraightVertPoleLongGeoKey,
                    geo_key_val_double) != SUCCESS)
        {
            IAS_LOG_ERROR("Setting GeoTIFF meridian projection pole (%lf)",
                    geo_key_val_double);
            GTIFFree(gtiff_vars.gtif);
            XTIFFClose(gtiff_vars.tif);
            ias_l1g_close_band(l1g_band);
            return ERROR;
        }
    }

    /* ALBERS specific parameters */
    if (l1g_meta->projection_code == ALBERS)
    {
        if (set_geo_key_short(&gtiff_vars, ProjCoordTransGeoKey,
                        CT_AlbersEqualArea) != SUCCESS)
        {
            IAS_LOG_ERROR("Setting GeoTIFF coordinate transformation for "
                "projection %d", l1g_meta->projection_code);
            GTIFFree(gtiff_vars.gtif);
            XTIFFClose(gtiff_vars.tif);
            ias_l1g_close_band(l1g_band);
            return ERROR;
        }

        /* Set standard parallel 1 Geo Key */
        if (ias_geo_convert_dms2deg(l1g_meta->projection_parameters[2],
                    &geo_key_val_double, "LAT") != SUCCESS)
        {
            IAS_LOG_ERROR("Setting GeoTIFF Std Parallel 1 (%lf)",
                    geo_key_val_double);
            GTIFFree(gtiff_vars.gtif);
            XTIFFClose(gtiff_vars.tif);
            ias_l1g_close_band(l1g_band);
            return ERROR;
        }
        if (set_geo_key_double(&gtiff_vars, ProjStdParallel1GeoKey ,
                    geo_key_val_double) != SUCCESS)
        {
            IAS_LOG_ERROR("Setting GeoTIFF Std Parallel 1 (%lf)",
                    geo_key_val_double);
            GTIFFree(gtiff_vars.gtif);
            XTIFFClose(gtiff_vars.tif);
            ias_l1g_close_band(l1g_band);
            return ERROR;
        }

        /* Set standard parallel 2 Geo Key */
        if (ias_geo_convert_dms2deg(l1g_meta->projection_parameters[3],
                    &geo_key_val_double, "LAT") != SUCCESS)
        {
            IAS_LOG_ERROR("Setting GeoTIFF Std Parallel 2 (%lf)",
                    geo_key_val_double);
            GTIFFree(gtiff_vars.gtif);
            XTIFFClose(gtiff_vars.tif);
            ias_l1g_close_band(l1g_band);
            return ERROR;
        }
        if (set_geo_key_double(&gtiff_vars, ProjStdParallel2GeoKey ,
                    geo_key_val_double) != SUCCESS)
        {
            IAS_LOG_ERROR("Setting GeoTIFF Std Parallel 2 (%lf)",
                    geo_key_val_double);
            GTIFFree(gtiff_vars.gtif);
            XTIFFClose(gtiff_vars.tif);
            ias_l1g_close_band(l1g_band);
            return ERROR;
        }

        /* Set proj nat origin long Geo Key */
        if (ias_geo_convert_dms2deg(l1g_meta->projection_parameters[4],
                    &geo_key_val_double, "LON") != SUCCESS)
        {
            IAS_LOG_ERROR("Setting GeoTIFF meridian projection pole (%lf)",
                    geo_key_val_double);
            GTIFFree(gtiff_vars.gtif);
            XTIFFClose(gtiff_vars.tif);
            ias_l1g_close_band(l1g_band);
            return ERROR;
        }
        if (set_geo_key_double(&gtiff_vars, ProjNatOriginLongGeoKey,
                    geo_key_val_double) != SUCCESS)
        {
            IAS_LOG_ERROR("Setting GeoTIFF meridian projection pole (%lf)",
                    geo_key_val_double);
            GTIFFree(gtiff_vars.gtif);
            XTIFFClose(gtiff_vars.tif);
            ias_l1g_close_band(l1g_band);
            return ERROR;
        }
    }

    /* Copy the actual band data/imagery */

    /* Allocate local buffer for a band of image data */
    bytes_wide = gtiff_vars.nsamples * (bit_depth / 8);
    lines_to_copy = (int)(MAX_BUF_SIZE / bytes_wide);

    /* For malloc(), total size is size of each element (size of a line)
       times the number of elements (number of lines) */
    buffer = (uint8 *)malloc(lines_to_copy * bytes_wide);
    if (buffer == NULL)
    {
        /* Could not successfully allocate the buffer */
        IAS_LOG_ERROR("Allocating band read buffer");
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }

    /* Copy the data */
    error_flag = 0;
    lines_remaining = gtiff_vars.nlines;

    /* Copy however many lines at a time the buffer allows and keep doing
       this until we've copied the entire image from the L1G to the GeoTIFF. */
    do
    {
        int start_line;     /* Line to start reading at */
        int start_sample;   /* Sample to start reading at */
        int num_lines;      /* Number of lines to read */
        int num_samples;    /* Number of samples to read */
        int first_line;     /* For this copy of a block, the start and end */
        int last_line;      /* Lines that bound the copy area */

        /* Make sure we don't copy too much the last time through */
        if (lines_to_copy > lines_remaining)
            lines_to_copy = lines_remaining;

        /* Set the offsets indicating the items to be copied */
        start_line = gtiff_vars.nlines - lines_remaining;
        start_sample = 0;
        num_lines = lines_to_copy;
        num_samples = gtiff_vars.nsamples;

        /* Set the values used for copying into the GeoTIFF file */
        first_line = start_line;
        last_line = first_line + num_lines - 1;

        /* Copy some lines of data (possibly all) */
        if (ias_l1g_read_image(l1g_band, 0, start_line, start_sample, num_lines,
                    num_samples, buffer) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading data from input file '%s'", input_filename);
            error_flag = 1;
        }
        else
        {
            int line;

            /* Loop through the lines, reading and writing each from the L1G
               out to the GeoTIFF */
            for (line = first_line; (line <= last_line); line++)
            {
                if (!TIFFWriteScanline(gtiff_vars.tif,
                            buffer + ((line - first_line) * bytes_wide), line,
                            0))
                {
                    IAS_LOG_ERROR("Writing data to GeoTIFF file, line %d",
                            line);
                    error_flag = 1;
                    break;
                }
            }
        }

        /* Update the number of lines remaining to be copied */
        lines_remaining = lines_remaining - num_lines;

    } while ((lines_remaining > 0) && !error_flag);
    /* Break out of loop when all lines copied or if there was an error */

    /* Free the copy buffer */
    free(buffer);
    buffer = NULL;

    /* Close the source L1G band */
    ias_l1g_close_band(l1g_band);
    l1g_band = NULL;

    /* Return failure if the copy failed */
    if (error_flag)
    {
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        return ERROR;
    }

    /* Since only one band per GeoTIFF file, go ahead and close the file
       and return SUCCESS as long as everything successful to this point */

    /* Write keys and close file */
    if (!GTIFWriteKeys(gtiff_vars.gtif))
    {
        IAS_LOG_ERROR("Writing GeoTIFF keys");
        GTIFFree(gtiff_vars.gtif);
        XTIFFClose(gtiff_vars.tif);
        return ERROR;
    }

    /* Note: GTIFFree and XTIFFClose do not return anything */
    GTIFFree(gtiff_vars.gtif);
    XTIFFClose(gtiff_vars.tif);

    return SUCCESS;
}

/* Get the number of bits in data type from the IAS_DATA_TYPE */
static int get_bit_depth
(
    IAS_DATA_TYPE data_type     /* I: the IAS enum type of the L1G type */
)
{
    /* DFP only should encounter 8, 16, or 32 bit data
       otherwise signal error */
    if (data_type == IAS_BYTE)
        return 8;
    else if ((data_type == IAS_I2) || (data_type == IAS_UI2))
        return 16;
    else if (data_type == IAS_R4)
        return 32;
    else
        return ERROR;
}

/* Get the GeoTIFF data type correlating to the IAS_DATA_TYPE */
static int get_gtiff_data_type
(
    IAS_DATA_TYPE data_type     /* I: the IAS enum type of the L1G type */
)
{
    /* Due to error checking after get_bit_depth, there are only a couple
       data types expected at this point */
    if (data_type == IAS_BYTE)
        return SAMPLEFORMAT_UINT;
    else if (data_type == IAS_I2)
        return SAMPLEFORMAT_INT;
    else if (data_type == IAS_UI2)
        return SAMPLEFORMAT_UINT;
    else
        return SAMPLEFORMAT_IEEEFP;
}

/* Set a given tiff tag that is of uint16 data type */
static int set_tifftag_uint16
(
    GTIFF_GENERATION_DATA *gtiff_vars,  /* I/O: GeoTIFF variables structure */
    int itag,                           /* I: The GeoTIFF tag */
    uint16 value                        /* I: the value to set the tag to */
)
{
    ttag_t tag;                 /* GeoTIFF tag */

    tag = (ttag_t)itag;

    /* Error if TIFF tag not valid or failure setting TIFF tag */
    if ((!itag) || (!value) || (!TIFFSetField(gtiff_vars->tif, tag, value)))
        return ERROR;

    return SUCCESS;
}

/* Set a given tiff tag that is of uint32 data type */
static int set_tifftag_uint32
(
    GTIFF_GENERATION_DATA *gtiff_vars,  /* I/O: GeoTIFF variables structure */
    int itag,                           /* I: The GeoTIFF tag */
    uint32 value                        /* I: the value to set the tag to */
)
{
    ttag_t tag;                 /* GeoTIFF tag */

    tag = (ttag_t)itag;

    /* Error if TIFF tag not valid or failure setting TIFF tag */
    if ((!itag) || (!value) || (!TIFFSetField(gtiff_vars->tif, tag, value)))
        return ERROR;

    return SUCCESS;
}

/* Set a given tiff tag that is an array of doubles */
static int set_tifftag_double_array
(
    GTIFF_GENERATION_DATA *gtiff_vars,  /* I/O: GeoTIFF variables structure */
    int itag,                           /* I: The GeoTIFF tag */
    double *values_array,               /* I: values array to set the tag to */
    int num_values                      /* I: number values in array */
)
{
    ttag_t tag;                 /* GeoTIFF tag */

    tag = (ttag_t)itag;

    /* Error if the TIFF tag wasn't valid or error setting TIFF tag */
    if ((!itag) || (!TIFFSetField(gtiff_vars->tif, tag, num_values,
                    values_array)))
    {
        return ERROR;
    }

    return SUCCESS;
}

/* Set a given tiff tag that is of short data type */
static int set_geo_key_short
(
    GTIFF_GENERATION_DATA *gtiff_vars,  /* I/O: GeoTIFF variables structure */
    int geo_key,                        /* I: The GeoTIFF geo key */
    int value                           /* I: the value to set the key to */
)
{
    /* Error if the Geo Key wasn't valid, if our value to set is not valid,
       or if there is an error setting the Geo Key return error */
    if ((!geo_key) || (!value) || (!GTIFKeySet(gtiff_vars->gtif, geo_key,
                    TYPE_SHORT, 1, value)))
    {
        return ERROR;
    }

    return SUCCESS;
}

/* Set a given geo key that is of double data type */
static int set_geo_key_double
(
    GTIFF_GENERATION_DATA *gtiff_vars,  /* I/O: GeoTIFF variables structure */
    int geo_key,                        /* I: The GeoTIFF geo key */
    double value                        /* I: the value to set the key to */
)
{
    /* Error if the Geo Key wasn't valid, if our value to set is not valid,
       or if there is an error setting the Geo Key return error */
    if ((!geo_key) || (!GTIFKeySet(gtiff_vars->gtif, geo_key, TYPE_DOUBLE, 1,
                    value)))
    {
        return ERROR;
    }

    return SUCCESS;
}

/* Set a given geo key that is an ASCII value */
static int set_geo_key_ascii
(
    GTIFF_GENERATION_DATA *gtiff_vars,  /* I/O: GeoTIFF variables structure */
    int geo_key,                        /* I: The GeoTIFF geo key */
    const char *value                   /* I: the value to set the key to */
)
{
    /* If the Geo Key wasn't valid, if our value to set is not valid, or if
       there is an error setting the Geo Key return error - the value is */
    if ((!geo_key) || (!value) || (!GTIFKeySet(gtiff_vars->gtif, geo_key,
                    TYPE_ASCII, 0, value)))
    {
        return ERROR;
    }

    return SUCCESS;
}
