/*----------------------------------------------------------------------------
NAME:    create_packages

PURPOSE: TAR/Gzip each product to create a package and optionally generate
         MD5 checksum file for each package.
         Update the OMF with the name of each package generated.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>                         /* getcwd() */
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "ias_types.h"
#include "ias_satellite_attributes.h"       /* IAS_MAX_NBANDS */
#include "ias_logging.h"
#include "ias_l0r.h"
#include "ias_parm.h"
#include "format_and_package.h"
#include "config.h"
#define ADDITIONAL_L0R_FILES 3   /* Potential number of additional files that
                                    could be included in the L0 product files,
                                    beyond the band files (the metadata file,
                                    ancillary data file, and the MD5 checksum
                                    file). Need complete file list for packaging
                                    routine. */
#define MAX_NUM_L0R_FILES (IAS_L0R_MAX_BAND_FILES + ADDITIONAL_L0R_FILES)

/* Maximum length (plus healthy buffer) of string length needed for any of
   the L1G system commands - to keep our memory use at reasonable needs. */
#define L1G_SYS_CMD_LENGTH  1000

/* The /bin/sh shell that is called by the system call sets the return status
   to the return status of the last application run in the pipe instead of the
   first non-zero status encountered in the pipe as most people expect.  To
   properly detect errors with piped commands, the PIPESTATUS environment
   variable needs to be checked for each command.  This code uses piped
   commands for creating the tar file and the md5sum in a set of piped commands
   like the following (not all parameters shown):
    tar | tee filename | md5sum
   So, three members of PIPESTATUS need to be checked.  This happens in a 
   couple places in the routine, so the following define sets up the 
   status check.  Appending this string to the system command will combine
   all three status flags into the single flag returned to the system call. */
#define CHECK_PIPE_STATUS "; test ${PIPESTATUS[0]} -eq 0 " \
                    "-a ${PIPESTATUS[1]} -eq 0 -a ${PIPESTATUS[2]} -eq 0"

/* Local function prototypes */
static int get_file_size_and_date
(
    const char *filename,                   /* I: Name of file */
    long long *filesize,                    /* O: Size of file in bytes */
    char *filedate                          /* O: Date of file */
);

static int get_product_size
(
    char file_list[][MAX_FILE_LENGTH],     /* I: List of files in product */
    int number_files,                      /* I: Number of files in product */
    const char *file_path,                 /* I: Directory of product files */
    long long *product_size                /* O: Product size in bytes */
);

static int populate_l0r_file_list
(
    char l0r_file_list[][MAX_FILE_LENGTH], /* O: List of files for L0R image */
    int *number_l0r_files,                 /* O: Number of files for L0R image*/
    int *number_l0r_bands                  /* O: Number of bands for L0R image*/
);

static int read_and_fix_checksum_file
(
    const char *checksum_filename,  /* I: The checksum file to read */
    const char *target_filename,    /* I: The file that was checksummed */
    char *checksum                  /* O: The checksum string read */
);

int create_packages
(
    char l1g_file_list[][MAX_FILE_LENGTH],  /* I: L1G product files list */
    int number_l1g_files,                   /* I: Number files in files list */
    DFP_METADATA *dfp_meta                  /* I/O: dfp metadata structure */
)
{
    /* The list of L0R files for packaging commands */
    char l0r_file_list[MAX_NUM_L0R_FILES][MAX_FILE_LENGTH];
    int want_l0r_package;
    /* Contains multiple file names, so extend the size some */
    char l0r_sys_cmd[PATH_MAX * 3] = {""};
    int number_l0r_files = 0;
    int want_l1g_package;
    char l1g_sys_cmd[L1G_SYS_CMD_LENGTH] = {""};
    int want_md5_checksum;
    int i;
    int status;

    /* Setup local copies of the flags for easier use */
    want_l0r_package = get_l0r_package_flag();
    want_l1g_package = get_l1g_package_flag();
    want_md5_checksum = get_md5_checksum_flag();

    /* Initialize the command strings so we have something to later check
       against */
    l0r_sys_cmd[0] = l1g_sys_cmd[0] = '\0';

    /* Set up the L0R packaging file names */
    if (want_l0r_package)
    {
        /* The L0R basename is essentially the L0R ID, the base part of all
           the L0R file names. So we can tack right on to the end of it. */

        /* Prepare the package filename */
        status = snprintf(dfp_meta->l0r_package_file_name,
                sizeof(dfp_meta->l0r_package_file_name),
                "%s_L0R.tar.gz", get_l0r_basename());
        if (status < 0 || status >= sizeof(dfp_meta->l0r_package_file_name))
        {
            IAS_LOG_ERROR("Creating L0R package filename");
            return ERROR;
        }

        if(want_md5_checksum)
        {
            /* Prepare the checksum filename */
            status = snprintf(dfp_meta->l0r_checksum_file_name,
                    sizeof(dfp_meta->l0r_checksum_file_name), "%s_L0R_MD5.txt",
                    get_l0r_basename());
            if (status < 0 || status 
                    >= sizeof(dfp_meta->l0r_checksum_file_name))
            {
                IAS_LOG_ERROR("Creating L0R checksum filename");
                return ERROR;
            }
        }

        if (populate_l0r_file_list(l0r_file_list, &number_l0r_files,
                    &dfp_meta->l0r_band_count) != SUCCESS)
        {
            IAS_LOG_ERROR("Getting list of L0R files");
            return ERROR;
        }
    }

    /* Set up the L1G packaging file names */
    if (want_l1g_package)
    {
        const char *basename = determine_l1_basename(dfp_meta);
        /* Prepare the package filename */
        status = snprintf(dfp_meta->l1_package_file_name,
                sizeof(dfp_meta->l1_package_file_name),
                "%s.tar.gz", basename);
        if (status < 0 || status >= sizeof(dfp_meta->l1_package_file_name))
        {
            IAS_LOG_ERROR("Creating L1G package filename");
            return ERROR;
        }

        if(want_md5_checksum)
        {
            /* Prepare the checksum filename */
            status = snprintf(dfp_meta->l1_checksum_file_name,
                    sizeof(dfp_meta->l1_checksum_file_name), "%s_MD5.txt",
                    basename);
            if (status < 0 || status >= sizeof(dfp_meta->l1_checksum_file_name))
            {
                IAS_LOG_ERROR("Creating L1G checksum filename");
                return ERROR;
            }
        }
    }

    /* Build the packaging system command to package the L0R product with
       options specified */
    if (want_l0r_package)
    {
        /* Contains multiple file names, so extend the size some */
        char working_cmd[PATH_MAX * 3];
        int l0r_has_path = 0;

        /* Determine if the L0R files are local or not */
        if (strcmp(get_l0r_path(), "") != 0)
            l0r_has_path = 1;

        if (!want_md5_checksum)
        {
            if (l0r_has_path)
            {
                /* The L0R is not in local directory, so use tar -C option
                   so files will be placed into local directory upon
                   decompression, per requirements. */
                status = snprintf(l0r_sys_cmd, sizeof(l0r_sys_cmd),
                        "tar --use-compress-program=%s -chf %s -C %s",
                        DEFAULT_ZIP_PROG, dfp_meta->l0r_package_file_name,
                        get_l0r_path());
                if (status < 0 || status >= sizeof(l0r_sys_cmd))
                {
                    IAS_LOG_ERROR("Creating L0R packaging system command "
                            "with non-local L0R");
                    return ERROR;
                }
            }
            else
            {
                status = snprintf(l0r_sys_cmd, sizeof(l0r_sys_cmd),
                        "tar --use-compress-program=%s -chf %s ",
                        DEFAULT_ZIP_PROG, dfp_meta->l0r_package_file_name);
                if (status < 0 || status >= sizeof(l0r_sys_cmd))
                {
                    IAS_LOG_ERROR("Creating L0R packaging system command "
                            "with local L0R");
                    return ERROR;
                }
            }

            for(i = 0; i < number_l0r_files; i++)
            {
                /* Tack on the next file name */
                strcpy(working_cmd, l0r_sys_cmd);
                status = snprintf(l0r_sys_cmd, sizeof(l0r_sys_cmd), "%s %s",
                        working_cmd, l0r_file_list[i]);
                if (status < 0 || status >= sizeof(l0r_sys_cmd))
                {
                    IAS_LOG_ERROR("Creating L0R packaging system command "
                            "while adding file %s", l0r_file_list[i]);
                    return ERROR;
                }
            }
        }
        /* MD5 checksum was requested, so add that to command string */
        else
        {
            if (l0r_has_path)
            {
                /* The L0R is not in local directory, so use tar -C option
                   so files will be placed into local directory upon
                   decompression, per requirements. */
                status = snprintf(l0r_sys_cmd, sizeof(l0r_sys_cmd),
                        "tar --use-compress-program=%s -ch -C %s",
                        DEFAULT_ZIP_PROG, get_l0r_path());
                if (status < 0 || status >= sizeof(l0r_sys_cmd))
                {
                    IAS_LOG_ERROR("Creating L0R packaging system command "
                            "with MD5 checksums and non-local L0R");
                    return ERROR;
                }
            }
            else
            {
                status = snprintf(l0r_sys_cmd, sizeof(l0r_sys_cmd),
                        "tar --use-compress-program=%s -ch ", DEFAULT_ZIP_PROG);
                if (status < 0 || status >= sizeof(l0r_sys_cmd))
                {
                    IAS_LOG_ERROR("Creating L0R packaging system command "
                            "with MD5 checksums and local L0R");
                    return ERROR;
                }
            }

            for(i = 0; i < number_l0r_files; i++)
            {
                /* Tack on the next file name */
                strcpy(working_cmd, l0r_sys_cmd);
                status = snprintf(l0r_sys_cmd, sizeof(l0r_sys_cmd), "%s %s",
                        working_cmd, l0r_file_list[i]);
                if (status < 0 || status >= sizeof(l0r_sys_cmd))
                {
                    IAS_LOG_ERROR("Creating L0R packaging system command "
                            "with MD5 checksums while adding file %s",
                            l0r_file_list[i]);
                    return ERROR;
                }
            }

            /* Tee the output of the TAR command to send it to a file and
               continue feeding it through stdout so we can pipe the output
               also to md5sum.  If the number of commands in the pipe change,
               the CHECK_PIPE_STATUS define needs to be adjusted. */
            strcpy(working_cmd, l0r_sys_cmd);
            status = snprintf(l0r_sys_cmd, sizeof(l0r_sys_cmd),
                    "%s | tee %s | md5sum - > %s" CHECK_PIPE_STATUS,
                    working_cmd, dfp_meta->l0r_package_file_name, 
                    dfp_meta->l0r_checksum_file_name);
            if (status < 0 || status >= sizeof(l0r_sys_cmd))
            {
                IAS_LOG_ERROR("Creating L0R packaging system command "
                        "while adding md5sum to command string");
                return ERROR;
            }
        }
    }

    /* Build the packaging system command to package the L1G product with
       options specified */
    if (want_l1g_package)
    {
        char working_cmd[L1G_SYS_CMD_LENGTH];

        if (!want_md5_checksum)
        {
            /* Tar/gzip the locally generated GeoTIFF L1G band files, the
               GeoTIFF quality band file (if created), and the ASCII metadata
               file into a single package. */
            status = snprintf(l1g_sys_cmd, sizeof(l1g_sys_cmd),
                    "tar --use-compress-program=%s -chf %s", DEFAULT_ZIP_PROG,
                    dfp_meta->l1_package_file_name);
            if (status < 0 || status >= sizeof(l1g_sys_cmd))
            {
                IAS_LOG_ERROR("Creating L1G packaging system command");
                return ERROR;
            }

            for(i = 0; i < number_l1g_files; i++)
            {
                /* Tack on the next file name */
                strcpy(working_cmd, l1g_sys_cmd);
                status = snprintf(l1g_sys_cmd, sizeof(l1g_sys_cmd), "%s %s",
                        working_cmd, l1g_file_list[i]);
                if (status < 0 || status >= sizeof(l1g_sys_cmd))
                {
                    IAS_LOG_ERROR("Creating L1G packaging system command "
                            "while adding file %s", l1g_file_list[i]);
                    return ERROR;
                }
            }
        }
        /* MD5 checksum was requested, so add that to command string */
        else
        {
            status = snprintf(l1g_sys_cmd, sizeof(l1g_sys_cmd),
                    "tar --use-compress-program=%s -ch", DEFAULT_ZIP_PROG);
            if (status < 0 || status >= sizeof(l1g_sys_cmd))
            {
                IAS_LOG_ERROR("Creating L0R packaging system command "
                        "with MD5 checksums and local L0R");
                return ERROR;
            }

            for(i = 0; i < number_l1g_files; i++)
            {
                /* Tack on the next file name */
                strcpy(working_cmd, l1g_sys_cmd);
                status = snprintf(l1g_sys_cmd, sizeof(l1g_sys_cmd), "%s %s",
                        working_cmd, l1g_file_list[i]);
                if (status < 0 || status >= sizeof(l1g_sys_cmd))
                {
                    IAS_LOG_ERROR("Creating L1G packaging system command "
                            "with MD5 checksums while adding file %s",
                            l1g_file_list[i]);
                    return ERROR;
                }
            }

            /* Tee the output of the TAR command to send it to a file and
               continue feeding it through stdout so we can pipe the output
               also to md5sum.  If the number of commands in the pipe change,
               the CHECK_PIPE_STATUS define needs to be adjusted. */
            strcpy(working_cmd, l1g_sys_cmd);
            status = snprintf(l1g_sys_cmd, sizeof(l1g_sys_cmd),
                    "%s | tee %s | md5sum - > %s" CHECK_PIPE_STATUS,
                    working_cmd, dfp_meta->l1_package_file_name, 
                    dfp_meta->l1_checksum_file_name);
            if (status < 0 || status >= sizeof(l1g_sys_cmd))
            {
                IAS_LOG_ERROR("Creating L1G packaging system command "
                        "while adding md5sum to command string");
                return ERROR;
            }
        }
    }

    /* Now execute the command built above */
    if (strlen(l0r_sys_cmd) > 0)
    {
        IAS_LOG_INFO("Preparing L0R product package");

        /* Execute the command */
        if (system(l0r_sys_cmd) != 0)
        {
            IAS_LOG_ERROR("Executing L0R packaging command");
            return ERROR;
        }

        if (get_file_size_and_date(dfp_meta->l0r_package_file_name, 
                    &dfp_meta->l0r_package_file_size,
                    dfp_meta->l0r_package_file_date) != SUCCESS)
        {
            IAS_LOG_ERROR("Getting L0r file size and date");
            return ERROR;
        }

        if (get_product_size(l0r_file_list, number_l0r_files, get_l0r_path(),
                    &dfp_meta->l0r_product_file_size) != SUCCESS)
        {
            IAS_LOG_ERROR("Getting L0R product file size");
            return ERROR;
        }

        if (want_md5_checksum)
        {
            /* The file name inside the checksum file is '-', so fix to
               give it the TAR gzip file name. Save checksum for l0r package */
            if (read_and_fix_checksum_file(dfp_meta->l0r_checksum_file_name,
                        dfp_meta->l0r_package_file_name,
                        dfp_meta->l0r_package_file_checksum) != SUCCESS)
            {
                IAS_LOG_ERROR("Reading l0r checksum file.");
                return ERROR;
            }

            if (get_file_size_and_date(dfp_meta->l0r_checksum_file_name, 
                        &dfp_meta->l0r_checksum_file_size,
                        dfp_meta->l0r_checksum_file_date) != SUCCESS)
            {
                IAS_LOG_ERROR("Getting L0r checksum file size and date");
                return ERROR;
            }
        }
    }

    if (strlen(l1g_sys_cmd) > 0)
    {
        IAS_LOG_INFO("Preparing L1G product package");

        /* Execute the command */
        if (system(l1g_sys_cmd) != 0)
        {
            IAS_LOG_ERROR("Executing L1g packaging command");
            return ERROR;
        }

        if (get_file_size_and_date(dfp_meta->l1_package_file_name, 
                    &dfp_meta->l1_package_file_size,
                    dfp_meta->l1_package_file_date) != SUCCESS)
        {
            IAS_LOG_ERROR("Getting L1 file size and date");
            return ERROR;
        }

        if (get_product_size(l1g_file_list, number_l1g_files, ".",
                    &dfp_meta->l1_product_file_size) != SUCCESS)
        {
            IAS_LOG_ERROR("Getting L1 product file size");
            return ERROR;
        }

        if (want_md5_checksum)
        {
            /* The file name inside the checksum file is '-', so fix to
               give it the TAR gzip file name. Save checksum for l1g package */
            if (read_and_fix_checksum_file(dfp_meta->l1_checksum_file_name,
                        dfp_meta->l1_package_file_name,
                        dfp_meta->l1_package_file_checksum) != SUCCESS)
            {
                IAS_LOG_ERROR("Reading l1g checksum file.");
                return ERROR;
            }

            if (get_file_size_and_date(dfp_meta->l1_checksum_file_name, 
                        &dfp_meta->l1_checksum_file_size,
                        dfp_meta->l1_checksum_file_date) != SUCCESS)
            {
                IAS_LOG_ERROR("Getting L0r checksum file size");
                return ERROR;
            }
        }
    }

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:    get_file_size_and_date

PURPOSE: Gets the size of a file in bytes and the date in YYYY/MM/DD format

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int get_file_size_and_date
(
    const char *filename,              /* I: Name of file */
    long long *filesize,               /* O: Size of file in bytes */
    char *filedate                     /* O: Date of file */
)
{
    struct stat st;
    struct tm *my_local_time;

    if (stat(filename, &st) != 0)
    {
        IAS_LOG_ERROR("Getting file statistics for %s", filename);
        return ERROR;
    }

    *filesize = st.st_size;

    my_local_time = localtime(&st.st_mtime);

    if (my_local_time == NULL)
    {
        IAS_LOG_ERROR("Getting file date for %s", filename);
        return ERROR;
    }

    sprintf(filedate,"%04d/%02d/%02d", my_local_time->tm_year + 1900,
            my_local_time->tm_mon + 1, my_local_time->tm_mday);

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:    get_product_size

PURPOSE: Given a file list, calculate the total file size.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int get_product_size
(
    char file_list[][MAX_FILE_LENGTH],    /* I: list of files in product */
    int number_files,                     /* I: number of files in product */
    const char *base_directory,           /* I: directory of product files */
    long long *product_size
)
{
    int i;
    int status;
    long long total_file_size = 0;
    char local_file_name[PATH_MAX];
    struct stat st;

    for (i=0; i < number_files; i++)
    {
        status = snprintf(local_file_name, sizeof(local_file_name) ,"%s/%s",
                base_directory, file_list[i]);
        if (status < 0 || status >= sizeof(local_file_name))
        {
            IAS_LOG_ERROR("Setting path for %s", file_list[i]);
            return ERROR;
        }
        if (stat(local_file_name, &st) != 0)
        {
            IAS_LOG_ERROR("Getting filesize for %s", local_file_name);
            return ERROR;
        }
        else
        {
            total_file_size = total_file_size + st.st_size;
        }
    }

    *product_size = total_file_size;
    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:    populate_l0r_file_list

PURPOSE: Given a directory to a L0R product, gathers all file names

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int populate_l0r_file_list
(
    char l0r_file_list[][MAX_FILE_LENGTH],
                                    /* O: List of files for the L0R image */
    int *number_l0r_files,          /* O: Number of files for this L0R image */
    int *number_l0r_bands           /* O: Number of bands for L0R image */
)
{
    L0RIO *l0r_ptr;
    IAS_L0R_FILE_METADATA l0r_meta;     /* Structure for L0R file names */
    int local_number_l0r_bands = 0;     /* Local number of bands */
    int local_number_l0r_files = 0;     /* Local number of l0r files */
    int i;
    int status;

    /* Start the file and band counts at zero */
    *number_l0r_files = 0;
    *number_l0r_bands = 0;

    /* Initialize access to the L0R data */
    l0r_ptr =  ias_l0r_open(get_l0r_basename(), get_l0r_path(), 0);
    if (!l0r_ptr)
    {
        IAS_LOG_ERROR("Opening image %s%s", get_l0r_path(),
                get_l0r_basename());
        return ERROR;
    }

    /* Set read access for the L0R metadata */
    if(ias_l0r_open_metadata(l0r_ptr, IAS_READ) != SUCCESS)
    {
        IAS_LOG_ERROR("Accessing L0R metadata");
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Get the L0R file names */
    if(ias_l0r_get_file_metadata(l0r_ptr, &l0r_meta) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading L0R file names");
        ias_l0r_close_metadata(l0r_ptr);
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Close the read access to the L0R metadata */
    if(ias_l0r_close_metadata(l0r_ptr) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing L0R metadata");
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Close all access to the L0R and free any resources */
    if(ias_l0r_close(l0r_ptr) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing L0R image %s%s", get_l0r_path(),
                get_l0r_basename());
        return ERROR;
    }

    /* Copy the file names from the L0R metadata structure to our local list */
    /* Get the checksum file file name, if it exists for this product */
    if (strlen(l0r_meta.checksum_file.name) > 0)
    {
        status = snprintf(l0r_file_list[local_number_l0r_files++],
                sizeof(l0r_file_list[0]), "%s", l0r_meta.checksum_file.name);
        if (status < 0 || status >= sizeof(l0r_file_list[0]))
        {
            IAS_LOG_ERROR("Storing L0R checksum filename");
            return ERROR;
        }
    }

    /* Get the metadata file file name, if it exists for this product */
    if (strlen(l0r_meta.metadata_file.name) > 0)
    {
        status = snprintf(l0r_file_list[local_number_l0r_files++],
                sizeof(l0r_file_list[0]), "%s", l0r_meta.metadata_file.name);
        if (status < 0 || status >= sizeof(l0r_file_list[0]))
        {
            IAS_LOG_ERROR("Storing L0R metadata filename");
            return ERROR;
        }
    }

    /* Get the ancillary file file name, if it exists for this product */
    if (strlen(l0r_meta.ancillary_file.name) > 0)
    {
        status = snprintf(l0r_file_list[local_number_l0r_files++],
                sizeof(l0r_file_list[0]), "%s", l0r_meta.ancillary_file.name);
        if (status < 0 || status >= sizeof(l0r_file_list[0]))
        {
            IAS_LOG_ERROR("Storing L0R ancillary filename");
            return ERROR;
        }
    }

    /* Get the file names of the band files that are present for this product */
    for (i = 0; i < IAS_L0R_MAX_BAND_FILES; i++)
    {
        if (strlen(l0r_meta.band_files[i].name) > 0)
        {
            status = snprintf(l0r_file_list[local_number_l0r_files++],
                    sizeof(l0r_file_list[0]), "%s",
                    l0r_meta.band_files[i].name);
            if (status < 0 || status >= sizeof(l0r_file_list[0]))
            {
                IAS_LOG_ERROR("Storing L0R band %d filename", i);
                return ERROR;
            }
            local_number_l0r_bands++;
        }
    }

    if (local_number_l0r_files == 0)
    {
        /* Didn't find any files */
        IAS_LOG_ERROR("Retrieving list of L0R files, based on L0R %s%s",
                get_l0r_path(), get_l0r_basename());
        return ERROR;
    }

    *number_l0r_files = local_number_l0r_files;
    *number_l0r_bands = local_number_l0r_bands;

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:    read_and_fix_checksum_file

PURPOSE: Reads checksum from previously generated checksum file. Writes out
         new checksum file replacing the '-' with the correct file name.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int read_and_fix_checksum_file
(
    const char *checksum_filename,  /* I: The checksum file to read */
    const char *target_filename,    /* I: The file that was checksummed */
    char *checksum                  /* O: The checksum string read */
)
{
    FILE *file;                     /* Checksum file handle */
    FILE *out_file;                 /* Temp checksum file handle */
    int status;                     /* Return value from fscanf */
    char output_filename[] = "checksum_file.tmp";  /* temp checksum file */

    file = fopen(checksum_filename, "r");
    if (file == NULL)
    {
        IAS_LOG_ERROR("Opening checksum file for reading: %s",
                checksum_filename);
        fclose(file);
        return ERROR;
    }

    /* Checksum is first field in file */
    status = fscanf(file, "%32s", checksum);
    if (status == 0)
    {
        IAS_LOG_ERROR("Reading checksum from file: %s",
                checksum_filename);
        fclose(file);
        return ERROR;
    }

    fclose(file);

    out_file  = fopen(output_filename, "w");
    if (out_file == NULL)
    {
        IAS_LOG_ERROR("Opening temp checksum file for writing: %s",
                output_filename);
        fclose(file);
        return ERROR;
    }

    if (fprintf(out_file,"%s  %s\n", checksum, target_filename) < 0)
    {
        IAS_LOG_ERROR("Writing temp checksum file: %s", output_filename);
        fclose(out_file);
        return ERROR;
    }

    fclose(out_file);

    if (rename(output_filename, checksum_filename) != 0)
    {
        IAS_LOG_ERROR("Renaming %s to %s", checksum_filename, output_filename);
        return ERROR;
    }

    return SUCCESS;
}
