#ifndef DFP_METADATA_H
#define DFP_METADATA_H

#include <limits.h>
#include "ias_const.h"
#include "ias_satellite_attributes.h"
#include "ias_sensor_model.h"

/* Set a reasonable size limit for maximum filename lengths of files created */
#define MAX_FILE_LENGTH (IAS_L0R_SCENE_ID_LENGTH + 50)
#define SCENE_CENTER_SIZE 30
#define DATE_LENGTH 12
#define CHECKSUM_LENGTH 33
#define PRODUCT_ID_LENGTH 41
#define PROJECT_LENGTH 32

typedef struct dfp_metadata
{
    const struct IAS_CPF_TIRS_THERMAL_CONSTANTS *tirs_thermal_constants;    
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char product_request_id[IAS_PRODUCT_REQUEST_ID_SIZE];
    char product_id[PRODUCT_ID_LENGTH + 1];
    char l0r_package_file_name[MAX_FILE_LENGTH];
    char l0r_package_file_checksum[CHECKSUM_LENGTH];
    char l0r_package_file_date[DATE_LENGTH];
    char l0r_checksum_file_name[MAX_FILE_LENGTH];
    char l0r_checksum_file_date[DATE_LENGTH];
    long long l0r_product_file_size;
    long long l0r_package_file_size;
    long long l0r_checksum_file_size;
    int l0r_band_count;
    char l1_package_file_name[MAX_FILE_LENGTH];
    char l1_package_file_checksum[CHECKSUM_LENGTH];
    char l1_package_file_date[DATE_LENGTH];
    char l1_checksum_file_name[MAX_FILE_LENGTH];
    char l1_checksum_file_date[DATE_LENGTH];
    long long l1_package_file_size;
    long long l1_checksum_file_size;
    long long l1_product_file_size;
    char ang_filename[PATH_MAX];
    char dem_filename[PATH_MAX];
    char elevation_source[8];
    char spacecraft_id[IAS_SPACECRAFT_NAME_LENGTH + 1];
    char sensor_id[IAS_SENSOR_ID_SIZE];
    int target_wrs_path;
    int target_wrs_row;
    char l0r_oli_present[2];  /* "Y" or "N" */
    char l0r_tirs_present[2]; /* "Y" or "N" */
    char l1_oli_present[2];   /* "Y" or "N" */
    char l1_tirs_present[2];  /* "Y" or "N" */
    char scene_center_time[SCENE_CENTER_SIZE];
    double ul_lat_lon[2];   /* order is [0]: lat, [1]: long */
    double ur_lat_lon[2];   /* order is [0]: lat, [1]: long */
    double ll_lat_lon[2];   /* order is [0]: lat, [1]: long */
    double lr_lat_lon[2];   /* order is [0]: lat, [1]: long */
    int pan_lines;          /* number of lines for band 8 */
    int pan_samples;        /* number of samples for band 8 */
    int reflective_lines;   /* number of lines for bands 1-7, 9 */
    int reflective_samples; /* number of samples for bands 1-7, 9 */
    int thermal_lines;      /* number of lines for bands 10-11 */
    int thermal_samples;    /* number of samples for bands 10-11 */
    /* filenames are indexed by band index; quality band at [IAS_MAX_NBANDS] */
    char geotiff_filenames[IAS_MAX_NBANDS + 1][MAX_FILE_LENGTH];
    char bpf_filename_oli[PATH_MAX];
    char bpf_filename_tirs[PATH_MAX];
    char cpf_filename[PATH_MAX];
    char rlut_filename[PATH_MAX];
    double cloud_cover_score;
    double snow_ice_score;
    double cloud_cover_land_score;
    int ground_control_points_model;
    double geometric_rmse_model;
    int ground_control_points_verify;
    int ground_control_points_version; /* Overall GCP version number */
    double geometric_rmse_verify;
    double model_y_rmse;
    double model_x_rmse;
    char l1r_filename[PATH_MAX];
    char grid_filename[PATH_MAX];
    double grid_cell_size_pan;          /* band 8 */
    double grid_cell_size_reflective;   /* bands 1-7, 9 */
    double grid_cell_size_thermal;      /* bands 10-11 */
    char orientation[9];                /* "NORTH-UP" or "NOMINAL" */
    char map_projection[4];
    char resample_type[18]; /* CUBIC_CONVOLUTION, NEAREST_NEIGHBOR, or
                               BILINEAR */
    int collection_number;
    char ssm_position_status[IAS_SENSOR_SSM_POSITION_STATUS_SIZE];
    IAS_SSM_POSITION_QUALITY ssm_position_quality;
    int band_saturated[IAS_MAX_NBANDS]; /* Band saturation flags */
    int have_saturated_bands_flag; /* Flag to indicated there are any bands with
                                      saturation */
    int process_band[IAS_MAX_NBANDS]; /* Bands to process by index */
    char stray_light_correction_source[IAS_COLLECT_TYPE_SIZE];
    char is_oli_lower_12_bits_flag; /* OLI truncation mode (1 = Lower 12 bits,
                                                            0 = Upper 12 bits)*/
} DFP_METADATA;

/* Much of this is L1G metadata found when opening each band. Save since
   some of it is needed for the L1G metadata file generation. */
typedef struct l1g_band_metadata
{
    int num_lines;
    int num_samples;
    IAS_L1G_BAND_METADATA band_meta;
    int band_valid;                     /* flag used by write_metadata() */
} L1G_BAND_METADATA;

#endif
