/**************************************************************************
NAME:          format_and_package (main)

PURPOSE:
Main application to format and package an output product.

Uses the OMF, L0R, L1R, L1Gt/L1T, and quality band to prepare and package
the L0R and/or L1T product(s). Converts the L1Gt/L1T and quality band image
files in the L1G format to GeoTIFF format. Extracts metadata from the OMF and
the L1R and L1G images to populate the necessary GeoTIFF tags as well as an
ASCII metadata file for the L1G.

RETURNS:
Type = int
Value         Description
-----         -----------
EXIT_SUCCESS  Successful completion
EXIT_FAILURE  Operation failed
***************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include "ias_types.h"
#include "ias_logging.h"
#include "ias_satellite_attributes.h"       /* IAS_MAX_NBANDS */
#include "ias_l0r_constants.h"
#include "ias_l0r_structures.h"
#include "gctp.h"
#include "ias_geo.h"
#include "ias_parm.h"
#include "format_and_package.h"
#include "dfp_metadata.h"
#include "ias_angle_gen.h"
#include "sync_directory.h"

#define ADDITIONAL_L1G_FILES 3   /* potential number of additional files that
                                    could be included in the L1G product files,
                                    beyond the band files (the quality band,
                                    the L1G metadata file, and Angle Coeff
                                    file). Need complete file list for 
                                    packaging routine. */
#define MAX_NUM_L1G_FILES (IAS_MAX_NBANDS + ADDITIONAL_L1G_FILES)

/*----------------------------------------------------------------------------
NAME:    validate_l1g_meta

PURPOSE: Checks for anything in the L1G metadata that is relevant to this
         application and should be a fatal error - do now before spending
         too much processing time.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int validate_l1g_meta
(
    L1GIO *l1g_ptr,                         /* I: File pointer for L1G image */
    const IAS_L1G_FILE_METADATA *l1g_meta,  /* I: Metadata to read from L1G */
    DFP_METADATA *dfp_meta                  /* O: Local L1G metadata struct */
)
{
    int band_index;

    if (l1g_meta->spheroid_code != WGS84_SPHEROID)
    {
        IAS_LOG_ERROR("Unexpected spheroid code in L1G: %d",
                l1g_meta->spheroid_code);
        return ERROR;
    }
    if (l1g_meta->projection_code == UTM)
        strcpy(dfp_meta->map_projection, "UTM");
    else if (l1g_meta->projection_code == PS)
        strcpy(dfp_meta->map_projection, "PS");
    else if (l1g_meta->projection_code == ALBERS)
        strcpy(dfp_meta->map_projection, "AEA");
    else
    {
        IAS_LOG_ERROR("Unexpected map projection code in L1G: %d",
                l1g_meta->projection_code);
        return ERROR;
    }

    /* Set the DFP resampling option value */
    if (strcmp(l1g_meta->resample_type, "CC") == 0)
    {
        if (snprintf(dfp_meta->resample_type, sizeof(dfp_meta->resample_type),
            "CUBIC_CONVOLUTION") < 0)
        {
            IAS_LOG_ERROR("Specifying resample type");
            return ERROR;
        }
    }
    else if (strcmp(l1g_meta->resample_type, "NN") == 0)
    {
        if (snprintf(dfp_meta->resample_type, sizeof(dfp_meta->resample_type),
            "NEAREST_NEIGHBOR") < 0)
        {
            IAS_LOG_ERROR("Specifying resample type");
            return ERROR;
        }
    }
    else if (strcmp(l1g_meta->resample_type, "BI") == 0)
    {
        /* This is not really supported now but may be in the future */
        if (snprintf(dfp_meta->resample_type, sizeof(dfp_meta->resample_type),
            "BILINEAR") < 0)
        {
            IAS_LOG_ERROR("Specifying resample type");
            return ERROR;
        }
    }
    else
    {
        IAS_LOG_ERROR("Invalid RESAMPLING_OPTION %s", l1g_meta->resample_type);
        return ERROR;
    }

    /* Initialize flags for the L1G metadata */
    strcpy(dfp_meta->l1_oli_present, "N");
    strcpy(dfp_meta->l1_tirs_present, "N");

    /* Initialize the L1G metadata band filenames;
       IAS_MAX_NBANDS+1 size for the quality band also (if present) */
    for (band_index = 0; band_index <= IAS_MAX_NBANDS; band_index++)
        strcpy(dfp_meta->geotiff_filenames[band_index], "");

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:    update_omf

PURPOSE: Adds the L0R & L1G package TARball filenames to the OMF file for
         either or both that were generated.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int update_omf
(
    const char *l0r_package_filename,   /* I: L0R package file name or NULL */
    const char *l1g_package_filename    /* I: L1G package file name or NULL */
)
{
    int write_count;
    int write_status;

    /* Setup the structure for the library for the parameters to add to OMF */
    IAS_PARM_DECLARE_TABLE(omf_param_out_table, 2);

    IAS_PARM_ADD_STRING(omf_param_out_table, L0R_PACKAGE,
            "L0R package file", IAS_PARM_REQUIRED,
            0, NULL, 0, NULL, (char *)l0r_package_filename,
            sizeof(l0r_package_filename), 1);
    IAS_PARM_SET_COUNT_WRITE(omf_param_out_table, L0R_PACKAGE, 1);

    IAS_PARM_ADD_STRING(omf_param_out_table, L1G_PACKAGE,
            "L1G package file", IAS_PARM_REQUIRED,
            0, NULL, 0, NULL, (char *)l1g_package_filename,
            sizeof(l1g_package_filename), 1);
    IAS_PARM_SET_COUNT_WRITE(omf_param_out_table, L1G_PACKAGE, 1);

    /* Use the Parameter File I/O library to update the OMF file */
    if ((l0r_package_filename != NULL) && (l1g_package_filename != NULL))
    {
        write_count = IAS_PARM_GET_TABLE_SIZE(omf_param_out_table);
        write_status = ias_parm_update(get_omf_filename(), omf_param_out_table,
                write_count);
    }
    else if ((l0r_package_filename != NULL) && (l1g_package_filename == NULL))
    {
        write_count = 1;
        write_status = ias_parm_update(get_omf_filename(),
                &omf_param_out_table[0], write_count);
    }
    else if ((l0r_package_filename == NULL) && (l1g_package_filename != NULL))
    {
        write_count = 1;
        write_status = ias_parm_update(get_omf_filename(),
                &omf_param_out_table[1], write_count);
    }
    else
        write_status = ERROR;

    if (write_status != SUCCESS)
    {
        IAS_LOG_ERROR("Updating OMF file: %s", get_omf_filename());
        return ERROR;
    }

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:    cleanup_old_output

PURPOSE: Deletes old output files named with the Product ID. This is needed
         because the product ID contains the generation date which will
         change.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int cleanup_old_output
(
    const char * prod_id /* I: The files with this product ID (wildcarding
                               generation date) will be cleaned up. */
)
{
    const int GEN_DATE_LOCATION = 3; /* This constant indicates which underscore
                  separated "group" the generation date can be found in. If the
                  product ID formats changes this may need to as well. */

    char delete_command[1024];
    char prod_id_wildcarded[PRODUCT_ID_LENGTH + 1];
    char * gen_date = NULL;
    int index;
    int status;
    int underscores_located = 0;

    /* Bypass this function if we were not requested to name with the
     * Product ID (files named with Scene ID will be overwritten )*/
    if (! get_named_with_prod_id())
        return SUCCESS;

    strncpy(prod_id_wildcarded, prod_id, PRODUCT_ID_LENGTH);
    /* Force last byte to \0, in case we ran out of room */
    prod_id_wildcarded[PRODUCT_ID_LENGTH] = '\0';

    /* Find the location of the underscore before the generation date */
    gen_date = strpbrk(prod_id_wildcarded, "_");
    for (underscores_located = 0; underscores_located < GEN_DATE_LOCATION &&
         gen_date != NULL; underscores_located++)
    {
        gen_date = strpbrk(gen_date + 1, "_");
    }
    if (gen_date == NULL)
    {
        IAS_LOG_ERROR("Failed to find generation date group");
        return ERROR;
    }

    /* Replace the 8 characters after the found underscore with '?' */
    for (index = 1; index <= 8; index++)
        gen_date[index] = '?';


    /* Build and execute the delete command */
    status = snprintf(delete_command, sizeof(delete_command),
        "/bin/rm -f %s*TIF %s*.tar.gz %s*.txt",
        prod_id_wildcarded,prod_id_wildcarded,prod_id_wildcarded);
    if (status < 0 || status >= sizeof(delete_command))
    {
        IAS_LOG_ERROR("Failed to build delete command");
        return ERROR;
    }

    if (system(delete_command) != 0)
    {
        IAS_LOG_ERROR("Failed to execute delete command");
        return ERROR;
    }

    return SUCCESS;
}

int main(int argc, char *argv[])
{
    DFP_METADATA dfp_meta;
    IAS_L1G_FILE_METADATA l1g_meta;
    IAS_L0R_SCENE_METADATA l0r_scene_meta;
    IAS_L0R_INTERVAL_METADATA l0r_int_meta;
    L1G_BAND_METADATA l1g_band_meta[IAS_MAX_NBANDS];
    IAS_CPF *cpf = NULL;
    const IAS_CPF_FILE_ATTRIBUTES *cpf_file_atrrib = NULL;
    char l1g_file_list[MAX_NUM_L1G_FILES][MAX_FILE_LENGTH];
    int control_band_index;         /* Band to use for band metadata */
    int l1g_file_count = 0;
    int process_quality_band = 0;   /* Flag to process this band or not */
    int status;                     /* Var for capturing function statuses */
    int l0r_meta_available;         /* Flag if L0R meta available; relevant in
                                       case of L1G product only requested
                                       (not L1G package) */
    time_t rawtime;              /* used to determine current date */
    struct tm *current_time;    /* A pointer to store the return value from
                                   gmtime */
    struct tm our_current_time; /* A local copy of the current_time struct
                                   so it doesn't get messed with */

    /* Prepare log file */
    status = ias_log_initialize("format_and_package");
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the logging library");
        exit(EXIT_FAILURE);
    }

    /* Verify the parameter filename was provided */
    if (argc < 2)
    {
        IAS_LOG_ERROR("Usage format_and_package <parameter file>");
        exit(EXIT_FAILURE);
    }

    /* Start with a clean dfp_meta structure */
    memset(&dfp_meta, 0, sizeof(dfp_meta));

    /* Read the parameters from the input parameter file and OMF */
    status = read_parameters(argv[1], &dfp_meta, &l0r_scene_meta,
            &l0r_int_meta, &l0r_meta_available);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading input parameters");
        exit(EXIT_FAILURE);
    }

    IAS_LOG_INFO("DMS Format & Package started");

    if (strcmp(get_quality_band_filename(), "") != 0)
    {
        /* Set the quality band processing flag if the filename was provided */
        process_quality_band = 1;
    }

    /* Setup the current date */
    rawtime = time(0);
    if (rawtime == ((time_t)-1))
    {
        IAS_LOG_ERROR("Getting current time error: %s", strerror(errno));
        return ERROR;
    }
    current_time = gmtime(&rawtime);
    if (current_time == NULL)
    {
        IAS_LOG_ERROR("Call to gmtime failed");
        return ERROR;
    }
    our_current_time = *current_time;

    /* If L1G product (GeoTIFF) or package requested, convert it to a
       GeoTIFF and generate the metadata file. Do the GeoTIFF conversion if
       the L1G package was requested, no matter if the L1G product flag is
       set or not. */
    if (get_l1g_product_flag() || get_l1g_package_flag())
    {
        L1GIO *l1g_ptr;
        char l1g_meta_filename[MAX_FILE_LENGTH];
        const char *l1g_filename;
        int band_list_index;
        int band_index;
        int band_number;
        IAS_DATA_TYPE data_type;
        int num_scas;

        if (strlen(get_sun_angle_corrected_l1g_input_filename()) > 0)
        {
            IAS_LOG_INFO("Overiding normal L1G with the Sun Angle corrected "
                "L1G");
            l1g_filename = get_sun_angle_corrected_l1g_input_filename();
        }
        else
        {
            l1g_filename = get_l1g_input_filename();
        }

        /* Open the L1G image for metadata and imagery access */
        l1g_ptr = ias_l1g_open_image(l1g_filename, IAS_READ);
        if (l1g_ptr == NULL)
        {
            IAS_LOG_ERROR("Opening image %s", l1g_filename);
            exit(EXIT_FAILURE);
        }

        status = ias_l1g_get_file_metadata(l1g_ptr, &l1g_meta);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Retrieving file metadata for %s", l1g_filename);
            ias_l1g_close_image(l1g_ptr);
            exit(EXIT_FAILURE);
        }

        status = validate_l1g_meta(l1g_ptr, &l1g_meta, &dfp_meta);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Validating and preparing L1G metadata for file %s",
                    l1g_filename);
            ias_l1g_close_image(l1g_ptr);
            exit(EXIT_FAILURE);
        }

        /* Read the CPF */
        cpf = ias_cpf_read(dfp_meta.cpf_filename);
        if (!cpf)
        {
            IAS_LOG_ERROR("Opening CPF file %s", dfp_meta.cpf_filename );
            ias_l1g_close_image(l1g_ptr);
            exit(EXIT_FAILURE);
        }

        /* Get the focal plane parameter group from the CPF */
        dfp_meta.tirs_thermal_constants = 
            ias_cpf_get_tirs_thermal_constants(cpf);
        if (!dfp_meta.tirs_thermal_constants)
        {
            IAS_LOG_ERROR("Getting TIRS thermal constants from CPF %s",
                dfp_meta.cpf_filename);
            ias_l1g_close_image(l1g_ptr);
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }

        /* Get the collection number from the CPF file atrributes */
        cpf_file_atrrib = ias_cpf_get_file_attributes(cpf);
        if (!cpf_file_atrrib)
        {
            IAS_LOG_ERROR("Getting file atrributes from CPF %s",
                dfp_meta.cpf_filename);
            ias_l1g_close_image(l1g_ptr);
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }
        dfp_meta.collection_number = cpf_file_atrrib->collection_number;

        /* Assume that neither OLI nor TIRS level 1 bands are present */
        strcpy(dfp_meta.l1_oli_present, "N");
        strcpy(dfp_meta.l1_tirs_present, "N");

        /* Collect some information for all bands needed for L1G metadata
           creation */
        for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
        {
            l1g_band_meta[band_index].band_valid = 0;

            band_number = ias_sat_attr_convert_band_index_to_number(band_index);
            if (band_number == ERROR)
            {
                IAS_LOG_ERROR("Converting band index %d to band number",
                        band_index);
                ias_l1g_close_image(l1g_ptr);
                ias_cpf_free(cpf);
                exit(EXIT_FAILURE);
            }

            if (ias_l1g_is_band_present(l1g_ptr, band_number))
            {
                IAS_SENSOR_ID sensor_id;

                status = ias_l1g_get_band_size(l1g_ptr, band_number, &data_type,
                        &num_scas, &l1g_band_meta[band_index].num_lines,
                        &l1g_band_meta[band_index].num_samples);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Getting band information for band %d",
                            band_number);
                    ias_l1g_close_image(l1g_ptr);
                    ias_cpf_free(cpf);
                    exit(EXIT_FAILURE);
                }
                if (num_scas != 1)
                {
                    IAS_LOG_ERROR("SCA-separated image not supported");
                    ias_l1g_close_image(l1g_ptr);
                    ias_cpf_free(cpf);
                    exit(EXIT_FAILURE);
                }

                status = ias_l1g_get_band_metadata(l1g_ptr, band_number,
                        &l1g_band_meta[band_index].band_meta);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Retrieving L1G band metadata for band %d of "
                            "image %s", band_number, l1g_filename);
                    ias_l1g_close_image(l1g_ptr);
                    ias_cpf_free(cpf);
                    exit(EXIT_FAILURE);
                }
                l1g_band_meta[band_index].band_valid = 1;

                /* Set the correct sensor as being present in the output
                   product */
                sensor_id = ias_sat_attr_get_sensor_id_from_band_number(
                        band_number);
                if (sensor_id == IAS_OLI)
                    strcpy(dfp_meta.l1_oli_present, "Y");
                else if (sensor_id == IAS_TIRS)
                    strcpy(dfp_meta.l1_tirs_present, "Y");
            }
        }

        /* Build the product ID to use now that enough information is collected
           in the dfp_metadata */
        status = build_product_id(&dfp_meta, &l0r_scene_meta, &l1g_meta,
                &our_current_time, dfp_meta.product_id);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to build product ID");
            ias_l1g_close_image(l1g_ptr);
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }

        /* Remove any previously created products that might have been created
           by a previous run */
        status = cleanup_old_output(dfp_meta.product_id);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to clean old output.");
            ias_l1g_close_image(l1g_ptr);
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }

        /* Convert on a band-by-band basis; loop over all bands requested in
           the parameter file bands list (which contains all bands if no bands
           list was given) */
        for (band_list_index = 0; band_list_index < get_l1g_nbands();
                band_list_index++)
        {
            const int *band_list;
            band_list = get_l1g_bands();

            band_number = band_list[band_list_index];
            band_index = ias_sat_attr_convert_band_number_to_index(band_number);
            if (band_index == ERROR)
            {
                IAS_LOG_ERROR("Converting band number %d to band index",
                        band_number);
                ias_l1g_close_image(l1g_ptr);
                ias_cpf_free(cpf);
                exit(EXIT_FAILURE);
            }

            /* Convert this L1G band if it is in the L1G image. Otherwise,
               issue a warning. */
            if (ias_l1g_is_band_present(l1g_ptr, band_number))
            {
                status = convert_band(l1g_ptr, &l1g_meta, band_number,
                        &l1g_band_meta[band_index].band_meta,
                        l1g_file_list[l1g_file_count++], &dfp_meta);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Converting band %d to GeoTIFF", band_number);
                    ias_l1g_close_image(l1g_ptr);
                    ias_cpf_free(cpf);
                    exit(EXIT_FAILURE);
                }
            }
            else
            {
                IAS_LOG_WARNING("Band number %d is not in %s; skipping",
                        band_number, l1g_filename);
            }
        }

        /* Close the input L1G file */
        ias_l1g_close_image(l1g_ptr);

        /* If quality band created then convert it to GeoTIFF also */
        if (process_quality_band)
        {
            L1GIO *qualband_ptr;
            IAS_L1G_FILE_METADATA qualband_file_meta;
            IAS_L1G_BAND_METADATA qualband_band_meta;

            /* Open the quality band image for metadata and imagery access */
            qualband_ptr = ias_l1g_open_image(get_quality_band_filename(),
                    IAS_READ);
            if (qualband_ptr == NULL)
            {
                IAS_LOG_ERROR("Opening image %s", get_quality_band_filename());
                ias_cpf_free(cpf);
                exit(EXIT_FAILURE);
            }

            status = ias_l1g_get_file_metadata(qualband_ptr,
                    &qualband_file_meta);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Retrieving file metadata for %s",
                        get_quality_band_filename());
                ias_l1g_close_image(qualband_ptr);
                ias_cpf_free(cpf);
                exit(EXIT_FAILURE);
            }

            status = ias_l1g_get_band_metadata(qualband_ptr, 1,
                    &qualband_band_meta);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Retrieving L1G band metadata for band 1 of "
                        "image %s", get_quality_band_filename());
                ias_l1g_close_image(qualband_ptr);
                ias_cpf_free(cpf);
                exit(EXIT_FAILURE);
            }

            /* Convert the quality band */
            status = convert_band(qualband_ptr, &qualband_file_meta, 0,
                    &qualband_band_meta, l1g_file_list[l1g_file_count++],
                    &dfp_meta);

            /* Close the image before error checking since that is done no
               matter the success or failure of convert_band */
            ias_l1g_close_image(qualband_ptr);

            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Converting quality band to GeoTIFF");
                ias_cpf_free(cpf);
                exit(EXIT_FAILURE);
            }
        }


        /* Move the angle coefficients file to the correct name */
        if (strlen(dfp_meta.ang_filename) > 0)
        {
            char new_ang_filename[PATH_MAX]; /* New ANG filename */
            int status;

            /* Create the new ANG filename */
            if (ias_angle_gen_create_ang_filename(
                determine_l1_basename(&dfp_meta),
                sizeof(new_ang_filename), new_ang_filename) != SUCCESS)
            {
                IAS_LOG_ERROR("Creating the ANG coefficients filename");
                ias_cpf_free(cpf);
                return ERROR;
            }

            /* Unlink the existing angle coefficient file if it exists */
            status = unlink(new_ang_filename);
            if (status != 0 && errno != ENOENT)
             {
                IAS_LOG_ERROR("Unlinking the existing angle coefficient file "
                    "%s", new_ang_filename);
                ias_cpf_free(cpf);
                return ERROR;
            } 

            /* Link the angle coefficient filename to the new name */
            if (symlink(dfp_meta.ang_filename, new_ang_filename) != 0)
            {
                IAS_LOG_ERROR("Linking the angle coefficient file %s to the "
                    "new name %s", dfp_meta.ang_filename, new_ang_filename);
                ias_cpf_free(cpf);
                return ERROR;
            }

            /* Save the metadata file name for later inclusion in the package */
            if (get_include_angle_coeff_file_flag())
            {
                IAS_LOG_INFO("Creating the angle coefficients file");
                strcpy(l1g_file_list[l1g_file_count++], new_ang_filename); 
            }
            
            /* Copy the correctly formatted angle coefficient filename to be
               written to the metadata */
            status = snprintf(dfp_meta.ang_filename, 
                sizeof(dfp_meta.ang_filename), new_ang_filename); 
            if (status < 0 || status >= sizeof(dfp_meta.ang_filename))
            {
                IAS_LOG_ERROR("Copying the angle coefficient filename "
                    "to the metadata struct");
                ias_cpf_free(cpf);
                return ERROR;   
            }
        }

        /* Create the L1G metadata file */
        status = write_metadata(&l0r_scene_meta, &l0r_int_meta,
                l0r_meta_available, &l1g_meta, l1g_band_meta, &dfp_meta,
                &our_current_time, &control_band_index, l1g_meta_filename);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Writing L1G ASCII metadata (MTL) file");
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }

        /* Save the metadata file name for later processing */
        strcpy(l1g_file_list[l1g_file_count++], l1g_meta_filename);

        /* Free CPF memory */
        dfp_meta.tirs_thermal_constants = NULL;
        ias_cpf_free(cpf);
        cpf = NULL;
    }

    /* Package each product requested and update the OMF with package names */
    if (get_l0r_package_flag() || get_l1g_package_flag())
    {
        /* Sync the current directory metadata if requested */
        if (get_want_directory_synced())
        {
            if (sync_directory(".") != SUCCESS)
            {
                IAS_LOG_ERROR("Failed syncing the work order directory");
                exit(EXIT_FAILURE);
            }
        }

        if (create_packages(l1g_file_list, l1g_file_count,
                    &dfp_meta) != SUCCESS)
        {
            IAS_LOG_ERROR("Finishing product packaging");
            exit(EXIT_FAILURE);
        }

        /* Create the database load file for transfer_product to use */
        status = write_product_metadata(&l0r_scene_meta, &l1g_meta,
                l1g_band_meta, &dfp_meta, control_band_index);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Writing database load file");
            exit(EXIT_FAILURE);
        }

        /* Update OMF with package names */

        if (get_l0r_package_flag() && get_l1g_package_flag())
        {
            if (update_omf(dfp_meta.l0r_package_file_name,
                        dfp_meta.l1_package_file_name) != SUCCESS)
            {
                IAS_LOG_ERROR("Writing L0R & L1G package TARball names to OMF");
                return ERROR;
            }
        }
        else if (get_l0r_package_flag())
        {
            if (update_omf(dfp_meta.l0r_package_file_name, NULL) != SUCCESS)
            {
                IAS_LOG_ERROR("Writing L0R package TARball name to OMF");
                return ERROR;
            }
        }
        else if (get_l1g_package_flag())
        {
            if (update_omf(NULL, dfp_meta.l1_package_file_name) != SUCCESS)
            {
                IAS_LOG_ERROR("Writing L1G package TARball name to OMF");
                return ERROR;
            }
        }
    }

    IAS_LOG_INFO("Successful completion of DMS Format & Package");

    return EXIT_SUCCESS;
}
