/*----------------------------------------------------------------------------
NAME:    read_parameters

PURPOSE: Read the input application parameter and OMF files and provides
         accessor routines for the parameter values

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <stdio.h>
#include <time.h>
#include <errno.h>
#include "ias_types.h"
#include "ias_structures.h"
#include "ias_const.h"
#include "ias_miscellaneous.h"
#include "ias_satellite_attributes.h"
#include "ias_logging.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "ias_l0r.h"
#include "format_and_package.h"

/* --------------------------------------------------------------------- */
/* ------ These variables are returned by the accessor functions ------- */

/* application parameter file */
static int l1g_bands[IAS_MAX_NBANDS];
static int want_l0r_package;
static int want_l1g_product;
static int want_l1g_package;
static int want_md5_checksum;
static int want_angle_coeff_file;
static int want_named_prod_id;
static double tier_1_rmse_threshold;
static int want_directory_synced;

/* OMF file */
static char omf_filename[PATH_MAX];
static char l1g_filename[PATH_MAX];
static char sun_angle_corrected_l1g_filename[PATH_MAX];
static char quality_band_filename[PATH_MAX];

static int l1g_nbands = 0;
static int processing_pass = 0;
static int satellite_number = 0;

static char l0r_path[PATH_MAX];
static char l0r_basename[MAX_FILE_LENGTH];
static char l1g_output_basename[MAX_FILE_LENGTH];
static char project[33]; /* LCMAP, etc. */

/* local function prototypes */
static int get_l0r_metadata
(
    const char *l0r_filename,                  /* I: L0R filename */
    DFP_METADATA *dfp_meta,                    /* IO: DFP metadata structure */
    IAS_L0R_SCENE_METADATA *l0r_scene_meta,    /* O: L0R scene metadata
                                                     structure */
    IAS_L0R_INTERVAL_METADATA *l0r_int_meta,   /* O: L0R interval metadata
                                                     structure */
    int *l0r_meta_available                    /* O: Flag indicating if L0R
                                                     metadata was successfully
                                                     retrieved */
);

int read_parameters
(
    const char *parm_file,              /* I: Name of the parameter file */
    DFP_METADATA *dfp_meta,             /* O: Meta for L1G from misc sources */
    IAS_L0R_SCENE_METADATA *l0r_scene_meta,  /* O: For L1G meta file setup */
    IAS_L0R_INTERVAL_METADATA *l0r_int_meta, /* O: For L1G meta file setup */
    int *l0r_meta_available             /* O: Flag indicating if L0R metadata
                                              was successfully retrieved */
)
{
    int status;  /* Routine return status */
    int i;
    int count;   /* Count of elements in the parameter definition tables */
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char l0r_filename[PATH_MAX];
    char sensor[IAS_L0R_SENSOR_ID_LENGTH+1];  /* sensor from omf file */
    int satellite_id;             /* Current satellite id */
    int satellite_number_bands;   /* Number of normal bands for the satellite */
    int local_bands[IAS_MAX_NBANDS];/* Band list from param file */
    int local_nbands;             /* Number bands in local_bands */
    char tmp_string[PATH_MAX];    /* Used for manipulating filename strings */
    char *tmp_ptr = NULL;         /* Used for manipulating filename strings */
    int all_band_count;           /* Count of all OLI & TIRS bands possible */
    int all_bands[IAS_MAX_NBANDS];/* List of all OLI & TIRS band numbers */
    int ssm_quality;              /* Quality of the SSM positions */
    int band_saturation_count;    /* Number of bands with saturation */
    int saturation_band_list[IAS_MAX_NBANDS];

    /* --------------------------------------------------------------------- */
    /* -- Table definition for things from the application parameter file -- */

    IAS_PARM_DECLARE_TABLE(parms, 12);

    /* Add the work order ID parameter */
    IAS_PARM_WORK_ORDER_ID(parms, work_order_id, sizeof(work_order_id), 0);

    /* Add the processing pass number parameter */
    IAS_PARM_PROCESSING_PASS(parms, &processing_pass, sizeof(processing_pass),
            0);

    /* Add the list of band numbers parameter */
    IAS_PARM_BAND_LIST(parms, IAS_PARM_OPTIONAL, local_bands,
            sizeof(local_bands), 0);

    /* Add the L0R package wanted flag parameter */
    int default_disabled_flag[] = {0};
    IAS_PARM_ADD_BOOLEAN(parms, L0R_PACKAGE, "Flag to indicate if L0R package "
            "should be created (1 = Create package, 0 = Do not create package)",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, default_disabled_flag,
            &want_l0r_package, sizeof(want_l0r_package), 0);

    /* Add the L1G product (convert to GeoTIFF) wanted flag parameter */
    int default_enabled_flag[] = {1};
    IAS_PARM_ADD_BOOLEAN(parms, L1G_PRODUCT, "Flag to indicate if the L1G "
            "product (GeoTIFF) should be created (1 = Create product, "
            "0 = Do not create product)", IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
            default_enabled_flag, &want_l1g_product, sizeof(want_l1g_product),
            0);

    /* Add the L1G package wanted flag parameter */
    IAS_PARM_ADD_BOOLEAN(parms, L1G_PACKAGE, "Flag to indicate if L1G package "
            "should be created (1 = Create package, 0 = Do not create package)",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, default_disabled_flag,
            &want_l1g_package, sizeof(want_l1g_package), 0);

    /* Add the name using product ID flag parameter */
    IAS_PARM_ADD_BOOLEAN(parms, NAME_USING_PRODUCT_ID, "Flag to indicate if the"
            " package should be named with the Product ID (1 = Use Product ID, "
            "0 = Use Scene ID)",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, default_disabled_flag,
            &want_named_prod_id, sizeof(want_named_prod_id), 0);

    /* Add the MD5 checksum of package(s) wanted flag parameter */
    int default_md5[] = {1};
    IAS_PARM_ADD_BOOLEAN(parms, MD5_CHECKSUM_PACKAGES, "Flag to indicate if "
            "an MD5 checksum file should be created for any output "
            "package(s) (1 = Create checksum, 0 = Do not create checksum)",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, default_md5,
            &want_md5_checksum, sizeof(want_md5_checksum), 0);

    /* Add the include angle flag parameter */
    int default_include_ang[] = {0};
    IAS_PARM_ADD_BOOLEAN(parms, INCLUDE_ANG_FILE_FLAG, "Flag to indicate if "
        "the angle coefficient file should be included in the L1G package. "
        "(1 = yes, 0 = no)",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, default_include_ang,
        &want_angle_coeff_file, sizeof(want_angle_coeff_file), 0);

    /* Add the parameter indicating which project the product is for */
    const char *default_project[] = {""};
    IAS_PARM_ADD_STRING(parms, PROJECT, "Project the product is created for. "
        "Allows for customization to the tier naming and delivery location",
        IAS_PARM_OPTIONAL, 0, NULL, 1, default_project,
        project, sizeof(project), 1);

    /* Add the parameter indicating the required geodetic accuracy for a
       Tier 1 product */
    const double default_tier1_threshold[] = {12.0};
    IAS_PARM_ADD_DOUBLE(parms, TIER_1_RMSE_THRESHOLD, "RMSE value required"
        " to be considered a Tier 1 product", IAS_PARM_OPTIONAL,
        IAS_PARM_NOT_ARRAY, 1, 0.1, 100.0, 1, default_tier1_threshold, 
        &tier_1_rmse_threshold, sizeof(tier_1_rmse_threshold),
        0);

    int default_directory_sync[] = {0};
    IAS_PARM_ADD_BOOLEAN(parms, SYNC_DIRECTORY_FLAG, "Flag to request an "
        "fsync operation be performed on the work order directory to work "
        "around tar commands reporting an error about file times changing "
        "on Lustre file systems under heavy load. (1 = fsync, 0 = no fsync)",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, default_directory_sync,
        &want_directory_synced, sizeof(want_directory_synced), 0);

    /* --------------------------------------------------------------------- */
    /* -------------- Read the application parameter file ------------------ */
    /* Need to do this now, since need the processing pass for the parameters
       from the OMF. */

    /* Calculate the number of items in the application parameter file */
    count = IAS_PARM_GET_TABLE_SIZE(parms);

    /* Call the ias_parm_provide_help routine to provide help for the
       parameters if requested; also verifies the ODL file exists */
    status = ias_parm_provide_help(parm_file, parms, count,
            IAS_INPUT_PARAMETERS);
    if (status)
    {
        /* either help was provided or an error occurred */
        if (status == 1)
        {
            /* If the status of the provide help is 1 the user asked for help.
               Since the parameter file name contains the help request, copy
               it to the OMF file name so the next call will ask for the same
               help */
            strcpy(omf_filename, parm_file);
        }
        else
        {
            IAS_LOG_ERROR("Providing application parameter file help");
            return ERROR;
        }
    }
    else
    {
        /* Read the parameter file */
        status = ias_parm_read(parm_file, parms, count);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading parameters from: %s", parm_file);
            return ERROR;
        }

        /* Build the OMF file name */
        status = snprintf(omf_filename, sizeof(omf_filename), "%s.omf",
                work_order_id);
        if (status < 0 || status >= sizeof(omf_filename))
        {
            IAS_LOG_ERROR("Building the OMF filename");
            return ERROR;
        }
    }

    /* --------------------------------------------------------------------- */
    /* -------------- Table definition for things from the OMF ------------- */
    IAS_PARM_DECLARE_TABLE(omf_parms, 33);

    /* Add the product request ID */
    IAS_PARM_PRODUCT_REQUEST_ID(omf_parms, dfp_meta->product_request_id,
            sizeof(dfp_meta->product_request_id), 0);

    /* Add the L0R image parameter */
    IAS_PARM_L0R_FILENAME(omf_parms, IAS_PARM_REQUIRED, l0r_filename,
            sizeof(l0r_filename), 0);

    /* Add the L1R image parameter */
    IAS_PARM_L1R_FILENAME(omf_parms, IAS_PARM_OPTIONAL, dfp_meta->l1r_filename,
            sizeof(dfp_meta->l1r_filename), 0);

    /* Add the L1G image parameter */
    IAS_PARM_L1G_FILENAME_FROM_PASS(omf_parms, IAS_PARM_OPTIONAL, l1g_filename,
            sizeof(l1g_filename), 0, processing_pass);

    /* Add the L1G image parameter */
    IAS_PARM_SUN_ANGLE_CORRECTED_L1G_FILENAME(omf_parms, IAS_PARM_OPTIONAL,
            sun_angle_corrected_l1g_filename,
            sizeof(sun_angle_corrected_l1g_filename), 0);

    /* Add read for the LANDSAT_SCENE_ID */
    strcpy(l1g_output_basename, "");
    IAS_PARM_LANDSAT_SCENE_ID(omf_parms, IAS_PARM_OPTIONAL, l1g_output_basename,
            sizeof(l1g_output_basename), 1);

    /* Add the quality band image parameter */
    IAS_PARM_QUALITY_BAND_FILENAME(omf_parms, IAS_PARM_OPTIONAL,
            quality_band_filename, sizeof(quality_band_filename));

    /* Add the DEM file parameter */
    IAS_PARM_DEM_FILENAME_FROM_PASS(omf_parms, IAS_PARM_OPTIONAL,
            dfp_meta->dem_filename, sizeof(dfp_meta->dem_filename), 0,
            processing_pass);

    /* Add the ANG file parameter */
    IAS_PARM_ANGLE_COEF_FILENAME_FROM_PASS(omf_parms, IAS_PARM_OPTIONAL,
        dfp_meta->ang_filename, sizeof(dfp_meta->ang_filename), 1,
        processing_pass);

    /* Add the CPF parameter */
    IAS_PARM_CPF(omf_parms, IAS_PARM_OPTIONAL, dfp_meta->cpf_filename,
            sizeof(dfp_meta->cpf_filename), 0);

    /* Add the OLI BPF parameter */
    IAS_PARM_BPF_OLI(omf_parms, IAS_PARM_OPTIONAL, dfp_meta->bpf_filename_oli,
            sizeof(dfp_meta->bpf_filename_oli), 0);

    /* Add the TIRS BPF parameter */
    IAS_PARM_BPF_TIRS(omf_parms, IAS_PARM_OPTIONAL, dfp_meta->bpf_filename_tirs,
            sizeof(dfp_meta->bpf_filename_tirs), 0);

    /* Add the RLUT filename parameter */
    IAS_PARM_RLUT(omf_parms, IAS_PARM_OPTIONAL, dfp_meta->rlut_filename,
            sizeof(dfp_meta->rlut_filename), 0);

    /* Add the grid filename parameter */
    IAS_PARM_GRID_FILENAME_FROM_PASS(omf_parms, IAS_PARM_OPTIONAL,
            dfp_meta->grid_filename, sizeof(dfp_meta->grid_filename), 0,
            processing_pass);

    /* Add the satellite number parameter */
    /* The parameter I/O library defaults this to the current satellite
       number if this is not in the OMF */
    IAS_PARM_SATELLITE(omf_parms, &satellite_number, sizeof(satellite_number),
            0);

    /* Add the target path/row parameters */
    IAS_PARM_TARGET_WRS_PATH(omf_parms, IAS_PARM_OPTIONAL, 
            &dfp_meta->target_wrs_path, sizeof(dfp_meta->target_wrs_path), 0);
    IAS_PARM_TARGET_WRS_ROW(omf_parms, IAS_PARM_OPTIONAL, 
            &dfp_meta->target_wrs_row, sizeof(dfp_meta->target_wrs_row), 0);

    /* Add the spacecraft ID parameter */
    IAS_PARM_SPACECRAFT(omf_parms, IAS_PARM_OPTIONAL,
            dfp_meta->spacecraft_id, sizeof(dfp_meta->spacecraft_id), 0);

    IAS_PARM_CLOUD_COVER_SCORE(omf_parms, &dfp_meta->cloud_cover_score,
            sizeof(dfp_meta->cloud_cover_score));

    IAS_PARM_CLOUD_COVER_LAND_SCORE(omf_parms, 
            &dfp_meta->cloud_cover_land_score,
            sizeof(dfp_meta->cloud_cover_land_score));

    IAS_PARM_SNOW_ICE_SCORE(omf_parms, &dfp_meta->snow_ice_score,
            sizeof(dfp_meta->snow_ice_score));
   
    double default_double_invalid[] = {-1.0};
    IAS_PARM_GEOMETRIC_RMSE_MODEL_Y(omf_parms, 
            /* default to -1 to indicate the RMSE not available */
            1, default_double_invalid,
            &dfp_meta->model_y_rmse, sizeof(dfp_meta->model_y_rmse), 0);

    IAS_PARM_GEOMETRIC_RMSE_MODEL_X(omf_parms,
            /* default to -1 to indicate the RMSE not available */
            1, default_double_invalid,
            &dfp_meta->model_x_rmse, sizeof(dfp_meta->model_x_rmse),
            0);

    IAS_PARM_SENSOR(omf_parms, IAS_PARM_REQUIRED, sensor, sizeof(sensor), 1);

    IAS_PARM_TIRS_SSM_POSITION_STATUS(omf_parms, IAS_PARM_OPTIONAL,
            dfp_meta->ssm_position_status,
            sizeof(dfp_meta->ssm_position_status));
 
    /* PARAM 28: */
    int default_ground_control_points_model[] = {0};
    IAS_PARM_GROUND_CONTROL_POINTS_MODEL(omf_parms, 1, 
            default_ground_control_points_model,
            &dfp_meta->ground_control_points_model,
            sizeof(dfp_meta->ground_control_points_model));

    /* PARAM 29: */
    IAS_PARM_GEOMETRIC_RMSE_MODEL(omf_parms, 1, default_double_invalid,
            &dfp_meta->geometric_rmse_model,
            sizeof(dfp_meta->geometric_rmse_model));

    /* PARAM 30: */
    int default_ground_control_points_verify[] = {0};
    IAS_PARM_GROUND_CONTROL_POINTS_VERIFY(omf_parms, 1,
            default_ground_control_points_verify,
            &dfp_meta->ground_control_points_verify,
            sizeof(dfp_meta->ground_control_points_verify));

    /* PARAM 31: */
    IAS_PARM_GEOMETRIC_RMSE_VERIFY(omf_parms, 1, default_double_invalid,
            &dfp_meta->geometric_rmse_verify,
            sizeof(dfp_meta->geometric_rmse_verify));
    
    /* PARAM 32: */
    IAS_PARM_GROUND_CONTROL_POINTS_VERSION(omf_parms, IAS_PARM_OPTIONAL,
            &dfp_meta->ground_control_points_version,
            sizeof(dfp_meta->ground_control_points_version), 0);

    /* SSM Position Quality */
    IAS_PARM_SSM_POSITION_QUALITY(omf_parms, &ssm_quality, sizeof(ssm_quality),
        0);

    IAS_PARM_SATURATED_BANDS_LIST(omf_parms, IAS_PARM_OPTIONAL,
        saturation_band_list, sizeof(saturation_band_list), 0);

    /* Stray light correction source */
    IAS_PARM_TIRS_STRAY_LIGHT_CORRECTION_SOURCE(omf_parms, 
        IAS_PARM_OPTIONAL, dfp_meta->stray_light_correction_source, 
        sizeof(dfp_meta->stray_light_correction_source), 0);

    /* --------------------------------------------------------------------- */
    /* -------------- Read the contents defined from the OMF --------------- */

    /* See if the user wanted help or a template, this time with the OMF file */
    count = IAS_PARM_GET_TABLE_SIZE(omf_parms);
    status = ias_parm_provide_help(omf_filename, omf_parms, count,
            IAS_OMF_PARAMETERS);
    if (status)
    {
        if (status == 1) /* help was successful */
            exit(EXIT_SUCCESS);
        else
        {
            IAS_LOG_ERROR("Providing OMF file help");
            return ERROR;
        }
    }

    /* Read the OMF file */
    status = ias_parm_read(omf_filename, omf_parms, count);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from OMF file: %s", omf_filename);
        return ERROR;
    }

    /* Save the SSM quality */
    dfp_meta->ssm_position_quality = ssm_quality;

    /* Get metadata for transfer_product to use */
    status = snprintf(dfp_meta->work_order_id, sizeof(dfp_meta->work_order_id),
            "%s", work_order_id);
    if (status < 0 || status >= sizeof(dfp_meta->work_order_id))
    {
        IAS_LOG_ERROR("Getting work order id");
        return ERROR;
    }

    /* --------------------------------------------------------------------- */
    /* -------------- Do some validation of values retrieved --------------- */

    /* Make sure at least one output product or package was requested,
       otherwise the application was not asked to do anything! */
    if (!want_l0r_package && !want_l1g_product && !want_l1g_package)
    {
        IAS_LOG_WARNING("No output formatting or packaging indicated by "
                "parameters");
        return SUCCESS;
    }

    /* Check for conflicting L1G parameters */
    if (!want_l1g_product && want_l1g_package)
    {
        IAS_LOG_WARNING("Parameters conflict: L1G_PRODUCT disabled; "
                "L1G_PACKAGE enabled. L1G product (GeoTIFFs) will be created "
                "anyway, to fulfill L1G packaging request.");
    }

    /* Ensure parameters needed for each processing case were provided */
    if (want_l0r_package && (strcmp(l0r_filename, "") == 0))
    {
        IAS_LOG_ERROR("L0R package requested with invalid L0R filename in the "
                "parameter file");
        return ERROR;
    }

    if (want_l1g_product || want_l1g_package)
    {
        if (strcmp(l1g_filename, "") == 0)
        {
            IAS_LOG_ERROR("L1G product or package requested with invalid L1G "
                    "filename in the OMF file");
            return ERROR;
        }

        if (strcmp(dfp_meta->l1r_filename, "") == 0)
        {
            IAS_LOG_ERROR("L1G product or package requested with invalid L1R "
                    "filename in the OMF file");
            return ERROR;
        }

        /* Make sure the OMF optional parameters have a known default value
           if they weren't provided */
        int parm_count;

        parm_count = IAS_PARM_GET_COUNT_READ(omf_parms, DEM_FILENAME);
        if (parm_count != 1)
            strcpy(dfp_meta->dem_filename, "");
        parm_count = IAS_PARM_GET_COUNT_READ(omf_parms, CAL_PARM_FILENAME);
        if (parm_count != 1)
            strcpy(dfp_meta->cpf_filename, "");
        parm_count = IAS_PARM_GET_COUNT_READ(omf_parms, BIAS_PARM_FILENAME_OLI);
        if (parm_count != 1)
            strcpy(dfp_meta->bpf_filename_oli, "");
        parm_count = IAS_PARM_GET_COUNT_READ(omf_parms,
            BIAS_PARM_FILENAME_TIRS);
        if (parm_count != 1)
            strcpy(dfp_meta->bpf_filename_tirs, "");
        parm_count = IAS_PARM_GET_COUNT_READ(omf_parms, RLUT_FILENAME);
        if (parm_count != 1)
            strcpy(dfp_meta->rlut_filename, "");
        parm_count = IAS_PARM_GET_COUNT_READ(omf_parms, GRID_FILENAME);
        if (parm_count != 1)
            strcpy(dfp_meta->grid_filename, "");
        parm_count = IAS_PARM_GET_COUNT_READ(omf_parms, TARGET_WRS_PATH);
        if (parm_count != 1)
            dfp_meta->target_wrs_path = -1;
        parm_count = IAS_PARM_GET_COUNT_READ(omf_parms, TARGET_WRS_ROW);
        if (parm_count != 1)
            dfp_meta->target_wrs_row = -1;
        parm_count = IAS_PARM_GET_COUNT_READ(omf_parms, 
            GROUND_CONTROL_POINTS_VERSION);
        if (parm_count != 1)
            dfp_meta->ground_control_points_version = -1;
    }

    /* Convert the satellite number to a satellite id */
    satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(
            satellite_number);
    if (satellite_id == ERROR)
    {
        IAS_LOG_ERROR("Converting the satellite number to a satellite ID");
        return ERROR;
    }

    /* Get sensor types present */
    if (strstr(sensor, "OLI") != NULL)
    {
        strcpy(dfp_meta->l0r_oli_present, "Y");
    }
    else
    {
        strcpy(dfp_meta->l0r_oli_present, "N");
    }
   
    if (strstr(sensor, "TIRS") != NULL)
    {
        strcpy(dfp_meta->l0r_tirs_present, "Y");
    }
    else
    {
        strcpy(dfp_meta->l0r_tirs_present, "N");
    }
    
    /* Check if the SSM position status was present */
    if (IAS_PARM_GET_COUNT_READ(omf_parms, TIRS_SSM_POSITION_STATUS) != 1)
        strcpy(dfp_meta->ssm_position_status, "");

    /* Initialize satellite attributes */
    if (ias_sat_attr_initialize(satellite_id) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the satellite attributes library");
        return ERROR;
    }

    /* Check if any bands were reported */
    band_saturation_count = IAS_PARM_GET_COUNT_READ(omf_parms,
        SATURATED_BANDS_LIST);
    if (band_saturation_count == 0)
    {
        dfp_meta->have_saturated_bands_flag = FALSE;
    }
    else
    {
        dfp_meta->have_saturated_bands_flag = TRUE;

        /* Initialize the band saturation list */
        for (i = 0; i < IAS_MAX_NBANDS; i++)
            dfp_meta->band_saturated[i] = 0;

        /* If -1 is provided as the first entry in the list of bands, that
           means no bands had saturation, but we want to report it in the
           MTL file */
        if (saturation_band_list[0] != -1)
        {
            /* Convert the band numbers to band index */
            for (i = 0; i < band_saturation_count; i++)
            {
                /* Convert the band number to an index */
                int band_index = ias_sat_attr_convert_band_number_to_index(
                    saturation_band_list[i]);
                if (band_index == ERROR)
                {
                    IAS_LOG_ERROR("Invalid band number (%d) in saturation band "
                        "list", saturation_band_list[i]);
                    return ERROR;
                }
                dfp_meta->band_saturated[band_index] = 1;
            }
        }
    }

    /* Validate the band list if L1G product or packaging requested
       (parameter not applicable to L0R packaging) */
    if (want_l1g_product || want_l1g_package)
    {
        /* Find the number of normal bands for the satellite */
        satellite_number_bands = ias_sat_attr_get_normal_band_count();
        if ((satellite_number_bands < 1) ||
                (satellite_number_bands > IAS_MAX_NBANDS))
        {
            IAS_LOG_ERROR("Invalid number of bands: %d.",
                    satellite_number_bands);
            return ERROR;
        }

        /* Set how many bands there are to process for the L1G. Note that this
           is only for the image bands, not including blind band or the VRPs
           and not applicable for the L0R package. */
        local_nbands = IAS_PARM_GET_COUNT_READ(parms, BAND_LIST);

        /* Get the full list of bands possible, then save for using if a
           band list was not provided.
           Unavailable bands will ultimately be skipped (does not cause fatal
           error), so safe to default to try to process all of them. */
        if (ias_sat_attr_get_sensor_band_numbers(IAS_MAX_SENSORS,
                    IAS_NORMAL_BAND, 0, all_bands, IAS_MAX_NBANDS,
                    &all_band_count) != SUCCESS)
        {
            IAS_LOG_ERROR("Initializing full bands list");
            return ERROR;
        }
        if (all_band_count > IAS_MAX_NBANDS)
        {
            IAS_LOG_ERROR("Initializing count of all OLI and TIRS bands");
            return ERROR;
        }

        if (local_nbands == 0)
        {
            /* No bands were specified in the parameter file, so set the
               DFP up to try to process all OLI and TIRS bands */
            l1g_nbands = all_band_count;
            for (i = 0; i < all_band_count; i++)
            {
                l1g_bands[i] = all_bands[i];
            }
        }

        /* Initialize the process band array */
        for (i = 0; i < IAS_MAX_NBANDS; i++)
            dfp_meta->process_band[i] = 0;

        /* Process BAND_LIST, if it was provided in the parameter file.
           Get rid of duplicate bands, and issue warning if there are any */
        for (i = 0; i < local_nbands; i++)
        {
            int band_index;

            /* Convert the band number to an index */
            band_index = ias_sat_attr_convert_band_number_to_index(
                    local_bands[i]);
            if (band_index == ERROR)
            {
                IAS_LOG_ERROR("Invalid band number (%d) in parameter file",
                        local_bands[i]);
                return ERROR;
            }

            /* Make sure the band isn't already in the list */
            if (!dfp_meta->process_band[band_index])
            {
                /* Put band in list and flag that it has been added
                   to the list */
                l1g_bands[l1g_nbands++] = local_bands[i];
                dfp_meta->process_band[band_index] += 1;
            }
            else if (dfp_meta->process_band[band_index] == 1)
            {
                /* only print the warning message once for each band */
                dfp_meta->process_band[band_index] += 1;

                IAS_LOG_WARNING("Band %d appears multiple times in the "
                        "BAND_LIST; duplicates ignored",
                        local_bands[i]);
            }
        }
    }

    /* Ensure the process_band array is only 0s and 1s */
    for (i = 0; i < IAS_MAX_NBANDS; i++)
    {
        if (dfp_meta->process_band[i] > 1)
        {
            dfp_meta->process_band[i] = 1;
        }
    }

    /* Do some filename manipulations to get strings that are valuable to us */

    /* The L0R is used yet, even if packaging of it was not requested (for
       scene metadata), so setup the basename and path no matter what
       format/package options were selected */
    strcpy(tmp_string, l0r_filename);
    tmp_ptr = strrchr(tmp_string, '/');
    if (tmp_ptr == NULL)
    {
        /* No path component, filename is full string */
        strcpy(l0r_basename, l0r_filename);
    }
    else
    {
        /* Copy just the filename component out of the string */
        tmp_ptr++;
        strcpy(l0r_basename, tmp_ptr);
    }
    tmp_ptr = NULL;

    /* tmp_string still contains l0r_filename */
    tmp_ptr = strrchr(tmp_string, '/');
    if (tmp_ptr == NULL)
    {
        /* No path component, set it to an empty string */
        strcpy(l0r_path, "");
    }
    else
    {
        /* Terminate the string after the end of the path and copy */
        *tmp_ptr = '\0';
        if ((strlen(tmp_string) + 1) > sizeof(l0r_path))
        {
            /* Catch if the path length would overflow our final l0r
               system command buffer of PATH_MAX size. Limit the l0r_path
               length here, early in the application, to quit before
               time-consuming processing is done and then potentially
               discover the string length error after a long waste of time. */
            IAS_LOG_ERROR("L0R file location path length too long");
            return ERROR;
        }
        strcpy(l0r_path, tmp_string);
    }
    tmp_ptr = NULL;

    /* Get the L0R interval and scene metadata which provide values for the
       product metadata */
    status = get_l0r_metadata(l0r_filename, dfp_meta, l0r_scene_meta,
            l0r_int_meta, l0r_meta_available);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving L0R metadata");
        return ERROR;
    }

    /* Only need to manipulate the L1G filename if the L1G product or package
       was requested */
    if (want_l1g_product || want_l1g_package)
    {
        /* default to landsat scene id */
        strcpy(l1g_output_basename, l0r_scene_meta->landsat_scene_id);

        /* Since the L1G package was not requested, and the L0R metadata
           is not available fall back to using the source L1G filename as 
           the basename for the output L1G files (GeoTIFFs). This way, an
           L1G can be converted to GeoTIFF with no requirement for an L0R 
           since this may be the desired capabilities of the DFP at times.
        */
        if (!want_l1g_package && !(*l0r_meta_available))
        {
            char l1g_input_basename[MAX_FILE_LENGTH];

            /* Get the basename of the L1G filename */
            strcpy(tmp_string, l1g_filename);
            tmp_ptr = strrchr(tmp_string, '/');
            if (tmp_ptr == NULL)
            {
                /* No path component, filename is full string */
                strcpy(l1g_input_basename, l1g_filename);
            }
            else
            {
                /* Copy just the filename out of the string */
                tmp_ptr++;
                strcpy(l1g_input_basename, tmp_ptr);
            }

            tmp_ptr = strrchr(l1g_input_basename, '.');
            /* Check if found a '.' in the filename; unlikely to have a filename
               like that, but potentially possible */
            if (tmp_ptr != NULL)
            {
                *tmp_ptr = '\0';
                tmp_ptr = NULL;
            }

            strcpy(l1g_output_basename, l1g_input_basename);
        }
    }

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:    get_l0r_metadata

PURPOSE: Reads the provided L0R metadata for the Landsat Scene ID.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int get_l0r_metadata
(
    const char *l0r_filename,                /* I: L0R filename */
    DFP_METADATA *dfp_meta,                  /* IO: DFP metadata structure */
    IAS_L0R_SCENE_METADATA *l0r_scene_meta,  /* O: L0R scene metadata 
                                                   structure */
    IAS_L0R_INTERVAL_METADATA *l0r_int_meta, /* O: L0R interval metadata
                                                   structure  */
    int *l0r_meta_available                  /* O: Flag indicating if L0R
                                                   metadata was successfully
                                                   retrieved */
)
{
    L0RIO *l0r_ptr;
    int status;
    int record_count;
    int index;
    int have_scene_id = (strlen(l1g_output_basename) > 0);
    int found_scene = 0;
    int header_present;

    /* Initialize the lower bits flag */
    dfp_meta->is_oli_lower_12_bits_flag = 0;

    /* Assume l0r data not available until we successfully retrieve it */
    *l0r_meta_available = 0;

    l0r_ptr = ias_l0r_open(l0r_basename, l0r_path, IAS_L0R_COMPRESSION_DEFAULT);
    if (!l0r_ptr)
    {
        if (!want_l1g_package)
        {
            /* Since the L1G package was not requested, allow the retrieval
               of the L0R metadata to fail gracefully. */
            IAS_LOG_WARNING("Opening L0R %s; unable to determine certain L1G "
                    "metadata", l0r_filename);
            return SUCCESS;
        }
        IAS_LOG_ERROR("Opening L0R %s", l0r_filename);
        return ERROR;
    }

    /* Open the L0R header */
    if (ias_l0r_open_header(l0r_ptr, IAS_READ) != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to check open the L0R header for file %s",
            l0r_filename);
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Check if the header is available to check */
    if (ias_l0r_is_oli_image_header_present(l0r_ptr, &header_present)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to check if OLI image header is present for "
            "file %s", l0r_filename);
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Read the header if available */
    if (header_present)
    {
        IAS_L0R_OLI_IMAGE_HEADER oli_header;

        /* Open the header and check the truncation mode */
        if (ias_l0r_get_oli_image_header(l0r_ptr, &oli_header) != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to get the OLI image header for "
                "file %s", l0r_filename);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }

        /* Check if the lower 12 bits is set */
        dfp_meta->is_oli_lower_12_bits_flag =
            oli_header.image_data_truncation_setting;
    }

    /* Close the header */
    if (ias_l0r_close_header(l0r_ptr) != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to close the L0R header for file %s",
            l0r_filename);
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }


    status = ias_l0r_open_metadata(l0r_ptr, IAS_READ);
    if (status != SUCCESS)
    {
        if (!want_l1g_package)
        {
            IAS_LOG_WARNING("Opening L0R metadata for file %s; unable to "
                    "determine certain L1G metadata", l0r_filename);
            ias_l0r_close(l0r_ptr);
            return SUCCESS;
        }
        IAS_LOG_ERROR("Opening L0R metadata for file %s", l0r_filename);
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Retrieve the L0R scene record count */
    if (ias_l0r_get_scene_metadata_records_count(l0r_ptr, &record_count) 
        != SUCCESS)
    {
        IAS_LOG_ERROR("Reading L0R scene count");
        ias_l0r_close_metadata(l0r_ptr);
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Search for the scene that we are processing. If we do not have a scene
       id we will take the first scene in the L0R */
    for (index = 0; index < record_count; index++)
    {
        /* Read each scene metadata */
        status = ias_l0r_get_scene_metadata(l0r_ptr, index, 1, l0r_scene_meta);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading L0R scene metadata for file %s", 
                l0r_filename);
            ias_l0r_close_metadata(l0r_ptr);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }

        /* Compare the scene id from the metadata to the scene id we are looking
           for, if we don't have a scene id take the first scene metadata */
        if (!have_scene_id || strcmp(l0r_scene_meta->landsat_scene_id, 
            l1g_output_basename) == 0)
        {
            found_scene = 1;
            break;
        }
    }

    /* Check if scene was specified but not found */
    if (have_scene_id && !found_scene)
    {
        IAS_LOG_ERROR("No scene metadata found for landsat scene id %s",
            l1g_output_basename);
        ias_l0r_close_metadata(l0r_ptr);
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    if (!found_scene)
    {
        if (want_l1g_package)
        {
            IAS_LOG_ERROR("No L0R scene metadata was found for file %s",
                l0r_filename);
            ias_l0r_close_metadata(l0r_ptr);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }
        else
        {
            IAS_LOG_WARNING("Reading L0R scene metadata for file %s; "
                "unable to determine certain L1G metadata", l0r_filename);
            return SUCCESS;
        }
    }
    else
    {
        /* Also do a sanity check of the target path/row against the OMF values,
           or use if OMF values weren't available */
        if (dfp_meta->target_wrs_path == -1)
            dfp_meta->target_wrs_path = l0r_scene_meta->target_wrs_path;
        if (dfp_meta->target_wrs_row == -1)
            dfp_meta->target_wrs_row = l0r_scene_meta->target_wrs_row;

        if (want_l1g_package
                && (dfp_meta->target_wrs_path != l0r_scene_meta->target_wrs_path
                || dfp_meta->target_wrs_row != l0r_scene_meta->target_wrs_row))
        {
            IAS_LOG_ERROR("L0R WRS target path/row values do not match OMF");
            ias_l0r_close_metadata(l0r_ptr);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }
    }

    status = ias_l0r_get_interval_metadata(l0r_ptr, l0r_int_meta);

    /* Done accessing the file */
    ias_l0r_close_metadata(l0r_ptr);
    ias_l0r_close(l0r_ptr);

    if (status != SUCCESS)
    {
        if (!want_l1g_package)
        {
            IAS_LOG_WARNING("Reading L0R interval metadata for file %s; unable"
                    "to determine certain L1G metadata", l0r_filename);
            return SUCCESS;
        }
        IAS_LOG_ERROR("Reading L0R interval metadata for file %s",
                l0r_filename);
        return ERROR;
    }

    /* Let caller know L0R data is available */
    *l0r_meta_available = 1;

    return SUCCESS;
}

/***********************************************************************
    Functions used to access the parameter file values
***********************************************************************/
/* Processing flags */
int get_l0r_package_flag(void)
{
    return want_l0r_package;
}

int get_l1g_product_flag(void)
{
    return want_l1g_product;
}

int get_l1g_package_flag(void)
{
    return want_l1g_package;
}

int get_md5_checksum_flag(void)
{
    return want_md5_checksum;
}

int get_include_angle_coeff_file_flag(void)
{
    return want_angle_coeff_file;
}

int get_named_with_prod_id(void)
{
    return want_named_prod_id;
}

/* Bands to process */
const int *get_l1g_bands(void)
{
    return l1g_bands;
}

int get_l1g_nbands(void)
{
    return l1g_nbands;
}

int get_processing_pass(void)
{
    return processing_pass;
}

/* File names */
const char *get_l0r_basename(void)
{
    return l0r_basename;
}

const char *get_l0r_path(void)
{
    return l0r_path;
}

const char *get_l1g_input_filename(void)
{
    return l1g_filename;
}

const char *get_project(void)
{
    return project;
}

const char *get_sun_angle_corrected_l1g_input_filename(void)
{
    return sun_angle_corrected_l1g_filename;
}

const char *get_l1g_output_basename(void)
{
    /* This is the Landsat Scene ID, if it was available, otherwise the
       basename of the input L1G filename. */
    return l1g_output_basename;
}

const char *get_omf_filename(void)
{
    return omf_filename;
}

const char *get_quality_band_filename(void)
{
    return quality_band_filename;
}

int get_satellite(void)
{
    return satellite_number;
}

double get_tier_1_rmse_threshold(void)
{
    return tier_1_rmse_threshold;
}

int get_want_directory_synced(void)
{
    return want_directory_synced;
}
