/*****************************************************************************
Name:           format_and_package_test

Purpose:
    The regression test for Format and Package.

******************************************************************************
                        Property of the U.S. Government
                            USGS EROS Data Center
*****************************************************************************/
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include "../format_and_package.h"
#include "cgreen/cgreen.h"
#include "ias_miscellaneous.h"

static char *unit_test_dir = NULL;

static char l1g_tiff[30] = {"LO80400332011265EDC01_B4.TIF"};
static char l1g_meta[31] = {"LO80400332011265EDC01_MTL.txt"};
static char l1g_meta_ref[40] = {"LO80400332011265EDC01_MTL_V8.txt"};
static char l1g_quality_band[31] = {"LO80400332011265EDC01_BQA.TIF"};
static char l1g_package[30] = {"LO80400332011265EDC01.tar.gz"};
static char l1g_package_checksum[30] = {"LO80400332011265EDC01_MD5.txt"};
static char l1g_package_file_list[24] = {"L1G_tar_file_list_v3.txt"};
static char l0r_package[36] = {"LO80400330332000227EDC00_L0R.tar.gz"};
static char l0r_package_checksum[40] = {"LO80400330332000227EDC00_L0R_MD5.txt"};
static char l0r_package_file_list[22] = {"L0R_tar_file_list.txt"};
static char product_data_file[17] = {"product_data.txt"};
static char product_data_file_ref[26] = {"product_data_v9.txt"};
static char product_data_file_l0r_ref[30] = {"product_data_l0r.txt"};
static char param_filename[15] = {"parameters.odl"};
static char omf_filename_orig[11] = {"WO1234.txt"};
static char omf_filename[11] = {"WO1234.omf"};
static char param_filename_l0r[19] = {"parameters_l0r.odl"};
static char omf_filename_l0r[15] = {"WO1234_l0r.omf"};
static char param_filename_no_op[21] = {"parameters_no_op.odl"};

static int diff_files(const char *file1, const char *file2);
static int remove_files(char *file_list[], int);
static int prepare_omf_file(char *omf_file_orig, char *omf_file);
static int prepare_ang_file(void);

static char param_float_filename[32] = {"parameters_float.odl"};
static char l1g_float_tiff[30] = {"LC81980472013089LGN00_B3.TIF"};
static char l1g_float_meta[31] = {"LC81980472013089LGN00_MTL.txt"};
static char l1g_float_package[30] = {"LC81980472013089LGN00.tar.gz"};
static char l1g_float_package_checksum[30] = {"LC81980472013089LGN00_MD5.txt"};
static char l1g_float_package_file_list[25] = {"L1G_tar_file_list_v2.txt"};

Ensure test_format_and_package(void);
Ensure test_l0r_package(void);
Ensure test_no_op(void);
Ensure test_format_and_package_float(void);
Ensure test_write_tirs_ssm_quality(void);
Ensure test_write_albers_metadata(void);

int main(int argc, char *argv[])
{
    char *file_remove_list[9];     

    TestSuite *suite = create_test_suite();

    file_remove_list[0] = l1g_package;
    file_remove_list[1] = l0r_package;
    file_remove_list[2] = l1g_package_checksum;
    file_remove_list[3] = l0r_package_checksum;
    file_remove_list[4] = product_data_file;
    file_remove_list[5] = l1g_tiff;
    file_remove_list[6] = l1g_meta;
    file_remove_list[7] = l1g_quality_band;
    file_remove_list[8] = omf_filename;

    add_test(suite, test_format_and_package_float);
    add_test(suite, test_format_and_package);
    add_test(suite, test_l0r_package);
    add_test(suite, test_no_op);
    add_test(suite, test_write_tirs_ssm_quality);
    add_test(suite, test_write_albers_metadata);

    unit_test_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (unit_test_dir == NULL)
    {
        printf("The IAS_UNIT_TEST_DATA_DIR environment variable is not set\n");
        exit(EXIT_FAILURE);
    }

    if (setenv("IAS_VERSION_OVERRIDE", "Y", 0) != 0)
    {
        printf("setenv failed / insufficient space in the environment\n");
        exit(EXIT_FAILURE);
    }

    /* Remove files from failed previous attempt */
    remove_files(file_remove_list, 9);

    if (argc > 1)
        return run_single_test(suite, argv[1], create_text_reporter());

    return run_test_suite(suite, create_text_reporter());
}

/* A test of an SCA-combined image */
Ensure test_format_and_package()
{
    int status;
    char command[PATH_MAX * 2];

    char omf_filename_orig[15] = "WO1234.txt.ang";
    char omf_filename[12] = "WO1234.omf";
    char ang_filename_orig[25] = "WO1234_ANG.txt.template";
    char ang_filename[30] = "LO80400332011265EDC01_ANG.txt";

    /* Prepare the control file names to compare against */
    char test_files_loc[PATH_MAX];
    char *file_remove_list[10];

    /* Files to remove created by this test */
    file_remove_list[0] = l1g_package;
    file_remove_list[1] = l0r_package;
    file_remove_list[2] = l1g_package_checksum;
    file_remove_list[3] = l0r_package_checksum;
    file_remove_list[4] = product_data_file;
    file_remove_list[5] = l1g_tiff;
    file_remove_list[6] = l1g_meta;
    file_remove_list[7] = l1g_quality_band;
    file_remove_list[8] = omf_filename;
    file_remove_list[9] = ang_filename;

    if (prepare_omf_file(omf_filename_orig, omf_filename) != SUCCESS)
    {
        assert_true_with_message(0, "Error preparing OMF file %s for unit "
                "test data directory path setup", omf_filename_orig);
        return;
    }

    if (prepare_ang_file() != 0)
    {
        assert_true_with_message(0, "Error preparing the ANG file from "
            "template");
        return;
    }

    /* Now run Format & Package to test it on this image */
    status = snprintf(command, sizeof(command), "../format_and_package %s",
            param_filename);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "to execute the Format & Package application");
        return;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Format & Package failed.  Command: %s",
            command);
        return;
    }

    /* Now compare the output files with ones previously generated */
    status = snprintf(test_files_loc, PATH_MAX,
           "%s/ias_base/dms/format_and_package/output/WO1234/", unit_test_dir);
    if (status < 0 || status >= PATH_MAX)
    {
        assert_true_with_message(0, "Error preparing path to control "
                "test files");
        return;
    }

    /* Check the band TIFF file */
    status = snprintf(command, sizeof(command), "diff -q %s %s%s", l1g_tiff,
            test_files_loc, l1g_tiff);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the L1G tiff file");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found with Format & "
                "Package output. Review %s and %s%s for details", l1g_tiff,
                test_files_loc, l1g_tiff);
    }

    /* Check the remaining L1G files */
    status = snprintf(command, sizeof(command),
            "sed"
            " -e 's|FILE_DATE = .*$|FILE_DATE = |'"
            " -e 's|PROCESSING_SOFTWARE_VERSION = \"IAS|"
            "PROCESSING_SOFTWARE_VERSION = \"LPGS|'"
            " %s | diff - %s%s",
            l1g_meta, test_files_loc, l1g_meta_ref);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the L1G metadata file");
        return;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found with Format & Package "
                "output. Review %s and %s%s for details", l1g_meta,
                test_files_loc, l1g_meta_ref);
        return;
    }

    status = snprintf(command, sizeof(command), "tar -tf %s | diff - %s%s",
            l1g_package, test_files_loc, l1g_package_file_list);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the L1G package file");
        return;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found with Format & Package "
                "output. Review %s and %s%s for details", l1g_package_file_list,
                test_files_loc, l1g_package_file_list);
    }

    /* Check the product_data.txt file, using sed to ignore unimportant
       differences and patch up some IAS versus LPGS differences. */
    status = snprintf(command, sizeof(command),
        "sed"
        " -e 's|L0RP_PACKAGE_FILE_DATE=.*$|L0RP_PACKAGE_FILE_DATE=|'"
        " -e 's|L0RP_PACKAGE_FILE_SIZE=.*$|L0RP_PACKAGE_FILE_SIZE=|'"
        " -e 's|L0RP_FILE_CHECKSUM=.*$|L0RP_FILE_CHECKSUM=|'"
        " -e 's|L0RP_CHECKSUM_FILE_DATE=.*$|L0RP_CHECKSUM_FILE_DATE=|'"
        " -e 's|L1_PACKAGE_FILE_DATE=.*$|L1_PACKAGE_FILE_DATE=|'"
        " -e 's|L1_PACKAGE_FILE_SIZE=.*$|L1_PACKAGE_FILE_SIZE=|'"
        " -e 's|L1_PACKAGE_FILE_CHECKSUM=.*$|L1_PACKAGE_FILE_CHECKSUM=|'"
        " -e 's|L1_CHECKSUM_FILE_DATE=.*$|L1_CHECKSUM_FILE_DATE=|'"
        " -e 's|PROCESSING_HOST=.*$|PROCESSING_HOST=|'"
        " -e 's|PROCESSING_SOFTWARE_VERSION=IAS|"
            "PROCESSING_SOFTWARE_VERSION=LPGS|'"
        " -e 's|PRODUCT_ID=.*$|"
            "PRODUCT_ID=LC08_L1GT_040033_20100825_DDDDDDDD_01_T1|'"
        " %s | diff - %s%s",
        product_data_file, test_files_loc, product_data_file_ref);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the product_data.txt file");
        return;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found with Format & Package "
                "output. Review %s and %s%s for details", product_data_file,
                test_files_loc, product_data_file_ref);
        return;
    }

    /* Check the L1 checksum file */
    status = snprintf(command, sizeof(command), 
             "awk '{print substr($0,33)}' %s | diff -q - %s%s"
             , l1g_package_checksum, test_files_loc, l1g_package_checksum);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the L1G checksum file");
        return;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found with Format & Package "
                "output. Checksum file %s is invalid", l1g_package_checksum);
    }

    /* Check the L0R package file by comparing the file lists */
    status = snprintf(command, sizeof(command), "tar -tf %s | diff -q - %s%s",
            l0r_package, test_files_loc, l0r_package_file_list);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the L0R package file");
        return;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found with Format & Package "
                "output. Review %s and %s%s for details", l0r_package,
                test_files_loc, l0r_package);
    }

    /* Check the L0R checksum file */
    status = snprintf(command, sizeof(command), 
             "awk '{print substr($0,33)}' %s | diff -q - %s%s"
             , l0r_package_checksum, test_files_loc, l0r_package_checksum);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the L0R checksum file");
        return;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found with Format & Package "
                "output. Checksum file %s is invalid", l0r_package_checksum);
    }

    /* Check the updated omf file */
    status = snprintf(command, sizeof(command), 
             "grep _PACKAGE %s | diff - %s%s"
             , omf_filename, test_files_loc, omf_filename);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the omf file");
        return;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found with Format & Package "
                "output. OMF file %s is invalid", omf_filename);
        return;
    }

    /* Check the L0R checksum file */
    status = snprintf(command, sizeof(command), "diff -q %s %s", 
        ang_filename, ang_filename_orig);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the ANG file");
        return;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found with Format & Package "
                "output. ANG file %s is invalid", ang_filename);
    }

    /* Remove files from this test */
    remove_files(file_remove_list, 10);

    /* Assert that test passed */
    assert_true(1);
    return;
}

Ensure test_l0r_package()
{
    int status;
    char command[PATH_MAX * 2];

    char omf_filename_orig[11] = "WO1234.txt";
    char omf_filename[11] = "WO1234.omf";
    char test_files_loc[PATH_MAX];
    char *file_remove_list[8];     

    /* Files to remove after test */
    file_remove_list[0] = l1g_package;
    file_remove_list[1] = l0r_package;
    file_remove_list[2] = l1g_package_checksum;
    file_remove_list[3] = l0r_package_checksum;
    file_remove_list[4] = product_data_file;
    file_remove_list[5] = l1g_tiff;
    file_remove_list[6] = l1g_meta;
    file_remove_list[7] = omf_filename;

    if (prepare_omf_file(omf_filename_orig, omf_filename) != SUCCESS)
    {
        assert_true_with_message(0, "Error preparing OMF file %s for unit "
                "test data directory path setup", omf_filename_orig);
        return;
    }

    /* Test where only a L0R package was requested */
    status = snprintf(command, sizeof(command), "../format_and_package %s",
            param_filename_l0r);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "to execute the Format & Package application");
        return;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Format & Package failed");
        return;
    }

    /* Now compare the output files with ones previously generated */
    status = snprintf(test_files_loc, PATH_MAX,
           "%s/ias_base/dms/format_and_package/output/WO1234/", unit_test_dir);
    if (status < 0 || status >= PATH_MAX)
    {
        assert_true_with_message(0, "Error preparing path to control "
                "test files");
        return;
    }

    status = snprintf(command, sizeof(command), "tar -tf %s | diff -q - %s%s",
            l0r_package, test_files_loc, l0r_package_file_list);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the L0R package file");
        return;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found with Format & Package "
                "output. Review %s and %s%s for details", l0r_package,
                test_files_loc, l0r_package);
    }

    /* Check the product_data.txt file */
    status = snprintf(command, sizeof(command),
        "sed -e 's|L1_PACKAGE_FILE_DATE=.*$|L1_PACKAGE_FILE_DATE=|'"
        " -e 's|L0RP_PACKAGE_FILE_SIZE=.*$|L0RP_PACKAGE_FILE_SIZE=|'"
        " -e 's|PROCESSING_HOST=.*$|PROCESSING_HOST=|'"
        " -e 's|L1_PACKAGE_FILE_CHECKSUM=.*$|L1_PACKAGE_FILE_CHECKSUM=|'"
        " -e 's|L1_PACKAGE_FILE_SIZE=.*$|L1_PACKAGE_FILE_SIZE=|'"
        " -e 's|L1_CHECKSUM_FILE_DATE=.*$|L1_CHECKSUM_FILE_DATE=|'"
        " -e 's|L0RP_PACKAGE_FILE_DATE=.*$|L0RP_PACKAGE_FILE_DATE=|'"
        " -e 's|L0RP_FILE_CHECKSUM=.*$|L0RP_FILE_CHECKSUM=|'"
        " -e 's|L0RP_CHECKSUM_FILE_DATE=.*$|L0RP_CHECKSUM_FILE_DATE=|'"
        " -e 's|PRODUCT_ID=.*$|"
            "PRODUCT_ID=LC08_L1TP_198047_20130330_DDDDDDDD_01_T1|'"
        " %s | diff - %s%s",
        product_data_file, test_files_loc, product_data_file_l0r_ref);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the product_data.txt file");
        return;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found with Format & Package "
                "output. Review %s and %s%s for details", product_data_file,
                test_files_loc, product_data_file_l0r_ref);
        return;
    }

    /* Check the updated omf file */
    status = snprintf(command, sizeof(command), 
             "grep _PACKAGE %s | diff - %s%s"
             , omf_filename, test_files_loc, omf_filename_l0r);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the omf file");
        return;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found with Format & Package "
                "output. OMF file %s is invalid", omf_filename);
        return;
    }

    /* Remove files created by this test */
    remove_files(file_remove_list, 8);

    /* Assert that test passed */
    assert_true(1);
    return;
}

Ensure test_no_op()
{
    int status;
    char command[PATH_MAX * 2];
    char *file_remove_list[2];     

    /* Files to remove after test */
    file_remove_list[0] = omf_filename;
    file_remove_list[1] = l1g_tiff;

    if (prepare_omf_file(omf_filename_orig, omf_filename) != SUCCESS)
    {
        assert_true_with_message(0, "Error preparing OMF file %s for unit "
                "test data directory path setup", omf_filename_orig);
        return;
    }

    /* No-op test where no formatting or packaging was requested */
    status = snprintf(command, sizeof(command), "../format_and_package %s",
            param_filename_no_op);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "to execute the Format & Package application");
        return;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Format & Package failed");
        return;
    }

    /* Clean up files */
    remove_files(file_remove_list, 2);

    /* Assert that test passed */
    assert_true(1);
    return;
}

Ensure test_format_and_package_float()
{
    int status;
    char command[PATH_MAX * 2];

    char omf_filename_orig[12] = "WOFLT.txt";
    char omf_filename[12] = "WOFLT.omf";
    char ref_product_data_file[26] = {"product_data_V5.txt"};
    char ref_l1g_float_meta[35] = {"LC81980472013089LGN00_MTL_V3.txt"};

    /* Prepare the control file names to compare against */
    char test_files_loc[PATH_MAX];
    char *file_remove_list[7];     

    /* Files to remove created by this test */
    file_remove_list[0] = l1g_float_package;
    file_remove_list[1] = l1g_float_package_checksum;
    file_remove_list[2] = product_data_file;
    file_remove_list[3] = l1g_float_tiff;
    file_remove_list[4] = l1g_float_meta;
    file_remove_list[5] = omf_filename;
    file_remove_list[6] = l1g_float_package_file_list;

    if (prepare_omf_file(omf_filename_orig, omf_filename) != SUCCESS)
    {
        assert_true_with_message(0, "Error preparing OMF file %s for unit "
                "test data directory path setup", omf_filename_orig);
        return;
    }

    /* Now run Format & Package to test it on this image */
    status = snprintf(command, sizeof(command), "../format_and_package %s",
            param_float_filename);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "to execute the Format & Package application");
        return;
    }
    else if (system(command) != 0)
    {
        assert_true_with_message(0, "Format & Package failed.  Command: %s",
            command);
        return;
    }

    /* Now compare the output files with ones previously generated */
    status = snprintf(test_files_loc, sizeof(test_files_loc),
        "%s/ias_base/dms/format_and_package/output/WOFLT/", unit_test_dir);
    if (status < 0 || status >= sizeof(test_files_loc))
    {
        assert_true_with_message(0, "Error preparing path to control "
                "test files");
        return;
    }

    /* Check the band TIFF file */
    status = snprintf(command, sizeof(command), "diff -q %s %s%s", 
        l1g_float_tiff, test_files_loc, l1g_float_tiff);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the L1G tiff file");
        return;
    }
    else if (system(command) != 0)
    {
        assert_true_with_message(0, "Differences found with Format & "
            "Package output. Review %s and %s%s for details", l1g_float_tiff,
            test_files_loc, l1g_float_tiff);
        return;
    }

    /* Check the remaining L1G files */
    status = snprintf(command, sizeof(command), "sed"
        " -e 's|FILE_DATE = .*$|FILE_DATE = |'"
        " -e 's|PROCESSING_SOFTWARE_VERSION = \"IAS|"
        "PROCESSING_SOFTWARE_VERSION = \"LPGS|'"
        " %s | diff -q - %s%s", l1g_float_meta, test_files_loc,
        ref_l1g_float_meta);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the L1G metadata file");
        return;
    }
    else if (system(command) != 0)
    {
        assert_true_with_message(0, "Differences found with Format & Package "
            "output. Review %s and %s%s for details", l1g_float_meta,
            test_files_loc, ref_l1g_float_meta);
        return;
    }

    status = snprintf(command, sizeof(command), "tar -tf %s | diff -q - %s%s",
        l1g_float_package, test_files_loc, l1g_float_package_file_list);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
            "for validating the L1G package file");
        return;
    }
    else if (system(command) != 0)
    {
        assert_true_with_message(0, "Differences found with Format & Package "
            "output. Review %s and %s%s for details", 
            l1g_float_package_file_list, test_files_loc, 
            l1g_float_package_file_list);
        return;
    }

    /* Check the product_data.txt file, using sed to ignore unimportant
       differences and patch up some IAS versus LPGS differences. */
    status = snprintf(command, sizeof(command),
        "sed"
        " -e 's|L0RP_PACKAGE_FILE_DATE=.*$|L0RP_PACKAGE_FILE_DATE=|'"
        " -e 's|L0RP_PACKAGE_FILE_SIZE=.*$|L0RP_PACKAGE_FILE_SIZE=|'"
        " -e 's|L0RP_FILE_CHECKSUM=.*$|L0RP_FILE_CHECKSUM=|'"
        " -e 's|L0RP_CHECKSUM_FILE_DATE=.*$|L0RP_CHECKSUM_FILE_DATE=|'"
        " -e 's|L1_PACKAGE_FILE_DATE=.*$|L1_PACKAGE_FILE_DATE=|'"
        " -e 's|L1_PACKAGE_FILE_SIZE=.*$|L1_PACKAGE_FILE_SIZE=|'"
        " -e 's|L1_PACKAGE_FILE_CHECKSUM=.*$|L1_PACKAGE_FILE_CHECKSUM=|'"
        " -e 's|L1_CHECKSUM_FILE_DATE=.*$|L1_CHECKSUM_FILE_DATE=|'"
        " -e 's|PROCESSING_HOST=.*$|PROCESSING_HOST=|'"
        " -e 's|PRODUCT_ID=.*$|"
            "PRODUCT_ID=LC08_L1TP_198047_20130330_DDDDDDDD_01_T1|'"
        " -e 's|PROCESSING_SOFTWARE_VERSION=IAS|"
            "PROCESSING_SOFTWARE_VERSION=LPGS|'"
        " %s | diff - %s%s",
        product_data_file, test_files_loc, ref_product_data_file);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the product_data.txt file");
        return;
    }
    else if (system(command) != 0)
    {
        assert_true_with_message(0, "Differences found with Format & Package "
                "output. Review %s and %s%s for details", product_data_file,
                test_files_loc, ref_product_data_file);
        return;
    }

     /* Check the L1 checksum file */
    status = snprintf(command, sizeof(command), "awk '{print substr($0,33)}' "
        "%s | diff -q - %s%s", l1g_float_package_checksum, test_files_loc, 
        l1g_float_package_checksum);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the L1G checksum file");
        return;
    }
    else if (system(command) != 0)
    {
        assert_true_with_message(0, "Differences found with Format & Package "
                "output. Checksum file %s is invalid", l1g_package_checksum);
        return;
    }
   
    /* Check the updated omf file */
    status = snprintf(command, sizeof(command), "grep _PACKAGE %s | diff "
        "- %s%s", omf_filename, test_files_loc, omf_filename);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the omf file");
        return;
    }
    else if (system(command) != 0)
    {
        assert_true_with_message(0, "Differences found with Format & Package "
                "output. OMF file %s is invalid", omf_filename);
        return;
    }

    /* Remove files from this test */
    remove_files(file_remove_list, 7);

    /* Assert that test passed */
    assert_true(1);
    return;
}

Ensure test_write_tirs_ssm_quality(void)
{
    int status;
    char command[PATH_MAX * 2];
    char test_files_loc[PATH_MAX];

    char omf_filename_orig[12] = "WOSSM.txt";
    char omf_filename[12] = "WOSSM.omf";

    char param_ssm_filename[19] = {"parameters_ssm.odl"};
    char ssm_l1g_meta[30] = {"LC81940522015091LGN00_MTL.txt"};
    char ssm_l1g_ang[30] = "LC81940522015091LGN00_ANG.txt";
    char ssm_l1g_meta_check[35] = {"LC81940522015091LGN00_MTL_V2.txt"};
    char b10_tiff[30] = "LC81940522015091LGN00_B10.TIF";
    char b11_tiff[30] = "LC81940522015091LGN00_B11.TIF";
    char bqa_tiff[30] = "LC81940522015091LGN00_BQA.TIF";
    char ssm_l1g_package[29] = "LC81940522015091LGN00.tar.gz";
    char ssm_l1g_package_checksum[30] = "LC81940522015091LGN00_MD5.txt";
    char ssm_l1g_package_file_list[PATH_MAX];

    char product_data_file[17] = {"product_data.txt"}; 
    char ref_file[PATH_MAX];

    /* Files to remove created by this test */
    char *file_remove_list[9];     
    file_remove_list[0] = omf_filename;
    file_remove_list[1] = product_data_file;
    file_remove_list[2] = b10_tiff;
    file_remove_list[3] = b11_tiff;
    file_remove_list[4] = bqa_tiff;
    file_remove_list[5] = ssm_l1g_package_checksum;
    file_remove_list[6] = ssm_l1g_meta;
    file_remove_list[7] = ssm_l1g_package;
    file_remove_list[8] = ssm_l1g_ang;

    if (prepare_omf_file(omf_filename_orig, omf_filename) != SUCCESS)
    {
        assert_true_with_message(0, "Error preparing OMF file %s for unit "
                "test data directory path setup", omf_filename_orig);
        return;
    }

    /* Now run Format & Package to test it on this image */
    status = snprintf(command, sizeof(command), "../format_and_package %s",
            param_ssm_filename);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "to execute the Format & Package application");
        return;
    }
    else if (system(command) != 0)
    {
        assert_true_with_message(0, "Format & Package failed. Command: %s",
            command);
        return;
    }

    /* Now compare the output files with expected results */
    status = snprintf(test_files_loc, PATH_MAX,
        "%s/ias_base/dms/format_and_package/output/WOSSM", unit_test_dir);
    if (status < 0 || status >= PATH_MAX)
    {
        assert_true_with_message(0, "Error preparing path to control "
                "test files");
        return;
    }

    /* Check the product_data.txt file, using sed to ignore unimportant
       differences and patch up some IAS versus LPGS differences. */
    status = snprintf(ref_file, PATH_MAX,
        "%s/product_data.txt", test_files_loc);
    if (status < 0 || status >= sizeof(ref_file))
    {
        assert_true_with_message(0, "Error formating reference product data "
            "file name");
        return;
    }

    status = snprintf(command, sizeof(command),
        "sed"
        " -e 's|L1_PACKAGE_FILE_DATE=.*$|L1_PACKAGE_FILE_DATE=|'"
        " -e 's|L1_PACKAGE_FILE_SIZE=.*$|L1_PACKAGE_FILE_SIZE=|'"
        " -e 's|L1_PACKAGE_FILE_CHECKSUM=.*$|L1_PACKAGE_FILE_CHECKSUM=|'"
        " -e 's|L1_CHECKSUM_FILE_DATE=.*$|L1_CHECKSUM_FILE_DATE=|'"
        " -e 's|PROCESSING_HOST=.*$|PROCESSING_HOST=|'"
        " -e 's|L1_PRODUCT_FILE_SIZE=.*$|L1_PRODUCT_FILE_SIZE=|'"
        " -e 's|PROCESSING_SOFTWARE_VERSION=.*$|PROCESSING_SOFTWARE_VERSION=|'"
        " -e 's|PRODUCT_ID=.*$|"
            "PRODUCT_ID=LC08_L1GT_194052_20150401_DDDDDDDD_01_RT|'"
        " %s | diff - %s",
        product_data_file, ref_file);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command to format"
            "the product data file");
        return;
    }
    else if(system(command) != 0)
    {
        assert_true_with_message(0, "Differences in %s and %s",
            product_data_file, ref_file);
        return;
    }

    /* Check the MTL file */
    status = snprintf(ref_file, PATH_MAX,
        "%s/%s", test_files_loc, ssm_l1g_meta_check);
    if (status < 0 || status >= sizeof(ref_file))
    {
        assert_true_with_message(0, "Error formating reference l1g metadata"
            "file name");
        return;
    }
    status = snprintf(command, sizeof(command),
            "sed"
            " -e 's|FILE_DATE = .*$|FILE_DATE = |'"
            " -e 's|PROCESSING_SOFTWARE_VERSION = .*$"
                "|PROCESSING_SOFTWARE_VERSION = |'"
            " %s | diff - %s",
            ssm_l1g_meta, ref_file);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the L1G metadata file");
        return;
    }
    else if(system(command) != 0)
    {
        assert_true_with_message(0, "Differences in %s and %s", ssm_l1g_meta,
            ref_file);
        return;
    }

    /* Check the TIF files */
    status = snprintf(ref_file, sizeof(ref_file), "%s/%s", 
        test_files_loc, b10_tiff);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error formatting file name "
                "of the B10 ref tiff file");
        return;
    }
    diff_files(b10_tiff, ref_file);

    status = snprintf(ref_file, sizeof(ref_file), "%s/%s", 
        test_files_loc, b11_tiff);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error formatting refrence filename"
                "for validating the B11 tiff file");
        return;
    }
    diff_files(b11_tiff, ref_file);

    status = snprintf(ref_file, sizeof(ref_file), "%s/%s", 
        test_files_loc, bqa_tiff);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error formatting refrence filename"
                "for validating the BQA tiff file");
        return;
    }
    diff_files(bqa_tiff, ref_file);

    /* Check the checksum file */
    status = snprintf(command, sizeof(command), "awk '{print substr($0,33)}' "
        "%s | diff - %s/%s", ssm_l1g_package_checksum, test_files_loc, 
        ssm_l1g_package_checksum);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the L1G checksum file");
        return;
    }
    else if (system(command) != 0)
    {
        assert_true_with_message(0, "Differences found with Format & Package "
                "output. Checksum file %s is invalid", l1g_package_checksum);
        return;
    }

    /* Check the package's file_list */
    status = snprintf(ssm_l1g_package_file_list,
        sizeof(ssm_l1g_package_file_list), "%s/%s", test_files_loc,
        "package_file_list.txt");
    status = snprintf(command, sizeof(command), "tar -tf %s | diff - %s",
            ssm_l1g_package, ssm_l1g_package_file_list);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the L1G package file");
        return;
    }
    else if (system(command) != 0)
    {
        assert_true_with_message(0, "Differences found with Format & Package "
                "output. Review %s and %s for details",
                ssm_l1g_package, ssm_l1g_package_file_list);
        return;
    }

    /* Assert that test passed */
    assert_true(1);

    /* Remove files from this test */
    remove_files(file_remove_list, 9);
}

Ensure test_write_albers_metadata(void)
{
    int status;
    char command[PATH_MAX * 2];
    char test_files_loc[PATH_MAX];

    char omf_filename_orig[10] = "WOALB.txt";
    char omf_filename[10] = "WOALB.omf";
    char param_albers_filename[22] = {"parameters_albers.odl"};
    char albers_l1g_meta[30] = {"LC80400302013078LGN00_MTL.txt"};
    char b4_tiff[30] = "LC80400302013078LGN00_B4.TIF";
    char md5_filename[40] = "LC80400302013078LGN00_MD5.txt";
    char tar_filename[40] = "LC80400302013078LGN00.tar.gz";
    char db_load_filename[20] = "product_data.txt";
    char package_list_filename[20] = "package_list.txt";

    char ref_file[PATH_MAX];

    /* Files to remove created by this test */
    char *file_remove_list[7];     
    file_remove_list[0] = omf_filename;
    file_remove_list[1] = b4_tiff;
    file_remove_list[2] = albers_l1g_meta;
    file_remove_list[3] = package_list_filename;
    file_remove_list[4] = db_load_filename;
    file_remove_list[5] = tar_filename;
    file_remove_list[6] = md5_filename;

    if (prepare_omf_file(omf_filename_orig, omf_filename) != SUCCESS)
    {
        assert_true_with_message(0, "Error preparing OMF file %s for unit "
                "test data directory path setup", omf_filename_orig);
        return;
    }

    /* Now run Format & Package to test it on this image */
    status = snprintf(command, sizeof(command), "../format_and_package %s",
            param_albers_filename);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "to execute the Format & Package application");
        return;
    }
    else if (system(command) != 0)
    {
        assert_true_with_message(0, "Format & Package failed. Command: %s",
            command);
        return;
    }

    /* Now compare the output files with expected results */
    status = snprintf(test_files_loc, PATH_MAX,
        "%s/ias_base/dms/format_and_package/output/WOALB/", unit_test_dir);
    if (status < 0 || status >= PATH_MAX)
    {
        assert_true_with_message(0, "Error preparing path to control "
                "test files");
        return;
    }

    /* Check the MTL file */
    status = snprintf(ref_file, PATH_MAX,
        "%s/%s", test_files_loc, albers_l1g_meta);
    if (status < 0 || status >= sizeof(ref_file))
    {
        assert_true_with_message(0, "Error formating reference l1g metadata"
            "file name");
        return;
    }
    status = snprintf(command, sizeof(command),
            "sed"
            " -e 's|FILE_DATE = .*$|FILE_DATE = |'"
            " -e 's|PROCESSING_SOFTWARE_VERSION = .*$"
                "|PROCESSING_SOFTWARE_VERSION = |'"
            " %s | diff - %s",
            albers_l1g_meta, ref_file);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the L1G metadata file");
        return;
    }
    else if(system(command) != 0)
    {
        assert_true_with_message(0, "Differences in %s and %s", albers_l1g_meta,
            ref_file);
        return;
    }

    /* Check the Band 4 TIF files */
    status = snprintf(ref_file, sizeof(ref_file), "%s%s", 
        test_files_loc, b4_tiff);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error formatting file name "
                "of the B8 ref tiff file");
        return;
    }
    diff_files(b4_tiff, ref_file);

    // Check the tar file
    status = snprintf(command, sizeof(command), "tar -tf %s | diff - %s%s",
            tar_filename, test_files_loc, package_list_filename);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the L1G package file");
        return;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found with Format & Package "
                "output. Review %s and %s%s for details", package_list_filename,
                test_files_loc, package_list_filename);
    }

    /* Check the product_data.txt file, using sed to ignore unimportant
       differences and patch up some IAS versus LPGS differences. */
    status = snprintf(command, sizeof(command),
        "sed"
        " -e 's|L1_PACKAGE_FILE_DATE=.*$|L1_PACKAGE_FILE_DATE=|'"
        " -e 's|L1_PACKAGE_FILE_SIZE=.*$|L1_PACKAGE_FILE_SIZE=|'"
        " -e 's|L1_PACKAGE_FILE_CHECKSUM=.*$|L1_PACKAGE_FILE_CHECKSUM=|'"
        " -e 's|L1_CHECKSUM_FILE_DATE=.*$|L1_CHECKSUM_FILE_DATE=|'"
        " -e 's|PROCESSING_HOST=.*$|PROCESSING_HOST=|'"
        " -e 's|L1_PRODUCT_FILE_SIZE=111181477|L1_PRODUCT_FILE_SIZE=111181478|'"
        " -e 's|PROCESSING_SOFTWARE_VERSION=.*$|"
            "PROCESSING_SOFTWARE_VERSION=|'"
        " -e 's|PRODUCT_ID=.*$|"
            "PRODUCT_ID=LC08_L1TP_040030_20130319_20160728_01_A1|'"
        " %s | diff - %s/%s",
        db_load_filename, test_files_loc, db_load_filename);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the product_data.txt file");
        return;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found with Format & Package "
                "output. Review %s and %s%s for details", db_load_filename,
                test_files_loc, db_load_filename);
        return;
    }

    /* Check the L1 checksum file */
    status = snprintf(command, sizeof(command), 
             "awk '{print substr($0,33)}' %s | diff - %s%s"
             , md5_filename, test_files_loc, md5_filename);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating the L1G checksum file");
        return;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found with Format & Package "
                "output. Checksum file %s is invalid", md5_filename);
        return;
    }

    /* Remove files from this test */
    remove_files(file_remove_list, 7);
}

/**************************************************************************
 *  Function: diff_files
 *  Purpose: Compare 2 files. Return SUCCESS if same, ERROR if different.
 **************************************************************************/
static int diff_files
(
    const char *file1,
    const char *file2
)
{
    char command[PATH_MAX * 2];
    int status = snprintf(command, sizeof(command), "diff -q %s %s", 
        file1, file2);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "for validating files %s and %s", file1, file2);
        return ERROR;
    }
    else if (system(command) != 0)
    {
        assert_true_with_message(0, "Differences found in files. "
            "Review %s and %s for details", file1, file2);
        return ERROR;
    }

    return SUCCESS;
}

/**************************************************************************
 *  Function: prepare_omf_file
 *  Purpose: replace unit test data directory environment variables with
 *           actual path
 **************************************************************************/
static int prepare_omf_file
(
    char *omf_filename_orig,
    char *omf_filename
)
{

    int status;
    char command[PATH_MAX * 2];

    /* Get the OMF file ready by replacing unit test data directory
       environment variables with the actual path */
    status = snprintf(command, sizeof(command), 
            "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' %s > %s", unit_test_dir,
            omf_filename_orig, omf_filename);
    if (status < 0 || sizeof(command) <= status)
    {
        assert_true_with_message(0, "Error preparing system command "
                "to prepare unit test data directory in OMF");
        return ERROR;
    }
    
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Error preparing OMF file %s for unit "
                "test data directory path setup", omf_filename_orig);
        return ERROR;
    }

    return SUCCESS;
}

/**************************************************************************
 *  Function: prepare_ang_file
 *  Purpose: copy the ang template file to the correct filename
 **************************************************************************/
static int prepare_ang_file
(
)
{
    int status;
    char command[PATH_MAX];

    /* Copy the ANG filename to the correct filename */
    status = snprintf(command, sizeof(command), 
        "cp WO1234_ANG.txt.template WO1234_ANG.txt");
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
                "to prepare ANG file copy");
        return ERROR;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Error copying the ANG file from template");
        return ERROR;
    }

    return SUCCESS;
}

/**************************************************************************
 *  Function: remove_files
 *  Purpose: removes a list of files from the local directory
 **************************************************************************/
static int remove_files
(
    char *file_list[],
    int num_files
)
{
    struct stat sb;
    int i;
    
    for (i = 0; i < num_files; i++ )
    {
        if (stat(file_list[i], &sb) == 0)
        {
            if (remove(file_list[i]))
            {
                printf("Error removing file: %s\n", file_list[i]);
                return ERROR;
            }
        }
    }

    return SUCCESS;
}
