/*----------------------------------------------------------------------------
NAME:    write_metadata

PURPOSE: Write values from the metadata structure to an ASCII metadata file
         for the L1G product. If the L1G package was not requested, forgiving
         enough to allow successful completion (with missing metadata) if any
         of the additional files metadata is retrieved from are not provided or
         cannot be successfully read.  This allows a lone L1G file to be run
         through the formatting aspect of the DFP to convert to a GeoTIFF.

RETURNS: L1G metadata filename;
         NULL if error
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include <math.h>
#include <errno.h>
#include "ias_const.h"
#include "ias_types.h"
#include "ias_logging.h"
#include "ias_l0r_constants.h"
#include "ias_miscellaneous.h"
#include "ias_l1r.h"
#include "ias_geo.h"
#include "gctp.h"
#include "ias_grid_io.h"
#include "format_and_package.h"
#include "dfp_metadata.h"

#define DATABASE_LOAD_FILE "product_data.txt"
#define HOSTNAME_LENGTH 21       /* Database host name length is 20 */

/*----------------------------------------------------------------------------
NAME:    adjust_tirs_image_quality_for_ssm_switch

PURPOSE: Adjust the TIRS image quality value to account for the dropped
         pixels when the image SSM position status is "SWITCHED"

RETURNS: Integer - Adjusted TIRS image quality value
------------------------------------------------------------------------------*/
static int adjust_tirs_image_quality_for_ssm_switch
(
    int image_quality,      /* I: Current TIRS image quality */
    double l0r_frame_count  /* I: Number of frames contained in the L0R */
)
{
    /* Using the current image quality adjust the quality by taking the 
       Standard number of frames (SNF) for TIRS and dividing by the Actual
       number of frames (ANF) and multiplying by the number
       of CRC failures in the scene (CRC) divided by the number of failures that
       drops the quality score by 1 point (CFBP). We are using the average 
       number of CRC failures for the SCA gap which result in a value of 1.5 
       Adjusted Image Quality = SNF/ANF * (CRC / CFBP) */
    int adjusted_image_quality = (image_quality - (int)round(((2621.0 
        / l0r_frame_count) * 1.5)));

    /* Ensure the image quality is not below 1 */
    if (adjusted_image_quality < 1)
        adjusted_image_quality = 1;
    
    return adjusted_image_quality;
}

/*----------------------------------------------------------------------------
NAME:    adjust_tirs_image_quality_for_ssm_quality

PURPOSE: Adjusts a tirs quality value based on an ssm quality flag and returns
    the modified tirs quality.

RETURNS: The adjusted quality
-----------------------------------------------------------------------------*/
int adjust_tirs_image_quality_for_ssm_quality
(
    int tirs_quality,
    IAS_SSM_POSITION_QUALITY ssm_position_quality
)
{
    int adjusted_quality = tirs_quality; /* Assume no adjustment needed */

    if (ssm_position_quality == IAS_SSM_POSITION_PREDICTED)
    {
        /* Predicted quality, so subtract 4 from the image quality */
        adjusted_quality = tirs_quality - 4;
    }
    else if (ssm_position_quality == IAS_SSM_POSITION_PRELIMINARY)
    {
        /* Preliminary quality, so subtract 2 from the image quality */
        adjusted_quality = tirs_quality - 2;
    }

    /* Ensure the image quality is not below 1 */
    if (adjusted_quality < 1)
        adjusted_quality = 1;

    return adjusted_quality;
}

/*----------------------------------------------------------------------------
NAME: get_ssm_quality_flag_name

PURPOSE: Each SSM quality flag value has an associated name.

RETURNS: The name of the quality flag.

-----------------------------------------------------------------------------*/
char *get_ssm_quality_flag_name
(
    IAS_SSM_POSITION_QUALITY ssm_quality
)
{
    if (ssm_quality == 0 || ssm_quality == 1)
    {
        return "PRELIMINARY";
    }
    else if (ssm_quality == 2)
    {
        return "FINAL";
    }
    else
    {
        /* The default value is "ACTUAL" */
        return "ACTUAL";
    }
}

/*----------------------------------------------------------------------------
NAME:    get_processing_software_version

PURPOSE: Gets processing_software_version which is a combination of the
         system running (IAS or LPGS) and the software version.
         Has the format "LPGS_XX.XX.XX" or "IAS_XX.XX.XX"

RETURNS: SUCCESS if processing software version retrieved successfully
         ERROR   if problem getting processing software version
-----------------------------------------------------------------------------*/
static int get_processing_software_version
(
    char *software_version,     /* O: software version */
    int version_size            /* I: bytes available in software version
                                      buffer */
)
{
    const static char ias_prefix[] = "IAS_";
    const static char lpgs_prefix[] = "LPGS_";
    const static char no_prefix[] = "";
    const char *prefix;
    int status;

    /* check if we are LPGS or IAS */
    switch (ias_misc_get_processing_system())
    {
        case IAS_PROCESSING_SYSTEM_IAS:
            prefix = ias_prefix;
            break;

        case IAS_PROCESSING_SYSTEM_LPGS:
            prefix = lpgs_prefix;
            break;

        default:
           prefix = no_prefix;
           break;
    }

    status = snprintf(software_version, version_size, "%s%s", prefix,
        ias_misc_get_software_version()); 
    if (status < 0 || status >= version_size)
    {
        IAS_LOG_ERROR("Buffer for software version too small");
        return ERROR;
    }

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:    get_oli_image_quality

PURPOSE: Calculates the OLI image quality

RETURNS: OLI image quality
-----------------------------------------------------------------------------*/
static int get_oli_image_quality
(
    const DFP_METADATA *dfp_meta,     /* I: Meta for L1G from misc sources */
    const IAS_L0R_SCENE_METADATA *l0r_scene_meta /* I: L0R scene metadata
                                                     structure */
)
{
    int image_quality_oli = l0r_scene_meta->image_quality_oli;

    /* Lower the image quality for daytimes scenes when the lower 12
       bits are used (if greater than a quality of 6 already) */
    if (dfp_meta->is_oli_lower_12_bits_flag
        && l0r_scene_meta->sun_elevation >= 0 && image_quality_oli > 6)
    {
       image_quality_oli = 6;
    }

    return image_quality_oli;
}

/*----------------------------------------------------------------------------
NAME:    get_tirs_image_quality

PURPOSE: Calculates the TIRS image quality

RETURNS: TIRS image quality
-----------------------------------------------------------------------------*/
static int get_tirs_image_quality
(
    const DFP_METADATA *dfp_meta,     /* I: Meta for L1G from misc sources */
    const IAS_L0R_SCENE_METADATA *l0r_scene_meta /* I: L0R scene metadata
                                                     structure */
)
{
    int image_quality_tirs = adjust_tirs_image_quality_for_ssm_quality(
        l0r_scene_meta->image_quality_tirs, dfp_meta->ssm_position_quality);

    if (strcmp(dfp_meta->ssm_position_status, "SWITCHED") == 0)
    {
        /* Adjust the quality value to account for the dropped
           pixels when the SSM switched modes */
        image_quality_tirs = adjust_tirs_image_quality_for_ssm_switch(
            image_quality_tirs, l0r_scene_meta->scene_stop_frame_tirs
            - l0r_scene_meta->scene_start_frame_tirs);
    }

    return image_quality_tirs;
}

/*----------------------------------------------------------------------------
NAME:    get_elevation_source

PURPOSE: Gets the elevation source from the DEM file

RETURNS: SUCCESS if elevation source retrieved successfully
                 or if not retrieved but L1G package not requested
                 (to allow GeoTIFF conversion of lone L1G)

         ERROR   if problem getting elevation source from DEM
-----------------------------------------------------------------------------*/
static int get_elevation_source
(
    const char *correction_type,/* I: The L1G file metadata correction type */
    int want_l1g_package,       /* I */
    DFP_METADATA *dfp_meta      /* I/O: DEM filename; elevation source saved */
)
{
    L1GIO *dem_ptr;                 /* I/O file pointer for DEM file */
    IAS_L1G_FILE_METADATA dem_meta; /* DEM file metadata to get elev source */
    int status;

    /* Initialize so we have a valid value in case of errors */
    strcpy(dfp_meta->elevation_source, "");

    if (strcmp(correction_type, "L1G") == 0)
    {
        /* If L1G product, there was no terrain correction applied, so
           elevation source is set to 'N' (no correction) */
        strcpy(dfp_meta->elevation_source, "N");
        return SUCCESS;
    }

    /* Check if DEM filename was provided in the OMF */
    if (strcmp(dfp_meta->dem_filename, "") == 0)
    {
        if (!want_l1g_package)
        {
            IAS_LOG_WARNING("DEM not provided; unable to determine elevation "
                    "source");
            return SUCCESS;
        }
        IAS_LOG_ERROR("DEM not provided; unable to determine elevation "
                "source");
        return ERROR;
    }

    dem_ptr = ias_l1g_open_image(dfp_meta->dem_filename, IAS_READ);
    if (dem_ptr == NULL)
    {
        if (!want_l1g_package)
        {
            IAS_LOG_WARNING("Opening DEM file %s; unable to determine "
                    "elevation source", dfp_meta->dem_filename);
            return SUCCESS;
        }
        IAS_LOG_ERROR("Opening DEM file %s", dfp_meta->dem_filename);
        return ERROR;
    }

    status = ias_l1g_get_file_metadata(dem_ptr, &dem_meta);

    /* Close the image no matter if successful or not since done with it */
    ias_l1g_close_image(dem_ptr);

    if (status != SUCCESS)
    {
        if (!want_l1g_package)
        {
            IAS_LOG_WARNING("Retrieving DEM file metadata for %s; unable to "
                    "determine elevation source", dfp_meta->dem_filename);
            return SUCCESS;
        }
        IAS_LOG_ERROR("Retrieving DEM file metadata for %s",
                dfp_meta->dem_filename);
        return ERROR;
    }

    if (strcasecmp(dem_meta.spacecraft, "GTOPO30") == 0)
    {
        strcpy(dfp_meta->elevation_source, "GTOPO30");
    }
    else if (strcasecmp(dem_meta.spacecraft, "GLS_DEM") == 0)
    {
        strcpy(dfp_meta->elevation_source, "GLS2000");
    }
    else if (strcasecmp(dem_meta.spacecraft, "RAMP") == 0)
    {
        strcpy(dfp_meta->elevation_source, "RAMP");
    }
    else
    {
        if (!want_l1g_package)
        {
            IAS_LOG_WARNING("DEM file contained unknown elevation source %s; "
                    "unable to determine elevation source using DEM file %s",
                    dem_meta.spacecraft, dfp_meta->dem_filename);
            return SUCCESS;
        }
        IAS_LOG_ERROR("DEM file contained unknown elevation source %s in "
                "file %s", dem_meta.spacecraft, dfp_meta->dem_filename);
        return ERROR;
    }

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:    get_scene_center_time

PURPOSE: Gets the scene center time from the L0R. If the L0R was previously
         determined to not be available, the L0R metadata structure will not
         contain valid data. Check the L0R meta available flag and if not set,
         let the routine return with the scene center time initialized to an
         empty string.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int get_scene_center_time
(
    const char *date_acquired,                  /* I */
    int l0r_meta_available,                     /* I */
    char scene_center_time[SCENE_CENTER_SIZE]   /* O */
)
{
    int status;
    char *tmp_ptr;

    /* Initialize so we have a valid value in case of errors */
    strcpy(scene_center_time, "");

    /* If L0R not provided (or unreadable) and the L1G package was not
       requested, be forgiving enough to let the L1G still be converted
       without errors. */
    if (!l0r_meta_available)
        return SUCCESS;

    /* Move pointer to ':' after the year, before day of year */
    tmp_ptr = strchr(date_acquired, ':');
    if (tmp_ptr == NULL)
    {
        IAS_LOG_ERROR("Extracting time component from L0R scene center time; "
                "unexpected string format");
        return ERROR;
    }

    /* Step past the ':' */
    tmp_ptr++;

    /* Move pointer further down the string past the day of year, before hour */
    tmp_ptr = strchr(tmp_ptr, ':');
    if (tmp_ptr == NULL)
    {
        IAS_LOG_ERROR("Extracting time component from L0R scene center time; "
                "unexpected string format");
        return ERROR;
    }

    /* Step past the ':' */
    tmp_ptr++;

    /* Limit the size copied to one less than max size. The strcat done to
       add on the 'Z' ("zulu time" - DFCB requirement) can then be done
       confidently also. */

    status = snprintf(scene_center_time, (SCENE_CENTER_SIZE - 1), "%s",
            tmp_ptr);
    if (status < 0 || status >= (SCENE_CENTER_SIZE - 1))
    {
        IAS_LOG_ERROR("Setting scene center time; snprintf");
        return ERROR;
    }
    strcat(scene_center_time, "Z");

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:    print_tirs_thermal_constants

PURPOSE: Gets the TIRS thermal constants from CPF and print their values to the
         metadata file.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int print_tirs_thermal_constants
(
    FILE *file,                               /* I */
    const struct IAS_CPF_TIRS_THERMAL_CONSTANTS *tirs_thermal_constants    
                                              /* I */
)
{
    int band_list[IAS_TIRS_MAX_NBANDS]; /* list of band numbers */
    int band_index;
    int nbands;                    /* number of tirs normal bands */
    int status;                    /* Status of return from function */


    /* get tirs band info */
    status = ias_sat_attr_get_sensor_band_numbers(IAS_TIRS, IAS_NORMAL_BAND, 
                                                  0, band_list,
                                                  IAS_TIRS_MAX_NBANDS, &nbands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting TIRS sat band attributes");
        return ERROR;
    }

    /* move the tirs data to the common structure elements */
    for (band_index = 0; band_index < nbands; band_index++)
    {
        if (fprintf(file, "    K1_CONSTANT_BAND_%d = %.4f\n" 
                          "    K2_CONSTANT_BAND_%d = %.4f\n", 
                    band_list[band_index], 
                    tirs_thermal_constants->k1_constant[band_index],
                    band_list[band_index], 
                    tirs_thermal_constants->k2_constant[band_index]) < 0)
        {
            IAS_LOG_ERROR("Writing TIRS thermal constant values for band %d",
                           band_list[band_index]);
            return ERROR;
        }
    }

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:    setup_corners_array

PURPOSE: Sets up an array of corner points from input L1G band metadata
         structure in this order: ULY,ULX,URY,URX,LLY,LLX,LRY,LRX.

RETURNS: nothing
-----------------------------------------------------------------------------*/
static void setup_corners_array
(
    const IAS_L1G_BAND_METADATA *band_meta,   /* I */
    double *comparison_corners          /* O */
)
{
    comparison_corners[0] = band_meta->upper_left_y;
    comparison_corners[1] = band_meta->upper_left_x;
    comparison_corners[2] = band_meta->upper_right_y;
    comparison_corners[3] = band_meta->upper_right_x;
    comparison_corners[4] = band_meta->lower_left_y;
    comparison_corners[5] = band_meta->lower_left_x;
    comparison_corners[6] = band_meta->lower_right_y;
    comparison_corners[7] = band_meta->lower_right_x;
}

/*----------------------------------------------------------------------------
NAME:    get_l1g_meta

PURPOSE: Gets some more metadata needed from the L1G file

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int get_l1g_meta
(
    const IAS_L1G_FILE_METADATA *l1g_meta,          /* I: L1G file meta */
    const L1G_BAND_METADATA *l1g_band_meta,         /* I */
    DFP_METADATA *dfp_meta,                         /* IO */
    int *control_band_index                         /* O */
)
{
    double tol = 1.5;       /* meters - 1.5 is roughly a tenth of a pixel */
    int i;
    int band_number;
    int band_index;
    int status;
    double control_corners[8];  /* ULY,ULX,URY,URX,LLY,LLX,LRY,LRX */
    int in_units;           /* Code for the input projection units */
    double geo_parms[IAS_PROJ_PARAM_SIZE];  /* Geo projection parameters */
    int set_reflective = 0;
    int set_thermal = 0;
    IAS_GEO_PROJ_TRANSFORMATION *map2degrees = NULL;
    IAS_PROJECTION map_proj;
    IAS_PROJECTION geo_proj;

    *control_band_index = -1;

    /* Find the first valid band to use as the control band to compare the
       other bands' corners against */
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        if (l1g_band_meta[band_index].band_valid)
        {
            *control_band_index = band_index;
            break;
        }
    }

    if (*control_band_index == -1)
    {
        IAS_LOG_ERROR("Finding valid band to verify corner coordinates of "
                "L1G bands");
        return ERROR;
    }

    /* Save the control bands projection coordinates for easy comparison */
    control_corners[0]
        = l1g_band_meta[*control_band_index].band_meta.upper_left_y;
    control_corners[1]
        = l1g_band_meta[*control_band_index].band_meta.upper_left_x;
    control_corners[2]
        = l1g_band_meta[*control_band_index].band_meta.upper_right_y;
    control_corners[3]
        = l1g_band_meta[*control_band_index].band_meta.upper_right_x;
    control_corners[4]
        = l1g_band_meta[*control_band_index].band_meta.lower_left_y;
    control_corners[5]
        = l1g_band_meta[*control_band_index].band_meta.lower_left_x;
    control_corners[6]
        = l1g_band_meta[*control_band_index].band_meta.lower_right_y;
    control_corners[7]
        = l1g_band_meta[*control_band_index].band_meta.lower_right_x;

    /* Compare all corners of bands after the first valid band for significant
       differences to indicate an error */
    for (band_index = (*control_band_index + 1); band_index < IAS_MAX_NBANDS;
            band_index++)
    {
        if (l1g_band_meta[band_index].band_valid)
        {
            double comparison_corners[8];

            setup_corners_array(&l1g_band_meta[band_index].band_meta,
                    comparison_corners);

            /* Loop over the 8 individual corner point components to compare
               them to the control band */
            for (i = 0; i < 8; i++)
            {
                double diff;
                /* Do the validation for this coordinate component */
                diff = comparison_corners[i] - control_corners[i];

                if (fabs(diff) > tol)
                {
                    IAS_LOG_ERROR("Corner point exceeds tolerance from control "
                            "band data");
                    return ERROR;
                }
            }
        }
    }

    /* Corner projection coordinates (X/Y) passed tolerance validations, so
       transform to get the product corner coordinates (lat/long) */
    status = ias_geo_get_units(l1g_meta->projection_units, &in_units);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Converting projection units %s to unit code",
                l1g_meta->projection_units);
        return ERROR;
    }

    /* Initialize geo_parms */
    for (i = 0; i < IAS_PROJ_PARAM_SIZE; i++)
    {
        geo_parms[i] = 0.0;
    }

    ias_geo_set_projection(GEO, NULLZONE, DEGREE, l1g_meta->spheroid_code,
            geo_parms, &geo_proj);
    ias_geo_set_projection(l1g_meta->projection_code, l1g_meta->zone_code,
            in_units, l1g_meta->spheroid_code, l1g_meta->projection_parameters,
            &map_proj);

    map2degrees = ias_geo_create_proj_transformation(&map_proj, &geo_proj);
    if (!map2degrees)
    {
        IAS_LOG_ERROR("Creating projection transformation");
        return ERROR;
    }

    /* Loop over the 4 corner coordinates and transform each */
    for (i = 0; i < 4; i++)
    {
        /* Setup pointers and the error message string to avoid rebuilding
           the transformation function call 4 times for each corner. Doesn't
           save a lot of code (line count), but hopefully increases
           readability and understanding. */
        double x = 0;
        double y = 0;
        double *lat_ptr = NULL;
        double *lon_ptr = NULL;
        char err_corner[12] = {""};

        switch (i)
        {
            case 0:
                /* UL */
                x = l1g_band_meta[*control_band_index].band_meta.upper_left_x;
                y = l1g_band_meta[*control_band_index].band_meta.upper_left_y;
                lat_ptr = &dfp_meta->ul_lat_lon[0];
                lon_ptr = &dfp_meta->ul_lat_lon[1];
                strcpy(err_corner, "upper left");
                break;
            case 1:
                /* UR */
                x = l1g_band_meta[*control_band_index].band_meta.upper_right_x;
                y = l1g_band_meta[*control_band_index].band_meta.upper_right_y;
                lat_ptr = &dfp_meta->ur_lat_lon[0];
                lon_ptr = &dfp_meta->ur_lat_lon[1];
                strcpy(err_corner, "upper right");
                break;
            case 2:
                /* LL */
                x = l1g_band_meta[*control_band_index].band_meta.lower_left_x;
                y = l1g_band_meta[*control_band_index].band_meta.lower_left_y;
                lat_ptr = &dfp_meta->ll_lat_lon[0];
                lon_ptr = &dfp_meta->ll_lat_lon[1];
                strcpy(err_corner, "lower left");
                break;
            case 3:
                /* LR */
                x = l1g_band_meta[*control_band_index].band_meta.lower_right_x;
                y = l1g_band_meta[*control_band_index].band_meta.lower_right_y;
                lat_ptr = &dfp_meta->lr_lat_lon[0];
                lon_ptr = &dfp_meta->lr_lat_lon[1];
                strcpy(err_corner, "lower right");
                break;
            default:
                IAS_LOG_ERROR("Internal programming error during calculation "
                        "of corner coordinates");
                ias_geo_destroy_proj_transformation(map2degrees);
                return ERROR;
        }

        /* This gets us the lat/long from the X/Y */
        status = ias_geo_transform_coordinate(map2degrees, x, y,
                lon_ptr, lat_ptr);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Converting %s corner to the geographic projection",
                    err_corner);
            ias_geo_destroy_proj_transformation(map2degrees);
            return ERROR;
        }
    }

    ias_geo_destroy_proj_transformation(map2degrees);
    map2degrees = NULL;

    /* Get number of lines/samples information for each band type */
    /* Initialize values so we have a known value for anything unavailable */
    dfp_meta->pan_lines = -1;
    dfp_meta->pan_samples = -1;
    dfp_meta->reflective_lines = -1;
    dfp_meta->reflective_samples = -1;
    dfp_meta->thermal_lines = -1;
    dfp_meta->thermal_samples = -1;

    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        if (l1g_band_meta[band_index].band_valid)
        {
            band_number = ias_sat_attr_convert_band_index_to_number(band_index);
            if (band_number == ERROR)
            {
                IAS_LOG_ERROR("Converting band index %d to band number",
                        band_index);
                return ERROR;
            }

            /* Get lines/samples from first valid band that is encountered
               of each band type */
            const IAS_BAND_ATTRIBUTES *band_attribs;

            band_attribs = ias_sat_attr_get_band_attributes(band_number);
            if (band_attribs == NULL)
            {
                IAS_LOG_ERROR("Getting band attributes for band %d",
                        band_number);
                return ERROR;
            }

            /* Set the number of lines and samples from the spectral type */
            if (band_attribs->spectral_type == IAS_SPECTRAL_PAN)
            {
                dfp_meta->pan_lines = l1g_band_meta[band_index].num_lines;
                dfp_meta->pan_samples = l1g_band_meta[band_index].num_samples;
            }
            if (!set_reflective
                    && ((band_attribs->spectral_type == IAS_SPECTRAL_VNIR)
                        || (band_attribs->spectral_type == IAS_SPECTRAL_SWIR)))
            {
                dfp_meta->reflective_lines
                    = l1g_band_meta[band_index].num_lines;
                dfp_meta->reflective_samples
                    = l1g_band_meta[band_index].num_samples;
                set_reflective = 1;
            }
            if (!set_thermal
                    && (band_attribs->spectral_type == IAS_SPECTRAL_THERMAL))
            {
                dfp_meta->thermal_lines = l1g_band_meta[band_index].num_lines;
                dfp_meta->thermal_samples
                    = l1g_band_meta[band_index].num_samples;
                set_thermal = 1;
            }
        }
    }

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:    convert_date_to_odl_style

PURPOSE: Given a date string of the format YYYY_MM_DD, where _ is any single
         printable non-digit character, convert the date to an ODL complient
         date. (YYYY-MM-DD)

RETURNS: SUCCESS if the string could be converted, otherwise ERROR
-----------------------------------------------------------------------------*/
static int convert_date_to_odl_style
(
    const char *old_date,    /* I */
    char *new_date           /* O */
)
{
    const int date_length = 11; /* length of YYYY-MM-DD + one for NULL term */
    const int delimiter_count = 2;
    const int delimiter_positions[] = {4,7}; /* These positions
                                             are non-digits for valid input */
    const int digit_count = 8;
    const int digit_positions[] = {0,1,2,3,5,6,8,9}; /* These
                                        positions are digits for valid input */

    int i = 0;

    if (old_date == NULL || new_date == NULL)
    {
        IAS_LOG_ERROR("Null pointer passed");
        return ERROR;
    }

    if (strlen(old_date) < (date_length-1))
    {
        IAS_LOG_ERROR("Date string passed (%s) is to short", old_date);
        return ERROR;
    }

    /* Verify the delimitors */
    for (i = 0; i < delimiter_count; i++)
    {
        int position = delimiter_positions[i];
        char check_char = old_date[position];

        if (check_char < ' ' || (check_char >= '0' && check_char <= '9'))
        {
            IAS_LOG_ERROR("Invalid deliminator of %c found in position "
               "%i of date", check_char, position);
            return ERROR;
        }
    }

    /* Verify the digits */
    for (i = 0; i < digit_count; i++)
    {
        int position = digit_positions[i];
        char check_char = old_date[position];

        if (!(check_char >= '0' && check_char <= '9'))
        {
            IAS_LOG_ERROR("Invalid digit of %c found in position "
               "%i of date", check_char, position);
            return ERROR;
        }
    }

    /* copy to the output string */
    strncpy(new_date,old_date,date_length-1);
    new_date[date_length-1] = '\0';

    /* Finally, change the delimitors */
    for (i = 0; i < delimiter_count; i++)
    {
        int position = delimiter_positions[i];

        new_date[position] = '-';
    }

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:    extract_filename_from_path

PURPOSE: Given a string with filename, isolate the filename from path component

RETURNS: nothing
-----------------------------------------------------------------------------*/
static void extract_filename_from_path
(
    const char *input_path_filename,    /* I */
    char *filename                      /* O */
)
{
    char *tmp_ptr = NULL;

    /* This works with empty strings also - just returns an empty string
       also for the filename only component of the input string */

    tmp_ptr = strrchr(input_path_filename, '/');
    if (tmp_ptr == NULL)
    {
        /* No path component, filename is full string */
        strcpy(filename, input_path_filename);
    }
    else
    {
        /* Copy just the filename out of the string */
        tmp_ptr++;
        strcpy(filename, tmp_ptr);
    }
}


/*----------------------------------------------------------------------------
NAME:    translate_data_type

PURPOSE: Convert the data type as shown below:
         L1T  -> L1TP
         L1Gt -> L1GT
         L1G  -> L1GS

RETURNS: SUCCESS - The input was valid
         ERROR   - The input was not valid. The output buffer is not populated.
-----------------------------------------------------------------------------*/
int translate_data_type
(
    const char *input_type,    /* I: Type: L1T/L1Gt/L1G */
    char *output_type          /* O: Type: L1TP/L1GT/L1GS */
)
{
    if (strcmp(input_type, "L1T") == 0)
        strcpy(output_type, "L1TP");
    else if (strcmp(input_type, "L1Gt") == 0)
        strcpy(output_type, "L1GT");
    else if (strcmp(input_type, "L1G") == 0)
        strcpy(output_type, "L1GS");
    else
    {
        return ERROR;
    }
    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:    build_product_id

PURPOSE: Builds the Product ID using various avaialble metadata.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int build_product_id
(
    const DFP_METADATA *dfp_meta,     /* I: Meta for L1G from misc sources */
    const IAS_L0R_SCENE_METADATA *l0r_scene_meta, /* I: L0R scene metadata
                                                     structure */
    const IAS_L1G_FILE_METADATA *l1g_meta,        /* I: L1G file meta */
    const struct tm *curr_time,                   /* I: Current Time */
    char *product_id       /* O: The Product ID. Provided buffer should be at
                                  least PRODUCT_ID_LENGTH + 1 bytes */
)
{
    char data_type[5] = "";
    char current_date[DATE_LENGTH] = "";
    char tier[3] = "";
    char sensor[2] = "";
    int status;
    IAS_DATETIME scene_time;

    if (translate_data_type(l1g_meta->correction_type, data_type) != SUCCESS)
    {
        IAS_LOG_ERROR("Failed to translate data type %s",
            l1g_meta->correction_type);
        return ERROR;
    }

    determine_tier(dfp_meta, l1g_meta, tier);

    
    if (strftime(current_date, sizeof(current_date), "%Y%m%d", curr_time) == 0)
    {
        IAS_LOG_ERROR("Unable to build date");
        return ERROR;
    }
    if (strcmp(dfp_meta->l0r_oli_present, "Y") == 0 &&
       strcmp(dfp_meta->l0r_tirs_present, "Y") == 0)
    {
       strcpy(sensor, "C");
    }
    else if (strcmp(dfp_meta->l0r_oli_present, "Y") == 0)
       strcpy(sensor, "O");
    else if (strcmp(dfp_meta->l0r_tirs_present, "Y") == 0)
       strcpy(sensor, "T");
    else
    {
        IAS_LOG_ERROR("Unable to determine sensor char");
        return ERROR;
    }
    status = ias_misc_parse_datetime_string(l0r_scene_meta->date_acquired,
                                            IAS_DATETIME_L0R_FORMAT,
                                            &scene_time);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to parseL0R  date: %s",
                      l0r_scene_meta->date_acquired);
        return ERROR;
    }
    status = snprintf(product_id, PRODUCT_ID_LENGTH,
             "L%s%02d_%4s_%03d%03d_%04d%02d%02d_%s_%02d_%s",
             sensor,
             get_satellite(),
             data_type,
             l0r_scene_meta->wrs_path,
             l0r_scene_meta->wrs_row,
             scene_time.year,
             scene_time.month,
             scene_time.day_of_month,
             current_date,
             dfp_meta->collection_number,
             tier);
    if (status < 0 || status >= PRODUCT_ID_LENGTH)
    {
        IAS_LOG_ERROR("Failed to populate product_id");
        return ERROR;
    }
    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:    determine_l1_basename

PURPOSE: Determines what the name the L1 output files with, the scene ID or
         the product ID.

RETURNS: Either the Product ID or the Scene ID, depending on if
         NAME_WITH_PRODUCT_ID is set.
-----------------------------------------------------------------------------*/
const char *determine_l1_basename
(
    const DFP_METADATA *dfp_meta      /* I: Meta for L1G from misc sources */
)
{
    if (get_named_with_prod_id())
        return dfp_meta->product_id;
    else
        return get_l1g_output_basename();
}

/*----------------------------------------------------------------------------
NAME:    determine_tier

PURPOSE: Determines the tier of the data

RETURNS: Nothing
-----------------------------------------------------------------------------*/
void determine_tier
(
    const DFP_METADATA *dfp_meta,      /* I: Meta for L1G from misc sources */
    const IAS_L1G_FILE_METADATA *l1g_meta,          /* I: L1G file meta */
    char *tier                  /* O: The tier. Provided buffer should be at
                                     at least 3 bytes */
)
{
    if (strcmp(dfp_meta->l1_tirs_present,"Y") == 0 &&
        dfp_meta->ssm_position_quality != IAS_SSM_POSITION_FINAL &&
        dfp_meta->ssm_position_quality != IAS_SSM_POSITION_ACTUAL)
    {
        strcpy(tier, "RT");
        /* Nothing else to consider if we don't have actual/final 
         * SSM positions */
        return;
    }

    /* If an L1T was made, check whether the RMSE value is good enough to
       consider it a Tier 1 product */
    if (dfp_meta->ground_control_points_model > 0 &&
       dfp_meta->geometric_rmse_model <= get_tier_1_rmse_threshold())
    {
        strcpy(tier, "T1");
    }
    else
    {
        strcpy(tier, "T2");
    }

    /* Check the project for being LCMAP to change the tier to indicate
       Albers projection */
    if (strcmp(get_project(), "LCMAP") == 0)
    {
        /* Replace the first letter with A to indicate Albers */
        tier[0] = 'A';
    }
}

/*----------------------------------------------------------------------------
NAME:    print_parameter

PURPOSE: Prints the passed parameter name and value to the file. It also
         prints the specified group name if the value of group_already_outputed
         is not 1. After printing the group name, group_already_outputed
         is set to 1, so subsequent calls using the same variable
         will not have the group outputed.

RETURNS: SUCCESS if the value printed to the file successfully
         ERROR   on general error
-----------------------------------------------------------------------------*/
static int print_parameter
(
    FILE *file,                            /* I */
    const char *parm_name_and_value,       /* I */
    const char *group_name,                /* I */
    int *group_already_outputed            /* I/O */
)
{
    /*  Print the group name, if needed */
    if (*group_already_outputed != 1)
    {
        if (fprintf(file, "  GROUP = %s\n",group_name) < 0)
            return ERROR;
        *group_already_outputed = 1;
    }

    /*  Print the parm name and value */
    if (fprintf(file,"    %s\n", parm_name_and_value) < 0)
    {
        return ERROR;
    }

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:    print_radiometric_group

PURPOSE: To output the MIN_MAX_RADIANCE, MIN_MAX_REFLECTANCE, and
         MIN_MAX_PIXEL_VALUE groups

RETURNS: SUCCESS if values printed to the file successfully
         ERROR   on general error
-----------------------------------------------------------------------------*/
static int print_radiometric_group
(
    FILE *file,                             /* I */
    const L1G_BAND_METADATA *l1g_band_meta, /* I */
    const char *group_name                  /* I */
)
{
    /* variables used by print_parameter helper function */
    int group_outputed = 0;
    const int parm_nv_size = 250;
    char parm_name_and_value[parm_nv_size];
    parm_name_and_value[0] = '\0';

    int band_index;

    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        int qcal_min;     /* quantize_cal_min_band_x value */
        int qcal_max;     /* quantize_cal_max_band_x value  */
        int band_number;
        int print_error = 0;
        int status;
        const IAS_L1G_BAND_METADATA *band_meta 
            = &l1g_band_meta[band_index].band_meta;

        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Converting band index %d to band number",
                    band_index);
            return ERROR;
        }

        /* Get the QCALMIN and QCALMAX values */
        if (ias_sat_attr_get_quantization_cal_min(band_number, &qcal_min)
            != SUCCESS)
        {
            IAS_LOG_ERROR("Getting qcal min for band %d", band_number);
            return ERROR;
        }
        if (ias_sat_attr_get_quantization_cal_max(band_number, &qcal_max)
            != SUCCESS)
        {
            IAS_LOG_ERROR("Getting qcal max for band %d", band_number);
            return ERROR;
        }

        if (strcmp(group_name, "MIN_MAX_RADIANCE") == 0)
        {
            if (l1g_band_meta[band_index].band_valid)
            {
                float radiance_value;
                radiance_value
                    = band_meta->spectral_radiance_scaling_gain
                    * qcal_max
                    + band_meta->spectral_radiance_scaling_offset; 
                status = snprintf(parm_name_and_value, parm_nv_size,
                        "RADIANCE_MAXIMUM_BAND_%d = %.5f",
                        band_number, radiance_value);
                if ((status < 0) || (status >= parm_nv_size))
                {
                    IAS_LOG_ERROR("Error filling parameter buffer");
                    return ERROR;
                }
                if (print_parameter(file, parm_name_and_value,
                        group_name, &group_outputed) == ERROR)
                {
                    print_error++;
                }

                radiance_value
                    = band_meta->spectral_radiance_scaling_gain
                    * qcal_min
                    + band_meta->spectral_radiance_scaling_offset; 
                status = snprintf(parm_name_and_value, parm_nv_size,
                        "RADIANCE_MINIMUM_BAND_%d = %.5f",
                        band_number, radiance_value);
                if ((status < 0) || (status >= parm_nv_size))
                {
                    IAS_LOG_ERROR("Error filling parameter buffer");
                    return ERROR;
                }
                if (print_parameter(file, parm_name_and_value,
                        group_name, &group_outputed) == ERROR)
                {
                    print_error++;
                }
            }
        }
        else if (strcmp(group_name, "MIN_MAX_REFLECTANCE") == 0)
        {
            const IAS_BAND_ATTRIBUTES *band_attribs;

            band_attribs = ias_sat_attr_get_band_attributes(band_number);
            if (band_attribs == NULL)
            {
                IAS_LOG_ERROR("Getting band attributes for band %d",
                        band_number);
                return ERROR;
            }

            if (band_attribs->sensor_id == IAS_OLI
                && l1g_band_meta[band_index].band_valid)
            {
                float reflectance_value;
                reflectance_value = band_meta->reflectance_scaling_gain
                    * qcal_max + band_meta->reflectance_scaling_offset; 
                status = snprintf(parm_name_and_value, parm_nv_size,
                        "REFLECTANCE_MAXIMUM_BAND_%d = %.6f",
                        band_number, reflectance_value);
                if ((status < 0) || (status >= parm_nv_size))
                {
                    IAS_LOG_ERROR("Error filling parameter buffer");
                    return ERROR;
                }
                if (print_parameter(file, parm_name_and_value,
                        group_name, &group_outputed) == ERROR)
                {
                    print_error++;
                }

                reflectance_value = band_meta->reflectance_scaling_gain
                    * qcal_min + band_meta->reflectance_scaling_offset; 
                status = snprintf(parm_name_and_value, parm_nv_size,
                        "REFLECTANCE_MINIMUM_BAND_%d = %.6f",
                        band_number, reflectance_value);
                if ((status < 0) || (status >= parm_nv_size))
                {
                    IAS_LOG_ERROR("Error filling parameter buffer");
                    return ERROR;
                }
                if (print_parameter(file, parm_name_and_value,
                        group_name, &group_outputed) == ERROR)
                {
                    print_error++;
                }
            }
        }
        else if (strcmp(group_name, "MIN_MAX_PIXEL_VALUE") == 0)
        {
            if (l1g_band_meta[band_index].band_valid)
            {
                status = snprintf(parm_name_and_value, parm_nv_size,
                        "QUANTIZE_CAL_MAX_BAND_%d = %d",
                        band_number, qcal_max);
                if ((status < 0) || (status >= parm_nv_size))
                {
                    IAS_LOG_ERROR("Error filling parameter buffer");
                    return ERROR;
                }
                if (print_parameter(file, parm_name_and_value,
                        group_name, &group_outputed) == ERROR)
                {
                    print_error++;
                }

                status = snprintf(parm_name_and_value, parm_nv_size,
                        "QUANTIZE_CAL_MIN_BAND_%d = %d",
                        band_number, qcal_min);
                if ((status < 0) || (status >= parm_nv_size))
                {
                    IAS_LOG_ERROR("Error filling parameter buffer");
                    return ERROR;
                }
                if (print_parameter(file, parm_name_and_value,
                        group_name, &group_outputed) == ERROR)
                {
                    print_error++;
                }
            }
        }

        if (print_error)
        {
            IAS_LOG_ERROR("Writing radiometric group values; fprintf");
            return ERROR;
        }
    }

    /* Print the End tag for the group, if we printed at least one
    parameter (if we didn't, no output goes to the file)  */
    if (group_outputed == 1)
    {
        if (fprintf(file, "  END_GROUP = %s\n", group_name) < 0)
        {
            IAS_LOG_ERROR("Unable to output END_GROUP tag for %s", group_name);
            return ERROR;
        }
    }

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:    print_radiometric_rescaling

PURPOSE: Print radiometric rescaling values to the metadata file

RETURNS: SUCCESS if values printed to the file successfully
         ERROR   on general error
-----------------------------------------------------------------------------*/
static int print_radiometric_rescaling
(
    FILE *file,                                     /* I */
    const L1G_BAND_METADATA *l1g_band_meta          /* I */
)
{
    int param_index;
    
    /* variables used by print_parameter helper function */
    int group_outputed = 0;
    const char group_name[] = "RADIOMETRIC_RESCALING";
    const int parm_nv_size = 250;
    char parm_name_and_value[parm_nv_size];
    parm_name_and_value[0] = '\0';

    /* Loop through the parameters that need to be output */
    for (param_index = 0; param_index < 4; param_index++)
    {
        int band_index;

        for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
        {
            const IAS_BAND_ATTRIBUTES *band_attribs;
            int band_number;
            int print_error = 0;
            int status;
            const IAS_L1G_BAND_METADATA *band_meta 
                = &l1g_band_meta[band_index].band_meta;

            band_number = ias_sat_attr_convert_band_index_to_number(band_index);
            if (band_number == ERROR)
            {
                IAS_LOG_ERROR("Converting band index %d to band number",
                        band_index);
                return ERROR;
            }

            band_attribs = ias_sat_attr_get_band_attributes(band_number);
            if (band_attribs == NULL)
            {
                IAS_LOG_ERROR("Getting band attributes for band %d",
                        band_number);
                return ERROR;
            }

            /* Print the key = value pairs to the file; reflectance values
               only exist for OLI, so skip over TIRS bands for these */
            switch (param_index)
            {
                /* Print each type of value with different formatting */
                case 0:
                    if (l1g_band_meta[band_index].band_valid)
                    {
                        status = snprintf(parm_name_and_value,
                                parm_nv_size, "RADIANCE_MULT_BAND_%d = %.4E",
                                band_number,
                                band_meta->spectral_radiance_scaling_gain);
                        if ((status < 0) || (status >= parm_nv_size))
                        {
                            IAS_LOG_ERROR("Error filling parameter buffer");
                            return ERROR;
                        }
                        if (print_parameter(file, parm_name_and_value,
                                group_name, &group_outputed) == ERROR)
                        {
                            print_error++;
                        }
                    }
                    break;
                case 1:
                    if (l1g_band_meta[band_index].band_valid)
                    {
                        status = snprintf(parm_name_and_value,
                                parm_nv_size, "RADIANCE_ADD_BAND_%d = %.5f",
                                band_number,
                                band_meta->spectral_radiance_scaling_offset);
                        if ((status < 0) || (status >= parm_nv_size))
                        {
                            IAS_LOG_ERROR("Error filling parameter buffer");
                            return ERROR;
                        }
                        if (print_parameter(file, parm_name_and_value,
                                group_name, &group_outputed) == ERROR)
                        {
                            print_error++;
                        }
                    }
                    break;
                case 2:
                    if (band_attribs->sensor_id == IAS_OLI)
                    {
                        if (l1g_band_meta[band_index].band_valid)
                        {
                            status = snprintf(parm_name_and_value,
                                    parm_nv_size,
                                    "REFLECTANCE_MULT_BAND_%d = %.4E",
                                    band_number,
                                    band_meta->reflectance_scaling_gain);
                            if ((status < 0) || (status >= parm_nv_size))
                            {
                                IAS_LOG_ERROR("Error filling parameter buffer");
                                return ERROR;
                            }
                            if (print_parameter(file, parm_name_and_value,
                                    group_name, &group_outputed) == ERROR)
                            {
                                print_error++;
                            }
                        }
                    }
                    break;
                case 3:
                    if (band_attribs->sensor_id == IAS_OLI)
                    {
                        if (l1g_band_meta[band_index].band_valid)
                        {
                            status = snprintf(parm_name_and_value,
                                    parm_nv_size,
                                    "REFLECTANCE_ADD_BAND_%d = %.6f",
                                    band_number,
                                    band_meta->reflectance_scaling_offset);
                            if ((status < 0) || (status >= parm_nv_size))
                            {
                                IAS_LOG_ERROR("Error filling parameter buffer");
                                return ERROR;
                            }
                            if (print_parameter(file, parm_name_and_value,
                                    group_name, &group_outputed) == ERROR)
                            {
                                print_error++;
                            }
                        }
                    }
                    break;
                default:
                    IAS_LOG_ERROR("Internal programming error during writing "
                            "of radiometric rescaling values");
                    return ERROR;
            }

            if (print_error)
            {
                IAS_LOG_ERROR("Writing radiometric rescaling values; fprintf");
                return ERROR;
            }
        }
    }

    /* Print the End tag for the group, if we printed at least one
    parameter (if we didn't, no output goes to the file)  */
    if (group_outputed == 1)
    {
        if (fprintf(file, "  END_GROUP = %s\n", group_name) < 0)
        {
            IAS_LOG_ERROR("Unable to output END_GROUP tag for %s", group_name);
            return ERROR;
        }
    }

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:    get_l1r_meta

PURPOSE: Gets the L1R file metadata values from the L1R file for the L1G
         metadata - radiometric correction flags, earth to sun distance

RETURNS: SUCCESS if metadata retrieved successfully
                 or if not retrieved but L1G package not requested
                 (to allow GeoTIFF conversion of lone L1G)

         ERROR   if problem getting metadata from L1R file
-----------------------------------------------------------------------------*/
static int get_l1r_meta
(
    const char *l1r_filename,               /* I */
    int want_l1g_package,                   /* I */
    int *l1r_meta_available,                /* O */
    IAS_L1R_FILE_METADATA *l1r_file_meta    /* O */
)
{
    L1RIO *l1r_ptr;
    int status;

    if (!ias_l1r_is_l1r_file(l1r_filename))
    {
        if (!want_l1g_package)
        {
            /* Since L1G not requested, allow this to continue */
            IAS_LOG_WARNING("File not an L1R file: %s; unable to determine "
                    "radiometric correction flags", l1r_filename);
            return SUCCESS;
        }
        IAS_LOG_ERROR("File not an L1R file: %s", l1r_filename);
        return ERROR;
    }

    l1r_ptr = ias_l1r_open_image(l1r_filename, IAS_READ);
    if (l1r_ptr == NULL)
    {
        if (!want_l1g_package)
        {
            /* Since the L1G package was not requested, allow L1R to not be
               available and the conversion to GeoTIFF still succeed */
            IAS_LOG_WARNING("Opening L1R %s; unable to determine radiometric "
                    "correction flags", l1r_filename);
            return SUCCESS;
        }
        IAS_LOG_ERROR("Opening L1R %s", l1r_filename);
        return ERROR;
    }

    status = ias_l1r_get_file_metadata(l1r_ptr, l1r_file_meta);

    /* done with the L1R file so close it before checking status */
    ias_l1r_close_image(l1r_ptr);

    if (status != SUCCESS)
    {
        if (!want_l1g_package)
        {
            /* Since the L1G package was not requested, allow L1R to not be
               available and the conversion to GeoTIFF still succeed */
            IAS_LOG_WARNING("Reading L1R file metadata for %s; unable to "
                    "determine radiometric correction flags", l1r_filename);
            return SUCCESS;
        }
        IAS_LOG_ERROR("Reading L1R file metadata for %s", l1r_filename);
        return ERROR;
    }

    *l1r_meta_available = 1;
    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:    read_grid_info

PURPOSE: Retrieves metadata values from the grid file

RETURNS: SUCCESS if metadata retrieved successfully
                 or if not retrieved but L1G package not requested
                 (to allow GeoTIFF conversion of lone L1G)

         ERROR   if problem getting metadata from grid file
-----------------------------------------------------------------------------*/
static int read_grid_info
(
    int want_l1g_package,           /* I */
    DFP_METADATA *dfp_meta          /* IO */
)
{
    IAS_GRID_TYPE grid;              /* Grid structure */
    int band_index;
    int band_number;

    /* Initialize so we have a known value to compare against in case of
       errors */
    dfp_meta->grid_cell_size_pan = -1;
    dfp_meta->grid_cell_size_reflective = -1;
    dfp_meta->grid_cell_size_thermal = -1;

    /* Initialize to empty string so we have a valid value in case of errors */
    strcpy(dfp_meta->orientation, "");

    /* Check if the grid filename was provided in the OMF */
    if (strcmp(dfp_meta->grid_filename, "") == 0)
    {
        if (!want_l1g_package)
        {
            IAS_LOG_WARNING("Grid not provided; unable to determine image "
                    "orientation and grid cell sizes");
            return SUCCESS;
        }
        IAS_LOG_ERROR("Grid not provided; unable to determine image "
                "orientation and grid cell sizes");
        return ERROR;
    }

    /* Initialize the grid structure */
    if (ias_grid_initialize(&grid, 0, 0) != SUCCESS)
    {
        if (!want_l1g_package)
        {
            IAS_LOG_WARNING("Error initializing the grid structure; unable to "
                    "determine image orientation and grid cell sizes using "
                    "grid file %s", dfp_meta->grid_filename);
            return SUCCESS;
        }
        IAS_LOG_ERROR("Error initializing the grid structure");
        return ERROR;
    }

    if (ias_grid_header_read(dfp_meta->grid_filename, &grid) != SUCCESS)
    {
        if (!want_l1g_package)
        {
            IAS_LOG_WARNING("Error reading the grid header from %s; unable to "
                    "determine image orientation and grid cell sizes",
                    dfp_meta->grid_filename);
            return SUCCESS;
        }
        IAS_LOG_ERROR("Error reading the grid header from %s",
                dfp_meta->grid_filename);
        return ERROR;
    }

    if (ias_grid_band_header_read(dfp_meta->grid_filename, &grid) != SUCCESS)
    {
        if (!want_l1g_package)
        {
            IAS_LOG_WARNING("Error reading the grid band header from %s; "
                    "unable to determine image orientation and grid cell sizes",
                    dfp_meta->grid_filename);
            ias_grid_free(&grid);
            return SUCCESS;
        }
        IAS_LOG_ERROR("Error reading the grid band header from %s",
                dfp_meta->grid_filename);
        ias_grid_free(&grid);
        return ERROR;
    }

    /* Get grid cell sizes for each band type */
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        int set_reflective = 0;
        int set_thermal = 0;
        const IAS_BAND_ATTRIBUTES *band_attribs = NULL;

        if (grid.bands_available[band_index])
        {
            band_number = ias_sat_attr_convert_band_index_to_number(band_index);
            if (band_number == ERROR)
            {
                if (!want_l1g_package)
                {
                    IAS_LOG_WARNING("Converting band index %d to band number; "
                            "unable to determine image orientation and grid "
                            "cell sizes using grid file %s", band_index,
                            dfp_meta->grid_filename);
                    ias_grid_free(&grid);
                    return SUCCESS;
                }
                IAS_LOG_ERROR("Converting band index %d to band number",
                        band_index);
                ias_grid_free(&grid);
                return ERROR;
            }
            band_attribs = ias_sat_attr_get_band_attributes(band_number);
            if (band_attribs == NULL)
            {
                if (!want_l1g_package)
                {
                    IAS_LOG_WARNING("Getting band attributes for band %d; "
                            "unable to determine image orientation and grid "
                            "cell sizes using grid file %s", band_number,
                            dfp_meta->grid_filename);
                    ias_grid_free(&grid);
                    return SUCCESS;
                }
                IAS_LOG_ERROR("Getting band attributes for band %d",
                        band_number);
                ias_grid_free(&grid);
                return ERROR;
            }

            /* Set the pixel sizes, using the spectral band attribs type */
            if (band_attribs->spectral_type == IAS_SPECTRAL_PAN)
            {
                dfp_meta->grid_cell_size_pan
                    = grid.gridbands[band_index][0].pixsize;
            }
            if (!set_reflective
                    && ((band_attribs->spectral_type == IAS_SPECTRAL_VNIR)
                        || (band_attribs->spectral_type == IAS_SPECTRAL_SWIR)))
            {
                dfp_meta->grid_cell_size_reflective
                    = grid.gridbands[band_index][0].pixsize;
                set_reflective = 1;
            }
            if (!set_thermal
                    && (band_attribs->spectral_type == IAS_SPECTRAL_THERMAL))
            {
                dfp_meta->grid_cell_size_thermal
                    = grid.gridbands[band_index][0].pixsize;
                set_thermal = 1;
            }
        }
    }

    /* Set orientation from frame_type */
    switch(grid.frame_type)
    {
        case IAS_GEOBOX:
        case IAS_PROJBOX:
        case IAS_UL_SIZE:
        case IAS_MINBOX:
        case IAS_MAXBOX:
        case IAS_LUNAR_MINBOX:
        case IAS_LUNAR_MAXBOX:
        case IAS_STELLAR_FRAME:
            strcpy(dfp_meta->orientation, "NORTH_UP");
            break;
        case IAS_PATH_ORIENTED:
        case IAS_PATH_MINBOX:
        case IAS_PATH_MAXBOX:
            strcpy(dfp_meta->orientation, "NOMINAL");
            break;
        default:
            /* Catches IAS_UNKNOWN_FRAME_TYPE and anything else unexpected */
            if (!want_l1g_package)
            {
                IAS_LOG_WARNING("Unknown grid frame type %d; unable to "
                        "determine image orientation using grid file %s",
                        grid.frame_type, dfp_meta->grid_filename);
                ias_grid_free(&grid);
                return SUCCESS;
            }
            IAS_LOG_ERROR("Unknown grid frame type %d in grid %s",
                    grid.frame_type, dfp_meta->grid_filename);
            ias_grid_free(&grid);
            return ERROR;
    }

    /* Free grid */
    if (ias_grid_free(&grid) != SUCCESS)
        IAS_LOG_WARNING("Error freeing the grid");

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:    write_metadata

PURPOSE: Creates the L1G metadata file, writing the metadata from various
         sources to the L1G metadata file.

RETURNS: L1G metadata filename;
         NULL if error
-----------------------------------------------------------------------------*/
int write_metadata
(
    const IAS_L0R_SCENE_METADATA *l0r_scene_meta,   /* I: L0R scene meta */
    const IAS_L0R_INTERVAL_METADATA *l0r_int_meta,  /* I: L0R interval meta */
    int l0r_meta_available,                         /* I: Flag */
    const IAS_L1G_FILE_METADATA *l1g_meta,          /* I: L1G file meta */
    const L1G_BAND_METADATA *l1g_band_meta,         /* I: L1G band meta */
    DFP_METADATA *dfp_meta,                         /* I: Misc sources meta */
    const struct tm *curr_time,                     /* I: Current Time */
    int *control_band_index,                        /* O: Control band index */
    char *l1g_meta_filename                        /* O: Metadata filename */
)
{
    FILE *file;                 /* L1G metadata file handle */
    char date[21];              /* used to determine current date */
    char correction_type_converted[5] = "";
    char collection_category[3] = ""; /* Data Tier (RT, T1, T2, A1, or A2) */
    char spacecraft_id_uc[IAS_SPACECRAFT_NAME_LENGTH + 1] = "";
    char filename[MAX_FILE_LENGTH]; /* filename extracted from path */
    int status;
    int i;
    IAS_L1R_FILE_METADATA l1r_file_meta;
    int l1r_meta_available = 0;
    int band_index;
    int print_error = 0;
    int want_l1g_package = get_l1g_package_flag();
    char temp_odl_date[11] = "";
    char software_version[30] = "";

    IAS_LOG_INFO("Creating the L1G metadata file");

    if (strftime(date, sizeof(date), "%Y-%m-%dT%H:%M:%SZ",
        curr_time) == 0)
    {
        IAS_LOG_ERROR("Getting current date");
        return ERROR;
    }

    /* Create the metadata file name */
    status = snprintf(l1g_meta_filename, MAX_FILE_LENGTH, "%s_MTL.txt",
                      determine_l1_basename(dfp_meta));
    if (status < 0 || status >= MAX_FILE_LENGTH)
    {
        IAS_LOG_ERROR("Creating L1G metadata filename");
        return ERROR;
    }

    file = fopen(l1g_meta_filename, "w");
    if (file == NULL)
    {
        IAS_LOG_ERROR("Opening L1G metadata file for writing: %s",
                l1g_meta_filename);
        return ERROR;
    }

    /* L1_METADATA_FILE group */
    /* METADATA_FILE_INFO group */
    if (fprintf(file, "GROUP = L1_METADATA_FILE\n"
                "  GROUP = METADATA_FILE_INFO\n") < 0)
    {
        print_error++;
    }

    if (fprintf(file, "    ORIGIN = \"%s\"\n",
            "Image courtesy of the U.S. Geological Survey") < 0)
    {
        print_error++;
    }

    if (fprintf(file, "    REQUEST_ID = \"%s\"\n",
            dfp_meta->product_request_id) < 0)
    {
        print_error++;
    }

    if (!l0r_meta_available)
    {
        if (fprintf(file, "    LANDSAT_SCENE_ID = \"\"\n") < 0)
        {
            print_error++;
        }
    }
    else
    {
        if (fprintf(file, "    LANDSAT_SCENE_ID = \"%s\"\n",
                    l0r_scene_meta->landsat_scene_id) < 0)
        {
            print_error++;
        }
    }

    /* Add the product ID only if it was requested we name the output 
       using it. */
    if (get_named_with_prod_id())
    {
        if (fprintf(file, "    LANDSAT_PRODUCT_ID = \"%s\"\n",
                    dfp_meta->product_id) < 0)
        {
            print_error++;
        }
    }

    if (fprintf(file, "    COLLECTION_NUMBER = %02d\n",
                dfp_meta->collection_number) < 0)
    {
        print_error++;
    }

    if (!l0r_meta_available)
    {
        if (fprintf(file, "    FILE_DATE = %s\n"
                    "    STATION_ID = \"\"\n", date) < 0)
        {
            print_error++;
        }
    }
    else
    {
        if (fprintf(file, "    FILE_DATE = %s\n"
                    "    STATION_ID = \"%s\"\n",
                    date, l0r_int_meta->station_id) < 0)
        {
            print_error++;
        }
    }

    if (get_processing_software_version(software_version,
            sizeof(software_version)) != SUCCESS)
    {
        IAS_LOG_ERROR("Creating software version");
        fclose(file);
        return ERROR;
    }
    if (fprintf(file, "    PROCESSING_SOFTWARE_VERSION = \"%s\"\n"
                "  END_GROUP = METADATA_FILE_INFO\n", software_version) < 0)
    {
        print_error++;
    }
    /* End METADATA_FILE_INFO group */

    /* PRODUCT_METADATA group */

    /* -------------- PREPARE DATA --------------- */
    status = translate_data_type(l1g_meta->correction_type,
                                 correction_type_converted);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Failed to translate data type %s",
            l1g_meta->correction_type);
        fclose(file);
        return ERROR;
    }

    status = get_elevation_source(l1g_meta->correction_type, want_l1g_package,
            dfp_meta);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Preparing elevation source from DEM");
        fclose(file);
        return ERROR;
    }

    /* Ensure spacecraft ID is upper-case for consistency output to file */
    strcpy(spacecraft_id_uc, dfp_meta->spacecraft_id);
    ias_misc_convert_to_uppercase(spacecraft_id_uc);

    status = get_scene_center_time(l0r_scene_meta->date_acquired,
            l0r_meta_available, dfp_meta->scene_center_time);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting the scene center time");
        fclose(file);
        return ERROR;
    }

    status = get_l1g_meta(l1g_meta, l1g_band_meta, dfp_meta, &band_index);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting L1G metadata");
        fclose(file);
        return ERROR;
    }

    /* Save the control band index for write_product_metadata */
    *control_band_index = band_index;

    status = convert_date_to_odl_style(
        l1g_meta->capture_date,temp_odl_date);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to change date style");
        fclose(file);
        return ERROR;
    }

    /* Determine which sensors are present */
    if (dfp_meta->l1_oli_present[0] == 'Y'
        && dfp_meta->l1_tirs_present[0] == 'Y')
    {
        snprintf(dfp_meta->sensor_id, sizeof(dfp_meta->sensor_id), "OLI_TIRS");
    }
    else if (dfp_meta->l1_oli_present[0] == 'Y')
    {
        snprintf(dfp_meta->sensor_id, sizeof(dfp_meta->sensor_id), "OLI");
    }
    else if (dfp_meta->l1_tirs_present[0] == 'Y')
    {
        snprintf(dfp_meta->sensor_id, sizeof(dfp_meta->sensor_id), "TIRS");
    }

    /* Determine the collection category */
    determine_tier(dfp_meta, l1g_meta, collection_category);

    /* -------------- PRINT DATA --------------- */
    if (fprintf(file, "  GROUP = PRODUCT_METADATA\n"
                "    DATA_TYPE = \"%s\"\n"
                "    COLLECTION_CATEGORY = \"%s\"\n"
                "    ELEVATION_SOURCE = \"%s\"\n"
                "    OUTPUT_FORMAT = \"GEOTIFF\"\n"
                "    SPACECRAFT_ID = \"%s\"\n"
                "    SENSOR_ID = \"%s\"\n"
                "    WRS_PATH = %d\n"
                "    WRS_ROW = %d\n"
                "    NADIR_OFFNADIR = \"%s\"\n"
                "    TARGET_WRS_PATH = %d\n"
                "    TARGET_WRS_ROW = %d\n"
                "    DATE_ACQUIRED = %s\n"
                "    SCENE_CENTER_TIME = \"%s\"\n"
                "    CORNER_UL_LAT_PRODUCT = %.5f\n"
                "    CORNER_UL_LON_PRODUCT = %.5f\n"
                "    CORNER_UR_LAT_PRODUCT = %.5f\n"
                "    CORNER_UR_LON_PRODUCT = %.5f\n"
                "    CORNER_LL_LAT_PRODUCT = %.5f\n"
                "    CORNER_LL_LON_PRODUCT = %.5f\n"
                "    CORNER_LR_LAT_PRODUCT = %.5f\n"
                "    CORNER_LR_LON_PRODUCT = %.5f\n"
                "    CORNER_UL_PROJECTION_X_PRODUCT = %.3f\n"
                "    CORNER_UL_PROJECTION_Y_PRODUCT = %.3f\n"
                "    CORNER_UR_PROJECTION_X_PRODUCT = %.3f\n"
                "    CORNER_UR_PROJECTION_Y_PRODUCT = %.3f\n"
                "    CORNER_LL_PROJECTION_X_PRODUCT = %.3f\n"
                "    CORNER_LL_PROJECTION_Y_PRODUCT = %.3f\n"
                "    CORNER_LR_PROJECTION_X_PRODUCT = %.3f\n"
                "    CORNER_LR_PROJECTION_Y_PRODUCT = %.3f\n",
                    correction_type_converted,
                    collection_category,
                    dfp_meta->elevation_source,
                    spacecraft_id_uc,
                    dfp_meta->sensor_id,
                    l0r_scene_meta->wrs_path,
                    l0r_scene_meta->wrs_row,
                    l0r_scene_meta->nadir_offnadir,
                    dfp_meta->target_wrs_path,
                    dfp_meta->target_wrs_row,
                    temp_odl_date,
                    dfp_meta->scene_center_time,
                    dfp_meta->ul_lat_lon[0],
                    dfp_meta->ul_lat_lon[1],
                    dfp_meta->ur_lat_lon[0],
                    dfp_meta->ur_lat_lon[1],
                    dfp_meta->ll_lat_lon[0],
                    dfp_meta->ll_lat_lon[1],
                    dfp_meta->lr_lat_lon[0],
                    dfp_meta->lr_lat_lon[1],
                    l1g_band_meta[band_index].band_meta.upper_left_x,
                    l1g_band_meta[band_index].band_meta.upper_left_y,
                    l1g_band_meta[band_index].band_meta.upper_right_x,
                    l1g_band_meta[band_index].band_meta.upper_right_y,
                    l1g_band_meta[band_index].band_meta.lower_left_x,
                    l1g_band_meta[band_index].band_meta.lower_left_y,
                    l1g_band_meta[band_index].band_meta.lower_right_x,
                    l1g_band_meta[band_index].band_meta.lower_right_y) < 0)
    {
        print_error++;
    }
    if (strlen(l1g_meta->sun_angle_correction) > 0 &&
        strcmp("NONE", l1g_meta->sun_angle_correction) != 0)
    {
        if (fprintf(file, "    SUN_ANGLE_CORRECTION = \"%s\"\n",
            l1g_meta->sun_angle_correction) < 0)
            print_error++;
    }
    if (dfp_meta->pan_lines != -1)
    {
        if (fprintf(file, "    PANCHROMATIC_LINES = %d\n",
            dfp_meta->pan_lines) < 0)
            print_error++;
    }
    if (dfp_meta->pan_samples != -1)
    {
        if (fprintf(file, "    PANCHROMATIC_SAMPLES = %d\n",
            dfp_meta->pan_samples) < 0)
            print_error++;
    }
    if (dfp_meta->reflective_lines != -1)
    {
        if (fprintf(file, "    REFLECTIVE_LINES = %d\n",
                    dfp_meta->reflective_lines) < 0)
            print_error++;
    }
    if (dfp_meta->reflective_samples != -1)
    {
        if (fprintf(file, "    REFLECTIVE_SAMPLES = %d\n",
                    dfp_meta->reflective_samples) < 0)
            print_error++;
    }
    if (dfp_meta->thermal_lines != -1)
    {
        if (fprintf(file, "    THERMAL_LINES = %d\n", dfp_meta->thermal_lines)
                < 0)
            print_error++;
    }
    if (dfp_meta->thermal_samples != -1)
    {
        if (fprintf(file, "    THERMAL_SAMPLES = %d\n",
                    dfp_meta->thermal_samples) < 0)
            print_error++;
    }
    for (i = 0; i < IAS_MAX_NBANDS; i++)
    {
        int band_number;

        /* If the file name is empty, just move on the next file 
           (we don't want empty filenames in the MTL */
        if (strlen(dfp_meta->geotiff_filenames[i] ) < 1)
        {
            continue;
        }

        band_number = ias_sat_attr_convert_band_index_to_number(i);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Converting band index %d to band number", i);
            fclose(file);
            return ERROR;
        }

        if (fprintf(file, "    FILE_NAME_BAND_%d = \"%s\"\n", band_number,
                dfp_meta->geotiff_filenames[i]) < 0)
        {
            print_error++;
        }
    }

    /* The quality band geotiff name is stored at the end of the filenames
       array, at index IAS_MAX_NBANDS */
    if (strlen(dfp_meta->geotiff_filenames[IAS_MAX_NBANDS]) > 0)
    {
        if (fprintf(file, "    FILE_NAME_BAND_QUALITY = \"%s\"\n",
                    dfp_meta->geotiff_filenames[IAS_MAX_NBANDS]) < 0)
        {
            print_error++;
        }
    }
    
    /* Angle coefficient filename */
    if (get_include_angle_coeff_file_flag()
        && (strlen(dfp_meta->ang_filename) > 0))
    {
        if (fprintf(file, "    ANGLE_COEFFICIENT_FILE_NAME = \"%s\"\n",
                    dfp_meta->ang_filename) < 0)
        {
            print_error++;
        }  
    }

    if (fprintf(file, "    METADATA_FILE_NAME = \"%s\"\n",
                l1g_meta_filename) < 0)
    {
        print_error++;
    }

    extract_filename_from_path(dfp_meta->cpf_filename, filename);
    if (fprintf(file, "    CPF_NAME = \"%s\"\n", filename) < 0)
        print_error++;
    extract_filename_from_path(dfp_meta->bpf_filename_oli, filename);
    if (strlen(filename) > 0)
    {
        if (fprintf(file, "    BPF_NAME_OLI = \"%s\"\n", filename) < 0)
            print_error++;
    }
    extract_filename_from_path(dfp_meta->bpf_filename_tirs, filename);
    if (strlen(filename) > 0)
    {
        if (fprintf(file, "    BPF_NAME_TIRS = \"%s\"\n", filename) < 0)
            print_error++;
    }
    extract_filename_from_path(dfp_meta->rlut_filename, filename);
    if (fprintf(file, "    RLUT_FILE_NAME = \"%s\"\n"
                "  END_GROUP = PRODUCT_METADATA\n",
                filename) < 0)
    {
        print_error++;
    }
    /* End PRODUCT_METADATA group */

    /* Get L1R metadata here to grab earth sun distance */
    status = get_l1r_meta(dfp_meta->l1r_filename, want_l1g_package,
            &l1r_meta_available, &l1r_file_meta);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting L1R metadata");
        fclose(file);
        return ERROR;
    }

    /* IMAGE_ATTRIBUTES group */
    if (fprintf(file, "  GROUP = IMAGE_ATTRIBUTES\n") < 0)
        print_error++;
    if (dfp_meta->cloud_cover_score < 0.0)
    {
        if (fprintf(file, "    CLOUD_COVER = -1\n") < 0)
            print_error++;
    }
    else
    {
        if (fprintf(file, "    CLOUD_COVER = %.2f\n",
                    dfp_meta->cloud_cover_score) < 0)
        {
            print_error++;
        }
    }

    if (dfp_meta->cloud_cover_land_score < 0.0)
    {
        if (fprintf(file, "    CLOUD_COVER_LAND = -1\n") < 0)
            print_error++;
    }
    else
    {
        if (fprintf(file, "    CLOUD_COVER_LAND = %.2f\n",
                    dfp_meta->cloud_cover_land_score) < 0)
        {
            print_error++;
        }
    }

    if (l0r_meta_available)
    {
        if (dfp_meta->l1_oli_present[0] == 'Y')
        {
            if (fprintf(file, "    IMAGE_QUALITY_OLI = %d\n",
                        get_oli_image_quality(dfp_meta, l0r_scene_meta)) < 0)
            {
                print_error++;
            }
        }

        if (dfp_meta->l1_tirs_present[0] == 'Y')
        {
            /* Output the image quality */
            if (fprintf(file, "    IMAGE_QUALITY_TIRS = %d\n",
                        get_tirs_image_quality(dfp_meta, l0r_scene_meta)) < 0)
            {
                print_error++;
            }

            /* Output the SSM model quality */
            if (fprintf(file, "    TIRS_SSM_MODEL = \"%s\"\n",
                get_ssm_quality_flag_name(dfp_meta->ssm_position_quality))
                < 0)
            {
                print_error++;
            }
            /* Output the SSM position status if defined */
            if (strlen(dfp_meta->ssm_position_status) > 0)
            {
                if (fprintf(file, "    TIRS_SSM_POSITION_STATUS = \"%s\"\n", 
                        dfp_meta->ssm_position_status) < 0)
                {
                    print_error++;   
                }
            }

            if (strlen(dfp_meta->stray_light_correction_source) > 0)
            {
                if (fprintf(file, "    TIRS_STRAY_LIGHT_CORRECTION_SOURCE = "
                            "\"%s\"\n",
                            dfp_meta->stray_light_correction_source) < 0)
                {
                    print_error++;
                }
            }
        }

        if (fprintf(file, "    ROLL_ANGLE = %.3f\n"
                    "    SUN_AZIMUTH = %.8f\n"
                    "    SUN_ELEVATION = %.8f\n",
                    l0r_int_meta->roll_angle,
                    l0r_scene_meta->sun_azimuth,
                    l0r_scene_meta->sun_elevation) < 0)
        {
            print_error++;
        }
    }

    if (l1r_meta_available)
    {
        if (fprintf(file, "    EARTH_SUN_DISTANCE = %.7f\n",
                    l1r_file_meta.earth_sun_distance) < 0)
        {
            print_error++;
        }
    }

    /* If OLI is present and saturated bands are supposed to be reported,
       include the info in the MTL file (no need to check TIRS since TIRS
       is not checked for saturation) */
    if ((dfp_meta->l1_oli_present[0] == 'Y')
        && dfp_meta->have_saturated_bands_flag)
    {
        /* Print out Y or N if the band has saturation */
        for (i = 0; i < IAS_MAX_NBANDS; i++)
        {
            int band_number;
            int should_report_saturation_flag;

            /* Skip this band if not being processed */
            if (!dfp_meta->process_band[i])
            {
                continue;
            }

            band_number = ias_sat_attr_convert_band_index_to_number(i);
            if (band_number == ERROR)
            {
                IAS_LOG_ERROR("Converting band index %d to band number", i);
                fclose(file);
                return ERROR;
            }

            /* Check which bands should be reported */
            if (ias_sat_attr_report_saturation_for_band(band_number,
                &should_report_saturation_flag) != SUCCESS)
            {
                IAS_LOG_ERROR("Getting band saturation report flag for band "
                    "number %d", band_number);
                fclose(file);
                return ERROR;
            }

            /* Continue if saturation should not be reported */
            if (should_report_saturation_flag)
            {
                if (fprintf(file,
                        "    SATURATION_BAND_%d = \"%s\"\n", band_number,
                        dfp_meta->band_saturated[i] == 0 ? "N" : "Y") < 0)
                {
                    print_error++;
                }
            }
        }
    }

    /* Only output the GEOMETRIC_RMSE_MODEL values if they were able to 
       be generated (precision model created) */
    if (dfp_meta->ground_control_points_model > 0)
    {
        /* The ground control points version is only useful if a precision
           model was successfully made */
        if (dfp_meta->ground_control_points_version != -1)
        {
            if (fprintf(file, "    GROUND_CONTROL_POINTS_VERSION = %d\n",
                        dfp_meta->ground_control_points_version) < 0)
                print_error++;
        }

        /* Output the number of points used to generate the precision model.  It
           will be zero if a precision model could not be created.  Note the
           terrible name for this parameter comes from the LMDD. */
        if (fprintf(file, "    GROUND_CONTROL_POINTS_MODEL = %d\n",
                    dfp_meta->ground_control_points_model) < 0)
        {
            print_error++;
        }

        if (dfp_meta->geometric_rmse_model >= 0)
        {
            if (fprintf(file, "    GEOMETRIC_RMSE_MODEL = %.3lf\n",
                        dfp_meta->geometric_rmse_model) < 0)
                print_error++;
        }
        if (dfp_meta->model_y_rmse >= 0)
        {
            if (fprintf(file, "    GEOMETRIC_RMSE_MODEL_Y = %.3lf\n",
                        dfp_meta->model_y_rmse) < 0)
                print_error++;
        }
        if (dfp_meta->model_x_rmse >= 0)
        {
            if (fprintf(file, "    GEOMETRIC_RMSE_MODEL_X = %.3lf\n",
                        dfp_meta->model_x_rmse) < 0)
                print_error++;
        }
    }

    /* Only output the GEOMETRIC_RMSE_VERIFY if it was successfully
       calculated */
    if (dfp_meta->ground_control_points_verify > 0)
    {
        /* Output the number of points used by geometric characterization to 
           generate the verification point stats */
        if (fprintf(file, "    GROUND_CONTROL_POINTS_VERIFY = %d\n",
                    dfp_meta->ground_control_points_verify) < 0)
        {
            print_error++;
        }

        if (dfp_meta->geometric_rmse_verify >= 0)
        {
            if (fprintf(file, "    GEOMETRIC_RMSE_VERIFY = %.3lf\n",
                        dfp_meta->geometric_rmse_verify) < 0)
                print_error++;
        }
    }

    if (dfp_meta->l1_oli_present[0] == 'Y')
    {
        const char *truncation_mode =
            dfp_meta->is_oli_lower_12_bits_flag ? "LOWER" : "UPPER";
        if (fprintf(file, "    TRUNCATION_OLI = \"%s\"\n",
                    truncation_mode) < 0)
        {
            print_error++;
        }
    }

    if (fprintf(file, "  END_GROUP = IMAGE_ATTRIBUTES\n") < 0)
    {
        print_error++;
    }
    /* End IMAGE_ATTRIBUTES group */

    /* MIN_MAX_RADIANCE group */
    status = print_radiometric_group(file, l1g_band_meta, "MIN_MAX_RADIANCE");
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing min/max radiance values to metadata "
                "file %s", l1g_meta_filename);
        fclose(file);
        return ERROR;
    }
    /* End MIN_MAX_RADIANCE group */

    /* MIN_MAX_REFLECTANCE group */
    status =
        print_radiometric_group(file, l1g_band_meta, "MIN_MAX_REFLECTANCE");
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing min/max reflectance values to metadata "
                "file %s", l1g_meta_filename);
        fclose(file);
        return ERROR;
    }
    /* End MIN_MAX_REFLECTANCE group */

    /* MIN_MAX_PIXEL_VALUE group */
    status =
        print_radiometric_group(file, l1g_band_meta, "MIN_MAX_PIXEL_VALUE");
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing min/max pixel values to metadata "
                "file %s", l1g_meta_filename);
        fclose(file);
        return ERROR;
    }
    /* End MIN_MAX_PIXEL_VALUE group */

    /* RADIOMETRIC_RESCALING group */
    status = print_radiometric_rescaling(file, l1g_band_meta);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing radiometric rescaling values to metadata "
                "file %s", l1g_meta_filename);
        fclose(file);
        return ERROR;
    }
    /* End RADIOMETRIC_RESCALING group */

    /* TIRS_THERMAL_CONSTANTS group */
    if (dfp_meta->l1_tirs_present[0] == 'Y')
    {
        if (fprintf(file, "  GROUP = TIRS_THERMAL_CONSTANTS\n") < 0)
            print_error++;
        status = print_tirs_thermal_constants(file, 
                 dfp_meta->tirs_thermal_constants);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Writing TIRS thermal constant values to metadata "
                    "file %s", l1g_meta_filename);
            fclose(file);
            return ERROR;
        }
        if (fprintf(file, "  END_GROUP = TIRS_THERMAL_CONSTANTS\n") < 0)
            print_error++;
    }
    /* End TIRS_THERMAL_CONSTANTS group */

    /* PROJECTION_PARAMETERS group */
    if (fprintf(file, "  GROUP = PROJECTION_PARAMETERS\n"
                "    MAP_PROJECTION = \"%s\"\n"
                "    DATUM = \"%s\"\n"
                "    ELLIPSOID = \"WGS84\"\n",
                dfp_meta->map_projection,
                l1g_meta->datum) < 0)
    {
        print_error++;
    }
    if (l1g_meta->projection_code == UTM)
    {
        /* UTM_PROJECTION_PARAMETERS */
        if (fprintf(file, "    UTM_ZONE = %d\n",
                    l1g_meta->zone_code) < 0)
        {
            print_error++;
        }
        /* End UTM_PROJECTION_PARAMETERS */
    }
    else if (l1g_meta->projection_code == PS)
    {
        double vertical_longitude;
        double true_scale_latitude;

        /* Convert the VERTICAL_LON_FROM_POLE and TRUE_SCALE_LAT values
           from DMS to decimal format */
        status = ias_geo_convert_dms2deg(l1g_meta->projection_parameters[4],
            &vertical_longitude, "LON");
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Converting VERTICAL_LON_FROM_POLE from DMS "
                "to degrees");
            fclose(file);
            return ERROR;
        }
        status = ias_geo_convert_dms2deg(l1g_meta->projection_parameters[5],
            &true_scale_latitude, "LAT");
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Converting TRUE_SCALE_LAT from DMS to degrees");
            fclose(file);
            return ERROR;
        }

        /* PS_PROJECTION_PARAMETERS group.  Note that the first two are lat/lon
           values that are reported to 5 decimal places.  The false easting
           and northing are output as integers. */
        if (fprintf(file, "    VERTICAL_LON_FROM_POLE = %.5f\n"
                    "    TRUE_SCALE_LAT = %.5f\n"
                    "    FALSE_EASTING = %.0f\n"
                    "    FALSE_NORTHING = %.0f\n",
                    vertical_longitude,
                    true_scale_latitude,
                    l1g_meta->projection_parameters[6],
                    l1g_meta->projection_parameters[7]) < 0)
        {
            print_error++;
        }
    }
    else /* ALBERS */
    {
        double latitude_origin;
        double central_meridian_longitude;
        double standard_parallel_1_lat;
        double standard_parallel_2_lat;

        status = ias_geo_convert_dms2deg(l1g_meta->projection_parameters[5],
            &latitude_origin, "LAT");
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Converting ORIGIN_LAT from DMS to degrees");
            fclose(file);
            return ERROR;
        }

        /* Convert the central meridian long and Lat origin values
           from DMS to decimal format */
        status = ias_geo_convert_dms2deg(l1g_meta->projection_parameters[4],
            &central_meridian_longitude, "LON");
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Converting CENTRAL_MERIDIAN_LON from DMS "
                "to degrees");
            fclose(file);
            return ERROR;
        }

        status = ias_geo_convert_dms2deg(l1g_meta->projection_parameters[2],
            &standard_parallel_1_lat, "LAT");
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Converting STD_PARALLEL_1_LAT from DMS to degrees");
            fclose(file);
            return ERROR;
        }

        status = ias_geo_convert_dms2deg(l1g_meta->projection_parameters[3],
            &standard_parallel_2_lat, "LAT");
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Converting STD_PARALLEL_1_LAT from DMS to degrees");
            fclose(file);
            return ERROR;
        }

        /* ALBERS_PROJECTION_PARAMETERS group.  Note that the first four are
           lat/lon values that are reported to 5 decimal places.  The false
           easting and northing are output as integers. */
        if (fprintf(file, "    ORIGIN_LAT = %.5f\n"
                    "    CENTRAL_MERIDIAN_LON = %.5f\n"
                    "    STANDARD_PARALLEL_1_LAT = %.5f\n"
                    "    STANDARD_PARALLEL_2_LAT = %.5f\n"
                    "    FALSE_EASTING = %.0f\n"
                    "    FALSE_NORTHING = %.0f\n",
                    latitude_origin,
                    central_meridian_longitude,
                    standard_parallel_1_lat,
                    standard_parallel_2_lat,
                    l1g_meta->projection_parameters[6],
                    l1g_meta->projection_parameters[7]) < 0)
        {
            print_error++;
        }
    }

    status = read_grid_info(want_l1g_package, dfp_meta);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Preparing metadata from grid");
        fclose(file);
        return ERROR;
    }
    if (dfp_meta->grid_cell_size_pan > 0)
    {
        if (fprintf(file, "    GRID_CELL_SIZE_PANCHROMATIC = %.2f\n",
                    dfp_meta->grid_cell_size_pan) < 0)
        {
            print_error++;
        }
    }
    if (dfp_meta->grid_cell_size_reflective > 0)
    {
        if (fprintf(file, "    GRID_CELL_SIZE_REFLECTIVE = %.2f\n",
                    dfp_meta->grid_cell_size_reflective) < 0)
        {
            print_error++;
        }
    }
    if (dfp_meta->grid_cell_size_thermal > 0)
    {
        if (fprintf(file, "    GRID_CELL_SIZE_THERMAL = %.2f\n",
                    dfp_meta->grid_cell_size_thermal) < 0)
        {
            print_error++;
        }
    }
    if (fprintf(file, "    ORIENTATION = \"%s\"\n"
                "    RESAMPLING_OPTION = \"%s\"\n"
                "  END_GROUP = PROJECTION_PARAMETERS\n",
                dfp_meta->orientation,
                dfp_meta->resample_type) < 0)
    {
        print_error++;
    }
    /* End PROJECTION_PARAMETERS group */

    if (fprintf(file, "END_GROUP = L1_METADATA_FILE\n") < 0)
        print_error++;
    /* End L1_METADATA_FILE group */

    if (fprintf(file, "END\n") < 0)
        print_error++;

    status = fclose(file);
    if (status != 0)
    {
        IAS_LOG_ERROR("Closing L1G metadata file");
        return ERROR;
    }

    if (print_error)
    {
        IAS_LOG_ERROR("Writing to L1G metadata file");
        return ERROR;
    }

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:    write_product_metadata

PURPOSE: Creates the product metadata file, writing the metadata from various
         sources to the database load file.

RETURNS: SUCCESS or
         ERROR 
-----------------------------------------------------------------------------*/
int write_product_metadata
(
    const IAS_L0R_SCENE_METADATA *l0r_scene_meta,   /* I: L0R scene meta */
    const IAS_L1G_FILE_METADATA *l1g_meta,          /* I: L1G file meta */
    const L1G_BAND_METADATA *l1g_band_meta,         /* I: L1G band meta */
    DFP_METADATA *dfp_meta,                         /* I: Misc sources meta */
    int band_index                                  /* I: Index of band to */
                                                    /*    for reported */
                                                    /*    metadata */
)
{
    FILE *file;                 /* Product_data file handle */
    char filename[PATH_MAX];    /* Temp value for filenames */
    char hostname[HOSTNAME_LENGTH];          /* Processing host name */
    char data_type[IAS_L1G_CORRECTION_TYPE_SIZE] = "";
    int print_error = 0;        /* Number of errors writing to file */
    char collection_category[3] = ""; /* Data Tier (RT, T1, T2, A1, or A2) */
    int want_l0r_package;       /* Write out metadata for L0R package */
    int want_l1g_package;       /* Write out metadata for L1 package */
    IAS_DATETIME date_time;     /* IAS format date from L0R */
    int status;
    double vertical_longitude;
    double true_scale_latitude;

    want_l0r_package = get_l0r_package_flag();
    want_l1g_package = get_l1g_package_flag();

    IAS_LOG_INFO("Creating the database load file: %s", DATABASE_LOAD_FILE);

    if (want_l1g_package)
    {
        /* Convert the VERTICAL_LON_FROM_POLE and TRUE_SCALE_LAT values
           from DMS to decimal format */
        status = ias_geo_convert_dms2deg(l1g_meta->projection_parameters[4],
            &vertical_longitude, "LON");
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Converting VERTICAL_LON_FROM_POLE from DMS "
                "to degrees");
            return ERROR;
        }

        status = ias_geo_convert_dms2deg(l1g_meta->projection_parameters[5],
            &true_scale_latitude, "LAT");
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Converting TRUE_SCALE_LAT from DMS to degrees");
            return ERROR;
        }
    }

    /* Get hostname */
    if (gethostname(hostname, HOSTNAME_LENGTH) != 0)
    {
        IAS_LOG_ERROR("Getting host name while creating load file: %s",
                DATABASE_LOAD_FILE);
        return ERROR;
    }

    file = fopen(DATABASE_LOAD_FILE, "w");
    if (file == NULL)
    {
        IAS_LOG_ERROR("Opening database load file: %s for writing",
            DATABASE_LOAD_FILE);
        return ERROR;
    }

    if (fprintf(file,"WORK_ORDER_ID=%s\n", dfp_meta->work_order_id) <0)
    {
        print_error++;
    }

    if (fprintf(file,"LANDSAT_SCENE_ID=%s\n", l0r_scene_meta->landsat_scene_id)
            < 0)
    {
        print_error++;
    }

    /* Add in the PROJECT */
    if (fprintf(file,"PROJECT=%s\n", get_project()) < 0)
    {
        print_error++;
    }

    /* Parse the datetime string provided by the L0R */
    if (ias_misc_parse_datetime_string(l0r_scene_meta->date_acquired,
                IAS_DATETIME_L0R_FORMAT, &date_time) == ERROR)
    {
        IAS_LOG_ERROR("Parsing L0R datetime string: %s",
            l0r_scene_meta->date_acquired);
        print_error++;
    }

    /* Convert the date to a oracle date string when printing */
    if (fprintf(file, "DATE_ACQUIRED=%04d/%02d/%02d\n", date_time.year,
                date_time.month, date_time.day_of_month) < 0) 
    {
        print_error++;
    }

    if (fprintf(file, "WRS_PATH=%d\n", l0r_scene_meta->wrs_path) < 0)
    {
        print_error++;
    }

    if (fprintf(file, "WRS_ROW=%d\n", l0r_scene_meta->wrs_row) < 0)
    {
        print_error++;
    }

    if (fprintf(file, "PROCESSING_HOST=%s\n", hostname) < 0)
    {
        print_error++;
    }

    if (want_l0r_package) 
    {
        if (fprintf(file, "L0R_OLI_DATA_PRESENT=%s\n",
                     dfp_meta->l0r_oli_present) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "L0R_TIRS_DATA_PRESENT=%s\n",
                     dfp_meta->l0r_tirs_present) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "L0RP_PACKAGE_NAME=%s\n", 
                    dfp_meta->l0r_package_file_name) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "L0RP_PACKAGE_FILE_DATE=%s\n", 
                    dfp_meta->l0r_package_file_date) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "L0RP_PACKAGE_FILE_SIZE=%lld\n", 
                    dfp_meta->l0r_package_file_size) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "L0RP_FILE_CHECKSUM=%s\n", 
                    dfp_meta->l0r_package_file_checksum) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "L0RP_PRODUCT_FILE_SIZE=%lld\n", 
                    dfp_meta->l0r_product_file_size) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "L0RP_CHECKSUM_FILE_NAME=%s\n", 
                    dfp_meta->l0r_checksum_file_name) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "L0RP_CHECKSUM_FILE_DATE=%s\n", 
                    dfp_meta->l0r_checksum_file_date) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "L0RP_CHECKSUM_FILE_SIZE=%lld\n", 
                    dfp_meta->l0r_checksum_file_size) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "SUBSETTER_VERSION_L0RP=%s\n", 
                    l0r_scene_meta->subsetter_version_l0rp) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "SCENE_BANDS_L0RP=%d\n", 
                    dfp_meta->l0r_band_count) < 0)
        {
            print_error++;
        }
    }

    if (want_l1g_package) 
    {
        char software_version[30];

        if (fprintf(file,"LANDSAT_PRODUCT_ID=%s\n", dfp_meta->product_id) <0)
        {
            print_error++;
        }

        determine_tier(dfp_meta, l1g_meta, collection_category);
        if (fprintf(file,"COLLECTION_CATEGORY=%s\n", collection_category) < 0)
        {
            print_error++;
        }

        if (fprintf(file,"COLLECTION_NUMBER=%u\n", dfp_meta->collection_number)
                < 0)
        {
            print_error++;
        }

        if (fprintf(file, "L1_OLI_DATA_PRESENT=%s\n", dfp_meta->l1_oli_present)
                     < 0)
        {
            print_error++;
        }

        if (fprintf(file, "L1_TIRS_DATA_PRESENT=%s\n",
                     dfp_meta->l1_tirs_present) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "L1_PACKAGE_NAME=%s\n", 
                    dfp_meta->l1_package_file_name) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "L1_PACKAGE_FILE_SIZE=%lld\n",
                    dfp_meta->l1_package_file_size) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "L1_PACKAGE_FILE_DATE=%s\n",
                    dfp_meta->l1_package_file_date) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "L1_PRODUCT_FILE_SIZE=%lld\n", 
                    dfp_meta->l1_product_file_size) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "L1_PACKAGE_FILE_CHECKSUM=%s\n",
                    dfp_meta->l1_package_file_checksum) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "L1_CHECKSUM_FILE_NAME=%s\n",
                    dfp_meta->l1_checksum_file_name) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "L1_CHECKSUM_FILE_DATE=%s\n",
                    dfp_meta->l1_checksum_file_date) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "L1_CHECKSUM_FILE_SIZE=%lld\n",
                    dfp_meta->l1_checksum_file_size) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "CORNER_UL_LAT_PRODUCT=%.5f\n",
                    dfp_meta->ul_lat_lon[0]) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "CORNER_UL_LON_PRODUCT=%.5f\n",
                    dfp_meta->ul_lat_lon[1]) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "CORNER_UR_LAT_PRODUCT=%.5f\n",
                    dfp_meta->ur_lat_lon[0]) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "CORNER_UR_LON_PRODUCT=%.5f\n",
                    dfp_meta->ur_lat_lon[1]) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "CORNER_LL_LAT_PRODUCT=%.5f\n",
                    dfp_meta->ll_lat_lon[0]) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "CORNER_LL_LON_PRODUCT=%.5f\n",
                    dfp_meta->ll_lat_lon[1]) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "CORNER_LR_LAT_PRODUCT=%.5f\n",
                    dfp_meta->lr_lat_lon[0]) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "CORNER_LR_LON_PRODUCT=%.5f\n",
                    dfp_meta->lr_lat_lon[1]) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "SCENE_COORDINATE_TYPE=%s\n", "PRODUCT") < 0)
        {
            print_error++;
        }

        if (fprintf(file, "SCENE_CENTER_LAT=%.5f\n", 
                    l0r_scene_meta->scene_center.lat) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "SCENE_CENTER_LON=%.5f\n", 
                    l0r_scene_meta->scene_center.lng) < 0)
        {
            print_error++;
        }

        extract_filename_from_path(dfp_meta->bpf_filename_oli, filename);
        if (fprintf(file, "BPF_NAME_OLI=%s\n", filename) < 0)
        {
            print_error++;
        }

        extract_filename_from_path(dfp_meta->bpf_filename_tirs, filename);
        if (fprintf(file, "BPF_NAME_TIRS=%s\n", filename) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "VERTICAL_LON_FROM_POLE=%.5f\n", 
                    vertical_longitude) < 0)
        {
            print_error++;
        }

        if (dfp_meta->cloud_cover_score < 0.0)
        {
            if (fprintf(file, "CLOUD_COVER=-1\n") < 0)
            {
                print_error++;
            }
        }
        else
        {
            if (fprintf(file, "CLOUD_COVER=%.2f\n",
                        dfp_meta->cloud_cover_score) < 0)
            {
                print_error++;
            }
        }
 
        if (dfp_meta->cloud_cover_land_score < 0.0)
        {
            if (fprintf(file, "CLOUD_COVER_LAND=-1\n") < 0)
            {
                print_error++;
            }
        }
        else
        {
            if (fprintf(file, "CLOUD_COVER_LAND=%.2f\n",
                        dfp_meta->cloud_cover_land_score) < 0)
            {
                print_error++;
            }
        }

        if (dfp_meta->snow_ice_score < 0.0)
        {
            if (fprintf(file, "SNOW_ICE_COVER=-1\n") < 0)
            {
                print_error++;
            }
        }
        else
        {
            if (fprintf(file, "SNOW_ICE_COVER=%.2f\n",
                        dfp_meta->snow_ice_score) < 0)
            {
                print_error++;
            }
        }

        if (dfp_meta->l1_oli_present[0] == 'Y')
        {
            if (fprintf(file, "IMAGE_QUALITY_OLI=%d\n",
                        get_oli_image_quality(dfp_meta, l0r_scene_meta)) < 0)
            {
                print_error++;
            }
        }

        if (dfp_meta->l1_tirs_present[0] == 'Y')
        {
            if (fprintf(file, "IMAGE_QUALITY_TIRS=%d\n",
                        get_tirs_image_quality(dfp_meta, l0r_scene_meta)) < 0)
            {
                print_error++;
            }

            if (fprintf(file, "TIRS_SSM_MODEL=%s\n",
                get_ssm_quality_flag_name(dfp_meta->ssm_position_quality)) < 0)
            {
                print_error++;
            }
        }

        extract_filename_from_path(dfp_meta->cpf_filename, filename);
        if (fprintf(file, "CPF_NAME=%s\n", filename) < 0)
        {
            print_error++;
        }

        if (translate_data_type(l1g_meta->correction_type, data_type) !=
            SUCCESS)
        {
            IAS_LOG_ERROR("Failed to translate data type %s",
                l1g_meta->correction_type);
            print_error++;
        }

        if (fprintf(file, "L1_DATA_TYPE=%s\n", data_type) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "ELEVATION_SOURCE=%s\n", 
                    dfp_meta->elevation_source) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "OUTPUT_FORMAT=GEOTIFF\n") < 0)
        {
            print_error++;
        }

        if (fprintf(file, "FALSE_EASTING=%.0f\n", 
                    l1g_meta->projection_parameters[6]) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "FALSE_NORTHING=%.0f\n", 
                    l1g_meta->projection_parameters[7]) < 0)
        {
            print_error++;
        }

        if (dfp_meta->ground_control_points_model > 0)
        {
            if (fprintf(file, "GROUND_CONTROL_POINTS_MODEL=%d\n",
                        dfp_meta->ground_control_points_model) < 0)
            {
                print_error++;
            }
        }
        else
        {
            if (fprintf(file, "GROUND_CONTROL_POINTS_MODEL=\n") < 0)
            {
                print_error++;
            }
        }

        /* The ground control points version is valid if it is zero or
           greater */
        if (dfp_meta->ground_control_points_version >= 0)
        {
            if (fprintf(file, "GROUND_CONTROL_POINTS_VERSION=%d\n",
                        dfp_meta->ground_control_points_version) < 0)
            {
                print_error++;
            }
        }
        else
        {
            if (fprintf(file, "GROUND_CONTROL_POINTS_VERSION=\n") < 0)
            {
                print_error++;
            }
        }

        if (dfp_meta->geometric_rmse_model < 0)
        {
            if (fprintf(file, "GEOMETRIC_RMSE_MODEL=\n") < 0)
                print_error++;
        }
        else
        {
            if (fprintf(file, "GEOMETRIC_RMSE_MODEL=%.3f\n",
                    dfp_meta->geometric_rmse_model) < 0)
            {
                print_error++;
            }                
        }

        if (dfp_meta->model_y_rmse < 0)
        {
            if (fprintf(file, "GEOMETRIC_RMSE_MODEL_Y=\n") < 0)
            {
                print_error++;
            }
        }
        else
        {
            if (fprintf(file, "GEOMETRIC_RMSE_MODEL_Y=%.7f\n", 
                    dfp_meta->model_y_rmse) < 0)
            {
                print_error++;
            }
        }

        if (dfp_meta->model_x_rmse < 0)
        {
            if (fprintf(file, "GEOMETRIC_RMSE_MODEL_X=\n") < 0)
            {
                print_error++;
            }
        }
        else
        {
            if (fprintf(file, "GEOMETRIC_RMSE_MODEL_X=%.7f\n", 
                        dfp_meta->model_x_rmse) < 0)
            {
                print_error++;
            }
        }

        if (dfp_meta->grid_cell_size_pan == -1.0)
        {
            if (fprintf(file, "GRID_CELL_SIZE_PANCHROMATIC=\n") < 0)
            {
                print_error++;
            }
        }
        else
        {
            if (fprintf(file, "GRID_CELL_SIZE_PANCHROMATIC=%.2f\n",
                        dfp_meta->grid_cell_size_pan) < 0)
            {
                print_error++;
            }
        }

        if (dfp_meta->grid_cell_size_reflective == -1.0)
        {
            if (fprintf(file, "GRID_CELL_SIZE_REFLECTIVE=\n") < 0)
            {
                print_error++;
            }
        }
        else
        {
            if (fprintf(file, "GRID_CELL_SIZE_REFLECTIVE=%.2f\n",
                        dfp_meta->grid_cell_size_reflective) < 0)
            {
                print_error++;
            }
        }

        if (dfp_meta->grid_cell_size_thermal == -1.0)
        {
            if (fprintf(file, "GRID_CELL_SIZE_THERMAL=\n") < 0)
            {
                print_error++;
            }
        }
        else
        {
            if (fprintf(file, "GRID_CELL_SIZE_THERMAL=%.2f\n",
                        dfp_meta->grid_cell_size_thermal) < 0)
            {
                print_error++;
            }
        }

        if (get_processing_software_version(software_version,
                sizeof(software_version)) != SUCCESS)
        {
            IAS_LOG_ERROR("Creating software version");
            fclose(file);
            return ERROR;
        }

        if (fprintf(file, "PROCESSING_SOFTWARE_VERSION=%s\n", 
                software_version) < 0)
        {
            print_error++;
        }
        
        if (fprintf(file, "MAP_PROJECTION=%s\n",
                dfp_meta->map_projection) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "ORIENTATION=%s\n", dfp_meta->orientation) < 0)
        {
            print_error++;
        }

        if (dfp_meta->pan_lines == -1)
        {
            if (fprintf(file, "PANCHROMATIC_LINES=0\n") < 0)
            {
                print_error++;
            }
        }
        else
        {
            if (fprintf(file, "PANCHROMATIC_LINES=%d\n",
                dfp_meta->pan_lines) < 0)
            {
                print_error++;
            }
        }
        if (dfp_meta->pan_samples == -1)
        {
            if (fprintf(file, "PANCHROMATIC_SAMPLES=0\n") < 0)
            {
                print_error++;
            }
        }
        else
        {
            if (fprintf(file, "PANCHROMATIC_SAMPLES=%d\n",
                dfp_meta->pan_samples) < 0)
            {
                print_error++;
            }
        }
        if (dfp_meta->reflective_lines == -1)
        {
            if (fprintf(file, "REFLECTIVE_LINES=0\n") < 0)
            {
                print_error++;
            }
        }
        else
        {
            if (fprintf(file, "REFLECTIVE_LINES=%d\n",
                        dfp_meta->reflective_lines) < 0)
            {
                print_error++;
            }
        }

        if (dfp_meta->reflective_samples == -1)
        {
            if (fprintf(file, "REFLECTIVE_SAMPLES=0\n") < 0)
            {
                print_error++;
            }
        }
        else
        {
            if (fprintf(file, "REFLECTIVE_SAMPLES=%d\n",
                        dfp_meta->reflective_samples) < 0)
            {
                print_error++;
            }
        }

        if (dfp_meta->thermal_lines == -1)
        {
            if (fprintf(file, "THERMAL_LINES=0\n") < 0)
            {
                print_error++;
            }
        }
        else
        {
            if (fprintf(file, "THERMAL_LINES=%d\n", 
                        dfp_meta->thermal_lines) < 0)
            {
                print_error++;
            }
        }

        if (dfp_meta->thermal_samples == -1)
        {
            if (fprintf(file, "THERMAL_SAMPLES=0\n") < 0)
            {
                print_error++;
            }
        }
        else
        {
            if (fprintf(file, "THERMAL_SAMPLES=%d\n",
                        dfp_meta->thermal_samples) < 0)
            {
                print_error++;
            }
        }

        if (fprintf(file, "RESAMPLING_OPTION=%s\n", dfp_meta->resample_type)
            < 0)
        {
            print_error++;
        }

        extract_filename_from_path(dfp_meta->rlut_filename, filename);
        if (fprintf(file, "RLUT_FILE_NAME=%s\n", filename) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "SCENE_BANDS_L1=%d\n", get_l1g_nbands()) < 0)
        {
            print_error++;
        }

        /* Scene status will always be 'G' for any scenes we process */
        if (fprintf(file, "SCENE_STATUS=%c\n", 'G') < 0)
        {
            print_error++;
        }

        if (fprintf(file, "SUBSETTER_VERSION_L1=%s\n", 
                    l0r_scene_meta->subsetter_version_l0rp) < 0)
        {
            print_error++;
        }

        if (fprintf(file, "TRUE_SCALE_LAT=%.5f\n", 
                    true_scale_latitude) < 0)
        {
            print_error++;
        }

        if (l1g_meta->zone_code <= 0)
        {
            if (fprintf(file, "UTM_ZONE=\n") < 0)
            {
                print_error++;
            }
        }
        else if (fprintf(file, "UTM_ZONE=%d\n", l1g_meta->zone_code) < 0)
        {
            print_error++;
        }
    }

    if (fclose(file) != 0)
    {
        IAS_LOG_ERROR("Closing database load file: %s", DATABASE_LOAD_FILE);
        return ERROR;
    }

    if (print_error)
    {
        IAS_LOG_ERROR("Writing to database load file: %s", DATABASE_LOAD_FILE);
        return ERROR;
    }

    return SUCCESS;
}
