/******************************************************************************
NAME: GCPresample (main)

PURPOSE:
Main driver for the GCPresample task.  This application resamples the custom
GCPs, located in reference chips, to a 1G image in the UTM projection.
The UTM Zone is based on the zone of the L1G image.  The results of this
task may be used in GCPcorrelate.

RETURN VALUE:
Type = int
Value        Description
-----        -----------
EXIT_FAILURE Error in resampling; GCP files may not have been produced
             and GCP library may not have been updated
EXIT_SUCCESS Successful resampling

******************************************************************************/
#include "ias_logging.h"
#include "GCPresample.h"

int main(int argc, char *argv[])
{
    APP_PARMS parm_data;             /* Parameters read from parameter file
                                        or OMF */
    int status;                      /* Check on return on functions */
    char prog_name[PATH_MAX];        /* Name of the program */
    char parm_filename[PATH_MAX];    /* Name of the parameter file */

    /* Get the name of the program */
    strcpy(prog_name, "gcpresample");

    /* Initialize the IAS logging library */
    status = ias_log_initialize(prog_name);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing log");
        exit(EXIT_FAILURE);
    }

    /* Check the number of arguments
       We want the program name and the ODL file for args */
    if (argc != 2)
    {
        IAS_LOG_ERROR("Usage: %s ODL_Parameter_File", prog_name);
        exit(EXIT_FAILURE);
    }

    /* Store the name of the parameter file */
    strcpy(parm_filename, argv[1]);

    /* Get parameters and setup information */
    memset(&parm_data, 0, sizeof(APP_PARMS));
    status = get_gcpresample_parameters(parm_filename, &parm_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting required parameter values");
        exit(EXIT_FAILURE);
    }

    /* Log initiation */
    IAS_LOG_INFO("Initiated");

    /* Get GCP information from the GCP library using specified criteria */
    status = resample_gcps(parm_data.img_name, parm_data.gcp_dir,
            parm_data.projection_source, parm_data.target_projection, 
            parm_data.utm_zone, parm_data.use_tarball);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting GCP information");
        exit(EXIT_FAILURE);
    }

    IAS_LOG_INFO("Successful completion");

    exit(EXIT_SUCCESS);
}
