#!/usr/bin/perl -w

#############################################################################
#
# NAME:            get_all_gcps.pl
#
# PURPOSE:         Uses the get_gcps_wrapper.pl script to create a directory
#                  of all GCP chips into WRS path/row directory structure.
#
# RETURNS:         0 if processing completes successfully
#                  non-zero value if an error is encountered
#
# NOTES:           A parameter --help is provided to display a
#                  usage message.
#
############################################################################
use strict;
use DBI;
use Getopt::Long;                        # For GetOptions()
use File::Basename;
use File::Temp qw(tempfile);             # Function to make temporary files
use ias_const;                           # Defines IAS status parameters
                                         # SUCCESS and ERROR
use Logger;                              # Log messages
use ias_processing_system;
use ias_db;				 # Database routines

# Declare program constant
use constant PROG => basename($0);

# Forward subroutine declarations
sub ShowUsage();                         # Prints help / usage message then
                                         # quits
sub ProcessOptions();                    # Processes command line options
sub PurgeWRSPath($);                     # Removes a WRS path from tree if
                                         # it's empty.
# Basic parameter file information
my $gcp_parameters = '';                 # most of the GCP wrapper parameters
                                         # are just accumulated from the 
                                         # command line and passed along
my $gcp_home = '';                       # GCP top level directory under
                                         # which the WRS path directories
                                         # are created, and the WRS row
                                         # directories under them


# Basic GCP information
my $gcp_type = '';                      # GCP type (optional for filtering)
my $gcp_source = 'ALL';                 # GCP source (optional here and can be 
                                        # overridden, but must be specified for 
                                        # get_gcps_wrapper.pl)

# WRS path/row ranges (default to all, but can be limited for filtering)
my $wrs_start_path = 1;
my $wrs_start_row = 1;
my $wrs_end_path = 233;
my $wrs_end_row = 248;

# Chip resolution to retrieve
my $resolution;

# Tarball compression type
my $compression = "NONE";

# Write tarball option
my $write_tarball = 1;

# GCP Looping query fetch results 
my $wrs_path;
my $wrs_row;

# variable for detecting when a new WRS path is fetched from query
my $last_path_dir = "";

# option to purge each WRS row directory prior to fetching chips
my $purge_dir = 0; 

# Miscellaneous temporary variables
my $status;
my $error;

# Initialize the ias_processing_system module
if (ias_init_system() != SUCCESS)
{
    print STDERR PROG . "$ias_error_msg\n";
    exit FAILURE;
}

# Set up the logger
my $logger = Logger->new("STDOUT");
if (!$logger)
{
    print STDERR PROG . ": ERROR:  Failed to establish logger object for "
        . "STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name(PROG);
set_default_log($logger);

# If no options have been entered on the command line, quit
if (scalar @ARGV < 1)
{
    my $PROGRAM = PROG;
    $error = "Not enough parameters specified on command line.";
    $logger->fatal_msg(__FILE__, __LINE__, $error);
    $error = "To display the USAGE message, enter '$PROGRAM --help' "
           . "at the command line";
    $logger->info_msg(__FILE__, __LINE__, $error);
    $logger->close_log();
    exit FAILURE;
}

# Process the command line options
$status = ProcessOptions();
if ($status != SUCCESS)
{
    $error = "Command-line options could not be processed";
    $logger->fatal_msg(__FILE__, __LINE__, $error);
    $logger->close_log();
    exit FAILURE;
}

# create the top-level directory, as needed
my $rc = system( "\\mkdir -p $gcp_home" );
if ($rc != SUCCESS)
{
   $error = "Error creating GCP top-level directory $gcp_home";
   $logger->fatal_msg(__FILE__, __LINE__, $error);
   $logger->close_log();
   exit FAILURE;
}

# move to the top-level directory, location of the WRS path subdirectories
if (! chdir( $gcp_home ) )
{
    $error = "Error changing to GCP directory $gcp_home";
    $logger->fatal_msg(__FILE__, __LINE__, $error);
    $logger->close_log();
    exit FAILURE;
}

# Get the username and password for the L8_OLITIRS database
my ($user_id, $password) = ias_db_GetDBConnectionInfo('L8_OLITIRS');
# Connect to the database
my $dbh = ias_db_ConnectToDB($user_id, $password);
if (!defined $dbh)
{
    $logger->fatal_msg(__FILE__, __LINE__,
        "Unable to connect to database: $ias_db_error");
    $logger->close_log();
    exit FAILURE;
}

$dbh->{RaiseError}=1; # Enable error checking (and then die)
$dbh->{PrintError}=0; # Disable warnings (so no doubling up on errors)
$dbh->{ShowErrorStatement}=1; # Show the SQL statement

my @execute_parms;  

# create a query for all unique WRS path / rows based on parameters 
my $gcp_sql = qq{
          SELECT unique wrs_path, wrs_row
          FROM ground_control_points
          WHERE
      };

# At this point, the GCP source has been verified to be defined and is
# 'GLS', 'DOQ', 'TM6' or 'GRI'.  It's required, so start the subquery WHERE
# clause with it.
my $use_and = '';
if ($gcp_source ne 'ALL')
{
    $gcp_sql .= " source = ? ";
    $use_and = "AND";
    push @execute_parms, $gcp_source;
}

# Add the GCP type to the query if it's specified as either 'CONTROL' or
# 'VALIDATION'.  If it's both, we don't need to add it to the query.
if ($gcp_type ne "BOTH")
{
    # Add to sql query.
    $gcp_sql .= " $use_and type = ? ";
    $use_and = "AND";

    # Add type to execute statement.
    push @execute_parms, $gcp_type;
}

# Add the WRS start/stop parameters
$gcp_sql .= " $use_and wrs_path BETWEEN ? AND ?  AND wrs_row BETWEEN ? AND ?".
           " ORDER BY wrs_path, wrs_row";
push @execute_parms, ($wrs_start_path, $wrs_end_path, $wrs_start_row, 
    $wrs_end_row);

# Prepare the query.
my $sth_gcp = $dbh->prepare($gcp_sql);
if (!$sth_gcp)
{
    $error = "Unable to prepare point query: $DBI::errstr";
    $logger->fatal_msg(__FILE__, __LINE__, $error);
    $logger->close_log();
    $dbh->disconnect();
    exit FAILURE;
}

# execute the query with the list of output parameters
if (!$sth_gcp->execute(@execute_parms))
{
    $error = "Unable to execute point query: $DBI::errstr";
    $logger->fatal_msg(__FILE__, __LINE__, $error);
    $logger->close_log();
    $sth_gcp->finish();
    $dbh->disconnect();
    exit FAILURE;
}

# loop through each path/row result fetched
my @fields;
while (@fields = $sth_gcp->fetchrow_array())
{
    my $wrs_path = $fields[0];
    my $wrs_row = $fields[1];

    # create directory names with leading 0's
    my $path_dir = sprintf( "%03d", $wrs_path );
    my $row_dir = sprintf( "%03d", $wrs_row );

    # Have we fetched a new WRS path?
    if ( ($last_path_dir ne "") && ($path_dir ne $last_path_dir) )
    {
        # remove the last WRS path directory if there are no rows in it.
        PurgeWRSPath($last_path_dir);
    }
    # remember the current WRS path directory name for the next fetch
    $last_path_dir = $path_dir;

    # create the WRS path directory, if needed
    my $rc = system( "\\mkdir -p $path_dir" );
    if ($rc != SUCCESS)
    {
        $error = "Error creating GCP path directory $path_dir";
        $logger->fatal_msg(__FILE__, __LINE__, $error);
        $logger->close_log();
        $sth_gcp->finish();
        $dbh->disconnect();
        exit FAILURE;
    }

    # move to the WRS path directory (the wrapper must be executed here).
    if (! chdir( $path_dir ) )
    {
        $error = "Error changing to GCP path directory $path_dir";
        $logger->fatal_msg(__FILE__, __LINE__, $error);
        $logger->close_log();
        $sth_gcp->finish();
        $dbh->disconnect();
        exit FAILURE;
    }
     
    # see if existing directories should be purged (deleted).
    if ($purge_dir)
    {
        my $cmd = "\\rm -rf $row_dir";
        $rc = system($cmd);
        if ($rc != SUCCESS)
        {
            $error = "Error deleting $gcp_home/$path_dir/$row_dir";
            $logger->fatal_msg(__FILE__, __LINE__, $error);
            $logger->close_log();
            $sth_gcp->finish();
            $dbh->disconnect();
            exit FAILURE;
        }
    }

    # execute the wrapper in the path directory and output to the row directory
    my $cmd = "\\get_gcps_wrapper.pl $gcp_parameters "
        . "--wrs_path $wrs_path --wrs_row $wrs_row "
        . "--include_inactive --output $row_dir";
        

    $logger->info_msg(__FILE__, __LINE__, "Retrieving GCP chips for path "
                . "$path_dir / row $row_dir");

    # Call 'get_gcps_wrapper.pl' for the path/row with the full list of 
    # parameters
    $rc = system($cmd);
    if ($rc != SUCCESS)
    {
        $error = "Error retrieving GCP chips";
        $logger->fatal_msg(__FILE__, __LINE__, $error);
        $logger->close_log();
        $sth_gcp->finish();
        $dbh->disconnect();
        exit FAILURE;
    }
 
    # see how many files were created in the row directory
    my @row_files = <$row_dir/*>;
    my $row_file_count = @row_files;

    # if there's one (presumably the GCPLib file) or less, remove the directory
    if( $row_file_count <= 1 )
    {
        $logger->info_msg( __FILE__, __LINE__, "Removing empty row $row_dir" );
        my $cmd = "\\rm -rf $row_dir";
        $rc = system($cmd);
        if ($rc != SUCCESS)
        {
            $error = "Error removing $row_dir";
            $logger->fatal_msg(__FILE__, __LINE__, $error);
            $logger->close_log();
            $sth_gcp->finish();
            $dbh->disconnect();
            exit FAILURE;
        }
    }

    # go back up a level (to the GCP top level)
    if (! chdir( ".." ) )
    {
        $error = "Error leaving GCP directory $path_dir";
        $logger->fatal_msg(__FILE__, __LINE__, $error);
        $logger->close_log();
        $sth_gcp->finish();
        $dbh->disconnect();
        exit FAILURE;
    }
} # end of path/row query loop

# Check the last WRS path directory to see if it's empty
if ($last_path_dir ne "")
{
    PurgeWRSPath($last_path_dir);
}

# Cleanup.
$sth_gcp->finish();
if (!ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK))
{
    $logger->warn_msg(__FILE__, __LINE__,
        "Unable to disconnect from database: $ias_db_error");
}
# Log completion message.
$logger->info_msg(__FILE__, __LINE__, 'Successful completion');
$logger->close_log();

exit SUCCESS;


##############################################################################
#
# NAME:     ProcessOptions()
#
# PURPOSE:  Using the GetOptions call in the Getopt::Long module, this routine
#           parses the argument list, flagging any options found in the
#           parameter list.  If an invalid option is given or the help option
#           is requested, a usage message is shown and the program exits
#           immediately.
#
# RETURNS:  SUCCESS status code if parameter processing succeeded
#           ERROR   status code if the GetOptions call fails 
#
# NOTES:    Mandatory parameters are noted in the GetOptions call as
#           <parm_name>=<parm_type> => \${variable_name}.  Optional
#           parameters are noted in the call as
#           <parm_name>:<parm_type> => \${variable_name}
#
##############################################################################
sub ProcessOptions()
{

    my $help = '';
    my $origCnt = scalar @ARGV;
    my $result = GetOptions(
            "help" => \$help,
            "gcp_home=s" => \$gcp_home,
            "gcp_source:s" => \$gcp_source,
            "gcp_type:s" => \$gcp_type,
            "wrs_start_path:i" => \$wrs_start_path,
            "wrs_start_row:i" => \$wrs_start_row,
            "wrs_end_path:i" => \$wrs_end_path,
            "wrs_end_row:i" => \$wrs_end_row,
            "resolution:i" => \$resolution,
            "purge_dir" => \$purge_dir,
            "write_tarball:i" => \$write_tarball,
            "compression=s" => \$compression
            );
    if (!$result)
    {
        $error = "Failed to read command line parameters";
        $logger->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }

    # Display help/usage information if requested, then quit.
    if ($help)
    {
        ShowUsage();
        exit SUCCESS;
    }

    # Validate required parameters

    # Check that compression is valid
    $compression = uc($compression);
    if ($compression ne "NONE" && $compression ne "GZIP"
        && $compression ne "BZIP")
    {
        $logger->fatal_msg("Compression must be specified as NONE, BZIP "
            . "or GZIP");
        return ERROR;
    }

    # Check that write tarball is valid
    if ($write_tarball != 0 && $write_tarball != 1)
    {
        $logger->fatal_msg("Write tarball must be either 0 or 1");
        return ERROR;
    }

    # GCP source
    if (not (($gcp_source eq 'GLS') || ($gcp_source eq 'DOQ')
        || ($gcp_source eq 'TM6') || ($gcp_source eq 'GRI') ||
        ($gcp_source eq 'ALL')))
    {
        $error = "gcp_source must be specified as 'GLS', 'DOQ', 'TM6', "
            . "'GRI' or 'ALL'";
        $logger->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }
    $gcp_parameters = "--gcp_source $gcp_source";

    # GCP home directory
    if ($gcp_home eq '')
    {
        $error = "gcp_home parameter not specified";
        $logger->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }

    # WRS start/end path and row.  Can be used to subset the query.
    if (($wrs_start_path < 1) || ($wrs_start_path > 233))
    {
        $error = "wrs_start_path must be between 1 and 233";
        $logger->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }

    if (($wrs_start_row < 1) || ($wrs_start_row > 248))
    {
        $error = "wrs_start_row must be between 1 and 248";
        $logger->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }

    if (($wrs_end_path < 1) || ($wrs_end_path > 233))
    {
        $error = "wrs_end_path must be between 1 and 233";
        $logger->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }

    if (($wrs_end_row < 1) || ($wrs_end_row > 248))
    {
        $error = "wrs_end_row must be between 1 and 248";
        $logger->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }

    if ($wrs_start_path > $wrs_end_path)
    {
        $error = "wrs_end_path must be greater than wrs_start_path";
        $logger->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }

    if ($wrs_start_row > $wrs_end_row)
    {
        $error = "wrs_end_row must be greater than wrs_start_row";
        $logger->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }


    # Validate optional parameters if they have been defined, otherwise,
    # set their default values


    # GCP type
    if ($gcp_type)
    {
        if (not (($gcp_type eq 'BOTH') || ($gcp_type eq 'CONTROL')
            || ($gcp_type eq 'VALIDATION')))
        {
            $error = "Invalid gcp_type $gcp_type specified";
            $logger->fatal_msg(__FILE__, __LINE__, $error);
            $logger->info_msg(__FILE__, __LINE__, "Valid gcp_type values are "
               . "CONTROL, VALIDATION, or BOTH");
            return ERROR;
        }
    }
    else
    {
        $gcp_type = 'BOTH';
    }
    $gcp_parameters .= " --gcp_type $gcp_type";

    # Resolution
    if (defined $resolution)
    {
        $gcp_parameters .= " --resolution $resolution";
    }    

    # Tarball
    if (defined $write_tarball)
    {
        $gcp_parameters .= " --write_tarball $write_tarball";
    }

    # Compression type
    if (defined $compression)
    {
       $gcp_parameters .= " --compression $compression";
    }

    if ($purge_dir)
    {
        $logger->info_msg( __FILE__, __LINE__, "Purging current directories" );
    }

    # if parameters checkout, return SUCCESS
    return SUCCESS;
}


##############################################################################
#
# NAME:    ShowUsage()
#
# PURPOSE: Shows the help or usage directives for the program.
#
# RETURNS: Nothing
#
##############################################################################
sub ShowUsage()
{
    my $PROGRAM = PROG;

    print "\n$PROGRAM loops through all WRS path/rows and runs "
        . "'get_gcps_wrapper.pl'.\n\n";
    print "$PROGRAM command line options:\n\n";
    print "--help:           Displays this help summary.\n\n";
    print "REQUIRED PARAMETERS\n";
    print "--gcp_home:       Required parameter specifying the location of\n"
        . "                  the top GCP directory where the retrieved image\n"
        . "                  chips are stored.  The WRS path/row\n"
        . "                  subdirectories are created under this\n"
        . "                  directory.\n\n";
    print "OPTIONAL PARAMETERS\n";
    print "--gcp_source:     Optional parameter specifying the GCP data\n"
        . "                  source.  Valid values are 'GLS', 'DOQ', TM6,\n"
        . "                  'GRI' or 'ALL'. Defaults to 'ALL'.\n";
    print "--gcp_type:       Optional parameter specifying the GCP usage\n"
        . "                  type. Valid values are 'CONTROL', 'VALIDATION',\n"
        . "                  or 'BOTH'. The default value is 'BOTH'.\n";
    print "--wrs_start_path: Optional parameter to override the default \n"
        . "                  starting WRS path of 1. Valid values are between\n"
        . "                  1 and 233.\n";
    print "--wrs_start_row:  Optional parameter to override the default \n"
        . "                  starting WRS row of 1. Valid values are between\n"
        . "                  1 and 248.\n";
    print "--wrs_end_path:   Optional parameter to override the default\n"
        . "                  ending WRS path of 233. Valid values are between\n"
        . "                  between 1 and 233 and must be >= start path, if\n"
        . "                  specified.\n";
    print "--wrs_end_row:    Optional parameter to override the default \n"
        . "                  ending WRS row of 248. Valid values are between\n"
        . "                  1 and 248 and must be >= start row, if\n"
        . "                  specified.\n";
    print "--resolution:     Optional parameter specifying the resolution\n"
        . "                  Valid values are 15, 30 and 60 meter chips.\n";
    print "--purge_dir:      Optional parameter to purge the output row\n"
        . "                  directories prior to extracting chips.  The\n" 
        . "                  default leaves existing files (which could be\n"
        . "                  overwritten).\n";
    print "--write_tarball:  Optional parameter indication whether image\n" 
        . "                  chips should be written as a single tarball\n"
        . "                  (option 1) or written as individual files\n" 
        . "                  (option 0). Valid options are (1 or 0) with\n" 
        . "                  default being 1.\n";
    print "--compression:    Optional parameter used when writing out image \n"
        . "                  chips as a tarball. Valid values are NONE,\n" 
        . "                  BZIP, and GZIP. The default is NONE. \n\n";
}

##############################################################################
#
# NAME:    PurgeWRSPath()
#
# PURPOSE: Removes a WRS path from tree if it's empty.
#
# RETURNS: Nothing
#
##############################################################################
sub PurgeWRSPath($)
{
    my $wrs_path_dir = shift;
    # see how many row directories remain in the WRS path directory
    my @row_dirs = <$wrs_path_dir/*>;
    my $row_dir_count = @row_dirs;

    # if there are no row directories, remove the WRS path directory
    if( $row_dir_count <= 0 )
    {
        $logger->info_msg( __FILE__, __LINE__, 
                           "Removing empty WRS path $wrs_path_dir" );
        my $cmd = "\\rm -rf $wrs_path_dir";
        $rc = system($cmd);
        if ($rc != SUCCESS)
        {
            $error = "Error removing $wrs_path_dir";
            $logger->fatal_msg(__FILE__, __LINE__, $error);
            $logger->close_log();
            $sth_gcp->finish();
            $dbh->disconnect();
            exit FAILURE;
        }
    }
}

