/******************************************************************************
NAME: get_gcpresample_parameters

PURPOSE:
Get parameters from the parameter file

DEVELOPMENT HISTORY:

RETURN VALUE:
Type = int
Value    Description
-----    -----------
ERROR    Error retrieving parameters (check parameter file)
SUCCESS  Parameters retrieved

******************************************************************************/
#include "ias_logging.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "GCPresample.h"

int get_gcpresample_parameters
(
    const char *parm_filename,  /* I: Parameter filename */
    APP_PARMS *parm_data        /* O: Parameters retrieved from
                                      parameter file or OMF */
)
{
    int read_count;             /* Parameter table size */
    int read_status;            /* Read function return status */
    int status;
    char temp_lib_loc[PATH_MAX]; /* Where temporary library will be located */

    /* Set up and read the application parameter table */
    IAS_PARM_DECLARE_TABLE(app_parm_table, 6);

    /* Work order directory */
    IAS_PARM_ADD_STRING(app_parm_table, WO_DIRECTORY,
            "Work order directory", IAS_PARM_REQUIRED,
            0, NULL, 0, NULL, temp_lib_loc, sizeof(temp_lib_loc), 1);

    /* Work order ID */
    IAS_PARM_WORK_ORDER_ID(app_parm_table, parm_data->work_order_id,
            sizeof(parm_data->work_order_id), 1);

    /* GCP directory */
    IAS_PARM_GCP_DIRECTORY(app_parm_table, IAS_PARM_REQUIRED,
            parm_data->gcp_dir, sizeof(parm_data->gcp_dir));

    /* L1G image name */
    /* Did not use the standard definition of L1G_FILENAME because this
       case is optional.  This parameter is used to allow flexibility
       to the cal val team but will generally not be used */
    IAS_PARM_ADD_STRING(app_parm_table, L1G_FILENAME,
            "Name of the L1G image file optionally used to "
            "retrieve projection information",
            IAS_PARM_OPTIONAL,
            0, NULL,
            0, NULL,
            parm_data->img_name,
            sizeof(parm_data->img_name), 0);

    /* Target projection source */
    const char *valid_sources[] = {"L1G", "OMF"};
    IAS_PARM_ADD_STRING(app_parm_table, TARGET_PROJECTION_SOURCE,
            "Source of projection: L1G or OMF", IAS_PARM_OPTIONAL,
            2, valid_sources, 0, NULL, parm_data->projection_source,
            sizeof(parm_data->projection_source), 0);

    /* Flag to use a tarball for the image chips */
    int default_package_chips_flag[] = {1};
    IAS_PARM_ADD_INT(app_parm_table, PACKAGE_CHIPS, "Flag to write the image "
            "chips to a single tarball instead of separately (1 = use tarball, "
            "0 = use separate files)", IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
            1, 0, 1, 1, default_package_chips_flag,
            &parm_data->use_tarball, sizeof(parm_data->use_tarball), 0);

    read_count  = IAS_PARM_GET_TABLE_SIZE(app_parm_table);
    /* Call the routine to provide help for the parameters */
    status = ias_parm_provide_help(parm_filename, app_parm_table, read_count,
            IAS_INPUT_PARAMETERS);
    if (status)
    {
        /* If the status of the provide help is 1 the user asked for help.
           Since the parameter file name contains the help request, copy it
           to the OMF file name so the next call will ask for the same help. */
        if (status == 1)
        {
            strcpy(parm_data->omf_filename, parm_filename);
        }
        else if (status == ERROR)
        {
            IAS_LOG_ERROR("Providing help for application");
            exit(EXIT_FAILURE);
        }
    }
    else
    {
        read_status = ias_parm_read(parm_filename, app_parm_table,
                read_count);
        if (read_status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading application parameters from file %s",
                    parm_filename);
            return ERROR;
        }

        /* Build the OMF name from the work order id and work order dir */
        sprintf(parm_data->omf_filename, "%s/%s.omf", temp_lib_loc,
                parm_data->work_order_id);
    }

    /* Set up and read the OMF parameter table */
    IAS_PARM_DECLARE_TABLE(omf_parm_table, 2);

    IAS_PARM_TARGET_PROJECTION(omf_parm_table, &parm_data->target_projection,
            sizeof(parm_data->target_projection), 1);

    /* Don't use standard parameter here as the range is different */
    IAS_PARM_ADD_INT(omf_parm_table, UTM_ZONE, "UTM Zone", IAS_PARM_REQUIRED,
            IAS_PARM_NOT_ARRAY, 1, 1, 60, 0, NULL,
            &parm_data->utm_zone, sizeof(parm_data->utm_zone), 1);

    read_count = IAS_PARM_GET_TABLE_SIZE(omf_parm_table);

    /* See if the user wanted help or a template with the OMF file. */
    status = ias_parm_provide_help(parm_data->omf_filename, omf_parm_table,
            read_count, IAS_OMF_PARAMETERS);
    if (status)
    {
        if (status == 1)
        {
            /* Help was successful */
            exit(EXIT_SUCCESS);
        }
        else if (status == ERROR)
        {
            /* Help failed */
            IAS_LOG_ERROR("Proving help for application");
            exit(EXIT_FAILURE);
        }
    }

    read_status = ias_parm_read(parm_data->omf_filename, omf_parm_table,
            read_count);
    if (read_status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading OMF parameters from file %s",
                parm_data->omf_filename);
        return ERROR;
    }

    return SUCCESS;
}
