#!/usr/bin/perl -w

#############################################################################
#
# NAME:           get_gcps.pl
#
# PURPOSE:        Get GCP metadata and its associated image chips
#
# RETURNS:        0 on success
#                 1 on failure
#
# Inputs and Outputs:
# I:  Parameter file:  contains the following parameters:
#     - work order directory
#     - work order ID
#     - GCP Source (GLS, DOQ, TM6 or GRI)
#     - GCP Type   (optional -- VALIDATION, CONTROL, or BOTH)
#     - WRS2 Path  (optional)
#     - WRS2 Row   (optional)
#     - begin date (optional -- image chip data acquired no earlier than this
#                               date)
#     - end date   (optional -- image chip data acquired no later than this
#                               date)
#
# I:  OMF file:  contains the following parameters:
#     - Target projection (optional).  Current valid values are (1 (UTM), 6 
#       (PS), or 22 (SOM))
#     - "Bounding box" lat/lon corner coordinates (optional)
#     - UTM zone code (optional, validated if target projection is specified as
#       UTM)
#
# O:  GCP output directory name written to OMF
# O:  GCP file containing list of GCPs
# O:  GCP metadata and image chips, reprojected if necessary
#
#############################################################################

use strict;
use DBI;
use File::Basename;
use Logger;         
use ias_processing_system;               # Defines $ias_error_msg
use ias_const;                           # Defines SUCCESS and ERROR
use ias_gcp qw(ias_gcp_write_gcplib ias_gcp_tar_image_chips IAS_GCP_BZIP2 
    IAS_GCP_GZIP IAS_GCP_NO_COMPRESSION IAS_GCP_IMAGE_CHIPS_FILENAME
    ias_gcp_create_tar_filename IAS_GCP_NO_MD5SUM);
use ias_db;

# Declare program constant
use constant PROG => basename($0);

# Subroutine forward declarations
sub parse_parm($);
sub parse_omf($);

# Coordinates for scene boundary
my $upper_left_lat;
my $upper_left_lon;
my $upper_right_lat;
my $upper_right_lon;
my $lower_right_lat;
my $lower_right_lon;
my $lower_left_lat;
my $lower_left_lon;

# Option variables 
my $write_tarball = 1;      # Tarball or individual chips option.  The default
                            # is to use a tarball.
my $compression = IAS_GCP_NO_COMPRESSION; # Compression type for the tarball

# The main parameter file that is passed in on the command-line
my $parm_name = '';
my $gcp_directory = '';        # GCP directory
my $omf_name = '';             # OMF filename
my $requested_gcp_type = '';   # GCP type
my $gcp_source = '';           # GCP source
my $source_database = '';      # Source database, which allows pointing at a
                               # alternative database for the gcps
my @chip_sensors = ();         # Image sensors to look for in the query
my $wo_directory = '';
my $work_order_id = '';
my $wrs_path;                  # Query ground_control_points table by path
my $wrs_row;                   # Query ground_control_points table by row
my $resolution_x;              # Query by chip resolution x
my $resolution_y;              # Query by chip resolution y 
my $include_inactive = 0;      # Flag to include inactive GCPs and chips in the
                               # results

# OMF parameter file
my $target_projection = 0;

# Date structures for date range queries against date_acquired in table
# image_chips
my $start_date;              # Date after which queried image chip(s) were
                             # acquired
my $end_date;                # Date before which queried image chip(s) were
                             # acquired
my $converted_date_acquired; # GCP date acquired converted to be in the same
                             # format as start_date and end_date

# SQL variables
my $dbh;                       # Database handle
my $sth_gcp;                   # GCP statement handle
my $sth_image_chips;           # Image chips statement handle

# Miscellaneous temporary variables
my $input_line         = '';   # Temporary line read from file
my $prev_gcp_proj      = '';   # Projection of previous GCP
my @gcp_data;                  # Array of GCP hashes
my $status;
my %corners_present = ();      # Hash containing key-value pairs representing
                               # the presence of specified corner
                               # coordinates.  If --corners is specified,
                               # ALL coordinate sets must be entered

# Initialize the IAS processing system module
if (ias_init_system() == ERROR)
{
    print STDERR PROG . ": $ias_error_msg\n";
    exit FAILURE;
}

# Set up the logger
my $logger = Logger->new("STDOUT");
if (!$logger)
{
    print STDERR PROG . ": ERROR: Failed to establish logger object for"
        . " STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name(PROG);
set_default_log($logger);

# Check the instance variable
my $processing_instance = $ENV{IAS_INSTANCE_NAME};
if (!$processing_instance)
{
    $logger->fatal_msg("The IAS_INSTANCE_NAME enironment variable is not set");
    $logger->close_log();
    exit FAILURE;
}

# Verify one parameter was passed to this routine -- the name of the
# processing parameter file
if (scalar @ARGV != 1)
{
    $logger->fatal_msg("$0 requires one parameter, the name of the processing"
        . " parameter file");
    $logger->close_log();
    exit FAILURE;
}

# Grab the parameter file that was passed in
$parm_name = $ARGV[0];

# Validate the parameter file's existence
if (!-e $parm_name)
{
    $logger->fatal_msg("The $parm_name parameter file specified does not"
        . " exist");
    $logger->close_log();
    exit FAILURE;
}

# Parse the parameter file.
$status = parse_parm($parm_name);
if ($status != SUCCESS)
{
    $logger->fatal_msg("Error parsing parameter file $parm_name");
    $logger->close_log();
    exit FAILURE;
}

# Parse the OMF.
$omf_name = $wo_directory . "/" . $work_order_id . ".omf";
$status = parse_omf($omf_name);
if ($status != SUCCESS)
{
    $logger->fatal_msg("Unable to parse OMF file $omf_name");
    $logger->close_log();
    exit FAILURE;
}

# One or more of the following can be specified in the parameter files:
#    i.  Lat/lon corner coordinates (all four corners must be specified)
#   ii.  WRS2 path AND WRS2 row (both must be specified)
#
my $have_bounding_box = 0;
my $have_wrs_pathrow  = 0;

# If both path and row have been specified, we've got a useable path/row
# combination.
if ($wrs_path && $wrs_row)
{
    $have_wrs_pathrow = 1;
}

# If we've specified lat/lon coordinates for a bounding box, check to make
# sure we've got the correct number (4); if we have any nonzero number but
# 4, that's an error.  If we haven't specified coordinates for a bounding
# box, and we haven't specified a WRS path/row combination, that's also an
# error.
if (%corners_present)
{
    my $corners_present_size = keys %corners_present;
    if ($corners_present_size == 4)
    {
        $have_bounding_box = 1;
    }
    else
    {
        $logger->fatal_msg("Invalid number of corner coordinate points"
            . " specified:  $corners_present_size");
        $logger->close_log();
        exit FAILURE;
    }
}
else
{
    if (!$have_wrs_pathrow)
    {
        $logger->fatal_msg("No location information provided for use in"
            . " selecting GCP image chips.  Specify a complete set of"
            . " lat/lon corner points (4) and/or a WRS path/row combination"
            . " (both path and row must be specified)");
        $logger->close_log();
        exit FAILURE;
    }
}

# Issue a starting message.
$logger->info_msg("Starting GCP retrieval");

# If the output GCP directory does not exist, create it.
if (!(-d $gcp_directory))
{
    $status = system("mkdir -p $gcp_directory");
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Unable to make directory $gcp_directory for"
            . " the output GCPs");
        $logger->close_log();
        exit FAILURE;
    }
}

# Get the username and password for the COMMON database
my ($user_id, $password) = ias_db_GetDBConnectionInfo();

# Connect to the database
$dbh = ias_db_ConnectToDB($user_id, $password);
if (!defined $dbh)
{
    $logger->fatal_msg("Unable to connect to database");
    $logger->close_log();
    exit FAILURE;
}

my %attributes = (
    "GCP_DATA_VERSION_$source_database" => "",
    "GROUND_CONTROL_POINTS_$source_database" => "" ,
    "IMAGE_CHIPS_$source_database" => "",
    );

# Grab the synonym names
$status = ias_db_GetProcessingInstanceAttributes(\%attributes,
    $processing_instance, $dbh, $logger);
if ($status != SUCCESS)
{
    $logger->fatal_msg("Unable to retrieve processing instance attributes");
    $logger->close_log();
    $dbh->disconnect();
    exit FAILURE;
}

# Use the lookup synonyms for this collection and data source
my $image_chips_table = $attributes{"IMAGE_CHIPS_$source_database"};
my $gcp_data_version_table = $attributes{"GCP_DATA_VERSION_$source_database"};
my $ground_control_points_table
    = $attributes{"GROUND_CONTROL_POINTS_$source_database"};

# Get current GCP Data Version from database
my $overall_gcp_version = get_gcp_version($dbh);
if (!defined $overall_gcp_version)
{
    $logger->fatal_msg("Unable to retrieve current overall GCP version from"
        . " the database");
    $logger->close_log();
    $dbh->disconnect();
    exit FAILURE;
}

# Set LongReadLen to control the max length of long column types (BLOB/CLOB).
# The DBI docs state this should be set to a value slightly larger than the
# largest field value.  Allow for the largest chips which are 64x64 with 4 
# byte samples and add a few bytes.
$dbh->{LongReadLen} = (4 * 64 * 64) + 16;

# Some notes on the Oracle spatial query elements:
# SHAPE = Oracle locator column in the ground_control_points table.
# 2003 = Geometry Type. 2003 indicates a two dimensional polygon or
#     surface search
# 8307 = Spatial Reference Identifier (SRID). It is an integer value key used
#     to identify a coordinate system to be associated with the geometry.  8307
#     indicates a WGS 84 longitude/latitude search.
# NULL = SDO_POINT object type.
# SDO_elem_info_array = This attribute lets you know how to interpret the
#     ordinates stored in the SDO_ORDINATES attribute. 1 indicates the
#     starting offset of where the first ordinate for the element is stored.
#     1003 indicates the type of element, in this case an interior polygon ring.
#     The last 1 indicates how the sequence of ordinates for the element are
#     interpreted.
#
# The coordinates of the bounding box area being queried are specified
# in the sdo_ordinate_array as follows:
#                  SDO_ORDINATE_ARRAY(
#                      <upper left longitude>, <upper left latitude>,
#                      <upper right longitude>, <upper right latitude>,
#                      <lower right longitude>, <lower right latitude>,
#                      <lower left longitude>, <lower left latitude>,
#                      <upper left longitude>, <upper left latitude>)
# Note the first coordinate is repeated at the end.

# Note the latitude, longitude, and elevation are converted to
# force a zero to be prepended before the decimal if the value is between
# 1 and 0 or -1 and 0.

# Prepare the point query.
my @execute_parms;
my $gcp_sql = qq{
    SELECT ground_control_points_id, active,
           TO_CHAR(date_updated, 'MM-DD-YYYY'), source,
           TO_CHAR(latitude,'9990.9999999'),
           TO_CHAR(longitude,'9990.9999999'),
           TO_CHAR(elevation,'9990.999'), type,
           update_reason, version, wrs_path, wrs_row
      FROM $ground_control_points_table
      WHERE
};

my $include_and = '';

if ($have_bounding_box)
{
    # We've specified a lat/lon bounding box.  Add this to the
    # query.
    $gcp_sql .= qq{ sdo_filter(
              SHAPE,
              SDO_geometry(
                  2003, 8307, NULL,
                  SDO_elem_info_array(1,1003,1),
                  SDO_ORDINATE_ARRAY(
                       ?, ?,
                       ?, ?,
                       ?, ?,
                       ?, ?,
                       ?, ?)),
              'querytype = WINDOW') = 'TRUE'};
    $include_and = "AND";

    # Add these to the execution parameters list.  To "close" the polygon,
    # enter the upper-left longitude and latitude as the last set of
    # coordinates.
    push @execute_parms, ($upper_left_lon, $upper_left_lat,
        $upper_right_lon, $upper_right_lat, $lower_right_lon,
        $lower_right_lat, $lower_left_lon, $lower_left_lat,
        $upper_left_lon, $upper_left_lat);
}

if ($have_wrs_pathrow)
{
    # Include the WRS path and row in the where clause
    $gcp_sql .= " $include_and wrs_path = ? ";
    $gcp_sql .= " AND wrs_row = ? ";

    # Add the path/row to the parameters list
    push @execute_parms, ($wrs_path, $wrs_row);
}

# Add the GCP type to the query if it's specified as either 'CONTROL' or
# 'VALIDATION'.  If it's both, we don't need to add it to the query.
if ($requested_gcp_type ne "BOTH")
{
    # Add to sql query.
    $gcp_sql .= " AND type = ? ";

    # Add to execute statement.
    push @execute_parms, $requested_gcp_type;
}

# At this point, the GCP source has been verified to be defined and is 'GLS',
# 'DOQ', 'TM6', 'GRI' or 'ALL'.  If it isn't 'ALL', add to the where clause to
#filter by the source.
if ($gcp_source ne 'ALL')
{
    $gcp_sql .= "AND source = ? ";
    push @execute_parms, $gcp_source;
}

# Filter out inactive GCPs unless inactive points have been requested
if (!$include_inactive)
{
    $gcp_sql .= "AND active = 'Y' ";
}

# Sort the GCPs by the ID.  This is done to make sure the GCPs are always
# in a consistent order.  If they aren't in a consistent order, the
# precision solution can be impacted in a very minor way since it is
# slightly sensitive to the order of the points that correlate.
$gcp_sql .= " ORDER BY ground_control_points_id";

# Prepare and execute the query.
$sth_gcp = $dbh->prepare($gcp_sql);
if (!$sth_gcp)
{
    $logger->fatal_msg("Unable to prepare point query: $DBI::errstr");
    $logger->close_log();
    $dbh->disconnect();
    exit FAILURE;
}

if (!$sth_gcp->execute(@execute_parms))
{
    $logger->fatal_msg("Unable to execute point query: $DBI::errstr");
    $logger->close_log();
    $sth_gcp->finish();
    $dbh->disconnect();
    exit FAILURE;
}

# Variables for the ground_control_points spatial GCP query.
my ($pt_id, $gcp_active, $gcp_date_updated, $source, $lat, $lon, $elev,
    $gcp_type, $update_reason, $gcp_version, $gcp_wrs_path, $gcp_wrs_row);

if (!$sth_gcp->bind_columns(undef, \$pt_id, \$gcp_active, \$gcp_date_updated,
        \$source, \$lat, \$lon, \$elev, \$gcp_type, \$update_reason,
        \$gcp_version, \$gcp_wrs_path, \$gcp_wrs_row))
{
    $logger->fatal_msg("Unable to bind columns on gcp statement: $DBI::errstr");
    $logger->close_log();
    $sth_gcp->finish();
    $sth_image_chips->finish();
    $dbh->disconnect();
    exit FAILURE;
}


# Prepare the chip query statement.
my @chip_execute_parms;
my $image_chips_sql = qq{
    SELECT image_chips_id, ground_control_points_id, version, active,
           TO_CHAR(date_acquired, 'MM-DD-YYYY'),
           TO_CHAR(date_updated, 'MM-DD-YYYY'), reference_line,
           reference_sample, projection, projection_y, projection_x, zone_code,
           pixel_size_x, pixel_size_y, lines, samples, sensor,
           abs_rel_flag, chip_data_type, select_method, moravec_rank,
           source_wrs_path, source_wrs_row, chip_data 
    FROM $image_chips_table
    WHERE ground_control_points_id = ?
};

# Filter out inactive chips unless inactive points have been requested
if (!$include_inactive)
{
    $image_chips_sql .= " AND active = 'Y' ";
}

# Check if RESOLUTION parm has be defined
if (defined $resolution_x && defined $resolution_y)
{
    $image_chips_sql .= " AND pixel_size_x = ? AND pixel_size_y = ?";
    push @chip_execute_parms, $resolution_x;
    push @chip_execute_parms, $resolution_y;
}

# Check if the SENSOR parm has been defined
if ($chip_sensors[0] ne "ALL")
{
    # Build the where clause sensor list
    $image_chips_sql .= " AND sensor in (?";

    # Add enough parameters for the rest of the sensors in the list
    for (my $i = 1; $i < scalar @chip_sensors; $i++)
    {
        $image_chips_sql .= ", ?";
    }
    
    # Finish the sql statement
    $image_chips_sql .= ")"; 

    # Add the sensors to the parameter list
    push @chip_execute_parms, @chip_sensors;
}

# Make sure the image chips are sorted in ascending order for consistency
$image_chips_sql .= " ORDER BY image_chips_id";

# Prepare the chip query
$sth_image_chips = $dbh->prepare($image_chips_sql);
if (!$sth_image_chips)
{
    $logger->fatal_msg("Unable to prepare chip query: $DBI::errstr");
    $logger->close_log();
    $sth_gcp->finish();
    $dbh->disconnect();
    exit FAILURE;
}

# The query we're using is executed in a loop (for each ground control
# point ID).  The begin and/or end dates are identical for each point.  In
# order for these to be properly "bound" in the query, they should be
# set at the same time as the ground control point ID (i.e. each time the
# loop iterates).

# Set up some query variables and bind the query columns to them.

# Variables for the image_chips chip information query.
my ($chip_name, $ground_control_points_id, $chip_version, $chip_active,
    $date_acquired, $chip_date_updated, $reference_line, $reference_sample,
    $projection, $projection_y, $projection_x, $zone_code, $pixel_size_x,
    $pixel_size_y, $lines, $samples, $sensor, $abs_rel_flag, $chip_data, 
    $chip_data_type, $select_method, $moravec_rank, $chip_wrs_path,
    $chip_wrs_row) = '';

if (!$sth_image_chips->bind_columns(undef, \$chip_name,
            \$ground_control_points_id, \$chip_version, \$chip_active,
            \$date_acquired, \$chip_date_updated, \$reference_line,
            \$reference_sample, \$projection, \$projection_y, \$projection_x,
            \$zone_code, \$pixel_size_x, \$pixel_size_y, \$lines, \$samples,
            \$sensor, \$abs_rel_flag, \$chip_data_type, \$select_method,
            \$moravec_rank, \$chip_wrs_path, \$chip_wrs_row, \$chip_data))
{
    $logger->fatal_msg("Unable to bind columns on image chip statement:"
        . " $DBI::errstr");
    $logger->close_log();
    $sth_gcp->finish();
    $sth_image_chips->finish();
    $dbh->disconnect();
    exit FAILURE;
}

# Loop through each query result, writing out the chip data using the
# appropriate filename from the database.
my $chip_count = 0;
while ($sth_gcp->fetch())
{
    # Format the point id with leading zeroes
    $pt_id = sprintf("%010d", $pt_id);

    # Push ID to front of the array
    unshift @chip_execute_parms, $pt_id;

    # Execute the query
    if (!$sth_image_chips->execute(@chip_execute_parms))
    {
        $logger->fatal_msg("Unable to execute image chip query for GCP id"
            . " $pt_id: $DBI::errstr");
        $logger->close_log();
        $sth_gcp->finish();
        $sth_image_chips->finish();
        $dbh->disconnect();
        exit FAILURE;
    }
    
    # Remove ID from the front of the array so its ready for the next ID
    shift @chip_execute_parms;
    
    # Loop over the image chips found
    while ($sth_image_chips->fetch())
    {
        # Exclude image chips that lie outside of the user-specified date range.
        # First convert the date from MM-DD-YYYY format to YYYYMMDD format.
        # This allows easy comparisons.  start and end date are in this format
        # already
        my $converted_date_acquired;
        if ($date_acquired =~ m/(\d\d)-(\d\d)-(\d\d\d\d)/)
        {
            $converted_date_acquired = "$3$1$2";
        }
        else
        {
            $logger->fatal_msg("Invalid format for date_acquired"
                . " $date_acquired; should be in format MM-DD-YYYY");
            $logger->close_log();
            $sth_gcp->finish();
            $sth_image_chips->finish();
            $dbh->disconnect();
            exit FAILURE;
        }

        # Now compare the converted date acquired to the start and end dates
        if ($start_date)
        {
            next if ($converted_date_acquired < $start_date);
        }
        if ($end_date)
        {
            next if ($converted_date_acquired > $end_date);
        }

        # Increment the "valid" chip count.
        $chip_count++;

        # Periodically output a message
        if ($chip_count % 100 == 0)
        {
            $logger->info_msg("Retrieving chip # $chip_count");
        }

        my $gcp_record = {};
        my $output_filename = "$gcp_directory/$chip_name";

        # Verify all GCPs have the same projection
        if (($prev_gcp_proj ne "") && ($projection ne "")
            && ($prev_gcp_proj ne $projection))
        {
            $logger->fatal_msg("Not all GCPs are of the same projection");
            $logger->fatal_msg("Previous GCP was $prev_gcp_proj, Current"
                . " GCP is $projection");
            $logger->close_log();
            $sth_gcp->finish();
            $sth_image_chips->finish();
            $dbh->disconnect();
            exit FAILURE;
        }

        # Verify the data type of the current chip is supported by our software
        if (($chip_data_type ne 'BYTE') && ($chip_data_type ne 'I*2')
            && ($chip_data_type ne 'UI*2') && ($chip_data_type ne 'R*4'))
        {
            $logger->fatal_msg("Unsupported chip data type: $chip_data_type");
            $logger->fatal_msg("Previous GCP was $prev_gcp_proj, Current"
                . " GCP is $projection");
            $logger->close_log();
            $sth_gcp->finish();
            $sth_image_chips->finish();
            $dbh->disconnect();
            exit FAILURE;
        }

        # strip the white space left from the to_char operation in the query
        $elev =~ s/^\s*(.*?)\s*$/$1/;
        $lat =~ s/^\s*(.*?)\s*$/$1/;
        $lon =~ s/^\s*(.*?)\s*$/$1/;

        # Add vars to gcp_record hash
        $gcp_record->{'point_id'}               = $pt_id;
        $gcp_record->{'point_active'}           = $gcp_active;
        $gcp_record->{'point_date_updated'}     = $gcp_date_updated;
        $gcp_record->{'chip_name'}              = $chip_name;
        $gcp_record->{'reference_line'}         = $reference_line;
        $gcp_record->{'reference_sample'}       = $reference_sample;
        $gcp_record->{'latitude'}               = $lat;
        $gcp_record->{'longitude'}              = $lon;
        $gcp_record->{'projection_x'}           = $projection_x;
        $gcp_record->{'projection_y'}           = $projection_y;
        $gcp_record->{'elevation'}              = $elev;
        $gcp_record->{'update_reason'}          = $update_reason;
        $gcp_record->{'point_version'}          = $gcp_version;
        $gcp_record->{'point_path'}             = $gcp_wrs_path;
        $gcp_record->{'point_row'}              = $gcp_wrs_row;
        $gcp_record->{'pixel_size_x'}           = $pixel_size_x;
        $gcp_record->{'pixel_size_y'}           = $pixel_size_y;
        $gcp_record->{'chip_size_lines'}        = $lines;
        $gcp_record->{'chip_size_samples'}      = $samples;
        $gcp_record->{'source'}                 = $source;
        $gcp_record->{'point_type'}             = $gcp_type;
        $gcp_record->{'projection'}             = $projection;
        $gcp_record->{'zone'}                   = $zone_code;
        $gcp_record->{'chip_acq_date'}          = $date_acquired;
        $gcp_record->{'abs_rel_flag'}           = $abs_rel_flag;
        $gcp_record->{'chip_sensor'}            = $sensor;
        $gcp_record->{'chip_data_type'}         = $chip_data_type;
        $gcp_record->{'chip_active'}            = $chip_active;
        $gcp_record->{'chip_date_updated'}      = $chip_date_updated;
        $gcp_record->{'chip_version'}           = $chip_version;
        $gcp_record->{'chip_path'}              = $chip_wrs_path;
        $gcp_record->{'chip_row'}               = $chip_wrs_row;
        $gcp_record->{'chip_select_method'}     = $select_method;
        $gcp_record->{'moravec_rank'}           = $moravec_rank;

        if (!$write_tarball)
        {        
            # Write the image chip to a file
            my $chipfile_status = open(OUTPUT_CHIP, "> $output_filename");
            if (!defined($chipfile_status))
            {
                $logger->fatal_msg("Can't open output file: $output_filename");
                $sth_gcp->finish();
                $sth_image_chips->finish();
                $dbh->disconnect();
                $logger->close_log();
                exit FAILURE;
            }
            $chipfile_status = print OUTPUT_CHIP $chip_data;
            if ($chipfile_status == 0)
            {
                $logger->fatal_msg("Writing to $output_filename");
                $sth_gcp->finish();
                $sth_image_chips->finish();
                $dbh->disconnect();
                $logger->close_log();
                exit FAILURE;
            }
            $chipfile_status = close(OUTPUT_CHIP);
            if (!$chipfile_status)
            {
                $logger->fatal_msg("Closing chip file $output_filename");
                $sth_gcp->finish();
                $sth_image_chips->finish();
                $dbh->disconnect();
                $logger->close_log();
                exit FAILURE;
            }
        }
        else
        {
            $gcp_record->{'image_chip'} = $chip_data;
        }

        # Add gcp_record hash to gcp_data array
        push @gcp_data, $gcp_record;

        $prev_gcp_proj = $projection;
    }
}

# Free up the database resources
$sth_gcp->finish();
$sth_image_chips->finish();
$dbh->disconnect();

$logger->info_msg("Retrieved a total of $chip_count chips");

# Write out the tarball if option is enabled
if ($write_tarball)
{
    my $filename = ias_gcp_create_tar_filename(IAS_GCP_IMAGE_CHIPS_FILENAME, 
        $compression);
    if (!defined $filename)
    {   
        $logger->fatal_msg("Creating tarball filename");
        $logger->close_log();
        exit FAILURE;
    }

    my $status = ias_gcp_tar_image_chips(\@gcp_data,
        "$gcp_directory/$filename", $compression, IAS_GCP_NO_MD5SUM);
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Unable to write out image chips tarball");
        $logger->close_log();
        exit FAILURE;    
    }

    # Remove all image chip data from records
    foreach my $gcp (@gcp_data)
    {
        delete $gcp->{'image_chip'};
    }
}

# Write the GCP library summary file for the retrieved set of chips.
my $gcp_filename = "$gcp_directory/GCPLib";
$status = ias_gcp_write_gcplib($gcp_filename, $overall_gcp_version,
    \@gcp_data);
if ($status != 0)
{
    $logger->fatal_msg("Unable to write GCPLib file");
    $logger->close_log();
    exit FAILURE;
}

# Update the OMF.
$status = update_omf($omf_name, $gcp_directory, $overall_gcp_version);
if ($status != 0)
{
    $logger->fatal_msg("Unable to update OMF file $omf_name");
    $logger->close_log();
    exit FAILURE;
}

# Only run the resampler if GCP chips were found and the source and target
# projections are UTM.
# All GCP projections should be the same, just look at the last GCP's
# projection.
if (($chip_count > 0) && ($projection eq "UTM") && ($target_projection == 1))
{
    # Run the resampler
    $status = system("$ENV{IAS_BIN}/gcpresample", "$parm_name");
    if ($status != 0)
    {
        $logger->fatal_msg("Call to gcpresample application failed");
        $logger->close_log();
        exit FAILURE;
    }
}

# Issue a completion message (mainly to help determine how long the app
# took to run).
$logger->info_msg("GCP retrieval completed successfully.");

# Free up the logger resources
$logger->close_log();

exit SUCCESS;


##############################################################################
# Subroutines below
##############################################################################



##############################################################################
#
# Name: parse_parm
#
# Purpose: Parse a parameter file that contains parameters needed by DGC.
#
# Inputs and Outputs:
# I: parameter filename
# O: Populated global variables for WO directory, GCP directory,
#    WO ID, GCP Type, GCP Source
# O: (Optional) populated global variables for beginning date, ending date,
#    WRS2 path, and WRS2 row
#
# Returns SUCCESS or ERROR
#
# Notes:  If the parameter file processed here was created when running the
#         script 'get_gcps_wrapper.pl', the parameters will be checked a
#         second time after they're read from the parameter file (in
#         'get_gcps_wrapper.pl', the parameters are validated after they're
#         read from the command line before they're written to the
#         parameter file).
#
##############################################################################
sub parse_parm($)
{
    # Get the parameter file name.
    my $parm_filename = shift;

    # Open the file to read.
    my $status = open (INPUT, "< $parm_filename");
    if (!defined($status))
    {
        $logger->fatal_msg("Can't open input parameter file: $parm_filename");
        return ERROR;
    }

    # Process each line in the file.
    while ($input_line = <INPUT>)
    {
        # Work order directory
        if ($input_line =~ /^\s*WO_DIRECTORY\s*=\s*"?(.*)"?/)
        {
            if ($1 =~ /([\w\.\/]*)/)
            {
                $wo_directory = $1;
            }
            else
            {
                $logger->fatal_msg("Invalid work order directory string");
                return ERROR;
            }
        }

        # Work order ID
        if ($input_line =~ /^\s*WORK_ORDER_ID\s*=\s*"?(.*)"?/)
        {
            if ($1 =~ /(\w+)/)
            {
                $work_order_id = $1;
            }
            else
            {
                $logger->fatal_msg("Invalid work order ID string");
                return ERROR;
            }
        }

        # GCP directory
        if ($input_line =~ /^\s*GCP_DIRECTORY\s*=\s*"?(.*)"?/)
        {
            if ($1 =~ /(([\w\/]*\.?\w*))/)
            {
                $gcp_directory = $1;
            }
            else
            {
                $logger->fatal_msg("Invalid GCP directory string");
                return ERROR;
            }
        }

        # GCP type (CONTROL or VALIDATION or BOTH)
        if ($input_line =~ /^\s*GCP_TYPE\s*=\s*"?(.*)"?/)
        {
            if ($1 =~ /(\w+)/)
            {
                $requested_gcp_type = $1;
            }
            else
            {
                $logger->fatal_msg("Invalid GCP type specification");
                return ERROR;
            }
        }

        # GCP source (GLS or DOQ or TM6 or GRI)
        if ($input_line =~ /^\s*GCP_SOURCE\s*=\s*"?(\w+)"?/)
        {
            $gcp_source = $1;
        }

        # Source database, which allows querying a different database
        # for the gcps, image_chips, and gcp version
        if ($input_line =~ /^\s*GCP_SOURCE_DATABASE\s*=\s*"?(.*)"?/)
        {
            if ($1 =~ /(\w+)/)
            {
                $source_database = $1;
            }
            else
            {
                $logger->fatal_msg("Invalid GCP data source specification");
                return ERROR;
            }
        }

        # Starting acquisition date.  If present, convert to a string in the
        # form of YYYYMMDD for easy comparison.
        if ($input_line =~ /^\s*START_DATE\s*=\s*"(.*)"/)
        {
            if ($1 =~ /(\d{4})\-(\d{2})\-(\d{2})/)
            {
                $start_date = "$1$2$3";
            }
            else
            {
                $logger->fatal_msg("Invalid start date string format");
                return ERROR;
            }
        }

        # Ending acquisition date
        if ($input_line =~ /^\s*END_DATE\s*=\s*"(.*)"/)
        {
            if ($1 =~ /(\d{4})\-(\d{2})\-(\d{2})/)
            {
                $end_date = "$1$2$3";
            }
            else
            {
                $logger->fatal_msg("Invalid end date string format");
                return ERROR;
            }
        }

        # WRS path
        if ($input_line =~ /^\s*WRS_PATH\s*=\s*(.*)/)
        {
            if ($1 =~ /(\d+)/)
            {
                $wrs_path = $1;
            }
            else
            {
                $logger->fatal_msg("Invalid WRS path format");
                return ERROR;
            }
        }

        # WRS row
        if ($input_line =~ /^\s*WRS_ROW\s*=\s*(.*)/)
        {
            if ($1 =~ /(\d+)/)
            {
                $wrs_row = $1;
            }
            else
            {
                $logger->fatal_msg("Invalid WRS row format");
                return ERROR;
            }
        }

        # Resolution
        if ($input_line =~ /^\s*RESOLUTION\s*=\s*(.*)/)
        {
            my $value = $1;
            if ($value =~ /(-?\d{1,3}\.?\d{0,})\s*/)
            {
                # A single resolution was provided, so apply it to both X and Y
                $resolution_x = $1;
                $resolution_y = $1;
            }
            elsif ($value 
                =~ /\((-?\d{1,3}\.?\d{0,})\s*,\s*(-?\d{1,3}\.?\d{0,}\s*)\)/x)
            {
                # An array of two resolutions for X, Y were provided
                $resolution_x = $1;
                $resolution_y = $2;
            }
            else
            {
                $logger->fatal_msg("Invalid RESOLUTION format '$value'");
                return ERROR;
            }
        }

        # Image Sensor
        if ($input_line =~ /^\s*SENSOR\s*=\s*(.*)\s*$/)
        {
            my $value = $1;
            if ($value =~ /\(?([^(].*[^)])\)?/)
            {
                @chip_sensors = split(',', $1);
                # strip spaces and quotes from the values
                foreach my $sensor (@chip_sensors)
                {
                    $sensor =~ s/^\s+|\s+$//g;
                    $sensor =~ s/"//g;
                }
            }
            else
            {
                $logger->fatal_msg("Invalid SENSOR format '$value'");
                return ERROR;
            }
        }

        # The INCLUDE_INACTIVE parameter allows retrieving inactive GCPs and
        # image chips, which are normally filtered out by the queries
        if ($input_line =~ /^\s*INCLUDE_INACTIVE\s*=\s*(.*)\s*$/)
        {
            my $value = $1;
            if ($value == 1 || $value == 0)
            {
                $include_inactive = $value;
            }
            else
            {
                $logger->fatal_msg("Invalid INCLUDE_INACTIVE value: '$value' -"
                    . " Zero or one expected");
                return ERROR;
            }
        }

        # Read the optional parameter indicating whether to put the image chips
        # in a tarball
        if ($input_line =~ /^\s*PACKAGE_CHIPS\s*=\s*(.*)\s*$/)
        {
            my $value = $1;
            if ($value == 1 || $value == 0)
            {
                $write_tarball = $value;
            }
            else
            {
                $logger->fatal_msg("Invalid PACKAGE_CHIPS value: '$value' -"
                    . " Zero or one expected");
                return ERROR;
            }
        }

        # Read the optional parameter indicating which compression method,
        # if any, to use on the image chips tarball
        if ($input_line =~ /^\s*PACKAGE_COMPRESSION\s*=\s*"?(\w+)"?/)
        {
            my $value = uc($1);

            # Change compression into a usable format
            if ($value eq "NONE")
            {
                $compression = IAS_GCP_NO_COMPRESSION;
            }
            elsif ($value eq "GZIP")
            {
                $compression = IAS_GCP_GZIP;
            }
            elsif ($value eq "BZIP")
            {
                $compression = IAS_GCP_BZIP2;
            }
            else
            {
                $logger->fatal_msg("Invalid PACKAGE_COMPRESSION value: "
                    . "'$value' - NONE, GZIP, or BZIP expected");
                return ERROR;
            }
        }
    }

    # Close the input file
    close (INPUT);

    #***********************************************************************
    #                     PARAMETER VALIDATION
    #***********************************************************************
    # Validate the work order directory and work order ID
    if ($wo_directory eq '')
    {
        $wo_directory = '.';
    }

    # Validate the GCP type and source
    if ($requested_gcp_type eq '')
    {
        $requested_gcp_type = 'BOTH';
    }
    else
    {
        if (not (($requested_gcp_type eq 'BOTH')
            || ($requested_gcp_type eq 'CONTROL')
            || ($requested_gcp_type eq 'VALIDATION')))
        {
            $logger->fatal_msg("GCP type must be 'CONTROL', 'VALIDATION',"
                . " or 'BOTH'");
            return ERROR;
        }
    }
    if (not (($gcp_source eq 'GLS') || ($gcp_source eq 'DOQ')
        || ($gcp_source eq 'TM6') || ($gcp_source eq 'GRI') ||
        ($gcp_source eq 'ALL')))
    {
        $logger->fatal_msg("GCP source must be 'GLS', 'DOQ', 'TM6', 'GRI' "
                         . "or 'ALL'");
        return ERROR;
    }

    # Default the gcp data source to the primary source
    if ($source_database eq '')
    {
        $source_database = 'PRIMARY';
    }

    # If sensor not specified then set to ALL
    if (scalar @chip_sensors == 0)
    {
        push @chip_sensors, "ALL";
    }

    # Check whether the calculated end date occurs after the calculated
    # start date.
    if ($start_date && $end_date)
    {
        if ($end_date <= $start_date)
        {
            $logger->fatal_msg("End date $end_date should occur after start "
                . " date $start_date");
            return ERROR;
        }
    }

    # Check whether the WRS path and/or row have been specified.  If the
    # path has been specified but the row hasn't, or vice versa, treat it
    # as an error -- both must be specified
    if ($wrs_path)
    {
        if (($wrs_path < 1) || ($wrs_path > 233))
        {
            $logger->fatal_msg("WRS path must be between 1 and 233");
            return ERROR;
        }

        if ($wrs_row)
        {
            if (($wrs_row < 1) || ($wrs_row > 248))
            {
               $logger->fatal_msg("WRS row must be between 1 and 248");
               return ERROR;
            }
        }
        else
        {
            $logger->fatal_msg("WRS row must be specified if WRS path is"
                . " specified");
            return ERROR;
        }
    }
    else
    {
        if ($wrs_row)
        {
            $logger->fatal_msg("WRS path must be specified if WRS row is"
                . " specified");
            return ERROR;
        }
    }

    return SUCCESS;
}

##############################################################################
#
# Name: parse_omf
#
# Purpose: Parse an OMF that contains parameters needed by DGC.
#
# Inputs and Outputs:
# I: OMF filename
# O: Populated global variables for the 4 lat/lon coordinates
# O: Populated global variables for the target projection and UTM zone
#    code
#
# Returns SUCCESS or ERROR
#
# Notes:  If the OMF file processed here was created when running the
#         script 'get_gcps_wrapper.pl', the parameters will be checked a
#         second time after they're read from the parameter file (in
#         'get_gcps_wrapper.pl', the parameters are validated after they're
#         read from the command line before they're written to the
#         parameter file).
#
##############################################################################
sub parse_omf($)
{
    my $utm_zone;

    # Get the parameter file name.
    my $omf_filename = shift;

    # Open the file.
    my $status = open (INPUTOMF, "< $omf_filename");
    if (!defined($status))
    {
        $logger->fatal_msg("Can't open input OMF file: $omf_filename");
        return ERROR;
    }

    # Process each line in the parameter file.
    while ($input_line = <INPUTOMF>)
    {
        if ($input_line =~ /^\s*UL_BOUNDARY_LAT_LON\s*=\s*(.*)/)
        {
            if ($1 =~ /\((-?\d{1,3}\.?\d{0,})\s*,\s*(-?\d{1,3}\.?\d{0,}\s*)\)/x)
            {
                $upper_left_lat = $1;
                $upper_left_lon = $2;
                $corners_present{UL} = 1;

            }
            else
            {
                $logger->fatal_msg("Invalid corner coordinate set format");
                return ERROR;
            }
        }

        if ($input_line =~ /^\s*UR_BOUNDARY_LAT_LON\s*=\s*(.*)/)
        {
            if ($1 =~ /\((-?\d{1,3}\.?\d{0,})\s*,\s*(-?\d{1,3}\.?\d{0,}\s*)\)/x)
            {
                $upper_right_lat = $1;
                $upper_right_lon = $2;
                $corners_present{UR} = 1;
            }
            else
            {
                $logger->fatal_msg("Invalid corner coordinate set format");
                return ERROR;
            }
        }
        if ($input_line =~ /^\s*LL_BOUNDARY_LAT_LON\s*=\s*(.*)/)
        {
            if ($1 =~ /\((-?\d{1,3}\.?\d{0,})\s*,\s*(-?\d{1,3}\.?\d{0,}\s*)\)/x)
            {
                $lower_left_lat = $1;
                $lower_left_lon = $2;
                $corners_present{LL} = 1;
            }
            else
            {
                $logger->fatal_msg("Invalid corner coordinate set format");
                return ERROR;
            }
        }
        if ($input_line =~ /^\s*LR_BOUNDARY_LAT_LON\s*=\s*(.*)/)
        {
            if ($1 =~ /\((-?\d{1,3}\.?\d{0,})\s*,\s*(-?\d{1,3}\.?\d{0,}\s*)\)/x)
            {
                $lower_right_lat = $1;
                $lower_right_lon = $2;
                $corners_present{LR} = 1;
            }
            else
            {
                $logger->fatal_msg("Invalid corner coordinate set format");
                return ERROR;
            }
        }

        # Target projection
        if ($input_line =~ /^\s*TARGET_PROJECTION\s*=\s*(.*)/)
        {
            if ($1 =~ /(\d+)/)
            {
                $target_projection = $1;
            }
            else
            {
                $logger->fatal_msg("Invalid target projection format");
                return ERROR;
            }
        }

        # UTM zone
        if ($input_line =~ /^\s*UTM_ZONE\s*=\s*(.*)/)
        {
            if ($1 =~ /(\d+)/)
            {
                $utm_zone = $1;
            }
            else
            {
                $logger->fatal_msg("Invalid UTM zone format");
                return ERROR;
            }
        }
    }

    # Close the OMF file
    $status = close (INPUTOMF);
    if (!$status)
    {
        $logger->fatal_msg("OMF file $omf_name did not close properly");
        return ERROR;
    }

    # Validate the target projection parameter if it was specified.
    # If it's set to 1 (representing UTM projection), validate the
    # UTM zone parameter.
    if (defined $target_projection)
    {
        # Check target projection for a value in the valid range.
        if (($target_projection < 0) || ($target_projection > 31))
        {
            $logger->fatal_msg("Invalid target projection parameter"
                . " $target_projection, should be between 0 and 31");
            return ERROR;
        }

        # Check the UTM zone if the target projection specifies UTM.
        if ($target_projection == 1)
        {
            if (!defined $utm_zone)
            {
                $logger->fatal_msg("OMF UTM zone code parameter is not"
                    . " defined");
                return ERROR;
            }
            else
            {
                if (($utm_zone < 1) || ($utm_zone > 60))
                {
                    $logger->fatal_msg("Specified UTM zone code $utm_zone is"
                        . " out of valid range");
                    return ERROR;
                }
            }
        }
    }

    return SUCCESS;
}

##############################################################################
#
# Name: update_omf
#
# Purpose: Update the omf with the GCP_DIRECTORY name
#
# Inputs and Outputs:
# I: OMF filename, GCP directory, Overall GCP version
# O: GCP_DIRECTORY parameter written to OMF
#
# Returns SUCCESS or ERROR
#
##############################################################################
sub update_omf
{
    my $omf_filename = shift;
    my $gcp_directory = shift;
    my $overall_gcp_version = shift;

    $status = open (INPUTOMF, "<$omf_filename");
    if (!defined($status))
    {
        $logger->fatal_msg("Can't open input OMF file $omf_filename"
            . " for reading");
        return ERROR;
    }
    my @lines = <INPUTOMF>;
    close (INPUTOMF);

    $status = open (OUTPUTOMF, "> $omf_filename");
    if (!defined($status))
    {
        $logger->fatal_msg("Can't open input OMF file $omf_filename"
            . " for writing");
        return ERROR;
    }
    foreach my $line (@lines)
    {
        # watch for end of parameters
        if ($line =~ /^END_OBJECT/)
        {
            if (! (print OUTPUTOMF "  GCP_DIRECTORY = \"$gcp_directory\"\n"))
            {
                $logger->fatal_msg("Cannot write GCP_DIRECTORY information"
                    . " to OMF file");
                close (OUTPUTOMF);
                return ERROR;
            }

            if (! (print OUTPUTOMF "  GROUND_CONTROL_POINTS_VERSION ="
                . " $overall_gcp_version\n"))
            {
                $logger->fatal_msg("Cannot write GROUND_CONTROL_POINTS_VERSION"
                    . " information to OMF file");
                close (OUTPUTOMF);
                return ERROR;
            }
        }

        # don't write out an old version of the GCP_DIRECTORY or
        # GROUND_CONTROL_POINTS_VERSION
        if ($line !~ /GCP_DIRECTORY/ 
            && $line !~ /GROUND_CONTROL_POINTS_VERSION/)
        {
            printf OUTPUTOMF $line;
        }
    }

    # Close the OMF file.
    $status = close (OUTPUTOMF);
    if (!$status)
    {
        $logger->fatal_msg("Output OMF file $omf_filename did not close"
            . " properly");
        return  ERROR;
    }
    return SUCCESS;
}

##############################################################################
# Name: get_gcp_version
#
# Purpose: Get the current gcp version from the database
#
# Parameters:
#   Database handle
#
# Returns:
#   Version number on SUCCESS
#   Undef on ERROR
#
##############################################################################
sub get_gcp_version
{
    my $dbh = shift;    # Database handle
    my $version_sth;    # Sql handle
    my $version;        # Version number

    my $sql = qq{SELECT MAX(VERSION) FROM $gcp_data_version_table};

    # Prepare the SQL query
    if (! ($version_sth = $dbh->prepare($sql)))
    {
        $logger->fatal_msg("Unable to prepare the GCP version query:"
            . " $DBI::errstr");
        return undef;
    }

    # Execute the query
    if (! ($version_sth->execute()))
    {
        $logger->fatal_msg("Unable to execute GCP version query:"
            . " $DBI::errstr");
        $version_sth->finish();
        return undef;
    }

    $version = $version_sth->fetchrow_array;
    
    $version_sth->finish();

    return $version;
}
