#!/usr/bin/perl -w

#############################################################################
#
# NAME:            get_gcps_wrapper.pl
#
# PURPOSE:         Allows user to specify processing parameter(s) on
#                  the command line.  The command line is parsed and
#                  the specified parameter(s) are written to temporary
#                  ODL-formatted parameter and OMF files.  Once the
#                  parameter files are generated, the script calls
#                  'get_gcps.pl'.  When processing is completed,
#                  the temporary parameter files are removed.
#
# RETURNS:         0 if processing completes successfully
#                  non-zero value if an error is encountered
#
# NOTES:           A parameter --help is provided to display a
#                  usage message.
#
############################################################################
use strict;
use Getopt::Long;                        # For GetOptions()
use File::Basename;
use File::Temp qw(tempfile);             # Function to make temporary files
use ias_const;                           # Defines IAS status parameters
                                         # SUCCESS and ERROR
use Logger;                              # Log messages */
use ias_processing_system;

# Declare program constant
use constant PROG => basename($0);

# Forward subroutine declarations
sub ShowUsage();                         # Prints help / usage message then
                                         # quits
sub ProcessOptions();                    # Processes command line options
sub WriteParameterFile($$);              # Writes parameters to ODL file
sub WriteOMFFile($);                     # Writes parameters to OMF file

# Coordinates for scene boundary
my $upper_left_lat;
my $upper_left_lon;
my $upper_right_lat;
my $upper_right_lon;
my $lower_right_lat;
my $lower_right_lon;
my $lower_left_lat;
my $lower_left_lon;

# Basic parameter file information
my $wo_directory = '.';        # Work order directory
my $work_order_id = '';        # Work order ID
my $gcp_directory = '';        # GCP directory
my $gcp_type = '';             # GCP type
my $gcp_source = '';           # GCP source
my $resolution;                # Chip resolution to retrieve
my $include_inactive;          # Flag to include the inactive GCPs and chips

# Other command-line query parameters
my @coordinates;               # 8 values (4 corner points)

# Additional parameter file information
my $start_date_string = '';  # Start date parameter string
my $start_date;              # Calculated begin date used for comparison
my $end_date_string = '';    # End date parameter string
my $end_date;                # Calculated end date used for comparison
my $wrs_path;                # Query ground_control_points table by path
my $wrs_row;                 # Query ground_control_points table by row
my $target_projection;
my $utm_zone;
my $compression = "NONE";    # Tarball compression type
my $write_tarball = 1;       # Write tarball option

# Miscellaneous temporary variables
my $status;

# Initialize the ias_processing_system module
if (ias_init_system() != SUCCESS)
{
    print STDERR PROG . "$ias_error_msg\n";
    exit FAILURE;
}

# Set up the logger
my $logger = Logger->new("STDOUT");
if (!$logger)
{
    print STDERR PROG . ": ERROR:  Failed to establish logger object for "
        . "STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name(PROG);
set_default_log($logger);

# If no options have been entered on the command line, quit
if (scalar @ARGV < 1)
{
    $logger->fatal_msg("No parameters specified on command line.  To display"
        . " the USAGE message, enter 'get_gcps_wrapper.pl --help' at the"
        . " command line");
    $logger->close_log();
    exit FAILURE;
}

# Process the command line options
$status = ProcessOptions();
if ($status != SUCCESS)
{
    $logger->fatal_msg("Command-line options could not be processed");
    $logger->close_log();
    exit FAILURE;
}

# Create temporary OMF file and ODL parameter files.  Set UNLINK option to
# force deletion of the temporary files when the script completes.

my $curr_dir = '.';
my ($omf_fh, $omf_tempname) = tempfile(DIR => $curr_dir, SUFFIX => ".omf",
    UNLINK => 1);

# Create the OMF file first.
$status = WriteOMFFile($omf_fh);
if ($status != SUCCESS)
{
    $logger->fatal_msg("Temporary OMF parameter file could not be generated");
    $logger->close_log();
    close ($omf_fh);
    exit FAILURE;
}

# Close the OMF file.  Check to make sure it closed successfully.
$status = close ($omf_fh);
if (!$status)
{
    $logger->fatal_msg("Error closing temporary OMF parameter file");
    $logger->close_log();
    exit FAILURE;
}

# Create the processing parameter file.  Extract the name of the OMF
# file to use as the work order ID.
my ($odl_fh, $odl_tempname) = tempfile(DIR => $curr_dir, SUFFIX => ".odl",
    UNLINK => 1);

$status = WriteParameterFile($odl_fh, fileparse($omf_tempname, ".omf"));
if ($status != SUCCESS)
{
    $logger->fatal_msg("Temporary parameter file could not be created");
    $logger->close_log();
    exit FAILURE;
}

# Close the parameter file.  Check to make sure it closed successfully.
$status = close ($odl_fh);
if (!$status)
{
    $logger->fatal_msg("Error closing temporary parameter file");
    $logger->close_log();
    exit FAILURE;
}

# Now that we have parameter files to work with, call 'get_gcps.pl'
my $rc = system("get_gcps.pl $odl_tempname");
if ($rc != SUCCESS)
{
    $logger->fatal_msg("Error retrieving GCP chips");
    $logger->close_log();
    exit FAILURE;
}

# Cleanup.
$logger->close_log();

exit SUCCESS;


##############################################################################
# Subroutines below
##############################################################################


##############################################################################
#
# NAME:     WriteParameterFile
#
# PURPOSE:  Takes the appropriate command-line parameters read from
#           ProcessOptions() and writes them to an ODL parameter file.
#
# RETURNS:  Integer status code of SUCCESS or ERROR
#
##############################################################################
sub WriteParameterFile($$)
{
    # Get the ODL parameter file handle from the argument list.
    my $file_handle = shift;

    # Get the temporary work order ID string from the argument list.
    my $temp_wo_id = shift;

    # Construct the file contents as a string.
    my $param_file_string = "OBJECT = PARAMETERS\n";
    if ($wo_directory)
    {
        $param_file_string .= "  WO_DIRECTORY = \"$wo_directory\"\n";
    }
    if ($temp_wo_id)
    {
        $param_file_string .= "  WORK_ORDER_ID = \"$temp_wo_id\"\n";
    }
    if ($gcp_directory)
    {
        $param_file_string .= "  GCP_DIRECTORY = \"$gcp_directory\"\n";
    }
    if ($gcp_type)
    {
        $param_file_string .= "  GCP_TYPE = \"$gcp_type\"\n";
    }
    $param_file_string .= "  GCP_SOURCE = \"$gcp_source\"\n";

    if ($start_date_string)
    {
        $param_file_string .= "  START_DATE = \"$start_date_string\"\n";
    }
    if ($end_date_string)
    {
        $param_file_string .= "  END_DATE = \"$end_date_string\"\n";
    }
    if ($wrs_path)
    {
        $param_file_string .= "  WRS_PATH = $wrs_path\n";
        $param_file_string .= "  WRS_ROW = $wrs_row\n";
    }
    if (defined $resolution)
    {
        $param_file_string .= "  RESOLUTION = ($resolution, $resolution)\n";
    }
    if (defined $include_inactive)
    {
        $param_file_string .= "  INCLUDE_INACTIVE = 1\n";
    }
    if (defined $write_tarball)
    {
        $param_file_string .= "  PACKAGE_CHIPS = $write_tarball\n";
    }
    if (defined $compression)
    {
        $param_file_string .= "  PACKAGE_COMPRESSION = \"$compression\"\n";
    }

    # Add the closing information.
    $param_file_string .= "END_OBJECT = PARAMETERS\nEND\n";

    # Write the file contents.
    if (! (print $file_handle $param_file_string))
    {
        $logger->fatal_msg("Cannot write contents to the temporary"
            . " parameter file");
        return ERROR;
    }

    return SUCCESS;
}

##############################################################################
#
# NAME:     WriteOMFFile
#
# PURPOSE:  Takes the appropriate command-line parameters read from
#           ProcessOptions() and writes them to an OMF parameter file.
#
# RETURNS:  Integer status code of SUCCESS or ERROR
#
##############################################################################
sub WriteOMFFile($)
{
    # Get the OMF parameter file name from the argument list
    my $file_handle = shift;

    # Write all the statements to go into the file.
    my $param_file_string = "OBJECT = OMF\n  SCRIPT_NAME = OMF\n";

    if (($upper_left_lat && $upper_left_lon)
        && ($upper_right_lat && $upper_right_lon)
        && ($lower_left_lat && $lower_left_lon)
        && ($lower_right_lat && $lower_right_lon))
    {
        $param_file_string .= "  UL_BOUNDARY_LAT_LON = ($upper_left_lat, "
            . "$upper_left_lon)\n";
        $param_file_string .= "  UR_BOUNDARY_LAT_LON = ($upper_right_lat, "
            . "$upper_right_lon)\n";
        $param_file_string .= "  LL_BOUNDARY_LAT_LON = ($lower_left_lat, "
            . "$lower_left_lon)\n";
        $param_file_string .= "  LR_BOUNDARY_LAT_LON = ($lower_right_lat, "
            . "$lower_right_lon)\n";
    }

    # Add the target projection information if it's available.
    if ($target_projection)
    {
        $param_file_string .= "  TARGET_PROJECTION = $target_projection\n";
    }

    # Add the UTM zone information if it's available.
    if ($utm_zone)
    {
        $param_file_string .= "  UTM_ZONE = $utm_zone\n";
    }

    # Add the closing information.
    $param_file_string .= "END_OBJECT = OMF\nEND\n";

    # Write the contents to the file.
    if (! (print $file_handle $param_file_string))
    {
        $logger->fatal_msg("Cannot write contents to the temporary OMF file");
        return ERROR;
    }

    return SUCCESS;
}

##############################################################################
#
# NAME:     ProcessOptions()
#
# PURPOSE:  Using the GetOptions call in the Getopt::Long module, this routine
#           parses the argument list, flagging any options found in the
#           parameter list.  If an invalid option is given or the help option
#           is requested, a usage message is shown and the program exits
#           immediately.
#
# RETURNS:  SUCCESS status code if parameter processing succeeded
#           ERROR   status code if the GetOptions call fails 
#
# NOTES:    Mandatory parameters are noted in the GetOptions call as
#           <parm_name>=<parm_type> => \${variable_name}.  Optional
#           parameters are noted in the call as
#           <parm_name>:<parm_type> => \${variable_name}
#
##############################################################################
sub ProcessOptions()
{
    my $help = '';
    my $origCnt = scalar @ARGV;
    my $result = GetOptions(
            "help" => \$help,
            "gcp_source=s" => \$gcp_source,
            "target:i" => \$target_projection,
            "corners:f{8,8}" => \@coordinates,
            "gcp_type:s" => \$gcp_type,
            "output:s" => \$gcp_directory,
            "start_date:s" => \$start_date_string,
            "end_date:s" => \$end_date_string,
            "wrs_path:i" => \$wrs_path,
            "wrs_row:i" => \$wrs_row,
            "resolution:i" => \$resolution,
            "utm_zone:i" => \$utm_zone,
            "include_inactive" => \$include_inactive,
            "write_tarball:i" => \$write_tarball,
            "compression=s" => \$compression
            );
    if (!$result)
    {
        $logger->fatal_msg("Failed to read command line parameters");
        return ERROR;
    }

    # Display help/usage information if requested, then quit.
    if ($help)
    {
        ShowUsage();
        exit SUCCESS;
    }

    # Validate required parameters

    # Check that compression is valid
    $compression = uc($compression);
    if ($compression ne "NONE" && $compression ne "GZIP"
        && $compression ne "BZIP")
    {
        $logger->fatal_msg("Compression must be specified as NONE, BZIP "
            . "or GZIP");
        return ERROR;
    }

    # Check that write tarball is valid
    if ($write_tarball != 0 && $write_tarball != 1)
    {
        $logger->fatal_msg("Write tarball must be either 0 or 1");
        return ERROR;
    }

    # GCP source
    if ($gcp_source eq '')
    {
        $logger->fatal_msg("GCP source parameter not specified");
        return ERROR;
    }
    if (not (($gcp_source eq 'GLS') || ($gcp_source eq 'DOQ')
        || ($gcp_source eq 'TM6') || ($gcp_source eq 'GRI') ||
        ($gcp_source eq 'ALL')))
    {
        $logger->fatal_msg("GCP source must be 'GLS', 'DOQ', 'TM6', 'GRI' "
                         . "or 'ALL'");
        return ERROR;
    }

    # GCP output directory
    if ($gcp_directory eq '')
    {
        $gcp_directory = "output";
    }

    # Target projection parameter.  If specified, check for values in the
    # range 0 < $target_projection < 31 (inclusive), in order to allow for
    # possible expanded use in the future.
    if (defined $target_projection)
    {
        if (($target_projection < 0) || ($target_projection > 31))
        {
            $logger->fatal_msg("If specified, target projection must be a value"
                . " between 0 and 31");
            return ERROR;
        }

        # If the target projection was specified as UTM (value == 1),
        # validate the UTM zone parameter.
        if ($target_projection == 1)
        {
            if (!defined $utm_zone)
            {
                $logger->fatal_msg("OMF UTM zone code parameter is not"
                    . " defined");
                return ERROR;
            }
            else
            {
                if (($utm_zone < 1) || ($utm_zone > 60))
                {
                    $logger->fatal_msg("Specified UTM zone code $utm_zone is"
                        . " out of valid range");
                    return ERROR;
                }
            }
        }
    }

    # Validate optional parameters if they have been defined, otherwise,
    # set their default values

    # Unload the corner coordinates
    if (@coordinates)
    {
       ($upper_left_lat, $upper_left_lon, $upper_right_lat, $upper_right_lon,
        $lower_left_lat, $lower_left_lon, $lower_right_lat,
        $lower_right_lon) = @coordinates;
    }

    # GCP type
    if ($gcp_type)
    {
        if (not (($gcp_type eq 'BOTH') || ($gcp_type eq 'CONTROL')
            || ($gcp_type eq 'VALIDATION')))
        {
            $logger->fatal_msg("Invalid GCP type $gcp_type specified");
            return ERROR;
        }
    }
    else
    {
        $gcp_type = 'BOTH';
    }

    # GCP directory
    $gcp_directory = 'output' if ($gcp_directory eq '');

    # WRS path and row.  Both must be specified on the command line.
    # If only one is specified, consider it an error
    if ($wrs_path)
    {
        if (($wrs_path < 1) || ($wrs_path > 233))
        {
            $logger->fatal_msg("Path must be between 1 and 233");
            return ERROR;
        }

        if ($wrs_row)
        {
            if (($wrs_row < 1) || ($wrs_row > 248))
            {
                $logger->fatal_msg("Row must be between 1 and 248");
                return ERROR;
            }
        }
        else
        {
            $logger->fatal_msg("Both path and row must be specified");
            return ERROR;
        }
    }
    else
    {
        if ($wrs_row)
        {
            $logger->fatal_msg("Both path and row must be specified");
            return ERROR;
        }
    }

    # The start and end acquisition dates used to filter the image chips.  The
    # expected format is 'YYYY-MM-DD' (where a leading 0 is required for
    # single-digit month/day numbers); the time is assumed to be midnight
    # (i.e. 00:00:00).  If the format is valid, calculate the corresponding
    # date from the date string.  If both dates are calculated, check to see if
    # the end date occurs after the start date; if it doesn't, consider it an
    # an error.
    #
    # Use 'defined' and 'length' to establish that i) the date string
    # has been defined, and ii) that it's not empty ('').
    if (defined $start_date_string and length($start_date_string))
    {
        if ($start_date_string =~ /^(\d{4})\-(\d{2})\-(\d{2})/)
        {
            my @start_date_components = ($1, $2, $3);
            $start_date = $start_date_components[0] * 10000
                + $start_date_components[1] * 100 + $start_date_components[2];
        }
        else
        {
            $logger->fatal_msg("Start date $start_date_string is not in the"
                . " required format 'YYYY-MM-DD'");
            return ERROR;
        }
    }
    if (defined $end_date_string and length($end_date_string))
    {
        if ($end_date_string =~ /^(\d{4})\-(\d{2})\-(\d{2})/)
        {
            my @end_date_components = ($1, $2, $3);
            $end_date = $end_date_components[0] * 10000
                + $end_date_components[1] * 100 + $end_date_components[2];
        }
        else
        {
            $logger->fatal_msg("End date $end_date_string is not in the"
                . " required format 'YYYY-MM-DD'");
            return ERROR;
        }
    }
    if (defined $start_date && defined $end_date)
    {
        if ($end_date <= $start_date)
        {
            $logger->fatal_msg("End date $end_date occurs before/on the start"
                . " date $start_date");
            return ERROR;
        }
    }

    return SUCCESS;
}

##############################################################################
#
# NAME:    ShowUsage()
#
# PURPOSE: Shows the help or usage directives for the program.
#
# RETURNS: Nothing
#
##############################################################################
sub ShowUsage()
{
    my $PROGRAM = PROG;

    print "\n$PROGRAM retrieves parameters from the command line, builds "
        . "the parameter files needed to run 'get_gcps.pl', then runs "
        . "'get_gcps.pl'.\n\n";
    print "$PROGRAM command line options:\n\n";
    print "--help:      Displays this help summary.\n";
    print "--gcp_source:Required parameter specifying the GCP data source.\n"
        . "             Valid values are 'GLS', 'DOQ', 'TM6', 'GRI' or\n"
        . "             'ALL'.\n";
    print "--target:    Optional parameter indicating the target projection.\n"
        . "             Valid values are between 0 and 31.  Currently\n"
        . "             defined values are 1 (UTM), 6 (PS), or 22 (SOM).\n";
    print "--corners:   Optional parameter to specify a 'bounding box' based\n"
        . "             on corner lat/lon coordinates.  These coordinates\n"
        . "             are specified in the following order, separated by\n"
        . "             spaces:  upper left lat, upper left lon, upper right\n"
        . "             lat, upper right lon, lower left lat, lower left\n"
        . "             lon, lower right lat, lower right lon.\n";
    print "--gcp_type:  Optional parameter specifying the GCP usage type.\n"
        . "             Valid values are 'CONTROL', 'VALIDATION', or 'BOTH'.\n"
        . "             The default value is 'BOTH'.\n";
    print "--wrs_path:  Optional parameter to filter GCP's to a single WRS\n"
        . "             path.  Valid values are between 1 and 233.  If\n"
        . "             --wrs_path is specified, --wrs_row must also be\n"
        . "             specified.\n";
    print "--wrs_row:   Optional parameter to filter GCP's to a single WRS\n"
        . "             row.  Valid values are between 1 and 248.  If\n"
        . "             --wrs_row is specified, --wrs_path must also be\n"
        . "             specified.\n";
    print "--start_date:Optional parameter specifying the acquisition date\n"
        . "             after which the GCP image chip data were acquired.\n"
        . "             The expected format is 'YYYY-MM-DD'.\n";
    print "--end_date:  Optional parameter specifying the acquisition date\n"
        . "             before which the GCP image chip data were\n"
        . "             acquired.  The expected format is 'YYYY-MM-DD'.\n"
        . "             If --start_date is also specified, --end_date must\n"
        . "             occur later in time.\n";
    print "--resolution:Optional parameter specifying the resolution\n"
        . "             Valid values are 15, 30 and 60 meter chips.\n";
    print "--output:    Optional parameter specifying the location of the\n"
        . "             output directory where the retrieved image chips are\n"
        . "             stored.  If not specified, the value defaults to\n"
        . "             'output' in the directory in which this script is\n"
        . "             run.\n";
    print "--utm_zone:  Optional parameter indicating the UTM zone code of\n"
        . "             the the GCP image chip(s).  Valid values\n"
        . "             are between 1 and 60.\n";
    print "--include_inactive: Optional parameter indicating the output\n"
        . "             should include inactive GCPs and chips instead of\n"
        . "             filtering them out.\n";
    print "--write_tarball: Optional parameter indication whether image chips\n"
        . "             should be written as a single tarball (option 1)\n"
        . "             or written as individual files (option 0). Valid\n"
        . "             options are (1 or 0) with default being 1.\n";
    print "--compression: Optional parameter used when writing out image \n"
        . "             chips as a tarball. Valid values are NONE, BZIP, and \n"
        . "             GZIP.  The default is NONE. \n\n"; 
}
