/*****************************************************************************
NAME: resample_chip

PURPOSE:
Resample GCP chip

RETURN VALUE:
Type = int

Value    Description
-----    -----------
SUCCESS  output GCP written
ERROR    error generating statistics

*******************************************************************************/
#include <string.h>
#include <limits.h>
#include "gctp.h"              /* for UTM zone code */
#include "ias_geo.h"
#include "ias_logging.h"
#include "ias_gcp.h"
#include "GCPresample.h"

int resample_chip
(
    IAS_GCP_IMAGE_CHIP_IO *in_chip_io, /* I: Input image chip container */
    IAS_GCP_IMAGE_CHIP_IO *out_chip_io,/* I: Output image chip container */
    int izone,              /* I: Input UTM Zone */
    int icode,              /* I: Input projection code */
    double *iparm,          /* I: Input projection parameters */
    int ispheroid,          /* I: Input spheroid code */
    int iunit,              /* I: Input projection units */
    int ocode,              /* I: Input projection code */
    int ozone,              /* I: Output UTM Zone */
    double *oparm,          /* I: Output projection parameters */
    int ospheroid,          /* I: Output spheroid code */
    int ounit,              /* I: Output projection units */
    IAS_GCP_RECORD *gcp_lib /* I/O: GCP libarary */
)
{
    int ii, jj;                /* Counters */
    double oproj_x;            /* Output projection x */
    double oproj_y;            /* Output projection y */
    double ul_x_gcp;           /* Upper left x of GCP */
    double ul_y_gcp;           /* Upper left y of GCP */
    double ul_x_new;           /* New Upper left x of GCP */
    double ul_y_new;           /* New Upper left y of GCP */
    IAS_R4_TYPE *ibuf;         /* Input resampled buffer */
    void *obuf_max;            /* Max sized output resampled buffer */
    IAS_BYTE_TYPE *obuf_byte;  /* Byte output buffer pointer */
    IAS_I2_TYPE *obuf_i2;      /* Signed 16-bit output buffer pointer */
    IAS_UI2_TYPE *obuf_ui2;    /* Unsigned 16-bit output buffer pointer */
    IAS_R4_TYPE *obuf_r4;      /* 4-byte floating point output buffer pointer */
    int line_size;             /* Chip line size */
    int samp_size;             /* Chip sample size */
    int size;                  /* GCP chip size (line_size x samp_size) */
    IAS_PROJECTION input_proj;
    IAS_PROJECTION output_proj;
    IAS_GEO_PROJ_TRANSFORMATION *in2out;
    IAS_GEO_PROJ_TRANSFORMATION *out2in;

    /* Set chip size, makes code easier to read */
    line_size = gcp_lib->chip_size_lines;
    samp_size = gcp_lib->chip_size_samples;
    size = line_size * samp_size;

    /* Input buffer needs to be floating point (R*4) for reading */
    ibuf = (IAS_R4_TYPE *)malloc(size * sizeof(IAS_R4_TYPE));
    if(ibuf == NULL)
    {
        IAS_LOG_ERROR("Allocating array");
        return ERROR;
    }

    /* Allocate the max size (R*4) output resampled buffer */
    obuf_max = malloc(size * sizeof(IAS_R4_TYPE));
    if(obuf_max == NULL)
    {
        IAS_LOG_ERROR("Allocating array");
        free(ibuf);
        return ERROR;
    }

    /* Set all the different data type buffer pointers to point at the
       max sized data buffer */
    obuf_byte = obuf_max;
    obuf_i2 = obuf_max;
    obuf_ui2 = obuf_max;
    obuf_r4 = obuf_max;

    /* Read GCP chip */
    if (ias_gcp_read_image_chip(in_chip_io, gcp_lib->chip_name, line_size,
            samp_size, gcp_lib->chip_data_type, IAS_GCP_CONVERT_TO_FLOAT, ibuf)
            != SUCCESS)
    {
        IAS_LOG_ERROR("Reading GCP image %s", gcp_lib->chip_name);
        free(ibuf);
        free(obuf_max);
        return ERROR;
    }

    /* Calculate current upper left projection coordinate of chip */
    ul_x_gcp = gcp_lib->projection_x - gcp_lib->reference_sample
        * gcp_lib->pixel_size_x;
    ul_y_gcp = gcp_lib->projection_y + gcp_lib->reference_line
        * gcp_lib->pixel_size_y;

    /* Build the two projection objects involved */
    ias_geo_set_projection(icode, izone, iunit, ispheroid, iparm, &input_proj);
    ias_geo_set_projection(ocode, ozone, ounit, ospheroid, oparm, &output_proj);

    /* Create the transformation from input to output */
    in2out = ias_geo_create_proj_transformation(&input_proj, &output_proj);
    if (!in2out)
    {
        IAS_LOG_ERROR("Creating input to output projection transformation");
        free(ibuf);
        free(obuf_max);
        return ERROR;
    }

    /* Find new coordinate for GCP location */
    if (ias_geo_transform_coordinate(in2out, gcp_lib->projection_x,
            gcp_lib->projection_y, &oproj_x, &oproj_y) != SUCCESS)
    {
        IAS_LOG_ERROR("Transforming GCP coordinate to target projection");
        ias_geo_destroy_proj_transformation(in2out);
        free(ibuf);
        free(obuf_max);
        return ERROR;
    }
    gcp_lib->projection_y = oproj_y;
    gcp_lib->projection_x = oproj_x;

    ias_geo_destroy_proj_transformation(in2out);
    in2out = NULL;

    /* Create the transformation from output to input */
    out2in = ias_geo_create_proj_transformation(&output_proj, &input_proj);
    if (!out2in)
    {
        IAS_LOG_ERROR("Creating output to input projection transformation");
        free(ibuf);
        free(obuf_max);
        return ERROR;
    }

    /* Calculate new 'floating point' upper left */
    ul_x_new = gcp_lib->projection_x - gcp_lib->reference_sample
        * gcp_lib->pixel_size_x;
    ul_y_new = gcp_lib->projection_y + gcp_lib->reference_line
        * gcp_lib->pixel_size_y;

    /* Set chip upper left coordinate to whole pixel size */
    ul_x_new = round(ul_x_new / gcp_lib->pixel_size_x) * gcp_lib->pixel_size_x;
    ul_y_new = round(ul_y_new / gcp_lib->pixel_size_y) * gcp_lib->pixel_size_y;

    /* Find new line sample location of GCP location */
    gcp_lib->reference_sample = (oproj_x - ul_x_new) / gcp_lib->pixel_size_x;
    gcp_lib->reference_line = (ul_y_new - oproj_y) / gcp_lib->pixel_size_y;

    /* Resample chip */
    for (ii = 0; ii < line_size; ii++)
    {
        /* Calculate the Y projection coordinate */
        double proj_y = ul_y_new - ii * gcp_lib->pixel_size_y;

        for (jj = 0; jj < samp_size; jj++)
        {
            int lindex;     /* Line index to GCP chip */
            int sindex;     /* Sample index to GCP chip */
            double oline;   /* New line point location in original chip */
            double osamp;   /* New sample point location in original chip */
            double dn;      /* Resampled DN value */
            double proj_x;  /* Input project X location */
            int output_index; /* Ouput buffer index */
 
            /* Calculate the projection X coordinate */
            proj_x = ul_x_new + jj * gcp_lib->pixel_size_x; 

            /* Transform the coordinates to the input projection */
            if (ias_geo_transform_coordinate(out2in, proj_x, proj_y,
                    &oproj_x, &oproj_y) != SUCCESS)
            {
                IAS_LOG_ERROR("Transforming coordinate");
                ias_geo_destroy_proj_transformation(out2in);
                free(ibuf);
                free(obuf_max);
                return ERROR;
            }

            /* Calculate line,sample location in GCP chip */
            oline = (ul_y_gcp - oproj_y) / gcp_lib->pixel_size_y;
            osamp = (oproj_x - ul_x_gcp) / gcp_lib->pixel_size_x;

            /* Find integer location in original GCP */
            lindex = (int)floor(oline);
            sindex = (int)floor(osamp);

            /* If the point is within the chip boundaries then resample this 
               point using bilinear interpolation */
            if ((lindex >= 0) && (lindex < line_size - 1) && (sindex >= 0)
                    && (sindex < samp_size - 1))
            {
                /* Find floating point location in original GCP */
                double fline = oline - lindex;
                double fsamp = osamp - sindex;

                /* Calculate the DN using bilinear interpolation */
                dn = (1.0 - fline) * (1.0 - fsamp)
                    * ibuf[lindex * samp_size + sindex]
                    + fline * (1.0 - fsamp)
                    * ibuf[(lindex + 1) * samp_size + sindex]
                    + (1.0 - fline) * fsamp
                    * ibuf[lindex * samp_size + sindex + 1]
                    + fline * fsamp
                    * ibuf[(lindex + 1) * samp_size + sindex + 1];
            }
            else
                dn = 0.0;

            /* Calculate the output buffer index */
            output_index = ii * samp_size + jj;

            /* Set output sample dn. Simply truncating the data is sufficient
               (instead of the more correct rounding) */
            if (gcp_lib->chip_data_type == IAS_BYTE)
            {
                if (dn < 0.0)
                    obuf_byte[output_index] = 0;
                else if (dn > (double)UCHAR_MAX)
                    obuf_byte[output_index] = UCHAR_MAX;
                else
                    obuf_byte[output_index] = dn;
            }
            else if (gcp_lib->chip_data_type == IAS_I2)
            {
                if (dn < (double)SHRT_MIN)
                    obuf_i2[output_index] = SHRT_MIN;
                else if (dn > (double)SHRT_MAX)
                    obuf_i2[output_index] = SHRT_MAX;
                else
                    obuf_i2[output_index] = dn;
            }
            else if (gcp_lib->chip_data_type == IAS_UI2)
            {
                if (dn < 0.0)
                    obuf_ui2[output_index] = 0;
                else if (dn > (double)USHRT_MAX)
                    obuf_ui2[output_index] = USHRT_MAX;
                else
                    obuf_ui2[output_index] = dn;
            }
            else if (gcp_lib->chip_data_type == IAS_R4)
            {
                obuf_r4[output_index] = dn;
            }
            else
            {
                IAS_LOG_ERROR("Unsupported GCP chip data type %d",
                        gcp_lib->chip_data_type);
                ias_geo_destroy_proj_transformation(out2in);
                free(ibuf);
                free(obuf_max);
                return ERROR;
            }
        }
    }

    ias_geo_destroy_proj_transformation(out2in);
    out2in = NULL;

    /* Resampled data is now in the output buffer, so can free input buffer */
    free(ibuf);
    ibuf = NULL;

    /* Write out the resampled data to the image chip */
    if (ias_gcp_write_image_chip(out_chip_io, gcp_lib->chip_name,
            line_size, samp_size, gcp_lib->chip_data_type, obuf_max) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing resampled image chip %s", gcp_lib->chip_name);
        free(obuf_max);
        return ERROR;
    }

    free(obuf_max);

    return SUCCESS;
}
