/******************************************************************************
NAME: resample_gcps

PURPOSE:
Resample the GCPs from the custom GCP library to the UTM zone of the L1G image

RETURN VALUE:
Type = int
Value    Description
-----    -----------
ERROR    Error reading GCP information
SUCCESS  GCPs read

******************************************************************************/
#include <stdlib.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include "gctp.h"               /* for UTM definition */
#include "ias_logging.h"
#include "ias_geo.h"            /* for WGS84_SPHEROID definition */
#include "ias_gcp.h"
#include "GCPresample.h"

/* right-shift macro for system() calls */
#define RIGHT_SHIFT(status) (((status)>>8) & 0xff)

int resample_gcps
(
    char *img_name,            /* I: Name of 1g image */
    char *gcp_dir,             /* I: Directory containing GCP data */
    char *projection_source,   /* I: Source of projection (OMF or L1G) */
    int target_projection,     /* I: Target projection from OMF */
    int utm_zone,              /* I: UTM zone from OMF */
    int use_tarball            /* I: Flag to write chips to a tarball */
)
{
    int ii;                      /* Loop counter */
    IAS_GCP_RECORD *tmp_lib=NULL; /* Pointer for chip information */
    int tnumgcp;                 /* Number of chips in tmp_lib */
    char sys_cmd[PATH_MAX];      /* System command for mv & rm */
    double iparms[IAS_PROJ_PARAM_SIZE];/* Image projection parameters */
    int icode;                   /* L1G projection code */
    int ispheroid;               /* L1G scheroid */
    int iunit;                   /* L1G units */
    int izone;                   /* L1G UTM zone */
    double gparms[IAS_PROJ_PARAM_SIZE];/* L1G projection parameters */
    int gcode;                   /* Chip projection code */
    int gspheroid;               /* Chip spheroid */
    int gzone;                   /* Chip UTM zone */
    char gcplib_filename[PATH_MAX]; /* Input GCPLib file */
    int dir_mode = S_IRWXU;      /* Temp working directory mode */
    int chip_resampled_flag;     /* Flag to that chips have been resampled */
    L1GIO *img;                  /* L1G image */
    int status;                  /* Status returned from routines */
    IAS_L1G_FILE_METADATA fmd;   /* L1G file metadata */
    int overall_gcp_version;     /* Overall GCP version number */
    IAS_GCP_IMAGE_CHIP_IO *in_chip_io; /* Input image chip container */
    IAS_GCP_IMAGE_CHIP_IO *out_chip_io = NULL;/* Output image chip container */
    IAS_GCP_CONTAINER_TYPE container_type; /* Output container type */
    char out_container_name[PATH_MAX];/* Resampled chip container name */
    char in_container_name[PATH_MAX]; /* Input chip container name */
    void *copy_chip_buffer = NULL;/* Buffer used when copying chips between
                                     containers */
    int copy_chip_buffer_size = 0;/* Size of the copy_chip_buffer */

    /* If projection_source is L1G */
    status = strcmp(projection_source, "L1G");
    if (status == 0)
    {
        IAS_LOG_INFO("Using L1G metadata for parameters");

        /* Get metadata from L1G image */
        img = ias_l1g_open_image(img_name, IAS_READ);
        if (img == NULL)
        {
            IAS_LOG_ERROR("Opening L1G file");
            return ERROR;
        }
        status = ias_l1g_get_file_metadata(img, &fmd);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Getting L1G metadata");
            return ERROR;
        }
        status = ias_l1g_close_image(img);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Closing L1G file");
            return ERROR;
        }

        /* Get the target UTM zone, and params needed for reprojection */
        izone = fmd.zone_code;
        ispheroid = fmd.spheroid_code;
        gspheroid = ispheroid;

        for (ii = 0; ii < IAS_PROJ_PARAM_SIZE; ii++)
        {
            iparms[ii] = fmd.projection_parameters[ii];
            gparms[ii] = fmd.projection_parameters[ii];
        }

        if (fmd.projection_code != UTM)
        {
            IAS_LOG_ERROR("Should only run this version of GCPRESAMPLE on UTM");
            return ERROR;
        }

        status = ias_geo_get_units((char *)fmd.projection_units, &iunit);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Illegal unit from 1G image metadata");
            return ERROR;
        }

        icode = fmd.projection_code;
        gcode = fmd.projection_code;
    }
    else /* If projection_source is OMF or empty, use OMF */
    {
        /* Assign UTM zone from OMF */
        IAS_LOG_INFO("Using OMF metadata for parameters");
        izone = utm_zone;
        ispheroid = gspheroid = WGS84_SPHEROID;

        for (ii = 0; ii < IAS_PROJ_PARAM_SIZE; ii++)
        {
            iparms[ii] = 0;
            gparms[ii] = 0;
        }

        if (target_projection != UTM)
        {
            IAS_LOG_ERROR("Should only run this version of GCPRESAMPLE on UTM");
            return ERROR;
        }

        status = ias_geo_get_units((char *)"METERS", &iunit);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Illegal unit from OMF metadata");
            return ERROR;
        }

        icode = target_projection;
        gcode = target_projection;
    }

    /* Construct the input GCPLib file path */
    sprintf(gcplib_filename, "%s/%s", gcp_dir, GCPLIB_NAME);

    /* Read the GCPLib file and extract needed information */
    status = ias_gcp_read_gcplib(gcplib_filename, &tmp_lib, &tnumgcp, 
        &overall_gcp_version);
    if (status != SUCCESS)
    {
        free(tmp_lib);
        IAS_LOG_ERROR("Reading GCP library file %s", gcplib_filename);
        return ERROR;
    }

    /* Prepare to create the correct type of output image chip container */
    if (!use_tarball)
    {
        /* Not using a tarball, so create a temporary directory */
        container_type = IAS_GCP_CONTAINER_DIR;

        /* Make the temporary resampled GCP directory */
        status = snprintf(out_container_name, sizeof(out_container_name),
            "%s/%s", gcp_dir, TEMP_GCP_DIR);
        if (status <= 0 || status >= sizeof(out_container_name))
        {
            IAS_LOG_ERROR("Container name buffer is not large enough");
            free(tmp_lib);
            return ERROR;
        }

        status = mkdir(out_container_name, dir_mode);
        if (status != SUCCESS)
        {
            if (errno != EEXIST)
            {
                free(tmp_lib);
                IAS_LOG_ERROR("Unable to make temporary resampled GCP path %s",
                    out_container_name);
                return ERROR;
            }
        }

        strcpy(in_container_name, gcp_dir);
    }
    else
    {
        /* Using a tarball, so prepare to create that */
        container_type = IAS_GCP_CONTAINER_TARBALL;

        status = snprintf(out_container_name, sizeof(out_container_name),
            "%s/resamp_%s.tar", gcp_dir, IAS_GCP_IMAGE_CHIP_FILENAME);
        if (status <= 0 || status >= sizeof(out_container_name))
        {
            IAS_LOG_ERROR("Container name buffer is not large enough");
            free(tmp_lib);
            return ERROR;
        }

        status = snprintf(in_container_name, sizeof(in_container_name),
            "%s/%s.tar", gcp_dir, IAS_GCP_IMAGE_CHIP_FILENAME);
        if (status <= 0 || status >= sizeof(in_container_name))
        {
            IAS_LOG_ERROR("Container name buffer is not large enough");
            free(tmp_lib);
            return ERROR;
        }
    }

    /* Create the output image chip container */
    out_chip_io = ias_gcp_open_image_chip_container(out_container_name, 
        container_type, IAS_WRITE);
    if (!out_chip_io)
    {
        IAS_LOG_ERROR("Failed to open ouput image chip container: %s",
            out_container_name);
        free(tmp_lib);
        return ERROR;
    }


    /* Open the input image chip container */
    in_chip_io = ias_gcp_open_image_chip_container(in_container_name,
        container_type, IAS_READ);
    if (!in_chip_io)
    {
        IAS_LOG_ERROR("Failed to open input image chip container: %s",
            gcp_dir);
        ias_gcp_close_image_chip_container(out_chip_io);
        free(tmp_lib);
        return ERROR;
    }

    /* Run through the GCPs one by one and resample if zones do not match */
    chip_resampled_flag = 0;
    for (ii = 0; ii < tnumgcp; ii++)
    {
        /* Do a quick sanity check to make sure no GCPs with unsupported
           data types got into the library. They should have been identified
           and caused an error when the GCPs were retrieved from the database,
           but if GCPresample is being run on a library from another source,
           we should check the types here too to catch any unsupported chips
           as soon as possible. */
        if ((tmp_lib[ii].chip_data_type != IAS_BYTE)
                && (tmp_lib[ii].chip_data_type != IAS_I2)
                && (tmp_lib[ii].chip_data_type != IAS_UI2)
                && (tmp_lib[ii].chip_data_type != IAS_R4))
        {
            IAS_LOG_ERROR("Unsupported chip data type: %d",
                    tmp_lib[ii].chip_data_type);
            ias_gcp_close_image_chip_container(in_chip_io);
            ias_gcp_close_image_chip_container(out_chip_io);
            free(copy_chip_buffer);
            free(tmp_lib);
            return ERROR;
        }

        /* If the zone of the L1G doesn't match that of the GCP, resample */
        if (izone != tmp_lib[ii].zone)
        {
            chip_resampled_flag = 1;

            IAS_LOG_INFO("Zones mismatch, resampling %s from %d to %d",
                    tmp_lib[ii].chip_name, tmp_lib[ii].zone, izone);

            gzone = tmp_lib[ii].zone;
            status = resample_chip(in_chip_io, out_chip_io, gzone, gcode,
                    gparms, gspheroid, iunit, icode, izone, iparms, ispheroid,
                    iunit, &tmp_lib[ii]);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Resampling chip %s", tmp_lib[ii].chip_name);
                ias_gcp_close_image_chip_container(in_chip_io);
                ias_gcp_close_image_chip_container(out_chip_io);
                free(copy_chip_buffer);
                free(tmp_lib);
                return ERROR;
            }
            /* Set the GCPLib zone to the zone it was resampled to */
            tmp_lib[ii].zone = izone;
        }
        else if (use_tarball)
        {
            /* The chip doesn't need to be resampled but it still needs to be
               copied unmodified from the input container to the output
               container if the container is a tarball */

            int chip_bytes;     /* Bytes required to store this chip */

            /* Calculate the worst case size for the chip buffer (4 bytes per
               pixel).  Could get the exact number of bytes per pixel based on
               the data type, but it isn't worth saving a few bytes of memory */
            chip_bytes = 4 * tmp_lib[ii].chip_size_lines
                * tmp_lib[ii].chip_size_samples;

            /* Reallocate the buffer if it isn't big enough */
            if (chip_bytes > copy_chip_buffer_size)
            {
                void *new_buffer;

                new_buffer = realloc(copy_chip_buffer, chip_bytes);
                if (!new_buffer)
                {
                    IAS_LOG_ERROR("Failed reallocating %d bytes for a chip "
                        "buffer", chip_bytes);
                    ias_gcp_close_image_chip_container(in_chip_io);
                    ias_gcp_close_image_chip_container(out_chip_io);
                    free(copy_chip_buffer);
                    free(tmp_lib);
                    return ERROR;
                }
                copy_chip_buffer = new_buffer;
                copy_chip_buffer_size = chip_bytes;
            }

            /* Read the original chip */
            if (ias_gcp_read_image_chip(in_chip_io, tmp_lib[ii].chip_name,
                    tmp_lib[ii].chip_size_lines, tmp_lib[ii].chip_size_samples,
                    tmp_lib[ii].chip_data_type, IAS_GCP_NO_CONVERSION,
                    copy_chip_buffer) != SUCCESS)
            {
                IAS_LOG_ERROR("Failed reading image chip %s from %s",
                    tmp_lib[ii].chip_name, gcp_dir);
                ias_gcp_close_image_chip_container(in_chip_io);
                ias_gcp_close_image_chip_container(out_chip_io);
                free(copy_chip_buffer);
                free(tmp_lib);
                return ERROR;
             }

            /* Write the chip to the new container */
            if (ias_gcp_write_image_chip(out_chip_io, tmp_lib[ii].chip_name,
                    tmp_lib[ii].chip_size_lines, tmp_lib[ii].chip_size_samples,
                    tmp_lib[ii].chip_data_type, copy_chip_buffer) != SUCCESS)
            {
                IAS_LOG_ERROR("Failed writing image chip %s to %s",
                    tmp_lib[ii].chip_name, out_container_name);
                ias_gcp_close_image_chip_container(in_chip_io);
                ias_gcp_close_image_chip_container(out_chip_io);
                free(copy_chip_buffer);
                free(tmp_lib);
                return ERROR;
            }
        }
    }

    /* Free the buffer used to copy chips */
    free(copy_chip_buffer);
    copy_chip_buffer = NULL;

    /* Done with the image chip containers, so close them */
    if (ias_gcp_close_image_chip_container(in_chip_io) != SUCCESS)
    {
        IAS_LOG_ERROR("Failed to close the input chip container %s", 
            in_container_name);
        ias_gcp_close_image_chip_container(out_chip_io);
        free(tmp_lib);
        return ERROR;
    }
    in_chip_io = NULL;
    if (ias_gcp_close_image_chip_container(out_chip_io) != SUCCESS)
    {
        IAS_LOG_ERROR("Failed to close the output chip container %s", 
            out_container_name);
        free(tmp_lib);
        return ERROR;
    }
    out_chip_io = NULL;

    /* If any chips were changed, write a new GCPLib file and move the updated
       non-tarred image chips to the main directory */
    if (chip_resampled_flag)
    {
        /* Overwrite the original GCPLib file */
        IAS_LOG_INFO("Writing updated GCPLib file to %s", gcplib_filename);
        status = ias_gcp_write_gcplib(gcplib_filename, tmp_lib, tnumgcp, 
            overall_gcp_version);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Creating GCPLib %s", gcplib_filename);
            free(tmp_lib);
            return ERROR;
        }

        /* If not using a tarball, move the resampled chips to the main
           directory */
        if (!use_tarball)
        {
            /* Move everything back up to the GCP directory and whack the
               temporary dir */
            sprintf(sys_cmd, "mv %s/%s/* %s", gcp_dir, TEMP_GCP_DIR, gcp_dir);
            status = RIGHT_SHIFT(system(sys_cmd));
            if (status != 0)
            {
                IAS_LOG_ERROR("Unable to move resampled GCPs");
                free(tmp_lib);
                return ERROR;
            }
        }
    }

    /* If using a tarball, replace the original tarball with the new file.
       This is done even if no chips were resampled since it would take more
       code to decide whether to move the file or remove the new one. */
    if (use_tarball)
    {
        /* Move the new tarball to replace the old one */
        if (rename(out_container_name, in_container_name) != 0)
        {
            IAS_LOG_ERROR("Error moving new image chip tarball %s to %s",
                out_container_name, in_container_name);
            free(tmp_lib);
            return ERROR;
        }
    }

    free(tmp_lib);
    tmp_lib = NULL;

    /* If not using a tarball, clean up the empty temporary directory */
    if (!use_tarball)
    {
        if (rmdir(out_container_name) != 0)
        {
            IAS_LOG_ERROR("Error removing temporary directory %s : %s",
                out_container_name, strerror(errno));
            return ERROR;
        }
    }

    IAS_LOG_INFO("GCP Resampling complete");

    return SUCCESS;
}
