#!/usr/bin/perl -w

# temporary instructions for setting up the environment:
# need to source the setup_db file with a parameter for this to work:
# source oli_dev/setup/setup_db olidev

use strict;
use DBI;
use ias_const;
use ias_db;                 # provides routines for accessing the database

my $output_dir = "output";
my $resampled_gcps_dir = "resampled_gcps";
my $md5_file = "md5.out";
my $odl_file = "testWO.odl";
my $omf_file = "testWO.omf";
my $omf_file_ref_before = "testWO.omf.ref_before";
my $omf_file_ref_after = "testWO.omf.ref_after";

# Testing GCP's requires access to the database.  If the environment variable
# ENABLE_DB_TESTS is not defined, skip this test and return a successful status.
if (!defined $ENV{ENABLE_DB_TESTS})
{
    print "\n";
    print __FILE__ . ": Tests not performed, database tests are disabled.\n";
    print "\n";
    exit SUCCESS;
}

# Run sub to get current gcp version
my $overall_gcp_version = get_gcp_version();
if (!defined $overall_gcp_version)
{
    print "Error retrieving overall GCP version from database\n";
    exit FAILURE;
}

# Test not using a tarball
print "Testing writing individual image chips\n";
if (run_get_gcps(0) != SUCCESS)
{
    print "Error getting GCPs\n";
    exit FAILURE;
}

# Test again using a tarball
print "Testing putting the image chips into a tarball\n";
if (run_get_gcps(1) != SUCCESS)
{
    print "Error getting GCPs\n";
    exit FAILURE;
}

exit SUCCESS;

##############################################################################
# Name: run_get_gcps
#
# Purpose: Run get_gcps.pl and check the results
#
##############################################################################
sub run_get_gcps
{
    my $use_tarball = shift;

    # clean up directory of old output/test data
    if (-d $output_dir)
    {
        system("rm -r $output_dir");
    }
    printf "creating $output_dir directory\n";
    system("mkdir $output_dir");

    # clean up old files
    unlink("$omf_file");
    system("cp $omf_file_ref_before $omf_file");

    # Generate the ODL file with the correct PACKAGE_CHIPS option
    my $replace_text = "sed -e \"s/PACKAGE_CHIPS/PACKAGE_CHIPS = "
        ."$use_tarball/\" $odl_file.template > $odl_file";
    my $rc = system($replace_text);
    if ($rc != 0)
    {
        print "Error replacing PACKAGE_CHIPS in $odl_file.template\n";
        return ERROR;
    }

    # call the Perl script to get the GCPs
    # that will call the gcpresample application afterward
    $ENV{IAS_BIN} = "..";
    $rc = system("../get_gcps.pl", "$odl_file");
    if ($rc != 0)
    {
        print "Error calling get_gcps.pl\n";
        return ERROR;
    }

    # Fix the version number in the GCPLib file
    $replace_text = "sed -e \"s/VERSION = ".$overall_gcp_version
        ."/VERSION = 1/\" ".$output_dir."/GCPLib >".$output_dir
        ."/GCPLib.verify";
    $rc = system($replace_text);
    if ($rc != 0)
    {
        print "Error replacing VERSION in file\n";
        return ERROR;
    }

    # If using a tarball, extract the chips
    if ($use_tarball == 1)
    {
        $rc = system("cd $output_dir && tar xf image_chips.tar");
        if ($rc != 0)
        {
            print "Error untarring $output_dir/image_chips.tar\n";
            return ERROR;
        }
    }

    # md5sum the output data to make sure it's correct
    $rc = system("md5sum -c $md5_file");
    if ($rc == 0)
    {
        print "md5sum test passed\n";

        # compare omf files
        my $rc = system("diff --brief $omf_file $omf_file_ref_after");
        if ($rc == 0)
        {
            print "OMF diff test passed\n";
            return SUCCESS;
        }
        else
        {
            print "OMF diff test failed\n";
            return ERROR;
        }
    }
    else
    {
        print "md5sum test failed\n";
        return ERROR;
    }
}

##############################################################################
# Name: get_gcp_version
#
# Purpose: Get the current GCP version from the database
#
##############################################################################
sub get_gcp_version
{
    # get the username and password for the sensor database
    my ($user_id, $password) = ias_db_GetDBConnectionInfo("COMMON");

    # connect to the database
    my $dbh = ias_db_ConnectToDB($user_id, $password);
    if (!$dbh)
    {
        print "Error connecting to database\n";
        exit FAILURE;
    }

    my $sql = qq{SELECT VERSION FROM GCP_DATA_VERSION};
    my $version_sth;

    if (! ($version_sth = $dbh->prepare($sql)))
    {
        $dbh->disconnect;
        return undef;
    }

    # execute the query
    if (! ($version_sth->execute()))
    {
        $version_sth->finish;
        $dbh->disconnect;
        return undef;
    }

    my $version = $version_sth->fetchrow_array;

    # Close connection
    $version_sth->finish;
    $dbh->disconnect;

    return $version;
}
