#!/usr/bin/perl -w

############################################################################
#
# NAME:     test_get_gcp_wrapper.pl
#
# PURPOSE:  Implements unit testing of the PERL script 'get_gcps_wrapper.pl'.
#
# RETURNS:  Nothing
#
# NOTES:    Two methods for retrieving a set of GCP image chips are
#           considered:
#             i.  Through specification of 'bounding box' corner coordinates
#            ii.  Through specification of a WRS path/row
############################################################################

use strict;
use DBI;
use ias_const;
use ias_db;                 # provides routines for accessing the database
use File::Spec;

# Subroutine forward declarations
sub cleanupOutputDirectory($);

# Test GCP output directories and reference GCPLib file names.
my $bbox_output_dir         = "output_bbox";
my $prow_output_dir         = "output_pathrow";
my $ref_GCPLib_bbox         = "ref_GCPLib_file_bbox";
my $ref_GCPLib_prow         = "ref_GCPLib_file_prow";

# Miscellaneous variables.
my $rc;                          # system command return status
my $exit_status = SUCCESS;


# Set up the test environment to look in the source code directory for the
# tools before checking the rest of the directory path.
my $test_env = 'PATH=../:${PATH}';


# Pulls 5 chips from the current database.
my $test_command_line_bbox = "get_gcps_wrapper.pl "
    . "--corners 43.7 -96.9 43.7 -96.6 43.4 -96.9 43.4 -96.6 "
    . "--gcp_type CONTROL --gcp_source GLS --start_date 2002-01-01 "
    . "--end_date 2002-12-31 --output output_bbox";

# Pulls 278 chips from the current database.
my $test_command_line_prow = "get_gcps_wrapper.pl "
    . "--gcp_type CONTROL --gcp_source GLS --wrs_path 30 --wrs_row 29 "
    . "--start_date 2002-01-01 --end_date 2002-12-31 --resolution 30 "
    . "--output output_pathrow";

# Access to the database is required for this test.  If we don't have
# it, inform the user and quit here.
if (!defined $ENV{'ENABLE_DB_TESTS'})
{
    print "\n";
    print __FILE__, ":  Test not performed database tests are disabled.\n";
    print "\n";
    exit SUCCESS;
}


#**************************************************************************
#                         INITIALIZATION
#**************************************************************************

# Clean up any previous instances of the output directory.
cleanupOutputDirectory($bbox_output_dir);
cleanupOutputDirectory($prow_output_dir);

#**************************************************************************
#                        GET CURRENT GCP VERSION                           
#**************************************************************************
# get the username and password for the sensor database
my ($user_id, $password) = ias_db_GetDBConnectionInfo("COMMON");

# connect to the database
my $dbh = ias_db_ConnectToDB($user_id, $password);
if (!$dbh)
{
    print STDOUT "Connecting to database.\n";
    $exit_status = FAILURE
}

# Run sub to get current gcp version
my $overall_gcp_version = get_gcp_version($dbh);
    
# Close connection
$dbh->disconnect;

if (!defined $overall_gcp_version)
{
    print "Error retrieving overall GCP version from database\n";
    exit FAILURE;
}

#**************************************************************************
#                 FIRST TEST -- SPECIFYING BOUNDING BOX
#**************************************************************************/

# Run the test.
$rc = system("$test_env $test_command_line_bbox");
if ($rc != 0)
{
    print STDOUT "GCP retrieval using bounding box corner points failed.\n";
    $exit_status = FAILURE;
}

my $replace_text = "sed -e \"s/VERSION = ".$overall_gcp_version
    ."/VERSION = 1/\" ".$bbox_output_dir."/GCPLib > "
    . $bbox_output_dir."/GCPLib.verify";
$rc = system($replace_text);
if ($rc != 0)
{
    print STDOUT "Creating bounding box GCPLib verification file.\n";
    $exit_status = FAILURE;
}

# Compare the GCPLib generated for the test vs. the expected result
my $sys_command_line = "diff --brief $bbox_output_dir/GCPLib.verify ". 
    "$ref_GCPLib_bbox";
$rc = system($sys_command_line);
if ($rc != 0)
{
    print STDOUT "Bounding box GCPLib file comparison failed.\n";
    $exit_status = FAILURE;
}
else
{
    print STDOUT "Bounding box GCPLib file comparison passed.\n";
}

#**************************************************************************
#                 SECOND TEST -- SPECIFYING WRS PATH/ROW
#**************************************************************************/

# Run the test.
$rc = system("$test_env $test_command_line_prow");
if ($rc != 0)
{
    print STDOUT "GCP retrieval using WRS path/row failed.\n";
    $exit_status = FAILURE;
}

$replace_text = "sed -e \"s/VERSION = ".$overall_gcp_version
    ."/VERSION = 1/\" ". $prow_output_dir."/GCPLib > "
    . $prow_output_dir."/GCPLib.verify";
$rc = system($replace_text);
if ($rc != 0)
{
    print STDOUT "Creating WRS path/row GCPLib verification file.\n";
    $exit_status = FAILURE;
}

# Compare the GCPLib generated for the test vs. the expected result
$sys_command_line = "diff --brief $prow_output_dir/GCPLib.verify ".
    "$ref_GCPLib_prow";
$rc = system($sys_command_line);
if ($rc != 0)
{
    print STDOUT "WRS path/row GCPLib file comparison failed.\n";
    $exit_status = FAILURE;
}
else
{
    print STDOUT "WRS path/row GCPLib file comparison passed.\n";
}

#**************************************************************************
#                          CLEAN UP
#**************************************************************************

# Clean up the current output directories.
if ($exit_status == SUCCESS)
{
    cleanupOutputDirectory($bbox_output_dir);
    cleanupOutputDirectory($prow_output_dir);
}

exit $exit_status;



#--------------------------------------------------------------------------
# NAME:       cleanupOutputDirectory
#
# PURPOSE:    Cleans up instances of the specified output directory
#
# RETURNS:    Nothing
#--------------------------------------------------------------------------
sub cleanupOutputDirectory($)
{
    # Get the output directory name to clean up.
    my $output_dir = shift;

    print STDOUT "cleaning up previous $output_dir directory contents...\n";

    # Open the specified output directory.
    if (-d $output_dir)
    {
        opendir (DIR, $output_dir) or die $!;

        # Get the list of files in the output directory.
        my @files = grep {!/^\.{1,2}$/} readdir(DIR);
        closedir (DIR);

        # Delete each file.
        foreach my $file (@files)
        {
            unlink("$output_dir/$file") or carp $!;
        }

        # Remove the directory itself.
        rmdir ($output_dir) || die("error in deleting $output_dir directory");
    }
}

##############################################################################
# Name: get_gcp_version
#
# Purpose: Get the current GCP version from the database
#
##############################################################################
sub get_gcp_version
{
    my ($dbh) = @_;

    my $version_sth;

    # Get the default logger object from the library 
    my $sql = qq{SELECT VERSION FROM GCP_DATA_VERSION};

    if (! ($version_sth = $dbh->prepare($sql)))
    {
        return undef;
    }

    # execute the query
    if (! ($version_sth->execute()))
    {
        $version_sth->finish;
        return undef;
    }

    my $version = $version_sth->fetchrow_array;

    return $version;
}
