/****************************************************************************
 NAME:                    adjust_elev_to_wgs84

 PURPOSE:      Adjusts the elevation values in the input DEM cell
               to values relative to the WGS84 ellipsoid, rather than
               relative to mean sea level

 RETURNS:      Integer status code of SUCCESS or ERROR
****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>                /* PATH_MAX #defined */
#include <math.h>                  /* floor()           */

#include "ias_logging.h"
#include "ias_const.h"

#include "retrieve_elevation.h"
#include "get_geoid_height.h"
#include "adjust_elev_to_wgs84.h"
#include "round.h"

int adjust_elev_to_wgs84
(
    const char *dem_dir,    /* I: DEM data parent directory */
    GTOPO30_CELL *cell      /* I/O: cell of elevation data to adjust */
)
{
    short nodata = cell->nodata;
    short *data_ptr = cell->data;

    GTOPO30_CELL geoid;

    char geoid_hfile[PATH_MAX];        /* Name of geoid data header file */
    char geoid_dfile[PATH_MAX];        /* Name of geoid data DEM file */

    double lat;                        /* Current latitude */
    double lon;                        /* Current longitude */

    int geoid_sep;                     /* Geoid height */
    int line_index;                    /* DEM data array offsets/indices */
    int sample_index;
    int line;                          /* Loop counters */
    int sample;
    

    /* For now, assume the DEM data parent directory is known.  This
       way allows it to be passed in */
    /* The geoid separation table is in the directory pointed to by the IASDEM
       environment variable.  For right now, pass in the DEM directory
       name.                */

    /* Construct the full names of the geoid header and DEM data files */
    strcpy(geoid_hfile, dem_dir);
    strcat(geoid_hfile, "/GEOID.HDR");
    strcpy(geoid_dfile, dem_dir);
    strcat(geoid_dfile, "/GEOID.DEM");

    /* Initialize the geoid cell data structure */
    memset(&geoid, 0, sizeof(GTOPO30_CELL));

    /* Read the geoid separation header */
    if (read_gtopohdr(geoid_hfile, &geoid) != SUCCESS)
    {
        IAS_LOG_ERROR("Cannot read geoid separation header data in "
                      "%s", geoid_hfile);
        return ERROR;
    }

    /* Read the geoid separation data */
    if (read_gtopodem(geoid_dfile, &geoid) != SUCCESS)
    {
        IAS_LOG_ERROR("Cannot read geoid separation data in %s",
                      geoid_dfile);
        free(geoid.data);
        return ERROR;
    }

    /* Apply the Geoid Separation Table to the elevation data */
    for (line = 0; line < cell->nrows; line++)
    {
        line_index = line * cell->ncols;
        lat = cell->ulymap - (line * cell->ydim);

        for (sample = 0; sample < cell->ncols; sample++)
        {
            sample_index = line_index + sample;

            if (data_ptr[sample_index] != nodata)
            {
                lon = cell->ulxmap + (sample * cell->xdim);
                geoid_sep
                    = (int)floor(get_geoid_height(&geoid, lat, lon) + 0.5);
                data_ptr[sample_index] += geoid_sep;
            }
        }
    }

    /* Free the data in the Geoid Separation Table structure */
    free(geoid.data);

    return SUCCESS;
}
