/*****************************************************************************
Name: get_cell_intersection

Description: Function to determine which part of the input cell intersects
    with the requested mosaic output area.

Returns:
    The upper left line, sample in the input cell where the intersection starts
    The upper left line, sample in the output mosaic where the intersection 
        starts
    The number of lines and samples in the intersecting area

*****************************************************************************/
#include "retrieve_elevation.h"
#include "get_cell_intersection.h"
#include "modulus.h"
#include "round.h"

void get_cell_intersection
(
    const GTOPO30_CELL *input_cell,  /* I: input elevation cell */
    const GTOPO30_CELL *output_cell, /* I: output cell */
    int *ic_ul_line,                 /* O: input cell upper-left line */
    int *ic_ul_sample,               /* O: input cell upper-left sample */
    int *oc_ul_line,                 /* O: output cell upper-left line */
    int *oc_ul_sample,               /* O: output cell upper-left sample */
    int *lines,                      /* O: lines in intersecting area */
    int *samples                     /* O: samples in intersecting area */
)
{
    /* NOTE: To make comparisons and calculations more simple, oc_lrxmap and
     * ic_lrxmap are not normalized longitude values.  Thus, they may have
     * values 180.0 or greater. */

    /* calculate the output cell lower right corner map coordinates */
    double oc_lrxmap = output_cell->ulxmap 
                     + (output_cell->ncols - 1) * output_cell->xdim;
    double oc_lrymap = output_cell->ulymap 
                     - (output_cell->nrows - 1) * output_cell->ydim;
    /* calculate the input cell lower right corner map coordinates */
    double ic_lrxmap = input_cell->ulxmap 
                     + (input_cell->ncols - 1) * input_cell->xdim;
    double ic_lrymap = input_cell->ulymap 
                     - (input_cell->nrows - 1) * input_cell->ydim;

    if (longitude_norm(input_cell->ulxmap - output_cell->ulxmap) > 0)
    {
        /* the input cell starts to the right of the output cell left edge,
           so start at the first sample of the input cell and the correct
           sample in the output cell */
        *ic_ul_sample = 0;
        *oc_ul_sample
            = round_value(fmodulus((input_cell->ulxmap - output_cell->ulxmap),
                                   360) /output_cell->xdim);
    }
    else
    {
        /* the input cell starts outside the left edge of the output cell,
           so start at the correct sample of the input cell and the first
           sample of the output cell */
        *ic_ul_sample
            = round_value(fmodulus((output_cell->ulxmap - input_cell->ulxmap),
                                   360) /input_cell->xdim);
        *oc_ul_sample = 0;
    }

    /* calculate the number of samples across the intersect area, depending
       on whether the right side of the input cell falls inside of or outside
       of the output cell */
    if (longitude_norm(ic_lrxmap - oc_lrxmap) > 0)
    {
        *samples = output_cell->ncols - *oc_ul_sample;
    }
    else
    {
        *samples = input_cell->ncols - *ic_ul_sample;
    }

    if (input_cell->ulymap <= output_cell->ulymap)
    {
        /* the input cell starts below the top edge of the output cell,
           so start at the first line of the input cell and the correct 
           line in the output cell */
        *ic_ul_line = 0;
        *oc_ul_line = round_value((output_cell->ulymap - input_cell->ulymap)
                                  /output_cell->ydim);
    }
    else
    {
        /* the input cell starts above the top edge of the output cell,
           so start at the correct line of the input cell and the first
           line of the output cell */
        *ic_ul_line = round_value((input_cell->ulymap - output_cell->ulymap)
                                  /input_cell->ydim);
        *oc_ul_line = 0;
    }

    /* calculate the number of lines across the intersect area, depending
       on whether the bottom of the input cell falls inside of or outside
       of the output cell */
    if (ic_lrymap < oc_lrymap)
    {
        *lines = output_cell->nrows - *oc_ul_line;
    }
    else
    {
        *lines = input_cell->nrows - *ic_ul_line;
    }
}

