/***************************************************************************
 NAME:                   get_geoid_height

 PURPOSE:       Computes the geoid height separation for the specified
                lat/long coordinates

 RETURNS:       Geoid separation in meters
****************************************************************************/
#include <math.h>                  /* floor() and ceil()  */

#include "retrieve_elevation.h"
#include "gtopo30.h"
#include "get_geoid_height.h"

double get_geoid_height
(
    GTOPO30_CELL *gcell,            /* I/O: Geoid separation data cell */
    double lat,                     /* I: Current latitude in the range of
                                       [-90.0, 90] degrees */
    double lon                      /* I: Current longitude in the range of
                                       [-180.0, 180.0] degrees */
)
{
    int i;                      /* Array indices for the initial (SW)
                                   geoid height in the DEM data cell */
    int j;
    int ncols;                  /* # of columns in the table */

    double x;                   /* Rectangular coordinates corresponding
                                   to the given longitude and latitude */
    double y;
    double height;              /* Returned geoid height */
    double gh_sw;               /* Geoid heights for the SW, SE, NE,
                                   and NW points in the DEM data cell,
                                   respectively */
    double gh_se;
    double gh_ne;
    double gh_nw;

    double scaled_lat;           /* temp variable to save division */
    double scaled_lon;           /* temp variable to save division */



    /* Compute initial (SW) cell position indices i and j */
    lat = NORTH_LATITUDE_LIMIT - lat;
    scaled_lat = lat / gcell->ydim;
    i = (int)floor(scaled_lat);

    lon += EAST_LONGITUDE_LIMIT;   /* Ensures longitudes between +- 180 deg */
    scaled_lon = lon / gcell->xdim;
    j = (int)floor(scaled_lon);
    if (j < 0)
    {
        /* Converts to a positive index value */
        j += (int)(2 * EAST_LONGITUDE_LIMIT);
    }

    /* Get the geoid heights of the surrounding points */
    ncols = gcell->ncols;
    gh_ne = gcell->data[i*ncols + j + 1];  /* NE Point */
    gh_nw = gcell->data[i*ncols + j];      /* NW Point */
    gh_sw = gcell->data[(i+1)*ncols + j];          /* SW Point */
    gh_se = gcell->data[(i+1)*ncols + j + 1];      /* SE Point */

    /* Data is half-degree; normalize the fractional
       step to range from 0 to 1 */
    x = scaled_lon - floor(scaled_lon);
    y = scaled_lat -  floor(scaled_lat);

    /* Determine the full geoid height separation from the
       (x, y) coordinates and the component geoid heights.  Use
       bilinear interpolation to get a more "refined" estimate */
    height = ((1.0 - x) * y * gh_sw) + ((1.0 - x) * (1.0 - y) * gh_nw);
    height += (x * y * gh_se) + (x * (1.0 - y) * gh_ne);

    /* Return geoid height separation value */
    return height;
}

