/*****************************************************************************
 NAME:                       get_gls_data

 DESCRIPTION: Function to build the mosaic'ed DEM from gls DEM data.

 RETURNS:     Integer status code of SUCCESS, WARNING  or ERROR
              Mosaiced DEM data is passed back to the calling routine
              via pointer
*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>              /* floor() and ceil() */


#include "ias_logging.h"
#include "ias_const.h"

#include "dem_types.h"
#include "retrieve_elevation.h"
#include "get_gls_tile_list.h"
#include "get_gls_data.h"
#include "gtopo30.h"
#include "load_gls.h"
#include "adjust_elev_to_wgs84.h"
#include "get_cell_intersection.h"
#include "modulus.h"


int get_gls_data
(
    const char *dem_root_dir, /* I: DEM data root directory */
    double ullat,             /* I: Upper-left latitude of bounding box */
    double ullon,             /* I: Upper-left longitude of bounding box */
    double lrlat,             /* I: Lower-right latitude of bounding box */
    double lrlon,             /* I: Lower-right longitude of bounding box */
    GTOPO30_CELL *elevation_cell      /* O:  Mosaiced DEM data */
)
{
    GTOPO30_CELL current_cell;

    int ic_ul_line;                /* Input cell ul line number */
    int ic_ul_sample;              /* Input cell ul sample number */
    int oc_ul_line;                /* Output cell ul line number */
    int oc_ul_sample;              /* Output cell ul sample number */
    int lines;                     /* Number of lines intersecting area */
    int ic_index;                  /* Index into input cell data buffer */
    int oc_index;                  /* Index into output cell data buffer */
    int samples;                   /* Number of samples in intersecting
                                      area */
    int line;                      /* Current line */
    int status;
    int i;                         /* Loop counter */
    int tile;                      /* current tile number */
    int num_elevation_pts;
    int tile_count = 0;            /* Total number of needed tiles */
    int missing_tiles = 0;         /* Total number of missing tiles */

    char dem_dir[PATH_MAX];        /* Full path to DEM data */

    char **tile_names = NULL;      /* Tile list */


    /* Get the size (in bytes) of the data pointer for the elevation cell */
    size_t elev_data_size =  sizeof(*(elevation_cell->data));

    /* Get the size (in bytes) of the entire input cell data structure */
    size_t curr_cell_size = sizeof(current_cell);

    /* Get the gls tile names needed for the output area.  For the LDCM
       IAS, tiles won't be staged from the silo--they'll be retrieved from
       the local cache */
    tile_names = get_gls_tile_list(ullat, ullon, lrlat, lrlon,
                                   &tile_count);
    if (!tile_names || !tile_count)
    {
        IAS_LOG_ERROR("Retrieving required GLS DEM tile list");
        return ERROR;
    }
    
    IAS_LOG_INFO("Mosaicing GLS DEM tiles...");

    /* Get to the GLS DEM data directory */
    snprintf(dem_dir, sizeof(dem_dir), "%s/%s", dem_root_dir, GLSDEM_DIR);

    /* Process each tile */
    for (tile = 0; tile < tile_count; tile++)
    {

        /* initialize the current cell structure */
        memset(&current_cell, 0, curr_cell_size);

        /* Read the current cell */
        status = load_gls(dem_dir, tile_names[tile], &current_cell,
                          &missing_tiles);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading data for tile %s", tile_names[tile]);

            /* Free the output cell buffer if it's been previously
               allocated */
            for (i = 0; i < tile_count; i++)
            {
                free(tile_names[i]);
            }
            free(tile_names);
            return ERROR;
        }

        /* Convert the map coordinates from arc-seconds to decimal degrees */
        current_cell.ulxmap /= DEG2ARCSEC;
        current_cell.ulymap /= DEG2ARCSEC;
        current_cell.xdim /= DEG2ARCSEC;
        current_cell.ydim /= DEG2ARCSEC;

        /* Trace data from each input tile */
        IAS_LOG_DEBUG("nrows           = %d\nnrows           = %d\n"
                      "nbands          = %d\nnbits           = %d\n"
                      "band row bytes  = %d\ntotal row bytes = %d\n"
                      "bandgap bytes   = %d\nul xmap         = %lf\n"
                      "ul ymap         = %lf\nxdim            = %lf\n"
                      "ydim            = %lf\n", current_cell.nrows,
                      current_cell.ncols, current_cell.nbands,
                      current_cell.nbits, current_cell.bandrowbytes,
                      current_cell.totalrowbytes,
                      current_cell.bandgapbytes, current_cell.ulxmap,
                      current_cell.ulymap, current_cell.xdim,
                      current_cell.ydim);

        /* When processing the first tile, populate the output header info
           and allocate memory for the elevation data */
        if (tile == 0)
        {
            /* Set the data source */
            elevation_cell->source = GLSDEM_SOURCE;

            /* Set up the output cell header values */
            strcpy(elevation_cell->byteorder, current_cell.byteorder);
            strcpy(elevation_cell->layout, current_cell.layout);
            elevation_cell->xdim         = current_cell.xdim;
            elevation_cell->ydim         = current_cell.ydim;
            elevation_cell->nodata       = current_cell.nodata;
            elevation_cell->nbands       = DMS_DRE_DEM_NBANDS;
            elevation_cell->nbits        = DMS_DRE_DEM_NBITS;
            elevation_cell->bandgapbytes = current_cell.bandgapbytes;

            /* Set the upper-left corner to an even multiple of the pixel
               size */
            elevation_cell->ulxmap
            = floor(ullon / elevation_cell->xdim) * elevation_cell->xdim;
            elevation_cell->ulymap
            = ceil(ullat / elevation_cell->xdim) * elevation_cell->ydim;

            /* Note: if the added value (0.5) is ever increased (say to 1.5)
             * then get_gls_tile_list will have to be revisited.  Consider
             * the following situation:
             *     elevation_cell->ulxmap = 0.0, lrlon = 0.99987
             * Then, elevation_cell->ncols gets 3602, which will be too wide
               for the one tile that will be returned by get_gls_tile_list.
             *
             * As it is, we only get away without modifying get_gls_tile_list
             * because the gls tiles have one column/row of overlap between
             * the tiles (making them 3601x3601 rather than 3600x3600).
             */

            /* Set the number of rows and columns of DEM data */
            elevation_cell->nrows
            = (int)ceil(((elevation_cell->ulymap-lrlat)/elevation_cell->ydim)
                        + 0.5);
            elevation_cell->ncols
            = (int)ceil((fmodulus((lrlon - elevation_cell->ulxmap), 360)
                        / elevation_cell->xdim) + 0.5);

            /* Set the additional size information */
            elevation_cell->bandrowbytes
            = elevation_cell->ncols * (elevation_cell->nbits / 8);

            elevation_cell->totalrowbytes
            = elevation_cell->bandrowbytes * elevation_cell->nbands;

            /* Do a temporary calculation of the number of elevation points
               to confirm that the allocated memory buffer will be able to
               handle all of them without going into overflow.  Do the
               calculation in double precision to allow for numbers
               larger than the largest signed integer.  */
            if ((double)elevation_cell->nrows * elevation_cell->ncols
                > INT_MAX)
            {
                IAS_LOG_ERROR("Requested elevation image is too large to "
                    "support (%d lines by %d samples)", elevation_cell->nrows,
                    elevation_cell->ncols);
                for (i = 0; i < tile_count; i++)
                {
                    free(tile_names[i]);
                    tile_names[i] = NULL;
                }
                free(tile_names);
                free(current_cell.data);
                return ERROR;
            }

            /* The number of points can be processed without overflow, so
               proceed */
            num_elevation_pts
            = elevation_cell->nrows * elevation_cell->ncols;

            /* Now... allocate the DEM data buffer */
            elevation_cell->data
                = malloc(num_elevation_pts * elev_data_size);
            if (elevation_cell->data == NULL)
            {
                IAS_LOG_ERROR("Allocating mosaiced DEM data buffer");
                for (i = 0; i < tile_count; i++)
                {
                    free(tile_names[i]);
                    tile_names[i] = NULL;
                }
                free(tile_names);
                free(current_cell.data);
                return ERROR;
            }

            /* initialize the output cell to an illegal value */
            for (i = 0; i < num_elevation_pts; i++)
            {
                elevation_cell->data[i] = ILLEGAL_ELEV;
            }
        }   /* END -- IF block processing first tile */
        else if (elevation_cell->nodata == ILLEGAL_ELEV)
        {
            /* If the first cell was missing, allow any later found cells to
               override mosaiced DEM's nodata marker */
            elevation_cell->nodata = current_cell.nodata;
        }

        /* Find the intersection of the current cell and the output cell */
        get_cell_intersection(&current_cell, elevation_cell, &ic_ul_line, 
                              &ic_ul_sample, &oc_ul_line, &oc_ul_sample,
                              &lines, &samples);

        /* Extract the required data from the current elevation cell and
           copy it into the output elevation cell */
        if ((lines > 0) && (samples > 0))
        {
            for (line = 0; line < lines; line++)
            {
                oc_index = (line + oc_ul_line) * elevation_cell->ncols;
                ic_index = (line + ic_ul_line) * current_cell.ncols;

                memcpy(&elevation_cell->data[oc_index + oc_ul_sample],
                       &current_cell.data[ic_index + ic_ul_sample], 
                       samples * elev_data_size);
            }
        }

        /* Free the current read cell */
        free(current_cell.data);
        current_cell.data = NULL;

    }  /* END -- Processing tile loop */

    /* Free up the tile list -- don't need it anymore */
    for (i = 0; i < tile_count; i++)
    {
        free(tile_names[i]);
    }
    free(tile_names);

    if (missing_tiles >= tile_count) 
    {
        /* we have nothing but zero-data */
        IAS_LOG_WARNING("No tiles found");
        free(elevation_cell->data);
        return WARNING;
    }

    /* Make sure there are no remaining illegal values in the output data
       (to make sure the entire area has been filled)*/
    num_elevation_pts = elevation_cell->nrows * elevation_cell->ncols;
    for (i = 0; i < num_elevation_pts; i++)
    {
        if (elevation_cell->data[i] == ILLEGAL_ELEV)
        {
            IAS_LOG_ERROR("Missing data in output cell, offset=%d", i+1);
            free(elevation_cell->data);
            return ERROR;
        }
    }

    /* Adjust the elevation values to be relative to the WGS84 ellipsoid
       instead of mean sea level */
    snprintf(dem_dir, sizeof(dem_dir), "%s/%s", dem_root_dir, "geoid");

    status = adjust_elev_to_wgs84(dem_dir, elevation_cell);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Adjusting elevation values to the WGS84 ellipsoid");
        free(elevation_cell->data);
        return ERROR;
    }

    /* Print out output DEM header information in debug mode */
    IAS_LOG_DEBUG("nrows           = %d\nncols           = %d\n"
                  "nbands          = %d\nnbits           = %d\n"
                  "band row bytes  = %d\ntotal row bytes = %d\n"
                  "bandgap bytes   = %d\nul xmap         = %lf\n"
                  "ul ymap         = %lf\nxdim            = %lf\n"
                  "ydim            = %lf\n", elevation_cell->nrows,
                  elevation_cell->ncols, elevation_cell->nbands,
                  elevation_cell->nbits, elevation_cell->bandrowbytes,
                  elevation_cell->totalrowbytes,
                  elevation_cell->bandgapbytes, elevation_cell->ulxmap,
                  elevation_cell->ulymap, elevation_cell->xdim,
                  elevation_cell->ydim);

    /* Done */
    return SUCCESS;
}
