/*****************************************************************************
 NAME:                         get_gls_tile_list

 DESCRIPTION: Function to determine which GLS DEM tiles intersect the
              indicated latitute/longitude bounding box and return the
              list of tiles.

 RETURNS:     Pointer to tile name list if successful
              NULL pointer on error

              Tile count is passed back to the calling routine via
              pointer
*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>         /* floor() and/or ceil() */

#include "ias_const.h"
#include "ias_logging.h"

#include "retrieve_elevation.h"
#include "make_gls_tile_name.h"
#include "get_gls_tile_list.h"
#include "modulus.h"


char **get_gls_tile_list
(
    double ullat,                   /* I: Scene upper-left lat coordinate */
    double ullon,                   /* I: Scene upper-left lon coordinate */
    double lrlat,                   /* I: Scene lower-right lat coordinate */
    double lrlon,                   /* I: Scene lower-right lon
                                       coordinate */
    int    *num_tiles               /* O: Number of required tiles to
                                       mosaic corresponding output area */
)
{
    char **tile_names = NULL; /* Returned tile list */

    int total_tiles;          /* Total number of tiles required */
    int tile_count = 0;       /* Tile counter */
    int i;                    /* Loop counter */

    /* GLS tiles are named with a format like n47w118 where the lat/long
       in the file name are the lower-left corner of the tile.  Each tile
       covers a 1-degree by 1-degree area. */
    int start_lat     = (int)floor(ullat);
    int end_lat       = (int)floor(lrlat);
    int start_lon     = (int)floor(longitude_norm(ullon));
    int end_lon       = (int)floor(longitude_norm(lrlon));
    int too_far_lon   = (int)floor(longitude_norm(lrlon + 1.0));
    int current_lat;
    int current_lon;
    int lat_range = start_lat - end_lat + 1;
    int lon_range = modulus((end_lon - start_lon + 1), (int)DEG2ARCSEC);

    /* Allocate memory for the tile name pointers */
    total_tiles = lat_range * lon_range;
    tile_names = malloc(total_tiles * sizeof(char *));
    if (!tile_names)
    {
        IAS_LOG_ERROR("Allocating tile name data buffer");
        return NULL;
    }

    for (current_lat = start_lat; current_lat >= end_lat; current_lat--)
    {
        /* We are working under the assumption that the caller does not want
           to circle the entire globe.  too_far_lon is the next tile to the
           East of the last tile we want. */
        for (current_lon = start_lon; current_lon != too_far_lon;
             current_lon = floor(longitude_norm(current_lon + 1.0)))
        {
            /* get the full path to the tile */
            tile_names[tile_count] = calloc(GLSDEM_TILE_NAME_SIZE,
                                       sizeof(char));
            if (tile_names[tile_count] == NULL)
            {
                IAS_LOG_ERROR("Allocating individual tile name buffer");
                for (i = 0; i < tile_count; i++)
                {
                    free(tile_names[i]);
                }
                return NULL;
            }

            /* Make the tile name for the current lat/lon coordinates */
            make_gls_tile_name(current_lat, current_lon,
                               tile_names[tile_count]);

            /* Increment the tile count */
            tile_count++;
        }
    }

    /* return the count of tiles found and the names of the tiles */
    *num_tiles = tile_count;

    /* Done */
    return tile_names;
}
