/*****************************************************************************
Name: get_gtopo30_data

Description: Function to build the mosaic'ed DEM from GTOPO30 DEM data.

Returns:
    ERROR/SUCCESS indication
    mosaic'ed DEM data

*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>             /* floor() and ceil()   */

#include "ias_logging.h"
#include "ias_const.h"

#include "retrieve_elevation.h"
#include "gtopo30.h"
#include "get_gtopo30_tile_list.h"
#include "get_gtopo30_data.h"
#include "adjust_elev_to_wgs84.h"
#include "get_cell_intersection.h"
#include "modulus.h"


int get_gtopo30_data
(
    const char *dem_root_dir,   /* I: DEM data root directory */
    double ullat,          /* I: Upper-left latitude of bounding box */
    double ullon,          /* I: Upper-left longitude of bounding box */
    double lrlat,          /* I: Lower-right latitude of bounding box */
    double lrlon,          /* I: Lower-right longitude of bounding box */
    GTOPO30_CELL *elevation_cell  /* O: Elevation data read */
)
{
    GTOPO30_CELL current_cell;     /* Input cell for a GTOPO30 DEM
                                      "tile" */

    int ic_ul_line;                /* Input cell ul line number */
    int ic_ul_sample;              /* Input cell ul sample number */
    int oc_ul_line;                /* Output cell ul line number */
    int oc_ul_sample;              /* Output cell ul sample number */
    int lines;                     /* Number of lines intersecting area */
    int ic_index;                  /* Index into input cell data buffer */
    int oc_index;                  /* Index into output cell data buffer */
    int samples;                   /* Number of samples in intersecting
                                      area */
    int line;                      /* Current line */
    int status;
    int i;
    int tile;                      /* Current tile number */
    int num_elevation_pts;
    int tile_count;                /* Total number of needed tiles */

    char dem_dir[PATH_MAX];        /* Full path to DEM data */

    char **tile_names = NULL;      /* Tile list */

    short nodata;
    short *data_ptr = NULL;        /* Pointer to DEM image data */

    /* Get the size (in bytes) of the pointer for the data cell */
    size_t elev_data_size = sizeof(*(elevation_cell->data));

    /* Get the total size of the current data cell */
    size_t curr_cell_size = sizeof(current_cell);

    /* get the GTOPO30 tile names needed for the output area */
    tile_names = get_gtopo30_tile_list(ullat, ullon, lrlat, lrlon,
                                       &tile_count);
    if (tile_names == NULL)
    {
        IAS_LOG_ERROR("Retrieving GTOPO30 tile list");
        return ERROR;
    }

    IAS_LOG_INFO("Mosaicing GTOPO30 DEM tiles...");

    /* Get to the GLS DEM data directory */
    snprintf(dem_dir, sizeof(dem_dir), "%s/%s", dem_root_dir, GTOPO_DIR);

    /* Process each of the tiles */
    for (tile = 0; tile < tile_count; tile++)
    {
        /* initialize the current cell structure */
        memset(&current_cell, 0, curr_cell_size);

        /* read the current cell */
        if (load_gtopo(dem_dir, tile_names[tile], &current_cell) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading data for tile %s", tile_names[tile]);

            /* Free the output cell buffer if it has been previously
               allocated.  Not in original code.  */
            for (i = 0; i < tile_count; i++)
            {
                free(tile_names[i]);
            }
            free(tile_names);
            return ERROR;
        }

        /* When processing the first tile, populate the output header info
           and allocate memory for the elevation data */
        if (tile == 0)
        {
            /* Set the data source */
            elevation_cell->source = GTOPO30_SOURCE;

            /* Set up the output cell header values */
            elevation_cell->xdim         = current_cell.xdim;
            elevation_cell->ydim         = current_cell.ydim;
            elevation_cell->nodata       = current_cell.nodata;
            elevation_cell->nbands       = DMS_DRE_DEM_NBANDS;
            elevation_cell->nbits        = DMS_DRE_DEM_NBITS;
            elevation_cell->bandgapbytes = current_cell.bandgapbytes;

            /* Set the upper-left corner to a multiple of the pixel size */
            elevation_cell->ulxmap
               = floor(ullon / elevation_cell->xdim) * elevation_cell->xdim;
            elevation_cell->ulxmap -= (elevation_cell->xdim / 2.0);
            
            elevation_cell->ulymap
                = ceil(ullat / elevation_cell->ydim) * elevation_cell->ydim;
            elevation_cell->ulymap += (elevation_cell->ydim / 2.0);

            /* Set the number of rows and columns in the DEM image */
            elevation_cell->nrows
                = ceil(((elevation_cell->ulymap - lrlat)
                     / elevation_cell->ydim) + 1.5);

            elevation_cell->ncols
                = ceil((fmodulus((lrlon - elevation_cell->ulxmap), 360)
                    / elevation_cell->xdim) + 1.5);

            /* Set the data buffer size information */
            elevation_cell->bandrowbytes
                = elevation_cell->ncols * (elevation_cell->nbits / 8);

            elevation_cell->totalrowbytes
                = elevation_cell->bandrowbytes * elevation_cell->nbands;

            num_elevation_pts
                = elevation_cell->nrows * elevation_cell->ncols;

            /* Now... allocate the output cell data buffer */
            elevation_cell->data
                = malloc(num_elevation_pts * elev_data_size);
            if (elevation_cell->data == NULL)
            {
                IAS_LOG_ERROR("Allocating elevation cell data buffer");
                for (i = 0; i < tile_count; i++)
                {
                    free(tile_names[i]);
                }
                free(tile_names);
                return ERROR;
            }

            /* Initialize the output cell to an illegal value */
            for (i = 0; i < num_elevation_pts; i++)
            {
                elevation_cell->data[i] = ILLEGAL_ELEV;
            }
        }  /* END -- IF block to process first tile */

        
        /* Find the intersection of the current cell and the output cell */
        get_cell_intersection(&current_cell, elevation_cell, &ic_ul_line,
                              &ic_ul_sample, &oc_ul_line, &oc_ul_sample,
                              &lines, &samples);

        /* Extract the required data from the current elevation cell and
           copy it into the output elevation cell */
        if ((lines > 0) && (samples > 0))
        {
            for (line = 0; line < lines; line++)
            {
                oc_index = (line + oc_ul_line) * elevation_cell->ncols;
                ic_index = (line + ic_ul_line) * current_cell.ncols;

                memcpy(&elevation_cell->data[oc_index + oc_ul_sample],
                       &current_cell.data[ic_index + ic_ul_sample], 
                       samples * elev_data_size);
            }
        }

        /* Trace data from each input tile */
        IAS_LOG_DEBUG("nrows           = %d\nncols           = %d\n"
                      "nbands          = %d\nnbits           = %d\n"
                      "band row bytes  = %d\ntotal row bytes = %d\n"
                      "bandgap bytes   = %d\nul xmap         = %lf\n"
                      "ul ymap         = %lf\nxdim            = %lf\n"
                      "ydim            = %lf\n", current_cell.nrows,
                      current_cell.ncols, current_cell.nbands,
                      current_cell.nbits, current_cell.bandrowbytes,
                      current_cell.totalrowbytes,
                      current_cell.bandgapbytes, current_cell.ulxmap,
                      current_cell.ulymap, current_cell.xdim,
                      current_cell.ydim);

        /* free the current read cell */
        free(current_cell.data);
        current_cell.data = NULL;

    } /* END -- Tile processing loop */


    /* Free the tile names list buffer.  Set everything to NULL so we
       find unexpected attempts to access this array (via system crash) */
    for (i = 0; i < tile_count; i++)
    {
        free(tile_names[i]);
    }
    free(tile_names);

    /* Make sure there are no remaining illegal values in the output data
       (to make sure the entire area has been filled)*/
    num_elevation_pts = elevation_cell->nrows * elevation_cell->ncols;
    for (i = 0; i < num_elevation_pts; i++)
    {
        if (elevation_cell->data[i] == ILLEGAL_ELEV)
        {
            IAS_LOG_ERROR("Missing data in output cell, offset %d",
                           i+1);
            free(elevation_cell->data);  /* Avoid possible memory leak */
            return ERROR;
        }
    }

    /* For the elevation values that are no-data, change them to zero since
       GTOPO30 no-data values are ocean areas */
    nodata = elevation_cell->nodata;
    data_ptr = elevation_cell->data;
    for (i = 0; i < num_elevation_pts; i++)
    {
        if (data_ptr[i] == nodata)
        {
            data_ptr[i] = 0;
        }
    }

    /* Adjust the elevation values to be relative to the WGS84 ellipsoid
       instead of mean sea level */
    snprintf(dem_dir, sizeof(dem_dir), "%s/%s", dem_root_dir, "geoid");

    status = adjust_elev_to_wgs84(dem_dir, elevation_cell);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Adjusting elevation values to the WGS84 ellipsoid");
        free(elevation_cell->data);
        return ERROR;
    }

    /* Print out output DEM header information in debug mode */
    IAS_LOG_DEBUG("nrows           = %d\nncols           = %d\n"
                  "nbands          = %d\nnbits           = %d\n"
                  "band row bytes  = %d\ntotal row bytes = %d\n"
                  "bandgap bytes   = %d\nul xmap         = %lf\n"
                  "ul ymap         = %lf\nxdim            = %lf\n"
                  "ydim            = %lf\n", elevation_cell->nrows,
                  elevation_cell->ncols, elevation_cell->nbands,
                  elevation_cell->nbits, elevation_cell->bandrowbytes,
                  elevation_cell->totalrowbytes,
                  elevation_cell->bandgapbytes, elevation_cell->ulxmap,
                  elevation_cell->ulymap, elevation_cell->xdim,
                  elevation_cell->ydim);

    /* Done */
    return SUCCESS;
}
