/*****************************************************************************
Name: get_gtopo30_tile_list

Description: Function to determine which GTOPO30 tiles intersect the indicated
    latitute/longitude bounding box and return the list of tiles.

Returns:
    NULL on error
    Otherwise, the pointer to the list of tile names and the count of tiles

*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ias_logging.h"

#include "retrieve_elevation.h"
#include "get_gtopo30_tile_list.h"
#include "modulus.h"

#define GTOPO30_TILE_SET_CUTOFF_LATITUDE -60.0

double LON_LOCATIONS_NORTH[MAX_LON_TILES_NORTH]
    = {-180.0, -140.0, -100.0, -60.0, -20.0, 20.0, 60.0, 100.0, 140.0};
double LON_LOCATIONS_SOUTH[MAX_LON_TILES_SOUTH]
    = {-180.0, -120.0, -60.0, 0.0, 60.0, 120.0};

char **get_gtopo30_tile_list
(
    double ullat,       /* I: upper-left latitude of bounding box */
    double ullon,       /* I: upper-left longitude of bounding box */
    double lrlat,       /* I: lower-right latitude of bounding box */
    double lrlon,       /* I: lower-right longitude of bounding box */
    int *tile_count     /* O: count of tiles in returned list */
)
{
    char n_s;                     /* "n" or "s" */
    char e_w;                     /* "e" or "w" */
    char name_buffer[GTOPO_NAME_BUFFER_LENGTH];
    char *tile_names[NUM_GTOPO30_TILES];       /* 33 GTOPO30 tiles */
    char **returned_names = NULL;

    double curr_lat = ullat;
    double *lon_locations = NULL; /* Either LON_LOCATIONS_SOUTH
                                     LON_LON_TILES_NORTH  */
    double next_lat;

    int max_lon_tiles;            /* Either MAX_LON_TILES_SOUTH or
                                     MAX_LON_TILES_NORTH */
    int tile_lon;
    int tile_lon_west_idx = 0;    /* Western index in lon_locations array */
    int tile_lon_east_idx = 0;    /* Eastern index in lon_locations array */
    int tile_lat;

    int tiles = 0;                /* Loop counters */
    int i;



    /* The tiles do not necessarily go all the way to the boundary.  For
     * example, the w100n90 tile only goes to -99.995833333334 longitude.
     * Thus, if our UL longitude falls between -99.995833333334 and -100.0 then
     * we will need to use w140n90 to get the far left edge.  To account for
     * this we'll artificially pad the East and West boundaries by 1 degree.
     *
     * Note: this does not seem to cause problems on the North and South
     * borders, so we will no worry about them.
     */
    ullon = longitude_norm(ullon - 1.0);
    lrlon = longitude_norm(lrlon + 1.0);

    /* There is a condition which could cause problems on the North and South
     * boundaries as well.  This is if a scene comes very near the edge, but
     * does not quite get there.  It could be that adding 1.5 to nrows before
     * taking the ceiling (see calculation of nrows in get_gtopo30_data.c)
     * causes the nrows to be large enough that we need data from the tile to
     * the South, even though the scene stopped just before that tile.  To
     * address this we will grow North and South by 1 degree as we did in the
     * East and West directions above.  Note, the same problem can occur in
     * the East and West directions, but was already addressed when adding one
     * to the longitudes above.
     */
    ullat = IAS_MIN((ullat + 1.0), NORTH_LATITUDE_LIMIT);
    lrlat = IAS_MAX((lrlat - 1.0), SOUTH_LATITUDE_LIMIT);

    if (curr_lat >= NORTH_LATITUDE_LIMIT)
    {
        curr_lat = 89.9;
    }

    while (curr_lat >= lrlat)
    {
        if (curr_lat > GTOPO30_TILE_SET_CUTOFF_LATITUDE)
        {
            tile_lat = ((int)((curr_lat + 60.0)/50.0)) * 50 - 10;
            next_lat = (double)tile_lat - 50.000001;
            if (next_lat < GTOPO30_TILE_SET_CUTOFF_LATITUDE)
                next_lat = GTOPO30_TILE_SET_CUTOFF_LATITUDE;
        }
        else
        {
            tile_lat = GTOPO30_TILE_SET_CUTOFF_LATITUDE;
            next_lat = -91.0;
        }

        if (tile_lat < 0)
        {
            n_s = 's';
            tile_lat = -tile_lat;
        }
        else
        {
            n_s = 'n';
        }

        /* determine if we are dealing with the Northern or Southern sets of
         * tiles */
        if (curr_lat > GTOPO30_TILE_SET_CUTOFF_LATITUDE)
        {
            max_lon_tiles = MAX_LON_TILES_NORTH;
            lon_locations  = LON_LOCATIONS_NORTH;
        }
        else 
        {
            max_lon_tiles = MAX_LON_TILES_SOUTH;
            lon_locations  = LON_LOCATIONS_SOUTH;
        }

        /* determine what tiles will be needed in the East-West direction */
        for ( i = 0; i < max_lon_tiles; i++ )
        {
            if ((double)lon_locations[i] <= ullon)
            {
                tile_lon_west_idx = i;
            }
            if ((double)lon_locations[i] <= lrlon)
            {
                tile_lon_east_idx = i;
            }
        }
        if (tile_lon_east_idx < tile_lon_west_idx)
        {
            tile_lon_east_idx += max_lon_tiles;
        }

        /* for each tile in East-West direction: record the name */
        for (i = tile_lon_west_idx; i <= tile_lon_east_idx; i++)
        {
            tile_lon = lon_locations[i % max_lon_tiles];
            if (tile_lon <= 0)
            {
                tile_lon = -tile_lon;
                e_w = 'w';
            }
            else
            {
                e_w = 'e';
            }
           
            /* Build the tile name and copy it to the secondary
               name list */
            snprintf(name_buffer, sizeof(name_buffer), "%c%03d%c%02d",
                     e_w, tile_lon, n_s, tile_lat);

            tile_names[tiles] = strdup(name_buffer);
            if (!tile_names[tiles])
            {
                /* error allocating memory for a new tile name, so free all
                   the ones allocated so far and return an error */
                for (i = 0; i < tiles; i++)
                {
                    free(tile_names[i]);
                }
                IAS_LOG_ERROR("Cannot allocate working string memory for "
                              "tile name");
                return NULL;
            }
            tiles++;
        }

        curr_lat = next_lat;
    }

    /* copy the working tile name list to the final returned list */
    returned_names = malloc(sizeof(char *) * tiles);
    if (returned_names == NULL)
    {
        /* error allocating memory for the name buffer, so free the names
           and return an error */
        for (i = 0; i < tiles; i++)
        {
            free(tile_names[i]);
        }
        IAS_LOG_ERROR("Error allocating final name list buffer\n");
        return NULL;
    }
    for (i = 0; i < tiles; i++)
    {
        returned_names[i] = tile_names[i];
    }

    /* return the count of tiles found and the names of the tiles */
    *tile_count = tiles;

    return returned_names;
}
