/***************************************************************************
 NAME:                         get_parameters

 PURPOSE:   Reads the work order parameter file to get the work order
            ID/directory and other needed processing parameters.
            Then reads the OMF file to get the scene bounding box 
            coordinates

 RETURNS:   Integer status code of SUCCESS or ERROR

****************************************************************************/
#include <string.h>

#include "ias_const.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "ias_logging.h"

#include "retrieve_elevation.h"
#include "get_parameters.h"

int get_parameters
(
    const char *wo_params_filename,      /* I: Name of WO parameter
                                            file */
    DMS_DRE_WO_PARAMS *wo_param_data,    /* O: Populated work
                                            parameters */
    DMS_DRE_OMF_PARAMS *omf_param_data   /* O: Populated OMF
                                            parameters */
)
{
    int read_count;         /* Parameter table size */
    int status;             /* Read function return status */

                                             /* default DEM_FILENAME */
    const char *default_filename[] = {"source_dem.h5"};

    char omf_params_filename[PATH_MAX];

    /* As used in setup_work_order.  The parameter order is expected to
       remain identical (i.e. [0]=latitude, [1]=longitude) */
    double default_lat_lons[2] = {0.0, 0.0};

    /* Set up and read the work order parameter table */
    IAS_PARM_DECLARE_TABLE(wo_param_table, MAX_NUM_WO_PARAMS);

    /* Common parameters first */
    IAS_PARM_WORK_ORDER_ID(wo_param_table, wo_param_data->work_order_id,
                           sizeof(wo_param_data->work_order_id), 0);

    IAS_PARM_ADD_STRING(wo_param_table, WO_DIRECTORY,
                     "IAS Work Order Directory", IAS_PARM_REQUIRED,
                     0, NULL, 0, NULL, wo_param_data->work_order_directory,
                     sizeof(wo_param_data->work_order_directory), 0);

    /* Get the name of the DEM file */
    IAS_PARM_ADD_STRING(wo_param_table, DEM_FILENAME,
                        "Name of output DEM file", IAS_PARM_OPTIONAL,
                        0, NULL, 1, default_filename,
                        wo_param_data->dem_filename,
                        sizeof(wo_param_data->dem_filename), 1);

    /* If help was requested, provide it.  Copy the WO parameter file
       name to the OMF parameter file name, so the help request is 
       processed in the same way */
    read_count = IAS_PARM_GET_TABLE_SIZE(wo_param_table);

    if (ias_parm_provide_help(wo_params_filename, wo_param_table,
                              read_count, IAS_INPUT_PARAMETERS))
    {
        sprintf(omf_params_filename, "%s", wo_params_filename);
    }
    else
    {
        status = ias_parm_read(wo_params_filename, wo_param_table,
                               read_count);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading parameters from file %s",
                    wo_params_filename);
            return ERROR;
        }

        /* Construct the OMF parameter file name */
        sprintf(omf_params_filename, "%s.omf", wo_param_data->work_order_id);
    }


    /* Now set up and read the OMF parameter table */
    IAS_PARM_DECLARE_TABLE(omf_param_table, MAX_NUM_OMF_PARAMS);

    /* Read the boundary information */
    IAS_PARM_ADD_DOUBLE(omf_param_table, UL_BOUNDARY_LAT_LON,
        "Upper left corner lat/lon coordinate array ([0]=lat, [1]=lon)",
        IAS_PARM_REQUIRED, IAS_PARM_ARRAY, 1, -180.0, 180.0, 1,
        default_lat_lons, (double *)&(omf_param_data->upper_left_lat_lon),
        sizeof(omf_param_data->upper_left_lat_lon), 2);

    IAS_PARM_ADD_DOUBLE(omf_param_table, UR_BOUNDARY_LAT_LON,
        "Upper right corner lat/lon coordinate array ([0]=lat, [1]=lon)",
        IAS_PARM_REQUIRED, IAS_PARM_ARRAY, 1, -180.0, 180.0, 1,
        default_lat_lons, (double *)&(omf_param_data->upper_right_lat_lon),
        sizeof(omf_param_data->upper_right_lat_lon), 2);

    IAS_PARM_ADD_DOUBLE(omf_param_table, LL_BOUNDARY_LAT_LON,
        "Lower left corner lat/lon coordinate array ([0]=lat, [1]=lon)",
        IAS_PARM_REQUIRED, IAS_PARM_ARRAY, 1, -180.0, 180.0, 1,
        default_lat_lons, (double *)&(omf_param_data->lower_left_lat_lon),
        sizeof(omf_param_data->lower_left_lat_lon), 2);

    IAS_PARM_ADD_DOUBLE(omf_param_table, LR_BOUNDARY_LAT_LON,
        "Lower right corner lat/lon coordinate array ([0]=lat, [1]=lon)",
        IAS_PARM_REQUIRED, IAS_PARM_ARRAY, 1, -180.0, 180.0, 1,
        default_lat_lons, (double *)&(omf_param_data->lower_right_lat_lon),
        sizeof(omf_param_data->lower_right_lat_lon), 2);


    /* Get the total number of items to read from the OMF parameter table */
    read_count  = IAS_PARM_GET_TABLE_SIZE(omf_param_table);

    /* See if the user requested help information.  If so, quit the
       application, otherwise read the OMF parameter file */
    status = ias_parm_provide_help(omf_params_filename, omf_param_table,
                   read_count, IAS_OMF_PARAMETERS);
    if (status)
    {
        if (status == ERROR)
            return ERROR;
        else
            return HELP_SUCCESS;
    }

    status = ias_parm_read(omf_params_filename, omf_param_table, read_count);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading OMF parameter file");
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}

