/*****************************************************************************
 NAME:                    get_ramp_data

 PURPOSE:    Generates a symbolic link to the RAMP DEM data file that is
             placed in the work order directory.  Also reads the
             DEM band metadata to determine the elevation range

 RETURNS:    Integer status code of SUCCESS or ERROR

 NOTES:      This is an "extended" version of the Landsat 7 ETM+ IAS/
             ALIAS get_ramp_data routine.  The biggest extension is
             the additional code to read an HDF5-formatted DEM data file
             (using the new IAS L1G I/O library functions) and its
             associated metadata to return the elevation range.

*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>

#include "ias_logging.h"
#include "ias_l1g.h"

#include "retrieve_elevation.h"
#include "get_ramp_data.h"
#include "find_minmax_elevation.h"
#include "round.h"

int get_ramp_data
(
    const char *ramp_filename, /* I: Full path to RAMP DEM data file */
    const char *wo_dir,      /* I: Target work order directory */
    const char *wo_dem_name, /* I: Name of DEM data file in the WO directory */
    short *max_elevation,    /* O: Maximum RAMP DEM elevation */
    short *min_elevation     /* O: Minimum RAMP DEM elevation */
)
{
    char link_name[PATH_MAX];

    int status;                           /* Function status return code */
    int band_number = 1;                  /* Band to open */

    L1GIO *l1g_img = NULL;                /* L1G image data handle */
    IAS_L1G_BAND_METADATA band_metadata;  /* Band metadata data structure */

    /* We have the (expected) name of the DEM file and its directory
       path.  Do a "test" open.  If this fails, return an ERROR--we
       can't say we've found the file */

    l1g_img = ias_l1g_open_image(ramp_filename, IAS_READ);
    if (l1g_img == NULL)
    {
        IAS_LOG_ERROR("Opening DEM data file name %s", ramp_filename);
        return ERROR;
    }

    /* Initialize the metadata structures */
    memset(&band_metadata, 0, sizeof(IAS_L1G_BAND_METADATA));

    /* Read the band and file metadata, then close the image handle */
    status = ias_l1g_get_band_metadata(l1g_img, band_number, &band_metadata);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading band metadata for DEM dataset");
        ias_l1g_close_image(l1g_img);
        return ERROR;
    }

    status = ias_l1g_close_image(l1g_img);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Closing DEM data file %s", ramp_filename);
        return ERROR;
    }

    /* Pick out the elevation range.  First, check to see whether the
       recorded values are even valid */
    if (!band_metadata.pixel_range_valid)
    {
        IAS_LOG_ERROR("Elevation values are not considered valid");
        return ERROR;
    }
    *max_elevation = (short)round_value(band_metadata.maximum_pixel_value);
    *min_elevation = (short)round_value(band_metadata.minimum_pixel_value);

    /* We have a valid ramp DEM data file, and we've gotten the elevation
       data out of it.  Now create a symbolic link to the DEM file name in
       the work order directory */

    /* Make the name of the symbolic link to be the name of the ramp DEM file */
    sprintf(link_name, "%s/%s", wo_dir, wo_dem_name);

    /* Attempt to create the link.  If we can't, symlink will set errno to
       the value representing the reason we can't.  There's a lot of
       error conditions, so we need to go through the list */
    if (symlink(ramp_filename, link_name)) /* Return of 0 is SUCCESS,
                                              anything else is an ERROR */
    {
        switch (errno)
        {
            case EACCES:
                IAS_LOG_ERROR("Write access to %s denied or path component "
                              "isn't searchable", wo_dir);
                break;
            case EEXIST:
                IAS_LOG_ERROR("Link already exists");
                break;
            case EFAULT:
                IAS_LOG_ERROR("%s and/or %s point outiside of accessible "
                        "address space", ramp_filename, wo_dir);
                break;
            case EIO:
                IAS_LOG_ERROR("I/O error");
                break;
            case ENAMETOOLONG:
                IAS_LOG_ERROR("%s and/or %s is too long for a symbolic link",
                        ramp_filename, wo_dir);
                break;
            case ENOENT:
                IAS_LOG_ERROR("Component in %s doesn't exist or is dangling, "
                        "or %s is empty string", wo_dir, ramp_filename);
                break;
            case ENOMEM:
                IAS_LOG_ERROR("Insufficient kernel memory available to "
                              "create symbolic link");
                break;
            case ENOSPC:
                IAS_LOG_ERROR("Not enough disk space to create symbolic "
                              "link");
                break;
            case ENOTDIR:
                IAS_LOG_ERROR("Component in %s is not a valid directory",
                        wo_dir);
                break;
            case EPERM:
                IAS_LOG_ERROR("Filesystem containing %s doesn't support "
                        "creation of symbolic links", wo_dir);
                break;
            case EROFS:
                IAS_LOG_ERROR("%s is on a read-only filesystem", wo_dir);
                break;
            default:
                IAS_LOG_ERROR("Unidentified error in symlink() call");
                break;
        }

        return ERROR;
    }

    /* Done */
    return SUCCESS;
}
