/***************************************************************************
 NAME:               GTOPO30

 PURPOSE:    Implements routines to load GTOPO30 header and data files.

****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>               /* toupper() */
#include <limits.h>              /* PATH_MAX #define */
#include <math.h>
#include <unistd.h>

#include "ias_const.h"
#include "ias_logging.h"

#include "retrieve_elevation.h"
#include "gtopo30.h"




/*------------------------------------------------------------------------
 NAME:                load_gtopo

 PURPOSE:   Loads data from a GTOPO30 DEM tile into memory

 RETURNS:   Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int load_gtopo
(
    const char *dem_dir,               /* I: DEM data root directory */
    const char *cell_name,             /* I: Base name of GTOPO30 tile */
    GTOPO30_CELL *cell                 /* O: Fully populated GTOPO30_CELL
                                          data structure */
)
{
    char gtopodir[PATH_MAX];
    char command[PATH_MAX * 2];
    char hfile[GTOPO_NAME_SIZE];
    char dfile[GTOPO_NAME_SIZE];

    /* Construct the desired header and DEM file names.
       get_gtopo30_tile_list() returns a list of tile names in the form
       
                 e(w)LLLn(s)ll,       where 
       
       - e(w) indicate east(west) longitude (with leading 0 if needed)
       - LLL is the longitude value for the tile
       - n(s) indicate north(south) latitude
       - ll is the latitude value for the tile */
    strcpy(hfile, cell_name);
    hfile[0] = (char)toupper(cell_name[0]);  /* Convert 'e', 'w' to
                                                capital letters */
    hfile[4] = (char)toupper(cell_name[4]);  /* Convert 'n', 's' to
                                                captial letters */
    strcpy(dfile, hfile);
    strcat(hfile, ".HDR");
    strcat(dfile, ".DEM");

    IAS_LOG_DEBUG("Header file name:\t%s\n", hfile);
    IAS_LOG_DEBUG("DEM file name:\t\t%s\n", dfile);

    /* Construct the GTOPO data directory name */
    strcpy(gtopodir, dem_dir);
    strcat(gtopodir, "/");

    /* Decompress and and extract the needed (compressed) data files */
    snprintf(command, sizeof(command), "/bin/tar zxf %s%s.tar.gz %s %s",
             gtopodir, cell_name, dfile, hfile);
    if (system(command) != 0)
    {
        IAS_LOG_ERROR("Extracting cell %s", cell_name);
        return ERROR;
    }


    /* Read the header file */
    if (read_gtopohdr(hfile, cell) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading header file %s", hfile);
        return ERROR;
    }

    /* Read the data (elevation) file */
    if (read_gtopodem(dfile, cell) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading cell elevations from data file %s",
                      dfile);
        return ERROR;
    }

    /* Cleanup of GTOPO header and data files is automatic, in order to save
       space */
    if (unlink(hfile))
    {
        IAS_LOG_ERROR("Deleting tile header file %s", hfile);
        return ERROR;
    }
    if (unlink(dfile))
    {
        IAS_LOG_ERROR("Deleting tile data file   %s", dfile);
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}




/*------------------------------------------------------------------------
 NAME:                   read_gtopohdr

 PURPOSE:   Reads a GTOPO30 DEM formatted header file and copies the
            information to the corresponding members of a GTOPO30_CELL
            data structure

 RETURNS:   Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int read_gtopohdr
(
    const char *fname,              /* I: header file name */
    GTOPO30_CELL *cell              /* O: Populated data structure */
)
{
    FILE *fd;                             /* Header file "handle" */
    char buffer[GTOPO_HDR_BUFFER_SIZE];   /* Line buffer */
    int records_parsed = 0;               /* Header record counter */

    /* Open the header file */
    if ((fd = fopen(fname, "rt")) == NULL)
    {
        IAS_LOG_ERROR("Opening header file %s", fname);
        return ERROR;
    }
    
    /* Read until EOF encountered, or the current line doesn't read
       properly */
    int status;
    while (!feof(fd))
    {
        if (fgets(buffer, GTOPO_HDR_BUFFER_SIZE, fd) == NULL)
        {
            IAS_LOG_DEBUG("End of header file reached");
            break;
        }
        
        status = -1;

        if (!strncmp(buffer, "BYTEORDER", 9))
        {
            status = sscanf(&buffer[9], "%3s", cell->byteorder);
            cell->byteorder[3] = '\0';
        }
        else if (!strncmp(buffer, "LAYOUT", 6))
        {
            status = sscanf(&buffer[6], "%3s", cell->layout);
            cell->layout[3] = '\0';
        }
        else if (!strncmp(buffer, "NROWS", 5))
        {
            status = sscanf(&buffer[5], "%d", &cell->nrows);
        }
        else if (!strncmp(buffer, "NCOLS", 5)) 
        {
            status = sscanf(&buffer[5], "%d", &cell->ncols);
        }
        else if (!strncmp(buffer, "NBANDS", 6))
        {
            status = sscanf(&buffer[6], "%d", &cell->nbands);
        }
        else if (!strncmp(buffer, "NBITS", 5))
        {
            status = sscanf(&buffer[5], "%d", &cell->nbits);
        }
        else if (!strncmp(buffer, "BANDROWBYTES", 12)) 
        {
            status = sscanf(&buffer[12], "%d", &cell->bandrowbytes);
        }
        else if (!strncmp(buffer, "TOTALROWBYTES", 13))
        {
            status = sscanf(&buffer[13], "%d", &cell->totalrowbytes);
        }
        else if (!strncmp(buffer, "BANDGAPBYTES", 12))
        {
            status = sscanf(&buffer[12], "%d", &cell->bandgapbytes);
        }
        else if (!strncmp(buffer, "NODATA", 6))
        {
            status = sscanf(&buffer[6], "%hd", &cell->nodata);
        }
        else if (!strncmp(buffer, "ULXMAP", 6))
        {
            status = sscanf(&buffer[6], "%lf", &cell->ulxmap);
        }
        else if (!strncmp(buffer, "ULYMAP", 6))
        {
            status = sscanf(&buffer[6], "%lf", &cell->ulymap);
        }
        else if (!strncmp(buffer, "XDIM", 4))
        {
            status = sscanf(&buffer[4], "%lf", &cell->xdim);
        }
        else if (!strncmp(buffer, "YDIM", 4))
        {
            status = sscanf(&buffer[4], "%lf", &cell->ydim);
        }
        else
        {
            IAS_LOG_ERROR("Invalid record in header file\n:%s:",
                          buffer);
        }

        /* detect if a record was not parsed successfully */
        if (status != 1)
        {
            if (status == 0)
            {
                IAS_LOG_ERROR("Parsing record in header file: %s\n",
                              buffer);
            }
            fclose(fd);
            return ERROR;
        }

        /* Count a successfully read record */
        records_parsed++;
    }
    fclose(fd);

    /* If the correct number of records wasn't parsed, it's an error */
    if (records_parsed != GTOPO_NUM_HDR_RECORDS)
    {
        IAS_LOG_ERROR("Incorrect number of header records parsed:  %d",
                      records_parsed);
        return ERROR;
    }

    /* Validate a couple of fields from the header file to confirm they
       are supported values (or at least read correctly) */
    if (strcmp(cell->layout,"BIL") != 0) 
    {
        IAS_LOG_ERROR("Unsupported LAYOUT value %s", cell->layout);
        return ERROR;
    }
    if ((strcmp(cell->byteorder, "I") != 0) 
        && (strcmp(cell->byteorder, "M") != 0))
    {
        IAS_LOG_ERROR("Unsupported BYTEORDER value %s",
                      cell->byteorder);
        return ERROR;
    }
    if (cell->bandgapbytes != 0)
    {
        IAS_LOG_ERROR("Invalid BANDGAPBYTES value %d, should be 0",
                      cell->bandgapbytes);
        return ERROR;
    }

    return SUCCESS;
}


/*------------------------------------------------------------------------
 NAME:                   read_gtopodem

 PURPOSE:   Reads a GTOPO30 DEM data file and copies the
            information to the corresponding members of a GTOPO30_CELL
            data structure

 RETURNS:   Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int read_gtopodem
(
    const char *fname,              /* I: data file name */
    GTOPO30_CELL *cell              /* O: Populated data structure */
)
{
    int num_elevation_pts;          /* Number of elevation data points */
    FILE *fd;                       /* Data file "handle" */


    /* See how much data we need to read */
    num_elevation_pts = cell->nrows * cell->ncols;
    if (num_elevation_pts <= 0)
    {
        IAS_LOG_ERROR("Invalid DEM array size %d", num_elevation_pts);
        return ERROR;
    }

    /* Allocate the necessary memory.  Use malloc rather than calloc
       because the data buffer will shortly be filled */
    cell->data = malloc(num_elevation_pts * sizeof(*(cell->data)));
    if (cell->data == NULL)
    {
        IAS_LOG_ERROR("Allocating DEM data buffer");
        return ERROR;
    }

    /* Open the data file */
    if ((fd = fopen(fname, "rb")) == NULL)
    {
        IAS_LOG_ERROR("Opening DEM data file %s", fname);
        free(cell->data);
        return ERROR;
    }

    /* Read the data */
    if (fread(cell->data, sizeof(*(cell->data)), num_elevation_pts, fd)
        != num_elevation_pts)
    {
        IAS_LOG_ERROR("Reading DEM data");
        fclose(fd);
        free(cell->data);
        return ERROR;
    }

    fclose(fd);

#ifdef LITTLE_ENDIAN_ORDER
    /* On "little-endian" machines, swap the byte order if the source byte
       order is "M" (Motorola) so the data is converted from big-endian
       order to little-endian order */
    if (cell->byteorder[0] == 'M')
    {
        int i;
        for (i = 0; i < num_elevation_pts; i++)
        {
            SWAP_BYTES_16(cell->data[i]);
        }
    }
#else
    /* On "big-endian" machines, swap the byte order if the source byte
       order is "I" (Intel) so the data is converted from little-endian 
       order to big-endian order */
    if (cell->byteorder[0] == 'I')
    {
        int i;
        for (i = 0; i < num_elevation_pts; i++)
        {
            SWAP_BYTES_16(cell->data[i]);
        }
    }
#endif

    /* Done */
    return SUCCESS;
}
