#ifndef GTOPO30_H
#define GTOPO30_H

#include "../../config.h"

/* Various constants */
#define GTOPO30_SOURCE 1
#define GLSDEM_SOURCE  2

/* define a macro to swap 16-bit integers if needed */
#ifdef HAVE_LITTLE_ENDIAN
#define LITTLE_ENDIAN_ORDER
/* Intel style processors are little-endian, so swap bytes since gtopo
   files are written in big-endian order */
#define SWAP_BYTES_16(x) \
{ \
    unsigned short int data_macro = *(unsigned short int*)&(x); \
    data_macro = ((data_macro & 0xff00) >> 8) |    \
    ((data_macro & 0x00ff) << 8);     \
    *(unsigned short int*)&(x) = data_macro;       \
}
#else
/* other common processors are big-endian, so no swapping needed */
#define SWAP_BYTES_16(x)
#endif


/* Various constants */
#define GTOPO_HDR_BUFFER_SIZE 80
#define GTOPO_NAME_SIZE       16
#define GTOPO_NUM_HDR_RECORDS 14


/* GTOPO30 Cell Data Structure.  This is likely based on most of the
   information contained in the GTOPO30 tile header file (along with
   a data buffer to hold the actual DEM data).  */
typedef struct gtopo30_cell 
{
    char byteorder[4];             /* Order in which bytes are read/written
                                      (MSB, LSB) or (LSB, MSB), where
                                      MSB=most significant byte
                                      LSB=least significant byte
                                      It's critical this is correctly
                                      established -- otherwise, bad
                                      things can happen... */
    char layout[4];                /* How the lines of DEM data for each
                                      band are arranged -- "BSQ" (band
                                      sequential), "BIL" (band interleaved
                                      by line), or "BIP" (band interleaved
                                      by pixel) are possible values */
    unsigned int nrows;            /* Number of lines of DEM data */
    unsigned int ncols;            /* Number of samples / line */
    unsigned int nbands;           /* # of bands in DEM image (1) */
    unsigned int nbits;            /* # of bytes representing each data
                                      value (multiples of 8) */
    unsigned int bandrowbytes;     /* ncols * nbits *(1 byte/8 bits) */
    unsigned int totalrowbytes;    /* Generally same as bandrowbytes */
    unsigned int bandgapbytes;     /* # of bytes spacing between bands */
    short nodata;                  /* Flag value indicating pixel does not
                                      have valid elevation data associated
                                      with it */
    double ulxmap;                 /* Upper-left corner x coordinate for
                                      the given map projection */ 
    double ulymap;                 /* Upper-left corner y coordinate for
                                      given map projection */
    double xdim;                   /* x-direction pixel size */
    double ydim;                   /* y-direction pixel size */
    int source;                    /* DEM data "source" (i.e. GTOPO30 or
                                      GLS) */
    short *data;                   /* DEM data buffer */
} GTOPO30_CELL;


/* Load from disk a specified GTOPO30 cell to a data structure */
int load_gtopo
(
    const char *dem_dir,               /* I: DEM data root directory */
    const char *cell_name,             /* I: Base name of GTOPO30 tile */
    GTOPO30_CELL *cell                 /* O: Fully populated GTOPO30_CELL
                                          data structure */
);


/* Read the GTOPO30 header file */
int read_gtopohdr
(
    const char *fname,              /* I: header file name */
    GTOPO30_CELL *cell              /* O: Populated data structure */
);


/* Read the GTOPO30 elevation (DEM) file */
int read_gtopodem
(
    const char *fname,              /* I: data file name */
    GTOPO30_CELL *cell              /* O: Populated data structure */
);


#endif

