/******************************************************************************
NAME:           hdf4_l1g_hdf5

PURPOSE:
Converts ALIAS HDF4 formated image to OLI HDF5 formated image.  This only works
on the xxx ALIAS HDF4 libraries, i.e., ti does not convert SCA separated image
formats between HDF4 and HDF5.  This routine was setup mostly to convert the
ALIAS DEMs and DOQs.

USAGE:  hdf4_l1g_hdf5 <HDF 4> <HDF 5>

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
FAILURE  Operation failed

NOTES:

ALGORITHM REFERENCES:
See ias_l1g.h for a description of the grid structure.

******************************************************************************/
#include "hdf4_l1g_hdf5.h"

int main(int argc, char *argv[])
{
   char progname[STRLEN];       /* Module name */
   char msg[STRLEN];            /* Messaging string */
   IAS_L1G_BAND_METADATA *bmd;  /* L1g band meta data */
   IAS_L1G_FILE_METADATA fmd;   /* L1g file meta data */
   L1GIO *img5;                 /* L1G image pointer */
   L1G_BAND_IO *l1g_band;       /* file descriptor for the file */
   IAS_DATA_TYPE dtype5;        /* data type of each band */
   int band_nsca;               /* number of scas in the band */
   int nl[NBANDS];              /* number of lines per band */
   int ns[NBANDS];              /* number of samples per band */
   int bb,nn;                   /* Counters */
   xxx_DDRControl_TYPE *ddr_control;    /* 1G image's DDR */
   xxx_DDR_TYPE ddr;            /* ith DDR structure */
   char hdf4_name[STRLEN];      /* Input HDF4 file name */
   char hdf5_name[STRLEN];      /* Output HDF5 file name */
   char band_name[STRLEN];      /* Band metadata name */
   char ddr_name[STRLEN];       /* HDF4 DDR file name */
   HDFIO *img4;                 /* HDF4 image pointer */
   int nddrs;                   /* Number of bands stored in HDF4 DDR */
   int l0ra_band_lut[] = {8,1,2,3,4,5,6,7,9};
   /* Not exactly sure these l1g band look-ups are correct */
   int l1g_band_lut[] = {2,1,5,4,3,7,6,0,8};
   int nbands = 1;
   DataType dtype4[NBANDS];     /* data type for HDF4 imagery */
   BandNumber bandno[NBANDS];   /* HDF4 band enumerated type */
   int size;                    /* Image size */
   int line;                    /* Counter */

   void    *ibuf_g   = NULL;    /* generic image buffer */
   uint8   *ibuf_i8  = NULL;    /* buffer depending on the data type of */
   int16   *ibuf_i16 = NULL;    /*   the 1G image: BYTE (uint8), */
   int32   *ibuf_i32 = NULL;    /*   I2 (int16), I4 (int32), or */
   float32 *ibuf_f32 = NULL;    /*   R4 (float32) */

   FILE *ofp;

   /* Check the arguments and input file name */
   if (argc != 3 )
   {
      printf("Usage: hdf4_l1g_hdf5 <HDF4-file> <HDF5-file>\n");
      exit (EXIT_FAILURE);
   }

   band_nsca = 1;

   strcpy(hdf4_name,argv[1]);
   strcpy(hdf5_name,argv[2]);
   printf("File names: %s -> %s\n",hdf4_name,hdf5_name);

   strcpy(progname,"HDF4_L1G_HDF5");

   /* Open L1g file */
   img5 = ias_l1g_open_image(hdf5_name, IPE_WRITE);
   if (img5 == NULL)
   {
      sprintf("Error opening file %s\n",hdf5_name);
      xxx_LogStatus(progname, __FILE__, __LINE__, msg);
      exit(EXIT_FAILURE);
   }

   xxx_create_DDR_name (ddr_name, hdf4_name);
   /* Read in the full DDR HDF structure */
   ddr_control = xxx_ReadDDR (ddr_name, msg);
   if (ddr_control == NULL)
   {
      xxx_LogStatus(progname, __FILE__,__LINE__,"Error reading ddr");
      exit(EXIT_FAILURE);
   }

   nddrs = ddr_control->Count;

   bmd = (IAS_L1G_BAND_METADATA *)malloc( sizeof( IAS_L1G_BAND_METADATA ) * nddrs );
   if( bmd == NULL )
   {
      xxx_LogStatus(progname, __FILE__,__LINE__,"Error allocating band metadata");
      exit(EXIT_FAILURE);
   }

   for( bb=0 ; bb<ddr_control->Count ; bb++ )
   {
      ddr = ddr_control->p_DDR[bb];

      /* bandno[0] = axx_parse_user_band(ddr.BandNumber); */
      bandno[0] = l1g_band_lut[ddr.BandNumber-1];
      printf("Opening band %d,%d\n",ddr.BandNumber,bandno[0]);
      img4 = xxx_open_1g_image(hdf4_name, READ, &nbands, bandno, 
                    dtype4, nl, ns);
      if (img4 == NULL)
      {
         sprintf(msg,"Error accessing input image %s",hdf4_name);
         xxx_LogStatus(progname,__FILE__,__LINE__,msg);
         exit(EXIT_FAILURE);
      }

      size = nl[0] * ns[0];
      /* Allocate input buffer.  */
      if (dtype4[0] == BYTE)
      {
         ibuf_i8  = (uint8 *)malloc(ns[0] * sizeof(uint8));
         ibuf_g = ibuf_i8;
      }
      else if (dtype4[0] == I2)
      {
         ibuf_i16 = (int16 *)malloc(ns[0] * sizeof(int16));
         ibuf_g = ibuf_i16;
      }
      else if (dtype4[0] == I4)
      {
         ibuf_i32 = (int32 *)malloc(ns[0] * sizeof(int32));
         ibuf_g = ibuf_i32;
      }
      else if (dtype4[0] == R4)
      {
         ibuf_f32 = (float32 *)malloc(ns[0] * sizeof(float32));
         ibuf_g = ibuf_f32;
      }
      else
      {
         strcpy(msg,"Error unknown input data type");
         xxx_LogStatus(progname, __FILE__, __LINE__,msg);
         ias_l1g_close_image(img5);
         return(ERROR);
      }
      if (ibuf_g == NULL )
      {
         strcpy(msg,"Error allocating memory for input data buffer");
         xxx_LogStatus(progname, __FILE__, __LINE__,msg);
         ias_l1g_close_image(img5);
         return(ERROR);
      }

      /* Data types should be the same? */
      dtype5 = dtype4[0];
      /* Open band dataset */
      /* nl[0] = nl[0] + 1; */
      /* ns[0] = ns[0] + 1; */
      /* if( ias_l1g_open_band(img5, l0ra_band_lut[bb], &dtype5, &band_nsca, &nl[0], &ns[0]) */
      printf("Size of image file %d %d\n",nl[0],ns[0]);
      l1g_band = ias_l1g_open_band(img5, bb+1, &dtype5, &band_nsca, &nl[0], &ns[0]);
      if( l1g_band == NULL )
      {
         sprintf(msg,"Image %s does not contain band %d", hdf5_name, bb);
         xxx_LogStatus(progname,__FILE__,__LINE__,msg);
         ias_l1g_close_image(img5);
         exit(EXIT_FAILURE);
      }

      ofp = fopen("temp.img","wb");
      for( line=0 ; line<nl[0] ; line++ )
      {
         if( xxx_read_1g_image(img4, bandno[0], line, 0, 1, ns[0], 0, 
                 ibuf_g) != SUCCESS )
         {
            free(ibuf_g);
            sprintf(msg,"Error reading image %s",hdf4_name);
            xxx_LogStatus(progname,__FILE__,__LINE__,msg);
            xxx_close_1g_image(img4);
            ias_l1g_close_image(img5);
            exit(EXIT_FAILURE);
         }
         if( ias_l1g_write_image((const L1G_BAND_IO *)l1g_band, 0, line, 0, 1, ns[0], ibuf_g )
                  != SUCCESS )
         {
            ias_l1g_close_band(l1g_band);
            ias_l1g_close_image(img5);
            strcpy(msg,"Error writing in output band");
            xxx_LogStatus(progname,__FILE__,__LINE__,msg);
            exit(EXIT_FAILURE);
         }
         fwrite(ibuf_g,sizeof(int16),ns[0],ofp);
      }
      fclose(ofp);

      /* Close the image file, probably should check status */
      xxx_close_1g_image(img4);

      /* Not sure the data types match up here */

      if( bb == 0 )
      {
         fmd.projection_code = ddr.ProjectCode;
         fmd.zone_code = ddr.ZoneCode;
         strcpy(fmd.datum,ddr.DatumCode);
         fmd.spheroid_code = ddr.SpheroidCode;       
         strcpy(fmd.projection_units,ddr.ProjectionUnits);
         for( nn=0 ; nn<IAS_L1G_PROJECTION_COEFFICIENTS ; nn++ )
             fmd.projection_parameters[nn] = ddr.ProjectionCoef[nn];
         fmd.wrs_path = ddr.WRSPath;
         fmd.wrs_row = ddr.WRSRows;
         fmd.roll_angle = 0;    
         strcpy(fmd.spacecraft,ddr.SpaceCraftSource);
         strcpy(fmd.collection_type,"Earth");
         switch (ddr.CorrectionType)
         {
            case xxx_Systematic:
                strcpy(fmd.correction_type,"L1G");
                break;
            case xxx_Precision:
                strcpy(fmd.correction_type,"L1P");
                break;
            case xxx_Terrain:
                strcpy(fmd.correction_type,"L1T");
                break;
        }
        strcpy(fmd.capture_direction,ddr.CaptureDirection);
        strcpy(fmd.capture_date,ddr.CaptureDate);
        strncpy(fmd.capture_time,ddr.CaptureTime,32);
        strcpy(fmd.resample_type,ddr.ResampleType);

         strcpy(fmd.software_version,"0.0.0");
         strcpy(fmd.ingest_software_version,"0.0.0");
      }
      bmd[bb].upper_left_y  = ddr.UpperLeft[0];
      bmd[bb].upper_left_x  = ddr.UpperLeft[1];
      bmd[bb].lower_left_y  = ddr.LowerLeft[0];
      bmd[bb].lower_left_x  = ddr.LowerLeft[1];
      bmd[bb].upper_right_y = ddr.UpperRight[0];
      bmd[bb].upper_right_x = ddr.UpperRight[1];
      bmd[bb].lower_right_y = ddr.LowerRight[0];
      bmd[bb].lower_right_x = ddr.LowerRight[1];

      bmd[bb].projection_distance_y = ddr.ProjectionDistY;
      bmd[bb].projection_distance_x = ddr.ProjectionDistX;

      bmd[bb].maximum_pixel_value = ddr.MaxPixelValue;
      bmd[bb].minimum_pixel_value = ddr.MinPixelValue;
      bmd[bb].pixel_range_valid = 1;
      bmd[bb].maximum_radiance = ddr.MaxRadianceValue;
      bmd[bb].minimum_radiance = ddr.MinRadianceValue;
      bmd[bb].spectral_radiance_scaling_offset = 0.0;
      bmd[bb].spectral_radiance_scaling_gain = 0.0;
      bmd[bb].radiance_valid = 0;
      bmd[bb].reflectance_scaling_offset = 0.0;
      bmd[bb].reflectance_scaling_gain = 0.0;
      strcpy(bmd[bb].instrument_source,ddr.InstrumentSource);

      bmd[bb].band_number = ddr.BandNumber;
      sprintf(band_name,"BAND%d",bmd[bb].band_number);
      strcpy(bmd[bb].band_name,band_name);
      ias_l1g_close_band(l1g_band);
   }

   /* Write the file metadata */
   if( ias_l1g_set_file_metadata(img5, &fmd) != SUCCESS )
   {
      ias_l1g_close_image(img5);
      xxx_LogStatus(progname, __FILE__,__LINE__,"Error reading file metadata");
      exit(EXIT_FAILURE);
   }

   /* Write the band metadata */
   if(  ias_l1g_set_band_metadata(img5, bmd, nddrs) != SUCCESS )
   {
      ias_l1g_close_image(img5);
      xxx_LogStatus(progname, __FILE__,__LINE__,"Error reading band metadata");
      exit(EXIT_FAILURE);
   }

   ias_l1g_close_image(img5);

   exit (SUCCESS);
}
