/***************************************************************************
 NAME:                         load_gls

 PURPOSE:           Loads a GLS tile file into memory.  If the tile file
                    can't be found, a "fake" tile dataset is created
                    where the tile DEM data is all 0's

 RETURNS:           Integer status code of SUCCESS (file loads/fake data
                    can be created) or ERROR (file cannot be loaded/fake
                    data cannot be created)
***************************************************************************/
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>              /* PATH_MAX #define */

#include "ias_logging.h"
#include "ias_const.h"

#include "retrieve_elevation.h"
#include "load_gls.h"
#include "gtopo30.h"


int load_gls
(
    const char *dem_dir,               /* I: DEM data root directory */
    const char *cell_name,             /* I: Name of DEM data tile */
    GTOPO30_CELL *cell,                /* O: DEM data */
    int *missing_tiles                 /* O: Count of missing tiles */
)
{
    char header_file[PATH_MAX];
    char data_file[PATH_MAX];
    char tile_basename[GLSDEM_TILE_NAME_SIZE];

    /* GLSDEM tile dimensions:
       dims[0] -- number of lines     (rows)
       dims[1] -- number of samples   (cols)
       dims[2] -- total number of bytes per row
       dims[3] -- x dimension
       dims[4] -- y dimension      */
    int dims[5] = {1201, 1201, 2401, 3.0, 3.0};

    int missing_tile_flag = 0;        /* Flag indicating current tile is
                                         missing, perhaps a "water" tile */

    int status;


    /* Debug check on IAS DEM parent directory */
    IAS_LOG_DEBUG("DEM parent directory:\t%s", dem_dir);

    /* If the name for a given tile is less than 10 characters, it
       may/may not be a water tile.  The Landsat version had
       special handling for this case that we'll (for now) preserve */
    strcpy(tile_basename, cell_name);
    strcat(tile_basename, ".hdr");
    if (strlen(tile_basename) < (GLSDEM_TILE_NAME_SIZE - 1))
        missing_tile_flag = 1;

    /* Construct the desired header and data file names */
    snprintf(header_file, sizeof(header_file), "%s/%s.hdr",
             dem_dir, cell_name);
    snprintf(data_file, sizeof(data_file), "%s/%s.bil",
             dem_dir, cell_name);
    
    /* Debug -- print out full path/name of tile header and data files */
    IAS_LOG_DEBUG("Tile header:\t%s\n", header_file);
    IAS_LOG_DEBUG("Tile data:\t%s\n", data_file);

    /* Check to see if the header file exists. If it does not it's likely a
       water tile. */
    status = access(header_file, R_OK);
    if (status != 0)
    {
        IAS_LOG_WARNING("Reading cell header file %s", header_file);
        missing_tile_flag = 1;
    }

    /* "Fake" a tile data cell with 0's for the DEM data */
    if (missing_tile_flag)
    {
        int lat;
        int lon;

        char n_s;             /* Character indicating north latitude
                                 ('n') or south latitude ('s') */
        char e_w;             /* Character indicating east longitude
                                 ('e') or west longitude ('w') */

        /* Increment the missing tile count */
        (*missing_tiles)++;

        /* Report this tile as missing */
        IAS_LOG_WARNING("Tile %s not found -- could be a 'water' tile",
                        tile_basename);

        /* Parse the basename for the location */
        if (sscanf(cell_name, "%c%2d%c%3d", &n_s, &lat, &e_w, &lon) != 4)
        {
            IAS_LOG_ERROR("Determining lat/lon coordinates from missing "
                          "tile %s", tile_basename);
            return ERROR;
        }
        if (n_s == 's')
            lat = -lat;
        if (e_w == 'w')
            lon = -lon;

        /* Fill the header with appropriate information */
        /* the length of byteorder and layout are 4 characters in
           the cell structure so don't copy more than that */
        strncpy(cell->byteorder, "I", 4);
        strncpy(cell->layout, "BIL", 4);
        cell->nrows         = dims[0];
        cell->ncols         = dims[1];
        cell->nbands        = DMS_DRE_DEM_NBANDS;
        cell->nbits         = DMS_DRE_DEM_NBITS;
        cell->bandrowbytes  = dims[2];
        cell->totalrowbytes = dims[2];
        cell->bandgapbytes  = 0;
        cell->ulxmap        = lon * (int)DEG2ARCSEC;
        cell->ulymap        = (lat + 1) * (int)DEG2ARCSEC;
        cell->xdim          = dims[3];
        cell->ydim          = dims[4];
        cell->nodata        = ILLEGAL_ELEV;

        /* Allocate space for the tile data.  Using calloc initializes
           it with the fill value we want (0)  */
        cell->data = calloc((dims[0] * dims[1]), sizeof(*(cell->data)));
        if (cell->data == NULL)
        {
            IAS_LOG_ERROR("Allocating missing tile DEM data buffer");
            return ERROR;
        }

        return SUCCESS;
    }      /* END -- Block processing for "missing" tile(s) */


                /* We've got an existing tile, so... */

    /* 1.  Read the header file */
    status = read_gtopohdr(header_file, cell);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading cell header file %s", header_file);
        return ERROR;
    }

    /* 2.  Read the data (elevation) file */
    status = read_gtopodem(data_file, cell);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading cell elevation data file %s",
                      data_file);
        return ERROR;
    } 
    /* END -- Block processing for existing tile(s) */
    /* Done */
    return SUCCESS;
}
