/*************************************************************************
 NAME:                      modulus

 PURPOSE:    Module implementing routines to compute modulus values
             for various inputs

 ALGORITHM REFERENCES:  - Wikipedia (modular arithmetic entry)
                        - Any text on number theory or abstract algebra

 NOTES:      The routines in this module compute the equivalence class
             to which a number belongs in Zn; in other words, we compute
             the modulus b, less than n and greater than zero, such that
             a = b + k*n for some integer k.  The routine fmodulus departs
             from the formal definition in that we allow a (and b) to be
             non-integer values.  Thus, we are not technically working in
             Zn (which includes only those integers from 0 to n-1).
***************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "modulus.h"


/*------------------------------------------------------------------------
 NAME:                      modulus

 PURPOSE:  Computes the value j such that i is an element of the
           equivalence class Zj (which contains all integers equivalent
           to j, 0 <= j < n).  More simply, returns the integer value j
           such that i-j is divisible by n

 RETURNS:  j
--------------------------------------------------------------------------*/
int modulus
(
    int i,                           /* I: The number for which we're
                                        interested in determining the
                                        equivalence class */
    int n                            /* I: The specified modulus */
)
{
    while (i < 0)
    {
        i += n;
    }

    while (i > n)
    {
        i -= n;
    }

    return i;
}




/*------------------------------------------------------------------------
 NAME:                   fmodulus

 PURPOSE:      Computes the floating point value y such that x-y is
               divisible by n.

 RETURNS:      y

 NOTES:        The numbers x and y are said to be "congruent modulo n"
               (technically not true, however, since they're not integer
               values).  Still, the concept is useful in working with
               longitudes, since -179 longitude is congruent to 181 modulo n.
               More useful, is the fact that -179 - 179 is equivalent to 2
               modulo n, which is to say that the -179 longitude line is 2
               degrees east of the 179 longitude line.  This makes width
               determinations somewhat easier and allows one to perform 
               calculations which can ignore the fact that the numbers
               associated with longitude are discontinuous--in fact, they
               are not continuous mod 360.
---------------------------------------------------------------------------*/
double fmodulus
(
    double x,                         /* I: The number for which we're
                                         interested in determining the
                                         equivalent value between 0 and n */
    int n                             /* I: The specified modulus */
)
{
    double dn = (double)n;

    while (x < 0.0)
    {
        x += dn;
    }

    while (x > dn)
    {
        x -= dn;
    }

    return x;
}




/*------------------------------------------------------------------------
 NAME:                longitude_norm

 PURPOSE:       Calculates the "canonical longitude" for the specified
                longitude value.

 RETURNS:       Equivalent longitude value in the range [-180.0, 180.0)
                degrees
--------------------------------------------------------------------------*/
double longitude_norm
(
    double longitude                       /* I: Distance east of
                                              the "Prime Meridian" (0 deg.
                                              longitude) to the place in
                                              question */
)
{
    while (longitude < -180.0)
    {
        longitude += 360.0;
    }

    while (longitude >= 180.0)
    {
        longitude -= 360.0;
    }

    return longitude;
}

