/****************************************************************************
 NAME:                       retrieve_elevation

 PURPOSE:      "Driver" routine to retrieve a DEM for an area bounded by
               a specified set of lat/lon coordinates, prior to its
               use during geometric processing.  The DEM output
               will be compatible with the L1G image data format.

 RETURNS:      Exits with a status of EXIT_SUCCESS for successful
               completion, or EXIT_FAILURE if any "fatal" error condition
               is encountered.

 Survey
****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "ias_const.h"
#include "ias_logging.h"
#include "ias_l0r.h"
#include "ias_l1g.h"
#include "ias_geo.h"

#include "dem_types.h"
#include "retrieve_elevation.h"
#include "get_parameters.h"
#include "modulus.h"
#include "gtopo30.h"
#include "get_ramp_data.h"
#include "get_gtopo30_data.h"
#include "get_gls_data.h"
#include "find_minmax_elevation.h"
#include "write_l1g.h"
#include "update_omf.h"
#include "verify_overlap.h"

#define MAXBOX_PADDING          0.2      /* The amount of padding in degrees
                                            to add around the maxbox window */

/***********************************************************************
Name: maxbox

Purpose: Performs a maxbox operation on the 4 corners and returns the 
    UL, LR bounding box.

Returns: nothing

************************************************************************/
static void maxbox
(
    const DMS_DRE_OMF_PARAMS *omf_params, /* I: 4 corners read from OMF */
    double *maxbox_ul_lat,                /* O: maxbox corners */
    double *maxbox_ul_lon,
    double *maxbox_lr_lat,
    double *maxbox_lr_lon
)
{
    int i;
    int source_crosses_180 = 0;
    double lats[4];
    double lons[4];
    double min_lat;
    double max_lat;
    double min_lon;
    double max_lon;

    lats[0] = omf_params->upper_left_lat_lon[0];
    lats[1] = omf_params->upper_right_lat_lon[0];
    lats[2] = omf_params->lower_left_lat_lon[0];
    lats[3] = omf_params->lower_right_lat_lon[0];

    lons[0] = omf_params->upper_left_lat_lon[1];
    lons[1] = omf_params->upper_right_lat_lon[1];
    lons[2] = omf_params->lower_left_lat_lon[1];
    lons[3] = omf_params->lower_right_lat_lon[1];

    /* check to see if longitude crosses 180 degree boundary */
    /* 4 is degrees */
    source_crosses_180 = ias_geo_does_cross_180(4, lons); 

    if (source_crosses_180)
    {
        int corner;

        for (corner = 0; corner < 4; corner ++)
        {
            if (lons[corner] < 0)
            {
                ias_geo_add_once_around( 4, &(lons[corner]));
            }
        }
    }

    /* find the min/maxs */
    min_lat = lats[0];
    max_lat = lats[0];
    min_lon = lons[0];
    max_lon = lons[0];

    for (i = 1; i < 4; i++)
    {
        if (lats[i] < min_lat)
            min_lat = lats[i];
        if (lats[i] > max_lat)
            max_lat = lats[i];
        if (lons[i] < min_lon)
            min_lon = lons[i];
        if (lons[i] > max_lon)
            max_lon = lons[i];
    }

    *maxbox_ul_lat = max_lat;
    *maxbox_ul_lon = min_lon;
    *maxbox_lr_lat = min_lat;
    *maxbox_lr_lon = max_lon;
}

/***********************************************************************
                            MAIN APPLICATION
************************************************************************/
int main(int argc, char *argv[])
{
    char *dem_parent_dir = NULL;       /* DEM data directory "parent" */
    char dem_dir[PATH_MAX];            /* Full path DEM data directory */
    char ramp_filename[PATH_MAX];      /* Full path to RAMP DEM data */
    char omf_params_filename[PATH_MAX];  /* Name of OMF parameter file
                                                 (based on work order ID,
                                                 expected to be in current
                                                 work order directory) */
    char process_name[IAS_L0R_ALGORITHM_NAME_LENGTH]; /* Application name */

    GTOPO30_CELL elevation_cell;       /* Output: elevation data cell   */

    int status;                        /* Function return status */
    int count;                         /* snprintf return value */
    IAS_ELEV_SOURCES_TYPE  dataset_id = UNKNOWN;

    short max_elevation;               /* Elevation range */
    short min_elevation;

    double lon_width;                  /* "Width" of scene in terms of
                                          longitude coordinates */

    double ul_lat;
    double ul_lon;
    double lr_lat;
    double lr_lon;

    struct stat buff;

    /*New data structures to contain desired work order and OMF
      parameters */
    DMS_DRE_WO_PARAMS wo_params;
    DMS_DRE_OMF_PARAMS omf_params;

    /* Initialize the IAS logging library.  Currently assumes the log is
       written to the directory running this application, which is the
       work order directory.  */
    strcpy(process_name, "retrieve_elevation");
    ias_log_initialize(process_name);

    /* Count the number of command-line arguments.  The only required
       command-line argument now is the name of the ODL-based parameter
       file. */
    if (argc != 2)
    {
        IAS_LOG_ERROR("USAGE:  %s <name of parameter file>", process_name);
        return EXIT_FAILURE;
    }

    /* Read the parameter file. */
    memset(&wo_params, 0, sizeof(DMS_DRE_WO_PARAMS));
    memset(&omf_params, 0, sizeof(DMS_DRE_OMF_PARAMS));

    IAS_LOG_DEBUG("Reading processing parameters...");
    status = get_parameters(argv[1], &wo_params, &omf_params);
    if (status != SUCCESS)
    {
        /* Either help was provided (exit with no error), or the parameter
           file(s) couldn't be read (exit with an error) */
        if (status == HELP_SUCCESS)
        {
            return EXIT_SUCCESS;
        }
        else
        {
            IAS_LOG_ERROR("Parameter processing failed");
            return EXIT_FAILURE;
        }
    }

    /* Get the maxbox coordinates */
    maxbox(&omf_params, &ul_lat, &ul_lon, &lr_lat, &lr_lon);

    /* push the bounds of the box out in all directions to account for the
       fact the corner points may not be completely accurate. */
    ul_lat += MAXBOX_PADDING;
    ul_lon -= MAXBOX_PADDING;
    lr_lat -= MAXBOX_PADDING;
    lr_lon += MAXBOX_PADDING;

    /* Verify the given boundary information is valid.  First, make sure the
       maxbox coordinates are within the range -90.0 to 90.0.  When these were
       read in, both lat/lon coordinates were validated against the range
       -180.0 to 180.0 */
    if (ul_lat > NORTH_LATITUDE_LIMIT)
    {
        ul_lat = NORTH_LATITUDE_LIMIT;
    }
    else if (ul_lat < SOUTH_LATITUDE_LIMIT)
    {
        ul_lat = SOUTH_LATITUDE_LIMIT;
    }
    if (lr_lat > NORTH_LATITUDE_LIMIT)
    {
        lr_lat = NORTH_LATITUDE_LIMIT;
    }
    else if (lr_lat < SOUTH_LATITUDE_LIMIT)
    {
        lr_lat = SOUTH_LATITUDE_LIMIT;
    }

    /* Do not allow requested area to be greater than MAX_SCENE_WIDTH from
       West edge to East edge */
    lon_width = (lr_lon - ul_lon);
    lon_width = fmodulus(lon_width, 360);
    if (lon_width > (double)MAX_SCENE_WIDTH)
    {
        IAS_LOG_ERROR("Scene too wide:  %lf, should be less than %lf",
                lon_width, (double)MAX_SCENE_WIDTH);
        return EXIT_FAILURE;
    }


    /* Reasonable scene boundary, so... */

    /* Initialize the DEM data "cell" in case we're retrieving GLS or
       GTOPO30 DEM data */
    memset(&elevation_cell, 0, sizeof(GTOPO30_CELL));

    /* Get the parent DEM data directory */
    dem_parent_dir = getenv("IAS_DATA_DIR");
    if (dem_parent_dir == NULL)
    {
        IAS_LOG_ERROR("Locating DEM data directory");
        return EXIT_FAILURE;
    }
    sprintf(dem_dir, "%s/dem/", dem_parent_dir);

    /* Check for existence of an existing DEM and remove it if there is one */
    status = stat(wo_params.dem_filename, &buff);
    if (status == 0) /* File exists--remove it */
    {
        unlink(wo_params.dem_filename);
    }

    /* Check the bounding box coordinates to see what DEM dataset to
       use.  */
    if (ul_lat <= RAMP_SOUTH_LATITUDE)
    {
        /* Construct the full path to the DEM data file */
        count = snprintf(ramp_filename, sizeof(ramp_filename),
            "%s/%s/RAMP_200_DEM.h5", dem_dir, RAMP_DIR);
        if (count < 0 || count >= sizeof(ramp_filename))
        {
            IAS_LOG_ERROR("Building RAMP DEM full path filename");
            return EXIT_FAILURE;
        }

        status = verify_overlap(ramp_filename, 1, ul_lat, ul_lon, lr_lat,
            lr_lon);
        if (status != SUCCESS)
        {
            /* Check for non-overlapping else fail */
            if (status != DOES_NOT_OVERLAP)
            {
                /* Error during the verify processing */
                IAS_LOG_ERROR("Verifying RAMP DEM overlap");
                return EXIT_FAILURE;
            }

            /* Didn't error so must not overlap --- Default to GTOPO30 */
            IAS_LOG_INFO("Falling back to the gtopo30 DEM dataset");
            dataset_id = GTOPO30;

            /* Get the mosaiced DEM containing the scene area */
            status = get_gtopo30_data(dem_dir, ul_lat, ul_lon, lr_lat, lr_lon,
                    &elevation_cell);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Retrieving mosaiced GTOPO30 DEM data set");
                return EXIT_FAILURE;
            }
        }
        else
        {
            /* RAMP DEM-- upper left latitude < -60.0 (60.0 S) */
            dataset_id = RAMP_DEM;

            /* Get the ramp data file name and the elevation range info */
            status = get_ramp_data(ramp_filename,
                wo_params.work_order_directory, wo_params.dem_filename,
                &max_elevation, &min_elevation);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Retrieving RAMP DEM link and elevation range"
                    " data");
                return EXIT_FAILURE;
            }
        }
    }
    else if (((ul_lat <= GLSDEM_SOUTH_LATITUDE)
                && (ul_lat > RAMP_SOUTH_LATITUDE)
                && (lr_lat <= GLSDEM_SOUTH_LATITUDE)
                && (lr_lat > RAMP_SOUTH_LATITUDE))
            || ((ul_lat >= GLSDEM_NORTH_LATITUDE)
                && (lr_lat >= GLSDEM_NORTH_LATITUDE)))
    {
        /* GTOPO30 DEM -- upper left and lower right latitudes between
           -53.0 and -60.0 (53.0 S and 60.0 S) OR
           upper left and lower right latitudes above
           83.0 N*/
        dataset_id = GTOPO30;

        /* Get the mosaiced DEM containing the scene area */
        status = get_gtopo30_data(dem_dir, ul_lat, ul_lon, lr_lat, lr_lon,
                &elevation_cell);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Retrieving mosaiced GTOPO30 DEM data set");
            return EXIT_FAILURE;
        }
    }
    else
    {
        /* GLS DEM. */
        dataset_id = GLSDEM_90;

        /* Get the mosaiced DEM containing the scene area */
        status = get_gls_data(dem_dir, ul_lat, ul_lon, lr_lat, lr_lon,
                &elevation_cell);
        if (status == WARNING)
        {
            /* Fall back to the gtopo30 DEM dataset over water */
            IAS_LOG_INFO("Fall back to the gtopo30 DEM dataset");
            dataset_id = GTOPO30;
            status = get_gtopo30_data(dem_dir, ul_lat, ul_lon, lr_lat, lr_lon,
                &elevation_cell);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Retrieving mosaiced GLS & GTOPO30 DEM data set");
                return EXIT_FAILURE;
            }
        }
        else if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Retrieving mosaiced GLS DEM data set");
            return EXIT_FAILURE;
        }
    }

    /* For GTOPO30 or GLS DEM data:
       a.  Find the minimum and maximum elevation in the subsetted DEM
       data
       b.  Write the subsetted DEM data to a file in the work order
       directory.  */
    if ((dataset_id == GTOPO30) || (dataset_id == GLSDEM_90))
    {
        status = find_minmax_elevation(elevation_cell.data,
                elevation_cell.nrows,
                elevation_cell.ncols,
                &min_elevation,
                &max_elevation);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Determining elevation range");
            free(elevation_cell.data);
            return EXIT_FAILURE;
        }

        /* Get the DEM file name from the work order parameters */

        status = write_l1g(elevation_cell, min_elevation, max_elevation,
                wo_params.dem_filename);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Writing DEM data file %s", wo_params.dem_filename);
            free(elevation_cell.data);
            return EXIT_FAILURE;
        }
    }   /* END IF Block -- Processing GTOPO30/GLS DEM data */


    /* Create the OMF file name locally, based on work order ID.  Then
       update it with the DEM name and elevation range */
    sprintf(omf_params_filename, "%s.omf", wo_params.work_order_id);

    status = update_omf(omf_params_filename, wo_params.dem_filename,
            max_elevation, min_elevation);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Updating OMF file %s", omf_params_filename);
        free(elevation_cell.data);
        return EXIT_FAILURE;
    }

    /* Free memory and generally clean up*/
    free(elevation_cell.data);

    /* Done */
    IAS_LOG_INFO("DEM data retrieval successfully completed.");
    return EXIT_SUCCESS;
}
