#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "cgreen.h"

#include "ias_const.h"
#include "ias_logging.h"

#include "../retrieve_elevation.h"
#include "../get_parameters.h"
#include "../modulus.h"
#include "../round.h"
#include "../adjust_elev_to_wgs84.h"
#include "../get_geoid_height.h"
#include "../get_cell_intersection.h"
#include "../gtopo30.h"
#include "../get_gls_data.h"
#include "../load_gls.h"
#include "../get_gls_tile_list.h"
#include "../make_gls_tile_name.h"
#include "../get_gtopo30_data.h"
#include "../find_minmax_elevation.h"

/*
#define DEBUG_PRINT
*/
#ifdef DEBUG_PRINT
#undef DEBUG_PRINT
#endif

Ensure test_find_minmax_elevation()
{
    const char simulated_dem_filename[] = "simulatedDEM.img";

    int nlines = 100;
    int nsamps = 150;
    int num_bytes_read;
    int status;

    short *DEM_data_ptr = NULL;
    short read_min;
    short read_max;
    short min_test_elevation = -3762;
    short max_test_elevation = 4081;

    FILE *test_data_fptr;

    /* Allocate data buffer */
    DEM_data_ptr = calloc((nlines * nsamps), sizeof(*DEM_data_ptr));
    assert_true(DEM_data_ptr != NULL);
    if (DEM_data_ptr == NULL)
    {
        printf("System error -- memory allocation failed\n");
        return;
    }

    /* Populate data buffer */
    test_data_fptr = fopen(simulated_dem_filename, "rb");
    assert_true(test_data_fptr != NULL);
    if (test_data_fptr == NULL)
    {
        printf("Can't open simulated DEM data file\n");
        free(DEM_data_ptr);
        return;
    }

    printf("Reading DEM image...\n");
    num_bytes_read = fread(DEM_data_ptr, sizeof(*DEM_data_ptr),
                           (nlines * nsamps), test_data_fptr);
    fclose(test_data_fptr);
    assert_equal(num_bytes_read, (nlines * nsamps));
    if (num_bytes_read != (nlines * nsamps))
    {
        printf("Error reading simulated DEM image\n");
        free(DEM_data_ptr);
        return;
    }
    printf("DEM image successfully read...\n");

    /* Call find_minmax_elevation() to determine the elevation range */
    status = find_minmax_elevation(DEM_data_ptr, nlines, nsamps,
                                   &read_min, &read_max);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Error determining DEM elevation range\n");
        free(DEM_data_ptr);
        return;
    }
    printf("Min elevation:\t%d\n", read_min);
    printf("Max elevation:\t%d\n", read_max);

    /* We've got an elevation range, let's see if it's "accurate" */
    assert_equal(min_test_elevation, read_min);
    assert_equal(max_test_elevation, read_max);
    if ((min_test_elevation != read_min) 
        || (max_test_elevation != read_max))
    {
        printf("Elevation range determined from DEM does not match "
               "known min/max values\n");
        free(DEM_data_ptr);
        return;
    }

    /* Clean up */
    free(DEM_data_ptr);
}




Ensure test_get_gtopo30_data()
{
    const char wo_params_filename[] = "I1003.odl";
    char *dem_parent_dir = NULL;

    DMS_DRE_WO_PARAMS wo_params;
    DMS_DRE_OMF_PARAMS omf_params;

    int status;

    GTOPO30_CELL test_cell;

    /* Clean slate */
    memset(&wo_params, 0, sizeof(DMS_DRE_WO_PARAMS));
    memset(&omf_params, 0, sizeof(DMS_DRE_OMF_PARAMS));

    /* Get the WO and OMF parameters */
    status
        = get_parameters(wo_params_filename, &wo_params, &omf_params);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Parameter read failed\n");
        return;
    }

    /* Build the DEM data path */
    dem_parent_dir = getenv("IAS_DATA_DIR");
    assert_true(dem_parent_dir != NULL);
    if (dem_parent_dir == NULL)
    {
        printf("Cannot locate main IAS data directory\n");
        return;
    }
    strcat(dem_parent_dir, "/dem");

    /* Build the GTOPO30 DEM mosaic for the given scene coordinates*/
    memset(&test_cell, 0, sizeof(GTOPO30_CELL));
    status = get_gtopo30_data(dem_parent_dir,
                              omf_params.upper_left_lat_lon[0],
                              omf_params.upper_left_lat_lon[1],
                              omf_params.lower_right_lat_lon[0],
                              omf_params.lower_right_lat_lon[1],
                              &test_cell);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("get_gtopo30_data() failed\n");
        return;
    }
    assert_true(test_cell.data != NULL);
    if (test_cell.data == NULL)
    {
        printf("mosaiced DEM image not generated\n");
        return;
    }

    /* Checks on population of data structure */

    /* 1.  Did values get written to the data structure members? */
    assert_not_equal(test_cell.nrows, 0);
    assert_not_equal(test_cell.ncols, 0);
    assert_not_equal(test_cell.nbits, 0);
    assert_not_equal(test_cell.bandrowbytes, 0);
    assert_not_equal(test_cell.totalrowbytes, 0);
    assert_equal(test_cell.bandgapbytes, 0);

    /* 2.  If so, were "reasonable" values written? */
    assert_true(test_cell.nrows > 0);
    assert_true(test_cell.ncols > 0);
    assert_true(test_cell.xdim > 0.0);
    assert_true(test_cell.ydim > 0.0);
    assert_true(test_cell.ulxmap <= omf_params.upper_left_lat_lon[1]);
    assert_true(test_cell.ulymap >= omf_params.upper_left_lat_lon[0]);

    /* Clean up */
    free(test_cell.data);
}

Ensure test_get_gtopo30_data_180()
{
    const char wo_params_filename[] = "I1004.odl";
    char *dem_parent_dir = NULL;

    DMS_DRE_WO_PARAMS wo_params;
    DMS_DRE_OMF_PARAMS omf_params;

    int status;

    GTOPO30_CELL test_cell;

    /* Clean slate */
    memset(&wo_params, 0, sizeof(DMS_DRE_WO_PARAMS));
    memset(&omf_params, 0, sizeof(DMS_DRE_OMF_PARAMS));

    /* Get the WO and OMF parameters */
    status
        = get_parameters(wo_params_filename, &wo_params, &omf_params);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Parameter read failed\n");
        return;
    }

    /* Build the DEM data path */
    dem_parent_dir = getenv("IAS_DATA_DIR");
    assert_true(dem_parent_dir != NULL);
    if (dem_parent_dir == NULL)
    {
        printf("Cannot locate main IAS data directory\n");
        return;
    }
    strcat(dem_parent_dir, "/dem");

    /* Build the GTOPO30 DEM mosaic for the given scene coordinates*/
    memset(&test_cell, 0, sizeof(GTOPO30_CELL));
    status = get_gtopo30_data(dem_parent_dir,
                              omf_params.upper_left_lat_lon[0],
                              omf_params.upper_left_lat_lon[1],
                              omf_params.lower_right_lat_lon[0],
                              omf_params.lower_right_lat_lon[1],
                              &test_cell);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("get_gtopo30_data() failed\n");
        return;
    }
    assert_true(test_cell.data != NULL);
    if (test_cell.data == NULL)
    {
        printf("mosaiced DEM image not generated\n");
        return;
    }

    /* Checks on population of data structure */

    /* 1.  Did values get written to the data structure members? */
    assert_equal(test_cell.nrows, 123);
    assert_equal(test_cell.ncols, 483);
    assert_equal(test_cell.nbits, 16);
    assert_equal(test_cell.bandrowbytes, 966);
    assert_equal(test_cell.totalrowbytes, 966);
    assert_equal(test_cell.bandgapbytes, 0);

    /* 2.  If so, were "reasonable" values written? */
    assert_double_equal(test_cell.xdim, 0.0083333333);
    assert_double_equal(test_cell.ydim, 0.0083333333);
    assert_double_equal(test_cell.ulxmap, 177.937500);
    assert_double_equal(test_cell.ulymap, -18.512500);

#ifdef DEBUG_PRINT
    printf("bandgapbytes: %d\n", test_cell.bandgapbytes);
    printf("nbits: %d\n", test_cell.nbits);
    printf("bandrowbytes: %d  totalrowbytes: %d\n",test_cell.bandrowbytes,
            test_cell.totalrowbytes);
    printf("test_cell.nrows: %d   test_call.ncols: %d\n",
            test_cell.nrows, test_cell.ncols);
    printf("test_cell.xdim: %1.9lf   test_call.ydim: %1.9lf\n",
            test_cell.xdim, test_cell.ydim);
    printf("test_cell.ulxmap: %lf test_cell.ulymap: %lf\n",
        test_cell.ulxmap, test_cell.ulymap);
    printf("omf_params.upper_left_lat_lon[1]: %lf"
           "  omf_params.upper_left_lat_long[0]: %lf\n", 
          omf_params.upper_left_lat_lon[1], omf_params.upper_left_lat_lon[0]);

#endif

    /* Clean up */
    free(test_cell.data);
}




Ensure test_get_gls_data()
{
    const char wo_params_filename[] = "I1003.odl";
    char *dem_parent_dir = NULL;

    int status;

    DMS_DRE_WO_PARAMS wo_params;
    DMS_DRE_OMF_PARAMS omf_params;

    GTOPO30_CELL test_cell;


    /* Clean slate */
    memset(&wo_params, 0, sizeof(DMS_DRE_WO_PARAMS));
    memset(&omf_params, 0, sizeof(DMS_DRE_OMF_PARAMS));

    /* Get the WO and OMF parameters */
    status
        = get_parameters(wo_params_filename, &wo_params, &omf_params);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Parameter read failed\n");
        return;
    }

    /* Build the DEM data path */
    dem_parent_dir = getenv("IAS_DATA_DIR");
    assert_true(dem_parent_dir != NULL);
    if (dem_parent_dir == NULL)
    {
        printf("Cannot locate main IAS data directory\n");
        return;
    }
    strcat(dem_parent_dir, "/dem");

    /* Build the GLSDEM mosiac for the given scene coordinates */
    memset(&test_cell, 0, sizeof(GTOPO30_CELL));
    status = get_gls_data(dem_parent_dir,
                          omf_params.upper_left_lat_lon[0],
                          omf_params.upper_left_lat_lon[1],
                          omf_params.lower_right_lat_lon[0],
                          omf_params.lower_right_lat_lon[1],
                          &test_cell);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("get_gls_data() failed\n");
        return;
    }
    assert_true(test_cell.data != NULL);
    if (test_cell.data == NULL)
    {
        printf("mosaiced DEM image not generated\n");
        return;
    }

    /* Checks on population of data structure */

    /* 1.  Did values get written to the data structure members? */
    assert_not_equal(test_cell.nrows, 0);
    assert_not_equal(test_cell.ncols, 0);
    assert_not_equal(test_cell.nbits, 0);
    assert_not_equal(test_cell.bandrowbytes, 0);
    assert_not_equal(test_cell.totalrowbytes, 0);
    assert_equal(test_cell.bandgapbytes, 0);

    /* 2.  If so, were "reasonable" values were written? */
    assert_true(test_cell.nrows > 0);
    assert_true(test_cell.ncols > 0);
    assert_true(test_cell.xdim > 0.0);
    assert_true(test_cell.ydim > 0.0);
    assert_true(test_cell.ulxmap <= omf_params.upper_left_lat_lon[1]);
    assert_true(test_cell.ulymap >= omf_params.upper_left_lat_lon[0]);

    /* Clean up */
    free(test_cell.data);
}

/* Verify that if the first tile is unavailable (e.g.: water data), data is not
   just set to 0 elevation but instead is adjusted for the WGS84 ellipsoid */
Ensure test_get_gls_data_water()
{
    const char wo_params_filename[] = "I10011.odl";
    char *dem_parent_dir = NULL;
    int data_index;

    int status;

    DMS_DRE_WO_PARAMS wo_params;
    DMS_DRE_OMF_PARAMS omf_params;

    GTOPO30_CELL test_cell;

    /* Clean slate */
    memset(&wo_params, 0, sizeof(DMS_DRE_WO_PARAMS));
    memset(&omf_params, 0, sizeof(DMS_DRE_OMF_PARAMS));

    /* Get the WO and OMF parameters */
    status
        = get_parameters(wo_params_filename, &wo_params, &omf_params);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Parameter read failed\n");
        return;
    }

    /* Build the DEM data path */
    dem_parent_dir = getenv("IAS_DATA_DIR");
    assert_true(dem_parent_dir != NULL);
    if (dem_parent_dir == NULL)
    {
        printf("Cannot locate main IAS data directory\n");
        return;
    }
    strcat(dem_parent_dir, "/dem");

    /* Build the GLSDEM mosaic for the given scene coordinates */
    memset(&test_cell, 0, sizeof(GTOPO30_CELL));
    status = get_gls_data(dem_parent_dir,
                          omf_params.upper_left_lat_lon[0],
                          omf_params.upper_left_lat_lon[1],
                          omf_params.lower_right_lat_lon[0],
                          omf_params.lower_right_lat_lon[1],
                          &test_cell);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("get_gls_data() failed\n");
        return;
    }
    assert_true(test_cell.data != NULL);
    if (test_cell.data == NULL)
    {
        printf("mosaiced DEM image not generated\n");
        return;
    }

    /* Verify nodata marker was set, and was not left at original setting */
    assert_true(test_cell.nodata != 0);
    assert_true(test_cell.nodata != ILLEGAL_ELEV);

    /* Verify water elevation data is adjusted from original 0 setting */
    for (data_index = 90; data_index < 100; data_index++)
    {
        assert_true(test_cell.data[data_index] != 0);
    }

    /* Clean up */
    free(test_cell.data);
}

Ensure test_get_gls_data_180()
{
    const char wo_params_filename[] = "I1004.odl";
    char *dem_parent_dir = NULL;

    int status;

    DMS_DRE_WO_PARAMS wo_params;
    DMS_DRE_OMF_PARAMS omf_params;

    GTOPO30_CELL test_cell;


    /* Clean slate */
    memset(&wo_params, 0, sizeof(DMS_DRE_WO_PARAMS));
    memset(&omf_params, 0, sizeof(DMS_DRE_OMF_PARAMS));

    /* Get the WO and OMF parameters */
    status
        = get_parameters(wo_params_filename, &wo_params, &omf_params);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Parameter read failed\n");
        return;
    }

    /* Build the DEM data path */
    dem_parent_dir = getenv("IAS_DATA_DIR");
    assert_true(dem_parent_dir != NULL);
    if (dem_parent_dir == NULL)
    {
        printf("Cannot locate main IAS data directory\n");
        return;
    }
    strcat(dem_parent_dir, "/dem");

    /* Build the GLSDEM mosiac for the given scene coordinates */
    memset(&test_cell, 0, sizeof(GTOPO30_CELL));
    status = get_gls_data(dem_parent_dir,
                          omf_params.upper_left_lat_lon[0],
                          omf_params.upper_left_lat_lon[1],
                          omf_params.lower_right_lat_lon[0],
                          omf_params.lower_right_lat_lon[1],
                          &test_cell);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("get_gls_data() failed\n");
        return;
    }
    assert_true(test_cell.data != NULL);
    if (test_cell.data == NULL)
    {
        printf("mosaiced DEM image not generated\n");
        return;
    }

    /* Checks on population of data structure */

    /* 1.  Did correct values get written to the data structure members? */
    assert_equal(test_cell.nrows, 1202);
    assert_equal(test_cell.ncols, 4801);
    assert_equal(test_cell.nbits, 16);
    assert_equal(test_cell.bandrowbytes, 9602);
    assert_equal(test_cell.totalrowbytes, 9602);
    assert_equal(test_cell.bandgapbytes, 0);

    assert_double_equal(test_cell.xdim, 0.00083333333);
    assert_double_equal(test_cell.ydim, 0.00083333333);
    assert_double_equal(test_cell.ulxmap, 177.946667);
    assert_double_equal(test_cell.ulymap, -18.519167);

#ifdef DEBUG_PRINT
    printf("bandgapbytes: %d\n", test_cell.bandgapbytes);
    printf("nbits: %d\n", test_cell.nbits);
    printf("bandrowbytes: %d  totalrowbytes: %d\n",test_cell.bandrowbytes,
            test_cell.totalrowbytes);
    printf("test_cell.nrows: %d   test_call.ncols: %d\n",
            test_cell.nrows, test_cell.ncols);
    printf("test_cell.xdim: %1.9lf   test_call.ydim: %1.9lf\n",
            test_cell.xdim, test_cell.ydim);
    printf("test_cell.ulxmap: %lf test_cell.ulymap: %lf\n",
        test_cell.ulxmap, test_cell.ulymap);
    printf("omf_params.upper_left_lat_lon[1]: %lf"
          "  omf_params.upper_left_lat_long[0]: %lf\n", 
          omf_params.upper_left_lat_lon[1], omf_params.upper_left_lat_lon[0]);

#endif

    /* Clean up */
    free(test_cell.data);
}

Ensure test_get_geoid_height()
{

    GTOPO30_CELL geoid;          /* elevation data */

    char *dem_parent_dir = NULL;
    char geoid_hfile[PATH_MAX];
    char geoid_dfile[PATH_MAX];
    double lat;
    double lon;
    double geo_height;
    int ncols;
    int status;


    dem_parent_dir = getenv("IAS_DATA_DIR");
    assert_true(dem_parent_dir != NULL);
    if (dem_parent_dir == NULL)
    {
       return;
    }

    sprintf(geoid_hfile, "%s/dem/geoid/GEOID.HDR", dem_parent_dir);
    sprintf(geoid_dfile, "%s/dem/geoid/GEOID.DEM", dem_parent_dir);


    /* Initialize the geoid cell data structure */
    memset(&geoid, 0, sizeof(GTOPO30_CELL));

    /* Read the geoid separation header */
    status = read_gtopohdr(geoid_hfile, &geoid);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Cannot read geoid separation header data in "
                      "%s", geoid_hfile);
        return;
    }

    /* Read the geoid separation data */
    status = read_gtopodem(geoid_dfile, &geoid);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Cannot read geoid separation data in %s",
                      geoid_dfile);
        free(geoid.data);
        return;
    }


    /* set some points to known values */

    ncols = geoid.ncols;
    geoid.data[2*ncols + 360 + 1] = 10;     /* NE point */
    geoid.data[2*ncols + 360] = 10;         /* NW point */
    geoid.data[3*ncols + 360] = 20;         /* SW point */
    geoid.data[3*ncols + 360 + 1] = 20;     /* SE point */

    /* check latitude values */
    
    lat = 89.0;
    lon = 0;
    geo_height = get_geoid_height(&geoid, lat, lon);
    assert_double_equal(geo_height, 10.0);

    lat = 88.875;
    geo_height = get_geoid_height(&geoid, lat, lon);
    assert_double_equal(geo_height, 12.5);

    lat = 88.75;
    geo_height = get_geoid_height(&geoid, lat, lon);
    assert_double_equal(geo_height, 15.0);

    lat = 88.49898393;
    geo_height = get_geoid_height(&geoid, lat, lon);
    assert_double_equal(geo_height, 19.997968);

    lat = 88.5;
    geo_height = get_geoid_height(&geoid, lat, lon);
    assert_double_equal(geo_height, 20.0);

    lat = -37.5;
    geo_height = get_geoid_height(&geoid, lat, lon);
    assert_double_equal(geo_height, 17.0);

    lat = -38.0;
    geo_height = get_geoid_height(&geoid, lat, lon);
    assert_double_equal(geo_height, 17.0);

    lat = -90;
    geo_height = get_geoid_height(&geoid, lat, lon);
    assert_double_equal(geo_height, -29.0);

    lat = 90;
    geo_height = get_geoid_height(&geoid, lat, lon);
    assert_double_equal(geo_height, 13.0);


    /* check longitude values */

    geoid.data[2*ncols + 360 + 1] = 20;     /* NE point */
    geoid.data[2*ncols + 360] = 10;         /* NW point */
    geoid.data[3*ncols + 360] = 10;         /* SW point */
    geoid.data[3*ncols + 360 + 1] = 20;     /* SE point */

    lat = 89.0;
    lon = 0.0;
    geo_height = get_geoid_height(&geoid, lat, lon);
    assert_double_equal(geo_height, 10.0);

    lat = 89.0;
    lon = 0.5;
    geo_height = get_geoid_height(&geoid, lat, lon);
    assert_double_equal(geo_height, 20.0);

    lat = 89.0;
    lon = 0.25;
    geo_height = get_geoid_height(&geoid, lat, lon);
    assert_double_equal(geo_height, 15.0);

    /* should get a height of 7.5 for both -180 and 180 degrees lon */
    lat = 65.25;
    lon = -180.0;
    geo_height = get_geoid_height(&geoid, lat, lon);
    assert_double_equal(geo_height, 7.5);

    lat = 65.25;
    lon = 180.0;
    geo_height = get_geoid_height(&geoid, lat, lon);
    assert_double_equal(geo_height, 7.5);

}


int main(int argc, char *argv[])
{
    /* Set the logging level to the default IAS_LOG_INFO--this will
       turn off the IAS_LOG_DEBUG messages */
    ias_log_initialize("dre_function_test");

    TestSuite *suite = create_test_suite();

    add_test(suite, test_find_minmax_elevation);
    add_test(suite, test_get_gls_data);
    add_test(suite, test_get_gls_data_water);
    add_test(suite, test_get_gls_data_180);
    add_test(suite, test_get_gtopo30_data);
    add_test(suite, test_get_gtopo30_data_180);
    add_test(suite, test_get_geoid_height);

    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }

    return run_test_suite(suite, create_text_reporter());
}
