#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "cgreen.h"

#include "ias_const.h"
#include "ias_logging.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"

#include "../get_parameters.h"
#include "../update_omf.h"

#define MAX_PARAM_FILENAME_LENGTH 30

Ensure test_get_parameters()
{
    DMS_DRE_WO_PARAMS    wo_params;
    DMS_DRE_OMF_PARAMS   omf_params;

    const char wo_params_filename[MAX_PARAM_FILENAME_LENGTH]  = "I1003.odl";

    int status;

    /* Clean slate */
    memset(&wo_params, 0, sizeof(DMS_DRE_WO_PARAMS));
    memset(&omf_params, 0, sizeof(DMS_DRE_OMF_PARAMS));

    /* Try the read */
    status = get_parameters(wo_params_filename, &wo_params, &omf_params);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Can't read parameter file %s\n", wo_params_filename);
        return;
    }

    /* Basic checks on the data */
    assert_true(strcmp(wo_params.dem_filename,"dem_data.h5") == 0);

    assert_double_equal(omf_params.upper_left_lat_lon[0], 44.52);
    assert_double_equal(omf_params.upper_left_lat_lon[1], -95.947);
    assert_double_equal(omf_params.upper_right_lat_lon[0], 44.02);
    assert_double_equal(omf_params.upper_right_lat_lon[1], -92.447);
    assert_double_equal(omf_params.lower_left_lat_lon[0], 41.02);
    assert_double_equal(omf_params.lower_left_lat_lon[1], -96.447);
    assert_double_equal(omf_params.lower_right_lat_lon[0], 40.52);
    assert_double_equal(omf_params.lower_right_lat_lon[1], -92.947);
}

Ensure test_get_parameters_default()
{
    DMS_DRE_WO_PARAMS    wo_params;
    DMS_DRE_OMF_PARAMS   omf_params;

    const char wo_params_filename[MAX_PARAM_FILENAME_LENGTH]  = "I1003d.odl";

    int status;

    /* Clean slate */
    memset(&wo_params, 0, sizeof(DMS_DRE_WO_PARAMS));
    memset(&omf_params, 0, sizeof(DMS_DRE_OMF_PARAMS));

    /* Try the read */
    status = get_parameters(wo_params_filename, &wo_params, &omf_params);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Can't read parameter file %s\n", wo_params_filename);
        return;
    }

    /* Check for default DEM_FILENAME on the data */
    assert_true(strcmp(wo_params.dem_filename,"source_dem.h5") == 0);
}

Ensure test_update_omf()
{
    const char omf_params_filename[MAX_PARAM_FILENAME_LENGTH]  = "I1003.omf";
    char  test_dem_filename[MAX_PARAM_FILENAME_LENGTH];

    short max_elevation = 2500;
    short min_elevation = -52;

    int  test_max_elevation;
    int  test_min_elevation;

    int read_count;
    int status;

    DMS_DRE_OMF_PARAMS  omf_params;

    /* Write the required information to the OMF file */
    status = update_omf(omf_params_filename, "dem_data.h5",
                        max_elevation, min_elevation);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Elevation update failed\n");
        return;
    }

    /* Read the updated OMF file to see if the info was written correctly */
    memset(&omf_params, 0, sizeof(DMS_DRE_OMF_PARAMS));
    IAS_PARM_DECLARE_TABLE(omf_params_compare, MAX_NUM_OMF_PARAMS);

    IAS_PARM_ADD_STRING(omf_params_compare, DEM_FILENAME,
                        "DEM data file", IAS_PARM_REQUIRED,
                        0, NULL, 0, NULL, test_dem_filename,
                        sizeof(test_dem_filename), 0);
    IAS_PARM_ADD_INT(omf_params_compare, MIN_ELEVATION,
                     "Minimum scene DEM elevation", IAS_PARM_REQUIRED,
                     IAS_PARM_NOT_ARRAY, 0, 0, 0, 0, NULL,
                     &test_min_elevation, sizeof(test_min_elevation), 0);
    IAS_PARM_ADD_INT(omf_params_compare, MAX_ELEVATION,
                     "Maximum scene DEM elevation", IAS_PARM_REQUIRED,
                     IAS_PARM_NOT_ARRAY, 0, 0, 0, 0, NULL,
                     &test_max_elevation, sizeof(test_max_elevation), 0);

    read_count = IAS_PARM_GET_TABLE_SIZE(omf_params_compare);
    assert_true(read_count == 3);

    status = ias_parm_read(omf_params_filename, omf_params_compare,
                           read_count);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Something's wrong with the read\n");
        return;
    }

    assert_string_equal(test_dem_filename, "dem_data.h5");
    assert_equal(test_max_elevation, max_elevation);
    assert_equal(test_min_elevation, min_elevation);
}




int main(int argc, char *argv[])
{
    /* Make sure to turn off any IAS_LOG_DEBUG messages from displaying
       by setting the default logging level to IAS_LOG_INFO */
    ias_log_initialize("dre_parameter_io_test");

    TestSuite *suite = create_test_suite();

    add_test(suite, test_get_parameters);
    add_test(suite, test_get_parameters_default);
    add_test(suite, test_update_omf);

    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
