/******************************************************************************
 NAME:                    dre_read_ramp_data_test

 PURPOSE:  CGreen unit testing for reading the RAMP DEM dataset

 RETURNS:  CGreen status

******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "cgreen.h"

#include "ias_const.h"
#include "ias_types.h"
#include "ias_logging.h"
#include "ias_l1g.h"

#include "../retrieve_elevation.h"
#include "../get_parameters.h"
#include "../get_ramp_data.h"
#include "../find_minmax_elevation.h"
#include "../verify_overlap.h"


           /* KLUDGE KLUDGE KLUDGE KLUDGE KLUDGE KLUDGE */

    /* The RAMP DEM is currently located on /edclxs110/development2.
       I initially had the script runtest.csh set IAS_DATA_DIR to that
       location in the runtest.csh driver script.  Unfortunately...  
       that's not where the rest of the DEM data other tests need is
       located.  I tried to add directories for the gtopo30 and gls dem
       tiles in order to copy the tiles there (so everybody can access all
       the DEM data from a common location), but the permissions are set
       so I can't.

       If permissions can't be changed on that directory...  A kludge is to
       set IAS_DATA_DIR from each test (to cover the possibility that it's
       just being run by itself), before trying to access it.  If the test
       order is changed such that this is not the last test run,
       IAS_DATA_DIR will need to be reset to point back to its original
       location (might be a good idea to do this even if this is the last
       test run).  This can be done by saving the current setting, set
       IAS_DATA_DIR with the new value for testing, then reset back with
       the original value prior to exiting this test.   */

Ensure test_read_ramp_metadata()
{
    const char ramp_dem_filename[30] = "RAMP_200_DEM.h5";

    char *dem_parent_dir = NULL;
    char output_filename[PATH_MAX];

    int status;

    L1GIO *l1g_img = NULL;
    IAS_L1G_FILE_METADATA file_metadata;
    IAS_L1G_BAND_METADATA band_metadata;


    /* Get the name of the DEM data root directory */
    dem_parent_dir = getenv("IAS_DATA_DIR");
    if (dem_parent_dir == NULL)
    {
        printf("Cannot find DEM data root directory\n");
        return;
    }
    strcat(dem_parent_dir, "/dem/ramp");

    /* Build the name of the RAMP DEM data file */
    sprintf(output_filename, "%s/%s", dem_parent_dir, ramp_dem_filename);

    /* Open the RAMP DEM file */
    l1g_img = ias_l1g_open_image(output_filename, IAS_READ);
    assert_true(l1g_img != NULL);
    if (l1g_img == NULL)
    {
        printf("Cannot open RAMP DEM data file %s\n", output_filename);
        return;
    }

    /* Prepare to read the metadata */
    memset(&file_metadata, 0, sizeof(IAS_L1G_FILE_METADATA));
    memset(&band_metadata, 0, sizeof(IAS_L1G_BAND_METADATA));

    /* Read the file metadata */
    status = ias_l1g_get_file_metadata(l1g_img, &file_metadata);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot read RAMP DEM file metadata\n");
        ias_l1g_close_image(l1g_img);
        return;
    }

    /* Read the band metadata */
    status = ias_l1g_get_band_metadata(l1g_img, 1, &band_metadata);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot read RAMP DEM band metadata\n");
        ias_l1g_close_image(l1g_img);
        return;
    }

    /* The DDR data record for the HDF4.x DEM has been split between the
       file and band metadata structures.  Make sure the data was copied
       correctly when the file was converted from HDF4.x format to HDF5
       format. */

    /* Check the file metadata first */
    assert_equal(file_metadata.projection_code, 6);
    assert_equal(file_metadata.zone_code, 0);
    assert_equal(file_metadata.spheroid_code, 12);
    assert_equal(file_metadata.wrs_path, 122);
    assert_equal(file_metadata.wrs_row, 1);

    assert_string_equal(file_metadata.software_version, "0.0.0");
    assert_string_equal(file_metadata.ingest_software_version, "0.0.0");
    assert_string_equal(file_metadata.collection_type, "Earth");
    assert_string_equal(file_metadata.projection_units, "METERS");
    assert_string_equal(file_metadata.datum, "WGS_84");
    assert_string_equal(file_metadata.spacecraft, "RAMP");
    assert_string_equal(file_metadata.capture_direction, "both");

    assert_double_equal(file_metadata.projection_parameters[0], 6378137.000);
    assert_double_equal(file_metadata.projection_parameters[1],
            6356752.3140000002);
    assert_double_equal(file_metadata.projection_parameters[2], 0);
    assert_double_equal(file_metadata.projection_parameters[3], 0);
    assert_double_equal(file_metadata.projection_parameters[4], 0);
    assert_double_equal(file_metadata.projection_parameters[5],
            -71000000.000000);
    assert_double_equal(file_metadata.projection_parameters[6], 0);
    assert_double_equal(file_metadata.projection_parameters[7], 0);
    assert_double_equal(file_metadata.projection_parameters[8], 0);
    assert_double_equal(file_metadata.projection_parameters[9], 0);
    assert_double_equal(file_metadata.projection_parameters[10], 0);
    assert_double_equal(file_metadata.projection_parameters[11], 0);
    assert_double_equal(file_metadata.projection_parameters[12], 0);
    assert_double_equal(file_metadata.projection_parameters[13], 0);
    assert_double_equal(file_metadata.projection_parameters[14], 0);

    /* Now check the band metadata */
    assert_equal(band_metadata.band_number, 1);
    assert_equal(band_metadata.pixel_range_valid, 1);
    assert_equal(band_metadata.radiance_valid, 0);
    assert_equal(band_metadata.reflectance_valid, 0);

    assert_double_equal(band_metadata.upper_left_y, 2457900.000);
    assert_double_equal(band_metadata.upper_left_x, -2867900.000);
    assert_double_equal(band_metadata.lower_left_y, -2457900.000);
    assert_double_equal(band_metadata.lower_left_x, -2867900.000);
    assert_double_equal(band_metadata.upper_right_y, 2457900.000);
    assert_double_equal(band_metadata.upper_right_x, 2867900.000);
    assert_double_equal(band_metadata.lower_right_y, -2457900.000);
    assert_double_equal(band_metadata.lower_right_x, 2867900.000);
    assert_double_equal(band_metadata.projection_distance_y, 200.0);
    assert_double_equal(band_metadata.projection_distance_x, 200.0);
    assert_double_equal(band_metadata.maximum_pixel_value, 5008.000);
    assert_double_equal(band_metadata.minimum_pixel_value, -67.000);
    assert_double_equal(band_metadata.spectral_radiance_scaling_offset, 0);
    assert_double_equal(band_metadata.spectral_radiance_scaling_gain, 0);
    assert_double_equal(band_metadata.reflectance_scaling_offset, 0);
    assert_double_equal(band_metadata.reflectance_scaling_gain, 0);

    assert_string_equal(band_metadata.band_name, "BAND1");
    assert_string_equal(band_metadata.instrument_source, "VARIOUS");

    /* Checks are finished.  Clean up */
    ias_l1g_close_image(l1g_img);

    /* Done */
}




Ensure test_read_ramp_image_size()
{
    const char ramp_dem_filename[30] = "RAMP_200_DEM.h5";

    char *dem_parent_dir = NULL;
    char output_filename[PATH_MAX];

    int band_number = 1;
    int nscas;
    int line_count;
    int sample_count;

    IAS_DATA_TYPE dtype;

    L1GIO *l1g_img = NULL;
    L1G_BAND_IO *l1g_band = NULL;


    /* Get the name of the DEM data root directory */
    dem_parent_dir = getenv("IAS_DATA_DIR");
    assert_true(dem_parent_dir != NULL);
    if (dem_parent_dir == NULL)
    {
        printf("Cannot locate main IAS data directory\n");
        return;
    }
    strcat(dem_parent_dir, "/dem/ramp");

    /* Build the name of the RAMP DEM data file */
    sprintf(output_filename, "%s/%s", dem_parent_dir, ramp_dem_filename);

    /* Open the RAMP DEM file */
    l1g_img= ias_l1g_open_image(output_filename, IAS_READ);
    assert_true(l1g_img != NULL);
    if (l1g_img == NULL)
    {
        printf("Cannot open RAMP DEM data file %s\n", output_filename);
        return;
    }

    /* Open the single DEM band */
    l1g_band = ias_l1g_open_band(l1g_img, band_number, &dtype, &nscas,
           &line_count, &sample_count);
    assert_true(l1g_band != NULL);
    if (l1g_band == NULL)
    {
        printf("Cannot open band 1 DEM image\n");
        ias_l1g_close_image(l1g_img);
        return;
    }

    /* Check to make sure the dimensions are valid */
    assert_equal(dtype, IAS_I2);
    assert_equal(nscas, 1);
    assert_equal(line_count, 24580);
    assert_equal(sample_count, 28680);

    /* Clean up */
    ias_l1g_close_band(l1g_band);
    ias_l1g_close_image(l1g_img);

    /* Done */
}

Ensure test_get_ramp_data()
{
    const char ramp_dem_filename[30] = "RAMP_200_DEM.h5";
    const char ramp_link_filename[30] = "RAMP_200_link.h5";

    char *dem_parent_dir = NULL;
    char wo_dir[PATH_MAX] = ".";
    char full_path_ramp[PATH_MAX];

    short max_elevation;
    short min_elevation;

    int status;
    struct stat buff;

    /* Get the name of the DEM data root directory */
    dem_parent_dir = getenv("IAS_DATA_DIR");
    assert_true(dem_parent_dir != NULL);
    if (dem_parent_dir == NULL)
    {
        printf("Cannot locate main IAS data directory\n");
        return;
    }
    snprintf(full_path_ramp, sizeof(full_path_ramp), "%s/dem/ramp/%s",
        dem_parent_dir, ramp_dem_filename);

    /* Get the current working directory */
    getcwd(wo_dir, PATH_MAX);

    unlink(ramp_link_filename); /* Make sure it is not present */

    /* Test getting the ramp data.  When this test completes, check to make
       sure a link to the DEM data directory has been placed in the
       "work order" directory (the test directory where this is run from) */
    status = get_ramp_data(full_path_ramp, wo_dir, ramp_link_filename,
        &max_elevation, &min_elevation);
    assert_true(status == SUCCESS);

    /* Verify that the link to the DEM was created and then remove it */
    status = stat(ramp_link_filename, &buff);
    if (status == 0) /* File exists--remove it */
    {
        unlink(ramp_link_filename);
    }
    else
    {
        assert_true(FALSE); /* Force a failure */
    }

    /* Check the returned elevation data */
    assert_equal(max_elevation, 5008);
    assert_equal(min_elevation, -67);
}

Ensure test_ramp_boundaries()
{
    const char ramp_dem_filename[30] = "RAMP_200_DEM.h5";
    char *dem_parent_dir = NULL;
    char full_path_ramp[PATH_MAX];

    int status;

    /* Get the name of the DEM data root directory */
    dem_parent_dir = getenv("IAS_DATA_DIR");
    assert_true(dem_parent_dir != NULL);
    if (dem_parent_dir == NULL)
    {
        printf("Cannot locate main IAS data directory\n");
        return;
    }
    snprintf(full_path_ramp, sizeof(full_path_ramp), "%s/dem/ramp/%s",
        dem_parent_dir, ramp_dem_filename);

    /* Fail - Upper Left, Upper Right, Center */
    status = verify_overlap(full_path_ramp, 1, -66.0, -10.0, -68.0, 0.0);
    assert_true(status == DOES_NOT_OVERLAP);

    /* Fail - Upper Left, Upper Right, Center - Different side of RAMP DEM */
    status = verify_overlap(full_path_ramp, 1, -61.0, 70.0, -65.0, 80.0);
    assert_true(status == DOES_NOT_OVERLAP);

    /* PASS - Within RAMP DEM */
    status = verify_overlap(full_path_ramp, 1, -70.0, 175.0, -75.0, 185.0);
    assert_true(status == SUCCESS);
}

/* This routine isn't directly used in the unit test.  It was written to 
   call the verify_overlap routine for an entire grid of locations and dump
   out the longitude/latitude of every point that does not overlap the RAMP
   file (or with a small change it can dump out the points that overlap).
   The output from running this test can be dumped to a file and directly
   loaded into envi as a set of vectors and used to confirm that the 
   verify_overlap routine is working correctly over a large area.
   It is currently disabled since it isn't meant to be a long-term unit test. */
#if 0
Ensure dump_grid_points()
{
    double lat_loop;
    const char ramp_dem_filename[30] = "RAMP_200_DEM.h5";
    char *dem_parent_dir = NULL;
    char full_path_ramp[PATH_MAX];
    int status;

    /* Get the name of the DEM data root directory */
    dem_parent_dir = getenv("IAS_DATA_DIR");
    assert_true(dem_parent_dir != NULL);
    if (dem_parent_dir == NULL)
    {
        printf("Cannot locate main IAS data directory\n");
        return;
    }
    status = snprintf(full_path_ramp, sizeof(full_path_ramp), "%s/dem/ramp/%s",
        dem_parent_dir, ramp_dem_filename);
    if (status < 0 || status >= sizeof(full_path_ramp))
    {
        assert_true(0);
        return;
    }

    for (lat_loop = -50.0; lat_loop > -88.0; lat_loop -= 1.0)
    {
        double lon_loop;
        double upper_lat = lat_loop;
        double lower_lat = lat_loop - 1.0;

        ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

        for (lon_loop = -180.0; lon_loop < 180.0; lon_loop += 1.0)
        {
            double left_lon = lon_loop;
            double right_lon = lon_loop + 1;

            status = verify_overlap(full_path_ramp, 1, upper_lat, left_lon,
                lower_lat, right_lon);
            if (status == DOES_NOT_OVERLAP)
                printf("%f %f\n", left_lon + 0.5, upper_lat - 0.5);
            else if (status == ERROR)
            {
                assert_true(0);
                return;
            }
        }
    }

}
#endif

int main(int argc, char *argv[])
{
    /* Set the logging level to IAS_LOG_INFO to avoid the IAS_LOG_DEBUG
       messages in the routines */
    ias_log_initialize("dre_read_ramp_dem_test");

    TestSuite *suite = create_test_suite();

    add_test(suite, test_read_ramp_metadata);
    add_test(suite, test_read_ramp_image_size);
    add_test(suite, test_get_ramp_data);
    add_test(suite, test_ramp_boundaries);

    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
