#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>

#include "cgreen.h"

#include "ias_const.h"
#include "ias_types.h"
#include "ias_logging.h"
#include "ias_l1g.h"

#include "../retrieve_elevation.h"
#include "../get_parameters.h"
#include "../modulus.h"
#include "../round.h"
#include "../adjust_elev_to_wgs84.h"
#include "../get_geoid_height.h"
#include "../get_cell_intersection.h"
#include "../gtopo30.h"
#include "../get_gls_data.h"
#include "../load_gls.h"
#include "../get_gls_tile_list.h"
#include "../make_gls_tile_name.h"
#include "../get_gtopo30_data.h"
#include "../get_gtopo30_tile_list.h"
#include "../find_minmax_elevation.h"
#include "../write_l1g.h"

Ensure test_write_gls_l1g()
{
    const char wo_params_filename[] = "I1003.odl";
    char *dem_parent_dir = NULL;
    char output_filename[PATH_MAX];

    short min_elevation;
    short max_elevation;

    int status;

    DMS_DRE_WO_PARAMS wo_params;
    DMS_DRE_OMF_PARAMS omf_params;

    GTOPO30_CELL glsdem_cell;


    L1GIO *img_ptr = NULL;
    IAS_L1G_BAND_METADATA  band_metadata;

    /* Clean slate */
    memset(&wo_params, 0, sizeof(DMS_DRE_WO_PARAMS));
    memset(&omf_params, 0, sizeof(DMS_DRE_OMF_PARAMS));

    /* Get the WO and OMF parameters */
    status
        = get_parameters(wo_params_filename, &wo_params, &omf_params);
    assert_true(status == SUCCESS);
    if (status !=  SUCCESS)
    {
        printf("Error reading WO and/or OMF parameters\n");
        return;
    }

    /* Get the name of the DEM data root directory */
    dem_parent_dir = getenv("IAS_DATA_DIR");
    assert_true(dem_parent_dir != NULL);
    if (dem_parent_dir == NULL)
    {
        printf("Cannot locate main IAS data directory\n");
        return;
    }
    strcat(dem_parent_dir, "/dem");

    /* Build the GLS_DEM data mosaic for the scene coordinates
       specified in the OMF */
    memset(&glsdem_cell, 0, sizeof(GTOPO30_CELL));
    status = get_gls_data(dem_parent_dir, omf_params.upper_left_lat_lon[0],
                          omf_params.upper_left_lat_lon[1],
                          omf_params.lower_right_lat_lon[0],
                          omf_params.lower_right_lat_lon[1],
                          &glsdem_cell);
    assert_true(status == SUCCESS);
    assert_not_equal(glsdem_cell.data, NULL);
    if ((status !=  SUCCESS) || (glsdem_cell.data == NULL))
    {
        printf("get_gls_data() failed\n");
        free(glsdem_cell.data);
        return;
    }

    /* Get the elevation range */
    status = find_minmax_elevation(glsdem_cell.data,
                                   glsdem_cell.nrows,
                                   glsdem_cell.ncols,
                                   &min_elevation,
                                   &max_elevation);
    assert_true(status == SUCCESS);
    if (status !=  SUCCESS)
    {
        printf("find_minmax_elevation() failed\n");
        free(glsdem_cell.data);
        return;
    }

    /* Here's what we've been leading up to...  Write the DEM data to
       a file in the "simulated" work order directory */
    sprintf(output_filename, "dem_data_gls.h5");
    status = write_l1g(glsdem_cell, min_elevation, max_elevation,
                       output_filename);
    assert_true(status == SUCCESS);
    if (status !=  SUCCESS)
    {
        printf("write_l1g() failed\n");
        free(glsdem_cell.data);
        return;
    }

    /* Now... read the band metadata in order to compare it to what was
       written */
    img_ptr = ias_l1g_open_image(output_filename, IAS_READ);
    if (img_ptr == NULL)
    {
        printf("Error opening created DEM image file for reading\n");
        free(glsdem_cell.data);
        return;
    }
    status = ias_l1g_get_band_metadata(img_ptr, 1, &band_metadata);
    assert_true(status == SUCCESS);
    if (status !=  SUCCESS)
    {
        printf("Error reading band metadata for DEM image file\n");
        ias_l1g_close_image(img_ptr);
        free(glsdem_cell.data);
        return;
    }

    /* Various checks to see if what was calculated was written
       "faithfully" to the metadata */
    assert_true(fabs(band_metadata.upper_left_x
                         - (glsdem_cell.ulxmap * (int)DEG2ARCSEC)) < 1e-2);
    if (fabs(band_metadata.upper_left_x
                 - (glsdem_cell.ulxmap * (int)DEG2ARCSEC)) >= 1e-2)
    {
        printf("Upper left X:  %lf vs %lf\n", band_metadata.upper_left_x,
               (glsdem_cell.ulxmap * (int)DEG2ARCSEC));
    }
    assert_true(fabs(band_metadata.upper_left_y
                         - (glsdem_cell.ulymap * (int)DEG2ARCSEC))
                < 1e-2);
    if (fabs(band_metadata.upper_left_y
                 - (glsdem_cell.ulymap * (int)DEG2ARCSEC)) >= 1e-2)
    {
        printf("Upper left Y:  %lf vs %lf\n", band_metadata.upper_left_y,
               (glsdem_cell.ulymap * (int)DEG2ARCSEC));
    }

    assert_true(fabs(band_metadata.projection_distance_x
                - ((int)DEG2ARCSEC * glsdem_cell.xdim)) < 1e-2);
    if (fabs(band_metadata.projection_distance_x
                - ((int)DEG2ARCSEC * glsdem_cell.xdim)) >= 1e-2)
    {
        printf("Projection distance X:  %lf vs %lf\n",
               band_metadata.projection_distance_x,
               (glsdem_cell.xdim * (int)DEG2ARCSEC));
    }

    assert_true(fabs(band_metadata.projection_distance_y
                - ((int)DEG2ARCSEC * glsdem_cell.ydim)) < 1e-2);
    if (fabs(band_metadata.projection_distance_y
                - ((int)DEG2ARCSEC * glsdem_cell.ydim)) >= 1e-2)
    {
        printf("Projection distance Y:  %lf vs %lf\n",
               band_metadata.projection_distance_y,
               (glsdem_cell.ydim * (int)DEG2ARCSEC));
    }

    assert_true(fabs(band_metadata.maximum_pixel_value - (double)max_elevation)
                < 1e-2);
    if (fabs(band_metadata.maximum_pixel_value - (double)max_elevation) >= 1e-2)
    {
        printf("Maximum elevation:  %lf vs %lf\n",
               band_metadata.maximum_pixel_value, (double)max_elevation);
    }

    assert_true(fabs(band_metadata.minimum_pixel_value - (double)min_elevation)
                < 1e-2);
    if (fabs(band_metadata.minimum_pixel_value - (double)min_elevation) >= 1e-2)
    {
        printf("Minimum elevation:  %lf vs %lf\n",
               band_metadata.minimum_pixel_value, (double)min_elevation);
    }

    /* Clean up */
    free(glsdem_cell.data);
    ias_l1g_close_image(img_ptr);
}




Ensure test_write_gtopo30_l1g()
{
    const char wo_params_filename[] = "I1003.odl";
    char *path = NULL;
    char *dem_parent_dir = NULL;
    char output_filename[PATH_MAX];

    short min_elevation;
    short max_elevation;

    int status;

    DMS_DRE_WO_PARAMS    wo_params;
    DMS_DRE_OMF_PARAMS   omf_params;

    GTOPO30_CELL gtopo30_cell;

    L1GIO *img_ptr = NULL;

    IAS_L1G_BAND_METADATA  band_metadata;

    /* Get the path for the test data directory */
    path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (path == NULL)
    {
        printf("Cannot find test data directory path\n");
        return;
    }

    /* Clean slate */
    memset(&wo_params, 0, sizeof(DMS_DRE_WO_PARAMS));
    memset(&omf_params, 0, sizeof(DMS_DRE_OMF_PARAMS));

    /* Get the WO and OMF parameters */
    status
        = get_parameters(wo_params_filename, &wo_params, &omf_params);
    assert_true(status == SUCCESS);
    if (status !=  SUCCESS)
    {
        printf("Error reading WO and/or OMF parameters\n");
        return;
    }

    /* Get the name of the DEM data root directory */
    dem_parent_dir = getenv("IAS_DATA_DIR");
    assert_true(dem_parent_dir != NULL);
    if (dem_parent_dir == NULL)
    {
        printf("Cannot locate main IAS data directory\n");
        return;
    }
    strcat(dem_parent_dir, "/dem");

    /* Build the GLS_DEM data mosaic for the scene coordinates
       specified in the OMF */
    memset(&gtopo30_cell, 0, sizeof(GTOPO30_CELL));
    status = get_gtopo30_data(dem_parent_dir,
                              omf_params.upper_left_lat_lon[0],
                              omf_params.upper_left_lat_lon[1],
                              omf_params.lower_right_lat_lon[0],
                              omf_params.lower_right_lat_lon[1],
                              &gtopo30_cell);
    assert_true(status == SUCCESS);
    assert_not_equal(gtopo30_cell.data, NULL);
    if ((status !=  SUCCESS) || (gtopo30_cell.data == NULL))
    {
        printf("get_gtopo30_data() failed\n");
        if (gtopo30_cell.data)
        {
            free(gtopo30_cell.data);
            gtopo30_cell.data = NULL;
            return;
        }
    }

    /* Get the elevation range */
    status = find_minmax_elevation(gtopo30_cell.data,
                                   gtopo30_cell.nrows,
                                   gtopo30_cell.ncols,
                                   &min_elevation,
                                   &max_elevation);
    assert_true(status == SUCCESS);
    if (status !=  SUCCESS)
    {
        printf("find_minmax_elevation() failed\n");
        free(gtopo30_cell.data);
        return;
    }

    /* Here's what we've been leading up to...  Write the DEM data to
       a file in the "simulated" work order directory */
    sprintf(output_filename, "dem_data_gtopo30.h5");
    status = write_l1g(gtopo30_cell, min_elevation, max_elevation,
                       output_filename);
    assert_true(status == SUCCESS);
    if (status !=  SUCCESS)
    {
        printf("write_l1g() failed\n");
        free(gtopo30_cell.data);
        return;
    }

    /* Now... read the band metadata in order to compare it to what was
       written */
    img_ptr = ias_l1g_open_image(output_filename, IAS_READ);
    if (img_ptr == NULL)
    {
        printf("Error opening created DEM image file for reading\n");
        free(gtopo30_cell.data);
        return;
    }
    status = ias_l1g_get_band_metadata(img_ptr, 1, &band_metadata);
    assert_true(status == SUCCESS);
    if (status !=  SUCCESS)
    {
        printf("Error reading band metadata for DEM image file\n");
        ias_l1g_close_image(img_ptr);
        free(gtopo30_cell.data);
        return;
    }

    /* Various checks to see if what was calculated was written
       "faithfully" to the metadata */
    assert_true(fabs(band_metadata.upper_left_x
                          - (gtopo30_cell.ulxmap * (int)DEG2ARCSEC)) < 1e-2);
    if (fabs(band_metadata.upper_left_x
                          - (gtopo30_cell.ulxmap * (int)DEG2ARCSEC)) >= 1e-2)
    {
        printf("Upper left X:  %lf vs %lf\n", band_metadata.upper_left_x,
               (gtopo30_cell.ulxmap * (int)DEG2ARCSEC));
    }
    assert_true(fabs(band_metadata.upper_left_y
                         - (gtopo30_cell.ulymap * (int)DEG2ARCSEC)) < 1e-2);
    if (fabs(band_metadata.upper_left_y
                         - (gtopo30_cell.ulymap * (int)DEG2ARCSEC)) >= 1e-2)
    {
        printf("Upper right Y:  %lf vs %lf\n", band_metadata.upper_left_y,
               (gtopo30_cell.ulymap * (int)DEG2ARCSEC));
    }

    assert_true(fabs(band_metadata.projection_distance_x
                         - ((int)DEG2ARCSEC * gtopo30_cell.xdim)) < 1e-2);
    if (fabs(band_metadata.projection_distance_x
                 - ((int)DEG2ARCSEC * gtopo30_cell.xdim)) >= 1e-2)
    {
        printf("Projection distance X:  %lf vs %lf\n",
               band_metadata.projection_distance_x,
               (gtopo30_cell.xdim * (int)DEG2ARCSEC));
    }

    assert_true(fabs(band_metadata.projection_distance_y
                         - ((int)DEG2ARCSEC * gtopo30_cell.ydim)) < 1e-2);
    if (fabs(band_metadata.projection_distance_y
                 - ((int)DEG2ARCSEC * gtopo30_cell.ydim)) >= 1e-2)
    {
        printf("Projection distance Y:  %lf vs %lf\n",
                band_metadata.projection_distance_y,
               (gtopo30_cell.ydim * (int)DEG2ARCSEC));
    }

    assert_true(fabs(band_metadata.maximum_pixel_value
                        - (double)max_elevation) < 1e-2);
    if (fabs(band_metadata.maximum_pixel_value
                 - (double)max_elevation) >= 1e-2)
    {
        printf("Maximum elevation:  %lf vs %lf\n",
               band_metadata.maximum_pixel_value,
               (double)max_elevation);
    }

    assert_true(fabs(band_metadata.minimum_pixel_value
                        - (double)min_elevation) < 1e-2);
    if (fabs(band_metadata.minimum_pixel_value
                 - (double)min_elevation) >= 1e-2)
    {
        printf("Minimum elevation:  %lf vs %lf\n",
               band_metadata.minimum_pixel_value,
               (double)min_elevation);
    }

    /* Clean up */
    free(gtopo30_cell.data);
    ias_l1g_close_image(img_ptr);
}


int main(int argc, char *argv[])
{
    /* Set the logging level to IAS_LOG_INFO to avoid the IAS_LOG_DEBUG
       messages in the routines */
    ias_log_initialize("dre_write_dem_test");

    TestSuite *suite = create_test_suite();

    add_test(suite, test_write_gls_l1g);
    add_test(suite, test_write_gtopo30_l1g);

    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}


