
/***********************************************************************
Name: verify_overlap

Purpose: Checks to see if the L1G band covers the maxbox

Returns: Type = int
    Value             Description
    -------           --------------------------
    SUCCESS           Successful completion - The corners overlap enough
    ERROR             Operation failed
    DOES_NOT_OVERLAP  Does not overlap

************************************************************************/

#include <string.h>

#include "gctp.h"
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_l1g.h"
#include "ias_geo.h"
#include "ias_math.h"

#include "verify_overlap.h"

int verify_overlap
(
    const char *full_path_name, /* I: Full path of the file to compare */
    int band_number,            /* I: Band to compare */
    double maxbox_ul_lat,       /* I: Maxbox corners in degrees to compare */
    double maxbox_ul_lon,       /*    against the file */
    double maxbox_lr_lat,
    double maxbox_lr_lon
)
{
    int status;                          /* Function status return code */
    int index;                           /* Loop control */

    L1GIO *l1g_img = NULL;               /* L1G image data handle */
    IAS_L1G_BAND_METADATA band_metadata; /* Band metadata data */
    IAS_L1G_FILE_METADATA file_metadata; /* File metadata data */

    /*** Projection Variables ***/
    IAS_PROJECTION dem_projection;
    IAS_PROJECTION geographic_projection;
    double geographic_projection_parameters[IAS_PROJ_PARAM_SIZE];
    IAS_GEO_PROJ_TRANSFORMATION *geographic2dem = NULL;
    int units; /* METER, DEGREE */

    /* Polygon center in lat lon */
    double center_lat;
    double center_lon;

    /*** Maxbox values in map coordinates ***/
    double ul_map_x;
    double ul_map_y;
    double ur_map_x;
    double ur_map_y;
    double lr_map_x;
    double lr_map_y;
    double ll_map_x;
    double ll_map_y;
    double center_map_x;
    double center_map_y;
    double polygon_x[4]; /* Only four corners in the L1G data */
    double polygon_y[4];

    int points_found_within_data;

     /* Determine the center of the polygon */
    center_lat = (maxbox_ul_lat + maxbox_lr_lat) / 2;
    center_lon = (maxbox_ul_lon + maxbox_lr_lon) / 2;

   /* Open the L1G filename */
    l1g_img = ias_l1g_open_image(full_path_name, IAS_READ);
    if (l1g_img == NULL)
    {
        IAS_LOG_ERROR("Opening data file name %s", full_path_name);
        return ERROR;
    }

    /* Initialize the metadata structures */
    memset(&band_metadata, 0, sizeof(IAS_L1G_BAND_METADATA));
    memset(&file_metadata, 0, sizeof(IAS_L1G_FILE_METADATA));

    /* Read the band and file metadata, then close the image handle */
    status = ias_l1g_get_band_metadata(l1g_img, band_number, &band_metadata);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading band metadata from %s band_number %d",
            full_path_name, band_number);
        ias_l1g_close_image(l1g_img);
        return ERROR;
    }
    status = ias_l1g_get_file_metadata(l1g_img, &file_metadata);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading file metadata from %s", full_path_name);
        ias_l1g_close_image(l1g_img);
        return ERROR;
    }

    status = ias_l1g_close_image(l1g_img);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Closing data file %s", full_path_name);
        return ERROR;
    }

    /* Initialize the geographic projection information */
    for (index = 0; index < IAS_PROJ_PARAM_SIZE; index++)
    {
        geographic_projection_parameters[index] = 0.0;
    }

    /* Set the geographic projection */
    ias_geo_set_projection(GEO, 0, DEGREE, WGS84_SPHEROID,
        geographic_projection_parameters, &geographic_projection);

    /* Check if meters or degrees or fail */
    status = ias_geo_get_units(file_metadata.projection_units, &units);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Invalid units %s", file_metadata.projection_units);
        return ERROR;
    }
    if (units != METER && units != DEGREE)
    {
        IAS_LOG_ERROR("Unsupported units %s", file_metadata.projection_units);
        return ERROR;
    }

    /* Set the DEM projection */
    ias_geo_set_projection(file_metadata.projection_code,
        file_metadata.zone_code, units, file_metadata.spheroid_code,
        file_metadata.projection_parameters, &dem_projection);

    /* Set the transformation for the point */
    geographic2dem = ias_geo_create_proj_transformation(
        &geographic_projection, &dem_projection);
    if (geographic2dem == NULL)
    {
        IAS_LOG_ERROR("Creating geographic projection");
        return ERROR;
    }

    /* Perform the transformations for the corners and the center of the
       polygon */
    if (ias_geo_transform_coordinate(geographic2dem,
        maxbox_ul_lon, maxbox_ul_lat, &ul_map_x, &ul_map_y) != SUCCESS)
    {
        IAS_LOG_ERROR("Failed to transform upper left coordinates");
        ias_geo_destroy_proj_transformation(geographic2dem);
        return ERROR;
    }
    if (ias_geo_transform_coordinate(geographic2dem,
        maxbox_lr_lon, maxbox_ul_lat, &ur_map_x, &ur_map_y) != SUCCESS)
    {
        IAS_LOG_ERROR("Failed to transform upper right coordinates");
        ias_geo_destroy_proj_transformation(geographic2dem);
        return ERROR;
    }
    if (ias_geo_transform_coordinate(geographic2dem,
        maxbox_lr_lon, maxbox_lr_lat, &lr_map_x, &lr_map_y) != SUCCESS)
    {
        IAS_LOG_ERROR("Failed to transform lower right coordinates");
        ias_geo_destroy_proj_transformation(geographic2dem);
        return ERROR;
    }
    if (ias_geo_transform_coordinate(geographic2dem,
        maxbox_ul_lon, maxbox_lr_lat, &ll_map_x, &ll_map_y) != SUCCESS)
    {
        IAS_LOG_ERROR("Failed to transform lower left coordinates");
        ias_geo_destroy_proj_transformation(geographic2dem);
        return ERROR;
    }
    if (ias_geo_transform_coordinate(geographic2dem,
        center_lon, center_lat, &center_map_x, &center_map_y) != SUCCESS)
    {
        IAS_LOG_ERROR("Failed to transform center coordinates");
        ias_geo_destroy_proj_transformation(geographic2dem);
        return ERROR;
    }

    /* Cleanup the transformation */
    ias_geo_destroy_proj_transformation(geographic2dem);
    geographic2dem = NULL;

    IAS_LOG_DEBUG("Maxbox Values ulx(%lf) uly(%lf)", ul_map_x, ul_map_y);
    IAS_LOG_DEBUG("Maxbox Values urx(%lf) ury(%lf)", ur_map_x, ur_map_y);
    IAS_LOG_DEBUG("Maxbox Values lrx(%lf) lry(%lf)", lr_map_x, lr_map_y);
    IAS_LOG_DEBUG("Maxbox Values llx(%lf) lly(%lf)", ll_map_x, ll_map_y);
    IAS_LOG_DEBUG("Center Values x(%lf) y(%lf)", center_map_x, center_map_y);

    IAS_LOG_DEBUG("L1G Corner Values ulx(%lf) uly(%lf)",
        band_metadata.upper_left_x, band_metadata.upper_left_y);
    IAS_LOG_DEBUG("L1G Corner Values urx(%lf) ury(%lf)",
        band_metadata.upper_right_x, band_metadata.upper_right_y);
    IAS_LOG_DEBUG("L1G Corner Values lrx(%lf) lry(%lf)",
        band_metadata.lower_right_x, band_metadata.lower_right_y);
    IAS_LOG_DEBUG("L1G Corner Values llx(%lf) lly(%lf)",
        band_metadata.lower_left_x, band_metadata.lower_left_y);

    /* Setup the x polygon values */
    polygon_x[0] = band_metadata.upper_left_x;
    polygon_x[1] = band_metadata.upper_right_x;
    polygon_x[2] = band_metadata.lower_right_x;
    polygon_x[3] = band_metadata.lower_left_x;

    /* Setup the y polygon values */
    polygon_y[0] = band_metadata.upper_left_y;
    polygon_y[1] = band_metadata.upper_right_y;
    polygon_y[2] = band_metadata.lower_right_y;
    polygon_y[3] = band_metadata.lower_left_y;

    points_found_within_data = 0; /* Initialize it */
    /* Check if UPPER_LEFT_CORNER is within the L1G polygon */
    if (!ias_math_point_in_polygon(4, polygon_x, polygon_y, ul_map_x, ul_map_y))
    {
        IAS_LOG_WARNING("Upper Left outside data ulx(%lf) uly(%lf)",
            ul_map_x, ul_map_y);
    }
    else
    {
        points_found_within_data++;
    }

    /* Check if UPPER_RIGHT_CORNER is within the L1G polygon */
    if (!ias_math_point_in_polygon(4, polygon_x, polygon_y, ur_map_x, ur_map_y))
    {
        IAS_LOG_WARNING("Upper Right outside data urx(%lf) ury(%lf)",
            ur_map_x, ur_map_y);
    }
    else
    {
        points_found_within_data++;
    }

    /* Check if POLYGON_CENTER is within the L1G polygon */
    if (!ias_math_point_in_polygon(4, polygon_x, polygon_y,
        center_map_x, center_map_y))
    {
        IAS_LOG_WARNING("Polygon center outside data pcx(%lf) pcy(%lf)",
            center_map_x, center_map_y);
    }
    else
    {
        points_found_within_data++;
    }

    /* Check if LOWER_RIGHT_CORNER is within the L1G polygon */
    if (!ias_math_point_in_polygon(4, polygon_x, polygon_y, lr_map_x, lr_map_y))
    {
        IAS_LOG_WARNING("Lower Right outside data lrx(%lf) lry(%lf)",
            lr_map_x, lr_map_y);
    }
    else
    {
        points_found_within_data++;
    }

    /* Check if LOWER_LEFT_CORNER is within the L1G polygon */
    if (!ias_math_point_in_polygon(4, polygon_x, polygon_y, ll_map_x, ll_map_y))
    {
        IAS_LOG_WARNING("Lower Left outside data llx(%lf) lly(%lf)",
            ll_map_x, ll_map_y);
    }
    else
    {
        points_found_within_data++;
    }

    if (points_found_within_data < 3)
    {
        IAS_LOG_WARNING("Insufficient overlap: Only %d points within data"
            " when at least 3 are required", points_found_within_data);
        return DOES_NOT_OVERLAP;
    }

    return SUCCESS;
};

