/****************************************************************************
 NAME:                      write_l1g

 PURPOSE:     Takes a GTOPO30/GLS DEM data cell for a scene and writes it
              to an L1G-formatted HDF5 data file

 RETURnsamples:     Integer status code of SUCCESS or ERROR

 NOTES:       This code was originally implemented for the Landsat 7
              ETM+ IAS.  Changes from the core algorithm include the
              following:

              1.  The output HDF file is written in HDF5 format, rather
              than HDF4.x

              2.  The previous "xxx_" library function calls have
              been replaced with the new L1G library "ias_l1g_"
              function calls.

              2.  The concept of a DDR file for LDCM has gone away.
              The required information is contained in file-related
              and band-related metadata "attributes" that are written to
              the image data file.  Consequently, the DDR data structure(s)
              are not used here, but the new L1G file and band metadata
              structures are.  The corresponding code has been updated
              to reflect this.

****************************************************************************/
#include <string.h>

#include "ias_logging.h"
#include "ias_l1g.h"
#include "ias_geo.h"

#include "retrieve_elevation.h"
#include "gtopo30.h"
#include "write_l1g.h"


/* Local #defines */
#ifndef DEFAULT_DATUM
#define DEFAULT_DATUM "WGS_84"
#endif

#define DMS_DRE_DEM_MAX_XDIM 648000.0


int write_l1g
(
    const GTOPO30_CELL gcell,            /* I: GTOPO30 cell to write */
    short min_elevation,                 /* I: Minimum DEM elevation */
    short max_elevation,                 /* I: Maximum DEM elevation */
    const char *output_filename          /* I: DEM data file name */
)
{
    int status;                     /* Function return status       */
    L1GIO *img_ptr = NULL;          /* L1G image data               */
    L1G_BAND_IO *band_io = NULL;    /* L1G band data                */
    IAS_L1G_FILE_METADATA file_metadata; /* L1G file metadata structure  */
    IAS_L1G_BAND_METADATA band_metadata; /* L1G band metadata structure  */
    int band_number;                /* Band number                  */
    IAS_DATA_TYPE dtype;            /* Band data type               */
    int nlines;                     /* Number of lines per band     */
    int nsamples;                   /* Number of samples per line   */
    int nscas;                      /* Number of SCAs in image (1)  */
    int start_line;                 /* Starting line and sample     */
    int start_sample;

    double delta_right_x;
    double delta_left_y;


    /* Initialize metadata data structures */
    memset(&file_metadata, 0, sizeof(IAS_L1G_FILE_METADATA));
    memset(&band_metadata, 0, sizeof(IAS_L1G_BAND_METADATA));

    /* Load some stuff from the GTOPO30 cell */
    band_number = 1;    /* The band number is 1 because: a) this is elevation
                           data not image data; and b) there is only one
                           band */
    nscas  = 1;
    nlines = gcell.nrows;     /* HDF lines are GTOPO30 rows */
    nsamples = gcell.ncols;   /* HDF samples are GTOPO30 columnsamples */
    dtype = IAS_I2;           /* Output data type is 16-bit (signed)
                                 integer */


    /* Open the HDF5 DEM image */
    img_ptr = ias_l1g_open_image(output_filename, IAS_WRITE);
    if (img_ptr == NULL)
    {
        IAS_LOG_ERROR("Creating DEM data file %s", output_filename);
        return ERROR;
    }

    /* Set up the band data */
    band_io = ias_l1g_open_band(img_ptr, band_number, &dtype, &nscas,
                                &nlines, &nsamples);
    if (band_io == NULL)
    {
        IAS_LOG_ERROR("Setting DEM image band information");
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    /* Write the DEM data cell to the HDF5 image data file.  Since there's
       onlinesy one SCA, the SCA number is 0 (that's how it's referenced in
       the L1G library).  Line/sample counting starts at 0, rather than 1 */
    start_line   = 0;
    start_sample = 0;
    status
        = ias_l1g_write_image(band_io, 0, start_line, start_sample, nlines,
                              nsamples, gcell.data);
    if (status != SUCCESS)
    {
        ias_l1g_close_band(band_io);
        ias_l1g_close_image(img_ptr);
        IAS_LOG_ERROR("Writing band %d to DEM data file %s",
                      band_number, output_filename);
        return ERROR;
    }

    /* Close the image band */
    status = ias_l1g_close_band(band_io);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Closing image band %d of DEM data file %s",
                      band_number,
                      output_filename);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    /* Populate the metadata information, then write it to the DEM data file
       as well.  Zero values should have been taken care of with earlier
       memset() calls to zero out the data structures.  */
       
    /*  First, the file metadata... */
    file_metadata.spheroid_code = WGS84_SPHEROID;      /* WGS84 */
    strcpy(file_metadata.datum, DEFAULT_DATUM);
    strcpy(file_metadata.projection_units, "SECONDS");
    strcpy(file_metadata.collection_type, "EARTH");

    if (gcell.source == GTOPO30_SOURCE)
    {
        strcpy(file_metadata.spacecraft, "GTOPO30");
    }
    else if (gcell.source == GLSDEM_SOURCE)
    {
        strcpy(file_metadata.spacecraft, "GLS_DEM");
    }

    /* Now the band metadata */
    band_metadata.band_number = band_number;

    band_metadata.upper_left_y = (int)DEG2ARCSEC * gcell.ulymap;
    band_metadata.upper_left_x = (int)DEG2ARCSEC * gcell.ulxmap;
    if (band_metadata.upper_left_x >= DMS_DRE_DEM_MAX_XDIM)
    {
        band_metadata.upper_left_x -= (2 * DMS_DRE_DEM_MAX_XDIM);
    }

    delta_left_y = gcell.ulymap - (gcell.ydim * (gcell.nrows - 1));
    band_metadata.lower_left_y = (int)DEG2ARCSEC * delta_left_y;
    band_metadata.lower_left_x = band_metadata.upper_left_x;

    band_metadata.upper_right_y = band_metadata.upper_left_y;

    delta_right_x = gcell.ulxmap + (gcell.xdim * (gcell.ncols - 1));
    band_metadata.upper_right_x = (int)DEG2ARCSEC * delta_right_x;
    if (band_metadata.upper_right_x >= DMS_DRE_DEM_MAX_XDIM)
    {
        band_metadata.upper_right_x -= (2 * DMS_DRE_DEM_MAX_XDIM);
    }

    band_metadata.lower_right_y = band_metadata.lower_left_y;
    band_metadata.lower_right_x = band_metadata.upper_right_x;

    band_metadata.projection_distance_y = (int)DEG2ARCSEC * gcell.ydim;
    band_metadata.projection_distance_x = (int)DEG2ARCSEC * gcell.xdim;

    /* KEY:  The pixel_range_valid flag MUST be set, otherwise, the min/max
             pixel value data are not considered "valid", and
             ias_l1g_get_band_metadata() won't copy them over to the 
             populated metadata structure.  A similar situation exists with
             the radiance_valid flag and the reflectance_scaling_offset and
             reflectance_scaling_gain values */
    band_metadata.pixel_range_valid = 1;
    band_metadata.maximum_pixel_value = (double)max_elevation;
    band_metadata.minimum_pixel_value = (double)min_elevation;

    /* Write the file metadata */
    status = ias_l1g_set_file_metadata(img_ptr, &file_metadata);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing file metadata to DEM data file %s\n",
                      output_filename);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    /* Write the band metadata */
    status = ias_l1g_set_band_metadata(img_ptr, &band_metadata, 1);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing band metadata to DEM data file %s\n",
                      output_filename);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

     /* Close the DEM file*/
    status = ias_l1g_close_image(img_ptr);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Closing DEM data file %s\n", output_filename);
        return ERROR;
    }

    /* Done */
    return SUCCESS;
}
