#!/usr/bin/perl -w
#------------------------------------------------------------------------------
# Name: set_characterization_validity.pl
#
# Description:  Updates the characterization valid flags for the GEOMETRIC and
# RADIOMETRIC_WORK_ORDER_COMMON tables.
#
#------------------------------------------------------------------------------
use strict;
# Global routines
use DBI;
use Logger; # Log routines
use ias_processing_system;
use ias_const;
use ias_db; # Database routines

# Handlers
$SIG{__DIE__} = \&DIEHandler;

# Forward declarations
sub processGeometric($);
sub processRadiometric($);
sub commitTransaction();
sub getWorkOrderCommonInfo($$);
sub setCharacterizationAsValid($$);
sub setCharacterizationAsInvalid($$);
sub setCharacterizationValidity($$$);
sub getMatchingScenesOrIntervals($$$$);
sub getMatchingScenesOrIntervalsForProcedure($$$$);
sub validateTypeParameter($);
sub parseParameters($);

# Declare program constant
use constant PROG => 'set_characterization_validity.pl';
use constant GEOMETRIC => 'GEOMETRIC';
use constant RADIOMETRIC => 'RADIOMETRIC';

# Work order source(s)
use constant IAS => 'I';
use constant LPGS => 'L';
use constant INGEST => 'IS';
use constant IS_VALID => 'Y';
use constant IS_INVALID => 'N';

# Courtesy variables
my $status;

# Begin main module
if (ias_init_system() == ERROR)
{
    $ias_error_msg = 'Undefined error' if !defined $ias_error_msg;
    print STDERR PROG . ": $ias_error_msg\n";
    exit FAILURE; # Exits with status = 1.
}

# Set up logger
my $logger = Logger->new('STDOUT', 'SCV');
if (!$logger)
{
    print STDERR PROG . ": Unable to create logger object for STDOUT.\n";
    exit FAILURE;
}
set_default_log($logger);

# Grab the parameter file that was passed in
if ($#ARGV != 0) # 1 parameter passed in
{
    $logger->fatal_msg("$0 requires a parameter file argument.");
    $logger->close_log();
    exit FAILURE;
}
# The main parameter file that is passed in on the command-line
my $parameter_filename = $ARGV[0];

# Validate the parameter file's existence
if (!-e $parameter_filename)
{
    $logger->fatal_msg("Parameter file '$parameter_filename' does not exist.");
    $logger->close_log();
    exit FAILURE;
}
# Parse the parameter file into a hash reference of input parameters
my $parms = parseParameters($parameter_filename);
if (!defined $parms)
{
    $logger->fatal_msg("Unable to parse parameter file: '$parameter_filename'");
    $logger->close_log();
    exit FAILURE;
}
# Make sure we picked up the necessary parameters for this program to process
if (!defined $parms->{WORK_ORDER_ID})
{
    $logger->fatal_msg("WORK_ORDER_ID not found in parameter file: " . 
        "'$parameter_filename'");
    $logger->close_log();
    exit FAILURE;
}
if (!defined $parms->{OVERWRITE_RPS_CHARACTERIZATION})
{
    # If this wasn't specified as a parameter, we default this to 0.
    $parms->{OVERWRITE_RPS_CHARACTERIZATION} = 0;
}

$logger->info_msg("Starting $0 for WORK_ORDER_ID: $parms->{WORK_ORDER_ID}");

# Get the username and password for the L8_OLITIRS database
my ($user_id, $password) = ias_db_GetDBConnectionInfo('L8_OLITIRS');
# Connect to the database
my $dbh = ias_db_ConnectToDB($user_id, $password);
if (!defined $dbh)
{
    $logger->fatal_msg("Unable to connect to database: $ias_db_error");
    $logger->close_log();
    exit FAILURE;
}

$dbh->{RaiseError}=1; # Enable error checking (and then die)
$dbh->{PrintError}=0; # Disable warnings (so no doubling up on errors)
$dbh->{ShowErrorStatement}=1; # Show the SQL statement

# Check for geometric characterization on this work order.
($status, my $common) = getWorkOrderCommonInfo(GEOMETRIC,
    $parms->{WORK_ORDER_ID});
if (!defined $status)
{
    $logger->fatal_msg('getWorkOrderCommonInfo/GEOMETRIC failed');
    $logger->close_log();
    $dbh->rollback();
    $dbh->disconnect();
    exit FAILURE;
}
if ($status == 1)
{
    # If we find a row for it, apply the validity rules for geometric.
    $status = processGeometric($common);
    if (!defined $status)
    {
        $logger->fatal_msg('processGeometric failed');
        $logger->close_log();
        $dbh->rollback();
        $dbh->disconnect();
        exit FAILURE;
    }
}
# Check for radiometric characterization on this work order.
($status, $common) = getWorkOrderCommonInfo(RADIOMETRIC,
    $parms->{WORK_ORDER_ID});
if (!defined $status)
{
    $logger->fatal_msg('getWorkOrderCommonInfo/RADIOMETRIC failed');
    $logger->close_log();
    $dbh->rollback();
    $dbh->disconnect();
    exit FAILURE;
}
if ($status == 1)
{
    # If we find a row for it, apply the validity rules for radiometric.
    $status = processRadiometric($common);
    if (!defined $status)
    {
        $logger->fatal_msg('processRadiometric failed');
        $logger->close_log();
        $dbh->rollback();
        $dbh->disconnect();
        exit FAILURE;
    }
}
commitTransaction(); # Will rollback if it encounters an error
if (!ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK))
{
    $logger->warn_msg("Unable to disconnect from database: $ias_db_error");
}
# Log completion message.
$logger->info_msg('Successful completion');
$logger->close_log();

exit SUCCESS;

#------------------------------------------------------------------------------
# Name: processRadiometric($)
#
# Description: When an LPGS work order completes, the valid flag should be
# marked 'Y' (IS_VALID) if there are no existing valid records for this scene.
# The valid flag should be marked 'N' (IS_INVALID) if there is an existing
# valid record.
#
# When an IAS work order completes, the valid flag should be marked 'Y' if
# the OVERWRITE_RPS_CHARACTERIZATION work order parameter is set or if there
# are no existing valid records for this scene/interval.  If the new work order
# is marked 'Y' then previous IAS/LPGS work orders for that same scene/interval
# should be marked 'N'.  Existing Ingest work orders (based on interval ID,
# where work_order_id suffix is 'IS') should be left alone.
#
# Returns with: undef (if error), 1 if successful.
#------------------------------------------------------------------------------
sub processRadiometric($)
{
    my $common = shift; # Get hash reference to work order common information
    my $status; # Return value
    my $validflag = IS_INVALID;
    if (!defined $common)
    {
        $logger->fatal_msg('Null parameter passed to function: ' .
            (caller(0))[3]);
        return undef;
    }
    # Verify that this is a IAS or LPGS work order
    if (!defined $common->{SOURCE}
        || ($common->{SOURCE} ne IAS && $common->{SOURCE} ne LPGS))
    {
        $logger->fatal_msg('An IAS or LPGS work order is expected for ' . 
            'function: ' . (caller(0))[3]);
        return undef;
    }
    my $scenes = getMatchingScenesOrIntervals(
        $common->{TYPE},
        $common->{LANDSAT_SCENE_ID},
        $common->{LANDSAT_INTERVAL_ID},
        $common->{WO_COMMON_ID});
    if (!defined $scenes)
    {
        $logger->fatal_msg('getMatchingScenesOrIntervals failed');
        return undef;
    }
    # See if another matching scene or interval says that it is 'valid'
    my $anotherValid = 0;
    foreach my $s (@$scenes)
    {
        if (($s->{VALID} eq IS_VALID) && ($s->{SOURCE} ne INGEST))
        {
            $anotherValid = 1;
            last;
        }
    }
    # Always mark the work order as 'Valid' if it is a LPGS order
    if ($common->{SOURCE} eq LPGS)
    {
        $validflag = IS_VALID;
    }
    # IAS orders need additional checks to be marked as 'Valid'
    elsif (!$anotherValid || $parms->{OVERWRITE_RPS_CHARACTERIZATION})
    {
        $validflag = IS_VALID;
    }
    $status = setCharacterizationValidity(
        $common->{TYPE},
        $common->{WO_COMMON_ID},
        $validflag);
    if (!defined $status || $status != 1)
    {
        $logger->fatal_msg('setCharacterizationValidity failed');
        return undef;
    }
    if ($validflag eq IS_VALID)
    {
        # All previous IAS/LPGS work orders for that same scene/interval should
        # be marked 'N'.
        foreach my $scene (@$scenes)
        {
            if ($scene->{SOURCE})
            {
                # Skip over Ingest Work orders
                next if ($scene->{SOURCE} eq INGEST);
                if ($scene->{VALID} ne IS_INVALID)
                {
                    $status = setCharacterizationAsInvalid(
                        $common->{TYPE},
                        $scene->{WO_COMMON_ID});
                    if (!defined $status || $status != 1)
                    {
                        $logger->fatal_msg('setCharacterizationAsInvalid ' . 
                            'failed');
                        return undef;
                    }
                }
            }
        }
    }

    return 1;
}
#------------------------------------------------------------------------------
# Name: processGeometric($)
#
# Description: All IAS work orders that complete successfully should have the
# valid flag marked 'Y' (IS_VALID).
#
# When a LPGS work order completes, the valid flag should be marked 'Y'.
# Previous LPGS work orders for that same scene (same or different version)
# should be marked 'N' (IS_INVALID).
#
# Returns with: undef (if error), 1 if successful.
#------------------------------------------------------------------------------
sub processGeometric($)
{
    my $common = shift; # Get hash reference to work order common information
    my $status; # Return value
    if (!defined $common)
    {
        $logger->fatal_msg('Null parameter passed to function: ' .
            (caller(0))[3]);
        return undef;
    }
    # Verify that this is a IAS or LPGS work order
    if (!defined $common->{SOURCE}
        || ($common->{SOURCE} ne IAS && $common->{SOURCE} ne LPGS))
    {
        $logger->fatal_msg('An IAS or LPGS work order is expected for ' . 
            'function: ' . (caller(0))[3]);
        return undef;
    }
    # Set the VALID column on this geometric work order common row to 'Y'.
    $status = setCharacterizationAsValid(
        $common->{TYPE},
        $common->{WO_COMMON_ID});
    if (!defined $status || $status != 1)
    {
        $logger->fatal_msg('setCharacterizationAsValid failed');
        return undef;
    }

    my $scenes;
    if ($common->{SOURCE} eq IAS)
    {
        $scenes = getMatchingScenesOrIntervalsForProcedure(
            $common->{TYPE},
            $common->{LANDSAT_SCENE_ID},
            $common->{LANDSAT_INTERVAL_ID},
            $common->{WO_COMMON_ID});
    }
    else #LPGS
    {
        $scenes = getMatchingScenesOrIntervals(
            $common->{TYPE},
            $common->{LANDSAT_SCENE_ID},
            $common->{LANDSAT_INTERVAL_ID},
            $common->{WO_COMMON_ID});
    }
    if (!defined $scenes)
    {
        $logger->fatal_msg('getMatchingScenesOrIntervals failed');
        return undef;
    }
    # Mark all other work orders that match this scene as invalid, as this
    # latest change supercedes all others.
    foreach my $scene (@$scenes)
    {
        # Skip over Ingest Work orders
        next if (!defined $scene->{SOURCE} || $scene->{SOURCE} eq INGEST);

        # Only affect the system being used currently ie IAS or LPGS
        if ($common->{SOURCE} eq $scene->{SOURCE})
        {
            if ($scene->{VALID} ne IS_INVALID)
            {
                $status = setCharacterizationAsInvalid(
                    $common->{TYPE},
                    $scene->{WO_COMMON_ID});
                if (!defined $status || $status != 1)
                {
                    $logger->fatal_msg('setCharacterizationAsInvalid failed');
                    return undef;
                }
            }
        }
    }
    return 1;
}
#------------------------------------------------------------------------------
# Name: commitTransaction()
#
# Description: Commits the current transaction, and rolls it back if it fails.
#
# Returns with: 1 (if successful)
#------------------------------------------------------------------------------
sub commitTransaction()
{
    eval
    {
        # Remove our DIE Handler (if we die, eval will catch it).
        local $SIG{__DIE__}; 
        $dbh->commit;
    };
    if ($@)
    {
        $dbh->rollback();
        $logger->fatal_msg("Unable to commit transaction due to: $@");
        exit FAILURE;
    }
    return 1;
}
#------------------------------------------------------------------------------
# Name: getWorkOrderCommonInfo($$)
#
# Description: Gets the primary key to the <GEOMETRIC/RADIOMETRIC>_WO_COMMON
# table.
#
# Returns with:
#    1) Return status: (undef=error, 0=no row found, 1=row found)
#    2) Hash reference
#------------------------------------------------------------------------------
sub getWorkOrderCommonInfo($$)
{
    # Validate incoming parameters
    my ($type, $woid) = @_;
    return (undef, undef) if (!validateTypeParameter($type));
    if (!defined $woid)
    {
        $logger->fatal_msg('Null WORK_ORDER_ID parameter passed to function: ' .
            (caller(0))[3]);
        return (undef, undef);
    }
    my ($query, $sth);
    $query = "SELECT $type".'_WO_COMMON_ID, WORK_ORDER_ID, '.
        'LANDSAT_SCENE_ID, LANDSAT_INTERVAL_ID '.
        "FROM $type".'_WORK_ORDER_COMMON '.
        'WHERE WORK_ORDER_ID = :WOID';
    $sth = $dbh->prepare($query);
    $sth->bind_param(':WOID', $woid);
    $sth->execute;

    my %common; # Work Order common information
    ($common{WO_COMMON_ID},
    $common{WORK_ORDER_ID},
    $common{LANDSAT_SCENE_ID},
    $common{LANDSAT_INTERVAL_ID}) = $sth->fetchrow_array();
    $common{TYPE} = $type;
    $common{SOURCE} = parseWOID($common{WORK_ORDER_ID});
    undef $sth;
    return (1, \%common) if (defined $common{WO_COMMON_ID});
    # Otherwise, no row was found
    return (0, \%common);
}
#------------------------------------------------------------------------------
# Name: setCharacterizationAsValid($$)
#
# Description: Convenience function (sends 'Y' as 'valid' parameter).
#------------------------------------------------------------------------------
sub setCharacterizationAsValid($$)
{
    return setCharacterizationValidity($_[0], $_[1], IS_VALID);
}
#------------------------------------------------------------------------------
# Name: setCharacterizationAsInvalid($$)
#
# Description: Convenience function (sends 'N' as 'valid' parameter).
#------------------------------------------------------------------------------
sub setCharacterizationAsInvalid($$)
{
    return setCharacterizationValidity($_[0], $_[1], IS_INVALID);
}
#------------------------------------------------------------------------------
# Name: setCharacterizationValidity($$$)
#
# Description: Given the table type, and ID, this routine uses SQL to update
# the VALID column on the <GEOMETRIC/RADIOMETRIC>_WO_COMMON table.
#
# Returns with: undef (if error), otherwise, number of rows updated.
#
# Note: 1 row is expected to be updated.
#------------------------------------------------------------------------------
sub setCharacterizationValidity($$$)
{
    # Validate incoming parameters
    my ($type, $id, $valid) = @_;
    return undef if (!validateTypeParameter($type));
    if (!defined $id || !defined $valid)
    {
        $logger->fatal_msg('Null parameter(s) passed to function: ' . 
            (caller(0))[3]);
        return undef;
    }
    if ($valid ne IS_VALID && $valid ne IS_INVALID)
    {
        $logger->fatal_msg('Invalid value for VALID parameter in function: ' . 
            (caller(0))[3]);
        return undef;
    }
    my ($st, $sth);
    $st = "UPDATE $type".'_WORK_ORDER_COMMON '.
        'SET VALID = :VALID '.
        "WHERE $type".'_WO_COMMON_ID = :ID';
    $sth = $dbh->prepare($st);
    $sth->bind_param(':VALID', $valid);
    $sth->bind_param(':ID', $id);
    my $status = $sth->execute();
    undef $sth;
    if ($status == 0)
    {
        $logger->fatal_msg("Could not update VALID field with key ($id) " .  
            "in table $type" . '_WORK_ORDER_COMMON.');
    }
    else
    {
        $logger->info_msg("Updated VALID field to '$valid' for key ($id) " . 
            "in table $type" . '_WORK_ORDER_COMMON.');
    }
    return $status;
}
#------------------------------------------------------------------------------
# Name: getMatchingScenesOrIntervals($$$$)
#
# Description: Given the table type, a sceneID or IntervalID, and a primary key
# to exclude, this routine finds any other scenes/intervals that match the
# scene_id or interval_id furnished (with the same or different version
# sequences).
#
# Returns with: undef (if error), otherwise an array of hashes of key info.
#------------------------------------------------------------------------------
sub getMatchingScenesOrIntervals($$$$)
{
    # Validate incoming parameters
    my ($type, $scene_id, $interval_id, $wo_common_id) = @_;
    return undef if (!validateTypeParameter($type));
    if (!defined $wo_common_id || (!defined $scene_id && !defined $interval_id))
    {
        $logger->fatal_msg('Null parameter(s) passed to function: ' . 
            (caller(0))[3]);
        return undef;
    }
    # Strip off the version number from the scene or interval id and use this
    # as our 'match'ing substring.
    my $match;
    if (defined $scene_id)
    {
        $match = substr($scene_id, 0, 19) . '%';
    }
    else
    {
        $match = substr($interval_id, 0, 22). '%';
    }
    my ($query, $sth);
    $query = "SELECT $type".'_WO_COMMON_ID, VALID, WORK_ORDER_ID '.
        "FROM $type".'_WORK_ORDER_COMMON ';
    if (defined $scene_id)
    {
        $query .= 'WHERE LANDSAT_SCENE_ID like :MATCH ';
    }
    else
    {
        $query .= 'WHERE LANDSAT_INTERVAL_ID like :MATCH ' .
            'AND LANDSAT_SCENE_ID IS NULL';
    }
    $sth = $dbh->prepare($query);
    $sth->bind_param(':MATCH', $match);
    $sth->execute;
    my @scenes;
    while (my ($id, $valid, $work_order_id) = $sth->fetchrow_array())
    {
        next if ($id == $wo_common_id); # Skip over our key to exclude
        my $source = parseWOID($work_order_id);
        push (@scenes,
            {WO_COMMON_ID => $id,
            VALID => $valid,
            WORK_ORDER_ID => $work_order_id,
            SOURCE => $source
            });
    }
    undef $sth;
    return \@scenes;
}
#------------------------------------------------------------------------------
# Name: getMatchingScenesOrIntervalsForProcedure($$$$)
#
# Description: Given the table type, a sceneID or IntervalID, and a primary key
# to exclude, this routine finds any other scenes/intervals that match the
# scene_id or interval_id furnished (with the same or different version
# sequences) AND the work order was for same procedure as the Work Order
# Common ID passed.
#
# Returns with: undef (if error), otherwise an array of hashes of key info.
#------------------------------------------------------------------------------
sub getMatchingScenesOrIntervalsForProcedure($$$$)
{
    # Validate incoming parameters
    my ($type, $scene_id, $interval_id, $wo_common_id) = @_;
    return undef if (!validateTypeParameter($type));
    if (!defined $wo_common_id || (!defined $scene_id && !defined $interval_id))
    {
        $logger->fatal_msg('Null parameter(s) passed to function: ' . 
            (caller(0))[3]);
        return undef;
    }

    # Strip off the version number from the scene or interval id and use this
    # as our 'match'ing substring.
    my $match;
    if (defined $scene_id)
    {
        $match = substr($scene_id, 0, 19) . '%';
    }
    else
    {
        $match = substr($interval_id, 0, 22). '%';
    }

    my ($query, $sth);
    # Get the procedure ID for this work_order_common_id
    $query = "SELECT PROCEDURE_ID FROM WORK_ORDERS wo, $type"
           . '_WORK_ORDER_COMMON com '
           . 'WHERE '.$type.'_WO_COMMON_ID = :COM_ID AND wo.WORK_ORDER_ID ='
           . 'com.WORK_ORDER_ID';
    $sth = $dbh->prepare($query);
    if (not $sth)
    {
        $logger->fatal_msg("prepare error : $DBI::errstr");
        return undef;
    }
    $sth->bind_param(':COM_ID', $wo_common_id);
    if (not $sth->execute())
    {
        $logger->fatal_msg("execute error : $DBI::errstr");
        $sth->finish();
        return undef;
    }
    my ($procedure_id) = $sth->fetchrow_array();
    if (!defined $procedure_id)
    {
        $logger->fatal_msg("Failed to get Procedure ID for ID: $wo_common_id");
        return undef;
    }

    $query = "SELECT $type".'_WO_COMMON_ID, com.VALID, com.WORK_ORDER_ID '
           . "FROM $type".'_WORK_ORDER_COMMON com, WORK_ORDERS wo WHERE '
           . 'wo.WORK_ORDER_ID = com.WORK_ORDER_ID AND wo.PROCEDURE_ID = '
           . ':PROC_ID';
    if (defined $scene_id)
    {
        $query .= ' AND LANDSAT_SCENE_ID like :MATCH ';
    }
    else
    {
        $query .= ' AND LANDSAT_INTERVAL_ID like :MATCH ' .
                  'AND LANDSAT_SCENE_ID IS NULL';
    }
    $sth->finish();
    $sth = $dbh->prepare($query);
    if (not $sth)
    {
        $logger->fatal_msg("prepare error : $DBI::errstr");
        return undef;
    }
    $sth->bind_param(':MATCH', $match);
    $sth->bind_param(':PROC_ID', $procedure_id);
    if (not $sth->execute())
    {
        $logger->fatal_msg("execute error : $DBI::errstr");
        $sth->finish();
        return undef;
    }
    my @scenes;
    while (my ($id, $valid, $work_order_id) = $sth->fetchrow_array())
    {
        next if ($id == $wo_common_id); # Skip over our key to exclude
        my $source = parseWOID($work_order_id);
        push (@scenes,
            {WO_COMMON_ID => $id,
            VALID => $valid,
            WORK_ORDER_ID => $work_order_id,
            SOURCE => $source
            });
    }
    $sth->finish();
    return \@scenes;
}
#------------------------------------------------------------------------------
# Name: validateTypeParameter($)
#
# Description: Verifies the type parameter is GEOMETRIC or RADIOMETRIC.
#
# Returns with: undef (if error), otherwise 1 (success).
#------------------------------------------------------------------------------
sub validateTypeParameter($)
{
    my $type = shift;
    if (!defined $type)
    {
        $logger->fatal_msg('Null TYPE parameter passed to function: ' . 
            (caller(0))[3]);
        return undef;
    }
    if ($type ne GEOMETRIC && $type ne RADIOMETRIC)
    {
        $logger->fatal_msg('Invalid value for TYPE parameter in function: ' .
            (caller(0))[3]);
        return undef;
    }
    return 1;
}
#------------------------------------------------------------------------------
# Name: parseWOID($)
#
# Description: Given a work order ID, this routine returns the source subsystem.
#
# Returns with: undef (if null), otherwise it will be IAS, LPGS or INGEST, or ''
# (if unknown)
#------------------------------------------------------------------------------
sub parseWOID($)
{
    my $woid = shift;
    return undef if (!defined $woid);

    return INGEST if ($woid =~ /^(\d+)IS$/);

    # Note: If we ever add a third processing type, this code must be changed,
    # as we can no longer rely on figuring what it is from what it isn't.
    if ($isLPGS)
    {
        return ($woid =~ /^$ENV{IAS_SYSTEM_ID}(\d+)$/) ? LPGS : IAS;
    }
    elsif ($isIAS)
    {
        return ($woid =~ /^$ENV{IAS_SYSTEM_ID}(\d+)$/) ? IAS : LPGS;
    }
    return '';
}
#------------------------------------------------------------------------------
# Name: parseParameters($)
#
# Description: Given the parameter filename, this routine finds the assigned
# WORK_ORDER_ID and WORK_ORDER_DIRECTORY values.
#
# Returns with: undef (if error), otherwise 1 (success).
#------------------------------------------------------------------------------
sub parseParameters($)
{
    my $parm_filename = shift;
    my $status = open (INPUT, "<$parm_filename");
    if (!defined($status))
    {
        $logger->fatal_msg("Can't open input parameter file:'$parm_filename'.");
        return undef;
    }
    my %parms;
    while (my $line = <INPUT>)
    {
        if ($line =~ /^\s*WORK_ORDER_ID\s*=\s*("(\w+)"|\w+)\s*$/)
        {
            if (defined $2)
            {
                $parms{WORK_ORDER_ID} = $2;
            }
            else
            {
                $parms{WORK_ORDER_ID} = $1;
            }
        }
        elsif ($line =~/^\s*OVERWRITE_RPS_CHARACTERIZATION\s*=\s*(\d+)\s*$/)
        {
            $parms{OVERWRITE_RPS_CHARACTERIZATION} = $1;
        }
        last if (exists $parms{WORK_ORDER_ID} &&
            exists $parms{OVERWRITE_RPS_CHARACTERIZATION});
    }
    close (INPUT);
    return \%parms;
}
#------------------------------------------------------------------------------
# Name: DIEHandler
#
# Description: Traps the die signal, and logs the error message to our
# log file, and shuts the script down.  The error could be a script or DB
# error.
#
# Exits with: FAILURE (error status = 1)
#------------------------------------------------------------------------------
sub DIEHandler
{
    my ($errormessage) = @_; # I: Die error message
    eval
    {
        local $SIG{__DIE__};
        if ($logger->{OPEN})
        {
            $logger->fatal_msg($errormessage);
        }
    };
    exit FAILURE;
}
