#!/usr/bin/perl -w
#------------------------------------------------------------------------------
# Name: test_set_char_validity.pl
#
# Description:  Tests the update of the characterization valid flags for the
# GEOMETRIC and RADIOMETRIC_WORK_ORDER_COMMON tables.
#
#------------------------------------------------------------------------------
use strict;
# Global routines
use DBI;
use ias_processing_system;
use ias_const;
use ias_db; # Database routines
use Getopt::Long;
use Test::More;

# Forward declarations
sub insertWorkOrder($;$$);
sub deleteWorkOrders();
sub getResults();
sub commitTransaction();
sub createParameterFile($$);
sub compareArrayofHashes($$);
sub compareHashes($$);
sub convertToIntervals();
sub launchScript();
sub firstTest($);
sub secondTest($);
sub thirdTest($);
sub fourthTest($);
sub fifthTest($);
sub ShowUsage($);
sub ProcessOptions();
sub insertHistogramStatsEarth($);
sub countHistogramStatsEarth($);

# Handlers
$SIG{__DIE__} = \&DIEHandler;

# Constants
use constant PROG => 'test_set_char_validity.pl';
use constant GEOMETRIC => 'GEOMETRIC';
use constant RADIOMETRIC => 'RADIOMETRIC';

# Global variables
my @work_orders = ('I10001', 'I10002', 'I10003', 'L20001', 'L20002',
'30001IS');
my $ias_system_id;
my $debug_scenario;
my $version = 0;
# Courtesy variables
my $status;

ProcessOptions();

# If database tests are disabled, skip this test
if (!defined $ENV{ENABLE_DB_TESTS})
{
    print STDERR "Database tests are disabled... Skipping $0\n";
    exit SUCCESS;
}

my $instance = $ENV{IAS_INSTANCE_NAME};
if (!defined $instance)
{
    print STDERR "Unable to find the default instance. " .
        "Set IAS_INSTANCE_NAME.\n";
    exit ERROR;
}

# Set up the number of tests we are going to run
if ($debug_scenario)
{
    if ($debug_scenario > 1)
    {
        plan tests => $debug_scenario - 1;
    }
}
else
{
    plan tests => 5;
}

# Begin main module
if (ias_init_system() == ERROR)
{
    $ias_error_msg = 'Undefined error' if !defined $ias_error_msg;
    print STDERR PROG . ": $ias_error_msg\n";
    exit FAILURE; # Actually will exit with 255 (Test::Simple does this)
}

# Get the username and password for the L8_OLITIRS database
my ($user_id, $password) = ias_db_GetDBConnectionInfo('L8_OLITIRS');
# Connect to the database
my $dbh = ias_db_ConnectToDB($user_id, $password);
if (!defined $dbh)
{
    print STDERR ("Unable to connect to database: $ias_db_error\n");
    exit FAILURE;
}

# Check to see whether the IAS_SYSTEM_ID environment variable has been set
# (it's set when 'ias/ias_setup' or 'lpgs/lpgs_setup' is sourced).
# We can't assume that the IAS and/or LPGS setup scripts have been sourced
# prior to this point, and theoretically, this test should be runnable
# independent of whether they have been sourced.  If IAS_SYSTEM_ID isn't
# defined, just default to the IAS value for the duration of this test.
$ias_system_id = $ENV{'IAS_SYSTEM_ID'};
if (not defined($ias_system_id))
{
    print
        "'IAS_SYSTEM_ID' environment variable not set, defaulting to IAS\n";
    $ENV{'IAS_SYSTEM_ID'} = 'I';
}

$dbh->{RaiseError}=1; # Enable error checking (and then die)
$dbh->{PrintError}=0; # Disable warnings (so no doubling up on errors)
$dbh->{ShowErrorStatement}=1; # Show the SQL statement

# Remove any previously used test data from database...
deleteWorkOrders();

# Run our test series...
ok(firstTest('I10001') == 0);
ok(secondTest('L20001') == 0);
ok(thirdTest('L20002') == 0);
ok(fourthTest('I10002') == 0);
ok(fifthTest('I10003') == 0);

# Clean up our work
deleteWorkOrders();
system("rm", "DSC.odl");

if (!ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK))
{
    print ("Unable to disconnect from database: $ias_db_error");
}
exit SUCCESS;

#------------------------------------------------------------------------------
# Name: firstTest($)
#
# Description: Tests the scenario where an IAS work order (I10001) does not have
# a matching scene anywhere.
# With parameter: OVERWRITE_RPS_CHARACTERIZATION = 0.
#
# Exits: 0 (SUCCESS), or other value (failure of some kind)
#---------------------------------------------------------------------------
sub firstTest($)
{
    my $workorder = shift;
    # Create a DSC.odl file to work with
    return FAILURE unless createParameterFile($workorder, 0) == SUCCESS;
    # Put rows out there in the common tables.
    return FAILURE unless insertWorkOrder($workorder) == SUCCESS;
    # Put rows in the histogram stats earth table.
    return FAILURE unless insertHistogramStatsEarth($workorder) == SUCCESS;

    # Make sure this IAS work order inserted its records correctly
    if (countHistogramStatsEarth($workorder) != 30)
    {
        print "-- Error work order $workorder doesn't have correct number of " .
            "histogram statistics earth records\n";
        return FAILURE;
    }

    # Invoke set_characterization_validity.pl script
    exit SUCCESS if ($debug_scenario == 1);

    $status = launchScript();
    if ($status == 0)
    {
        # Get the validity data from the database
        my $actual = getResults();
        my @expected = (
            {SOURCE => GEOMETRIC, WORK_ORDER_ID => $workorder, VALID => 'Y'},
            {SOURCE => RADIOMETRIC, WORK_ORDER_ID => $workorder, VALID => 'Y'}
        );
        # Compare the actual results to the expected output
        return compareArrayofHashes($actual, \@expected);
    }
    else
    {
        return $status;
    }
}

#------------------------------------------------------------------------------
# Name: secondTest($)
#
# Description: Tests the scenario where an LPGS work order (L20001) has a
# matching IAS scene (I10001).  (For previous values, see above).
# With parameter: OVERWRITE_RPS_CHARACTERIZATION = 0.
#
# Exits: 0 (SUCCESS), or other value (failure of some kind)
#---------------------------------------------------------------------------
sub secondTest($)
{
    my $workorder = shift;
    # Create a DSC.odl file to work with
    return FAILURE unless createParameterFile($workorder, 0) == SUCCESS;
    # Put rows out there in the common tables.
    return FAILURE unless insertWorkOrder($workorder) == SUCCESS;
    # Put rows in the histogram stats earth table.
    return FAILURE unless insertHistogramStatsEarth($workorder) == SUCCESS;

    # Make sure this LPGS work order inserted its records correctly
    if (countHistogramStatsEarth($workorder) != 30)
    {
        print "-- Error work order $workorder doesn't have correct number of " .
            "histogram statistics earth records\n";
        return FAILURE;
    }

    # Invoke set_characterization_validity.pl script
    exit SUCCESS if ($debug_scenario == 2);

    $status = launchScript();
    if ($status == 0)
    {
        # Get the validity data from the database
        my $actual = getResults();
        my @expected = (
            {SOURCE => GEOMETRIC, WORK_ORDER_ID => 'I10001', VALID => 'Y'},
            {SOURCE => GEOMETRIC, WORK_ORDER_ID => $workorder, VALID => 'Y'},
            {SOURCE => RADIOMETRIC, WORK_ORDER_ID => 'I10001', VALID => 'N'},
            {SOURCE => RADIOMETRIC, WORK_ORDER_ID => $workorder, VALID => 'Y'}
        );
        # Compare the actual results to the expected output
        return compareArrayofHashes($actual, \@expected);

        # Make sure this LPGS work order didn't delete the records from the 
        # previous IAS work order 
        if (countHistogramStatsEarth('I10001') != 30)
        {
            print "-- Error work order $workorder removed work order I10001's ".
                "histogram statistics earth records\n";
            return FAILURE;
        }
    }
    else
    {
        return $status;
    }
}
#------------------------------------------------------------------------------
# Name: thirdTest()
#
# Description: Tests the scenario where a LPGS work order (L20002) has matching
# IAS and LPGS scenes (I10001, L20001). (For previous values, see above).
# With parameter: OVERWRITE_RPS_CHARACTERIZATION = 1.
#
# Exits: 0 (SUCCESS), or other value (failure of some kind)
#---------------------------------------------------------------------------
sub thirdTest($)
{
    my $workorder = shift;
    # Create a DSC.odl file to work
    return FAILURE unless createParameterFile($workorder, 0) == SUCCESS;
    # Put rows out there in the common tables.
    return FAILURE unless insertWorkOrder($workorder) == SUCCESS;
    # Put rows in the histogram stats earth table.
    return FAILURE unless insertHistogramStatsEarth($workorder) == SUCCESS;
    
    # Make sure this LPGS work order inserted its records correctly
    if (countHistogramStatsEarth($workorder) != 30)
    {
        print "-- Error work order $workorder doesn't have correct number of " .
            "histogram statistics earth records\n";
        return FAILURE;
    }

    # Invoke set_characterization_validity.pl script
    exit SUCCESS if ($debug_scenario == 3);

    $status = launchScript();
    if ($status == 0)
    {
        # Get the validity data from the database
        my $actual = getResults();
        my @expected = (
            {SOURCE => GEOMETRIC, WORK_ORDER_ID => 'I10001', VALID => 'Y'},
            {SOURCE => GEOMETRIC, WORK_ORDER_ID => 'L20001', VALID => 'N'},
            {SOURCE => GEOMETRIC, WORK_ORDER_ID => $workorder, VALID => 'Y'},
            {SOURCE => RADIOMETRIC, WORK_ORDER_ID => 'I10001', VALID => 'N'},
            {SOURCE => RADIOMETRIC, WORK_ORDER_ID => 'L20001', VALID => 'N'},
            {SOURCE => RADIOMETRIC, WORK_ORDER_ID => $workorder, VALID => 'Y'}
        );
        
        # Make sure this LPGS work order did not delete the records from the 
        # previous LPGS work order (this is no longer done)
        if (countHistogramStatsEarth('L20001') != 30)
        {
            print "-- Error work order $workorder didn't remove work order " .
                "L20001's histogram statistics earth records\n";
            return FAILURE;
        }

        # Make sure this LPGS work order didn't delete the records from the 
        # previous IAS work order 
        if (countHistogramStatsEarth('I10001') != 30)
        {
            print "-- Error work order $workorder removed work order I10001's ".
                "histogram statistics earth records\n";
            return FAILURE;
        }

        # Compare the actual results to the expected output
        return compareArrayofHashes($actual, \@expected);
    }
    else
    {
        return $status;
    }
}
#------------------------------------------------------------------------------
# Name: fourthTest($)
#
# Description: Tests the scenario where an IAS work order (I10002) has matching
# IAS and LPGS scenes (I10001, L20001, L20002). (For previous values, see
# above).
#
# With parameter: OVERWRITE_RPS_CHARACTERIZATION = 1.
#
# Exits: 0 (SUCCESS), or other value (failure of some kind)
#---------------------------------------------------------------------------
sub fourthTest($)
{
    my $workorder = shift;
    # Create a DSC.odl file to work with
    return FAILURE unless createParameterFile($workorder, 1) == SUCCESS;
    # Put rows out there in the common tables.
    return FAILURE unless insertWorkOrder($workorder) == SUCCESS;
    # Invoke set_characterization_validity.pl script

    exit SUCCESS if ($debug_scenario == 4);

    $status = launchScript();
    if ($status == 0)
    {
        # Get the validity data from the database
        my $actual = getResults();
        my @expected = (
            {SOURCE => GEOMETRIC, WORK_ORDER_ID => 'I10001', VALID => 'N'},
            {SOURCE => GEOMETRIC, WORK_ORDER_ID => $workorder, VALID => 'Y'},
            {SOURCE => GEOMETRIC, WORK_ORDER_ID => 'L20001', VALID => 'N'},
            {SOURCE => GEOMETRIC, WORK_ORDER_ID => 'L20002', VALID => 'Y'},
            {SOURCE => RADIOMETRIC, WORK_ORDER_ID => 'I10001', VALID => 'N'},
            {SOURCE => RADIOMETRIC, WORK_ORDER_ID => $workorder, VALID => 'Y'},
            {SOURCE => RADIOMETRIC, WORK_ORDER_ID => 'L20001', VALID => 'N'},
            {SOURCE => RADIOMETRIC, WORK_ORDER_ID => 'L20002', VALID => 'N'}
        );

        # Make sure this IAS work order didnt delete the records from the 
        # previous LPGS work order
        if (countHistogramStatsEarth('L20002') != 30)
        {
            print "-- Error work order $workorder removed work order L20002's ".
                "histogram statistics earth records\n";
            return FAILURE;
        }

        # Compare the actual results to the expected output
        return compareArrayofHashes($actual, \@expected);
    }
    else
    {
        return $status;
    }
}
#------------------------------------------------------------------------------
# Name: fifthTest($)
#
# Description: Tests the scenario where an IAS work order (I10003) has matching
# IAS and LPGS intervals (I10001, I10002, L20001, L20002), and a matching
# INGEST interval (30001IS). (For previous values, see above).
#
# With parameter: OVERWRITE_RPS_CHARACTERIZATION = 1.
#
# Exits: 0 (SUCCESS), or other value (failure of some kind)
#---------------------------------------------------------------------------
sub fifthTest($)
{
    my $workorder = shift;
    # Create a DSC.odl file to work with
    return FAILURE unless createParameterFile($workorder, 1) == SUCCESS;
    # Put rows out there in the common tables. This has a different procedure
    # than the other work orders, so I10002 Geo will remain valid
    return FAILURE unless insertWorkOrder($workorder, 'Y',
                                          'Not an L1T') == SUCCESS;
    # Put an Ingest order out there in the common tables (as VALID) to see
    # if it remains that way.
    return FAILURE unless insertWorkOrder('30001IS', 'Y') == SUCCESS;
    # Null out the LANDSAT_SCENE_IDS on the radiometric_work_order_common table.
    convertToIntervals();

    exit SUCCESS if ($debug_scenario == 5);

    # Invoke set_characterization_validity.pl script
    $status = launchScript();
    if ($status == 0)
    {
        # Get the validity data from the database
        my $actual = getResults();
        my @expected = (
            {SOURCE => GEOMETRIC, WORK_ORDER_ID => '30001IS', VALID => 'Y'},
            {SOURCE => GEOMETRIC, WORK_ORDER_ID => 'I10001', VALID => 'N'},
            {SOURCE => GEOMETRIC, WORK_ORDER_ID => 'I10002', VALID => 'Y'},
            {SOURCE => GEOMETRIC, WORK_ORDER_ID => $workorder, VALID => 'Y'},
            {SOURCE => GEOMETRIC, WORK_ORDER_ID => 'L20001', VALID => 'N'},
            {SOURCE => GEOMETRIC, WORK_ORDER_ID => 'L20002', VALID => 'Y'},
            {SOURCE => RADIOMETRIC, WORK_ORDER_ID => '30001IS', VALID => 'Y'},
            {SOURCE => RADIOMETRIC, WORK_ORDER_ID => 'I10001', VALID => 'N'},
            {SOURCE => RADIOMETRIC, WORK_ORDER_ID => 'I10002', VALID => 'N'},
            {SOURCE => RADIOMETRIC, WORK_ORDER_ID => $workorder, VALID => 'Y'},
            {SOURCE => RADIOMETRIC, WORK_ORDER_ID => 'L20001', VALID => 'N'},
            {SOURCE => RADIOMETRIC, WORK_ORDER_ID => 'L20002', VALID => 'N'}
        );

        # Make sure this IAS work order didnt delete the records from the 
        # previous LPGS work order
        if (countHistogramStatsEarth('L20002') != 30)
        {
            print "-- Error work order $workorder removed work order L20002's ".
                "histogram statistics earth records\n";
            return FAILURE;
        }

        # Compare the actual results to the expected output
        return compareArrayofHashes($actual, \@expected);
    }
    else
    {
        return $status;
    }
}
#------------------------------------------------------------------------------
# Name: launchScript()
#
# Description: Invokes the set_characterization_validity.pl script located in
# the parent directory.
#
# Exits: 0 (SUCCESS), or other value (failure of some kind)
#------------------------------------------------------------------------------
sub launchScript()
{
    my $script = 'set_characterization_validity.pl';
    my $output = `../$script DSC.odl`;
    if ($? == -1)
    {
        print STDERR "$script failed to launch.\n";
        return FAILURE;
    }
    else
    {
        my $status = $? >> 8;
        if ($status != 0)
        {
            print STDERR "$script failed with status: $status\n";
        }
        return $status;
    }
}
#------------------------------------------------------------------------------
# Name: getResults()
#
# Description: Returns the geometric and radiometric validity data (in an array
# of hashes) for the work orders in global variable: @work_orders.
#
# Exits: Array of hashes (may be empty).
#------------------------------------------------------------------------------
sub getResults()
{
    my $text = join ', ', map {qq/'$_'/} @work_orders;
    my ($st, $sth, $status);

    $st = q{SELECT 'GEOMETRIC' AS SOURCE, WORK_ORDER_ID, VALID FROM }.
        'GEOMETRIC_WORK_ORDER_COMMON '.
        "WHERE WORK_ORDER_ID IN ($text) ".
        'UNION ALL '.
        q{SELECT 'RADIOMETRIC' AS SOURCE, WORK_ORDER_ID, VALID FROM }.
        'RADIOMETRIC_WORK_ORDER_COMMON '.
        "WHERE WORK_ORDER_ID IN ($text) ".
        'ORDER BY SOURCE ASC, WORK_ORDER_ID ASC';
    $sth = $dbh->prepare($st);
    $status = $sth->execute();
    my @results = ();
    while (1)
    {
        my ($source, $work_order_id, $valid) = $sth->fetchrow_array();
        last if (!defined $work_order_id);
        push (@results,
            {SOURCE => $source,
            WORK_ORDER_ID => $work_order_id,
            VALID => $valid
            });
    }
    undef $sth;
    return \@results;
}
#------------------------------------------------------------------------------
# Name: insertWorkOrder($)
#
# Description: Creates geometric and radiometric header rows in the
# GEOMETRIC/RADIOMETRIC_WORK_ORDER_COMMON tables.
#
# Exits with: SUCCESS, otherwise it returns (FAILURE) or dies.
#------------------------------------------------------------------------------
sub insertWorkOrder($;$$)
{
    # Validate incoming parameters
    my ($work_order_id, $valid, $procedure) = @_;
    if (!defined $work_order_id)
    {
        die 'Undefined parameter passed to function: '. (caller(0))[3];
    }
    # Default the VALID column to 'N', unless we are requesting 'Y'.
    $valid = 'N' unless (defined $valid && $valid eq 'Y');
    $procedure = 'L1T Product' unless defined $procedure;
    my $versionstring = sprintf('%02d', $version);
    my ($st, $sth);
    for my $type (GEOMETRIC, RADIOMETRIC)
    {
        $st = "INSERT INTO $type".'_WORK_ORDER_COMMON ('.
        'COLLECTION_TYPE, LANDSAT_INTERVAL_ID, LANDSAT_SCENE_ID, '.
        'SENSOR_ID, VALID, WORK_ORDER_ID, TARGET_WRS_PATH, '.
        'TARGET_WRS_ROW) VALUES ('.
        q{'EARTH_IMAGING','LC80250380382001108XXX'||:VERSION, }.
        q{'LC80250382001108XXX'||:VERSION, }.
        q{'OLITIRS', :VALID, :WORK_ORDER_ID, 24, 38)};
        $sth = $dbh->prepare($st);
        $sth->bind_param(':VALID', $valid);
        $sth->bind_param(':WORK_ORDER_ID', $work_order_id);
        $sth->bind_param(':VERSION', $versionstring);
        my $status = $sth->execute();
        undef $sth;
        if (!$status)
        {
            print STDERR
                ("Could not insert row with work order id ($work_order_id) ".
                "into table $type"."_WORK_ORDER_COMMON.\n");
            $dbh->rollback();
            return FAILURE;
        }
    }
    
    $st = qq{INSERT INTO work_orders
               (work_order_id, archived, date_entered, deleted,
                priority, requester, request_type, state, act_compl_date,
                date_l0R_requested, ephem_type,
                req_start_date, customer_order_nbr, customer_unit_nbr,
                procedure_id, product_request_id,
                cpf_override, ephem_file, wo_comment, wo_path, cpf_name,
                satellite, sensor, scene_id, l0r_data_path, l0r_data_status,
                output_db_schema, processing_instance)
            VALUES
               (:WORK_ORDER_ID, 0, sysdate, 0, 
                5, 'unit_test', 'S', 'P', NULL, 
                NULL, NULL,
                sysdate, NULL, NULL,
                :PROCEDURE, NULL, 
                NULL, NULL, NULL, NULL, NULL,
                'L8', 'OLITIRS', 'LC80250382001108XXX'||:VERSION,NULL,
                0, 'P', '$instance')};
    $sth = $dbh->prepare($st);
    $sth->bind_param(':WORK_ORDER_ID', $work_order_id);
    $sth->bind_param(':PROCEDURE', $procedure);
    $sth->bind_param(':VERSION', $versionstring);
    my $status = $sth->execute();
    undef $sth;
    if (!$status)
    {
        print STDERR
            ("Could not insert row with work order id ($work_order_id) ".
            "into table WORK_ORDERS.\n");
        $dbh->rollback();
        return FAILURE;
    }
    commitTransaction(); # Bails if this fails...
    $version++;
    return SUCCESS;
}
#------------------------------------------------------------------------------
# Name: insertHistogramStatsEarth($)
#
# Description: Inserts 30 records into the HISTOGRAM_STATISTICS_EARTH table
#              given a work order id that exists in the 
#              RADIOMETRIC_WORK_ORDER_COMMON table.
# Exits with: SUCCESS, otherwise it returns (FAILURE) or dies.
#------------------------------------------------------------------------------
sub insertHistogramStatsEarth($)
{
    # Validate incoming parameters
    my $work_order_id = shift;
    if (!defined $work_order_id)
    {
        die 'Undefined parameter passed to function: '. (caller(0))[3];
    }

    # Retrieve the radiometric work order common id for the work order id
    my ($query, $sth);
    $query = "SELECT RADIOMETRIC_WO_COMMON_ID from " . 
        "RADIOMETRIC_WORK_ORDER_COMMON where WORK_ORDER_ID = :WOID";
    $sth = $dbh->prepare($query);
    $sth->bind_param(':WOID', $work_order_id);
    $sth->execute();
    my $rwocid = $sth->fetchrow_array();
    undef $sth;

    if (!defined $rwocid)
    {
        die 'No radiometric work order common id found for work order '.
            'id ' . $work_order_id;
    }
    
    # Insert the same stats for each band 1-30 for the current radiometric
    # work order common id
    for my $band (1..30)
    {
        $query = qq(INSERT into HISTOGRAM_STATISTICS_EARTH 
            (RADIOMETRIC_WO_COMMON_ID, BAND_ID, SENSOR_CHIP_ASSEMBLY_ID, 
            DETECTOR_ID, SCENE_PIXEL_MIN, SCENE_PIXEL_MAX, SCENE_PIXEL_STDEV,  
            SCENE_PIXEL_MEAN, SCENE_PIXEL_SKEWNESS, SCENE_PIXEL_KURTOSIS,
            NBR_OF_FRAMES, MEAN_SQUARED_RESPONSE, PROCESSING_LEVEL, 
            ADJACENT_CORRELATION, LINESET) VALUES 
            (:RWOCID, :BAND, 1, 226, 675.966064, 6879.95557, 846.354675,
            1482.16809,	3.10192251, 11.2044849,	7001, 2913575, 'CTBL', 
            2909896, 2));

        $sth = $dbh->prepare($query);
        $sth->bind_param(':RWOCID', $rwocid);
        $sth->bind_param(':BAND', $band);
        my $status = $sth->execute();
        undef $sth;
        if (!$status)
        {
            print STDERR
                ("Could not insert row with rad common id ($rwocid) for ".
                "band number $band into table HISTOGRAM_STATISTICS_EARTH.\n");
            $dbh->rollback();
            return FAILURE;
        }
    }
    commitTransaction(); # Bails if this fails...
    return SUCCESS;
}
#------------------------------------------------------------------------------
# Name: countHistogramStatsEarth($)
#
# Description: Returns the number of records in the HISTOGRAM_STATISTICS_EARTH
#              table for the given work order id.
# Exits with: Record Count, or on error returns undef
#------------------------------------------------------------------------------
sub countHistogramStatsEarth($)
{
    # Validate incoming parameters
    my $work_order_id = shift;
    if (!defined $work_order_id)
    {
        die 'Undefined parameter passed to function: '. (caller(0))[3];
    }

    # Retrieve the radiometric work order common id for the work order id
    my ($query, $sth);
    $query = "SELECT RADIOMETRIC_WO_COMMON_ID from " . 
        "RADIOMETRIC_WORK_ORDER_COMMON where WORK_ORDER_ID = :WOID";
    $sth = $dbh->prepare($query);
    $sth->bind_param(':WOID', $work_order_id);
    $sth->execute();
    my $rwocid = $sth->fetchrow_array();
    undef $sth;
    if (!defined $rwocid)
    {
        die 'No radiometric work order common id found for work order '.
            'id ' . $work_order_id;
    }

    # Count the number of records for the id selected
    $query = "SELECT count(RADIOMETRIC_WO_COMMON_ID) as count from " . 
        "HISTOGRAM_STATISTICS_EARTH where RADIOMETRIC_WO_COMMON_ID = :RWOCID";
    $sth = $dbh->prepare($query);
    $sth->bind_param(':RWOCID', $rwocid);
    my $status = $sth->execute(); 
    my $count = $sth->fetchrow_array();
    undef $sth; 
    if ($status)
    {
        return int($count);
    }
    {
        die "Error retrieving record count for id ($rwocid)";
    }
}
#------------------------------------------------------------------------------
# Name: deleteWorkOrders()
#
# Description: Delete all the rows in the
# GEOMETRIC/RADIOMETRIC_WORK_ORDER_COMMON table matching the WORK_ORDER_IDs in
# in our global array: @work_orders.
#
# Exits with: SUCCESS, otherwise it bails out.
#------------------------------------------------------------------------------
sub deleteWorkOrders()
{
    my $text = join ', ', map {qq/'$_'/} @work_orders;
    my ($st, $sth, $status);
    for my $type (GEOMETRIC, RADIOMETRIC)
    {
        $st = "DELETE $type" . '_WORK_ORDER_COMMON '.
        "WHERE WORK_ORDER_ID IN ($text)";
        $sth = $dbh->prepare($st);
        $sth->execute();
        undef $sth;
    }
    $st = "DELETE FROM WORK_ORDERS ".
    "WHERE WORK_ORDER_ID IN ($text)";
    $sth = $dbh->prepare($st);
    $sth->execute();
    undef $sth;
    commitTransaction(); # This bails if this fails...
    return SUCCESS;
}
#------------------------------------------------------------------------------
# Name: convertToIntervals()
#
# Description: Nulls out the LANDSAT_SCENE_ID column for all the rows in the
# RADIOMETRIC_WORK_ORDER_COMMON table matching the WORK_ORDER_IDs in
# in our global array: @work_orders.  This changes those 'scenes' into
# 'intervals'.
#
# Exits with: SUCCESS, otherwise it bails out.
#------------------------------------------------------------------------------
sub convertToIntervals()
{
    my $text = join ', ', map {qq/'$_'/} @work_orders;
    my ($st, $sth, $status);
    $st = 'UPDATE RADIOMETRIC_WORK_ORDER_COMMON '.
    'SET LANDSAT_SCENE_ID = NULL '.
    "WHERE WORK_ORDER_ID IN ($text)";
    $sth = $dbh->prepare($st);
    $sth->execute();
    undef $sth;
    commitTransaction(); # This bails if this fails...
    return SUCCESS;
}
#------------------------------------------------------------------------------
# Name: commitTransaction()
#
# Description: Commits the current transaction, and rolls it back if it fails.
#
# Exits with: SUCCESS, otherwise it perform a rollback and bails out.
#------------------------------------------------------------------------------
sub commitTransaction()
{
    eval
    {
        local $SIG{__DIE__};
        $dbh->commit;
    };
    if ($@)
    {
        $dbh->rollback();
        print STDERR ("Unable to commit transaction due to: $@");
        exit FAILURE;
    }
    return SUCCESS;
}
#------------------------------------------------------------------------------
# Name: createParameterFile($$)
#
# Description: Creates a DSC.odl file with the values needed for running the
# set_characterization_validity.pl script.
#
# Exits with: SUCCESS, dies otherwise
#------------------------------------------------------------------------------
sub createParameterFile($$)
{
     # Validate incoming parameters
    my ($work_order_id, $overwrite) = @_; # I: Work order ID;
                                 # I: Overwrite RPS Characterization vaue
    if (!defined $work_order_id || !defined $overwrite)
    {
        die 'Undefined parameter(s) passed to function: '. (caller(0))[3];
    }
    open (OUTFILE, ">DSC.odl")
        or die "Can't open DSC.odl file for output: $!";
    print OUTFILE "OBJECT = DSC\n";
    print OUTFILE "  WORK_ORDER_ID = $work_order_id\n";
    print OUTFILE "  OVERWRITE_RPS_CHARACTERIZATION = $overwrite\n";
    print OUTFILE "END_OBJECT = DSC\n";
    print OUTFILE "END\n";
    close OUTFILE;
    return SUCCESS;
}
#------------------------------------------------------------------------------
# Name: compareArrayofHashes($$)
#
# Description: Compares two arrays of hashes to see if they contain the same
# data.
#
# Exits with: SUCCESS (same), FAILURE (different).
#------------------------------------------------------------------------------
sub compareArrayofHashes($$)
{
    my ($left, $right) = @_;  # I: Array references
    # Verify we have the same number of elements in the array
    return FAILURE unless (scalar @$left == scalar @$right);
    # Compare each hash element
    for (my $index = 0; $index < scalar @$left; $index++)
    {
        return FAILURE unless
            compareHashes($left->[$index], $right->[$index]) == SUCCESS;
    }
    return SUCCESS;
}
#------------------------------------------------------------------------------
# Name: compareHashes($$)
#
# Description: Compares two hash structures to see if they contain the same
# keys and values.
#
# Exits with: SUCCESS (same), FAILURE (different).
#------------------------------------------------------------------------------
sub compareHashes($$)
{
    my ($left, $right) = @_; # I: Hash references

    # Verify the same number of keys in the hash
    return FAILURE unless (scalar keys %$left == scalar keys %$right);
    # Compare each key
    foreach my $key (keys %$left)
    {
        return FAILURE unless defined $right->{$key}
            && $left->{$key} eq $right->{$key};
    }
    return SUCCESS; # 100% match
}
#------------------------------------------------------------------------------
# Name: ProcessOptions()
#
# Description:  Using Getopt::Long, this routine parses the argument list,
# flagging any options found in the parameter list.  If an invalid option is
# given or the help option is requested, a usage message is shown and the
# program exits immediately.
#
#------------------------------------------------------------------------------
sub ProcessOptions()
{
    my $help;
    my $result = GetOptions(
        'debug|d=i' => \$debug_scenario,
        "help|h" => \$help);
    ShowUsage(1) if (!$result);
    ShowUsage(0) if $help;
    $debug_scenario = 0 unless (defined $debug_scenario
        && $debug_scenario >= 1 && $debug_scenario <= 5);
}
#------------------------------------------------------------------------------
# Name: ShowUsage($)
#
# Description: Shows the help or usage directives for the program.
#
# Exits with parameter value.
#------------------------------------------------------------------------------
sub ShowUsage($)
{
    my $exitvalue = shift; # I: Return code
    print << "EOF";
This program (test_set_char_validity.pl) runs unit tests against the
set_characterization_validity.pl script.
Options:
  --debug (-d) <test number>
        Allows you to stage the data (including creating a DSC.odl file and
inserting rows into the database) for manually debugging of the 
set_characterization_validity.pl script.  The script will exit just prior to the
test scenario (1 - 5) specified.
  --help (-h)
        Prints a help summary.
EOF
    exit($exitvalue);
}

#------------------------------------------------------------------------------
# Name: DIEHandler
#
# Description: Traps the die signal and issues an error message to STDERR.
#
# Exits with: FAILURE (return status = 1)
#------------------------------------------------------------------------------
sub DIEHandler
{
    my ($errormessage) = @_; # I: Die error message
    eval
    {
        local $SIG{__DIE__};
        chomp $errormessage;
        print STDERR "$errormessage\n";
    };
    exit FAILURE;
}
#------------------------------------------------------------------------------
# Name: END
#
# Description: Closes any resources used by this script.
#
#----------------------------------------------------------------------------
END
{
    eval
    {
        local $SIG{__DIE__};
        if (defined $dbh)
        {
            $dbh->rollback();
            $dbh->disconnect;
        }
    };
}
