/******************************************************************
Name: assign_bpf_name

Purpose: Retreives the BPF name from the BPF and changes the current
         BPF to the "real" name.

Notes: The BPF is currently a temporary name based on the day/time.  Within
       the BPF, file attributes contains the standard file name.  The BPF
       name will change from the temporary name to this new file name.

Returns: SUCCESS or ERROR
*******************************************************************/

#include <string.h>
#include <stdlib.h>
#include "setup_work_order.h"
#include "ias_logging.h"
#include "ias_bpf.h"

int assign_bpf_name
(
    char path[PATH_MAX] /* IO: Path to a BPF */
)
{
    IAS_BPF *                              bpf       = NULL;
    const struct IAS_BPF_FILE_ATTRIBUTES * file_attr = NULL;

    /* Read the BPF */
    bpf = ias_bpf_read(path);
    if (bpf == NULL)
    {
        IAS_LOG_ERROR("Cannot read BPF file %s", path);
        return ERROR;
    }

    /* Get the file attributes group */
    file_attr = ias_bpf_get_file_attributes(bpf);
    if (file_attr == NULL)
    {
        IAS_LOG_ERROR("Getting File Attributes group from BPF %s", path);
        ias_bpf_free(bpf);
        return ERROR;
    }

    /* If the BPF name is different, rename it */
    if (strcmp(path, file_attr->file_name) != 0)
    {
        if (strlen(file_attr->file_name) >= PATH_MAX)
        {
            long int my_size = (long int) strlen(file_attr->file_name);
            IAS_LOG_ERROR(
                "BPF file name exceeds buffer size, requires %lu characters",
                my_size);
            ias_bpf_free(bpf);
            return ERROR;
        }
        IAS_LOG_DEBUG("Rename BPF %s -> %s", path, file_attr->file_name);
        if (rename(path, file_attr->file_name))
        {
            IAS_LOG_ERROR("Unable to rename BPF %s -> %s", path,
                file_attr->file_name);
            ias_bpf_free(bpf);
            return ERROR;
        }

        /* This strcpy() is protected by the if statement above.  If the new
           path is too long, the rename action shouldn't be attempted either. */
        strcpy(path, file_attr->file_name);
    }
    ias_bpf_free(bpf);
    return SUCCESS;
}
