/******************************************************************
Name: calc_l0r_corners

Purpose: Calculate the corners for an L0R scene, using the worst case
point if both OLI and TIRS are present.

Returns: SUCCESS or ERROR
*******************************************************************/
#include "ias_logging.h"
#include "ias_geo.h"
#include "gctp.h"
#include "setup_work_order.h"

double get_lat_value(double center, double oli, double tirs);
double get_lng_value(double center, double oli, double tirs, int cross_180);

/* 
 * Calculates corner points for an L0R scene.  If both OLI and TIRS corners are
 * present, use the worst case points. If only OLI or TIRS are available uses
 * the set of points available. 
 *
 * Returns: SUCCESS or ERROR
 */

int calc_l0r_corners
(
    DSW_DATA *dsw_data /* I/O: selected scene metadata */
)
{
    IAS_L0R_CORNERS corners; /* O: calculated corners */

    if (dsw_data == NULL)
    {
        IAS_LOG_ERROR("dsw_data is a null pointer");
        return ERROR;
    }

    /* If there are no scene records present, skip scene corner processing */
    if (dsw_data->record_count > 0)
    {

        if (dsw_data->sensor_present_oli && dsw_data->sensor_present_tirs) 
        {
            /* A few temp variables to simplify variable access */
            double center_lat = dsw_data->scene_center.lat;
            double center_lng = dsw_data->scene_center.lng;
            double oli_val, tirs_val;

            /* Determine if the scene crosses 180 longitude */
            double lng_corners[4] = {
                dsw_data->corners_oli.upleft.lng,
                dsw_data->corners_oli.upright.lng,
                dsw_data->corners_oli.loright.lng,
                dsw_data->corners_oli.loleft.lng,
            };

            int cross_180 = ias_geo_does_cross_180(DEGREE, lng_corners);

            /* If the OLI bands don't cross 180, check if the TIRS bands do */
            if (!cross_180)
            {
                lng_corners[0] = dsw_data->corners_tirs.upleft.lng;
                lng_corners[1] = dsw_data->corners_tirs.upright.lng;
                lng_corners[2] = dsw_data->corners_tirs.loright.lng;
                lng_corners[3] = dsw_data->corners_tirs.loleft.lng;

                cross_180 = ias_geo_does_cross_180(DEGREE, lng_corners);
            }
    
            /* Find UPPER LEFT Latitude */
            oli_val = dsw_data->corners_oli.upleft.lat;
            tirs_val = dsw_data->corners_tirs.upleft.lat;

            corners.upleft.lat = get_lat_value(center_lat, oli_val, tirs_val);
        
            /* Find UPPER RIGHT Latitude */
            oli_val = dsw_data->corners_oli.upright.lat;
            tirs_val = dsw_data->corners_tirs.upright.lat;

            corners.upright.lat = get_lat_value(center_lat, oli_val, tirs_val);
        
            /* Find LOWER LEFT Latitude */
            oli_val = dsw_data->corners_oli.loleft.lat;
            tirs_val = dsw_data->corners_tirs.loleft.lat;

            corners.loleft.lat = get_lat_value(center_lat, oli_val, tirs_val);
        
            /* Find LOWER RIGHT Latitude */
            oli_val = dsw_data->corners_oli.loright.lat;
            tirs_val = dsw_data->corners_tirs.loright.lat;

            corners.loright.lat = get_lat_value(center_lat, oli_val, tirs_val);
     
            /* Find UPPER LEFT Longitude */
            oli_val = dsw_data->corners_oli.upleft.lng;
            tirs_val = dsw_data->corners_tirs.upleft.lng;
        
            corners.upleft.lng = get_lng_value(center_lng, oli_val, tirs_val, 
                cross_180);
 
            /* Find UPPER RIGHT Longitude */
            oli_val = dsw_data->corners_oli.upright.lng;
            tirs_val = dsw_data->corners_tirs.upright.lng;
        
            corners.upright.lng = get_lng_value(center_lng, oli_val, tirs_val,
                cross_180);
 
            /* Find LOWER LEFT Longitude */
            oli_val = dsw_data->corners_oli.loleft.lng;
            tirs_val = dsw_data->corners_tirs.loleft.lng;
 
            corners.loleft.lng = get_lng_value(center_lng, oli_val, tirs_val,
                cross_180);

            /* Find LOWER RIGHT Longitude */
            oli_val = dsw_data->corners_oli.loright.lng;
            tirs_val = dsw_data->corners_tirs.loright.lng;
        
            corners.loright.lng = get_lng_value(center_lng, oli_val, tirs_val,
                cross_180);
        }
        else if (dsw_data->sensor_present_oli)
        {
            corners = dsw_data->corners_oli;   
        }
        else if (dsw_data->sensor_present_tirs)
        {
            corners = dsw_data->corners_tirs;   
        }
        else
        {
            IAS_LOG_ERROR("Neither OLI nor TIRS present in scene metadata");
            return ERROR;
        }

        /* Copy the primary set of corners */
        dsw_data->upper_left_lat_lon[0] = corners.upleft.lat;
        dsw_data->upper_left_lat_lon[1] = corners.upleft.lng;
        dsw_data->upper_right_lat_lon[0] = corners.upright.lat;
        dsw_data->upper_right_lat_lon[1] = corners.upright.lng;
        dsw_data->lower_left_lat_lon[0] = corners.loleft.lat;
        dsw_data->lower_left_lat_lon[1] = corners.loleft.lng;
        dsw_data->lower_right_lat_lon[0] = corners.loright.lat;
        dsw_data->lower_right_lat_lon[1] = corners.loright.lng;
    }

    return SUCCESS;
}

/* Determine latitude value based on center point, oli, and tirs 
   latitude values */
double get_lat_value(double center, double oli, double tirs)
{
    double lat; /* return value */
    
    /* Use the center point to determine which of the OLI or TIRS corner points
       should be used. If the corners are greater than the center, use the 
       greater of the OLI or TIRS points.  If the corners are less than the 
       center, use the lesser of the OLI or TIRS points. */
    if (oli > center || tirs > center)
        lat = (oli > tirs) ? oli : tirs;
    else
        lat = (oli < tirs) ? oli : tirs;
    
    return lat;
}

/* Determine longitude value based on center point, oli, and tirs 
   longitude values */
double get_lng_value(double center, double oli, double tirs, int cross_180)
{
    double lng;  /* return value */

    /* if the scene crosses 180, convert all longitudes to positive values */
    if (cross_180)
    {
        if (center < 0.0)
            center += 360.0;

        if (oli < 0.0)
            oli += 360.0;

        if (tirs < 0.0)
            tirs += 360.0;
    }

    /* Use the center point to determine which of the OLI or TIRS corner points
       should be used. If the corners are greater than the center, use the 
       greater of the OLI or TIRS points.  If the corners are less than the 
       center, use the lesser of the OLI or TIRS points. */
    if (oli > center || tirs > center)
        lng = (oli > tirs) ? oli : tirs;
    else
        lng = (oli < tirs) ? oli : tirs;

    /* if necessary, covert longitude back to negative value */
    if (lng > 180.0)
        lng -= 360.0;

    return lng;
}

