/**************************************************************************
 NAME:          extend_corners

 PURPOSE:       Extends the corners in the metadata based on CPF field of view
                (FOV).

 RETURNS:       Status code of SUCCESS or ERROR
***************************************************************************/
#include "ias_logging.h"
#include "setup_work_order.h"

int extend_corners
(
    DSW_DATA *dsw_data,      /* I/O: Scene metadata for bands */
    const struct IAS_CPF_EARTH_CONSTANTS *earth_parameters,
                             /* I: Earth constants */
    const struct IAS_CPF_ORBIT_PARAMETERS *orbit_parameters,
                             /* I: Orbit parameters */
    const struct IAS_CPF_FOCAL_PLANE *focal_plane
                             /* I: Focal plane parameters */
)
{
    IAS_L0R_CORNERS new_oli;         /* Extended OLI corners */
    IAS_L0R_CORNERS new_tirs;        /* Extended TIRS corners */
    IAS_EARTH_CHARACTERISTICS earth; /* Structure of Earth constants */
    double fov_at;                   /* Along track field of view */
    double fov_xt;                   /* Across track field of view */
    int status;                      /* General routine status */

    /* Extend corners is only needed for Earth scenes */
    if (dsw_data->acquisition_type != IAS_EARTH)
    {
        return SUCCESS;
    }

    IAS_LOG_DEBUG("Extend Corners started");

    /* Check for at least one sensor */
    if ((dsw_data->sensor_present_oli != 1)
        && (dsw_data->sensor_present_tirs != 1))
    {
        IAS_LOG_ERROR("Neither OLI nor TIRS is present");
        return ERROR;
    }

    /* Debugging information -- before corner extension */
    IAS_LOG_DEBUG("Before:\n");
    IAS_LOG_DEBUG("OLI:\nUL: %lf,%lf\tUR: %lf,%lf\nLL: %lf,%lf\tLR: %lf,%lf\n",
        dsw_data->corners_oli.upleft.lat, dsw_data->corners_oli.upleft.lng,
        dsw_data->corners_oli.upright.lat, dsw_data->corners_oli.upright.lng,
        dsw_data->corners_oli.loleft.lat, dsw_data->corners_oli.loleft.lng,
        dsw_data->corners_oli.loright.lat, dsw_data->corners_oli.loright.lng);

    IAS_LOG_DEBUG("TIRS:\nUL: %lf,%lf\tUR: %lf,%lf\nLL: %lf,%lf\tLR: %lf,%lf\n",
        dsw_data->corners_tirs.upleft.lat, dsw_data->corners_tirs.upleft.lng,
        dsw_data->corners_tirs.upright.lat, dsw_data->corners_tirs.upright.lng,
        dsw_data->corners_tirs.loleft.lat, dsw_data->corners_tirs.loleft.lng,
        dsw_data->corners_tirs.loright.lat, dsw_data->corners_tirs.loright.lng);
    /* Load the Earth characteristics structure */
    earth.ut1_utc_correction = 0.0; /* not used here */
    earth.pole_wander_x = 0.0;      /* not used here */
    earth.pole_wander_y = 0.0;      /* not used here */
    earth.semi_major_axis = earth_parameters->semi_major_axis;
    earth.semi_minor_axis = earth_parameters->semi_minor_axis;
    earth.eccentricity = earth_parameters->eccentricity;
    earth.earth_angular_velocity = earth_parameters->earth_angular_velocity;
    earth.speed_of_light = earth_parameters->speed_of_light;

    /* Debugging information:  off-nadir (roll) angle */
    IAS_LOG_DEBUG("Roll Angle: %lf\n", dsw_data->off_nadir_angle);

    /* Extend the OLI corners */
    if (dsw_data->sensor_present_oli == 1)
    {
        /* Calculate the OLI instrument fields of view */
        status = find_fov(IAS_OLI, focal_plane, &fov_at, &fov_xt);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Computing instrument FOVs based on CPF");
            return ERROR;
        }

        status = move_corners(&dsw_data->corners_oli,
            dsw_data->off_nadir_angle, &dsw_data->scene_center, &earth, 
            orbit_parameters->nominal_orbit_radius * 1000.0, fov_at, fov_xt, 
            &new_oli);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("OLI corner extension failed");
            return ERROR;
        }
        else
        {
            dsw_data->corners_oli = new_oli;
        }
    }

    /* Extend the TIRS corners */
    if (dsw_data->sensor_present_tirs == 1)
    {
        /* Calculate the TIRS instrument fields of view */
        status = find_fov(IAS_TIRS, focal_plane, &fov_at, &fov_xt);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Computing instrument FOVs based on CPF");
            return ERROR;
        }

        status = move_corners(&dsw_data->corners_tirs, 
            dsw_data->off_nadir_angle, &dsw_data->scene_center, &earth, 
            orbit_parameters->nominal_orbit_radius * 1000.0, fov_at, fov_xt,
            &new_tirs);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("TIRS corner extension failed");
            return ERROR;
        }
        else
        {
            dsw_data->corners_tirs = new_tirs;
        }
    }

    /* Debugging information -- after corner extension */
    IAS_LOG_DEBUG("After:\n");
    IAS_LOG_DEBUG("OLI:\nUL: %lf,%lf\tUR: %lf,%lf\nLL: %lf,%lf\tLR: %lf,%lf\n",
        dsw_data->corners_oli.upleft.lat, dsw_data->corners_oli.upleft.lng,
        dsw_data->corners_oli.upright.lat, dsw_data->corners_oli.upright.lng,
        dsw_data->corners_oli.loleft.lat, dsw_data->corners_oli.loleft.lng,
        dsw_data->corners_oli.loright.lat, dsw_data->corners_oli.loright.lng);

    IAS_LOG_DEBUG("TIRS:\nUL: %lf,%lf\tUR: %lf,%lf\nLL: %lf,%lf\tLR: %lf,%lf\n",
        dsw_data->corners_tirs.upleft.lat, dsw_data->corners_tirs.upleft.lng,
        dsw_data->corners_tirs.upright.lat, dsw_data->corners_tirs.upright.lng,
        dsw_data->corners_tirs.loleft.lat, dsw_data->corners_tirs.loleft.lng,
        dsw_data->corners_tirs.loright.lat, dsw_data->corners_tirs.loright.lng);

    IAS_LOG_DEBUG("Extend Corners finished");

    return SUCCESS;
}
