/**************************************************************************
 NAME:                      find_fov

 PURPOSE:       
 Calculate the fields of view for a given sensor using CPF parameters.

 RETURNS:       
 Status code of SUCCESS or ERROR
***************************************************************************/
#include "ias_logging.h"
#include "ias_cpf.h"

int find_fov
(
    IAS_SENSOR_ID sensor, /* I: ID of sensor whose fields of view will be 
                             calculated */ 
    const IAS_CPF_FOCAL_PLANE *focal, /* I: Focal plane parameters from the
                             CPF */
    double *fov_at,       /* O: Along-track field of view */
    double *fov_xt        /* O: Across-track field of view */
)
{
    double at_angle_min;  /* Minimum along-track FOV angle */
    double at_angle_max;  /* Maximum along-track FOV angle */
    double xt_angle_min;  /* Minimum across-track FOV angle */
    double xt_angle_max;  /* Maximum across-track FOV angle */
    double at_angle;      /* Along-track FOV angle for one point */
    double xt_angle;      /* Across-track FOV angle for one point */
    int band_index;       /* Band index */
    int band_numbers[IAS_MAX_NBANDS]; /* List of band numbers for a sensor */
    int band_list_index;  /* Index into band_numbers list */
    int num_sensor_bands; /* Total number of bands for the sensor */
    int sca_index;        /* SCA index */
    int number_of_scas;   /* Number of SCAs for a band */
    int status;           /* Function return status */

    /* Calculate the instrument fields of view */
    at_angle_min = xt_angle_min = 1.0e6;
    at_angle_max = xt_angle_max = -1.0e6;

    /* Get list of bands for the sensor */ 
    status = ias_sat_attr_get_sensor_band_numbers(sensor, IAS_NORMAL_BAND, 
        0, band_numbers, IAS_MAX_NBANDS, &num_sensor_bands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting normal band numbers for sensor %d", sensor);
        return ERROR;
    }

    for (band_list_index = 0; band_list_index < num_sensor_bands; 
         band_list_index++)
    {
        /* Get the band index for the current band number in the list */
        band_index = ias_sat_attr_convert_band_number_to_index(
            band_numbers[band_list_index]);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Invalid band index for band number %d",
                band_numbers[band_list_index]);
            return ERROR;
        }

        /* Get the number of SCAs for the band */
        number_of_scas = ias_sat_attr_get_scas_per_band(
            band_numbers[band_list_index]);
        if (number_of_scas == ERROR)
        {
            IAS_LOG_ERROR("Invalid number of SCAs retrieved for band %d",
                band_numbers[band_list_index]);
            return ERROR;
        }

        for (sca_index = 0; sca_index < number_of_scas; sca_index++)
        {
            at_angle = focal->along_los_legendre[band_index][sca_index][0]
                     + focal->along_los_legendre[band_index][sca_index][1]
                     + focal->along_los_legendre[band_index][sca_index][2]
                     + focal->along_los_legendre[band_index][sca_index][3];
            if (at_angle_min > at_angle) 
                at_angle_min = at_angle;
            if (at_angle_max < at_angle) 
                at_angle_max = at_angle;

            xt_angle = focal->across_los_legendre[band_index][sca_index][0]
                     + focal->across_los_legendre[band_index][sca_index][1]
                     + focal->across_los_legendre[band_index][sca_index][2]
                     + focal->across_los_legendre[band_index][sca_index][3];
            if (xt_angle_min > xt_angle) 
                xt_angle_min = xt_angle;
            if (xt_angle_max < xt_angle) 
                xt_angle_max = xt_angle;

            at_angle = focal->along_los_legendre[band_index][sca_index][0]
                     - focal->along_los_legendre[band_index][sca_index][1]
                     + focal->along_los_legendre[band_index][sca_index][2]
                     - focal->along_los_legendre[band_index][sca_index][3];
            if (at_angle_min > at_angle) 
                at_angle_min = at_angle;
            if (at_angle_max < at_angle) 
                at_angle_max = at_angle;

            xt_angle = focal->across_los_legendre[band_index][sca_index][0]
                     - focal->across_los_legendre[band_index][sca_index][1]
                     + focal->across_los_legendre[band_index][sca_index][2]
                     - focal->across_los_legendre[band_index][sca_index][3];
            if (xt_angle_min > xt_angle) 
                xt_angle_min = xt_angle;
            if (xt_angle_max < xt_angle) 
                xt_angle_max = xt_angle;
        }
    }

    *fov_at = at_angle_max - at_angle_min;
    *fov_xt = xt_angle_max;
    if (-xt_angle_min > xt_angle_max) 
        *fov_xt = -xt_angle_min;

    return SUCCESS;
}
