/******************************************************************
Name: generate_omf

Purpose: Writes out the OMF file

Returns: SUCCESS or ERROR
*******************************************************************/

#include <string.h>
#include "gctp.h"                      /* For UTM definition */
#include "ias_logging.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "setup_work_order.h"

int generate_omf
(
    DSW_DATA *dsw_data                  /* I: Parameters and values retrieved 
                                              from parameter file and L0R */
)
{
    FILE *out_fp;                       /* Output file pointer */
    int write_status;                   /* Status from writing parm file */
    int write_count;                    /* Total of parms to write */
    int status;                         /* Function return status */

    /* Open a file to write the OMF */
    out_fp = fopen(dsw_data->omf_filename,"w");
    if (out_fp == NULL)
    {
        IAS_LOG_ERROR("Opening OMF file for writing: %s",
                dsw_data->omf_filename);
        return ERROR;
    }

    /* Write a basic OMF shell */
    status = fprintf(out_fp,"OBJECT = OMF\n"
            "END_OBJECT = OMF\n"
            "END\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing OMF %s", dsw_data->omf_filename);
        fclose(out_fp);
        return ERROR;
    }

    /* Close the OMF file */
    status = fclose(out_fp);
    if (status != 0)
    {
        IAS_LOG_ERROR("Closing OMF %s", dsw_data->omf_filename);
        return ERROR;
    }

    /* If the CPF, BPF and RLUT are included, write the names */
    if (dsw_data->retrieve_calibration_files_flag)
    {
        /* Set up the parameter table */
        IAS_PARM_DECLARE_TABLE(omf_parm_table, 2);

        /* Add the CPF filename */
        IAS_PARM_CPF(omf_parm_table, IAS_PARM_REQUIRED, dsw_data->cpf_filename,
            sizeof(dsw_data->cpf_filename), 1);
        IAS_PARM_SET_COUNT_WRITE(omf_parm_table, CAL_PARM_FILENAME, 1);

        /* Add the OLI BPF filename */
        if (dsw_data->sensor_present_oli)
        {
            /* Set up the parameter table */
            IAS_PARM_DECLARE_TABLE(omf_parm_table, 1);

            IAS_PARM_BPF_OLI(omf_parm_table, IAS_PARM_OPTIONAL,
                dsw_data->bpf_filename_oli, sizeof(dsw_data->bpf_filename_oli),
                1);
            IAS_PARM_SET_COUNT_WRITE(omf_parm_table, BIAS_PARM_FILENAME_OLI, 1);
            /* Get the write count at this point, then write the parameters 
               out to file */
            write_count = IAS_PARM_GET_TABLE_SIZE(omf_parm_table);
            write_status = ias_parm_update(dsw_data->omf_filename,
                omf_parm_table, write_count);
            if (write_status != SUCCESS)
            {
                IAS_LOG_ERROR("Writing general parameters to OMF file %s",
                    dsw_data->omf_filename);
                return ERROR;
            }
        }

        /* Add the TIRS BPF filename */
        if (dsw_data->sensor_present_tirs)
        {
            /* Set up the parameter table */
            IAS_PARM_DECLARE_TABLE(omf_parm_table, 1);

            IAS_PARM_BPF_TIRS(omf_parm_table, IAS_PARM_OPTIONAL,
                dsw_data->bpf_filename_tirs, 
                sizeof(dsw_data->bpf_filename_tirs), 1);
            IAS_PARM_SET_COUNT_WRITE(omf_parm_table, BIAS_PARM_FILENAME_TIRS, 
                1);

            /* Get the write count at this point, then write the parameters 
               out to file */
            write_count = IAS_PARM_GET_TABLE_SIZE(omf_parm_table);
            write_status = ias_parm_update(dsw_data->omf_filename,
                omf_parm_table, write_count);
            if (write_status != SUCCESS)
            {
                IAS_LOG_ERROR("Writing general parameters to OMF file %s",
                    dsw_data->omf_filename);
                return ERROR;
            }
        }

        /* Add the RLUT filename */
        IAS_PARM_RLUT(omf_parm_table, IAS_PARM_REQUIRED, 
            dsw_data->rlut_filename, sizeof(dsw_data->rlut_filename), 1);
        IAS_PARM_SET_COUNT_WRITE(omf_parm_table, RLUT_FILENAME, 1);

        /* Get the write count at this point, then write the parameters 
           out to file */
        write_count = IAS_PARM_GET_TABLE_SIZE(omf_parm_table);
        write_status = ias_parm_update(dsw_data->omf_filename,
            omf_parm_table, write_count);
        if (write_status != SUCCESS)
        {
            IAS_LOG_ERROR("Writing general parameters to OMF file %s",
                dsw_data->omf_filename);
            return ERROR;
        }
    }   /* End of section with calibration files included */

    /* This section is included in all products */

    /* Set up the parameter table */
    IAS_PARM_DECLARE_TABLE(omf_parm_table, 10);

    /* Add the L0R filename */
    IAS_PARM_L0R_FILENAME(omf_parm_table, IAS_PARM_REQUIRED,
            dsw_data->l0r_filename, sizeof(dsw_data->l0r_filename), 1);
    IAS_PARM_SET_COUNT_WRITE(omf_parm_table, L0R_FILENAME, 1);

    /* Add the software version */
    IAS_PARM_ADD_STRING(omf_parm_table, SOFTWARE_VERSION, "Software version",
            IAS_PARM_REQUIRED, 0, NULL, 0, NULL, dsw_data->sw_version,
            sizeof(dsw_data->sw_version), 0);
    IAS_PARM_SET_COUNT_WRITE(omf_parm_table, SOFTWARE_VERSION, 1);

    /* Add the spacecraft name */
    IAS_PARM_SPACECRAFT(omf_parm_table, IAS_PARM_REQUIRED,
            dsw_data->spacecraft_id, sizeof(dsw_data->spacecraft_id), 1);
    IAS_PARM_SET_COUNT_WRITE(omf_parm_table, SPACECRAFT, 1);

    /* Add the processing center name */
    IAS_PARM_ADD_STRING(omf_parm_table, PROCESSING_CENTER,
            "Processing center name", IAS_PARM_REQUIRED, 0, NULL, 0, NULL, 
            dsw_data->processing_center, sizeof(dsw_data->processing_center),
            1);
    IAS_PARM_SET_COUNT_WRITE(omf_parm_table, PROCESSING_CENTER, 1);

    /* Add the maximum processors to use */
    IAS_PARM_MAX_PROCESSORS_TO_USE(omf_parm_table, IAS_PARM_REQUIRED,
            &dsw_data->max_processors, sizeof(dsw_data->max_processors), 1);
    IAS_PARM_SET_COUNT_WRITE(omf_parm_table, MAX_PROCESSORS_TO_USE, 1);

    /* Add the acquisition type */
    IAS_PARM_ACQUISITION_TYPE(omf_parm_table, &dsw_data->acquisition_type,
            sizeof(dsw_data->acquisition_type), 1);
    IAS_PARM_SET_COUNT_WRITE(omf_parm_table, ACQUISITION_TYPE, 1);

    /* Add the Product Request ID */
    IAS_PARM_PRODUCT_REQUEST_ID(omf_parm_table,
        dsw_data->product_request_id,
        sizeof(dsw_data->product_request_id), 1);
    IAS_PARM_SET_COUNT_WRITE(omf_parm_table, PRODUCT_REQUEST_ID, 1);

    /* Add the sensor ID */
    IAS_PARM_ADD_STRING(omf_parm_table, SENSOR, "Sensor ID", IAS_PARM_REQUIRED,
            0, NULL, 0, NULL, dsw_data->sensor_id_from_l0r,
            sizeof(dsw_data->sensor_id_from_l0r), 0);
    IAS_PARM_SET_COUNT_WRITE(omf_parm_table, SENSOR, 1);

    /* Add the satellite ID */
    IAS_PARM_ADD_INT(omf_parm_table, SATELLITE, "Satellite ID",
            IAS_PARM_REQUIRED, IAS_PARM_NOT_ARRAY, 0, 0, 0, 0, NULL,
            &dsw_data->satellite_id, sizeof(dsw_data->satellite_id), 1);
    IAS_PARM_SET_COUNT_WRITE(omf_parm_table, SATELLITE, 1);

    /* Add the acquisition date */
    IAS_PARM_ADD_STRING(omf_parm_table, ACQUISITION_DATE, "Acquisition Date",
            IAS_PARM_REQUIRED, 0, NULL, 0, NULL,
            dsw_data->acquisition_date_string,
            sizeof(dsw_data->acquisition_date_string), 1);
    IAS_PARM_SET_COUNT_WRITE(omf_parm_table, ACQUISITION_DATE, 1);

    /* Get the write count at this point for the general set of
       parameters, then write the parameters out to file */
    write_count = IAS_PARM_GET_TABLE_SIZE(omf_parm_table);
    write_status = ias_parm_update(dsw_data->omf_filename,
        omf_parm_table, write_count);
    if (write_status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing general parameters to OMF file %s",
            dsw_data->omf_filename);
        return ERROR;
    }

    /* If we're processing an Earth scene, get the remaining Earth
       material together as a separate parameter table and write it to the
       OMF.  Otherwise, skip this section */
    if (dsw_data->acquisition_type == IAS_EARTH)
    {
        /* Only include corners if the calibration files were retrieved */
        if (dsw_data->retrieve_calibration_files_flag)
        {
            double default_scene_boundary_lat_lon[] = {0.0, 0.0};

            /* Declare a new table */
            IAS_PARM_DECLARE_TABLE(earth_parm_table, 4);

            /* Add the upper left lat lon */
            IAS_PARM_ADD_DOUBLE(earth_parm_table, UL_BOUNDARY_LAT_LON,
                "Scene boundary - upper left lat/lon", IAS_PARM_REQUIRED,
                IAS_PARM_ARRAY, 0, 0.0, 0.0, 1, default_scene_boundary_lat_lon, 
                (double *)&dsw_data->upper_left_lat_lon,
                sizeof(dsw_data->upper_left_lat_lon), 1);
            IAS_PARM_SET_COUNT_WRITE(earth_parm_table, UL_BOUNDARY_LAT_LON, 2);

            /* Add the upper right lat lon */
            IAS_PARM_ADD_DOUBLE(earth_parm_table, UR_BOUNDARY_LAT_LON,
                "Scene boundary - upper right lat/lon", IAS_PARM_REQUIRED,
                IAS_PARM_ARRAY, 0, 0.0, 0.0, 1, default_scene_boundary_lat_lon, 
                (double *)&dsw_data->upper_right_lat_lon,
                sizeof(dsw_data->upper_right_lat_lon), 1);
            IAS_PARM_SET_COUNT_WRITE(earth_parm_table, UR_BOUNDARY_LAT_LON, 2);

            /* Add the lower left lat lon */
            IAS_PARM_ADD_DOUBLE(earth_parm_table, LL_BOUNDARY_LAT_LON,
                "Scene boundary - lower left lat/lon", IAS_PARM_REQUIRED,
                IAS_PARM_ARRAY, 0, 0.0, 0.0, 1, default_scene_boundary_lat_lon, 
                (double *)&dsw_data->lower_left_lat_lon,
                sizeof(dsw_data->lower_left_lat_lon), 1);
            IAS_PARM_SET_COUNT_WRITE(earth_parm_table, LL_BOUNDARY_LAT_LON, 2);

            /* Add the lower right lat lon */
            IAS_PARM_ADD_DOUBLE(earth_parm_table, LR_BOUNDARY_LAT_LON,
                "Scene boundary - lower right lat/lon", IAS_PARM_REQUIRED,
                IAS_PARM_ARRAY, 0, 0.0, 0.0, 1, default_scene_boundary_lat_lon, 
                (double *)&dsw_data->lower_right_lat_lon,
                sizeof(dsw_data->lower_right_lat_lon), 1);
            IAS_PARM_SET_COUNT_WRITE(earth_parm_table, LR_BOUNDARY_LAT_LON, 2);

            /* Get the write count for the Earth-specific parameter table */
            write_count = IAS_PARM_GET_TABLE_SIZE(earth_parm_table);

            /* Write the information to the OMF */
            write_status = ias_parm_update(dsw_data->omf_filename, 
                earth_parm_table, write_count);
            if (write_status != SUCCESS)
            {
                IAS_LOG_ERROR(
                    "Writing Corner parameters to OMF file %s",
                    dsw_data->omf_filename);
                return ERROR;
            }
        }   /* End of if retrieved calibration files for Earth scenes */

        /* Declare a new table for all Earth products */
        IAS_PARM_DECLARE_TABLE(earth_parm_table, 5);

        /* Add the target WRS path */
        IAS_PARM_TARGET_WRS_PATH(earth_parm_table, IAS_PARM_REQUIRED,
            &dsw_data->path, sizeof(dsw_data->path), 1);
        IAS_PARM_SET_COUNT_WRITE(earth_parm_table, TARGET_WRS_PATH, 1);

        /* Add the target WRS row */
        IAS_PARM_TARGET_WRS_ROW(earth_parm_table, IAS_PARM_REQUIRED,
            &dsw_data->row, sizeof(dsw_data->row), 1);
        IAS_PARM_SET_COUNT_WRITE(earth_parm_table, TARGET_WRS_ROW, 1);

        /* Add the projection */
        IAS_PARM_TARGET_PROJECTION(earth_parm_table, &dsw_data->projection,
            sizeof(dsw_data->projection), 1);
        IAS_PARM_SET_COUNT_WRITE(earth_parm_table, TARGET_PROJECTION, 1);

        /* Add the UTM Zone */ 
        IAS_PARM_UTM_ZONE(earth_parm_table, IAS_PARM_REQUIRED, 0, NULL,
            &dsw_data->utm_zone, sizeof(dsw_data->utm_zone), 1);
        IAS_PARM_SET_COUNT_WRITE(earth_parm_table, UTM_ZONE, 1);

        /* Add the Landsat Scene ID */
        IAS_PARM_LANDSAT_SCENE_ID(earth_parm_table, IAS_PARM_OPTIONAL,
            dsw_data->landsat_scene_id, sizeof(dsw_data->landsat_scene_id),
            0);
        IAS_PARM_SET_COUNT_WRITE(earth_parm_table, LANDSAT_SCENE_ID, 1);

        /* Get the write count for the Earth-specific parameter table */
        write_count = IAS_PARM_GET_TABLE_SIZE(earth_parm_table);

        /* Write the information to the OMF */
        write_status = ias_parm_update(dsw_data->omf_filename, 
            earth_parm_table, write_count);
        if (write_status != SUCCESS)
        {
            IAS_LOG_ERROR("Writing Earth collection parameters to OMF file %s",
                dsw_data->omf_filename);
            return ERROR;
        }
    }   /* END -- Earth parameter table write */

    return SUCCESS;
}
