/******************************************************************
Name: insert_geometric_row

Purpose: Insert a record into the geometric work order common table using 
         information from the work order parameters, L0R metadata,
         and software version. Sensor ID should be set based on the
         type of work order (based on a parameter in the parameter file), not 
         from the input data.

Returns: SUCCESS or ERROR
*******************************************************************/

#include <stdlib.h>
#include <string.h>
#include "ias_logging.h"
#include "ias_db.h"
#include "setup_work_order.h"
#include "ias_miscellaneous.h"

int insert_geometric_row
(
    const DSW_DATA *dsw_data,          /* I: Parameters and values retrieved 
                                             from parameter file and L0R */
    struct ias_db_connection *db       /* I: Database connection */
)
{
    int status;                        /* Return status */
    char sql[1024];                    /* Buffer for sql commands */
    char buffer[IAS_SQL_RTN_MSGLEN];   /* Buffer for sql return messages */
    struct ias_db_query *query = NULL; /* DB query structure */
    char temp_sensor_id[IAS_L0R_SENSOR_ID_LENGTH];/* Temporary string for 
                                                     sensor ID */
    char date_acquired[DATE_LENGTH];   /* String for format-converted date */
    char path_str[10];                 /* path used to build sql statemet */
    char row_str[10];                  /* row used to build sql statement */
    char char_id[IAS_CHAR_ID_SIZE];    /* Key value in
                                        RADIOMETRIC_WORK_ORDER_COMMON table
                                        (needed to associate processing
                                        with the correct work order)*/
    char *processing_system = NULL;    /* Processing system environment */
    char sw_version_lpgs[IAS_SOFTWARE_VERSION_SIZE + 2];
                                       /* LPGS software version plus quotes */
    char sw_version_ias[IAS_SOFTWARE_VERSION_SIZE + 2];
                                       /* IAS software version plus quotes */

    /* If the GEOMETRY_SENSOR_ID_OVERRIDE parameter is present in the 
       parameter file, use that instead of the L0R value to populate the 
       SENSOR_ID field of the GEOMETRIC_WORK_ORDER_COMMON table */
    if (strcmp(dsw_data->geometry_sensor_id_override, "") != 0)
    {
        strncpy(temp_sensor_id, dsw_data->geometry_sensor_id_override,
            sizeof(temp_sensor_id));
    }
    else
    {
        strncpy(temp_sensor_id, dsw_data->sensor_id_from_l0r,
            sizeof(temp_sensor_id));
    }
    /* make sure it is zero-terminated */
    temp_sensor_id[sizeof(temp_sensor_id) - 1] = '\0';

    /* Convert the date to an Oracle time string */
    snprintf(date_acquired, sizeof(date_acquired),
        "%04d-%02d-%02d %02d:%02d:%02d",
        dsw_data->acquisition_date.year, 
        dsw_data->acquisition_date.month, 
        dsw_data->acquisition_date.day_of_month,
        dsw_data->acquisition_date.hour, 
        dsw_data->acquisition_date.minute, 
        (int)dsw_data->acquisition_date.second);

    /* Only Earth facing scenes will have a path/row */
    if (dsw_data->acquisition_type == IAS_EARTH)
    {
        sprintf(path_str, "'%d'", dsw_data->path);
        sprintf(row_str, "'%d'", dsw_data->row);
    }
    else
    {
        sprintf(path_str, "NULL");
        sprintf(row_str, "NULL");
    }

    status = ias_misc_get_char_id(db, "GEOMETRIC_WORK_ORDER_COMMON",
                "GEOMETRIC_WO_COMMON_ID", dsw_data->work_order_id,
                char_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to get the Geometric Characterization ID");
        return ERROR;
    }

    /* Load the software version based on the PROCESSING_SYSTEM env */
    processing_system = getenv("PROCESSING_SYSTEM");
    if (processing_system == NULL)
    {
        IAS_LOG_ERROR("PROCESSING_SYSTEM environment variable is not "
            "defined to either IAS or LPGS");
        return ERROR;
    }
    else if (strncmp(processing_system, "IAS", 3) == 0)
    {
        sprintf(sw_version_ias, "'%s'", dsw_data->sw_version);
        sprintf(sw_version_lpgs, "NULL");
    }
    else if (strncmp(processing_system, "LPGS", 4) == 0)
    {
        sprintf(sw_version_lpgs, "'%s'", dsw_data->sw_version);
        sprintf(sw_version_ias, "NULL");
    }
    else
    {
        IAS_LOG_ERROR("PROCESSING_SYSTEM of '%s' is not the expected "
            "IAS or LPGS value", processing_system);
        return ERROR;
    }

    /* Build the SQL statement */
    sprintf(sql, "UPDATE GEOMETRIC_WORK_ORDER_COMMON "
                 "SET "
                 " DATE_ACQUIRED = to_date('%s', 'YYYY-MM-DD HH24:MI:SS'), "
                 " DATE_PROCESSED = SYSDATE, "
                 " SENSOR_ID = '%s', "
                 " PROCESSING_CENTER = '%s', "
                 " TARGET_WRS_PATH = %s, "     /* path already quoted */
                 " TARGET_WRS_ROW = %s, "      /* row already quoted */
                 " SOFTWARE_VERSION_IS = '%s', "
                 " SOFTWARE_VERSION_LPGS = %s, "
                 " SOFTWARE_VERSION_IAS = %s, "
                 " OFF_NADIR_ANGLE = '%f', "
                 " COLLECTION_TYPE = '%s', "
                 " LANDSAT_SCENE_ID = '%s', "
                 " L0RP_IMAGE_FILE = '%s', "
                 " CPF_NAME = '%s', "
                 " LANDSAT_INTERVAL_ID = '%s', "
                 " INTEGRATION_TIME_OLI_PAN = '%f', "
                 " INTEGRATION_TIME_OLI_MS = '%f', "
                 " INTEGRATION_TIME_TIRS = '%f', "
                 " STATION_ID = '%s', "
                 " COLLECTION_NUMBER = %d "
                 " WHERE GEOMETRIC_WO_COMMON_ID = '%s'",
                 date_acquired,
                 temp_sensor_id,
                 dsw_data->processing_center,
                 path_str,
                 row_str,
                 dsw_data->is_version,
                 sw_version_lpgs,
                 sw_version_ias,
                 dsw_data->off_nadir_angle,
                 dsw_data->collection_type,
                 dsw_data->landsat_scene_id,
                 dsw_data->l0r_id,    /* use filename without path */
                 dsw_data->cpf_filename,
                 dsw_data->landsat_interval_id,
                 dsw_data->integration_time_oli_pan,
                 dsw_data->integration_time_oli_ms,
                 dsw_data->integration_time_tirs,
                 dsw_data->station_id,
                 dsw_data->collection_number,
                 char_id
           );

    /* Perform the query */
    query = ias_db_perform_query(db, sql);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, buffer, sizeof(buffer));
        IAS_LOG_ERROR("Inserting: %s", buffer);
        ias_db_query_close(query);
        query = NULL;
        return ERROR;
    }

    /* Close the query */
    ias_db_query_close(query);
    query = NULL;

    return SUCCESS;
}
