/******************************************************************
Name: insert_radiometric_row

Purpose: Insert a record into the radiometric work order common table using 
         information from the L0R metadata and software version.

Returns: SUCCESS or ERROR
*******************************************************************/

#include <stdlib.h>
#include <string.h>
#include "ias_logging.h"
#include "setup_work_order.h"
#include "ias_db.h"
#include "ias_miscellaneous.h"

int insert_radiometric_row
(
    const DSW_DATA *dsw_data,          /* I: Parameters and values retrieved 
                                             from parameter file and L0R */
    struct ias_db_connection *db       /* I: Database connection */
)
{
    int status;                        /* Return status */
    char sql[1024];                    /* Buffer for sql commands */
    char buffer[IAS_SQL_RTN_MSGLEN];   /* Buffer for sql return messages */
    struct ias_db_query *query = NULL; /* DB query structure */
    char date_acquired[DATE_LENGTH];   /* String for format-converted date */
    char path_str[10];                 /* path used to build sql statemet */
    char row_str[10];                  /* row used to build sql statement */
    char char_id[IAS_CHAR_ID_SIZE];   /* Key value in
                                        RADIOMETRIC_WORK_ORDER_COMMON table
                                        (needed to associate processing
                                        with the correct work order)*/
    char *processing_system = NULL;    /* Processing system environment */
    char sw_version_lpgs[IAS_SOFTWARE_VERSION_SIZE + 2];
                                       /* LPGS software version plus quotes */
    char sw_version_ias[IAS_SOFTWARE_VERSION_SIZE + 2];
                                       /* IAS software version plus quotes */
    const char *oli_bpf_name = NULL;
    const char *tirs_bpf_name = NULL;
    const char *rlut_name = NULL;

    /* Convert the date to an Oracle time string */
    snprintf(date_acquired, sizeof(date_acquired),
        "%04d-%02d-%02d %02d:%02d:%02d",
        dsw_data->acquisition_date.year, 
        dsw_data->acquisition_date.month, 
        dsw_data->acquisition_date.day_of_month,
        dsw_data->acquisition_date.hour, 
        dsw_data->acquisition_date.minute, 
        (int)dsw_data->acquisition_date.second);

    /* Only Earth facing scenes will have a path/row */
    if (dsw_data->acquisition_type == IAS_EARTH)
    {
        sprintf(path_str, "'%d'", dsw_data->path);
        sprintf(row_str, "'%d'", dsw_data->row);
    }
    else
    {
        sprintf(path_str, "NULL");
        sprintf(row_str, "NULL");
    }

    status = ias_misc_get_char_id(db, "RADIOMETRIC_WORK_ORDER_COMMON",
                "RADIOMETRIC_WO_COMMON_ID", dsw_data->work_order_id,
                char_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to get the Radiometric Characterization ID");
        return ERROR;
    }

    /* Load the software version based on the PROCESSING_SYSTEM env */
    processing_system = getenv("PROCESSING_SYSTEM");
    if (processing_system == NULL)
    {
        IAS_LOG_ERROR("PROCESSING_SYSTEM environment variable is not "
            "defined to either IAS or LPGS");
        return ERROR;
    }
    else if (strncmp(processing_system, "IAS", 3) == 0)
    {
        sprintf(sw_version_ias, "'%s'", dsw_data->sw_version);
        sprintf(sw_version_lpgs, "NULL");
    }
    else if (strncmp(processing_system, "LPGS", 4) == 0)
    {
        sprintf(sw_version_lpgs, "'%s'", dsw_data->sw_version);
        sprintf(sw_version_ias, "NULL");
    }
    else
    {
        IAS_LOG_ERROR("PROCESSING_SYSTEM of '%s' is not the expected "
            "IAS or LPGS value", processing_system);
        return ERROR;
    }

    /* The BPF, CPF, and RLUT fields in the structure may contain full
       paths, we want just the filename. So use strrchr() to find the
       final slash, if we have one */
    oli_bpf_name = strrchr(dsw_data->bpf_filename_oli, '/');
    if (oli_bpf_name)
    {
        /* Move pointer past the slash */
        oli_bpf_name++; 
    }
    else
    {
        /* Name only, just use the pointer from the structure */
        oli_bpf_name = dsw_data->bpf_filename_oli;  
    }

    tirs_bpf_name = strrchr(dsw_data->bpf_filename_tirs, '/');
    if (tirs_bpf_name)
        tirs_bpf_name++; 
    else
        tirs_bpf_name = dsw_data->bpf_filename_tirs;

    rlut_name = strrchr(dsw_data->rlut_filename, '/');
    if (rlut_name)
        rlut_name++; 
    else
        rlut_name = dsw_data->rlut_filename;  

    /* Build the SQL statement */
    sprintf(sql, "UPDATE RADIOMETRIC_WORK_ORDER_COMMON "
                 "SET "
                 " DATE_ACQUIRED = to_date('%s', 'YYYY-MM-DD HH24:MI:SS'), "
                 " SENSOR_ID = '%s', "
                 " SOFTWARE_VERSION_IS = '%s', "
                 " SOFTWARE_VERSION_LPGS = %s, "
                 " SOFTWARE_VERSION_IAS = %s, "
                 " COLLECTION_TYPE = '%s', "
                 " LANDSAT_SCENE_ID = '%s', "
                 " INTEGRATION_TIME_OLI_PAN = '%f', "
                 " INTEGRATION_TIME_OLI_MS = '%f', "
                 " OLI_DETECTOR_SELECT_TABLE = '%d', "
                 " OLI_DETECTOR_SELECT_TABLE_ID = '%d', "
                 " TIRS_DETECTOR_MAP_ID = '%u', "
                 " LANDSAT_INTERVAL_ID = '%s', "
                 " DATE_PROCESSED = SYSDATE, "
                 " PROCESSING_CENTER = '%s', "
                 " TARGET_WRS_PATH = %s, "         /* path already quoted */
                 " TARGET_WRS_ROW = %s, "          /* row already quoted */
                 " L0RP_IMAGE_FILE = '%s', "
                 " CPF_NAME = '%s', "
                 " INTEGRATION_TIME_TIRS = '%f', "
                 " IMAGE_START_TIME = '%.6f', "
                 " IMAGE_STOP_TIME = '%.6f', "
                 " IMAGE_DATA_TRUNCATION_SETTING = '%d', "
                 " STATION_ID = '%s', "
                 " COLLECTION_NUMBER = %d, "
                 " OLI_BPF_NAME = '%s', "
                 " TIRS_BPF_NAME = '%s', "
                 " RLUT_NAME = '%s' "
                 " WHERE RADIOMETRIC_WO_COMMON_ID = '%s'",
                 date_acquired,
                 dsw_data->sensor_id_from_l0r,
                 dsw_data->is_version,
                 sw_version_lpgs,
                 sw_version_ias,
                 dsw_data->collection_type,
                 dsw_data->landsat_scene_id,
                 dsw_data->integration_time_oli_pan,
                 dsw_data->integration_time_oli_ms,
                 dsw_data->oli_detector_select_table,
                 dsw_data->oli_detector_select_table_id,
                 dsw_data->detector_map_id_tirs,
                 dsw_data->landsat_interval_id,
                 dsw_data->processing_center,
                 path_str,
                 row_str,
                 dsw_data->l0r_id,   /* use filename without path */
                 dsw_data->cpf_filename,
                 dsw_data->integration_time_tirs,
                 dsw_data->image_start_time_sec_since_j2000,
                 dsw_data->image_stop_time_sec_since_j2000,
                 dsw_data->image_data_truncation_setting,
                 dsw_data->station_id,
                 dsw_data->collection_number,
                 oli_bpf_name,
                 tirs_bpf_name,
                 rlut_name,
                 char_id
           );

    /* Perform the query */
    query = ias_db_perform_query(db, sql);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, buffer, sizeof(buffer));
        IAS_LOG_ERROR("Inserting: %s", buffer);
        ias_db_query_close(query);
        query = NULL;
        return ERROR;
    }

    /* Close the query */
    ias_db_query_close(query);
    query = NULL;

    return SUCCESS;
}
