/******************************************************************
Name: insert_scenes_table_row

Purpose: Insert a record into the SCENES table using 
         information from the L0R metadata.

Returns: SUCCESS or ERROR
*******************************************************************/
#include <stdio.h>
#include "ias_logging.h"
#include "setup_work_order.h"
#include "ias_db.h"

int insert_scenes_table_row
(
    const DSW_DATA *dsw_data,          /* I: Parameters and values retrieved 
                                             from parameter file and L0R */
    struct ias_db_connection *db       /* I: Database connection */
)
{
    int status;                        /* Return status */
    char sql[1024];                    /* Buffer for sql commands */
    char buffer[IAS_SQL_RTN_MSGLEN];   /* Buffer for sql return messages */
    struct ias_db_query *query = NULL; /* DB query structure */
    char date_acquired[DATE_LENGTH];   /* String for format-converted date */
    char path_str[10];                 /* Path used to build sql statemet */
    char row_str[10];                  /* Row used to build sql statement */

    /* Convert the date to an Oracle time string */
    status = snprintf(date_acquired, sizeof(date_acquired),
        "%04d-%02d-%02d %02d:%02d:%02d",
        dsw_data->acquisition_date.year, 
        dsw_data->acquisition_date.month, 
        dsw_data->acquisition_date.day_of_month,
        dsw_data->acquisition_date.hour, 
        dsw_data->acquisition_date.minute, 
        (int)dsw_data->acquisition_date.second);
    if (status < 0 || status >= sizeof(date_acquired))
    {
        IAS_LOG_ERROR("Creating the date acquired string");
        return ERROR;
    }

    /* Only Earth facing scenes will have a path/row */
    if (dsw_data->acquisition_type == IAS_EARTH)
    {
        sprintf(path_str, "'%d'", dsw_data->path);
        sprintf(row_str, "'%d'", dsw_data->row);
    }
    else
    {
        sprintf(path_str, "NULL");
        sprintf(row_str, "NULL");
    }

    /* Delete any existing rows for the work order id */
    status = snprintf(sql, sizeof(sql), "Delete from scenes "
        "where work_order_id = '%s'", dsw_data->work_order_id);
    if (status < 0 || status >= sizeof(sql))
    {
        IAS_LOG_ERROR("Creating the SQL delete statement");
        return ERROR;
    }

    /* Perform the delete of existing records*/
    query = ias_db_perform_query(db, sql);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, buffer, sizeof(buffer));
        IAS_LOG_ERROR("Deleting: %s", buffer);
        ias_db_query_close(query);
        query = NULL;
        return ERROR;
    }
    ias_db_query_close(query);
    query = NULL;

    /* Create the insert statement */
    status = snprintf(sql, sizeof(sql), "insert into scenes "
        "(work_order_id, date_acquired, target_wrs_path, target_wrs_row) "
        "values ('%s', to_date('%s', 'YYYY-MM-DD HH24:MI:SS'), " 
        "%s, %s) ", dsw_data->work_order_id, date_acquired, path_str, row_str);
    if (status < 0 || status >= sizeof(sql))
    {
        IAS_LOG_ERROR("Creating the SQL insert statement");
        return ERROR;
    }

    /* Perform the insert */
    query = ias_db_perform_query(db, sql);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, buffer, sizeof(buffer));
        IAS_LOG_ERROR("Inserting: %s", buffer);
        ias_db_query_close(query);
        query = NULL;
        return ERROR;
    }

    /* Close the query */
    ias_db_query_close(query);
    query = NULL;

    return SUCCESS;
}
