/**************************************************************************
 NAME:                      move_corners

 PURPOSE:       Calculate the new OLI or TIRS scene corners.

 RETURNS:       Status code of SUCCESS or ERROR
***************************************************************************/
#include <math.h>
#include "ias_logging.h"
#include "ias_los_model.h"
#include "ias_geo.h"

int move_corners
(
    const IAS_L0R_CORNERS *old_corners, /* I: Original scene corners */
    double roll_angle,               /* I: Scene off-nadir angle */
    const IAS_DBL_LAT_LONG *scene_center, /* I: Scene center coordinates */
    const IAS_EARTH_CHARACTERISTICS *earth, /* I: Earth parameters */
    double orbit_radius,             /* I: Nominal orbit radius */
    double fov_at,                   /* I: Along-track field of view */
    double fov_xt,                   /* I: Across-track field of view */
    IAS_L0R_CORNERS *new_corners     /* O: New scene corners */
)
{
    IAS_QUATERNION left_quaternion;  /* Corner rotation quaternion */
    IAS_QUATERNION right_quaternion; /* Corner rotation quaternion */
    IAS_QUATERNION left_quaternion_conjugated;  /* Conjugated quaternion */
    IAS_QUATERNION right_quaternion_conjugated; /* Conjugated quaternion */
    IAS_VECTOR x_ul;                 /* Upper left corner point in ECEF */
    IAS_VECTOR x_ur;                 /* Upper right corner point in ECEF */
    IAS_VECTOR x_ll;                 /* Lower left corner point in ECEF */
    IAS_VECTOR x_lr;                 /* Lower right corner point in ECEF */
    IAS_VECTOR right_rotation_axis;  /* Axis of rotation for right side */
    IAS_VECTOR left_rotation_axis;   /* Axis of rotation for left side */
    double flattening;               /* Earth flattening */
    double earth_radius;             /* Earth radius at scene center */
    double magnitude;                /* Vector magnitude */
    double left_distance;            /* Distance to the left edge of scene */
    double right_distance;           /* Distances to right edge of scene */
    double left_beta_angle;          /* Angle to rotate left edge of scene */
    double right_beta_angle;         /* Angle to rotate right edge of scene */
    double rotation_matrix[3][3];    /* Rotation matrix to transform vectors */
    double geodetic_latitude;        /* Geodetic position latitude */
    double geodetic_longitude;       /* Geodetic position longitude */
    double geocentric_latitude;      /* Geocentric latitude */
    double degrees_per_radian;       /* Degrees to radians conversion */
    double radians_per_degree;       /* Radians to degree conversion */

    /* Calculate constants */
    flattening = (earth->semi_major_axis - earth->semi_minor_axis)
        / earth->semi_major_axis;
    degrees_per_radian = ias_math_get_degrees_per_radian(); 
    radians_per_degree = ias_math_get_radians_per_degree(); 

    /* Convert scene center to ECEF to get local Earth radius */
    ias_geo_convert_geod2cart(scene_center->lat * radians_per_degree,
                              scene_center->lng * radians_per_degree,
                              0.0, earth->semi_major_axis, flattening, &x_ul);

    earth_radius = ias_math_compute_vector_length(&x_ul);

    /* Convert corners to ECEF unit vectors */
    flattening = earth->semi_minor_axis/earth->semi_major_axis
               * earth->semi_minor_axis/earth->semi_major_axis;

    /* Upper Left */
    geocentric_latitude = atan(tan(old_corners->upleft.lat * radians_per_degree)
        * flattening);
    geodetic_longitude = old_corners->upleft.lng * radians_per_degree;
    x_ul.x = cos(geodetic_longitude) * cos(geocentric_latitude);
    x_ul.y = sin(geodetic_longitude) * cos(geocentric_latitude);
    x_ul.z = sin(geocentric_latitude);

    /* Upper Right */
    geocentric_latitude = atan(tan(old_corners->upright.lat 
        * radians_per_degree) * flattening);
    geodetic_longitude = old_corners->upright.lng * radians_per_degree;
    x_ur.x = cos(geodetic_longitude) * cos(geocentric_latitude);
    x_ur.y = sin(geodetic_longitude) * cos(geocentric_latitude);
    x_ur.z = sin(geocentric_latitude);

    /* Lower Left */
    geocentric_latitude = atan(tan(old_corners->loleft.lat * radians_per_degree)
        * flattening);
    geodetic_longitude = old_corners->loleft.lng * radians_per_degree;
    x_ll.x = cos(geodetic_longitude) * cos(geocentric_latitude);
    x_ll.y = sin(geodetic_longitude) * cos(geocentric_latitude);
    x_ll.z = sin(geocentric_latitude);

    /* Lower Right */
    geocentric_latitude = atan(tan(old_corners->loright.lat 
        * radians_per_degree) * flattening);
    geodetic_longitude = old_corners->loright.lng * radians_per_degree;
    x_lr.x = cos(geodetic_longitude) * cos(geocentric_latitude);
    x_lr.y = sin(geodetic_longitude) * cos(geocentric_latitude);
    x_lr.z = sin(geocentric_latitude);

    /* Calculate axes of rotation for left and right sides of scene */
    ias_math_compute_3dvec_cross(&x_ul, &x_ll, &left_rotation_axis);
    magnitude = ias_math_compute_vector_length(&left_rotation_axis);

    if (magnitude > 0.0)
    {
        left_rotation_axis.x /= magnitude;
        left_rotation_axis.y /= magnitude;
        left_rotation_axis.z /= magnitude;
    }
    else
    {
        IAS_LOG_ERROR("Computing axis of rotation for left side of scene");
        return(ERROR);
    }
    ias_math_compute_3dvec_cross(&x_ur, &x_lr, &right_rotation_axis);
    magnitude = ias_math_compute_vector_length(&right_rotation_axis);
    if (magnitude > 0.0)
    {
        right_rotation_axis.x /= magnitude;
        right_rotation_axis.y /= magnitude;
        right_rotation_axis.z /= magnitude;
    }
    else
    {
        IAS_LOG_ERROR("Computing axis of rotation for right side of scene");
        return ERROR;
    }

    /* Calculate ranges to edges of scene */
    magnitude = orbit_radius * sin(roll_angle * radians_per_degree - fov_xt);
    left_distance = orbit_radius * cos(roll_angle * radians_per_degree - fov_xt)
        - sqrt(earth_radius * earth_radius - magnitude * magnitude);
    magnitude = orbit_radius * sin(roll_angle * radians_per_degree + fov_xt);
    right_distance = orbit_radius * cos(roll_angle * radians_per_degree 
        + fov_xt) - sqrt(earth_radius * earth_radius - magnitude * magnitude);

    /* Calculate angular extensions required on left and right edges */
    left_beta_angle = asin(left_distance * sin(fov_at) / earth_radius);
    right_beta_angle = asin(right_distance * sin(fov_at) / earth_radius);

    /* Construct rotation quaternions for left and right edges */
    magnitude = sin(left_beta_angle / 2.0);
    left_quaternion.vector.x = left_rotation_axis.x * magnitude;
    left_quaternion.vector.y = left_rotation_axis.y * magnitude;
    left_quaternion.vector.z = left_rotation_axis.z * magnitude;
    left_quaternion.scalar = cos(left_beta_angle / 2.0);
    magnitude = sin(right_beta_angle / 2.0);
    right_quaternion.vector.x = right_rotation_axis.x * magnitude;
    right_quaternion.vector.y = right_rotation_axis.y * magnitude;
    right_quaternion.vector.z = right_rotation_axis.z * magnitude;
    right_quaternion.scalar = cos(right_beta_angle / 2.0);

    /* Rotate the corner vectors */

    /* Positive rotation of LL corner */
    ias_math_conjugate_quaternion(&left_quaternion,
        &left_quaternion_conjugated);
    ias_math_convert_quaternion_to_euler(&left_quaternion_conjugated, 
        rotation_matrix);
    ias_math_transform_3dvec(&x_ll, rotation_matrix, &left_rotation_axis);
    x_ll.x = left_rotation_axis.x;
    x_ll.y = left_rotation_axis.y;
    x_ll.z = left_rotation_axis.z;

    /* Negative rotation of UL corner */
    ias_math_convert_quaternion_to_euler(&left_quaternion, rotation_matrix);
    ias_math_transform_3dvec(&x_ul, rotation_matrix, &left_rotation_axis);
    x_ul.x = left_rotation_axis.x;
    x_ul.y = left_rotation_axis.y;
    x_ul.z = left_rotation_axis.z;

    /* Positive rotation of LR corner */
    ias_math_conjugate_quaternion(&right_quaternion,
        &right_quaternion_conjugated);
    ias_math_convert_quaternion_to_euler(&right_quaternion_conjugated, 
        rotation_matrix);
    ias_math_transform_3dvec(&x_lr, rotation_matrix, &right_rotation_axis);
    x_lr.x = right_rotation_axis.x;
    x_lr.y = right_rotation_axis.y;
    x_lr.z = right_rotation_axis.z;

    /* Negative rotation of UR corner */
    ias_math_convert_quaternion_to_euler(&right_quaternion, rotation_matrix);
    ias_math_transform_3dvec(&x_ur, rotation_matrix, &right_rotation_axis);
    x_ur.x = right_rotation_axis.x;
    x_ur.y = right_rotation_axis.y;
    x_ur.z = right_rotation_axis.z;

    /* Convert updated position vectors to geodetic */

    /* Upper Left */
    if (ias_geo_convert_cart2sph(&x_ul, &geocentric_latitude, 
        &geodetic_longitude, &earth_radius) == ERROR)
    {
        IAS_LOG_ERROR("Converting updated upper left corner to geodetic");
        return ERROR;
    }
    geodetic_latitude = atan(tan(geocentric_latitude) / flattening);
    new_corners->upleft.lat = geodetic_latitude * degrees_per_radian;
    new_corners->upleft.lng = geodetic_longitude * degrees_per_radian;

    /* Upper Right */
    if (ias_geo_convert_cart2sph(&x_ur, &geocentric_latitude, 
        &geodetic_longitude, &earth_radius) == ERROR)
    {
        IAS_LOG_ERROR("Converting updated upper right corner to geodetic");
        return ERROR;
    }
    geodetic_latitude = atan(tan(geocentric_latitude) / flattening);
    new_corners->upright.lat = geodetic_latitude * degrees_per_radian;
    new_corners->upright.lng = geodetic_longitude * degrees_per_radian;

    /* Lower Left */
    if (ias_geo_convert_cart2sph(&x_ll, &geocentric_latitude, 
        &geodetic_longitude, &earth_radius) == ERROR)
    {
        IAS_LOG_ERROR("Converting updated lower left corner to geodetic");
        return ERROR;
    }
    geodetic_latitude = atan(tan(geocentric_latitude) / flattening);
    new_corners->loleft.lat = geodetic_latitude * degrees_per_radian;
    new_corners->loleft.lng = geodetic_longitude * degrees_per_radian;

    /* Lower Right */
    if (ias_geo_convert_cart2sph(&x_lr, &geocentric_latitude, 
        &geodetic_longitude, &earth_radius) == ERROR)
    {
        IAS_LOG_ERROR("Converting updated lower right corner to geodetic");
        return ERROR;
    }
    geodetic_latitude = atan(tan(geocentric_latitude) / flattening);
    new_corners->loright.lat = geodetic_latitude * degrees_per_radian;
    new_corners->loright.lng = geodetic_longitude * degrees_per_radian;

    return SUCCESS;
}
