/******************************************************************
Name: read_l0r_metadata

Purpose: Acquire the L0R Product location from the parameter file
         Create link to the L0R Product directory in the work order directory
         Read the L0R metadata

Returns: SUCCESS or ERROR
*******************************************************************/

#include <string.h>
#include <stdlib.h>
#include <unistd.h>                    /* For symlink */
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "setup_work_order.h"

#define L0R_LINK_NAME "l0r_symlink"

int read_l0r_metadata
(
    DSW_DATA *dsw_data                 /* I/O: Parameters and values retrieved 
                                             from parameter file and L0R */
)
{
    int status;                     /* Status returned from function call */
    char l0r_product_dir[PATH_MAX]; /* Directory path for the L0R */
    char *ptr;                      /* For string manipulations */
    char l0r_id[PATH_MAX];          /* L0R image ID */
    L0RIO *l0r_ptr;                 /* L0R file access pointer */
    int record_count;               /* Number of records present */
    int oli_time_found = 0;         /* OLI start/stop times found */
    int scene_metadata_available = 0; /* Flag that indicates the scene metadata 
                                         is available */
    int index;                      /* Generic loop index */
    int have_scene_id;         /* Flag to indicate if the scene id is present */
    uint32_t oli_start_frame = 0;
    uint32_t oli_frame_count = 0;
    uint32_t tirs_start_frame = 0;
    uint32_t tirs_frame_count = 0;
    int tirs_start_index = 0;
    int oli_start_index = 0;

    /* L0R header and metadata */
    IAS_L0R_INTERVAL_METADATA interval_metadata;
    IAS_L0R_OLI_IMAGE_HEADER oli_image_header;
    IAS_L0R_SCENE_METADATA *scene_metadata = NULL;
    IAS_L0R_SCENE_METADATA *cur_scene = NULL;

    /* Make sure there's an L0R file to open */
    if (strcmp(dsw_data->l0r_filename, "") == 0)
    {
        IAS_LOG_ERROR("L0R filename empty");
        return ERROR;
    }

    memset(&interval_metadata, 0, sizeof(interval_metadata));
    memset(&oli_image_header, 0, sizeof(oli_image_header));

    /* Split the L0R base file name into the directory path and ID
       components */
    strcpy (l0r_product_dir, dsw_data->l0r_filename);
    ptr = strrchr(l0r_product_dir, '/');
    if (ptr == NULL) /* Not found */
    {
        IAS_LOG_ERROR("Slash not found in L0R product directory");
        return ERROR;
    }
    strncpy(l0r_id, ++ptr, PATH_MAX);
    *ptr = '\0';

    /* Save the new L0R filename */
    sprintf(dsw_data->l0r_filename, "%s/%s", L0R_LINK_NAME, l0r_id);

    /* Save the L0R Id */
    sprintf(dsw_data->l0r_id, "%s", l0r_id);

    /* Remove an existing link, the status of unlink has not been checked for
       errors since it will normally return an error since the link won't 
       exist. -- DSW might have failed part way through and is being re-run */
    unlink(L0R_LINK_NAME);

    /* Create link to the L0R product directory in the work order directory */
    status = symlink(l0r_product_dir, L0R_LINK_NAME);
    if (status != 0)
    {
        IAS_LOG_ERROR("Creating link to L0R product to %s", l0r_product_dir);
        return ERROR;
    }

    /* Open the L0R file */
    l0r_ptr = ias_l0r_open(l0r_id, L0R_LINK_NAME, 0);
    if (!l0r_ptr)
    {
        IAS_LOG_ERROR("Opening L0R file %s/%s", L0R_LINK_NAME, l0r_id);
        return ERROR;
    }

    IAS_LOG_INFO("Reading L0R metadata from %s/%s", L0R_LINK_NAME, l0r_id);

    /* Open read access to the L0R metadata */
    status = ias_l0r_open_metadata(l0r_ptr, IAS_READ);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Opening read access to L0R metadata");
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Open read access to the L0R header */
    status = ias_l0r_open_header(l0r_ptr, IAS_READ);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Opening read access to L0R header");
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Retrieve the L0R interval metadata */
    status = ias_l0r_get_interval_metadata(l0r_ptr, &interval_metadata);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading L0R interval metadata");
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Sensor ID  -- read from L0R, not to be confused with 
       geometry_sensor_id_override read from the parameter file.
       Set the sensor_present flags accordingly */

    /* Note L0R has an underscore for OLI_TIRS, but the DB table does not 
       allow that -- it must be OLI, TIRS, or OLITIRS. */
    if (strcmp(interval_metadata.sensor_id, "OLI_TIRS") == 0)
    {
        strncpy(dsw_data->sensor_id_from_l0r, "OLITIRS",
            sizeof(dsw_data->sensor_id_from_l0r));

        /* Set the sensor present flags.    For earth images, these settings
           should be considered tentative--the scene metadata will be used to
           make a final determination.*/
        dsw_data->sensor_present_oli = TRUE;
        dsw_data->sensor_present_tirs = TRUE;
    }
    else    /* OLI or TIRS only */
    {
        strncpy(dsw_data->sensor_id_from_l0r, interval_metadata.sensor_id,
            sizeof(dsw_data->sensor_id_from_l0r));

        /* Set the sensor present flags for either OLI or TIRS.  Again, for
           earth images these settings should be considered tentative until
           the scene metadata sensor_present flags are checked.  For
           calibration collects, these settings should be considered
           valid. */
        if (strcmp(interval_metadata.sensor_id, "OLI") == 0)
        {
            dsw_data->sensor_present_oli = TRUE;
            dsw_data->sensor_present_tirs = FALSE;
        }
        else if (strcmp(interval_metadata.sensor_id, "TIRS") == 0)
        {
            dsw_data->sensor_present_oli = FALSE;
            dsw_data->sensor_present_tirs = TRUE;
        }
        else
        {
            /* It should not get here as it's neither OLI nor TIRS */
            IAS_LOG_ERROR("Unable to recognize the sensor id %s from the "
                "interval metadata file", interval_metadata.sensor_id);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }
    }

    dsw_data->sensor_id_from_l0r[sizeof(dsw_data->sensor_id_from_l0r) - 1] 
        = '\0';

    /* Retrieve the L0R scene record count */
    status = ias_l0r_get_scene_metadata_records_count(l0r_ptr, &record_count);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading L0R scene count");
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Check if the scene id is present */
    have_scene_id = (strlen(dsw_data->landsat_scene_id) > 0);

    /* Only read the scene metadata if available */
    if (record_count > 0)
    {
        /* Allocate memory for the scene metadata */
        scene_metadata = malloc(record_count * sizeof(*scene_metadata));
        if (!scene_metadata)
        {
            IAS_LOG_ERROR("Allocating memory for %d records of scene "
                "metadata", record_count);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }

        /* Read the metadata */
        if (ias_l0r_get_scene_metadata(l0r_ptr, 0, record_count,
            scene_metadata) != SUCCESS)
        {
            IAS_LOG_ERROR("Retrieving %d records from the L0R scene metadata",
                record_count);
            ias_l0r_close(l0r_ptr);
            free(scene_metadata);
            return ERROR;
        }
    }

    /* If more than 1 scene is present find the correct start and
       end times */
    if (record_count > 1)
    {
        /* Determine the TIRS and OLI index starts */
        for (oli_start_index = 0; oli_start_index < record_count;
             oli_start_index++)
        {
            if (scene_metadata[oli_start_index].sensor_present_oli)
                break;
        }
        for (tirs_start_index = 0; tirs_start_index < record_count;
             tirs_start_index++)
        {
            if (scene_metadata[tirs_start_index].sensor_present_tirs)
                break;
        }
    }

    /* Loop through all of the scene metadata records for the correct scene */
    for (index = 0; index < record_count; index++)
    {
        cur_scene = &scene_metadata[index];

        /* Check if the landsat scene id is present else use the first scene */
        if (!have_scene_id || strcmp(dsw_data->landsat_scene_id,
            cur_scene->landsat_scene_id) == 0)
        {
            if (!have_scene_id)
            {
                /* Store the current Landsat scene ID */
                strncpy(dsw_data->landsat_scene_id, cur_scene->landsat_scene_id,
                    sizeof(dsw_data->landsat_scene_id));
                dsw_data->landsat_scene_id[sizeof(
                    dsw_data->landsat_scene_id) - 1] = '\0';
            }

            /* Set scene_metadata_available flag to 1 after retrieving L0R
               scene metadata */
            scene_metadata_available = 1;
            break;
        }
    }

    /* Check that the scene metadata was found when the scene ID is specified */
    if (record_count > 0 && !scene_metadata_available)
    {
        if (have_scene_id)
        {
            IAS_LOG_ERROR("Scene metadata for scene ID %s not found", 
                dsw_data->landsat_scene_id);
        }
        else
        {
            IAS_LOG_ERROR("No scene metadata was found");
        }
        ias_l0r_close(l0r_ptr);
        free(scene_metadata);
        return ERROR;
    }

    /* Copy the scene metadata if it is available */
    if (scene_metadata_available)
    {
        /* Use the scene metadata 'sensor_present' flags to "officially"
           establish which sensor(s) imaged for this collect.  If there's
           a partial scene acquired from one sensor, this will override the
           settings established when 'OLI_TIRS' is the sensor ID flag in the
           interval metadata. */
        if (cur_scene->sensor_present_oli)
        {
            oli_start_frame = cur_scene->scene_start_frame_oli;
            oli_frame_count = cur_scene->scene_stop_frame_oli -
                oli_start_frame + 1;
            oli_start_frame -= scene_metadata[
                oli_start_index].scene_start_frame_oli;

            dsw_data->sensor_present_oli = TRUE;
        }
        else
        {
            dsw_data->sensor_present_oli = FALSE;
        }

        if (cur_scene->sensor_present_tirs)
        {
            tirs_start_frame = cur_scene->scene_start_frame_tirs;
            tirs_frame_count = cur_scene->scene_stop_frame_tirs -
                tirs_start_frame + 1;
            tirs_start_frame -= scene_metadata[
                tirs_start_index].scene_start_frame_tirs;

            dsw_data->sensor_present_tirs = TRUE;
        }
        else
        {
            dsw_data->sensor_present_tirs = FALSE;
        }

        /* Copy the tirs and oli corners for the current scene */
        dsw_data->corners_oli = cur_scene->corners_oli;
        dsw_data->corners_tirs = cur_scene->corners_tirs;

        /* Off nadir angle (roll angle) */
        dsw_data->off_nadir_angle = cur_scene->roll_angle;
        
        /* Path and row */
        dsw_data->path = cur_scene->target_wrs_path;
        dsw_data->row = cur_scene->target_wrs_row;

        /* Latitude of scene center */
        dsw_data->scene_center.lat = cur_scene->scene_center.lat;
        dsw_data->scene_center.lng = cur_scene->scene_center.lng;

        /* Get the date acquired from the scene metadata */
        status = ias_misc_parse_datetime_string(cur_scene->date_acquired,
            IAS_DATETIME_L0R_FORMAT, &dsw_data->acquisition_date);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Parsing scene metadata acquisition date string %s",
                cur_scene->date_acquired);
            free(scene_metadata);
            return ERROR;
        }
        strncpy(dsw_data->acquisition_date_string, cur_scene->date_acquired,
            sizeof(dsw_data->acquisition_date_string));
        dsw_data->acquisition_date_string[
            sizeof(dsw_data->acquisition_date_string) - 1] = '\0';
    }
    else
    {
        /* Fall back to the interval metadata date_acquired. */
        status = ias_misc_parse_datetime_string(
            interval_metadata.date_acquired,
            IAS_DATETIME_L0R_FORMAT, &dsw_data->acquisition_date);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Parsing interval metadata acquisition date string"
                " %s", interval_metadata.date_acquired);
            free(scene_metadata);
            return ERROR;
        }
        strncpy(dsw_data->acquisition_date_string,
            interval_metadata.date_acquired,
            sizeof(dsw_data->acquisition_date_string));
        dsw_data->acquisition_date_string[
            sizeof(dsw_data->acquisition_date_string) - 1] = '\0';
    }

    /* Free the scene metadata */
    free(scene_metadata);
    scene_metadata = NULL;
    cur_scene = NULL;

    /* record count */
    dsw_data->record_count = record_count;

    /* Initialize the frame header related information in case it isn't in
       the file */
    dsw_data->integration_time_oli_ms = 0;
    dsw_data->integration_time_oli_pan = 0;
    dsw_data->oli_detector_select_table = 0;
    dsw_data->oli_detector_select_table_id = 0;
    dsw_data->integration_time_tirs = 0;
    dsw_data->image_start_time_sec_since_j2000 = 0;
    dsw_data->image_stop_time_sec_since_j2000 = 0;

    /* Get the count of OLI frame header records */
    status = ias_l0r_get_oli_frame_headers_records_count(l0r_ptr, 
                                                         &record_count);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting L0R OLI image header count");
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* If there are any OLI frame header records, get the info needed */
    if (record_count > 0)
    {
        int index;
        double oli_start_time;
        double oli_stop_time;
        IAS_L0R_OLI_FRAME_HEADER *frame_headers;
        int start_frame = 0;
        int end_frame = record_count;

        /* Retrieve the L0R OLI image header */
        status = ias_l0r_get_oli_image_header(l0r_ptr, &oli_image_header);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading L0R OLI image header");
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }

        /* Integration time is in units of microseconds in the L0R image
           header; convert to milliseconds for the database */
        dsw_data->integration_time_oli_ms
                = oli_image_header.ms_integration_time * 1.0e-3;
        dsw_data->integration_time_oli_pan
                = oli_image_header.pan_integration_time * 1.0e-3;

        /* OLI detector slect table and its ID */
        dsw_data->oli_detector_select_table 
                = (int) oli_image_header.current_detector_select_table;
        dsw_data->oli_detector_select_table_id 
                = (int) oli_image_header.detector_select_table_id_number;

        /* Image data truncation setting */
        dsw_data->image_data_truncation_setting 
                = oli_image_header.image_data_truncation_setting;

        /* Fill flag for the Image Header */
        if(IAS_L0R_FILLED & oli_image_header.frame_header.frame_status)
            dsw_data->image_header_filled = 1;
        else
            dsw_data->image_header_filled = 0;

        /* Allocate space for the frame headers */
        frame_headers = malloc(sizeof(*frame_headers) * record_count);
        if (!frame_headers)
        {
            IAS_LOG_ERROR("Allocating memory for OLI frame headers");
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }

        /* Get the OLI frame headers so the time stamps can be used */
        status = ias_l0r_get_oli_frame_headers(l0r_ptr, 0, record_count,
                    frame_headers);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading OLI frame headers");
            free(frame_headers);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }

        /* Set the start and end frames if scene id is present */
        if (scene_metadata_available)
        {
            start_frame = oli_start_frame;
            end_frame = oli_start_frame + oli_frame_count;
        }

        /* Find the first valid time in the frame headers */
        for (index = start_frame; index < end_frame; index++)
        {
            if (frame_headers[index].frame_status & IAS_L0R_VERIFIED)
            {
                oli_start_time = IAS_L0R_CONVERT_TIME_TO_SECONDS_SINCE_J2000(
                    frame_headers[index].l0r_time);
                break;
            }
        }
        if (index >= end_frame)
        {
            IAS_LOG_ERROR("Unable to find valid OLI start time");
            free(frame_headers);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }

        /* Find the last valid time in the frame headers */
        for (index = end_frame - 1; index >= start_frame; index--)
        {
            if (frame_headers[index].frame_status & IAS_L0R_VERIFIED)
            {
                oli_stop_time = IAS_L0R_CONVERT_TIME_TO_SECONDS_SINCE_J2000(
                    frame_headers[index].l0r_time);
                break;
            }
        }
        if (index < start_frame)
        {
            IAS_LOG_ERROR("Unable to find valid OLI stop time");
            free(frame_headers);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }
        free(frame_headers);
        frame_headers = NULL;

        /* Save the start and stop times */
        dsw_data->image_start_time_sec_since_j2000 = oli_start_time;
        dsw_data->image_stop_time_sec_since_j2000 = oli_stop_time;
        oli_time_found = 1;
    }

    /* Get the count of TIRS frame header records */
    status = ias_l0r_get_tirs_frame_headers_records_count(l0r_ptr,
                                                          &record_count);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting L0R TIRS frame header count");
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* If there are any TIRS frame header records, get the info needed */
    if (record_count > 0)
    {
        int index;
        double tirs_start_time;
        double tirs_stop_time;
        IAS_L0R_TIRS_FRAME_HEADER *frame_headers;
        int start_frame = 0;
        int end_frame = record_count;

        /* Allocate space for the frame headers */
        frame_headers = malloc(sizeof(*frame_headers) * record_count);
        if (!frame_headers)
        {
            IAS_LOG_ERROR("Allocating memory for TIRS frame headers");
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }

        /* Retrieve the L0R TIRS frame headers */
        status = ias_l0r_get_tirs_frame_headers(l0r_ptr, 0, record_count, 
                                                frame_headers);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading L0R TIRS frame headers");
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }

        /* Integration time is in units of microseconds in the L0R frame
           header; convert to milliseconds for the database */
        dsw_data->integration_time_tirs
                = frame_headers[0].integration_duration * 1.0e-3;

        /* Set the start and end frames if scene id is present */
        if (scene_metadata_available)
        {
            start_frame = tirs_start_frame;
            end_frame = tirs_start_frame + tirs_frame_count;
        }

        /* Find the first valid time in the frame headers */
        for (index = start_frame; index < end_frame; index++)
        {
            if (frame_headers[index].frame_status & IAS_L0R_VERIFIED)
            {
                tirs_start_time = IAS_L0R_CONVERT_TIME_TO_SECONDS_SINCE_J2000(
                    frame_headers[index].l0r_time);
                break;
            }
        }
        if (index >= end_frame)
        {
            IAS_LOG_ERROR("Unable to find valid TIRS start time");
            free(frame_headers);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }

        /* Find the last valid time in the frame headers */
        for (index = end_frame - 1; index >= start_frame; index--)
        {
            if (frame_headers[index].frame_status & IAS_L0R_VERIFIED)
            {
                tirs_stop_time = IAS_L0R_CONVERT_TIME_TO_SECONDS_SINCE_J2000(
                    frame_headers[index].l0r_time);
                break;
            }
        }
        if (index < start_frame)
        {
            IAS_LOG_ERROR("Unable to find valid TIRS stop time");
            free(frame_headers);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }
        free(frame_headers);
        frame_headers = NULL;

        /* Get the start and stop times */
        if (oli_time_found)
        {
            if (tirs_start_time < dsw_data->image_start_time_sec_since_j2000)
                dsw_data->image_start_time_sec_since_j2000 = tirs_start_time;

            if (tirs_stop_time > dsw_data->image_stop_time_sec_since_j2000)
                dsw_data->image_stop_time_sec_since_j2000 = tirs_stop_time;
        }
        else
        {
            dsw_data->image_start_time_sec_since_j2000 = tirs_start_time;
            dsw_data->image_stop_time_sec_since_j2000 = tirs_stop_time;
        }
    }

    /* Close the L0R */
    status = ias_l0r_close(l0r_ptr);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Closing L0R");
        return ERROR;
    }

    /* Collection type */
    strncpy(dsw_data->collection_type, interval_metadata.collection_type,
        sizeof(dsw_data->collection_type));
    dsw_data->collection_type[sizeof(dsw_data->collection_type) - 1] = '\0';

    /* Satellite ID */
    dsw_data->satellite_id = interval_metadata.satellite;

    /* IS software version number */
    strncpy(dsw_data->is_version, interval_metadata.is_version,
        sizeof(dsw_data->is_version));
    dsw_data->is_version[sizeof(dsw_data->is_version) - 1] = '\0';

    /* Map acquistion type to collection type */
    if (strcmp(interval_metadata.collection_type, "EARTH_IMAGING") == 0)
    {
        dsw_data->acquisition_type = IAS_EARTH;
    }
    else if (strcmp(interval_metadata.collection_type, "LUNAR") == 0)
    {
        dsw_data->acquisition_type = IAS_LUNAR;
    }
    else if (strcmp(interval_metadata.collection_type, "STELLAR") == 0)
    {
        dsw_data->acquisition_type = IAS_STELLAR;
    }
    else
    {
        /* Not any of the collections GPS is interested in, so group all
           the others into "other" for RPS */
        dsw_data->acquisition_type = IAS_OTHER_ACQUISITION;
    }

    /* Spacecraft ID */
    strncpy(dsw_data->spacecraft_id, interval_metadata.spacecraft_id,
        sizeof(dsw_data->spacecraft_id));
    dsw_data->spacecraft_id[sizeof(dsw_data->spacecraft_id) - 1] = '\0';

    /* TIRS Detector Map ID */
    dsw_data->detector_map_id_tirs = interval_metadata.detector_map_id_tirs;

    /* Landsat interval ID */
    strncpy(dsw_data->landsat_interval_id, 
        interval_metadata.landsat_interval_id,
        sizeof(dsw_data->landsat_interval_id));
    dsw_data->landsat_interval_id[sizeof(dsw_data->landsat_interval_id) - 1] 
        = '\0';

    /* For calibration scenes use the landsat_cal_interval_id instead of
       landsat_interval_id. Check if the landsat_interval_id is empty. If it is
       this is a calibration scene. Can't use string length becuase
       landsat_interval_id could be spaces so check for 'L' which should be in 
       all landsat_interval_ids. */
    if (strchr(dsw_data->landsat_interval_id, 'L') == NULL)
    {
        strncpy(dsw_data->landsat_interval_id, 
                interval_metadata.landsat_cal_interval_id,
                sizeof(dsw_data->landsat_interval_id));
        dsw_data->landsat_interval_id[sizeof(dsw_data->landsat_interval_id)-1]
            = '\0';
    }

    /* Copy the ground station ID */
    strncpy(dsw_data->station_id, interval_metadata.station_id,
        IAS_L0R_STATION_ID_LENGTH);
    dsw_data->station_id[IAS_L0R_STATION_ID_LENGTH] = '\0';

    return SUCCESS;
}

