/******************************************************************
Name: read_parameters

Purpose: Read the application parameter file, storing values in the 
         DSW_DATA structure

Returns: SUCCESS or ERROR
*******************************************************************/

#include <string.h>
#include <stdlib.h>
#include "ias_logging.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "setup_work_order.h"


int read_parameters
(
    const char *parm_filename,    /* I: Parameter filename */
    DSW_DATA *dsw_data            /* O: Parameters and values retrieved from 
                                        parameter file and L0R */
)
{

    int read_count;               /* Parameter table size */
    int return_status;            /* Function return status */
    int default_utm_zone[] = {0}; /* Default UTM zone */
    int default_radiometric_char_flag[] = {1};  /* Default RPS char flag */
    int default_geometric_char_flag[] = {1};    /* Default GPS char flag */
    int default_use_three_tirs_scenes_flag[] = {0}; /* Default 3 scene flag */
    int default_retrieve_cal_file_flag[] = {1}; /* Default retrieval flag */
    int default_scenes_row_flag[] = {0}; /* Default insert SCENES row flag */

    /* Set up and read the application parameter table */
    IAS_PARM_DECLARE_TABLE(app_parm_table, 17);

    /* L0R filename */
    IAS_PARM_L0R_FILENAME(app_parm_table, IAS_PARM_REQUIRED,
            dsw_data->l0r_filename, sizeof(dsw_data->l0r_filename), 1);

    /* UTM zone */
    IAS_PARM_UTM_ZONE(app_parm_table, IAS_PARM_OPTIONAL, 1, default_utm_zone,
            &dsw_data->utm_zone, sizeof(dsw_data->utm_zone), 1);

    /* Note: not using the standard parameters for CPF, BPF, and RLUT
       as these are used as an override and have different help text */
    /* CPF filename */
    IAS_PARM_ADD_STRING(app_parm_table, CAL_PARM_FILENAME,
            "This parameter is provided for overriding the normal calibration "
            "parameter file", IAS_PARM_OPTIONAL, 0, NULL, 0, NULL, 
            dsw_data->cpf_filename, sizeof(dsw_data->cpf_filename), 1);

    /* BPF filenames */
    IAS_PARM_ADD_STRING(app_parm_table, BIAS_PARM_FILENAME_OLI,
            "This parameter is provided for overriding the normal OLI bias "
            "parameter file", IAS_PARM_OPTIONAL, 0, NULL, 0, NULL,
            dsw_data->bpf_filename_oli, sizeof(dsw_data->bpf_filename_oli), 1);
    IAS_PARM_ADD_STRING(app_parm_table, BIAS_PARM_FILENAME_TIRS,
            "This parameter is provided for overriding the normal TIRS bias "
            "parameter file", IAS_PARM_OPTIONAL, 0, NULL, 0, NULL,
            dsw_data->bpf_filename_tirs, sizeof(dsw_data->bpf_filename_tirs),
            1);

    /* RLUT filename */
    IAS_PARM_ADD_STRING(app_parm_table, RLUT_FILENAME,
            "This parameter is provided for overriding the normal response "
            "linearization look up table parameter file", IAS_PARM_OPTIONAL,
            0, NULL, 0, NULL, dsw_data->rlut_filename,
            sizeof(dsw_data->rlut_filename), 1);

    /* Flag to retrieve calibration files */
    IAS_PARM_ADD_BOOLEAN(app_parm_table, RETRIEVE_CALIBRATION_FILES_FLAG,
            "Parameter flag to determine if the CPF, BPF, and RLUT are to "
            "be retrieved for future processing.  1 - Retrieve the files. "
            "0 - do not retrieve the files", IAS_PARM_OPTIONAL,
            IAS_PARM_NOT_ARRAY, default_retrieve_cal_file_flag,
            &dsw_data->retrieve_calibration_files_flag,
            sizeof(dsw_data->retrieve_calibration_files_flag), 1);

    /* Flag to create radiometric characterization */
    IAS_PARM_ADD_BOOLEAN(app_parm_table, CREATE_RADIOMETRIC_CHAR_FLAG,
            "Parameter for populating radiometric WO common table to create a "
            "radiometric characterization ID", IAS_PARM_OPTIONAL,
            IAS_PARM_NOT_ARRAY, default_radiometric_char_flag,
            &dsw_data->radiometric_char_flag,
            sizeof(dsw_data->radiometric_char_flag), 1);

    /* Flag to create geometric characterization */
    IAS_PARM_ADD_BOOLEAN(app_parm_table, CREATE_GEOMETRIC_CHAR_FLAG,
            "Parameter for populating geometric WO common table to create a "
            "geometric characterization ID", IAS_PARM_OPTIONAL,
            IAS_PARM_NOT_ARRAY, default_geometric_char_flag,
            &dsw_data->geometric_char_flag,
            sizeof(dsw_data->geometric_char_flag), 1);

    /* Flag to create SCENES row */
    IAS_PARM_ADD_BOOLEAN(app_parm_table, INSERT_SCENES_ROW_FLAG,
            "Parameter for populating the SCENES table",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, default_scenes_row_flag,
            &dsw_data->insert_scenes_row_flag,
            sizeof(dsw_data->insert_scenes_row_flag), 1);

    /* Max processors to use */
    IAS_PARM_MAX_PROCESSORS_TO_USE(app_parm_table, IAS_PARM_OPTIONAL,
            &dsw_data->max_processors, sizeof(dsw_data->max_processors), 1);

    /* Work order ID */
    IAS_PARM_WORK_ORDER_ID(app_parm_table, dsw_data->work_order_id,
            sizeof(dsw_data->work_order_id), 1);

    /* Work order directory */
    IAS_PARM_ADD_STRING(app_parm_table, WO_DIRECTORY, "Work order directory",
            IAS_PARM_REQUIRED, 0, NULL, 0, NULL, dsw_data->work_order_dir,
            sizeof(dsw_data->work_order_dir), 1);

    /* Product Request ID */
    IAS_PARM_PRODUCT_REQUEST_ID(app_parm_table, dsw_data->product_request_id,
            sizeof(dsw_data->product_request_id), 1);

    /* Landsat Scene ID */
    /* Optional due to the fact that it will fallback to the method of 
       selecting the first record in the metadata instead of searching for
       the landsat scene id */
    strcpy(dsw_data->landsat_scene_id, "");
    IAS_PARM_LANDSAT_SCENE_ID(app_parm_table, IAS_PARM_OPTIONAL, 
            dsw_data->landsat_scene_id, sizeof(dsw_data->landsat_scene_id), 1);

    /* Flag to indicate that 3 TIRS scenes will be requested */
    IAS_PARM_USE_THREE_TIRS_SCENES_FLAG(app_parm_table, IAS_PARM_OPTIONAL,
            default_use_three_tirs_scenes_flag,
            &dsw_data->use_three_tirs_scenes_flag,
            sizeof(dsw_data->use_three_tirs_scenes_flag), 1);

    /* Sensor ID that overrides the L0R sensor_id parameter and is 
       used to populate the GEOMETRIC_WORK_ORDER_COMMON table */
    const char *valid_geo_override[] = {"OLITIRS", "OLI", "TIRS"};
    IAS_PARM_ADD_STRING(app_parm_table, GEOMETRY_SENSOR_ID_OVERRIDE,
            "Sensor ID which overrides sensor_id from L0r; this is used to "
            "populate the SENSOR_ID field in the geometry work order table",
            IAS_PARM_OPTIONAL, 3, valid_geo_override, 0, NULL, 
            dsw_data->geometry_sensor_id_override,
            sizeof(dsw_data->geometry_sensor_id_override), 0);

    read_count = IAS_PARM_GET_TABLE_SIZE(app_parm_table);

    /* Call the routine to provide help for the parameters if requested */
    return_status = ias_parm_provide_help(parm_filename, app_parm_table, 
            read_count, IAS_INPUT_PARAMETERS);
    if (return_status == ERROR)
    {
        /* An error occurred generating help */
        IAS_LOG_ERROR("With providing parameter help");
        return ERROR;
    }
    else if (return_status == 1)
    {
        /* Exit if help requested */
        exit(EXIT_SUCCESS);
    }

    /* Read the parameter file */
    return_status = ias_parm_read(parm_filename, app_parm_table, read_count);
    if (return_status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading application parameters from file %s",
                parm_filename);
        return ERROR;
    }

    /* Build the OMF name from the work order ID */
    sprintf(dsw_data->omf_filename, "%s.omf", dsw_data->work_order_id);

    return SUCCESS;
}

/******************************************************************
Name: calculate_utm_zone

Purpose: Calulate the UTM Zone based on the longitude 

Returns: The UTM Zone from 1 to 60
*******************************************************************/
int calculate_utm_zone
(
    double longitude            /* I: Longitude to calculate UTM zone for */
)
{
    longitude += 180.0;

    /* "wrap around" invalid ranges */
    while (longitude >= 360.0)
        longitude -= 360.0;
    while (longitude < 0.0)
        longitude += 360.0;

    return (int) (longitude / 6) + 1;
}
