#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "setup_work_order.h"

#define SERVICE_BPF        "/dataservices/rest/bpf?operation=getBiasParmFile"
#define SERVICE_CPF        "/dataservices/rest/cpf?operation=getCalParmFile"
#define SERVICE_FILE_NAME  "&filename="
#define SERVICE_NEAR_MATCH "&nearMatch=true&date="
#define SERVICE_SENSOR_OLI "&sensor=OLI"
#define SERVICE_SENSOR_TIRS "&sensor=TIRS"

/*------------------------------------------------------------------------------
NAME: retrieve_bpf_cpf

PURPOSE:
    Retrieve either a BPF or CPF from the CPS.  This function will return a flag
    indicating if the file should be renamed to the file name within the BPF or
    CPF.

    The renaming of the file is not performed within this function to avoid
    multiple reads of the CPF within this application for performance reasons.
    The calling function should call assign_bpf_name() or assign_cpf_name()
    accordingly.

RETURNS:
    This function returns a status of SUCCESS if the requested file is
    successfully retrieved from the CPS, otherwise ERROR is returned.
------------------------------------------------------------------------------*/
int retrieve_bpf_cpf (
    const char * date_time,   /* I:   Acquisition date time */
    int file_type,            /* I:   What type of file are we retrieving? */
    const char * service_url, /* I:   URL to the CPS */
    char path[PATH_MAX],      /* I/O: BPF or CPF file to be retrieved */
    int * rename_flag         /* O:   Should the file be renamed? */
    )
{
    char url[4096];
    int  satNum;
    int  status;
    int  url_size;

    *rename_flag = FALSE;

    /* Retrieve the satellite number so it can be used in the temporary names
       used when retrieving the BPF and CPF files. */
    satNum = ias_sat_attr_get_satellite_number();
    if (satNum == ERROR)
    {
        IAS_LOG_ERROR("Obtaining satellite number");
        return ERROR;
    }

    /* If the requested file path contains a '/' character, the path is
       interpreted as either an absolute path or a relative path to a file on
       disk.  If the file exists, return a SUCCESS status otherwise return an
       ERROR status to indicate the file does not exist. */
    if (strchr(path, '/') != NULL)
    {
        if (access(path, F_OK) == 0)
            return SUCCESS;
        IAS_LOG_ERROR("File does not exist: %s", path);
        return ERROR;
    }

    if (strcmp(path, "") == 0)
    {
        int path_size;

        /* If no name has been specified at all, attempt to retrieve the nearest
           match for the scene's acquisition date and time. */
        if (file_type == RETRIEVE_BPF_OLI)
        {
            url_size = snprintf(url, sizeof(url), "%s%s%s%s%s", service_url,
                SERVICE_BPF, SERVICE_SENSOR_OLI, SERVICE_NEAR_MATCH, date_time);
            path_size = snprintf(path, PATH_MAX, "LO%dBPF%s", satNum,
                date_time);
        }
        else if (file_type == RETRIEVE_BPF_TIRS)
        {
            url_size = snprintf(url, sizeof(url), "%s%s%s%s%s", service_url,
                SERVICE_BPF, SERVICE_SENSOR_TIRS, SERVICE_NEAR_MATCH,
                date_time);
            path_size = snprintf(path, PATH_MAX, "LT%dBPF%s", satNum,
                date_time);
        }
        else
        {
            url_size = snprintf(url, sizeof(url), "%s%s%s%s", service_url,
                SERVICE_CPF, SERVICE_NEAR_MATCH, date_time);
            path_size = snprintf(path, PATH_MAX, "CPF%s", date_time);
        }

        if ((path_size < 0) || (path_size >= PATH_MAX))
        {
            IAS_LOG_ERROR("File path truncated by snprintf()");
            return ERROR;
        }
        *rename_flag = TRUE;
    }
    else
    {
        /* A specific file has been named.  Attempt to retrieve that file from
           the CPS */
        if (file_type == RETRIEVE_BPF_OLI)
        {
            url_size = snprintf(url, sizeof(url), "%s%s%s%s%s", service_url,
                SERVICE_BPF, SERVICE_SENSOR_OLI, SERVICE_FILE_NAME, path);
        }
        else if (file_type == RETRIEVE_BPF_TIRS)
        {
            url_size = snprintf(url, sizeof(url), "%s%s%s%s%s", service_url,
                SERVICE_BPF, SERVICE_SENSOR_TIRS, SERVICE_FILE_NAME, path);
        }
        else
        {
            url_size = snprintf(url, sizeof(url), "%s%s%s%s", service_url,
                SERVICE_CPF, SERVICE_FILE_NAME, path);
        }
        
        *rename_flag = TRUE;
    }

    /* Ensure the URL created to retrieve the file was formed properly */
    if ((url_size < 0) || (url_size >= sizeof(url)))
    {
        IAS_LOG_ERROR("URL string truncated by snprintf()");
        return ERROR;
    }

    /* Retrieve the requested file */
    status = ias_misc_retrieve_file_from_url(url, path);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving file from CPS: %s", url);
        return ERROR;
    } 

    /* Verify the file exists */
    if (access(path, F_OK) != 0)
    {
        IAS_LOG_ERROR("File not found after retrieving from CPS: %s", path);
        return ERROR;
    }
    return SUCCESS;
}
