/******************************************************************************
NAME: setup_work_order (main)

PURPOSE:    
Main driver for the DSW. It does the initial setup for a work order.
For LDCM, that setup includes:
    Retrieving the CPF
    Retrieving the BPF (if not TIRS-only products)
    Retrieving the RLUT (if not TIRS-only products)
    Selecting the correct projection for the product
    Retrieving the correct UTM zone if the projection is UTM
    Generating the initial OMF
    Create link to the L0R directory
    Inserting a row into the Radiometric and/or Geometric Work Order Common
        tables if specified in the parameter file

RETURN VALUE:
Type = int
Value        Description
-----        -----------
EXIT_FAILURE    Error in setting up work order
EXIT_SUCCESS    Successful setup of work order
******************************************************************************/

#include <string.h>
#include <unistd.h>               /* For symlink */
#include <stdlib.h>

#include "ias_logging.h"
#include "ias_db.h"
#include "ias_db_get_connect_info.h"
#include "ias_miscellaneous.h"
#include "ias_satellite_attributes.h"
#include "gctp.h"                 /* For UTM definition */
#include "setup_work_order.h"

int main(int argc, char *argv[])
{
    int status = 0;               /* Status returned from function call */
    char *processing_center;      /* Holds processing center env var */
    DSW_DATA dsw_data;            /* Parameters and values read from parameter 
                                     file and L0R */
    const char *ias_version;      /* IAS software version */
    char date_acquired[DATE_LENGTH]; /* String for format-converted date */

    /* CPF variables */
    IAS_CPF *cpf = NULL;
    const struct IAS_CPF_DETECTOR_STATUS *det_status = NULL;
    const struct IAS_CPF_EARTH_CONSTANTS *earth_consts = NULL;
    const struct IAS_CPF_ORBIT_PARAMETERS *orbit_parms = NULL;
    const struct IAS_CPF_FOCAL_PLANE *focal_plane = NULL;
    const struct IAS_CPF_FILE_ATTRIBUTES *file_attr = NULL;

    /* Initialize the IAS logging library */
    status = ias_log_initialize("setup_work_order");
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing log");
        exit(EXIT_FAILURE);
    }

    /* Check the number of arguments
       We want the program name and the ODL file for args */
    if (argc != 2)
    {
        IAS_LOG_ERROR("Usage: %s ODL_Parameter_File", argv[0]);
        exit(EXIT_FAILURE);
    }

    memset(&dsw_data, 0, sizeof(dsw_data));

    /* Read parameter file -- argv[1] should be the parameter file name */
    status = read_parameters(argv[1], &dsw_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters");
        exit(EXIT_FAILURE);
    }

    /* Initialize the satellite attributes.  Note, making the assumption here
       that L8 is the only supported one for now.  If more need to be 
       supported, it will need to be a parameter to the application since
       the satellite attributes library needs to be initialized for the
       L0R library. */
    if (ias_sat_attr_initialize(IAS_L8) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attributes");
        return ERROR;
    }

    /* Log initiation */
    IAS_LOG_INFO("Initiated");

    /* Get PROCESSING_CENTER environment variable */
    processing_center = getenv("PROCESSING_CENTER");
    if (processing_center == NULL)
    {
        IAS_LOG_ERROR("Getting PROCESSING_CENTER env variable");
        exit(EXIT_FAILURE);
    }
    status = snprintf(dsw_data.processing_center,
        sizeof(dsw_data.processing_center), "%s", processing_center);
    if ((status < 0) || (status >= sizeof(dsw_data.processing_center)))
    {
        IAS_LOG_ERROR("Processing center string truncated by snprintf");
        exit(EXIT_FAILURE);
    }

    /* Read L0R metadata and create L0R directory link */
    status = read_l0r_metadata(&dsw_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading L0R metadata");
        exit(EXIT_FAILURE);
    }

    /* Convert the date to a CPF/BPF/RLUT time string */
    status = snprintf(date_acquired, sizeof(date_acquired),
        "%04d-%02d-%02dT%02d:%02d:%02d", dsw_data.acquisition_date.year,
        dsw_data.acquisition_date.month, dsw_data.acquisition_date.day_of_month,
        dsw_data.acquisition_date.hour, dsw_data.acquisition_date.minute,
        (int)dsw_data.acquisition_date.second);
    if ((status < 0) || (status >= sizeof(date_acquired)))
    {
        IAS_LOG_ERROR("Date acquired string truncated by snprintf");
        exit(EXIT_FAILURE);
    }

    if (dsw_data.retrieve_calibration_files_flag)
    {
        int rename_bpf;
        int rename_cpf;
        char target_cpf_name[PATH_MAX];
        const char *services_location; /* Holds services location env var */

        /* Get IAS_SERVICES environment variable */
        services_location = getenv("IAS_SERVICES");
        if (services_location == NULL)
        {
            IAS_LOG_ERROR("Getting IAS_SERVICES env variable");
            exit(EXIT_FAILURE);
        }
        status = snprintf(dsw_data.services_location, 
            sizeof(dsw_data.services_location), "%s", services_location);
        if ((status < 0) || (status >= sizeof(dsw_data.services_location)))
        {
            IAS_LOG_ERROR("Services location string truncated by snprintf "
                "(%d which is larger than %d)", status,
                (int)sizeof(dsw_data.services_location));
            exit(EXIT_FAILURE);
        }

        /* Retrieve the CPF */
        status = retrieve_bpf_cpf(date_acquired, RETRIEVE_CPF,
            dsw_data.services_location, dsw_data.cpf_filename, &rename_cpf);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Retrieving CPF");
            exit(EXIT_FAILURE);
        }

        /* Retrieve the OLI BPF */
        if (dsw_data.sensor_present_oli)
        {
            status = retrieve_bpf_cpf(date_acquired, RETRIEVE_BPF_OLI,
                dsw_data.services_location, dsw_data.bpf_filename_oli,
                &rename_bpf);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Retrieving OLI BPF");
                exit(EXIT_FAILURE);
            }

            /* Rename the OLI BPF to match the file name stored within itself */
            if (rename_bpf &&
                (assign_bpf_name(dsw_data.bpf_filename_oli) != SUCCESS))
            {
                IAS_LOG_ERROR("Renaming TIRS BPF file");
                exit(EXIT_FAILURE);
            }
        }

        /* Retrieve the TIRS BPF */
        if (dsw_data.sensor_present_tirs)
        {
            status = retrieve_bpf_cpf(date_acquired, RETRIEVE_BPF_TIRS,
                dsw_data.services_location, dsw_data.bpf_filename_tirs,
                &rename_bpf);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Retrieving TIRS BPF");
                exit(EXIT_FAILURE);
            }

            /* Rename the TIRS BPF to match the file name stored within itself
               */
            if (rename_bpf &&
                (assign_bpf_name(dsw_data.bpf_filename_tirs) != SUCCESS))
            {
                IAS_LOG_ERROR("Renaming TIRS BPF file");
                exit(EXIT_FAILURE);
            }
        }

        /* Cache the RLUT */
        status = ias_misc_retrieve_rlut(date_acquired,
            dsw_data.services_location, dsw_data.rlut_filename,
            sizeof(dsw_data.rlut_filename));
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Caching RLUT file");
            exit(EXIT_FAILURE);
        }

        /* Read the CPF, then get the OLI detector select table and its ID and
           check their values against the corresponding values in the L0R OLI
           image header */
        cpf = ias_cpf_read(dsw_data.cpf_filename);
        if (!cpf)
        {
            IAS_LOG_ERROR("Can't read CPF file %s", dsw_data.cpf_filename);
            exit(EXIT_FAILURE);    
        }

        det_status = ias_cpf_get_det_status(cpf);
        if (!det_status)
        {
            IAS_LOG_ERROR("Can't get the CPF Detector Status group");
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }

        if (dsw_data.sensor_present_oli)
        {
            if( dsw_data.image_header_filled == 1)
            {
                IAS_LOG_WARNING("The image header was filled, therefore "
                    "its detector select table values will not be checked "
                    "against the values in the CPF");
            }
            else if ((det_status->oli_detector_select_table != 
                dsw_data.oli_detector_select_table) || 
                (det_status->oli_detector_select_table_id != 
                dsw_data.oli_detector_select_table_id))
            {
                /* Print out error message and exit out gracefully */
                IAS_LOG_ERROR("L0R %s OLI detector select table %d and id %d "
                            "do not match CPF values of %d and %d",
                            dsw_data.l0r_filename, 
                            dsw_data.oli_detector_select_table, 
                            dsw_data.oli_detector_select_table_id, 
                            det_status->oli_detector_select_table, 
                            det_status->oli_detector_select_table_id);
                ias_cpf_free(cpf);
                exit(EXIT_FAILURE);    
            }
        }

        if (dsw_data.sensor_present_tirs)
        {
            if ((det_status->tirs_detector_map_id != 
                dsw_data.detector_map_id_tirs))
            {
                /* Print out error message and exit out gracefully */
                IAS_LOG_ERROR("L0R %s TIRS detector map id %d does not "
                            "match CPF value of %d",
                            dsw_data.l0r_filename,
                            dsw_data.detector_map_id_tirs,
                            det_status->tirs_detector_map_id);
                ias_cpf_free(cpf);
                exit(EXIT_FAILURE);    
            }
        }

        /* Get the earth constants, orbit parameter, and focal plane
        information from the CPF */
        earth_consts = ias_cpf_get_earth_const(cpf);
        if (earth_consts == NULL)
        {
            IAS_LOG_ERROR("Getting earth constant information from CPF");
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }
 
        orbit_parms = ias_cpf_get_orbit(cpf);
        if (orbit_parms == NULL)
        {
            IAS_LOG_ERROR("Getting orbit parameter information from CPF");
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }
 
        focal_plane = ias_cpf_get_focal_plane(cpf);
        if (focal_plane == NULL)
        {
            IAS_LOG_ERROR("Getting earth constant information from CPF");
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }

        /* Extend the corners as needed */
        status = extend_corners(&dsw_data, earth_consts, orbit_parms,
            focal_plane);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Extending corners");
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }

        /* Calculate corner points considering both OLI and TIRS corners */
        status = calc_l0r_corners(&dsw_data);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating corners");
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }

        /* Read the file attributes to get the collection number and file
           name */
        file_attr = ias_cpf_get_file_attributes(cpf);
        if (file_attr == NULL)
        {
            IAS_LOG_ERROR("Retrieving file attributes from CPF: %s",
                    dsw_data.cpf_filename);
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }

        /* Copy the collection number from the CPF */
        dsw_data.collection_number = file_attr->collection_number;

        /* If the CPF needs be renamed, get the target name before freeing
           the CPF (since it should not be open when it is renamed) */
        if (rename_cpf)
        {
            if (strlen(file_attr->file_name) >= PATH_MAX)
            {
                IAS_LOG_ERROR("CPF file name exceeds buffer size");
                ias_cpf_free(cpf);
                exit(EXIT_FAILURE);
            }

            strcpy(target_cpf_name, file_attr->file_name);
        }

        /* Free the CPF resources */
        ias_cpf_free(cpf);
        cpf = NULL;

        /* If necessary, rename the CPF to the target name */
        if (rename_cpf
            && (assign_cpf_name(target_cpf_name, dsw_data.cpf_filename)
            != SUCCESS))
        {
            IAS_LOG_ERROR("Renaming CPF file");
            exit(EXIT_FAILURE);
        }
    }

    /* Get the software version */
    ias_version = ias_misc_get_software_version();
    if (ias_version == NULL)
        strcpy(dsw_data.sw_version, "");
    else
    {
        status = snprintf(dsw_data.sw_version, sizeof(dsw_data.sw_version), 
            "%s", ias_version);
        if ((status < 0) || (status >= sizeof(dsw_data.sw_version)))
        {
            IAS_LOG_ERROR("Software version string truncated by snprintf");
            exit(EXIT_FAILURE);
        }
    }

    /* Set the target projection for Earth collections */
    if (dsw_data.acquisition_type == IAS_EARTH)
    {
        /* Latitude > -63 and the target WRS row is in the normal range
           (i.e. not 880-889 or 990-999 for off-nadir at the poles) means
           UTM projection. */
        if ((dsw_data.scene_center.lat > -63.0) && (dsw_data.row <= 248))
        {
            /* Use Universal Transverse Mercator (UTM) */
            dsw_data.projection = UTM;
        }
        else
        {
            /* Latitude <= -63 or off-nadir polar WRS row number means
               it is a polar region, so use Polar Stereographic */
            dsw_data.projection = PS; /* 6 = Polar Stereographic (PS) */
        }
    }

    /* Setup DB connection for retrieving the UTM zone, updating the
       radiometric/geometric work order common tables, and updating the scenes
       table (but only if it is needed) */
    if ((dsw_data.radiometric_char_flag == 1)
        || (dsw_data.geometric_char_flag == 1)
        || (dsw_data.insert_scenes_row_flag == 1)
        || (dsw_data.projection == UTM && dsw_data.utm_zone == 0))
    {
        /* Database connection variables */
        char *database_name = NULL;
        char *user_name = NULL;
        char *password = NULL;
        char *host = NULL;
        struct ias_db_connection *db = NULL;

        /* Initialize the database library */
        status = ias_db_initialize_database_lib();
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("From ias_db_initialize_database_lib");
            exit(EXIT_FAILURE);
        }

        /* Get the database connection info */
        status = ias_db_get_connect_info(&database_name, &user_name,
            &password, &host);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("From ias_db_get_connect_info");
            free(database_name);
            free(user_name);
            free(password);
            free(host);
            ias_db_close_database_lib();
            exit(EXIT_FAILURE);
        }

        /* Connect to the database */
        db = ias_db_connect_to_database(database_name, user_name,
            password, host);

        /* Free the database connection information */
        free(database_name);
        free(user_name);
        free(password);
        free(host);

        /* Confirm the DB connection succeeded */
        if (!db)
        {
            IAS_LOG_ERROR("From ias_db_connect_to_database: %s",
                ias_db_connect_last_error(db));
            ias_db_close_database_lib();
            exit(EXIT_FAILURE);
        }

        /* For the UTM projection, get the UTM zone if one was not provided in
           the parameters */
        if ((dsw_data.projection == UTM) && (dsw_data.utm_zone == 0))
        {
            /* retrieve UTM zone using path and row */
            status = retrieve_utm_zone(&dsw_data, db);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Unable to determine the UTM zone for "
                    "path %d, row %d", dsw_data.path, dsw_data.row);
                ias_db_close_connection(db);
                ias_db_close_database_lib();
                exit(EXIT_FAILURE);
            }
        }

        /* Before starting a transaction, make sure a database insert needs
           to be done */
        if ((dsw_data.radiometric_char_flag == 1)
            || (dsw_data.geometric_char_flag == 1)
            || (dsw_data.insert_scenes_row_flag == 1))
        {
            /* Start the transaction */
            status = ias_db_start_transaction(db);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("From ias_db_start_transaction: %s", 
                    ias_db_connect_last_error(db));
                ias_db_close_connection(db);
                ias_db_close_database_lib();
                exit(EXIT_FAILURE);
            }

            /* If the parameter specifying population of the radiometric WO
               common table is set and insert a record into the radiometric
               work order common table */
            if (dsw_data.radiometric_char_flag == 1)
            {
                /* Insert row into radiometric WO Common tables */
                status = insert_radiometric_row(&dsw_data, db);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Inserting radiometric row");
                    ias_db_rollback_transaction(db);
                    ias_db_close_connection(db);
                    ias_db_close_database_lib();
                    exit(EXIT_FAILURE);
                }
            }

            /* If the parameter specifying population of the geometric WO common
               table is set, insert a record into the geometric work order
               common table */
            if (dsw_data.geometric_char_flag == 1)
            {
                /* Insert row into geometric WO Common tables */
                status = insert_geometric_row(&dsw_data, db);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Inserting geometric row");
                    ias_db_rollback_transaction(db);
                    ias_db_close_connection(db);
                    ias_db_close_database_lib();
                    exit(EXIT_FAILURE);
                }
            }

            /* Insert the scenes row */
            if (dsw_data.insert_scenes_row_flag)
            {
                if (insert_scenes_table_row(&dsw_data, db) != SUCCESS)
                {
                    IAS_LOG_ERROR("Inserting scenes row");
                    ias_db_rollback_transaction(db);
                    ias_db_close_connection(db);
                    ias_db_close_database_lib();
                    exit(EXIT_FAILURE);
                }
            }

            /* Commit the transaction */
            status = ias_db_commit_transaction(db);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Committing transaction: %s",
                        ias_db_connect_last_error(db));
                ias_db_close_connection(db);
                ias_db_close_database_lib();
                exit(EXIT_FAILURE);
            }
        }

        /* Close the connection */
        ias_db_close_connection(db);
        ias_db_close_database_lib();
    }

    /* Calculate the UTM zone using a formula. This is mostly intended for use
       by ICs */
    if (dsw_data.utm_zone == -1)
    {
        dsw_data.utm_zone = calculate_utm_zone(dsw_data.scene_center.lng);
    }

    /* Generate OMF */
    status = generate_omf(&dsw_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Generating OMF");
        exit(EXIT_FAILURE);
    }

    IAS_LOG_INFO("Successful completion");

    return EXIT_SUCCESS;
}
