#ifndef _SETUPWORKORDER_H_
#define _SETUPWORKORDER_H_

#include <limits.h>                /* PATH_MAX definition */
#include "ias_const.h"             /* SUCCESS, ERROR definition */
#include "ias_l0r.h"               /* For IAS_SENSOR_ID_LENGTH definition */
#include "ias_db.h"                /* For ias_db_connection definition */
#include "ias_cpf.h"               /* For CPF definition */ 
#include "ias_los_model.h"         /* For IAS_EARTH_CHARACTERISTICS */


/* Flags indicating the type of parameter file that is to be retrieved from the
   CPS */
#define RETRIEVE_BPF_OLI  1
#define RETRIEVE_BPF_TIRS 2
#define RETRIEVE_CPF      3

#define SPACECRAFT_ID_LENGTH 80

/* Processing center is currently either "EROS", "SDSU", or "GSFC", but is 20
   characters in the DB column */
#define PROCESSING_CENTER_NAME_LENGTH 20

/* Web services location (defined by IAS_SERVICES environment variable) for BPF,
   CPF, and RLUT retrieval */
#define SERVICES_LOCATION_LENGTH 200

#define ACQUISITION_TYPE_LENGTH 80
#define DATE_LENGTH 20             /* Date format: YYYY-MM-DD HH:MM:SS */


/* Application parameter file parameters and L0R metadata values */
typedef struct dsw_data 
{
    char l0r_filename[PATH_MAX];   /* L0R product location */
    char l0r_id[PATH_MAX];         /* L0R filename (excluding path) */
    int  utm_zone;                 /* UTM zone */
    char omf_filename[PATH_MAX];   /* OMF filename */
    char cpf_filename[PATH_MAX];   /* CPF filename */
    char bpf_filename_oli[PATH_MAX];/* BPF filename */
    char bpf_filename_tirs[PATH_MAX];/* BPF filename */
    char rlut_filename[PATH_MAX];  /* RLUT filename */
    char sw_version[IAS_SOFTWARE_VERSION_SIZE]; /* IAS software version */
    char is_version[IAS_SOFTWARE_VERSION_SIZE]; /* L0R IS Software version */
    int collection_number;         /* Collection number of IAS software */
    int  path;                     /* WRS path */
    int  row;                      /* WRS row */
    int  projection;               /* Projection, PS (6) or UTM (1) */
    char spacecraft_id[SPACECRAFT_ID_LENGTH]; /* Spacecraft name */
    char processing_center[PROCESSING_CENTER_NAME_LENGTH]; /* Processing center 
                                                              name */
    char services_location[SERVICES_LOCATION_LENGTH]; /* Services location */
    int  max_processors;           /* Max processors to use */
    int acquisition_type;          /* Acquistion type: should be IAS_EARTH,
                                      IAS_LUNAR or IAS_STELLAR */
    IAS_DATETIME acquisition_date; /* Acquistion date */
    char acquisition_date_string[IAS_L0R_DATE_LENGTH+1];
                                   /* Acquistion date string representation*/
    double off_nadir_angle;        /* Scene off-nadir roll angle */
    char collection_type[IAS_L0R_COLLECTION_TYPE_LENGTH+1]; /* Type of data in 
                                      this collection: should be EARTH, LUNAR 
                                      or STELLAR */
    char landsat_scene_id[IAS_L0R_SCENE_ID_LENGTH+1]; /* Unique Landsat scene 
                                                         identifier */
    double integration_time_oli_pan;/* OLI integration time in milliseconds 
                                        for the pan band */
    double integration_time_oli_ms; /* OLI integration time in milliseconds
                                       for the multispectral band */
    double integration_time_tirs;   /* TIRS integration time in milliseconds */
    char sensor_id_from_l0r[IAS_L0R_SENSOR_ID_LENGTH+1]; /* Sensor used to 
                                      capture interval -- read from L0R */
    int image_header_filled;     /* Flag to keep track if the Image Header
                                    was filled. 1 if it was, 0 if it was not */
    char geometry_sensor_id_override[IAS_L0R_SENSOR_ID_LENGTH+1]; /* Sensor used
                                      to capture interval -- read from parameter
                                      file; overrides sensor_id_from_l0r for 
                                      writing to geometry work order table */
    int  satellite_id;             /* Satellite ID */
    char work_order_id[IAS_WORKORDER_ID_SIZE]; /* Uniquely identifies the
                                                  work order */
    char product_request_id
        [IAS_PRODUCT_REQUEST_ID_SIZE]; /* Uniquely identifies the LPGS
                                          Product Request ID*/
    char work_order_dir[PATH_MAX]; /* Work order directory */
    int  radiometric_char_flag;    /* Flag to create radiometric
                                      characterization ID */
    int  geometric_char_flag;      /* Flag to create geometric
                                      characterization ID */
    int  insert_scenes_row_flag;   /* Flag to create row in SCENES table */
    int  retrieve_calibration_files_flag; /* Flag to determine if the CPF, BPF,
                                      and RLUT files are to be retrieved */
    char landsat_interval_id[IAS_L0R_INTERVAL_ID_LENGTH+1]; /* Landsat interval 
                                                               identifier */
    int oli_detector_select_table; /* OLI detector map identifier */
    int oli_detector_select_table_id; /* OLI detector map identifier */
    unsigned int detector_map_id_tirs; /* TIRS detector map identifier */
    IAS_DBL_LAT_LONG scene_center; /* Scene center latitude and longitude from
                                      L0R */
    double upper_left_lat_lon[2];  /* X/Y Scene boundary coordinates */
    double upper_right_lat_lon[2];
    double lower_left_lat_lon[2];
    double lower_right_lat_lon[2];
    IAS_L0R_CORNERS corners_oli;   /* OLI scene corner points */
    IAS_L0R_CORNERS corners_tirs;  /* TIRS scene corner points */
    int sensor_present_oli;        /* 1 if data for OLI is present, 0 if not */
    int sensor_present_tirs;       /* 1 if data for TIRS is present, 0 if not */
    double image_start_time_sec_since_j2000; /* Image start time */
    double image_stop_time_sec_since_j2000;  /* Image stop time */
    int image_data_truncation_setting; /* Upper or lower 12 bits for image 
                                          pixels only - 0 = upper, 1 = lower */
    int record_count;              /* Number of records in L0R metadata */
    char station_id[IAS_L0R_STATION_ID_LENGTH+1];/* Ground station ID */
    int use_three_tirs_scenes_flag;/* Flag to indicate 3 tirs scenes are used */
} DSW_DATA;

/* Local functions */
int read_parameters
(
    const char *parm_filename,     /* I: Parameter filename */
    DSW_DATA *dsw_data             /* O: Parameters and values retrieved from 
                                         parameter file and L0R */
);

int read_l0r_metadata
(
    DSW_DATA *dsw_data             /* I/O: Parameters and values retrieved from 
                                         parameter file and L0R */
);

int retrieve_bpf_cpf
(
    const char * date_time,   /* I:   Acquisition date and time */
    int file_type,            /* I:   What type of file is to be retrieved */
    const char * service_url, /* I:   URL to the CPS service */
    char path[PATH_MAX],      /* I/O: BPF or CPF file to be retrieved */
    int * rename_flag         /* O:   Should the file be renamed? */
);

int generate_omf
(
    DSW_DATA *dsw_data             /* I: Parameters and values retrieved from 
                                         parameter file and L0R */
);

int insert_radiometric_row
(
    const DSW_DATA *dsw_data,      /* I: Parameters and values retrieved from 
                                         parameter file and L0R */
    struct ias_db_connection *db   /* I: Database connection */
);

int insert_geometric_row
(
    const DSW_DATA *dsw_data,      /* I: Parameters and values retrieved from 
                                         parameter file and L0R */
    struct ias_db_connection *db   /* I: Database connection */
);

int insert_scenes_table_row
(
    const DSW_DATA *dsw_data,          /* I: Parameters and values retrieved 
                                             from parameter file and L0R */
    struct ias_db_connection *db       /* I: Database connection */
);

int calculate_utm_zone
(
    double longitude            /* I: Longitude to calculate UTM zone for */
);

int retrieve_utm_zone
(
    DSW_DATA *dsw_data,            /* I/O: Parameters and values retrieved 
                                           from parameter file and L0R */
    struct ias_db_connection *db   /* I: Database connection */
);

int assign_cpf_name
(
    const char *target_cpf_name,/* I: File name to use for the renamed CPF */
    char *path                  /* I/O: Path to the CPF */
);

int assign_bpf_name
(
    char path[PATH_MAX] /* IO: Path to a BPF */
);

int extend_corners
(
    DSW_DATA *dsw_data,      /* I/O: Scene metadata for bands */
    const struct IAS_CPF_EARTH_CONSTANTS *earth_parameters,
                             /* I: Earth constants */
    const struct IAS_CPF_ORBIT_PARAMETERS *orbit_parameters,
                             /* I: Orbit parameters */
    const struct IAS_CPF_FOCAL_PLANE *focal_plane
                             /* I: Focal plane parameters */
);

int find_fov
(
    IAS_SENSOR_ID sensor, /* I: ID of sensor whose fields of view will be
                             calculated */
    const IAS_CPF_FOCAL_PLANE *focal, /* I: Focal plane parameters from the
                             CPF */
    double *fov_at,       /* O: Along-track field of view */
    double *fov_xt        /* O: Across-track field of view */
);

int move_corners
(
    const IAS_L0R_CORNERS *old_corners, /* I: Original scene corners */
    double roll_angle,               /* I: Scene off-nadir angle */
    const IAS_DBL_LAT_LONG *scene_center, /* I: Scene center coordinates */
    const IAS_EARTH_CHARACTERISTICS *earth, /* I: Earth parameters */
    double o_radius,                 /* I: Nominal orbit radius */
    double fov_at,                   /* I: Along-track field of view */
    double fov_xt,                   /* I: Across-track field of view */
    IAS_L0R_CORNERS *new_corners     /* O: New scene corners */
);

int calc_l0r_corners
(
    DSW_DATA *dsw_data /* I/O: selected scene metadata */
);

#endif
