/**\file ot_subroutines.h
 * Header file for application that implements the subsetting of L0Ra data
 * to product L0Rp data
 */
/*-----------------------------------------------------------------------------

 NAME:     ot_subroutines.h

 PURPOSE:  Header file for the application that implements the subsetting of
           L0Ra data to produce L0Rp data

 ALGORITHM REFERENCES: None

 -----------------------------------------------------------------------------*/
#ifndef OT_SUBROUTINES_H
#define OT_SUBROUTINES_H

#include <dirent.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include "ots_version.h"

#include "ias_logging.h"
#include "ias_types.h"
#include "ias_l0r.h"
#include "ias_satellite_attributes.h"

/**
 * Struct for the L0R (a/p) information 
 */
typedef struct 
{
    L0RIO *l0r_data;             /**< Pointer to the L0R data for File I/O */
    char l0r_prefix[PATH_MAX];   /**< Prefix for the L0R files */
    char l0r_location[PATH_MAX]; /**< L0R location */
    int read_write_status;       /**< Read/write status when opening files */
}L0R_INFORMATION;

int check_file_system
(
    const char *l0ra_location, /* I: L0Ra dataset location */
    const char *l0rp_location, /* I: L0Rp dataset location */
    int *same_file_system      /* O: Flag indicating if L0Ra and L0Rp
                                     data is on the same file system */
);

int copy_ancillary_data
(
    L0R_INFORMATION l0ra_data, /* I: Pointer to the L0Ra structure for I/O */
    L0R_INFORMATION l0rp_data  /* I: Pointer to the L0Rp structure for I/O */
);

int copy_offsets
(
    L0RIO *l0ra_data, /* I: Pointer to the L0Ra structure for I/O */
    L0RIO *l0rp_data, /* I: Pointer to the L0Rp structure for I/O */
    const int band_number,  /* I: Current band number processing */
    const IAS_BAND_ATTRIBUTES *band_attributes /* I: Current band attributes */
);

int create_l0rp_metadata
(
    L0R_INFORMATION struct_l0ra_metadata, /* I/O: L0Ra info struct */
    L0R_INFORMATION struct_l0rp_metadata, /* I/O: L0Rp info struct */
    const int start_line,                 /* I: Subset beginning line */
    const int stop_line                   /* I: Subset ending line */
);

int generate_checksum_file
(
    const char *l0rp_location, /* I: Output L0Rp dataset location */
    const char *l0r_prefix     /* I: L0Rp file prefix */
);

int get_file_metadata
(
    L0RIO *l0ra_data,          /* I/O: Struct pointer for the L0Ra File Input */
    L0RIO *l0rp_data,          /* I/O: Struct pointer for the L0Rp File 
                                       Output */
    const char *l0rp_location, /* I: L0Rp output location */
    const char *l0rp_prefix    /* I: L0Rp file prefix */
);

int get_interval_metadata
(
    L0RIO *l0ra_data, /* I: Pointer to struct for the L0Ra File Input */
    L0RIO *l0rp_data  /* I: Pointer to struct for the L0Rp File Output */
);

int get_scene_metadata
(
    L0RIO *l0ra_data, /* I: Pointer to a struct for the L0Ra file input */
    L0RIO *l0rp_data, /* I: Pointer to a struct for the L0Rp file input */
    int start_line,   /* I: Subset beginning line */
    int stop_line     /* I: Subset ending line */
);

int get_start_stop_lines_from_scene
(
    L0RIO *l0r_data,       /* I: Pointer to the L0Ra data */
    const char *scene_id,  /* I: Scene ID for desired subset */
    int *start_line,       /* O: Beginning line for desired subset */
    int *stop_line         /* O: Ending line for desired subset */
);

char *get_version_number();

int initialize_l0r_data
(   
    L0R_INFORMATION *struct_l0r_data /* I/O: Struct for L0R data */
);

void initialize_L0R_information
(
    const char *l0r_prefix,    /* I: L0R file prefix */
    const char *l0r_location,  /* I: L0R location */
    int l0ra,                  /* I: Flag indicating L0Ra or L0Rp */
    L0R_INFORMATION *l0r_info  /* O: Struct for the L0R Information */
);

int interval_request
(
    const char *l0ra_location, /* I: Input L0Ra location */
    const char *l0rp_location, /* I: Input L0Rp location */
    char *interval_id          /* I: Landsat Interval ID */
);

int subset_band_image_data
(
    L0RIO *l0ra_data,     /* I/O: Pointer to struct for L0Ra data */
    L0RIO *l0rp_data,     /* I/O: Pointer to struct for L0Ra data */
    const int start_line, /* I: Beginning line of desired subset */
    const int stop_line,  /* I: Ending line of desired subset */
    int remove_tirs_secondary_bands /* Flag indicating whether TIRS Redundant
                                       bands should be removed */
);

int subset_by_scene
(
    const char *l0ra_prefix,   /* I: L0Ra file prefix */
    const char *l0rp_scene_id, /* I: L0Rp file prefix, Landsat Scene ID */
    const char *l0ra_location, /* I: L0Ra dataset location */
    const char *l0rp_location, /* I: L0Rp dataset location */
    int remove_tirs_secondary_bands /* I: Flag indicating whether TIRS Redundant
                                          bands should be removed */
);

int terminate_l0r_access
(
    L0R_INFORMATION *struct_l0r_data /* I/O: L0R information struct */
);

#endif
