/**\file subsetter_io_test.c
 * Implements the unit tests of the subsetting of the L0Ra data to
 * produce L0Rp data
 */
/*-----------------------------------------------------------------------------

 NAME: subsetter_io_test.c

 PURPOSE: Implements the unit tests of the subsetting of L0Ra data to produce 
          L0Rp data

 ALGORITHM REFERENCES: NONE

 NOTES:

 -----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>
#include <limits.h>

#include "cgreen/cgreen.h"
#include "ias_l0r.h"
#include "ias_types.h"
#include "ot_subroutines.h"

const int number_of_lines = 500;
/* note: number_of_lines should be at least 6 times the size
   of cycle number */
const int cycle_number = 40;
/* note: cycle_number * 1000 + # of dectectors must be smaller 
    than 65536 (2^16).
    Every cycle_number of lines the generated image data will start over */
const int line_count = 1;
char l0ra_location[PATH_MAX];
char l0rp_location[PATH_MAX];

/**
 * Tests the get_command_line_parameters function in the main routine
 */
static void test_subsetter_get_command_line_parameters()
{
    char command[PATH_MAX * 3];/* Command to test the command line
                                  parameters */
    int status = SUCCESS;      /* Status variable for checking function calls */

    /* Subset scene 1 with out specifying the L0Ra prefix
           (it will come from the interval) */
    snprintf(command, 
        PATH_MAX*3,  "../ot_subsetter \"test\" \"%s\" \"%s\" --scene "
        "\"SCENE_ID_1\"", l0ra_location, l0rp_location);
    IAS_LOG_INFO("Running system command: %s", command);
    status = system(command);
    if (status == -1 || WEXITSTATUS(status) < 0)
    {
        IAS_LOG_ERROR("Error encountered when running command: %s",
            command);
    }
    assert_true(status == SUCCESS);
   
    /* Subset scene 2 and specify the L0Ra prefix */
    snprintf(command, 
        PATH_MAX*3, "../ot_subsetter \"test\" \"%s\" \"%s\" --scene "
        "\"SCENE_ID_2\" --l0ra_file_prefix \"test\" "
        "--remove_tirs_secondary_bands", l0ra_location, l0rp_location);
    IAS_LOG_INFO("Running system command: %s", command);
    status = system(command);
    if (status == -1 || WEXITSTATUS(status) < 0)
    {
        IAS_LOG_ERROR("Error encountered when running command: %s",
            command);
    }
    assert_true(status == SUCCESS);

    /* Test subset by interval */
    snprintf(command,
        PATH_MAX*3, "../ot_subsetter \"test\" \"%s\" \"%s%s\"",
        l0ra_location, l0rp_location, "/interval_request");
    IAS_LOG_INFO("Running system command: %s", command);
    status = system(command);
    if (status == -1 || WEXITSTATUS(status) < 0)
    {
        IAS_LOG_ERROR("Error encountered when running command: %s",
            command);
    }
    assert_true(status == SUCCESS);
}

/**
 * Generates (and tests the generation of) L0Ra data
 *
 * The test image and offsets data is using a counting pattern to
 * ease with verification and troubleshooting.  The test header data
 * is randomly generated (except for the J2000 and the line_status),
 * and has no significant meaning or value.
 */
static void test_generate_l0ra()
{
    int band_number = 0;       /* Current band number processing */
    int i = 0;                 /* Looping variable */
    int failure_flag = 0;      /* Flag indicating failure encountered */
    int line = 0;              /* Looping variable for current line */
    int number_of_detectors = 0; /* Total number of detectors */
    int status;                /* Status variable to check function calls */

    uint16_t *image = NULL;    /* Image data buffer */
    uint16_t *top_offsets = NULL;    /* Top offsets buffer */
    uint16_t *bottom_offsets = NULL; /* Bottom offsets buffer */

    L0RIO *l0ra_data = NULL; /* Pointer to L0Ra data */

    /* Band attributes to be obtained from satellite attributes library */
    const IAS_BAND_ATTRIBUTES *band_attributes = NULL;
    const IAS_SATELLITE_ATTRIBUTES *landsat8_attributes = NULL;

    /* Number of lines to write to hdf5 file */
    int num_lines_to_write = number_of_lines;

    /* Structures from L0R I/O Library */
    IAS_L0R_EPHEMERIS ephemeris;
    memset(&ephemeris, 0, sizeof(IAS_L0R_EPHEMERIS));

    IAS_L0R_OLI_IMAGE_HEADER oli_frame_header;
    memset(&oli_frame_header, 0, sizeof(IAS_L0R_OLI_IMAGE_HEADER));

    IAS_L0R_OLI_FRAME_HEADER oli_line_header;
    memset(&oli_frame_header, 0, sizeof(IAS_L0R_OLI_FRAME_HEADER));

    IAS_L0R_TIRS_FRAME_HEADER tirs_line_header;
    memset(&oli_frame_header, 0, sizeof(IAS_L0R_TIRS_FRAME_HEADER));

    /* Open access to L0Ra data */
    l0ra_data = ias_l0r_open("test", l0ra_location, IAS_COMPRESSION_ON);
    assert_true(l0ra_data != NULL);

    /* Initialize attributes using satellite attributes library */
    ias_sat_attr_initialize(IAS_L8);
    landsat8_attributes = ias_sat_attr_get_attributes();
    assert_true(landsat8_attributes != NULL);

    /* Open all the bands */
    for (band_number = 1; band_number <= landsat8_attributes->total_bands;
        band_number++)
    {
        status = ias_l0r_open_band(l0ra_data, band_number, IAS_WRITE);
        assert_true(status == SUCCESS);
    }

    /* Open the header file */
    status = ias_l0r_open_header(l0ra_data, IAS_WRITE);
    assert_true(status == SUCCESS);

    /* Open the ancillary file */
    ias_l0r_open_ancillary(l0ra_data, IAS_WRITE);
    status = ias_l0r_set_ancillary_ephemeris(l0ra_data, 0, 1, &ephemeris);
    assert_true(status == SUCCESS);

/* added by bryce */
    IAS_L0R_ATTITUDE attitude;
    memset(&attitude, 0, sizeof(IAS_L0R_ATTITUDE));
    //attitude.time_from_J2000.days_from_J2000 = 0;
    //attitude.time_from_J2000.seconds = 0;
    attitude.l0r_time.days_from_J2000 = 0;
    attitude.l0r_time.seconds_of_day = 0;
    attitude.inertial_to_body.vector.x = 0;
    attitude.inertial_to_body.vector.y = 0;
    attitude.inertial_to_body.vector.z = 0;
    attitude.inertial_to_body.scalar = 0;
    //attitude.angular_vel_rad_sec.x = 0;
    //attitude.angular_vel_rad_sec.y = 0;
    //attitude.angular_vel_rad_sec.z = 0;
    //status = ias_l0r_anc_write_attitude(l0ra_data, 0, 1, &attitude);
    status = ias_l0r_set_ancillary_attitude(l0ra_data, 0, 1, &attitude);
    assert_true(status == SUCCESS);

    //IAS_L0R_IRU iru_data;
    IAS_L0R_IMU iru_data;
    //memset(&iru_data, 0, sizeof(IAS_L0R_IRU));
    memset(&iru_data, 0, sizeof(IAS_L0R_IMU));
    //iru_data.time_from_J2000.days_from_J2000 = 0;
    //iru_data.time_from_J2000.seconds = 0;
    iru_data.l0r_time.days_from_J2000 = 0;
    iru_data.l0r_time.seconds_of_day = 0;
/*
    iru_data.gyro_failure_status_flags = 0;
    for (i = 0; i < IAS_L0R_IRU_NUM_ANGLES; i++)
    {
        iru_data.gyro_angles[i].delta_time_tag_usec = 0;
        iru_data.gyro_angles[i].data_validity = 0;
        for (j = 0; j <  IAS_L0R_NUM_GYROS; j++)
        {
            iru_data.gyro_angles[i].integrated_angle[j] = 0;
        }
    }
*/
    //status = ias_l0r_anc_write_iru(l0ra_data, 0, 1, &iru_data);
    //status = ias_l0r_set_ancillary_iru(l0ra_data, 0, 1, &iru_data);
    status = ias_l0r_set_ancillary_imu(l0ra_data, 0, 1, &iru_data);
    assert_true(status == SUCCESS);
/* added by bryce */
#if 0
#endif

    /* Populate frame header number of lines */
    oli_frame_header.length_of_image = number_of_lines;

    /* Write the frame header to the file */
    status = ias_l0r_set_oli_image_header(l0ra_data, &oli_frame_header);
    assert_true(status == SUCCESS);

    /* Fill every pixel in each band file with data of the following format:
       image = ((line number % cycle_number) * 1000) + 
           detector number within the line
       offset 1 = detector number within the line  
       offset 2 = detector number within the line + 10000 */
    for (band_number = 1; band_number <= landsat8_attributes->total_bands;
        band_number++)
    {
        /* Get the band attributes */
        band_attributes = ias_sat_attr_get_band_attributes(band_number);
        if (band_attributes->band_type == IAS_PAN_BAND)
        {
            num_lines_to_write = 2 * number_of_lines;
        }
        else
        {
            num_lines_to_write = number_of_lines;
        }

        /* Get the number of detectors */
        number_of_detectors = band_attributes->detectors_per_sca *
            band_attributes->scas;

        /* Allocate memory for image data */
        image = (uint16_t*)malloc(number_of_detectors * sizeof(uint16_t));
        assert_true(image != NULL);

        /* Allocate memory for top offsets */
        top_offsets = (uint16_t*)malloc(number_of_detectors * sizeof(uint16_t));
        assert_true(top_offsets != NULL);

        /* Allocate memory for bottom offsets */
        bottom_offsets = (uint16_t*)malloc(number_of_detectors * 
            sizeof(uint16_t));
        assert_true(bottom_offsets != NULL);

        /* Populate the offsets */
        for (i = 0; i < (band_attributes->detectors_per_sca *
            band_attributes->scas); i++)
        {
            top_offsets[i] = i;
            bottom_offsets[i] = i + 10000;
        }
        /* Only write offsets for non-VRP bands */
        if ((band_attributes->band_classification & IAS_VRP_BAND) == 0)
        {
            /* Write top offsets */
            status = ias_l0r_set_top_detector_offsets(l0ra_data, band_number,
                top_offsets);
            assert_true(status == SUCCESS);
            /* Write bottom offsets */
            status = ias_l0r_set_bottom_detector_offsets(l0ra_data, band_number,
                bottom_offsets);
            assert_true(status == SUCCESS);
        }

        /* Populate the image data */
        for (line = 0; line < num_lines_to_write; line++)
        {
            /* Fill the line data */
            for (i = 0; i < (band_attributes->detectors_per_sca *
                band_attributes->scas); i++)
            {
                image[i] = (line % cycle_number * 1000) + i;
            }
            /* Write the line data */
            status = ias_l0r_set_band_lines(l0ra_data, band_number, line,
                line_count, image);
            if (status != SUCCESS)
            {
                failure_flag = 1;
            }

            /* Only create line headers when encounter band 1, or else they
               will be overwritten with each subsequent band */
            if (band_number == 1)
            {
                /* Write the OLI line header*/
                oli_line_header.frame_number = line;
                oli_line_header.l0r_time.days_from_J2000 = 8500;
                oli_line_header.l0r_time.seconds_of_day = line * 10;
                oli_line_header.blind_data_included_in_this_line = 
                    rand()%UINT8_MAX;
                for (i = 0; i < IAS_L0R_FRAME_HEADER_RESERVED_LENGTH; i++)
                {
                    oli_line_header.reserved[i] = rand()%UINT8_MAX;
                }
                oli_line_header.crc = rand()%UINT32_MAX;
                oli_line_header.line_status = 0;
        
                status = ias_l0r_set_oli_frame_headers(l0ra_data, line,
                    1, &oli_line_header);
                if (status != SUCCESS)
                {
                    failure_flag = 1;
                }
            
                /* Write TIRS line header */
                tirs_line_header.sync_byte = rand()%UINT8_MAX;
                tirs_line_header.l0r_time.days_from_J2000 = 8500;
                tirs_line_header.l0r_time.seconds_of_day = line * 10;
                for (i = 0; i < IAS_L0R_TIRS_FRAME_HEADER_RESERVED_SPACE; i++)
                {
                    tirs_line_header.reserved[i] = rand()%UINT8_MAX;
                }
                tirs_line_header.data_set_type = rand()%UINT8_MAX;
                tirs_line_header.integration_duration = rand()%UINT8_MAX;
                for (i = 0; i < sizeof(IAS_L0R_TIRS_FRAME_HEADER_OFFSETS); i++)
                {
                    tirs_line_header.offsets[i] = rand()%UINT8_MAX;
                } 
                status = ias_l0r_set_tirs_frame_headers(l0ra_data, line, 1,
                    &tirs_line_header);
                if (status != SUCCESS)
                {
                    failure_flag = 1;
                }
            }
        }

        /* Cleanup the resources used with line read/write */
        free(image);
        image = NULL;
        free(top_offsets);
        top_offsets = NULL;
        free(bottom_offsets);
        bottom_offsets = NULL;
    }
    if (failure_flag == 1)
    {
        assert_true(1 == 0);
        failure_flag = 0;
    }

    /* Close all the bands */
    for (band_number = 1; band_number <= landsat8_attributes->total_bands;
        band_number++)
    {
        status = ias_l0r_close_band(l0ra_data, band_number);
        assert_true(status == SUCCESS);
    }
    /* Close access to the header */
    status = ias_l0r_close_header(l0ra_data);
    assert_true(status == SUCCESS);

    /* Close access to the ancilliary */
    status = ias_l0r_close_ancillary(l0ra_data);
    assert_true(status == SUCCESS);

    /* Close access to the L0Ra data */
    status = ias_l0r_close(l0ra_data);
    assert_true(status == SUCCESS);
}

/**
 * Generates (and tests the generation of) L0Ra metadata
 *
 * Some of the metadata values generated are hard
 * coded to their expected values, and others are randomly 
 * generated, such as the lat and lon parameters.
 */
static void test_generate_mta()
{
    int i = 0;  /* Looping variable */
    int status = SUCCESS; /* Status variable for checking function calls */
    char temp_file_name[IAS_L0R_FILE_NAME_LENGTH];/* Temp file name for
                                                          L0R files */

    /* Pointer to l0r data files*/
    L0RIO *l0r_data = NULL;

    srand(time(NULL));

    /* Metadata Structures from L0R I/O Library */
    IAS_L0R_FILE_METADATA file_metadata;
    IAS_L0R_INTERVAL_METADATA interval_metadata;    
    IAS_L0R_SCENE_METADATA scene1_metadata;
    IAS_L0R_SCENE_METADATA scene2_metadata;
    IAS_L0R_SCENE_METADATA scene3_metadata;

    /* Set all values in structures to 0 using memset */
    memset(&file_metadata, 0, sizeof(IAS_L0R_FILE_METADATA));
    memset(&interval_metadata, 0, sizeof(IAS_L0R_INTERVAL_METADATA));
    memset(&scene1_metadata, 0, sizeof(IAS_L0R_SCENE_METADATA));
    memset(&scene2_metadata, 0, sizeof(IAS_L0R_SCENE_METADATA));
    memset(&scene3_metadata, 0, sizeof(IAS_L0R_SCENE_METADATA));

    /* For each band file */
    for (i = 0; i < IAS_L0R_MAX_BAND_FILES; i++)
    {
        /* Create name */
        snprintf(temp_file_name, IAS_L0R_FILE_NAME_LENGTH, "test_B%02d",i+1);
        strncpy(file_metadata.band_files[i].name, temp_file_name,
               IAS_L0R_FILE_NAME_LENGTH);
        /* Create file date */
        snprintf(temp_file_name, IAS_L0R_FILE_NAME_LENGTH, 
            "band %02d date", i+1);
        strncpy(file_metadata.band_files[i].date, temp_file_name,
            IAS_L0R_FILEDATE_LENGTH);
        /* File Version */
//        file_metadata.band_files[i].version = 1;
        /* File Size */
        file_metadata.band_files[i].size = 0;
    }
    /* Create ancillary file name, date, version and size */
    strncpy(file_metadata.ancillary_file.name, "test_ANC",
            IAS_L0R_FILE_NAME_LENGTH);
    strncpy(file_metadata.ancillary_file.date, "ANC date",
            IAS_L0R_FILEDATE_LENGTH);
//    file_metadata.ancillary_file.version = 1;
    file_metadata.ancillary_file.size = 0;

    /* Checksum file name */
//    strncpy(file_metadata.checksum_file_name, "CKSUM",
//    strncpy(file_metadata.checksum, "CKSUM",
//            IAS_L0R_FILE_NAME_LENGTH);

    /* Metadata file name, date, version and size */
    strncpy(file_metadata.metadata_file.name, "test_MTA",
        IAS_L0R_FILE_NAME_LENGTH);
    strncpy(file_metadata.metadata_file.date, "MTA date",
        IAS_L0R_FILEDATE_LENGTH);
//    file_metadata.metadata_file.version = 0;

    /* Populate interval metadata fields */
    strncpy(interval_metadata.landsat_interval_id, "LO80010010032014125LGN01",
        IAS_L0R_INTERVAL_ID_LENGTH);
    strncpy(interval_metadata.spacecraft_id, "8", IAS_L0R_SPACECRAFT_ID_LENGTH);
    strncpy(interval_metadata.sensor_id, "OLI", IAS_L0R_SENSOR_ID_LENGTH);
    interval_metadata.satellite = 8;
    strncpy(interval_metadata.data_type, "OLI_L0Ra", IAS_L0R_DATA_TYPE_LENGTH);
    strncpy(interval_metadata.collection_type, "C", 
        IAS_L0R_COLLECTION_TYPE_LENGTH);
    interval_metadata.wrs_starting_path = 1;
    interval_metadata.wrs_starting_row  = 1;
    interval_metadata.wrs_ending_row    = 10;
    strncpy(interval_metadata.ancillary_start_time, "0", IAS_L0R_TIME_LENGTH);
    strncpy(interval_metadata.ancillary_stop_time, "5000", IAS_L0R_TIME_LENGTH);
    strncpy(interval_metadata.start_time_oli, "10", IAS_L0R_TIME_LENGTH);
    strncpy(interval_metadata.stop_time_oli, "4990", IAS_L0R_TIME_LENGTH);
    strncpy(interval_metadata.start_time_tirs, "10", IAS_L0R_TIME_LENGTH);
    strncpy(interval_metadata.stop_time_tirs, "4990", IAS_L0R_TIME_LENGTH);
    strncpy(interval_metadata.cpf_name, "TEST_CPF", IAS_L0R_CPF_NAME_LENGTH);
    interval_metadata.interval_frames_oli    = number_of_lines;
    interval_metadata.interval_frames_tirs    = number_of_lines;
    interval_metadata.corners.upleft.x  = rand()%UINT32_MAX;
    interval_metadata.corners.upleft.y  = rand()%UINT32_MAX;
    interval_metadata.corners.upright.x = rand()%UINT32_MAX;
    interval_metadata.corners.upright.y = rand()%UINT32_MAX;
    interval_metadata.corners.loleft.x  = rand()%UINT32_MAX;
    interval_metadata.corners.loleft.y  = rand()%UINT32_MAX;
    interval_metadata.corners.loright.x = rand()%UINT32_MAX;
    interval_metadata.corners.loright.y = rand()%UINT32_MAX;
    interval_metadata.wrs_type   = 2;
    interval_metadata.wrs_scenes = 3;
    interval_metadata.wrs_scenes_full    = 2;
    interval_metadata.wrs_scenes_partial = 1;
    strncpy(interval_metadata.is_version, "1.0", IAS_SOFTWARE_VERSION_SIZE);
    strncpy(interval_metadata.hostname, "HOSTNAME", IAS_L0R_HOSTNAME_LENGTH);
    strncpy(scene1_metadata.date_acquired, "2009:120:05:21:34.9457395",
        IAS_L0R_DATE_LENGTH);
   
    /* Populate scene metadata fields */
    strncpy(scene1_metadata.landsat_scene_id, "SCENE_ID_1",
        IAS_L0R_SCENE_ID_LENGTH);
/*
    scene1_metadata.target_path = 1;
    scene1_metadata.target_row  = 1;
*/
/* added by bryce */
scene1_metadata.target_wrs_path = 30;
scene1_metadata.target_wrs_row = 30;
/* added by bryce */

    scene1_metadata.roll_angle  = 12;
    strncpy(scene1_metadata.start_time, "0", IAS_L0R_TIME_LENGTH);
    strncpy(scene1_metadata.date_acquired, "2009:120:05:21:34.9457395",
        IAS_L0R_DATE_LENGTH);
    strncpy(scene1_metadata.stop_time, "250", IAS_L0R_TIME_LENGTH);
    scene1_metadata.scene_center.lat  = 50;
    scene1_metadata.scene_center.lng  = 50;
/*
    scene1_metadata.corners.upleft.x  = rand()%UINT16_MAX;
    scene1_metadata.corners.upleft.y  = rand()%UINT16_MAX;
    scene1_metadata.corners.upright.x = rand()%UINT16_MAX;
    scene1_metadata.corners.upright.y = rand()%UINT16_MAX;
    scene1_metadata.corners.loleft.x  = rand()%UINT16_MAX;
    scene1_metadata.corners.loleft.y  = rand()%UINT16_MAX;
    scene1_metadata.corners.loright.x = rand()%UINT16_MAX;
    scene1_metadata.corners.loright.y = rand()%UINT16_MAX;
*/
/* added by bryce */
/*
    scene1_metadata.corners.upleft.x = 11497;
    scene1_metadata.corners.upleft.y = 63892;
    scene1_metadata.corners.upright.x = 24453;
    scene1_metadata.corners.upright.y = 3711;
    scene1_metadata.corners.loleft.x = 3093;
    scene1_metadata.corners.loleft.y = 56163;
    scene1_metadata.corners.loright.x = 54215;
    scene1_metadata.corners.loright.y = 31064;
*/
    scene1_metadata.corners.upleft.y = 43.7;
    scene1_metadata.corners.upleft.x = -96.9;
    scene1_metadata.corners.upright.y = 43.7;
    scene1_metadata.corners.upright.x = -96.6;
    scene1_metadata.corners.loleft.y = 43.4;
    scene1_metadata.corners.loleft.x = -96.9;
    scene1_metadata.corners.loright.y = 43.4;
    scene1_metadata.corners.loright.x = -96.6;
/* added by bryce */

    //scene1_metadata.horizontal_display_shift = 0;
    scene1_metadata.sun_azimuth   = 45;
    scene1_metadata.sun_elevation = 82;
    strncpy(scene1_metadata.full_partial_scene, "FULL", 
        IAS_L0R_SCENE_FULL_PARTIAL_LENGTH);
    strncpy(scene1_metadata.day_night, "DAY", IAS_L0R_DAY_NIGHT_LENGTH);
//    scene1_metadata.scene_center_frame = 5;
//    scene1_metadata.scene_start_frame  = 0;
//    scene1_metadata.scene_stop_frame   = 10; 
    scene1_metadata.time_code_errors  = 0;
    scene1_metadata.missing_frames = 2;

    strncpy(scene2_metadata.landsat_scene_id, "SCENE_ID_2",
        IAS_L0R_SCENE_ID_LENGTH);
    scene2_metadata.target_wrs_path = 1;
    scene2_metadata.target_wrs_row  = 2;
    scene2_metadata.roll_angle  = 12;
    strncpy(scene2_metadata.start_time, "250", IAS_L0R_TIME_LENGTH);
    strncpy(scene2_metadata.date_acquired, "2009:120:05:21:34.9457395",
        IAS_L0R_DATE_LENGTH);
    strncpy(scene2_metadata.stop_time, "500", IAS_L0R_TIME_LENGTH);
    scene2_metadata.scene_center.lat  = 50;
    scene2_metadata.scene_center.lng  = 50;
/*
    scene2_metadata.corners.upleft.x  = rand()%UINT16_MAX;
    scene2_metadata.corners.upleft.y  = rand()%UINT16_MAX;
    scene2_metadata.corners.upright.x = rand()%UINT16_MAX;
    scene2_metadata.corners.upright.y = rand()%UINT16_MAX;
    scene2_metadata.corners.loleft.x  = rand()%UINT16_MAX;
    scene2_metadata.corners.loleft.y  = rand()%UINT16_MAX;
    scene2_metadata.corners.loright.x = rand()%UINT16_MAX;
    scene2_metadata.corners.loright.y = rand()%UINT16_MAX;
*/
/* added by bryce */
/*
    scene2_metadata.corners.upleft.x = 11497;
    scene2_metadata.corners.upleft.y = 63892;
    scene2_metadata.corners.upright.x = 24453;
    scene2_metadata.corners.upright.y = 3711;
    scene2_metadata.corners.loleft.x = 3093; 
    scene2_metadata.corners.loleft.y = 56163;
    scene2_metadata.corners.loright.x = 54215;
    scene2_metadata.corners.loright.y = 31064;
*/
    scene2_metadata.corners.upleft.y = 43.7;
    scene2_metadata.corners.upleft.x = -96.9;
    scene2_metadata.corners.upright.y = 43.7;
    scene2_metadata.corners.upright.x = -96.6;
    scene2_metadata.corners.loleft.y = 43.4;
    scene2_metadata.corners.loleft.x = -96.9;
    scene2_metadata.corners.loright.y = 43.4;
    scene2_metadata.corners.loright.x = -96.6;
/* added by bryce */

    //scene2_metadata.horizontal_display_shift = 0;
    scene2_metadata.sun_azimuth   = 45;
    scene2_metadata.sun_elevation = 82;
    strncpy(scene2_metadata.full_partial_scene, "FULL", 
        IAS_L0R_SCENE_FULL_PARTIAL_LENGTH);
    strncpy(scene2_metadata.day_night, "DAY", IAS_L0R_DAY_NIGHT_LENGTH);
//    scene2_metadata.scene_center_frame = 15;
//    scene2_metadata.scene_start_frame  = 11;
//    scene2_metadata.scene_stop_frame   = 20;
    scene2_metadata.time_code_errors = 0;
    scene2_metadata.missing_frames = 0;

    strncpy(scene3_metadata.landsat_scene_id, "SCENE_ID_3",
        IAS_L0R_SCENE_ID_LENGTH);
    scene3_metadata.target_wrs_path = 1;
    scene3_metadata.target_wrs_row  = 3;
    scene3_metadata.roll_angle  = 12;
    strncpy(scene3_metadata.start_time, "500", IAS_L0R_TIME_LENGTH);
    strncpy(scene3_metadata.date_acquired, "2009:120:05:21:34.9457395",
        IAS_L0R_DATE_LENGTH);
    strncpy(scene3_metadata.stop_time, "750", IAS_L0R_TIME_LENGTH);
    scene3_metadata.scene_center.lat  = 50;
    scene3_metadata.scene_center.lng  = 50;
/*
    scene3_metadata.corners.upleft.x  = rand()%UINT16_MAX;
    scene3_metadata.corners.upleft.y  = rand()%UINT16_MAX;
    scene3_metadata.corners.upright.x = rand()%UINT16_MAX;
    scene3_metadata.corners.upright.y = rand()%UINT16_MAX;
    scene3_metadata.corners.loleft.x  = rand()%UINT16_MAX;
    scene3_metadata.corners.loleft.y  = rand()%UINT16_MAX;
    scene3_metadata.corners.loright.x = rand()%UINT16_MAX;
    scene3_metadata.corners.loright.y = rand()%UINT16_MAX;
*/
/* added by bryce */
/*
    scene3_metadata.corners.upleft.x = 11497;
    scene3_metadata.corners.upleft.y = 63892;
    scene3_metadata.corners.upright.x = 24453;
    scene3_metadata.corners.upright.y = 3711;
    scene3_metadata.corners.loleft.x = 3093;
    scene3_metadata.corners.loleft.y = 56163;
    scene3_metadata.corners.loright.x = 54215;
    scene3_metadata.corners.loright.y = 31064;
*/
    scene3_metadata.corners.upleft.y = 43.7;
    scene3_metadata.corners.upleft.x = -96.9;
    scene3_metadata.corners.upright.y = 43.7;
    scene3_metadata.corners.upright.x = -96.6;
    scene3_metadata.corners.loleft.y = 43.4;
    scene3_metadata.corners.loleft.x = -96.9;
    scene3_metadata.corners.loright.y = 43.4;
    scene3_metadata.corners.loright.x = -96.6;
/* added by bryce */

    //scene3_metadata.horizontal_display_shift = 0;
    scene3_metadata.sun_azimuth = 45;
    scene3_metadata.sun_elevation = 82;
    strncpy(scene3_metadata.full_partial_scene, "PARTIAL",
        IAS_L0R_SCENE_FULL_PARTIAL_LENGTH);
    strncpy(scene3_metadata.day_night, "DAY", IAS_L0R_DAY_NIGHT_LENGTH);
//    scene3_metadata.scene_center_frame = 25;
//    scene3_metadata.scene_start_frame  = 21;
//    scene3_metadata.scene_stop_frame   = 30;
    scene3_metadata.time_code_errors  = 0;
    scene3_metadata.missing_frames = 0;

/* added by bryce */
//strncpy(interval_metadata.date_acquired, "2009-03-01T23:59:59",
strncpy(interval_metadata.date_acquired, "2009:060:23:59:59",
    sizeof(interval_metadata.date_acquired));
scene1_metadata.target_wrs_path = 30;
scene1_metadata.target_wrs_row = 30;
strncpy(scene1_metadata.collection_type, "EARTH",
    sizeof(scene1_metadata.collection_type));
//strncpy(scene1_metadata.date_acquired, "2009-03-01T23:59:59",
//IAS_L0R_DATE_LENGTH);
strncpy(scene1_metadata.date_acquired, "2009:060:23:59:59",IAS_L0R_DATE_LENGTH);
scene2_metadata.target_wrs_path = 30;
scene2_metadata.target_wrs_row = 30;
strncpy(scene2_metadata.collection_type, "EARTH",
    sizeof(scene2_metadata.collection_type));
//strncpy(scene2_metadata.date_acquired, "2009-03-01T23:59:59",
//IAS_L0R_DATE_LENGTH);
strncpy(scene2_metadata.date_acquired, "2009:060:23:59:59",IAS_L0R_DATE_LENGTH);
scene3_metadata.target_wrs_path = 30;
scene3_metadata.target_wrs_row = 30;
strncpy(scene3_metadata.collection_type, "EARTH",
    sizeof(scene3_metadata.collection_type));
//strncpy(scene3_metadata.date_acquired, "2009-03-01T23:59:59",
//IAS_L0R_DATE_LENGTH);
strncpy(scene3_metadata.date_acquired, "2009:060:23:59:59",IAS_L0R_DATE_LENGTH);
/* added by bryce */
    
    /* Open access to L0R data */
    l0r_data = ias_l0r_open("test", l0ra_location, IAS_COMPRESSION_ON);
    assert_true(l0r_data != NULL);

    /* Open metadata file */
    ias_l0r_open_metadata(l0r_data, IAS_WRITE);

    /* Write the file metadata */
    status = ias_l0r_set_file_metadata(l0r_data, &file_metadata);
    assert_true(status == SUCCESS);

    /* Write the interval metadata */
    status = ias_l0r_set_interval_metadata(l0r_data, &interval_metadata);
    assert_true(status == SUCCESS);

    /* Write scene 1 metadata */
    status = ias_l0r_set_scene_metadata(l0r_data, 0, 1, &scene1_metadata);
    assert_true(status == SUCCESS);

    /* Write scene 2 metadata */
    status = ias_l0r_set_scene_metadata(l0r_data, 1, 1, &scene2_metadata);
    assert_true(status == SUCCESS);

    /* Write scene 3 metadata */
    status = ias_l0r_set_scene_metadata(l0r_data, 2, 1, &scene3_metadata);
    assert_true(status == SUCCESS);

    /* Close metadata file */
    status = ias_l0r_close_metadata(l0r_data);
    assert_true(status == SUCCESS);

    /* Close access to the L0R data */
    status = ias_l0r_close(l0r_data);
    assert_true(status == SUCCESS);

    /* Generate checksum file */
    status = generate_checksum_file(l0ra_location, "test");
    assert_true(status == SUCCESS);
}

/**
 * Verifies that the subset_band_image_data subsets the data correctly
 */
static void test_subset_by_line_number()
{
    int band_number = 0;           /* Looping variable for bands */
    int l0ra_offsets_present = 0;  /* Flag indicating if offsets are present
                                      in L0Ra data */
    int l0rp_offsets_present = 0;  /* Flag indicating if offsets are present
                                      in L0Rp data */
    int number_of_detectors = 0;   /* Total number of detectors */
    int start_line = 0;            /* Line to start subsetting data */
    int stop_line  = 0;            /* Line to stop subsetting data */
    int status;                    /* Status variable to check function calls */

    uint16_t *image = NULL;        /* Image buffer */
    uint16_t *top_offsets = NULL;  /* Top offsets buffer */
    uint16_t *bottom_offsets = NULL;/* Bottom offsets buffer */
    uint16_t *image_subset = NULL;  /* Image subset buffer */
    uint16_t *top_offsets_subset = NULL; /* Top offsets subset buffer */
    uint16_t *bottom_offsets_subset = NULL; /* Bottom offsets subset buffer */

    unsigned int size = 0; /* Size of offsets dataset */

    /* Band attributes obtained from satellite attributes library */
    const IAS_BAND_ATTRIBUTES* band_attributes = NULL;
    const IAS_SATELLITE_ATTRIBUTES* landsat8_attributes = NULL;

    /* Initialize satellite attributes */
    status = ias_sat_attr_initialize(IAS_L8);
    landsat8_attributes = ias_sat_attr_get_attributes();
    assert_true(landsat8_attributes != NULL);

    srand(time(NULL));

    /* Calculate a random range of L0Ra lines to subset
       which are near the edges of the L0Ra */
    start_line = cycle_number * (rand() % 3);
    stop_line = number_of_lines - (cycle_number * (rand() % 3));
    /* Use the last line as the stop line, if an invalid value was 
       generated */
    if (stop_line >= number_of_lines)
        stop_line = number_of_lines - 1;

    char dir_name[PATH_MAX];
    snprintf (dir_name, PATH_MAX, "%s/line_subset", l0rp_location);

    /* Initialize access to the L0Ra data */
    L0R_INFORMATION l0ra_data;
    initialize_L0R_information("test", l0ra_location, IAS_READ, &l0ra_data);
    status = initialize_l0r_data(&l0ra_data);
    assert_true(status == SUCCESS);

    /* Initialize access to the L0Ra data */
    L0R_INFORMATION l0rp_data;
    initialize_L0R_information("test", dir_name, IAS_WRITE, &l0rp_data);
    status = initialize_l0r_data(&l0rp_data);
    assert_true(status == SUCCESS);

    /* Copy the L0Ra ancillary data */
    status = copy_ancillary_data(l0ra_data, l0rp_data);
    assert_true(status == SUCCESS);

    /* Subset the L0ra data by calling subset_band_image_data */
    status = subset_band_image_data(l0ra_data.l0r_data, l0rp_data.l0r_data, 
        start_line, stop_line, 0);
    assert_true(status == SUCCESS);

    /* Now check that the L0Ra data was subsetted correctly */
    for (band_number = 1; band_number <= landsat8_attributes->total_bands;
         band_number++)
    {
        /* Get band attributes for current band */ 
        band_attributes = ias_sat_attr_get_band_attributes(band_number);

        /* Get number of detectors */
        number_of_detectors = band_attributes->detectors_per_sca *
            band_attributes->scas;
 
        /* Set offsets present flag to 0 */
        l0ra_offsets_present = 0;
        l0rp_offsets_present = 0;

        /* Allocate memory for image buffer for L0Ra and L0Rp */
        image = (uint16_t*)malloc(number_of_detectors * sizeof(uint16_t));
        assert_true(image != NULL);
        image_subset = (uint16_t*)malloc(number_of_detectors * 
            sizeof(uint16_t));
        assert_true(image_subset != NULL);

        /* Allocate memory for top offsets buffer for L0Ra and L0Rp */
        top_offsets = (uint16_t*)malloc(number_of_detectors * sizeof(uint16_t));
        assert_true(top_offsets != NULL);
        top_offsets_subset = (uint16_t*)malloc(number_of_detectors * 
            sizeof(uint16_t));
        assert_true(top_offsets_subset != NULL);

        /* Allocate memory for bottom offsets buffer for L0Ra and L0Rp */
        bottom_offsets = (uint16_t*)malloc(number_of_detectors * 
            sizeof(uint16_t));
        assert_true(bottom_offsets != NULL);
        bottom_offsets_subset = (uint16_t*)malloc(number_of_detectors * 
            sizeof(uint16_t));
        assert_true(bottom_offsets_subset != NULL);

        /* Open each individual L0Ra band file */
        status = ias_l0r_open_band(l0ra_data.l0r_data, band_number, IAS_READ);
        assert_true(status == SUCCESS);

        /* Open each individual L0Rp band file */
        status = ias_l0r_open_band(l0rp_data.l0r_data, band_number, IAS_READ);
        assert_true(status == SUCCESS);

        /* Verify the first line of the L0rp */
        /* Get first line of L0Rp data */
        status = ias_l0r_get_band_lines(l0rp_data.l0r_data,band_number, 0, 1,
           number_of_detectors, image_subset);
        assert_true(status == SUCCESS);
        /* Get first line of L0Ra data */
        status = ias_l0r_get_band_lines(l0ra_data.l0r_data,band_number, 
            start_line, 1, number_of_detectors, image);
        assert_true(status == SUCCESS);

        /* Compare L0Ra and L0Rp data */
        assert_true(memcmp(image, image_subset, sizeof(image)) == 0);

        /* verify the last line of the L0Rp */
        /* Get last line of L0Rp data */
        status = ias_l0r_get_band_lines(l0rp_data.l0r_data,band_number,
           stop_line-start_line, 1, number_of_detectors, image_subset);
        assert_true(status == SUCCESS);
        /* Get last line of L0Ra data */
        status = ias_l0r_get_band_lines(l0ra_data.l0r_data,band_number, 
           stop_line, 1, number_of_detectors, image);
        assert_true(status == SUCCESS);

        /* Compare L0Ra and L0Rp data */
        assert_true(memcmp(image, image_subset, sizeof(image)) == 0);

        /* Free memory used */
        free(image);
        image = NULL;
        free(image_subset);
        image_subset = NULL;

        /* Check the offsets */
        /* First see if the offsets dataset exists in the L0Ra data */
        status = ias_l0r_get_band_offset_records_count(l0ra_data.l0r_data,
                band_number, &size);
        assert_true(status == SUCCESS);
        /* If size is > 0, then the datasets offset exists in L0Ra data */
        if (size > 0)
        {
            /* Set flag that the offsets exist */
            l0ra_offsets_present = 1;
          
            /* Get the top offsets from the L0Ra data */
            status = ias_l0r_get_top_detector_offsets(l0ra_data.l0r_data,
                band_number, top_offsets, number_of_detectors);
            assert_true(status == SUCCESS);

            /* Get the bottom offsets from the L0Ra data */
            status = ias_l0r_get_bottom_detector_offsets(l0ra_data.l0r_data,
                band_number, bottom_offsets, number_of_detectors);
            assert_true(status == SUCCESS);
        }
        /* Now see if the offsets dataset exists in the L0Rp data */
        status = ias_l0r_get_band_offset_records_count(l0rp_data.l0r_data, 
            band_number, &size);
        assert_true(status == SUCCESS);
        /* If size is > 0, then the datasets offset exists in L0Rp data */
        if (size > 0)
        {
            /* Set flag that the offsets exist */
            l0rp_offsets_present = 1;

            /* Get the top offsets from the L0Ra data */
            status = ias_l0r_get_top_detector_offsets(l0rp_data.l0r_data,
                band_number, top_offsets_subset, number_of_detectors);
            assert_true(status == SUCCESS);

            /* Get the bottom offsets from the L0Ra data */
            status = ias_l0r_get_bottom_detector_offsets(l0rp_data.l0r_data,
                band_number, bottom_offsets_subset, number_of_detectors);
            assert_true(status == SUCCESS);
        }
        
        /* If offsets exist for both L0Ra and L0Rp, make sure they were
           they are identical */
        if (l0ra_offsets_present && l0rp_offsets_present)
        {
            /* Check the top offsets */
            assert_true(memcmp(top_offsets, top_offsets_subset, 
                sizeof(top_offsets)) == 0);

            /* Check the bottom offsets */
            assert_true(memcmp(bottom_offsets, bottom_offsets_subset, 
                sizeof(bottom_offsets)) == 0);
        }
        else
            assert_true(!l0ra_offsets_present && !l0rp_offsets_present);


        /* Close the indivudal L0Ra band files */
        status = ias_l0r_close_band(l0ra_data.l0r_data, band_number);
        assert_true(status == SUCCESS);

        /* Close the indivudal L0Rp band files */
        status = ias_l0r_close_band(l0rp_data.l0r_data, band_number);
        assert_true(status == SUCCESS);

        /* Clean up memory used */
        free(top_offsets);
        top_offsets = NULL;
        free(top_offsets_subset);
        top_offsets_subset = NULL;
        free(bottom_offsets);
        bottom_offsets = NULL;
        free(bottom_offsets_subset);
        bottom_offsets_subset = NULL;
    }
    /* Terminate access to L0Ra data */
    status = terminate_l0r_access(&l0ra_data);
    assert_true(status == SUCCESS);

    /* Terminate access to L0Rp data */
    status = terminate_l0r_access(&l0rp_data);
    assert_true(status == SUCCESS);
}

/**
 * Tests the subset by scene id option
 */
static void test_subset_by_scene_id()
{
    int band_number = 0;  /* Looping variable for bands */
    int line_num_to_verify = 0; /* Line number to verify, this will be different
                                   for the pan band, as it is so much bigger 
                                   than the other bands */
    int start_line = 0;  /* Start line for verifications for pan band */

    int status = SUCCESS; /* Status variable for checking function calls */

    uint16_t *image = NULL; /* Image buffer */
    uint16_t *image_subset = NULL; /* Image subset buffer */

    /* Band attributes obtained from satellite attributes library */
    const IAS_BAND_ATTRIBUTES* band_attributes = NULL;
    const IAS_SATELLITE_ATTRIBUTES* landsat8_attributes = NULL;
    /* Initialize satellite attributes */
    status = ias_sat_attr_initialize(IAS_L8);
    landsat8_attributes = ias_sat_attr_get_attributes();
    assert_true(landsat8_attributes != NULL);

    /* Directory names for each scene */
    char scene_1_dir_name[PATH_MAX];
    char scene_2_dir_name[PATH_MAX];
    char scene_3_dir_name[PATH_MAX];

    srand(time(NULL));

    /* Subset each of the scenes into it's own directory */

    /* Scene 1 */
    snprintf(scene_1_dir_name, PATH_MAX, "%s/scene_1", l0rp_location);
    /* Call subset_by_scene for scene 1 */
    status = subset_by_scene("test", "SCENE_ID_1", l0ra_location, 
        scene_1_dir_name, 0);
    assert_true(status == SUCCESS);

    /* Scene 2 */
    snprintf(scene_2_dir_name, PATH_MAX, "%s/scene_2", l0rp_location);
    /* Call subset_by_scene for scene 2 */
    status = subset_by_scene("test", "SCENE_ID_2", l0ra_location, 
        scene_2_dir_name, 0);
    assert_true(status == SUCCESS);

    /* Scene 3 */
    snprintf(scene_3_dir_name, PATH_MAX, "%s/scene_3", l0rp_location);
    /* Call subset_by_scene for scene 3 */
    status = subset_by_scene("test", "SCENE_ID_3", l0ra_location, 
        scene_3_dir_name, 0);
    assert_true(status == SUCCESS);

    /* Initialize access to the L0Ra data */
    L0R_INFORMATION l0ra_data;
    initialize_L0R_information("test", l0ra_location, IAS_READ, &l0ra_data);
    status = initialize_l0r_data(&l0ra_data);
    assert_true(status == SUCCESS);

    /* Initialize access to each scene's L0Rp data */
    L0R_INFORMATION scene_1_data;
    L0R_INFORMATION scene_2_data;
    L0R_INFORMATION scene_3_data;
    initialize_L0R_information("SCENE_ID_1", scene_1_dir_name, IAS_WRITE, 
        &scene_1_data);
    initialize_L0R_information("SCENE_ID_2", scene_2_dir_name, IAS_WRITE, 
        &scene_2_data);
    initialize_L0R_information("SCENE_ID_3", scene_3_dir_name, IAS_WRITE, 
        &scene_3_data);
    status = initialize_l0r_data(&scene_1_data);
    assert_true(status == SUCCESS);
    status = initialize_l0r_data(&scene_2_data);
    assert_true(status == SUCCESS);
    status = initialize_l0r_data(&scene_3_data);
    assert_true(status == SUCCESS);

    for (band_number = 1; band_number <= landsat8_attributes->total_bands;
        band_number++)
    {
        /* Get current band's attributes */
        band_attributes = ias_sat_attr_get_band_attributes(band_number);

        /* Allocate memory for image data.  This needs to be done inside the
           loop because some bands have different detectors_per_sca etc */
        image = (uint16_t*)malloc(band_attributes->detectors_per_sca *
            band_attributes->scas * sizeof(uint16_t));
        assert_true(image != NULL);
        image_subset = (uint16_t*)malloc(band_attributes->detectors_per_sca *
            band_attributes->scas * sizeof(uint16_t));
        assert_true(image_subset != NULL);

        /* Open band file for both L0Ra data and L0Rp data */
        status = ias_l0r_open_band(l0ra_data.l0r_data, band_number, IAS_READ);
        assert_true(status == SUCCESS);

        /* Scene 1 */
        status = ias_l0r_open_band(scene_1_data.l0r_data, band_number,
            IAS_READ);
        assert_true(status == SUCCESS);
        /* Scene 2 */
        status = ias_l0r_open_band(scene_2_data.l0r_data, band_number,
            IAS_READ);
        assert_true(status == SUCCESS);
        /* Scene 3 */
        status = ias_l0r_open_band(scene_3_data.l0r_data, band_number,
            IAS_READ);
        assert_true(status == SUCCESS);
        
        /* Verify the scenes are subsetted correctly
           The values in the ias_l0r_get_band_lines function
           are hardcoded to match the values used to populate
           the MTA file */ 
  
        /* Verify Scene 1, first line in dataset */
        line_num_to_verify = 0;
        status = ias_l0r_get_band_lines(scene_1_data.l0r_data, band_number, 
             0, 1, band_attributes->detectors_per_sca * band_attributes->scas,
             image_subset);
        assert_true(status == SUCCESS);
        status = ias_l0r_get_band_lines(l0ra_data.l0r_data, band_number,
            0, 1, band_attributes->detectors_per_sca * band_attributes->scas, 
            image);
        assert_true(status == SUCCESS);
        assert_true(memcmp(image, image_subset, sizeof(image)) == 0);

        /* Verify Scene 1, last line in dataset */
        line_num_to_verify = 10;
        if (band_attributes->band_type == IAS_PAN_BAND)
        {
            /* Modify line number to verify if we are on the pan band,
               since it is twice as big as the other bands */
            line_num_to_verify = 21;
        }
        status = ias_l0r_get_band_lines(scene_1_data.l0r_data, band_number, 
            10, 1, band_attributes->detectors_per_sca * band_attributes->scas, 
            image_subset);
        assert_true(status == SUCCESS);
        status = ias_l0r_get_band_lines(l0ra_data.l0r_data, band_number,
            10, 1, band_attributes->detectors_per_sca * band_attributes->scas, 
            image);
        assert_true(status == SUCCESS);
        assert_true(memcmp(image, image_subset, sizeof(image)) == 0);

        /* Verify Scene 2 */
        line_num_to_verify = 11;
        if (band_attributes->band_type == IAS_PAN_BAND)
        {
            /* Modify line number to verify if we are on the pan band,
               since it is twice as big as the other bands */
            line_num_to_verify = 22;
        }
        status = ias_l0r_get_band_lines(scene_2_data.l0r_data, band_number, 0,
            1, band_attributes->detectors_per_sca * band_attributes->scas, 
            image_subset);
        assert_true(status == SUCCESS);
        status = ias_l0r_get_band_lines(l0ra_data.l0r_data, band_number,
            line_num_to_verify,
            1, band_attributes->detectors_per_sca * band_attributes->scas, 
            image);
        assert_true(status == SUCCESS);
        assert_true(memcmp(image, image_subset, sizeof(image)) == 0);

        line_num_to_verify = 20;
        start_line = 9;
        if (band_attributes->band_type == IAS_PAN_BAND)
        {
            /* Modify line number to verify if we are on the pan band,
               since it is twice as big as the other bands.  */
            line_num_to_verify = 41;
            start_line = 19;
        }
        status = ias_l0r_get_band_lines(scene_2_data.l0r_data, band_number, 
            start_line, 1, 
            band_attributes->detectors_per_sca * band_attributes->scas, 
            image_subset);
        assert_true(status == SUCCESS);
        status = ias_l0r_get_band_lines(l0ra_data.l0r_data, band_number,
            line_num_to_verify, 
            1, band_attributes->detectors_per_sca * band_attributes->scas, 
            image);
        assert_true(status == SUCCESS);
        assert_true(memcmp(image, image_subset, sizeof(image)) == 0);

        /* Verify Scene 3 */
        line_num_to_verify = 21;
        if (band_attributes->band_type == IAS_PAN_BAND)
        {
            /* Modify line number to verify if we are on the pan band,
               since it is twice as big as the other bands */
            line_num_to_verify = 42;
        }
        status = ias_l0r_get_band_lines(scene_3_data.l0r_data, band_number, 0, 
            1, band_attributes->detectors_per_sca * band_attributes->scas, 
            image_subset);
        assert_true(status == SUCCESS);
        status = ias_l0r_get_band_lines(l0ra_data.l0r_data, band_number,
            line_num_to_verify, 1, 
            band_attributes->detectors_per_sca * band_attributes->scas, 
            image);
        assert_true(status == SUCCESS);
        assert_true(memcmp(image, image_subset, sizeof(image)) == 0);

        line_num_to_verify = 30;
        start_line = 9;
        if (band_attributes->band_type == IAS_PAN_BAND)
        {
            /* Modify line number to verify if we are on the pan band,
               since it is twice as big as the other bands */
            
            line_num_to_verify = 61;
            start_line = 19;
        }
        status = ias_l0r_get_band_lines(scene_3_data.l0r_data, band_number, 
            start_line, 1, 
            band_attributes->detectors_per_sca * band_attributes->scas, 
            image_subset);
        assert_true(status == SUCCESS);
        status = ias_l0r_get_band_lines(l0ra_data.l0r_data, band_number,
            line_num_to_verify, 1,
            band_attributes->detectors_per_sca * band_attributes->scas, 
            image);
        assert_true(status == SUCCESS);
        assert_true(memcmp(image, image_subset, sizeof(image)) == 0);

        /* Free memory used */
        free(image);
        image = NULL;
        free(image_subset);
        image_subset = NULL;

        /* Close the indivudal band files */
        status = ias_l0r_close_band(l0ra_data.l0r_data, band_number);
        assert_true(status == SUCCESS);

        status = ias_l0r_close_band(scene_1_data.l0r_data, band_number);
        assert_true(status == SUCCESS);

        status = ias_l0r_close_band(scene_2_data.l0r_data, band_number);
        assert_true(status == SUCCESS);

        status = ias_l0r_close_band(scene_3_data.l0r_data, band_number);
        assert_true(status == SUCCESS);
    }
    /* Terminate access to all the data */
    status = terminate_l0r_access(&l0ra_data);
    status = terminate_l0r_access(&scene_1_data);
    status = terminate_l0r_access(&scene_2_data);
    status = terminate_l0r_access(&scene_3_data);
}

/**
 * Tests that the tirs redundant bands get dropped when specified
 */
static void test_subset_by_scene_id_drop_tirs_red()
{
    int band_number = 0;  /* Looping variable for bands */
    int status = SUCCESS; /* Status variable for checking function calls */

    uint16_t *image = NULL; /* Image buffer */
    uint16_t *image_subset = NULL; /* Image subset buffer */

    /* Band attributes obtained from satellite attributes library */
    const IAS_BAND_ATTRIBUTES* band_attributes = NULL;
    const IAS_SATELLITE_ATTRIBUTES* landsat8_attributes = NULL;
    /* Initialize satellite attributes */
    status = ias_sat_attr_initialize(IAS_L8);
    landsat8_attributes = ias_sat_attr_get_attributes();
    assert_true(landsat8_attributes != NULL);

    char scene_1_dir_name[PATH_MAX];

    srand(time(NULL));

    /* Subset into it's own directory */
    snprintf(scene_1_dir_name, PATH_MAX, 
        "%s/scene_1_no_tirs_red", l0rp_location);
    subset_by_scene("test", "SCENE_ID_1", l0ra_location, scene_1_dir_name,
        1);

    /* Initialize access to the L0Ra data */
    L0R_INFORMATION l0ra_data;
    initialize_L0R_information("test", l0ra_location, IAS_READ, &l0ra_data);
    status = initialize_l0r_data(&l0ra_data);
    assert_true(status == SUCCESS);

    /* Initialize access to the L0Rp data */
    L0R_INFORMATION scene_1_data;
    initialize_L0R_information("SCENE_ID_1", scene_1_dir_name, IAS_WRITE, 
        &scene_1_data);
    status = initialize_l0r_data(&scene_1_data);
    assert_true(status == SUCCESS);

    for (band_number = 1; band_number <= landsat8_attributes->total_bands;
         band_number++)
    {
        /* Get current band's attributes */
        band_attributes = ias_sat_attr_get_band_attributes(band_number);

        /* Skip TIRS redundant bands, since they were dropped */
        if ((band_attributes->band_classification
             & IAS_SECONDARY_BAND) != 0)
        {
            continue;
        }

        /* Allocate memory for image data.  This needs to be done inside the
           loop because some bands have different detectors_per_sca etc */
        image = (uint16_t*)malloc(band_attributes->detectors_per_sca *
            band_attributes->scas * sizeof(uint16_t));
        assert_true(image != NULL);
        image_subset = (uint16_t*)malloc(band_attributes->detectors_per_sca *
            band_attributes->scas * sizeof(uint16_t));
        assert_true(image_subset != NULL);

        /* Open the L0Ra band file */
        status = ias_l0r_open_band(l0ra_data.l0r_data, band_number, IAS_READ);
        assert_true(status == SUCCESS);

        /* Open the L0Rp band file */
        status = ias_l0r_open_band(scene_1_data.l0r_data, band_number,
            IAS_READ);
        assert_true(status == SUCCESS);

        /* Verify the scenes are subsetted correctly
           The values in the ias_l0r_get_band_lines function
           are hardcoded to match the values used to populate
           the MTA file */ 
  
        /* Verify Scene 1 */
        status = ias_l0r_get_band_lines(scene_1_data.l0r_data,band_number, 0, 1,
            band_attributes->detectors_per_sca * band_attributes->scas, 
            image_subset);
        assert_true(status == SUCCESS);
        status = ias_l0r_get_band_lines(l0ra_data.l0r_data,band_number,
            0, 1, band_attributes->detectors_per_sca * band_attributes->scas, 
            image);
        assert_true(status == SUCCESS);
        assert_true(memcmp(image, image_subset, sizeof(image)) == 0);

        status = ias_l0r_get_band_lines(scene_1_data.l0r_data,band_number, 10,
            1, band_attributes->detectors_per_sca * band_attributes->scas, 
            image_subset);
        assert_true(status == SUCCESS);
        status = ias_l0r_get_band_lines(l0ra_data.l0r_data,band_number,
            10, 1, band_attributes->detectors_per_sca * band_attributes->scas, 
            image);
        assert_true(status == SUCCESS);
        assert_true(memcmp(image, image_subset, sizeof(image)) == 0);

        /* Cleanup memory used */
        free(image);
        image = NULL;
        free(image_subset);
        image_subset = NULL;

        /* Close the indivudal L0Ra band files */
        status = ias_l0r_close_band(l0ra_data.l0r_data, band_number);
        assert_true(status == SUCCESS);

        /* Close the indivudal L0Rp band files */
        status = ias_l0r_close_band(scene_1_data.l0r_data, band_number);
        assert_true(status == SUCCESS);
    }
    /* Terminate access to the L0R data */
    status = terminate_l0r_access(&l0ra_data);
    status = terminate_l0r_access(&scene_1_data);
}

/**
 * Cleans up the L0Ra and L0Rp directories if they exist
 */
void cleanup()
{
    char command[PATH_MAX + 256];
    int status = 0;  /* status variable for checking function calls */

    /* Remove the L0Ra location */
    status = snprintf(command, PATH_MAX+256, "rm -rf %s", l0ra_location);
    if (status > 0)
    {
        status = system(command);
        if (status == -1 || WEXITSTATUS(status) < 0)
        {
            IAS_LOG_ERROR("Error encountered when cleaning up output "
                "directories.");
        }
    }

    /* Remove the L0Rp location */
    status = snprintf(command, PATH_MAX+256, "rm -rf %s", l0rp_location);
    if (status > 0)
    {
        status = system(command);
        if (status == -1 || WEXITSTATUS(status) < 0)
        {
            IAS_LOG_ERROR("Error encountered when cleaning up output "
                "directories.");
        }
    }
}

/**
 * creates the L0Ra and L0Rp test directories if needed
 */
int create_test_environment()
{
    struct stat stat_p;  /* Used to stat the l0ra and l0rp directories */
    char command[PATH_MAX * 2];
    int status = 0;  /* status variable for checking function calls */

    /* Check to see if the l0ra location already exists, if so, clean it up */
    if (stat(l0ra_location, &stat_p) == 0)
    {
        cleanup();
    }
    /* Now create the l0ra location */
    snprintf (command, PATH_MAX*2, "mkdir %s\n", l0ra_location); 
    status = system (command);
    if (status == -1 || WEXITSTATUS(status) < 0)
    {
        IAS_LOG_ERROR("Error encountered when running command %s ",
            command);
        return ERROR;
    }

    /* Check to see if the l0rp location already exists, if not, create it and
       all subdirectories needed to run the unit tests */
    if (stat(l0rp_location, &stat_p) == -1)
    {
        snprintf (command, PATH_MAX*2, "mkdir %s", l0rp_location); 
        status = system (command);
        if (status == -1 || WEXITSTATUS(status) < 0)
        {
            IAS_LOG_ERROR("Error encountered when running command %s ",
                command);
            return ERROR;
        }

        char interval_request_dir[PATH_MAX];
        snprintf (interval_request_dir, PATH_MAX,
            "%s/interval_request", l0rp_location);
        snprintf (command, PATH_MAX*2, "mkdir %s", interval_request_dir); 
        status = system (command);
        if (status == -1 || WEXITSTATUS(status) < 0)
        {
            IAS_LOG_ERROR("Error encountered when running command %s ",
                command);
            return ERROR;
        }

        char line_subset_dir[PATH_MAX];
        snprintf (line_subset_dir, PATH_MAX, "%s/line_subset", l0rp_location);
        snprintf (command, PATH_MAX*2, "mkdir %s", line_subset_dir); 
        status = system (command);
        if (status == -1 || WEXITSTATUS(status) < 0)
        {
            IAS_LOG_ERROR("Error encountered when running command %s ",
                command);
            return ERROR;
        }

        char scene_1_dir[PATH_MAX];
        snprintf (scene_1_dir, PATH_MAX, "%s/scene_1", l0rp_location);
        snprintf (command, PATH_MAX*2, "mkdir %s", scene_1_dir); 
        status = system (command);
        if (status == -1 || WEXITSTATUS(status) < 0)
        {
            IAS_LOG_ERROR("Error encountered when running command %s ",
                command);
            return ERROR;
        }

        char scene_2_dir[PATH_MAX];
        snprintf (scene_2_dir, PATH_MAX, "%s/scene_2", l0rp_location);
        snprintf (command, PATH_MAX*2, "mkdir %s", scene_2_dir); 
        status = system (command);
        if (status == -1 || WEXITSTATUS(status) < 0)
        {
            IAS_LOG_ERROR("Error encountered when running command %s ",
                command);
            return ERROR;
        }

        char scene_3_dir[PATH_MAX];
        snprintf (scene_3_dir, PATH_MAX, "%s/scene_3", l0rp_location);
        snprintf (command, PATH_MAX*2, "mkdir %s", scene_3_dir); 
        status = system (command);
        if (status == -1 || WEXITSTATUS(status) < 0)
        {
            IAS_LOG_ERROR("Error encountered when running command %s ",
                command);
            return ERROR;
        }

        char scene_no_tirs_dir[PATH_MAX];
        snprintf (scene_no_tirs_dir, PATH_MAX,
            "%s/scene_1_no_tirs_red", l0rp_location);
        snprintf (command, PATH_MAX*2, "mkdir %s", scene_no_tirs_dir); 
        status = system (command);
        if (status == -1 || WEXITSTATUS(status) < 0)
        {
            IAS_LOG_ERROR("Error encountered when running command %s ",
                command);
            return ERROR;
        }

        char time_subset_dir[PATH_MAX];
        snprintf (time_subset_dir, PATH_MAX, "%s/time_subset", l0rp_location);
        snprintf (command, PATH_MAX*2, "mkdir %s", time_subset_dir); 
        status = system (command);
        if (status == -1 || WEXITSTATUS(status) < 0)
        {
            IAS_LOG_ERROR("Error encountered when running command %s ",
                command);
            return ERROR;
        }
    }
    return SUCCESS;
}

/**
 * Calls each individual test routine to perform unit tests
 */
int main(int argc, char **argv)
{
    char *username;       /* $USER environment variable */
    int return_value;     /* Return value from running unit tests */
    int status = SUCCESS; /* Status variable for system calls */


    /* Get user name environment variable */
    username = getenv("USER");
    if (username == NULL)
    {
        IAS_LOG_ERROR("Error initializing username\n");
        return ERROR;
    }
#if 0
    /* Hard code the l0ra location on data1 */
    strncpy(l0ra_location, "/edclxs113/data1/ingest/", PATH_MAX);
    strncat(l0ra_location, username, PATH_MAX);
    strncat(l0ra_location, "/l0ra_location/", PATH_MAX);

    /* Hard code the l0rp location on data1 */
    strncpy(l0rp_location, "/edclxs113/data1/ingest/", PATH_MAX);
    strncat(l0rp_location, username, PATH_MAX);
    strncat(l0rp_location, "/l0rp_location/", PATH_MAX);
#endif
    /* Hard code the l0ra location on data1 */
    strncpy(l0ra_location, "/edclxs110/data1/", PATH_MAX);
    strncat(l0ra_location, username, PATH_MAX);
    strncat(l0ra_location, "/l0ra_location/", PATH_MAX);

    /* Hard code the l0rp location on data1 */
    strncpy(l0rp_location, "/edclxs110/data1/", PATH_MAX);
    strncat(l0rp_location, username, PATH_MAX);
    strncat(l0rp_location, "/l0rp_location/", PATH_MAX);

    /* Initialize the ias logging library */
    status = ias_log_initialize("Subsetter Test");
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Error encountered when initializing the ias logging "
            "library");
        return ERROR;
    }
    ias_log_set_output_level(IAS_LOG_LEVEL_DEBUG);

    /* Call routine to create the test directories needed for the unit tests */
    status = create_test_environment();
    /* If an error occurred when creating the test environment, no reason to
       continue processing */
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Error encountered when creating test "
            "directories");
        return ERROR;
    }

    TestSuite *suite = create_test_suite();
    TestReporter *reporter = create_text_reporter();

    /* Add each unit test */
    add_test(suite, test_generate_l0ra);
    add_test(suite, test_generate_mta);
#if 0
    add_test(suite, test_subset_by_line_number);
    add_test(suite, test_subset_by_scene_id);
    add_test(suite, test_subset_by_scene_id_drop_tirs_red);
    add_test(suite, test_subsetter_get_command_line_parameters);
#endif

    /* Run the unit tests */
    return_value = run_test_suite(suite, reporter);

    /* Clean the output directories if all the tests passed */
    if (reporter->failures == 0)
    {
        IAS_LOG_DEBUG("No failures reported, calling cleanup to remove the "
            "L0Ra and L0Rp  directories\n");
        cleanup();
    }
    return return_value;
}
