#include <stdio.h>
#include <string.h>

#include "cgreen.h"

/* IAS includes */
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_l0r.h"
#include "gctp.h"
#include "setup_work_order.h"


/* Test 4 extend_corners scenarios based on CalVal test data and results:
   - OLI/TIRS with roll angle of 0
   - OLI/TIRS with positive roll angle
   - OLI/TIRS with negative roll angle
   - TIRS only with roll angle of 0 */
Ensure extend_corners_test()
{
    int status;
    DSW_DATA dsw_data;
    IAS_CPF *cpf = NULL;
    const struct IAS_CPF_EARTH_CONSTANTS *earth_consts = NULL;
    const struct IAS_CPF_ORBIT_PARAMETERS *orbit_parms = NULL;
    const struct IAS_CPF_FOCAL_PLANE *focal_plane = NULL;

    /* Make sure the dsw_data structure is initialized */
    memset(&dsw_data, 0, sizeof(dsw_data));

    status = ias_sat_attr_initialize(IAS_L8);
    assert_equal_with_message(status, SUCCESS,
        "Could not initialize satellite attributes library");

    /* Set up the CPF filename */
    strncpy(dsw_data.cpf_filename,
        "./dataservices/rest/cpf?operation=getCalParmFile&filename="
        "LC08CPF_20090101_20090331_01.03", PATH_MAX);

    /* Read the CPF, and get the parameters required for extending the
       corners */
    cpf = ias_cpf_read(dsw_data.cpf_filename);
    if (cpf == NULL)
    {
        assert_true_with_message(FALSE, "Error reading CPF");
        return;
    }
    earth_consts = ias_cpf_get_earth_const(cpf);
    orbit_parms = ias_cpf_get_orbit(cpf);
    focal_plane = ias_cpf_get_focal_plane(cpf);
    assert_not_equal_with_message(earth_consts, NULL, "Error reading "
        "IAS_CPF_EARTH_CONSTANTS group");
    assert_not_equal_with_message(orbit_parms, NULL, "Error reading "
        "IAS_CPF_ORBIT_PARAMETERS group");
    assert_not_equal_with_message(focal_plane, NULL, "Error reading "
        "IAS_CPF_FOCAL_PLANE group");
    if (earth_consts == NULL || orbit_parms == NULL || focal_plane == NULL)
    {
        ias_cpf_free(cpf);
        return;
    }

    /* Test OLI and TIRS missing */
    dsw_data.sensor_present_oli = 0;
    dsw_data.sensor_present_tirs = 0;

    /* Temporarily disable error logging */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    status = extend_corners(&dsw_data, earth_consts, orbit_parms, focal_plane);

    /* Reenable error logging */
    ias_log_set_output_level(IAS_LOG_LEVEL_ERROR);

    assert_equal_with_message(status, ERROR, "Missing OLI and TIRS did not "
        "return ERROR");

    /* Test OLI/TIRS with 0 roll angle */

    /* setup the corners */
    dsw_data.corners_oli.upleft.lat = 35.492777354424717;
    dsw_data.corners_oli.upleft.lng = -116.08837060593612;
    dsw_data.corners_oli.upright.lat = 35.102379226736716;
    dsw_data.corners_oli.upright.lng = -114.04332071271998;
    dsw_data.corners_oli.loleft.lat = 34.135432249255857;
    dsw_data.corners_oli.loleft.lng = -116.45451433069125;
    dsw_data.corners_oli.loright.lat = 33.744570149942078;
    dsw_data.corners_oli.loright.lng = -114.44290120476886;
    dsw_data.corners_tirs.upleft.lat = 35.492777354424717;
    dsw_data.corners_tirs.upleft.lng = -116.08837060593612;
    dsw_data.corners_tirs.upright.lat = 35.102379226736716;
    dsw_data.corners_tirs.upright.lng = -114.04332071271998;
    dsw_data.corners_tirs.loleft.lat = 34.135432249255857;
    dsw_data.corners_tirs.loleft.lng = -116.45451433069125;
    dsw_data.corners_tirs.loright.lat = 33.744570149942078;
    dsw_data.corners_tirs.loright.lng = -114.44290120476886;

    /* setup the sensor present flags, scene center and off-nadir angle */
    dsw_data.sensor_present_oli = 1;
    dsw_data.sensor_present_tirs = 1;
    dsw_data.scene_center.lat = 34.618789745089842;
    dsw_data.scene_center.lng = -115.25727671352905;
    dsw_data.off_nadir_angle = 0.000000;

    status = extend_corners(&dsw_data, earth_consts, orbit_parms, focal_plane);
    
    assert_equal_with_message(status, SUCCESS, "OLI/TIRS 0 roll angle test "
        "did not return SUCCESS");
    assert_double_equal_with_message(35.675849703187062,
        dsw_data.corners_oli.upleft.lat, "UL.lat for OLI did not match "
        "expected results in OLI/TIRS 0 roll angle check");
    assert_double_equal_with_message(-116.03802360746211,
        dsw_data.corners_oli.upleft.lng, "UL.lng for OLI did not match "
        "expected results in OLI/TIRS 0 roll angle check");
    assert_double_equal_with_message(35.284587945314833,
        dsw_data.corners_oli.upright.lat, "UR.lat for OLI did not match "
        "expected results in OLI/TIRS 0 roll angle check"); 
    assert_double_equal_with_message(-113.98866427434382,
        dsw_data.corners_oli.upright.lng, "UR.lng for OLI did not match "
        "expected results in OLI/TIRS 0 roll angle check"); 
    assert_double_equal_with_message(33.95213000046509,
        dsw_data.corners_oli.loleft.lat, "LL.lat for OLI did not match "
        "expected results in OLI/TIRS 0 roll angle check");
    assert_double_equal_with_message(-116.50302963622276,
        dsw_data.corners_oli.loleft.lng, "LL.lng for OLI did not match "
        "expected results in OLI/TIRS 0 roll angle check"); 
    assert_double_equal_with_message(33.562102205755551,
        dsw_data.corners_oli.loright.lat, "LR.lat for OLI did not match "
        "expected results in OLI/TIRS 0 roll angle check"); 
    assert_double_equal_with_message(-114.49559807903398,
        dsw_data.corners_oli.loright.lng, "LR.lng for OLI did not match "
        "expected results in OLI/TIRS 0 roll angle check"); 
    assert_double_equal_with_message(36.039084417509351,
        dsw_data.corners_tirs.upleft.lat, "UL.lat for TIRS did not match "
        "expected results in OLI/TIRS 0 roll angle check");
    assert_double_equal_with_message(-115.93742459527616,
        dsw_data.corners_tirs.upleft.lng, "UL.lng for TIRS did not match "
        "expected results in OLI/TIRS 0 roll angle check");
    assert_double_equal_with_message(35.646098222280912,
        dsw_data.corners_tirs.upright.lat, "UR.lat for TIRS did not match "
        "expected results in OLI/TIRS 0 roll angle check");
    assert_double_equal_with_message(-113.87946870364975,
        dsw_data.corners_tirs.upright.lng, "UR.lng for TIRS did not match "
        "expected results in OLI/TIRS 0 roll angle check");
    assert_double_equal_with_message(33.588276967933666,
        dsw_data.corners_tirs.loleft.lat, "LL.lat for TIRS did not match "
        "expected results in OLI/TIRS 0 roll angle check");
    assert_double_equal_with_message(-116.59870170258117,
        dsw_data.corners_tirs.loleft.lng, "LL.lng for TIRS did not match "
        "expected results in OLI/TIRS 0 roll angle check"); 
    assert_double_equal_with_message(33.199895532955985,
        dsw_data.corners_tirs.loright.lat, "LR.lat for TIRS did not match "
        "expected results in OLI/TIRS 0 roll angle check");
    assert_double_equal_with_message(-114.5995288507016,
        dsw_data.corners_tirs.loright.lng, "LR.lng for TIRS did not match "
        "expected results in OLI/TIRS 0 roll angle check");

    /* Test OLI/TIRS with positive roll angle */

    /* setup the corners */
    dsw_data.corners_oli.upleft.lat = 35.197957214342878;
    dsw_data.corners_oli.upleft.lng = -114.06990901769714;
    dsw_data.corners_oli.upright.lat = 34.736773153167078;
    dsw_data.corners_oli.upright.lng = -111.86713841788888;
    dsw_data.corners_oli.loleft.lat = 33.845031962837957;
    dsw_data.corners_oli.loleft.lng = -114.46831994916761;
    dsw_data.corners_oli.loright.lat = 33.400398303319889;
    dsw_data.corners_oli.loright.lng = -112.2958307921746;
    dsw_data.corners_tirs.upleft.lat = 35.197957214342878;
    dsw_data.corners_tirs.upleft.lng = -114.06990901769714;
    dsw_data.corners_tirs.upright.lat = 34.736773153167078;
    dsw_data.corners_tirs.upright.lng = -111.86713841788888;
    dsw_data.corners_tirs.loleft.lat = 33.845031962837957;
    dsw_data.corners_tirs.loleft.lng = -114.46831994916761;
    dsw_data.corners_tirs.loright.lat = 33.400398303319889;
    dsw_data.corners_tirs.loright.lng = -112.2958307921746;

    /* setup the sensor present flags, scene center and off-nadir angle */
    dsw_data.sensor_present_oli = 1;
    dsw_data.sensor_present_tirs = 1;
    dsw_data.scene_center.lat = 34.295040158416953; 
    dsw_data.scene_center.lng = -113.17529954423206; 
    dsw_data.off_nadir_angle = 15.000000;

    status = extend_corners(&dsw_data, earth_consts, orbit_parms, focal_plane);
    
    assert_equal_with_message(status, SUCCESS, "OLI/TIRS positive roll angle "
        "test did not return SUCCESS");
    assert_double_equal_with_message(35.379909849798381,
        dsw_data.corners_oli.upleft.lat, "UL.lat for OLI did not match "
        "expected results in OLI/TIRS positive roll angle check");
    assert_double_equal_with_message(-114.01529244390575,
        dsw_data.corners_oli.upleft.lng, "UL.lng for OLI did not match "
        "expected results in OLI/TIRS positive roll angle check");
    assert_double_equal_with_message(34.933169599412459,
        dsw_data.corners_oli.upright.lat, "UR.lat for OLI did not match "
        "expected results in OLI/TIRS positive roll angle check"); 
    assert_double_equal_with_message(-111.80293511019269,
        dsw_data.corners_oli.upright.lng, "UR.lng for OLI did not match "
        "expected results in OLI/TIRS positive roll angle check"); 
    assert_double_equal_with_message(33.662820723625494,
        dsw_data.corners_oli.loleft.lat, "LL.lat for OLI did not match "
        "expected results in OLI/TIRS positive roll angle check");
    assert_double_equal_with_message(-114.52097772610546,
        dsw_data.corners_oli.loleft.lng, "LL.lng for OLI did not match "
        "expected results in OLI/TIRS positive roll angle check"); 
    assert_double_equal_with_message(33.203685298948699,
        dsw_data.corners_oli.loright.lat, "LR.lat for OLI did not match "
        "expected results in OLI/TIRS positive roll angle check"); 
    assert_double_equal_with_message(-112.3577727913917,
        dsw_data.corners_oli.loright.lng, "LR.lng for OLI did not match "
        "expected results in OLI/TIRS positive roll angle check"); 
    assert_double_equal_with_message(35.741289772381471,
        dsw_data.corners_tirs.upleft.lat, "UL.lat for TIRS did not match "
        "expected results in OLI/TIRS positive roll angle check");
    assert_double_equal_with_message(-113.90606023162704,
        dsw_data.corners_tirs.upleft.lng, "UL.lng for TIRS did not match "
        "expected results in OLI/TIRS positive roll angle check");
    assert_double_equal_with_message(35.322356551640439,
        dsw_data.corners_tirs.upright.lat, "UR.lat for TIRS did not match "
        "expected results in OLI/TIRS positive roll angle check");
    assert_double_equal_with_message(-111.67476007856136,
        dsw_data.corners_tirs.upright.lng, "UR.lng for TIRS did not match "
        "expected results in OLI/TIRS positive roll angle check");
    assert_double_equal_with_message(33.300744870045506,
        dsw_data.corners_tirs.loleft.lat, "LL.lat for TIRS did not match "
        "expected results in OLI/TIRS positive roll angle check");
    assert_double_equal_with_message(-114.62493806830807,
        dsw_data.corners_tirs.loleft.lng, "LL.lng for TIRS did not match "
        "expected results in OLI/TIRS positive roll angle check"); 
    assert_double_equal_with_message(32.813630649051333,
        dsw_data.corners_tirs.loright.lat, "LR.lat for TIRS did not match "
        "expected results in OLI/TIRS positive roll angle check");
    assert_double_equal_with_message(-112.47974891741025,
        dsw_data.corners_tirs.loright.lng, "LR.lng for TIRS did not match "
        "expected results in OLI/TIRS positive roll angle check");

    /* Test OLI/TIRS with negative roll angle */

    /* setup the corners */
    dsw_data.corners_oli.upleft.lat = 35.749221733843697;
    dsw_data.corners_oli.upleft.lng = -118.14408422104036;
    dsw_data.corners_oli.upright.lat = 35.378441608084685;
    dsw_data.corners_oli.upright.lng = -115.92314596547499;
    dsw_data.corners_oli.loleft.lat = 34.402736458598156;
    dsw_data.corners_oli.loleft.lng = -118.47318704324742;
    dsw_data.corners_oli.loright.lat = 34.016007762900415;
    dsw_data.corners_oli.loright.lng = -116.29267874283691;
    dsw_data.corners_tirs.upleft.lat = 35.749221733843697;
    dsw_data.corners_tirs.upleft.lng = -118.14408422104036;
    dsw_data.corners_tirs.upright.lat = 35.378441608084685;
    dsw_data.corners_tirs.upright.lng = -115.92314596547499;
    dsw_data.corners_tirs.loleft.lat = 34.402736458598156;
    dsw_data.corners_tirs.loleft.lng = -118.47318704324742;
    dsw_data.corners_tirs.loright.lat = 34.016007762900415;
    dsw_data.corners_tirs.loright.lng = -116.29267874283691;

    /* setup the sensor present flags, scene center and off-nadir angle */
    dsw_data.sensor_present_oli = 1;
    dsw_data.sensor_present_tirs = 1;
    dsw_data.scene_center.lat = 34.886601890856738;
    dsw_data.scene_center.lng = -117.20827399314993;
    dsw_data.off_nadir_angle = -14.000000;

    status = extend_corners(&dsw_data, earth_consts, orbit_parms, focal_plane);
    
    assert_equal_with_message(status, SUCCESS, "OLI/TIRS negative roll angle "
        "test did not return SUCCESS");
    assert_double_equal_with_message(35.946977685304645,
        dsw_data.corners_oli.upleft.lat, "UL.lat for OLI did not match "
        "expected results in OLI/TIRS negative roll angle check");
    assert_double_equal_with_message(-118.09479913957583,
        dsw_data.corners_oli.upleft.lng, "UL.lng for OLI did not match "
        "expected results in OLI/TIRS negative roll angle check");
    assert_double_equal_with_message(35.5608370564944,
        dsw_data.corners_oli.upright.lat, "UR.lat for OLI did not match "
        "expected results in OLI/TIRS negative roll angle check"); 
    assert_double_equal_with_message(-115.87271101130965,
        dsw_data.corners_oli.upright.lng, "UR.lng for OLI did not match "
        "expected results in OLI/TIRS negative roll angle check"); 
    assert_double_equal_with_message(34.204763410477987,
        dsw_data.corners_oli.loleft.lat, "LL.lat for OLI did not match "
        "expected results in OLI/TIRS negative roll angle check");
    assert_double_equal_with_message(-118.52065734403033,
        dsw_data.corners_oli.loleft.lng, "LL.lng for OLI did not match "
        "expected results in OLI/TIRS negative roll angle check"); 
    assert_double_equal_with_message(33.83338111414875,
        dsw_data.corners_oli.loright.lat, "LR.lat for OLI did not match "
        "expected results in OLI/TIRS negative roll angle check"); 
    assert_double_equal_with_message(-116.34128154091479,
        dsw_data.corners_oli.loright.lng, "LR.lng for OLI did not match "
        "expected results in OLI/TIRS negative roll angle check"); 
    assert_double_equal_with_message(36.33893292813066,
        dsw_data.corners_tirs.upleft.lat, "UL.lat for TIRS did not match "
        "expected results in OLI/TIRS negative roll angle check");
    assert_double_equal_with_message(-117.99636643790693,
        dsw_data.corners_tirs.upleft.lng, "UL.lng for TIRS did not match "
        "expected results in OLI/TIRS negative roll angle check");
    assert_double_equal_with_message(35.923081643958113,
        dsw_data.corners_tirs.upright.lat, "UR.lat for TIRS did not match "
        "expected results in OLI/TIRS negative roll angle check");
    assert_double_equal_with_message(-115.77184290153973,
        dsw_data.corners_tirs.upright.lng, "UR.lng for TIRS did not match "
        "expected results in OLI/TIRS negative roll angle check");
    assert_double_equal_with_message(33.812213188650325,
        dsw_data.corners_tirs.loleft.lat, "LL.lat for TIRS did not match "
        "expected results in OLI/TIRS negative roll angle check");
    assert_double_equal_with_message(-118.6141156536127,
        dsw_data.corners_tirs.loleft.lng, "LL.lng for TIRS did not match "
        "expected results in OLI/TIRS negative roll angle check"); 
    assert_double_equal_with_message(33.470515065193219,
        dsw_data.corners_tirs.loright.lat, "LR.lat for TIRS did not match "
        "expected results in OLI/TIRS negative roll angle check");
    assert_double_equal_with_message(-116.4372242718249,
        dsw_data.corners_tirs.loright.lng, "LR.lng for TIRS did not match "
        "expected results in OLI/TIRS negative roll angle check");

    /* Test TIRS only */

    /* setup the TIRS corners */
    dsw_data.corners_oli.upleft.lat = 0.00000;
    dsw_data.corners_oli.upleft.lng = 0.00000;
    dsw_data.corners_oli.upright.lat = 0.00000;
    dsw_data.corners_oli.upright.lng = 0.00000;
    dsw_data.corners_oli.loleft.lat = 0.00000;
    dsw_data.corners_oli.loleft.lng = 0.00000;
    dsw_data.corners_oli.loright.lat = 0.00000;
    dsw_data.corners_oli.loright.lng = 0.00000;
    dsw_data.corners_tirs.upleft.lat = 36.92592750754725;
    dsw_data.corners_tirs.upleft.lng = -115.67278701794972;
    dsw_data.corners_tirs.upright.lat = 36.537588999539743;
    dsw_data.corners_tirs.upright.lng = -113.6241927161967;
    dsw_data.corners_tirs.loleft.lat = 35.517679496407645;
    dsw_data.corners_tirs.loleft.lng = -116.06321211578953;
    dsw_data.corners_tirs.loright.lat = 35.13068326696245;
    dsw_data.corners_tirs.loright.lng = -114.05082924635826;

    /* setup the sensor present flags, scene center, and off-nadir angle */
    dsw_data.sensor_present_oli = 0;
    dsw_data.sensor_present_tirs = 1;
    dsw_data.scene_center.lat = 36.027969817614277;
    dsw_data.scene_center.lng = -114.85275527407354; 
    dsw_data.off_nadir_angle = 0.000000;

    status = extend_corners(&dsw_data, earth_consts, orbit_parms, focal_plane);
 
    assert_equal_with_message(status, SUCCESS, "TIRS only test did not return "
        "SUCCESS");
    assert_double_equal_with_message(37.472196690040825,
        dsw_data.corners_tirs.upleft.lat,  
        "UL.lat did not match expected results in TIRS only check");
    assert_double_equal_with_message(-115.51735434562642,
        dsw_data.corners_tirs.upleft.lng,  
        "UL.lng did not match expected results in TIRS only check");
    assert_double_equal_with_message(37.081127292671667,
        dsw_data.corners_tirs.upright.lat,  
        "UR.lat did not match expected results in TIRS only check");
    assert_double_equal_with_message(-113.45508043308152,
        dsw_data.corners_tirs.upright.lng,  
        "UR.lng did not match expected results in TIRS only check");
    assert_double_equal_with_message(34.970488031470381,
        dsw_data.corners_tirs.loleft.lat,  
        "LL.lat did not match expected results in TIRS only check");
    assert_double_equal_with_message(-116.21112961862777,
        dsw_data.corners_tirs.loleft.lng,  
        "LL.lng did not match expected results in TIRS only check");
    assert_double_equal_with_message(34.586100319070724,
        dsw_data.corners_tirs.loright.lat,  
        "LR.lat did not match expected results in TIRS only check");
    assert_double_equal_with_message(-114.21189449005747,
        dsw_data.corners_tirs.loright.lng,  
        "LR.lng did not match expected results in TIRS only check");

    /* Clean up */
    ias_cpf_free(cpf);
}

Ensure calc_l0r_corners_test()
{
    int status;
    DSW_DATA dsw_data;

    /* Set the record count */
    dsw_data.record_count = 1;

    /* initial the scene center point */
    dsw_data.scene_center.lat = 47.1;
    dsw_data.scene_center.lng = -55.5;

    /* setup the OLI corners */
    dsw_data.corners_oli.upleft.lat = 60.1;
    dsw_data.corners_oli.upleft.lng = -65.2;
    dsw_data.corners_oli.upright.lat = 60.1;
    dsw_data.corners_oli.upright.lng = -45.3;
    dsw_data.corners_oli.loleft.lat = 40.2;
    dsw_data.corners_oli.loleft.lng = -65.2;
    dsw_data.corners_oli.loright.lat = 40.2;
    dsw_data.corners_oli.loright.lng = -45.3;

    /* setup the TIRS corners */
    dsw_data.corners_tirs.upleft.lat = 62.1;
    dsw_data.corners_tirs.upleft.lng = -66.2;
    dsw_data.corners_tirs.upright.lat = 59.1;
    dsw_data.corners_tirs.upright.lng = -46.3;
    dsw_data.corners_tirs.loleft.lat = 41.2;
    dsw_data.corners_tirs.loleft.lng = -66.2;
    dsw_data.corners_tirs.loright.lat = 39.2;
    dsw_data.corners_tirs.loright.lng = -46.3;

    /* Test OLI and TIRS missing */
    dsw_data.sensor_present_oli = 0;
    dsw_data.sensor_present_tirs = 0;

    /* Temporarily disable error logging */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    status = calc_l0r_corners(&dsw_data);

    /* Reenable error logging */
    ias_log_set_output_level(IAS_LOG_LEVEL_ERROR);

    assert_equal_with_message(status, ERROR, "Missing OLI and TIRS did not "
        "return ERROR");

    /* Test OLI only */
    dsw_data.sensor_present_oli = 1;
    dsw_data.sensor_present_tirs = 0;

    status = calc_l0r_corners(&dsw_data);
    
    assert_equal_with_message(status, SUCCESS, "OLI only test did not return "
        "SUCCESS");
    assert_double_equal_with_message(dsw_data.upper_left_lat_lon[0], 
        dsw_data.corners_oli.upleft.lat, 
        "UL.lat - OLI corner points were not used in OLI only check");
    assert_double_equal_with_message(dsw_data.upper_left_lat_lon[1], 
        dsw_data.corners_oli.upleft.lng, 
        "UL.lng - OLI corner points were not used in OLI only check");
    assert_double_equal_with_message(dsw_data.upper_right_lat_lon[0], 
        dsw_data.corners_oli.upright.lat, 
        "UR.lat - OLI corner points were not used in OLI only check");
    assert_double_equal_with_message(dsw_data.upper_right_lat_lon[1], 
        dsw_data.corners_oli.upright.lng, 
        "UR.lng - OLI corner points were not used in OLI only check");
    assert_double_equal_with_message(dsw_data.lower_right_lat_lon[0], 
        dsw_data.corners_oli.loright.lat, 
        "LR.lat - OLI corner points were not used in OLI only check");
    assert_double_equal_with_message(dsw_data.lower_right_lat_lon[1], 
        dsw_data.corners_oli.loright.lng, 
        "LR.lng - OLI corner points were not used in OLI only check");
    assert_double_equal_with_message(dsw_data.lower_left_lat_lon[0], 
        dsw_data.corners_oli.loleft.lat, 
        "LL.lat - OLI corner points were not used in OLI only check");
    assert_double_equal_with_message(dsw_data.lower_left_lat_lon[1], 
        dsw_data.corners_oli.loleft.lng, 
        "LL.lng - OLI corner points were not used in OLI only check");

    /* Test TIRS only */
    dsw_data.sensor_present_oli = 0;
    dsw_data.sensor_present_tirs = 1;
   
    status = calc_l0r_corners(&dsw_data);
    
    assert_equal_with_message(status, SUCCESS, "TIRS only test did not return "
        "SUCCESS");
    assert_double_equal_with_message(dsw_data.upper_left_lat_lon[0], 
        dsw_data.corners_tirs.upleft.lat, 
        "UL.lat - TIRS corner points were not used in TIRS only check");
    assert_double_equal_with_message(dsw_data.upper_left_lat_lon[1], 
        dsw_data.corners_tirs.upleft.lng, 
        "UL.lng - TIRS corner points were not used in TIRS only check");
    assert_double_equal_with_message(dsw_data.upper_right_lat_lon[0], 
        dsw_data.corners_tirs.upright.lat, 
        "UR.lat - TIRS corner points were not used in TIRS only check");
    assert_double_equal_with_message(dsw_data.upper_right_lat_lon[1], 
        dsw_data.corners_tirs.upright.lng, 
        "UR.lng - TIRS corner points were not used in TIRS only check");
    assert_double_equal_with_message(dsw_data.lower_right_lat_lon[0], 
        dsw_data.corners_tirs.loright.lat, 
        "LR.lat - TIRS corner points were not used in TIRS only check");
    assert_double_equal_with_message(dsw_data.lower_right_lat_lon[1], 
        dsw_data.corners_tirs.loright.lng, 
        "LR.lng - TIRS corner points were not used in TIRS only check");
    assert_double_equal_with_message(dsw_data.lower_left_lat_lon[0], 
        dsw_data.corners_tirs.loleft.lat, 
        "LL.lat - TIRS corner points were not used in TIRS only check");
    assert_double_equal_with_message(dsw_data.lower_left_lat_lon[1], 
        dsw_data.corners_tirs.loleft.lng, 
        "LL.lng - TIRS corner points were not used in TIRS only check");

    /* Test combined OLI and TIRS - Test 1 - UL in UL position */
    dsw_data.sensor_present_oli = 1;
    dsw_data.sensor_present_tirs = 1;
   
    status = calc_l0r_corners(&dsw_data);
    
    assert_equal_with_message(status, SUCCESS, "OLI/TIRS combined Test 1 did "
        "not return SUCCESS");
    assert_double_equal_with_message(dsw_data.upper_left_lat_lon[0],
        dsw_data.corners_tirs.upleft.lat,
        "OLI/TIRS Test 1 - UL.lat is incorrect");
    assert_double_equal_with_message(dsw_data.upper_left_lat_lon[1],
        dsw_data.corners_tirs.upleft.lng,
        "OLI/TIRS Test 1 - UL.lng is incorrect");
    assert_double_equal_with_message(dsw_data.upper_right_lat_lon[0],
        dsw_data.corners_oli.upright.lat,
        "OLI/TIRS Test 1 - UR.lat is incorrect");
    assert_double_equal_with_message(dsw_data.upper_right_lat_lon[1],
        dsw_data.corners_oli.upright.lng,
        "OLI/TIRS Test 1 - UR.lng is incorrect");
    assert_double_equal_with_message(dsw_data.lower_right_lat_lon[0],
        dsw_data.corners_tirs.loright.lat,
        "OLI/TIRS Test 1 - LR.lat is incorrect");
    assert_double_equal_with_message(dsw_data.lower_right_lat_lon[1],
        dsw_data.corners_oli.loright.lng,
        "OLI/TIRS Test 1 - LR.lng is incorrect");
    assert_double_equal_with_message(dsw_data.lower_left_lat_lon[0],
        dsw_data.corners_oli.loleft.lat,
        "OLI/TIRS Test 1 - LL.lat is incorrect");
    assert_double_equal_with_message(dsw_data.lower_left_lat_lon[1],
        dsw_data.corners_tirs.loleft.lng,
        "OLI/TIRS Test 1 - LL.lng is incorrect");

    /* Test scene crossing the -180 latitude - Test 2 */
    /* Setup new corner points, use existing latitudes */
    dsw_data.scene_center.lng = -175.5;

    /* setup the OLI corners */
    dsw_data.corners_oli.upleft.lng = 179.2;
    dsw_data.corners_oli.upright.lng = -165.3;
    dsw_data.corners_oli.loleft.lng = 170.2;
    dsw_data.corners_oli.loright.lng = -170.3;

    /* setup the TIRS corners */
    dsw_data.corners_tirs.upleft.lng = -178.2;
    dsw_data.corners_tirs.upright.lng = -163.3;
    dsw_data.corners_tirs.loleft.lng = 175.2;
    dsw_data.corners_tirs.loright.lng = -171.3;

    status = calc_l0r_corners(&dsw_data);
    
    assert_equal_with_message(status, SUCCESS, "OLI/TIRS combined Test 2 did "
        "not return SUCCESS");
    assert_double_equal_with_message(dsw_data.upper_left_lat_lon[1],
        dsw_data.corners_oli.upleft.lng,
        "OLI/TIRS Test 2 - UL.lng is incorrect");
    assert_double_equal_with_message(dsw_data.upper_right_lat_lon[1],
        dsw_data.corners_tirs.upright.lng,
        "OLI/TIRS Test 2 - UR.lng is incorrect");
    assert_double_equal_with_message(dsw_data.lower_right_lat_lon[1],
        dsw_data.corners_oli.loright.lng,
        "OLI/TIRS Test 2 - LR.lng is incorrect");
    assert_double_equal_with_message(dsw_data.lower_left_lat_lon[1],
        dsw_data.corners_oli.loleft.lng,
        "OLI/TIRS Test 2 - LL.lng is incorrect");

    /* Test scene crossing the Prime Meridian - Test 3 */
    /* Setup new corner points, use existing latitudes */
    dsw_data.scene_center.lng = -2.5;

    /* setup the OLI corners */
    dsw_data.corners_oli.upleft.lng = -7.0;
    dsw_data.corners_oli.upright.lng = 1.3;
    dsw_data.corners_oli.loleft.lng = -7.0;
    dsw_data.corners_oli.loright.lng = -1.2;

    /* setup the TIRS corners */
    dsw_data.corners_tirs.upleft.lng = -6.2;
    dsw_data.corners_tirs.upright.lng = 3.0;
    dsw_data.corners_tirs.loleft.lng = -7.0;
    dsw_data.corners_tirs.loright.lng = 2.0;

    status = calc_l0r_corners(&dsw_data);
    
    assert_equal_with_message(status, SUCCESS, "OLI/TIRS combined Test 3 did "
        "not return SUCCESS");
    assert_double_equal_with_message(dsw_data.upper_left_lat_lon[1],
        dsw_data.corners_oli.upleft.lng,
        "OLI/TIRS Test 3 - UL.lng is incorrect");
    assert_double_equal_with_message(dsw_data.upper_right_lat_lon[1],
        dsw_data.corners_tirs.upright.lng,
        "OLI/TIRS Test 3 - UR.lng is incorrect");
    assert_double_equal_with_message(dsw_data.lower_right_lat_lon[1],
        dsw_data.corners_tirs.loright.lng,
        "OLI/TIRS Test 3 - LR.lng is incorrect");
    assert_double_equal_with_message(dsw_data.lower_left_lat_lon[1],
        dsw_data.corners_oli.loleft.lng,
        "OLI/TIRS Test 3 - LL.lng is incorrect");

    /* Test OLI/TIRS corners on opposite sides of center point - Test 4 */

    /* setup center point */
    dsw_data.scene_center.lat = 45.0;
    dsw_data.scene_center.lng = 45.0;

    /* setup the OLI corners */
    dsw_data.corners_oli.upleft.lat = 55.0;
    dsw_data.corners_oli.upleft.lng = 44.0;
    dsw_data.corners_oli.upright.lat = 44.0;
    dsw_data.corners_oli.upright.lng = 55.0;
    dsw_data.corners_oli.loright.lat = 35.0;
    dsw_data.corners_oli.loright.lng = 46.0;
    dsw_data.corners_oli.loleft.lat = 45.0;
    dsw_data.corners_oli.loleft.lng = 35.0;

    /* setup the TIRS corners */
    dsw_data.corners_tirs.upleft.lat = 56.0;
    dsw_data.corners_tirs.upleft.lng = 46.0;
    dsw_data.corners_tirs.upright.lat = 46.0;
    dsw_data.corners_tirs.upright.lng = 56.0;
    dsw_data.corners_tirs.loright.lat = 34.0;
    dsw_data.corners_tirs.loright.lng = 44.0;
    dsw_data.corners_tirs.loleft.lat = 44.0;
    dsw_data.corners_tirs.loleft.lng = 34.0;

    status = calc_l0r_corners(&dsw_data);
    
    assert_equal_with_message(status, SUCCESS, "OLI/TIRS combined Test 4 did "
        "not return SUCCESS");
    assert_double_equal_with_message(dsw_data.upper_left_lat_lon[0],
        dsw_data.corners_tirs.upleft.lat,
        "OLI/TIRS Test 4 - UL.lat is incorrect");
    assert_double_equal_with_message(dsw_data.upper_left_lat_lon[1],
        dsw_data.corners_tirs.upleft.lng,
        "OLI/TIRS Test 4 - UL.lng is incorrect");
    assert_double_equal_with_message(dsw_data.upper_right_lat_lon[0],
        dsw_data.corners_tirs.upright.lat,
        "OLI/TIRS Test 4 - UR.lat is incorrect");
    assert_double_equal_with_message(dsw_data.upper_right_lat_lon[1],
        dsw_data.corners_tirs.upright.lng,
        "OLI/TIRS Test 4 - UR.lng is incorrect");
    assert_double_equal_with_message(dsw_data.lower_right_lat_lon[0],
        dsw_data.corners_tirs.loright.lat,
        "OLI/TIRS Test 4 - LR.lat is incorrect");
    assert_double_equal_with_message(dsw_data.lower_right_lat_lon[1],
        dsw_data.corners_oli.loright.lng,
        "OLI/TIRS Test 4 - LR.lng is incorrect");
    assert_double_equal_with_message(dsw_data.lower_left_lat_lon[0],
        dsw_data.corners_tirs.loleft.lat,
        "OLI/TIRS Test 4 - LL.lat is incorrect");
    assert_double_equal_with_message(dsw_data.lower_left_lat_lon[1],
        dsw_data.corners_tirs.loleft.lng,
        "OLI/TIRS Test 4 - LL.lng is incorrect");
}

Ensure read_l0r_metadata_test()
{
    char data_dir[PATH_MAX] =
        "ias_base/dms/setup_work_order/l0rp_OLITIRS_partials";
    char l0r_name[100] = "LC80350102013107LGN00";
    char sys_command[100];
    char *test_data_dir = NULL;
    char *processing_center = NULL;
    char *ias_services = NULL;
    DSW_DATA test_dsw;
    int status;

    /* Initialize */
    memset(&test_dsw, 0, sizeof(DSW_DATA));

    status = ias_sat_attr_initialize(IAS_L8);
    assert_equal_with_message(status, SUCCESS,
        "Could not initialize satellite attributes library");

    /* Get some required environment settings. */
    processing_center = getenv("PROCESSING_CENTER");
    ias_services      = getenv("IAS_SERVICES");
    test_data_dir     = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!(processing_center || ias_services || test_data_dir))
    {
        assert_true_with_message(FALSE, "Error locating required environment "
            "variables");
        return;
    }

    /* Provide some initial information to the test DSW data structure. */
    status = snprintf(test_dsw.processing_center,
        sizeof(test_dsw.processing_center), "%s", processing_center);
    if ((status < 0) || (status >= sizeof(test_dsw.processing_center)))
    {
        assert_true_with_message(FALSE, "Error writing processing center "
            "to work order data structure");
        return;
    }

    status = snprintf(test_dsw.services_location,
        sizeof(test_dsw.services_location), "%s", ias_services);
    if ((status < 0) || (status >= sizeof(test_dsw.services_location)))
    {
        assert_true_with_message(FALSE, "Error writing IAS services "
            "information to work order data structure");
        return;
    }

    status = snprintf(test_dsw.l0r_filename, sizeof(test_dsw.l0r_filename),
        "%s/%s/%s/%s", test_data_dir, data_dir, l0r_name, l0r_name);
    if ((status < 0) || (status >= sizeof(test_dsw.l0r_filename)))
    {
        assert_true_with_message(FALSE, "Error writing test L0R file name "
            "to work order data structure");
        return;
    }

    /* Read the L0R metadata. */
    status = read_l0r_metadata(&test_dsw);
    assert_equal_with_message(status, SUCCESS, "Error reading L0R metadata "
        "for test scene %s", l0r_name);

    /* Check the sensor_present_xxx fields in the data structure.  The dataset
       tested here represents a case where the sensor id in the interval
       metadata is 'OLI_TIRS', but the scene metadata indicates that just the
       OLI sensor collected an image.  The values should be TRUE (1) for
       test_dsw.sensor_oli_present and FALSE (0) for
       test_dsw.sensor_present_tirs.

       See the note in 'read_l0r_metadata.c' about why 'OLITIRS'
       is the recorded sensor ID rather than 'OLI_TIRS'... */
    assert_string_equal_with_message(test_dsw.sensor_id_from_l0r, "OLITIRS",
        "Sensor ID from L0R metadata should be 'OLITIRS' for test scene %s",
        l0r_name);
    assert_equal_with_message(test_dsw.sensor_present_oli, TRUE,
        "OLI sensor should be present for test scene %s", l0r_name);
    assert_equal_with_message(test_dsw.sensor_present_tirs, FALSE,
        "TIRS sensor should not be present for test scene %s", l0r_name);

    /* Also check the first value of the corners.  The OLI corners
       should be set (since the OLI collected an image); the TIRS corners
       should should not be set (since TIRS didn't collect an image). */
    assert_not_equal_with_message(test_dsw.corners_oli.upleft.lat, 0,
        "OLI corners should be set");
    assert_not_equal_with_message(test_dsw.corners_oli.upleft.lng, 0,
        "OLI corners should be set");
    assert_equal_with_message(test_dsw.corners_tirs.upleft.lat, 0,
        "TIRS corners should be set");
    assert_equal_with_message(test_dsw.corners_tirs.upleft.lng, 0,
        "TIRS corners should be set");

    /* Remove the 'l0r_symlink' entry created by 'read_l0r_metadata'. */
    status = snprintf(sys_command, sizeof(sys_command), "rm l0r_symlink");
    if ((status < 0) || (status >= sizeof(sys_command)))
    {
        assert_true_with_message(FALSE, "Error creating 'rm l0r_symlink' "
            "command string");
        return;
    }
    status = system(sys_command);
    assert_equal_with_message(status, 0, "Error deleting test 'l0r_symlink' "
        "output");
}

/* Tests out the calculate_utm_zone function */
Ensure utm_zone_calc_test()
{
    double longitudes[] = {46.5, -46.5, 90.1, 89.1, -98.0, -141.2,-180.0,
                           180.0, 179.9};
    int expected_zone[] = {38, 23, 46, 45, 14, 7, 1, 1, 60};
    int i;

    for (i=0; i < sizeof(longitudes) / sizeof(double); i++)
    {
        char error[1024] = "";
        double lng =  longitudes[i];
        sprintf(error, "Wrong zone for longitude %f.", lng);
        assert_equal_with_message(calculate_utm_zone(lng), expected_zone[i],
                                  error);
    }
}

int main(int argc, char **argv)
{
    TestSuite *suite = create_test_suite();

    add_test(suite, extend_corners_test);
    add_test(suite, calc_l0r_corners_test);
    add_test(suite, read_l0r_metadata_test);
    add_test(suite, utm_zone_calc_test);
    return run_test_suite(suite, create_text_reporter());
}
