#!/usr/bin/perl -w

##############################################################################
#
# This script loads a record file into the database for the specified schema.
#
# Usage:
#     ./load_records <db connection info> <record file> [env var replacement]
#
#     <db connection info> == typically $IAS_DB_COM
#            <record file> == full path to the file
#    [env var replacement] == an environment variable to string replace
#
##############################################################################
#
# Record File Format
#
# line 1 = tablename
# line 2 = comma seperated list of column names
# line 3 = pipe seperated list of values
#   .
#   .
#   .
# line N = pipe seperated list of values
#
##############################################################################

use strict;

use ias_const;
use ias_db;

my $dbh;
my $sth;
my $userID;
my $password;
my $sql;
my $db_connection_info;
my $record_filename;
my $env_var_defined = 0;
my $env_var_name;
my $env_var_value;

# test for the command line parameters
if (not defined $ARGV[0])
{
    print "Error missing database connection information\n";
    exit FAILURE;
}
if (not defined $ARGV[1])
{
    print "Error missing records filename\n";
    exit FAILURE;
}
if (defined $ARGV[2])
{
    $env_var_defined = 1;
    $env_var_name = $ARGV[2];
}

# test for the existence of the file
if (not -e $ARGV[1])
{
    print "Error record file [$ARGV[1]] does not exist\n";
    exit FAILURE;
}

# Make sure the collection number is defined
my $instance = $ENV{IAS_INSTANCE_NAME};
if (!defined $instance)
{
    print "Error IAS_INSTANCE_NAME not set to valid value";
    exit FAILURE;
}

$db_connection_info = $ARGV[0] ;
$record_filename = $ARGV[1] ;

if ($env_var_defined)
{
    $env_var_value = $ENV{$env_var_name};

    if (not defined $env_var_value)
    {
        print "Error environment variable specified but does not exist\n";
        exit FAILURE;
    }

    if ($env_var_value eq "")
    {
        print "Error environment variable defined but does not have a value\n";
        exit FAILURE;
    }
}

# get the username and password for the database
($userID, $password) = split('/', $db_connection_info);

# connect to the database
$dbh = ias_db_ConnectToDB($userID, $password);
if (not $dbh)
{
    print "Error connecting to database: $ias_db_error\n";
    exit FAILURE;
}

# open the input data file
open(RECORD_FILE, "<$record_filename") or exit FAILURE;

# process the table name
my $table_name = <RECORD_FILE>;
chomp($table_name);

# process the column names
my $columns = <RECORD_FILE>;
chomp($columns);
my @COLUMNS = split(',', $columns);

# setup the sql insert prefix
my $sql_prefix = "insert into $table_name (";
my $first = 0;
foreach my $column (@COLUMNS)
{
    if ($first)
    {
        $sql_prefix .= ", ";
    }

    $sql_prefix .= "$column";
    $first = 1;
}
$sql_prefix .= ") values (";
my $column_count = $#COLUMNS;

# process the records in the input file
while (<RECORD_FILE>)
{
    my $line = $_;
    chomp($line);

    # Replace the collection with a valid one
    $line =~ s/COLLECTION_REPLACE/$instance/g;

    # if defined replace the environment variable
    if ($env_var_defined)
    {
        $line =~ s/$env_var_name/$env_var_value/g;
    }

    # get the column values
    my @VALUES = split('\|', $line);

    # add the values to the sql
    $first = 0;
    $sql = $sql_prefix;
    foreach my $value (@VALUES)
    {
        if ($first)
        {
            $sql .= ", ";
        }

        $sql .= "'$value'";
        $first = 1;
    }
    # For some reason empty pipes at the tail end don't get counted in the
    # split command above.  This for loop fixes them.  I don't get it, I
    # don't care, and I don't have time to fix it.
    for (my $i = 0; $i < $column_count - $#VALUES; $i++)
    {
        $sql .= ", ''";
    }
    $sql .= ")";

    # DEBUG printf $sql . "\n";

    # prepare the sql
    $sth = $dbh->prepare($sql);
    if (not $sth)
    {
        print "Error failed to prepare sql [$sql]: $DBI::errstr\n";
        exit FAILURE;
    }

    # execute the sql
    if (not $sth->execute())
    {
        print "Error failed to execute sql [$sql]: $DBI::errstr\n";
        exit FAILURE;
    }

    $sth->finish();
}

# close the input file
close(RECORD_FILE);

# disconnect from the database
if (not ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_COMMIT))
{
    print "Warning: Failed disconnect from database: $ias_db_error\n";
}

# terminate the application
exit SUCCESS;

