/*****************************************************************************
NAME: ancillary_data_preprocessor (main)

PURPOSE:    
    Main driver for the ADP which prepares the ancillary data provided by the
    spacecraft for use by subsequent geometric components.

This includes:
    1) Identifying the time window appropriate to the associated image data
    2) Quality checking the incoming data to identify and remove outliers
    3) Applying scale factors from the CPF to convert the relevant ancillary
       data fields to engineering units
    4) Processing the spacecraft attitude and ephemeris data to construct 
       consistent attitude and ephemeris time histories for the data set

The ADP will:
    Read L0Rp and CPF filenames and locations of L0Ra collection from OMF
    Read CPF
    Read L0R ancillary data
    Preprocess ancillary ephemeris data
    Preprocess ancillary attitude data
    Create the output ephemeris and attitude data set
    Write ancillary data to file
    Insert characterization data
    Write ancillary data filename to OMF

RETURN VALUE: Type = int
    Value           Description
    -----           -----------
    EXIT_FAILURE    Error in preparing the ancillary data
    EXIT_SUCCESS    Successful in preparing the ancillary data
*****************************************************************************/

#include <stdlib.h>
#include <string.h>

#include "ias_logging.h"
#include "ias_parm.h"
#include "ias_ancillary.h"
#include "ancillary_data_preprocessor.h"

int main(int argc, char *argv[])
{
    int status = ERROR;           /* Status returned from function call */
    ADP_DATA adp_data;            /* Parameters read from application
                                     parameter file and OMF */
    char prog_name[PATH_MAX];     /* Name of the program */

    IAS_CPF *cpf = NULL;          /* Structure for CPF */
    const IAS_CPF_EARTH_CONSTANTS *earth_constants = NULL; /* CPF Earth 
                                                              constants */ 
    IAS_L0R_EPHEMERIS *ephemeris_data = NULL;
    IAS_L0R_ATTITUDE *attitude_data = NULL;
    IAS_L0R_IMU *l0r_imu = NULL;
    IAS_ANC_EPHEMERIS_DATA *anc_ephemeris_data = NULL;
    IAS_ANC_ATTITUDE_DATA *anc_attitude_data = NULL;

    int ephemeris_size;           /* Number of ephemeris records */
    int attitude_size;            /* Number of attitude records */
    int imu_size;                 /* Number of IMU records */

    /* Get the name of the program */
    strcpy(prog_name, "ancillary_data_preprocessor");

    /* Initialize the IAS logging library */
    status = ias_log_initialize(prog_name);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing log");
        exit(EXIT_FAILURE);
    }

    /* Check the number of arguments.
       We want the program name and the ODL file for args. */
    if (argc != 2)
    {
        IAS_LOG_ERROR("Usage: %s ODL_Parameter_File", prog_name);
        exit(EXIT_FAILURE);
    }

    memset(&adp_data, 0, sizeof(adp_data));

    /* Read the parameters */
    status = read_parameters(argv[1], &adp_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters");
        exit(EXIT_FAILURE);
    }

    /* Initialize the satellite attributes library */
    status = ias_sat_attr_initialize(adp_data.satellite_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing sat_attr");
        exit(EXIT_FAILURE);
    }

    /* Log initiation */
    IAS_LOG_INFO("Initiated");

    /* Read the CPF */
    IAS_LOG_DEBUG("Reading CPF file %s", adp_data.cpf_filename);
    cpf = ias_cpf_read(adp_data.cpf_filename);
    if (cpf == NULL)
    {
        IAS_LOG_ERROR("Cannot read CPF contents");
        exit(EXIT_FAILURE);
    }

    /* Get the CPF earth constants */
    earth_constants = ias_cpf_get_earth_const(cpf);
    if (earth_constants == NULL)
    {
        IAS_LOG_ERROR("Reading CPF earth constants");
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }

    /* Read l0r metadata */
    status = read_l0r(adp_data.l0r_filename, 
        &earth_constants->leap_seconds_data, &ephemeris_data,
        &attitude_data, &l0r_imu, &ephemeris_size, 
        &attitude_size, &imu_size, &adp_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading L0R metadata");
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }
    IAS_LOG_DEBUG("Ephemeris size: %u, Attitude size: %u", ephemeris_size, 
        attitude_size);

    /* Call the Ancillary library routine(s) to (pre)process the 
       ancillary data */
    status = ias_ancillary_preprocess(cpf, attitude_data,
        attitude_size, ephemeris_data, ephemeris_size, l0r_imu, imu_size,
        adp_data.image_start_time, adp_data.image_stop_time,
        adp_data.acquisition_type, &anc_attitude_data, &anc_ephemeris_data,
        &adp_data.number_of_bad_ephemeris_points,
        &adp_data.number_of_bad_attitude_points);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Pre-processing ancillary data");
        free(anc_ephemeris_data);
        free(anc_attitude_data);
        free(ephemeris_data);
        free(attitude_data);
        free(l0r_imu);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }

    /* Free the CPF memory */
    ias_cpf_free(cpf);
    cpf = NULL;

    /* Write ancillary information to file/disk */
    status = ias_ancillary_write(adp_data.anc_filename, anc_attitude_data, 
                                 anc_ephemeris_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing ancillary data to %s", adp_data.anc_filename);
        free(ephemeris_data);
        free(attitude_data);
        free(l0r_imu);
        free(anc_ephemeris_data);
        free(anc_attitude_data);
        exit(EXIT_FAILURE);
    }

    if (adp_data.characterization_flag)
    {
        /* Finish populating the adp_data */
        adp_data.number_ephemeris_points 
            = anc_ephemeris_data->number_of_samples;
        adp_data.ephemeris_start_time[0] 
            = anc_ephemeris_data->utc_epoch_time[0];
        adp_data.ephemeris_start_time[1] 
            = anc_ephemeris_data->utc_epoch_time[1];
        adp_data.ephemeris_start_time[2] 
            = anc_ephemeris_data->utc_epoch_time[2];
    
        adp_data.number_of_attitude_points 
            = anc_attitude_data->number_of_samples;
        adp_data.attitude_start_time[0] = anc_attitude_data->utc_epoch_time[0];
        adp_data.attitude_start_time[1] = anc_attitude_data->utc_epoch_time[1];
        adp_data.attitude_start_time[2] = anc_attitude_data->utc_epoch_time[2];
    
        /* Write characterization data to the database */
        if (insert_characterization_data(&adp_data) != SUCCESS)
        {
            IAS_LOG_ERROR("Writing characterization data to the database");
            free(ephemeris_data);
            free(attitude_data);
            free(l0r_imu);
            free(anc_ephemeris_data);
            free(anc_attitude_data);
            exit(EXIT_FAILURE);
        }
    }

    /* Write ancillary filename to OMF */
    IAS_PARM_DECLARE_TABLE(parms, 1);

    /* Add the ancillary filename parameter */
    IAS_PARM_ADD_STRING(parms, ANCILLARY_FILENAME, "Ancillary data filename",
        IAS_PARM_REQUIRED, 0, NULL, 1, NULL,
        adp_data.anc_filename, sizeof(adp_data.anc_filename), 0);
    IAS_PARM_SET_COUNT_WRITE(parms, ANCILLARY_FILENAME, 1);

    /* Write the ancillary filename parameter */
    status = ias_parm_update(adp_data.omf_filename, parms, 1);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing ancillary filename to OMF");
        free(ephemeris_data);
        free(attitude_data);
        free(l0r_imu);
        free(anc_ephemeris_data);
        free(anc_attitude_data);
        exit(EXIT_FAILURE);
    }

    /* Free the rest of the structures */
    free(ephemeris_data);
    free(attitude_data);
    free(l0r_imu);
    free(anc_ephemeris_data);
    free(anc_attitude_data);

    IAS_LOG_INFO("Successful completion");

    return EXIT_SUCCESS;
}

