/*****************************************************************************
Name: insert_characterization_data

Purpose: Insert a record into the IAS characterization DB

Returns: Type = int
    Value    Description
    -----    -----------
    SUCCESS  Successful completion
    ERROR    Operation failed
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "ias_const.h"
#include "ias_logging.h"
#include "ias_db.h"
#include "ias_db_insert.h"
#include "ias_db_get_connect_info.h"
#include "ias_miscellaneous.h"
#include "ancillary_data_preprocessor.h"

int insert_characterization_data
(
    const ADP_DATA *adp_data /* I: Parameters retrieved from parameter file */
)
{
    char sql[2048];                    /* Buffer for sql commands */
    char buffer[IAS_SQL_RTN_MSGLEN];   /* Buffer for sql return messages */
    struct ias_db_query *query = NULL; /* DB query structure */
    int status = ERROR;                /* Return values */
    char *database_name = NULL;        /* Database name pointer */
    char *user_name = NULL;            /* User name pointer */
    char *password = NULL;             /* Password pointer */
    char *host = NULL;                 /* Host pointer */
    char geometric_wo_common_id[IAS_CHAR_ID_SIZE];      /* Primary key in
                                          GEOMETRIC_WORK_ORDER_COMMON to
                                          associate processing with the WO */
    struct ias_db_connection *db = NULL; /* Db connection struct */


    /* Get the database connection info */
    status = ias_db_get_connect_info(&database_name, &user_name,
                                     &password, &host);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("From ias_db_get_connect_info");
        return ERROR;
    }

    IAS_LOG_DEBUG("database_name %s", database_name);
    IAS_LOG_DEBUG("user_name %s", user_name);
    IAS_LOG_DEBUG("password %s", password);
    IAS_LOG_DEBUG("adp_data->work_order_id %s", adp_data->work_order_id);

    /* Connect to the database */
    db = ias_db_connect_with_transaction(database_name, user_name,
                                         password, host);

    /* Free the database connection information */
    free(database_name);
    free(user_name);
    free(password);
    free(host);
    database_name = NULL;
    user_name = NULL;
    password = NULL;
    host = NULL;

    /* Get the database ID corresponding to the current work order */
    status = ias_misc_get_char_id(db,
        "geometric_work_order_common", "geometric_wo_common_id",
        adp_data->work_order_id, geometric_wo_common_id);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Retrieving database ID corresponding to current "
                      "geometric work order %s", adp_data->work_order_id);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Delete the previous information from MODEL_GENERATION. */
    status = snprintf(sql, sizeof(sql),
                 "DELETE FROM model_generation WHERE geometric_wo_common_id=%s",
                 geometric_wo_common_id);
    if (status < 0 || status >= sizeof(sql))
    {
        IAS_LOG_ERROR("Creating sql command");
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Perform the query */
    query = ias_db_perform_query(db, sql);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, buffer, IAS_SQL_RTN_MSGLEN);
        IAS_LOG_ERROR("Inserting: %s", buffer);
        ias_db_query_close(query);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Close the query */
    ias_db_query_close(query);

    /* Build the SQL statement */
    status = snprintf(sql, sizeof(sql),
                 "INSERT INTO model_generation "
                 "(geometric_wo_common_id,"
                 " nbr_attitude_points,"
                 " attitude_start_year,"
                 " attitude_start_day,"
                 " attitude_start_seconds,"
                 " nbr_bad_attitude_points,"
                 " nbr_ephemeris_points,"
                 " ephemeris_start_year,"
                 " ephemeris_start_day,"
                 " ephemeris_start_seconds,"
                 " nbr_bad_ephemeris_points)"
                 " values "
                 "(%s, %d, %f, %f, %f,"
                 " %d, %d, %f, %f, %f, %d)",
                 geometric_wo_common_id,
                 adp_data->number_of_attitude_points,
                 adp_data->attitude_start_time[0],
                 adp_data->attitude_start_time[1],
                 adp_data->attitude_start_time[2],
                 adp_data->number_of_bad_attitude_points,
                 adp_data->number_ephemeris_points,
                 adp_data->ephemeris_start_time[0],
                 adp_data->ephemeris_start_time[1],
                 adp_data->ephemeris_start_time[2],
                 adp_data->number_of_bad_ephemeris_points);
    if (status < 0 || status >= sizeof(sql))
    {
        IAS_LOG_ERROR("Creating sql command");
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Perform the query */
    query = ias_db_perform_query(db, sql);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, buffer, 200);
        IAS_LOG_ERROR("Inserting: %s", buffer);
        ias_db_query_close(query);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Close the query */
    ias_db_query_close(query);

    /* The insert worked, so commit the transaction */
    status = ias_db_disconnect_with_transaction(db, IAS_DB_COMMIT);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Committing transaction and disconnecting from database");
    }

    return status;
}

