/*****************************************************************************
Name: read_l0r

Purpose: Reads all needed ancillary samples into variables or structures.

Returns: Type = int
    Value    Description
    -----    -----------
    SUCCESS  Successful completion
    ERROR    Operation failed
*****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "ias_const.h"
#include "ias_types.h"
#include "ias_logging.h"
#include "ias_l0r.h"
#include "ias_ancillary.h"
#include "ancillary_data_preprocessor.h"


int read_l0r
(
    const char *l0r_filename,            /* I: Name of the L0R file */
    const IAS_MATH_LEAP_SECONDS_DATA *leap_seconds_data, /* I: Leap seconds */
    IAS_L0R_EPHEMERIS **ephemeris_data,  /* O: Pointer to ephemeris data
                                               structure */
    IAS_L0R_ATTITUDE **attitude_data,    /* O: Pointer to attitude data
                                               structure */
    IAS_L0R_IMU **imu_data,              /* O: Pointer to IMU data structure */
    int *ephemeris_size,                 /* O: Number of ephemeris records */
    int *attitude_size,                  /* O: Number of attitude records */
    int *imu_size,                       /* O: Number of IMU records */
    ADP_DATA *adp_data                   /* O: Parameters retrieved from 
                                               parameter file and L0R */
)
{
    int oli_frame_header_size;   /* Number of OLI frame header records */
    int tirs_frame_header_size;  /* Number of TIRS frame header records */
    int status;                  /* Status returned from function call */
    int first_valid;             /* First line record with valid time stamp */
    int last_valid;              /* Last line record with valid time stamp */
    double oli_start_total_secs = 0.0; /* Start OLI line header seconds */
    double oli_end_total_secs = 0.0;   /* End OLI line header seconds */
    double tirs_start_total_secs = 0.0;/* Start TIRS line header seconds */
    double tirs_end_total_secs = 0.0;  /* End TIRS line header seconds */
    double start_total_secs;     /* Start seconds from line header for scene */
    double end_total_secs;       /* End seconds from line header for scene */
    L0RIO *l0r_ptr = NULL;       /* L0R file access pointer */
    IAS_L0R_OLI_FRAME_HEADER *oli_frame_header = NULL;
    IAS_L0R_TIRS_FRAME_HEADER *tirs_frame_header = NULL;

    IAS_LOG_DEBUG("Reading L0R: %s", l0r_filename);

    /* Open the L0R file */
    l0r_ptr = ias_l0r_open_from_filename(l0r_filename, 
        IAS_L0R_COMPRESSION_DEFAULT);
    if (l0r_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening L0R file %s", l0r_filename);
        return ERROR;
    }

    /* Open read access to the L0R ancillary data */
    status = ias_l0r_open_ancillary(l0r_ptr, IAS_READ);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Opening read access to L0R ancillary from file %s", 
            l0r_filename);
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Get the ephemeris record count */
    status = ias_l0r_get_ancillary_ephemeris_records_count(l0r_ptr, 
        ephemeris_size);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving L0R ephemeris size from file %s", 
            l0r_filename);
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Verify there are records available */
    if (*ephemeris_size <= 0)
    {
        IAS_LOG_ERROR("No L0R ephemeris records in file %s", 
            l0r_filename );
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Allocate memory for the ephemeris data */
    *ephemeris_data = malloc(sizeof(**ephemeris_data) * (*ephemeris_size));
    if (*ephemeris_data == NULL)
    {
        IAS_LOG_ERROR("Allocating ephemeris_data");
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Get the ephemeris data */
    status = ias_l0r_get_ancillary_ephemeris(l0r_ptr, 0, *ephemeris_size, 
        *ephemeris_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading L0R ephemeris from file %s", l0r_filename);
        ias_l0r_close(l0r_ptr);
        free(*ephemeris_data);
        return ERROR;
    }

    /* Get the attitude record count */
    status = ias_l0r_get_ancillary_attitude_records_count(l0r_ptr, 
        attitude_size);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving L0R attitude size from file %s", 
            l0r_filename);
        ias_l0r_close(l0r_ptr);
        free(*ephemeris_data);
        return ERROR;
    }

    /* Verify there are records available */
    if (*attitude_size <= 0)
    {
        IAS_LOG_ERROR("No L0R attitude records for file %s", l0r_filename);
        ias_l0r_close(l0r_ptr);
        free(*ephemeris_data);
        return ERROR;
    }

    /* Allocate memory for the attitude data */
    *attitude_data = malloc(sizeof(**attitude_data) * (*attitude_size));
    if (*attitude_data == NULL)
    {
        IAS_LOG_ERROR("Allocating attitude_data");
        ias_l0r_close(l0r_ptr);
        free(*ephemeris_data);
        return ERROR;
    }

    /* Get the attitude data */
    status = ias_l0r_get_ancillary_attitude(l0r_ptr, 0, *attitude_size, 
        *attitude_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading L0R attitude from file %s", l0r_filename);
        ias_l0r_close(l0r_ptr);
        free(*ephemeris_data);
        free(*attitude_data);
        return ERROR;
    }

    /* Get the IMU record count */
    status = ias_l0r_get_ancillary_imu_records_count(l0r_ptr, imu_size);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving L0R IMU size from file %s", l0r_filename);
        ias_l0r_close(l0r_ptr);
        free(*ephemeris_data);
        free(*attitude_data);
        return ERROR;
    }

    /* Verify there are records available */
    if (*imu_size <= 0)
    {
        IAS_LOG_ERROR("No L0R IMU records in file %s", l0r_filename);
        ias_l0r_close(l0r_ptr);
        free(*ephemeris_data);
        free(*attitude_data);
        return ERROR;
    }

    /* Allocate memory for the IMU data */
    *imu_data = malloc(sizeof(**imu_data) * (*imu_size));
    if (*imu_data == NULL)
    {
        IAS_LOG_ERROR("Allocating attitude_data from file %s", l0r_filename);
        ias_l0r_close(l0r_ptr);
        free(*ephemeris_data);
        free(*attitude_data);
        return ERROR;
    }

    /* Get the IMU data */
    status = ias_l0r_get_ancillary_imu(l0r_ptr, 0, *imu_size, *imu_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading L0R IRU from file %s", l0r_filename);
        ias_l0r_close(l0r_ptr);
        free(*ephemeris_data);
        free(*attitude_data);
        free(*imu_data);
        return ERROR;
    }

    /* Open read access to the L0R header */
    status = ias_l0r_open_header(l0r_ptr, IAS_READ);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Opening read access to L0R header from file %s", 
            l0r_filename);
        ias_l0r_close(l0r_ptr);
        free(*ephemeris_data);
        free(*attitude_data);
        free(*imu_data);
        return ERROR;
    }

    /* Get the OLI frame header record count */
    status = ias_l0r_get_oli_frame_headers_records_count(l0r_ptr,
        &oli_frame_header_size);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving OLI L0R frame header size from file %s", 
            l0r_filename);
        ias_l0r_close(l0r_ptr);
        free(*ephemeris_data);
        free(*attitude_data);
        free(*imu_data);
        return ERROR;
    }

    /* Allocate memory for the OLI frame header data */
    if (oli_frame_header_size > 0)
    {
        oli_frame_header = malloc(sizeof(*oli_frame_header) 
            * oli_frame_header_size);
        if (oli_frame_header == NULL)
        {
            IAS_LOG_ERROR("Allocating OLI frame header");
            ias_l0r_close(l0r_ptr);
            free(*ephemeris_data);
            free(*attitude_data);
            free(*imu_data);
            return ERROR;
        }

        /* Get the OLI frame headers */
        status = ias_l0r_get_oli_frame_headers(l0r_ptr, 0, 
            oli_frame_header_size, oli_frame_header);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading OLI L0R frame header from file %s", 
                l0r_filename);
            ias_l0r_close(l0r_ptr);
            free(*ephemeris_data);
            free(*attitude_data);
            free(*imu_data);
            return ERROR;
        }

        /* Find the first valid frame header */
        first_valid = 0;
        while (first_valid < oli_frame_header_size)
        {
            if ((oli_frame_header[first_valid].frame_status 
                    & IAS_L0R_VERIFIED) == IAS_L0R_VERIFIED)
                break;
            first_valid++;
        }
    
        last_valid = oli_frame_header_size - 1;
        while (last_valid >= 0 )
        {
            if ((oli_frame_header[last_valid].frame_status 
                    & IAS_L0R_VERIFIED) == IAS_L0R_VERIFIED)
                break;
            last_valid--;
        }
    
        if (last_valid < 0 || first_valid >= oli_frame_header_size 
            || first_valid >= last_valid)
        {
            IAS_LOG_ERROR("Finding OLI start times for imagery from file %s", 
                l0r_filename);
            ias_l0r_close(l0r_ptr);
            free(oli_frame_header);
            free(*ephemeris_data);
            free(*attitude_data);
            free(*imu_data);
            return ERROR;
        }
    
        IAS_LOG_DEBUG("OLI Scene bounds");

        oli_start_total_secs = IAS_L0R_CONVERT_TIME_TO_SECONDS_SINCE_J2000(
            oli_frame_header[first_valid].l0r_time);
        IAS_LOG_DEBUG("Start %f ", oli_start_total_secs);

        oli_end_total_secs = IAS_L0R_CONVERT_TIME_TO_SECONDS_SINCE_J2000(
            oli_frame_header[last_valid].l0r_time);
        IAS_LOG_DEBUG("End   %f ", oli_end_total_secs);

        /* Free the OLI frame header */
        free(oli_frame_header);
        oli_frame_header = NULL;
    }

    /* Get the TIRS frame header record count */
    status = ias_l0r_get_tirs_frame_headers_records_count(l0r_ptr,
        &tirs_frame_header_size);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving TIRS L0R frame header size from file %s", 
            l0r_filename);
        ias_l0r_close(l0r_ptr);
        free(*ephemeris_data);
        free(*attitude_data);
        free(*imu_data);
        return ERROR;
    }

    /* Allocate memory for the TIRS frame header data */
    if (tirs_frame_header_size > 0)
    {
        tirs_frame_header = malloc(sizeof(*tirs_frame_header) 
            * tirs_frame_header_size);
        if (tirs_frame_header == NULL)
        {
            IAS_LOG_ERROR("Allocating TIRS frame header");
            ias_l0r_close(l0r_ptr);
            free(*ephemeris_data);
            free(*attitude_data);
            free(*imu_data);
            return ERROR;
        }

        /* Get the TIRS frame headers */
        status = ias_l0r_get_tirs_frame_headers(l0r_ptr, 0, 
            tirs_frame_header_size, tirs_frame_header);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading TIRS L0R frame header from file %s", 
                l0r_filename);
            ias_l0r_close(l0r_ptr);
            free(tirs_frame_header);
            free(*ephemeris_data);
            free(*attitude_data);
            free(*imu_data);
            return ERROR;
        }

        /* Find the first valid frame header */
        first_valid = 0;
        while (first_valid < tirs_frame_header_size)
        {
            if ((tirs_frame_header[first_valid].frame_status 
                    & IAS_L0R_VERIFIED) == IAS_L0R_VERIFIED)
                break;
            first_valid++;
        }

        last_valid = tirs_frame_header_size - 1;
        while (last_valid >= 0 )
        {
            if ((tirs_frame_header[last_valid].frame_status 
                    & IAS_L0R_VERIFIED) == IAS_L0R_VERIFIED)
                break;
            last_valid--;
        }

        if (last_valid < 0 || first_valid >= tirs_frame_header_size 
            || first_valid >= last_valid)
        {
            IAS_LOG_ERROR("Finding TIRS start times for imagery from file %s", 
                l0r_filename);
            ias_l0r_close(l0r_ptr);
            free(tirs_frame_header);
            free(*ephemeris_data);
            free(*attitude_data);
            free(*imu_data);
            return ERROR;
        }

        IAS_LOG_DEBUG("TIRS Scene bounds");
    
        tirs_start_total_secs = IAS_L0R_CONVERT_TIME_TO_SECONDS_SINCE_J2000(
            tirs_frame_header[first_valid].l0r_time);
        IAS_LOG_DEBUG("Start %f ", tirs_start_total_secs);

        tirs_end_total_secs = IAS_L0R_CONVERT_TIME_TO_SECONDS_SINCE_J2000(
            tirs_frame_header[last_valid].l0r_time);
        IAS_LOG_DEBUG("End   %f ", tirs_end_total_secs);

        /* Free the TIRS frame header */
        free(tirs_frame_header);
        tirs_frame_header = NULL;
    }

    /* If there exists both OLI and TIRS data, figure out largest window
       to encompass both */
    if ((oli_frame_header_size > 0) && (tirs_frame_header_size > 0))
    {
        if (oli_start_total_secs < tirs_start_total_secs)
            start_total_secs = oli_start_total_secs;
        else
            start_total_secs = tirs_start_total_secs;
       
        if (oli_end_total_secs < tirs_end_total_secs)
            end_total_secs = tirs_end_total_secs;
        else
            end_total_secs = oli_end_total_secs;
    }
    else if (oli_frame_header_size > 0)
    {
            start_total_secs = oli_start_total_secs;
            end_total_secs = oli_end_total_secs;
    }
    else if (tirs_frame_header_size > 0)
    {
            start_total_secs = tirs_start_total_secs;
            end_total_secs = tirs_end_total_secs;
    }
    else
    {
        IAS_LOG_ERROR("No OLI or TIRS start and end times for file %s", 
            l0r_filename);
        ias_l0r_close(l0r_ptr);
        free(*ephemeris_data);
        free(*attitude_data);
        free(*imu_data);
        return ERROR;
    }

    IAS_LOG_DEBUG("Scene bounds: start %f seconds, end %f seconds",
        start_total_secs, end_total_secs);

    /* Initialize the leap seconds */
    if (ias_math_init_leap_seconds(start_total_secs, leap_seconds_data))
    {
        IAS_LOG_ERROR("Computing leap second conversion for l0r_header from "
            "file %s", l0r_filename);
        ias_l0r_close(l0r_ptr);
        free(*ephemeris_data);
        free(*attitude_data);
        free(*imu_data);
        return ERROR;
    }

    status = ias_math_convert_j2000_seconds_to_year_doy_sod(start_total_secs, 
        adp_data->image_start_time);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Converting J2000 seconds %lf to Year, DOY, SOD format",
            start_total_secs);
        ias_l0r_close(l0r_ptr);
        free(*ephemeris_data);
        free(*attitude_data);
        free(*imu_data);
        return ERROR;
    }

    status = ias_math_convert_j2000_seconds_to_year_doy_sod(end_total_secs, 
        adp_data->image_stop_time);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Converting J2000 seconds %lf to Year, DOY, SOD format",
            end_total_secs);
        ias_l0r_close(l0r_ptr);
        free(*ephemeris_data);
        free(*attitude_data);
        free(*imu_data);
        return ERROR;
    }

    /* Close the L0R file */
    status = ias_l0r_close(l0r_ptr);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Closing L0R from file %s", l0r_filename);
        free(*ephemeris_data);
        free(*attitude_data);
        free(*imu_data);
        return ERROR;
    }

    return SUCCESS;
}

