/*****************************************************************************
Name: read_parameters

Purpose: Read the application parameter file, storing values in the 
         ADP_DATA structure

Returns: Type = int
    Value    Description
    -----    -----------
    SUCCESS  Successful completion
    ERROR    Operation failed
*****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "ias_const.h"
#include "ias_logging.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "ancillary_data_preprocessor.h"

#define DEFAULT_ANCILLARY_FILENAME "ancillary_file.h5"

int read_parameters
(
    const char *parm_filename,  /* I: Parameter filename */
    ADP_DATA *adp_data          /* O: Parameters retrieved from parameter
                                      file, OMF and CPF */
)
{
    int read_count = 0;         /* Parameter table size */
    int status = 0;             /* Function return status */
    int satellite_number;       /* Satellite number being processed */

    /* Set up and read the application parameter table */
    IAS_PARM_DECLARE_TABLE(app_parm_table, 3);

    /* Get the work order ID */
    IAS_PARM_WORK_ORDER_ID(app_parm_table, adp_data->work_order_id,
            sizeof(adp_data->work_order_id), 0);

    /* Get the ancillary output filename */
    const char *default_ancillary_filename[] = {DEFAULT_ANCILLARY_FILENAME};
    IAS_PARM_ADD_STRING(app_parm_table, ANCILLARY_FILENAME,
            "Output ancillary data filename", IAS_PARM_OPTIONAL,
            0, NULL, 1, default_ancillary_filename, 
            adp_data->anc_filename,
            sizeof(adp_data->anc_filename), 1);

    /* Get the flag to write (or not write) data to the database */
    IAS_PARM_STORE_CHARACTERIZATION_IN_DB(app_parm_table, IAS_PARM_OPTIONAL,
            &adp_data->characterization_flag, 
            sizeof(adp_data->characterization_flag), 0);

    IAS_LOG_DEBUG("Reading application parameters from file %s",
            parm_filename);

    read_count  = IAS_PARM_GET_TABLE_SIZE(app_parm_table);
    /* Call the routine to provide help for the parameters if requested 
       also verifies the file exists */
    status = ias_parm_provide_help(parm_filename, app_parm_table, read_count,
            IAS_INPUT_PARAMETERS);
    if (status) /* either help was provided or an error occurred */
    {
        if (status == 1)
        {
            /* If the status of the provide help is 1 the user asked for help.
               Since the parameter file name contains the help request, copy it
               to the OMF file name so the next call will ask for the same 
               help. */
            strcpy(adp_data->omf_filename, parm_filename);
        }
        else
        {
            IAS_LOG_ERROR("Providing application parameter file help");
            return ERROR;
        }
    }
    else
    {
        status = ias_parm_read(parm_filename, app_parm_table, read_count);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading application parameters from file %s",
                    parm_filename);
            return ERROR;
        }

        /* Build the OMF name from the work order ID */
        sprintf(adp_data->omf_filename, "%s.omf", adp_data->work_order_id);
    }

    /* Set up and read the OMF parameter table */
    IAS_PARM_DECLARE_TABLE(omf_parm_table, 5);

    /* Get the CPF */
    IAS_PARM_CPF(omf_parm_table, IAS_PARM_REQUIRED,
            adp_data->cpf_filename,
            sizeof(adp_data->cpf_filename), 1);

    /* Get the LOR filename */
    IAS_PARM_L0R_FILENAME(omf_parm_table, IAS_PARM_REQUIRED,
            adp_data->l0r_filename, sizeof(adp_data->l0r_filename), 1);

    /* Get the acquisition type */
    IAS_PARM_ACQUISITION_TYPE(omf_parm_table, &adp_data->acquisition_type,
            sizeof(adp_data->acquisition_type), 1);

    /* Add the satellite number parameter */
    IAS_PARM_SATELLITE(omf_parm_table, &satellite_number,
            sizeof(satellite_number), 1);

    IAS_LOG_DEBUG("Reading OMF parameters from file %s",
            adp_data->omf_filename);

    read_count  = IAS_PARM_GET_TABLE_SIZE(omf_parm_table);
    /* Call the routine to provide help for the parameters if requested 
       also verifies the file exists */
    status = ias_parm_provide_help(adp_data->omf_filename, omf_parm_table,
            read_count, IAS_OMF_PARAMETERS);
    if (status)
    {
        if (status == 1) /* help was successful */
            exit(EXIT_SUCCESS);
        else
        {
            IAS_LOG_ERROR("Providing OMF file help");
            return ERROR;
        }
    }

    /* Read the OMF file */
    status = ias_parm_read(adp_data->omf_filename, omf_parm_table, read_count);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading OMF parameters from file %s",
                adp_data->omf_filename);
        return ERROR;
    }

    /* Get the satellite ID from the satellite number provided */
    adp_data->satellite_id =
        ias_sat_attr_get_satellite_id_from_satellite_number(satellite_number);

    if (adp_data->satellite_id != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to determine satellite ID from satellite "
                "number %d", satellite_number);
        return ERROR;
    }

    return SUCCESS;
}

