#!/usr/bin/perl -w

# Test script for ADP

use strict;
use ias_const;

my $command;
my $data_dir = "$ENV{IAS_UNIT_TEST_DATA_DIR}";
my $enable_db_tests_set = 0;
my $parm_file_db = "testWO_1.odl";
my $parm_file_no_db = "testWO_1_no_db.odl";
my $parm_file;
my $omf_file = "WO1299.omf";
my $omf_file_ref = "WO1299.omf.ref";
my $omf_file_verify = "WO1299.omf.verify";
my $omf_file_verify_ref = "WO1299.omf.verify.ref";
my $hdf_file = "ancillary_file.h5";
my $hdf_file_ref = $data_dir . "/ias_base/gps/ancillary_data_preprocessor/ancillary_file.h5.ref";
my $db_dump_file_1 = "GEOMETRIC_WORK_ORDER_COMMON.db.results";
my $db_dump_file_verify_1 = "GEOMETRIC_WORK_ORDER_COMMON.db.verify";
my $db_dump_file_2 = "MODEL_GENERATION.db.results";
my $db_dump_file_verify_2 = "MODEL_GENERATION.db.verify";
my $rc;

# clean up old file(s)
unlink ("$omf_file");

# Check whether or not the ENABLE_DB_TESTS environment variable has been set.
if (defined $ENV{ENABLE_DB_TESTS})
{
    # if it's been set, delete any old database records
    $command = "./delete_schema.sh";
    $rc = system ($command);
    if ($rc != 0)
    {
        print "Cleanup of old ADP database records failed\n";
        exit FAILURE;
    }

    # set a local flag indicating ENABLE_DB_TESTS has been set
    $enable_db_tests_set = 1;

    # use the parameter file with database writes enabled
    $parm_file = $parm_file_db;
}
else
{
    # print a message indicating database tests are disabled
    print "\n";
    print __FILE__ . ":  Database tests are disabled.\n";
    print "\n";

    # set a local flag indicating ENABLE_DB_TESTS hasn't been set
    $enable_db_tests_set = 0;

    # use the parameter file with database writes disabled
    $parm_file = $parm_file_no_db;
}

# make OMF files usable, inserting test data directory
$command = "cp $omf_file_ref $omf_file";
$rc = system ($command);
if ($rc != 0)
{
    print "OMF file copy failed.\n";
    exit FAILURE;
}
$command = "cp $omf_file_verify_ref $omf_file_verify";
$rc = system ($command);
if ($rc != 0)
{
    print "OMF verification file copy failed.\n";
    exit FAILURE;
}

# escape the backslashes
$data_dir =~ s/\//\\\//g;

# call sed to replace the string with the actual test data directory
my $sed_replace_string = "s/IAS_UNIT_TEST_DATA_DIR/$data_dir/";
$rc = system("sed", "-i", $sed_replace_string, $omf_file);
if ($rc != 0)
{
    print "SED replacement of string '$sed_replace_string' in $omf_file "
        . "failed.\n";
    exit FAILURE;
}
$rc = system("sed", "-i", $sed_replace_string, $omf_file_verify);
if ($rc != 0)
{
    print "SED replacement of string '$sed_replace_string' in "
        . "$omf_file_verify failed.\n";
    exit FAILURE;
}

# run the ADP with the test odl file
$command = "../ancillary_data_preprocessor $parm_file";
$rc = system ($command);
if ($rc != 0)
{
    print "Call to ADP application failed.\n";
    exit FAILURE;
}

# compare HDF output files, allowing minor differences since they can 
# happen normally between 32 and 64-bit builds
$command = "h5diff -p 0.00000001 $hdf_file $hdf_file_ref";
print "Running: $command\n";
$rc = system ($command);
if ($rc != 0)
{
    print "ADP HDF diff test failed.\n";
    exit FAILURE;
}
else
{
    print "ADP HDF diff test passed.\n";
}

# compare OMF files
$command = "diff -b --brief $omf_file $omf_file_verify";
$rc = system ($command);
if ($rc != 0)
{
    print "ADP OMF diff test failed.\n";
    exit FAILURE;
}
else
{
    print "ADP OMF diff test passed.\n";
}

# if database tests are disabled, exit here with a success status
if ($enable_db_tests_set == 0)
{
    print "Non-database ADP tests passed.\n";
    exit SUCCESS;
}

# dump the database records
$command = "./dump_schema.sh";
$rc = system ($command);
if ($rc != 0)
{
    print "ADP database dump failed.\n";
    exit FAILURE;
}
else
{
    print "ADP database dump passed.\n";
}

# delete the database records

$command = "./delete_schema.sh";
$rc = system ($command);
if ($rc != 0)
{
    print "ADP database cleanup failed.\n";
    exit FAILURE;
}
else
{
    print "ADP database cleanup passed.\n";
}

# diff the database files
$command = "diff --brief $db_dump_file_1 $db_dump_file_verify_1";
my $rc_diff1 = system ($command);
if ($rc_diff1 != 0)
{
    print "ADP diff $db_dump_file_1 failed.\n";
}

$command = "diff --brief $db_dump_file_2 $db_dump_file_verify_2";
my $rc_diff2 = system ($command);
if ($rc_diff2 != 0)
{
    print "ADP diff $db_dump_file_2 failed.\n";
}

if (($rc_diff1 != 0) || ($rc_diff2 != 0))
{
    print "ADP database diffs failed.\n";
    exit FAILURE;
}
else
{
    print "ADP database diffs passed.\n";
}

# exit with a SUCCESS status
print "All ADP tests passed.\n";
exit SUCCESS;
