/******************************************************************************
NAME:           b2b_char

PURPOSE:  To implement a band to band characterization between the reference
and search bands.

RETURN VALUE:   Type = int
Value         Description
-----         -----------
EXIT_SUCCESS  Successful completion
EXIT_FAILURE  Operation failed

NOTES:
The b2b_char application ultimately performs a statistical measure of all the
geometric shifts relating the measured offsets among all bands (T-distribution),
but this portion of the application does not implement the final outlier
detection.  That is done in the Tdist code.  This routine is used to do the
correlation, then output those results to a file.  Tdist will read the results
from the correlation and implement the T-distribution final outlier detection.

If no bands are specified, all LDCM bands will be used.

ALGORITHM REFERENCES:
See ias_gcp.h for a description of the IAS_GCP_RECORD structure.
See ias_l1g.h for a description of the band metadata structure.
See ias_structures.h for a description of the IAS_LNG_LS structure.

******************************************************************************/
#include <stdlib.h>
#include <string.h>

#include "b2b_char.h"
#include "ias_logging.h"              /* IAS logging library          */
#include "ias_satellite_attributes.h" /* Satellite attributes library */

int main (int argc, char *argv[])
{
    /* Variable declarations.  */
    int i;                           /* Counter                               */
    int stat;                        /* Status return value                   */
    int nbands;                      /* # of bands                            */
    int nscas;                       /* # of SCA indexes in sca_indexes       */
    int nline_pts;                   /* # GCP points in line direction        */
    int nsamp_pts;                   /* # GCP points in the sample direction  */
    int numgcps[IAS_MAX_NSCAS];      /* Total # GCP points                    */
    int abs_corr_coeff;              /* flag to use abs value of correlation
                                        coefficients in grey correlation      */
    enum TIEPT_TYPE tiept_type;      /* Variable to specify the type of tiepts
                                        either regular spaced tiepoints or 
                                        grid based tiepoints                  */
    float min_corr;                  /* Minimum correlation strength          */
    float max_disp;                  /* Maximum correlation displacement      */
    float invalid_threshold;         /* Invalid threshold precentage for  
                                        buffer                                */
    float valid_image_max;           /* DNs above this are invalid            */
    float valid_image_min;           /* DNs below this are invalid            */
    IAS_CORRELATION_FIT_TYPE fit_method; /* Correlation result fit method     */
    int band_indexes[IAS_MAX_NBANDS];/* indexes of bands for the correlation  */
    int sca_indexes[IAS_MAX_NSCAS];  /* SCA indexes for the correlation       */
    struct IAS_LNG_LS chip_size;     /* Size of the chip for correlation      */
    char parameter_file[PATH_MAX];   /* input parameter filename              */
    char filename_1G[PATH_MAX];      /* 1G image filename                     */
    char filename_0R[PATH_MAX];      /* 0R image filename                     */
    char output_filename[PATH_MAX];  /* Name of B2B output report file        */
    char grid_filename[PATH_MAX];    /* Name of grid file                     */
    char cpf_filename[PATH_MAX];     /* Name of Calibration parameter file    */
    char omf_file_name[PATH_MAX];    /* Name of omf file                      */
    char work_order_id[IAS_WORKORDER_ID_SIZE]; /* Work order ID string      */
    IAS_L1G_BAND_METADATA band_metadata[IAS_MAX_NBANDS]; /* Array of band 
                                        metadata                              */
    int nl;                          /* Number of lines                       */
    int ns;                          /* Number of samples                     */
    float range[2];                  /* Valid pixel range                     */
    int reduce_flag;                 /* Flag for creating reduced image file  */
    IAS_GCP_RECORD *sca_gcp[IAS_MAX_NSCAS]; /* Array of GCP points for 30m 
                                        band resolution                       */
    int pan_band_number;             /* PAN band number for the satellite     */
    int pan_band_index;              /* PAN band index for the satellite      */


    /* Initialize the logging library */
    if (ias_log_initialize("b2b_char") != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing logging library");
        exit(EXIT_FAILURE);
    }

    /* Check the arguments and get the parameter_file  from the arguments. */
    if (argc != 2)
    {
        IAS_LOG_ERROR("Usage: b2b_char parameter_file");
        exit(EXIT_FAILURE);
    }
    else
        strcpy(parameter_file, argv[1]);

    /* Get the b2b_char parameters. */
    stat = get_parms(parameter_file, filename_1G, filename_0R, output_filename, 
        grid_filename, cpf_filename, &reduce_flag, band_indexes, &nbands, 
        sca_indexes, &nscas, &nline_pts, &nsamp_pts, &chip_size, &tiept_type, 
        &min_corr, &max_disp, range, &invalid_threshold, &fit_method, 
        &abs_corr_coeff, work_order_id, omf_file_name); 
    if (stat == ERROR)
    {
        IAS_LOG_ERROR("Reading the parameters from the ODL file.");
        exit(EXIT_FAILURE);
    }

    /* Get the PAN band number. */
    pan_band_number = ias_sat_attr_get_band_number_from_type(IAS_PAN_BAND);
    if ((pan_band_number < 1) || (pan_band_number > IAS_MAX_NBANDS))
    {
        IAS_LOG_ERROR("Invalid PAN band number %d.", pan_band_number);
        exit(EXIT_FAILURE);
    }

    /* Get the PAN band index. */
    pan_band_index = ias_sat_attr_convert_band_number_to_index(pan_band_number);
    if ((pan_band_index < 0) || (pan_band_index > IAS_MAX_NBANDS - 1))
    {
        IAS_LOG_ERROR("Invalid PAN band index %d.", pan_band_index);
        exit(EXIT_FAILURE);
    }

    /* Print the starting message.  */
    IAS_LOG_INFO("Initiate ...");

    valid_image_min = range[0];
    valid_image_max = range[1];

    /* Initialize SCA GCP records structure. */
    for (i = 0; i < IAS_MAX_NSCAS; i++) 
    {
        sca_gcp[i] = NULL;
    }

    /* Read and compare the metadata for the bands. */
    stat = compare_metadata(filename_1G, band_indexes, nbands, band_metadata,
        &nl, &ns);
    if (stat == ERROR)
    {
        IAS_LOG_ERROR("Reading and comparing the band metadata.");
        exit(EXIT_FAILURE);
    }

    if (reduce_flag)
    {
        /* Check if a band in the list is panchromatic. */
        for (i = 0; i < nbands; i++)
        {
            if (band_indexes[i] == pan_band_index)
                break;
        }

        /* Downsample the image if a band on the band list is panchromatic. */
        if (i != nbands)
        {
            IAS_LOG_INFO("Perform downsampling ...");
            stat = downsample(filename_1G, band_indexes, nbands); 
            if (stat == ERROR)
            {
                IAS_LOG_ERROR("Downsampling the image.");
                exit(EXIT_FAILURE);
            }
        }
    }
    else
    {
        /* This should only happen during manual testing. */
        IAS_LOG_WARNING("Not creating reduced image file!\n");
    }

    IAS_LOG_INFO("Creating tiepoints ...");
    /* Create tiepoints for the band combinations.  If a grid file is given
       choose regular spaced GCPs by stepping through the grid.  If no grid 
       file was given choose regular spaced GCPs by stepping through the 
       output image frame.  */
    if (tiept_type == GRID_BASED)
    {
        stat = create_tiepts_grid(band_indexes, nbands, sca_indexes, nscas,
            band_metadata, nline_pts, nsamp_pts, chip_size, grid_filename,
            pan_band_index, numgcps, sca_gcp);

        if (stat != SUCCESS)
        {
            IAS_LOG_ERROR("Creating the regular-spaced tiepoints.");
            exit(EXIT_FAILURE);
        }
    }
    else /* tiept type is REGULAR_SPACE */
    {
        for (i = 0; i < IAS_MAX_NSCAS; i++ )
        {
            numgcps[i] = nline_pts * nsamp_pts;

            stat = create_tiepts(band_indexes, nbands, band_metadata, nl, ns, 
                nline_pts, nsamp_pts, chip_size, pan_band_index, &sca_gcp[i]);  
            if (stat == ERROR)
            {
                IAS_LOG_ERROR("Creating the regular-spaced tiepoints.");
                exit(EXIT_FAILURE);
            }
        }
    }

    IAS_LOG_INFO("Perform correlation ...");
    /* Implement the correlation on unique band pairs.  */
    stat = b2b_corr(sca_gcp, numgcps, filename_1G, filename_0R, band_indexes, 
        nbands, band_metadata, sca_indexes, nscas, output_filename, min_corr,
        max_disp, fit_method, abs_corr_coeff, invalid_threshold,
        valid_image_max, valid_image_min, work_order_id);
    if (stat == ERROR)
    {
        IAS_LOG_ERROR("Correlating the bands.");
        for (i = 0; i < IAS_MAX_NSCAS; i++)
        {
            free(sca_gcp[i]);
        }
        exit(EXIT_FAILURE);
    }

    /* Free the used pointers.  */
    for (i = 0; i < IAS_MAX_NSCAS; i++)
    {
        free(sca_gcp[i]);
    }

    /* update omf file with output file name */
    if (update_omf(omf_file_name, output_filename) != SUCCESS)
    {
        IAS_LOG_ERROR("Updating omf file.");
        exit(EXIT_FAILURE);
    }

    /* Print the completion message.  */
    IAS_LOG_INFO("Successful completion.");

    return EXIT_SUCCESS;
} /* done */
