/******************************************************************************
NAME:       b2b_corr

PURPOSE:  This routine will implement the band-to-band correlation on the
          unique band pairs.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
This routine will open the required images for each unique band pair.  Then
the correlation will be implemented for each GCP point in the images.  The
final results will be output to a single file with all band pair combination
results.

ALGORITHM REFERENCES:

******************************************************************************/
#include <stdlib.h>
#include <string.h>
#include "ias_math.h"
#include "ias_math_parallel_corr.h"   /* Parallel correlator          */
#include "ias_logging.h"              /* IAS logging library          */
#include "ias_miscellaneous.h"        /* Miscellaneous libraries      */
#include "ias_satellite_attributes.h" /* Satellite attributes library */
#include "b2b_char.h"

int b2b_corr 
(
    IAS_GCP_RECORD *sca_gcp[IAS_MAX_NSCAS], /* I: Array of GCP points        */
    const int *numgcps,           /* I: Number of GCPs                       */
    const char *l1g_img_name,     /* I: Name of 1G image                     */
    const char *l0r_img_name,     /* I: Name of 0R image                     */
    const int *band_indexes,      /* I: Indexes of bands for the correlation */
    const int nbands,             /* I: Number of band indexes               */
    const IAS_L1G_BAND_METADATA band_metadata[IAS_MAX_NBANDS],
                                  /* I: Array of band data                   */
    const int sca_indexes[IAS_MAX_NSCAS],/* I: SCA indexes to search         */
    const int nscas,              /* I: Number of SCA indexes                */
    const char *output_filename,  /* I: Name of output data file             */ 
    const float min_corr,         /* I: Minimum correlation strength         */
    const float max_disp,         /* I: Maximum correlation displacement     */
    const IAS_CORRELATION_FIT_TYPE fit_method, /* I: Correlation result fit 
                                        method                               */
    const int abs_corr_coeff,     /* I: flag to use abs value of correlation
                                        coefficients in grey correlation     */
    const float invalid_threshold,/* I: Invalid threshold percentage for 
                                        buffer                               */
    const float valid_image_max,  /* I: DNs above this are invalid           */
    const float valid_image_min,  /* I: DNs below this are invalid           */
    const char *work_order_id     /* I: Work Order ID for report header      */
)
{
    IAS_GCP_RECORD *gcplib;         /* list of GCPs (either gcp30 or gcp60) */
    IAS_GCP_RESULTS *gcp;           /* GCP data structures (1 per pt)       */
    L1GIO *ref_img_ptr;             /* reference 1G image pointer           */
    L1GIO *srch_img_ptr;            /* search 1G image pointer              */
    FILE *output;                   /* GCP output ASCII file pointer        */
    IAS_DATA_TYPE ref_dtype;        /* reference data type                  */
    IAS_DATA_TYPE srch_dtype;       /* search data type                     */
    char rimg_name[PATH_MAX];       /* reference image name                 */
    char simg_name[PATH_MAX];       /* search image name                    */
    int ref_nsca;                   /* number of SCAs in reference image    */ 
    int srch_nsca;                  /* number of SCAs in search image       */
    int ref_nl;                     /* number of lines in reference image   */
    int ref_ns;                     /* number of samps in reference image   */
    int srch_nl;                    /* number of lines in search image      */
    int srch_ns;                    /* number of samps in search image      */
    int i;                          /* looping variables                    */
    int j;                          /* looping variables                    */
    int curr;                       /* current GCP                          */
    int curref;                     /* looping variable for reference bands */
    int cursrch;                    /* looping variable for search bands    */
    int curgcp;                     /* looping variable for GCPs            */
    int rref;                       /* reference band reference position    */
    int sref;                       /* search band reference position       */
    int sca_index;                  /* SCA index                            */ 
    int sca_number;                 /* SCA number; 0 indicates SCA-combined */
    int sca_loop_idx;               /* Loop index into SCA list             */
    int num_correlated;             /* number of GCPs that correlated       */
    int status;                     /* function return status               */
    int done[IAS_MAX_NBANDS][IAS_MAX_NBANDS]; /* array of flags to specify 
                                       the correlation has been done for 
                                       this band combination                */
    float mean_offset[2];           /* mean of the deltas for the current
                                       band combination (line, sample)      */
    double *offset[2];              /* 2D array to store the offsets for the
                                       standard deviation calculation       */
    double stddev_offset[2];        /* std dev of the deltas for the current
                                       band combination (line, sample)      */
    int maxgcps;                    /* largest number of GCPs across SCAs   */
    IAS_REPORT_HEADER *header = NULL;/* report header structure              */
    int reference_band_number;      /* Reference band                       */
    int search_band_number;         /* Search band                          */
    double reference_pixel_resolution;/* Pixel resolution of reference band */
    double search_pixel_resolution; /* Pixel resolution of search band      */
    IAS_PARALLEL_CORRELATOR_TYPE *correlator_ptr = NULL;
                                    /* pointer to the parallel correlator
                                       data structure                       */
    IAS_CORRELATION_RESULT_TYPE *result_ptr = NULL;
                                    /* pointer to the correlation results   */
    int max_chip_lines = 0;         /* maximum number of lines in a chip    */
    int max_chip_samples = 0;       /* maximum number of samples in a chip  */


    /* Allocate space for GCP data (one point per gcplib point).  */
    maxgcps = numgcps[sca_indexes[0]];
    for (i = 0; i < nscas; i++)
    {
        if (maxgcps < numgcps[sca_indexes[i]])
        {
            maxgcps = numgcps[sca_indexes[i]];
        }
    }

    gcp = malloc(maxgcps * sizeof(IAS_GCP_RESULTS));
    if (gcp == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for %d GCP data points.", maxgcps);
        return ERROR;
    }

    /* Open the GCP output file for writing.  */
    output = fopen(output_filename, "wt");
    if (output == NULL)
    {
        IAS_LOG_ERROR("Opening the GCP data output file %s.", output_filename);
        free(gcp);
        return ERROR;
    }

    /* Output the standard header to the file.  */
    header = ias_misc_initialize_gps_report_header(l1g_img_name, l0r_img_name,
            work_order_id, "B2B_CHAR");
    if (!header)
    {
        IAS_LOG_ERROR("Initializing GPS file header.");
        free(gcp);
        fclose(output);
        return ERROR;
    }

    /* Write the header information to the report file */
    status = ias_misc_write_gps_report_header(output, header);
    ias_misc_free_report_header(header);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing to the GPS output file.");
        free(gcp);
        fclose(output);
        return ERROR;
    }

    /* Output the image and band numbers to the file. */
    if (fprintf(output, "Reference image: %s\n"
                        "Reference bands: ", l1g_img_name) < 0)
    {
        IAS_LOG_ERROR("Writing to the output file.");
        free(gcp);
        fclose(output);
        return ERROR;
    }

    for (i = 0; i < nbands; i++)
    {
        reference_band_number = ias_sat_attr_convert_band_index_to_number(
            band_indexes[i]);
        if ((reference_band_number < 1) || 
            (reference_band_number > IAS_MAX_NBANDS))
        {
            IAS_LOG_ERROR("Invalid band number %d.", reference_band_number);
            free(gcp);
            fclose(output);
            return ERROR;
        }
            
        if (fprintf(output, " %d", reference_band_number) < 0)
        {
            IAS_LOG_ERROR("Writing to the output file.");
            free(gcp);
            fclose(output);
            return ERROR;
        }
    }

    if (fprintf(output, "\nSearch bands: ") < 0)
    {
        IAS_LOG_ERROR("Writing to the output file.");
        free(gcp);
        fclose(output);
        return ERROR;
    }

    for (i = 0; i < nbands; i++)
    {
        search_band_number = ias_sat_attr_convert_band_index_to_number(
            band_indexes[i]);
        if ((search_band_number < 1) || (search_band_number > IAS_MAX_NBANDS))
        {
            IAS_LOG_ERROR("Invalid band number %d.", search_band_number);
            free(gcp);
            fclose(output);
            return ERROR;
        }

        if (fprintf(output, " %d", search_band_number) < 0)
        {
            IAS_LOG_ERROR("Writing to the output file.");
            free(gcp);
            fclose(output);
            return ERROR;
        }
    }

    if (fprintf(output,
                "\n\nPoint  Ref_Line    Ref_Sample      Ref_Lat"
                "      Ref_Long     Ref_Elev   Search_Line Search_Sample   "
                "Delta_Y      Delta_X      Flag Coef    "
                "    Ref_Band  Search_Band Ref_SCA Search_SCA            "
                "Ref_Image                                 "
                "  Search_Image\n") < 0)
    {
        IAS_LOG_ERROR("Writing to the output file.");
        free(gcp);
        fclose(output);
        return ERROR;
    }

    /* Find the largest chip size for line and sample. */
    for (sca_loop_idx = 0; sca_loop_idx < nscas; sca_loop_idx++)
    {
        for (curgcp = 0; curgcp < maxgcps; curgcp++)
        {
            if (sca_gcp[sca_loop_idx][curgcp].chip_size_lines > max_chip_lines)
                max_chip_lines = sca_gcp[sca_loop_idx][curgcp].chip_size_lines;
            if (sca_gcp[sca_loop_idx][curgcp].chip_size_samples > 
                max_chip_samples)
                max_chip_samples = 
                    sca_gcp[sca_loop_idx][curgcp].chip_size_samples;
        }
    }

    /* Allocate memory for the results of the correlation. */
    result_ptr = malloc(maxgcps * sizeof(IAS_CORRELATION_RESULT_TYPE));
    if (result_ptr == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for the correlation results");
        free(gcp);
        fclose(output);
        return ERROR;
    }

    /* Initialize the parallel correlator. */
    /* set the check edge and low peak flags*/
    correlator_ptr = ias_math_init_parallel_correlator(1, 0, 1, 0,
        max_chip_lines, max_chip_samples, max_chip_lines, 
        max_chip_samples, result_ptr);
    if (correlator_ptr == NULL)
    {
        IAS_LOG_ERROR("Creating the parallel correlator");
        free(result_ptr);
        free(gcp);
        fclose(output);
        return ERROR;
    }

    /* Loop through the reference bands.  */
    for (sca_loop_idx = 0; sca_loop_idx < nscas; sca_loop_idx++)
    {
        sca_index = sca_indexes[sca_loop_idx];

        /* Set the GCP record pointer. */
        gcplib = sca_gcp[sca_loop_idx];

        /* Initialize the flags.  */
        for (i = 0; i < IAS_MAX_NBANDS; i++)
        {
            for (j = 0; j < IAS_MAX_NBANDS; j++)
            {
                done[i][j] = FALSE;
            }
        }

        for (curref = 0; curref < nbands; curref++)
        {
            /* Look up pixel resolution of the reference band. */
            reference_band_number = ias_sat_attr_convert_band_index_to_number(
                band_indexes[curref]);
            if ((reference_band_number < 1) || 
                (reference_band_number > IAS_MAX_NBANDS))
            {
                IAS_LOG_ERROR("Invalid band number %d.", reference_band_number);
                free(result_ptr);
                free(gcp);
                fclose(output);
                ias_math_close_parallel_correlator(correlator_ptr);
                return ERROR;
            }

            reference_pixel_resolution =
                band_metadata[curref].projection_distance_x;

            /* Get the band reference position for the reference bands.  */
            rref = band_indexes[curref];

            /* Loop through the search bands.  */
            for (cursrch = 0; cursrch < nbands; cursrch++)
            {
                /* Look up pixel resolution of the search band. */
                search_band_number = ias_sat_attr_convert_band_index_to_number(
                    band_indexes[cursrch]);
                if ((search_band_number < 1) || 
                    (search_band_number > IAS_MAX_NBANDS))
                {
                    IAS_LOG_ERROR("Invalid band number %d.",
                        search_band_number);
                    free(result_ptr);
                    free(gcp);
                    fclose(output);
                    ias_math_close_parallel_correlator(correlator_ptr);
                    return ERROR;
                }

                search_pixel_resolution =
                    band_metadata[cursrch].projection_distance_x;

                /* Get the band reference position for the search bands, 
                   initialize the valid band combination flag to FALSE, and 
                   initialize needed values to 0.  */
                sref = band_indexes[cursrch];
                num_correlated = 0;
                mean_offset[0] = 0.0;
                mean_offset[1] = 0.0;

                /* Process the bands if they are a valid combination and the
                   combination has not been previously processed. */
                if (!done[rref][sref] && !done[sref][rref] && rref != sref)
                {
                    if (reference_pixel_resolution == search_pixel_resolution)
                    {
                        /* Implement the correlation on the original image
                           for both reference and search bands. */

                        /* Assign the 1G reference image. */
                        strcpy(rimg_name, l1g_img_name);

                        /* Assign the 1G search image. */
                        strcpy(simg_name, l1g_img_name);
                    }
                    else if (reference_pixel_resolution >
                             search_pixel_resolution)
                    {
                        /* Implement the correlation on the original image
                           for the reference band and the downsampled image
                           for the search band. */

                        /* Assign the 1G reference image. */
                        strcpy(rimg_name, l1g_img_name);

                        /* Assign the 1G search image. */
                        strcpy(simg_name, REDUCE_FILE);
                    }
                    else if (reference_pixel_resolution <
                             search_pixel_resolution)
                    {
                        /* Implement the correlation on the downsampled image
                           for the reference band and the original image
                           for the search band. */

                        /* Assign the 1G reference image. */
                        strcpy(rimg_name, REDUCE_FILE);
 
                        /* Assign the 1G search image. */
                        strcpy(simg_name, l1g_img_name);
                    }

                    IAS_LOG_INFO("Correlating band %d to %d for SCA %d.",
                        reference_band_number, search_band_number,
                        sca_index + 1);

                    /* Open the 1G reference image for extraction of windows. */
                    ref_img_ptr = ias_l1g_open_image(rimg_name, IAS_READ);
                    if (ref_img_ptr == NULL)
                    {
                        IAS_LOG_ERROR("Image %s cannot be opened", rimg_name);
                        free(gcp);
                        fclose(output);
                        free(result_ptr);
                        ias_math_close_parallel_correlator(correlator_ptr);
                        return ERROR;
                    }

                    if (ias_l1g_get_band_size(ref_img_ptr, 
                        reference_band_number, &ref_dtype, &ref_nsca, &ref_nl, 
                        &ref_ns) != SUCCESS)
                    {
                        IAS_LOG_ERROR("Getting band %d size from image %s.",
                            reference_band_number, rimg_name);
                        free(gcp);
                        fclose(output);       
                        free(result_ptr);
                        ias_math_close_parallel_correlator(correlator_ptr);
                        ias_l1g_close_image(ref_img_ptr);
                        return ERROR;
                    }

                    /* Open the 1G search image for extraction of windows. */
                    srch_img_ptr = ias_l1g_open_image(simg_name, IAS_READ);
                    if (srch_img_ptr == NULL)
                    {
                        IAS_LOG_ERROR("Image %s cannot be opened", simg_name);
                        free(gcp);
                        ias_l1g_close_image(ref_img_ptr);
                        fclose(output);
                        free(result_ptr);
                        ias_math_close_parallel_correlator(correlator_ptr);
                        return ERROR;
                    }

                    if (ias_l1g_get_band_size(srch_img_ptr, search_band_number, 
                        &srch_dtype, &srch_nsca, &srch_nl, &srch_ns) 
                        != SUCCESS)
                    {
                        IAS_LOG_ERROR("Getting band %d size from image %s.",
                            search_band_number, simg_name);
                        free(gcp);
                        ias_l1g_close_image(srch_img_ptr);
                        ias_l1g_close_image(ref_img_ptr);
                        fclose(output);
                        free(result_ptr);
                        ias_math_close_parallel_correlator(correlator_ptr);
                        return ERROR;
                    }

                    /* Check the data type for both the search and reference 
                       image. Only valid 1G output types are BYTE, I2, and UI2.
                       Since the search image is the same image (different
                       band) as the reference image, we don't need to verify 
                       the data type for the search image.  */

                    if ((ref_dtype != IAS_BYTE) && (ref_dtype != IAS_I2) &&
                         (ref_dtype != IAS_UI2))
                    {
                        IAS_LOG_ERROR("Invalid 1G reference image data type.  "
                                "Valid type are: BYTE, I2, UI2.");
                        free(gcp);
                        ias_l1g_close_image(srch_img_ptr);
                        ias_l1g_close_image(ref_img_ptr);
                        fclose(output);
                        free(result_ptr);
                        ias_math_close_parallel_correlator(correlator_ptr);
                        return ERROR;
                    }

                    if ((srch_dtype != IAS_BYTE) && (srch_dtype != IAS_I2) &&
                        (srch_dtype != IAS_UI2))
                    {
                        IAS_LOG_ERROR("Invalid 1G search image data type.  "
                                "Valid type are: BYTE, I2, UI2.");
                        free(gcp);
                        ias_l1g_close_image(srch_img_ptr);
                        ias_l1g_close_image(ref_img_ptr);
                        fclose(output);
                        free(result_ptr);
                        ias_math_close_parallel_correlator(correlator_ptr);
                        return ERROR;
                    }

                    /* Verify the number of lines and samples is the same for 
                       both the reference and search images.  */
                    if (ref_nl != srch_nl)
                    {
                        IAS_LOG_ERROR(
                                "Number of lines in reference image is not "
                                "equal to that in search image.");
                        free(gcp);
                        ias_l1g_close_image(srch_img_ptr);
                        ias_l1g_close_image(ref_img_ptr);
                        fclose(output);
                        free(result_ptr);
                        ias_math_close_parallel_correlator(correlator_ptr);
                        return ERROR;
                    }

                    if (ref_ns != srch_ns)
                    {
                        IAS_LOG_ERROR(
                                "Number of samples in reference image is not "
                                "equal to that in search image.");
                        free(gcp);
                        ias_l1g_close_image(srch_img_ptr);
                        ias_l1g_close_image(ref_img_ptr);
                        fclose(output);
                        free(result_ptr);
                        ias_math_close_parallel_correlator(correlator_ptr);
                        return ERROR;
                    }

                    /* Extract the windows and correlate the images.  */
                    status = process_gcp(gcplib, numgcps[sca_index], 
                        ref_img_ptr, srch_img_ptr, band_indexes[curref], 
                        band_indexes[cursrch], sca_index, ref_dtype, srch_dtype,
                        ref_nl, ref_ns, min_corr, max_disp, fit_method, gcp, 
                        abs_corr_coeff, invalid_threshold, valid_image_max, 
                        valid_image_min, correlator_ptr, result_ptr);
                    if (status == ERROR)
                    {
                        IAS_LOG_ERROR("Correlating band %d and %d.",
                            reference_band_number, search_band_number);
                        free(gcp);
                        ias_l1g_close_image(srch_img_ptr);
                        ias_l1g_close_image(ref_img_ptr);
                        fclose(output);
                        free(result_ptr);
                        ias_math_close_parallel_correlator(correlator_ptr);
                        return ERROR;
                    }

                    /* Write out all the GCPs and the correlation data.  */

                    if (ref_nsca == 1)
                    {
                        /* SCA-combined image */
                        sca_number = 0;
                    }
                    else
                    {
                        sca_number = sca_index + 1;
                    }

                    for (i = 0; i < numgcps[sca_index]; i++)
                    {
                        if (gcp[i].accept_flag)
                        {
                            if (fprintf (output,
                                "%3s   %11.6lf  %11.6lf  %11.6lf  %11.6lf  "
                                "%11.6lf  %11.6lf  %11.6lf  %11.6lf  %11.6lf  "
                                "%3d  %11.6lf  %3d  %8d  %10d  %6d  %s  %s\n",
                                gcplib[i].point_id, 
                                gcplib[i].reference_line, 
                                gcplib[i].reference_sample, 
                                gcplib[i].latitude, 
                                gcplib[i].longitude, 
                                gcplib[i].elevation, 
                                gcplib[i].reference_line, 
                                gcplib[i].reference_sample, 
                                gcp[i].fit_line_offset, 
                                gcp[i].fit_sample_offset,
                                gcp[i].accept_flag,
                                gcp[i].correlation_coefficient,
                                reference_band_number,
                                search_band_number,
                                sca_number, sca_number, 
                                rimg_name, simg_name) < 0)
                            {
                                IAS_LOG_ERROR("Writing to the output file.");
                                free(gcp);
                                ias_l1g_close_image(srch_img_ptr);
                                ias_l1g_close_image(ref_img_ptr);
                                fclose(output);
                                return ERROR;
                            }

                            /* Add the fit_offset to the mean_offset if the 
                               point correlated.  */
                            num_correlated++;
                            mean_offset[0] += gcp[i].fit_line_offset;
                            mean_offset[1] += gcp[i].fit_sample_offset;
                        }
                    }

                    if (num_correlated < 1)
                    {
                        /* No points correlated in this band/SCA. */
                        IAS_LOG_WARNING("No GCPs correlated.");
                    }
                    else
                    {
                        /* Output the number of GCPs that correlated.  */
                        IAS_LOG_INFO("%d GCPs correlated out of %d.", 
                                num_correlated, numgcps[sca_index]);
                    }

                    /* If any values correlated then calculate the mean and 
                       std dev.  */
                    if (num_correlated > 1)
                    {
                        /* Output the mean offset values for the current band 
                           combination.  */
                        mean_offset[0] /= (float) num_correlated;
                        mean_offset[1] /= (float) num_correlated;
                        IAS_LOG_INFO("The mean fit_offset is %f, %f "
                            "(line, sample).", mean_offset[0], mean_offset[1]);

                        /* Malloc space to hold the correlated fit_offsets for 
                           the std dev calculation.  */
                        offset[0] = (double *)malloc(
                            num_correlated * sizeof(double));
                        if (offset[0] == NULL)
                        {
                            IAS_LOG_ERROR (
                                "Allocating memory for the offset array.");
                            free(gcp);
                            ias_l1g_close_image(srch_img_ptr);
                            ias_l1g_close_image(ref_img_ptr);
                            fclose(output);
                            free(result_ptr);
                            ias_math_close_parallel_correlator(correlator_ptr);
                            return ERROR;
                        }
                        offset[1] = (double *)malloc(
                            num_correlated * sizeof(double));
                        if (offset[1] == NULL)
                        {
                            IAS_LOG_ERROR(
                                "Allocating memory for the offset array.");
                            free(gcp);
                            free(offset[0]);
                            ias_l1g_close_image(srch_img_ptr);
                            ias_l1g_close_image(ref_img_ptr);
                            fclose(output);
                            free(result_ptr);
                            ias_math_close_parallel_correlator(correlator_ptr);
                            return ERROR;
                        }

                        /* Output the standard deviation for the current band 
                           combinations.  */
                        for (i = 0, curr = 0; i < numgcps[sca_index]; i++)
                        {
                            if (gcp[i].accept_flag)
                            {
                                offset[0][curr] = gcp[i].fit_line_offset;
                                offset[1][curr] = gcp[i].fit_sample_offset;
                                curr++;
                            }
                        }
                        if (ias_math_compute_stdev(offset[0], num_correlated, 
                            mean_offset[0], &stddev_offset[0]) == ERROR)
                        {
                            IAS_LOG_ERROR(
                                    "Calculating the standard deviation of "
                                    "the line offsets.");
                            free(gcp);
                            free(offset[0]);
                            free(offset[1]);
                            ias_l1g_close_image(srch_img_ptr);
                            ias_l1g_close_image(ref_img_ptr);
                            fclose(output);
                            free(result_ptr);
                            ias_math_close_parallel_correlator(correlator_ptr);
                            return ERROR;
                        }
                        if (ias_math_compute_stdev(offset[1], num_correlated, 
                            mean_offset[1], &stddev_offset[1]) == ERROR)
                        {
                            IAS_LOG_ERROR(
                                    "calculating the standard deviation of "
                                    "the sample offsets.");
                            free(gcp);
                            free(offset[0]);
                            free(offset[1]);
                            ias_l1g_close_image(srch_img_ptr);
                            ias_l1g_close_image(ref_img_ptr);
                            fclose(output);
                            free(result_ptr);
                            ias_math_close_parallel_correlator(correlator_ptr);
                            return ERROR;
                        }
                                
                        IAS_LOG_INFO("The standard deviation fit_offset is "
                                "%f, %f (line, sample).",
                                stddev_offset[0], stddev_offset[1]);

                        /* Free up the offset pointers for the next set.  */
                        free(offset[0]);
                        free(offset[1]);
                    }

                    /* Close the L1G images.  */
                    if (ias_l1g_close_image(ref_img_ptr) == ERROR)
                    {
                        IAS_LOG_ERROR("Closing 1G reference image.");
                        free(gcp);
                        ias_l1g_close_image(srch_img_ptr);
                        fclose(output);
                        free(result_ptr);
                        ias_math_close_parallel_correlator(correlator_ptr);
                        return ERROR;
                    }
                    if (ias_l1g_close_image(srch_img_ptr) == ERROR)
                    {
                        IAS_LOG_ERROR("Closing 1G search image.");
                        free(gcp);
                        fclose(output);
                        free(result_ptr);
                        ias_math_close_parallel_correlator(correlator_ptr);
                        return ERROR;
                    }

                    /* Set the done flag to TRUE.  */
                    done[rref][sref] = TRUE;

                } /* if valid combination */
            } /* for cursrch */
        } /* for curref */
    } /* SCA loop */

    /* Allow the parallel correlator to finish. */
    if (ias_math_close_parallel_correlator(correlator_ptr) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing the parallel correlator");
        free(gcp);
        free(result_ptr);
        fclose(output);
        return ERROR;
    }

    /* Free the memory for the correlation results. */
    free(result_ptr);

    /* Free up the IAS_GCP_RESULTS pointer and close the output file.  */
    free(gcp);
    status = fclose(output);
    if (status != 0)
    {
        IAS_LOG_ERROR("Closing output report file.");
        return ERROR;
    }

    return SUCCESS;
}
