/******************************************************************************
NAME:           compare_metadata

PURPOSE:
Read the metadata and compare the metadata information to verify that it is the
same for all bands.  The pixel sizes need to be a factor of two with reference
to the band resolutions.  An array of band metadata and a structure of image
metadata is returned on successful completion of this routine.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
B2B characterization requires that the 15m resolution be 2x as large as the 
30m resolution. If this is not the case, the application will exit with a 
failure.

A band-referenced array of metadata is built.  This is similar 
to the manner of storing the different band elements of the grid structure.

ALGORITHM REFERENCES:
See ias_l1g.h for a description of the file and band metadata structures.

******************************************************************************/
#include <string.h>
#include "gctp.h"
#include "b2b_char.h"
#include "ias_logging.h"              /* IAS logging library          */
#include "ias_satellite_attributes.h" /* Satellite attributes library */

#define TOLERANCE 0.000001 /* The tolerance to use for comparing
                              floating point data. */

int compare_metadata 
(
    const char *l1g_filename,     /* I: L1G filename                          */
    const int *band_indexes,      /* I: Array of band indexes                 */
    const int nbands,             /* I: Number of band indexes in the array   */
    IAS_L1G_BAND_METADATA band_metadata[IAS_MAX_NBANDS], /* O: Array of band 
                                        data                                  */
    int *nl,                      /* O: Number of lines                       */
    int *ns                       /* O: Number of samples                     */
)
{
    int i;                     /* loop counter                             */
    int stat;                  /* error status                             */
    double ratio;               /* Ratio for the pixel sizes between the
                                   first and ith band */
    IAS_L1G_FILE_METADATA file_metadata; /* file-wide metatdata            */
    int first_band_number;     /* First band                               */
    int current_band_number;   /* Current band                             */

    /* Read the L1G metadata for the band array and for the file-wide 
       metadata. */ 
    stat = read_metadata(l1g_filename, band_indexes, nbands, &file_metadata,
        band_metadata, nl, ns);
    if (stat == ERROR)
    {
        IAS_LOG_ERROR("Reading the file and band metadata.");
        return ERROR;
    }

    /* Check that the input image projection is UTM or SOM.  Produce a
       warning, not an error, if it's not one of these projections.  */ 
    if ((file_metadata.projection_code != UTM) &&
        (file_metadata.projection_code != SOM))
    {
        IAS_LOG_WARNING("Image %s is not in UTM or SOM projection", 
            l1g_filename);
    }

    /* Check that the input image correction type is precision-terrain 
       corrected, or "L1T".  Produce a warning, not an error, otherwise. */
    if (strcmp(file_metadata.correction_type, "L1T") != 0)
    {
        IAS_LOG_WARNING("Image %s is not L1T corrected", l1g_filename);
    }

    /* Check that the input image is nadir-viewing.  Produce a warning, not 
       an error, otherwise. */
    if ((file_metadata.roll_angle > IAS_OFF_NADIR_BOUNDARY) ||
        (file_metadata.roll_angle < -IAS_OFF_NADIR_BOUNDARY))
    {
        IAS_LOG_WARNING("Image %s is not nadir-viewing; roll angle is %f", 
            l1g_filename, file_metadata.roll_angle);
    }

    /* Check the pixel sizes.  Visible band pixel sizes must be 2x pan band
     * pixel sizes. */ 

    /* Look up pixel resolution of the first band. */
    first_band_number = ias_sat_attr_convert_band_index_to_number(
        band_indexes[0]);
    if ((first_band_number < 1) || (first_band_number > IAS_MAX_NBANDS))
    {
        IAS_LOG_ERROR("Invalid band_number %d", first_band_number); 
        return ERROR;
    }

    /* Check the other bands' pixel sizes against band 1. */
    for (i = 1; i < nbands; i++)
    {
        /* Look up pixel resolution of the current band. */
        current_band_number = ias_sat_attr_convert_band_index_to_number(
            band_indexes[i]);
        if ((current_band_number < 1) || (current_band_number > IAS_MAX_NBANDS))
        {
            IAS_LOG_ERROR("Invalid band_number %d", current_band_number); 
            return ERROR;
        }

        /* Determine the ratio of the first band with the current band.
           Only x is used here because x and y should be the same. */
        ratio = band_metadata[0].projection_distance_x /
                band_metadata[i].projection_distance_x;

        /* Search for invalid band combinations. */
        if (((ratio > 0.5 + TOLERANCE) || (ratio < 0.5 - TOLERANCE)) &&
            ((ratio > 1.0 + TOLERANCE) || (ratio < 1.0 - TOLERANCE)) &&
            ((ratio > 2.0 + TOLERANCE) || (ratio < 2.0 - TOLERANCE)))
        {
            IAS_LOG_ERROR("Invalid ratio in pixel sizes between bands "
                "%d and %d.", first_band_number, current_band_number);
            return ERROR;
        }
    }

    return SUCCESS;
}
