/******************************************************************************
NAME:           create_tiepts

PURPOSE:
Create tiepoints for a band combination.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
This routine will malloc the appropriate space needed for the GCP library
pointers.

ALGORITHM REFERENCES:

******************************************************************************/
#include <stdlib.h>
#include "b2b_char.h"
#include "ias_logging.h"              /* IAS logging library          */
#include "ias_satellite_attributes.h" /* Satellite attributes library */

int create_tiepts 
(
    const int *band_indexes,    /* I: Band indexes                          */
    const int nbands,           /* I: Number of indexes in band index array */
    const IAS_L1G_BAND_METADATA band_metadata[IAS_MAX_NBANDS], /* I: Array 
                                      of band data                          */
    const int nl,               /* I: Number of lines                       */
    const int ns,               /* I: Number of samples                     */
    const int nline_pts,        /* I: Number of line tiepoints to create    */
    const int nsamp_pts,        /* I: Number of sample tiepoints to create  */
    const struct IAS_LNG_LS chip_size,/* I: Chip size of the tiepoints to 
                                      create                                */
    int pan_band_index,         /* I: Index of the PAN band                 */
    IAS_GCP_RECORD **gcp30      /* O: Array of GCP Library tiepoints in row- 
                                      major form                            */
)
{
    int stat;                  /* Return status                             */
    int  numgcps;              /* Total number of GCPs in the GCP library   */
    int reference_count;       /* Reference loop counter                    */
    int reference_index;       /* Reference band index                      */

    /* Initialize the GCP pointer to NULL.  Then it can be used as a flag
     * to determine if the GCP record has been created.  */
    *gcp30 = NULL;

    /* The total number of GCPs is nline_pts * nsamp_pts.  */
    numgcps = nline_pts * nsamp_pts;

    /* Loop through the band combinations.  At most, there will only be one 
       GCP record created. */ 
    for (reference_count = 0; reference_count < nbands; reference_count++)
    {
        reference_index = band_indexes[reference_count];

        /* Don't allow PAN band to define the reference tiepoint spacing */
        if (reference_index == pan_band_index)
            continue;

        /* If the GCP record hasn't already been done ... */
        if (*gcp30 == NULL)
        {
            /* Determine the regular-spaced tiepoints. */
            *gcp30 = malloc(numgcps * sizeof (IAS_GCP_RECORD));
            if (*gcp30 == NULL)
            {
                IAS_LOG_ERROR("Allocating space for the GCP30 tiepoints.");
                return ERROR;
            }

            stat = det_tiepts(&band_metadata[reference_count], nl, ns,
                nline_pts, nsamp_pts, &chip_size, *gcp30);
            if (stat != SUCCESS)
            {
                IAS_LOG_ERROR("Getting GCP30 tiepoints.");
                free(*gcp30);
                return ERROR;
            }
        }

        /* If the GCP record has been created, then we're done. */
        if (*gcp30 != NULL)
            break;

    }  /* for reference_count */

    return SUCCESS;
}
