/******************************************************************************
NAME:           create_tiepts_grid

PURPOSE:
Create tiepoints for all band combinations.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
This routine will malloc the appropriate space needed for the IAS_GCP_RECORD
pointers.

ALGORITHM REFERENCES:

******************************************************************************/
#include <stdlib.h>
#include "b2b_char.h"
#include "ias_logging.h"        /* IAS logging library           */
#include "gctp.h"
#include "ias_satellite_attributes.h" /* Satellite attributes library */

int create_tiepts_grid 
(
    const int *band_indexes,    /* I: Band indexes                           */
    const int nbands,           /* I: Number of band indexes in array        */
    const int sca_indexes[IAS_MAX_NSCAS], /* I: SCA indexes to search        */
    const int nscas,            /* I: Number of SCA indexes in array         */
    const IAS_L1G_BAND_METADATA band_metadata[IAS_MAX_NBANDS], /* I: Array
                                      of band data                           */
    const int nline_pts,        /* I: Number of line tiepoints to create     */
    const int nsamp_pts,        /* I: Number of sample tiepoints to create   */
    const struct IAS_LNG_LS chip_size,/* I: Chip size of the tiepoints to 
                                      create                                 */
    const char *grid_filename,  /* I: Grid file name                         */
    int pan_band_index,         /* I: Index of the PAN band                  */
    int *num_gcps,              /* O: Number of GCPs                         */
    IAS_GCP_RECORD *sca_gcp[IAS_MAX_NSCAS] /* O: Arrays of GCPs for each SCA */
)
{
    int k;                /* Loop variable                                   */
    int sca_index;        /* SCA index                                       */
    int sca_loop_idx;     /* Loop index into SCA list                        */
    int stat;             /* Return status                                   */
    int  numgcps;         /* Total number of GCPs in the GCP library         */
    IAS_GRID_TYPE grid;   /* the grid structure                              */
    int reference_count;  /* Reference loop counter                          */
    int reference_index;  /* Reference band index                            */
    int band_numbers[IAS_MAX_NBANDS];  /* List of band numbers               */

    /* Initialize the GCP pointers to NULL, then they can be used as a flag
     * to determine if the GCP structure has been created.  */
    for (sca_index = 0; sca_index < nscas; sca_index++)
    {
        sca_gcp[sca_index] = NULL;
    }

    /* Make a list of band numbers corresponding to the band indexes. */
    for (k = 0; k < nbands; k++)
    {
        band_numbers[k] = ias_sat_attr_convert_band_index_to_number(
            band_indexes[k]);
        if ((band_numbers[k] < 1) || (band_numbers[k] > IAS_MAX_NBANDS))
        {
            IAS_LOG_ERROR("Invalid band number %d.", band_numbers[k]);
            return ERROR;
        }
    }

    /* Read grid file.  */
    stat = ias_grid_read(grid_filename, band_numbers, nbands, &grid);
    if (stat != SUCCESS)
    {
        IAS_LOG_ERROR("Reading grid file: '%s'", grid_filename);
        return ERROR;
    }

    if (grid.frame_type == IAS_PATH_ORIENTED)
    {
        IAS_LOG_ERROR("Currently grid option only works with North-up images");
        return ERROR;
    }

    /* The total number of GCPs is nline_pts * nsamp_pts.  */
    numgcps = nline_pts * nsamp_pts;

    for (sca_loop_idx = 0; sca_loop_idx < nscas; sca_loop_idx++)
    {
        sca_index = sca_indexes[sca_loop_idx];

        for (reference_count = 0; reference_count < nbands; reference_count++)
        {
            reference_index = band_indexes[reference_count];

            /* Don't allow PAN band to define the reference tiepoint spacing */
            if (reference_index == pan_band_index)
                continue;

            /* If the GCP record hasn't already been done ...  */
            if (sca_gcp[sca_loop_idx] == NULL)
            {
                /* Determine the regular-spaced tiepoints. */ 
                sca_gcp[sca_loop_idx] = 
                    malloc(numgcps * sizeof(IAS_GCP_RECORD)); 
                if (sca_gcp[sca_loop_idx] == NULL)
                {
                    IAS_LOG_ERROR("Allocating space.");
                    for (k = 0; k < sca_loop_idx; k++)
                    {
                        free(sca_gcp[k]);
                    }
                    return ERROR;
                }

                stat = grid_det_tiepts(&band_metadata[reference_count], 
                    nline_pts, nsamp_pts, &chip_size, sca_index, 
                    reference_index, &grid, &num_gcps[sca_index],
                    sca_gcp[sca_loop_idx]);
                    
                if (stat == ERROR)
                {
                    IAS_LOG_ERROR("Getting tiepoints.");
                    for (k = 0; k < sca_loop_idx; k++)
                    {
                        free(sca_gcp[k]);
                    }
                    return ERROR;
                }
                IAS_LOG_INFO("SCA %d Number of GCPs %d\n", sca_index + 1, 
                    num_gcps[sca_index]);
            }

            /* If the GCP records have been created, then we're done.  */
            if (sca_gcp[nscas-1] != NULL)
            {
                break;
            }
        }  /* for reference_count */
    }

    /* Free the grid.  */
    if (ias_grid_free(&grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Freeing the grid structure.");
        return ERROR;
    }

    return SUCCESS;
}
