/******************************************************************************
NAME:           det_tiepts

PURPOSE:
Determine the tiepoints for a specified band resolution.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
This routine will make sure that the chip windows for the determined tiepoints
don't fall outside the image.  Thus for any given tiepoint, its chip window
(according to the chip_size) will be completely inside the image for the
specified band resolution.

The corner points for the tiepoints are found in the following manner:
  UL(x) = Chip_Size/2               UR(x) = nsamps - Chip_Size/2
  UL(y) = Chip_Size/2               UR(y) = Chip_Size/2
  LL(x) = Chip_Size/2               LR(x) = nsamps - Chip_Size/2
  LL(y) = nlines - Chip_Size/2      LR(y) = nlines - Chip_Size/2

ALGORITHM REFERENCES:

******************************************************************************/
#include "b2b_char.h"
#include "ias_logging.h"          /* IAS logging library           */


int det_tiepts 
(
    const IAS_L1G_BAND_METADATA *band_metadata, /* I: band data             */ 
    const int nl,                     /* I: Number of lines                 */
    const int ns,                     /* I: Number of samples               */
    const int nline_pts,              /* I: Number of line tiepoints to 
                                            create                          */
    const int nsamp_pts,              /* I: Number of sample tiepoints to 
                                            create                          */
    const struct IAS_LNG_LS *chip_size, /* I: Chip size of the tiepoints to 
                                            create                          */
    IAS_GCP_RECORD *gcps        /* O: Array of GCP Library tiepoints in 
                                      row-major format                      */
)
{
    int i;                    /* Looping variable                           */
    int j;                    /* Looping variable                           */
    int status;               /* status of return from function             */
    int count;                /* Number of GCPs                             */
    double min_x;             /* Minimum x tiepoint value within the image  */
    double max_x;             /* Maximum x tiepoint value within the image  */
    double min_y;             /* Minimum y tiepoint value within the image  */
    double max_y;             /* Maximum y tiepoint value within the image  */
    double x_spacing;         /* Line spacing between chips                 */
    double y_spacing;         /* Sample spacing between chips               */
    double input_line;        /* Input line location for GCP location       */
    double input_sample;      /* Input sample location for GCP location     */


    /* If only one line or sample point then return with an ERROR.  */
    if (nline_pts <= 1)
    {
        IAS_LOG_ERROR("This routine requires more than one line point.");
        return ERROR;
    }
    if (nsamp_pts <= 1)
    {
        IAS_LOG_ERROR("This routine requires more than one sample point.");
        return ERROR;
    }

    /* Find the min and max corner point values for the chips inside of the
     * image.  */
    min_x = chip_size->samp / 2.0;
    max_x = ns - chip_size->samp/2.0;
    min_y = chip_size->line / 2.0;
    max_y = nl - chip_size->line/2.0;

    /* Determine the line and sample spacings for all other points between
     * the min and max points.  */
    x_spacing = (max_x - min_x) / (nsamp_pts - 1);
    y_spacing = (max_y - min_y) / (nline_pts - 1);

    /* Fill in the complete GCP structures.  */
    for (i = 0, count = 0; i < nline_pts; i++)
    {
        if (i == nline_pts - 1)
            input_line = max_y; 
        else
            input_line = min_y + y_spacing * i; 

        for (j = 0; j < nsamp_pts; j++, count++)
        {
            if (j == nsamp_pts - 1)
                input_sample = max_x; 
            else
                input_sample = min_x + x_spacing * j; 

            status = sprintf(gcps[count].point_id, "%d", count);
            if (status < 1)
            {
                IAS_LOG_ERROR("In sprintf");
                return ERROR;
            }
            gcps[count].reference_line = input_line;
            gcps[count].reference_sample = input_sample;
            gcps[count].pixel_size_x = band_metadata->projection_distance_x;  
            gcps[count].pixel_size_y = band_metadata->projection_distance_y;  
            gcps[count].chip_size_lines = chip_size->line;
            gcps[count].chip_size_samples = chip_size->samp;
            gcps[count].latitude = 0.0;
            gcps[count].longitude = 0.0;
            gcps[count].elevation = 0.0;
        }
    }

    return SUCCESS;
}
