/******************************************************************************
NAME:           downsample

PURPOSE:
Downsamples all the necessary images by a factor of two and writes
the images to a file.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
The Gaussian Pyramid routine uses a 5 x 5 low pass filter.  The shape of the
convolution function is controlled by the weight.  The bands that are
downsampled will appear in the current directory with the name of
REDUCE_FILE.  

ALGORITHM REFERENCES:

******************************************************************************/
#include <stdlib.h>
#include <string.h>

#include "b2b_char.h"
#include "ias_logging.h"              /* IAS logging library          */
#include "ias_satellite_attributes.h" /* Satellite attributes library */
#include "ias_math.h"                 /* IAS Math library             */

/* Create a reduced resolution PAN band. */
int downsample 
(
    const char *in_name,   /* I: Original image                           */
    const int band_indexes[IAS_MAX_NBANDS], /* I: Band numbers            */
    const int nbands       /* I: Number of bands                          */
)
{
    int cur;               /* Current band                                  */
    int band_index;        /* Band index                                    */
    int counter;           /* Weight counter                                */ 
    int i;                 /* Loop variable                                 */
    int j;                 /* Loop variable                                 */
    int down = 2;          /* Downsampling scaling factor                   */
    double delta;          /* Scaling ratio                                 */
    double *weights;       /* Resampling weights                            */
    int number_down_weights; /* Number of resampling weights                */
    double x;              /* "Positions" for scaling weights               */
    double y;              /* "Positions" for scaling weights               */
    double wx;             /* Resampling weights for x/y directions         */
    double wy;             /* Resampling weights for x/y directions         */
    double sum = 0.0;      /* Sum of weights                                */
    int band_number;       /* Band number version of band                   */
    int pan_band_number;   /* PAN band number for the satellite             */


    number_down_weights = 5 + 4 * ( down - 1 );
    delta = 1.0 / (double)down;
    weights = (double *)malloc(sizeof(double) * number_down_weights 
        * number_down_weights);
    if (weights == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for downsample weights.");
        return ERROR;
    }
    for (j = 0, counter = 0, y =- 2.0; j < number_down_weights; j++)
    {
        wy = ias_math_cubic_convolution(-0.5, y);
        for (i = 0, x = -2.0; i < number_down_weights; i++, counter++)
        {
            wx = ias_math_cubic_convolution(-0.5, x);
            weights[counter] = wy * wx;
            sum = sum + weights[counter];
            x = x + delta;
        }
        y = y + delta;
    } 
    for (i = 0; i < number_down_weights * number_down_weights; i++)
    {
        weights[i] = weights[i] / sum;
    }

#ifdef PRINT_WEIGHTS
    FILE *fp;              /* Weights file pointer */
    IAS_LOG_INFO("SUM %f", sum);
    fp = fopen("weights.dat", "wb");
    for (j = 0, counter = 0; j < number_down_weights; j++)
    {
        for (i = 0; i < number_down_weights; i++, counter++)
        {
            fprintf(fp, "%d %d %f\n", j, i, weights[counter]);
        }
        fprintf(fp, "\n");
    }
    fclose(fp);
#endif

    /* Print a message to notify we are downsampling.  */
    IAS_LOG_INFO("Downsampling ...");

    /* Get the PAN band number. */
    pan_band_number = ias_sat_attr_get_band_number_from_type(IAS_PAN_BAND);
    if ((pan_band_number < 1) || (pan_band_number > IAS_MAX_NBANDS))
    {
        IAS_LOG_ERROR("Invalid band number %d.", pan_band_number);
        free(weights);
        return ERROR;
    }

    /* Loop through the bands.  */
    for (cur = 0; cur < nbands; cur++)
    {
        /* Get the current band index. */
        band_index = band_indexes[cur];
        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        if ((band_number < 1) || (band_number > IAS_MAX_NBANDS))
        {
            IAS_LOG_ERROR("Invalid band number %d.", band_number);
            free(weights);
            return ERROR;
        }

        /* Reduce the input file for the specified band by a factor of 2 and
         * output it to REDUCE_FILE.  Only do the reduction if the reduce flag
         * is set.  */
        if (band_number == pan_band_number)
        {
            IAS_LOG_INFO("Pan band found, creating reduced image file...");
            if (reduce(in_name, REDUCE_FILE, band_indexes[cur], down, weights, 
                number_down_weights) == ERROR)
            {
                IAS_LOG_ERROR("Downsampling the 1g image %s for band number"
                    " %d.", in_name, band_number);
                free(weights);
                return ERROR;
            }
        }
    }  /* for cur */

    free(weights);

    return SUCCESS;
}
