/******************************************************************************
NAME:           get_parms

PURPOSE:
Get the B2B characterization parameters from the input parameter file.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
This routine opens the input parameter files, reads the B2B characterization
parameters, and returns the parameters.  It also reads the CPF and retrieves
the min_corr, max_disp, chip_size, range, threshold, and fit_method from it.
Min_corr, max_disp, chip_size, range, and threshold CPF values are overridden 
by parameter file values if they are present there.

ALGORITHM REFERENCES:
******************************************************************************/
#include <stdlib.h>
#include <string.h> 

#include "b2b_char.h"
#include "ias_parm.h"                      /* parameter library */
#include "ias_parm_standard_parameters.h"  /* parameter table standard fields */
#include "ias_cpf.h"                       /* CPF library */
#include "ias_satellite_attributes.h"      /* Satellite attributes library */
#include "ias_math.h"                      /* insertion sort used */

#define B2B_SCA_ERROR -1        /* This indicates that an entry in the SCA 
                                   list is not a valid SCA. */
#define DEFAULT_REDUCE_FLAG 1   /* Default is to reduce */

int get_parms 
(
    const char *param_filename,   /* I: Parameter file name                   */
    char *file_name_1G,           /* O: File name for the 1G image            */
    char *file_name_0R,           /* O: File name for the 1G image            */
    char *output_file_name,       /* O: Name of the output file               */
    char *grid_file_name,         /* O: Grid file name                        */
    char *cpf_file_name,          /* O: CPF file name                         */
    int *reduce_flag,             /* O: Flag to indicate reduced image should 
                                        be created                            */
    int band_indexes[IAS_MAX_NBANDS], /* O: Band indexes                      */
    int *nbands,                  /* O: Number of band indexes                */
    int sca_indexes[IAS_MAX_NSCAS], /* O: SCA indexes to search               */
    int *nscas,                   /* O: Number of SCA indexes                 */
    int *nline_pts,               /* O: Number of line tiepoints              */
    int *nsamp_pts,               /* O: Number of samp tiepoints              */
    struct IAS_LNG_LS *chip_size, /* O: Height and width of the chips         */
    enum TIEPT_TYPE *tiept_type,  /* O: Type of tiepts to use.                */
    float *min_corr,              /* O: Minimum correlation strength          */
    float *max_disp,              /* O: Maximum correlation displacement      */
    float *range,                 /* O: Range of valid pixels                 */
    float *threshold,             /* O: Percentage of valid pixels            */
    IAS_CORRELATION_FIT_TYPE *fit_method, /* O: Correlation result fit method */
    int *abs_corr_coeff,          /* O: Flag to use abs value of correlation
                                        coefficients in grey correlation      */
    char *wo_id,                  /* O: Work order ID string                  */
    char *omf_file_name           /* O: omf file name                         */
)
{
    int lunar_sca_number;       /* Optional OMF parameter indicating which
                                   SCA imaged the moon (lunar collects
                                   only) */
    int i;                      /* Loop variable.                             */
    int j;                      /* Loop variable.                             */
    int curr_band_index;        /* counter for verifying bands present        */
    int count;                  /* Number of values returned                  */
    int tiept;                  /* Type of the tiepoint                       */
    int band_number;            /* Band number for verifying bands            */
    static int band_number_list[IAS_MAX_NBANDS]; /* Band numbers to process */
    static int sca_number_list[IAS_MAX_NSCAS]; /* SCA number list to process */
    int satellite_number_bands; /* Number of normal bands for the satellite   */
    IAS_CPF *cpf;               /* Calibration parameter file constants       */
    const struct IAS_CPF_B2B_ASSESSMENT *cpf_b2b;
    int status;                 /* Status returned from functions             */
    int processing_pass;        /* The current processing pass                */
    int sat_num;                /* The satellite number                       */
    IAS_SATELLITE_ID sat_id;    /* The satellite ID                           */
    int sca_count_in_image = B2B_SCA_ERROR;/* Number of SCAs in image */

    /* Various file name storage - local storage is used because sizeof doesn't
       work on function parameter arrays. */
    char local_file_name_1G[PATH_MAX];
    char local_file_name_0R[PATH_MAX];
    char local_output_file_name[PATH_MAX];
    char local_grid_file_name[PATH_MAX];
    char local_cpf_file_name[PATH_MAX];
    char local_wo_id[IAS_WORKORDER_ID_SIZE];

    /* Local buffer used to translate between IAS_LNG_LS* and parameter I/O
       interface. */
    int local_chip_size[2];

    /* Local buffers used to translate between float* and parameter I/O
       interface. */
    double local_range[2];
    double local_min_corr;
    double local_max_disp;
    double local_threshold;

    L1GIO * l1g_image;  /* pointer to l1g image structure for verifying bands */

    /* Initialize band_number_list. */
    for (i = 0; i < IAS_MAX_NBANDS; i++)
        band_number_list[i] = 0;

    /* Initialize SCA list to be empty. */
    for (i = 0; i < IAS_MAX_NSCAS; i++)
    {
        sca_number_list[i] = B2B_SCA_ERROR;
        sca_indexes[i] = B2B_SCA_ERROR;
    }

    /* Start parameter table definition */
    IAS_PARM_DECLARE_TABLE(parms, 14);

    /* Get the work order ID */
    IAS_PARM_WORK_ORDER_ID(parms, local_wo_id, sizeof(local_wo_id), 0);

    /* Bands and SCAs are optional parameters.  If they aren't there, the full
       list/range is used. */

    /* Get the list of bands. */
    IAS_PARM_BAND_LIST(parms, IAS_PARM_OPTIONAL, band_number_list, 
        sizeof(band_number_list), 1);

    /* Get the list of SCAs. */
    IAS_PARM_ADD_INT(parms, SCA_LIST, "List of SCAs to process.  Overrides the"
        " LUNAR_SCA_NUMBER in the OMF file.  If neither is specified, defaults"
        " to all SCAs", IAS_PARM_OPTIONAL, IAS_PARM_ARRAY, 1, 1, 14, 0, NULL,
        sca_number_list, sizeof(sca_number_list), 0);

    /* Get the tiept_type. */
    int default_tiept_type[] = {1};
    IAS_PARM_TIEPT_TYPE(parms, default_tiept_type, &tiept, sizeof(tiept), 1);

    /* Get the nline_pts. */ 
    int default_nline_pts[] = {25};
    IAS_PARM_NLINE_PTS(parms, 1, default_nline_pts, nline_pts,
        sizeof(*nline_pts));

    /* Get the nsamp_pts. */ 
    int default_nsamp_pts[] = {25};
    IAS_PARM_NSAMP_PTS(parms, 1, default_nsamp_pts, nsamp_pts,
        sizeof(*nsamp_pts));

    /* Get the minimum correlation value. */ 
    IAS_PARM_MIN_CORR(parms, &local_min_corr, sizeof(local_min_corr));

    /* Get the maximum displacement value. */ 
    IAS_PARM_MAX_DISP(parms, &local_max_disp, sizeof(local_max_disp));

    /* Get the valid pixel percentage threshold value. */ 
    IAS_PARM_THRESHOLD(parms, &local_threshold, sizeof(local_threshold));

    /* Get the chip size values. */ 
    IAS_PARM_CHIP_SIZE(parms, local_chip_size, sizeof(local_chip_size));

    /* Get the range values. */ 
    IAS_PARM_RANGE(parms, local_range, sizeof(local_range));

    /* Get the output_file_name.  */
    const char *default_output_file_name[] = {"b2b_char.dat"};
    IAS_PARM_ADD_STRING(parms, OUTPUT_FILENAME,
        "Output data file name", IAS_PARM_OPTIONAL,
        0, NULL, 1, default_output_file_name,
        local_output_file_name, sizeof(local_output_file_name), 0);

    /* Get the reduce_flag to determine whether or not downsampling should 
       occur.  This is used for testing purposes (the reduce process can take
       a lot of time) and for consistency with CalVal prototype software. 
       It is not for production use. */
    int default_reduce_flag[] = {DEFAULT_REDUCE_FLAG};
    IAS_PARM_ADD_INT(parms, REDUCE_FLAG, 
        "Switch to allow skip of downsampling during testing",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 0, 1, 1,
        default_reduce_flag, reduce_flag, sizeof(*reduce_flag), 0);

    /* Get processing pass. */
    IAS_PARM_PROCESSING_PASS(parms, &processing_pass, sizeof(processing_pass),
        0);

    /* Calculate the number of items in the parameter file. */
    count = IAS_PARM_GET_TABLE_SIZE(parms);

    /* Call the ias_parm_provide_help routine to provide help for the
       parameters if requested; also verifies the ODL file exists. */
    status = ias_parm_provide_help(param_filename, parms, count,
                                   IAS_INPUT_PARAMETERS);
    if (status == 1)
    {
        /* If the status of the provide help is 1 the user asked for help.
           Since the parameter file name contains the help request, copy it
           to the OMF file name so the next call will ask for the same help. */
        strcpy(omf_file_name, param_filename);
    }
    else if (status == ERROR) /* Help failed. */
    {
        IAS_LOG_ERROR("Providing help using application parameter file %s",
            param_filename);
        exit(EXIT_FAILURE);
    }
    else
    {
        /* Read the parameter file. */
        if (ias_parm_read(param_filename, parms, count) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading parameters from: %s", param_filename);
            return ERROR;
        }

        /* Transfer the various local arrays to the function parameters. */
        strcpy(output_file_name, local_output_file_name);
        strcpy(wo_id, local_wo_id);
        chip_size->line = local_chip_size[0];
        chip_size->samp = local_chip_size[1];
        range[0] = local_range[0];
        range[1] = local_range[1];
        *min_corr = (float)local_min_corr;
        *max_disp = (float)local_max_disp;
        *threshold = (float)local_threshold;

        /* Build the OMF file name. */
        sprintf(omf_file_name, "%s.omf", wo_id);
    }

    /* Now read and check the OMF. */

    /* Start parameter table definition */
    IAS_PARM_DECLARE_TABLE(omf_parms, 6);

    /* Get the L1G file name. */
    IAS_PARM_L1G_FILENAME_FROM_PASS(omf_parms, IAS_PARM_REQUIRED,
        local_file_name_1G, sizeof(local_file_name_1G), 1, processing_pass);

    /* Get the name of the Calibration parameter file. */
    IAS_PARM_CPF(omf_parms, IAS_PARM_REQUIRED, local_cpf_file_name, 
        sizeof(local_cpf_file_name), 1);

    /* Get the L0R file name. */
    IAS_PARM_L0R_FILENAME(omf_parms, IAS_PARM_OPTIONAL, local_file_name_0R, 
        sizeof(local_file_name_0R), 1);

    /* Get the name of the grid file. */
    IAS_PARM_GRID_FILENAME_FROM_PASS(omf_parms, IAS_PARM_REQUIRED,
        local_grid_file_name, sizeof(local_grid_file_name), 1,
        processing_pass);

    /* Get the satellite number */
    IAS_PARM_SATELLITE(omf_parms, &sat_num, sizeof(sat_num), 1);

    /* For lunar collects, get the SCA number with the lunar
       image if it's available.  It's assumed to be unavailable by
       default */
    IAS_PARM_LUNAR_SCA_NUMBER(omf_parms, &lunar_sca_number, 
        sizeof(lunar_sca_number));

    /* See if the user wanted help or a template with the OMF file. */
    status = ias_parm_provide_help(omf_file_name, omf_parms,
        IAS_PARM_GET_TABLE_SIZE(omf_parms), IAS_OMF_PARAMETERS);
    if (status)
    {
        if (status == 1) /* Help was successful. */
        {
            exit(EXIT_SUCCESS);
        }
        else if (status == ERROR) /* Help failed. */
        {
            IAS_LOG_ERROR("Providing help using OMF %s", omf_file_name);
            exit(EXIT_FAILURE);
        }
    }

    /* Read the OMF file */
    count = IAS_PARM_GET_TABLE_SIZE(omf_parms);
    if (ias_parm_read(omf_file_name, omf_parms, count) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from OMF file: %s", omf_file_name);
        return ERROR;
    }

    strcpy(file_name_1G, local_file_name_1G);
    strcpy(file_name_0R, local_file_name_0R);
    strcpy(cpf_file_name, local_cpf_file_name);
    strcpy(grid_file_name, local_grid_file_name);

    /*==========================================================================
     * Validate the parameters read from the ODL file
     *========================================================================*/

    /* Using the satellite number read from the OMF file, initialize the
       satellite attributes library */
    sat_id = ias_sat_attr_get_satellite_id_from_satellite_number(sat_num);
    if (sat_id == ERROR)
    {
        IAS_LOG_ERROR("Unsupported satellite number: %d", sat_num);
        return ERROR;
    }
    if (ias_sat_attr_initialize(sat_id) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attribute library");
        return ERROR;
    }

    /* Find the number of normal bands for the satellite. */
    satellite_number_bands = ias_sat_attr_get_normal_band_count();
    if ((satellite_number_bands < 1) || 
        (satellite_number_bands > IAS_MAX_NBANDS))
    {
        IAS_LOG_ERROR("Invalid number of bands %d.", satellite_number_bands);
        return ERROR;
    }

    /* Exit if the output filename is empty. */
    if (strcmp(output_file_name, "") == 0)
    {
        IAS_LOG_ERROR("Missing output filename in the parameter file.");
        return ERROR;
    }

    /* Find the number of bands returned.  The first 0 encountered signifies the
       end of the band list. */
    for (*nbands = 0; *nbands < satellite_number_bands; (*nbands)++)
    {
        if (band_number_list[*nbands] == 0)
        {
            /* found the end of the band list */
            break;
        }
    }

    /* Make sure there are no duplicate bands.  Error out if there are. */
    for (i = 0; i < (*nbands - 1); i++)
    {
        if (band_number_list[i] == band_number_list[i+1])
        {
            IAS_LOG_ERROR("Duplicate band number (%d) in parameter file.",
                band_number_list[i]);
            return ERROR;
        }
    }

    /* Validate the bands input. */
    if (*nbands > 0)
    {
        for (i = 0; i < *nbands; i++)
        {
            if ((band_number_list[i] <= 0) ||
                (band_number_list[i] > satellite_number_bands))
            {
                IAS_LOG_ERROR("Invalid band number (%d) in parameter file.",
                    band_number_list[i]);
                return ERROR;
            }
        }
    }

    /* Fill in the band indexes.  If the user did not specify any bands then
       fill in all satellite_number_bands bands. */
    if (*nbands > 0)
    {
        for (i = 0; i < *nbands; i++)
        {
            band_indexes[i] =
                ias_sat_attr_convert_band_number_to_index(band_number_list[i]);
            if ((band_indexes[i] < 0) || (band_indexes[i] > IAS_MAX_NBANDS - 1))
            {
                IAS_LOG_ERROR("Invalid band index %d.", band_indexes[i]);
                return ERROR;
            }
        }
    }
    else
    {
        /* Fill in all satellite_number_bands bands. */
        for (i = 0; i < satellite_number_bands; i++)
            band_indexes[i] = i;
        *nbands = satellite_number_bands;
    }

    /* Make sure there are at least 2 bands.  You must have at least 2 for a
       band combination.  Check here to ensure the bands are unique. */
    if (*nbands <= 1)
    {
        IAS_LOG_ERROR("Cannot correlate band onto itself.");
        return ERROR;
    }

    /* If the LUNAR_SCA_NUMBER parameter was found in the OMF and the SCA_LIST
       parameter was not found in the input parameters, use the
       LUNAR_SCA_NUMBER as the SCA to process.  If the SCA_LIST was provided,
       that overrides the OMF parameter automatically. */
    if ((IAS_PARM_GET_COUNT_READ(omf_parms, LUNAR_SCA_NUMBER) > 0)
        && (IAS_PARM_GET_COUNT_READ(parms, SCA_LIST) < 1))
    {
        sca_number_list[0] = lunar_sca_number;
        for (i = 1; i < IAS_MAX_NSCAS; i++)
        {
            sca_number_list[i] = B2B_SCA_ERROR;
        }
    }

    /* Find the number of SCAs returned.  The first B2B_SCA_ERROR encountered
       signifies the end of the SCA list. */
    for (*nscas = 0; *nscas < IAS_MAX_NSCAS; (*nscas)++)
    {
        if (sca_number_list[*nscas] == B2B_SCA_ERROR)
        {
            /* found the end of the SCA list */
            break;
        }
    }

    /* Sort the SCAs in ascending order. */
    ias_math_insertion_sort_integer_array(*nscas, sca_number_list);

    /* Make sure there are no duplicate SCAs.  Error out if there are. */
    for (i = 0; i < (*nscas - 1); i++)
    {
        if (sca_number_list[i] == sca_number_list[i+1])
        {
            IAS_LOG_ERROR("Duplicate SCA number (%d) in parameter file.",
                sca_number_list[i]);
            return ERROR;
        }
    }

    /* Validate the SCAs input. */
    if (*nscas > 0)
    {
        for (i = 0; i < *nscas; i++)
        {
            if ((sca_number_list[i] <= 0) ||
                (sca_number_list[i] > IAS_MAX_NSCAS))
            {
                IAS_LOG_ERROR("Invalid SCA number (%d) in parameter file.",
                    sca_number_list[i]);
                return ERROR;
            }
        }
    }

    /* Fill in the SCA indexes.  If the user did not specify any SCAs then fill
       in all IAS_MAX_NSCAS SCAs. */
    if (*nscas > 0)
    {
        for (i = 0; i < *nscas; i++)
            sca_indexes[i] = sca_number_list[i] - 1;

        /* Fill the rest of the data elements with an invalid SCA number. */
        for (i = *nscas; i < IAS_MAX_NSCAS; i++)
            sca_indexes[i] = B2B_SCA_ERROR;
    }
    else
    {
        /* Fill in all IAS_MAX_NSCAS SCAs. */
        for (i = 0; i < IAS_MAX_NSCAS; i++)
            sca_indexes[i] = i;
        *nscas = IAS_MAX_NSCAS;
    }

    /* Validate the tiept_type. */
    if (tiept == 1) /* Regular-spaced tiepts */
    {
        *tiept_type = REGULAR_SPACED;
    }
    else if (tiept == 2) /* Read the tiept file. */
    {
        *tiept_type = GRID_BASED;
    }
    else
    {
        IAS_LOG_ERROR(
            "Invalid tiepoint type was specified in the parameter file.");
        return ERROR;
    }

    /*==========================================================================
     * Validate the parameters read from the OMF file
     *========================================================================*/

    /* Exit if the grid filename is empty and it's needed. */
    if (strcmp(grid_file_name, "") == 0)
    {
        if (*tiept_type == GRID_BASED)
        {
            IAS_LOG_ERROR("Retrieving grid filename from the OMF file.");
            return ERROR;
        }
    }

    /* Exit if the L1G filename is empty. */
    if (strcmp(file_name_1G, "") == 0)
    {
        IAS_LOG_ERROR("No L1G filename provided in the OMF file.");
        return ERROR;
    }

    /* Do a check to see if the bands in the list exist */

    /* Open the input image */
    l1g_image = ias_l1g_open_image(local_file_name_1G, IAS_READ);
    if (l1g_image == NULL)
    {
        IAS_LOG_ERROR("Can't open image %s", local_file_name_1G);
        return ERROR;
    }

    curr_band_index = 0;
    for (i = 0; i < *nbands; i++)
    {
        band_number = ias_sat_attr_convert_band_index_to_number(
                          band_indexes[curr_band_index]);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Can't convert band index %d to number",
                          band_indexes[curr_band_index]);
            ias_l1g_close_image(l1g_image);
            return ERROR;
        }
        if (!ias_l1g_is_band_present(l1g_image, band_number))
        {
            IAS_LOG_WARNING("Band number %d is not present in L1G image - "
                "skipping it", band_number);
             /* band not present, remove it from list */ 
            for (j = curr_band_index + 1; j < *nbands; j++)
            {
                band_indexes[j-1] = band_indexes[j];
            }
        }
        else
        {
            /* If the number of SCAs hasn't been retrieved from the file, get
               it now */
            if (sca_count_in_image == B2B_SCA_ERROR)
            {
                IAS_DATA_TYPE data_type;
                int lines;
                int samples;

                if (ias_l1g_get_band_size(l1g_image, band_number, &data_type,
                        &sca_count_in_image, &lines, &samples) != SUCCESS)
                {
                    IAS_LOG_ERROR("Failed to read the band size from band "
                        "number %d of %s", band_number, local_file_name_1G);
                    ias_l1g_close_image(l1g_image);
                    return ERROR;
                }
            }

            curr_band_index++;
        }
    }
    *nbands = curr_band_index;

    ias_l1g_close_image(l1g_image);
    l1g_image = NULL;

    /* If the maximum SCA index requested is greater than or equal to the count
       of SCAS, there is a problem with the parameters.  The most likely cause
       of this is using an SCA combined image but requesting SCA numbers like
       it is SCA separated. */
    if (sca_indexes[*nscas - 1] >= sca_count_in_image)
    {
        IAS_LOG_ERROR("The maximum SCA number requested for processing is %d, "
            "but the L1G image %s SCA count is %d", sca_indexes[*nscas - 1] + 1,
            local_file_name_1G, sca_count_in_image);
        return ERROR;
    }

    /* Exit if the CPF filename is empty. */
    if (strcmp (cpf_file_name, "") == 0)
    {
        IAS_LOG_ERROR("No CPF filename provided in the OMF file.");
        return ERROR;
    }

    /* Now read and check the CPF. */

    /* Read the CPF. */
    cpf = ias_cpf_read(cpf_file_name);
    if (!cpf)
    {
        IAS_LOG_ERROR("Cannot open CPF file %s for reading.", cpf_file_name);
        return ERROR;
    }

    /* Get the needed group from the CPF */
    cpf_b2b = ias_cpf_get_b2b_assessment(cpf);
    if (!cpf_b2b)
    {
        IAS_LOG_ERROR("Getting B2B Assessment group from CPF %s",
            cpf_file_name);
        ias_cpf_free(cpf);
        return ERROR;
    }

    /* Retrieve values from the CPF as needed. */ 
    if (IAS_PARM_GET_COUNT_READ(parms, CHIP_SIZE) < 1)
        chip_size->line = cpf_b2b->corr_window_size[0];  
    if (IAS_PARM_GET_COUNT_READ(parms, CHIP_SIZE) < 2)
        chip_size->samp = cpf_b2b->corr_window_size[1];
    if (IAS_PARM_GET_COUNT_READ(parms, RANGE) < 1)
        range[0] = cpf_b2b->corr_pix_min;  
    if (IAS_PARM_GET_COUNT_READ(parms, RANGE) < 2)
        range[1] = cpf_b2b->corr_pix_max;  
    if (IAS_PARM_GET_COUNT_READ(parms, MIN_CORR) == 0)
        *min_corr = cpf_b2b->min_corr_strength;
    if (IAS_PARM_GET_COUNT_READ(parms, MAX_DISP) == 0)
        *max_disp = cpf_b2b->max_displacement_offset;
    if (IAS_PARM_GET_COUNT_READ(parms, THRESHOLD) == 0)
        *threshold = cpf_b2b->fill_threshold;
    *fit_method = cpf_b2b->corr_fit_method;

    /* Free CPF memory. */
    ias_cpf_free(cpf);

    /* Get the abs_corr_coeffs flag.  This value has been taken out of the
       parameters and has been hardcoded to TRUE. */
    *abs_corr_coeff = TRUE;
    return SUCCESS;
}
