/******************************************************************************
NAME:           grid_det_tiepts

PURPOSE:
Determine the tiepoints for a specified band resolution.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
This routine will make sure that the chip windows for the determined tiepoints
don't fall outside the grid.  Thus for any given tiepoint, its chip window
(according to the chip_size) will be completely inside the grid for the
specified band resolution.

ALGORITHM REFERENCES:
******************************************************************************/
#include "b2b_char.h"
#include "ias_logging.h"          /* IAS logging library           */

int grid_det_tiepts
(
    const IAS_L1G_BAND_METADATA *band_metadata, /* I: band data             */
    const int nline_pts,        /* I: Number of line tiepoints to create    */
    const int nsamp_pts,        /* I: Number of sample tiepoints to create  */
    const struct IAS_LNG_LS *chip_size, /* I: Chip size of the tiepoints to 
                                      create                                */
    const int sca_index,        /* I: SCA index                             */
    const int band_index,       /* I: band index                            */
    const IAS_GRID_TYPE *grid,  /* I: OLI grid structure                    */
    int *numgcps,               /* O: Number of GCPs                        */
    IAS_GCP_RECORD *gcps        /* O: GCP structure                         */
)
{
    int i;                 /* Looping variable                              */
    int j;                 /* Looping variable                              */
    int status;            /* status of return from function                */
    int count;             /* Number of GCPs                                */
    double x_spacing;      /* Line spacing between chips                    */
    double y_spacing;      /* Sample spacing between chips                  */
    int nl;                /* Number of "raw input space" lines             */
    int ns;                /* Number of "raw input space" samples           */
    int onl;               /* Number of output space lines                  */
    int ons;               /* Number of output space samples                */
    double search_line;    /* Line location in L1G/L1T                      */
    double search_samp;    /* Samples location in L1G/L1T                   */
    double elev = 0.0;     /* Elevation of point, set to zero               */
    double iline;          /* Input line location for GCP location          */
    double isamp;          /* Input sample location for GCP location        */


    /* If only one line or sample point then return with an ERROR.  */
    if (nline_pts <= 1)
    {
        IAS_LOG_ERROR("This routine requires more than one line point.");
        return ERROR;
    }
    if (nsamp_pts <= 1)
    {
        IAS_LOG_ERROR("This routine requires more than one sample point.");
        return ERROR;
    }

    /* Determine the line and sample spacings for all other points between
     * the min and max points.  */
    nl = grid->gridbands[band_index][sca_index].in_lines[grid->
        gridbands[band_index][sca_index].ngrid_lines - 1];
    ns = grid->gridbands[band_index][sca_index].in_samps[grid->
        gridbands[band_index][sca_index].ngrid_samps - 1];
    onl = grid->gridbands[band_index][sca_index].lines; 
    ons = grid->gridbands[band_index][sca_index].samps; 

    x_spacing = (ns - chip_size->samp - 1) / (nsamp_pts - 1);
    y_spacing = (nl - chip_size->line - 1) / (nline_pts - 1);

    /* Fill in the complete GCP structures.  */
    for (i = 0, count = 0; i < nline_pts; i++)
    {
        if (i == nline_pts - 1)
            iline = nl - chip_size->line / 2.0 - 1.0; 
        else
            iline = chip_size->line / 2.0 + 1 + y_spacing * i; 

        for (j = 0; j < nsamp_pts; j++)
        {
            if (j == nsamp_pts - 1)
                isamp = ns - chip_size->samp / 2.0 - 1.0; 
            else
                isamp = chip_size->samp / 2.0 + 1 + x_spacing * j; 

            if (get_output_ls(&grid->gridbands[band_index][sca_index], elev, 
                iline, isamp, &search_line, &search_samp) != SUCCESS)
            {
               IAS_LOG_ERROR("Mapping point %f,%f to output.", iline, isamp);
            }

            if (search_line > chip_size->line 
                && search_line< (onl - chip_size->line - 1) 
                && search_samp > chip_size->samp 
                && search_samp < (ons - chip_size->samp - 1))
            {
                status = sprintf(gcps[count].point_id, "%d", count);
                if (status < 1)
                {
                    IAS_LOG_ERROR("In sprintf");
                    return ERROR;
                }
                gcps[count].reference_line = search_line;
                gcps[count].reference_sample = search_samp;
                gcps[count].pixel_size_x = 
                    band_metadata->projection_distance_x;  
                gcps[count].pixel_size_y = 
                    band_metadata->projection_distance_y;  
                gcps[count].chip_size_lines = chip_size->line;
                gcps[count].chip_size_samples = chip_size->samp;
                gcps[count].latitude = 0.0;
                gcps[count].longitude = 0.0;
                gcps[count].elevation = 0.0;
                count++;
            }
        }
    }

    *numgcps = count;

    return SUCCESS;
}
