/******************************************************************************
NAME:       process_gcp

PURPOSE:  Process all the GCPs (extract the image window for both the
          reference and search bands and perform the correlation for each
          point).

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
Correlation routines want things in sample, line order, so the fit_offset
returned are horizontal (sample) and then vertical (line) offsets.  GCP data
file contains fit_offset in line, sample order.

ALGORITHM REFERENCES:

******************************************************************************/
#include <stdlib.h>
#include "b2b_char.h"

#include "ias_math_parallel_corr.h"   /* Parallel correlator          */
#include "ias_logging.h"              /* IAS logging library          */
#include "ias_geo.h"                  /* Geographic library           */
#include "ias_satellite_attributes.h" /* Satellite attributes library */

int process_gcp 
(
    const IAS_GCP_RECORD *gcp_rec,/* I: List of GCPs                       */
    const int numgcps,            /* I: Number of GCPs                     */
    L1GIO *ref_img_ptr,           /* I: Reference 1G image pointer         */
    L1GIO *srch_img_ptr,          /* I: Search 1G image pointer            */
    const int ref_band_index,     /* I: Reference band index               */
    const int srch_band_index,    /* I: Search band index                  */
    const int sca_index,          /* I: SCA index to extract imagery from  */
    const IAS_DATA_TYPE ref_dtype,/* I: Data type of reference image       */
    const IAS_DATA_TYPE srch_dtype,/* I: Data type of search image         */
    const int nl,                 /* I: Number of lines in both images     */
    const int ns,                 /* I: Number of samples in both images   */
    const float min_corr,         /* I: Minimum correlation strength       */
    const float max_disp,         /* I: Maximum correlation displacement   */
    const IAS_CORRELATION_FIT_TYPE fit_method, /* I: Correlation result fit 
                                        method                             */
    IAS_GCP_RESULTS *gcp,         /* O: GCP data structures (1 per pt) 
                                        for the correlation results        */
    const int abs_corr_coeff,     /* I: flag to use the abs value of the
                                        correlation coeffs in grey 
                                        correlation                        */
    const float invalid_threshold,/* I: invalid threshold percentage for 
                                        buffer                             */
    const float valid_image_max,  /* I: DNs above this are invalid         */
    const float valid_image_min,  /* I: DNs below this are invalid         */
    IAS_PARALLEL_CORRELATOR_TYPE *correlator_ptr, /* I: Pointer to     
                                        the parallel correlator structure  */
    IAS_CORRELATION_RESULT_TYPE *result_ptr /* O: Pointer to the 
                                        correlation results                */
)
{
    int curgcp;                 /* GCP counter                          */
    int win_size_line_sample[2]; /* chip size to extract (line, samp)   */
    int win_size_sample_line[2]; /* chip size to extract (samp, line)   */
    static float *srch_buf;     /* float search correlation window      */
    static float *ref_buf;      /* float reference correlation window   */
    float nom_off[2]={0.0, 0.0};/* nominal correlation offset           */
    int status;                 /* function return status               */
    static int need_buffers = 1;/* flag to indicate buffers are needed 
                                   for the reference and search chip    */
    int ref_band_number;        /* 1-relative reference band number     */
    int srch_band_number;       /* 1-relative search band number        */
    IAS_DATA_TYPE dtype;        /* Data type dummy parameter            */
    int num_scas;               /* Num SCAs dummy parameter             */
    int nlines;                 /* Num lines dummy parameter            */
    int nsamps;                 /* Num samples dummy parameter          */
    L1G_BAND_IO *ref_l1g_band, *srch_l1g_band; /* L1G_BAND_IO pointers  */
    int sca_number;             /* SCA number                           */
    int chip_size;              /* The size of the current chip         */
    int upper_line;             /* The upper line in the image for the chip */
    int left_sample;            /* The left sample in the image for the chip */

    /* Initialize sca_number. */
    sca_number = sca_index + 1;

    /* initialize the results */
    for (curgcp = 0; curgcp < numgcps; curgcp++)
    {
        result_ptr[curgcp].valid = FALSE;
        result_ptr[curgcp].fit_offset[0] = 0.0;
        result_ptr[curgcp].fit_offset[1] = 0.0;
    }

    /* Open the reference band. */
    ref_band_number = ias_sat_attr_convert_band_index_to_number(ref_band_index);
    if ((ref_band_number < 1) || (ref_band_number > IAS_MAX_NBANDS))
    {
        IAS_LOG_ERROR("Invalid band %d", ref_band_number);
        return ERROR;
    }

    ref_l1g_band = ias_l1g_open_band(ref_img_ptr, ref_band_number, &dtype, 
                                     &num_scas, &nlines, &nsamps);
    if (ref_l1g_band == NULL)
    {
        IAS_LOG_ERROR("Opening band %d, sca %d, for extracting",
                 ref_band_number, sca_number);
        return ERROR;
    }

    /* Open the search band. */
    srch_band_number = ias_sat_attr_convert_band_index_to_number(
        srch_band_index);
    if ((srch_band_number < 1) || (srch_band_number > IAS_MAX_NBANDS))
    {
        IAS_LOG_ERROR("Invalid band %d", srch_band_number);
        return ERROR;
    }

    srch_l1g_band = ias_l1g_open_band(srch_img_ptr, srch_band_number, &dtype, 
                                      &num_scas, &nlines, &nsamps);
    if (srch_l1g_band == NULL)
    {
        IAS_LOG_ERROR("Opening band %d, sca %d, for extracting",
                 srch_band_number, sca_number);
        return ERROR;
    }

    /* Process each GCP.  */
    for (curgcp = 0; curgcp < numgcps; curgcp++)
    {
        /* get buffers for the chips if needed */
        if (need_buffers)
        {
            if (ias_math_get_corr_chip_buffers(correlator_ptr, &ref_buf, 
                &srch_buf) != SUCCESS)
            {
                IAS_LOG_ERROR("Getting buffers for chips");
                ias_l1g_close_band(ref_l1g_band);
                ias_l1g_close_band(srch_l1g_band);
                return ERROR;
            }

            need_buffers = 0;
        }

        /* Setup the size variables for the current gcp. */
        win_size_line_sample[0] = (int) gcp_rec[curgcp].chip_size_lines;
        win_size_line_sample[1] = (int) gcp_rec[curgcp].chip_size_samples;
        chip_size = win_size_line_sample[0] * win_size_line_sample[1];
        /* The grey correlation expects reference and search windows sizes to
           be in samp, line order, so we need to flip our window size values. */
        win_size_sample_line[0] = win_size_line_sample[1];
        win_size_sample_line[1] = win_size_line_sample[0];

        /* Extract the window from the 1G reference image.  The extract_window
           function converts the window to a buffer of float data. */
        status = ias_geo_extract_window(nl, ns, sca_index, ref_dtype,
               gcp_rec[curgcp].reference_line, 
               gcp_rec[curgcp].reference_sample,
               win_size_line_sample, ref_l1g_band, ref_buf,
               &upper_line, &left_sample);
        if (status == OUTSIDE)
        {
            IAS_LOG_INFO("GCP %s is outside reference image.", 
                           gcp_rec[curgcp].point_id);
            continue;
        }
        if (status == ERROR)
        {
            IAS_LOG_ERROR("Extracting the reference window for GCP %d.",
                     curgcp);
            ias_l1g_close_band(ref_l1g_band);
            ias_l1g_close_band(srch_l1g_band);
            return ERROR;
        }

        /* Extract the window from the 1G search image.  The extract_window
           function takes care of converting the window to a buffer of float 
           data.  */
        status = ias_geo_extract_window(nl, ns, sca_index, srch_dtype,
            gcp_rec[curgcp].reference_line, 
            gcp_rec[curgcp].reference_sample, 
            win_size_line_sample, srch_l1g_band, srch_buf,
            &upper_line, &left_sample);
        if (status == OUTSIDE)
        {
            IAS_LOG_INFO("GCP %s is outside search image.", 
                           gcp_rec[curgcp].point_id);
            continue;
        }
        if (status == ERROR)
        {
            IAS_LOG_ERROR("Extracting the search window for GCP %d.",
                     curgcp);
            ias_l1g_close_band(ref_l1g_band);
            ias_l1g_close_band(srch_l1g_band);
            return ERROR;
        }

        /* If the reference or search buffers have more than 
           <invalid_threshold>% out of range data, then don't 
           correlate them. */
        if (ias_math_check_pixels_in_range(ref_buf, chip_size, 
                    invalid_threshold, valid_image_max, valid_image_min) &&
            ias_math_check_pixels_in_range(srch_buf, chip_size, 
                    invalid_threshold, valid_image_max, valid_image_min))
        {
            /* Do the correlation */
            status = ias_math_submit_chip_to_corr(correlator_ptr, curgcp, 
                srch_buf, ref_buf, win_size_sample_line, win_size_sample_line,
                min_corr, fit_method, max_disp, nom_off, abs_corr_coeff);
            if (status == ERROR)
            {
                IAS_LOG_ERROR("In the grey correlation for GCP %d.", 
                    curgcp);
                ias_math_parallel_correlator_wait_for_results(correlator_ptr);
                ias_l1g_close_band(ref_l1g_band);
                ias_l1g_close_band(srch_l1g_band);
                return ERROR;
            }

            /* flag that the next chips will need new buffers */
            need_buffers = 1;
        }
        else
        {
            result_ptr[curgcp].valid = FALSE;
        }
    }

    /* Wait for the parallel correlator to finish. */
    if (ias_math_parallel_correlator_wait_for_results(correlator_ptr) 
        != SUCCESS)
    {
        IAS_LOG_ERROR("Waiting for parallel correlator results");
        ias_l1g_close_band(ref_l1g_band);
        ias_l1g_close_band(srch_l1g_band);
        return ERROR;    
    }

    /* extract the correlation results returned by the parallel correlator */
    for (curgcp = 0; curgcp < numgcps; curgcp++)
    {
        /* Check the returned flag to see if this correlation succeeded */
        if (result_ptr[curgcp].valid)
        {
            /* correlation succeeded */
            gcp[curgcp].accept_flag = TRUE;
            gcp[curgcp].fit_line_offset = result_ptr[curgcp].fit_offset[1];
            gcp[curgcp].fit_sample_offset = result_ptr[curgcp].fit_offset[0];
            gcp[curgcp].correlation_coefficient = result_ptr[curgcp].strength;
        }
        else
        {
            /* CORRELATION FAILURE */
            gcp[curgcp].accept_flag = FALSE;
            gcp[curgcp].fit_line_offset = 0.0;
            gcp[curgcp].fit_sample_offset = 0.0;
            gcp[curgcp].correlation_coefficient = 0.0; 
        }
    }

    /* Close the bands. */
    if (ias_l1g_close_band(ref_l1g_band) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing band %d, sca %d, for extracting",
                 ref_band_number, sca_number);
        ias_l1g_close_band(srch_l1g_band);
        return ERROR;
    }

    if (ias_l1g_close_band(srch_l1g_band) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing band %d, sca %d, for extracting",
                 srch_band_number, sca_number);
        return ERROR;
    }

    return SUCCESS;
}

