/******************************************************************************
NAME:           read_metadata

PURPOSE:
Read metadata for the specified bands.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
This routine reads the image and band metadata and makes sure each specified
band is in this structure.  The input bands array will specify which bands
the user wants to process.  The nbands parameter specifies the total number
of bands in the bands array.  The band metadata array should be of size 
MAX_BANDs, since there are MAX_BANDS possible band metadata structures to 
read from the file.  The metadata will be placed in the appropriate reference 
position for its corresponding band number.  This is similar to the manner of 
storing the different band elements of the grid structure.  If the specified 
band is not in the metadata, then read_metadata will return a fatal error 
status.

ALGORITHM REFERENCES:
See ias_l1g.h for a description of the image and band metadata structures.

******************************************************************************/
#include "b2b_char.h"
#include "ias_logging.h"              /* IAS logging library          */
#include "ias_satellite_attributes.h" /* Satellite attributes library */

int read_metadata 
(
    const char *l1g_filename,    /* I: L1G file name                         */
    const int *band_indexes,     /* I: Array of bands indexes to read        */
    const int nbands,            /* I: Number of indexes in the array        */
    IAS_L1G_FILE_METADATA *file_metadata, /* O: file-wide metadata           */
    IAS_L1G_BAND_METADATA band_metadata[IAS_MAX_NBANDS], /* O: Array of band 
                                        data                                 */
    int *nl,                      /* O: Number of lines                      */
    int *ns                       /* O: Number of samples                    */
)
{
    int i;                     /* loop counter                             */
    int sca;                   /* SCA number; just used to fill in param   */
    IAS_DATA_TYPE dtype;       /* Data type for the image                  */
    L1GIO *l1g_image;          /* L1G image pointer                        */
    int band_number;           /* 1-relative band number                   */


    /* Open the input image */
    l1g_image = ias_l1g_open_image(l1g_filename, IAS_READ);
    if (l1g_image == NULL)
    {
        IAS_LOG_ERROR("Opening the 1g image %s.", l1g_filename);
        return ERROR;
    }    

    /* Read input file's band metadata and number of lines and samples. */
    for (i = 0; i < nbands; i++)
    {
        band_number = ias_sat_attr_convert_band_index_to_number(
            band_indexes[i]);
        if ((band_number < 1) || (band_number > IAS_MAX_NBANDS))
        {
            IAS_LOG_ERROR("Invalid band number %d.", band_number);
            ias_l1g_close_image(l1g_image);
            return ERROR;
        }

        if (ias_l1g_get_band_metadata(l1g_image, band_number, &band_metadata[i])
                != SUCCESS)
        {
            IAS_LOG_ERROR("Reading metadata for band number %d from file %s.",
                band_number, l1g_filename);
            ias_l1g_close_image(l1g_image);
            return ERROR;
        }

        /* Get the number of lines and samples plus the datatype. */
        if (ias_l1g_get_band_size(l1g_image, band_number, &dtype, &sca,
            nl, ns) != SUCCESS)
        {
            IAS_LOG_ERROR("Getting size of band number %d from image %s.", 
                band_number, l1g_filename);
            ias_l1g_close_image(l1g_image);
            return ERROR;
        }
    }

    /* Read input file metadata. */
    if (ias_l1g_get_file_metadata(l1g_image, file_metadata) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading file metadata from file %s.", l1g_filename);
        ias_l1g_close_image(l1g_image);
        return ERROR;
    }

    /* Close the image data file. */
    
    if ((ias_l1g_close_image(l1g_image)) == ERROR)
    {
        IAS_LOG_ERROR("Closing image %s.", l1g_filename);
        ias_l1g_close_image(l1g_image);
        return ERROR;
    }

    return SUCCESS;
}
