/******************************************************************************
NAME:           reduce

PURPOSE:
Downsamples an image by convolving it with an oversample cubic convolution
weight function.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

******************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "b2b_char.h"
#include "ias_logging.h"              /* IAS logging library          */
#include "ias_satellite_attributes.h" /* Satellite attributes library */

int reduce 
(
    const char *in_name,   /* I: Original image                           */
    const char *out_name,  /* I: Output image name                        */
    const int band_index,  /* I: band index to process                    */
    const int down_scale,  /* I: Scaling of imagery                       */
    const double *weights, /* I: Resampleing weights                      */
    const int number_of_weights  /* I: Size of convolving function        */
)
{
    L1GIO *out_image;            /* Downsampled image                     */
    L1GIO *in_image;             /* Original image                        */
    int pos_line;                /* Current line during resampling        */
    int pos_samp;                /* Current line during resampling        */
    int in_nlines;               /* Number of lines in the input image    */
    int in_nsamps;               /* Number of samples in the input image  */
    int out_nlines;              /* Number of lines in the output image   */
    int out_nsamps;              /* Number of samples in the output image */
    int line;                    /* Line counter                          */
    int out_line;                /* Line counter                          */
    int samp;                    /* Sample counter                        */
    int out_samp;                /* Sample counter                        */
    int weight_counter;          /* Resampling weight counter             */
    int i;                       /* Loop counter                          */
    int j;                       /* Loop counter                          */
    IAS_DATA_TYPE in_dtype;      /* Data type of the input image          */
    IAS_DATA_TYPE out_dtype;     /* Data type of the output image         */

    void *img_buff=NULL;         /* Generic image buffer                  */
    unsigned char *in_buff_uchar=NULL; /* Image buffer if dtype is BYTE   */
    short int *in_buff_short_int=NULL; /* Image buffer if dtype is I2     */
                                       /* Image buffer if dtype is UI2    */
    unsigned short int *in_buff_short_uint=NULL;

    void *out_img_buff=NULL;     /* Generic image buffer                  */
    unsigned char *out_buff_uchar=NULL; /* Image buffer if dtype is BYTE  */
    short int *out_buff_short_int=NULL; /* Image buffer if dtype is I2    */
                                        /* Image buffer if dtype is UI2   */
    unsigned short int *out_buff_short_uint=NULL;

    IAS_L1G_FILE_METADATA file_metadata;  /* file-wide metadata           */
    IAS_L1G_BAND_METADATA band_metadata;  /* per-band metadata            */
    int band_number;                      /* 1-relative band number       */
    int sca_counter;             /* Counter for the SCAs                  */
    int num_scas;                /* Number of SCAs to set up in the image */
    float temp = 0.0;
    double total;                /* Sum of applied resampling kernal      */
    L1G_BAND_IO *in_l1g_band, *out_l1g_band; /* L1G_BAND_IO pointers */


    /* Open the input image */
    in_image = ias_l1g_open_image(in_name, IAS_READ);
    if (in_image == NULL)
    {
        IAS_LOG_ERROR("Opening the 1g image %s.", in_name);
        return ERROR;
    }

    /* Read input file metadata. */ 
    if (ias_l1g_get_file_metadata(in_image, &file_metadata) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading file metadata");
        ias_l1g_close_image(in_image);
        return ERROR;
    }

    /* Read input file's band metadata. */
    band_number = ias_sat_attr_convert_band_index_to_number(band_index);
    if ((band_number < 1) || (band_number > IAS_MAX_NBANDS))
    {
        IAS_LOG_ERROR("Invalid band number %d", band_number);
        ias_l1g_close_image(in_image);
        return ERROR;
    }
    if (ias_l1g_get_band_metadata(in_image, band_number, &band_metadata)
            != SUCCESS)
    {
        IAS_LOG_ERROR("Reading band metadata");
        ias_l1g_close_image(in_image);
        return ERROR;
    }

    /* Open the input image band. */
    in_l1g_band = ias_l1g_open_band(in_image, band_number, &in_dtype, &num_scas,
                                 &in_nlines, &in_nsamps);
    if (in_l1g_band == NULL)
    {
        IAS_LOG_ERROR("Opening band for reading");
        ias_l1g_close_image(in_image);
        return ERROR;
    }

    /* The output image will have half as many lines and samples and be
       of data type dtype. */
    out_nlines = (in_nlines - 1) / 2 + 1;
    out_nsamps = (in_nsamps - 1) / 2 + 1;

    /* Fix the projection distance values to match the change in lines. */
    band_metadata.projection_distance_x *= 2;
    band_metadata.projection_distance_y *= 2;

    /* For now set output data type to be equal to input data type */
    out_dtype = in_dtype;

    /* Show the image size. */ 
    IAS_LOG_INFO("%s size is %d,%d", out_name, out_nlines, out_nsamps);

    /* Open the output image */
    out_image = ias_l1g_open_image(out_name, IAS_WRITE);
    if (out_image == NULL)
    {
        IAS_LOG_ERROR("Opening the 1g image %s.", out_name);
        ias_l1g_close_band(in_l1g_band);
        ias_l1g_close_image(in_image);
        return ERROR;
    }

    /* Write output file's file and band metadata. */
    if ((ias_l1g_set_file_metadata(out_image, &file_metadata) != SUCCESS)
        || (ias_l1g_set_band_metadata(out_image, &band_metadata, 1)
        != SUCCESS))
    {
        IAS_LOG_ERROR("Updating metadata in %s", out_name);
        ias_l1g_close_band(in_l1g_band);
        ias_l1g_close_image(in_image);
        ias_l1g_close_image(out_image);
        return ERROR;
    }

    /* Open the output image band. */
    out_l1g_band = ias_l1g_open_band(out_image, band_number, &out_dtype,
                                     &num_scas, &out_nlines, &out_nsamps);
    if (out_l1g_band == NULL)
    {
        IAS_LOG_ERROR("Opening band for writing");
        ias_l1g_close_band(in_l1g_band);
        ias_l1g_close_image(in_image);
        ias_l1g_close_image(out_image);
        return ERROR;
    }

    /* Malloc space for the input data buffers.  */
    if (in_dtype == IAS_I2)
    {
        in_buff_short_int = malloc(in_nlines * in_nsamps * sizeof(short int));
        img_buff = in_buff_short_int;
    }
    else if (in_dtype == IAS_UI2)
    {
        in_buff_short_uint = malloc(in_nlines * in_nsamps * 
                               sizeof(unsigned short int));
        img_buff = in_buff_short_uint;
    }
    else if (in_dtype == IAS_BYTE)
    {
        in_buff_uchar = malloc(in_nlines * in_nsamps * sizeof(unsigned char));
        img_buff = in_buff_uchar;
    }
    else
    {
        IAS_LOG_ERROR("In the data type for the 1g image %s."
                "  Valid data types are BYTE, I2 and UI2.", in_name);
        ias_l1g_close_band(in_l1g_band);
        ias_l1g_close_image(in_image);
        ias_l1g_close_band(out_l1g_band);
        ias_l1g_close_image(out_image);
        return ERROR;
    }

    if (img_buff == NULL)
    {
        IAS_LOG_ERROR("Allocating buffer space.");
        ias_l1g_close_band(in_l1g_band);
        ias_l1g_close_image(in_image);
        ias_l1g_close_band(out_l1g_band);
        ias_l1g_close_image(out_image);
        return ERROR;
    }

    /* Malloc space for the output data buffers. */
    if (out_dtype == IAS_I2)
    {
        out_buff_short_int = malloc(out_nsamps * sizeof(short int));
        out_img_buff = out_buff_short_int;
    }
    else if (out_dtype == IAS_UI2)
    {
        out_buff_short_uint = malloc(out_nsamps * sizeof(unsigned short int));
        out_img_buff = out_buff_short_uint;
    }
    else if (out_dtype == IAS_BYTE)
    {
        out_buff_uchar = malloc(out_nsamps * sizeof(unsigned char));
        out_img_buff = out_buff_uchar;
    }
    else
    {
        IAS_LOG_ERROR("In the data type for the 1g image %s."
                "  Valid data types are BYTE, I2 and UI2.", in_name);
        free(img_buff);
        ias_l1g_close_band(in_l1g_band);
        ias_l1g_close_image(in_image);
        ias_l1g_close_band(out_l1g_band);
        ias_l1g_close_image(out_image);
        return ERROR;
    }

    if (out_img_buff == NULL)
    {
        IAS_LOG_ERROR("Allocating buffer space.");
        free(img_buff);
        ias_l1g_close_band(in_l1g_band);
        ias_l1g_close_image(in_image);
        ias_l1g_close_band(out_l1g_band);
        ias_l1g_close_image(out_image);
        return ERROR;
    }

    for (sca_counter = 0; sca_counter < num_scas; sca_counter++)
    {
        /* Display some downsampling information. */
        IAS_LOG_INFO("Downsampling SCA %02d", sca_counter + 1);

        /* Read the input image band. */ 
        if (ias_l1g_read_image(in_l1g_band, sca_counter, 0, 0,
            in_nlines, in_nsamps, img_buff) == ERROR) 
            
        {
            IAS_LOG_ERROR("Reading 1g image %s.", in_name);
            free(img_buff);
            free(out_img_buff);
            ias_l1g_close_band(in_l1g_band);
            ias_l1g_close_image(in_image);
            ias_l1g_close_band(out_l1g_band);
            ias_l1g_close_image(out_image);
            return ERROR;
        }

        for (line = 0, out_line = 0; line < in_nlines; 
             line = line + down_scale, out_line++)
        {
            for (samp = 0, out_samp = 0; samp < in_nsamps; 
                 samp = samp + down_scale, out_samp++)
            {
                total = 0.0;
                weight_counter = 0;
                for (i =- (number_of_weights - 1)/2; 
                     i <= (number_of_weights - 1)/2 ; 
                     i++)
                {
                    if (line + i < 0)
                        pos_line = -line - i;
                    else if (line + i >= in_nlines)
                        pos_line = 2 * in_nlines - line - i - 1; 
                    else
                        pos_line = line + i;
                    for (j =- (number_of_weights - 1)/2; 
                         j <= (number_of_weights - 1)/2 ; j++)
                    {
                        if (samp + j < 0)
                            pos_samp = -samp - j;
                        else if (samp + j >= in_nsamps)
                            pos_samp = 2 * in_nsamps - samp - j - 1; 
                        else
                            pos_samp = samp + j;

                        if (in_dtype == IAS_I2)
                            temp = (float) in_buff_short_int[in_nsamps
                                * pos_line + pos_samp];
                        else if (in_dtype == IAS_UI2)
                            temp = (float) in_buff_short_uint[in_nsamps
                                * pos_line + pos_samp];
                        else if (in_dtype == IAS_BYTE)
                            temp = (float) in_buff_uchar[in_nsamps
                                * pos_line + pos_samp];

                        total = total + temp * weights[weight_counter];
                        weight_counter++;
                    }
                }

                if (out_dtype == IAS_I2)
                {
                    if (total < SHRT_MIN)
                        out_buff_short_int[out_samp] = SHRT_MIN;
                    else if (total > SHRT_MAX)
                        out_buff_short_int[out_samp] = SHRT_MAX;
                    else
                        out_buff_short_int[out_samp] = total;
                }
                else if (out_dtype == IAS_UI2)
                {
                    if (total < 0)
                        out_buff_short_uint[out_samp] = 0;
                    else if (total > USHRT_MAX)
                        out_buff_short_uint[out_samp] = USHRT_MAX;
                    else
                        out_buff_short_uint[out_samp] = total;
                }
                else if (out_dtype == IAS_BYTE)
                {
                    if (total < 0)
                       out_buff_uchar[out_samp] = 0;
                    else if (total > UCHAR_MAX)
                       out_buff_uchar[out_samp] = UCHAR_MAX;
                    else
                       out_buff_uchar[out_samp] = total;
                }
            }

            /* Write the output image band. */
            if (ias_l1g_write_image(out_l1g_band, sca_counter, out_line, 0, 
                1/*lines*/, out_nsamps, (void *) out_img_buff) == ERROR)
            {
                IAS_LOG_ERROR("Writing the 1g image %s.", out_name);
                free(img_buff);
                free(out_img_buff);
                ias_l1g_close_band(in_l1g_band);
                ias_l1g_close_image(in_image);
                ias_l1g_close_band(out_l1g_band);
                ias_l1g_close_image(out_image);
                return ERROR;
            } 
        }
    } /* SCA counter */

    /* Free the buffers.  */
    free(img_buff);
    free(out_img_buff);

    /* Close the output image band. */
    if (ias_l1g_close_band(out_l1g_band) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing band %d in the 1g image %s.", 
                band_number, out_name);
        ias_l1g_close_band(in_l1g_band);
        ias_l1g_close_image(in_image);
        ias_l1g_close_image(out_image);
        return ERROR;
    }

    /* Close the L1G images */
    if (ias_l1g_close_image(out_image) == ERROR)
    {
        IAS_LOG_ERROR("Closing the 1g image %s.", out_name);
        ias_l1g_close_band(in_l1g_band);
        ias_l1g_close_image(in_image);
        return ERROR;
    }

    /* Close the input image band. */
    if (ias_l1g_close_band(in_l1g_band) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing band %d in the 1g image %s.",
            band_number, in_name);
        ias_l1g_close_image(in_image);
        return ERROR;
    }

    if (ias_l1g_close_image(in_image) == ERROR)
    {
        IAS_LOG_ERROR("Closing the 1g image %s.", in_name);
        return ERROR;
    }

    return SUCCESS;
}
