/***************************************************************************
 NAME:        lunar_b2b_char_get_parms_test

 PURPOSE:     Calls the 'get_parms' function to retrieve parameters for
              running 'b2b_char' on a lunar collect, and verifies that
              the parameter 'LUNAR_SCA' in the OMF file was read in
              and properly handled.

****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "ias_satellite_attributes.h"
#include "b2b_char.h"
#include "cgreen.h"

#define B2B_SCA_ERROR -1    /* Invalid SCA value in sca_indices array */

Ensure test_lunar_b2b_char_get_parms()
{
    struct IAS_LNG_LS chip_size;
    enum TIEPT_TYPE tiept_type;
    IAS_CORRELATION_FIT_TYPE fit_method;

    char omf_filename[PATH_MAX];
    char L1G_filename[PATH_MAX];
    char L0R_filename[PATH_MAX];
    char grid_filename[PATH_MAX];
    char cpf_filename[PATH_MAX];
    char output_filename[PATH_MAX];
    char work_order_id[IAS_WORKORDER_ID_SIZE];

    float range[2];
    float min_corr;
    float max_displacement;
    float invalid_threshold;
    int sca_indices[IAS_MAX_NSCAS];
    int band_indices[IAS_MAX_NBANDS];
    int i;                                     /* loop counter */
    int abs_corr_coeff;
    int reduce_flag = 0;
    int number_of_bands;
    int number_of_scas;
    int number_of_line_tiepts;
    int number_of_sample_tiepts;
    int status;


    printf("Checking read of 'LUNAR_SCA' OMF parameter...\n");

    /* Call 'get_parms' for the case where the OMF has the Lunar SCA
       parameter but the ODL does not */
    status = get_parms("B2BCHAR.odl", L1G_filename, L0R_filename,
        output_filename, grid_filename, cpf_filename, &reduce_flag,
        band_indices, &number_of_bands, sca_indices, &number_of_scas,
        &number_of_line_tiepts, &number_of_sample_tiepts, &chip_size,
        &tiept_type, &min_corr, &max_displacement, range,
        &invalid_threshold, &fit_method, &abs_corr_coeff, work_order_id,
        omf_filename);

    assert_equal_with_message(status, SUCCESS, "Error retrieving "
        "parameters from parameter file");

    /* The 'LUNAR_SCA' number was read in.  The only SCA to search
       should be represented by the index corresponding to the lunar SCA
       number.  Check the SCA index array to confirm there's only one valid
       entry (in the first element), and that the number of SCAs to search
       should be 1. */
    assert_equal_with_message(1, number_of_scas,
        "Number of SCAs returned %d does not match expected number %d",
        number_of_scas, 1);
    assert_equal(sca_indices[0], 6);

    /* Now the case where the parameter file overrides the OMF with the first
       9 SCAs requested */
    status = get_parms("override_lunar.odl", L1G_filename, L0R_filename,
        output_filename, grid_filename, cpf_filename, &reduce_flag,
        band_indices, &number_of_bands, sca_indices, &number_of_scas,
        &number_of_line_tiepts, &number_of_sample_tiepts, &chip_size,
        &tiept_type, &min_corr, &max_displacement, range,
        &invalid_threshold, &fit_method, &abs_corr_coeff, work_order_id,
        omf_filename);

    assert_equal_with_message(status, SUCCESS, "Error retrieving "
        "parameters from parameter file");
    assert_equal_with_message(9, number_of_scas,
        "Number of SCAs returned %d does not match expected number %d",
        number_of_scas, 9);

    for (i = 0; i < IAS_MAX_NSCAS; i++)
    {
        if (i < 9)
            assert_equal(i, sca_indices[i]);
        else
            assert_equal(sca_indices[i], B2B_SCA_ERROR);
    }

    /* Now the case where the OMF does not contain LUNAR_SCA_NUMBER and the
       parameter file does not have the SCA_LIST parameter, so it should
       default to all 14 SCAs. */
    status = get_parms("no_sca_list.odl", L1G_filename, L0R_filename,
        output_filename, grid_filename, cpf_filename, &reduce_flag,
        band_indices, &number_of_bands, sca_indices, &number_of_scas,
        &number_of_line_tiepts, &number_of_sample_tiepts, &chip_size,
        &tiept_type, &min_corr, &max_displacement, range,
        &invalid_threshold, &fit_method, &abs_corr_coeff, work_order_id,
        omf_filename);

    assert_equal_with_message(status, SUCCESS, "Error retrieving "
        "parameters from parameter file");
    assert_equal_with_message(14, number_of_scas,
        "Number of SCAs returned %d does not match expected number %d",
        number_of_scas, 14);

    for (i = 0; i < IAS_MAX_NSCAS; i++)
    {
        assert_equal(i, sca_indices[i]);
    }
}


int main(int argc, char *argv[])
{
    /* Construct the test suite */
    TestSuite *suite = create_test_suite();
    add_test(suite, test_lunar_b2b_char_get_parms);

    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    else
    {
        return run_test_suite(suite, create_text_reporter());
    }
}
