/****************************************************************************
NAME: bandcal

PURPOSE:
  This program finds the relative band placement by finding the Legendre
  polynomials between estimated partial derivatives associated with each band 
  and the measured values.  This will give relative band placement values that 
  will be converted to IFOVs (microradians), then these values can be used to
  better register the bands for a systematic product.

RETURN VALUE:
Type = int
Value    Description
-----    -----------
EXIT_SUCCESS  Successful completion 
EXIT_FAILURE  Operation failed

*******************************************************************************/
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>                 /* For strcmp */

#include "gctp.h"                   /* UTM, SOM */

#include "ias_const.h"
#include "ias_types.h"
#include "ias_structures.h"
#include "ias_cpf.h"
#include "ias_geo.h"                /* Used to check effective dates */
#include "ias_satellite_attributes.h" /* Satellite attributes library */
#include "ias_miscellaneous.h"      /* Miscellaneous libraries */
#include "ias_logging.h"            /* Logging library */
#include "ias_model_io.h"           /* IAS model I/O library */
#include "ias_los_model.h"
#include "ias_grid.h"
#include "ias_grid_io.h"

#include "bandcal.h"                /* Bandcal header */
#include "save_to_database.h"

#define TOLERANCE 0.001             /* Grid resolution tolerance in meters */
#define EFFECTIVE_DATE_SIZE 20      /* Size of effective date string */
#define BAND_SCA_NOT_PRESENT -1     /* Identifies the Band and SCA combination
                                       as not being present */

static int check_grid_resolutions
(
    const IAS_GRID_TYPE *grid,       /* I: Grid that will be checked */
    int num_bands,                   /* I: Total number of bands */
    int band_numbers[IAS_MAX_NBANDS] /* I: List of band numbers */
);

int main(int argc, char *argv[]) 
{
    char dem_filename[PATH_MAX];    /* Name of the DEM file */
    char grid_file[PATH_MAX];       /* Name of grid file */
    char model_file[PATH_MAX];      /* Name of model file */
    char b2b_char_file[PATH_MAX];   /* Name of input band to band char file */
    char b2b_cal_file[PATH_MAX];    /* Name of output band to band cal file */
    char cpf_filename[PATH_MAX];    /* Name of the calibration parameter file */
    char output_cpf_filename[PATH_MAX]; /* Name of file for writing Band 
                                       Calibration parameters CPF group */
    char l0r_file[PATH_MAX];        /* L0R image filename */
    char l1g_file[PATH_MAX];        /* L1G image filename */
    int cpf_group_flag;             /* Flag to write CPF group to file */
    int tie_point_list_flag;        /* Flag to write tie point list to file */
    int minimum_points;             /* Minimum points for omitted bands */
    int sca_index;                  /* Index into SCAs array */
    int combo_index;                /* Index into band combination arrays */
    int number_of_combinations;     /* Number of band combinations */
    double new_sca_coef_x[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS];   /* New along-track Legendre coefficients */
    double new_sca_coef_y[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS];   /* New across-track Legendre coefficients */
    IAS_CPF *cpf;                    /* CPF parameters */
    const struct IAS_CPF_FILE_ATTRIBUTES *cpf_file_atts; 
                                    /* CPF File Attributes group */
    int start_effective_date = 0;   /* Starting effective date (YYYYMMDD) for 
                                       CPF update file */
    int end_effective_date = 0;     /* Ending effective date (YYYYMMDD) for CPF 
                                       update file */
    char start_effective_date_str[EFFECTIVE_DATE_SIZE]; /* CPF string for start
                                       effective date */
    char end_effective_date_str[EFFECTIVE_DATE_SIZE]; /* CPF string for end 
                                       effective date */
    int year;                       /* year extracted from effective date */
    int month;                      /* month extracted from effective date */
    int day;                        /* day extracted from effective date */
    int number_of_scas;             /* Number of SCAs to process */
    int sca_indexes[IAS_MAX_NSCAS]; /* SCA indexes in B2B char file */
    IAS_GRID_TYPE grid;             /* Geometric resampling grid structure */
    POINT *points = NULL;           /* B2B correlation information */
    int num_points;                 /* Number of points in correlation file */
    BANDCOMBO bandcombo;            /* Possible band combinations */

    int reference_indexes[NCOMBO] = /* Reference part of band combinations */
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     1, 1, 1, 1, 1, 1, 1, 1, 1,
     2, 2, 2, 2, 2, 2, 2, 2,
     3, 3, 3, 3, 3, 3, 3,
     4, 4, 4, 4, 4, 4,
     5, 5, 5, 5, 5,
     6, 6, 6, 6,
     7, 7, 7,
     8, 8,
     9};

    int search_indexes[NCOMBO] =    /* Search part of band combinations */
    {1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
     2, 3, 4, 5, 6, 7, 8, 9, 10,
     3, 4, 5, 6, 7, 8, 9, 10,
     4, 5, 6, 7, 8, 9, 10,
     5, 6, 7, 8, 9, 10,
     6, 7, 8, 9, 10,
     7, 8, 9, 10,
     8, 9, 10,
     9, 10,
     10}; 

    int combo_present[NCOMBO];      /* Flags to show if band combination 
                                       present */
    int point;                      /* Loop variable */
    int ref_band_index;             /* Reference band index */
    int search_band_index;          /* Search band index */
    int ref_band_number;            /* Reference band number */
    int search_band_number;         /* Search band number */
    double along_track_solution[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
                          [IAS_LOS_LEGENDRE_TERMS]; 
    double across_track_solution[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
                           [IAS_LOS_LEGENDRE_TERMS];
                                    /* The solution: offsets for band centers */
    int nsca_in_file;               /* Number of SCAs in B2B char file */
    IAS_IMAGE dem_image;            /* Structure to hold DEM information */
    int num_coeffs;                 /* Number of coefficients to solve for */
    IAS_LOS_MODEL *model;           /* Model structure */
    char work_order_id[IAS_WORKORDER_ID_SIZE]; /* Work order ID string */
    struct MAIN_DB_SAVE_INFO main_db_save; /* I: top-level data to save */
    struct BAND_SCA_DB_SAVE_INFO band_sca_db_save; /* I: Band/SCA-level data to
                                       save */
    struct PRE_POST_FIT_DB_SAVE_INFO pre_post_fit_db_save; /* I: Pre and post
                                       fit data to save */
    int store_in_database_flag;     /* Flag to indicate algorithm results 
                                       should be written to the database */
    int num_band_sca_db_records = 0; /* Number of band/SCA records to save to 
                                       database */
    int num_pre_post_fit_db_records = 0; /* Number of pre/post fit records to 
                                       save to database */
    IAS_SATELLITE_ID satellite_id;  /* Satellite ID */
    int unit_code;                  /* Projection unit code */
    int status;                     /* Status of return from functions */

    int current_resolution; /* For searching for the lowest resolution */
    int dem_band_index;     /* Index of the band to use for reading the DEM */
    int num_bands;          /* Total number of bands */
    int band_numbers[IAS_MAX_NBANDS]; /* List of band numbers */
    int band;               /* Band loop counter */
    int band_index;         /* Band index */ 
    int coeff;              /* Coefficient loop counter */

    char msg_string[512];   /* To setup messages */
    const char *sensor_id_string; /* The name of the sensor */

    /* Initialize the logging library */
    if (ias_log_initialize("bandcal") != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing logging library");
        return EXIT_FAILURE;
    }

    /* Initialize the DB save structures */
    memset(&main_db_save, 0, sizeof(main_db_save));
    memset(&band_sca_db_save, 0, sizeof(band_sca_db_save));
    memset(&pre_post_fit_db_save, 0, sizeof(pre_post_fit_db_save));

    /* Check the number of arguments.  We want the program name and the WO 
       parameter file as arguments */
    if (argc != 2)
    {
        IAS_LOG_ERROR("Usage: bandcal parameter_file");
        return EXIT_FAILURE;
    }

    /* Initialization */
    for (band = 0; band < IAS_MAX_NBANDS; band++)
    {
        for (sca_index = 0; sca_index < IAS_MAX_NSCAS; sca_index++)
        {
            for (coeff = 0; coeff < IAS_LOS_LEGENDRE_TERMS; coeff++)
            {
                new_sca_coef_x[band][sca_index][coeff] = 0.0;
                new_sca_coef_y[band][sca_index][coeff] = 0.0;
            }
        }
    }

    /* Get parameters from the WO parameter file */
    status = get_parameters(argv[1], dem_filename, l0r_file, l1g_file,
        grid_file, model_file, b2b_char_file, b2b_cal_file, cpf_filename,
        &num_coeffs, &store_in_database_flag, &cpf_group_flag,
        &tie_point_list_flag, &minimum_points, &start_effective_date,
        &end_effective_date, work_order_id, &satellite_id); 
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving parameters");
        return EXIT_FAILURE;
    } 

    /* Message to the user */
    IAS_LOG_INFO("Initialized");

    /* Initialize satellite attributes */
    status = ias_sat_attr_initialize(satellite_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attributes library");
        return EXIT_FAILURE;
    }

    /* Initialize combo_present */
    for (combo_index = 0; combo_index < NCOMBO; combo_index++)
    {
        combo_present[combo_index] = FALSE;
    }

    /* Setup the band combinations that we expect */
    for (combo_index = 0; combo_index < NCOMBO; combo_index++)
    {
        for (sca_index = 0; sca_index < IAS_MAX_NSCAS; sca_index++)
        {
            bandcombo.reference[combo_index][sca_index] = BAND_SCA_NOT_PRESENT;
            bandcombo.search[combo_index][sca_index] = BAND_SCA_NOT_PRESENT;
        }
    }

    /* If we are going to generate a CPF report file, then make sure we have
       valid effective dates. */
    if (cpf_group_flag)
    {
        /* If the effective date parameters were not read from the parameter
           file, get them from the CPF */
        if ((start_effective_date == 0) || (end_effective_date == 0))
        {
            /* Read the CPF */
            cpf = ias_cpf_read(cpf_filename);
            if (!cpf)
            {
                IAS_LOG_ERROR("Cannot open CPF file %s for reading", 
                    cpf_filename);
                return EXIT_FAILURE;
            }

            /* Get the file attributes group */
            cpf_file_atts = ias_cpf_get_file_attributes(cpf);
            if (!cpf_file_atts)
            {
                IAS_LOG_ERROR("Getting File Attributes group from CPF %s",
                    cpf_filename);
                ias_cpf_free(cpf);
                return EXIT_FAILURE;
            }

            /* Get the effective dates */
            strcpy(start_effective_date_str, 
                cpf_file_atts->effective_date_begin);
            strcpy(end_effective_date_str, cpf_file_atts->effective_date_end);

            /* Format the effective dates and put them in integers.  They are
               in form YYYY-MM-DDTHH:MM:SS.  We want YYYYMMDD */
            status = sscanf(start_effective_date_str, "%4d-%2d-%2d", &year, 
                &month, &day);
            if (status != 3)
            {
                IAS_LOG_ERROR("Could not scan start effective date from "
                    "string %s", start_effective_date_str);
                ias_cpf_free(cpf);
                return EXIT_FAILURE;
            }
            start_effective_date = 10000 * year + 100 * month + day;

            status = sscanf(end_effective_date_str, "%4d-%2d-%2d", &year, 
                &month, &day);
            if (status != 3)
            {
                IAS_LOG_ERROR("Could not scan end effective date from string "
                     "%s", end_effective_date_str);
                ias_cpf_free(cpf);
                return EXIT_FAILURE;
            }
            end_effective_date = 10000 * year + 100 * month + day;

            /* Free CPF memory */
            ias_cpf_free(cpf);
        }

        /* Check start and end effective dates for CPF group report file */
        if (ias_geo_check_start_end_date(start_effective_date, 
            end_effective_date) != SUCCESS)
        {
            IAS_LOG_ERROR("In start effective date %d or end effective date "
                "%d", start_effective_date, end_effective_date);
            return EXIT_FAILURE;
        }
    } /* END - cpf_group_flag */

    /* Read the grid file */
    status = ias_grid_read(grid_file, NULL, 0, &grid);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading grid file: '%s'", grid_file);
        return EXIT_FAILURE;
    }

    /* Read the model file */
    model = ias_model_read(model_file);
    if (model == NULL)
    {
        IAS_LOG_ERROR("Reading model file %s", model_file);
        ias_grid_free(&grid);
        return EXIT_FAILURE;
    }

    /* Check that the projection is UTM or SOM.  Produce a warning, not an
       error, if it's not one of these projections */
    if (model->acquisition_type == IAS_EARTH)
    {
        if ((grid.code != UTM) && (grid.code != SOM))
        {
            IAS_LOG_WARNING("Grid %s is not in UTM or SOM projection", 
                grid_file);
        }
    }

    /* The JPL setup is only needed for Lunar/Stellar acquisitions */
    if (model->acquisition_type == IAS_STELLAR || 
        model->acquisition_type == IAS_LUNAR)
    {
        if (ias_geo_initialize_novas() != SUCCESS)
        {
            IAS_LOG_ERROR("Initializing the NOVAS library");
            ias_grid_free(&grid);
            ias_los_model_free(model);
            model = NULL;
            return EXIT_FAILURE;
        }
    }

    /* Check for units of grid. */ 
    status = ias_geo_get_units(grid.units, &unit_code);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Converting projection units %s to unit code",
            grid.units);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        model = NULL;
        return EXIT_FAILURE;
    }

    if (model->acquisition_type == IAS_EARTH)
    {
        if (unit_code != METER)
        {
            IAS_LOG_ERROR("Grid file must have units of meters for IAS_EARTH "
                "acquisition type; units are '%s'", grid.units);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            model = NULL;
            return EXIT_FAILURE;
        }

        /* The DEM is optional, but it's required for Earth scenes */
        if (strcmp(dem_filename, "") == 0)
        {
            IAS_LOG_ERROR("Input DEM filename must be given for IAS_EARTH "
                "acquisition type");
            ias_grid_free(&grid);
            ias_los_model_free(model);
            model = NULL;
            return EXIT_FAILURE;
        }
    }
    else
    {
        if (unit_code != SECOND)
        {
            IAS_LOG_ERROR("Grid file must have units of radians for "
                "acquisition types other than IAS_EARTH; units are '%s'",
                grid.units);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            model = NULL;
            return EXIT_FAILURE;
        }
    }

    /* Read the band to band characterization output file, which contains the 
       band to band correlations */
    points = read_b2b_char_output(b2b_char_file, &num_points, &nsca_in_file, 
        sca_indexes, &bandcombo); 
    if (points == NULL)
    {
        IAS_LOG_ERROR("Reading band to band characterization file %s",
                b2b_char_file);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        model = NULL;
        return EXIT_FAILURE;
    }
    number_of_scas = nsca_in_file;

    /* Get band info */
    status = ias_sat_attr_get_sensor_band_numbers(bandcombo.sensor_id,
        IAS_NORMAL_BAND, 0, band_numbers, IAS_MAX_NBANDS, &num_bands);
    if (status != SUCCESS)
    {
        strncpy(msg_string, "Getting normal band numbers for the %s sensor",
                sizeof(msg_string));
        sensor_id_string = ias_sat_attr_get_sensor_name(bandcombo.sensor_id);

        IAS_LOG_ERROR(msg_string, sensor_id_string);

        ias_grid_free(&grid);
        ias_los_model_free(model);
        model = NULL;
        points = NULL;
        return EXIT_FAILURE;
    }

    /* Check the grid resolutions. */
    status = check_grid_resolutions(&grid, num_bands, band_numbers);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Checking grid resolutions");
        ias_grid_free(&grid);
        ias_los_model_free(model);
        model = NULL;
        points = NULL;
        return EXIT_FAILURE;
    }

    /* Check for all band combinations in the B2B Char output file.  We
       expect the band combinations to be in a particular order.  For example,
       we expect r=1 s=2 but not r=2 s=1. */
    for (point = 0; point < num_points; point++)
    {
        ref_band_index = ias_sat_attr_convert_band_number_to_index(
            points[point].reference_band_number);
        if (ref_band_index == ERROR)
        {
            IAS_LOG_ERROR("band_index not found for reference_band_number[%d]", 
                points[point].reference_band_number);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            model = NULL;
            free(points);
            points = NULL;
            return EXIT_FAILURE;
        }

        search_band_index = ias_sat_attr_convert_band_number_to_index(
            points[point].search_band_number);
        if (search_band_index == ERROR)
        {
            IAS_LOG_ERROR("band_index not found for search_band_number[%d]", 
                points[point].search_band_number);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            model = NULL;
            free(points);
            points = NULL;
            return EXIT_FAILURE;
        }

        for (combo_index = 0; combo_index < NCOMBO; combo_index++)
        {
            if (((reference_indexes[combo_index] == ref_band_index) &&
                 (search_indexes[combo_index] == search_band_index)) ||
                ((reference_indexes[combo_index] == search_band_index) &&
                 (search_indexes[combo_index] == ref_band_index)))
            {
                combo_present[combo_index] = TRUE;
            }
        }
    }

    /* Checking to see if we missed a band combination */
    /* Also find the number of band combinations present */
    number_of_combinations = 0;
    for (combo_index = 0; combo_index < NCOMBO; combo_index++)
    {
        if (combo_present[combo_index] != TRUE)
        {
            ref_band_number = ias_sat_attr_convert_band_index_to_number(
                reference_indexes[combo_index]);
            if (ref_band_number == ERROR)
            {
                IAS_LOG_ERROR(
                    "band_number not found for reference band_index[%d]",
                    reference_indexes[combo_index]);
                ias_grid_free(&grid);
                ias_los_model_free(model);
                model = NULL;
                free(points);
                points = NULL;
                return EXIT_FAILURE;
            }

            search_band_number = ias_sat_attr_convert_band_index_to_number(
                search_indexes[combo_index]);
            if (search_band_number == ERROR)
            {
                IAS_LOG_ERROR(
                    "band_number not found for search band_index[%d]",
                    search_indexes[combo_index]);
                ias_grid_free(&grid);
                ias_los_model_free(model);
                model = NULL;
                free(points);
                points = NULL;
                return EXIT_FAILURE;
            }

            IAS_LOG_WARNING("Missing band combination ref=%d, search=%d", 
                ref_band_number, search_band_number);
        }
        else
        {
            number_of_combinations++;
        }
    }

    if (model->acquisition_type == IAS_EARTH)
    {
        /* Loop through the bands, comparing resolutions to figure out which
           band to use for reading the DEM.
           This will choose the first lowest resolution band.
           (aka. the largest pixel value) */
        current_resolution = 0;
        dem_band_index = -1;
        for (band = 0; band < num_bands; band++)
        {
            /* Get the band index */
            band_index = ias_sat_attr_convert_band_number_to_index(
                             band_numbers[band]);
            if (band_index == ERROR)
            {
                IAS_LOG_ERROR("band_index not found for band_number[%d]",
                    band_numbers[band]);
                ias_grid_free(&grid);
                ias_los_model_free(model);
                model = NULL;
                free(points);
                points = NULL;
                return EXIT_FAILURE;
            }

            if (grid.bands_available[band_index])
            {
                if (grid.gridbands[band_index][0].pixsize > current_resolution)
                {
                    current_resolution = grid.gridbands[band_index][0].pixsize;
                    dem_band_index = band_index;
                }
            }
        }

        if (dem_band_index == -1)
        {
            IAS_LOG_ERROR("Unable to find best resolution in %s", dem_filename);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            model = NULL;
            free(points);
            points = NULL;
            return EXIT_FAILURE;
        }

        /* Read the DEM using information from the first lowest resolution
           band as reference. */
        if (ias_misc_read_coregistered_single_band_l1g(dem_filename,
                              &grid.corners, 
                              grid.gridbands[dem_band_index][0].lines,
                              grid.gridbands[dem_band_index][0].samps, 
                              grid.gridbands[dem_band_index][0].pixsize, 
                              &dem_image, IAS_I2) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading DEM from %s", dem_filename);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            model = NULL;
            free(points);
            points = NULL;
            return EXIT_FAILURE;
        }
    }
    else
    {
        dem_image.data = NULL;
    }

    /* Find new focal plane placement along and across track values */
    if (solve_focal_plane(l1g_file, l0r_file, b2b_cal_file, &grid, model,
                &dem_image, num_points, number_of_scas, sca_indexes,
                minimum_points, &bandcombo, num_coeffs, points,
                along_track_solution, across_track_solution, new_sca_coef_x,
                new_sca_coef_y, number_of_combinations, tie_point_list_flag,
                store_in_database_flag, &band_sca_db_save, 
                &pre_post_fit_db_save, work_order_id, &num_band_sca_db_records,
                &num_pre_post_fit_db_records) != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating focal plane placements");
        ias_grid_free(&grid);
        ias_los_model_free(model);
        model = NULL;
        free(points);
        points = NULL;
        free(dem_image.data);
        dem_image.data = NULL;
        return EXIT_FAILURE;
    }

    /* Free the structures */ 
    ias_grid_free(&grid);
    free(points);
    points = NULL;
    free(dem_image.data);
    dem_image.data = NULL;

    /* The JPL setup is only needed for Lunar/Stellar acquisitions */
    if (model->acquisition_type == IAS_STELLAR || 
        model->acquisition_type == IAS_LUNAR)
    {
        if (ias_geo_shutdown_novas() != SUCCESS)
        {
            IAS_LOG_ERROR("Shutting down the NOVAS library");
            ias_los_model_free(model);
            return EXIT_FAILURE;
        }
    }

    ias_los_model_free(model);
    model = NULL;

    /* If storing results in the database is requested */
    if (store_in_database_flag)
    {
        IAS_LOG_INFO("Writing data to database");

        /* Prepare main database save structure */ 
        strcpy(main_db_save.wo_id, work_order_id);
        main_db_save.nbr_of_legendre_coefficients = num_coeffs;

        /* Store results in the database */
        if (save_to_database(&main_db_save, &band_sca_db_save,
                             num_band_sca_db_records, &pre_post_fit_db_save, 
                             num_pre_post_fit_db_records) != SUCCESS)
        {
            IAS_LOG_ERROR("Writing data to database");
            return EXIT_FAILURE;
        }
    }

    /* Output the updated CPF parameters */
    if (cpf_group_flag)
    {
        /* Form the ASCII CPF Group filename */
        if (sprintf(output_cpf_filename, "bandcal_parameters_%08d_%08d.cpf",
                    start_effective_date, end_effective_date) < 0)
        {
            IAS_LOG_ERROR("Constructing the output CPF Group file name for "
                "dates %d to %d", start_effective_date, end_effective_date);
            return EXIT_FAILURE;
        }

        /* Write parameters to ASCII CPF Group file */  
        if (write_sca_parameters_cpf(output_cpf_filename, new_sca_coef_x, 
                                     new_sca_coef_y, &bandcombo) != SUCCESS)
        {
            IAS_LOG_ERROR("Writing the updated CPF Group file %s",
                output_cpf_filename);
            return EXIT_FAILURE;
        }
    }

    /* Notify the user of the program completion */
    IAS_LOG_INFO("Successful completion");
    return EXIT_SUCCESS;
}

/****************************************************************************
Name: check_grid_resolutions

Purpose: Local routine to verify that the grid resolutions are consistent
    across bands and SCAs

Returns: SUCCESS or ERROR
****************************************************************************/
static int check_grid_resolutions
(
    const IAS_GRID_TYPE *grid,       /* I: Grid that will be checked */
    int num_bands,                   /* I: Total number of bands */
    int band_numbers[IAS_MAX_NBANDS] /* I: List of band numbers */
)
{
    int sca_index;                  /* SCA index */
    int band;                       /* Band loop counter */
    int band_index;                 /* Band index */ 
    int process_first_band = 1;     /* Flag marking first band processed */
    double first_grid_pixsize = 0.0;  /* Pixel size of first band from grid */
    double scaling_factor;          /* Scaling factor to compare pixel sizes
                                       for bands with different resolutions */
    double scaled_pixsize;          /* Pixel size with scaling factor applied */

    /* Loop through the bands, comparing them to the first band */
    for (band = 0; band < num_bands; band++)
    {
        /* Get the band index */
        band_index = ias_sat_attr_convert_band_number_to_index(
                         band_numbers[band]);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("band_index not found for band_number[%d]", 
                band_numbers[band]);
            return ERROR;
        }

        /* Skip any bands that aren't available in the grid */
        if (!grid->bands_available[band_index])
            continue;

        /* Get the pixel size and grid pixel size for the first band in the
           band_numbers list */
        if (process_first_band)
        {
            first_grid_pixsize = grid->gridbands[band_index][0].pixsize;
            process_first_band = 0;
            continue;
        }

        /* Determine the scaling factor to use when comparing bands to the
           first band */
        scaling_factor = first_grid_pixsize 
            / grid->gridbands[band_index][0].pixsize;

        /* Check that the scaled grid pixsizes match the first band's pixsizes
           for all SCAs, within a tolerance */
        for (sca_index = 1;
             sca_index < grid->scas_per_band[band_index];
             sca_index++)
        {
            scaled_pixsize = grid->gridbands[band_index][sca_index].pixsize
                * scaling_factor;

            if ((scaled_pixsize + TOLERANCE < first_grid_pixsize) ||
                (scaled_pixsize - TOLERANCE > first_grid_pixsize))
            {
                IAS_LOG_ERROR("Scaled grid resolution %3.2f for band number "
                    "%d SCA index %d and grid resolution %3.2f for "
                    "band number %d SCA index 0 should match with tolerance "
                    "of %f meters", scaled_pixsize, band_numbers[band], 
                    sca_index, first_grid_pixsize, band_numbers[0], 
                    TOLERANCE);
                return ERROR;

            }
        }
    }

    return SUCCESS;
}
