#ifndef BANDCAL_H
#define BANDCAL_H

#include <limits.h>         /* For PATH_MAX */

#include "ias_grid_io.h"    /* Grid I/O library */
#include "ias_los_model.h"  /* IAS model library */
#include "save_to_database.h" /* Bandcal database save header */
#include "residuals_stats.h" /* Statistics structures */
#include "ias_satellite_attributes.h"

#define NCOMBO 55           /* Number of band combinations: 11 bands, 2 bands
                               per combination: 11!/(2! * (11-2)!) = 55 */

#define CHIP_NAME_SIZE 100  /* maximum size of image chip names */
#define CHIP_NAME_LENGTH (CHIP_NAME_SIZE - 1)

typedef struct POINT
{
    int point_id;   /* Point identification number */
    float reference_line; /* Line in reference image */
    float reference_samp; /* Sample in reference image */
    float reference_lat; /* Latitude (degrees) of point in reference image */
    float reference_lon; /* Longitude (degrees) of point in reference image */
    float reference_elev; /* Elevation (meters) of point in reference image */
    float search_line; /* Line in search image */
    float search_samp; /* Sample in search image */
    float dx;       /* Difference in x */
    float dy;       /* Difference in y */
    int validity_flag; /* Valid/invalid flag */
    float correlation_coefficient; /* Correlation peak strength */
    int reference_band_number; /* Band number for the reference image */
    int search_band_number; /* Band number for the search image */
    int reference_sca_number; /* Reference SCA number */
    int search_sca_number; /* Search SCA number - currently same as reference */
    char reference_img[CHIP_NAME_SIZE]; /* Image chip name for reference band */
    char search_img[CHIP_NAME_SIZE]; /* Image chip name for search band */
} POINT;

typedef struct BANDCOMBO
{
    int reference[NCOMBO][IAS_MAX_NSCAS]; /* Reference band number in 
                                             combination */
    int search[NCOMBO][IAS_MAX_NSCAS];    /* Search band number in 
                                             combination */
    int num_points[NCOMBO][IAS_MAX_NSCAS];/* Number of points in combination */
    IAS_SENSOR_ID sensor_id;              /* The sensor of the input data */
    int band_sca_not_present[IAS_MAX_NBANDS][IAS_MAX_NSCAS];
                                          /* Identifies if the band/sca
                                             combination is missing from the
                                             input data */
} BANDCOMBO;

/* function prototypes */

int get_parameters
(
    const char parm_file[PATH_MAX], /* I: Name of the parameter file */
    char dem_file[PATH_MAX],    /* O: Name of the DEM file */
    char l0r_file[PATH_MAX],    /* O: Name of the L0R file */
    char l1g_file[PATH_MAX],    /* O: Name of the L1G file */
    char grid_file[PATH_MAX],   /* O: Name of grid file */
    char model_file[PATH_MAX],  /* O: Name of model file */
    char b2b_char_file[PATH_MAX],/* O: Name of input band to band char file */
    char b2b_cal_file[PATH_MAX], /* O: Name of output band to band cal file */
    char *cpf_file_name,        /* O: Name of the CPF */
    int *ncoeff,                /* O: Number of coefficients to solve for */
    int *store_in_database_flag,/* O: Flag to write data to database */
    int *cpf_group_flag,        /* O: Flag to write CPF group to file */
    int *tie_point_list_flag,   /* O: Flag to write tie point list to report */
    int *minimum_points,        /* O: Minimum points to zero omitted bands */
    int *start_effective_date,  /* O: Starting effective date (YYYYMMDD) for
                                      CPF update file */
    int *end_effective_date,    /* O: Ending effective date (YYYYMMDD) for CPF
                                      update file */
    char *work_order_id,        /* O: Work order ID string */
    IAS_SATELLITE_ID *satellite_id /* O: Satellite ID */
);

int solve_focal_plane
(
    const char l1g[PATH_MAX], /* I: Name of the L1G for report data */
    const char l0r[PATH_MAX], /* I: L0R filename */
    const char b2b_cal_file[PATH_MAX], /* I: Name of band to band report file */
    const IAS_GRID_TYPE *grid, /* I: Grid with all bands */
    const IAS_LOS_MODEL *model, /* I: Model */
    const IAS_IMAGE *dem,   /* I: DEM data */
    int num_points,         /* I: Total number of points */
    int nsca,               /* I: Number of SCAs */
    const int *sca_indexes, /* I: List of SCA indexes to process */
    int minimum_points,     /* I: Minimum points to process */
    const BANDCOMBO *bandcombo, /* I: Band combinations */
    int ncoeff,             /* I: Number of coefficients to solve for */
    const POINT *points,    /* I: Points from B2B char file */
    double sol_along_track[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* O: Solution offsets in line direction */ 
    double sol_across_track[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* O: Solution offsets in sample direction */
    double new_legendre_along[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* O: New Legendre coefficients along track 
                                  direction */
    double new_legendre_across[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* O: New Legendre coefficients across track 
                                  direction */
    int ncombo,             /* I: Number of band combinations to process */
    int tie_point_list_flag, /* I: Flag for creating tie point report */
    int store_in_database_flag, /* I: Flag for saving results in database */
    struct BAND_SCA_DB_SAVE_INFO* band_sca_db_save, /* O: Band/SCA-level data
                                  to save */
    struct PRE_POST_FIT_DB_SAVE_INFO* pre_pst_fit_db_save, /* O: Pre and post
                                  fit data to save */
    const char *work_order_id, /* I: Work Order ID for report header */
    int *num_band_sca_db_records, /* O: Number of band/SCA records to save to 
                                  the database */
    int *num_pre_post_fit_db_records /* O: Number of pre/post fit records to 
                                  save to the database */
);

int get_los_errors
(
   int reference_band_index, /* I: Reference band index */
   int search_band_index,   /* I: Search band index */
   int sca_index,           /* I: SCA index */
   double reference_line,   /* I: Reference line location, output space */
   double reference_samp,   /* I: Reference samp location, output space */
   double search_line,      /* I: Search line location, output space */
   double search_samp,      /* I: Search samp location, output space */
   const IAS_IMAGE *dem,    /* I: DEM information */
   const IAS_GRID_BAND_TYPE *reference_grid_band_ptr, /* I: Reference grid 
                                  band */
   const IAS_GRID_BAND_TYPE *search_grid_band_ptr, /* I: Search grid band */
   const IAS_LOS_MODEL *model, /* I: Model information */
   int reference_nl,        /* I: Reference input number of lines */
   int reference_ns,        /* I: Reference input number of samples */
   int search_nl,           /* I: Search input number of lines */
   int search_ns,           /* I: Search input number of samples */
   int num_coeffs,          /* I: Number of coefficients to solve for */
   double *reference_input_line, /* O: Line location for reference, input 
                                  space */
   double *reference_input_samp, /* O: Sample location for reference, input 
                                  space */
   double *search_input_line, /* O: Line location for search, input space */
   double *search_input_samp, /* O: Sample location for search, input space */
   double *los_err_line,    /* O: LOS error in IFOV for line direction */
   double *los_err_samp     /* O: LOS error in IFOV for sample direction */
);

POINT *read_b2b_char_output
(
    const char b2b_char_file[PATH_MAX], /* I: Name of output B2B_char file */
    int *num_rec,           /* O: Number of records in B2B char output file */
    int *nsca,              /* O: Number of SCAs in B2B char output file */
    int *sca_number,        /* O: SCA numbers */
    BANDCOMBO *bandcombo    /* O: Number of points for each bandcombo */
);

int get_effective_ifov
(
    double sline,           /* I: Search input line number */
    double ssamp,           /* I: Search input sample number */
    int sband_index,        /* I: Search input band index */
    int sca,                /* I: SCA number */
    double elev,            /* I: Target elevation (at line/sample) */
    const IAS_LOS_MODEL *model, /* I: Model information */
    const IAS_GRID_BAND_TYPE *sgrid_ptr, /* I: Search grid band info */
    double *ifov            /* O: Effective IFOV */
);

int write_sca_parameters_cpf
(
    const char *cpf_filename, /* I: CPF name for writing the group */
    double x_coeff_corr[IAS_MAX_NBANDS][IAS_MAX_NSCAS][IAS_LOS_LEGENDRE_TERMS],
                              /* I: X Legendre coefficient corrections */
    double y_coeff_corr[IAS_MAX_NBANDS][IAS_MAX_NSCAS][IAS_LOS_LEGENDRE_TERMS],
                              /* I: Y Legendre coefficient corrections */
    const BANDCOMBO *bandcombo /* I: Band combinations */
);

int create_bandcal_report
(
    const char l1g[PATH_MAX],    /* I: Name of the L1G for report data */
    const char l0r[PATH_MAX],    /* I: Name of the L0R HDF file */
    const char b2b_cal_file[PATH_MAX], /* I: Name of band to band report file */
    int ncoeff,                  /* I: Number of coefficients to solve for */
    double sol_along_track[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* I: Solution offsets in line direction */
    double sol_across_track[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* I: Solution offsets in sample direction */
    double new_legendre_along[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* I: New Legendre coefficients along track
                                       direction */
    double new_legendre_across[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* I: New Legendre coefficients across track
                                       direction */
    const FIT_RESIDUAL_STATS_TYPE fit_stats[IAS_MAX_NSCAS], /* Fit Statistics */
    const RESIDUALS_PNTS *res,   /* I: Holds offsets between the search and 
                                       reference locations between 2 bands */
    int point_count,             /* I: Number of residual points */
    int tie_point_list_flag,     /* I: Flag for creating tie point report */
    const char *work_order_id,   /* I: Work Order ID for report header */
    const BANDCOMBO *bandcombo,  /* I: Band combinations */
    int nsca                     /* I: Number of SCAs */
);

#endif
