/*****************************************************************************
NAME: create_bandcal_report

PURPOSE:
Creates the bandcal report file.

RETURN VALUE: Type = int
    Value    Description
    -----    -----------
    SUCCESS  Successful completion
    ERROR    Operation failed
*****************************************************************************/

#include <limits.h>
#include <stdio.h>
#include <string.h>

#include "ias_const.h"
#include "ias_logging.h"         /* Logging library */
#include "ias_miscellaneous.h"   /* For report header routines */
#include "ias_satellite_attributes.h" /* Satellite attributes library */

#include "residuals_stats.h"
#include "bandcal.h"

int create_bandcal_report
(
    const char l1g[PATH_MAX],    /* I: Name of the L1G for report data */
    const char l0r[PATH_MAX],    /* I: Name of the L0R HDF file */
    const char b2bcal_file[PATH_MAX], /* I: Name of band to band report file */
    int ncoeff,                  /* I: Number of coefficients to solve for */
    double sol_along_track[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* I: Solution offsets in line direction */
    double sol_across_track[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* I: Solution offsets in sample direction */
    double new_legendre_along[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* I: New Legendre coefficients along track 
                                       direction */
    double new_legendre_across[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* I: New Legendre coefficients across track 
                                       direction */
    const FIT_RESIDUAL_STATS_TYPE fit_stats[IAS_MAX_NSCAS], /* I: Statistics */
    const RESIDUALS_PNTS *res,   /* I: Tie points structures */
    int point_count,             /* I: The count of points */
    int tie_point_list_flag,     /* I: Flag for creating tie point report */
    const char *work_order_id,   /* I: Work Order ID for report header */
    const BANDCOMBO *bandcombo,  /* I: Band combinations */
    int nsca                     /* I: Number of SCAs */
)
{
    FILE *ofile = NULL;    /* Output file for the B2B cal report */
    int status;            /* Return status for function */
    int res_point;         /* Loop variable */
    int coeff;             /* Loop variable */
    int band;              /* Loop variable */
    int sca_index;         /* SCA index */
    int band_index;        /* Band iEXIT_FAILUREndex */
    int num_bands;         /* Total number of bands in sensor */
    int band_numbers[IAS_MAX_NBANDS]; /* List of band numbers */
    IAS_REPORT_HEADER *header = NULL; /* Report header structure */
    char msg_string[512];  /* To setup messages */
    const char *sensor_id_string; /* The name of the sensor */

    /* Get band info */
    status = ias_sat_attr_get_sensor_band_numbers(bandcombo->sensor_id,
        IAS_NORMAL_BAND, 0, band_numbers, IAS_MAX_NBANDS, &num_bands);
    if (status != SUCCESS)
    {
        strncpy(msg_string, "Getting normal band numbers for the %s sensor",
                sizeof(msg_string));
        sensor_id_string = ias_sat_attr_get_sensor_name(bandcombo->sensor_id);

        IAS_LOG_ERROR(msg_string, sensor_id_string);

        return ERROR;
    }

    /* Open the band to band calibration output file */
    ofile = fopen(b2bcal_file, "wt");
    if (ofile == NULL)
    {
        IAS_LOG_ERROR("Opening Band to Band Calibration output file: %s", 
                b2bcal_file);
        return ERROR;
    }

    /* Output the report header and write it to the output file */
    header = ias_misc_initialize_gps_report_header(l1g, l0r, work_order_id,
            "BANDCAL");
    if (!header)
    {
        IAS_LOG_ERROR("Initializing the GPS report header for file %s", 
            b2bcal_file);
        fclose(ofile);
        return ERROR;
    }

    status = ias_misc_write_gps_report_header(ofile, header);
    ias_misc_free_report_header(header);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing GPS header to report file %s", b2bcal_file);
        fclose(ofile);
        return ERROR;
    }
    header = NULL;

    /* Write report */
    status = fprintf(ofile, "Number of Legendre coefficients %d\n"
        "Correction coefficients\n"
        "SCA                       Pre-Fit"
            "                                           Post-fit\n"
        "          Mean             RMSE            StDev"
            "            Mean             RMSE          StDev\n",
        ncoeff);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing to report file %s", b2bcal_file);
        fclose(ofile);
        return ERROR;
    }

    for (sca_index = 0; sca_index < nsca; sca_index++)
    {
        status = fprintf(ofile,
                "%d\t%e\t%e\t%e\t%e\t%e\t%e\tAcross track\n"
                "%d\t%e\t%e\t%e\t%e\t%e\t%e\tAlong track\n",
                sca_index + 1,
                fit_stats[sca_index].pre_fit_x.mean,
                fit_stats[sca_index].pre_fit_x.rmse, 
                fit_stats[sca_index].pre_fit_x.stdev,
                fit_stats[sca_index].post_fit_x.mean, 
                fit_stats[sca_index].post_fit_x.rmse,
                fit_stats[sca_index].post_fit_x.stdev,
                sca_index + 1,
                fit_stats[sca_index].pre_fit_y.mean,
                fit_stats[sca_index].pre_fit_y.rmse, 
                fit_stats[sca_index].pre_fit_y.stdev,
                fit_stats[sca_index].post_fit_y.mean, 
                fit_stats[sca_index].post_fit_y.rmse,
                fit_stats[sca_index].post_fit_y.stdev);
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing sca_index[%d] to report file %s",
                          sca_index, b2bcal_file);
            fclose(ofile);
            return ERROR;
        }
    }

    status = fprintf(ofile, "Coefficient updates\nSCA\tBand\tDir\t");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing to report file %s", b2bcal_file);
        fclose(ofile);
        return ERROR;
    }

    for (coeff = 0; coeff < ncoeff; coeff++)
    {
        status = fprintf(ofile, "L%d\t\t", coeff + 1);
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing coefficient[%d] to report file %s",
                          coeff + 1, b2bcal_file);
            fclose(ofile);
            return ERROR;
        }
    }

    status = fprintf(ofile, "\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing to report file %s", b2bcal_file);
        fclose(ofile);
        return ERROR;
    }

    for (sca_index = 0; sca_index < nsca; sca_index++)
    {
        for (band = 0; band < num_bands; band++)
        {
            /* Get the band index */
            band_index = ias_sat_attr_convert_band_number_to_index(
                band_numbers[band]);
            if (band_index == ERROR)
            {
                IAS_LOG_ERROR("Invalid band_index[%d] for band_number[%d]", 
                              band_index, band_numbers[band]);
                fclose(ofile);
                return ERROR;
            }

            /* Skip what we didn't process */
            if (bandcombo->band_sca_not_present[band_index][sca_index] == TRUE)
                continue;

            /* ------------------------------------------------------- */
            status = fprintf(ofile, "%d\t%d\tAlong\t", sca_index + 1, 
                band_numbers[band]);
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing sca_index[%d], band_index[%d] to "
                              "report file %s", sca_index, band_index,
                              b2bcal_file);
                fclose(ofile);
                return ERROR;
            }

            for (coeff = 0; coeff < ncoeff; coeff++)
            {
                status = fprintf(ofile, "%e\t", 
                    sol_along_track[band_index][sca_index][coeff]);
                if (status < 0)
                {
                    IAS_LOG_ERROR("Writing sca_index[%d], band_index[%d] to "
                                  "report file %s", sca_index, band_index,
                                  b2bcal_file);
                    fclose(ofile);
                    return ERROR;
                }
            }

            /* ------------------------------------------------------- */
            status = fprintf(ofile, "\n%d\t%d\tAcross\t", sca_index + 1, 
                band_numbers[band]);
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing sca_index[%d], band_index[%d] to "
                              "report file %s", sca_index, band_index,
                              b2bcal_file);
                fclose(ofile);
                return ERROR;
            }

            for (coeff = 0; coeff < ncoeff; coeff++)
            {
                status = fprintf(ofile, "%e\t", 
                    sol_across_track[band_index][sca_index][coeff]);
                if (status < 0)
                {
                    IAS_LOG_ERROR("Writing sca_index[%d], band_index[%d] to "
                                  "report file %s", sca_index, band_index,
                                  b2bcal_file);
                    fclose(ofile);
                    return ERROR;
                }
            }

            status = fprintf(ofile, "\n");
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing sca_index[%d], band_index[%d] to "
                              "report file %s", sca_index, band_index,
                              b2bcal_file);
                fclose(ofile);
                return ERROR;
            }
        }
    }

    /* Print new CPF coefficients in report file */
    status = fprintf(ofile, "New CPF coefficients\nSCA\tBand\tDir\t");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing to report file %s", b2bcal_file);
        fclose(ofile);
        return ERROR;
    }

    for (coeff = 0; coeff < ncoeff; coeff++)
    {
        status = fprintf(ofile, "L%d\t\t", coeff + 1);
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing coefficient[%d] to report file %s",
                          coeff + 1, b2bcal_file);
            fclose(ofile);
            return ERROR;
        }
    }

    status = fprintf(ofile, "\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing to report file %s", b2bcal_file);
        fclose(ofile);
        return ERROR;
    }

    for (sca_index = 0; sca_index < nsca; sca_index++)
    {
        for (band = 0; band < num_bands; band++)
        {
            /* Get the band index */
            band_index = ias_sat_attr_convert_band_number_to_index(
                band_numbers[band]);
            if (band_index == ERROR)
            {
                IAS_LOG_ERROR("Invalid band_index[%d] for band_number[%d]", 
                              band_index, band_numbers[band]);
                fclose(ofile);
                return ERROR;
            }

            /* Skip what we didn't process */
            if (bandcombo->band_sca_not_present[band_index][sca_index] == TRUE)
                continue;

            status = fprintf(ofile, "%d\t%d\tAlong\t", sca_index + 1, 
                band_numbers[band]);
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing sca_index[%d], band_index[%d] to "
                              "report file %s", sca_index, band_index,
                              b2bcal_file);
                fclose(ofile);
                return ERROR;
            }
            for (coeff = 0; coeff < ncoeff; coeff++)
            {
                status = fprintf(ofile, "%e\t", 
                    new_legendre_along[band_index][sca_index][coeff]);
                if (status < 0)
                {
                    IAS_LOG_ERROR("Writing sca_index[%d], band_index[%d] to "
                                  "report file %s", sca_index, band_index,
                                  b2bcal_file);
                    fclose(ofile);
                    return ERROR;
                }
            }
            status = fprintf(ofile, "\n%d\t%d\tAcross\t", sca_index + 1, 
                band_numbers[band]);
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing sca_index[%d], band_index[%d] to "
                              "report file %s", sca_index, band_index,
                              b2bcal_file);
                fclose(ofile);
                return ERROR;
            }
            for (coeff = 0; coeff < ncoeff; coeff++)
            {
                status = fprintf(ofile, "%e\t", 
                    new_legendre_across[band_index][sca_index][coeff]);
                if (status < 0)
                {
                    IAS_LOG_ERROR("Writing sca_index[%d], band_index[%d] to "
                                  "report file %s", sca_index, band_index,
                                  b2bcal_file);
                    fclose(ofile);
                    return ERROR;
                }
            }
            status = fprintf(ofile, "\n");
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing sca_index[%d], band_index[%d] to "
                              "report file %s", sca_index, band_index,
                              b2bcal_file);
                fclose(ofile);
                return ERROR;
            }
        }
    }

    /* Print tie point information */
    if (tie_point_list_flag == TRUE)
    {
        status = fprintf(ofile, "Point residuals\n\t\tOutput\t\t\t\tInput\t"
            "\t\t\t\t\t\t\tBands\t\tMeasured\t\t\tPre-delta\t\t\tPost-delta"
            "\nID\tSCA\tRline\t\tRsamp\t\tRline\t\tRsamp\t\tSline\t\tSsamp\t"
            "\tRband\tSband\tDY\t\tDX\t\tLine\t\tSamp\t\tLine\t\tSamp\n");
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing to report file %s", b2bcal_file);
            fclose(ofile);
            return ERROR;
        }

        for (res_point = 0; res_point < point_count; res_point++)
        {
           status = fprintf(ofile, "%d\t%d\t%f\t%f\t%f\t%f\t%f\t%f\t%d\t%d"
               "\t%f\t%f\t%e\t%e\t%e\t%e\n",
               res[res_point].point_id,
               res[res_point].sca_number,
               res[res_point].ref_output_line,
               res[res_point].ref_output_samp,
               res[res_point].ref_input_line,
               res[res_point].ref_input_samp,
               res[res_point].search_input_line,
               res[res_point].search_input_samp,
               res[res_point].ref_band,
               res[res_point].search_band,
               res[res_point].dy,
               res[res_point].dx,
               res[res_point].pre_los_err_line,
               res[res_point].pre_los_err_samp,
               res[res_point].post_los_err_line,
               res[res_point].post_los_err_samp);
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing tie point residual[%d] to "
                              "report file %s", res_point, b2bcal_file);
                fclose(ofile);
                return ERROR;
            }
        }
    }

    /* Close the band to band output report file */
    status = fclose(ofile);
    if (status != 0)
    {
        IAS_LOG_ERROR("Closing band to band output file: %s", b2bcal_file);
        return ERROR;
    }

    return SUCCESS;
}
