/*********************************************************************
NAME: get_effective_ifov

PURPOSE:
Calculate effective along track IFOV for a given line.  IFOV is based on
change in output space based on a single line change in input space.  Earth
acquisitions are scaled by the distance from the satellite to the target.

RETURN VALUE:
Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed
******************************************************************************/

#include <math.h>

#include "ias_types.h"      /* Acquisition type definitions */
#include "ias_const.h"      /* SUCCESS/ERROR definitions */
#include "ias_structures.h"
#include "ias_math.h"       /* Math library */
#include "ias_geo.h"        /* Geographic library */
#include "ias_grid.h"       /* Grid library */
#include "ias_spacecraft_model.h"
#include "ias_los_model.h"  /* Model library */
#include "ias_sensor_model.h" /* Sensor model library */
#include "ias_logging.h"    /* Logging library */
#include "ias_satellite_attributes.h" /* Satellite attributes library */

#include "bandcal.h"        /* Bandcal application */

int get_effective_ifov
(
    double search_line,     /* I: Search input line number */
    double search_samp,     /* I: Search input sample number */
    int search_band_index,  /* I: Search input band index */
    int sca_index,          /* I: SCA index */
    double elev,            /* I: Target elevation (at line/sample) */
    const IAS_LOS_MODEL *model, /* I: Model information */
    const IAS_GRID_BAND_TYPE *search_grid_ptr, /* I: Search grid band info */
    double *ifov            /* O: Effective IFOV */
)
{
    IAS_SENSOR_DETECTOR_TYPE dettype = IAS_NOMINAL_DETECTOR; /* Detector type */
    double lon;             /* Target longitude */
    double latd;            /* Target latitude */
    double time;            /* Time in imagery */
    IAS_VECTOR satpos;      /* Spacecraft position */
    IAS_VECTOR satvel;      /* Spacecraft velocity */
    double eph_time;        /* Delta time for ephemeris of image time */
    double alt;             /* Satellite altitude */
    double out_samp1, out_samp2; /* Output sample locations 1 input line 
                                    apart */
    double out_line1, out_line2; /* Output line locations 1 input line apart */
    double diff_samp;       /* Difference in output sample locations */
    double diff_line;       /* Difference in output line locations */
    double dist;            /* Total change in distance */
    IAS_VECTOR los;         /* Line of sight vector */
    IAS_VECTOR target;      /* Target vector */
    double flat;            /* Flattening of earth */
    int search_band_number; /* Band number corresponding to sband_index */

    /* Get the search band number */
    search_band_number = ias_sat_attr_convert_band_index_to_number(
            search_band_index);
    if (search_band_number == ERROR)
    {
        IAS_LOG_ERROR("Invalid band_number[%d] for search band_index[%d]", 
                search_band_number, search_band_index);
        return ERROR;
    }

    /* Map input space point to output space */
    if (ias_grid_ils2ols_at_elevation(search_grid_ptr, elev, search_line, 
                search_samp, &out_line1, &out_samp1) != SUCCESS)
    {
        IAS_LOG_ERROR("Mapping SCA number %d, band number %d, L1R line %f, "
                "L1R samp %f to output space", sca_index + 1,
                search_band_number, search_line, search_samp);
        return ERROR;
    }

    /* Map input space point 1 input space line over to output space */
    if (ias_grid_ils2ols_at_elevation(search_grid_ptr, elev, search_line + 1.0,
                search_samp, &out_line2, &out_samp2) != SUCCESS) 
    {
        IAS_LOG_ERROR("Mapping SCA number %d, band number %d, L1R line %f, "
                "L1R samp %f to output space", sca_index + 1,
                search_band_number, search_line + 1, search_samp);
        return ERROR;
    }

    /* Find change in output pixel */
    diff_line = out_line1 - out_line2;
    diff_samp = out_samp1 - out_samp2;

    /* Find total change in output, or distance */
    dist = sqrt(diff_line * diff_line + diff_samp * diff_samp);

    if (model->acquisition_type == IAS_EARTH)
    {
        /* Calculate target position */
        if (ias_los_model_input_line_samp_to_geodetic(search_line, search_samp,
                    search_band_index, sca_index, elev, model, dettype, NULL,
                    &latd, &lon) != SUCCESS)
        {
            IAS_LOG_ERROR("Finding target of SCA number %d, band number %d, "
                    "L1R line %f, L1R samp %f", sca_index, search_band_number, 
                    search_line, search_samp);
            return ERROR;
        }

        /* Find time for target location */
        if (ias_sensor_find_time(search_line, search_samp, search_band_index, 
                    sca_index, &model->sensor, dettype, &time) != SUCCESS)
        {
            IAS_LOG_ERROR("Finding time of SCA number %d, band number %d, L1R "
                    "line %f, L1R samp %f", sca_index, search_band_number, 
                    search_line, search_samp);
            return ERROR;
        }

        /* Find position and velocity for current location */
        if (ias_math_get_time_difference( 
                    model->sensor.bands[search_band_index].utc_epoch_time,
                    model->spacecraft.ephemeris.utc_epoch_time, &eph_time)
                != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating the time difference between image and "
                    "ephemeris epoch times");
            return ERROR;
        }
        eph_time += time;

        ias_sc_model_get_position_and_velocity_at_time(
                &model->spacecraft.ephemeris, model->acquisition_type,
                eph_time, &satpos, &satvel);

        /* Scale by distance to earth */
        flat = (model->earth.semi_major_axis - model->earth.semi_minor_axis) /
            model->earth.semi_major_axis;

        ias_geo_convert_geod2cart(latd, lon, elev, model->earth.semi_major_axis,
                flat, &target);

        los.x = target.x - satpos.x;
        los.y = target.y - satpos.y;
        los.z = target.z - satpos.z;
        alt = sqrt(los.x * los.x + los.y * los.y + los.z * los.z);
        if (alt == 0.0)
        {
            IAS_LOG_ERROR("Variable alt is 0 causing division by 0");
            return ERROR;
        }
        *ifov = (search_grid_ptr->pixsize * dist) / alt;
    }
    else /* Nothing to do for lunar image */
    {
        *ifov = search_grid_ptr->pixsize * dist
            * ias_math_get_arcsec_to_radian_conversion();
    }

    return SUCCESS;
}
