/****************************************************************************
NAME: get_los_errors 

PURPOSE:
Finds the adjustment needed between search and reference LOS vectors.  This
is needed for least squares. 

RETURN VALUE:
Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

*******************************************************************************/
#include <math.h>

#include "ias_types.h"      /* Acquisition type definitions */
#include "ias_const.h"      /* SUCCESS/ERROR definitions */
#include "ias_structures.h"
#include "ias_logging.h"    /* Logging library */
#include "ias_miscellaneous.h" /* Miscellaneous library */
#include "ias_math.h"
#include "ias_grid.h"
#include "ias_los_model.h"
#include "ias_sensor_model.h"

#include "bandcal.h"

int get_los_errors
(
    int reference_band_index, /* I: Reference band index */
    int search_band_index,   /* I: Search band index */
    int sca_index,           /* I: SCA index */
    double reference_line,   /* I: Reference line location, output space */
    double reference_samp,   /* I: Reference sample location, output space */
    double search_line,      /* I: Search line location, output space */
    double search_samp,      /* I: Search sample location, output space */
    const IAS_IMAGE *dem,    /* I: DEM information */
    const IAS_GRID_BAND_TYPE *reference_grid_band_ptr, /* I: Reference grid 
                                   band */
    const IAS_GRID_BAND_TYPE *search_grid_band_ptr, /* I: Search grid band */
    const IAS_LOS_MODEL *model, /* I: Model information */
    int reference_nl,        /* I: Reference input number of lines */
    int reference_ns,        /* I: Reference input number of samples */
    int search_nl,           /* I: Search input number of lines */
    int search_ns,           /* I: Search input number of samples */
    int num_coeffs,          /* I: Number of coefficients to solve for */
    double *reference_input_line, /* O: Line location for reference, input 
                                   space */
    double *reference_input_samp, /* O: Sample location for reference, input 
                                   space */
    double *search_input_line, /* O: Line location for search, input space */
    double *search_input_samp, /* O: Sample location for search, input space */
    double *los_err_line,    /* O: LOS error in IFOV for line direction */
    double *los_err_samp     /* O: LOS error in IFOV for sample direction */
)
{
    double elev;            /* Elevation at point */
    double reference_iline; /* Reference input line location */
    double reference_isamp; /* Reference input sample location */
    double riline_from_search_grid; /* Reference input line location using 
                                       search grid */
    double risamp_from_search_grid; /* Reference input sample location using 
                                       search grid */
    double search_iline;    /* Search input line location using search grid */
    double search_isamp;    /* Search input sample location using search grid */
    double search_x;        /* Search x input space coordinate component */
    double search_y;        /* Search y input space coordinate component */
    double nom_search_x;    /* Search x input space coordinate component */
    double nom_search_y;    /* Search y input space coordinate component */
    double norm_det;        /* Normalized detector within an SCA */
    double reference_pix_size; /* I: Reference pixel size */
    double search_pix_size; /* I: Search pixel size */
    double ifov;            /* Effective instantaneous field of view */
    double scale;           /* Ratio of search and reference band resolutions */
    double sqrt_of_sum_of_squares; /* Square root of the sum of squares for the
                                      x, y, and z components */
    double nom_search_z;    /* Nominal search z */
    double search_z;        /* Search z input space coordinate component */
    IAS_SENSOR_SCA_MODEL *model_search_sca; /* Pointer to search band and 
                                               current SCA in the model */
    int status;             /* Return status */

    /* Get elevation for point.  This is needed to find mapping coefficients */
    if (model->acquisition_type == IAS_EARTH)
    {
        reference_pix_size = reference_grid_band_ptr->pixsize; 
        if (ias_misc_read_elevation_at_line_sample(dem, reference_line, 
            reference_samp, reference_pix_size, reference_pix_size, &elev) 
            != SUCCESS)
        {
            IAS_LOG_ERROR("Getting elevation for reference line %f, sample %f",
                reference_line, reference_samp);
            return ERROR;
        }
    }
    else
    {
        elev = 0.0;
    }

    /* Find reference input line/sample for output line/sample using reference
       grid */
    status = ias_grid_3d_ols2ils(reference_grid_band_ptr, reference_line, 
        reference_samp, &reference_iline, &reference_isamp, elev);

    if ((status != SUCCESS)
        || (reference_iline < 1) || (reference_iline >= (reference_nl - 2))
        || (reference_isamp < 1) || (reference_isamp >= (reference_ns - 2)))
    {
        IAS_LOG_ERROR("Reference point could not be found in input space for "
            "output space line %f, sample %f using reference grid", 
            reference_line, reference_samp);
        return ERROR;
    }

    /* Save reference input line and sample */
    *reference_input_line = reference_iline;
    *reference_input_samp = reference_isamp;

    /* Find reference input line/sample for output line/sample using search
       grid */
    scale = model->sensor.bands[reference_band_index].sampling_char
        .lines_per_frame
        / model->sensor.bands[search_band_index].sampling_char.lines_per_frame;
    reference_line = reference_line / scale;
    reference_samp = reference_samp / scale;

    status = ias_grid_3d_ols2ils(search_grid_band_ptr, reference_line, 
        reference_samp, &riline_from_search_grid, &risamp_from_search_grid, 
        elev);

    if ((status != SUCCESS)
        || (riline_from_search_grid < 1) || (riline_from_search_grid 
        >= (search_nl - 2)) || (risamp_from_search_grid < 1) 
        || (risamp_from_search_grid >= (search_ns - 2)))
    {
        IAS_LOG_ERROR("Reference point could not be found in input space for "
            "output space line %f, sample %f using search grid", 
            reference_line, reference_samp);
        return ERROR;
    }

    /* Set up a pointer to current search band and SCA in the model */
    model_search_sca = &model->sensor.bands[search_band_index].scas[sca_index];

    /* Compute the normalized detector number */
    norm_det = 2.0 * risamp_from_search_grid 
        / ((double)model_search_sca->detectors - 1.0) - 1.0;

    /* Evaluate the Legendre polynomials from the model */
    nom_search_x = ias_math_eval_legendre(norm_det,
                                 model_search_sca->sca_coef_x, num_coeffs);
    nom_search_y = ias_math_eval_legendre(norm_det,
                                 model_search_sca->sca_coef_y, num_coeffs);

    /* Get elevation for point.  This is needed to find mapping coefficients */
    if (model->acquisition_type == IAS_EARTH)
    {
        search_pix_size = search_grid_band_ptr->pixsize;
        if (ias_misc_read_elevation_at_line_sample(dem, search_line, 
            search_samp, search_pix_size, search_pix_size, &elev) != SUCCESS)
        {
            IAS_LOG_ERROR("Getting elevation for search line %f, sample %f",
                search_line, search_samp);
            return ERROR;
        }
    }
    else
    {
        elev = 0.0;
    }

    status = ias_grid_3d_ols2ils(search_grid_band_ptr, search_line, 
        search_samp, &search_iline, &search_isamp, elev);

    if ((status != SUCCESS) ||
        (search_iline < 1) ||
        (search_iline >= (search_nl - 2)) ||
        (search_isamp < 1) ||
        (search_isamp >= (search_ns - 2)))
    {
        IAS_LOG_ERROR("Search point could not be found in input space for "
            "output space search line %f, sample %f using search grid", 
            search_line, search_samp);
        return ERROR;
    }

    /* Save values for external reporting */
    *search_input_line = search_iline;
    *search_input_samp = search_isamp;

    /* Compute the normalized detector number */
    norm_det = 2.0 * search_isamp / ((double)model_search_sca->detectors - 1.0)
        - 1.0;

    /* Evaluate the Legendre polynomials from the model */
    search_x = ias_math_eval_legendre(norm_det,
                             model_search_sca->sca_coef_x, num_coeffs);
    search_y = ias_math_eval_legendre(norm_det,
                             model_search_sca->sca_coef_y, num_coeffs);

    /* Calculate the errors */
    search_z = 1.0;
    sqrt_of_sum_of_squares = sqrt(search_x * search_x +
                                  search_y * search_y + 
                                  search_z * search_z);
    if (sqrt_of_sum_of_squares == 0)
    {
        IAS_LOG_ERROR("The square root of the sum of squares for search x, "
            "y, and z components for band index %d SCA index %d is 0, leading "
            "to divide by 0", search_band_index, sca_index);
        return ERROR;
    }
    search_x = search_x / sqrt_of_sum_of_squares;
    search_y = search_y / sqrt_of_sum_of_squares;
    search_z = search_z / sqrt_of_sum_of_squares;

    nom_search_z = 1.0;
    sqrt_of_sum_of_squares = sqrt(nom_search_x * nom_search_x + nom_search_y 
        * nom_search_y + nom_search_z * nom_search_z);
    if (sqrt_of_sum_of_squares == 0)
    {
        IAS_LOG_ERROR("The square root of the sum of squares for search x, "
            "y, and z components for band index %d SCA index %d is 0, leading "
            "to divide by 0", search_band_index, sca_index);
        return ERROR;
    }
    nom_search_x = nom_search_x / sqrt_of_sum_of_squares;
    nom_search_y = nom_search_y / sqrt_of_sum_of_squares;
    nom_search_z = nom_search_z / sqrt_of_sum_of_squares;

    /* Calculate the effective IFOV based on target and satellite info */
    if (get_effective_ifov(search_iline, search_isamp, search_band_index, 
        sca_index, elev, model, search_grid_band_ptr, &ifov) != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating IFOV for search line %f, sample %f, band "
            "index %d, SCA index %d", search_iline, search_isamp, 
            search_band_index, sca_index);
        return ERROR;
    }

    *los_err_line = search_x / search_z - nom_search_x / nom_search_z 
        + (search_iline - riline_from_search_grid) * ifov; 
    *los_err_samp = search_y / search_z - nom_search_y / nom_search_z;

    return SUCCESS;
}
