/*****************************************************************************
NAME: get_parameters

PURPOSE:
Get the Bandcal parameters from the input parameter file and OMF file. 

RETURN VALUE:
Type = int
Value        Description
-----        -----------
SUCCESS      Successful completion 
ERROR        Operation failed 
EXIT_SUCCESS If user asked for help
******************************************************************************/

#include <limits.h>
#include <stdlib.h>
#include <string.h>

#include "ias_const.h"
#include "ias_logging.h"
#include "ias_parm.h"            /* Parameter library */
#include "ias_parm_standard_parameters.h" /* Parameter table standard fields */
#include "ias_satellite_attributes.h" /* Satellite attributes library */

#include "bandcal.h"

#define NO_MINIMUM_POINTS_PARAMETER -1 /* Minimum Points parameter wasn't set */

int get_parameters
(
    const char parm_file[PATH_MAX], /* I: Name of the parameter file */
    char dem_file[PATH_MAX],     /* O: Name of the DEM file */
    char l0r_file[PATH_MAX],     /* O: Name of the L0R file */
    char l1g_file[PATH_MAX],     /* O: Name of the L1G file */
    char grid_file[PATH_MAX],    /* O: Name of grid file */
    char model_file[PATH_MAX],   /* O: Name of model file */
    char b2bchar_file[PATH_MAX], /* O: Name of input band to band char file */
    char b2bcal_file[PATH_MAX],  /* O: Name of output band to band cal file */
    char *cpf_file_name,         /* O: CPF file name */
    int *num_coeffs,             /* O: Number of coefficients to solve for */
    int *store_in_database_flag, /* O: Flag to write data to database */
    int *cpf_group_flag,         /* O: Flag to write CPF group to file */
    int *tie_point_list_flag,    /* O: Flag to write tie point list to report */
    int *minimum_points,         /* O: Minimum points to zero omitted bands */
    int *start_effective_date,   /* O: Starting effective date (YYYYMMDD) for 
                                       CPF update file */
    int *end_effective_date,     /* O: Ending effective date (YYYYMMDD) for CPF 
                                       update file */
    char *work_order_id,         /* O: Work order ID string */
    IAS_SATELLITE_ID *satellite_id /* O: Satellite ID */
)
{
    int count;                   /* Number of elements */
    char omf_file_name[PATH_MAX]; /* Name of OMF */
    int satellite_number;        /* Satellite being processed */
    int processing_pass;         /* Processing flow pass */
    int status;                  /* Status returned from functions */

    /* Various file name storage - local storage is used because sizeof doesn't
       work on function parameter arrays */
    char local_dem_file[PATH_MAX] = "";
    char local_l0r_file[PATH_MAX] = "";
    char local_l1g_file[PATH_MAX] = "";
    char local_grid_file[PATH_MAX] = "";
    char local_model_file[PATH_MAX] = "";
    char local_b2bchar_file[PATH_MAX] = "";
    char local_b2bcal_file[PATH_MAX] = "";
    char local_cpf_file_name[PATH_MAX] = "";
    char local_wo_id[IAS_WORKORDER_ID_SIZE] = "";

    /* Start parameter table definition */
    IAS_PARM_DECLARE_TABLE(parms, 10);

    /* Get the work order ID */
    IAS_PARM_WORK_ORDER_ID(parms, local_wo_id, sizeof(local_wo_id), 1); 

    /* Get the name of band to band calibration's report file */
    const char *default_b2bcal_file[] = {"bandcal.rpt"};
    IAS_PARM_ADD_STRING(parms, B2B_CAL_FILENAME, 
            "Report made by Band Accuracy Assessment Calibration",
            IAS_PARM_OPTIONAL, 0, NULL, 1, default_b2bcal_file,
            local_b2bcal_file, sizeof(local_b2bcal_file), 1);

    /* Get the number of coefficients to solve for */
    int default_num_coeffs[] = {3};
    IAS_PARM_ADD_INT(parms, NUM_COEFF, "Number of Legendre coefficients",
            IAS_PARM_REQUIRED, IAS_PARM_NOT_ARRAY, 1, 1, 4, 1,
            default_num_coeffs, num_coeffs, sizeof(*num_coeffs), 1);

    /* Get minimum points used to remove omitted bands from calculations */
    *minimum_points = NO_MINIMUM_POINTS_PARAMETER;
    IAS_PARM_ADD_INT(parms, MINIMUM_NUMBER_POINTS, 
            "Minimum points threshold used to omit bands from calculations",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 0, 0, 0, 0, NULL,
            minimum_points, sizeof(*minimum_points), 0);

    /* Get the flag to write (or not write) data to the database */
    IAS_PARM_STORE_CHARACTERIZATION_IN_DB(parms, IAS_PARM_OPTIONAL,
            store_in_database_flag, sizeof(*store_in_database_flag), 0);

    /* Get the flag to write (or not write) the CPF group */
    int default_cpf_group_flag[] = {FALSE};
    IAS_PARM_ADD_INT(parms, CPF_GROUP_FLAG, "ASCII CPF Group output switch", 
            IAS_PARM_REQUIRED, IAS_PARM_NOT_ARRAY, 1, 0, 1, 1,
            default_cpf_group_flag, cpf_group_flag, sizeof(*cpf_group_flag), 1);

    /* Get the flag to write (or not write) the report's tie point list */
    int default_tie_point_list_flag[] = {FALSE};
    IAS_PARM_ADD_INT(parms, TIE_POINT_LISTING_FLAG, 
            "Tie point list output switch", IAS_PARM_REQUIRED,
            IAS_PARM_NOT_ARRAY, 1, 0, 1, 1, default_tie_point_list_flag,
            tie_point_list_flag, sizeof(*tie_point_list_flag), 1);

    /* Get the effective start date */
    IAS_PARM_ADD_INT(parms, START_EFFECTIVE_DATE,
            "Effective start date for CPF group output file, expected format "
            "YYYYMMDD", IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 0, 0, 0, 0, NULL,
            start_effective_date, sizeof(*start_effective_date), 0);

    /* Get the effective end date */
    IAS_PARM_ADD_INT(parms, END_EFFECTIVE_DATE,
            "Effective end date for CPF group output file, expected format "
            "YYYYMMDD", IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 0, 0, 0, 0, NULL,
            end_effective_date, sizeof(*end_effective_date), 0);

    /* Get processing pass */
    IAS_PARM_PROCESSING_PASS(parms, &processing_pass, sizeof(processing_pass), 
            0);

    /* Calculate the number of items in the parameter file */
    count = IAS_PARM_GET_TABLE_SIZE(parms);

    /* Call the ias_parm_provide_help routine to provide help for the
       parameters if requested; also verifies the parameter file exists */
    status = ias_parm_provide_help(parm_file, parms, count, 
            IAS_INPUT_PARAMETERS);
    if (status == 1)
    {
        /* If the status of the provide help is 1 the user asked for help.
           Since the parameter file name contains the help request, copy it
           to the OMF file name so the next call will ask for the same help. */
        strcpy(omf_file_name, parm_file);
    }
    else if (status == ERROR) /* Help failed */
    {
        IAS_LOG_ERROR("Providing help using application parameter file %s",
                parm_file);
        return ERROR;
    }
    else
    {
        /* Read the parameter file. */
        if (ias_parm_read(parm_file, parms, count) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading parameters from: %s", parm_file);
            return ERROR;
        }

        /* Transfer the various local arrays to the function parameters. */
        strcpy(b2bcal_file, local_b2bcal_file);
        strcpy(work_order_id, local_wo_id);

        if (*minimum_points == NO_MINIMUM_POINTS_PARAMETER)
        {
            IAS_LOG_WARNING("Minimum number of points value not retrieved "
                    "from parameter file %s; will derive it from the Band "
                    "Characterization report", parm_file);
        }
        else
        {
            if (*minimum_points < 0)
            {
                IAS_LOG_ERROR("Invalid minimum number of points %d retrieved "
                        "from parameter file %s", *minimum_points, parm_file);
                return ERROR;
            }
        }

        /* Build the OMF file name */
        sprintf(omf_file_name, "%s.omf", work_order_id);
    }

    /* Now read and check the OMF */

    /* Start parameter table definition */
    IAS_PARM_DECLARE_TABLE(omf_parms, 8);

    /* Get the name of the DEM file */
    IAS_PARM_DEM_FILENAME_FROM_PASS(omf_parms, IAS_PARM_OPTIONAL, 
            local_dem_file, sizeof(local_dem_file), 1, processing_pass); 

    /* Get the name of the L0R file */
    IAS_PARM_L0R_FILENAME(omf_parms, IAS_PARM_OPTIONAL, local_l0r_file,
            sizeof(local_l0r_file), 1);

    /* Get the name of the L1G file */
    IAS_PARM_L1G_FILENAME_FROM_PASS(omf_parms, IAS_PARM_OPTIONAL,
            local_l1g_file, sizeof(local_l1g_file), 1, processing_pass);

    /* Get the name of the grid file */ 
    IAS_PARM_GRID_FILENAME_FROM_PASS(omf_parms, IAS_PARM_REQUIRED, 
            local_grid_file, sizeof(local_grid_file), 1, processing_pass); 

    /* Get the name of the model file */ 
    IAS_PARM_MODEL_FILENAME_FROM_PASS(omf_parms, IAS_PARM_REQUIRED, 
            local_model_file, sizeof(local_model_file), 1, processing_pass);

    /* Get the CPF filename */
    IAS_PARM_CPF(omf_parms, IAS_PARM_REQUIRED, local_cpf_file_name,
            sizeof(local_cpf_file_name), 1);

    /* Get the satellite number */
    IAS_PARM_SATELLITE(omf_parms, &satellite_number,
            sizeof(satellite_number), 1);

    /* Get the name of band to band characterization's residual report file */
    IAS_PARM_B2B_CHAR_RES_FILENAME(omf_parms, local_b2bchar_file,
            sizeof(local_b2bchar_file), 1);

    /* See if the user wanted help or a template with the OMF file */
    status = ias_parm_provide_help(omf_file_name, omf_parms,
            IAS_PARM_GET_TABLE_SIZE(omf_parms), IAS_OMF_PARAMETERS);
    if (status)
    {
        if (status == 1) /* Help was successful */
        {
            exit(EXIT_SUCCESS);
        }
        else if (status == ERROR) /* Help failed */
        {
            IAS_LOG_ERROR("Providing help using OMF %s", omf_file_name);
            return ERROR;
        }
    }

    /* Read the OMF file */
    count = IAS_PARM_GET_TABLE_SIZE(omf_parms);
    if (ias_parm_read(omf_file_name, omf_parms, count) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from OMF file: %s", omf_file_name);
        return ERROR;
    }

    /* Transfer the various local arrays to the function parameters */
    strcpy(dem_file, local_dem_file);
    strcpy(l0r_file, local_l0r_file);
    strcpy(l1g_file, local_l1g_file);
    strcpy(model_file, local_model_file);
    strcpy(grid_file, local_grid_file);
    strcpy(cpf_file_name, local_cpf_file_name);
    strcpy(b2bchar_file, local_b2bchar_file);

    *satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(
            satellite_number);
    if (*satellite_id == ERROR)
    {
        IAS_LOG_ERROR("Unable to determine satellite ID from satellite "
                "number %d", satellite_number);
        return ERROR;
    }

    return SUCCESS;
}

