/*******************************************************************************
NAME:   read_b2b_char_output

PURPOSE:
        Read the band to band characterization output file.

RETURN VALUE:
Type=POINT *
Value           Description
-----           -----------
not NULL        pointer to data that was read correctly
NULL            error reading band to band characterization output file

NOTES:
The POINT *point must be freed outside of the function

*******************************************************************************/

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>         /* For strcmp */

#include "ias_const.h"      /* TRUE/FALSE/ERROR definitions */
#include "ias_logging.h"    /* Logging library */
#include "ias_satellite_attributes.h" /* Satellite attributes library */

#include "bandcal.h"

#define OUTLIER 0
#define INVALID_NUM_RECORDS -1 /* A valid number of records hasn't been read */
#define MAX_LINE_SIZE 250   /* Limit the possible size of a file line */

POINT *read_b2b_char_output
(
    const char b2b_char_file[PATH_MAX], /* I: Name of output B2B_char file */
    int *number_of_records, /* O: Number of records in B2B char output file */
    int *number_of_scas,    /* O: Number of SCAs in B2B char output file */
    int *sca_indexes,       /* O: List of SCA indexes */
    BANDCOMBO *bandcombo    /* O: Number of points for each bandcombo */
)
{
    POINT *point;           /* Ptr to array of points from the B2B char file */
    char tmp[MAX_LINE_SIZE * 2]; /* Temporary variable for reading strings */
    char number_of_records_str[MAX_LINE_SIZE] = ""; /* Match text after "number
                               of records" text in report file */
    int point_index;        /* Point index */
    int combo_count;        /* Combination number in bandcombo variable */
    int combo_index;        /* Band combination index */
    int number_of_points;   /* Number of points in a particular band combo */
    int ref_band_number;    /* Reference band number */
    int search_band_number; /* Search band number */
    FILE *file = NULL;      /* Filepointer for the B2B char output file */
    int status;             /* Status of return from function */
    int point_count[NCOMBO][IAS_MAX_NSCAS];
                            /* Number of points for each band combination */ 
    int sca_index;          /* SCA index */
    int current_sca_number; /* Currently read SCA number */
    int first_search_band;  /* First search band after start of new combo */
    int first_ref_band;     /* First refer band after start of new combo */
    int ref_band_index = 0; /* Reference band index */
    int search_band_index = 0; /* Search band index */
    int sca_flag;           /* Flag for checking the SCA number */
    int oli_combo_count;
    int tirs_combo_count;

    /* Initialize the point counts used for validation */
    for (combo_index = 0; combo_index < NCOMBO; combo_index++)
    {
        for (sca_index = 0; sca_index < IAS_MAX_NSCAS; sca_index++)
        {
            point_count[combo_index][sca_index] = 0;
        }
    }

    for (ref_band_index = 0; ref_band_index < IAS_MAX_NBANDS; ref_band_index++)
    {
        for (sca_index = 0; sca_index < IAS_MAX_NSCAS; sca_index++)
        {
            bandcombo->band_sca_not_present[ref_band_index][sca_index] = TRUE;
        }
    }

    *number_of_scas = 0;
    *number_of_records = INVALID_NUM_RECORDS; 

    /* Open the band to band characterization output file */
    file = fopen(b2b_char_file, "r");
    if (file == NULL)
    {
        IAS_LOG_ERROR("Opening band to band characterization file: %s",
                b2b_char_file);
        return NULL;
    }

    /* Read the number of records in the file */
    while (*number_of_records == INVALID_NUM_RECORDS)
    {
        if (fgets(tmp, MAX_LINE_SIZE, file) == NULL)
        {
            IAS_LOG_ERROR("Reading header of file: '%s'", b2b_char_file);
            fclose(file);
            return NULL;
        }

        status = sscanf(tmp, "%d %17c", number_of_records, 
            number_of_records_str);

        /* If there weren't 2 matches or the text didn't match, but the number
           was updated */
        if (((status != 2) ||
             (strcmp("Number of records", number_of_records_str) != 0)) &&
            (*number_of_records != INVALID_NUM_RECORDS)) 
        {
            /* Disregard numbers read in a non-matching line */
            *number_of_records = INVALID_NUM_RECORDS;
        }
    }

    /* Allocate the memory for the POINT structure.  This may be bigger than
       the actual number of non outlier points. */
    point = malloc(sizeof(POINT) * (*number_of_records));
    if (point == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for %d records for POINT structure",
            *number_of_records);
        fclose(file);
        return NULL;
    }

    /* Initialize the number of points to zero */
    for (combo_index = 0; combo_index < NCOMBO; combo_index++)
    {
        for (sca_index = 0; sca_index < IAS_MAX_NSCAS; sca_index++)
        {
            bandcombo->num_points[combo_index][sca_index] = 0.0;
        }
    }

    /* Read the band to band characterization shifts */
    /* Read the header */
    if (fgets(tmp, MAX_LINE_SIZE, file) == NULL)
    {
        IAS_LOG_ERROR("Reading header of file: '%s'", b2b_char_file);
        fclose(file);
        free(point);
        return NULL;
    }

    /* Initialize a bunch of variables */
    combo_count = 0;
    point_index = 0;
    sca_index = 0;
    sca_flag = 0;
    oli_combo_count = 0;
    tirs_combo_count = 0;
    first_search_band = 0;
    first_ref_band = 0;

    /* Read each record of the data */
    while (point_index < *number_of_records)
    {
        IAS_SENSOR_ID ref_sensor_id;

        if (sca_index * combo_count > NCOMBO * IAS_MAX_NSCAS)
        {
            IAS_LOG_WARNING("Reached maximum number of combinations for "
                   "maximum number of SCAs at point index %d", point_index);
            break;  
        }

        /* Read a line of data if it is the header for a band combo, pull out 
           the info, else read the next line for the next record of data */
        if (fgets(tmp, MAX_LINE_SIZE * 2, file) == NULL)
        {
            IAS_LOG_ERROR("Reading record from file %s at point index %d", 
                b2b_char_file, point_index); 
            fclose(file);
            free(point);
            return NULL;
        }

        /* "Number" is the beginning of the line that states we're at a new 
           band combo */
        while (strncmp(tmp, "Number", 6) == 0) /* if at a new band combo */
        {
            /* Pull out the reference and search band numbers and the number 
               of points for that band combo */
            status = sscanf(tmp, "Number of points for the band combination "
                    "reference =%d search =%d: %d points SCA %d", 
                    &ref_band_number, &search_band_number, &number_of_points, 
                    &current_sca_number);
            if (status != 4)
            {
                IAS_LOG_ERROR("Reading B2B characterization file: %s at point "
                    "index %d\n", b2b_char_file, point_index); 
                fclose(file);
                free(point);
                return NULL;
            }

            /* Validate SCA number */
            if (current_sca_number < 1 || current_sca_number > IAS_MAX_NSCAS)
            {
                IAS_LOG_ERROR("Invalid SCA number %d for band combination "
                    "ref=%d search=%d at point index %d", current_sca_number, 
                     ref_band_number, search_band_number, point_index); 
                fclose(file);
                free(point);
                return NULL;
            }

            sca_index = current_sca_number - 1;

            if (point_index == 0)
            {
                /* Initialize to the starting control values. */
                first_search_band = search_band_number;
                first_ref_band = ref_band_number;
                sca_flag = 1;
            }
            else
            {
                if (first_search_band == search_band_number &&
                    first_ref_band == ref_band_number)
                {
                    /* Due to the organization of the input file... combo and
                       number of scas are set/reset here.

                       Ref Band 8 - Srch Band 1 - SCA 1
                       Ref Band 8 - Srch Band 2 - SCA 1
                       Ref Band 8 - Srch Band 3 - SCA 1
                         ...
                       Ref Band 8 - Srch Band 1 - SCA 2
                       Ref Band 8 - Srch Band 2 - SCA 2
                       Ref Band 8 - Srch Band 3 - SCA 2
                         ...
                       Ref Band 8 - Srch Band 1 - SCA 3
                       Ref Band 8 - Srch Band 2 - SCA 3
                       Ref Band 8 - Srch Band 3 - SCA 3
                    */
                    combo_count = 0;
                    (*number_of_scas)++;
                    sca_flag = 1;
                }
                else
                {
                    combo_count++;
                }
            }

            ref_band_index = ias_sat_attr_convert_band_number_to_index(
                ref_band_number);
            if (ref_band_index == ERROR)
            {
                IAS_LOG_ERROR(
                    "Invalid band_index[%d] for reference band_number[%d]", 
                    ref_band_index, ref_band_number);
                fclose(file);
                free(point);
                return NULL;
            }

             /* Figure out if the reference band is OLI or TIRS and increment
               the correct counter */
            ref_sensor_id =
                ias_sat_attr_get_sensor_id_from_band_number(ref_band_number);
            if ((ref_sensor_id == IAS_INVALID_SENSOR_ID)
                || (ref_sensor_id > IAS_TIRS))
            {
                IAS_LOG_ERROR("Determining sensor ID for reference "
                    "band_number[%d]", ref_band_number);
                fclose(file);
                free(point);
                return NULL;
            }
            if (ref_sensor_id == IAS_TIRS)
                tirs_combo_count++;
            else
                oli_combo_count++;

           search_band_index = ias_sat_attr_convert_band_number_to_index(
                search_band_number);
            if (search_band_index == ERROR)
            {
                IAS_LOG_ERROR(
                    "Invalid band_index[%d] for search band_number[%d]", 
                    search_band_index, search_band_number);
                fclose(file);
                free(point);
                return NULL;
            }

            /* Define combination */
            bandcombo->reference[combo_count][sca_index] = ref_band_number;
            bandcombo->search[combo_count][sca_index] = search_band_number;
            bandcombo->num_points[combo_count][sca_index] = number_of_points;
            bandcombo->band_sca_not_present[ref_band_index][sca_index] = FALSE;
            bandcombo->band_sca_not_present[search_band_index][sca_index] =
                FALSE;

            /* Get the first line of the next record, since the last one isn't
               a point entry for the upcoming section of code */
            if (fgets(tmp, MAX_LINE_SIZE * 2, file) == NULL)
            {
                IAS_LOG_ERROR("Reading record from file: %s at point index %d",
                    b2b_char_file, point_index); 
                fclose(file);
                free(point);
                return NULL;
            }
        }/* End while strncmp */

        if (number_of_points > 0) /* If we have any points to work with */
        {
            char reference_img[PATH_MAX];
            char search_img[PATH_MAX];

            /* Fill the point structure with this point */
            status = sscanf(tmp, 
                "%d %f %f %f %f %f %f %f %f %f %d %f %d %d %d %d %s %s", 
                &point[point_index].point_id, 
                &point[point_index].reference_line, 
                &point[point_index].reference_samp, 
                &point[point_index].reference_lat, 
                &point[point_index].reference_lon, 
                &point[point_index].reference_elev, 
                &point[point_index].search_line, 
                &point[point_index].search_samp, 
                &point[point_index].dy,
                &point[point_index].dx, 
                &point[point_index].validity_flag, 
                &point[point_index].correlation_coefficient, 
                &point[point_index].reference_band_number, 
                &point[point_index].search_band_number, 
                &point[point_index].reference_sca_number, 
                &point[point_index].search_sca_number, 
                reference_img, search_img);
            if (status != 18)
            {
                IAS_LOG_ERROR("Could not read line of data at point index %d; "
                    "sscanf status %d", point_index, status); 
                fclose(file);
                free(point);
                return NULL;
            }

            /* Make sure the reference and search image names aren't longer
               than the structure supports */
            if ((strlen(reference_img) > CHIP_NAME_LENGTH)
                || (strlen(search_img) > CHIP_NAME_LENGTH))
            {
                IAS_LOG_ERROR("The reference or search image is longer than "
                    "the supported length");
                fclose(file);
                free(point);
                return NULL;
            }
            /* Now that we know the image names fit, copy them to the 
               point record */
            strcpy(point[point_index].reference_img, reference_img);
            strcpy(point[point_index].search_img, search_img);

            if (sca_flag)
            {
                if (point[point_index].search_sca_number 
                    != point[point_index].reference_sca_number)
                {
                    IAS_LOG_ERROR("SCA numbers %d, %d (search, reference) at "
                        "point index %d in file %s should match", 
                        point[point_index].search_sca_number,
                        point[point_index].reference_sca_number, point_index,
                        b2b_char_file);
                    fclose(file);
                    free(point);
                    return NULL;
                }

                if (current_sca_number != point[point_index].search_sca_number)
                {
                    IAS_LOG_ERROR("SCA numbers %d, %d (header SCA number, "
                        "search) at point index %d in file %s should match", 
                        current_sca_number,
                        point[point_index].search_sca_number, point_index,
                        b2b_char_file);
                    fclose(file);
                    free(point);
                    return NULL;
                }

                sca_indexes[*number_of_scas] 
                    = point[point_index].reference_sca_number - 1;
                sca_flag = 0;
            }

            /* If the point is an outlier, then it shouldn't be in the points
               we use (we won't increment and it'll be written over) */
            if (point[point_index].validity_flag != OUTLIER)
            {
                point_index++;
                point_count[combo_count][sca_index]++;
            }
            else
            {
                /* If the point was an outlier we won't output as many as
                   we expected, so correct the number of output points */
                bandcombo->num_points[combo_count][sca_index] -= 1;
                *number_of_records -= 1;
            }
        }
    } /* End while point_index */

    /* When the algorithm above finishes, we are off by 1.  So add it here. */
    (*number_of_scas)++;

    /* Verify that the numbers listed in the file are what's actually in the
       file.  Only check SCAs that are actually in the file */
    for (sca_index = 0; sca_index < *number_of_scas; sca_index++)
    {
        for (combo_index = 0; combo_index < combo_count; combo_index++)
        {
            if (bandcombo->num_points[combo_index][sca_index] != 
                point_count[combo_index][sca_index])
            {
                IAS_LOG_ERROR("Number of points listed (%d) in file %s does "
                    "not equal number of points (%d) actually in file for the "
                    "band combination reference band number %d, search band "
                    "number %d, SCA %d\n", 
                    bandcombo->num_points[combo_index][sca_index], 
                    b2b_char_file, point_count[combo_index][sca_index], 
                    bandcombo->reference[combo_index][sca_index], 
                    bandcombo->search[combo_index][sca_index], sca_index + 1);
                fclose(file);
                free(point);
                return NULL;
            }
        } /* End for combo_index */
    } /* End for sca_index */

    /* Verify that we only have one sensor in the input data */
    if (oli_combo_count > 0 && tirs_combo_count > 0)
    {
        IAS_LOG_ERROR("Both OLI and TIRS have band combinations present");
        fclose(file);
        free(point);
        return NULL;
    }
    else if (oli_combo_count > 0)
    {
        bandcombo->sensor_id = IAS_OLI;
    }
    else if (tirs_combo_count > 0)
    {
        bandcombo->sensor_id = IAS_TIRS;
    }
    else
    {
        IAS_LOG_ERROR("No band/sca combinations present");
        fclose(file);
        free(point);
        return NULL;
    }

    /* Close the file */
    if (fclose(file) != 0)
    {
        IAS_LOG_ERROR("Closing band to band characterization file: %s",
                b2b_char_file);
        free(point);
        return NULL;
    }

    return point;
}
