/******************************************************************************
NAME: save_to_database

PURPOSE:
Sends algorithm report data to the database.

RETURNS: SUCCESS or ERROR
******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "ias_const.h"
#include "ias_miscellaneous.h" /* Used for ias_misc_get_char_id */
#include "ias_db_insert.h"     /* Database library */
#include "ias_db_insert_transaction_using_table.h" /* Database library */
#include "ias_db_get_connect_info.h" /* Database library */
#include "ias_logging.h"       /* IAS logging library */
#include "save_to_database.h"  /* Header definitions for this file */

#define SQL_QUERYLEN 256

int save_to_database
(
    struct MAIN_DB_SAVE_INFO* main_db_save, /* I: top-level data to save */
    struct BAND_SCA_DB_SAVE_INFO* band_sca_db_save, /* I: Band/SCA-level data 
                                  to save */
    int num_band_sca_db_records, /* I: The number of band/SCA DB records */
    struct PRE_POST_FIT_DB_SAVE_INFO* pre_post_fit_db_save, /* I: Pre and post 
                                  fit data to save */
    int num_pre_post_fit_db_records /* I: The number of pre/post fit DB
                                      records */
)
{
    int status;                   /* Function return status */
    int main_table_size;          /* Size of the main database table */
    int band_sca_table_size;      /* Size of the Band/SCA database table */
    int pre_post_fit_table_size;  /* Size of the Pre/Post fit database table */
    char *database_name = NULL;   /* Database name */
    char *user_name = NULL;       /* Database user name */
    char *password = NULL;        /* Database password */
    char *host = NULL;            /* Database host */
    char main_database_table_name[] = "band_calibration"; /* This table has
                                     records that apply to the whole
                                     application run */
    char band_sca_database_table_name[] = "band_calibration_band_sca"; /* This
                                     table has per-band, per-SCA records */
    char pre_post_fit_database_table_name[] = "band_cal_prepst_fit_detail"; 
                                  /* This table has pre-fit and post-fit
                                     records */
    char geometric_wo_common_id[IAS_CHAR_ID_SIZE];     
                                  /* Primary key in GEOMETRIC_WORK_ORDER_COMMON
                                     to associate processing with the WO */
    struct ias_db_connection *db = NULL; /* Database connection */
    char query_string[SQL_QUERYLEN]; /* SQL statement string */
    struct ias_db_query *query = NULL; /* SQL query */
    char error_message[IAS_SQL_RTN_MSGLEN]; /* SQL query error message */

    /* Get the database connection info */
    if (ias_db_get_connect_info(&database_name, &user_name, &password, &host)
        != SUCCESS)
    {
        /* Failed getting the database connection info */
        IAS_LOG_ERROR("Getting database connection info");
        return ERROR;
    }

    /* Connect to the database for a transaction */
    db = ias_db_connect_with_transaction(database_name, user_name, password,
             host);

    /* Free the database connection information */
    free(database_name);
    free(user_name);
    free(password);
    free(host);

    if (db == NULL)
    {
        IAS_LOG_ERROR("Connecting to database");
        ias_db_close_database_lib();
        return ERROR;
    }

    /* Get the database ID corresponding to the current work order */
    status = ias_misc_get_char_id(db,
        "GEOMETRIC_WORK_ORDER_COMMON", "geometric_wo_common_id",
        main_db_save->wo_id, geometric_wo_common_id);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Retrieving database ID corresponding to current "
                      "geometric work order %s", main_db_save->wo_id);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Put the Geometric WO common ID in the main table */
    strncpy(main_db_save->geometric_wo_common_id, geometric_wo_common_id, 
        sizeof(main_db_save->geometric_wo_common_id));
    main_db_save->geometric_wo_common_id[sizeof(
        main_db_save->geometric_wo_common_id) - 1] = '\0';

    /* Build the top-level table of data to save */
    IAS_DB_INSERT_TABLE main_table[] =
    {
        IAS_DB_TABLE_STRUCT_STRING(main_db_save, geometric_wo_common_id),
        IAS_DB_TABLE_STRUCT_INT(main_db_save, nbr_of_legendre_coefficients)
    };

    main_table_size = sizeof(main_table) / sizeof(IAS_DB_INSERT_TABLE);

    /* Delete the previous records */
    status = snprintf(query_string, sizeof(query_string),
                      "delete from %s where geometric_wo_common_id=%s",
                      main_database_table_name, geometric_wo_common_id);
    if (status < 0 || status >= sizeof(query_string))
    {
        IAS_LOG_ERROR("Creating string to delete records from %s",
                      main_database_table_name);
        return ERROR;
    }

    query = ias_db_perform_query(db, query_string);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, error_message, sizeof(
            error_message));
        IAS_LOG_ERROR("Query of %s failed: %s", main_database_table_name,
            error_message);
        IAS_LOG_ERROR("Deleting records from %s where geometric_wo_common_id="
            "%s", main_database_table_name, geometric_wo_common_id);
        ias_db_query_close(query);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Close query */
    ias_db_query_close(query);
    query = NULL;

    /* Insert the data into the main database table */
    status = ias_db_insert_using_table(db, main_database_table_name, main_table,
        main_table_size, 1);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Inserting to %s DB table", main_database_table_name);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Put the Geometric WO common ID in the Band/SCA table */
    strncpy(band_sca_db_save->geometric_wo_common_id, geometric_wo_common_id, 
        sizeof(band_sca_db_save->geometric_wo_common_id));
    band_sca_db_save->geometric_wo_common_id[sizeof(
        band_sca_db_save->geometric_wo_common_id) - 1] = '\0';

    /* Build the Band/SCA-level table of data to save */
    IAS_DB_INSERT_TABLE band_sca_table[] =
    {
        IAS_DB_TABLE_STRUCT_STRING(band_sca_db_save, geometric_wo_common_id),
        IAS_DB_TABLE_STRUCT_INT_ARRAY(band_sca_db_save, 
            band_id),
        IAS_DB_TABLE_STRUCT_INT_ARRAY(band_sca_db_save, 
            sensor_chip_assembly_id),
        IAS_DB_TABLE_STRUCT_INT_ARRAY(band_sca_db_save, nbr_of_pnts_used),

        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save,
            upds_alng_trk_leg_poly_0),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save,
            upds_alng_trk_leg_poly_1),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save,
            upds_alng_trk_leg_poly_2),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save,
            upds_alng_trk_leg_poly_3),

        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save,
            upds_acrss_trk_leg_poly_0),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save,
            upds_acrss_trk_leg_poly_1),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save,
            upds_acrss_trk_leg_poly_2),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save,
            upds_acrss_trk_leg_poly_3),

        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save, 
            new_alng_trk_leg_poly_0),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save, 
            new_alng_trk_leg_poly_1),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save, 
            new_alng_trk_leg_poly_2),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save, 
            new_alng_trk_leg_poly_3),

        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save,
            new_acrss_trk_leg_poly_0),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save,
            new_acrss_trk_leg_poly_1),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save,
            new_acrss_trk_leg_poly_2),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save,
            new_acrss_trk_leg_poly_3)
    };

    band_sca_table_size = sizeof(band_sca_table) / sizeof(IAS_DB_INSERT_TABLE);

    /* Delete the previous records */
    status = snprintf(query_string, sizeof(query_string),
                      "delete from %s where geometric_wo_common_id=%s",
                      band_sca_database_table_name, geometric_wo_common_id);
    if (status < 0 || status >= sizeof(query_string))
    {
        IAS_LOG_ERROR("Creating string to delete records from %s",
                      band_sca_database_table_name);
        return ERROR;
    }

    query = ias_db_perform_query(db, query_string);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, error_message, sizeof(
            error_message));
        IAS_LOG_ERROR("Query of %s failed: %s", band_sca_database_table_name,
            error_message);
        IAS_LOG_ERROR("Deleting records from %s where geometric_wo_common_id="
            "%s", band_sca_database_table_name, geometric_wo_common_id);
        ias_db_query_close(query);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Close query */
    ias_db_query_close(query);
    query = NULL;

    /* Insert the data into the SCA database table */
    status = ias_db_insert_using_table(db, band_sca_database_table_name, 
        band_sca_table, band_sca_table_size, num_band_sca_db_records);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Inserting to %s DB table", band_sca_database_table_name);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Put the Geometric WO common ID in the SCA table */
    strncpy(pre_post_fit_db_save->geometric_wo_common_id, 
        geometric_wo_common_id, 
        sizeof(pre_post_fit_db_save->geometric_wo_common_id));
    pre_post_fit_db_save->geometric_wo_common_id[sizeof(
        pre_post_fit_db_save->geometric_wo_common_id) - 1] = '\0';

    /* Build the SCA-level table of data to save */
    IAS_DB_INSERT_TABLE pre_post_fit_table[] =
    {
        IAS_DB_TABLE_STRUCT_STRING(pre_post_fit_db_save, 
            geometric_wo_common_id),
        IAS_DB_TABLE_STRUCT_INT_ARRAY(pre_post_fit_db_save, 
            sensor_chip_assembly_id),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(pre_post_fit_db_save,
            pre_fit_res_alng_trk_mean),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(pre_post_fit_db_save,
            pre_fit_res_alng_trk_std_dev),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(pre_post_fit_db_save,
            pre_fit_res_alng_trk_rmse),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(pre_post_fit_db_save,
            pre_fit_res_acrss_trk_mean),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(pre_post_fit_db_save,
            pre_fit_res_acrss_trk_std_dev),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(pre_post_fit_db_save,
            pre_fit_res_acrss_trk_rmse),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(pre_post_fit_db_save,
            pst_fit_res_alng_trk_mean),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(pre_post_fit_db_save,
            pst_fit_res_alng_trk_std_dev),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(pre_post_fit_db_save,
            pst_fit_res_alng_trk_rmse),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(pre_post_fit_db_save,
            pst_fit_res_acrss_trk_mean),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(pre_post_fit_db_save,
            pst_fit_res_acrss_trk_std_dev),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(pre_post_fit_db_save,
            pst_fit_res_acrss_trk_rmse)
    };

    pre_post_fit_table_size = sizeof(pre_post_fit_table)
        / sizeof(IAS_DB_INSERT_TABLE);

    /* Delete the previous records */
    status = snprintf(query_string, sizeof(query_string),
                      "delete from %s where geometric_wo_common_id=%s",
                      pre_post_fit_database_table_name,
                      geometric_wo_common_id);
    if (status < 0 || status >= sizeof(query_string))
    {
        IAS_LOG_ERROR("Creating string to delete records from %s",
                      pre_post_fit_database_table_name);
        return ERROR;
    }

    query = ias_db_perform_query(db, query_string);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, error_message, sizeof(
            error_message));
        IAS_LOG_ERROR("Query of %s failed: %s", 
            pre_post_fit_database_table_name, error_message);
        IAS_LOG_ERROR("Deleting records from %s where geometric_wo_common_id="
            "%s", pre_post_fit_database_table_name, geometric_wo_common_id);
        ias_db_query_close(query);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Close query */
    ias_db_query_close(query);
    query = NULL;

    /* Insert the data into the Pre and Post fit database table */
    status = ias_db_insert_using_table(db, pre_post_fit_database_table_name, 
        pre_post_fit_table, pre_post_fit_table_size, 
        num_pre_post_fit_db_records);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Inserting to %s DB table", 
            pre_post_fit_database_table_name);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* The inserts worked, so commit the transaction */
    status = ias_db_disconnect_with_transaction(db, IAS_DB_COMMIT);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Committing transaction and disconnecting from database");
    }

    return status;
}
