/*****************************************************************************
NAME: solve_focal_plane

PURPOSE: Determines the line-of-sight Legendre polynomial updates based on
         band-to-band residuals (or measured offsets).

Algorithm references:
    OLI Band-to-Band Calibration Algorithm ADD
    TIRS Band-to-Band Calibration Algorithm ADD

    Please refer to these ADD's for more details on the meaning of the main
    matrix and vector variable names used here.
*****************************************************************************/

#include <limits.h>
#include <stdlib.h>
#include <math.h>

#include "ias_const.h"
#include "ias_structures.h"
#include "ias_logging.h"         /* Logging library */
#include "ias_satellite_attributes.h" /* Satellite attributes library */
#include "ias_math.h"
#include "ias_grid.h"
#include "ias_los_model.h"

#include "save_to_database.h"
#include "residuals_stats.h"
#include "bandcal.h"

#define X_Y 2               /* Dimension size for x & y directions */
#define X_ALONG  0          /* Array elements for along offset direction */
#define Y_ACROSS 1          /* Array elements for across offset direction */ 
#define MAX_LEGENDRE_COEFFS 4 /* Maximum number of Legendre coefficients */

static int collect_database_data
(
    double sol_along_track[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* I: Solution offsets in line direction */
    double sol_across_track[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* I: Solution offsets in sample direction */
    double new_legendre_along[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* I: New Legendre coefficients along track
                                       direction */
    double new_legendre_across[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* I: New Legendre coefficients across track
                                       direction */
    const FIT_RESIDUAL_STATS_TYPE fit_stats[IAS_MAX_NSCAS], /* I: Statistics */
    int point_count[IAS_MAX_NBANDS][IAS_MAX_NSCAS], /* I: Number of points used
                                    per band/SCA */
    int nsca,                    /* I: Number of SCAs */
    int num_bands,               /* I: Total number of bands in sensor */
    int band_numbers[IAS_MAX_NBANDS], /* I: List of band numbers */
    const BANDCOMBO *bandcombo,  /* I: Band combinations */
    int *num_band_sca_records,   /* O: The number of band/SCA DB records */
    int *num_fit_records,        /* O: The number of pre/post fit DB records */
    struct BAND_SCA_DB_SAVE_INFO* band_sca_db_save, /* O: Band/SCA-level data
                                       to save */
    struct PRE_POST_FIT_DB_SAVE_INFO* pre_post_fit_db_save /* O: Pre and post
                                       fit data to save */
);

static void calculate_statistics
(
    STATS_TYPE *stats, /* I/O: Collection of statistics */
    int num_points,    /* I: Number of points used to generate statistics */
    double scale       /* I: Scale used in calculating statistics */
);

static int populate_residuals
(
    int nsca,                    /* I: Number of SCAs */
    int number_of_combinations,  /* I: Number of band combinations to process */
    const POINT *points,         /* I: Points from B2B char file */
    int num_points,              /* I: Total number of points */
    const IAS_GRID_TYPE *grid,   /* I: Grid with all bands */
    const IAS_LOS_MODEL *model,  /* I: Model */
    const IAS_IMAGE *dem,        /* I: DEM data */
    int tie_point_list_flag,     /* I: Flag for creating tie point report */
    const int *sca_indexes,      /* I: List of SCA indexes to process */
    int num_coeffs,              /* I: Number of coefficients to solve for */
    double sol_along_track[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* I: Solution offsets in line direction */
    double sol_across_track[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* I: Solution offsets in sample direction */
    int nl[IAS_MAX_NBANDS][IAS_MAX_NSCAS], /* I: Number lines in input data */
    int ns[IAS_MAX_NBANDS][IAS_MAX_NSCAS], /* I: Number samples in input data */
    FIT_RESIDUAL_STATS_TYPE fit_stats[IAS_MAX_NSCAS], /* O: Statistics */
    RESIDUALS_PNTS *res          /* O: Offsets between search and reference */
);

static void setup_legendre_coefficients
(
    int num_coeffs,               /* I: Number of coefficients to solve for */
    double normalized_detector,   /* I: Normalized detector value */
    double legr_coeff[MAX_LEGENDRE_COEFFS] /* O: Legendre matrix */
);

/***************************************************************************
NAME: solve_focal_plane

PURPOSE: Determines the line-of-sight Legendre polynomial updates based on
         band-to-band residuals (or measured offsets).

RETURN VALUE: Type = int
    Value    Description
    -----    -----------
    SUCCESS  Successful completion
    ERROR    Operation failed
*******************************************************************************/
int solve_focal_plane
(
    const char l1g[PATH_MAX],    /* I: Name of the L1G for report data */
    const char l0r[PATH_MAX],    /* I: Name of the L0R HDF file */
    const char b2bcal_file[PATH_MAX], /* I: Name of band to band report file */
    const IAS_GRID_TYPE *grid,   /* I: Grid with all bands */
    const IAS_LOS_MODEL *model,  /* I: Model */
    const IAS_IMAGE *dem,        /* I: DEM data */
    int num_points,              /* I: Total number of points */
    int nsca,                    /* I: Number of SCAs */
    const int *sca_indexes,      /* I: List of SCA indexes to process */
    int minimum_points,          /* I: Minimum points to process */
    const BANDCOMBO *bandcombo,  /* I: Band combinations */
    int num_coeffs,              /* I: Number of coefficients to solve for */
    const POINT *points,         /* I: Points from B2B char file */
    double along_track_solution[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* O: Solution offsets in line direction */
    double across_track_solution[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* O: Solution offsets in sample direction */
    double new_legendre_along[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* O: New Legendre coefficients along track 
                                       direction */
    double new_legendre_across[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* O: New Legendre coefficients across track 
                                       direction */
    int number_of_combinations,  /* I: Number of band combinations to process */
    int tie_point_list_flag,     /* I: Flag for creating tie point report */
    int store_in_database_flag,  /* I: Flag for saving results in database */
    struct BAND_SCA_DB_SAVE_INFO* band_sca_db_save, /* O: Band/SCA-level data
                                       to save */
    struct PRE_POST_FIT_DB_SAVE_INFO* pre_post_fit_db_save, /* O: Pre and post
                                       fit data to save */
    const char *work_order_id,   /* I: Work Order ID for report header */
    int *num_band_sca_records,   /* O: The number of band/SCA DB records */
    int *num_fit_records         /* O: The number of pre/post fit DB records */
)
{
    double ref_line;       /* Reference line location */
    double ref_samp;       /* Reference sample location */
    double search_line;    /* Search line location */
    double search_samp;    /* Search sample location */
    int status;            /* Return status for function */
    int ref_band_index;    /* Index of band that is being used as reference */
    int search_band_index; /* Index of band that is being used as seach */
    int ref_band_lines_per_frame; /* Lines per frame for reference band */
    int search_band_lines_per_frame; /* Lines per frame for search band */
    int row;               /* Row loop variable */
    int col;               /* Column loop variable */
    int array_element;     /* Combined row and column array loop variable */
    int coeff_index;       /* Loop variable */
    int coeff;             /* Loop variable */
    int minimum_points_offset; /* Minimum points offset */
    int sca_counter;       /* Counter for SCA loops */
    int combo_counter;     /* Counter for band combinations */
    int sca_index;         /* SCA index */
    int band_index;        /* Band index */
    int band;              /* Band loop counter */
    int count;             /* Number of residual points */
    double norm_det;       /* Normalized detector within an SCA */
    int nl[IAS_MAX_NBANDS][IAS_MAX_NSCAS]; /* Number of lines in input data */
    int ns[IAS_MAX_NBANDS][IAS_MAX_NSCAS]; /* Number of samples in input data */
    double los_err_samp;   /* Difference in LOS between reference and search 
                              in sample direction */   
    double los_err_line;   /* Difference in LOS between reference and search 
                              in line direction */    
    double ref_input_samp; /* Sample location in input space for reference 
                              point */ 
    double ref_input_line; /* Line location in input space for reference 
                              point */
    double search_input_samp; /* Sample location in input space for search 
                              point */ 
    double search_input_line; /* Line location in input space for search 
                              point */
    double *A;             /* Matrix for least squares solution */
    double Y[X_Y];         /* Observation vector */
    double C[X_Y * X_Y] = {16.0, 0.0, 0.0, 16.0}; /* weight matrix */
    double *AT;            /* Transpose of observation coefficient matrix */
    double *ATC;           /* (transpose of A)*C */
    double *N;             /* Normal matrix */
    double *L;             /* Sequence of observation vector */
    double *Nsum;          /* Sum of the normal matrices for each pt */
    double *Lsum;          /* Sum of the observation vectors */
    double *Lsum_pointer = NULL; /* Solution pointer to Lsum */
    double *work_vector;   /* Work vector for QR routines */
    int num_a_elements;    /* Number of columns in A matrix */
    /* 00's denote constraint variables */
    double *A00;           /* Observation coefficient matrix constraint */
    double *AT00;          /* Transpose of observation coef matrix constraint */
    double C00[X_Y * X_Y] = {10000.0, 0.0, 0.0, 10000.0};
    double Y00[X_Y] = {0, 0}; /* Observation vector for constraint */
    double *ATC00;         /* (transpose of A00)*C00 */
    double legr_coeff[MAX_LEGENDRE_COEFFS]; /* X variable for Legendre matrix */
    int first_reference_band_number; /* First reference for current combo */
    int first_search_band_number; /* First search for current combination */
    double scale;          /* Scale used in calculating statistics */
    FIT_RESIDUAL_STATS_TYPE fit_stats[IAS_MAX_NSCAS]; /* Statistics structure */
    RESIDUALS_PNTS *residuals = NULL; /* Tie point offsets structure */
    int point_count[IAS_MAX_NBANDS][IAS_MAX_NSCAS]; /* Number of points used 
                              per band/SCA */
    int band_offsets[IAS_MAX_NBANDS]; /* Enforce pan band in first columns */
    int offset_count;      /* Keep track of which offset is being set */
    int pan_band_number;   /* first band number for the OLI setup of matrices */
    int pan_band_index = 0; /* first band index for the OLI setup of matrices */
    int band_numbers[IAS_MAX_NBANDS]; /* List of band numbers */
    int num_bands;         /* Total number of bands in sensor */

    /* Get band info */
    status = ias_sat_attr_get_sensor_band_numbers(bandcombo->sensor_id,
        IAS_NORMAL_BAND, 0, band_numbers, IAS_MAX_NBANDS, &num_bands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting sensor's normal band numbers");
        return ERROR;
    }

    if (bandcombo->sensor_id == IAS_OLI)
    {
        /* For OLI the pan band has to be first in the matrices */
        /* Get the pan band number */
        pan_band_number =
            ias_sat_attr_get_band_number_from_type(IAS_PAN_BAND);
        if (pan_band_number == ERROR)
        {
            IAS_LOG_ERROR("Invalid PAN band number found using satellite "
                "attributes library");
            return ERROR;
        }

        /* Get the pan band index */
        pan_band_index = ias_sat_attr_convert_band_number_to_index(
            pan_band_number);
        if (pan_band_index == ERROR)
        {
            IAS_LOG_ERROR("Invalid band index for PAN band number %d", 
                pan_band_number);
            return ERROR;
        }

        band_offsets[pan_band_index] = 0;
        offset_count = 1;
    }
    else if (bandcombo->sensor_id == IAS_TIRS)
    {
        offset_count = 0;
    }
    else
    {
        /* Should never get here */
        IAS_LOG_ERROR("Invalid sensor");
        return ERROR;
    }

    /* Build band offsets table.  */
    for (band = 0; band < num_bands; band++)
    {
        /* Get the band index */
        band_index = ias_sat_attr_convert_band_number_to_index(
            band_numbers[band]);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Invalid band index for band number %d",
                band_numbers[band]);
            return ERROR;
        }

        if (bandcombo->sensor_id == IAS_OLI)
        {
            /* We already took care of the pan band */
            if (band_index != pan_band_index)
            {
                band_offsets[band_index] = offset_count++;
            }
        }
        else
        {
            band_offsets[band_index] = offset_count++;
        }
    }

    /* Initialize solution, point count, and missing bands */
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        for (sca_index = 0; sca_index < IAS_MAX_NSCAS; sca_index++)
        {
            point_count[band_index][sca_index] = 0;
            for (coeff = 0; coeff < IAS_LOS_LEGENDRE_TERMS; coeff++)
            {
                along_track_solution[band_index][sca_index][coeff] = 0.0;
                across_track_solution[band_index][sca_index][coeff] = 0.0;
            }
        }
    }

    /* Initialize the statistics structure */
    for (sca_index = 0; sca_index < IAS_MAX_NSCAS; sca_index++)
    {
        fit_stats[sca_index].points = minimum_points;

        fit_stats[sca_index].pre_fit_x.mean = 0;
        fit_stats[sca_index].pre_fit_y.mean = 0;
        fit_stats[sca_index].pre_fit_x.stdev = 0;
        fit_stats[sca_index].pre_fit_y.stdev = 0;
        fit_stats[sca_index].pre_fit_x.rmse = 0;
        fit_stats[sca_index].pre_fit_y.rmse = 0;
        fit_stats[sca_index].pre_fit_x.sum = 0;
        fit_stats[sca_index].pre_fit_y.sum = 0;
        fit_stats[sca_index].pre_fit_x.sum_sqr_err = 0;
        fit_stats[sca_index].pre_fit_y.sum_sqr_err = 0;
        fit_stats[sca_index].post_fit_x.mean = 0;
        fit_stats[sca_index].post_fit_y.mean = 0;
        fit_stats[sca_index].post_fit_x.stdev = 0;
        fit_stats[sca_index].post_fit_y.stdev = 0;
        fit_stats[sca_index].post_fit_x.rmse = 0;
        fit_stats[sca_index].post_fit_y.rmse = 0;
        fit_stats[sca_index].post_fit_x.sum = 0;
        fit_stats[sca_index].post_fit_y.sum = 0;
        fit_stats[sca_index].post_fit_x.sum_sqr_err = 0;
        fit_stats[sca_index].post_fit_y.sum_sqr_err = 0;
    }

    /* Set number of lines and samples arrays; used for convenience */
    for (band = 0; band < num_bands; band++)
    {
        /* Get the band index */
        band_index = ias_sat_attr_convert_band_number_to_index(
            band_numbers[band]);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Invalid band index for band number %d",
                band_numbers[band]);
            return ERROR;
        }

        for (sca_counter = 0; sca_counter < nsca; sca_counter++)
        {
            sca_index = sca_indexes[sca_counter];
            nl[band_index][sca_index] = grid->gridbands[band_index][sca_index].
                in_lines[grid->gridbands[band_index][sca_index].ngrid_lines-1];
            ns[band_index][sca_index] = grid->gridbands[band_index][sca_index].
                in_samps[grid->gridbands[band_index][sca_index].ngrid_samps-1];
        }
    }

    /* Initialize arrays for LSQ */
    num_a_elements = num_coeffs * num_bands * 2;
    A = (double *)malloc(sizeof(double) * num_a_elements * X_Y);
    if (A == NULL)
    {
        IAS_LOG_ERROR("Allocating matrix A\n");
        return ERROR;
    }
    AT = (double *)malloc(sizeof(double) * num_a_elements * X_Y);
    if (AT == NULL)
    {
        IAS_LOG_ERROR("Allocating matrix AT\n");
        free(A);
        return ERROR;
    }
    ATC = (double *)malloc(sizeof(double) * num_a_elements * X_Y);
    if (ATC == NULL)
    {
        IAS_LOG_ERROR("Allocating matrix ATC\n");
        free(A);
        free(AT);
        return ERROR;
    }
    N = (double *)malloc(sizeof(double) * num_a_elements * num_a_elements);
    if (N == NULL)
    {
        IAS_LOG_ERROR("Allocating normal matrix N\n");
        free(A);
        free(AT);
        free(ATC);
        return ERROR;
    }
    L = (double *)malloc(sizeof(double) * num_a_elements);
    if (L == NULL)
    {
        IAS_LOG_ERROR("Allocating observation vector L\n");
        free(A);
        free(N);
        free(AT);
        free(ATC);
        return ERROR;
    }
    Nsum = (double *)malloc(sizeof(double) * num_a_elements * num_a_elements);
    if (Nsum == NULL)
    {
        IAS_LOG_ERROR("Allocating matrix Nsum\n");
        free(L);
        free(A);
        free(N);
        free(AT);
        free(ATC);
        return ERROR;
    }
    Lsum = (double *)malloc(sizeof(double) * num_a_elements);
    if (Lsum == NULL)
    {
        IAS_LOG_ERROR("Allocating vector Lsum\n");
        free(L);
        free(A);
        free(N);
        free(Nsum);
        free(AT);
        free(ATC);
        return ERROR;
    }
    A00 = (double *)malloc(sizeof(double) * X_Y * num_a_elements);
    if (A00 == NULL)
    {
        IAS_LOG_ERROR("Allocating matrix A00\n");
        free(L);
        free(A);
        free(N);
        free(Nsum);
        free(Lsum);
        free(AT);
        free(ATC);
        return ERROR;
    }
    AT00 = (double *)malloc(sizeof(double) * X_Y * num_a_elements);
    if (AT00 == NULL)
    {
        IAS_LOG_ERROR("Allocating matrix AT00\n");
        free(L);
        free(A);
        free(A00);
        free(N);
        free(Nsum);
        free(Lsum);
        free(AT);
        free(ATC);
        return ERROR;
    }
    ATC00 = (double *)malloc(sizeof(double) * X_Y * num_a_elements);
    if (ATC00 == NULL)
    {
        IAS_LOG_ERROR("Allocating matrix ATC00\n");
        free(L);
        free(A);
        free(A00);
        free(AT00);
        free(Lsum);
        free(N);
        free(Nsum);
        free(AT);
        free(ATC);
        return ERROR;
    }
    /* Setup work vector pointer for QR routines */
    work_vector = malloc(num_a_elements * sizeof(double));
    if (work_vector == NULL)
    {
        IAS_LOG_ERROR("Allocating work vector work_vector");
        free(L);
        free(A);
        free(A00);
        free(AT00);
        free(ATC00);
        free(Lsum);
        free(N);
        free(Nsum);
        free(AT);
        free(ATC);
        return ERROR;
    }

    count = 0;
    for (sca_counter = 0; sca_counter < nsca; sca_counter++)
    {
        sca_index = sca_indexes[sca_counter];

        /* Zero arrays */
        for (row = 0; row < num_a_elements; row++)
        {
            Lsum[row] = 0.0;
        }
        for (array_element = 0;
             array_element < num_a_elements * num_a_elements;
             array_element++)
        {
            Nsum[array_element] = 0.0;
        }
        for (band = 0; band < num_bands; band++)
        {
            /* Get the band index */
            band_index = ias_sat_attr_convert_band_number_to_index(
                band_numbers[band]);
            if (band_index == ERROR)
            {
                IAS_LOG_ERROR("Invalid band index for band number %d",
                    band_numbers[band]);
                free(L);
                free(A);
                free(A00);
                free(AT00);
                free(ATC00);
                free(Lsum);
                free(N);
                free(Nsum);
                free(AT);
                free(ATC);
                free(work_vector);
                return ERROR;
            }
            point_count[band_index][sca_index] = 0;
        }

        /* Set up pan band constraint */
        for (array_element = 0; array_element < X_Y * num_a_elements; 
             array_element++)
        {
            A00[array_element] = 0.0;
            AT00[array_element] = 0.0;
            ATC00[array_element] = 0.0;
        }

        combo_counter = 0;
        while (combo_counter < number_of_combinations)
        {
            ref_band_index = ias_sat_attr_convert_band_number_to_index(
                points[count].reference_band_number);
            if (ref_band_index == ERROR)
            {
                IAS_LOG_ERROR("Invalid band index for reference band number "
                    "%d", points[count].reference_band_number);
                free(L);
                free(A);
                free(A00);
                free(AT00);
                free(ATC00);
                free(Lsum);
                free(N);
                free(Nsum);
                free(AT);
                free(ATC);
                free(work_vector);
                return ERROR;
            }

            search_band_index = ias_sat_attr_convert_band_number_to_index(
                points[count].search_band_number);
            if (search_band_index == ERROR)
            {
                IAS_LOG_ERROR("Invalid band index for search band number %d", 
                    points[count].search_band_number);
                free(L);
                free(A);
                free(A00);
                free(AT00);
                free(ATC00);
                free(Lsum);
                free(N);
                free(Nsum);
                free(AT);
                free(ATC);
                free(work_vector);
                return ERROR;
            }

            first_reference_band_number = points[count].reference_band_number;
            first_search_band_number = points[count].search_band_number;

            if (bandcombo->band_sca_not_present[ref_band_index][sca_index]
                    == TRUE ||
                bandcombo->band_sca_not_present[search_band_index][sca_index]
                    == TRUE)
            {
                IAS_LOG_ERROR("Bands %d %d were listed as missing",
                    points[count].reference_band_number,
                    points[count].search_band_number);
                free(L);
                free(A);
                free(A00);
                free(AT00);
                free(ATC00);
                free(Lsum);
                free(N);
                free(Nsum);
                free(AT);
                free(ATC);
                free(work_vector);
                return ERROR;
            }

            while ((first_reference_band_number ==
                       points[count].reference_band_number) &&
                   (first_search_band_number ==
                       points[count].search_band_number) &&
                   ((sca_index + 1) == points[count].reference_sca_number) &&
                   (count < num_points))
            {
                /* Find location of correlation point in both the reference 
                   and search output spaces */
                ref_line = points[count].reference_line;
                ref_samp = points[count].reference_samp;
                ref_band_lines_per_frame = model->sensor.bands[ref_band_index].
                    sampling_char.lines_per_frame;
                search_band_lines_per_frame = 
                    model->sensor.bands[search_band_index].sampling_char.
                    lines_per_frame;

                if (ref_band_lines_per_frame == 0 || 
                    search_band_lines_per_frame == 0)
                {
                    IAS_LOG_ERROR("Lines per frame in model is 0 for "
                        "reference band number %d and/or search band number "
                        "%d", points[count].reference_band_number,
                        points[count].search_band_number);
                    free(L);
                    free(A);
                    free(A00);
                    free(AT00);
                    free(ATC00);
                    free(Lsum);
                    free(N);
                    free(Nsum);
                    free(AT);
                    free(ATC);
                    free(work_vector);
                    return ERROR;
                }
                    
                ref_line *= ref_band_lines_per_frame 
                    / search_band_lines_per_frame; 
                ref_samp *= ref_band_lines_per_frame 
                    / search_band_lines_per_frame; 

                search_line = points[count].search_line + points[count].dy;
                search_samp = points[count].search_samp + points[count].dx;

                point_count[ref_band_index][sca_index]++;
                point_count[search_band_index][sca_index]++;

                /* Determine LOS errors */
                if (get_los_errors(ref_band_index, search_band_index, 
                        sca_index, ref_line, ref_samp, search_line, 
                        search_samp, dem, 
                        &grid->gridbands[ref_band_index][sca_index], 
                        &grid->gridbands[search_band_index][sca_index],
                        model,
                        nl[ref_band_index][sca_index], 
                        ns[ref_band_index][sca_index], 
                        nl[search_band_index][sca_index], 
                        ns[search_band_index][sca_index],
                        num_coeffs,
                        &ref_input_line, 
                        &ref_input_samp,
                        &search_input_line,
                        &search_input_samp,
                        &los_err_line,
                        &los_err_samp) != SUCCESS)
                {
                    IAS_LOG_ERROR("Could not find LOS errors for point %d id "
                        "%d SCA number %d reference band index %d search band "
                        "index %d reference line %f reference sample %f search "
                        "line %f search sample %f\n", count, 
                        points[count].point_id, sca_index + 1, ref_band_index, 
                        search_band_index, ref_line, ref_samp, search_line, 
                        search_samp);
                    free(L);
                    free(A);
                    free(A00);
                    free(AT00);
                    free(ATC00);
                    free(Lsum);
                    free(N);
                    free(Nsum);
                    free(AT);
                    free(ATC);
                    free(work_vector);
                    return ERROR;
                }

                Y[X_ALONG] = los_err_line;
                Y[Y_ACROSS] = los_err_samp;

                /* Fill in pre-fit statistics */
                fit_stats[sca_index].pre_fit_x.sum += los_err_samp; 
                fit_stats[sca_index].pre_fit_y.sum += los_err_line;
                fit_stats[sca_index].pre_fit_x.sum_sqr_err += los_err_samp 
                    * los_err_samp;
                fit_stats[sca_index].pre_fit_y.sum_sqr_err += los_err_line 
                    * los_err_line;
                fit_stats[sca_index].points++;

                /* Initialize the matrices to zero */
                for (row = 0; row < num_a_elements; row++)
                {
                    L[row] = 0.0;
                    for (col = 0; col < num_a_elements; col++)
                    {
                        N[row * num_a_elements + col] = 0.0;
                    }
                    for (col = 0; col < X_Y; col++)
                    {
                        AT[row * X_Y + col] = 0.0;
                        ATC[row * X_Y + col] = 0.0;
                    }/* End for col */
                }/* End for row */
                for (row = 0; row < X_Y; row++)
                {
                    for (col = 0; col < num_a_elements; col++)
                    {
                        A[row * num_a_elements + col] = 0.0;
                    }
                }

                /* Check for a divide by 0 */
                if ((model->sensor.bands[ref_band_index].scas[sca_index]
                    .detectors == 1)
                || (model->sensor.bands[search_band_index].scas[sca_index]
                    .detectors == 1))
                {
                    IAS_LOG_ERROR("Number of detectors in model is 1 for "
                        "reference band number %d and/or search band number "
                        "%d, leading to divide by 0", 
                        points[count].reference_band_number,
                        points[count].search_band_number);
                    free(L);
                    free(A);
                    free(A00);
                    free(AT00);
                    free(ATC00);
                    free(Lsum);
                    free(N);
                    free(Nsum);
                    free(AT);
                    free(ATC);
                    free(work_vector);
                    return ERROR;
                }

                /* Compute the normalized detector number */
                norm_det = 2.0 * ref_input_samp 
                    / ((double)model->sensor.bands[ref_band_index]
                      .scas[sca_index].detectors - 1.0) - 1.0;

                /* Form the Legendre terms into the A matrix for this point */
                setup_legendre_coefficients(num_coeffs, norm_det, legr_coeff);

                /* Fill in the reference band's partials */
                for (coeff = 0; coeff < num_coeffs; coeff++)
                {
                    A[X_ALONG * num_a_elements + num_coeffs 
                        * band_offsets[ref_band_index] * X_Y + coeff] = 
                        -legr_coeff[coeff];
                    A[Y_ACROSS * num_a_elements + num_coeffs 
                        * band_offsets[ref_band_index] * X_Y + coeff 
                        + num_coeffs] = -legr_coeff[coeff];
                }

                /* Compute the normalized detector number */
                norm_det = 2.0 * search_input_samp 
                    / ((double)model->sensor.bands[search_band_index].
                    scas[sca_index].detectors - 1.0) - 1.0;

                /* Form the Legendre terms into the A matrix for this point */
                setup_legendre_coefficients(num_coeffs, norm_det, legr_coeff);

                for (coeff = 0; coeff < num_coeffs; coeff++)
                {
                    /* Fill in the search band's partials */
                    A[X_ALONG * num_a_elements + num_coeffs 
                        * band_offsets[search_band_index] * X_Y + coeff] 
                        = legr_coeff[coeff];
                    A[Y_ACROSS * num_a_elements + num_coeffs 
                        * band_offsets[search_band_index] * X_Y + coeff 
                        + num_coeffs] 
                        = legr_coeff[coeff];
                }              

                ias_math_transpose_matrix(A, AT, X_Y, num_a_elements);

                /* Find A(transpose)*C */
                status = ias_math_multiply_matrix(AT, C, ATC, num_a_elements, 
                        X_Y, X_Y, X_Y);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Multiplying matrices AT and C");
                    free(L);
                    free(A);
                    free(A00);
                    free(AT00);
                    free(ATC00);
                    free(Lsum);
                    free(N);
                    free(Nsum);
                    free(AT);
                    free(ATC);
                    free(work_vector);
                    return ERROR;
                }

                /* Find normal matrix, N= A(transpose)*C*A */
                status = ias_math_multiply_matrix(ATC, A, N, num_a_elements, 
                        X_Y, X_Y, num_a_elements);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Multiplying matrices ATC and A");
                    free(L);
                    free(A);
                    free(A00);
                    free(AT00);
                    free(ATC00);
                    free(Lsum);
                    free(N);
                    free(Nsum);
                    free(AT);
                    free(ATC);
                    free(work_vector);
                    return ERROR;
                }

                /* Add the normal matrix info to the Sum */
                for (row = 0; row < num_a_elements; row++)
                {
                    for (col = 0; col < num_a_elements; col++)
                    {
                        Nsum[row * num_a_elements + col] += 
                            N[row * num_a_elements + col];
                    }
                }

                /* Find the sequence of observations L= A(transpose)*C*Y */
                status = ias_math_multiply_matrix(ATC, Y, L, num_a_elements, 
                        X_Y, X_Y, 1);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Multiplying matrices ATC and Y");
                    free(L);
                    free(A);
                    free(A00);
                    free(AT00);
                    free(ATC00);
                    free(Lsum);
                    free(N);
                    free(Nsum);
                    free(AT);
                    free(ATC);
                    free(work_vector);
                    return ERROR;
                }

                /* Add the L info to the Sum of the L info */
                for (row = 0; row < num_a_elements; row++)
                {
                    Lsum[row] += L[row];
                }

                count++;
            }
            combo_counter++;
        }

        /* Calculate pre-fit statistics */
        scale = 1.0 / ((double)fit_stats[sca_index].points 
            * ((double)fit_stats[sca_index].points - 1.0));
        calculate_statistics(&fit_stats[sca_index].pre_fit_x,
            fit_stats[sca_index].points, scale);
        calculate_statistics(&fit_stats[sca_index].pre_fit_y,
            fit_stats[sca_index].points, scale);

        /* Set diagonals of Nsum to minimum points for missing bands */
        for (band = 0; band < num_bands; band++)
        {
            /* Get the band index */
            band_index = ias_sat_attr_convert_band_number_to_index(
                band_numbers[band]);
            if (band_index == ERROR)
            {
                IAS_LOG_ERROR("Invalid band index for band number %d",
                    band_numbers[band]);
                free(L);
                free(A);
                free(A00);
                free(AT00);
                free(ATC00);
                free(Lsum);
                free(N);
                free(Nsum);
                free(AT);
                free(ATC);
                free(work_vector);
                return ERROR;
            }

            /* If the band wasn't found, process it as an omitted band */
            if (bandcombo->band_sca_not_present[band_index][sca_index] == TRUE)
            {
                /* Process all rows/columns for the band index */
                for (coeff = 0; coeff < num_coeffs; coeff++)
                {
                    /* Set minimum points along diagonal extent of omitted band
                       index.  Both along and across elements are set. 
                       num_a_elements is num_coeffs * num_bands * 2 */
                    minimum_points_offset = coeff + band_offsets[band_index] 
                        * num_coeffs * 2;
                    Nsum[minimum_points_offset * num_a_elements 
                        + minimum_points_offset] = minimum_points;
                    minimum_points_offset = coeff + band_offsets[band_index] 
                        * num_coeffs * 2 + num_coeffs;
                    Nsum[minimum_points_offset * num_a_elements 
                        + minimum_points_offset] = minimum_points;
                }
            }
        }

        for (coeff = 0; coeff < num_coeffs; coeff++)
        {
            /* Initialize the constraint matrix */
            for (row = 0; row < X_Y * num_a_elements; row++)
                A00[row] = 0.0;
            A00[coeff] = 1.0;
            A00[num_a_elements + num_coeffs + coeff] = 1.0;

            /* Add the constraint */
            /* Find the transpose of A00 */
            ias_math_transpose_matrix(A00, AT00, X_Y, num_a_elements);

            /* Find A(transpose)*C */
            status = ias_math_multiply_matrix(AT00, C00, ATC00, num_a_elements,
                    X_Y, X_Y, X_Y);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Multiplying matrices AT00 and C00");
                free(L);
                free(A);
                free(A00);
                free(AT00);
                free(ATC00);
                free(Lsum);
                free(N);
                free(Nsum);
                free(AT);
                free(ATC);
                free(work_vector);
                return ERROR;
            }

            /* Find N = A(transpose)*C*A */
            status = ias_math_multiply_matrix(ATC00, A00, N, num_a_elements, 
                    X_Y, X_Y, num_a_elements);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Multiplying matrices ATC00 and A00");
                free(L);
                free(A);
                free(A00);
                free(AT00);
                free(ATC00);
                free(Lsum);
                free(N);
                free(Nsum);
                free(AT);
                free(ATC);
                free(work_vector);
                return ERROR;
            }

            /* Add the constraint N to the Sum */
            for (row = 0; row < num_a_elements; row++)
            {
                for (col = 0; col < num_a_elements; col++)
                {
                    Nsum[row * num_a_elements + col] += 
                        N[row * num_a_elements + col];
                }
            }

            /* Find L = A(transpose)*C*Y */
            status = ias_math_multiply_matrix(ATC00, Y00, L, num_a_elements, 
                    X_Y, X_Y, 1);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Multiplying matrices ATC00 and Y00");
                free(L);
                free(A);
                free(A00);
                free(AT00);
                free(ATC00);
                free(Lsum);
                free(N);
                free(Nsum);
                free(AT);
                free(ATC);
                free(work_vector);
                return ERROR;
            }

            /* Add the constraint L info to the Sum */
            for (row = 0; row < num_a_elements; row++)
            {
                Lsum[row] += L[row]; 
            }
        }

        /* Solve for the band offsets */
        status = ias_math_matrix_QRfactorization(Nsum, num_a_elements, 
            num_a_elements, work_vector, 0);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Preparing for least squares solution using Nsum");
            free(L);
            free(A);
            free(A00);
            free(AT00);
            free(ATC00);
            free(Lsum);
            free(N);
            free(Nsum);
            free(AT);
            free(ATC);
            free(work_vector);
            return ERROR;
        }

        status = ias_math_matrix_QRsolve(Nsum, num_a_elements, num_a_elements, 
            work_vector, Lsum, 0);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Finding least squares solution using Nsum");
            free(L);
            free(A);
            free(A00);
            free(AT00);
            free(ATC00);
            free(Lsum);
            free(N);
            free(Nsum);
            free(AT);
            free(ATC);
            free(work_vector);
            return ERROR;
        }

        /* Setup the solution pointer */
        Lsum_pointer = &Lsum[0];

        /* Save the solution */
        for (band = 0; band < num_bands; band++)
        {
            /* Get the band index */
            band_index = ias_sat_attr_convert_band_number_to_index(
                band_numbers[band]);
            if (band_index == ERROR)
            {
                IAS_LOG_ERROR("Invalid band index for band number %d",
                    band_numbers[band]);
                free(L);
                free(A);
                free(A00);
                free(AT00);
                free(ATC00);
                free(Lsum);
                free(N);
                free(Nsum);
                free(AT);
                free(ATC);
                free(work_vector);
                return ERROR;
            }

            for (coeff = 0; coeff < num_coeffs; coeff++)
            {
                along_track_solution[band_index][sca_index][coeff]
                    = -Lsum_pointer[band_offsets[band_index] * X_Y * num_coeffs 
                    + coeff];
                across_track_solution[band_index][sca_index][coeff]
                    = -Lsum_pointer[band_offsets[band_index] * X_Y * num_coeffs 
                    + coeff + num_coeffs];
            }
        }
    }

    free(A);
    free(AT);
    free(ATC);
    free(A00);
    free(AT00);
    free(ATC00);
    free(N);
    free(L);
    free(Nsum);
    free(Lsum);
    free(work_vector);

    /* If printing of residuals was requested, allocate residual pointer */
    if (tie_point_list_flag)
    {
        residuals = (RESIDUALS_PNTS *)malloc(sizeof(RESIDUALS_PNTS) * count);
        if (residuals == NULL)
        {
            IAS_LOG_INFO("Allocating %d records for tie point residual pointer",
                count);
            return ERROR;
        }
    }

    /* Populate the residuals structure for all tie points */
    status = populate_residuals(nsca, number_of_combinations, points, 
        num_points, grid, model, dem, tie_point_list_flag, sca_indexes, 
        num_coeffs, along_track_solution, across_track_solution, nl, ns,
        fit_stats, residuals);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Populating residuals");
        free(residuals);
        return ERROR;
    }

    /* Determine new Legendre LOS coefficients */
    for (band = 0; band < num_bands; band++)
    {
        /* Get the band index */
        band_index = ias_sat_attr_convert_band_number_to_index(
            band_numbers[band]);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Invalid band index for band number %d",
                band_numbers[band]);
            free(residuals);
            return ERROR;
        }

        for (sca_index = 0; sca_index < nsca; sca_index++)
        {
            for (coeff = 0; coeff < num_coeffs; coeff++)
            {
                new_legendre_along[band_index][sca_index][coeff]
                    = model->sensor.bands[band_index].scas[sca_index].
                    sca_coef_x[coeff]
                    + along_track_solution[band_index][sca_index][coeff];
            }

            for (coeff_index = coeff; coeff_index < MAX_LEGENDRE_COEFFS; 
                coeff_index++)
            {
                new_legendre_along[band_index][sca_index][coeff_index]
                    = model->sensor.bands[band_index].scas[sca_index].
                    sca_coef_x[coeff_index];
            }

            for (coeff = 0; coeff < num_coeffs; coeff++)
            {
                new_legendre_across[band_index][sca_index][coeff]
                    = model->sensor.bands[band_index].scas[sca_index].
                    sca_coef_y[coeff] 
                    + across_track_solution[band_index][sca_index][coeff];
            }

            for (coeff_index = coeff; coeff_index < MAX_LEGENDRE_COEFFS; 
                coeff_index++)
            {
                new_legendre_across[band_index][sca_index][coeff_index]
                    = model->sensor.bands[band_index].scas[sca_index].
                    sca_coef_y[coeff_index];
            }
        }
    }

    /* Create the bandcal report file */
    status = create_bandcal_report(l1g, l0r, b2bcal_file, num_coeffs, 
        along_track_solution, across_track_solution, new_legendre_along, 
        new_legendre_across, fit_stats, residuals, count, tie_point_list_flag, 
        work_order_id, bandcombo, nsca);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Creating Band Calibration report");
        free(residuals);
        return ERROR;
    }

    /* Free residuals structure */
    free(residuals);

    /* Save data to structures so it can be written to the database later */
    if (store_in_database_flag)
    {
        status = collect_database_data(along_track_solution,
            across_track_solution, new_legendre_along, new_legendre_across,
            fit_stats, point_count, nsca, num_bands, band_numbers, bandcombo,
            num_band_sca_records, num_fit_records, band_sca_db_save,
            pre_post_fit_db_save);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Collecting data to be saved to database");
            return ERROR;
        }
    }

    return SUCCESS;
}

/***************************************************************************
NAME: collect_database_data

PURPOSE:
Collect data that is to be written to the database into structures that will
make that easier.

RETURN VALUE:
Type     = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed
*******************************************************************************/
static int collect_database_data
(
    double along_track_solution[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* I: Solution offsets in line direction */
    double across_track_solution[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* I: Solution offsets in sample direction */
    double new_legendre_along[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* I: New Legendre coefficients along track
                                        direction */
    double new_legendre_across[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* I: New Legendre coefficients across track
                                        direction */
    const FIT_RESIDUAL_STATS_TYPE fit_stats[IAS_MAX_NSCAS], /* I: Statistics */
    int point_count[IAS_MAX_NBANDS][IAS_MAX_NSCAS], /* I: Number of points used 
                                                          per band/SCA */
    int nsca,                    /* I: Number of SCAs */
    int num_bands,               /* I: Total number of bands in sensor */
    int band_numbers[IAS_MAX_NBANDS], /* I: List of band numbers */
    const BANDCOMBO *bandcombo,  /* I: Band combinations */
    int *num_band_sca_records,   /* O: The number of band/SCA DB records */
    int *num_fit_records,        /* O: The number of pre/post fit DB records */
    struct BAND_SCA_DB_SAVE_INFO* band_sca_db_save,
                                 /* O: Band/SCA-level data to save */
    struct PRE_POST_FIT_DB_SAVE_INFO* pre_post_fit_db_save
                                 /* O: Pre and post fit data to save */
)
{
    int sca_index;         /* SCA index */
    int band_index;        /* Band index */
    int band;              /* Band loop counter */

    *num_band_sca_records = 0;
    *num_fit_records = 0;

    for (sca_index = 0; sca_index < nsca; sca_index++)
    {
        /* Fill pre and post fit level database save structure */
        pre_post_fit_db_save->sensor_chip_assembly_id[sca_index] = sca_index 
            + 1;
        pre_post_fit_db_save->pre_fit_res_alng_trk_mean[sca_index] 
            = fit_stats[sca_index].pre_fit_y.mean;
        pre_post_fit_db_save->pre_fit_res_acrss_trk_mean[sca_index]
            = fit_stats[sca_index].pre_fit_x.mean;
        pre_post_fit_db_save->pre_fit_res_alng_trk_rmse[sca_index] 
            = fit_stats[sca_index].pre_fit_y.rmse;
        pre_post_fit_db_save->pre_fit_res_acrss_trk_rmse[sca_index]
            = fit_stats[sca_index].pre_fit_x.rmse;
        pre_post_fit_db_save->pre_fit_res_alng_trk_std_dev[sca_index] 
            = fit_stats[sca_index].pre_fit_y.stdev;
        pre_post_fit_db_save->pre_fit_res_acrss_trk_std_dev[sca_index] 
            = fit_stats[sca_index].pre_fit_x.stdev;
        pre_post_fit_db_save->pst_fit_res_alng_trk_mean[sca_index] 
            = fit_stats[sca_index].post_fit_y.mean;
        pre_post_fit_db_save->pst_fit_res_acrss_trk_mean[sca_index]
            = fit_stats[sca_index].post_fit_x.mean;
        pre_post_fit_db_save->pst_fit_res_alng_trk_rmse[sca_index] 
            = fit_stats[sca_index].post_fit_y.rmse;
        pre_post_fit_db_save->pst_fit_res_acrss_trk_rmse[sca_index]
            = fit_stats[sca_index].post_fit_x.rmse;
        pre_post_fit_db_save->pst_fit_res_alng_trk_std_dev[sca_index] 
            = fit_stats[sca_index].post_fit_y.stdev;
        pre_post_fit_db_save->pst_fit_res_acrss_trk_std_dev[sca_index] 
            = fit_stats[sca_index].post_fit_x.stdev;
        (*num_fit_records)++;

        /* Fill Band/SCA level database save structure */ 
        for (band = 0; band < num_bands; band++)
        {
            /* Get the band index */
            band_index = ias_sat_attr_convert_band_number_to_index(
                band_numbers[band]);
            if (band_index == ERROR)
            {
                IAS_LOG_ERROR("Invalid band index for band number %d",
                    band_numbers[band]);
                return ERROR;
            }

            /* Skip what we didn't process */
            if (bandcombo->band_sca_not_present[band_index][sca_index] == TRUE)
            {
                continue;
            }

            /* Fill Band/SCA level database save structure with global values */
            band_sca_db_save->band_id[*num_band_sca_records]
                = band_numbers[band];
            band_sca_db_save->sensor_chip_assembly_id[*num_band_sca_records]
                = sca_index + 1;
            band_sca_db_save->nbr_of_pnts_used[*num_band_sca_records]
                = point_count[band_index][sca_index];

            /* Fill Band/SCA level database save structure with the updates to 
               the Legendre coefficients */
            band_sca_db_save->upds_alng_trk_leg_poly_0[*num_band_sca_records]
                = along_track_solution[band_index][sca_index][0];
            band_sca_db_save->upds_alng_trk_leg_poly_1[*num_band_sca_records]
                = along_track_solution[band_index][sca_index][1];
            band_sca_db_save->upds_alng_trk_leg_poly_2[*num_band_sca_records]
                = along_track_solution[band_index][sca_index][2];
            band_sca_db_save->upds_alng_trk_leg_poly_3[*num_band_sca_records]
                = along_track_solution[band_index][sca_index][3];

            band_sca_db_save->upds_acrss_trk_leg_poly_0[*num_band_sca_records]
                = across_track_solution[band_index][sca_index][0];
            band_sca_db_save->upds_acrss_trk_leg_poly_1[*num_band_sca_records]
                = across_track_solution[band_index][sca_index][1];
            band_sca_db_save->upds_acrss_trk_leg_poly_2[*num_band_sca_records]
                = across_track_solution[band_index][sca_index][2];
            band_sca_db_save->upds_acrss_trk_leg_poly_3[*num_band_sca_records]
                = across_track_solution[band_index][sca_index][3];

            /* Fill Band/SCA level database save structure with the new 
               Legendre coefficients */
            band_sca_db_save->new_alng_trk_leg_poly_0[*num_band_sca_records]
                = new_legendre_along[band_index][sca_index][0];
            band_sca_db_save->new_alng_trk_leg_poly_1[*num_band_sca_records]
                = new_legendre_along[band_index][sca_index][1];
            band_sca_db_save->new_alng_trk_leg_poly_2[*num_band_sca_records]
                = new_legendre_along[band_index][sca_index][2];
            band_sca_db_save->new_alng_trk_leg_poly_3[*num_band_sca_records]
                = new_legendre_along[band_index][sca_index][3];

            band_sca_db_save->new_acrss_trk_leg_poly_0[*num_band_sca_records]
                = new_legendre_across[band_index][sca_index][0];
            band_sca_db_save->new_acrss_trk_leg_poly_1[*num_band_sca_records]
                = new_legendre_across[band_index][sca_index][1];
            band_sca_db_save->new_acrss_trk_leg_poly_2[*num_band_sca_records]
                = new_legendre_across[band_index][sca_index][2];
            band_sca_db_save->new_acrss_trk_leg_poly_3[*num_band_sca_records]
                = new_legendre_across[band_index][sca_index][3];

            (*num_band_sca_records)++;
        }
    }

    return SUCCESS;
}

/***************************************************************************
NAME: calculate_statistics

PURPOSE:
Calculate mean, sum, RMSE, and standard deviation 

RETURN VALUE: None
*******************************************************************************/
static void calculate_statistics
(
    STATS_TYPE *stats, /* I/O: Collection of statistics */
    int num_points,   /* I: Number of points used to generate statistics */
    double scale      /* I: Scale used in calculating statistics */
)
{
    stats->mean = stats->sum / num_points;
    stats->stdev = stats->sum_sqr_err / ((double)num_points - 1.0) - stats->sum 
        * stats->sum * scale;
    stats->stdev = sqrt(stats->stdev);
    stats->rmse = sqrt(stats->sum_sqr_err / ((double)num_points));
}

/***************************************************************************
NAME: populate_residuals

PURPOSE:
Populate the residuals structures for each tie point.  Also save post-fit
statistics.

RETURN VALUE:
Type     = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed
*******************************************************************************/
static int populate_residuals
(
    int nsca,                    /* I: Number of SCAs */
    int number_of_combinations,  /* I: Number of band combinations to process */
    const POINT *points,         /* I: Points from B2B char file */
    int num_points,              /* I: Total number of points */
    const IAS_GRID_TYPE *grid,   /* I: Grid with all bands */
    const IAS_LOS_MODEL *model,  /* I: Model */
    const IAS_IMAGE *dem,        /* I: DEM data */
    int tie_point_list_flag,     /* I: Flag for creating tie point report */
    const int *sca_indexes,      /* I: List of SCA indexes to process */
    int num_coeffs,              /* I: Number of coefficients to solve for */
    double along_track_solution[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* I: Solution offsets in line direction */
    double across_track_solution[IAS_MAX_NBANDS][IAS_MAX_NSCAS]
        [IAS_LOS_LEGENDRE_TERMS], /* I: Solution offsets in sample direction */
    int nl[IAS_MAX_NBANDS][IAS_MAX_NSCAS], /* I: Number lines in input data */
    int ns[IAS_MAX_NBANDS][IAS_MAX_NSCAS], /* I: Number samples in input data */
    FIT_RESIDUAL_STATS_TYPE fit_stats[IAS_MAX_NSCAS], /* O: Statistics */
    RESIDUALS_PNTS *residuals     /* O: Tie points offset structures */
)
{
    int sca_counter;       /* Counter for SCA loops */
    int combo_counter;     /* Counter for band combinations */
    int sca_index;         /* SCA index */
    int count;             /* Number of residual points */
    int ref_band_index;    /* Index of band that is being used as reference */
    int search_band_index; /* Index of band that is being used as seach */
    int first_reference_band_number; /* First reference for current combo */
    int first_search_band_number; /* First search for current combination */
    double ref_line;       /* Reference line location */
    double ref_samp;       /* Reference sample location */
    double search_line;    /* Search line location */
    double search_samp;    /* Search sample location */
    int ref_band_lines_per_frame; /* Lines per frame for reference band */
    int search_band_lines_per_frame; /* Lines per frame for search band */
    double los_err_samp;   /* Difference in LOS between reference and search 
                              in sample direction */   
    double los_err_line;   /* Difference in LOS between reference and search 
                              in line direction */    
    double ref_input_samp; /* Sample location in input space for reference 
                              point */ 
    double ref_input_line; /* Line location in input space for reference 
                              point */
    double search_input_samp; /* Sample location in input space for search 
                              point */ 
    double search_input_line; /* Line location in input space for search 
                              point */
    double refer_adj_line; /* Adjusted reference sample location */
    double refer_adj_samp; /* Adjusted reference line location */
    double search_adj_line; /* Adjusted search sample location */
    double search_adj_samp; /* Adjusted search line location */
    double norm_det;       /* Normalized detector within an SCA */
    double scale;          /* Scale used in calculating statistics */

    /* Populate residual structure for each point */
    count = 0;
    for (sca_counter = 0; sca_counter < nsca; sca_counter++)
    {
        sca_index = sca_indexes[sca_counter];

        combo_counter = 0;
        while (combo_counter < number_of_combinations)
        {
            ref_band_index = ias_sat_attr_convert_band_number_to_index(
                points[count].reference_band_number);
            if (ref_band_index == ERROR)
            {
                IAS_LOG_ERROR("Invalid band index for reference band number "
                    "%d", points[count].reference_band_number);
                return ERROR;
            }

            search_band_index = ias_sat_attr_convert_band_number_to_index(
                points[count].search_band_number);
            if (search_band_index == ERROR)
            {
                IAS_LOG_ERROR("Invalid band index for search band number %d", 
                    points[count].search_band_number);
                return ERROR;
            }

            first_reference_band_number = points[count].reference_band_number;
            first_search_band_number = points[count].search_band_number;
            while ((first_reference_band_number ==
                       points[count].reference_band_number)
                   && (first_search_band_number ==
                       points[count].search_band_number)
                   && ((sca_index + 1) == points[count].reference_sca_number)
                   && (count < num_points))
            {
                /* Find location of correlation point in both the reference and 
                   search output spaces */
                ref_line = points[count].reference_line;
                ref_samp = points[count].reference_samp;
                ref_band_lines_per_frame = model->sensor.bands[ref_band_index].
                    sampling_char.lines_per_frame;
                search_band_lines_per_frame = 
                    model->sensor.bands[search_band_index].sampling_char.
                    lines_per_frame;

                ref_line *= ref_band_lines_per_frame 
                    / search_band_lines_per_frame; 
                ref_samp *= ref_band_lines_per_frame 
                    / search_band_lines_per_frame; 

                search_line = points[count].search_line + points[count].dy;
                search_samp = points[count].search_samp + points[count].dx;

                /* Determine LOS errors */
                if (get_los_errors(ref_band_index, search_band_index, 
                        sca_index, ref_line, ref_samp, search_line, 
                        search_samp, dem, 
                        &grid->gridbands[ref_band_index][sca_index],
                        &grid->gridbands[search_band_index][sca_index],
                        model,
                        nl[ref_band_index][sca_index], 
                        ns[ref_band_index][sca_index], 
                        nl[search_band_index][sca_index], 
                        ns[search_band_index][sca_index],
                        num_coeffs,
                        &ref_input_line, 
                        &ref_input_samp,
                        &search_input_line,
                        &search_input_samp,
                        &los_err_line,
                        &los_err_samp) != SUCCESS)
                {
                    IAS_LOG_ERROR("Could not find LOS errors for point %d id "
                        "%d SCA %d reference band index %d search band index "
                        "%d reference line %f reference sample %f search line "
                        "%f search sample %f\n", count, points[count].point_id, 
                        sca_index, ref_band_index, search_band_index, ref_line, 
                        ref_samp, search_line, search_samp);
                    return ERROR;
                }

                /* Compute the normalized detector number */
                norm_det = 2.0 * ref_input_samp /
                    ((double)model->sensor.bands
                        [ref_band_index].scas[sca_index].detectors - 1.0) - 1.0;

                /* Form the Legendre terms for this point */
                refer_adj_line = ias_math_eval_legendre(norm_det,
                    along_track_solution[ref_band_index][sca_index],
                    num_coeffs);
                refer_adj_samp = ias_math_eval_legendre(norm_det,
                    across_track_solution[ref_band_index][sca_index],
                    num_coeffs);

                /* Compute the normalized detector number */
                norm_det = 2.0 * search_input_samp /
                    ((double)model->sensor.bands[search_band_index].
                        scas[sca_index].detectors - 1.0) - 1.0;

                /* Form the Legendre terms for this point */
                search_adj_line = ias_math_eval_legendre(norm_det,
                    along_track_solution[search_band_index][sca_index],
                    num_coeffs);
                search_adj_samp = ias_math_eval_legendre(norm_det,
                    across_track_solution[search_band_index][sca_index],
                    num_coeffs);

                /* Populate tie point residuals for later output */
                if (tie_point_list_flag)
                {
                    residuals[count].point_id = points[count].point_id;
                    residuals[count].sca_number = sca_index + 1;
                    residuals[count].ref_band = ref_band_index;
                    residuals[count].search_band = search_band_index;
                    residuals[count].ref_output_samp = ref_line;
                    residuals[count].ref_output_line = ref_samp;
                    residuals[count].ref_input_samp = ref_input_samp;
                    residuals[count].ref_input_line = ref_input_line;
                    residuals[count].search_input_samp = search_input_samp;
                    residuals[count].search_input_line = search_input_line;
                    residuals[count].dy = points[count].dy;
                    residuals[count].dx = points[count].dx;
                    residuals[count].pre_los_err_line = los_err_line;
                    residuals[count].pre_los_err_samp = los_err_samp;
                }

                /* Calculate updated remaining residuals of LOS errors */
                los_err_line = los_err_line - refer_adj_line + search_adj_line;
                los_err_samp = los_err_samp - refer_adj_samp + search_adj_samp;

                /* Save post-fit input space deltas/offsets */
                if (tie_point_list_flag)
                {
                    residuals[count].post_los_err_line = los_err_line;
                    residuals[count].post_los_err_samp = los_err_samp;
                }

                /* Populate statistics structure */
                fit_stats[sca_index].post_fit_x.sum += los_err_samp;
                fit_stats[sca_index].post_fit_y.sum += los_err_line;
                fit_stats[sca_index].post_fit_x.sum_sqr_err += los_err_samp 
                    * los_err_samp;
                fit_stats[sca_index].post_fit_y.sum_sqr_err += los_err_line  
                    * los_err_line;
                fit_stats[sca_index].points++;
                count++;
            }
            combo_counter++;
        }

        /* Calculate post-fit statistics */
        if (fit_stats[sca_index].points == 1)
        {
            IAS_LOG_INFO("There is only 1 fit statistics point for SCA index "
                "%d, leading to divide by 0", sca_index);
            return ERROR;
        }
        scale = 1.0 / ((double)fit_stats[sca_index].points 
            * ((double)fit_stats[sca_index].points - 1.0));
        calculate_statistics(&fit_stats[sca_index].post_fit_x,
            fit_stats[sca_index].points, scale);
        calculate_statistics(&fit_stats[sca_index].post_fit_y,
            fit_stats[sca_index].points, scale);
    }

    return SUCCESS;
}

/*****************************************************************************
NAME: setup_legendre_coefficients

PURPOSE: To populate the array of coefficients.

RETURN VALUE: None
*****************************************************************************/
static void setup_legendre_coefficients(
    int num_coeffs,               /* I: Number of coefficients to solve for */
    double norm_det,              /* I: Normalized detector value */
    double legr_coeff[MAX_LEGENDRE_COEFFS] /* O: Legendre matrix */
)
{
    int index;

    /* Form the Legendre terms into the A matrix for this point */

    /* Initialize for one coefficient then fix the rest. */
    legr_coeff[0] = 1.0;
    for (index = 1; index < MAX_LEGENDRE_COEFFS; index++)
    {
        legr_coeff[index] = 0.0;
    }

    if (num_coeffs > 1)
        legr_coeff[1] = norm_det;

    if (num_coeffs > 2)
        legr_coeff[2] = (1.5 * norm_det * norm_det - 0.5);

    if (num_coeffs > 3)
        legr_coeff[3] = norm_det * (2.5 * norm_det * norm_det - 1.5);
}

