#!/bin/csh -f

#************************************************************************
# NAME:    run_bandcal
#
# PURPOSE: Run a test of an individual bandcal scenario.  This requires
# several environment variables to be set up by the caller.
#
# RETURNS: 0 if successful, 1 on an error condition
#************************************************************************

# Set up the environment for the test.
set application = "bandcal"
set work_order_id = "${scenario}_I170"
set diff_suffix = "diffs"
set omf_suffix = "omf"
set parameter_file_suffix = "odl"
set test_dir = \
    "$IAS_UNIT_TEST_DATA_DIR/ias_base/gps/${application}/${scenario}"
set bandcal = ".././bandcal"
set omf_source = "${scenario}_${application}.${omf_suffix}"
set omf = "${work_order_id}.${omf_suffix}"
set parameter_file_source_db = \
    "${scenario}_${application}.${parameter_file_suffix}"
set parameter_file_source_no_db = \
    "${scenario}_${application}_no_db.${parameter_file_suffix}"
set parameter_file = "${work_order_id}.${parameter_file_suffix}"
set report_file = "${application}.rpt"
set cpf_group_file = "${application}_parameters_20000701_20000930.cpf"
set report_file_diffs = "${report_file}.${diff_suffix}"
set cpf_group_file_diffs = "${cpf_group_file}.${diff_suffix}" 

echo "Testing $scenario scenario."

# Ensure the test environment variable is set.
if ( ! $?IAS_UNIT_TEST_DATA_DIR ) then
    echo "IAS_UNIT_TEST_DATA_DIR isn't set"
    exit 1
endif

# Check whether database testing has been enabled.  If it has, set the
# parameter file to the database-enabled version, otherwise set the
# parameter file to the database-disabled version.  The database
# enabled/disabled versions of the parameter file differ only in the setting
# of the STORE_CHARACTERIZATION_IN_DB flag...
if ( $?ENABLE_DB_TESTS ) then
    # Sanity check to make sure the database user ID/password are set in the
    # environment
    if ( ! $?IAS_DB_L8_OLITIRS ) then
        echo "IAS_DB_L8_OLITIRS is not currently set"
        exit 1
    endif

    # Set up with database support
    set parameter_file_source = "${parameter_file_source_db}"

else
    # display message indicating database support is disabled
    echo
    echo `basename $0` ":  Database tests are disabled."
    echo

    # Set up without database support
    set parameter_file_source = "${parameter_file_source_no_db}"
endif

# Copy the OMF file and parameter file to the standard filename.
cp ${omf_source} ${omf} 
cp ${parameter_file_source} ${parameter_file} 

# Update the working copy of the selected OMF file so that cases where the
# IAS_UNIT_TEST_DATA_DIR string is present are replaced with the actual
# unit test directory, as defined in the environment variable of that name.

# Parse the environment variable so it is in a backslashed form that sed can
# handle.  In other words, replace all / with \/.  Otherwise the slashes in
# the filename path would be interpreted by sed as instructions rather than
#data.
set modified_directory=`echo $IAS_UNIT_TEST_DATA_DIR | sed 's@\/@\\\/@g'`

# Pass the OMF and Work order parameter file through sed.  Sed will replace 
# all cases of the string IAS_UNIT_TEST_DATA_DIR with the (now sed-digestible) 
# contents of that environment variable.
cat ${omf} | sed "s/IAS_UNIT_TEST_DATA_DIR/$modified_directory/g" > ${omf}_tmp
mv ${omf}_tmp ${omf}

cat ${parameter_file} | sed "s/IAS_UNIT_TEST_DATA_DIR/$modified_directory/g" \
    > ${parameter_file}_tmp
mv ${parameter_file}_tmp ${parameter_file}

# Run the Band Calibration algorithm using the input parameter file.  This will
# create a report file, an ASCII CPF group file, and database updates.
echo "Testing Band Calibration."
${bandcal} ${parameter_file}
if ( $status != 0 ) then
    echo "Failed to run $bandcal."
    exit 1
endif

# Compare the CPF group file to the unit test version.  Any difference will be
# treated as a failure.
echo "Comparing CPF group file $cpf_group_file to $test_dir/$cpf_group_file."
diff ${test_dir}/${cpf_group_file} ${cpf_group_file} > ${cpf_group_file_diffs}
if ( $status != 0 ) then
    echo "$cpf_group_file and $test_dir/$cpf_group_file files do not match."
    exit 1
endif

# Prepare to compare the Bandcal report file to the unit test version.  Exclude
# the report header contents, since that is tested with its own regression test
sed -e '/Number of Legendre coefficients/,$!'d ${report_file} \
    > ${report_file}_tmp
cp ${report_file} ${report_file}_sav
mv ${report_file}_tmp ${report_file}

# Compare the report file to the unit test version.  Any difference will be
# treated as a failure.
echo "Comparing report file $report_file to $test_dir/$report_file."
diff ${test_dir}/${report_file} ${report_file} > ${report_file_diffs}
if ( $status != 0 ) then
    echo "$report_file and $test_dir/$report_file files do not match."
        
    exit 1
endif

# Restore the original report file so we don't have an extra file on disk.
mv ${report_file}_sav ${report_file}

# If database testing is disabled, exit here with a success status.
if ( ! $?ENABLE_DB_TESTS ) then
    exit 0
endif

# Finish setting up the database information
set db_results = "_db_results.txt"
set main_db_table = "band_calibration"
set band_sca_db_table = "band_calibration_band_sca"
set prepst_fit_db_table = "band_cal_prepst_fit_detail"
set main_db_table_file = "${main_db_table}${db_results}"
set band_sca_db_table_file = "${band_sca_db_table}${db_results}"
set prepst_fit_db_table_file = "${prepst_fit_db_table}${db_results}"
set main_db_table_file_diffs = "${main_db_table_file}.${diff_suffix}"
set band_sca_db_table_file_diffs = "${band_sca_db_table_file}.${diff_suffix}"
set prepst_fit_db_table_file_diffs = \
    "${prepst_fit_db_table_file}.${diff_suffix}"
set main_table_sql_script = "main_db_table_script.sql"
set band_sca_table_sql_script = "band_sca_db_table_script.sql"
set prepst_fit_table_sql_script = "prepst_fit_db_table_script.sql"
set common_id_sql_script = "geometric_wo_common_id_script.sql"
set common_id_file = "geometric_wo_common_id.txt"

# Get the geometric common ID.#
echo "set term off" > $common_id_sql_script
echo "set echo on" >> $common_id_sql_script
echo "set trimout on" >> $common_id_sql_script
echo "set trimspool on" >> $common_id_sql_script
echo "set heading off" >> $common_id_sql_script
echo "set feedback off" >> $common_id_sql_script
echo "set pause off" >> $common_id_sql_script
echo "set pagesize 0" >> $common_id_sql_script
echo "set linesize 500" >> $common_id_sql_script
echo "set space 0" >> $common_id_sql_script
echo "set colsep |" >> $common_id_sql_script
echo "set tab off" >> $common_id_sql_script
echo "set verify off" >> $common_id_sql_script
echo "spool ./${common_id_file}" >> $common_id_sql_script
echo "select work_order_id, geometric_wo_common_id " >> $common_id_sql_script
echo "from geometric_work_order_common " >> $common_id_sql_script
echo "where work_order_id='"$work_order_id"';" >> $common_id_sql_script
echo "spool off" >> $common_id_sql_script
echo "exit" >> $common_id_sql_script

# Query the database.
sqlplus $IAS_DB_L8_OLITIRS @$common_id_sql_script > & /dev/null

# Squeeze out the white space and select the lines beginning with the work
# order ID.
set common_id = `sed 's/  *//g' ./${common_id_file} | grep ^$work_order_id | awk -F'|' '{print $2}'`

# Create a script to query the main band calibration database table.  This is
# done here so that a separate script will not have to be maintained.  It will
# be used then removed at the end.
echo "set term off" > $main_table_sql_script
echo "set echo on" >> $main_table_sql_script
echo "set trimout on" >> $main_table_sql_script
echo "set trimspool on" >> $main_table_sql_script
echo "set heading off" >> $main_table_sql_script
echo "set feedback off" >> $main_table_sql_script
echo "set pause off" >> $main_table_sql_script
echo "set pagesize 0" >> $main_table_sql_script
echo "set linesize 500" >> $main_table_sql_script
echo "set space 0" >> $main_table_sql_script
echo "set colsep |" >> $main_table_sql_script
echo "set tab off" >> $main_table_sql_script
echo "set verify off" >> $main_table_sql_script
echo "spool ./${main_db_table_file}" >> $main_table_sql_script
echo "select nbr_of_legendre_coefficients " >> $main_table_sql_script
echo "from ${main_db_table} " >> $main_table_sql_script
echo "where geometric_wo_common_id=$common_id;"  >> $main_table_sql_script
echo "spool off" >> $main_table_sql_script
echo "exit" >> $main_table_sql_script

# Query the database.
sqlplus $IAS_DB_L8_OLITIRS @$main_table_sql_script > & /dev/null

# Find the line with a single number, the number of Legendre coefficients.
grep '^[[:space:]]*[0-9][[:space:]]*$' ./${main_db_table_file} > \
    ./${main_db_table_file}_tmp
cp ${main_db_table_file}_tmp ${main_db_table_file}

# Compare the results.
diff ${test_dir}/${main_db_table_file} . > ${main_db_table_file_diffs}
if ( $status != 0 ) then
    echo "Database results in file ${main_db_table_file} differ from" \
        "expected results"
    exit 1
endif

# Create a script to query the band/SCA level band calibration database table. 
# This is done here so that a separate script will not have to be maintained.  
# It will be used then removed at the end.
echo "set term off" > $band_sca_table_sql_script
echo "set echo on" >> $band_sca_table_sql_script
echo "set trimout on" >> $band_sca_table_sql_script
echo "set trimspool on" >> $band_sca_table_sql_script
echo "set heading off" >> $band_sca_table_sql_script
echo "set feedback off" >> $band_sca_table_sql_script
echo "set pause off" >> $band_sca_table_sql_script
echo "set pagesize 0" >> $band_sca_table_sql_script
echo "set linesize 500" >> $band_sca_table_sql_script
echo "set space 0" >> $band_sca_table_sql_script
echo "set colsep |" >> $band_sca_table_sql_script
echo "set tab off" >> $band_sca_table_sql_script
echo "set verify off" >> $band_sca_table_sql_script
echo "spool ./${band_sca_db_table_file}" >> $band_sca_table_sql_script
echo "select band_id, sensor_chip_assembly_id, nbr_of_pnts_used, " >> \
    $band_sca_table_sql_script
echo "upds_alng_trk_leg_poly_0, upds_alng_trk_leg_poly_1, " >> \
    $band_sca_table_sql_script
echo "upds_alng_trk_leg_poly_2, upds_acrss_trk_leg_poly_0, " >> \
    $band_sca_table_sql_script
echo "upds_acrss_trk_leg_poly_1, upds_acrss_trk_leg_poly_2, " >> \
    $band_sca_table_sql_script
echo "new_alng_trk_leg_poly_0, new_alng_trk_leg_poly_1, " >> \
    $band_sca_table_sql_script
echo "new_alng_trk_leg_poly_2, new_acrss_trk_leg_poly_0, " >> \
    $band_sca_table_sql_script
echo "new_acrss_trk_leg_poly_1, new_acrss_trk_leg_poly_2 " >> \
    $band_sca_table_sql_script
echo "from ${band_sca_db_table} " >> $band_sca_table_sql_script
echo "where geometric_wo_common_id=$common_id "  >> \
    $band_sca_table_sql_script
echo "order by band_id, sensor_chip_assembly_id; " >> \
    $band_sca_table_sql_script
echo "spool off" >> $band_sca_table_sql_script
echo "exit" >> $band_sca_table_sql_script

# Query the database.
sqlplus $IAS_DB_L8_OLITIRS @$band_sca_table_sql_script > & /dev/null

# Squeeze out the white space and select the line with pipes.
sed 's/  *//g' ./${band_sca_db_table_file} | grep '|' > \
    ./${band_sca_db_table_file}_tmp
sed 's/|/ /g' ./${band_sca_db_table_file}_tmp > ./${band_sca_db_table_file}

# Compare the results.
diff ${test_dir}/${band_sca_db_table_file} . > ${band_sca_db_table_file_diffs}
if ( $status != 0 ) then
    echo "Database results in file ${band_sca_db_table_file} differ from" \
        "expected results"
    exit 1
endif

# Create a script to query the pre and post fit band calibration database 
# table.  This is done here so that a separate script will not have to be 
# maintained.  It will be used then removed at the end.
echo "set term off" > $prepst_fit_table_sql_script
echo "set echo on" >> $prepst_fit_table_sql_script
echo "set trimout on" >> $prepst_fit_table_sql_script
echo "set trimspool on" >> $prepst_fit_table_sql_script
echo "set heading off" >> $prepst_fit_table_sql_script
echo "set feedback off" >> $prepst_fit_table_sql_script
echo "set pause off" >> $prepst_fit_table_sql_script
echo "set pagesize 0" >> $prepst_fit_table_sql_script
echo "set linesize 500" >> $prepst_fit_table_sql_script
echo "set space 0" >> $prepst_fit_table_sql_script
echo "set colsep |" >> $prepst_fit_table_sql_script
echo "set tab off" >> $prepst_fit_table_sql_script
echo "set verify off" >> $prepst_fit_table_sql_script
echo "spool ./${prepst_fit_db_table_file}" >> $prepst_fit_table_sql_script
echo "select sensor_chip_assembly_id, pre_fit_res_acrss_trk_mean, " >> \
    $prepst_fit_table_sql_script
echo "pre_fit_res_acrss_trk_rmse, pre_fit_res_acrss_trk_std_dev," >> \
    $prepst_fit_table_sql_script
echo "pre_fit_res_alng_trk_mean, pre_fit_res_alng_trk_rmse," >> \
    $prepst_fit_table_sql_script
echo "pre_fit_res_alng_trk_std_dev, pst_fit_res_acrss_trk_mean," >> \
    $prepst_fit_table_sql_script
echo "pst_fit_res_acrss_trk_rmse, pst_fit_res_acrss_trk_std_dev," >> \
    $prepst_fit_table_sql_script
echo "pst_fit_res_alng_trk_mean, pst_fit_res_alng_trk_rmse," >> \
    $prepst_fit_table_sql_script
echo "pst_fit_res_alng_trk_std_dev " >> $prepst_fit_table_sql_script
echo "from ${prepst_fit_db_table} " >> $prepst_fit_table_sql_script
echo "where geometric_wo_common_id=$common_id "  >> \
    $prepst_fit_table_sql_script
echo "order by sensor_chip_assembly_id; " >> $prepst_fit_table_sql_script
echo "spool off" >> $prepst_fit_table_sql_script
echo "exit" >> $prepst_fit_table_sql_script

# Query the database.
sqlplus $IAS_DB_L8_OLITIRS @$prepst_fit_table_sql_script > & /dev/null

# Squeeze out the white space and select the line with pipes.
sed 's/  *//g' ./${prepst_fit_db_table_file} | \
    grep '|' > ./${prepst_fit_db_table_file}_tmp
sed 's/|/ /g' ./${prepst_fit_db_table_file}_tmp > \
    ./${prepst_fit_db_table_file}

# Compare the results.
diff ${test_dir}/${prepst_fit_db_table_file} . > \
    ${prepst_fit_db_table_file_diffs}
if ( $status != 0 ) then
    echo "Database results in file ${prepst_fit_db_table_file} " \
        "differ from expected results"
    exit 1
endif

# Clean up some files.
rm ./${main_db_table_file} ./${main_db_table_file}_tmp \
    ./${band_sca_db_table_file} ./${band_sca_db_table_file}_tmp \
    ./${prepst_fit_db_table_file} ./${prepst_fit_db_table_file}_tmp \
    ./${common_id_file} ./${main_table_sql_script} \
    ./${band_sca_table_sql_script} ./${prepst_fit_table_sql_script} \
    ./${common_id_sql_script}

# Exit with success.
exit 0
