/******************************************************************************
NAME:           write_sca_parameters_cpf

PURPOSE:
write_sca_parameters_cpf writes a file containing the updates to the 
FOCAL_PLANE group of the CPF.

RETURNS:
Type = int
Value           Description
-----           -----------
ERROR           Error writing the CPF group 
SUCCESS         CPF successfully written

******************************************************************************/

#include <stdio.h>           /* File I/O prototypes */
#include <stdlib.h>          /* Standard C library */
#include <string.h>

#include "ias_const.h"
#include "ias_logging.h"     /* IAS logging library */
#include "ias_satellite_attributes.h" /* IAS satellite attributes library */

#include "bandcal.h"

int write_sca_parameters_cpf
(
    const char *cpf_group_filename, /* I: CPF name for writing the group */
    double x_coeff_corr[IAS_MAX_NBANDS][IAS_MAX_NSCAS][IAS_LOS_LEGENDRE_TERMS],
                             /* I: X Legendre coefficient corrections */
    double y_coeff_corr[IAS_MAX_NBANDS][IAS_MAX_NSCAS][IAS_LOS_LEGENDRE_TERMS],
                             /* I: Y Legendre coefficient corrections */
    const BANDCOMBO *bandcombo /* I: Band combinations */
)
{
    FILE *file_ptr;          /* CPF file to write the group to */
    int status;              /* Status returned by called routines */
    int sca_number;          /* SCA number */
    int sca_index;           /* SCA index corresponding to SCA number */
    int band_index = 0;      /* Band index corresponding to band number */
    int number_scas_for_band;/* Number of SCAs for the current band */
    int band_numbers[IAS_MAX_NBANDS]; /* List of band numbers */
    int num_normal_bands;    /* Total normal bands of satellite */
    int num_bands;           /* Total number of bands in sensor */
    int band;                /* Loop counter */
    char groupname[32];      /* groupname for the output */
    const char *sensor_id_string; /* The name of the sensor */

    /* Get total satellite band count */
    num_normal_bands = ias_sat_attr_get_normal_band_count();
    if (num_normal_bands == ERROR)
    {
        IAS_LOG_ERROR("Getting normal band count using satellite attributes "
            "library");
        return ERROR;
    }

    /* Get band info */
    status = ias_sat_attr_get_sensor_band_numbers(bandcombo->sensor_id,
        IAS_NORMAL_BAND, 0, band_numbers, IAS_MAX_NBANDS, &num_bands);
    if (status != SUCCESS)
    {
        sensor_id_string = ias_sat_attr_get_sensor_name(bandcombo->sensor_id);
        IAS_LOG_ERROR("Getting normal band numbers for the %s sensor",
            sensor_id_string);
        return ERROR;
    }

    switch(bandcombo->sensor_id)
    {
        case IAS_OLI:
            status = snprintf(groupname, sizeof(groupname), "OLI_FOCAL_PLANE");
            break;

        case IAS_TIRS:
            status = snprintf(groupname, sizeof(groupname), "TIRS_FOCAL_PLANE");
            break;

        default:
            /* shouldn't get here, but just-in-case */
            IAS_LOG_ERROR("Invalid sensor for processing bandcal %d",
                          bandcombo->sensor_id);
            return ERROR;
            break;
    }
    if ((status < 0) || (status >= sizeof(groupname)))
    {
        IAS_LOG_ERROR("groupname buffer is too small");
        return ERROR;
    }

    /* Open the CPF group ASCII file */
    file_ptr = fopen(cpf_group_filename, "wt");
    if (file_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening the CPF group output file - %s", 
            cpf_group_filename);
        return ERROR;
     }

    /* Write the group name */
    status = fprintf(file_ptr, "GROUP = %s\n", groupname);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing to the CPF group ASCII file %s", 
            cpf_group_filename);
        fclose(file_ptr);
        return ERROR;
    }    

    /* Write the updated along-track (X) Legendre coefficients */
    for (band = 0; band < num_bands; band++)
    {
        /* Get the band index */
        band_index = ias_sat_attr_convert_band_number_to_index(
            band_numbers[band]);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Invalid band index for band number %d", 
                band_numbers[band]);
            fclose(file_ptr);
            return ERROR;
        }

        /* Find the number of SCAs for the band */
        number_scas_for_band = ias_sat_attr_get_scas_per_band(
            band_numbers[band]);
        if (number_scas_for_band == ERROR)
        {
            IAS_LOG_ERROR("Invalid number of SCAs retrieved for band %d", 
                band_numbers[band]);
            return ERROR;
        }

        for (sca_index = 0; sca_index < number_scas_for_band; sca_index++)
        {
            sca_number = sca_index + 1;

            /* Skip what we didn't process */
            if (bandcombo->band_sca_not_present[band_index][sca_index] == TRUE)
            {
                continue;
            }

            status = fprintf(file_ptr, " Along_LOS_Legendre_B%02d_SCA%02d = "
                "(%E, %E,\n  %E, %E)\n",
                band_numbers[band], sca_number,
                    x_coeff_corr[band_index][sca_index][0],
                    x_coeff_corr[band_index][sca_index][1],
                    x_coeff_corr[band_index][sca_index][2],
                    x_coeff_corr[band_index][sca_index][3]);
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing to the CPF Group ASCII file %s", 
                    cpf_group_filename);
                fclose(file_ptr);
                return ERROR;
            }
        }
    }

    /* Write the updated across-track (Y) Legendre coefficients */
    for (band = 0; band < num_bands; band++)
    {
        /* Get the band index */
        band_index = ias_sat_attr_convert_band_number_to_index(
            band_numbers[band]);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Invalid band index for band number %d", 
                band_numbers[band]);
            return ERROR;
        }

        /* Find the number of SCAs for the band */
        number_scas_for_band = ias_sat_attr_get_scas_per_band(
            band_numbers[band]);
        if (number_scas_for_band == ERROR)
        {
            IAS_LOG_ERROR("Invalid number of SCAs retrieved for band %d", 
                band_numbers[band]);
            return ERROR;
        }

        for (sca_index = 0; sca_index < number_scas_for_band; sca_index++)
        {
            sca_number = sca_index + 1;

            /* Skip what we didn't process */
            if (bandcombo->band_sca_not_present[band_index][sca_index] == TRUE)
            {
                continue;
            }

            status = fprintf(file_ptr, " Across_LOS_Legendre_B%02d_SCA%02d = "
                "(%E, %E,\n  %E, %E)\n",
                band_numbers[band], sca_number,
                    y_coeff_corr[band_index][sca_index][0],
                    y_coeff_corr[band_index][sca_index][1],
                    y_coeff_corr[band_index][sca_index][2],
                    y_coeff_corr[band_index][sca_index][3]);
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing to the CPF Group ASCII file %s", 
                    cpf_group_filename);
                fclose(file_ptr);
                return ERROR;
            }
        }
    }

    /* Write the end of group tag */
    status = fprintf(file_ptr, "END_GROUP = %s\n", groupname);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing to the CPF Group ASCII file %s", 
            cpf_group_filename);
        fclose(file_ptr);
        return ERROR;
    }    
    
    /* Close the CPF group report file */
    if (fclose(file_ptr) != 0)
    {
        IAS_LOG_ERROR("Closing the CPF group report file %s", 
            cpf_group_filename);
        return ERROR;
    }
      
    return SUCCESS;
}

