/*----------------------------------------------------------------------------
   NAME:    correct_los_model_test.c

PURPOSE: Exercise various regression tests to ensure application is working
         as expected
  
  Contains the routines:
        
        test_read_parameters
        test_read_default_parameters
        test_precision_params_A: using band 8 SCA-combined J. Storey test data
        test_precision_params_B: using band 8 SCA-separated J. Storey test data
        test_precision_params_C: using band 5 SCA-combined M. Choate data
        test_precision_params_D: using band 5 SCA-separated M. Choate data
        test_database_store: using band 8 SCA-combined M. Choate data

Usage:  ./correct_los_model_test [test to run]

    If no test is provided on the command line all tests are run.  Providing
    a test name makes debugging easier.
-----------------------------------------------------------------------------*/
#include <string.h>
#include "cgreen/cgreen.h"
#include "ias_logging.h"
#include "ias_parm.h"
#include "ias_cpf.h"
#include "../correct_los_model.h"

static char *unit_test_dir = NULL;
static char test_directory[] = "ias_base/gps/correct_los_model";

Ensure test_read_default_parameters(void);
Ensure test_read_cpf_parameters(void);
Ensure test_precision_params_A(void);
Ensure test_precision_params_B(void);
Ensure test_precision_params_C(void);
Ensure test_precision_params_D(void);
Ensure test_database_store(void);
Ensure test_insufficient_gcps(void);

int main(int argc, char *argv[])
{
    char *enable_db_tests = NULL;
    int test_database = 0;

    /* See if the database tests should be enabled */
    enable_db_tests = getenv( "ENABLE_DB_TESTS" );
    if ( enable_db_tests )
    {
        test_database = 1;
    }
    else
    {
        test_database = 0;
        printf("\n");
        printf("%s:  Database tests are disabled.\n", __FILE__);
        printf("\n");
    }

    TestSuite *suite = create_test_suite();
    add_test(suite, test_read_default_parameters);
    add_test(suite, test_read_cpf_parameters);
    add_test(suite, test_precision_params_A);
    add_test(suite, test_precision_params_B);
    add_test(suite, test_precision_params_C);
    add_test(suite, test_precision_params_D);
    if (test_database)
        add_test(suite, test_database_store);
    add_test(suite, test_insufficient_gcps);

    unit_test_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (unit_test_dir == NULL)
    {
        printf("The IAS_UNIT_TEST_DATA_DIR environment variable is not set\n");
        exit(EXIT_FAILURE);
    }

    if (argc > 1)
        return run_single_test(suite, argv[1], create_text_reporter());

    return run_test_suite(suite, create_text_reporter());
}

Ensure test_read_default_parameters(void)
{
    int status;
    char command[PATH_MAX];
    char cpf_filename[PATH_MAX];
    PRECISION_CORRECTIONS precision_corr;
    double obs_sigma_apri[IAS_CPF_LOS_OBSERVATION_APRI_COUNT];
    
    IAS_CPF *cpf = NULL;

    printf("Test read default parameters.......\n");

    /* Remove existing parameter files */
    status = system("rm -f default_parameters.odl DL0831.omf");
    if (status)
    {
        assert_true_with_message(0, "Error removing default parameter files");
        return;
    }

    /* Attempt to read the parameters -- this should fail since the file
       isn't here.  Disable the logging until after the next error check */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = read_parameters("default_parameters.odl", &cpf);
    assert_not_equal_with_message(status, SUCCESS,
            "read_parameters should have returned ERROR");

    /* Copy the default parameter file from the test directory */
    sprintf(command, "cp %s/%s/parameter_test/default_parameters.odl .",
            unit_test_dir, test_directory);
    status = system(command);
    if (status)
    {
        assert_true_with_message(0, "Error copying default_parameters.odl "
                "from IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* Read the parameters -- this should fail since the OMF isn't here */
    status = read_parameters("default_parameters.odl", &cpf);
    assert_not_equal_with_message(status, SUCCESS,
            "read_parameters should have returned ERROR");

    /* Re-enable the logging */
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* prepare the test omf file */
    status = snprintf(command, PATH_MAX,
            "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' DL0831_template.omf > "
            "DL0831.omf", unit_test_dir);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Error preparing parameter test OMF file "
                "for unit test data directory path setup");
        return;
    }

    /* Read the parameters -- this should pass */
    status = read_parameters("default_parameters.odl", &cpf);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from read_parameters");
        return;
    }
    
    /* build cpf_filename and path for the getter test. */
    status = snprintf(cpf_filename, PATH_MAX, 
                    "%s/%s/parameter_test/parameters_tests.cpf",
                    unit_test_dir, test_directory);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error creating path to cpf file");
        return;
    }

    /* Test the getter routines */
    assert_true(get_store_in_database_flag());
    assert_string_equal(get_l1g_filename(), "l1g_filename");
    assert_string_equal(get_grid_filename(), "grid_filename");
    assert_string_equal(get_syst_model_filename(), "systematic_model");
    assert_string_equal(get_gcp_data_filename(), "measured_gcp_filename");
    assert_string_equal(get_solution_filename(), "out_solution");
    assert_string_equal(get_residual_filename(), "resudual_filename");
    assert_string_equal(get_cpf_filename(), cpf_filename);
    assert_double_equal(get_outlier_threshold(), 0.98);
    assert_equal(get_max_iterations(), 10);
    assert_true(get_time_rate_estimation_flag());
    assert_equal(get_parameter_option(), 2);
    assert_string_equal(get_dem_filename(), "dem_filename_pass_1");
    assert_true(get_terrain_flag());
    assert_string_equal(get_prec_model_filename(), "precision_model");
    set_terrain_flag(0);
    assert_false(get_terrain_flag());
    assert_string_equal(get_l0r_filename(), "l0r_filename");
    assert_string_equal(get_omf_filename(), "DL0831.omf");
    assert_string_equal(get_work_order_id(), "DL0831");
    assert_equal(get_processing_pass(), 1);

    /* default version */
    get_precision_corr_information (&precision_corr);
    assert_double_equal(precision_corr.apri_att_mean[0], 0.0);
    assert_double_equal(precision_corr.apri_att_sigma[0], 200.0);
    assert_double_equal(precision_corr.apri_eph_mean[0], 0.0);
    assert_double_equal(precision_corr.apri_eph_sigma[0], 200.0);
    assert_double_equal(precision_corr.apri_att_mean[1], 0.0);
    assert_double_equal(precision_corr.apri_att_sigma[1], 200.0);
    assert_double_equal(precision_corr.apri_eph_mean[1], 0.0);
    assert_double_equal(precision_corr.apri_eph_sigma[1], 200.0);
    assert_double_equal(precision_corr.apri_att_mean[2], 0.0);
    assert_double_equal(precision_corr.apri_att_sigma[2], 200.0);
    assert_double_equal(precision_corr.apri_eph_mean[2], 0.0);
    assert_double_equal(precision_corr.apri_eph_sigma[2], 200.0);
    assert_double_equal(precision_corr.apri_att_mean[3], 0.0);
    assert_double_equal(precision_corr.apri_att_sigma[3], 5.0);
    assert_double_equal(precision_corr.apri_eph_mean[3], 0.0);
    assert_double_equal(precision_corr.apri_eph_sigma[3], 5.0);
    assert_double_equal(precision_corr.apri_att_mean[4], 0.0);
    assert_double_equal(precision_corr.apri_att_sigma[4], 5.0);
    assert_double_equal(precision_corr.apri_eph_mean[4], 0.0);
    assert_double_equal(precision_corr.apri_eph_sigma[4], 5.0);
    assert_double_equal(precision_corr.apri_att_mean[5], 0.0);
    assert_double_equal(precision_corr.apri_att_sigma[5], 5.0);
    assert_double_equal(precision_corr.apri_eph_mean[5], 0.0);
    assert_double_equal(precision_corr.apri_eph_sigma[5], 5.0);

    /* default version */
    get_doq_observation_sigma_apri(obs_sigma_apri);
    assert_double_equal(obs_sigma_apri[0], 5.0);
    assert_double_equal(obs_sigma_apri[1], 5.0);

    remove("default_parameters.odl");
    remove("DL0831.omf");
    
    ias_cpf_free(cpf);

}  /* END test_read_parameters */

/* test the cpf parameters here */

Ensure test_read_cpf_parameters(void)
{
    int status;
    char command[PATH_MAX];
    PRECISION_CORRECTIONS precision_corr;
    double obs_sigma_apri[IAS_CPF_LOS_OBSERVATION_APRI_COUNT];

    IAS_CPF *cpf = NULL;

    printf("Test read parameters from CPF.......\n");

    /* Remove existing parameter files */
    status = system("rm -f cpf_parameters.odl DL0831.omf");
    if (status)
    {
        assert_true_with_message(0, "Error removing parameter files");
        return;
    }

    /* Copy the default parameter file from the test directory */
    sprintf(command, "cp %s/%s/parameter_test/cpf_parameters.odl .",
            unit_test_dir, test_directory);
    status = system(command);
    if (status)
    {
        assert_true_with_message(0, "Error copying cpf_parameters.odl "
                "from IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* prepare the test omf file */
    status = snprintf(command, PATH_MAX,
            "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' DL0831_template.omf > "
            "DL0831.omf", unit_test_dir);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Error preparing parameter test OMF file "
                "for unit test data directory path setup");
        return;
    }

    /* Read the parameters -- this should pass */
    status = read_parameters("cpf_parameters.odl", &cpf);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from CPF read_parameters");
        return;
    }
    
    /* cpf version */
    get_precision_corr_information (&precision_corr);
    assert_double_equal(precision_corr.apri_att_mean[0], 0.0);
    assert_double_equal(precision_corr.apri_att_sigma[0], 50.0);
    assert_double_equal(precision_corr.apri_eph_mean[0], 0.0);
    assert_double_equal(precision_corr.apri_eph_sigma[0], 8.0);
    assert_double_equal(precision_corr.apri_att_mean[1], 0.0);
    assert_double_equal(precision_corr.apri_att_sigma[1], 50.0);
    assert_double_equal(precision_corr.apri_eph_mean[1], 0.0);
    assert_double_equal(precision_corr.apri_eph_sigma[1], 8.0);
    assert_double_equal(precision_corr.apri_att_mean[2], 0.0);
    assert_double_equal(precision_corr.apri_att_sigma[2], 50.0);
    assert_double_equal(precision_corr.apri_eph_mean[2], 0.0);
    assert_double_equal(precision_corr.apri_eph_sigma[2], 8.0);
    assert_double_equal(precision_corr.apri_att_mean[3], 0.0);
    assert_double_equal(precision_corr.apri_att_sigma[3], 1.0);
    assert_double_equal(precision_corr.apri_eph_mean[3], 0.0);
    assert_double_equal(precision_corr.apri_eph_sigma[3], 0.5);
    assert_double_equal(precision_corr.apri_att_mean[4], 0.0);
    assert_double_equal(precision_corr.apri_att_sigma[4], 1.0);
    assert_double_equal(precision_corr.apri_eph_mean[4], 0.0);
    assert_double_equal(precision_corr.apri_eph_sigma[4], 0.5);
    assert_double_equal(precision_corr.apri_att_mean[5], 0.0);
    assert_double_equal(precision_corr.apri_att_sigma[5], 1.0);
    assert_double_equal(precision_corr.apri_eph_mean[5], 0.0);
    assert_double_equal(precision_corr.apri_eph_sigma[5], 0.5);

    get_doq_observation_sigma_apri(obs_sigma_apri);
    assert_double_equal(obs_sigma_apri[0], 5.1);
    assert_double_equal(obs_sigma_apri[1], 5.2);

    get_gls_observation_sigma_apri(obs_sigma_apri);
    assert_double_equal(obs_sigma_apri[0], 10.1);
    assert_double_equal(obs_sigma_apri[1], 10.2);

    remove("cpf_parameters.odl");
    remove("DL0831.omf");

    ias_cpf_free(cpf);
} /* END test the cpf parameters here */

Ensure test_precision_params_A(void)
{
    /* Using band 8 SCA-combined J. Storey test data */
    char command[PATH_MAX];
    int status;
    char param_filename[] = "parameters_band_8_storey_combined.odl";
    char prec_parm_filename[] = "precision_params_band_8_storey_combined.txt";
    char new_prec_parm_filename[]
            = "new_precision_params_band_8_storey_combined.txt";
    char our_residuals_filename[]
            = "los_model_band_8_storey_combined.residuals";
    char calval_residuals_filename[]
            = "calval_originals/precision.residuals";
    char our_solution_filename[] = "los_model_band_8_storey_combined.solution";
    char calval_solution_filename[] = "calval_originals/precision.solution";

    printf("Test precision params A.....\n");

    status = snprintf(command, PATH_MAX,
            "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' WOband8C_JS_template.omf > "
            "WOband8C_JS.omf", unit_test_dir);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Error preparing OMF file for unit "
                "test data directory path setup");
        return;
    }

    /* Run correct_los_model; capture precision parameters from DEBUG output */
    setenv("IAS_LOG_LEVEL", "DEBUG", 1);
    status = snprintf(command, PATH_MAX,
            "../correct_los_model %s | grep PRECISION_PARMS_TAG > %s",
            param_filename, new_prec_parm_filename);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Correct LOS Model failed");
        return;
    }

    /* Compare this test run precision parameters to saved parameters */
    status = snprintf(command, PATH_MAX, "diff %s %s", prec_parm_filename,
            new_prec_parm_filename);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found in precision "
                "parameters. Review %s and %s for details", prec_parm_filename,
                new_prec_parm_filename);
        return;
    }

    /* Look at the report files and make sure they match the expected
       results. The report files have header information and this has to be
       removed for the diff */
    status = snprintf(command, PATH_MAX, "sed '/Iteration 0/,$!'d "
            "%s | diff -q - %s/%s/band_8_storey_combined/%s",
            our_residuals_filename, unit_test_dir, test_directory,
            calval_residuals_filename);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found in precision "
                "residuals file %s", our_residuals_filename);
    }

    status = snprintf(command, PATH_MAX, "sed '/Reference Time/,$!'d "
            "%s | diff -q - %s/%s/band_8_storey_combined/%s",
            our_solution_filename, unit_test_dir, test_directory,
            calval_solution_filename);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found in precision "
                "residuals file %s", our_solution_filename);
    }
}  /* END test_precision_params_A */

Ensure test_precision_params_B(void)
{
    /* Using band 8 SCA-separated J. Storey test data */
    char command[PATH_MAX];
    int status;
    char param_filename[] = "parameters_band_8_storey_separated.odl";
    char prec_parm_filename[] = "precision_params_band_8_storey_separated.txt";
    char new_prec_parm_filename[]
            = "new_precision_params_band_8_storey_separated.txt";
    char our_residuals_filename[]
            = "los_model_band_8_storey_separated.residuals";
    char calval_residuals_filename[]
            = "calval_originals/precision.residuals";
    char our_solution_filename[] = "los_model_band_8_storey_separated.solution";
    char calval_solution_filename[] = "calval_originals/precision.solution";

    printf("Test precision params B......\n");

    status = snprintf(command, PATH_MAX,
            "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' WOband8S_JS_template.omf > "
            "WOband8S_JS.omf", unit_test_dir);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Error preparing OMF file for unit "
                "test data directory path setup");
        return;
    }

    /* Run correct_los_model; capture precision parameters from DEBUG output */
    setenv("IAS_LOG_LEVEL", "DEBUG", 1);
    status = snprintf(command, PATH_MAX,
            "../correct_los_model %s | grep PRECISION_PARMS_TAG > %s",
            param_filename, new_prec_parm_filename);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Correct LOS Model failed");
        return;
    }

    /* Compare this test run precision parameters to saved parameters */
    status = snprintf(command, PATH_MAX, "diff %s %s", prec_parm_filename,
            new_prec_parm_filename);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found in precision "
                "parameters. Review %s and %s for details", prec_parm_filename,
                new_prec_parm_filename);
        return;
    }

    /* Look at the report files and make sure they match the expected
       results. The report files have header information and this has to be
       removed for the diff */
    status = snprintf(command, PATH_MAX, "sed '/Iteration 0/,$!'d "
            "%s | diff -q - %s/%s/band_8_storey_separated/%s",
            our_residuals_filename, unit_test_dir, test_directory,
            calval_residuals_filename);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found in precision "
                "residuals file %s", our_residuals_filename);
    }

    status = snprintf(command, PATH_MAX, "sed '/Reference Time/,$!'d "
            "%s | diff -q - %s/%s/band_8_storey_separated/%s",
            our_solution_filename, unit_test_dir, test_directory,
            calval_solution_filename);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found in precision "
                "residuals file %s", our_solution_filename);
    }
}  /* END test_precision_params_B */

Ensure test_precision_params_C(void)
{
    /* Using band 5 SCA-combined M. Choate data */
    char command[PATH_MAX];
    int status;
    char param_filename[] = "parameters_band_5_combined.odl";
    char prec_parm_filename[] = "precision_params_band_5_combined.txt";
    char new_prec_parm_filename[] = "new_precision_params_band_5_combined.txt";
    char our_residuals_filename[] = "los_model_band_5_combined.residuals";
    char calval_residuals_filename[] = "calval_outputs/precision.residuals";
    char our_solution_filename[] = "los_model_band_5_combined.solution";
    char calval_solution_filename[] = "calval_outputs/precision.solution";

    printf("Test precision params C......\n");

    status = snprintf(command, PATH_MAX,
            "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' WOband5C_template.omf > "
            "WOband5C.omf", unit_test_dir);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Error preparing OMF file for unit "
                "test data directory path setup");
        return;
    }

    /* Run correct_los_model; capture precision parameters from DEBUG output */
    setenv("IAS_LOG_LEVEL", "DEBUG", 1);
    status = snprintf(command, PATH_MAX,
            "../correct_los_model %s | grep PRECISION_PARMS_TAG > %s",
            param_filename, new_prec_parm_filename);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Correct LOS Model failed");
        return;
    }

    /* Compare this test run precision parameters to saved parameters */
    status = snprintf(command, PATH_MAX, "diff %s %s", prec_parm_filename,
            new_prec_parm_filename);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found in precision "
                "parameters. Review %s and %s for details", prec_parm_filename,
                new_prec_parm_filename);
        return;
    }

    /* Look at the report files and make sure they match the expected
       results. The report files have header information and this has to be
       removed for the diff */
    status = snprintf(command, PATH_MAX, "sed '/Iteration 0/,$!'d "
            "%s | diff -q - %s/%s/band_5_combined/%s", our_residuals_filename,
            unit_test_dir, test_directory, calval_residuals_filename);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found in precision "
                "residuals file %s", our_residuals_filename);
    }

    status = snprintf(command, PATH_MAX, "sed '/Reference Time/,$!'d "
            "%s | diff -q - %s/%s/band_5_combined/%s", our_solution_filename,
            unit_test_dir, test_directory, calval_solution_filename);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found in precision "
                "residuals file %s", our_solution_filename);
    }
}  /* END test_precision_params_C */

Ensure test_precision_params_D(void)
{
    /* Using band 5 SCA-separated M. Choate data */
    char command[PATH_MAX];
    int status;
    char param_filename[] = "parameters_band_5_separated.odl";
    char prec_parm_filename[] = "precision_params_band_5_separated.txt";
    char new_prec_parm_filename[] = "new_precision_params_band_5_separated.txt";
    char our_residuals_filename[] = "los_model_band_5_separated.residuals";
    char calval_residuals_filename[] = "calval_outputs/precision.residuals";
    char our_solution_filename[] = "los_model_band_5_separated.solution";
    char calval_solution_filename[] = "calval_outputs/precision.solution";
    char omf_template_filename[] = "WOband5S_template.omf";
    char omf_filename[] = "WOband5S.omf";
    char omf_verify_template_filename[] = "WOband5S_template_verify.omf";
    char omf_verify_filename[] = "WOband5S_verify.omf";

    printf("Test precision params D......\n");

    status = snprintf(command, PATH_MAX,
            "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' %s > %s",
            unit_test_dir, omf_template_filename, omf_filename);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Error preparing OMF file %s for unit "
                "test data directory path setup", omf_filename);
        return;
    }

    status = snprintf(command, PATH_MAX,
            "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' %s > %s",
            unit_test_dir, omf_verify_template_filename, omf_verify_filename);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Error preparing OMF file %s for unit "
                "test data directory path setup", omf_filename);
        return;
    }

    /* Run correct_los_model; capture precision parameters from DEBUG output */
    setenv("IAS_LOG_LEVEL", "DEBUG", 1);
    status = snprintf(command, PATH_MAX,
            "../correct_los_model %s | grep PRECISION_PARMS_TAG > %s",
            param_filename, new_prec_parm_filename);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Correct LOS Model failed");
        return;
    }

    /* compare the OMF files to verify the PASS parameters are correct */
    status = snprintf(command, PATH_MAX, "diff %s %s", omf_filename,
            omf_verify_filename);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found in OMFs."
                " Review %s and %s for details", omf_filename,
                omf_verify_filename);
        return;
    }

    /* Compare this test run precision parameters to saved parameters */
    status = snprintf(command, PATH_MAX, "diff %s %s", prec_parm_filename,
            new_prec_parm_filename);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found in precision "
                "parameters. Review %s and %s for details", prec_parm_filename,
                new_prec_parm_filename);
        return;
    }

    /* Look at the report files and make sure they match the expected
       results. The report files have header information and this has to be
       removed for the diff */
    status = snprintf(command, PATH_MAX, "sed '/Iteration 0/,$!'d "
            "%s | diff -q - %s/%s/band_5_separated/%s", our_residuals_filename,
            unit_test_dir, test_directory, calval_residuals_filename);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found in precision "
                "residuals file %s", our_residuals_filename);
    }

    status = snprintf(command, PATH_MAX, "sed '/Reference Time/,$!'d "
            "%s | diff -q - %s/%s/band_5_separated/%s", our_solution_filename,
            unit_test_dir, test_directory, calval_solution_filename);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found in precision "
                "residuals file %s", our_solution_filename);
    }
}  /* END test_precision_params_D */

Ensure test_database_store(void)
{
    /* Using band 8 SCA-combined M. Choate data */
    char command[PATH_MAX];
    int status;
    char param_filename[] = "parameters_band_8_choate_combined.odl";

    printf("Test database store......\n");

    /* Prepare the OMF file for use by replacing the test directory path */
    status = snprintf(command, PATH_MAX, "sed 's|$IAS_UNIT_TEST_DATA_DIR|"
            "%s|' WOband8C_MC_template.omf > WOband8C_MC.omf", unit_test_dir);
    assert_true_with_message(status >= 0, "Error creating command to create "
            "OMF file");
    if (status < 0)
        return;

    status = system(command);
    assert_true_with_message(status == 0, "Error preparing OMF file for unit "
            "test data directory path setup");
    if (status != 0)
        return;

    /* Prepare the database.  A row containing information for the WO needs
       to be added if it doesn't exist. */
    status = system("./check_or_insert_woid.csh WOband8C_MC");
    assert_true_with_message(status == 0, "Error preparing database, "
            "unable to insert work_order_id WOband8C_MC");
    if (status != 0)
        return;

    /* Run the correct LOS model making sure the log level is at the 
       appropriate level by setting the environment variable.  Don't care
       if this fails since I just want to eliminate extra stuff in the
       output. */
    setenv("IAS_LOG_LEVEL", "ERROR", 1);
    status = snprintf(command, PATH_MAX, "../correct_los_model %s",
        param_filename);
    assert_true_with_message(status >= 0, "Error creating command to run "
        "correct_los_model %s", param_filename);
    if (status < 0)
        return;

    status = system(command);
    assert_true_with_message(status == 0, "Correct LOS Model failed");
    if (status != 0)
        return;

    /* Check the precision parameters */
    status = system("./retrieve_stats_from_db.csh "
            "WOband8C_MC_precision_model_correction");
    assert_true_with_message(status == 0, "Error checking precision stats");
    if (status != 0)
        return;

    /* Check the GCP results */
    status = system("./retrieve_stats_from_db.csh "
            "WOband8C_MC_precision_model_corr_gcps");
    assert_true_with_message(status == 0, "Error checking GCP results");

}  /* END test_database_store */

/* Test to confirm that not having enough GCPs correlate will result in the
   precision correction being considered to not succeed */
Ensure test_insufficient_gcps(void)
{
    /* Using band 5 SCA-combined M. Choate data */
    char command[PATH_MAX];
    int status;
    char param_filename[] = "insufficient_gcps.odl";
    char omf_filename[] = "InsufGCP.omf";
    int succeeded_flag;

    printf("Test insufficient GCPS.....\n");

    status = snprintf(command, PATH_MAX,
            "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' InsufGCP_template.omf > %s",
            unit_test_dir, omf_filename);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Error preparing OMF file for unit "
                "test data directory path setup");
        return;
    }

    /* Run correct_los_model */
    status = snprintf(command, PATH_MAX, "../correct_los_model %s > /dev/null",
            param_filename);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Correct LOS Model failed");
        return;
    }

    /* Get the precision succeeded flag from the omf file */
    IAS_PARM_DECLARE_TABLE(parms, 1);
    IAS_PARM_ADD_INT(parms, PRECISION_SUCCEEDED_FLAG_PASS_1, "",
            IAS_PARM_REQUIRED, IAS_PARM_NOT_ARRAY,
            0, 0, 0, /* no limits */
            0, NULL, /* default */
            &succeeded_flag, sizeof(succeeded_flag), 1);
    status = ias_parm_read(omf_filename, parms, 1);
    assert_equal(status, SUCCESS);
    assert_equal(succeeded_flag, 0);

}
