/* Standard Library Includes */
#include <stdlib.h>
#include <string.h>

/* IAS Library Includes */
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_angle_gen.h"
#include "ias_miscellaneous.h" 

/* Local Includes */
#include "create_angle_bands.h" 

/* Local Defines */ 
#define NUM_OUTPUT_SCAS 2 /* Number of output SCAs */

/*******************************************************************************
Name: calculate_angles

Purpose: Calculate the satellite and solar zenith and azimuth angles.

Note: Returns the output Azimuth and Zenith in radians for both the Solar and
      Satellite angles.

Return: SUCCESS / ERROR
 ******************************************************************************/
int calculate_angles
(
    IAS_ANGLE_GEN_METADATA *metadata,/* I: Angle coefficient metadata */
    IAS_IMAGE *dem,                  /* I: Elevation image */
    int line,                        /* I: L1T line coordinate */
    int samp,                        /* I: L1T sample coordinate */
    double pixel_size,               /* I: Pixel size */
    int band_number,                 /* I: Spectral band number */
    ANGLE_MODEL angle_model,         /* I: Model type used to generate angles */
    ANGLE_TYPE angle_type,           /* I: Type of angles to generate */
    double *sat_angles,              /* O: Satellite angles (radians) */
    double *sun_angles               /* O: Solar angles (radians) */
)       
{
    double *height_ptr = NULL; /* Height to evaluate */
    double height;             /* Height pulled from DEM */
    int band_index;            /* Band index */
    int status;                /* Status return flag */
    int outside_image_flag;    /* Outside image flag */

    band_index = ias_sat_attr_convert_band_number_to_index(band_number);
    if (band_index < 0)
    {
        IAS_LOG_ERROR("Error retrieving band index for band number %d",
            band_number);
        return ERROR;
    }

    /* Check if the DEM is available */
    if (dem->data)
    { 
        /* Grab the height at the current line/sample */
        if (ias_misc_read_elevation_at_line_sample(dem, line, samp,
            pixel_size, pixel_size, &height) != SUCCESS)
        {
            IAS_LOG_ERROR("Getting elevation for Band number %d at line %d "
                "sample %d", band_number, line, samp);
            return ERROR;
        }
        height_ptr = &height;
    }

    if (angle_type != AT_SOLAR)
    {
        /* Calculate the satellite viewing angles */
        if (angle_model == AM_RPC)
        {
            status = ias_angle_gen_calculate_angles_rpc(metadata, line, samp, 
                height_ptr, band_index, IAS_ANGLE_GEN_SATELLITE, 
                &outside_image_flag, sat_angles);
        }
        else
        {
            status = ias_angle_gen_calculate_angles_rigor(metadata, line, samp, 
                height_ptr,band_index, IAS_ANGLE_GEN_SATELLITE,
                &outside_image_flag, sat_angles);
        }

        if (status != SUCCESS) 
        {
            IAS_LOG_ERROR("Evaluating satellite angles for band number %d.",
                band_number);
            return ERROR;
        }
    }

    if (angle_type != AT_SATELLITE)
    {
        /* Calculate the solar angles */
        if (angle_model == AM_RPC)
        {
            status = ias_angle_gen_calculate_angles_rpc(metadata, line, samp, 
                height_ptr, band_index, IAS_ANGLE_GEN_SOLAR, 
                &outside_image_flag, sun_angles);
        }
        else
        {
            status = ias_angle_gen_calculate_angles_rigor(metadata, line, samp,
                height_ptr, band_index, IAS_ANGLE_GEN_SOLAR,
                &outside_image_flag, sun_angles);
        }

        if (status != SUCCESS) 
        {
            IAS_LOG_ERROR("Evaluating solar angles for band number %d.",
                band_number);
            return ERROR;
        }
    }

    return SUCCESS;
}

/*******************************************************************************
Name: get_frame

Purpose: Retrieve the band frame information.

Return: SUCCESS / ERROR
 ******************************************************************************/
int get_frame
(
    IAS_ANGLE_GEN_METADATA *metadata, /* I: Angle coefficient metadata */
    int band_index,                   /* I: Band index */
    ANGLE_FRAME *frame               /* O: Image frame info */
)        
{
    if (!metadata->band_present[band_index])
    {
        IAS_LOG_WARNING("Band index %d not present in metadata skipping...", 
            band_index);
        return ERROR;
    }

    /* Establish the output image file frame */
    frame->band_number = metadata->band_metadata[band_index].band_number;
    frame->num_lines = metadata->band_metadata[band_index].l1t_lines;
    frame->num_samps = metadata->band_metadata[band_index].l1t_samps;
    frame->projection = metadata->projection;
    frame->pixel_size = metadata->band_metadata[band_index].pixel_size;
    frame->ul_corner.x = metadata->corners.upleft.x;
    frame->ul_corner.y = metadata->corners.upleft.y;

    return SUCCESS;
}

/*******************************************************************************
Name: open_l1g_image

Purpose: Opens the Satellite or Solar l1g output image.

Return: SUCCESS / ERROR
 ******************************************************************************/
int open_l1g_image
(
    L1GIO **l1gio_ptr,        /* I: L1G I/O pointer */
    IAS_ANGLE_GEN_TYPE type,  /* I: Type of angles to generate */
    char *image_filename,     /* I: HDF image output filename */
    int ang_filename_size,    /* I: Angle filename size */
    char *ang_filename        /* O: Angle filename */
)
{
    const char *suffix;             /* Filename suffix */
    int status;                     /* Return placeholder */

    if (type == IAS_ANGLE_GEN_SATELLITE)
    {
        suffix = "SAT.h5";
    }
    else
    {
        suffix = "SUN.h5";
    }

    /* Construct output file name */
    status = snprintf(ang_filename, ang_filename_size, "%s_%s",
        image_filename, suffix);
    if (status < 0 || status >= ang_filename_size)
    {
        IAS_LOG_ERROR("Creating the image filename from filename %s", 
            image_filename);
        return ERROR;
    } 

    /* Open the output file */
    *l1gio_ptr = ias_l1g_open_image(ang_filename, IAS_WRITE);
    if (!*l1gio_ptr)
    {
        IAS_LOG_ERROR("Error opening output file %s", ang_filename);
        return ERROR;
    }
    
    return SUCCESS;
}

/*******************************************************************************
Name: write_image_l1g

Purpose: Write out the solar illumination or the satellite viewing 
         angle images as a l1g file.

Note: Azimuth angles are stored in SCA 0 and the Zenith angles are 
      stored in SCA 1.

Returns: 
     Type = integer
     SUCCESS / ERROR
 ******************************************************************************/
int write_l1g_image
(
    L1GIO *l1gio_ptr,     /* I: L1GIO pointer to write output to */
    void *azimuth,        /* I: Array of azimuth angles */
    void *zenith,         /* I: Array of zenith angles */
    int band_index,       /* I: Ouput band index */
    int num_lines,        /* I: Number of image lines */
    int num_samps         /* I: Number of image samples */
)        
{
    int band_number;                  /* Ouput band number */
    int num_scas = NUM_OUTPUT_SCAS;   /* Number of output SCAs */
    L1G_BAND_IO *l1g_band;            /* Ouput band pointer */
    IAS_DATA_TYPE data_type = IAS_I2; /* Ouput data type */ 

    band_number = ias_sat_attr_convert_band_index_to_number(band_index);
    if (band_number == ERROR)
    {
        IAS_LOG_ERROR("Converting band index %d to band number", band_index);
        return ERROR;
    }

    /* Open the correct band in the output file */
    l1g_band = ias_l1g_open_band(l1gio_ptr, band_number, &data_type,
        &num_scas, &num_lines, &num_samps);
    if (!l1g_band)
    {
        IAS_LOG_ERROR("Failed to open band number %d", band_number);
        return ERROR;
    }

    /* Write the Azimuth as the first SCA */
    if (ias_l1g_write_image(l1g_band, 0, 0, 0, num_lines, num_samps, azimuth) 
        != SUCCESS)
    {
        IAS_LOG_ERROR("Writing azimuth to the l1g file for band number %d", 
            band_number);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }

    /* Write the Zenith as the second SCA */
    if (ias_l1g_write_image(l1g_band, 1, 0, 0, num_lines, num_samps, zenith) 
        != SUCCESS)
    {
        IAS_LOG_ERROR("Writing zenith to the l1g file for band number %d", 
            band_number);
        ias_l1g_close_band(l1g_band);
        return ERROR;
    }
    
    /* Close the current band */
    if (ias_l1g_close_band(l1g_band) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing l1g output band for band number %d", 
            band_number);
        return ERROR;
    }
    l1g_band = NULL;

    return SUCCESS;
}

/*******************************************************************************
Name: write_l1g_metadata

Purpose: Write out the image file and band metadata 

Returns: 
     Type = integer
     SUCCESS / ERROR
 ******************************************************************************/
int write_l1g_metadata
(
    L1GIO *l1gio_ptr,                  /* I: L1GIO pointer to write output to */
    IAS_ANGLE_GEN_METADATA *metadata,  /* I: Angle coefficient metadata */
    ANGLE_PARAMETERS *parameters /* I: Processing parameters */
)       
{
    IAS_L1G_FILE_METADATA fmd;    /* file metadata */
    IAS_L1G_BAND_METADATA *bmd;   /* band metadata */
    int bands_to_output;
    const char *ias_version = NULL;
    int index;

    /* Allocate memory for the band metadata */
    bmd = malloc(sizeof(IAS_L1G_BAND_METADATA) * metadata->num_bands);
    if (!bmd)
    {
        IAS_LOG_ERROR("Allocating memory for band metadata");
        return ERROR;
    }
    memset(bmd, 0, sizeof(IAS_L1G_BAND_METADATA) * metadata->num_bands);
    memset(&fmd, 0, sizeof(fmd));

    /* Add the file metadata availaible*/
    fmd.projection_code = metadata->projection.proj_code;

    fmd.zone_code = metadata->projection.zone;

    strncpy(fmd.datum, metadata->datum, sizeof(fmd.datum));
    fmd.datum[sizeof(fmd.datum) - 1] = '\0';
    fmd.spheroid_code = metadata->projection.spheroid;

    strncpy(fmd.projection_units, metadata->units,
            sizeof(fmd.projection_units));
    fmd.projection_units[sizeof(fmd.projection_units) - 1] = '\0';
    for (index = 0; index < IAS_PROJ_PARAM_SIZE; index++)
    {
        fmd.projection_parameters[index] 
            = metadata->projection.parameters[index];
    }

    strncpy(fmd.spacecraft, metadata->spacecraft_id, sizeof(fmd.spacecraft));
    fmd.spacecraft[sizeof(fmd.spacecraft) - 1] = '\0';

    /* assume systematic correction */
    strncpy(fmd.correction_type, "L1G", sizeof(fmd.correction_type));
    fmd.correction_type[sizeof(fmd.correction_type) - 1] = '\0';

    /* Grab the software versions. */
    ias_version = ias_misc_get_software_version();
    if (ias_version == NULL)
        strncpy(fmd.software_version, "", sizeof(fmd.software_version));
    else
    {
        strncpy(fmd.software_version, ias_version,
                sizeof(fmd.software_version));
    }
    fmd.software_version[sizeof(fmd.software_version) - 1] = '\0';

    strncpy(fmd.sun_angle_correction, "NONE", sizeof(fmd.sun_angle_correction));
    fmd.sun_angle_correction[sizeof(fmd.sun_angle_correction) - 1] = '\0';

    fmd.earth_sun_distance = metadata->earth_sun_distance;

    for (index = 0, bands_to_output = 0; index < metadata->num_bands; index++)
    {
        const char *band_name;
        char complete_band_name[IAS_BAND_NAME_SIZE * 2];
        int status; 
        IAS_ANGLE_GEN_BAND *band_ptr = &metadata->band_metadata[index];

        if (!parameters->process_band[index] || !metadata->band_present[index])
            continue;

        /* Set the band number */
        bmd[bands_to_output].band_number = band_ptr->band_number;

        /* Grab the band name and apply it to the L1G */
        band_name = ias_sat_attr_convert_band_number_to_name(
            band_ptr->band_number);
        if (!band_name)
        {
            IAS_LOG_ERROR("Error retrieving band name for band number %d",
                band_ptr->band_number);
            free(bmd);
            return ERROR;
        }
        status = snprintf(complete_band_name, sizeof(complete_band_name), 
            "Angle_Band_%s", band_name);
        if (status < 0 || status >= sizeof(complete_band_name))
        {
            IAS_LOG_ERROR("Creating the band name for band number %d",
                band_ptr->band_number);
            return ERROR;
        }

        strncpy(bmd[bands_to_output].band_name, complete_band_name, 
            sizeof(bmd[bands_to_output].band_name));
        bmd[bands_to_output].band_name[
            sizeof(bmd[bands_to_output].band_name) - 1] = '\0';

        /* copy the projection window to the metadata */
        bmd[bands_to_output].upper_left_x = metadata->corners.upleft.x;
        bmd[bands_to_output].upper_left_y = metadata->corners.upleft.y;

        bmd[bands_to_output].upper_right_x = metadata->corners.upright.x;
        bmd[bands_to_output].upper_right_y = metadata->corners.upright.y;

        bmd[bands_to_output].lower_right_x = metadata->corners.loright.x;
        bmd[bands_to_output].lower_right_y = metadata->corners.loright.y;

        bmd[bands_to_output].lower_left_x = metadata->corners.loleft.x;
        bmd[bands_to_output].lower_left_y = metadata->corners.loleft.y;

        /* set the projected pixel size */
        bmd[bands_to_output].projection_distance_y = band_ptr->pixel_size;
        bmd[bands_to_output].projection_distance_x = band_ptr->pixel_size;

        /* Set the rescale and reflectance scaling parameters */
        bmd[bands_to_output].spectral_radiance_scaling_gain = 100;
        bmd[bands_to_output].spectral_radiance_scaling_offset = 0;
        bmd[bands_to_output].radiance_valid = TRUE;
        bmd[bands_to_output].reflectance_valid = FALSE;
                
        bands_to_output++;
    }

    /* write the metadata */
    if ((ias_l1g_set_file_metadata(l1gio_ptr, &fmd) != SUCCESS) ||
        (ias_l1g_set_band_metadata(l1gio_ptr, bmd, bands_to_output) != SUCCESS))
    {
        IAS_LOG_ERROR("Error updating metadata for L1G image");
        free(bmd);
        return ERROR;
    }
    free(bmd);

    return SUCCESS;
}
