/* Standard Library Includes */
#include <stdlib.h>
#include <libgen.h>
#include <string.h>
#include <math.h>

/* IAS Library Includes */
#include "ias_angle_gen.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_satellite_attributes.h"
#include "ias_math.h"

/* Local Includes */
#include "create_angle_bands.h"

/* Local defines */
#define USE_MESSAGE "Usage: create_angle_bands <parameter_file>"

/*******************************************************************************
Name: free_memory

Purpose: Frees up all of the memory for this application

Return: None
 ******************************************************************************/
static int free_memory
(
    IAS_ANGLE_GEN_METADATA *metadata,/* I: Angle coefficient metadata */
    IAS_IMAGE *dem,                  /* I: Elevation image */
    L1GIO *l1gio_sat_ptr,            /* I: Satellite L1G I/O pointer */
    L1GIO *l1gio_sun_ptr             /* I: Solar L1G I/O pointer */
)
{
    int return_status = SUCCESS; /* Return status */

    /* Free the ephemeris structure */
    ias_angle_gen_free(metadata);

    /* Free the projection transformation */
    ias_angle_gen_free_transformation(metadata);

    /* Free the elevation data */
    if (dem->data)
    {
        free(dem->data);
        dem->data = NULL;
    }
    
    /* Free the I/O pointers */
    if (l1gio_sat_ptr)
    {
        if (ias_l1g_close_image(l1gio_sat_ptr) != SUCCESS)
        {   
            IAS_LOG_ERROR("Closing satellite image");
            return_status = ERROR;
        }
        l1gio_sat_ptr = NULL;
    }
    if (l1gio_sun_ptr)
    {
        if (ias_l1g_close_image(l1gio_sun_ptr) != SUCCESS)
        {   
            IAS_LOG_ERROR("Closing solar image");
            return_status = ERROR;
        }
        l1gio_sun_ptr = NULL;
    }

    return return_status;
}

/**************************************************************************
NAME: create_angle_bands (main)

PURPOSE: Calculates satellite viewing and solar illumination zenith and 
         azimuth angles for each L1T pixel (with optional sub-sampling), 
         and writes the angles out to binary flat image files with 
         ENVI headers.      

RETURNS: EXIT_SUCCESS or EXIT_FAILURE
***************************************************************************/
int main
( 
    int argc, 
    char *argv[] 
)
{
    char *root_filename;                /* Scene root file name */
    int band_index;                     /* Metadata band index */
    int subsamp;                        /* Sub-sampling factor */
    int calc_sat_angles;                /* Local satellite angle flag */
    int calc_solar_angles;              /* Local solar angle flag */
    ANGLE_PARAMETERS parameters;        /* Parameters read in from file */
    IAS_ANGLE_GEN_METADATA metadata;    /* Angle metadata structure */
    IAS_IMAGE dem;                      /* DEM image data */
    char directory[PATH_MAX];           /* File directory */
    char basename[PATH_MAX];            /* File basename */
    char extension[PATH_MAX];           /* File extension */
    L1GIO *l1gio_sat_ptr = NULL;        /* Satellite L1G I/O pointer */
    L1GIO *l1gio_sun_ptr = NULL;        /* Solar L1G I/O pointer */
    double scaled_r2d; /* Radian to degree conversion scaled by 100 */
    char sat_ang_filename[PATH_MAX];    /* Satellite output filename */
    char sun_ang_filename[PATH_MAX];    /* Solar output filename */

    /* Initialize dem data */
    dem.data = NULL;

    /* Initialize conversion factor from radians to hundredths of degrees */
    scaled_r2d = ias_math_get_degrees_per_radian() * 100;

    /* Initialize the logging library */
    if (ias_log_initialize("create_angle_bands") != SUCCESS)
    {
        IAS_LOG_ERROR("Error initializing logging library");
        return EXIT_FAILURE;
    }

    /* If there are no arguments, generate a usage message */
    if (argc < 2)
    {
        IAS_LOG_ERROR(USE_MESSAGE);
        return EXIT_FAILURE;
    }

    /* Initialize the satellite attributes */
    if (ias_sat_attr_initialize(IAS_L8) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attributes library");
        return EXIT_FAILURE;
    }

    /* Read the parameters file passed as parameter */
    if (read_parameters(argv[1], &parameters) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameter file %s", argv[1]);
        return EXIT_FAILURE;
    }
    
    /* Issue message on type of model being used */
    if (parameters.angle_model == AM_RPC)
    {
        IAS_LOG_INFO("Using the Rational Polynomial Coefficient (RPC) model "
            "to create angles");
    }
    else
    {
        IAS_LOG_INFO("Using the Rigorous model to create angles");
    }

    /* Setup local sub sampling factor variable */
    subsamp = parameters.sub_sample_factor;
    IAS_LOG_INFO("Using sub-sampling factor of %d to create angles", subsamp);

    /* Extract the root file name */
    ias_misc_split_filename(parameters.metadata_filename, directory, basename,
        extension);
    root_filename = strtok(basename, "_");
    if (!root_filename)
    {
        IAS_LOG_ERROR("Determining the root filename");
        return EXIT_FAILURE;
    }

    /* Read the ANG metadata file */
    if (ias_angle_gen_read_ang(parameters.metadata_filename, &metadata) 
        != SUCCESS)
    {
        IAS_LOG_ERROR("Reading the metadata file %s.", 
            parameters.metadata_filename);
        return EXIT_FAILURE;
    }
    
    /* Check what angles should be calculated */
    calc_sat_angles = TRUE;
    calc_solar_angles = TRUE;
    if (parameters.angle_type == AT_SOLAR)
    {
        calc_sat_angles = FALSE;
    }
    else if (parameters.angle_type == AT_SATELLITE)
    {
        calc_solar_angles = FALSE;
    }

    /* Read the elevation data from DEM */
    if (parameters.use_dem_flag)
    {
        IAS_DATA_TYPE data_type_read;
        IAS_CORNERS *corners = &metadata.corners;
        double corner_diff; /* Max corner difference */

        IAS_LOG_INFO("Using DEM %s to create angles", parameters.dem_filename);

        /* Read the elevation file */
        if (ias_misc_read_single_band_l1g(parameters.dem_filename, 1,
            &data_type_read, &dem) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading the DEM %s", parameters.dem_filename);
            free_memory(&metadata, &dem, l1gio_sat_ptr, l1gio_sun_ptr);
            return EXIT_FAILURE;
        }

        if (data_type_read != IAS_I2)
        {
            IAS_LOG_ERROR("Invalid data type found in DEM");
            free_memory(&metadata, &dem, l1gio_sat_ptr, l1gio_sun_ptr);
            return EXIT_FAILURE;
        }

        /* Make sure the corners of the image are within 0.5 pixels */
        corner_diff = dem.pixel_size_x * 0.5;
        if ((fabs(corners->upleft.y - dem.corners.upleft.y) > corner_diff)
            || (fabs(corners->upleft.x - dem.corners.upleft.x) > corner_diff)
            || (fabs(corners->upright.y - dem.corners.upright.y) > corner_diff)
            || (fabs(corners->upright.x - dem.corners.upright.x) > corner_diff)
            || (fabs(corners->loleft.y - dem.corners.loleft.y) > corner_diff)
            || (fabs(corners->loleft.x - dem.corners.loleft.x) > corner_diff)
            || (fabs(corners->loright.y - dem.corners.loright.y) > corner_diff)
            || (fabs(corners->loright.x - dem.corners.loright.x) > corner_diff))
        {
            IAS_LOG_ERROR("Image corners do not match");
            free_memory(&metadata, &dem, l1gio_sat_ptr, l1gio_sun_ptr);
            return EXIT_FAILURE;
        }
    }
    else
    {
        IAS_LOG_INFO("Not using DEM to create angles");
    }

    
    /* Check if l1g output flag has been set and open the image pointer */
    if (parameters.output_as_l1g_flag)
    {
        if (calc_sat_angles)
        {
            if (open_l1g_image(&l1gio_sat_ptr, IAS_ANGLE_GEN_SATELLITE, 
                root_filename, sizeof(sat_ang_filename), sat_ang_filename)
                != SUCCESS)
            {   
                IAS_LOG_ERROR("Opening satellite l1g file");
                free_memory(&metadata, &dem, l1gio_sat_ptr, l1gio_sun_ptr);
                return EXIT_FAILURE;
            }
        }

        if (calc_solar_angles)
        {
            if (open_l1g_image(&l1gio_sun_ptr, IAS_ANGLE_GEN_SOLAR, 
                root_filename, sizeof(sun_ang_filename), sun_ang_filename)
                != SUCCESS)
            {   
                IAS_LOG_ERROR("Opening solar l1g file");
                free_memory(&metadata, &dem, l1gio_sat_ptr, l1gio_sun_ptr);
                return EXIT_FAILURE;
            }
        }
    }

    /* Initialize the metadata projection transformation */
    if (ias_angle_gen_initialize_transformation(&metadata) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the projection transformation");
        free_memory(&metadata, &dem, l1gio_sat_ptr, l1gio_sun_ptr);
        exit (EXIT_FAILURE);
    }

    /* Process the angles for each band */
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        short *sat_zn = NULL;           /* Satellite zenith angle */
        short *sat_az = NULL;           /* Satellite azimuth angle */
        short *sun_zn = NULL;           /* Solar zenith angle */
        short *sun_az = NULL;           /* Solar azimuth angle */
        int num_lines;                  /* Lines in output angle band */
        int num_samps;                  /* Samples in output angle band */
        int count;                      /* Output sample counter */
        ANGLE_FRAME frame;             /* Output image frame info */
        IAS_MISC_LINE_EXTENT *trim_lut; /* Image trim lookup table */
        size_t angle_size;              /* Angle array size */
        int line;                       /* Line index */
        int samp;                       /* Sample index */
        int band_number;                /* Band number */ 
        
        /* Retrieve the band number for current index */
        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Getting band number for band index %d", band_index);
            free_memory(&metadata, &dem, l1gio_sat_ptr, l1gio_sun_ptr);
            return EXIT_FAILURE;
        }

        /* Check if this band should be processed */
        if (!parameters.process_band[band_index])
             continue;
        

        /* Get framing information for this band if return is not successful
           warning has already been issued becuase the band is missing from the 
           metadata so continue onto the next band */
        if (get_frame(&metadata, band_index, &frame) != SUCCESS)
            continue;

        if (frame.band_number != band_number)
        {
            IAS_LOG_ERROR("Frame band number %d does not match expected band "
                "number %d", frame.band_number, band_number);
            free_memory(&metadata, &dem, l1gio_sat_ptr, l1gio_sun_ptr);
            return EXIT_FAILURE;
        }

        /* Calculate size of sub-sampled output image */
        num_lines = (frame.num_lines - 1) / subsamp + 1;
        num_samps = (frame.num_samps - 1) / subsamp + 1;
        IAS_LOG_INFO("Processing band number %d with %d lines and %d samples "
            "using %d as sub-sampling factor", band_number, num_lines,
            num_samps, subsamp);

        /* Retrieve the trim look up table to remove the scene crenulation */
        trim_lut = ias_misc_create_output_image_trim_lut(
            metadata.band_metadata[band_index].active_l1t_corner_lines, 
            metadata.band_metadata[band_index].active_l1t_corner_samps, 
            frame.num_lines, frame.num_samps);
        if (!trim_lut)
        {
            IAS_LOG_ERROR("Creating image trim lookup table for band number %d",
                band_number);
            free_memory(&metadata, &dem, l1gio_sat_ptr, l1gio_sun_ptr);
            return EXIT_FAILURE;
        }
       
        /* Calculate the angle sizes */
        angle_size = num_lines * num_samps * sizeof(short);

        /* Allocate the output buffers */
        if (calc_sat_angles)
        {
            sat_zn = (short *)malloc(angle_size);
            if (!sat_zn)
            {
                IAS_LOG_ERROR("Allocating satellite zenith angle array for "
                    "band number %d", band_number);
                free_memory(&metadata, &dem, l1gio_sat_ptr, l1gio_sun_ptr);
                free(trim_lut);
                return EXIT_FAILURE;
            }

            sat_az = (short *)malloc(angle_size);
            if (!sat_az)
            {
                IAS_LOG_ERROR("Allocating satellite azimuth angle array for "
                    "band number %d", band_number);
                free(sat_zn);
                free(trim_lut);
                free_memory(&metadata, &dem, l1gio_sat_ptr, l1gio_sun_ptr);
                return EXIT_FAILURE;
            }
        }

        if (calc_solar_angles)
        {
            sun_zn = (short *)malloc(angle_size);
            if (!sun_zn)
            {
                IAS_LOG_ERROR("Allocating solar zeith angle array for "
                    "band number %d", band_number);
                free(sat_zn);
                free(sat_az);
                free(trim_lut);
                free_memory(&metadata, &dem, l1gio_sat_ptr, l1gio_sun_ptr);
                return EXIT_FAILURE;
            }

            sun_az = (short *)malloc(angle_size);
            if (!sun_az)
            {
                IAS_LOG_ERROR("Allocating solar azimuth angle array for "
                    "band number %d", band_number);
                free(sat_zn);
                free(sat_az);
                free(sun_zn);
                free(trim_lut);
                free_memory(&metadata, &dem, l1gio_sat_ptr, l1gio_sun_ptr);
                return EXIT_FAILURE;
            }
        }

        /* Loop through the L1T lines and samples */
        for (line = 0, count = 0; line < frame.num_lines; line += subsamp)
        {
            if (!((line / subsamp) % 500))
            {
                IAS_LOG_INFO("Line %d processed", line / subsamp);
            }

            for (samp = 0; samp < frame.num_samps; samp += subsamp, count++)
            {
                double satang[2]; /* Satellite viewing angles */
                double sunang[2]; /* Solar viewing angles */

                if (samp <= trim_lut[line].start_sample || 
                    samp >= trim_lut[line].end_sample)
                {
                    if (calc_sat_angles)
                    {
                        sat_zn[count] = parameters.background;
                        sat_az[count] = parameters.background;
                    }
                    if (calc_solar_angles)
                    {
                        sun_zn[count] = parameters.background;
                        sun_az[count] = parameters.background;
                    }
                    continue;
                }
                
                /* Calculate the satellite and solar azimuth and zenith */
                if (calculate_angles(&metadata, &dem, line, samp, 
                    frame.pixel_size, band_number, parameters.angle_model,
                    parameters.angle_type, satang, sunang) != SUCCESS)
                {
                    IAS_LOG_ERROR("Evaluating angles in band number %d", 
                        band_number);
                    free(sat_zn);
                    free(sat_az);
                    free(sun_zn);
                    free(sun_az);
                    free(trim_lut);
                    free_memory(&metadata, &dem, l1gio_sat_ptr, l1gio_sun_ptr);
                    return EXIT_FAILURE;
                }

                /* Quantize the angles */
                if (calc_sat_angles)
                {
                    sat_zn[count] = (short)round(scaled_r2d 
                        * satang[IAS_ANGLE_GEN_ZENITH_INDEX]);
                    sat_az[count] = (short)round(scaled_r2d 
                        * satang[IAS_ANGLE_GEN_AZIMUTH_INDEX]);
                }
                if (calc_solar_angles)
                {
                    sun_zn[count] = (short)round(scaled_r2d 
                        * sunang[IAS_ANGLE_GEN_ZENITH_INDEX]);
                    sun_az[count] = (short)round(scaled_r2d 
                        * sunang[IAS_ANGLE_GEN_AZIMUTH_INDEX]);
                }
            }
        }
        
        /* Free the trim table */
        free(trim_lut);
        trim_lut = NULL;

        /* Write angles to output file */
        frame.num_lines = num_lines;
        frame.num_samps = num_samps;
        frame.pixel_size = frame.pixel_size * subsamp;
    
        /* Write out the satellite image */
        if (calc_sat_angles)
        {   
            int status; 
            if (parameters.output_as_l1g_flag)
            {
                status = write_l1g_image(l1gio_sat_ptr, sat_az, sat_zn,
                    band_index, frame.num_lines, frame.num_samps);
            }
            else
            {
                status = ias_angle_gen_write_image(root_filename, sat_az,
                    sat_zn, IAS_ANGLE_GEN_SATELLITE, band_index, 
                    frame.num_lines, frame.num_samps, frame.ul_corner, 
                    frame.pixel_size, &frame.projection);
            }
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Writing satellite angle for band number %d", 
                    band_number);
                free(sat_zn);
                free(sat_az);
                free(sun_zn);
                free(sun_az);
                free_memory(&metadata, &dem, l1gio_sat_ptr, l1gio_sun_ptr);
                return EXIT_FAILURE;
            }

            free(sat_zn);
            free(sat_az);
            sat_zn = NULL;      
            sat_az = NULL;
        }
    
        /* Write out the solar image */
        if (calc_solar_angles)
        {
            int status;
            if (parameters.output_as_l1g_flag)
            {
                status = write_l1g_image(l1gio_sun_ptr, sun_az, sun_zn,
                    band_index, frame.num_lines, frame.num_samps);
            }
            else
            {
                status = ias_angle_gen_write_image(root_filename, sun_az,
                    sun_zn, IAS_ANGLE_GEN_SOLAR, band_index, 
                    frame.num_lines, frame.num_samps, frame.ul_corner, 
                    frame.pixel_size, &frame.projection);
            }

            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Writing solar angles for band number %d",
                    band_number);
                free(sun_zn);
                free(sun_az);
                free_memory(&metadata, &dem, l1gio_sat_ptr, l1gio_sun_ptr);
                return EXIT_FAILURE;
            }

            free(sun_zn);
            free(sun_az);
            sun_zn = NULL;
            sun_az = NULL;
        }
    }
    
    /* Write out the L1G metadata */ 
    if (parameters.output_as_l1g_flag)
    {
        if (calc_sat_angles)
        {
            if (write_l1g_metadata(l1gio_sat_ptr, &metadata, &parameters) 
                != SUCCESS)
            {
                IAS_LOG_ERROR("Writing l1g metadata");
                free_memory(&metadata, &dem, l1gio_sat_ptr, l1gio_sun_ptr);
                return EXIT_FAILURE;
            }
        }

        if (calc_solar_angles)
        {
            if (write_l1g_metadata(l1gio_sun_ptr, &metadata, &parameters) 
                != SUCCESS)
            {
                IAS_LOG_ERROR("Writing l1g metadata");
                free_memory(&metadata, &dem, l1gio_sat_ptr, l1gio_sun_ptr);
                return EXIT_FAILURE;
            }

            /* Update the OMF with the sun angle filename */
            if (update_omf(get_omf_filename(), sun_ang_filename,
                get_processing_pass()) != SUCCESS)
            {
                IAS_LOG_ERROR("Updating the OMF file");
                free_memory(&metadata, &dem, l1gio_sat_ptr, l1gio_sun_ptr);
                return EXIT_FAILURE;
            }
        }
    }

    /* Release the metadata */
    if (free_memory(&metadata, &dem, l1gio_sat_ptr, l1gio_sun_ptr) != SUCCESS)
    {
        IAS_LOG_ERROR("Freeing up the application memory");
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}
