#ifndef _SIMPLE_VIEW_H_
#define _SIMPLE_VIEW_H_

/* Standard Library Includes */
#include <limits.h>

/* IAS Library Includes */
#include "ias_angle_gen.h"

/* Model used the generate the angles */
typedef enum angle_model
{
    AM_UNKNOWN = 0, /* Unknown model type */
    AM_RPC,         /* Rational polynomial coefficients model */
    AM_RIGOROUS     /* Rigorous vector calculation model */
} ANGLE_MODEL;

/* Type of angle */
typedef enum angle_type
{
    AT_UNKNOWN = 0, /* Unknown angle type */
    AT_BOTH,        /* Both Solar and Satellite angle types */
    AT_SATELLITE,   /* Satellite angle type */
    AT_SOLAR        /* Solar angle type */
} ANGLE_TYPE;

/* Used as median between API and main routine for the band metadata needed to
   write the image to file */
typedef struct angle_frame
{
    int band_number;        /* Band number */
    int num_lines;          /* Number of lines in frame */
    int num_samps;          /* Number of samples in frame */
    IAS_DBL_XY ul_corner;   /* Upper left corner coordinates */
    double parms[IAS_PROJ_PARAM_SIZE];/* Projection coefficients */
    double pixel_size;      /* Pixel size in meters */
    IAS_PROJECTION projection; /* Projection information */ 
} ANGLE_FRAME;

/* Parameters struct used to hold all of the read in ODL parameters */
typedef struct angle_parameters
{
    int process_band[IAS_MAX_NBANDS]; /* Process band index array */
    char metadata_filename[PATH_MAX]; /* Metadata filename */
    char dem_filename[PATH_MAX]; /* DEM filename */
    int use_dem_flag;            /* Flag used to check if DEM should be used */
    ANGLE_TYPE angle_type;       /* Type of angles to be generated */
    ANGLE_MODEL angle_model;     /* Model used to generate angles */
    int sub_sample_factor;       /* Sub-sampling factor to be used */
    short background;            /* Background value used for fill pixels */
    int output_as_l1g_flag;      /* Flag to indicate output as HDF format */
} ANGLE_PARAMETERS;

/************************ PROTOTYPES START ************************************/

int calculate_angles
(
    IAS_ANGLE_GEN_METADATA *metadata,/* I: Angle coefficient metadata */
    IAS_IMAGE *dem,                  /* I: Elevation image */
    int line,                        /* I: L1T line coordinate */
    int samp,                        /* I: L1T sample coordinate */
    double pixel_size,               /* I: Pixel size */
    int band_number,                 /* I: Spectral band number */
    ANGLE_MODEL angle_model,         /* I: Model type used to generate angles */
    ANGLE_TYPE angle_type,           /* I: Type of angles to generate */
    double *sat_angles,              /* O: Satellite angles (radians) */
    double *sun_angles               /* O: Solar angles (radians) */
);

int get_frame
(
    IAS_ANGLE_GEN_METADATA *metadata, /* I: Angle coefficient metadata */
    int band_index,                   /* I: Band index */
    ANGLE_FRAME *frame               /* O: Image frame info */
);

char *get_omf_filename
(
);

int get_processing_pass
(
);

int open_l1g_image
(
    L1GIO **l1gio_ptr,        /* I: L1G I/O pointer */
    IAS_ANGLE_GEN_TYPE type,  /* I: Type of angles to generate */
    char *image_filename,     /* I: HDF image output filename */
    int ang_filename_size,    /* I: Angle filename size */
    char *ang_filename        /* O: Angle filename */
);

int read_parameters
(
    const char *parameter_filename,         /* I: Parameter file name */
    ANGLE_PARAMETERS *parameters      /* O: Generate Chips parameters */
);

int write_l1g_image
(
    L1GIO *l1gio_ptr,     /* I: L1GIO pointer to write output to */
    void *azimuth,        /* I: Array of azimuth angles */
    void *zenith,         /* I: Array of zenith angles */
    int band_index,       /* I: Ouput band index */
    int num_lines,        /* I: Number of image lines */
    int num_samps         /* I: Number of image samples */
);

int write_l1g_metadata
(
    L1GIO *l1gio_ptr,                  /* I: L1GIO pointer to write output to */
    IAS_ANGLE_GEN_METADATA *metadata,  /* I: Angle coefficient metadata */
    ANGLE_PARAMETERS *parameters /* I: Processing parameters */
);

int update_omf
(
    const char *omf_filename,       /* I: Name of OMF */
    char *sun_angle_filename,       /* I: Sun Angle filename */
    int processing_pass             /* I: Number of WO processing flow pass */
);

#endif
