/* Standard C Includes */
#include <stdlib.h>
#include <string.h>

/* IAS Includes */
#include "ias_const.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"

/* Local Includes */
#include "create_angle_bands.h"

/* Local Variable Declarations */
static char local_angle_model[9];
static char local_angle_type[10];
static int local_processing_pass;
static int local_band_count;
static int local_band_numbers[IAS_MAX_NBANDS] = {ERROR};
static char local_omf_filename[PATH_MAX]; 
static char local_wo_id[IAS_WORKORDER_ID_SIZE];

/******************************************************************************
NAME: get_processing_pass

PURPOSE: Returns the processing pass

RETURN VALUE: Type = int
******************************************************************************/
int get_processing_pass
(
)
{
    return local_processing_pass;
}

/******************************************************************************
NAME: get_omf_filename

PURPOSE: Returns the current work order OMF filename

RETURN VALUE: Type = int
******************************************************************************/
char *get_omf_filename
(
)
{
    return local_omf_filename;
}

/******************************************************************************
NAME: process_parameters

PURPOSE: This function performs special processing on parameters which need
         to be translated after being read from the parameter or OMF files.

RETURN VALUE: Type = int
    Value     Description
    -----     -----------
    SUCCESS   The parameters were successfully processed
    ERROR     An error occurred processing the parameters
******************************************************************************/
static int process_parameters
(
    ANGLE_PARAMETERS *parameters /* I/O: Generation parameters */
)
{
    int index; /* Generic loop variable */
    int band_already_in_list[IAS_MAX_NBANDS];

    /* Convert the angle model from string to enumeration */
    if (strcmp(local_angle_model, "RPC") == 0)
    {
        parameters->angle_model = AM_RPC;
    }
    else
    {
        parameters->angle_model = AM_RIGOROUS;
    }

    /* Convert the angle type from string to enumeration */
    if (strcmp(local_angle_type, "BOTH") == 0)
    {
        parameters->angle_type = AT_BOTH;
    }
    else if (strcmp(local_angle_type, "SATELLITE") == 0)
    {
        parameters->angle_type = AT_SATELLITE;
    }
    else
    {
        parameters->angle_type = AT_SOLAR;
    }

    /* Initialize band list */
    for (index = 0; index < IAS_MAX_NBANDS; index++)
    {
        parameters->process_band[index] = 0;
        band_already_in_list[index] = 0;
    }

    for (index = 0; index < local_band_count; index++)
    {
        /* Convert the band number entered into a band index */
        int temp_band_index = ias_sat_attr_convert_band_number_to_index(    
                                local_band_numbers[index]);
        if ((index == 0) && (temp_band_index == ERROR))
        {
            /* No bands numbers entered, so it is an error */
            IAS_LOG_ERROR("Illegal band number %d", local_band_numbers[index]);
            return ERROR;
        }

        /* Make sure the band isn't already in the list */
        if (temp_band_index != ERROR)
        {
            if (!band_already_in_list[temp_band_index])
            {
                /* put band in list and flag that it has been added 
                   to the list */
                parameters->process_band[temp_band_index] = 1;
                band_already_in_list[temp_band_index] = 1;
            }
            else if (band_already_in_list[temp_band_index] == 1)
            {
                 /* only print the warning message once for each band */
                 band_already_in_list[temp_band_index] += 1;

                 IAS_LOG_WARNING("Warning: Band %d appears multiple times "
                            "in the BAND_LIST, only the first one was used",
                            local_band_numbers[index]);
            }
        }
    }

    return SUCCESS;
}

/******************************************************************************
NAME: read_omf_parameters

PURPOSE: This function reads all the parameters from the OMF file.

RETURN VALUE: Type = int
    Value    Description
    -----    -----------
    SUCCESS  The parameters were successfully read
    ERROR    An error occurred reading the parameters
******************************************************************************/
static int read_omf_parameters
(
    const char *omf_filename,           /* I : OMF filename that contains order 
                                               parameters */
    ANGLE_PARAMETERS *parameters        /* O: Generate Chips parameters */
)
{
    int count;               /* loop counter */
    int help_requested;      /* Help request variable */
 
    /* Now read and check the OMF. */
    /* start parameter table definition */
    IAS_PARM_DECLARE_TABLE(omf_parms, 2);

    IAS_PARM_ANGLE_COEF_FILENAME_FROM_PASS(omf_parms, IAS_PARM_REQUIRED,
        parameters->metadata_filename, sizeof(parameters->metadata_filename),
        1, local_processing_pass);
        
    IAS_PARM_DEM_FILENAME_FROM_PASS(omf_parms, IAS_PARM_OPTIONAL,
        parameters->dem_filename, sizeof(parameters->dem_filename), 1,
        local_processing_pass);

    /* calculate number of items in the omf_parms list */
    count = IAS_PARM_GET_TABLE_SIZE(omf_parms);

    help_requested = ias_parm_provide_help(omf_filename, omf_parms, count,
            IAS_OMF_PARAMETERS);
    if (help_requested)
    {
        /* Return ERROR if help requested */
        if (help_requested == TRUE)
        {
            /* successfully provided help, as requested, so just exit */
            exit(EXIT_SUCCESS);
        }
        IAS_LOG_ERROR("Error occurred providing OMF help");
        return ERROR;
    }

    if (ias_parm_read(omf_filename, omf_parms, count) != SUCCESS)
    {
        IAS_LOG_ERROR("Error reading OMF parameter file %s", omf_filename);
        return ERROR;
    }

    /* If the DEM flag was set check if DEM filename was passed if not 
       reset the flag */
    if (parameters->use_dem_flag &&
        (IAS_PARM_GET_COUNT_READ(omf_parms, DEM_FILENAME) == 0 
        || strlen(parameters->dem_filename) == 0))
    {
        parameters->use_dem_flag = 0;
    }

    /* Process all of the parameters */
    if (process_parameters(parameters) != SUCCESS)
    {
        IAS_LOG_ERROR("Processing the parameters");
        return ERROR;
    }

    return SUCCESS;
}

/*****************************************************************************
 NAME: read_parameters

 PURPOSE: Read parameters from the specified input ODL file.

 RETURN VALUE: Type = int
    Value    Description
    -------  --------------------------
    SUCCESS  Successful completion
    ERROR    Operation failed

*****************************************************************************/
int read_parameters
(
    const char *parameter_filename,         /* I: Parameter file name */
    ANGLE_PARAMETERS *parameters            /* O: Generate Chips parameters */
)
{
    int help_requested;                     /* Help request variable */
    int count;                              /* Table parameter count */
    int status;                             /* Routine return status */

    /* Zero the parameters */
    memset(parameters, 0, sizeof(ANGLE_PARAMETERS));

    /* Parameters table definition */
    IAS_PARM_DECLARE_TABLE(parameter_table, 9);

    /* -------------------------------------------------------------------- */
    /* Common parameters table definition for the parameter file */

    /* Get the work order ID */
    IAS_PARM_WORK_ORDER_ID(parameter_table, local_wo_id, sizeof(local_wo_id),
        1);

    /* Get processing pass. */
    IAS_PARM_PROCESSING_PASS(parameter_table, &local_processing_pass,
        sizeof(local_processing_pass), 0);

    IAS_PARM_BAND_LIST(parameter_table, IAS_PARM_REQUIRED, local_band_numbers, 
        sizeof(local_band_numbers), 1);

    const char *valid_angle_types[] = {"BOTH", "SATELLITE", "SOLAR"};
    const char *default_angle_type[] = {"BOTH"};
    IAS_PARM_ADD_STRING(parameter_table, ANGLE_GENERATION_TYPE,
        "Type of angles to generate. \"BOTH\" will generate both satellite and "
        "solar angles", IAS_PARM_REQUIRED, 3, valid_angle_types, 1, 
        default_angle_type, local_angle_type, sizeof(local_angle_type), 0);

    const char *valid_angle_models[] = {"RPC", "RIGOROUS"};
    const char *default_angle_model[] = {"RPC"};
    IAS_PARM_ADD_STRING(parameter_table, ANGLE_MODEL,
        "Angle model used to generate the satellite and solar angles. Rational "
        "Polynomial Coefficients (RPC) uses the coefficients from the angle "
        "coefficient file. Rigorous is a more precise model that will take "
        "longer to run",
        IAS_PARM_OPTIONAL, 2, valid_angle_models, 1, default_angle_model,
        local_angle_model, sizeof(local_angle_model), 0);

    const int default_sub_sample_factor[] = {1};
    IAS_PARM_ADD_INT(parameter_table, SUB_SAMPLE_FACTOR,
        "Defines the sub-sample factor used when calculating the satellite and "
        "solar angles", IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        1, 1, 50, 1, default_sub_sample_factor, &parameters->sub_sample_factor, 
        sizeof(parameters->sub_sample_factor), 0);

    const int default_read_dem_flag[] = {1};
    IAS_PARM_ADD_INT(parameter_table, USE_DEM_FLAG,
        "Flag to indicate if DEM should be used in angle generation",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        1, 0, 1, 1, default_read_dem_flag, &parameters->use_dem_flag, 
        sizeof(parameters->use_dem_flag), 0);

    const int default_fill_pixel[] = {0};
    int local_background;
    IAS_PARM_ADD_INT(parameter_table, BACKGROUND, "Value to use for the "
        "fill pixels. Input units used is degrees scaled by a factor of 100",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, SHRT_MIN, SHRT_MAX, 1, 
        default_fill_pixel, &local_background, sizeof(local_background), 0);

    const int default_output_as_l1g[] = {1};
    IAS_PARM_ADD_INT(parameter_table, OUTPUT_AS_L1G_FLAG,
        "Flag to indicate if image should be output in L1G format. 0 = Flat "
        "binary file with envi header, 1 = L1G format",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        1, 0, 1, 1, default_output_as_l1g, &parameters->output_as_l1g_flag, 
        sizeof(parameters->output_as_l1g_flag), 0);

    /* -------------------------------------------------------------------- */
    /* Read the parameter file */

    /* Calculate the number of items in the parameter file */
    count = IAS_PARM_GET_TABLE_SIZE(parameter_table);

    /* Call the ias_parm_provide_help routine to provide help for the
       parameters if requested; also verifies the ODL file exists */
    help_requested = ias_parm_provide_help(parameter_filename, parameter_table,
        count, IAS_INPUT_PARAMETERS);
    if (help_requested)
    {
        /* Either help was provided or an error occurred */
        if (help_requested == TRUE)
        {
            /* provide help for the OMF.  If it succeeds, it will exit.  If
               it returns, an error occurred, so return an error */
            read_omf_parameters(parameter_filename, parameters);
            return ERROR;
        }
        
        IAS_LOG_ERROR("Error occurred providing ODL help");
        return ERROR;
    }
  
    /* Read the parameter file */
    if (ias_parm_read(parameter_filename, parameter_table, count) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from: %s", parameter_filename);
        return ERROR;
    }
 
    /* Save the band count for later */
    local_band_count = IAS_PARM_GET_COUNT_READ(parameter_table, BAND_LIST);
    
    /* Convert the background to short value */
    parameters->background = (short)local_background;

    /* Build the OMF filename. */
    status = snprintf(local_omf_filename, sizeof(local_omf_filename), "%s.omf", 
        local_wo_id);
    if (status < 0 || status >= sizeof(local_omf_filename))
    {
        IAS_LOG_ERROR("Building OMF filename");
        return ERROR;
    }

    return read_omf_parameters(local_omf_filename, parameters);
}
