#!/bin/csh -f

################################################################################
# NAME: test_create_angle_bands
#
# PURPOSE: Set up the environment and run the create angle bands program and 
#          check that the results are correct.
#
################################################################################

# Set up the environment for the tests.
set ERROR = 1
set SUCCESS = 0

set verify_dir = "$IAS_UNIT_TEST_DATA_DIR/ias_base/gps/create_angle_bands"
set log = "test_create_angle_bands.log"
set test_cases = ("rpc_sat" "rpc_solar" "rpc_both" "rig_sat" "rig_solar" \
    "rig_both" "rpc_l1g_solar")
set application = "../create_angle_bands"

# Set the IAS_VERSION_OVERRIDE variable
setenv IAS_VERSION_OVERRIDE "Y"

# Clean up the log
echo "" > $log

# Replace the unit test directory in omf file
sed -e "s|VERIFY_DIR|$verify_dir|" template_tests.omf > tests.omf 

# Prepare all of the test odl files before testing
foreach test ($test_cases)
    # Replace the unit test directory in odl files
    sed -e "s|AAAAA|`echo $IAS_UNIT_TEST_DATA_DIR`|" odl/template_${test}.odl\
        > ${test}.odl 
end

############################## RPC SAT TEST ####################################
set org_file = "LC81650342013170LGN00_SAT.h5"
set verify_file = "LC81650342013170LGN00_SAT_V3.h5"
set cur_test = $test_cases[1]

echo "Running test ${cur_test}" | tee -a $log
${application} ${cur_test}.odl >> $log
if ( $status ) then
    echo "Simple view test ${cur_test} failed" | tee -a $log
    exit $ERROR
endif

echo "    - h5diff -q ${org_file} ${verify_dir}/${verify_file}"\
    | tee -a $log
h5diff -q ${org_file} ${verify_dir}/${verify_file} | tee -a $log
if ( $status ) then
    echo "        - ERROR: Files differ"
    exit $ERROR
endif

rm ${org_file}

############################## RPC SOLAR TEST ##################################
set org_file = "LC81650342013170LGN00_solar_B08.img"
set verify_file = "LC81650342013170LGN00_solar_B08_V2.img"
set cur_test = $test_cases[2]

echo "Running test ${cur_test}" | tee -a $log
${application} ${cur_test}.odl >> $log
if ( $status ) then
    echo "Simple view test ${cur_test} failed" | tee -a $log
    exit $ERROR
endif

# Verify the output images
echo "    - diff -q ${org_file} ${verify_dir}/${cur_test}_${verify_file}" \
    | tee -a $log
diff -q ${org_file} ${verify_dir}/${cur_test}_${verify_file}
if ($status ) then
    echo "        - ERROR: Files differ"
    exit $ERROR
endif
echo "    - diff -q ${org_file}.hdr ${verify_dir}/${cur_test}_" \
    "${org_file}.hdr" | tee -a $log
diff -q ${org_file}.hdr ${verify_dir}/${cur_test}_${org_file}.hdr
if ($status ) then
    echo "        - ERROR: Files differ" | tee -a $log
    exit $ERROR
endif

rm ${org_file}
rm ${org_file}.hdr

######################## RPC SAT & SOLAR TEST ##################################
set cur_test = $test_cases[3]

echo "Running test ${cur_test}" | tee -a $log
${application} ${cur_test}.odl >> $log
if ( $status ) then
    echo "Simple view test ${cur_test} failed" | tee -a $log
    exit $ERROR
endif

set ver = "_V2"
# Verify the solar and satellite images
foreach verify_file ("LC81650342013170LGN00_solar_B11" \
                     "LC81650342013170LGN00_sensor_B11")
    echo "    - diff -q ${verify_file}.img ${verify_dir}/${cur_test}_"\
        "${verify_file}${ver}.img" | tee -a $log
    diff -q ${verify_file}.img \
        ${verify_dir}/${cur_test}_${verify_file}${ver}.img
    if ($status ) then
        echo "        - ERROR: Files differ"
        exit $ERROR
    endif
    echo "    - diff -q ${verify_file}.img.hdr ${verify_dir}/${cur_test}_"\
        "${verify_file}.img.hdr" | tee -a $log
    diff -q ${verify_file}.img.hdr \
        ${verify_dir}/${cur_test}_${verify_file}.img.hdr
    if ($status ) then
        echo "        - ERROR: Files differ" | tee -a $log
        exit $ERROR
    endif

    rm ${verify_file}.img
    rm ${verify_file}.img.hdr
end

############################ RIGOROUS SAT TEST #################################
set verify_file = "LC81650342013170LGN00_sensor_B01.img"
set cur_test = $test_cases[4]

echo "Running test ${cur_test}" | tee -a $log
${application} ${cur_test}.odl >> $log
if ( $status ) then
    echo "Simple view test ${cur_test} failed" | tee -a $log
    exit $ERROR
endif

# Verify the output images
echo "    - diff -q ${verify_file} ${verify_dir}/${cur_test}_${verify_file}" \
    | tee -a $log
diff -q ${verify_file} ${verify_dir}/${cur_test}_${verify_file}
if ($status ) then
    echo "        - ERROR: Files differ"
    exit $ERROR
endif
echo "    - diff -q ${verify_file}.hdr ${verify_dir}/${cur_test}_"\
    "${verify_file}.hdr" | tee -a $log
diff -q ${verify_file}.hdr ${verify_dir}/${cur_test}_${verify_file}.hdr
if ($status ) then
    echo "        - ERROR: Files differ" | tee -a $log
    exit $ERROR
endif

rm ${verify_file}
rm ${verify_file}.hdr

########################### RIGOROUS SOLAR TEST ################################
set verify_file = "LC81650342013170LGN00_solar_B08.img"
set cur_test = $test_cases[5]

echo "Running test ${cur_test}" | tee -a $log
${application} ${cur_test}.odl >> $log
if ( $status ) then
    echo "Simple view test ${cur_test} failed" | tee -a $log
    exit $ERROR
endif

# Verify the output images
echo "    - diff -q ${verify_file} ${verify_dir}/${cur_test}_${verify_file}" \
    | tee -a $log
diff -q ${verify_file} ${verify_dir}/${cur_test}_${verify_file}
if ($status ) then
    echo "        - ERROR: Files differ"
    exit $ERROR
endif
echo "    - diff -q ${verify_file}.hdr ${verify_dir}/${cur_test}_"\
    "${verify_file}.hdr" | tee -a $log
diff -q ${verify_file}.hdr ${verify_dir}/${cur_test}_${verify_file}.hdr
if ($status ) then
    echo "        - ERROR: Files differ" | tee -a $log
    exit $ERROR
endif

rm ${verify_file}
rm ${verify_file}.hdr

###################### RIGOROUS SAT & SOLAR TEST  #############################
set cur_test = $test_cases[6]

echo "Running test ${cur_test}" | tee -a $log
${application} ${cur_test}.odl >> $log
if ( $status ) then
    echo "Simple view test ${cur_test} failed" | tee -a $log
    exit $ERROR
endif

# Verify the solar and satellite images
foreach verify_file ("LC81650342013170LGN00_solar_B11.img" \
                     "LC81650342013170LGN00_sensor_B11.img")
    echo "    - diff -q ${verify_file} ${verify_dir}/${cur_test}_"\
        "${verify_file}" | tee -a $log
    diff -q ${verify_file} ${verify_dir}/${cur_test}_${verify_file}
    if ($status ) then
        echo "        - ERROR: Files differ"
        exit $ERROR
    endif
    echo "    - diff -q ${verify_file}.hdr ${verify_dir}/${cur_test}_"\
        "${verify_file}.hdr" | tee -a $log
    diff -q ${verify_file}.hdr ${verify_dir}/${cur_test}_${verify_file}.hdr
    if ($status ) then
        echo "        - ERROR: Files differ" | tee -a $log
        exit $ERROR
    endif

    rm ${verify_file}
    rm ${verify_file}.hdr
end

########################## RPC L1G SOLAR TEST ##################################
set org_file = "LC81650342013170LGN00_SUN.h5"
set verify_file = "LC81650342013170LGN00_SUN_V2.h5"
set cur_test = $test_cases[7]

echo "Running test ${cur_test}" | tee -a $log
${application} ${cur_test}.odl >> $log
if ( $status ) then
    echo "Simple view test ${cur_test} failed" | tee -a $log
    exit $ERROR
endif

echo "    - h5diff -q ${org_file} ${verify_dir}/${verify_file}" | tee -a $log
h5diff -q ${org_file} ${verify_dir}/${verify_file} | tee -a $log
if ( $status ) then
    echo "        - ERROR: Files differ"
    exit $ERROR
endif

set generated_omf = "tests.omf"
echo "    - grep -q SUN_ANGLE_FILENAME_PASS_2 ="\
    "LC81650342013170LGN00_SUN.h5 $generated_omf"
grep -q 'SUN_ANGLE_FILENAME_PASS_2 = "LC81650342013170LGN00_SUN.h5"' \
    $generated_omf
if ( $status ) then
    echo "        - ERROR: \'SUN_ANGLE_FILENAME_PASS_2 = "\
        "\"LC81650342013170LGN00_SUN.h5\" not found in $generated_omf"
    exit $ERROR
endif

rm ${org_file}

########################### TESTS END ##########################################

# Tests have completed successfully 
echo "Tests have completed Successfully" | tee -a $log
echo "" | tee -a $log


exit $SUCCESS
