/******************************************************************************
NAME:           angle_to_map

PURPOSE:
This function will convert the WRS rotation angle (from geodetic north) to
a frame orientation angle in map coordinates.

RETURN VALUE:
Type = int (SUCCESS or ERROR)
******************************************************************************/
#include <math.h>

#include "ias_math.h"                 /* Math library */
#include "ias_geo.h"                  /* Geometric library */
#include "ias_logging.h"              /* Logging library */
#include "create_grid.h"

int angle_to_map
(
    IAS_GRID_TYPE *grid   /* I/O: Grid structure to examine/modify */
)
{
    int status;                 /* Return status from library functions */
    double azimuth;             /* Grid azimuth of geodetic north */
    double sclat, sclong;       /* WRS scene center in lat/long */
    double x1, y1;              /* Scene center in projection coords */
    double x2, y2;              /* Projection center after adding 1 
                                   microradian to the SC latitude. */
    const IAS_GEO_PROJ_TRANSFORMATION *geo2proj;/* projection transformation */

    geo2proj = get_proj_transform_degrees2proj();

    /* Give initial values to variables needed for processing. */
    sclat = grid->center_pt.lat;
    sclong = grid->center_pt.lng;

    /* Retrieve the projection coordinates of scene center.  */
    x1 = grid->proj_center_pt.x;
    y1 = grid->proj_center_pt.y;

    /* Add one microradian (0.2 arc seconds) to the WRS SC latitude and find 
       the projection coordinates, storing the results in x2 and y2.        */
    sclat += ((1.0 / 3600.0) * 0.2);

    status = ias_geo_transform_coordinate(geo2proj, sclong, sclat, &x2, &y2);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Transforming coordinate");
        return ERROR;
    }

    /* Now we have two slightly-differing scene centers in projection
       coordinates.  Compute the azimuth of the line that they form as
       the arctangent of (x2 - x1) / (y2 - y1).  This will represent the
       grid azimuth of geodetic north at the WRS scene center.             */
    azimuth = atan((x2 - x1) / (y2 - y1));
    /* Convert from radians to degrees */
    azimuth *= ias_math_get_degrees_per_radian();

    /* Finally, we add this angle to the WRS rotation angle to get the
       grid heading.  A standard framed scene puts the satellite direction
       of flight at the bottom of scene, so the scene orientation is the
       grid heading either + or - 180 degrees.  If:

       the grid heading is < 0, subtract 180 degrees.
       the grid heading is > 0, add 180 degrees.                             */
    grid->orient_angle = grid->rotate_angle + azimuth;
    if (grid->orient_angle < 0)
        grid->orient_angle -= 180.0;
    else if (grid->orient_angle >= 0)
        grid->orient_angle += 180.0;

    return SUCCESS;
}
