/******************************************************************************
NAME: build_grid

PURPOSE: Provides the main driver for calculating the per-band/SCA data in the
    grid.

******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "gctp.h"
#include "ias_threadpool.h"
#include "ias_logging.h"
#include "create_grid.h"

/* Define the structure to hold the parameters provided to the processing
   threads */
typedef struct params
{
    struct ias_threadpool *threadpool;/* Threadpool being used */
    const IAS_LOS_MODEL *model;       /* Model structure */
    IAS_GRID_TYPE *grid;              /* Grid structure */
    DET_OFFSET_STATS *det_offset_stats;/* Detector offset stats */
    double poly_x[IAS_MAX_NBANDS][4]; /* Per-band array of x polynomial
                                         coefficients */
    double poly_y[IAS_MAX_NBANDS][4]; /* Per-band array of y polynomial
                                         coefficients */
    double lunar_scale;         /* Scale used for pixel size where acquisition 
                                   type is lunar or stellar */
    int scas_remaining[IAS_MAX_NBANDS]; /* Number of SCAs left to generate
                                    for each band.  Used to determine when the
                                    bands can be written to the file. */
    int blank[IAS_MAX_NBANDS][IAS_MAX_NSCAS]; /* Flag to specify on a per-band,
                                   per-SCA basis if the output image is blank.
                                   If one input pixel maps to an output pixel,
                                   then the image is not blank */
    int thread_error;           /* Non-zero if any threads encountered an
                                   error */
    IAS_GRID_FILE *grid_file;   /* Grid file to write to */
} PARAMS;

/* Define the mutex for serializing access to the work assignment tracking
   structure */
static IAS_THREAD_MUTEX_TYPE work_assignment_mutex;

/* Define the structure to hold parameters to the write_grid_band routine */
typedef struct write_params
{
    int *scas_remaining;        /* Pointer to the per-band array of SCA
                                   remaining counts */
    int band_index;             /* band index to write */
    IAS_GRID_FILE *grid_file;   /* Grid file to write */
    IAS_GRID_TYPE *grid;        /* Grid structure */
} WRITE_PARAMS;

/******************************************************************************
Name: write_grid_band

Purpose: Keeps track of the SCAs completed for each band and writes the band
    to the grid file when all the SCAs have been completed.  This is called
    through the threadpool work queue and is always running on the same thread.
    So no locking is required when writing the file.

Returns:
    SUCCESS or ERROR
******************************************************************************/
static int write_grid_band
(
    void *message       /* I: input message from the work queue */
)
{
    WRITE_PARAMS *params = (WRITE_PARAMS *)message;

    /* Decrement the count of SCAs remaining for this band and if it hits zero,
       write the grid band to the file */
    params->scas_remaining[params->band_index]--;
    if (params->scas_remaining[params->band_index] == 0)
    {
        int band_number;

        /* convert the band index to a band number */
        band_number = ias_sat_attr_convert_band_index_to_number(
                        params->band_index);
        if (band_number ==  ERROR)
        {
            IAS_LOG_ERROR("Unable to convert band index %d to a band number",
                params->band_index);
            free(params);
            return ERROR;
        }

        IAS_LOG_INFO("Writing grid band number %d", band_number);

        /* Write grid band pointers */
        if (ias_grid_band_pointers_write(params->grid_file, params->grid,
            band_number) != SUCCESS)
        {
            IAS_LOG_ERROR("Writing grid band pointers to %s",
                params->grid->gridname);
            free(params);
            return ERROR;
        }

        /* Free the grid band memory */
        if (ias_grid_free_band(params->grid, band_number) != SUCCESS)
        {
            IAS_LOG_ERROR("Freeing memory for band number %d",
                band_number);
            free(params);
            return ERROR;
        }
    }

    free(params);
    return SUCCESS;
}

/******************************************************************************
Name: get_assigned_band_and_sca

Purpose: Gets the band and SCA index that should be processed next.

Returns:
    SUCCESS or ERROR
******************************************************************************/
static int get_assigned_band_and_sca
(
    const IAS_GRID_TYPE *grid,  /* I: grid pointer */
    int *assigned_band_index,   /* O: band index assigned */
    int *assigned_sca_index,    /* O: SCA index assigned */
    int *work_assigned          /* O: flag to indicate work was assigned */
)
{
    static int band_index = 0;  /* next band to assign */
    static int sca_index = 0;   /* next SCA to assign */

    /* Default to no work being assigned */
    *work_assigned = 0;

    /* Lock the work assignment mutex to protect against multiple threads
       calling this at the same time */
    if (IAS_THREAD_LOCK_MUTEX(&work_assignment_mutex) != 0)
    {
        IAS_LOG_ERROR("Locking work assignment mutex" );
        return ERROR;
    }

    /* Loop through the bands looking for one to assign */
    while (band_index < grid->nbands)
    {
        /* Only consider bands that are available */
        if (grid->bands_available[band_index])
        {
            while (sca_index < grid->scas_per_band[band_index])
            {
                /* Assign the current band and SCA */
                *assigned_band_index = band_index;
                *assigned_sca_index = sca_index;

                /* Increment the SCA index for the next caller */
                sca_index++;

                /* Unlock the work assignment mutex since sca_index and
                   band_index won't be changed anymore */
                if (IAS_THREAD_UNLOCK_MUTEX(&work_assignment_mutex) != 0)
                {
                    IAS_LOG_ERROR("Unlocking work assignment mutex" );
                    return ERROR;
                }

                *work_assigned = 1;

                return SUCCESS;
            }

            /* Reset the SCA index for the next band */
            sca_index = 0;
        }

        band_index++;
    }

    /* No work to assign, so just unlock the mutex */
    if (IAS_THREAD_UNLOCK_MUTEX(&work_assignment_mutex) != 0)
    {
        IAS_LOG_ERROR("Unlocking work assignment mutex" );
        return ERROR;
    }

    /* No work was assigned, but that is still considered successful */
    return SUCCESS;
}

/******************************************************************************
Name: build_grid_thread

Purpose: Calculates the per-band/SCA members of the grid.  Multiple instances
    of this routine are run in different threads since it is a CPU bound 
    operation.

Returns:
    SUCCESS or ERROR
******************************************************************************/
static int build_grid_thread
(
    void *params_in,/* I: pointer to input parameter structure */
    int thread_num  /* I: thread number for the thread running this routine */
)
{
    PARAMS *params = (PARAMS *)params_in;/* input parameter structure pointer */
    int band_index;           /* current band index */
    int band_number;          /* current band number */
    int sca_index;            /* current SCA index */
    int work_assigned;        /* flag that indicates work was assigned */

    /* Loop forever - when no more work is available, the loop will exit */
    while (1)
    {
        DET_OFFSET_STATS *stats = NULL;
        WRITE_PARAMS *write_params = NULL;

        /* Get the next band and SCA to work on */
        if (get_assigned_band_and_sca(params->grid, &band_index, &sca_index,
                &work_assigned) != SUCCESS)
        {
            IAS_LOG_ERROR("Getting assigned band and SCA");
            params->thread_error = thread_num + 1;
            return ERROR;
        }

        /* If no work was assigned, exit the loop */
        if (!work_assigned)
            break;

        /* If any threads have encountered errors, don't bother doing any
           more processing */
        if (ias_threadpool_get_worker_return_status(params->threadpool)
            != SUCCESS)
        {
            IAS_LOG_WARNING("Exiting thread number %d since another thread "
                "encountered an error", thread_num);
            return ERROR;
        }

        /* Get the band number for the index */
        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Unable to convert band index %d to a band number",
                band_index);
            params->thread_error = thread_num + 1;
            return ERROR;
        }

        /* Issue a progress message for each new band started */
        if (sca_index == 0)
        {
            IAS_LOG_INFO("Processing band number %d", band_number);
        }

        /* Calculate the output line/sample locations and jitter sensitivity */
        if (calc_mapping_and_sensitivity(band_index, sca_index, params->model,
                params->poly_x[band_index], params->poly_y[band_index],
                params->lunar_scale, params->grid, 
                &params->blank[band_index][sca_index]) != SUCCESS)
        {
            IAS_LOG_ERROR("Making the grid in output space");
            params->thread_error = thread_num + 1;
            return ERROR;
        }

        /* Calculate the mapping coefficients */
        if (calc_map_coefs(&params->grid->gridbands[band_index][sca_index])
            != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating mapping coefficients for band number "
                "%d, SCA number %d", band_number, sca_index + 1);
            params->thread_error = thread_num + 1;
            return ERROR;
        }

        /* Calculate the rough mapping coefficients */
        if (ias_grid_compute_rough_polynomial(
                &params->grid->gridbands[band_index][sca_index]) != SUCCESS)
        {
            IAS_LOG_ERROR("Computing rough polynomial for band number %d, "
                "SCA number %d", band_number, sca_index + 1);
            params->thread_error = thread_num + 1;
            return ERROR;
        }

        /* Calculate the detector offsets */
        if (calc_det_offsets(band_index, sca_index, params->model,
                params->poly_x[band_index], params->poly_y[band_index],
                params->lunar_scale, params->grid) != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating detector offset values for band "
                "number %d, SCA number %d", band_number, sca_index + 1);
            params->thread_error = thread_num + 1;
            return ERROR;
        }

        /* Calculate the detector offset statistics for this band */
        stats = &params->det_offset_stats[band_index * IAS_MAX_NSCAS
                    + sca_index];
        if (calc_det_offset_stats(
                &params->grid->gridbands[band_index][sca_index], stats)
                != SUCCESS)
        {
            IAS_LOG_ERROR("Failed to calculate detector offset statistics for "
                "band number %d, SCA number %d", band_number, sca_index + 1);
            params->thread_error = thread_num + 1;
            return ERROR;
        }
        stats->band_index = band_index;
        stats->sca_index = sca_index;

        /* Allocate the parameters for the message to write_grid_band */
        write_params = malloc(sizeof(*write_params));
        if (!write_params)
        {
            IAS_LOG_ERROR("Failed allocating memory for write grid message");
            params->thread_error = thread_num + 1;
            return ERROR;
        }

        /* Let the write grid routine know that this band/SCA is complete by
           inserting a call to write_grid_band via the work queue (which then
           gets executed on the main thread) */
        write_params->band_index = band_index;
        write_params->scas_remaining = params->scas_remaining;
        write_params->grid_file = params->grid_file;
        write_params->grid = params->grid;
        if (ias_threadpool_add_to_work_queue(params->threadpool,
            write_grid_band, write_params) != SUCCESS)
        {
            IAS_LOG_ERROR("Failed inserting write grid call into work queue");
            free(write_params);
            params->thread_error = thread_num + 1;
            return ERROR;
        }
    }

    return SUCCESS;
}


/******************************************************************************
Name: check_blank

Purpose: For each band, determine whether the band is blank (no grid pixels 
    fell within the output image), and if so issue a warning message.  If all 
    bands are blank, issue an error message and exit with an ERROR status.

Returns:
    SUCCESS or ERROR
******************************************************************************/
static int check_blank
(
    IAS_GRID_TYPE *grid, /* I: Grid structure */
    PARAMS params        /* I: Parameters to pass to the threads */
)
{
    int band_number;    /* Number for current band */ 
    int band_index;     /* Index of band_number */ 
    int sca_index;      /* Index of current SCA */ 
    int band_blank;     /* Flag to specify whether the current band is blank */
    int all_blank = 1;  /* Flag to specify whether all bands are blank */

    /* Loop through the bands to check if any grid pixels fell within the 
       output image */
    for (band_index = 0; band_index < grid->nbands; band_index++)
    {
        /* Only consider bands that are available */
        if (grid->bands_available[band_index])
        {
            /* Initialize the status for the next band */
            band_blank = 1;

            /* Check all SCAs for the current band */
            for (sca_index = 0; sca_index < grid->scas_per_band[band_index];
                sca_index++)
            {
                /* Check the blank status */
                if (params.blank[band_index][sca_index] == 0)
                {
                    /* This band isn't blank */
                    band_blank = 0;
                    break;
                }
            }

            /* If no grid pixels fell within the output image for this band,
               then issue a warning message */
            if (band_blank)
            { 
                /* Get the band number for the index */
                band_number = ias_sat_attr_convert_band_index_to_number(
                    band_index);
                if (band_number == ERROR)
                {
                    IAS_LOG_ERROR("Unable to convert band index %d to a band "
                        "number", band_index);
                    return ERROR;
                }

                IAS_LOG_WARNING("No pixels for band %d fall within the "
                    "output image.", band_number);
            }
            else
            { 
                /* This band was not blank, so they can't all be blank */
                all_blank = 0;
            }
        }
    }

    /* If no grid pixels fell within the output image for any bands, then exit
       with an error. */
    if (all_blank)
    {
        IAS_LOG_ERROR("No pixels for any band fall within the output image.");
        return ERROR;
    }

    return SUCCESS;
}

/******************************************************************************
Name: build_grid

Purpose: Driving routine for calculating the per-band/SCA members of the grid.
    A threadpool is created and the calculations are done by the
    build_grid_thread routine.

Returns:
    SUCCESS or ERROR
******************************************************************************/
int build_grid
(
    const IAS_LOS_MODEL *model, /* I: Model structure */
    double poly_x[IAS_MAX_NBANDS][4], /* I: Per-band array of x polynomial
                                            coefficients */
    double poly_y[IAS_MAX_NBANDS][4], /* I: Per-band array of y polynomial
                                            coefficients */
    int max_processors_to_use,  /* I: Maximum number of processors to use */
    IAS_GRID_TYPE *grid,        /* I/O: Grid structure */
    DET_OFFSET_STATS *det_offset_stats,/* O: detector offset stats as 1D array
                                     IAS_MAX_NBANDS * IAS_MAX_NSCAS in size */
    IAS_GRID_FILE *grid_file    /* I: pointer to open grid to write to */
)
{
    PARAMS params;      /* parameters to pass to the threads */
    int band_index;
    int sca_index;
    int status;
    int procs_to_use = max_processors_to_use;

    /* Set up the parameters */
    params.model = model;
    params.grid = grid;
    params.det_offset_stats = det_offset_stats;
    memcpy(&params.poly_x, poly_x, sizeof(params.poly_x));
    memcpy(&params.poly_y, poly_y, sizeof(params.poly_y));
    params.thread_error = 0;
    params.grid_file = grid_file;
    for (band_index = 0; band_index < grid->nbands; band_index++)
        params.scas_remaining[band_index] = grid->scas_per_band[band_index];
    for (band_index = 0; band_index < grid->nbands; band_index++)
    {
        for (sca_index = 0; sca_index < grid->scas_per_band[band_index]; 
            sca_index++)
        {
            params.blank[band_index][sca_index] = 1; 
        }
    }

    /* Get the scaling factor for lunar/stellar collections */
    if (model->acquisition_type == IAS_STELLAR || 
        model->acquisition_type == IAS_LUNAR)
    {
        int out_units;

        if (ias_geo_get_units(grid->units, &out_units) != SUCCESS)
        {
            IAS_LOG_ERROR("Converting projection units %s to unit code",
                grid->units);
            return ERROR;
        }

        if (out_units == RADIAN) 
            params.lunar_scale = 1.0;
        else if (out_units == SECOND)
            params.lunar_scale = ias_math_get_arcsec_to_radian_conversion();
        else if (out_units == DEGREE)
            params.lunar_scale = ias_math_get_radians_per_degree();
        else 
        {
            IAS_LOG_ERROR("Invalid projection units %s", grid->units);
            return ERROR;
        }
    }

    /* Create the work assignment mutex */
    if (IAS_THREAD_CREATE_MUTEX(&work_assignment_mutex) != 0)
    {
        IAS_LOG_ERROR("Creating work assignment mutex");
        return ERROR;
    }

    /* The NOVAS library used by the lunar projection is not thread-safe, so 
       limit to using a single thread for lunar.  An attempt was made
       to use locks around calls to the NOVAS library, but it didn't work
       since the library is riddled with static variables that maintain some
       state information between calls. */
    if (model->acquisition_type == IAS_LUNAR)
    {
        if (procs_to_use > 1)
            procs_to_use = 1;
    }

    /* If using more than one processor, make sure only thread-safe transforms
       are used */
    if (procs_to_use > 1)
        ias_geo_only_allow_threadsafe_transforms();

    /* Initialize the threadpool */
    params.threadpool = ias_threadpool_initialize(procs_to_use);
    if (!params.threadpool)
    {
        IAS_LOG_ERROR("Initializing the thread pool");
        IAS_THREAD_DESTROY_MUTEX(&work_assignment_mutex);
        return ERROR;
    }

    /* Enable the work queue on the threadpool to write the grid bands as they
       complete */
    if (ias_threadpool_enable_work_queue(params.threadpool) != SUCCESS)
    {
        IAS_LOG_ERROR("Enabling the thread pool work queue");
        IAS_THREAD_DESTROY_MUTEX(&work_assignment_mutex);
        ias_threadpool_destroy(params.threadpool);
        return ERROR;
    }

    /* Run the build_grid_thread routine using the threadpool */
    status = ias_threadpool_run_function(params.threadpool, build_grid_thread,
                                         &params);
    if ((status == ERROR) || (params.thread_error > 0))
    {
        IAS_LOG_ERROR("Error encountered while making the grid");
        IAS_THREAD_DESTROY_MUTEX(&work_assignment_mutex);
        ias_threadpool_destroy(params.threadpool);
        return ERROR;
    }

    /* Clean up the threadpool */
    ias_threadpool_destroy(params.threadpool);

    if (IAS_THREAD_DESTROY_MUTEX(&work_assignment_mutex) != 0)
    {
        IAS_LOG_ERROR("Destroying work assignment mutex");
        return ERROR;
    }

    /* Check if bands are blank */
    status = check_blank(grid, params);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Failed check for non-blank bands");
        return ERROR;
    }

    return SUCCESS;
}

