/******************************************************************************
NAME:       calc_center_and_rotation_angle

PURPOSE:
Given a path, row, and various entries from the CPF, this function will 
calculate the scene center in lat/long and the rotation angle for a nominal
WRS scene.  Earth_Constants and Orbit_Parameters hold the required CPF data.

RETURN VALUE:
Type = int (SUCCESS or ERROR)
******************************************************************************/
#include <math.h>

#include "ias_math.h"                 /* Math Library */
#include "ias_logging.h"              /* Logging library */
#include "create_grid.h"

int calc_center_and_rotation_angle
(
    int path,            /* I: WRS path number */
    double row,          /* I: WRS row number (may be fractional) */
    IAS_CPF *cpf,         /* I: Calibration parameters */
    IAS_GRID_TYPE *grid  /* I/O: Grid structure */
)
{
    double inclination;      /* Inclination angle in radians */
    double long_path1_row60; /* Longitude of path 1, row 60 */
    double sc_ang_rate;      /* Angular velocity of spacecraft */
    double central_angle;    /* Central travel angle for this row */
    double wrs_gclat;        /* WRS scene center geocentric latitude */
    double long_origin;      /* Longitude of row 60 for given path */
    double delta_long;       /* Longitude offset due to central ang. */
    double wrs_long;         /* WRS scene center longitude */
    double wrs_lat;          /* WRS scene center latitude */
    double heading_angle;    /* Scene center heading from north */
    double pi;               /* Calculated value of pi */
    double pi2;              /* Calculated value of pi*2 */
    double degrees_per_radian; /* Calculated value of degrees per radian */
    double earth_spin_rate;  /* calculated value of earth spin rate */
    const struct IAS_CPF_ORBIT_PARAMETERS *cpf_orbit;
    const struct IAS_CPF_EARTH_CONSTANTS *cpf_earth;


    /* first calculate constants needed */
    pi = ias_math_get_pi();
    pi2 = pi * 2.0;
    degrees_per_radian = ias_math_get_degrees_per_radian();

   /* Define the Earth's angular rotation rate.  We use the solar
      rotation rate rather than the sidereal rate as called for in the
      General Electric Landsat D PIR which describes the Worldwide
      Reference System.  This is to account for the orbital precession
      which is designed to make the orbit sun-synchronous.               */
    earth_spin_rate = pi2 / IAS_SEC_PER_DAY;

    /* get the CPF groups needed */
    cpf_orbit = ias_cpf_get_orbit(cpf);
    if (!cpf_orbit)
    {
        IAS_LOG_ERROR("Getting Orbit parameter group from CPF");
        return ERROR;
    }
    cpf_earth = ias_cpf_get_earth_const(cpf);
    if (!cpf_earth)
    {
        IAS_LOG_ERROR("Getting Earth constants group from CPF");
        return ERROR;
    }

    /* Convert input angles to radians.   */
    inclination = ias_math_get_radians_per_degree() 
                  * cpf_orbit->inclination_angle;
    long_path1_row60 = ias_math_get_radians_per_degree()
                       * cpf_orbit->long_path1_row60;

    /* Compute the spacecraft's angular rotation rate.   */
    sc_ang_rate = pi2 * (double) cpf_orbit->wrs_cycle_orbits /
        (double) (cpf_orbit->wrs_cycle_days * IAS_SEC_PER_DAY);

    /* Compute the central travel angle for this row.   */
    central_angle = pi2 * (row -
            (double) cpf_orbit->descending_node_row) /
        (double) cpf_orbit->scenes_per_orbit;

    /* Compute the WRS geocentric latitude.   */
    wrs_gclat = asin(-sin(central_angle) * sin(inclination));

    /* Compute the longitude of row 60 for this path.   */
    long_origin = long_path1_row60 - pi2 * (double) (path - 1) /
        (double) cpf_orbit->wrs_cycle_orbits;

    /* Compute the WRS longitude.   */
    delta_long = atan2(tan(wrs_gclat) / tan(inclination),
            cos(central_angle) / cos(wrs_gclat));
    wrs_long = long_origin - delta_long - central_angle * earth_spin_rate /
        sc_ang_rate;

    /* Ensure the longitude is in the range +/- pi.   */
    while (wrs_long > pi)
        wrs_long -= pi2;
    while (wrs_long < -pi)
        wrs_long += pi2;

    /* Compute the scene heading angle.   */
    heading_angle = atan2(cos(inclination) / cos(wrs_gclat),
            -cos(delta_long) * sin(inclination));

    /* Convert the WRS geocentric latitude to geodetic latitude.   */
    wrs_lat = atan(tan(wrs_gclat) * 
            (cpf_earth->semi_major_axis /
             cpf_earth->semi_minor_axis) *
            (cpf_earth->semi_major_axis /
             cpf_earth->semi_minor_axis));

    /* Convert all computed angles from radians to degrees.   */
    wrs_lat *= degrees_per_radian;
    wrs_long *= degrees_per_radian;
    heading_angle *= degrees_per_radian;

    /* Round WRS scene center lat/long to nearest whole arc minute.   */
    wrs_lat = floor(wrs_lat * 60.0 + 0.5) / 60.0;
    wrs_long = floor(wrs_long * 60.0 + 0.5) / 60.0;

    /* Give our results to the caller.   */
    grid->center_pt.lat = wrs_lat;
    grid->center_pt.lng = wrs_long;
    grid->rotate_angle = heading_angle + 180;

    return SUCCESS;
}
