/******************************************************************************
NAME:           calc_corner_point_output_space_locations

PURPOSE:
Calculate the locations of the 4 corner points in output space. 

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed
******************************************************************************/
#include "ias_geo.h"                  /* Geometric library */
#include "ias_logging.h"              /* Logging library */
#include "create_grid.h"

int calc_corner_point_output_space_locations
(
    const IAS_DBL_LS *ul_ls, /* I: Upper left line/sample for the corners */
    const IAS_DBL_LS *ur_ls, /* I: Upper right line/sample for the corners */
    const IAS_DBL_LS *ll_ls, /* I: Lower left line/sample for the corners */
    const IAS_DBL_LS *lr_ls, /* I: Lower right line/sample for the corners */
    int band_index,    /* I: Input band number */
    int sca_index,     /* I: SCA index */
    double elevation,  /* I: Target elevation (at line/sample) */
    const IAS_LOS_MODEL *model, /* I: LOS model information */
    IAS_GRID_TYPE *grid, /* I: Grid structure */
    IAS_DBL_XY *ul_xy, /* O: Upper left x/y for the corners */
    IAS_DBL_XY *ur_xy, /* O: Upper right x/y for the corners */
    IAS_DBL_XY *ll_xy, /* O: Lower left x/y for the corners */
    IAS_DBL_XY *lr_xy  /* O: Lower right x/y for the corners */
)
{
    int band_number;            /* Band number corresponding to band_index */
    struct IAS_DBL_LAT_LONG ul_ltlg; /* Upper left lat/long for the corners */
    struct IAS_DBL_LAT_LONG ur_ltlg; /* Upper right lat/long for the corners */
    struct IAS_DBL_LAT_LONG ll_ltlg; /* Lower left lat/long for the corners */
    struct IAS_DBL_LAT_LONG lr_ltlg; /* Lower right lat/long for the corners */
    IAS_SENSOR_DETECTOR_TYPE dettype = IAS_NOMINAL_DETECTOR;
    int status;                 /* Function return status */
    const IAS_GEO_PROJ_TRANSFORMATION *geo2proj;

    geo2proj = get_proj_transform_radians2proj();

    /* Get the band number. */
    band_number = ias_sat_attr_convert_band_index_to_number(band_index);
    if (band_number == ERROR)
    {
        IAS_LOG_ERROR("Invalid band number %d", band_number);
        return ERROR;
    }

    /* Find corner point latitude and longitude projection values. */
    status = ias_los_model_input_line_samp_to_geodetic(ul_ls->line, ul_ls->samp,
            band_index, sca_index, elevation, model, dettype, NULL,
            &ul_ltlg.lat, &ul_ltlg.lng);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating latitude and longitude for band %d, "
            "SCA %d, elevation %f upper left corner", band_number, 
            sca_index + 1, elevation);
        return ERROR;
    }

    status = ias_los_model_input_line_samp_to_geodetic(ur_ls->line, ur_ls->samp,
            band_index, sca_index, elevation, model, dettype, NULL,
            &ur_ltlg.lat, &ur_ltlg.lng);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating latitude and longitude for band %d, "
            "SCA %d, elevation %f upper right corner", band_number, 
            sca_index + 1, elevation);
        return ERROR;
    }

    status = ias_los_model_input_line_samp_to_geodetic(ll_ls->line, ll_ls->samp,
            band_index, sca_index, elevation, model, dettype, NULL,
            &ll_ltlg.lat, &ll_ltlg.lng);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating latitude and longitude for band %d, "
            "SCA %d, elevation %f lower left corner", band_number, 
            sca_index + 1, elevation);
        return ERROR;
    }

    status = ias_los_model_input_line_samp_to_geodetic(lr_ls->line, lr_ls->samp,
            band_index, sca_index, elevation, model, dettype, NULL,
            &lr_ltlg.lat, &lr_ltlg.lng);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating latitude and longitude for band %d, "
            "SCA %d, elevation %f lower right left corner", band_number, 
            sca_index + 1, elevation);
        return ERROR;
    }

    /* Calculate the location of the latitude and longitude in output space */
    status = ias_geo_transform_coordinate(geo2proj, ul_ltlg.lng, ul_ltlg.lat,
        &ul_xy->x, &ul_xy->y);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating location in output space for lat/long"
            "for band %d, SCA %d, elevation %f upper left corner", band_number, 
            sca_index + 1, elevation);
        return ERROR;
    }

    status = ias_geo_transform_coordinate(geo2proj, ur_ltlg.lng, ur_ltlg.lat,
        &ur_xy->x, &ur_xy->y);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating location in output space for lat/long "
            "for band %d, SCA %d, elevation %f upper right corner", 
            band_number, sca_index + 1, elevation);
        return ERROR;
    }

    status = ias_geo_transform_coordinate(geo2proj, ll_ltlg.lng, ll_ltlg.lat,
        &ll_xy->x, &ll_xy->y);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating location in output space for lat/long "
            "for band %d, SCA %d, elevation %f lower left corner", band_number, 
            sca_index + 1, elevation);
        return ERROR;
    }

    status = ias_geo_transform_coordinate(geo2proj, lr_ltlg.lng, lr_ltlg.lat,
        &lr_xy->x, &lr_xy->y);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating location in output space for lat/long "
            "for band %d, SCA %d, elevation %f lower right corner", 
            band_number, sca_index + 1, elevation);
        return ERROR;
    }

    return SUCCESS;
}
