/******************************************************************************
NAME:           calc_det_offsets

PURPOSE:
Computes the even/odd detector offset values.  Computes linear mapping
coefficients associated with detector offsets, and stores them in the grid.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

ALGORITHM REFERENCES:
See ias_grid.h for a description of the grid structure.
See ias_los_model.h for a description of the model structure.

******************************************************************************/
#include <math.h>
#include <string.h>                   /* String functions     */
#include "gctp.h"

#include "ias_math.h"                 /* Math library */
#include "ias_geo.h"                  /* Geometric library */
#include "ias_los_model.h"            /* LOS model library */
#include "ias_logging.h"              /* Logging library   */
#include "create_grid.h"

#define         CALC_DELAYS     1
#define         ODD_EVEN_TOL    0.05

/* Define a macro for switching an output message between an error and warning
   depending on the acquisition type.  There are some cases where something
   that is an error for Earth collects is only a warning for Lunar collections.
   If it is an error, the macro returns ERROR to the calling routine. */
#define OUTPUT_MESSAGE(format,...) \
{   \
    int log_level; \
    if (model->acquisition_type == IAS_EARTH) \
        log_level = IAS_LOG_LEVEL_ERROR; \
    else \
        log_level = IAS_LOG_LEVEL_WARN; \
    ias_log_message(log_level, __FILE__, __LINE__, format,##__VA_ARGS__); \
    if (model->acquisition_type == IAS_EARTH) \
        return ERROR; \
}

int calc_det_offsets 
(
    int band_index,             /* I: band index to process */
    int sca_index,              /* I: sca index to process */
    const IAS_LOS_MODEL *model, /* I: Model structure */
    double poly_x[4],           /* I: array of x polynomial coefficients */
    double poly_y[4],           /* I: array of y polynomial coefficients */
    double lunar_scale,         /* I: scale used for pixel size for 
                                      lunar/stellar acquisitions */
    IAS_GRID_TYPE *grid         /* I/O: Grid structure */
)
{
    int i, j;               /* Loop variables */
    int zplane;             /* Counter for elevation planes */
    double elev = 0.0;      /* Elevation at the current point */
    IAS_SENSOR_DETECTOR_TYPE dettype = IAS_MAXIMUM_DETECTOR;

    double *delta_line_oe;  /* Linear offset coefficients to be stored in 
                               grid */
    double *delta_samp_oe;  /* Linear offset coefficients to be stored in 
                               grid */
    double dline_even0;     /* Delta line, even sample, for elev=0 */
    double dline_odd0;      /* Delta line, odd sample, for elev=0 */
    double dsamp_even0;     /* Delta sample, even sample, for elev=0 */
    double dsamp_odd0;      /* Delta sample, odd sample, for elev=0 */
    double dline_evenH;     /* Delta line, even sample, for elev=MAX */
    double dline_oddH;      /* Delta line, odd sample, for elev=MAX */
    double dsamp_evenH;     /* Delta sample, even sample, for elev=MAX */
    double dsamp_oddH;      /* Delta sample, odd sample, for elev=MAX */

    double delta_line0;     /* Delta line (elev=0) scaled by pixels */
    double delta_lineH;     /* Delta line (elev=MAX) scaled by pixels */
    double delta_samp0;     /* Delta sample (elev=0) scaled by pixels */
    double delta_sampH;     /* Delta sample (elev=MAX) scaled by pixels */

    double current_line;    /* Current line for processing */
    double current_samp;    /* Current sample for processing */

    double maximum_detector_delay; /* Maximum detector delay (in IFOV) */
    IAS_LUNAR_PROJECTION *lunar_proj = NULL;


    /* Get the maximum detector delay from the model.  This is just to  
       make later code easier to read. */
    maximum_detector_delay = model->sensor.bands[band_index].
        sampling_char.maximum_detector_delay;

    if (model->acquisition_type == IAS_LUNAR)
    {
        /* Create the lunar projection for the current band/sca */
        lunar_proj = ias_los_model_create_lunar_projection(model,
                band_index, sca_index, dettype, lunar_scale, 1);
        if (!lunar_proj)
        {
            IAS_LOG_ERROR("Initializing lunar projection for band "
                    "index %d, SCA index %d", band_index, sca_index);
            return ERROR;
        }
    }

    delta_line_oe = grid->gridbands[band_index][sca_index].delta_line_oe;
    delta_samp_oe = grid->gridbands[band_index][sca_index].delta_samp_oe;

    /* Loop through all of the grid points to get the 
       latitude/longitude, latitude/longitude in output space, and 
       finally the line/sample in output space. Use the ACTUAL detector
       to do this.  */
    for (i = 0; i < grid->gridbands[band_index][sca_index].ngrid_lines; i++)
    {
        for (j = 0; j < grid->gridbands[band_index][sca_index].ngrid_samps; j++)
        {
            elev = 0.0;
            zplane = grid->gridbands[band_index][sca_index].zeroplane;

            current_line = grid->gridbands[band_index][sca_index].in_lines[i];
            current_samp = grid->gridbands[band_index][sca_index].in_samps[j];

            if (CALC_DELAYS)
            {
                if (calc_odd_or_even_offsets(model, lunar_proj, poly_x, poly_y,
                            band_index, sca_index, dettype, elev, lunar_scale,
                            current_line, current_samp, zplane, grid,
                            &dline_even0, &dsamp_even0, &dline_odd0,
                            &dsamp_odd0) != SUCCESS)
                {
                    IAS_LOG_ERROR("Mapping the detector back to input space");
                    return ERROR;
                }
            }
            else
            {
                dline_even0 = current_line;
                dsamp_even0 = current_samp;
                dline_odd0  = current_line;
                dsamp_odd0  = current_samp;
            }

            zplane = grid->gridbands[band_index][sca_index].nzplanes -1;
            elev = (zplane - grid->gridbands[band_index][sca_index].zeroplane)
                 * grid->gridbands[band_index][sca_index].zspacing;

            if (CALC_DELAYS)
            {
                if (calc_odd_or_even_offsets(model, lunar_proj, poly_x, poly_y,
                            band_index, sca_index, dettype, elev, lunar_scale, 
                            current_line, current_samp, zplane, grid,
                            &dline_evenH, &dsamp_evenH, &dline_oddH, 
                            &dsamp_oddH) != SUCCESS)
                {
                    IAS_LOG_ERROR("Mapping the detector back to input space");
                    return ERROR;
                }
            }
            else
            {
                dline_evenH = current_line;
                dsamp_evenH = current_samp;
                dline_oddH  = current_line;
                dsamp_oddH  = current_samp;
            }

            /* Solve for linear mapping coefficients associated with 
               detector delays. Store values into appropriate grid 
               array */
            delta_line0 = dline_odd0 / maximum_detector_delay;
            delta_samp0 = dsamp_odd0 / maximum_detector_delay;
            delta_lineH = dline_oddH / maximum_detector_delay;
            delta_sampH = dsamp_oddH / maximum_detector_delay;

            if (fabs(delta_line0 - (dline_even0 / 
                            maximum_detector_delay)) > ODD_EVEN_TOL)
            {
                /* Output a message that switches between an ERROR and a 
                   WARNING based on the collection type since lunar collects
                   consider this a warning while earth collects consider it
                   an error.  If it is an error, ERROR is returned, otherwise
                   it continues running*/
                OUTPUT_MESSAGE("Difference for elevation 0 coefficient %lf "
                    "for current line %lf exceeds tolerance limit %lf:  "
                    "observed difference is %lf",
                    delta_line0, current_line, ODD_EVEN_TOL,
                    fabs(delta_line0 - (dline_even0
                        / maximum_detector_delay)));
            }

            if (fabs(delta_samp0 - (dsamp_even0 / 
                            maximum_detector_delay)) > ODD_EVEN_TOL)
            {
                /* Output a message that switches between an ERROR and a 
                   WARNING based on the collection type since lunar collects
                   consider this a warning while earth collects consider it
                   an error.  If it is an error, ERROR is returned, otherwise
                   it continues running*/
                OUTPUT_MESSAGE("Difference for elevation 0 coefficient %lf "
                    "for current sample %lf exceeds tolerance limit %lf:  "
                    "observed difference is %lf",
                    delta_samp0, current_samp, ODD_EVEN_TOL,
                    fabs(delta_samp0 - (dsamp_even0
                        / maximum_detector_delay)));
            }

            if (fabs(delta_lineH - (dline_evenH / 
                            maximum_detector_delay)) > ODD_EVEN_TOL)
            {
                /* Output a message that switches between an ERROR and a 
                   WARNING based on the collection type since lunar collects
                   consider this a warning while earth collects consider it
                   an error.  If it is an error, ERROR is returned, otherwise
                   it continues running*/
                OUTPUT_MESSAGE("Difference for max. elevation coefficient %lf "
                    "for current line %lf exceeds tolerance limit %lf:  "
                    "observed difference is %lf",
                    delta_lineH, current_line, ODD_EVEN_TOL,
                    fabs(delta_lineH - (dline_evenH
                        / maximum_detector_delay)));
            }

            if (fabs(delta_sampH - (dsamp_evenH / 
                            maximum_detector_delay)) > ODD_EVEN_TOL)
            {
                /* Output a message that switches between an ERROR and a 
                   WARNING based on the collection type since lunar collects
                   consider this a warning while earth collects consider it
                   an error.  If it is an error, ERROR is returned, otherwise
                   it continues running*/
                OUTPUT_MESSAGE("Difference for max. elevation coefficient %lf "
                    "for current sample %lf exceeds tolerance limit %lf:  "
                    "observed difference is %lf",
                    delta_sampH, current_samp, ODD_EVEN_TOL,
                    fabs(delta_sampH - (dsamp_evenH
                        / maximum_detector_delay)));
            }

            /* Solve for line direction odd detector mapping
               coefficients */
            *delta_line_oe = delta_line0;
            delta_line_oe++;
            *delta_line_oe = (delta_lineH - delta_line0) / elev;
            delta_line_oe++;

            /* Solve for sample direction odd detector mapping
               coefficients */
            *delta_samp_oe = delta_samp0;
            delta_samp_oe++;
            *delta_samp_oe = (delta_sampH - delta_samp0) / elev;
            delta_samp_oe++;
        } /* ngrid_samps */
    } /* ngrid_lines */

    /* Free the lunar projection if it was created */
    if (lunar_proj)
    {
        ias_los_model_free_lunar_projection(lunar_proj);
        lunar_proj = NULL;
    }

    return SUCCESS;
}
