/******************************************************************************
NAME: calc_det_offsets_statistics

PURPOSE: Computes and prints the statistics on even/odd detector offset values.

******************************************************************************/
#include <stdlib.h>
#include <math.h>
#include "ias_math.h"                 /* Math library */
#include "ias_logging.h"              /* Logging library */
#include "create_grid.h"

#define     PRINT       0


/******************************************************************************
NAME: calc_det_offset_stats

PURPOSE: Computes the statistics on even/odd detector offset values for a
    single band/sca of the grid.

RETURNS:
    SUCCESS or ERROR

******************************************************************************/
int calc_det_offset_stats
(
    IAS_GRID_BAND_TYPE *gridband, /* I/O: Grid structure  */
    DET_OFFSET_STATS *stats       /* O: calculated statistics */
)
{
    int i, j, k;               /* Loop variables */
    int curr_pt;               /* Current point */
    int num;                   /* Number of grid points */
    double *line_oe;           /* Array of det offsets in line dir for odd or
                                  even detectors */
    double *samp_oe;           /* Array of det offsets in line dir for odd or
                                  even detectors */
    double sum_sqr_line_oe[2]; /* Sum of squares for detector offsets in line 
                                  dir for odd or even detectors */
    double sum_line_oe[2];     /* Sum of det offsets in line dir for odd or
                                  even detectors */
    double sum_sqr_samp_oe[2]; /* Sum of squares for detector offsets in sample
                                  dir for odd even or detectors */
    double sum_samp_oe[2];     /* Sum of det offsets in sample dir for odd or
                                  even detectors */
    double line_offset;        /* A single line detector offset from line_oe */
    double sample_offset;      /* A single sample detector offset from 
                                  samp_oe */
    double tempf;

    /* If printing of the stats hasn't been enabled, don't bother calculating
       them */
    if (!PRINT)
        return SUCCESS;

    /* Find the number of grid points for the band */
    num = gridband->ngrid_lines * gridband->ngrid_samps;
    if (num <= 0)
    {
        IAS_LOG_ERROR("In number of grid points");
        return ERROR;  
    }

    /* Allocate arrays for detector offsets */
    line_oe = (double *)malloc(sizeof(double) * 2 * num);
    if (line_oe == NULL)
    {
        IAS_LOG_ERROR("Allocating line_oe");
        return ERROR;  
    }
    samp_oe = (double *)malloc(sizeof(double) * 2 * num);
    if (samp_oe == NULL)
    {
        IAS_LOG_ERROR("Allocating samp_oe");
        free(line_oe);
        return ERROR;
    }

    /* Initialize summing variables */
    for (k = 0; k < 2; k++)
    {
        sum_sqr_line_oe[k]  = 0.0;
        sum_line_oe[k]      = 0.0;
        sum_sqr_samp_oe[k]  = 0.0;
        sum_samp_oe[k]      = 0.0;
    }

    for (i = 0, curr_pt = 0; i < gridband->ngrid_lines; i++)
    {
        for (j = 0; j < gridband->ngrid_samps; j++, curr_pt++)
        {
            for (k = 0; k < 2; k++ )
            {
                /* Use temporary variables for readability. */
                line_offset = gridband->delta_line_oe[2*curr_pt+k];
                sample_offset =  gridband->delta_samp_oe[2*curr_pt+k];

                /* Rearrange data into new buffers */
                line_oe[k*num+curr_pt] = line_offset;
                samp_oe[k*num+curr_pt] = sample_offset;

                /* Find sum of offsets */
                sum_line_oe[k] = sum_line_oe[k] + line_offset;
                sum_samp_oe[k] = sum_samp_oe[k] + sample_offset;

                /* Find sum of squares */
                sum_sqr_line_oe[k] = sum_sqr_line_oe[k]
                                   + line_offset * line_offset;
                sum_sqr_samp_oe[k] = sum_sqr_samp_oe[k]
                                   + sample_offset * sample_offset;
            }
        }
    }

    for (k = 0; k < 2; k++)
    {
        /* Calculate mean */
        stats->mean_line_oe[k] = sum_line_oe[k] / (double)num;
        stats->mean_samp_oe[k] = sum_samp_oe[k] / (double)num;

        /* Calculate RMS */
        stats->rms_line_oe[k] = sqrt(sum_sqr_line_oe[k] / (double)num);
        stats->rms_samp_oe[k] = sqrt(sum_sqr_samp_oe[k] / (double)num);

        /* Calculate standard deviation */
        tempf = sum_sqr_line_oe[k] - ((double)num * stats->mean_line_oe[k] 
                * stats->mean_line_oe[k]);
        if (tempf < 0.0)
            stats->stdev_line_oe[k] = 0.0;
        else
            stats->stdev_line_oe[k] = sqrt(tempf / ((double)num - 1.0));

        tempf = sum_sqr_samp_oe[k] - ((double)num * stats->mean_samp_oe[k] 
                * stats->mean_samp_oe[k]);

        if (tempf < 0.0)
            stats->stdev_samp_oe[k] = 0.0;
        else
            stats->stdev_samp_oe[k] = sqrt(tempf / ((double)num - 1.0));

        /* Sort elements */
        ias_math_heapsort_double_array(num, &line_oe[k*num]);
        ias_math_heapsort_double_array(num, &samp_oe[k*num]);

        /* Calculate median from sorted elements */
        stats->median_line_oe[k] = line_oe[k*num+num/2];
        stats->median_samp_oe[k] = samp_oe[k*num+num/2];

        /* Calculate maximum from sorted elements */
        stats->max_line_oe[k] = line_oe[k*num+num-1];
        stats->max_samp_oe[k] = samp_oe[k*num+num-1];

        /* Calculate minimum from sorted elements */
        stats->min_line_oe[k] = line_oe[k*num];
        stats->min_samp_oe[k] = samp_oe[k*num];
    } /* Odd/Even loop for stats */

    free(line_oe);
    free(samp_oe);

    /* Indicate the stats record is valid */
    stats->available = 1;

    return SUCCESS;
}

/******************************************************************************
NAME: print_det_offset_stats

PURPOSE: Prints the statistics computed by the calc_det_offset_stats routine.

RETURNS:
    SUCCESS or ERROR

******************************************************************************/
int print_det_offset_stats
(
    const DET_OFFSET_STATS *stats_array,/* I: statistics to print */
    int stats_size 
)
{
    int k;                     /* Loop variable */
    int index;
    FILE *fp;
    int status;                /* Function return status */

    /* If printing hasn't been enabled, just return */
    if (!PRINT)
        return SUCCESS;

    fp = fopen("detoffset.txt", "wb");
    if (!fp)
    {
        IAS_LOG_ERROR("Failed to open the detoffsets.txt file");
        return ERROR;
    }

    /* Loop through all detector offsets, printing the values to the output
       file */
    for (index = 0; index < stats_size; index++)
    {
        const DET_OFFSET_STATS *stats = &stats_array[index];

        /* Skip the records for bands that weren't processed */
        if (!stats->available)
            continue;

        fprintf(fp, "Band %d SCA %d\n", stats->band_index, stats->sca_index);

        for (k = 0; k < 2; k++)
        {
            if (k == 0)
                fprintf(fp," Bias\n");
            else
                fprintf(fp," Scale\n");
            fprintf(fp,"Minimum %e %e\n", stats->min_line_oe[k],
                                          stats->min_samp_oe[k]);

            fprintf(fp,"Maximum %e %e\n", stats->max_line_oe[k],
                                          stats->max_samp_oe[k]);

            fprintf(fp,"Median %e %e\n", stats->median_line_oe[k],
                                         stats->median_samp_oe[k]);

            fprintf(fp,"Mean %e %e\n", stats->mean_line_oe[k],
                                       stats->mean_samp_oe[k]);

            fprintf(fp,"Standard Deviation %e %e\n", 
                                        stats->stdev_line_oe[k],
                                        stats->stdev_samp_oe[k]);

            fprintf(fp,"RMS %e %e\n", stats->rms_line_oe[k],
                                      stats->rms_samp_oe[k]);
            fprintf(fp,"\n");
        } /* Odd/Even loop for stats */
    }

    status = fclose(fp);
    if (status != 0)
    {
        IAS_LOG_ERROR("Closing output file");
        return ERROR;
    }

    return SUCCESS;
}
