/******************************************************************************
NAME: calc_lunar_size

PURPOSE: Calculates the output size, in pixels, of a lunar acquisition.
    For a frame type of Lunar MAXBOX, max/min bounding area is found from the
    projected four corners of every band of every SCA.  For a frame type
    of Lunar MINBOX, max/min bounding area is found from only the SCA which
    contains the moon. The bounds is max/min of these projected values.

RETURNS: SUCCESS or ERROR

******************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <math.h>

#include "gctp.h"
#include "ias_los_model.h"
#include "ias_logging.h"
#include "create_grid.h"
#include "ias_types.h"

/* Define a point count of four to be used by both of the point arrays */
#define POINT_COUNT 4

/* Define locations for correctly positioning the point values into the
   polygon array.  ias_math_point_in_polygon requires the correct order */
#define UL_POINT 0
#define UR_POINT 1
#define LR_POINT 2
#define LL_POINT 3

int calc_lunar_size
(
    const IAS_LOS_MODEL *model,     /* I: model structure */
    IAS_GRID_TYPE *grid,            /* I/O: Grid structure */
    int *lunar_sca_number           /* O: SCA number containing the image
                                       of the moon */
)
{
    int band_index;
    int sca_index;
    IAS_SENSOR_DETECTOR_TYPE dettype
        = IAS_NOMINAL_DETECTOR; /* Detector placement */

    /* Max right ascension, declination, and distance */
    double max_rasc = -DBL_MAX;
    double max_dec = -DBL_MAX; 
    double max_dist = -DBL_MAX;
    /* Min right ascension, declination, and distance */
    double min_rasc = DBL_MAX;
    double min_dec = DBL_MAX;
    double min_dist = DBL_MAX;
    double scale;            /* Scale for converting pixel uints */
    double elev = 0.0;       /* Elevation at the current point */
    double in_lat;           /* Latitude in input space */
    double in_long;          /* Longitude in input space */
    double convert_to_degrees; /* Scale to allow conversion to degrees */
    double moon_right_asc;   /* Right Ascension of the center of the moon */
    double moon_declination; /* Declination of the center of the moon */

    /* Lunar variables */
    double rasc;             /* Apparent right ascension of the moon */
    double dec;              /* Declination of the moon */
    double sdist;            /* Scale of pixel due to change in distance
                                between moon and satellite */
    double d_lat;            /* Delta latitude between extremes */
    double d_long;           /* Delta longitude between extremes */
    int grid_units;          /* units from grid converted to an int */
    double max_pixsize = 0.0;/* largest pixel size in grid */
    int line_count;          /* lines in the computed frame */
    int sample_count;        /* samples in the computed frame */

    int found_moon_in_sca; /* To determine if the moon was found in the sca */
    int point_index;       /* Corner loop counter */

    /* Convert the grid projection units to a unit code for GCTP */
    if (ias_geo_get_units(grid->units, &grid_units) != SUCCESS)
    {
        IAS_LOG_ERROR("Converting projection units %s to unit code",
            grid->units);
        return ERROR;
    }

    /* Set the unit scaling factor */
    if (grid_units == RADIAN)
    {
        scale = 1.0;
    }
    else if (grid_units == SECOND)
    {
        scale = ias_math_get_arcsec_to_radian_conversion();
    }
    else if (grid_units == DEGREE)
    {
        scale = ias_math_get_radians_per_degree();
    }
    else
    {
        IAS_LOG_ERROR("Unsupported units: %s", grid->units);
        return ERROR;
    }

    /* Loop over all the bands to find the max extent of:
         IAS_LUNAR_MAXBOX: all the bands in the data
         IAS_LUNAR_MINBOX: all the SCAs that contain the moon in the data */
    for (band_index = 0; band_index < grid->nbands; band_index++)
    {
        int image_lines;
        double line[POINT_COUNT]; /* Four corner locations */

        /* Skip bands that are not available in the grid */
        if (!grid->bands_available[band_index])
        {
            continue;
        }

        /* Calculate the corner point line indices for this band */
        image_lines = model->sensor.bands[band_index].frame_count
            * model->sensor.bands[band_index].sampling_char.lines_per_frame;

        line[UL_POINT] = line[UR_POINT] = 0.0;
        line[LR_POINT] = line[LL_POINT] = image_lines - 1;

        found_moon_in_sca = FALSE;
        for (sca_index = 0; sca_index < grid->scas_per_band[band_index];
             sca_index++)
        {
            IAS_LUNAR_PROJECTION *lunar_proj;
            double samp[POINT_COUNT]; /* Four corner locations */

            /* Assign arrays as input to ias_math_point_in_polygon */
            /* The X polygon values specified as right-ascension */
            double polygon_rasc[POINT_COUNT];
            /* The Y polygon values specified as declination */
            double polygon_dec[POINT_COUNT];

            /* Initialize the lunar projection for this band/sca */
            lunar_proj = ias_los_model_create_lunar_projection(model,
                                band_index, sca_index, dettype, scale, 0);
            if (!lunar_proj)
            {
                IAS_LOG_ERROR("Initializing lunar projection "
                    "transformation for band index %d, SCA index %d",
                    band_index, sca_index);
                return ERROR;
            }

            /* Initialize sample locations for this band/SCA */
            samp[UL_POINT] = samp[LL_POINT] = 0.0;
            samp[UR_POINT] = samp[LR_POINT] =
                model->sensor.bands[band_index].scas[sca_index].detectors - 1;

            /* Loop on line and samples determining image extent */
            for (point_index = 0; point_index < POINT_COUNT; point_index++)
            {
                /* Current line/sample location */
                double cline;
                double csamp;

                /* Calculate the latitude and longitude of the current
                   grid point.  */
                cline = line[point_index];

                /* Note samples is already scaled to resolution. */
                csamp = samp[point_index];

                if (ias_los_model_input_line_samp_to_geodetic(cline, csamp,
                            band_index, sca_index, elev, model, dettype, NULL,
                            &in_lat, &in_long) != SUCCESS)
                {
                    IAS_LOG_ERROR("Calculating the geodetic location for "
                        "line %f, sample %f", cline, csamp);
                    ias_los_model_free_lunar_projection(lunar_proj);
                    lunar_proj = NULL;
                    return ERROR;
                }

                /* Find moon's position at current line/sample time */
                dec = in_lat;
                rasc = in_long;

                if (ias_los_model_transform_lunar_projection(lunar_proj,
                        cline, csamp, &dec, &rasc, &sdist) != SUCCESS)
                {
                    IAS_LOG_ERROR("Projecting transforming Lunar coordinates "
                        "for line %f, sample %f", cline, csamp);
                    ias_los_model_free_lunar_projection(lunar_proj);
                    lunar_proj = NULL;
                    return ERROR;
                }

                /* Find the min/max bounds for MAXBOX */
                if (grid->frame_type == IAS_LUNAR_MAXBOX)
                {
                    /* Min and max declination */
                    if (dec > max_dec)
                        max_dec  = dec;
                    if (dec < min_dec)
                        min_dec  = dec;

                    /* Min and max right-ascension */
                    if (rasc > max_rasc)
                        max_rasc = rasc;
                    if (rasc < min_rasc)
                        min_rasc = rasc;
                }
                else /* IAS_LUNAR_MINBOX */
                {
                    /* Calculated SCA corner points are stored within the
                       polygon arrays.  Later these bounding boxes will be
                       checked to see which bounding box contains the moon. */
                    polygon_rasc[point_index] = rasc;
                    polygon_dec[point_index] = dec;
                }

                /* Min and Max scale distance */
                if (sdist > max_dist)
                    max_dist  = sdist;
                if (sdist < min_dist)
                    min_dist  = sdist;
            } /* End point_index loop */

            /* For MINBOX we use the SCA which contains the moon.  So we
               search for the moon in each SCA and then use the corner points
               for that SCA to determine the min and max values. */
            if (grid->frame_type == IAS_LUNAR_MINBOX)
            {
                /* Find the declination and right ascension of the moon
                   center */ 
                if (sca_index == 0)
                {
                    /* Only need to find the moon using the first SCA, because
                       finding the values is based on time. */
                    ias_los_model_get_moon_center(lunar_proj, &moon_declination,
                        &moon_right_asc);
                }

                /* Free the lunar projection */
                ias_los_model_free_lunar_projection(lunar_proj);
                lunar_proj = NULL;

                /* Determine if this SCA contains the moon.  Note this is
                   based on reference location which is based on the time
                   of the center of SCA 7. Moon will change position slightly
                   throughout the acquisition. */

                /* Check to see if reference location of moon is within
                   bounding box of this SCA. */
                if (ias_math_point_in_polygon(POINT_COUNT, polygon_rasc,
                        polygon_dec, moon_right_asc, moon_declination) == TRUE)
                {
                    IAS_LOG_INFO("Moon is located in SCA %d", sca_index + 1);
                    found_moon_in_sca = TRUE;

                    /* Adjust the min and max values based on this SCA */
                    for (point_index = 0;
                         point_index < POINT_COUNT;
                         point_index++)
                    {
                        if (polygon_dec[point_index] > max_dec)
                            max_dec = polygon_dec[point_index];
                        if (polygon_dec[point_index] < min_dec)
                            min_dec = polygon_dec[point_index];
                        if (polygon_rasc[point_index] > max_rasc)
                            max_rasc = polygon_rasc[point_index];
                        if (polygon_rasc[point_index] < min_rasc)
                            min_rasc = polygon_rasc[point_index];
                    }

                    /* No more iterations are necessary for this band */
                    break;
                }
            }
            else /* MAXBOX */
            {
                /* Free the lunar projection */
                ias_los_model_free_lunar_projection(lunar_proj);
                lunar_proj = NULL;
            }
        } /* End SCA loop */

        /* Make sure we found the moon in the band */
        if (grid->frame_type == IAS_LUNAR_MINBOX)
        {
            if (found_moon_in_sca)
            {
                /* Save the SCA number containing the lunar image */
                *lunar_sca_number = (sca_index + 1);
            }
            else
            {
                IAS_LOG_ERROR("Moon was not found within any SCAs "
                    "for band index %d", band_index);
                return ERROR;
            }
        }

    } /* End Band loop */

    /* Calculate the distance scale */
    sdist = max_dist / min_dist;

    /* Search the available bands for the largest pixel size */
    for (band_index = 0; band_index < grid->nbands; band_index++)
    {
        if (grid->bands_available[band_index])
        {
            double pixsize = grid->gridbands[band_index][0].pixsize;
            if (pixsize > max_pixsize)
            {
                max_pixsize = pixsize;
            }
        }
    }

    /* Calculate the number of lines/samples at that pixel size */
    d_lat  = fabs(max_dec - min_dec);
    d_long = fabs(max_rasc - min_rasc);
    line_count = (2.0 * (int)((d_lat / (sdist * max_pixsize)) / 2.0)) + 1;
    sample_count = (2.0 * (int)((d_long / (sdist * max_pixsize)) / 2.0)) + 1;

    /* Set the grid corners based on the pixel size */
    grid->corners.upleft.x  = max_rasc;
    grid->corners.upleft.y  = max_dec;

    grid->corners.upright.x = grid->corners.upleft.x - (sample_count - 1.0)
        * max_pixsize;
    grid->corners.upright.y = grid->corners.upleft.y;

    grid->corners.loleft.y  = grid->corners.upleft.y - (line_count - 1.0)
        * max_pixsize;
    grid->corners.loleft.x  = grid->corners.upleft.x;

    grid->corners.loright.y = grid->corners.loleft.y;
    grid->corners.loright.x = grid->corners.upright.x;

    /* Initialize conversion factor */
    convert_to_degrees = scale * ias_math_get_degrees_per_radian();

    /* For each band and SCA, set the number of lines/samples scaled for the
       pixel size of that band */
    for (band_index = 0; band_index < grid->nbands; band_index++)
    {
        if (!grid->bands_available[band_index])
            continue;

        for (sca_index = 0; sca_index < grid->scas_per_band[band_index];
             sca_index++)
        {
            double pixsize = grid->gridbands[band_index][sca_index].pixsize;

            if (pixsize == max_pixsize)
            {
                grid->gridbands[band_index][sca_index].samps = sample_count;
                grid->gridbands[band_index][sca_index].lines = line_count;
            }
            else
            {
                grid->gridbands[band_index][sca_index].samps
                    = floor(max_pixsize/pixsize + 0.5) * (sample_count - 1) + 1;
                grid->gridbands[band_index][sca_index].lines
                    = floor(max_pixsize/pixsize + 0.5) * (line_count - 1) + 1;
            }
        } /* End SCA loop */

        /* Set the active area for each band to the full output area, and 
           convert to degrees */
        grid->active_corners[band_index].upleft.x = grid->corners.upleft.x 
            * convert_to_degrees; 
        grid->active_corners[band_index].upleft.y = grid->corners.upleft.y 
            * convert_to_degrees; 
        grid->active_corners[band_index].upright.x = grid->corners.upright.x 
            * convert_to_degrees; 
        grid->active_corners[band_index].upright.y = grid->corners.upright.y 
            * convert_to_degrees; 
        grid->active_corners[band_index].loleft.x = grid->corners.loleft.x 
            * convert_to_degrees; 
        grid->active_corners[band_index].loleft.y = grid->corners.loleft.y 
            * convert_to_degrees; 
        grid->active_corners[band_index].loright.x = grid->corners.loright.x 
            * convert_to_degrees; 
        grid->active_corners[band_index].loright.y = grid->corners.loright.y 
            * convert_to_degrees; 
    }

    return SUCCESS;
}
