/******************************************************************************
NAME: calc_mapping_and_sensitivity

PURPOSE:
Creates the output line/sample locations in the grid and calculates the jitter
sensitivity.

Notes:
    Includes the following routines:
        calc_mapping_and_sensitivity
        make_grid_point (local_routine)
        make_grid_sensitivity (local_routine)

    See ias_grid.h for a description of the grid structure.
    See ias_los_model.h for a description of the model structure.

******************************************************************************/
#include <string.h>                   /* String functions */

#include "gctp.h"                     /* For projection unit numbers */
#include "ias_geo.h"                  /* Geographic library */
#include "ias_los_model.h"            /* LOS model library */
#include "ias_math.h"                 /* Math library */
#include "ias_logging.h"              /* Logging library */
#include "create_grid.h"

/*************************************************************************
Name: make_grid_point 

Purpose: Calculates the input to output space mapping for a single grid point. 

Returns: SUCCESS or ERROR 

**************************************************************************/
static int make_grid_point
(
    double input_line,      /* I: Input space line coordinate */
    double input_sample,    /* I: Input space sample coordinate */
    const IAS_GRID_TYPE *grid, /* I: Grid structure */
    int band_index,         /* I: Band index */
    int sca_index,          /* I: Current SCA index */
    double elevation,       /* I: Current elevation plane height */
    const IAS_LOS_MODEL *model, /* I: LOS model information */
    double attitude_perturbation[3], /* I: Attitude perturbation vector for
                                  roll, pitch, and yaw in radians */
    double poly_x[4],       /* I: Array of x polynomial coefficients */
    double poly_y[4],       /* I: Array of y polynomial coefficients */
    const IAS_GEO_PROJ_TRANSFORMATION *geo2proj, /* projection transformation */
    double stellar_scale,   /* I: Angular scale factor for celestial
                                  acquisitions */
    const IAS_LUNAR_PROJECTION *lunar_proj, /* I: Lunar projection structure */
    double *output_line,    /* O: Output space line coordinate */
    double *output_sample   /* O: Output space sample coordinate */
)
{
    int status;             /* Return status */
    IAS_SENSOR_DETECTOR_TYPE detector_type = IAS_NOMINAL_DETECTOR;
    double distance_scale;  /* Scale of pixel due to change in distance
                               between moon and satellite */
    double in_lat;          /* Ground point latitude */
    double in_long;         /* Ground point longitude */
    double proj_x;          /* Projection X */
    double proj_y;          /* Projection Y */
    int band_number;        /* Reference band number */
    IAS_GRID_BAND_TYPE *grid_band; /* Current grid band */

    band_number = ias_sat_attr_convert_band_index_to_number(band_index);
    if (band_number == ERROR)
    {
        IAS_LOG_ERROR("Invalid band number for band index %d", band_index);
        return ERROR;
    }

    grid_band = &grid->gridbands[band_index][sca_index];

    /* Calculate the latitude and longitude of the current grid point. */
    status = ias_los_model_input_line_samp_to_geodetic(input_line, input_sample,
            band_index, sca_index, elevation, model, detector_type, 
            attitude_perturbation, &in_lat, &in_long);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating lat/long for band %d, SCA %d, elevation %f",
                band_number, sca_index + 1, elevation);
        return ERROR;
    }

    if (model->acquisition_type == IAS_EARTH)
    {
        /* Calculate the location of the latitude and longitude in output
           space. */
        status = ias_geo_transform_coordinate(geo2proj, in_long, in_lat,
                    &proj_x, &proj_y);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating the lat/long output space location for "
                    "band %d, SCA %d, elevation %f", band_number, sca_index + 1,
                    elevation);
            return ERROR;
        }
    
        /* Calculate the output line and sample. */
        if (grid->frame_type == IAS_PATH_ORIENTED
            || grid->frame_type == IAS_PATH_MINBOX
            || grid->frame_type == IAS_PATH_MAXBOX)
        { 
            /* Path-oriented */
            ias_math_eval_poly(grid_band->degree, poly_y, proj_x, 
                    proj_y, output_line);
            ias_math_eval_poly(grid_band->degree, poly_x, proj_x, 
                    proj_y, output_sample);
        }
        else
        { 
            /* North-up */
            *output_line = (grid->corners.upleft.y - proj_y) 
                / grid_band->pixsize;
            *output_sample = (proj_x - grid->corners.upleft.x) 
                / grid_band->pixsize;
        }
    }
    else if (model->acquisition_type == IAS_STELLAR)
    {
        in_lat /= stellar_scale;
        in_long /= stellar_scale;
        CONVERT_LUNAR_LAT_LONG_TO_LS(in_lat, in_long, grid->corners.upleft.x, 
                grid->corners.upleft.y, grid_band->pixsize, *output_line, 
                *output_sample);
    }
    else if (model->acquisition_type == IAS_LUNAR)
    {
        if (ias_los_model_transform_lunar_projection(lunar_proj, input_line, 
                input_sample, &in_lat, &in_long, &distance_scale) != SUCCESS)
        {
            IAS_LOG_ERROR("Projecting Lunar coordinates for line %f, sample %f",
                input_line, input_sample);
            return ERROR;
        }

        CONVERT_LUNAR_LAT_LONG_TO_LS(in_lat, in_long, grid->corners.upleft.x, 
                grid->corners.upleft.y, grid_band->pixsize, *output_line, 
                *output_sample);
    }
    else
    {
        IAS_LOG_ERROR("Unknown acquisition type %d", model->acquisition_type);
        return ERROR;
    }

    return SUCCESS;
}

/*************************************************************************
Name: make_grid_sensitivity 

Purpose: Calculates the roll-pitch-yaw to input space line/sample jitter
sensitivity coefficients for one grid point.  It calls 
ias_los_model_input_line_samp_to_geodetic (through make_grid_point) while 
varying the spacecraft attitude, the input space line number, and the input
space sample number to determine the corresponding output space sensitivity.
It then finds the input space offsets that provide the same effect in output
space as a given attitude perturbation, yielding the input space correction
needed to compensate for a unit jitter disturbance for each spacecraft axis.

Returns: SUCCESS or ERROR 

**************************************************************************/
static int make_grid_sensitivity
(
    double input_line,      /* I: Input space line coordinate */
    double input_sample,    /* I: Input space sample coordinate */
    const IAS_GRID_TYPE *grid, /* I: Grid structure */
    int band_index,         /* I: Band index */
    int sca_index,          /* I: Current SCA index */
    double elevation,       /* I: Current elevation plane height */
    const IAS_LOS_MODEL *model, /* I: LOS model information */
    int attitude_axis,      /* I: Attitude axis to analyze */
    double poly_x[4],       /* I: Array of x polynomial coefficients */
    double poly_y[4],       /* I: Array of y polynomial coefficients */
    const IAS_GEO_PROJ_TRANSFORMATION *geo2proj, /* projection transformation */
    double stellar_scale,   /* I: Angular scale factor for celestial
                                  acquisitions */
    double baseline_output_space_line, /* I: Baseline output space line 
                                  coordinate */
    double baseline_output_space_sample, /* I: Baseline output space sample 
                                  coordinate */
    const IAS_LUNAR_PROJECTION *lunar_proj, /* I: Lunar projection structure */
    double *line_sensitivity, /* O: Line direction sensitivity coefficient */
    double *sample_sensitivity /* O: Sample direction sensitivity
                                  coefficient */
)
{
    int status;             /* Status return flag */
    double output_lines[3]; /* Output line values for computing sensitivity */
    double output_samps[3]; /* Output sample values for computing sensitivity */
    double working_input_line; /* Input line location; value gets perturbed */
    double working_input_samp; /* Input sample location; value gets perturbed */
    double line_perturbation; /* Line perturbation */
    double sample_perturbation; /* Sample perturbation */
    double attitude_perturbation[3]; /* Attitude perturbation vector for roll
                               (x), pitch (y), and yaw (z) in radians */
    double input_line_to_output_line; /* Input line to output line 
                               sensitivity */
    double input_line_to_output_sample; /* Input line to output sample 
                               sensitivity */
    double input_sample_to_output_line; /* Input sample to output line 
                               sensitivity */
    double input_sample_to_output_sample; /* Input sample to output sample 
                               sensitivity */
    double output_line_to_input_line; /* Output line to input line 
                               sensitivity */
    double output_line_to_input_sample; /* Output line to input sample 
                               sensitivity */
    double output_sample_to_input_line; /* Output sample to input line 
                               sensitivity */
    double output_sample_to_input_sample; /* Output sample to input sample 
                               sensitivity */
    double determinant;     /* Determinant of sensitivity matrix */
    int band_number;        /* Reference band number */

    band_number = ias_sat_attr_convert_band_index_to_number(band_index);
    if (band_number == ERROR)
    {
        IAS_LOG_ERROR("Invalid band number for band index %d", band_index);
        return ERROR;
    }

    /* Initialize attitude perturbations for this axis */
    attitude_perturbation[0] = 0.0;
    attitude_perturbation[1] = 0.0;
    attitude_perturbation[2] = 0.0;
    attitude_perturbation[attitude_axis] = 1.0e-6;  /* 1 microradian 
       perturbation */

    /* Initialize the input location */
    working_input_line = input_line;
    working_input_samp = input_sample;

    /* Find the delta input line and samp required to yield the original
       output line and samp using the perturbed model */

    /* Compute the corresponding output point for the perturbed model */
    status = make_grid_point(working_input_line, working_input_samp, grid, 
            band_index, sca_index, elevation, model, attitude_perturbation, 
            poly_x, poly_y,geo2proj, stellar_scale, lunar_proj, 
            &output_lines[0], &output_samps[0]);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Generating current line and sample sensitivity for "
                "band %d SCA %d", band_number, sca_index + 1);
        return ERROR;
    }

    /* Compute the line sensitivity */
    line_perturbation = 1.0;
    if (input_line > 0) 
    {
        line_perturbation = -1.0;
    }
    status = make_grid_point((working_input_line + line_perturbation), 
            working_input_samp, grid, band_index, sca_index, elevation, model, 
            attitude_perturbation, poly_x, poly_y, geo2proj, stellar_scale, 
            lunar_proj, &output_lines[1], &output_samps[1]);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Generating line sensitivity for band %d SCA %d", 
                band_number, sca_index + 1);
        return ERROR;
    }

    /* Compute the sample sensitivity */
    sample_perturbation = 1.0;
    if (input_sample > 0) 
    {
        sample_perturbation = -1.0;
    }
    status = make_grid_point(working_input_line, (working_input_samp + 
            sample_perturbation), grid, band_index, sca_index, elevation, 
            model, attitude_perturbation, poly_x, poly_y, geo2proj, 
            stellar_scale, lunar_proj, &output_lines[2], &output_samps[2]);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Generating sample sensitivity for band %d SCA %d", 
                band_number, sca_index + 1);
        return ERROR;
    }

    /* Compute the output sensitivities to input deltas */
    output_line_to_input_line = (output_lines[1] - output_lines[0]) 
        / line_perturbation;
    output_line_to_input_sample = (output_lines[2] - output_lines[0]) 
        / sample_perturbation;
    output_sample_to_input_line = (output_samps[1] - output_samps[0]) 
        / line_perturbation;
    output_sample_to_input_sample = (output_samps[2] - output_samps[0]) 
        / sample_perturbation;

    /* Invert to compute the input sensitivities to output deltas */
    determinant = output_line_to_input_line * output_sample_to_input_sample 
            - output_line_to_input_sample * output_sample_to_input_line;
    input_line_to_output_line = output_sample_to_input_sample / determinant;
    input_line_to_output_sample = -output_line_to_input_sample / determinant;
    input_sample_to_output_line = -output_sample_to_input_line / determinant;
    input_sample_to_output_sample =  output_line_to_input_line / determinant;

    /* Compute the corresponding input space offset */
    line_perturbation = input_line_to_output_line * (output_lines[0] 
            - baseline_output_space_line) + input_line_to_output_sample 
            * (output_samps[0] - baseline_output_space_sample);
    sample_perturbation = input_sample_to_output_line * (output_lines[0] 
            - baseline_output_space_line) + input_sample_to_output_sample 
            * (output_samps[0] - baseline_output_space_sample);
    working_input_line -= line_perturbation;
    working_input_samp -= sample_perturbation;

    *line_sensitivity = (working_input_line - input_line) 
            / attitude_perturbation[attitude_axis];
    *sample_sensitivity = (working_input_samp - input_sample) 
            / attitude_perturbation[attitude_axis];

    return SUCCESS;
}

/*************************************************************************
Name: calc_mapping_and_sensitivity

Purpose: Establish the input to output mappings.  Invoke make_grid_point for
each point to compute the mapping, and then invoke make_grid_sensitivity for
each point to compute the jitter sensitivity coefficients.

Returns: SUCCESS or ERROR 

**************************************************************************/
int calc_mapping_and_sensitivity 
(
    int band_index,             /* I: band index */
    int sca_index,              /* I: SCA index */
    const IAS_LOS_MODEL *model, /* I: Model structure */
    double poly_x[4],           /* I: array of x polynomial coefficients */
    double poly_y[4],           /* I: array of y polynomial coefficients */
    double lunar_scale,         /* I: scale used for pixel size for 
                                      lunar/stellar acquisitions */
    IAS_GRID_TYPE *grid,        /* I/O: Grid structure */
    int *blank                  /* O: Flag to specify if no input pixels map to
                                      an output pixel */
                                   
)
{
    int band_number;          /* Reference band number */
    int status;               /* Return status */
    int i, j;                 /* Loop variables */
    int attitude_axis;        /* Attitude axis */
    int zplane;               /* Counter for elevation planes */
    int curr_pt;              /* Current point */
    double elevation = 0.0;   /* Elevation at the current point */
    double line;              /* Line in output space */
    double samp;              /* Sample in output space */
    double baseline_output_space_line; /* Base output space line coordinate */
    double baseline_output_space_sample; /* Base output space sample 
                                 coordinate */
    double attitude_perturbation[3]; /* Attitude pertubation vector (r,p,y) in 
                                 radians */
    double sensitivity_coefficients[2][3]; /* Attitude sensitivity coefficient 
                                 matrix */
    double stellar_scale = 0; /* Scale used for pixel size where acquisition 
                                 is stellar */
    int blank_flag;           /* Local flag for input pixels mapping to output 
                                 pixels */
    IAS_SENSOR_DETECTOR_TYPE detector_type = IAS_NOMINAL_DETECTOR;
    IAS_LUNAR_PROJECTION *lunar_proj = NULL;
    IAS_GRID_BAND_TYPE *grid_band; /* Current grid band */
    const IAS_GEO_PROJ_TRANSFORMATION *geo2proj;/* projection transformation */

    geo2proj = get_proj_transform_radians2proj();

    band_number = ias_sat_attr_convert_band_index_to_number(band_index);
    if (band_number == ERROR)
    {
        IAS_LOG_ERROR("Invalid band index %d", band_index);
        return ERROR;
    }

    grid_band = &grid->gridbands[band_index][sca_index];

    /* Initialize blank flag */
    blank_flag = 1;

    if (model->acquisition_type == IAS_LUNAR)
    {
        /* Create the lunar projection for the current band/sca */
        lunar_proj = ias_los_model_create_lunar_projection(model,
                        band_index, sca_index, detector_type, lunar_scale, 1);
        if (!lunar_proj)
        {
            IAS_LOG_ERROR("Initializing lunar projection for band "
                    "index %d, SCA index %d", band_index, sca_index);
            return ERROR;
        }
    }

    /* Loop through elevation planes */
    for (zplane = 0, curr_pt = 0; zplane < grid_band->nzplanes; zplane++)
    {
        /* Calculate elevation for plane */
        elevation = (zplane - grid_band->zeroplane) * grid_band->zspacing;
                         
        /* Loop through all of the grid points to get the 
           latitude/longitude, latitude/longitude in output space, 
           and finally the line/sample in output space. */
        for (i = 0; i < grid_band->ngrid_lines; i++)
        {
            for (j = 0; j < grid_band->ngrid_samps; j++, curr_pt++)
            {
                /* Initialize attitude perturbation. */
                attitude_perturbation[0] = 0.0;
                attitude_perturbation[1] = 0.0;
                attitude_perturbation[2] = 0.0;

                /* Generate the current grid point. */ 
                status = make_grid_point((double)grid_band->in_lines[i],
                        (double)grid_band->in_samps[j], grid, 
                        band_index, sca_index, elevation, model, 
                        attitude_perturbation, poly_x, poly_y, 
                        geo2proj, stellar_scale, lunar_proj, &line, &samp);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Generating grid point %d,%d for "
                            "band %d, SCA %d, elevation plane %d", i, j,
                            band_number, sca_index + 1, zplane);
                    ias_los_model_free_lunar_projection(lunar_proj);
                    return ERROR;
                }

                grid_band->out_lines[curr_pt] = line;
                grid_band->out_samps[curr_pt] = samp;
                baseline_output_space_line = line;
                baseline_output_space_sample = samp;

                /* Compute the attitude sensitivity using a loop. */
                for (attitude_axis = 0; attitude_axis < 3; attitude_axis++)
                {
                    status = make_grid_sensitivity(grid_band->in_lines[i], 
                            grid_band->in_samps[j], grid, band_index,
                            sca_index, elevation, model, attitude_axis, poly_x,
                            poly_y, geo2proj, stellar_scale, 
                            baseline_output_space_line, 
                            baseline_output_space_sample, lunar_proj,
                            &sensitivity_coefficients[0][attitude_axis], 
                            &sensitivity_coefficients[1][attitude_axis]);
                    if (status != SUCCESS)
                    {
                        IAS_LOG_ERROR("Generating sensitivity for "
                                "grid point %d,%d for band %d, SCA %d, "
                                "elevation plane %d", i, j, band_number,
                                sca_index + 1, zplane);
                        ias_los_model_free_lunar_projection(lunar_proj);
                        return ERROR;
                    }
                }

                /* If the image is currently blank and the output 
                   line/sample fall within the specified output image
                   then flag the image as not blank. */
                if (blank_flag && (line > 0 && line < grid_band->lines)  
                        && (samp > 0 && samp < grid_band->samps)) 
                {
                    blank_flag = 0;
                }

                /* Record the line and sample sensitivities for this point. */
                grid_band->line_sensitivity[curr_pt].x 
                    = sensitivity_coefficients[0][0];
                grid_band->line_sensitivity[curr_pt].y 
                    = sensitivity_coefficients[0][1];
                grid_band->line_sensitivity[curr_pt].z 
                    = sensitivity_coefficients[0][2];
                grid_band->samp_sensitivity[curr_pt].x 
                    = sensitivity_coefficients[1][0];
                grid_band->samp_sensitivity[curr_pt].y 
                    = sensitivity_coefficients[1][1];
                grid_band->samp_sensitivity[curr_pt].z 
                    = sensitivity_coefficients[1][2];
            } 
        }
    }

    /* Free the lunar projection if it was created */
    if (lunar_proj)
    {
        ias_los_model_free_lunar_projection(lunar_proj);
        lunar_proj = NULL;
    }

    /* Pass local flag back to caller */
    *blank = blank_flag;

    return SUCCESS;
}

