/******************************************************************************
NAME:           calc_odd_or_even_offsets

PURPOSE:
Computes the even/odd detector offset values.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

ALGORITHM REFERENCES:
See ias_grid.h for a description of the grid structure.
See ias_los_model.h for a description of the model structure.

******************************************************************************/
#include "create_grid.h"
#include "ias_geo.h"                  /* Geometric library */
#include "ias_los_model.h"            /* LOS model library */
#include "ias_math.h"                 /* Math library */
#include "ias_logging.h"              /* Logging library */

/* Calculate the offsets for the even or odd case. */
int calc_offsets 
(
    const IAS_LOS_MODEL *model, /* I: Model structure */
    const IAS_LUNAR_PROJECTION *lunar_proj,/* I: Lunar projection information */
    double poly_x[4],        /* I: array of x polynomial coefficients */
    double poly_y[4],        /* I: array of y polynomial coefficients */
    int band_index,          /* I: Band index */
    int sca_index,           /* I: SCA index */
    IAS_SENSOR_DETECTOR_TYPE dettype, /* I: Detector type for forward model 
                                   calculations */
    double elev,             /* I: Current elevation */
    double scale,            /* I: Scale for lunar projections */
    int gline,               /* I: Input line location */
    int gsamp,               /* I: Input sample location */
    int zplane,              /* I: Elevation plane */
    int samp_oe,             /* I: Closest even/odd detector sample number */
    IAS_GRID_TYPE *grid,     /* I/O: Grid structure */
    double *zline,           /* O: Line det Offset value */
    double *zsamp            /* O: Sample det Offset value */
)
{
    int stat;                /* Return status */
    double in_lat;           /* Latitude in input space */
    double in_long;          /* Longitude in input space */
    double proj_x;           /* Projection X */
    double proj_y;           /* Projection Y */
    double line;             /* Line in output space */
    double samp;             /* Sample in output space */
    double iline;            /* Line in input space */
    double isamp;            /* Sample in input space */
    double sdist;            /* Scale of pixel due to change in distance 
                                between moon and satellite */
    double lline;            /* Input line for calculating moon's position */
    double lsamp;            /* Input sample for calculating moon's position */

    /* Calculate the latitude and longitude of the detector. */
    stat = ias_los_model_input_line_samp_to_geodetic((double)gline,
            (double)samp_oe, grid->gridbands[band_index][sca_index].band_index,
            sca_index, elev, model, dettype, NULL, &in_lat, &in_long);
    if (stat != SUCCESS)
    {
        IAS_LOG_ERROR("In the ias_los_model_input_line_samp_to_geodetic call");
        return ERROR;
    }

    if (model->acquisition_type == IAS_EARTH)
    {    
        const IAS_GEO_PROJ_TRANSFORMATION *geo2proj;
        geo2proj = get_proj_transform_radians2proj();

        /* Calculate the location of the latitude and longitude in output
           space. */
        stat = ias_geo_transform_coordinate(geo2proj, in_long, in_lat,  
                    &proj_x, &proj_y);
        if (stat != SUCCESS)
        {
            IAS_LOG_ERROR("Transforming coordinate to target projection");
            return ERROR;
        }

        /* Calculate the output line and sample.  */
        if (grid->frame_type == IAS_PATH_ORIENTED || 
            grid->frame_type == IAS_PATH_MINBOX   ||
            grid->frame_type == IAS_PATH_MAXBOX)
        { /* Path-oriented */
            ias_math_eval_poly(grid->gridbands[band_index][sca_index].degree, 
                poly_y, proj_x, proj_y, &line);
            ias_math_eval_poly(grid->gridbands[band_index][sca_index].degree, 
                poly_x, proj_x, proj_y, &samp);
        }
        else
        { /* North-up */
            line = (grid->corners.upleft.y - proj_y)
                / grid->gridbands[band_index][sca_index].pixsize;
            samp = (proj_x - grid->corners.upleft.x)
                / grid->gridbands[band_index][sca_index].pixsize;
        }
    }
    else if (model->acquisition_type == IAS_STELLAR)
    {
        in_lat  /= scale;
        in_long /= scale;

        CONVERT_LUNAR_LAT_LONG_TO_LS(in_lat, in_long,
                grid->corners.upleft.x, grid->corners.upleft.y,
                grid->gridbands[band_index][sca_index].pixsize,
                line, samp);
    }
    else /* IAS_LUNAR */
    {
        /* Find moon's position at current line/sample time. */
        lline = (double)gline;
        lsamp = (double)samp_oe;

        if (ias_los_model_transform_lunar_projection(lunar_proj, 
                lline, lsamp, &in_lat, &in_long, &sdist) != SUCCESS)
        {
            IAS_LOG_ERROR("Projecting lunar coord to L/S");
            return ERROR;
        }

        CONVERT_LUNAR_LAT_LONG_TO_LS(in_lat, in_long,
                grid->corners.upleft.x, grid->corners.upleft.y,
                grid->gridbands[band_index][sca_index].pixsize,
                line, samp);
    }

    if (ias_grid_ols2ils(&grid->gridbands[band_index][sca_index], 
        line, samp, &iline, &isamp, zplane) != SUCCESS)
    {
        IAS_LOG_ERROR("Mapping detector to input space (ias_grid_ols2ils) "
            "for band_index %d sca_index %d input line %d input sample %d",
            band_index, sca_index, gline, gsamp);
            
        return ERROR;
    }

    *zline = (double)gline - iline;
    *zsamp = (double)samp_oe - isamp;

    return SUCCESS;
}


int calc_odd_or_even_offsets 
(
    const IAS_LOS_MODEL *model, /* I: model structure                        */
    const IAS_LUNAR_PROJECTION *lunar_proj,/* I: Lunar projection information */
    double poly_x[4],        /* I: array of x polynomial coefficients */
    double poly_y[4],        /* I: array of y polynomial coefficients */
    int band_index,          /* I: Band index                                */
    int sca_index,           /* I: SCA index                                 */
    IAS_SENSOR_DETECTOR_TYPE dettype, /* I: detector type for forward model 
                                   calculations                              */
    double elev,             /* I: Current elevation                         */
    double scale,            /* I: Scale for lunar projections               */
    int gline,               /* I: Input line location                       */
    int gsamp,               /* I: Input sample location                     */
    int zplane,              /* I: Elevation plane                           */
    IAS_GRID_TYPE *grid,     /* I/O: Grid structure                          */
    double *zline_even,      /* O: Line/even det offset value                */
    double *zsamp_even,      /* O: Sample/even det offset value              */
    double *zline_odd,       /* O: Line/odd det Offset value                 */
    double *zsamp_odd        /* O: Sample/odd det Offset value               */
)
{
    int odd_samp;            /* Closest odd detector sample number           */
    int even_samp;           /* Closest even detector sample number          */

    /* Find the closest even and odd detector/sample numbers. */ 
    even_samp = 2 * ((int)(gsamp/2.0));
    odd_samp = even_samp + 1;

    /* Compute even detector offsets. */
    if (calc_offsets(model, lunar_proj, poly_x, poly_y, band_index, sca_index,
        dettype, elev, scale, gline, gsamp, zplane, 
        even_samp, grid, zline_even, zsamp_even) 
        != SUCCESS)
    {
        IAS_LOG_ERROR("Computing even detector offsets");
        return ERROR;
    }

    /* Compute odd detector offsets. */
    if (calc_offsets(model, lunar_proj, poly_x, poly_y, band_index, sca_index,
        dettype, elev, scale, gline, gsamp, zplane, 
        odd_samp, grid, zline_odd, zsamp_odd) != SUCCESS)
    {
        IAS_LOG_ERROR("Computing odd detector offsets");
        return ERROR;
    }

    return SUCCESS;
}
