/******************************************************************************
NAME:       calc_path_oriented_frame

PURPOSE:
Given the center point and rotation angle, this function will calculate the
image corner coordinates in an SOM or UTM product.  It also calculates the
first-order polynomial coefficients which map output line/sample coords to
their corresponding output projection coordinates.

RETURN VALUE:
Type = int (SUCCESS or ERROR)

NOTES:
This routine will determine the frame in output space for the path-oriented
product.  The frame is calculated for each band, but the frame must be the
same for every band.  If not, then the application will error out.

Support for the path oriented option is not a requirement.  This file is
being left only for convenience for CalVal.
******************************************************************************/
#include <math.h>
#include "ias_geo.h"                  /* Geometric library */
#include "ias_math.h"                 /* Math library */
#include "ias_logging.h"              /* Logging library */
#include "create_grid.h"

int calc_path_oriented_frame
(
    IAS_GRID_TYPE *grid,    /* I/O: Grid structure */
    int units,              /* I: Projection units */
    double poly_x[IAS_MAX_NBANDS][4], /* O: Per-band sample-to-projection
                                            coefficients */
    double poly_y[IAS_MAX_NBANDS][4]  /* O: Per-band line-to-projection
                                            coefficients */
)
{
    double line_center;          /* Center line in output space */
    double samp_center;          /* Center sample in output space */
    double ul_dx, ul_dy;         /* Upper-left delta meters from center */
    double ur_dx, ur_dy;         /* Upper-right delta meters from center */
    double ll_dx, ll_dy;         /* Lower-left delta meteres from center */
    double lr_dx, lr_dy;         /* Lower-right delta meteres from center */
    double x, y;                 /* Scene center in projection coords */
    double x1, y1;               /* Auxiliary point projection coords */
    double lat1, lng1;           /* Auxiliary point latitude longitude */
    double ul_xp, ul_yp;         /* Upper-left projection coordinates */
    double ur_xp, ur_yp;         /* Upper-right projection coordinates */
    double ll_xp, ll_yp;         /* Lower-left projection coordinates */
    double lr_xp, lr_yp;         /* Lower-right projection coordinates */
    double ul_xp_old, ul_yp_old; /* Old upper-left projection coordinates */
    double ur_xp_old, ur_yp_old; /* Old upper-right projection coordinates */
    double ll_xp_old, ll_yp_old; /* Old lower-left projection coordinates */
    double lr_xp_old, lr_yp_old; /* Old lower-right projection coordinates */
    double rot_angle;            /* Rotation angle */
    double map_angle;            /* Map projection angle true north */
    double sin_angle;            /* Sine of the rotation angle */
    double cos_angle;            /* Cosine of the rotation angle */
    double radians_per_degree;   /* calculate radians per degree */
    int band_index;           /* Band index */
    int first_band_index = -1;/* first available band index */
    int sca_index;            /* SCA index */
    int i;                    /* Loop counter */
    int nl, ns;               /* Number of output lines/samples */
    int status;               /* Status returned from library funcs */
    int search_band_index;    /* index for searching for a band with the same
                                 pixel size */
    const IAS_GEO_PROJ_TRANSFORMATION *geo2proj;/* projection transformation */

    geo2proj = get_proj_transform_degrees2proj();

    /* Initialize these to avoid a compiler warning */
    ul_xp_old = ul_yp_old = ur_xp_old = ur_yp_old = ll_xp_old = ll_yp_old =
        lr_xp_old = lr_yp_old = 0.0;

    /* Get calculated value of radians per degree */
    radians_per_degree = ias_math_get_radians_per_degree();

    /* Loop through all the bands.   */
    for (band_index = 0; band_index < grid->nbands; band_index++)
    {
        int poly_copy_index = -1;

        /* skip bands not available */
        if (!grid->bands_available[band_index])
            continue;

        /* Check to see whether it has already been calculated for a
           previous band with the same pixel size.  If there is one, save
           the index to it. */
        for (search_band_index = 0; search_band_index < band_index;
             search_band_index++)
        {
            if (grid->bands_available[search_band_index])
            {
                if (grid->gridbands[search_band_index][0].pixsize
                    == grid->gridbands[band_index][0].pixsize)
                {
                    poly_copy_index = search_band_index;
                    break;
                }
            }
        }

        /* Get the reference position and the number of lines and samples
           in the current band resolution.                                  */
        for (sca_index = 0; sca_index < grid->scas_per_band[band_index];
             sca_index++)
        {
            nl = grid->gridbands[band_index][sca_index].lines;
            ns = grid->gridbands[band_index][sca_index].samps;

            line_center = (double) (nl - 1.0) / 2.0;
            samp_center = (double) (ns - 1.0) / 2.0;

            /* Transform WRS scene center to the target projection */
            status = ias_geo_transform_coordinate(geo2proj,
                grid->center_pt.lng, grid->center_pt.lat, &x, &y);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Transforming the scene center to the target "
                    "projection");
                return ERROR;
            }

            /* Construct a point just north of the scene center for use in
               computing the angle between true north and grid north.        */
            lat1 = grid->center_pt.lat + EPSILON;
            lng1 = grid->center_pt.lng;
            status = ias_geo_transform_coordinate(geo2proj,
                lng1, lat1, &x1, &y1);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Transforming a coordinate to the target "
                    "projection");
                return ERROR;
            }

            /* Compute the grid rotation angle.   */
            map_angle = atan2(x1 - x, y1 - y);

            /* Precompute the sine and cosine of the rotation angle.   */
            rot_angle = (grid->rotate_angle) * radians_per_degree + map_angle;
            sin_angle = sin(rot_angle);
            cos_angle = cos(rot_angle);

            /* Calculate the upper left coordinates.   */
            ul_dx = -samp_center * 
                grid->gridbands[band_index][sca_index].pixsize;
            ul_dy = line_center * 
                grid->gridbands[band_index][sca_index].pixsize;
            ul_xp =  ul_dx * cos_angle + ul_dy * sin_angle + x;
            ul_yp = -ul_dx * sin_angle + ul_dy * cos_angle + y;

            /* Calculate the upper right coordinates.   */
            ur_dx = ((ns-1.0) - samp_center) *
                grid->gridbands[band_index][sca_index].pixsize;
            ur_dy = line_center * 
                grid->gridbands[band_index][sca_index].pixsize;
            ur_xp =  ur_dx * cos_angle + ur_dy * sin_angle + x;
            ur_yp = -ur_dx * sin_angle + ur_dy * cos_angle + y;

            /* Calculate the lower right coordinates.   */
            lr_dx = ((ns-1.0) - samp_center) 
                * grid->gridbands[band_index][sca_index].pixsize;
            lr_dy = (line_center - (nl-1.0)) 
                * grid->gridbands[band_index][sca_index].pixsize;
            lr_xp =  lr_dx * cos_angle + lr_dy * sin_angle + x;
            lr_yp = -lr_dx * sin_angle + lr_dy * cos_angle + y;

            /* Calculate the lower left coordinates.   */
            ll_dx = -samp_center * 
                grid->gridbands[band_index][sca_index].pixsize;
            ll_dy = (line_center - (nl-1.0)) 
                * grid->gridbands[band_index][sca_index].pixsize;
            ll_xp =  ll_dx * cos_angle + ll_dy * sin_angle + x;
            ll_yp = -ll_dx * sin_angle + ll_dy * cos_angle + y;

            /* Check these corner points with the last band's corner point. */
            if (first_band_index == -1)
            {
                /* If the first band then set the old vals = current vals */
                ul_xp_old = ul_xp;
                ul_yp_old = ul_yp;
                ur_xp_old = ur_xp;
                ur_yp_old = ur_yp;
                ll_xp_old = ll_xp;
                ll_yp_old = ll_yp;
                lr_xp_old = lr_xp;
                lr_yp_old = lr_yp;

                first_band_index = band_index;

                /* Store the corners in the grid structure */
                grid->corners.upleft.x  = ul_xp;
                grid->corners.upleft.y  = ul_yp;
                grid->corners.upright.x = ur_xp;
                grid->corners.upright.y = ur_yp;
                grid->corners.loleft.x  = ll_xp;
                grid->corners.loleft.y  = ll_yp;
                grid->corners.loright.x = lr_xp;
                grid->corners.loright.y = lr_yp;
                grid->proj_center_pt.x = x;
                grid->proj_center_pt.y = y;
            }
            else if ((ul_xp != ul_xp_old || ul_yp != ul_yp_old ||
                      ur_xp != ur_xp_old || ur_yp != ur_yp_old ||
                      ll_xp != ll_xp_old || ll_yp != ll_yp_old ||
                      lr_xp != lr_xp_old || lr_yp != lr_yp_old))
            {
                IAS_LOG_ERROR(
                   "The corner points for bands %d and %d do not match",
                   ias_sat_attr_convert_band_index_to_number(band_index),
                   ias_sat_attr_convert_band_index_to_number(first_band_index));
                return ERROR;
            }

            /* only calculate the following for the first SCA of each band */
            if (sca_index != 0)
                continue;

            /* Calculate the polynomials for all available band resolutions. */
            if (poly_copy_index != -1)
            {
                /* A previous band has the same pixel size, so copy the poly
                   values instead of calculating the same values again */
                for (i = 0; i < 4; i++)
                {
                    poly_x[band_index][i] = poly_x[poly_copy_index][i];
                    poly_y[band_index][i] = poly_y[poly_copy_index][i];
                }
            }
            else
            {
                /* Now that we have calculated the corners in projection 
                   coordinates, calculate the first order polynomial
                   coefficients which map output line/samples to output
                   projection coordinates.  The 4x3 matrix is populated as
                   shown below, QR-factorized and then Ax = b is solved for two
                   different vectors b; one for samples, the other for lines.
                   The matrix A depicted below is in column-major format, which
                   is how QRfactorization interprets the array by default. */
                double A[12];     /* Matrix used by QRsolve for LSQ fit */
                double v[4];      /* Vector used in above operation */
                double px[4];     /* Poly_x used for calculations */
                double py[4];     /* Poly_y used for calculations */

                A[0] = 1.0;
                A[1] = 1.0;
                A[2] = 1.0;
                A[3] = 1.0;
                A[4]  = ul_xp;
                A[5]  = ur_xp;   /*     +-                                 -+ */
                A[6]  = lr_xp;   /*     |  1  upper-left  X  upper-left  Y  | */
                A[7]  = ll_xp;   /* A = |  1  upper-right X  upper-right Y  | */
                A[8]  = ul_yp;   /*     |  1  lower-right X  lower-right Y  | */
                A[9]  = ur_yp;   /*     |  1  lower-left  X  lower-left  Y  | */
                A[10] = lr_yp;   /*     +-                                 -+ */
                A[11] = ll_yp;

                /* Obtain QR factorization */
                ias_math_matrix_QRfactorization(A, 4, 3, v, 0); 

                /* Vector one: Find the coefficients to transform output 
                   samples to output coordinates in projection space.  Note 
                   that the call to ias_math_matrix_QRsolve() modifies the 
                   values at poly_x for part of the output (this vector 
                   becomes the before-mentioned coefficients.)               */
                px[0] = 0.0;
                px[1] = ns - 1.0;
                px[2] = ns - 1.0;
                px[3] = 0.0;
                ias_math_matrix_QRsolve(A, 4, 3, v, px, 0);

                /* Vector two: Find the coefficients to transform output lines
                   to output coordinates in projection space.               */
                py[0] = 0.0;
                py[1] = 0.0;
                py[2] = nl - 1.0;
                py[3] = nl - 1.0;
                ias_math_matrix_QRsolve(A, 4, 3, v, py, 0);

                /* Copy the poly's into their respective band resolution.   */
                for (i = 0; i < 4; i++)
                {
                    poly_x[band_index][i] = px[i];
                    poly_y[band_index][i] = py[i];
                }
            }
        }
    }

    return SUCCESS;
}
