/******************************************************************************
NAME: calc_path_maxbox_oriented_frame

PURPOSE: This routine calculates the path-oriented frame for maxbox.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

ALGORITHM REFERENCES:
This algorithm was created by combining the logic in both path_oriented.c
and minbox.c.  Almost all the code is from these two modules.
******************************************************************************/
#include <stdlib.h>
#include <string.h>                   /* String functions */
#include <math.h>

#include "ias_geo.h"                  /* Geographic library */
#include "ias_math.h"                 /* Math library */
#include "ias_logging.h"              /* Logging library */
#include "create_grid.h"


int calc_path_oriented_maxbox_frame 
(
    const IAS_LOS_MODEL *model, /* I: model structure */
    const int *num_lines,    /* I: Number of lines in model for each band */ 
    const int *num_samps,    /* I: Number of samps in model for each band */ 
    IAS_GRID_TYPE *grid,     /* I/O: Grid structure */
    double poly_x[IAS_MAX_NBANDS][4], /* O: Per-band array of x polynomial
                                            coefficients */
    double poly_y[IAS_MAX_NBANDS][4]  /* O: Per-band array of y polynomial
                                            coefficients */
)
{
    int i;                   /* Loop control */
    int nl;                  /* Number of lines in the scene */
    int ns;                  /* Number of samples in the scene */
    int zplane;              /* Counter for elevation planes */
    int max_pixel_band_index = -1; /* Band index for the a band using the 
                                maximum pixel size */
    int first_band_index = -1;/* First available band index */
    int band_index;          /* Band index */
    int band_number;         /* Band number corresponding to band_index */
    int band_loop_index;     /* Band loop counter */
    int sca_index;           /* SCA index */
    double pxmax;            /* Maximum X projection coordinate */
    double pxmin;            /* Minimum X projection coordinate */
    double pymax;            /* Maximum Y projection coordinate */
    double pymin;            /* Minimum Y projection coordinate */
    double elev = 0.0;       /* Elevation at the current point */
    struct IAS_DBL_LS ul_ls; /* Upper left line/sample for the corners */
    struct IAS_DBL_LS ur_ls; /* Upper right line/sample for the corners */
    struct IAS_DBL_LS ll_ls; /* Lower left line/sample for the corners */
    struct IAS_DBL_LS lr_ls; /* Lower right line/sample for the corners */
    struct IAS_DBL_XY ul_xy; /* Upper left x/y for the corners */
    struct IAS_DBL_XY ur_xy; /* Upper right x/y for the corners */
    struct IAS_DBL_XY ll_xy; /* Lower left x/y for the corners */
    struct IAS_DBL_XY lr_xy; /* Lower right x/y for the corners */
    double line_center;      /* Center line in output space */
    double samp_center;      /* Center sample in output space */
    double ul_dx, ul_dy;     /* Upper-left delta meters from center */
    double ur_dx, ur_dy;     /* Upper-right delta meters from center */
    double ll_dx, ll_dy;     /* Lower-left delta meteres from center */
    double lr_dx, lr_dy;     /* Lower-right delta meteres from center */
    double x, y;             /* Scene center in projection coords */
    double x1, y1;           /* Auxiliary point projection coords */
    double lat1, lng1;       /* Auxiliary point latitude longitude */
    double ul_xp = 0;        /* Upper-left projection x coordinate */
    double ul_yp = 0;        /* Upper-left projection y coordinate */
    double ur_xp = 0;        /* Upper-right projection x coordinate */
    double ur_yp = 0;        /* Upper-right projection y coordinate */
    double ll_xp = 0;        /* Lower-left projection x coordinate */
    double ll_yp = 0;        /* Lower-left projection y coordinate */
    double lr_xp = 0;        /* Lower-right projection x coordinate */
    double lr_yp = 0;        /* Lower-right projection y coordinate */
    double ul_xp_old = 0;    /* Old upper-left projection x coordinate */
    double ul_yp_old = 0;    /* Old upper-left projection y coordinate */
    double ur_xp_old = 0;    /* Old upper-right projection x coordinate */
    double ur_yp_old = 0;    /* Old upper-right projection y coordinate */
    double ll_xp_old = 0;    /* Old lower-left projection x coordinate */
    double ll_yp_old = 0;    /* Old lower-left projection y coordinate */
    double lr_xp_old = 0;    /* Old lower-right projection x coordinate */
    double lr_yp_old = 0;    /* Old lower-right projection y coordinate */
    double rot_angle;        /* Rotation angle */
    double map_angle;        /* Map projection angle true north */
    double sin_angle;        /* Sine of the rotation angle */
    double cos_angle;        /* Cosine of the rotation angle */
    int status;              /* Status returned from library funcs */
    int out_nl;              /* Number of lines for optimal band */
    int out_ns;              /* Number of sample for optimal band */
    double dx;               /* Sample/X change due to rotation */
    double dy;               /* Line/Y change due to rotation */
    double xp;               /* New sample/X due to rotation */
    double yp;               /* New line/Y due to rotation */
    int pad_nl;              /* Padding of lines due to WRS/center offset */
    int pad_ns;              /* Padding of samples due to WRS/center offset */
    double pad_x = 0;
    double pad_y = 0;
    double max_pixsize = 0;  /* Pixel size to be used for padding */
    int first_available_sca; /* Is this the first available sca? */
    int search_band_index;   /* index for searching for a band with the same
                                pixel size */
    const IAS_GEO_PROJ_TRANSFORMATION *geo2proj;/* projection transformation */

    geo2proj = get_proj_transform_degrees2proj();

    /* Transform the center coordinate to the target projection */
    status = ias_geo_transform_coordinate(geo2proj, grid->center_pt.lng, 
        grid->center_pt.lat, &x, &y);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Transforming center point to target projection");
        return ERROR;
    }

    /* Construct a point just north of the scene center for use in
       computing the angle between true north and grid north. */
    lat1 = grid->center_pt.lat + EPSILON;
    lng1 = grid->center_pt.lng;
    status = ias_geo_transform_coordinate(geo2proj, lng1, lat1, &x1, &y1 );
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Transforming coordinate to target projection");
        return ERROR;
    }

    /* Compute the grid rotation angle. */
    map_angle = atan2(x1 - x, y1 - y);

    /* Precompute the sine and cosine of the rotation angle. */
    rot_angle = (grid->rotate_angle) * ias_math_get_radians_per_degree()
                 + map_angle;
    sin_angle = sin(rot_angle);
    cos_angle = cos(rot_angle);

    first_available_sca = TRUE;
    for (band_loop_index = 0; band_loop_index < grid->nbands; 
         band_loop_index++)
    {
        /* skip bands not available */
        if (!grid->bands_available[band_loop_index])
            continue;

        for (sca_index = 0; sca_index < grid->scas_per_band[band_loop_index];
             sca_index++)
        {
            /* Choose biggest pixsize to pad corners */
            if (first_available_sca)
            {
                max_pixsize
                    = grid->gridbands[band_loop_index][sca_index].pixsize;
                max_pixel_band_index = band_loop_index;

            }
            else if (max_pixsize < 
                     grid->gridbands[band_loop_index][sca_index].pixsize)
            {
                max_pixsize
                    = grid->gridbands[band_loop_index][sca_index].pixsize;
                max_pixel_band_index = band_loop_index;
            }

            for (zplane = 0; 
                 zplane < grid->gridbands[band_loop_index][sca_index].nzplanes;
                 zplane++)
            {
                elev = (zplane - 
                    grid->gridbands[band_loop_index][sca_index].zeroplane) *
                    grid->gridbands[band_loop_index][sca_index].zspacing;

                /* Find the offsets for the first and last lines in the 
                   scene.  These will determine the corner points in 
                   line/sample.  */
                nl = num_lines[band_loop_index];
                ns = num_samps[band_loop_index];

                /* Get the band index */
                band_index =
                    grid->gridbands[band_loop_index][sca_index].band_index;

                /* Get the band number */
                band_number = ias_sat_attr_convert_band_index_to_number(
                    band_index);
                if (band_number == ERROR)
                {
                    IAS_LOG_ERROR("Invalid band number %d", band_number);
                    return ERROR;
                }

                /* Find the sample values for the corner points. */
                ul_ls.samp = 0;
                ll_ls.samp = 0;
                ur_ls.samp = ns - 1;
                lr_ls.samp = ns - 1;

                /* Find the top and bottom line values for the corner points. */
                ul_ls.line = ur_ls.line = 0;
                ll_ls.line = lr_ls.line = nl - 1;

                /* For the 4 corner points, find the location in output 
                   space */
                status = calc_corner_point_output_space_locations(&ul_ls, 
                    &ur_ls, &ll_ls, &lr_ls, band_index, sca_index, elev, model,
                    grid, &ul_xy, &ur_xy, &ll_xy, &lr_xy);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Calculating output space locations for "
                        "the corner points for band %d SCA %d elevation %f",
                        band_number, sca_index + 1, elev);
                    return ERROR;
                }

                /* Find the minimum and maximum x and y corners.  The corners
                   need to be rotated to a space that is orthogonal to the
                   path oriented X/Y axis.                                 */
                dx = (ul_xy.x - x);
                dy = (ul_xy.y - y);
                xp = dx * cos_angle - dy * sin_angle + x;
                yp = dx * sin_angle + dy * cos_angle + y;
                if (sca_index == 0 && zplane == 0 && first_available_sca)
                {
                    pxmin = xp;
                    pymin = yp;
                    pxmax = xp;
                    pymax = yp;
                }
                else
                {
                    if (xp > pxmax)
                        pxmax = xp;
                    if (xp < pxmin)
                        pxmin = xp;
                    if (yp > pymax)
                        pymax = yp;
                    if (yp < pymin)
                        pymin = yp;
                }

                dx = (ur_xy.x - x);
                dy = (ur_xy.y - y);
                xp = dx * cos_angle - dy * sin_angle + x;
                yp = dx * sin_angle + dy * cos_angle + y;
                if (xp > pxmax)
                    pxmax = xp;
                if (xp < pxmin)
                    pxmin = xp;
                if (yp > pymax)
                    pymax = yp;
                if (yp < pymin)
                    pymin = yp;

                dx = (lr_xy.x - x);
                dy = (lr_xy.y - y);
                xp = dx * cos_angle - dy * sin_angle + x;
                yp = dx * sin_angle + dy * cos_angle + y;
                if (xp > pxmax)
                    pxmax = xp;
                if (xp < pxmin)
                    pxmin = xp;
                if (yp > pymax)
                    pymax = yp;
                if (yp < pymin)
                    pymin = yp;

                dx = (ll_xy.x - x);
                dy = (ll_xy.y - y);
                xp = dx * cos_angle - dy * sin_angle + x;
                yp = dx * sin_angle + dy * cos_angle + y;
                if (xp > pxmax)
                    pxmax = xp;
                if (xp < pxmin)
                    pxmin = xp;
                if (yp > pymax)
                    pymax = yp;
                if (yp < pymin)
                    pymin = yp;
            } /* Z plane loop */

            first_available_sca = FALSE;
        } /* SCA loop */ 
    } /* Band loop */

    /* Pad corners to PADVAL * pixel size. */
    status = pad_corners(max_pixsize, &pxmax, &pxmin, &pymax, &pymin);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("In the pad_corners call");
        return ERROR;
    }

    /* Calculate output lines and samples according to new max/min.  */
    out_nl = (int) (((pymax - pymin) / max_pixsize) + 1.5);
    out_ns = (int) (((pxmax - pxmin) / max_pixsize) + 1.5);

    /* Pad according to nominal scene center to WRS center. */
    pad_x = IAS_MAX(abs(pxmax - x), abs(pxmin - x)) - (pxmax - pxmin) / 2.0;
    pad_y = IAS_MAX(abs(pymax - y), abs(pymin - y)) - (pymax - pymin) / 2.0;
    pad_nl = (int) ((fabs(pad_y) / max_pixsize) + 1.5);
    pad_ns = (int) ((fabs(pad_x) / max_pixsize) + 1.5);
    out_nl = out_nl + 2 * pad_nl;
    out_ns = out_ns + 2 * pad_ns;

    /* Make sure output lines and samples are an odd number */
    if (out_nl%2 == 0)
        out_nl = out_nl + 1;
    if (out_ns%2 == 0)
        out_ns = out_ns + 1;

    /* Store NL and NS in grid structure. */
    for (band_index = 0; band_index < grid->nbands; band_index++)
    {
        /* skip bands not available */
        if (!grid->bands_available[band_index])
            continue;

        for (sca_index = 0; sca_index < grid->scas_per_band[band_index];
             sca_index++)
        {
            grid->gridbands[band_index][sca_index].lines = 
                (int) ((grid->gridbands[max_pixel_band_index][sca_index].pixsize
                / grid->gridbands[band_index][sca_index].pixsize)
                * (out_nl - 1)) + 1;
            grid->gridbands[band_index][sca_index].samps = 
                (int) ((grid->gridbands[max_pixel_band_index][sca_index].pixsize
                / grid->gridbands[band_index][sca_index].pixsize)
                * (out_ns - 1)) + 1;
        }
    }

    sin_angle = sin(rot_angle);
    cos_angle = cos(rot_angle);

    first_available_sca = TRUE;
    for (band_loop_index = 0; band_loop_index < grid->nbands; 
         band_loop_index++)
    {
        int poly_copy_index = -1; /* index of band of the same resolution to
                                     use as the source for copying previously
                                     calculated poly_x and poly_y values */

        /* skip bands not available */
        if (!grid->bands_available[band_loop_index])
            continue;

        if (first_available_sca == TRUE)
        {
            first_band_index = band_loop_index;
        }

        /* Check to see whether it has already been calculated for a
           previous band with the same pixel size.  If there is one, save
           the index to it. */
        for (search_band_index = 0; search_band_index < band_loop_index;
             search_band_index++)
        {
            if (grid->bands_available[search_band_index])
            {
                if (grid->gridbands[search_band_index][0].pixsize
                    == grid->gridbands[band_loop_index][0].pixsize)
                {
                    poly_copy_index = search_band_index;
                    break;
                }
            }
        }

        for (sca_index = 0; sca_index < grid->scas_per_band[band_loop_index];
             sca_index++)
        {
            /* Get the reference position and the number of lines and samples
               in the current band resolution.                               */
            nl = grid->gridbands[band_loop_index][sca_index].lines;
            ns = grid->gridbands[band_loop_index][sca_index].samps;

            line_center = (double) (nl - 1.0) / 2.0;
            samp_center = (double) (ns - 1.0) / 2.0;

            /* Calculate the upper left coordinates.   */
            ul_dx = -samp_center * 
                grid->gridbands[band_loop_index][sca_index].pixsize;
            ul_dy = line_center * 
                grid->gridbands[band_loop_index][sca_index].pixsize;
            ul_xp =  ul_dx * cos_angle + ul_dy * sin_angle + x;
            ul_yp = -ul_dx * sin_angle + ul_dy * cos_angle + y;

            /* Calculate the upper right coordinates.   */
            ur_dx = ((ns-1.0) - samp_center) *
                grid->gridbands[band_loop_index][sca_index].pixsize;
            ur_dy = line_center * 
                grid->gridbands[band_loop_index][sca_index].pixsize;
            ur_xp =  ur_dx * cos_angle + ur_dy * sin_angle + x;
            ur_yp = -ur_dx * sin_angle + ur_dy * cos_angle + y;

            /* Calculate the lower right coordinates.   */
            lr_dx = ((ns-1.0) - samp_center) *
                grid->gridbands[band_loop_index][sca_index].pixsize;
            lr_dy = (line_center - (nl-1.0)) *
                grid->gridbands[band_loop_index][sca_index].pixsize;
            lr_xp =  lr_dx * cos_angle + lr_dy * sin_angle + x;
            lr_yp = -lr_dx * sin_angle + lr_dy * cos_angle + y;

            /* Calculate the lower left coordinates.   */
            ll_dx = -samp_center * 
                grid->gridbands[band_loop_index][sca_index].pixsize;
            ll_dy = (line_center - (nl-1.0)) *
                grid->gridbands[band_loop_index][sca_index].pixsize;
            ll_xp =  ll_dx * cos_angle + ll_dy * sin_angle + x;
            ll_yp = -ll_dx * sin_angle + ll_dy * cos_angle + y;

            /* Check these corner points with the last first calculated corner
               points. */
            if (first_available_sca)
            { 
                first_available_sca = FALSE;

                /* If only 1 band then don't keep track of each band corner.  If
                   the first band then set the old vals = current vals. */
                ul_xp_old = ul_xp;
                ul_yp_old = ul_yp;
                ur_xp_old = ur_xp;
                ur_yp_old = ur_yp;
                ll_xp_old = ll_xp;
                ll_yp_old = ll_yp;
                lr_xp_old = lr_xp;
                lr_yp_old = lr_yp;
            }
            else if ((ul_xp != ul_xp_old || ul_yp != ul_yp_old ||
                      ur_xp != ur_xp_old || ur_yp != ur_yp_old ||
                      ll_xp != ll_xp_old || ll_yp != ll_yp_old ||
                      lr_xp != lr_xp_old || lr_yp != lr_yp_old) &&
                    grid->nbands != 1)
            {
                IAS_LOG_ERROR(
                    "The corner points for bands %d and %d do not match\n",
                    ias_sat_attr_convert_band_index_to_number(band_loop_index),
                    ias_sat_attr_convert_band_index_to_number(
                        first_band_index));
                return ERROR;
            }

            /* only calculate the following for the first SCA of each band */
            if (sca_index != 0)
                continue;

            /* Calculate the polynomials for all available band resolutions. */
            if (poly_copy_index != -1)
            {
                /* A previous band has the same pixel size, so copy the poly
                   values instead of calculating the same values again */
                for (i = 0; i < 4; i++)
                {
                    poly_x[band_loop_index][i] = poly_x[poly_copy_index][i];
                    poly_y[band_loop_index][i] = poly_y[poly_copy_index][i];
                }
            }
            else
            {
                /* Now that we have calculated the corners in projection 
                   coordinates, calculate the first order polynomial
                   coefficients which map output line/samples to output
                   projection coordinates.  The 4x3 matrix is populated as
                   shown below, QR-factorized and then Ax = b is solved for two
                   different vectors b; one for samples, the other for lines.
                   The matrix A depicted below is in column-major format, which
                   is how QRfactorization interprets the array by default. */
                double A[12];      /* Matrix used by QRsolve for LSQ fit */
                double v[4];       /* Vector used in above operation */
                double px[4];      /* Poly_x used for calculations */
                double py[4];      /* Poly_y used for calculations */

                A[0]  = 1.0;
                A[1]  = 1.0;
                A[2]  = 1.0;
                A[3]  = 1.0;
                A[4]  = ul_xp;
                A[5]  = ur_xp;  /*     +-                                 -+ */
                A[6]  = lr_xp;  /*     |  1  upper-left  X  upper-left  Y  | */
                A[7]  = ll_xp;  /* A = |  1  upper-right X  upper-right Y  | */
                A[8]  = ul_yp;  /*     |  1  lower-right X  lower-right Y  | */
                A[9]  = ur_yp;  /*     |  1  lower-left  X  lower-left  Y  | */
                A[10] = lr_yp;  /*     +-                                 -+ */
                A[11] = ll_yp;

                /* Obtain QR factorization */
                ias_math_matrix_QRfactorization(A, 4, 3, v, 0); 

                /* Vector one: Find the coefficients to transform output samples
                   to output coordinates in projection space.  Note that the 
                   call to ias_math_matrix_QRsolve() modifies the values at 
                   poly_x for part of the output (this vector becomes the 
                   before-mentioned coefficients.)                           */
                px[0] = 0.0;
                px[1] = ns - 1.0;
                px[2] = ns - 1.0;
                px[3] = 0.0;
                ias_math_matrix_QRsolve(A, 4, 3, v, px, 0);

                /* Vector two: Find the coefficients to transform output lines
                   to output coordinates in projection space.                */
                py[0] = 0.0;
                py[1] = 0.0;
                py[2] = nl - 1.0;
                py[3] = nl - 1.0;
                ias_math_matrix_QRsolve(A, 4, 3, v, py, 0);

                /* Copy the poly's into their respective band resolution.   */
                for (i = 0; i < 4; i++)
                {
                    poly_x[band_loop_index][i] = px[i];
                    poly_y[band_loop_index][i] = py[i];
                }
            }
        } /* SCA counter */

    } /* Band counter */

    /* Store the corners in the grid structure.   */
    grid->corners.upleft.x  = ul_xp;
    grid->corners.upleft.y  = ul_yp;
    grid->corners.upright.x = ur_xp;
    grid->corners.upright.y = ur_yp;
    grid->corners.loleft.x  = ll_xp;
    grid->corners.loleft.y  = ll_yp;
    grid->corners.loright.x = lr_xp;
    grid->corners.loright.y = lr_yp;
    grid->proj_center_pt.x = x;
    grid->proj_center_pt.y = y;

    /* Convert the corners to lat/long, and set the active area corners */
    status = set_grid_active_corners(grid);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Converting corners to lat/long and setting active "
            "area corners");
        return ERROR;
    }

    return SUCCESS;
}
