/******************************************************************************
NAME: calc_path_oriented_minbox_frame

PURPOSE: Creates path-oriented frame for minbox.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

ALGORITHM REFERENCES:
This algorithm was created by combining the logic in both path_oriented.c
and minbox.c.  Almost all the code is from these two modules.
******************************************************************************/
#include <stdlib.h>
#include <string.h>                   /* String functions */
#include <math.h>

#include "ias_geo.h"                  /* Geographic library */
#include "ias_math.h"                 /* Math library */
#include "ias_logging.h"              /* Logging library */
#include "create_grid.h"

int calc_path_oriented_minbox_frame 
(
    const IAS_LOS_MODEL *model, /* I: Model structure */
    int opt_band_index,   /* I: Index of "optimal" band to work on */
    IAS_GRID_TYPE *grid,  /* I/O: Grid structure */
    double poly_x[IAS_MAX_NBANDS][4], /* O: Per-band array of x polynomial
                                            coefficients */
    double poly_y[IAS_MAX_NBANDS][4]  /* O: Per-band array of y polynomial
                                            coefficients */
)
{
    int i;                   /* Loop control */
    int nl;                  /* Number of lines in the scene */
    int ns;                  /* Number of samples in the scene */
    int ref_band_index;      /* Index for the reference band */
    int band_index;          /* Band index */
    int band_number;         /* Band number for band index */
    int band_loop_index;     /* Band loop counter */
    int first_band;          /* First available band */
    int sca_index;           /* SCA index */
    double pxmax;            /* Maximum X projection coordinate */
    double pxmin;            /* Minimum X projection coordinate */
    double pymax;            /* Maximum Y projection coordinate */
    double pymin;            /* Minimum Y projection coordinate */
    struct IAS_DBL_LAT_LONG ul_ltlg; /* Upper left lat/long for the corners */
    struct IAS_DBL_LAT_LONG ur_ltlg; /* Upper right lat/long for the corners */
    struct IAS_DBL_LAT_LONG ll_ltlg; /* Lower left lat/long for the corners */
    struct IAS_DBL_LAT_LONG lr_ltlg; /* Lower right lat/long for the corners */
    struct IAS_DBL_XY ul_xy; /* Upper left x/y for the corners */
    struct IAS_DBL_XY ur_xy; /* Upper right x/y for the corners */
    struct IAS_DBL_XY ll_xy; /* Lower left x/y for the corners */
    struct IAS_DBL_XY lr_xy; /* Lower right x/y for the corners */
    double line_center;      /* Center line in output space */
    double samp_center;      /* Center sample in output space */
    double ul_dx, ul_dy;     /* Upper-left delta meters from center */
    double ur_dx, ur_dy;     /* Upper-right delta meters from center */
    double ll_dx, ll_dy;     /* Lower-left delta meteres from center */
    double lr_dx, lr_dy;     /* Lower-right delta meteres from center */
    double x, y;             /* Scene center in projection coords */
    double x1, y1;           /* Auxiliary point projection coords */
    double lat1, lng1;       /* Auxiliary point latitude longitude */
    double ul_xp = 0;        /* Upper-left projection x coordinate */
    double ul_yp = 0;        /* Upper-left projection y coordinate */
    double ur_xp = 0;        /* Upper-right projection x coordinate */
    double ur_yp = 0;        /* Upper-right projection y coordinate */
    double ll_xp = 0;        /* Lower-left projection x coordinate */
    double ll_yp = 0;        /* Lower-left projection y coordinate */
    double lr_xp = 0;        /* Lower-right projection x coordinate */
    double lr_yp = 0;        /* Lower-right projection y coordinate */
    double ul_xp_old = 0;    /* Old upper-left projection x coordinate */
    double ul_yp_old = 0;    /* Old upper-left projection y coordinate */
    double ur_xp_old = 0;    /* Old upper-right projection x coordinate */
    double ur_yp_old = 0;    /* Old upper-right projection y coordinate */
    double ll_xp_old = 0;    /* Old lower-left projection x coordinate */
    double ll_yp_old = 0;    /* Old lower-left projection y coordinate */
    double lr_xp_old = 0;    /* Old lower-right projection x coordinate */
    double lr_yp_old = 0;    /* Old lower-right projection y coordinate */
    double rot_angle = 0;    /* Rotation angle */
    double map_angle = 0;    /* Map projection angle true north */
    double sin_angle = 0;    /* Sine of the rotation angle */
    double cos_angle = 0;    /* Cosine of the rotation angle */
    int status;              /* Status returned from library functions */
    int out_nl;              /* Number of lines for optimal band */
    int out_ns;              /* Number of sample for optimal band */
    double dx;               /* Sample/X change due to rotation */
    double dy;               /* Line/Y change due to rotation */
    double xp;               /* New sample/X due to rotation */
    double yp;               /* New line/Y due to rotation */
    int pad_nl;              /* Padding of lines due to WRS/center offset */
    int pad_ns;              /* Padding of samples due to WRS/center offset */
    int first_sca;           /* Flag showing if it's the first sca in loop */
    int search_band_index;   /* index for searching for a band with the same
                                pixel size */
    double radians_per_degree; /* calculated radians per degree */
    const IAS_GEO_PROJ_TRANSFORMATION *degrees2proj; /* Projection
                                 transformation parameters */

    /* Calculate radians per degree */
    radians_per_degree = ias_math_get_radians_per_degree();

    /* Determine the reference band index.  Implement this on the optimal
     * band.  This will be the band for which the frame is calculated, but
     * the frame will be the same for all bands.  */
    ref_band_index = opt_band_index;

    first_band = TRUE;
    for (band_index = 0; band_index < grid->nbands; band_index++)
    {
        /* Skip bands not available. */
        if (!grid->bands_available[band_index])
            continue;

        /* Get the band number. */
        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Invalid band number for band index %d", band_index);
            return ERROR;
        }

        /* Find the corner points' latitude and longitude projection values. */
        ul_ltlg.lng = grid->active_corners[band_index].upleft.x;
        ul_ltlg.lat = grid->active_corners[band_index].upleft.y;
        ur_ltlg.lng = grid->active_corners[band_index].upright.x;
        ur_ltlg.lat = grid->active_corners[band_index].upright.y;
        ll_ltlg.lng = grid->active_corners[band_index].loleft.x;
        ll_ltlg.lat = grid->active_corners[band_index].loleft.y;
        lr_ltlg.lng = grid->active_corners[band_index].loright.x;
        lr_ltlg.lat = grid->active_corners[band_index].loright.y;

        /* Initialize projection transformation. */
        degrees2proj = get_proj_transform_degrees2proj();

        /* Calculate the location of the latitude and longitude in output space.
           Units stored in the grid active area corners are in degrees */
        status = ias_geo_transform_coordinate(degrees2proj, ul_ltlg.lng,
            ul_ltlg.lat, &ul_xy.x, &ul_xy.y);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating location in output space for lat/long "
                "for band %d upper left corner", band_number);
            return ERROR;
        }

        status = ias_geo_transform_coordinate(degrees2proj, ur_ltlg.lng,
            ur_ltlg.lat, &ur_xy.x, &ur_xy.y);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating location in output space for lat/long "
                "for band %d upper right corner", band_number);
            return ERROR;
        }

        status = ias_geo_transform_coordinate(degrees2proj, ll_ltlg.lng,
            ll_ltlg.lat, &ll_xy.x, &ll_xy.y);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating location in output space for lat/long "
                "for band %d lower left corner", band_number);
            return ERROR;
        }

        status = ias_geo_transform_coordinate(degrees2proj, lr_ltlg.lng,
            lr_ltlg.lat, &lr_xy.x, &lr_xy.y);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating location in output space for lat/long "
                "for band %d lower right corner", band_number);
            return ERROR;
        }

        /* Transform the WRS scene center to the target projection */
        status = ias_geo_transform_coordinate(degrees2proj, grid->center_pt.lng,
            grid->center_pt.lat, &x, &y);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Transforming the scene center to the target "
                "projection");
            return ERROR;
        }

        /* Construct a point just north of the scene center for use in
           computing the angle between true north and grid north.       */
        lat1 = grid->center_pt.lat + EPSILON;
        lng1 = grid->center_pt.lng;
        status = ias_geo_transform_coordinate(degrees2proj, lng1, lat1, &x1, 
            &y1);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Transforming coordinate to the target projection");
            return ERROR;
        }

        /* Compute the grid rotation angle. */
        map_angle = atan2(x1 - x, y1 - y);

        /* Precompute the sine and cosine of the rotation angle. */
        rot_angle = grid->rotate_angle * radians_per_degree + map_angle;
        sin_angle = sin(rot_angle);
        cos_angle = cos(rot_angle);

        /* Find the minimum and maximum x and y corners.  The corners need to
           be rotated to a space that is orthogonal to the path oriented X/Y
           axis. */
        dx = (ul_xy.x - x);
        dy = (ul_xy.y - y);
        xp = dx * cos_angle - dy * sin_angle + x;
        yp = dx * sin_angle + dy * cos_angle + y;
        if (first_band)
        {
            pxmin = xp;
            pymin = yp;
            pxmax = xp;
            pymax = yp;
        }
        else
        {
            if (xp > pxmax)
                pxmax = xp;
            if (xp < pxmin)
                pxmin = xp;
            if (yp > pymax)
                pymax = yp;
            if (yp < pymin)
                pymin = yp;
        }

        dx = (ur_xy.x - x);
        dy = (ur_xy.y - y);
        xp = dx * cos_angle - dy * sin_angle + x;
        yp = dx * sin_angle + dy * cos_angle + y;
        if( xp > pxmax )
            pxmax = xp;
        if( xp < pxmin )
            pxmin = xp;
        if( yp > pymax )
            pymax = yp;
        if( yp < pymin )
            pymin = yp;

        dx = (lr_xy.x - x);
        dy = (lr_xy.y - y);
        xp = dx * cos_angle - dy * sin_angle + x;
        yp = dx * sin_angle + dy * cos_angle + y;
        if (xp > pxmax)
            pxmax = xp;
        if (xp < pxmin)
            pxmin = xp;
        if (yp > pymax)
            pymax = yp;
        if (yp < pymin)
            pymin = yp;

        dx = (ll_xy.x - x);
        dy = (ll_xy.y - y);
        xp = dx * cos_angle - dy * sin_angle + x;
        yp = dx * sin_angle + dy * cos_angle + y;
        if (xp > pxmax)
            pxmax = xp;
        if (xp < pxmin)
            pxmin = xp;
        if (yp > pymax)
            pymax = yp;
        if (yp < pymin)
            pymin = yp;

        first_band = FALSE;
    }

    /* Pad corners to PADVAL * pixel size. */
    sca_index = 0; /* Pixel sizes should be the same */
    status = pad_corners(grid->gridbands[ref_band_index][sca_index].pixsize, 
        &pxmax, &pxmin, &pymax, &pymin);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("In the pad_corners call");
        return ERROR;
    }

    /* Calculate output lines and samples according to new max/min.  */
    out_nl = (int) (((pymax - pymin) / 
        grid->gridbands[ref_band_index][sca_index].pixsize) + 1.5);
    out_ns = (int) (((pxmax - pxmin) / 
        grid->gridbands[ref_band_index][sca_index].pixsize) + 1.5);

    /* Pad according to nominal scene center to WRS center. */
    xp = IAS_MAX(abs(pxmax - x), abs(pxmin - x) ) - (pxmax - pxmin) / 2.0;
    yp = IAS_MAX(abs(pymax - y), abs(pymin - y) ) - (pymax - pymin) / 2.0;
    pad_nl = (int) ((fabs(yp) / 
        grid->gridbands[ref_band_index][sca_index].pixsize) + 1.5);
    pad_ns = (int) ((fabs(xp) / 
        grid->gridbands[ref_band_index][sca_index].pixsize) + 1.5);
    out_nl = out_nl + 2 * pad_nl;
    out_ns = out_ns + 2 * pad_ns;

    /* Make sure output lines and samples are ad odd number */
    if (out_nl%2 == 0)
        out_nl = out_nl + 1;
    if (out_ns%2 == 0)
        out_ns = out_ns + 1;


    /* Store NL and NS in grid structure. */
    for (band_index = 0; band_index < grid->nbands; band_index++)
    {
        /* skip bands not available */
        if (!grid->bands_available[band_index])
            continue;

        for (sca_index = 0; sca_index < grid->scas_per_band[band_index];
             sca_index++)
        {
            grid->gridbands[band_index][sca_index].lines = 
                (int) ((grid->gridbands[ref_band_index][sca_index].pixsize / 
                grid->gridbands[band_index][sca_index].pixsize) * 
                (out_nl - 1)) + 1;
            grid->gridbands[band_index][sca_index].samps = 
                (int) ((grid->gridbands[ref_band_index][sca_index].pixsize / 
                grid->gridbands[band_index][sca_index].pixsize) * 
                (out_ns - 1)) + 1;
        }
    }

    sin_angle = sin(rot_angle);
    cos_angle = cos(rot_angle);

    /* Initialize first sca flag. */
    first_sca = TRUE;

    /* Loop through all the bands.   */
    for (band_loop_index = 0; band_loop_index < grid->nbands; 
        band_loop_index++)
    {
        int poly_copy_index = -1; /* index of band of the same resolution to
                                     use as the source for copying previously
                                     calculated poly_x and poly_y values */

        /* skip bands not available */
        if (!grid->bands_available[band_loop_index])
            continue;

        /* Check to see whether it has already been calculated for a
           previous band with the same pixel size.  If there is one, save
           the index to it. */
        for (search_band_index = 0; search_band_index < band_loop_index;
             search_band_index++)
        {
            if (grid->bands_available[search_band_index])
            {
                if (grid->gridbands[search_band_index][0].pixsize
                    == grid->gridbands[band_loop_index][0].pixsize)
                {
                    poly_copy_index = search_band_index;
                    break;
                }
            }
        }

        for (sca_index = 0; sca_index < grid->scas_per_band[band_loop_index];
             sca_index++) 
        {
            /* Get the reference position and the number of lines and samples
               in the current band resolution.                             */
            nl = grid->gridbands[band_loop_index][sca_index].lines;
            ns = grid->gridbands[band_loop_index][sca_index].samps;

            line_center = (double) (nl - 1.0) / 2.0;
            samp_center = (double) (ns - 1.0) / 2.0;

            /* Calculate the upper left coordinates.   */
            ul_dx = -samp_center * 
                grid->gridbands[band_loop_index][sca_index].pixsize;
            ul_dy = line_center * 
                grid->gridbands[band_loop_index][sca_index].pixsize;
            ul_xp =  ul_dx * cos_angle + ul_dy * sin_angle + x;
            ul_yp = -ul_dx * sin_angle + ul_dy * cos_angle + y;

            /* Calculate the upper right coordinates.   */
            ur_dx = ((ns-1.0) - samp_center) *
                grid->gridbands[band_loop_index][sca_index].pixsize;
            ur_dy = line_center * 
                grid->gridbands[band_loop_index][sca_index].pixsize;
            ur_xp =  ur_dx * cos_angle + ur_dy * sin_angle + x;
            ur_yp = -ur_dx * sin_angle + ur_dy * cos_angle + y;

            /* Calculate the lower right coordinates.   */
            lr_dx = ((ns-1.0) - samp_center) 
                * grid->gridbands[band_loop_index][sca_index].pixsize;
            lr_dy = (line_center - (nl-1.0)) 
                * grid->gridbands[band_loop_index][sca_index].pixsize;
            lr_xp =  lr_dx * cos_angle + lr_dy * sin_angle + x;
            lr_yp = -lr_dx * sin_angle + lr_dy * cos_angle + y;

            /* Calculate the lower left coordinates.   */
            ll_dx = -samp_center * 
                grid->gridbands[band_loop_index][sca_index].pixsize;
            ll_dy = (line_center - (nl-1.0)) *
                grid->gridbands[band_loop_index][sca_index].pixsize;
            ll_xp =  ll_dx * cos_angle + ll_dy * sin_angle + x;
            ll_yp = -ll_dx * sin_angle + ll_dy * cos_angle + y;

            /* Save the first corners calculated and compare the rest to them
               to make sure they are all the same. */
            if (first_sca == TRUE)
            {  
                first_sca = FALSE;
                ul_xp_old = ul_xp;
                ul_yp_old = ul_yp;
                ur_xp_old = ur_xp;
                ur_yp_old = ur_yp;
                ll_xp_old = ll_xp;
                ll_yp_old = ll_yp;
                lr_xp_old = lr_xp;
                lr_yp_old = lr_yp;
            }
            else if ((ul_xp != ul_xp_old || ul_yp != ul_yp_old ||
                      ur_xp != ur_xp_old || ur_yp != ur_yp_old ||
                      ll_xp != ll_xp_old || ll_yp != ll_yp_old ||
                      lr_xp != lr_xp_old || lr_yp != lr_yp_old) &&
                    grid->nbands != 1)
            {
                IAS_LOG_ERROR(
                    "The corner points for bands %d and %d do not match",
                    ias_sat_attr_convert_band_index_to_number(band_loop_index),
                    ias_sat_attr_convert_band_index_to_number(ref_band_index));
                return ERROR;
            }

            /* only calculate the following for the first SCA of each band */
            if (sca_index != 0)
                continue;

            /* Calculate the polynomials for all available band resolutions. */
            if (poly_copy_index != -1)
            {
                /* A previous band has the same pixel size, so copy the poly
                   values instead of calculating the same values again */
                for (i = 0; i < 4; i++)
                {
                    poly_x[band_loop_index][i] = poly_x[poly_copy_index][i];
                    poly_y[band_loop_index][i] = poly_y[poly_copy_index][i];
                }
            }
            else
            {
                /* Now that we have calculated the corners in projection 
                   coordinates, calculate the first order polynomial 
                   coefficients which map output line/samples to output 
                   projection coordinates.  The 4x3 matrix is populated as 
                   shown below, QR-factorized and then Ax = b is solved for two
                   different vectors b; one for samples, the other for lines. 
                   The matrix A depicted below is in column-major format, 
                   which is how QRfactorization interprets the array by 
                   default. */
                double A[12];      /* Matrix used by QRsolve for LSQ fit */
                double v[4];       /* Vector used in above operation */
                double px[4];      /* Poly_x used for calculations */
                double py[4];      /* Poly_y used for calculations */

                A[0]  = 1.0;
                A[1]  = 1.0;
                A[2]  = 1.0;
                A[3]  = 1.0;
                A[4]  = ul_xp;
                A[5]  = ur_xp;  /*     +-                                 -+ */
                A[6]  = lr_xp;  /*     |  1  upper-left  X  upper-left  Y  | */
                A[7]  = ll_xp;  /* A = |  1  upper-right X  upper-right Y  | */
                A[8]  = ul_yp;  /*     |  1  lower-right X  lower-right Y  | */
                A[9]  = ur_yp;  /*     |  1  lower-left  X  lower-left  Y  | */
                A[10] = lr_yp;  /*     +-                                 -+ */
                A[11] = ll_yp;

                /* Obtain QR factorization */
                ias_math_matrix_QRfactorization(A, 4, 3, v, 0); 

                /* Vector one: Find the coefficients to transform output samples
                   to output coordinates in projection space.  Note that the 
                   call to ias_math_matrix_QRsolve() modifies the values at 
                   poly_x for part of the output (this vector becomes the 
                   before-mentioned coefficients.)                           */
                px[0] = 0.0;
                px[1] = ns - 1.0;
                px[2] = ns - 1.0;
                px[3] = 0.0;
                ias_math_matrix_QRsolve(A, 4, 3, v, px, 0);

                /* Vector two: Find the coefficients to transform output lines
                   to output coordinates in projection space.                */
                py[0] = 0.0;
                py[1] = 0.0;
                py[2] = nl - 1.0;
                py[3] = nl - 1.0;
                ias_math_matrix_QRsolve(A, 4, 3, v, py, 0);

                /* Copy the poly's into their respective band resolution.   */
                for (i = 0; i < 4; i++)
                {
                    poly_x[band_loop_index][i] = px[i];
                    poly_y[band_loop_index][i] = py[i];
                }
            }

        } /* Loop for SCAs */

    } /* Loop for bands */

    /* Store the corners in the grid structure.   */
    grid->corners.upleft.x  = ul_xp;
    grid->corners.upleft.y  = ul_yp;
    grid->corners.upright.x = ur_xp;
    grid->corners.upright.y = ur_yp;
    grid->corners.loleft.x  = ll_xp;
    grid->corners.loleft.y  = ll_yp;
    grid->corners.loright.x = lr_xp;
    grid->corners.loright.y = lr_yp;
    grid->proj_center_pt.x = x;
    grid->proj_center_pt.y = y;

    return SUCCESS;
}
