/******************************************************************************
NAME:           create_grid

PURPOSE:
To generate the mapping grid defining a transformation from the image's
satellite perspective to a user specified output projection.

RETURN VALUE:   Type = int
Value         Description
-----         -----------
EXIT_SUCCESS  Successful completion
EXIT_FAILURE  Operation failed

NOTES:
  Generates a mapping grid that defines a transformation from
  the image's satellite perspective to a user specified output
  projection.  This routine will read the model file.
  Then the output framing coordinates are determined from the user's
  input.  The frame may be map-oriented or path-oriented.  Once the
  frame is determined in output space, the input space is gridded.
  Then the grid in input space is mapped to the output space.
  Transformation coefficients to transform a grid cell from input to
  output space are determined, as well as coefficients to transform
  a grid cell from output to input space.  These coefficients are
  stored in the grid for later use in resampling and band-to-band
  calibration.  The rough mapping polynomial coefficients are also
  calculated and stored in the grid for later use by resampling. 
  The grid structure is then written to an output file.

  The grid references the upper-left corner to be (0,0).

  For all framing, the optimal band list is retrieved from the CPF.

  Create Grid will search the list for the most optimal band that was
  specified.  This band information will then be used for minbox
  and path-oriented framing options.

ALGORITHM REFERENCES:
See ias_grid.h for a description of the grid structure.
See ias_los_model.h for a description of the model structure.
See ias_cpf.h for a description of the CPF structure.

******************************************************************************/
#include <unistd.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>                   /* String functions */
#include "gctp.h"                     /* Projection parameters */
#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "ias_geo.h"
#include "create_grid.h"

/* Degree of the rough mapping polynomial. */
#define ROUGH_DEGREE 1

static IAS_GEO_PROJ_TRANSFORMATION *geo_degrees2proj = NULL;
/* Projection transformation to go from the geographic projection in degrees to
   the selected output projection */

static IAS_GEO_PROJ_TRANSFORMATION *geo_radians2proj = NULL;
/* Projection transformation to go from the geographic projection in radians to
   the selected output projection */

static IAS_GEO_PROJ_TRANSFORMATION *geo_proj2degrees = NULL;
/* Projection transformation to go from the selected output projection to 
    the geographic projection in degrees  */

const IAS_GEO_PROJ_TRANSFORMATION *get_proj_transform_degrees2proj()
{
    return geo_degrees2proj;
}

const IAS_GEO_PROJ_TRANSFORMATION *get_proj_transform_radians2proj()
{
    return geo_radians2proj;
}

const IAS_GEO_PROJ_TRANSFORMATION *get_proj_transform_proj2degrees()
{
    return geo_proj2degrees;
}

int main(int argc, char *argv[])
{
    /* Variable declarations. */
    IAS_GRID_TYPE grid;           /* Grid structure */
    IAS_LOS_MODEL *model;         /* model structure */
    IAS_CPF *cpf;                  /* CPF constants */
    char cal_parm_file[PATH_MAX]; /* Calibration parameter filename */
    char model_filename[PATH_MAX];/* Model filename */
    char omf_filename[PATH_MAX];  /* OMF filename */
    char dem_filename[PATH_MAX];  /* Optional DEM filename */
    struct IAS_DBL_XY frame_coords[2]; /* User-entered frame coordinates;
                                     there are 2 sets of X/Y pairs */
    struct IAS_DBL_LS lscoords;   /* User-entered l/s coordinates */
    int band_index;               /* Band index */
    int sca_index;                /* SCA index */
    int computed_zone = -1;       /* Computed zone value for the scene */
    int path;                     /* WRS path */
    int curr_band_index;          /* Current band reference position */
    int number_lines[IAS_MAX_NBANDS]; /* Number of lines in the model */
    int number_samples[IAS_MAX_NBANDS]; /* Number of samples in the model */
    double row;                   /* WRS row */
    double poly_x[IAS_MAX_NBANDS][4]; /* Per-band X polynomial coefficients
                                         (4 coefs for each band resolution) */
    double poly_y[IAS_MAX_NBANDS][4]; /* Per-band Y polynomial coefficients
                                         (4 coefs for each band resolution) */
    int optimal_band_index;       /* Optimal band index of all specified
                                     bands for grid processing */
    int optimal_band_number;      /* Band number for optimal band index */
    int band_numbers[IAS_MAX_NBANDS]; /* Unique user-selected bands */
    DET_OFFSET_STATS det_offset_stats[IAS_MAX_NBANDS * IAS_MAX_NSCAS];
                                  /* detector offset stats */
    int nbands;                   /* Number of bands in the band indexes 
                                     array */
    int zone_valid;               /* Flag to specify if the UTM zone is 
                                     valid */
    int num;                      /* Number of z planes */
    int zeroplane;                /* Zero plane */ 
    float zmax = 0;               /* Maximum elevation */
    float zmin = 0;               /* Minimum elevation */
    int processing_pass;          /* Pass through WO processing flow */
    int max_procs_to_use;         /* Maximum number of processors to use */
    int satellite_number_bands;   /* Number of normal bands for the sat */
    int line_density;             /* Grid line density for band type */
    int sample_density;           /* Grid sample density for band type */
    int elevation_density;        /* Grid elevation density for band type */
    int band_number;              /* Band number */
    int lunar_sca_number = -1;    /* SCA number containing image of the
                                     moon; initialize to -1 to ensure this
                                     parameter is not processed if it isn't
                                     a lunar collect or is not available */
    int no_elevation_required_flag; /* Flag that indicates no elevation is
                                       required.  Mainly an indication that the
                                       grid is being created for stray light
                                       correction and some checks can be
                                       relaxed. */
    const struct IAS_CPF_EARTH_CONSTANTS *cpf_earth;
    const struct IAS_CPF_GEO_SYSTEM *cpf_geo_sys;
    IAS_GRID_FILE *grid_file;


    /* Initialize the detector offset stats */
    memset(det_offset_stats, 0, sizeof(det_offset_stats));

    /* Initialize logging library */
    if (ias_log_initialize("create_grid") != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing logging library");
        return EXIT_FAILURE;
    }

    /* Check arguments and get the parameter file name from the arguments. */
    if (argc != 2)
    {
        IAS_LOG_ERROR("Usage: create_grid parameter_file");
        return EXIT_FAILURE;
    }

    /* Get the grid parameters. */
    if (get_parms_and_initialize_grid(argv[1], cal_parm_file, model_filename, 
            omf_filename, dem_filename, frame_coords, &lscoords,
            &path, &row, &zmax, &zmin, &processing_pass, band_numbers, &nbands,
            &max_procs_to_use, &no_elevation_required_flag, &grid, &model)
            != SUCCESS)
    {
        IAS_LOG_ERROR("Reading the parameters from the parameter file");
        return EXIT_FAILURE;
    }

    /* Print the starting message. */
    IAS_LOG_INFO("Initiate Create Grid...");

    /* If there aren't any bands to process, error out */
    if (nbands <= 0)
    {
        IAS_LOG_ERROR("None of the requested bands are present in the model");
        ias_grid_free(&grid);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }

    /* Set number of SCAs to actual value from model. */
    for (band_index = 0; band_index < model->sensor.band_count; band_index++)
    {
        grid.scas_per_band[band_index] 
            = model->sensor.bands[band_index].sca_count;
    }

    /* Read the CPF. */
    cpf = ias_cpf_read(cal_parm_file);
    if (!cpf)
    {
        IAS_LOG_ERROR("Cannot read CPF file %s", cal_parm_file);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }

    /* Get the needed groups from the CPF. */
    cpf_earth = ias_cpf_get_earth_const(cpf);
    if (!cpf_earth)
    {
        IAS_LOG_ERROR("Retrieving Earth constants group from CPF file %s", 
            cal_parm_file);
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }
    cpf_geo_sys = ias_cpf_get_geo_system(cpf);
    if (!cpf_geo_sys)
    {
        IAS_LOG_ERROR("Retrieving Geo System group from CPF file %s",
            cal_parm_file);
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }

    /* For lunar and stellar acquisitions, initialize the number of leap
       seconds since the geo library routines that are called for lunar and
       stellar processing will need it initialized */
    if (model->acquisition_type == IAS_LUNAR
        || model->acquisition_type == IAS_STELLAR)
    {
        int found = 0;

        /* Find the first available band for the UTC epoch time */
        for (band_index = 0; band_index < model->sensor.band_count;
            band_index++)
        {
            if (model->sensor.bands[band_index].band_present)
            {
                found = 1;
                break;
            }
        }
        if (!found)
        {
            IAS_LOG_ERROR("No bands found in the model");
            ias_cpf_free(cpf);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            return EXIT_FAILURE;
        }

        /* Initialize the leap seconds using the UTC epoch time for the band */
        if (ias_math_init_leap_seconds_from_UTC_time(
            model->sensor.bands[band_index].utc_epoch_time,
            &cpf_earth->leap_seconds_data) != SUCCESS)
        {
            IAS_LOG_ERROR("Initializing leap seconds");
            ias_cpf_free(cpf);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            return EXIT_FAILURE;
        }
    }

    /* Store CPF's spheriod and datum constants in the grid. */
    grid.spheroid = cpf_earth->spheroid_code;
    strcpy(grid.datum, cpf_earth->datum);

    /* If the elevation range is invalid, get the range from a provided DEM
       file */
    if ((zmin == INVALID_ELEVATION) || (zmax == INVALID_ELEVATION))
    {
        /* Read the minimum and maximum elevation values from the DEM. */
        if (get_dem_elevation_range(dem_filename, &zmin, &zmax) != SUCCESS)
        {
            IAS_LOG_ERROR("Failed to read elevation values from DEM %s",
                dem_filename);
            ias_cpf_free(cpf);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            return EXIT_FAILURE;
        }
    }

    IAS_LOG_INFO("Elevation range %8.2f %8.2f", zmin, zmax);

    /* Fix number of planes, and change zmin and zmax, so that
       a zero plane exists. */
    if (zmin < 0.0 && zmax < 0.0)
    {
        zmax = 0.0;
    }
    else if (zmax > 0.0 && zmin > 0.0)
    {
        zmin = 0.0;
    }

    /* Find the number of normal bands for the satellite. */
    satellite_number_bands = ias_sat_attr_get_normal_band_count();
    if (satellite_number_bands == ERROR) 
    {
        IAS_LOG_ERROR("Invalid number of bands %d retrieved for normal "
            "band count", satellite_number_bands);
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }

    /* Store the grid density, degree, poly_lines, poly samps, and SCA number
       in grid. */
    for (band_index = 0; band_index < satellite_number_bands; band_index++)
    {
        float band_zmax; /* Maximum elevation for band */

        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Invalid band number %d", band_number);
            ias_cpf_free(cpf);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            return EXIT_FAILURE;
        }

        const IAS_BAND_ATTRIBUTES *band_attributes =
            ias_sat_attr_get_band_attributes(band_number);
        if (band_attributes == NULL)
        {
            IAS_LOG_ERROR("Retrieving attributes for band %d", band_number);
            ias_cpf_free(cpf);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            return EXIT_FAILURE;
        }

        if ((band_attributes->spectral_type == IAS_SPECTRAL_VNIR)
            || (band_attributes->spectral_type == IAS_SPECTRAL_SWIR))
        {
            line_density = cpf_geo_sys->ms_grid_density_lines;
            sample_density = cpf_geo_sys->ms_grid_density_samples; 
            elevation_density = cpf_geo_sys->ms_grid_density_elev; 
        }
        else if (band_attributes->spectral_type == IAS_SPECTRAL_PAN)
        {
            line_density = cpf_geo_sys->pan_grid_density_lines;
            sample_density = cpf_geo_sys->pan_grid_density_samples; 
            elevation_density = cpf_geo_sys->pan_grid_density_elev; 
        }
        else if (band_attributes->spectral_type == IAS_SPECTRAL_THERMAL)
        {
            line_density = cpf_geo_sys->tirs_grid_density_lines;
            sample_density = cpf_geo_sys->tirs_grid_density_samples; 
            elevation_density = cpf_geo_sys->tirs_grid_density_elev;
        }
        else
        {
            IAS_LOG_ERROR("Invalid spectral type.");
            ias_cpf_free(cpf);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            return EXIT_FAILURE;
        }

        /* If zmin == zmax, which the zeroplane check should limit to the case
           where both are 0, keep the zmin value and make the band's zmax value
           zmin + the band's elevation_density. */
        band_zmax = zmax;
        if (zmin == band_zmax)
        {
            band_zmax = zmin + elevation_density;
        }

        /* band_zmax + 1 is used when calculating the number of elevation 
           planes because the highest elevation plane must be greater than 
           all elevations in the image when calling routine ols2ils. */
        num = (int)(ceil((band_zmax + 1) / elevation_density) + 
            ceil((-zmin) / elevation_density)) + 1;
        zeroplane = (int)(ceil)((-zmin / elevation_density));

        for (sca_index = 0; sca_index < grid.scas_per_band[band_index]; 
            sca_index++)
        {
            grid.gridbands[band_index][sca_index].grid_cell_size.line = 
                line_density;
            grid.gridbands[band_index][sca_index].grid_cell_size.samp = 
                sample_density;
            grid.gridbands[band_index][sca_index].zspacing = 
                elevation_density;
            grid.gridbands[band_index][sca_index].degree = ROUGH_DEGREE;
            grid.gridbands[band_index][sca_index].inv_cell_lines = 
                1.0 / line_density;
            grid.gridbands[band_index][sca_index].inv_cell_samps = 
                1.0 / sample_density;
            grid.gridbands[band_index][sca_index].sca_number = sca_index + 1;
            grid.gridbands[band_index][sca_index].nzplanes = num;
            grid.gridbands[band_index][sca_index].zeroplane = zeroplane;
        }
    }

    /* Read the number of lines and samples from the model. */
    set_number_lines_samples(model, number_lines, number_samples);

    /* Determine the number of grid lines and samples. */
    set_number_grid_lines_samples(number_lines, number_samples, &grid);

    /* Set the grid correction type and grid satellite number. */
    grid.correction_type = model->correction_type;
    grid.satnum = model->satellite_number; 

    /* Malloc space for the grid pointers. */
    if (ias_grid_malloc(&grid, band_numbers, nbands) != SUCCESS)
    {
        IAS_LOG_ERROR("Mallocing space for the grid");
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }

    /* If the TARGET_PROJECTION is UTM, verify the correct zone was given.  
     * If the user-entered zone is 0, use the calculated zone and continue. */
    if (grid.code == UTM)
    {
        /* Use the information from the first band available. */
        for (curr_band_index = 0; curr_band_index < grid.nbands; 
            curr_band_index++)
        {
            /* skip bands not available */
            if (grid.bands_available[curr_band_index])
                break;
        }

        /* Use the center SCA for validating the zone */
        sca_index = grid.scas_per_band[curr_band_index] / 2;

        if (validate_utm_zone(number_lines[curr_band_index], 
            number_samples[curr_band_index], curr_band_index, sca_index, 
            model, grid.zone, &computed_zone, &zone_valid) != SUCCESS) 
        {
            IAS_LOG_ERROR("Validating the UTM zone");
            ias_cpf_free(cpf);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            return EXIT_FAILURE;
        }

        /* If the zone is zero then use the calculated zone. */
        if (grid.zone == 0)
        {
            IAS_LOG_INFO("Automatically computing UTM zone.  "
                "Computed zone is %d", computed_zone);
            grid.zone = computed_zone;
        }

        /* If the zone is not valid then exit the application. */
        else if (!zone_valid)
        {
            /* Only consider a zone mismatch to be an error if not making the
               grid for stray light correction (which is when no elevation is
               required).  The method to calculate the UTM zone for comparison
               isn't really valid for the stray light correction grid since the
               number of lines in the model is for the 3 scenes of TIRS data
               (some which may be partial scenes). */
            if (!no_elevation_required_flag)
            {
                IAS_LOG_ERROR("User-entered zone is %d.  Computed zone is %d", 
                    grid.zone, computed_zone);
                IAS_LOG_ERROR("The zone provided does not correspond to the "
                    "image scene center longitude");
                ias_cpf_free(cpf);
                ias_grid_free(&grid);
                ias_los_model_free(model);
                return EXIT_FAILURE;
            }
            else
            {
                /* Provide a warning instead */
                IAS_LOG_WARNING("User-entered zone is %d.  Computed zone is %d",
                    grid.zone, computed_zone);
                IAS_LOG_WARNING("The zone provided does not correspond to the "
                    "image scene center longitude");
            }
        }
    }

    /* Determine the "most optimal" band. */
    if (determine_optimal_band(cpf, grid.bands_available, &optimal_band_index) 
        != SUCCESS) 
    {
        IAS_LOG_ERROR("Optimal band was not found");
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }

    optimal_band_number = ias_sat_attr_convert_band_index_to_number(
        optimal_band_index);
    if (optimal_band_number == ERROR)
    {
        IAS_LOG_ERROR("Invalid optimal band number %d", optimal_band_number);
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }

    /* Calculate the active image area bounds, based on optimal band. */
    if (ias_los_model_calc_scene_corners(model, optimal_band_number, 0, 
        number_lines[optimal_band_index] - 1, 0.0, 
        &grid.active_corners[optimal_band_index]) != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating active image area for optimal band "
            "number %d", optimal_band_number);
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }

    for (band_index = 0; band_index < grid.nbands; band_index++)
    {
        /* Skip bands not available */
        if (!grid.bands_available[band_index])
            continue;

        /* Skip the optimal band, which we already set. */
        if (band_index == optimal_band_index)
            continue;

        /* Copy active corners from the optimal band to the current band. */
        grid.active_corners[band_index] 
            = grid.active_corners[optimal_band_index];
    }

    if (model->acquisition_type == IAS_STELLAR || 
        model->acquisition_type == IAS_LUNAR)
    {
        /* Initialize the novas library */
        if (ias_geo_initialize_novas() != SUCCESS)
        {
            IAS_LOG_ERROR("Initializing the NOVAS library");
            ias_cpf_free(cpf);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            return EXIT_FAILURE;
        }
    }

    if (model->acquisition_type == IAS_EARTH)
    {
        IAS_PROJECTION geo_proj;
        double geo_parms[IAS_PROJ_PARAM_SIZE];
        IAS_PROJECTION target_proj;
        int units;
        int i;

        /* Create the projection transformation from geographic to the target
           projection */
        if (ias_geo_get_units(grid.units, &units) != SUCCESS)
        {
            IAS_LOG_ERROR("Converting projection units %s to unit code",
                grid.units);
            ias_cpf_free(cpf);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            return EXIT_FAILURE;
        }

        for (i = 0; i < IAS_PROJ_PARAM_SIZE; i++)
            geo_parms[i] = 0.0;
        ias_geo_set_projection(GEO, NULLZONE, DEGREE, grid.spheroid, geo_parms,
            &geo_proj);
        ias_geo_set_projection(grid.code, grid.zone, units, grid.spheroid,
            grid.projprms, &target_proj);

        geo_degrees2proj = ias_geo_create_proj_transformation(&geo_proj,
            &target_proj);
        if (!geo_degrees2proj)
        {
            IAS_LOG_ERROR("Creating degrees to map projection transformation");
            ias_cpf_free(cpf);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            return EXIT_FAILURE;
        }

        geo_proj2degrees = ias_geo_create_proj_transformation(&target_proj,
            &geo_proj);
        if (!geo_proj2degrees)
        {
            IAS_LOG_ERROR("Creating map projection to degrees transformation");
            ias_cpf_free(cpf);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            ias_geo_destroy_proj_transformation(geo_degrees2proj);
            return EXIT_FAILURE;
        }

        ias_geo_set_projection(GEO, NULLZONE, RADIAN, grid.spheroid, geo_parms,
            &geo_proj);
        geo_radians2proj = ias_geo_create_proj_transformation(&geo_proj,
            &target_proj);
        if (!geo_radians2proj)
        {
            IAS_LOG_ERROR("Creating radians to map projection transformation");
            ias_cpf_free(cpf);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            ias_geo_destroy_proj_transformation(geo_degrees2proj);
            ias_geo_destroy_proj_transformation(geo_proj2degrees);
            return EXIT_FAILURE;
        }

        /* Determine the frame in output space. */
        switch (grid.frame_type)
        {
            case IAS_GEOBOX:
            case IAS_PROJBOX:
            case IAS_UL_SIZE:
                /* Determine the "most optimal" band. */
                if (determine_optimal_band(cpf, grid.bands_available, 
                    &optimal_band_index) != SUCCESS) 
                {
                    IAS_LOG_ERROR("Optimal band was not found");
                    ias_cpf_free(cpf);
                    ias_grid_free(&grid);
                    ias_los_model_free(model);
                    ias_geo_destroy_proj_transformation(geo_degrees2proj);
                    ias_geo_destroy_proj_transformation(geo_radians2proj);
                    ias_geo_destroy_proj_transformation(geo_proj2degrees);
                    return EXIT_FAILURE;
                }

                if (get_frame_north_up(&grid, optimal_band_index, frame_coords,
                    lscoords) != SUCCESS)
                {
                    IAS_LOG_ERROR("Determining the north-up frame");
                    ias_cpf_free(cpf);
                    ias_grid_free(&grid);
                    ias_los_model_free(model);
                    ias_geo_destroy_proj_transformation(geo_degrees2proj);
                    ias_geo_destroy_proj_transformation(geo_radians2proj);
                    ias_geo_destroy_proj_transformation(geo_proj2degrees);
                    return EXIT_FAILURE;
                }
                break;

            case IAS_MINBOX:
                /* Determine the "most optimal" band. */
                if (determine_optimal_band(cpf, grid.bands_available, 
                    &optimal_band_index) != SUCCESS)
                {
                    IAS_LOG_ERROR("Optimal band was not found");
                    ias_cpf_free(cpf);
                    ias_grid_free(&grid);
                    ias_los_model_free(model);
                    ias_geo_destroy_proj_transformation(geo_degrees2proj);
                    ias_geo_destroy_proj_transformation(geo_radians2proj);
                    ias_geo_destroy_proj_transformation(geo_proj2degrees);
                    return EXIT_FAILURE;
                }

                if (get_frame_minbox(model, optimal_band_index, &grid) 
                    != SUCCESS)
                {
                    IAS_LOG_ERROR("Determining the minbox frame");
                    ias_cpf_free(cpf);
                    ias_grid_free(&grid);
                    ias_los_model_free(model);
                    ias_geo_destroy_proj_transformation(geo_degrees2proj);
                    ias_geo_destroy_proj_transformation(geo_radians2proj);
                    ias_geo_destroy_proj_transformation(geo_proj2degrees);
                    return EXIT_FAILURE;
                }
                break;

            case IAS_MAXBOX:
                if (get_frame_maxbox(model, number_lines, number_samples, 
                    &grid) != SUCCESS)
                {
                    IAS_LOG_ERROR("Determining the maxbox frame");
                    ias_cpf_free(cpf);
                    ias_grid_free(&grid);
                    ias_los_model_free(model);
                    ias_geo_destroy_proj_transformation(geo_degrees2proj);
                    ias_geo_destroy_proj_transformation(geo_radians2proj);
                    ias_geo_destroy_proj_transformation(geo_proj2degrees);
                    return EXIT_FAILURE;
                }
                break;

            case IAS_PATH_ORIENTED:
                if (get_frame_path_oriented(&grid, cpf, path, row, poly_x, 
                    poly_y) != SUCCESS)
                {
                    IAS_LOG_ERROR("Determining the path_oriented frame");
                    ias_cpf_free(cpf);
                    ias_grid_free(&grid);
                    ias_los_model_free(model);
                    ias_geo_destroy_proj_transformation(geo_degrees2proj);
                    ias_geo_destroy_proj_transformation(geo_radians2proj);
                    ias_geo_destroy_proj_transformation(geo_proj2degrees);
                    return EXIT_FAILURE;
                }
                break;

            case IAS_PATH_MINBOX:
                /* Determine the "most optimal" band. */ 
                if (determine_optimal_band(cpf, grid.bands_available, 
                    &optimal_band_index) != SUCCESS)
                {
                    IAS_LOG_ERROR("Optimal band was not found");
                    ias_cpf_free(cpf);
                    ias_grid_free(&grid);
                    ias_los_model_free(model);
                    ias_geo_destroy_proj_transformation(geo_degrees2proj);
                    ias_geo_destroy_proj_transformation(geo_radians2proj);
                    ias_geo_destroy_proj_transformation(geo_proj2degrees);
                    return EXIT_FAILURE;
                }
                if (get_frame_path_maxmin_box(model, optimal_band_index, 
                    number_lines, number_samples, cpf, &grid, path, row, 
                    poly_x, poly_y) != SUCCESS)
                {
                    IAS_LOG_ERROR("Determining the max/min box frame");
                    ias_cpf_free(cpf);
                    ias_grid_free(&grid);
                    ias_los_model_free(model);
                    ias_geo_destroy_proj_transformation(geo_degrees2proj);
                    ias_geo_destroy_proj_transformation(geo_radians2proj);
                    ias_geo_destroy_proj_transformation(geo_proj2degrees);
                    return EXIT_FAILURE;
                }
                break;

            case IAS_PATH_MAXBOX:
                if (get_frame_path_maxmin_box(model, optimal_band_index, 
                    number_lines, number_samples, cpf, &grid, path, row, 
                    poly_x, poly_y) != SUCCESS)
                {
                    IAS_LOG_ERROR("Determining the path_maxbox frame");
                    ias_cpf_free(cpf);
                    ias_grid_free(&grid);
                    ias_los_model_free(model);
                    ias_geo_destroy_proj_transformation(geo_degrees2proj);
                    ias_geo_destroy_proj_transformation(geo_radians2proj);
                    ias_geo_destroy_proj_transformation(geo_proj2degrees);
                    return EXIT_FAILURE;
                }
                break;

            default:
            {
                IAS_LOG_ERROR("Incorrect framing type was specified");
                ias_cpf_free(cpf);
                ias_grid_free(&grid);
                ias_los_model_free(model);
                ias_geo_destroy_proj_transformation(geo_degrees2proj);
                ias_geo_destroy_proj_transformation(geo_radians2proj);
                ias_geo_destroy_proj_transformation(geo_proj2degrees);
                return EXIT_FAILURE;
            }
        }
    }
    else if (model->acquisition_type == IAS_LUNAR)
    {
        if (calc_lunar_size(model, &grid, &lunar_sca_number) != SUCCESS)
        {
            IAS_LOG_ERROR("Framing lunar acquisition");
            ias_geo_shutdown_novas();
            ias_cpf_free(cpf);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            return EXIT_FAILURE;
        }
    }
#if 0 /* TODO: We don't have a requirement for stellar acquisitions, but
        leaving this in as an indication of what would need to be done for
        stellar processing to be added */
    else if (model->acquisition_type == IAS_STELLAR)
    {
        if (calc_stellar_size(model, &grid) != SUCCESS)
        {
            IAS_LOG_ERROR("Framing lunar/stellar acquisition");
            ias_geo_shutdown_novas();
            ias_cpf_free(cpf);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            return EXIT_FAILURE;
        }
    }
#endif
    else
    {
        IAS_LOG_ERROR("Unknown acquisition type %d", model->acquisition_type);
        /* TODO: Remove this when IAS_STELLAR is added to the switch */
        if (model->acquisition_type == IAS_STELLAR)
            ias_geo_shutdown_novas();
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }

    /* Set the grid lines in input space.  */
    set_grid_input_lines_samples(number_lines, number_samples, &grid);

    /* Create the grid file */
    grid_file = ias_grid_create(grid.gridname);
    if (!grid_file)
    {
        IAS_LOG_ERROR("Error creating grid file %s", grid.gridname);
        if (model->acquisition_type == IAS_STELLAR || 
            model->acquisition_type == IAS_LUNAR)
            ias_geo_shutdown_novas();
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        ias_geo_destroy_proj_transformation(geo_degrees2proj);
        ias_geo_destroy_proj_transformation(geo_radians2proj);
        ias_geo_destroy_proj_transformation(geo_proj2degrees);
        return EXIT_FAILURE;
    }

    /* Populate the main information into the grid */
    if (build_grid(model, poly_x, poly_y, max_procs_to_use, &grid,
            det_offset_stats, grid_file) != SUCCESS)
    {
        IAS_LOG_ERROR("Making the grid");
        if (model->acquisition_type == IAS_STELLAR || 
            model->acquisition_type == IAS_LUNAR)
            ias_geo_shutdown_novas();
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        ias_geo_destroy_proj_transformation(geo_degrees2proj);
        ias_geo_destroy_proj_transformation(geo_radians2proj);
        ias_geo_destroy_proj_transformation(geo_proj2degrees);
        ias_grid_close(grid_file);
        return EXIT_FAILURE;
    }

    /* Close planetary ephemeris file. */
    if (model->acquisition_type == IAS_STELLAR || 
        model->acquisition_type == IAS_LUNAR)
    {
        if (ias_geo_shutdown_novas() != SUCCESS)
        {
            IAS_LOG_ERROR("Shutting down the NOVAS library");
            ias_cpf_free(cpf);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            ias_geo_destroy_proj_transformation(geo_degrees2proj);
            ias_geo_destroy_proj_transformation(geo_radians2proj);
            ias_geo_destroy_proj_transformation(geo_proj2degrees);
            ias_grid_close(grid_file);
            return EXIT_FAILURE;
        }
    }

    /* Print the even/odd detector offset statistics */
    if (print_det_offset_stats(det_offset_stats,
            IAS_MAX_NBANDS * IAS_MAX_NSCAS) != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating detector offset statistics");
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        ias_geo_destroy_proj_transformation(geo_degrees2proj);
        ias_geo_destroy_proj_transformation(geo_radians2proj);
        ias_geo_destroy_proj_transformation(geo_proj2degrees);
        ias_grid_close(grid_file);
        return EXIT_FAILURE;
    }

    /* destroy the transformations since they should no longer be needed */
    ias_geo_destroy_proj_transformation(geo_degrees2proj);
    geo_degrees2proj = NULL;
    ias_geo_destroy_proj_transformation(geo_radians2proj);
    geo_radians2proj = NULL;
    ias_geo_destroy_proj_transformation(geo_proj2degrees);
    geo_proj2degrees = NULL;

    /* Write the grid structure out to a file. */
    IAS_LOG_INFO("Writing output grid file headers");

    if (ias_grid_header_write(grid_file, &grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing grid header: %s", grid.gridname);
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        ias_grid_close(grid_file);
        return EXIT_FAILURE;
    }

    /* Write grid band header */
    if (ias_grid_band_header_write(grid_file, &grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing grid band header: %s", grid.gridname);
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        ias_grid_close(grid_file);
        return EXIT_FAILURE;
    }

    /* Close the grid file */
    if (ias_grid_close(grid_file) != SUCCESS)
    {
        IAS_LOG_ERROR("Error closing grid file: %s", grid.gridname);
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }

    /* Write the grid filename and the lunar SCA number (if available) to
       the OMF */
    if (update_omf(omf_filename, grid.gridname, processing_pass,
        lunar_sca_number) != SUCCESS)
    {
        IAS_LOG_ERROR("Error updating OMF file: %s", omf_filename);
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }

    /* Free the CPF structure. */
    ias_cpf_free(cpf);

    /* Free the grid. */
    if (ias_grid_free(&grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Freeing the grid structure");
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }

    /* Free the model. */
    ias_los_model_free(model);

    /* Print the completion message. */
    IAS_LOG_INFO("Successful completion");

    return EXIT_SUCCESS;
} /* done */
