#ifndef _CREATE_GRID_H_
#define _CREATE_GRID_H_

#include <limits.h>                   /* PATH_MAX definition  */

#include "ias_geo.h"                  /* Lunar projection type */
#include "ias_grid_io.h"              /* Grid I/O library */
#include "ias_model_io.h"             /* IAS model I/O library */

#define EPSILON 0.00001     /* Epsilon value to add to the center point */
#define NUM_FRAME_COORDINATES        4 /* 2 sets of Y/X coordinates */

/* This macro is used to convert Lunar and Stellar lat/long values to output
   image line, sample values.  Note:  The osamp calculation looks backwards but
   is OK because we frame lunar scenes with the maximum right ascension at the
   UL corner, making the coordinate system lefthanded so that the output frame
   is oriented as if we are looking up at the sky.   For earth looking
   projections, we are looking down so the sign is in the opposite direction. */
#define CONVERT_LUNAR_LAT_LONG_TO_LS(lat, lng, upleft_x, upleft_y, pixsize, \
    oline, osamp) \
{ \
    (oline) = ((upleft_y) - (lat)) / (pixsize); \
    (osamp) = ((upleft_x) - (lng)) / (pixsize); \
}

/* Define a structure to store the detector offset statistics */
typedef struct det_offset_stats
{
    int available;             /* This set of stats is available */
    int band_index;
    int sca_index;
    double min_line_oe[2];     /* Min of det offset in line dir for odd or
                                  even detectors (bias,scale) */
    double min_samp_oe[2];     /* Min of det offset in sample dir for odd or
                                  even detectors (bias,scale) */
    double max_line_oe[2];     /* Max of det offset in line dir for odd or
                                  even detectors (bias,scale) */
    double max_samp_oe[2];     /* Max of det offset in sample dir for odd or
                                  even detectors (bias,scale) */
    double median_line_oe[2];  /* Median of det offset in line dir for odd or
                                  even detectors (bias,scale) */
    double median_samp_oe[2];  /* Median of det offset in sample dir for odd or
                                  even detectors (bias,scale) */
    double mean_line_oe[2];    /* Mean of det offset in line dir for odd or 
                                  even detectors (bias,scale) */
    double mean_samp_oe[2];    /* Mean of det offset in sample dir for odd or
                                  even detectors (bias,scale) */
    double stdev_line_oe[2];   /* Stdev of det offset in line dir for odd or
                                  even detectors (bias,scale) */
    double stdev_samp_oe[2];   /* Stdev of det offset in sample dir for odd or
                                  even detectors (bias,scale) */
    double rms_line_oe[2];     /* RMS of det offset in line dir for odd or
                                  even detectors (bias,scale) */
    double rms_samp_oe[2];     /* RMS of det offset in sample dir for odd or
                                  even detectors (bias,scale) */
} DET_OFFSET_STATS;

/* Prototype definitions for the grid routines. */
int angle_to_map 
(
    IAS_GRID_TYPE *grid     /* I/O: Grid structure to examine/modify */
);

int calc_center_and_rotation_angle 
(
    int path,               /* I: WRS path number */
    double row,             /* I: WRS row number (may be fractional) */
    IAS_CPF *cpf,            /* I: Calibration parameters */
    IAS_GRID_TYPE *grid     /* I/O: Grid structure */
);

int calc_det_offsets 
(
    int band_index,             /* I: band index to process */
    int sca_index,              /* I: sca index to process */
    const IAS_LOS_MODEL *model, /* I: Model structure */
    double poly_x[4],           /* I: array of x polynomial coefficients */
    double poly_y[4],           /* I: array of y polynomial coefficients */
    double lunar_scale,         /* I: scale used for pixel size for 
                                      lunar/stellar acquisitions */
    IAS_GRID_TYPE *grid         /* I/O: Grid structure */
);

int calc_det_offset_stats
(
    IAS_GRID_BAND_TYPE *gridband, /* I/O: Grid structure  */
    DET_OFFSET_STATS *stats       /* O: calculated statistics */
);

int print_det_offset_stats
(
    const DET_OFFSET_STATS *stats_array,/* I: statistics to print */
    int stats_size 
);

int calc_lunar_size
(
    const IAS_LOS_MODEL *model,     /* I: model structure */
    IAS_GRID_TYPE *grid,            /* I/O: Grid structure */
    int *lunar_sca_number           /* O: SCA number containing the image
                                       of the moon */
);

int calc_map_coefs 
(
    IAS_GRID_BAND_TYPE *gridband /* I/O: Grid band structure */
);

int calc_odd_or_even_offsets
(
    const IAS_LOS_MODEL *model, /* I: model structure */
    const IAS_LUNAR_PROJECTION *lunar_proj,/* I: Lunar projection information */
    double poly_x[4],        /* I: array of x polynomial coefficients */
    double poly_y[4],        /* I: array of y polynomial coefficients */
    int band_index,          /* I: Band index */
    int sca_index,           /* I: SCA index */
    IAS_SENSOR_DETECTOR_TYPE dettype, /* I: detector type for forward 
                                   model calculations */
    double elev,             /* I: Current elevation */
    double scale,            /* I: Scale for lunar projections */
    int gline,               /* I: Input line location */
    int gsamp,               /* I: Input sample location */
    int zplane,              /* I: Elevation plane */
    IAS_GRID_TYPE *grid,     /* I/O: Grid structure */
    double *zline_even,      /* O: Line/even det offset value */
    double *zsamp_even,      /* O: Sample/even det offset value */
    double *zline_odd,       /* O: Line/odd det Offset value */
    double *zsamp_odd        /* O: Sample/odd det Offset value */
);

int calc_path_oriented_frame 
(
    IAS_GRID_TYPE *grid,     /* I/O: Grid structure */
    int units,               /* I: Projection units */
    double poly_x[IAS_MAX_NBANDS][4], /* O: Per-band sample-to-projection
                                            coefficients */
    double poly_y[IAS_MAX_NBANDS][4]  /* O: Per-band line-to-projection
                                            coefficients */
);

int calc_rough_map_coefs 
(
    IAS_GRID_TYPE *grid        /* I/O: Grid structure */
);

void set_grid_input_lines_samples 
(
    const int num_lines[IAS_MAX_NBANDS], /* I: Number of input lines */
    const int num_samps[IAS_MAX_NBANDS], /* I: Number of input samples */
    IAS_GRID_TYPE *grid                  /* I/O: Grid structure */
);

void set_number_grid_lines_samples 
(
    const int num_lines[IAS_MAX_NBANDS], /* I: Number of input lines */
    const int num_samps[IAS_MAX_NBANDS], /* I: Number of input samples */
    IAS_GRID_TYPE *grid                  /* I/O: Grid structure */
);

int determine_optimal_band 
(
    IAS_CPF *cpf,                /* I: CPF constants */
    const int *bands_available, /* I: Array of band available flags */
    int *optimal_band_index     /* O: Optimal band index in the list of bands */
);

int get_parms_and_initialize_grid 
(
    const char *param_filename,     /* I: Parameter file name */
    char cal_parm_file[PATH_MAX],   /* O: Calibration parameter filename */
    char model_file_name[PATH_MAX], /* O: Model file name */
    char omf_file_name[PATH_MAX],   /* O: OMF file name */
    char dem_file_name[PATH_MAX],   /* O: Optional DEM filename */
    struct IAS_DBL_XY frame_coords[NUM_FRAME_COORDINATES/2], /* O: User-input 
                                        frame coordinates to define the frame 
                                        of the output area */
    struct IAS_DBL_LS *lscoords,    /* O: Image coordinates of the first coord
                                        pair entered in frame_coords when proj
                                        params are being specified directly */
    int *path,                      /* O: WRS path number of the image */
    double *row,                    /* O: WRS (fractional) row number */
    float *zmax,                    /* O: Elevation plane maximum */
    float *zmin,                    /* O: Elevation plane minimum */
    int *processing_pass,           /* O: Pass through WO processing flow */
    int band_numbers[IAS_MAX_NBANDS], /* O: Unique user-selected bands */
    int *number_bands,              /* O: # of unique user-selected bands */
    int *max_processors_to_use,     /* O: Maximum number of processors to use */
    int *no_elevation_required_flag,/* O: Flag that indicates no elevation is
                                          required.  Mainly an indication that
                                          the grid is being created for stray
                                          light correction and some checks can
                                          be relaxed. */
    IAS_GRID_TYPE *grid,            /* O: Grid structure to be filled */
    IAS_LOS_MODEL **model           /* O: Model structure */
);

int calc_mapping_and_sensitivity 
(
    int band_index,             /* I: band index */
    int sca_index,              /* I: SCA index */
    const IAS_LOS_MODEL *model, /* I: Model structure */
    double poly_x[4],           /* I: array of x polynomial coefficients */
    double poly_y[4],           /* I: array of y polynomial coefficients */
    double lunar_scale,         /* I: scale used for pixel size for 
                                      lunar/stellar acquisitions */
    IAS_GRID_TYPE *grid,        /* I/O: Grid structure */
    int *blank                  /* Flag to specify if the output image is
                                   blank.  If one input pixel maps to an output
                                   pixel, then the image is not blank */

);

int build_grid
(
    const IAS_LOS_MODEL *model, /* I: Model structure */
    double poly_x[IAS_MAX_NBANDS][4], /* I: Per-band array of x polynomial
                                            coefficients */
    double poly_y[IAS_MAX_NBANDS][4], /* I: Per-band array of y polynomial
                                            coefficients */
    int max_processors_to_use,  /* I: Maximum number of processors to use */
    IAS_GRID_TYPE *grid,        /* I/O: Grid structure */
    DET_OFFSET_STATS *det_offset_stats,/* O: detector offset stats as 1D array
                                     IAS_MAX_NBANDS * IAS_MAX_NSCAS in size */
    IAS_GRID_FILE *grid_file    /* I: pointer to open grid to write to */
);

int get_frame_minbox
(
    const IAS_LOS_MODEL *model, /* I: Model structure */
    int opt_band_index,      /* I: Index of "optimal" band to work on */
    IAS_GRID_TYPE *grid      /* I/O: Grid structure */
);

int get_frame_maxbox
(
    const IAS_LOS_MODEL *model, /* I: Model structure */
    const int *num_lines,       /* I: Number lines in model for each band */ 
    const int *num_samps,       /* I: Number samps in model for each band */ 
    IAS_GRID_TYPE *grid         /* I/O: Grid structure */
);

int get_frame_north_up
(
    IAS_GRID_TYPE *grid,        /* I/O: Grid structure */
    int opt_band_index,         /* I: "Optimal" band index to work on */
    struct IAS_DBL_XY frame_coords[2],/* I: User-input frame coordinates to 
                                      define the frame of the output area */
    struct IAS_DBL_LS lscoords  /* I: Image coordinates of the first coord
                                      pair entered in frame_coords when proj
                                      params are being specified directly */
);

int pad_corners 
(
    double pixsize,          /* I: Pixel size for the coordinates */
    double *pxmax,           /* I/O: Maximum X projection coordinate */
    double *pxmin,           /* I/O: Minimum X projection coordinate */
    double *pymax,           /* I/O: Maximum Y projection coordinate */
    double *pymin            /* I/O: Minimum Y projection coordinate */
);

int get_frame_path_oriented
(
    IAS_GRID_TYPE *grid,     /* I/O: Grid structure */
    IAS_CPF *cpf,             /* I: CPF constants */
    int path,                /* I: Path of the image */
    double row,              /* I: Fractional row of the image */
    double poly_x[IAS_MAX_NBANDS][4],/* O: Per-band X polynomial coefficients */
    double poly_y[IAS_MAX_NBANDS][4] /* O: Per-band Y polynomial coefficients */
);

void set_number_lines_samples 
(
    const IAS_LOS_MODEL *model, /* I: Model structure */
    int nlines[IAS_MAX_NBANDS], /* O: Number of lines in image in
                                   band-referenced order */
    int nsamps[IAS_MAX_NBANDS]  /* O: Number of samples in image in 
                                   band-reference order */
);

int validate_utm_zone 
(
    int nlines,          /* I: Number of lines in image */
    int nsamps,          /* I: Number of samples in image */
    int band_index,      /* I: Index for the current band */
    int sca_index,       /* I: Index for the current SCA */
    const IAS_LOS_MODEL *model, /* I: Model structure */
    int zone,            /* I: Zone value to validate */
    int *computed_zone,  /* O: Computed zone value for the scene */
    int *valid           /* O: Flag to specify if the UTM zone was valid */
);

int get_frame_path_maxmin_box
(
    const IAS_LOS_MODEL *model, /* I: Model structure */
    int opt_band_index,  /* I: Optimal band index , used with MINBOX */
    const int *num_lines,  /* I: Number lines in model for optimal band */ 
    const int *num_samps,  /* I: Number samps in model for optimal band */ 
    IAS_CPF *cpf,         /* I: CPF to determine WRS lat/long and angle */
    IAS_GRID_TYPE *grid, /* I/O: Grid structure to use/modify */
    int path,            /* I: WRS path of product to provide */
    double row,          /* I: WRS row of product (may be fractional) */
    double poly_x[IAS_MAX_NBANDS][4], /* O: Per-band array of x polynomial
                                            coefficients */
    double poly_y[IAS_MAX_NBANDS][4]  /* O: Per-band array of y polynomial
                                            coefficients */
);

int calc_path_oriented_maxbox_frame 
(
    const IAS_LOS_MODEL *model, /* I: Model structure */
    const int *num_lines, /* I: Number lines in model for each band */ 
    const int *num_samps, /* I: Number samps in model for each band */ 
    IAS_GRID_TYPE *grid, /* I/O: Grid structure */
    double poly_x[IAS_MAX_NBANDS][4], /* O: Per-band array of x polynomial
                                            coefficients */
    double poly_y[IAS_MAX_NBANDS][4]  /* O: Per-band array of y polynomial
                                            coefficients */
);

int calc_path_oriented_minbox_frame 
(
    const IAS_LOS_MODEL *model, /* I: Model structure */
    int opt_band_index,  /* I: "Optimal" band index to work on */
    IAS_GRID_TYPE *grid, /* I/O: Grid structure */
    double poly_x[IAS_MAX_NBANDS][4], /* O: Per-band array of x polynomial
                                            coefficients */
    double poly_y[IAS_MAX_NBANDS][4]  /* O: Per-band array of y polynomial
                                            coefficients */
);

int update_omf
(
    const char *omf_filename,  /* I: Name of OMF */
    const char *grid_filename, /* I: Name of the grid data file */ 
    int processing_pass,       /* I: Number of WO processing flow pass */
    int lunar_sca_number       /* I: SCA number containing image of the
                                  moon; pass in -1 if not available */    
);

int get_dem_elevation_range
(
    char dem_file_name[PATH_MAX], /* I: DEM file name */
    float *zmin,                  /* O: Minimum elevation value in the DEM */
    float *zmax                   /* O: Maximimum elevation value in the DEM */
);

int calc_corner_point_output_space_locations
(
    const IAS_DBL_LS *ul_ls, /* I: Upper left line/sample for the corners */
    const IAS_DBL_LS *ur_ls, /* I: Upper right line/sample for the corners */
    const IAS_DBL_LS *ll_ls, /* I: Lower left line/sample for the corners */
    const IAS_DBL_LS *lr_ls, /* I: Lower right line/sample for the corners */
    int band_index,    /* I: Input band number */
    int sca_index,     /* I: SCA index */
    double elevation,  /* I: Target elevation (at line/sample) */
    const IAS_LOS_MODEL *model, /* I: LOS model information */
    IAS_GRID_TYPE *grid, /* I: Grid structure */
    IAS_DBL_XY *ul_xy, /* O: Upper left x/y for the corners */
    IAS_DBL_XY *ur_xy, /* O: Upper right x/y for the corners */
    IAS_DBL_XY *ll_xy, /* O: Lower left x/y for the corners */
    IAS_DBL_XY *lr_xy  /* O: Lower right x/y for the corners */
);

int set_grid_active_corners
(
    IAS_GRID_TYPE *grid     /* I/O: Grid structure */
);

const IAS_GEO_PROJ_TRANSFORMATION *get_proj_transform_degrees2proj();
const IAS_GEO_PROJ_TRANSFORMATION *get_proj_transform_radians2proj();
const IAS_GEO_PROJ_TRANSFORMATION *get_proj_transform_proj2degrees();

#endif
