/******************************************************************************
NAME:           det_opt_band

PURPOSE:
To determine the optimal band in the specified list of bands.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
  For path-oriented and minbox framing, the optimal band to use is
  retrieved from the CPF.

  This routine will search the list for the most optimal band that
  was specified.  This band information will then be used for minbox
  and path-oriented framing options.

******************************************************************************/
#include <string.h>                   /* String functions     */

#include "ias_logging.h"              /* Logging library */
#include "create_grid.h"

int determine_optimal_band 
(
    IAS_CPF *cpf,                /* I: CPF constants */
    const int *bands_available, /* I: Array of band available flags */
    int *opt_band_index         /* O: Optimal band index in the list of bands */
)
{
    int i;                      /* Loop variable */
    int found_opt_band = FALSE; /* Boolean to specify if optimal band found */
    int band_number;            /* Band number */
    int band_index;             /* Band index */
    const struct IAS_CPF_GEO_SYSTEM *geo_system;

    /* Get the focal plane parameters from the CPF */
    geo_system = ias_cpf_get_geo_system(cpf);
    if (!geo_system)
    {
        IAS_LOG_ERROR("Getting Geo System parameters from CPF");
        return ERROR;
    }

    /* Look for the current optimal band available */
    for (i = 0; i < IAS_MAX_NBANDS; i++)
    {
        /* Get the next band number in the CPF optimal band list. */
        band_number = geo_system->optimal_band_order[i];
        if ((band_number < 1) || (band_number > IAS_MAX_NBANDS))
        {
            IAS_LOG_ERROR("Invalid band number %d", band_number);
            return ERROR;
        }

        band_index = ias_sat_attr_convert_band_number_to_index(band_number);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Invalid band index for band number %d", 
                band_number);
            return ERROR;
        }

        /* See if the band index is available. */
        if (bands_available[band_index])
        {
            *opt_band_index = band_index;
            found_opt_band = TRUE;
            break;
        }
    }

    if (!found_opt_band)
    {
        IAS_LOG_ERROR("Optimal band was not found");
        *opt_band_index = -1;
        return ERROR;
    }

    return SUCCESS;
}
