/******************************************************************************
NAME:           get_dem_elevation_range

PURPOSE:
Get the minimum and maximum elevations from the DEM. 

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed
******************************************************************************/
#include <limits.h>               /* PATH_MAX definition */
#include <stdlib.h>               /* malloc definition */

#include "ias_logging.h"          /* IAS logging library */
#include "ias_l1g.h"              /* L1G I/O library */
#include "create_grid.h"

#define DEM_FILL            -9999 /* DEM data fill value */

int get_dem_elevation_range
(
    char dem_file_name[PATH_MAX], /* I: DEM file name */
    float *zmin,                  /* O: Minimum elevation value in the DEM */
    float *zmax                   /* O: Maximimum elevation value in the DEM */
)
{
    L1GIO *dem_ptr;               /* File descriptor for the DEM file */
    L1G_BAND_IO *dem_band;        /* Band for the DEM file */
    int dem_band_number = 1;      /* DEM band number */
    IAS_DATA_TYPE dem_dtype;      /* Datatype of DEM */
    int dem_number_scas;          /* Number of SCAs in DEM file */
    int dem_number_lines;         /* Number of lines in DEM file */
    int dem_number_samples;       /* Number of samples in DEM file */
    short *dem_data;              /* Data read from the DEM file */
    short dem_minimum;            /* Minimum DEM data value */
    short dem_maximum;            /* Maximum DEM data value */
    int i;                        /* Loop variable */

   /* Open the DEM image */
    dem_ptr = ias_l1g_open_image(dem_file_name, IAS_READ);
    if (dem_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening DEM data file %s", dem_file_name);
        return ERROR;
    }

    /* Open the correct band in the DEM file */
    dem_band = ias_l1g_open_band(dem_ptr, dem_band_number, &dem_dtype,
        &dem_number_scas, &dem_number_lines, &dem_number_samples);
    if (dem_band == NULL)
    {
        IAS_LOG_ERROR("Opening band %d for DEM file %s", dem_band_number,
            dem_file_name);
        ias_l1g_close_image(dem_ptr);
        return ERROR;
    }

    /* Verify data type is short */
    if (dem_dtype != IAS_I2)
    {
        IAS_LOG_ERROR("DEM file %s band %d does not contain short "
            "integer data", dem_file_name, dem_band_number);
        ias_l1g_close_band(dem_band);
        ias_l1g_close_image(dem_ptr);
        return ERROR;
    }

    /* Allocate array to hold all of DEM */
    dem_data = (short *)malloc(sizeof(short) * dem_number_lines *
        dem_number_samples);
    if (dem_data == NULL)
    {
        IAS_LOG_ERROR("Allocating pointer for data in DEM file %s "
           "band %d", dem_file_name, dem_band_number);
        ias_l1g_close_band(dem_band);
        ias_l1g_close_image(dem_ptr);
        return ERROR;
    }

    if (ias_l1g_read_image((const L1G_BAND_IO *)dem_band, 0, 0, 0,
        dem_number_lines, dem_number_samples, dem_data) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading imagery from DEM file %s band %d",
            dem_file_name, dem_band_number);
        ias_l1g_close_band(dem_band);
        ias_l1g_close_image(dem_ptr);
        free(dem_data);
        return ERROR;
    }

    /* Close L1G image band */
    if (ias_l1g_close_band(dem_band) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing DEM file %s band %d", dem_file_name,
            dem_band_number);
        free(dem_data);
        ias_l1g_close_image(dem_ptr);
        return ERROR;
    }

    /* Close image file */
    if (ias_l1g_close_image(dem_ptr) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing DEM file %s", dem_file_name);
        free(dem_data);
        return ERROR;
    }

    /* Start out finding a valid DEM input elevation */
    for (i = 0; i < dem_number_lines * dem_number_samples; i++)
    {
       /* Check to make sure we are not initializing to a fill value */
       if (dem_data[i] != DEM_FILL)
          break;
    }

    if (i == dem_number_lines * dem_number_samples)
    {
        IAS_LOG_ERROR("No valid DEM data available in file %s", dem_file_name);
        free(dem_data);
        return ERROR;
    }

    /* Initialize maximum and minimum */
    dem_minimum = dem_data[i];
    dem_maximum = dem_data[i];

    /* Find the maximum and minimum DEM data values, skipping fill */
    for (i = 0; i < dem_number_lines * dem_number_samples; i++)
    {
        if (dem_data[i] != DEM_FILL)
        {
            if (dem_data[i] > dem_maximum)
                dem_maximum = dem_data[i];
            if (dem_data[i] < dem_minimum)
                dem_minimum = dem_data[i];
        }
    }

    /* Check that a valid maximum and valid minimum were found */
    if (dem_maximum == DEM_FILL || dem_minimum == DEM_FILL)
    {
        IAS_LOG_ERROR("Only fill data available in DEM file %s",
                dem_file_name);
        free(dem_data);
        return ERROR;
    }

    /* Log results of DEM check */
    IAS_LOG_INFO("DEM range %d %d", dem_minimum, dem_maximum);
    free(dem_data);

    /* Set values to be returned */
    *zmin = dem_minimum;
    *zmax = dem_maximum;

    return SUCCESS;
}
