/******************************************************************************
NAME:           get_frame_maxbox

PURPOSE:
Determines the frame in output space for the maxbox north-up product.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed
******************************************************************************/
#include <string.h>                   /* String functions */

#include "ias_geo.h"                  /* Geometric library */
#include "ias_logging.h"              /* Logging library */
#include "create_grid.h"

int get_frame_maxbox
(
    const IAS_LOS_MODEL *model, /* I: Model structure */
    const int *num_lines,     /* I: Number of lines in model for each band */
    const int *num_samps,     /* I: Number of samps in model for each band */
    IAS_GRID_TYPE *grid       /* I/O: Grid structure */
)
{
    int nl;                   /* Number of lines in the scene */
    int ns;                   /* Number of samples in the scene */
    int stat;                 /* Function return status */
    int band_index;           /* Band index */
    int band_number;          /* Band number corresponding to band_index */
    double pxmax;             /* Maximum X projection coordinate */
    double pxmin;             /* Minimum X projection coordinate */
    double pymax;             /* Maximum Y projection coordinate */
    double pymin;             /* Minimum Y projection coordinate */
    double elev = 0.0;        /* Elevation at the current point */
    struct IAS_DBL_LS ul_ls;  /* Upper left line/sample for the corners */
    struct IAS_DBL_LS ur_ls;  /* Upper right line/sample for the corners */
    struct IAS_DBL_LS ll_ls;  /* Lower left line/sample for the corners */
    struct IAS_DBL_LS lr_ls;  /* Lower right line/sample for the corners */
    struct IAS_DBL_XY ul_xy;  /* Upper left x/y for the corners */
    struct IAS_DBL_XY ur_xy;  /* Upper right x/y for the corners */
    struct IAS_DBL_XY ll_xy;  /* Lower left x/y for the corners */
    struct IAS_DBL_XY lr_xy;  /* Lower right x/y for the corners */
    double pixsize = 0;       /* Pixel size to be used for padding */
    int sca_index;            /* SCA index */
    int zplane;               /* Counter for elevation planes */
    int first_available_band; /* Is this the first available band? */ 

    first_available_band = TRUE;
    for (band_index = 0; band_index < grid->nbands; band_index++)
    {
        /* skip bands not available */
        if (!grid->bands_available[band_index])
            continue;

        /* Get the band number. */
        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Invalid band number %d", band_number);
            return ERROR;
        }

        for (sca_index = 0; sca_index < grid->scas_per_band[band_index];
             sca_index++)
        {
            /* Choose biggest pixsize to pad the corners */
            if (first_available_band)
                pixsize = grid->gridbands[band_index][sca_index].pixsize;
            else if (pixsize < grid->gridbands[band_index][sca_index].pixsize)
                pixsize = grid->gridbands[band_index][sca_index].pixsize;

            for (zplane = 0; 
                 zplane < grid->gridbands[band_index][sca_index].nzplanes; 
                 zplane++)
            {
                elev = (zplane - 
                    grid->gridbands[band_index][sca_index].zeroplane) * 
                    grid->gridbands[band_index][sca_index].zspacing;

                /* Find the offsets for the first and last lines in the 
                   scene.  These will determine the corner points in 
                   line/sample.  */
                nl = num_lines[band_index];
                ns = num_samps[band_index];

                /* Find the corner points. */
                ul_ls.samp = 0.0;
                ul_ls.line = 0.0;

                ur_ls.samp = ns - 1; 
                ur_ls.line = 0;

                lr_ls.samp = ns - 1;
                lr_ls.line = nl - 1;

                ll_ls.samp = 0;
                ll_ls.line = nl - 1;

                /* For the 4 corner points, find the location in output
                   space */
                stat = calc_corner_point_output_space_locations(&ul_ls, &ur_ls,
                    &ll_ls, &lr_ls, band_index, sca_index, elev, model, grid,
                    &ul_xy, &ur_xy, &ll_xy, &lr_xy);
                if (stat != SUCCESS)
                {
                    IAS_LOG_ERROR("Calculating output space locations for "
                        "the corner points for band %d SCA %d elevation %f",
                        band_number, sca_index + 1, elev);
                    return ERROR;
                }

                /* Find the minimum and maximum x and y corners. */
                if (sca_index == 0 && zplane == 0 && first_available_band)
                    pxmin = ul_xy.x;
                else if (ul_xy.x < pxmin)
                    pxmin = ul_xy.x;
                if (ll_xy.x < pxmin)
                    pxmin = ll_xy.x;
                if (ur_xy.x < pxmin)
                    pxmin = ur_xy.x;
                if (lr_xy.x < pxmin)
                    pxmin = lr_xy.x;

                if (sca_index == 0 && zplane == 0 && first_available_band)
                    pxmax = ul_xy.x;
                else if (ul_xy.x > pxmax)
                    pxmax = ul_xy.x;
                if (ll_xy.x > pxmax)
                    pxmax = ll_xy.x;
                if (ur_xy.x > pxmax)
                    pxmax = ur_xy.x;
                if (lr_xy.x > pxmax)
                    pxmax = lr_xy.x;

                if (sca_index == 0 && zplane == 0 && first_available_band)
                    pymin = ul_xy.y;
                else if (ul_xy.y < pymin)
                    pymin = ul_xy.y;
                if (ll_xy.y < pymin) 
                    pymin = ll_xy.y;
                if (ur_xy.y < pymin)
                    pymin = ur_xy.y;
                if (lr_xy.y < pymin)
                    pymin = lr_xy.y;

                if (sca_index == 0 && zplane == 0 && first_available_band)
                    pymax = ul_xy.y;
                else if (ul_xy.y > pymax)
                    pymax = ul_xy.y;
                if (ll_xy.y > pymax)
                    pymax = ll_xy.y;
                if (ur_xy.y > pymax)
                    pymax = ur_xy.y;
                if (lr_xy.y > pymax)
                    pymax = lr_xy.y;
            }
        }

        first_available_band = FALSE;
    }

    /* Pad corners to PADVAL * pixel size.  */
    stat = pad_corners(pixsize, &pxmax, &pxmin, &pymax, &pymin);
    if (stat != SUCCESS)
    {
        IAS_LOG_ERROR("In the pad_corners call");
        return ERROR;
    }

    /* Find the number of lines and samples in output space.
     * This must be implemented for each band.
     * NOTE: This equation adds 1 to the (x-y)/pixsize, but to round we need
     * to add another .5 since the (int) truncates to an integer value.    */
    for (band_index = 0; band_index < grid->nbands; band_index++)
    {
        /* skip bands not available */
        if (!grid->bands_available[band_index])
            continue;

        for (sca_index = 0; sca_index < grid->scas_per_band[band_index];
             sca_index++)
        {
            grid->gridbands[band_index][sca_index].lines = 
                (int) (((pymax - pymin) /
                grid->gridbands[band_index][sca_index].pixsize) + 1.5);
            grid->gridbands[band_index][sca_index].samps = 
                (int) (((pxmax - pxmin) /
                grid->gridbands[band_index][sca_index].pixsize) + 1.5);
        }
    }

    /* Pack the corner projection coordinates in the frame structure */
    grid->corners.upleft.x = grid->corners.loleft.x = pxmin;
    grid->corners.upright.x = grid->corners.loright.x = pxmax;
    grid->corners.loleft.y = grid->corners.loright.y = pymin;
    grid->corners.upleft.y = grid->corners.upright.y = pymax;

    /* Convert the corners to lat/long, and set the active area corners */
    stat = set_grid_active_corners(grid);
    if (stat != SUCCESS)
    {
        IAS_LOG_ERROR("Converting corners to lat/long and setting active "
            "area corners");
        return ERROR;
    }

    return SUCCESS;
}
