/******************************************************************************
NAME:           get_frame_minbox

PURPOSE:
Determines the frame in output space for the minbox north-up product.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
This routine will determine the frame in output space for the MINBOX north-up
product.  The actual frame is determined based on the optimal band's pixel
size, but the frame is the same for every band.

Support for the minbox option is not a requirement.  This file is being left
only for convenience for CalVal.

ALGORITHM REFERENCES:
See ias_grid.h for a description of the grid structure.
See ias_los_model.h for a description of the model structure.

******************************************************************************/
#include <string.h>                   /* String functions */

#include "ias_geo.h"                  /* Geographic library */
#include "ias_logging.h"              /* Logging library */
#include "create_grid.h"

int get_frame_minbox
(
    const IAS_LOS_MODEL *model, /* I: Model structure */
    int opt_band_index,       /* I: "Optimal" band to work on */
    IAS_GRID_TYPE *grid       /* I/O: Grid structure */
)
{
    int sca_index;            /* SCA index */
    int stat;                 /* Function return status */
    int band_index;           /* Band index */ 
    int first_band;           /* First available band */
    int band_number;          /* Band number for band index */ 
    double pixsize = 0;       /* Pixel size to be used for padding */
    double pxmax;             /* Maximum X projection coordinate */
    double pxmin;             /* Minimum X projection coordinate */
    double pymax;             /* Maximum Y projection coordinate */
    double pymin;             /* Minimum Y projection coordinate */
    struct IAS_DBL_XY ul_xy;  /* Upper left x/y for the corners */
    struct IAS_DBL_XY ur_xy;  /* Upper right x/y for the corners */
    struct IAS_DBL_XY ll_xy;  /* Lower left x/y for the corners */
    struct IAS_DBL_XY lr_xy;  /* Lower right x/y for the corners */
    struct IAS_DBL_LAT_LONG ul_ltlg; /* Upper left lat/long for the corners */
    struct IAS_DBL_LAT_LONG ur_ltlg; /* Upper right lat/long for the corners */
    struct IAS_DBL_LAT_LONG ll_ltlg; /* Lower left lat/long for the corners */
    struct IAS_DBL_LAT_LONG lr_ltlg; /* Lower right lat/long for the corners */
    const IAS_GEO_PROJ_TRANSFORMATION *degrees2proj; /* Projection 
                                 transformation parameters */

    /* Find the minimum bounding rectangle for active areas for each band. */
    first_band = TRUE;
    for (band_index = 0; band_index < grid->nbands; band_index++)
    {
        /* Skip bands not available */
        if (!grid->bands_available[band_index])
            continue;

        /* Choose biggest pixsize to pad the corners */
        if (first_band)
            pixsize = grid->gridbands[band_index][0].pixsize;
        else if (pixsize < grid->gridbands[band_index][0].pixsize)
            pixsize = grid->gridbands[band_index][0].pixsize;

        /* Get the band number. */
        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Invalid band number for band index %d", band_index);
            return ERROR;
        }

        ul_ltlg.lat = grid->active_corners[band_index].upleft.y;
        ul_ltlg.lng = grid->active_corners[band_index].upleft.x;
        ur_ltlg.lat = grid->active_corners[band_index].upright.y;
        ur_ltlg.lng = grid->active_corners[band_index].upright.x;
        ll_ltlg.lat = grid->active_corners[band_index].loleft.y;
        ll_ltlg.lng = grid->active_corners[band_index].loleft.x;
        lr_ltlg.lat = grid->active_corners[band_index].loright.y;
        lr_ltlg.lng = grid->active_corners[band_index].loright.x;

        /* Initialize projection transformation. */
        degrees2proj = get_proj_transform_degrees2proj();

        /* Calculate the location of the latitude and longitude in output space.
           Units stored in the grid active area corners are in degrees */
        stat = ias_geo_transform_coordinate(degrees2proj, ul_ltlg.lng, 
            ul_ltlg.lat, &ul_xy.x, &ul_xy.y);
        if (stat != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating location in output space for lat/long "
                "for band %d upper left corner", band_number);
            return ERROR;
        }

        stat = ias_geo_transform_coordinate(degrees2proj, ur_ltlg.lng, 
            ur_ltlg.lat, &ur_xy.x, &ur_xy.y);
        if (stat != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating location in output space for lat/long "
                "for band %d upper right corner", band_number);
            return ERROR;
        }

        stat = ias_geo_transform_coordinate(degrees2proj, ll_ltlg.lng, 
            ll_ltlg.lat, &ll_xy.x, &ll_xy.y);
        if (stat != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating location in output space for lat/long "
                "for band %d lower left corner", band_number);
            return ERROR;
        }

        stat = ias_geo_transform_coordinate(degrees2proj, lr_ltlg.lng, 
            lr_ltlg.lat, &lr_xy.x, &lr_xy.y);
        if (stat != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating location in output space for lat/long "
                "for band %d lower right corner", band_number);
            return ERROR;
        }

        /* Find the minimum and maximum x and y corners. */
        if (first_band)
            pxmin = ul_xy.x;
        else if (ul_xy.x < pxmin)
            pxmin = ul_xy.x;
        if (ll_xy.x < pxmin)
            pxmin = ll_xy.x;
        if (ur_xy.x < pxmin)
            pxmin = ur_xy.x;
        if (lr_xy.x < pxmin)
            pxmin = lr_xy.x;

        if (first_band)
            pxmax = ul_xy.x;
        else if (ul_xy.x > pxmax)
            pxmax = ul_xy.x;
        if (ll_xy.x > pxmax)
            pxmax = ll_xy.x;
        if (ur_xy.x > pxmax)
            pxmax = ur_xy.x;
        if (lr_xy.x > pxmax)
            pxmax = lr_xy.x;

        if (first_band)
            pymin = ul_xy.y;
        else if (ul_xy.y < pymin)
            pymin = ul_xy.y;
        if (ll_xy.y < pymin)
            pymin = ll_xy.y;
        if (ur_xy.y < pymin)
            pymin = ur_xy.y;
        if (lr_xy.y < pymin)
            pymin = lr_xy.y;

        if (first_band)
            pymax = ul_xy.y;
        else if (ul_xy.y > pymax)
            pymax = ul_xy.y;
        if (ll_xy.y > pymax)
            pymax = ll_xy.y;
        if (ur_xy.y > pymax)
            pymax = ur_xy.y;
        if (lr_xy.y > pymax)
            pymax = lr_xy.y;

        first_band = FALSE;
    }

    /* Pad corners to PADVAL * pixel size. */
    stat = pad_corners(pixsize, &pxmax, &pxmin, &pymax, &pymin);
    if (stat != SUCCESS)
    {
        IAS_LOG_ERROR("In the pad_corners call");
        return ERROR;
    }
    IAS_LOG_INFO("%f %f %f %f", pxmin, pxmax, pymin, pymax);

    /* Find the number of lines and samples in output space.
     * This must be implemented for each band.
     * NOTE: This equation adds 1 to the (x-y)/pixsize, but to round we need
     * to add another .5 since the (int) truncates to an integer value.  */
    for (band_index = 0; band_index < grid->nbands; band_index++)
    {
        for (sca_index = 0; sca_index < grid->scas_per_band[band_index];
             sca_index++)
        {
            grid->gridbands[band_index][sca_index].lines = 
                (int) (((pymax - pymin) /
                grid->gridbands[band_index][sca_index].pixsize) + 1.5);
            grid->gridbands[band_index][sca_index].samps = 
                (int) (((pxmax - pxmin) /
                grid->gridbands[band_index][sca_index].pixsize) + 1.5);
        }
    }

    /* Pack the corner projection coordinates in the frame structure */
    grid->corners.upleft.x = grid->corners.loleft.x = pxmin;
    grid->corners.upright.x = grid->corners.loright.x = pxmax;
    grid->corners.loleft.y = grid->corners.loright.y = pymin;
    grid->corners.upleft.y = grid->corners.upright.y = pymax;

    return SUCCESS;
}
