/******************************************************************************
NAME:           get_frame_calc_north_up

PURPOSE:
Determines the frame in output space for the north-up product.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
This routine will determine the frame in output space for the GEOBOX, PROJBOX,
and UL_SIZE north-up products.  The actual frame is determined based on the
optimal band's pixel size, but the frame is the same for every band.

ALGORITHM REFERENCES:
See ias_grid.h for a description of the grid structure.

******************************************************************************/
#include <string.h>               /* String functions */

#include "ias_geo.h"              /* Geographic library */
#include "ias_logging.h"          /* Logging library */
#include "create_grid.h"

int get_frame_north_up
(
    IAS_GRID_TYPE *grid,          /* I/O: Grid structure */
    int opt_band_index,           /* I: "Optimal" band index to work on */
    struct IAS_DBL_XY frame_coords[2],/* I: User-input frame coordinates to 
                                        define the frame of the output area 
                                        (upper left followed by lower right 
                                        corner) */
    struct IAS_DBL_LS lscoords    /* I: Image coordinates of the first coord
                                        pair entered in frame_coords when proj
                                        params are being specified directly */
)
{
    double pxmax;        /* Maximum X projection coordinate */
    double pxmin;        /* Minimum X projection coordinate */
    double pymax;        /* Maximum Y projection coordinate */
    double pymin;        /* Minimum Y projection coordinate */
    int sca_index;       /* SCA index */
    int out_units;       /* Code for the output projection units */
    int status;          /* Function return status */
    int ref_band_index;  /* Reference band index */
    int band_index;      /* Band index */
    const IAS_GEO_PROJ_TRANSFORMATION *geo2proj;/* projection transformation */

    geo2proj = get_proj_transform_degrees2proj();

    /* Determine the reference band number.  Implement this on the optimal
     * band.  This will be the band for which the frame is calculated, but the
     * frame will be the same for all bands.  */
    ref_band_index = opt_band_index;

    /* Determine the frame type */
    if (grid->frame_type == IAS_GEOBOX)
    { /* GEOBOX */
        /* Convert the grid projection units to a unit code for GCTP.  */
        status = ias_geo_get_units(grid->units, &out_units);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Converting projection units %s to unit code",
                grid->units);
            return ERROR;
        }

        /* Determine output projection space minimums and maximums.  The input
         * and output spheroid code will be the same.  */
        status = ias_geo_compute_map_edge(geo2proj, frame_coords[0].x, 
                frame_coords[1].x, frame_coords[0].y, frame_coords[1].y, 
                &pxmin, &pxmax, &pymin, &pymax);
        if (status != SUCCESS) 
        {
            IAS_LOG_ERROR("In the ias_geo_compute_map_edge call in north_up");
            return ERROR;
        }

        /* Pad corners to PADVAL * pixel size.  */
        status = pad_corners(grid->gridbands[ref_band_index][0].pixsize, 
            &pxmax, &pxmin, &pymax, &pymin);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("In the pad_corners call");
            return ERROR;
        }

        /* Find the number of lines and samples in output space.
         * This must be implemented for each band.
         * NOTE: This equation adds 1 to (x-y)/pixsize, but to round we need
         * to add another .5 since the (int) truncates to an integer value. */
        for (band_index = 0; band_index < grid->nbands; band_index++)
        {
            /* skip bands not available */
            if (!grid->bands_available[band_index])
                continue;

            for (sca_index = 0; sca_index < grid->scas_per_band[band_index];
                 sca_index++)
            {
                grid->gridbands[band_index][sca_index].lines = 
                    (int) (((pymax - pymin) /
                    grid->gridbands[band_index][sca_index].pixsize) + 1.5);
                grid->gridbands[band_index][sca_index].samps = 
                    (int) (((pxmax - pxmin) /
                    grid->gridbands[band_index][sca_index].pixsize) + 1.5);
            }
        }
    }
    else if (grid->frame_type == IAS_PROJBOX)
    { /* PROJBOX */
        /* If output frame was given in projection coordinates, for two 
           points */
        /* I'm going to assume there is at least one SCA for reference 
           band and the pixel size is the same for every SCA with a band.  */
        pymax = frame_coords[0].y 
            + (lscoords.line * grid->gridbands[ref_band_index][0].pixsize);
        pxmin = frame_coords[0].x 
            - (lscoords.samp * grid->gridbands[ref_band_index][0].pixsize);
        pymin = frame_coords[1].y;
        pxmax = frame_coords[1].x;

        /* Find the number of lines and samples in output space.
         * This must be implemented for each band.
         * NOTE: This equation adds 1 to (x-y)/pixsize, but to round we need
         * to add another .5 since the (int) truncates to an integer value.  */
        for (band_index = 0; band_index < grid->nbands; band_index++)
        {
            /* skip bands not available */
            if (!grid->bands_available[band_index])
                continue;

            for(sca_index = 0; sca_index < grid->scas_per_band[band_index];
                sca_index++)
            {
                grid->gridbands[band_index][sca_index].lines = 
                    (int) (((pymax - pymin) /
                    grid->gridbands[band_index][sca_index].pixsize) + 1.5);
                grid->gridbands[band_index][sca_index].samps = 
                    (int) (((pxmax - pxmin) /
                    grid->gridbands[band_index][sca_index].pixsize) + 1.5);
            }
        }
    }
    else if (grid->frame_type == IAS_UL_SIZE)
    { /* UL_SIZE */
        /* If output frame was given in projection coordinates, 1 pt & 
            nlines/samps */
        /* I'm going to assume there is at least one SCA for reference 
           band and the pixel size is the same for every SCA with a band.  */
        pymax = frame_coords[0].y 
            + (lscoords.line * grid->gridbands[ref_band_index][0].pixsize);
        pxmin = frame_coords[0].x 
            - (lscoords.samp * grid->gridbands[ref_band_index][0].pixsize);
        pymin = pymax - ((grid->gridbands[ref_band_index][0].lines - 1) 
            * grid->gridbands[ref_band_index][0].pixsize);
        pxmax = pxmin + ((grid->gridbands[ref_band_index][0].samps - 1) 
            * grid->gridbands[ref_band_index][0].pixsize);
    }
    else
    {
        IAS_LOG_ERROR("Unsupported frame type: %d", grid->frame_type);
        return ERROR;
    }

    /* Pack the corner projection coordinates in the frame structure */
    grid->corners.upleft.x = grid->corners.loleft.x = pxmin;
    grid->corners.upright.x = grid->corners.loright.x = pxmax;
    grid->corners.upleft.y = grid->corners.upright.y = pymax;
    grid->corners.loleft.y = grid->corners.loright.y = pymin;

    return SUCCESS;
}
