/******************************************************************************
NAME: get_frame_path_maxmin_box

PURPOSE: This routine will provide a path-oriented product whose frame is
either the optimal band(MINBOX) or large enough to contain all bands(MAXBOX).

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

ALGORITHM REFERENCES:
This algorithm was created by combining the logic in the in both
path_oriented.c and minbox.c.  Almost all the code is from these
two modules.

******************************************************************************/
#include <string.h>                   /* String functions     */

#include "ias_logging.h"              /* Logging library      */
#include "ias_geo.h"                  /* Geographic library   */
#include "create_grid.h"

int get_frame_path_maxmin_box
(
    const IAS_LOS_MODEL *model, /* I: Model structure */
    int opt_band_index,     /* I: Optimal band index, used with MINBOX */
    const int *num_lines,   /* I: Number of lines in model for optimal band */ 
    const int *num_samps,   /* I: Number of samps in model for optimal band */
    IAS_CPF *cpf,            /* I: CPF to determine WRS lat/long and angle */
    IAS_GRID_TYPE *grid,    /* I/O: Grid structure to use/modify */
    int path,               /* I: WRS path of product to provide */
    double row,             /* I: WRS row of product (may be fractional) */
    double poly_x[IAS_MAX_NBANDS][4], /* O: Per-band array of x polynomial
                                            coefficients */
    double poly_y[IAS_MAX_NBANDS][4]  /* O: Per-band array of y polynomial
                                            coefficients */
)
{
    int stat;               /* Return status */

    /* Fill in the grid structure with the framing information for the
       path-oriented product described by the path/row parameters and
       the projection information in the grid structure.                 */
    stat = calc_center_and_rotation_angle(path, row, cpf, grid);
    if (stat != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating the center point and rotation angle");
        return ERROR;
    }

    if (grid->frame_type == IAS_PATH_MINBOX)
    {
        stat = calc_path_oriented_minbox_frame(model, opt_band_index, grid, 
            poly_x, poly_y);
        if (stat != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating the path-oriented frame");
            return ERROR;
        }
    }
    else
    {
        stat = calc_path_oriented_maxbox_frame(model, num_lines, num_samps,
                   grid, poly_x, poly_y);
        if (stat != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating the path-oriented frame");
            return ERROR;
        }
    }

    stat = angle_to_map(grid);
    if (stat != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating the angle for mapping");
        return ERROR;
    }

    return SUCCESS;
}
