/******************************************************************************
NAME:           get_frame_path_oriented

PURPOSE:
This routine will provide a path-oriented product that is framed to a nominal
WRS scene.  The user specifies only the projection, pixel size, and the path
and row of the scene.  

RETURN VALUE:
Type = int (SUCCESS or ERROR)

Support for the path oriented option is not a requirement.  This file is
being left only for convenience for CalVal.

ALGORITHM REFERENCES:
Landsat 7 Image Assessment System Geometric Processing Subsystem Detailed
Design Specification, algorithms on pages 57-62.
******************************************************************************/
#include <string.h>                   /* String functions */

#include "ias_logging.h"              /* Logging library */
#include "ias_geo.h"                  /* Geographic library */
#include "create_grid.h"

int get_frame_path_oriented
(
    IAS_GRID_TYPE *grid, /* I/O: Grid structure to use/modify */
    IAS_CPF *cpf,         /* I: CPF to determine WRS lat/long and angle */
    int path,            /* I: WRS path of product to provide */
    double row,          /* I: WRS row of product (may be fractional) */
    double poly_x[IAS_MAX_NBANDS][4],/* O: Per-band X polynomial coefficients */
    double poly_y[IAS_MAX_NBANDS][4] /* O: Per-band Y polynomial coefficients */
)
{
    int units;           /* Integral representation of units */
    int stat;            /* Return status */

    /* Convert the units string in the grid structure to a
       numerical representation for GCTP.                    */
    stat = ias_geo_get_units(grid->units, &units);
    if (stat != SUCCESS)
    {
        IAS_LOG_ERROR("Converting projection units %s to unit code",
            grid->units);
        return ERROR;
    }

    /* Fill in the grid structure with the framing information for the
       path-oriented product described by the path/row parameters and
       the projection information in the grid structure.                 */
    stat = calc_center_and_rotation_angle(path, row, cpf, grid);
    if (stat != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating the center point and rotation angle");
        return ERROR;
    }

    stat = calc_path_oriented_frame(grid, units, poly_x, poly_y);
    if (stat != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating the path-oriented frame");
        return ERROR;
    }

    stat = angle_to_map(grid);
    if (stat != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating the angle for mapping");
        return ERROR;
    }

    return SUCCESS;
}
