/******************************************************************************
NAME:           get_parms_and_initialize_grid

PURPOSE:
Get the grid parameters from the input parameter file and OMF file.
Initialize some of the basic grid values.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
This routine will open the input parameter file, read the grid parameters,
close the parameter file, and return the parameters and the grid.
Pixel sizes < 20.0 refer to 15m band resolution.  Pixel sizes >= 20.0 and
<= 50.0 refer to 30m band resolution.  Anything larger is considered a 100m
band.

ALGORITHM REFERENCES:
See ias_grid.h for a description of the grid structure.
******************************************************************************/
#include <stdlib.h>
#include <string.h>                   /* String functions */
#include <math.h>                     /* For fmod */

#include "gctp.h"                     /* Projection parameter definitions */
#include "ias_const.h"                 /* IAS constants; contains defined
                                          INVALID_ELEVATION value */
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "ias_geo.h"                   /* Geometric library */
#include "ias_miscellaneous.h"         /* IAS miscellaneous library */
#include "ias_math.h"                  /* Math library */
#include "create_grid.h"

#define WRS_SCENE_SIZE_LINES      5965 /* Number of lines in WRS scene */
#define WRS_SCENE_SIZE_SAMPS      6967 /* Number of samps in WRS scene */

#define NUM_PIXEL_SIZES              3 /* Reflective, pan, and thermal */
#define NUM_LS_COORDINATES           2 /* Line and sample */
#define INVALID_COORDINATE 999999999.0 /* Indicates coordinate not chosen */


int calc_size
(
    const float pixsize[NUM_PIXEL_SIZES],     /* I: pixel sizes */
    const int default_dim[NUM_PIXEL_SIZES],   /* I: default dimension to use */
    int frame_type,                           /* I: frame type */
    int dim[NUM_PIXEL_SIZES]                  /* I/O: dimension to update */
);

int get_parms_and_initialize_grid
(
    const char *param_filename,     /* I: Parameter file name */
    char cal_parm_file[PATH_MAX],   /* O: Calibration parameter filename */
    char model_file_name[PATH_MAX], /* O: Model file name */
    char omf_file_name[PATH_MAX],   /* O: OMF file name */
    char dem_file_name[PATH_MAX],   /* O: Optional DEM filename */
    struct IAS_DBL_XY frame_coords[NUM_FRAME_COORDINATES/2], /* O: User-input 
                                        frame coordinates to define the frame 
                                        of the output area */
    struct IAS_DBL_LS *lscoords,    /* O: Image coordinates of the first coord
                                        pair entered in frame_coords when proj
                                        params are being specified directly */
    int *path,                      /* O: WRS path number of the image */
    double *row,                    /* O: WRS (fractional) row number */
    float *zmax,                    /* O: Elevation plane maximum */
    float *zmin,                    /* O: Elevation plane minimum */
    int *processing_pass,           /* O: Pass through WO processing flow */
    int band_numbers[IAS_MAX_NBANDS], /* O: Unique user-selected bands */
    int *number_bands,              /* O: # of unique user-selected bands */
    int *max_processors_to_use,     /* O: Maximum number of processors to use */
    int *no_elevation_required_flag,/* O: Flag that indicates no elevation is
                                          required.  Mainly an indication that
                                          the grid is being created for stray
                                          light correction and some checks can
                                          be relaxed. */
    IAS_GRID_TYPE *grid,            /* O: Grid structure to be filled */
    IAS_LOS_MODEL **model           /* O: Model structure */
)
{
    char projection_units[IAS_UNITS_SIZE]; /* Projection units */
    double frm_coords[NUM_FRAME_COORDINATES]; /* Y/X coords of the frame 
                                     points (2 sets) */
    float ls_coords[NUM_LS_COORDINATES]; /* Line/sample coords of the point */
    float pixsize[NUM_PIXEL_SIZES]; /* Pixel size (30m, 15m, 30m) */
    double projection_parameters[IAS_PROJ_PARAM_SIZE]; /* Projection 
                                     parameters */
    int i;                        /* Loop variable */
    int coordinate;               /* Loop variable for coordinates */
    int sca_index;                /* SCA index */
    int nl[NUM_PIXEL_SIZES];      /* Num of lines of the band resolutions 
                                    (30m, 15m) */
    int ns[NUM_PIXEL_SIZES];      /* Num of samps of the band resolutions
                                    (30m, 15m) */
    int row_int;                  /* WRS row - integer part */
    double row_frac;              /* WRS row - fractional part */
    int satellite_number;         /* Satellite number being processed */
    int satellite_number_bands;   /* Number of normal bands for the satellite */
    int band_number;              /* Band number */
    int band_index;               /* Band index */
    IAS_SPECTRAL_TYPE spectral_type[IAS_MAX_NBANDS]; /* Bands' spectral types */
    int disable_tirs_bands = 0;   /* Flag to disable TIRS bands */

    const int lines_in_scene[NUM_PIXEL_SIZES] = 
    {
        WRS_SCENE_SIZE_LINES,               /* 30 meters */
        2 * WRS_SCENE_SIZE_LINES - 1,       /* 15 meters */
        3 * WRS_SCENE_SIZE_LINES / 10 + 1,  /* 100 meters */
    };
    const int samps_in_scene[NUM_PIXEL_SIZES] = 
    {
        WRS_SCENE_SIZE_SAMPS,               /* 30 meters */
        2 * WRS_SCENE_SIZE_SAMPS - 1,       /* 15 meters */
        3 * WRS_SCENE_SIZE_SAMPS / 10 + 1,  /* 100 meters */
    };
    const char *valid_units[] = {"RADIANS", "FEET", "METERS", "SECONDS", 
        "DEGREES"};               /* Valid PROJ_UNITS or COORD_UNITS values */
    int status;                   /* Status returned from functions */
    double remainder_check;       /* Remainder for checking tolerance */
    double division_tolerance;    /* Floating point division tolerance */

    /* Application parameter file values */
    char local_wo_id[IAS_WORKORDER_ID_SIZE]; /* Work order ID */
    int local_band_numbers[IAS_MAX_NBANDS]; /* Array of band numbers */ 
    double local_pixel_sizes[NUM_PIXEL_SIZES]; /* Pixel sizes */
    char local_projection_units[IAS_UNITS_SIZE]; /* Projection units */
    int local_frame_type;                  /* Frame type */
    double local_frame_coordinates[NUM_FRAME_COORDINATES] = 
        {INVALID_COORDINATE, INVALID_COORDINATE, 
        INVALID_COORDINATE, INVALID_COORDINATE}; /* Frame coordinates */
    double local_ls_coordinates[NUM_LS_COORDINATES]; /* Line and sample 
                                              coordinates */
    int local_nl[NUM_PIXEL_SIZES];         /* Number of lines */
    int local_ns[NUM_PIXEL_SIZES];         /* Number of samples */
    int local_wrs_path;                    /* WRS path */
    int local_wrs_row_int;                 /* WRS row - integer part */
    double local_projection_parameters[IAS_PROJ_PARAM_SIZE]; /* Projection 
                                              parameters */
    double omf_projection_parameters[IAS_PROJ_PARAM_SIZE] = {0}; /* Projection 
                                              parameters */
    int local_processing_pass;             /* Processing flow pass */ 
    char local_omf_file_name[PATH_MAX];    /* OMF name */
    int local_projection_code_override = 0;/* Target projection override */ 
    int override_projection_with_pass = 0; /* Target projection override
                                              for pass */
    char disable_tirs_starting_date[IAS_L0R_DATE_LENGTH] = "";
                                           /* Disable TIRS starting date */

    /* OMF values */
    char local_cpf_file_name[PATH_MAX];    /* CPF file name */
    int local_utm_zone;                    /* UTM zone */
    int local_projection_code;             /* Projection code */
    int local_max_elevation = 0;           /* Elevation plane maximum as read
                                              from DEM */
    int local_min_elevation = 0;           /* Elevation plane minimum as read
                                              from DEM */
    char local_grid_file_name[GRIDNAME_SIZE]; /* Grid file name */
    char local_model_file_name[PATH_MAX];  /* Model file name */
    char local_dem_file_name[PATH_MAX] = ""; /* DEM file name */
    int satellite_id;                      /* current satellite id */
    char acquisition_date[IAS_L0R_DATE_LENGTH] = ""; /* Acquisition date */
    double default_projection_parameters[] = {6378137.0, 6356752.314245, 0.0, 
        0.0, 0.0, -71000000.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};

    /* Initialize the zmax and zmin */
    *zmax = INVALID_ELEVATION;
    *zmin = INVALID_ELEVATION;

    /* Initialize band lists. */
    for (i = 0; i < IAS_MAX_NBANDS; i++)
    {
        local_band_numbers[i] = 0;
        band_numbers[i] = 0;
    }

    /* start parameter table definition */
    IAS_PARM_DECLARE_TABLE(parms, 17);

    /* Get the work order ID */
    IAS_PARM_ADD_STRING(parms, WORK_ORDER_ID,
        "Identification string for Work Order", IAS_PARM_REQUIRED,
        0, NULL, 0, NULL, local_wo_id, sizeof(local_wo_id), 1);

    /* Get the list of bands. */
    IAS_PARM_BAND_LIST(parms, IAS_PARM_OPTIONAL, local_band_numbers, 
        sizeof(local_band_numbers), 1);

    /* Get the pixel sizes.  This is read for all frame_types.  The minimum
       value for the range is zero to support processing lunar data where the
       pixel size is entered in radians, degrees, or seconds which are very
       close to zero. */
    double default_pixel_size[NUM_PIXEL_SIZES] = {30.0, 15.0, 30.0};
    IAS_PARM_ADD_DOUBLE(parms, PIXEL_SIZE, 
        "Output image pixel size for (reflective, pan, thermal) in units "
        "specified by PROJ_UNITS", 
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY, 1, 0.0, 120.0, 3,
        default_pixel_size, local_pixel_sizes, sizeof(local_pixel_sizes),
        NUM_PIXEL_SIZES);

    /* Get the projection units. */
    const char *default_proj_units[] = {"METERS"};
    IAS_PARM_ADD_STRING(parms, PROJ_UNITS, 
        "Projection distance (pixel size) units", IAS_PARM_OPTIONAL, 5,  
        valid_units, 1, default_proj_units, local_projection_units, 
        sizeof(local_projection_units), 1);

    /* Get the frame type.  This is only needed for Earth scenes; 
       otherwise it's derived from the model. */ 
    int default_frame_type[] = {IAS_MAXBOX};
    IAS_PARM_ADD_INT(parms, FRAME_TYPE,
        "Output image frame computation method: 1=Geobox, 2=Projbox, "
        "3=UL size, 4=Minbox, 5=Maxbox, 6=Path oriented, "
        "7=Path Minbox, 8=Path Maxbox, 9=Lunar Minbox, 10=Lunar Maxbox, "
        "11=Stellar frame; the entry is needed in the parameter file only "
        "for Earth scenes", IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 1, 11,
        1, default_frame_type, &local_frame_type, sizeof(local_frame_type), 1);

    /* Get the frame coordinates.  They are entered in the parameter file as
       Y then X for each point, up to two points (and possibly no points).
       Read 2 sets for IAS_GEOBOX and IAS_PROJBOX.  Read only 1 set for
       IAS_UL_SIZE. */
    IAS_PARM_ADD_DOUBLE(parms, FRAME_COORDS, 
        "Frame corner or center coordinate as (north/lat, east/long); 2 "
        "Y and X pairs are needed for IAS_GEOBOX and IAS_PROJBOX; 1 pair "
        "is needed for IAS_UL_SIZE", IAS_PARM_OPTIONAL, IAS_PARM_ARRAY, 1, 
        -180000000.0, 180000000.0, 0, NULL, 
        local_frame_coordinates, sizeof(local_frame_coordinates), 0);

    /* Get the line and sample coordinates.  This is only needed if the frame 
       type is IAS_PROJBOX or IAS_UL_SIZE.  The lscoords are entered in the 
       parameter file as line then sample. */
    double default_ls_coordinates[] = {0.0, 0.0};
    IAS_PARM_ADD_DOUBLE(parms, LSCOORDS, 
        "Line and sample coordinates relative to the optimal band "
        "corresponding to FRAME_COORDS; required for IAS_PROJBOX and " 
        "IAS_UL_SIZE framing", IAS_PARM_OPTIONAL, IAS_PARM_ARRAY, 1,
        -1000000.0, 1000000.0, 2, default_ls_coordinates, local_ls_coordinates, 
        sizeof(local_ls_coordinates), 2);

    /* Get the number of lines.  This is only needed if the frame type is
       IAS_UL_SIZE or IAS_PATH_ORIENTED. */
    int default_nl[] = {0, 0, 0};
    IAS_PARM_ADD_INT(parms, NLINES, 
        "Number of lines in output frame for (reflective, pan, thermal); "
        "required for IAS_UL_SIZE and IAS_PATH_ORIENTED framing",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY, 1, 0, 1000000, 3, default_nl,
        local_nl, sizeof(local_nl), NUM_PIXEL_SIZES);

    /* Get the number of samples.  This is only needed if the frame type is
       IAS_UL_SIZE or IAS_PATH_ORIENTED. */
    int default_ns[] = {0, 0, 0};
    IAS_PARM_ADD_INT(parms, NSAMPS, 
        "Number of samples in output frame for (reflective, pan, thermal); "
        "required for IAS_UL_SIZE and IAS_PATH_ORIENTED framing",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY, 1, 0, 1000000, 3, default_ns,
        local_ns, sizeof(local_ns), NUM_PIXEL_SIZES);

    /* Get the fractional WRS row.  This is only needed for path-oriented 
       scenes. */
    double default_wrs_row_frac[] = {0.0};
    IAS_PARM_ADD_DOUBLE(parms, FRAC_ROW, 
        "Fractional row increment to use in path scene framing; required for "
        "IAS_PATH_ORIENTED, IAS_PATH_MINBOX, IAS_PATH_MAXBOX framing",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 0, 1, 1, 
        default_wrs_row_frac, &row_frac, sizeof(row_frac), 1);

    /* Get the projection parameters.  This is only needed if the projection
       code is not UTM, SOM or PS. Remove the default here, only use the default
       projection parameters once we are certain that none were listed in the
       ODL file and none were listed in the OMF */
    IAS_PARM_ADD_DOUBLE(parms, PROJ_PARMS, 
        "Target projection parameters:  Required for projections other than "
        "UTM, SOM, and PS", IAS_PARM_OPTIONAL, IAS_PARM_ARRAY, 0, 0, 0, 
        0, 0, local_projection_parameters, sizeof(local_projection_parameters), 
        IAS_PROJ_PARAM_SIZE);

    /* Get the name of the grid file. */
    IAS_PARM_GRID_FILENAME(parms, IAS_PARM_REQUIRED, 
        local_grid_file_name, sizeof(local_grid_file_name), 1); 

    /* Get processing pass. */
    IAS_PARM_PROCESSING_PASS(parms, &local_processing_pass, 
        sizeof(local_processing_pass), 1);

    /* Get target projection. Defining separately, as the standard parameter
       is required, this is optional. */
    IAS_PARM_ADD_INT(parms, TARGET_PROJECTION, 
        "Target projection override -- this parameter, if present, "
        "will override the target projection in the OMF. "
        "1 = UTM, 3 = ALBERS, 6 = PS, and 22 = SOM.", 
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 0, MAXPROJ, 0, NULL, 
        &local_projection_code_override, 
        sizeof(local_projection_code_override), 0);

    /* Get optional disable TIRS starting date parameter */
    IAS_PARM_ADD_STRING(parms, DISABLE_TIRS_STARTING_ON_DATE,
       "Date to start disabling the TIRS bands in YYYY:DDD format where YYYY "
       "is the 4 digit year and DDD is the day of the year",
       IAS_PARM_OPTIONAL, 0, NULL, 0, NULL, disable_tirs_starting_date,
       sizeof(disable_tirs_starting_date), 1);

    /* Get the no elevation required flag */
    int default_no_elevation_required_flag[] = {0};
    IAS_PARM_ADD_INT(parms, NO_ELEVATION_REQUIRED_FLAG,
        "Flag to default min/max elevation to 0 if they are not available. "
        "This flag is used for TIRS stray light only.",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 0, 1, 1,
        default_no_elevation_required_flag, no_elevation_required_flag,
        sizeof(*no_elevation_required_flag), 0);

    /* Calculate the number of items in the parameter list. */
    int count = IAS_PARM_GET_TABLE_SIZE(parms);

    /* Call ias_parm_provide_help to provide help for the parameters
       if requested; it also verifies the parameter file exists. */
    status = ias_parm_provide_help(param_filename, parms, count, 
                                   IAS_INPUT_PARAMETERS);
    if (status)
    {
        /* If the status of the provide help is 1 the user got help.  Since
           the parameter file name contains the help request, copy it to the
           OMF file name so the next call will ask for the same help. */
        if (status == 1)
        {
            strcpy(local_omf_file_name, param_filename);
        }
        else if (status == ERROR) /* Help failed. */
        {
            IAS_LOG_ERROR("Providing help using application parameter file %s",
                param_filename);
            exit(EXIT_FAILURE);
        }
    }
    else
    {
        if (ias_parm_read(param_filename, parms, count) == ERROR)
        {
            /* Error reading required item, log an error */
            IAS_LOG_ERROR("Reading application parameter file %s",
                param_filename);
            return ERROR;
        }

        /* Build the OMF file name. */
        sprintf(local_omf_file_name, "%s.omf", local_wo_id);
    }

    /* Now read and check the OMF. */

    /* start parameter table definition */
    IAS_PARM_DECLARE_TABLE(omf_parms, 14);

    /* Get the name of the Calibration parameter file. */
    IAS_PARM_CPF(omf_parms, IAS_PARM_REQUIRED, local_cpf_file_name, 
        sizeof(local_cpf_file_name), 1);

    /* Get the maximum number of processors to use */
    IAS_PARM_MAX_PROCESSORS_TO_USE(omf_parms, IAS_PARM_OPTIONAL,
                                   max_processors_to_use,
                                   sizeof(*max_processors_to_use), 1);

    /* Get the UTM zone.  This is only needed if the TARGET_PROJECTION is 1 
       (UTM). */
    int default_utm_zone[] = {0};
    IAS_PARM_UTM_ZONE(omf_parms, IAS_PARM_OPTIONAL, 1, default_utm_zone,
        &local_utm_zone, sizeof(local_utm_zone), 1);

    /* Get the projection code.  This is only needed for Earth scenes.  The
       default is for UTM. */ 
    int default_projection_code[] = {1};
    IAS_PARM_ADD_INT(omf_parms, TARGET_PROJECTION, "Target projection; " 
        "required for Earth scenes. 1 = UTM, 3 = ALBERS, 6 = PS, and 22 = SOM", 
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 
        0, MAXPROJ, 1, default_projection_code, &local_projection_code, 
        sizeof(local_projection_code), 1);

    /* See if the override with pass is present */
    IAS_PARM_ADD_INT_WITH_PASS(omf_parms, OVERRIDE_PROJECTION,
        "Override target projection (1 = UTM, 3 = ALBERS, 6 = PS, "
        "and 22 = SOM)", IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        1, 0, 30, 0, NULL, &override_projection_with_pass,
        sizeof(override_projection_with_pass), 1, local_processing_pass);

    /* Are the override projection parameters with pass present */
    IAS_PARM_ADD_DOUBLE_WITH_PASS(omf_parms, OVERRIDE_PROJ_PARMS,
        "Override projection parameters", IAS_PARM_OPTIONAL,
        IAS_PARM_ARRAY, 0, 0, 0, 0, NULL,
        omf_projection_parameters, sizeof(omf_projection_parameters),
        IAS_PROJ_PARAM_SIZE, local_processing_pass);

    /* Get maximum elevation.  The source of the elevation will be the
       parameter file first, OMF as backup, and DEM as last backup */
    IAS_PARM_MAX_ELEVATION(omf_parms, IAS_PARM_OPTIONAL, 0, NULL,
        &local_max_elevation, sizeof(local_max_elevation), 1);

    /* Get minimum elevation.  The source of the elevation will be the
       parameter file first, OMF as backup, and DEM as last backup */
    IAS_PARM_MIN_ELEVATION(omf_parms, IAS_PARM_OPTIONAL, 0, NULL,
        &local_min_elevation, sizeof(local_min_elevation), 1);

    /* Get the name of the model file. */
    IAS_PARM_MODEL_FILENAME_FROM_PASS(omf_parms, IAS_PARM_REQUIRED,
        local_model_file_name, sizeof(local_model_file_name), 1,
        local_processing_pass);

    /* Get the name of the DEM file. */
    IAS_PARM_DEM_FILENAME_FROM_PASS(omf_parms, IAS_PARM_OPTIONAL,
        local_dem_file_name, sizeof(local_dem_file_name), 1,
        local_processing_pass);

    /* Get the WRS path.  This is only needed for path-oriented, SOM scenes. */
    IAS_PARM_TARGET_WRS_PATH(omf_parms, IAS_PARM_OPTIONAL, &local_wrs_path,
            sizeof(local_wrs_path), 1);

    /* Get the WRS row.  This is only needed for path-oriented scenes. */
    IAS_PARM_TARGET_WRS_ROW(omf_parms, IAS_PARM_OPTIONAL, &local_wrs_row_int, 
            sizeof(local_wrs_row_int), 1);

    /* Add the satellite number parameter */
    IAS_PARM_SATELLITE(omf_parms, &satellite_number, 
                       sizeof(satellite_number), 1);

    /* Get the acquisition date from omf */
    IAS_PARM_ADD_STRING(omf_parms, ACQUISITION_DATE,
       "Acquisition date", IAS_PARM_OPTIONAL,
       0, NULL, 0, NULL, acquisition_date, sizeof(acquisition_date), 1);

    /* See if the user wanted help or a template with the OMF file. */
    status = ias_parm_provide_help(local_omf_file_name, omf_parms,
        IAS_PARM_GET_TABLE_SIZE(omf_parms), IAS_OMF_PARAMETERS);
    if (status)
    {
        if (status == 1) /* Help was successful. */
        {
            exit(EXIT_SUCCESS);
        }
        else if (status == ERROR) /* Help failed. */
        {
            IAS_LOG_ERROR("Providing help using OMF %s", local_omf_file_name);
            exit(EXIT_FAILURE);
        }
    }

    /* Read the OMF file */
    count = IAS_PARM_GET_TABLE_SIZE(omf_parms);
    if (ias_parm_read(local_omf_file_name, omf_parms, count) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from OMF file: %s", 
            local_omf_file_name);
        return ERROR;
    }

    /* Convert the satellite number to a satellite id */
    satellite_id =
        ias_sat_attr_get_satellite_id_from_satellite_number(satellite_number);
    if (satellite_id == ERROR)
    {
        IAS_LOG_ERROR("Converting the satellite number to a satellite id");
        return ERROR;
    }

    /* Initialize satellite attributes. */
    if (ias_sat_attr_initialize(satellite_id) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the satellite attributes library");
        return ERROR;
    }

    /* Find the number of normal bands for the satellite. */
    satellite_number_bands = ias_sat_attr_get_normal_band_count();
    if ((satellite_number_bands < 1) ||
        (satellite_number_bands > IAS_MAX_NBANDS))
    {
        IAS_LOG_ERROR("Invalid number of bands %d.", satellite_number_bands);
        return ERROR;
    }

    /* Initialize the grid */
    status = ias_grid_initialize(grid, IAS_MAX_NBANDS, IAS_MAX_NSCAS);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing IAS Grid");
        return ERROR;
    }

    /* Initialize the WRS scene center and rotation angle */
    grid->center_pt.lat = 0;
    grid->center_pt.lng = 0;
    grid->proj_center_pt.x = 0.0;
    grid->proj_center_pt.y = 0.0;
    grid->rotate_angle = 0;
    grid->orient_angle = 0.0;

    /* Transfer local values to the grid and other output locations. */
    strcpy(cal_parm_file, local_cpf_file_name);
    strcpy(grid->gridname, local_grid_file_name);
    strcpy(model_file_name, local_model_file_name);
    strcpy(dem_file_name, local_dem_file_name);
    strcpy(omf_file_name, local_omf_file_name);

    for (i = 0; i < satellite_number_bands; i++)
    {
        band_numbers[i] = local_band_numbers[i];
    }
    for (i = 0; i < NUM_PIXEL_SIZES; i++)
    {
        pixsize[i] = local_pixel_sizes[i];
    }
    strcpy (projection_units, local_projection_units);
    grid->frame_type = local_frame_type;

    /* Use target projection (override) in the parameter file if provided */
    if (IAS_PARM_GET_COUNT_READ(parms, TARGET_PROJECTION) > 0)
    {
        grid->code = local_projection_code_override;
    }
    /* If the override_projection_with_pass is present and no override from
       parameter file  was found, use the override from the OMF instead of the
       original projection */
    else if (override_projection_with_pass > 0)
    {
        grid->code = override_projection_with_pass;
    }
    else
    {
        grid->code = local_projection_code;
    }
    for (i = 0; i < NUM_FRAME_COORDINATES; i++)
    {
        frm_coords[i] = local_frame_coordinates[i];
    }
    for (i = 0; i < NUM_LS_COORDINATES; i++)
    {
        ls_coords[i] = local_ls_coordinates[i];
    }
    for (i = 0; i < NUM_PIXEL_SIZES; i++)
    {
        nl[i] = local_nl[i];
        ns[i] = local_ns[i];
    }
    if (grid->code == UTM)
    {
        grid->zone = local_utm_zone;
    }
    *path = local_wrs_path;
    row_int = local_wrs_row_int;

    /* If the projection parameters were read from odl file, use those. They
        will get priority */
    if (IAS_PARM_GET_COUNT_READ(parms, PROJ_PARMS) > 0) 
    {
        for (i = 0; i < IAS_PROJ_PARAM_SIZE; i++)
        {
            projection_parameters[i] = local_projection_parameters[i];
        }
    }
    /* Projection parameters were read from OMF */
    else if (IAS_PARM_GET_COUNT_READ(omf_parms, OVERRIDE_PROJ_PARMS) > 0)
    {
        for (i = 0; i < IAS_PROJ_PARAM_SIZE; i++)
        {
            projection_parameters[i] = omf_projection_parameters[i];
        }
    }
    /* If no projection parameters were read from ODL file, and no projection
       parameters were read from the OMF, use the default projection
       parameters */
    else
    {
        for (i = 0; i < IAS_PROJ_PARAM_SIZE; i++)
        {
            projection_parameters[i] = default_projection_parameters[i];
        }
    }

    /* Set the MIN/MAX elevation using the OMF values */
    if (IAS_PARM_GET_COUNT_READ(omf_parms, MIN_ELEVATION) > 0)
    {
        *zmin = local_min_elevation;
    }
    if (IAS_PARM_GET_COUNT_READ(omf_parms, MAX_ELEVATION) > 0)
    {
        *zmax = local_max_elevation;
    }

    *processing_pass = local_processing_pass;

    /* Perform verifications on the various values, and migrate more of them
       to their final output locations (like the grid structure). */

    /* Verify that the grid filename is present. */
    if (!strcmp(grid->gridname, ""))
    {
        IAS_LOG_ERROR("No grid filename provided in the parameter file.");
        ias_grid_free(grid);
        return ERROR;
    }

    /* Verify that the model filename is present. */
    if (!strcmp(model_file_name, ""))
    {
        IAS_LOG_ERROR("No model filename provided in the OMF.");
        ias_grid_free(grid);
        return ERROR;
    }

    /* Get the model. */
    *model = ias_model_read(model_file_name);
    if (*model == NULL)
    {
        IAS_LOG_ERROR("Reading the model file %s.", model_file_name);
        ias_grid_free(grid);
        return ERROR;
    }

    /* Find the number of bands returned.  The first 0 encountered
       signifies the end of the band list. */
    *number_bands = IAS_PARM_GET_COUNT_READ(parms, BAND_LIST);

    /* Sort the bands so they are put in ascending order in the grid. */
    ias_math_insertion_sort_integer_array(*number_bands, band_numbers);

    /* Make sure there are no duplicate bands, and error out if there are. */
    for (i = 0; i < (*number_bands - 1); i++)
    {
        if (band_numbers[i] == band_numbers[i+1])
        {
            IAS_LOG_ERROR("Duplicate band number (%d) in parameter file.",
                band_numbers[i]);
            ias_los_model_free(*model);
            *model = NULL;
            ias_grid_free(grid);
            return ERROR;
        }
    }

    /* Validate that the bands input fall within a valid range.  If the user 
       did not specify any bands, fill the array with all 
       satellite_number_bands bands. */
    if (*number_bands > 0)
    {
        for (i = 0; i < *number_bands; i++)
        {
            if ((band_numbers[i] <= 0) || 
                (band_numbers[i] > satellite_number_bands))
            {
                IAS_LOG_ERROR("Invalid band number (%d) in parameter file.",
                    band_numbers[i]);
                ias_los_model_free(*model);
                *model = NULL;
                ias_grid_free(grid);
                return ERROR;
            }
        }
    }
    else
    {   /* Fill in all satellite_number_bands bands. */
        for (i = 1; i <= satellite_number_bands; i++)
        {
            band_numbers[i] = i; 

        }
        *number_bands = satellite_number_bands;
    }

    /* If a date to start disabling TIRS bands in the output was provided,
       check whether the acquisition date is on or after the scene acquisition
       date */
    if (IAS_PARM_GET_COUNT_READ(parms, DISABLE_TIRS_STARTING_ON_DATE) > 0)
    {
        if ((IAS_PARM_GET_COUNT_READ(omf_parms, ACQUISITION_DATE) > 0))
        {
            IAS_DATETIME acq_date_time; /* Structure for acquisition date and
                                           time values */
            IAS_DATETIME disable_date_time; /* Structure for disable TIRS
                                               starting on date and time
                                               values */
            char local_disable_tirs_starting_date[IAS_L0R_DATE_LENGTH];

            /* Append an hour/minute/second to the disable date to allow it
               to be parsed by the IAS library routine */
            status = snprintf(local_disable_tirs_starting_date,
                IAS_L0R_DATE_LENGTH, "%s%s",disable_tirs_starting_date,
                ":0:0:0.00000");
            if (status < 0 || status >= IAS_L0R_DATE_LENGTH)
            {
                IAS_LOG_ERROR("Concatenating disable TIRS starting date "
                    "string");
                ias_los_model_free(*model);
                *model = NULL;
                ias_grid_free(grid);
                return ERROR;
            }

            /* Parse the date and time string for acquisition date */
            if (ias_misc_parse_datetime_string(acquisition_date,
                IAS_DATETIME_L0R_FORMAT, &acq_date_time) != SUCCESS)
            {
                IAS_LOG_ERROR("Parsing acquisition date string %s",
                    acquisition_date);
                ias_los_model_free(*model);
                *model = NULL;
                ias_grid_free(grid);
                return ERROR;
            }

            /* Parse the date and time string for disable TIRS starting date */
            if (ias_misc_parse_datetime_string(local_disable_tirs_starting_date,
                IAS_DATETIME_L0R_FORMAT, &disable_date_time) != SUCCESS)
            {
                IAS_LOG_ERROR("Parsing disable TIRS starting date string %s",
                    disable_tirs_starting_date);
                ias_los_model_free(*model);
                *model = NULL;
                ias_grid_free(grid);
                return ERROR;
            }

            /* Check the acquisiton date with disable TIRS starting date,
               if acquisition date is on or after the disable TIRS starting on
               date then set the disable_tirs_band to 1 */
            if ((acq_date_time.year > disable_date_time.year) ||
                 ((acq_date_time.year == disable_date_time.year) &&
                  (acq_date_time.day_of_year >= disable_date_time.day_of_year)))
            {
                IAS_LOG_INFO("TIRS bands are disabled after date %s, so not "
                     "including TIRS bands",disable_tirs_starting_date);
                disable_tirs_bands = 1;
            }
        }
        else
        {
            IAS_LOG_ERROR("No acquisition date is present in the OMF");
            ias_los_model_free(*model);
            *model = NULL;
            ias_grid_free(grid);
            return ERROR;
        }
    }

    /* Verify the bands in the list are available in the model.  If they
       aren't, remove them from the band list and issue a warning.  This is
       done to allow specifying all the bands, but falling back to only the
       bands that are present if only OLI or TIRS data is present. */
    band_index = 0;
    for (i = 0; i < *number_bands; i++)
    {
        int index = ias_sat_attr_convert_band_number_to_index(band_numbers[i]);
        if (index == ERROR)
        {
            IAS_LOG_ERROR("Invalid band number (%d)", band_numbers[i]);
            ias_los_model_free(*model);
            *model = NULL;
            ias_grid_free(grid);
            return ERROR;
        }
        if ((*model)->sensor.bands[index].band_present)
        {
            /* If the disable_tirs_band flag is set and sensor for the current
               band is TIRS, don't add the band to the final list */
            if (disable_tirs_bands)
            {
                /* Get the sensor id */
                IAS_SENSOR_ID sensor_id;
                sensor_id = ias_sat_attr_get_sensor_id_from_band_number(
                    band_numbers[i]);
                if (sensor_id == IAS_INVALID_SENSOR_ID)
                {
                    IAS_LOG_ERROR("Retrieving sensor ID for band number %d",
                        band_numbers[i]);
                    ias_los_model_free(*model);
                    *model = NULL;
                    ias_grid_free(grid);
                    return ERROR;
                }

                /* It is a TIRS band, so issue a warning that it is being
                   skipped and continue with the next band */
                if (sensor_id == IAS_TIRS)
                {
                    IAS_LOG_INFO("Removing TIRS band number %d since TIRS "
                       "bands are disabled for this acquisition date",
                        band_numbers[i]);
                    continue;
                }
            }

            /* This band is present, so keep it in the list */
            band_numbers[band_index] = band_numbers[i];
            band_index++;
        }
        else
        {
            /* The band isn't present, so issue a warning and don't increment
               band_index so that the next one that is present will overwrite
               it. */
            IAS_LOG_WARNING("Band number %d is not in the model, so skipping"
                " it", band_numbers[i]);
        }
    }

    /* save the new number of bands */
    *number_bands = band_index;

    /* The grid has slots even for bands not selected. */
    grid->nbands = satellite_number_bands;

    /* Fill in band_index in the grid. */
    for (i = 0; i < grid->nbands; i++)
    {
        band_index = i;
        for (sca_index = 0; sca_index < grid->scas_per_band[band_index];
             sca_index++)
        {
            grid->gridbands[band_index][sca_index].band_index = band_index;
        }
    }

    /* Put the units in the grid. */
    ias_misc_convert_to_uppercase(projection_units);
    strcpy(grid->units, projection_units);

    /* Look up the spectral type for the various bands. */
    for (i = 0; i < satellite_number_bands; i++)
    {
        band_number = ias_sat_attr_convert_band_index_to_number(i);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Invalid band number %d.", band_number);
            ias_los_model_free(*model);
            *model = NULL;
            ias_grid_free(grid);
            return ERROR;
        }

        const IAS_BAND_ATTRIBUTES *band_attributes = 
            ias_sat_attr_get_band_attributes(band_number);
        if (band_attributes == NULL)
        {
            IAS_LOG_ERROR("Retrieving attributes for band %d", band_number);
            ias_los_model_free(*model);
            *model = NULL;
            ias_grid_free(grid);
            return ERROR;
        }

        spectral_type[i] = band_attributes->spectral_type;
    }

    /* Set the acquisition type. */
    grid->acquisition_type = (*model)->acquisition_type; 

    /* Check to see what kind of acquisition it is. */
    if ((*model)->acquisition_type == IAS_EARTH)
    {
        /* Verify and transfer FRAME_COORDS. */
        if (grid->frame_type == IAS_GEOBOX || grid->frame_type == IAS_PROJBOX)
        {
            if (frm_coords[0] == INVALID_COORDINATE ||
                frm_coords[1] == INVALID_COORDINATE ||
                frm_coords[2] == INVALID_COORDINATE ||
                frm_coords[3] == INVALID_COORDINATE)
            {
                IAS_LOG_ERROR("Invalid FRAME_COORDS value.  For IAS_GEOBOX "
                    "and IAS_PROJBOX, 4 values (2 pairs) must be specified");
                ias_los_model_free(*model);
                *model = NULL;
                ias_grid_free(grid);
                return ERROR;
            }

            if ((frm_coords[0] == 0.0) && (frm_coords[1] == 0.0))
            {
                /* If both of these are zero, it is likely the parameters were
                   not provided correctly.  However, since it is possible (but
                   extremely unlikely) that it could be valid.  So, issue a
                   warning message to help figure out what is wrong when
                   things likely fail later */
                IAS_LOG_WARNING("FRAME_COORDS are (0.0, 0.0) which is likely "
                    "incorrect.  Attempting to continue in case they are "
                    "correct.");
            }

            /* Fill the frame_coords structure appropriately. */
            if (grid->code == SOM)
            {
                frame_coords[0].y = -frm_coords[1];
                frame_coords[0].x = frm_coords[0];
                frame_coords[1].y = -frm_coords[3];
                frame_coords[1].x = frm_coords[2];
            }
            else
            {
                frame_coords[0].y = frm_coords[0];
                frame_coords[0].x = frm_coords[1];
                frame_coords[1].y = frm_coords[2];
                frame_coords[1].x = frm_coords[3];
            }

            /* Check for mismatch between frame coordinates and pixel size */
            for (i = 0; i < NUM_PIXEL_SIZES; i++)
            {
                /* Define the scale of error to tolerate: 1/100th of a pixel */
                division_tolerance = pixsize[i] / 100.0;

                /* Exit if difference between coordinates isn't evenly divisible
                   by the pixel sizes for both x and y directions */
                remainder_check = fmod(fabs(frame_coords[0].x 
                    - frame_coords[1].x), pixsize[i]);

                if ((remainder_check >= division_tolerance) && 
                    (remainder_check <= pixsize[i] - division_tolerance))  
                {
                    IAS_LOG_ERROR("Inconsistent x frame coordinates %f and %f "
                        "and pixel size %f.  The difference between the frame "
                        "coordinates must be evenly divisible by the pixel "
                        "size.", frame_coords[0].x, frame_coords[1].x,
                        pixsize[i]);
                    ias_los_model_free(*model);
                    *model = NULL;
                    ias_grid_free(grid);
                    return ERROR;
                }

                remainder_check = fmod(fabs(frame_coords[0].y 
                    - frame_coords[1].y), pixsize[i]);

                if ((remainder_check >= division_tolerance) && 
                    (remainder_check <= pixsize[i] - division_tolerance))  
                {
                    IAS_LOG_ERROR("Inconsistent y frame coordinates %f and %f "
                        "and pixel size %f.  The difference between the frame "
                        "coordinates must be evenly divisible by the pixel "
                        "size.", frame_coords[0].y, frame_coords[1].y,
                        pixsize[i]);
                    ias_los_model_free(*model);
                    *model = NULL;
                    ias_grid_free(grid);
                    return ERROR;
                }

                /* Give a warning if the corner coordinates aren't divisible 
                   by the pixel size, for both x and y directions */ 
                for (coordinate = 0; coordinate < 2; coordinate++)
                {
                    remainder_check = fmod(fabs(frame_coords[coordinate].x), 
                        pixsize[i]);

                    if ((remainder_check >= division_tolerance) && 
                        (remainder_check <= pixsize[i] - division_tolerance))  
                    {
                        IAS_LOG_WARNING("Inconsistent x frame coordinate %f "
                            "and pixel size %f.  The frame coordinate should "
                            "be evenly divisible by the pixel size.",
                            frame_coords[coordinate].x, pixsize[i]);
                    }

                    remainder_check = fmod(fabs(frame_coords[coordinate].y), 
                        pixsize[i]);

                    if ((remainder_check >= division_tolerance) && 
                        (remainder_check <= pixsize[i] - division_tolerance))  
                    {
                        IAS_LOG_WARNING("Inconsistent y frame coordinate %f "
                            "and pixel size %f.  The frame coordinate should "
                            "be evenly divisible by the pixel size.",
                            frame_coords[coordinate].y, pixsize[i]);
                    }
                }
            }
        }
        else if (grid->frame_type == IAS_UL_SIZE)
        {
            if (frm_coords[0] == INVALID_COORDINATE ||
                frm_coords[1] == INVALID_COORDINATE)
            {
                IAS_LOG_ERROR("Invalid FRAME_COORDS value.  For IAS_UL_SIZE, "
                    "2 values (1 pair) must be specified");
                ias_los_model_free(*model);
                *model = NULL;
                ias_grid_free(grid);
                return ERROR;
            }

            /* Fill the frame_coords structure appropriately. */
            if (grid->code == SOM)
            {
                frame_coords[0].y = -frm_coords[1];
                frame_coords[0].x = frm_coords[0];
            }
            else
            {
                frame_coords[0].y = frm_coords[0];
                frame_coords[0].x = frm_coords[1];
            }
        }

        /* If the TARGET_PROJECTION != GEOGRAPHIC, then the proj_units 
           must be in METERS or FEET. */
        if (grid->code != 0)
        {
            if ((strcmp(grid->units, "FEET") != 0) && 
                (strcmp(grid->units, "METERS") != 0))
            { /* Invalid proj units were given for the proj code. */
                IAS_LOG_ERROR("Since a non-Geographic TARGET_PROJECTION is in " 
                    "the parameter file PROJ_UNITS should be FEET or METERS");
                ias_los_model_free(*model);
                *model = NULL;
                ias_grid_free(grid);
                return ERROR;
            }
        }

        /* Set ls_coords if the frame type is PROJBOX or UL_SIZE. */
        if (grid->frame_type == IAS_PROJBOX || grid->frame_type == IAS_UL_SIZE)
        {
            /* Fill the lscoords structure appropriately.  */
            lscoords->line = ls_coords[0];
            lscoords->samp = ls_coords[1];
        }

        /* Set the number of lines.  Only do for UL_SIZE - required, and
           PATH_ORIENTED - optional.  The order of the NLINES should be
           30m reflective, 15m pan, 30/100m thermal. */
        if (grid->frame_type == IAS_UL_SIZE ||
            grid->frame_type == IAS_PATH_ORIENTED)
        {
            if (calc_size(pixsize, lines_in_scene, grid->frame_type, nl)
                != SUCCESS)
            {
                IAS_LOG_ERROR("Calculating the number of lines in the output "
                              "image");
                ias_los_model_free(*model);
                *model = NULL;
                ias_grid_free(grid);
                return ERROR;
            }

            /* Fill in the nlines for each band (selected or not). */
            for (i = 0; i < satellite_number_bands; i++)
            {
                for (sca_index = 0; sca_index < grid->scas_per_band[i];
                     sca_index++)
                {
                    if ((spectral_type[i] == IAS_SPECTRAL_VNIR)
                        || (spectral_type[i] == IAS_SPECTRAL_SWIR))
                    {
                        grid->gridbands[i][sca_index].lines = nl[0];
                    }
                    else if (spectral_type[i] == IAS_SPECTRAL_PAN)
                    {
                        grid->gridbands[i][sca_index].lines = nl[1];
                    }
                    else if (spectral_type[i] == IAS_SPECTRAL_THERMAL)
                    {
                        grid->gridbands[i][sca_index].lines = nl[2];
                    }
                    else
                    {
                        IAS_LOG_ERROR("Unhandled spectral type of %d for band "
                                      "index %d", spectral_type[i], i);
                        ias_los_model_free(*model);
                        *model = NULL;
                        ias_grid_free(grid);
                        return ERROR;
                    }
                }
            }
        }

        /* Set the number of samples.  Only do for UL_SIZE - required, 
           and PATH_ORIENTED - optional.  The order of the NSAMPS should 
           be 30m, 15m. */
        if (grid->frame_type == IAS_UL_SIZE ||
            grid->frame_type == IAS_PATH_ORIENTED)
        {
            if (calc_size(pixsize, samps_in_scene, grid->frame_type, ns)
                != SUCCESS)
            {
                IAS_LOG_ERROR("Calculating the number of samples in the output "
                              "image");
                ias_los_model_free(*model);
                *model = NULL;
                ias_grid_free(grid);
                return ERROR;
            }

            /* Fill in the nsamps for each band (selected or not). */
            for (i = 0; i < satellite_number_bands; i++)
            {
                for (sca_index = 0; sca_index < grid->scas_per_band[i];
                     sca_index++)
                {
                    if ((spectral_type[i] == IAS_SPECTRAL_VNIR)
                        || (spectral_type[i] == IAS_SPECTRAL_SWIR))
                    {
                        grid->gridbands[i][sca_index].samps = ns[0];
                    }
                    else if (spectral_type[i] == IAS_SPECTRAL_PAN)
                    {
                        grid->gridbands[i][sca_index].samps = ns[1];
                    }
                    else if (spectral_type[i] == IAS_SPECTRAL_THERMAL)
                    {
                        grid->gridbands[i][sca_index].samps = ns[2];
                    }
                    else
                    {
                        IAS_LOG_ERROR("Unhandled spectral type of %d for band "
                                      "index %d", spectral_type[i], i);
                        ias_los_model_free(*model);
                        *model = NULL;
                        ias_grid_free(grid);
                        return ERROR;
                    }
                }
            }
        }

        /* Verify the path was read for PATH_ORIENTED and SOM projections. */
        if (grid->frame_type == IAS_PATH_ORIENTED || 
            grid->code == SOM ||
            grid->frame_type == IAS_PATH_MINBOX ||
            grid->frame_type == IAS_PATH_MAXBOX)
        {
            if (IAS_PARM_GET_COUNT_READ(omf_parms, TARGET_WRS_PATH) == 0)
            {
                IAS_LOG_ERROR("TARGET_WRS_PATH not found in the OMF file");
                ias_los_model_free(*model);
                *model = NULL;
                ias_grid_free(grid);
                return ERROR;
            }
        }
    
        /* Verify the row_int was read for PATH_ORIENTED. */
        if (grid->frame_type == IAS_PATH_ORIENTED 
            || grid->frame_type == IAS_PATH_MINBOX 
            || grid->frame_type == IAS_PATH_MAXBOX)
        {
            if (IAS_PARM_GET_COUNT_READ(omf_parms, TARGET_WRS_ROW) == 0)
            {
                IAS_LOG_ERROR("TARGET_WRS_ROW not found in the OMF file");
                ias_los_model_free(*model);
                *model = NULL;
                ias_grid_free(grid);
                return ERROR;
            }

            /* There is a similar check when the parameter is read, but this
               one captures the exact floating point range needed. */
            if (row_frac >= 1.0 || row_frac < 0.0)
            {
                IAS_LOG_ERROR("In the fractional part of the WRS row in the "
                    "parameter file");
                ias_los_model_free(*model);
                *model = NULL;
                ias_grid_free(grid);
                return ERROR;
            }

            /* ... else put the wrs_row into the row parameter. */
            *row = row_int + row_frac;
        }

        /* Set the projection parameters based on the projection requested */
        if (grid->code == UTM)
        {
            /* UTM doesn't really use the projection parameters, so zero fill */
            for (i = 0; i < IAS_PROJ_PARAM_SIZE; i++)
            {
                grid->projprms[i] = 0.0;
            }
        }
        else if (grid->code == SOM)
        {
            /* Define the SOM projection using the satellite number option */
            for (i = 0; i < IAS_PROJ_PARAM_SIZE; i++)
            {
                grid->projprms[i] = 0.0;
            }
            grid->projprms[2] = 8.0;              /* satellite number */
            grid->projprms[3] = (double) *path;
            grid->projprms[12] = 1.0;             /* SOMB             */
            /* This step includes functionality consistent with an altered 
               version of GCTPC.  This version of GCTPC is different from the 
               released version.  We added a pflag to SOMB and changed the way 
               the flag performs.  This must be evaluated when a new version 
               of GCTPC is adapted.  Add the pflag value to the parameter array
               according to row.  For the northern rows, set to start flag (0).
               For the southern hemisphere rows set the start flag to (1).  */
            if (*row >= 183)
                grid->projprms[10] = 1;    /* End of path */
            else
                grid->projprms[10] = 0;    /* Start of path */
        }
        else if (grid->code == PS)
        {
            /* For Polar Stereographic, select the north or south pole version
               based on the target row */
            for (i = 0; i < IAS_PROJ_PARAM_SIZE; i++)
            {
                grid->projprms[i] = 0.0;
            }
            /* Semi-major and minor axes */
            grid->projprms[0] = 6378137.0;
            grid->projprms[1] = 6356752.314245;

            if (IAS_PARM_GET_COUNT_READ(omf_parms, TARGET_WRS_ROW) == 0)
            {
                IAS_LOG_ERROR("TARGET_WRS_ROW not found in the OMF file");
                ias_los_model_free(*model);
                *model = NULL;
                ias_grid_free(grid);
                return ERROR;
            }

            /* Latitude of true scale, encoded in DMS, assuming north pole */
            grid->projprms[5] = 71000000.0;

            /* If close to the south pole (WRS row in the southern hemisphere
               or one of the special off-nadir values at the south pole),
               use the latitude of true scale for the south pole. */
            if ((row_int > 60 && row_int < 188) || (row_int >= 900))
                grid->projprms[5] = -71000000.0;
        }
        else
        {
            /* Fall back to using the provided projection parameters if
               not one of the other projections */
            for (i = 0; i < IAS_PROJ_PARAM_SIZE; i++)
            {
                grid->projprms[i] = projection_parameters[i];
            }
        }

    } /* End of Earth based acquisition */
    else if ((*model)->acquisition_type == IAS_STELLAR || 
             (*model)->acquisition_type == IAS_LUNAR)
    {
        if ((*model)->acquisition_type == IAS_STELLAR)
        {
            grid->frame_type = IAS_STELLAR_FRAME;
        }
        else
        {
            if ((grid->frame_type != IAS_LUNAR_MINBOX) && 
                (grid->frame_type != IAS_LUNAR_MAXBOX))
            {
                IAS_LOG_WARNING("Lunar Frame Type, %d, is not valid for "
                    "processing a lunar image, set to default LUNAR_MINBOX",
                    grid->frame_type);
                grid->frame_type = IAS_LUNAR_MINBOX;
             }
        }

        /* set the projection code to GEO and zone to 0 for lunar */
        grid->code = GEO;
        grid->zone = 0;

        for (i = 0; i < IAS_PROJ_PARAM_SIZE; i++)
        {
            grid->projprms[i] = 0.0;
        }

        /* For non-Earth acquisitions, the units should be in radians, 
           seconds, or degrees. */
        if (strcmp(grid->units, "RADIANS") != 0 &&
            strcmp(grid->units, "SECONDS") != 0 &&
            strcmp(grid->units, "DEGREES") != 0)
        {
            IAS_LOG_ERROR(
                "Invalid proj units (%s) were given in the parameter file",
                grid->units);
            ias_los_model_free(*model);
            *model = NULL;
            ias_grid_free(grid);
            return ERROR;
        }

        /* Set the elevation range to a default range for lunar so it
           doesn't try to read a DEM or use more than 2 elevation planes */
        *zmin = 0;
        *zmax = 500;

    }
    else
    {
        IAS_LOG_ERROR("An unknown acquisition type (%d) was given in the "
            "model file", (*model)->acquisition_type);
        ias_los_model_free(*model);
        *model = NULL;
        ias_grid_free(grid);
        return ERROR;
    }

    /* Fill in the pixsizes in the grid.  It is assumed that the grid band 
       index matches that of the satellite attributes library. */
    for (i = 0; i < satellite_number_bands; i++)
    {
        for (sca_index = 0; sca_index < grid->scas_per_band[i]; sca_index++)
        {
            if ((spectral_type[i] == IAS_SPECTRAL_VNIR)
                || (spectral_type[i] == IAS_SPECTRAL_SWIR))
            {
                grid->gridbands[i][sca_index].pixsize = pixsize[0];
            }
            else if (spectral_type[i] == IAS_SPECTRAL_PAN)
            {
                grid->gridbands[i][sca_index].pixsize = pixsize[1];
            }
            else if (spectral_type[i] == IAS_SPECTRAL_THERMAL)
            {
                grid->gridbands[i][sca_index].pixsize = pixsize[2];
            }
            else
            {
                IAS_LOG_ERROR("Unhandled spectral type of %d for band "
                              "index %d", spectral_type[i], i);
                ias_los_model_free(*model);
                *model = NULL;
                ias_grid_free(grid);
                return ERROR;
            }
        }
    }

    /* If the maximum and minimum elevation values were not in the parameter
       file, use the DEM for these values if it's available.  Otherwise, use
       default values. */
    if ((*zmin == INVALID_ELEVATION) || (*zmax == INVALID_ELEVATION))
    {
        if (*no_elevation_required_flag)
        {
            *zmin = 0;
            *zmax = 0;
        }
        else if (strcmp(dem_file_name, "") == 0)
        {
            IAS_LOG_ERROR("Neither elevation values nor DEM filename are "
                "present in the OMF");
            ias_los_model_free(*model);
            *model = NULL;
            ias_grid_free(grid);
            return ERROR;
        }
    }

    /* Initialize the dtype to DOUBLE.  */
    grid->dtype = IAS_R8;

    return SUCCESS;
}

/****************************************************************************
Name: calc_size

Purpose: Local routine to help calculate the size of the output image given a
    pixel size, default dimensions, frame type, and input dimensions for each
    resolution.  

Returns: SUCCESS or ERROR

Notes:
    - This routine was created to clean up cal/val provided code that we
      don't really have requirements to support.  However, to have any hope of
      it working it needed fixing when TIRS support was added.  This appears
      to work as expected, but don't spend a lot of time maintaining it until
      we have requirements for the functionality.
****************************************************************************/
int calc_size
(
    const float pixsize[NUM_PIXEL_SIZES],     /* I: pixel sizes */
    const int default_dim[NUM_PIXEL_SIZES],   /* I: default dimension to use */
    int frame_type,                           /* I: frame type */
    int dim[NUM_PIXEL_SIZES]                  /* I/O: dimension to update */
)
{
    int num_set = 0;
    int value;
    int i;

    /* determine how many NL parameters are set */
    for (i = 0; i < NUM_PIXEL_SIZES; i++)
    {
        if (dim[i] != 0)
            num_set++;
    }

    if (num_set == 0 && frame_type == IAS_UL_SIZE)
    {  
        /* If the lines were not provided or they were all zero. */
        IAS_LOG_ERROR("In the number of lines in the parameter file");
        return ERROR;
    }

    /* If number of lines or samples is not specified and frame_type is
       path-oriented, use the nominal image size */
    else if (num_set == 0)
    {
        for (i = 0; i < NUM_PIXEL_SIZES; i++)
        {
            if (pixsize[i] < 20.0)        /* 15m resolution */
                dim[i] = default_dim[1];
            else if (pixsize[i] > 50.0)   /* 100m resolution */
                dim[i] = default_dim[2];
            else                          /* 30m resolution */
                dim[i] = default_dim[0];
        }
    }

    /* Use the available resolutions to determine the value of the
       unspecified nlines.  If one value is specified but other values 
       are not, then calculate the other values based on the given 
       values.  Use the pixel size of the band to determine its 
       resolution. */
    if ((num_set > 0) && (num_set < NUM_PIXEL_SIZES))
    {
        int index_found = 0;
        float pixsize_found;

        /* At least one dim was specified for a band resolution.
           Find the first specified number of lines.  */
        for (i = 0; i < NUM_PIXEL_SIZES; i++)
        {
            if (dim[i] != 0)
            {
                index_found = i;
                break;
            }
        }
        /* Use this to determine the other values. */
        value = dim[index_found];
        pixsize_found = pixsize[index_found];

        for (i = 0; i < NUM_PIXEL_SIZES; i++)
        {
            /* If the line value was not specified, then calculate it 
               from the first specified line value and its band 
               resolution. */
            if (dim[i] == 0)
            {
                if (pixsize[i] < pixsize_found)
                {
                    dim[i] = pixsize_found / pixsize[i] * value
                           - (int)((pixsize_found - pixsize[i])/pixsize[i]);
                }
                else if (pixsize[i] > pixsize_found)
                {
                    dim[i] = pixsize_found / pixsize[i] * value + 1;
                }
                else
                {
                    dim[i] = value;
                }
            }
        }
    }

    return SUCCESS;
}
