/******************************************************************************
NAME:           pad_corners

PURPOSE:
Pads the input corners by a defined factor of the pixel size.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
The x/y min and max values are input for the corner locations.  These values
are padded by PADVAL * the pixel size.  The newly padded x/y min and max
values are returned, replacing the original values.
******************************************************************************/
#include <math.h>

#include "ias_logging.h"              /* Logging library */
#include "create_grid.h"

/* Value for padding the corner points.  This is 10.0 to be consistent with
Landsat and to provide a safe margin to account for the detector geometry. */
#define PADVAL 10.0

int pad_corners 
(
    double pixsize,  /* I: Pixel size for the coordinates */
    double *pxmax,   /* I/O: Maximum X projection coordinate */
    double *pxmin,   /* I/O: Minimum X projection coordinate */
    double *pymax,   /* I/O: Maximum Y projection coordinate */
    double *pymin    /* I/O: Minimum Y projection coordinate */
)
{
    int ixmin;       /* Pixel size increment to be added to pxmin */
    int ixmax;       /* Pixel size increment to be added to pxmax */
    int iymin;       /* Pixel size increment to be added to pymin */
    int iymax;       /* Pixel size increment to be added to pymax */

    /* Pixel size of 0.0 will result in a divide by zero.  */
    if (pixsize == 0.0)
    {
        IAS_LOG_ERROR("Pixel size is zero");
        return ERROR;
    }

    /* Pad corners to PADVAL * pixel size.  */
    ixmin = (int) floor(*pxmin / (PADVAL * pixsize));
    *pxmin = (double) ixmin * PADVAL * pixsize;
    ixmax = (int) ceil(*pxmax / (PADVAL * pixsize));
    *pxmax = (double) ixmax * PADVAL * pixsize;
    iymin = (int) floor(*pymin / (PADVAL * pixsize));
    *pymin = (double) iymin * PADVAL * pixsize;
    iymax = (int) ceil(*pymax / (PADVAL * pixsize));
    *pymax = (double) iymax * PADVAL * pixsize;

    return SUCCESS;
}
