/******************************************************************************
NAME: set_grid_active_corners

PURPOSE: This routine converts the original grid corner points to lat/long, and
    sets them in the grid active_corners structure.  It is used for the maxbox
    framing options, which use the original corner locations.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed
******************************************************************************/
#include "ias_geo.h"                  /* Geographic library */
#include "ias_logging.h"              /* Logging library */
#include "create_grid.h"

int set_grid_active_corners 
(
    IAS_GRID_TYPE *grid     /* I/O: Grid structure */
)
{
    int band_index;            /* Band index */
    struct IAS_CORNERS active_corners; /* Active area corners in lat/long */
    const IAS_GEO_PROJ_TRANSFORMATION *proj2degrees; /* Projection 
                                transformation parameters */
    int status;               /* Function return status */


    /* Convert the product corners to latitude/longitude */

    /* initialize projection transformation */
    proj2degrees = get_proj_transform_proj2degrees();

    /* Upper Left */
    status = ias_geo_transform_coordinate(proj2degrees, grid->corners.upleft.x,
        grid->corners.upleft.y, &active_corners.upleft.x, 
        &active_corners.upleft.y);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating location in lat/long for output space "
            "upper left corner"); 
        return ERROR;
    }

    /* Upper Right */
    status = ias_geo_transform_coordinate(proj2degrees, grid->corners.upright.x,
        grid->corners.upright.y, &active_corners.upright.x, 
        &active_corners.upright.y);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating location in lat/long for output space "
            "upper right corner"); 
        return ERROR;
    }

    /* Lower Left */
    status = ias_geo_transform_coordinate(proj2degrees, grid->corners.loleft.x,
        grid->corners.loleft.y, &active_corners.loleft.x, 
        &active_corners.loleft.y);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating location in lat/long for output space "
            "lower left corner"); 
        return ERROR;
    }

    /* Lower Right */
    status = ias_geo_transform_coordinate(proj2degrees, grid->corners.loright.x,
        grid->corners.loright.y, &active_corners.loright.x, 
        &active_corners.loright.y);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating location in lat/long for output space "
            "lower right corner"); 
        return ERROR;
    }

    /* Set all of the active areas equal to the product corners */
    for (band_index = 0; band_index < grid->nbands; band_index++)
    {
        /* skip bands not available */
        if (!grid->bands_available[band_index])
            continue;

        grid->active_corners[band_index] = active_corners;
    }

    return SUCCESS;
}
