/******************************************************************************
NAME:           set_grid_input_lines_samples

PURPOSE:
Determine the grid lines and samples in the image.

RETURN VALUE:   None 

NOTES:
This routine will set in_lines and in_samps of the grid, based on
the specified band numbers in the grid.  The grid_cell_size.line and
grid_cell_size.samp in the grid specify the size of the grid cells and are
used to calculate the positioning for the grid lines and samples.  The
num_lines and num_samps arrays must be in band-referenced order.

ALGORITHM REFERENCES:
See ias_grid.h for a description of the grid structure.

******************************************************************************/
#include "create_grid.h"

void set_grid_input_lines_samples 
(
    const int num_lines[IAS_MAX_NBANDS],   /* I: Number of input lines     */
    const int num_samples[IAS_MAX_NBANDS], /* I: Number of input samples   */ 
    IAS_GRID_TYPE *grid                    /* I/O: Grid structure          */
)
{
    int j;                     /* Loop counter                             */
    int sca_index;             /* SCA index                                */
    int band_index;            /* Band index                               */ 
    int count;                 /* Counter for the number of lines/samps    */
    int number_lines;          /* Number of lines for current band         */
    int number_samples;        /* Number of samples for current band       */
    int number_scas;           /* Number of SCAS for current band */
    IAS_GRID_BAND_TYPE *grid_band; /* Pointer to the current grid band     */

    /* Loop through bands in the grid.  Note line/sample are zero-relative */ 
    for (band_index = 0; band_index < grid->nbands; band_index++)
    {
        /* Skip bands that aren't present */
        if (!grid->bands_present[band_index])
            continue;

        grid_band = grid->gridbands[band_index]; 
        number_lines = num_lines[band_index];
        number_samples = num_samples[band_index];
        number_scas = grid->scas_per_band[band_index];

        for (sca_index = 0; sca_index < number_scas; sca_index++)
        {
            /* Fill the grid lines */
            grid_band[sca_index].in_lines[0] = 0;
            for (j = grid_band[sca_index].grid_cell_size.line, count = 1;   
                 j < number_lines;
                 j += grid_band[sca_index].grid_cell_size.line)
            {
                grid_band[sca_index].in_lines[count++] = j - 1;
            }
            grid_band[sca_index].in_lines[count] = number_lines - 1;

            /* Fill the grid samples */
            grid_band[sca_index].in_samps[0] = 0;
            for (j = grid_band[sca_index].grid_cell_size.samp, count = 1;  
                 j < number_samples;
                 j += grid_band[sca_index].grid_cell_size.samp)
            {
                grid_band[sca_index].in_samps[count++] = j - 1; 
            }
            grid_band[sca_index].in_samps[count] = number_samples - 1;
        }
    }
}
