#include <stdlib.h>
#include <string.h>
#include "ias_types.h"
#include "ias_const.h"
#include "ias_logging.h"
#include "cgreen/cgreen.h"
#include "create_grid.h"

#define NUM_PIXEL_SIZES              3 /* Reflective, pan, and thermal */
#define WRS_SCENE_SIZE_LINES      5965 /* Number of lines in WRS scene */

int calc_size
(
    const float pixsize[NUM_PIXEL_SIZES],     /* I: pixel sizes */
    const int default_dim[NUM_PIXEL_SIZES],   /* I: default dimension to use */
    int frame_type,                           /* I: frame type */
    int dim[NUM_PIXEL_SIZES]                  /* I/O: dimension to update */
);

Ensure test_calc_size()
{
    /* test the calc_size routine for the lines (sample functionality is the
       same - just passing different parameters) */
    const int lines_in_scene[NUM_PIXEL_SIZES] = 
    {
        WRS_SCENE_SIZE_LINES,               /* 30 meters */
        2 * WRS_SCENE_SIZE_LINES - 1,       /* 15 meters */
        3.0 * WRS_SCENE_SIZE_LINES / 10.0 + 1,  /* 100 METERS */
    };
    float pixsize[NUM_PIXEL_SIZES]; /* Pixel size (30m, 15m, 30m) */
    int nl[NUM_PIXEL_SIZES];      /* Num of lines of the band resolutions
                                    (30m, 15m, 100m) */
    int status;

    /* test using the default sizes */
    pixsize[0] = 30.0;
    pixsize[1] = 15.0;
    pixsize[2] = 100.0;
    nl[0] = 0;
    nl[1] = 0;
    nl[2] = 0;

    status = calc_size(pixsize, lines_in_scene, IAS_PATH_ORIENTED, nl);
    assert_equal(status, SUCCESS);
    assert_equal(nl[0], lines_in_scene[0]);
    assert_equal(nl[1], lines_in_scene[1]);
    assert_equal(nl[2], lines_in_scene[2]);

    /* test the defaults again, this time with thermal being 30m */
    pixsize[0] = 30.0;
    pixsize[1] = 15.0;
    pixsize[2] = 30.0;
    nl[0] = 0;
    nl[1] = 0;
    nl[2] = 0;

    status = calc_size(pixsize, lines_in_scene, IAS_PATH_ORIENTED, nl);
    assert_equal(status, SUCCESS);
    assert_equal(nl[0], lines_in_scene[0]);
    assert_equal(nl[1], lines_in_scene[1]);
    assert_equal(nl[2], lines_in_scene[0]); /* this is 30m size now */

    /* now test just setting the first number of lines */
    pixsize[0] = 30.0;
    pixsize[1] = 15.0;
    pixsize[2] = 100.0;
    nl[0] = lines_in_scene[0];
    nl[1] = 0;
    nl[2] = 0;

    status = calc_size(pixsize, lines_in_scene, IAS_PATH_ORIENTED, nl);
    assert_equal(status, SUCCESS);
    assert_equal(nl[0], lines_in_scene[0]);
    assert_equal(nl[1], lines_in_scene[1]);
    assert_equal(nl[2], lines_in_scene[2]);

    /* test the first number of lines with the third pixel size matching the
       first */
    pixsize[0] = 30.0;
    pixsize[1] = 15.0;
    pixsize[2] = 30.0;
    nl[0] = lines_in_scene[0];
    nl[1] = 0;
    nl[2] = 0;

    status = calc_size(pixsize, lines_in_scene, IAS_PATH_ORIENTED, nl);
    assert_equal(status, SUCCESS);
    assert_equal(nl[0], lines_in_scene[0]);
    assert_equal(nl[1], lines_in_scene[1]);
    assert_equal(nl[2], lines_in_scene[0]); /* should be the 30m size now */

    /* now test the second being set */
    pixsize[0] = 30.0;
    pixsize[1] = 15.0;
    pixsize[2] = 100.0;
    nl[0] = 0;
    nl[1] = lines_in_scene[1];
    nl[2] = 0;

    status = calc_size(pixsize, lines_in_scene, IAS_PATH_ORIENTED, nl);
    assert_equal(status, SUCCESS);
    assert_equal(nl[0], lines_in_scene[0]);
    assert_equal(nl[1], lines_in_scene[1]);
    assert_equal(nl[2], lines_in_scene[2]);

    /* now test the second being set and thermal being 30m */
    pixsize[0] = 30.0;
    pixsize[1] = 15.0;
    pixsize[2] = 30.0;
    nl[0] = 0;
    nl[1] = lines_in_scene[1];
    nl[2] = 0;

    status = calc_size(pixsize, lines_in_scene, IAS_PATH_ORIENTED, nl);
    assert_equal(status, SUCCESS);
    assert_equal(nl[0], lines_in_scene[0]);
    assert_equal(nl[1], lines_in_scene[1]);
    assert_equal(nl[2], lines_in_scene[0]); /* should be the 30m size now */

    /* now test the third being set */
    pixsize[0] = 30.0;
    pixsize[1] = 15.0;
    pixsize[2] = 90.0;
    nl[0] = 0;
    nl[1] = 0;
    nl[2] = 2;

    status = calc_size(pixsize, lines_in_scene, IAS_PATH_ORIENTED, nl);
    assert_equal(status, SUCCESS);
    assert_equal(nl[0], 4);
    assert_equal(nl[1], 7);
    assert_equal(nl[2], 2);

    /* now test the third being set to the 30m size*/
    pixsize[0] = 30.0;
    pixsize[1] = 15.0;
    pixsize[2] = 30.0;
    nl[0] = 0;
    nl[1] = 0;
    nl[2] = lines_in_scene[0];

    status = calc_size(pixsize, lines_in_scene, IAS_PATH_ORIENTED, nl);
    assert_equal(status, SUCCESS);
    assert_equal(nl[0], lines_in_scene[0]);
    assert_equal(nl[1], lines_in_scene[1]);
    assert_equal(nl[2], lines_in_scene[0]); /* should be 30m size now */

    /* now test with 2 being set */
    pixsize[0] = 30.0;
    pixsize[1] = 15.0;
    pixsize[2] = 100.0;
    nl[0] = lines_in_scene[0];
    nl[1] = lines_in_scene[1];
    nl[2] = 0;

    status = calc_size(pixsize, lines_in_scene, IAS_PATH_ORIENTED, nl);
    assert_equal(status, SUCCESS);
    assert_equal(nl[0], lines_in_scene[0]);
    assert_equal(nl[1], lines_in_scene[1]);
    assert_equal(nl[2], lines_in_scene[2]);

    /* now test with 2 being set and thermal set to 30m */
    pixsize[0] = 30.0;
    pixsize[1] = 15.0;
    pixsize[2] = 30.0;
    nl[0] = lines_in_scene[0];
    nl[1] = lines_in_scene[1];
    nl[2] = 0;

    status = calc_size(pixsize, lines_in_scene, IAS_PATH_ORIENTED, nl);
    assert_equal(status, SUCCESS);
    assert_equal(nl[0], lines_in_scene[0]);
    assert_equal(nl[1], lines_in_scene[1]);
    assert_equal(nl[2], lines_in_scene[0]); /* should be 30m size now */

    /* check the lowest resolution for a simple case */
    pixsize[0] = 30.0;
    pixsize[1] = 15.0;
    pixsize[2] = 100.0;
    nl[0] = 0;
    nl[1] = 13;
    nl[2] = 0;

    status = calc_size(pixsize, lines_in_scene, IAS_PATH_ORIENTED, nl);
    assert_equal(status, SUCCESS);
    assert_equal(nl[0], 7);
    assert_equal(nl[1], 13);
    assert_equal(nl[2], 2);

    /* check the middle resolution for a simple case */
    pixsize[0] = 30.0;
    pixsize[1] = 15.0;
    pixsize[2] = 90.0;
    nl[0] = 4;
    nl[1] = 0;
    nl[2] = 0;

    status = calc_size(pixsize, lines_in_scene, IAS_PATH_ORIENTED, nl);
    assert_equal(status, SUCCESS);
    assert_equal(nl[0], 4);
    assert_equal(nl[1], 7);
    assert_equal(nl[2], 2);
}

int main(int argc, char *argv[])
{
    TestSuite *suite = create_test_suite();
    add_test(suite, test_calc_size);
    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
