/******************************************************************************
NAME:           validate_utm_zone

PURPOSE:
This routine will validate the UTM zone that the user entered as a parameter.
The scene center longitude will be used for this verification.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed
******************************************************************************/
#include <stdlib.h>

#include "ias_math.h"
#include "ias_logging.h"              /* Logging library */
#include "create_grid.h"

int validate_utm_zone 
(
    int nlines,          /* I: Number of lines in image */
    int nsamps,          /* I: Number of samples in image */
    int band_index,      /* I: Index for the current band */
    int sca_index,       /* I: Index for the current SCA */
    const IAS_LOS_MODEL *model, /* I: model structure */
    int zone,            /* I: Zone value to validate */
    int *computed_zone,  /* O: Computed zone value for the scene */
    int *valid           /* O: Flag to specify if the UTM zone was valid */
)
{
    IAS_DBL_LS center_ls;             /* Center point line/sample */
    IAS_DBL_LAT_LONG center_lat_long; /* Center point lat/long */
    double elev = 0.0;                /* Elevation at the current point */
    IAS_SENSOR_DETECTOR_TYPE dettype = IAS_NOMINAL_DETECTOR;

    /* Determine the center line/sample value.  */
    center_ls.line = nlines / 2.0;
    center_ls.samp = nsamps / 2.0;

    /* Determine the lat/long for the center point.  */
    if (ias_los_model_input_line_samp_to_geodetic(center_ls.line,
                center_ls.samp, band_index, sca_index, elev, model, dettype,
                NULL, &center_lat_long.lat, &center_lat_long.lng) != SUCCESS)
    {
        IAS_LOG_ERROR("In the ias_los_model_input_line_samp_to_geodetic call");
        *valid = FALSE;
        return ERROR;
    }

    /* Convert the latitude and longitude from Radian to Degrees.  */
    center_lat_long.lat *= ias_math_get_degrees_per_radian();
    center_lat_long.lng *= ias_math_get_degrees_per_radian();

    /* Compute the scene zone based on the scene center longitude.  Verify
     * that the zone falls between 1 and 60.  */
    *computed_zone = (int) ((center_lat_long.lng + 180.0) / 6.0) + 1;
    if (*computed_zone < 1)
        *computed_zone += 60;
    if (*computed_zone > 60)
        zone -= 60;

    /* Compare the input zone to the computed zone.  Input zones that fall
     * within one zone from the computed zone will be valid.  If the user-
     * entered zone is 1 and the computed zone is 60 (or vice versa), then
     * this should be a valid user-entered zone since they are only one
     * zone apart.  */
    if ((zone == 1 && *computed_zone == 60) ||
        (zone == 60 && *computed_zone == 1))
        *valid = TRUE;
    else if (zone - *computed_zone > 1)
        *valid = FALSE;  
    else
        *valid = TRUE;

    return SUCCESS;
}
