/**************************************************************************
 NAME:          create_los_model (main)

 PURPOSE:       Initializes a model structure from application parameters,
                values known from the satellite attributes, and values
                retrieved from the CPF, L0R, and ancillary data file.

 RETURNS:       Status code of EXIT_SUCCESS or EXIT_FAILURE
***************************************************************************/
#include <stdlib.h>
#include <string.h>
#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "ias_ancillary_io.h"
#include "ias_los_model.h"
#include "ias_model_io.h"
#include "ias_miscellaneous.h"
#include "ias_cpf.h"
#include "create_los_model.h"

void copy_params_to_model(MODEL_APP_PARAMS *app_params, IAS_LOS_MODEL *model);
void free_l0r_data_dyn_mem(IAS_LOS_MODEL_L0R_DATA *l0r_data);
int set_cpf_and_l0r_data_in_los_model(int ssm_encoder_value, IAS_CPF *cpf,
        IAS_LOS_MODEL_L0R_DATA *l0r_data, const char *ssm_position_filename,
        IAS_SSM_POSITION_QUALITY *ssm_position_quality, IAS_LOS_MODEL *model);
void log_app_params(MODEL_APP_PARAMS *app_params);

int main(int argc, char *argv[])
{
    MODEL_APP_PARAMS app_params;            /* Parameter file and OMF values */
    IAS_LOS_MODEL *model;                   /* The main model structure */
    IAS_ANC_EPHEMERIS_DATA *ephemeris_data; /* Ancillary ephemeris pointer */
    IAS_ANC_ATTITUDE_DATA *attitude_data;   /* Ancillary attitude pointer */
    IAS_CPF *cpf;                           /* CPF structure pointer */
    IAS_LOS_MODEL_L0R_DATA l0r_data;        /* Struct to compile L0R data */
    int status;                             /* General routine return status */
    const char *sw_version_ptr;             /* Pointer to SW version string */
    const char *tirs_ssm_position_status;   /* Pointer to TIRS SSM position */
    IAS_SSM_POSITION_QUALITY ssm_position_quality;/* TIRS SSM estimated position
                                                    quality */

    /* Prepare log file */
    status = ias_log_initialize("create_los_model");
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing logging library");
        return EXIT_FAILURE;
    }

    /* Verify the command line includes the application parameter file */
    if (argc != 2)
    {
        IAS_LOG_ERROR("USAGE:  create_los_model <parameter file name>");
        return EXIT_FAILURE;
    }

    /* Store the application parameter file name in the app_params */
    strcpy(app_params.parameter_filename, argv[1]);

    /* Read the parameters from the input parameter files (app param and OMF) */
    status = read_parameters(&app_params);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading input parameters");
        return EXIT_FAILURE;
    }

    /* Initialize the satellite attributes */
    status = ias_sat_attr_initialize(app_params.satellite_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attributes");
        return EXIT_FAILURE;
    }

    IAS_LOG_INFO("Create LOS Model started");

    /* Only does something if log level set to IAS_LOG_LEVEL_DEBUG */
    if (IAS_LOG_DEBUG_ENABLED())
        log_app_params(&app_params);

    /* Initialize the model structure */
    model = ias_los_model_initialize(app_params.acquisition_type);
    if (!model)
    {
        IAS_LOG_ERROR("Initializing model");
        return EXIT_FAILURE;
    }

    /* Copy the values from the application parameters that go directly into
       the model as-is. */
    copy_params_to_model(&app_params, model);

    /* Allocate the CPF data structure and read it */
    cpf = ias_cpf_read(app_params.cpf_filename);
    if (cpf == NULL)
    {
        IAS_LOG_ERROR("Opening CPF %s", app_params.cpf_filename);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }

    /* Initialize the IAS_LOS_MODEL_L0R_DATA structure to have NULLs and
       zeros to compare against later on. */
    memset(&l0r_data, 0, sizeof(l0r_data));

    /* Read the L0R */
    status = read_l0r(&app_params, &l0r_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading L0R %s", app_params.l0r_filename);
        free_l0r_data_dyn_mem(&l0r_data);
        ias_cpf_free(cpf);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }

    /* Set the L0R and CPF data into the model. Error check this after
       releasing memory resources that will be released whether this
       routine was successful or not. */
    status = set_cpf_and_l0r_data_in_los_model(app_params.ssm_encoder_value, 
        cpf, &l0r_data, app_params.ssm_position_filename, &ssm_position_quality,
        model);

    /* Free IAS_LOS_MODEL_L0R_DATA memory allocated dynamically by read_l0r */
    free_l0r_data_dyn_mem(&l0r_data);

    /* Check the status of setting the L0R and CPF data into the model */
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Setting CPF and L0R data into model structure");
        ias_cpf_free(cpf);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }

    /* Read the Ancillary Data File */
    status = ias_ancillary_read(app_params.anc_filename, &attitude_data,
            &ephemeris_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading ancillary data file %s",
                app_params.anc_filename);
        ias_cpf_free(cpf);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }

    /* Set the attitude and ephemeris values from the ancillary data file
       into the model structure. Error check this after releasing memory
       resources that will be released whether this routine was successful
       or not. */
    status = ias_sc_model_set_ancillary(ephemeris_data, attitude_data,
            &model->spacecraft);

    /* Clean up resources that are no longer needed */
    free(ephemeris_data);
    free(attitude_data);
    ephemeris_data = NULL;
    attitude_data = NULL;

    /* Check the status of setting the ancillary data into the model */
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Setting Ancillary data into model");
        ias_cpf_free(cpf);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }

    /* Setup the jitter model table by separating the attitude data into
       low and high frequency, based on the cut-off in the CPF. Error check
       this after releasing memory resources that will be released whether
       this routine was successful or not. */
    status = ias_los_model_build_jitter_table(cpf, model);

    /* Release the CPF resources, done with that now */
    ias_cpf_free(cpf);
    cpf = NULL;

    /* Check the status of building the jitter table */
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Building jitter model table");
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }

    /* Set the correction type of the model to systematic, which is what
       the Model Init application creates. */
    model->correction_type = IAS_SYSTEMATIC;

    /* Set the software version that created this model */
    sw_version_ptr = ias_misc_get_software_version();
    strcpy(model->sw_version, sw_version_ptr);

    /*== DEBUG LOGGING ======================================================*/
    IAS_LOG_DEBUG("====> SYSTEMATIC MODEL VALUES REPORT <====");
    IAS_LOG_DEBUG("Model correction type: %d", model->correction_type);
    IAS_LOG_DEBUG("Model created with IAS software version %s",
            model->sw_version);
    /*== END DEBUG ==========================================================*/

    /* Apply the precision parameters, if provided, to the model */
    if (app_params.precision_flag)
    {
        status = ias_los_model_apply_precision_params(
                &app_params.precision_model, model);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Applying precision parameters to model");
            ias_los_model_free(model);
            return EXIT_FAILURE;
        }
    }

    /* Write the model to disk */
    status = ias_model_write(app_params.model_filename, model);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing model to disk as file %s",
                app_params.model_filename);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }

    /* Update the database with the characterization data, if requested */
    if (app_params.characterization_flag)
    {
        status = write_characterization_data(model, &app_params,
                &l0r_data.l0r_sensor_data);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Writing to characterization database");
            ias_los_model_free(model);
            return EXIT_FAILURE;
        }
    }

    /* Convert the TIRS ssm position status to string for output to the 
       omf file */
    if (model->sensor.sensors[IAS_TIRS].sensor_present)
    {
        if (ias_sensor_convert_ssm_position_status_to_string(
            model->sensor.sensors[IAS_TIRS].ssm_model->position_status,
            &tirs_ssm_position_status) != SUCCESS)
        {
            IAS_LOG_ERROR("Converting TIRS SSM position status to string");
            ias_los_model_free(model);
            return EXIT_FAILURE;
        }
    }
    else
    {
        tirs_ssm_position_status = "";
    }

    /* Clean up resources that are no longer needed */
    ias_los_model_free(model);

    /* Update the OMF with the model filename */
    status = update_omf(&app_params, ssm_position_quality,
        tirs_ssm_position_status);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing model filename to the OMF %s",
                app_params.omf_filename);
        return EXIT_FAILURE;
    }

    IAS_LOG_INFO("Create LOS Model finished");

    return EXIT_SUCCESS;
}

/*************************************************************************
Name: copy_params_to_model

Purpose: Copies values from the given application parameters structure to
    the model structure for values that are directly applicable to the model.

Returns:
    nothing

Notes:
    If precision parameters not provided, they will be zeros in the model
    from the initialization of the model structure in the library. At the
    stage the precision parameters are then applied, if they are zeros, no
    corrections are applied and the precision values will match the
    non-precision attitude and ephemeris values in the model.
**************************************************************************/
void copy_params_to_model
(
    MODEL_APP_PARAMS *app_params,   /* I: application parameters */
    IAS_LOS_MODEL *model            /* O: parameters populated into model */
)
{
    int i;

    /* Some basic values from the application parameter file */
    model->acquisition_type = app_params->acquisition_type;

    if (app_params->precision_flag)
    {
        /* Input precision parameters */
        model->spacecraft.precision.seconds_from_image_epoch
            = app_params->precision_model.seconds_from_image_epoch;
        model->spacecraft.precision.ephemeris_order
            = app_params->precision_model.ephemeris_order;
        model->spacecraft.precision.attitude_order
            = app_params->precision_model.attitude_order;
        for (i = 0; i < IAS_PRECISION_MAX_POLY_COEFF; i++)
        {
            model->spacecraft.precision.x_corr[i]
                = app_params->precision_model.x_corr[i];
            model->spacecraft.precision.y_corr[i]
                = app_params->precision_model.y_corr[i];
            model->spacecraft.precision.z_corr[i]
                = app_params->precision_model.z_corr[i];
            model->spacecraft.precision.roll_corr[i]
                = app_params->precision_model.roll_corr[i];
            model->spacecraft.precision.pitch_corr[i]
                = app_params->precision_model.pitch_corr[i];
            model->spacecraft.precision.yaw_corr[i]
                = app_params->precision_model.yaw_corr[i];
        }
    }
}

/*************************************************************************
Name: free_l0r_data_dyn_mem

Purpose: Frees up the dynamically allocated memory within the
    IAS_LOS_MODEL_L0R_DATA structure.

Returns:
    Nothing

Notes:
**************************************************************************/
void free_l0r_data_dyn_mem
(
    IAS_LOS_MODEL_L0R_DATA *l0r_data    /* the struct to free dynamic mem */
)
{
    int i;      /* general looping counter */

    /* Free IAS_LOS_MODEL_L0R_DATA memory allocated dynamically by read_l0r */
    free(l0r_data->l0r_sensor_data.oli_data.frame_headers);
    l0r_data->l0r_sensor_data.oli_data.frame_headers = NULL;

    free(l0r_data->l0r_sensor_data.tirs_data.frame_headers);
    l0r_data->l0r_sensor_data.tirs_data.frame_headers = NULL;

    free(l0r_data->l0r_sensor_data.tirs_data.ssm_records);
    l0r_data->l0r_sensor_data.tirs_data.ssm_records = NULL;

    /* Since we may have data for one or more sensors, and we don't
       want to get complicated about what band indeces the data could
       have been allocated at, just loop over all possible bands, and
       free what appears to have been allocated */
    if (l0r_data->l0r_sensor_data.l0r_detector_offsets)
    {
        for(i = 0; i < IAS_MAX_NBANDS; i++)
        {
            free(l0r_data->l0r_sensor_data.l0r_detector_offsets[i]);
            l0r_data->l0r_sensor_data.l0r_detector_offsets[i] = NULL;
        }

        free(l0r_data->l0r_sensor_data.l0r_detector_offsets);
        l0r_data->l0r_sensor_data.l0r_detector_offsets = NULL;
    }
}

/*************************************************************************
Name: set_cpf_and_l0r_data_in_los_model

Purpose: Sets model components from the L0R, then sets the sensor model
    components and earth characteristics from the CPF. Made this function
    to wrap up these function calls to ensure they are executed in the
    proper order since there are assumptions and dependencies on this
    calling order.

Returns:
    nothing

Notes:
    Meant to isolate this set of function calls, in order to ensure the
    dependencies that exist are maintained and understood (these comments).
    The following functions are used in conjunction with each other to be
    called in the following order:
    -> ias_los_model_set_cpf()
    -> ias_los_model_set_l0r()
    Dependencies:
    ias_los_model_set_cpf() sets CPF info into the model needed by
    ias_los_model_set_l0r().
**************************************************************************/
int set_cpf_and_l0r_data_in_los_model
(
    int ssm_encoder_value,              /* I: Replacement SSM encoder value */
    IAS_CPF *cpf,                       /* I: CPF structure pointer */
    IAS_LOS_MODEL_L0R_DATA *l0r_data,   /* I: Struct to compile L0R data */
    const char *ssm_position_filename,  /* I: Estimated SSM Position Filename */
    IAS_SSM_POSITION_QUALITY *ssm_position_quality,/* O: SSM positions quality 
                                                         flag */
    IAS_LOS_MODEL *model                /* O: Parameters populated into model */
)
{
    int status;

    /* Set the sensor model information that comes from the CPF */
    status = ias_los_model_set_cpf(cpf, &l0r_data->l0r_sensor_data, model);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Setting CPF information into model");
        return ERROR;
    }

    /* Set the image values from the L0R into the model structure */
    status = ias_los_model_set_l0r(ssm_encoder_value, cpf, l0r_data, 
        ssm_position_filename, ssm_position_quality, model);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Setting L0R data into model");
        return ERROR;
    }

    return SUCCESS;
}

/*************************************************************************
Name: log_app_params

Purpose: Prints debugging log data for the application parameter values.

Returns:
    nothing

Notes:
**************************************************************************/
void log_app_params
(
    MODEL_APP_PARAMS *app_params    /* I: Parameter file and OMF values */
)
{
    /*== DEBUG LOGGING ======================================================*/
    IAS_LOG_DEBUG("====> APPLICATION PARAMETERS REPORT <====");
    /* Application Parameter file */
    IAS_LOG_DEBUG("Work order ID: %s", app_params->work_order_id);
    IAS_LOG_DEBUG("Characterization flag: %d",
            app_params->characterization_flag);
    IAS_LOG_DEBUG("Output model file name: %s",
            app_params->model_filename);
    IAS_LOG_DEBUG("Precision reference time: %lf",
            app_params->precision_model.seconds_from_image_epoch);
    IAS_LOG_DEBUG("Precision ephemeris order: %d",
            app_params->precision_model.ephemeris_order);
    IAS_LOG_DEBUG("Precision X corrections: %lf, %lf",
            app_params->precision_model.x_corr[0],
            app_params->precision_model.x_corr[1]);
    IAS_LOG_DEBUG("Precision Y corrections: %lf, %lf",
            app_params->precision_model.y_corr[0],
            app_params->precision_model.y_corr[1]);
    IAS_LOG_DEBUG("Precision Z corrections: %lf, %lf",
            app_params->precision_model.z_corr[0],
            app_params->precision_model.z_corr[1]);
    IAS_LOG_DEBUG("Precision attitude order: %d",
            app_params->precision_model.attitude_order);
    IAS_LOG_DEBUG("Precision roll corrections: %lf, %lf",
            app_params->precision_model.roll_corr[0],
            app_params->precision_model.roll_corr[1]);
    IAS_LOG_DEBUG("Precision pitch corrections: %lf, %lf",
            app_params->precision_model.pitch_corr[0],
            app_params->precision_model.pitch_corr[1]);
    IAS_LOG_DEBUG("Precision yaw corrections: %lf, %lf",
            app_params->precision_model.yaw_corr[0],
            app_params->precision_model.yaw_corr[1]);

    /* OMF */
    IAS_LOG_DEBUG("CPF file: %s", app_params->cpf_filename);
    IAS_LOG_DEBUG("Estimated SSM Position file: %s", 
        app_params->ssm_position_filename);
    IAS_LOG_DEBUG("L0R base file name: %s", app_params->l0r_filename);
    IAS_LOG_DEBUG("Ancillary data file: %s", app_params->anc_filename);
    IAS_LOG_DEBUG("Acquisition type: %d", app_params->acquisition_type);
    /*== END DEBUG ==========================================================*/
}
