/**************************************************************************
 NAME:                      read_l0r

 PURPOSE:       Reads the values needed from the L0R into memory

 RETURNS:       Status code of SUCCESS or ERROR

 NOTES: Allocated memory is free'd at higher level if there is an error
        Assumes satellite attributes have been initialized
***************************************************************************/
#include <stdlib.h>
#include <string.h>
#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "ias_l0r.h"
#include "create_los_model.h"

int read_l0r
(
    const MODEL_APP_PARAMS *app_params, /* I: Parameter file and OMF values */
    IAS_LOS_MODEL_L0R_DATA *l0r_data    /* O: Compiled L0R data for model */
)
{
    L0RIO *l0r_ptr;                         /* L0R file access pointer */
    int num_frames;                         /* Number of image frames */
    uint16_t *l0r_detector_offsets;         /* Detector shifts from L0R */
    int total_detectors;                    /* Total number dets across SCA */
    int status;                             /* General routine status */
    int band_count;                         /* Gotten from sat attr lib */
    int band_index;                         /* Band array index for looping */
    int sensor_index;                       /* Sensor looping */
    int sca_count;                          /* Number of SCAs for band */
    int num_dets_per_sca;                   /* Number detectors/SCA for band */
    int band_num;                           /* Satellite attribs band number */
    IAS_L0R_INTERVAL_METADATA int_meta;     /* Interval metadata for bands */
    IAS_SENSOR_L0R_DATA *l0r_sensor_data = &l0r_data->l0r_sensor_data;
    L0R_FILE_TYPE l0r_type;                 /* Local enum type for L0R file */
    int tirs_start_frame = 0;               /* Local TIRS starting frame */
    int tirs_frame_count = 0;               /* TIRS override frame count */
    int oli_start_frame = 0;                /* Local OLI starting frame */
    int oli_frame_count = 0;                /* OLI overrid frame count */
    int pick_correct_scene = FALSE;         /* Flag that indicates the correct
                                               scene needs to be picked in the
                                               metadata */
    int have_scene_id = (strlen(app_params->landsat_scene_id) > 0);

    l0r_ptr = ias_l0r_open_from_filename(app_params->l0r_filename,
            IAS_L0R_COMPRESSION_ON);
    if (l0r_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening L0R file %s", app_params->l0r_filename);
        return ERROR;
    }

    /* Open read access to the L0R metadata */
    status = ias_l0r_open_metadata(l0r_ptr, IAS_READ);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Opening read access to L0R metadata in file %s",
                app_params->l0r_filename);
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Initialize L0R sensor and band flags */
    l0r_sensor_data->oli_data.sensor_present = 0;
    l0r_sensor_data->tirs_data.sensor_present = 0;

    /* Determine what sensors the L0R contains imagery for */
    status = ias_l0r_get_interval_metadata(l0r_ptr, &int_meta);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Opening read access to L0R interval metadata in file %s",
                app_params->l0r_filename);
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Set the sensor flags for sensors that collected data for this image */
    if(strcmp(int_meta.sensor_id, "OLI") == 0)
        l0r_sensor_data->oli_data.sensor_present = 1;
    else if (strcmp(int_meta.sensor_id, "TIRS") == 0)
        l0r_sensor_data->tirs_data.sensor_present = 1;
    else if (strcmp(int_meta.sensor_id, "OLI_TIRS") == 0)
    {
        l0r_sensor_data->oli_data.sensor_present = 1;
        l0r_sensor_data->tirs_data.sensor_present = 1;
    }
    else
    {
        /* Sanity check */
        IAS_LOG_ERROR("Source L0R %s does not contain OLI or TIRS data",
                app_params->l0r_filename);
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Determine if the L0R is an interval (L0Ra) or scene (L0Rp) file */
    if (strcmp(int_meta.data_type, "OLI_L0RA") == 0
            || strcmp(int_meta.data_type, "TIRS_L0RA") == 0
            || strcmp(int_meta.data_type, "OLI_TIRS_L0RA") == 0)
    {
        l0r_type = L0RA;
    }
    else if (strcmp(int_meta.data_type, "OLI_L0RP") == 0
            || strcmp(int_meta.data_type, "TIRS_L0RP") == 0
            || strcmp(int_meta.data_type, "OLI_TIRS_L0RP") == 0)
    {
        l0r_type = L0RP;
    }
    else
    {
        IAS_LOG_ERROR("Unable to determine source L0R file type from %s",
                int_meta.data_type);
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* If a full L0RA is present and a scene id has not been specified 
       get the path and row from the interval metadata */
    if (!have_scene_id && l0r_type == L0RA)
    {
        /* For L0Ra products, get the WRS path/row from the interval metadata */
        l0r_data->wrs_path = int_meta.wrs_starting_path;
        l0r_data->wrs_row = int_meta.wrs_starting_row;
    }
    else
    {
        IAS_L0R_SCENE_METADATA *scene_metadata; /* Scene metadata for bands */
        int record_count;                  /* L0Rp record count */
        int index;                         /* Generic loop index */
        int tirs_start_index = 0;
        int oli_start_index = 0;

        /* Retrieve the L0R scene record count */
        if (ias_l0r_get_scene_metadata_records_count(l0r_ptr, 
            &record_count) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading L0R scene count");
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }

        /* Allocate memory for the scene metadata */
        scene_metadata = malloc(record_count * sizeof(*scene_metadata));
        if (!scene_metadata)
        {
            IAS_LOG_ERROR("Allocating memory for %d records of scene "
                "metadata", record_count);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }

        /* Read the metadata */
        if (ias_l0r_get_scene_metadata(l0r_ptr, 0, record_count,
            scene_metadata) != SUCCESS)
        { 
            IAS_LOG_ERROR("Retrieving %d records from the L0R scene metadata",
                record_count);
            ias_l0r_close(l0r_ptr);
            free(scene_metadata);
            return ERROR;
        }

        /* If more than 1 scene is present and not processing the full 
           interval override the frames to pull out the single scene
           from the interval */
        if (record_count > 1 && !app_params->use_three_tirs_scenes_flag)
        {
            pick_correct_scene = TRUE;

            /* Determine the TIRS and OLI index starts */
            for (oli_start_index = 0; oli_start_index < record_count; 
                 oli_start_index++)
            {
                if (scene_metadata[oli_start_index].sensor_present_oli)
                    break;
            }
            for (tirs_start_index = 0; tirs_start_index < record_count; 
                 tirs_start_index++)
            {
                if (scene_metadata[tirs_start_index].sensor_present_tirs)
                    break;           
            }
        }
        
        
        /* For a L0Rp product, retrieve the WRS path/row. If the scene id is
           not specified use the first scene in the metadata  */
        for (index = 0; index < record_count; index++)
        {
            IAS_L0R_SCENE_METADATA *cur_scene = &scene_metadata[index];

            /* Check for the correct record to use */
            if (!have_scene_id || strcmp(app_params->landsat_scene_id, 
                cur_scene->landsat_scene_id) == 0)
            {
                /* Scene id has been found or no scene id was specified so using
                   the first record in the metadata */
                l0r_data->wrs_path = cur_scene->target_wrs_path;
                l0r_data->wrs_row = cur_scene->target_wrs_row;

                /* And check the 'sensor_present_xxx' flags as well for which
                   sensor(s) actually imaged data for the collect.  This can
                   override the settings for an OLI_TIRS sensor ID established
                   earlier in cases where an OLI-only or TIRS-only partial scene
                   has been acquired. */
                if (cur_scene->sensor_present_oli)
                {
                    l0r_sensor_data->oli_data.sensor_present = 1;

                    if (pick_correct_scene)
                    {
                        /* Extract the needed data from the scene metadata */
                        oli_start_frame = cur_scene->scene_start_frame_oli;
                        oli_frame_count = cur_scene->scene_stop_frame_oli 
                            - oli_start_frame + 1;

                        /* Make all the line numbers relative to the first */
                        oli_start_frame -= scene_metadata[
                            oli_start_index].scene_start_frame_oli;
                    }
                }
                else
                {
                    l0r_sensor_data->oli_data.sensor_present = 0;
                }

                if (cur_scene->sensor_present_tirs)
                {
                    l0r_sensor_data->tirs_data.sensor_present = 1;

                    if (pick_correct_scene)
                    {
                        /* Extract the needed data from the scene metadata */
                        tirs_start_frame = cur_scene->scene_start_frame_tirs;
                        tirs_frame_count = cur_scene->scene_stop_frame_tirs 
                            - tirs_start_frame + 1;

                        /* Make all the line numbers relative to the first */
                        tirs_start_frame -= scene_metadata[
                            tirs_start_index].scene_start_frame_tirs;
                    }
                }
                else
                {
                    l0r_sensor_data->tirs_data.sensor_present = 0;
                }
                
                /* No need to check the rest of the records */
                break;
            }
        }  
        free(scene_metadata);
        scene_metadata = NULL;

        /* Check if the metadata was found */
        if (index >= record_count)
        {
            if (have_scene_id)
            {
                IAS_LOG_ERROR("Scene metadata for scene ID %s not found", 
                    app_params->landsat_scene_id);
            }
            else
            {
                IAS_LOG_ERROR("No scene metadata was found");
            }
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }
    }

    /* Retrieve the flags for each band being present or not in the image */
    band_count = ias_sat_attr_get_normal_band_count();
    if (band_count == ERROR)
    {
        IAS_LOG_ERROR("Retrieving the band count from satellite attributes");
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }
    for (band_index = 0; band_index < band_count; band_index++)
    {
        int band_present;
        band_num = ias_sat_attr_convert_band_index_to_number(band_index);
        if (band_num == ERROR) /* set to ERROR if conversion fails */
        {
            IAS_LOG_ERROR("Using satellite attributes to convert band index %d "
                    "to band number", band_index);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }
        status = ias_l0r_get_band_present(l0r_ptr, band_num, &band_present);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Retrieving band present information from L0R "
                    "file %s", app_params->l0r_filename);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }
        if (band_present)
            l0r_sensor_data->bands_present[band_index] = 1;
        else
            l0r_sensor_data->bands_present[band_index] = 0;
    }

    /* Get frame header information */
    status = ias_l0r_open_header(l0r_ptr, IAS_READ);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Opening read access to L0R headers in file %s",
                app_params->l0r_filename);
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Initialize the image header flag */
    l0r_sensor_data->image_header_flag = 0;

    if (l0r_sensor_data->oli_data.sensor_present)
    {
        /* See if the OLI image header is present */
        status = ias_l0r_is_oli_image_header_present(l0r_ptr, &num_frames);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Could not determine if OLI image header is "
                "present in file %s", app_params->l0r_filename);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }
        if (num_frames <= 0)
        {
            IAS_LOG_WARNING("OLI image header not available from L0R file %s",
                app_params->l0r_filename);
        }
        else
        {
            /* Read the image header and set the flag indicating the image 
               header was available */
            status = ias_l0r_get_oli_image_header(l0r_ptr,
                    &l0r_sensor_data->image_header);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Could not get OLI image header from L0R file %s",
                    app_params->l0r_filename);
                ias_l0r_close(l0r_ptr);
                return ERROR;
            }
            else
                l0r_sensor_data->image_header_flag = 1;
        }

        /* Get number of timestamps to allocate frame headers structure (there 
           is one timestamp per frame) */
        status = ias_l0r_get_oli_frame_headers_records_count(l0r_ptr,
                &num_frames);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Retrieving number of OLI frames from L0R file %s",
                    app_params->l0r_filename);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }

        l0r_sensor_data->oli_data.num_frames = num_frames;
        if (pick_correct_scene)
        {
            if (oli_frame_count > num_frames)
            {
                IAS_LOG_ERROR("Number of calculated OLI frames (%d) is greater "
                    "than the number of frames present in the header file",
                    oli_frame_count);
                ias_l0r_close(l0r_ptr);
                return ERROR;
            }
            l0r_sensor_data->oli_data.num_frames = oli_frame_count;
        }
    }
    if (l0r_sensor_data->tirs_data.sensor_present)
    {
        /* read the TIRS frame header record count */
        status = ias_l0r_get_tirs_frame_headers_records_count(l0r_ptr,
                &num_frames);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Retrieving number of TIRS frames from L0R file %s",
                    app_params->l0r_filename);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }
        
        l0r_sensor_data->tirs_data.num_frames = num_frames;
        if (pick_correct_scene)
        {                
            if (tirs_frame_count > num_frames)
            {
                IAS_LOG_ERROR("Number of calculated TIRS frames (%d) is "
                    "greater than the number of frames present in the header "
                    "file", tirs_frame_count);
                ias_l0r_close(l0r_ptr);
                return ERROR;
            }
            l0r_sensor_data->tirs_data.num_frames = tirs_frame_count;
        }
    }

    /*== DEBUG LOGGING ======================================================*/
    IAS_LOG_DEBUG("====> READ L0R REPORT <====");
    IAS_LOG_DEBUG("WRS path: %d", l0r_data->wrs_path);
    IAS_LOG_DEBUG("WRS row: %d", l0r_data->wrs_row);
    IAS_LOG_DEBUG("Number frames in OLI bands: %d",
            l0r_sensor_data->oli_data.num_frames);
    IAS_LOG_DEBUG("Starting frame for OLI bands: %d",
            oli_start_frame);
    IAS_LOG_DEBUG("Number frames in TIRS bands: %d",
            l0r_sensor_data->tirs_data.num_frames);
    IAS_LOG_DEBUG("Starting frame for TIRS bands: %d",
            tirs_start_frame);
    /*== END DEBUG ==========================================================*/

    if (l0r_sensor_data->oli_data.sensor_present)
    {
        /* Allocate the OLI frame headers data structures. Sized with number of
           timestamps because there is one timestamp per frame of OLI data. */
        l0r_sensor_data->oli_data.frame_headers
            = malloc(sizeof(*l0r_sensor_data->oli_data.frame_headers)
                    * l0r_sensor_data->oli_data.num_frames);
        if (l0r_sensor_data->oli_data.frame_headers == NULL)
        {
            IAS_LOG_ERROR("Allocating OLI frame headers structures");
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }

        /* Read only the frame header information that is needed */
        status = ias_l0r_get_oli_frame_headers(l0r_ptr, oli_start_frame,
                l0r_sensor_data->oli_data.num_frames,
                l0r_sensor_data->oli_data.frame_headers);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading OLI frame header records in L0R file %s",
                    app_params->l0r_filename);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }
    }

    if (l0r_sensor_data->tirs_data.sensor_present)
    {
        int ssm_count;

        /* Allocate the TIRS frame headers data structures. Sized with number of
           timestamps because there is one timestamp per frame of TIRS data. */
        l0r_sensor_data->tirs_data.frame_headers
            = malloc(sizeof(*l0r_sensor_data->tirs_data.frame_headers)
                    * l0r_sensor_data->tirs_data.num_frames);
        if (l0r_sensor_data->tirs_data.frame_headers == NULL)
        {
            IAS_LOG_ERROR("Allocating TIRS frame headers structures");
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }

        /* Read only the frame header information that is needed */
        status = ias_l0r_get_tirs_frame_headers(l0r_ptr, tirs_start_frame,
                l0r_sensor_data->tirs_data.num_frames,
                l0r_sensor_data->tirs_data.frame_headers);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading TIRS frame header records in L0R file %s",
                    app_params->l0r_filename);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }

        /* Open the L0R ancillary data for reading the TIRS scene select mirror
           (SSM) information */
        status = ias_l0r_open_ancillary(l0r_ptr, IAS_READ);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Opening ancillary file for L0R file %s",
                    app_params->l0r_filename);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }

        /* Get the count of the SSM records */
        status = ias_l0r_get_ancillary_tirs_ssm_records_count(l0r_ptr,
                    &ssm_count);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Opening ancillary file for L0R file %s",
                    app_params->l0r_filename);
            ias_l0r_close_ancillary(l0r_ptr);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }

        /* Allocate memory for the SSM records */
        l0r_sensor_data->tirs_data.num_ssm_records = ssm_count;
        l0r_sensor_data->tirs_data.ssm_records = malloc(ssm_count 
                * sizeof(*l0r_sensor_data->tirs_data.ssm_records));
        if (l0r_sensor_data->tirs_data.ssm_records == NULL)
        {
            IAS_LOG_ERROR("Allocating TIRS SSM array");
            ias_l0r_close_ancillary(l0r_ptr);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }

        /* Read the SSM records */
        status = ias_l0r_get_ancillary_tirs_ssm(l0r_ptr, 0, ssm_count,
                    l0r_sensor_data->tirs_data.ssm_records);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading TIRS SSM data from L0R file %s",
                    app_params->l0r_filename);
            ias_l0r_close_ancillary(l0r_ptr);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }

        /* Close the ancillary data without checking for errors since we only
           read the file and closing shouldn't cause important errors */
        ias_l0r_close_ancillary(l0r_ptr);
    }

    /* Allocate memory for the pointers to the band arrays of offsets */
    l0r_sensor_data->l0r_detector_offsets = malloc(sizeof(l0r_detector_offsets)
            * band_count);
    if (l0r_sensor_data->l0r_detector_offsets == NULL)
    {
        IAS_LOG_ERROR("Allocating L0R offsets array");
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Initialize the detector offsets pointers for all possible bands */
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        l0r_sensor_data->l0r_detector_offsets[band_index] = NULL;
    }

    /* Read and save the L0R offsets values per detector, per band */
    for (sensor_index = 0; sensor_index < IAS_MAX_SENSORS; sensor_index++)
    {
        int band_numbers[IAS_MAX_NBANDS];   /* Band numbers for sensor */
        int num_sensor_bands;               /* Number of band #s in array */
        int band_num_index;                 /* Index to array of band #s */

        if ((sensor_index == IAS_OLI
                    && l0r_sensor_data->oli_data.sensor_present)
                || (sensor_index == IAS_TIRS
                    && l0r_sensor_data->tirs_data.sensor_present))
        {
            status = ias_sat_attr_get_sensor_band_numbers(sensor_index,
                    IAS_NORMAL_BAND, 0, &band_numbers[0], IAS_MAX_NBANDS,
                    &num_sensor_bands);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Using satellite attributes to retrieve bands "
                        "for this sensor");
                ias_l0r_close(l0r_ptr);
                return ERROR;
            }

            for (band_num_index = 0; band_num_index < num_sensor_bands;
                    band_num_index++)
            {
                const IAS_BAND_ATTRIBUTES *band_attribs;

                /* Get band attributes for this band */
                band_attribs = ias_sat_attr_get_band_attributes(
                            band_numbers[band_num_index]);
                /* NULL ptr if band attribs not found */
                if (band_attribs == NULL)
                {
                    IAS_LOG_ERROR("Using satellite attributes to retrieve band "
                            "attributes for band number %d",
                            band_numbers[band_num_index]);
                    ias_l0r_close(l0r_ptr);
                    return ERROR;
                }
                band_index = band_attribs->band_index;

                /* Open access to the image band metadata */
                status = ias_l0r_open_band(l0r_ptr,
                        band_numbers[band_num_index], IAS_READ);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Opening band %d for reading from L0R "
                            "file %s", band_numbers[band_num_index],
                            app_params->l0r_filename);
                    ias_l0r_close(l0r_ptr);
                    return ERROR;
                }

                /* Get the number of SCAs for this band and the number of
                   detectors per SCA (it's the same for each SCA within the
                   same band) */
                sca_count = band_attribs->scas;
                num_dets_per_sca = band_attribs->detectors_per_sca;

                total_detectors = sca_count * num_dets_per_sca;
                l0r_detector_offsets = malloc(sizeof(*l0r_detector_offsets)
                        * total_detectors);
                if (l0r_detector_offsets == NULL)
                {
                    IAS_LOG_ERROR("Allocating L0R offsets array");
                    ias_l0r_close(l0r_ptr);
                    return ERROR;
                }

                /* Read the detector offsets from the L0R */
                status = ias_l0r_get_top_detector_offsets(l0r_ptr,
                        band_numbers[band_num_index], l0r_detector_offsets,
                        total_detectors);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Reading detector offsets in L0R file %s",
                            app_params->l0r_filename);
                    ias_l0r_close(l0r_ptr);
                    return ERROR;
                }

                l0r_sensor_data->l0r_detector_offsets[band_index]
                    = l0r_detector_offsets;
                l0r_detector_offsets = NULL; /* unassign pointer */
            }
        }
    }

    ias_l0r_close(l0r_ptr); /* closes headers, metadata, ancillary, & bands */

    return SUCCESS;
}
