/*----------------------------------------------------------------------------
NAME:    read_parameters

PURPOSE: Read the input application parameter file and the system OMF.
    Populate the app_params with parameters.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "ias_satellite_attributes.h"
#include "create_los_model.h"

int read_parameters
(
    MODEL_APP_PARAMS *app_params  /* I: parameter and OMF file names
                                     O: parameters listed in input files */
)
{
    int status;                 /* Routine return status */
    int count;                  /* Count of elements in the parameter definition
                                   tables. Also used for precision. */
    int i;                      /* Generic loop counter */
    double precision_dbl_default = 0.0; /* Default to use for precision parms */
    int satellite_number;       /* Satellite number being processed */
                                /* default model filename */
    const char *default_model_filename[] = {"systematic_model.h5"};
    char ssm_position_filename_override[PATH_MAX] = "";
    int default_use_three_tirs_scenes_flag[] = {0}; /* Default 3 scene flag */

    /* Make sure the precision parameters start at 0 and only get modified
       if provided. This way if not provided, when copying to the model
       structure, we are sure to set zeros for those potential values not
       specifically set. */
    memset(&app_params->precision_model, 0,
            sizeof(app_params->precision_model));

    /* --------------------------------------------------------------------- */
    /* -- Table definition for things from the application parameter file -- */

    IAS_PARM_DECLARE_TABLE(parms, 14);

    /* Add the work order ID parameter */
    IAS_PARM_ADD_STRING(parms, WORK_ORDER_ID, "Work order ID",
            IAS_PARM_REQUIRED,
            /* no limits */
            0, NULL,
            /* no default */
            0, NULL,
            app_params->work_order_id, sizeof(app_params->work_order_id), 1);

    /* Add the characterization flag parameter */
    IAS_PARM_STORE_CHARACTERIZATION_IN_DB(parms, IAS_PARM_OPTIONAL,
            &app_params->characterization_flag,
            sizeof(app_params->characterization_flag), 1);

    /* Add the output model file name parameter */
    IAS_PARM_ADD_STRING(parms, MODEL_FILENAME, "Output model file name",
            IAS_PARM_OPTIONAL,
            /* no limits */
            0, NULL,
            /* provide a default filename */
            1, default_model_filename,
            app_params->model_filename, sizeof(app_params->model_filename),
            1);

    /* Add the precision reference time parameter */
    IAS_PARM_ADD_DOUBLE(parms, PRECISION_REF_TIME, "Precision reference time",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
            /* no limits */
            0, 0, 0,
            /* has one default value */
            1, &precision_dbl_default,
            &app_params->precision_model.seconds_from_image_epoch,
            sizeof(app_params->precision_model.seconds_from_image_epoch), 1);

    /* Add the precision x corrections parameter */
    IAS_PARM_ADD_DOUBLE(parms, PRECISION_X_CORR_PARAMS, "Precision x "
            "correction parameters",
            IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
            /* no limits */
            0, 0, 0,
            /* no default */
            0, NULL,
            app_params->precision_model.x_corr,
            sizeof(app_params->precision_model.x_corr), 0);

    /* Add the precision y corrections parameter */
    IAS_PARM_ADD_DOUBLE(parms, PRECISION_Y_CORR_PARAMS, "Precision y "
            "correction parameters",
            IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
            /* no limits */
            0, 0, 0,
            /* no default */
            0, NULL,
            app_params->precision_model.y_corr,
            sizeof(app_params->precision_model.y_corr), 0);

    /* Add the precision z corrections parameter */
    IAS_PARM_ADD_DOUBLE(parms, PRECISION_Z_CORR_PARAMS, "Precision z "
            "correction parameters",
            IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
            /* no limits */
            0, 0, 0,
            /* no default */
            0, NULL,
            app_params->precision_model.z_corr,
            sizeof(app_params->precision_model.z_corr), 0);

    /* Add the precision roll corrections parameter */
    IAS_PARM_ADD_DOUBLE(parms, PRECISION_ROLL_CORR_PARAMS, "Precision roll "
            "correction parameters",
            IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
            /* no limits */
            0, 0, 0,
            /* no default */
            0, NULL,
            app_params->precision_model.roll_corr,
            sizeof(app_params->precision_model.roll_corr), 0);

    /* Add the precision pitch corrections parameter */
    IAS_PARM_ADD_DOUBLE(parms, PRECISION_PITCH_CORR_PARAMS, "Precision pitch "
            "correction parameters",
            IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
            /* no limits */
            0, 0, 0,
            /* no default */
            0, NULL,
            app_params->precision_model.pitch_corr,
            sizeof(app_params->precision_model.pitch_corr), 0);

    /* Add the precision yaw corrections parameter */
    IAS_PARM_ADD_DOUBLE(parms, PRECISION_YAW_CORR_PARAMS, "Precision yaw "
            "correction parameters",
            IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
            /* no limits */
            0, 0, 0,
            /* no default */
            0, NULL,
            app_params->precision_model.yaw_corr,
            sizeof(app_params->precision_model.yaw_corr), 0);

    /* Add the SSM user defined encoder value */
    const int default_encoder_value[] = {0}; 
    IAS_PARM_ADD_INT(parms, SSM_ENCODER_VALUE, "User defined SSM encoder value "
            "used to replace the existing encoder values (0 indicates the "
            "encoder value will not be replaced)",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
            /* no limits */
            0, 0, 0,
            /* no default */
            1, default_encoder_value,
            &app_params->ssm_encoder_value,
            sizeof(app_params->ssm_encoder_value), 0);

    /* Add the processing pass # parameter */
    IAS_PARM_PROCESSING_PASS(parms, &app_params->processing_pass,
            sizeof(app_params->processing_pass), 0);

    /* Add the TIRS ESTIMATED SSM FILENAME override parameter */
    IAS_PARM_ADD_STRING(parms, ESTIMATED_SSM_POSITION_FILENAME,
            "Parameter provided for overriding the TIRS SSM estimated "
            "position file",
            IAS_PARM_OPTIONAL,
            /* no limits */
            0, NULL,
            /* no default */
            0, NULL,
            ssm_position_filename_override,
            sizeof(ssm_position_filename_override), 1);

    /* Flag to indicate that a TIRS subinterval might be passed in */
    IAS_PARM_USE_THREE_TIRS_SCENES_FLAG(parms, IAS_PARM_OPTIONAL,
            default_use_three_tirs_scenes_flag,
            &app_params->use_three_tirs_scenes_flag,
            sizeof(app_params->use_three_tirs_scenes_flag), 1);

    /* --------------------------------------------------------------------- */
    /* -------------- Table definition for things from the OMF ------------- */

    IAS_PARM_DECLARE_TABLE(omf_parms, 7);

    /* Add the CPF parameter */
    IAS_PARM_CPF(omf_parms, IAS_PARM_REQUIRED, app_params->cpf_filename,
            sizeof(app_params->cpf_filename), 1);

    /* Add the L0R file name parameter */
    IAS_PARM_L0R_FILENAME(omf_parms, IAS_PARM_REQUIRED,
            app_params->l0r_filename, sizeof(app_params->l0r_filename), 1);

    /* Add the TIRS ESTIMATED SSM FILENAME parameter which will be used
       if an override was not provided in the overall odl file */
    IAS_PARM_ESTIMATED_SSM_POSITION_FILENAME(omf_parms,
            app_params->ssm_position_filename,
            sizeof(app_params->ssm_position_filename), 1);

    /* Add the ancillary metadata file name parameter */
    IAS_PARM_ADD_STRING(omf_parms, ANCILLARY_FILENAME,
            "Ancillary data preprocessor metadata file name",
            IAS_PARM_REQUIRED,
            /* no limits */
            0, NULL,
            /* no default */
            0, NULL,
            app_params->anc_filename, sizeof(app_params->anc_filename), 1);

    /* Add the acquisition type parameter */
    int acq_default_value = 0;
    IAS_PARM_ADD_INT(omf_parms, ACQUISITION_TYPE, "Acquisition type with 0 "
            "indicating earth, 1 indicating lunar, and 2 indicating stellar",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
            /* has a limited range of 0-2 */
            1, 0, 2,
            /* default is 0 for an earth acquisition if not specified */
            1, &acq_default_value,
            &app_params->acquisition_type,
            sizeof(app_params->acquisition_type), 1);

    /* Add the satellite number parameter */
    IAS_PARM_SATELLITE(omf_parms, &satellite_number, 
            sizeof(satellite_number), 1);

    /* Add read for the LANDSAT_SCENE_ID */
    strcpy(app_params->landsat_scene_id, "");
    IAS_PARM_LANDSAT_SCENE_ID(omf_parms, IAS_PARM_OPTIONAL,
            app_params->landsat_scene_id, sizeof(app_params->landsat_scene_id),
            1);
    /* --------------------------------------------------------------------- */
    /* -------------- Read the contents defined from each file ------------- */

    /* See if the user wanted help or a template */
    count = IAS_PARM_GET_TABLE_SIZE(parms);
    status = ias_parm_provide_help(app_params->parameter_filename, parms, count,
            IAS_INPUT_PARAMETERS);
    if (status)
    {
        /* either help was provided or an error occurred */
        if (status == 1)
        {
            /* If the status of the provide help is 1 the user asked for help.
               Since the parameter file name contains the help request, copy
               it to the OMF file name so the next call will ask for the same
               help */
            strcpy(app_params->omf_filename, app_params->parameter_filename);
        }
        else
        {
            IAS_LOG_ERROR("Providing application parameter file help");
            return ERROR;
        }
    }
    else
    {
        /* Read the parameter file */
        status = ias_parm_read(app_params->parameter_filename, parms, count);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading parameters from: %s",
                    app_params->parameter_filename);
            return ERROR;
        }

        /* Build the OMF file name */
        sprintf(app_params->omf_filename, "%s.omf", app_params->work_order_id);
    }

    /* See if the user wanted help or a template this time with the OMF file */
    count = IAS_PARM_GET_TABLE_SIZE(omf_parms);
    status = ias_parm_provide_help(app_params->omf_filename, omf_parms, count,
            IAS_OMF_PARAMETERS);
    if (status)
    {
        if (status == 1) /* help was successful */
            exit(EXIT_SUCCESS);
        else
        {
            IAS_LOG_ERROR("Providing OMF file help");
            return ERROR;
        }
    }

    /* Read the OMF file */
    status = ias_parm_read(app_params->omf_filename, omf_parms, count);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from OMF file: %s",
                app_params->omf_filename);
        return ERROR;
    }

    /* Check that the input file names necessary were provided */
    if (strlen(app_params->model_filename) == 0)
    {
        IAS_LOG_ERROR("Output model file name not set properly");
        return ERROR;
    }
    if (strlen(app_params->l0r_filename) == 0)
    {
        IAS_LOG_ERROR("Input L0R file name not set properly");
        return ERROR;
    }

    /* Was an override parameter provided in the overall ODL file? */
    if (IAS_PARM_GET_COUNT_READ(parms, ESTIMATED_SSM_POSITION_FILENAME) > 0)
    {
        /* Override provided, copy it over the OMF value */
        strncpy(app_params->ssm_position_filename, 
                ssm_position_filename_override, PATH_MAX);
        app_params->ssm_position_filename[PATH_MAX - 1] = '\0';

        IAS_LOG_INFO("Overriding the estimated TIRS SSM position file "
            "with '%s'", app_params->ssm_position_filename);
    }

    /* No ssm position filename provided */
    if (strlen(app_params->ssm_position_filename) == 0)
    {
        IAS_LOG_INFO("Estimated SSM file name was not provided");
        app_params->ssm_position_filename[0] = '\0';
    }

    /* Do some checks on the values */

    /* Set the precision flag based on whether any of the precision values
       were provided */
    app_params->precision_flag = 0;
    count = IAS_PARM_GET_COUNT_READ(parms, PRECISION_REF_TIME);
    count += IAS_PARM_GET_COUNT_READ(parms, PRECISION_X_CORR_PARAMS);
    count += IAS_PARM_GET_COUNT_READ(parms, PRECISION_Y_CORR_PARAMS);
    count += IAS_PARM_GET_COUNT_READ(parms, PRECISION_Z_CORR_PARAMS);
    count += IAS_PARM_GET_COUNT_READ(parms, PRECISION_ROLL_CORR_PARAMS);
    count += IAS_PARM_GET_COUNT_READ(parms, PRECISION_PITCH_CORR_PARAMS);
    count += IAS_PARM_GET_COUNT_READ(parms, PRECISION_YAW_CORR_PARAMS);
    if (count > 0)
    {
        int order;

        /* If the parameters were in the param file check to see if any
           were non-zeros, then set our precision flag to apply the params
           to the model. Even if provided, but all zeros, no need to apply
           to the model - it would be a waste of time for basically a no-op. */
        if (app_params->precision_model.seconds_from_image_epoch != 0.0)
            app_params->precision_flag = 1;
        for (i = 0; i < IAS_PRECISION_MAX_POLY_COEFF; i++)
        {
            if (app_params->precision_model.x_corr[i] != 0.0
                    || app_params->precision_model.y_corr[i] != 0.0
                    || app_params->precision_model.z_corr[i] != 0.0
                    || app_params->precision_model.roll_corr[i] != 0.0
                    || app_params->precision_model.pitch_corr[i] != 0.0
                    || app_params->precision_model.yaw_corr[i] != 0.0)
                app_params->precision_flag = 1;
        }
        if (app_params->precision_flag)
        {
            /* Found some precision params, so set attitude and ephemeris
               "order" (really, the number of coefficients), the correction
               model "order" is one less than the number of coefficients.
               i.e. 2 coefficients is a 1st order (linear) correction model,
               3 coefficients is a 2nd order (quadratic) correction model. */
            order = 0;
            count = IAS_PARM_GET_COUNT_READ(parms, PRECISION_X_CORR_PARAMS);
            if (count > order)
                order = count;
            count = IAS_PARM_GET_COUNT_READ(parms, PRECISION_Y_CORR_PARAMS);
            if (count > order)
                order = count;
            count = IAS_PARM_GET_COUNT_READ(parms, PRECISION_Z_CORR_PARAMS);
            if (count > order)
                order = count;
            app_params->precision_model.ephemeris_order = order;

            order = 0;
            count = IAS_PARM_GET_COUNT_READ(parms, PRECISION_ROLL_CORR_PARAMS);
            if (count > order)
                order = count;
            count = IAS_PARM_GET_COUNT_READ(parms,PRECISION_PITCH_CORR_PARAMS);
            if (count > order)
                order = count;
            count = IAS_PARM_GET_COUNT_READ(parms, PRECISION_YAW_CORR_PARAMS);
            if (count > order)
                order = count;
            app_params->precision_model.attitude_order = order;

            if (app_params->precision_model.ephemeris_order
                    > IAS_PRECISION_MAX_POLY_COEFF)
            {
                IAS_LOG_ERROR("Invalid precision ephemeris order");
                return ERROR;
            }

            if (app_params->precision_model.attitude_order
                    > IAS_PRECISION_MAX_POLY_COEFF)
            {
                IAS_LOG_ERROR("Invalid precision attitude order");
                return ERROR;
            }
        }
    }

    /* Get the satellite ID from the satellite number provided */
    app_params->satellite_id =
        ias_sat_attr_get_satellite_id_from_satellite_number(satellite_number);

    if (app_params->satellite_id == ERROR)
    {
        IAS_LOG_ERROR("Unable to determine satellite ID from satellite "
                "number %d", satellite_number);
        return ERROR;
    }

    return SUCCESS;
}
