#include <stdlib.h>
#include <string.h>
#include "ias_model_io.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_db_get_connect_info.h"
#include "ias_db_query.h"
#include "../create_los_model.h"
#include "cgreen/cgreen.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"

#define SQL_STR_MAX_LENGTH      600

Ensure create_los_model_test(void);
Ensure create_los_model_ssm_quality_test(void);
Ensure create_los_model_no_ssm_name_test(void);
Ensure test_forward_model(void);
Ensure tirs_create_los_model_test(void);
Ensure test_tirs_forward_model(void);
Ensure estimated_encoder_test(void);
Ensure user_input_value_encoder_test(void);
Ensure mode_switch_encoder_test(void);
Ensure create_los_model_ssm_override_test(void);
Ensure multiple_tirs_scene_data_test(void);

typedef struct findtime_inputs
{
    int band;
    int sca;
    double line;
    double sample;
    IAS_SENSOR_DETECTOR_TYPE dettype;
} FINDTIME_INPUTS;

typedef struct model_inputs
{
    int band;
    int sca;
    double line;
    double sample;
    IAS_SENSOR_DETECTOR_TYPE dettype;
} MODEL_INPUTS;

int test_database;
char *unit_test_dir;

int main(int argc, char *argv[])
{
    char *enable_db_tests;
    TestSuite *suite = create_test_suite();
    const char *unit_test_dir_name = "IAS_UNIT_TEST_DATA_DIR";

    /* Set the IAS version override environment variable */
    if (setenv("IAS_VERSION_OVERRIDE", "Y", 1) != 0)
    {
        printf("Error:Unable to set IAS_VERSION_OVERRIDE environment variable");
        return ERROR;
    }
    
    /* Retrieve the unit test directory */
    unit_test_dir = getenv(unit_test_dir_name);
    if (!unit_test_dir)
    {
        printf("Error: Retrieving unit test directory path");
        return ERROR;
    }

    /* See if the database tests should be enabled */
    enable_db_tests = getenv("ENABLE_DB_TESTS");
    if (enable_db_tests)
    {
        test_database = 1;
    }
    else
    {
        test_database = 0;
        printf("\n");
        printf("%s:  Database tests are disabled.\n", __FILE__);
        printf("\n");
    }

    add_test(suite, create_los_model_no_ssm_name_test);
    add_test(suite, create_los_model_ssm_override_test);
    add_test(suite, create_los_model_ssm_quality_test);
    add_test(suite, create_los_model_test);
    add_test(suite, test_forward_model);
    add_test(suite, tirs_create_los_model_test);
    add_test(suite, test_tirs_forward_model);
    add_test(suite, estimated_encoder_test);
    add_test(suite, user_input_value_encoder_test);
    add_test(suite, mode_switch_encoder_test);
    add_test(suite, multiple_tirs_scene_data_test);
    ias_log_initialize("create_los_model_test");
    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}

/* helper routine to initialize the database library and make a connection */
struct ias_db_connection *connect_to_database()
{
    struct ias_db_connection *db = NULL;/* Database connect object */
    char *database_name = NULL;         /* Database connect info - name */
    char *user_name = NULL;             /* Database connect info - user name */
    char *password = NULL;              /* Database connect info - password */
    char *host = NULL;                  /* Database connect info - host */
    int status;

    /* Initialize the satellite attributes library */
    ias_sat_attr_initialize(IAS_L8);

    /* Initialize the database library */
    status = ias_db_initialize_database_lib();
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error initializing database connection");
        return NULL;
    }

    /* Get the database connection info */
    status = ias_db_get_connect_info(&database_name, &user_name, &password,
            &host);
    if (status != SUCCESS)
    {
        assert_true_with_message(0,
                "Error getting database connection information");
        ias_db_close_database_lib();
        return NULL;
    }

    /* Connect to the database */
    db = ias_db_connect_to_database(database_name, user_name, password, host);

    /* Free the database connection information */
    free(database_name);
    free(user_name);
    free(password);
    free(host);

    if (!db)
    {
        assert_true_with_message(0,
                "Error establishing connection to database");
        ias_db_close_database_lib();
        return NULL;
    }

    return db;
}

/* helper routine to shut down the database connection and library */
void disconnect_database
(
    struct ias_db_connection *db    /* I: database connection to close */
)
{
    ias_db_close_connection(db);
    ias_db_close_database_lib();
}


/* Setup a row in the database, as if ADP had been run previously, so this
   model creation can have a row to update in the DB at the end.  Returns
   the characterization id for later use (0 if error). */
int prepare_db
(
    struct ias_db_connection *db, /* I: Database connection */
    const char *wo_id,            /* I: work order id */
    char *char_id                 /* O: geometric characterization id */
)
{
    struct ias_db_query *query = NULL;  /* Database query structure */
    char db_error_msg[IAS_SQL_RTN_MSGLEN]; /* SQL return messages */
    char sql[SQL_STR_MAX_LENGTH];       /* Buffer for SQL command string */
    int status;                         /* General routine return status */
    int rows_modified;                  /* Number of rows changed by DB query */

    /* Start the transaction */
    status = ias_db_start_transaction(db);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error starting the database transaction");
        return ERROR;
    }

    /* Get the work order "common id" */
    status = ias_misc_get_char_id(db, "geometric_work_order_common",
            "geometric_wo_common_id", wo_id, char_id);
    if (status == ERROR)
    {
        assert_true_with_message(0, "Error getting work order common ID from "
                "work order ID %s for geometric processing", wo_id);
        return ERROR;
    }

    /* Check if the needed test row in the DB already exists.
       Quit now if it's already there (create_los_model updates it).
       Add the row if it's not htere (so create_los_model update won't fail). */
    status = snprintf(sql, SQL_STR_MAX_LENGTH,
            "SELECT GEOMETRIC_WO_COMMON_ID FROM MODEL_GENERATION "
            "WHERE GEOMETRIC_WO_COMMON_ID = %s;",
            char_id
            );
    if (status < 0 || status > SQL_STR_MAX_LENGTH)
    {
        assert_true_with_message(0, "Error query is too long");
        return ERROR;
    }

    /* Perform the query */
    query = ias_db_perform_query(db, sql);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_msg,
                sizeof(db_error_msg));
        assert_true_with_message(0,
                "Error updating database; performing update: %s", db_error_msg);
        ias_db_query_close(query);
        return ERROR;
    }

    /* Check if the query returned any results */
    if(ias_db_query_next(query))
    {
        /* Found the row in the DB, so no need to create it for the
           create_los_model test to be able to update the row */
        ias_db_query_close(query);
        return SUCCESS;
    }

    /* Close the query */
    ias_db_query_close(query);
    query = NULL;

    /* A row like ADP would have created does not exist in the database for
       the test work order, so create it so create_los_model test can
       successfully update it. */
    /* Build an SQL statement to mimic the row ADP would create */
    status = snprintf(sql, SQL_STR_MAX_LENGTH,
            "INSERT INTO MODEL_GENERATION "
            "( "
            " GEOMETRIC_WO_COMMON_ID, "
            " NBR_ATTITUDE_POINTS, "
            " ATTITUDE_START_YEAR, "
            " ATTITUDE_START_DAY, "
            " ATTITUDE_START_SECONDS, "
            " NBR_BAD_ATTITUDE_POINTS, "
            " NBR_EPHEMERIS_POINTS, "
            " EPHEMERIS_START_YEAR, "
            " EPHEMERIS_START_DAY, "
            " EPHEMERIS_START_SECONDS, "
            " NBR_BAD_EPHEMERIS_POINTS "
            ") "
            " values "
            "( "
            " '%s', "
            " '%d', "
            " '%d', "
            " '%d', "
            " '%d', "
            " '%d', "
            " '%d', "
            " '%d', "
            " '%d', "
            " '%d', "
            " '%d' "
            ");",
        char_id,
        2,
        2008,
        128,
        45621,
        0,
        2,
        2008,
        128,
        45621,
        0
            );
    if (status < 0 || status > SQL_STR_MAX_LENGTH)
    {
        assert_true_with_message(0, "Error query is too long");
        return ERROR;
    }

    /* Perform the query */
    query = ias_db_perform_query(db, sql);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_msg,
                sizeof(db_error_msg));
        assert_true_with_message(0,
                "Error updating database; performing update: %s", db_error_msg);
        ias_db_query_close(query);
        return ERROR;
    }

    /* Check if one and only one row affected by the query, as expected */
    rows_modified = ias_db_query_get_modified_rows(query);
    if (rows_modified != 1)
    {
        /* -1 is special value indicating an error */
        if (rows_modified == -1)
        {
            assert_true_with_message(0,
                    "Error retrieving number of database rows modified");
        }
        else
        {
            assert_true_with_message(0, "Updating database; number of rows "
                    "modified incorrect; expected 1, got %d", rows_modified);
        }
        ias_db_query_close(query);
        return ERROR;
    }

    /* Close the query */
    ias_db_query_close(query);
    query = NULL;

    /* Commit the transaction */
    status = ias_db_commit_transaction(db);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error committing transaction: %s",
                ias_db_connect_last_error(db));
        return ERROR;
    }

    return SUCCESS;
}

typedef struct EXPECTED_DB_VALUES
{
    int oli_lines;
    int oli_start_year;
    int oli_start_day;
    double oli_start_seconds;
    int oli_bad_lines;
    double oli_delta_time;
    int tirs_lines;
    int tirs_start_year;
    int tirs_start_day;
    double tirs_start_seconds;
    int tirs_bad_lines;
    double tirs_delta_time;
    int tirs_bad_ssm_times;
    int tirs_bad_ssm_encoder_values;
} EXPECTED_DB_VALUES;

/* helper routine to verify the database contents are as expected */
int verify_database_contents
(
    struct ias_db_connection *db,   /* I: database connection to use */
    char *char_id,                  /* I: characterization id */
    const EXPECTED_DB_VALUES *expected_values /* I: expected DB values */
)
{
    char sql[2000];     /* select statement */
    struct ias_db_query *query = NULL;
    int status;
    EXPECTED_DB_VALUES db_values;

    /* build the query statement */
    if (snprintf(sql, sizeof(sql),
        "SELECT "
            "NBR_IMAGE_LINES_OLI, "
            "IMAGE_START_YEAR_OLI, "
            "IMAGE_START_DAY_OLI, "
            "IMAGE_START_SECONDS_OLI, "
            "NBR_BAD_IMAGE_LINES_OLI, "
            "DELTA_TIME_OLI, "
            "NBR_IMAGE_LINES_TIRS, "
            "IMAGE_START_YEAR_TIRS, "
            "IMAGE_START_DAY_TIRS, "
            "IMAGE_START_SECONDS_TIRS, "
            "NBR_BAD_IMAGE_LINES_TIRS, "
            "DELTA_TIME_TIRS, "
            "NBR_BAD_SSM_TIME_CODES_TIRS, "
            "NBR_BAD_SSM_ENCODER_VAL_TIRS "
        "FROM MODEL_GENERATION where GEOMETRIC_WO_COMMON_ID = %s;",
        char_id) >= sizeof(sql))
    {
        assert_true_with_message(0, "Error building sql statement");
        return ERROR;
    }

    /* query the database for the inserted data */
    query = ias_db_perform_query(db, sql);
    assert_true(query != NULL);
    if (!query)
        return ERROR;

    status = ias_db_query_was_successful(query);
    assert_true(status != 0);
    if (status == 0)
    {
        ias_db_query_close(query);
        return ERROR;
    }

    /* get the first row */
    status = ias_db_query_next(query);
    assert_true(status);

    /* get the returned results */
    status = ias_db_query_int_value(query, &db_values.oli_lines, 0);
    assert_equal(status, SUCCESS);
    status = ias_db_query_int_value(query, &db_values.oli_start_year, 1);
    assert_equal(status, SUCCESS);
    status = ias_db_query_int_value(query, &db_values.oli_start_day, 2);
    assert_equal(status, SUCCESS);
    status = ias_db_query_double_value(query, &db_values.oli_start_seconds, 3);
    assert_equal(status, SUCCESS);
    status = ias_db_query_int_value(query, &db_values.oli_bad_lines, 4);
    assert_equal(status, SUCCESS);
    status = ias_db_query_double_value(query, &db_values.oli_delta_time, 5);
    assert_equal(status, SUCCESS);
    status = ias_db_query_int_value(query, &db_values.tirs_lines, 6);
    assert_equal(status, SUCCESS);
    status = ias_db_query_int_value(query, &db_values.tirs_start_year, 7);
    assert_equal(status, SUCCESS);
    status = ias_db_query_int_value(query, &db_values.tirs_start_day, 8);
    assert_equal(status, SUCCESS);
    status = ias_db_query_double_value(query, &db_values.tirs_start_seconds, 9);
    assert_equal(status, SUCCESS);
    status = ias_db_query_int_value(query, &db_values.tirs_bad_lines, 10);
    assert_equal(status, SUCCESS);
    status = ias_db_query_double_value(query, &db_values.tirs_delta_time, 11);
    assert_equal(status, SUCCESS);
    status = ias_db_query_int_value(query, &db_values.tirs_bad_ssm_times, 12);
    assert_equal(status, SUCCESS);
    status = ias_db_query_int_value(query,
                &db_values.tirs_bad_ssm_encoder_values, 13);
    assert_equal(status, SUCCESS);

    ias_db_query_close(query);
    query = NULL;

    /* compare the data to the expected values */
    assert_equal(db_values.oli_lines, expected_values->oli_lines);
    assert_equal(db_values.oli_start_year, expected_values->oli_start_year);
    assert_equal(db_values.oli_start_day, expected_values->oli_start_day);
    assert_double_equal(db_values.oli_start_seconds,
                        expected_values->oli_start_seconds);
    assert_double_equal(db_values.oli_delta_time,
                        expected_values->oli_delta_time);
    assert_equal(db_values.tirs_lines, expected_values->tirs_lines);
    assert_equal(db_values.tirs_start_year, expected_values->tirs_start_year);
    assert_equal(db_values.tirs_start_day, expected_values->tirs_start_day);
    assert_double_equal(db_values.tirs_start_seconds,
                        expected_values->tirs_start_seconds);
    assert_double_equal(db_values.tirs_delta_time,
                        expected_values->tirs_delta_time);
    assert_equal(db_values.tirs_bad_ssm_times,
                        expected_values->tirs_bad_ssm_times);
    assert_equal(db_values.tirs_bad_ssm_encoder_values,
                        expected_values->tirs_bad_ssm_encoder_values);
    
    return SUCCESS;
}

Ensure create_los_model_test(void)
{
    int status;
    struct ias_db_connection *db = NULL;/* Database connect object */
    char char_id[IAS_CHAR_ID_SIZE];  
    char param_filename[PATH_MAX] = "parameters_no_db.odl";
    char command[PATH_MAX * 2];

    if (test_database)
    {
        strcpy(param_filename, "parameters.odl");

        db = connect_to_database();
        assert_true(db != NULL);
        if (!db)
        {
            disconnect_database(db);
            return;
        }

        status = prepare_db(db, "WO1234_new", char_id);
        assert_true_with_message(status == SUCCESS,
                    "Error preparing DB for test");
        if (status != SUCCESS)
        {
            /* Don't know state of DB, so do not attempt create_los_model,
               since it probably will also encounter DB errors */
            disconnect_database(db);
            return;
        }
    }

    sprintf(command, "../create_los_model %s", param_filename);
    status = system(command);
    assert_true_with_message(status == 0, "Create LOS Model failed");

    if (test_database)
    {
        EXPECTED_DB_VALUES expected =
            {6050, 2003, 148, 65067.366800, 0, 0.004236,
                0, 0, 0, 0.0, 0, 0.0, 0, 0 };

        /* make sure the database was properly updated */
        verify_database_contents(db, char_id, &expected);

        /* disconnect from the database */
        disconnect_database(db);
    }

    status = system("../create_los_model parameters_precision.odl");
    assert_true_with_message(status == 0, "Create LOS Model failed");
    assert_equal(status, 0);
}

Ensure create_los_model_ssm_override_test(void)
{
    char est_ssm_filename[PATH_MAX] = "ssm_override_test.data";
    char orig_ssm_filename[PATH_MAX] = "est_ssm_encoder_quality.data_new";
    int status;
    MODEL_APP_PARAMS app_params; /* Parameter file and OMF values */

    /* Disable all output since there will be an info log message about the 
       override */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* Read the parameters from the input parameter files (app param and OMF) */
    strcpy(app_params.parameter_filename, "ssm_override.odl");
    status = read_parameters(&app_params);
    assert_equal(status, SUCCESS);

    /* Verify that the est ssm filename is the override file name and not
       the one listed in the associated OMF for this test:
       ESTIMATED_SSM_POSITION_FILENAME = "est_ssm_encoder_quality.data_new" */
    assert_string_equal(est_ssm_filename, app_params.ssm_position_filename);
    assert_string_not_equal(orig_ssm_filename,
        app_params.ssm_position_filename);
}

Ensure create_los_model_ssm_quality_test(void)
{
    char command[PATH_MAX * 2];
    char est_ssm_filename[PATH_MAX] = "est_ssm_encoder_quality.data";
    char param_filename[PATH_MAX] = "quality_tests.odl";
    char omf_filename[PATH_MAX] = "WOQUAL_new.omf";
    int count;
    int ssm_position_quality;
    int status;

    /* disable all output since the tests are hitting error conditions */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    IAS_PARM_DECLARE_TABLE(parms, 1);
    IAS_PARM_ADD_INT(parms, SSM_POSITION_QUALITY,
        "TIRS SSM Position Quality", IAS_PARM_OPTIONAL,
        IAS_PARM_NOT_ARRAY, 1, 0, 3, 0, NULL,
        &ssm_position_quality, sizeof(ssm_position_quality), 0);
    count = IAS_PARM_GET_TABLE_SIZE(parms);
    assert_equal(count, 1);

    /* Update file, set quality = 2 then run test */
    sprintf(command, "./convert_ssm_encoder_file.pl %s 2", est_ssm_filename);
    status = system(command);
    assert_true_with_message(status == 0, "convert_ssm_encoder_file.pl failed");
    sprintf(command, "../create_los_model %s", param_filename);
    status = system(command);
    assert_true_with_message(status == 0, "Create LOS Model failed");
    status = ias_parm_read(omf_filename, parms, count);
    assert_true_with_message(status == 0, "IAS PARM READ Failed");
    assert_equal(ssm_position_quality, 2);

    /* Update file, set quality = 1 then run test */
    sprintf(command, "./convert_ssm_encoder_file.pl %s 1", est_ssm_filename);
    status = system(command);
    assert_true_with_message(status == 0, "convert_ssm_encoder_file.pl failed");
    sprintf(command, "../create_los_model %s", param_filename);
    status = system(command);
    assert_true_with_message(status == 0, "Create LOS Model failed");
    status = ias_parm_read(omf_filename, parms, count);
    assert_true_with_message(status == 0, "IAS PARM READ Failed");
    assert_equal(ssm_position_quality, 1);

    /* Update file, set quality = 0 then run test */
    sprintf(command, "./convert_ssm_encoder_file.pl %s 0", est_ssm_filename);
    status = system(command);
    assert_true_with_message(status == 0, "convert_ssm_encoder_file.pl failed");
    sprintf(command, "../create_los_model %s", param_filename);
    status = system(command);
    assert_true_with_message(status == 0, "Create LOS Model failed");
    status = ias_parm_read(omf_filename, parms, count);
    assert_true_with_message(status == 0, "IAS PARM READ Failed");
    assert_equal(ssm_position_quality, 0);

    /* Change a few qualities to 1 but leave 2's as well */
    sprintf(command, "./convert_ssm_encoder_file.pl %s 1 EVERY_OTHER",
        est_ssm_filename);
    status = system(command);
    assert_true_with_message(status == 0, "convert_ssm_encoder_file.pl failed");
    sprintf(command, "../create_los_model %s", param_filename);
    status = system(command);
    assert_true_with_message(status == 0, "Create LOS Model failed");
    status = ias_parm_read(omf_filename, parms, count);
    assert_true_with_message(status == 0, "IAS PARM READ Failed");
    assert_equal(ssm_position_quality, 1);

    /* Change a few qualities to 0 but leave 2's as well */
    sprintf(command, "./convert_ssm_encoder_file.pl %s 0 EVERY_OTHER",
        est_ssm_filename);
    status = system(command);
    assert_true_with_message(status == 0, "convert_ssm_encoder_file.pl failed");
    sprintf(command, "../create_los_model %s", param_filename);
    status = system(command);
    assert_true_with_message(status == 0, "Create LOS Model failed");
    status = ias_parm_read(omf_filename, parms, count);
    assert_true_with_message(status == 0, "IAS PARM READ Failed");
    assert_equal(ssm_position_quality, 0);

    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);
}

Ensure create_los_model_no_ssm_name_test(void)
{
    char command[PATH_MAX * 2];
    char param_filename[PATH_MAX] = "no_ssm_tests.odl";
    int status;

    /* disable all output since the tests are hitting error conditions */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* This parameter file does not contain an estimated ssm file, even
       though the processing would use one if present.  The fact that it
       completes successfully tests that the library can handle a missing 
       filename.  The read parameters routine sets it to an empty string. */
    sprintf(command, "../create_los_model %s", param_filename);
    status = system(command);
    assert_true_with_message(status == 0, "Create LOS Model failed");
    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);
}

Ensure test_forward_model()
{
    double tarlatd;
    double tarlong;
    double tarelev = 1000.0;

    IAS_LOS_MODEL *model = NULL;
    char model_filename[PATH_MAX];
    /* the following values were obtained by running the cal/val prototype
       code. this file is linked in from the los_model forward model unit test
       data. */
    #include "forward_model_test_data.h"
    int count = sizeof(inputs)/sizeof(MODEL_INPUTS);
    int index;

    ias_sat_attr_initialize(IAS_L8);

    /* get the test model */

    /* build the model filename */
    sprintf(model_filename, "ias_model.h5");

    model = ias_model_read(model_filename);
    assert_true(model != NULL);
    if (!model)
        return;

    /* disable all output since the tests are hitting error conditions */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    for (index = 0; index < count; index++)
    {
        MODEL_INPUTS *input = &inputs[index];
        int status;

        status = ias_los_model_input_line_samp_to_geodetic(input->line,
                input->sample, input->band, input->sca, tarelev, model,
                input->dettype, NULL, &tarlatd, &tarlong);

        if (status != output_status[index])
        {
            /* only use asserts if there is a failure due to the cgreen limit
               of 16384 asserts per run */
            assert_equal_with_message(status, output_status[index],
                    "failed status - expected %d, got %d for index %d",
                    output_status[index], status, index);
        }
        else if (status != ERROR)
        {
            /* use a cgreen function to verify the doubles are equal to 8
               significant digits */
            int valid = doubles_are_equal(output_tarlatd[index], tarlatd)
                && doubles_are_equal(output_tarlong[index], tarlong);
            if (!valid)
            {
                assert_equal_with_message(valid, 1, 
                        "output error - expected (%.8e, %.8e) "
                        "got (%.8e, %.8e) for index %d", output_tarlatd[index],
                        output_tarlong[index], tarlatd, tarlong, index);
            }
        }
    }

    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    ias_los_model_free(model);
}

Ensure tirs_create_los_model_test(void)
{
    int status;
    struct ias_db_connection *db = NULL;/* Database connect object */
    char char_id[IAS_CHAR_ID_SIZE];  
    char param_filename[PATH_MAX] = "tirs_parameters_no_db.odl";
    char command[PATH_MAX * 2];

    if (test_database)
    {
        strcpy(param_filename, "tirs_parameters.odl");

        db = connect_to_database();
        assert_true(db != NULL);
        if (!db)
        {
            disconnect_database(db);
            return;
        }

        status = prepare_db(db, "WO_TIRS_new", char_id);
        assert_true_with_message(status == SUCCESS,
                        "Error preparing DB for test");
        if (status != SUCCESS)
        {
            /* Don't know state of DB, so do not attempt create_los_model, since
               it probably will also encounter DB errors */
            disconnect_database(db);
            return;
        }
    }

    sprintf(command, "../create_los_model %s", param_filename);
    status = system(command);
    assert_true_with_message(status == 0, "TIRS Create LOS Model failed");

    if (test_database)
    {
        EXPECTED_DB_VALUES expected =
            {0, 0, 0, 0.0, 0, 0.0,
             2291, 2000, 227, 65506.609566, 0, 0.014286, 0, 0};

        /* make sure the database was properly updated */
        verify_database_contents(db, char_id, &expected);

        /* disconnect from the database */
        disconnect_database(db);
    }
}

Ensure test_tirs_forward_model()
{
    double tarlatd;
    double tarlong;
    double tarelev = 1000.0;

    IAS_LOS_MODEL *model = NULL;
    char model_filename[PATH_MAX];
    /* the following values were obtained by running the cal/val prototype
       code. this file is linked in from the los_model forward model unit test
       data. */
    #include "tirs_forward_model_test_data.h"
    int count = sizeof(inputs)/sizeof(MODEL_INPUTS);
    int index;

    ias_sat_attr_initialize(IAS_L8);

    /* get the test model */

    /* build the model filename */
    sprintf(model_filename, "tirs_model.h5");

    model = ias_model_read(model_filename);
    assert_true(model != NULL);
    if (!model)
        return;

    /* disable all output since the tests are hitting error conditions */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    for (index = 0; index < count; index++)
    {
        MODEL_INPUTS *input = &inputs[index];
        int status;

        status = ias_los_model_input_line_samp_to_geodetic(input->line,
                input->sample, input->band, input->sca, tarelev, model,
                input->dettype, NULL, &tarlatd, &tarlong);

        if (status != output_status[index])
        {
            /* only use asserts if there is a failure due to the cgreen limit
               of 16384 asserts per run */
            assert_equal_with_message(status, output_status[index],
                    "failed status - expected %d, got %d for index %d",
                    output_status[index], status, index);
        }
        else if (status != ERROR)
        {
            /* use a cgreen function to verify the doubles are equal to 8
               significant digits */
            int valid = doubles_are_equal(output_tarlatd[index], tarlatd)
                && doubles_are_equal(output_tarlong[index], tarlong);
            if (!valid)
            {
                assert_equal_with_message(valid, 1, 
                        "output error - expected (%.8e, %.8e) "
                        "got (%.8e, %.8e) for index %d", output_tarlatd[index],
                        output_tarlong[index], tarlatd, tarlong, index);
            }
        }
    }

    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    ias_los_model_free(model);
}

Ensure estimated_encoder_test(void)
{
    int status;
    char param_filename[PATH_MAX] = "encoder_off.odl";
    char model_filename[PATH_MAX] = "systematic_model_encoff_estimated.h5";
    char comparison_model_filename[PATH_MAX] = 
        "systematic_model_encoff_estimated_v2.h5";
    char command[PATH_MAX * 2];
    
    /* Build the Create LOS model run command */
    status = snprintf(command, sizeof(command), "../create_los_model %s", 
        param_filename);
    if (status < 0 || status >= sizeof(command))
    {   
        assert_true_with_message(0, "Error: Creating execution command");
        return;
    }

    /* Run Create LOS model */
    status = system(command);
    assert_true_with_message(status == 0, "Error: Creating LOS model");

    status = snprintf(command, sizeof(command), "h5diff -q -n 1 %s/ias_base"
        "/gps/create_los_model/ENCODER_OFF/%s ./%s", 
        unit_test_dir, comparison_model_filename, model_filename);
    if (status < 0 || status >= sizeof(command))
    {   
        assert_true_with_message(0, "Error: Creating ssm encoder path");
        return;
    }

    status = system(command);
    assert_true_with_message(status == 0, "Error: Files differ");
}

Ensure user_input_value_encoder_test(void)
{
    int status;
    char param_filename[PATH_MAX] = "encoder_off_user_input.odl";
    char model_filename[PATH_MAX] = "systematic_model_encoff_user.h5";
    char comparison_model_filename[PATH_MAX] = 
        "systematic_model_encoff_user_v2.h5";
    char command[PATH_MAX * 2];
    
    /* Build the Create LOS model run command */
    status = snprintf(command, sizeof(command), "../create_los_model %s", 
        param_filename);
    if (status < 0 || status >= sizeof(command))
    {   
        assert_true_with_message(0, "Error: Creating execution command");
        return;
    }

    /* Run Create LOS model */
    status = system(command);
    assert_true_with_message(status == 0, "Error: Creating LOS model");

    /* Verify the model */
    status = snprintf(command, sizeof(command), "h5diff -q -n 1 %s/ias_base"
        "/gps/create_los_model/ENCODER_OFF/%s ./%s", 
        unit_test_dir, comparison_model_filename, model_filename);
    if (status < 0 || status >= sizeof(command))
    {   
        assert_true_with_message(0, "Error: Creating ssm encoder path");
        return;
    }

    status = system(command);
    assert_true_with_message(status == 0, "Error: Files differ");
}

Ensure mode_switch_encoder_test(void)
{
    int status;
    char param_filename[PATH_MAX] = "encoder_switch_mode.odl";
    char model_filename[PATH_MAX] = "systematic_model_mode_switch.h5";
    char comparison_model_filename[PATH_MAX] = 
        "systematic_model_mode_switch_v2.h5";
    char command[PATH_MAX * 2];
   
    /* Build the Create LOS model run command */
    status = snprintf(command, sizeof(command), "../create_los_model %s", 
        param_filename);
    if (status < 0 || status >= sizeof(command))
    {   
        assert_true_with_message(0, "Error: Creating execution command");
        return;
    }

    /* Run Create LOS model */
    status = system(command);
    assert_true_with_message(status == 0, "Error: Creating LOS model");

    /* Verify the model */
    status = snprintf(command, sizeof(command), "h5diff -q -n 1 %s/ias_base"
        "/gps/create_los_model/ENCODER_MODE_4_TO_0/%s ./%s", 
        unit_test_dir, comparison_model_filename, model_filename);
    if (status < 0 || status >= sizeof(command))
    {   
        assert_true_with_message(0, "Error: Creating ssm encoder path");
        return;
    }

    status = system(command);
    assert_true_with_message(status == 0, "Error: Files differ");
}

Ensure multiple_tirs_scene_data_test(void)
{
    int status;
    char param_filename[PATH_MAX] = "multiple_scene.odl";
    char model_filename[PATH_MAX] = "multiple_scene_model.h5";
    char comparison_model_filename[PATH_MAX] = "multiple_scene_model_v2.h5";
    char command[PATH_MAX * 2];
   
    /* Build the Create LOS model run command */
    status = snprintf(command, sizeof(command), "../create_los_model %s", 
        param_filename);
    if (status < 0 || status >= sizeof(command))
    {   
        assert_true_with_message(0, "Error: Creating execution command");
        return;
    }

    /* Run Create LOS model */
    status = system(command);
    assert_true_with_message(status == 0, "Error: Creating LOS model");

    /* Verify the model */
    status = snprintf(command, sizeof(command), "h5diff -q -n 1 %s/ias_base"
        "/gps/create_los_model/TIRS_THREE_SCENE/%s ./%s", 
        unit_test_dir, comparison_model_filename, model_filename);
    if (status < 0 || status >= sizeof(command))
    {   
        assert_true_with_message(0, "Error: Creating ssm encoder path");
        return;
    }

    status = system(command);
    assert_true_with_message(status == 0, "Error: Files differ");
}
