/*----------------------------------------------------------------------------
NAME:    write_characterization_data

PURPOSE: Writes the data used for characterization to the IAS database

RETURNS: SUCCESS or ERROR

Notes:
    Assumes ADP was run prior to this application, creating the intial row
    in the database model_generation table for this work order.
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_db_get_connect_info.h"
#include "ias_satellite_attributes.h"
#include "create_los_model.h"

#define SQL_STR_MAX_LENGTH      600

int write_characterization_data
(
    const IAS_LOS_MODEL *model,         /* I: completely populated model */
    const MODEL_APP_PARAMS *app_params, /* I: parameter file and OMF values */
    const IAS_SENSOR_L0R_DATA *l0r_data /* I: L0R file values */
)
{
    struct ias_db_connection *db = NULL;/* Database connect object */
    struct ias_db_query *query = NULL;  /* Database query structure */
    char db_error_msg[IAS_SQL_RTN_MSGLEN]; /* SQL return messages */
    char *database_name = NULL;         /* Database connect info - name */
    char *user_name = NULL;             /* Database connect info - user name */
    char *password = NULL;              /* Database connect info - password */
    char *host = NULL;                  /* Database connect info - host */
    char sql[SQL_STR_MAX_LENGTH];       /* Buffer for SQL command string */
    int status;                         /* General routine return status */
    int rows_modified;                  /* Number of rows changed by DB query */
    char char_id[IAS_CHAR_ID_SIZE];   /* Geometric common work order ID */
    int sensor_index;
    int num_lines_oli = 0;
    int num_lines_tirs = 0;
    int epoch_year_oli = 0;
    int epoch_year_tirs = 0;
    int epoch_day_oli = 0;
    int epoch_day_tirs = 0;
    double epoch_seconds_oli = 0;
    double epoch_seconds_tirs = 0;
    int num_bad_time_codes_oli = 0;
    int num_bad_time_codes_tirs = 0;
    double frame_time_oli = 0;
    double frame_time_tirs = 0;
    int num_bad_ssm_times = 0;
    int num_bad_ssm_encoder_values = 0;

    /* Initialize the database library */
    status = ias_db_initialize_database_lib();
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing database connection");
        return ERROR;
    }

    /* Get the database connection info */
    status = ias_db_get_connect_info(&database_name, &user_name, &password,
            &host);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting database connection information");
        ias_db_close_database_lib();
        return ERROR;
    }

    /* Connect to the database */
    db = ias_db_connect_to_database(database_name, user_name, password, host);

    /* Free the database connection information */
    free(database_name);
    free(user_name);
    free(password);
    free(host);

    if (!db)
    {
        IAS_LOG_ERROR("Establishing connection to database");
        ias_db_close_database_lib();
        return ERROR;
    }

    /* Start the transaction */
    status = ias_db_start_transaction(db);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Starting the database transaction");
        ias_db_close_connection(db);
        ias_db_close_database_lib();
        return ERROR;
    }

    /* Get the work order "common id" */
    status = ias_misc_get_char_id(db, "geometric_work_order_common",
            "geometric_wo_common_id", app_params->work_order_id, char_id);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Getting work order common ID from work order ID %s "
                "for geometric processing", app_params->work_order_id);
        ias_db_close_connection(db);
        ias_db_close_database_lib();
        return ERROR;
    }

    /* Prepare the characterization values for the IAS database update */
    for (sensor_index = 0; sensor_index < IAS_MAX_SENSORS; sensor_index++)
    {
        if ((sensor_index == IAS_OLI
                    && model->sensor.sensors[IAS_OLI].sensor_present)
                || (sensor_index == IAS_TIRS
                    && model->sensor.sensors[IAS_TIRS].sensor_present))
        {
            IAS_SENSOR_BAND_MODEL *band;        /* Current band */
            int band_numbers[IAS_MAX_NBANDS];   /* Band numbers for sensor */
            int num_sensor_bands;               /* Number of band #s in array */
            int band_num_index;                 /* Index to array of band #s */
            int band_index;

            status = ias_sat_attr_get_sensor_band_numbers(sensor_index,
                    IAS_NORMAL_BAND, 0, band_numbers, IAS_MAX_NBANDS,
                    &num_sensor_bands);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Using satellite attributes to retrieve band "
                        "for sensor %d", sensor_index);
                ias_db_close_connection(db);
                ias_db_close_database_lib();
                return ERROR;
            }

            for (band_num_index = 0; band_num_index < num_sensor_bands;
                    band_num_index++)
            {
                /* Get band index number from the band number */
                band_index = ias_sat_attr_convert_band_number_to_index(
                        band_numbers[band_num_index]);
                if (band_index == ERROR) /* set to ERROR if conversion fails */
                {
                    IAS_LOG_ERROR("Using satellite attributes to convert band "
                            "number %d to band index",
                            band_numbers[band_num_index]);
                    ias_db_close_connection(db);
                    ias_db_close_database_lib();
                    return ERROR;
                }

                band = &model->sensor.bands[band_index];

                /* Compute the total lines */
                if (band->band_present)
                {
                    if (sensor_index == IAS_OLI)
                    {
                        /* The DB field probably should be renamed to
                           NBR_IMAGE_FRAMES_OLI, so just keep this in
                           mind and don't multiple this frame count by
                           number of lines per frame to attempt to get an
                           actual line count. This could be a problem if
                           have an image where only a pan band, to avoid
                           the number of "lines" calculated won't be twice
                           as much as other scenes with no real indication
                           that the lines are "pan band lines" instead of
                           "ms band lines". */
                        num_lines_oli = band->frame_count;
                        epoch_year_oli = (int)band->utc_epoch_time[0];
                        epoch_day_oli = (int)band->utc_epoch_time[1];
                        epoch_seconds_oli = band->utc_epoch_time[2];
                        num_bad_time_codes_oli
                            = l0r_data->oli_data.num_bad_frame_times;
                        frame_time_oli = band->sampling_char.sampling_time;
                        break;
                    }
                    else if (sensor_index == IAS_TIRS)
                    {
                        num_lines_tirs = band->frame_count;
                        epoch_year_tirs = (int)band->utc_epoch_time[0];
                        epoch_day_tirs = (int)band->utc_epoch_time[1];
                        epoch_seconds_tirs = band->utc_epoch_time[2];
                        num_bad_time_codes_tirs
                            = l0r_data->tirs_data.num_bad_frame_times;
                        frame_time_tirs = band->sampling_char.sampling_time;
                        num_bad_ssm_times 
                            = l0r_data->tirs_data.num_bad_ssm_times;
                        num_bad_ssm_encoder_values
                            = l0r_data->tirs_data.num_bad_ssm_encoder_values;
                        break;
                    }
                }
            }
        }
    }

    /* Build the SQL statement */
    /* This is an update for the row the ADP should have already created 
       for this work order common ID */
    status = snprintf(sql, SQL_STR_MAX_LENGTH,
            "UPDATE MODEL_GENERATION "
            " SET "
            " NBR_IMAGE_LINES_OLI = %d, "
            " NBR_IMAGE_LINES_TIRS = %d, "
            " IMAGE_START_YEAR_OLI = %d, "
            " IMAGE_START_YEAR_TIRS = %d, "
            " IMAGE_START_DAY_OLI = %d, "
            " IMAGE_START_DAY_TIRS = %d, "
            " IMAGE_START_SECONDS_OLI = %lf, "
            " IMAGE_START_SECONDS_TIRS = %lf, "
            " NBR_BAD_IMAGE_LINES_OLI = %d, "
            " NBR_BAD_IMAGE_LINES_TIRS = %d, "
            " DELTA_TIME_OLI = %lf, "
            " DELTA_TIME_TIRS = %lf, "
            " NBR_BAD_SSM_TIME_CODES_TIRS = %d, "
            " NBR_BAD_SSM_ENCODER_VAL_TIRS = %d "
            " WHERE "
            " GEOMETRIC_WO_COMMON_ID = %s;",
            num_lines_oli,
            num_lines_tirs,
            epoch_year_oli,
            epoch_year_tirs,
            epoch_day_oli,
            epoch_day_tirs,
            epoch_seconds_oli,
            epoch_seconds_tirs,
            num_bad_time_codes_oli,
            num_bad_time_codes_tirs,
            frame_time_oli,
            frame_time_tirs,
            num_bad_ssm_times,
            num_bad_ssm_encoder_values,
            char_id
                );
    if (status < 0 || status >= SQL_STR_MAX_LENGTH)
    {
        IAS_LOG_ERROR("Query is too long");
        ias_db_close_connection(db);
        ias_db_close_database_lib();
        return ERROR;
    }

    /* Perform the query */
    query = ias_db_perform_query(db, sql);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_msg,
                sizeof(db_error_msg));
        IAS_LOG_ERROR("Updating database; performing update: %s", db_error_msg);
        ias_db_query_close(query);
        ias_db_close_connection(db);
        ias_db_close_database_lib();
        return ERROR;
    }

    /* Check if one and only one row affected by the query, as expected */
    rows_modified = ias_db_query_get_modified_rows(query);
    if (rows_modified != 1) 
    {
        /* -1 is special value indicating an error */
        if (rows_modified == -1)
            IAS_LOG_ERROR("Retrieving number of database rows modified");
        else
        {
            IAS_LOG_ERROR("Updating database; number of rows modified "
                    "incorrect; expected 1, got %d", rows_modified);
        }
        ias_db_query_close(query);
        ias_db_close_connection(db);
        ias_db_close_database_lib();
        return ERROR;
    }

    /* Close the query */
    ias_db_query_close(query);
    query = NULL;

    /* Commit the transaction */
    status = ias_db_commit_transaction(db);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Committing transaction: %s",
                ias_db_connect_last_error(db));
        ias_db_close_connection(db);
        ias_db_close_database_lib();
        return ERROR;
    }

    /* Close the connection */
    ias_db_close_connection(db);
    ias_db_close_database_lib();

    return SUCCESS;
}
