#!/usr/bin/perl
################################################################################
# Name: dump_estimated_tirs_ssm_positions.pl
#
# Purpose:
#   This script will query the tirs_ssm_pos_estimates table for estimated
#   ssm positions which it will then output to a file. How it decides to select
#   which ssm positions is detailed below.
#
# Usage:
#   See the help routine or run ./dump_estimated_tirs_ssm_positions.pl --help
################################################################################

use strict;
use ias_const;
use ias_db;
use ias_odl;
use DBI;
use Logger;
use File::Basename;
use Getopt::Long;
use Time::Piece;
use Time::Seconds;

############################# BEGIN MAIN #######################################

# Set up logging
my $prog_name = basename($0);
my $logger = Logger->new("STDOUT");
if (not defined $logger)
{
    print STERR "Error: failed to establish logger for STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name($prog_name);
set_default_log($logger);

my $omf_filename;
my ($help, $odl_file, $output_file, $start_date, $end_date,
    $as_of_date, $quiet);
if (parse_cmd_line_args(\$help, \$odl_file, \$output_file, \$start_date,
    \$end_date, \$as_of_date, \$quiet) != SUCCESS)
{
    $logger->fatal_msg("Error parsing command line parameters");
    $logger->close_log();
    exit FAILURE;
}

if ($help)
{
    help_output();
    exit SUCCESS;
}

# If an odl file was sent then we read our other parameters from that.
if (defined $odl_file and read_parameters_from_odl($odl_file, \$output_file,
        \$start_date, \$end_date, \$omf_filename) != SUCCESS)
{
    $logger->fatal_msg("Could not read parameters from parameter file "
        . "$odl_file");
    $logger->close_log();
    exit FAILURE;
}

if (not defined $as_of_date)
{
    my $now = gmtime();
    $as_of_date = get_year_doy_seconds_from_datetime($now);
}

# Issue a log message for the time period being retrieved
if (!$quiet)
{
    $logger->info_msg("Getting estimated TIRS SSM positions for the time "
        . "period $start_date to $end_date");
}

# Get results from database
my @est_positions;
if (select_est_ssm_positions(
    $start_date, $end_date, $as_of_date, \@est_positions) != SUCCESS)
{
    $logger->fatal_msg("Selecting estimated ssm positions");
    $logger->close_log();
    exit FAILURE;
}

# Write results to file
if (write_file($output_file, $start_date, $end_date, $as_of_date,
    \@est_positions) != SUCCESS)
{
    $logger->fatal_msg("Could not write positions to file $output_file");
    $logger->close_log();
    exit FAILURE;
}

# Add the output_filename to the omf
if (defined $odl_file
    and update_omf($omf_filename, $output_file) != SUCCESS)
{
    $logger->fatal_msg("Could not add output filename to $omf_filename");
    $logger->close_log();
    return ERROR;
}

if (!$quiet)
{
    $logger->info_msg("Esimated SSM positions successfully retrieved");
}
$logger->close_log();

exit SUCCESS;

################################### END MAIN ###################################

################################################################################
# Name: help_output
#
# Description: Prints help message
#
# Arguments:
#   None
#
# Returns:
#   None
################################################################################
sub help_output()
{
    my $filename = basename($0);

    # Ignore the poor indentation. It's necessary for proper format in terminal.
    my $help_message = "Usage: $filename [odl_file] | "
        . "(--output_file=output_filename [--start_date=date] "
        . "[--end_date=date] [--as_of_date=date] [--quiet])\n";
    $help_message .= << "END_MESSAGE";

General note: Day of year starts at 1 being January 1st. Hours are 0-23, minutes
    are 0-59, and seconds are 0-60.

--help: Show this message
odl_file: No other arguments should be sent if an ODL file is provided. This
    should contain an ESTIMATED_FILE field, a WO_DIRECTORY field, and a
    WORK_ORDER_ID field. The ESTIMATED_FILE is the file where the records will
    be written. WO_DIRECTORY and WORK_ORDER_ID are used to determine the OMF's
    name. The OMF should contain a ACQUISITION_DATE field under an OMF object.
    This field should be formatted like YYYY:DOY:HH:MM:SS. Records from the
    database will be written to the ESTIMATED_FILE should their dates fall
    within one week of the ACQUISITION_DATE.
--output_file: File which will hold the records. This must be provided if an
    ODL file is not.
--start_date: The starting date/time (inclusive) for the estimated position
    records to include in the file. Expected format is YYYY:DOY:SSSSS.
    Defaults to earliest available date if not provided.  Example: 2014:38:357
--end_date: The ending date/time (exclusive) for the estimated position records
    to include in the file. Expected format is YYYY:DOY:SSSSS.  Defaults to the
    latest available date if not provided.
--as_of_date: By default, the currently active records are written to the file.
    If this parameter is provided, it will include the records that were active
    at the requested date/time. Expected format is YYYY:DOY:SSSSS or
    YYYY:MM:DD:HH:MI:SS.
--quiet: Do not write non-error message to stdout to allow for running in a
    cron job.
END_MESSAGE

    print $help_message;
}

################################################################################
# Name: parse_cmd_line_args
#
# Description:
#   Reads in the command line arguments
#
# Arguments:
#   O: help flag reference
#   O: odl filename reference
#   O: output filename reference
#   O: start date reference
#   O: end date reference
#   O: as of date reference
#   O: quiet flag to silence non-error output
#
# Returns:
#   SUCCESS/ERROR
################################################################################
sub parse_cmd_line_args($$$$$$)
{
    my $parse_status;
    my $help = shift;
    my $odl_file = shift;
    my $output_file = shift;
    my $start_date = shift;
    my $end_date = shift;
    my $as_of_date = shift;
    my $quiet = shift;

    $parse_status = GetOptions(
                               "help" => $help,
                               "output_file=s" => $output_file,
                               "start_date=s" => $start_date,
                               "end_date=s" => $end_date,
                               "as_of_date=s" => $as_of_date,
                               "quiet" => $quiet
    );

    # GetOptions returns true when parsing went OK
    if (!$parse_status)
    {
        return ERROR;
    }

    if (defined $$help)
    {
        return SUCCESS;
    }

    # After parsing the arguments with GetOptions, there should be either 0 or
    # 1 command line argument left.  If there is a single argument remaining,
    # it must be a raw odl name
    if (scalar @ARGV < 0 || scalar @ARGV > 1)
    {
        $logger->fatal_msg("Unexpected number of command line parameters");
        return ERROR;
    }
    if (scalar @ARGV == 1)
    {
       $$odl_file = $ARGV[0];
    }

    # Validate command line args
    if (defined $$odl_file)
    {
        if (! -e $$odl_file)
        {
            $logger->fatal_msg("ODL file $$odl_file does not exist");
            return ERROR;
        }

        if (defined $$output_file or defined $$start_date or defined $$end_date
            or defined $$as_of_date)
        {
            $logger->fatal_msg("All parameters should be read from the "
                . "provided parameter file $$odl_file");
            return ERROR;
        }
    }
    else # ODL filename was not provided
    {
        if (!defined $$output_file)
        {
            $logger->fatal_msg("Output filename is required with a parameter "
                . "file");
            return ERROR;
        }

        if (defined $$start_date
            and not is_valid_year_doy_seconds($$start_date))
        {
            $logger->fatal_msg("Invalid start_date: $$start_date");
            return ERROR;
        }
        if (defined $$end_date and not is_valid_year_doy_seconds($$end_date))
        {
            $logger->fatal_msg("Invalid end_date: $$end_date");
            return ERROR;
        }
        if (defined $$as_of_date)
        {
            if (is_valid_conventional_date($$as_of_date))
            {
                # Convert to year_doy_seconds
                my $format = "%Y:%m:%d:%H:%M:%S";
                my $time = Time::Piece->strptime($$as_of_date, $format);
                $$as_of_date = get_year_doy_seconds_from_datetime($time);
            }
            elsif (not is_valid_year_doy_seconds($$as_of_date))
            {
                $logger->fatal_msg("Invalid as_of_date: $$as_of_date");
                return ERROR;
            }
        }
    }

    return SUCCESS;
}

################################################################################
# Name: read_parameters_from_odl
#
# Description:
#   Reads output filename, start date, end date, and determines OMF filename
#   from odl file.
#
# Arguments:
#   I: ODL filename
#   O: Output filename reference
#   O: Start date reference
#   O: End date reference
#   O: OMF filename reference
#
# Returns:
#   SUCCESS/ERROR
################################################################################
sub read_parameters_from_odl($$$$$)
{
    my $odl_filename = shift;
    my $output_filename_ref = shift;
    my $start_date_ref = shift;
    my $end_date_ref = shift;
    my $omf_filename_ref = shift;

    my $odl = ias_odl->new();
    if ($odl->load_odl($odl_filename) != SUCCESS)
    {
        $logger->fatal_msg("Error loading odl $odl_filename");
        return ERROR;
    }

    # Determine group name, and verify there is only one group
    my @odl_group_names = $odl->groups();
    if (scalar @odl_group_names != 1)
    {
        $logger->fatal_msg("Incorrectly formatted ODL $odl_filename. Should " .
            "only have one group");
        return ERROR;
    }
    my $odl_group_name = $odl_group_names[0];

    # Read in output filename
    my $field = "$odl_group_name.ESTIMATED_SSM_POSITION_FILENAME";
    $$output_filename_ref = $odl->get_parameter_value($field);
    if (not defined $$output_filename_ref)
    {
        $logger->fatal_msg("No $field field in $odl_filename");
        return ERROR;
    }

    # Find omf filename
    my $work_order_directory = $odl->get_parameter_value(
        "$odl_group_name.WO_DIRECTORY");
    if (not defined $work_order_directory)
    {
        $work_order_directory = ".";
    }

    my $work_order_id = $odl->get_parameter_value(
        "$odl_group_name.WORK_ORDER_ID");
    if (not defined $work_order_id)
    {
        $logger->fatal_msg("No WORK_ODER_ID field in $odl_filename");
        return ERROR;
    }

    # Load omf
    $$omf_filename_ref = "$work_order_directory/$work_order_id.omf";
    my $omf = ias_odl->new();
    if ($omf->load_odl($$omf_filename_ref) != SUCCESS)
    {
        $logger->fatal_msg("Error loading omf $$omf_filename_ref");
        return ERROR;
    }

    # Determine group name, and there should be only one group
    my @omf_group_names = $omf->groups();
    if (scalar @omf_group_names != 1)
    {
        $logger->fatal_msg("Incorrectly formatted OMF $$omf_filename_ref. " .
            "only have one group");
    }
    my $omf_group_name = $omf_group_names[0];

    # Read acquisition date from the omf. This is formatted like
    # YYYY:DOY:HH:MM:SS.SSSSSSS
    my $acq_date = $omf->get_parameter_value(
        "$omf_group_name.ACQUISITION_DATE");
    if (not defined $acq_date)
    {
        $logger->fatal_msg("No $omf_group_name.ACQUISITION_DATE field in omf " .
            "$$omf_filename_ref");
        return ERROR;
    }

    # Strip any ending fractional seconds off the end of the acquisition date
    # since strptime does not accept fractional seconds
    my $stripped_acq_date = $acq_date;
    $stripped_acq_date =~ s/\.\d*$//;

    # Check for format YYYY:DOY:HH:MM:SS
    if (not $stripped_acq_date =~ /^\d{4}:\d{1,3}:\d{1,2}:\d{1,2}:\d{1,2}$/)
    {
        $logger->fatal_msg("Acquistion date $acq_date not in proper " .
            "YYYY:DOY:HH:MM:SS.SSSSS format");
        return ERROR;
    }

    my $acq_datetime = Time::Piece->strptime($stripped_acq_date,
        "%Y:%j:%H:%M:%S");

    # Start date and end date are 1 week on either side of acquisition date
    my $start_date = $acq_datetime - (7 * ONE_DAY);
    my $end_date = $acq_datetime + (7 * ONE_DAY);

    # We only wish to match based on year and doy, so start date is has 0
    # seconds and the end date is set to maximum seconds per day.
    my $start_year_doy_seconds_format = "%Y:%j:0";
    my $end_year_doy_seconds_format = "%Y:%j:86399";
    $$start_date_ref = $start_date->strftime($start_year_doy_seconds_format);
    $$end_date_ref = $end_date->strftime($end_year_doy_seconds_format);

    return SUCCESS;
}

################################################################################
# Name: update_omf
#
# Description:
#   Adds the output filename to the OMF under the
#   ESTIMATED_SSM_POSITION_FILENAME field. This field will be overwritten if it
#   exists, created if it does not.
#
# Arguments:
#   I: OMF name
#   I: Estimated SSM position filename
#
# Returns:
#   SUCCESS/ERROR
################################################################################
sub update_omf($$)
{
    my $omf_filename = shift;
    my $output_filename = shift;

    # Read all lines
    if (!defined open(OMFIN, "<", $omf_filename))
    {
        $logger->fatal_msg("Cannot open $omf_filename for reading");
        return ERROR;
    }

    my @lines = <OMFIN>;
    close(OMFIN);

    # Write parameter
    if (!defined open(OMFOUT, ">", $omf_filename))
    {
        $logger->fatal_msg("Cannot open $omf_filename for writing");
        return ERROR;
    }

    # Rewrite whole OMF
    foreach my $line (@lines)
    {
        # Watch for end of line parameters
        if ($line =~ /^END_OBJECT/)
        {
            print OMFOUT "  ESTIMATED_SSM_POSITION_FILENAME = " .
                "\"$output_filename\"\n";
        }

        # We don't want to write an old filename
        if ($line !~ /ESTIMATED_SSM_POSITION_FILENAME/)
        {
            print OMFOUT $line;
        }
    }

    if (!close(OMFOUT))
    {
        $logger->fatal_msg("Could not close $omf_filename properly");
        return ERROR;
    }
    return SUCCESS;
}

################################################################################
# Name: select_est_ssm_positions
#
# Description:
#   Selects year, day_of_year, seconds_of_day, encoder_position, and
#   quality_flag from the tirs_ssm_pos_estimates table
#
# Arguments:
#   I: Start date as string formatted like year:doy:sod
#   I: End date as string formatted like year:doy:sod
#   I: As of date formatted like year:doy:sod
#   O: Reference to array which will hold results. Should be empty at start
#
# Returns:
#   SUCCESS/ERROR
################################################################################
sub select_est_ssm_positions($$$$)
{
    my $start_date = shift;
    my $end_date = shift;
    my $as_of_date = shift;
    my $est_positions = shift;

    my @all_est_positions;

    my $select_clause = "select year, day_of_year, seconds_of_day, " .
        "encoder_position, quality_flag";
    my $from_clause = "from tirs_ssm_pos_estimates";

    # The where clause depends on which arguments are defined
    my $where_clause = "where ";
    my @bind_vars;
    if (defined $as_of_date)
    {
        $where_clause .= "(date_added <= to_date(?, 'YYYY:DDD:SSSSS') " .
            "and (date_disabled is null " .
                "or (to_date(?, 'YYYY:DDD:SSSSS') < date_disabled))) ";

        push @bind_vars, $as_of_date;
        push @bind_vars, $as_of_date;
    }
    else
    {
        $where_clause .= "(date_disabled is null) "
    }

    if (defined $start_date)
    {
        my @split_start = split(":", $start_date);
        my $start_year = $split_start[0];
        my $start_doy = $split_start[1];
        $where_clause .= "and (? < year or (? = year and ? <= day_of_year)) ";

        push @bind_vars, $start_year;
        push @bind_vars, $start_year;
        push @bind_vars, $start_doy;
    }

    if (defined $end_date)
    {
        my @split_end = split(":", $end_date);
        my $end_year = $split_end[0];
        my $end_doy = $split_end[1];
        $where_clause .= "and (year < ? or (? = year and day_of_year <= ?)) ";

        push @bind_vars, $end_year;
        push @bind_vars, $end_year;
        push @bind_vars, $end_doy;
    }

    my $order_clause = "order by year, day_of_year, seconds_of_day";

    my $final_query = "$select_clause $from_clause $where_clause $order_clause";

    my ($user_id, $password) = ias_db_GetDBConnectionInfo("L8_OLITIRS");
    my $dbh = ias_db_ConnectToDB($user_id, $password);
    if (!defined($dbh))
    {
        $logger->fatal_msg("Could not connect to database");
        return ERROR;
    }

    my $sth = $dbh->prepare($final_query);
    if (!defined($sth))
    {
        ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
        $logger->fatal_msg("Preparing query");
        return ERROR;
    }

    $sth->execute(@bind_vars);
    while (my @row = $sth->fetchrow())
    {
        push @all_est_positions, \@row;
    }

    $sth->finish();

    ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);

    # Remove any entries falling outside our start end date ranges
    my $start_dt;
    my $end_dt;
    if (defined $start_date)
    {
        $start_dt = convert_year_doy_seconds_to_datetime($start_date);
    }
    if (defined $end_date)
    {
        $end_dt = convert_year_doy_seconds_to_datetime($end_date);
    }
    foreach my $est_position (@all_est_positions)
    {
        my $est_position_date_string = "@$est_position[0]:@$est_position[1]:" .
            "@$est_position[2]";
        my $est_position_dt = convert_year_doy_seconds_to_datetime(
            $est_position_date_string);
        if ((not defined $start_date
                or (defined $start_date and $start_dt <= $est_position_dt))
            and (not defined $end_date
                or (defined $end_date and $end_dt > $est_position_dt)))
        {
            push @$est_positions, $est_position;
        }
    }

    return SUCCESS;
}

################################################################################
# Name: write_file
#
# Description:
#   Writes an array of estimated positions to a file.
#   If the requested start date is not provided, then the timestamp of the first
#   position will be used. Same goes for the end date.
#
# Arguments:
#   I: filename. This will be completely overwritten if it already exists.
#   I: Requested start date as year_doy_seconds
#   I: Requested end date
#   I: Creation date as year_doy_seconds
#   I: Reference to an array of estimated position references. Each index should
#      be an array like "year doy sod est_ssm_pos quality_flag"
#
# Returns:
#   SUCCESS/ERROR
################################################################################
sub write_file($$$$$)
{
    my $filename = shift;
    my $start_date = shift;
    my $end_date = shift;
    my $creation_date = shift;
    my $est_positions = shift;
    my $num_positions = @$est_positions;

    if (not defined $start_date)
    {
        # If positions were sent in
        if (scalar @$est_positions > 0)
        {
            my $first_pos = @$est_positions[0];
            $start_date = "@$first_pos[0]:@$first_pos[1]:@$first_pos[2]";
        }
        else # Set to hardcoded launch date of Feb 11, 2013
        {
            $start_date = "2013:42:0";
        }
    }
    if (not defined $end_date)
    {
        # If positions were sent in
        if (scalar @$est_positions > 0)
        {
            my $last_pos = @$est_positions[$num_positions-1];
            $end_date = "@$last_pos[0]:@$last_pos[1]:@$last_pos[2]";
        }
        else # There are no est positions, so set end date to now
        {
            my $end_date_time = gmtime();
            $end_date = get_year_doy_seconds_from_datetime($end_date_time);
        }
    }

    # Open file for writing
    if (!defined open (OUTFILE, '>', $filename))
    {
        $logger->fatal_msg("Error opening $filename for output");
        return ERROR;
    }

    my $header =
        "FORMAT_VERSION = 1\n" .
        "CREATION_DATE = $creation_date\n" .
        "START_DATE = $start_date\n" .
        "END_DATE = $end_date\n" .
        "NUM_RECORDS = $num_positions\n" .
        "# YEAR DOY SECONDS EST_SSM_POS QUALITY";
    print OUTFILE "$header\n";

    foreach my $est_position (@$est_positions)
    {
        printf OUTFILE "%4s %3s %5s %8s %s\n", @$est_position[0],
            @$est_position[1], @$est_position[2], @$est_position[3],
            @$est_position[4];
    }

    if (!close(OUTFILE))
    {
        $logger->fatal_msg("Closing $filename");
        return ERROR;
    }

    return SUCCESS;
}

################################################################################
# Name: is_valid_year_doy_seconds
#
# Description:
#   Returns true if the string is in format YYYY:DOY:SSSSS. False otherwise.
#   Limits are not checked; only digit counts.
#
# Arguments:
#   I: test string
#
# Returns:
#   true or false
################################################################################
sub is_valid_year_doy_seconds($)
{
    my $year_doy_seconds = shift;
    return $year_doy_seconds =~ /^\d{4}:\d{1,3}:\d{1,5}$/;
}

################################################################################
# Name: is_valid_conventional_date
#
# Description:
#   Returns true if the string is in format YYYY:MM:DD:HH:MM:SS. False otherwise
#   Leading zeroes not required in date. Limits are not checked, only digit
#   counts. No fractions of seconds allowed.
#
# Arguments:
#   I: test string
#
# Returns:
#   true or false
################################################################################
sub is_valid_conventional_date($)
{
    my $conv_date = shift;
    return $conv_date =~ /^\d{4}:\d{1,2}:\d{1,2}:\d{1,2}:\d{1,2}:\d{1,2}$/;
}

################################################################################
# Name: get_year_doy_seconds_from_datetime
#
# Description:
#   Returns a date like YEAR:DOY:SSSSS from a time piece object
#
# Arguments:
#   I: Time piece
#
# Return:
#   String formatted like YEAR:DOY:SSSSS
################################################################################
sub get_year_doy_seconds_from_datetime($)
{
    my $tp = shift;
    my $year = $tp->year;

    # Dates are stored in the database with 1 being Jan 1st, but Time::Piece
    # has 0 as Jan 1st.
    my $doy = $tp->yday + 1;
    my $sod = $tp->sec + $tp->min * 60 + $tp->hour * 3600;

    return "$year:$doy:$sod";
}

################################################################################
# Name: convert_year_doy_seconds_to_datetime
#
# Description:
#   Returns a datetime object
#
# Arguments:
#   I: string formatted like YEAR:DOY:SSSSS
#
# Return:
#   Datetime
################################################################################
sub convert_year_doy_seconds_to_datetime($)
{
    my $year_doy_seconds = shift;

    my $format = "%Y:%j";
    my @arr = split(":", $year_doy_seconds);
    my $year = $arr[0];
    my $doy = $arr[1];
    my $sod = $arr[2];

    my $final_time = Time::Piece->strptime("$year:$doy", $format);
    $final_time = $final_time + $sod;
    return $final_time;
}

