#!/usr/bin/env python3

import unittest
from subprocess import check_call
from datetime import datetime
from ias.db import database as db
from ias.misc import satellite_attributes as sat_attr
from os import system
import filecmp


class TestDumpEstimatedPositions(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        cls.db_conn = cls.connect_to_database()

        cls.est_pos_table_name = "TIRS_SSM_POS_ESTIMATES"
        cls.pos_models_table_name = "TIRS_SSM_POS_MODEL_PARAMS"
        cls.switch_events_table_name = "TIRS_SSM_MODE_SWITCH_EVENTS"
        cls.as_of_date = datetime.strptime(
            "2016-02-18 15:48:52", "%Y-%m-%d %H:%M:%S")

        # Clean up previous failed runs
        cls.clean_up_test_data()

        # Load test data
        cls.load_data(cls.switch_events_table_name)
        cls.load_data(cls.pos_models_table_name)
        cls.load_data(cls.est_pos_table_name)

    @classmethod
    def tearDownClass(cls):
        cls.clean_up_test_data()
        cls.db_conn.close(db.DISCONNECT_COMMIT)

    @classmethod
    def connect_to_database(cls):
        """ Connect to the database using the currently configured
            satellite/sensor schema.  The caller is expected to have properly
            initialized the satellite attributes library prior to calling this
            routine.
        """
        # Get the satellite attributes
        sa = sat_attr.get_default()

        # Connect to the database sensor schema
        db_connection = db.DatabaseConnection()
        db_connection.set_login_using_satellite_sensor(
            sa.get_satellite_name(), sa.get_satellite_sensor_name())
        db_connection.connect()

        # return the DB connection
        return db_connection

    @classmethod
    def load_data(cls, table_name):
        check_call([
            "sqlldr $IAS_DB_L8_OLITIRS control={0} > temp_out.txt".format(
                table_name)], shell=True)

    @classmethod
    def clean_up_test_data(cls):
        # This classmethod is responsible for deleting any test data which may
        # have been loaded by the setup. This basically amounts to anything
        # with a year of <= 2003.

        # Get all switch event and model ids from before 2003
        switch_event_ids = cls.get_pre_2003_switch_event_ids()
        model_ids = cls.get_pre_2003_model_ids()

        # Build queries
        # Delete any estimated ssm positions if their year is less than or
        # equal to 2003
        delete_estimated_positions_query = (
            "delete from {0} where year <= 2003").format(
                cls.est_pos_table_name)

        # Delete any ssm pos model whose switch event id matches those inserted
        # Look in TIRS_SSM_MODE_SWITCH_EVENTS.ldr and
        # TIRS_SSM_POS_MODEL_PARAMS.ldr for these values
        delete_model_query = (
            "delete from {0} where model_id in ({1})").format(
                cls.pos_models_table_name, ", ".join(model_ids))

        # Delete any switch event whose id matches those we found
        delete_switch_events_query = (
            "delete from {0} where switch_event_id in ({1})").format(
                cls.switch_events_table_name, ", ".join(switch_event_ids))

        cursor = cls.db_conn.execute(delete_estimated_positions_query)
        if len(model_ids) > 0:
            cursor = cls.db_conn.execute(delete_model_query)

        if len(switch_event_ids) > 0:
            cursor = cls.db_conn.execute(delete_switch_events_query)

        cursor.close()
        cls.db_conn.commit()

    @classmethod
    def get_pre_2003_switch_event_ids(cls):
        switch_event_id_select_query = (
            "select switch_event_id from {0} where year <= 2003".format(
                cls.switch_events_table_name))
        cursor = cls.db_conn.execute(switch_event_id_select_query)
        switch_event_ids = [str(x[0]) for x in cursor.fetchall()]
        return switch_event_ids

    @classmethod
    def get_pre_2003_model_ids(cls):
        model_ids = []
        switch_event_ids = cls.get_pre_2003_switch_event_ids()

        if len(switch_event_ids) > 0:
            model_id_select_query = (
                "select model_id from {0} where switch_event_id " +
                "in ({1})").format(
                    cls.pos_models_table_name, ", ".join(switch_event_ids))
            cursor = cls.db_conn.execute(model_id_select_query)
            model_ids = [str(x[0]) for x in cursor.fetchall()]
        return model_ids

    def test_invalid_args(self):
        # This test is to ensure invalid combinations of command line arguments
        # never result in a successful run.
        prog = "../dump_estimated_tirs_ssm_positions.pl"

        valid_start_date = "2015:2:1234"
        valid_end_date = "2015:3:1234"
        valid_as_of_date = "2015:02:13:14:45:37"
        valid_output_file = "test_output.txt"
        valid_odl_file = "odl"

        # Each invalid command should redirect its stdout to temp_out.txt.
        # The console is just way to noisy otherwise.
        invalid_command_flags = [
            "",
            "i_dont_exist.odl",
            "invalid_wo_directory.odl",
            "{0} --start_date={1}".format(valid_odl_file, valid_start_date),
            "invalid_wo_id.odl",
            "valid_odl.odl --start_date={0}".format(valid_start_date),
            "valid_odl.odl --output_file={0}".format(valid_output_file),
            "valid_odl.odl --end_date={0}".format(valid_end_date),
            "valid_odl.odl --as_of_date={0}".format(valid_as_of_date),
            "valid_odl.odl --as_of_date={0}".format(valid_as_of_date),
            "--start_date=words",
            "--end_date=words",
            "--as_of_date=words",
        ]

        for invalid_flags in invalid_command_flags:
            command = prog + " " + invalid_flags + "> temp_out.txt"
            error_msg = "{0} should have failed".format(command)
            status = system(command)
            self.assertEqual(status, 256, command)

    def test_valid_commands(self):
        # This test runs valid commands and compares the generated dump file
        # with expected output.
        prog = "../dump_estimated_tirs_ssm_positions.pl"
        outfile = "compare.txt"
        odl_file = "odl"
        start_date = "2003:15:00"
        start_date2 = "2003:14:777"
        end_date = "2003:22:5001"
        end_date2 = "2003:40:6001"
        formatted_as_of_date = self.as_of_date.strftime("%Y:%m:%d:%H:%M:%S")

        # Each command should redirect its stdout to temp_out.txt
        command_flags = [

            # Base command. All non disabled positions should be dumped
            # Start and ending dates of first and last positions should be
            # written.
            # Is an issue with post existing 2003 data.
            "--output_file={0}".format(outfile),

            # Acquisition date from odl. Only non disabled positions within
            # 7 days of this date should be dumped.
            # Not an issue with post 2003 data
            "{0}".format(odl_file),

            # Predetermined start date. All non-disabled estimated
            # positions with timestamps after this date should be dumped.
            # Will be an issue with existing post 2003 data
            "--output_file={0} --start_date={1}".format(outfile, start_date),

            # Predetermined end date. All non-disabled positions with
            # timestamps before this date should be dumped. Should not
            # be an issue with existing post 2003 data.
            "--output_file={0} --end_date={1}".format(outfile, end_date),

            # Start and end date. Only non-disabled positions between these
            # two dates inclusive should be dumped.
            # Not an issue with existing post 2003 data.
            "--output_file={0} --start_date={1} --end_date={2}".format(
                outfile, start_date, end_date),

            # Start and end dates which don't directly equal existing
            # timestamps. The utility should write out the requested start
            # and end times.
            # Not an issue with existing post 2003 data.
            "--output_file={0} --start_date={1} --end_date={2}".format(
                outfile, start_date2, end_date2),

            # As of date. The middle two positions from
            # disable_positions in the setup should be dumped along with
            # all non-disabled positions.
            # Is an issue with existing post 2003 data.
            "--output_file={0} --as_of_date={1}".format(
                outfile, formatted_as_of_date),

            # As of date combined with a date range. All non-disabled
            # positions after 2003 15 should be dumped, and the disabled
            # position at 2003 40 6000 should be dumped. Nothing else.
            # Not an issue with existing post 2003 data.
            (
                "--output_file={0} --as_of_date={1} " +
                "--start_date={2} --end_date={3}").format(
                    outfile, formatted_as_of_date, start_date, end_date2),
        ]

        # List of files containing expected output. These files are expected to
        # have a different creation date than the generated files.
        expected_dir = "./expected_output/"
        expected_files = [
            expected_dir + x + ".txt" for x in [
                "base_command",  # flag
                "acquisition_date",
                "start_date",  # flag
                "end_date",
                "start_end_date",
                "different_requested_date",
                "as_of_date",  # flag
                "as_of_date_with_date_range",
            ]
        ]

        # Some commands will grab existing post-2003 data. True flags
        # correspond to those commands.
        issue_flags = [
            True, False, True, False, False, False, True, False]

        # Run the command to generate a dump file and compare it with its
        # associated expected output.
        error_msg = "Each command should be paired with an expected file"
        self.assertEqual(len(command_flags), len(expected_files), error_msg)
        error_msg = "Each command should be paired with an issue flag"
        self.assertEqual(len(command_flags), len(issue_flags), error_msg)

        # Run each command. If the command doesn't grab post 2003 data, then
        # we can directly compare the files. If it does, then we read all
        # the estimated ssm positions from the expected and generated files,
        # and make sure the expected file is a subset of the generated file.
        for command_flag, exp_file, issue in zip(
                command_flags, expected_files, issue_flags):
            command = prog + " " + command_flag + " > temp_out.txt"
            status = system(command)
            self.assertEqual(status, 0, "Command {0} failed".format(command))

            if not issue:
                error_msg = "{0} does not match {1}".format(outfile, exp_file)
                self.assertTrue(
                    self.compare_files(outfile, exp_file), error_msg)
            else:
                error_msg = "{0} is not a subset of {1}".format(
                    outfile, exp_file)
                self.assertTrue(
                    self.is_subset_file(exp_file, outfile), error_msg)

    def compare_files(self, compare, expected):
        """ We need a separate function to compare the files because we have
            no interest in comparing the creation dates. Of course those will
            be different.
        """
        clines = []
        elines = []
        with open(compare) as c:
            clines = c.readlines()
        with open(expected) as e:
            elines = e.readlines()

        clines = clines[2:]
        elines = elines[2:]

        return (clines == elines)

    def is_subset_file(self, subset, full):
        """ Determines if the each position in the subset file is in full.
        """
        sub_lines = []
        full_lines = []

        with open(subset) as s:
            sub_lines = s.readlines()
        with open(full) as f:
            full_lines = f.readlines()

        sub_lines = sub_lines[6:]
        full_lines = full_lines[6:]

        for sub_line in sub_lines:
            if sub_line not in full_lines:
                return False
        return True

    def test_odl_writing_output_filename_to_omf(self):
        # When the dump script is run, the name of the dump file should be
        # written to the omf once and only once.
        prog = "../dump_estimated_tirs_ssm_positions.pl"
        write_test_dir = "./omf_write_test/"
        odl_filename = write_test_dir + "odl"
        base_omf_filename = write_test_dir + "base.omf"
        final_omf_filename = write_test_dir + "final.omf"
        expected_omf_filename = write_test_dir + "expected.omf"
        run_command = "{0} {1} > temp_out.txt".format(
            prog, odl_filename)

        # Copy the base omf file to final. Base does not have the
        # ESTIMATED_SSM_POSITION_FILENAME field
        status = system("cp {0} {1}".format(
            base_omf_filename, final_omf_filename))
        self.assertEqual(status, 0)

        # Run dump script
        status = system(run_command)
        self.assertEqual(status, 0)

        # The output filename should have been written to the omf once
        error_msg = "{0} and {1} are not equal".format(
            final_omf_filename, expected_omf_filename)
        self.assertTrue(
            filecmp.cmp(final_omf_filename, expected_omf_filename), error_msg)

        # Run dump script again to ensure we don't write twice
        system(run_command)
        self.assertEqual(status, 0)

        self.assertTrue(
            filecmp.cmp(final_omf_filename, expected_omf_filename), error_msg)

        # Remove final omf file
        system("rm {0}".format(final_omf_filename, expected_omf_filename))
        self.assertEqual(status, 0)


if __name__ == '__main__':
    sat_attr.set_default(sat_attr.get_satellite_id_from_number(8))
    unittest.main(verbosity=2)
