#ifndef _FOCAL_PLANE_H_
#define _FOCAL_PLANE_H_

/* Include files for Focal Plane calibration */
#include <limits.h>             /* For PATH_MAX */
#include "ias_geo.h"            /* For OUTSIDE definition */
#include "ias_sensor_model.h"   /* for IAS_LOS_LEGENDRE_TERMS */

#define NUM_TIRS_SCAS 3         /* Maximum number of SCAs for TIRS bands */

#define LOCAL_LOS_LEGENDRE_TERMS 3 /* This defines the number of Legendre
    terms to process at runtime in this application.  It overrides
    IAS_LOS_LEGENDRE_TERMS, but that global value is still used for sizing. */

/* Identifies the processing type for the algorithm run */
typedef enum alignment_processing_type 
{
    OLI_FOCAL_PLANE_ALIGNMENT = 0, /* Focal Plane Alignment */
    TIRS_ALIGNMENT = 1          /* TIRS Alignment */
} ALIGNMENT_PROCESSING_TYPE;

/* Note that the character strings read from the ODL file must provide room 
   for the NULL terminator (i.e., if it's always a 3-character code, you must
   make it char[4] to leave room for the NULL) */
typedef struct alignment_parms 
{
    int path;                   /* WRS path */
    int row;                    /* WRS row */
    char cpf_name[PATH_MAX];    /* Name of CPF */
    char terr_img_name[PATH_MAX]; /* Name of input 1T image */
    char ref_img_name[PATH_MAX]; /* Name of reference image */
    char l0r_img_name[PATH_MAX]; /* Name of L0R image */
    char precision_grid_name[PATH_MAX]; /* Name of input precision grid used
                                   to make terr_img_name */
    char model_file_name[PATH_MAX]; /* Name of input model file used to
                                   make terr_img_name */
    char dem_img_name[PATH_MAX]; /* Name of input DEM image used to make
                                   terr_img_name */
    char doq_img_name[PATH_MAX]; /* Name of input DOQ image used if the  
                                   reference image isn't manually selected */
    float min_corr;             /* Minimum acceptable correlation strength */
    float max_disp;             /* Maximum allowed diagonal displacement from
                                   nominal tiepoint loc to location found by
                                   correlation */
    int corr_win_nl;            /* # of lines in correlation windows */
    int corr_win_ns;            /* # of samples in correlation windows */
    double threshold;           /* Fill threshold percentage */
    int pts_per_cell;           /* # of points per cell to correlate */
    int fit_order;              /* Order of Legendre polynomial fit to perform
                                   perform:  0, 1, or 2 */
    char correlation_file_name[PATH_MAX]; /* Name of output (temp) correlation
                                   file */
    double outlier_tolerance;   /* Tolerance value for outliers */
    int write_cpf_flag;         /* Flag whether to create CPF report */
    char report_file_name[PATH_MAX]; /* Name of output report file */
    int start_effective_date;   /* Starting effective date (YYYYMMDD) for CPF
                                   update file */
    int end_effective_date;     /* Ending effective date (YYYYMMDD) for CPF
                                   update file */
    float point_weight;         /* Weight to assign to each correlated point
                                   in the least squares fit */
    float constraint_weight;    /* Weight used to build constraint matrix */
    IAS_CORRELATION_FIT_TYPE fit_method; /* Correlation fit method */
    char work_order_id[IAS_WORKORDER_ID_SIZE]; /* Work order id */
    int store_in_database_flag; /* Flag to indicate if results should be 
                                   written to database */
    double along_trend_threshold; /* Along track database results metrics 
                                   threshold */
    double across_trend_threshold; /* Across track database results metrics 
                                   threshold */
    IAS_SATELLITE_ID satellite_id; /* Satellite ID */
    int constraint_flag;        /* Solution constraint option: 0 = nominal,
                                   1 = no alignment */
    ALIGNMENT_PROCESSING_TYPE processing_type; /* Algorithm processing type:
                                   OLI_FOCAL_PLANE_ALIGNMENT or TIRS_ALIGNMENT
                                   This is only used in the Setup Phase; 
                                   Legendre Phase code isn't shared */
} ALIGNMENT_PARMS;

/* Identifies whether the point is active or in some other state */
typedef enum active_flag 
{
    ACTIVE_FLAG_START = 0,       /* Start of range to report on */
    ACTIVE = 0,                  /* Point is active, so process it */
    CORR_FAIL = 1,               /* Correlation failed for the point */
    OUTLIER = 2,                 /* Point is an outlier */ 
    ACTIVE_FLAG_END = 2,         /* End of range to report on */
    OUTSIDE_IMAGE = OUTSIDE      /* Point is outside the output image - do not
                                    report on these */
} ACTIVE_FLAG;

/* Information tracked for each correlation point: */
typedef struct corr_pt_info 
{
    int sca_index;               /* SCA index (0-relative) */
    int grid_column;             /* Grid column number (0-relative) */
    int grid_row;                /* Grid row number (0-relative) */
    IAS_DBL_LS nom_os_pt;        /* Nominal output space point */
    IAS_DBL_LS ref_os_pt;        /* Correlated output space ref pt */
    IAS_DBL_LS srch_os_pt;       /* Correlated output space srch pt */
    IAS_DBL_LS ref_is_pt;        /* Ref point in input space, terrain
                                    correction removed */
    IAS_DBL_LS srch_is_pt;       /* Srch point in input space, terrain
                                    correction removed */
    IAS_DBL_LS los_err;          /* Normalized line-of-sight error */
    IAS_DBL_LS los_err_pix;      /* Line-of-sight error in pixels */
    double correlation_accuracy; /* Correlation accuracy */
    ACTIVE_FLAG active_flag;     /* Point active flag */
    double pt_weight;            /* Point weight for use in fit */
    IAS_DBL_LS fit_residual;     /* Residual from fit of pts */
} CORR_PT_INFO;

/* Local functions */
int get_parameters
(
    int argc,                    /* I: Command-line argument count */
    char *argv[],                /* I: Command-line arguments */
    ALIGNMENT_PARMS *parms,      /* O: Parameters retrieved from parameter
                                       file and OMF */
    IAS_CPF **cpf                /* O: CPF structure */
); 

CORR_PT_INFO *read_correlation_info
(
    const char *correlation_file_name, /* I: Pointer to correlation file name */
    int *points_ptr,             /* O: Number of points read */
    int *scas_ptr,               /* O: Number of SCAs in the data */
    int *cells_ptr               /* O: Number of cells across a scan */
);

int write_sca_parameters_cpf
(
    const char *cpf_filename, /* I: CPF filename for writing the group */
    double x_coeff_corr[][IAS_LOS_LEGENDRE_TERMS], /* I: X Legendre
                                    coefficient corrections */
    double y_coeff_corr[][IAS_LOS_LEGENDRE_TERMS], /* I: Y Legendre
                                    coefficient corrections */
    IAS_CPF *cpf_ptr,         /* I: current CPF for the information that
                                    hasn't been updated by this executable */
    IAS_SENSOR_ID sensor      /* I: ID of sensor whose CPF group is written */
);


#endif
