/******************************************************************************
NAME:           get_parameters

PURPOSE:        
Get parameters from the parameter file, OMF, and CPF. 

RETURN VALUE:
Type = int
Value        Description
-----        -----------
ERROR        Error retrieving parameters from parameter file, OMF, and/or CPF
SUCCESS      Parameters retrieved
EXIT_SUCCESS If user asked for help

NOTE: This file is shared by the Setup and Legendre phase Focal Plane 
    Alignment applications.
    The caller must free the CPF structure when done using it.
******************************************************************************/
#include <stdlib.h>
#include <string.h>              /* String library */

#include "ias_logging.h"         /* IAS logging library */
#include "ias_parm.h"            /* Parameter library */
#include "ias_parm_standard_parameters.h" /* Parameter table standard fields */
#include "focal_plane.h"

/* Local types & defines */
#define DEFAULT_PTS_PER_CELL 4   /* Default parameter values */
#define DEFAULT_OUTLIER_TOLERANCE 0.95
#define DEFAULT_CORR_FILENAME "focal_plane.dat"
#define DEFAULT_REPORT_FILENAME "focal_plane.rpt"
#define MINIMUM_CORR_WIN_NL 3
#define MINIMUM_CORR_WIN_NS 8
#define EFFECTIVE_DATE_SIZE 20   /* Size of effective date string */


int get_cpf_effective_dates
(
    IAS_CPF *cpf,             /* I: Current CPF with the date information */
    int* start_effective_date, /* O: Starting effective date (YYYYMMDD) */
    int* end_effective_date   /* O: Ending effective date (YYYYMMDD) */
);

int get_parameters
(
    int argc,                    /* I: Command-line argument count */
    char *argv[],                /* I: Command-line arguments */
    ALIGNMENT_PARMS *parameters, /* O: Parameters retrieved from parameter file
                                     and OMF */
    IAS_CPF **cpf                /* O: CPF structure */
)
{
    int count;                   /* Number of items in parameter list */
    char parm_file[PATH_MAX];    /* Name of the parameter file */
    char omf_file_name[PATH_MAX]; /* Name of OMF */
    int status;                  /* Status returned from functions */
    const struct IAS_CPF_FOCAL_PLANE_CAL *cpf_fp_cal; /* CPF Alignment 
                                    calibration group */
    const struct IAS_CPF_TIRS_ALIGN_CAL *cpf_tirs_align_cal; /* CPF TIRS 
                                    Alignment calibration group */
    int processing_pass = 0;     /* Processing flow pass */
    int satellite_number;        /* Satellite number being processed */


    /* Local buffer between parameter I/O interface and ALIGNMENT_PARMS 
       structure */
    int local_chip_size[2];

    /* Local buffer between parameter I/O interface and 
       ALIGNMENT_PROCESSING_TYPE */
    int processing_type;

    /* Local buffers used to translate between and parameter I/O interface */
    double local_min_corr;
    double local_max_disp;

    /* Make sure a parameter file was specified */
    if (argc < 2)
    {
        IAS_LOG_ERROR("Usage: %s parameter_file", argv[0]);
        return ERROR;
    }

    strcpy(parm_file, argv[1]);

    /* Make sure any optional filename parameters are initialized */
    memset(parameters, 0, sizeof(*parameters));

    /* Start parameter table definition */
    IAS_PARM_DECLARE_TABLE(parms, 17);

    /* Get the work order ID */
    IAS_PARM_WORK_ORDER_ID(parms, parameters->work_order_id, 
            sizeof(parameters->work_order_id), 0); 

    /* Get the name of the REFERENCE_IMAGE file */
    IAS_PARM_ADD_STRING(parms, REFERENCE_IMAGE, 
            "Reference image to correlate to", IAS_PARM_OPTIONAL, 0, NULL, 0,
            NULL, parameters->ref_img_name, sizeof(parameters->ref_img_name),
            0);

    /* Get the name of the correlation file */ 
    const char *default_correlation_file_name[] = {DEFAULT_CORR_FILENAME};
    IAS_PARM_ADD_STRING(parms, CORRELATION_FILENAME, "Name of output "
            "correlation data file", IAS_PARM_OPTIONAL, 0, NULL, 1, 
            default_correlation_file_name, parameters->correlation_file_name, 
            sizeof(parameters->correlation_file_name), 1);

    /* Get the name of the report file */ 
    const char *default_report_file_name[] = {DEFAULT_REPORT_FILENAME};
    IAS_PARM_ADD_STRING(parms, REPORT_FILENAME, "Name of output report file", 
            IAS_PARM_OPTIONAL, 0, NULL, 1, default_report_file_name, 
            parameters->report_file_name, sizeof(parameters->report_file_name),
            1);

    /* Get the number of tie points per grid cell */
    int default_pts_per_cell[] = {DEFAULT_PTS_PER_CELL};
    IAS_PARM_ADD_INT(parms, PTS_PER_CELL, "Points to correlate per grid cell",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 0, 50, 1,
            default_pts_per_cell, &parameters->pts_per_cell, 
            sizeof(parameters->pts_per_cell), 0);

    /* Get the confidence level */
    double default_outlier_tolerance[] = {DEFAULT_OUTLIER_TOLERANCE};
    IAS_PARM_ADD_DOUBLE(parms, OUTLIER_TOLERANCE, "Outlier confidence "
            "threshold", IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 0.0, 
            1.0, 1, default_outlier_tolerance, &parameters->outlier_tolerance, 
            sizeof(parameters->outlier_tolerance), 0);

    /* Get the flag to write (or not write) the CPF group */
    int default_write_cpf_flag[] = {FALSE};
    IAS_PARM_ADD_INT(parms, CPF_GROUP_FLAG, "CPF Group output switch",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 0, 1, 1,
            default_write_cpf_flag, &parameters->write_cpf_flag, 
            sizeof(parameters->write_cpf_flag), 0);

    /* Get the flag to write (or not write) data to the database */
    IAS_PARM_STORE_CHARACTERIZATION_IN_DB(parms, IAS_PARM_OPTIONAL,
            &parameters->store_in_database_flag, 
            sizeof(parameters->store_in_database_flag), 0);

    /* Get the minimum correlation value */
    IAS_PARM_MIN_CORR(parms, &local_min_corr, sizeof(local_min_corr));

    /* Get the maximum displacement value */
    IAS_PARM_MAX_DISP(parms, &local_max_disp, sizeof(local_max_disp));

    /* Get the valid pixel percentage threshold value */
    IAS_PARM_THRESHOLD(parms, &parameters->threshold,
            sizeof(parameters->threshold));

    /* Get the chip size values */
    IAS_PARM_CHIP_SIZE(parms, local_chip_size, sizeof(local_chip_size));

    /* Get the effective start date */
    IAS_PARM_ADD_INT(parms, START_EFFECTIVE_DATE, "Effective start date for "
            "CPF group output file, expected format YYYYMMDD",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 0, 0, 0, 0, NULL, 
            &parameters->start_effective_date, 
            sizeof(parameters->start_effective_date), 0);

    /* Get the effective end date */
    IAS_PARM_ADD_INT(parms, END_EFFECTIVE_DATE, "Effective end date for CPF "
            "group output file, expected format YYYYMMDD", IAS_PARM_OPTIONAL, 
            IAS_PARM_NOT_ARRAY, 0, 0, 0, 0, NULL,
            &parameters->end_effective_date, 
            sizeof(parameters->end_effective_date), 0);

    /* Get the processing type - Focal Plane Alignment or TIRS Alignment */
    int default_processing_type[] = {OLI_FOCAL_PLANE_ALIGNMENT};
    IAS_PARM_ADD_INT(parms, PROCESSING_TYPE, "Type of algorithm processing: "
            "0 - Focal Plane Alignment; 1 - TIRS Alignment", IAS_PARM_REQUIRED,
            IAS_PARM_NOT_ARRAY, 1, 0, 1, 1, default_processing_type,
            &processing_type, sizeof(processing_type), 0);

    /* Get processing pass */
    IAS_PARM_PROCESSING_PASS(parms, &processing_pass, sizeof(processing_pass),
            0);

    /* Get the Legendre coefficient constraint type (used only for TIRS 
       Alignment) */
    int default_constraint_flag[] = {0};
    IAS_PARM_ADD_INT(parms, CONSTRAINT_FLAG, "Legendre coefficient constraint "
            "flag: 0 - constrain Legendre coefficients (solve all parameters); "
            "1 - constrain angles (solve Legendre only)", IAS_PARM_OPTIONAL,
            IAS_PARM_NOT_ARRAY, 1, 0, 1, 1, default_constraint_flag,
            &parameters->constraint_flag, sizeof(parameters->constraint_flag),
            0);

    /* Read and check the WO parameter file */

    /* Calculate the number of items in the WO parameter file */
    count = IAS_PARM_GET_TABLE_SIZE(parms);

    /* Call the ias_parm_provide_help routine to provide help for the
       parameters if requested; also verifies the ODL file exists */
    status = ias_parm_provide_help(parm_file, parms, count,
            IAS_INPUT_PARAMETERS);
    if (status)
    {
        /* If the status of the provide help is 1 the user got help.  Since
           the parameter file name contains the help request, copy it to the
           OMF file name so the next call will ask for the same help */
        if (status == 1)
        {
            strcpy(omf_file_name, parm_file);
        }
        else if (status == ERROR) /* Help failed */
        {
            IAS_LOG_ERROR("Providing help using application parameter file %s",
                    parm_file);
            exit(EXIT_FAILURE);
        }
    }
    else
    {
        /* Read the parameter file */
        if (ias_parm_read(parm_file, parms, count) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading application parameter file: %s", parm_file);
            return ERROR;
        }

        /* Transfer the various local values to the temporary structure */
        parameters->processing_type = processing_type;
        parameters->min_corr = (float)local_min_corr;
        parameters->max_disp = (float)local_max_disp;
        parameters->corr_win_nl = local_chip_size[0];
        parameters->corr_win_ns = local_chip_size[1];

        /* Build the OMF file name */
        sprintf(omf_file_name, "%s.omf", parameters->work_order_id);
    }

    /* Start OMF table definition */
    IAS_PARM_DECLARE_TABLE(omf_parms, 10);

    /* Get the name of the Calibration parameter file - don't use the
       standard_parms version because that one isn't REQUIRED */
    IAS_PARM_CPF(omf_parms, IAS_PARM_REQUIRED, parameters->cpf_name, 
            sizeof(parameters->cpf_name), 1);

    /* Get the L0R filename */
    IAS_PARM_L0R_FILENAME(omf_parms, IAS_PARM_OPTIONAL,
            parameters->l0r_img_name, sizeof(parameters->l0r_img_name), 1);

    /* Get the name of the DEM file */
    IAS_PARM_DEM_FILENAME_FROM_PASS(omf_parms, IAS_PARM_OPTIONAL,
            parameters->dem_img_name, sizeof(parameters->dem_img_name), 1, 
            processing_pass);

    /* Get the name of the DOQ file */
    IAS_PARM_DOQ_FILENAME(omf_parms, parameters->doq_img_name, 
            sizeof(parameters->doq_img_name), 1);

    /* Get the name of the L1G file */
    IAS_PARM_L1G_FILENAME_FROM_PASS(omf_parms, IAS_PARM_OPTIONAL,
            parameters->terr_img_name, sizeof(parameters->terr_img_name), 1,
            processing_pass);

    /* Get the name of the grid file */
    IAS_PARM_GRID_FILENAME_FROM_PASS(omf_parms, IAS_PARM_OPTIONAL, 
            parameters->precision_grid_name, 
            sizeof(parameters->precision_grid_name), 1, processing_pass);

    /* Get the name of the model file */
    IAS_PARM_MODEL_FILENAME_FROM_PASS(omf_parms, IAS_PARM_OPTIONAL, 
            parameters->model_file_name, sizeof(parameters->model_file_name), 1,
            processing_pass);

    /* Get the WRS path */
    IAS_PARM_TARGET_WRS_PATH(omf_parms, IAS_PARM_OPTIONAL, &parameters->path, 
            sizeof(parameters->path), 1);

    /* Get the WRS row */
    IAS_PARM_TARGET_WRS_ROW(omf_parms, IAS_PARM_OPTIONAL, &parameters->row, 
            sizeof(parameters->row), 1);

    /* Get the satellite */
    IAS_PARM_SATELLITE(omf_parms, &satellite_number, sizeof(satellite_number), 
            1);

    /* Read and check the OMF */

    /* See if the user wanted help or a template with the OMF file */
    status = ias_parm_provide_help(omf_file_name, omf_parms,
            IAS_PARM_GET_TABLE_SIZE(omf_parms), IAS_OMF_PARAMETERS);
    if (status)
    {
        if (status == 1) /* Help was successful */
        {
            exit(EXIT_SUCCESS);
        }
        else if (status == ERROR) /* Help failed */
        {
            IAS_LOG_ERROR("Providing help using OMF %s", omf_file_name);
            exit(EXIT_FAILURE);
        }
    }

    /* Read the OMF file */
    count = IAS_PARM_GET_TABLE_SIZE(omf_parms);
    if (ias_parm_read(omf_file_name, omf_parms, count) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from OMF file: %s", omf_file_name);
        return ERROR;
    }

    /* Use the DOQ if REFERENCE_IMAGE is not set. */
    if (IAS_PARM_GET_COUNT_READ(parms, REFERENCE_IMAGE) == 0)
    {
        if (IAS_PARM_GET_COUNT_READ(omf_parms, DOQ_FILENAME) == 1)
        {
            strcpy(parameters->ref_img_name, parameters->doq_img_name);
        }
        else
        {
            IAS_LOG_ERROR("The reference image could not be defined because "
                    "neither the REFERENCE_IMAGE nor the DOQ_FILENAME were "
                    "present.");
            return ERROR;
        }
    }

    /* Get the satellite ID */
    parameters->satellite_id = 
        ias_sat_attr_get_satellite_id_from_satellite_number(satellite_number); 
    if (parameters->satellite_id == ERROR)
    {
        IAS_LOG_ERROR("Unable to determine satellite ID from satellite "
                "number %d", satellite_number);
    }

    /* Initialize satellite attributes */
    status = ias_sat_attr_initialize(parameters->satellite_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attributes library");
        return ERROR;
    }

    /* Open the CPF */
    *cpf = ias_cpf_read(parameters->cpf_name);
    if (!*cpf)
    {
        IAS_LOG_ERROR("Cannot open CPF file %s for reading",
                parameters->cpf_name);
        return ERROR;
    }

    /* If updated CPF values will be written to file, the effective dates are
       needed */ 
    if (parameters->write_cpf_flag)
    {
        /* If the effective date parameters were not read from the appplication
           parameter file, read them from the CPF */
        if ((IAS_PARM_GET_COUNT_READ(parms, START_EFFECTIVE_DATE) == 0) ||
                (IAS_PARM_GET_COUNT_READ(parms, END_EFFECTIVE_DATE) == 0))
        {
            status = get_cpf_effective_dates(*cpf, 
                    &parameters->start_effective_date,
                    &parameters->end_effective_date);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Getting effective start and stop dates from "
                        "the CPF");
                ias_cpf_free(*cpf);
                *cpf = NULL;
                return ERROR;
            }
        }

        /* Check the start and end effective dates, whether read from the CPF
           or from the application parameter file */
        if (ias_geo_check_start_end_date(parameters->start_effective_date,
                    parameters->end_effective_date) != SUCCESS)
        {
            IAS_LOG_ERROR("In start effective date %d or end effective date %d",
                    parameters->start_effective_date, 
                    parameters->end_effective_date);
            ias_cpf_free(*cpf);
            *cpf = NULL;
            return ERROR;
        }
    }

    /* Get the CPF groups needed */
    if (parameters->processing_type == OLI_FOCAL_PLANE_ALIGNMENT)
    {
        cpf_fp_cal = ias_cpf_get_focal_plane_cal(*cpf);
        if (!cpf_fp_cal)
        {
            IAS_LOG_ERROR("Getting Focal Plane Calibration group from CPF");
            ias_cpf_free(*cpf);
            *cpf = NULL;
            return ERROR;
        }

        /* Retrieve values from the CPF as needed */
        parameters->fit_order = cpf_fp_cal->fit_order;
        parameters->point_weight = cpf_fp_cal->tie_point_weight;
        parameters->fit_method = cpf_fp_cal->corr_fit_method;
        parameters->along_trend_threshold = 
            cpf_fp_cal->along_postfit_rmse_threshold;
        parameters->across_trend_threshold = 
            cpf_fp_cal->across_postfit_rmse_threshold;
        if (IAS_PARM_GET_COUNT_READ(parms, CHIP_SIZE) < 1)
        {
            parameters->corr_win_nl = cpf_fp_cal->corr_window_size[0];
        }
        if (IAS_PARM_GET_COUNT_READ(parms, CHIP_SIZE) < 2)
        {
            parameters->corr_win_ns = cpf_fp_cal->corr_window_size[1];
        }
        if (IAS_PARM_GET_COUNT_READ(parms, MIN_CORR) == 0)
        {
            parameters->min_corr = cpf_fp_cal->min_corr_strength;
        }
        if (IAS_PARM_GET_COUNT_READ(parms, MAX_DISP) == 0)
        {
            parameters->max_disp = cpf_fp_cal->max_displacement_offset;
        }
        if (IAS_PARM_GET_COUNT_READ(parms, THRESHOLD) == 0)
        {
            parameters->threshold = cpf_fp_cal->fill_threshold;
        }

        /* Check for a valid fit order value */
        if ((parameters->fit_order < 0) || (parameters->fit_order > 2))
        {
            IAS_LOG_ERROR("Invalid fit order specified--must be 0, 1, or 2");
            ias_cpf_free(*cpf);
            *cpf = NULL;
            return ERROR;
        }
    }
    else /* processing_type == TIRS_ALIGNMENT */
    {
        cpf_tirs_align_cal = ias_cpf_get_tirs_align_cal(*cpf);
        if (!cpf_tirs_align_cal)
        {
            IAS_LOG_ERROR("Getting TIRS Alignment Calibration group from CPF");
            ias_cpf_free(*cpf);
            *cpf = NULL;
            return ERROR;
        }

        /* Retrieve values from the CPF as needed */
        parameters->fit_order = cpf_tirs_align_cal->fit_order;
        parameters->point_weight = cpf_tirs_align_cal->tie_point_weight;
        parameters->fit_method = cpf_tirs_align_cal->corr_fit_method;
        parameters->constraint_weight = 
            cpf_tirs_align_cal->align_constraint_weight;
        parameters->along_trend_threshold =
            cpf_tirs_align_cal->along_postfit_rmse_threshold;
        parameters->across_trend_threshold =
            cpf_tirs_align_cal->across_postfit_rmse_threshold;
        if (IAS_PARM_GET_COUNT_READ(parms, CHIP_SIZE) < 1)
        {
            parameters->corr_win_nl = cpf_tirs_align_cal->corr_window_size[0];
        }
        if (IAS_PARM_GET_COUNT_READ(parms, CHIP_SIZE) < 2)
        {
            parameters->corr_win_ns = cpf_tirs_align_cal->corr_window_size[1];
        }
        if (IAS_PARM_GET_COUNT_READ(parms, MIN_CORR) == 0)
        {
            parameters->min_corr = cpf_tirs_align_cal->min_corr_strength;
        }
        if (IAS_PARM_GET_COUNT_READ(parms, MAX_DISP) == 0)
        {
            parameters->max_disp = cpf_tirs_align_cal->max_displacement_offset;
        }
        if (IAS_PARM_GET_COUNT_READ(parms, THRESHOLD) == 0)
        {
            parameters->threshold = cpf_tirs_align_cal->fill_threshold;
        }

        /* Check for a valid fit order value */
        if ((parameters->fit_order < 0) || (parameters->fit_order > 3))
        {
            IAS_LOG_ERROR("Invalid fit order specified--must be 0, 1, 2, or 3");
            ias_cpf_free(*cpf);
            *cpf = NULL;
            return ERROR;
        }
    }

    /* Check the weight to assign to each point in the least squares fit */
    if (parameters->point_weight <= 0.0)
    {
        IAS_LOG_ERROR("Invalid tie point weight");
        ias_cpf_free(*cpf);
        *cpf = NULL;
        return ERROR;
    }

    /* Check for valid fit method */
    if ((parameters->fit_method < IAS_FIT_ELLIP_PARA) || 
        (parameters->fit_method > IAS_FIT_ROUND))
    {
        IAS_LOG_ERROR("Invalid correlation fit method in CPF");
        ias_cpf_free(*cpf);
        *cpf = NULL;
        return ERROR;
    }

    /* Make sure the window size is large enough */
    if (parameters->corr_win_nl < MINIMUM_CORR_WIN_NS) 
    {
        IAS_LOG_ERROR("CORR_WIN_NL must be at least %d; it is %d",
                MINIMUM_CORR_WIN_NL, parameters->corr_win_nl); 
        ias_cpf_free(*cpf);
        *cpf = NULL;
        return ERROR;
    }

    if (parameters->corr_win_ns < MINIMUM_CORR_WIN_NS) 
    {
        IAS_LOG_ERROR("CORR_WIN_NS must be at least %d; it is %d",
                MINIMUM_CORR_WIN_NS, parameters->corr_win_ns); 
        ias_cpf_free(*cpf);
        *cpf = NULL;
        return ERROR;
    }


    return SUCCESS;
}


/******************************************************************************
NAME:           get_cpf_effective_dates 

PURPOSE:
Get the start and end effective dates from the CPF

RETURNS:
Type = int
Value           Description
-----           -----------
ERROR           Error getting the dates from the CPF
SUCCESS         Dates successfully read 
******************************************************************************/

int get_cpf_effective_dates 
(
    IAS_CPF *cpf,           /* I: Current CPF with the date information */ 
    int *start_effective_date, /* O: Starting effective date (YYYYMMDD) */
    int *end_effective_date /* O: Ending effective date (YYYYMMDD) */
)
{
    const struct IAS_CPF_FILE_ATTRIBUTES *cpf_file_atts; /* CPF File Attributes
                                                            group */
    int year;               /* Year extracted from effective date */
    int month;              /* Month extracted from effective date */
    int day;                /* Day extracted from effective date */
    int status;             /* Status returned by called routines */

    /* Get the file attributes group */
    cpf_file_atts = ias_cpf_get_file_attributes(cpf);
    if (!cpf_file_atts)
    {
        IAS_LOG_ERROR("Getting File Attributes group from CPF");
        return ERROR;
    }

    /* Format the effective dates and put them in integers.  They are
       in form YYYY-MM-DDTHH:MM:SS.  We want YYYYMMDD */
    status = sscanf(cpf_file_atts->effective_date_begin, "%4d-%2d-%2d", &year, 
            &month, &day);
    if (status != 3)
    {
        IAS_LOG_ERROR("Could not scan start effective date from string %s",
                cpf_file_atts->effective_date_begin);
        return ERROR;
    }
    *start_effective_date = 10000 * year + 100 * month + day;

    status = sscanf(cpf_file_atts->effective_date_end, "%4d-%2d-%2d", &year, 
            &month, &day);
    if (status != 3)
    {
        IAS_LOG_ERROR("Could not scan end effective date from string %s",
                cpf_file_atts->effective_date_end);
        return ERROR;
    }
    *end_effective_date = 10000 * year + 100 * month + day;

    return SUCCESS;
}

