/******************************************************************************
NAME:           create_focal_plane_report

PURPOSE:
create_focal_plane_report generates a report of the results of the 
SCA Legendre polynomial fit calculations.  The report is written to a file.

RETURNS:
Type = int
Value           Description
-----           -----------
ERROR           Error generating the report
SUCCESS         Report successfully generated

ALGORITHM REFERENCES:
OLI Focal Plane Alignment Calibration ADD
******************************************************************************/

#include <string.h>               /* String library */
#include <stdio.h>                /* fprintf prototype */

#include "ias_miscellaneous.h"
#include "ias_logging.h"          /* IAS Logging library */
#include "create_focal_plane_report.h" /* Prototype for this routine */

#define UNKNOWN_IDX 4             /* Unknown state message string index */
#define SCA_MSG_SIZE 7            /* SCA message size; format SCA##n */
#define NUM_STATE_MSGS 5          /* Number of point state messages */

/* Local prototype */
static int write_coeffs
(
    FILE *file_ptr,               /* I: Pointer to file to write */
    const char *title_ptr,        /* I: Title for these coefficients */
    const int sca_num,            /* I: SCA number */
    const double *coeffs          /* I: Set of coefficients to write */
);

int create_focal_plane_report
(
    const char *report_file_name, /* I: Report file name */
    const char *terr_image_name,  /* I: Name of input 1T image */
    const char *ref_image_name,   /* I: Name of reference image */
    const char *l0r_image_name,   /* I: Name of L0R image */
    int nscas,                    /* I: Number of SCAs */
    double sca_coef_x[IAS_MAX_NSCAS][IAS_LOS_LEGENDRE_TERMS], /* I: Existing X 
                                        Legendre coefficients */
    double sca_coef_y[IAS_MAX_NSCAS][IAS_LOS_LEGENDRE_TERMS], /* I: Existing Y 
                                        Legendre coefficients */
    double error_legendre_x[IAS_MAX_NSCAS][IAS_LOS_LEGENDRE_TERMS], 
                                  /* I: Error X Legendre coefficients */
    double error_legendre_y[IAS_MAX_NSCAS][IAS_LOS_LEGENDRE_TERMS], 
                                  /* I: Error Y Legendre coefficients */
    double updated_legendre_x[IAS_MAX_NSCAS][IAS_LOS_LEGENDRE_TERMS], 
                                  /* I: Updated X coefficients */
    double updated_legendre_y[IAS_MAX_NSCAS][IAS_LOS_LEGENDRE_TERMS], 
                                  /* I: Updated Y coefficients */
    CORR_PT_INFO *fp_corr_pts_ptr,/* I: Focal plane correlation points */
    int points,                   /* I: Number of focal plane correlation 
                                        points */
    const POST_FIT_RESIDUAL_STATS_TYPE *residuals_ptr, /* I: Calculated 
                                        residual information */
    float confidence_level,       /* I: Confidence level used for outlier 
                                        rejection */
    int fit_order,                /* I: Flag that indicates the polynomial 
                                        order that was used in the solution */
    const char *work_order_id     /* I: Work Order ID for report header */
)
{
    FILE *report_file;            /* Pointer to the report file */
    int sca_index;                /* SCA index */
    int index;                    /* Loop counter */
    int state_index;              /* Index into state message */
    const char *state_msg[NUM_STATE_MSGS] = {"Active", "CorrFail", "Outlier", 
                "Outside", "Unknown"}; /* Message for active flag */
    char sca_msg[SCA_MSG_SIZE];   /* Message for SCA index */
    IAS_REPORT_HEADER *header = NULL; /* Report header structure */
    int status;                   /* Status returned by called routines */

    /* Open the report file for writing */
    report_file = fopen(report_file_name, "w");
    if (report_file == NULL)
    {
        IAS_LOG_ERROR("Opening the report file - %s", report_file_name);
        return ERROR;
    }

    /* Output the standard header */
    header = ias_misc_initialize_gps_report_header(terr_image_name,
            l0r_image_name, work_order_id, "FOCAL_PLANE");
    if (!header)
    {
        IAS_LOG_ERROR("Initializing GPS header for report %s", 
            report_file_name);
        fclose(report_file);
        return ERROR;
    }

    status = ias_misc_write_gps_report_header(report_file, header);
    ias_misc_free_report_header(header);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing GPS header to report file %s",
            report_file_name);
        fclose(report_file);
        return ERROR;
    }
    header = NULL;

    /* Write the name of the input 1T image to the report file */    
    status = fprintf(report_file, "Level 1T image: %s\n", terr_image_name);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    /* Write the name of the reference image to the report file */
    status = fprintf(report_file, "DOQ Reference Image: %s\n", ref_image_name);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    for (sca_index = 0; sca_index < nscas; sca_index++)
    {
        /* Write the existing coefficients to the report file */
        status = write_coeffs(report_file,
            "SCA %d Existing X Legendre coeffs:  ", sca_index + 1, 
            sca_coef_x[sca_index]);
        if (status < 0)
        {
            /* Error returned from fprintf */
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            fclose(report_file);
            return ERROR;
        }
     
        status = write_coeffs(report_file,
            "SCA %d Existing Y Legendre coeffs:  ", sca_index + 1, 
            sca_coef_y[sca_index]); 
        if (status < 0)
        {
            /* Error returned from fprintf */
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            fclose(report_file);
            return ERROR;
        }
     
        /* Write the Legendre error coefficients to the report file */
        status = write_coeffs(report_file,
            "SCA %d Error Legendre X coeffs:     ", sca_index + 1, 
            error_legendre_x[sca_index]); 
        if (status < 0)
        {
            /* Error returned from fprintf */
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            fclose(report_file);
            return ERROR;
        }
        
        status = write_coeffs(report_file,
            "SCA %d Error Legendre Y coeffs:     ", sca_index + 1, 
            error_legendre_y[sca_index]);
        if (status < 0)
        {
            /* Error returned from fprintf */
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            fclose(report_file);
            return ERROR;
        }
    
        /* Write the updated legendre coefficients to the report file */
        status = write_coeffs(report_file,
            "SCA %d Updated Legendre X coeffs:   ", sca_index + 1, 
            updated_legendre_x[sca_index]); 
        if (status < 0)
        {
            /* Error returned from fprintf */
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            fclose(report_file);
            return ERROR;
        }

        status = write_coeffs(report_file,
            "SCA %d Updated Legendre Y coeffs:   ", sca_index + 1, 
            updated_legendre_y[sca_index]);
        if (status < 0)
        {
            /* Error returned from fprintf */
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            fclose(report_file);
            return ERROR;
        }
    }

    /* Write the residual statistics information to the report file */
    status = fprintf(report_file,"\n"
                     "                             Pre-fit            "
                     "             Post-fit\n"
                     "                         (microradians)         "
                     "          (microradians)\n"
                     "            Mean      RMSE     StdDev       Mean"
                     "      RMSE     StdDev\n");
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    for (sca_index = 0; sca_index < nscas; sca_index++)
    {
        status = fprintf(report_file,"SCA%02d X: % f % f % f    % f % f % f\n",
            sca_index + 1, residuals_ptr[sca_index].delta_x.mean * 1E6,
            residuals_ptr[sca_index].delta_x.rms * 1E6,
            residuals_ptr[sca_index].delta_x.stdev * 1E6,
            residuals_ptr[sca_index].residual_x.mean * 1E6,
            residuals_ptr[sca_index].residual_x.rms * 1E6,
            residuals_ptr[sca_index].residual_x.stdev * 1E6);
        if (status < 0)
        {
            /* Error returned from fprintf */
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            fclose(report_file);
            return ERROR;
        }

        status = fprintf(report_file,"SCA%02d Y: % f % f % f    % f % f % f\n",
            sca_index + 1, residuals_ptr[sca_index].delta_y.mean * 1E6,
            residuals_ptr[sca_index].delta_y.rms * 1E6,
            residuals_ptr[sca_index].delta_y.stdev * 1E6,
            residuals_ptr[sca_index].residual_y.mean * 1E6,
            residuals_ptr[sca_index].residual_y.rms * 1E6,
            residuals_ptr[sca_index].residual_y.stdev * 1E6);

        if (status < 0)
        {
            /* Error returned from fprintf */
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            fclose(report_file);
            return ERROR;
        }
    }

    /* Write the confidence level, fit order,  and the number of points 
       used for the statistics to the report file */
    status = fprintf(report_file,"\nConfidence Level: %f\nFit Order: %d\n",
        confidence_level, fit_order);

    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    for (sca_index = 0; sca_index < nscas; sca_index++)
    {
        status = fprintf(report_file,"%d points used to fit SCA%02d\n",
            residuals_ptr[sca_index].points, sca_index + 1);
        if (status < 0)
        {
            /* Error returned from fprintf */
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            fclose(report_file);
            return ERROR;
        }
    }

    /* Write the correlation information for each point to the report file */
    if (fprintf(report_file,"\nFocal Plane Calibration correlation points:\n") 
        < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    /* Write the correlation information headers to the report file */
    if (fprintf(report_file,
        "                    Nominal  Nominal   Pixel    Pixel"
        "    Angle    Angle             Angle     Angle \n"
        "                    Output   Output    Error    Error"
        "    Error    Error            Fit Res   Fit Res\n"
        "SCA   Chip Col       Line     Samp     Line     Samp "
        "    Line     Samp    State     Line      Samp\n"
        "_____ ____ ___     ________ ________ ________ ________"
        " ________ ________ ________ _________ _________\n") < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    for (index = 0; index < points; index++)
    {
        /* Get the current active flag and make sure it points to a legal 
           state_msg entry */
        state_index = fp_corr_pts_ptr[index].active_flag;
        if ((state_index < ACTIVE_FLAG_START) || 
            (state_index > ACTIVE_FLAG_END))
        {
            /* Skip points outside the image or with unknown state */
            continue;
        }

        sca_index = fp_corr_pts_ptr[index].sca_index;
        sprintf(sca_msg, "SCA%02d", sca_index + 1); 
        if ((sca_index < 0) || (sca_index > (nscas - 1)))
        {
            strcpy(sca_msg, "");  /* Point to empty direction */
        }

        status = fprintf(report_file, "%5s %2d %4d      %8.2f %8.2f %8.2f "
            "%8.2f %8.2f %8.2f %8s % 8.6f % 8.6f\n", sca_msg,
            fp_corr_pts_ptr[index].sca_index,
            fp_corr_pts_ptr[index].grid_column,
            fp_corr_pts_ptr[index].nom_os_pt.line,
            fp_corr_pts_ptr[index].nom_os_pt.samp,
            fp_corr_pts_ptr[index].los_err_pix.line,
            fp_corr_pts_ptr[index].los_err_pix.samp,
            fp_corr_pts_ptr[index].los_err.line*1E6,
            fp_corr_pts_ptr[index].los_err.samp*1E6,
            state_msg[state_index],
            fp_corr_pts_ptr[index].fit_residual.line*1E6,
            fp_corr_pts_ptr[index].fit_residual.samp*1E6);
        if (status < 0)
        {
            /* Error returned from fprintf */
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            fclose(report_file);
            return ERROR;
        }
    }

    if (fclose(report_file) != 0)
    {
        /* Error returned from fclose */
        IAS_LOG_ERROR("Closing report file %s", report_file_name);
        return ERROR;
    }

    return SUCCESS;
}

/******************************************************************************
NAME:           write_coeffs

PURPOSE: Writes an entire set of coefficients to the indicated output file.

RETURNS:
Type = int
Value              Description
-----              -----------
   <0              Error writing coefficients
    0 or greater   Coefficients successfully written

ALGORITHM REFERENCES:
OLI Focal Plane Alignment Calibration ADD
******************************************************************************/
/* routine to write out an entire set of coefficients */
static int write_coeffs
(
    FILE *file_ptr,               /* I: Pointer to file to write     */
    const char *title_ptr,        /* I: Title for these coefficients */
    const int sca_number,         /* I: SCA number                   */
    const double *coeffs          /* I: Set of coefficients to write */
)
{
    int status;                   /* Return status                   */

    /* Write the existing mirror coefficients to the report file */
    status = fprintf(file_ptr, title_ptr, sca_number);
    if (status < 0)
    {
        return status;
    }

    /* Write the coefficients */
    status = fprintf(file_ptr, "% 9.6E % 9.6E % 9.6E\n", coeffs[0], coeffs[1], 
        coeffs[2]);

    return status;
}
