/******************************************************************************
NAME:           filter_outliers

PURPOSE:
filter_outliers separates the focal plane correlation data into groups for
each SCA for the X (sample) and Y (line) directions.  It then finds the 
standard deviation for the points in each group.  Outlier rejection is then
performed on the points based on the tolerance selected by the user and the
Student's T distribution.

RETURNS:
Type = int
Value           Description
-----           -----------
ERROR           Error filtering the outliers
SUCCESS         Successfully filtered the outliers

ALGORITHM REFERENCES:
OLI Focal Plane Alignment Calibration ADD
******************************************************************************/

#include <string.h>            /* String library */
#include <stdlib.h>            /* malloc/free prototypes */
#include <math.h>              /* Math library */

#include "ias_logging.h"       /* IAS logging library */
#include "ias_math.h"          /* IAS math library */
#include "filter_outliers.h"   /* Prototype for this routine */

/* TYPE DEFINITIONS */

/* statistics maintained for each grid cell */
typedef struct local_stats_type
{
    double sum;           /* Sum of data */
    double sum_sqr;       /* Sum of squared data */
    double mean;          /* Mean of data */
    double std_dev;       /* Standard deviation of data */
    double min_interval;  /* Minimum data point accepted */
    double max_interval;  /* Maximum data point accepted */
} LOCAL_STATS_TYPE;

/* Holds the outlier statistics for the X and Y directions */
typedef struct outlier_stats_type
{
    int points;              /* Points included in this set of data */
    LOCAL_STATS_TYPE x;      /* Statistics in the sample direction */
    LOCAL_STATS_TYPE y;      /* Statistics in the line direction */
} OUTLIER_STATS_TYPE;

/* Type to track outlier information for a grid cell */
typedef struct
{
    int no_outliers;         /* Flag to indicate that no outliers are left
                                in this cell */
    int point_index;         /* Correlation point number of the most extreme
                                outlier for this cell */
    double error_magnitude;  /* Magnitude of the error for the current 
                                outlier */
} OUTLIER_TRACKING_TYPE;

/* This is used to create a simple hash table for the t distribution values.
   Previously this routine repeatedly called the ias_math_compute_t_confidence 
   routine to calculate the same values.  This hash increased the speed of the 
   executable by a factor of 20. */
typedef struct 
{
    int key;        /* Hash table key (in this case the degrees of freedom) */
    double t_val;   /* Stored t distribution value */
} HASH_T_VAL_TYPE;

int filter_outliers
(
    int cells_per_sca,            /* I: Number of cells in an SCA */
    double outlier_tolerance,     /* I: Outlier tolerance */
    int corr_points,              /* I: Number of focal plane correlation
                                        points */
    CORR_PT_INFO *fp_corr_pts_ptr /* I/O: Focal plane correlation points */
)
{
    OUTLIER_STATS_TYPE *stats_ptr; /* Pointer to stats information */
    int index;                     /* Loop counter */
    int cell_index;                /* Index to cell statistics */
    double temp_x;                 /* Temporary variable for sample */
    double temp_y;                 /* Temporary variable for line */
    double t_dist;                 /* T distribution value for the given 
                                      tolerance */
    double temp;                   /* Temp values */
    int outlier_found = 1;         /* Flag to indicate outliers are still 
                                      present in the data */
    OUTLIER_TRACKING_TYPE *track_ptr; /* Pointer to outlier tracking 
                                      information */
    HASH_T_VAL_TYPE hashed_t_val[256]; /* Simple hash table for t distribution
                                      values.  256 is generally a bad size for
                                      a hash table, but for this simple case
                                      it doesn't really matter. */
    int hash_index;                /* Index into the hash table */
    
    /* Initialize the hash table key contents to illegal values so it appears 
       empty as it should */
    memset(hashed_t_val, 0xff, sizeof(hashed_t_val));

    /* Allocate memory for gathering the statistics on the cells */
    stats_ptr = malloc(cells_per_sca * sizeof(OUTLIER_STATS_TYPE) * 
        IAS_MAX_NSCAS);
    if (stats_ptr == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for %d outlier statistics sets", 
            cells_per_sca);
        return ERROR;
    }
    
    /* Allocate memory for tracking outliers */
    track_ptr = malloc(cells_per_sca * sizeof(OUTLIER_TRACKING_TYPE) *
        IAS_MAX_NSCAS);
    if (track_ptr == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for %d outlier tracking sets",
            cells_per_sca);
        free(stats_ptr);
        return ERROR;
    }
    /* Initialize the no outliers flag */
    memset(track_ptr, 0, cells_per_sca * sizeof(OUTLIER_TRACKING_TYPE) * 
        IAS_MAX_NSCAS);
    
    /* Zero out the statistics information */
    for (index = 0; index < cells_per_sca * IAS_MAX_NSCAS; index++)
    {
        stats_ptr[index].points = 0;
        stats_ptr[index].x.sum = 0.0;
        stats_ptr[index].x.sum_sqr = 0.0;
        stats_ptr[index].y.sum = 0.0;
        stats_ptr[index].y.sum_sqr = 0.0;
    }
    
    /* Collect the sum and sum squared statistics for each active point in 
       the cells across the set of SCAs */
    for (index = 0; index < corr_points; index++)
    {
        if (fp_corr_pts_ptr[index].active_flag == ACTIVE)
        {
            /* Calculate the cell index */
            cell_index = fp_corr_pts_ptr[index].grid_column + 
                        (fp_corr_pts_ptr[index].sca_index * cells_per_sca);
           
            /* If the cell_index is illegal, report an error and return */
            if ((cell_index < 0) || 
                (cell_index > cells_per_sca * IAS_MAX_NSCAS))
            {
                
                IAS_LOG_ERROR("Illegal grid column or SCA number at point %d",
                    index);
                free(stats_ptr);
                free(track_ptr);
                return ERROR;
            }

            /* Get the LOS errors */
            temp_x = fp_corr_pts_ptr[index].los_err.line;
            temp_y = fp_corr_pts_ptr[index].los_err.samp;
            
            /* Calculate the sum and sum of squares */
            stats_ptr[cell_index].x.sum += temp_x;
            stats_ptr[cell_index].x.sum_sqr += temp_x * temp_x;
            stats_ptr[cell_index].y.sum += temp_y;
            stats_ptr[cell_index].y.sum_sqr += temp_y * temp_y;
           
            /* Keep track of the number of points */
            stats_ptr[cell_index].points++;
        }
    }

    /* Eliminate outliers, at most one per grid cell in a pass, until no 
       more are found */
    while (outlier_found)
    {
        /* Flag that no outliers were found yet */
        outlier_found = 0;
       
        /* Find the minimum and maximum interval points for each cell */
        for (cell_index = 0; cell_index < cells_per_sca * IAS_MAX_NSCAS; 
             cell_index++)
        {
            OUTLIER_STATS_TYPE *cell_ptr; /* Pointer to stats information 
                                             for 1 cell */
            track_ptr[cell_index].point_index = -1;
            track_ptr[cell_index].error_magnitude = 0.0;

            if (track_ptr[cell_index].no_outliers)
                continue;
                
            cell_ptr = &stats_ptr[cell_index];
        
            /* If there are not enough points in this cell, set the min/max
               interval to force all points to be thrown away */
            if (cell_ptr->points < 2)
            {
                cell_ptr->x.min_interval = 1.0;
                cell_ptr->x.max_interval = 0.0;
                cell_ptr->y.min_interval = 1.0;
                cell_ptr->y.max_interval = 0.0;
                continue;
            }
            
            /* Calculate the mean and standard deviation */
            cell_ptr->x.mean = cell_ptr->x.sum / cell_ptr->points;
            cell_ptr->y.mean = cell_ptr->y.sum / cell_ptr->points;

            /* If the value inside the square root is less than zero then
               issue an error. */
            temp = cell_ptr->x.sum_sqr - (cell_ptr->points *
                   cell_ptr->x.mean * cell_ptr->x.mean);
            if (temp < 0.0)
            {
                IAS_LOG_ERROR("temp for x is %f; it should not be negative "
                    "since it is used in a square root calculation.", temp);
                free(stats_ptr);
                free(track_ptr);
                return ERROR;
            }
            else
            {
                cell_ptr->x.std_dev = sqrt(temp/(cell_ptr->points - 1));
            }

            temp = cell_ptr->y.sum_sqr - (cell_ptr->points * 
                   cell_ptr->y.mean * cell_ptr->y.mean);
            if (temp < 0.0)
            {
                IAS_LOG_ERROR("temp for y is %f; it should not be negative "
                    "since it is used in a square root calculation.", temp);
                free(stats_ptr);
                free(track_ptr);
                return ERROR;
            }
            else
            {
                cell_ptr->y.std_dev = sqrt(temp/(cell_ptr->points - 1));
            }

            /* Check if t distribution value is in the hash table */
            hash_index = (cell_ptr->points - 1) & 0xFF;
            if (hashed_t_val[hash_index].key == (cell_ptr->points - 1))
            {
                /* The value in the hash table is the desired one */
                t_dist = hashed_t_val[hash_index].t_val;
            }
            else
            {
                /* Get the T distribution value for the given tolerance and 
                   degrees of freedom */
                t_dist = ias_math_compute_t_confidence(outlier_tolerance, 
                    cell_ptr->points - 1);

                hashed_t_val[hash_index].key = cell_ptr->points - 1;
                hashed_t_val[hash_index].t_val = t_dist;
            }

            /* Calculate the minimum and maximum interval points */
            temp_x = t_dist * cell_ptr->x.std_dev;
            temp_y = t_dist * cell_ptr->y.std_dev;
            cell_ptr->x.min_interval = cell_ptr->x.mean - temp_x;
            cell_ptr->x.max_interval = cell_ptr->x.mean + temp_x;
            cell_ptr->y.min_interval = cell_ptr->y.mean - temp_y;
            cell_ptr->y.max_interval = cell_ptr->y.mean + temp_y;
        }

        /* For each correlation point, check if the point should be thrown 
           away because it is outside of the selected tolerance */
        for (index = 0; index < corr_points; index++)
        {
            if (fp_corr_pts_ptr[index].active_flag == ACTIVE)
            {
                /* Calculate the cell index */
                cell_index = fp_corr_pts_ptr[index].grid_column + 
                            (fp_corr_pts_ptr[index].sca_index * cells_per_sca);

                if (track_ptr[cell_index].no_outliers)
                    continue;
                    
                /* If the point's LOS error falls outside either the X
                   or Y min/max interval more than the current maximum 
                   error for its SCA cell, log it as the current maximum 
                   outlier */
                temp_x = fp_corr_pts_ptr[index].los_err.line;
                temp_y = fp_corr_pts_ptr[index].los_err.samp;

                if (temp_x < stats_ptr[cell_index].x.min_interval)
                {
                    temp = stats_ptr[cell_index].x.min_interval - temp_x;
                    if (temp > track_ptr[cell_index].error_magnitude)
                    {
                        track_ptr[cell_index].error_magnitude = temp;
                        track_ptr[cell_index].point_index = index;
                    }
                }
                else if (temp_x > stats_ptr[cell_index].x.max_interval)
                {
                    temp = temp_x - stats_ptr[cell_index].x.max_interval;
                    if (temp > track_ptr[cell_index].error_magnitude)
                    {
                        track_ptr[cell_index].error_magnitude = temp;
                        track_ptr[cell_index].point_index = index;
                    }
                }
               
                /* Note this is not an else condition on purpose.  The
                   algorithm to find a point to throw away as an outlier
                   treats along and across equally */
                if (temp_y < stats_ptr[cell_index].y.min_interval)
                {
                    temp = stats_ptr[cell_index].y.min_interval - temp_y;
                    if (temp > track_ptr[cell_index].error_magnitude)
                    {
                        track_ptr[cell_index].error_magnitude = temp;
                        track_ptr[cell_index].point_index = index;
                    }
                }
                else if (temp_y > stats_ptr[cell_index].y.max_interval)
                {
                    temp = temp_y - stats_ptr[cell_index].y.max_interval;
                    if (temp > track_ptr[cell_index].error_magnitude)
                    {
                        track_ptr[cell_index].error_magnitude = temp;
                        track_ptr[cell_index].point_index = index;
                    }
                }
            }  /* Is active point */
        } /* Loop through all points */

        /* Eliminate the maximum outliers from the data set by setting their
           active flags to "outlier" */
        for (cell_index = 0; cell_index < cells_per_sca * IAS_MAX_NSCAS; 
             cell_index++)
        {
            if (track_ptr[cell_index].point_index >= 0)
            {
                index = track_ptr[cell_index].point_index;
                fp_corr_pts_ptr[index].active_flag = OUTLIER;

                /* Remove the effects of the outlier point from statistics */
                temp_x = fp_corr_pts_ptr[index].los_err.line;
                temp_y = fp_corr_pts_ptr[index].los_err.samp;
                
                /* Remove the sum and sum of squares from the data for the
                   rejected outlier point */
                stats_ptr[cell_index].x.sum -= temp_x;
                stats_ptr[cell_index].x.sum_sqr -= temp_x * temp_x;
                stats_ptr[cell_index].y.sum -= temp_y;
                stats_ptr[cell_index].y.sum_sqr -= temp_y * temp_y;
               
                /* Decrement the number of points for this cell */
                stats_ptr[cell_index].points--;
                
                /* Flag that an outlier was found so another pass through
                   the outlier rejection is needed */
                outlier_found = 1;
            }
            else
            {
                /* Flag that no more outliers remain in this cell */
                track_ptr[cell_index].no_outliers = 1;
            }
        }
        
    } /* End while */
    
    /* Free the memory allocated for the statistics information */
    free(stats_ptr);
    free(track_ptr);

    IAS_LOG_INFO("Outlier detection and rejection is complete.");

    return SUCCESS;
}
