/******************************************************************************
NAME:           fit_polynomials

PURPOSE:
fit_polynomials performs the weighted least squares fit of the correlation 
data points to find the Legendre error polynomial.  

RETURNS:
Type = int
Value           Description
-----           -----------
ERROR           Error fitting the Legendre error polynomial 
SUCCESS         Successfully fitted the Legendre error polynomial

NOTES:
- This routine calculates the weighted least squares fit of the correlated data 
  points (using the angular error) to a Legendre error polynomial.  
- Through some mathematical and statistical magic, the problem to solve in
  this routine ends up being TWAX = TWY where we want to solve for X with:
      A being the [corr_points x IAS_LOS_LEGENDRE_TERMS] matrix of terms from 
      the equation
        x= 2 * (time in scan to this point/measured time in whole scan) - 1 
        from the correlation data points
      W being the weight matrix including the weight for each correlation data
        point of the A matrix (only the diagonal contains weights for that row)
      T being the transpose of the A matrix
      X being the [IAS_LOS_LEGENDRE_TERMS x 1] matrix with the Legendre 
        coefficients we are looking for
      Y being the [corr_points x 1] matrix of angular error for each 
        correlation data point corresponding to the rows in the A matrix
  This problem needs to be solved for both the along and across scan 
  directions, and for the forward and reverse scans.  This routine will solve 
  for both the along and across scan directions for either the forward or 
  reverse scans.
- Solving the above equation for X yields: X = (TWA)^-1 * TWY
- An optimization is used to keep the size of the matrices to a managable 
  level.  Since the W matrix (weight) only contains data on the diagonal the 
  TWA matrix can be accumulated in a 
  [IAS_LOS_LEGENDRE_TERMS x IAS_LOS_LEGENDRE_TERMS] matrix and the TWY matrix 
  can be accumulated in a [IAS_LOS_LEGENDRE_TERMS x 1] matrix.
- The TW and TWA matrices are the same for both along and across the scan.  
  This is used to optimize the code.
  
ALGORITHM REFERENCES:
- OLI Focal Plane Alignment Calibration ADD 
- Numerical Recipes in C
******************************************************************************/

#include "ias_logging.h"     /* IAS logging library */
#include "ias_math.h"        /* IAS math library */
#include "fit_polynomials.h" /* Prototype for this routine */

int fit_polynomials
(
    int corr_points,         /* I: Number of focal plane correlation points */
    int sca_index,           /* I: SCA index to solve */
    int fit_order,           /* I: Order of polynomial to fit (0-2) */
    int sca_num_det,         /* I: Number of detectors in the SCA */
    const CORR_PT_INFO *fp_corr_pts_ptr,/* I: Focal plane correlation points */
    double *error_coeffs_x,  /* O: Legendre error coeffs along track */
    double *error_coeffs_y   /* O: Legendre error coeffs across track */
)
{
    int index;               /* Loop counter */
    int term;                /* Loop counter */
    int row, column;         /* Loop counters */
    double TWA[IAS_LOS_LEGENDRE_TERMS * IAS_LOS_LEGENDRE_TERMS]={0}; 
                             /* [IAS_LOS_LEGENDRE_TERMS x 
                                IAS_LOS_LEGENDRE_TERMS] accumulated matrix for 
                                the data points */
    double A[IAS_LOS_LEGENDRE_TERMS]; /* [1 x IAS_LOS_LEGENDRE_TERMS] matrix of 
                                Legendre terms for the current point */
    double TW[IAS_LOS_LEGENDRE_TERMS];/* [IAS_LOS_LEGENDRE_TERMS x 1] matrix of 
                                A transposed and multiplied by the weight of 
                                the current point */
    double TWY_x[IAS_LOS_LEGENDRE_TERMS] ={0}; /* [IAS_LOS_LEGENDRE_TERMS x 1] 
                                matrix of error terms sample direction */
    double TWY_y[IAS_LOS_LEGENDRE_TERMS] ={0}; /* [IAS_LOS_LEGENDRE_TERMS x 1] 
                                matrix of error terms line direction */
    double weight;           /* Temporary weight holder */
    double v[IAS_LOS_LEGENDRE_TERMS]; /* Work vector for QR/QRsolve */
    int points_used = 0;     /* Count of points factored into the solution */
    
    /* Make sure the number of points is not less than or equal to zero */
    if (corr_points <= 0)
    {
        IAS_LOG_ERROR("Number of correlation points %d is <= 0", corr_points);
        return ERROR;
    }

    /* Check for later divide by 0 */
    if (sca_num_det == 1)
    {
        IAS_LOG_ERROR("Number of detectors is 1, leading to divide by 0");
    }
    
    /* For each of the active points in the current SCA */
    for (index = 0; index < corr_points; index++)
    {
        if ((fp_corr_pts_ptr[index].active_flag == ACTIVE) &&
            (fp_corr_pts_ptr[index].sca_index == sca_index))
        {
            /* Update the count of the points used */
            points_used++;

            /* Form the Legendre terms into the A matrix for this point */
            A[0] = 1.0;
            /* Calculate normalized detector coordinate as
               2 * (input space x / number of detectors) - 1 */
            A[1] = 2 * (fp_corr_pts_ptr[index].srch_is_pt.samp /
                (sca_num_det - 1)) - 1;

            for (term = 2; term < LOCAL_LOS_LEGENDRE_TERMS; term++)
            {
                A[term]= ((2 * term - 1) * A[1]
                    * A[term - 1] - (term - 1) * A[term - 2])/(double)term;
            }

            /* Form the TW matrix by multiplying the transpose of A by the
               point's weight */
            weight = fp_corr_pts_ptr[index].pt_weight;
            for (term = 0; term < LOCAL_LOS_LEGENDRE_TERMS; term++)
            {
                TW[term] = A[term] * weight;
            }

            /* Accumulate just the upper half of the TWA matrix since it
               is symmetrical around the diagonal */
            for (row = 0; row < LOCAL_LOS_LEGENDRE_TERMS; row++)
            {
                for (column = row; column < LOCAL_LOS_LEGENDRE_TERMS; column++)
                {
                    TWA[column + row * LOCAL_LOS_LEGENDRE_TERMS] += TW[row] * 
                        A[column];
                }
            }

            /* Accumulate the TWY matrix in the line and sample directions */
            for (row = 0; row < LOCAL_LOS_LEGENDRE_TERMS; row++)
            {
                TWY_x[row] += TW[row] * fp_corr_pts_ptr[index].los_err.line;
                TWY_y[row] += TW[row] * fp_corr_pts_ptr[index].los_err.samp;
            }
        } /* End of active data point */
    } /* End of looping through all the correlation data points */
 
    /* Verify enough points were found to fit (minimum is the degrees of
       freedom - LOCAL_LOS_LEGENDRE_TERMS) */
    if (points_used < LOCAL_LOS_LEGENDRE_TERMS)
    {
        /* Not enough points used, generate an error message and error out */
        IAS_LOG_ERROR("Only %d correlation points found to be usable "
                "when a minimum of %d are needed", points_used, 
                LOCAL_LOS_LEGENDRE_TERMS);
        return ERROR;
    }
    
    /* Copy the upper half of the TWA matrix to the lower half in preparation 
       for the QR call */
    for (row = 0; row < (LOCAL_LOS_LEGENDRE_TERMS - 1); row++)
    {
        for (column = row + 1; column < LOCAL_LOS_LEGENDRE_TERMS; column++)
        {
            TWA[row + column * LOCAL_LOS_LEGENDRE_TERMS] = 
                TWA[column + row * LOCAL_LOS_LEGENDRE_TERMS];
        }
    }
    
    /* Perform the QR decomposition on the TWA matrix */
    if (ias_math_matrix_QRfactorization(TWA, LOCAL_LOS_LEGENDRE_TERMS /*rows*/, 
        LOCAL_LOS_LEGENDRE_TERMS/*columns*/, v, 1 /* row major format */) 
        != SUCCESS)  
    {
        /* Error solving for the least squares fit */
        IAS_LOG_ERROR("Calculating least squares fit for SCA %d", 
            sca_index + 1);
        return ERROR;
    }
    
    /* Solve for the TWY X coefficients */
    if (ias_math_matrix_QRsolve(TWA, LOCAL_LOS_LEGENDRE_TERMS /*rows*/, 
        LOCAL_LOS_LEGENDRE_TERMS /*columns*/, v, TWY_x, 1 /*v is already 
        constructed*/) != SUCCESS)  
    {
        /* Error solving for the least squares fit */
        IAS_LOG_ERROR("Calculating least squares fit for SCA %d",
            sca_index + 1);
        return ERROR;
    }

    /* Solve for the TWY Y coefficients */
    if (ias_math_matrix_QRsolve(TWA, LOCAL_LOS_LEGENDRE_TERMS /*rows*/, 
        LOCAL_LOS_LEGENDRE_TERMS /*columns*/, v, TWY_y, 1 /*v is already 
        constructed*/) != SUCCESS) 
    {
        /* Error solving for the least squares fit */
        IAS_LOG_ERROR("Calculating least squares fit for SCA %d",
            sca_index + 1);
        return ERROR;
    }
  
    /* Copy the X and Y error coefficients from the solutions.  Note: It's
       fine to compute full order and then zero out the higher order terms
       here because they're Legendre coefficients */
    for (index = 0; index < IAS_LOS_LEGENDRE_TERMS; index++)
    {
        if (index <= fit_order)
        {
            error_coeffs_x[index] = TWY_x[index];
            error_coeffs_y[index] = TWY_y[index];
        }
        else
        {
            error_coeffs_x[index] = 0.0;
            error_coeffs_y[index] = 0.0;
        }
    }
    
    return SUCCESS;
}
