/******************************************************************************
NAME:           generate_legendre_polynomials (main)

PURPOSE:        
Main driver for the generate polynomial portion of focal plane calibration.  
The generate Legendre portion reads the results of the focal plane setup, 
filters the outliers, fits the data to a legendre polynomial, updates the 
SCA models, and generates output reports.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
EXIT_FAILURE    Error generating Legendre; no output data produced 
EXIT_SUCCESS    Successful; Legendre polynomials generated
******************************************************************************/
#include <stdlib.h>                       /* Standard C library */
#include <string.h>                       /* String library */

#include "ias_logging.h"                  /* IAS logging library */
#include "ias_model_io.h"                 /* Model I/O library */
#include "focal_plane.h"                  /* Prototypes for the Focal Plane
                                             executables */
#include "calculate_post_fit_residuals.h" /* calculate_post_fit_residuals 
                                             prototype */
#include "create_focal_plane_report.h"    /* create_focal_plane_report 
                                             prototype */
#include "filter_outliers.h"              /* filter_outliers prototype */
#include "calculate_point_weights.h"      /* calculate_point_weights 
                                             prototype */
#include "fit_polynomials.h"              /* fit_polynomials prototype */
#include "save_to_database.h"             /* DB save structure */

int main(int argc, char *argv[])
{
    ALIGNMENT_PARMS parms;      /* Parameters read from parameter file */
    CORR_PT_INFO *fp_corr_pts;  /* Information about each correlation point */
    IAS_CPF *cpf;                /* CPF parameters */
    IAS_LOS_MODEL *model;       /* Model structure */
    int status;                 /* Function return status */
    int ncells;                 /* Number of cells across a SCA */
    int sca_index;              /* SCA counter/index */
    int nscas;                  /* Number of SCAs in this band (input space) */
    int sca_num_det;            /* Number of detectors per SCA */
    int npts;                   /* Number of correlation points to work with */
    int pan_band_number;        /* Number of reference band (pan) */
    int pan_band_index;         /* Index of reference band (pan) */
    char output_cpf_filename[PATH_MAX]; /* Name of file for writing OLI 
                                   focal plane parameters CPF group */
    double sca_coef_x[IAS_MAX_NSCAS][IAS_LOS_LEGENDRE_TERMS]; /* Along-track 
                                   SCA LOS Legendre coefficients */
    double sca_coef_y[IAS_MAX_NSCAS][IAS_LOS_LEGENDRE_TERMS]; /* Across-track 
                                   SCA LOS Legendre coefficients */
    double error_coeffs_x[IAS_MAX_NSCAS][IAS_LOS_LEGENDRE_TERMS]; 
                                /* Along-track error Legendre coefficients */
    double error_coeffs_y[IAS_MAX_NSCAS][IAS_LOS_LEGENDRE_TERMS]; 
                                /* Across-track error Legendre coefficients */
    double new_sca_coef_x[IAS_MAX_NSCAS][IAS_LOS_LEGENDRE_TERMS]; 
                                /* New along-track Legendre coefficients */
    double new_sca_coef_y[IAS_MAX_NSCAS][IAS_LOS_LEGENDRE_TERMS]; 
                                /* New across-track Legendre coefficients */
    double mean_x_fit = 0.0;    /* Mean best-fit x offset */
    double mean_y_fit = 0.0;    /* Mean best-fit y offset */
    POST_FIT_RESIDUAL_STATS_TYPE residual_stats[IAS_MAX_NSCAS];
    struct main_db_save_info main_db_save_info; /* Main DB save information */ 
    struct sca_db_save_info sca_db_save_info; /* SCA level DB save info */
    double db_save_threshold_x;  /* Threshold to limit trending */
    double db_save_threshold_y;  /* Threshold to limit trending */
    int db_save_threshold_flag = 1; /* Flag to indicate results should
                                   be written to the database */
    int coef_index;             /* Legendre coefficient index */

    /* Initialize the DB save structures */
    memset(&main_db_save_info, 0, sizeof(main_db_save_info));
    memset(&sca_db_save_info, 0, sizeof(sca_db_save_info));

    /* Initialize the logging library */
    if (ias_log_initialize("generate_legendre_polynomials") != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing logging library");
        return EXIT_FAILURE;
    }

    /* Get parameters and setup information from get_focal_plane_parms
       which is shared with the focal plane setup executable */
    status = get_parameters(argc, argv, &parms, &cpf); 
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting input parameters");
        return EXIT_FAILURE;
    }

    /* Initialize satellite attributes */
    status = ias_sat_attr_initialize(parms.satellite_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attributes library");
        ias_cpf_free(cpf); 
        return EXIT_FAILURE;
    }

    /* Get the PAN band number, which will be the reference band */
    pan_band_number = ias_sat_attr_get_band_number_from_type(IAS_PAN_BAND);
    if (pan_band_number == ERROR)
    {
        IAS_LOG_ERROR("Invalid pan band number %d", pan_band_number);
        ias_cpf_free(cpf); 
        return EXIT_FAILURE;
    }

    /* Get the PAN band index */
    pan_band_index = ias_sat_attr_convert_band_number_to_index(
        pan_band_number);
    if (pan_band_index == ERROR)
    {
        IAS_LOG_ERROR("Invalid band index for pan band number %d", 
            pan_band_number);
        ias_cpf_free(cpf); 
        return EXIT_FAILURE;
    }

    /* Log initiation */
    IAS_LOG_INFO("Initiated");

    /* Load the DB save thresholds and convert to radians */
    db_save_threshold_x = parms.along_trend_threshold * 1e-6;
    db_save_threshold_y = parms.across_trend_threshold * 1e-6;

    /* Read the model file */
    model = ias_model_read(parms.model_file_name);
    if (model == NULL)
    {
        IAS_LOG_ERROR("Reading the model file %s", parms.model_file_name);
        ias_cpf_free(cpf); 
        return EXIT_FAILURE;
    }

    /* Get the existing pan band SCA polynomials from the model */
    for (sca_index = 0; sca_index < 
        model->sensor.bands[pan_band_index].sca_count; sca_index++)
    {
        for (coef_index = 0; coef_index < LOCAL_LOS_LEGENDRE_TERMS; 
             coef_index++)
        {
            sca_coef_x[sca_index][coef_index] = 
                model->sensor.bands[pan_band_index].scas[sca_index].
                sca_coef_x[coef_index];
                
            sca_coef_y[sca_index][coef_index] = 
                model->sensor.bands[pan_band_index].scas[sca_index].
                sca_coef_y[coef_index]; 
        }
    }

    /* Read the correlation results from the focal plane setup executable */
    fp_corr_pts = read_correlation_info(parms.correlation_file_name, &npts, 
        &nscas, &ncells);

    if (fp_corr_pts == NULL)
    {
        IAS_LOG_ERROR("Reading correlation information from %s",
            parms.correlation_file_name); 
        ias_cpf_free(cpf); 
        ias_los_model_free(model);
        model = NULL;
        return EXIT_FAILURE;
    }

    /* Filter the outliers in the correlation data */
    if (filter_outliers(ncells, parms.outlier_tolerance, npts, fp_corr_pts) 
        != SUCCESS)
    {
        IAS_LOG_ERROR("Filtering the outlier points from the correlation "
                "information from %s", parms.correlation_file_name);
        free(fp_corr_pts);
        ias_cpf_free(cpf); 
        ias_los_model_free(model);
        model = NULL;
        return EXIT_FAILURE;
    }

    /* Weight the remaining correlation points */
    calculate_point_weights(parms.point_weight, npts, fp_corr_pts);

    /* Fit each SCA */
    for (sca_index = 0; sca_index < nscas; sca_index++)
    {
        /* Get the number of detectors for this SCA for the PAN band */
        sca_num_det = model->sensor.bands[pan_band_index].
            scas[sca_index].detectors;

        /* Fit the correlation results to a Legendre error polynomial */
        if (fit_polynomials(npts, sca_index, parms.fit_order, sca_num_det, 
                fp_corr_pts, error_coeffs_x[sca_index], 
                error_coeffs_y[sca_index]) != SUCCESS)
        {
            IAS_LOG_ERROR("Performing a least squares fit on the correlation "
                          "points in SCA %d",(sca_index+1));
            free(fp_corr_pts);
            ias_cpf_free(cpf); 
            ias_los_model_free(model);
            model = NULL;
            return EXIT_FAILURE;
        }

        /* Accumulate mean offset fit values */
        mean_x_fit += error_coeffs_x[sca_index][0];
        mean_y_fit += error_coeffs_y[sca_index][0];

        /* Calculate the post fit residual statistics */
        if (calculate_post_fit_residuals(error_coeffs_x[sca_index],
                error_coeffs_y[sca_index], npts, sca_index, sca_num_det, 
                fp_corr_pts, &residual_stats[sca_index], db_save_threshold_x, 
                db_save_threshold_y, &db_save_threshold_flag) != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating scan post-fit residual statistics for "
                "SCA %d", sca_index + 1);
            free(fp_corr_pts);
            ias_cpf_free(cpf); 
            ias_los_model_free(model);
            model = NULL;
            return EXIT_FAILURE;
        }
    }

    /* Free the model */
    ias_los_model_free(model);
    model = NULL;

    /* Compute the mean fit offsets */
    mean_x_fit /= (double)nscas;
    mean_y_fit /= (double)nscas;

    /* Update the SCA coefficients */
    for (sca_index = 0; sca_index < nscas; sca_index++)
    {
        /* Remove the mean fit offset (since it's just residual precision 
           fit error) */
        error_coeffs_x[sca_index][0] -= mean_x_fit;
        error_coeffs_y[sca_index][0] -= mean_y_fit;

        /* Apply correction Legendre to original polynomials */
        /* Along track */
        new_sca_coef_x[sca_index][0] = sca_coef_x[sca_index][0] + 
            error_coeffs_x[sca_index][0];
        new_sca_coef_x[sca_index][1] = sca_coef_x[sca_index][1] + 
            error_coeffs_x[sca_index][1];
        new_sca_coef_x[sca_index][2] = sca_coef_x[sca_index][2] + 
            error_coeffs_x[sca_index][2];

        /* Across track */
        new_sca_coef_y[sca_index][0] = sca_coef_y[sca_index][0] + 
            error_coeffs_y[sca_index][0];
        new_sca_coef_y[sca_index][1] = sca_coef_y[sca_index][1] + 
            error_coeffs_y[sca_index][1];
        new_sca_coef_y[sca_index][2] = sca_coef_y[sca_index][2] + 
            error_coeffs_y[sca_index][2];
    }

    /* Output the focal plane report */
    if (create_focal_plane_report(parms.report_file_name, parms.terr_img_name, 
            parms.ref_img_name, parms.l0r_img_name, nscas, sca_coef_x, 
            sca_coef_y, error_coeffs_x, error_coeffs_y, new_sca_coef_x, 
            new_sca_coef_y, fp_corr_pts, npts, residual_stats,
            parms.outlier_tolerance, parms.fit_order, parms.work_order_id) 
            != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the focal plane calibration report");
        free(fp_corr_pts);
        ias_cpf_free(cpf); 
        return EXIT_FAILURE;
    }

    /* Free correlation points */
    free(fp_corr_pts);
    fp_corr_pts = NULL;

    /* Output the updated CPF parameters if requested */
    if (parms.write_cpf_flag)
    {
        if (sprintf(output_cpf_filename, "focal_plane_parameters_%08d_%08d.cpf",
                parms.start_effective_date, parms.end_effective_date) < 0)
        {
            IAS_LOG_ERROR("Constructing the output CPF group file name for "
                "dates %d to %d", parms.start_effective_date,
                parms.end_effective_date);
            ias_cpf_free(cpf); 
            return EXIT_FAILURE;
        }

        if (write_sca_parameters_cpf(output_cpf_filename, error_coeffs_x, 
            error_coeffs_y, cpf, IAS_OLI) != SUCCESS)
        {
            IAS_LOG_ERROR("Writing the updated CPF group report file %s",
                output_cpf_filename);
            ias_cpf_free(cpf); 
            return EXIT_FAILURE;
        }
    }

    /* Free CPF */
    ias_cpf_free(cpf); 
    cpf = NULL;

    /* If storing results in the database is requested, and not rejected by 
       the DB save thresholds */
    if ((db_save_threshold_flag) && (parms.store_in_database_flag))
    {
        IAS_LOG_INFO("Writing results to database");

        /* Populate database structures */
        prepare_for_database_save(sca_coef_x, sca_coef_y, error_coeffs_x,
            error_coeffs_y, new_sca_coef_x, new_sca_coef_y, residual_stats,
            parms.work_order_id, parms.ref_img_name, parms.outlier_tolerance,
            parms.fit_order, mean_x_fit, mean_y_fit, nscas, &main_db_save_info,
            &sca_db_save_info);

        if (save_to_database(&main_db_save_info, &sca_db_save_info, nscas)
            != SUCCESS)
        {
            IAS_LOG_ERROR("Storing results in database");
            return EXIT_FAILURE;
        }
    }

    IAS_LOG_INFO("Successful completion");

    return EXIT_SUCCESS;
}
