/******************************************************************************
NAME:           read_correlation_info

PURPOSE:
read_correlation_info reads the correlation results generated by the Setup
phase executable.

RETURNS:
Type = CORR_PT_INFO *
Value           Description
-----           -----------
NULL            Error reading the correlation information
non-NULL        Successful; correlation information read

NOTES:
- The correlation file is written as a binary file.  The first integer in
  the file is the number of correlation data records written to the file.  The 
  correlation data records are written as binary data with a single call to 
  fwrite.  So the read function will read the data in that format.
- The calling routine is responsible for freeing the memory allocated to the
  corr_pts array when it is done with it.
******************************************************************************/
#include <stdlib.h>        /* For malloc */

#include "ias_logging.h"   /* IAS Logging library */
#include "focal_plane.h"   /* Prototype for this routine */

CORR_PT_INFO *read_correlation_info
(
    const char *correlation_file_name, /* I: Pointer to correlation file name */
    int *points_ptr,       /* O: Number of points read */
    int *scas_ptr,         /* O: Number of SCAs in the data */
    int *cells_ptr         /* O: Number of cells across an SCA */
)
{
    int max_sca_index = 0; /* Maximum SCA index in correlation data */
    int last_cell = 0;     /* Last cell in a SCA */
    int sca_index;         /* Temporary SCA index */
    FILE *corr_file_ptr;   /* File structure for correlation data I/O */
    int points;            /* Correlation points contained in the file */
    int index;             /* Loop counter */
    CORR_PT_INFO *corr_pts = NULL; /* Pointer to correlation points read */

    /* Open the correlation file for reading in binary mode */
    corr_file_ptr = fopen(correlation_file_name, "rb");
    if (corr_file_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening the correlation data file, %s",
            correlation_file_name);
        return NULL;
    }

    /* Read the number of correlation points from the file */
    if (fread(&points, sizeof(int), 1, corr_file_ptr) != 1)
    {
        /* Error reading the number of points */
        IAS_LOG_ERROR("Reading the number of correlation points from %s",
            correlation_file_name);
        fclose(corr_file_ptr);
        return NULL;
    }
    *points_ptr = points;

    /* Verify there are some points in the data */
    if (points <= 0)
    {
        /* An illegal number of points was read; generate an error */
        IAS_LOG_ERROR("In the number of points read from %s",
            correlation_file_name);
        fclose(corr_file_ptr);
        return NULL;
    }
    
    /* Allocate memory for the correlation points */
    corr_pts = malloc(points * sizeof(CORR_PT_INFO));
    if (corr_pts == NULL)
    {
        /* Error allocating memory */
        IAS_LOG_ERROR("Allocating memory for %d correlation points", points);
        fclose(corr_file_ptr);
        return NULL;
    }
    
    /* Read the correlation points from the file */
    if (fread(corr_pts, sizeof(CORR_PT_INFO), points, corr_file_ptr) != points)
    {
        /* Error reading the correlation points */
        IAS_LOG_ERROR("Reading the correlation points from %s",
            correlation_file_name);
        free(corr_pts);
        fclose(corr_file_ptr);
        return NULL;
    }
    
    /* Close the correlation data file */
    if (fclose(corr_file_ptr) != 0)
    {
        /* Note: This is only a warning since data is only being read from
           the file and it's already been successfully read */
        IAS_LOG_WARNING("%s did not close properly", correlation_file_name);
    }

    /* Find the max and min SCA numbers and the number of cells across the SCA
       in the data */
    for (index = 0; index < points; index++)
    {
        sca_index = corr_pts[index].sca_index;
        if (sca_index > max_sca_index)
            max_sca_index = sca_index;

        if (corr_pts[index].grid_column > last_cell)
            last_cell = corr_pts[index].grid_column;
    }
    
    /* Return the number of cells in a scan, adding 1 since last_cell holds
       the last cell index which is 1 less then the number of cells across the
       scan */
    *cells_ptr = last_cell + 1;
    
    /* Calculate the number of SCAs the data covers */
    *scas_ptr = max_sca_index + 1;

    IAS_LOG_INFO("Correlation data loaded");

    return corr_pts;
}
