/******************************************************************************
NAME: calc_input_space_errors

PURPOSE:
Calculate the errors in input space pixels.  This is done by first 
correlating in output space and converting the correlated locations to input 
space.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
ERROR           Error getting correlation info (error allocating memory, 
                correlating, or calculating input space coords)
SUCCESS         Successful correlation and calculation of input space coords

NOTES:
This part of the program sets the following fields in the correlation point
structures:
 sca_index
 grid_column
 grid_row
 nom_os_pt
 ref_os_pt
 srch_os_pt
 ref_is_pt
 srch_is_pt
 active_flag
 correlation_accuracy
******************************************************************************/

#include "ias_logging.h"               /* IAS Logging library */
#include "focal_plane_setup.h"

int calc_input_space_errors
(
    const char *ref_img_name, /* I: Name of the reference image */
    const char *terr_img_name, /* I: Name of the 1T (search) image */
    const char *dem_name,   /* I: Name of the DEM image */
    int corr_win_nl,        /* I: # of lines in correlation window */
    int corr_win_ns,        /* I: # of samples in correlation window */
    int abs_corr_coeff,     /* I: Flag to use abs of correlation coeff */
    float min_corr,         /* I: Minimum correlation strength */
    float max_disp,         /* I: Maximum correlation displacement */
    IAS_CORRELATION_FIT_TYPE fit_method, /*I: Correlation result fit 
                                  method */
    const IAS_GRID_TYPE *grid, /* I: Grid structure, used for point 
                                  selection and inverse mapping */
    float invalid_threshold, /* I: Invalid threshold percentage */ 
    int band_number,        /* I: Band number (pan) */
    CORR_PT_INFO *fp_corr_pts, /* O: Focal plane correlation point info */
    int npts                /* O: Number of focal plane corr points */
)
{
    /* Do correlation at each point */
    if (perform_correlation(fp_corr_pts, npts, ref_img_name, terr_img_name, 
        corr_win_nl, corr_win_ns, min_corr, max_disp, fit_method, 
        abs_corr_coeff, invalid_threshold, band_number) != SUCCESS)
    {
        IAS_LOG_ERROR("Correlating points for reference image %s, search "
            "image %s", ref_img_name, terr_img_name);
        return ERROR;
    }

    /* Calculate input space coords for output space coords */
    IAS_LOG_INFO("Mapping points to input space");
    if (map_coords_to_input_space(fp_corr_pts, dem_name, npts, grid,
        band_number) != SUCCESS) 
    {
        IAS_LOG_ERROR("Mapping points from output space to input space");
        return ERROR;
    }

    return SUCCESS;
}
