/******************************************************************************
NAME:           calc_los_errors

PURPOSE:
Using the input space error measures, calculate the line-of-sight errors.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
ERROR           Error calculating LOS for point
SUCCESS         LOS errors calculated

NOTES:
  This function uses the following fields:
        ref_os_pt
        srch_os_pt
        ref_is_pt
        srch_is_pt
  This function calculates the following fields:
        los_err
******************************************************************************/
#include "ias_logging.h"       /* IAS Logging library */
#include "focal_plane_setup.h"

int calc_los_errors
(
    CORR_PT_INFO *fp_corr_pts, /* I/O: Focal plane correlation point info */
    int npts,                  /* I: Number of focal plane corr points */
    const IAS_LOS_MODEL *model, /* I: Precision model structure */
    int band_number           /* I: Band number */
)
{
    int i;                     /* Looping variable */
    IAS_VECTOR ref_los;        /* Reference point line-of-sight */
    IAS_VECTOR srch_los;       /* Search point line-of-sight */
    int band_index;            /* Index corresponding to band number */
    int status;                /* Function return status */

    /* Find the band index corresponding to the band number */
    band_index = ias_sat_attr_convert_band_number_to_index(
        band_number);
    if (band_index == ERROR)
    {
        IAS_LOG_ERROR("Invalid band index for band number %d", band_number);
        return ERROR;
    }

    for (i = 0; i < npts; i++)
    {
        if (fp_corr_pts[i].active_flag == ACTIVE)
        {
            /* Calculate the line-of-sight errors in pixels, which is just 
               the difference between reference and search line/sample */
            fp_corr_pts[i].los_err_pix.line = fp_corr_pts[i].ref_is_pt.line -  
                fp_corr_pts[i].srch_is_pt.line;

            fp_corr_pts[i].los_err_pix.samp = fp_corr_pts[i].ref_is_pt.samp - 
                fp_corr_pts[i].srch_is_pt.samp;

            /* Find the LOS for this point using the reference point */
            status = ias_sensor_find_los_vector(fp_corr_pts[i].sca_index, 
                fp_corr_pts[i].ref_is_pt.samp, IAS_NOMINAL_DETECTOR, 
                &model->sensor.bands[band_index], &ref_los);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Calculation of LOS vector failed at SCA number "
                        "%d, Band number %d, Detector %.8e", 
                        fp_corr_pts[i].sca_index + 1, band_number, 
                        fp_corr_pts[i].ref_is_pt.samp);
                return ERROR;
            }

            /* Find the LOS for this point using the search point */
            status = ias_sensor_find_los_vector(fp_corr_pts[i].sca_index, 
                fp_corr_pts[i].srch_is_pt.samp, IAS_NOMINAL_DETECTOR, 
                &model->sensor.bands[band_index], &srch_los);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Calculation of LOS vector failed at SCA number "
                        "%d, Band number %d, Detector %.8e", 
                        fp_corr_pts[i].sca_index + 1, band_number, 
                        fp_corr_pts[i].srch_is_pt.samp);
                return ERROR;
            }

            /* Calculate the line-of-sight errors */
            fp_corr_pts[i].los_err.line = ref_los.x / ref_los.z - 
                srch_los.x / srch_los.z + (fp_corr_pts[i].los_err_pix.line *
                model->sensor.bands[band_index].sampling_char.along_ifov);

            fp_corr_pts[i].los_err.samp = ref_los.y / ref_los.z - 
                srch_los.y / srch_los.z; 
                
        }
    }

    return SUCCESS;
}
