/******************************************************************************
NAME:    check_images_match

PURPOSE:
Check to make sure the 1T, reference, and DEM images all match.  The
corners of all the images should match to within half a pixel, and the 
reference and 1T should be the same resolution (thus the same size).

RETURN VALUE:
Type = int
Value    Description
-----    -----------
ERROR    Images don't match, or problem with metadata
SUCCESS  The 3 images match, may be used for focal plane calibration

NOTES:
This function is an up-front check to make sure all the images work together,
before correlation is attempted.

**Assume the reference image and DEM are one-band images
******************************************************************************/

#include <math.h>
#include <string.h>                                /* String library */

#include "ias_satellite_attributes.h"
#include "ias_logging.h"        /* IAS Logging library */
#include "focal_plane_setup.h"

int check_images_match
(
    const char *ref_img_name,   /* I: Name of the reference image */
    const char *srch_img_name,  /* I: Name of the 1T image */
    const char *dem_img_name,   /* I: Name of the DEM image */
    int srch_band_number        /* I: Band number of search image */
)
{
    int band_number;            /* Current band number */
    int number_of_bands;        /* Number of bands present */
    int band_index;             /* Index for band number */
    int images_match_flag = SUCCESS; /* Flag whether the images match */

    int ref_nl;                 /* Number of reference lines */
    int ref_ns;                 /* Number of reference samples */
    int srch_nl;                /* Number of search lines */
    int srch_ns;                /* Number of search samples */
    int dem_nl;                 /* Number of DEM lines */
    int dem_ns;                 /* Number of DEM samples */
    int nscas;                  /* Number of SCAs */

    double corner_off_y;        /* Amount corner can be off in y */
    double corner_off_x;        /* Amount corner can be off in x */

    IAS_DATA_TYPE dtype;        /* Image data type code */

    L1GIO *ref_id;              /* Reference image file handle */
    L1GIO *srch_id;             /* Search image file handle */
    L1GIO *dem_id;              /* DEM file handle */

    IAS_L1G_FILE_METADATA ref_fmd;  /* Reference image file metadata */
    IAS_L1G_FILE_METADATA srch_fmd; /* Search image file metadata */
    IAS_L1G_BAND_METADATA ref_bmd;  /* Reference image band metadata */
    IAS_L1G_BAND_METADATA srch_bmd; /* Search image band metadata */
    IAS_L1G_BAND_METADATA dem_bmd;  /* DEM band metadata */

    /* Open reference image. */
    ref_id = ias_l1g_open_image(ref_img_name, IAS_READ);
    if (ref_id == NULL)
    {
        IAS_LOG_ERROR("Opening reference image %s", ref_img_name);
        return ERROR;
    }

    /* Initialize band number to an invalid value. */
    band_number = 0;
 
    /* Get the number of bands. */
    number_of_bands = ias_l1g_get_band_count(ref_id);
    if ((number_of_bands == ERROR) || (number_of_bands < 1))
    {
        IAS_LOG_ERROR("Failed to retrieve the band count for reference image "
            "%s", ref_img_name);
        ias_l1g_close_image(ref_id);
        ref_id = NULL;
        return ERROR;
    }
    else 
    {
        if (number_of_bands > 1)
        {
            IAS_LOG_WARNING("There is more than 1 band in reference image %s",
                ref_img_name);
        }
    }

    /* Find the identity of the reference band. */
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Invalid band number for band index %d", 
                band_index);
            ias_l1g_close_image(ref_id);
            ref_id = NULL;
            return ERROR;
        }

        if (ias_l1g_is_band_present(ref_id, band_number))
        {
            break; 
        }
    }

    /* Get reference image file metadata. */
    if (ias_l1g_get_file_metadata(ref_id, &ref_fmd) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading reference file %s metadata", ref_img_name);
        ias_l1g_close_image(ref_id);
        ref_id = NULL;
        return ERROR;
    }

    /* Get reference image band metadata for the band of interest. */
    if (ias_l1g_get_band_metadata(ref_id, band_number, &ref_bmd) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading reference file %s band %d metadata", 
            ref_img_name, band_number);
        ias_l1g_close_image(ref_id);
        ref_id = NULL;
        return ERROR;
    }

    /* Get reference band size. */
    if (ias_l1g_get_band_size(ref_id, band_number, &dtype, &nscas, 
        &ref_nl, &ref_ns) != SUCCESS)
    {
        IAS_LOG_ERROR("Getting reference image %s band %d size", ref_img_name, 
            band_number);
        ias_l1g_close_image(ref_id);
        ref_id = NULL;
        return ERROR;
    }

    /* Close the reference image. */
    if (ias_l1g_close_image(ref_id) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing reference image %s", ref_img_name);
        return ERROR;
    }
    ref_id = NULL;

    /* Open search image. */
    srch_id = ias_l1g_open_image(srch_img_name, IAS_READ);
    if (srch_id == NULL)
    {
        IAS_LOG_ERROR("Opening search image %s", srch_img_name);
        return ERROR;
    }

    /* Get search image file metadata. */
    if (ias_l1g_get_file_metadata(srch_id, &srch_fmd) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading search file %s metadata", srch_img_name);
        ias_l1g_close_image(srch_id);
        srch_id = NULL;
        return ERROR;
    }

    /* Get search image band metadata for the band of interest. */
    if (ias_l1g_get_band_metadata(srch_id, srch_band_number, &srch_bmd) 
        != SUCCESS)
    {
        IAS_LOG_ERROR("Reading search file %s band %d metadata", 
            srch_img_name, srch_band_number);
        ias_l1g_close_image(srch_id);
        srch_id = NULL;
        return ERROR;
    }

    /* Get search band size. */
    if (ias_l1g_get_band_size(srch_id, srch_band_number, &dtype, &nscas, 
        &srch_nl, &srch_ns) != SUCCESS)
    {
        IAS_LOG_ERROR("Getting search image %s band %d size", srch_img_name, 
            srch_band_number);
        ias_l1g_close_image(srch_id);
        srch_id = NULL;
        return ERROR;
    }

    /* Close the search image. */
    if (ias_l1g_close_image(srch_id) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing search image %s", srch_img_name);
        return ERROR;
    }
    srch_id = NULL;

    /* Open DEM image. */
    dem_id = ias_l1g_open_image(dem_img_name, IAS_READ);
    if (dem_id == NULL)
    {
        IAS_LOG_ERROR("Opening DEM image %s", dem_img_name);
        return ERROR;
    }

    /* Initialize band number to an invalid value. */
    band_number = 0;
 
    /* Get the number of bands. */
    number_of_bands = ias_l1g_get_band_count(dem_id);
    if ((number_of_bands == ERROR) || (number_of_bands < 1))
    {
        IAS_LOG_ERROR("Failed to retrieve the band count from DEM file %s",
            dem_img_name);
        ias_l1g_close_image(dem_id);
        dem_id = NULL;
        return ERROR;
    }
    else 
    {
        if (number_of_bands > 1)
        {
            IAS_LOG_WARNING("There is more than 1 band in DEM file %s",
                dem_img_name);
        }
    }

    /* Find the identity of the DEM band. */
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Invalid band number for band index %d", 
                band_index);
            ias_l1g_close_image(dem_id);
            dem_id = NULL;
            return ERROR;
        }

        if (ias_l1g_is_band_present(dem_id, band_number))
        {
            break; 
        }
    }

    /* Get DEM image band metadata for the band of interest. */
    if (ias_l1g_get_band_metadata(dem_id, band_number, &dem_bmd) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading DEM file %s band %d metadata", dem_img_name, 
            band_number);
        ias_l1g_close_image(dem_id);
        dem_id = NULL;
        return ERROR;
    }

    /* Get DEM band size. */
    if (ias_l1g_get_band_size(dem_id, band_number, &dtype, &nscas, &dem_nl, 
        &dem_ns) != SUCCESS) 
    {
        IAS_LOG_ERROR("Getting DEM image %s band %d size", dem_img_name, 
            band_number);
        ias_l1g_close_image(dem_id);
        dem_id = NULL;
        return ERROR;
    }

    /* Close the DEM image. */
    if (ias_l1g_close_image(dem_id) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing DEM image %s", dem_img_name);
        return ERROR;
    }
    dem_id = NULL;

    /* Check that the search and reference images are nadir-viewing.  Produce 
       warning messages if not. */
    if ((srch_fmd.roll_angle > IAS_OFF_NADIR_BOUNDARY) ||
        (srch_fmd.roll_angle < -IAS_OFF_NADIR_BOUNDARY))
    {
        IAS_LOG_WARNING("Search image %s is not nadir-viewing; roll angle is "
            "%f", srch_img_name, srch_fmd.roll_angle);
    }

    if ((ref_fmd.roll_angle > IAS_OFF_NADIR_BOUNDARY) ||
        (ref_fmd.roll_angle < -IAS_OFF_NADIR_BOUNDARY))
    {
        IAS_LOG_WARNING("Reference image %s is not nadir-viewing; roll angle "
            "is %f", ref_img_name, ref_fmd.roll_angle);
    }

    /* Check that the search and reference images are L1T, and issue warnings
       if not. */
    if (strcasecmp(srch_fmd.correction_type, "L1T") != 0)
    {
        IAS_LOG_WARNING("Search image %s is not terrain-corrected", 
            srch_img_name);
    }

    if (strcasecmp(ref_fmd.correction_type, "L1T") != 0)
    {
        IAS_LOG_WARNING("Reference image %s is not terrain-corrected", 
            ref_img_name);
    }

    /* Check that 1T and reference image corners are within 0.5 pixel. */
    corner_off_y = srch_bmd.projection_distance_y * 0.5;
    corner_off_x = srch_bmd.projection_distance_x * 0.5;
    if ((fabs(srch_bmd.upper_left_y  - ref_bmd.upper_left_y)  > corner_off_y) ||
        (fabs(srch_bmd.upper_left_x  - ref_bmd.upper_left_x)  > corner_off_x) ||
        (fabs(srch_bmd.lower_left_y  - ref_bmd.lower_left_y)  > corner_off_y) ||
        (fabs(srch_bmd.lower_left_x  - ref_bmd.lower_left_x)  > corner_off_x) ||
        (fabs(srch_bmd.upper_right_y - ref_bmd.upper_right_y) > corner_off_y) ||
        (fabs(srch_bmd.upper_right_x - ref_bmd.upper_right_x) > corner_off_x) ||
        (fabs(srch_bmd.lower_right_y - ref_bmd.lower_right_y) > corner_off_y) ||
        (fabs(srch_bmd.lower_right_x - ref_bmd.lower_right_x) > corner_off_x))
    {
        IAS_LOG_ERROR("Search image %s corners do not match reference image "
            "%s corners", srch_img_name, ref_img_name);
        images_match_flag = ERROR;
    }

    /* Check that 1T and reference image sizes match. */
    if ((srch_nl != ref_nl) || (srch_ns != ref_ns))
    {
        IAS_LOG_ERROR("Search image %s size does not match reference image "
            "%s size", srch_img_name, ref_img_name);
        images_match_flag = ERROR;
    }

    /* DEM can be different resolution so just check corners, not sizes.
       Check that the corners are within a half of a DEM pixel. */
    corner_off_y = dem_bmd.projection_distance_y * 0.5;
    corner_off_x = dem_bmd.projection_distance_x * 0.5;
    if ((fabs(srch_bmd.upper_left_y  - dem_bmd.upper_left_y)  > corner_off_y) ||
        (fabs(srch_bmd.upper_left_x  - dem_bmd.upper_left_x)  > corner_off_x) ||
        (fabs(srch_bmd.lower_left_y  - dem_bmd.lower_left_y)  > corner_off_y) ||
        (fabs(srch_bmd.lower_left_x  - dem_bmd.lower_left_x)  > corner_off_x) ||
        (fabs(srch_bmd.upper_right_y - dem_bmd.upper_right_y) > corner_off_y) ||
        (fabs(srch_bmd.upper_right_x - dem_bmd.upper_right_x) > corner_off_x) ||
        (fabs(srch_bmd.lower_right_y - dem_bmd.lower_right_y) > corner_off_y) ||
        (fabs(srch_bmd.lower_right_x - dem_bmd.lower_right_x) > corner_off_x)) 
    {
        IAS_LOG_ERROR("Search image %s corners do not match DEM image %s "
            "corners", srch_img_name, dem_img_name);
        images_match_flag = ERROR;
    }

    return images_match_flag;
}
