/******************************************************************************
NAME:           focal_plane_setup (main)

PURPOSE:        
Main driver for the setup portion of Focal Plane Calibration.  
The setup portion consists of correlating points between a search image
and a reference image, and converting the correlation offsets into
line-of-sight deviations that can be used to model each SCA's detector
array, modelled by a quadratic Legendre polynomial.
The results of this program are used in the second portion of Focal Plane 
Calibration, the generation of new Legendre polynomials.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
EXIT_FAILURE    Error in setup; no output data file produced 
EXIT_SUCCESS    Successful setup

NOTES:
This program creates a temporary output file that is read by the second 
portion of Focal Plane Calibration.  
******************************************************************************/

#include <stdlib.h>

#include "ias_logging.h"            /* IAS Logging library */
#include "ias_model_io.h"           /* Model I/O library */
#include "focal_plane_setup.h"

int main
(
    int argc, 
    char *argv[]
)
{
    ALIGNMENT_PARMS parms;          /* Parameters read from parameter file;
                                       done to preserve the sizeof() arrays
                                       when passing into a function */
    CORR_PT_INFO *fp_corr_pts = NULL; /* Information about each correlation
                                       point */
    IAS_CPF *cpf;                   /* CPF parameters */
    IAS_LOS_MODEL *model;           /* Model structure */
    IAS_GRID_TYPE grid;             /* Grid for reference band, including
                                       header */
    int grid_band_number;           /* Band number read for grid file */
    int status;                     /* Function return status */
    int npts = 0;                   /* Number of correlation points */
    int abs_corr_coeff = FALSE;     /* Flag indicating the absolute value 
                                       of correlation coeffs is not used */

    /* Initialize the logging library */
    if (ias_log_initialize("focal_plane_setup") != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing logging library");
        exit(EXIT_FAILURE);
    }

    /* Get parameters and correlation information */
    status = get_parameters(argc, argv, &parms, &cpf); 
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting parameters");
        exit(EXIT_FAILURE);
    }

    /* Log initiation */
    IAS_LOG_INFO("Initiated");

    /* Initialize satellite attributes */
    status = ias_sat_attr_initialize(parms.satellite_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attributes library");
        exit(EXIT_FAILURE);
    }

    /* Done with CPF */
    ias_cpf_free(cpf);

    /* Get the reference band number, which will be the grid band */
    if (parms.processing_type == OLI_FOCAL_PLANE_ALIGNMENT)
    {
        grid_band_number = ias_sat_attr_get_band_number_from_type(IAS_PAN_BAND);
        if (grid_band_number == ERROR)
        {
            IAS_LOG_ERROR("Invalid PAN band number");
            exit(EXIT_FAILURE);
        }
    }
    else /* TIRS Alignment */
    {
        grid_band_number = ias_sat_attr_get_band_number_from_type(
            IAS_THERMAL1_BAND);
        if (grid_band_number == ERROR)
        {
            IAS_LOG_ERROR("Invalid 10.8um thermal band number");
            exit(EXIT_FAILURE);
        }
    }

    /* Make sure the images all match up */
    if (check_images_match(parms.ref_img_name, parms.terr_img_name, 
                           parms.dem_img_name, grid_band_number) != SUCCESS) 
    {
        IAS_LOG_ERROR("Focal Plane Calibration setup images %s (reference), "
            "%s (search), and %s (DEM) don't match", parms.ref_img_name,
            parms.terr_img_name, parms.dem_img_name);
        exit(EXIT_FAILURE);
    }

    /* Get the grid and related info */
    IAS_LOG_INFO("Setting up grid"); 
    if (set_up_grid(parms.precision_grid_name, grid_band_number, 
        parms.processing_type, &grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Setting up grid %s band number %d", 
            parms.precision_grid_name, grid_band_number);
        exit(EXIT_FAILURE);
    }

    /* Get the model */
    model = ias_model_read(parms.model_file_name);
    if (model == NULL)
    {
        IAS_LOG_ERROR("Reading model file %s", parms.model_file_name);
        ias_grid_free(&grid);
        exit(EXIT_FAILURE);
    }

    /* Select the correlation points (pts_per_cell per grid cell) */
    IAS_LOG_INFO("Selecting correlation points");
    fp_corr_pts = select_corr_pts(&grid, parms.pts_per_cell, grid_band_number,
        &npts); 
    if (fp_corr_pts == NULL)
    {
        IAS_LOG_ERROR("Selecting correlation points from grid using band "
            "number %d", grid_band_number);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        model = NULL;
        exit(EXIT_FAILURE);
    }

    /* Do correlation & then calculate errors in input space lines/samples */
    if (calc_input_space_errors(parms.ref_img_name, 
            parms.terr_img_name, parms.dem_img_name,
            parms.corr_win_nl, parms.corr_win_ns, abs_corr_coeff, 
            parms.min_corr, parms.max_disp, parms.fit_method, 
            &grid, parms.threshold, grid_band_number, fp_corr_pts, 
            npts) != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating input space errors using reference image "
            "%s and search image %s", parms.ref_img_name, parms.terr_img_name);
        free(fp_corr_pts);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        model = NULL;
        exit(EXIT_FAILURE);
    }

    /* After correlation & initializing terrain table we're done with the 
       grid */
    ias_grid_free(&grid);

    /* Calculate line-of-sight errors from the input space error info */
    if (calc_los_errors(fp_corr_pts, npts, model, grid_band_number) 
            != SUCCESS) 
    {
        IAS_LOG_ERROR("Calculating line-of-sight errors");
        free(fp_corr_pts);
        ias_los_model_free(model);
        model = NULL;
        exit(EXIT_FAILURE);
    }

    /* Output the correlation info to the temp file */
    if (output_correlation_info(parms.correlation_file_name,
                                fp_corr_pts, npts) != SUCCESS)
    {
        IAS_LOG_ERROR("Saving correlation data");
        free(fp_corr_pts);
        ias_los_model_free(model);
        model = NULL;
        exit(EXIT_FAILURE);
    }

    /* Finish up */
    free(fp_corr_pts);
    ias_los_model_free(model);
    model = NULL;

    IAS_LOG_INFO("Successful completion");

    return EXIT_SUCCESS;
}
