/******************************************************************************
NAME: map_coords_to_input_space

PURPOSE:
Use the inverse coefficients in the grid to calculate the input space 
line/sample for each output space line/sample.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
ERROR           Error converting a point from output space to input space
SUCCESS         Coordinates mapped to input space

NOTES:
This function calculates the ref_is_pt corresponding to each ref_os_pt and the 
srch_is_pt corresponding to each srch_os_pt.
******************************************************************************/
#include <stdlib.h>

#include "ias_miscellaneous.h"  /* IAS miscellaneous libraries */
#include "ias_logging.h"        /* IAS Logging library */
#include "focal_plane_setup.h"


int map_coords_to_input_space
(
    CORR_PT_INFO *fp_corr_pts,  /* I/O: Focal plane correlation points */
    const char *dem_name,       /* I: DEM file name */
    int npts,                   /* I: Number of focal plane corr points */
    const IAS_GRID_TYPE *grid,  /* I: Grid structure for all bands */
    int band_number             /* I: Band number (pan) */
)
{
    int i;                      /* Looping variable */
    int band_index;             /* Index of band */
    int sca_index;              /* Index of SCA */
    long input_nl;              /* Input space number of lines */
    long input_ns;              /* Input space number of samples */
    long output_nl;             /* Output space number of lines */
    long output_ns;             /* Output space number of samples */
    double pixsize;             /* Terrain/image pixel size */
    double ref_elev;            /* Reference point elevation */
    double srch_elev;           /* Search point elevation */
    IAS_IMAGE dem_image;        /* DEM data */

    /* Look up the band index */
    band_index = ias_sat_attr_convert_band_number_to_index(band_number);
    if (band_index == ERROR)
    {
        IAS_LOG_ERROR("Invalid band index for band number %d", band_number);
        return ERROR;
    }

    /* Set the input image size using the grid info */
    input_nl = grid->gridbands[band_index][0].
        in_lines[grid->gridbands[band_index][0].ngrid_lines-1];
    input_ns = grid->gridbands[band_index][0].
        in_samps[grid->gridbands[band_index][0].ngrid_samps-1];
    output_nl = grid->gridbands[band_index][0].lines;
    output_ns = grid->gridbands[band_index][0].samps;
    pixsize = grid->gridbands[band_index][0].pixsize;

    /* First read the DEM */
    if (ias_misc_read_coregistered_single_band_l1g(dem_name, &grid->corners,
            output_nl, output_ns, pixsize, &dem_image, IAS_I2) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading DEM from %s", dem_name);
        return ERROR;
    }

    /* Then map the points back to input space */
    for (i = 0; i < npts; i++)
    {
        int status;
        if (fp_corr_pts[i].active_flag == ACTIVE)
        {
            /* Get the point elevation from the DEM */
            if (ias_misc_read_elevation_at_line_sample(&dem_image, 
                fp_corr_pts[i].ref_os_pt.line,
                fp_corr_pts[i].ref_os_pt.samp, pixsize, pixsize, &ref_elev) 
                != SUCCESS)
            {
                IAS_LOG_ERROR("Getting elevation from %s at line %f sample %f",
                    dem_name, fp_corr_pts[i].ref_os_pt.line,
                    fp_corr_pts[i].ref_os_pt.samp);
                free(dem_image.data);
                return ERROR;
            }

            /* Set SCA index */
            sca_index = fp_corr_pts[i].sca_index;

            /* Convert the reference point to input space */
            status = ias_grid_3d_ols2ils(
                &grid->gridbands[band_index][sca_index],
                fp_corr_pts[i].ref_os_pt.line, fp_corr_pts[i].ref_os_pt.samp, 
                &fp_corr_pts[i].ref_is_pt.line, &fp_corr_pts[i].ref_is_pt.samp, 
                ref_elev);

            if (status == ERROR)
            {
                IAS_LOG_ERROR("Mapping output to input space for band index "
                    "%d SCA index %d input line %f input sample %f",
                    band_index, sca_index, fp_corr_pts[i].ref_os_pt.line,
                    fp_corr_pts[i].ref_is_pt.samp);
                free(dem_image.data);
                return ERROR;
            }

            if ((status != SUCCESS)
                || (fp_corr_pts[i].ref_is_pt.line < 1)
                || (fp_corr_pts[i].ref_is_pt.line >= (input_nl - 2))
                || (fp_corr_pts[i].ref_is_pt.samp < 1)
                || (fp_corr_pts[i].ref_is_pt.samp >= (input_ns - 2)))
            {
                IAS_LOG_DEBUG("Calculating reference input line/sample "
                    "for point %d at output line/sample %f %f",
                    i, fp_corr_pts[i].ref_os_pt.line, 
                    fp_corr_pts[i].ref_os_pt.samp);

                /* Outside Image */
                fp_corr_pts[i].active_flag = OUTSIDE_IMAGE;
                fp_corr_pts[i].ref_os_pt.line  = 0.0;
                fp_corr_pts[i].ref_os_pt.samp  = 0.0;
                fp_corr_pts[i].srch_os_pt.line = 0.0;
                fp_corr_pts[i].srch_os_pt.samp = 0.0;
                fp_corr_pts[i].ref_is_pt.line  = 0.0;
                fp_corr_pts[i].ref_is_pt.samp  = 0.0;
                fp_corr_pts[i].srch_is_pt.line = 0.0;
                fp_corr_pts[i].srch_is_pt.samp = 0.0;
                continue;
            }

            /* Get the point elevation from the DEM */
            if (ias_misc_read_elevation_at_line_sample(&dem_image, 
                fp_corr_pts[i].srch_os_pt.line, fp_corr_pts[i].srch_os_pt.samp,
                pixsize, pixsize, &srch_elev) != SUCCESS)
            {
                IAS_LOG_ERROR("Getting elevation from %s at line %f sample %f",
                    dem_name, fp_corr_pts[i].srch_os_pt.line,
                    fp_corr_pts[i].srch_os_pt.samp);
                free(dem_image.data);
                return ERROR;
            }

            /* Convert the search point to input space */
            status = ias_grid_3d_ols2ils(
                &grid->gridbands[band_index][sca_index],
                fp_corr_pts[i].srch_os_pt.line,
                fp_corr_pts[i].srch_os_pt.samp, 
                &fp_corr_pts[i].srch_is_pt.line,
                &fp_corr_pts[i].srch_is_pt.samp, 
                srch_elev);

            if (status == ERROR)
            {
                IAS_LOG_ERROR("Mapping output to input space for band index "
                    "%d SCA index %d input line %f input sample %f",
                    band_index, sca_index, fp_corr_pts[i].srch_os_pt.line,
                    fp_corr_pts[i].srch_is_pt.samp);
                free(dem_image.data);
                return ERROR;
            }

            if ((status != SUCCESS)
                || (fp_corr_pts[i].srch_is_pt.line < 1)
                || (fp_corr_pts[i].srch_is_pt.line >= (input_nl - 2))
                || (fp_corr_pts[i].srch_is_pt.samp < 1)
                || (fp_corr_pts[i].srch_is_pt.samp >= (input_ns - 2)))
            {
                IAS_LOG_DEBUG("Correlation failure for point %d at output "
                   "line/sample %f %f, setting corresponding correlation "
                   "point info to zero",
                   i, fp_corr_pts[i].srch_os_pt.line,
                   fp_corr_pts[i].srch_os_pt.samp);

                /* CORRELATION FAILURE */
                fp_corr_pts[i].active_flag = CORR_FAIL;
                fp_corr_pts[i].ref_os_pt.line  = 0.0;
                fp_corr_pts[i].ref_os_pt.samp  = 0.0;
                fp_corr_pts[i].srch_os_pt.line = 0.0;
                fp_corr_pts[i].srch_os_pt.samp = 0.0;
                fp_corr_pts[i].ref_is_pt.line  = 0.0;
                fp_corr_pts[i].ref_is_pt.samp  = 0.0;
                fp_corr_pts[i].srch_is_pt.line = 0.0;
                fp_corr_pts[i].srch_is_pt.samp = 0.0;
            }
        }
    }

    /* Free the DEM memory */
    free(dem_image.data);

    return SUCCESS;
}
