/******************************************************************************
NAME: output_correlation_info

PURPOSE:
Dump out the correlation information to a file.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
ERROR           Error creating output file
SUCCESS         Successully created file

NOTES:
The correlation points are dumped to a binary file, so the second phase of 
Focal Plane Calibration (Legendre polynomial generation) can read them 
directly back in.  First a long integer is written to indicate the number of 
records, then all the records are written.

This is NOT a human-readable (ASCII) file.  It is only used to transport
information from the first phase of calibration to the second.

If the file exists, it will be overwritten if permissions allow.
******************************************************************************/
#include "ias_logging.h"               /* IAS Logging library */
#include "focal_plane_setup.h"

int output_correlation_info
(
    const char *correlation_file_name, /* I: Name of output file */
    const CORR_PT_INFO *fp_corr_pts,   /* I: Correlation point info */
    int npts                           /* I: Number of points */
)
{
    FILE *corrfile;                    /* Pointer to output file */

    /* Open the file */
    corrfile = fopen(correlation_file_name, "wb");
    if (corrfile == NULL)
    {
        IAS_LOG_ERROR("Opening output data file %s", correlation_file_name);
        return ERROR;
    }

    /* First write a single int indicating the number of records */
    if (fwrite(&npts, sizeof(npts), 1, corrfile) != 1)
    {
        IAS_LOG_ERROR("Writing record count to output data file %s",
            correlation_file_name);
        fclose(corrfile);
        return ERROR;
    }

    /* Write all the structures out */
    if (fwrite(fp_corr_pts, sizeof(CORR_PT_INFO), (size_t)npts, corrfile) !=
        npts)
    {
        IAS_LOG_ERROR("Writing correlation information to output data file %s",
            correlation_file_name);
        fclose(corrfile);
        return ERROR;
    }

    /* Done; close file */
    if (fclose(corrfile) != 0)
    {
        IAS_LOG_ERROR("Closing output data file %s", correlation_file_name);
        return ERROR;
    }

    return SUCCESS;
}
