/******************************************************************************
NAME: perform_correlation

PURPOSE:
Do correlation at each point.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
ERROR           Error trying to correlate a point
SUCCESS         All points attempted correlation

NOTES:
This function updates the following fields in the CORR_PT_INFO structures:

        ref_os_pt
        srch_os_pt
        active_flag (based on correlation failure/success or outside image)

Pan band of the 1T image and the first band of the reference image are used.
The low_peak_flag returned from correlation is not checked.
******************************************************************************/
#include <stdlib.h>

#include "ias_math_parallel_corr.h" /* parallel correlator */
#include "ias_satellite_attributes.h"
#include "ias_logging.h"            /* IAS Logging library */
#include "focal_plane_setup.h"

int perform_correlation
(
    CORR_PT_INFO *fp_corr_pts,  /* I/O: Correlation point information */
    int npts,                   /* I: Total number of points */
    const char *ref_img,        /* I: Name of reference image */
    const char *srch_img,       /* I: Name of search (1T) image */
    int corr_win_nl,            /* I: Line size of corr window */
    int corr_win_ns,            /* I: Sample size of corr window */
    float min_corr,             /* I: Minimum correlation strength */
    float max_disp,             /* I: Maximum correlation displacement */
    IAS_CORRELATION_FIT_TYPE fit_method, /* I: Correlation result fit method */
    int abs_corr_coeff,         /* I: Flag to use abs value of corr coeffs */
    float invalid_threshold,    /* I: Invalid threshold percent */
    int srch_band_number        /* I: Search (1T) image band number */
)
{
    L1GIO *ref_id;              /* Reference image file handle */
    L1GIO *srch_id;             /* Search image file handle */

    int status;                 /* Function return status */
    int cursamp;                /* Sample within correlation window */

    int curpt;                  /* Point looping variable */
    int count0;                 /* Count of 0 fill values */
    int window_size[2];         /* Size of correlation window NOTE:
                                   samps, lines for use by grey_corr */
    int extract_window_size[2]; /* Size of correlation window  NOTE:
                                   lines, samps for use by extract_window */

    float *ref_buf = NULL;      /* Buffer of ref img pixels */
    float *srch_buf = NULL;     /* Buffer of srch img pixels */
    float nom_off[2];           /* Nominal correlation offset */
    int ref_band_index;         /* Reference image band index */
    int ref_band_number;        /* Reference image band number */

    IAS_DATA_TYPE ref_dtype;    /* Reference image data type */
    IAS_DATA_TYPE srch_dtype;   /* Search image data type */
    int num_scas;               /* Number of SCAs in band */
    int ref_nl;                 /* Reference number of lines in band */
    int ref_ns;                 /* Reference number of samples in band */
    int srch_nl;                /* Search number of lines in band */
    int srch_ns;                /* Search number of samples in band */
    L1G_BAND_IO *ref_l1g_band;  /* Reference L1G_BAND_IO pointer */
    L1G_BAND_IO *srch_l1g_band; /* Search L1G_BAND_IO pointer */

    IAS_PARALLEL_CORRELATOR_TYPE *correlator_ptr = NULL; /* Pointer to 
                                   parallel correlator data structure */
    IAS_CORRELATION_RESULT_TYPE *result_ptr = NULL; /* Pointer to the 
                                   correlation results */
    int need_buffers = 1;       /* Flag to indicate buffers are needed 
                                   for the reference and search chip */
    int upper_line;             /* The upper line in the image for the chip */
    int left_sample;            /* The left sample in the image for the chip */

    /* Set up correlation info and image buffers */
    window_size[0] = corr_win_ns;
    window_size[1] = corr_win_nl;
    extract_window_size[1] = corr_win_ns;
    extract_window_size[0] = corr_win_nl;

    /* Reference and search windows are the same size, assumed to line up */
    nom_off[0] = 0.0; 
    nom_off[1] = 0.0; 

    /* Open the reference image */
    ref_id = ias_l1g_open_image(ref_img, IAS_READ);
    if (ref_id == NULL)
    {
        IAS_LOG_ERROR("Opening reference image %s", ref_img);
        return ERROR;
    }

    /* Just use the 1st band found of the reference image */
    for (ref_band_index = 0; ref_band_index < IAS_MAX_NBANDS; ref_band_index++)
    {
        ref_band_number = ias_sat_attr_convert_band_index_to_number(
            ref_band_index);
        if (ref_band_number == ERROR)
        {
            IAS_LOG_ERROR("Invalid band number for reference band index %d",
                ref_band_index);
            ias_l1g_close_image(ref_id);
            return ERROR;
        }

        if (ias_l1g_is_band_present(ref_id, ref_band_number))
        {
             /* Found the band; move along */
             break;
        }
    }

    if (ref_band_index == IAS_MAX_NBANDS)
    { 
        IAS_LOG_ERROR("Reference image %s does not contain a band", ref_img);
        ias_l1g_close_image(ref_id);
        return ERROR;
    }

    ref_l1g_band = ias_l1g_open_band(ref_id, ref_band_number, &ref_dtype, 
        &num_scas, &ref_nl, &ref_ns); 
    if (ref_l1g_band == NULL)
    {
        IAS_LOG_ERROR("Opening band %d of reference image %s", 
            ref_band_number, ref_img);
        ias_l1g_close_image(ref_id);
        return ERROR;
    }

    if (num_scas != 1)
    {
        IAS_LOG_ERROR("Reference image %s must be SCA-combined", ref_img);
        ias_l1g_close_band(ref_l1g_band);
        ias_l1g_close_image(ref_id);
        return ERROR;
    }

    /* Open the search image */
    srch_id = ias_l1g_open_image(srch_img, IAS_READ);
    if (srch_id == NULL)
    {
        IAS_LOG_ERROR("Opening search image %s", srch_img);
        ias_l1g_close_band(ref_l1g_band);
        ias_l1g_close_image(ref_id);
        return ERROR;
    }

    /* Look specifically for the pan band of the search (1T) image */
    srch_l1g_band = ias_l1g_open_band(srch_id, srch_band_number, &srch_dtype, 
        &num_scas, &srch_nl, &srch_ns);
    if (srch_l1g_band == NULL)
    {
        IAS_LOG_ERROR("Opening band %d of search image %s", srch_band_number, 
            srch_img); 
        ias_l1g_close_band(ref_l1g_band);
        ias_l1g_close_image(ref_id);
        ias_l1g_close_image(srch_id);
         return ERROR;
    }

    /* Allocate memory for the results of the correlation */
    result_ptr = malloc(npts * sizeof(IAS_CORRELATION_RESULT_TYPE));
    if (result_ptr == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for correlation results for up to "
            "%d points", npts);
        ias_l1g_close_band(ref_l1g_band);
        ias_l1g_close_image(ref_id);
        ias_l1g_close_band(srch_l1g_band);
        ias_l1g_close_image(srch_id);
        return ERROR;
    }

    /* Initialize the parallel correlator */
    correlator_ptr = ias_math_init_parallel_correlator(1, 1, 0,
                            1 /* don't check low peak*/,
                            window_size[1],window_size[0],
                            window_size[1],window_size[0], result_ptr);
    if (correlator_ptr == NULL)
    {
        IAS_LOG_ERROR("Creating the parallel correlator");
        free(result_ptr);
        ias_l1g_close_band(ref_l1g_band);
        ias_l1g_close_image(ref_id);
        ias_l1g_close_band(srch_l1g_band);
        ias_l1g_close_image(srch_id);
        return ERROR;
    }

    /* Initialize the results */
    for (curpt = 0; curpt < npts; curpt++)
    {
        result_ptr[curpt].valid = FALSE;
        result_ptr[curpt].fit_offset[0] = 0.0;
        result_ptr[curpt].fit_offset[1] = 0.0;
    }

    /* Correlate each point */
    for (curpt = 0; curpt < npts; curpt++)
    {
        /* Occasional status message */
        if ((curpt % 5000) == 0)
        {
            IAS_LOG_INFO("Correlating point %d", curpt);
        }

        /* Get buffers for the chips if needed */
        if (need_buffers)
        {
            if (ias_math_get_corr_chip_buffers(correlator_ptr, &ref_buf, 
                &srch_buf) != SUCCESS)
            {
                IAS_LOG_ERROR("Getting buffers for chips for search point %d", 
                    curpt);
                ias_math_close_parallel_correlator(correlator_ptr);
                free(result_ptr);
                ias_l1g_close_band(ref_l1g_band);
                ias_l1g_close_image(ref_id);
                ias_l1g_close_band(srch_l1g_band);
                ias_l1g_close_image(srch_id);
                return ERROR;
            }
            need_buffers = 0;
        }

        /* Extract the window surrounding the point for the search image */
        status = ias_geo_extract_window(srch_nl, srch_ns, 
                   fp_corr_pts[curpt].sca_index, srch_dtype,
                   fp_corr_pts[curpt].nom_os_pt.line, 
                   fp_corr_pts[curpt].nom_os_pt.samp, 
                   extract_window_size, srch_l1g_band, srch_buf,
                   &upper_line, &left_sample);

        if (status == ERROR)
        {
            IAS_LOG_ERROR("Extracting window around search point %d (%f %f)",
                curpt, fp_corr_pts[curpt].nom_os_pt.line, 
                fp_corr_pts[curpt].nom_os_pt.samp);
            ias_math_close_parallel_correlator(correlator_ptr);
            free(result_ptr);        
            ias_l1g_close_band(ref_l1g_band);
            ias_l1g_close_image(ref_id);
            ias_l1g_close_band(srch_l1g_band);
            ias_l1g_close_image(srch_id);
            return ERROR;
        }
        else if (status == OUTSIDE)
        {
            fp_corr_pts[curpt].active_flag = OUTSIDE_IMAGE;
            continue;
        }

       /* Check the search image window for fill values (over invalid_
          threshold%, don't correlate).  Always check for fill value of 0.0 */
        for (cursamp=0, count0=0; cursamp<window_size[0] * window_size[1];
             cursamp++)
        {
            if (srch_buf[cursamp] == 0.0) 
                count0++;
        }
        if (count0 > (window_size[0] * window_size[1] * invalid_threshold))
        {
            fp_corr_pts[curpt].active_flag = OUTSIDE_IMAGE;
            continue;
        }

        /* Extract the window surrounding the point for the reference image */
        status = ias_geo_extract_window(ref_nl, ref_ns, 0, ref_dtype,
                   fp_corr_pts[curpt].nom_os_pt.line, 
                   fp_corr_pts[curpt].nom_os_pt.samp, 
                   extract_window_size, ref_l1g_band, ref_buf,
                   &upper_line, &left_sample);

        if (status == ERROR)
        {
            IAS_LOG_ERROR(
                "Extracting window around reference point %d (%f %f)",
                curpt, fp_corr_pts[curpt].nom_os_pt.line, 
                fp_corr_pts[curpt].nom_os_pt.samp);
            ias_math_close_parallel_correlator(correlator_ptr);
            free(result_ptr);        
            ias_l1g_close_band(ref_l1g_band);
            ias_l1g_close_image(ref_id);
            ias_l1g_close_band(srch_l1g_band);
            ias_l1g_close_image(srch_id);
            return ERROR;
        }
        else if (status == OUTSIDE)
        {
            /* Depending on how the image was framed, a point selected from 
               the grid may not appear in the output image.  Grid cells are 
               created for for the entire width of each scan, but all the 
               input data may not be framed into the output space. If this 
               point was not framed in, that's ok-just mark it as outside so 
               we don't try to do any more processing on this point & go on */
            fp_corr_pts[curpt].active_flag = OUTSIDE_IMAGE;
            continue;
        }

        /* Do the correlation */
        status = ias_math_submit_chip_to_corr(correlator_ptr, curpt, srch_buf, 
            ref_buf, window_size, window_size, min_corr, fit_method, max_disp, 
            nom_off, abs_corr_coeff);
        if (status == ERROR)
        {
            IAS_LOG_ERROR("Correlating point %d (%f %f)",
                curpt, fp_corr_pts[curpt].nom_os_pt.line, 
                fp_corr_pts[curpt].nom_os_pt.samp);
            ias_math_close_parallel_correlator(correlator_ptr);
            free(result_ptr);        
            ias_l1g_close_band(ref_l1g_band);
            ias_l1g_close_image(ref_id);
            ias_l1g_close_band(srch_l1g_band);
            ias_l1g_close_image(srch_id);
            return ERROR;
        }

        /* Flag that the next chips will need new buffers */
        need_buffers = 1;
    }

    /* Close bands and images while correlator runs */
    if (ias_l1g_close_band(ref_l1g_band) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing reference image %s band", ref_img);
        ias_l1g_close_image(ref_id);
        ias_l1g_close_band(srch_l1g_band);
        ias_l1g_close_image(srch_id);
        return ERROR;
    }

    if (ias_l1g_close_image(ref_id) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing reference image %s", ref_img);
        ias_l1g_close_band(srch_l1g_band);
        ias_l1g_close_image(srch_id);
        return ERROR;
    }

    if (ias_l1g_close_band(srch_l1g_band) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing search image %s band", srch_img);
        ias_l1g_close_image(srch_id);
        return ERROR;
    }

    if (ias_l1g_close_image(srch_id) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing search image %s", srch_img);
        return ERROR;
    }

    /* Wait for the parallel correlator to finish */
    if (ias_math_close_parallel_correlator(correlator_ptr) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing the parallel correlator");
        free(result_ptr);
        return ERROR;    
    }

    /* Extract the correlation results returned by the parallel correlator */
    for (curpt = 0; curpt < npts; curpt++)
    {
        /* Don't update point state if it was outside the image */
        if (fp_corr_pts[curpt].active_flag != OUTSIDE_IMAGE)
        {
            /* Check the returned flag to see if this correlation succeeded */
            if (result_ptr[curpt].valid)
            {
                /* Correlation succeeded */
                fp_corr_pts[curpt].active_flag = ACTIVE;

                /* Since the images are co-registered, the offset can be 
                   directly applied to the nominal since predicted ref 1/s
                   = predicted srch l/s */
                fp_corr_pts[curpt].ref_os_pt.line = 
                    fp_corr_pts[curpt].nom_os_pt.line;
                fp_corr_pts[curpt].ref_os_pt.samp = 
                    fp_corr_pts[curpt].nom_os_pt.samp;

                fp_corr_pts[curpt].srch_os_pt.line = 
                    fp_corr_pts[curpt].nom_os_pt.line + 
                    (double)result_ptr[curpt].fit_offset[1];
                fp_corr_pts[curpt].srch_os_pt.samp = 
                    fp_corr_pts[curpt].nom_os_pt.samp +
                    (double)result_ptr[curpt].fit_offset[0];
                fp_corr_pts[curpt].correlation_accuracy = 
                    result_ptr[curpt].strength;
            }
            else
            {
                /* CORRELATION FAILURE */
                fp_corr_pts[curpt].active_flag = CORR_FAIL;
                fp_corr_pts[curpt].ref_os_pt.line  = 0.0;
                fp_corr_pts[curpt].ref_os_pt.samp  = 0.0;
                fp_corr_pts[curpt].srch_os_pt.line = 0.0;
                fp_corr_pts[curpt].srch_os_pt.samp = 0.0;
            }
        }
    }

    /* Free the memory for the results */
    free(result_ptr);

    return SUCCESS;
}

