/******************************************************************************
NAME: select_corr_pts

PURPOSE:
Select pts_per_grid nominal correlation points from the center line of each 
grid cell (output space).

RETURN VALUE:
Type = CORR_PT_INFO *
Value           Description
-----           -----------
pointer to array of CORR_PT_INFO structures
NULL            Error allocating memory for points

NOTES:
This function allocates memory for correlation points.  The caller is 
responsible for freeing this memory when done with it.

There will be pts_per_cell equally-spaced points per grid cell.  For example,
if there are 4 points per cell, they will be placed as such:
  +-----------+
  |     |     |
  |  *  |  *  |
  |     |     |
  |-----|-----|
  |     |     |
  |  *  |  *  |
  |     |     |
  +-----------+
The cell is divided into 4 sub-cells, and each sub-cell is divided in half to
place the point in the middle, yielding 8 sections, with points at 1/8, 3/8, 
5/8, and 7/8.

This function assigns nom_os_pt, sca_index, grid_column, and grid_row.
The rest of the point information is initialized to all zeros.
This includes fields that are enumerated types such as active flag, so the
active flag is initialized to Active.
******************************************************************************/
#include <stdlib.h>
#include <math.h>

#include "ias_logging.h"        /* IAS Logging library */
#include "focal_plane_setup.h"

CORR_PT_INFO *select_corr_pts
(
    const IAS_GRID_TYPE *grid,  /* I: Grid structure */
    int pts_per_cell,           /* I: Num points per cell to correlate at */
    const int band_number,      /* I: Band number (pan) */
    int *npts                   /* O: Total # of pts in image to correlate */
)
{
    CORR_PT_INFO *fp_corr_pts; /* Array of structures allocated & returned */
    static CORR_PT_INFO corr_pt_init; /* Static local structure to zero-
                           initialize all fields (use default initializer 
                           for static variables) */
    int cell_rloc;      /* Row location within cell counter */
    int cell_cloc;      /* Column location within cell counter */
    int cell_ind;       /* Cell index */
    int curpt;          /* Point counter */
    int currow;         /* Grid row counter */
    int curcell;        /* Cell within row counter */
    int ncols;          /* Number of columns in grid */
    int nrows;          /* Number of rows in grid */
    int nsca;           /* Number of SCAs in grid */
    int ncol_pts;       /* Number of columns of points in each cell */
    int nrow_pts;       /* Number of rows of points in each cell */
    int curr_sca_index; /* Current SCA */
    int band_index;     /* Index of band */
    float rfrac;        /* Fractional row grid cell location */
    float cfrac;        /* Fractional column grid cell location */
    double l1, l2, l3, l4; /* Grid corner line locations */
    double s1, s2, s3, s4; /* Grid corner sample locations */

    /* Look up the band index */
    band_index = ias_sat_attr_convert_band_number_to_index(band_number);
    if (band_index == ERROR)
    {
        IAS_LOG_ERROR("Invalid band index for band number %d", band_number);
        *npts = 0;
        return NULL;
    }

    /* Figure out how many points we will have & allocate memory for them */
    nsca = grid->scas_per_band[band_index];

    /* Set up tie point pattern for each grid cell */
    ncol_pts = (int)ceil(sqrt((double)pts_per_cell));
    nrow_pts = (int)ceil((double)pts_per_cell / (double)ncol_pts);
    pts_per_cell = ncol_pts * nrow_pts;
    *npts = (grid->gridbands[band_index][0].ngrid_lines - 1) * nrow_pts *
            (grid->gridbands[band_index][0].ngrid_samps - 1) * ncol_pts *
            nsca;
    fp_corr_pts = malloc(*npts * sizeof(CORR_PT_INFO));
    if (fp_corr_pts == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for information for %d correlation "
            "points", *npts);
        *npts = 0;
        return NULL;
    }

    IAS_LOG_INFO("%d by %d tie points in each grid cell", ncol_pts, nrow_pts);

    IAS_LOG_INFO("Setting up to correlate %d points (%d grid rows "
        "x %d grid cells per row x %d points per cell)", *npts,
        grid->gridbands[band_index][0].ngrid_lines - 1,
        grid->gridbands[band_index][0].ngrid_samps - 1,
        pts_per_cell * pts_per_cell);

    /* Calculate the nominal location for each point.  Use bilinear
       interpolation to compute the point locations. */
    curpt = 0;
    for (curr_sca_index = 0; curr_sca_index < nsca; curr_sca_index++)
    {
        ncols = grid->gridbands[band_index][curr_sca_index].ngrid_samps;
        nrows = grid->gridbands[band_index][curr_sca_index].ngrid_lines;
        for (currow = 0; currow < nrows - 1; currow++)
        {
            for (curcell = 0; curcell < ncols - 1; curcell++)
            {
               /*
                * The corners are l1, s1 ----------- l3, s3
                *                   |                   |
                *                   |                   |
                *                   |                   |
                *                 l2, s2 ----------- l4, s4
                */
                cell_ind = grid->gridbands[band_index][curr_sca_index].
                    zeroplane * ncols * nrows + currow * ncols + curcell;
                l1 = grid->gridbands[band_index][curr_sca_index].
                    out_lines[cell_ind];
                s1 = grid->gridbands[band_index][curr_sca_index].
                    out_samps[cell_ind];
                l2 = grid->gridbands[band_index][curr_sca_index].
                    out_lines[cell_ind+ncols];
                s2 = grid->gridbands[band_index][curr_sca_index].
                    out_samps[cell_ind+ncols];
                l3 = grid->gridbands[band_index][curr_sca_index].
                    out_lines[cell_ind+1];
                s3 = grid->gridbands[band_index][curr_sca_index].
                    out_samps[cell_ind+1];
                l4 = grid->gridbands[band_index][curr_sca_index].
                    out_lines[cell_ind+ncols+1];
                s4 = grid->gridbands[band_index][curr_sca_index].
                    out_samps[cell_ind+ncols+1];

                for (cell_rloc = 0; cell_rloc < nrow_pts; cell_rloc++)
                {
                    /* Compute the grid cell row fractional location */
                    rfrac = (double)(2*cell_rloc + 1) / (double)(2*nrow_pts);

                    for (cell_cloc = 0; cell_cloc < ncol_pts; 
                         cell_cloc++, curpt++)
                    {
                        /* First assign the local structure to it to zero out 
                           fields */
                        fp_corr_pts[curpt] = corr_pt_init;

                        /* Compute the grid cell column fractional location */
                        cfrac = (double)(2*cell_cloc + 1) / 
                                (double)(2*ncol_pts);

                        /* Compute the tie point coordinates using bilinear 
                           interpolation */
                        fp_corr_pts[curpt].nom_os_pt.line = 
                            l1 * (1.0-cfrac) * (1.0-rfrac) +
                            l2 * (1.0-cfrac) * rfrac       +
                            l3 * cfrac       * (1.0-rfrac) +
                            l4 * cfrac       * rfrac;

                        fp_corr_pts[curpt].nom_os_pt.samp = 
                            s1 * (1.0-cfrac) * (1.0-rfrac) +
                            s2 * (1.0-cfrac) * rfrac       +
                            s3 * cfrac       * (1.0-rfrac) +
                            s4 * cfrac       * rfrac;

                        fp_corr_pts[curpt].sca_index = curr_sca_index;
                        fp_corr_pts[curpt].grid_column = curcell;
                        fp_corr_pts[curpt].grid_row = currow;
                    }
                }
            }
        }
    }

    return fp_corr_pts;
}
