/******************************************************************************
NAME: set_up_grid

PURPOSE:
Read the grid, do some validity checks, and set up some variables.
The whole grid structure (pan band plus header info) is returned.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
ERROR           Error reading grid or not a precision grid
SUCCESS         Successfully read precision grid

NOTES:
The whole grid structure is returned so the caller can call ias_grid_free to
free all memory allocated when the grid was read.
******************************************************************************/
#include "ias_logging.h"                /* IAS Logging library */
#include "focal_plane_setup.h"

int set_up_grid
(
    const char *grid_name,              /* I: Precision grid filename */
    int band_number,                    /* I: Band number (pan) */
    ALIGNMENT_PROCESSING_TYPE processing_type, /* I: Algorithm processing 
                                              type */
    IAS_GRID_TYPE *grid                 /* O: Whole grid read from file */
)
{
    int bandcount = 1;                  /* Read only 1 band from grid */

    /* First read the grid header & reference band */
    if (ias_grid_read(grid_name, &band_number, bandcount, grid) != SUCCESS) 
    {
        IAS_LOG_ERROR("Reading grid file %s", grid_name);
        return ERROR;
    }

    /* Make sure it's a precision grid */
    if (grid->correction_type != IAS_PRECISION)
    {
        if (processing_type == OLI_FOCAL_PLANE_ALIGNMENT)
        {
            /* OLI Focal Plane Alignment must be registered to a reference DOQ
               image, so a non-precision grid implies an error-level problem */
            ias_grid_free(grid);
            IAS_LOG_ERROR("Grid %s is not a precision grid", grid_name);
            return ERROR;
        }
        else /* processing_type == TIRS_ALIGNMENT */
        {
            /* TIRS Alignment needs to be registered to a corresponding OLI
               image, which could be a systematic terrain corrected image */
            IAS_LOG_WARNING("Grid %s is not a precision grid", grid_name);
        }
    }

    return SUCCESS;
}
