/******************************************************************************
NAME:           calculate_post_fit_residuals

PURPOSE:
calculate_post_fit_residuals calculates the residual statistics after the 
Legendre polynomial coefficients have been fit.

RETURNS:
Type = int
Value           Description
-----           -----------
ERROR           Error calculating statistics
SUCCESS         Statistics successfully calculated

ALGORITHM REFERENCES:
OLI TIRS Alignment Calibration ADD
******************************************************************************/

#include <math.h>

#include "ias_logging.h"        /* IAS Logging library */
#include "calculate_post_fit_residuals.h" /* Definitions for this module */

int calculate_post_fit_residuals
(
    const double *original_legendre_coeffs_x, /* I: IAS_LOS_LEGENDRE_TERMS
                              original Legendre coefficients in the line
                              direction */
    const double *original_legendre_coeffs_y, /* I: IAS_LOS_LEGENDRE_TERMS
                              original Legendre coefficients in the sample
                              direction */
    int points,                 /* I: Number of correlation points */
    int sca_index,              /* I: SCA index for calculating residuals */
    int sca_num_det,            /* I: Number of detectors in each SCA */
    CORR_PT_INFO *corr_pts_ptr, /* I/O: Correlation points to calculate post-
                                      fit residuals on */
    POST_FIT_RESIDUAL_STATS_TYPE *residuals_ptr, /* O: Calculated residual 
                                      information */
    double db_save_threshold_x, /*  I: Threshold to limit saving to DB */
    double db_save_threshold_y, /*  I: Threshold to limit saving to DB */
    int *store_in_database_flag, /* I/O: Flag used to indicate whether data 
                                      should be stored in the database */
    void (*calculate_fit) (const double, const double, 
        const PARAMETERS_FIT_CALCULATION *,
        double *, double *), /* I: Function pointer */  
    PARAMETERS_FIT_CALCULATION *fit_cal, /* I/O: Struct for fit calculation */
    int app_flag             /* I: Identifies which application called this
                                   function.
                                   1: tirs_alignment_legendre
                                   2: estimate_tirs_ssm_angle */ 
)
{
    /* Statistical sums of the delta and residual information (X and Y) */
    double sum_delta_x = 0.0;
    double sum_delta_y = 0.0;
    double sum_delta_x_sqr = 0.0;
    double sum_delta_y_sqr = 0.0;
    double sum_residual_x = 0.0;
    double sum_residual_y = 0.0;
    double sum_residual_x_sqr = 0.0;
    double sum_residual_y_sqr = 0.0;

    int index;                  /* Loop counter */
    double temp;                /* Temporary variable */
    int stat_point_count = 0;   /* Count of points used for statistics */

    /* Loop through all the scan mirror correlation points */
    for (index = 0; index < points; index++)
    {
        CORR_PT_INFO *corr_ptr; /* Cached pointer to current point */
        double x_coordinate;    /* Current LOS x coordinate from original 
                                   Legendre coefficients */
        double y_coordinate;    /* Current LOS y coordinate from original 
                                   Legendre coefficients */
        double normalized_det;  /* Normalized detector location */
        double los_err_line;    /* LOS error for line from corr_ptr */
        double los_err_samp;    /* LOS error for sample from corr_ptr */
        double fit_diff_line;   /* Difference between fit and line LOS error */
        double fit_diff_samp;   /* Difference between fit and samp LOS error */
        double fit_x;           /* Fit calculation */
        double fit_y;           /* Fit calculation */

        corr_ptr = &corr_pts_ptr[index];

        /* If the current point is from the correct SCA and is an active
           point, collect statistics */
        if ((corr_ptr->sca_index == sca_index) && (corr_ptr->active_flag 
            == ACTIVE))
        {
            /* Get the sample statistics */
            los_err_line = corr_ptr->los_err.line;
            sum_delta_x += los_err_line;
            sum_delta_x_sqr += (los_err_line * los_err_line);

            /* Calculate the X (line) residuals */
            normalized_det = 2 * (corr_ptr->ref_is_pt.samp / (sca_num_det - 1))
                - 1;
            x_coordinate = ias_math_eval_legendre(normalized_det, 
                original_legendre_coeffs_x, IAS_LOS_LEGENDRE_TERMS);
            y_coordinate = ias_math_eval_legendre(normalized_det, 
                original_legendre_coeffs_y, IAS_LOS_LEGENDRE_TERMS);

            fit_cal->normalized_det = normalized_det;
             
            /* Calculate the modeled LOS angle correction */
            calculate_fit(x_coordinate, y_coordinate, fit_cal, &fit_x, &fit_y);

            /* Keep track of some values for later statistics calculations */
            fit_diff_line = fit_x - los_err_line;
            corr_ptr->fit_residual.line = fit_diff_line;
            sum_residual_x += fit_diff_line;
            sum_residual_x_sqr += (fit_diff_line * fit_diff_line);
            
            /* Get the Y (sample) statistics */
            los_err_samp = corr_ptr->los_err.samp;
            sum_delta_y += los_err_samp;
            sum_delta_y_sqr += (los_err_samp * los_err_samp);

            /* Keep track of some values for later statistics calculations */
            fit_diff_samp = fit_y - los_err_samp;
            corr_ptr->fit_residual.samp = fit_diff_samp;
            sum_residual_y += fit_diff_samp;
            sum_residual_y_sqr += (fit_diff_samp * fit_diff_samp);

            /* Count the points used to generate the statistics */
            stat_point_count++;
        }
    }

    /* Check if enough points have been gathered to calculate statistics on. 
       This condition will only be only a warning for SSM calibration
       (estimate_tirs_ssm_angle), Don't necessarily need measurements from 
       all three SCAs */
    if (stat_point_count < 2)
    {
        /* app_flag = 2, estimate_tirs_ssm_angle application */
        if (app_flag == 2)
        {
            /* Not enough points have been used to calculate statistics on, 
               generate a warning and return */
            IAS_LOG_WARNING("There are not enough points to calculate post-fit "
                            "residual statistics for SCA number %d",
                            sca_index + 1);
            residuals_ptr->points = stat_point_count;
            return SUCCESS;
        }
        else
        {
            /* Not enough points have been used to calculate statistics on, 
               generate an error and return */
            IAS_LOG_ERROR("There are not enough points to calculate post-fit "
                          "residual statistics");
            return ERROR;
        }

    }

    /* Save the statistics information */
    residuals_ptr->points = stat_point_count;

    /* Calculate the delta line statistics */
    residuals_ptr->delta_x.mean = sum_delta_x/stat_point_count;
    residuals_ptr->delta_x.rms = sqrt(sum_delta_x_sqr/stat_point_count);

    /* Find the standard deviation, setting it to zero if the square root part
       is negative */
    temp = sum_delta_x_sqr - (stat_point_count * residuals_ptr->delta_x.mean 
        * residuals_ptr->delta_x.mean);
        
    if (temp < 0.0)
        residuals_ptr->delta_x.stdev = 0.0;
    else
        residuals_ptr->delta_x.stdev = sqrt(temp / (stat_point_count - 1));

    /* Calculate the delta sample statistics */
    residuals_ptr->delta_y.mean = sum_delta_y / stat_point_count;
    residuals_ptr->delta_y.rms = sqrt(sum_delta_y_sqr / stat_point_count); 

    /* Find the standard deviation, setting it to zero if the square root part
       is negative */
    temp = sum_delta_y_sqr - (stat_point_count * residuals_ptr->delta_y.mean
        * residuals_ptr->delta_y.mean);
        
    if (temp < 0.0)
        residuals_ptr->delta_y.stdev = 0.0;
    else
        residuals_ptr->delta_y.stdev = sqrt(temp / (stat_point_count - 1));

    /* Calculate the residual line statistics */
    residuals_ptr->residual_x.mean = sum_residual_x / stat_point_count;
    residuals_ptr->residual_x.rms = sqrt(sum_residual_x_sqr / stat_point_count);

    /* Find the standard deviation, setting it to zero if the square root part
       is negative */
    temp = sum_residual_x_sqr - (stat_point_count
        * residuals_ptr->residual_x.mean * residuals_ptr->residual_x.mean);
    if (temp < 0.0)
        residuals_ptr->residual_x.stdev = 0.0;
    else
        residuals_ptr->residual_x.stdev = sqrt(temp / (stat_point_count - 1));

    /* Calculate the residual sample statistics */
    residuals_ptr->residual_y.mean = sum_residual_y / stat_point_count;
    residuals_ptr->residual_y.rms = sqrt(sum_residual_y_sqr / stat_point_count);

    /* Turn off saving to DB if root mean squared (RMS) values are too high */
    if (*store_in_database_flag)
    {
        if ((residuals_ptr->residual_x.rms > db_save_threshold_x)
            || (residuals_ptr->residual_y.rms > db_save_threshold_y))
        {
            *store_in_database_flag = 0;
            if (residuals_ptr->residual_x.rms > db_save_threshold_x)
            {
                IAS_LOG_INFO("SCA %02d x RMS fit %f exceeds tolerance %f, so "
                    "saving to database suppressed", sca_index + 1, 
                    residuals_ptr->residual_x.rms, db_save_threshold_x);
            }

            if (residuals_ptr->residual_y.rms > db_save_threshold_y)
            {
                IAS_LOG_INFO("SCA %02d y RMS fit %f exceeds tolerance %f, so "
                    "saving to database suppressed", sca_index + 1, 
                    residuals_ptr->residual_y.rms, db_save_threshold_y);
            }
        }
    }

    /* Find the standard deviation, setting it to zero if the square root part
       is negative */
    temp = sum_residual_y_sqr - (stat_point_count
        * residuals_ptr->residual_y.mean * residuals_ptr->residual_y.mean);
    if (temp < 0.0)
        residuals_ptr->residual_y.stdev = 0.0;
    else
        residuals_ptr->residual_y.stdev = sqrt(temp / (stat_point_count - 1));

    return SUCCESS;
}
