/******************************************************************************
NAME:           create_report_tirs_alignment_legendre

PURPOSE:
create_report_tirs_alignment_legendre generates a report of the results of 
the SCA Legendre polynomial fit calculations.  The report is written to a file.

RETURNS:
Type = int
Value           Description
-----           -----------
ERROR           Error generating the report
SUCCESS         Report successfully generated

ALGORITHM REFERENCES:
TIRS Alignment Calibration ADD
******************************************************************************/

#include <string.h>               /* String library */
#include <stdio.h>                /* fprintf prototype */

#include "ias_miscellaneous.h"    /* IAS Miscellaneous library */
#include "ias_logging.h"          /* IAS Logging library */
#include "create_report_tirs_alignment_legendre.h" /* Prototype for 
                                                      this routine */
#include "ias_structures.h"       /* IAS_VECTOR */
#include "ias_geo.h"              /* IAS geometric library */

#define SCA_MSG_SIZE 7            /* SCA message size; format SCA##n */
#define NUM_STATE_MSGS 5          /* Number of point state messages */

/* Local prototype */
static int write_coeffs
(
    FILE *file_ptr,               /* I: Pointer to file to write */
    const char *title_ptr,        /* I: Title for these coefficients */
    int sca_num,                  /* I: SCA number */
    const double *coeffs          /* I: Set of coefficients to write */
);

int create_report_tirs_alignment_legendre
(
    const char *report_file_name, /* I: Report file name */
    const char *terr_image_name,  /* I: Name of input 1T image */
    const char *ref_image_name,   /* I: Name of reference image */
    const char *l0r_image_name,   /* I: Name of L0R image */
    int nscas,                    /* I: Number of SCAs */
    double sca_coef_x[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS], /* I: Existing X 
                                        Legendre coefficients */
    double sca_coef_y[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS], /* I: Existing Y 
                                        Legendre coefficients */
    double error_legendre_x[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS], 
                                  /* I: Error X Legendre coefficients */
    double error_legendre_y[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS], 
                                  /* I: Error Y Legendre coefficients */
    double updated_legendre_x[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS], 
                                  /* I: Updated X coefficients */
    double updated_legendre_y[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS], 
                                  /* I: Updated Y coefficients */
    const CORR_PT_INFO *corr_pts_ptr, /* I: Correlation points */
    int points,                   /* I: Number of correlation points */
    const POST_FIT_RESIDUAL_STATS_TYPE *residuals_ptr, /* I: Calculated 
                                        residual information */
    double confidence_level,      /* I: Confidence level used for outlier 
                                        rejection */
    int fit_order,                /* I: Flag that indicates the polynomial 
                                        order that was used in the solution */
    const char *work_order_id,    /* I: Work Order ID for report header */
    double mean_x_fit,            /* I: Mean best-fit x offset */
    double mean_y_fit,            /* I: Mean best-fit y offset */
    struct main_db_save_info* main_db_save  /* I: top-level data for DB */
)
{
    FILE *report_file;            /* Pointer to the report file */
    int sca_index;                /* SCA index */
    double alignment_matrix[3][3]; /* Temporary alignment matrix */
    IAS_REPORT_HEADER *header = NULL; /* Report header structure */
    IAS_VECTOR state_vector;      /* x, y, z here are roll, pitch, and yaw */
    int status;                   /* Status returned by called routines */

    /* Open the report file for writing */
    report_file = fopen(report_file_name, "w");
    if (report_file == NULL)
    {
        IAS_LOG_ERROR("Opening the report file - %s", report_file_name);
        return ERROR;
    }

    /* Output the standard header */
    header = ias_misc_initialize_gps_report_header(terr_image_name,
            l0r_image_name, work_order_id, "TIRS_ALIGNMENT");
    if (!header)
    {
        IAS_LOG_ERROR("Initializing GPS header for report %s", 
            report_file_name);
        fclose(report_file);
        return ERROR;
    }

    status = ias_misc_write_gps_report_header(report_file, header);
    ias_misc_free_report_header(header);
    header = NULL;
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing GPS header to report file %s",
            report_file_name);
        fclose(report_file);
        return ERROR;
    }
    header = NULL;

    /* Write the name of the input 1T image to the report file */    
    status = fprintf(report_file, "Level 1Gt image: %s\n", terr_image_name);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    /* Write the name of the reference image to the report file */
    status = fprintf(report_file, "DOQ Reference Image: %s\n", ref_image_name);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    /* Write the original TIRS alignment angles */
    status = fprintf(report_file, "\nOriginal TIRS-to-OLI alignment angles:\n"
        "\tRoll:  %13.6E radians\n", main_db_save->original_tirs_oli_roll);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }
    status = fprintf(report_file, "\tPitch: %13.6E radians\n", 
        main_db_save->original_tirs_oli_pitch);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }
    status = fprintf(report_file, "\tYaw:   %13.6E radians\n", 
        main_db_save->original_tirs_oli_yaw);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    /* Write the TIRS alignment corrections */
    status = fprintf(report_file, "\nTIRS-to-OLI alignment corrections:\n"
        "\tRoll correction:  %13.6E radians\n", 
        main_db_save->est_tirs_oli_roll_corr);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }
    status = fprintf(report_file, "\tPitch correction: %13.6E radians\n", 
        main_db_save->est_tirs_oli_pitch_corr);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }
    status = fprintf(report_file, "\tYaw correction:   %13.6E radians\n", 
        main_db_save->est_tirs_oli_yaw_corr);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    /* Write the new TIRS alignment angles */
    status = fprintf(report_file,"\nNew TIRS-to-OLI alignment angles:\n\t"
        "Roll:  %13.6E radians\n", main_db_save->updated_tirs_oli_roll);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }
    status = fprintf(report_file,"\tPitch: %13.6E radians\n", 
        main_db_save->updated_tirs_oli_pitch);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }
    status = fprintf(report_file,"\tYaw:   %13.6E radians\n", 
        main_db_save->updated_tirs_oli_yaw);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    /* Calculate and write out the old TIRS2OLI alignment matrix */
    state_vector.x = main_db_save->original_tirs_oli_roll;
    state_vector.y = main_db_save->original_tirs_oli_pitch;
    state_vector.z = main_db_save->original_tirs_oli_yaw;
    ias_geo_convert_rpy_to_matrix(&state_vector, alignment_matrix);
    status = fprintf(report_file, "\nOld TIRS-to-OLI Alignment Matrix:\n\t"
        "%16.9E, %16.9E, %16.9E\n\t%16.9E, %16.9E, %16.9E\n\t%16.9E, %16.9E, "
        "%16.9E\n", alignment_matrix[0][0], alignment_matrix[0][1], 
        alignment_matrix[0][2], alignment_matrix[1][0], alignment_matrix[1][1],
        alignment_matrix[1][2], alignment_matrix[2][0], alignment_matrix[2][1],
        alignment_matrix[2][2]);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    /* Calculate and write out the new TIRS2OLI alignment matrix */
    state_vector.x = main_db_save->updated_tirs_oli_roll;
    state_vector.y = main_db_save->updated_tirs_oli_pitch;
    state_vector.z = main_db_save->updated_tirs_oli_yaw;
    ias_geo_convert_rpy_to_matrix(&state_vector, alignment_matrix);
    status = fprintf(report_file, "\nNew TIRS-to-OLI Alignment Matrix:\n\t"
        "%16.9E, %16.9E, %16.9E\n\t%16.9E, %16.9E, %16.9E\n\t%16.9E, %16.9E, "
        "%16.9E\n", alignment_matrix[0][0], alignment_matrix[0][1], 
        alignment_matrix[0][2], alignment_matrix[1][0], alignment_matrix[1][1],
        alignment_matrix[1][2], alignment_matrix[2][0], alignment_matrix[2][1],
        alignment_matrix[2][2]);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    /* Write the mean along- and across-track offsets to the report file */
    status = fprintf(report_file, "\nMean along-track offset:  %13.6E\nMean "
        "across-track offset: %13.6E\n\n", mean_x_fit, mean_y_fit); 
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    for (sca_index = 0; sca_index < nscas; sca_index++)
    {
        /* Write the existing coefficients to the report file */
        status = write_coeffs(report_file,
            "SCA %d Existing X Legendre coeffs:  ", sca_index + 1, 
            sca_coef_x[sca_index]);
        if (status < 0)
        {
            /* Error returned from fprintf */
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            fclose(report_file);
            return ERROR;
        }
     
        status = write_coeffs(report_file,
            "SCA %d Existing Y Legendre coeffs:  ", sca_index + 1, 
            sca_coef_y[sca_index]); 
        if (status < 0)
        {
            /* Error returned from fprintf */
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            fclose(report_file);
            return ERROR;
        }
     
        /* Write the Legendre error coefficients to the report file */
        status = write_coeffs(report_file,
            "SCA %d Error Legendre X coeffs:     ", sca_index + 1, 
            error_legendre_x[sca_index]); 
        if (status < 0)
        {
            /* Error returned from fprintf */
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            fclose(report_file);
            return ERROR;
        }
        
        status = write_coeffs(report_file,
            "SCA %d Error Legendre Y coeffs:     ", sca_index + 1, 
            error_legendre_y[sca_index]);
        if (status < 0)
        {
            /* Error returned from fprintf */
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            fclose(report_file);
            return ERROR;
        }
    
        /* Write the updated legendre coefficients to the report file */
        status = write_coeffs(report_file,
            "SCA %d Updated Legendre X coeffs:   ", sca_index + 1, 
            updated_legendre_x[sca_index]); 
        if (status < 0)
        {
            /* Error returned from fprintf */
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            fclose(report_file);
            return ERROR;
        }

        status = write_coeffs(report_file,
            "SCA %d Updated Legendre Y coeffs:   ", sca_index + 1, 
            updated_legendre_y[sca_index]);
        if (status < 0)
        {
            /* Error returned from fprintf */
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            fclose(report_file);
            return ERROR;
        }
    }
   
    /* Write the residuals statistics in the report file */ 
    if (write_stats_report_file(report_file, report_file_name, nscas,
        fit_order, points, confidence_level, residuals_ptr, corr_pts_ptr)
        != SUCCESS)
    {
        /* Error message already output by now */
        fclose(report_file);
        return ERROR;
    }

    /* Close report file */
    if (fclose(report_file) != 0)
    {
        /* Error returned from fclose */
        IAS_LOG_ERROR("Closing report file %s", report_file_name);
        return ERROR;
    }
    
    return SUCCESS;
  
}

/******************************************************************************
NAME:           write_stats_report_file

PURPOSE: Writes the residual statistics information to the report file

RETURNS:
Type = int
Value           Description
-----           -----------
ERROR           Error writing the residual statistics 
SUCCESS         Success writing the residual statistics

ALGORITHM REFERENCES:
TIRS Alignment Calibration ADD and TIRS SSM Calibration ADD

******************************************************************************/
/* routine to write out an entire set of coefficients */
int write_stats_report_file
(
    FILE *report_file,             /* I: Pointer to report file name */
    const char *report_file_name,  /* I: Report file name */
    int nscas,                     /* I: Number of SCAs */
    int fit_order,                 /* I: Flag that indicates the polynomial
                                         order that was used in the solution */
    int points,                    /* I: Number of correlation points */
    double confidence_level,       /* I: Confidence level used for outlier
                                         rejection */
    const POST_FIT_RESIDUAL_STATS_TYPE *residuals_ptr, /* I: Calulcated 
                                         residual information */
    const CORR_PT_INFO *corr_pts_ptr /* I: Correlation points */
)
{
    int sca_index;                /* SCA index */
    int index;                    /* Loop counter */
    int state_index;              /* Index into state message */
    const char *state_msg[NUM_STATE_MSGS] = {"Active", "CorrFail", "Outlier", 
                "Outside", "Unknown"}; /* Message for active flag */
    char sca_msg[SCA_MSG_SIZE];   /* Message for SCA index */
    int status;                   /* Status returned by called routines */
    
    /* Write the residual statistics information to the report file */
    status = fprintf(report_file,"\n"
            "                             Pre-fit            "
            "             Post-fit\n"
            "                         (microradians)         "
            "          (microradians)\n"
            "            Mean      RMSE     StdDev       Mean"
            "      RMSE     StdDev\n");
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        return ERROR;
    }

    for (sca_index = 0; sca_index < nscas; sca_index++)
    {
        status = fprintf(report_file,"SCA%02d X: % f % f % f    % f % f % f\n",
            sca_index + 1, residuals_ptr[sca_index].delta_x.mean * 1E6,
            residuals_ptr[sca_index].delta_x.rms * 1E6,
            residuals_ptr[sca_index].delta_x.stdev * 1E6,
            residuals_ptr[sca_index].residual_x.mean * 1E6,
            residuals_ptr[sca_index].residual_x.rms * 1E6,
            residuals_ptr[sca_index].residual_x.stdev * 1E6);
        if (status < 0)
        {
            /* Error returned from fprintf */
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            return ERROR;
        }

        status = fprintf(report_file,"SCA%02d Y: % f % f % f    % f % f % f\n",
            sca_index + 1, residuals_ptr[sca_index].delta_y.mean * 1E6,
            residuals_ptr[sca_index].delta_y.rms * 1E6,
            residuals_ptr[sca_index].delta_y.stdev * 1E6,
            residuals_ptr[sca_index].residual_y.mean * 1E6,
            residuals_ptr[sca_index].residual_y.rms * 1E6,
            residuals_ptr[sca_index].residual_y.stdev * 1E6);

        if (status < 0)
        {
            /* Error returned from fprintf */
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            return ERROR;
        }
    }
    
    /* If the fit order is valid (>= 0), include it in the report file */
    if (fit_order >= 0)
    {
        /* Write the confidence level and fit order to the report file */
        status = fprintf(report_file,"\nConfidence Level: %f\nFit Order: %d\n",
            confidence_level, fit_order);
    }
    else
    {
        /* Fit order not valid, so exclude it from the report.  Just write
           the confidence level used for the statistics to the report file */
        status = fprintf(report_file,"\nConfidence Level: %f\n",
            confidence_level);
    }
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        return ERROR;
    }

    /* Write the number of points used for each SCA */
    for (sca_index = 0; sca_index < nscas; sca_index++)
    {
        status = fprintf(report_file,"%d points used to fit SCA%02d\n",
            residuals_ptr[sca_index].points, sca_index + 1);
        if (status < 0)
        {
            /* Error returned from fprintf */
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            return ERROR;
        }
    }

    /* Write the correlation information for each point to the report file
       using fit_order flag. fit_order flag ranges from 0 to 3 for
       tirs_alignment_legendre applicationa and -1 for 
       estimate_tirs_ssm_angle application. */
    if (fit_order >= 0)
    {
        if (fprintf(report_file,
            "\nTIRS Alignment Calibration correlation points:\n") < 0)
        {
            /* Error returned from fprintf */
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            return ERROR;
        }
    }
    else
    {
        if (fprintf(report_file,
            "\nTIRS SSM Calibration correlation points:\n") < 0)
        {
            /* Error returned from fprintf */
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            return ERROR;
        }
    }

    /* Write the correlation information headers to the report file */
    if (fprintf(report_file,
        "                    Nominal  Nominal   Pixel    Pixel"
        "    Angle    Angle             Angle     Angle \n"
        "                    Output   Output    Error    Error"
        "    Error    Error            Fit Res   Fit Res\n"
        "SCA   Chip Col       Line     Samp     Line     Samp "
        "    Line     Samp    State     Line      Samp\n"
        "_____ ____ ___     ________ ________ ________ ________"
        " ________ ________ ________ _________ _________\n") < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        return ERROR;
    }

    for (index = 0; index < points; index++)
    {
        /* Get the current active flag and make sure it points to a legal 
           state_msg entry */
        state_index = corr_pts_ptr[index].active_flag;
        if ((state_index < ACTIVE_FLAG_START) || 
            (state_index > ACTIVE_FLAG_END))
        {
            /* Skip points outside the image or with unknown state */
            continue;
        }

        sca_index = corr_pts_ptr[index].sca_index;
        sprintf(sca_msg, "SCA%02d", sca_index + 1); 
        if ((sca_index < 0) || (sca_index > (nscas - 1)))
        {
            strcpy(sca_msg, "");  /* Point to empty direction */
        }

        status = fprintf(report_file, "%5s %2d %4d      %8.2f %8.2f %8.2f "
            "%8.2f %8.2f %8.2f %8s % 8.6f % 8.6f\n", sca_msg,
            corr_pts_ptr[index].sca_index,
            corr_pts_ptr[index].grid_column,
            corr_pts_ptr[index].nom_os_pt.line,
            corr_pts_ptr[index].nom_os_pt.samp,
            corr_pts_ptr[index].los_err_pix.line,
            corr_pts_ptr[index].los_err_pix.samp,
            corr_pts_ptr[index].los_err.line * 1E6,
            corr_pts_ptr[index].los_err.samp * 1E6,
            state_msg[state_index],
            corr_pts_ptr[index].fit_residual.line * 1E6,
            corr_pts_ptr[index].fit_residual.samp * 1E6);
        if (status < 0)
        {
            /* Error returned from fprintf */
            IAS_LOG_ERROR("Writing to report file %s", report_file_name);
            return ERROR;
        }
    }

    return SUCCESS;
}

/******************************************************************************
NAME:           write_coeffs

PURPOSE: Writes an entire set of coefficients to the indicated output file.

RETURNS:
Type = int
Value              Description
-----              -----------
   <0              Error writing coefficients
    0 or greater   Coefficients successfully written

ALGORITHM REFERENCES:
TIRS Alignment Calibration ADD
******************************************************************************/
static int write_coeffs
(
    FILE *file_ptr,               /* I: Pointer to file to write */
    const char *title_ptr,        /* I: Title for these coefficients */
    int sca_number,               /* I: SCA number */
    const double *coeffs          /* I: Set of coefficients to write */
)
{
    int status;                   /* Return status                   */

    /* Write the existing mirror coefficients to the report file */
    status = fprintf(file_ptr, title_ptr, sca_number);
    if (status < 0)
    {
        return status;
    }

    /* Write the coefficients */
    status = fprintf(file_ptr, "% 9.6E % 9.6E % 9.6E % 9.6E\n", coeffs[0], 
        coeffs[1], coeffs[2], coeffs[3]);

    return status;
}
