/******************************************************************************
NAME:           create_report_tirs_ssm_angle

PURPOSE:
create_report_tirs_ssm_angle generates a report of the results of the tirs
ssm calculations.  The report is written to a file.

RETURNS:
Type = int
Value           Description
-----           -----------
ERROR           Error generating the report
SUCCESS         Report successfully generated

ALGORITHM REFERENCES:
TIRS SSM Calibration ADD
******************************************************************************/

#include <string.h>               /* String library */
#include <stdio.h>                /* fprintf prototype */
#include <math.h>

#include "ias_miscellaneous.h"    /* IAS Miscellaneous library */
#include "ias_logging.h"          /* IAS Logging library */

#include "create_report_tirs_alignment_legendre.h"  /* Prototype for
                                                       write_stats_report_file
                                                       function */ 
#include "create_report_tirs_ssm_angle.h" /* Prototype for 
                                        create_report_tirs_ssm_angle routine */

int create_report_tirs_ssm_angle
(
    const char *report_file_name,  /* I: Report file name */
    const char *search_image_name, /* I: Name of input 1T image */
    const char *ref_image_name,    /* I: Name of reference image */
    const char *l0r_image_name,    /* I: Name of L0R image */
    int nscas,                     /* I: Number of SCAs */
    const double *ssm_epoch,      /* I: Time epoch for estimated angles */
    double ssm_angle_correction,   /* I: Estimated SSM angle correction */
    double net_ssm_angle,          /* I: Net new estimated angle */
    double pitch_angle_correction, /* I: Estimated pitch alignment
                                          angle correction */
    const CORR_PT_INFO *corr_pts_ptr, /* I: Correlation points */
    int points,                    /* I: Number of correlation points */
    const POST_FIT_RESIDUAL_STATS_TYPE *residuals_ptr, /* I: Calculated 
                                         residual information */
    double confidence_level,       /* I: Confidence level used for outlier 
                                         rejection */
    const char *work_order_id      /* I: Work Order ID for report header */
)
{
    FILE *report_file;             /* Pointer to the report file */
    int ssm_angle_counts;          /* Angle correction converted
                                     to counts */
    IAS_REPORT_HEADER *header = NULL; /* Report header structure */
    int status;                    /* Status returned by called routines */
    int fit_order = -1;            /* Polynomial order is not needed for the
                                      estimate_tirs_ssm_angle application
                                      so it is set to -1 which indicates that
                                      it is not in the report file */

    /* Open the report file for writing */
    report_file = fopen(report_file_name, "w");
    if (report_file == NULL)
    {
        IAS_LOG_ERROR("Opening the report file - %s", report_file_name);
        return ERROR;
    }

    /* Output the standard header */
    header = ias_misc_initialize_gps_report_header(search_image_name,
            l0r_image_name, work_order_id, "SSM ANGLE ESTIMATION");
    if (!header)
    {
        IAS_LOG_ERROR("Initializing GPS header for report %s", 
            report_file_name);
        fclose(report_file);
        return ERROR;
    }

    status = ias_misc_write_gps_report_header(report_file, header);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing GPS header to report file %s",
            report_file_name);
        fclose(report_file);
        ias_misc_free_report_header(header);
        return ERROR;
    }
    ias_misc_free_report_header(header);
    header = NULL;

    /* Write the name of the input 1T image to the report file */    
    status = fprintf(report_file, "TIRS Search Image: %s\n", search_image_name);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    /* Write the name of the reference image to the report file */
    status = fprintf(report_file, "OLI Reference Image: %s\n", ref_image_name);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    /* Write the epoch of the estimated angles */
    status = fprintf(report_file, "\nTIRS SSM angle estimation epoch:\n"
        "\tYear,DOY,SOD:  %6.1lf,%05.1lf,%09.3lf\n",
        ssm_epoch[0], ssm_epoch[1], ssm_epoch[2]);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }
    
    /* Write the TIRS SSM angle correction */
    status = fprintf(report_file, "\nTIRS SSM rotation angle correction:\n"
        "\tSSM correction:  %13.6E radians\n", ssm_angle_correction);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }
   
    /* Calculate ssm angle count, SSM counter is 24 bit for a full
       rotation.  This equation is angle * (1<<24) / (2 * pi). */
    ssm_angle_counts = (int) round(ssm_angle_correction * pow(2.0,21.0) / 
        atan(1.0));
    status = fprintf(report_file, "\nTIRS SSM rotation angle correction:\n"
        "\tSSM correction:  %d counts\n", ssm_angle_counts);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    /* Write the new net TIRS SSM angle */
    status = fprintf(report_file, "\nNet TIRS SSM rotation angle (from "
        "nadir):\n\tSSM angle:  %13.6E radians\n", net_ssm_angle);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    /* Calculate ssm angle count, SSM counter is 24 bit for a full
       rotation.  This equation is angle * (1<<24) / (2 * pi). */
    ssm_angle_counts = (int) round(net_ssm_angle * pow(2.0,21.0) / 
        atan(1.0));
    status = fprintf(report_file, "\nNet TIRS SSM rotation angle (from "
        "nadir):\n\tSSM angle:  %d counts\n", ssm_angle_counts);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    /* Write the TIRS pitch alignment angle correction */
    status = fprintf(report_file, "\nTIRS pitch alignment angle correction:\n"
        "\tPitch correction:  %13.6E radians\n", pitch_angle_correction);
    if (status < 0)
    {
        /* Error returned from fprintf */
        IAS_LOG_ERROR("Writing to report file %s", report_file_name);
        fclose(report_file);
        return ERROR;
    }

    /* Write the residuals statistics in the report file */
    if (write_stats_report_file(report_file, report_file_name, nscas,
        fit_order, points, confidence_level, residuals_ptr, corr_pts_ptr)
        != SUCCESS)
    {
        /* Error message already output by now */
        fclose(report_file);
        return ERROR;
    }

    /* Close report file */
    if (fclose(report_file) != 0)
    {
        /* Error returned from fclose */
        IAS_LOG_ERROR("Closing report file %s", report_file_name);
        return ERROR;
    }

    return SUCCESS;
}
