/******************************************************************************
NAME:           estimate_tirs_ssm_angle (main)

PURPOSE:        
Main driver for the mirror rotation angle estimation portion of TIRS SSM angle
correction. The estimate SSM angle portion reads the results of the setup step, 
filters the outliers, fits the data to a SSM rotation angle, updates the 
SSM trending data, and generates output reports.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
EXIT_FAILURE    Error generating angle; no output data produced 
EXIT_SUCCESS    Successful; SSM correction angle generated
******************************************************************************/
#include <stdlib.h>             /* Standard C library */
#include <string.h>             /* String library */

#include "ias_logging.h"        /* IAS logging library */
#include "ias_model_io.h"       /* Model I/O library */

#include "focal_plane.h"        /* Main algorithm structures */
#include "read_parameters.h"    /* Read parameters */ 
#include "calculate_post_fit_residuals.h" /* calculate_post_fit_residuals 
                                            prototype */
#include "create_report_tirs_ssm_angle.h" /* create_report prototype */
#include "filter_outliers.h"    /* filter_outliers prototype */
#include "calculate_point_weights.h" /* calculate_point_weights prototype */
#include "fit_ssm_angle.h"      /* fit_ssm_angle prototype */
#include "save_estimated_ssm_to_database.h"  /* DB save structure */


/******************************************************************************
NAME:           tirs_ssm_angle_fit_cal

PURPOSE:
tirs_ssm_angle_fit_cal calculates the fit x and fit y for
estimate_tirs_ssm_angle application.

RETURNS: nothing

ALGORITHM REFERENCES:
TIRS SSM Calibration ADD
******************************************************************************/
void tirs_ssm_angle_fit_cal
(
    const double x_coordinate, /* I: Current LOS x coordinate */
    const double y_coordinate, /* I: Current LOS y coordinate */
    const PARAMETERS_FIT_CALCULATION *fit_cal, /* I: Pointer to fit calculation
                                                     parameters */
    double *fit_x,             /* O: Fit calculation x */
    double *fit_y              /* O: Fit calculation y */
)             
{
    *fit_x = -y_coordinate * fit_cal->ssm_angle_correction 
        - fit_cal->pitch_angle_correction;

    *fit_y = (x_coordinate - 1.0) * fit_cal->ssm_angle_correction;
}

/* main function */
int main(int argc, char *argv[])
{
    ALIGNMENT_PARMS parms;      /* Parameters read from parameter file */
    CORR_PT_INFO *corr_pts;     /* Information about each correlation point */
    IAS_CPF *cpf;               /* CPF parameters */
    IAS_LOS_MODEL *model;       /* Model structure */
    int ncells;                 /* Number of cells across a SCA */
    int sca_index;              /* SCA counter/index */
    int nscas;                  /* Number of SCAs in this band (input space) */
    int sca_num_det;            /* Number of detectors per SCA */
    int npts;                   /* Number of correlation points to work with */
    int thermal1_band_number;   /* Number of reference band (TIRS_THERMAL1) */
    int thermal1_band_index;    /* Index of reference band (TIRS_THERMAL1) */
    double sca_coef_x[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS]; /* Along-track 
                                   SCA LOS Legendre coefficients */
    double sca_coef_y[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS]; /* Across-track 
                                   SCA LOS Legendre coefficients */
    POST_FIT_RESIDUAL_STATS_TYPE residual_stats[NUM_TIRS_SCAS];
    struct main_db_trend_info main_db_trend_info;/* Main DB save information */ 
    struct sca_db_trend_info sca_db_trend_info; /* SCA level DB save info */
    double db_save_threshold_x; /* Threshold to limit database saving */
    double db_save_threshold_y; /* Threshold to limit database saving */
    int db_save_threshold_flag = 1; /* Flag to indicate results should be
                                   written to the database */
    IAS_SENSOR_SCENE_SELECT_MIRROR_MODEL *ssm;  /* Pointer to SSM model */
    double ssm_angle_correction; /* TIRS SSM mirror rotation angle correction */
    double net_ssm_angle;        /* Estimated total TIRS SSM mirror rotation 
                                    angle */
    double pitch_angle_correction; /* Pitch alignment angle correction */
    double ssm_epoch[3];	/* Time epoch for this SSM angle estimate */
    int ssm_index;		/* SSM model record index */
    PARAMETERS_FIT_CALCULATION fit; /* Structure for fit calculation */
    int app_flag = 2;   /* Applicaton flag = 2 identifies the 
                               estimate_tirs_ssm_angle application */

    /* Initialize the DB save structures */
    memset(&main_db_trend_info, 0, sizeof(main_db_trend_info));
    memset(&sca_db_trend_info, 0, sizeof(sca_db_trend_info));

    /* Initialize the logging library */
    if (ias_log_initialize("estimate_tirs_ssm_angle") != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing logging library");
        return EXIT_FAILURE;
    }

    /* Read parameters and inputs */
    if (read_parameters(argc, argv, &parms, &corr_pts, &cpf, &model,
        &thermal1_band_number, &thermal1_band_index, &sca_num_det, &npts,
        &nscas, &ncells, &db_save_threshold_x, &db_save_threshold_y,
        sca_coef_x, sca_coef_y) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading input parameters");
        return EXIT_FAILURE;
    }

    /* Free CPF since it isn't needed for this application (since the parameter
       reading routine is shared) */ 
    ias_cpf_free(cpf);
    cpf = NULL;

    /* Set pointer to SSM model */
    ssm = model->sensor.sensors[IAS_TIRS].ssm_model;

    /* Fit correlation results to a set of Legendre error polynomials */
    if (fit_ssm_angle(npts, nscas, sca_num_det, corr_pts,
        sca_coef_x, sca_coef_y, &ssm_angle_correction,
        &pitch_angle_correction) != SUCCESS)
    {
        IAS_LOG_ERROR("Performing a least squares fit on the correlation "
                      "points");
        free(corr_pts);
        ias_los_model_free(model);
        model = NULL;
        return EXIT_FAILURE;
    }

    /* Extract the input SSM angle */
    ssm_epoch[0] = ssm->utc_epoch_time[0];
    ssm_epoch[1] = ssm->utc_epoch_time[1];
    ssm_epoch[2] = ssm->utc_epoch_time[2];
    ssm_index = 0;
    while ( ssm_index<ssm->ssm_record_count &&
            ssm->records[ssm_index].seconds_from_epoch < 0.0 ) ssm_index++;
    net_ssm_angle = ssm_angle_correction + ssm->records[ssm_index].mirror_angle;
    ssm_epoch[2] += ssm->records[ssm_index].seconds_from_epoch;

    fit.ssm_angle_correction = ssm_angle_correction;
    fit.pitch_angle_correction = pitch_angle_correction;

    /* Fit each SCA */
    for (sca_index = 0; sca_index < nscas; sca_index++)
    {
        /* Get the number of detectors for this SCA for the THERMAL1 band */
        sca_num_det = model->sensor.bands[thermal1_band_index].
            scas[sca_index].detectors;

        /* Calculate the post fit residual statistics */
        if (calculate_post_fit_residuals(sca_coef_x[sca_index],
            sca_coef_y[sca_index], npts, sca_index, 
            sca_num_det, corr_pts, &residual_stats[sca_index], 
            db_save_threshold_x, db_save_threshold_y, &db_save_threshold_flag,
            tirs_ssm_angle_fit_cal, &fit, app_flag) != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating post-fit residual statistics for "
                "SCA %d", sca_index + 1);
            free(corr_pts);
            ias_los_model_free(model);
            model = NULL;
            return EXIT_FAILURE;
        }
    }

    /* Free the model */
    ias_los_model_free(model);
    model = NULL;

    /* Output the TIRS SSM angle estimation report */
    if (create_report_tirs_ssm_angle(parms.report_file_name, 
        parms.terr_img_name, parms.ref_img_name, parms.l0r_img_name, 
        nscas, ssm_epoch, ssm_angle_correction, net_ssm_angle,
        pitch_angle_correction, corr_pts, npts, residual_stats,
        parms.outlier_tolerance, parms.work_order_id) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the TIRS SSM angle estimation report");
        free(corr_pts);
        return EXIT_FAILURE;
    }

    /* Free correlation points */
    free(corr_pts);
    corr_pts = NULL;

    /* If valid encoder data is associated with the telemetry then this scene
       should not be trended */
    if (ssm->position_quality == IAS_SSM_POSITION_ACTUAL)
    {
        db_save_threshold_flag = 0;
        IAS_LOG_INFO("Scene processed using actual SSM telemetry, so saving "
            "to database suppressed");
    }

    /* If storing results in the database is requested, and not rejected by 
       the DB save thresholds */
    if ((db_save_threshold_flag) && (parms.store_in_database_flag))
    {
        IAS_LOG_INFO("Writing results to database");

        /* Populate database structures */
        prepare_estimated_ssm_for_database_trend(
            ssm_epoch, ssm_angle_correction, net_ssm_angle,
            pitch_angle_correction, residual_stats, parms.outlier_tolerance,
            parms.work_order_id, parms.ref_img_name, nscas,
            &main_db_trend_info, &sca_db_trend_info);

        /* Send data in database structures to database */
        if (save_estimated_ssm_to_database(&main_db_trend_info,
            &sca_db_trend_info, nscas) != SUCCESS)
        {
            IAS_LOG_ERROR("Storing results in database");
            return EXIT_FAILURE;
        }
    }

    IAS_LOG_INFO("Successful completion");

    return EXIT_SUCCESS;
}
