/******************************************************************************
NAME:           fit_polynomials

PURPOSE:
fit_polynomials performs a constrained least squares fit of the correlation 
data points to find the Legendre error polynomial coefficients and the 
corrections to the TIRS alignment angles.  

RETURNS:
Type = int
Value           Description
-----           -----------
ERROR           Error fitting the Legendre error polynomial and alignment 
                angles
SUCCESS         Successfully fitted the Legendre error polynomial and 
                alignment angles

NOTES:
- This routine calculates the weighted least squares fit of the correlated data 
  points (using the angular error) to a Legendre error polynomial.  
- An optimization is used to keep the size of the matrices to a managable 
  level.  Since the weight matrix only contains data on the diagonal the TWA
  matrix can be accumulated in a [IAS_LOS_LEGENDRE_TERMS x 
  IAS_LOS_LEGENDRE_TERMS] matrix and the TWY matrix can be accumulated in a 
  [IAS_LOS_LEGENDRE_TERMS x 1] matrix.
- See the ADD for detailed definitions of matrices like C, W, and A.  Multi-
  letter matrices indicate the formula to calculate the matrix (T => transpose)

ALGORITHM REFERENCES:
- TIRS Alignment Calibration ADD
- Numerical Recipes in C
******************************************************************************/

#include "ias_logging.h"     /* IAS logging library */
#include "ias_math.h"        /* IAS math library */
#include "fit_polynomials.h" /* Prototype for this routine */

#define NUMBER_UNKNOWNS (3 + 2 * NUM_TIRS_SCAS * IAS_LOS_LEGENDRE_TERMS)
                             /* Number of unknowns to solve for; 3 is for the 
                                alignment angles; 2 accounts for along and 
                                across track Legendre coefficients */

int fit_polynomials
(
    int corr_points,         /* I: Number of correlation points */
    int number_scas,         /* I: Number of SCAs to solve */
    int fit_order,           /* I: Order of polynomial to fit (0-3) */
    int sca_num_det,         /* I: Number of detectors in the SCA */
    const CORR_PT_INFO *corr_pts_ptr,/* I: Correlation points */
    int constraint,          /* I: Constraint option: 0 = nominal; 1 = no 
                                   alignment */
    int constraint_weight,   /* I: Constraint weight from the CPF */
    double current_coeffs_x[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS], /* I: 
                                   Current along-track Legendre coefficients */
    double current_coeffs_y[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS], /* I: 
                                   Current across-track Legendre coefficients */
    double error_coeffs_x[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS], /* O: 
                                   Along-track Legendre error coefficients */
    double error_coeffs_y[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS], /* O:
                                   Across-track Legendre error coefficients */
    double *alignment_corrections /* O: TIRS-to-OLI alignment roll-pitch-yaw */
)
{
    int index;               /* Loop counter - correlation point index */
    int term;                /* Loop counter - Legendre coefficient terms */
    int unknown;             /* Loop counter - unknown to solve for */
    int sca_index;           /* SCA index */
    int row;                 /* Loop counter */
    int column;              /* Loop counter */
    double TWA[NUMBER_UNKNOWNS * NUMBER_UNKNOWNS] = {0}; /* Accumulated matrix 
                                for the data points */ 
    double A[2][NUMBER_UNKNOWNS]; /* Matrix of Legendre terms for the current
                                point */
    double TW[NUMBER_UNKNOWNS][2]; /* Matrix of A transposed and multiplied by
                                the weight of current point */
    double C[3][NUMBER_UNKNOWNS]; /* Constraint matrix */
    double CW[NUMBER_UNKNOWNS][3]; /* Matrix of C transposed and multiplied by
                                the constraint weight */
    double TWY[NUMBER_UNKNOWNS] = {0}; /* Matrix of error terms */
    double L[IAS_LOS_LEGENDRE_TERMS] = {0}; /* Legendre weights */
    double weight;           /* Temporary weight holder */
    double x_point;          /* Current x point LOS from Legendre 
                                coefficients */
    double y_point;          /* Current y point LOS from Legendre 
                                coefficients */
    double v[NUMBER_UNKNOWNS]; /* Work vector for QR/QRsolve */
    int points_used = 0;     /* Count of points factored into the solution */

    /* Make sure the number of points is not less than or equal to zero */
    if (corr_points <= 0)
    {
        IAS_LOG_ERROR("Number of correlation points %d is <= 0", corr_points);
        return ERROR;
    }

    /* Check for later divide by 0 */
    if (sca_num_det == 1)
    {
        IAS_LOG_ERROR("Number of detectors is 1, leading to divide by 0");
    }

    /* For each of the active points in the current SCA */
    for (index = 0; index < corr_points; index++)
    {
        if (corr_pts_ptr[index].active_flag == ACTIVE)
        {
            /* Update the count of the points used */
            points_used++;

            /* Find the SCA for the point */
            sca_index = corr_pts_ptr[index].sca_index;

            /* Initialize A matrix for this point */
            for (unknown = 0; unknown < NUMBER_UNKNOWNS; unknown++)
            {
                A[0][unknown] = 0.0;
                A[1][unknown] = 0.0;
            }

            /* The next several sections form the Legendre terms and load the A
               matrix for this point, 1 Legendre coefficient at a time.  To
               understand indexing for A, see the ADD for the layout of A */

            /* Set the A(1,j) matrix cells for the 1st Legendre coefficient */
            L[0] = 1.0;
            A[0][3 + sca_index * 2 * IAS_LOS_LEGENDRE_TERMS] = L[0];
            A[1][3 + sca_index * 2 * IAS_LOS_LEGENDRE_TERMS 
                + IAS_LOS_LEGENDRE_TERMS] = L[0];

            /* Account for the 1st Legendre coefficient's contribution to the
               current point's x and y coordinates */
            x_point = current_coeffs_x[sca_index][0] * L[0];
            y_point = current_coeffs_y[sca_index][0] * L[0];

            /* Calculate normalized detector coordinate */
            L[1] = 2 * (corr_pts_ptr[index].srch_is_pt.samp / (sca_num_det 
                - 1)) - 1;

            /* Set the A(1,j) matrix cells for the 2nd Legendre coefficient */
            A[0][3 + sca_index * 2 * IAS_LOS_LEGENDRE_TERMS + 1] = L[1];
            A[1][3 + sca_index * 2 * IAS_LOS_LEGENDRE_TERMS 
                + IAS_LOS_LEGENDRE_TERMS + 1] = L[1];

            /* Account for the 2nd Legendre coefficient's contribution to the
               current point's x and y coordinates */
            x_point += current_coeffs_x[sca_index][1] * L[1];
            y_point += current_coeffs_y[sca_index][1] * L[1];

            /* Account for the remaining Legendre terms */
            for (term = 2; term < IAS_LOS_LEGENDRE_TERMS; term++)
            {
               L[term] = ((2 * term - 1) * L[1] * L[term - 1] - (term - 1) 
                   * L[term - 2])/(double)term;

               /* Set the A(1,j) matrix cells for the remaining Legendre 
                  coefficients */
               A[0][3 + sca_index * 2 * IAS_LOS_LEGENDRE_TERMS + term] 
                   = L[term];
               A[1][3 + sca_index * 2 * IAS_LOS_LEGENDRE_TERMS 
                   + IAS_LOS_LEGENDRE_TERMS + term] = L[term];

               /* Account for the remaining Legendre coefficients' contribution
                  to the current point's x and y coordinates */
               y_point += current_coeffs_y[sca_index][term] * L[term];
               x_point += current_coeffs_x[sca_index][term] * L[term];
            }

            if (constraint == 0)
            {
                /* Initialize non-0 parts of the 2x3 [A(0)] submatrix of A */
                A[0][1] = -1.0;
                A[1][0] = 1.0;
                A[0][2] = y_point;
                A[1][2] = -x_point;
            }

            /* Form the TW matrix by multiplying the transpose of A by the
               point's weight */
            weight = corr_pts_ptr[index].pt_weight;
            for (unknown = 0; unknown < NUMBER_UNKNOWNS; unknown++)
            {
               TW[unknown][0] = A[0][unknown] * weight;
               TW[unknown][1] = A[1][unknown] * weight;
            }

            /* Accumulate just the upper half of the TWA matrix since it
               is symmetrical around the diagonal */
            for (row = 0; row < NUMBER_UNKNOWNS; row++)
            {
                for (column = row; column < NUMBER_UNKNOWNS; column++)
                {
                    TWA[column + row * NUMBER_UNKNOWNS] += TW[row][0] * 
                        A[0][column];
                    TWA[column + row * NUMBER_UNKNOWNS] += TW[row][1] * 
                        A[1][column];
                }
            }

            /* Accumulate the TWY matrix in the line and sample directions */
            for (row = 0; row < NUMBER_UNKNOWNS; row++)
            {
                TWY[row] += TW[row][0] * corr_pts_ptr[index].los_err.line;
                TWY[row] += TW[row][1] * corr_pts_ptr[index].los_err.samp;
            }
        } /* End of active data point */
    } /* End of looping through all the correlation data points */

    /* Add solution constraints */

    /* Initialize C matrix */
    for (unknown = 0; unknown < NUMBER_UNKNOWNS; unknown++)
    {
        C[0][unknown] = 0.0;
        C[1][unknown] = 0.0;
        C[2][unknown] = 0.0;
    }

    /* Set constraints based on the option selected */
    if (constraint == 0) /* Use nominal constraint for full solution */
    {
        /* These constraints suppress global roll-pitch-yaw effects in the 
           Legendre coefficients.  See matrix (2-7) in the ADD for details */
        for (sca_index = 0; sca_index < NUM_TIRS_SCAS; sca_index++)
        {
            /* Roll constraint */
            C[0][3 + sca_index * 2 * IAS_LOS_LEGENDRE_TERMS 
                + IAS_LOS_LEGENDRE_TERMS] = 1.0;
            C[0][3 + sca_index * 2 * IAS_LOS_LEGENDRE_TERMS 
                + IAS_LOS_LEGENDRE_TERMS + 2] = -0.5;

            /* Pitch constraint */
            C[1][3 + sca_index * 2 * IAS_LOS_LEGENDRE_TERMS] = 1.0;
            C[1][3 + sca_index * 2 * IAS_LOS_LEGENDRE_TERMS + 2] = -0.5;
        }

        /* Yaw constraint */
        C[2][3 + 0 * 2 * IAS_LOS_LEGENDRE_TERMS] = 1.0;       /* SCA 1 */
        C[2][3 + 0 * 2 * IAS_LOS_LEGENDRE_TERMS + 2] = -0.5;  /* SCA 1 */
        C[2][3 + 2 * 2 * IAS_LOS_LEGENDRE_TERMS] = -1.0;      /* SCA 3 */
        C[2][3 + 2 * 2 * IAS_LOS_LEGENDRE_TERMS + 2] = 0.5;   /* SCA 3 */
    }
    else
    {
        /* These constraints suppress the roll-pitch-yaw alignment solution.
           See matrix (2-8) in the ADD for details */
        C[0][0] = 1.0;
        C[1][1] = 1.0;
        C[2][2] = 1.0;
    }

    /* Form the CW matrix by multiplying the transpose of C by the constraint
       weight */
    for (unknown = 0; unknown < NUMBER_UNKNOWNS; unknown++)
    {
        CW[unknown][0] = C[0][unknown] * constraint_weight;
        CW[unknown][1] = C[1][unknown] * constraint_weight;
        CW[unknown][2] = C[2][unknown] * constraint_weight;
    }

    /* Accumulate just the upper half of the TWA matrix since it is
       symmetrical around the diagonal */
    for (row = 0; row < NUMBER_UNKNOWNS; row++)
    {
        for (column = row; column < NUMBER_UNKNOWNS; column++)
        {
            TWA[column + row * NUMBER_UNKNOWNS] += CW[row][0] * C[0][column];
            TWA[column + row * NUMBER_UNKNOWNS] += CW[row][1] * C[1][column];
            TWA[column + row * NUMBER_UNKNOWNS] += CW[row][2] * C[2][column];
        }
    }

    /* Verify enough points were found to fit (minimum is the degrees of
       freedom - NUMBER_UNKNOWNS) */
    if (points_used < NUMBER_UNKNOWNS)
    {
        /* Not enough points used, generate an error message and error out */
        IAS_LOG_ERROR("Only %d correlation points found to be usable "
                "when a minimum of %d are needed", points_used, 
                NUMBER_UNKNOWNS);
        return ERROR;
    }

    /* Copy the upper half of the TWA matrix to the lower half in preparation 
       for the QR call */
    for (row = 0; row < (NUMBER_UNKNOWNS - 1); row++)
    {
        for (column = row + 1; column < NUMBER_UNKNOWNS; column++)
        {
            TWA[row + column * NUMBER_UNKNOWNS] = 
                TWA[column + row * NUMBER_UNKNOWNS];
        }
    }

    /* Perform the QR decomposition on the TWA matrix */
    if (ias_math_matrix_QRfactorization(TWA, NUMBER_UNKNOWNS /* rows */, 
        NUMBER_UNKNOWNS /* columns */, v, 1 /* row major format */) != SUCCESS)
    {
        /* Error solving for the least squares fit */
        IAS_LOG_ERROR("Calculating least squares fit"); 
        return ERROR;
    }

    /* Solve for the TWY coefficients */
    if (ias_math_matrix_QRsolve(TWA, NUMBER_UNKNOWNS /* rows */, 
        NUMBER_UNKNOWNS /* columns */, v, TWY, 1 /* v is already 
        constructed */) != SUCCESS)  
    {
        /* Error solving for the least squares fit */
        IAS_LOG_ERROR("Calculating least squares fit");
        return ERROR;
    }

    /* Copy the X and Y error coefficients from the solutions.  Note: It's
       fine to compute full order and then zero out the higher order terms
       here because they're Legendre coefficients */
    for (sca_index = 0; sca_index < NUM_TIRS_SCAS; sca_index++)
    {
        for (index = 0; index < IAS_LOS_LEGENDRE_TERMS; index++)
        {
            if (index <= fit_order)
            {
                error_coeffs_x[sca_index][index] = TWY[3 + sca_index * 2 
                    * IAS_LOS_LEGENDRE_TERMS + index];
                error_coeffs_y[sca_index][index] = TWY[3 + sca_index * 2 
                    * IAS_LOS_LEGENDRE_TERMS + IAS_LOS_LEGENDRE_TERMS 
                    + index];
            }
            else
            {
                error_coeffs_x[sca_index][index] = 0.0;
                error_coeffs_y[sca_index][index] = 0.0;
            }
        }
    }

    /* Copy the roll-pitch-yaw alignment corrections to the output */
    alignment_corrections[0] = TWY[0];
    alignment_corrections[1] = TWY[1];
    alignment_corrections[2] = TWY[2];

    return SUCCESS;
}
