/******************************************************************************
NAME:           fit_ssm_angle

PURPOSE:
fit_ssm_angle performs a least squares fit of the correlation 
data points to find the best fit adjustment to the SSM rotation angle. 

RETURNS:
Type = int
Value           Description
-----           -----------
ERROR           Error fitting the SSM angle
SUCCESS         Successfully fitted the SSM angle

NOTES:
- This routine calculates the weighted least squares fit of the correlated data 
  points (using the angular error) to a scene select mirror rotation angle.
- See the ADD for detailed definitions of matrices like C, W, and A.  Multi-
  letter matrices indicate the formula to calculate the matrix (T => transpose)

ALGORITHM REFERENCES:
- TIRS SSM Angle Estimation ADD
- Numerical Recipes in C
******************************************************************************/

#include <math.h>            /* fabs prototype */
#include "ias_logging.h"     /* IAS logging library */
#include "ias_math.h"        /* IAS math library */
#include "fit_ssm_angle.h"   /* Prototype for this routine */

#define	MIN_NUMBER_POINTS	10

int fit_ssm_angle
(
    int corr_points,        /* I: Number of correlation points */
    int number_scas,        /* I: Number of SCAs to solve */
    int sca_num_det,        /* I: Number of detectors in the SCA */
    const CORR_PT_INFO *corr_pts_ptr, /* I: Correlation points */
    double current_coeffs_x[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS], /* I: 
                                  Current along-track Legendre coefficients */
    double current_coeffs_y[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS], /* I: 
                                  Current across-track Legendre coefficients */
    double *ssm_angle_correction,  /* O: TIRS SSM rotation angle correction */
    double *pitch_angle_correction /* O: Pitch alignment angle correction */
)
{
    int index;              /* Loop counter - correlation point index */
    int term;               /* Loop counter - Legendre coefficient terms */
    int sca_index;          /* SCA index */
    double TWA[2][2];       /* Accumulated matrix for the data points
                               reduces to scalar for single unknown */ 
    double A[2][2];         /* Matrix of Legendre terms for the current point */
    double TW[2][2];        /* Matrix of A transposed and multiplied by
                               the weight of current point */
    double TWY[2];          /* Matrix of error terms */
    double L[IAS_LOS_LEGENDRE_TERMS]; /* Legendre weights */
    double weight;          /* Temporary weight holder */
    double det;             /* Determinant of normal equation matrix */
    double x_point;         /* Current x point LOS from Legendre coefficients */
    double y_point;         /* Current y point LOS from Legendre coefficients */
    int points_used = 0;    /* Count of points factored into the solution */

    /* Make sure the number of points is not less than or equal to zero */
    if (corr_points <= 0)
    {
        IAS_LOG_ERROR("Number of correlation points %d is <= 0", corr_points);
        return ERROR;
    }

    /* Check for later divide by 0 */
    if (sca_num_det == 1)
    {
        IAS_LOG_ERROR("Number of detectors is 1, leading to divide by 0");
        return ERROR;
    }

    /* Initialize TWA matrix */
    TWA[0][0] = 0.0;
    TWA[1][0] = 0.0;
    TWA[0][1] = 0.0;
    TWA[1][1] = 0.0;

    /* Initialize TWY matrix */
    TWY[0] = 0.0;
    TWY[1] = 0.0;

    /* For each of the active points in the current SCA */
    for (index = 0; index < corr_points; index++)
    {
        if (corr_pts_ptr[index].active_flag == ACTIVE)
        {
            /* Update the count of the points used */
            points_used++;

            /* Find the SCA for the point */
            sca_index = corr_pts_ptr[index].sca_index;

            /* The next several sections form the Legendre terms and load the A
               matrix for this point, 1 Legendre coefficient at a time.  To
               understand indexing for A, see the ADD for the layout of A */
            L[0] = 1.0;

            /* Account for the 1st Legendre coefficient's contribution to the
               current point's x and y coordinates */
            x_point = current_coeffs_x[sca_index][0] * L[0];
            y_point = current_coeffs_y[sca_index][0] * L[0];

            /* Calculate normalized detector coordinate */
            L[1] = 2 * (corr_pts_ptr[index].srch_is_pt.samp / 
                (sca_num_det - 1)) - 1;

            /* Account for the 2nd Legendre coefficient's contribution to the
               current point's x and y coordinates */
            x_point += current_coeffs_x[sca_index][1] * L[1];
            y_point += current_coeffs_y[sca_index][1] * L[1];

            /* Account for the remaining Legendre terms */
            for (term = 2; term < IAS_LOS_LEGENDRE_TERMS; term++)
            {
               L[term] = ((2 * term - 1) * L[1] * L[term - 1] - (term - 1) 
                   * L[term - 2])/(double)term;

               /* Account for the remaining Legendre coefficients' contribution
                  to the current point's x and y coordinates */
               x_point += current_coeffs_x[sca_index][term] * L[term];
               y_point += current_coeffs_y[sca_index][term] * L[term];
            }

            /* Initialize non-0 parts of the A matrix */
            A[0][0] = -y_point;
            A[1][0] = x_point - 1.0;
            A[0][1] = -1.0;
            A[1][1] = 0.0;

            /* Form the TW matrix by multiplying the transpose of A by the
               point's weight */
            weight = corr_pts_ptr[index].pt_weight;
            TW[0][0] = A[0][0] * weight;
            TW[1][0] = A[0][1] * weight;
            TW[0][1] = A[1][0] * weight;
            TW[1][1] = A[1][1] * weight;

            /* Accumulate the TWA matrix */
            TWA[0][0] += TW[0][0] * A[0][0] + TW[0][1] * A[1][0];
            TWA[0][1] += TW[0][0] * A[0][1] + TW[0][1] * A[1][1];
            TWA[1][0] += TW[1][0] * A[0][0] + TW[1][1] * A[1][0];
            TWA[1][1] += TW[1][0] * A[0][1] + TW[1][1] * A[1][1];

            /* Accumulate the TWY matrix in the line and sample directions */
            TWY[0] += TW[0][0] * corr_pts_ptr[index].los_err.line + 
                TW[0][1] * corr_pts_ptr[index].los_err.samp;
            TWY[1] += TW[1][0] * corr_pts_ptr[index].los_err.line + 
                TW[1][1] * corr_pts_ptr[index].los_err.samp;

        } /* End of active data point */
    } /* End of looping through all the correlation data points */

    /* Verify enough points were found to fit (minimum is MIN_NUMBER_POINTS) */
    if (points_used < MIN_NUMBER_POINTS)
    {
        /* Not enough points used, generate an error message and error out */
        IAS_LOG_ERROR("Only %d correlation points found to be usable "
                "when a minimum of %d are needed", points_used, 
                MIN_NUMBER_POINTS);
        return ERROR;
    }

    /* Solve for the SSM angle in TWY */
    det = TWA[0][0]*TWA[1][1] - TWA[0][1]*TWA[1][0];

   /* Check if det is less than 1.0e-6 */
    if ( fabs(det) < 1.0e-6 )
    {
        /* Error solving for the least squares fit */
        IAS_LOG_ERROR("Calculating least squares fit");
        return ERROR;
    }
    /* Copy the SSM mirror angle correction to the output */
    *ssm_angle_correction   =  TWA[1][1]/det * TWY[0] - TWA[0][1]/det * TWY[1];
    *pitch_angle_correction = -TWA[1][0]/det * TWY[0] + TWA[0][0]/det * TWY[1];

    return SUCCESS;
}
