/******************************************************************************
NAME:           generate_legendre_polynomials (main)

PURPOSE:        
Main driver for the generate polynomial portion of TIRS Alignment calibration.  
The generate Legendre portion reads the results of the TIRS Alignment setup, 
filters the outliers, fits the data to a Legendre polynomial, updates the 
SCA models, and generates output reports.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
EXIT_FAILURE    Error generating Legendre; no output data produced 
EXIT_SUCCESS    Successful; Legendre polynomials generated
******************************************************************************/
#include <stdlib.h>             /* Standard C library */
#include <string.h>             /* String library */

#include "ias_logging.h"        /* IAS logging library */
#include "ias_model_io.h"       /* Model I/O library */
#include "focal_plane.h"        /* Main algorithm structures */
#include "read_parameters.h"    /* Read parameters */ 
#include "calculate_post_fit_residuals.h" /* calculate_post_fit_residuals 
                                   prototype */
#include "create_report_tirs_alignment_legendre.h" /* create_report prototype */
#include "filter_outliers.h"    /* filter_outliers prototype */
#include "calculate_point_weights.h" /* calculate_point_weights prototype */
#include "fit_polynomials.h"    /* fit_polynomials prototype */
#include "update_alignment_parameters.h" /* update_alignment_parameters
                                   prototype */
#include "write_alignment_matrix.h" /* write_alignment_matrix prototype */
#include "save_to_database.h"   /* DB save structure */


/******************************************************************************
NAME:           tirs_alignment_legendre_fit_cal

PURPOSE:
tirs_alignment_legendre_fit_cal calculates the fit x and fit y for
tirs_alignment_legendre application.

RETURNS: nothing

ALGORITHM REFERENCES:
OLI TIRS Alignment Calibration ADD
******************************************************************************/
void tirs_alignment_legendre_fit_cal 
(
    const double x_coordinate,  /* I: Current LOS x coordinate from original 
                                      Legendre coefficients */
    const double y_coordinate,  /* I:  Current LOS y coordinate from original
                                       Legendre coefficients */
    const PARAMETERS_FIT_CALCULATION *fit_cal, /* I: Pointer to fit calculation
                                                     parameters */
    double *fit_x,              /* O: Fit calculation x */
    double *fit_y               /* O: Fit calculation y */
)
{
    /* Calculate the modeled LOS angle corrections by multiplying the
       design matrix by the solution vector (alignment_corrections).
       See the ADD definition for design matrix A to understand the
       terms.  The design matrix is set up this way because the x and
       y coordinates have different sensitivities to roll, pitch, yaw,
       and Legendre coefficients.  For example, the x coordinate is
       sensitive to yaw, and that sensitivity increases with cross-
       track distance; hence the y * alignment_corrections[2] term */
    *fit_x = y_coordinate * fit_cal->alignment_corrections[2] 
        - fit_cal->alignment_corrections[1] + ias_math_eval_legendre(
        fit_cal->normalized_det, fit_cal->error_legendre_coeffs_x, 
        IAS_LOS_LEGENDRE_TERMS);
    
    /* Calculate the modeled LOS angle corrections by multiplying the
       design matrix by the solution vector (alignment_corrections) */    
    *fit_y = fit_cal->alignment_corrections[0] - x_coordinate 
        * fit_cal->alignment_corrections[2] + ias_math_eval_legendre(
        fit_cal->normalized_det, fit_cal->error_legendre_coeffs_y, 
        IAS_LOS_LEGENDRE_TERMS);
}

/* main function */
int main(int argc, char *argv[])
{
    ALIGNMENT_PARMS parms;      /* Parameters read from parameter file */
    CORR_PT_INFO *corr_pts;     /* Information about each correlation point */
    IAS_CPF *cpf;               /* CPF parameters */
    const struct IAS_CPF_ATTITUDE_PARAMETERS *attitude_parameters; /* Attitude 
                                   parameters from the CPF */
    double acs2tirs[3][3];      /* ACS to TIRS rotation matrix */
    double acs2oli[3][3];       /* ACS to OLI rotation matrix */
    double tirs2oli_new[3][3];  /* Updated TIRS to OLI rotation matrix */
    IAS_LOS_MODEL *model;       /* Model structure */
    int ncells;                 /* Number of cells across a SCA */
    int sca_index;              /* SCA counter/index */
    int nscas;                  /* Number of SCAs in this band (input space) */
    int sca_num_det;            /* Number of detectors per SCA */
    int npts;                   /* Number of correlation points to work with */
    int thermal1_band_number;   /* Number of reference band (TIRS_THERMAL1) */
    int thermal1_band_index;    /* Index of reference band (TIRS_THERMAL1) */
    char output_cpf_filename[PATH_MAX]; /* Name of files for writing alignment 
                                   and coefficients CPF groups */
    double sca_coef_x[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS]; /* Along-track 
                                   SCA LOS Legendre coefficients */
    double sca_coef_y[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS]; /* Across-track 
                                   SCA LOS Legendre coefficients */
    double error_coeffs_x[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS]; 
                                /* Along-track error Legendre coefficients */
    double error_coeffs_y[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS]; 
                                /* Across-track error Legendre coefficients */
    double new_sca_coef_x[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS]; 
                                /* New along-track Legendre coefficients */
    double new_sca_coef_y[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS]; 
                                /* New across-track Legendre coefficients */
    double mean_x_fit = 0.0;    /* Mean best-fit x offset */
    double mean_y_fit = 0.0;    /* Mean best-fit y offset */
    POST_FIT_RESIDUAL_STATS_TYPE residual_stats[NUM_TIRS_SCAS];
    struct main_db_save_info main_db_save_info; /* Main DB save information */ 
    struct sca_db_save_info sca_db_save_info; /* SCA level DB save info */
    double db_save_threshold_x; /* Threshold to limit database saving */
    double db_save_threshold_y; /* Threshold to limit database saving */
    int db_save_threshold_flag = 1; /* Flag to indicate results should be
                                   written to the database */
    int coef_index;             /* Legendre coefficient index */
    double alignment_corrections[3]; /* TIRS-to-OLI roll-pitch-yaw alignment
                                   corrections */
    int row;                    /* Loop counter for matrix rows */ 
    int col;                    /* Loop counter for matrix columns */ 
    PARAMETERS_FIT_CALCULATION fit; /* Structure for fit calculation */
    int app_flag = 1;           /* Applicaton flag = 1 identifies the 
                                   tirs_alignment_legendre */
   
    /* Initialize the DB save structures */
    memset(&main_db_save_info, 0, sizeof(main_db_save_info));
    memset(&sca_db_save_info, 0, sizeof(sca_db_save_info));

    /* Initialize the logging library */
    if (ias_log_initialize("generate_legendre_polynomials") != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing logging library");
        return EXIT_FAILURE;
    }

    if (read_parameters(argc, argv, &parms, &corr_pts, &cpf, &model,
        &thermal1_band_number, &thermal1_band_index, &sca_num_det, &npts,
        &nscas, &ncells, &db_save_threshold_x, &db_save_threshold_y,
        sca_coef_x, sca_coef_y) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading input parameters");
        return EXIT_FAILURE;
    }

    /* Fit correlation results to a set of Legendre error polynomials */
    if (fit_polynomials(npts, nscas, parms.fit_order, sca_num_det, corr_pts,
        parms.constraint_flag, parms.constraint_weight, sca_coef_x, sca_coef_y,
        error_coeffs_x, error_coeffs_y, alignment_corrections) != SUCCESS)
    {
        IAS_LOG_ERROR("Performing a least squares fit on the correlation "
                      "points");
        free(corr_pts);
        ias_cpf_free(cpf); 
        ias_los_model_free(model);
        model = NULL;
        return EXIT_FAILURE;
    }

    /* Fit each SCA */
    for (sca_index = 0; sca_index < nscas; sca_index++)
    {
        /* Get the number of detectors for this SCA for the THERMAL1 band */
        sca_num_det = model->sensor.bands[thermal1_band_index].
            scas[sca_index].detectors;

        /* Accumulate mean offset fit values */
        mean_x_fit += error_coeffs_x[sca_index][0];
        mean_y_fit += error_coeffs_y[sca_index][0];
        
        /* Assign values for the structure parameters_fit_calculation */
        fit.error_legendre_coeffs_x = error_coeffs_x[sca_index];
        fit.error_legendre_coeffs_y = error_coeffs_y[sca_index];
        fit.alignment_corrections = alignment_corrections;         
       
         
        /* Calculate the post fit residual statistics */
        if (calculate_post_fit_residuals(sca_coef_x[sca_index],
            sca_coef_y[sca_index], npts, sca_index,
            sca_num_det, corr_pts, &residual_stats[sca_index], 
            db_save_threshold_x, db_save_threshold_y, &db_save_threshold_flag, 
            tirs_alignment_legendre_fit_cal, &fit, app_flag) != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating scan post-fit residual statistics for "
                "SCA %d", sca_index + 1);
            free(corr_pts);
            ias_cpf_free(cpf); 
            ias_los_model_free(model);
            model = NULL;
            return EXIT_FAILURE;
        }
    }

    /* Free the model */
    ias_los_model_free(model);
    model = NULL;

    /* Check for divide by 0 */
    if (nscas == 0)
    {
        IAS_LOG_ERROR("Number of SCAs is 0, causing a divide by 0");
        free(corr_pts);
        ias_cpf_free(cpf); 
        return EXIT_FAILURE;
    }

    /* Compute the mean fit offsets */
    mean_x_fit /= (double)nscas;
    mean_y_fit /= (double)nscas;

    /* Update the SCA coefficients */
    for (sca_index = 0; sca_index < nscas; sca_index++)
    {
        for (coef_index = 0; coef_index < IAS_LOS_LEGENDRE_TERMS; coef_index++)
        {
            /* Along track */
            new_sca_coef_x[sca_index][coef_index] 
                = sca_coef_x[sca_index][coef_index] 
                + error_coeffs_x[sca_index][coef_index];
            /* Across track */
            new_sca_coef_y[sca_index][coef_index] 
                = sca_coef_y[sca_index][coef_index] 
                + error_coeffs_y[sca_index][coef_index];
        }
    }

    /* Get the CPF attitude parameters group */
    attitude_parameters = ias_cpf_get_attitude_params(cpf);
    if (!attitude_parameters)
    {
        IAS_LOG_ERROR("Getting attitude parameters group from CPF");
        ias_cpf_free(cpf);
        return EXIT_FAILURE;
    }

    /* Copy the values to avoid the const on the CPF structure */
    for (row = 0; row < 3; row++)
    {
        for (col = 0; col < 3; col++)
        {
            acs2tirs[row][col]
                = attitude_parameters->acs_to_tirs_rotation_matrix[row][col];
            acs2oli[row][col]
                = attitude_parameters->acs_to_oli_rotation_matrix[row][col];
        }
    }

    /* Compute new alignment matrices */
    if (update_alignment_parameters(alignment_corrections, acs2tirs, acs2oli,
        tirs2oli_new, &main_db_save_info) != SUCCESS) 
    {
        IAS_LOG_ERROR("Updating the CPF attitude parameters group");
        ias_cpf_free(cpf); 
        return EXIT_FAILURE;
    }

    /* Output the TIRS Alignment report */
    if (create_report_tirs_alignment_legendre(parms.report_file_name, 
        parms.terr_img_name, parms.ref_img_name, parms.l0r_img_name, 
        nscas, sca_coef_x, sca_coef_y,
        error_coeffs_x, error_coeffs_y, new_sca_coef_x, new_sca_coef_y, 
        corr_pts, npts, residual_stats, parms.outlier_tolerance, 
        parms.fit_order, parms.work_order_id, mean_x_fit, mean_y_fit, 
        &main_db_save_info) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the TIRS Alignment calibration report");
        free(corr_pts);
        ias_cpf_free(cpf); 
        return EXIT_FAILURE;
    }

    /* Free correlation points */
    free(corr_pts);
    corr_pts = NULL;

    /* Output the updated CPF parameters if requested */
    if (parms.write_cpf_flag)
    {
        /* Write the updated alignment matrix to a CPF group fragment */
        if (sprintf(output_cpf_filename, "ATTITUDE_PARAMETERS_%08d_%08d.cpf", 
            parms.start_effective_date, parms.end_effective_date) < 0)
        {
            IAS_LOG_ERROR("Constructing the attitude parameter CPF group file "
                "name for dates %d to %d", parms.start_effective_date,
                parms.end_effective_date);
            ias_cpf_free(cpf); 
            return EXIT_FAILURE;
        }

        if (write_alignment_matrix(output_cpf_filename, tirs2oli_new, acs2oli)
            != SUCCESS)
        {
            IAS_LOG_ERROR("Writing new attitude parameters CPF group");
            ias_cpf_free(cpf); 
            return EXIT_FAILURE;
        }

        /* Write the updated Legendre coefficients to a CPF group fragment */
        if (sprintf(output_cpf_filename, "TIRS_FOCAL_PLANE_%08d_%08d.cpf",
            parms.start_effective_date, parms.end_effective_date) < 0)
        {
            IAS_LOG_ERROR("Constructing the Legendre coefficient CPF group "
                "file name for dates %d to %d", parms.start_effective_date,
                parms.end_effective_date);
            ias_cpf_free(cpf); 
            return EXIT_FAILURE;
        }

        if (write_sca_parameters_cpf(output_cpf_filename, error_coeffs_x, 
            error_coeffs_y, cpf, IAS_TIRS) != SUCCESS)
        {
            IAS_LOG_ERROR("Writing the updated CPF group report file %s",
                output_cpf_filename);
            ias_cpf_free(cpf); 
            return EXIT_FAILURE;
        }
    }

    /* Free CPF */
    ias_cpf_free(cpf); 
    cpf = NULL;

    /* If storing results in the database is requested, and not rejected by 
       the DB save thresholds */
    if ((db_save_threshold_flag) && (parms.store_in_database_flag))
    {
        IAS_LOG_INFO("Writing results to database");

        /* Populate database structures */
        prepare_for_database_save(sca_coef_x, sca_coef_y, error_coeffs_x,
            error_coeffs_y, new_sca_coef_x, new_sca_coef_y, residual_stats, 
            parms.work_order_id, parms.ref_img_name, parms.outlier_tolerance,
            parms.fit_order, nscas, &main_db_save_info, &sca_db_save_info); 

        /* Send data in database structures to database */
        if (save_to_database(&main_db_save_info, &sca_db_save_info, nscas)
            != SUCCESS)
        {
            IAS_LOG_ERROR("Storing results in database");
            return EXIT_FAILURE;
        }
    }

    IAS_LOG_INFO("Successful completion");

    return EXIT_SUCCESS;
}
