/******************************************************************************
NAME:          read_parameters 

PURPOSE:
Reads all the input parameters, model, corelation information, filter outliers,
and calculates point weight. This function is used in both applications.

RETURNS:
Type = int
Value           Description
-----           -----------
ERROR           Error
SUCCESS         Success

******************************************************************************/

#include <stdlib.h>
#include "ias_logging.h"    /* IAS logging library */
#include "filter_outliers.h"
#include "calculate_point_weights.h"
#include "read_parameters.h"

int read_parameters
(
    int argc,               /* I: Command line arguments count */
    char *argv[],           /* I: Command line arguments */
    ALIGNMENT_PARMS *parms, /* I: Parameters read from parameter file */
    CORR_PT_INFO **corr_pts,/* O: Information about each correlation point */
    IAS_CPF **cpf,          /* O: CPF parameters */
    IAS_LOS_MODEL **model,  /* O: Model structure */
    int *thermal1_band_number,/* O: Number of reference 
                                      band (TIRS_THERMAL1) */
    int *thermal1_band_index, /* O: Index of reference band (TIRS_THERMAL1) */
    int *sca_num_det,       /* O: Number of detectors per SCA */
    int *npts,              /* O: Number of correlation points to work with */
    int *nscas,             /* O Number of SCAs in this band (input space) */
    int *ncells,            /* O: Number of cells across a SCA */
    double *db_save_threshold_x, /* O: Threshold to limit database saving */
    double *db_save_threshold_y, /* O: Threshold to limit database saving */
    double sca_coef_x[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS],/* I: Along-track 
                                   SCA LOS Legendre coefficients */
    double sca_coef_y[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS] /* I: Across-track 
                                   SCA LOS Legendre coefficients */
)
{
    int status;      /* Function return status */
    int sca_index;   /* SCA counter/index */
    int coef_index;  /* Legendre coefficient index */

    /* Get parameters and setup information from get_parameters which is shared
       among Focal Plane and TIRS alignment setup and Legendre phases */
    status = get_parameters(argc, argv, parms, cpf); 
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting input parameters");
        return ERROR;
    }

    /* Get the TIRS_THERMAL1 band number, which will be the reference band */
    *thermal1_band_number = ias_sat_attr_get_band_number_from_type(
        IAS_THERMAL1_BAND);
    if (*thermal1_band_number == ERROR)
    {
        IAS_LOG_ERROR("Invalid thermal 1 band number %d", 
            *thermal1_band_number);
        ias_cpf_free(*cpf); 
        return ERROR;
    }

    /* Get the TIRS_THERMAL1 band index */
    *thermal1_band_index = ias_sat_attr_convert_band_number_to_index(
        *thermal1_band_number);
    if (*thermal1_band_index == ERROR)
    {
        IAS_LOG_ERROR("Invalid band index for thermal 1 band number %d", 
            *thermal1_band_number);
        ias_cpf_free(*cpf); 
        return ERROR;
    }

    /* Log initiation */
    IAS_LOG_INFO("Initiated");

    /* Load the DB save thresholds and convert to radians */
    *db_save_threshold_x = parms->along_trend_threshold * 1e-6;
    *db_save_threshold_y = parms->across_trend_threshold * 1e-6;

    /* Read the model file */
    *model = ias_model_read(parms->model_file_name);
    if (*model == NULL)
    {
        IAS_LOG_ERROR("Reading the model file %s", parms->model_file_name);
        ias_cpf_free(*cpf); 
        return ERROR;
    }

    /* Get the number of detectors for SCA 0 */
    *sca_num_det = 
        (*model)->sensor.bands[*thermal1_band_index].scas[0].detectors;

    /* Get the existing thermal 1 band SCA polynomials from the model */
    for (sca_index = 0; sca_index 
        < (*model)->sensor.bands[*thermal1_band_index].sca_count; sca_index++)
    {
        for (coef_index = 0; coef_index < IAS_LOS_LEGENDRE_TERMS; coef_index++)
        {
            sca_coef_x[sca_index][coef_index] = 
                (*model)->sensor.bands[*thermal1_band_index].scas[sca_index].
                sca_coef_x[coef_index];

            sca_coef_y[sca_index][coef_index] = 
                (*model)->sensor.bands[*thermal1_band_index].scas[sca_index].
                sca_coef_y[coef_index]; 
        }
    }

    /* Read the correlation results from the TIRS Alignment or 
       TIRS SSM setup phase */
    *corr_pts = read_correlation_info(parms->correlation_file_name, npts, 
        nscas, ncells);
    if (*corr_pts == NULL)
    {
        IAS_LOG_ERROR("Reading correlation information from %s",
            parms->correlation_file_name); 
        ias_cpf_free(*cpf); 
        ias_los_model_free(*model);
        (*model) = NULL;
        return ERROR;
    }

    /* Filter the outliers in the correlation data */
    if (filter_outliers(*ncells, parms->outlier_tolerance, *npts, *corr_pts) 
        != SUCCESS)
    {
        IAS_LOG_ERROR("Filtering the outlier points from the correlation "
                "information from %s", parms->correlation_file_name);
        free(*corr_pts);
        ias_cpf_free(*cpf); 
        ias_los_model_free(*model);
        (*model) = NULL;
        return ERROR;
    }

    /* Weight the remaining correlation points */
    calculate_point_weights(parms->point_weight, *npts, *corr_pts);

    return SUCCESS;
}

