/*****************************************************************************
NAME:      save_estimated_ssm_to_database.c

PURPPOSE:
Send algorithm report data to the database for TIRS SSM Calibration algorithm.

RETURNS: SUCCESS or ERROR
******************************************************************************/
/* Standard library headers */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* IAS library headers */
#include "ias_db_insert.h"        /* Database library */
#include "ias_db_insert_transaction_using_table.h"   /* Database library */
#include "ias_db_get_connect_info.h" /* Database library */
#include "ias_miscellaneous.h"    /* Used for ias_misc_get_char_id */
#include "ias_logging.h"          /* Header definitions for this file */

/* Local header files */
#include "save_estimated_ssm_to_database.h" /* Header definitions to save
                                               the data in database for
                                               estimate_tirs_ssm_angle
                                               application */

#define MAX_DELETE_QUERY_LENGTH 85 /* Maximum size of delete query including
                                      ample space for table name and WO ID */

int save_estimated_ssm_to_database
(
    struct main_db_trend_info* main_db_trend, /* I/O: Top-level data to save */
    struct sca_db_trend_info* sca_db_trend,   /* I/O: SCA-level data to save */
    int number_sca_records                    /* I: Number of records to submit 
                                                    for sca database */
)
{
    int status;     /* Function return status */
    int main_table_size;    /* Size of the main database table */
    int sca_table_size;     /* Size of the SCA database table */
    char *database_name = NULL; /* Database name */
    char *user_name = NULL;     /* Database user name */
    char *password = NULL;      /* Database password */
    char *host = NULL;          /* Database host */
    char main_database_table_name[] = "tirs_ssm_estimation"; /* This table has
                                     records that apply to the whole
                                     application run */
    char sca_database_table_name[] = "tirs_ssm_estimation_sca"; /* This table
                                     has per-SCA records */
    char geometric_wo_common_id[IAS_CHAR_ID_SIZE]; /* Promary key in
                                     GEOMETRIC_WORK_ORDER_COMMON table to
                                     associate processing with the WO */
    struct ias_db_connection *db = NULL; /* Database connection */
    char query_string[MAX_DELETE_QUERY_LENGTH]; /* SQL statement sting */
    struct ias_db_query *query = NULL; /* SQL query */
    char error_message[IAS_SQL_RTN_MSGLEN]; /* SQL query error message */

    /* Get the database connection info */
    if (ias_db_get_connect_info(&database_name, &user_name, &password, &host)
        != SUCCESS)
    {
        /* Failed getting the database connection info */
        IAS_LOG_ERROR("Getting database connection information");
        return ERROR;
    }
    
    /* Connect to the database for a transaction */
    db = ias_db_connect_with_transaction(database_name, user_name, password,
            host);

    /* Free the database connection information */
    free(database_name);
    database_name = NULL;
    free(user_name);
    user_name = NULL;
    free(password);
    password = NULL;
    free(host);
    host = NULL;

    if (db == NULL)
    {
        IAS_LOG_ERROR("Connecting to database");
        return ERROR;
    }

    /* Get the database ID corresponding to the current work order */
    status = ias_misc_get_char_id(db,
        "GEOMETRIC_WORK_ORDER_COMMON", "geometric_wo_common_id",
        main_db_trend->wo_id, geometric_wo_common_id);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Retrieving database ID corresponding to current "
            "geometric work order %s", main_db_trend->wo_id);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Build the top-level table of data to save */
    IAS_DB_INSERT_TABLE main_table[] =
    {
        IAS_DB_TABLE_STRING("geometric_wo_common_id", geometric_wo_common_id),
        IAS_DB_TABLE_STRUCT_STRING(main_db_trend, reference_image),
        IAS_DB_TABLE_STRUCT_INT(main_db_trend, number_of_scas),
        IAS_DB_TABLE_STRUCT_DOUBLE(main_db_trend, ssm_epoch_year),
        IAS_DB_TABLE_STRUCT_DOUBLE(main_db_trend, ssm_epoch_doy),
        IAS_DB_TABLE_STRUCT_DOUBLE(main_db_trend, ssm_epoch_sod),
        IAS_DB_TABLE_STRUCT_DOUBLE(main_db_trend, ssm_angle_correction),
        IAS_DB_TABLE_STRUCT_DOUBLE(main_db_trend, net_ssm_angle),
        IAS_DB_TABLE_STRUCT_DOUBLE(main_db_trend, pitch_angle_correction),
        IAS_DB_TABLE_STRUCT_DOUBLE(main_db_trend, confidence_level),
    };

    main_table_size = sizeof(main_table) / sizeof(IAS_DB_INSERT_TABLE);

    /* Delete the previous records */
    status = snprintf(query_string, sizeof(query_string), "DELETE FROM %s "
        "WHERE geometric_wo_common_id=%s", main_database_table_name,
        geometric_wo_common_id);
    if (status < 0 || status >= sizeof(query_string))
    {
        IAS_LOG_ERROR("Preparing query string: DELETE FROM %s where "
            "geometric_wo_common_id=%s", main_database_table_name,
            geometric_wo_common_id);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }
   
    query = ias_db_perform_query(db, query_string);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, error_message, sizeof(
            error_message));
        IAS_LOG_ERROR("Query of %s failed: %s", main_database_table_name,
            error_message);
        IAS_LOG_ERROR("Deleting records from %s where geometric_wo_common_id="
            "%s", main_database_table_name, geometric_wo_common_id);
        ias_db_query_close(query);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Close query */
    ias_db_query_close(query);
    query = NULL;
    
    /* Insert the data into the main database table */
    status = ias_db_insert_using_table(db, main_database_table_name,
        main_table, main_table_size, 1);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Inserting to %s DB table", main_database_table_name);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Build the SCA-level table of data to save */
    IAS_DB_INSERT_TABLE sca_table[] =
    {
        IAS_DB_TABLE_STRING("geometric_wo_common_id", geometric_wo_common_id),
        IAS_DB_TABLE_STRUCT_INT_ARRAY(sca_db_trend, sca_number),
        IAS_DB_TABLE_STRUCT_INT_ARRAY(sca_db_trend, nbr_of_pnts_used),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_trend, pre_fit_alng_trk_mean),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_trend, pre_fit_alng_trk_stdev),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_trend, pre_fit_alng_trk_rmse),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_trend, pre_fit_acrss_trk_mean),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_trend, pre_fit_acrss_trk_stdev),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_trend, pre_fit_acrss_trk_rmse),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_trend, post_fit_alng_trk_mean),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_trend, post_fit_alng_trk_stdev),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_trend, post_fit_alng_trk_rmse),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_trend, post_fit_acrss_trk_mean),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_trend, 
            post_fit_acrss_trk_stdev),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_trend, post_fit_acrss_trk_rmse),
    };

    sca_table_size = sizeof(sca_table) / sizeof(IAS_DB_INSERT_TABLE);

    /* Delete the previous records */
    status = snprintf(query_string, sizeof(query_string),
        "DELETE FROM %s WHERE "
        "geometric_wo_common_id=%s", sca_database_table_name,
        geometric_wo_common_id);
    if (status < 0 || status >= sizeof(query_string))
    {
        IAS_LOG_ERROR("Preparing query string: DELETE FROM %s WHERE "
            "geometric_wo_common_id=%s", sca_database_table_name,
            geometric_wo_common_id);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    query = ias_db_perform_query(db, query_string);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, error_message, sizeof(
            error_message));
        IAS_LOG_ERROR("Query of %s failed: %s", sca_database_table_name,
            error_message);
        IAS_LOG_ERROR("Deleting records from %s where geometric_wo_common_id="
            "%s", sca_database_table_name, geometric_wo_common_id);
        ias_db_query_close(query);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Close query */
    ias_db_query_close(query);
    query = NULL;

    /* Insert the data into the SCA database table */
    status = ias_db_insert_using_table(db, sca_database_table_name, sca_table,
        sca_table_size, number_sca_records);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Inserting to %s DB table", sca_database_table_name);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* The inserts worked, so commit the transaction */
    status = ias_db_disconnect_with_transaction(db, IAS_DB_COMMIT);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Committing transaction and disconnecting from database");
    }

    return status;
}


/******************************************************************************
NAME:      prepare_estimated_ssm_for_database_trend

PURPOSE:
Populate the database structures.

RETURNS: none 
******************************************************************************/
void prepare_estimated_ssm_for_database_trend
(
    const double *ssm_epoch,       /* I: Time epoch for estimated angles */
    double ssm_angle_correction,   /* I: Estimated SSM angle correction */
    double net_ssm_angle,          /* I: Net new estimated angle */
    double pitch_angle_correction, /* I: Estimated pitch alignment angle
                                         correction */ 
    const POST_FIT_RESIDUAL_STATS_TYPE *residual_stats, /* I: Calculated
                                          residual information */
    double confidence_level,       /* I: Confidence level used for outlier
                                          rejection */
    const char *work_order_id,     /* I: Work order ID for report header */
    const char *ref_image_name,    /* I: Name of reference image */
    int number_sca_records,            /* I: The number of DB records */
    struct main_db_trend_info* main_db_trend, /* O: Populated main table data */
    struct sca_db_trend_info* sca_db_trend    /* O: Populated SCA table data */
)
{
    int sca_index;  /* SCA counter/index */

    /* Populate the SCA-level database results structure */
    for (sca_index = 0; sca_index < number_sca_records; sca_index++)
    {
        /* Miscellaneous fields - SCA and number of points used */
        sca_db_trend->nbr_of_pnts_used[sca_index] = 
            residual_stats[sca_index].points;
        sca_db_trend->sca_number[sca_index] = sca_index + 1;
        
        /* Get fit statistics information to save to the Db */
        
        /* Pre fit and scale it from radians to micro-radians */
        sca_db_trend->pre_fit_alng_trk_mean[sca_index] =
            residual_stats[sca_index].delta_x.mean * 1E6;
        sca_db_trend->pre_fit_alng_trk_stdev[sca_index] =
            residual_stats[sca_index].delta_x.stdev * 1E6;
        sca_db_trend->pre_fit_alng_trk_rmse[sca_index] =
            residual_stats[sca_index].delta_x.rms * 1E6;

        sca_db_trend->pre_fit_acrss_trk_mean[sca_index] = 
            residual_stats[sca_index].delta_y.mean * 1E6;
        sca_db_trend->pre_fit_acrss_trk_stdev[sca_index] = 
            residual_stats[sca_index].delta_y.stdev * 1E6;
        sca_db_trend->pre_fit_acrss_trk_rmse[sca_index] = 
            residual_stats[sca_index].delta_y.rms * 1E6;

        /* Post fit and scale it from radians to micro-radians */
        sca_db_trend->post_fit_alng_trk_mean[sca_index] = 
            residual_stats[sca_index].residual_x.mean * 1E6;
        sca_db_trend->post_fit_alng_trk_stdev[sca_index] = 
            residual_stats[sca_index].residual_x.stdev * 1E6;
        sca_db_trend->post_fit_alng_trk_rmse[sca_index] =
            residual_stats[sca_index].residual_x.rms * 1E6;

        sca_db_trend->post_fit_acrss_trk_mean[sca_index] =
            residual_stats[sca_index].residual_y.mean * 1E6;
        sca_db_trend->post_fit_acrss_trk_stdev[sca_index] =
            residual_stats[sca_index].residual_y.stdev * 1E6;
        sca_db_trend->post_fit_acrss_trk_rmse[sca_index] =
            residual_stats[sca_index].residual_y.rms * 1E6;
    }        
    snprintf(main_db_trend->wo_id, sizeof(main_db_trend->wo_id), "%s",
        work_order_id);
    snprintf(main_db_trend->reference_image,
        sizeof(main_db_trend->reference_image), "%s", ref_image_name);

    main_db_trend->number_of_scas = number_sca_records; 
    main_db_trend->ssm_epoch_year = ssm_epoch[0];
    main_db_trend->ssm_epoch_doy = ssm_epoch[1];
    main_db_trend->ssm_epoch_sod = ssm_epoch[2];
    main_db_trend->ssm_angle_correction = ssm_angle_correction;
    main_db_trend->net_ssm_angle = net_ssm_angle;
    main_db_trend->pitch_angle_correction = pitch_angle_correction;
    main_db_trend->confidence_level = confidence_level;
}
