/******************************************************************************
NAME:          save_to_database.c 

PURPOSE:
Sends algorithm report data to the database.

RETURNS: SUCCESS or ERROR
******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "ias_db_insert.h"     /* Database library */
#include "ias_db_insert_transaction_using_table.h" /* Database library */
#include "ias_db_get_connect_info.h" /* Database library */
#include "ias_miscellaneous.h" /* Used for ias_misc_get_char_id */
#include "ias_logging.h"       /* IAS logging library */
#include "save_to_database.h"  /* Header definitions for this file */
#include "focal_plane.h"       /* Header definitions for whole application */

#define MAX_DELETE_QUERY_LENGTH 85 /* Maximum size of delete query including
                                  ample space for table name and WO ID */

int save_to_database
(
    struct main_db_save_info* main_db_save, /* I: Top-level data to save */
    struct sca_db_save_info* sca_db_save, /* I: SCA-level data to save */
    int number_records            /* I: Number of records to submit */
)
{
    int status;                   /* Function return status */
    int main_table_size;          /* Size of the main database table */
    int sca_table_size;           /* Size of the SCA database table */
    char *database_name = NULL;   /* Database name */
    char *user_name = NULL;       /* Database user name */
    char *password = NULL;        /* Database password */
    char *host = NULL;            /* Database host */
    char main_database_table_name[] = "tirs_alignment"; /* This table has
                                     records that apply to the whole
                                     application run */
    char sca_database_table_name[] = "tirs_alignment_sca"; /* This table has
                                     per-SCA records */ 
    char geometric_wo_common_id[IAS_CHAR_ID_SIZE];     
                                  /* Primary key in GEOMETRIC_WORK_ORDER_COMMON
                                     to associate processing with the WO */
    struct ias_db_connection *db = NULL; /* Database connection */
    char query_string[MAX_DELETE_QUERY_LENGTH]; /* SQL statement string */
    struct ias_db_query *query = NULL; /* SQL query */
    char error_message[IAS_SQL_RTN_MSGLEN]; /* SQL query error message */

    /* Get the database connection info */
    if (ias_db_get_connect_info(&database_name, &user_name, &password, &host)
        != SUCCESS)
    {
        /* Failed getting the database connection info */
        IAS_LOG_ERROR("Getting database connection information");
        return ERROR;
    }

    /* Connect to the database for a transaction */
    db = ias_db_connect_with_transaction(database_name, user_name, password,
             host);

    /* Free the database connection information */
    free(database_name);
    free(user_name);
    free(password);
    free(host);

    if (db == NULL)
    {
        IAS_LOG_ERROR("Connecting to database");
        return ERROR;
    }

    /* Get the database ID corresponding to the current work order */
    status = ias_misc_get_char_id(db,
        "GEOMETRIC_WORK_ORDER_COMMON", "geometric_wo_common_id",
        main_db_save->wo_id, geometric_wo_common_id);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Retrieving database ID corresponding to current "
            "geometric work order %s", main_db_save->wo_id);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Put the Geometric WO common ID in the main table */
    strncpy(main_db_save->geometric_wo_common_id, geometric_wo_common_id, 
        sizeof(main_db_save->geometric_wo_common_id));
    main_db_save->geometric_wo_common_id[sizeof(
        main_db_save->geometric_wo_common_id) - 1] = '\0';

    /* Build the top-level table of data to save */
    IAS_DB_INSERT_TABLE main_table[] =
    {
        IAS_DB_TABLE_STRUCT_STRING(main_db_save, geometric_wo_common_id),
        IAS_DB_TABLE_STRUCT_STRING(main_db_save, reference_image),
        IAS_DB_TABLE_STRUCT_INT(main_db_save, number_of_scas),
        IAS_DB_TABLE_STRUCT_DOUBLE(main_db_save, original_tirs_oli_roll),
        IAS_DB_TABLE_STRUCT_DOUBLE(main_db_save, original_tirs_oli_pitch),
        IAS_DB_TABLE_STRUCT_DOUBLE(main_db_save, original_tirs_oli_yaw),
        IAS_DB_TABLE_STRUCT_DOUBLE(main_db_save, est_tirs_oli_roll_corr),
        IAS_DB_TABLE_STRUCT_DOUBLE(main_db_save, est_tirs_oli_pitch_corr),
        IAS_DB_TABLE_STRUCT_DOUBLE(main_db_save, est_tirs_oli_yaw_corr),
        IAS_DB_TABLE_STRUCT_DOUBLE(main_db_save, updated_tirs_oli_roll),
        IAS_DB_TABLE_STRUCT_DOUBLE(main_db_save, updated_tirs_oli_pitch),
        IAS_DB_TABLE_STRUCT_DOUBLE(main_db_save, updated_tirs_oli_yaw),
        IAS_DB_TABLE_STRUCT_INT(main_db_save, fit_order),
        IAS_DB_TABLE_STRUCT_DOUBLE(main_db_save, outlier_threshold),
    };

    main_table_size = sizeof(main_table) / sizeof(IAS_DB_INSERT_TABLE);

    /* Delete the previous records */
    status = snprintf(query_string, sizeof(query_string), "DELETE FROM %s "
        "WHERE geometric_wo_common_id=%s", main_database_table_name, 
        geometric_wo_common_id);
    if (status < 0 || status >= sizeof(query_string))
    {
        IAS_LOG_ERROR("Preparing query string: DELETE FROM %s WHERE "
            "geometric_wo_common_id=%s", main_database_table_name,
            geometric_wo_common_id);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    query = ias_db_perform_query(db, query_string);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, error_message, sizeof(
            error_message));
        IAS_LOG_ERROR("Query of %s failed: %s", main_database_table_name,
            error_message);
        IAS_LOG_ERROR("Deleting records from %s where geometric_wo_common_id="
            "%s", main_database_table_name, geometric_wo_common_id);
        ias_db_query_close(query);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Close query */
    ias_db_query_close(query);
    query = NULL;

    /* Insert the data into the main database table */
    status = ias_db_insert_using_table(db, main_database_table_name, main_table,
        main_table_size, 1);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Inserting to %s DB table", main_database_table_name);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Put the Geometric WO common ID in the SCA table */
    strncpy(sca_db_save->geometric_wo_common_id, geometric_wo_common_id, 
        sizeof(sca_db_save->geometric_wo_common_id));
    sca_db_save->geometric_wo_common_id[sizeof(
        sca_db_save->geometric_wo_common_id) - 1] = '\0';

    /* Build the SCA-level table of data to save */
    IAS_DB_INSERT_TABLE sca_table[] =
    {
        IAS_DB_TABLE_STRUCT_STRING(sca_db_save, geometric_wo_common_id),
        IAS_DB_TABLE_STRUCT_INT_ARRAY(sca_db_save, sensor_chip_assembly_id),
        IAS_DB_TABLE_STRUCT_INT_ARRAY(sca_db_save, nbr_of_pnts_used),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            upds_alng_trk_leg_poly_0),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            upds_alng_trk_leg_poly_1),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            upds_alng_trk_leg_poly_2),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            upds_alng_trk_leg_poly_3),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            upds_acrss_trk_leg_poly_0),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            upds_acrss_trk_leg_poly_1),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            upds_acrss_trk_leg_poly_2),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            upds_acrss_trk_leg_poly_3),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, new_alng_trk_leg_poly_0),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, new_alng_trk_leg_poly_1),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, new_alng_trk_leg_poly_2),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, new_alng_trk_leg_poly_3),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            new_acrss_trk_leg_poly_0),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            new_acrss_trk_leg_poly_1),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            new_acrss_trk_leg_poly_2),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            new_acrss_trk_leg_poly_3),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            pst_fit_res_alng_trk_mean),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            pst_fit_res_alng_trk_std_dev),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            pst_fit_res_alng_trk_rmse),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            pst_fit_res_acrss_trk_mean),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            pst_fit_res_acrss_trk_std_dev),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            pst_fit_res_acrss_trk_rmse),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            original_alng_trk_leg_poly_0),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            original_alng_trk_leg_poly_1),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            original_alng_trk_leg_poly_2),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            original_alng_trk_leg_poly_3),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            original_acrss_trk_leg_poly_0),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            original_acrss_trk_leg_poly_1),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            original_acrss_trk_leg_poly_2),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            original_acrss_trk_leg_poly_3),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            pre_fit_res_alng_trk_mean),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            pre_fit_res_alng_trk_std_dev),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            pre_fit_res_alng_trk_rmse),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            pre_fit_res_acrss_trk_mean),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            pre_fit_res_acrss_trk_std_dev),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(sca_db_save, 
            pre_fit_res_acrss_trk_rmse),
    };

    sca_table_size = sizeof(sca_table) / sizeof(IAS_DB_INSERT_TABLE);

    /* Delete the previous records */
    status = snprintf(query_string, sizeof(query_string), 
        "DELETE FROM %s WHERE "
        "geometric_wo_common_id=%s", sca_database_table_name, 
        geometric_wo_common_id);
    if (status < 0 || status >= sizeof(query_string))
    {
        IAS_LOG_ERROR("Preparing query string: DELETE FROM %s WHERE "
            "geometric_wo_common_id=%s", sca_database_table_name,
            geometric_wo_common_id);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    query = ias_db_perform_query(db, query_string);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, error_message, sizeof(
            error_message));
        IAS_LOG_ERROR("Query of %s failed: %s", sca_database_table_name,
            error_message);
        IAS_LOG_ERROR("Deleting records from %s where geometric_wo_common_id="
            "%s", sca_database_table_name, geometric_wo_common_id);
        ias_db_query_close(query);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Close query */
    ias_db_query_close(query);
    query = NULL;

    /* Insert the data into the SCA database table */
    status = ias_db_insert_using_table(db, sca_database_table_name, sca_table, 
        sca_table_size, number_records);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Inserting to %s DB table", sca_database_table_name);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* The inserts worked, so commit the transaction */
    status = ias_db_disconnect_with_transaction(db, IAS_DB_COMMIT);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Committing transaction and disconnecting from database");
    }

    return status;
}

/******************************************************************************
NAME:          prepare_for_database_save

PURPOSE:
Populate the database structures.

RETURNS: none 
******************************************************************************/
void prepare_for_database_save
(
    double sca_coef_x[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS], /* I: Existing X
                                        Legendre coefficients */
    double sca_coef_y[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS], /* I: Existing Y
                                        Legendre coefficients */
    double error_coef_x[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS], /* I: Error
                                        X Legendre coefficients */
    double error_coef_y[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS], /* I: Error
                                        Y Legendre coefficients */
    double new_sca_coef_x[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS],
                                  /* I: Updated X coefficients */
    double new_sca_coef_y[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS],
                                  /* I: Updated Y coefficients */
    const POST_FIT_RESIDUAL_STATS_TYPE *residual_stats, /* I: Calculated
                                        residual information */
    const char *work_order_id,    /* I: Work Order ID for report header */
    const char *ref_image_name,   /* I: Name of reference image */
    double confidence_level,       /* I: Confidence level used for outlier
                                        rejection */
    int fit_order,                /* I: Flag that indicates the polynomial
                                        order that was used in the solution */
    int number_records,           /* I: The number of DB records */
    struct main_db_save_info* main_db_save, /* O: Populated main table data */
    struct sca_db_save_info* sca_db_save /* O: Populated SCA table data */
)
{
    int sca_index;                /* SCA counter/index */

    /* Populate the SCA-level database results structure */
    for (sca_index = 0; sca_index < number_records; sca_index++)
    {
        /* Miscellaneous fields - SCA and number of points used */
        sca_db_save->sensor_chip_assembly_id[sca_index] = sca_index + 1;
        sca_db_save->nbr_of_pnts_used[sca_index] =
            residual_stats[sca_index].points;

        /* Get along/across track poly information to save to the DB */

        /* Original coefficients */
        sca_db_save->original_alng_trk_leg_poly_0[sca_index] =
            sca_coef_x[sca_index][0];
        sca_db_save->original_alng_trk_leg_poly_1[sca_index] =
            sca_coef_x[sca_index][1];
        sca_db_save->original_alng_trk_leg_poly_2[sca_index] =
            sca_coef_x[sca_index][2];
        sca_db_save->original_alng_trk_leg_poly_3[sca_index] =
            sca_coef_x[sca_index][3];
        sca_db_save->original_acrss_trk_leg_poly_0[sca_index] =
            sca_coef_y[sca_index][0];
        sca_db_save->original_acrss_trk_leg_poly_1[sca_index] =
            sca_coef_y[sca_index][1];
        sca_db_save->original_acrss_trk_leg_poly_2[sca_index] =
            sca_coef_y[sca_index][2];
        sca_db_save->original_acrss_trk_leg_poly_3[sca_index] =
            sca_coef_y[sca_index][3];

        /* Differences in coefficients */
        sca_db_save->upds_alng_trk_leg_poly_0[sca_index] =
            error_coef_x[sca_index][0];
        sca_db_save->upds_alng_trk_leg_poly_1[sca_index] =
            error_coef_x[sca_index][1];
        sca_db_save->upds_alng_trk_leg_poly_2[sca_index] =
            error_coef_x[sca_index][2];
        sca_db_save->upds_alng_trk_leg_poly_3[sca_index] =
            error_coef_x[sca_index][3];
        sca_db_save->upds_acrss_trk_leg_poly_0[sca_index]=
            error_coef_y[sca_index][0];
        sca_db_save->upds_acrss_trk_leg_poly_1[sca_index]=
            error_coef_y[sca_index][1];
        sca_db_save->upds_acrss_trk_leg_poly_2[sca_index]=
            error_coef_y[sca_index][2];
        sca_db_save->upds_acrss_trk_leg_poly_3[sca_index]=
            error_coef_y[sca_index][3];

        /* New coefficients */
        sca_db_save->new_alng_trk_leg_poly_0[sca_index] =
            new_sca_coef_x[sca_index][0];
        sca_db_save->new_alng_trk_leg_poly_1[sca_index] =
            new_sca_coef_x[sca_index][1];
        sca_db_save->new_alng_trk_leg_poly_2[sca_index] =
            new_sca_coef_x[sca_index][2];
        sca_db_save->new_alng_trk_leg_poly_3[sca_index] =
            new_sca_coef_x[sca_index][3];
        sca_db_save->new_acrss_trk_leg_poly_0[sca_index] =
            new_sca_coef_y[sca_index][0];
        sca_db_save->new_acrss_trk_leg_poly_1[sca_index] =
            new_sca_coef_y[sca_index][1];
        sca_db_save->new_acrss_trk_leg_poly_2[sca_index] =
            new_sca_coef_y[sca_index][2];
        sca_db_save->new_acrss_trk_leg_poly_3[sca_index] =
            new_sca_coef_y[sca_index][3];

        /* Get fit statistics information to save to the DB */

        /* Pre fit */
        sca_db_save->pre_fit_res_alng_trk_mean[sca_index] =
            residual_stats[sca_index].delta_x.mean * 1E6;
        sca_db_save->pre_fit_res_alng_trk_std_dev[sca_index] =
            residual_stats[sca_index].delta_x.stdev * 1E6;
        sca_db_save->pre_fit_res_alng_trk_rmse[sca_index] =
            residual_stats[sca_index].delta_x.rms * 1E6;
        sca_db_save->pre_fit_res_acrss_trk_mean[sca_index] =
            residual_stats[sca_index].delta_y.mean * 1E6;
        sca_db_save->pre_fit_res_acrss_trk_std_dev[sca_index] =
            residual_stats[sca_index].delta_y.stdev * 1E6;
        sca_db_save->pre_fit_res_acrss_trk_rmse[sca_index] =
            residual_stats[sca_index].delta_y.rms * 1E6;

        /* Post fit */
        sca_db_save->pst_fit_res_alng_trk_mean[sca_index] =
            residual_stats[sca_index].residual_x.mean * 1E6;
        sca_db_save->pst_fit_res_alng_trk_std_dev[sca_index] =
            residual_stats[sca_index].residual_x.stdev * 1E6;
        sca_db_save->pst_fit_res_alng_trk_rmse[sca_index] =
            residual_stats[sca_index].residual_x.rms * 1E6;
        sca_db_save->pst_fit_res_acrss_trk_mean[sca_index] =
            residual_stats[sca_index].residual_y.mean * 1E6;
        sca_db_save->pst_fit_res_acrss_trk_std_dev[sca_index] =
            residual_stats[sca_index].residual_y.stdev * 1E6;
        sca_db_save->pst_fit_res_acrss_trk_rmse[sca_index] =
            residual_stats[sca_index].residual_y.rms * 1E6;
    }

    snprintf(main_db_save->wo_id, sizeof(main_db_save->wo_id), "%s", 
        work_order_id);
    snprintf(main_db_save->reference_image,
        sizeof(main_db_save->reference_image), "%s", ref_image_name);
    main_db_save->number_of_scas = number_records;
    main_db_save->outlier_threshold = confidence_level;
    main_db_save->fit_order = fit_order;
}


