#ifndef SAVE_TO_DATABASE_H
#define SAVE_TO_DATABASE_H

#include <limits.h>      /* PATH_MAX */

#include "focal_plane.h" 
#include "calculate_post_fit_residuals.h" /* POST_FIT_RESIDUAL_STATS_TYPE
                               definition */

/* Define a structure for the information to be collected and inserted into
   the TIRS_ALIGNMENT table */
struct main_db_save_info
{
    char wo_id[IAS_WORKORDER_ID_SIZE]; /* This one isn't in the database */
    char geometric_wo_common_id[IAS_CHAR_ID_SIZE];  
    char reference_image[PATH_MAX + 1];
    int number_of_scas;
    double original_tirs_oli_roll;
    double original_tirs_oli_pitch;
    double original_tirs_oli_yaw;
    double est_tirs_oli_roll_corr;
    double est_tirs_oli_pitch_corr;
    double est_tirs_oli_yaw_corr;
    double updated_tirs_oli_roll;
    double updated_tirs_oli_pitch;
    double updated_tirs_oli_yaw;
    int fit_order;
    double outlier_threshold;
} main_db_save_info;

/* Define a structure for the information to be collected and inserted into
   table TIRS_ALIGNMENT_SCA */
struct sca_db_save_info
{
    char geometric_wo_common_id[IAS_CHAR_ID_SIZE];  
    int sensor_chip_assembly_id[NUM_TIRS_SCAS];
    int nbr_of_pnts_used[NUM_TIRS_SCAS];
    double upds_alng_trk_leg_poly_0[NUM_TIRS_SCAS]; /* Updates to Legendre 
        polynomials: original + updates = new */
    double upds_alng_trk_leg_poly_1[NUM_TIRS_SCAS];
    double upds_alng_trk_leg_poly_2[NUM_TIRS_SCAS];
    double upds_alng_trk_leg_poly_3[NUM_TIRS_SCAS];
    double upds_acrss_trk_leg_poly_0[NUM_TIRS_SCAS];
    double upds_acrss_trk_leg_poly_1[NUM_TIRS_SCAS];
    double upds_acrss_trk_leg_poly_2[NUM_TIRS_SCAS];
    double upds_acrss_trk_leg_poly_3[NUM_TIRS_SCAS];
    double new_alng_trk_leg_poly_0[NUM_TIRS_SCAS];
    double new_alng_trk_leg_poly_1[NUM_TIRS_SCAS];
    double new_alng_trk_leg_poly_2[NUM_TIRS_SCAS];
    double new_alng_trk_leg_poly_3[NUM_TIRS_SCAS];
    double new_acrss_trk_leg_poly_0[NUM_TIRS_SCAS];
    double new_acrss_trk_leg_poly_1[NUM_TIRS_SCAS];
    double new_acrss_trk_leg_poly_2[NUM_TIRS_SCAS];
    double new_acrss_trk_leg_poly_3[NUM_TIRS_SCAS];
    double original_alng_trk_leg_poly_0[NUM_TIRS_SCAS];
    double original_alng_trk_leg_poly_1[NUM_TIRS_SCAS];
    double original_alng_trk_leg_poly_2[NUM_TIRS_SCAS];
    double original_alng_trk_leg_poly_3[NUM_TIRS_SCAS];
    double original_acrss_trk_leg_poly_0[NUM_TIRS_SCAS];
    double original_acrss_trk_leg_poly_1[NUM_TIRS_SCAS];
    double original_acrss_trk_leg_poly_2[NUM_TIRS_SCAS];
    double original_acrss_trk_leg_poly_3[NUM_TIRS_SCAS];
    double pre_fit_res_alng_trk_mean[NUM_TIRS_SCAS];
    double pre_fit_res_alng_trk_std_dev[NUM_TIRS_SCAS];
    double pre_fit_res_alng_trk_rmse[NUM_TIRS_SCAS];
    double pre_fit_res_acrss_trk_mean[NUM_TIRS_SCAS];
    double pre_fit_res_acrss_trk_std_dev[NUM_TIRS_SCAS];
    double pre_fit_res_acrss_trk_rmse[NUM_TIRS_SCAS];
    double pst_fit_res_alng_trk_mean[NUM_TIRS_SCAS];
    double pst_fit_res_alng_trk_std_dev[NUM_TIRS_SCAS];
    double pst_fit_res_alng_trk_rmse[NUM_TIRS_SCAS];
    double pst_fit_res_acrss_trk_mean[NUM_TIRS_SCAS];
    double pst_fit_res_acrss_trk_std_dev[NUM_TIRS_SCAS];
    double pst_fit_res_acrss_trk_rmse[NUM_TIRS_SCAS];
} db_save_info;

void prepare_for_database_save 
(
    double sca_coef_x[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS], /* I: Existing X
                                        Legendre coefficients */
    double sca_coef_y[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS], /* I: Existing Y
                                        Legendre coefficients */
    double error_coef_x[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS], /* I: Error X
                                        Legendre coefficients */
    double error_coef_y[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS], /* I: Error Y
                                        Legendre coefficients */
    double new_coef_x[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS], /* I: Updated X
                                        coefficients */
    double new_coef_y[NUM_TIRS_SCAS][IAS_LOS_LEGENDRE_TERMS], /* I: Updated Y
                                        coefficients */
    const POST_FIT_RESIDUAL_STATS_TYPE *residual_stats, /* I: Calculated
                                        residual information */
    const char *work_order_id,    /* I: Work Order ID for report header */
    const char *ref_image_name,   /* I: Name of reference image */
    double confidence_level,      /* I: Confidence level used for outlier
                                        rejection */
    int fit_order,                /* I: Flag that indicates the polynomial
                                        order that was used in the solution */
    int number_records,           /* I: The number of DB records */
    struct main_db_save_info* main_db_save, /* O: Populated main table data */
    struct sca_db_save_info* sca_db_save /* O: Populated SCA table data */
);

int save_to_database
(
    struct main_db_save_info* main_db_save, /* I/O: Main table data */
    struct sca_db_save_info* sca_db_save, /* I/O: SCA table data */
    int number_records            /* I: The number of DB records */
);

#endif
