#!/bin/csh -f

#************************************************************************
# NAME:    test_tirs_alignment
#
# PURPOSE: Run a test of an individual TIRS focal plane alignment test
#          scenario.  This requires several environment variables to be
#          set up by the caller.
#
# RETURNS: 0 if successful, 1 on an error condition
#************************************************************************

# Set up the environment for the test.
set application = "tirs_alignment"
set work_order_id = "I904"
set diff_suffix = "diffs"
set omf_suffix = "omf"
set parameter_file_suffix = "odl"
set test_dir = "$IAS_UNIT_TEST_DATA_DIR/ias_base/gps/focal_plane/${application}"
set setup_phase = "../../setup/focal_plane_setup"
set legendre_phase = "../tirs_alignment_legendre"
set omf_source = "${application}.${omf_suffix}"
set omf = "${work_order_id}.${omf_suffix}"
set parameter_file_source_db = "${application}.${parameter_file_suffix}"
set parameter_file_source_no_db = \
    "${application}_no_db.${parameter_file_suffix}"
set parameter_file = "${work_order_id}.${parameter_file_suffix}"
set setup_phase_output_file = "tirs_align_corr.dat"
set legendre_phase_report_file = "tirs_align.rpt"
set cpf_attitude_group_file = "ATTITUDE_PARAMETERS_20030401_20030630.cpf" 
set cpf_focal_plane_group_file = "TIRS_FOCAL_PLANE_20030401_20030630.cpf" 
set setup_phase_output_file_diffs = \
    "${setup_phase_output_file}.${diff_suffix}"
set legendre_phase_report_file_diffs = \
    "${legendre_phase_report_file}.${diff_suffix}"
set cpf_attitude_group_file_diffs = \
    "${cpf_attitude_group_file}.${diff_suffix}"
set cpf_focal_plane_group_file_diffs = \
    "${cpf_focal_plane_group_file}.${diff_suffix}"
setenv IAS_VERSION_OVERRIDE "Y"

# Ensure the test environment is set.
if ( ! $?IAS_UNIT_TEST_DATA_DIR ) then
    echo "IAS_UNIT_TEST_DATA_DIR isn't set"
    exit 1
endif

# Check whether database testing has been enabled.  If it has, set the
# parameter file to the database-enabled version, otherwise set the
# parameter file to the database-disabled version.  The database
# enabled/disabled versions of the parameter file differ only in the setting
# the STORE_CHARACTERIZATION_IN_DB flag...
if ( $?ENABLE_DB_TESTS ) then
    #use the parameter file with database support
    set parameter_file_source = "${parameter_file_source_db}"

    # Sanity check to make sure the database user ID/password are set in the
    # environment
    if ( ! $?IAS_DB_L8_OLITIRS ) then
        echo "IAS_DB_L8_OLITIRS is not currently set"
        exit 1
    endif
else
    # display message indicating database support is disabled
    echo
    echo `basename $0` ":  Database tests are disabled."
    echo

    # use the parameter file without database support
    set parameter_file_source = "${parameter_file_source_no_db}"
endif

# Copy the OMF file and parameter file to the standard filename.
cp ${omf_source} ${omf} 
cp ${parameter_file_source} ${parameter_file} 

# Update the working copy of the selected OMF file so that cases where the
# IAS_UNIT_TEST_DATA_DIR string is present are replaced with the actual
# unit test directory, as defined in the environment variable of that name.
#
# Parse the environment variable so it's in a backslashed form that sed can
# handle.  In other words, replace all / with \/.  Otherwise the slashes in
# the filename path would be interpreted by sed as instructions rather than
# data. 
set modified_directory=`echo $IAS_UNIT_TEST_DATA_DIR | sed 's@\/@\\\/@g'`

# Pass the OMF and Work order parameter file through sed.  Sed will replace 
# all cases of the string IAS_UNIT_TEST_DATA_DIR with the (now sed-digestible) 
# contents of that environment variable.
cat ${omf} | sed "s/IAS_UNIT_TEST_DATA_DIR/$modified_directory/g" > ${omf}_tmp
mv ${omf}_tmp ${omf}

cat ${parameter_file} | sed "s/IAS_UNIT_TEST_DATA_DIR/$modified_directory/g" \
    > ${parameter_file}_tmp
mv ${parameter_file}_tmp ${parameter_file}

# Run the TIRS Alignment Calibration algorithm Setup Phase using the input
# parameter file.  This will create a machine-readable intermediate report
# file.
echo "Testing TIRS Alignment Setup Phase."
${setup_phase} ${parameter_file}
if ( $status != 0 ) then
    echo "Failed to run $setup_phase."
    exit 1
endif

# Compare the Setup Phase output file to the unit test version.  Any 
# difference will be treated as a failure.
echo "Comparing Setup Phase output file $setup_phase_output_file to " \
    "$test_dir/$setup_phase_output_file."
diff ${test_dir}/${setup_phase_output_file} ${setup_phase_output_file} > \
    ${setup_phase_output_file_diffs}
if ( $status != 0 ) then
    echo "$setup_phase_output_file and $test_dir/$setup_phase_output_file " \
         "files do not match."
    exit 1
endif

# Run the TIRS Alignment Calibration algorithm Legendre Phase using the input 
# parameter file.  This will create a report file, 2 CPF group files, and 
# updated rows in the database. 
echo "Testing TIRS Alignment Legendre Phase."
${legendre_phase} ${parameter_file}
if ( $status != 0 ) then
    echo "Failed to run $legendre_phase."
    exit 1
endif

# Compare the Legendre Phase CPF attitude group file to the unit test version.
# Any difference will be treated as a failure.
echo "Comparing Legendre Phase CPF ATTITUDE_PARAMETERS group file" \
    "$cpf_attitude_group_file to $test_dir/$cpf_attitude_group_file."
diff ${test_dir}/${cpf_attitude_group_file} ${cpf_attitude_group_file} > \
    ${cpf_attitude_group_file_diffs}
if ( $status != 0 ) then
    echo "$cpf_attitude_group_file and $test_dir/$cpf_attitude_group_file " \
        "files do not match."
    exit 1
endif

# Compare the Legendre Phase CPF Legendre coefficent group file to the unit 
# test version.  Any difference will be treated as a failure.
echo "Comparing Legendre Phase CPF TIRS_FOCAL_PLANE group file" \
    "$cpf_focal_plane_group_file to $test_dir/$cpf_focal_plane_group_file."
diff ${test_dir}/${cpf_focal_plane_group_file} \
    ${cpf_focal_plane_group_file} > ${cpf_focal_plane_group_file_diffs}
if ( $status != 0 ) then
    echo "$cpf_focal_plane_group_file and " \
        "$test_dir/$cpf_focal_plane_group_file files do not match."
    exit 1
endif

# Prepare to compare the Legendre Phase report file to the unit test version.
# Exclude the report header, since it has a regression test of its own.
sed -e '/Level 1Gt image/,$!'d -e "s/$modified_directory//g" \
    ${legendre_phase_report_file} > ${legendre_phase_report_file}_tmp
cp ${legendre_phase_report_file} ${legendre_phase_report_file}_sav
mv ${legendre_phase_report_file}_tmp ${legendre_phase_report_file}

# Compare the Legendre Phase report file to the unit test version.  Any
# difference will be treated as a failure.
echo "Comparing Legendre Phase report file $legendre_phase_report_file to" \
    "$test_dir/$legendre_phase_report_file."
diff ${test_dir}/${legendre_phase_report_file} ${legendre_phase_report_file} \
    > ${legendre_phase_report_file_diffs}
if ( $status != 0 ) then
    echo "$legendre_phase_report_file and " \
        "$test_dir/$legendre_phase_report_file files do not match."
    exit 1
endif

# Restore the original report file so we don't have an extra file on disk.
mv ${legendre_phase_report_file}_sav ${legendre_phase_report_file}

# If database support is disabled, exit here with a success status.
if ( ! $?ENABLE_DB_TESTS ) then
    exit 0
endif

# Finish setting up the database information
set db_results = "_db_results.txt"
set main_db_table = "tirs_alignment"
set sca_db_table = "tirs_alignment_sca"
set main_db_table_file = "${main_db_table}${db_results}"
set sca_db_table_file = "${sca_db_table}${db_results}"
set main_db_table_file_diffs = "${main_db_table_file}.${diff_suffix}"
set sca_db_table_file_diffs = "${sca_db_table_file}.${diff_suffix}"
set main_table_sql_script = "main_db_table_script.sql"
set sca_table_sql_script = "sca_db_table_script.sql"
set common_id_sql_script = "geometric_wo_common_id_script.sql"
set common_id_file = "geometric_wo_common_id.txt"

# Get the geometric common ID.
echo "set term off" > $common_id_sql_script
echo "set echo on" >> $common_id_sql_script
echo "set trimout on" >> $common_id_sql_script
echo "set trimspool on" >> $common_id_sql_script
echo "set heading off" >> $common_id_sql_script
echo "set feedback off" >> $common_id_sql_script
echo "set pause off" >> $common_id_sql_script
echo "set pagesize 0" >> $common_id_sql_script
echo "set linesize 500" >> $common_id_sql_script
echo "set space 0" >> $common_id_sql_script
echo "set colsep |" >> $common_id_sql_script
echo "set tab off" >> $common_id_sql_script
echo "set verify off" >> $common_id_sql_script
echo "spool ./${common_id_file}" >> $common_id_sql_script
echo "select work_order_id, geometric_wo_common_id " >> $common_id_sql_script
echo "from geometric_work_order_common " >> $common_id_sql_script
echo "where work_order_id='"$work_order_id"';" >> $common_id_sql_script
echo "spool off" >> $common_id_sql_script
echo "exit" >> $common_id_sql_script

# Query the database.
sqlplus $IAS_DB_L8_OLITIRS @$common_id_sql_script > & /dev/null

# Squeeze out the white space and select the lines beginning with the work
# order ID.
set common_id = `sed 's/  *//g' ./${common_id_file} | grep ^$work_order_id | awk -F'|' '{print $2}'`

# Create a script to query the main TIRS alignment database table.  This is
# done here so that a separate script will not have to be maintained.  It will
# be used then removed at the end.
echo "set term off" > $main_table_sql_script
echo "set echo on" >> $main_table_sql_script
echo "set trimout on" >> $main_table_sql_script
echo "set trimspool on" >> $main_table_sql_script
echo "set heading off" >> $main_table_sql_script
echo "set feedback off" >> $main_table_sql_script
echo "set pause off" >> $main_table_sql_script
echo "set pagesize 0" >> $main_table_sql_script
echo "set linesize 1000" >> $main_table_sql_script
echo "set space 0" >> $main_table_sql_script
echo "set colsep |" >> $main_table_sql_script
echo "set tab off" >> $main_table_sql_script
echo "set verify off" >> $main_table_sql_script
echo "spool ./${main_db_table}_db_results.txt" >> $main_table_sql_script
echo "select reference_image, number_of_scas, " >> $main_table_sql_script
echo "original_tirs_oli_roll, " >> $main_table_sql_script
echo "original_tirs_oli_pitch, " >> $main_table_sql_script
echo "original_tirs_oli_yaw, " >> $main_table_sql_script
echo "est_tirs_oli_roll_corr, " >> $main_table_sql_script
echo "est_tirs_oli_pitch_corr, " >> $main_table_sql_script
echo "est_tirs_oli_yaw_corr, " >> $main_table_sql_script
echo "updated_tirs_oli_roll, " >> $main_table_sql_script
echo "updated_tirs_oli_pitch, " >> $main_table_sql_script
echo "updated_tirs_oli_yaw, " >> $main_table_sql_script
echo "fit_order, outlier_threshold " >> $main_table_sql_script
echo "from ${main_db_table} " >> $main_table_sql_script
echo "where geometric_wo_common_id=$common_id;"  >> $main_table_sql_script
echo "spool off" >> $main_table_sql_script
echo "exit" >> $main_table_sql_script

# Query the database.
sqlplus $IAS_DB_L8_OLITIRS @$main_table_sql_script > & /dev/null

# Squeeze out the white space and select the line with pipes.
sed -e 's/  *//g' -e "s/$modified_directory//g" ./${main_db_table_file} | grep '|' > ./${main_db_table_file}_tmp
sed 's/|/ /g' ./${main_db_table_file}_tmp > ./${main_db_table_file}

# Compare the results.
diff ${test_dir}/${main_db_table_file} . > ${main_db_table_file_diffs}
if ( $status != 0 ) then
    echo "Results in database table ${main_db_table_file} differ from" \
        "expected results"
    exit 1
endif

# Create a script to query the SCA-level TIRS alignment database table.  This
# is done here so that a separate script will not have to be maintained.  It
# will be used then removed at the end.
echo "set term off" > $sca_table_sql_script
echo "set echo on" >> $sca_table_sql_script
echo "set trimout on" >> $sca_table_sql_script
echo "set trimspool on" >> $sca_table_sql_script
echo "set heading off" >> $sca_table_sql_script
echo "set feedback off" >> $sca_table_sql_script
echo "set pause off" >> $sca_table_sql_script
echo "set pagesize 0" >> $sca_table_sql_script
echo "set linesize 1500" >> $sca_table_sql_script
echo "set space 0" >> $sca_table_sql_script
echo "set colsep |" >> $sca_table_sql_script
echo "set tab off" >> $sca_table_sql_script
echo "set verify off" >> $sca_table_sql_script
echo "spool ./${sca_db_table}_db_results.txt" >> $sca_table_sql_script
echo "select sensor_chip_assembly_id, " >> $sca_table_sql_script
echo "nbr_of_pnts_used, upds_alng_trk_leg_poly_0, " >> \
    $sca_table_sql_script
echo "upds_alng_trk_leg_poly_1, " >> $sca_table_sql_script
echo "upds_alng_trk_leg_poly_2, " >> $sca_table_sql_script
echo "upds_alng_trk_leg_poly_3, " >> $sca_table_sql_script
echo "upds_acrss_trk_leg_poly_0, " >> $sca_table_sql_script
echo "upds_acrss_trk_leg_poly_1, " >> $sca_table_sql_script
echo "upds_acrss_trk_leg_poly_2, " >> $sca_table_sql_script
echo "upds_acrss_trk_leg_poly_3, " >> $sca_table_sql_script
echo "new_alng_trk_leg_poly_0, " >> $sca_table_sql_script
echo "new_alng_trk_leg_poly_1, " >> $sca_table_sql_script
echo "new_alng_trk_leg_poly_2, " >> $sca_table_sql_script
echo "new_alng_trk_leg_poly_3, " >> $sca_table_sql_script
echo "new_acrss_trk_leg_poly_0, " >> $sca_table_sql_script
echo "new_acrss_trk_leg_poly_1, " >> $sca_table_sql_script
echo "new_acrss_trk_leg_poly_2, " >> $sca_table_sql_script
echo "new_acrss_trk_leg_poly_3, " >> $sca_table_sql_script
echo "pst_fit_res_alng_trk_mean, " >> $sca_table_sql_script
echo "pst_fit_res_alng_trk_std_dev, " >> $sca_table_sql_script
echo "pst_fit_res_alng_trk_rmse, " >> $sca_table_sql_script
echo "pst_fit_res_acrss_trk_mean, " >> $sca_table_sql_script
echo "pst_fit_res_acrss_trk_std_dev, " >> $sca_table_sql_script
echo "pst_fit_res_acrss_trk_rmse, " >> $sca_table_sql_script
echo "original_alng_trk_leg_poly_0, " >> $sca_table_sql_script
echo "original_alng_trk_leg_poly_1, " >> $sca_table_sql_script
echo "original_alng_trk_leg_poly_2, " >> $sca_table_sql_script
echo "original_alng_trk_leg_poly_3, " >> $sca_table_sql_script
echo "original_acrss_trk_leg_poly_0, " >> $sca_table_sql_script
echo "original_acrss_trk_leg_poly_1, " >> $sca_table_sql_script
echo "original_acrss_trk_leg_poly_2, " >> $sca_table_sql_script
echo "original_acrss_trk_leg_poly_3, " >> $sca_table_sql_script
echo "pre_fit_res_alng_trk_mean, " >> $sca_table_sql_script
echo "pre_fit_res_alng_trk_std_dev, " >> $sca_table_sql_script
echo "pre_fit_res_alng_trk_rmse, " >> $sca_table_sql_script
echo "pre_fit_res_acrss_trk_mean, " >> $sca_table_sql_script
echo "pre_fit_res_acrss_trk_std_dev, " >> $sca_table_sql_script
echo "pre_fit_res_acrss_trk_rmse from ${sca_db_table} " >> \
    $sca_table_sql_script
echo "where geometric_wo_common_id=$common_id;"  >> $sca_table_sql_script
echo "spool off" >> $sca_table_sql_script
echo "exit" >> $sca_table_sql_script

# Query the database.
sqlplus $IAS_DB_L8_OLITIRS @$sca_table_sql_script > & /dev/null

# Squeeze out the white space and select the line with pipes.
sed 's/  *//g' ./${sca_db_table_file} | grep '|' > ./${sca_db_table_file}_tmp
sed 's/|/ /g' ./${sca_db_table_file}_tmp > ./${sca_db_table_file}

# Compare the results.
diff ${test_dir}/${sca_db_table_file} . > ${sca_db_table_file_diffs}
if ( $status != 0 ) then
    echo "Results in database table ${sca_db_table_file} differ from" \
        "expected results"
    exit 1
endif

# Clean up some files.
rm ./${main_db_table_file} ./${main_db_table_file}_tmp \
    ./${sca_db_table_file} ./${sca_db_table_file}_tmp ./${common_id_file} \
    ./${main_table_sql_script} ./${sca_table_sql_script} \
    ./${common_id_sql_script}

# Exit with success
exit 0
