/******************************************************************************
NAME:           update_alignment_parameters

PURPOSE:
update_alignment_parameters computes new TIRS_To_OLI and Attitude_To_TIRS
alignment matrices and creates a new ATTITUDE_PARAMETERS group of the CPF file.

RETURNS:
Type = int
Value       Description
-----       -----------
ERROR       Error writing the CPF file
SUCCESS     CPF successfully written
******************************************************************************/
#include <math.h>
#include "update_alignment_parameters.h" /* Prototype for this function */
#include "ias_logging.h"      /* IAS logging library */
#include "ias_geo.h"          /* IAS geometric library */

int update_alignment_parameters
(
    const double talign[3],   /* I: TIRS-to-OLI alignment corrections */
    double acs2tirs[3][3],    /* I: ACS to TIRS rotation matrix */
    double acs2oli[3][3],     /* I: ACS to OLI rotation matrix */
    double tirs2oli_new[3][3], /* O: Updated TIRS to OLI rotation matrix */
    struct main_db_save_info* main_db_save /* O: top-level data for DB */
)
{
    double delta_align[3][3]; /* Alignment correction matrix */
    double tirs2acs[3][3];    /* TIRS-to-Attitude matrix */
    double tirs2oli_old[3][3]; /* Original TIRS-to-OLI matrix */
    double rpy_old[3];        /* Original TIRS-to-OLI alignment angles */
    double rpy_new[3];        /* New TIRS-to-OLI alignment angles */
    IAS_VECTOR state_vector; /* x, y, z here are roll, pitch, and yaw */

    /* Put the alignment corrections in a vector */
    state_vector.x = talign[0]; /* Roll */
    state_vector.y = talign[1]; /* Pitch */
    state_vector.z = talign[2]; /* Yaw */

    /* Construct the alignment correction matrix */
    ias_geo_convert_rpy_to_matrix(&state_vector, delta_align);

    /* Calculate the old TIRS-to-OLI matrix */
    if (ias_math_invert_matrix(*acs2tirs, *tirs2acs, 3) != SUCCESS)
    {
        IAS_LOG_ERROR("Inverting ACS-to-TIRS matrix");
        return ERROR;
    }
    ias_math_multiply_3x3_matrix(acs2oli, tirs2acs, tirs2oli_old);

    /* Calculate the new TIRS-To-OLI matrix */
    ias_math_multiply_3x3_matrix(tirs2oli_old, delta_align, tirs2oli_new);

    /* Check for divide by 0 */
    if ((tirs2oli_old[2][2] == 0) || (tirs2oli_old[0][0] == 0) || 
        (tirs2oli_new[2][2] == 0) || (tirs2oli_new[0][0] == 0))
    {
        IAS_LOG_ERROR("A TIRS to OLI matrix value that is used in a "
            "denominator is 0, causing a divide by 0");
        return ERROR;
    }

    /* Back out the old and new net TIRS-to-OLI alignment angles */
    rpy_old[0] = atan(-tirs2oli_old[2][1] / tirs2oli_old[2][2]);
    rpy_old[1] = asin(tirs2oli_old[2][0]);
    rpy_old[2] = atan(-tirs2oli_old[1][0] / tirs2oli_old[0][0]);
    rpy_new[0] = atan(-tirs2oli_new[2][1] / tirs2oli_new[2][2]);
    rpy_new[1] = asin(tirs2oli_new[2][0]);
    rpy_new[2] = atan(-tirs2oli_new[1][0] / tirs2oli_new[0][0]);
    main_db_save->original_tirs_oli_roll = rpy_old[0];
    main_db_save->original_tirs_oli_pitch = rpy_old[1];
    main_db_save->original_tirs_oli_yaw = rpy_old[2];
    main_db_save->est_tirs_oli_roll_corr = talign[0];
    main_db_save->est_tirs_oli_pitch_corr = talign[1];
    main_db_save->est_tirs_oli_yaw_corr = talign[2];
    main_db_save->updated_tirs_oli_roll = rpy_new[0];
    main_db_save->updated_tirs_oli_pitch = rpy_new[1];
    main_db_save->updated_tirs_oli_yaw = rpy_new[2];

    return SUCCESS;
}
