/******************************************************************************
NAME:           write_alignment_matrix

PURPOSE:
write_alignment_matrix writes the new alignment matrix 
(ACS_to_TIRS_Rotation_Matrix in the CPF) to a file.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
ERROR           Error writing alignment matrix
SUCCESS         Successfully wrote the alignment matrix

NOTES:
This portion of the CPF will be combined with other portions which have the
same effective date.

Only the alignment matrix is written.  The rest of the ATTITUDE_PARAMETERS
group is not written.  The CPF editor can include a portion of a group, so the 
whole group isn't needed.
******************************************************************************/
#include "write_alignment_matrix.h" /* Prototype */
#include "ias_math.h"               /* Math routines */ 
#include "ias_logging.h"            /* IAS logging library */

int write_alignment_matrix
(
    const char *cpf_group_filename, /* I: Output CPF file name */
    double oli2tirs[3][3],          /* I: Starts as new TIRS to OLI alignment 
                                          matrix; transposed to OLI to TIRS */
    double acs2oli[3][3]            /* I: ACS to OLI rotation matrix */
)
{
    FILE *file_ptr;         /* Output CPF group file pointer */
    int status;             /* Return status of functions called */
    double acs2tirs[3][3];  /* New Attitude_To_TIRS matrix */

    /* Open the CPF group ASCII file */
    file_ptr = fopen(cpf_group_filename, "wt");
    if (file_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening the CPF group output file %s",
            cpf_group_filename);
        return ERROR;
     }

    /* Write the ATTITUDE_PARAMETERS CPF group header */ 
    status = fprintf(file_ptr, "GROUP = ATTITUDE_PARAMETERS\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing ATTITUDE_PARAMETERS header");
        fclose(file_ptr);
        return ERROR;
    }

    /* Calculate the new Attitude_To_TIRS matrix from the old Attitude_To_OLI 
       matrix and the new TIRS_To_OLI matrix */
    ias_math_transpose_3x3_matrix(oli2tirs);
    ias_math_multiply_3x3_matrix(oli2tirs, acs2oli, acs2tirs);

    /* Write the new Attitude_To_TIRS matrix */
    status = fprintf(file_ptr, "  ACS_To_TIRS_Rotation_Matrix = "
        "(%13.10lf, %13.10lf, %13.10lf, \n%13.10lf, %13.10lf, %13.10lf, "
        "%13.10lf, \n%13.10lf, %13.10lf)\nEND_GROUP = ATTITUDE_PARAMETERS\n",
        acs2tirs[0][0], acs2tirs[0][1], acs2tirs[0][2], acs2tirs[1][0], 
        acs2tirs[1][1], acs2tirs[1][2], acs2tirs[2][0], acs2tirs[2][1], 
        acs2tirs[2][2]);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing to the CPF Group ASCII file %s",
            cpf_group_filename);
        fclose(file_ptr);
        return ERROR;
    }

    /* Close the CPF group report file */
    if (fclose(file_ptr) != 0)
    {
        IAS_LOG_ERROR("Closing the CPF group report file %s",
            cpf_group_filename);
        return ERROR;
    }

    return SUCCESS;
}
