/******************************************************************************
NAME:           write_sca_parameters_cpf

PURPOSE:
write_sca_parameters_cpf writes a file containing the updated TIRS_FOCAL_PLANE 
or OLI_FOCAL_PLANE group of the CPF.

RETURNS:
Type = int
Value           Description
-----           -----------
ERROR           Error writing the CPF 
SUCCESS         CPF successfully written
******************************************************************************/

#include <stdlib.h>          /* Standard C library */
#include <stdio.h>           /* File I/O prototypes */
#include <string.h>          /* sprintf prototypes */

#include "ias_logging.h"     /* IAS logging library */
#include "focal_plane.h"     /* Prototype for this function */

int write_sca_parameters_cpf
(
    const char *cpf_filename, /* I: CPF name for writing the group */
    double x_coeff_corr[][IAS_LOS_LEGENDRE_TERMS], /* I: X Legendre
                                   coefficient corrections */
    double y_coeff_corr[][IAS_LOS_LEGENDRE_TERMS], /* I: Y Legendre
                                   coefficient corrections */
    IAS_CPF *cpf_ptr,        /* I: Current CPF for the information that hasn't
                                   been updated by this executable */
    IAS_SENSOR_ID sensor     /* I: ID of sensor whose CPF group is written */
)
{
    FILE *file_ptr;          /* CPF file to write the group to */
    int status;              /* Status returned by called routines */
    int sca_index;           /* SCA index corresponding to SCA number */
    int band_index = 0;      /* Band index corresponding to band number */
    int satellite_number_scas[IAS_MAX_NBANDS]; /* Number of SCAs for the 
                                satellite on a per-band basis */
    const struct IAS_CPF_FOCAL_PLANE *cpf_focal_plane;
    int band_numbers[IAS_MAX_NBANDS]; /* List of band numbers */
    int num_normal_bands;    /* Total normal bands of satellite */
    int num_sensor_bands;    /* Total number of bands for the sensor */
    int i;                   /* Loop counter */
    char group_name[20];     /* Name of the group as shown in the ASCII file */

    if (sensor == IAS_OLI)
    {
        strcpy(group_name, "OLI_FOCAL_PLANE");
    }
    else if (sensor == IAS_TIRS)
    {
        strcpy(group_name, "TIRS_FOCAL_PLANE");
    }
    else
    {
        IAS_LOG_ERROR("Invalid sensor type %d", sensor);
        return ERROR;
    }

    /* Get total satellite band count */
    num_normal_bands = ias_sat_attr_get_normal_band_count();
    if (num_normal_bands == ERROR)
    {
        IAS_LOG_ERROR("Getting normal band count using satellite attributes "
            "library");
        return ERROR;
    }

    /* Get band information for the sensor */
    status = ias_sat_attr_get_sensor_band_numbers(sensor, IAS_NORMAL_BAND, 0,
        band_numbers, num_normal_bands, &num_sensor_bands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting sensor's normal band numbers");
        return ERROR;
    }

    /* Find the number of SCAs for each band */
    for (i = 0; i < num_sensor_bands; i++)
    {
        /* Get the band index */
        band_index = ias_sat_attr_convert_band_number_to_index(band_numbers[i]);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Invalid band index for band number %d", 
                band_numbers[i]);
            return ERROR;
        }

        /* Find the number of SCAs for the band */
        satellite_number_scas[band_index] = ias_sat_attr_get_scas_per_band(
            band_numbers[i]); 
            
        if (satellite_number_scas[band_index] == ERROR)
        {
            IAS_LOG_ERROR("Invalid number of SCAs returned for band number %d",
                band_numbers[i]);
            return ERROR;
        }
    }

    /* Get the groups needed from the CPF */
    cpf_focal_plane = ias_cpf_get_focal_plane(cpf_ptr);
    if (!cpf_focal_plane)
    {
        IAS_LOG_ERROR("Getting Focal Plane parameters from CPF");
        return ERROR;
    }

    /* Open the CPF group ASCII file */
    file_ptr = fopen(cpf_filename, "w");
    if (file_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening the CPF group output file %s", cpf_filename);
        return ERROR;
    }

    /* Write the group name */
    status = fprintf(file_ptr, "GROUP = %s\n", group_name);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing to the CPF Group ASCII file %s", cpf_filename);
        fclose(file_ptr);
        return ERROR;
    }

    /* Write the updated along-track (X) Legendre coefficients */
    for (i = 0; i < num_sensor_bands; i++)
    {
        /* Get the band index */
        band_index = ias_sat_attr_convert_band_number_to_index(band_numbers[i]);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Invalid band index for band number %d", 
                band_numbers[i]);
            fclose(file_ptr);
            return ERROR;
        }

        for (sca_index = 0; sca_index < satellite_number_scas[band_index]; 
            sca_index++)
        {
            status = fprintf(file_ptr, " Along_LOS_Legendre_B%02d_SCA%02d"
                " = (%7.8E,%7.8E,\n  %7.8E,%7.8E)\n", band_numbers[i], 
                sca_index + 1, cpf_focal_plane->along_los_legendre[band_index]
                    [sca_index][0] + x_coeff_corr[sca_index][0],
                cpf_focal_plane->along_los_legendre[band_index]
                    [sca_index][1] + x_coeff_corr[sca_index][1],
                cpf_focal_plane->along_los_legendre[band_index]
                    [sca_index][2] + x_coeff_corr[sca_index][2],
                cpf_focal_plane->along_los_legendre[band_index]
                    [sca_index][3] + x_coeff_corr[sca_index][3]);
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing to the CPF Group ASCII file %s", 
                    cpf_filename);
                fclose(file_ptr);
                return ERROR;
            }
        }
    }

    /* Write the updated across-track (Y) Legendre coefficients */
    for (i = 0; i < num_sensor_bands; i++)
    {
        /* Get the band index */
        band_index = ias_sat_attr_convert_band_number_to_index(band_numbers[i]);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Invalid band index for band number %d", 
                band_numbers[i]);
            fclose(file_ptr);
            return ERROR;
        }

        for (sca_index = 0; sca_index < satellite_number_scas[band_index]; 
            sca_index++)
        {
            status = fprintf(file_ptr, " Across_LOS_Legendre_B%02d_SCA%02d"
                " = (%7.8E,%7.8E,\n  %7.8E,%7.8E)\n", band_numbers[i], 
                sca_index + 1, cpf_focal_plane->across_los_legendre[band_index]
                    [sca_index][0] + y_coeff_corr[sca_index][0],
                cpf_focal_plane->across_los_legendre[band_index]
                    [sca_index][1] + y_coeff_corr[sca_index][1],
                cpf_focal_plane->across_los_legendre[band_index]
                    [sca_index][2] + y_coeff_corr[sca_index][2],
                cpf_focal_plane->across_los_legendre[band_index]
                    [sca_index][3] + y_coeff_corr[sca_index][3]);
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing to the CPF Group ASCII file %s", 
                    cpf_filename);
                fclose(file_ptr);
                return ERROR;
            }
        }
    }

    /* Write the end of group tag */
    status = fprintf(file_ptr, "END_GROUP = %s\n", group_name);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing to the CPF Group ASCII file %s", cpf_filename);
        fclose(file_ptr);
        return ERROR;
    }

    /* Close the CPF group report file */
    if (fclose(file_ptr) != 0)
    {
        IAS_LOG_ERROR("Closing the CPF group report file %s", cpf_filename);
        return ERROR;
    }

    return SUCCESS;
}

