/******************************************************************************
NAME: GCPcorrelate (main)

PURPOSE:
Main driver for the GCPcorrelate task.  This application correlates GCPs
located in reference chips to a 1G image.  The results of this task may be
used in precision model updating.

RETURN VALUE:
Type = int
Value        Description
-----        -----------
EXIT_FAILURE    Error in correlation; GCP data file may not have been produced
                and GCP library may not have been updated
EXIT_SUCCESS    Successful correlations

******************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <limits.h>                /* For PATH_MAX */
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_gcp.h"
#include "ias_parm.h"
#include "ias_satellite_attributes.h"
#include "ias_parm_standard_parameters.h"
#include "GCPcorrelate.h"

#define GCPLIB_NAME "GCPLib"

int main(int argc, char *argv[])
{
    GCP_CORR_PARMS parm_data;       /* Parameters read from param file */
    IAS_GCP_RECORD *gcp_lib = NULL; /* Array of GCP library record structures */
    int numgcp = 0;                 /* Number of GCPs that fit the criteria */
    int status;                     /* Function return status */
    int abs_corr_coeff = FALSE;     /* The flag indicating do not take the
                                       absolute value of correlation coeffs */
    /* The following are criteria the correlation must
       meet to consider it a "successful" correlation: */
    float min_corr,                 /* Minimum correlation strength required */
          fill,                     /* DN fill value */
          max_disp;                 /* Maximum allowable displacement */
    CorrFitMethod fit_method;       /* Correlation fit method */
    float threshold;                /* Percent valid pixels */
    char prog_name[PATH_MAX];       /* Name of the program */
    char parm_filename[PATH_MAX];   /* Name of the parameter file */
    int overall_gcp_version;        /* Overall GCP version read from file */

    /* Get the name of the program */
    strcpy(prog_name, "gcpcorrelate");

    /* Initialize the IAS logging library */
    status = ias_log_initialize(prog_name);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing log");
        exit(EXIT_FAILURE);
    }

    /* Check the number of arguements
       We want the program name and the ODL file for args */
    if (argc != 2)
    {
        IAS_LOG_ERROR("Usage: %s ODL_Parameter_File", prog_name);
        exit(EXIT_FAILURE);
    }

    /* Store the name of the parameter file */
    strcpy(parm_filename, argv[1]);

    memset(&parm_data, 0, sizeof(parm_data));

    /* Get parameters and setup information */
    status = get_gcpcorrelate_parameters(parm_filename, &parm_data,
            &min_corr, &max_disp, &fill, &threshold, &fit_method);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting GCPcorrelate parameters");
        exit(EXIT_FAILURE);
    }

    /* Log initiation */
    IAS_LOG_INFO("Initiated");

    /* Status messaging */
    IAS_LOG_INFO("Reading %s", parm_data.gcplib_filename);

    /* Get GCP information from the GCP library using specified criteria */
    status = ias_gcp_read_gcplib_filtered(parm_data.gcplib_filename,
            parm_data.begin_date, parm_data.end_date, parm_data.season,
            parm_data.chip_source, parm_data.chip_type, &gcp_lib, &numgcp,
            &overall_gcp_version);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading GCP library file %s", parm_data.gcplib_filename);
        exit(EXIT_FAILURE);
    }

    /* Status messaging */
    IAS_LOG_INFO("Read %d GCPs", numgcp);

    /* Do the GCP processing (correlation) and update the GCP library */
    status = process_gcp(gcp_lib, numgcp, parm_data.gcp_dir, parm_data.img_name,
            parm_data.srch_win_size, parm_data.gcp_data_filename, IAS_MAX_NSCAS,
            abs_corr_coeff, min_corr, max_disp, fill, threshold,
            parm_data.y_offset, parm_data.x_offset, fit_method,
            parm_data.band_number);
    if (status != SUCCESS)
    {
        free(gcp_lib);
        IAS_LOG_ERROR("Correlating GCPs");
        exit(EXIT_FAILURE);
    }

    /* Write GPC measurements filename to OMF */
    IAS_PARM_DECLARE_TABLE(parms, 1);

    /* Add the GCP measurements filename parameter */
    IAS_PARM_MEASURED_GCP_FILENAME_FROM_PASS(parms, parm_data.gcp_data_filename,
            sizeof(parm_data.gcp_data_filename), 0, parm_data.processing_pass);
    IAS_PARM_SET_COUNT_WRITE(parms, MEASURED_GCP_FILENAME, 1);

    /* Write the GCP measurements filename parameter */
    status = ias_parm_update(parm_data.omf_filename, parms, 1);
    if (status != SUCCESS)
    {
        free(gcp_lib);
        IAS_LOG_ERROR("Writing GCP measurements filename to OMF");
        exit(EXIT_FAILURE);
    }

    /* Finish up */
    free(gcp_lib);

    IAS_LOG_INFO("Successful completion");

    return EXIT_SUCCESS;
}
