/******************************************************************************
NAME: get_gcpcorrelate_parameters

PURPOSE:
Get parameters from the ODL parameter file and the system table

RETURN VALUE:
Type = int
Value        Description
-----        -----------
ERROR        Error retrieving parameters (check ODL file and/or system table)
SUCCESS      Parameters retrieved

******************************************************************************/
#include <stdlib.h>
#include <ctype.h>                         /* for toupper and tolower */
#include <string.h>
#include "ias_logging.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "ias_cpf.h"
#include "GCPcorrelate.h"

#define EARLIEST_YEAR 1960
#define LATEST_YEAR 3000
#define GCPLIB_NAME "GCPLib"

int get_gcpcorrelate_parameters
(
    const char *parm_filename,  /* I: Parameter filename */
    GCP_CORR_PARMS *parm_data,  /* O: Parameters retrieved from parm file*/
    float *min_corr,            /* O: Minimum correlation strength */
    float *max_disp,            /* O: Maximum correlation displacement */
    float *fill,                /* O: DN fill value */
    float *threshold,           /* O: DN fill Percentage */
    CorrFitMethod *fit_method   /* O: Correlation result fit method */
)
{
    char *tmp_sources[IAS_GCP_NUM_CHIP_SOURCES]; /* temporary char ptrs for
                                                    string arrays */
    char *tmp_seasons[IAS_GCP_NUM_SEASONS]; /* Temporary char ptrs for string
                                               arrays */
    int i;                       /* Looping variable */
    int table_size;              /* Parameter table size */
    int status;                  /* Function return status */
    int tmp_int[2];              /* Temporary int to test values read */
    IAS_CPF *cpf;                /* Structure for the cpf */
    int satellite_number;        /* Satellite number for initializing satellite
                                    attributes */
    int satellite_id;            /* Current satellite id */
    const struct IAS_CPF_GCP_CORRELATION *cpf_gcp;
    /* GCP Correlation CPF parameters */

    /* Local buffer is used to translate between float* and parameter I/O
       interface. */
    double local_min_corr;
    double local_max_disp;
    double local_threshold;


    IAS_PARM_DECLARE_TABLE(app_parm_table, 15);

    /* Get processing pass */
    IAS_PARM_PROCESSING_PASS(app_parm_table, &parm_data->processing_pass,
            sizeof(parm_data->processing_pass), 1);

    /* Work order ID */
    IAS_PARM_ADD_STRING(app_parm_table, WORK_ORDER_ID,
            "Work order ID", IAS_PARM_REQUIRED,
            0, NULL, 0, NULL, parm_data->work_order_id,
            sizeof(parm_data->work_order_id), 1);

    /* Size of correlation search window */
    int default_search_window_size[] = {64, 64};
    IAS_PARM_ADD_INT(app_parm_table, SRCH_WIN_SIZE,
            "Size of the correlation search window -- "
            "two integers in line, sample order",
            IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
            1, 1, 256, 2, default_search_window_size,
            parm_data->srch_win_size,
            sizeof(parm_data->srch_win_size), 2);

    /* Name of output GCP data file */
    const char *default_measured_gcp[] = {"measured.gcp"};
    IAS_PARM_ADD_STRING(app_parm_table, MEASURED_GCP,
            "Measured GCP output filename", IAS_PARM_OPTIONAL,
            0, NULL, 1, default_measured_gcp,
            parm_data->gcp_data_filename,
            sizeof(parm_data->gcp_data_filename), 1);

    /* Begin date for GCP selection */
    int default_begin_date[] = {0, 0};
    IAS_PARM_ADD_INT(app_parm_table, BEGIN_DATE,
            "Begin date for GCP selection -- two integers in "
            "month, year order", IAS_PARM_OPTIONAL,
            IAS_PARM_ARRAY,
            0, 0, 0, 2, default_begin_date,
            parm_data->begin_date,
            sizeof(parm_data->begin_date), 2);

    /* End date for GCP selection */
    int default_end_date[] = {0, 0};
    IAS_PARM_ADD_INT(app_parm_table, END_DATE,
            "End date for GCP selection -- two integers in "
            "month, year order", IAS_PARM_OPTIONAL,
            IAS_PARM_ARRAY,
            0, 0, 0, 2, default_end_date,
            parm_data->end_date,
            sizeof(parm_data->end_date), 0);

    /* Season for GCP selection -- allow a max of 4 seasons */
    for (i = 0; i < IAS_GCP_NUM_SEASONS; i++)
    {
        tmp_seasons[i] = NULL;
        strcpy(parm_data->season[i], "");
    }

    const char *valid_seasons[] = {"WINTER", "SPRING", "SUMMER", "FALL"};
    IAS_PARM_ADD_STRING_ARRAY(app_parm_table, SEASONS,
            "Seasons, used to filter GCPs used: WINTER, "
            "SPRING, SUMMER, FALL",
            IAS_PARM_OPTIONAL, 4, valid_seasons, 0, NULL,
            tmp_seasons,
            sizeof(tmp_seasons), 0);

    /* Chip source for GCP selection */
    for (i = 0; i < IAS_GCP_NUM_CHIP_SOURCES; i++)
    {
        tmp_sources[i] = NULL;
        strcpy(parm_data->chip_source[i], "");
    }

    const char *valid_sources[] = {"DOQ", "GLS", "TM6", "GRI"};
    IAS_PARM_ADD_STRING_ARRAY(app_parm_table, GCP_SOURCE,
            "Array of chip sources to use: DOQ, GLS, TM6 or GRI",
            IAS_PARM_OPTIONAL, 4, valid_sources, 0, NULL,
            tmp_sources,
            sizeof(tmp_sources), 0);

    const char *valid_types[] = {"CONTROL", "VALIDATION", "BOTH"};
    const char *default_type[] = {"CONTROL"};
    IAS_PARM_ADD_STRING(app_parm_table, GCP_TYPE,
            "GCP Type: CONTROL, VALIDATION, or BOTH",
            IAS_PARM_OPTIONAL, 3, valid_types, 1, default_type,
            parm_data->chip_type, sizeof(parm_data->chip_type), 1);

    int default_band_number[] = {6}; /* Should default to SWIR1 band */
    IAS_PARM_ADD_INT(app_parm_table, CORRELATION_BAND,
            "Band number to correlate", IAS_PARM_REQUIRED,
            IAS_PARM_NOT_ARRAY,
            1, 1, IAS_MAX_NBANDS, 1, default_band_number,
            &parm_data->band_number,
            sizeof(parm_data->band_number), 1);

    /* Get user entered offsets if available */
    int default_y_offset[] = {0};/* Default y offset of GCP image */
    IAS_PARM_ADD_INT(app_parm_table, Y_OFFSET,
            "Line/Y offset entered by user", IAS_PARM_OPTIONAL,
            IAS_PARM_NOT_ARRAY,
            1, -512, 512, 1, default_y_offset,
            &parm_data->y_offset,
            sizeof(parm_data->y_offset), 0);

    int default_x_offset[] = {0};/* Default x offset of GCP image */
    IAS_PARM_ADD_INT(app_parm_table, X_OFFSET,
            "Sample/X offset entered by user", IAS_PARM_OPTIONAL,
            IAS_PARM_NOT_ARRAY,
            1, -512, 512, 1, default_x_offset,
            &parm_data->x_offset,
            sizeof(parm_data->x_offset), 0);

    /* Get the minimum correlation value */
    IAS_PARM_MIN_CORR(app_parm_table, &local_min_corr, sizeof(local_min_corr));

    /* Get the maximum displacement value */
    IAS_PARM_MAX_DISP(app_parm_table, &local_max_disp, sizeof(local_max_disp));

    /* Get the valid pixel percentage threshold value */
    IAS_PARM_THRESHOLD(app_parm_table, &local_threshold,
            sizeof(local_threshold));

    table_size = IAS_PARM_GET_TABLE_SIZE(app_parm_table);

    /* Call the routine to provide help for the parameters if requested */
    status = ias_parm_provide_help(parm_filename, app_parm_table, table_size,
            IAS_INPUT_PARAMETERS);
    if (status)
    {
        /* If the status of the provide help is 1 the user asked for help.
           Since the parameter file name contains the help request, copy it
           to the OMF file name so the next call will ask for the same help. */
        if (status == 1)
        {
            strcpy(parm_data->omf_filename, parm_filename);
        }
        else if (status == ERROR) /* Help failed */
        {
            IAS_LOG_ERROR("Providing help for application");
            exit(EXIT_FAILURE);
        }
    }
    else
    {
        status = ias_parm_read(parm_filename, app_parm_table, table_size);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading application parameters from file %s",
                    parm_filename);
            return ERROR;
        }

        /* Build the OMF name from the work order id and work order dir */
        sprintf(parm_data->omf_filename, "%s.omf", parm_data->work_order_id);

        /* Transfer the various local values to the function parameters */
        *min_corr = (float)local_min_corr;
        *max_disp = (float)local_max_disp;
        *threshold = (float)local_threshold;

        /* Season for GCP selection -- allow a max of 4 seasons */
        /* Copy the returned seasons into the structure */
        for (i = 0; i < IAS_GCP_NUM_SEASONS; i++)
        {
            if (tmp_seasons[i] != NULL)
            {
                /* Copy a max of 6 characters and NULL-terminate */
                strncpy(parm_data->season[i],tmp_seasons[i], 6);
                parm_data->season[i][6] = '\0';
            }
        }
        /* Free the temp strings */
        for (i = 0; i < IAS_GCP_NUM_SEASONS; i++)
        {
            free(tmp_seasons[i]);
            tmp_seasons[i] = NULL;
        }

        for (i = 0; i < IAS_GCP_NUM_CHIP_SOURCES; i++)
        {
            if (tmp_sources[i] != NULL)
            {
                /* Copy a maximum of IAS_GCP_SOURCE_SIZE characters and
                   NULL-terminate */
                strncpy(parm_data->chip_source[i], tmp_sources[i],
                        IAS_GCP_SOURCE_SIZE);
                parm_data->chip_source[i][IAS_GCP_SOURCE_SIZE-1] = '\0';
            }
        }
        /* Free the temp strings */
        for (i = 0; i < IAS_GCP_NUM_CHIP_SOURCES; i++)
        {
            free(tmp_sources[i]);
            tmp_sources[i] = NULL;
        }

        /* If BOTH is provided for the chip type, convert it to an empty
           string since that is what the GCP reading routine expects */
        if (strcmp(parm_data->chip_type, "BOTH") == 0)
            strcpy(parm_data->chip_type, "");
    }

    /* Set up and read the OMF parameter table */
    IAS_PARM_DECLARE_TABLE(omf_parm_table, 4);

    IAS_PARM_L1G_FILENAME_FROM_PASS(omf_parm_table, IAS_PARM_REQUIRED,
            parm_data->img_name, sizeof(parm_data->img_name), 0,
            parm_data->processing_pass);

    /* Read CPF name */
    IAS_PARM_CPF(omf_parm_table, IAS_PARM_REQUIRED,
            parm_data->cpf_filename,
            sizeof(parm_data->cpf_filename), 1);

    /* Read GCP directory */
    IAS_PARM_GCP_DIRECTORY(omf_parm_table, IAS_PARM_REQUIRED,
            parm_data->gcp_dir, sizeof(parm_data->gcp_dir));

    /* Add the satellite number parameter */
    IAS_PARM_SATELLITE(omf_parm_table, &satellite_number,
            sizeof(satellite_number), 1);

    table_size = IAS_PARM_GET_TABLE_SIZE(omf_parm_table);

    /* Call the routine to provide help for the parameters if requested */
    status = ias_parm_provide_help(parm_data->omf_filename, omf_parm_table,
            table_size, IAS_OMF_PARAMETERS);
    if (status)
    {
        if (status == 1) /* Help was successful */
        {
            exit(EXIT_SUCCESS);
        }
        else if (status == ERROR) /* Help failed */
        {
            IAS_LOG_ERROR("Providing help for OMF");
            exit(EXIT_FAILURE);
        }
    }

    status = ias_parm_read(parm_data->omf_filename, omf_parm_table, table_size);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading OMF parameters from file %s",
                parm_data->omf_filename);
        return ERROR;
    }

    /* Convert the satellite number to a satellite id */
    satellite_id =
        ias_sat_attr_get_satellite_id_from_satellite_number(satellite_number);
    if (satellite_id == ERROR)
    {
        IAS_LOG_ERROR("Converting the satellite number to a satellite id");
        return ERROR;
    }

    /* Initialize satellite attributes. */
    if (ias_sat_attr_initialize(satellite_id) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the satellite attributes library");
        return ERROR;
    }

    /* Verify size of correlation search window is a power of 2 */
    tmp_int[0] = parm_data->srch_win_size[0];
    tmp_int[1] = parm_data->srch_win_size[1];

    while (tmp_int[0] > 2)
    {
        if ((tmp_int[0] % 2) == 1) break;
        tmp_int[0] /= 2;
    }
    while (tmp_int[1] > 2)
    {
        if ((tmp_int[1] % 2) == 1) break;
        tmp_int[1] /= 2;
    }
    if ((tmp_int[0] % 2) || (tmp_int[1] % 2))
    {
        IAS_LOG_ERROR("Invalid SRCH_WIN_SIZE specified (must be power of 2) "
                "Search window size provided is: %d by %d",
                parm_data->srch_win_size[0], parm_data->srch_win_size[1]);
        return ERROR;
    }

    /* Verify begin date */
    if ((parm_data->begin_date[0] < 0)            /* check month */
            || (parm_data->begin_date[0] > 12)
            || (parm_data->begin_date[1] < 0)     /* check year */
            || ((parm_data->begin_date[1] > 0)
                && (parm_data->begin_date[1] < EARLIEST_YEAR))
            || (parm_data->begin_date[1] > LATEST_YEAR))
    {
        IAS_LOG_ERROR("Invalid BEGIN_DATE specified: %d%d",
                parm_data->begin_date[0], parm_data->begin_date[1]);
        return ERROR;
    }

    /* Verify end date */
    if ((parm_data->end_date[0] < 0)                    /* check month */
            || (parm_data->end_date[0] > 12)
            || (parm_data->end_date[1] < 0)             /* check year */
            || ((parm_data->end_date[1] > 0)
                && (parm_data->end_date[1] < EARLIEST_YEAR))
            || (parm_data->end_date[1] > LATEST_YEAR)
            || ((parm_data->begin_date[0] > 0)          /* begin after end? */
                && (parm_data->begin_date[1] > 0)
                && (parm_data->end_date[0] > 0)
                && (parm_data->end_date[1] > 0)
                && ((parm_data->begin_date[1]*100+parm_data->begin_date[0])
                    > (parm_data->end_date[1]*100+parm_data->end_date[0]))))
    {
        IAS_LOG_ERROR("Invalid END_DATE specified: %d%d",
                parm_data->end_date[0], parm_data->end_date[1]);
        return ERROR;
    }


    /* Build GCPLib file location */
    sprintf(parm_data->gcplib_filename, "%s/%s", parm_data->gcp_dir,
            GCPLIB_NAME);

    IAS_LOG_DEBUG("Opening CPF file %s for reading", parm_data->cpf_filename);

    cpf = ias_cpf_read(parm_data->cpf_filename);
    if (!cpf)
    {
        IAS_LOG_ERROR("Cannot open CPF file %s for reading",
                parm_data->cpf_filename);
        return ERROR;
    }
    cpf_gcp = ias_cpf_get_gcp_correlation(cpf);
    if (!cpf_gcp)
    {
        IAS_LOG_ERROR("Cannot obtain GCP correlation group from CPF file %s",
                parm_data->cpf_filename);
        ias_cpf_free(cpf);
        return ERROR;
    }
    *fit_method = cpf_gcp->corr_fit_method;
    if (IAS_PARM_GET_COUNT_READ(app_parm_table, MAX_DISP) == 0)
    {
        *max_disp = cpf_gcp->max_displacement_offset;
    }
    if (IAS_PARM_GET_COUNT_READ(app_parm_table, MIN_CORR) == 0)
    {
        *min_corr = cpf_gcp->min_corr_strength;
    }
    if (IAS_PARM_GET_COUNT_READ(app_parm_table, THRESHOLD) == 0)
    {
        *threshold = cpf_gcp->fill_threshold;
    }
    *fill = cpf_gcp->corr_fill_value;

    IAS_LOG_DEBUG("Minimum correlate %f",*min_corr);
    IAS_LOG_DEBUG("Maximum Displacement %f",*max_disp);
    IAS_LOG_DEBUG("Fit method %d",*fit_method);
    IAS_LOG_DEBUG("Fill value %f",*fill);
    IAS_LOG_DEBUG("Threshold %f",*threshold);

    ias_cpf_free(cpf);

    return SUCCESS;
}
