/******************************************************************************
NAME: process_gcp

PURPOSE:
Process all the GCPs (extract the chip, the image window, and perform the
correlation for each point, and then write the results to the GCP data file).

******************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "gctp.h"              /* for UTM zone code */
#include "ias_l1g.h"
#include "ias_geo.h"
#include "ias_logging.h"
#include "ias_gcp.h"
#include "ias_math_parallel_corr.h"
#include "ias_types.h"
#include "ias_const.h"
#include "ias_miscellaneous.h"
#include "ias_satellite_attributes.h"
#include "GCPcorrelate.h"

#define DUMP_CHIPS 0             /* Dump chip information to disk. If on, it 
                                    uses subdirectory chips, which may need to 
                                    be hand-created, e.g.: in a make check */
#define OUTSIDE 99               /* code for a GCP outside the image */

/* check_fill prototype */
int check_fill
(
    const float *window,      /* I: Image window */
    int  size,                /* I: Size of the window (nlines * nsamps) */
    float fill_value,         /* I: Fill value to check for */
    float fill_thresh         /* I: Threshold for fill data */
);


/******************************************************************************
NAME: compare_gcp_record_y_coords

PURPOSE:
Routine to compare two GCP records to find which has the higher Y projection
coordinate.  Used to sort GCP records in decreasing Y coordinate order.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
 -1      The 1st record's Y projection is greater than the 2nd
  1      The 1st record's Y projection is less than the 2nd
  0      The 1st record's Y projection is not greater than nor less than the 2nd

******************************************************************************/
int compare_gcp_record_y_coords
(
    const void *r1,
    const void *r2
)
{
    IAS_GCP_RECORD *g1 = (IAS_GCP_RECORD *)r1;
    IAS_GCP_RECORD *g2 = (IAS_GCP_RECORD *)r2;

    if (g1->projection_y > g2->projection_y)
        return -1;
    else if (g1->projection_y < g2->projection_y)
        return 1;
    else
        return 0;
}

/******************************************************************************
NAME: process_gcp

PURPOSE:
Process all the GCPs (extract the chip, the image window, and perform the
correlation for each point, and then write the results to the GCP data file).

RETURN VALUE:
Type = int
Value        Description
-----        -----------
ERROR        Unable to attempt correlation
SUCCESS      Points successfully correlated

NOTES:
1: Correlation routines want things in sample, line order,
   so the fit_offset returned are horizontal (sample) and then vertical
   (line) offsets.  GCP data file contains fit_offset in line, sample order.
2: To calculate the correlated location, know 2 things:
   1. The correlate routines return the offset from the ref window to
      the search window, which is also the offset from the reference point
      to the actual point in the search window.
   2. The integer location of the predicted location roughly corresponds
      to the integer location of the reference location.
      We need to report the predicted search line,sample of the reference
      point and the offset from the predicted point to the correlated point.
      So to get the correlated location, we start with the integer location
      of the predicted point because this corresponds to the integer part
      of the ref point (this is why gcp[num_used_gcp].fit_offset subtracts
      the fractional part of the predicted).  Then we add the fractional
      part of the ref coord because this is really the point we are going
      after.  Then we add the correlation fit_offset because this tells
      how the ref point relates to the location in the search window.

******************************************************************************/
int process_gcp
(
    IAS_GCP_RECORD *gcp_lib,  /* I: List of GCPs */
    int numgcp,               /* I: Number of GCPs */
    const char *gcp_dir,      /* I: Root directory of GCP libraries */
    const char *img_name,     /* I: Name of 1g image */
    int srch_win_size[2],     /* I: Correlation window size (lines,samps) */
    const char *gcp_data_filename,/* I: Name of output data file */
    int nsca,                 /* I: Number of SCAs */
    int abs_corr_coeff,       /* I: Flag for abs of corr coeffs in
                                    ias_math_correlate_grey*/
    float min_corr,           /* I: Minimum correlation strength */
    float max_disp,           /* I: Maximum correlation displacement */
    float fill,               /* I: Fill value to search for */
    float threshold,          /* I: Percentage of valid pixels need for
                                    correlation */
    int y_offset,             /* I: Line/Y user entered offset */
    int x_offset,             /* I: Sample/X user entered offset */
    CorrFitMethod fit_method, /* I: Correlation result fit method */
    int band_number           /* I: Band number to run correlate on */
)
{
    char name[PATH_MAX];      /* Name of 1G SDS */
    IAS_GCP_RESULTS *gcp;     /* GCP data structures (1 per pt) */
    L1GIO *img_ptr;           /* 1G image pointer */
    const char *band_name;    /* Relative band name */
    char proj_name[40];       /* Projection name of target file */
    IAS_DATA_TYPE band_dtype; /* Data type of open band */
    int band_nsca;            /* Number of scas in the band */
    int nl;                   /* Number of lines per band */
    int ns;                   /* Number of samples per band */
    int ref_win_size[2];      /* Size of the ref GCP chip (lines,samps) */
    int corr_ref_size[2];     /* Size of the ref GCP chip (samps,lines) */
    int corr_srch_size[2];    /* Size of the srch window (samps,lines) */
    int curgcp;               /* GCP counter */
    int gcp_inc;              /* Increment for displaying log status */
    int num_used_gcp = 0;     /* Number of GCPs used in correlation */
    int last_used_gcp = 0;    /* Last of GCPs used in correlation */
    int num_corr_gcp = 0;     /* Number of GCPs successfully correlated */
    float nom_off[2];         /* Nominal correlation offset (samps,lines) */
    float *srch_buf;          /* Float search correlation window */
    float *ref_buf;           /* Float reference correlation window */
    double corner_off_y;      /* Corner offsets */
    double corner_off_x;
    int need_buffers = 1;     /* Flag to indicate that buffers are needed for
                                 the reference and search chip */
    int status;               /* Function return status */
    /* Pointer to the parallel correlator data structure */
    IAS_PARALLEL_CORRELATOR_TYPE *correlator_ptr = NULL;
    /* Pointer to the correlation results */
    IAS_CORRELATION_RESULT_TYPE *result_ptr = NULL;
    long chips_in_image_count;/* Count of points inside image */
    int sca;                  /* Counter for SCAs */
    IAS_L1G_BAND_METADATA band_metadata;/* L1G band metadata */
    IAS_L1G_FILE_METADATA file_metadata;/* L1G file metadata */
    L1G_BAND_IO *l1g_band;    /* L1G_BAND_IO pointer */
    FILE *fp;                 /* File pointer */
    /* These variables are added to account for "non-centered" control points*/
    double hline_win_size;    /* Half the chip size (in lines) */
    double hsamp_win_size;    /* Half the chip size (in samples) */
    double center_off_line;   /* Offset from center of GCP point (in lines) */
    double center_off_samp;   /* Offset from center of GCP point (in samples) */
    double pred_center_line;  /* Predicted center line of chip in image */
    double pred_center_samp;  /* Predicted center sample of chip in image */
    float fill_value = fill;  /* Floating point fill value */
    int upper_line;           /* The upper line in the image for the chip */
    int left_sample;          /* The left sample in the image for the chip */
    IAS_GCP_IMAGE_CHIP_IO *chip_io; /* Image chip IO structure */

    /*
     * If no gcp points were provided, create an empty results file and return
     */
    if (numgcp == 0)
    {
        IAS_LOG_WARNING("No GCP points to correlate");

        status = ias_gcp_write_correlation_results(gcp_data_filename, NULL,
                numgcp);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Error creating empty results file.");
            return ERROR;
        }

        return SUCCESS;
    }

    /*
     * Find the biggest chip to allocate enough memory for the
     * srch and ref buffers
     */
    ref_win_size[0] = ref_win_size[1] = 0;
    for (curgcp = 0; curgcp < numgcp; curgcp++)
    {
        if (gcp_lib[curgcp].chip_size_lines > ref_win_size[0])
            ref_win_size[0] = (int)gcp_lib[curgcp].chip_size_lines;
        if (gcp_lib[curgcp].chip_size_samples > ref_win_size[1])
            ref_win_size[1] = (int)gcp_lib[curgcp].chip_size_samples;
    }

    /*
     * Open the 1G image for extraction of chips
     * Specifying a band will return its info in nl, ns, and dtype
     */
    band_name = ias_sat_attr_convert_band_number_to_name(band_number);
    if (band_name == NULL)
    {
        IAS_LOG_ERROR("No name for invalid band number %d", band_number);
        return ERROR;
    }

    img_ptr = ias_l1g_open_image(img_name, IAS_READ);
    if (img_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening Image %s", img_name);
        return ERROR;
    }

    /* Now open the band to get the needed info about it */
    l1g_band = ias_l1g_open_band(img_ptr, band_number, &band_dtype, &band_nsca,
            &nl, &ns);
    if (l1g_band == NULL)
    {
        IAS_LOG_ERROR("Image %s does not contain a pan band", img_name);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    /* If the number of scas in the band don't match the number of scas
       passed in, set nsca to the number of scas in the band */
    IAS_LOG_INFO("Processing %d SCA(s) in image %s", band_nsca, img_name);
    if (band_nsca != nsca)
    {
        nsca = band_nsca;
    }

    /* Allocate space for GCP data (one point per gcp_lib point) */
    gcp = malloc(nsca * numgcp * sizeof(*gcp));
    if (gcp == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for %d GCP data points",numgcp);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }
    memset(gcp, 0, sizeof(*gcp) * nsca * numgcp);

    /*
     * Check the data type
     * Only valid 1G output types are BYTE, I2, UI2, I4, and R4
     */
    if ((band_dtype != IAS_BYTE) && (band_dtype != IAS_I2)
            && (band_dtype != IAS_UI2) && (band_dtype != IAS_I4)
            && (band_dtype != IAS_R4))
    {
        IAS_LOG_ERROR("Image %s has an invalid 1G image data type", img_name);
        free(gcp);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    /* Read the file metadata */
    status = ias_l1g_get_file_metadata(img_ptr, &file_metadata);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading file metadata in image %s", img_name);
        free(gcp);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    /* Read the band metadata */
    status = ias_l1g_get_band_metadata(img_ptr, band_number, &band_metadata);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading band metadata in image %s", img_name);
        free(gcp);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    /* Do some basic checks on the image to make sure we can correctly
     * correlate-- Make sure the pixel size is filled in, and make
     * sure it was framed north-up (corner y's and x's match) so the GCP
     * location prediction works.
     */
    if ((band_metadata.projection_distance_y <= 0.0)
            || (band_metadata.projection_distance_x <= 0.0))
    {
        IAS_LOG_ERROR("Invalid pixel size (Y %f, X %f)",
                band_metadata.projection_distance_y,
                band_metadata.projection_distance_x);
        free(gcp);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    /* Check that corners are within 1/4 pixel */
    corner_off_y = band_metadata.projection_distance_y * 0.25;
    corner_off_x = band_metadata.projection_distance_x * 0.25;
    if ((fabs(band_metadata.upper_left_y - band_metadata.upper_right_y)
                > corner_off_y)
            || (fabs(band_metadata.upper_left_x - band_metadata.lower_left_x)
                > corner_off_x))
    {
        IAS_LOG_ERROR("Image %s is not in a north-up frame", img_name);
        free(gcp);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    /* If projection type is not UTM, issue a warning */
    if (file_metadata.projection_code != UTM)
    {
        IAS_LOG_WARNING("Image %s is not in UTM projection", img_name);
    }

    /* Convert the target file projection code to a string for easy comparison
       to the GCP projection information */
    switch (file_metadata.projection_code)
    {
        case UTM:
            strncpy(proj_name, "UTM", sizeof(proj_name));
            break;
        case PS:
            strncpy(proj_name, "PS", sizeof(proj_name));
            break;
        case SOM:
            strncpy(proj_name, "SOM", sizeof(proj_name));
            break;
        default:
            IAS_LOG_WARNING("Unsupported projection code: %d - no points "
                    "will correlate", file_metadata.projection_code);
            strncpy(proj_name, "UNKNOWN", sizeof(proj_name));
            break;
    }
    proj_name[sizeof(proj_name) - 1] = '\0';

    /* Sort the GCPs by decreasing projection Y coordinate since that will
       likely be roughly the equivalent of sorting them by predicted line.
       The purpose of this is to group the gcps in an order that might help
       improve the IO performance. */
    qsort(gcp_lib, numgcp, sizeof(*gcp_lib), compare_gcp_record_y_coords);

    /* Open the image chip container */
    chip_io = ias_gcp_open_image_chip_container(gcp_dir,
        IAS_GCP_CONTAINER_SEARCH, IAS_READ);
    if (chip_io == NULL)
    {
        IAS_LOG_ERROR("Failed to open the image chip container at %s", gcp_dir);
        free(gcp);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    /* Report which image chip container is being used */
    IAS_LOG_INFO("Using image chips from %s", 
        ias_gcp_get_image_chip_container_name(chip_io));

    /* Allocate memory for the results of the correlation */
    result_ptr = malloc(nsca * numgcp * sizeof(*result_ptr));
    if (result_ptr == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for the correlation results");
        ias_gcp_close_image_chip_container(chip_io);
        free(gcp);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    /* Initialize the parallel correlator */
    correlator_ptr = ias_math_init_parallel_correlator(1, 1, 1,
            1 /* check all flags */,
            ref_win_size[0], ref_win_size[1], srch_win_size[0],
            srch_win_size[1], result_ptr);
    if (correlator_ptr == NULL)
    {
        IAS_LOG_ERROR("Creating the parallel correlator");
        free(result_ptr);
        ias_gcp_close_image_chip_container(chip_io);
        free(gcp);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    /* Initialize the results */
    for (curgcp = 0; curgcp < band_nsca * numgcp; curgcp++)
    {
        result_ptr[curgcp].valid = FALSE;
        result_ptr[curgcp].fit_offset[0] = 0.0;
        result_ptr[curgcp].fit_offset[1] = 0.0;
        result_ptr[curgcp].strength = 0.0;
    }

    /* Process each GCP */
    num_used_gcp = 0;
    num_corr_gcp = 0;
    last_used_gcp = 0;
    for (sca = 0; sca < band_nsca; sca++)
    {
        chips_in_image_count = 0;
        /* Set up the increment to display status every 10 chips */
        gcp_inc = numgcp / 10;
        if (gcp_inc < 1)
            gcp_inc = 1;
        for (curgcp = 0; curgcp < numgcp; curgcp++)
        {
            /* Display status */
            if ((curgcp % gcp_inc) == 0)
            {
                IAS_LOG_INFO("Correlating GCP # %d to SCA %d", curgcp,sca);
            }

            /* Get buffers for the chips if needed */
            if (need_buffers)
            {
                if (ias_math_get_corr_chip_buffers(correlator_ptr, &ref_buf,
                            &srch_buf) != SUCCESS)
                {
                    IAS_LOG_ERROR("Getting buffers for chips");
                    ias_math_close_parallel_correlator(correlator_ptr);
                    free(result_ptr);
                    ias_gcp_close_image_chip_container(chip_io);
                    free(gcp);
                    ias_l1g_close_band(l1g_band);
                    ias_l1g_close_image(img_ptr);
                    return ERROR;
                }
                need_buffers = 0;
            }

            /* Make sure the resolution of the chip matches the resolution of
               the image */
            if ((band_metadata.projection_distance_y
                        != gcp_lib[curgcp].pixel_size_y)
                    || (band_metadata.projection_distance_x
                        != gcp_lib[curgcp].pixel_size_x))
            {
                IAS_LOG_ERROR("GCP %s pixel size (%f, %f) does not match "
                        "image pixel size (%f, %f)",
                        gcp_lib[curgcp].point_id, gcp_lib[curgcp].pixel_size_y,
                        gcp_lib[curgcp].pixel_size_x,
                        band_metadata.projection_distance_y,
                        band_metadata.projection_distance_x);
                ias_math_close_parallel_correlator(correlator_ptr);
                free(result_ptr);
                ias_gcp_close_image_chip_container(chip_io);
                free(gcp);
                ias_l1g_close_band(l1g_band);
                ias_l1g_close_image(img_ptr);
                return ERROR;
            }

            /* Copy information from the library point to the data point */
            strcpy(gcp[num_used_gcp].point_id, gcp_lib[curgcp].point_id);
            gcp[num_used_gcp].reference_line = gcp_lib[curgcp].reference_line;
            gcp[num_used_gcp].reference_sample
                = gcp_lib[curgcp].reference_sample;
            gcp[num_used_gcp].latitude = gcp_lib[curgcp].latitude;
            gcp[num_used_gcp].longitude = gcp_lib[curgcp].longitude;
            gcp[num_used_gcp].elevation = gcp_lib[curgcp].elevation;
            gcp[num_used_gcp].search_sca = sca;
            strcpy(gcp[num_used_gcp].reference_name, gcp_lib[curgcp].chip_name);
            strcpy(gcp[num_used_gcp].search_name, band_name);
            gcp[num_used_gcp].chip_data_type = gcp_lib[curgcp].chip_data_type;

            /* Make sure the GCP and the image are the same projection.  If
               they are not, skip the point. */
            if (strcmp(gcp_lib[curgcp].projection, proj_name) != 0)
            {
                IAS_LOG_WARNING("Skipping GCP id %s since it is not in the "
                        "same projection as the image",
                        gcp_lib[curgcp].point_id);
                continue;
            }

            /* Search band is the enumerated type so precision can just read &
               use it */
            strcpy(gcp[num_used_gcp].chip_source, gcp_lib[curgcp].source);
            gcp[num_used_gcp].search_band = band_number;

            /* Find half the chip size */
            hline_win_size = (double)gcp_lib[curgcp].chip_size_lines / 2.0 - 1;
            hsamp_win_size = (double)gcp_lib[curgcp].chip_size_samples / 2.0
                - 1;
            /* Find offset from center of GCP point */
            center_off_line = (gcp_lib[curgcp].reference_line - hline_win_size);
            center_off_samp = (gcp_lib[curgcp].reference_sample
                    - hsamp_win_size);

            /* Calculate the predicted search line and sample */
            get_search_line_samp(gcp_lib[curgcp].projection_y,
                    gcp_lib[curgcp].projection_x, &band_metadata,
                    &gcp[num_used_gcp].predicted_search_line,
                    &gcp[num_used_gcp].predicted_search_sample);

            /* Calculate location of GCP point in imagery (center of chip
               + offset) */
            pred_center_samp = gcp[num_used_gcp].predicted_search_sample
                - center_off_samp + x_offset;
            pred_center_line = gcp[num_used_gcp].predicted_search_line
                - center_off_line + y_offset;

            /* Extract the window from the 1G image
               the extract_window function takes care of converting the window
               to a buffer of float data */
            status = ias_geo_extract_window(nl, ns, sca, band_dtype,
                    pred_center_line, pred_center_samp,
                    srch_win_size, l1g_band, srch_buf,
                    &upper_line, &left_sample);
            if (status == OUTSIDE)
            {
                IAS_LOG_DEBUG("GCP %s is outside image",
                        gcp_lib[curgcp].point_id);
                continue;
            }

            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Extracting window from image %s that "
                        "covers GCP id %s", img_name, gcp_lib[curgcp].point_id);
                ias_math_close_parallel_correlator(correlator_ptr);
                free(result_ptr);
                ias_gcp_close_image_chip_container(chip_io);
                free(gcp);
                ias_l1g_close_band(l1g_band);
                ias_l1g_close_image(img_ptr);
                return ERROR;
            }

            /* Check the input window for fill data over the specified
               percentage */
            status = check_fill(srch_buf, srch_win_size[0]*srch_win_size[1],
                    fill_value, threshold);

            /* If fill data was over the specified amount, mark it as invalid */
            if (status)
            {
                result_ptr[num_used_gcp].valid = FALSE;
            }
            else
            {
                chips_in_image_count++;

                /* Read in the chip and convert to float */
                ref_win_size[0] = (int)gcp_lib[curgcp].chip_size_lines;
                ref_win_size[1] = (int)gcp_lib[curgcp].chip_size_samples;

                status = ias_gcp_read_image_chip(chip_io, 
                        gcp_lib[curgcp].chip_name, ref_win_size[0],
                        ref_win_size[1], gcp_lib[curgcp].chip_data_type,
                        IAS_GCP_CONVERT_TO_FLOAT, ref_buf);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Reading GCP image %s",
                        gcp_lib[curgcp].chip_name);
                    ias_math_close_parallel_correlator(correlator_ptr);
                    free(result_ptr);
                    ias_gcp_close_image_chip_container(chip_io);
                    free(gcp);
                    ias_l1g_close_band(l1g_band);
                    ias_l1g_close_image(img_ptr);
                    return ERROR;
                }

                if (DUMP_CHIPS)
                {
                    sprintf(name, "chips/rsca%dchip%d.img", sca,curgcp);
                    fp = fopen(name, "wb");
                    if (fp == NULL)
                    {
                        IAS_LOG_ERROR("Could not open file %s", name);
                        ias_math_close_parallel_correlator(correlator_ptr);
                        free(result_ptr);
                        ias_gcp_close_image_chip_container(chip_io);
                        free(gcp);
                        ias_l1g_close_band(l1g_band);
                        ias_l1g_close_image(img_ptr);
                        return ERROR;
                    }
                    fwrite(ref_buf, sizeof(float), (ref_win_size[0]
                                * ref_win_size[1]), fp);
                    fclose(fp);
                    ias_misc_write_envi_header(name, NULL, "chip", 
                            ref_win_size[0], ref_win_size[1], 1, 0.0, 0.0, 0.0,
                            0.0, gcp_lib[curgcp].point_id, IAS_R4);

                    sprintf(name, "chips/ssca%dchip%d.img", sca,curgcp);
                    fp = fopen(name, "wb");
                    if (fp == NULL)
                    {
                        IAS_LOG_ERROR("Could not open file %s", name);
                        ias_math_close_parallel_correlator(correlator_ptr);
                        free(result_ptr);
                        ias_gcp_close_image_chip_container(chip_io);
                        free(gcp);
                        ias_l1g_close_band(l1g_band);
                        ias_l1g_close_image(img_ptr);
                        return ERROR;
                    }
                    fwrite(srch_buf, sizeof(float), (srch_win_size[0]
                                * srch_win_size[1]), fp);
                    fclose(fp);
                    ias_misc_write_envi_header(name, NULL, "chip", 
                            srch_win_size[0], srch_win_size[1], 1, 0.0, 0.0, 
                            0.0, 0.0, "search", IAS_R4);
                }

                /* Do the correlation
                 * Swap the window sizes to sample,line order for correlation */
                corr_srch_size[0] = srch_win_size[1];
                corr_srch_size[1] = srch_win_size[0];
                corr_ref_size[0] = ref_win_size[1];
                corr_ref_size[1] = ref_win_size[0];
                /* samples */
                nom_off[0] = fabs(corr_srch_size[0] - corr_ref_size[0]) / 2.0;
                /* lines */
                nom_off[1] = fabs(corr_srch_size[1] - corr_ref_size[1]) / 2.0;

                status = ias_math_submit_chip_to_corr(correlator_ptr,
                        num_used_gcp, srch_buf, ref_buf, corr_srch_size,
                        corr_ref_size, min_corr, fit_method, max_disp, nom_off,
                        abs_corr_coeff);

                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Submitting a chip to the correlator");
                    ias_math_close_parallel_correlator(correlator_ptr);
                    free(result_ptr);
                    ias_gcp_close_image_chip_container(chip_io);
                    free(gcp);
                    ias_l1g_close_band(l1g_band);
                    ias_l1g_close_image(img_ptr);
                    return ERROR;
                }

                /* Save the data needed to extract the correlation results
                   correctly */
                gcp[num_used_gcp].fit_line_offset = -(pred_center_line
                    - (int)(pred_center_line));
                gcp[num_used_gcp].fit_sample_offset = -(pred_center_samp
                    - (int)(pred_center_samp));

                /* Flag that the next chips will need new buffers */
                need_buffers = 1;
            }

            /* If both windows were successfully extracted (the loop wasn't
             * continued before this point) increment num_used_gcp */
            num_used_gcp++;
        }

        /* Wait for the parallel correlator to finish */
        if (ias_math_parallel_correlator_wait_for_results(correlator_ptr)
                != SUCCESS)
        {
            IAS_LOG_ERROR("Waiting for correlation results");
            ias_math_close_parallel_correlator(correlator_ptr);
            free(result_ptr);
            ias_gcp_close_image_chip_container(chip_io);
            free(gcp);
            ias_l1g_close_band(l1g_band);
            ias_l1g_close_image(img_ptr);
            return ERROR;
        }

        IAS_LOG_INFO("Points inside image %ld, Number used %d, Last used %d",
                chips_in_image_count,num_used_gcp,last_used_gcp);

        /* Extract the correlation results returned by the parallel correlator
         */
        for (curgcp = last_used_gcp; curgcp < num_used_gcp; curgcp++)
        {
            /* Check the returned flag to see if this correlation succeeded */
            if (result_ptr[curgcp].valid)
            {
                /* Correlation succeeded */
                gcp[curgcp].accept_flag = TRUE;
                gcp[curgcp].fit_line_offset
                    += result_ptr[curgcp].fit_offset[1] + y_offset;
                gcp[curgcp].fit_sample_offset
                    += result_ptr[curgcp].fit_offset[0] + x_offset;
                num_corr_gcp++;
            }
            else
            {
                /* CORRELATION FAILURE */
                gcp[curgcp].accept_flag = FALSE;
                gcp[curgcp].fit_line_offset = 0.0;
                gcp[curgcp].fit_sample_offset = 0.0;
            }
            gcp[curgcp].correlation_coefficient = result_ptr[curgcp].strength;
            /* Reference band and sca are 0 regardless of validity */
            gcp[curgcp].reference_band = 0;
            gcp[curgcp].reference_sca = 0;
        }
        last_used_gcp = num_used_gcp;
    } /* SCA counter */

    /* Close the parallel correlator */
    if (ias_math_close_parallel_correlator(correlator_ptr) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing the parallel correlator");
        free(result_ptr);
        ias_gcp_close_image_chip_container(chip_io);
        free(gcp);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    /* Free the memory for the results */
    free(result_ptr);
    result_ptr = NULL;

    /* Close the image chip container */
    if (ias_gcp_close_image_chip_container(chip_io) != SUCCESS)
    {
        IAS_LOG_ERROR("Error closing the image chip container");
        free(gcp);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    /* If none of the GCPs correlated, display a warning */
    if (num_corr_gcp == 0)
    {
        IAS_LOG_WARNING("No points correlated successfully");
    }

    /* Write out all the GCPs */
    IAS_LOG_INFO("%d of %d points correlated", num_corr_gcp, numgcp);
    if (ias_gcp_write_correlation_results(gcp_data_filename, gcp, num_used_gcp)
            != SUCCESS)
    {
        IAS_LOG_ERROR("Error creating correlated results file");
        free(gcp);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    /* For LDCM processing, all GCPs are correlated against each SCA
       making it appear as if 14x the number of GCPs were being used
       for registration.  So, the number of correlation attempts would
       be 14x the number of GCPs available. Therefore, the number of
       used GCPs should be divisable by 14. */
    if (num_used_gcp % nsca != 0)
    {
        IAS_LOG_ERROR("SCAs have differing number of used GCPs");
        free(gcp);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    free(gcp);

    /* Close the 1G image */
    if ((ias_l1g_close_band(l1g_band) != SUCCESS)
            || (ias_l1g_close_image(img_ptr) != SUCCESS))
    {
        IAS_LOG_ERROR("Closing 1G image %s", img_name);
        return ERROR;
    }

    return SUCCESS;
}


/******************************************************************************
NAME: check_fill

PURPOSE:  Check the input window for fill data over the specified percentage.


RETURN VALUE:   Type = int
Value    Description
-----    -----------
  1      Fill data was over the specified amount.
  0      Fill data was not over the specified amount.

NOTES:
This routine is useful to determine if there is too much fill data in a
window.  If too much fill data exists, then the window might not be good
for correlating.  Fill data has a value of 0.0.

******************************************************************************/
int check_fill
(
    const float *window,          /* I: Image window */
    int  size,                    /* I: Size of the window (nlines * nsamps) */
    float fill_value,             /* I: Fill value to check for */
    float fill_thresh             /* I: Threshold for fill data */
)
{
    int i;                /* Loop variable */
    int count = 0;        /* Count of fill pixels */
    float fill;           /* Fraction of fill data  */

    /* The size is <= 0, then don't process. */
    if (size <= 0)
        return 0;

    /* Count the number of fill pixels in the window. */
    for (i = 0; i < size; i++)
    {
        if (window[i] == fill_value)
            count++;
    }

    /* Determine the amount of fill. */
    fill = (float)count / (float)size;

    /* If the fill is > the specified fill_thresh then return 1. */
    if (fill > fill_thresh)
        return 1;

    /* Otherwise return 0. */
    return 0;
}
