/*****************************************************************************
NAME:    geodetic_char.c         

PURPOSE: Program verifies the absolute accuracy of the systematic and
         precision products by analyzing the Ground Control Point (GCP)
         measurement residuals before and after the precision correction
         solution.

NOTES:   The last iteration data is used from the precision input file since
         this reflects the solution.  The first iteration reflects the
         initial state.

RETURNS: EXIT_SUCCESS or EXIT_FAILURE

ALGORITHM REFERENCES:
    OLI Geodetic Accuracy Assessment Algorithm, Algorithm Description 
    Document (ADD).

******************************************************************************
                        Property of the U.S. Government
                              USGS EROS Data Center
*******************************************************************************/
#include <stdlib.h>
#include <string.h>

#include "ias_logging.h"
#include "ias_misc_gcp_residuals.h"
#include "geodetic_char.h"

int main (int argc, char *argv[])
{
    IAS_MISC_GCP_RESIDUAL *pre_gcp_res;  /* pointer to the pre-fit GCP
                                            residual data   */
    IAS_MISC_GCP_RESIDUAL *post_gcp_res; /* pointer to the post-fit GCP
                                            residual data  */
    GCP_RES_STATS prefit_stats; /* pre-fit GCP residual statistics            */
    GCP_RES_STATS postfit_stats;/* post-fit GCP residual statistics           */
    int  num_pre_res;        /* number of pre-fit residual structures read    */
    int  num_post_res;       /* number of post-fit residual structures read   */
    int  num_valid;          /* number of non-outlier GCP residuals           */
    int  status;             /* status of return from function                */
    double *GCP_lat;         /* pointer to GCP residual latitude data         */
    double *GCP_lon;         /* pointer to GCP residual longitude data        */
    double *pre_across_res;  /* pointer to pre-fit across-track resid. data   */
    double *pre_along_res;   /* pointer to pre-fit along-track resid. data    */
    double *post_across_res; /* pointer to post-fit across-track resid. data  */
    double *post_along_res;  /* pointer to post-fit along-track resid. data   */
    DB_RESIDUAL_STATISTICS residual_stats;  /* Statistics to be stored in
                                               database */
    int store_stats_in_db;   /* flag to indicate that residual statistics
                                should be written to the database             */
    char gcp_source[IAS_GCP_SOURCE_SIZE];
    int band_number;         /* the first GCP band num for report header use */

    /* initialize the residual statistics info structure */
    memset(&residual_stats, 0, sizeof(DB_RESIDUAL_STATISTICS));

    /*---------------------------------------------------------------------
      Check for the parameter file name
      ----------------------------------------------------------------------*/
    ias_log_initialize( PROG_NAME );
    if (argc < 2)
    {
        IAS_LOG_ERROR( "Usage: geodetic geodetic_char_Parameter_File" );
        exit(EXIT_FAILURE);
    }

    /*----------------------------------------------------------------------
      Read the parameters from the geodetic char. Parameter File.
      ----------------------------------------------------------------------*/
    status = get_geod_char_parms(argv[1], &store_stats_in_db);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR( "Reading parameters from file: %s", argv[1] );
        exit(EXIT_FAILURE);
    }

    /*----------------------------------------------------------------------
      Exit with SUCCESS if correct_los_model failed to produce a precision model
      ----------------------------------------------------------------------*/
    if (get_precision_succeeded_flag() == 0)
    {
        IAS_LOG_INFO( "A precision model was not created.  Nothing for "
                      "geodetic to do." );
        exit(EXIT_SUCCESS);
    }

    /*  Program entry message */
    IAS_LOG_INFO( "Geodetic Accuracy Assessment initiated" );

    /*----------------------------------------------------------------------
      Read the pre- and then post-fit GCP measurement residuals from the 
      precision correction solution residual file.
      ----------------------------------------------------------------------*/
    status = ias_misc_read_gcp_residuals(get_sol_res_name(),
        IAS_GCP_RESIDUALS_FIRST_ITER, &pre_gcp_res, &num_pre_res);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR( "Reading pre_fit residuals" );
        exit(EXIT_FAILURE);
    }

    status = ias_misc_read_gcp_residuals(get_sol_res_name(),
        IAS_GCP_RESIDUALS_LAST_ITER, &post_gcp_res, &num_post_res);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR( "Reading post_fit residuals" );
        free(pre_gcp_res);
        exit(EXIT_FAILURE);
    }

    /* Save the GCP source */
    strcpy(gcp_source, post_gcp_res->gcp_source);

    /*----------------------------------------------------------------------
      Remove outliers and move the valid data to buffers.   
      ----------------------------------------------------------------------*/
    residual_stats.gcp_source[0] = '\0';
    status = remove_outliers(pre_gcp_res, post_gcp_res, num_pre_res,
            num_post_res, &num_valid, &GCP_lat, &GCP_lon, 
            &pre_across_res, &pre_along_res, &post_across_res,
            &post_along_res, &residual_stats);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR( "During removal of the outliers" );
        free(pre_gcp_res);
        free(post_gcp_res);
        exit(EXIT_FAILURE);
    }

    /*----------------------------------------------------------------------
      Free memory which was allocated in function ias_misc get_residual
      ----------------------------------------------------------------------*/ 
    /* First save the band from the first GCP */
    band_number = pre_gcp_res[0].band_number;

    free(pre_gcp_res);
    pre_gcp_res = NULL;
    free(post_gcp_res);
    post_gcp_res = NULL;

    /*----------------------------------------------------------------------
      Calculate mean, RMSE, standard dev., and the correlation coefficient
      for the pre-fit and post-fit residuals.
      ----------------------------------------------------------------------*/
    status = ias_geo_analyze_gcp_records(GCP_lat, GCP_lon, pre_along_res, 
             pre_across_res, num_valid, &prefit_stats); 
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR( "During calculation of pre-fit "
                       "GCP residual statistics" );
        free(GCP_lat);
        free(GCP_lon);
        free(pre_across_res);
        free(pre_along_res);
        free(post_across_res);
        free(post_along_res);
        exit(EXIT_FAILURE);
    } 

    status = ias_geo_analyze_gcp_records(GCP_lat, GCP_lon, post_along_res, 
             post_across_res, num_valid, &postfit_stats);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR( "During calculation of post-fit "
                       "GCP residual statistics" );
        free(GCP_lat);
        free(GCP_lon);
        free(pre_across_res);
        free(pre_along_res);
        free(post_across_res);
        free(post_along_res);
        exit(EXIT_FAILURE);
    } 
    /*----------------------------------------------------------------------
      Free memory which was allocated in function remove_outliers.          
      ----------------------------------------------------------------------*/
    free(GCP_lat);
    GCP_lat = NULL;
    free(GCP_lon);
    GCP_lon = NULL;
    free(pre_across_res);
    pre_across_res = NULL;
    free(pre_along_res);
    pre_along_res = NULL;
    free(post_across_res);
    post_across_res = NULL;
    free(post_along_res);
    post_along_res = NULL;

    /*----------------------------------------------------------------------
      Output residual statistics to output file
      ----------------------------------------------------------------------*/
    status = output_resid_stats(&prefit_stats, &postfit_stats, gcp_source,
            band_number);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR( "During output of residual statistics" );
        exit(EXIT_FAILURE);
    }

    /* Write the data to the residual statistics structure */
    residual_stats.nbr_of_vld_pnts = postfit_stats.num_points;
    residual_stats.mean_lat_of_gcps = postfit_stats.mean_lat;
    residual_stats.mean_lon_of_gcps = postfit_stats.mean_lon;
    residual_stats.pst_fit_mean_alng_trk = postfit_stats.mean_along_res;
    residual_stats.pst_fit_mean_acrss_trk = postfit_stats.mean_across_res;
    residual_stats.pst_fit_std_dev_alng_trk = postfit_stats.sd_along_res;
    residual_stats.pst_fit_std_dev_acrss_trk = postfit_stats.sd_across_res;
    residual_stats.pst_fit_rmse_alng_trk = postfit_stats.rmse_along_res;
    residual_stats.pst_fit_rmse_acrss_trk = postfit_stats.rmse_across_res;
    residual_stats.pst_fit_corr_coeff = postfit_stats.corr_coeff;
    residual_stats.pre_fit_mean_alng_trk = prefit_stats.mean_along_res;
    residual_stats.pre_fit_mean_acrss_trk = prefit_stats.mean_across_res;
    residual_stats.pre_fit_std_dev_alng_trk = prefit_stats.sd_along_res;
    residual_stats.pre_fit_std_dev_acrss_trk = prefit_stats.sd_across_res;
    residual_stats.pre_fit_rmse_alng_trk = prefit_stats.rmse_along_res;
    residual_stats.pre_fit_rmse_acrss_trk = prefit_stats.rmse_across_res;
    residual_stats.pre_fit_corr_coeff = prefit_stats.corr_coeff;

    /* Store the data in the database if requested */
    if (store_stats_in_db)
    {
        IAS_LOG_INFO( "Writing data to database" );

        if (write_residual_statistics_to_db(&residual_stats) != SUCCESS)
        {
            IAS_LOG_ERROR( "Writing data to database" );
            exit(EXIT_FAILURE);
        }
    }

    /*  Program successful exit message */
    IAS_LOG_INFO( "Successful Completion" );

    return EXIT_SUCCESS;
}
