/******************************************************************************
NAME:    get_geod_char_parms.c

PURPOSE: Gets the input parameters from the parameter file and OMF


RETURNS: SUCCESS or ERROR
******************************************************************************
                        Property of the U.S. Government
                            USGS EROS Data Center
*******************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <float.h>

#include "ias_types.h"       /* IAS_READ                       */
#include "ias_const.h"       /* SUCCESS/FAILURE definitions    */
#include "ias_logging.h"     /* ias_logging prototype          */
#include "geodetic_char.h"   /* get_geod_char_parms prototype  */
#include "ias_l1g.h"
#include "ias_satellite_attributes.h"
#include "ias_parm.h"        /* parameter library              */
#include "ias_parm_standard_parameters.h" /* standard fields   */


/* These variables are returned in the accessor functions */
static char omf_file_name[PATH_MAX];     /* name of the OMF file           */
static char sol_res_name[PATH_MAX];      /* precision corrected solution 
                                            residual filename              */ 
static char res_record_name[PATH_MAX];   /* precision corrected residual 
                                            record filename                */
static char l0r_image_name[PATH_MAX];    /* The L0R image filename */
static char l1g_image_name[PATH_MAX];    /* the level 1G image filename    */
static char work_order_id[IAS_WORKORDER_ID_SIZE]; /* Work order ID     */
static int processing_pass;
static int wrs_path = -1;
static int wrs_row = -1;
static double roll_angle = 0;         /* roll angle for off-nadir collections */
static char spacecraft[IAS_L1G_SPACECRAFT_SIZE];
static char collection_type[IAS_COLLECT_TYPE_SIZE];
                                      /* typically EARTH, LUNAR, or STELLAR */
static char capture_date[IAS_L1G_CAPTURE_DATE_SIZE];
                                      /* MM/DD/YYYY format */
static int precision_succeeded_flag;  /* Flag indicating precision solution */


int get_geod_char_parms
(
    const char *ias_parm_file, /* I: Name of the parameter file */
    int *store_in_database     /* O: flag for database storage:
                                     1 = yes, 0 = no */
)
{
    int status;            /* status of return from function             */
    int number_of_parameters; /* Number of parameters to read from
                                 parameter files */
    const char *res_record_name_default[] = {"geodetic_characterization.rpt"};
    int local_store_in_database_flag; /* database storage flag
                                         1 = yes 
                                         0 = no                         */
    int satellite_num;
    IAS_SATELLITE_ID satellite_id;

    /* Define the parameter table */
    IAS_PARM_DECLARE_TABLE(parms, 4);

    IAS_PARM_WORK_ORDER_ID(parms, work_order_id, sizeof(work_order_id), 1);

    IAS_PARM_PROCESSING_PASS(parms, &processing_pass, sizeof(processing_pass),
            0);

    IAS_PARM_ADD_STRING(parms, REPORT_FILENAME, "Name of output report file",
            IAS_PARM_OPTIONAL, 0, NULL, 1, res_record_name_default,
            res_record_name, sizeof(res_record_name), 0);

    IAS_PARM_STORE_CHARACTERIZATION_IN_DB(parms, IAS_PARM_OPTIONAL,
            &local_store_in_database_flag, sizeof(local_store_in_database_flag),
            0);

    /* See if the user wanted help or a template */
    status = ias_parm_provide_help(ias_parm_file, parms,
            IAS_PARM_GET_TABLE_SIZE(parms), IAS_INPUT_PARAMETERS);
    if (status == ERROR)
        return ERROR;
    else if (status == 1)
    {
        /* Copy the input parameter to the OMF file name */
        strcpy(omf_file_name, ias_parm_file);
    }
    else
    {
        /* Read the parameter file */
        number_of_parameters = IAS_PARM_GET_TABLE_SIZE(parms);
        status = ias_parm_read(ias_parm_file, parms, number_of_parameters);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading parameters from: %s", ias_parm_file);
            return ERROR;
        }

        /* Build the OMF file name */
        sprintf(omf_file_name, "%s.omf", work_order_id);
    }

    /* Define the OMF table */
    IAS_PARM_DECLARE_TABLE(omf_parms, 5);

    IAS_PARM_SATELLITE(omf_parms, &satellite_num, sizeof(satellite_num), 0);

    IAS_PARM_L0R_FILENAME(omf_parms, IAS_PARM_OPTIONAL, l0r_image_name,
            sizeof(l0r_image_name), 0);

    IAS_PARM_L1G_FILENAME_FROM_PASS(omf_parms, IAS_PARM_REQUIRED,
            l1g_image_name, sizeof(l1g_image_name), 0, processing_pass);

    IAS_PARM_PRECISION_RESIDUALS_FILENAME_FROM_PASS(omf_parms,
            sol_res_name, sizeof(sol_res_name), 0, processing_pass);

    IAS_PARM_PRECISION_SUCCEEDED_FLAG_FROM_PASS(omf_parms, 
            &precision_succeeded_flag, sizeof(precision_succeeded_flag),
            0, processing_pass);

    /* Again check for help with the OMF file */
    status = ias_parm_provide_help(omf_file_name, omf_parms,
            IAS_PARM_GET_TABLE_SIZE(omf_parms), IAS_OMF_PARAMETERS);
    if (status == ERROR)
        return ERROR;
    else if (status == 1)
    {
        exit(EXIT_SUCCESS);
    }

    /* Read the OMF file */
    number_of_parameters = IAS_PARM_GET_TABLE_SIZE(omf_parms);
    status = ias_parm_read(omf_file_name, omf_parms, number_of_parameters);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from: %s", omf_file_name);
        return ERROR;
    }

    /* Initialize the satellite attributes */
    satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(
            satellite_num);
    status = ias_sat_attr_initialize(satellite_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the satellite attributes library");
        return ERROR;
    }

    /*-----------------------------------------------------------------------
      If the L1G was specified get the information out of the file.
      -----------------------------------------------------------------------*/
    number_of_parameters = IAS_PARM_GET_COUNT_READ(omf_parms, L1G_FILENAME);
    if (number_of_parameters == 0)
    {
        /* Since not falling back to the L0R, if the L1G is not available we
           will lack required output metadata and we must fail. */
        IAS_LOG_ERROR("The L1G Image filename was not found in the OMF file");
        return ERROR;
    }
    else
    {
        L1GIO *l1g_image;        /* L1G image pointer */
        IAS_L1G_FILE_METADATA file_metadata;
        /*---------------------------------------------------------------------
          Read the L1G file in order to get WRS row and path values.
          -------------------------------------------------------------------*/
        l1g_image = ias_l1g_open_image(get_l1g_image_name(), IAS_READ);
        if (l1g_image == NULL)
        {
            IAS_LOG_ERROR("Image %s cannot be opened", get_l1g_image_name()); 
            return ERROR;
        }

        if (ias_l1g_get_file_metadata(l1g_image, &file_metadata) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading metadata");
            ias_l1g_close_image(l1g_image);
            return ERROR;
        }
        ias_l1g_close_image(l1g_image);
        l1g_image = NULL;

        /* Extract data needed for reporting */
        wrs_path = file_metadata.wrs_path;
        wrs_row = file_metadata.wrs_row;
        roll_angle = file_metadata.roll_angle;
        strcpy(spacecraft, file_metadata.spacecraft);
        strcpy(collection_type, file_metadata.collection_type);
        strcpy(capture_date, file_metadata.capture_date);
    }

    /* return the flag */
    *store_in_database = local_store_in_database_flag;

    return SUCCESS;
}


/***********************************************************************
    Functions used to access the ODL parameter file values
***********************************************************************/
const char *get_sol_res_name(void)
{
    return sol_res_name;
}

const char *get_res_record_name(void)
{
    return res_record_name;
}

const char *get_l0r_image_name(void)
{
    return l0r_image_name;
}

const char *get_l1g_image_name(void)
{
    return l1g_image_name;
}

const char *get_omf_file_name(void)
{
    return omf_file_name;
}

const char *get_work_order_id(void)
{
    return work_order_id;
}

int get_wrs_path(void)
{
    return wrs_path;
}

int get_wrs_row(void)
{
    return wrs_row;
}

double get_roll_angle(void)
{
    return roll_angle;
}

const char *get_capture_date(void)
{
    return capture_date;
}

const char *get_collection_type(void)
{
    return collection_type;
}

const char *get_spacecraft(void)
{
    return spacecraft;
}

int get_precision_succeeded_flag(void)
{
    return precision_succeeded_flag;
}
