/*****************************************************************************
NAME:    remove_outliers.c

PURPOSE: Function removes outlier data from the GCP residual information 
         and places the data to be used in buffers to be accessed by 
         later routines.

RETURNS: SUCCESS or ERROR

NOTES:   Memory allocated within this function must be freed by the
         calling routine.

******************************************************************************
                        Property of the U.S. Government
                            USGS EROS Data Center
*******************************************************************************/
#include <stdlib.h>             
#include <string.h>             

#include "ias_logging.h"        /* Logging Library                */
#include "ias_const.h"          /* ERROR/SUCCESS defines          */
#include "geodetic_char.h"      

/* flag value used for valid GCP's */
#define NOT_OUTLIER             1

int remove_outliers 
(
    const IAS_MISC_GCP_RESIDUAL *pre_gcp_res,  /* I: ptr to pre-fit GCP
                                                     residual data  */
    const IAS_MISC_GCP_RESIDUAL *post_gcp_res, /* I: ptr to post-fit GCP
                                                     residual data */

    int  num_pre_res,               /* I: number of pre-fit residual 
                                        structures                            */
    int  num_post_res,              /* I: number of post-fit residual 
                                          structures                          */
    int  *num_valid,                /* I: ptr to number valid GCP residuals   */

    double **GCP_lat,               /* O: GCP residual latitude data          */
    double **GCP_lon,               /* O: GCP residual longitude data         */
    double **pre_across_res,        /* O: pre-fit across-track residual data  */
    double **pre_along_res,         /* O: pre-fit along-track residual data   */
    double **post_across_res,       /* O: post-fit across-track residual data */
    double **post_along_res,        /* O: post-fit along-track residual 
                                          data                                */
    DB_RESIDUAL_STATISTICS *residual_stats/* O: Needed to set the gcp_source  */
)
{
    int  i;            /* loop variable                                       */
    int  index;        /* index for the data buffers                          */
    int  num_differ=0; /* number of differing pre- and post-fit outlier flags */

    /*----------------------------------------------------------------------
      Make sure there are the same number of pre- and post-fit structures
      and if the outlier flags differ, report the number that differ.
      ----------------------------------------------------------------------*/
    if (num_pre_res != num_post_res)
    {
        IAS_LOG_ERROR("Number of pre- and post-fit residuals differ");
        return ERROR;
    }

    for (*num_valid = 0, i = 0; i < num_post_res; i++)
    {
        if (post_gcp_res[i].outlier_flag == NOT_OUTLIER)
            (*num_valid)++;
        if (pre_gcp_res[i].outlier_flag != post_gcp_res[i].outlier_flag)
            num_differ++;
    }

    if (num_differ > 0)
    {
        IAS_LOG_WARNING( "%d of the pre- and post-fit outlier "
            "flags differ, will use post-fit value", num_differ);
    }

    /*----------------------------------------------------------------------
      Allocate space for the data to be used to compute the statistics for
      the GCP residuals.
      ----------------------------------------------------------------------*/
    if ((*GCP_lat = calloc(*num_valid, sizeof(double))) == NULL)
    {
        IAS_LOG_ERROR("Allocating GCP latitude memory");
        return ERROR;
    }
    if ((*GCP_lon = calloc(*num_valid, sizeof(double))) == NULL)
    {
        free(*GCP_lat);
        *GCP_lat = NULL;
        IAS_LOG_ERROR("Allocating GCP longitude memory");
        return ERROR;
    }
    if ((*pre_across_res = calloc(*num_valid, sizeof(double))) == NULL)
    {
        free(*GCP_lat);
        free(*GCP_lon);
        *GCP_lat=*GCP_lon = NULL;
        IAS_LOG_ERROR("Allocating pre-across residuals memory");
        return ERROR;
    }
    if ((*pre_along_res = calloc(*num_valid, sizeof(double))) == NULL)
    {
        free(*GCP_lat);
        free(*GCP_lon);
        free(*pre_across_res);
        *GCP_lat=*GCP_lon=*pre_across_res = NULL;
        IAS_LOG_ERROR("Allocating pre-along residuals memory");
        return ERROR;
    }
    if ((*post_across_res = calloc(*num_valid, sizeof(double))) == NULL)
    {
        free(*GCP_lat);
        free(*GCP_lon);
        free(*pre_across_res);
        free(*pre_along_res);
        *GCP_lat=*GCP_lon=*pre_across_res=*pre_along_res = NULL;
        IAS_LOG_ERROR("Allocating post-across residuals memory memory");
        return ERROR;
    }
    if ((*post_along_res = calloc(*num_valid, sizeof(double))) == NULL)
    {
        free(*GCP_lat);
        free(*GCP_lon);
        free(*pre_across_res);
        free(*pre_along_res);
        free(*post_across_res);
        *GCP_lat=*GCP_lon=*pre_across_res=*pre_along_res=*post_along_res = 
            NULL;
        IAS_LOG_ERROR("Allocating post-along residuals memory memory");
        return ERROR;
    }
    /*----------------------------------------------------------------------
      Fill the buffers with the valid data (not outliers).  We can check
      only the post-fit GCP residual outlier flag since we know the pre-fit
      is the same.
      ----------------------------------------------------------------------*/
    for (index = 0, i = 0; i < num_post_res; i++)
    {
        if (post_gcp_res[i].outlier_flag == NOT_OUTLIER)
        {
            (*GCP_lat)[index] = post_gcp_res[i].latitude;
            (*GCP_lon)[index] = post_gcp_res[i].longitude;
            (*pre_across_res)[index] = pre_gcp_res[i].across_track_residual;
            (*pre_along_res)[index] = pre_gcp_res[i].along_track_residual;
            (*post_across_res)[index] = post_gcp_res[i].across_track_residual;
            (*post_along_res)[index] = post_gcp_res[i].along_track_residual; 
            index++;

            /* If the GCP source hasn't been set, set it to the post-
               residuals source */
            if (strlen(residual_stats->gcp_source) == 0)
            {
                snprintf(residual_stats->gcp_source,
                    sizeof(residual_stats->gcp_source),
                    "%s", post_gcp_res[i].gcp_source);
            }
        }
    }

    return SUCCESS;
}
