#!/bin/csh -f

#***********************************************************************
# NAME:     check_results
# PURPOSE:  checks geodetic characterization results
# RETURNS:  0 if successful, 1 on an error condition
#***********************************************************************/

# Initialize work order ID and exit status values
set woid = $1
set exit_status = 0

# Ensure the test environment is set
if ( ! $?IAS_UNIT_TEST_DATA_DIR ) then
    echo "IAS_UNIT_TEST_DATA_DIR isn't set"
    exit 1
endif

# Check the report file -- because of the changes in header, only look at
# the lines from the 'GCP' on.
sed '/GCP/,$!'d geodetic.rpt | diff -q - \
    $IAS_UNIT_TEST_DATA_DIR/ias_base/gps/geodetic/expect_geodetic.rpt
if ( $status == 1 ) then
    echo "Report results differ from expected"
    set exit_status = 1
endif

# If database support is disabled, exit here with the status obtained from
# testing the report file
if ( ! $?ENABLE_DB_TESTS ) then
    echo
    echo `basename $0` ":  Database tests are disabled."
    echo
    exit $exit_status
endif

# Ensure the database login/password has been defined in the environment
if ( ! $?IAS_DB_L8_OLITIRS ) then
    echo "IAS_DB_L8_OLITIRS isn't set"
    exit 1
endif

# Get the geometric common ID
set sql_script = "retrieve_woid.sql"
echo "set term off" > $sql_script
echo "set echo on" >> $sql_script
echo "set trimout on" >> $sql_script
echo "set trimspool on" >> $sql_script
echo "set heading off" >> $sql_script
echo "set feedback off" >> $sql_script
echo "set pause off" >> $sql_script
echo "set pagesize 0" >> $sql_script
echo "set linesize 500" >> $sql_script
echo "set space 0" >> $sql_script
echo "set colsep |" >> $sql_script
echo "set tab off" >> $sql_script
echo "set verify off" >> $sql_script
echo "spool ./woid.txt" >> $sql_script
echo "select work_order_id, geometric_wo_common_id " >> $sql_script
echo "from geometric_work_order_common " >> $sql_script
echo "where work_order_id='"$woid"';" >> $sql_script
echo "spool off" >> $sql_script
echo "exit" >> $sql_script

# Query the database
sqlplus $IAS_DB_L8_OLITIRS @$sql_script > & /dev/null
rm $sql_script

# Squeeze out the white space and select the lines beginning with the 
# work order ID
set common_id = `sed 's/  *//g' ./woid.txt | grep ^$woid | awk -F'|' \
    '{print $2}'`

# Create a script to query the database ... this is done here so that a
# separate script will not have to be maintained.  It will be used then
# removed at the end.
set sql_script = "retrieve_stats.sql"
echo "set term off" > $sql_script
echo "set echo on" >> $sql_script
echo "set trimout on" >> $sql_script
echo "set trimspool on" >> $sql_script
echo "set heading off" >> $sql_script
echo "set feedback off" >> $sql_script
echo "set pause off" >> $sql_script
echo "set pagesize 0" >> $sql_script
echo "set linesize 500" >> $sql_script
echo "set space 0" >> $sql_script
echo "set colsep |" >> $sql_script
echo "set tab off" >> $sql_script
echo "set verify off" >> $sql_script
echo "spool ./db_results.txt" >> $sql_script
echo "select nbr_of_vld_pnts,  " >> $sql_script
echo "mean_lat_of_gcps, mean_lon_of_gcps, " >> $sql_script
echo "pst_fit_mean_acrss_trk, pst_fit_mean_alng_trk, " >> $sql_script
echo "pst_fit_std_dev_acrss_trk, pst_fit_std_dev_alng_trk, " >> $sql_script
echo "pst_fit_rmse_acrss_trk, pst_fit_rmse_alng_trk, " >> $sql_script
echo "pst_fit_corr_coeff, " >> $sql_script
echo "pre_fit_mean_acrss_trk, pre_fit_mean_alng_trk, " >> $sql_script
echo "pre_fit_std_dev_acrss_trk, pre_fit_std_dev_alng_trk, " >> $sql_script
echo "pre_fit_rmse_acrss_trk, pre_fit_rmse_alng_trk, " >> $sql_script
echo "pre_fit_corr_coeff, " >> $sql_script
echo "gcp_source from geodetic_characterization " >> $sql_script
echo "where geometric_wo_common_id=$common_id;"  >> $sql_script
echo "spool off" >> $sql_script
echo "exit" >> $sql_script

# Query the database
sqlplus $IAS_DB_L8_OLITIRS @$sql_script > & /dev/null

# Squeeze out the white space and select the line with pipes
sed 's/  *//g' ./db_results.txt | grep '|' > ./db_results.tmp
sed 's/|/ /g' ./db_results.tmp > ./db_results.txt

# Compare the results
diff -q $IAS_UNIT_TEST_DATA_DIR/ias_base/gps/geodetic/db_results.txt .
if ( $status == 1 ) then
    echo "Database results differ from expected"
    set exit_status = 1
endif

# Do some cleanup
rm ./db_results.txt ./db_results.tmp ./woid.txt $sql_script

# Exit with the current status value
exit $exit_status
