/*----------------------------------------------------------------------------
  NAME:    geodetic_test.c

  PURPOSE: Exercise the geodetic application routines.

  USAGE:   To execute all tests:
               geodetic_test
           To execute a single test:
               geodetic_test test_name

-----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "cgreen/cgreen.h"
#include "../geodetic_char.h"

static char *unit_test_dir = NULL;
static char test_directory[] = "ias_base/gps/geodetic";
static char *enable_db_tests = NULL;

Ensure test_get_parameters(void);
Ensure test_get_residuals(void);
Ensure test_remove_outliers(void);
Ensure test_full_application(void);
Ensure test_precision_succeeded_flag(void);

int main(int argc, char *argv[])
{
    /* Get the ENABLE_DB_TESTS environment variable value (if it exists) */
    enable_db_tests = getenv("ENABLE_DB_TESTS");

    TestSuite *suite = create_test_suite();
    add_test(suite, test_get_parameters);
    add_test(suite, test_get_residuals);
    add_test(suite, test_remove_outliers);
    add_test(suite, test_full_application);
    add_test(suite, test_precision_succeeded_flag);

    unit_test_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (unit_test_dir == NULL)
    {
        printf("The IAS_UNIT_TEST_DATA_DIR environment variable is not set\n");
        return EXIT_FAILURE;
    }

    if (argc == 2)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}


Ensure test_get_parameters(void)
{
    int status;
    FILE *ouf;
    char parameter_filename[] = "test_parameters";
    char work_order_id[] = "D8008";
    char omf_filename[15];
    int store_in_database;
    char l1g_filename[PATH_MAX]; 

    /* Create a test parameter file -- Note: on errors this file is left
       hanging around. */
    ouf = fopen(parameter_filename, "w");
    if (!ouf)
    {
        assert_true_with_message(0, "Error opening parameter file");
        return;
    }

    sprintf(l1g_filename, "%s/%s/oli_baseline_gcp.h5", unit_test_dir,
        test_directory);
    
    status = fprintf(ouf, "OBJECT = BASELINE_GCP\n" 
                           "SCRIPT_NAME = \"GEODETIC\"\n"
                           "WORK_ORDER_ID = \"D8008\"\n"
                           "PROCESSING_PASS = 1\n"
                           "REPORT_FILENAME = \"output_file\"\n"
                           "STORE_CHARACTERIZATION_IN_DB = 1\n"
                           "END_OBJECT = BASELINE_GCP\n" 
                           "END\n"); 
    if (status <= 0)
    {
        assert_true_with_message(0, "Error writing to parameter file");
        return;
    }

    status = fclose(ouf);
    if (status < 0)
    {
        assert_true_with_message(0, "Error closing parameter file");
        return;
    }

    /* Create an OMF file -- Note: on errors this file is left hanging
       around. */
    sprintf(omf_filename, "%s.omf", work_order_id);
    ouf = fopen(omf_filename, "w");
    if (!ouf)
    {
        assert_true_with_message(0, "Error opening OMF file");
        return;
    }

    status = fprintf(ouf, "OBJECT = OMF\n"
            "  L1G_FILENAME_PASS_1 = \"%s\"\n"
            "  PRECISION_RESIDUALS_FILENAME_PASS_1 = \"residuals\"\n"
            "  PRECISION_SUCCEEDED_FLAG_PASS_1 = 1\n"
            "END_OBJECT = OMF\nEND\n",
            l1g_filename);
    if (status <= 0)
    {
        assert_true_with_message(0, "Error writing to OMF file");
        return;
    }

    status = fclose(ouf);
    if (status < 0)
    {
        assert_true_with_message(0, "Error closing OMF file");
        return;
    }

    /* Call the routine to get the parameters */
    status = get_geod_char_parms(parameter_filename, &store_in_database);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from "
            "get_geod_char_parms");
        return;
    }

    assert_equal(strcmp("D8008", get_work_order_id()), 0);
    assert_equal(strcmp("residuals", get_sol_res_name()), 0);
    assert_equal(strcmp("output_file", get_res_record_name()), 0);
    assert_equal(40, get_wrs_path());
    assert_equal(33, get_wrs_row());

    /* Clean up */
    remove(parameter_filename);
    remove(omf_filename);
}


Ensure test_get_residuals(void)
{
    char *residuals_filename;
    char filename[] = "precision_com.residuals";
    int name_size;
    IAS_MISC_GCP_RESIDUAL *gcp_res;
    int number_of_residuals;
    int status;

    name_size = strlen(unit_test_dir) + strlen(test_directory) +
        strlen(filename) + 3; 
    residuals_filename = malloc(name_size * sizeof(char));
    if (!residuals_filename)
    {
        assert_true_with_message(0, "Error allocating residuals file name");
        return;
    }

    sprintf(residuals_filename, "%s/%s/%s", unit_test_dir, test_directory,
        filename);

    /* Get the pre-fit residuals -- that's what the 0 is specifying */
    status = ias_misc_read_gcp_residuals(residuals_filename, 0, &gcp_res,
        &number_of_residuals);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from get_residuals");
        free(residuals_filename);
        return;
    }

    /* Make sure that 196 records were returned */
    assert_equal(196, number_of_residuals);
    if (number_of_residuals != 196)
    {
        free(gcp_res);
        free(residuals_filename);
        return;
    }


    /* Spot check the structure members for the first record */
    assert_equal(gcp_res[0].band_number, 1);
    assert_equal(strcmp("040033002", gcp_res[0].point_id), 0);
    assert_double_equal( 65531.488819, gcp_res[0].seconds_from_epoch);
    assert_double_equal( -56.380, gcp_res[0].residual_y);

    /* Check an intermediate record */
    assert_equal(gcp_res[119].band_number, 1);
    assert_equal(strcmp(gcp_res[119].point_id, "040033191"), 0);
    assert_double_equal(gcp_res[119].predicted_line, 7236.000);
    assert_double_equal(gcp_res[119].predicted_sample, 3626.000);
    assert_double_equal(gcp_res[119].seconds_from_epoch, 65524.619814);
    assert_double_equal(gcp_res[119].latitude, 38.894868);
    assert_double_equal(gcp_res[119].longitude, -116.274794);
    assert_double_equal(gcp_res[119].height, 2276.088);
    assert_double_equal(gcp_res[119].across_track_angle, 0.086049);
    assert_double_equal(gcp_res[119].along_track_residual, -72.527);
    assert_double_equal(gcp_res[119].across_track_residual, -73.562);
    assert_double_equal(gcp_res[119].residual_y, -59.192);
    assert_double_equal(gcp_res[119].residual_x, -84.553);
    assert_equal(gcp_res[119].outlier_flag, 1);
    assert_equal(strcmp(gcp_res[119].gcp_source, "DOQ"), 0);

    /* Check all of the last record */
    assert_equal(gcp_res[195].band_number, 1);
    assert_equal(strcmp(gcp_res[195].point_id, "040033319"), 0);
    assert_double_equal(gcp_res[195].predicted_line, 8926.000);
    assert_double_equal(gcp_res[195].predicted_sample, 6161.000);
    assert_double_equal(gcp_res[195].seconds_from_epoch, 65527.801568);
    assert_double_equal(gcp_res[195].latitude, 38.662922);
    assert_double_equal(gcp_res[195].longitude, -115.840090);
    assert_double_equal(gcp_res[195].height, 2175.283);
    assert_double_equal(gcp_res[195].across_track_angle, 0.025467);
    assert_double_equal(gcp_res[195].along_track_residual, -64.994);
    assert_double_equal(gcp_res[195].across_track_residual, -71.663);
    assert_double_equal(gcp_res[195].residual_y, -52.076);
    assert_double_equal(gcp_res[195].residual_x, -81.454);
    assert_equal(gcp_res[195].outlier_flag, 1);
    assert_equal(strcmp(gcp_res[195].gcp_source, "DOQ"), 0);

    free(gcp_res);
    gcp_res = NULL;

    /* Now get the post-fit residuals -- the -1 parameter */
    status = ias_misc_read_gcp_residuals(residuals_filename, -1, &gcp_res,
        &number_of_residuals);
    free(residuals_filename);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from get_residuals");
        return;
    }

    /* Make sure that 196 records were returned */
    assert_equal(196, number_of_residuals);
    if (number_of_residuals != 196)
    {
        free(gcp_res);
        return;
    }

    /* Check the first record */
    assert_equal(gcp_res[0].band_number, 1);
    assert_equal(strcmp(gcp_res[0].point_id, "040033002"), 0);
    assert_double_equal(gcp_res[0].predicted_line, 10616.000);
    assert_double_equal(gcp_res[0].predicted_sample, 4471.000);
    assert_double_equal(gcp_res[0].seconds_from_epoch, 65531.488819);
    assert_double_equal(gcp_res[0].latitude, 38.437019);
    assert_double_equal(gcp_res[0].longitude, -116.134169);
    assert_double_equal(gcp_res[0].height, 1932.773);
    assert_double_equal(gcp_res[0].across_track_angle, 0.052876);
    assert_double_equal(gcp_res[0].along_track_residual, -2.856);
    assert_double_equal(gcp_res[0].across_track_residual, -0.416);
    assert_double_equal(gcp_res[0].residual_y, -2.745);
    assert_double_equal(gcp_res[0].residual_x, -0.887);
    assert_equal(gcp_res[0].outlier_flag, 1);
    assert_equal(strcmp(gcp_res[0].gcp_source, "DOQ"), 0);

    /* Check a record in the middle */
    assert_equal(gcp_res[72].band_number, 1);
    assert_equal(strcmp(gcp_res[72].point_id, "040033126"), 0);
    assert_double_equal(gcp_res[72].predicted_line, 6391.000);
    assert_double_equal(gcp_res[72].predicted_sample, 6161.000);
    assert_double_equal(gcp_res[72].seconds_from_epoch, 65521.639000);
    assert_double_equal(gcp_res[72].latitude, 39.005518);
    assert_double_equal(gcp_res[72].longitude, -115.834515);
    assert_double_equal(gcp_res[72].height, 2005.787);
    assert_double_equal(gcp_res[72].across_track_angle, 0.037468);
    assert_double_equal(gcp_res[72].along_track_residual, -4.074);
    assert_double_equal(gcp_res[72].across_track_residual, -4.034);
    assert_double_equal(gcp_res[72].residual_y, -3.343);
    assert_double_equal(gcp_res[72].residual_x, -4.653);
    assert_equal(gcp_res[72].outlier_flag, 1);
    assert_equal(strcmp(gcp_res[72].gcp_source, "DOQ"), 0);

    /* Check the last record */
    assert_equal(gcp_res[195].band_number, 1);
    assert_equal(strcmp(gcp_res[195].point_id, "040033319"), 0);
    assert_double_equal(gcp_res[195].predicted_line, 8926.000);
    assert_double_equal(gcp_res[195].predicted_sample, 6161.000);
    assert_double_equal(gcp_res[195].seconds_from_epoch, 65527.801568);
    assert_double_equal(gcp_res[195].latitude, 38.662922);
    assert_double_equal(gcp_res[195].longitude, -115.840090);
    assert_double_equal(gcp_res[195].height, 2175.283);
    assert_double_equal(gcp_res[195].across_track_angle, 0.025364);
    assert_double_equal(gcp_res[195].along_track_residual, -1.807);
    assert_double_equal(gcp_res[195].across_track_residual, 2.818);
    assert_double_equal(gcp_res[195].residual_y, -2.250);
    assert_double_equal(gcp_res[195].residual_x, 2.474);
    assert_equal(gcp_res[195].outlier_flag, 1);
    assert_equal(strcmp(gcp_res[195].gcp_source, "DOQ"), 0);

    free(gcp_res);
}


Ensure test_remove_outliers(void)
{
    char *residuals_filename;
    char filename[] = "precision.residuals";
    char expected_results_filename[] = "residuals_sans_outliers";
    char *expected_results_file;
    int name_size;
    IAS_MISC_GCP_RESIDUAL *pre_gcp_res;
    IAS_MISC_GCP_RESIDUAL *post_gcp_res;
    int number_of_residuals;
    int status;
    int  num_valid;
    double *GCP_lat;
    double *GCP_lon;
    double *pre_across_res;
    double *pre_along_res;
    double *post_across_res; 
    double *post_along_res;
    DB_RESIDUAL_STATISTICS residual_stats;
    FILE *exp_fid;
    double expected_lat;
    double expected_residual;
    int i;

    name_size = strlen(unit_test_dir) + strlen(test_directory) +
        strlen(filename) + 3; 
    residuals_filename = malloc(name_size * sizeof(char));
    if (!residuals_filename)
    {
        assert_true_with_message(0, "Error allocating residuals file name");
        return;
    }

    sprintf(residuals_filename, "%s/%s/%s", unit_test_dir, test_directory,
        filename);

    /* Get the pre-fit residuals -- that's what the 0 is specifying */
    status = ias_misc_read_gcp_residuals(residuals_filename, 0, &pre_gcp_res,
        &number_of_residuals);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from get_residuals");
        free(residuals_filename);
        return;
    }

    /* Make sure that 196 records were returned */
    assert_equal(196, number_of_residuals);
    if (number_of_residuals != 196)
    {
        free(pre_gcp_res);
        free(residuals_filename);
        return;
    }

    /* Get the post-fit residuals -- that's what the -1 is specifying */
    status = ias_misc_read_gcp_residuals(residuals_filename, -1,
        &post_gcp_res, &number_of_residuals);
    free(residuals_filename);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from get_residuals");
        free(pre_gcp_res);
        return;
    }

    /* Make sure that 196 records were returned */
    assert_equal(196, number_of_residuals);
    if (number_of_residuals != 196)
    {
        free(pre_gcp_res);
        free(post_gcp_res);
        return;
    }

    residual_stats.gcp_source[0] = '\0';
    status = remove_outliers(pre_gcp_res, post_gcp_res, number_of_residuals,
        number_of_residuals, &num_valid, &GCP_lat, &GCP_lon,
        &pre_across_res, &pre_along_res, &post_across_res,
        &post_along_res, &residual_stats);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from remove_outliers");
        free(pre_gcp_res);
        free(post_gcp_res);
        return;
    }

    /* These are no longer needed */
    free(pre_gcp_res);
    pre_gcp_res = NULL;
    free(post_gcp_res);
    post_gcp_res = NULL;

    if (num_valid != 137)
    {
        assert_true_with_message(0,
            "remove_outliers returned %d instead of 137 as expected");
        return;
    }
    assert_equal(137, num_valid);


    /* Build the name of the expected results file */
    name_size = strlen(unit_test_dir) + strlen(test_directory) +
        strlen(expected_results_filename) + 3; 
    expected_results_file = malloc(name_size * sizeof(char));
    if (!expected_results_file)
    {
        assert_true_with_message(0, "Error expected results file name");
        return;
    }

    sprintf(expected_results_file, "%s/%s/%s", unit_test_dir, test_directory,
        expected_results_filename);
    exp_fid = fopen(expected_results_file, "r");
    free(expected_results_file);
    if (!exp_fid)
    {
        assert_true_with_message(0, "Error opening: %s",
            expected_results_file);
        free(GCP_lat);
        free(GCP_lon);
        free(pre_across_res);
        free(pre_along_res);
        free(post_across_res); 
        free(post_along_res);
        return;
    }

    for (i = 0; i < num_valid; i++)
    {
        status = fscanf(exp_fid, "%lf %*f %*f %lf", &expected_lat,
            &expected_residual);
        if (status <= 0)
        {
            assert_true_with_message(0, "Error reading expected results");
            fclose(exp_fid);
            free(GCP_lat);
            free(GCP_lon);
            free(pre_across_res);
            free(pre_along_res);
            free(post_across_res); 
            free(post_along_res);
            return;
        }

        assert_double_equal(expected_lat, GCP_lat[i]);
        assert_double_equal(expected_residual, post_across_res[i]);
    }

    fclose(exp_fid);
    free(GCP_lat);
    free(GCP_lon);
    free(pre_across_res);
    free(pre_along_res);
    free(post_across_res); 
    free(post_along_res);
}


Ensure test_full_application(void)
{
    int status;
    FILE *ouf;
    char parameter_filename[] = "test_parameters";
    char work_order_id[] = "X8008";
    char omf_filename[15];
    char residual_filename[PATH_MAX];
    char l1g_filename[PATH_MAX];
    char command_string[256];

    /* Build the name of the precision residual file */
    sprintf(residual_filename, "%s/%s/precision.residuals",
        unit_test_dir, test_directory);

    /* Create a test parameter file -- Note: on errors this file is left
       hanging around. */
    ouf = fopen(parameter_filename, "w");
    if (!ouf)
    {
        assert_true_with_message(0, "Error opening parameter file");
        return;
    }

    status = fprintf(ouf, "OBJECT = TEST_PARAMETERS\n" 
        "SCRIPT_NAME = \"GEODETIC\"\n"
        "WORK_ORDER_ID = \"X8008\"\n"
        "PROCESSING_PASS = 1\n"
        "REPORT_FILENAME = \"geodetic.rpt\"\n"
        "END_OBJECT = TEST_PARAMETERS\n" 
        "END\n"); 
    if (status <= 0)
    {
        assert_true_with_message(0, "Error writing to parameter file");
        fclose(ouf);
        return;
    }

    /* Check ENABLE_DB_TESTS to determine whether to set writing the
       results to the database.  If ENABLE_DB_TESTS is set, do it, otherwise,
       don't */
    if (enable_db_tests == NULL)
    {
        printf("\n");
        printf("%s:  Database tests are disabled.\n", __FILE__);
        printf("\n"); 
        status = fprintf(ouf, "STORE_CHARACTERIZATION_IN_DB = 0\n");
    }
    else
    {
        status = fprintf(ouf, "STORE_CHARACTERIZATION_IN_DB = 1\n");
    }
    if (status <= 0)
    {
        assert_true_with_message(0, "Error writing to parameter file");
        fclose(ouf);
        return;
    }

    status = fclose(ouf);
    if (status < 0)
    {
        assert_true_with_message(0, "Error closing parameter file");
        return;
    }

    /* Create an OMF file -- Note: on errors this file is left hanging
       around. */
    sprintf(omf_filename, "%s.omf", work_order_id);
    ouf = fopen(omf_filename, "w");
    if (!ouf)
    {
        assert_true_with_message(0, "Error opening OMF file");
        return;
    }

    sprintf(l1g_filename, "%s/%s/oli_baseline_gcp.h5",
        unit_test_dir, test_directory);
    status = fprintf(ouf, "OBJECT = OMF\n"
            "  L1G_FILENAME_PASS_1 = \"%s\"\n"
            "  PRECISION_RESIDUALS_FILENAME_PASS_1 = \"%s\"\n"
            "  PRECISION_SUCCEEDED_FLAG_PASS_1 = 1\n"
            "END_OBJECT = OMF\nEND\n",
            l1g_filename, residual_filename);
    if (status <= 0)
    {
        assert_true_with_message(0, "Error writing to OMF file");
        fclose(ouf);
        return;
    }

    status = fclose(ouf);
    if (status < 0)
    {
        assert_true_with_message(0, "Error closing OMF file");
        return;
    }

    /* Execute the application */
    sprintf(command_string, "../geodetic %s", parameter_filename);
    status = system(command_string);
    if (status)
    {
        assert_true_with_message(0, "Error executing: %s", command_string);
        return;
    }

    /* Run a script to make sure database results and report results are
       good */
    status = system("./check_results.csh X8008");
    if (status)
    {
        assert_true_with_message(0, "./check_results.csh returned an error");
        return;
    }

    /* Clean up */
    status = system("rm test_parameters X8008.omf geodetic.rpt");
    if (status)
        assert_true_with_message(0, "Error removing files");
}

Ensure test_precision_succeeded_flag(void)
{
    int status;
    FILE *ouf;
    char parameter_filename[] = "test_parameters";
    char work_order_id[] = "P8008";
    char omf_filename[15];
    char residual_filename[PATH_MAX];
    char l1g_filename[PATH_MAX];
    char command_string[256];

    /* Create a test parameter file -- Note: on errors this file is left
       hanging around. */
    ouf = fopen(parameter_filename, "w");
    if (!ouf)
    {
        assert_true_with_message(0, "Error opening parameter file");
        return;
    }

    /* Specify a file that doesn't exist. Since we are testing the 
       precision succeeded flag being false, geodetic should exit 
       before seeing that file does not exist. */
    sprintf(residual_filename, "%s/%s/precision_missing.residuals",
        unit_test_dir, test_directory);

    status = fprintf(ouf, "OBJECT = TEST_PARAMETERS\n" 
                           "SCRIPT_NAME = \"GEODETIC\"\n"
                           "WORK_ORDER_ID = \"P8008\"\n"
                           "PROCESSING_PASS = 1\n"
                           "REPORT_FILENAME = \"geodetic.rpt\"\n"
                           "STORE_CHARACTERIZATION_IN_DB = 1\n"
                           "END_OBJECT = TEST_PARAMETERS\n" 
                           "END\n"); 
    if (status <= 0)
    {
        assert_true_with_message(0, "Error printing to parameter file");
        return;
    }

    status = fclose(ouf);
    if (status < 0)
    {
        assert_true_with_message(0, "Error closing parameter file");
        return;
    }

    /* Create an OMF file -- Note: on errors this file is left hanging
       around. */
    sprintf(omf_filename, "%s.omf", work_order_id);
    ouf = fopen(omf_filename, "w");
    if (!ouf)
    {
        assert_true_with_message(0, "Error opening OMF file");
        return;
    }

    sprintf(l1g_filename, "%s/%s/oli_baseline_gcp.h5",
        unit_test_dir, test_directory);
    status = fprintf(ouf, "OBJECT = OMF\n"
            "  L1G_FILENAME_PASS_1 = \"%s\"\n"
            "  PRECISION_RESIDUALS_FILENAME_PASS_1 = \"%s\"\n"
            "  PRECISION_SUCCEEDED_FLAG_PASS_1 = 0\n"
            "END_OBJECT = OMF\nEND\n",
            l1g_filename, residual_filename);
    if (status <= 0)
    {
        assert_true_with_message(0, "Error printing to OMF file");
        return;
    }

    status = fclose(ouf);
    if (status < 0)
    {
        assert_true_with_message(0, "Error closing OMF file");
        return;
    }

    /* Execute the application */
    sprintf(command_string, "../geodetic %s", parameter_filename);
    status = system(command_string);
    if (status)
    {
        assert_true_with_message(0, "Error executing: %s", command_string);
        return;
    }

    /* Clean up */
    status = system("rm test_parameters P8008.omf");
    if (status)
        assert_true_with_message(0, "Error removing files");
}
