/******************************************************************************
NAME:    write_residual_statistics_report.c

PURPOSE: Write the GCP residual statistics to the ASCII output file.  

RETURNS: SUCCESS or ERROR

*******************************************************************************
                        Property of the U.S. Government
                            USGS EROS Data Center
******************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <float.h>

#include "ias_logging.h"           /* Logging Library        */
#include "ias_const.h"             /* ERROR/SUCCESS defines  */
#include "ias_satellite_attributes.h"
#include "ias_miscellaneous.h"     /* Miscellaneous Library  */
#include "geodetic_char.h"      

int write_residual_statistics_report
(
    FILE *stats_fptr,              /* I: Name of the GCP residual stats. file */
    const GCP_RES_STATS *prefit_stats,  /* I: pre-fit GCP residual stats */
    const GCP_RES_STATS *postfit_stats, /* I: post-fit GCP residual stats */
    const char *gcp_source,             /* I: GCP source */
    int band_number                     /* I: The first pre-GCP band number */
)
{
    int status;                         /* status return of the function */
    IAS_REPORT_HEADER *header = NULL;   /* Output header structure       */
    char prog_name[PATH_MAX];           /* local copy of program name */

    strcpy(prog_name, PROG_NAME);

    /*---------------------------------------------------------------------
      Write the GPS-style output header.
      ---------------------------------------------------------------------*/
    header = ias_misc_initialize_gps_report_header(get_l1g_image_name(),
            get_l0r_image_name(), get_work_order_id(),
            ias_misc_convert_to_uppercase(prog_name));
    if (header == NULL)
    {
        IAS_LOG_ERROR("Initializing GPS report header");
        return ERROR;
    }

    status = ias_misc_write_gps_report_header(stats_fptr, header);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the GPS header for this report");
        ias_misc_free_report_header(header);
        return ERROR;
    }

    /* Release the report header memory allocated by that library */
    ias_misc_free_report_header(header);
    header = NULL;

    /*-----------------------------------------------------------------------
      Write the statistics column headers followed by the data
      -----------------------------------------------------------------------*/
    status = fprintf(stats_fptr,
            "GCP     # Valid  Mean Lat.     Mean Lon.\n"
            "Source  Points   (degrees)     (degrees)\n"
            "------  -------  ------------  ------------\n"
            "   %s     %4d  %12.6f  %12.6f\n\n\n",
            gcp_source, postfit_stats->num_points, postfit_stats->mean_lat,
            postfit_stats->mean_lon);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the column headers");
        return ERROR;
    }

    status = fprintf(stats_fptr,
            "                     Pre-fit Residual Statistics (meters)\n"
            "                     ------------------------------------\n\n"
            "      Mean                 RMSE               Std. Dev.        "
            "Corr. Coeff.\nacross    along      across    along      across"
            "    along       along vs.\ntrack     track      track     "
            "track      track     track       across\n--------  --------   "
            "--------  --------   --------  --------    ----------\n"
            "%8.3f  %8.3f   %8.3f  %8.3f   %8.3f  %8.3f     %9.4f\n\n\n",
            prefit_stats->mean_across_res, prefit_stats->mean_along_res,
            prefit_stats->rmse_across_res, prefit_stats->rmse_along_res, 
            prefit_stats->sd_across_res,   prefit_stats->sd_along_res, 
            prefit_stats->corr_coeff);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the pre-fit residual statistics data");
        return ERROR;
    }

    status = fprintf(stats_fptr,
            "                    Post-fit Residual Statistics (meters)\n"
            "                    -------------------------------------\n\n"
            "      Mean                 RMSE               Std. Dev.        "
            "Corr. Coeff.\nacross    along      across    along      across"
            "    along       along vs.\ntrack     track      track     "
            "track      track     track       across\n--------  --------   "
            "--------  --------   --------  --------    ----------\n"
            "%8.3f  %8.3f   %8.3f  %8.3f   %8.3f  %8.3f     %9.4f\n",
            postfit_stats->mean_across_res, postfit_stats->mean_along_res,
            postfit_stats->rmse_across_res, postfit_stats->rmse_along_res, 
            postfit_stats->sd_across_res,   postfit_stats->sd_along_res, 
            postfit_stats->corr_coeff);

    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the post-fit residual statistics data");
        return ERROR;
    }

    return SUCCESS;
}
