/*****************************************************************************
NAME:    write_residual_statistics_to_db.c

PURPOSE: Writes the residual statistics to the DB.

RETURNS: SUCCESS or ERROR

******************************************************************************
                       Property of the U.S. Government
                           USGS EROS Data Center
******************************************************************************/
#include <stdlib.h>
#include <string.h>

#include "ias_logging.h"
#include "ias_db.h"
#include "ias_db_insert.h"
#include "ias_db_insert_transaction_using_table.h"
#include "ias_db_get_connect_info.h"
#include "ias_miscellaneous.h"
#include "geodetic_char.h"

#define SQL_QUERYLEN 256

int write_residual_statistics_to_db
(
    DB_RESIDUAL_STATISTICS *residual_stats /* I: data to write to DB */
)
{
    char *database_name = NULL;
    char *user_name = NULL;
    char *password = NULL;
    char *host = NULL;
    int status;
    int table_size;
    struct ias_db_connection *db = NULL; /* Database connection */
    struct ias_db_query *query = NULL;
    char query_string[SQL_QUERYLEN];
    char error_message[IAS_SQL_RTN_MSGLEN];


    /* build the table of data to write to DB */
    IAS_DB_INSERT_TABLE table[] =
    {
        IAS_DB_TABLE_STRUCT_STRING( residual_stats, geometric_wo_common_id ),
        IAS_DB_TABLE_STRUCT_INT( residual_stats, nbr_of_vld_pnts ),
        IAS_DB_TABLE_STRUCT_DOUBLE( residual_stats, mean_lat_of_gcps ),
        IAS_DB_TABLE_STRUCT_DOUBLE( residual_stats, mean_lon_of_gcps ),
        IAS_DB_TABLE_STRUCT_DOUBLE( residual_stats, pst_fit_mean_alng_trk ),
        IAS_DB_TABLE_STRUCT_DOUBLE( residual_stats, pst_fit_mean_acrss_trk ),
        IAS_DB_TABLE_STRUCT_DOUBLE( residual_stats, pst_fit_std_dev_alng_trk ),
        IAS_DB_TABLE_STRUCT_DOUBLE( residual_stats, pst_fit_std_dev_acrss_trk ),
        IAS_DB_TABLE_STRUCT_DOUBLE( residual_stats, pst_fit_rmse_alng_trk ),
        IAS_DB_TABLE_STRUCT_DOUBLE( residual_stats, pst_fit_rmse_acrss_trk ),
        IAS_DB_TABLE_STRUCT_DOUBLE( residual_stats, pst_fit_corr_coeff ),
        IAS_DB_TABLE_STRUCT_DOUBLE( residual_stats, pre_fit_mean_alng_trk ),
        IAS_DB_TABLE_STRUCT_DOUBLE( residual_stats, pre_fit_mean_acrss_trk ),
        IAS_DB_TABLE_STRUCT_DOUBLE( residual_stats, pre_fit_std_dev_alng_trk ),
        IAS_DB_TABLE_STRUCT_DOUBLE( residual_stats, pre_fit_std_dev_acrss_trk ),
        IAS_DB_TABLE_STRUCT_DOUBLE( residual_stats, pre_fit_rmse_alng_trk ),
        IAS_DB_TABLE_STRUCT_DOUBLE( residual_stats, pre_fit_rmse_acrss_trk ),
        IAS_DB_TABLE_STRUCT_DOUBLE( residual_stats, pre_fit_corr_coeff ),
        IAS_DB_TABLE_STRUCT_STRING( residual_stats, gcp_source )
    };
    table_size = sizeof(table)/sizeof(IAS_DB_INSERT_TABLE);

    /* Get the database connection info */
    status = ias_db_get_connect_info( &database_name, &user_name,
        &password, &host );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Getting database connection info");
        ias_db_close_database_lib();
        return ERROR;
    }

    /* Connect to the database for a transaction */
    db = ias_db_connect_with_transaction( database_name, user_name,
        password, host );
    /* Free the resources before checking the connection */
    free( database_name );
    free( user_name );
    free( password );
    free( host );
    /* Check the connection */
    if ( !db )
    {
        IAS_LOG_ERROR("Connecting to database");
        ias_db_close_database_lib();
        return ERROR;
    }

    /* Get the common ID associated with this work order */
    status = ias_misc_get_char_id( db,
        "GEOMETRIC_WORK_ORDER_COMMON", "geometric_wo_common_id",
        get_work_order_id(), residual_stats->geometric_wo_common_id);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Retrieving geometric_wo_common_id for work order: %s",
            get_work_order_id());
        ias_db_close_connection( db );
        ias_db_close_database_lib();
        return ERROR;
    }

    /* Delete previous records */
    snprintf( query_string, sizeof(query_string) - 1, "delete from "
        "geodetic_characterization where geometric_wo_common_id=%s",
        residual_stats->geometric_wo_common_id );
    query = ias_db_perform_query( db, query_string );
    if ( !ias_db_query_was_successful( query ))
    {
        ias_db_query_get_error_message( query, error_message,
            sizeof(error_message));
        IAS_LOG_ERROR("Query of geodetic_characterization failed: %s",
            error_message );
        IAS_LOG_ERROR("Deleting records from geodetic_characterization where "
            "geometric_wo_common_id=%s",
            residual_stats->geometric_wo_common_id );
        ias_db_close_connection( db );
        ias_db_close_database_lib();
        return ERROR;
    }
    ias_db_query_close( query );
    query = NULL;

    /* Insert the data into the database */
    status = ias_db_insert_using_table( db, "geodetic_characterization",
        table, table_size, 1 );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Insert into geodetic_characterization failed");
        ias_db_close_connection( db );
        ias_db_close_database_lib();
        return ERROR;
    }

    status = ias_db_commit_transaction( db );
    if ( status != SUCCESS )
        IAS_LOG_ERROR("Committing database connection");
    ias_db_close_connection( db );
    ias_db_close_database_lib();

    return status;
}
