/*****************************************************************************
Name:           geometric_char (main)

Purpose:
    Geometric Accuracy Assessment (geometric characterization) analyzes the 
    results of the Ground control Point (GCP) measurements derived through
    correlation of the GCP image chips with a precision and terrain corrected 
    OLI L1G image. This algorithm provides a measure of the accuracy of the 
    output L1G product through direct comparison to an absolute ground control 
    source.  

Algorithm References:
    OLI Geometric Accuracy Assessment Algorithm, Algorithm Description Document

******************************************************************************
                        Property of the U.S. Government
                              USGS EROS Data Center
*******************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "ias_const.h"
#include "ias_satellite_attributes.h"
#include "ias_logging.h"
#include "ias_gcp.h"
#include "ias_geo.h"
#include "geometric_char.h"

int main (int argc, char *argv[])
{
    GCP_RES_STATS stats;            /* GCP residual statistics */
    int status;                     /* Status of return from function */
    double *gcp_lat;                /* Pointer to GCP residual latitude data */
    double *gcp_lon;                /* Pointer to GCP residual longitude data */
    double *along_res;              /* Pointer to along-scan residual data */
    double *across_res;             /* Pointer to across-scan residual data */
    DB_RESIDUAL_STATISTICS residual_stats; /* Data to write to database */
    IAS_GCP_RESULTS *gcp_results;   /* GCP results records */
    SCALED_FIT_OFFSETS *scaled_offsets; /* Scaled line and sample offsets */
    int num_gcp;                    /* Number of GCP results records */
    int valid_gcp_count;            /* Number of non-outlier GCP residuals */
    int gcp_index;                  /* GCP index */ 
    int band_index;                 /* Band index */
    int mem_alloc_size;             /* Size to be allocated to calculate once */
    double outlier_fraction;        /* Fraction of GCPs that are outliers */ 
    const struct IAS_CPF_GEO_SYSTEM *geo_system; /* GEO_SYSTEM CPF group */
    IAS_CPF *cpf;                   /* Calibration parameter file constants */
    double combined_rmse;           /* The combined along and across track
                                       rmse values */

    /* Initialize the residual statistics info structure */
    memset(&residual_stats, 0, sizeof(residual_stats));

    /* Initialize logging */
    status = ias_log_initialize("geometric");
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the logging library");
        exit(EXIT_FAILURE);
    }

    /* Check for the parameter file name */
    if (argc < 2)
    {
        IAS_LOG_ERROR("Usage: geometric <parameter_file>");
        exit(EXIT_FAILURE);
    }

    /* Read the parameters from the geometric char parameter file */
    status = get_geom_char_parms(argv[1]);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from file: %s", argv[1]);
        exit(EXIT_FAILURE);
    }

    /* Check whether precision model was created.  If not, exit */
    if (get_precision_succeeded_flag() == 0)
    {
        if (get_geometric_required_flag() == 0)
        {
            IAS_LOG_WARNING("A precision model was not created.  There is "
                    "nothing for geometric to do.");
            return EXIT_SUCCESS;
        }
        else
        {
            IAS_LOG_ERROR("A precision model was not created.  There is "
                    "nothing for geometric to do.");
            exit(EXIT_FAILURE);
        }
    }

    /* Some additional L1G metadata is also needed, so read that input data */
    status = get_l1g_header_data();
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving L1G header metadata from: %s",
                get_l1g_image_name());
        exit(EXIT_FAILURE);
    }

    /* Program entry message */
    IAS_LOG_INFO("Geometric Accuracy Assessment Characterization initiated");

    /* Read the CPF */
    cpf = ias_cpf_read(get_cpf_filename());
    if (cpf == NULL)
    {
        IAS_LOG_ERROR("Cannot open CPF file %s for reading",
                get_cpf_filename());
        exit(EXIT_FAILURE);
    }

    /* Read the measured GCP data */
    status = get_gcp_results(cpf, &gcp_results, &num_gcp);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading GCPs from file %s", get_measured_gcp_filename());
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }

    /* Get geometric parameters from the CPF */
    geo_system = ias_cpf_get_geo_system(cpf);
    if (geo_system == NULL)
    {
        IAS_LOG_ERROR("Reading Geometric system group from CPF");
        free(gcp_results);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }

    /* Check the number of correlated validation GCPs read */
    if (num_gcp < geo_system->minimum_number_correlated_validation_gcps) 
    {
        free(gcp_results);
        if (get_geometric_required_flag() == 0)
        {
            IAS_LOG_WARNING("Number of correlated validation GCPs read %d is "
                    "below the minimum threshold %d, so geometric results "
                    "should not be used", num_gcp, 
                    geo_system->minimum_number_correlated_validation_gcps); 
            ias_cpf_free(cpf);
            return EXIT_SUCCESS;
        }
        else
        {
            IAS_LOG_ERROR("Number of correlated validation GCPs read %d is "
                    "below the minimum threshold %d, so geometric results "
                    "should not be used", num_gcp, 
                    geo_system->minimum_number_correlated_validation_gcps); 
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }
    }

    /* Scale the GCP offsets from pixels to meters */
    mem_alloc_size = num_gcp * sizeof(SCALED_FIT_OFFSETS);
    scaled_offsets = malloc(mem_alloc_size);
    if (scaled_offsets == NULL)
    {
        IAS_LOG_ERROR("Allocating %d bytes for scaled GCP offsets",
                mem_alloc_size);
        free(gcp_results);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }

    for (gcp_index = 0; gcp_index < num_gcp; gcp_index++)
    {
        /* Find the band index from the GCP's band number */
        band_index = ias_sat_attr_convert_band_number_to_index(
                gcp_results[gcp_index].search_band);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Getting band index for band number %d\n",
                    gcp_results[gcp_index].search_band);
            free(gcp_results);
            free(scaled_offsets);
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }

        /* Check if we have the needed L1G band metadata to do the scaling */
        if (get_projection_distance_y(band_index) == 0
                || get_projection_distance_x(band_index) == 0)
        {
            /* Zero is a non-valid projection distance and will be encountered
               only if it really was invalid in the L1G or if we set it
               manually to zero in the get parameters routine for the case that
               the band metadata was not available for the band in the L1G. */
            IAS_LOG_ERROR("Missing or invalid projection distance for search "
                    "band %d", gcp_results[gcp_index].search_band);
            free(gcp_results);
            free(scaled_offsets);
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }

        /* Scale */
        scaled_offsets[gcp_index].line_offset
            = gcp_results[gcp_index].fit_line_offset
            * get_projection_distance_y(band_index);
        scaled_offsets[gcp_index].sample_offset
            = gcp_results[gcp_index].fit_sample_offset
            * get_projection_distance_x(band_index);

        if (get_write_stats_to_db_flag())
        {
            /* While looping over the bands for doing the scaling, grab the
               projection distance from the first GCP for the database output */
            if (gcp_index == 0)
            {
                residual_stats.pixel_size
                    = get_projection_distance_y(band_index);
            }
        }
    }

    /* Remove outliers */
    valid_gcp_count = num_gcp;
    status = remove_outliers(&gcp_results, &scaled_offsets, &valid_gcp_count, 
            &gcp_lat, &gcp_lon, &along_res, &across_res);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Removing GCP outliers");
        free(gcp_results);
        free(scaled_offsets);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    }

    /* Check if the percentage of outlier GCPs does not exceed the threshold */
    outlier_fraction = (double)(num_gcp - valid_gcp_count) / num_gcp;
    if (outlier_fraction > geo_system->max_percent_gcp_outliers)
    {
        free(gcp_lat);
        free(gcp_lon);
        free(along_res);
        free(across_res);
        free(gcp_results);
        free(scaled_offsets);
        if (get_geometric_required_flag() == 0)
        {
            IAS_LOG_WARNING("Fraction of GCPs that are outliers %f exceeds "
                    "threshold of %f, so geometric results should not be used",
                    outlier_fraction, geo_system->max_percent_gcp_outliers);
            ias_cpf_free(cpf);
            return EXIT_SUCCESS;
        }
        else
        {
            IAS_LOG_ERROR("Fraction of GCPs that are outliers %f exceeds "
                    "threshold of %f, so geometric results should not be used",
                    outlier_fraction, geo_system->max_percent_gcp_outliers);
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }
    }

    /* Calculate mean, RMSE, standard dev., and the correlation coefficient
       for the pre-fit residuals */
    status = ias_geo_analyze_gcp_records(gcp_lat, gcp_lon, along_res,
            across_res, valid_gcp_count, &stats);

    /* Free memory which was allocated in remove_outliers before checking
       status since it is no longer needed and can be freed now whether the
       previous routine was successful or not. */
    free(gcp_lat);
    gcp_lat = NULL;
    free(gcp_lon);
    gcp_lon = NULL;
    free(along_res);
    along_res = NULL;
    free(across_res);
    across_res = NULL;

    /* Check the return status of ias_geo_analyze_gcp_records() */
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating GCP residual statistics");
        free(gcp_results);
        free(scaled_offsets);
        ias_cpf_free(cpf);
        exit(EXIT_FAILURE);
    } 

    /* Check if the along and across track RMSE values exceed the thresholds */
    if (stats.rmse_along_res > geo_system->y_prefit_gcp_rms) 
    {
        free(gcp_results);
        free(scaled_offsets);
        if (get_geometric_required_flag() == 0)
        {
            IAS_LOG_WARNING("Along track RMSE %f exceeds threhold of %f, so "
                    "geometric results should not be used",
                    stats.rmse_along_res, geo_system->y_prefit_gcp_rms); 
            ias_cpf_free(cpf);
            return EXIT_SUCCESS;
        }
        else
        {
            IAS_LOG_ERROR("Along track RMSE %f exceeds threhold of %f, so "
                    "geometric results should not be used",
                    stats.rmse_along_res, geo_system->y_prefit_gcp_rms); 
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }
    }
    if (stats.rmse_across_res > geo_system->x_prefit_gcp_rms) 
    {
        free(gcp_results);
        free(scaled_offsets);
        if (get_geometric_required_flag() == 0)
        {
            IAS_LOG_WARNING("Across track RMSE %f exceeds threhold of %f, so "
                    "geometric results should not be used",
                    stats.rmse_across_res, geo_system->x_prefit_gcp_rms); 
            ias_cpf_free(cpf);
            return EXIT_SUCCESS;
        }
        else
        {
            IAS_LOG_ERROR("Across track RMSE %f exceeds threhold of %f, so "
                    "geometric results should not be used",
                    stats.rmse_across_res, geo_system->x_prefit_gcp_rms); 
            ias_cpf_free(cpf);
            exit(EXIT_FAILURE);
        }
    }

    /* Free CPF memory */
    ias_cpf_free(cpf);
    cpf = NULL;

    /* Output residual statistics to output file */
    status = write_residual_statistics_report(&stats, gcp_results,
            scaled_offsets);
    
    if (get_write_stats_to_db_flag())
    {
        /* Save the GCP source before cleaning up the GCP results */
        strcpy(residual_stats.gcp_source, gcp_results[0].chip_source);
    }

    /* Free memory for GCP results and scaled offsets allocated earlier before
       checking status since it is no longer needed and can be freed now whether
       the previous routine was successful or not. */
    free(gcp_results);
    gcp_results = NULL;
    free(scaled_offsets);
    scaled_offsets = NULL;

    /* Check the return status of writing the residual stats report */
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing residual statistics report");
        exit(EXIT_FAILURE);
    }

    /* If database output is requested, send the residual stats info to
       the database */
    if (get_write_stats_to_db_flag())
    {
        /* Setup the data in the residual statistics structure */
        /* GCP source and pixel size were setup previously */
        residual_stats.mean_lat_of_gcps = stats.mean_lat;
        residual_stats.mean_lon_of_gcps = stats.mean_lon;
        residual_stats.nbr_of_vld_pnts = stats.num_points;
        residual_stats.mean_alng_trk = stats.mean_along_res;
        residual_stats.mean_acrss_trk = stats.mean_across_res;
        residual_stats.rmse_alng_trk = stats.rmse_along_res;
        residual_stats.rmse_acrss_trk = stats.rmse_across_res;
        residual_stats.std_dev_alng_trk = stats.sd_along_res;
        residual_stats.std_dev_acrss_trk = stats.sd_across_res;
        residual_stats.correlation_coefficient = stats.corr_coeff;

        status = write_residual_statistics_to_db(&residual_stats);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Writing residual statistics data to database");
            exit(EXIT_FAILURE);
        }
    }

    /* Update the OMF with the RMSE along and across track residual
       statistics */
    combined_rmse = sqrt(stats.rmse_along_res * stats.rmse_along_res
                       + stats.rmse_across_res * stats.rmse_across_res);
    status = update_omf(stats.num_points, combined_rmse);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Cannot update OMF with rmse information");
        exit(EXIT_FAILURE);
    }

    /* Program successful exit message */
    IAS_LOG_INFO("Successful Completion");

    return EXIT_SUCCESS;
}
