#ifndef GEOMETRIC_CHAR_H
#define GEOMETRIC_CHAR_H

#include "ias_gcp.h"
#include "ias_geo.h"

typedef struct SCALED_FIT_OFFSETS
{
    double line_offset;     /* Scaled best-fit vertical (line) offset
                               of correlation peak */ 
    double sample_offset;   /* Scaled best-fit horizontal (sample)
                               offset of correlation peak */ 
} SCALED_FIT_OFFSETS;

/* Define a structure for the characterization information to be collected and 
   inserted into the characterization database table */
typedef struct db_residual_statistics
{
    /* Note: While some of these names are less than great, they must not
       be modified. They must match the field names used by the database
       for the insert_using_table database routine to work correctly. */
    char geometric_wo_common_id[IAS_CHAR_ID_SIZE];   
                                            /* Common work order ID for DB */
    char gcp_source[IAS_GCP_SOURCE_SIZE];   /* Source of GCPs */
    double mean_lat_of_gcps;                /* Mean latitude of valid GCPs
                                               (degrees) */
    double mean_lon_of_gcps;                /* Mean longitude of valid GCPs
                                               (degrees) */
    double pixel_size;                      /* Image pixel size for measured
                                               band (meters) */
    int nbr_of_vld_pnts;                    /* Number of GCPs accepted as
                                               valid */
    double mean_alng_trk;                   /* Mean of along track offsets
                                               (meters) */
    double mean_acrss_trk;                  /* Mean of across track offsets
                                               (meters) */
    double rmse_alng_trk;                   /* Root Mean Square Error of
                                               along track offsets (meters) */
    double rmse_acrss_trk;                  /* Root Mean Square Error of
                                               across track offsets (meters) */
    double std_dev_alng_trk;                /* Standard deviation of along
                                               track offsets (meters) */
    double std_dev_acrss_trk;               /* Standard deviation of across
                                               track offsets (meters) */
    double correlation_coefficient;         /* Correlation coefficient between
                                               along and across offsets */
} DB_RESIDUAL_STATISTICS;

/* Prototypes */
/**************/
int get_gcp_results
(
    IAS_CPF *cpf,                   /* I: Calibration Parm File */
    IAS_GCP_RESULTS **gcp_results,  /* O: GCP data records */
    int *num_gcp                    /* O: Number of GCPs read */
);

int get_geom_char_parms
(
    const char *parm_file   /* I: Name of the IAS Geometry Parameter File */
);

int get_l1g_header_data(void);

int remove_outliers 
(
    IAS_GCP_RESULTS **gcp_results,        /* I/O: Pointer to GCP measurement
                                                  results for outlier removal */
    SCALED_FIT_OFFSETS **scaled_offsets,  /* I/O: Pointer to scaled fit offset
                                                  results; modified to keep in
                                                  synch with the gcp_results as
                                                  outliers are removed */
    int *num_gcp,       /* I/O: Number of GCP measurement records on input,
                                number of non-outlier GCPs on output */ 
    double **gcp_lat,   /* O: GCP residual latitude data */
    double **gcp_lon,   /* O: GCP residual longitude data */
    double **along_res, /* O: Along track residual data */
    double **across_res /* O: Across track residual data */
);

int write_residual_statistics_report
(
    const GCP_RES_STATS *stats,     /* I: GCP residual statistics */
    const IAS_GCP_RESULTS *gcp_results,       /* I: Original GCP records */
    const SCALED_FIT_OFFSETS *scaled_offsets  /* I: Scaled fit offset results */
);

int write_residual_statistics_to_db
(
    DB_RESIDUAL_STATISTICS *residual_stats  /* I: Data to write to DB */
);

int update_omf
(
    int number_valid_points,  /* I: number of valid points */
    double combined_rmse      /* I: the combined rmse value */
);

/* Functions used to access parameters from the geometric_char parameter file,
   the OMF, and the L1G metadata */
const char *get_capture_date(void);
const char *get_cpf_filename(void);
const char *get_collection_type(void);
int get_geometric_required_flag(void);
const char *get_l0r_image_name(void);
const char *get_l1g_image_name(void);
const char *get_measured_gcp_filename(void);
const char *get_omf_filename(void);
double get_outlier_threshold(void);
int get_precision_succeeded_flag(void);
int get_processing_pass(void);
double get_projection_distance_x(int);
double get_projection_distance_y(int);
const char *get_report_filename(void);
int get_residual_report_flag(void);
const char *get_spacecraft(void);
double get_roll_angle(void);
int get_satellite_num(void);
const char *get_work_order_id(void);
int get_write_stats_to_db_flag(void);
int get_wrs_path(void);
int get_wrs_row(void);

#endif
