/*****************************************************************************
Name:           get_gcp_results

Purpose:
    Reads GCP results from measured GCP file.

Returns:
SUCCESS  The measured GCP results file was read successfully.
ERROR    Error trying to read the GCP file.

Note:
    The GCP data records are allocated within a function called by this
    routine. The caller is responsible to free the memory when done with it.
    If this routine errors, internally allocated memory is freed before
    returning.

******************************************************************************
                        Property of the U.S. Government
                            USGS EROS Data Center
*******************************************************************************/
#include <stdlib.h>
#include "ias_const.h"
#include "ias_gcp.h"
#include "ias_cpf.h"
#include "ias_logging.h"
#include "geometric_char.h"

int get_gcp_results
(
    IAS_CPF *cpf,                   /* I: Calibration Parm File */
    IAS_GCP_RESULTS **gcp_results,  /* O: GCP data records */
    int *num_gcp                    /* O: Number of GCPs read */
)
{
    double semi_major_axis;       /* Earth semi-major axis constant */
    double ellipticity;           /* Earth ellipticity constant */
    const struct IAS_CPF_EARTH_CONSTANTS *cpf_earth;
    int status;

    /* Get the earth parameters from the CPF */
    cpf_earth = ias_cpf_get_earth_const(cpf);
    if (cpf_earth == NULL)
    {
        IAS_LOG_ERROR("Reading Earth constants group from CPF");
        return ERROR;
    }

    /* Retrieve Earth semi-major axis and ellipticity values from CPF */
    semi_major_axis = cpf_earth->semi_major_axis; 
    ellipticity = cpf_earth->ellipticity;

    /* Read GCP results file */
    status = ias_gcp_read_correlation_results(get_measured_gcp_filename(), 
                semi_major_axis, ellipticity, gcp_results, num_gcp);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading GCP correlation results file %s",
                get_measured_gcp_filename());
        /* The memory for gcp_results was freed within the GCP library
           routine if an error occurred, so no need to free it here. */
        return ERROR;
    }
    if (*num_gcp < 0)
    {
        IAS_LOG_ERROR("%d valid ground control points found in GCP results "
                "file %s", *num_gcp, get_measured_gcp_filename());
        *gcp_results = NULL;
        return ERROR;
    }

    return SUCCESS;
}
