/*****************************************************************************
Name:           get_geom_char_parms

Purpose:
    Gets the parameters for geometric from the IAS Geometry parameter file
    (ODL Format) and the work order OMF file.
    Makes the data available through accessor functions.

Returns:
SUCCESS  Parameters were read from the parameter file.
ERROR    Error trying to read parameters.

******************************************************************************
                        Property of the U.S. Government
                            USGS EROS Data Center
*******************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "geometric_char.h"

#define DEFAULT_OUTLIER_THRESHOLD 0.95

/* --------------------------------------------------------------------- */
/* ------ These variables are returned by the accessor functions ------- */

/* Application parameter file parameters */
static char work_order_id[IAS_WORKORDER_ID_SIZE];
static double outlier_threshold;
static char report_filename[PATH_MAX];     /* Residual stats report filename */ 
static int residual_report_flag;
static int store_stats_in_db;              /* Flag for writing to DB */

/* OMF parameters */
static int satellite_num;
static char l0r_image_name[PATH_MAX];           /* The L0R image file name */
static char l1g_image_name[PATH_MAX];           /* The L1G image file name */
static char cpf_filename[PATH_MAX];  
static char measured_gcp_filename[PATH_MAX];
static int precision_succeeded_flag;   /* Flag for success of precision 
                                          solution */
static int geometric_required_flag;    /* Flag for success of geometric 
                                          processing being required */
static char omf_filename[PATH_MAX];    /* Name of OMF; from work order ID */

int get_geom_char_parms
(
    const char *parm_file   /* I: Name of the IAS Geometry Parameter File */
)
{
    int count;                      /* Number of elements */
    int status;                     /* Status returned from functions */
    /* Application parameter file parameters */
    int processing_pass;

    /* --------------------------------------------------------------------- */
    /* -- Table definition for things from the application parameter file -- */

    IAS_PARM_DECLARE_TABLE(parms, 7);

    /* PARAM 1: Get the work order ID.
       The standard parameter sets this as required. */
    IAS_PARM_WORK_ORDER_ID(parms, work_order_id, sizeof(work_order_id), 0);

    /* PARAM 2: Read the processing pass number.
       The standard parameters sets this as required and checks values. */
    IAS_PARM_PROCESSING_PASS(parms, &processing_pass, sizeof(processing_pass),
            0);

    /* PARAM 3: Read the outlier threshold.
       With default and valid range given to parameter definition, no
       additional validation needed. */
    double default_outlier_threshold[] = {DEFAULT_OUTLIER_THRESHOLD};
    IAS_PARM_ADD_DOUBLE(parms, OUTLIER_TOLERANCE,
            "GCP outlier probability confidence threshold",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
            /* has range flag, range low value, range high value */
            1, 0.0, 1.0,
            /* number defaults, default value pointer */
            1, default_outlier_threshold,
            &outlier_threshold, sizeof(outlier_threshold), 0);

    /* PARAM 4: Read the output report file name. */
    const char *default_report_filename[] = {"geometric.geo"};
    IAS_PARM_ADD_STRING(parms, REPORT_FILENAME,
            "Name of output report file", IAS_PARM_REQUIRED,
            /* valid count, valid pointer */
            0, NULL,
            /* number defaults, default value pointer */
            1, default_report_filename,
            report_filename, sizeof(report_filename), 0);

    /* PARAM 5: Read the residual report flag.
       Implementing as boolean, requires no further validation */
    int default_residual_report_flag = 1;
    IAS_PARM_ADD_BOOLEAN(parms, REPORT_RES_FLAG, "Flag to indicate if "
            "application should write the valid GCP residuals measurements to "
            "the output geometric report file or not: "
            "1 to write the data to the file, 0 don't write the data",
            IAS_PARM_REQUIRED, IAS_PARM_NOT_ARRAY,
            &default_residual_report_flag,
            &residual_report_flag, sizeof(residual_report_flag), 0);

    /* PARAM 6: Read the flag for writing the residual stats to
       characterization DB.
       The way this standard parameter works, it will default the parameter to
       1 (flag on) if this parameter is not provided.
       Also, implements it as a boolean type, so no additional validation
       needed. */
    IAS_PARM_STORE_CHARACTERIZATION_IN_DB(parms, IAS_PARM_OPTIONAL,
            &store_stats_in_db, sizeof(store_stats_in_db), 0);

    /* PARAM 7: Read the flag indicating whether successful geometric 
       processing is required.  If it isn't required, upon failure some 
       threshold checks may result in a success exit. */ 
    int default_geometric_required_flag = 1;
    IAS_PARM_ADD_INT(parms, GEOMETRIC_REQUIRED_FLAG, "Flag which indicates "
            "successful geometric processing is required",
            IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
            /* has a limited range of 0-1 */
            1, 0, 1,
            /* default is 1 -- required */
            1, &default_geometric_required_flag, &geometric_required_flag,
            sizeof(geometric_required_flag), 1);


    /* --------------------------------------------------------------------- */
    /* -------------- Read the application parameter file ------------------ */
    /* Need to do this now, since need the processing pass for the parameters
       from the OMF. */

    /* Calculate the number of items in the application parameter file */
    count = IAS_PARM_GET_TABLE_SIZE(parms);

    /* Call the ias_parm_provide_help routine to provide help for the
       parameters if requested; also verifies the ODL file exists */
    status = ias_parm_provide_help(parm_file, parms, count,
            IAS_INPUT_PARAMETERS);
    if (status)
    {
        /* either help was provided or an error occurred */
        if (status == 1)
        {
            /* If the status of the provide help is 1 the user asked for help.
               Since the parameter file name contains the help request, copy
               it to the OMF file name so the next call will ask for the same
               help */
            strcpy(omf_filename, parm_file);
        }
        else
        {
            IAS_LOG_ERROR("Providing application parameter file help");
            return ERROR;
        }
    }
    else
    {
        /* Read the parameter file */
        status = ias_parm_read(parm_file, parms, count);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading parameters from: %s", parm_file);
            return ERROR;
        }

        /* Build the OMF file name */
        sprintf(omf_filename, "%s.omf", work_order_id);
    }

    /* --------------------------------------------------------------------- */
    /* -------------- Table definition for things from the OMF ------------- */

    IAS_PARM_DECLARE_TABLE(omf_parms, 6);

    /* PARAM 1: Read the satellite number to init satellite attributes. */
    IAS_PARM_SATELLITE(omf_parms, &satellite_num, sizeof(satellite_num), 0);

    /* PARAM 2: Read the L0R image file name. */
    IAS_PARM_L0R_FILENAME(omf_parms, IAS_PARM_OPTIONAL, l0r_image_name,
            sizeof(l0r_image_name), 0);

    /* PARAM 3: Read the L1G image file name. */
    IAS_PARM_L1G_FILENAME_FROM_PASS(omf_parms, IAS_PARM_REQUIRED,
            l1g_image_name, sizeof(l1g_image_name), 0, processing_pass);

    /* PARAM 4: Read the CPF name. */
    IAS_PARM_CPF(omf_parms, IAS_PARM_REQUIRED, cpf_filename,
            sizeof(cpf_filename), 0);

    /* PARAM 5: Get the name of the measured GCP file name. */
    IAS_PARM_MEASURED_GCP_FILENAME_FROM_PASS(omf_parms, measured_gcp_filename,
            sizeof(measured_gcp_filename), 0, processing_pass);

    /* PARAM 6: Get the flag for success of precision solution step during the
       previous processing pass. */
    IAS_PARM_PRECISION_SUCCEEDED_FLAG_FROM_PASS(omf_parms,
            &precision_succeeded_flag, sizeof(precision_succeeded_flag), 0,
            processing_pass - 1);

    /* --------------------------------------------------------------------- */
    /* ------------- Read the parameters defined from the OMF -------------- */
    /* See if the user wanted help or a template, this time with the OMF file */
    count = IAS_PARM_GET_TABLE_SIZE(omf_parms);
    status = ias_parm_provide_help(omf_filename, omf_parms, count,
            IAS_OMF_PARAMETERS);
    if (status)
    {
        if (status == 1) /* help was successful */
            exit(EXIT_SUCCESS);
        else
        {
            IAS_LOG_ERROR("Providing OMF file help");
            return ERROR;
        }
    }

    /* Read the OMF file */
    status = ias_parm_read(omf_filename, omf_parms, count);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from OMF file: %s", omf_filename);
        return ERROR;
    }

    /* --------------------------------------------------------------------- */
    /* -------------- Do some validation of values retrieved --------------- */

    /* Exit if the L1G filename is empty */
    if (strcmp(l1g_image_name, "") == 0)
    {
        IAS_LOG_ERROR("Invalid L1G filename in the parameter file");
        return ERROR;
    }

    /* Exit if the CPF filename is empty */
    if (strcmp(cpf_filename, "") == 0)
    {
        IAS_LOG_ERROR("Invalid CPF filename in the parameter file");
        return ERROR;
    }

    /* Exit if the GCP measurements filename is empty */
    if (strcmp(measured_gcp_filename, "") == 0)
    {
        IAS_LOG_ERROR("Invalid GCP measurements filename in the parameter "
                "file");
        return ERROR;
    }

    return SUCCESS;
}

/***********************************************************************
    Functions used to access the parameter file values
***********************************************************************/
const char *get_cpf_filename(void)
{
    return cpf_filename;
}

const char *get_l0r_image_name(void)
{
    return l0r_image_name;
}

const char *get_l1g_image_name(void)
{
    return l1g_image_name;
}

const char *get_measured_gcp_filename(void)
{
    return measured_gcp_filename;
}

double get_outlier_threshold(void)
{
    return outlier_threshold;
}

const char *get_report_filename(void)
{
    return report_filename;
}

int get_residual_report_flag(void)
{
    return residual_report_flag;
}

int get_satellite_num(void)
{
    return satellite_num;
}

const char *get_work_order_id(void)
{
    return work_order_id;
}

int get_write_stats_to_db_flag(void)
{
    return store_stats_in_db;
}
const char *get_omf_filename(void)
{
    return omf_filename;
}

int get_precision_succeeded_flag(void)
{
    return precision_succeeded_flag;
}

int get_geometric_required_flag(void)
{
    return geometric_required_flag;
}

