/*****************************************************************************
Name:           get_l1g_header_data

Purpose:
    Gets additional information needed from the L1G metadata.
    Makes the data available through accessor functions.

Returns:
SUCCESS  L1G metadata was read.
ERROR    Error reading L1G metadata.

******************************************************************************
                        Property of the U.S. Government
                            USGS EROS Data Center
*******************************************************************************/
#include <string.h>
#include "ias_types.h"
#include "ias_const.h"
#include "ias_satellite_attributes.h"
#include "ias_logging.h"
#include "ias_l1g.h"
#include "geometric_char.h"

static double projection_distance_y[IAS_MAX_NBANDS];/* Per-band (index) pixel */
static double projection_distance_x[IAS_MAX_NBANDS];/* sizes in projection    */
                                                    /* coordinates            */
static int wrs_path;
static int wrs_row;
static double roll_angle;                           /* off-nadir collections */
static char spacecraft[IAS_L1G_SPACECRAFT_SIZE];
static char collection_type[IAS_COLLECT_TYPE_SIZE]; /* earth */
static char capture_date[IAS_L1G_CAPTURE_DATE_SIZE];/* MM/DD/YYYY format */

int get_l1g_header_data()
{
    L1GIO *img_ptr;                 /* L1G image pointer */
    int band_index;                 /* Band index */
    int band_number;                /* 1-based band number */
    int status;                     /* Status returned from functions */
    int satellite_number_bands;     /* Number of normal bands for satellite */
    IAS_SATELLITE_ID satellite_id;
    IAS_L1G_BAND_METADATA band_metadata;
    IAS_L1G_FILE_METADATA image_metadata;

    /* Initialize the satellite attributes */
    satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(
            get_satellite_num());
    status = ias_sat_attr_initialize(satellite_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the satellite attributes library");
        return ERROR;
    }

    /* Find the number of normal bands for the satellite */
    satellite_number_bands = ias_sat_attr_get_normal_band_count();

    /* --------------------------------------------------------------------- */
    /* ------ Read the L1G to retrieve essential values for processing ----- */
    img_ptr = ias_l1g_open_image(get_l1g_image_name(), IAS_READ);
    if (img_ptr == NULL)
    {
        IAS_LOG_ERROR("Image %s cannot be opened", get_l1g_image_name());
        return ERROR;
    }

    /* Check the image metadata to ensure provided a terrain-corrected image */
    status = ias_l1g_get_file_metadata(img_ptr, &image_metadata);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving image metadata for %s", get_l1g_image_name());
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }
    if(strcmp(image_metadata.correction_type, "L1T") != 0)
    {
        IAS_LOG_ERROR("Image %s is not terrain corrected",
                get_l1g_image_name());
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    /* Loop through all bands, reading pixel sizes for each band */
    for (band_index = 0; band_index < satellite_number_bands; band_index++)
    {
        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Converting band index %d to band number",
                    band_index);
            ias_l1g_close_image(img_ptr);
            return ERROR;
        }
        if (ias_l1g_is_band_present(img_ptr, band_number) != 1)
        {
            /* Should not require all bands metadata to be present since, at
               this point we do not yet know the search band (from the GCP
               measurements file). So just populate the non-present bands
               projection distance values with zero. When the GCP offset scaling
               is done, we can use this non-valid projection distance value to
               determine if a needed value was not available from the L1G. */
            projection_distance_y[band_index] = 0;
            projection_distance_x[band_index] = 0;
        }
        else
        {
            /* The band is in the L1G, so store the needed metadata */
            /* Get image band metadata for this band */
            status = ias_l1g_get_band_metadata(img_ptr, band_number,
                    &band_metadata);
            if (status != SUCCESS)
            {
                /* Again, don't error out just yet since we don't know yet if
                   we will need the band metadata for this band yet or not.
                   Store the value as a zero and later the code will validate
                   this value before using it and handling the error case at
                   the time of use. */
                projection_distance_y[band_index] = 0;
                projection_distance_x[band_index] = 0;
            }
            else
            {
                projection_distance_y[band_index]
                    = band_metadata.projection_distance_y;
                projection_distance_x[band_index]
                    = band_metadata.projection_distance_x; 
            }
        }
    }

    /* Close the L1G file */
    ias_l1g_close_image(img_ptr);
    img_ptr = NULL;

    /* Extract data needed for reporting */
    wrs_path = image_metadata.wrs_path;
    wrs_row = image_metadata.wrs_row;
    roll_angle = image_metadata.roll_angle;
    strcpy(spacecraft, image_metadata.spacecraft);
    strcpy(collection_type, image_metadata.collection_type);
    strcpy(capture_date, image_metadata.capture_date);

    return SUCCESS;
}

const char *get_capture_date(void)
{
    return capture_date;
}

const char *get_collection_type(void)
{
    return collection_type;
}

double get_projection_distance_x(band_index)
{
    return projection_distance_x[band_index];
}

double get_projection_distance_y(band_index)
{
    return projection_distance_y[band_index];
}

double get_roll_angle(void)
{
    return roll_angle;
}

const char *get_spacecraft(void)
{
    return spacecraft;
}

int get_wrs_path(void)
{
    return wrs_path;
}

int get_wrs_row(void)
{
    return wrs_row;
}
