/*****************************************************************************
Name:           remove_outliers

Purpose:
    Removes outliers from the GCP results.

Returns:
SUCCESS  Outliers were removed.
ERROR    Error removing outliers.

Notes:
    Memory allocated within this function must be freed by the calling routine.
    If this routine errors, internally allocated memory is freed before
    returning.

******************************************************************************
                        Property of the U.S. Government
                            USGS EROS Data Center
*******************************************************************************/
#include <stdlib.h> 
#include <math.h>
#include <string.h>
#include "ias_const.h"
#include "ias_gcp.h"
#include "ias_logging.h"
#include "ias_math.h"
#include "geometric_char.h"

/* Local routine prototypes */
void copy_gcp_results
(
    IAS_GCP_RESULTS *gcp_results_from,
    IAS_GCP_RESULTS *gcp_results_to
);

void copy_scaled_offsets 
(
    SCALED_FIT_OFFSETS *scaled_offsets_from,
    SCALED_FIT_OFFSETS *scaled_offsets_to
);

int remove_outliers 
(
    IAS_GCP_RESULTS **gcp_results,        /* I/O: Pointer to GCP measurement
                                                  results for outlier removal */
    SCALED_FIT_OFFSETS **scaled_offsets,  /* I/O: Pointer to scaled fit offset
                                                  results; modified to keep in
                                                  synch with the gcp_results as
                                                  outliers are removed */
    int *num_gcp,       /* I/O: Number of GCP measurement records on input,
                                number of non-outlier GCPs on output */ 
    double **gcp_lat,   /* O: GCP residual latitude data */
    double **gcp_lon,   /* O: GCP residual longitude data */
    double **along_res, /* O: Along track residual data */
    double **across_res /* O: Across track residual data */
)
{
    int mem_alloc_size;             /* Used for memory allocations */
    int valid_gcp_count;            /* Number of valid (non-outlier) GCPs */
    double *line_data;              /* Consolidated line data */
    double *sample_data;            /* Consolidated sample data */
    int gcp_index;                  /* Index into GCP structures */
    double line_mean;               /* Mean of line offsets */
    double sample_mean;             /* Mean of sample offsets */
    double line_std_dev;            /* Standard deviation of line offsets */
    double sample_std_dev;          /* Standard deviation of sample offsets */
    int outlier_found = 1;          /* Flag for outlier detection */
    double t_dist;                  /* T-distribution value for the given
                                       tolerance */
    double line_max_dev_allowed;    /* Maximum allowed deviation from mean */
    double sample_max_dev_allowed;  /* Maximum allowed deviation from mean */
    double max_line;                /* Mean offset value for farthest point
                                       from mean */
    double max_sample;              /* Mean offset value for farthest point
                                       from mean */
    int max_line_index;             /* Index of max_line */
    int max_sample_index;           /* Index of max_sample */
    double difference_from_mean;    /* Absolute magnitude of the difference
                                       from mean */
    int outlier_index;              /* Index of the outlier */
    double swap_temp;               /* Temporary holder during swap */
    IAS_GCP_RESULTS temp_gcp_results;       /* Temporary holder of GCP results
                                               during swap */
    SCALED_FIT_OFFSETS temp_scaled_offsets; /* Temporary holder of scaled 
                                               offsets during swap */
    int status;

    /* Allocate memory for consolidated line and sample data */
    mem_alloc_size = *num_gcp * sizeof(double);
    line_data = calloc(*num_gcp, sizeof(double));
    if (line_data == NULL)
    {
        IAS_LOG_ERROR("Allocating %d bytes for line data", mem_alloc_size);
        return ERROR;
    }

    sample_data = calloc(*num_gcp, sizeof(double));
    if (sample_data == NULL)
    {
        IAS_LOG_ERROR("Allocating %d bytes for sample data", mem_alloc_size);
        free(line_data);
        return ERROR;
    }

    /* Consolidate the GCP line and sample data. The purpose is to put
       the data in a form usable by the statistics functions. */
    for (gcp_index = 0; gcp_index < *num_gcp; gcp_index++)
    {
        line_data[gcp_index] = (*scaled_offsets)[gcp_index].line_offset;
        sample_data[gcp_index] = (*scaled_offsets)[gcp_index].sample_offset;
    }

    /* Initialize number of valid GCPs to the total number of GCPs */
    valid_gcp_count = *num_gcp;

    while (outlier_found)
    {
        /* Indicate that no outliers were found yet during this outlier
           rejection iteration. Initialize outlier flag to false */
        outlier_found = 0;

        /* Calculate the mean of the valid line offsets */
        status = ias_math_compute_mean(line_data, valid_gcp_count, &line_mean); 
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating the mean line offset");
            free(line_data);
            free(sample_data);
            return ERROR;
        }

        /* Calculate the mean of the valid sample offsets */
        status = ias_math_compute_mean(sample_data, valid_gcp_count,
                &sample_mean); 
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating the mean sample offset");
            free(line_data);
            free(sample_data);
            return ERROR;
        }

        /* Calculate the standard deviation of the valid line offsets */
        status = ias_math_compute_stdev(line_data, valid_gcp_count, line_mean, 
                &line_std_dev);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating the standard deviation of valid "
                    "line offsets");
            free(line_data);
            free(sample_data);
            return ERROR;
        }

        /* Calculate the standard deviation of the valid sample offsets */
        status = ias_math_compute_stdev(sample_data, valid_gcp_count,
                sample_mean, &sample_std_dev);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating the standard deviation of valid "
                    "sample offsets");
            free(line_data);
            free(sample_data);
            return ERROR;
        }

        /* Get the t distribution value for the given tolerance and
           degrees of freedom */
        t_dist = ias_math_compute_t_confidence(get_outlier_threshold(), 
                valid_gcp_count - 1);

        /* For line and sample, calculate the largest deviation from the mean 
           allowable for the specified degree of freedom and probability 
           threshold */
        line_max_dev_allowed = line_std_dev * t_dist;
        sample_max_dev_allowed = sample_std_dev * t_dist;

        /* For line and sample, find the valid data point value and location
           that is farthest from the mean */
        max_line = 0;
        max_sample = 0;
        max_line_index = 0;
        max_sample_index = 0;

        for (gcp_index = 0; gcp_index < valid_gcp_count; gcp_index++)
        {
            difference_from_mean = fabs(line_data[gcp_index] - line_mean);
            if (difference_from_mean > max_line)
            {
                max_line = difference_from_mean; 
                max_line_index = gcp_index;
            }

            difference_from_mean = fabs(sample_data[gcp_index] - sample_mean);
            if (fabs(sample_data[gcp_index] - sample_mean) > max_sample)
            {
                max_sample = difference_from_mean; 
                max_sample_index = gcp_index;
            }
        }

        /* Check if there's an outlier. If there is, process it and next
           loop pass check again for another outlier. */
        if ((max_line > line_max_dev_allowed)
                || (max_sample > sample_max_dev_allowed))
        {
            /* At least one outlier has been found */
            outlier_found = 1;

            /* Two outliers may have been detected. In that case decide which
               outlier should be marked. The if condition is equivalent to
               the following from the ADD, without division by 0 potential: 
               if ((max_sample / sample_std_dev) > (max_line / line_std_dev)) */
            if ((max_sample * line_std_dev) > (max_line * sample_std_dev))
            {
                /* The GCP flagged by sample data is marked as an outlier */
                outlier_index = max_sample_index;
            }
            else
            {
                /* The GCP flagged by line data is marked as an outlier */
                outlier_index = max_line_index;
            }

            /* Outliers are swapped to the end of the data arrays */
            swap_temp = line_data[outlier_index];
            line_data[outlier_index] = line_data[valid_gcp_count - 1];
            line_data[valid_gcp_count - 1] = swap_temp;

            swap_temp = sample_data[outlier_index];
            sample_data[outlier_index] = sample_data[valid_gcp_count - 1];
            sample_data[valid_gcp_count - 1] = swap_temp;

            /* The GCP results are kept in sync with this swapping process */ 
            copy_gcp_results(&(*gcp_results)[outlier_index], &temp_gcp_results);
            copy_gcp_results(&(*gcp_results)[valid_gcp_count - 1], 
                    &(*gcp_results)[outlier_index]);
            copy_gcp_results(&temp_gcp_results, 
                    &(*gcp_results)[valid_gcp_count - 1]);

            /* The scaled offsets are also kept in sync */ 
            copy_scaled_offsets(&(*scaled_offsets)[outlier_index], 
                    &temp_scaled_offsets);
            copy_scaled_offsets(&(*scaled_offsets)[valid_gcp_count - 1], 
                    &(*scaled_offsets)[outlier_index]);
            copy_scaled_offsets(&temp_scaled_offsets, 
                    &(*scaled_offsets)[valid_gcp_count - 1]);

            /* Adjust valid GCP count to reflect the latest outlier */
            valid_gcp_count--;
        }
    }

    /* Release the line and sample data copies used for outlier detection */
    free(line_data);
    free(sample_data);

    /* Allocate space for the data to be used to compute the statistics for
       the GCP residuals */
    mem_alloc_size = valid_gcp_count * sizeof(double);
    *gcp_lat = calloc(valid_gcp_count, sizeof(double));
    if (*gcp_lat == NULL)
    {
        IAS_LOG_ERROR("Allocating memory of size %d", mem_alloc_size);
        return ERROR;
    }
    *gcp_lon = calloc(valid_gcp_count, sizeof(double));
    if (*gcp_lon == NULL)
    {
        IAS_LOG_ERROR("Allocating memory of size %d", mem_alloc_size);
        free(*gcp_lat);
        *gcp_lat = NULL;
        return ERROR;
    }
    *along_res = calloc(valid_gcp_count, sizeof(double));
    if (*along_res == NULL)
    {
        IAS_LOG_ERROR("Allocating memory of size %d", mem_alloc_size);
        free(*gcp_lat);
        free(*gcp_lon);
        *gcp_lat = NULL;
        *gcp_lon = NULL;
        return ERROR;
    }
    *across_res = calloc(valid_gcp_count, sizeof(double));
    if (*across_res == NULL)
    {
        IAS_LOG_ERROR("Allocating memory of size %d", mem_alloc_size);
        free(*gcp_lat);
        free(*gcp_lon);
        free(*along_res);
        *gcp_lat = NULL;
        *gcp_lon = NULL;
        *along_res = NULL;
        return ERROR;
    }

    /* Fill the buffers with the valid data (not outliers) */
    for (gcp_index = 0; gcp_index < valid_gcp_count; gcp_index++)
    {
        (*gcp_lat)[gcp_index] = (*gcp_results)[gcp_index].latitude;
        (*gcp_lon)[gcp_index] = (*gcp_results)[gcp_index].longitude;

        /* Along-track equates to lines while across-track equates to samples */
        (*along_res)[gcp_index] = (*scaled_offsets)[gcp_index].line_offset; 
        (*across_res)[gcp_index] = (*scaled_offsets)[gcp_index].sample_offset;  
    }

    /* Report back on the number of valid GCPs */
    *num_gcp = valid_gcp_count;

    return SUCCESS;
}

/*****************************************************************************
NAME:           copy_gcp_results

PURPOSE:
        Makes a copy of an IAS_GCP_RESULTS structure

RETURNS:
        nothing
******************************************************************************/
void copy_gcp_results 
(
    IAS_GCP_RESULTS *gcp_results_from, /* I: Pointer to GCP measurement source 
                                             results */
    IAS_GCP_RESULTS *gcp_results_to    /* O: Pointer ptr to GCP measurement 
                                             destination */
)
{
    strcpy(gcp_results_to->point_id, gcp_results_from->point_id);
    gcp_results_to->reference_line = gcp_results_from->reference_line;
    gcp_results_to->reference_sample = gcp_results_from->reference_sample;
    gcp_results_to->latitude = gcp_results_from->latitude;
    gcp_results_to->longitude = gcp_results_from->longitude;
    gcp_results_to->elevation = gcp_results_from->elevation;
    gcp_results_to->gcp_position.x = gcp_results_from->gcp_position.x;
    gcp_results_to->gcp_position.y = gcp_results_from->gcp_position.y;
    gcp_results_to->gcp_position.z = gcp_results_from->gcp_position.z;
    gcp_results_to->predicted_search_line = 
        gcp_results_from->predicted_search_line;
    gcp_results_to->predicted_search_sample = 
        gcp_results_from->predicted_search_sample;
    gcp_results_to->fit_line_offset = gcp_results_from->fit_line_offset;
    gcp_results_to->fit_sample_offset = gcp_results_from->fit_sample_offset;
    gcp_results_to->accept_flag = gcp_results_from->accept_flag;
    strcpy(gcp_results_to->reference_name, gcp_results_from->reference_name);
    gcp_results_to->reference_band = gcp_results_from->reference_band;
    strcpy(gcp_results_to->search_name, gcp_results_from->search_name);
    gcp_results_to->search_band = gcp_results_from->search_band;
    strcpy(gcp_results_to->chip_source, gcp_results_from->chip_source);
    gcp_results_to->reference_sca = gcp_results_from->reference_sca;
    gcp_results_to->search_sca = gcp_results_from->search_sca;
}

/*****************************************************************************
NAME:           copy_scaled_offsets

PURPOSE:
        Makes a copy of a SCALED_FIT_OFFSETS structure

RETURNS:
        nothing
******************************************************************************/
void copy_scaled_offsets 
(
    SCALED_FIT_OFFSETS *scaled_offsets_from, /* I: Pointer ptr to scaled offset
                                                   source */
    SCALED_FIT_OFFSETS *scaled_offsets_to    /* O: Pointer to scaled offset 
                                                   destination */
)
{
    scaled_offsets_to->line_offset = scaled_offsets_from->line_offset;
    scaled_offsets_to->sample_offset = scaled_offsets_from->sample_offset;
}
