/*****************************************************************************
Name:           geometric_test

Purpose:
    The regression test for Geometric Accuracy Assessment. Tests at the
    application level with an SCA-combined and an SCA-separated image to
    verify output report files are as expected (as the reports contain the
    data of all the statistics that are the main data this application
    generates).

******************************************************************************
                        Property of the U.S. Government
                            USGS EROS Data Center
*******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include "../geometric_char.h"
#include "ias_const.h"           /* For GET_FILE_NAME macro */
#include "cgreen/cgreen.h"

static char *unit_test_dir = NULL;

Ensure test_sca_combined(void);
Ensure test_sca_separated(void);
Ensure test_database_store(void);

int main(int argc, char *argv[])
{
    /* Get the ENABLE_DB_TESTS environment variable */
    char *enable_db_tests = getenv("ENABLE_DB_TESTS");

    /* Construct the test suite */
    TestSuite *suite = create_test_suite();

    add_test(suite, test_sca_combined);
    add_test(suite, test_sca_separated);

    /* If database support is enabled, add the database test to the
       suite */
    if (enable_db_tests)
    {
        add_test(suite, test_database_store);
    }
    else
    {
        printf("\n");
        printf("%s:  Database tests are disabled.\n", __FILE__);
        printf("\n");
    }

    unit_test_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (setenv("IAS_VERSION_OVERRIDE", "Y", 0) != 0)
    {
        printf("setenv failed / insufficient space in the environment\n");
        exit(EXIT_FAILURE);
    }
    if (unit_test_dir == NULL)
    {
        printf("The IAS_UNIT_TEST_DATA_DIR environment variable is not set\n");
        exit(EXIT_FAILURE);
    }

    if (argc > 1)
        return run_single_test(suite, argv[1], create_text_reporter());

    return run_test_suite(suite, create_text_reporter());
}

Ensure test_sca_combined()
{
    int status;
    char command[PATH_MAX * 2];

    char param_comb_filename[PATH_MAX] = "parameters_comb.odl";
    char omf_comb_filename_orig[PATH_MAX] = "WO1234.omf";
    char omf_comb_filename_mod[PATH_MAX] = "WO1234_comb.omf";
    char output_comb_filename_control[PATH_MAX];
    char output_comb_filename_generated[PATH_MAX] = "geometric_comb.geo";

    /*------------------------------------------------------------*/
    /* Test execution of Geometric on the SCA-combined test image */
    /*------------------------------------------------------------*/

    /* First get the OMF file ready by replacing unit test data directory
       environment variables with the actual path */
    status = snprintf(command, sizeof(command),
            "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' %s > %s", unit_test_dir,
            omf_comb_filename_orig, omf_comb_filename_mod);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Error preparing OMF file %s for unit "
                "test data directory path setup", omf_comb_filename_orig);
        return;
    }

    /* Now run Geometric to test it on this image */
    status = snprintf(command, sizeof(command), "../geometric %s",
            param_comb_filename);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Geometric Accuracy Assessment failed");
        return;
    }

    /* Now compare the output report file with the one previously generated */
    /* Prepare the control file name to compare against */
    status = snprintf(output_comb_filename_control, PATH_MAX,
            "%s/ias_base/gps/geometric/combined/geometric_comb.geo",
            unit_test_dir);
    if (status < 0 || status >= PATH_MAX)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }

    status = snprintf(command, sizeof(command),
            "sed '/Pixel Size/,$!'d %s | diff -q - %s",
            output_comb_filename_generated, output_comb_filename_control);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found with Geometric "
                "assessment output. Review %s and %s for details",
                output_comb_filename_control, output_comb_filename_generated);
    }
    else
    {
        /* This test case completed successfully, so cleanup all
           generated files */
        remove(omf_comb_filename_mod);
        remove(output_comb_filename_generated);
    }
}

Ensure test_sca_separated()
{
    int status;
    char command[PATH_MAX * 2];

    char param_sep_filename[PATH_MAX] = "parameters_sep.odl";
    char omf_sep_filename_orig[PATH_MAX] = "WO1235.omf";
    char omf_sep_filename_mod[PATH_MAX] = "WO1235_sep.omf";
    char output_sep_filename_control[PATH_MAX];
    char output_sep_filename_generated[PATH_MAX] = "geometric_sep.geo";

    /*-------------------------------------------------------------*/
    /* Test execution of Geometric on the SCA-separated test image */
    /*-------------------------------------------------------------*/

    /* First get the OMF file ready by replacing unit test data directory
       environment variables with the actual path */
    status = snprintf(command, sizeof(command),
            "sed 's|$IAS_UNIT_TEST_DATA_DIR|"
            "%s|' %s > %s", unit_test_dir, omf_sep_filename_orig,
            omf_sep_filename_mod);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Error preparing OMF file %s for unit "
                "test data directory path setup", omf_sep_filename_orig);
        return;
    }

    /* Now run Geometric to test it on this image */
    status = snprintf(command, sizeof(command),
        "../geometric %s", param_sep_filename);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Geometric Accuracy Assessment failed");
        return;
    }

    /* Now compare the output report file with the one previously generated */
    /* Prepare the control file name to compare against */
    status = snprintf(output_sep_filename_control, PATH_MAX,
            "%s/ias_base/gps/geometric/separated/geometric_sep.geo",
            unit_test_dir);
    if (status < 0 || status >= PATH_MAX)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }

    status = snprintf(command, sizeof(command),
            "sed '/Pixel Size/,$!'d %s | diff -q - %s",
            output_sep_filename_generated, output_sep_filename_control);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found with Geometric "
                "assessment output. Review %s and %s for details",
                output_sep_filename_control, output_sep_filename_generated);
    }
    else
    {
        /* This test case completed successfully, so cleanup all
           generated files */
        remove(omf_sep_filename_mod);
        remove(output_sep_filename_generated);
    }
}

Ensure test_database_store()
{
    int status;
    char command[PATH_MAX * 2];

    char param_comb_filename[PATH_MAX] = "parameters_db.odl";
    char omf_comb_filename_orig[PATH_MAX] = "WO1234.omf";
    char omf_comb_filename_mod[PATH_MAX] = "WO1234_db.omf";
    char output_comb_filename_generated[PATH_MAX] = "geometric_db.geo";

    /*------------------------------------------------------------*/
    /* Test execution of Geometric on the SCA-combined test image */
    /*------------------------------------------------------------*/

    /* First get the OMF file ready by replacing unit test data directory
       environment variables with the actual path */
    status = snprintf(command, sizeof(command),
            "sed 's|$IAS_UNIT_TEST_DATA_DIR|"
            "%s|' %s > %s", unit_test_dir, omf_comb_filename_orig,
            omf_comb_filename_mod);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Error preparing OMF file %s for unit "
                "test data directory path setup", omf_comb_filename_orig);
        return;
    }

    /* Prepare the database. A row containing information for the WO needs
       to be added if it doesn't exist. */
    status = system("./check_or_insert_woid.csh WO1234_db");
    assert_true_with_message(status == 0, "Error preparing database, "
            "unable to insert work_order_id WO1234_db");
    if (status != 0)
        return;

    /* Now run Geometric to test it on this image */
    status = snprintf(command, sizeof(command), "../geometric %s",
            param_comb_filename);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Geometric Accuracy Assessment failed");
        return;
    }

    /* Check the precision parameters */
    status = system("./retrieve_stats_from_db.csh WO1234_db_geometric");
    assert_true_with_message(status == 0, "Error checking geometric stats");
    if (status != 0)
        return;
    else
    {
        /* This test case completed successfully, so cleanup all
           generated files */
        remove(omf_comb_filename_mod);
        remove(output_comb_filename_generated);
    }
}
