/******************************************************************************
Name:           write_residual_statistics_report

Purpose:
    Writes the GCP residual statistics to the ASCII output file.  

Returns:
SUCCESS   Successfully wrote to the file.
ERROR     Error occurred; statistics were not written.

*******************************************************************************
                        Property of the U.S. Government
                            USGS EROS Data Center
******************************************************************************/
#include <stdio.h>
#include <string.h>
#include "ias_const.h"
#include "ias_geo.h"
#include "ias_gcp.h"
#include "ias_miscellaneous.h"
#include "ias_satellite_attributes.h"
#include "ias_logging.h"
#include "geometric_char.h"

int write_residual_statistics_report
(
    const GCP_RES_STATS *stats,     /* I: GCP residual statistics */
    const IAS_GCP_RESULTS *gcp_results,       /* I: Original GCP records */
    const SCALED_FIT_OFFSETS *scaled_offsets  /* I: Scaled fit offset results */
)
{
    FILE *stats_fptr;                   /* GCP residual statistics file name */
    int status;                         /* Status return of the function */
    int gcp_index;                      /* Index into GCP data */
    int band_index;                     /* Band index */
    IAS_REPORT_HEADER *header = NULL;   /* Report header structure */

    /* Write the standard output header */
    header = ias_misc_initialize_gps_report_header(get_l1g_image_name(),
            get_l0r_image_name(), get_work_order_id(), "GEOMETRIC");
    if (header == NULL)
    {
        IAS_LOG_ERROR("Initializing the standard report header");
        return ERROR;
    }

    /* Open the file now to start this report header */
    stats_fptr = fopen(get_report_filename(), "w");
    if (stats_fptr == NULL)
    {
        IAS_LOG_ERROR("Opening residual statistics output file: %s", 
                get_report_filename());
        ias_misc_free_report_header(header);
        return ERROR;
    }

    status = ias_misc_write_gps_report_header(stats_fptr, header);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the standard header for this report");
        ias_misc_free_report_header(header);
        fclose(stats_fptr);
        return ERROR;
    }

    ias_misc_free_report_header(header);
    header = NULL;

    /* Get the search band information to access the pixel size info for it */
    band_index = ias_sat_attr_convert_band_number_to_index(
            gcp_results[0].search_band);
    if (band_index == ERROR)
    {
        IAS_LOG_ERROR("Getting band index from band number %d",
                gcp_results[0].search_band);
        fclose(stats_fptr);
        return ERROR;
    }

    /* Print out additional information called for by the ADD.
       FYI - Using the projection distance without validating it here
       knowingly, as it was checked during GCP offsets scaling. */
    status = fprintf(stats_fptr, "Pixel Size:           %4.2f,%4.2f\n\n\n",
            get_projection_distance_x(band_index),
            get_projection_distance_y(band_index));
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing additional header information");
        fclose(stats_fptr);
        return ERROR;
    }

    /* Write the statistics column headers followed by the data.
       Note: GCP source (chip_source) is the same for all GCPs, so just
       use the 1st one */
    status = fprintf(stats_fptr,
            "GCP     # Valid  Mean Lat.     Mean Lon.\n"
            "Source  Points   (degrees)     (degrees)\n"
            "------  -------  ------------  ------------\n"
            "   %3s     %4d  %12.6f  %12.6f\n\n\n",
            gcp_results[0].chip_source, stats->num_points, stats->mean_lat,
            stats->mean_lon);

    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the residual statistics file data");
        fclose(stats_fptr);
        return ERROR;
    }

    status = fprintf(stats_fptr,
            "                     Residual Statistics (meters)\n"
            "                     ----------------------------\n\n"
            "      Mean                 RMSE               Std. Dev.        "
            "Corr. Coeff.\nLine      Sample     Line      Sample     Line  "
            "    Sample      Line  vs.\nOffset    Offset     Offset    "
            "Offset     Offset    Offset      Sample\n--------  --------   "
            "--------  --------   --------  --------    ----------\n"
            "%8.3f  %8.3f   %8.3f  %8.3f   %8.3f  %8.3f     %9.4f\n\n\n",
            stats->mean_along_res, stats->mean_across_res,
            stats->rmse_along_res, stats->rmse_across_res,
            stats->sd_along_res, stats->sd_across_res,
            stats->corr_coeff);

    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the residual statistics file data");
        fclose(stats_fptr);
        return ERROR;
    }

    /* If requested, print the GCP residual measurements for each valid GCP */
    if (get_residual_report_flag())
    {
        status = fprintf(stats_fptr,
                "                                  GCP Measurements\n"
                "                                  ----------------\n\n"
                "Point      Lat.         Lon.         Height       Line Offset"
                "  Sample Offset  Band  SCA\n"
                "ID         (degrees)    (degrees)    (meters)     (meters)"
                "     (meters)       Num   Num\n"
                "---------  -----------  -----------  -----------  -----------"
                "  -----------    ----  ---\n");

        if (status < 0)
        {
            IAS_LOG_ERROR("Writing the residual statistics file data");
            fclose(stats_fptr);
            return ERROR;
        }

        for (gcp_index = 0; gcp_index < stats->num_points; gcp_index++)
        {
            status = fprintf(stats_fptr,
                    "%s  %11.6f  %11.6f  %11.6f  %11.6f  %11.6f      "
                    "%02d   %02d\n",
                    gcp_results[gcp_index].point_id, 
                    gcp_results[gcp_index].latitude, 
                    gcp_results[gcp_index].longitude, 
                    gcp_results[gcp_index].elevation, 
                    scaled_offsets[gcp_index].line_offset, 
                    scaled_offsets[gcp_index].sample_offset, 
                    gcp_results[gcp_index].search_band, 
                    gcp_results[gcp_index].search_sca);
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing the residual statistics file data");
                fclose(stats_fptr);
                return ERROR;
            }
        }
    }

    /* Close the file */
    if (fclose(stats_fptr) == EOF)
    {
        IAS_LOG_ERROR("Closing the residual statistics output file: %s",
                get_report_filename());
        return ERROR;
    }

    return SUCCESS;
}
