/*****************************************************************************
Name:           write_residual_statistics_to_db

Purpose:
    Writes the GCP residual statistics to the database.

Returns:
SUCCESS   Successfully wrote to the database.
ERROR     Error occurred; statistics were not written.

******************************************************************************
                        Property of the U.S. Government
                            USGS EROS Data Center
*******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_db.h"
#include "ias_db_insert.h"
#include "ias_db_table.h"
#include "ias_db_get_connect_info.h"
#include "ias_miscellaneous.h"
#include "geometric_char.h"

/* Define here since used multiple times, to help with bug avoidance, and to
   reduce maintenance time if the DB table name would ever change. */
#define TABLE_NAME "geometric_accuracy"

int write_residual_statistics_to_db
(
    DB_RESIDUAL_STATISTICS *residual_stats  /* I: Data to write to DB */
)
{
    int status;
    int table_size;
    char *database_name = NULL;
    char *user_name = NULL;
    char *password = NULL;
    char *host = NULL;
    struct ias_db_connection *db = NULL;    /* Database connection */
    struct ias_db_query *query = NULL;
    char query_string[100];
    char error_message[IAS_SQL_RTN_MSGLEN];

    /* Build the table of data to write to the DB */
    IAS_DB_INSERT_TABLE table[] =
    {
        IAS_DB_TABLE_STRUCT_STRING(residual_stats, geometric_wo_common_id),
        IAS_DB_TABLE_STRUCT_STRING(residual_stats, gcp_source),
        IAS_DB_TABLE_STRUCT_DOUBLE(residual_stats, mean_lat_of_gcps),
        IAS_DB_TABLE_STRUCT_DOUBLE(residual_stats, mean_lon_of_gcps),
        IAS_DB_TABLE_STRUCT_DOUBLE(residual_stats, pixel_size),
        IAS_DB_TABLE_STRUCT_INT(residual_stats, nbr_of_vld_pnts),
        IAS_DB_TABLE_STRUCT_DOUBLE(residual_stats, mean_alng_trk),
        IAS_DB_TABLE_STRUCT_DOUBLE(residual_stats, mean_acrss_trk),
        IAS_DB_TABLE_STRUCT_DOUBLE(residual_stats, rmse_alng_trk),
        IAS_DB_TABLE_STRUCT_DOUBLE(residual_stats, rmse_acrss_trk),
        IAS_DB_TABLE_STRUCT_DOUBLE(residual_stats, std_dev_alng_trk),
        IAS_DB_TABLE_STRUCT_DOUBLE(residual_stats, std_dev_acrss_trk),
        IAS_DB_TABLE_STRUCT_DOUBLE(residual_stats, correlation_coefficient)
    };
    table_size = sizeof(table)/sizeof(IAS_DB_INSERT_TABLE);

    /* Get the database connection info */
    status = ias_db_get_connect_info(&database_name, &user_name, &password,
            &host);
    if (status != SUCCESS)
    {
        /* Failed getting the database connection info */
        IAS_LOG_ERROR("Getting database connection info");
        return ERROR;
    }

    /* Connect to the database for a transaction */
    db = ias_db_connect_with_transaction(database_name, user_name, password,
            host);

    /* Free the database connection info before checking status of DB object */
    free(database_name);
    free(user_name);
    free(password);
    free(host);

    /* Check the connection */
    if (db == NULL)
    {
        IAS_LOG_ERROR("Connecting to database");
        ias_db_close_database_lib();
        return ERROR;
    }

    /* Get the common ID associated with this work order */
    status = ias_misc_get_char_id(db,
            "GEOMETRIC_WORK_ORDER_COMMON", "geometric_wo_common_id",
            get_work_order_id(), residual_stats->geometric_wo_common_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving geometric_wo_common_id for work order: %s",
                get_work_order_id());
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Delete previous records */
    status = snprintf(query_string, sizeof(query_string), "delete from "
            "%s where geometric_wo_common_id=%s", TABLE_NAME,
            residual_stats->geometric_wo_common_id);
    if (status < 0 || status >= sizeof(query_string))
    {
        IAS_LOG_ERROR("Preparing query string");
        IAS_LOG_ERROR("Deleting records from %s where "
                "geometric_wo_common_id=%s", TABLE_NAME,
                residual_stats->geometric_wo_common_id);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }
    query = ias_db_perform_query(db, query_string);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, error_message,
                sizeof(error_message));
        IAS_LOG_ERROR("Query of %s failed: %s", TABLE_NAME, error_message);
        IAS_LOG_ERROR("Deleting records from %s where "
                "geometric_wo_common_id=%s", TABLE_NAME,
                residual_stats->geometric_wo_common_id);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }
    ias_db_query_close(query);
    query = NULL;

    /* Insert the data into the database */
    status = ias_db_insert_using_table(db, TABLE_NAME, table, table_size, 1);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Insert into %s failed", TABLE_NAME);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    status = ias_db_disconnect_with_transaction(db, IAS_DB_COMMIT);
    if (status != SUCCESS)
        IAS_LOG_ERROR("Committing database connection");
    return status;
}
