/******************************************************************************
NAME:           calc_window_corners

PURPOSE:
The calc_window_corners function calculates the projection coordinates for
a window given the window's line/sample corners and the overall image 
projection window's line/sample corners.

******************************************************************************/

#include "calc_window_corners.h" /* prototypes for this module */
#include "ias_math.h" /* ias math matrix operation prototypes */

/* macro to calculate a coordinate given a point and the coefficients for
   an equation of the form: new_proj_coord = coeff0 + coeff1 * sample +
   coeff2 * line + coeff3 * line * sample */
#define find_coord(point,coeff) \
    coeff[0] + coeff[1] * point.x + coeff[2] * point.y + \
    coeff[3] * point.x * point.y

#define MATRIX_ELEMENTS 16
#define SINGLE_ELEMENTS 4

/******************************************************************************
NAME:           calc_window_corners

PURPOSE:        
The calc_window_corners function calculates the projection coordinates for
a window given the windows line/sample corners and the overall image 
projection window and line/sample corners.

RETURN VALUE:
The CORNERS structure at window_corner_proj_ptr is returned.

NOTES:
This routine makes the assumption that the window projection coordinates can
be found by using the equation form:
   Line_proj = a0 + a1 * sample_in + a2 * line_in + a3 * sample_in * line_in  
"A" matrix is packed with the sample/line and projection values of the four 
known corner points and used to solve for the coefficients.  Those 
coefficients are then used to calculate the window projection coordinates.

******************************************************************************/

void calc_window_corners 
(
    const struct IAS_CORNERS *known_corners_ptr,     
       /* I: known corner line/sample coordinates */
    const struct IAS_CORNERS *known_corners_proj_ptr, 
       /* I: known corner projection coordinates */
    const struct IAS_CORNERS *window_corners_ptr,  
       /* I: sub-window line/sample coordinates */
    struct IAS_CORNERS *window_corners_proj_ptr       
       /* O: sub-window projection coordinates */
)
{
   double A[MATRIX_ELEMENTS];           /* matrix with the four known corner 
                                           points' coefficients */
   double v[SINGLE_ELEMENTS];           /* decomposed vector */
   double line_coeff[SINGLE_ELEMENTS];  /* line coefficients matrix */
   double samp_coeff[SINGLE_ELEMENTS];  /* sample coefficients matrix */

   /* pack the A matrix with input lines/samples.  The order of the 
      corner points is upleft, upright, loleft, then loright. */
   A[0]  = 1.0;
   A[1]  = 1.0;
   A[2]  = 1.0;
   A[3]  = 1.0;

   /* samples */
   A[4]  = known_corners_ptr->upleft.x;
   A[5]  = known_corners_ptr->upright.x;
   A[6]  = known_corners_ptr->loleft.x;
   A[7]  = known_corners_ptr->loright.x;

   /* lines */
   A[8]  = known_corners_ptr->upleft.y;
   A[9]  = known_corners_ptr->upright.y;
   A[10] = known_corners_ptr->loleft.y;
   A[11] = known_corners_ptr->loright.y;

   /* matrix multiple terms */
   A[12] = A[4]*A[8];
   A[13] = A[5]*A[9];
   A[14] = A[6]*A[10];
   A[15] = A[7]*A[11];

   /* pack the sample coefficients vector with the known output samples */
   samp_coeff[0] = known_corners_proj_ptr->upleft.x;
   samp_coeff[1] = known_corners_proj_ptr->upright.x;
   samp_coeff[2] = known_corners_proj_ptr->loleft.x;
   samp_coeff[3] = known_corners_proj_ptr->loright.x;

   /* decompose the matrix */
   ias_math_matrix_QRfactorization( A, 4, 4, v, 0);

   /* solve for the sample coefficients */
   ias_math_matrix_QRsolve( A, 4, 4, v, samp_coeff, 0);

   /* pack the line coefficients  vector with the known output lines */
   line_coeff[0] = known_corners_proj_ptr->upleft.y;
   line_coeff[1] = known_corners_proj_ptr->upright.y;
   line_coeff[2] = known_corners_proj_ptr->loleft.y;
   line_coeff[3] = known_corners_proj_ptr->loright.y;

   /* solve for the line coefficients, reusing the same decomposed matrix */
   ias_math_matrix_QRsolve( A, 4, 4, v, line_coeff, 0 );

   /* find the window corner projection coordinates */
   window_corners_proj_ptr->upleft.x = 
                     find_coord(window_corners_ptr->upleft, samp_coeff);
   window_corners_proj_ptr->upleft.y = 
                     find_coord(window_corners_ptr->upleft, line_coeff);
   window_corners_proj_ptr->upright.x = 
                     find_coord(window_corners_ptr->upright, samp_coeff);
   window_corners_proj_ptr->upright.y = 
                     find_coord(window_corners_ptr->upright, line_coeff);
   window_corners_proj_ptr->loleft.x = 
                     find_coord(window_corners_ptr->loleft, samp_coeff);
   window_corners_proj_ptr->loleft.y = 
                     find_coord(window_corners_ptr->loleft, line_coeff);
   window_corners_proj_ptr->loright.x = 
                     find_coord(window_corners_ptr->loright, samp_coeff);
   window_corners_proj_ptr->loright.y = 
                     find_coord(window_corners_ptr->loright, line_coeff);
}
