/******************************************************************************
NAME:           create_metadata

PURPOSE: Create the output image metadata based on the input image metadata
         and grid information

RETURN VALUE:
Type = int
Value           Description
-----           -----------
ERROR           Error; image not opened
SUCCESS         Successfully updated the metadata

NOTES:
none

ALGORITHM REFERENCES:
none

******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ias_logging.h"             /* logging prototype */
#include "ias_const.h"               /* constants prototype */
#include "ias_l1g.h"                 /* L1G library prototype */
#include "ias_geom_grid.h"           /* Geometric Grid library prototype */
#include "ias_satellite_attributes.h"/* Satellite Attributes library 
                                        prototype */        
#include "ias_miscellaneous.h"       /* get software version prototype */
#include "get_resample_parameters.h" /* prototype for getting geometric resample
                                        parameters */ 
#include "calc_window_corners.h"     /* prototype for calculating window 
                                        corners */
#include "create_metadata.h"         /* prototype for this module */

int create_metadata
(
    const char *input_filename,          /* I: name of the input filename */
    const char *output_filename,         /* I: name of the output filename */
    const IAS_GEOM_GRID_TYPE *grid_ptr   /* I: pointer to grid */
)
{
    L1GIO *output_image;          /* output image */
    L1GIO *input_image;           /* input image */
    IAS_L1G_FILE_METADATA ofmd;   /* output file metadata */
    IAS_L1G_FILE_METADATA fmd;    /* input file metadata */
    IAS_L1G_BAND_METADATA *bmd;   /* input band metadata */
    IAS_L1G_BAND_METADATA *obmd;  /* output band metadata */
    int count;                    /* loop counter */
    int band_index;               /* band index */
    int band_number;              /* band number */
    int index;                    /* temporary index */
    int status;                   /* return status */
    const WINDOW_TYPE *output_window_ptr;
    struct IAS_CORNERS window_corners;/* window corners lines/samples */
    struct IAS_CORNERS image_corners; /* entire image corners */
    struct IAS_CORNERS proj_window;   /* projected window coordinates */

    memset(&ofmd, 0, sizeof(ofmd));

    /* open the input file */
    input_image = ias_l1g_open_image(input_filename, IAS_READ);
    if (!input_image)
    {
        IAS_LOG_ERROR("Unable to open %s to read metadata", input_filename);
        return ERROR;
    }
    if (ias_l1g_get_file_metadata(input_image, &fmd) != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to read file metadata from %s", input_filename);
        ias_l1g_close_image(input_image);
        return ERROR;
    }

    if (grid_ptr == NULL)
    {
        IAS_LOG_ERROR("Empty grid_ptr pointer");
        ias_l1g_close_image(input_image);
        return ERROR;    
    }

    ofmd.projection_code = grid_ptr->code;
    ofmd.zone_code = grid_ptr->zone;
    strncpy(ofmd.datum, grid_ptr->datum, sizeof(ofmd.datum));
    ofmd.spheroid_code = grid_ptr->spheroid;
    strncpy(ofmd.projection_units, grid_ptr->units,
            sizeof(ofmd.projection_units));
    for (index = 0; index < IAS_PROJ_PARAM_SIZE; index++)
        ofmd.projection_parameters[index] = grid_ptr->projprms[index];
    ofmd.wrs_path = fmd.wrs_path;
    ofmd.wrs_row = fmd.wrs_row;
    ofmd.roll_angle = fmd.roll_angle;
    strncpy(ofmd.spacecraft, fmd.spacecraft, sizeof(ofmd.spacecraft));
    strncpy(ofmd.capture_direction, fmd.capture_direction,
            sizeof(ofmd.capture_direction));
    strncpy(ofmd.capture_date, fmd.capture_date, sizeof(ofmd.capture_date));
    strncpy(ofmd.capture_time, fmd.capture_time, sizeof(ofmd.capture_time));

    /* Set the collection type */
    strncpy(ofmd.collection_type, fmd.collection_type, 
            sizeof(ofmd.correction_type));

    /* Set the resample type in the metadata */
    switch (get_resampling_type())
    {
        case CC_RESAMPLING:
            strncpy(ofmd.resample_type, "CC", sizeof(ofmd.resample_type));
            break;
        case MTF_RESAMPLING:
            strncpy(ofmd.resample_type, "MTF", sizeof(ofmd.resample_type));
            break;
        case NN_RESAMPLING:
            strncpy(ofmd.resample_type, "NN", sizeof(ofmd.resample_type));
            break;
        case BI_RESAMPLING:
            strncpy(ofmd.resample_type, "BI", sizeof(ofmd.resample_type));
            break;
        default:
            IAS_LOG_ERROR("Unable to determine resample type.");
            ias_l1g_close_image(input_image);
            return ERROR;
    }


    /* Get software versions, might need change later */
    strcpy(ofmd.software_version, ias_misc_get_software_version()); 
    strcpy(ofmd.ingest_software_version, fmd.ingest_software_version); 

    /* allocate memory for the band metadata */
    bmd = (IAS_L1G_BAND_METADATA *)malloc(sizeof(IAS_L1G_BAND_METADATA) 
                        * grid_ptr->nbands);
    obmd = (IAS_L1G_BAND_METADATA *)malloc(sizeof(IAS_L1G_BAND_METADATA) 
                        * grid_ptr->nbands);
    if (!bmd || !obmd)
    {
        IAS_LOG_ERROR("Allocating memory for band metadata");
        free(bmd);
        free(obmd);
        ias_l1g_close_image(input_image);
        return ERROR;
    }
    memset(obmd, 0, sizeof(IAS_L1G_BAND_METADATA) * grid_ptr->nbands);

    for (count = 0; count < grid_ptr->nbands; count++)
    {
        band_index = grid_ptr->band_index_list[count];
        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Converting band index %d to band number",
                          band_index);
            free(bmd);
            free(obmd);
            ias_l1g_close_image(input_image);
            return ERROR;
        }

        /* Read the band metadata */
        status = ias_l1g_get_band_metadata(input_image, band_number, bmd);
        if (status == ERROR)
        {
            IAS_LOG_ERROR("Getting input band metadata\n");
            free(bmd);
            free(obmd);
            ias_l1g_close_image(input_image);
            return ERROR;
        }

        /* put in band number and band name */
        obmd[count].band_number = band_number;
        strncpy(obmd[count].band_name, bmd[count].band_name,
                sizeof(obmd[count].band_name));

        /* set the output window information */
        output_window_ptr = get_output_window(band_index);

        /* set the projection coordinates for the output window */
        image_corners.upleft.x   = 0;
        image_corners.upleft.y   = 0;
        image_corners.upright.x  = 
                          grid_ptr->gridbands[count].output_image_samps-1;
        image_corners.upright.y  = 0;
        image_corners.loleft.x   = 0;
        image_corners.loleft.y   = 
                          grid_ptr->gridbands[count].output_image_lines-1;
        image_corners.loright.x  = 
                          grid_ptr->gridbands[count].output_image_samps-1;
        image_corners.loright.y  = 
                          grid_ptr->gridbands[count].output_image_lines-1;

        window_corners.upleft.x  = output_window_ptr->start_sample;
        window_corners.upleft.y  = output_window_ptr->start_line;
        window_corners.upright.x = output_window_ptr->end_sample-1;
        window_corners.upright.y = output_window_ptr->start_line;
        window_corners.loleft.x  = output_window_ptr->start_sample;
        window_corners.loleft.y  = output_window_ptr->end_line-1;
        window_corners.loright.x = output_window_ptr->end_sample-1;
        window_corners.loright.y = output_window_ptr->end_line-1;

        /* calculate the new projection window */
        calc_window_corners(&image_corners, &grid_ptr->corners, 
                            &window_corners, &proj_window);

        /* copy the projection window to the metadata */
        obmd[count].upper_left_x  = proj_window.upleft.x;
        obmd[count].lower_left_x  = proj_window.loleft.x;
        obmd[count].upper_right_x = proj_window.upright.x;
        obmd[count].lower_right_x = proj_window.loright.x;
        obmd[count].upper_left_y  = proj_window.upleft.y;
        obmd[count].lower_left_y  = proj_window.loleft.y;
        obmd[count].upper_right_y = proj_window.upright.y;
        obmd[count].lower_right_y = proj_window.loright.y;

        obmd[count].maximum_pixel_value = get_output_range_maximum();
        obmd[count].minimum_pixel_value = get_output_range_minimum();
        obmd[count].pixel_range_valid = 1;
        /* set the projected pixel size */
        obmd[count].projection_distance_y
                = grid_ptr->gridbands[band_index].pixsize_line;
        obmd[count].projection_distance_x 
                = grid_ptr->gridbands[band_index].pixsize_samp;

        /* set the min/max radiance values */
        obmd[count].minimum_radiance = get_output_range_minimum();
        obmd[count].maximum_radiance = get_output_range_maximum();
        obmd[count].radiance_valid = 1;
      
        /* set the spectral radiance values */
        obmd[count].spectral_radiance_scaling_offset = 
                        bmd[count].spectral_radiance_scaling_offset;
        obmd[count].spectral_radiance_scaling_gain = 
                        bmd[count].spectral_radiance_scaling_gain;
      
        /* set the reflectance scaling values */
        obmd[count].reflectance_scaling_offset = 
                        bmd[count].spectral_radiance_scaling_offset;
        obmd[count].reflectance_scaling_gain = 
                        bmd[count].spectral_radiance_scaling_gain;
        obmd[count].reflectance_valid = bmd[count].reflectance_valid;
      
        /* Might want to think about populating this with source
           instrument */
        strncpy(obmd[count].instrument_source, bmd[count].instrument_source,
                 sizeof(obmd[count].instrument_source));
    }

    /* Close the input image file */
    ias_l1g_close_image(input_image);
    input_image = NULL;

    /* open the output file */
    output_image = ias_l1g_open_image(output_filename, IAS_UPDATE);
    if (!output_image)
    {
        IAS_LOG_ERROR("Unable to open %s to update metadata", output_filename);
        free(bmd);
        free(obmd);
        return ERROR;
    }

    /* write the metadata */
    if ((ias_l1g_set_file_metadata(output_image, &ofmd) != SUCCESS)
        || (ias_l1g_set_band_metadata(output_image, obmd, grid_ptr->nbands) 
        != SUCCESS))
    {
        IAS_LOG_ERROR("Updating metadata in %s", output_filename);
        free(bmd);
        free(obmd);
        ias_l1g_close_image(output_image);
        return ERROR;
    }
    ias_l1g_close_image(output_image);
    free(bmd);
    free(obmd);

    /* return that the metadata was successfully written */
    return SUCCESS;
}
