/******************************************************************************
NAME:           geomresample

PURPOSE:
The geomresample module contains the "main" function for the geometric 
resampler.

NOTES:
- The scans parameter is converted to lines by multiplying by 
  SCAN_LINE_CONVERSION_FACTOR, which is 16.

******************************************************************************/
#include <stdlib.h>         /* exit prototype */
#include <stdio.h>          /* sprintf prototype */
#include "ias_types.h"      /* IAS_DATA_TYPE definitions */
#include "ias_const.h"      /* SUCCESS/ERROR definitions */
#include "ias_logging.h"    /* IAS logging prototype */
#include "ias_satellite_attributes.h"   /* IAS_MAX_NBANDS prototype */
#include "ias_l1g.h"        /* L1g image I/O */
#include "ias_math.h"       /* ias math prototype */
#include "ias_geom_grid.h"  /* geom grid structure */
#include "ias_threadpool.h" /* threadpool implementation */
#include "kernel.h"         /* Kernel_Setup prototype */
#include "warp.h"           /* warp prototype */
#include "get_resample_parameters.h"  /* read parameters routine */
#include "resample_image.h" /* setup_resample_output prototype */
#include "create_metadata.h"/* create_metadata prototype */
#include "update_omf.h"     /* write output image file into the OMF */

/******************************************************************************
NAME:           main

PURPOSE:
This is the main function for the resampler.  It controls the process of
resampling a level 1R (or 0R) image into a level 1G image using the 
grid.

Note: Macro ASSERT is defined in debug.h and is used for debug purpose only,
      it could be deleted if there is a need since it is not hurting anything. 
******************************************************************************/
#define SCAN_LINE_CONVERSION_FACTOR 16

int main(int argc, char *argv[]) 
{
    IAS_GEOM_GRID_TYPE grid;    /* image grid */
    int current_band;           /* current band counter */
    int band_index;             /* Band index */
    int status;                 /* status of return from function */
    int num_cpus = IAS_THREAD_GET_NUM_PROCESSORS(); /* number of processors */
    int procs_to_use;           /* number of processors to use */
    const int *band_index_ptr;  /* pointer to list of band indexs to 
                                   resample */
    int input_image_scans_to_buffer = 50; /* number of scans */
    int max_output_image_buffer = 500;    /* in MB */

    RESAMPLE_TYPE resampling_type;   /* current resampling type */
    L1GIO *output_l1g_io_ptr = NULL;      /* output L1G file id */
    L1G_BAND_IO *output_l1g_band = NULL;  /* output L1G band id */
    const WINDOW_TYPE *output_window_ptr; /* pointer to output window geometry*/
    OUTPUT_DATA_TYPE odtype;         /* selected output type */
    int grid_io_band_list[IAS_MAX_NBANDS];   
                                     /* list of bands for grid i/o operations*/
    int num_grid_io_bands;           /* number of bands to operate on for 
                                        grid i/o */
    struct ias_threadpool *threadpool; /* thread pool for multithreading */
    int buffer_full_output_image = 0; /* flag to buffer entire output image */
    int band_number;          /* Band number */
    int n_lines;              /* Number of output lines */
    int n_samps;              /* Number of output samples */
    IAS_DATA_TYPE dtype;      /* data type of each band */

    /* make sure the stdio output is only line buffered in case the output
       is piped to a file which is being viewed */
    ias_log_initialize("Geomresample");

    /* if there are no arguments, generate a usage message */
    if (argc < 2)
    {
        IAS_LOG_ERROR("Not enough arguments, should be called as "
                      "./geomresample parameter_filename");
        return EXIT_FAILURE;
    }

    /* read the resampler control parameters and initial the satellite
       attributes library */
    if (read_resample_parameters(argv[1]) != SUCCESS)
    {
        IAS_LOG_ERROR("Error reading parameters from %s", argv[1]);
        return EXIT_FAILURE;
    }

    /* generate entry message */
    IAS_LOG_INFO("Geometric Resample Initiated");

    dtype = odtype = get_output_data_type(); 
  
    /* get the resampling type to use and the list of bands to 
       perform the resampling on */
    resampling_type = get_resampling_type();
    band_index_ptr = get_band_list();

    /* Initialize the geom grid */
    status = ias_geom_grid_initialize(IAS_MAX_NBANDS, &grid);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing GEOM Grid");
        return EXIT_FAILURE; 
    }

    /* read the grid header information (includes all the information in the
       grid, except for the data which is pointed to by pointers) */
    IAS_LOG_INFO("Reading Grid");
    if (ias_read_geom_grid_header(get_geom_grid_filename(), &grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading the grid from %s",get_geom_grid_filename());
        return EXIT_FAILURE;
    }

    if (ias_read_geom_grid_band_header(get_geom_grid_filename(), &grid) 
                     != SUCCESS)
    {
        IAS_LOG_ERROR("Reading the grid band from %s",get_geom_grid_filename());
        return EXIT_FAILURE;
    }

    /* if the user did not request output windowing, set the output window
       to the output image size from the grid */
    if (!do_output_window())
    {
        for (band_index = 0; band_index < grid.nbands; band_index++)
        {
            set_output_window(grid.band_index_list[band_index], 0, 0,
                grid.gridbands[grid.band_index_list[band_index]].
                output_image_lines, grid.gridbands[grid.
                band_index_list[band_index]].output_image_samps);
        }
    }
    else
    {
        /* scale the output window relative to the reference window passed
           in for the first band in the band_list. */
        for (band_index = 0; band_index < grid.nbands; band_index++)
        {
            /* scale the window relative to the reference band 
               (the first one in the band list) */
            set_scaled_output_window(grid.band_index_list[band_index], 
                grid.gridbands[band_index_ptr[0]].pixsize_line/
                grid.gridbands[grid.band_index_list[band_index]].pixsize_line,
                grid.gridbands[band_index_ptr[0]].pixsize_samp/
                grid.gridbands[grid.band_index_list[band_index]].pixsize_samp);

            /* make sure the window is contained in the output grid - if
               it is not, generate an error message and exit */
            output_window_ptr = 
                    get_output_window(grid.band_index_list[band_index]);
            if ((output_window_ptr->start_line < 0) ||
                (output_window_ptr->start_sample < 0) ||
                (output_window_ptr->end_line >= 
                 grid.gridbands[grid.band_index_list[band_index]].
                 output_image_lines) ||
                (output_window_ptr->end_sample >=
                 grid.gridbands[grid.band_index_list[band_index]].
                 output_image_samps))
            {
                IAS_LOG_ERROR("Window specified falls outside output image for"
                              " band_index %d", band_index);
                return EXIT_FAILURE;
            }
        }
    }

    /* Open the HDF5 DEM image */
    output_l1g_io_ptr = ias_l1g_open_image(get_output_image_filename(), 
                                           IAS_WRITE);
    if (output_l1g_io_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening the output image file %s",
                      get_output_image_filename());
        return EXIT_FAILURE;
    }

    /* initialize the threadpool */
    procs_to_use = get_max_processors_to_use();
    if (procs_to_use >= num_cpus)
        procs_to_use = num_cpus;

    threadpool = ias_threadpool_initialize(procs_to_use);
    if (!threadpool)
    {
        IAS_LOG_ERROR("Error initializing the thread pool");
        return EXIT_FAILURE;
    }

    initialize_resample_image();

    /* setup the resampler for the output type and window specified in the
       input parameters */
    setup_resampf1p_output();

    /* process all the requested bands */
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        current_band = band_index_ptr[band_index];

        /* make sure a band is not attempted if it does not exist */
        if (current_band !=SUCCESS || current_band >= IAS_MAX_NBANDS)
             break;

        band_number = ias_sat_attr_convert_band_index_to_number(current_band);
        if (band_number != band_index + 1)
        {
            IAS_LOG_ERROR("Band index and band number conversion");
            ias_l1g_close_image(output_l1g_io_ptr);
            ias_geom_grid_free(grid_io_band_list, &num_grid_io_bands, &grid);
            ias_threadpool_destroy(threadpool);
            return EXIT_FAILURE;
        }

        IAS_LOG_INFO("Resampling band %d", band_number);

        /* read the grid data that pertains to this band */
        num_grid_io_bands = 1;
        if (ias_read_geom_grid_band_pointers(get_geom_grid_filename(), 
            &band_number, &num_grid_io_bands, &grid) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading the grid from %s",get_geom_grid_filename());
            /* close the open L1g HDF file since HDF is quite picky about 
               being closed */
            ias_l1g_close_image(output_l1g_io_ptr);
            ias_geom_grid_free(grid_io_band_list, &num_grid_io_bands, &grid);
            ias_threadpool_destroy(threadpool);
            return EXIT_FAILURE;
        }
        
        if (resampling_type != NN_RESAMPLING)
        { 
            if (Kernel_Setup(current_band, resampling_type, 
                   get_cubic_convolution_alpha()) != SUCCESS)
            {
                IAS_LOG_ERROR("Setting up the resampling kernel");
                /* close the open L1g HDF file since HDF is quite picky about 
                   being closed */
                ias_l1g_close_image(output_l1g_io_ptr);
                ias_geom_grid_free(grid_io_band_list, &num_grid_io_bands, 
                                   &grid);
                ias_threadpool_destroy(threadpool);
                return EXIT_FAILURE;
            }

        } /* not nearest neighbor */ 

        n_lines = grid.gridbands[current_band].output_image_lines;
        n_samps = grid.gridbands[current_band].output_image_samps;

       /* Set up the band data */
       int num_SCAs = 1;
       output_l1g_band = ias_l1g_open_band(output_l1g_io_ptr, band_number, 
                                      &dtype, &num_SCAs, &n_lines, &n_samps); 
        if (output_l1g_band == NULL)
        {
            IAS_LOG_ERROR("Opening input image band for information\n");
            ias_l1g_close_image(output_l1g_io_ptr);
            ias_geom_grid_free(grid_io_band_list, &num_grid_io_bands, &grid);
            return EXIT_FAILURE;
        }

        status = warp(threadpool, &grid, band_index, resampling_type,
                      output_l1g_band, input_image_scans_to_buffer * 
                      SCAN_LINE_CONVERSION_FACTOR, procs_to_use, 
                      max_output_image_buffer, buffer_full_output_image);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Call warp to perform resampling");
            /* close the open 1g HDF file since HDF is quite picky about 
               being closed */
            status = ias_l1g_close_band(output_l1g_band);
            ias_l1g_close_image(output_l1g_io_ptr);
            ias_geom_grid_free(grid_io_band_list, &num_grid_io_bands, &grid);
            ias_threadpool_destroy(threadpool);
            return EXIT_FAILURE;
        }

        /* Close the band */
        status = ias_l1g_close_band(output_l1g_band);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Closing the band\n");
            ias_l1g_close_image(output_l1g_io_ptr);
            ias_geom_grid_free(grid_io_band_list, &num_grid_io_bands, &grid);
            return EXIT_FAILURE;
        }

        /* free the grid.gridbands pointers */
        ias_geom_grid_band_free(&grid, band_number);

    } /* for (band_index = 0; band_index < NBANDS; band_index++) */

    /* Clean up */
    ias_threadpool_destroy(threadpool);
    cleanup_resample_image();

    /* close the output 1g image */
    if (ias_l1g_close_image(output_l1g_io_ptr) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing output image");
        return EXIT_FAILURE;
    }

   /*------------------------------------------------------------
        Write all ancillary info (proj info, corners, etc.)
    ------------------------------------------------------------*/
    if (create_metadata(get_input_image_filename(), get_output_image_filename(),
              &grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing output image metadata");
        return EXIT_FAILURE;
    }

    /* free the the geometeric grid memory allocations */
    free(grid.gridbands);

    /* clean up */
    Cleanup_Kernel();

    /* Write the Geometric Grid filename into OMF */
    update_omf(get_omf_filename(), get_output_image_filename(), 
               get_processing_pass(), get_doq_image_flag());

    /* generate exit message */
    IAS_LOG_INFO("Geometric Resample Successful Completion");

    return EXIT_SUCCESS;
}
