/******************************************************************************
NAME:           get_resample_parameters

PURPOSE:        
This module reads the input Work Order parameter file required by the 
geometric resampler.

ROUTINES:
    read_resample_parameters
    get_input_image_filename
    get_band_list
    get_output_image_filename
    get_geom_grid_filename
    get_omf_filename
    get_DEM_filename
    get_max_processors_to_use    
    get_output_data_type
    get_output_data_size
    do_output_window
    get_output_window
    set_output_window
    set_scaled_output_window
    get_resampling_type
    get_output_range_minimum
    get_output_range_maximum
    get_cubic_convolution_alpha
    get_fill_pixel
    get_processing_pass
    get_doq_image_flag
    get_satellite_number

******************************************************************************/
#include <limits.h>                   /* for PATH_MAX parameter */

#include <stdlib.h>                   /* standard library functions */
#include <string.h>                   /* string library functions */ 
#include <stdio.h>                    /* standard i/o library functions */ 
#include <stdint.h>                   /* standard integer type functions */
#include "ias_logging.h"              /* IAS logging prototype */
#include "ias_parm.h"                 /* IAS parameter file prototype */
#include "ias_parm_standard_parameters.h" /* IAS standard parameter file */
#include "ias_miscellaneous.h"        /* convert to uppercase prototype */
#include "ias_satellite_attributes.h" /* IAS_MAX_NBANDS definitions */
#include "get_resample_parameters.h"  /* prototypes for this module */
#include "debug.h"                    /* ASSERT macro definition */

/* NOTE: if there is a default value for the parameters read from the Work
         Order parameter file, they need to be set when the variable is 
         declared (unless the auto-zero initialization sets the correct 
         default value). */

/* list of bands to be processed */
static int band_list[IAS_MAX_NBANDS] = 
 {ERROR, ERROR, ERROR, ERROR, ERROR, ERROR, ERROR, ERROR, ERROR, ERROR, ERROR};

/* maximum procesors to use */
static int max_processors_to_use;

/* flag to indicate the output data should be windowed.  non-zero = window
   output data. */
static int window_flag;

/* holds the window dimensions for the selected output window */
static WINDOW_TYPE output_window[IAS_MAX_NBANDS];

/* minimum and maximum output values from the geometric resampler */
static double min_max_output[2];

/* cubic convolution alpha parameter */
static double alpha;

/* value to use for fill pixels */
static double fill_pixel;

/* resampling type to perform */
static RESAMPLE_TYPE resampling_type;

/* output data type */
static OUTPUT_DATA_TYPE output_type;

/* work order id */
static char work_order_id[IAS_WORKORDER_ID_SIZE];

/* Processing flow pass        */
static int processing_pass = 1; 

/* Processing DEM or DOQ input file */
static int doq_image_flag;

/* various file name storage */
static char input_image_filename[PATH_MAX];
static char output_image_filename[PATH_MAX];
static char geom_grid_filename[PATH_MAX];
static char geom_doq_grid_filename[PATH_MAX];
static char dem_filename[PATH_MAX];
static char omf_filename[PATH_MAX];

/* character string storage for the output data type read from the work order
   parameter file.  defaulted to byte. */
static char odtype[5] = "I*2";

/* resampling option read from the ODL file. */
static char resample_opt[4];

/* output window when windowing output data.  format is start line, start
   sample, number of lines, number of samples. */
static int window[4];

/* Target WRS path and row */
static int wrs_path;
static int wrs_row;

/* Satellite number */
static int satellite_number = 0;

/******************************************************************************
NAME:           read_resample_parameters

PURPOSE:        
This function reads all the parameters from Work Order parameter file that the 
geometric resampler requires.  All the necessary error checking is done on the 
parameters read (except checking to see if file names are legal/exist).

RETURN VALUE:
Type = int
Value    Description
-----    -----------
SUCCESS  The parameters were successfully read
ERROR    An error occurred reading the parameters

******************************************************************************/

int read_resample_parameters
(
    const char *wo_parameter_filename  /* I: Work Order parameter filename 
                                             that contains resample 
                                             parameters */
)
{
    char path[PATH_MAX];      /* directory path for the OMF */
    char *filename;           /* temporary filename pointer */
    int temporary_band_index; /* temporary band index */
    int count;                /* loop counter */
    int band_already_in_list[IAS_MAX_NBANDS] = {0}; 
                              /* flag to indicate a band is 
                                 already in the band list */
    int band_index;           /* index to next band number to insert */
    int status;               /* return variable */ 
    IAS_SATELLITE_ID sat_id;  /* Satellite ID */

    /* -- Table definition for things from the application parameter file -- */

    IAS_PARM_DECLARE_TABLE(parms, 12);

    /* Add the work order ID parameter */
    IAS_PARM_WORK_ORDER_ID(parms, work_order_id, sizeof(work_order_id), 0);

    /* Get processing pass. */
    IAS_PARM_PROCESSING_PASS(parms, &processing_pass, sizeof(processing_pass),
        0);

    int default_band_list[] = {1};
    /* Add the band list parameter */
    IAS_PARM_ADD_INT(parms, BAND_LIST, "Band list ",
        IAS_PARM_REQUIRED, IAS_PARM_ARRAY,
        /* no limits */
        0, 0, 0,
        /* has one default value */
        0, default_band_list,
        band_list, sizeof(band_list), 0);

    /* Add the output image filename parameter */
    IAS_PARM_ADD_STRING(parms, OUTPUT_IMAGE_FILENAME,
        "Resampled output image file name", IAS_PARM_REQUIRED,
        /* no limits */
        0, NULL,
        /* no default */
        0, NULL,
        output_image_filename, sizeof(output_image_filename), 0);

    const char *valid_odtype_options[] = {"BYTE", "I*2", "UI*2", "I*4", "R*4"};
    const char *default_odtype[] = {"I*2"};
    /* Add the od_type parameter */
    IAS_PARM_ADD_STRING(parms, ODTYPE, "Output data type",
        IAS_PARM_REQUIRED,
        /* has a list of valid options */
        5, valid_odtype_options,
        /* default value set to byte */
        1, default_odtype,        
        odtype, sizeof(odtype), 0);

    int default_window_flag[] = {0};
    /* Add the window flag parameter */
    IAS_PARM_ADD_INT(parms, WINDOW_FLAG, "Flag to indicate "
        "the output should be limited to a window",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has a limited range of 0-1 */
        1, 0, 1,
        /* has one default value */
        1, default_window_flag,
        &window_flag, sizeof(window_flag), 1);

    /* Add the window parameter */
    IAS_PARM_ADD_INT(parms, WINDOW, "Start line, start sample "
        "number of lines, and number of samples to use for the "
        "output window",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
        /* no limits */
        0, 0, 0,
        /* has no default value */
        0, NULL,
        window, sizeof(window), 0);

    const char *valid_resample_options[] = {"CC", "NN", "BI"};
    /* Add the resample parameter */
    IAS_PARM_ADD_STRING(parms, RESAMPLE, "Type of resampling "
        "to use: CC, NN or BI",
        IAS_PARM_REQUIRED, 
        /* three valid resample options */
        3, valid_resample_options,
        /* has no default value */
        0, NULL,
        resample_opt, sizeof(resample_opt), 0);

    /* Add the cubic convolution alpha parameter */
    double default_alpha[] = {-0.5};
    IAS_PARM_ADD_DOUBLE(parms, PCCALPHA, "Cubic convolution alpha parameter",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,  
        1, -1.0, 0.0,
        1, default_alpha,
        &alpha, sizeof(alpha), 1);

    /* Add the min and max output number */
    double default_minmax_output_dn[] = {-32768.0, 32767.0};
    IAS_PARM_ADD_DOUBLE(parms, MINMAX_OUTPUT, "Output digital value range "
        "represented by minimum and maximum values",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,  
        0, 0, 0,
        2, default_minmax_output_dn,
        min_max_output, sizeof(min_max_output), 2);

    double default_fill_pixel[] = {0.0};
    /* Add the window flag parameter */
    IAS_PARM_ADD_DOUBLE(parms, BACKGRND, "Background value",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* no limits */
        0, 0, 0,
        /* has one default value */
        1, default_fill_pixel,
        &fill_pixel, sizeof(fill_pixel), 1);

    /* Put the processing type - DEM or DOQ processing, in a flag */
    IAS_PARM_SOURCE_IMAGE_TYPE(parms, IAS_PARM_OPTIONAL, &doq_image_flag,
        sizeof(doq_image_flag), 1);

    /* See if the user wanted help or a template */
    status = ias_parm_provide_help(wo_parameter_filename, parms,
        IAS_PARM_GET_TABLE_SIZE(parms), IAS_INPUT_PARAMETERS);
    if (status == 1)
    {
        /* If the status of the provided help is 1 the user asked for help.
           Since the parameter file name contains the help request, copy it
           to the OMF file name so the next call will ask for the same help */
        strcpy(omf_filename, wo_parameter_filename);
    }
    else
    {
        /* Read the parameter file */
        count = IAS_PARM_GET_TABLE_SIZE(parms);
        status = ias_parm_read(wo_parameter_filename, parms, count);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading parameters from: %s",
                wo_parameter_filename);
            return ERROR;
        }

        /* Build the OMF file name */
        strcpy(path, wo_parameter_filename);
        filename = strrchr(path, '/');
        if (filename == NULL)     /* no directory portion in the path */
            sprintf(omf_filename, "%s.omf", work_order_id);           
        else
        {
            *filename = '\0';
            sprintf(omf_filename, "%s/%s.omf", path, work_order_id);
        }
    }

    /* -------------- Table definition for things from the OMF ------------- */

    IAS_PARM_DECLARE_TABLE(omf_parms, 7);

    /* Add the source dem filename parameter */
    IAS_PARM_ADD_STRING(omf_parms, DEM_FILENAME, "Source DEM file name",
        IAS_PARM_REQUIRED,
        /* no limits */
        0, NULL,
        /* no default */
        0, NULL,
        input_image_filename, sizeof(input_image_filename), 0);

    /* Add the geometric grid file parameter */
    IAS_PARM_ADD_STRING_WITH_PASS(omf_parms, GEOM_GRID_FILENAME, 
        "Geometric Grid filename",
        IAS_PARM_REQUIRED,
        /* no limits */
        0, NULL,
        /* no default */
        0, NULL,
        geom_grid_filename, sizeof(geom_grid_filename), 0, processing_pass);

    /* Add the geometric grid for the DOQ file parameter */
    IAS_PARM_DOQ_GRID_FILENAME(omf_parms, (char *)geom_doq_grid_filename, 
        sizeof(geom_doq_grid_filename), 0);

    /* Add the window flag parameter */
    IAS_PARM_MAX_PROCESSORS_TO_USE(omf_parms, IAS_PARM_OPTIONAL,
                                   &max_processors_to_use,
                                   sizeof(max_processors_to_use), 1);

    /* Get the WRS path */
    IAS_PARM_TARGET_WRS_PATH(omf_parms, IAS_PARM_OPTIONAL, &wrs_path,
        sizeof(wrs_path), 1);

    /* Get the WRS row */
    IAS_PARM_TARGET_WRS_ROW(omf_parms, IAS_PARM_OPTIONAL, &wrs_row,
        sizeof(wrs_row), 1);

    /* Get the satellite number */
    IAS_PARM_SATELLITE(omf_parms, &satellite_number, sizeof(satellite_number),
        1);

    /* See if the user wanted help or a template this time with the OMF file */
    status = ias_parm_provide_help(omf_filename, omf_parms,
        IAS_PARM_GET_TABLE_SIZE(omf_parms), IAS_OMF_PARAMETERS);
    if (status == 1)
    {
        exit(EXIT_SUCCESS);
    }
    else if (status == ERROR)
    {
        IAS_LOG_ERROR("Printing out parameters from OMF file: %s",
            omf_filename);
        return ERROR;
    }    

    /* Read the OMF file */
    count = IAS_PARM_GET_TABLE_SIZE(omf_parms);
    status = ias_parm_read(omf_filename, omf_parms, count);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from OMF file: %s",
            omf_filename);
        return ERROR;
    }

    /* parse the output data type */
    ias_misc_convert_to_uppercase(odtype);
    if (strcmp(odtype, "BYTE") == 0)
        output_type = BYTE_OUTPUT;
    else if (strcmp(odtype, "I*2") == 0)
        output_type = INT16_OUTPUT; 
    else if (strcmp(odtype, "UI*2") == 0)
        output_type = UINT16_OUTPUT; 
    else if (strcmp(odtype, "I*4") == 0)
        output_type = INT32_OUTPUT;
    else if (strcmp(odtype, "R*4") == 0)
        output_type = FLOAT_OUTPUT;
    else
    {
        IAS_LOG_ERROR("Unknown output type - %s\n",odtype);
        return ERROR;
    }

    /* parse the resampling type */
    ias_misc_convert_to_uppercase(resample_opt); 
    if (strcmp(resample_opt, "CC") == 0)
        resampling_type = CC_RESAMPLING;
    else if (strcmp(resample_opt, "NN") == 0)
        resampling_type = NN_RESAMPLING;
    /* disable the MTF resampling for now
    else if (strcmp(resample_opt, "MTF") == 0)
        resampling_type = MTF_RESAMPLING;  */
    else if (strcmp(resample_opt, "BI") == 0)
        resampling_type = BI_RESAMPLING;
    else
    {
        IAS_LOG_ERROR("Unknown resample option %s",resample_opt);
        return ERROR;
    }

    /* Initialize satellite attributes library */
    sat_id =
        ias_sat_attr_get_satellite_id_from_satellite_number(satellite_number);
    if (sat_id == ERROR)
    {
        IAS_LOG_ERROR("Unsupported satellite number: %d", satellite_number);
        return ERROR;  
    }
    status = ias_sat_attr_initialize(sat_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attributes library");
        return ERROR;  
    }

    /* parse the band number to the band index */
    band_index = 0;
    for (count = 0; count < IAS_MAX_NBANDS; count++)
    {
        /* convert the band number entered into the band index */
        if (band_list[count] != ERROR)
        {
            temporary_band_index = ias_sat_attr_convert_band_number_to_index(
                             band_list[count]);
            if ((count == 0) && (temporary_band_index == ERROR))
            {
                /* no band numbers entered, so it is an error */
                IAS_LOG_ERROR("Illegal band number %d",band_list[count]);
                return ERROR;
            }

            /* make sure the band isn't already in the list */
            if (temporary_band_index != ERROR)
            {
                if (!band_already_in_list[temporary_band_index])
                {
                    /* put band in list and flag that it has been added 
                       to the list */
                    band_list[band_index] = temporary_band_index;
                    band_already_in_list[temporary_band_index]++;
                    band_index++;
                }
                else if (band_already_in_list[temporary_band_index] == 1)
                {
                    /* only print the warning message once for each band */
                    band_already_in_list[temporary_band_index]++;

                    IAS_LOG_WARNING("Band %d appears multiple times "
                               "in the BAND_LIST, only the first one was used",
                                ias_sat_attr_convert_band_index_to_number(
                                temporary_band_index));
                }
            }
        }
    }

    /* fill in remaining band locations with the band error indication */
    for (count = band_index; count < IAS_MAX_NBANDS; count++)
    {
        band_list[count] = ERROR;
    }
       
    if (window_flag)
    {
        /* output windowing was selected, make sure values were 
           provided for the window size */ 
        if ((window[2] == 0) || (window[3] == 0))
        {
            /* error reading required item, log an error */
            IAS_LOG_ERROR("output window size not provided in Work Order"
                          " parameter parameter file"); 
            return ERROR;
        }

        /* initialize the output window data structure, relative to the
               first band specified in the input band list */
        set_output_window(band_list[0], window[0], window[1], window[2], 
                          window[3]);
    }

    if (doq_image_flag)
    {
        /* Make sure the WRS path and row are available */
        if ((IAS_PARM_GET_COUNT_READ(omf_parms, TARGET_WRS_PATH) == 0)
            || (IAS_PARM_GET_COUNT_READ(omf_parms, TARGET_WRS_ROW) == 0))
        {
            /* WRS path and/or row weren't selected, but are needed */
            IAS_LOG_ERROR("WRS path and/or row were not specified, "
                "but they are needed to build the DOQ image name");
            return ERROR;
        }

        /* Get the DOQ reference filename */
        status = ias_misc_get_doq_reference_filename(wrs_path, wrs_row,
            input_image_filename, sizeof(input_image_filename));
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to get the DOQ reference filename");
            return ERROR;
        }
            
        IAS_LOG_INFO("Using DOQ reference image %s", input_image_filename);
    }

    return SUCCESS;
}

/******************************************************************************
NAME:           various accessor functions

PURPOSE:        
These functions provide access to the parameters read. 

RETURN VALUE:
Varies for the function.

NOTES:
The accessor function method of providing parameters was selected since
passing all the parameters up and down a calling sequence is a 
maintenance nightmare.

ALGORITHM REFERENCES:
none

******************************************************************************/


char *get_input_image_filename(void)
{
    return input_image_filename;
}

const int *get_band_list(void)
{
    return band_list;
}

char *get_output_image_filename(void)
{
    return output_image_filename;
}

char *get_geom_grid_filename(void)
{
    if (doq_image_flag)
    {
        return geom_doq_grid_filename;
    }
    else
    {
        return geom_grid_filename;
    }
}

char *get_omf_filename(void)
{
    return omf_filename;
}

char *get_DEM_filename(void)
{
    return dem_filename;
}

int get_max_processors_to_use(void)
{
    return max_processors_to_use;
}

OUTPUT_DATA_TYPE get_output_data_type(void)
{
    return output_type;
}

int get_output_data_size(void)
{
    if (output_type == BYTE_OUTPUT)
        return sizeof(unsigned char);
    else if (output_type == INT16_OUTPUT)
        return sizeof(int16_t);
    else if (output_type == UINT16_OUTPUT)
        return sizeof(uint16_t);
    else if (output_type == INT32_OUTPUT)
        return sizeof(int32_t);
    else 
    {
        ASSERT(output_type == FLOAT_OUTPUT);
        return sizeof(float);
    }
}

int do_output_window(void)
{
   return window_flag;
}

const WINDOW_TYPE *get_output_window
(
    int band_index /* I: band index for the desired the output window */
)
{ 
    /* verify input band number is legal */
    ASSERT((band_index >= 0) && (band_index < IAS_MAX_NBANDS));

    return &output_window[band_index];
}

void set_output_window
(
    int band_index,  /* I: band index for the desired output window */
    int start_line,  /* I: starting line for the output window */
    int start_sample,/* I: starting sample for the output window */
    int num_lines,   /* I: number of lines in the output window */
    int num_samples  /* I: number of samples in the output window */
)
{
    /* verify input band number is legal */
    ASSERT((band_index >= 0) && (band_index < IAS_MAX_NBANDS));

    output_window[band_index].start_line = start_line;
    output_window[band_index].start_sample = start_sample;
    output_window[band_index].num_lines = num_lines;
    output_window[band_index].num_samples = num_samples;
    output_window[band_index].end_line = start_line + num_lines;
    output_window[band_index].end_sample = start_sample + num_samples;
}

void set_scaled_output_window
(
    int band_index,     /* I: band index for the desired output window */
    float line_scaling,  /* I: scaling factor for the window line size */
    float sample_scaling /* I: scaling factor for the window sample size */
)
{
    int refband = band_list[0]; /* the reference band for scaling
                                          is the first band in the band_list */
    /* verify input band number is legal */
    ASSERT((band_index >= 0) && (band_index < IAS_MAX_NBANDS));

    output_window[band_index].start_line 
                  = output_window[refband].start_line * line_scaling;
    output_window[band_index].start_sample 
                  = output_window[refband].start_sample * sample_scaling;
    output_window[band_index].num_lines 
                  = output_window[refband].num_lines * line_scaling;
    output_window[band_index].num_samples 
                  = output_window[refband].num_samples * sample_scaling;
    output_window[band_index].end_line 
          = output_window[band_index].start_line 
          + output_window[band_index].num_lines;
    output_window[band_index].end_sample 
          = output_window[band_index].start_sample 
          + output_window[band_index].num_samples;
}

RESAMPLE_TYPE get_resampling_type(void)
{
    return resampling_type;
}

float get_output_range_minimum(void)
{
    return min_max_output[0];
}

float get_output_range_maximum(void)
{
    return min_max_output[1];
}

float get_cubic_convolution_alpha(void)
{
    return alpha;
}

float get_fill_pixel(void)
{
   return fill_pixel;
}

int get_processing_pass(void)
{
    return processing_pass;
}

int get_doq_image_flag(void)
{
    return doq_image_flag;
}

int get_satellite_number(void)
{
    return satellite_number;
}
