#ifndef _OUTPUT_IMAGE_BUFFER_H_
#define _OUTPUT_IMAGE_BUFFER_H_


#include "ias_const.h"      /* const definition */
#include "ias_l1g.h"        /* l1g HDF I/O routines */
#include "get_resample_parameters.h" /* OUTPUT_WINDOW_TYPE definition */
#include "ias_threadsync.h" /* multiprocessing macros */


#ifdef USE_WRITE_THREAD
/* queue definition for the write queue */
typedef struct
{
    int row_index;      /* row index to write */
    int row_start;      /* starting row (or line) to write */
    int row_end;        /* ending row (or line) to write */
    
} WRITE_QUEUE_BUFFER_TYPE;
#endif

/* NOTE: for those not familiar with the "volatile" keyword.  The volatile
   keyword is a standard C keyword that tells the compiler the variable may
   change without its knowledge.  This essentially disables some optimizations
   on that variable that are not safe in a multiprocessing environment when
   multiple threads may be changing a variable. */

/* The OUTPUT_IMAGE_BUFFER_TYPE defines the data that is needed to track the
   state of the output image buffer. */
typedef struct
{
    L1G_BAND_IO *output_band_io;/* output HDF I/O file pointer */
    int band_index;             /* band index for this output buffer */
    int   threads;              /* number of threads buffer can accommodate */
    void  *data_buffer_ptr;     /* buffer for resampled data */
    int data_size;              /* number of bytes per sample in buffer */
    int bytes_per_line;         /* number of bytes per line in buffer */
    int row_buffer_size;        /* number of bytes in each row buffer */
    int last_row_index;         /* last legal row index */
    WINDOW_TYPE output_window;  /* pointer to output window geometry */
    int *sample_reached_ptr;    /* pointer to array that holds
                                   the sample reached in each line */
    int *sample_written_ptr;    /* pointer to array that holds the sample
                                   written in each line */
    int * volatile buffer_allocation_ptr; /* pointer to the array that tracks 
                                             the buffer allocations */
    int buffers;                /* number of buffers allocated for the data */
    volatile int next_buffer_to_alloc;   /* index of next row to allocate a 
                                            buffer to */
    int resample_dir;           /* direction of resampling */
    int full_buffer;            /* full image buffering being used flag 
                                   (non-zero=full image buffered) */
    int granularity;            /* number of lines for buffer granularity */
    int interleave;             /* interleave between partial write buffers
                                   when insufficient memory to buffer active
                                   output rows */
    int interleave_count;       /* running count of when next buffer interleave
                                   should occur */
    int full_buffer_writes;     /* count of entire row buffers written.  It is
                                   a measure of whether enough memory is being
                                   used for the output image */
    IAS_THREAD_MUTEX_TYPE file_io_mutex; 
                                /* mutex that controls access to file i/o */
    IAS_THREAD_MUTEX_TYPE buffer_mutex; /* mutex that controls access to buffer
                                   manipulation */

#ifdef USE_WRITE_THREAD
    IAS_THREAD_SEMAPHORE_TYPE data_available_sem; 
                                   /* semaphore to notify when data is
                                      available in the write queue */
    IAS_THREAD_MUTEX_TYPE q_mutex; /* mutex that controls access for inserting
                                      data into the write queue */
    WRITE_QUEUE_BUFFER_TYPE *q_ptr;/* pointer to the buffer queue */
    int q_insert_index;            /* index of next queue insert action */
    volatile    int q_remove_index;/* index of next queue remove action */
    int q_size;                 /* number of entries allocated for the queue */
    pthread_t write_thread_id;  /* write thread id */
    
#endif
    
} OUTPUT_IMAGE_BUFFER_TYPE;



OUTPUT_IMAGE_BUFFER_TYPE *initialize_output_image_buffer
(
    int active_lines,            /* I: number of lines active in output image */
    int band_index,              /* I: current band index */
    L1G_BAND_IO *output_band_io, /* I: output HDF I/O file pointer */
    int threads,                 /* I: number of threads to accommodate */
    int resample_dir,            /* I: resampling direction */
    int max_output_image_buffer, /* I: maximum memory to use for output 
                                       buffer */
    int buffer_full_output_image /* I: flag to buffer entire output image,
                                       ignoring max memory size */
);

int cleanup_output_image_buffer
(
    OUTPUT_IMAGE_BUFFER_TYPE *output_buffer_ptr, /* I: pointer to output image
                                                       buffer */
    int success  /* I: set to non-zero if no errors have occurred */
);

void *get_output_image_buffer
(
    OUTPUT_IMAGE_BUFFER_TYPE *output_buffer_ptr, /* I: output buffer pointer */
    int row_index,   /* I: row index (calculated by line number / GRANULARITY)*/
    int thread_num   /* I: thread number of calling thread */
);

int return_output_image_buffer
(
    OUTPUT_IMAGE_BUFFER_TYPE *output_buffer_ptr, 
                     /* I: output buffer state ptr */
    int row_index,   /* I: row index (calculated by line number / GRANULARITY)*/
    int row_start,   /* I: starting line number of row */
    int row_end      /* I: ending line number of row */
);

#endif
