/******************************************************************************
NAME:           read_input_image

PURPOSE:
The read_input_image module provides the routines needed to open, read, and
close an image for input.  The data read is converted to float type for use
by the resampler.

ROUTINES:
    open_input_image
    close_input_image
    read_input_line
    
ALGORITHM REFERENCES:
none

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include "ias_const.h"   /* SUCCESS/ERROR definitions */
#include "ias_logging.h" /* IAS logging prototype */
#include "ias_satellite_attributes.h" /* IAS_MAX_NBANDS */
#include "read_input_image.h" /* prototypes for this module */
#include "debug.h"         /* ASSERT macro definition */



/******************************************************************************
NAME:           open_input_image

PURPOSE:        
The open_input_image routine opens the input file to access the indicated
band number.

RETURN VALUE:
Type = IN_IMAGE_TYPE pointer
Value           Description
-----           -----------
NULL            Error; image not opened
non-NULL        Successfully opened image, file state data returned
The lines and samples in the source image are also returned.

******************************************************************************/

IN_IMAGE_TYPE *open_input_image
(
    char *input_filename,  /* I: input file name of image to open */
    int  band_index,       /* I: band_index of image to open */
    int  *in_nl_ptr,       /* O: number of lines in image */
    int  *in_ns_ptr        /* O: number of samples in image */
)
{
    int in_size;              /* bytes in one unit of input data */
    IN_IMAGE_TYPE *image_ptr; /* input image data */
    IAS_DATA_TYPE dtype[IAS_MAX_NBANDS];   /* data type of image */
    int  nl[IAS_MAX_NBANDS];          /* number of lines in each band */
    int  ns[IAS_MAX_NBANDS];          /* number of samples in each band */
    int number_of_scas = 1;
    int band_number;

    /* verify input parameters are valid */
    ASSERT((band_index >= 0) && (band_index < IAS_MAX_NBANDS));

    /* allocate space for the image state data */
    image_ptr = malloc(sizeof(IN_IMAGE_TYPE));
    if (image_ptr == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for the input image");
        return NULL;
    }

    /* initialize image state data fields */
    image_ptr->input_file = NULL;
    image_ptr->buf_ptr = NULL;
    image_ptr->band_index = band_index;
        
    /* Open the input DEM image */
    image_ptr->input_file = ias_l1g_open_image(input_filename, IAS_READ);
    if (image_ptr->input_file == NULL)
    {
        IAS_LOG_ERROR("Opening DEM data file %s\n", input_filename);
        return NULL;
    }

    band_number = ias_sat_attr_convert_band_index_to_number(band_index);

    /* Set up the band data */
    image_ptr->band_io = ias_l1g_open_band(image_ptr->input_file, band_number, 
                          &dtype[band_index], &number_of_scas, &nl[band_index], 
                          &ns[band_index]);
    if (image_ptr->band_io == NULL)
    {
        IAS_LOG_ERROR("Opening DEM image band for information\n");
        ias_l1g_close_image(image_ptr->input_file);
        return NULL;
    }

    /* The number of SCA should be 1 for now */
    if (number_of_scas != 1)
    {
        IAS_LOG_ERROR("Only 1 SCA can be processed for now\n");
        ias_l1g_close_image(image_ptr->input_file);
        return NULL;
    }

    /* get the input lines and samples */
    *in_nl_ptr = (int)nl[0];
    *in_ns_ptr = (int)ns[0];
    image_ptr->samples_per_line = (int)ns[0];


    /* get the input data type */

    switch (dtype[0])
    {
      case IAS_BYTE:
          image_ptr->in_dtype = BYTE_IN;
          in_size = sizeof(uint8_t);
          break;
      case IAS_I2 :
          image_ptr->in_dtype = INT16_IN;
          in_size = sizeof(int16_t);
          break;
      case IAS_UI2 :
          image_ptr->in_dtype = UINT16_IN;
          in_size = sizeof(uint16_t);
          break;
      case IAS_I4 :
          image_ptr->in_dtype = INT32_IN;
          in_size = sizeof(int32_t);
          break;
      case IAS_R4 :
          image_ptr->in_dtype = FLOAT_IN;
          in_size = sizeof(float);
          break;
      default :
          IAS_LOG_ERROR("Unrecognized data type in %s\n",input_filename);
          close_input_image(image_ptr);
          return NULL;
    }


    /* allocate memory for an intermediate buffer to allow the each line of 
       the input data type to be converted to float for the resampler.
       the allocation is done here to prevent the memory allocation
       from taking place for each set of lines read */
    image_ptr->buf_ptr = malloc (in_size * *in_ns_ptr);
    if (image_ptr->buf_ptr == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for scan buffer");
        close_input_image(image_ptr);
        return NULL;
    }

    return image_ptr;
}

/******************************************************************************
NAME:           close_input_image

PURPOSE:        
The close_input_image routine closes the input file and frees any memory 
allocated.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
SUCCESS         Successfully closed image
ERROR           Error; image not closed

******************************************************************************/

int close_input_image
(
    IN_IMAGE_TYPE *in_image_ptr  /* I/O: pointer to image to close */
)
{ 
    int status;
    /* free buffer memory */
    free(in_image_ptr->buf_ptr);
    in_image_ptr->buf_ptr = NULL;

    /* Close the band */
    status = ias_l1g_close_band(in_image_ptr->band_io);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Closing the band\n");
        ias_l1g_close_image(in_image_ptr->input_file);
        free(in_image_ptr);
        return ERROR;
    }

    if (ias_l1g_close_image(in_image_ptr->input_file) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing input image file");
        free(in_image_ptr);
        return ERROR;
    }

    /* free image buffer memory */
    free(in_image_ptr);

    return SUCCESS;
}

/******************************************************************************
NAME:           read_input_lines

PURPOSE:
The read_1r_scan routine reads the scan number requested from the input file,
converts the data read to float, and returns the data to the calling routine.

RETURN VALUE:
Type = int
Value           Description
-----           -----------
SUCCESS         Successfully read scan
ERROR           Error; scan not read

******************************************************************************/

int read_input_lines
(
    IN_IMAGE_TYPE *in_image_ptr, /* I/O: input file state pointer */
    int   start_line,            /* I: current start line number */
    int   num_lines,             /* I: number of lines to read */
    float *buffer                /* O: data buffer */
)
{
    int index;   /* sample loop counter */
    int count;   /* loop counter */
    int samps = in_image_ptr->samples_per_line; /* cache the sample per line */
    int data_index; /* index to data buffer */
    

    /* read each line, converting it to float */
    for (count = 0; count < num_lines; count++)
    {
        data_index = count * samps;
        
        /* read the input image line into the intermediate buffer */
        if (ias_l1g_read_image(in_image_ptr->band_io, 0,
                              start_line + count, 0 /*start sample*/,
                              1 /* num_lines*/, samps, in_image_ptr->buf_ptr)
                 != SUCCESS)
        {
            IAS_LOG_ERROR("Error reading line %d of input image",
                    start_line + count);
            return ERROR;
        }

        /* convert the data to float for the resampler */
        if (in_image_ptr->in_dtype == BYTE_IN)
        {
            unsigned char *buf_ptr = in_image_ptr->buf_ptr; 
            /* use source pointer of correct type */

            for (index = 0; index < samps; index++)
            {
                /* convert the byte data to float */
                buffer[data_index + index] = buf_ptr[index];
            }
        }
        else if (in_image_ptr->in_dtype == INT16_IN)
        {
            int16_t *buf_ptr = in_image_ptr->buf_ptr; 
            /* use source pointer of correct type */

            for (index = 0; index < samps; index++)
            {
                /* scale the integer data and save it to the real image buffer 
                   in float format */
                buffer[data_index+index] = buf_ptr[index];
            }
        }
        else if (in_image_ptr->in_dtype == UINT16_IN)
        {
            uint16_t *buf_ptr = in_image_ptr->buf_ptr; 
            /* use source pointer of correct type */

            for (index = 0; index < samps; index++)
            {
                /* scale the integer data and save it to the real image buffer 
                   in float format */
                buffer[data_index+index] = buf_ptr[index];
            }
        }
        else if (in_image_ptr->in_dtype == INT32_IN)
        {
            int32_t *buf_ptr = in_image_ptr->buf_ptr; 
            /* use source pointer of correct type */

            for (index = 0; index < samps; index++)
            {
                /* scale the integer data and save it to the real image buffer 
                   in float format */
                buffer[data_index+index] = buf_ptr[index];
            }
        }
        else
        {
            float *buf_ptr = in_image_ptr->buf_ptr; 
            /* use source pointer of correct type */

            /* verify the data type is correct in case a type is added */
            ASSERT(in_image_ptr->in_dtype == FLOAT_IN);

            for (index = 0; index < samps; index++)
            {
                /* scale the integer data and save it to the real image buffer 
                   in float format */
                buffer[data_index+index] = buf_ptr[index];
            }
        }    
    }
    
    return SUCCESS;

}

