/******************************************************************************
NAME:              resample_image

PURPOSE:
Perform geometric resampling on an image using cubic convolution 
or nearest neighbor resampling. 

******************************************************************************/
 
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include "ias_const.h"     /* SUCCESS/ERROR definitions */
#include "ias_logging.h"   /* IAS logging prototype */
#include "ias_structures.h"/* IAS structures prototype */
#include "ias_geom_grid.h" /* IAS_GEOM_GRID_BAND_TYPE definition */
#include "resample_image.h"/* prototypes for this module */
#include "scanman.h"       /* Get_Sample_Data prototype */ 
#include "debug.h"         /* ASSERT macro */
#include "warp.h"          /* RESAMPLE_THREAD_PARAMS definition */ 
#include "ias_threadpool.h"/* multiprocessing wrapper module */
#include "kernel.h"        /* Get_Resample_Weight_Table_Ptr prototype */
#include "output_image_buffer.h" /* output buffer management */

/* local versions of the minimum and maximum output range, adjusted for
   the currently selected output type. */
static double min_range;
static double max_range;

/* local copy of the output data type */
static OUTPUT_DATA_TYPE output_data_type;

/* mutex for assigning work to the threads running process_scas */
static IAS_THREAD_MUTEX_TYPE assignment_mutex;

/* variable for tracking the work to assign to threads */
static int assigned_work;

int initialize_resample_image()
{
    /* create atomic increment mutex */
    if (IAS_THREAD_CREATE_MUTEX(&assignment_mutex) != 0)
    {
        IAS_LOG_ERROR("Creating atomic increment");
        return ERROR;
    }

    assigned_work = 0;

    return SUCCESS;
}

void reset_resample_image()
{
    assigned_work = 0;
}

/* function to return the work assigned, or ERROR */
static int get_assigned_work()
{
    int count;

    /* lock the assignment mutex */
    if (IAS_THREAD_LOCK_MUTEX(&assignment_mutex) != 0)
    {
        IAS_LOG_ERROR("Locking atomic increment mutex");
        return ERROR;
    }

   /* get a local copy of the counter */
    count = assigned_work;

    /* increment the actual counter */
    assigned_work++;

    /* unlock the assignment mutex */
    if (IAS_THREAD_UNLOCK_MUTEX(&assignment_mutex) != 0)
    {
        IAS_LOG_ERROR("Unlocking atomic increment mutex");
        return ERROR;
    }

    return count;
}

void cleanup_resample_image()
{
    IAS_THREAD_DESTROY_MUTEX(&assignment_mutex);
}

/******************************************************************************
NAME:           setup_resampf1p_output

PURPOSE:        
setup_resampf1p_output configures the resampf1p module to generate the 
output image with the requested output data type and dynamic range.

RETURN VALUE:
Nothing is returned by this routine

NOTES:
- This routine uses the parameters read from the ODL parameter file to 
  create locally cached copies in the most efficient format for this module.
- This routine does not use the min/max values defined for the types in
  limits.h since those are machine specific.  The min/max values here are
  data dependent and cannot change for different machines.

******************************************************************************/

void setup_resampf1p_output(void)
{
    output_data_type = get_output_data_type();

    min_range = get_output_range_minimum();
    max_range = get_output_range_maximum();

    if (output_data_type == BYTE_OUTPUT)
    {
        /* byte output so adjust the min and max accordingly */
        if (min_range < 0.0)
            min_range = 0.0;
        else
            min_range = ceil(min_range);
        if (max_range > 255.0)
            max_range = 255.0;
        else
            max_range = floor(max_range);
    }
    else if (output_data_type == INT16_OUTPUT)
    {
        /* 2 byte integer output so adjust the min and max accordingly */
        if (min_range < -32768.0)
            min_range = -32768.0;
        else
            min_range = ceil(min_range);
        if (max_range > 32767.0)
            max_range = 32767.0;
        else
            max_range = floor(max_range);
    }
    else if (output_data_type == UINT16_OUTPUT)
    {
        /* unsigned 2 byte integer output so adjust the min and max
           accordingly */
        if (min_range < 0.0)
            min_range = 0.0;
        else
            min_range = ceil(min_range);
        if (max_range > 65535.0)
            max_range = 65535.0;
        else
            max_range = floor(max_range);
    }
    else if (output_data_type == INT32_OUTPUT)
    {
        /* byte output so adjust the min and max accordingly */
        if (min_range < -2147483648.0) 
            min_range = -2147483648.0;
        else
            min_range = ceil(min_range);
        if (max_range > 2147483647.0)
            max_range = 2147483647.0;
        else
            max_range = floor(max_range);
    }
    else
    {
        /* must be float output, if it's not an error should have been 
           obtained parsing the input, so just assert here during 
           development.  no range manipulations are needed for the float
           case */
        ASSERT (output_data_type == FLOAT_OUTPUT);
    }

}


/******************************************************************************
NAME:           convert_data

PURPOSE:                
convert_data converts the intermediate resampling results (currently stored as
doubles) and converts them to the requested output data type.

RETURN VALUE:
The type converted data is returned in the outbuf.

NOTES:
This routine was created to prevent the code having to be copied to the 
different specialized resampling routines.  This incurs a small calling 
penalty, but it is small when compared to the actual function.

******************************************************************************/

static void convert_data
(
    const double *mid_output_ptr, /* I: data to convert to the output type */
    void *outbuf,           /* O: output data buffer */
    int left_sample,        /* I: starting index in mid_output_ptr buffer */
    int right_sample,       /* I: ending index in mid_output_ptr buffer */
    long row_index          /* I: row index in outbuf */
)
{
    int l;  /* counter */

    /* convert the intermediate resampling results to the requested 
       output type */
    if (output_data_type == BYTE_OUTPUT)
    {
        unsigned char *char_ptr = outbuf;

        /* if buffering output, offset to the correct line and sample */
        char_ptr = &char_ptr[row_index + left_sample];

       /* convert the data, using unsigned rounding */
       for (l = left_sample; l < right_sample; l++)
           *char_ptr++ = (unsigned char)(*mid_output_ptr++ + 0.5);
    }
    else if (output_data_type == FLOAT_OUTPUT)
    {
        float *float_ptr = outbuf;

        /* if buffering output, offset to the correct line and sample */
        float_ptr = &float_ptr[row_index + left_sample];

        for (l = left_sample; l < right_sample; l++)
            *float_ptr++ = *mid_output_ptr++;
    }
    else if (output_data_type == INT16_OUTPUT)
    {
        short int *int16_ptr = outbuf;

        /* if buffering output, offset to the correct line and sample */

        int16_ptr = &int16_ptr[row_index + left_sample];

        for (l = left_sample; l < right_sample; l++)
        {
            /* convert the data, using signed rounding */
            if (*mid_output_ptr >= 0)
                *int16_ptr++ = (short int)(*mid_output_ptr++ + 0.5);
            else
                *int16_ptr++ = (short int)(*mid_output_ptr++ - 0.5);
        }
    }
    else if (output_data_type == UINT16_OUTPUT)
    {
        unsigned short int *uint16_ptr = outbuf;

        /* if buffering output, offset to the correct line and sample */

        uint16_ptr = &uint16_ptr[row_index + left_sample];

       /* convert the data, using unsigned rounding */
       for (l = left_sample; l < right_sample; l++)
           *uint16_ptr++ = (unsigned short int)(*mid_output_ptr++ + 0.5);
    }
    else
    {
        int *int32_ptr = outbuf;

        /* if buffering output, offset to the correct line and sample */
        int32_ptr = &int32_ptr[row_index + left_sample];

        ASSERT(output_data_type == INT32_OUTPUT);

        for (l = left_sample; l < right_sample; l++)
        {
            /* convert the data, using signed rounding */
            if (*mid_output_ptr >= 0.0)
                *int32_ptr++ = (int)(*mid_output_ptr++ + 0.5);
            else
                *int32_ptr++ = (int)(*mid_output_ptr++ - 0.5);
        }
    }
}

/******************************************************************************
NAME:           resample_image_cc

PURPOSE:
The specific resample routine to perform cubic convolution resampling.

RETURN VALUE:
The resampled data is returned.

******************************************************************************/

int resample_image_cc
(
    void * params,  /* I: pointer to input parameter structure. */
    int thread_num  /* I: thread number running this routine */
)
{
    /* create local copies of the parameters */
    RESAMPLE_THREAD_PARAMS *param_ptr = (RESAMPLE_THREAD_PARAMS *)params;
    const IAS_GEOM_GRID_BAND_TYPE *grid_band_ptr = param_ptr->grid_band_ptr;
    int  in_nl            = param_ptr->in_nl;
    int  in_ns            = param_ptr->in_ns;
    const SCAN_MAN_TYPE *scan_man_ptr = param_ptr->scan_man_ptr;
    int band_index = param_ptr->band_index;
    int resample_dir = param_ptr->resample_dir;
    int last_line_complete_index = param_ptr->last_line_complete_index;
    OUTPUT_IMAGE_BUFFER_TYPE *output_buffer_ptr =
                                  param_ptr->output_image_buffer_ptr;
    RESAMPLE_TYPE resample_type = param_ptr->resample_type;

    int new_last_line_index = 0; /* the last line 
                        completed index allows the resampler to skip over
                        the majority of the rows that have already been 
                        completely resampled, instead of checking each
                        line individually. */

    int row_index;/* row index for output buffer */
    int *sample_reached_ptr = output_buffer_ptr->sample_reached_ptr; 
                  /* pointer to array of sample numbers reached in resampling */

    void *outbuf;                 /* Output image buffer           */
    double *middle_buf_ptr;       /* pointer to intermediate buffer allocated
                                   to allow resampling to be done in doubles
                                   then converted to whichever output type
                                   was selected by the user */
    double *mid_output_ptr;       /* pointer to current location to save data
                                   to in the middle_buf_ptr */

    const float *data_ptr;        /* Input data pointer             */
    float marker = 0.0;
    IAS_COEFFICIENTS *coeff_ptr;  /* pointer to reverse coefficients in grid */

    int  ll, ls;          /* Resampling weight table locations    */
    int line_counter;     /* line counter                         */
    int  grid_cols;       /* Number of columns in grid            */
    const double *kernel_weight_table_ptr = Get_Resample_Weight_Table_Ptr(); 
                          /* pointer to kernel weight table */

    const double *w_ptr;        /* Resampling weight table              */
    double round = 1.0 / 64.0;  /* Rounding error in resampling incr.   */
    double dl, ds;        /* Fractional line/sample offset        */
    double d_iline, d_isamp;  /* Input line/sample                    */
    int    i_iline, i_isamp;  /* integer input line/sample */
    int    grid_row_index;  /* index to start of row in grid coefficients */
    int    grid_cell_index; /* index to cell in row for grid coefficients */
    int    oline, osamp;    /* output line/sample */
    double total;                   /* Resampled pixel output               */
    double fill = get_fill_pixel(); /* fill pixel */

    int multi_proc_index; /* index used to split the resampling task
                           between threads */
    int  row_start;       /* starting row */
    int  row_end;         /* ending row */
    int  left_kernel_samps = num_left_kernel_samples();
                          /* samples to the left of the kernel center */
    int  top_kernel_lines = num_top_kernel_lines();
                          /* lines to the top of the kernel center */

    const WINDOW_TYPE *output_window_ptr = get_output_window(band_index); 
                          /* pointer to the output window geometry */
    int changed;          /* flag to indicate something has changed in the 
                           row buffer */
    int last_sample_was_reached = 1;  /* flag to indicate the last sample has 
                                       been reached for all lines so far */
    int granularity = output_buffer_ptr->granularity; /* granularity of the 
                        number of rows to resample at a time */

    double max_in_line; /* maximum input space line legal for resampling */
    double min_in_line; /* minimum input space line legal for resampling */
    double max_in_samp; /* maximum input space sample legal for resampling */
    double min_in_samp; /* minimum input space sample legal for resampling */

    /* resampling kernel information */
    int  kernel_samples;
    int  kernel_lines;
    int  kernel_subpixel_steps;
    double doub_kernel_subpixel_steps;
    int  kernel_table_steps;
    int  kernel_size;   /* samples * lines */

    if ((resample_type == CC_RESAMPLING) || (resample_type == MTF_RESAMPLING))
    {
        /* get the kernel information */
        get_kernel_table_info(&kernel_subpixel_steps, &kernel_table_steps,
                            &kernel_samples, &kernel_lines);
        doub_kernel_subpixel_steps = (double)kernel_subpixel_steps;
        kernel_size = kernel_samples * kernel_lines;

        /* verify the kernel size is correct for the loop that 
           actually does the resampling, which is hard coded to use
           a 4 line by 4 sample kernel for speed */
        ASSERT((kernel_lines == 4) && (kernel_samples == 4));

        /* set the input image limits */
        min_in_line = top_kernel_lines;
        max_in_line = in_nl - num_bottom_kernel_lines();
        min_in_samp = left_kernel_samps;
        max_in_samp = in_ns - num_right_kernel_samples();
    }
    else
    {
        IAS_LOG_ERROR("Unknown resampling type");
        param_ptr->thread_error = thread_num + 1;
        return ERROR;
    }

    /* allocate space for the intermediate buffer */
    middle_buf_ptr = malloc(sizeof(double) * grid_band_ptr->output_image_samps);
    if (middle_buf_ptr == NULL)
    {
        IAS_LOG_ERROR("Allocating intermediate buffer");
        param_ptr->thread_error = thread_num + 1;
        return ERROR;
    }

    /* cache the number of grid columns */
    grid_cols = grid_band_ptr->grid_cols;

    /* set the data pointer to a non-null value to allow correct detection
       of no resampling occurring.  This value will never be used except
       as a flag */
    data_ptr = &marker;

#ifdef VERBOSE
    printf("thread %d last line complete %d\n",thread_num,new_last_line_index);
#endif

    /*--------------------------------------------------------------------
      Loop through the requested output image window, performing resampling
    --------------------------------------------------------------------*/
    for (multi_proc_index = get_assigned_work(); 
        multi_proc_index < output_window_ptr->num_lines; 
       multi_proc_index = get_assigned_work())
    {
        /* make sure the work assignment succeeded (worked when >= 0) */
        if (multi_proc_index < 0)
        {
            IAS_LOG_ERROR("Allocating intermediate buffer");
            param_ptr->thread_error = thread_num + 1;
            return ERROR;
        }
 
        /* If any threads have encountered errors, don't bother doing any
           more processing */
        if (ias_threadpool_get_worker_return_status(param_ptr->threadpool)
            != SUCCESS)
        {
            IAS_LOG_WARNING("Exiting thread number %d since another thread "
                "encountered an error", thread_num);
            return ERROR;
        }

        /* calculate the starting and ending row for this pass */
        if (resample_dir > 0)
        {
            row_index = multi_proc_index + last_line_complete_index;
      
            /* resampling from the top to the bottom of output space, so 
               calculate the rows based on the top of the image */
            row_start = row_index * granularity + output_window_ptr->start_line;
            row_end = row_start + granularity;
    
            /* if starting row is past end of requested output window,
               break out of the loop */ 
            if (row_start >= output_window_ptr->end_line)
                break;

            /* make sure ending row is limited to the last line of the output 
               window */
            if (row_end >= output_window_ptr->end_line)
                row_end = output_window_ptr->end_line;
        }
        else
        {
            row_index = output_buffer_ptr->last_row_index - multi_proc_index
                  - last_line_complete_index - 1;
      
            /* resampling from the bottom to the top of output space, so 
               calculate the rows based on the bottom of the image */
            row_start = row_index * granularity + output_window_ptr->start_line;
            row_end = row_start + granularity;
    
            /* if starting row is before the first line of requested output 
               window, break out of the loop */ 
            if (row_start < output_window_ptr->start_line)
                break;

            /* make sure ending row is limited to the last line of the output 
               window */
            if (row_end >= output_window_ptr->end_line)
                row_end = output_window_ptr->end_line;
        }

        /* set the line counter to zero to allow detection of when no resampling
           has occurred for this set of rows */
        line_counter = 0;

        /* set the changed flag to indicate no output has changed yet */
        changed = 0;
    
#ifdef VERBOSE
       /* generate status message */
       printf("thread %ld working on rows %ld to %ld\n",thread_num, row_start,
           row_end);
#endif 
   
       /* get a buffer for the output data */
       outbuf = get_output_image_buffer(output_buffer_ptr, row_index, 
                                        thread_num);
       if (outbuf == NULL)
       {
           IAS_LOG_ERROR("Getting an output buffer");
           param_ptr->thread_error = thread_num + 1;
           return ERROR;
       }    
    
       for (oline = row_start; oline < row_end; oline++)
       {
           /* start at beginning of intermediate buffer */
           mid_output_ptr = middle_buf_ptr;

           /* convert the output line into a grid row */
           grid_row_index = oline/grid_band_ptr->cell_lines * grid_cols;
      
           /* loop through the samples in the output window */ 
           for (osamp = sample_reached_ptr[oline]; 
               osamp < output_window_ptr->end_sample; osamp++)
           {
               /* convert the output sample into a grid col */
               grid_cell_index = grid_row_index + 
                                 osamp/grid_band_ptr->cell_samps;

               /* get the pointer to the grid coefficients */
               coeff_ptr = &grid_band_ptr->reverse_coeffs[grid_cell_index];
        
               /* using the grid's reverse coefficients, calculate the input 
                  space line/sample */
               d_isamp = coeff_ptr->a[0] + coeff_ptr->a[1] * osamp + 
                  coeff_ptr->a[2] * oline + coeff_ptr->a[3] * osamp * oline;
               d_iline = coeff_ptr->b[0] + coeff_ptr->b[1] * osamp + 
                  coeff_ptr->b[2] * oline + coeff_ptr->b[3] * osamp * oline;
                
               /* if the input line/sample are within the input image bounds */
               if ((d_iline >= min_in_line) && (d_iline <  max_in_line) &&
                   (d_isamp >= min_in_samp) && (d_isamp <  max_in_samp))
               {
                   /* convert the input line/sample into whole and fractional 
                      line/sample */
                   i_iline = (int)d_iline;
                   i_isamp = (int)d_isamp;
                   dl = d_iline - i_iline;
                   ds = d_isamp - i_isamp;
        
                   /* get the resampling kernel index for the fractional 
                      line/sample */
                   ll = (int)(doub_kernel_subpixel_steps * (round + dl));
                   ls = (int)(doub_kernel_subpixel_steps * (round + ds));

                   /* get pointer to weight table location to use */ 
                   w_ptr = kernel_weight_table_ptr + (ll * kernel_table_steps 
                           + ls) * kernel_size;

                   /* get the address of the data at input line/sample */
                    data_ptr = Get_Sample_Data(scan_man_ptr, 
                                         i_iline - top_kernel_lines, 
                                         i_isamp - left_kernel_samps);

                   /* if the data pointer is NULL (i.e. data not in memory)
                      break out of the loop */
                   if (data_ptr == NULL)
                   {
                        break;
                   }

                   /*----------------------------------------------------
                     Perform resampling.
                   ----------------------------------------------------*/
                   total = 0.0;     /* initialize total to zero */

                   /* loop through the kernel lines for resampling */
                   for (line_counter = -1; line_counter < 3; line_counter++)
                   {
                       /* unroll the inner loop of kernel (from samples -1 to 
                          < 3) */
                       total += w_ptr[0] * data_ptr[0] +
                       w_ptr[1] * data_ptr[1] +
                       w_ptr[2] * data_ptr[2] +
                       w_ptr[3] * data_ptr[3]; 

                       /* advance weight pointer and input sample pointer */
                       w_ptr = &w_ptr[4];
                       data_ptr += in_ns;
                   }

                   /* put new resampled value in intermediate buffer, limiting 
                      it to the output range */
                   if (total < min_range)
                       *mid_output_ptr++ = min_range;
                   else if (total > max_range)
                       *mid_output_ptr++ = max_range;
                   else
                        *mid_output_ptr++ = total;
      
               }
               else
               {
                    /* pixel is outside of input image so set the output pixel 
                       to fill */
                    *mid_output_ptr++ = fill;
               }
          } /* end of line loop */

          /* if new resampling done, convert the data to the requested output 
             type and save it */
          if (osamp != sample_reached_ptr[oline])
          {
              /* convert the intermediate data to the final data type for the
                 sample of the output window that have been resampled in this
                 pass */
              convert_data(middle_buf_ptr, outbuf, 
                   sample_reached_ptr[oline] - output_window_ptr->start_sample,
                   osamp-output_window_ptr->start_sample, 
                 (oline - row_start) * output_window_ptr->num_samples);

              sample_reached_ptr[oline] = osamp;
      
              changed = 1;
        
              /* if the current line has been completed, update the new last 
                 line index to hold the current multiprocessing index (another 
                 way to say "set of output lines"). */
              if (last_sample_was_reached)
              {
                  if (osamp >= output_window_ptr->end_sample)
                      new_last_line_index = multi_proc_index;
                  else
                      last_sample_was_reached = 0;
              }
          }
      } /* end of row loop */

      /* return the buffer to the buffer pool if something has changed.  This 
         routine takes care of writing the data when needed. */
      if (changed)
      {
          if (return_output_image_buffer(output_buffer_ptr,row_index,
                                     row_start,row_end) != SUCCESS)
          {
              param_ptr->thread_error = thread_num + 1;
              return ERROR;
          }
      }

      /* If resampling has halted due to missing input data, break out of the 
         loop since no more output can be generated until a new set of input 
         scans has been read.  The (data_ptr == NULL) indicates that resampling
         of the last line of the set was halted due to lack of data if it 
         evaluates to true.  Since the resampling of a set of lines always 
         occurs top to bottom, the (line_counter == 0) makes sure that none of 
         the lines were resampled. This wouldn't be needed if the lines were 
         stepped through in the same direction as the outer multiprocessing 
         loop, but this isn't done since the row for loop would be very 
         confusing. */
      if ((data_ptr == NULL) && (line_counter == 0))
          break;
        
    } /* end of multiprocessing loop */

    free(middle_buf_ptr);

    /* calculate the new last line index so it can be used for the next set of 
       scans.  Guarantee that the most conservative value from all the threads
       ends up being saved.  This is needed since in some remote cases the last
       line completed by one thread may be past the last line completed by a
       different thread. */
    new_last_line_index += last_line_complete_index;
  
    if (IAS_THREAD_LOCK_MUTEX(&assignment_mutex) != 0)
    {
        IAS_LOG_ERROR("Entering critical section");
        param_ptr->thread_error = thread_num + 1;
        return ERROR;
    }

    if (last_line_complete_index == 
                param_ptr->returned_last_line_complete_index)
    {
        /* no other thread has updated the last line completed index so just
           update it with this thread's value */
        param_ptr->returned_last_line_complete_index = new_last_line_index;
    }
    else
    {
        /* the last line completed index has already been updated by a different
           thread so check for the more conservative value */
        if (new_last_line_index < param_ptr->returned_last_line_complete_index)
            param_ptr->returned_last_line_complete_index = new_last_line_index;
    }

#ifdef VERBOSE
    printf("thread %d last line complete %d\n",thread_num,new_last_line_index);
#endif
  
    if (IAS_THREAD_UNLOCK_MUTEX(&assignment_mutex) != 0)
    {
        IAS_LOG_ERROR("Exiting critical section");
        param_ptr->thread_error = thread_num + 1;
        return ERROR;
    }

    return SUCCESS;
}


/******************************************************************************
NAME:           resample_image_nn

PURPOSE:
The specific resample routine to perform nearest neighbor resampling.

RETURN VALUE:
The resampled data is returned.

******************************************************************************/

int resample_image_nn
(
    void * params,  /* I: pointer to input parameter structure. */
    int thread_num  /* I: thread number running this routine */
)
{
    /* create local copies of the parameters */
    RESAMPLE_THREAD_PARAMS *param_ptr = (RESAMPLE_THREAD_PARAMS *)params;
    const IAS_GEOM_GRID_BAND_TYPE *grid_band_ptr = param_ptr->grid_band_ptr;
    int  in_nl            = param_ptr->in_nl;
    int  in_ns            = param_ptr->in_ns;
    const SCAN_MAN_TYPE *scan_man_ptr = param_ptr->scan_man_ptr;
    int band_index = param_ptr->band_index;
    int resample_dir = param_ptr->resample_dir;
    int last_line_complete_index = param_ptr-> last_line_complete_index;
    OUTPUT_IMAGE_BUFFER_TYPE *output_buffer_ptr =
                                   param_ptr->output_image_buffer_ptr;
    RESAMPLE_TYPE resample_type = param_ptr->resample_type;

    int new_last_line_index = 0; /* the last line 
                        completed index allows the resampler to skip over
                        the majority of the rows that have already been 
                        completely resampled, instead of checking each
                        line individually. */

    int row_index;/* row index for output buffer */
    int *sample_reached_ptr = output_buffer_ptr->sample_reached_ptr; 
                  /* pointer to array of sample numbers reached in resampling */

    void *outbuf;                 /* Output image buffer           */
    double *middle_buf_ptr;       /* pointer to intermediate buffer allocated
                                   to allow resampling to be done in doubles
                                   then converted to whichever output type
                                   was selected by the user */
    double *mid_output_ptr;       /* pointer to current location to save data
                                   to in the middle_buf_ptr */

    const float *data_ptr;        /* Input data pointer               */
    float marker = 0.0;
    IAS_COEFFICIENTS *coeff_ptr;  /* pointer to reverse coefficients in grid */

    int new_data_resampled; /* flag to indicate new data has been generated */
    int  grid_cols;                /* Number of columns in grid            */

    double d_iline, d_isamp;  /* Input line/sample                    */
    int    i_iline, i_isamp;  /* integer input line/sample */
    int    grid_row_index;    /* index to start of row in grid coefficients */
    int    grid_cell_index;   /* index to cell in row for grid coefficients */
    int    oline, osamp;      /* output line/sample */
    double total;                /* Resampled pixel output               */
    double fill = get_fill_pixel(); /* fill pixel */

    int multi_proc_index; /* index used to split the resampling task
                           between threads */
    int  row_start;       /* starting row */
    int  row_end;         /* ending row */

    const WINDOW_TYPE *output_window_ptr = get_output_window(band_index); 
                          /* pointer to the output window geometry */
    int changed;          /* flag to indicate something has changed in the 
                           row buffer */
    int last_sample_was_reached = 1;  /* flag to indicate the last sample has 
                                       been reached for all lines so far */
    int granularity = output_buffer_ptr->granularity; /* granularity of the 
                        number of rows to resample at a time */

    double max_in_line; /* maximum input space line legal for resampling */
    double min_in_line; /* minimum input space line legal for resampling */
    double max_in_samp; /* maximum input space sample legal for resampling */
    double min_in_samp; /* minimum input space sample legal for resampling */

    if (resample_type == NN_RESAMPLING)
    {
        min_in_line = 0;
        max_in_line = in_nl;
        min_in_samp = 0;
        max_in_samp = in_ns;
    }
    else
    {
        IAS_LOG_ERROR("Unknown resampling type");
        param_ptr->thread_error = thread_num + 1;
        return ERROR;
    }

    /* allocate space for the intermediate buffer */
    middle_buf_ptr = malloc(sizeof(double) * grid_band_ptr->output_image_samps);
    if (middle_buf_ptr == NULL)
    {
        IAS_LOG_ERROR("Allocating intermediate buffer");
        param_ptr->thread_error = thread_num + 1;
        return ERROR;
    }

    /* cache the number of grid columns */
    grid_cols = grid_band_ptr->grid_cols;

    /* set the data pointer to a non-null value to allow correct detection
       of no resampling occurring.  This value will never be used except
       as a flag */
    data_ptr = &marker;

#ifdef VERBOSE
    printf("thread %d last line complete %d\n",thread_num,new_last_line_index);
#endif

    /*--------------------------------------------------------------------
      Loop through the requested output image window, performing resampling
    --------------------------------------------------------------------*/
    for (multi_proc_index = get_assigned_work(); 
         multi_proc_index < output_window_ptr->num_lines; 
         multi_proc_index = get_assigned_work())
    {
        /* make sure the work assignment succeeded (worked when >= 0) */
        if (multi_proc_index < 0)
        {
            IAS_LOG_ERROR("Allocating intermediate buffer");
            param_ptr->thread_error = thread_num + 1;
            return ERROR;
        }
    
        /* If any threads have encountered errors, don't bother doing any
           more processing */
        if (ias_threadpool_get_worker_return_status(param_ptr->threadpool)
            != SUCCESS)
        {
            IAS_LOG_WARNING("Exiting thread number %d since another thread "
                "encountered an error", thread_num);
            return ERROR;
        }

        /* calculate the starting and ending row for this pass */
        if (resample_dir > 0)
        {
            row_index = multi_proc_index + last_line_complete_index;
      
            /* resampling from the top to the bottom of output space, so 
               calculate the rows based on the top of the image */
            row_start = row_index * granularity + output_window_ptr->start_line;
            row_end = row_start + granularity;
    
            /* if starting row is past end of requested output window,
               break out of the loop */ 
            if (row_start >= output_window_ptr->end_line)
                break;

            /* make sure ending row is limited to the last line of the output 
               window */
            if (row_end >= output_window_ptr->end_line)
                row_end = output_window_ptr->end_line;
        }
        else
        {
            row_index = output_buffer_ptr->last_row_index - multi_proc_index
                  - last_line_complete_index - 1;
      
            /* resampling from the bottom to the top of output space, so 
               calculate the rows based on the bottom of the image */
            row_start = row_index * granularity + output_window_ptr->start_line;
            row_end = row_start + granularity;
    
            /* if starting row is before the first line of requested output 
               window, break out of the loop */ 
            if (row_start < output_window_ptr->start_line)
                break;

            /* make sure ending row is limited to the last line of the output 
               window */
            if (row_end >= output_window_ptr->end_line)
                row_end = output_window_ptr->end_line;
         }

         /* clear the new data has been generated flag to allow detection of 
            when no resampling has occurred for this set of rows */
         new_data_resampled = 0;

         /* set the changed flag to indicate no output has changed yet */
         changed = 0;
    
#ifdef VERBOSE
        /* generate status message */
        printf("thread %ld working on rows %ld to %ld\n",thread_num, row_start,
           row_end);
#endif 
   
        /* get a buffer for the output data */
        outbuf = get_output_image_buffer(output_buffer_ptr, row_index, 
                                         thread_num);
        if (outbuf == NULL)
        {
            IAS_LOG_ERROR("Getting an output buffer");
            param_ptr->thread_error = thread_num + 1;
            return ERROR;
        }    
    
        for (oline = row_start; oline < row_end; oline++)
        {
            /* start at beginning of intermediate buffer */
            mid_output_ptr = middle_buf_ptr;

            /* convert the output line into a grid row */
            grid_row_index = oline/grid_band_ptr->cell_lines * grid_cols;
      
            /* loop through the samples in the output window */ 
            for (osamp = sample_reached_ptr[oline]; 
                osamp < output_window_ptr->end_sample; osamp++)
            {
                /* convert the output sample into a grid col */
                grid_cell_index = grid_row_index + 
                                  osamp/grid_band_ptr->cell_samps;

                /* get the pointer to the grid coefficients */
                coeff_ptr = &grid_band_ptr->reverse_coeffs[grid_cell_index];
        
                /* using the grid's reverse coefficients, calculate the input 
                   space line/sample */
                d_isamp = coeff_ptr->a[0] + coeff_ptr->a[1] * osamp + 
                    coeff_ptr->a[2] * oline + coeff_ptr->a[3] * osamp * oline;
                d_iline = coeff_ptr->b[0] + coeff_ptr->b[1] * osamp + 
                    coeff_ptr->b[2] * oline + coeff_ptr->b[3] * osamp * oline;
        
                /* add half a pixel to round to nearest neighbor */
                d_isamp += 0.5;
                d_iline += 0.5;
        
                /* if the input line/sample are within the input image bounds */
                if ((d_iline >= min_in_line) && (d_iline <  max_in_line) &&
                    (d_isamp >= min_in_samp) && (d_isamp <  max_in_samp))
                {
                    /* convert the input line/sample into an integer */
                    i_iline = (int)d_iline;
                    i_isamp = (int)d_isamp;
        
                    /* get the address of the data at input line/sample */
                    data_ptr = Get_Sample_Data(scan_man_ptr, i_iline, i_isamp);

                    /* if the data pointer is NULL (i.e. data not in memory)
                       break out of the loop */
                    if (data_ptr == NULL)
                    {
                        break;
                    }
          
                    /* set the new data generated flag */
                    new_data_resampled = 1;

                    /* get the nearest input pixel */
                    total = *data_ptr;

                    /* put new resampled value in intermediate buffer, limiting
                       it to the output range */
                    if (total < min_range)
                        *mid_output_ptr++ = min_range;
                    else if (total > max_range)
                        *mid_output_ptr++ = max_range;
                    else
                        *mid_output_ptr++ = total;
      
                 }
                 else
                 {
                     /* pixel is outside of input image so set the output pixel
                        to fill */
                     *mid_output_ptr++ = fill;
                 }
             } /* end of line loop */

             /* if new resampling done, convert the data to the requested
                output type and save it */
             if (osamp != sample_reached_ptr[oline])
             {
                 /* convert the intermediate data to the final data type for the
                    sample of the output window that have been resampled in this
                    pass */
                 convert_data(middle_buf_ptr, outbuf, 
                        sample_reached_ptr[oline] - 
                        output_window_ptr->start_sample,
                        osamp-output_window_ptr->start_sample, 
                        (oline - row_start) * output_window_ptr->num_samples);

                 sample_reached_ptr[oline] = osamp;
      
                 changed = 1;
        
                 /* if the current line has been completed, update the new last
                    line index to hold the current multiprocessing index 
                    (another way to say "set of output lines"). */
                 if (last_sample_was_reached)
                 {
                     if (osamp >= output_window_ptr->end_sample)
                          new_last_line_index = multi_proc_index;
                     else
                          last_sample_was_reached = 0;
                 }
             }
         } /* end of row loop */

         /* return the buffer to the buffer pool if something has changed.
            This routine takes care of writing the data when needed. */
         if (changed)
         {
             if (return_output_image_buffer(output_buffer_ptr,row_index,
                                     row_start,row_end) != SUCCESS)
             {
                 param_ptr->thread_error = thread_num + 1;
                 return ERROR;
             }
         }

         /* If resampling has halted due to missing input data, break out of
            the loop since no more output can be generated until a new set of
            input scans has been read.  The (data_ptr == NULL) indicates that
            resampling of the last line of the set was halted due to lack of 
            data if it evaluates to true.  Since the resampling of a set of 
            lines always occurs top to bottom, the (new_data_resampled == 0) 
            makes sure that none of the lines were resampled.  This wouldn't be
            needed if the lines were stepped through in the same direction as 
            the outer multiprocessing loop, but this isn't done since the row 
            for loop would be very confusing. */
         if ((data_ptr == NULL) && (new_data_resampled == 0))
             break;
        
      } /* end of multiprocessing loop */

      free(middle_buf_ptr);

    /* calculate the new last line index so it can be used for the next set of 
     scans.  Guarantee that the most conservative value from all the threads
     ends up being saved.  This is needed since in some remote cases the last
     line completed by one thread may be past the last line completed by a
     different thread. */
    new_last_line_index += last_line_complete_index;
  
    if (IAS_THREAD_LOCK_MUTEX(&assignment_mutex) != 0)
    {
        IAS_LOG_ERROR("Entering critical section");
        param_ptr->thread_error = thread_num + 1;
        return ERROR;
    }
  
    if (last_line_complete_index == 
                param_ptr->returned_last_line_complete_index)
    {
        /* no other thread has updated the last line completed index so just
           update it with this thread's value */
        param_ptr->returned_last_line_complete_index = new_last_line_index;
    }
    else
    {
        /* the last line completed index has already been updated by a different
         thread so check for the more conservative value */
        if (new_last_line_index < param_ptr->returned_last_line_complete_index)
            param_ptr->returned_last_line_complete_index = new_last_line_index;
    }

#ifdef VERBOSE
    printf("thread %d last line complete %d\n",thread_num,new_last_line_index);
#endif
  
    if (IAS_THREAD_UNLOCK_MUTEX(&assignment_mutex) != 0)
    {
        IAS_LOG_ERROR("Exiting critical section");
        param_ptr->thread_error = thread_num + 1;
        return ERROR;
    }

    return SUCCESS;
  
}

/******************************************************************************
NAME:           resample_image_bi

PURPOSE:
The specific resample routine to perform bilinear interpolation resampling.

RETURN VALUE:
The resampled data is returned.

NOTES:
none

ALGORITHM REFERENCES:
none

******************************************************************************/

int resample_image_bi
(
    void * params,  /* I: pointer to input parameter structure. */
    int thread_num  /* I: thread number running this routine */
)
{
    /* create local copies of the parameters */
    RESAMPLE_THREAD_PARAMS *param_ptr = (RESAMPLE_THREAD_PARAMS *)params;
    const IAS_GEOM_GRID_BAND_TYPE *grid_band_ptr = param_ptr->grid_band_ptr;
    int  in_nl            = param_ptr->in_nl;
    int  in_ns            = param_ptr->in_ns;
    const SCAN_MAN_TYPE *scan_man_ptr = param_ptr->scan_man_ptr;
    int band_index = param_ptr->band_index;
    int resample_dir = param_ptr->resample_dir;
    int last_line_complete_index = param_ptr->last_line_complete_index;
    OUTPUT_IMAGE_BUFFER_TYPE *output_buffer_ptr =
                                   param_ptr->output_image_buffer_ptr;
    RESAMPLE_TYPE resample_type = param_ptr->resample_type;

    int new_last_line_index = 0; /* the last line 
                        completed index allows the resampler to skip over
                        the majority of the rows that have already been 
                        completely resampled, instead of checking each
                        line individually. */

    int row_index;/* row index for output buffer */
    int *sample_reached_ptr = output_buffer_ptr->sample_reached_ptr; 
                  /* pointer to array of sample numbers reached in resampling */

    void *outbuf;                 /* Output image buffer           */
    double *middle_buf_ptr;       /* pointer to intermediate buffer allocated
                                   to allow resampling to be done in doubles
                                   then converted to whichever output type
                                   was selected by the user */
    double *mid_output_ptr;       /* pointer to current location to save data
                                   to in the middle_buf_ptr */

    const float *data_ptr;        /* Input data pointer               */
    float marker = 0.0;
    IAS_COEFFICIENTS *coeff_ptr;  /* pointer to reverse coefficients in grid */

    int new_data_resampled; /* flag to indicate new data has been generated */
    int grid_cols;                  /* Number of columns in grid            */

    double dl, ds;          /* Fractional line/sample offset        */
    double d_iline, d_isamp;  /* Input line/sample                    */
    int    i_iline, i_isamp;  /* integer input line/sample */
    int    grid_row_index;  /* index to start of row in grid coefficients */
    int    grid_cell_index; /* index to cell in row for grid coefficients */
    int    oline, osamp;    /* output line/sample */
    double total;                   /* Resampled pixel output               */
    double fill = get_fill_pixel(); /* fill pixel */

    int multi_proc_index; /* index used to split the resampling task
                           between threads */
    int  row_start;       /* starting row */
    int  row_end;         /* ending row */
    int  left_kernel_samps = num_left_kernel_samples();
                        /* samples to the left of the kernel center */
    int  top_kernel_lines = num_top_kernel_lines();
                        /* lines to the top of the kernel center */

    const WINDOW_TYPE *output_window_ptr = get_output_window(band_index); 
                        /* pointer to the output window geometry */
    int changed;        /* flag to indicate something has changed in the 
                           row buffer */
    int last_sample_was_reached = 1;  /* flag to indicate the last sample has 
                                       been reached for all lines so far */
    int granularity = output_buffer_ptr->granularity; /* granularity of the 
                        number of rows to resample at a time */

    double max_in_line; /* maximum input space line legal for resampling */
    double min_in_line; /* minimum input space line legal for resampling */
    double max_in_samp; /* maximum input space sample legal for resampling */
    double min_in_samp; /* minimum input space sample legal for resampling */

    /* resampling kernel information */
    int  kernel_samples;
    int  kernel_lines;
    int  kernel_subpixel_steps;
    int  kernel_table_steps;

    if (resample_type == BI_RESAMPLING)
    {
        /* get the kernel information */
        get_kernel_table_info(&kernel_subpixel_steps, &kernel_table_steps,
                            &kernel_samples, &kernel_lines);

        /* verify the kernel size is correct for bilinear resampling */
        ASSERT((kernel_lines == 2) && (kernel_samples == 2));

        /* set the input image limits */
        min_in_line = top_kernel_lines;
        max_in_line = in_nl - num_bottom_kernel_lines();
        min_in_samp = left_kernel_samps;
        max_in_samp = in_ns - num_right_kernel_samples();
    }
    else
    {
        IAS_LOG_ERROR("Unknown resampling type");
        param_ptr->thread_error = thread_num + 1;
        return ERROR;
    }

    /* allocate space for the intermediate buffer */
    middle_buf_ptr = malloc(sizeof(double) * grid_band_ptr->output_image_samps);
    if (middle_buf_ptr == NULL)
    {
        IAS_LOG_ERROR("Allocating intermediate buffer");
        param_ptr->thread_error = thread_num + 1;
        return ERROR;
    }

    /* cache the number of grid columns */
    grid_cols = grid_band_ptr->grid_cols;

    /* set the data pointer to a non-null value to allow correct detection
       of no resampling occurring.  This value will never be used except
       as a flag */
    data_ptr = &marker;

#ifdef VERBOSE
    printf("thread %d last line complete %d\n",thread_num,new_last_line_index);
#endif

    /*--------------------------------------------------------------------
      Loop through the requested output image window, performing resampling
    --------------------------------------------------------------------*/
    for (multi_proc_index = get_assigned_work(); 
        multi_proc_index < output_window_ptr->num_lines; 
        multi_proc_index = get_assigned_work())
    {
        /* make sure the work assignment succeeded (worked when >= 0) */
        if (multi_proc_index < 0)
        {
            IAS_LOG_ERROR("Allocating intermediate buffer");
            param_ptr->thread_error = thread_num + 1;
            return ERROR;
        }
    
        /* If any threads have encountered errors, don't bother doing any
           more processing */
        if (ias_threadpool_get_worker_return_status(param_ptr->threadpool)
            != SUCCESS)
        {
            IAS_LOG_WARNING("Exiting thread number %d since another thread "
                "encountered an error", thread_num);
            return ERROR;
        }

        /* calculate the starting and ending row for this pass */
        if (resample_dir > 0)
        {
            row_index = multi_proc_index + last_line_complete_index;
            /* resampling from the top to the bottom of output space, so 
               calculate the rows based on the top of the image */
            row_start = row_index * granularity + output_window_ptr->start_line;
            row_end = row_start + granularity;
    
            /* if starting row is past end of requested output window,
               break out of the loop */ 
               if (row_start >= output_window_ptr->end_line)
                   break;

           /* make sure ending row is limited to the last line of the output 
              window */
           if (row_end >= output_window_ptr->end_line)
               row_end = output_window_ptr->end_line;
        }
        else
        {
            row_index = output_buffer_ptr->last_row_index - multi_proc_index
                  - last_line_complete_index - 1;
      
            /* resampling from the bottom to the top of output space, so 
               calculate the rows based on the bottom of the image */
            row_start = row_index * granularity + output_window_ptr->start_line;
            row_end = row_start + granularity;
    
            /* if starting row is before the first line of requested output 
               window, break out of the loop */ 
            if (row_start < output_window_ptr->start_line)
                break;

            /* make sure ending row is limited to the last line of the output 
               window */
            if (row_end >= output_window_ptr->end_line)
                row_end = output_window_ptr->end_line;
        }

        /* clear the new data has been generated flag to allow detection of 
           when no resampling has occurred for this set of rows */
        new_data_resampled = 0;

        /* set the changed flag to indicate no output has changed yet */
        changed = 0;
    
#ifdef VERBOSE
        /* generate status message */
        printf("thread %ld working on rows %ld to %ld\n",thread_num, row_start,
                row_end);
#endif 
   
        /* get a buffer for the output data */
        outbuf = get_output_image_buffer(output_buffer_ptr, row_index, 
                                         thread_num);
        if (outbuf == NULL)
        {
            IAS_LOG_ERROR("Getting an output buffer");
            param_ptr->thread_error = thread_num + 1;
            return ERROR;
        }    
    
        for (oline = row_start; oline < row_end; oline++)
        {
            /* start at beginning of intermediate buffer */
            mid_output_ptr = middle_buf_ptr;

            /* convert the output line into a grid row */
            grid_row_index = oline/grid_band_ptr->cell_lines * grid_cols;
      
            /* loop through the samples in the output window */ 
            for (osamp = sample_reached_ptr[oline]; 
                 osamp < output_window_ptr->end_sample; osamp++)
            {
                /* convert the output sample into a grid col */
                grid_cell_index = grid_row_index + 
                                  osamp/grid_band_ptr->cell_samps;

                /* get the pointer to the grid coefficients */
                coeff_ptr = &grid_band_ptr->reverse_coeffs[grid_cell_index];
        
                /* using the grid's reverse coefficients, calculate the input 
                   space line/sample */
                d_isamp = coeff_ptr->a[0] + coeff_ptr->a[1] * osamp + 
                     coeff_ptr->a[2] * oline + coeff_ptr->a[3] * osamp * oline;
                d_iline = coeff_ptr->b[0] + coeff_ptr->b[1] * osamp + 
                     coeff_ptr->b[2] * oline + coeff_ptr->b[3] * osamp * oline;
                
                /* if the input line/sample are within the input image bounds */
                if ((d_iline >= min_in_line) && (d_iline <  max_in_line) &&
                    (d_isamp >= min_in_samp) && (d_isamp <  max_in_samp))
                {
                    /* convert the input line/sample into whole and fractional 
                       line/sample */
                    i_iline = (int)d_iline;
                    i_isamp = (int)d_isamp;
                    dl = d_iline - i_iline;
                    ds = d_isamp - i_isamp;

                    /* get the address of the data at input line/sample */
                    data_ptr = Get_Sample_Data(scan_man_ptr, i_iline - 
                        top_kernel_lines, i_isamp - left_kernel_samps);

                    /* if the data pointer is NULL (i.e. data not in memory)
                        break out of the loop */
                    if (data_ptr == NULL)
                    {
                        break;
                    }
          
                    /* set the new data generated flag */
                    new_data_resampled = 1;
          
                    /*----------------------------------------------------
                      Perform resampling.
                     ----------------------------------------------------*/
                    total = (1.0 - ds) * (1.0 - dl) * data_ptr[0] +
                         ds  * (1.0 - dl) * data_ptr[1] +
                         (1.0 - ds) *        dl  * data_ptr[in_ns] +
                         ds  *        dl  * data_ptr[in_ns + 1];
          

                   /* put new resampled value in intermediate buffer, limiting 
                      it to the output range */
                   if (total < min_range)
                       *mid_output_ptr++ = min_range;
                   else if (total > max_range)
                       *mid_output_ptr++ = max_range;
                   else
                       *mid_output_ptr++ = total;
      
                }
                else
                {
                    /* pixel is outside of input image so set the output pixel 
                       to fill */
                    *mid_output_ptr++ = fill;
                }
            } /* end of line loop */

            /* if new resampling done, convert the data to the requested output 
               type and save it */
            if (osamp != sample_reached_ptr[oline])
            {
                /* convert the intermediate data to the final data type for the
                   sample of the output window that have been resampled in this
                   pass */
                convert_data(middle_buf_ptr, outbuf, 
                   sample_reached_ptr[oline] - output_window_ptr->start_sample,
                   osamp-output_window_ptr->start_sample, 
                   (oline - row_start) * output_window_ptr->num_samples);

               sample_reached_ptr[oline] = osamp;
      
               changed = 1;
        
            /* if the current line has been completed, update the new last line 
               index to hold the current multiprocessing index (another way to 
               say "set of output lines"). */
            if (last_sample_was_reached)
            {
                if (osamp >= output_window_ptr->end_sample)
                    new_last_line_index = multi_proc_index;
                else
                    last_sample_was_reached = 0;
            }
        }
    } /* end of row loop */

    /* return the buffer to the buffer pool if something has changed.  This 
       routine takes care of writing the data when needed. */
    if (changed)
    {
        if (return_output_image_buffer(output_buffer_ptr,row_index,
                                     row_start,row_end) != SUCCESS)
        {
            param_ptr->thread_error = thread_num + 1;
            return ERROR;
        }
    }

    /* If resampling has halted due to missing input data, break out of the 
       loop since no more output can be generated until a new set of input 
       scans has been read.  The (data_ptr == NULL) indicates that resampling
       of the last line of the set was halted due to lack of data if it 
       evaluates to true.  Since the resampling of a set of lines always occurs
       top to bottom, the (new_data_resampled == 0) makes sure that none of the 
       lines were resampled.  This wouldn't be needed if the lines were stepped 
       through in the same direction as the outer multiprocessing loop, but this
       isn't done since the row for loop would be very confusing. */
      if ((data_ptr == NULL) && (new_data_resampled == 0))
          break;
        
    } /* end of multiprocessing loop */

    free(middle_buf_ptr);

    /* calculate the new last line index so it can be used for the next set of 
       scans.  Guarantee that the most conservative value from all the threads
       ends up being saved.  This is needed since in some remote cases the last
       line completed by one thread may be past the last line completed by a
       different thread. */
    new_last_line_index += last_line_complete_index;
 
    if (IAS_THREAD_LOCK_MUTEX(&assignment_mutex) != 0)
    {
        IAS_LOG_ERROR("Entering critical section");
        param_ptr->thread_error = thread_num + 1;
        return ERROR;
    }
  
    if (last_line_complete_index == 
               param_ptr->returned_last_line_complete_index)
    {
        /* no other thread has updated the last line completed index so just
           update it with this thread's value */
        param_ptr->returned_last_line_complete_index = new_last_line_index;
    }
    else
    {
        /* the last line completed index has already been updated by a different
           thread so check for the more conservative value */
        if (new_last_line_index < param_ptr->returned_last_line_complete_index)
            param_ptr->returned_last_line_complete_index = new_last_line_index;
    }

#ifdef VERBOSE
    printf("thread %d last line complete %d\n",thread_num,new_last_line_index);
#endif

    if (IAS_THREAD_UNLOCK_MUTEX(&assignment_mutex) != 0)
    {
        IAS_LOG_ERROR("Exiting critical section");
        param_ptr->thread_error = thread_num + 1;
        return ERROR;
    }

    return SUCCESS;
}
