/******************************************************************************
NAME:           scanman

PURPOSE:        
The scanman (short for Scan Manager) manages the buffer of scans currently 
residing in memory.  For flexibility, the number of scans to buffer is a 
parameter for initializing the scan manager.  This allows from a few scans
to an entire image to be buffered.  The name is not really accurate anymore 
since this generic version of the resampler manages the input image on
a line basis instead of a scan basis.  This is because the input image is 
not split into scans like the L7 input image.


ROUTINES:
    Initialize_Scan_Manager
    Get_Sample_Data
    Read_Next_Scan_Set
    Cleanup_Scan_Manager
    More_Scans_To_Read

NOTES:
The user must first initialize a scan manager, then read the first set of
lines in by calling the Read_Next_Scan_Set routine before attempting to 
get sample data.  Multiple scan managers may be created to read multiple 
bands, or even different areas of the same band.

ALGORITHM REFERENCES:
None

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "ias_const.h"   /* SUCCESS/ERROR, TRUE/FALSE definitions */
#include "ias_logging.h" /* IAS logging prototype */
#include "ias_satellite_attributes.h" /* IAS_MAX_NBANDS */
#include "scanman.h"    /* prototypes for this module */
#include "debug.h"      /* ASSERT macro */
#include "get_resample_parameters.h" /* get_0R_filename prototype */
#include "read_input_image.h" /* routines to read input image */

#define LINE_HYSTERESIS 160
/* defines the number of lines to use for hysteresis when reading input lines
   into the buffer.  When the next group of lines is read, this is the 
   number of lines that will overlap. */

/******************************************************************************
FUNCTION NAME: Initialize_Scan_Manager

PURPOSE:        
Initialize_Scan_Manager does the initial setup of an instance of a scan
manager.

RETURNS:
A pointer to the scan manager state structure is returned.  If an error
occurs, an error message is generated and NULL is returned.

******************************************************************************/

SCAN_MAN_TYPE *Initialize_Scan_Manager
(
    const IAS_GEOM_GRID_BAND_TYPE *grid_band_ptr, /* I: pointer to grid band */
    int  band_index,            /* I: band index to read */
    int  lines_to_buffer,       /* I: number of lines to buffer at a time */
    int  *samples_per_line_ptr, /* O: number of samples per line */
    int  *lines_in_image_ptr,   /* O: number of lines in the input image */
    int  kernel_size            /* I: number of lines in kernel */
)
{
    SCAN_MAN_TYPE *scan_man_ptr; /* pointer to allocated scan manager struct */
    double left_line;       /* line number of point on left of input image */
    double right_line;      /* line number of point on right of input image */

    /* verify there are enough lines in the buffer */
    ASSERT (lines_to_buffer > LINE_HYSTERESIS + 2);

    /* verify the band number is legal */
    ASSERT ((band_index >= 0) && (band_index < IAS_MAX_NBANDS));

    /* verify other parameters are legal */
    ASSERT (kernel_size >= 0);

    /* allocate memory for the scan manager structure */
    scan_man_ptr = (SCAN_MAN_TYPE *)malloc(sizeof(SCAN_MAN_TYPE));
    if (scan_man_ptr == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for scan manager structure");
        return NULL;
    }

    /* initialize members of the scan manager structure */
    scan_man_ptr->image_buffer_ptr = NULL;
    scan_man_ptr->grid_band_ptr = grid_band_ptr;
    scan_man_ptr->band_index = band_index;
    scan_man_ptr->lines_in_buffer = lines_to_buffer;
    scan_man_ptr->kernel_size = kernel_size;

    /* open the input file */
    scan_man_ptr->in_image_ptr = open_input_image(get_input_image_filename(), 
                                band_index, lines_in_image_ptr, 
                                samples_per_line_ptr);
    if (scan_man_ptr->in_image_ptr == NULL)
    {
        IAS_LOG_ERROR("Returned from open_input_image for file %s\n",
                 get_input_image_filename());
        return NULL;
    }

    scan_man_ptr->samples_per_line = *samples_per_line_ptr;
    scan_man_ptr->lines_in_image = *lines_in_image_ptr;

    /* allocate memory for the line buffer */
    scan_man_ptr->image_buffer_ptr = (float *)malloc(sizeof(float) * 
                                *samples_per_line_ptr * lines_to_buffer);
    if (scan_man_ptr->image_buffer_ptr == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for input image buffer");
        close_input_image(scan_man_ptr->in_image_ptr);
        free (scan_man_ptr);
        return NULL;
    }

    /* determine if the file should be read top to bottom or bottom to top.
       This is done by using the first set of in_lines in the grid.  If the 
       left input space line number is smaller than the right input space
       line number, the lines in input space slope upward relative to the 
       lines in output space as you go from left to right.  Therefore, the
       image needs to be read from the top to the bottom (forward).  
       Otherwise the image needs to be read from bottom to top (reverse). */
    left_line = grid_band_ptr->in_lines[0]; 
    right_line = grid_band_ptr->in_lines[grid_band_ptr->num_out_samps-1]; 
     
    /* set scan_direction as a result */
    if (left_line <= right_line)
    {
        scan_man_ptr->scan_direction = READ_SCAN_FORWARD;
    }
    else
    {
        scan_man_ptr->scan_direction = READ_SCAN_REVERSE;
    }

    /* flag that the data has not been read yet */  
    scan_man_ptr->data_valid = 0;
    
    return scan_man_ptr;
}

/******************************************************************************
FUNCTION NAME: Cleanup_Scan_Manager

PURPOSE:        
Cleanup_Scan_Manager frees the memory allocated for a scan manager structure.

NOTES:
- After this call, the scan_man_ptr will no longer contain valid data.

******************************************************************************/

void Cleanup_Scan_Manager
(
    SCAN_MAN_TYPE *scan_man_ptr /* I/O: pointer to scan manager to clean up */
)
{
    if (scan_man_ptr != NULL)
    {
        /* close the input file */
        close_input_image(scan_man_ptr->in_image_ptr);

        if (scan_man_ptr->image_buffer_ptr != NULL)
            free (scan_man_ptr->image_buffer_ptr);
        
        free (scan_man_ptr);
    }
}

/******************************************************************************
FUNCTION NAME: Read_Next_Scan_Set

PURPOSE:        
Read_Next_Scan_Set reads in the next set of lines that will fit into the
scan buffer.  

RETURNS:
The scans and associated members of the scan manager structure are updated.
The function returns a SUCCESS or ERROR indication.

NOTES:
This routine loads the next set of image lines into memory.

******************************************************************************/

int Read_Next_Scan_Set
(
    SCAN_MAN_TYPE *scan_man_ptr /* I/O: pointer to scan manager structure */
)
{
    int  start_line, end_line; /* starting and ending lines to be read */

    /* if a range of lines has not been previously read, initialize the
       line location variables */
    if (!scan_man_ptr->data_valid)
    {
        /* set the starting scan to match the direction the file must 
           be read in */
        if (scan_man_ptr->scan_direction == READ_SCAN_FORWARD)
            start_line = 0;
        else
            start_line = scan_man_ptr->lines_in_image - 
                         scan_man_ptr->lines_in_buffer;

       /* flag that the scan data is now valid */
       scan_man_ptr->data_valid = 1;
    }
    /* calculate range of lines to read */
    else if (scan_man_ptr->scan_direction == READ_SCAN_FORWARD)
    { 
        /* advance to the next set of lines, overlapping the lines
           with LINE_HYSTERESIS scans to allow for the kernel size */
        start_line = scan_man_ptr->start_line + scan_man_ptr->lines_in_buffer - 
                     LINE_HYSTERESIS;
    }
    else
    {
        /* verify the scan direction is legal */
        ASSERT(scan_man_ptr->scan_direction == READ_SCAN_REVERSE);

        /* fall back to the next set of lines, with LINE_HYSTERESIS as 
           above */
        start_line = scan_man_ptr->start_line - scan_man_ptr->lines_in_buffer +
                     LINE_HYSTERESIS;
    }
  
    end_line = start_line + scan_man_ptr->lines_in_buffer; 

    /* make sure the start and end lines are legal */
    if (start_line < 0)
        start_line = 0;

    if (end_line > scan_man_ptr->lines_in_image)
        end_line = scan_man_ptr->lines_in_image;

    scan_man_ptr->start_line = start_line;

    /* calculate cached values */
    scan_man_ptr->last_valid_line = end_line - scan_man_ptr->kernel_size + 1;

    /* log a status message to indicate how far the resampler is on 
       the current band */    
    IAS_LOG_INFO("resampling lines %d through %d",start_line,end_line);

    /* read in the new lines */
    if (read_input_lines(scan_man_ptr->in_image_ptr, start_line, 
                         end_line - start_line, scan_man_ptr->image_buffer_ptr) 
                         != SUCCESS)
    {
        IAS_LOG_ERROR("Reading from source image");
        return ERROR;
    }

    return SUCCESS;
}


/******************************************************************************
FUNCTION NAME: Get_Sample_Data

PURPOSE:        
Get_Sample_Data accepts a line/sample pair and returns
the address in memory where the sample resides.

RETURNS:
The address of the sample located at the line/sample is returned.  If the 
sample is not in memory or the kernel size will take exceed the samples
in memory, a NULL is returned.

NOTES:
The line and sample passed in should represent the upper left corner of the
resampling kernel location.

******************************************************************************/

const float *Get_Sample_Data
(
    const SCAN_MAN_TYPE *scan_man_ptr, /*I: pointer to scan manager structure */
    int  line,    /* I: line number of the upper left corner of the kernel */
    int  sample   /* I: sample number of the upper left corner of the kernel */
)
{
    int  index; /* index to the line/sample in the extended image */

    /* verify the input parameters */
    ASSERT ((line >= 0) && (line < scan_man_ptr->lines_in_image)); 
    ASSERT ((sample >= 0) && (sample < scan_man_ptr->samples_per_line));

    /* verify the requested line actually lies in memory */
    if ((line >= scan_man_ptr->start_line)
        && (line < scan_man_ptr->last_valid_line))
    {
        /* calculate the index relative to the beginning of the band */
        index = (line - scan_man_ptr->start_line)  * 
                scan_man_ptr->samples_per_line + sample;

        /* the index is in memory, so return the address */
        return &scan_man_ptr->image_buffer_ptr[index];
    }
  
    /* not in memory, so return NULL to indicate line/sample not available */
    return NULL;
}

/******************************************************************************
FUNCTION NAME: More_Scans_To_Read

PURPOSE:        
More_Scans_To_Read checks to see if there are more lines to be read from
the input band.

RETURNS:
TRUE: More lines of data to read.
FALSE: NO more data to read.

NOTES:
This routine allows the details of when an image is done being read to be
isolated to the Scan Manager module.

******************************************************************************/

int More_Scans_To_Read
(
    const SCAN_MAN_TYPE *scan_man_ptr /* I: pointer to scan manager structure */
)
{
    if (scan_man_ptr->scan_direction == READ_SCAN_FORWARD)
    {
        if (!scan_man_ptr->data_valid || 
            ((scan_man_ptr->start_line + scan_man_ptr->lines_in_buffer) <
             scan_man_ptr->lines_in_image))
        {
            /* return that there are more scans to read */
            return TRUE;
        }
        else
        {
            return FALSE;
        }
    }
    else
    {
        if (!scan_man_ptr->data_valid || 
            (scan_man_ptr->start_line > 0))
        {
            /* return that there are more scans to read */
            return TRUE;
        }
        else
        {
            return FALSE;
        }
    }
}
