#ifndef _SCANMAN_H_
#define _SCANMAN_H_

#include "ias_geom_grid.h"    /* IAS_GEOM_GRID_TYPE definition */
#include "read_input_image.h" /* IN_IMAGE_TYPE definition */
#include "ias_l1g.h"          /* L1g I/O routines */

/*
**
** TYPEDEFS 
**
*/

/* enumerated type to define which direction the raw image scans 
   are being read in.  forward = top of image to bottom. */
typedef enum
{
   READ_SCAN_FORWARD,
   READ_SCAN_REVERSE
} SCAN_DIR_TYPE;

/* The SCAN_MAN_TYPE defines the data that is needed to track the 
   state of a scan manager.  Unless otherwise stated, all members 
   apply to the input image. */
typedef struct
{
    /* pointer to the memory allocated to hold the image lines as they 
       are read into memory */
    float *image_buffer_ptr;

    /* pointer to the grid band */
    const IAS_GEOM_GRID_BAND_TYPE *grid_band_ptr;

    /* pointer to input image state data */
    IN_IMAGE_TYPE *in_image_ptr;

    /* band index held by the scan manager */
    int band_index;

    /* number of lines in the input image */
    int lines_in_image;

    /* number of lines held in the image buffer */
    int lines_in_buffer;

    /* number of samples per line */
    int samples_per_line;

    int start_line;
    /* the starting line held in the image buffer. */

    /* the number of lines in the kernel.  This is used to verify 
       that enough lines exist in the scan buffer for resampling */
    int kernel_size;

    /* the last_valid_line is the last valid line currently in the 
       scan buffer that may be used for resampling.  This is a cached
       value of the last line in the buffer with the number of gap
       lines and the kernel size subtracted out (gap since the last
       scan isn't extended and kernel since the number of lines in
       the kernel will be read). */ 
    int last_valid_line;
    
    /* holds the direction image should be read in */
    SCAN_DIR_TYPE scan_direction;
 
    /* holds a non-zero value if data has been read into the image buffer. */
    int data_valid;
 
} SCAN_MAN_TYPE;

SCAN_MAN_TYPE *Initialize_Scan_Manager
(
    const IAS_GEOM_GRID_BAND_TYPE *grid_band_ptr, /* I: pointer to grid band */
    int  band_index,            /* I: band index to read */
    int  lines_to_buffer,       /* I: number of lines to buffer at a time */
    int  *samples_per_line_ptr, /* O: number of samples per line */
    int  *lines_in_image_ptr,   /* O: number of lines in the input image */
    int  kernel_size            /* I: number of lines in kernel */
);

void Cleanup_Scan_Manager
(
    SCAN_MAN_TYPE *scan_man_ptr /* I/O: pointer to scan manager to clean up */
);

int Read_Next_Scan_Set
(
    SCAN_MAN_TYPE *scan_man_ptr /* I: pointer to scan manager structure */
);

const float *Get_Sample_Data
(
    const SCAN_MAN_TYPE *scan_man_ptr, /* I: pointer to scan manager structure*/
    int  line,    /* I: line number of the upper left corner of the kernel */
    int  sample   /* I: sample number of the upper left corner of the kernel */
);

int More_Scans_To_Read
(
    const SCAN_MAN_TYPE *scan_man_ptr /* I: pointer to scan manager structure */
);

#endif 

