/******************************************************************************
NAME:           warp

PURPOSE:
The warp module is a driver for the resampling process.  It initializes the 
input scan manager, reads the SLO data, and creates the multiple processor
threads to perform the resampling.

NOTES:
none

ALGORITHM REFERENCES:
none

******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/param.h>     /* MIN/MAX definitions */
#include "ias_const.h"     /* SUCCESS/ERROR definition */
#include "ias_logging.h"
#include "warp.h"          /* prototypes for this module */
#include "kernel.h"        /* get_lines/samples_in_kernel prototypes */
#include "resample_image.h"/* resampling routines */
#include "ias_threadpool.h"/* multiprocessing details */
#include "debug.h"         /* ASSERT macro */


int warp
( 
    struct ias_threadpool *threadpool,  /* I: threadpool to use */
    const IAS_GEOM_GRID_TYPE *grid_ptr, /* I: pointer to grid */
    int band_index,                     /* I: current band index   */
    RESAMPLE_TYPE resampling_type,/* I: type of resampling */
    L1G_BAND_IO *output_l1g_band, /* I: Output L1G band pointer */
    int  lines_to_buffer,         /* I: number of input lines to buffer */
    int  procs_to_use,            /* I: number of processors to use */
    int  max_output_image_buffer, /* I: max memory in MB to use for buffering
                                        the output image */
    int  buffer_full_output_image /* I: flag to indicate entire output image
                                        should be kept in memory during 
                                        resampling */
)
{
    int rs;                             /* return status of calls */
    const IAS_GEOM_GRID_BAND_TYPE *grid_band_ptr = 
                                   &grid_ptr->gridbands[band_index]; 
                                        /* pointer to grid band */

    SCAN_MAN_TYPE *scan_man_ptr;        /* pointer to scan manager structure */

    RESAMPLE_THREAD_PARAMS params;      /* parameters to pass to resampling
                                           threads */

    int  in_nl;                         /* input image number of lines */
    int  in_ns;                         /* input image number of samples */
    int  active_lines;                  /* number of lines active in the output
                                           image */
    double effective_inlines;           /* effective number of input lines
                                           corresponding to the number of 
                                           output lines */

    /* verify the resampling type is one known to this module */
    ASSERT ((resampling_type == CC_RESAMPLING) ||
            (resampling_type == MTF_RESAMPLING) ||
            (resampling_type == NN_RESAMPLING) ||
            (resampling_type == BI_RESAMPLING));

    /* set the parameters to pass to the resampling threads */
    params.threadpool = threadpool;
    params.grid_band_ptr = grid_band_ptr;
    params.band_index = band_index;
    params.resample_type = resampling_type;
    params.thread_error = 0;

    /* prepare scan manager for use */
    scan_man_ptr = Initialize_Scan_Manager(grid_band_ptr, band_index, 
                    lines_to_buffer, &in_ns, &in_nl, get_lines_in_kernel());
    if (scan_man_ptr == NULL)
    {
        IAS_LOG_ERROR("Encountered initializing the scan manager");
        return ERROR;
    }

    params.scan_man_ptr = scan_man_ptr;
    params.in_nl = in_nl;
    params.in_ns = in_ns;

    /* determine which direction resampling should occur in - assume
       it is in the same direction the lines are being read */
    if (scan_man_ptr->scan_direction == READ_SCAN_FORWARD)
        params.resample_dir = 1;
    else
        params.resample_dir = -1;

    /* invert the scan direction if the image is rotated to more than 
       90 degrees */
    if (grid_band_ptr->in_lines[0] > 
        grid_band_ptr->in_lines[grid_band_ptr->num_out_samps * 
            (grid_band_ptr->num_out_lines - 1)])
    {
       params.resample_dir *= -1;
    }

    /* calculate the number of lines that will be active in the output image.
       Since the grid is a reverse grid, figure out how many input lines 
       are active, then scale it to the output image size.  Multiple by 1.42
       (sqrt of 2) to handle the worst case rotation of the image */
    active_lines = abs((int)(grid_band_ptr->in_lines[0] - 
        grid_band_ptr->in_lines[grid_band_ptr->num_out_samps - 1]));
    effective_inlines
            = MAX(MAX(grid_band_ptr->in_lines[0],
                    grid_band_ptr->in_lines[grid_band_ptr->num_out_samps-1]),
                  MAX(grid_band_ptr->in_lines[grid_band_ptr->num_out_samps
                                        * (grid_band_ptr->num_out_lines - 1)],
                      grid_band_ptr->in_lines[grid_band_ptr->num_out_samps 
                                        * grid_band_ptr->num_out_lines - 1]))
            - MIN(MIN(grid_band_ptr->in_lines[0],
                      grid_band_ptr->in_lines[grid_band_ptr->num_out_samps-1]),
                  MIN(grid_band_ptr->in_lines[grid_band_ptr->num_out_samps 
                                        * (grid_band_ptr->num_out_lines - 1)],
                      grid_band_ptr->in_lines[grid_band_ptr->num_out_samps
                                        * grid_band_ptr->num_out_lines - 1]));
    active_lines = active_lines * (double)grid_band_ptr->output_image_lines 
                 / effective_inlines * sqrt(2.0);

    /* create the output image buffer */
    if (procs_to_use < 1)
        procs_to_use = 1;

    params.output_image_buffer_ptr = initialize_output_image_buffer(
                        active_lines, band_index, output_l1g_band,
                        procs_to_use, params.resample_dir,
                        max_output_image_buffer, buffer_full_output_image);
    if (params.output_image_buffer_ptr == NULL)
    {
        IAS_LOG_ERROR("Unable to create output image buffer");
        Cleanup_Scan_Manager(scan_man_ptr);
        return ERROR;
    }
   
    /* initialize the last line completed indices to zero so the resampling
       starts at the edge of the image */
    params.last_line_complete_index = 0;
    params.returned_last_line_complete_index = 0;

    while (More_Scans_To_Read(scan_man_ptr))
    {
        if (Read_Next_Scan_Set(scan_man_ptr) == ERROR)
        {
            IAS_LOG_ERROR("Reading set of input image lines");
            cleanup_output_image_buffer(params.output_image_buffer_ptr, 0);
            Cleanup_Scan_Manager(scan_man_ptr);
            return ERROR; 
        }

        reset_resample_image();

        switch (resampling_type)
        {
            case CC_RESAMPLING:
            case MTF_RESAMPLING:
                /* run the resampling routine using the threadpool */
                rs = ias_threadpool_run_function(threadpool, resample_image_cc,
                            (void *)&params);
                break;
            
            case NN_RESAMPLING:
                rs = ias_threadpool_run_function(threadpool, resample_image_nn,
                            (void *)&params);
                break;

            case BI_RESAMPLING:
                rs = ias_threadpool_run_function(threadpool, resample_image_bi,
                            (void *)&params);
                break;

            default:
                rs = 1; /* simulate error */
                break;
        }
        if ((rs != 0) || (params.thread_error != 0))
        {
            if (rs != 0)
                IAS_LOG_ERROR("Forking processes");
            else
            {
                IAS_LOG_ERROR("Encountered in resampling thread %d\n",
                        params.thread_error - 1);
            }    

            cleanup_output_image_buffer(params.output_image_buffer_ptr, 0);
            Cleanup_Scan_Manager(scan_man_ptr);
            return ERROR;
        }
    
        /* update the last line complete index */
        params.last_line_complete_index = 
                params.returned_last_line_complete_index;
    }

    /* close the output image buffer */
    if (cleanup_output_image_buffer(params.output_image_buffer_ptr, 1) == ERROR)
    {
        IAS_LOG_ERROR("Freeing the output image buffer");
        Cleanup_Scan_Manager(scan_man_ptr);
        return ERROR;
    }

    /* close the scan manager */ 
    Cleanup_Scan_Manager(scan_man_ptr); 

    return SUCCESS;
}
