/******************************************************************************
Name:                    compare_metadata

Purpose:
Read the file and band metadata to verify that it is the
same for the reference and search images for all bands specified.  Pointers
to reference and search file/band metadata structures for both the reference
and search images are returned on successful completion of this routine.

Returns:  Integer status code
          SUCCESS:  Successful completion
          ERROR:    Operation failed

Notes:
This routine will check the projection information, pixel size, and correction
types.  I2I characterization requires that this information be the same for
all the bands that are being characterized.  If any of the above are not
present, then the application will exit with a failure.

Another check is to verify that the images are terrain-corrected images.  If
the search and reference images are not terrain-corrected, then a warning
message will be printed to the log file.  A warning message will also be
printed if the correction_types are not the same for both images.

A band-referenced array of metadata is returned for both the reference and
search images.  This is similar to the manner of storing the different band 
elements of the grid structure.  These arrays should be of size MAX_BANDS.

Algorithm References:  See ias_l1g.h for a description of the file and
                       band metadata structures.

******************************************************************************/
#include <string.h>

#include "i2i_char.h"
#include "ias_logging.h"                /* Logging library                   */
#include "ias_satellite_attributes.h"   /* Satellite attributes library      */


int compare_metadata 
(
    const char *ref_l1g_name,  /* I: Reference L1G image filename            */
    const char *srch_l1g_name, /* I: Search L1G image filename               */
    const int *reference_band_indexes, /* I: Array of reference band indexes */
    const int *search_band_indexes, /* I: Array of search band indexes       */
    int number_bands,          /* I: Number of bands in the band indexes 
                                     array                                   */
    IAS_L1G_FILE_METADATA *ref_file_metadata, /* O: Reference image file 
                                     metadata                                */
    IAS_L1G_FILE_METADATA *srch_file_metadata, /* O: Search image file 
                                     metadata                                */
    IAS_L1G_BAND_METADATA *ref_band_metadata, /* O: Reference image band 
                                     metadata array                          */
    IAS_L1G_BAND_METADATA *srch_band_metadata, /* O: Search image band 
                                     metadata array                          */
    int *ref_number_lines,     /* O: Number of reference lines               */
    int *ref_number_samples,   /* O: Number of reference samples             */
    int *srch_number_lines,    /* O: Number of search lines                  */
    int *srch_number_samples   /* O: Number of search samples                */
)
{
    int i;                     /* Loop counter                               */
    int reference_band_number; /* Reference band number                      */
    int reference_band_index;  /* Corresponding reference position for
                                  the current band                           */
    int search_band_number;    /* Search band number                         */
    int search_band_index;     /* Corresponding search position for the
                                  the current band                           */
    int status;                /* Error status                               */

    /* Read the reference image metadata. */
    status = read_metadata(ref_l1g_name, reference_band_indexes, number_bands, 
        ref_number_lines, ref_number_samples, ref_file_metadata, 
        ref_band_metadata);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Reading reference image %s metadata", ref_l1g_name);
        return ERROR;
    }

    /* Read the search image metadata. */
    status = read_metadata(srch_l1g_name, search_band_indexes, number_bands, 
        srch_number_lines, srch_number_samples, srch_file_metadata, 
        srch_band_metadata);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Reading search image %s metadata", srch_l1g_name);
        return ERROR;
    }

    /* Check the reference image against the search image to verify the
       critical metadata information is the same. */

    /* Start with the file metadata information. */
    if (ref_file_metadata->projection_code != 
        srch_file_metadata->projection_code)
    {
        IAS_LOG_ERROR("Search image projection code %d does not match "
            "reference image projection code %d", 
            srch_file_metadata->projection_code, 
            ref_file_metadata->projection_code);
        return ERROR;
    }

    if (ref_file_metadata->zone_code != srch_file_metadata->zone_code)
    {
        IAS_LOG_ERROR("Search image zone code %d does not match reference "
            "image zone code %d", srch_file_metadata->zone_code, 
            ref_file_metadata->zone_code); 
        return ERROR;
    }

    /* Remove "_" in the datum string if it has for both search and reference 
       images */
    char *srch_ptr = NULL;
    while ((srch_ptr = strchr(srch_file_metadata->datum, '_')) != NULL)
        strcpy (srch_ptr, srch_ptr + 1);
    while ((srch_ptr = strchr(ref_file_metadata->datum, '_')) != NULL)
        strcpy (srch_ptr, srch_ptr + 1);
    if (strcasecmp(ref_file_metadata->datum, srch_file_metadata->datum))
    {
        IAS_LOG_ERROR("Search image datum %s does not match reference image "
            "datum %s", srch_file_metadata->datum, ref_file_metadata->datum); 
        return ERROR;
    }

    if (ref_file_metadata->spheroid_code != srch_file_metadata->spheroid_code)
    {
        IAS_LOG_ERROR("Search image spheroid code %d does not match reference "
            "image spheroid code %d", srch_file_metadata->spheroid_code, 
            ref_file_metadata->spheroid_code);
        return ERROR;
    }

    if (strcasecmp(ref_file_metadata->projection_units, 
                   srch_file_metadata->projection_units))
    {
        IAS_LOG_ERROR("Search image projection units %s do not match "
            "reference image projection units %s", 
            srch_file_metadata->projection_units, 
            ref_file_metadata->projection_units); 
        return ERROR;
    }

    if (strcasecmp(ref_file_metadata->correction_type, 
                   srch_file_metadata->correction_type))
    {
        IAS_LOG_WARNING("Search image correction type %s does not match "
            "reference image correction type %s", 
            srch_file_metadata->correction_type, 
            ref_file_metadata->correction_type); 
    }

    if (strcasecmp(ref_file_metadata->resample_type, 
                   srch_file_metadata->resample_type))
    {
        IAS_LOG_WARNING("Search image resampling type %s does not match "
            "reference image resampling type %s", 
            srch_file_metadata->resample_type, 
            ref_file_metadata->resample_type); 
    }

    if (ref_file_metadata->wrs_path != srch_file_metadata->wrs_path)
    {
        IAS_LOG_WARNING("Search image path %d does not match reference image "
            "path %d", srch_file_metadata->wrs_path, 
            ref_file_metadata->wrs_path);
    }

    if (ref_file_metadata->wrs_row != srch_file_metadata->wrs_row)
    {
        IAS_LOG_WARNING("Search image row %d does not match reference image "
            "row %d", srch_file_metadata->wrs_row, ref_file_metadata->wrs_row);
    }

    for (i = 0; i < IAS_PROJ_PARAM_SIZE; i++)
    {
        if (ref_file_metadata->projection_parameters[i] !=
            srch_file_metadata->projection_parameters[i])
        {
            IAS_LOG_ERROR("Search image projection parameter %d (value %lf) "
                "does not match reference image projection parameter %d "
                "(value %lf)", 
                i + 1, srch_file_metadata->projection_parameters[i], 
                i + 1, ref_file_metadata->projection_parameters[i]);
            return ERROR;
        }
    }

    /* Check that the search and reference images are nadir-viewing.  Produce 
       a warning, not an error, otherwise. */
    if ((srch_file_metadata->roll_angle > IAS_OFF_NADIR_BOUNDARY) ||
        (srch_file_metadata->roll_angle < -IAS_OFF_NADIR_BOUNDARY))
    {
        IAS_LOG_WARNING("Search image is not nadir-viewing; roll angle is %f",
            srch_file_metadata->roll_angle);
    }

    if ((ref_file_metadata->roll_angle > IAS_OFF_NADIR_BOUNDARY) ||
        (ref_file_metadata->roll_angle < -IAS_OFF_NADIR_BOUNDARY))
    {
        IAS_LOG_WARNING("Reference image is not nadir-viewing; roll angle is "
            "%f", ref_file_metadata->roll_angle);
    }

    /* Check that the search and reference images are L1Ts, and issue warnings
       if not. */
    if (strcasecmp(srch_file_metadata->correction_type, "L1T") != 0)
    {
        IAS_LOG_WARNING("Search image %s is not terrain-corrected",
            srch_l1g_name);
    }

    if (strcasecmp(ref_file_metadata->correction_type, "L1T") != 0)
    {
        IAS_LOG_WARNING("Reference image %s is not terrain-corrected",
            ref_l1g_name);
    }

    /* Now compare the band-specific metadata information. */
    for (i = 0; i < number_bands; i++)
    {
        /* Find reference positions. */
        reference_band_index = reference_band_indexes[i];
        reference_band_number = ias_sat_attr_convert_band_index_to_number(
            reference_band_index);
        if (reference_band_number == ERROR) 
        {
            IAS_LOG_ERROR("Invalid band number for reference band index %d", 
                reference_band_index);
            return ERROR;
        }

        /* Find search positions. */
        search_band_index = search_band_indexes[i];
        search_band_number = ias_sat_attr_convert_band_index_to_number(
            search_band_index);
        if (search_band_number == ERROR) 
        {
            IAS_LOG_ERROR("Invalid band number for search band index %d", 
                search_band_index); 
            return ERROR;
        }

        /* Check the pixel sizes. */
        if (ref_band_metadata[reference_band_index].projection_distance_y !=
            srch_band_metadata[search_band_index].projection_distance_y)
        {
            IAS_LOG_ERROR("Search image Y pixel size %lf does not match "
                "reference image Y pixel size %lf for search band %d, "
                "reference band %d", 
                srch_band_metadata[search_band_index].projection_distance_y, 
                ref_band_metadata[reference_band_index].projection_distance_y, 
                search_band_number, reference_band_number); 
            return ERROR;
        }

        if (ref_band_metadata[reference_band_index].projection_distance_x !=
            srch_band_metadata[search_band_index].projection_distance_x)
        {
            IAS_LOG_ERROR("Search image X pixel size %lf does not match "
                "reference image X pixel size %lf for search band %d, "
                "reference band %d", 
                srch_band_metadata[search_band_index].projection_distance_x, 
                ref_band_metadata[reference_band_index].projection_distance_x, 
                search_band_number, reference_band_number); 
            return ERROR;
        }
    }

    return SUCCESS;
}
