/******************************************************************************
NAME:           create_tiepts

PURPOSE:
Create tiepoints for all band resolutions for both the reference and search
images.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
This routine will malloc the appropriate space needed for the IAS_GCP_RECORD
pointers, ref_gcp15, srch_gcp15, ref_gcp30, and srch_gcp30.

******************************************************************************/
#include <stdlib.h>
#include "i2i_char.h"
#include "ias_logging.h"        /* Logging library                           */

int create_tiepts 
(
    enum TIEPT_TYPE tiept_type, /* I: Variable to specify the type of 
                                      tiepts                                 */
    const int *reference_band_indexes, /* I: Array of reference band indexes */
    const int *search_band_indexes, /* I: Array of search band indexes       */
    int number_bands,           /* I: Number of band indexes                 */
    const IAS_L1G_BAND_METADATA *ref_band_metadata, /* I: Reference image 
                                      metadata for the desired band          */
    const IAS_L1G_BAND_METADATA *srch_band_metadata, /* I: Search image 
                                      metadata for the desired band          */
    const int *ref_number_lines, /* I: Number of lines in reference image    */
    const int *ref_number_samples,/* I: Number of samples in reference image */
    const int *srch_number_lines, /* I: Number of lines in search image      */
    const int *srch_number_samples, /* I: Number of samples in search image  */
    int number_line_points,     /* I: Number of line tiepoints to create     */
    int number_sample_points,   /* I: Number of sample tiepoints to create   */
    const struct IAS_LNG_LS chip_size,/* I: Chip size of the tiepoints to 
                                      create                                 */
    IAS_DBL_XY *fract_offsetPTR_15, /* O: Pointer of image offset structure 
                                      for 15m band                           */
    IAS_DBL_XY *fract_offsetPTR_30, /* O: Pointer of image offset structure 
                                      for 30m band                           */
    IAS_GCP_RECORD **ref_gcp15, /* O: Array of ref GCP record tiepoints in 
                                      row-major form for 15m band 
                                      resolutions                            */
    IAS_GCP_RECORD **srch_gcp15,/* O: Array of srch GCP record tiepoints in 
                                      row-major form for 15m band 
                                      resolutions                            */
    IAS_GCP_RECORD **ref_gcp30, /* O: Array of ref GCP record tiepoints in 
                                      row-major form for 30m band 
                                      resolutions                            */
    IAS_GCP_RECORD **srch_gcp30 /* O: Array of srch GCP record tiepoints in 
                                      row-major form for 30m band 
                                      resolutions                            */
)
{
    int i;                      /* Loop variables                            */
    int GCP_record_size;        /* Total size of GCP record arrays           */
    int number_of_gcps;         /* Total number of GCP Library GCPs          */
    int reference_band_index;   /* Reference position for the arrays         */
    int search_band_index;      /* Search position for the arrays            */
    double projection_distance; /* Variable to make condition easier to read */
    int status;                 /* Return status                             */

    /* Initialize the GCP pointers to NULL, then they can be used as a flag
       to determine if the GCP15 or GCP30 structures have been created. */
    *ref_gcp15 = NULL;
    *srch_gcp15 = NULL;
    *ref_gcp30 = NULL;
    *srch_gcp30 = NULL;

    /* The total number of GCPs is number_line_points * number_sample_points. */
    number_of_gcps = number_line_points * number_sample_points;
    GCP_record_size = number_of_gcps * sizeof(IAS_GCP_RECORD);

    /* Loop through the bands checking the different band resolutions.  At
       most, there will only be one GCP15 and GCP30 tiepoint set created for
       both the reference and search images.  Anytime the ref GCP is created,
       the search GCP is also created. */
    for (i = 0; i < number_bands; i++)
    {
        /* Get the search and reference positions. */
        reference_band_index = reference_band_indexes[i];
        search_band_index = search_band_indexes[i];

        /* If GCP15 hasn't already been done and there is a 15m band, we need
           a GCP15 set of tiepoints. */
        projection_distance = 
            ref_band_metadata[reference_band_index].projection_distance_x;
        if (IS_PROJ_DIST(projection_distance, 15.0) && (*ref_gcp15 == NULL))
        {
            /* Determine the GCP15 regular-spaced tiepoints. */
            *ref_gcp15 = malloc(GCP_record_size);
            if (*ref_gcp15 == NULL)
            {
                IAS_LOG_ERROR("Allocating reference GCP15 buffer");
                free(*ref_gcp30);
                free(*srch_gcp30);
                return ERROR;
            }
            *srch_gcp15 = malloc(GCP_record_size);
            if (*srch_gcp15 == NULL)
            {
                IAS_LOG_ERROR("Allocating search GCP15 buffer");
                free(*ref_gcp15);
                free(*ref_gcp30);
                free(*srch_gcp30);
                return ERROR;
            }

            status = i2i_det_tiepts(&ref_band_metadata[reference_band_index], 
                &srch_band_metadata[search_band_index],
                ref_number_lines[i], ref_number_samples[i],
                srch_number_lines[i], srch_number_samples[i],
                number_line_points, number_sample_points, 
                chip_size, *ref_gcp15, *srch_gcp15, fract_offsetPTR_15);
            if (status == ERROR)
            {
                IAS_LOG_ERROR("Retrieving GCP15 tiepoints");
                free(*ref_gcp15);
                free(*srch_gcp15);
                free(*ref_gcp30);
                free(*srch_gcp30);
                return ERROR;
            }
        }

        /* If GCP30 hasn't already been done and there is a 30m band, we need 
           a GCP30 set of tiepoints. */
        else if (IS_PROJ_DIST(projection_distance, 30.0) 
            && (*ref_gcp30 == NULL))
        {
            /* Determine the GCP30 regular-spaced tiepoints. */
            *ref_gcp30 = malloc(GCP_record_size);
            if (*ref_gcp30 == NULL)
            {
                IAS_LOG_ERROR("Allocating reference GCP30 buffer");
                free(*ref_gcp15);
                free(*srch_gcp15);
                return ERROR;
            }
            *srch_gcp30 = malloc(GCP_record_size);
            if (*srch_gcp30 == NULL)
            {
                IAS_LOG_ERROR("Allocating search GCP30 buffer");
                free(*ref_gcp15);
                free(*srch_gcp15);
                free(*ref_gcp30);
                return ERROR;
            }

            status = i2i_det_tiepts(&ref_band_metadata[reference_band_index], 
                &srch_band_metadata[search_band_index],
                ref_number_lines[i], ref_number_samples[i],
                srch_number_lines[i], srch_number_samples[i],
                number_line_points, number_sample_points, 
                chip_size, *ref_gcp30, *srch_gcp30, fract_offsetPTR_30);
            if (status == ERROR)
            {
                IAS_LOG_ERROR("Retrieving GCP30 tiepoints");
                free(*ref_gcp15);
                free(*srch_gcp15);
                free(*ref_gcp30);
                free(*srch_gcp30);
                return ERROR;
            }
        }
    }  /* for i */

    return SUCCESS;
}
