/******************************************************************************
NAME:          get_gcp_line_sample 

PURPOSE:
Determine the line/sample of the GCPs for a specified band resolution.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:

******************************************************************************/
#include <string.h>

#include "i2i_char.h"
#include "ias_geo.h"                  /* Geographic library                  */
#include "ias_miscellaneous.h"        /* Miscellaneous IAS libraries         */
#include "ias_logging.h"              /* Logging library                     */

int get_gcp_line_sample
(
    const IAS_L1G_BAND_METADATA *band_metadata, /* I: Image band metadata    */
    const IAS_GEO_PROJ_TRANSFORMATION *degrees2map, /* I: projection
                                            transformation to use */
    const int number_lines,           /* I: Number of lines in image         */
    const int number_samples,         /* I: Number of samples in image       */
    const int number_of_gcps,         /* I: Number of GCPs in the GCP 
                                            library                          */
    const IAS_GCP_RECORD *in_gcps,    /* I: Input array of GCP record 
                                            tiepoints in row-major format    */
    IAS_GCP_RECORD *gcps              /* O: Output array of GCP record 
                                            tiepoints in row-major format    */
)
{
    int i;                /* Looping variable                                */
    double poly_x[4];     /* Sample-to-projection coefficients for north-up  */
    double poly_y[4];     /* Line-to-projection coefficients for north-up    */
    int degree;           /* Degree of the polynomials                       */
    int stat;             /* Function return status                          */
    int year;             /* year extracted from date in GCP record          */
    int month;            /* month extracted from date in GCP record         */
    int day;              /* day extracted from date in GCP record           */
    int status;           /* the status of the return from a funtion         */
    double corners[2][4]; /* latitude and longitude for the four corners     */


    /* Check metadata input. */
    if (band_metadata == NULL)
    {
        IAS_LOG_ERROR("Band metadata is NULL");
        return ERROR;
    }
    
    /* Set up the corners structure needed for the compute_poly call. */
    corners[1][0] = band_metadata->upper_left_x;
    corners[1][1] = band_metadata->upper_right_x;
    corners[1][2] = band_metadata->lower_right_x;
    corners[1][3] = band_metadata->lower_left_x;
    corners[0][0] = band_metadata->upper_left_y;
    corners[0][1] = band_metadata->upper_right_y;
    corners[0][2] = band_metadata->lower_right_y;
    corners[0][3] = band_metadata->lower_left_y;

    /* Get the polynomials for the eval_poly call. */
    degree = 1;
    stat = ias_geo_compute_proj2ls_poly(number_samples, number_lines, corners, 
        poly_x, poly_y);
    if (stat == ERROR)
    {
        IAS_LOG_ERROR("Calculating the poly_x and poly_y values");
        return ERROR;
    }

    /* Loop through the GCPs and calculate the line/sample values. */
    for (i = 0; i < number_of_gcps; i++)
    {
        /* Copy the data from the input GCP that we want to keep. */
        strcpy (gcps[i].point_id, in_gcps[i].point_id);
        strcpy (gcps[i].chip_name, in_gcps[i].chip_name);
        gcps[i].latitude = in_gcps[i].latitude;
        gcps[i].longitude = in_gcps[i].longitude;
        gcps[i].elevation = in_gcps[i].elevation;
        gcps[i].chip_size_lines = in_gcps[i].chip_size_lines;
        gcps[i].chip_size_samples = in_gcps[i].chip_size_samples;
        strcpy (gcps[i].source, in_gcps[i].source);
        strcpy (gcps[i].date, in_gcps[i].date);

        /* Check the date of the GCPLib */
        status = sscanf(gcps[i].date, IAS_GCP_DATE_FORMAT, &month, &day, &year);
        if (status != 3)
        {
            IAS_LOG_ERROR("In date %s for GCP point %s", gcps[i].date,
                gcps[i].point_id);
            return ERROR;
        }
        status = ias_misc_check_year_month_day(year, month, day);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("In date %s for GCP point %s", gcps[i].date,
                gcps[i].point_id);
            return ERROR;
        }

        /* Fill in pixel size. */ 
        gcps[i].pixel_size_x = band_metadata->projection_distance_x;
        gcps[i].pixel_size_y = band_metadata->projection_distance_y;

        /* Calculate the proj x/y values from the GCPLib lat/long using the 
           projection information. */
        stat = ias_geo_transform_coordinate(degrees2map, gcps[i].longitude, 
               gcps[i].latitude, &gcps[i].projection_x, &gcps[i].projection_y);
        if (stat == ERROR)
        {
            IAS_LOG_ERROR("In the transform projection call for GCP point %s",
                gcps[i].point_id);
            return ERROR;
        }

        /* Calculate output line and sample.  This routine will use polynomial
           calculations with both the path-oriented and north-up products. */
        ias_math_eval_poly(degree, poly_y, gcps[i].projection_x, 
            gcps[i].projection_y, &gcps[i].reference_line);
        ias_math_eval_poly(degree, poly_x, gcps[i].projection_x, 
            gcps[i].projection_y, &gcps[i].reference_sample);
    }

    return SUCCESS;
}
