/******************************************************************************
NAME:           get_parms

PURPOSE:
Get the i2i characterization parameters from the input parameter file.


RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
This routine opens the input parameter files, reads the I2I characterization
parameters, and returns the parameters.  It also reads the CPF and retrieves 
the min_corr, max_disp, chip_size, range, threshold, and fit method from it.
Min_corr, max_disp, chip_size, range, and threshold CPF values are overridden
by parameter file values if they are present there.

******************************************************************************/
#include <stdlib.h>
#include <string.h>

#include "i2i_char.h"
#include "ias_parm.h"                     /* parameter library               */
#include "ias_cpf.h"                      /* CPF library                     */
#include "ias_math.h"                     /* Math library                    */
#include "ias_parm_standard_parameters.h" /* parameter table standard fields */

#define INVALID_PARAMETER -1              /* Invalid parameter value         */

int set_band_indexes
(
    const int band_number_list[IAS_MAX_NBANDS], /* I: Band number list source 
                                                for band indexes             */
    int check_duplicates,                 /* I: Flag for checking for dupes  */ 
    int maximum_number_bands,             /* I: Maximum allowed # of bands   */
    int band_indexes[IAS_MAX_NBANDS],     /* O: Band index list              */
    int *number_bands                     /* O: # bands in band index list   */
);

int get_parms 
(
    const char *param_filename,   /* I: Parameter file name                  */
    char *ref_L0R,                /* O: L0R HDF filename                     */
    char *ref_1G,                 /* O: OLI file name for the 1G ref image   */
    char *srch_1G,                /* O: OLI file name for the 1G srch image  */
    char *gcplib_file_name,       /* O: Name of the tiepoint file            */
    char *cpf_file_name,          /* O: CPF file name                        */
    char *output_file_name,       /* O: Name of the output file              */
    int reference_band_indexes[IAS_MAX_NBANDS], /* O: Reference band indexes */
    int search_band_indexes[IAS_MAX_NBANDS], /* O: Search band indexes       */
    int *number_bands,            /* O: Number of bands in the lists         */
    int *number_line_points,      /* O: Number of line tiepoints             */
    int *number_sample_points,    /* O: Number of sample tiepoints           */
    struct IAS_LNG_LS *chip_size, /* O: Height and width of the chips        */
    enum TIEPT_TYPE *tiept_type,  /* O: Type of tiepts to use                */
    float *min_corr,              /* O: Minimum correlation strength         */
    float *max_disp,              /* O: Maximum correlation displacement     */
    float *range,                 /* O: Range of valid pixels                */
    float *threshold,             /* O: Percentage of valid pixels           */
    IAS_CORRELATION_FIT_TYPE *fit_method, /* O: Correlation result fit 
                                        method                               */
    int *abs_corr_coeff,          /* O: Flag to use absolute value of the
                                        grey correlation coeff               */
    char *wo_id,                  /* O: Work order ID string                 */
    char *omf_file_name           /* O: omf file name                        */
)
{
    int i;                        /* Loop variable                           */
    int count;                    /* Number of values returned               */
    int tiept;                    /* Type of the tiepoint                    */
    char *IMG_ptr;                /* Pointer to the DOQ image directory      */
    int reference_band_number_list[IAS_MAX_NBANDS]; /* Reference band list 
                                     read from file                          */
    int search_band_number_list[IAS_MAX_NBANDS]; /* Search band list read 
                                     from file                               */
    int wrs_path;                 /* WRS path                                */
    int wrs_row;                  /* WRS row                                 */
    char l1g_image[PATH_MAX];     /* L1G image filename                      */
    int satellite_number_bands;   /* Number normal bands for the satellite   */
    int reference_number_bands;   /* Save a copy of the number of ref bands  */
    IAS_CPF *cpf;                 /* Calibration parameter file constants    */
    const struct IAS_CPF_I2I_ASSESSMENT *cpf_i2i;
    int status;                   /* Status returned from functions          */
    int processing_pass;          /* The current processing pass             */
    int sat_num;                  /* The satellite number                    */
    IAS_SATELLITE_ID sat_id;      /* The satellite ID                        */

    /* Define various file name storage variables.  Local storage is used 
       because sizeof doesn't work on function parameter arrays. */
    char local_ref_L0R[PATH_MAX];
    char local_ref_1G[PATH_MAX];
    char local_srch_1G[PATH_MAX];
    char local_gcp_directory[PATH_MAX];
    char local_cpf_file_name[PATH_MAX];
    char local_output_file_name[PATH_MAX];
    char local_l1g_image[PATH_MAX];
    char local_wo_id[IAS_WORKORDER_ID_SIZE];

    /* Local buffer is used to translate between IAS_LNG_LS* and parameter I/O
       interface. */
    int local_chip_size[2];

    /* Local buffer is used to translate between float* and parameter I/O
       interface. */
    double local_range[2];
    double local_min_corr;
    double local_max_disp;
    double local_threshold;

    /* Initialize WRS path and row so we know if they weren't selected. */
    wrs_path = INVALID_PARAMETER;
    wrs_row = INVALID_PARAMETER;

    /* Initialize band number lists. */
    for (i = 0; i < IAS_MAX_NBANDS; i++)
    {
        reference_band_number_list[i] = 0;
        search_band_number_list[i] = 0;
    }

    /* Define the parameter table. */
    IAS_PARM_DECLARE_TABLE(parms, 15);

    /* Get the work order ID. */
    IAS_PARM_WORK_ORDER_ID(parms, local_wo_id, sizeof(local_wo_id), 1);

    /* Get the list of reference bands.  Don't use standard band list since I2I
        has special default band list requirements (no cirrus and no TIRS). */
    int default_reference_band_number_list[] = {1,2,3,4,5,6,7,8};
    IAS_PARM_ADD_INT(parms, REFERENCE_BAND_LIST, 
            "Defines reference bands to process", IAS_PARM_OPTIONAL,
            IAS_PARM_ARRAY, 1, 1, 11, 8, default_reference_band_number_list, 
            reference_band_number_list, sizeof(reference_band_number_list), 1); 

    /* Get the list of search bands.  Don't use standard band list since I2I
        has special default band list requirements (no cirrus and no TIRS). */
    int default_search_band_number_list[] = {1,2,3,4,5,6,7,8};
    IAS_PARM_ADD_INT(parms, SEARCH_BAND_LIST, "Defines search bands to process",
            IAS_PARM_OPTIONAL, IAS_PARM_ARRAY, 1, 1, 11, 8, 
            default_search_band_number_list, search_band_number_list, 
            sizeof(search_band_number_list), 1); 
 
    /* Get the L1G reference image file name. */
    strcpy (local_ref_1G, "");
    IAS_PARM_ADD_STRING(parms, L1G_REF_IMAGE, 
            "Reference image that supersedes the default Path/Row DOQ", 
            IAS_PARM_OPTIONAL, 0, NULL, 0, NULL, local_ref_1G, 
            sizeof(local_ref_1G), 0);

    /* Get the L1G search image file name. */
    strcpy (local_srch_1G, "");
    IAS_PARM_ADD_STRING(parms, L1G_SRCH_IMAGE, 
            "Search image that supersedes the default L1G image", 
            IAS_PARM_OPTIONAL, 0, NULL, 0, NULL, local_srch_1G, 
            sizeof(local_srch_1G), 0);

    /* Get the tiept_type. */
    int default_tiept_type[] = {1};
    IAS_PARM_TIEPT_TYPE(parms, default_tiept_type, &tiept, sizeof(tiept), 1);

    /* Get the number_line_points and number_sample_points.  These are only 
       needed if the tiept_type is REGULAR_SPACED. */
    int default_number_line_points[] = {25};
    IAS_PARM_NLINE_PTS(parms, 1, default_number_line_points,
            number_line_points, sizeof(*number_line_points));

    int default_number_sample_points[] = {25};
    IAS_PARM_NSAMP_PTS(parms, 1, default_number_sample_points,
            number_sample_points, sizeof(*number_sample_points));

    /* Get the minimum correlation value. */
    IAS_PARM_MIN_CORR(parms, &local_min_corr, sizeof(local_min_corr));

    /* Get the maximum displacement value. */
    IAS_PARM_MAX_DISP(parms, &local_max_disp, sizeof(local_max_disp));

    /* Get the valid pixel percentage threshold value. */
    IAS_PARM_THRESHOLD(parms, &local_threshold, sizeof(local_threshold));

    /* Get the chip size values. */
    IAS_PARM_CHIP_SIZE(parms, local_chip_size, sizeof(local_chip_size));

    /* Get the range values. */
    IAS_PARM_RANGE(parms, local_range, sizeof(local_range));

    /* Get the output_file_name. */
    const char *default_output_file_name[] = {"i2i_char.dat"};
    IAS_PARM_ADD_STRING(parms, I2I_CHAR_OUTPUT_FILENAME,
            "Output data file name", IAS_PARM_OPTIONAL,
            0, NULL, 1, default_output_file_name, 
            local_output_file_name, sizeof(local_output_file_name), 1); 

    IAS_PARM_PROCESSING_PASS(parms, &processing_pass, sizeof(processing_pass),
            0);

    /* Calculate the number of items in the parameter file. */
    count = IAS_PARM_GET_TABLE_SIZE(parms);

    /* Call the ias_parm_provide_help routine to provide help for the
       parameters if requested.  This also verifies the ODL file exists. */
    status = ias_parm_provide_help(param_filename, parms, count,
            IAS_INPUT_PARAMETERS);
    if (status == 1)
    {
        /* If the status of the provide help is 1 the user got help.  Since
           the parameter file name contains the help request, copy it to the
           OMF file name so the next call will ask for the same help. */
        strcpy(omf_file_name, param_filename);
    }
    else if (status == ERROR) /* Help failed. */
    {
        IAS_LOG_ERROR("Providing help using application parameter file %s",
                param_filename);
        exit(EXIT_FAILURE);
    }
    else
    {
        /* Read the parameter file. */
        if (ias_parm_read(param_filename, parms, count) != SUCCESS)
        {
            IAS_LOG_ERROR("reading parameters from: %s", param_filename);
            return ERROR;
        }

        /* Transfer the various local arrays to the function parameters. */
        strcpy(ref_1G, local_ref_1G);
        strcpy(srch_1G, local_srch_1G);
        strcpy(output_file_name, local_output_file_name);
        strcpy(wo_id, local_wo_id);
        chip_size->line = local_chip_size[0];
        chip_size->samp = local_chip_size[1];
        range[0] = local_range[0];
        range[1] = local_range[1];
        *min_corr = (float)local_min_corr;
        *max_disp = (float)local_max_disp;
        *threshold = (float)local_threshold;

        /* Build the OMF file name. */
        sprintf(omf_file_name, "%s.omf", wo_id);
    }

    /* Now read and check the OMF. */

    /* Define the parameter table. */
    IAS_PARM_DECLARE_TABLE(omf_parms, 7);

    IAS_PARM_CPF(omf_parms, IAS_PARM_REQUIRED, local_cpf_file_name, 
            sizeof(local_cpf_file_name), 1);

    IAS_PARM_L0R_FILENAME(omf_parms, IAS_PARM_OPTIONAL, local_ref_L0R,
            sizeof(local_ref_L0R), 1);

    /* Get the L1G image, which is only needed if the L1G search image isn't
       specified. */ 
    IAS_PARM_L1G_FILENAME_FROM_PASS(omf_parms, IAS_PARM_OPTIONAL, 
            local_l1g_image, sizeof(local_l1g_image), 0, processing_pass);

    /* Get the GCP directory.  It's only used if the user entered 
       tiepoint_type of READ_FILE. */
    /* GCP directory */
    IAS_PARM_GCP_DIRECTORY(omf_parms, IAS_PARM_OPTIONAL, local_gcp_directory, 
            sizeof(local_gcp_directory));

    /* Get the WRS path (only needed if ref_1G is not specified) */
    IAS_PARM_TARGET_WRS_PATH(omf_parms, IAS_PARM_OPTIONAL, &wrs_path,
            sizeof(wrs_path), 0);

    /* Get the WRS row (only needed if ref_1G is not specified). */
    IAS_PARM_TARGET_WRS_ROW(omf_parms, IAS_PARM_OPTIONAL, &wrs_row,
            sizeof(wrs_row), 0);

    /* Get the satellite number */
    IAS_PARM_SATELLITE(omf_parms, &sat_num, sizeof(sat_num), 1);

    /* See if the user wanted help or a template with the OMF file. */
    status = ias_parm_provide_help(omf_file_name, omf_parms,
            IAS_PARM_GET_TABLE_SIZE(omf_parms), IAS_OMF_PARAMETERS);
    if (status)
    {
        if (status == 1) /* Help was successful. */
        {
            exit(EXIT_SUCCESS);
        }
        else if (status == ERROR) /* Help failed. */
        {
            IAS_LOG_ERROR("Providing help using OMF %s", omf_file_name);
            exit(EXIT_FAILURE);
        }
    }

    /* Read the OMF file. */
    count = IAS_PARM_GET_TABLE_SIZE(omf_parms);
    if (ias_parm_read(omf_file_name, omf_parms, count) != SUCCESS)
    {
        IAS_LOG_ERROR("reading parameters from OMF file: %s", omf_file_name);
        return ERROR;
    }

    /*==========================================================================
     * Validate the ODL parameters
     *========================================================================*/

    /* Exit if the output filename is empty. */
    if (strcmp(output_file_name, "") == 0)
    {
        IAS_LOG_ERROR("No output filename provided in the parameter file");
        return ERROR;
    }

    /* Using the satellite number read from the OMF file, initialize the
       satellite attributes library */
    sat_id = ias_sat_attr_get_satellite_id_from_satellite_number(sat_num);
    if (sat_id == ERROR)
    {
        IAS_LOG_ERROR("Unsupported satellite number: %d", sat_num);
        return ERROR;
    }
    if (ias_sat_attr_initialize(sat_id) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attribute library");
        return ERROR;
    }

    /* Find the number of normal bands for the satellite. */
    satellite_number_bands = ias_sat_attr_get_normal_band_count();
    if ((satellite_number_bands < 1) ||
        (satellite_number_bands > IAS_MAX_NBANDS))
    {
        IAS_LOG_ERROR("Invalid number of bands %d", satellite_number_bands);
        return ERROR;
    }

    /* Set the reference band indexes.  Don't check for duplicates. */
    status = set_band_indexes(reference_band_number_list, 0,
        satellite_number_bands, reference_band_indexes, number_bands); 
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Could not set the reference band indexes");
        return ERROR;
    }
    reference_number_bands = *number_bands;

    /* Set the search band indexes.  Check for duplicates. */
    status = set_band_indexes(search_band_number_list, 1,
        satellite_number_bands, search_band_indexes, number_bands); 
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Could not set the search band indexes");
        return ERROR;
    }

    /* Make sure that the number of bands is the same for the reference and
       search band lists */
    if (reference_number_bands != *number_bands)
    {
        IAS_LOG_ERROR(
            "The reference and search band lists are not the same size");
        return ERROR;
    }

    /* Validate the tiept_type. */
    if (tiept == 1) /* Build regular-spaced tiepts. */
        *tiept_type = REGULAR_SPACED;
    else if (tiept == 2) /* Read the tiept file. */
        *tiept_type = READ_FILE;

    /*==========================================================================
     * Validate the OMF parameters
     *========================================================================*/

    /* Get the DOQ if L1G_REF_IMAGE is not set. */
    if (strcmp(ref_1G, "") == 0)
    {
        /* Assign the DOQ file name. */
        IMG_ptr = getenv("IAS_IMG");
        if ((IMG_ptr != NULL) && (strlen(IMG_ptr) > 0))
        {
            if ((wrs_path == INVALID_PARAMETER) ||
                (wrs_row == INVALID_PARAMETER))
            {
                /* WRS path and/or row weren't selected, but are needed. */
                IAS_LOG_ERROR("WRS path and/or row were not specified, "
                        "but they are needed to build the DOQ image name.");
                return ERROR;
            }

            /* Construct the input DOQ image filename. */
            if (sprintf(ref_1G, "%s/ias%03d%03d_UTM_doq.h5", IMG_ptr, wrs_path,
                    wrs_row) < 0)
            {
                /* There was an error constructing the DOQ file name. */
                IAS_LOG_ERROR("constructing the DOQ file name");
                return ERROR;
            }
        }
        /* We can not get the path for the DOQ's, so log an error. */
        else
        {
            IAS_LOG_ERROR("retrieving the DOQ path");
            return ERROR;
        }
    }

    /* Transfer the various local arrays to the function parameters. */
    strcpy(ref_L0R, local_ref_L0R);
    strcpy(cpf_file_name, local_cpf_file_name);
    strcpy(l1g_image, local_l1g_image);

    if (*tiept_type == READ_FILE)
    {
        if (strcmp(local_gcp_directory, "") == 0)
        {
            IAS_LOG_ERROR("No GCP directory provided in OMF file %s",
                    omf_file_name);
            return ERROR;
        }

        /* Form the GCP library filename with the full path based on the GCP
           directory and the standard filename. */ 
        sprintf(gcplib_file_name, "%s/GCPLib", local_gcp_directory);
    }

    /* Exit if the CPF filename is empty. */
    if (strcmp(cpf_file_name, "") == 0)
    {
        IAS_LOG_ERROR("No CPF filename provided in OMF file %s",
                omf_file_name);
        return ERROR;
    }

    /* If the L1G_SRCH_IMAGE is not set, use the L1G_IMAGE.  If the L1G_IMAGE is
       also not set, issue an error. */
    if (strcmp(srch_1G, "") == 0)
    {
        if (strcmp(l1g_image, "") == 0)
        {
            IAS_LOG_ERROR("No L1G filename provided in OMF file %s",
                    omf_file_name);
            return ERROR;
        }
        else
        {
            sprintf(srch_1G, "%s", l1g_image);
        }
    }

    /* Now read and check the CPF. */

    /* Open the CPF. */
    cpf = ias_cpf_read(cpf_file_name);
    if (!cpf)
    {
        IAS_LOG_ERROR("Cannot read CPF file %s", cpf_file_name);
        return ERROR;
    }

    /* Get the needed groups from the CPF. */
    cpf_i2i = ias_cpf_get_i2i_assessment(cpf);
    if (!cpf_i2i)
    {
        IAS_LOG_ERROR("getting I2I assessment group from CPF %s", 
                cpf_file_name);
        ias_cpf_free(cpf);
        return ERROR;
    }

    /* Retrieve values from the CPF as needed. */
    if (IAS_PARM_GET_COUNT_READ(parms, CHIP_SIZE) < 1)
    {
        chip_size->line = cpf_i2i->corr_window_size[0];
    }
    if (IAS_PARM_GET_COUNT_READ(parms, CHIP_SIZE) < 2)
    {
        chip_size->samp = cpf_i2i->corr_window_size[1];
    }
    if (IAS_PARM_GET_COUNT_READ(parms, RANGE) < 1)
    {
        range[0] = cpf_i2i->corr_pix_min;
    }
    if (IAS_PARM_GET_COUNT_READ(parms, RANGE) < 2)
    {
        range[1] = cpf_i2i->corr_pix_max;
    }
    if (IAS_PARM_GET_COUNT_READ(parms, MIN_CORR) == 0)
    {
        *min_corr = cpf_i2i->min_corr_strength;
    }
    if (IAS_PARM_GET_COUNT_READ(parms, MAX_DISP) == 0)
    {
        *max_disp = cpf_i2i->max_displacement_offset;
    }
    if (IAS_PARM_GET_COUNT_READ(parms, THRESHOLD) == 0)
    {
        *threshold = cpf_i2i->fill_threshold;
    }
    *fit_method = cpf_i2i->corr_fit_method;

    /* Free CPF memory. */
    ias_cpf_free(cpf);

    /* Set the absolute correlation coefficient to FALSE. */
    *abs_corr_coeff = FALSE;

    return SUCCESS;
}

/****************************************************************************
Name: set_band_indexes

Purpose: Local routine to set the band indexes, as well as the number of bands
in the band index list, based on the input band list

Returns: SUCCESS or ERROR
******************************************************************************/
int set_band_indexes
(
    const int band_number_list[IAS_MAX_NBANDS], /* I: Band number list source 
                                                for band indexes             */
    int check_duplicates,                 /* I: Flag for checking for dupes  */ 
    int maximum_number_bands,             /* I: Maximum allowed # of bands   */
    int band_indexes[IAS_MAX_NBANDS],     /* O: Band index list              */
    int *number_bands                     /* O: # bands in band index list   */
)
{
    int i;                                /* Band loop variable              */
    int j;                                /* Band loop variable              */

    /* Find the number of bands in the band number list.  The first 0 
       encountered signifies the end of the band number list. */
    for (*number_bands = 0; *number_bands < maximum_number_bands; 
        (*number_bands)++)
    {
        if (band_number_list[*number_bands] == 0)
        {
            /* We found the end of the band number list. */
            break;
        }
    }

    /* Make sure there are no duplicate bands.  Error out if there are. */
    if (check_duplicates)
    {
        for (i = 0; i < (*number_bands - 1); i++)
        {
            for (j = i + 1; j < *number_bands; j++)
            {
                 if (band_number_list[i] == band_number_list[j])
                 {
                     IAS_LOG_ERROR("Duplicate band number (%d) in parameter "
                         "file", band_number_list[i]);
                     return ERROR;
                 }
            }
        }
    }

    /* Fill in the band indexes. */
    for (i = 0; i < *number_bands; i++)
    {
        band_indexes[i] = ias_sat_attr_convert_band_number_to_index(
            band_number_list[i]);
        if (band_indexes[i] == ERROR)
        {
            IAS_LOG_ERROR("Invalid band index for band number %d", 
                band_number_list[i]);
            return ERROR;
        }
    }

    return SUCCESS;
}
