/******************************************************************************
NAME:           i2i_char

PURPOSE:  To implement an image to image characterization between the reference
and search images for the specified bands.


RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
FAILURE  Operation failed

NOTES:
The i2i_char application ultimately performs a statistical measure of all the
geometric shifts relating the measured offsets among all bands (T-distribution),
but this portion of the application does not implement the final outlier
detection.  That is done in the T-distribution code.  This routine is used to 
do the correlation, then output those results to a file.  T-Distribution reads 
the results from the correlation and implements the T-distribution final 
outlier detection.

If no bands are specified all OLI bands will be characterized.

ALGORITHM REFERENCES:
See ias_gcp.h for a description of the IAS_GCP_RECORD structure.
See ias_structures.h for a description of the LNG_LS structure.
******************************************************************************/
#include <stdlib.h>
#include <string.h>

#include "i2i_char.h"
#include "ias_logging.h"           /* Logging library                        */

int main (int argc, char *argv[])
{
    /* Variable declarations.  */
    int status;                    /* Status return value                    */
    int number_bands;              /* # of bands to characterize for both 
                                      reference and search                   */
    int number_line_tiepoints;     /* # GCP points in line direction         */
    int number_sample_tiepoints;   /* # GCP points in the sample direction   */
    int number_of_gcps = 0;        /* Total # GCP points                     */
    int abs_corr_coeff;            /* flag to use the absolute value for
                                      correlation coeffs in grey correlation */
    enum TIEPT_TYPE tiept_type;    /* Variable to specify the type of tiepts:
                                        - Regular spaced tiepoints
                                        - Tiepoints read from a file         */
    float min_corr;                /* Minimum correlation strength           */
    float max_disp;                /* Maximum correlation displacement       */
    float threshold;               /* Invalid threshold precentage for
                                      buffer                                 */
    float range[2];                /* DNs above/below this are invalid       */
    IAS_CORRELATION_FIT_TYPE fit_method;  /* Correlation result fit method   */
    int reference_band_indexes[IAS_MAX_NBANDS]; /* Reference bands to 
                                      correlate                              */
    int search_band_indexes[IAS_MAX_NBANDS]; /* Reference bands to correlate */
    struct IAS_LNG_LS chip_size;   /* Size of the chip for correlation       */
    char ref_L0R[PATH_MAX];        /* L0R HDF filename                       */
    char gcplib_filename[PATH_MAX]; /* Name of GCPLib file (if needed)       */
    char output_filename[PATH_MAX]; /* Name of output file                   */
    char ref_l1g_filename[PATH_MAX]; /* Name of reference L1G image file     */
    char srch_l1g_filename[PATH_MAX];/* Name of search L1G image file        */
    char cpf_filename[PATH_MAX];   /* Name of the calibration parameter file */
    char omf_file_name[PATH_MAX];  /* Name of the omf file                   */
    char work_order_id[IAS_WORKORDER_ID_SIZE]; /* Work order ID string     */
    IAS_L1G_BAND_METADATA ref_band_metadata[IAS_MAX_NBANDS]; /* Reference 
                                      image band metadata array              */
    IAS_L1G_BAND_METADATA srch_band_metadata[IAS_MAX_NBANDS];  /* Search 
                                      image band metadata array              */
    IAS_L1G_FILE_METADATA ref_file_metadata;/* Reference image file metadata */
    IAS_L1G_FILE_METADATA srch_file_metadata; /* Search image file metadata  */
    IAS_GCP_RECORD *ref_gcp15 = NULL; /* Ref GCP points array for 15m band   */
    IAS_GCP_RECORD *srch_gcp15 = NULL;/* Search GCP points array for 15m 
                                         band                                */
    IAS_GCP_RECORD *ref_gcp30 = NULL; /* Ref GCP points array for 30m band   */
    IAS_GCP_RECORD *srch_gcp30 = NULL;/* Search GCP points array for 30m 
                                         band                                */
    IAS_DBL_XY fract_offset_15;    /* Two element structure containing the 
                                      fractional parts of the search image 
                                      offset from the reference image for 
                                      the 15m band.                          */
    IAS_DBL_XY fract_offset_30;    /* Two element structure containing the 
                                      fractional parts of the search image 
                                      offset from the reference image for 
                                      the 30m band.                          */
    int ref_number_lines[IAS_MAX_NBANDS]; /* Number of reference image lines */
    int ref_number_samples[IAS_MAX_NBANDS]; /* Number of reference image 
                                      samples                                */
    int srch_number_lines[IAS_MAX_NBANDS]; /* Number of search image lines   */
    int srch_number_samples[IAS_MAX_NBANDS]; /* Number of search image 
                                      samples                                */

    /* Initialize the logging library. */
    if (ias_log_initialize("i2i_char") != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing logging library");
        exit(EXIT_FAILURE);
    }

    /* Check the arguments. */
    if (argc != 2)
    {
        IAS_LOG_ERROR("USAGE:  i2i_char < parameter file name >");
        exit(EXIT_FAILURE);
    }

    /* Initialize the fractional offsets. */
    fract_offset_15.x = 0.0;
    fract_offset_15.y = 0.0;
    fract_offset_30.x = 0.0;
    fract_offset_30.y = 0.0;

    /* Get the i2i_char parameters and initialize the satellite attributes */
    status = get_parms(argv[1], ref_L0R, ref_l1g_filename, srch_l1g_filename,
                       gcplib_filename, cpf_filename, output_filename,
                       reference_band_indexes, search_band_indexes, 
                       &number_bands, &number_line_tiepoints, 
                       &number_sample_tiepoints, &chip_size, &tiept_type, 
                       &min_corr, &max_disp, range, &threshold, &fit_method, 
                       &abs_corr_coeff, work_order_id, omf_file_name);  
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Reading work order parameters");
        exit(EXIT_FAILURE);
    }

    /* Print the starting message.  */
    IAS_LOG_INFO("Initiating image registration accuracy characterization...");

    /* Read and compare the reference and search metadata. */
    status = compare_metadata(ref_l1g_filename, srch_l1g_filename, 
                              reference_band_indexes, search_band_indexes, 
                              number_bands, &ref_file_metadata, 
                              &srch_file_metadata, ref_band_metadata, 
                              srch_band_metadata, ref_number_lines, 
                              ref_number_samples, srch_number_lines, 
                              srch_number_samples);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Reading/comparing metadata information");
        exit(EXIT_FAILURE);
    }

    /* Get the tiepoints.  */
    if (tiept_type == REGULAR_SPACED)   
    {
        /* Create regular-spaced tiepoints for the band resolutions. */
        number_of_gcps = number_line_tiepoints * number_sample_tiepoints;
        status =  create_tiepts(tiept_type, reference_band_indexes, 
                                search_band_indexes, number_bands, 
                                ref_band_metadata, srch_band_metadata, 
                                ref_number_lines, ref_number_samples, 
                                srch_number_lines, srch_number_samples, 
                                number_line_tiepoints, number_sample_tiepoints,
                                chip_size, &fract_offset_15, &fract_offset_30, 
                                &ref_gcp15, &srch_gcp15, &ref_gcp30, 
                                &srch_gcp30);
        if (status == ERROR)
        {
            IAS_LOG_ERROR("Creating regular-spaced tiepoints");
            exit(EXIT_FAILURE);
        }
    }
    else if (tiept_type == READ_FILE)   /* Read the tiepoint file. */
    {
        /* Read the GCP library file and determine the reference and search 15m
           GCPs, 30m GCPs, and 60m GCPs if necessary. */
        status = read_gcps(gcplib_filename, reference_band_indexes, 
            search_band_indexes, number_bands, ref_band_metadata, 
            srch_band_metadata, &ref_file_metadata, &srch_file_metadata, 
            ref_number_lines, ref_number_samples, srch_number_lines, 
            srch_number_samples, &number_of_gcps, &ref_gcp15, &srch_gcp15, 
            &ref_gcp30, &srch_gcp30);
        if (status == ERROR)
        {
            IAS_LOG_ERROR("Reading GCP Library tiepoints");
            exit(EXIT_FAILURE);
        }
    }
    else
    {
        IAS_LOG_ERROR("Invalid tiepoint data source");
        exit(EXIT_FAILURE);
    }

    /* Implement the correlation on all the bands. */
    status = i2i_corr(ref_gcp15, srch_gcp15, ref_gcp30, srch_gcp30, 
                      fract_offset_15, fract_offset_30, number_of_gcps,
                      ref_L0R, ref_l1g_filename, srch_l1g_filename, 
                      ref_band_metadata, srch_band_metadata,
                      reference_band_indexes, search_band_indexes, number_bands,
                      output_filename, min_corr, max_disp, fit_method, 
                      abs_corr_coeff, threshold, range[1], range[0], 
                      work_order_id);

    /* Free the used pointers. */
    free(ref_gcp15);
    free(srch_gcp15);
    free(ref_gcp30);
    free(srch_gcp30);

    if (status == ERROR)
    {
        IAS_LOG_ERROR("Correlating reference and search images");
        exit(EXIT_FAILURE);
    }

    /* update omf file with output file name */
    if (update_omf(omf_file_name, output_filename) != SUCCESS)
    {
        IAS_LOG_ERROR("Updating omf file.");
        exit(EXIT_FAILURE);
    }

    /* Print the completion message. */
    IAS_LOG_INFO("Image registration accuracy characterization successfully "
                 "completed");

    return EXIT_SUCCESS;
} /* done */
