#ifndef _I2I_CHAR_H_
#define _I2I_CHAR_H_

#include "ias_l1g.h"                  /* L1G I/O library                     */
#include "ias_gcp.h"                  /* GCP library                         */
#include "ias_satellite_attributes.h" /* Satellite attributes library        */
#include "ias_math_parallel_corr.h"   /* Parallel correlator                 */
#include "ias_geo.h"

enum TIEPT_TYPE {REGULAR_SPACED, READ_FILE};

#define TOLERANCE 0.000001     /* Tolerance used for comparing float data    */

/* Check if the projection distance is approximately at a specified value */
#define IS_PROJ_DIST(dist, target) \
        ((dist > target - TOLERANCE) && (dist < target + TOLERANCE))

/* Prototype definitions for the I2I_CHAR routines */
int compare_metadata
(
    const char *ref_l1g_name,  /* I: Reference L1G image filename            */
    const char *srch_l1g_name, /* I: Search L1G image filename               */
    const int *reference_band_indexes, /* I: Array of reference band indexes */
    const int *search_band_indexes, /* I: Array of search band indexes       */
    const int number_bands,    /* I: Number of bands in the band indexes   
                                     array                                   */
    IAS_L1G_FILE_METADATA *ref_file_metadata, /* O: Reference image file 
                                     metadata                                */ 
    IAS_L1G_FILE_METADATA *srch_file_metadata, /* O: Search image file 
                                     metadata                                */
    IAS_L1G_BAND_METADATA *ref_band_metadata, /* O: Reference image band 
                                     metadata array                          */
    IAS_L1G_BAND_METADATA *srch_band_metadata, /* O: Search image band 
                                     metadata array                          */
    int *ref_number_lines,     /* O: Number of reference lines               */
    int *ref_number_samples,   /* O: Number of reference samples             */
    int *srch_number_lines,    /* O: Number of search lines                  */
    int *srch_number_samples   /* O: Number of search samples                */
);

int create_tiepts 
(
    enum TIEPT_TYPE tiept_type, /* I: Variable to specify the type of 
                                      tiepts                                 */
    const int *reference_band_indexes, /* I: Array of reference band indexes */
    const int *search_band_indexes, /* I: Array of search band indexes       */
    int number_bands,           /* I: Number of band indexes                 */
    const IAS_L1G_BAND_METADATA *ref_band_metadata,  /* I: Reference image 
                                      band metadata array                    */
    const IAS_L1G_BAND_METADATA *srch_band_metadata, /* I: Search image band 
                                      metadata array                         */
    const int *ref_number_lines, /* I: Number of lines in reference image    */
    const int *ref_number_samples,/* I: Number of samples in reference image */
    const int *srch_number_lines, /* I: Number of lines in search image      */
    const int *srch_number_samples, /* I: Number of samples in search image  */
    int number_line_tiepoints,  /* I: Number of line tiepoints to create     */
    int number_sample_tiepoints, /* I: Number of sample tiepoints to create  */
    const struct IAS_LNG_LS chip_size, /* I: Chip size of the tiepoints to 
                                      create                                 */
    IAS_DBL_XY *fract_offsetPTR_15, /* O: Pointer to two element structure 
                                      containing fractional parts of search
                                      image offset from reference image for
                                      the 15m band.                          */
    IAS_DBL_XY *fract_offsetPTR_30, /* O: Pointer to two element structure 
                                      containing fractional parts of search
                                      image offset from reference image for
                                      the 30m band.                          */
    IAS_GCP_RECORD **ref_gcp15, /* O: Array of ref GCPLib tiepoints in row-
                                      major form for 15m band resolutions    */
    IAS_GCP_RECORD **srch_gcp15,/* O: Array of srch GCPLib tiepoints in row-
                                      major form for 15m band resolutions    */
    IAS_GCP_RECORD **ref_gcp30, /* O: Array of ref GCPLib tiepoints in row-
                                      major form for 30m band resolutions    */
    IAS_GCP_RECORD **srch_gcp30 /* O: Array of srch GCPLib tiepoints in row-
                                      major form for 30m band resolutions    */
);

int get_parms 
(
    const char *param_filename,   /* I: Parameter file name                  */
    char *ref_L0R,                /* O: L0R HDF filename                     */
    char *ref_1G,                 /* O: OLI file name for the 1G ref image   */
    char *srch_1G,                /* O: OLI file name for the 1G srch image  */
    char *gcplib_file_name,       /* O: Name of the tiepoint file            */
    char *cpf_file_name,          /* O: Name of the CPF                      */
    char *output_file_name,       /* O: Name of the output file              */
    int reference_band_indexes[IAS_MAX_NBANDS], /* O: Reference band indexes */
    int search_band_indexes[IAS_MAX_NBANDS], /* O: Search band indexes       */
    int *number_bands,            /* O: Number of reference bands            */
    int *number_line_tiepoints,   /* O: Number of line tiepoints             */
    int *number_sample_tiepoints, /* O: Number of samp tiepoints             */
    struct IAS_LNG_LS *chip_size, /* O: Height and width of the chips        */
    enum TIEPT_TYPE *tiept_type,  /* O: Type of tiepts to use.               */
    float *min_corr,              /* O: Minimum correlation strength         */
    float *max_disp,              /* O: Maximum correlation displacement     */
    float *range,                 /* O: Range of valid pixels                */
    float *threshold,             /* O: Percentage of valid pixels           */
    IAS_CORRELATION_FIT_TYPE *fit_method, /* O: Correlation result fit 
                                        method                               */
    int *abs_corr_coeff,          /* O: Flag absolute correlation coeff to 
                                        be used                              */
    char *wo_id,                  /* O: Work order ID string                 */
    char *omf_file_name           /* O: omf file name                        */
);

int i2i_corr 
(
    IAS_GCP_RECORD *ref_gcp15,   /* I: Array of ref GCP points for 15m band 
                                       resolution                            */
    IAS_GCP_RECORD *srch_gcp15,  /* I: Array of search GCP points for 15m 
                                       band resolution                       */
    IAS_GCP_RECORD *ref_gcp30,   /* I: Array of ref GCP points for 30m band 
                                       resolution                            */
    IAS_GCP_RECORD *srch_gcp30,  /* I: Array of search GCP points for 30m 
                                       band resolution                       */
    IAS_DBL_XY fract_offset_15,  /* I: Two element structure containing the
                                       fractional offsets of search image
                                       from reference image for 15m band.    */
    IAS_DBL_XY fract_offset_30,  /* I: Two element structure containing the
                                       fractional offsets of search image
                                       from reference image for 30m band.    */
    int number_of_gcps,          /* I: Number of GCPs                        */
    const char *ref_L0R,         /* I: Name of the reference L0R HDF         */
    const char *ref_img_name,    /* I: Name of 1G OLI reference image        */
    const char *srch_img_name,   /* I: Name of 1G OLI search image           */
    const IAS_L1G_BAND_METADATA *ref_band_metadata, /* I: Reference image
                                      band metadata                          */
    const IAS_L1G_BAND_METADATA *srch_band_metadata, /* I: Search image band
                                      metadata                               */
    const int *reference_band_indexes, /* I: Array of reference band indexes */
    const int *search_band_indexes, /* I: Array of search band indexes       */
    int number_bands,            /* I: Number of band indexes                */
    const char *output_filename, /* I: Name of output data file              */
    float min_corr,              /* I: Minimum correlation strength          */
    float max_disp,              /* I: Maximum correlation displacement      */
    IAS_CORRELATION_FIT_TYPE fit_method, /* I: Correlation result fit 
                                       method                                */
    int abs_corr_coeff,          /* I: Flag to use abs corr coefs            */
    float invalid_threshold,     /* I: Invalid threshold percentage for 
                                       const buffer                          */
    float valid_image_max,       /* I: DNs above this are invalid            */
    float valid_image_min,       /* I: DNs below this are invalid            */
    const char *work_order_id    /* I: Work Order ID for report header       */
);

int i2i_det_tiepts 
(
    const IAS_L1G_BAND_METADATA *ref_band_metadata,  /* I: Reference image 
                                      metadata for the desired band          */
    const IAS_L1G_BAND_METADATA *srch_band_metadata, /* I: Search image 
                                      metadata for the desired band          */
    int ref_number_lines,       /* I: Number of reference lines              */
    int ref_number_samples,     /* I: Number of reference samples            */
    int srch_number_lines,      /* I: Number of search lines                 */
    int srch_number_samples,    /* I: Number of search samples               */
    int number_line_tiepoints,  /* I: Number of line tiepoints to create     */
    int number_sample_tiepoints, /* I: Number of sample tiepoints to create  */
    struct IAS_LNG_LS chip_size,/* I: Chip size of the tiepoints to create   */
    IAS_GCP_RECORD *ref_gcps,   /* O: Array of reference GCPLib tiepoints in
                                      row-major format                       */
    IAS_GCP_RECORD *srch_gcps,  /* O: Array of search GCPLib tiepoints in
                                      row-major format                       */
    IAS_DBL_XY *fract_offsetPTR /* O: Pointer to two element structure 
                                      containing the fractional parts of the 
                                      search image offset from the reference
                                      image.                                 */
);

int process_gcp 
(
    const IAS_GCP_RECORD *ref_gcp_lib,  /* I: List of reference GCPs         */ 
    const IAS_GCP_RECORD *srch_gcp_lib, /* I: List of search GCPs            */
    int number_of_gcps,           /* I: Number of GCPs                       */
    L1G_BAND_IO *ref_l1g_band,    /* I: Reference 1G band pointer            */
    L1G_BAND_IO *srch_l1g_band,   /* I: Search 1G band pointer               */
    int reference_band_index,     /* I: Reference band number to process     */
    int search_band_index,        /* I: Search band number to process        */
    IAS_DATA_TYPE ref_dtype,      /* I: Data type of reference image         */
    IAS_DATA_TYPE srch_dtype,     /* I: Data type of search image            */
    int ref_number_lines,         /* I: Number of lines in ref image         */
    int ref_number_samples,       /* I: Number of samples in ref image       */
    int srch_number_lines,        /* I: Number of lines in srch image        */
    int srch_number_samples,      /* I: Number of samples in srch image      */
    float min_corr,               /* I: Minimum correlation strength         */
    float max_disp,               /* I: Maximum correlation displacement     */
    IAS_CORRELATION_FIT_TYPE fit_method, /* I: Correlation result fit 
                                        method                               */
    IAS_GCP_RESULTS *gcp,         /* O: GCP data structures (1 per pt)      
                                        for the correlation results          */
    IAS_DBL_XY fract_offset,      /* I: Fractional parts of mean offsets
                                        of search image relative to ref
                                        image.                               */
    int abs_corr_coeff,           /* I: Flag to use abs of corr coeffs       */
    float invalid_threshold,      /* I: Invalid threshold percentage for
                                        buffer                               */
    float valid_image_max,        /* I: DNs above this are invalid           */
    float valid_image_min,        /* I: DNs below this are invalid           */
    IAS_PARALLEL_CORRELATOR_TYPE *correlator_ptr, /* I: Pointer to the
                                        parallel correlator structure        */
    IAS_CORRELATION_RESULT_TYPE *result_ptr /* O: Pointer to the
                                        correlation results                  */
);

int read_metadata
(
    const char *filename,        /* I: L1G image file name                   */
    const int *band_indexes,     /* I: Array of band indexes                 */
    int number_bands,            /* I: Number of bands in the band indexes 
                                       array                                 */
    int *number_lines,           /* O: Number of lines in image              */
    int *number_samples,         /* O: Number of samples in image            */
    IAS_L1G_FILE_METADATA *file_metadata, /* I/O: Image file metadata        */
    IAS_L1G_BAND_METADATA *band_metadata  /* I/O: Image band metadata        */
);

int read_gcps 
(
    const char *gcp_filename,    /* I: GCP file name                         */
    const int *reference_band_indexes, /* I: Reference band indexes          */
    const int *search_band_indexes, /* I: Search band indexes                */
    int number_bands,            /* I: Number of band indexes                */
    const IAS_L1G_BAND_METADATA *ref_band_metadata,  /* I: Reference image 
                                       metadata in band-referenced order     */
    const IAS_L1G_BAND_METADATA *srch_band_metadata, /* I: Search image 
                                       metadata in band-referenced order     */
    const IAS_L1G_FILE_METADATA *ref_file_metadata, /* I: Reference image file 
                                       metadata                              */
    const IAS_L1G_FILE_METADATA *srch_file_metadata, /* I: Search image file 
                                       metadata                              */
    const int *ref_number_lines, /* I: Number of lines in reference image    */
                                       
    const int *ref_number_samples,/* I: Number of samples in reference image */
    const int *srch_number_lines, /* I: Number of lines in search image      */
    const int *srch_number_samples, /* I: Number of samples in search image  */
    int  *number_of_gcps,        /* O: Number of GCPs                        */
    IAS_GCP_RECORD **ref_gcp15,  /* O: Ref GCPs for 30m band resolution      */
    IAS_GCP_RECORD **srch_gcp15, /* O: Search GCPs for 30m band resolution   */
    IAS_GCP_RECORD **ref_gcp30,  /* O: Ref GCPs for 30m band resolution      */
    IAS_GCP_RECORD **srch_gcp30  /* O: Search GCPs for 30m band resolution   */
);

int get_gcp_line_sample
(
    const IAS_L1G_BAND_METADATA *band_metadata, /* I: Image band metadata    */
    const IAS_GEO_PROJ_TRANSFORMATION *degrees2map, /* I: projection
                                            transformation to use */
    const int number_lines,           /* I: Number of lines in image         */
    const int number_samples,         /* I: Number of samples in image       */
    const int number_of_gcps,         /* I: Number of GCPs in the GCP 
                                            library                          */
    const IAS_GCP_RECORD *in_gcps,    /* I: Input array of GCP record 
                                            tiepoints in row-major format    */
    IAS_GCP_RECORD *gcps              /* O: Output array of GCP record 
                                            tiepoints in row-major format    */
);

int update_omf
(
   char *omf_file_name,           /* I: omf file name */
   char *output_file_name         /* I: b2bchar output file name */
);


#endif
