/******************************************************************************
NAME:       i2i_corr

PURPOSE:  This routine will implement the image-to-image correlation on the
          specified bands.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
This routine will open the required images for each band.  Then the
correlation will be implemented for each GCP point in the images.  The final
results will be output to a single file with all band results.

******************************************************************************/
#include <stdlib.h>
#include <string.h>

#include "i2i_char.h"
#include "ias_miscellaneous.h"        /* Miscellaneous IAS libraries         */
#include "ias_logging.h"              /* Logging library                     */
#include "ias_math_parallel_corr.h"   /* Parallel correlator                 */
#include "ias_satellite_attributes.h" /* Satellite attributes library        */

int i2i_corr 
(
    IAS_GCP_RECORD *ref_gcp15,    /* I: Array of ref GCP points for 15m band 
                                        resolution                           */ 
    IAS_GCP_RECORD *srch_gcp15,   /* I: Array of search GCP points for 15m 
                                        band resolution                      */
    IAS_GCP_RECORD *ref_gcp30,    /* I: Array of ref GCP points for 30m 
                                        band resolution                      */
    IAS_GCP_RECORD *srch_gcp30,   /* I: Array of search GCP points for 30m 
                                        band resolution                      */
    IAS_DBL_XY fract_offset_15,   /* I: Two element structure containing the
                                        fractional parts of the search image
                                        offset from the reference image for
                                        the 15m band.                        */
    IAS_DBL_XY fract_offset_30,   /* I: Two element structure containing the 
                                        fractional parts of the search image
                                        offset from the reference image for
                                        the 30m band.                        */
    int number_of_gcps,           /* I: Number of GCPs                       */
    const char *ref_L0R,          /* I: Name of the reference L0R HDF        */
    const char *ref_img_name,     /* I: Name of 1G OLI reference image       */
    const char *srch_img_name,    /* I: Name of 1G OLI search image          */
    const IAS_L1G_BAND_METADATA *ref_band_metadata, /* I: Reference image
                                      band metadata                          */
    const IAS_L1G_BAND_METADATA *srch_band_metadata, /* I: Search image band
                                      metadata                               */
    const int *reference_band_indexes, /* I: Array of reference band indexes */ 
    const int *search_band_indexes, /* I: Array of search band indexes       */ 
    int number_bands,             /* I: Number of band indexes               */
    const char *output_filename,  /* I: Name of output data file             */ 
    float min_corr,               /* I: Minimum correlation strength         */
    float max_disp,               /* I: Maximum correlation displacement     */
    IAS_CORRELATION_FIT_TYPE fit_method, /* I: Correlation result fit method */ 
    int abs_corr_coeff,           /* I: flag to use absolute corr coeffs     */
    float invalid_threshold,      /* I: Invalid threshold percentage for 
                                        buffer                               */
    float valid_image_max,        /* I: DNs above this are invalid           */
    float valid_image_min,        /* I: DNs below this are invalid           */
    const char *work_order_id     /* I: Work Order ID for report header      */
)
{
    IAS_GCP_RECORD *ref_gcplib;   /* list of reference GCPs (gcp15, gcp30)   */
    IAS_GCP_RECORD *srch_gcplib;  /* list of search GCPs (gcp15, gcp30)      */
    IAS_GCP_RESULTS *gcp;         /* GCP data structures (1 per pt)          */
    L1GIO *ref_img_ptr;           /* reference 1G image pointer              */
    L1GIO *srch_img_ptr;          /* search 1G image pointer                 */
    FILE *output;                 /* GCP output ASCII file pointer           */
    IAS_DATA_TYPE ref_dtype[IAS_MAX_NBANDS];  /* reference data type         */
    IAS_DATA_TYPE srch_dtype[IAS_MAX_NBANDS]; /* search data type            */
    int reference_band_number;    /* Reference band number                   */
    int search_band_number;       /* Search band number                      */
    int band_index;               /* Band index                              */
    int reference_band_index;     /* Reference band index                    */
    int search_band_index;        /* Search band index                       */
    double projection_distance;   /* Make "if" condition easier to read      */
    int found_15m;                /* Flag indicating a 15m band is present   */
    int found_30m;                /* Flag indicating a 30m band is present   */
    int one = 1;                  /* number of bands to read                 */
    int ref_number_lines[IAS_MAX_NBANDS]; /* number of lines in reference 
                                     image                                   */
    int ref_number_samples[IAS_MAX_NBANDS]; /* number of samps in reference 
                                     image                                   */
    int srch_number_lines[IAS_MAX_NBANDS]; /* number of lines in search 
                                     image                                   */
    int srch_number_samples[IAS_MAX_NBANDS]; /* number of samps in search
                                     image */
    int i;                        /* looping variable                        */
    int j;                        /* looping variable                        */
    int curr;                     /* looping variable for bands              */
    int curr_gcp;                 /* GCP counter                             */
    int num_correlated;           /* number of GCPs that correlated          */
    int status;                   /* function return status                  */

    IAS_DBL_XY fract_offset;      /* Fractional parts of mean offsets of
                                     current band.                           */
    double *offset_line;          /* array to store line offsets for the
                                     standard deviation calculation          */
    double *offset_sample;        /* array to store sample offsets for the
                                     standard deviation calculation          */
    double mean_offset_line;      /* mean of line deltas for band combo      */
    double mean_offset_sample;    /* mean of sample deltas for band combo    */
    double stddev_offset_line;    /* std dev of line deltas for band combo   */ 
    double stddev_offset_sample;  /* std dev of sample deltas for band combo */ 
    L1G_BAND_IO *ref_l1g_band;    /* reference L1G band pointer              */ 
    L1G_BAND_IO *srch_l1g_band;   /* search L1G band pointer                 */
    IAS_REPORT_HEADER *header = NULL; /* report header structure             */
    IAS_PARALLEL_CORRELATOR_TYPE *correlator_ptr = NULL; /* pointer to the 
                                     parallel correlator data structure      */
    IAS_CORRELATION_RESULT_TYPE *result_ptr = NULL; /* pointer to the 
                                     correlation results                     */
    int ref_max_chip_lines = 0;   /* maximum number of lines in a
                                     reference chip                          */
    int ref_max_chip_samples = 0; /* maximum number of samples in a
                                     reference chip                          */
    int search_max_chip_lines = 0; /* maximum number of lines in a
                                     reference chip                          */
    int search_max_chip_samples = 0; /* maximum number of samples in a
                                     reference chip                          */

    /* Allocate space for GCP data (one point per GCP library point). */
    gcp = malloc(number_of_gcps * sizeof(IAS_GCP_RESULTS));
    if (gcp == NULL)
    {
        IAS_LOG_ERROR("Allocating buffer for %d GCP data points", 
            number_of_gcps);
        return ERROR;
    }

    /* Open the GCP output file for writing. */
    output = fopen(output_filename, "wt");
    if (output == NULL)
    {
        IAS_LOG_ERROR("Opening GCP output data file %s", output_filename);
        free(gcp);
        return ERROR;
    }

    /* Initialize the report header and write it to the output file */
    header = ias_misc_initialize_gps_report_header(srch_img_name, ref_L0R,
            work_order_id, "I2I_CHAR");
    if (!header)
    {
        IAS_LOG_ERROR("Initializing the GPS output file");
        free(gcp);
        fclose(output);
        return ERROR;
    }
    status = ias_misc_write_gps_report_header(output, header);
    ias_misc_free_report_header(header);

    if (status != SUCCESS) 
    {
        IAS_LOG_ERROR("Writing to the GPS output file");
        free(gcp);
        fclose(output);
        return ERROR;
    }

    /* Output the reference and search image and band numbers to the file. */
    if (fprintf(output,
               "Reference image: %s\n"
               "Search image: %s\n"
               "Reference bands: ",
               ref_img_name, srch_img_name) < 0)
    {
        IAS_LOG_ERROR("Writing reference/search information to data file");
        free(gcp);
        fclose(output);
        return ERROR;
    }

    for (i = 0; i < number_bands; i++)
    {
        reference_band_number = ias_sat_attr_convert_band_index_to_number(
            reference_band_indexes[i]);
        if (reference_band_number == ERROR)
        {
            IAS_LOG_ERROR("Invalid band number for reference band index %d", 
                reference_band_indexes[i]);
            free(gcp);
            fclose(output);
            return ERROR;
        }

        if (fprintf(output, " %d", reference_band_number) < 0)
        {
            IAS_LOG_ERROR("Writing reference band information to data file");
            free(gcp);
            fclose(output);
            return ERROR;
        }
    }

    if (fprintf(output, "\nSearch bands: ") < 0)
    {
        IAS_LOG_ERROR("Writing search information to data file");
        free(gcp);
        fclose(output);
        return ERROR;
    }
               
    for (i = 0; i < number_bands; i++)
    {
        search_band_number = ias_sat_attr_convert_band_index_to_number(
            search_band_indexes[i]);
        if (search_band_number == ERROR)
        {
            IAS_LOG_ERROR("Invalid band number for search band index %d", 
                search_band_indexes[i]);
            free(gcp);
            fclose(output);
            return ERROR;
        }

        if (fprintf(output, " %d", search_band_number) < 0)
        {
            IAS_LOG_ERROR("Writing search band information to data file");
            free(gcp);
            fclose(output);
            return ERROR;
        }
    }

    if (fprintf(output,
                "\n\nPoint       Ref_Line     Ref_Sample        Ref_Lat"
                "     Ref_Long   Ref_Elev    Search_Line   Search_Sample   "
                "Delta_Y      Delta_X       Flag  Coef     "
                "  Ref_Band  Search_Band  Ref_SCA Search_SCA          "
                "Ref_Image                              "
                "Search_Image\n") < 0)
    {
        IAS_LOG_ERROR("Writing GCP information to data file");
        free(gcp);
        fclose(output);
        return ERROR;
    }

    /* Open L1G images for reading. */
    ref_img_ptr = ias_l1g_open_image(ref_img_name, IAS_READ);
    if (ref_img_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening reference image %s", ref_img_name);
        free(gcp);
        fclose(output);
        return ERROR;
    }
    srch_img_ptr = ias_l1g_open_image(srch_img_name, IAS_READ);
    if (srch_img_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening search image %s", srch_img_name);
        ias_l1g_close_image(ref_img_ptr);
        free(gcp);
        fclose(output);
        return ERROR;
    }

    /* Find the largest reference and search sizes for both line and sample.  
       Do this for all band resolutions so we only need to initialize the 
       parallel correlator once.  However, only do this for sizes that are
       represented in the band lists. The projection distances should be the
       same for reference and search image bands, so only check reference. */
    found_15m = 0;
    found_30m = 0;
    for (band_index = 0; band_index < number_bands; band_index++)
    {
        reference_band_index = reference_band_indexes[band_index];
        projection_distance =
            ref_band_metadata[reference_band_index].projection_distance_x;
        if (IS_PROJ_DIST(projection_distance, 15.0))
        {
            found_15m = 1;
        }
        if (IS_PROJ_DIST(projection_distance, 30.0))
        {
            found_30m = 1;
        }
    }

    for (curr_gcp = 0; curr_gcp < number_of_gcps; curr_gcp++)
    {
        if (found_15m)
        {
            if (ref_gcp15[curr_gcp].chip_size_lines > ref_max_chip_lines)
                ref_max_chip_lines = ref_gcp15[curr_gcp].chip_size_lines;
            if (ref_gcp15[curr_gcp].chip_size_samples > ref_max_chip_samples)
                ref_max_chip_samples = ref_gcp15[curr_gcp].chip_size_samples;
            if (srch_gcp15[curr_gcp].chip_size_lines > search_max_chip_lines)
                search_max_chip_lines = srch_gcp15[curr_gcp].chip_size_lines;
            if (srch_gcp15[curr_gcp].chip_size_samples > 
                search_max_chip_samples)
                search_max_chip_samples = 
                    srch_gcp15[curr_gcp].chip_size_samples;
        }

        if (found_30m)
        {
            if (ref_gcp30[curr_gcp].chip_size_lines > ref_max_chip_lines)
                ref_max_chip_lines = ref_gcp30[curr_gcp].chip_size_lines;
            if (ref_gcp30[curr_gcp].chip_size_samples > ref_max_chip_samples)
                ref_max_chip_samples = ref_gcp30[curr_gcp].chip_size_samples;
            if (srch_gcp30[curr_gcp].chip_size_lines > search_max_chip_lines)
                search_max_chip_lines = srch_gcp30[curr_gcp].chip_size_lines;
            if (srch_gcp30[curr_gcp].chip_size_samples > 
                search_max_chip_samples)
                search_max_chip_samples = 
                    srch_gcp30[curr_gcp].chip_size_samples;
        }
    }

    /* Allocate memory for the results of the correlation. */
    result_ptr = malloc(number_of_gcps * sizeof(IAS_CORRELATION_RESULT_TYPE));
    if (result_ptr == NULL)
    {
        IAS_LOG_ERROR("Allocating correlation results buffer");
        ias_l1g_close_image(ref_img_ptr);
        ias_l1g_close_image(srch_img_ptr);
        free(gcp);
        fclose(output);
        return ERROR;
    }

    /* Initialize the parallel correlator. */
    correlator_ptr = ias_math_init_parallel_correlator(
            1, 1, 1, 1 /*check all flags*/,
            ref_max_chip_lines, ref_max_chip_samples,
            search_max_chip_lines, search_max_chip_samples,
            result_ptr);
    if (correlator_ptr == NULL)
    {
        ias_l1g_close_image(ref_img_ptr);
        ias_l1g_close_image(srch_img_ptr);
        free(gcp);
        fclose(output);
        free(result_ptr);
        IAS_LOG_ERROR("Creating the parallel correlator");
        return ERROR;
    }

    /* Process each band. */
    for (curr = 0; curr < number_bands; curr++)
    {
        /* Initialize needed values to 0. */
        num_correlated = 0;
        mean_offset_line = 0.0;
        mean_offset_sample = 0.0;

        /* Get the band numbers. */
        reference_band_index = reference_band_indexes[curr];
        reference_band_number = ias_sat_attr_convert_band_index_to_number(
            reference_band_index);
        if (reference_band_number == ERROR) 
        {
            IAS_LOG_ERROR("Invalid band number for reference band index %d", 
                reference_band_index);
            ias_l1g_close_image(ref_img_ptr);
            ias_l1g_close_image(srch_img_ptr);
            free(result_ptr);
            free(gcp);
            fclose(output);
            ias_math_close_parallel_correlator(correlator_ptr);
            return ERROR;
        }

        search_band_index = search_band_indexes[curr];
        search_band_number = ias_sat_attr_convert_band_index_to_number(
            search_band_index);
        if (search_band_number == ERROR) 
        {
            IAS_LOG_ERROR("Invalid band number for search band index %d", 
                search_band_index);
            ias_l1g_close_image(ref_img_ptr);
            ias_l1g_close_image(srch_img_ptr);
            free(result_ptr);
            free(gcp);
            fclose(output);
            ias_math_close_parallel_correlator(correlator_ptr);
            return ERROR;
        }

        IAS_LOG_INFO("Correlating reference band %d and search band %d", 
            reference_band_number, search_band_number);

        /* Open the current L1G reference image band for window extraction. */
        ref_l1g_band = ias_l1g_open_band(ref_img_ptr, reference_band_number,
                                   &ref_dtype[curr], &one,
                                   &ref_number_lines[curr], 
                                   &ref_number_samples[curr]);
        if (ref_l1g_band == NULL)
        {
            IAS_LOG_ERROR("Opening reference band %d", reference_band_number);
            ias_l1g_close_image(ref_img_ptr);
            ias_l1g_close_image(srch_img_ptr);
            free(result_ptr);
            free(gcp);
            fclose(output);
            ias_math_close_parallel_correlator(correlator_ptr);
            return ERROR;
        }

        IAS_LOG_INFO("Reference %s NL %d NS %d", ref_img_name, 
            ref_number_lines[curr], ref_number_samples[curr]);

        /* Open the current L1G search image band for window extraction. */
        srch_l1g_band = ias_l1g_open_band(srch_img_ptr, search_band_number,
                                   &srch_dtype[curr], &one,
                                   &srch_number_lines[curr], 
                                   &srch_number_samples[curr]);
        if (srch_l1g_band == NULL)
        {
            IAS_LOG_ERROR("Opening search band %d", search_band_number);
            ias_l1g_close_band(ref_l1g_band);
            ias_l1g_close_image(ref_img_ptr);
            ias_l1g_close_image(srch_img_ptr);
            free(result_ptr);
            free(gcp);
            fclose(output);
            ias_math_close_parallel_correlator(correlator_ptr);
            return ERROR;
        }

        IAS_LOG_INFO("Search %s NL %d NS %d", srch_img_name, 
            srch_number_lines[curr], srch_number_samples[curr]);

        /* Set the GCPLib pointer to the appropriate GCPs. */
        projection_distance =
            ref_band_metadata[reference_band_index].projection_distance_x;
        if (IS_PROJ_DIST(projection_distance, 15.0))
        {
            ref_gcplib = ref_gcp15;
            srch_gcplib = srch_gcp15;
            fract_offset = fract_offset_15;
        }
        else
        if (IS_PROJ_DIST(projection_distance, 30.0))
        {
            ref_gcplib = ref_gcp30;
            srch_gcplib = srch_gcp30;
            fract_offset = fract_offset_30;
        }
        else
        {
            IAS_LOG_ERROR("Invalid projection distance %f for reference "
                "band number %d", projection_distance, reference_band_number);
            ias_l1g_close_band(ref_l1g_band);
            ias_l1g_close_image(ref_img_ptr);
            ias_l1g_close_band(srch_l1g_band);
            ias_l1g_close_image(srch_img_ptr);
            free(result_ptr);
            free(gcp);
            fclose(output);
            ias_math_close_parallel_correlator(correlator_ptr);
            return ERROR;
        }

        /* Check the data type for both the search and reference images.
           Only valid 1G output types are BYTE and I2. */
        if ((ref_dtype[curr] != IAS_BYTE) && (ref_dtype[curr] != IAS_I2) &&
            (ref_dtype[curr] != IAS_UI2))
        {
            IAS_LOG_ERROR("Invalid L1G reference image data type");
            ias_l1g_close_band(ref_l1g_band);
            ias_l1g_close_image(ref_img_ptr);
            ias_l1g_close_band(srch_l1g_band);
            ias_l1g_close_image(srch_img_ptr);
            free(result_ptr);
            free(gcp);
            fclose(output);
            ias_math_close_parallel_correlator(correlator_ptr);
            return ERROR;
        }

        if ((srch_dtype[curr] != IAS_BYTE) && (srch_dtype[curr] != IAS_I2) &&
            (srch_dtype[curr] != IAS_UI2))
        {
            IAS_LOG_ERROR("Invalid L1G search image data type");
            ias_l1g_close_band(ref_l1g_band);
            ias_l1g_close_image(ref_img_ptr);
            ias_l1g_close_band(srch_l1g_band);
            ias_l1g_close_image(srch_img_ptr);
            free(result_ptr);
            free(gcp);
            fclose(output);
            ias_math_close_parallel_correlator(correlator_ptr);
            return ERROR;
        }

        /* Extract the windows and correlate the images. */
        status = process_gcp(ref_gcplib, srch_gcplib, number_of_gcps, 
                     ref_l1g_band, srch_l1g_band, reference_band_number,
                     search_band_number, ref_dtype[curr], srch_dtype[curr], 
                     ref_number_lines[curr], ref_number_samples[curr], 
                     srch_number_lines[curr], srch_number_samples[curr],
                     min_corr, max_disp, fit_method, gcp, fract_offset, 
                     abs_corr_coeff, invalid_threshold, valid_image_max, 
                     valid_image_min, correlator_ptr, result_ptr);

        if (status == ERROR)
        {
            IAS_LOG_ERROR("Correlating reference %s and search %s", 
                ref_img_name, srch_img_name);
            ias_l1g_close_band(ref_l1g_band);
            ias_l1g_close_image(ref_img_ptr);
            ias_l1g_close_band(srch_l1g_band);
            ias_l1g_close_image(srch_img_ptr);
            free(result_ptr);
            free(gcp);
            fclose(output);
            ias_math_close_parallel_correlator(correlator_ptr);
            return ERROR;
        }

        /* Close the current band in the reference image. */
        if (ias_l1g_close_band(ref_l1g_band) != SUCCESS)
        {
            IAS_LOG_ERROR("Closing reference band %d", reference_band_number);
            ias_l1g_close_image(ref_img_ptr);
            ias_l1g_close_band(srch_l1g_band);
            ias_l1g_close_image(srch_img_ptr);
            free(result_ptr);
            free(gcp);
            fclose(output);
            ias_math_close_parallel_correlator(correlator_ptr);
            return ERROR;
        }

        /* Close the current band in the search image. */
        if (ias_l1g_close_band(srch_l1g_band) != SUCCESS)
        {
            IAS_LOG_ERROR("Closing search band %d", search_band_number);
            ias_l1g_close_image(ref_img_ptr);
            ias_l1g_close_image(srch_img_ptr);
            free(result_ptr);
            free(gcp);
            fclose(output);
            ias_math_close_parallel_correlator(correlator_ptr);
            return ERROR;
        }

        /* Write the correlation results to the output file. */
        for (i = 0; i < number_of_gcps; i++)
        { 
            if (gcp[i].accept_flag)
            {
                /* Write a single line for the report.  The 0's are for
                   consistency with the B2B report which displays SCA numbers
                   there.  Tdist reads both reports using a single format. */ 
                if (fprintf(output,
                        "%8s   %12.6lf  %12.6lf  %11.6lf  %11.6lf  %11.6lf  "
                        "%12.6lf  %12.6lf  %11.6lf  %11.6lf  %3d  %11.6lf  "
                        "%3d  %8d             0       0  %s  %s\n",
                        ref_gcplib[i].point_id, 
                        ref_gcplib[i].reference_line,
                        ref_gcplib[i].reference_sample, 
                        ref_gcplib[i].latitude,
                        ref_gcplib[i].longitude, 
                        ref_gcplib[i].elevation,
                        srch_gcplib[i].reference_line,
                        srch_gcplib[i].reference_sample,
                        gcp[i].fit_line_offset, 
                        gcp[i].fit_sample_offset,
                        gcp[i].accept_flag, 
                        gcp[i].correlation_coefficient, 
                        reference_band_number, 
                        search_band_number, 
                        ref_img_name,
                        srch_img_name) < 0)
                {
                    IAS_LOG_ERROR("Writing correlation results to data file");
                    ias_l1g_close_image(ref_img_ptr);
                    ias_l1g_close_image(srch_img_ptr);
                    free(result_ptr);
                    free(gcp);
                    fclose(output);
                    ias_math_close_parallel_correlator(correlator_ptr);
                    return ERROR;
                }

                /* Add the fit_offset to the mean_offsets if the point
                   correlated. */
                num_correlated++;
                mean_offset_line += gcp[i].fit_line_offset;
                mean_offset_sample += gcp[i].fit_sample_offset;
            }
        }

        /* Output the number of GCPs that correlated. */
        IAS_LOG_INFO("%d of %d GCPs correlated successfully",
                     num_correlated, number_of_gcps);

        /* If any values correlated then calculate the mean and std dev. */
        if (num_correlated > 1)
        {
            mean_offset_line /= (double)num_correlated;
            mean_offset_sample /= (double)num_correlated;
            IAS_LOG_INFO("Mean fit offset is %f, %f, (line, sample).",
                         mean_offset_line, mean_offset_sample);

            /* Malloc space to hold the correlated fit_offsets for the 
               standard deviation calculation. */
            offset_line = malloc(num_correlated * sizeof(double));
            if (offset_line == NULL)
            {
                IAS_LOG_ERROR("Allocating line offset array buffer");
                ias_l1g_close_image(ref_img_ptr);
                ias_l1g_close_image(srch_img_ptr);
                free(result_ptr);
                free(gcp);
                fclose(output);
                ias_math_close_parallel_correlator(correlator_ptr);
                return ERROR;
            }
            offset_sample = malloc(num_correlated * sizeof(double));
            if (offset_sample == NULL)
            {
                IAS_LOG_ERROR("Allocating sample offset array buffer");
                ias_l1g_close_image(ref_img_ptr);
                ias_l1g_close_image(srch_img_ptr);
                free(offset_line);
                free(result_ptr);
                free(gcp);
                fclose(output);
                ias_math_close_parallel_correlator(correlator_ptr);
                return ERROR;
            }

            /* Output the standard deviation for the current band 
               combinations. */
            for (i = 0, j = 0; i < number_of_gcps; i++)
            {
                if (gcp[i].accept_flag)
                {
                    offset_line[j] = gcp[i].fit_line_offset;
                    offset_sample[j] = gcp[i].fit_sample_offset;
                    j++;
                }
            }
            
            status = ias_math_compute_stdev(offset_line, num_correlated, 
                mean_offset_line, &stddev_offset_line);
            if (status == ERROR)
            {
                IAS_LOG_ERROR("Calculating std deviation of line offsets");
                ias_l1g_close_image(ref_img_ptr);
                ias_l1g_close_image(srch_img_ptr);
                free(offset_line);
                free(offset_sample);
                free(result_ptr);
                free(gcp);
                fclose(output);
                ias_math_close_parallel_correlator(correlator_ptr);
                return ERROR;
            }
            status = ias_math_compute_stdev(offset_sample, num_correlated, 
                mean_offset_sample, &stddev_offset_sample);
            if (status == ERROR)
            {
                IAS_LOG_ERROR("Calculating std deviation of sample offsets");
                ias_l1g_close_image(ref_img_ptr);
                ias_l1g_close_image(srch_img_ptr);
                free(offset_line);
                free(offset_sample);
                free(result_ptr);
                free(gcp);
                fclose(output);
                ias_math_close_parallel_correlator(correlator_ptr);
                return ERROR;

            }
            IAS_LOG_INFO("Standard deviation fit_offset is: %f, %f "
                         "(line, sample).", stddev_offset_line,
                         stddev_offset_sample);
            /* Free the offset pointers for the next set. */
            free(offset_line);
            free(offset_sample);
        }

    } /* for current band */

    /* Allow the parallel correlator to finish. */
    if (ias_math_close_parallel_correlator(correlator_ptr) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing the parallel correlator");
        ias_l1g_close_image(ref_img_ptr);
        ias_l1g_close_image(srch_img_ptr);
        free(result_ptr);
        free(gcp);
        fclose(output);
        return ERROR;
    }

    /* Close the L1G images. */
    if (ias_l1g_close_image(srch_img_ptr) == ERROR)
    {
        IAS_LOG_ERROR("Closing search image %s", srch_img_name);
        ias_l1g_close_image(ref_img_ptr);
        free(result_ptr);
        free(gcp);
        fclose(output);
        return ERROR;
    }
    if (ias_l1g_close_image(ref_img_ptr) == ERROR)
    {
        IAS_LOG_ERROR("Closing reference image %s", ref_img_name);
        free(result_ptr);
        free(gcp);
        fclose(output);
        return ERROR;
    }

    /* Free the memory for the correlation results. */
    free(result_ptr);

    /* Free the IAS_GCP_RESULTS pointer and close the output file. */
    free(gcp);
    status = fclose(output);
    if (status != 0)
    {
        IAS_LOG_ERROR("Closing output report file");
        return ERROR;
    }

    return SUCCESS;
}
