/******************************************************************************
NAME:           i2i_det_tiepts

PURPOSE:
Determine the I2I tiepoints for a specified band resolution.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
This routine will convert the search corners to reference line/sample space.
Then the routine will determine the overlapping area between both the
reference and search images in reference line/sample space.  This overlapping
area is gridded into regular-spaced tiepoints.  The offset between the search
and reference images must be subtracted from the ref GCP values to get the
corresponding search GCP.

This routine will make sure that the chip windows for the determined tiepoints
don't fall outside the image.  Thus for any given tiepoint, its chip window
(according to the chip_size) will be completely inside the image for the
specified band resolution.

The corner points for the tiepoints are found in the following manner:
  UL(x) = min.x + Chip_Size/2       UR(x) = max.x - Chip_Size/2
  UL(y) = min.y + Chip_Size/2       UR(y) = min.y + Chip_Size/2
  LL(x) = min.x + Chip_Size/2       LR(x) = max.x - Chip_Size/2
  LL(y) = max.y - Chip_Size/2       LR(y) = max.y - Chip_Size/2

******************************************************************************/
#include "i2i_char.h"
#include "ias_logging.h"      /* Logging library                             */
#include "ias_geo.h"

int i2i_det_tiepts 
(
    const IAS_L1G_BAND_METADATA *ref_band_metadata, /* I: Reference image 
                                       metadata for the desired band         */
    const IAS_L1G_BAND_METADATA *srch_band_metadata, /* I: Search image 
                                       metadata for the desired band         */
    int ref_number_lines,        /* I: Number of reference lines             */
    int ref_number_samples,      /* I: Number of reference samples           */
    int srch_number_lines,       /* I: Number of search lines                */
    int srch_number_samples,     /* I: Number of search samples              */
    int number_line_points,      /* I: Number of line tiepoints to create    */
    int number_sample_points,    /* I: Number of sample tiepoints to create  */
    struct IAS_LNG_LS chip_size, /* I: Chip size of the tiepoints to create  */
    IAS_GCP_RECORD *ref_gcps,    /* O: Array of reference GCP record 
                                       tiepoints in row-major format         */
    IAS_GCP_RECORD *srch_gcps,   /* O: Array of search GCP record tiepoints 
                                       in row-major format                   */
    IAS_DBL_XY *fract_offsetPTR  /* O: Pointer to two element structure which
                                       contains the fraction parts of the
                                       offsets of the search image from the
                                       reference image.                      */
)
{
    int i;                 /* Looping variable                               */
    int j;                 /* Looping variable                               */
    int count;             /* Number of GCPs                                 */

    int min_x;             /* Minimum x tiepoint value within the image      */
    int max_x;             /* Maximum x tiepoint value within the image      */
    int min_y;             /* Minimum y tiepoint value within the image      */
    int max_y;             /* Maximum y tiepoint value within the image      */
    int x_spacing;         /* Line spacing between chips                     */
    int y_spacing;         /* Sample spacing between chips                   */

    double rpoly_x[4];     /* Sample-to-projection coefficients for ref      */
    double rpoly_y[4];     /* Line-to-projection coefficients for ref        */
    double spoly_x[4];     /* Sample-to-projection coefficients for search   */
    double spoly_y[4];     /* Line-to-projection coefficients for search     */

    int degree;            /* Degree of the polynomials                      */
    int status;            /* Function return status                         */

    IAS_DBL_XY ref_ul;     /* Ref UL corner in ref line/sample space         */
    IAS_DBL_XY ref_ur;     /* Ref UR corner in ref line/sample space         */
    IAS_DBL_XY ref_ll;     /* Ref LL corner in ref line/sample space         */
    IAS_DBL_XY ref_lr;     /* Ref LR corner in ref line/sample space         */
    IAS_DBL_XY srch_ul;    /* Srch UL corner in ref line/sample space        */
    IAS_DBL_XY srch_ur;    /* Srch UR corner in ref line/sample space        */
    IAS_DBL_XY srch_ll;    /* Srch LL corner in ref line/sample space        */
    IAS_DBL_XY srch_lr;    /* Srch LR corner in ref line/sample space        */

    IAS_LNG_XY offset;     /* Offset between ref and srch line/samp values   */

    double corners[2][4];  /* latitude and longitude for the four corners    */
    FILE *fp = NULL;       /* File pointer for report "points.dat"           */

    int input_line;        /* Input line location for GCP location           */
    int input_sample;      /* Input sample location for GCP location         */


#ifdef PRINT_GCPS

    fp = fopen("points.dat", "w");
    if (fp == NULL)
    {
        IAS_LOG_ERROR("Creating points.dat file");
        return ERROR;
    }
#endif

    /* If only one line or sample point then return with an ERROR. */
    if (number_line_points <= 1)
    {
        IAS_LOG_ERROR("This routine requires more than one line point");
        if (fp)
            fclose(fp);
        return ERROR;
    }
    if (number_sample_points <= 1)
    {
        IAS_LOG_ERROR("This routine requires more than one sample per line");
        if (fp)
           fclose(fp);
        return ERROR;
    }

    /* Set up the corners structure needed for the compute_poly call. */
    corners[1][0] = ref_band_metadata->upper_left_x;
    corners[1][1] = ref_band_metadata->upper_right_x;
    corners[1][2] = ref_band_metadata->lower_right_x;
    corners[1][3] = ref_band_metadata->lower_left_x;
    corners[0][0] = ref_band_metadata->upper_left_y;
    corners[0][1] = ref_band_metadata->upper_right_y;
    corners[0][2] = ref_band_metadata->lower_right_y;
    corners[0][3] = ref_band_metadata->lower_left_y;

    /* Get the ref and srch polynomials for the eval_poly call. */
    degree = 1;
    status = ias_geo_compute_proj2ls_poly(ref_number_samples, ref_number_lines, 
        corners, rpoly_x, rpoly_y);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Calculating reference poly_x and poly_y");
        if (fp)
            fclose(fp);
        return ERROR;
    }

    /* Set up the corners structure needed for the compute_poly call. */
    corners[1][0] = srch_band_metadata->upper_left_x;
    corners[1][1] = srch_band_metadata->upper_right_x;
    corners[1][2] = srch_band_metadata->lower_right_x;
    corners[1][3] = srch_band_metadata->lower_left_x;
    corners[0][0] = srch_band_metadata->upper_left_y;
    corners[0][1] = srch_band_metadata->upper_right_y;
    corners[0][2] = srch_band_metadata->lower_right_y;
    corners[0][3] = srch_band_metadata->lower_left_y;

    status = ias_geo_compute_proj2ls_poly(srch_number_samples,
        srch_number_lines, corners, spoly_x, spoly_y);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating search poly_x and poly_y");
        if (fp)
           fclose(fp);
        return ERROR;
    }

    /* Determine the search corners in reference line/sample space.  Note: The 
       offset between the reference and search images in line/samp is the
       line/samp value of the ul point of the search image in reference 
       space. */
    ias_math_eval_poly(degree, rpoly_y, srch_band_metadata->upper_left_x,
                       srch_band_metadata->upper_left_y, &srch_ul.y);
    ias_math_eval_poly(degree, rpoly_x, srch_band_metadata->upper_left_x,
                       srch_band_metadata->upper_left_y, &srch_ul.x);
    offset.y = (int)(srch_ul.y);  /* These are the X and Y offsets between */
    offset.x = (int)(srch_ul.x);  /* the reference and search images.      */

    /* Save fractional part of offsets. */
    fract_offsetPTR->x = srch_ul.x - (double)offset.x;
    fract_offsetPTR->y = srch_ul.y - (double)offset.y;

    ias_math_eval_poly(degree, rpoly_y, srch_band_metadata->lower_left_x,
                       srch_band_metadata->lower_left_y, &srch_ll.y);
    ias_math_eval_poly(degree, rpoly_x, srch_band_metadata->lower_left_x,
                       srch_band_metadata->lower_left_y, &srch_ll.x);
    ias_math_eval_poly(degree, rpoly_y, srch_band_metadata->upper_right_x,
                       srch_band_metadata->upper_right_y, &srch_ur.y);
    ias_math_eval_poly(degree, rpoly_x, srch_band_metadata->upper_right_x,
                       srch_band_metadata->upper_right_y, &srch_ur.x);
    ias_math_eval_poly(degree, rpoly_y, srch_band_metadata->lower_right_x,
                       srch_band_metadata->lower_right_y, &srch_lr.y);
    ias_math_eval_poly(degree, rpoly_x, srch_band_metadata->lower_right_x,
                       srch_band_metadata->lower_right_y, &srch_lr.x);

    /* Determine the reference corners in reference line/sample space. */
   ias_math_eval_poly(degree, rpoly_y, ref_band_metadata->upper_left_x,
                       ref_band_metadata->upper_left_y, &ref_ul.y);
    ias_math_eval_poly(degree, rpoly_x, ref_band_metadata->upper_left_x,
                       ref_band_metadata->upper_left_y, &ref_ul.x);
    ias_math_eval_poly(degree, rpoly_y, ref_band_metadata->lower_left_x,
                       ref_band_metadata->lower_left_y, &ref_ll.y);
    ias_math_eval_poly(degree, rpoly_x, ref_band_metadata->lower_left_x,
                       ref_band_metadata->lower_left_y, &ref_ll.x);
    ias_math_eval_poly(degree, rpoly_y, ref_band_metadata->upper_right_x,
                       ref_band_metadata->upper_right_y, &ref_ur.y);
    ias_math_eval_poly(degree, rpoly_x, ref_band_metadata->upper_right_x,
                       ref_band_metadata->upper_right_y, &ref_ur.x);
    ias_math_eval_poly(degree, rpoly_y, ref_band_metadata->lower_right_x,
                       ref_band_metadata->lower_right_y, &ref_lr.y);
    ias_math_eval_poly(degree, rpoly_x, ref_band_metadata->lower_right_x,
                       ref_band_metadata->lower_right_y, &ref_lr.x);

    /* Determine the overlapping image space between the search and reference
       images. */
    if (ref_ul.x > srch_ul.x)
        min_x = (int)ref_ul.x;
    else
        min_x = (int)srch_ul.x;

    if (ref_ur.x < srch_ur.x)
        max_x = (int)ref_ur.x;
    else
        max_x = (int)srch_ur.x;

    if (ref_ul.y > srch_ul.y)
        min_y = (int)ref_ul.y;
    else
        min_y = (int)srch_ul.y;

    if (ref_ll.y < srch_ll.y)
        max_y = (int)ref_ll.y;
    else
        max_y = (int)srch_ll.y;

    /* Find the min and max corner point values for the chips inside of the
     * overlapping image.  These are the four corners which the grid will
     * be based off of. */
    min_x += chip_size.samp / 2 + 1;
    max_x -= chip_size.samp / 2 - 1;
    min_y += chip_size.line / 2 + 1;
    max_y -= chip_size.line / 2 - 1;

    /* Determine the line and sample spacings for all other points between
     * the min and max points. */
    x_spacing = (max_x - min_x) / (number_sample_points - 1);
    y_spacing = (max_y - min_y) / (number_line_points - 1);

    IAS_LOG_INFO("Offsets %d %d", offset.x, offset.y);

    /* Fill in the complete GCP structures. */
    for (i = 0, count = 0; i < number_line_points; i++)
    {
        /* Determine the reference line GCP values. */
        if (i == number_line_points - 1)
            input_line = max_y;
        else
            input_line = min_y + y_spacing * i;

        for (j = 0; j < number_sample_points; j++, count++)
        {
            /* Determine the reference sample GCP values. */
            if (j == number_sample_points - 1)
                input_sample = max_x;
            else
                input_sample = min_x + x_spacing * j;

            /* Fill in the reference GCPS.  */
#ifdef PRINT_GCPS
            if (fprintf(fp, "%d ", count) < 0)
            {
                IAS_LOG_WARNING("Failure printing GCP count to file");
            }
#endif

            status = sprintf (ref_gcps[count].point_id, "%d", count);

            ref_gcps[count].reference_line = input_line;
            ref_gcps[count].reference_sample = input_sample;
            ref_gcps[count].pixel_size_x = 
                ref_band_metadata->projection_distance_x;
            ref_gcps[count].pixel_size_y = 
                ref_band_metadata->projection_distance_y;
            ref_gcps[count].chip_size_lines = chip_size.line;
            ref_gcps[count].chip_size_samples = chip_size.samp;
            ref_gcps[count].latitude = 0.0;
            ref_gcps[count].longitude = 0.0;
            ref_gcps[count].elevation = 0.0;

#ifdef PRINT_GCPS
            if (fprintf(fp, "Ref %f %f %f %f ", 
                ref_gcps[count].reference_line,
                ref_gcps[count].reference_sample, 
                ref_gcps[count].chip_size_lines, 
                ref_gcps[count].chip_size_samples) < 0)
            {
                IAS_LOG_WARNING("Failure printing search GCP %s to file", 
                    ref_gcps[count].point_id);
            }
#endif

            /* Fill in the search GCPs. */
            status = sprintf(srch_gcps[count].point_id, "%d", count);

            srch_gcps[count].reference_line = input_line - offset.y;
            srch_gcps[count].reference_sample = input_sample - offset.x;
            srch_gcps[count].pixel_size_x = 
                srch_band_metadata->projection_distance_x;
            srch_gcps[count].pixel_size_y = 
                srch_band_metadata->projection_distance_y;
            srch_gcps[count].chip_size_lines = chip_size.line;
            srch_gcps[count].chip_size_samples = chip_size.samp;
            srch_gcps[count].latitude = 0.0;
            srch_gcps[count].longitude = 0.0;
            srch_gcps[count].elevation = 0.0;

#ifdef PRINT_GCPS
            if (fprintf(fp, "Sear %f %f %f %f\n", 
                srch_gcps[count].reference_line,
                srch_gcps[count].reference_sample,
                srch_gcps[count].chip_size_lines,
                srch_gcps[count].chip_size_samples) < 0)
            {
                IAS_LOG_WARNING("Failure printing search GCP %s to file", 
                    srch_gcps[count].point_id);
            }
#endif
        }
    }

#ifdef PRINT_GCPS
    fclose(fp);
#endif

    return SUCCESS;
}
