/******************************************************************************
NAME:       process_gcp

PURPOSE:  Process all the GCPs (extract the image window for both the
          reference and search images and perform the correlation for each
          point).

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
Correlation routines want things in sample, line order, so the fit_offset
returned are horizontal (sample) and then vertical (line) offsets.  GCP data
file contains fit_offset in line, sample order.

******************************************************************************/
#include <stdlib.h>
#include "i2i_char.h"
#include "ias_geo.h"              /* Geographic library                      */
#include "ias_logging.h"          /* Logging library                         */

int process_gcp 
(
    const IAS_GCP_RECORD *ref_gcp_lib,  /* I: List of reference GCPs         */
    const IAS_GCP_RECORD *srch_gcp_lib, /* I: List of search GCPs            */
    int number_of_gcps,           /* I: Number of GCPs                       */
    L1G_BAND_IO *ref_l1g_band,    /* I: Reference 1G band pointer            */
    L1G_BAND_IO *srch_l1g_band,   /* I: Search 1G band pointer               */
    int reference_band_number,    /* I: Reference band number to process     */
    int search_band_number,       /* I: Search band number to process        */
    IAS_DATA_TYPE ref_dtype,      /* I: Data type of reference image         */
    IAS_DATA_TYPE srch_dtype,     /* I: Data type of search image            */
    int ref_number_lines,         /* I: Number of lines in ref image         */
    int ref_number_samples,       /* I: Number of samples in ref image       */
    int srch_number_lines,        /* I: Number of lines in srch image        */
    int srch_number_samples,      /* I: Number of samples in srch image      */
    float min_corr,               /* I: Minimum correlation strength         */
    float max_disp,               /* I: Maximum correlation displacement     */
    IAS_CORRELATION_FIT_TYPE fit_method, /* I: Correlation result fit 
                                        method                               */
    IAS_GCP_RESULTS *gcp,         /* O: GCP data structures (1 per pt) 
                                        for the correlation results          */
    IAS_DBL_XY fract_offset,      /* I: Structure containing fractional 
                                        pixel mean offsets of the search 
                                        image from the reference image       */
    int abs_corr_coeff,           /* I: flag to use absolute correlation
                                        coeffs for grey correlation          */
    float invalid_threshold,      /* I: Invalid threshold percentage for 
                                        buffer                               */
    float valid_image_max,        /* I: DNs above this are invalid           */
    float valid_image_min,        /* I: DNs below this are invalid           */
    IAS_PARALLEL_CORRELATOR_TYPE *correlator_ptr, /* I: Pointer to
                                        the parallel correlator structure    */
    IAS_CORRELATION_RESULT_TYPE *result_ptr /* O: Pointer to the
                                        correlation results                  */
)
{
    int curgcp;                   /* GCP counter                             */
    int win_size[2];              /* chip size to extract (line, samp)       */
    int ref_win_size[2];          /* chip size to extract (samp, line)       */
    int srch_win_size[2];         /* chip size to extract (samp, line)       */
    float nom_off[2]={0.0, 0.0};  /* nominal correlation offset              */
    int status;                   /* function return status                  */
    int rstatus;                  /* function return status                  */
    int sstatus;                  /* function return status                  */
    int sca = 0;                  /* Only SCA combined imagery to process    */
    long count;                   /* Count of GCPs falling in reference      */
    int upper_line;              /* The upper line in the image for the chip */
    int left_sample;            /* The left sample in the image for the chip */

    /* These are static so the buffers can be managed and remembered 
       during multiple calls to prevent memory leak.                         */
    static float *srch_buf;       /* float search correlation window         */
    static float *ref_buf;        /* float reference correlation window      */
    static int need_buffers = 1;  /* flag to indicate buffers are needed 
                                     for the reference and search chip       */

    /* Initialize the results. */
    for (curgcp = 0; curgcp < number_of_gcps; curgcp++)
    {
        result_ptr[curgcp].valid = FALSE;
        result_ptr[curgcp].fit_offset[0] = 0.0;
        result_ptr[curgcp].fit_offset[1] = 0.0;
    }

    /* Process each GCP. */
    count = 0;
    for (curgcp = 0; curgcp < number_of_gcps; curgcp++)
    {
        /* Create a status message for bands and GCPs, but only do every
           50th one to moderate the amount of output generated. */
        if ((curgcp % 50) == 0)
        {
            IAS_LOG_INFO("Processing reference band %d search band %d GCP %d", 
                reference_band_number, search_band_number, curgcp);
        }

        /* Get buffers for the chips if needed. */
        if (need_buffers)
        {
            status = ias_math_get_corr_chip_buffers(correlator_ptr, &ref_buf,
                                               &srch_buf);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Retrieving chip buffers");
                return ERROR;
            }

            need_buffers = 0;
        }

        win_size[0] = (int)ref_gcp_lib[curgcp].chip_size_lines;
        win_size[1] = (int)ref_gcp_lib[curgcp].chip_size_samples;
        status = ias_geo_extract_window(ref_number_lines, ref_number_samples, 
            sca, ref_dtype,
            ref_gcp_lib[curgcp].reference_line,
            ref_gcp_lib[curgcp].reference_sample,
            win_size, ref_l1g_band, ref_buf,
            &upper_line, &left_sample);
        if (status == OUTSIDE)
        {
            IAS_LOG_WARNING("GCP %s is outside reference image",
                            ref_gcp_lib[curgcp].point_id);
            continue;
        }
        if (status == ERROR)
        {
            IAS_LOG_ERROR("Extracting reference window for GCP %s",
                          ref_gcp_lib[curgcp].point_id);
            free(srch_buf);
            free(ref_buf);
            return ERROR;
        }

        /* Extract the window from the 1G search image.  The extract_window
           function takes care of converting the window to a buffer of float 
           data.  */
        win_size[0] = (int) srch_gcp_lib[curgcp].chip_size_lines;
        win_size[1] = (int) srch_gcp_lib[curgcp].chip_size_samples;
        status = ias_geo_extract_window(srch_number_lines, srch_number_samples, 
            sca, srch_dtype, 
            srch_gcp_lib[curgcp].reference_line,
            srch_gcp_lib[curgcp].reference_sample,
            win_size, srch_l1g_band, srch_buf,
            &upper_line, &left_sample);
        if (status == OUTSIDE)
        {
            IAS_LOG_WARNING("GCP %s is outside search image",
                            srch_gcp_lib[curgcp].point_id);
            continue;
        }
        else if (status == ERROR)
        {
            IAS_LOG_ERROR("Extracting search window for GCP %s",
                          srch_gcp_lib[curgcp].point_id);
            free(srch_buf);
            free(ref_buf);
            return ERROR;
        }

        /* If the reference or search buffers are greater than
           <invalid_threshold>% out of range data, don't correlate them. */
        sstatus = ias_math_check_pixels_in_range(srch_buf,
                (int)srch_gcp_lib[curgcp].chip_size_samples
                *(int)srch_gcp_lib[curgcp].chip_size_lines,
                invalid_threshold, valid_image_max, valid_image_min);
        rstatus = ias_math_check_pixels_in_range(ref_buf,
                (int)ref_gcp_lib[curgcp].chip_size_samples
                *(int)ref_gcp_lib[curgcp].chip_size_lines,
                invalid_threshold, valid_image_max, valid_image_min);

        if (sstatus && rstatus)
        {
            count++;

            /* Grey correlation expects the reference and search windows sizes
               to be in samp, line order, so we need to flip our window size 
               values. */
            ref_win_size[0] = (int)ref_gcp_lib[curgcp].chip_size_samples; 
            ref_win_size[1] = (int)ref_gcp_lib[curgcp].chip_size_lines; 
            srch_win_size[0] = (int)srch_gcp_lib[curgcp].chip_size_samples; 
            srch_win_size[1] = (int)srch_gcp_lib[curgcp].chip_size_lines; 

            /* Do the correlation. */
            status = ias_math_submit_chip_to_corr(correlator_ptr, curgcp, 
                srch_buf, ref_buf, srch_win_size, ref_win_size, 
                min_corr, fit_method, max_disp, nom_off, abs_corr_coeff);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Grey correlation error for GCP point %d",
                              curgcp + 1);
                ias_math_parallel_correlator_wait_for_results(correlator_ptr);
                return ERROR;
            }

            /* Flag that the next chips will need new buffers. */
            need_buffers = 1;
        }
        else
        {
            result_ptr[curgcp].valid = FALSE;
        }
    }

    IAS_LOG_INFO("Number of GCPs falling in reference (excluding fill): %ld",
        count);

    /* Wait for the parallel correlator to finish. */
    status = ias_math_parallel_correlator_wait_for_results(correlator_ptr);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Waiting for parallel correlator results");
        return ERROR;    
    }

    /* Extract the correlation results returned by the parallel correlator. */
    for (curgcp = 0; curgcp < number_of_gcps; curgcp++)
    {
        gcp[curgcp].reference_line = srch_gcp_lib[curgcp].reference_line;
        gcp[curgcp].reference_sample = srch_gcp_lib[curgcp].reference_sample;
        gcp[curgcp].predicted_search_line =
            srch_gcp_lib[curgcp].reference_line;
        gcp[curgcp].predicted_search_sample =
            srch_gcp_lib[curgcp].reference_sample;

        /* Check the returned flag to see if this correlation succeeded. */
        if (result_ptr[curgcp].valid)
        {
            /* Correlation succeeded. */
            gcp[curgcp].accept_flag = TRUE;
            gcp[curgcp].correlation_coefficient = result_ptr[curgcp].strength;
            gcp[curgcp].fit_line_offset = result_ptr[curgcp].fit_offset[1] 
                + fract_offset.y;
            gcp[curgcp].fit_sample_offset = result_ptr[curgcp].fit_offset[0]
                + fract_offset.x;
        }
        else
        {
            /* There was a correlation failure. */
            gcp[curgcp].accept_flag = FALSE;
            gcp[curgcp].correlation_coefficient = 0.0;
            gcp[curgcp].fit_line_offset = 0.0;
            gcp[curgcp].fit_sample_offset = 0.0;
        }
    }

    return SUCCESS;
}
