/******************************************************************************
NAME:           read_gcps

PURPOSE:
Read GCP record file for the specified band resolutions.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
This routine reads ALL of the gcps in the GCP library file.
It is assumed the GCP library file contains the proj x/y values for the GCPs.
Space is malloced for the gcp15 and gcp30 pointers.

******************************************************************************/
#include <stdlib.h>
#include "i2i_char.h"
#include "ias_gcp.h"             /* GCP library                              */
#include "ias_logging.h"         /* Logging library                          */
#include "gctp.h"

int read_gcps 
(
    const char *gcp_filename,    /* I: GCP file name                         */
    const int *reference_band_indexes, /* I: Reference band indexes          */
    const int *search_band_indexes, /* I: Search band indexes                */
    int number_bands,            /* I: Number of band indexes                */
    const IAS_L1G_BAND_METADATA *ref_band_metadata, /* I: Reference image
                                       band metadata array                   */
    const IAS_L1G_BAND_METADATA *srch_band_metadata, /* I: Search image band 
                                       metadata array                        */
    const IAS_L1G_FILE_METADATA *ref_file_metadata, /* I: Reference image file 
                                       metadata                              */
    const IAS_L1G_FILE_METADATA *srch_file_metadata, /* I: Search image file 
                                       metadata                              */
    const int *ref_number_lines, /* I: Number of lines in reference image    */
    const int *ref_number_samples,/* I: Number of samples in reference image */
    const int *srch_number_lines, /* I: Number of lines in search image      */
    const int *srch_number_samples, /* I: Number of samples in search image  */
    int *number_of_gcps,         /* O: Number of GCPs                        */
    IAS_GCP_RECORD **ref_gcp15,  /* O: Ref GCPs for 15m band resolution      */
    IAS_GCP_RECORD **srch_gcp15, /* O: Search GCPs for 15m band resolution   */
    IAS_GCP_RECORD **ref_gcp30,  /* O: Ref GCPs for 30m band resolution      */
    IAS_GCP_RECORD **srch_gcp30  /* O: Search GCPs for 30m band resolution   */
)
{
    int i;                       /* Loop variables                           */
    int reference_band_index;    /* Reference position for the arrays        */
    int search_band_index;       /* Search position for the arrays           */
    double projection_distance;  /* Make if conditions easier to read        */
    int status;                  /* Return status                            */
    int GCP_record_size;         /* Total size of GCP record points          */
    IAS_GCP_RECORD *gcp_record;  /* Generic GCP record                       */
    IAS_GEO_PROJ_TRANSFORMATION *ref_degrees2map = NULL; 
                /* Projection transformation to convert degrees to the
                   reference file L1G map projection */
    IAS_GEO_PROJ_TRANSFORMATION *srch_degrees2map = NULL; 
                /* Projection transformation to convert degrees to the
                   search file L1G map projection */
    double geo_parms[IAS_PROJ_PARAM_SIZE]; 
    IAS_PROJECTION geo_proj;
    IAS_PROJECTION map_proj;
    int ref_out_units;          /* Code for the output projection units */
    int srch_out_units;         /* Code for the output projection units */
    int overall_gcp_version;    /* Overall GCP version read in from Lib */
    /* Initialize the pointers to NULL. */
    *ref_gcp15 = NULL;
    *ref_gcp30 = NULL;
    *srch_gcp15 = NULL;
    *srch_gcp30 = NULL;

    /* Convert the projection units to a unit code for GCTP. */
    status = ias_geo_get_units(ref_file_metadata->projection_units,
        &ref_out_units);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Converting projection units %s to unit code",
            ref_file_metadata->projection_units);
        return ERROR;
    }
    status = ias_geo_get_units(srch_file_metadata->projection_units,
        &srch_out_units);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Converting projection units %s to unit code",
            srch_file_metadata->projection_units);
        return ERROR;
    }

    /* Create the geographic projection */
    for (i = 0; i < IAS_PROJ_PARAM_SIZE; i++)
        geo_parms[i] = 0.0;
    ias_geo_set_projection(GEO, NULLZONE, DEGREE, 
        ref_file_metadata->spheroid_code, geo_parms, &geo_proj);

    /* Create the map projection for the reference file */
    ias_geo_set_projection(ref_file_metadata->projection_code,
        ref_file_metadata->zone_code, ref_out_units,
        ref_file_metadata->spheroid_code,
        ref_file_metadata->projection_parameters, &map_proj);

    /* Create the transformation for the reference file */
    ref_degrees2map = ias_geo_create_proj_transformation(&geo_proj,
        &map_proj);
    if (!ref_degrees2map)
    {
        IAS_LOG_ERROR("Failed to create the reference projection "
            "transformation");
        return EXIT_FAILURE;
    }

    /* Create the map projection for the search file */
    ias_geo_set_projection(GEO, NULLZONE, DEGREE, 
        srch_file_metadata->spheroid_code, geo_parms, &geo_proj);
    ias_geo_set_projection(srch_file_metadata->projection_code,
        srch_file_metadata->zone_code, srch_out_units,
        srch_file_metadata->spheroid_code,
        srch_file_metadata->projection_parameters, &map_proj);

    /* Create the transformation for the search file */
    srch_degrees2map = ias_geo_create_proj_transformation(&geo_proj,
        &map_proj);
    if (!srch_degrees2map)
    {
        IAS_LOG_ERROR("Failed to create the search projection transformation");
        ias_geo_destroy_proj_transformation(ref_degrees2map);
        return EXIT_FAILURE;
    }

    /* Read the entire specified GCP record file.  Nothing will be filtered 
       out. */
    status = ias_gcp_read_gcplib(gcp_filename, &gcp_record, number_of_gcps,
        &overall_gcp_version);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Reading tiepoints from GCP Library %s", gcp_filename);
        ias_geo_destroy_proj_transformation(ref_degrees2map);
        ias_geo_destroy_proj_transformation(srch_degrees2map);
        return ERROR;
    }

    GCP_record_size = (*number_of_gcps) * sizeof(IAS_GCP_RECORD);

    for (i = 0; i < number_bands; i++)
    {
        /* Get the search and reference positions. */
        reference_band_index = reference_band_indexes[i];
        search_band_index = search_band_indexes[i];

        /* Determine the reference line and sample values for the needed
           resolutions. */

        /* If GCP15 hasn't already been done and a 15m band is present then we
           need a GCP15 set of tiepoints. */
        projection_distance =
            ref_band_metadata[reference_band_index].projection_distance_x;
        if (IS_PROJ_DIST(projection_distance, 15.0) && (*ref_gcp15 == NULL))
        {
            /* Determine the GCP15 tiepoints. */
            *ref_gcp15 = malloc(GCP_record_size);
            if (*ref_gcp15 == NULL)
            {
                IAS_LOG_ERROR("Allocating reference GCP15 buffer");
                ias_geo_destroy_proj_transformation(ref_degrees2map);
                ias_geo_destroy_proj_transformation(srch_degrees2map);
                return ERROR;
            }

            /* Convert the current reference GCPs to GCPs for the 15m band. */
            status = get_gcp_line_sample(
                &ref_band_metadata[reference_band_index], ref_degrees2map,
                ref_number_lines[i], ref_number_samples[i],
                *number_of_gcps, gcp_record, *ref_gcp15);
            if (status == ERROR)
            {
                IAS_LOG_ERROR("Determining reference GCP15 line/sample "
                              "values");
                free(*ref_gcp15);
                free(*srch_gcp15);
                free(*ref_gcp30);
                free(*srch_gcp30);
                free(gcp_record);
                ias_geo_destroy_proj_transformation(ref_degrees2map);
                ias_geo_destroy_proj_transformation(srch_degrees2map);
                return ERROR;
            }
        }

        /* If GCP30 hasn't already been done and there is a 30m band then we 
           need a GCP30 set of tiepoints. */
        else if (IS_PROJ_DIST(projection_distance, 30.0) 
            && (*ref_gcp30 == NULL))
        {
            /* Determine the GCP30 tiepoints. */
            *ref_gcp30 = malloc(GCP_record_size);
            if (*ref_gcp30 == NULL)
            {
                IAS_LOG_ERROR("Allocating reference GCP30 buffer");
                ias_geo_destroy_proj_transformation(ref_degrees2map);
                ias_geo_destroy_proj_transformation(srch_degrees2map);
                return ERROR;
            }

            /* Convert the current reference GCPs to GCPs for the visible 
               bands. */
            status = get_gcp_line_sample(
                &ref_band_metadata[reference_band_index], 
                ref_degrees2map, ref_number_lines[i], ref_number_samples[i],
                *number_of_gcps, gcp_record, *ref_gcp30);
            if (status == ERROR)
            {
                IAS_LOG_ERROR("Determining reference GCCP15 line/sample "
                              "values");
                free(*ref_gcp15);
                free(*srch_gcp15);
                free(*ref_gcp30);
                free(*srch_gcp30);
                free(gcp_record);
                ias_geo_destroy_proj_transformation(ref_degrees2map);
                ias_geo_destroy_proj_transformation(srch_degrees2map);
                return ERROR;
            }
        }

        /* Determine the search line and sample values for the needed
           resolutions. */

        /* If GCP15 hasn't already been done and a 15m band is present, we 
           need a GCP15 set of tiepoints.  */
        projection_distance =
            srch_band_metadata[search_band_index].projection_distance_x;
        if (IS_PROJ_DIST(projection_distance, 15.0) && (*srch_gcp15 == NULL))
        {
            /* Determine the GCP15 tiepoints. */
            *srch_gcp15 = malloc(GCP_record_size);
            if (*srch_gcp15 == NULL)
            {
                IAS_LOG_ERROR("Allocating search GCP15 buffer");
                ias_geo_destroy_proj_transformation(ref_degrees2map);
                ias_geo_destroy_proj_transformation(srch_degrees2map);
                return ERROR;
            }

            /* Convert the current search GCPs to GCPs for the 15m band. */
            status = get_gcp_line_sample(&srch_band_metadata[search_band_index],
                srch_degrees2map, srch_number_lines[i], 
                srch_number_samples[i], *number_of_gcps, gcp_record, 
                *srch_gcp15);
            if (status == ERROR)
            {
                IAS_LOG_ERROR("Determining search GCP15 line/sample values");
                free(*ref_gcp15);
                free(*srch_gcp15);
                free(*ref_gcp30);
                free(*srch_gcp30);
                free(gcp_record);
                ias_geo_destroy_proj_transformation(ref_degrees2map);
                ias_geo_destroy_proj_transformation(srch_degrees2map);
                return ERROR;
            }
        }

        /* If GCP30 hasn't already been done and a 30m band is present, we 
           need a GCP30 set of tiepoints. */
        else if (IS_PROJ_DIST(projection_distance, 30.0)
            && (*srch_gcp30 == NULL))
        {
            /* Determine the GCP30 tiepoints. */
            *srch_gcp30 = malloc(GCP_record_size);
            if (*srch_gcp30 == NULL)
            {
                IAS_LOG_ERROR("Allocating search GCP30 buffer");
                ias_geo_destroy_proj_transformation(ref_degrees2map);
                ias_geo_destroy_proj_transformation(srch_degrees2map);
                return ERROR;
            }

            /* Convert the current search GCPs to GCPs for the visible bands. */
            status = get_gcp_line_sample(&srch_band_metadata[search_band_index],
                srch_degrees2map, srch_number_lines[i], 
                srch_number_samples[i], *number_of_gcps, gcp_record, 
                *srch_gcp30);
            if (status == ERROR)
            {
                IAS_LOG_ERROR("Determining search GCP30 line/sample values");
                free(*ref_gcp15);
                free(*srch_gcp15);
                free(*ref_gcp30);
                free(*srch_gcp30);
                free(gcp_record);
                ias_geo_destroy_proj_transformation(ref_degrees2map);
                ias_geo_destroy_proj_transformation(srch_degrees2map);
                return ERROR;
            }
        }
    }

    /* Free the GCP record. */
    free(gcp_record);

    /* Cleanup the projection transformations */
    ias_geo_destroy_proj_transformation(ref_degrees2map);
    ias_geo_destroy_proj_transformation(srch_degrees2map);

    return SUCCESS;
}
