/******************************************************************************
Name:           read_metadata

Purpose:  Read the file and band metadata for a specified L1G image.

Notes:
This routine opens the L1G image file and reads both the file metadata
and the band metadata for all user-specified bands listed in the "*bands"
array, in the given order. The number_bands parameter specifies the total 
number of bands in the array.

Algorithm References:
The code is a modified version of the routine read_ddrs originally
developed for the Landsat 7 IAS (and used essentially as-is in the ALIAS).

See ias_l1g.h for a description of the file and band metadata structures.

******************************************************************************/

#include "i2i_char.h"
#include "ias_logging.h"        /* Logging library                           */

int read_metadata
(
    const char *filename,       /* I: L1G image file name                    */
    const int *band_indexes,    /* I: Array of bands indexes to read         */
    int number_bands,           /* I: Number of bands in the band indexes
                                      array                                  */
    int *number_lines,          /* O: Number of lines in imagery             */
    int *number_samples,        /* O: Number of samples in imagery           */
    IAS_L1G_FILE_METADATA *file_metadata, /* O: File metadata structure      */
    IAS_L1G_BAND_METADATA *band_metadata  /* O: Array of band metadata 
                                      structures                             */
)
{
    int band_number;            /* "True" OLI band number                    */
    int i;                      /* loop counter                              */
    int band_index;             /* corresponding reference position for the
                                   band numbers                              */
    int status;                 /* function return status code               */

    L1GIO *img_ptr;             /* Pointer to L1G image data                 */
    IAS_DATA_TYPE dtype;        /* data type of each band                    */
    int band_nsca;              /* number of SCAs in the band                */


    /* Open the image for reading. */
    img_ptr = ias_l1g_open_image(filename, IAS_READ);
    if (img_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening image %s", filename);
        return ERROR;
    }

    /* Get the file metadata. */
    status = ias_l1g_get_file_metadata(img_ptr, file_metadata);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Reading file metadata from file %s", filename);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    /* Loop through the bands and place them in their corresponding
       reference position. */
    for (i = 0; i < number_bands; i++)
    {
        band_index = band_indexes[i];
        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        if ((band_number < 1) || (band_number > IAS_MAX_NBANDS))
        {
            IAS_LOG_ERROR("Invalid band number %d", band_number);
            ias_l1g_close_image(img_ptr);
            return ERROR;
        }

        /* Get the band metadata. */
        status = ias_l1g_get_band_metadata(img_ptr, band_number,
                                           &band_metadata[band_index]);
        if (status == ERROR)
        {
            IAS_LOG_ERROR("Reading metadata for band number %d from file %s", 
                band_number, filename);
            ias_l1g_close_image(img_ptr);
            return ERROR;
        }

        /* Get the number of lines and samples. */
        if (ias_l1g_get_band_size(img_ptr, band_number, &dtype, &band_nsca,
            &number_lines[i], &number_samples[i]) != SUCCESS)
        {
            IAS_LOG_ERROR("Getting size of band number %d from image %s.",
                band_number, filename);
            ias_l1g_close_image(img_ptr);
            return ERROR;
        }
    }

    /* Close the image data file. */
    status = ias_l1g_close_image(img_ptr);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Closing image %s", filename);
        return ERROR;
    }

    return SUCCESS;
}
