#!/bin/csh -e
#############################################################################
# Name: export_l8_angles
#
# Description: This creates a l8_angles.tgz file with all the software to build
#   the l8_angles application.  After building the tarball, it tests to make
#   sure the software compiles.
#
# Parameters: 
#       1. Optional: SVN Repository url
#       2. Optional: Save output directory by adding "save" as a parameter
#           
# Examples: 
#       1. ./export_l8_angles save
#       2. ./export_l8_angles "svn_url" save
#       3. ./export_l8_angles "svn_url"
###############################################################################
set ERROR = 1
set SUCCESS = 0
set SRCROOT = "../../../.."
set remove_dir = 1
set use_svn_info = 1

# Retrieve the parameters 
if ( $#argv == 1 ) then
    if ( $1 == "save" ) then
        set remove_dir = 0
    else
        set use_svn_info = 0
    endif
else if ( $#argv == 2 ) then
    set use_svn_info = 0
    if ( $2 == "save" ) then
        set remove_dir = 0
    endif
endif

if ( $use_svn_info ) then
    set svn_repo = `svn info | grep 'Repository Root:' | cut -d ' ' -f 3`
    set svn_repo = "$svn_repo/trunk/COTSsrc"
else
    set svn_repo = $1
endif

echo using "$svn_repo"
# Complete set of required library files to use the l8 angles application
set library_files = (\
    "/ias_lib/misc/geo/ias_geo_convert_dms2deg.c" \
    "/ias_lib/misc/satellite_attributes/ias_satellite_attributes.c" \
    "/ias_lib/misc/satellite_attributes/ias_satellite_attributes.h" \
    "/ias_lib/misc/satellite_attributes/local_defines.h"\
    "/ias_lib/misc/satellite_attributes/landsat8.c"\
    "/ias_lib/misc/odl/ias_odl.h" \
    "/ias_lib/misc/odl/ias_odl_free_tree.c" \
    "/ias_lib/misc/odl/ias_odl_get_field.c" \
    "/ias_lib/misc/odl/ias_odl_read_tree.c" \
    "/ias_lib/misc/angle_generation/ias_angle_gen_utilities.c" \
    "/ias_lib/misc/angle_generation/ias_angle_gen_find_scas.c" \
    "/ias_lib/misc/angle_generation/ias_angle_gen_initialize.c" \
    "/ias_lib/misc/angle_generation/ias_angle_gen_read_ang.c" \
    "/ias_lib/misc/angle_generation/ias_angle_gen_calculate_angles_rpc.c" \
    "/ias_lib/misc/angle_generation/ias_angle_gen_private.h" \
    "/ias_lib/misc/angle_generation/ias_angle_gen_write_image.c" \
    "/ias_lib/misc/angle_generation/ias_angle_gen_distro.h" \
    "/ias_lib/misc/math/ias_math.h" \
    "/ias_lib/misc/math/ias_math_compute_unit_vector.c" \
    "/ias_lib/misc/math/ias_math_compute_vector_length.c" \
    "/ias_lib/misc/math/ias_math_find_line_segment_intersection.c" \
    "/ias_lib/misc/ias_structures.h" \
    "/ias_lib/misc/ias_types.h" \
    "/ias_lib/misc/ias_const.h" \
    "/ias_lib/misc/ias_logging.h" \
    "/ias_lib/misc/ias_logging.c" \
    "/ias_lib/misc/miscellaneous/ias_misc_write_envi_header.c" \
    "/ias_lib/misc/miscellaneous/ias_misc_create_output_image_trim_lut.c" \
    "/ias_lib/misc/miscellaneous/ias_misc_convert_to_uppercase.c" \
    "/ias_lib/io/parameter_file_io/ias_parm_map_odl_type.c" \
    "/ias_lib/io/parameter_file_io/ias_parm_provide_help.c" \
    "/ias_lib/io/parameter_file_io/ias_parm_read.c" \
    "/ias_lib/io/parameter_file_io/ias_parm_private.h" \
    "/ias_lib/io/parameter_file_io/ias_parm.h" \
    "/ias_lib/io/parameter_file_io/ias_parm_check_ranges.c")

# COTS files that will be exported via SVN
set cots_odl_files = (\
    "lablib3.c" \
    "/lablib3.h" \
    "/toolbox.h")

# The l8_angles application files that need to be copied to the distrubtion
# package and compiled
set l8_angles_files = (\
    "../angles_api.c" \
    "../l8_angles.c" \
    "../l8_angles.h")

# Create temporary directory for files to be placed and compiled
set base_dir = "./l8_angles"
set copy_dir = "./l8_angles/ias_lib"
set tar_file = "./l8_angles.tgz"

# Check if the base directory already exists. If it does remove it.
if ( -d $base_dir ) then 
    rm -rf $base_dir
    if ( $status ) then
        echo "Error removing old $base_dir directory\n"
        exit $ERROR
    endif 
endif

# Create the base l8_angles directory
mkdir $base_dir
if ( $status ) then
    echo "Error making $base_dir directory\n"
    exit $ERROR
endif

# Create the ias_lib directory
mkdir $copy_dir
if ( $status ) then
    echo "Error making $copy_dir directory\n"
    exit $ERROR
endif 

# Copy each of the library files needed
foreach file ($library_files)
    cp "${SRCROOT}${file}" "${copy_dir}/"
    if ( $status) then
        echo "Error copying file: $file\n"
        exit $ERROR
    endif
end

# Change to the local ias_lib directory to export the COTS files
cd $copy_dir

# Export each of the ODL files
foreach file ($cots_odl_files)
    svn export "$svn_repo/ODL/${file}"
    if ( $status ) then
        echo "Error exporting $file"
        exit $ERROR
    endif
end

# Change back to the main directory
cd ../../

# Copy each of the l8_angles source files
foreach file ($l8_angles_files)
    cp "${file}" "${base_dir}/"
    if ( $status) then
        echo "Error copying file: $file\n"
        exit $ERROR
    endif
end

# Copy the locally create make files and makeshift header file
cp "Makefile_l8_angles" $base_dir/Makefile
cp "README.txt" $base_dir/README.txt
cp "Makefile_ias_lib" $copy_dir/Makefile
cp "ias_angle_gen_includes.h" $copy_dir/
cp "config.h" $copy_dir/

# Changes back to the ias_lib directory
cd $copy_dir
set SRCROOT = "../../../../../../"

# Replace the use of gctp.h with the makeshift header
sed -i 's/gctp.h/ias_angle_gen_includes.h/g' ias_misc_write_envi_header.c

# ias_geo.h needs to be dynamically created 
echo "#ifndef IAS_GEO_H" > ias_geo.h
echo "#define IAS_GEO_H" >> ias_geo.h
foreach cmd (\
    "grep -B 0 'IAS_GEO_PROJ_TRANSFORMATION;' " \
    "grep -B 9 'IAS_PROJECTION;' " \
    "grep -A 6 ias_geo_convert_dms2deg ")
    eval "$cmd $SRCROOT/ias_lib/misc/geo/ias_geo.h >> ias_geo.h"
end
echo "#endif" >> ias_geo.h

# ias_miscellaneous.h also needs to be dynamically created 
echo "#ifndef IAS_MISCELLANEOUS_H" > ias_miscellaneous.h
echo "#define IAS_MISCELLANEOUS_H" >> ias_miscellaneous.h
echo '#include "ias_structures.h"' >> ias_miscellaneous.h
echo '#include "ias_types.h"' >> ias_miscellaneous.h
echo '#include "ias_const.h"' >> ias_miscellaneous.h
echo '#include "ias_geo.h"' >> ias_miscellaneous.h
echo '#include "ias_math.h"' >> ias_miscellaneous.h
foreach cmd (\
    "grep -B 4 'IAS_MISC_LINE_EXTENT;' " \
    "grep -A 6 'ias_misc_create_output_image_trim_lut' " \
    "grep -A 25 'int ias_misc_write_envi_header' " \
    "grep -A 3 'ias_misc_convert_to_uppercase' ")
    eval "$cmd $SRCROOT/ias_lib/misc/miscellaneous/ias_miscellaneous.h >> ias_miscellaneous.h"
end
echo "#endif" >> ias_miscellaneous.h

# Everything has succeeded so tar up the distribution package
cd ../../

# Check if old distribution tar ball exists
if (-e $tar_file) then
    rm $tar_file
    if ( $status ) then
        echo "Removing old tar file $tar_file"
        exit $ERROR
    endif
endif

# Create new tar ball for distribution
tar -czvf $tar_file $base_dir
if ( $status ) then
    echo "Error creating tar ball of l8_angles distrubtion package"
    exit $ERROR
endif
cd $copy_dir

# Try making the new distribution library 
make
if ( $status ) then
    echo "Error making distrubution library"
    exit $ERROR
endif
cd ..

# The library succeeded try making the l8_angles application
make
if ( $status ) then
    echo "Error making distrubution l8_angles"
    exit $ERROR
endif
cd ..

# Cleanup if not otherwise stated
if ( $remove_dir ) then
    rm -rf $base_dir
endif

exit $SUCCESS
