/******************************************************************************
NAME:           build_geom_grid

PURPOSE:
build_geom_grid builds an inverse geometric grid given the source projection 
information and target projection information.  The out_lines and out_samps are 
filled with a regular geometric grid.  The in_lines and in_samps are calculated 
by using ias_geo_compute_proj2proj_poly to project the out_lines, out_samps 
intersections from the output projection to the input projection.

RETURN VALUE:   
Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

******************************************************************************/

#include "ias_const.h"         /* SUCCESS/ERROR definition */
#include "ias_logging.h"       /* ias logging prototype */
#include "ias_geo.h"           /* ias geo prototypes */
#include "ias_math.h"          /* ias math prototypes */
#include "ias_l1g.h"           /* ias l1g prototype */
#include "build_geom_grid.h"   /* prototypes for this module */
#include "make_geom_grid.h"    /* PROJECTION_INFO structure definition */


int build_geom_grid
(
    const PROJECTION_INFO *proj_info,/* I: input image projection information */
    const IAS_GEO_PROJ_TRANSFORMATION *target2source, /* I: target to source
                                       projection transformation */
    int band_index,              /* I: grid band index to build */
    IAS_GEOM_GRID_TYPE *grid_ptr,/* I/O: grid to build, already containing
                                         the target image information */
    int source_crosses_180,      /* I: should be set if the source scene is 
                                       one that crosses the 180 degree line */
    int target_crosses_180       /* I: should be set if the target scene is one
                                       that crosses the 180 degree line */
)
{
    int count;         /* loop counter */
    int line_index;    /* line loop counter */
    int samp_index;    /* sample loop counter */
    int temp_index;    /* temporary index */
    IAS_GEOM_GRID_BAND_TYPE *grid_band_ptr = &grid_ptr->gridbands[band_index];
                       /* cached pointer to the grid band */
    double out_line;   /* current output space line */
    double out_samp;   /* current output space sample */
    double out_proj_x; /* output space projection x */
    double out_proj_y; /* output space projection y */
    double in_proj_x;  /* input space projection x */
    double in_proj_y;  /* input space projection y */
    double in_line;    /* current input space line */
    double in_samp;    /* current input space samp */
    double out_ls_to_proj_x[4];/*output line/sample to x projection polynomial*/
    double out_ls_to_proj_y[4];/*output line/sample to y projection polynomial*/
    double in_proj_to_line[4]; /* input projection to line polynomial */
    double in_proj_to_samp[4]; /* input projection to sample polynomial */

    double source_proj_x[4];   /* source frame corner to x projection  */
    double source_proj_y[4];   /* source frame corner to y projection */
    double target_proj_x[4];   /* target frame corner to x projection */
    double target_proj_y[4];   /* target frame corner to y projection */
    int source_units;
    int target_units;

    /* convert the projection units to int */
    if (ias_geo_get_units(proj_info->projection_units, &source_units)
        != SUCCESS) 
    {
        IAS_LOG_ERROR("Parsing the source units from: %s",
               proj_info->projection_units);
        return ERROR;
    }
    
    if (ias_geo_get_units(grid_ptr->units, &target_units) != SUCCESS)
    {
        IAS_LOG_ERROR("Parsing the target units from: %s", grid_ptr->units);
        return ERROR;
    }

    
    /* fill the out lines and out samps arrays */
    for (count = 0; count < grid_band_ptr->num_out_lines; count++)
        grid_band_ptr->out_lines[count] = count * grid_band_ptr->cell_lines;

    for (count = 0; count < grid_band_ptr->num_out_samps; count++)
        grid_band_ptr->out_samps[count] = count * grid_band_ptr->cell_samps;
        

    /* calculate the polynomial coefficients for the output image
       (line,sample) to projection coordinates transformation */
    source_proj_y[0] = 0; /* upper left line */
    source_proj_y[1] = 0; /* upper right line */
    source_proj_y[2] = grid_band_ptr->output_image_lines-1;/* lower left line */
    source_proj_y[3] = grid_band_ptr->output_image_lines-1;/* lower right line*/
    source_proj_x[0] = 0; /* upper left sample */ 
    source_proj_x[1] = grid_band_ptr->output_image_samps-1;/* upper right samp*/
    source_proj_x[2] = 0; /* lower left sample */
    source_proj_x[3] = grid_band_ptr->output_image_samps-1;/* lower right samp*/
    target_proj_y[0] = grid_ptr->corners.upleft.y;
    target_proj_y[1] = grid_ptr->corners.upright.y;
    target_proj_y[2] = grid_ptr->corners.loleft.y;
    target_proj_y[3] = grid_ptr->corners.loright.y;
    target_proj_x[0] = grid_ptr->corners.upleft.x;
    target_proj_x[1] = grid_ptr->corners.upright.x;
    target_proj_x[2] = grid_ptr->corners.loleft.x;
    target_proj_x[3] = grid_ptr->corners.loright.x;
    
    /* if this is a 180 degree crossing scene, we will want to make
       sure all the longitudes get shifted so they are positive */
    if (target_crosses_180)
    {
        int corner = 0;

        for (corner = 0; corner < 4; corner++)
            if (target_proj_x[corner] < 0)
            {
                ias_geo_add_once_around(target_units, &(target_proj_x[corner]));
            }
    }

    if (ias_geo_compute_proj2proj_poly(source_proj_x, source_proj_y, 
                                       target_proj_x, target_proj_y, 
                                       1 /* include x*y */,
                                       out_ls_to_proj_x, out_ls_to_proj_y) 
        != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating the transform polynomial coefficients");
        return ERROR;
    } 
    
    /* calculate the polynomial coefficients for the input image
       projection coordinates to (line,sample) transformation */
    source_proj_y[0] = proj_info->upper_left_y;
    source_proj_y[1] = proj_info->upper_right_y;
    source_proj_y[2] = proj_info->lower_left_y;
    source_proj_y[3] = proj_info->lower_right_y;  
    source_proj_x[0] = proj_info->upper_left_x;
    source_proj_x[1] = proj_info->upper_right_x;
    source_proj_x[2] = proj_info->lower_left_x;
    source_proj_x[3] = proj_info->lower_right_x;
    target_proj_y[0] = 0; /* upper left line */
    target_proj_y[1] = 0; /* upper right line */
    target_proj_y[2] = proj_info->lines-1;/*lower left line */
    target_proj_y[3] = proj_info->lines-1;/*lower right line*/
    target_proj_x[0] = 0; /* upper left sample */ 
    target_proj_x[1] 
            = proj_info->samples-1;/*upper right samp*/
    target_proj_x[2] = 0;          /* lower left sample */
    target_proj_x[3] 
            = proj_info->samples-1;/*lower right samp*/
    
    /* if this is a 180 degree crossing scene, we will want to make
       sure all the longitudes get shifted so they are positive */
    if (source_crosses_180)
    {
        int corner = 0;

        for (corner = 0; corner < 4; corner++)
            if (source_proj_x[corner] < 0)
            {
                ias_geo_add_once_around(source_units, &(source_proj_x[corner]));
            }
    }

    if (ias_geo_compute_proj2proj_poly(source_proj_x, source_proj_y, 
                                        target_proj_x,
                                target_proj_y, 1 /* include x*y */,
                                in_proj_to_samp, in_proj_to_line) != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating the transform polynomial coefficients");
        return ERROR;
    }

    /* calculate the in lines and in samps arrays */
    for (line_index = 0; line_index < grid_band_ptr->num_out_lines;
         line_index++)
    {
        out_line = grid_band_ptr->out_lines[line_index];
        
        for (samp_index = 0; samp_index < grid_band_ptr->num_out_samps;
             samp_index++)
        {
            out_samp = grid_band_ptr->out_samps[samp_index];

            /* convert the output space line/sample into projection space */
            out_proj_x = ias_math_eval_poly_xy(1,out_ls_to_proj_x,out_samp,
                                               out_line);
            out_proj_y = ias_math_eval_poly_xy(1,out_ls_to_proj_y,out_samp,
                                               out_line);

            /* transform the output space projection into input space 
               projection */
            if (ias_geo_transform_coordinate(target2source, 
                             out_proj_x, out_proj_y, &in_proj_x,
                             &in_proj_y) != SUCCESS)
            {
                IAS_LOG_ERROR("Transforming projection");
                return ERROR;
            } 

            /* if this is a 180 degree crossing scene, we will want to make
               sure all the longitudes get shifted so they are positive */
            if (source_crosses_180)
            {
                if (in_proj_x < 0)
                {
                    ias_geo_add_once_around(source_units, &in_proj_x);
                }
            }
               
            /* convert the input space projection into an input line/sample */
            in_samp = ias_math_eval_poly_xy(1,in_proj_to_samp,in_proj_x,
                                            in_proj_y);
            in_line = ias_math_eval_poly_xy(1,in_proj_to_line,in_proj_x,
                                            in_proj_y);

            /* save the input line/sample */
            temp_index = line_index * grid_band_ptr->num_out_samps + samp_index;
            grid_band_ptr->in_lines[temp_index] = in_line;
            grid_band_ptr->in_samps[temp_index] = in_samp;
        }
    }

    return SUCCESS;
}
