/******************************************************************************
NAME:           calc_coeffs

PURPOSE:
calc_coeffs calculates the reverse coefficients and fills them into the 
geometric grid.

RETURN VALUE:   
Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

******************************************************************************/

#include "ias_const.h"      /* SUCCESS/ERROR definitions */
#include "ias_logging.h"    /* ias logging prototype */
#include "ias_geo.h"        /* ias geo prototypes */
#include "calc_coeffs.h"    /* prototype for this routine */


int calc_coeffs
(
    IAS_GEOM_GRID_BAND_TYPE *grid_band_ptr /* I/O: grid band to calculate 
                                                   coefficients for */
)
{
    int row;            /* grid row working on */
    int col;            /* grid column working on */
    int in_index_top;   /* in_line index of the top of the current grid cell */
    int in_index_bottom;/*in_line index of the bottom of the current grid 
                           cell   */
    int grid_cell_index;/* index to the current grid cell in the coeffs array*/
    
    double source_x[4]; /* corner samples of the output space */
    double source_y[4]; /* corner lines of the output space */
    double target_x[4]; /* corner sample of the input space */
    double target_y[4]; /* corner lines of the input space */

    /* loop through the rows of the grid */
    for (row = 0; row < grid_band_ptr->grid_rows; row++)
    {
        /* pack the output lines for calculating the reverse coefficients since
           these don't change for each row (the grid is regular in output 
           space) */
        source_y[0] = grid_band_ptr->out_lines[row];
        source_y[1] = grid_band_ptr->out_lines[row];
        source_y[2] = grid_band_ptr->out_lines[row+1];
        source_y[3] = grid_band_ptr->out_lines[row+1];

        /* loop through the columns of the grid */
        for (col = 0; col < grid_band_ptr->grid_cols; col++)
        {
            /* pack the output samples for calculating the reverse 
               coefficients */
            source_x[0] = grid_band_ptr->out_samps[col];
            source_x[1] = grid_band_ptr->out_samps[col+1];
            source_x[2] = grid_band_ptr->out_samps[col];
            source_x[3] = grid_band_ptr->out_samps[col+1];

            /* calculate indices for the in line/samp arrays */
            in_index_top = row * grid_band_ptr->num_out_samps + col;
            in_index_bottom = in_index_top + grid_band_ptr->num_out_samps;

            /* pack the input lines for calculating the reverse coefficients */
            target_y[0] = grid_band_ptr->in_lines[in_index_top];
            target_y[1] = grid_band_ptr->in_lines[in_index_top+1];
            target_y[2] = grid_band_ptr->in_lines[in_index_bottom];
            target_y[3] = grid_band_ptr->in_lines[in_index_bottom+1];

            /* pack the input samples for calculating the reverse 
               coefficients */
            target_x[0] = grid_band_ptr->in_samps[in_index_top];
            target_x[1] = grid_band_ptr->in_samps[in_index_top+1];
            target_x[2] = grid_band_ptr->in_samps[in_index_bottom];
            target_x[3] = grid_band_ptr->in_samps[in_index_bottom+1];

            /* calculate the grid cell index */
            grid_cell_index = row * grid_band_ptr->grid_cols + col;
            
            /* call the gxx_calc_transform_poly routine to solve for the 
               inverse coefficients, storing the results in the reverse
               coefficient array */
            if (ias_geo_compute_proj2proj_poly(source_x, source_y, target_x, 
                    target_y, 1,
                    grid_band_ptr->reverse_coeffs[grid_cell_index].a,
                    grid_band_ptr->reverse_coeffs[grid_cell_index].b) 
                != SUCCESS) 
            {
                IAS_LOG_ERROR("Calculating reverse coefficients");
                return ERROR;
            }
        }
    }

    return SUCCESS;
}
