/******************************************************************************
NAME:           get_input_projection_information

PURPOSE:
eck the input file type (grid, geom grid, or L1G file and then get the
input projection information.

RETURN VALUE:   
Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

******************************************************************************/
#include <stdio.h>          
#include "ias_types.h"      /* IAS_FRAME_TYPE definition */
#include "ias_const.h"      /* SUCCESS/ERROR, TRUE/FALSE definitions */
#include "ias_logging.h"    /* ias_logging prototype */
#include "get_proj_info.h"  /* prototypes for get_proj_info */
#include "ias_grid_io.h"    /* prottotype for grid file checking */
#include "ias_geom_grid.h"  /* prototype for geometric grid file checking */
#include "ias_l1g.h"        /* prototype for L1G file checking */
#include "get_input_projection_information.h"  /* prototypes for this module */

/* types of files recognized for reading projection information */

int get_input_projection_information
(
    char *filename,   /* I: file to get projection info from */
    int *band_ptr,    /* I/O: band index to use from file - if -1, 
                              search for the desired band */
    PROJECTION_INFO *proj_info,     /* O: pointer to the projection info */    
    IAS_FRAME_TYPE *frame_type_ptr, /* O: pointer to the frame type */
    IAS_DBL_LAT_LONG *center_pt_ptr,/* O: pointer to the center point */
    IAS_DBL_XY *proj_center_pt_ptr, /* O: pointer to the projection center 
                                          point */
    double *rotate_angle_ptr,   /* O: pointer to the rotation angle */
    double *orient_angle_ptr    /* O: pointer to the orientation angle */
)
{
    int status;

    /* Check if the filename is a grid file */
    status = ias_grid_is_grid_file(filename);
    if (status == TRUE)
    {
        IAS_LOG_DEBUG("The input file %s is grid file", filename);
        if( get_proj_info_grid(filename, band_ptr, proj_info, frame_type_ptr,
                                 center_pt_ptr, proj_center_pt_ptr,
                                 rotate_angle_ptr, orient_angle_ptr)
                                 != SUCCESS )
        {
            /* generate an error message and return error */
            IAS_LOG_ERROR("Getting proj info from grid file: %s", filename);
            return ERROR;
        }
    }
    else 
    {
        /* Check if the filename is a geometric grid file */
        status = ias_grid_is_geom_grid_file(filename);
        if (status == TRUE)
        {
            IAS_LOG_DEBUG("The input file %s is geom grid file", filename);
            if( get_proj_info_geom(filename, band_ptr, proj_info, 
                             frame_type_ptr, center_pt_ptr, 
                             proj_center_pt_ptr, rotate_angle_ptr, 
                             orient_angle_ptr) != SUCCESS )
            {
                /* generate an error message and return error */
                IAS_LOG_ERROR("Getting proj info from geom grid file: %s",
                               filename);
                return ERROR;
            }
        }
        else
        {
            /* Check if the filename is l1g file */
            status = ias_l1g_is_l1g_file(filename);
            if (status == FALSE)
            {
                /* Not identified as any file type, so give a warning */
                IAS_LOG_WARNING("Unable to identify file type of %s, so it "
                    "will be assumed that it is an L1G file", filename);
            }

            IAS_LOG_DEBUG("The input file %s is l1g file", filename);
            if( get_proj_info_l1g(filename, band_ptr, proj_info, 
                                 frame_type_ptr, center_pt_ptr, 
                                 proj_center_pt_ptr, rotate_angle_ptr, 
                                 orient_angle_ptr) != SUCCESS )
            {
                /* generate an error message and return error */
                IAS_LOG_ERROR("Getting proj info from l1g file: %s", filename); 
                return ERROR;
            }
        }
    }
    return SUCCESS;
}
