/******************************************************************************
NAME:           get_make_geom_grid_parameters

PURPOSE:        
This module reads the ODL file input parameters required by the make_geom_grid
executable.  It also provides accessor routines to the parameters read.

ROUTINES:
    read_make_geom_grid_parameters
    get_source_filename
    get_grid_filename
    get_geom_grid_filename
    get_cell_lines
    get_cell_samples
    get_target_band_index_list
    get_source_band_index_list
    get_num_bands
    get_processing_pass
    get_doq_image_flag

******************************************************************************/

#include <stdlib.h>                
#include <stdio.h>                
#include <string.h>                
#include <limits.h>                         /* for PATH_MAX */
#include "ias_logging.h"           
#include "ias_const.h"           
#include "ias_parm.h"                       /*ias parameter file prototype */
#include "ias_parm_standard_parameters.h"
#include "ias_satellite_attributes.h"       /* IAS_MAX_BANDS definitions */
#include "ias_miscellaneous.h"
#include "get_make_geom_grid_parameters.h"  /* prototypes for this module */

/* VARIABLE DECLARATIONS */

/* The band lists are in band_number form in the application parameter file */
static int source_band_number_list[IAS_MAX_NBANDS] = 
 {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1};
static int target_band_number_list[IAS_MAX_NBANDS] = 
 {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1};

/* The band lists are converted to band_indexes when used in the application */
static int source_band_index_list[IAS_MAX_NBANDS] = 
 {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1};
static int target_band_index_list[IAS_MAX_NBANDS] = 
 {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1};

/* number of bands contained in the band lists */
static int nbands;

/* various file name storage */
static char source_filename[PATH_MAX];
static char grid_filename[PATH_MAX];
static char work_order_id[IAS_WORKORDER_ID_SIZE];
static char geom_grid_filename[PATH_MAX];
static char omf_filename[PATH_MAX];

/* Processing flow pass        */
static int processing_pass;     

/* Processing DEM or DOQ input file */
static int doq_image_flag;

static int cell_lines;
static int cell_samples;
/* size of each grid cell */

/* Target WRS path and row */
static int wrs_path;
static int wrs_row;

/******************************************************************************
NAME:           read_make_geom_grid_parameters

PURPOSE:        
This function reads all the parameters from ODL file that make_geom_grid
requires.  All the necessary error checking is done on the parameters read 
(except checking to see if file names are legal/exist).

RETURN VALUE:
Type = int
Value    Description
-----    -----------
SUCCESS  The parameters were successfully read
ERROR    An error occurred reading the parameters

******************************************************************************/

int read_make_geom_grid_parameters
(
    const char *parameter_filename /* INPUT: ODL filename that contains 
                                             parameters */
)
{
    int count;                /* loop counter */
    int num_source_bands;     /* count of source bands specified */
    int num_target_bands;     /* count of target bands specified */
    int source_band;          /* current source band */
    int target_band;          /* current target band */
    int status;               /* return variable */
    char path[PATH_MAX];    /* path for the input parameter files */
    char *filename = NULL;    /* filename of the input parameter file */
    int sat_num;              /* Satellite number */
    IAS_SATELLITE_ID sat_id;  /* Satellite ID */

    /* -- Table definition for things from the application parameter file -- */

    IAS_PARM_DECLARE_TABLE(parms, 8);

    /* Add the work order ID parameter */
    IAS_PARM_WORK_ORDER_ID(parms, work_order_id, sizeof(work_order_id), 0);

    /* Get processing pass. */
    IAS_PARM_PROCESSING_PASS(parms, &processing_pass,
                             sizeof(processing_pass), 0);

    /* Add the grid file parameter */
    IAS_PARM_ADD_STRING(parms, GEOM_GRID_FILENAME, "Output Geom Grid filename",
        IAS_PARM_REQUIRED,
        /* no limits */
        0, NULL,
        /* no default */
        0, NULL,
        geom_grid_filename, sizeof(geom_grid_filename), 0);

    int default_cell_lines = 50;
    /* Add the cell lines parameter */
    IAS_PARM_ADD_INT(parms, CELL_LINES, "Number of lines in a cell",
        IAS_PARM_REQUIRED, IAS_PARM_NOT_ARRAY,
        /* has a limited range of 2-256 */
        1, 2, 256,
        /* has one default value */
        1, &default_cell_lines,
        &cell_lines, sizeof(cell_lines), 0);

    int default_cell_samples = 50;
    /* Add the cell samples parameter */
    IAS_PARM_ADD_INT(parms, CELL_SAMPLES, "Number of samples in a cell",
        IAS_PARM_REQUIRED, IAS_PARM_NOT_ARRAY,
        /* has a limited range of 2-256 */
        1, 2, 256,
        /* has one default value */
        1, &default_cell_samples,
        &cell_samples, sizeof(cell_samples), 0);

    /* Add the source band list parameter */
    IAS_PARM_ADD_INT(parms, SOURCE_BAND_NUMBER_LIST, "Source band list",
        IAS_PARM_REQUIRED, IAS_PARM_ARRAY,
        /* has no limits */
        0, 0, 0,
        /* has no default value */
        0, NULL,
        source_band_number_list, sizeof(source_band_number_list), 0);

    /* Add the target band list parameter */
    IAS_PARM_ADD_INT(parms, TARGET_BAND_NUMBER_LIST, "Target band list - "
        "entering 0 indicates the file's first available band will be used",
        IAS_PARM_REQUIRED, IAS_PARM_ARRAY,
        /* has no limits */
        0, 0, 0,
        /* has no default value */
        0, NULL,
        target_band_number_list, sizeof(target_band_number_list), 0);

    /* Put the processing type - DEM or DOQ processing, in a flag */
    IAS_PARM_SOURCE_IMAGE_TYPE(parms, IAS_PARM_OPTIONAL, &doq_image_flag,
        sizeof(doq_image_flag), 1);

    /* See if the user wanted help or a template */
    status = ias_parm_provide_help(parameter_filename, parms,
        IAS_PARM_GET_TABLE_SIZE(parms), IAS_INPUT_PARAMETERS);

    if (status == 1)
    {
        /* If the status of the provide help is 1 the user asked for help.
           Since the parameter file name contains the help request, copy it
           to the omf file name so the next call will ask for the same help */
        strcpy(omf_filename, parameter_filename);
    }
    else
    {
        /* Read the parameter file */
        count = IAS_PARM_GET_TABLE_SIZE(parms);
        status = ias_parm_read(parameter_filename, parms, count);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading parameters from: %s",
                parameter_filename);
            return ERROR;
        }

        /* Build the omf file name */
        strcpy(path, parameter_filename);
        filename = strrchr(path, '/');
        if (filename == NULL)     /* no directory portion in the path */
            sprintf(omf_filename, "%s.omf", work_order_id);           
        else
        {
            *filename = '\0';
            sprintf(omf_filename, "%s/%s.omf", path, work_order_id);
        }
    }

    /* -------------- Table definition for things from the OMF ------------- */

    IAS_PARM_DECLARE_TABLE(omf_parms, 5);

    /* Add the source dem filename parameter */
    IAS_PARM_ADD_STRING(omf_parms, DEM_FILENAME, "Source DEM file name",
        IAS_PARM_REQUIRED,
        /* no limits */
        0, NULL,
        /* no default */
        0, NULL,
        source_filename, sizeof(source_filename), 0);

    /* Get the name of the grid file. */
    IAS_PARM_GRID_FILENAME_FROM_PASS(omf_parms, IAS_PARM_REQUIRED, 
        grid_filename, sizeof(grid_filename), 0, 
        processing_pass);   

    /* Get the WRS path */
    IAS_PARM_TARGET_WRS_PATH(omf_parms, IAS_PARM_OPTIONAL, &wrs_path,
        sizeof(wrs_path), 1);

    /* Get the WRS row */
    IAS_PARM_TARGET_WRS_ROW(omf_parms, IAS_PARM_OPTIONAL, &wrs_row,
        sizeof(wrs_row), 1);

    /* Get the satellite number */
    IAS_PARM_SATELLITE(omf_parms, &sat_num, sizeof(sat_num), 1);

    /* See if the user wanted help or a template this time with the OMF file */
    status = ias_parm_provide_help(omf_filename, omf_parms,
        IAS_PARM_GET_TABLE_SIZE(omf_parms), IAS_OMF_PARAMETERS);
    if (status == 1)
    {
        exit(1);
    }    

    /* Read the OMF file */
    count = IAS_PARM_GET_TABLE_SIZE(omf_parms);
    status = ias_parm_read(omf_filename, omf_parms, count);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from OMF file: %s",
            omf_filename);
        return ERROR;
    }

    /* Using the satellite number read from the OMF file, initialize the
       satellite attributes library */
    sat_id = ias_sat_attr_get_satellite_id_from_satellite_number(sat_num);
    if (sat_id == ERROR)
    {
        IAS_LOG_ERROR("Unsupported satellite number: %d", sat_num);
        return ERROR;
    }
    if (ias_sat_attr_initialize(sat_id) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attribute library");
        return ERROR;
    }

    /* make sure the band lists are not zero in length and are the same 
       length */
    num_source_bands = IAS_PARM_GET_COUNT_READ(parms, SOURCE_BAND_NUMBER_LIST);
    num_target_bands = IAS_PARM_GET_COUNT_READ(parms, TARGET_BAND_NUMBER_LIST);

    for (source_band = 0; source_band < num_source_bands; source_band++)
    {
        /* convert the band number to band index */
        source_band_index_list[source_band] 
            = ias_sat_attr_convert_band_number_to_index(
                source_band_number_list[source_band]);
        if (source_band_index_list[source_band] == ERROR)
        {
            IAS_LOG_ERROR("Converting source band number %d to a band index",
                source_band_number_list[source_band]);
            return ERROR;
        }
    }
 
    /* check if the target band list has a zero as its first entry, which
       indicates the list should not be used and that the desired target 
       band should be searched for by the get_proj_info routines */
    if (target_band_number_list[0] == 0)
    {
        /* flag the target band should be searched for */
        num_target_bands = 1;

        /* change the search flag into an illegal band number so
           the get_proj_info routines know to search for the band */
        target_band_index_list[0] = -1;

        /* make sure the list is terminated after the search flag */
        if ((target_band_number_list[1] > 0) && (target_band_number_list[1] 
              < IAS_MAX_NBANDS))
        {
            /* list not terminated so issue a warning message */
            IAS_LOG_ERROR("Error: TARGET_BAND_NUMBER_LIST may not contain any"
                          " bands if the search flag of zero is used");
            return ERROR;
        }

        /* make sure only a single source band is present since searching
           for the target band */
        if (num_source_bands != 1)
        {
            IAS_LOG_ERROR( "Error: only one source band may be specified"
                          " if searching for the target band");
            return ERROR;
        }
    }
    else
    {
        /* not searching for target band so parse the list */
        for (target_band=0; target_band < num_target_bands; target_band++)
        {
            /* convert the band number to band index */
            target_band_index_list[target_band] 
                = ias_sat_attr_convert_band_number_to_index(
                    target_band_number_list[target_band]);
            if (target_band_index_list[target_band] == ERROR)
            {
                IAS_LOG_ERROR("Converting target band number %d to a band "
                    "index", target_band_number_list[target_band]);
                return ERROR;
            }
        }

        if ((num_target_bands != num_source_bands) || (num_target_bands == 0))
        {
            IAS_LOG_ERROR("number of target and source bands do not match");
            return ERROR;
        }
    }

    /* save the number of bands count */
    nbands = num_target_bands;

    /* verify the number of cell lines is legal */
    if (cell_lines < 2)
    {
        IAS_LOG_ERROR("at least 2 cell lines must be specified");
        return ERROR;
    }

    if (cell_samples < 2)
    {
        IAS_LOG_ERROR("Error: at least 2 cell samples must be specified");
        return ERROR;
    }

    if (doq_image_flag)
    {
        /* Make sure the WRS path and row are available */
        if ((IAS_PARM_GET_COUNT_READ(omf_parms, TARGET_WRS_PATH) == 0)
            || (IAS_PARM_GET_COUNT_READ(omf_parms, TARGET_WRS_ROW) == 0))
        {
            /* WRS path and/or row weren't selected, but are needed */
            IAS_LOG_ERROR("WRS path and/or row were not specified, "
                "but they are needed to build the DOQ image name.");
            return ERROR;
        }

        /* Get the DOQ reference filename */
        status = ias_misc_get_doq_reference_filename(wrs_path, wrs_row,
            source_filename, sizeof(source_filename));
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to get the DOQ reference filename");
            return ERROR;
        }

        IAS_LOG_INFO("Using DOQ reference image %s", source_filename);
    }
    else 
    {
        /* if no source DEM file name was provided, log an error */
        if (strlen(source_filename) == 0)
        {
            IAS_LOG_ERROR("No source DEM file provided");
            return ERROR;                      
        }
    }

    if (strlen(grid_filename) == 0)
    {
        /* issue a warning that the source frame filename wasn't provided
           so it is being constructed */
        IAS_LOG_ERROR("No target grid file provided");
        return ERROR;                      
    }

    return SUCCESS;
}

/******************************************************************************
NAME:           various accessor functions

PURPOSE:        
These functions provide access to the parameters read. 

RETURN VALUE:
Varies for the function.

NOTES:
The accessor function method of providing parameters was selected since
passing all the parameters up and down a calling sequence is a 
maintenance nightmare.

******************************************************************************/


char *get_source_filename(void)
{
    return source_filename;
}

char *get_grid_filename(void)
{
    return grid_filename;
}

char *get_geom_grid_filename(void)
{
    return geom_grid_filename;
}

char *get_omf_filename(void)
{
    return omf_filename;
}

int get_cell_lines(void)
{
    return cell_lines;        
}

int get_cell_samples(void)
{
    return cell_samples;
}

int *get_target_band_index_list(void)
{
    return target_band_index_list;
}

int *get_source_band_index_list(void)
{
    return source_band_index_list;
}

int get_num_bands(void)
{
    return nbands;
}

int get_processing_pass(void)
{
    return processing_pass;
}

int get_doq_image_flag(void)
{
    return doq_image_flag;
}

