/******************************************************************************
NAME:           get_proj_info_geom

PURPOSE:
get_proj_info_geom reads the projection information needed to build a geometric
grid from the file name indicated.  The file read is a geometric grid file.

RETURN VALUE:   
Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
- If the first available band according in the geom file is desired, set the
  band index (*band_index_ptr) to -1.
- If the band search is used, the resulting projection will always be 
  a 30 meter projection.  If the band is specified, the projection will be
  unchanged from the original projection.  This behavior is a little odd, 
  but will work best for this application.

******************************************************************************/
#include <stdio.h>           
#include <string.h>          /* memcpy prototype */
#include "gctp.h"
#include "ias_types.h"       /* IAS_FRAME_TYPE definition */
#include "ias_const.h"       /* SUCCESS/ERROR definitions */
#include "ias_logging.h"     /* IAS Logging prototype */
#include "ias_satellite_attributes.h" /* NBANDS definition */
#include "ias_geo.h"         /* prototype in geometric library */
#include "ias_l1g.h"         /* prototype in l1g library */
#include "ias_geom_grid.h"   /* IAS_GRID_TYPE definition */
#include "get_proj_info.h"   /* prototypes for this module */

int get_proj_info_geom
(
    char *filename,   /* I: file to get projection info from */
    int *band_index_ptr, /* I/O: band index to use from file - if -1, search 
                              for the desired band */
    PROJECTION_INFO *proj_info,     /* O: pointer to the projection info */    
    IAS_FRAME_TYPE *frame_type_ptr, /* O: pointer to the frame type */
    IAS_DBL_LAT_LONG *center_pt_ptr,/* O: pointer to the center point */
    IAS_DBL_XY *proj_center_pt_ptr, /* O: pointer to the projection center 
                                          point */
    double *rotate_angle_ptr,   /* O: pointer to the rotation angle */
    double *orient_angle_ptr    /* O: pointer to the orientation angle */
)
{
    int count;                    /* loop counter */
    IAS_GEOM_GRID_TYPE geom_grid; /* geometric grid */
    int band_index;               /* geometric grid band index */
    int band_count = 0;           /* number of bands to free (0 means all) */
            
    /* Initialize the geom grid */
    if (ias_geom_grid_initialize(IAS_MAX_NBANDS, &geom_grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing IAS Geom Grid");
        return ERROR; 
    }

    /* IAS GEOM GRID file, so read the grid header information into memory */
    if (ias_read_geom_grid_header(filename, &geom_grid) != SUCCESS)
    {
        /* generate an error message and return error */
        IAS_LOG_ERROR("Reading geometric grid file: %s",filename);
        ias_geom_grid_free(NULL, &band_count, &geom_grid);
        return ERROR;
    }

    if (ias_read_geom_grid_band_header(filename, &geom_grid) != SUCCESS)
    {
        /* generate an error message and return error */
        IAS_LOG_ERROR("Reading geometric grid file: %s",filename);
        ias_geom_grid_free(NULL, &band_count, &geom_grid);
        return ERROR;
    }

    /* if a specific band was not requested, use the first available band in 
       the geometric grid, which will be at the first index */
    if (*band_index_ptr == -1)
    {
        *band_index_ptr = geom_grid.band_index_list[0];
    }

    /* set band index to geometric grid band index */ 
    band_index = *band_index_ptr;

    /* find the correct band in the geometric grid */
    for (count = 0; count < geom_grid.nbands; count++)
    {
        /* check if correct band */
        if (geom_grid.band_index_list[count] == band_index)
            break;
    }

    if (count >= geom_grid.nbands)
    {
        /* band not present in the grid */
        IAS_LOG_ERROR("Band %d not present in the geometric grid %s", 
                       band_index, filename);
        ias_geom_grid_free(NULL, &band_count, &geom_grid);
        return ERROR;
    }
    
    /* convert the grid information into a PROJECTION INFORMATION structure */
    proj_info->band_index = band_index;
    proj_info->lines = geom_grid.gridbands[band_index].output_image_lines;
    proj_info->samples= geom_grid.gridbands[band_index].output_image_samps;
    proj_info->projection_code = geom_grid.code;
    proj_info->zone_code = geom_grid.zone;
    proj_info->spheroid_code = geom_grid.spheroid;
    proj_info->upper_left_y = geom_grid.corners.upleft.y;
    proj_info->upper_left_x = geom_grid.corners.upleft.x;
    proj_info->lower_left_y = geom_grid.corners.loleft.y;
    proj_info->lower_left_x = geom_grid.corners.loleft.x;
    proj_info->upper_right_y = geom_grid.corners.upright.y;
    proj_info->upper_right_x = geom_grid.corners.upright.x;
    proj_info->lower_right_y = geom_grid.corners.loright.y;
    proj_info->lower_right_x = geom_grid.corners.loright.x;
    proj_info->projection_distance_y 
                  = geom_grid.gridbands[band_index].pixsize_line;
    proj_info->projection_distance_x 
                  = geom_grid.gridbands[band_index].pixsize_samp;

    memcpy(proj_info->datum,geom_grid.datum, sizeof(proj_info->datum));
    memcpy(proj_info->projection_units,geom_grid.units,
            sizeof(proj_info->projection_units));
    for (count = 0; count < IAS_PROJ_PARAM_SIZE; count++)
    {
        proj_info->projection_parameters[count] = geom_grid.projprms[count];
    }

    /* get the frame information */
    *frame_type_ptr = geom_grid.frame_type;
    *center_pt_ptr = geom_grid.center_pt;
    *proj_center_pt_ptr = geom_grid.proj_center_pt;
    *rotate_angle_ptr = geom_grid.rotate_angle;
    *orient_angle_ptr = geom_grid.orient_angle;

    /* free the temporary grid */
    ias_geom_grid_free(NULL, &band_count, &geom_grid);

    return SUCCESS;
}
