/******************************************************************************
NAME:           get_proj_info_grid

PURPOSE:
get_proj_info_grid reads the projection information needed to build a geometric
grid from the file name indicated.  The file read is a grid file.

RETURN VALUE:   
Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
- If the first available band according in the grid file is desired, set the 
  band index (*band_index_ptr) to -1.
- If the band search is used, the resulting projection will always be 
  a 30 meter projection.  If the band is specified, the projection will be
  unchanged from the original projection.  This behavior is a little odd, 
  but will work best for this application.

******************************************************************************/

#include <stdio.h>          /* sprintf prototype */
#include <string.h>         /* sprintf prototype */
#include "gctp.h"
#include "ias_types.h"      /* IAS_FRAME_TYPE definition */
#include "ias_const.h"      /* SUCCESS/ERROR definitions */
#include "ias_satellite_attributes.h"      /* NABNDS/NSCA definitions */
#include "ias_logging.h"    /* ias logging prototype */
#include "ias_geo.h"        /* ias geo prototype */
#include "ias_l1g.h"        /* ias l1g prototype */
#include "ias_geom_grid.h"  /* ias geom grid prototype */
#include "ias_grid_io.h"    /* ias grid prototype */
#include "get_proj_info.h"  /* prototypes for this module */

int get_proj_info_grid
(
    char *filename,   /* I: file to get projection info from */
    int *band_index_ptr, /* I/O: band index to use from file - if -1, search 
                              for the desired band */
    PROJECTION_INFO *proj_info,     /* O: pointer to the projection info */    
    IAS_FRAME_TYPE *frame_type_ptr, /* O: pointer to the frame type */
    IAS_DBL_LAT_LONG *center_pt_ptr,/* O: pointer to the center point */
    IAS_DBL_XY *proj_center_pt_ptr, /* O: pointer to the projection center 
                                          point */
    double *rotate_angle_ptr,   /* O: pointer to the rotation angle */
    double *orient_angle_ptr    /* O: pointer to the orientation angle */
)
{
    int count;              /* loop counter */
    IAS_GRID_TYPE grid;     /* OLI grid */
    int enum_band_index;    /* band index */
    int sca;                /* SCA number */
        
    /* Initialize the grid */
    if (ias_grid_initialize(&grid, IAS_MAX_NBANDS, IAS_MAX_NSCAS) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing IAS Grid");
        return ERROR; 
    }

    /* IAS GRID file, so read the grid header information into memory */
    if (ias_grid_header_read(filename, &grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Error reading the grid header file: %s",filename);
        ias_grid_free(&grid);
        return ERROR;
    }

    /* IAS GRID file, so read the grid band header information into memory */
    if (ias_grid_band_header_read(filename, &grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Error reading the grid band header file: %s",filename);
        ias_grid_free(&grid);
        return ERROR;
    }

    /* if a specific band was not requested, search grid for the first 
       available band */
    if (*band_index_ptr == -1)
    {
        for (*band_index_ptr = 0; *band_index_ptr < grid.nbands; 
            (*band_index_ptr)++)
        {
            /* found an available band */
            if (grid.bands_available[*band_index_ptr])
                break;
        }

        if (*band_index_ptr >= grid.nbands)
        {
            IAS_LOG_ERROR("No available bands found in grid file %s", filename);
            ias_grid_free(&grid);
            return ERROR;
        }
    }

    /* make sure the band is in the grid */
    enum_band_index = *band_index_ptr;
    if (!grid.bands_available[enum_band_index])
    {
        /* band not present in the grid */
        IAS_LOG_ERROR("Band index %d not present in the grid %s", 
                       enum_band_index, filename);
        ias_grid_free(&grid);
        return ERROR;
    }
     
    for (sca = 1; sca < grid.scas_per_band[enum_band_index]; sca++)
    {
        if( grid.gridbands[enum_band_index][0].lines != 
            grid.gridbands[enum_band_index][sca].lines )
        {
            IAS_LOG_ERROR("Lines are not the same for all SCAs in %s",
                          filename);
            ias_grid_free(&grid);
            return ERROR;
        }
        if( grid.gridbands[enum_band_index][0].samps != 
            grid.gridbands[enum_band_index][sca].samps )
        {
            IAS_LOG_ERROR("Samples are not the same for all SCAs in %s",
                          filename);
            ias_grid_free(&grid);
            return ERROR;
        }
        if( grid.gridbands[enum_band_index][0].pixsize != 
            grid.gridbands[enum_band_index][sca].pixsize )
        {
            IAS_LOG_ERROR("Pixel size are not the same for all SCAs in %s",
                          filename);
            ias_grid_free(&grid);
            return ERROR;
        }
    }

    /* convert the grid information into the projection information structure 
        format */
    proj_info->band_index = *band_index_ptr;
    proj_info->lines = grid.gridbands[enum_band_index][0].lines;
    proj_info->samples= grid.gridbands[enum_band_index][0].samps;
    proj_info->projection_code = grid.code;
    proj_info->zone_code = grid.zone;
    proj_info->spheroid_code = grid.spheroid;
    proj_info->upper_left_y = grid.corners.upleft.y;
    proj_info->upper_left_x = grid.corners.upleft.x;
    proj_info->lower_left_y = grid.corners.loleft.y;
    proj_info->lower_left_x = grid.corners.loleft.x;
    proj_info->upper_right_y = grid.corners.upright.y;
    proj_info->upper_right_x = grid.corners.upright.x;
    proj_info->lower_right_y = grid.corners.loright.y;
    proj_info->lower_right_x = grid.corners.loright.x;
    proj_info->projection_distance_y 
            = grid.gridbands[enum_band_index][0].pixsize;
    proj_info->projection_distance_x 
            = grid.gridbands[enum_band_index][0].pixsize;
    memcpy(proj_info->datum,grid.datum, sizeof(proj_info->datum));
    memcpy(proj_info->projection_units,grid.units,
           sizeof(proj_info->projection_units));
    for (count = 0; count < IAS_PROJ_PARAM_SIZE; count++)
    {
        proj_info->projection_parameters[count] = grid.projprms[count];
    }

    /* get the frame information */
    *frame_type_ptr = grid.frame_type;
    *center_pt_ptr = grid.center_pt;
    *proj_center_pt_ptr = grid.proj_center_pt;
    *rotate_angle_ptr = grid.rotate_angle;
    *orient_angle_ptr = grid.orient_angle;

    /* free the allocated memory for grid */
    ias_grid_free(&grid);

    return SUCCESS;
}
