/******************************************************************************
NAME:           get_proj_info_l1g

PURPOSE:
get_proj_info_l1g reads the projection information needed to build a geometric 
grid from the file name indicated.  The file read is an L1G file.

RETURN VALUE:   
Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

NOTES:
- If the first available band in the L1G file is desired, set the band index
  (*band_index_ptr) to -1.
- If the band search is used, the resulting projection will always be 
  a 30 meter projection.  If the band is specified, the projection will be
  unchanged from the original projection.  This behavior is a little odd, 
  but will work best for this application.

******************************************************************************/

#include <stdio.h>          
#include <stdlib.h>         /* free prototype */ 
#include <string.h>         /* memcpy prototype */
#include "gctp.h"
#include "ias_types.h"      /* IAS_DATA_TYPE, IAS_ACCESS_MODE, IAS_FRAME_TYPE
                               definitions */
#include "ias_const.h"      /* SUCCESS/ERROR definitions */
#include "ias_logging.h"    /* IAS logging prototype */
#include "ias_satellite_attributes.h"      /* NABNDS/NSCA definitions */
#include "ias_geo.h"        /* ias geo prototype */
#include "ias_geom_grid.h"  /* IAS_GRID_TYPE definition */
#include "get_proj_info.h"  /* prototypes for this module */

int get_proj_info_l1g
(
    char *filename,   /* I: file to get projection info from */
    int *band_index_ptr, /* I/O: band index to use from file - if -1, search 
                               for the desired band */
    PROJECTION_INFO *proj_info,     /* O: pointer to the projection info */    
    IAS_FRAME_TYPE *frame_type_ptr, /* O: pointer to the frame type */
    IAS_DBL_LAT_LONG *center_pt_ptr,/* O: pointer to the center point */
    IAS_DBL_XY *proj_center_pt_ptr, /* O: pointer to the projection center 
                                          point */
    double *rotate_angle_ptr,   /* O: pointer to the rotation angle */
    double *orient_angle_ptr    /* O: pointer to the orientation angle */
)
{
    L1GIO *img_ptr = NULL;
    L1G_BAND_IO *band_io = NULL;
    int count;                  /* loop counter */
    int band_number;            /* geometric grid band number */
    int lines, samples, number_of_scas;
    IAS_DATA_TYPE data_type;    
    IAS_L1G_FILE_METADATA *file_metadata = NULL;
    IAS_L1G_BAND_METADATA *band_metadata = NULL;
    int status;

    /* Open the HDF5 DEM image */
    img_ptr = ias_l1g_open_image(filename, IAS_READ);
    if (img_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening DEM data file %s\n", filename);
        return ERROR;
    }

    /* if a specific band was not requested, search L1G DEM for the first 
       available band */
    if (*band_index_ptr == -1)
    {
        for (*band_index_ptr = 0; *band_index_ptr < IAS_MAX_NBANDS; 
            (*band_index_ptr)++)
        {
            band_number = 
                ias_sat_attr_convert_band_index_to_number(*band_index_ptr);
            if (band_number == ERROR)
            {
                IAS_LOG_ERROR("Invalid band number for band index %d", 
                    *band_index_ptr);
                ias_l1g_close_image(img_ptr);
                return ERROR;
            }

            if (ias_l1g_is_band_present(img_ptr, band_number))
            {
                break;
            }
        }

        if (*band_index_ptr >= IAS_MAX_NBANDS)
        {
            IAS_LOG_ERROR("No bands found in file %s", filename);
            ias_l1g_close_image(img_ptr);
            return ERROR;
        }
    }
    else
    {
        band_number = 
            ias_sat_attr_convert_band_index_to_number(*band_index_ptr);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Invalid band number for band index %d", 
                *band_index_ptr);
            ias_l1g_close_image(img_ptr);
            return ERROR;
        }
    }

    status = ias_l1g_get_band_size(img_ptr, band_number, &data_type,
                                   &number_of_scas, &lines, &samples);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Getting band size information\n");
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    /* Allocate memory for file_metadata */
    file_metadata = malloc(sizeof(IAS_L1G_FILE_METADATA));
    if (file_metadata == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for file metadata");
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

     /* Read the file metadata */
    status = ias_l1g_get_file_metadata(img_ptr, file_metadata);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Reading file metadata\n");
        free(file_metadata);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    /* Set up the band data */
    int num_SCAs = 1;
    band_io = ias_l1g_open_band(img_ptr, band_number, &data_type, &num_SCAs, 
                                &lines, &samples);
    if (band_io == NULL)
    {
        IAS_LOG_ERROR("Opening DEM image band for information\n");
        free(file_metadata);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    /* Allocate memory for band_metadata */
    band_metadata = malloc(sizeof(IAS_L1G_BAND_METADATA));
    if (band_metadata == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for band metadata");
        free(file_metadata);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }    

    /* Read the band metadata */
    status = ias_l1g_get_band_metadata(img_ptr, band_number, band_metadata);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Getting band metadata\n");
        free(file_metadata);
        free(band_metadata);
        ias_l1g_close_band(band_io);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    /* Close the band */
    status = ias_l1g_close_band(band_io);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Closing the band\n");
        free(file_metadata);
        free(band_metadata);
        ias_l1g_close_image(img_ptr);
        return ERROR;
    }

    /* Close the image */
    status = ias_l1g_close_image(img_ptr);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Closing the DEM data file\n");
        free(file_metadata);
        free(band_metadata);
        return ERROR;
    }

    /* convert the grid information into a PROJECTION INFORMATION structure */
    proj_info->band_index = *band_index_ptr;
    proj_info->lines = lines;
    proj_info->samples= samples;
    proj_info->projection_code = file_metadata->projection_code;
    proj_info->zone_code = file_metadata->zone_code;
    proj_info->spheroid_code = file_metadata->spheroid_code;
    proj_info->upper_left_y = band_metadata->upper_left_y;
    proj_info->upper_left_x = band_metadata->upper_left_x;
    proj_info->lower_left_y = band_metadata->lower_left_y;
    proj_info->lower_left_x = band_metadata->lower_left_x;
    proj_info->upper_right_y = band_metadata->upper_right_y;
    proj_info->upper_right_x = band_metadata->upper_right_x;
    proj_info->lower_right_y = band_metadata->lower_right_y;
    proj_info->lower_right_x = band_metadata->lower_right_x;
    proj_info->projection_distance_y 
                  = band_metadata->projection_distance_y;
    proj_info->projection_distance_x 
                  = band_metadata->projection_distance_x;

    memcpy(proj_info->datum,file_metadata->datum, sizeof(proj_info->datum));
    memcpy(proj_info->projection_units,file_metadata->projection_units,
            sizeof(proj_info->projection_units));
    for (count = 0; count < IAS_PROJ_PARAM_SIZE; count++)
    {
        proj_info->projection_parameters[count] 
                       = file_metadata->projection_parameters[count];
    }

    /* If the projection is SOM, swap the corners since the SOM projection
       is defined with Y increasing moving east and X increasing moving south */
    if (proj_info->projection_code == SOM)
    {
        double temp;

        temp = proj_info->upper_left_x;
        proj_info->upper_left_x = proj_info->upper_left_y;
        proj_info->upper_left_y = -temp;
        temp = proj_info->lower_left_x;
        proj_info->lower_left_x = proj_info->lower_left_y;
        proj_info->lower_left_y = -temp;
        temp = proj_info->upper_right_x;
        proj_info->upper_right_x = proj_info->upper_right_y;
        proj_info->upper_right_y = -temp;
        temp = proj_info->lower_right_x;
        proj_info->lower_right_x = proj_info->lower_right_y;
        proj_info->lower_right_y = -temp;
    }

    /* Give it a valid frame type to allow I/O */
    *frame_type_ptr = IAS_MAXBOX;
    center_pt_ptr->lat = 0.0;
    center_pt_ptr->lng = 0.0;
    proj_center_pt_ptr->x = 0.0;
    proj_center_pt_ptr->y = 0.0;
    *rotate_angle_ptr = 0.0;
    *orient_angle_ptr = 0.0;

    /* free allocated memory for file and band metadata */
    free(file_metadata);
    free(band_metadata);

    return SUCCESS;
}
