/******************************************************************************
NAME:           make_geom_grid (main)

PURPOSE:
This is the main routine for making a geometric grid.

RETURN VALUE:   
Type = int
Value         Description
-----         -----------
SUCCESS  Successful completion
ERROR    Operation failed


******************************************************************************/

#include <stdlib.h>          /* EXIT_SUCCESS definition */
#include <stdio.h>           /* sprintf prototype */
#include <string.h>          /* memcpy prototype */
#include "ias_logging.h"     /* ias logging prototype */
#include "ias_types.h"       /* IAS_FRAME_TYPE definition */
#include "ias_const.h"       /* SUCCESS definition */
#include "ias_geom_grid.h"   /* ias geom grid prototypes */
#include "ias_satellite_attributes.h"      /* NABNDS/NSCA definitions */
#include "ias_l1g.h"         /* ias l1g prototypes */
#include "ias_math.h"        /* ias math prototypes */
#include "ias_geo.h"         /* ias geo prototypes */
#include "make_geom_grid.h"  /* prototype definitions */
#include "get_make_geom_grid_parameters.h" /* read_make_geom_grid_parameters
                                              prototype */
#include "get_proj_info.h"   /* get_proj_info prototype */
#include "build_geom_grid.h" /* build_geom_grid prototype */
#include "calc_coeffs.h"     /* calc_coeffs prototype */
#include "get_input_projection_information.h"
#include "gctp.h"
#include "update_omf.h"

int main(int argc, char *argv[])
{
    PROJECTION_INFO *source_ptr = NULL;  /* source image projection info 
                                            structure */
    PROJECTION_INFO target;  /* target image projection info structure */
    IAS_GEOM_GRID_TYPE geom_grid; /* geometric grid */
    IAS_FRAME_TYPE frame_type;  /* frame type */
    IAS_DBL_LAT_LONG center_pt; /* frame center point */
    IAS_DBL_XY proj_center_pt;  /* frame projection center point */
    int band_counter;        /* band counter to create grid for */
    int num_bands;           /* number of bands to create */
    int *source_band_index_list = NULL;   
                             /* pointer to the source band index list */
    int *target_band_index_list = NULL;   
                             /* pointer to the target band index list */
    int count;               /* loop counter */
    int geom_band_count = 0;    /* variable used to call geom grid
                                    malloc/free routines */
    int band_index;             /* band index for calling 
                                   get_proj_info */
    int grid_band = 0;          /* holds the source band for entering info
                                   into the grid */
    double rotate_angle;        /* frame rotation angle */
    double orient_angle;        /* frame orientation angle */

    /* Define arrays for determining the transformation from projection
       coordinates to image plane (or line-sample) coordinates */
    double source_proj_x[4];
    double source_proj_y[4];
    double image_proj_x[4];
    double image_proj_y[4];
    double target_proj_x[4];
    double target_proj_y[4];
    double proj_to_samp[4];
    double proj_to_line[4];

    /* Define variables for source-target image overlap test. */
    double source_image_x_max;
    double source_image_y_max;

    /* Define variables for projection related information */
    int target_units;
    int source_units;
    IAS_PROJECTION source_proj;
    IAS_PROJECTION target_proj;
    IAS_GEO_PROJ_TRANSFORMATION *target2source = NULL;
    /* Projection transformation to go from the target projection to the source
       projection */


    int source_crosses_180 = 0; /* Gets set if it is determined that the
                                   source scene crosses the 180 degree
                                   boundary */
    int target_crosses_180 = 0; /* Set if it is determined that the target 
                                   scene crosses the 180 degree boundary */
    int status;                 /* return variable */
    int band_number = 0;        /* band number */

    /* Initialize the log library */
    ias_log_initialize("Makegeomgrid");

    /* generate entry message */
    IAS_LOG_INFO("Initiated");

    /* verify the command line includes the odl parameter file */
    if (argc != 2)
    {
        IAS_LOG_ERROR("correct use: %s odl_parameter_file",argv[0]);
        return EXIT_FAILURE;
    }

    /* read the parameters from the odl file */
    if (read_make_geom_grid_parameters(argv[1]) != SUCCESS)
    {
        IAS_LOG_ERROR("Error reading parameters from: %s",argv[1]);
        return EXIT_FAILURE;
    }

    /* Initialize the geom grid */
    status = ias_geom_grid_initialize(IAS_MAX_NBANDS, &geom_grid);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing IAS Grid");
        return EXIT_FAILURE; 
    }

    /* get the band information */
    num_bands = get_num_bands();
    source_band_index_list = get_source_band_index_list();
    target_band_index_list = get_target_band_index_list();

    /* allocate memory for the Projection Info Structure */
    source_ptr = malloc(num_bands * sizeof(PROJECTION_INFO));
    if (source_ptr == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for the source projection "
                       "information");
        ias_geom_grid_free(&band_number, &geom_band_count, &geom_grid);
        return EXIT_FAILURE;
    }
    
    for (band_counter = 0; band_counter < num_bands; band_counter++)
    {
        int target_band_num = 0; /* Variable for storing the band number
                             this index maps to */

        /* set the source band index for calling get_proj_info */
        band_index = (int)source_band_index_list[band_counter];
        
        /* read the source image (DEM or DOQ) file */
        if (get_input_projection_information(get_source_filename(), 
                &band_index, &source_ptr[band_counter], &frame_type, 
                &center_pt, &proj_center_pt, &rotate_angle, &orient_angle) 
                != SUCCESS)
        {
            IAS_LOG_ERROR("Reading projection information from: %s",
                    get_source_filename());
            ias_geo_destroy_proj_transformation(target2source);
            free(source_ptr);
            ias_geom_grid_free(&band_number, &geom_band_count, &geom_grid);
            return EXIT_FAILURE;
        }

        /* set the target band index for calling get_proj_info */
        band_index = target_band_index_list[band_counter];

        /* read the target image (Grid) file */
        if (get_input_projection_information(get_grid_filename(), 
                &band_index, &target, &frame_type, &center_pt,
                &proj_center_pt, &rotate_angle, &orient_angle) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading projection information from: %s",
                    get_grid_filename());
            ias_geo_destroy_proj_transformation(target2source);
            free(source_ptr);
            ias_geom_grid_free(&band_number, &geom_band_count, &geom_grid);
            return EXIT_FAILURE;
        }

        /* set the target_band_list, in case it changed in the call to
           get_input_projection_information */
        target_band_index_list[band_counter] = band_index;

        target_band_num = ias_sat_attr_convert_band_index_to_number(band_index);
        if (target_band_num == ERROR)
        {
            IAS_LOG_ERROR("Unable to convert band_index: %i to a band number",
                band_index);
        }

        IAS_LOG_INFO("Using band number %d from the TARGET_FRAME_FILE",
                        target_band_num);

        /* fill in the geometric grid data that doesn't change for each band
           if on the first band */
        if (band_counter == 0)
        {
            /* Declare target image corners for all 4 corners */
            IAS_DBL_XY target_line_sample[4];
            int corner_index;
            double sample_min;
            double sample_max;
            double line_min;
            double line_max;

            /* we'll get the units upfront; this information will be needed
               when determining if we are crossing the 180 degree boundary */

            if (ias_geo_get_units(target.projection_units, &target_units)
                != SUCCESS)   
            {
                IAS_LOG_ERROR("Parsing the target units from: %s",
                       target.projection_units);
                free(source_ptr);
                ias_geom_grid_free(&band_number, &geom_band_count, &geom_grid);
                return EXIT_FAILURE;
            }

            if (ias_geo_get_units(source_ptr[band_counter].projection_units,
                &source_units) != SUCCESS)
            {
                IAS_LOG_ERROR("Parsing the source units from: %s",
                       source_ptr[band_counter].projection_units);
                free(source_ptr);
                ias_geom_grid_free(&band_number, &geom_band_count, &geom_grid);
                return EXIT_FAILURE;
            }
  
            if ((target_units == DMS) || (source_units == DMS))
            {
                IAS_LOG_ERROR("Unable to properly handle DMS angles.");
                free(source_ptr);
                ias_geom_grid_free(&band_number, &geom_band_count, &geom_grid);
                return EXIT_FAILURE;
            }

            /* make ready to calculate the polynomial coefficients for the
               source image projection coordinates to (line,sample)
               transformation */
            source_proj_y[0] = source_ptr[band_counter].upper_left_y;
            source_proj_y[1] = source_ptr[band_counter].upper_right_y;
            source_proj_y[2] = source_ptr[band_counter].lower_left_y;
            source_proj_y[3] = source_ptr[band_counter].lower_right_y;  
            source_proj_x[0] = source_ptr[band_counter].upper_left_x;
            source_proj_x[1] = source_ptr[band_counter].upper_right_x;
            source_proj_x[2] = source_ptr[band_counter].lower_left_x;
            source_proj_x[3] = source_ptr[band_counter].lower_right_x;
            source_image_x_max = source_ptr[band_counter].samples - 1;
            source_image_y_max = source_ptr[band_counter].lines - 1;
            image_proj_y[0]  = 0; /* upper left line */
            image_proj_y[1]  = 0; /* upper right line */
            image_proj_y[2]  = source_image_y_max; /* lower left line */
            image_proj_y[3]  = source_image_y_max; /* lower right line */
            image_proj_x[0]  = 0; /* upper left sample */ 
            image_proj_x[1]  = source_image_x_max; /* upper right samp*/
            image_proj_x[2]  = 0; /* lower left sample */
            image_proj_x[3]  = source_image_x_max; /* lower right samp*/
    
            /* check if input scene crosses the 180 degree boundary */
            source_crosses_180 = ias_geo_does_cross_180(source_units,
                source_proj_x);

            /* Set up the projection definitions for the target and source
               projections */
            ias_geo_set_projection(target.projection_code, target.zone_code,
                target_units, target.spheroid_code,
                target.projection_parameters, &target_proj);
            ias_geo_set_projection(source_ptr[band_counter].projection_code,
                source_ptr[band_counter].zone_code, source_units,
                source_ptr[band_counter].spheroid_code,
                source_ptr[band_counter].projection_parameters, &source_proj);

            /* Create a projection transformation for going from the target
               to the source */
            target2source = ias_geo_create_proj_transformation(&target_proj,
                &source_proj);
            if (!target2source)
            {
                IAS_LOG_ERROR("Creating target to source projection "
                    "transformation");
                free(source_ptr);
                ias_geom_grid_free(&band_number, &geom_band_count, &geom_grid);
                return EXIT_FAILURE;
            }

            /* Convert the corners of the target image from the projected
               coordinates specified by target to those specified by
               source_ptr[band_counter].  */
            if (ias_geo_transform_coordinate(target2source,
                    target.upper_left_x, target.upper_left_y,
                    &(target_proj_x[0]), &(target_proj_y[0])) != SUCCESS)
            {
                IAS_LOG_ERROR("Transforming projection for upper left");
                ias_geo_destroy_proj_transformation(target2source);
                free(source_ptr);
                ias_geom_grid_free(&band_number, &geom_band_count, &geom_grid);
                return EXIT_FAILURE;
            }

            if (ias_geo_transform_coordinate(target2source,
                    target.upper_right_x, target.upper_right_y,
                    &(target_proj_x[1]), &(target_proj_y[1])) != SUCCESS)
            {
                IAS_LOG_ERROR("Transforming projection for upper right");
                ias_geo_destroy_proj_transformation(target2source);
                free(source_ptr);
                ias_geom_grid_free(&band_number, &geom_band_count, &geom_grid);
                return EXIT_FAILURE;
            }

            if (ias_geo_transform_coordinate(target2source, target.lower_left_x,
                    target.lower_left_y, &(target_proj_x[2]),
                    &(target_proj_y[2])) != SUCCESS)
            {
                IAS_LOG_ERROR("Transforming projection for lower left");
                ias_geo_destroy_proj_transformation(target2source);
                free(source_ptr);
                ias_geom_grid_free(&band_number, &geom_band_count, &geom_grid);
                return EXIT_FAILURE;
            }

            if (ias_geo_transform_coordinate(target2source,
                    target.lower_right_x, target.lower_right_y,
                    &(target_proj_x[3]), &(target_proj_y[3])) != SUCCESS)
            {
                IAS_LOG_ERROR("Transforming projection for lower right");
                ias_geo_destroy_proj_transformation(target2source);
                free(source_ptr);
                ias_geom_grid_free(&band_number, &geom_band_count, &geom_grid);
                return EXIT_FAILURE;
            }
          
            /* check if output scene crosses the 180 degree boundary */
            target_crosses_180 = ias_geo_does_cross_180(source_units,
                target_proj_x);

            /* if corners are on opposite sides of the 180 degree line for
               target or source, we need to bring the points on the East side 
               back around to the positive */
            if ((source_crosses_180) || (target_crosses_180))
            {
                int corner;

                for (corner = 0; corner < 4; corner++)
                {
                    if (source_proj_x[corner] < 0)
                    {
                        ias_geo_add_once_around(source_units, 
                            &(source_proj_x[corner]));
                    }
                    if (target_proj_x[corner] < 0)
                    {
                        ias_geo_add_once_around(source_units,
                            &(target_proj_x[corner]));
                    }
                }
            }

            /* calculate the polynomial coefficients for the source image
               projection coordinates to (line,sample) transformation */
            if (ias_geo_compute_proj2proj_poly(source_proj_x,
                   source_proj_y, image_proj_x, image_proj_y, 
                   1 /* include x*y */,
                   proj_to_samp, proj_to_line) != SUCCESS)
            {
                IAS_LOG_ERROR("Calculating the transform polynomial " 
                              "coefficients");
                ias_geo_destroy_proj_transformation(target2source);
                free(source_ptr);
                ias_geom_grid_free(&band_number, &geom_band_count, &geom_grid);
                return EXIT_FAILURE;
            }

            /* convert the space projection of target image corners into
               line/sample */
            target_line_sample[0].x = ias_math_eval_poly_xy(1, proj_to_samp,
                                target_proj_x[0], target_proj_y[0]);
            target_line_sample[0].y = ias_math_eval_poly_xy(1, proj_to_line,
                                target_proj_x[0], target_proj_y[0]);
            target_line_sample[1].x = ias_math_eval_poly_xy(1, proj_to_samp,
                                target_proj_x[1], target_proj_y[1]);
            target_line_sample[1].y = ias_math_eval_poly_xy(1, proj_to_line,
                                target_proj_x[1], target_proj_y[1]);

            target_line_sample[2].x = ias_math_eval_poly_xy(1, proj_to_samp,
                                target_proj_x[2], target_proj_y[2]);
            target_line_sample[2].y = ias_math_eval_poly_xy(1, proj_to_line,
                                target_proj_x[2], target_proj_y[2]);
            target_line_sample[3].x = ias_math_eval_poly_xy(1, proj_to_samp,
                                target_proj_x[3], target_proj_y[3]);
            target_line_sample[3].y = ias_math_eval_poly_xy(1, proj_to_line,
                                target_proj_x[3], target_proj_y[3]);
 
            /* Find the min/max target line/sample */
            sample_max = sample_min = target_line_sample[0].x;
            line_max = line_min = target_line_sample[0].y;
            for (corner_index = 1; corner_index < 4; corner_index++)
            {
                if (target_line_sample[corner_index].x < sample_min)
                    sample_min = target_line_sample[corner_index].x;
                else if (target_line_sample[corner_index].x > sample_max)
                    sample_max = target_line_sample[corner_index].x;

                if (target_line_sample[corner_index].y < line_min)
                    line_min = target_line_sample[corner_index].y;
                else if (target_line_sample[corner_index].y > line_max)
                    line_max = target_line_sample[corner_index].y;
            }

            /* Clamp the min/max values to a valid range */
            if (sample_min < 0.0)
                sample_min = 0.0;
            if (sample_max > source_image_x_max)
                sample_max = source_image_x_max;
            if (line_min < 0.0)
                line_min = 0.0;
            if (line_max > source_image_y_max)
                line_max = source_image_y_max;

            /* Perform test to ensure at least one-pixel of overlap in 
               both the line and sample directions of 
               the source and target images.    */

            /* Check source and target coordinates to ensure that the
               target image overlaps the source image*/
            if (((target_line_sample[0].x > 1.0 && target_line_sample[0].x
                      < (source_image_x_max - 1.0))
                  && (target_line_sample[0].y > 1.0 && target_line_sample[0].y
                      < (source_image_y_max - 1.0)))
               || ((target_line_sample[1].x > 1.0 && target_line_sample[1].x
                      < (source_image_x_max - 1.0))
                  && (target_line_sample[1].y > 1.0 && target_line_sample[1].y
                      < (source_image_y_max - 1.0)))
               || ((target_line_sample[2].x > 1.0 && target_line_sample[2].x
                      < (source_image_x_max - 1.0))
                  && (target_line_sample[2].y > 1.0 && target_line_sample[2].y
                     < (source_image_y_max - 1.0)))
               || ((target_line_sample[3].x > 1.0 && target_line_sample[3].x
                      < (source_image_x_max - 1.0))
                  && (target_line_sample[3].y > 1.0 && target_line_sample[3].y
                      < (source_image_y_max - 1.0))))
            {
                /* At least one corner of the target falls inside the source,
                   so consider it OK, just issue a debug message */
                IAS_LOG_DEBUG("Source contains at least one target "
                              "file corner");
            }
            /* If at least one corner didn't fall inside the source, make sure
               there is at least some overlap */
            else if (((sample_max - sample_min) < 1.0)
                     || ((line_max - line_min) < 1.0)) 
            {
                /* No overlap, so something is wrong */
                IAS_LOG_ERROR("Disjoint source and target images");
                ias_geo_destroy_proj_transformation(target2source);
                free(source_ptr);
                ias_geom_grid_free(&band_number, &geom_band_count, &geom_grid);
                return EXIT_FAILURE;
            }

            geom_grid.spheroid = target.spheroid_code;
            geom_grid.zone = target.zone_code;
            geom_grid.code = target.projection_code;

            memset(geom_grid.gridname, 0, sizeof(geom_grid.gridname));
            strcpy(geom_grid.gridname,get_geom_grid_filename());
            memcpy(geom_grid.units, target.projection_units,
                   sizeof(geom_grid.units));
            memcpy(geom_grid.datum, target.datum,
                   sizeof(geom_grid.datum));

            for (count = 0; count < IAS_PROJ_PARAM_SIZE; count++)
            {
                geom_grid.projprms[count] =
                    target.projection_parameters[count];
            }
            geom_grid.corners.upleft.y = target.upper_left_y; 
            geom_grid.corners.upleft.x = target.upper_left_x;
            geom_grid.corners.loleft.y = target.lower_left_y;
            geom_grid.corners.loleft.x = target.lower_left_x;
            geom_grid.corners.upright.y = target.upper_right_y;
            geom_grid.corners.upright.x = target.upper_right_x;
            geom_grid.corners.loright.y = target.lower_right_y;
            geom_grid.corners.loright.x = target.lower_right_x;
            
            geom_grid.gridtype = REVERSE_GEOMETRIC_GRID;
            
            geom_grid.frame_type = frame_type;
            geom_grid.center_pt = center_pt;
            geom_grid.proj_center_pt = proj_center_pt;
            geom_grid.rotate_angle = rotate_angle;
            geom_grid.orient_angle = orient_angle;
        }
        
        /* use the source band for the band entered in the grid */
        grid_band = source_band_index_list[band_counter];

        /* add the band specific information to the grid */
        geom_grid.gridbands[grid_band].band_number = grid_band + 1;
        geom_grid.gridbands[grid_band].gridtype = REVERSE_GEOMETRIC_GRID;
        geom_grid.gridbands[grid_band].cell_lines = get_cell_lines();
        geom_grid.gridbands[grid_band].cell_samps = get_cell_samples();

        geom_grid.gridbands[grid_band].pixsize_line = 
            target.projection_distance_y;
        geom_grid.gridbands[grid_band].pixsize_samp = 
            target.projection_distance_x;

        geom_grid.gridbands[grid_band].output_image_lines = 
                                    target.lines;
        geom_grid.gridbands[grid_band].output_image_samps = 
                                    target.samples;
        geom_grid.gridbands[grid_band].num_out_lines = 
            (target.lines + get_cell_lines() - 1)
            /get_cell_lines() + 1;
        geom_grid.gridbands[grid_band].num_out_samps =
            (target.samples + get_cell_samples() - 1)
            /get_cell_samples() + 1;
        geom_grid.gridbands[grid_band].num_in_lines = 
            geom_grid.gridbands[grid_band].num_out_lines *
            geom_grid.gridbands[grid_band].num_out_samps;
        geom_grid.gridbands[grid_band].num_in_samps = 
            geom_grid.gridbands[grid_band].num_in_lines;
        geom_grid.gridbands[grid_band].grid_rows = 
            geom_grid.gridbands[grid_band].num_out_lines - 1;
        geom_grid.gridbands[grid_band].grid_cols =
            geom_grid.gridbands[grid_band].num_out_samps - 1;
        geom_grid.gridbands[grid_band].forward_present = 0;
        geom_grid.gridbands[grid_band].reverse_present = 1;
        geom_grid.gridbands[grid_band].degree = 0;
    }

    /* allocate memory for the geometric grid */
    band_number = grid_band + 1;
    geom_band_count = num_bands;

    if (ias_geom_grid_malloc(&band_number, &geom_band_count, &geom_grid)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Allocating memory for the geometric grid (band %d)",
                band_counter);
        ias_geo_destroy_proj_transformation(target2source);
        free(source_ptr);
        ias_geom_grid_free(&band_number, &geom_band_count, &geom_grid);
        return EXIT_FAILURE;
    }

    /* build the grid for each band */
    for (band_counter = 0; band_counter < num_bands; band_counter++)
    {
        /* build the geometric grid */
        if (build_geom_grid(&source_ptr[band_counter], target2source,
                            source_band_index_list[band_counter],
                            &geom_grid, source_crosses_180,
                            target_crosses_180) != SUCCESS)
        {
            IAS_LOG_ERROR("Building band %d of the geometric grid", 
                          band_counter);
            ias_geo_destroy_proj_transformation(target2source);
            free(source_ptr);
            ias_geom_grid_free(&band_number, &geom_band_count, &geom_grid);
            return EXIT_FAILURE;
        }

        /* calculate the coefficients for the geometric grid */
        if (calc_coeffs
            (&geom_grid.gridbands[source_band_index_list[band_counter]])
            != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating grid coefficients");
            ias_geo_destroy_proj_transformation(target2source);
            free(source_ptr);
            ias_geom_grid_free(&band_number, &geom_band_count, &geom_grid);
            return EXIT_FAILURE;
        }

        /* increment the number of bands in the grid */
        geom_grid.nbands++;

        /* set the band index in the bandlist member */
        geom_grid.band_index_list[band_counter] 
                     = source_band_index_list[band_counter];
    }

    /* Free the projection transformation */
    ias_geo_destroy_proj_transformation(target2source);
    target2source = NULL;

    /* write the geometric grid */
    if (ias_write_geom_grid(get_geom_grid_filename(), &geom_grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the geometric grid to %s",
                      get_geom_grid_filename());
        free(source_ptr);
        ias_geom_grid_free(&band_number, &geom_band_count, &geom_grid);
        return EXIT_FAILURE;
    }
   
    /* free the memory allocated to the geometric grid */
    ias_geom_grid_free(&band_number, &geom_band_count, &geom_grid);

    /* free the memory allocated to the source project information */
    free(source_ptr);

    /* Write the Geometric Grid filename into OMF */
    update_omf(get_omf_filename(), get_geom_grid_filename(), 
               get_processing_pass(), get_doq_image_flag());

    /* generate exit message */
    IAS_LOG_INFO("Successful Completion");

    return EXIT_SUCCESS;
}
