/*****************************************************************************
NAME: akima

PURPOSE: Evaluates the akima function for the input information.

RETURN VALUE:
    Returns the calculated value.

ALGORITHM REFERENCES:
    Refer to the ADD.

*****************************************************************************/

#include <math.h>

double akima
(
    double *y,      /* Grey Scale values for interpolation */
    double *x,      /* X Locations for pixels */
    double xp       /* X coordinate of interpolation point */
)
{
    double m[5];         /* Slopes associated with line segments */
    double p[4];         /* Polynomial coefficients for akima interpolant */
    double t1,t2;        /* Temporary variables */
    double x_1,x_2,x_3;  /* X, X squared, X cubed */
    double pix;
    double inv_x_diff;

    inv_x_diff = 1.0 / (x[3] - x[2]);

    /* Find slopes of line segments, assuming numerator is never zero */
    m[0] = (y[1] - y[0]) / (x[1] - x[0]);
    m[1] = (y[2] - y[1]) / (x[2] - x[1]);
    m[2] = (y[3] - y[2]) * inv_x_diff;
    m[3] = (y[4] - y[3]) / (x[4] - x[3]);
    m[4] = (y[5] - y[4]) / (x[5] - x[4]);

    /* Find slopes of points to left and right of interpolated point */
    if (m[0] == m[1] && m[2] == m[3])
    {
        /* Slope is undefined */
        t1 = 0.5 * (m[1] + m[2]);
    }
    else
    {
        double s1 = fabs(m[3] - m[2]);
        double s2 = fabs(m[1] - m[0]);
        t1 = (s1 * m[1] + s2 * m[2]) / (s1 + s2);
    }
    if (m[1] == m[2] && m[3] == m[4])
    {
        /* Slope is undefined */
        t2 = 0.5 * (m[2] + m[3]);
    }
    else
    {
        double s1 = fabs(m[4] - m[3]);
        double s2 = fabs(m[2] - m[1]);
        t2 = (s1 * m[2] + s2 * m[3]) / (s1 + s2);
    }

    /* Calculate polynomial coefficients */
    p[0] = y[2];
    p[1] = t1;
    p[2] = (3.0 * m[2] - 2.0 * t1 - t2) * inv_x_diff;
    p[3] = (t1 + t2 - 2.0 * m[2]) * inv_x_diff * inv_x_diff;

    /* calculate x^2 and x^3 for later use */
    x_1 = xp;
    x_2 = x_1 * x_1;
    x_3 = x_2 * x_1;

    /* Find interpolated value */
    pix = p[0] + p[1] * x_1 + p[2] * x_2 + p[3] * x_3;

    return pix;
}
