/*****************************************************************************
NAME: buffer_output_image

PURPOSE:
  The buffer_output_image module provides an interface to write an output
  L1G image as a series of tiles.  Some intelligent buffering is performed
  under the assumption that tiles will be written in the order from the top to
  the bottom of the image.  Writes are clustered in as large of blocks as
  possible.  This helps performance by minimizing how often the HDF library
  mutex needs to be held.

ROUTINES:
    open_output_image
    open_output_band
    write_output_image_data
    reset_output_buffer
    flush_output_buffer
    close_output_band
    close_output_image
    get_l1gio_ptr

*****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include "ias_l1g.h"
#include "ias_satellite_attributes.h"
#include "ias_logging.h"
#include "get_resample_parameters.h"
#include "track_tile.h"
#include "buffer_output_image.h"

/* This structure contains the state information for the output buffer */
struct output_buffer
{
    L1GIO *l1gio_ptr;         /* file descriptor for the file */
    int bytes_per_pixel;      /* bytes per pixel */
    int lines;                /* lines in the currently open band */
    int samples;              /* samples in the currently open band */
    void *buf;                /* pointer to the image buffer */
    int buffer_rows;          /* rows of tiles in allocated buffer */
    int current_row;          /* current start row of buffer */
    char filename[PATH_MAX];  /* name of image file */
    int output_start_line;    /* current band output window start line */
    int sca_index;            /* current SCA being written */
    IAS_THREAD_MUTEX_TYPE buffer_mutex;
                              /* mutex for access to the members of this 
                                 structure */
    L1G_BAND_IO *l1g_band;    /* L1G_BAND_IO pointer */
};

/* number of rows of output tile buffers to maintain to ensure the file
   is written to disk in a linear fashion so reads are relatively fast */
#define ROWS_TO_BUFFER 8

/*****************************************************************************
NAME: open_output_image

PURPOSE: open_output_image opens an image file for writing.

RETURNS:
    pointer to a OUTPUT_BUFFER structure if successful
    NULL if there is an error

*****************************************************************************/
OUTPUT_BUFFER *open_output_image
(
    const char *filename    /* I: name of image file */
)
{
    OUTPUT_BUFFER *img = NULL;

    /* allocate memory for the OUTPUT_BUFFER structure */
    img = malloc(sizeof(*img));
    if (img == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for the output buffer");
        return NULL;
    }
    memset(img, 0, sizeof(*img));

    /* save information in the OUTPUT_BUFFER structure for future use */
    strncpy(img->filename,filename,PATH_MAX);
    img->filename[PATH_MAX - 1] = '\0';

    /* create the mutex for the buffer access */
    if (IAS_THREAD_CREATE_MUTEX(&img->buffer_mutex) != 0)
    {
        IAS_LOG_ERROR("Error creating HDF file mutex");
        free(img);
        return NULL;
    }

    /* open the output file */
    img->l1gio_ptr = ias_l1g_open_image(filename, IAS_WRITE);
    if (img->l1gio_ptr == NULL)
    {
        IAS_LOG_ERROR("Error opening output file %s", filename);
        IAS_THREAD_DESTROY_MUTEX(&img->buffer_mutex);
        free(img);
        return NULL;
    }

    return img;
}

/*****************************************************************************
NAME: open_output_band

PURPOSE: open_output_band opens the requested band in the output file and
    prepares to write to it.

RETURNS:
    SUCCESS/ERROR

*****************************************************************************/
int open_output_band
(
    OUTPUT_BUFFER *img,     /* I: image to write */
    int band_index,         /* I: band index to start writing */
    int number_of_scas,     /* I: number of scas to include in the band */
    int lines,              /* I: number of lines in the band */
    int samples,            /* I: number of samples in the band */
    IAS_DATA_TYPE data_type /* I: data type for the output file */
)
{
    int band_number;
    long long buf_size;         /* size of buffer to allocate */

    /* make sure there isn't already a band open for writing since only a
       single band at a time is supported */
    if (img->l1g_band != NULL)
    {
        IAS_LOG_ERROR("Band already open for %s", img->filename);
        return ERROR;
    }

    /* get the band number for using the L1G IO library */
    band_number = ias_sat_attr_convert_band_index_to_number(band_index);
    if (band_number == ERROR)
    {
        IAS_LOG_ERROR("Invalid band index: %d", band_index);
        return ERROR;
    }

    /* open the correct band in the output file */
    img->l1g_band = ias_l1g_open_band(img->l1gio_ptr, band_number, 
                        &data_type, &number_of_scas, &lines, &samples);
    if (img->l1g_band == NULL)
    {
        IAS_LOG_ERROR("Failed to open band number %d in %s for writing",
                      band_number, img->filename);
        return ERROR;
    }

    img->lines = lines;
    img->samples = samples;

    /* determine the number of bytes per pixel */
    if (data_type == IAS_BYTE)
        img->bytes_per_pixel = 1;
    else if ((data_type == IAS_I2) || (data_type == IAS_UI2))
        img->bytes_per_pixel = 2;
    else if ((data_type == IAS_I4) || (data_type == IAS_UI4) 
             || (data_type == IAS_R4))
    {
        img->bytes_per_pixel = 4;
    }
    else
        img->bytes_per_pixel = 8;

    /* allocate memory for a cache of rows of tiles.  This is done since linux 
       apparently deals poorly with writing partial lines often.  When writing 
       tiles to the file, the file reads at about 10% of the capability of the 
       hard drive due to excessive seeking on disk.  It also allows minimizing
       the number of serialized calls to the HDF library. */
    buf_size = (long long)img->bytes_per_pixel * samples 
             * get_output_tile_size() * ROWS_TO_BUFFER;
    img->buf = malloc(buf_size);
    if (img->buf == NULL)
    {
        IAS_LOG_ERROR("Allocating image buffer memory");
        ias_l1g_close_band(img->l1g_band);
        img->l1g_band = NULL;
        return ERROR;
    }

    return SUCCESS;
}

/*****************************************************************************
NAME: reset_output_buffer

PURPOSE: reset_output_buffer resets the output buffer tracking for a new SCA.

RETURNS:
    nothing

*****************************************************************************/
void reset_output_buffer
(
    OUTPUT_BUFFER *img,     /* I: pointer to image structure */
    int sca_index           /* I: sca_index to write */
)
{
    img->buffer_rows = ROWS_TO_BUFFER;
    img->sca_index = sca_index;
    img->current_row = 0;
    img->output_start_line = 0;
}

/*****************************************************************************
NAME: write_output_image_data

PURPOSE: write_output_image_data writes a tile of imagery to the output image,
    caching data in an output buffer as needed to group writes together.

RETURNS:
    SUCCESS/ERROR

*****************************************************************************/
int write_output_image_data
(
    OUTPUT_BUFFER *img,     /* I: pointer to image structure */
    const TILE_TYPE *tile,  /* I: pointer to tile location */
    IAS_THREAD_MUTEX_TYPE *hdf_io_mutex, /* I: HDF I/O mutex */
    void *tile_buf          /* I: pointer to the tile to write */
)
{
    int line;
    const unsigned char *tile_buffer = (const unsigned char *)tile_buf;
    int bytes_per_img_line = img->bytes_per_pixel * img->samples;
    int tile_lines = get_output_tile_size();
    int row_index = (tile->start_line - img->output_start_line)/tile_lines;
    int status;

    /* lock the buffer mutex to prevent multiple routines from modifying 
       buffer tracking variables concurrently */
    if (IAS_THREAD_LOCK_MUTEX(&img->buffer_mutex) != 0)
    {
        IAS_LOG_ERROR("Error locking buffer mutex");
        return ERROR;
    }

    /* Possible optimization: The locking done in this routine could be
       damaging the parallelization in the resampler since all threads are
       taking the lock just to copy data to the buffer.  Meaning only one
       thread can be copying data at a time, even if no other buffer
       manipulation is happening.  It should be possible to use a read/write
       lock to speed things up.  The read lock would be taken to start with
       assuming it will only be copying data without any buffer manipulations.
       If it is found that some buffers need to be flushed, the write lock
       would then be taken, the buffers written and buffer tracking info
       updated, and the lock released. */

    /* Possible optimization: to make the buffering a bit more intelligent,
       each row could be made a separately allocated buffer with some tracking
       information to remember how many tiles have been copied into it.  Once
       the last tile has been copied to it, the write to the file can occur.
       The only lock, other than the HDF lock, could be a quick lock to
       decrement the count of tiles in each row (probably can get by with a
       single lock for all rows).  The thread that sets the remaining tile
       count for a row to zero can write the row with just the HDF lock taken
       as long as the tile count isn't decremented until after the data is
       copied to the row.  The row buffer could also be freed at that time, or
       recycled for another row. */

    /* flush the rows until the tile just received will fit in one.  Note that
       this could be inefficient if a very strange pattern of tiles returned
       happens (i.e. one thread is not returning tiles while some
       are returning them rapidly - empty or partially filled buffers could be 
       getting flushed to disk.  However this should be rare.) */
    while (row_index >= (img->current_row + img->buffer_rows))
    {
        /* Possible optimization: to allow minimizing the calls into the HDF
           library, could try to group multiple rows together in a single
           write.  Probably would still need to loop to handle the case where
           the buffer wraps around */

        /* lock access to the HDF library since it isn't thread-safe */
        if (IAS_THREAD_LOCK_MUTEX(hdf_io_mutex) != 0)
        {
            IAS_LOG_ERROR("Error locking HDF I/O mutex");
            IAS_THREAD_UNLOCK_MUTEX(&img->buffer_mutex);
            return ERROR;
        }

        status = ias_l1g_write_image(img->l1g_band, img->sca_index,
                img->current_row * tile_lines, 0, tile_lines, img->samples,
                &((unsigned char *)img->buf)[(img->current_row % 
                  img->buffer_rows) * tile_lines * bytes_per_img_line]);

        /* allow other threads to have access to the HDF library */
        if (IAS_THREAD_UNLOCK_MUTEX(hdf_io_mutex) != 0)
        {
            IAS_LOG_ERROR("Error unlocking HDF I/O mutex");
            IAS_THREAD_UNLOCK_MUTEX(&img->buffer_mutex);
            return ERROR;
        }

        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Error writing lines to SCA %d, start line %d, "
                    "number of lines %d", img->sca_index,
                    img->current_row * tile_lines, tile_lines);
            IAS_THREAD_UNLOCK_MUTEX(&img->buffer_mutex);
            return ERROR;
        }

        img->current_row++;
    }

    /* if the tile is for a row that has already been flushed to disk, write
       just this tile.  This occurring should be very rare. */
    if (row_index < img->current_row)
    {
        /* warn when a lagging tile is written - really debug for now */
        IAS_LOG_WARNING("Lagging tile written for row %d at %d,%d\n", row_index,
                tile->start_line, tile->start_sample);

        /* lock access to the HDF library since it isn't thread-safe */
        if (IAS_THREAD_LOCK_MUTEX(hdf_io_mutex) != 0)
        {
            IAS_LOG_ERROR("Error locking HDF I/O mutex");
            IAS_THREAD_UNLOCK_MUTEX(&img->buffer_mutex);
            return ERROR;
        }

        status =  ias_l1g_write_image(img->l1g_band, img->sca_index,
                    (tile->start_line - img->output_start_line),
                    tile->start_sample, tile->num_lines,
                    tile->num_samples, tile_buf);

        /* allow other threads to have access to the HDF library */
        if (IAS_THREAD_UNLOCK_MUTEX(hdf_io_mutex) != 0)
        {
            IAS_LOG_ERROR("Error unlocking HDF I/O mutex");
            IAS_THREAD_UNLOCK_MUTEX(&img->buffer_mutex);
            return ERROR;
        }

        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Write to %s failed", img->filename);
            IAS_THREAD_UNLOCK_MUTEX(&img->buffer_mutex);
            return ERROR;
        }
    }
    else if ((row_index >= img->current_row) && 
        (row_index < (img->current_row + img->buffer_rows)))
    {
        /* the tile is in the currently buffered rows, so just copy the data
           to the buffer */
        int buf_index = (row_index % img->buffer_rows) * bytes_per_img_line
                      * tile_lines + tile->start_sample * img->bytes_per_pixel;
        int bytes_per_tile_line = img->bytes_per_pixel * tile->num_samples;

        /* copy the tile to the row buffers */
        for (line = 0; line < tile->num_lines; line++)
        {
            memcpy(&((unsigned char *)img->buf)[buf_index],
                &tile_buffer[line * bytes_per_tile_line], bytes_per_tile_line);

            buf_index += bytes_per_img_line;
        }
    }
    else
    {
        /* This is an error.  The while loop at the top of the routine 
           should make it impossible for this code to be reached. */
        IAS_LOG_ERROR("Error - unexpected output tile");
        IAS_THREAD_UNLOCK_MUTEX(&img->buffer_mutex);
        return ERROR;
    }

    /* release the buffer mutex */
    if (IAS_THREAD_UNLOCK_MUTEX(&img->buffer_mutex) != 0)
    {
        IAS_LOG_ERROR("Error unlocking HDF file io mutex");
        return ERROR;
    }

    return SUCCESS;
}

/*****************************************************************************
NAME: flush_output_buffer

PURPOSE: flush_output_buffer flushes any remaining buffered data for the 
    current band to disk.

RETURNS:
    SUCCESS/ERROR

NOTES:
    - No locking is done here since it is assumed that no other threads will
      be active when this is called.

*****************************************************************************/
int flush_output_buffer
(
    OUTPUT_BUFFER *img  /* I: pointer to image structure */
)
{
    int bytes_per_img_line = img->bytes_per_pixel * img->samples;
    int tile_lines = get_output_tile_size();
    int rows = (img->lines + tile_lines - 1)/tile_lines;
    int row;

    /* flush the remaining buffers to disk */
    for (row = img->current_row; row < rows; row++)
    {
        int lines_to_write;
        int status;
        
        if (row == (rows - 1))
        {
            /* last row, so maybe not a full set of lines */
            lines_to_write = img->lines - row * tile_lines;
        }
        else
        {
            lines_to_write = tile_lines;
        }

        status =  ias_l1g_write_image(img->l1g_band, img->sca_index,
                        row * tile_lines, 0, lines_to_write, img->samples,
                        &((unsigned char *)img->buf)[(row % img->buffer_rows) 
                        * tile_lines * bytes_per_img_line]);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Error writing to output file");
            return ERROR;
        }
    }

    return SUCCESS;
}

/*****************************************************************************
NAME: close_output_band

PURPOSE: close_output_band closes the current output band.

RETURNS:
    SUCCESS/ERROR

*****************************************************************************/
int close_output_band
(
    OUTPUT_BUFFER *img  /* I: pointer to image to close */
)
{
    /* free the buffer since the next band may need a different sized buffer */
    free(img->buf);
    img->buf = NULL;

    /* close the current band */
    if (ias_l1g_close_band(img->l1g_band) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing output band");
        return ERROR;
    }
    img->l1g_band = NULL;

    return SUCCESS;
}

/*****************************************************************************
NAME: close_output_image

PURPOSE: close_output_image closes an image opened with open_output_image

RETURNS:
    SUCCESS/ERROR

*****************************************************************************/
int close_output_image
(
    OUTPUT_BUFFER *img  /* I: pointer to image to close */
)
{
    int status;

    status = ias_l1g_close_image(img->l1gio_ptr);
    IAS_THREAD_DESTROY_MUTEX(&img->buffer_mutex);
    free(img->buf);
    free(img);

    return status;
}

/*****************************************************************************

NAME: get_l1gio_ptr

PURPOSE: To return the L1GIO pointer so that it can be used outside of this
         library.

RETURNS: Pointer to the L1GIO structure.

*****************************************************************************/
L1GIO *get_l1gio_ptr
(
    OUTPUT_BUFFER *img  /* I: pointer to image structure */
)
{
    return img->l1gio_ptr;
}

