/*****************************************************************************
NAME: get_cpf_parameters

PURPOSE:  This module reads the CPF file and grabs parameters required by
          the resampler.

ROUTINES:
    read_cpf_parameters
    get_radiance_rescale_parms
*****************************************************************************/

#include <stdlib.h>
#include "ias_logging.h"
#include "ias_cpf.h"
#include "get_cpf_parameters.h"  /* prototypes for this module */

/* VARIABLE DECLARATIONS */

static struct IAS_CPF_RADIANCE_RESCALE local_radiance_rescale_parms;
static const struct IAS_CPF_EARTH_CONSTANTS *cpf_earth = NULL;
static IAS_CPF *cpf = NULL;

/*****************************************************************************
NAME: read_cpf_parameters

PURPOSE: This function reads all the parameters from the CPF that the
         resampler requires.

RETURN VALUE: Type = int
    Value    Description
    -----    -----------
    SUCCESS  The parameters were successfully read
    ERROR    An error occurred reading the parameters
*****************************************************************************/
int read_cpf_parameters
(
    const char *cpf_filename,      /* I: CPF filename */
    int radiance_rescale_flag      /* I: Flag to signal reading cpf
                                          for rescale factors */
)
{
    int index;
    const struct IAS_CPF_RADIANCE_RESCALE *radiance_rescale = NULL;

    /* Read the cpf */
    cpf = ias_cpf_read(cpf_filename);
    if (!cpf)
    {
        IAS_LOG_ERROR("Opening the CPF %s", cpf_filename);
        return ERROR;
    }

    /* Retrieve the cpf constants */
    cpf_earth = ias_cpf_get_earth_const(cpf);
    if (!cpf_earth)
    {
        IAS_LOG_ERROR("Retrieving Earth constants group from the CPF");
        return ERROR;
    }

    /* Open the CPF */
    if (radiance_rescale_flag)
    {
        /* Get the radiance rescale parameters from the CPF */
        radiance_rescale = ias_cpf_get_radiance_rescale(cpf);
        if (!radiance_rescale)
        {
            IAS_LOG_ERROR("Reading radiance rescale parameters from the CPF");
            ias_cpf_free(cpf);
            return ERROR;
        }

        /* Store the radiance rescale parameters in the local variable */
        for(index = 0; index < IAS_MAX_NBANDS; index++)
        {
            local_radiance_rescale_parms.gain[index] =
                radiance_rescale->gain[index];
            local_radiance_rescale_parms.bias[index] =
                radiance_rescale->bias[index];
        }
    }
    else
    {
        for(index = 0; index < IAS_MAX_NBANDS; index++)
        {
            local_radiance_rescale_parms.gain[index] = 1.0;
            local_radiance_rescale_parms.bias[index] = 0.0;
        }
    }

    return SUCCESS;
}

/******************************************************************************
NAME: various accessor functions

PURPOSE: These functions provide access to the parameters read. 

RETURN VALUE: Varies for the function.

NOTES:
    The accessor function method of providing parameters was selected since
    passing all the parameters up and down a calling sequence is a 
    maintenance nightmare.
******************************************************************************/

double get_radiance_rescale_gain(int band_index)
{
    return local_radiance_rescale_parms.gain[band_index];
}

double get_radiance_rescale_bias(int band_index)
{
    return local_radiance_rescale_parms.bias[band_index];
}

void free_resample_cpf()
{   
    /* Free the CPF */
    ias_cpf_free(cpf);
    cpf = NULL;
    cpf_earth = NULL;
}

const struct IAS_CPF_EARTH_CONSTANTS *get_cpf_earth_constants()
{
    /* Check that the CPF has been read */
    if (!cpf)
    {
        IAS_LOG_ERROR("CPF has not been initialized");
        return NULL;
    }

    return cpf_earth;
}
