/****************************************************************************
NAME: get_input_image

PURPOSE: This module implements an intermediate layer to open the input image.
    The input image can be either an L1R or an L0R image.  This set of routines
    hides the details about which type of file is being read from the routines
    that just care about reading imagery.

NOTES:
    - Performing all the image IO through this layer also allows for easily
      modifying how the input image data is buffered.  Currently, each tile
      is read from the input file using the respective image I/O library.
      Due to the underlying HDF library calls needing to be serialized, this
      could become a bottleneck to the processing.  If needed, some extra
      caching logic can be added here to try to minimize how often the image IO
      routines are called.
*****************************************************************************/
#include <stdlib.h>
#include <string.h>         /* for memset */
#include "ias_logging.h"
#include "ias_types.h"
#include "ias_const.h"
#include "ias_satellite_attributes.h"
#include "get_input_image.h"
#include "get_cpf_parameters.h"
#include "get_resample_parameters.h"

/* Local Defines */
#define BASE_RESOLUTION 30.0

/* local versions of the radiance rescaling factors to apply. */
static float scaling_multiplier = 1.0;
static float scaling_offset = 0.0;
static float qcal_min;
static float qcal_max;

/******************************************************************************
NAME: apply_scaling_factors

PURPOSE:
    Apply the radiance rescaling factors to the input value.  The values
    are not checked against QCALMIN or QCALMAX since that needs to be done
    after the resampling process combines overlapped SCAs.

RETURN VALUE:
    The input value scaled by the scaling factors.
******************************************************************************/
static inline float apply_scaling_factors
(
    float input_value /* I: The value to apply scaling to */
)
{
    float output_value;

    output_value = (input_value - scaling_offset) * scaling_multiplier;

    return output_value;
}

/******************************************************************************
NAME: wrapped_apply_scaling_factors

PURPOSE:
    Wrapper for the test code.  Since the real function is declared inline
    for speed reasons within the core resampling.

RETURN VALUE:
    The input value scaled by the scaling factors.
******************************************************************************/
float wrapped_apply_scaling_factors
(
    float input_value /* The value to apply scaling to */
)
{
    float output_value = apply_scaling_factors(input_value);

    /* The QCALMIN and QCALMAX checks are needed since complete processing
       through the resampler is not performed. */
    if (output_value < qcal_min)
    {
        output_value = qcal_min;
    }
    else if (output_value > qcal_max)
    {
        output_value = qcal_max;
    }

    return output_value;
}

/******************************************************************************
NAME: setup_qcal_min_max

PURPOSE:
    Set up the qcal min/max values for the data conversion

RETURN VALUE:
    SUCCESS or ERROR
******************************************************************************/
int setup_qcal_min_max
(
    int band_number         /* I: band number to set up */
)
{
    int quant_value;

    /* Set the qcal min/max values, converting them to floats since that
       is most efficient for applying them */
    if (ias_sat_attr_get_quantization_cal_min(band_number, &quant_value)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Error getting the QCAL min value for band number %d",
            band_number);
        return ERROR;
    }
    qcal_min = quant_value;

    if (ias_sat_attr_get_quantization_cal_max(band_number, &quant_value)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Error getting the QCAL max value for band number %d",
            band_number);
        return ERROR;
    }
    qcal_max = quant_value;

    return SUCCESS;
}

/*****************************************************************************
NAME: open_input_image

PURPOSE: Determines whether the input file is an L1R or L0R and opens it
    appropriately.

RETURNS:
    NULL on an error
    pointer to an INPUT_IMAGE structure if it succeeds

*****************************************************************************/
INPUT_IMAGE *open_input_image
(
    const char *l1r_filename,   /* I: L1R filename to open */
    const char *l0r_filename    /* I: L0R filename to open if the L1R file name
                                      isn't provided. */
)
{
    INPUT_IMAGE *input_image;
    const char *filename;

    /* if the L1R filename is null or blank, use the L0R instead */
    if ((l1r_filename == NULL) || (l1r_filename[0] == '\0'))
    {
        /* no L1R filename provided so assume there is an L0R name provided */
        filename = l0r_filename;
    }
    else
    {
        /* use the L1R filename since it was provided */
        filename = l1r_filename;
    }

    input_image = malloc(sizeof(*input_image));
    if (!input_image)
    {
        IAS_LOG_ERROR("Allocating memory for the input image structure");
        return NULL;
    }

    /* initialize the input image structure */
    memset(input_image, 0, sizeof(*input_image));
    input_image->band_number = -1;
    input_image->io_lock = NULL;

    /* determine whether the filename points to a L1R or L0R image (assumes it
       is an L0R if the L1R test fails) */
    if (ias_l1r_is_l1r_file(filename))
        input_image->is_l1r = 1;
    else
        input_image->is_l1r = 0;

    if (input_image->is_l1r)
    {
        /* open the L1R image */
        input_image->l1r_file = ias_l1r_open_image(filename, IAS_READ);
        if (!input_image->l1r_file)
        {
            IAS_LOG_ERROR("Error opening L1R file %s for reading", filename);
            free(input_image);
            return NULL;
        }
    }
    else
    {
        /* open the L0R image */
        input_image->l0r_file = ias_l0r_open_from_filename(filename, 
                                    IAS_L0R_COMPRESSION_DEFAULT);
        if (!input_image->l0r_file)
        {
            IAS_LOG_ERROR("Error opening L0R file %s for reading", filename);
            free(input_image);
            return NULL;
        }
    }

    return input_image;
}

/*****************************************************************************
NAME: set_input_image_io_lock

PURPOSE: Sets the mutex to use for locking access to the file.  Note that this
    is set from an external source since the underlying HDF5 library requires
    access to the library to be serialized.  So, all open files need to use the
    same lock.

RETURNS:
    SUCCESS/ERROR

NOTES:
    This lock is only used in the read_input_image_data routine.  All other
    routines are assumed to be called from a non-threaded context.

*****************************************************************************/
void set_input_image_io_lock
(
    INPUT_IMAGE *image,             /* I: input image pointer */
    IAS_THREAD_MUTEX_TYPE *io_lock  /* I: pointer to the mutex to use for 
                                          serializing access to the file */
)
{
    image->io_lock = io_lock;
}

/*****************************************************************************
NAME: open_input_band

PURPOSE: Opens the correct band for reading.

RETURNS:
    SUCCESS/ERROR

*****************************************************************************/
int open_input_band
(
    INPUT_IMAGE *image,     /* I: input image pointer */
    int band_number         /* I: band number to open */
)
{
    /* make sure there isn't already an open band */
    if (image->band_number != -1)
    {
        IAS_LOG_ERROR("Opening a band when one is already open");
        return ERROR;
    }

    if (setup_qcal_min_max(band_number) != SUCCESS)
    {
        IAS_LOG_ERROR("Error getting the QCAL min/max values for band "
            "number %d", band_number);
        return ERROR;
    }

    if (image->is_l1r)
    {
        IAS_DATA_TYPE data_type;

        /* open the L1R band */
        image->l1r_band = ias_l1r_open_band(image->l1r_file, band_number,
            &data_type, &image->scas, &image->nl, &image->ns);
        if (!image->l1r_band)
        {
            IAS_LOG_ERROR("Error opening band number %d of L1R image",
                          band_number);
            return ERROR;
        }

        /* make sure the datatype is floating point */
        if (data_type != IAS_R4)
        {
            IAS_LOG_ERROR("L1R image data type is not floating point");
            ias_l1r_close_band(image->l1r_band);
            image->l1r_band = NULL;
            return ERROR;
        }
    }
    else
    {
        IAS_L0R_BAND_ATTRIBUTES band_attributes;

        /* open the L0R band */
        if (ias_l0r_open_band(image->l0r_file, band_number, IAS_READ)
            != SUCCESS)
        {
            IAS_LOG_ERROR("Error opening band number %d of L0R image",
                          band_number);
            return ERROR;
        }

        /* get the band size information */
        if (ias_l0r_get_band_attributes(image->l0r_file, band_number, 
                &band_attributes) != SUCCESS)
        {
            IAS_LOG_ERROR("Error getting L0R attributes for band number %d",
                          band_number);
            ias_l0r_close_band(image->l0r_file, band_number);
            return ERROR;
        }

        image->scas = band_attributes.sca_count;
        image->nl = band_attributes.number_of_lines;
        image->ns = band_attributes.detectors_per_sca;
    }
    
    /* Check if buffer message has been issued */
    if (!image->buffer_message_issued)
    {
        BUFFER_TYPE buffer_type; /* Buffer type parameter */

        /* Change the messaged issued to true */
        image->buffer_message_issued = TRUE;
        
        /* Check if the band buffering is not set to default */
        buffer_type = get_buffer_type();
        if (buffer_type != DEFAULT_BUFFER)
        {
            if (buffer_type == FULL_BAND_BUFFER)
            {
                IAS_LOG_INFO("Band buffering was overidden to use Full Band "
                    "buffering");
                image->read_whole_band = TRUE;
            }
            else
            {
                IAS_LOG_INFO("Band buffering was overidden to use Tile "
                    "buffering");
                image->read_whole_band = FALSE;
            }
        }
        else
        {
            const IAS_BAND_ATTRIBUTES *attributes; /* Current Band attributes */
            int scaled_threshold; /* Scaled threshold per image resolution */

            attributes = ias_sat_attr_get_band_attributes(band_number);
            scaled_threshold = (int)(BASE_RESOLUTION 
                / attributes->pixel_resolution 
                * get_input_image_buffer_threshold() - ((BASE_RESOLUTION 
                - attributes->pixel_resolution) / attributes->pixel_resolution) 
                + 0.5);

            /* Check if the entire band should be read at a time */
            image->read_whole_band = TRUE;
            if (image->nl >= scaled_threshold)
            {
                IAS_LOG_INFO("Not using full band buffering, %d lines found "
                    "in Band %d were above the threshold of %d lines", 
                    image->nl, band_number, scaled_threshold);
                image->read_whole_band = FALSE;
            }
        }
    }

    /* Read entire band if band is below the threshold */
    if (image->read_whole_band)
    {
        /* now that the band has been opened, read the SCAs into buffers. */
        int sca_index;
        int image_pixel_count = image->nl * image->ns;
        size_t datatype_size; /* Band datatype size */

        /* Determine the allocation size */
        datatype_size = sizeof(float);
        if (!image->is_l1r)
        {
            datatype_size = sizeof(uint16_t);
        }

        for (sca_index = 0; sca_index < image->scas; sca_index++)
        {
            void *output;       /* Band pointer */
            int index;          /* Loop variable */

            /* allocate memory for the SCA buffer */
            image->band_data[sca_index] = malloc(image_pixel_count 
                * datatype_size);           
            if (!image->band_data[sca_index])
            {
                IAS_LOG_ERROR("Error allocating band_data buffer");
                for (index = 0; index < sca_index; index++)
                {
                    free(image->band_data[index]);
                    image->band_data[index] = NULL;
                }
                return ERROR;
            }

            /* assign to a pointer to be used by the conversions */
            output = image->band_data[sca_index];

            /* read the L1R or L0R data into the buffer */
            if (image->is_l1r)
            {
                if (ias_l1r_read_image(image->l1r_band, sca_index, 0, 0,
                        image->nl, image->ns, output) != SUCCESS)
                {
                    IAS_LOG_ERROR("Error reading input file for SCA index %d, "
                                  "start line %d, start sample %d, number of "
                                  "lines %d, number of samples %d",
                                  sca_index, 0, 0, image->nl, image->ns);
                    for (index = 0; index <= sca_index; index++)
                    {
                        free(image->band_data[index]);
                        image->band_data[index] = NULL;
                    }
                    ias_l1r_close_band(image->l1r_band);
                    image->l1r_band = NULL;
                    return ERROR;
                }
            }
            else
            {
                /* read the L0R image into the output buffer as uint16_t where
                   it is later converted to float in-place */
                if (ias_l0r_get_band_lines_sca(image->l0r_file, band_number,
                    sca_index, 0, 0, image->nl, image->ns, output)
                    != SUCCESS)
                {
                    IAS_LOG_ERROR("Error reading input file for SCA index %d, "
                                  "start line %d, start sample %d, number of "
                                  "lines %d, number of samples %d", sca_index,
                                  0, 0, image->nl, image->ns);
                    for (index = 0; index <= sca_index; index++)
                    {
                        free(image->band_data[index]);
                        image->band_data[index] = NULL;
                    }
                    ias_l0r_close_band(image->l0r_file, band_number);
                    return ERROR;
                }
            }
        }
    }

    /* save the open band number */
    image->band_number = band_number;

    return SUCCESS;
}

/*****************************************************************************
NAME: read_input_image_data

PURPOSE: Reads the requested tile from the indicated SCA index.  If it is
    reading an L0R image, it also converts the 16-bit int data to float.

RETURNS:
    SUCCESS/ERROR

*****************************************************************************/
int read_input_image_data
(
    INPUT_IMAGE *image,         /* I: input image pointer */
    int sca_index,              /* I: SCA index to read */
    const TILE_TYPE *tile,      /* I: specification of the tile to read */
    float *tile_buffer          /* I/O: pointer to a buffer for the read data */
)
{
    int index;      /* current index being converted */
    int size = tile->num_lines * tile->num_samples; /* pixel count to convert */
    float *output;  /* pointer to the output data during conversion */

    /* Copy the data from the band buffer into the tile buffer while
     * applying scaling factors and doing conversions during the copy */
    if (image->read_whole_band)
    {
        int start_line_index = tile->start_line * image->ns 
            + tile->start_sample; /* Start index */
        int sample; /* Sample loop variable */
        int line;   /* Line loop variable */

        if (image->is_l1r)
        { 
            float *sca = (float *)image->band_data[sca_index];
            float *input_data;
                
            /* Copy the data while applying the scaling factors */
            for (line = 0; line < tile->num_lines; line++)
            {
                input_data = &sca[start_line_index + line * image->ns];
                output = &tile_buffer[line * tile->num_samples];

                for (sample = 0; sample < tile->num_samples; sample++)
                {
                    output[sample] = apply_scaling_factors(input_data[sample]);
                }
            }
        }
        else
        {
            uint16_t *sca = (uint16_t *)image->band_data[sca_index];
            uint16_t *input_data;
                            
            /* Convert the data right in the current buffer.  This assumes 
               the tile buffer is sized for floats.  The conversion is done
               from the end of the buffer to the beginning of the same  
               buffer to prevent needing a secondary buffer. */
            for (line = tile->num_lines - 1; line >= 0 ; line--)
            {
                input_data = &sca[start_line_index + line * image->ns];
                output = &tile_buffer[line * tile->num_samples];

                for (sample = tile->num_samples - 1; sample >= 0; sample--)
                {
                    output[sample] = input_data[sample];
                }
            }
        }
    }
    else
    {
        int status; /* Status placeholder */
        output = (float *)tile_buffer;

        /* lock the IO lock */
        if (IAS_THREAD_LOCK_MUTEX(image->io_lock) != SUCCESS)
        {
            IAS_LOG_ERROR("Error locking input image I/O mutex");
            return ERROR;
        }

        /* Read the image into the input buffer */
        if (image->is_l1r)
        {
            status = ias_l1r_read_image(image->l1r_band, sca_index, 
                tile->start_line, tile->start_sample, tile->num_lines, 
                tile->num_samples, tile_buffer);
         }
        else
        {
            status = ias_l0r_get_band_lines_sca(image->l0r_file, 
                    image->band_number, sca_index, tile->start_line, 
                    tile->start_sample, tile->num_lines, tile->num_samples, 
                    (uint16_t *)tile_buffer);
        }
        
        /* Check if the image was read successfully */
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Error reading input file for SCA index %d, start"
                " line %d, start sample %d, number of lines %d, number of"
                " samples %d", sca_index, tile->start_line, 
                tile->start_sample, tile->num_lines, tile->num_samples);
            IAS_THREAD_UNLOCK_MUTEX(image->io_lock);
            return ERROR;
        }

        /* unlock the IO lock */
        if (IAS_THREAD_UNLOCK_MUTEX(image->io_lock) != SUCCESS)
        {
            IAS_LOG_ERROR("Error unlocking I/O mutex");
            return ERROR;
        }

        if (image->is_l1r)
        {
            /* Apply the scaling factors here */
            for (index = 0; index < size; index++)
            {
                output[index] = apply_scaling_factors(output[index]);
            }
        }
        else
        {
            /* convert the data right in the current buffer.  This assumes 
               the tile buffer is sized for floats.  The conversion is done
               from the end of the buffer to the beginning of the same  
               buffer to prevent needing a secondary buffer. */
            uint16_t *input = (uint16_t *)tile_buffer;
            for (index = size - 1; index >= 0; index--)
            {
                /* convert the data from short int to float */
                output[index] = input[index];
            }
        }
    }

    return SUCCESS;
}

/*****************************************************************************
NAME: close_input_band

PURPOSE: Closes the currently open band in the input image.

RETURNS:
    SUCCESS/ERROR

*****************************************************************************/
int close_input_band
(
    INPUT_IMAGE *image      /* I: input image pointer */
)
{
    /* make sure there is actually an open band */
    if (image->band_number == -1)
    {
        IAS_LOG_ERROR("Closing band when one is not open");
        return ERROR;
    }

    if (image->read_whole_band)
    {
        /* free the memory allocated for the input image */
        int sca_index;
        for (sca_index = 0; sca_index< image->scas; sca_index++)
        {
            free(image->band_data[sca_index]);
            image->band_data[sca_index] = NULL;
        }
    }

    if (image->is_l1r)
    {
        /* close the L1R band */
        if (ias_l1r_close_band(image->l1r_band) != SUCCESS)
        {
            IAS_LOG_ERROR("Error closing band number %d of L1R image",
                          image->band_number);
            return ERROR;
        }
        image->l1r_band = NULL;
    }
    else
    {
        /* close the L0R band */
        if (ias_l0r_close_band(image->l0r_file, image->band_number) != SUCCESS)
        {
            IAS_LOG_ERROR("Error closing band number %d of L0R image",
                          image->band_number);
            return ERROR;
        }
    }
    image->band_number = -1;

    return SUCCESS;
}

/*****************************************************************************
NAME: close_input_image

PURPOSE: Closes an input image.

RETURNS:
    SUCCESS/ERROR

*****************************************************************************/
int close_input_image
(
    INPUT_IMAGE *image      /* I: input image pointer */
)
{
    int status = SUCCESS;

    /* warn if there is an open band */
    if (image->band_number != -1)
    {
        IAS_LOG_WARNING("Closing the input file with a band still open");
    }

    if (image->is_l1r)
    {
        /* close the L1R image */
        status = ias_l1r_close_image(image->l1r_file);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Error closing L1R input image");
        }
    }
    else
    {
        /* close the L0R image */
        status = ias_l0r_close(image->l0r_file);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Error closing L0R input image");
        }
    }

    free(image);

    return status;
}

/*****************************************************************************
NAME: setup_radiance_parameters

PURPOSE: Setup the local variables with the rescaling parameter values.

RETURNS: Nothing.  It sets the local variables required for the rescaling.
*****************************************************************************/
void setup_radiance_parameters
(
    int band_index /* I: current band */
)
{ 
    scaling_multiplier = get_radiance_rescale_gain(band_index);
    scaling_offset = get_radiance_rescale_bias(band_index);

    /* invert the gain so that the division turns into a multiplication in
       the core resampler code. */
    scaling_multiplier = 1.0F/scaling_multiplier;
}
