/******************************************************************************
NAME:           get_resample_parameters

PURPOSE:        
This module reads the work order and OMF parameter files required by the
resampler.

ROUTINES:
    read_resample_parameters
    read_omf_parameters
    process_parameters
    get_acquisition_type
    get_band_index_list
    get_output_range_minimum
    get_output_range_maximum
    get_cubic_convolution_alpha
    get_fill_pixel
    get_create_ang_flag
    get_1R_filename
    get_1G_filename
    get_grid_filename
    do_terrain_correction
    do_sca_combine
    get_DEM_filename
    get_output_data_type
    get_output_data_size
    get_resampling_type
    get_0R_filename
    get_CPF_filename
    get_model_filename
    get_processing_pass
    get_omf_filename
    get_satellite_id

******************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <limits.h>                   /* PATH_MAX definition */
#include "ias_satellite_attributes.h"
#include "ias_miscellaneous.h"
#include "ias_logging.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "get_resample_parameters.h"  /* prototypes for this module */
#include "debug.h"                    /* ASSERT macro definition */

/*
**
** VARIABLE DECLARATIONS
**
*/

/* list of band indices to be processed */
static int local_band_index_list[IAS_MAX_NBANDS];

/* flag to indicate that radiance rescale from the cpf should be applied
   non-zero = apply radiance rescale factors from cpf */
static int local_radiance_rescale_flag;

/* flag to indicate terrain correction should be done. non-zero = do terrain 
   correction */
static int local_terrain_flag;

/* flag to indicate the SCAs should be combined into the output image instead
   of writing separate images for each of the SCAs */
static int local_combine_sca_in_output;

/* minimum and maximum output values from the resampler */
static double local_min_max_output[2];

/* cubic convolution alpha parameter */
static double local_alpha;

/* create ANG file flag */
static int local_create_ang_flag;

/* value to use for fill pixels */
static double local_fill_pixel;

/* resampling type to perform */
static RESAMPLE_TYPE local_resampling_type;

/* output data type */
static IAS_DATA_TYPE local_output_type;

/* The current processing pass. */
static int local_processing_pass;

/* The maximum number of processors to utilize. */
static int local_max_processors_to_use;

static int local_band_count;
static int local_band_numbers[IAS_MAX_NBANDS] = {ERROR};

static char local_wo_id[IAS_WORKORDER_ID_SIZE]; /* Work order ID */

static IAS_SATELLITE_ID local_satellite_id; /* Satellite ID */

/* various file name storage */
static char local_1R_filename[PATH_MAX];
static char local_1G_filename[PATH_MAX];
static char local_grid_filename[PATH_MAX];
static char local_dem_filename[PATH_MAX];
static char local_model_filename[PATH_MAX];
static char local_cpf_filename[PATH_MAX];
static char local_L0R_filename[PATH_MAX];
static char local_omf_filename[PATH_MAX];  /* OMF name */

static int local_wrs_path = 0;
static int local_wrs_row = 0;

/* character string storage for the output data type read from the parameter
   file.  defaulted to byte. */
static char local_odtype[5];

/* resampling option read from the parameter file. */
static char local_resample_opt[4];

/* acquisition type for the data being processed */
static IAS_ACQUISITION_TYPE acquisition_type;

/* buffer type to be used when reading the bands */
static BUFFER_TYPE local_buffer_type;
static char local_buffer_opt[10];

/* threshold to use when default buffer is selected */
static int local_input_image_buffer_threshold;

/* Parameters related to the SCA overlap characterization algorithm */
static char local_sca_overlap_sca_option[20];
static int sca_overlap_char_type;
static char sca_overlap_char_filename[PATH_MAX];

/******************************************************************************
NAME: read_resample_parameters

PURPOSE: This function reads all the parameters from the parameter file
         that the resampler requires.

RETURN VALUE: Type = int
    Value    Description
    -----    -----------
    SUCCESS  The parameters were successfully read
    ERROR    An error occurred reading the parameters

******************************************************************************/
int read_resample_parameters
(
    const char *parm_filename /* I: parameter filename that contains resample 
                                    parameters */
)
{
    int count;                /* loop counter */
    int status;

    /* start parameter table definition */
    IAS_PARM_DECLARE_TABLE(parms, 17);

    /* Get the work order ID */
    IAS_PARM_WORK_ORDER_ID(parms, local_wo_id,
                           sizeof(local_wo_id), 1);

    /* Get processing pass. */
    IAS_PARM_PROCESSING_PASS(parms, &local_processing_pass,
                             sizeof(local_processing_pass), 0);

    IAS_PARM_BAND_LIST(parms, IAS_PARM_REQUIRED, local_band_numbers, 
                            sizeof(local_band_numbers), 1);

    IAS_PARM_L1G_FILENAME(parms, local_1G_filename,
                          sizeof(local_1G_filename), 1);

    int default_radiance_rescale_flag[] = {1};
    IAS_PARM_ADD_INT(parms, APPLY_RADIANCE_RESCALING,
        "Flag to indicate radiance rescaling should be applied to the "
        "input L1R image (1 = Apply rescaling, 0 = Do not apply rescaling)",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has a limited range of 0-1*/
        1, 0, 1,
        /* has one default value */
        1, default_radiance_rescale_flag,
        &local_radiance_rescale_flag, 
        sizeof(local_radiance_rescale_flag), 0);

    int default_terrain_flag[] = {1};
    IAS_PARM_ADD_INT(parms, TERRAIN_FLAG,
        "Flag to indicate terrain correction should be done",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has a limited range of 0-1 */
        1, 0, 1,
        /* has one default value */
        1, default_terrain_flag,
        &local_terrain_flag, sizeof(local_terrain_flag), 0);

    int default_combine_sca_in_output[] = {1};
    IAS_PARM_ADD_INT(parms, COMBINE_SCAS,
        "Flag to indicate the SCAs should be combined into the output image",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has a limited range of 0-1 */
        1, 0, 1,
        /* has one default value */
        1, default_combine_sca_in_output,
        &local_combine_sca_in_output, sizeof(local_combine_sca_in_output), 0);

    /* define the output type to use and use UI*2 as the default.  Note that
       if the default changes, the default min/max should change as well */
    const char *valid_odtype_options[] = {"UI*2", "BYTE", "I*2", "I*4", "R*4"};
    const char *default_odtype[] = {"UI*2"};
    IAS_PARM_ADD_STRING(parms, ODTYPE,
        "Character string storage for the output data type",
        IAS_PARM_OPTIONAL,
        /* has a list of valid options */
        5, valid_odtype_options,
        /* has a default value */
        1, default_odtype,
        local_odtype, sizeof(local_odtype), 0);

    /* define the default output range, assuming it is UI*2.  If the default
       output type changes, this will need to be updated. */
    double default_min_max_output[] = {0.0, 65535.0};
    IAS_PARM_ADD_DOUBLE(parms, MINMAX_OUTPUT_DN,
        "Minimum and maximum output values from the resampler.  This may be "
        "further restricted by the radiance rescaling QCALMIN and QCALMAX.",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
        /* has no range constraints */
        0, 0.0, 0.0,
        /* has two default values */
        2, default_min_max_output,
        local_min_max_output, sizeof(local_min_max_output), 2);

    const char *valid_resampling_options[] = {"CC", "NN"};
    const char *default_resampling_option[] = {"CC"};
    IAS_PARM_ADD_STRING(parms, RESAMPLE, "Resampling method to use, CC or NN",
        IAS_PARM_OPTIONAL,
        /* has a list of valid options */
        2, valid_resampling_options,
        /* has no default value */
        1, default_resampling_option,
        local_resample_opt, sizeof(local_resample_opt), 0);

    double default_alpha[] = {-0.5};
    IAS_PARM_ADD_DOUBLE(parms, PCCALPHA, "Cubic convolution alpha parameter",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has range from -1.0 to 0.0 */
        1, -1.0, 0.0,
        /* has one default value */
        1, default_alpha,
        &local_alpha, sizeof(local_alpha), 0);

    const char *valid_buffer_options[] = {"DEFAULT", "FULL_BAND", "TILE"};
    const char *default_buffer_option[] = {"DEFAULT"};
    IAS_PARM_ADD_STRING(parms, INPUT_IMAGE_BUFFER_METHOD, "Input image "
        "buffering method to use. The default selects full buffering for "
        "input images whose number of lines falls below the INPUT_IMAGE_"
        "BUFFER_TILE_THRESHOLD and tiling for any other input images", 
        IAS_PARM_OPTIONAL,
        /* has a list of 3 valid options */
        3, valid_buffer_options,
        /* Specify the default value */
        1, default_buffer_option,
        local_buffer_opt, sizeof(local_buffer_opt), 0);

    int default_threshold[] = {10500};
    IAS_PARM_ADD_INT(parms, INPUT_IMAGE_BUFFER_TILE_THRESHOLD, "The number "
        "of lines in a 30m band to use as a threshold before the tile "
        "buffering method is used",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has range from 0 to 100,000 */
        1, 0, 100000,
        /* has one default value */
        1, default_threshold, &local_input_image_buffer_threshold, 
        sizeof(local_input_image_buffer_threshold), 0);

    int default_create_ang_flag[] = {0};
    IAS_PARM_ADD_INT(parms, CREATE_ANG_FLAG, "Flag to indicate that the ANG "
        "file should/shouldn't be created [1 = create; 0 = don't create] ",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has a limited range of 0-1 */
        1, 0, 1,
        /* has one default value */
        1, default_create_ang_flag, &local_create_ang_flag,
        sizeof(local_create_ang_flag), 0);

    /* for the fill pixel, don't constrain the range */
    double default_fill_pixel = 0.0;
    IAS_PARM_BACKGROUND(parms, &default_fill_pixel, &local_fill_pixel,
        sizeof(local_fill_pixel));

    const char *valid_sca_overlap_char_options[] = {"NONE", "LIMITED", "FULL"};
    const char *default_sca_overlap_char_option[] = {"NONE"};
    IAS_PARM_ADD_STRING(parms, SCA_OVERLAP_CHAR_TYPE, "Defines the type of "
        "data to collect for the SCA overlap characterization algorithm",
        IAS_PARM_OPTIONAL, 
        /* has a list of 3 valid options */
        3, valid_sca_overlap_char_options,
        /* has one default value */
        1, default_sca_overlap_char_option, local_sca_overlap_sca_option, 
        sizeof(local_sca_overlap_sca_option), 0);

    IAS_PARM_SCA_OVERLAP_CHAR_FILENAME(parms, IAS_PARM_OPTIONAL,
        sca_overlap_char_filename, sizeof(sca_overlap_char_filename), 1);

    /* calculate number of items in the parms list */
    count = IAS_PARM_GET_TABLE_SIZE(parms);

    /* call the ias_parm_provide_help routine to provide help for the 
       parameters if requested */
    status = ias_parm_provide_help(parm_filename, parms, count,
                                   IAS_INPUT_PARAMETERS);
    if (status)
    {
        if (status == 1)
        {
            /* provide help for the OMF.  If it succeeds, it will exit.  If
               it returns, an error occurred, so return an error */
            read_omf_parameters(parm_filename);
            return ERROR;
        }
        else
            return ERROR;
    }

    if (ias_parm_read(parm_filename, parms, count) == SUCCESS)
    {
        /* Save the band count for later */
        local_band_count = IAS_PARM_GET_COUNT_READ(parms, BAND_LIST);

        /* Build the OMF filename. */
        sprintf(local_omf_filename, "%s.omf", local_wo_id);
    }
    else
    {
        IAS_LOG_ERROR("Error reading parameter file %s", parm_filename);
        return ERROR;
    }

    return read_omf_parameters(local_omf_filename);
}

/******************************************************************************
NAME: read_omf_parameters

PURPOSE: This function reads all the parameters from the OMF file that the 
         resampler requires.

RETURN VALUE: Type = int
    Value    Description
    -----    -----------
    SUCCESS  The parameters were successfully read
    ERROR    An error occurred reading the parameters
******************************************************************************/
int read_omf_parameters
(
    const char *omf_filename   /* I : OMF filename that contains order 
                                      parameters */
)
{
    int count;                /* loop counter */
    int status;
    int satellite_number;

    /* Now read and check the OMF. */
    /* start parameter table definition */
    IAS_PARM_DECLARE_TABLE(omf_parms, 11);

    IAS_PARM_SATELLITE(omf_parms, &satellite_number,
            sizeof(satellite_number), 1);

    IAS_PARM_MAX_PROCESSORS_TO_USE(omf_parms, IAS_PARM_OPTIONAL,
            &local_max_processors_to_use,
            sizeof(local_max_processors_to_use), 1);

    IAS_PARM_CPF(omf_parms, IAS_PARM_REQUIRED, local_cpf_filename, 
            sizeof(local_cpf_filename), 1);

    IAS_PARM_L0R_FILENAME(omf_parms, IAS_PARM_REQUIRED, local_L0R_filename,
            sizeof(local_L0R_filename), 1);

    IAS_PARM_L1R_FILENAME(omf_parms, IAS_PARM_REQUIRED, local_1R_filename,
            sizeof(local_1R_filename), 1);

    IAS_PARM_TARGET_WRS_PATH(omf_parms, IAS_PARM_OPTIONAL, &local_wrs_path,
            sizeof(local_wrs_path), 1);
    IAS_PARM_TARGET_WRS_ROW(omf_parms, IAS_PARM_OPTIONAL, &local_wrs_row, 
            sizeof(local_wrs_row), 1);

    IAS_PARM_DEM_FILENAME_FROM_PASS(omf_parms, IAS_PARM_OPTIONAL,
            local_dem_filename, sizeof(local_dem_filename), 1,
            local_processing_pass);

    IAS_PARM_GRID_FILENAME_FROM_PASS(omf_parms, IAS_PARM_REQUIRED,
            local_grid_filename, sizeof(local_grid_filename), 1,
            local_processing_pass);

    IAS_PARM_MODEL_FILENAME_FROM_PASS(omf_parms, IAS_PARM_REQUIRED,
            local_model_filename, sizeof(local_model_filename), 1,
            local_processing_pass);

    IAS_PARM_ACQUISITION_TYPE(omf_parms, &acquisition_type,
            sizeof(acquisition_type), 1);

    /* calculate number of items in the omf_parms list */
    count = IAS_PARM_GET_TABLE_SIZE(omf_parms);

    status = ias_parm_provide_help(omf_filename, omf_parms, count,
            IAS_OMF_PARAMETERS);
    if (status)
    {
        /* Return ERROR if help requested */
        if (status == 1)
        {
            /* successfully provided help, as requested, so just exit */
            exit(EXIT_SUCCESS);
        }
        IAS_LOG_ERROR("Error occurred providing OMF help");
        return ERROR;
    }

    if (ias_parm_read(omf_filename, omf_parms, count) == SUCCESS)
    {
        /* Save for use outside this file. */
        local_satellite_id
            = ias_sat_attr_get_satellite_id_from_satellite_number(
                    satellite_number);
        if (local_satellite_id == ERROR)
        {
            IAS_LOG_ERROR( "Unable to determine satellite ID "
                    "from satellite number %d", satellite_number);

            return ERROR;
        }
    }
    else
    {
        IAS_LOG_ERROR("Error reading OMF parameter file %s", omf_filename);

        return ERROR;
    }
    return SUCCESS;
}

/******************************************************************************
NAME: process_parameters

PURPOSE: This function performs special processing on parameters which need
         to be translated after being read from the parameter or OMF files.

RETURN VALUE: Type = int
    Value    Description
    -----    -----------
    SUCCESS  The parameters were successfully processed
    ERROR    An error occurred processing the parameters
******************************************************************************/
int process_parameters()
{
    int count = 0;
    int temp_band_index = 0;
    int band_index = 0;
    int band_already_in_list[IAS_MAX_NBANDS] = {0}; /* flag for a band is 
                                 already in the band list */

    /* verify the parameters that are required only if some flag turns on
       an option */
    if (local_terrain_flag)
    { 
        /* terrain correction requested, so make sure the DEM name
           was provided */
        if (local_dem_filename[0] == '\0')
        {
            /* error reading required item, log an error */
            IAS_LOG_ERROR("DEM name not provided in the parameter file");
            return ERROR;
        }
    }

    /* parse the output data type */
    ias_misc_convert_to_uppercase(local_odtype);
    if (strcmp(local_odtype, "UI*2") == 0)
        local_output_type = IAS_UI2;
    else if (strcmp(local_odtype, "BYTE") == 0)
        local_output_type = IAS_BYTE;
    else if (strcmp(local_odtype, "I*2") == 0)
        local_output_type = IAS_I2; 
    else if (strcmp(local_odtype, "I*4") == 0)
        local_output_type = IAS_I4;
    else if (strcmp(local_odtype, "R*4") == 0)
        local_output_type = IAS_R4;
    else
    {
        IAS_LOG_ERROR("Unknown output type - %s\n",local_odtype);

        return ERROR;
    }

    /* parse the resampling type */
    ias_misc_convert_to_uppercase(local_resample_opt); 
    if (strcmp(local_resample_opt, "CC") == 0)
        local_resampling_type = CC_RESAMPLING;
    else if (strcmp(local_resample_opt, "NN") == 0)
        local_resampling_type = NN_RESAMPLING;
    else
    {
        IAS_LOG_ERROR("Unknown resample option %s",local_resample_opt);
        return ERROR;
    }

    /* parse the buffer type */
    ias_misc_convert_to_uppercase(local_buffer_opt); 
    if (strcmp(local_buffer_opt, "DEFAULT") == 0)
        local_buffer_type = DEFAULT_BUFFER;
    else if (strcmp(local_buffer_opt, "FULL_BAND") == 0)
        local_buffer_type = FULL_BAND_BUFFER;
    else if (strcmp(local_buffer_opt, "TILE") == 0)
        local_buffer_type = TILE_BUFFER;
    else
    {
        IAS_LOG_ERROR("Unknown buffer option %s",local_buffer_opt);
        return ERROR;
    }

    band_index = 0;
    for (count = 0; count < local_band_count; count++)
    {
        /* Convert the band number entered into a band index */
        temp_band_index = ias_sat_attr_convert_band_number_to_index(    
                                local_band_numbers[count]);
        if ((count == 0) && (temp_band_index == ERROR))
        {
            /* no bands numbers entered, so it is an error */
            IAS_LOG_ERROR("Illegal band number %d",local_band_numbers[count]);
            return ERROR;
        }

        /* make sure the band isn't already in the list */
        if (temp_band_index != ERROR)
        {
            if (!band_already_in_list[temp_band_index])
            {
                /* put band in list and flag that it has been added 
                   to the list */
                local_band_index_list[band_index] = temp_band_index;
                band_already_in_list[temp_band_index] += 1;
                band_index++;
            }
            else if (band_already_in_list[temp_band_index] == 1)
            {
                 /* only print the warning message once for each band */
                 band_already_in_list[temp_band_index] += 1;

                 IAS_LOG_WARNING("Warning: Band %d appears multiple times "
                            "in the BAND_LIST, only the first one was used",
                            local_band_numbers[count]);
            }
        }
    }

    /* fill in remaining band locations with the band error indication */
    for (count = band_index; count < IAS_MAX_NBANDS; count++)
    {
        local_band_index_list[count] = ERROR;
    }

    /* Make sure the target WRS path/row were read if the acquisition is
       earth-looking */
    if (acquisition_type == IAS_EARTH)
    {
        if (local_wrs_path == 0 || local_wrs_row == 0)
        {
            IAS_LOG_ERROR(
                "Required TARGET_WRS_PATH/ROW missing from the OMF file");
            return ERROR;
        }
    }
    else if (local_create_ang_flag)
    {
        IAS_LOG_WARNING("Create angle coefficients flag set for non-Earth "
            "scene");
        local_create_ang_flag = FALSE; 
    }

    /* Convert the SCA overlap characterization option into the enumerated
       type */
    if (strcmp(local_sca_overlap_sca_option, "LIMITED") == 0)
        sca_overlap_char_type = SCA_OVERLAP_CHAR_LIMITED;
    else if (strcmp(local_sca_overlap_sca_option, "FULL") == 0)
        sca_overlap_char_type = SCA_OVERLAP_CHAR_FULL;
    else
        sca_overlap_char_type = SCA_OVERLAP_CHAR_NONE;

    return SUCCESS;
}

/******************************************************************************
NAME:           various accessor functions

PURPOSE:        
These functions provide access to the parameters read. 

RETURN VALUE:
Varies for the function.

NOTES:
The accessor function method of providing parameters was selected since
passing all the parameters up and down a calling sequence is a 
maintenance nightmare.

******************************************************************************/

char *get_1R_filename(void)
{
    return local_1R_filename;
}

IAS_ACQUISITION_TYPE get_acquisition_type()
{
    return acquisition_type;
}

const int *get_band_index_list(void)
{
    return local_band_index_list;
}

char *get_1G_filename(void)
{
    return local_1G_filename;
}

char *get_grid_filename(void)
{
    return local_grid_filename;
}

int do_terrain_correction(void)
{
    return local_terrain_flag;
}

int do_apply_cpf_rad_rescale(void)
{
    return local_radiance_rescale_flag;
}

int do_sca_combine(void)
{
    return local_combine_sca_in_output;
}

char *get_DEM_filename(void)
{
    return local_dem_filename;
}

IAS_DATA_TYPE get_output_data_type(void)
{
    return local_output_type;
}

int get_output_data_size(void)
{
    if (local_output_type == IAS_BYTE)
        return sizeof(unsigned char);
    else if (local_output_type == IAS_I2)
        return sizeof(short int);
    else if (local_output_type == IAS_UI2)
        return sizeof(unsigned short int);
    else if (local_output_type == IAS_I4)
        return sizeof(int);
    else 
    {
        ASSERT(local_output_type == IAS_R4);
        return sizeof(float);
    }
}

RESAMPLE_TYPE get_resampling_type(void)
{
    return local_resampling_type;
}

BUFFER_TYPE get_buffer_type(void)
{
    return local_buffer_type;
}

int get_input_image_buffer_threshold(void)
{
    return local_input_image_buffer_threshold;
}

float get_output_range_minimum(void)
{
    return local_min_max_output[0];
}

float get_output_range_maximum(void)
{
    return local_min_max_output[1];
}

float get_cubic_convolution_alpha(void)
{
    return local_alpha;
}

int get_create_ang_flag(void)
{
    return local_create_ang_flag;
}

float get_fill_pixel(void)
{
   return local_fill_pixel;
}

char *get_model_filename(void)
{
    return local_model_filename;
}

char *get_0R_filename(void)
{
    return local_L0R_filename;
}

char *get_CPF_filename(void)
{
    return local_cpf_filename;
}

int get_processing_pass(void)
{
    return local_processing_pass;
}

int get_max_processors_to_use(void)
{
    return local_max_processors_to_use;
}

char* get_omf_filename(void)
{
    return local_omf_filename;
}

IAS_SATELLITE_ID get_satellite_id(void)
{
    return local_satellite_id;
}

SCA_OVERLAP_CHAR_TYPE get_sca_overlap_char_type(void)
{
    return sca_overlap_char_type;
}

char *get_sca_overlap_char_filename(void)
{
    return sca_overlap_char_filename;
}

int get_wrs_path(void)
{
    return local_wrs_path;
}

int get_wrs_row(void)
{
    return local_wrs_row;
}

