/******************************************************************************
NAME:           kernel

PURPOSE:        
The kernel module contains the routines needed to create and manage the 
resampling kernel.  The kernel is the weight table used for spatially 
resampling pixels.

ROUTINES:
    kernel_setup
    cleanup_kernel
    get_resample_weight_table_ptr
    get_lines_in_kernel
    get_samples_in_kernel
    num_left_kernel_samples
    num_right_kernel_samples
    num_top_kernel_lines
    num_bottom_kernel_lines
    get_kernel_step_size

NOTES:
- The kernel module will build the resampling weight table for cubic
  convolution resampling.

******************************************************************************/

#include <stdio.h>         /* i/o routines */
#include <stdlib.h>        /* memory management routines */
#include <math.h>          /* fabs routine */
#include "ias_const.h"     /* SUCCESS/ERROR definition */
#include "ias_logging.h"   /* logging library */
#include "ias_math.h"      /* cubic convolution routine */
#include "kernel.h"        /* prototypes for this module */
#include "debug.h"         /* ASSERT macro */

/*
**
** MACROS
**
*/

#define SUBPIXEL_STEPS 64
/* defines the number of subpixel steps to include in the kernel weights. */

#define SUBPIXEL_TABLE_ENTRIES (SUBPIXEL_STEPS + 1)
/* defines the number of entries in the kernel weight table for the subpixel
   steps.  This is one more than the number of SUBPIXEL steps since the 
   fractional line number is rounded to the nearest 1/32 subpixel in the
   resampler.  To avoid having to check for rounding up to the next whole 
   line number, the extra subpixel step is included in the table. */

#define LINES_IN_KERNEL 4
/* defines the number of lines in the resampling kernel. This is the normal 
   cubic convolution kernel size. */

#define SAMPLES_IN_KERNEL 1
/* defines the number of samples in the resampling kernel.  Since the
   resampling only uses cubic convolution in the line direction, there is only
   a single value in the sample direction. */

/*
**
** MODULE VARIABLES
**
*/

static double *kernel_weights_ptr;
/* kernel_weights_ptr is a pointer to the table of resampling weights for
   cubic convolution resampling.

   The table is laid out as a 2-dimensional array with the following indices:
    table[line subpixel offset (size SUBPIXEL_TABLE_ENTRIES)]
         [kernel line (size LINES_IN_KERNEL)]
   
*/

static int lines_in_kernel = 1;
/* number of lines in the kernel.  Its default initialization is one line
   (assuming the nearest neighbor kernel).  This is set to LINES_IN_KERNEL
   when kernel_setup is called. */

static int samples_in_kernel = 1;
/* number of samples in the kernel.  Its default initialization is one sample
   (assuming the nearest neighbor kernel).  This is set to SAMPLES_IN_KERNEL
   when kernel_setup is called. */

static int kernel_size_accessed;
/* the kernel_size_accessed is set to a non-zero value when either of the 
   functions to return a kernel size parameter is called.  This allows
   a debugging check to be put into the kernel_setup routine to make
   sure the kernel size isn't accessed before kernel_setup is called and
   changes the kernel size.  kernel_size_access bit 0 means the kernel
   size has been accessed, and bit 1 means the kernel has been setup. */

/*
**
** LOCAL PROTOTYPES
**
*/

static void create_resampling_kernel_1d
(
    double *weight_ptr, /* I/O: pointer to weights table */
    double alpha        /* I: cc alpha parameter */
);

/******************************************************************************
NAME:           kernel_setup

PURPOSE:                
kernel_setup initializes the row weights for creating the extended image and
creates the resampling weight table for the current resampling method.

RETURN VALUE:
Either SUCCESS or ERROR are returned, depending on whether the kernel 
setup is successful or not.

ALGORITHM REFERENCES:
none

******************************************************************************/

int kernel_setup
(
    double cc_alpha /* I: cubic convolution alpha parameter */
)
{
    int kernel_table_size; /* number of weights in the kernel weight table */

    /* verify the kernel size hasn't been accessed before the kernel_setup
       is done the first time */
    ASSERT(kernel_size_accessed != 1);
    kernel_size_accessed |= 2;  /* set bit 1 */

    /* initialize the lines and samples in the kernel */
    lines_in_kernel = LINES_IN_KERNEL;
    samples_in_kernel = SAMPLES_IN_KERNEL;

    /* if the kernel_weight_ptr has already been allocated from a previous
       call, free it since the size may change for the current call */
    free(kernel_weights_ptr);

    /* calculate the size of the kernel weight table */
    kernel_table_size = lines_in_kernel * SUBPIXEL_TABLE_ENTRIES;
    kernel_weights_ptr = (double *)malloc(sizeof(double) * kernel_table_size);
    if (kernel_weights_ptr == NULL)
    {
        IAS_LOG_ERROR("Error allocating kernel weights");
        return ERROR;
    }

    create_resampling_kernel_1d(kernel_weights_ptr, cc_alpha);

    return SUCCESS;
}

/******************************************************************************
NAME:           cleanup_kernel

PURPOSE:                
cleanup_kernel cleans up the effects of the kernel_setup routine.

RETURN VALUE:
nothing is returned


ALGORITHM REFERENCES:
none

******************************************************************************/

void cleanup_kernel(void)
{
    /* free the space used for the resampling weights */
    free (kernel_weights_ptr);
    kernel_weights_ptr = NULL;
}


/******************************************************************************
NAME: create_resampling_kernel_1d

PURPOSE:
create_resampling_kernel_1d builds the cubic convolution weight table for the
resampling kernel.

RETURNS:
    Nothing

******************************************************************************/
static void create_resampling_kernel_1d
(
    double *weight_ptr, /* I/O: pointer to weights table */
    double alpha        /* I: cc alpha parameter      */
)
{
    int subpixel_line;             /* Counters */
    double dx;                     /* Sub-pixel spacing for cubic */
    double *w_ptr;                 /* pointer to current weight table */
    double x;
    int kernel_line;
    int top_line;
    int bottom_line;

    /* set temporary pointer to the weight table */
    w_ptr = weight_ptr;

    /* get the kernel dimensions */
    top_line = -num_top_kernel_lines();
    bottom_line = num_bottom_kernel_lines();

    for (subpixel_line = 0; subpixel_line <= SUBPIXEL_STEPS; subpixel_line++)
    {
        /* calculate subpixel line step */
        dx = subpixel_line / (double)(SUBPIXEL_STEPS);

        for (kernel_line = top_line; kernel_line <= bottom_line; kernel_line++)
        {
            /* adjust the kernel line location for the current
               subpixel step. */
            x = (double)kernel_line - dx;

            /* get the weight for the current pixel */
            *w_ptr++ = ias_math_cubic_convolution(alpha, x);

        } /* kernel lines */
    } /* subpixel lines */

}

/******************************************************************************
NAME:           get_resample_weight_table_ptr

PURPOSE:
get_resample_weight_table_ptr provides access to the resampling kernel weight
table.

RETURN VALUE:
The pointer to the resampling weight table is returned.

NOTES:
The pointer to the weight table could be passed back to the calling routine when
it is created, but this is a little more efficient.

ALGORITHM REFERENCES:
none

******************************************************************************/

const double *get_resample_weight_table_ptr(void)
{
    ASSERT (kernel_weights_ptr != NULL);
    return kernel_weights_ptr;
}

/******************************************************************************
NAME:           get_lines_in_kernel, get_samples_in_kernel

PURPOSE:
The get_lines_in_kernel and get_samples_in_kernel routines provide access 
to the number of lines and samples in the resampling kernel.

RETURN VALUE:
Type = int
The number of lines/samples in the resampling kernel is returned

NOTES:
none

ALGORITHM REFERENCES:
none

******************************************************************************/

int get_lines_in_kernel(void)
{
    /* set the accessed flag to allow debugging check */
    kernel_size_accessed |= 1;

    return lines_in_kernel;
}

int get_samples_in_kernel(void)
{
    /* set the accessed flag to allow debugging check */
    kernel_size_accessed |= 1;

    return samples_in_kernel;
}

/* these routines return the number of pixels to the left, right, top, and
   bottom of the kernel center */

int num_left_kernel_samples(void)
{
    return ((samples_in_kernel) - ((samples_in_kernel)/2) - 1);
}

int num_right_kernel_samples(void)
{
    return (samples_in_kernel/2);
}

int num_top_kernel_lines(void)
{
    return ((lines_in_kernel) - ((lines_in_kernel)/2) - 1);
}

int num_bottom_kernel_lines(void)
{
    return (lines_in_kernel/2);
}

double get_kernel_step_size(void)
{
    return (1.0/(double)SUBPIXEL_STEPS);
}

void get_kernel_table_info
(
    int *subpixel_steps, /* O: number of subpixel steps in table */
    int *table_entries,  /* O: number of table entries for a whole pixel
                                step (one more than subpixel_steps */
    int *kernel_width,   /* O: width of kernel (samples across kernel) */
    int *kernel_height   /* O: height of kernel (lines across kernel) */
)
{
    /* verify this routine is only called for cubic convolution */
    ASSERT (samples_in_kernel == SAMPLES_IN_KERNEL);

   *subpixel_steps = SUBPIXEL_STEPS;
   *table_entries = SUBPIXEL_TABLE_ENTRIES;
   *kernel_width = samples_in_kernel;
   *kernel_height = lines_in_kernel;
}
