/******************************************************************************
NAME: process_band

PURPOSE: Resamples a single band.

RETURNS:
    SUCCESS/ERROR

******************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "ias_const.h"               /* SUCCESS/ERROR definitions */
#include "ias_logging.h"
#include "get_resample_parameters.h" /* read parameters routine */
#include "resample_image.h"          /* resampling routine prototypes */
#include "ias_threadpool.h"          /* multithreaded definitions */
#include "process_band.h"
#include "get_input_image.h"
#include "track_tile.h"
#include "buffer_output_image.h"
#include "ias_miscellaneous.h"

int process_band
(
    struct ias_threadpool *threadpool, /* I: resampling thread pool */
    int band_index,                    /* I: band index being processed */
    IAS_GRID_TYPE *grid,               /* I: image grid */
    INPUT_IMAGE *input_image,          /* I: input image I/O structure */
    OUTPUT_BUFFER *output_image_ptr,   /* I: output file I/O structure */
    const IAS_LOS_MODEL *model_ptr,    /* I: input model */
    IAS_IMAGE *dem_image,              /* I: DEM data */
    SCA_OVERLAP_IO *sca_overlap_file   /* I: SCA overlap char file (can be NULL
                                             if not writing the file) */
)
{
    int status;                      /* Return status */
    int combine_scas = do_sca_combine();
    int sca_index;                   /* SCA counter */
    int sca_end_loop;                /* value to stop looping on SCAs */
    double trim_lines[4];            /* Lookup table line corners */
    double trim_samps[4];            /* Lookup table samp corners */
    RESAMPLE_THREAD_PARAMS params;/* parameters to pass to resampling threads */
    RESAMPLE_TYPE resampling_type = get_resampling_type();
    int band_number;                 /* Current band number */
    IAS_MISC_LINE_EXTENT *trim_lut;  /* Image trim lookup table */
    double min_det_offsets[IAS_MAX_NSCAS]; /* minimum per-SCA detector offset */
    double max_det_offsets[IAS_MAX_NSCAS]; /* maximum per-SCA detector offset */

    /* select the resampling function to use */
    switch (resampling_type)
    {
        case CC_RESAMPLING:
            params.resample_function = cc_resample;
            break;

        case NN_RESAMPLING:
            params.resample_function = nn_resample;
            break;

        default:
            IAS_LOG_ERROR("Unsupported resampling type: %d", resampling_type);
            return ERROR;
    }

    /* if combining SCAs, only loop once; otherwise loop for nscas */
    if (combine_scas)
        sca_end_loop = 1;
    else
        sca_end_loop = grid->scas_per_band[band_index];

    /* get the band number */
    band_number = ias_sat_attr_convert_band_index_to_number(band_index);
    if (band_number == ERROR)
    {
        IAS_LOG_ERROR("Error converting band index %d to a band number",
                      band_index);
        return ERROR;
    }

    /* provide a message for each band processed */
    IAS_LOG_INFO("Resampling band number %d", band_number);

    /* Get the maximum line offsets for each of the SCAs in this band */
    if (ias_los_model_get_min_max_detector_offsets(model_ptr, band_number, 
            min_det_offsets, max_det_offsets) != SUCCESS)
    {
        IAS_LOG_ERROR("Error getting the maximum detector offsets for band "
            "number %d", band_number);
        return ERROR;
    }
    for (sca_index = 0; sca_index < grid->scas_per_band[band_index];
        sca_index++)
    {
        params.max_line_offsets[sca_index] = (int)ceil(
            fabs(max_det_offsets[sca_index] - min_det_offsets[sca_index]));
    }

    /* Calculate the trim box for the the lookup table */
    if (ias_misc_calculate_trim_box(band_index, 
        &grid->active_corners[band_index], grid, trim_lines, trim_samps)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating the trim box for band number %d", 
            band_number);
        return ERROR;
    }

    /* Create the trimming lookup table for the output image to allow trimming
       off the extra SCA data */
    trim_lut = ias_misc_create_output_image_trim_lut(trim_lines, trim_samps,
        grid->gridbands[band_index][0].lines,
        grid->gridbands[band_index][0].samps);
    if (!trim_lut)
    {
        IAS_LOG_ERROR("Failed attempting to create the SCA trimming table for "
                "band number %d", band_number);
        return ERROR;
    }

    /* open the input band */
    if (open_input_band(input_image, band_number) != SUCCESS)
    {
        IAS_LOG_ERROR("Error opening Band %d for L1R file %s", band_number,
                      get_1R_filename());
        free(trim_lut);
        return ERROR;
    }
    params.input_image = input_image;

    /* do some consistency checking on the L1R file versus the grid */
    if (input_image->scas != grid->scas_per_band[band_index])
    {
        IAS_LOG_ERROR("Expected %d SCAs in %s, but found %d for band number %d",
                      grid->scas_per_band[band_index], get_1R_filename(),
                      input_image->scas, band_number);
        close_input_band(input_image);
        free(trim_lut);
        return ERROR;
    }


    /* start writing the current band/sca */
    if (open_output_band(output_image_ptr, band_index, sca_end_loop,
            grid->gridbands[band_index][0].lines,
            grid->gridbands[band_index][0].samps, get_output_data_type())
            != SUCCESS)
    {
        IAS_LOG_ERROR("Error setting up write for the current band/sca");
        close_input_band(input_image);
        free(trim_lut);
        return ERROR;
    }

    /* If an SCA overlap characterization file has been provided, open the
       current band in the file */
    if (sca_overlap_file
        && (open_sca_overlap_band(sca_overlap_file, band_number) != SUCCESS))
    {
        IAS_LOG_ERROR("Error starting band in the SCA overlap file");
        close_output_band(output_image_ptr);
        close_input_band(input_image);
        free(trim_lut);
        return ERROR;
    }

    /* create a mutex to control access to the HDF library since it is
       completely unsafe for threads (can't even read and write to
       different files without problems) */
    if (IAS_THREAD_CREATE_MUTEX(&params.hdf_io_mutex) != 0)
    {
        IAS_LOG_ERROR("Error creating HDF I/O mutex");
        close_sca_overlap_band(sca_overlap_file);
        close_output_band(output_image_ptr);
        close_input_band(input_image);
        free(trim_lut);
        return ERROR;
    }

    /* set the HDF library lock in the input image */
    set_input_image_io_lock(input_image, &params.hdf_io_mutex);

    /* set the parameters to pass to the resampling threads */
    params.threadpool = threadpool;
    params.output_image_ptr = output_image_ptr;
    params.grid_ptr = grid;
    params.dem_image = dem_image;
    params.combine_scas = combine_scas;
    params.band_index = band_index;
    params.model_ptr = model_ptr;
    params.trim_lut = trim_lut;
    params.sca_overlap_file = sca_overlap_file;

    /* Start looping on SCAs since data is organized by SCA */
    for (sca_index = 0; sca_index < sca_end_loop; sca_index++)
    {
        if (initialize_tile_tracking(
                grid->gridbands[band_index][sca_index].lines,
                grid->gridbands[band_index][sca_index].samps)
                != SUCCESS)
        {
            IAS_LOG_ERROR("Error initializing the tile tracking module");
            close_sca_overlap_band(sca_overlap_file);
            close_output_band(output_image_ptr);
            close_input_band(input_image);
            IAS_THREAD_DESTROY_MUTEX(&params.hdf_io_mutex);
            free(trim_lut);
            return ERROR;
        }

        /* reset the output buffer for the current SCA */
        reset_output_buffer(output_image_ptr, sca_index);

        /* issue an appropriate informational message for each loop */
        if (!combine_scas)
        {
            IAS_LOG_INFO("Resampling band number %d, SCA number %d",
                    band_number, sca_index + 1);
        }

        /* set the parameters to pass to the resampling threads */
        params.sca = sca_index;
        params.thread_error = 0;

        /* perform the resampling */
        status = ias_threadpool_run_function(threadpool, resample_image,
                                             &params);

        if ((status == ERROR) || (params.thread_error > 0))
        {
            IAS_LOG_ERROR("Error encountered while resampling");
            close_sca_overlap_band(sca_overlap_file);
            close_output_band(output_image_ptr);
            close_input_band(input_image);
            IAS_THREAD_DESTROY_MUTEX(&params.hdf_io_mutex);
            free(trim_lut);
            return ERROR;
        }

        if (flush_output_buffer(output_image_ptr) != SUCCESS)
        {
            IAS_LOG_ERROR("Flushing output buffer for the current band/sca");
            close_sca_overlap_band(sca_overlap_file);
            close_output_band(output_image_ptr);
            close_input_band(input_image);
            IAS_THREAD_DESTROY_MUTEX(&params.hdf_io_mutex);
            free(trim_lut);
            return ERROR;
        }

    } /* SCA loop */

    /* destroy the HDF IO mutex */
    if (IAS_THREAD_DESTROY_MUTEX(&params.hdf_io_mutex) != 0)
    {
        IAS_LOG_ERROR("Error destroying HDF I/O mutex");
        close_sca_overlap_band(sca_overlap_file);
        close_output_band(output_image_ptr);
        close_input_band(input_image);
        free(trim_lut);
        return ERROR;
    }

    /* If the SCA overlap char file is being written, close the current band */
    if (sca_overlap_file    
        && (close_sca_overlap_band(sca_overlap_file) != SUCCESS))
    {
        IAS_LOG_ERROR("Error finishing band in the SCA overlap char file");
        close_output_band(output_image_ptr);
        close_input_band(input_image);
        free(trim_lut);
        return ERROR;
    }

    /* stop writing the current band */
    if (close_output_band(output_image_ptr) != SUCCESS)
    {
        IAS_LOG_ERROR("Error finishing write for the current band");
        close_input_band(input_image);
        free(trim_lut);
        return ERROR;
    }

    /* close the input image */
    if (close_input_band(input_image) != SUCCESS)
    {
        IAS_LOG_ERROR("Error closing band number %d in L1R %s",
                      band_number, get_1R_filename());
        free(trim_lut);
        return ERROR;
    }

    if (cleanup_tile_tracking() != SUCCESS)
    {
        IAS_LOG_ERROR("Error cleaning up tile tracking");
        free(trim_lut);
        return ERROR;
    };

    free(trim_lut);

    return SUCCESS;
}
