/******************************************************************************
NAME:           resample

PURPOSE:
The resample module contains the "main" function for the resampler.

NOTES:
none

******************************************************************************/

#include <stdlib.h>             /* exit prototype */
#include <stdio.h>              /* sprintf prototype */
#include <string.h>
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_threadsync.h"     /* IAS_THREAD_GET_NUM_PROCESSORS */
#include "ias_grid_io.h"
#include "ias_model_io.h"
#include "ias_miscellaneous.h"  /* ias_misc_read_dem prototype */
#include "ias_threadpool.h"
#include "get_resample_parameters.h"  /* read parameters routine */
#include "get_cpf_parameters.h" /* read CPF parameters routine */
#include "buffer_output_image.h"
#include "kernel.h"             /* kernel prototypes */
#include "process_band.h"
#include "resample_image.h"
#include "update_l1g_metadata.h"/* update_l1g_metadata prototype */
#include "update_omf.h"         /* update_omf prototype */
#include "write_envi_hdr.h"
#include "ias_angle_gen.h"
#include "calculate_angle_coeffs.h"
#include "sca_overlap_io.h"

/*
**
** MACROS
**
*/
#define USE_MESSAGE "Usage: resample <parameter_file>"
#define NUM_BACKGROUND_THREADS 1 /* Number of threads used in the angle
                                    coefficients threadpool */

/******************************************************************************
NAME:           main

PURPOSE:
This is the main function for the resampler.  It controls the process of
resampling a level 1R (or 0R) image into a level 1G image using the 
grid.

NOTES:
none

******************************************************************************/

int main(int argc, char *argv[]) 
{
    const int *orig_band_index_list_ptr; 
                            /* pointer to original list of bands to resample */
    int band_index_list[IAS_MAX_NBANDS];
                            /* list of requested bands confirmed present */
    int procs_to_use;
    int num_cpus = IAS_THREAD_GET_NUM_PROCESSORS();/* number of processors */
    IAS_GRID_TYPE grid;              /* grid structure */
    int index;
    int num_bands_processed = 0;     /* Number of bands processed */
    int band_index;
    int create_ang_flag;             /* Flag to create ANG file */
    int sca_index;                   /* SCA counter */
    IAS_IMAGE dem_image;             /* DEM data */
    struct ias_threadpool *threadpool_ptr;
    struct ias_threadpool *background_thread; /* Background thread */
    OUTPUT_BUFFER *output_image_ptr;
    INPUT_IMAGE *input_image_ptr;
    IAS_LOS_MODEL *model_ptr;
    IAS_ANGLE_GEN_METADATA metadata; /* Angle metadata struct */
    char ang_filename[PATH_MAX];     /* Angle coefficient filename */
    COEFF_THREAD_PARAMS coeff_params;/* Band Coefficient threading parameters */
    SCA_OVERLAP_IO *sca_overlap_file = NULL;/* SCA overlap char output file */

    /* initialize the dem_image structure to make it easier to clean up */
    memset(&dem_image, 0, sizeof(dem_image));
    
    /* initialize the ANG variables to make cleanup easier */
    background_thread = NULL;
    metadata.transformation = NULL;
    metadata.ephemeris = NULL;
    metadata.solar_vector = NULL;
    metadata.transformation = NULL;

    /* initialize the logging library */
    if (ias_log_initialize("Resample") != SUCCESS)
    {
        IAS_LOG_ERROR("Error initializing logging library");
        exit(EXIT_FAILURE);
    }

    /* if there are no arguments, generate a usage message */
    if (argc < 2)
    {
        IAS_LOG_ERROR(USE_MESSAGE);
        exit(EXIT_FAILURE);
    }
    
    /* read the resampler control parameters */
    if (read_resample_parameters(argv[1]) != SUCCESS)
    {
        IAS_LOG_ERROR("Error reading parameters from %s", argv[1]);
        exit(EXIT_FAILURE);
    }
    
    /* generate entry message, after the parameter are read so that the
       help provided won't include this message */
    IAS_LOG_INFO("Initiated");

    /* Initialize the satellite attributes. */
    if (ias_sat_attr_initialize(get_satellite_id()) != SUCCESS)
    {
        IAS_LOG_ERROR("Error initializing the satellite attributes library");
        exit(EXIT_FAILURE);
    }

    /* Do special processing on some parameters that were read. */
    if (process_parameters() != SUCCESS)
    {
        IAS_LOG_ERROR("Error processing parameters");
        exit(EXIT_FAILURE);
    }
    
    /* Get the flag and filename for creating the ANG file */
    create_ang_flag = get_create_ang_flag();   

    /* get the list of bands to perform the resampling on */
    orig_band_index_list_ptr = get_band_index_list();

    /* Get the number of processors to use */
    procs_to_use = get_max_processors_to_use();

    /* Adjust for the number of CPUs */
    if (procs_to_use >= num_cpus)
        procs_to_use = num_cpus;

    /* initialize the grid structure */
    if (ias_grid_initialize(&grid, 0, 0) != SUCCESS)
    {
        IAS_LOG_ERROR("Error initializing the grid structure");
        exit(EXIT_FAILURE);
    }

    if (ias_grid_header_read(get_grid_filename(), &grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Error reading the grid header from %s",
                          get_grid_filename());
        exit(EXIT_FAILURE);
    }

    if (ias_grid_band_header_read(get_grid_filename(), &grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Error reading the grid band header from %s",
                           get_grid_filename());
        ias_grid_free(&grid);
        exit(EXIT_FAILURE);
    }

    /* initialize the band list */
    for (index = 0; index < IAS_MAX_NBANDS; index++)
        band_index_list[index] = ERROR;

    /* verify the grid has data for all the bands requested by the user,
       building the list of confirmed bands */
    for (index = 0; index < IAS_MAX_NBANDS; index++)
    {
        /* check if at the end of the band list */
        if (orig_band_index_list_ptr[index] == ERROR)
            break;

        /* verify the requested band is in the grid */
        if (!grid.bands_available[orig_band_index_list_ptr[index]])
        {
            IAS_LOG_WARNING("Band number %d is not present in grid - skipping "
                    "it", ias_sat_attr_convert_band_index_to_number(
                    orig_band_index_list_ptr[index]));
        }
        else
        {
            /* copy the band number to the final list */
            band_index_list[num_bands_processed]
                = orig_band_index_list_ptr[index];
            num_bands_processed++;
        }
    }

    /* if no bands to process, error out */
    if (num_bands_processed < 1)
    {
        IAS_LOG_ERROR("No bands to process");
        ias_grid_free(&grid);
        exit(EXIT_FAILURE);
    }

    /* do a quick consistency check on the grid */
    for (band_index = 0; band_index < grid.nbands; band_index++)
    {
        if (!grid.bands_available[band_index])
            continue;

        for (sca_index = 1; sca_index < grid.scas_per_band[band_index];
             sca_index++)
        {
            if ((grid.gridbands[band_index][0].samps 
                    != grid.gridbands[band_index][sca_index].samps) ||
                (grid.gridbands[band_index][0].lines 
                    != grid.gridbands[band_index][sca_index].lines))
            {
                IAS_LOG_ERROR("Error SCAs have differing lines/samples");
                ias_grid_free(&grid);
                exit(EXIT_FAILURE);
            }
        }
    }

    if (do_terrain_correction())
    {
        /* use the first band being resampled to establish the size the DEM
           image should be */
        band_index = band_index_list[0];

        /* read the DEM into memory */
        if (ias_misc_read_coregistered_single_band_l1g(get_DEM_filename(),
                        &grid.corners, 
                        grid.gridbands[band_index][0].lines,
                        grid.gridbands[band_index][0].samps, 
                        grid.gridbands[band_index][0].pixsize,
                        &dem_image, IAS_I2) == ERROR)
        {
            IAS_LOG_ERROR("Error reading DEM from %s", get_DEM_filename());
            ias_grid_free(&grid);
            exit(EXIT_FAILURE);
        }
    }

    /* read the model */
    model_ptr = ias_model_read(get_model_filename());
    if (!model_ptr)
    {
        IAS_LOG_ERROR("Error reading model file %s", get_model_filename());
        free(dem_image.data);
        ias_grid_free(&grid);
        exit(EXIT_FAILURE);
    }

    /* open the input image */
    input_image_ptr = open_input_image(get_1R_filename(), get_0R_filename());
    if (!input_image_ptr)
    {
        IAS_LOG_ERROR("Error opening L1R file %s for reading", 
                      get_1R_filename());
        ias_los_model_free(model_ptr);
        free(dem_image.data);
        ias_grid_free(&grid);
        exit(EXIT_FAILURE);
    }

    /* open the 1g output file */
    output_image_ptr = open_output_image(get_1G_filename());
    if (output_image_ptr == NULL)
    {
        IAS_LOG_ERROR("Error opening output image: %s", get_1G_filename());
        close_input_image(input_image_ptr);
        ias_los_model_free(model_ptr);
        free(dem_image.data);
        ias_grid_free(&grid);
        exit(EXIT_FAILURE);
    }

    /* initialize the threadpool */
    threadpool_ptr = ias_threadpool_initialize(procs_to_use);
    if (!threadpool_ptr)
    {
        IAS_LOG_ERROR("Error initializing the thread pool");
        close_output_image(output_image_ptr);
        close_input_image(input_image_ptr);
        ias_los_model_free(model_ptr);
        free(dem_image.data);
        ias_grid_free(&grid);
        exit(EXIT_FAILURE);
    }

    /* set up resampling kernel */
    if (get_resampling_type() != NN_RESAMPLING)
    { 
        if (kernel_setup(get_cubic_convolution_alpha()) == ERROR)
        {
            IAS_LOG_ERROR("Setting up the resampling kernel");
            close_output_image(output_image_ptr);
            close_input_image(input_image_ptr);
            ias_threadpool_destroy(threadpool_ptr);
            ias_los_model_free(model_ptr);
            free(dem_image.data);
            ias_grid_free(&grid);
            exit(EXIT_FAILURE);
        }
    }
    
    /* read the CPF parameters */
    if (read_cpf_parameters(get_CPF_filename(), do_apply_cpf_rad_rescale()) 
        != SUCCESS)
    { 
        IAS_LOG_ERROR("Error reading CPF parameters from %s",
            get_CPF_filename());
        close_output_image(output_image_ptr);
        close_input_image(input_image_ptr);
        ias_los_model_free(model_ptr);
        ias_threadpool_destroy(threadpool_ptr);
        free(dem_image.data);
        ias_grid_free(&grid);
        cleanup_kernel();
        exit(EXIT_FAILURE);
    }

    if (create_ang_flag)
    {
        const struct IAS_CPF_EARTH_CONSTANTS *ecpf = get_cpf_earth_constants();

        /* Initialize the leap seconds using the UTC ephemeris epoch time */
        if (ias_math_init_leap_seconds_from_UTC_time(
            model_ptr->spacecraft.ephemeris.utc_epoch_time, 
            &ecpf->leap_seconds_data) != SUCCESS)
        {
            IAS_LOG_ERROR("Initializing leap seconds");
            close_output_image(output_image_ptr);
            close_input_image(input_image_ptr);
            ias_los_model_free(model_ptr);
            ias_threadpool_destroy(threadpool_ptr);
            free(dem_image.data);
            ias_grid_free(&grid);
            cleanup_kernel();
            free_resample_cpf();
            exit(EXIT_FAILURE);
        }

        /* Initialize the threadpool */
        background_thread = ias_threadpool_initialize(NUM_BACKGROUND_THREADS);
        if (!background_thread)
        {
            IAS_LOG_ERROR("Error initializing the background thread");
            close_output_image(output_image_ptr);
            close_input_image(input_image_ptr);
            ias_los_model_free(model_ptr);
            ias_threadpool_destroy(threadpool_ptr);
            free(dem_image.data);
            ias_grid_free(&grid);
            cleanup_kernel();
            free_resample_cpf();
            exit(EXIT_FAILURE);
        }

        /* Connect to NOVAS for sun angle computation */
        if (ias_geo_initialize_novas() != SUCCESS)
        {
            IAS_LOG_ERROR("Initializing the novas library");
            close_output_image(output_image_ptr);
            close_input_image(input_image_ptr);
            ias_los_model_free(model_ptr);
            ias_threadpool_destroy(threadpool_ptr);
            free(dem_image.data);
            ias_grid_free(&grid);
            cleanup_kernel();
            ias_threadpool_destroy(background_thread);
            free_resample_cpf();
            exit(EXIT_FAILURE);
        }

        /* Pull the necessary data from the grid and model */
        if (ias_angle_gen_populate_metadata(ecpf->semi_major_axis, 
            ecpf->semi_minor_axis, &grid, model_ptr, &metadata) != SUCCESS)
        {
            IAS_LOG_ERROR("Retrieving the angle coefficient metadata");
            close_output_image(output_image_ptr);
            close_input_image(input_image_ptr);
            ias_los_model_free(model_ptr);
            ias_threadpool_destroy(threadpool_ptr);
            free(dem_image.data);
            ias_grid_free(&grid);
            cleanup_kernel();
            ias_threadpool_destroy(background_thread);
            free_resample_cpf();
            exit(EXIT_FAILURE);
        }

        /* Disconnect from NOVAS */
        if (ias_geo_shutdown_novas() != SUCCESS)
        { 
            IAS_LOG_ERROR("Shutting down the novas library");
            close_output_image(output_image_ptr);
            close_input_image(input_image_ptr);
            ias_los_model_free(model_ptr);
            ias_threadpool_destroy(threadpool_ptr);
            free(dem_image.data);
            ias_grid_free(&grid);
            cleanup_kernel();
            ias_threadpool_destroy(background_thread);
            free_resample_cpf();
            ias_angle_gen_free(&metadata);
            exit(EXIT_FAILURE);
        }

        /* Initialize the projection transformation */
        if (ias_angle_gen_initialize_transformation(&metadata) != SUCCESS)
        {
            IAS_LOG_ERROR("Initializing the transformation");
            close_output_image(output_image_ptr);
            close_input_image(input_image_ptr);
            ias_los_model_free(model_ptr);
            ias_threadpool_destroy(threadpool_ptr);
            free(dem_image.data);
            ias_grid_free(&grid);
            cleanup_kernel();
            ias_threadpool_destroy(background_thread);
            free_resample_cpf();
            ias_angle_gen_free(&metadata);
            ias_angle_gen_free_transformation(&metadata);
            exit (EXIT_FAILURE);
        }

        /* Set up the coefficient parameters */
        coeff_params.grid_ptr = &grid;
        coeff_params.model_ptr = model_ptr;
        coeff_params.thread_error = 0;
        coeff_params.metadata = &metadata;
    }
    
    /* Cpf is no longer needed */
    free_resample_cpf();

    /* Open the SCA overlap characterization output file if it has been
       requested */
    if (get_sca_overlap_char_type() != SCA_OVERLAP_CHAR_NONE)
    {
        int success = 1;

        /* Open the SCA overlap file */
        sca_overlap_file = open_sca_overlap_file(
            get_sca_overlap_char_filename(), get_sca_overlap_char_type());

        /* If the SCA overlap file opened successfully, enable the work queue
           on the threadpool since it will be used to write to the file */
        if (!sca_overlap_file)
            success = 0;
        else if (ias_threadpool_enable_work_queue(threadpool_ptr) != SUCCESS)
            success = 0;

        /* If either operation failed, clean up */
        if (!success)
        {
            IAS_LOG_ERROR("Error opening SCA overlap characterization output "
                "file %s", get_sca_overlap_char_filename());
            close_sca_overlap_file(sca_overlap_file);
            cleanup_kernel();
            close_output_image(output_image_ptr);
            close_input_image(input_image_ptr);
            ias_threadpool_destroy(threadpool_ptr);
            ias_los_model_free(model_ptr);
            free(dem_image.data);
            ias_grid_free(&grid);
            if (background_thread)
                ias_threadpool_destroy(background_thread);
            ias_angle_gen_free(&metadata);
            ias_angle_gen_free_transformation(&metadata);
            exit(EXIT_FAILURE);
        }
    }

    /* process all the requested bands */
    for (index = 0; index < num_bands_processed; index++)
    {
        int band_number;
        int status;

        band_index = band_index_list[index];
        coeff_params.band_index = band_index;

        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Failed to convert band index %d to band number",
                          band_index);
            cleanup_kernel();
            close_output_image(output_image_ptr);
            close_input_image(input_image_ptr);
            ias_threadpool_destroy(threadpool_ptr);
            ias_los_model_free(model_ptr);
            free(dem_image.data);
            ias_grid_free(&grid);
            if (background_thread)
                ias_threadpool_destroy(background_thread);
            ias_angle_gen_free(&metadata);
            ias_angle_gen_free_transformation(&metadata);
            close_sca_overlap_file(sca_overlap_file);
            exit(EXIT_FAILURE);
        }

        /* Set up the range constraints on the output.  Only apply radiance
           rescaling if the input image is an L1R and the parameters indicate
           radiance rescaling should be done. */
        if (setup_resamp_output(input_image_ptr->is_l1r 
                && do_apply_cpf_rad_rescale(), band_number) != SUCCESS)
        {
            IAS_LOG_ERROR("Failed to setup resampling output for band "
                "number %d", band_number);
            cleanup_kernel();
            close_output_image(output_image_ptr);
            close_input_image(input_image_ptr);
            ias_threadpool_destroy(threadpool_ptr);
            ias_los_model_free(model_ptr);
            free(dem_image.data);
            ias_grid_free(&grid);
            if (background_thread)
                ias_threadpool_destroy(background_thread);
            ias_angle_gen_free(&metadata);
            ias_angle_gen_free_transformation(&metadata);
            close_sca_overlap_file(sca_overlap_file);
            exit(EXIT_FAILURE);
        }

        /* Get pointer information for the band being processed */
        if (ias_grid_band_pointers_read(get_grid_filename(), &grid, band_number)
                != SUCCESS)
        {
            IAS_LOG_ERROR("Error reading the grid pointers from %s",
                          get_grid_filename());
            cleanup_kernel();
            close_output_image(output_image_ptr);
            close_input_image(input_image_ptr);
            ias_threadpool_destroy(threadpool_ptr);
            ias_los_model_free(model_ptr);
            free(dem_image.data);
            ias_grid_free(&grid);
            if (background_thread)
                ias_threadpool_destroy(background_thread);
            ias_angle_gen_free(&metadata);
            ias_angle_gen_free_transformation(&metadata);
            close_sca_overlap_file(sca_overlap_file);
            exit(EXIT_FAILURE);
        }

        if (create_ang_flag)
        {
            /* Calculate the coefficients on the background thread */
            if (ias_threadpool_start_function(background_thread, 
                calculate_angle_coeffs, &coeff_params) != SUCCESS)
            {
                IAS_LOG_ERROR("Starting the calculate coefficients function "
                    "for band number %d", band_number);
                cleanup_kernel();
                close_output_image(output_image_ptr);
                close_input_image(input_image_ptr);
                ias_threadpool_destroy(threadpool_ptr);
                ias_los_model_free(model_ptr);
                free(dem_image.data);
                ias_grid_free(&grid);
                ias_threadpool_destroy(background_thread);
                ias_angle_gen_free(&metadata);
                ias_angle_gen_free_transformation(&metadata);
                close_sca_overlap_file(sca_overlap_file);
                exit(EXIT_FAILURE);
            }
        }

        /* initialize the radiance rescale parameters for this band */
        setup_radiance_parameters(band_index);

        /* Process this band and save the return status to be check after the
           angle coefficients thread has finished so nothing is freed up before
           it finished */
        status = process_band(threadpool_ptr, band_index, &grid, 
            input_image_ptr, output_image_ptr, model_ptr, &dem_image,
            sca_overlap_file);

        if (create_ang_flag)
        {
            /* Wait for background thread if not finished */
            if (ias_threadpool_wait_for_completion(background_thread) 
                != SUCCESS || coeff_params.thread_error > 0)
            {
                IAS_LOG_ERROR("Calculating the coefficients for band number %d",
                    band_number);
                cleanup_kernel();
                close_output_image(output_image_ptr);
                close_input_image(input_image_ptr);
                ias_threadpool_destroy(threadpool_ptr);
                ias_los_model_free(model_ptr);
                free(dem_image.data);
                ias_grid_free(&grid);
                ias_threadpool_destroy(background_thread);
                ias_angle_gen_free(&metadata);
                ias_angle_gen_free_transformation(&metadata);
                close_sca_overlap_file(sca_overlap_file);
                exit(EXIT_FAILURE);
            }
        }
        
        /* Now check the status of process_band() */
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Error resampling band %d", band_number);
            cleanup_kernel();
            close_output_image(output_image_ptr);
            close_input_image(input_image_ptr);
            ias_los_model_free(model_ptr);
            free(dem_image.data);
            ias_grid_free(&grid);
            ias_threadpool_destroy(threadpool_ptr);
            if (background_thread)
                ias_threadpool_destroy(background_thread);
            ias_angle_gen_free(&metadata);
            ias_angle_gen_free_transformation(&metadata);
            close_sca_overlap_file(sca_overlap_file);
            exit(EXIT_FAILURE);
        }

        /* Free grid */
        if (ias_grid_free_band(&grid, band_number) != SUCCESS)
        {
            IAS_LOG_ERROR("Error freeing the grid");
            cleanup_kernel();
            close_output_image(output_image_ptr);
            close_input_image(input_image_ptr);
            ias_los_model_free(model_ptr);
            free(dem_image.data);
            ias_threadpool_destroy(threadpool_ptr);
            if (background_thread)
                ias_threadpool_destroy(background_thread);
            ias_angle_gen_free(&metadata);
            ias_angle_gen_free_transformation(&metadata);
            close_sca_overlap_file(sca_overlap_file);
            exit(EXIT_FAILURE);
        }
    }

    /* Cleanup the ANG variables if used */
    if (background_thread)
    {
        /* Destroy the background thread */
        ias_threadpool_destroy(background_thread);
        background_thread = NULL;
    }

    /* Free the metadata projection transformation */
    ias_angle_gen_free_transformation(&metadata);

    /* Destroy the main thread pool */
    ias_threadpool_destroy(threadpool_ptr);
    threadpool_ptr = NULL;

    /* free the memory storing the DEM data */
    free(dem_image.data);
    dem_image.data = NULL;

    /* clean up */
    cleanup_kernel();

    /* free the model */
    ias_los_model_free(model_ptr);
    model_ptr = NULL;

    /* update the image metadata before closing the input and output images */
    if (update_l1g_metadata(output_image_ptr, input_image_ptr, &grid,
        num_bands_processed, band_index_list, sizeof(metadata.landsat_scene_id),
        sizeof(metadata.spacecraft_id), metadata.landsat_scene_id,
        metadata.spacecraft_id) == ERROR)
    {
        IAS_LOG_ERROR("Error updating metadata in file %s", get_1G_filename());
        close_output_image(output_image_ptr);
        close_input_image(input_image_ptr);
        ias_angle_gen_free(&metadata);
        ias_grid_free(&grid);
        close_sca_overlap_file(sca_overlap_file);
        exit(EXIT_FAILURE);
    }

    if (create_ang_flag)
    {
        char directory[PATH_MAX];   /* L1G directory path */
        char root[PATH_MAX];        /* L1G base name */
        char base_ang[PATH_MAX];    /* ANG base name */
        int status;                 /* Routine return variable */

        /* Create the base ANG filename from the L1G filename. base_ang is used
         * as a placeholder for the extension since it will not be used */
        ias_misc_split_filename(get_1G_filename(), directory, root, base_ang);
        status = snprintf(base_ang, sizeof(base_ang), "%s%s", directory, root);
        if (status < 0 || status >= sizeof(base_ang))
        {
            IAS_LOG_ERROR("Creating the base ANG filename");
            close_output_image(output_image_ptr);
            close_input_image(input_image_ptr);
            ias_angle_gen_free(&metadata);
            ias_grid_free(&grid);
            close_sca_overlap_file(sca_overlap_file);
            exit(EXIT_FAILURE);
        } 

        /* Create the coefficients metadata filename */
        if (ias_angle_gen_create_ang_filename(base_ang, sizeof(ang_filename), 
            ang_filename) != SUCCESS)
        {
            IAS_LOG_ERROR("Creating the ANG filename");
            close_output_image(output_image_ptr);
            close_input_image(input_image_ptr);
            ias_angle_gen_free(&metadata);
            ias_grid_free(&grid);
            close_sca_overlap_file(sca_overlap_file);
            exit(EXIT_FAILURE);          
        }
        
        /* Write out the coefficients metadata file */
        if (ias_angle_gen_write_ang(ang_filename, &metadata) != SUCCESS)
        {
            IAS_LOG_ERROR("Writing out coefficients metadata file %s", 
                ang_filename);
            close_output_image(output_image_ptr);
            close_input_image(input_image_ptr);
            ias_angle_gen_free(&metadata);
            ias_grid_free(&grid);
            close_sca_overlap_file(sca_overlap_file);
            exit(EXIT_FAILURE);
        }

        /* Free up the ephemeris data */
        ias_angle_gen_free(&metadata);
    }

    /* free grid */
    if (ias_grid_free(&grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Error freeing the grid");
        close_output_image(output_image_ptr);
        close_input_image(input_image_ptr);
        close_sca_overlap_file(sca_overlap_file);
        exit(EXIT_FAILURE);
    }

    /* close the input image */
    if (close_input_image(input_image_ptr) != SUCCESS)
    {
        IAS_LOG_ERROR("Error closing L1R %s", get_1R_filename());
        close_output_image(output_image_ptr);
        close_sca_overlap_file(sca_overlap_file);
        exit(EXIT_FAILURE);
    }

    /* write out an envi header file to help loading the file in envi */
    if (write_envi_hdr(output_image_ptr) != SUCCESS)
    {
        IAS_LOG_ERROR("Error writing envi header file for %s",
                      get_1G_filename());
        close_output_image(output_image_ptr);
        close_sca_overlap_file(sca_overlap_file);
        exit(EXIT_FAILURE);
    }

    /* close the output 1g image after updating the metadata */
    if (close_output_image(output_image_ptr) == ERROR)
    {
        IAS_LOG_ERROR("Error closing output image");
        close_sca_overlap_file(sca_overlap_file);
        exit(EXIT_FAILURE);
    }
    
    /* Close the SCA overlap characterization file */
    if (sca_overlap_file
        && (close_sca_overlap_file(sca_overlap_file) != SUCCESS))
    {
        IAS_LOG_ERROR("Error closing SCA overlap characterization file");
        exit(EXIT_FAILURE);
    }

    /* Write the L1G filename to the OMF along with the ANG filename */
    if (update_omf(get_omf_filename(), get_1G_filename(), ang_filename,
            get_sca_overlap_char_filename(), get_processing_pass()) != SUCCESS)
    {
        IAS_LOG_ERROR("Failed to update OMF file %s", get_omf_filename());
        exit(EXIT_FAILURE);
    }

    /* generate exit message */
    IAS_LOG_INFO("Successful Completion");

    return EXIT_SUCCESS;
}
