/******************************************************************************
NAME:           resample_image

PURPOSE:
Perform geometric resampling on an image using cubic convolution or nearest 
neighbor resampling.

******************************************************************************/
 
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "ias_logging.h"
#include "ias_grid.h"
#include "ias_los_model.h"
#include "ias_geo.h"
#include "ias_miscellaneous.h"
#include "resample_image.h"/* prototypes for this module */
#include "debug.h"         /* ASSERT macro */
#include "kernel.h"        /* get_resample_weight_table_ptr prototype */
#include "get_resample_parameters.h"
#include "get_output_image_buffer.h"
#include "get_cpf_parameters.h"
#include "akima.h"
#include "track_tile.h"
#include "transform_window.h"
#include "sca_overlap_io.h"

/* This structure is used to track the SCAs that need to be combined to make
   an output tile */
typedef struct overlapping_scas
{
    int sca_index;          /* SCA index for this record */
    TILE_TYPE input_tile;   /* L1R image input tile for this SCA */

    /* The L1R line/sample locations are used to track the locations for the
       SCA overlap characterization algorithm */
    float *l1r_lines;       /* L1R line for every pixel in the output tile */
    float *l1r_samples;     /* L1R sample for every pixel in the output tile */
} OVERLAPPING_SCAS;

/* Define the structure to hold parameters to the write_sca_overlap routine */
typedef struct sca_overlap_write_params
{
    SCA_OVERLAP_IO *sca_overlap_file; /* SCA file pointer */
    SCA_OVERLAP_PTR records;          /* Pointer to the records to write */
    int record_count;                 /* Count of records to write */
} SCA_OVERLAP_WRITE_PARAMS;

/* local versions of the minimum and maximum output range, adjusted for
   the currently selected output type. */
static double min_range;
static double max_range;

/* local copy of the output data type */
static IAS_DATA_TYPE output_data_type;


/******************************************************************************
NAME:           setup_resamp_output

PURPOSE:        
setup_resamp_output configures the resample_image module to generate the 
output image with the requested output data type and dynamic range.

RETURNS:
    SUCCESS or ERROR

NOTES:
- This routine uses the parameters read from the parameter file to 
  create locally cached copies in the most efficient format for this module.
- This routine does not use the min/max values defined for the types in
  limits.h since those are machine specific.  The min/max values here are
  data dependent and cannot change for different machines.

******************************************************************************/

int setup_resamp_output
(
    int apply_radiance_rescale_range,/* I: Determine if the radiance rescale
                                           range should be applied. */
    int band_number                  /* I: band number to set up for */
)
{
    int qcal_min;     /* quantine_cal_min_band_x value */
    int qcal_max;     /* quantine_cal_max_band_x value  */

    output_data_type = get_output_data_type();

    min_range = get_output_range_minimum();
    max_range = get_output_range_maximum();

    /* Get the QCALMIN and QCALMAX values */
    if (ias_sat_attr_get_quantization_cal_min(band_number, &qcal_min)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Error getting qcal min value for band number %d",
            band_number);
        return ERROR;
    }
    if (ias_sat_attr_get_quantization_cal_max(band_number, &qcal_max)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Error getting qcal max value for band number %d",
            band_number);
        return ERROR;
    }

    /* adjust the output range for radiance rescaling */
    if (apply_radiance_rescale_range)
    {
        if (min_range < qcal_min)
            min_range = qcal_min;

        if (max_range > qcal_max)
            max_range = qcal_max;
    }

    if (output_data_type == IAS_BYTE)
    {
        /* byte output so adjust the min and max accordingly */
        if (min_range < 0.0)
            min_range = 0.0;
        else
            min_range = ceil(min_range);
        if (max_range > 255.0)
            max_range = 255.0;
        else
            max_range = floor(max_range);
    }
    else if (output_data_type == IAS_I2)
    {
        /* 2 byte integer output so adjust the min and max accordingly */
        if (min_range < -32768.0)
            min_range = -32768.0;
        else
            min_range = ceil(min_range);
        if (max_range > 32767.0)
            max_range = 32767.0;
        else
            max_range = floor(max_range);
    }
    else if (output_data_type == IAS_UI2)
    {
        /* unsigned 2 byte integer output so adjust the min and max 
           accordingly */
        if (min_range < 0.0)
            min_range = 0.0;
        else
            min_range = ceil(min_range);
        if (max_range > 65535.0)
            max_range = 65535.0;
        else
            max_range = floor(max_range);
    }
    else if (output_data_type == IAS_I4)
    {
        /* byte output so adjust the min and max accordingly */
        if (min_range < -2147483648.0) 
            min_range = -2147483648.0;
        else
            min_range = ceil(min_range);
        if (max_range > 2147483647.0)
            max_range = 2147483647.0;
        else
            max_range = floor(max_range);
    }
    else
    {
        /* must be float output, if it's not an error should have been 
           obtained parsing the input, so just assert here during 
           development.  no range manipulations are needed for the float
           case */
        ASSERT (output_data_type == IAS_R4);
    }

    return SUCCESS;
}

/******************************************************************************
NAME:           convert_data

PURPOSE:                
convert_data converts the intermediate resampling results (currently stored as
doubles) to the requested output data type.

RETURN VALUE:
The type converted data is returned in the out_buf.

******************************************************************************/
static void convert_data
(
    const double *input_ptr,/* I: data to convert to the output type */
    void *out_buf,          /* O: output data buffer */
    int nl,                 /* I: Number of lines in ouput */
    int ns                  /* I: Number of samples in output */
)
{
    int l;  /* counter */
    int pixel_count = nl * ns;

    /* convert the intermediate resampling results to the requested 
       output type */
    if (output_data_type == IAS_UI2)
    {
        unsigned short int *uint16_ptr = out_buf;

        for (l = 0; l < pixel_count; l++)
        {
            /* convert the data, ignoring the sign when rounding since all
               values should be positive */
            *uint16_ptr++ = (unsigned short int)(*input_ptr++ + 0.5);
        }
    }
    else if (output_data_type == IAS_I2)
    {
        short int *int16_ptr = out_buf;

        for (l = 0; l < pixel_count; l++)
        {
            /* convert the data, using signed rounding */
            if (*input_ptr >= 0)
                *int16_ptr++ = (short int)(*input_ptr++ + 0.5);
            else
                *int16_ptr++ = (short int)(*input_ptr++ - 0.5);
        }
    }
    else if (output_data_type == IAS_BYTE)
    {
        unsigned char *char_ptr = out_buf;

        /* convert the data, using unsigned rounding */
        for (l = 0; l < pixel_count; l++)
            *char_ptr++ = (unsigned char)(*input_ptr++ + 0.5);
    }
    else if (output_data_type == IAS_R4)
    {
        float *float_ptr = out_buf;

        /* convert the data, no rounding required */
        for (l = 0; l < pixel_count; l++)
            *float_ptr++ = (float)*input_ptr++;
    }
    else
    {
        int *int32_ptr = out_buf;

        ASSERT(output_data_type == INT32_OUTPUT);

        for (l = 0; l < pixel_count; l++)
        {
            /* convert the data, using signed rounding */
            if (*input_ptr >= 0.0)
                *int32_ptr++ = (int)(*input_ptr++ + 0.5);
            else
                *int32_ptr++ = (int)(*input_ptr++ - 0.5);
        }
    }
}

/******************************************************************************
NAME: wrapped_convert_data

PURPOSE:                
    Wrapper for the test code.  Since the real function is declared static.

RETURN VALUE:
    The type converted data is returned in the out_buf.
******************************************************************************/
void wrapped_convert_data
(
    const double *input_ptr,/* I: data to convert to the output type */
    void *out_buf,          /* O: output data buffer */
    int nl,                 /* I: Number of lines in ouput */
    int ns                  /* I: Number of samples in output */
)
{
    /* set the output type for the test code */
    output_data_type = IAS_UI2;

    convert_data(input_ptr, out_buf, nl, ns);
}

/******************************************************************************
NAME: combine_sca_tiles

PURPOSE: combine_sca_tiles combines the output tiles for the different SCAs
    by averaging the pixels.  The combined data is returned in the first tile
    of the resample_ptr buffer.  Also, collect the information for the SCA
    overlap characterization algorithm.

NOTES:
    - The output tile at the right and bottom edges might not be even multiples
      of the tile size, so the loops in this routine can't be condensed into
      a single loop over all the pixels.
******************************************************************************/
static void combine_sca_tiles
(
    double *resample_ptr,   /* I/O: pointer to num_scas tiles to combine */
    int num_scas,           /* I: number of SCAs to combine */
    int tile_size,          /* I: lines/samples in tile */
    const TILE_TYPE *output_tile, /* I: output tile location */
    double fill,            /* I: value used for fill pixels */
    const OVERLAPPING_SCAS *overlap_scas, /* I: Overlapping SCA information */
    SCA_OVERLAP_PTR overlap_records, /* O: SCA overlap char records generated */
    int *overlap_record_index /* I/O: Index to add records into overlap_records
                                      (updated on return) */
)
{
    int line, samp, current_sca;
    int tile_index = tile_size * tile_size;
    int record_index = *overlap_record_index;
    SCA_OVERLAP_CHAR_TYPE overlap_char_type = get_sca_overlap_char_type();
    int do_overlap = (overlap_char_type != SCA_OVERLAP_CHAR_NONE);
    int do_full_overlap = (overlap_char_type == SCA_OVERLAP_CHAR_FULL);

    for (line = 0; line < output_tile->num_lines; line++)
    {
        int line_index = line * output_tile->num_samples;

        for (samp = 0; samp < output_tile->num_samples; samp++)
        {
            int overlap_index[num_scas]; /* Overlapping SCA indices */
            double l1r_value[num_scas];  /* DN from each overlapping SCA */
            int count = 0;
            double total = 0.0;
            int pixel_index = line_index + samp;

            /* combine the data from each of the SCAs that are present */
            for (current_sca = 0; current_sca < num_scas; current_sca++)
            {
                int sca_index = current_sca * tile_index;
                double data = resample_ptr[sca_index + pixel_index];
                if (data != fill)
                {
                    /* Collect the SCA overlap index and L1R pixel value from
                       each SCA for the SCA overlap char algorithm */
                    if (do_overlap)
                    {
                        overlap_index[count] = current_sca;
                        l1r_value[count] = data;
                    }

                    /* Sum the pixels for taking the average later */
                    total += data;
                    count++;
                }
            }

            /* if there were non-fill pixels, set the output pixel to the
               average value (note that if there weren't any non-fill pixels
               then nothing is done since the pixel in the buffer is already
               set to the fill value) */
            if (count != 0)
            {
                resample_ptr[pixel_index] = total / count;

                /* If collecting overlap statistics and there were exactly two
                   pixels combined, collect the SCA overlap characterization
                   information.  Note that it should be impossible for any
                   given pixel to have more two SCAs contributing to an output
                   pixel, but if it were to happen, those would be ignored
                   here, as intended. */
                if (do_overlap && count == 2)
                {
                    const OVERLAPPING_SCAS *sca1
                            = &overlap_scas[overlap_index[0]];
                    const OVERLAPPING_SCAS *sca2
                            = &overlap_scas[overlap_index[1]];
                    float l1r_line1 = sca1->l1r_lines[pixel_index];
                    float l1r_line2 = sca2->l1r_lines[pixel_index];

                    /* Save all the information if the full set of information
                       has been requested */
                    if (do_full_overlap)
                    {
                        SCA_OVERLAP_RECORD *rec
                            = &overlap_records.full_ptr[record_index];

                        /* Calculate the L1T line/sample location */
                        rec->l1t_line = output_tile->start_line + line;
                        rec->l1t_sample = output_tile->start_sample + samp;

                        /* Set the SCA zone number, which is just the lowest
                           SCA number (which is always the first one here) */
                        rec->sca_zone = sca1->sca_index + 1;

                        /* Get required fields for the first SCA */
                        rec->l1r_value[0] = l1r_value[0];
                        rec->l1r_line[0] = l1r_line1;
                        rec->l1r_sample[0] = sca1->l1r_samples[pixel_index];

                        /* Get required fields for the second SCA */
                        rec->l1r_value[1] = l1r_value[1];
                        rec->l1r_line[1] = l1r_line2;
                        rec->l1r_sample[1] = sca2->l1r_samples[pixel_index];

                        record_index++;
                    }
                    /* For the limited case, only save information for lines
                       that are near integer line numbers in at least one SCA
                       in the L1R image */
                    else if ((fabsf(l1r_line1 - round(l1r_line1)) < 0.1)
                             || (fabsf(l1r_line2 - round(l1r_line2)) < 0.1))
                    {
                        LIMITED_SCA_OVERLAP_RECORD *rec
                            = &overlap_records.limited_ptr[record_index];

                        /* Set the SCA zone number, which is just the lowest
                           SCA number (which is always the first one here) */
                        rec->sca_zone = sca1->sca_index + 1;

                        /* Get required fields for the first SCA */
                        rec->l1r_value[0] = l1r_value[0];
                        rec->l1r_line[0] = l1r_line1;

                        /* Get required fields for the second SCA */
                        rec->l1r_value[1] = l1r_value[1];
                        rec->l1r_line[1] = l1r_line2;

                        record_index++;
                    }
                }
            }
        }
    }
    *overlap_record_index = record_index;
}

/******************************************************************************
NAME: do_ols2ils

PURPOSE:                
do_ols2ils is a wrapper function to convert an output line/sample into an
input line/sample using the elevation if available and the correct ols2ils
routine.

RETURN VALUE:
SUCCESS if the ols2ils function was successful
ERROR   if the ols2ils function was not successful
IAS_GRID_ILS_NOT_CALCULATED if no error was returned, but the ols2ils routine
        could not calculate the ILS

******************************************************************************/
int do_ols2ils
(
    const IAS_GRID_BAND_TYPE *grid_band_ptr, /* I: current grid */
    int terrain_flag,       /* I: terrain correction flag */
    IAS_IMAGE *dem,         /* I: DEM data if doing terrain correction */
    int oline,              /* I: output space line */
    int osamp,              /* I: output space sample */
    double *iline,          /* O: input space line */
    double *isamp           /* O: input space sample */
)
{
    int status;

    if (terrain_flag)
    {
        double elev = 0.0;    /* Elevation of point */

        /* doing terrain correction, so get elevation for image */
        if (ias_misc_read_elevation_at_line_sample(dem, oline, osamp,
                              grid_band_ptr->pixsize,
                              grid_band_ptr->pixsize, &elev) != SUCCESS)
        {
            IAS_LOG_ERROR("Error getting elevation");
            return ERROR;
        }

        /* calculate the input line/sample using the elevation */
        status = ias_grid_3d_ols2ils(grid_band_ptr, oline, osamp, iline, isamp, 
                                     elev);
    }
    else
    {
        /* no elevation correction, so just find the input line/sample using
           the zero plane */
        status = ias_grid_ols2ils(grid_band_ptr, oline, osamp, iline, isamp,
                                  grid_band_ptr->zeroplane);
    }

    return status;
}

/******************************************************************************
NAME: calculate_jitter

PURPOSE: 
compute the line and sample direction jitter corrections for the current input 
line and sample location.  These corrections are the adjustments to the input 
space interpolation location required to compensate for the high frequency 
jitter present at the time of observation.

RETURN VALUE:
none

******************************************************************************/
static void calculate_jitter
(
    double line_delta,       /* I: fractional index */
    int jitter_index,        /* I: index into jitter table */
    int jitter_cell_index,   /* I: index into grid jitter sensitivity tables */
    const IAS_GRID_BAND_TYPE *grid_band_ptr, /* pointer to grid band */
    const IAS_VECTOR *jitter_table, /* I: jitter disturbance table */
    double *line_jitter_adjustment, /* O: line jitter adjustment */
    double *sample_jitter_adjustment /* O: sample jitter adjustment */
)
{
    double sample_jitter;    /* sample jitter correction adjusted for effects
                                of line jitter */
    double line_jitter;      /* error in line coordinate due to jitter */
    double sample_velocity;  /* sample velocity - rate of change of sample 
                                jitter considering line effects */
    double line_velocity;    /* line velocity - rate of change of line 
                                jitter */
    IAS_VECTOR jitter_table_difference; /* difference between 2 jitter table
                                entries */
    IAS_VECTOR *line_sensitivity; /* points to line_sensitivity cell in grid */
    IAS_VECTOR *samp_sensitivity; /* points to samp_sensitivity cell in grid */
    int next_jitter_index;   /* the index after jitter_index */
    double line_delta_difference; /* difference between line_delta and 1.0 */ 

    /* point to grid sensitivity structures.  This is for readability only */
    line_sensitivity = &grid_band_ptr->line_sensitivity[jitter_cell_index];
    samp_sensitivity = &grid_band_ptr->samp_sensitivity[jitter_cell_index];

    next_jitter_index = jitter_index + 1;
    line_delta_difference = 1.0 - line_delta;

    /* calculate sample jitter adjustment */
    sample_jitter = (samp_sensitivity->x * jitter_table[jitter_index].x 
        + samp_sensitivity->y * jitter_table[jitter_index].y 
        + samp_sensitivity->z * jitter_table[jitter_index].z) 
        * line_delta_difference
        + (samp_sensitivity->x * jitter_table[next_jitter_index].x 
        + samp_sensitivity->y * jitter_table[next_jitter_index].y 
        + samp_sensitivity->z * jitter_table[next_jitter_index].z) * line_delta;

    /* find the error in the line coordinate due to line jitter */
    line_jitter = (line_sensitivity->x * jitter_table[jitter_index].x 
        + line_sensitivity->y * jitter_table[jitter_index].y 
        + line_sensitivity->z * jitter_table[jitter_index].z) 
        * line_delta_difference
        + (line_sensitivity->x * jitter_table[next_jitter_index].x 
        + line_sensitivity->y * jitter_table[next_jitter_index].y 
        + line_sensitivity->z * jitter_table[next_jitter_index].z) * line_delta;

    /* find the difference between 2 jitter table entries */
    jitter_table_difference.x = jitter_table[next_jitter_index].x 
        - jitter_table[jitter_index].x;
    jitter_table_difference.y = jitter_table[next_jitter_index].y 
        - jitter_table[jitter_index].y;
    jitter_table_difference.z = jitter_table[next_jitter_index].z 
        - jitter_table[jitter_index].z;

    /* find the rate of change of sample jitter with line coordinate */
    sample_velocity = samp_sensitivity->x * jitter_table_difference.x 
        + samp_sensitivity->y * jitter_table_difference.y 
        + samp_sensitivity->z * jitter_table_difference.z;

    /* find the rate of change of line jitter with line coordinate */
    line_velocity = line_sensitivity->x * jitter_table_difference.x 
        + line_sensitivity->y * jitter_table_difference.y 
        + line_sensitivity->z * jitter_table_difference.z;

    /* find the line and sample jitter correction adjusted for the effects of 
       line jitter.  Both sample_velocity and line_velocity represent the rate
       of change of sample/line (respectivity) jitter with line coordinate */
    *sample_jitter_adjustment = sample_jitter + line_jitter * sample_velocity;
    *line_jitter_adjustment = line_jitter + line_jitter * line_velocity;
}

/******************************************************************************
NAME: calculate_jitter_sample 

PURPOSE: 
compute the sample direction jitter corrections for the current input line and
sample location.  These corrections are the adjustments to the input space 
interpolation location required to compensate for the high frequency jitter
present at the time of observation.

NOTE:
this is a simplified version of calculate_jitter that computes only the sample
direction jitter correction.  It is implemented as a separate function for 
processing efficiency because it is invoked more frequently than 
calculate_jitter.

RETURN VALUE:
sample jitter correction adjusted for the effects of line jitter

******************************************************************************/
static double calculate_jitter_sample
(
    double line_delta,       /* I: fractional index */
    int jitter_index,        /* I: index into jitter table */
    int jitter_cell_index,   /* I: index into grid jitter sensitivity tables */
    const IAS_GRID_BAND_TYPE *grid_band_ptr, /* pointer to grid band */
    const IAS_VECTOR *jitter_table /* I: jitter disturbance table */
)
{
    double sample_jitter;    /* sample jitter correction adjusted for effects
                                of line jitter */
    double line_jitter;      /* error in line coordinate due to jitter */
    double sample_velocity;  /* sample velocity - rate of change of sample 
                                jitter considering line effects */
    IAS_VECTOR *line_sensitivity; /* points to line_sensitivity cell in grid */
    IAS_VECTOR *samp_sensitivity; /* points to samp_sensitivity cell in grid */
    int next_jitter_index;   /* the index after jitter_index */
    double line_delta_difference; /* difference between line_delta and 1.0 */ 

    /* point to grid sensitivity structures.  This is for readability only */
    line_sensitivity = &grid_band_ptr->line_sensitivity[jitter_cell_index];
    samp_sensitivity = &grid_band_ptr->samp_sensitivity[jitter_cell_index];

    next_jitter_index = jitter_index + 1;
    line_delta_difference = 1.0 - line_delta;

    /* calculate sample jitter adjustment */
    sample_jitter = (samp_sensitivity->x * jitter_table[jitter_index].x 
        + samp_sensitivity->y * jitter_table[jitter_index].y 
        + samp_sensitivity->z * jitter_table[jitter_index].z) 
        * line_delta_difference 
        + (samp_sensitivity->x * jitter_table[next_jitter_index].x 
        + samp_sensitivity->y * jitter_table[next_jitter_index].y 
        + samp_sensitivity->z * jitter_table[next_jitter_index].z) * line_delta;

    /* find the error in the line coordinate due to line jitter */
    line_jitter = (line_sensitivity->x * jitter_table[jitter_index].x 
        + line_sensitivity->y * jitter_table[jitter_index].y 
        + line_sensitivity->z * jitter_table[jitter_index].z) 
        * line_delta_difference
        + (line_sensitivity->x * jitter_table[next_jitter_index].x 
        + line_sensitivity->y * jitter_table[next_jitter_index].y 
        + line_sensitivity->z * jitter_table[next_jitter_index].z) * line_delta;

    /* find the rate of change of sample jitter with line coordinate */
    sample_velocity = samp_sensitivity->x * (jitter_table[next_jitter_index].x 
        - jitter_table[jitter_index].x) + samp_sensitivity->y
        * (jitter_table[next_jitter_index].y - jitter_table[jitter_index].y)
        + samp_sensitivity->z * (jitter_table[next_jitter_index].z 
        - jitter_table[jitter_index].z);

    /* find the sample jitter correction by multiplying the line jitter by
       the rate of change of sample jitter with line coordinate */
    sample_jitter += line_jitter * sample_velocity;

    return sample_jitter;
}

/******************************************************************************
NAME: adjust_sample 

PURPOSE: 
calculate jitter adjustment for a given sample

RETURN VALUE:
sample jitter correction adjusted for the effects of line jitter

******************************************************************************/
static double adjust_sample
(
    const double *along_track_detector_offset, /* I: along-track detector 
                                    offsets */
    const double *across_track_detector_offset, /* I: across-track detector 
                                    offsets */
    int cell_index,           /* I: cell index into the grid band structure */
    double elevation,         /* I: elevation for the current pixel */
    int current_isamp,        /* I: current sample number */
    double jitter_scale,      /* I: scale factor to allow the MS bands to use 
                                    the PAN jitter table */
    int jitter_cell_index,    /* I: index into grid jitter sensitivity tables */
    const IAS_GRID_BAND_TYPE *grid_band_ptr, /* pointer to grid band */
    const IAS_VECTOR *jitter_table, /* I: jitter disturbance table */
    int jitter_max_index,     /* I: maximum index into the jitter table */ 
    double d_iline,           /* I: double input line */
    double delta_line_oe,     /* I: detector odd/even offset/delay in line
                                    direction */
    int nom_fill              /* I: nominal fill */ 
)
{
    double jitter_line;       /* Adjusted line position to perform jitter 
                                 look-up */
    int jitter_index;         /* Jitter table index */
    double jitter_line_delta; /* Fractional jitter sample offset */
    double scale;             /* Scale to find the line */
    double adjusted_sample;   /* sample jitter correction adjusted for effects
                                 of line jitter */
    double delta_samp_oe;     /* detector odd/even offset/delay in sample 
                                 direction */

    /* find the delta values in the line and sample direction for the current
       sample.  The even and odd detectors are calculated from different grid
       entries.  The grid holds 2 coefficients that need to be applied */ 
    scale = floor(along_track_detector_offset[current_isamp] + 0.5);
    delta_samp_oe = grid_band_ptr->delta_samp_oe[cell_index] + elevation
        * grid_band_ptr->delta_samp_oe[cell_index + 1];
    delta_samp_oe *= scale;

    /* apply detector and terrain offsets to the current sample location */ 
    adjusted_sample = current_isamp 
        - across_track_detector_offset[current_isamp] - delta_samp_oe;

    /* calculate the index into the jitter table for the current image line */
    jitter_line = d_iline + delta_line_oe * scale - nom_fill 
        - along_track_detector_offset[current_isamp];
    jitter_index = (int)(jitter_scale * jitter_line);

    /* if the jitter index isn't within the range of the jitter table, set it
       to the appropriate minimum or maximum value */
    if (jitter_index < 1)
        jitter_index = 1;
    if (jitter_index > jitter_max_index - 1)
        jitter_index = jitter_max_index - 1;

    /* calculate the fractional jitter table index */
    jitter_line_delta = jitter_scale * jitter_line - floor(jitter_scale 
        * jitter_line);

    adjusted_sample -= calculate_jitter_sample(jitter_line_delta, jitter_index,
        jitter_cell_index, grid_band_ptr, jitter_table);

    return adjusted_sample;
}

/******************************************************************************
Name: allocate_overlapping_scas

Purpose: Allocates the structure for tracking information about the 
    overlapping SCAs.

Returns:
    Pointer to the allocated structure, or NULL if an error occurs
******************************************************************************/
static OVERLAPPING_SCAS *allocate_overlapping_scas
(
    int tiles_to_allocate,  /* I: Number of tiles to allocate */
    int tile_size,          /* I: Size of tile in one dimension */
    int include_l1r_locations_flag /* I: Flag to include the fields for
                                         tracking the L1R locations for the
                                         SCA overlap characterization alg */
)
{
    OVERLAPPING_SCAS *overlap_scas;
    int tile_entries = tile_size * tile_size;
    float *buffer;
    int i;

    /* Allocate memory for tracking the SCAs that intersect the current tile */
    overlap_scas = malloc(tiles_to_allocate * sizeof(*overlap_scas));
    if (overlap_scas == NULL)
    {
        IAS_LOG_ERROR("Error allocating SCA list buffer");
        return NULL;
    }

    /* If requested, allocate the buffer for tracking the L1R line and sample
       for the SCA overlap characterization.  A single buffer is allocated
       which is then divided up between the tiles. */
    if (include_l1r_locations_flag)
    {
        buffer = malloc(2 * tiles_to_allocate * tile_entries * sizeof(*buffer));
        if (buffer == NULL)
        {
            IAS_LOG_ERROR("Error allocating L1R location buffer");
            free(overlap_scas);
            return NULL;
        }

        /* For each tile requested, assign a portion of the buffer for the L1R
           lines and samples */
        for (i = 0; i < tiles_to_allocate; i++)
        {
            overlap_scas[i].l1r_lines = &buffer[2 * i * tile_entries];
            overlap_scas[i].l1r_samples = &buffer[(2 * i + 1) * tile_entries];
        }
    }
    else
    {
        /* L1R location buffers not requested, so make sure the pointers are
           NULL */
        for (i = 0; i < tiles_to_allocate; i++)
        {
            overlap_scas[i].l1r_lines = NULL;
            overlap_scas[i].l1r_samples = NULL;
        }
    }

    return overlap_scas;
}

/******************************************************************************
Name: free_overlapping_scas

Purpose: Frees an overlapping SCA structure

Returns: Nothing
******************************************************************************/
static void free_track_scas
(
    OVERLAPPING_SCAS *overlap_scas  /* I: Pointer to the SCA overlap records */
)
{
    if (overlap_scas)
    {
        /* Free the L1R line and sample buffer, which is pointed to by the
           first entry of the l1r_lines member */
        free(overlap_scas[0].l1r_lines);
        overlap_scas[0].l1r_lines = NULL;
        free(overlap_scas);
    }
}

/******************************************************************************
Name: write_sca_overlap

Purpose: Calls the write_sca_overlap_records routine.  This is called
    through the threadpool work queue and is always running on the same thread.
    So no locking is required when writing the file.

Returns:
    SUCCESS or ERROR
******************************************************************************/
static int write_sca_overlap
(
    void *message       /* I: Input message from the work queue */
)
{
    SCA_OVERLAP_WRITE_PARAMS *params = (SCA_OVERLAP_WRITE_PARAMS *)message;
    int status;

    /* Call the real write routine */
    status = write_sca_overlap_records(params->sca_overlap_file,
                params->records, params->record_count);

    /* Free the records and message since ownership belongs to this routine */
    free(params->records.full_ptr);
    free(params);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing sca overlap characterization records");
        return ERROR;
    }

    return SUCCESS;
}

/******************************************************************************
Name: submit_sca_overlap_write

Purpose: Inserts requests to write data to the SCA overlap file to the
    threadpool work queue.  This allows the writes to happen on the main thread
    without needing to worry about locking access to the file. 

Returns:
    SUCCESS or ERROR
******************************************************************************/
static int submit_sca_overlap_write
(
    struct ias_threadpool *threadpool,  /* I: Theadpool to submit the work to */
    SCA_OVERLAP_IO *sca_overlap_file,   /* I: SCA overlap file pointer */
    SCA_OVERLAP_PTR sca_overlap_records,/* I: Pointer to records to write */
    int record_count                    /* I: Count of records to write */
)
{
    SCA_OVERLAP_WRITE_PARAMS *write_params;

    /* Allocate a write parameters message */
    write_params = malloc(sizeof(*write_params));
    if (!write_params)
    {
        IAS_LOG_ERROR("Error allocating write parameters buffer");
        return ERROR;
    }

    /* Set the write message contents */
    write_params->sca_overlap_file = sca_overlap_file;
    write_params->records = sca_overlap_records;
    write_params->record_count = record_count;

    /* Add the call to write the data to the work queue */
    if (ias_threadpool_add_to_work_queue(threadpool, write_sca_overlap,
                write_params) != SUCCESS)
    {
        IAS_LOG_ERROR("Failed inserting write SCA overlap call into work "
            "queue");
        free(write_params);
        return ERROR;
    }

    return SUCCESS;
}

/******************************************************************************
NAME: resample_image

PURPOSE:
resample_image is the resampling driver routine.  It performs all the 
operations that need to be done, no matter what type of resampling is being
performed.  That includes handling the multithreading and doing image I/O.

RETURN VALUE:
The thread_error member of the parameters is set to the number of the thread
encountering an error.  If no errors are encountered, the thread_error member
is not modified.

******************************************************************************/

int resample_image
(
    void *params,   /* I: pointer to input parameter structure */
    int thread_num  /* I: thread number for the thread running this routine */
)
{
    /* create local copies of the parameters */
    RESAMPLE_THREAD_PARAMS *param_ptr = (RESAMPLE_THREAD_PARAMS *)params;
    RESAMPLE_FUNCTION resample_function = param_ptr->resample_function;
    const IAS_GRID_TYPE *grid_ptr = param_ptr->grid_ptr;
    const IAS_LOS_MODEL *model_ptr = param_ptr->model_ptr;
    int sca = param_ptr->sca;
    int combine_scas = param_ptr->combine_scas;
    IAS_IMAGE *dem = param_ptr->dem_image;
    int band_index = param_ptr->band_index;
    INPUT_IMAGE *input_image = param_ptr->input_image;
    OUTPUT_BUFFER *output_image_ptr = param_ptr->output_image_ptr;
    IAS_THREAD_MUTEX_TYPE *hdf_io_mutex = &param_ptr->hdf_io_mutex;
    int tiles_to_allocate;

    void *out_buf;                /* Output image buffer */
    void *fill_buf;               /* buffer of all fill */
    void *write_buf_ptr;          /* pointer to buffer to write */
    int oline, osamp;             /* output line/sample */
    double fill = get_fill_pixel(); /* fill pixel */
    double *resample_ptr;

    int terrain_flag;         /* Flag to indicate terrain processing */

    int tile_size = get_output_tile_size(); /* output tile size width/height */
    TILE_TYPE output_tile;   /* current output tile location */
    int tile_number;         /* current output tile number */
    float *input_buf;        /* input tile buffer */
    int input_buf_size;      /* size of the input_buf in bytes */

    int band_number;         /* current band number */
    const IAS_BAND_ATTRIBUTES *band_attributes = NULL; /* attributes of band
                                band_number */
    int sensor_id;           /* id of the sensor the band belongs to */
    double jitter_scale;     /* scale factor to allow the MS bands to use the
                                PAN jitter table */
    int tile_pixels = tile_size * tile_size;

    /* The following variables are used to track the SCA overlap
       characterization information */
    SCA_OVERLAP_PTR sca_overlap_records = {NULL};
    SCA_OVERLAP_CHAR_TYPE overlap_type = get_sca_overlap_char_type();
    int overlap_record_size = 0;
    int do_overlap = 1;
    int allocated_sca_overlap_records = 0;
    int sca_overlap_record_count = 0;

    /* Determine whether the SCA overlap characterization has been requested
       and how big the overlap record should be */
    if (overlap_type == SCA_OVERLAP_CHAR_LIMITED)
        overlap_record_size = sizeof(LIMITED_SCA_OVERLAP_RECORD);
    else if (overlap_type == SCA_OVERLAP_CHAR_FULL)
        overlap_record_size = sizeof(SCA_OVERLAP_RECORD);
    else
        do_overlap = 0;

    /* get the band number */
    band_number = ias_sat_attr_convert_band_index_to_number(band_index);
    if (band_number == ERROR)
    {
        IAS_LOG_ERROR("Retrieving band number for band index %d", band_index);
        param_ptr->thread_error = thread_num + 1;
        return ERROR;
    }

    /* get the band attributes */
    band_attributes = ias_sat_attr_get_band_attributes(band_number);
    if (!band_attributes)
    {
        IAS_LOG_ERROR("Retrieving band attributes for band number %d",
                band_number);
        param_ptr->thread_error = thread_num + 1;
        return ERROR;
    }

    /* get the sensor Id */
    sensor_id = band_attributes->sensor_id;

    /* get the jitter scale */
    jitter_scale = 
        model_ptr->sensor.sensors[sensor_id].jitter_entries_per_frame /
        model_ptr->sensor.bands[band_index].sampling_char.lines_per_frame;

    /* allocate memory for the output image tile */
    out_buf = get_output_image_buffer(tile_size, tile_size);
    if (out_buf == NULL)
    {
        IAS_LOG_ERROR("Error allocating output buffer");
        param_ptr->thread_error = thread_num + 1;
        return ERROR;
    }

    /* determine the number of tiles to allocate for resampling.  One for each
       SCA if combining the output, otherwise a single buffer is enough. */
    if (combine_scas)
        tiles_to_allocate = grid_ptr->scas_per_band[band_index]; 
    else
        tiles_to_allocate = 1;

    /* allocate a buffer for the resampling output for the required number
       of tiles */
    resample_ptr = malloc(tiles_to_allocate * sizeof(double)
                          * tile_size * tile_size);
    if (resample_ptr == NULL)
    {
        IAS_LOG_ERROR("Error allocating intermediate buffer");
        free(out_buf);
        param_ptr->thread_error = thread_num + 1;
        return ERROR;
    }

    /* Allocate memory for tracking the SCAs that intersect the current tile */
    OVERLAPPING_SCAS *overlap_scas = allocate_overlapping_scas(
            tiles_to_allocate, tile_size, do_overlap);
    if (overlap_scas == NULL)
    {
        IAS_LOG_ERROR("Error allocating SCA list buffer");
        free(resample_ptr);
        free(out_buf);
        param_ptr->thread_error = thread_num + 1;
        return ERROR;
    }

    /* allocate memory for the input tile buffer, assuming that twice the size
       of the output tile should be large enough for a starting guess */
    input_buf_size = 2 * tile_size * tile_size * sizeof(float);
    input_buf = malloc(input_buf_size);
    if (input_buf == NULL)
    {
        IAS_LOG_ERROR("Error allocating input buffer");
        free(out_buf);
        free(resample_ptr);
        free_track_scas(overlap_scas);
        param_ptr->thread_error = thread_num + 1;
        return ERROR;
    }

    /* create a buffer of all fill as an optimization for when an output
       tile is all fill */
    fill_buf = get_output_image_buffer(tile_size, tile_size);
    if (fill_buf == NULL)
    {
        IAS_LOG_ERROR("Error allocating fill buffer");
        free(input_buf);
        free(out_buf);
        free(resample_ptr);
        free_track_scas(overlap_scas);
        param_ptr->thread_error = thread_num + 1;
        return ERROR;
    }
    for (oline = 0; oline < tile_size; oline++)
    {
        int line_index = oline * tile_size;
        for (osamp = 0; osamp < tile_size; osamp++)
        {
            resample_ptr[line_index + osamp] = fill;
        }
    }
    /* Convert the fill data from double to output data type */
    convert_data(resample_ptr, fill_buf, tile_size, tile_size);

    terrain_flag = do_terrain_correction();

    /* while there are tiles left to be assigned, do the resampling for each
       tile */
    while (get_assigned_output_tile(&output_tile, &tile_number) == 1)
    {
        int current_sca;    /* SCA loop counter */
        int end_sca;        /* ending SCA for the SCA loop */
        int valid_scas = 0; /* count of SCAs that were not completely fill
                               in the output tile */
        int overlap_index = 0;
        int sca_loop;

        /* If any threads have encountered errors, don't bother doing any
           more processing */
        if (ias_threadpool_get_worker_return_status(param_ptr->threadpool)
            != SUCCESS)
        {
            IAS_LOG_WARNING("Exiting thread number %d since another thread "
                "encountered an error", thread_num);
            param_ptr->thread_error = thread_num + 1;
            free(out_buf);
            free(resample_ptr);
            free_track_scas(overlap_scas);
            free(input_buf);
            free(fill_buf);
            return ERROR;
        }

        /* if combining SCAs, loop over them all; otherwise, just do the 
           SCA passed in */
        if (combine_scas)
        {
            sca = 0;
            end_sca = grid_ptr->scas_per_band[band_index];
        }
        else
            end_sca = sca + 1;

        /* Determine which SCAs overlap this output tile */
        for (current_sca = sca; current_sca < end_sca; current_sca++)
        {
            /* get the grid pointer for the current SCA and band */
            const IAS_GRID_BAND_TYPE *grid_band_ptr = 
                    &grid_ptr->gridbands[band_index][current_sca];
            TILE_TYPE input_tile;

            /* calculate the input tile needed to create the assigned
               output tile */
            if (calculate_input_tile_to_do(grid_band_ptr, input_image,
                    terrain_flag, dem, param_ptr->max_line_offsets[current_sca],
                    &output_tile, &input_tile) == ERROR)
            {
                IAS_LOG_ERROR("Calculating input tile for output tile start "
                    "line %d end line %d start sample %d end sample %d SCA %d",
                    output_tile.start_line, output_tile.end_line, 
                    output_tile.start_sample, output_tile.end_sample, 
                    current_sca);
                param_ptr->thread_error = thread_num + 1;
                free(input_buf);
                free(out_buf);
                free(resample_ptr);
                free_track_scas(overlap_scas);
                free(fill_buf);
                return ERROR;
            }

            /* if the needed input tile is outside the input image, skip to
               the next SCA */
            if ((input_tile.num_lines <= 0) || (input_tile.num_samples <= 0))
                continue;

            /* Save the overlapping SCA info */
            overlap_scas[overlap_index].sca_index = current_sca;
            overlap_scas[overlap_index].input_tile = input_tile;
            overlap_index++;
        }

        /* Process each of the overlapping SCAs */
        for (sca_loop = 0; sca_loop < overlap_index; sca_loop++)
        {
            current_sca = overlap_scas[sca_loop].sca_index;
            TILE_TYPE input_tile = overlap_scas[sca_loop].input_tile;
            int size;
            const IAS_GRID_BAND_TYPE *grid_band_ptr = 
                    &grid_ptr->gridbands[band_index][current_sca];
            float *l1r_lines = NULL;
            float *l1r_samples = NULL;

            /* If the SCA overlap characterization info has been requested
               and more than one SCA is being processed, set up the pointers
               for the L1R lines and samples */
            if (do_overlap && overlap_index > 1)
            {
                l1r_lines = overlap_scas[valid_scas].l1r_lines;
                l1r_samples = overlap_scas[valid_scas].l1r_samples;
            }

            /* an input tile that is inside the image was successfully 
               found, so read that tile into memory.  First calculate the size
               of the needed input buffer */
            size = input_tile.num_lines * input_tile.num_samples 
                    * sizeof(float);

            /* if the current buffer isn't large enough, allocate a larger
               buffer */
            if (size > input_buf_size)
            {
                free(input_buf);
                input_buf_size = size;
                input_buf = malloc(input_buf_size);
                if (input_buf == NULL)
                {
                    IAS_LOG_ERROR("Error re-allocating input buffer");
                    param_ptr->thread_error = thread_num + 1;
                    free(out_buf);
                    free(fill_buf);
                    free(resample_ptr);
                    free_track_scas(overlap_scas);
                    return ERROR;
                }
            }

            /* read the input image tile into the buffer */
            if (read_input_image_data(input_image, current_sca, 
                      &input_tile, input_buf) != SUCCESS)
            {
                IAS_LOG_ERROR("Error reading input tile");
                free(out_buf);
                free(fill_buf);
                free(resample_ptr);
                free_track_scas(overlap_scas);
                free(input_buf);
                param_ptr->thread_error = thread_num + 1;
                return ERROR;
            }

            /* call the correct resampling routine */
            if (resample_function(grid_band_ptr, model_ptr, current_sca, 
                    band_index, &output_tile, dem, &input_tile, input_buf, 
                    input_image->nl, input_image->ns,
                    param_ptr->max_line_offsets[current_sca], jitter_scale,
                    param_ptr->trim_lut,
                    &resample_ptr[valid_scas * tile_size * tile_size],
                    l1r_lines, l1r_samples) == ERROR)
            {
                IAS_LOG_ERROR("Error with resample_function");
                free(out_buf);
                free(fill_buf);
                free(resample_ptr);
                free_track_scas(overlap_scas);
                free(input_buf);
                param_ptr->thread_error = thread_num + 1;
                return ERROR;
            }

            /* another valid (i.e. non-fill) SCA was processed */
            valid_scas++;

        } /* end of SCA loop */

        if (valid_scas == 0)
        {
            /* no valid SCAs were processed, so use the fill buffer for
               writing */
            write_buf_ptr = fill_buf;
        }
        else
        {
            /* non-fill data present, so use the output buffer to write */
            write_buf_ptr = out_buf;

            if (combine_scas && (valid_scas > 1))
            {
                /* If collecting the SCA overlap char information, handle the
                   record allocation and writing */
                if (do_overlap)
                {
                    /* Make sure there are enough free SCA overlap records to
                       store the potential information */
                    if (sca_overlap_record_count + tile_pixels 
                            > allocated_sca_overlap_records)
                    {
                        /* If there records, submit them for writing */
                        if (sca_overlap_record_count > 0)
                        {
                            if (submit_sca_overlap_write(param_ptr->threadpool,
                                    param_ptr->sca_overlap_file,
                                    sca_overlap_records,
                                    sca_overlap_record_count) != SUCCESS)
                            {
                                IAS_LOG_ERROR("Submitting SCA overlap records "
                                        "for write");
                                free(out_buf);
                                free(fill_buf);
                                free(resample_ptr);
                                free_track_scas(overlap_scas);
                                free(input_buf);
                                param_ptr->thread_error = thread_num + 1;
                                return ERROR;
                            }

                            sca_overlap_records.full_ptr = NULL;
                            sca_overlap_record_count = 0;
                            allocated_sca_overlap_records = 0;
                        }

                        /* Allocate a new set of records */
                        allocated_sca_overlap_records = tile_pixels * 10;
                        sca_overlap_records.full_ptr
                            = malloc(allocated_sca_overlap_records
                                    * overlap_record_size);
                        if (sca_overlap_records.full_ptr == NULL)
                        {
                            IAS_LOG_ERROR("Error allocating SCA overlap record "
                                    "buffer");
                            free(out_buf);
                            free(fill_buf);
                            free(resample_ptr);
                            free_track_scas(overlap_scas);
                            free(input_buf);
                            param_ptr->thread_error = thread_num + 1;
                            return ERROR;
                        }
                    }
                }

                /* combining SCAs and more than one SCA intersected this output 
                   tile, so average the contents of the valid SCAs into a single
                   output buffer */
                combine_sca_tiles(resample_ptr, valid_scas, tile_size, 
                    &output_tile, fill, overlap_scas, sca_overlap_records,
                    &sca_overlap_record_count);
            }

            /* Convert data from double to output data type */
            convert_data(resample_ptr, out_buf, output_tile.num_lines, 
                         output_tile.num_samples);
        }

        /* Write output line to 1G HDF */
        if (write_output_image_data(output_image_ptr, &output_tile, 
                            hdf_io_mutex, write_buf_ptr) != SUCCESS)
        {
            IAS_LOG_ERROR("Error writing data");
            free(out_buf);
            free(fill_buf);
            free(resample_ptr);
            free_track_scas(overlap_scas);
            free(input_buf);
            param_ptr->thread_error = thread_num + 1;
            return ERROR;
        }

    } /* end of assigned tile loop */

    free(input_buf);
    free(out_buf);
    free(fill_buf);
    free(resample_ptr);
    free_track_scas(overlap_scas);

    /* If there SCA overlap records left, write them now */
    if (sca_overlap_record_count > 0)
    {
        if (submit_sca_overlap_write(param_ptr->threadpool,
                param_ptr->sca_overlap_file, sca_overlap_records,
                sca_overlap_record_count) != SUCCESS)
        {
            IAS_LOG_ERROR("Submitting SCA overlap records for write");
            param_ptr->thread_error = thread_num + 1;
            return ERROR;
        }
    }

    return SUCCESS;
}

/******************************************************************************
NAME: cc_resample

PURPOSE:
    cc_resample performs cubic convolution resampling of the input
    image data while applying the detector delays and offsets.

RETURN VALUE:
    SUCCESS if the output tile is successfully created
    ERROR if the output tile could not be created

******************************************************************************/

int cc_resample
(
    const IAS_GRID_BAND_TYPE *grid_band_ptr, /* I: pointer to grid band */
    const IAS_LOS_MODEL *model_ptr,/* I: model */
    int current_sca,               /* I: current SCA being resampled */
    int band_index,                /* I: current band index being resampled */
    const TILE_TYPE *output_tile,  /* I: current output tile location */
    IAS_IMAGE *dem,                /* I: DEM image data */
    const TILE_TYPE *input_tile,   /* I: current input tile location */
    const float *input_buf,        /* I: input buffer pointer */
    int input_image_nl,            /* I: number of lines in input image */
    int input_image_ns,            /* I: number of samples in input image */
    int max_line_offset,           /* I: maximum even/odd offset in the line
                                         direction for this SCA */
    double jitter_scale,           /* I: band scale factor for jitter table */ 
    const IAS_MISC_LINE_EXTENT *trim_lut,/* I: Lookup table for trimming off the
                                               extra SCA data */
    double *resample_ptr,          /* I/O: pointer to output buffer */
    float *l1r_lines,              /* I/O: pointer to L1R lines sampled for
                                           each pixel in the output tile for
                                           the SCA overlap char algorithm */
    float *l1r_samples             /* I/O: pointer to L1R lines sampled for
                                           each pixel in the output tile for
                                           the SCA overlap char algorithm */
)
{
    const double *kernel_weight_table_ptr = get_resample_weight_table_ptr();
                                  /* pointer to kernel weight table */
    double round = 1.0 / 128.0;   /* Rounding error in resampling incr. */
    int oline;                    /* output line */
    double fill = get_fill_pixel(); /* fill pixel */

    /* resampling kernel information */
    int kernel_samples;
    int kernel_lines;
    int kernel_subpixel_steps;
    int kernel_table_steps;

    double elev = 0.0;        /* Initialize elevation to zero */
    int status;               /* Return status */
    int terrain_flag;         /* Flag to indicate terrain processing */

    int start_line;       /* starting line for the current tile */
    int end_line;         /* ending line for the current tile */

    int ncols = grid_band_ptr->ngrid_samps - 1;
    int sensor_id;        /* id of the sensor the band belongs to */
    IAS_VECTOR *jitter_table; /* jitter table from the model */
    int jitter_max_index; /* maximum index in jitter table from the model */
    int band_number;      /* current band number */
    const IAS_BAND_ATTRIBUTES *band_attributes = NULL; /* attributes of band 
                             band_number */

    /* cache sensor offset and fill information from the model */
    const double *along_track_detector_offset = model_ptr->sensor.
            bands[band_index].scas[current_sca].detector_offsets_along_track;
    const double *across_track_detector_offset = model_ptr->sensor.
            bands[band_index].scas[current_sca].detector_offsets_across_track;
    const int *l0r_fill = model_ptr->sensor.bands[band_index].scas[current_sca].
            l0r_detector_offsets;
    int nom_fill = model_ptr->sensor.bands[band_index].scas[current_sca].
            nominal_fill;

    /* get the band number */
    band_number = ias_sat_attr_convert_band_index_to_number(band_index);
    if (band_number == ERROR)
    {
        IAS_LOG_ERROR("Retrieving band number for band index %d", band_index);
        return ERROR;
    }

    /* get the band attributes */
    band_attributes = ias_sat_attr_get_band_attributes(band_number);
    if (!band_attributes)
    {
        IAS_LOG_ERROR("Retrieving band attributes for band number %d",
                band_number);
        return ERROR;
    }

    /* get the sensor Id */
    sensor_id = band_attributes->sensor_id;

    /* get jitter information from the model */
    jitter_table = model_ptr->sensor.sensors[sensor_id].jitter_table;
    jitter_max_index = model_ptr->sensor.sensors[sensor_id].jitter_table_count 
        - 1;

    /* get the kernel information */
    get_kernel_table_info(&kernel_subpixel_steps, &kernel_table_steps,
            &kernel_samples, &kernel_lines);

    terrain_flag = do_terrain_correction();

    /* loop over the assigned lines */
    start_line = output_tile->start_line;
    end_line = output_tile->end_line;

    /* loop over the lines in the tile */
    for (oline = start_line; oline < end_line; oline++)
    {
        /* cache the index into the output tile buffer for the current line */
        int line_index = (oline - start_line) * output_tile->num_samples;
        int osamp;             /* output sample */

        for (osamp = output_tile->start_sample; 
             osamp < output_tile->end_sample; osamp++)
        {
            int row;                      /* row index in grid */
            int col;                      /* column index in grid */
            double d_iline;               /* double input line */
            double d_isamp;               /* double input sample */
            int zplane;                   /* current Z plane index */

            /* calculate the index to the output pixel in the buffer */
            int output_pixel_index = line_index + osamp 
                                   - output_tile->start_sample;

            /* See if output location is within trimmed boundary */
            if (osamp <= trim_lut[oline].start_sample 
                || osamp >= trim_lut[oline].end_sample)
            {
                /* The sample is trimmed, so set it to the fill value and
                   go to the next sample.  Note the continue is used to
                   prevent indenting a large block of code */
                resample_ptr[output_pixel_index] = fill;
                continue;
            }

            /* get elevation at the output pixel */
            if (terrain_flag)
            {
                if (ias_misc_read_elevation_at_line_sample(dem, (double)oline,
                            (double)osamp, grid_band_ptr->pixsize, 
                            grid_band_ptr->pixsize, &elev) != SUCCESS)
                {
                    IAS_LOG_ERROR("Error getting elevation");
                    return ERROR;
                }
                status = ias_grid_3d_ols2ils(grid_band_ptr, oline, osamp, 
                                             &d_iline, &d_isamp, elev);
                zplane = (int)floor(elev / grid_band_ptr->zspacing) 
                    + grid_band_ptr->zeroplane;
            }
            else
            {
                status = ias_grid_ols2ils(grid_band_ptr, oline, osamp, &d_iline,
                                    &d_isamp, grid_band_ptr->zeroplane);
                zplane = grid_band_ptr->zeroplane;
            }

            row = (int)(d_iline * grid_band_ptr->inv_cell_lines);
            col = (int)(d_isamp * grid_band_ptr->inv_cell_samps);

            /* if the ols2ils was successful and a legal row and column
               in the grid were calculated, continue creating this
               output pixel (note the row/column checks are needed to
               make sure we will not index into a non-existing grid cell) */
            if ((status == SUCCESS) 
                && (row >= 0) && (row < grid_band_ptr->ngrid_lines)
                && (col >= 0) && (col < grid_band_ptr->ngrid_samps)
                && d_isamp >= 6 && d_isamp < input_image_ns-8)
            {
                double adjusted_samp;
                int start_samp;
                double delta_line_oe; /* Detector odd/even offset/delay in 
                                         line direction */
                double delta_samp_oe; /* Detector odd/even offset/delay in 
                                         sample direction */
                int cell_index;       /* Array index into grid structure */
                int jitter_cell_index; /* Array index into grid structure
                                         for jitter sensitivity coefficients */
                int i_isamp;          /* Integer input sample */

                /* calculate the index into the delta odd/even members
                   of the grid - note the "* 2" is needed since there
                   are two coefficients for each cell */
                cell_index = (row * ncols + col) * 2;
                i_isamp = (int)d_isamp;

                /* calculate the jitter grid cell offset */
                jitter_cell_index = zplane * grid_band_ptr->ngrid_samps
                    * grid_band_ptr->ngrid_lines + row 
                    * grid_band_ptr->ngrid_samps + col;
                delta_line_oe = grid_band_ptr->delta_line_oe[cell_index]
                    + elev * grid_band_ptr->delta_line_oe[cell_index + 1];
                delta_samp_oe = grid_band_ptr->delta_samp_oe[cell_index]
                    + elev * grid_band_ptr->delta_samp_oe[cell_index + 1];

                /* get the input pixel location after adjusting for the
                   detector and elevation offsets and jitter effects */
                adjusted_samp = adjust_sample(along_track_detector_offset, 
                    across_track_detector_offset, cell_index, elev, i_isamp, 
                    jitter_scale, jitter_cell_index, grid_band_ptr, 
                    jitter_table, jitter_max_index, d_iline, delta_line_oe, 
                    nom_fill);

                if (d_isamp < adjusted_samp)
                {
                    /* if these adjustments lead to the actual input sample 
                       being greater than the original input sample location, 
                       move the sample range one more sample to the left */
                    adjusted_samp = adjust_sample(along_track_detector_offset, 
                        across_track_detector_offset, cell_index, elev, 
                        i_isamp - 1, jitter_scale, jitter_cell_index, 
                        grid_band_ptr, jitter_table, jitter_max_index, d_iline,
                        delta_line_oe, nom_fill);

                    /* the input sample is less than the adjusted value,
                       so to apply the akima algorithm, input samples
                       starting 3 before the current one are needed */
                    if (d_isamp > adjusted_samp)
                    {
                        start_samp = i_isamp - 3;
                    }
                    else
                    {
                        start_samp = i_isamp - 4;
                    }
                }
                else
                {
                    /* get the input pixel location for the next sample after
                       adjusting for the detector and elevation offsets and 
                       jitter effects */
                    adjusted_samp = adjust_sample(along_track_detector_offset, 
                        across_track_detector_offset, cell_index, elev, 
                        i_isamp + 1, jitter_scale, jitter_cell_index, 
                        grid_band_ptr, jitter_table, jitter_max_index, d_iline, 
                        delta_line_oe, nom_fill);

                    /* if the input sample is greater than the adjusted
                       sample, start one pixel to the left, otherwise
                       start two pixels to the left */
                    if (d_isamp > adjusted_samp)
                        start_samp = i_isamp - 1;
                    else
                        start_samp = i_isamp - 2;
                }

                /* if the input line/sample are within the input image 
                   bounds (including the ragged top/bottom edges).
                   Note: the max_line_offset is set based on the worst case
                   detector offset and the +2 and +4 are based on the
                   dimensions of the resampling kernel. */
                if ( (d_iline >= (max_line_offset + 2)) &&
                     (d_iline <  (input_image_nl-(max_line_offset + 4))) &&
                     (start_samp >= 0) && (start_samp <  (input_image_ns - 6)))
                {
                    int itile_samps = input_tile->num_samples;
                    int samp_counter;   /* Counter for samples */
                    double total;       /* Resampled pixel output */
                    double ds;          /* fractional sample offset */
                    double resampled_det[6]; /* resampled detector values
                                        obtained by using CC on each detector
                                        for 4 lines */
                    double det_location[6];  /* Pixel locations for Akima */

                    /* make sure the input tile covers the samples needed.  
                       If not, there is likely a bug in the input image tile 
                       calculation */
                    if ((start_samp < input_tile->start_sample)
                        || (start_samp >= (input_tile->end_sample - 6)))
                    {
                        IAS_LOG_DEBUG("Sample start %d",start_samp);
                        IAS_LOG_DEBUG("Tile sample bounds %d, %d",
                            input_tile->start_sample, input_tile->end_sample);
                        if (start_samp < input_tile->start_sample)
                        {
                            IAS_LOG_DEBUG("Start sample error %d,%d",
                                start_samp, input_tile->start_sample);
                        }
                        else
                        {
                            IAS_LOG_DEBUG("End sample error %d,%d",
                                start_samp, input_tile->end_sample);
                        }
                        IAS_LOG_ERROR("Input image buffering problem");
                        return ERROR;
                    }

                    /* Save the L1R line and sample location to generate 
                       statistics for the SCA overlap characterization
                       algorithm, if requested */
                    if (l1r_lines)
                    {
                        l1r_lines[output_pixel_index] = d_iline;
                        l1r_samples[output_pixel_index] = d_isamp;
                    }

                    /* find the delta value in the line direction.  The grid
                       holds two coefficients that need to be applied. */
                    delta_line_oe = grid_band_ptr->delta_line_oe[cell_index]
                        + elev * grid_band_ptr->delta_line_oe[cell_index + 1];

                    /* Perform resampling */

                    /* Apply 1-d CCW in line direction, extra lines are
                       resampled to account for a sample shift of greater 
                       than +/-1. */
                    for (samp_counter = 0; samp_counter < 6; samp_counter++)
                    {
                        int current_iline; /* line to apply cubic 1-d */
                        int current_isamp; /* sample needed for Akima */
                        const double *w;   /* Resampling weight table */
                        int  ll;    /* Resampling weight table locations */
                        int pixel_index;
                        double dl;         /* fractional line offset */
                        double scaled_line;
                        double scaled_samp;
                        double line_jitter_adjustment; /* jitter adjustment */
                        double sample_jitter_adjustment; /* jitter adjustment */
                        double jitter_line; /* adjusted line position to 
                                              perform jitter look-up */
                        int jitter_index;  /* jitter table index */
                        double jitter_line_delta; /* fractional jitter sample 
                                              offset */
                        double scale;

                        current_isamp = start_samp + samp_counter;

                        /* Sum detector (static) offset and Odd/Even offset
                           (dynamic) */

                        /* Scale odd/even sample adjustment */
                        scale = floor(
                            along_track_detector_offset[current_isamp]+0.5);
                        scaled_samp = delta_samp_oe * scale;

                        /* Calculate total adjustment according to the pixel
                           location on focal plane and in L0r/L1r file. */
                        scaled_line = (delta_line_oe * scale) 
                                + l0r_fill[current_isamp] - nom_fill
                                - along_track_detector_offset[current_isamp];

                        /* Compute the jitter effect for this detector */
                        jitter_line = d_iline + scaled_line;
                        jitter_index = (int)(jitter_scale * (jitter_line 
                            - l0r_fill[current_isamp]));

                        /* if the jitter index isn't within the range of the 
                           jitter table, set it to the appropriate minimum or 
                           maximum value */
                        if (jitter_index < 1)
                            jitter_index = 1;
                        if (jitter_index > jitter_max_index - 1)
                            jitter_index = jitter_max_index - 1;

                        /* calculate the fractional jitter table index */
                        jitter_line_delta = jitter_scale * jitter_line 
                            - floor(jitter_scale * jitter_line);

                        calculate_jitter(jitter_line_delta, jitter_index, 
                            jitter_cell_index, grid_band_ptr, jitter_table,
                            &line_jitter_adjustment, &sample_jitter_adjustment);

                        scaled_samp += sample_jitter_adjustment; 
                        scaled_line += line_jitter_adjustment;

                        /* Set current line to input line */
                        current_iline = (int)(d_iline + scaled_line);

                        /* if the current line doesn't fall in the input
                           tile calculated, there is a problem */
                        if (((current_iline - 1) < input_tile->start_line)
                           || ((current_iline + 2) >= input_tile->end_line))
                        {
                            IAS_LOG_DEBUG("Current line %d, tile %d,%d "
                                "delta %f (%f)", current_iline,
                                input_tile->start_line,
                                input_tile->end_line, scaled_line,
                                d_iline);
                            if (current_iline - 1 < input_tile->start_line)
                            {
                                IAS_LOG_DEBUG("Start line error %d,%d",
                                current_iline - 1, input_tile->start_line);
                            }
                            else
                            {
                                IAS_LOG_DEBUG("End line error %d,%d",
                                    current_iline + 2,
                                    input_tile->end_line);
                            }
                            IAS_LOG_ERROR("Input image buffering problem");
                            return ERROR;
                        }

                        /* Determine new sub-pixel adjustment */
                        dl = d_iline + scaled_line - (double)current_iline;

                        /* Find 1D resampling kernel */
                        ll = (int)(kernel_lines 
                           * (int)(kernel_subpixel_steps * (round + dl)));
                        w = &(kernel_weight_table_ptr[ll]);

                        /* Resample only in the line direction using
                           cubic weights */
                        pixel_index = (current_iline - input_tile->start_line) 
                            * itile_samps 
                            + (current_isamp - input_tile->start_sample);

                        resampled_det[samp_counter]
                              = w[0] * input_buf[pixel_index - itile_samps]
                              + w[1] * input_buf[pixel_index] 
                              + w[2] * input_buf[pixel_index + itile_samps]
                              + w[3] * input_buf[pixel_index + 2 * itile_samps];

                        /* Store corresponding sample location.  This 
                           is needed for Akima interpolation */
                        det_location[samp_counter] = current_isamp 
                            - across_track_detector_offset[current_isamp] 
                            - scaled_samp;

                    } /* Sample counter */

                    /* perform the Akima interpolation */
                    ds = d_isamp - det_location[2];
                    total = akima(&resampled_det[0],&det_location[0],ds);

                    /* put new resampled value in intermediate buffer, 
                       limiting it to the output range */
                    if (total < min_range)
                        resample_ptr[output_pixel_index] = min_range;
                    else if (total > max_range)
                        resample_ptr[output_pixel_index] = max_range;
                    else
                        resample_ptr[output_pixel_index] = total;

                } /* Check to see if within an SCA for samples */
                else
                {
                    /* line or sample is out of range: pixel is outside of 
                       input image so set the output pixel to fill */
                    resample_ptr[output_pixel_index] = fill;
                }
            }
            else
            {
                if (status != ERROR)
                {
                    /* grid cell is out of range: pixel is outside of input 
                       image so set the output pixel to fill */
                    resample_ptr[output_pixel_index] = fill;
                }
                else
                {
                    IAS_LOG_ERROR("Mapping output to input space for band "
                        "index %d SCA index %d output line %d output sample "
                        "%d elevation %f", band_index, current_sca, oline, 
                        osamp, elev);
                    return ERROR;
                }
            }
        } /* end of samp loop */
    } /* end of line loop */

    return SUCCESS;
}

/******************************************************************************
NAME: nn_resample

PURPOSE:
    nn_resample performs nearest neighbor resampling of the input image data 
    without applying the detector delays and offsets.

RETURN VALUE:
    SUCCESS if the output tile is successfully created
    ERROR if the output tile could not be created

******************************************************************************/
int nn_resample
(
    const IAS_GRID_BAND_TYPE *grid_band_ptr, /* I: pointer to grid band */
    const IAS_LOS_MODEL *model_ptr,/* I: model */
    int current_sca,               /* I: current SCA being resampled */
    int band_index,                /* I: index into band array */
    const TILE_TYPE *output_tile,  /* I: current output tile location */
    IAS_IMAGE *dem,                /* I: DEM image data */
    const TILE_TYPE *input_tile,   /* I: current input tile location */
    const float *input_buf,        /* I: input buffer pointer */
    int input_image_nl,            /* I: number of lines in input image */
    int input_image_ns,            /* I: number of samples in input image */
    int max_line_offset,           /* I: maximum even/odd offset in the line
                                         direction for this SCA */
    double jitter_scale,           /* I: band scale factor for jitter table */ 
    const IAS_MISC_LINE_EXTENT *trim_lut,/* I: Lookup table for trimming off the
                                         extra SCA data */
    double *resample_ptr,          /* I/O: pointer to output buffer */
    float *l1r_lines,              /* I/O: pointer to L1R lines sampled for
                                           each pixel in the output tile for
                                           the SCA overlap char algorithm */
    float *l1r_samples             /* I/O: pointer to L1R lines sampled for
                                           each pixel in the output tile for
                                           the SCA overlap char algorithm */
)
{
    double d_iline, d_isamp;      /* Input line/sample */
    int oline, osamp;             /* output line/sample */
    double fill = get_fill_pixel(); /* fill pixel */

    int status;           /* function return status */

    double elev = 0.0;    /* Initialize elevation to zero */
    int terrain_flag;     /* flag to indicate terrain processing */

    int start_line;       /* starting line for the current tile */
    int end_line;         /* ending line for the current tile */

    IAS_VECTOR *jitter_table; /* jitter table from the model */
    int jitter_max_index; /* maximum index in jitter table from the model */
    int ncols = grid_band_ptr->ngrid_samps - 1; /* number of columns */
    int sensor_id;        /* id of the sensor the band belongs to */
    int band_number;      /* current band number */
    const IAS_BAND_ATTRIBUTES *band_attributes = NULL; /* attributes of band 
                             band_number */

    /* set up the parameters for delay */
    const double *along_track_detector_offset = model_ptr->sensor.
            bands[band_index].scas[current_sca].detector_offsets_along_track;
    const double *across_track_detector_offset = model_ptr->sensor.
            bands[band_index].scas[current_sca].detector_offsets_across_track;
    const int *l0r_fill = model_ptr->sensor.bands[band_index].scas[current_sca].
            l0r_detector_offsets;
    int nom_fill = model_ptr->sensor.bands[band_index].scas[current_sca].
            nominal_fill;

    terrain_flag = do_terrain_correction();

    /* get the band number */
    band_number = ias_sat_attr_convert_band_index_to_number(band_index);
    if (band_number == ERROR)
    {
        IAS_LOG_ERROR("Retrieving band number for band index %d", band_index);
        return ERROR;
    }

    /* get the band attributes */
    band_attributes = ias_sat_attr_get_band_attributes(band_number);
    if (!band_attributes)
    {
        IAS_LOG_ERROR("Retrieving band attributes for band number %d",
                band_number);
        return ERROR;
    }

    /* get the sensor Id and spectral type */
    sensor_id = band_attributes->sensor_id;

    /* initialize jitter model parameters */
    jitter_table = model_ptr->sensor.sensors[sensor_id].jitter_table;
    jitter_max_index = model_ptr->sensor.sensors[sensor_id].jitter_table_count 
        - 1;

    /* loop over the assigned lines */
    start_line = output_tile->start_line;
    end_line = output_tile->end_line;

    for (oline = start_line; oline < end_line; oline++)
    {
        /* cache the index into the output tile buffer for the current line */
        int line_index = (oline - start_line) * output_tile->num_samples;

        for (osamp = output_tile->start_sample; 
             osamp < output_tile->end_sample; osamp++)
        {
            int row;               /* row index in grid */
            int col;               /* column index in grid */
            int zplane;            /* current Z plane index */
            double dl;             /* fractional line offset */

            /* calculate the index to the output pixel in the buffer */
            int output_pixel_index = line_index + osamp 
                                   - output_tile->start_sample;

            /* See if output location is within trimmed boundary */
            if (osamp <= trim_lut[oline].start_sample 
                || osamp >= trim_lut[oline].end_sample)
            {
                /* The sample is trimmed, so set it to the fill value and
                   go to the next sample.  Note the continue is used to
                   prevent indenting a large block of code */
                resample_ptr[output_pixel_index] = fill;
                continue;
            }

            /* get elevation for image */
            if (terrain_flag)
            {
                if (ias_misc_read_elevation_at_line_sample(dem, (double)oline,
                            (double)osamp, grid_band_ptr->pixsize,
                            grid_band_ptr->pixsize, &elev) != SUCCESS)
                {
                    IAS_LOG_ERROR("Getting elevation");
                    return ERROR;
                }
                status = ias_grid_3d_ols2ils(grid_band_ptr, oline, osamp, 
                                             &d_iline, &d_isamp, elev);
                zplane = (int)floor(elev / grid_band_ptr->zspacing) 
                    + grid_band_ptr->zeroplane;
            }
            else
            {
                elev = 0.0;
                status = ias_grid_ols2ils(grid_band_ptr, oline, osamp, &d_iline,
                                         &d_isamp, grid_band_ptr->zeroplane);
                zplane = grid_band_ptr->zeroplane;
            }

            row = (int)(d_iline * grid_band_ptr->inv_cell_lines);
            col = (int)(d_isamp * grid_band_ptr->inv_cell_samps);

            /* if the ols2ils was successful and a legal row and column
               in the grid were calculated, continue creating this
               output pixel (note the row/column checks are needed to
               make sure we will not index into a non-existing grid cell) */
            if ((status == SUCCESS) 
                && (row >= 0) && (row < grid_band_ptr->ngrid_lines)
                && (col >= 0) && (col < grid_band_ptr->ngrid_samps)
                && d_isamp >= 1 && d_isamp < input_image_ns-3)
            {
                double adjusted_samp; /* Sample adjusted for detector and 
                                         elevation offsets and jitter */ 
                int start_samp;       /* Start sample */
                double delta_line_oe; /* Detector odd/even offset/delay in 
                                         line direction */
                double delta_samp_oe; /* Detector odd/even offset/delay in 
                                         sample direction */
                int cell_index;       /* Array index into grid structure */
                int jitter_cell_index; /* Array index into grid structure
                                         for jitter sensitivity coefficients */
                int i_isamp;          /* Integer input sample */
                double total;         /* resampled pixel output */

                /* calculate the index into the delta odd/even members
                   of the grid - note the "* 2" is needed since there
                   are two coefficients for each cell */
                cell_index = (row * ncols + col) * 2;
                i_isamp = (int)d_isamp;

                /* calculate jitter grid cell offset */
                jitter_cell_index = zplane * grid_band_ptr->ngrid_samps 
                    * grid_band_ptr->ngrid_lines + row 
                    * grid_band_ptr->ngrid_samps + col;

                delta_line_oe = grid_band_ptr->delta_line_oe[cell_index]
                    + elev * grid_band_ptr->delta_line_oe[cell_index + 1];
                delta_samp_oe = grid_band_ptr->delta_samp_oe[cell_index]
                    + elev * grid_band_ptr->delta_samp_oe[cell_index + 1];

                /* get the input pixel location after adjusting for the
                   detector and elevation offsets and jitter effects */
                adjusted_samp = adjust_sample(along_track_detector_offset, 
                    across_track_detector_offset, cell_index, elev, i_isamp, 
                    jitter_scale, jitter_cell_index, grid_band_ptr, 
                    jitter_table, jitter_max_index, d_iline, delta_line_oe, 
                    nom_fill);

                if (d_isamp < adjusted_samp)
                {
                    /* if these adjustments lead to the actual input sample 
                       being greater than the original input sample location, 
                       move the sample range one more sample to the left */
                    adjusted_samp = adjust_sample(along_track_detector_offset, 
                        across_track_detector_offset, cell_index, elev, 
                        i_isamp - 1, jitter_scale, jitter_cell_index, 
                        grid_band_ptr, jitter_table, jitter_max_index, d_iline,
                        delta_line_oe, nom_fill);

                    /* the input sample is less than the adjusted value,
                       so to apply the akima algorithm, input samples
                       starting 1 before the current one are needed */
                    if (d_isamp > adjusted_samp)
                    {
                        start_samp = i_isamp - 1;
                    }
                    else
                    {
                        start_samp = i_isamp - 2;
                    }
                }
                else
                {
                    /* get the input pixel location for the next sample after
                       adjusting for the detector and elevation offsets and 
                       jitter effects */
                    adjusted_samp = adjust_sample(along_track_detector_offset, 
                        across_track_detector_offset, cell_index, elev, 
                        i_isamp + 1, jitter_scale, jitter_cell_index, 
                        grid_band_ptr, jitter_table, jitter_max_index, d_iline,
                        delta_line_oe, nom_fill);

                    /* if the input sample is greater than the adjusted
                       sample, start one pixel to the right, otherwise
                       start at the current sample */
                    if (d_isamp > adjusted_samp)
                        start_samp = i_isamp + 1;
                    else
                        start_samp = i_isamp;
                }

                /* if the input line/sample are within the input image 
                   bounds (including the ragged top/bottom edges) */
                if ((d_iline >= (max_line_offset + 1)) 
                    && (d_iline < (input_image_nl - (max_line_offset + 3))) 
                    && (start_samp >= 0) && (start_samp < (input_image_ns - 2)))
                {
                    int itile_samps = input_tile->num_samples;
                    int samp_counter;   /* counter for samples */
                    double det_location[2]; /* pixel location values */
                    double resampled_det[2]; /* resampled detector values */

                    /* make sure the input tile covers the samples needed.  
                       If not, there is likely a bug in the input image tile 
                       calculation */
                    if ((start_samp < input_tile->start_sample)
                        || (start_samp >= (input_tile->end_sample - 2)))
                    {
                        IAS_LOG_DEBUG("Sample start %d", start_samp);
                        IAS_LOG_DEBUG("Tile sample bounds %d, %d",
                            input_tile->start_sample, input_tile->end_sample);
                        if (start_samp < input_tile->start_sample)
                        {
                            IAS_LOG_DEBUG("Start sample error %d,%d",
                                start_samp, input_tile->start_sample);
                        }
                        else
                        {
                            IAS_LOG_DEBUG("End sample error %d,%d",
                                start_samp, input_tile->end_sample);
                        }
                        IAS_LOG_ERROR("Input image buffering problem");
                        return ERROR;
                    }

                    /* Save the L1R line and sample location to generate 
                       statistics for the SCA overlap characterization
                       algorithm, if requested */
                    if (l1r_lines)
                    {
                        l1r_lines[output_pixel_index] = d_iline;
                        l1r_samples[output_pixel_index] = d_isamp;
                    }

                    /* make sure the input tile covers the samples needed.
                       If not, there is likely a bug in the input image tile
                       calculation */
                    if (start_samp > 0 && start_samp < input_image_ns - 2)
                    {
                       /* find the delta value in the line and sample 
                          directions.  Note the even and odd detectors are
                          calculated from different grid entries.  The grid 
                          holds two coefficients that need to be applied. */
                       delta_line_oe = grid_band_ptr->delta_line_oe[cell_index]
                           + elev 
                           * grid_band_ptr->delta_line_oe[cell_index + 1];
                       delta_samp_oe = grid_band_ptr->delta_samp_oe[cell_index]
                           + elev 
                           * grid_band_ptr->delta_samp_oe[cell_index + 1];

                       /* perform resampling */
                       for (samp_counter = 0; samp_counter < 2; samp_counter++)
                       {
                           int current_iline;  /* line to apply cubic 1-d */
                           int current_isamp;  /* sample needed for Akima */
                           int pixel_index;    /* index into input buffer */
                           double scaled_line; /* line offset with scale 
                                                  applied */
                           double scaled_samp; /* sample offset with scale 
                                                  applied */
                           double scale;       /* scale to apply to lines and
                                                  samples */
                           double jitter_line_delta; /* fractional jitter 
                                                  sample offset */
                           int jitter_index;   /* jitter table index */
                           double line_jitter_adjustment; /* jitter line 
                                                  adjustment */
                           double sample_jitter_adjustment; /* jitter sample 
                                                  adjustment */
                           double jitter_line; /* adjusted line position to 
                                                  perform jitter look-up */

                           current_isamp = start_samp + samp_counter;

                           /* scale odd/even sample adjustment */
                           scale = floor(along_track_detector_offset
                               [current_isamp] + 0.5);
                           scaled_samp = delta_samp_oe * scale;

                           /* calculate total adjustment according to the pixel
                              location on focal plane and in L0r/L1r file. */
                           scaled_line = (delta_line_oe * scale) 
                               + l0r_fill[current_isamp] - nom_fill
                               - along_track_detector_offset[current_isamp];

                           /* compute the jitter effect for this detector */
                           jitter_line = d_iline + scaled_line;
                           jitter_index = (int)(jitter_scale * (jitter_line 
                               - l0r_fill[current_isamp]));

                            /* calculate the fractional jitter table index */
                            jitter_line_delta = jitter_scale * jitter_line 
                                - floor(jitter_scale * jitter_line);

                            /* if the jitter index isn't within the range of 
                               the jitter table, set it to the appropriate 
                               minimum or maximum value */
                            if (jitter_index < 1)
                                jitter_index = 1;
                            if (jitter_index > jitter_max_index - 1)
                                jitter_index = jitter_max_index - 1;

                            calculate_jitter(jitter_line_delta, jitter_index, 
                                jitter_cell_index, grid_band_ptr, jitter_table,
                                &line_jitter_adjustment, 
                                &sample_jitter_adjustment);

                            scaled_samp += sample_jitter_adjustment; 
                            scaled_line += line_jitter_adjustment;

                            /* sum detector (static) offset and odd/even offset
                               (dynamic) */

                            /* set current line to input line */
                            current_iline = (int)(d_iline + scaled_line);

                            /* if the current line doesn't fall in the input
                               tile calculated, there is a problem */
                            if ((current_iline < input_tile->start_line)
                               || ((current_iline + 1) >= input_tile->end_line))
                            {
                                IAS_LOG_DEBUG("Current line %d, tile %d,%d "
                                    "delta %f (%f)", current_iline,
                                    input_tile->start_line,
                                    input_tile->end_line, scaled_line,
                                    d_iline);
                                if (current_iline < input_tile->start_line)
                                {
                                    IAS_LOG_DEBUG("Start line error %d,%d",
                                    current_iline, input_tile->start_line);
                                }
                                else
                                {
                                    IAS_LOG_DEBUG("End line error %d,%d",
                                        current_iline + 1,
                                        input_tile->end_line);
                                }
                                IAS_LOG_ERROR("Input image buffering problem");
                                return ERROR;
                            }

                            /* determine new sub-pixel adjustment */
                            dl = d_iline + scaled_line - (double)current_iline;

                            /* Resample in the line direction */
                            pixel_index = (current_iline 
                                - input_tile->start_line) * itile_samps 
                                + (current_isamp - input_tile->start_sample);

                            if (dl < 0.5)
                            {
                                resampled_det[samp_counter]
                                    = input_buf[pixel_index];
                            }
                            else
                            {
                                resampled_det[samp_counter]
                                    = input_buf[pixel_index + itile_samps];
                            }

                            /* Store corresponding sample location.  This 
                               is needed for interpolation */
                            det_location[samp_counter] = current_isamp 
                                - across_track_detector_offset[current_isamp] 
                                - scaled_samp;
                        }

                        /* perform the sample interpolation */
                        if (fabs(d_isamp - det_location[0])
                            < fabs(d_isamp - det_location[1]))
                        {
                            total = resampled_det[0];
                        }
                        else
                        {
                            total = resampled_det[1];
                        }

                        /* put new resampled value in intermediate buffer,
                           limiting it to the output range */
                        if (total < min_range)
                            resample_ptr[output_pixel_index] = min_range;
                        else if (total > max_range)
                            resample_ptr[output_pixel_index] = max_range;
                        else
                        {
                            resample_ptr[output_pixel_index] = total;
                        }

                    }
                    else
                    {
                        /* pixel is outside of input image so set the output 
                           pixel to fill */
                        resample_ptr[output_pixel_index] = fill;
                    }
                }
                else
                {
                    /* pixel is outside of input image so set the output pixel
                       to fill */
                    resample_ptr[output_pixel_index] = fill;
                }
            }
            else
            {
                if (status != ERROR)
                {
                    /* pixel is outside of input image so set the output pixel 
                       to fill */
                    resample_ptr[output_pixel_index] = fill;
                }
                else
                {
                    IAS_LOG_ERROR("Mapping output to input space for band "
                        "index %d SCA index %d output line %d output sample "
                        "%d elevation %f", band_index, current_sca, oline, 
                        osamp, elev);
                    return ERROR;
                }
            }
        } /* end of sample loop */
    } /* end of line loop */

    return SUCCESS;
}

/******************************************************************************
NAME: get_min_range_value

PURPOSE: To provide access to the min_range value used by the resampling.

RETURN VALUE: The min range value.
******************************************************************************/
double get_min_range_value()
{
    return min_range;
}

/******************************************************************************
NAME: get_max_range_value

PURPOSE: To provide access to the max_range value used by the resampling.

RETURN VALUE: The max range value.
******************************************************************************/
double get_max_range_value()
{
    return max_range;
}

