#ifndef RESAMPLE_IMAGE_H
#define RESAMPLE_IMAGE_H

#include "ias_los_model.h"
#include "ias_threadpool.h"
#include "get_input_image.h"
#include "buffer_output_image.h"
#include "ias_miscellaneous.h"
#include "sca_overlap_io.h"

/* define the type for the resampling functions */
typedef int (*RESAMPLE_FUNCTION)
(
    const IAS_GRID_BAND_TYPE *grid_band_ptr, /* I: pointer to grid band */
    const IAS_LOS_MODEL *model_ptr,/* I: model */
    int current_sca,               /* I: current sca being resampled */
    int band_index,                /* I: index into band array */
    const TILE_TYPE *output_tile,  /* I: current output tile location */
    IAS_IMAGE *dem,                /* I: DEM image data */
    const TILE_TYPE *input_tile,   /* I: current input tile location */
    const float *input_buf,        /* I: input buffer pointer */
    int input_image_nl,            /* I: number of lines in input image */
    int input_image_ns,            /* I: number of samples in input image */
    int max_line_offset,           /* I: maximum even/odd offset in the line
                                         direction for this SCA */
    double jitter_scale,           /* I: band scale factor for jitter table */
    const IAS_MISC_LINE_EXTENT *trim_lut,/* I: Lookup table for trimming off the
                                         extra SCA data */
    double *resample_ptr,          /* I/O: pointer to output buffer */
    float *l1r_lines,              /* I/O: pointer to L1R lines sampled for
                                           each pixel in the output tile for
                                           the SCA overlap char algorithm */
    float *l1r_samples             /* I/O: pointer to L1R lines sampled for
                                           each pixel in the output tile for
                                           the SCA overlap char algorithm */
);

typedef struct
{
    struct ias_threadpool *threadpool;
    int thread_error;         
    const IAS_GRID_TYPE *grid_ptr;
    const IAS_LOS_MODEL *model_ptr;
    OUTPUT_BUFFER *output_image_ptr; 
    int sca;
    IAS_IMAGE *dem_image;
    INPUT_IMAGE *input_image;
    IAS_THREAD_MUTEX_TYPE hdf_io_mutex;
    int max_line_offsets[IAS_MAX_NSCAS]; /* maximum even/odd offset in the line
                                            direction for each SCA */
    int combine_scas;
    int band_index;
    RESAMPLE_FUNCTION resample_function;
    const IAS_MISC_LINE_EXTENT *trim_lut;
    SCA_OVERLAP_IO *sca_overlap_file;
} RESAMPLE_THREAD_PARAMS;

int setup_resamp_output
(
    int apply_radiance_rescale_range,/* I: Determine if the radiance rescale
                                           range should be applied. */
    int band_number                  /* I: band number to set up for */
);

int do_ols2ils
(
    const IAS_GRID_BAND_TYPE *grid_band_ptr, /* I: current grid */
    int terrain_flag,       /* I: terrain correction flag */
    IAS_IMAGE *dem,         /* I: DEM data if doing terrain correction */
    int oline,              /* I: output space line */
    int osamp,              /* I: output space sample */
    double *iline,          /* O: input space line */
    double *isamp           /* O: input space sample */
);

void wrapped_convert_data
(
    const double *input_ptr,/* I: data to convert to the output type */
    void *outbuf,           /* O: output data buffer */
    int nl,                 /* I: Number of lines in ouput */
    int ns                  /* I: Number of samples in output */
);

int cc_resample
(
    const IAS_GRID_BAND_TYPE *grid_band_ptr, /* I: pointer to grid band */
    const IAS_LOS_MODEL *model_ptr,/* I: model */
    int current_sca,               /* I: current sca being resampled */
    int band_index,                /* I: current band index being resampled */
    const TILE_TYPE *output_tile,  /* I: current output tile location */
    IAS_IMAGE *dem,                /* I: DEM image data */
    const TILE_TYPE *input_tile,   /* I: current input tile location */
    const float *input_buf,        /* I: input buffer pointer */
    int input_image_nl,            /* I: number of lines in input image */
    int input_image_ns,            /* I: number of samples in input image */
    int max_line_offset,           /* I: maximum even/odd offset in the line
                                         direction for this SCA */
    double jitter_scale,           /* I: band scale factor for jitter table */
    const IAS_MISC_LINE_EXTENT *trim_lut,/* I: Lookup table for trimming off the
                                         extra SCA data */
    double *resample_ptr,          /* I/O: pointer to output buffer */
    float *l1r_lines,              /* I/O: pointer to L1R lines sampled for
                                           each pixel in the output tile for
                                           the SCA overlap char algorithm */
    float *l1r_samples             /* I/O: pointer to L1R lines sampled for
                                           each pixel in the output tile for
                                           the SCA overlap char algorithm */
);

int nn_resample
(
    const IAS_GRID_BAND_TYPE *grid_band_ptr, /* I: pointer to grid band */
    const IAS_LOS_MODEL *model_ptr,/* I: model */
    int current_sca,               /* I: current sca being resampled */
    int band_index,                /* I: index into band array */
    const TILE_TYPE *output_tile,  /* I: current output tile location */
    IAS_IMAGE *dem,                /* I: DEM image data */
    const TILE_TYPE *input_tile,   /* I: current input tile location */
    const float *input_buf,        /* I: input buffer pointer */
    int input_image_nl,            /* I: number of lines in input image */
    int input_image_ns,            /* I: number of samples in input image */
    int max_line_offset,           /* I: maximum even/odd offset in the line
                                         direction for this SCA */
    double jitter_scale,           /* I: band scale factor for jitter table */
    const IAS_MISC_LINE_EXTENT *trim_lut,/* I: Lookup table for trimming off the
                                         extra SCA data */
    double *resample_ptr,          /* I/O: pointer to output buffer */
    float *l1r_lines,              /* I/O: pointer to L1R lines sampled for
                                           each pixel in the output tile for
                                           the SCA overlap char algorithm */
    float *l1r_samples             /* I/O: pointer to L1R lines sampled for
                                           each pixel in the output tile for
                                           the SCA overlap char algorithm */
);

int resample_image
(
    void *params,   /* I: pointer to input parameter structure */
    int thread_num  /* I: thread number for the thread running this routine */
);

void calculate_max_line_offsets
(
    const IAS_LOS_MODEL *model_ptr,     /* I: model */
    int band_index,                     /* I: band index */
    int max_line_offsets[IAS_MAX_NSCAS] /* O: maximum even/odd offsets in the
                                              line direction for each SCA */
);

/* Access to the min and max values used by the resampling.  They are needed
   by the L1G metadata. */
double get_min_range_value();
double get_max_range_value();

#endif
