/****************************************************************************
Name: sca_overlap_io

Purpose: Provides the routines to write the SCA overlap characterization data
    to a binary file.

File format description (binary file):
    File type tag : int
    Record type (i.e. full/limited): int
    For each band:
        Band start tag : int
        Band number: int
        Record count: int
        SCA_OVERLAP_RECORDS (record count of them)
            - The records can be the full or limited records
    File end tag : int



****************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include "ias_const.h"
#include "ias_logging.h"
#include "sca_overlap_io.h"

/* These are random tags that are used to mark different sections of the
   file */
#define FILE_TYPE_TAG 0x216677FA
#define BAND_START_TAG 0x5A5AA5A5
#define FILE_END_TAG 0x2352CE03

struct sca_overlap_io
{
    FILE *file;             /* Open file pointer */
    int band_number;        /* Current band number */
    long band_file_offset;  /* File offset to the current band header */
    int record_count;       /* Count of records written to the current band */
    SCA_OVERLAP_CHAR_TYPE overlap_type;
                            /* The type of overlap data being written to the
                               file (limited/full) */
};

/****************************************************************************
Name: open_sca_overlap_file

Purpose: Opens an SCA overlap characterization data file for writing.  The
    file type tag and type of SCA data indication are written as a file header.

Returns: Pointer to an IO structure, or NULL on error.

****************************************************************************/
SCA_OVERLAP_IO *open_sca_overlap_file
(
    const char *filename,   /* I: File name to open */
    SCA_OVERLAP_CHAR_TYPE overlap_type /* I: Specifies the type of overlap data
                                             to write to the file */
)
{
    SCA_OVERLAP_IO *io = NULL;
    int file_type = FILE_TYPE_TAG;

    /* Make sure the overlap type is valid */
    if (overlap_type == SCA_OVERLAP_CHAR_NONE)
    {
        IAS_LOG_ERROR("Attempted to open the SCA overlap characterization "
            "file when no characterization was requested");
        return NULL;
    }

    /* Make sure the overlap type is valid */
    if (overlap_type != SCA_OVERLAP_CHAR_LIMITED
        && overlap_type != SCA_OVERLAP_CHAR_FULL)
    {
        IAS_LOG_ERROR("Unrecognized SCA overlap characterization type %d",
            overlap_type);
        return NULL;
    }

    /* Allocate memory for the structure used to track the file info and
       initialize it */
    io = malloc(sizeof(*io));
    if (!io)
    {
        IAS_LOG_ERROR("Failed allocating memory for the SCA overlap IO "
            "structure");
        return NULL;
    }
    io->band_number = -1;
    io->band_file_offset = -1;
    io->record_count = 0;
    io->overlap_type = overlap_type;

    /* Open the file in binary write mode */
    io->file = fopen(filename, "wb");
    if (!io->file)
    {
        IAS_LOG_ERROR("Failed opening SCA overlap characterization file %s",
            filename);
        free(io);
        return NULL;
    }

    /* Write the file type tag to help identify the file */
    if (fwrite(&file_type, sizeof(file_type), 1, io->file) != 1)
    {
        IAS_LOG_ERROR("Failed writing file type to %s", filename);
        fclose(io->file);
        free(io);
        return NULL;
    }

    /* Write the type of SCA overlap data contained in the file */
    if (fwrite(&overlap_type, sizeof(overlap_type), 1, io->file) != 1)
    {
        IAS_LOG_ERROR("Failed writing overlap type to %s", filename);
        fclose(io->file);
        free(io);
        return NULL;
    }

    return io;
}

/****************************************************************************
Name: close_sca_overlap_file

Purpose: Writes the file end tag to the SCA overlap characterization file,
    closes it, and frees the associated structure memory.

Returns: SUCCESS/ERROR

****************************************************************************/
int close_sca_overlap_file
(
    SCA_OVERLAP_IO *io      /* I: File to close */
)
{
    int return_status = SUCCESS;
    int file_end_tag = FILE_END_TAG;

    /* Allow a NULL pointer to be provided and just ignore it to make
       error handling easier */
    if (!io)
    {
        return SUCCESS;
    }

    /* Write the end of file tag */
    if (fwrite(&file_end_tag, sizeof(file_end_tag), 1, io->file) != 1)
    {
        IAS_LOG_ERROR("Failed writing the file end tag");
        return_status = ERROR;
    }

    /* Close the file */
    if (fclose(io->file) != 0)
    {
        IAS_LOG_ERROR("Error closing SCA overlap characterization file");
        return_status = ERROR;
    }

    free(io);

    return return_status;
}

/****************************************************************************
Name: open_sca_overlap_band

Purpose: Prepares an open SCA overlap characterization file for writing
    information for a band by writing the band start tag, band number, and
    placeholder record count.  The starting file offset is saved to allow
    updating the actual record count later.

Returns: SUCCESS/ERROR

****************************************************************************/
int open_sca_overlap_band
(
    SCA_OVERLAP_IO *io,     /* I/O: Pointer to the previously opened file */
    int band_number         /* I: Band number to write */
)
{
    int start_tag = BAND_START_TAG;

    /* Initialize the info to track the band */
    io->band_number = band_number;
    io->record_count = 0;

    /* Save the current file position so it can be returned to later to 
       update the record count */
    io->band_file_offset = ftell(io->file);
    if (io->band_file_offset == -1)
    {
        IAS_LOG_ERROR("Failed getting current SCA overlap file position");
        return ERROR;
    }

    /* Write the start of band tag, band number, and placeholder for the 
       record count */
    if ((fwrite(&start_tag, sizeof(start_tag), 1, io->file) != 1)
        || (fwrite(&io->band_number, sizeof(io->band_number), 1, io->file) != 1)
        || (fwrite(&io->record_count, sizeof(io->record_count), 1, io->file)
            != 1))
    {
        IAS_LOG_ERROR("Failed writing start of band record to an SCA overlap "
            "file");
        return ERROR;
    }

    return SUCCESS;
}

/****************************************************************************
Name: close_sca_overlap_band

Purpose: Closes the current band in an SCA overlap characterization file.
    The band header information is re-written to update the record count.

Returns: SUCCESS/ERROR

****************************************************************************/
int close_sca_overlap_band
(
    SCA_OVERLAP_IO *io      /* I/O: Pointer to the previously opened file */
)
{
    int start_tag = BAND_START_TAG;

    /* Allow a NULL pointer to be provided and just ignore it to make
       error handling easier */
    if (!io)
    {
        return SUCCESS;
    }

    /* Make sure a band was really opened */
    if (io->band_number < 1 || io->band_file_offset < 0)
    {
        IAS_LOG_ERROR("Close band routine called without first calling the "
            "open band routine");
        return ERROR;
    }

    /* Return to the starting location for the band so the record count can
       be updated */
    if (fseek(io->file, io->band_file_offset, SEEK_SET) != 0)
    {
        IAS_LOG_ERROR("Failed seeking band start in SCA overlap file");
        return ERROR;
    }

    /* Write the start of band tag, band number, and placeholder for the 
       record count */
    if ((fwrite(&start_tag, sizeof(start_tag), 1, io->file) != 1)
        || (fwrite(&io->band_number, sizeof(io->band_number), 1, io->file) != 1)
        || (fwrite(&io->record_count, sizeof(io->record_count), 1, io->file)
            != 1))
    {
        IAS_LOG_ERROR("Failed writing start of band record to an SCA overlap "
            "file");
        return ERROR;
    }

    /* Go back to the end of the file */
    if (fseek(io->file, 0, SEEK_END) != 0)
    {
        IAS_LOG_ERROR("Failed seeking the end of the SCA overlap file");
        return ERROR;
    }

    /* Reset the band tracking info */
    io->band_number = -1;
    io->band_file_offset = -1;
    io->record_count = 0;

    return SUCCESS;
}

/****************************************************************************
Name: write_sca_overlap_records

Purpose: Writes a set of records to the file.  The set of recrods can be
    either a full or limited set of data.

Returns: SUCCESS/ERROR

****************************************************************************/
int write_sca_overlap_records
(
    SCA_OVERLAP_IO *io,         /* I: File to write */
    const SCA_OVERLAP_PTR records_ptr, /* I: Set of records to write (can point
                                             to full or limited records) */
    int record_count            /* I: Count of records to write */
)
{
    int record_size;

    /* Make sure a band was really opened */
    if (io->band_number < 1 || io->band_file_offset < 0)
    {
        IAS_LOG_ERROR("Write record routine called without first calling the "
            "open band routine");
        return ERROR;
    }

    /* Decide which record size to write */
    if (io->overlap_type == SCA_OVERLAP_CHAR_LIMITED)
        record_size = sizeof(LIMITED_SCA_OVERLAP_RECORD);
    else
        record_size = sizeof(SCA_OVERLAP_RECORD);

    /* Write the set of records to the file */
    if (fwrite(records_ptr.full_ptr, record_size, record_count, io->file)
        != record_count)
    {
        IAS_LOG_ERROR("Error writing %d records to SCA overlap file",
            record_count);
        return ERROR;
    }

    /* Keep track of the records written so the band header can be updated
       when the band is closed */
    io->record_count += record_count;

    return SUCCESS;
}
