#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "ias_satellite_attributes.h"
#include "ias_l1r.h"
#include "ias_types.h"
#include "ias_const.h"

int create_rad_rescale_test_l1r_hdf(char *output_name, int num_lines)
{
    const IAS_SATELLITE_ATTRIBUTES *sat_attr = NULL;

    int num_bands;
    int num_scas;
    int num_det;
    int band_index;
    int band_number;
    int sca_index;
    int status;
    int data_size;
    int index;

    IAS_DATA_TYPE data_type = IAS_R4;

    L1RIO *output_image = NULL;
    L1R_BAND_IO *out_band_io = NULL;
    IAS_L1R_FILE_METADATA file_metadata;
    IAS_L1R_BAND_METADATA band_metadata[1];

    float *output_buffer = NULL;

    /* Initialize the satellite attributes. */
    if (ias_sat_attr_initialize(IAS_L8) != SUCCESS)
    {
        printf("Error initializing the satellite attributes library");
        return ERROR;
    }

    /* Grab the attributes so things can be figured out. */
    sat_attr = ias_sat_attr_get_attributes();
    if (sat_attr == NULL)
    {
        printf("Error initializing the satellite attributes library");
        return ERROR;
    }

    /* Open the output image. */
    output_image = ias_l1r_open_image(output_name, IAS_WRITE);
    if (output_image == NULL)
    {
        printf("Cannot open %s for writing\n", output_name);
        return ERROR;
    }

#if DEBUG == 9
    printf("%d %d %d\n", sat_attr->sensors,
                         sat_attr->bands,
                         sat_attr->total_bands);
#endif

    num_bands = sat_attr->bands;

    for (band_index = 0; band_index < num_bands; band_index++)
    {
        band_number = sat_attr->band_attributes[band_index].band_number;

#if DEBUG == 9
        printf("Processing Band %d.........\n", band_number);
#endif

        num_scas = sat_attr->band_attributes[band_index].scas;
        num_det = sat_attr->band_attributes[band_index].detectors_per_sca;

        /* Open the current band to write to. */
        out_band_io = ias_l1r_open_band(output_image, band_number,
                                        &data_type, &num_scas, &num_lines,
                                        &num_det);

        if (out_band_io == NULL)
        {
            printf("Cannot set up band %d for writing\n", band_number);
            ias_l1r_close_image(output_image);
            return ERROR;
        }

        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
#if DEBUG == 9
            printf("  Processing SCA %d[%d][%d].........\n",
                   sca_index+1, num_det, num_lines);
#endif

            data_size = num_det*num_lines;

            /* Allocate the correct memory amount for the current sca. */
            output_buffer = calloc(data_size, sizeof(*output_buffer));

            /* Set the values for the data. */
            for (index = 0; index < data_size; index++)
            {
               output_buffer[index] = 42.0;
            }

            /* Write the data to the file. */
            status = ias_l1r_write_image(out_band_io, sca_index, 0, 0,
                                         num_lines, num_det, output_buffer);

            if (status)
            {
#if DEBUG == 9
                printf("Cannot write SCA image %d to output file\n",
                       (sca_index + 1));
#endif

                ias_l1r_close_band(out_band_io);
                ias_l1r_close_image(output_image);
                free(output_buffer);
                return ERROR;
            }

            /* Free the output buffer. */
            free(output_buffer);
        } /* End SCA Loop */

        /* Create some dummy band metadata. */
        band_metadata[0].band_number = band_number;
        sprintf(band_metadata[0].band_name, "B%02d", band_number);
        strcpy(band_metadata[0].instrument_source, "Test Data");
        strcpy(band_metadata[0].discontinuity_factor_source,
            "CPF");
        band_metadata[0].reflectance_to_radiance_coefficient = 9.9;

        ias_l1r_set_band_metadata(output_image, band_metadata, 1);

        /* Close the current band. */
        ias_l1r_close_band(out_band_io);
    } /* End Band Loop */

    /* Set the file metadata. */
    strcpy(file_metadata.software_version, "SW V-0.0");
    strcpy(file_metadata.ingest_software_version, "V-1.1");
    strcpy(file_metadata.collect_type, "EARTH");
    file_metadata.band_average_gain_applied = 0;
    file_metadata.relative_gain_applied = 1;
    file_metadata.sca_discontinuity_correction_applied = 1;
    file_metadata.inoperable_detector_fill_applied = 0;
    file_metadata.saturated_pixel_correction_applied = 1;
    file_metadata.residual_striping_correction_applied = 0;
    file_metadata.reflectance_conversion_applied = 1;
    strcpy(file_metadata.custom_rad_processing_steps_applied, "");
    strcpy(file_metadata.custom_rad_processing_steps_applied,
        "Test_Data_Steps");

    status = ias_l1r_set_file_metadata(output_image, &file_metadata);
    if (status)
    {
        printf("Error setting file metadata");
    }

    /* Close the output image. */
    ias_l1r_close_image(output_image);

    return SUCCESS;
}

