#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "cgreen/cgreen.h"

#include "ias_satellite_attributes.h"
#include "ias_threadpool.h"
#include "resample_image.h"
#include "get_cpf_parameters.h"
#include "get_resample_parameters.h"

#include "create_rad_rescale_test_l1r_hdf.h"

#define DATA_SIZE 289

#define DEBUG 0

/* shorten the type by creating a typedef for it */
typedef unsigned short int MY_OUT_TYPE;

/* declare some constant strings */
const char *calval_data_dir = "radiance_rescaling_calval_data";
const char *simulated_data_dir = "radiance_rescaling_data";
const char *cpf_name = "radiance_rescaling.cpf";
const char *mask_name = "radresc01_fillmask.data";
const char *oli_name = "radresc01_oli_input.data";
const char *oli_calval_result_name = "radresc01_oli_calval_results.data";
const char *tirs_calval_result_name = "radresc01_tirs_calval_results.data";

/* declare variables used by both tests */
static char *mask_data = NULL;
static double *oli_data = NULL;

/* process the OLI cal val test data */
Ensure test_oli_algorithm()
{
    char file_path[PATH_MAX];
    FILE *oli_calval_result_fd = NULL;
    MY_OUT_TYPE *oli_calval_result_data = NULL;
    double *oli_result_data = NULL;
    MY_OUT_TYPE *oli_short_data = NULL;
    int index = 0;

    /* allocate the memory needed */
    oli_calval_result_data = malloc(DATA_SIZE * sizeof(MY_OUT_TYPE));
    oli_result_data = malloc(DATA_SIZE * sizeof(double));
    oli_short_data = malloc(DATA_SIZE * sizeof(MY_OUT_TYPE));

    /* initialize the memory */
    memset(oli_calval_result_data, 0, DATA_SIZE * sizeof(MY_OUT_TYPE));
    memset(oli_result_data, 0, DATA_SIZE * sizeof(double));
    memset(oli_short_data, 0, DATA_SIZE * sizeof(MY_OUT_TYPE));

    /* load the OLI calval results */
    sprintf(file_path, "%s/%s", calval_data_dir, oli_calval_result_name);
    oli_calval_result_fd = fopen(file_path, "r");
    fread(oli_calval_result_data, DATA_SIZE, sizeof(MY_OUT_TYPE),
          oli_calval_result_fd);
    fclose(oli_calval_result_fd);

    /* read the cpf parameters */
    sprintf(file_path, "%s/%s", calval_data_dir, cpf_name);
    if (read_cpf_parameters(file_path, 1) != SUCCESS)
    {
        assert_true_with_message(0, "Error reading cpf file: %s", file_path);
        return;
    }
    /* patch the test data since it was generated using double math when
       it should have been done using floats.  calval plans to eventually
       fix it to use floats.  Until then, tweak this one value to allow the
       tests to pass */
    oli_data[106] -= 0.000001;

    /* process the OLI data */
    setup_radiance_parameters(0);
    setup_qcal_min_max(1);

    for (index = 0; index < DATA_SIZE; index++)
    {
#if DEBUG == 1 || DEBUG == 2
        printf("Mask = %d\n", mask_data[index]);
        printf("Index = %d %d\n", index/17, index%17);
#endif
        if (mask_data[index] != 0)
        {
            oli_result_data[index] =
                wrapped_apply_scaling_factors(oli_data[index]);
#if DEBUG == 2
            printf("OLI Source = %0.15lf\n", oli_data[index]);
            printf("OLI Result = %0.15lf\n", oli_result_data[index]);
#endif
        }
    }

    /* convert the data to unsigned short int(I*2) */
    wrapped_convert_data(oli_result_data, oli_short_data, 1, DATA_SIZE);

    /* test the OLI results */
    for (index = 0; index < DATA_SIZE; index++)
    {
#if DEBUG == 3
        if (oli_short_data[index] != oli_calval_result_data[index])
        {
            printf("OLI Source = %0.15lf\n", oli_data[index]);
            printf("OLI Result = %0.15lf\n", oli_result_data[index]);
            printf("Mask = %d\n", mask_data[index]);
            printf("OLI Short Result = %d\n", oli_short_data[index]);
            printf("OLI CalVal Result = %d\n", oli_calval_result_data[index]);
            printf("Index = %d (%d %d)\n", index, index/17, index%17);
        }
#endif
        assert_equal(oli_short_data[index], oli_calval_result_data[index]);
    }

    free(oli_calval_result_data);
    free(oli_result_data);
    free(oli_short_data);
}

/* process the TIRS cal val test data */
Ensure test_tirs_algorithm()
{
    char file_path[PATH_MAX];
    double *tirs_data = NULL;
    FILE *tirs_calval_result_fd = NULL;
    MY_OUT_TYPE *tirs_calval_result_data = NULL;
    double *tirs_result_data = NULL;
    MY_OUT_TYPE *tirs_short_data = NULL;
    int index = 0;

    /* allocate the memory needed */
    tirs_data = malloc(DATA_SIZE * sizeof(double));
    tirs_calval_result_data = malloc(DATA_SIZE * sizeof(MY_OUT_TYPE));
    tirs_result_data = malloc(DATA_SIZE * sizeof(double));
    tirs_short_data = malloc(DATA_SIZE * sizeof(MY_OUT_TYPE));

    /* initialize the memory */
    memset(tirs_data, 0, DATA_SIZE * sizeof(char));
    memset(tirs_calval_result_data, 0, DATA_SIZE * sizeof(MY_OUT_TYPE));
    memset(tirs_result_data, 0, DATA_SIZE * sizeof(double));
    memset(tirs_short_data, 0, DATA_SIZE * sizeof(MY_OUT_TYPE));

    /* generate the TIRS test data from the OLI test data
       the TIRS data is the OLI data multiplied by 10 */
    for (index = 0; index < DATA_SIZE; index++)
    {
        tirs_data[index] = oli_data[index] * 10.0;
    }

    /* load the TIRS results */
    sprintf(file_path, "%s/%s", calval_data_dir, tirs_calval_result_name);
    tirs_calval_result_fd = fopen(file_path, "r");
    fread(tirs_calval_result_data, DATA_SIZE, sizeof(MY_OUT_TYPE),
          tirs_calval_result_fd);
    fclose(tirs_calval_result_fd);

    /* read the cpf parameters */
    sprintf(file_path, "%s/%s", calval_data_dir, cpf_name);
    if (read_cpf_parameters(file_path, 1) != SUCCESS)
    {
        assert_true_with_message(0, "Error reading cpf file: %s", file_path);
        return;
    }

    /* patch the test data since it was generated using double math when
       it should have been done using floats.  calval plans to eventually
       fix it to use floats.  Until then, tweak this one value to allow the
       tests to pass */
    tirs_data[59] -= 0.000001;

    /* process the TIRS data */
    setup_radiance_parameters(1);
    setup_qcal_min_max(10);
    for (index = 0; index < DATA_SIZE; index++)
    {
#if DEBUG == 4 || DEBUG == 5
        printf("Mask = %d\n", mask_data[index]);
        printf("Index = %d %d\n", index/17, index%17);
#endif
        if (mask_data[index] != 0)
        {
            tirs_result_data[index] =
                wrapped_apply_scaling_factors(tirs_data[index]);
#if DEBUG == 5
            printf("TIRS Source = %0.15lf\n", tirs_data[index]);
            printf("TIRS Result = %0.15lf\n", tirs_result_data[index]);
#endif
        }
    }

    /* convert the data to unsigned short int(I*2) */
    wrapped_convert_data(tirs_result_data, tirs_short_data, 1, DATA_SIZE);

    /* test the TIRS results */
    for (index = 0; index < DATA_SIZE; index++)
    {
#if DEBUG == 6
        if (tirs_short_data[index] != tirs_calval_result_data[index])
        {
            printf("TIRS Source = %0.15lf\n", tirs_data[index]);
            printf("TIRS Result = %0.15lf\n", tirs_result_data[index]);
            printf("Mask = %d\n", mask_data[index]);
            printf("TIRS Short Result = %d\n", tirs_short_data[index]);
            printf("TIRS CalVal Result = %d\n", tirs_calval_result_data[index]);
            printf("Index = %d (%d %d)\n", index, index/17, index%17);
        }
#endif
        assert_equal(tirs_short_data[index], tirs_calval_result_data[index]);
    }

    free(tirs_data);
    free(tirs_calval_result_data);
    free(tirs_result_data);
    free(tirs_short_data);
}

/* Process a simulated scene using the read_input_image_data routine. */
static void test_read_l1r(int radiance_rescaling_on)
{
    char L1R_name[PATH_MAX] = "./RadianceRescale_L1R.h5";
    const IAS_SATELLITE_ATTRIBUTES *sat_attr = NULL;
    const int num_lines = 10;

    char file_path[PATH_MAX];
    char cmd[PATH_MAX];
    INPUT_IMAGE *input_image;
    int index;
    int band_index;
    int band_number;
    int num_scas;
    int sca_index;
    int num_det;
    int data_size = 0;
    double expected_value;

    IAS_THREAD_MUTEX_TYPE hdf_io_mutex;

    TILE_TYPE tile;
    float *tile_buffer = NULL;

    /* set the expected value depending on whether rescaling is done */
    if (radiance_rescaling_on)
        expected_value = 20.5;
    else
        expected_value = 42.0;

    /* create the test L1R image */
    if (create_rad_rescale_test_l1r_hdf(L1R_name, num_lines) != SUCCESS)
    {
        assert_true_with_message(0, "Create test L1R image failed, %s", 
                                 L1R_name);
        return;
    }

    /* Grab the attributes so things can be figured out. */
    sat_attr = ias_sat_attr_get_attributes();
    if (sat_attr == NULL)
    {
        assert_true_with_message(0, "Get satellite attirbutes failed");
        return;
    }

    /* read the cpf parameters */
    sprintf(file_path, "%s/%s", simulated_data_dir, cpf_name);
    if (read_cpf_parameters(file_path, radiance_rescaling_on) != SUCCESS)
    {
        assert_true_with_message(0, "Error reading cpf file: %s", file_path);
        return;
    }

    /* open the input image */
    input_image = open_input_image(L1R_name, NULL);
    if (!input_image)
    {
        assert_true_with_message(0, "Open input image failed: %s", L1R_name);
        return;
    }

    /* create a mutex to control access to the HDF library since it is
       completely unsafe for threads (can't even read and write to
       different files without problems) */
    if (IAS_THREAD_CREATE_MUTEX(&hdf_io_mutex) != 0)
    {
        close_input_image(input_image);
        assert_true_with_message(0, "Create control access mutex to HDF lib"
                                " failed.");
        return;
    }

    /* set the HDF library lock in the input image */
    set_input_image_io_lock(input_image, &hdf_io_mutex);

    /* process all the requested bands */
    /* FIXME - When TIRS support is added to the resampler, this could be
       updated to test TIRS */
    for (band_index = 0; band_index < 9; band_index++)
    {
        /* initialize the radiance rescale parameters for this band */
        setup_radiance_parameters(band_index);

        band_number = band_index+1;

#if DEBUG == 7 || DEBUG == 8
        printf("Processing Band %d\n", band_number);
#endif

        /* open the input band */
        if (open_input_band(input_image, band_number) != SUCCESS)
        {
            assert_true_with_message(0, "Open input band: %d failed, image: %s",
                                    band_number, input_image);
            return;
        }

        num_scas = sat_attr->band_attributes[band_index].scas;
        num_det = sat_attr->band_attributes[band_index].detectors_per_sca;

        data_size = num_det*num_lines;

        tile_buffer = calloc(data_size, sizeof(*tile_buffer));

        tile.start_line = 0;
        tile.start_sample = 0;
        tile.end_line = num_lines-1;
        tile.end_sample = num_det-1;
        tile.num_lines = num_lines;
        tile.num_samples = num_det;

        /* process each sca */
        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
#if DEBUG == 7 || DEBUG == 8
            printf("  Processing SCA %d\n", sca_index+1);
#endif

            read_input_image_data(input_image, sca_index,
                                  &tile, tile_buffer);

            /* test each value and only assert the failed ones */
            /* because of a cgreen bug that only allows so many asserts */
            for (index = 0; index < data_size; index++)
            {
#if DEBUG == 8
                printf("  Value = %d[%f]\n", index, tile_buffer[index]);
#endif

                if (tile_buffer[index] != expected_value)
                {
                    assert_equal(tile_buffer[index], expected_value);
                }
            }
        }

        free(tile_buffer);

        /* close the input image */
        if (close_input_band(input_image) != SUCCESS)
        {
            assert_true_with_message(0, "Close input band %d failed: %s", 
                                    band_number, input_image);
            return;
        }
    }

    /* close the input image */
    if (close_input_image(input_image) != SUCCESS)
    {
        assert_true_with_message(0, "Close input image failed: %s",
                                    input_image);
        return;
    }

    /* remove the test L1R image */
    sprintf(cmd, "rm %s", L1R_name);
    system(cmd);
}

/* run with rescaling turned on */
Ensure test_read_input_image()
{
    test_read_l1r(1);
}

/* run with rescaling turned off */
Ensure test_rescale_off()
{
    test_read_l1r(0);
}

int main(int argc, char **argv)
{
    int return_result = SUCCESS;
    char file_path[PATH_MAX];
    FILE *mask_fd = NULL;
    FILE *oli_fd = NULL;

    TestSuite *suite = create_test_suite();

    add_test(suite, test_oli_algorithm);
    add_test(suite, test_tirs_algorithm);
    add_test(suite, test_read_input_image);
    add_test(suite, test_rescale_off);

    /* allocate memory */
    mask_data = malloc(DATA_SIZE * sizeof(char));
    oli_data = malloc(DATA_SIZE * sizeof(double));

    /* initialize the memory */
    memset(mask_data, 0, DATA_SIZE * sizeof(char));
    memset(oli_data, 0, DATA_SIZE * sizeof(char));

    /* load the mask data */
    sprintf(file_path, "%s/%s", calval_data_dir, mask_name);
    mask_fd = fopen(file_path, "r");
    fread(mask_data, DATA_SIZE, sizeof(char), mask_fd);
    fclose(mask_fd);

    /* load the OLI test data */
    sprintf(file_path, "%s/%s", calval_data_dir, oli_name);
    oli_fd = fopen(file_path, "r");
    fread(oli_data, DATA_SIZE, sizeof(double), oli_fd);
    fclose(oli_fd);

    /* initialize the satellite attributes */
    if (ias_sat_attr_initialize(IAS_L8) != SUCCESS)
    {
        printf("Initialize satellite attributes failed.\n");
        exit(EXIT_FAILURE);
    }

    if (argc > 1)
        return_result = run_single_test(suite, argv[1], create_text_reporter());
    else
        return_result = run_test_suite(suite, create_text_reporter());

    free(mask_data);
    free(oli_data);

    return return_result;
}

