/*****************************************************************************
NAME: track_tile

PURPOSE:
  The track_tile module provides a mechanism for tracking which output tiles
  have been assigned to resampling threads and the next one that should be
  assigned.

ROUTINES:
    initialize_tile_tracking
    cleanup_tile_tracking
    get_output_tile_size
    get_assigned_output_tile

NOTES:
    none

*****************************************************************************/

#include <stdio.h>
#include "ias_const.h"
#include "ias_threadsync.h"
#include "ias_logging.h"
#include "track_tile.h"

static int tile_size = 200; /* size of the tiles assigned */
static int num_tiles;       /* total number of tiles to create */
static int tiles_across;    /* number of tiles across the image */
static int current_tile;    /* current tile working on */
static int status_period;   /* period for issuing status message */
static TILE_TYPE output_window; /* output window geometry */
static IAS_THREAD_MUTEX_TYPE tile_mutex; 
                            /* mutex to control access to the current_tile */
static int tile_mutex_initialized = 0; /* flag to track whether the tile_mutex
                                          has been initialized */

/*****************************************************************************
NAME: initialize_tile_tracking

PURPOSE:        
initialize_tile_tracking initializes the tile tracking mechanism for the 
current band.

RETURN VALUE:
SUCCESS     When initialization is successful
ERROR       When initialization is not successful

*****************************************************************************/

int initialize_tile_tracking
(
    int image_line_count,       /* I: number of lines in the image */
    int image_sample_count      /* I: number of samples across the image */
)
{
    int tile_rows;

    output_window.start_line = 0;
    output_window.end_line = image_line_count;

    output_window.start_sample = 0;
    output_window.end_sample = image_sample_count;

    output_window.num_lines = output_window.end_line - output_window.start_line;
    output_window.num_samples
            = output_window.end_sample - output_window.start_sample;

    current_tile = 0;
    tiles_across = (output_window.num_samples + tile_size - 1)/tile_size;
    tile_rows = (output_window.num_lines + tile_size - 1)/tile_size;
    num_tiles = tile_rows * tiles_across;
    status_period = num_tiles/5;

    /* create the mutex for access to the current_tile */
    if (!tile_mutex_initialized)
    {
        if (IAS_THREAD_CREATE_MUTEX(&tile_mutex) != 0)
        {
            IAS_LOG_ERROR("Error creating tile tracking mutex");
            return ERROR;
        }
        tile_mutex_initialized = 1;
    }

    return SUCCESS;
}

/*****************************************************************************
NAME: cleanup_tile_tracking

PURPOSE:        
cleanup_tile_tracking cleans up the resources used for tile tracking.

RETURNS:
    SUCCESS or ERROR

*****************************************************************************/
int cleanup_tile_tracking()
{
    if (tile_mutex_initialized)
    {
        tile_mutex_initialized = 0;
        if (IAS_THREAD_DESTROY_MUTEX(&tile_mutex) != 0)
        {
            IAS_LOG_ERROR("Failed destroying the tile tracking mutex");
            return ERROR;
        }
    }
    return SUCCESS;
}

/*****************************************************************************
NAME: get_output_tile_size

PURPOSE:        
get_output_tile_size provides a method for external modules to learn the size
of the output tiles assigned (tiles are assumed to be square nominally).

RETURN VALUE:
Returns the number of lines/samples in the tile

*****************************************************************************/
int get_output_tile_size(void)
{
    return (tile_size);
}

/*****************************************************************************
NAME: get_tile_by_index

PURPOSE:        
get_tile_by_index is an internal function that populates the output tile 
location given a tile index.

RETURN VALUE:
SUCCESS     When the tile location is successfully populated
ERROR       When no more tiles remain to be assigned

*****************************************************************************/
static int get_tile_by_index
(
    int tile_index,         /* I: index of tile to get */
    TILE_TYPE *out_tile     /* O: tile location */
)
{
    int row;
    int column;

    if (tile_index >= num_tiles)
    {
        /* no tile to assign, so return ERROR to indicate that but don't
           output an error message since it isn't really an error in the larger
           scope of things */
        return ERROR;
    }

    row = tile_index / tiles_across;
    column = tile_index - row * tiles_across;

    /* assign the tile to return */
    out_tile->start_line = output_window.start_line + row * tile_size;
    out_tile->start_sample = output_window.start_sample + column * tile_size;
    out_tile->end_line = out_tile->start_line + tile_size;
    out_tile->end_sample = out_tile->start_sample + tile_size;

    /* limit the ending line and sample to be inside the image */
    if (out_tile->end_line > output_window.end_line)
        out_tile->end_line = output_window.end_line;
    if (out_tile->end_sample > output_window.end_sample)
        out_tile->end_sample = output_window.end_sample;
    out_tile->num_lines = out_tile->end_line - out_tile->start_line;
    out_tile->num_samples = out_tile->end_sample - out_tile->start_sample;

    return SUCCESS;
}

/*****************************************************************************
NAME: get_assigned_output_tile

PURPOSE:        
get_assigned_output_tile returns the next tile that needs to be created in 
the output image.

RETURN VALUE:
1           When a tile is successfully assigned
0           When a no more tiles remain to be assigned
ERROR       When an error occurs

NOTES:
none

ALGORITHM REFERENCES:
none

*****************************************************************************/
int get_assigned_output_tile
(
    TILE_TYPE *out_tile,    /* O: output tile location */
    int *tile_index         /* O: tile index assigned */
)
{
    int tile_assigned = 0;  /* assume no tile to assign */

    /* lock the current tile mutex */
    if (IAS_THREAD_LOCK_MUTEX(&tile_mutex) != 0)
    {
        IAS_LOG_ERROR("Error locking the tile mutex");
        return ERROR;
    }

    *tile_index = current_tile;
    current_tile++;

    /* unlock the current tile mutex */
    if (IAS_THREAD_UNLOCK_MUTEX(&tile_mutex) != 0)
    {
        IAS_LOG_ERROR("Error unlocking the tile mutex");
        return ERROR;
    }

    if (get_tile_by_index(*tile_index, out_tile) == SUCCESS)
    {
        tile_assigned = 1;
    }

#if DEBUG
    /* periodically issue a status message */
    if (tile_assigned && ((*tile_index == num_tiles) || 
        ((*tile_index % status_period) == 0)))
    {
        IAS_LOG_INFO("Resampling tile %d of %d at %d, %d\n", *tile_index,
                num_tiles, out_tile->start_line, out_tile->start_sample);
    }
#endif

    return tile_assigned;
}
