/*****************************************************************************
NAME: transform_window

PURPOSE:        
The transform_window module contains the routines to transform windows defined
with line/samples between input and output space.


ROUTINES:
    check_point
    transform_ow2iw
    calculate_input_tile_to_do

NOTES:
None

ALGORITHM REFERENCES:
None

*****************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "ias_logging.h"
#include "kernel.h"
#include "resample_image.h"
#include "get_resample_parameters.h"
#include "transform_window.h"

#define NORMAL_DISTANCE 10
/* normal distance allowed between check points when determining min/max 
   line/sample values */
#define BREAK_DISTANCE 2
/* distance allowed between check points when determining min/max line/sample 
   values when one of the corners involved is in a break area */
#define FIND_DISTANCE_LIMIT(a,b) ((valid_corner[(a)] && valid_corner[(b)]) ? \
    NORMAL_DISTANCE : BREAK_DISTANCE)
/* macro to chose between the normal and in-break distance based on whether the
   corner points both have a valid mapping back to input space */
#define WINDOW_NOT_CALCULATED -3
/* return status indicating the input space window was not calculated */

/*****************************************************************************
FUNCTION NAME: check_point

PURPOSE:
Given two output space line/sample pairs, check along the line connecting them
for new input space max/min values.

RETURNS:
    SUCCESS or ERROR
    The min/max line and sample of the input space window and the required 
    line overlap are returned as parameters.

*****************************************************************************/
static int check_point
(
    const IAS_GRID_BAND_TYPE *grid_band_ptr, /* I: grid to use */
    int terrain_flag,       /* I: terrain correction flag (1=correct terrain) */
    IAS_IMAGE *dem_img,     /* I: DEM image for terrain correction */
    const int output_lines[2],   /* I: lines and samples of the window      */
    const int output_samples[2], /* I: corners in output space              */
    int *min_line_ptr,     /* I/O: min line number in transformed window    */
    int *max_line_ptr,     /* I/O: max line number in transformed window    */ 
    int *min_sample_ptr,   /* I/O: min sample number in transformed window  */
    int *max_sample_ptr,   /* I/O: max sample number in transformed window  */
    int *min_overlap,      /* I/O: amount of line overlap for top of
                                                                input space */
    int *max_overlap,      /* I/O: amount of line overlap for bottom of
                                                                input space */
    int distance           /* I: distance to allow between check points */
)
{
    int midpt_line, midpt_sample; /* output space midpoint */
    int new_out_lines[2];       /* new output line/sample pairs to check */
    int new_out_samples[2];
    double d_input_line;        /* line and sample in input space */
    double d_input_sample;
    double dtemp;               /* temporary value */
    int status;                 /* function return status */


    /* Do we need to check the midpoint for this pair?  How large (in output
       space) is the point separation?
       ---------------------------------------------------------------------*/
    dtemp = sqrt((double)(output_lines[0] - output_lines[1])*
            (double)(output_lines[0] - output_lines[1]) +
            (double)(output_samples[0] - output_samples[1])*
            (double)(output_samples[0] - output_samples[1]));
    if (dtemp < distance)
        return SUCCESS;

    /* Find the output space midpoint */
    midpt_line = (output_lines[0] + output_lines[1])/2;
    midpt_sample = (output_samples[0] + output_samples[1])/2;

    /* Convert the output space midpoint to input space */
    status = do_ols2ils(grid_band_ptr, terrain_flag, dem_img, midpt_line,
                        midpt_sample, &d_input_line, &d_input_sample);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Converting output space midpoint line %d sample %d "
            "to input space", midpt_line, midpt_sample);
        return ERROR;
    }
    else if (status == IAS_GRID_ILS_NOT_CALCULATED)
    {
        d_input_line = *min_line_ptr;
        d_input_sample = *min_sample_ptr;
    }

    /* Will the midpoint change the min/max line and sample values?
       If so, check the amount of line overlap required. */
    if ((int)d_input_line < *min_line_ptr)
    {
        *min_overlap += *min_line_ptr - (int)d_input_line;
        *min_line_ptr = (int)d_input_line;
    }
    if ((int)d_input_line > *max_line_ptr)
    {
        *max_overlap += (int)d_input_line - *max_line_ptr;
        *max_line_ptr = (int)d_input_line;
    }
    if ((int)d_input_sample < *min_sample_ptr)
        *min_sample_ptr = (int)d_input_sample;
    if ((int)d_input_sample > *max_sample_ptr)
        *max_sample_ptr = (int)d_input_sample;

    /* Check the two new line segments created */
    new_out_lines[0] = midpt_line;
    new_out_samples[0] = midpt_sample;

    new_out_lines[1] = output_lines[0];
    new_out_samples[1] = output_samples[0];
    if (check_point(grid_band_ptr, terrain_flag, dem_img, new_out_lines,
                new_out_samples, min_line_ptr, max_line_ptr, min_sample_ptr, 
                max_sample_ptr, min_overlap, max_overlap, distance) == ERROR)
    {
        /* Don't log an error since this can be one of many recursive calls */
        return ERROR;
    }

    new_out_lines[1] = output_lines[1];
    new_out_samples[1] = output_samples[1];
    if (check_point(grid_band_ptr, terrain_flag, dem_img, new_out_lines,
                new_out_samples, min_line_ptr, max_line_ptr, min_sample_ptr, 
                max_sample_ptr, min_overlap, max_overlap, distance) == ERROR)
    {
        /* Don't log an error since this can be one of many recursive calls */
        return ERROR;
    }

    return SUCCESS;
}


/*****************************************************************************
FUNCTION NAME: transform_ow2iw

PURPOSE:        
transform_ow2iw transforms a window in output space into a square window in 
input space that fully contains the output space window.

RETURNS:
    SUCCESS, ERROR, or WINDOW_NOT_CALCULATED

*****************************************************************************/

static int transform_ow2iw
(
    const IAS_GRID_BAND_TYPE *grid_band_ptr, /* I: grid to use */
    int terrain_flag,       /* I: terrain correction flag (1=correct terrain) */
    IAS_IMAGE *dem_img,     /* I: DEM image for terrain correction */
    int max_line_offset,    /* I: maximum even/odd line offset */
    const TILE_TYPE *output_window_ptr, /* I: pointer to output window */
    TILE_TYPE *input_window_ptr /* O: pointer to input space window */
)
{
    int min_overlap = 0;      /* number of overlap lines required along
                                 top of input image                         */
    int max_overlap = 0;      /* number of overlap lines required along
                                 bottom of input image                      */
    int loop;                 /* loop counter                               */
    int output_lines[4];      /* lines and samples of the window corners in */
    int output_samples[4];    /* output space                               */
    int temp_line, temp_samp; /* temporary line/sample pair                 */
    double d_input_lines[4];  /* lines and samples in input space           */
    double d_input_samples[4];
    int min_line;             /* minimum line number in transformed window   */
    int max_line;             /* maximum line number in transformed window   */ 
    int min_sample;           /* minimum sample number in transformed window */
    int max_sample;           /* maximum sample number in transformed window */
    int valid_corner[4] = {0,0,0,0};
    int valid_count = 0;
    int distance;
    int status;               /* function return status */

    /* Get the output window into a useable structure */
    output_lines[0] = output_window_ptr->start_line;
    output_lines[1] = output_window_ptr->start_line;
    output_lines[2] = output_window_ptr->end_line - 1;
    output_lines[3] = output_window_ptr->end_line - 1;
    output_samples[0] = output_window_ptr->start_sample;
    output_samples[1] = output_window_ptr->end_sample - 1;
    output_samples[2] = output_window_ptr->start_sample;
    output_samples[3] = output_window_ptr->end_sample - 1;

    /* Convert each of the output space corners into an input space pixel */
    for (loop = 0; loop < 4; loop++)
    {
        status = do_ols2ils(grid_band_ptr, terrain_flag, dem_img, 
                       output_lines[loop], output_samples[loop], 
                       &d_input_lines[loop], &d_input_samples[loop]);
        if (status == ERROR)
        {
            IAS_LOG_ERROR("Converting output space line %d sample %d "
                "to input space", output_lines[loop], output_samples[loop]);
            return ERROR;
        }
        else if (status == SUCCESS)
        {
            valid_corner[loop] = 1;
            valid_count++;
        }
        /* Otherwise ILS wasn't calculated so don't increment valid count */
    }

    if (valid_count == 0)
    {
        /* do not output an error message since returning this status here is 
           done to communicate with the caller that the mapping failed.  The 
           calling routine needs to decide whether that is a scenario that 
           deserves an error message */
        return WINDOW_NOT_CALCULATED;
    }

    /* Find the min/max line and sample from the corners */
    min_line = 100000000;
    max_line = -100000000;
    min_sample = 100000000;
    max_sample = -100000000;
    for (loop = 0; loop < 4; loop++)
    {
        if (!valid_corner[loop])
            continue;
        if (d_input_lines[loop] < min_line)
            min_line = d_input_lines[loop];
        if (d_input_lines[loop] > max_line)
            max_line = d_input_lines[loop];

        if (d_input_samples[loop] < min_sample)
            min_sample = d_input_samples[loop];
        if (d_input_samples[loop] > max_sample)
            max_sample = d_input_samples[loop];
    }

    /* Check along the output space boundaries for input space lines and
       samples that modify the max/min values.
       -----------------------------------------------------------------*/
    /* Top */
    distance = FIND_DISTANCE_LIMIT(0,1);
    if (check_point(grid_band_ptr, terrain_flag, dem_img, output_lines,
                output_samples, &min_line, &max_line, &min_sample, &max_sample,
                &min_overlap, &max_overlap, distance) == ERROR)
    {
        IAS_LOG_ERROR("Checking output points line %d sample %d and line %d "
            "sample %d", output_lines[0], output_samples[0], output_lines[1], 
            output_samples[1]);
        return ERROR;
    }
    /* Bottom */
    distance = FIND_DISTANCE_LIMIT(2,3);
    if (check_point(grid_band_ptr, terrain_flag, dem_img, &output_lines[2],
                &output_samples[2], &min_line, &max_line, &min_sample,
                &max_sample, &min_overlap, &max_overlap, distance) == ERROR)
    {
        IAS_LOG_ERROR("Checking output points line %d sample %d and line %d "
            "sample %d", output_lines[2], output_samples[2], output_lines[3], 
            output_samples[3]);
        return ERROR;
    }
    /* Exchange UR and LL output points. */
    temp_line = output_lines[1];
    temp_samp = output_samples[1];
    output_lines[1] = output_lines[2];
    output_samples[1] = output_samples[2];
    output_lines[2] = temp_line;
    output_samples[2] = temp_samp;
    /* Left */
    distance = FIND_DISTANCE_LIMIT(0,2);
    if (check_point(grid_band_ptr, terrain_flag, dem_img, output_lines,
                output_samples, &min_line, &max_line, &min_sample, &max_sample,
                &min_overlap, &max_overlap, distance) == ERROR)
    {
        IAS_LOG_ERROR("Checking output points line %d sample %d and line %d "
            "sample %d", output_lines[0], output_samples[0], output_lines[1], 
            output_samples[1]);
        return ERROR;
    }
    /* Right */
    distance = FIND_DISTANCE_LIMIT(1,3);
    if (check_point(grid_band_ptr, terrain_flag, dem_img, &output_lines[2],
                &output_samples[2], &min_line, &max_line, &min_sample,
                &max_sample, &min_overlap, &max_overlap, distance) == ERROR)
    {
        IAS_LOG_ERROR("Checking output points line %d sample %d and line %d "
            "sample %d", output_lines[2], output_samples[2], output_lines[3], 
            output_samples[3]);
        return ERROR;
    }

    /* Add one to the max values to account for integer truncation */
    max_line += 1;
    max_sample += 1;

    /* Adjust the window size for the kernel size (add 1 to be conservative)
       and the maximum impact of the detector offsets and a little extra in
       the sample direction. */
    min_line -= (num_top_kernel_lines() + 1 + max_line_offset);
    max_line += (num_bottom_kernel_lines() + 1 + max_line_offset);
    min_sample -= (num_left_kernel_samples() + 1 + 8);
    max_sample += (num_right_kernel_samples() + 1 + 8);

    /* return the input window needed */
    input_window_ptr->start_line = min_line;
    input_window_ptr->start_sample = min_sample;
    input_window_ptr->end_line = max_line + 1;
    input_window_ptr->end_sample = max_sample + 1;
    input_window_ptr->num_lines = 
            input_window_ptr->end_line - input_window_ptr->start_line;
    input_window_ptr->num_samples = 
            input_window_ptr->end_sample - input_window_ptr->start_sample;

    IAS_LOG_DEBUG("Input window parameters\n"
                  "   start line/sample %d,%d\n"
                  "   end line/sample %d,%d\n"
                  "   number of lines %d\n"
                  "   number of samples %d",
                  input_window_ptr->start_line,input_window_ptr->start_sample,
                  input_window_ptr->end_line,input_window_ptr->end_sample,
                  input_window_ptr->num_lines,
                  input_window_ptr->num_samples);

    return SUCCESS;
}

/*****************************************************************************
FUNCTION NAME: calculate_input_tile_to_do

PURPOSE:        
calculate_input_tile_to_do calculates the input window that fully contains the 
output space window.

RETURNS:
    SUCCESS or ERROR

*****************************************************************************/
int calculate_input_tile_to_do
(
    const IAS_GRID_BAND_TYPE *grid_band_ptr, /* I: grid to use */
    INPUT_IMAGE *input_img,     /* I: input image to use (for nl, ns) */
    int terrain_flag,   /* I: terrain correction flag (1=correct terrain) */
    IAS_IMAGE *dem_img,         /* I: DEM image for terrain correction */
    int max_line_offset,        /* I: maximum even/odd line offset */
    const TILE_TYPE *out_tile,  /* I: output space tile to do */
    TILE_TYPE *in_tile          /* O: input space tile */
)
{
    int in_lines = input_img->nl;
    int in_samples = input_img->ns;
    int status;                 /* function return status */

    /* Transform the output tile line/sample to input image coordinates */
    status = transform_ow2iw(grid_band_ptr, terrain_flag, dem_img, 
             max_line_offset, out_tile, in_tile);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Transforming output tile start line %d start sample %d "
            "end line %d end sample %d to input coordinates", 
            out_tile->start_line, out_tile->start_sample, out_tile->end_line,
            out_tile->end_sample);
        return ERROR;
    }
    if (status == WINDOW_NOT_CALCULATED)
    {
        /* tile transformation failed, so no mapping could be found from
           output space to input space for this tile.  Just return a dummy 
           tile to let the output tile be filled with fill values. */
        IAS_LOG_WARNING("Tile transform failed for tile at %d, %d",
                        out_tile->start_line, out_tile->start_sample);
        in_tile->start_line = 0;
        in_tile->end_line = 1;
        in_tile->start_sample = 0;
        in_tile->end_sample = 1;
        in_tile->num_lines = in_tile->end_line - in_tile->start_line;
        in_tile->num_samples = in_tile->end_sample - in_tile->start_sample;
        return SUCCESS;
    }

    /* limit the tile to the extents of the real input image */
    if (in_tile->start_line < 0)
        in_tile->start_line = 0;
    if (in_tile->start_line > in_lines)
        in_tile->start_line = in_lines;
    if (in_tile->end_line < 0)
        in_tile->end_line = 0;
    if (in_tile->end_line > in_lines)
        in_tile->end_line = in_lines;
    if (in_tile->start_sample < 0)
        in_tile->start_sample = 0;
    if (in_tile->start_sample > in_samples)
        in_tile->start_sample = in_samples;
    if (in_tile->end_sample < 0)
        in_tile->end_sample = 0;
    if (in_tile->end_sample > in_samples)
        in_tile->end_sample = in_samples;
    in_tile->num_lines = in_tile->end_line - in_tile->start_line;
    in_tile->num_samples = in_tile->end_sample - in_tile->start_sample;

    return SUCCESS;
}
