/******************************************************************************
NAME: update_l1g_metadata

PURPOSE: Update the L1G output image metadata.

RETURNS
    SUCCESS or ERROR

******************************************************************************/
#include <stdlib.h>
#include <string.h>
#include "ias_logging.h"
#include "ias_l1g.h"
#include "ias_l0r.h"
#include "ias_satellite_attributes.h"
#include "ias_miscellaneous.h"
#include "get_resample_parameters.h" /* resampler control parameters */
#include "debug.h"                   /* ASSERT macro */
#include "buffer_output_image.h"
#include "get_input_image.h"
#include "get_cpf_parameters.h"
#include "resample_image.h"
#include "update_l1g_metadata.h"     /* prototypes for this module */

int update_l1g_metadata
(
    OUTPUT_BUFFER *output_image_ptr,/* I: pointer to the output image */
    INPUT_IMAGE *input_image_ptr,   /* I: pointer to the input image */
    const IAS_GRID_TYPE *grid_ptr,  /* I: pointer to grid */
    int bands_to_output,            /* I: number of bands to write */
    const int *band_indexes,        /* I: pointer to list of bands to output */
    size_t scene_id_size,           /* I: size of landsat scene id buffer */
    size_t spacecraft_size,         /* I: size of spacecraft id buffer */
    char *landsat_scene_id,         /* O: pointer to scene id output */
    char *spacecraft_id             /* O: pointer to spacecraft id output */
)
{
    L0RIO *l0r_img = NULL;
    L1RIO *l1r_img = NULL;
    L1GIO *l1g_img = NULL;
    IAS_L1G_FILE_METADATA fmd;    /* file metadata */
    IAS_L1G_BAND_METADATA *bmd;   /* band metadata */
    int close_l0r;                /* flag to indicate L0R should be closed */
    int count;                    /* loop counter */
    int band_index;               /* band index */
    int band_number;              /* band number */
    int index;                    /* temporary index */
    int have_l1r = 0;
    int scene_count = 0;
                                     
    IAS_L0R_INTERVAL_METADATA l0r_interval_metadata;
    const char *ias_version = NULL;
    IAS_DATETIME date_time;
    const char *satellite_name = NULL;

    memset(&fmd, 0, sizeof(fmd));

    have_l1r = input_image_ptr->is_l1r;

    /* Retrieve the pointer to the L1RIO and L0RIO data structures */
    if (have_l1r)
    {
        IAS_L1R_FILE_METADATA l1r_fmd; /* L1R input file metadata */

        l1r_img = input_image_ptr->l1r_file;
        
        /* Retrieve the L1R file metadata */
        if (ias_l1r_get_file_metadata(l1r_img, &l1r_fmd) != SUCCESS)
        {
            IAS_LOG_ERROR("Retrieving the L1R file metadata");
            return ERROR;
        }
        fmd.earth_sun_distance = l1r_fmd.earth_sun_distance;

        /* Open and assign the L0R here */
        l0r_img = ias_l0r_open_from_filename(get_0R_filename(),
                                             IAS_L0R_COMPRESSION_DEFAULT);
        if (l0r_img == NULL)
        {
            IAS_LOG_ERROR("Opening L0R");
            return ERROR;
        }
        close_l0r = 1;
    }
    else
    {
        fmd.earth_sun_distance = 0.0;
        l0r_img = input_image_ptr->l0r_file;
        close_l0r = 0;
    }
    
    /* Copy the sun angle correction */
    strncpy(fmd.sun_angle_correction, "NONE", sizeof(fmd.sun_angle_correction));
    fmd.sun_angle_correction[sizeof(fmd.sun_angle_correction) - 1] = '\0';

    /* Grab the path row from the omf file */
    fmd.wrs_path = get_wrs_path();
    fmd.wrs_row = get_wrs_row();

    /* open the L0R metadata file */
    if (ias_l0r_open_metadata(l0r_img, IAS_READ) != SUCCESS)
    {
        IAS_LOG_ERROR("Opening L0R metadata");
        if (close_l0r)
            ias_l0r_close(l0r_img);
        return ERROR;
    }

    /* Grab the interval metadata to use later. */
    if (ias_l0r_get_interval_metadata(l0r_img, &l0r_interval_metadata)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Reading L0R interval metadata");
        ias_l0r_close_metadata(l0r_img);
        if (close_l0r)
            ias_l0r_close(l0r_img);
        return ERROR;
    }

    /* Copy the spacecraft id */
    strncpy(spacecraft_id, l0r_interval_metadata.spacecraft_id,
        spacecraft_size);
    spacecraft_id[spacecraft_size - 1] = '\0';
    ias_misc_convert_to_uppercase(spacecraft_id);

    /* Get the count of scenes in the metadata */
    if (ias_l0r_get_scene_metadata_records_count(l0r_img, &scene_count) 
        != SUCCESS)
    {
        IAS_LOG_ERROR("Getting L0R scene metadata size");
        ias_l0r_close_metadata(l0r_img);
        if (close_l0r)
            ias_l0r_close(l0r_img);
        return ERROR;
    }

    if (scene_count > 0)
    {
        int scene_index = 0;
        IAS_L0R_SCENE_METADATA *l0r_metadata_scenes = NULL;

        /* Allocate memory for the scene metadata */
        l0r_metadata_scenes = malloc(sizeof(IAS_L0R_SCENE_METADATA)
            * scene_count);
        if (l0r_metadata_scenes == NULL)
        {
            IAS_LOG_ERROR("Allocating memory for L0R scene metadata");
            ias_l0r_close_metadata(l0r_img);
            if (close_l0r)
                ias_l0r_close(l0r_img);
            return ERROR;
        }
        memset(l0r_metadata_scenes, 0, sizeof(IAS_L0R_SCENE_METADATA)
                * scene_count);

        /* Read all the scenes in */
        if (ias_l0r_get_scene_metadata(l0r_img, 0, scene_count,
                                       l0r_metadata_scenes) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading L0R scene metadata");
            free(l0r_metadata_scenes);
            ias_l0r_close_metadata(l0r_img);
            if (close_l0r)
                ias_l0r_close(l0r_img);
            return ERROR;
        }

        /* Find the scene metadata for the path/row being processed */
        for (scene_index = 0; scene_index < scene_count; scene_index++)
        {
            if (fmd.wrs_path 
                    == l0r_metadata_scenes[scene_index].target_wrs_path
                && fmd.wrs_row 
                    == l0r_metadata_scenes[scene_index].target_wrs_row)
            {
                break;
            }
        }
        if (scene_index == scene_count) /* Make sure the scene was found */
        {
            IAS_LOG_ERROR("L0R scene metadata not found");
            free(l0r_metadata_scenes);
            return ERROR;
        }

        fmd.roll_angle = l0r_metadata_scenes[scene_index].roll_angle;
        fmd.sun_azimuth = l0r_metadata_scenes[scene_index].sun_azimuth;
        fmd.sun_elevation = l0r_metadata_scenes[scene_index].sun_elevation;
        fmd.sun_angles_valid = 1;

        /* Copy the landsat scene id */
        strncpy(landsat_scene_id, 
            l0r_metadata_scenes[scene_index].landsat_scene_id,
            scene_id_size);
        landsat_scene_id[scene_id_size - 1] = '\0';

        /* Parse the datetime string provided by the L0R */
        if (ias_misc_parse_datetime_string(
                l0r_metadata_scenes[scene_index].date_acquired,
                IAS_DATETIME_L0R_FORMAT, &date_time) != SUCCESS)
        {
            IAS_LOG_ERROR("Parsing L0R scene metadata date acquired string");
            free(l0r_metadata_scenes);
            return ERROR;
        }

        free(l0r_metadata_scenes);
        l0r_metadata_scenes = NULL;
    }
    else
    {
        /* The scene metadata wasn't available (i.e. Lunar collection), so
           set the roll angle to zero and get the date acquired from the
           interval metadata */
        fmd.roll_angle = 0.0;
        fmd.sun_azimuth = 0.0;
        fmd.sun_elevation = 0.0;
        fmd.sun_angles_valid = 0;

        if (ias_misc_parse_datetime_string(l0r_interval_metadata.date_acquired,
                IAS_DATETIME_L0R_FORMAT, &date_time) != SUCCESS)
        {
            IAS_LOG_ERROR("Parsing L0R interval date acquired string");
            return ERROR;
        }
    }

    /* close the L0R metadata */
    ias_l0r_close_metadata(l0r_img);
    if (close_l0r)
    {
        ias_l0r_close(l0r_img);
        l0r_img = NULL;
    }

    /* Retrieve the pointer to the L1GIO */
    l1g_img = get_l1gio_ptr(output_image_ptr);

    fmd.projection_code = grid_ptr->code;

    fmd.zone_code = grid_ptr->zone;

    strncpy(fmd.datum, grid_ptr->datum, sizeof(fmd.datum));
    fmd.datum[sizeof(fmd.datum) - 1] = '\0';
    fmd.spheroid_code = grid_ptr->spheroid;

    strncpy(fmd.projection_units, grid_ptr->units,
            sizeof(fmd.projection_units));
    fmd.projection_units[sizeof(fmd.projection_units) - 1] = '\0';
    for (index = 0; index < IAS_PROJ_PARAM_SIZE; index++)
    {
        fmd.projection_parameters[index] = grid_ptr->projprms[index];
    }

    /* Set the spacecraft Id */
    satellite_name = ias_sat_attr_get_satellite_name();

    if (strcmp(satellite_name, "Unknown") == 0)
    {
        IAS_LOG_ERROR("Unable to determine satellite name.");
        return ERROR;
    }
    strncpy(fmd.spacecraft, satellite_name, sizeof(fmd.spacecraft));
    fmd.spacecraft[sizeof(fmd.spacecraft) - 1] = '\0';

    strncpy(fmd.collection_type, l0r_interval_metadata.collection_type,
            sizeof(fmd.collection_type));
    fmd.collection_type[sizeof(fmd.collection_type) - 1] = '\0';

    /* Figure out if the scene is ascending or descending.
       Ascending if (row >= 123 and row <= 246)
       I shortened the if in the following code to not perform the =. */
    if (fmd.wrs_row > 122 && fmd.wrs_row < 247)
    {
        strncpy(fmd.capture_direction, "Ascending",
                sizeof(fmd.capture_direction));
    }
    else
    {
        strncpy(fmd.capture_direction, "Descending",
                sizeof(fmd.capture_direction));
    }
    fmd.capture_direction[sizeof(fmd.capture_direction) - 1] = '\0';


    /* Set the date and time values for the L1G */
    snprintf(fmd.capture_date, sizeof(fmd.capture_date), "%04d/%02d/%02d",
            date_time.year, date_time.month, date_time.day_of_month);
    snprintf(fmd.capture_time, sizeof(fmd.capture_time), "%02d:%02d:%02d",
            date_time.hour, date_time.minute, (int)date_time.second);

    /* Set the correction type */
    if (grid_ptr->correction_type == IAS_PRECISION)
    {
        strncpy(fmd.correction_type, "L1T", sizeof(fmd.correction_type));
    }
    else if (do_terrain_correction())
    {
        strncpy(fmd.correction_type, "L1Gt", sizeof(fmd.correction_type));
    }
    else
    {
        /* assume systematic correction */
        strncpy(fmd.correction_type, "L1G", sizeof(fmd.correction_type));
    }
    fmd.correction_type[sizeof(fmd.correction_type) - 1] = '\0';

    /* Set the resample type */
    switch (get_resampling_type())
    {
        case CC_RESAMPLING :
            strncpy(fmd.resample_type, "CC", sizeof(fmd.resample_type));
            break;

        case NN_RESAMPLING :
            strncpy(fmd.resample_type, "NN", sizeof(fmd.resample_type));
            break;

        default:
            IAS_LOG_ERROR("Unable to determine resample type.");
            return ERROR;
            break;
    }
    fmd.resample_type[sizeof(fmd.resample_type) - 1] = '\0';

    /* Grab the software versions. */
    ias_version = ias_misc_get_software_version();
    if (ias_version == NULL)
        strncpy(fmd.software_version, "", sizeof(fmd.software_version));
    else
    {
        strncpy(fmd.software_version, ias_version,
                sizeof(fmd.software_version));
    }
    fmd.software_version[sizeof(fmd.software_version) - 1] = '\0';

    strncpy(fmd.ingest_software_version, l0r_interval_metadata.is_version,
            sizeof(fmd.ingest_software_version));
    fmd.ingest_software_version[sizeof(fmd.ingest_software_version) - 1] = '\0';

    /* Allocate memory for the band metadata */
    bmd = malloc(sizeof(*bmd) * bands_to_output);
    if (!bmd)
    {
        IAS_LOG_ERROR("Allocating memory for band metadata");
        return ERROR;
    }
    memset(bmd, 0, sizeof(IAS_L1G_BAND_METADATA) * bands_to_output);

    for (count = 0; count < bands_to_output; count++)
    {
        const char *band_name;
        const IAS_BAND_ATTRIBUTES *band_attributes = NULL;
        const char *sensor_name = NULL;

        band_index = band_indexes[count];

        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Error retrieving band number for band index %d",
                          band_index);
            free(bmd);
            return ERROR;
        }

        band_attributes = ias_sat_attr_get_band_attributes(band_number);
        if (!band_attributes)
        {
            IAS_LOG_ERROR("Error retrieving band attributes for band "
                          "number %d", band_number);
            free(bmd);
            return ERROR;
        }

        /* Set the band number */
        bmd[count].band_number = band_number;

        /* Grab the band name and apply it to the L1G */
        band_name = ias_sat_attr_convert_band_number_to_name(band_number);
        if (!band_name)
        {
            IAS_LOG_ERROR("Error retrieving band name for band number %d",
                          band_number);
            free(bmd);
            return ERROR;
        }
        strncpy(bmd[count].band_name, band_name, sizeof(bmd[count].band_name));
        bmd[count].band_name[sizeof(bmd[count].band_name) - 1] = '\0';

        /* copy the projection window to the metadata */
        bmd[count].upper_left_x = grid_ptr->corners.upleft.x;
        bmd[count].upper_left_y = grid_ptr->corners.upleft.y;

        bmd[count].upper_right_x = grid_ptr->corners.upright.x;
        bmd[count].upper_right_y = grid_ptr->corners.upright.y;

        bmd[count].lower_right_x = grid_ptr->corners.loright.x;
        bmd[count].lower_right_y = grid_ptr->corners.loright.y;

        bmd[count].lower_left_x = grid_ptr->corners.loleft.x;
        bmd[count].lower_left_y = grid_ptr->corners.loleft.y;

        /* set the projected pixel size */
        bmd[count].projection_distance_y
                = grid_ptr->gridbands[band_index][0].pixsize;
        bmd[count].projection_distance_x 
                = grid_ptr->gridbands[band_index][0].pixsize;

        /* set the dynamic range of the output pixels */
        bmd[count].maximum_pixel_value = get_max_range_value();
        bmd[count].minimum_pixel_value = get_min_range_value();
        bmd[count].pixel_range_valid = TRUE;

        /* init these to false */
        bmd[count].radiance_valid = FALSE;
        bmd[count].reflectance_valid = FALSE;
        if (have_l1r)
        {
            double ref_to_rad_coefficient = 0.0;
            double scaling_multiplier = 1.0;
            double scaling_offset = 0.0;
            IAS_L1R_BAND_METADATA l1r_band_metadata;

            /* get the reflectance to radiance coefficient value from the
               L1R metadata for the current band */
            if (ias_l1r_get_band_metadata(l1r_img, band_number,
                    &l1r_band_metadata) != SUCCESS)
            {
                IAS_LOG_ERROR("Failed to read band %d metadata from %s",
                              band_number, get_1R_filename());
                free(bmd);
                return ERROR;
            }
            ref_to_rad_coefficient =
                l1r_band_metadata.reflectance_to_radiance_coefficient;

            /* get the radiance rescale parameters from the CPF */
            scaling_multiplier = get_radiance_rescale_gain(band_index);
            scaling_offset = get_radiance_rescale_bias(band_index);

            /* set these to a no-op and not valid */
            bmd[count].reflectance_scaling_gain = 1.0;
            bmd[count].reflectance_scaling_offset = 0.0;
            bmd[count].reflectance_valid = FALSE;

            bmd[count].radiance_valid = TRUE; /* fixed to not valid below */

            if (band_attributes->sensor_id == IAS_OLI)
            {
                bmd[count].spectral_radiance_scaling_gain =
                    ref_to_rad_coefficient * scaling_multiplier;
                bmd[count].spectral_radiance_scaling_offset =
                    ref_to_rad_coefficient * scaling_offset;

                bmd[count].reflectance_scaling_gain = scaling_multiplier;
                bmd[count].reflectance_scaling_offset = scaling_offset;

                bmd[count].reflectance_valid = TRUE;
            }
            else if (band_attributes->sensor_id == IAS_TIRS)
            {
                bmd[count].spectral_radiance_scaling_gain = scaling_multiplier;
                bmd[count].spectral_radiance_scaling_offset = scaling_offset;
            }
            else
            {
                /* fix so that it is not valid */
                bmd[count].radiance_valid = FALSE;
            }

            /* calculate the min/max radiance values */
            bmd[count].maximum_radiance = bmd[count].maximum_pixel_value
                * bmd[count].spectral_radiance_scaling_gain
                + bmd[count].spectral_radiance_scaling_offset;
            bmd[count].minimum_radiance = bmd[count].minimum_pixel_value
                * bmd[count].spectral_radiance_scaling_gain
                + bmd[count].spectral_radiance_scaling_offset;
        }

        /* Get and set the sensor name */
        sensor_name = ias_sat_attr_get_sensor_name(band_attributes->sensor_id);
        strncpy(bmd[count].instrument_source, sensor_name,
                sizeof(bmd[count].instrument_source));
        bmd[count].instrument_source[sizeof(bmd[count].instrument_source) - 1]
            = '\0';
    }

    /* write the metadata */
    if ((ias_l1g_set_file_metadata(l1g_img, &fmd) != SUCCESS) ||
        (ias_l1g_set_band_metadata(l1g_img, bmd, bands_to_output) != SUCCESS))
    {
        IAS_LOG_ERROR("Error updating metadata for L1G image");
        free(bmd);
        return ERROR;
    }
    free(bmd);

    /* return that the metadata was successfully written */
    return SUCCESS;
}

