/******************************************************************************
NAME: write_envi_hdr.c 

PURPOSE: Generate an Envi header file to allow easy loading of the file into
    ENVI during development.  Since ENVI doesn't yet support HDF5 version 1.8,
    this is only useful after dumping an HDF5 band to a flat file and copying
    the header to the correct file name.

RETURNS:
    SUCCESS or ERROR

NOTES:
    To dump a flat file that can be used with the header, h5dump can be used:
       h5dump -b LE -d </band name> -o <flat file name> <hdf5 filename>

******************************************************************************/
#include "ias_types.h"          /* IAS_DATA_TYPE definition */
#include "ias_const.h"          /* SUCCESS/ERROR definitions */
#include "ias_logging.h"        /* logging library */
#include "ias_l1g.h"           
#include "ias_geo.h"            /* IAS GEO library, for IAS_PROJECTION */
#include "ias_miscellaneous.h"  /* IAS miscellaneous library, for 
                                   ias_misc_write_envi_hdr call */
#include "write_envi_hdr.h"

int write_envi_hdr
( 
    OUTPUT_BUFFER *output_image_ptr /* I: pointer to the output image */
)
{
    L1GIO *l1g_image = NULL;
    IAS_L1G_FILE_METADATA file_metadata;
    IAS_L1G_BAND_METADATA band_metadata;
    int band_count;
    IAS_DATA_TYPE data_type;
    int line_count;
    int sample_count;
    int sca_count;
    int band_number;
    int band_index;
    IAS_PROJECTION projection_information;

    /* Retrieve the pointer to the L1GIO */
    l1g_image = get_l1gio_ptr(output_image_ptr);

    /* get the count of the bands in the image */
    band_count = ias_l1g_get_band_count(l1g_image);
    if ((band_count == ERROR) || (band_count < 1))
    {
        IAS_LOG_ERROR("Failed to retrieve the L1G band count");
        return ERROR;
    }

    /* find the first available band (likely a multispectral band) */
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Converting band index %d to band number", 
                band_index);
            return ERROR;
        }
        if (ias_l1g_is_band_present(l1g_image, band_number)) 
            break;
    }
    
    if (band_index >= IAS_MAX_NBANDS)
    {
        IAS_LOG_ERROR("No bands found in L1G file");
        return ERROR;
    }

    /* get the size of the band found */
    if (ias_l1g_get_band_size(l1g_image, band_number, &data_type, &sca_count,
            &line_count, &sample_count) != SUCCESS)
    {
        IAS_LOG_ERROR("Error retrieving the band size for band number %d",
                      band_number);
        return ERROR;
    }

    /* get the file metadata */
    if (ias_l1g_get_file_metadata(l1g_image, &file_metadata) != SUCCESS)
    {
        IAS_LOG_ERROR("Failed retrieving the L1G file metadata");
        return ERROR;
    }

    /* get the band metadata */
    if (ias_l1g_get_band_metadata(l1g_image, band_number, &band_metadata)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Failed retrieving L1G band metadata for band number %d",
                      band_number);
        return ERROR;
    }

    /* set up subset of projection information needed to write the ENVI file */
    projection_information.proj_code = file_metadata.projection_code;
    projection_information.zone = file_metadata.zone_code;

    /* write the ENVI header file */
    if (ias_misc_write_envi_header("l1g", &projection_information,
            "Image file", line_count, sample_count, band_count,  
            band_metadata.upper_left_x, band_metadata.upper_left_y, 
            band_metadata.projection_distance_x,
            band_metadata.projection_distance_y, NULL, data_type) != SUCCESS)
    {
        IAS_LOG_ERROR("Failed to write ENVI header file");
        return ERROR;
    }

    return SUCCESS;
}

