/*****************************************************************************
NAME:  filter_outliers.c

PURPOSE:  This routine filters the outliers from the array of input GCP data
          using the input tolerance value for the confidence level.

RETURN VALUE:  Type = void

NOTES:
    This routine finds the x and y GCP offsets that are the furthest from
    their mean offset.  The offset (either x or y) that is furthest from their
    mean offset is the point that is checked to be an outlier.  The process
    continues until no more outliers are found, or there are less than 2
    non-outlier points left.
*****************************************************************************/
#include <math.h>
#include "ias_math.h"
#include "ias_logging.h"  /* IAS logging library */
#include "tdist.h"

void filter_outliers 
(
    double tolerance,     /* I: Tolerance value for the t-dist */
    int sca,              /* I: SCA number */
    int num_gcp_results,  /* I: Number of IAS_GCP_RESULTS structs in the
                                array */
    IAS_GCP_RESULTS *gcp_results /* I/O: Array of IAS_GCP_RESULTS structures */
)
{
    int npts;            /* Number of non-outlier points */
    int rejected_pt;     /* Flag to specify if any points were rejected */
    int max_x_index = 0; /* Index of GCP with max x distance from mean */
    int max_y_index = 0; /* Index of GCP with max y distance from mean */
    int throw_index;     /* Index of GCP to be thrown */
    int i;               /* Loop variable */
    double sum_x;        /* Sum of the x non-outliers */
    double sum_y;        /* Sum of the y non-outliers */
    double sum_sqr_x;    /* Sum of the square of the x non-outliers */
    double sum_sqr_y;    /* Sum of the square of the y non-outliers */
    double mean_x;       /* Mean of the x non-outliers */
    double mean_y;       /* Mean of the y non-outliers */
    double stddev_x;     /* Standard deviation of the x non-outliers */
    double stddev_y;     /* Standard deviation of the y non-outliers */
    double delta_x;      /* Cutoff value for the x GCP points */
    double delta_y;      /* Cutoff value for the y GCP points */
    double max_x;        /* Maximum x distance from mean */
    double max_y;        /* Maximum y distance from mean */
    double err_x;        /* Error between the max_x and cutoff value */
    double err_y;        /* Error between the max_y and cutoff value */
    double thresh;       /* Threshold value for the t-dist */
    int degrees_of_freedom; /* Number of non-outlier points - 1 */

    /* Output message to specify which band combination is being filtered */
    IAS_LOG_INFO("Filtering outliers for search SCA %d reference band %d and "
                 "search band %d.", sca, gcp_results[0].reference_band, 
                 gcp_results[0].search_band);

    /* Initialize the variables */
    rejected_pt = TRUE;
    npts = 0;
    sum_x = 0.0;
    sum_y = 0.0;
    sum_sqr_x = 0.0;
    sum_sqr_y = 0.0;

    /* Find the number of non-outliers.  Also calculate the sums and sum of 
       the squares of these non-outliers */
    for (i = 0; i < num_gcp_results; i++)
    {
        if (gcp_results[i].accept_flag)
        {
            npts++;
            sum_x += gcp_results[i].fit_sample_offset;
            sum_y += gcp_results[i].fit_line_offset;
            sum_sqr_x += gcp_results[i].fit_sample_offset 
                * gcp_results[i].fit_sample_offset;
            sum_sqr_y += gcp_results[i].fit_line_offset
                * gcp_results[i].fit_line_offset;
        }
    }

    /* As long as there are at least 2 points and the last iteration produced
       a rejected point, look for another outlier */
    while (npts >= 2 && rejected_pt)
    {
        /* Clear the rejected flag */
        rejected_pt = FALSE;

        /* Initialize degrees of freedom */
        degrees_of_freedom = npts - 1;

        /* Compute the standard deviation of the non-outliers */
        mean_x = sum_x / npts;
        mean_y = sum_y / npts;
        stddev_x = sqrt ((sum_sqr_x - sum_x * mean_x) / degrees_of_freedom);
        stddev_y = sqrt ((sum_sqr_y - sum_y * mean_y) / degrees_of_freedom);

        /* Compute the threshold value using the tolerance value and degrees of
           freedom */
        thresh = ias_math_compute_t_confidence(tolerance, degrees_of_freedom);

        /* Compute the cutoff values from the threshold */
        delta_x = thresh * stddev_x;
        delta_y = thresh * stddev_y;

        /* Find the points with the largest difference from the mean */
        max_x = 0.0;
        max_y = 0.0;

        for (i = 0; i < num_gcp_results; i++)
        {
            /* Only consider the points that are not outliers */
            if (gcp_results[i].accept_flag)
            {
                if (fabs(gcp_results[i].fit_sample_offset - mean_x) > max_x)
                {
                    max_x = fabs(gcp_results[i].fit_sample_offset - mean_x);
                    max_x_index = i;
                }

                if (fabs(gcp_results[i].fit_line_offset - mean_y) > max_y)
                {
                    max_y = fabs(gcp_results[i].fit_line_offset - mean_y);
                    max_y_index = i;
                }
            }
        }

        /* Reject the point that is furthest from its cutoff value */
        err_x = max_x - delta_x;
        err_y = max_y - delta_y;
        if (stddev_x != 0.0 && stddev_y != 0.0)
        {
            if (err_x > 0.0 || err_y > 0.0)
            {
                /* Throw the point with the largest error */
                if (err_x / stddev_x > err_y / stddev_y)
                    throw_index = max_x_index;
                else
                    throw_index = max_y_index;

                /* Set the rejected flag and the throwaway's accept flag */
                rejected_pt = TRUE;
                gcp_results[throw_index].accept_flag = 0;

                /* Remove the offset values from the non-outlier statistics */
                npts--;
                sum_x -= gcp_results[throw_index].fit_sample_offset;
                sum_y -= gcp_results[throw_index].fit_line_offset;
                sum_sqr_x -= gcp_results[throw_index].fit_sample_offset 
                    * gcp_results[throw_index].fit_sample_offset;
                sum_sqr_y -= gcp_results[throw_index].fit_line_offset
                    * gcp_results[throw_index].fit_line_offset;
            }
        }
    }
}
