/*****************************************************************************
NAME:  get_gcpdata.c

PURPOSE:  This routine reads the specified GCP results file.  The information
          will be returned in a IAS_GCP_RESULTS structure.

RETURN VALUE:  Type = int
    Value    Description
    -------  ---------------------
    SUCCESS  Successful completion
    ERROR    Operation failed

NOTES:
    This routine was written to read the GCP results files which are output
    from band-to-band and image-to-image characterization.  It may be used for 
    other GCP results files as long as the file format remains the same.

    The caller is responsible for freeing gcp_results and corr_gcps.
*****************************************************************************/
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "ias_logging.h"      /* IAS logging library */
#include "tdist.h"

#define GCP_DATA_STRLEN (2 * (PATH_MAX + 80)) /* Leave space for 2 file paths
                                                 and numerous other fields */

int get_gcp_data 
(
    const char *gcp_results_file, /* I: Name of the GCP results file */
    char report_name[REPORT_NAME_SIZE], /* O: Name of the report */
    char ref_img[PATH_MAX],   /* O: Reference image name */
    char srch_img[PATH_MAX],  /* O: Search image name */
    int *nref_bands,          /* O: Number of reference bands */
    int *ref_bands,           /* O: Array of reference band numbers */
    int *nsrch_bands,         /* O: Number of search bands */
    int *srch_bands,          /* O: Array of search band numbers */
    int *nband_combos,        /* O: Number of GCP band combinations */
    int *num_gcp_results,     /* O: Number of IAS_GCP_RESULTS structs in the
                                    array */
    IAS_GCP_RESULTS **gcp_results, /* O: Address of an array of 
                                         IAS_GCP_RESULTS structures */
    int **corr_gcps           /* O: Array of the number of correlated GCPs */
)
{
    FILE *fptr;             /* File pointer for the GCP data file */
    char *line;             /* Line read from the file */
    char *lineptr;          /* Pointer to the line */
    char word[GCP_DATA_STRLEN];  /* Word read from the file */
    char buffer[GCP_DATA_STRLEN]; /* Buffer for reading the file */
    int i;                  /* Loop variable */
    int curr_combo;         /* Current band combination */
    int num_nonoutliers;    /* Number of nonoutliers in the current band 
                               combination */
    long gcp_results_offset;/* Relative offset of the GCP results in the 
                               file */
    int length;             /* Length of string */
    int rs;                 /* Return status */

    /* Initialize the number of IAS_GCP_RESULTS structures, reference bands, 
       and search bands to 0 */
    *num_gcp_results = 0;
    *nref_bands = 0;
    *nsrch_bands = 0;
    *nband_combos = 0;

    /* Open the GCP data file */
    if ((fptr = fopen(gcp_results_file, "rt")) == NULL)
    {
        IAS_LOG_ERROR("Opening GCP data file %s.", gcp_results_file);
        return ERROR;
    }

    /* Read the first line and disregard it. This is the
       "date <processing center> time" line. */
    if (fgets(buffer, GCP_DATA_STRLEN, fptr) == NULL)
    {
        IAS_LOG_ERROR("Reading report header while retrieving GCP data.");
        fclose(fptr);
        return ERROR;
    }

    /* Read the name of the report from the header. This will be the name of 
       the report in the output header */
    if (fscanf(fptr, "%s", word) == EOF)
    {
        IAS_LOG_ERROR("Reading report header while retrieving GCP data.");
        fclose(fptr);
        return ERROR;
    };
    strncpy(report_name, word, REPORT_NAME_SIZE);
    if (fscanf(fptr, "%s", word) == EOF)
    {
        IAS_LOG_ERROR("Reading report header while retrieving GCP data.");
        fclose(fptr);
        return ERROR;
    };
    while (strcmp(word, "Report"))
    {
        strcat(report_name, " ");
        strcat(report_name, word);
        if (fscanf(fptr, "%s", word) == EOF)
        {
            IAS_LOG_ERROR("Reading report header while retrieving GCP data.");
            fclose(fptr);
            return ERROR;
        };
    }

    /* Check the size of the report name */
    length = strlen(report_name);
    if (length > REPORT_NAME_SIZE)
    {
        IAS_LOG_ERROR("Reading report header while retrieving GCP data. The "
            "report name should be <= %d characters, but it is %d characters",
            REPORT_NAME_SIZE, length);
        fclose(fptr);
        return ERROR;
    }

    /* Clear the rest of the line */
    if (fgets(buffer, GCP_DATA_STRLEN, fptr) == NULL)
    {
        IAS_LOG_ERROR("Reading report header while retrieving GCP data.");
        fclose(fptr);
        return ERROR;
    }

    /* Find the line that starts off with "Reference image".  This is the 
       beginning of the line that contains the reference image */
    line = fgets(buffer, GCP_DATA_STRLEN, fptr);
    if (line == NULL)
    {
        IAS_LOG_ERROR("Reading report header while retrieving GCP data.");
        fclose(fptr);
        return ERROR;
    }

    while (strlen(line) < 15 || strncmp(line, "Reference image", 15))
    {
        line = fgets(buffer, GCP_DATA_STRLEN, fptr);
        if (line == NULL)
        {
            IAS_LOG_ERROR("Reading report header while retrieving GCP data.");
            fclose(fptr);
            return ERROR;
        }
    }

    /* Read the reference image name */ 
    if (sscanf(line, "%s %s %s", word, word, ref_img) != 3)
    {
        IAS_LOG_ERROR("Reading reference image while retrieving GCP data.");
        fclose(fptr);
        return ERROR;
    }


    /* Check the size of the reference image name */ 
    length = strlen(ref_img);
    if (length > PATH_MAX)
    {
        IAS_LOG_ERROR("Reading reference image name while retrieving GCP "
            "data. The reference image name should be <= %d characters, but "
            "it is %d characters.", PATH_MAX, length);
        fclose(fptr);
        return ERROR;
    }

    /* If this is an I2I_CHAR report, then read the search image line */ 
    if (!strcmp(report_name, "I2I_CHAR"))
    {
        /* Read the search image name */
        if (fscanf(fptr, "%s %s %s", word, word, srch_img) == EOF)
        {
            IAS_LOG_ERROR("Reading image names while retrieving GCP data.");
            fclose(fptr);
            return ERROR;
        };

        /* Check the size of the reference image name */ 
        length = strlen(srch_img);
        if (length > PATH_MAX)
        {
            IAS_LOG_ERROR("Reading search image name while retrieving GCP "
                "data. The search image name should be <= %d characters, but "
                "it is %d characters.", PATH_MAX, length);
            fclose(fptr);
            return ERROR;
        }

        /* Clear the rest of the line */
        if (fgets(buffer, GCP_DATA_STRLEN, fptr) == NULL)
        {
            IAS_LOG_ERROR("Reading search image name while retrieving GCP "
                "data.");
            fclose(fptr);
            return ERROR;
        }
    }

    /* Read the reference bands line */
    line = fgets(buffer, GCP_DATA_STRLEN, fptr);
    if (line == NULL)
    {
        IAS_LOG_ERROR("Reading references bands line while retrieving GCP "
            "data.");
        fclose(fptr);
        return ERROR;
    }

    /* Find the position of the first band number.  We are assuming there is
       at least one band number listed */
    i = 0;
    length = strlen(line);
    while (i < length && !isdigit(line[i++]));

    /* Read the reference band numbers */
    lineptr = &line[i - 1]; /* i was incremented when digit was found */
    while (i < length && (sscanf(lineptr, "%d", 
        &ref_bands[(*nref_bands)++]) == 1))
    {
        /* Move until we are after the digit */
        while (i < length && isdigit(line[i++]));

        /* Find the next digit */
        while (i < length && !isdigit(line[i++]));

        lineptr = &line[i - 1];
        /* i was incremented when the digit was found */
    }

    /* Read the search bands line */
    line = fgets(buffer, GCP_DATA_STRLEN, fptr);
    if (line == NULL)
    {
        IAS_LOG_ERROR("Reading search bands line while retrieving GCP data.");
        fclose(fptr);
        return ERROR;
    }

    /* Find the position of the first band number.  We are assuming there is
       at least one band number listed */
    i = 0;
    length = strlen(line);
    while (i < length && !isdigit(line[i++]));

    /* Read the search band numbers */
    lineptr = &line[i - 1]; /* i was incremented when digit was found */
    while (i < length && (sscanf(lineptr, "%d", 
        &srch_bands[(*nsrch_bands)++]) == 1))
    {
        /* Move until we are after the digit */
        while (i < length && isdigit(line[i++]));

        /* Find the next digit */
        while (i < length && !isdigit(line[i++]));

        lineptr = &line[i - 1];
        /* i was incremented when the digit was found */
    }

    /* Find the line that starts off with Point.  This is the label line for
       the GCP data lines */
    line = fgets(buffer, GCP_DATA_STRLEN, fptr);
    if (line == NULL)
    {
        IAS_LOG_ERROR("Finding label line while retrieving GCP data.");
        fclose(fptr);
        return ERROR;
    }

    while (strlen(line) < 5 || strncmp(line, "Point", 5))
    {
        line = fgets(buffer, GCP_DATA_STRLEN, fptr);
        if (line == NULL)
        {
            IAS_LOG_ERROR("Finding label line while retrieving GCP data.");
            fclose(fptr);
            return ERROR;
        }
    }

    /* Get the current offset in the file so that we may return to this
       offset */
    gcp_results_offset = ftell(fptr);

    /* Count the number of lines left in the file */
    while (fgets(buffer, GCP_DATA_STRLEN, fptr) != NULL)
        (*num_gcp_results)++;

    /* Return to the offset of the GCP data */
    if (fseek(fptr, gcp_results_offset, SEEK_SET))
    {
        IAS_LOG_ERROR("Retrieving GCP data.");
        fclose(fptr);
        return ERROR;
    }

    /* Allocate space for the GCP data structures.  There is one structure for
       each GCP data line in the file */
    *gcp_results = calloc(*num_gcp_results, sizeof(IAS_GCP_RESULTS));
    if (*gcp_results == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for the GCP data.");
        fclose(fptr);
        return ERROR;
    }

    /* Read the GCP data. There are 18 elements that are to be read per line */
    for (i = 0; i < *num_gcp_results; i++)
    {
        rs = fscanf(fptr, "%s %lf %lf %lf %lf %lf %lf %lf %lf %lf %d %lf %d "
                 "%d %d %d %s %s", (*gcp_results)[i].point_id, 
                 &(*gcp_results)[i].reference_line, 
                 &(*gcp_results)[i].reference_sample, 
                 &(*gcp_results)[i].latitude, &(*gcp_results)[i].longitude, 
                 &(*gcp_results)[i].elevation, 
                 &(*gcp_results)[i].predicted_search_line, 
                 &(*gcp_results)[i].predicted_search_sample, 
                 &(*gcp_results)[i].fit_line_offset, 
                 &(*gcp_results)[i].fit_sample_offset, 
                 &(*gcp_results)[i].accept_flag, 
                 &(*gcp_results)[i].correlation_coefficient, 
                 &(*gcp_results)[i].reference_band, 
                 &(*gcp_results)[i].search_band, 
                 &(*gcp_results)[i].reference_sca, 
                 &(*gcp_results)[i].search_sca, 
                 (*gcp_results)[i].reference_name, 
                 (*gcp_results)[i].search_name);
        if (rs != 18)
        {
            if (i == 0)
            {
                IAS_LOG_ERROR("Not a complete set of GCP data for first "
                    "point");
            }
            else
            {
                IAS_LOG_ERROR("Not a complete set of GCP data for point after "
                    " point ID %s", (*gcp_results)[i - 1].point_id);
            }
            free(*gcp_results);
            *gcp_results = NULL;
            fclose(fptr);
            return ERROR;
        }

        /* Determine the number of band combinations.  Thus if this reference 
           band and search band is the same as the last, then it is not a new 
           combination.  Don't look at this for the first GCP */
        if (i != 0 && ((*gcp_results)[i].reference_band != 
                       (*gcp_results)[i - 1].reference_band
                    || (*gcp_results)[i].search_band != 
                       (*gcp_results)[i - 1].search_band
                    || (*gcp_results)[i].reference_sca != 
                       (*gcp_results)[i - 1].reference_sca))
            (*nband_combos)++;
    }
    (*nband_combos)++; /* Started counting at 0 */

    /* Allocate space for the array which contains the number of correlated 
       GCPs */
    *corr_gcps = calloc(*nband_combos, sizeof(int));
    if (*corr_gcps == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for the correlated GCPs.");
        free(*gcp_results);
        *gcp_results = NULL;
        fclose(fptr);
        return ERROR;
    }

    /* Determine the number of correlated points for each different band
       combination */
    num_nonoutliers = 0;
    curr_combo = 0;
    for (i = 0; i < *num_gcp_results; i++)
    {
        /* Check if this point is a nonoutlier */
        if ((*gcp_results)[i].accept_flag)
            num_nonoutliers++;

        /* Determine the number of band combinations.  Thus if this reference 
           band and search band is the same as the last, then it is not a new 
           combination.  Don't look at this for the first GCP */
        if (i != 0 && ((*gcp_results)[i].reference_band != 
                       (*gcp_results)[i - 1].reference_band 
                    || (*gcp_results)[i].search_band != 
                       (*gcp_results)[i - 1].search_band
                    || (*gcp_results)[i].reference_sca != 
                       (*gcp_results)[i - 1].reference_sca))
        {
            (*corr_gcps)[curr_combo] = num_nonoutliers;
            num_nonoutliers = 0; /* Reset the number of nonoutliers */
            curr_combo++;
        }
    }

    /* Fill in the number of correlated points for the last band combo */
    (*corr_gcps)[curr_combo] = num_nonoutliers;

    rs = fclose(fptr);
    if (rs != 0)
    {
        IAS_LOG_ERROR("Closing input GCP report file.");
        free(*gcp_results);
        *gcp_results = NULL;
        free(*corr_gcps);
        *corr_gcps = NULL;
        return ERROR;
    }

    return SUCCESS;
}
