/*****************************************************************************
NAME:  GETPAR

PURPOSE:  Get the input parameter from the TDIST application parameter file. 

RETURN VALUE:  Type = int
    Value    Description
    -------  ---------------------
    SUCCESS  Successful completion 
    ERROR    Operation failed 
*******************************************************************************/
#include <stdlib.h>
#include <stddef.h>
#include <string.h>

#include "ias_parm.h"                     /* parameter library */
#include "ias_parm_standard_parameters.h" /* parameter table standard fields */
#include "tdist.h"

/* defaults */
#define DEFAULT_TOLERANCE 0.95

/* invalid values */
#define INVALID_TOLERANCE -1.0

int get_parameters
(
    const char parm_file[PATH_MAX],     /* I: WO parameter file */
    char gcp_results_filename[PATH_MAX], /* O: Input residuals file made by an
                                              algorithm like B2B or I2I */
    char *cpf_filename,                 /* O: CPF file */
    double *tolerance,                  /* O: threshold tolerance */
    char statistics_filename[PATH_MAX], /* O: Statistics file */
    int *statistics_flag,               /* O: Flag to create statistics file */ 
    char residual_filename[PATH_MAX],   /* O: Residuals file */
    int *residual_flag,                 /* O: Flag to create residuals file  */ 
    char l0r_filename[PATH_MAX],        /* O: L0R file */
    char l1g_filename[PATH_MAX],        /* O: L1G file */
    struct main_db_save_info *main_db_save_info, /* O: Structure containing 
                                              information to save to the DB */
    int *store_in_database_flag,        /* O: Flag to indicate if a DB save 
                                              is performed (1= yes, 0 = no) */
    IAS_SATELLITE_ID *satellite_id,     /* O: Satellite ID */
    char omf_file_name[PATH_MAX],       /* O: omf file name */
    char input_gcp_source[4]            /* O: input gcp source either B2B or
                                              I2I */
)
{
    int count;                          /* Number of elements */
    int status;                         /* Status returned from functions */
    int processing_pass;                /* Processing flow pass */

    /* Various file name storage - local storage is used because sizeof doesn't
       work on function parameter arrays */
    char local_cpf_filename[PATH_MAX];
    char local_wo_id[IAS_WORKORDER_ID_SIZE];
    char local_residual_filename[PATH_MAX];
    char local_b2b_char_output_filename[PATH_MAX];
    char local_i2i_char_output_filename[PATH_MAX];
    char local_statistics_filename[PATH_MAX];
    char local_l0r_filename[PATH_MAX];
    char local_l1g_filename[PATH_MAX];
    char local_gcp_source[4]; /* should be either B2B or I2I */
    int satellite_number; /* Satellite number being processed */

    /* Start parameter table definition */
    IAS_PARM_DECLARE_TABLE(parms, 7);

    /* Get the work order ID */
    IAS_PARM_WORK_ORDER_ID(parms, local_wo_id, sizeof(local_wo_id), 0);  

    /* Get the flag to write (or not write) data to the database */
    IAS_PARM_STORE_CHARACTERIZATION_IN_DB(parms, IAS_PARM_OPTIONAL,
        store_in_database_flag, sizeof(*store_in_database_flag), 0);

    /* Get the processing pass */
    IAS_PARM_PROCESSING_PASS(parms, &processing_pass, sizeof(processing_pass),
        0);

    /* Get the confidence level */
    *tolerance = INVALID_TOLERANCE;
    double default_tolerance[] = {DEFAULT_TOLERANCE};
    IAS_PARM_ADD_DOUBLE(parms, OUTLIER_TOLERANCE, "Outlier threshold",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 0.0, 1.0, 1, 
        default_tolerance, tolerance, sizeof(*tolerance), 0);

    /* Get the type of input file */
    const char *valid_gcp_sources[] = {"B2B","I2I"};
    IAS_PARM_ADD_STRING(parms, INPUT_GCP_SOURCE, "Used to define which "
        "parameter to read from the OMF file. (B2B) retrieves "
        "B2B_CHAR_OUTPUT_FILENAME. (I2I) retrieves I2I_CHAR_OUTPUT_FILENAME"
        ,IAS_PARM_REQUIRED, 2, valid_gcp_sources, 0, NULL, local_gcp_source,
        sizeof(local_gcp_source), 1);

    /* Get the name of output residual filename */
    IAS_PARM_ADD_STRING(parms, RESIDUAL_FILENAME, "Output control point "
        "residual file name", IAS_PARM_OPTIONAL, 0, NULL, 0, NULL,
        local_residual_filename, sizeof(local_residual_filename), 1);

    /* Get the name of output statistics filename */
    IAS_PARM_ADD_STRING(parms, STATISTICS_FILENAME, 
        "Correlation statistics file name",
        IAS_PARM_OPTIONAL, 0, NULL, 0, NULL, local_statistics_filename, 
        sizeof(local_statistics_filename), 1);

    /* Calculate the number of items in the parameter file */
    count = IAS_PARM_GET_TABLE_SIZE(parms);

    /* Call the ias_parm_provide_help routine to provide help for the
       parameters if requested; also verifies the ODL file exists */
    status = ias_parm_provide_help(parm_file, parms, count, 
        IAS_INPUT_PARAMETERS);
    if (status == 1)
    {
        /* If the status of the provide help is 1 the user asked for help.
           Since the parameter file name contains the help request, copy it to
           the OMF file name so the next call will ask for the same help */
        strncpy(omf_file_name, parm_file, PATH_MAX);
    }
    else if (status == ERROR) /* Help failed */
    {
        IAS_LOG_ERROR("Providing help using application parameter file %s",
            parm_file);
        return ERROR;
    }
    else
    {
        /* Read the parameters and validate them as needed */
        if (ias_parm_read(parm_file, parms, count) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading parameters from: %s", parm_file);
            return ERROR;
        }

        /* Transfer the various local arrays to the function parameters */
        strncpy(main_db_save_info->wo_id, local_wo_id, IAS_WORKORDER_ID_SIZE);
        strncpy(residual_filename, local_residual_filename, PATH_MAX);
        strncpy(statistics_filename, local_statistics_filename, PATH_MAX);

        
        if (IAS_PARM_GET_COUNT_READ(parms, RESIDUAL_FILENAME) == 0)
        {
            *residual_flag = 0;
        }
        else
        {
            *residual_flag = 1;
        }

        if (IAS_PARM_GET_COUNT_READ(parms, STATISTICS_FILENAME) == 0)
        {
            *statistics_flag = 0;
        }
        else
        {
            *statistics_flag = 1;
        }

        /* Build the OMF file name */
        sprintf(omf_file_name, "%s.omf", local_wo_id);
    }

    /* Now read and check the OMF */

    /* Start parameter table definition */
    IAS_PARM_DECLARE_TABLE(omf_parms, 6);

    /* Get the name of input GCP results filename */
    IAS_PARM_B2B_CHAR_OUTPUT_FILENAME(omf_parms, local_b2b_char_output_filename,
        sizeof(local_b2b_char_output_filename),1);

    IAS_PARM_I2I_CHAR_OUTPUT_FILENAME(omf_parms, local_i2i_char_output_filename,
        sizeof(local_i2i_char_output_filename),1);

    /* Get the name of the Calibration parameter file */
    IAS_PARM_CPF(omf_parms, IAS_PARM_REQUIRED, local_cpf_filename, 
        sizeof(local_cpf_filename), 1);

    /* Get the L0R file name */
    IAS_PARM_L0R_FILENAME(omf_parms, IAS_PARM_OPTIONAL, local_l0r_filename,
        sizeof(local_l0r_filename), 1);

    /* Get the L1G file name */
    IAS_PARM_L1G_FILENAME_FROM_PASS(omf_parms, IAS_PARM_REQUIRED,
            local_l1g_filename, sizeof(local_l1g_filename), 1, processing_pass);

    IAS_PARM_SATELLITE(omf_parms, &satellite_number, sizeof(satellite_number),
        1); 

    /* See if the user wanted help or a template with the OMF file */
    status = ias_parm_provide_help(omf_file_name, omf_parms,
        IAS_PARM_GET_TABLE_SIZE(omf_parms), IAS_OMF_PARAMETERS);
    if (status)
    {
        if (status == 1) /* Help was successful */
        {
            exit(EXIT_SUCCESS);
        }
        else if (status == ERROR) /* Help failed */
        {
            IAS_LOG_ERROR("Providing help using OMF %s", omf_file_name);
            return ERROR;
        }
    }

    /* Read the OMF file */
    count = IAS_PARM_GET_TABLE_SIZE(omf_parms);
    if (ias_parm_read(omf_file_name, omf_parms, count) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from OMF file: %s", omf_file_name);
        return ERROR;
    }

    strncpy(l0r_filename, local_l0r_filename, PATH_MAX);
    strncpy(l1g_filename, local_l1g_filename, PATH_MAX);
    strncpy(cpf_filename, local_cpf_filename, PATH_MAX);
    strncpy(input_gcp_source, local_gcp_source,4);

    if (strcmp(local_gcp_source, "B2B") == 0)
    {
        if(IAS_PARM_GET_COUNT_READ(omf_parms, B2B_CHAR_OUTPUT_FILENAME) == 0)
        {
            IAS_LOG_ERROR("No B2B_CHAR_OUTPUT_FILENAME provided in the"
                          " parameter file."); 
            return ERROR;
        }
        strncpy(gcp_results_filename, local_b2b_char_output_filename, PATH_MAX);
    }
    else
    {
        if(IAS_PARM_GET_COUNT_READ(omf_parms, I2I_CHAR_OUTPUT_FILENAME) == 0)
        {
            IAS_LOG_ERROR("No I2I_CHAR_OUTPUT_FILENAME provided in the"
                          " parameter file."); 
            return ERROR;
        }
        strncpy(gcp_results_filename, local_i2i_char_output_filename, PATH_MAX);
    }

    /* Exit if the CPF filename is empty */
    if (strcmp(cpf_filename, "") == 0)
    {
        IAS_LOG_ERROR("No CPF filename provided in the parameter file.");
        return ERROR;
    }

    /* Look up the satellite ID */
    *satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(
        satellite_number);
    if (*satellite_id == ERROR)
    {
        IAS_LOG_ERROR("Unable to determine satellite ID from satellite "
            "number %d", satellite_number);
        return ERROR;
    }

    return SUCCESS;
}
