/*****************************************************************************
NAME:  put_gcpstats.c

PURPOSE:  This routine will calculate and output the GCP statistics to the
          specified file.

RETURN VALUE:  Type = int
    Value    Description
    -------  ---------------------
    SUCCESS  Successful completion
    ERROR    Operation failed
*****************************************************************************/
#include <stdlib.h>
#include <string.h>

#include "ias_l1g.h"          /* L1G I/O libarary */
#include "ias_miscellaneous.h" /* Miscellaneous IAS libraries */
#include "ias_logging.h"      /* IAS logging library */
#include "ias_cpf.h"          /* IAS CPF library */
#include "save_to_database.h" /* Set up database inserts */
#include "tdist.h"

int put_gcp_stats 
(
    const char *out_gcp_stats_file, /* I: Name of the GCP statistics output
                                    file */ 
    const char *report_name,  /* I: Name of the report */
    const char *l1g_file,     /* I: L1G filename for report header data */
    const char *ref_l0r,      /* I: Name of the reference L0R image */
    const char *ref_img,      /* I: Reference image name */
    const char *srch_img,     /* I: Search image name */
    const char *cpf_filename, /* I: CPF filename */
    const int *ncorr_gcps,    /* I: Array of the number of correlated GCPs for
                                    each band combination */
    double threshold,         /* I: Threshold of the t-dist routine */
    int nref_bands,           /* I: Number of reference bands */
    const int *ref_bands,     /* I: Array of reference band numbers */
    int nsrch_bands,          /* I: Number of search bands */
    const int *srch_bands,    /* I: Array of search band numbers */
    int num_gcp_results,      /* I: Number of IAS_GCP_RESULTS structs in the
                                    array */
    const IAS_GCP_RESULTS *gcp_results, /* I: Array of IAS_GCP_RESULTS 
                                    structures */
    char *char_type,          /* I: "I2I" or "B2B" to tell the routine if this 
                                    is image to image characterization or band
                                    to band characterization since tdist is 
                                    used in both cases */
    int *save_to_db_flag,     /* I/O: Flag used to indicate whether results 
                                      will be saved to the DB */
    struct main_db_save_info *main_db_save_info, /* I/O: Info to save to DB */
    struct band_sca_db_save_info *band_sca_db_save_info, /* I/O: Info to save 
                                    to the database */
    int *number_db_records    /* O: The number of records to add to the array
                                    of structures to be saved to the DB */
)
{
    FILE *fptr;               /* File pointer for the GCP data file */
    int i;                    /* Loop variable */
    int j;                    /* Loop variable */
    int band_start;           /* Starting GCP of the current band combo */
    int curr_combo;           /* Current band combination */
    int curr_gcp;             /* Current GCP */
    int num_gcps;             /* Number of gcps in the current band combo */
    int num_nonoutliers;      /* Number of non-outlier GCPs */
    int band_sca_index = 0;   /* Counter tracking # of band/SCA DB records */
    double min_x;             /* Minimum x non-outlier */
    double min_y;             /* Minimum y non-outlier */
    double max_x;             /* Maximum x non-outlier */
    double max_y;             /* Maximum y non-outlier */
    double mean_x;            /* Mean of the x non-outliers */
    double mean_y;            /* Mean of the y non-outliers */
    double median_x;          /* Median of the x non-outliers */
    double median_y;          /* Median of the y non-outliers */
    double rmse_x;            /* RMSE of the x non-outliers */
    double rmse_y;            /* RMSE of the y non-outliers */
    double stddev_x;          /* Standard deviation of the x non-outliers */
    double stddev_y;          /* Standard deviation of the y non-outliers */
    double *line_res;         /* Array of valid line residuals */
    double *samp_res;         /* Array of valid sample residuals */
    int ref_band_number;      /* Reference band# to shorten code lines only */
    int srch_band_number;     /* Search band# to shorten code lines only */
    int sca_number;           /* SCA# to shorten code lines only */
    int ref_band_index;       /* Reference band index */
    int sca_index;            /* SCA index */
    double b2b_db_save_metrics_line[IAS_MAX_NBANDS][IAS_MAX_NSCAS]; /* limit
                                 saving to the DB */
    double b2b_db_save_metrics_sample[IAS_MAX_NBANDS][IAS_MAX_NSCAS]; /* limit
                                 saving to the DB */
    double i2i_db_save_metrics_line[IAS_MAX_NBANDS]; /* limit DB saves */
    double i2i_db_save_metrics_sample[IAS_MAX_NBANDS]; /* limit DB saves */
    int number_normal_bands;  /* Number of normal bands for the satellite */
    int number_bands;         /* Number of bands for the satellite */
    int band_list[IAS_MAX_NBANDS]; /* List of band numbers */
    int band_number;          /* Band number to find number of SCAs for */
    int number_scas;          /* Number of SCA for a particular band */
    IAS_CPF *cpf;             /* Calibration parameter file constants */
    IAS_REPORT_HEADER *header = NULL; /* Report header structure */
    int status;               /* Function return status */
 
    /* Open the GCP statistics file for writing */
    fptr = fopen(out_gcp_stats_file, "w");
    if (fptr == NULL)
    {
        IAS_LOG_ERROR("Opening the GCP statistics file %s.",
                out_gcp_stats_file);
        return ERROR;
    }

    /* Output the standard header to the file */
    header = ias_misc_initialize_gps_report_header(l1g_file, ref_l0r, 
        main_db_save_info->wo_id, report_name);
    if (!header)
    {
        IAS_LOG_ERROR("Writing to the GCP statistics output file.");
        fclose(fptr);
        return ERROR;
    }

    status = ias_misc_write_gps_report_header(fptr, header);
    ias_misc_free_report_header(header);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing to the GCP statistics output file.");
        fclose(fptr);
        return ERROR;
    }

    /* If this is an I2I_CHAR report, then output the reference image, search
       image, and reference bands label */
    if (!strcmp(char_type, "I2I"))
    {
        if (fprintf(fptr, "Reference image: %s\nSearch image: %s\n"
                          "Reference bands: ", ref_img, srch_img) < 0)
        {
            IAS_LOG_ERROR("Writing to the GCP statistics output file.");
            fclose(fptr);
            return ERROR;
        }
    }
    /* If this is a B2B_CHAR report, then output the reference image and
       reference band label */
    else if (!strcmp(char_type, "B2B"))
    {
        if (fprintf(fptr, "Reference image: %s\nReference bands: ",
                    ref_img) < 0)
        {
            IAS_LOG_ERROR("Writing to the GCP statistics output file.");
            fclose(fptr);
            return ERROR;
        }
    }

    /* Output the reference bands, search band label, and search bands */
    for (i = 0; i < nref_bands; i++)
    {
        if (fprintf(fptr, " %d", ref_bands[i]) < 0)
        {
            IAS_LOG_ERROR("Writing to the GCP statistics output file.");
            fclose(fptr);
            return ERROR;
        }
    }

    if (fprintf(fptr, "\nSearch bands: ") < 0)
    {
        IAS_LOG_ERROR("Writing to the GCP statistics output file.");
        fclose(fptr);
        return ERROR;
    }

    for (i = 0; i < nsrch_bands; i++)
    {
        if (fprintf (fptr, " %d", srch_bands[i]) < 0)
        {
            IAS_LOG_ERROR("Writing to the GCP statistics output file.");
            fclose(fptr);
            return ERROR;
        }
    }

    line_res = (double *)malloc(sizeof(double) * num_gcp_results);
    if (line_res == NULL)
    {
        IAS_LOG_ERROR("Writing to the GCP statistics output file."); 
        fclose(fptr);
        return ERROR;
    }

    samp_res = (double *)malloc(sizeof(double) * num_gcp_results);
    if (samp_res == NULL)
    {
        IAS_LOG_ERROR("Writing to the GCP statistics output file."); 
        fclose(fptr);
        free(line_res);
        return ERROR;
    }

    /* Output the threshold value */
    if (fprintf(fptr, "\nThreshold: %lf\n\n", threshold) < 0)
    {
        IAS_LOG_ERROR("Writing to the GCP statistics output file.");
        fclose(fptr);
        free(line_res);
        free(samp_res);
        return ERROR;
    }

    if (*save_to_db_flag)
    {
        const struct IAS_CPF_B2B_ASSESSMENT *cpf_b2b;
        const struct IAS_CPF_I2I_ASSESSMENT *cpf_i2i;

        /* Read the CPF */
        cpf = ias_cpf_read(cpf_filename);
        if (!cpf)
        {
            IAS_LOG_ERROR("Cannot open CPF file %s for reading.", cpf_filename);
            fclose(fptr);
            free(line_res);
            free(samp_res);
            return ERROR;
        }

        /* Retrieve DB save metric values from the CPF */
        if (!strcmp(char_type, "I2I"))
        {
            /* Get the I2I group needed from the CPF */
            cpf_i2i = ias_cpf_get_i2i_assessment(cpf);
            if (!cpf_i2i)
            {
                IAS_LOG_ERROR("Getting I2I Assessment group from CPF");
                ias_cpf_free(cpf);
                fclose(fptr);
                free(line_res);
                free(samp_res);
                return ERROR;
            }

            /* Populate the DB save metrics structure with values from CPF */
            for (i = 0; i < IAS_MAX_NBANDS; i++)
            {
                i2i_db_save_metrics_line[i] = cpf_i2i->trend_threshold_line[i]; 
                i2i_db_save_metrics_sample[i] 
                    = cpf_i2i->trend_threshold_sample[i]; 
            }
        }

        if (!strcmp(char_type, "B2B"))
        {
            /* Get the B2B group needed from the CPF */
            cpf_b2b = ias_cpf_get_b2b_assessment(cpf);
            if (!cpf_b2b)
            {
                IAS_LOG_ERROR("Getting B2B Assessment group from CPF");
                ias_cpf_free(cpf);
                fclose(fptr);
                free(line_res);
                free(samp_res);
                return ERROR;
            }

            /* Get total satellite band count */
            number_normal_bands = ias_sat_attr_get_normal_band_count();
            if ((number_normal_bands < 1) ||
                (number_normal_bands > IAS_MAX_NBANDS))
            {
                IAS_LOG_ERROR("Invalid number of bands %d.", 
                    number_normal_bands);
                ias_cpf_free(cpf);
                fclose(fptr);
                free(line_res);
                free(samp_res);
                return ERROR;
            }

            /* Get band information */
            if (ias_sat_attr_get_sensor_band_numbers(IAS_MAX_SENSORS,
                    IAS_NORMAL_BAND, 0, band_list, number_normal_bands,
                    &number_bands) != SUCCESS)
            {
                IAS_LOG_ERROR("Error getting band numbers");
                ias_cpf_free(cpf);
                fclose(fptr);
                free(line_res);
                free(samp_res);
                return ERROR;
            }

            /* Populate the DB save metrics structure with values from CPF */
            for (i = 0; i < number_bands; i++)
            {
                band_number = ias_sat_attr_convert_band_index_to_number(i);
                if (band_number == ERROR)
                {
                    IAS_LOG_ERROR("Invalid band number for band index %d.", i);
                    ias_cpf_free(cpf);
                    fclose(fptr);
                    free(line_res);
                    free(samp_res);
                    return ERROR;
                }

                /* Find the number of SCAs for this band */
                number_scas = ias_sat_attr_get_scas_per_band(band_number);
                if (number_scas == ERROR)
                {
                    IAS_LOG_ERROR("Invalid number of SCAs for band number %d.",
                        band_number);
                    ias_cpf_free(cpf);
                    fclose(fptr);
                    free(line_res);
                    free(samp_res);
                    return ERROR;
                }

                for (j = 0; j < number_scas; j++)
                {
                    b2b_db_save_metrics_line[i][j] 
                        = cpf_b2b->trend_threshold_line[i][j];
                    b2b_db_save_metrics_sample[i][j] 
                        = cpf_b2b->trend_threshold_sample[i][j];
                }
            }
        }

        /* Free CPF memory. */
        ias_cpf_free(cpf);
    }

    /* Start looping through the band combinations.  Start looking at the
       second GCP data point, but include the first point in the statistics.
       Loop until all GCP data points have been checked for outliers */
    curr_gcp = 0;
    curr_combo = 0;
    while (curr_gcp < num_gcp_results)
    {
        /* Print the reference and search band numbers, plus the SCA */
        if (fprintf(fptr, "SCA: %d, Reference Band: %d, Search Band: %d\n",
                gcp_results[curr_gcp].reference_sca, 
                gcp_results[curr_gcp].reference_band, 
                gcp_results[curr_gcp].search_band) < 0)
        {
            IAS_LOG_ERROR("Writing to the GCP statistics output file.");
            fclose(fptr);
            free(line_res);
            free(samp_res);
            return ERROR;
        }

        /* Initialize the statistics variables to the first point in this band
           combination, if it isn't an outlier.  Otherwise, initialize to 0 */
        if (gcp_results[curr_gcp].accept_flag)
        {
            max_x = gcp_results[curr_gcp].fit_sample_offset;
            max_y = gcp_results[curr_gcp].fit_line_offset;
            min_x = gcp_results[curr_gcp].fit_sample_offset;
            min_y = gcp_results[curr_gcp].fit_line_offset;
            num_nonoutliers = 1;

            line_res[num_nonoutliers-1] = 
                gcp_results[curr_gcp].fit_line_offset;
            samp_res[num_nonoutliers-1] = 
                gcp_results[curr_gcp].fit_sample_offset;
        }
        else
        {
            max_x = -999.0;
            max_y = -999.0;
            min_x = 999.0;
            min_y = 999.0;
            num_nonoutliers = 0;
        }

        /* While the current reference and search bands and the next reference
           and search bands are the same, then we are still in the same band
           combination.  Keep track of the number of GCPs and the sum of the
           outliers for this band combination.  Start comparing the second
           point in the band combination with the first point in the band
           combination. */

        /* Keep track of the starting GCP for this combination */
        band_start = curr_gcp;  
        curr_gcp++;             /* Skip the first point */
        num_gcps = 1;

        while (curr_gcp < num_gcp_results 
               && gcp_results[curr_gcp].reference_band == 
                  gcp_results[curr_gcp-1].reference_band 
               && gcp_results[curr_gcp].search_band == 
                  gcp_results[curr_gcp-1].search_band
               && gcp_results[curr_gcp].reference_sca == 
                  gcp_results[curr_gcp-1].reference_sca)
        {
            /* Sum the offsets for the non-outliers */
            if (gcp_results[curr_gcp].accept_flag)
            {
                /* Check if this offset is the min or max */
                if (gcp_results[curr_gcp].fit_sample_offset < min_x)
                    min_x = gcp_results[curr_gcp].fit_sample_offset;
                if (gcp_results[curr_gcp].fit_line_offset < min_y)
                    min_y = gcp_results[curr_gcp].fit_line_offset;
                if (gcp_results[curr_gcp].fit_sample_offset > max_x)
                    max_x = gcp_results[curr_gcp].fit_sample_offset;
                if (gcp_results[curr_gcp].fit_line_offset > max_y)
                    max_y = gcp_results[curr_gcp].fit_line_offset;

                line_res[num_nonoutliers] = 
                    gcp_results[curr_gcp].fit_line_offset;
                samp_res[num_nonoutliers] = 
                    gcp_results[curr_gcp].fit_sample_offset;

                num_nonoutliers++;
            }

            num_gcps++;
            curr_gcp++;
        }

        /* Compute the standard deviation of the non-outliers for the current
           band combination */
        if (num_nonoutliers > 1)
        {
            if (ias_math_compute_mean(line_res, num_nonoutliers, &mean_y) != 
                SUCCESS)
            {
                IAS_LOG_ERROR("Calculating mean of along-scan residuals.");
                fclose(fptr);
                free(line_res);
                free(samp_res);
                return ERROR;
            }

            if (ias_math_compute_mean(samp_res, num_nonoutliers, &mean_x) != 
                SUCCESS)
            {
                IAS_LOG_ERROR("Calculating mean of across-scan residuals.");
                fclose(fptr);
                free(line_res);
                free(samp_res);
                return ERROR;
            }

            if (ias_math_compute_rmse(line_res, num_nonoutliers, &rmse_y) != 
                SUCCESS)
            {
                IAS_LOG_ERROR("Calculating the along-scan RMSE.");
                fclose(fptr);
                free(line_res);
                free(samp_res);
                return ERROR;
            }

            if (ias_math_compute_rmse(samp_res, num_nonoutliers, &rmse_x) != 
                SUCCESS)
            {
                IAS_LOG_ERROR("Calculating the across-scan RMSE.");
                fclose(fptr);
                free(line_res);
                free(samp_res);
                return ERROR;
            }

            if (ias_math_compute_stdev(line_res, num_nonoutliers, mean_y, 
                &stddev_y) != SUCCESS)
            {
                IAS_LOG_ERROR("Calculating the along-scan standard deviation.");
                fclose(fptr);
                free(line_res);
                free(samp_res);
                return ERROR;
            }

            if (ias_math_compute_stdev(samp_res, num_nonoutliers, mean_x, 
                &stddev_x) != SUCCESS)
            {
                IAS_LOG_ERROR("Calculating the across-scan standard "
                    "deviation.");
                fclose(fptr);
                free(line_res);
                free(samp_res);
                return ERROR;
            }
        }
        else
        {
            mean_x = 0.0;
            mean_y = 0.0;
            rmse_x = 0.0;
            rmse_y = 0.0;
            stddev_x = 0.0;
            stddev_y = 0.0;
        }

        /* Compute the median of the GCPs */
        if (num_nonoutliers > 0)
        {
            /* Sort the arrays */
            ias_math_heapsort_double_array(num_nonoutliers, samp_res);
            ias_math_heapsort_double_array(num_nonoutliers, line_res);

            /* The median is the middle element in the sorted arrays */
            median_x = samp_res[num_nonoutliers/2];
            median_y = line_res[num_nonoutliers/2];
        }
        else
        {
            median_x = 0;
            median_y = 0;
        }

        /* Output the statistics for the current band combination */
        if (fprintf(fptr,
                   "Total GCPs: %d, Correlated GCPs: %d, Valid GCPs: %d\n"
                   "Min   (line, sample): %lf, %lf\n"
                   "Mean  (line, sample): %lf, %lf\n"
                   "Max   (line, sample): %lf, %lf\n"
                   "Median(line, sample): %lf, %lf\n"
                   "StDev (line, sample): %lf, %lf\n"
                   "RMSE  (line, sample): %lf, %lf\n\n",
                   num_gcps, ncorr_gcps[curr_combo], num_nonoutliers, min_y, 
                   min_x, mean_y, mean_x, max_y, max_x, median_y, median_x, 
                   stddev_y, stddev_x, rmse_y, rmse_x) < 0)
        {
            IAS_LOG_ERROR("Writing to the GCP statistics output file.");
            fclose(fptr);
            free(line_res);
            free(samp_res);
            return ERROR;
        }

        /* These are just to shorten some long code lines */
        ref_band_number = gcp_results[band_start].reference_band;
        srch_band_number = gcp_results[band_start].search_band;
        sca_number = gcp_results[band_start].reference_sca;

        ref_band_index = ias_sat_attr_convert_band_number_to_index(
            ref_band_number); 
        if ((ref_band_index < 0) || (ref_band_index > IAS_MAX_NBANDS - 1))
        {
            IAS_LOG_ERROR("Invalid band index %d.", ref_band_index);
            fclose(fptr);
            free(line_res);
            free(samp_res);
            return ERROR;
        }

        if (sca_number == 0)
        {
            /* If SCA-combined image */
            sca_index = 0;
        }
        else
        {
            sca_index = sca_number - 1;
        }

        if (*save_to_db_flag)
        {
            /* Check the DB save metrics.  Override the user-specified 
               save_to_db_flag if necessary */
            if (!strcmp(char_type, "I2I"))
            {
                if ((rmse_x > i2i_db_save_metrics_sample[ref_band_index]) ||
                    (rmse_y > i2i_db_save_metrics_line[ref_band_index])) 
                {
                    *save_to_db_flag = 0;
                }
            }

            if (!strcmp(char_type, "B2B"))
            {
                if ((stddev_x > 
                    b2b_db_save_metrics_sample[ref_band_index][sca_index]) ||
                    (stddev_y > 
                    b2b_db_save_metrics_line[ref_band_index][sca_index])) 
                {
                    *save_to_db_flag = 0;
                }
            }
        }

        /* If DB saving is requested, set up the band/SCA level DB structure */
        if (*save_to_db_flag)
        {
            band_sca_db_save_info->reference_band_id[band_sca_index] = 
                ref_band_number;
            band_sca_db_save_info->search_band_id[band_sca_index] = 
                srch_band_number;
            band_sca_db_save_info->sensor_chip_assembly_id[band_sca_index] = 
                sca_number;
            band_sca_db_save_info->nbr_of_pnts_corr[band_sca_index] = 
               num_gcps;
            band_sca_db_save_info->nbr_of_pnts_successfully_corr
                [band_sca_index] = ncorr_gcps[curr_combo];
            band_sca_db_save_info->nbr_of_vld_pnts[band_sca_index] = 
                num_nonoutliers;
            band_sca_db_save_info->rmse_line[band_sca_index] = rmse_y;
            band_sca_db_save_info->rmse_sample[band_sca_index] = rmse_x;
            band_sca_db_save_info->std_dev_line[band_sca_index] = stddev_y;
            band_sca_db_save_info->std_dev_sample[band_sca_index] = stddev_x;
            band_sca_db_save_info->min_line[band_sca_index] = min_y;
            band_sca_db_save_info->min_sample[band_sca_index] = min_x;
            band_sca_db_save_info->max_line[band_sca_index] = max_y;
            band_sca_db_save_info->max_sample[band_sca_index] = max_x;
            band_sca_db_save_info->med_line[band_sca_index] = median_y;
            band_sca_db_save_info->med_sample[band_sca_index] = median_x;
            band_sca_db_save_info->mean_line[band_sca_index] = mean_y;
            band_sca_db_save_info->mean_sample[band_sca_index] = mean_x;

            band_sca_index++;
        }

        curr_combo++;
    }

    if (*save_to_db_flag)
    {
        /* Set up the top-level DB structure */
        main_db_save_info->t_distribution_threshold = threshold;
        strcpy(main_db_save_info->reference_image, ref_img);

        /* Return the number of records to save to the DB */
        *number_db_records = band_sca_index;
    }

    free(samp_res);
    free(line_res);

    status = fclose(fptr);
    if (status != 0)
    {
        IAS_LOG_ERROR("Closing statistics file.");
        return ERROR;
    }

    return SUCCESS;
}
