/******************************************************************************
NAME:  put_resdata.c

PURPOSE:  This routine will output the GCP residual data to the specified file.

RETURN VALUE:  Type = int
    Value    Description
    -------  ---------------------
    SUCCESS  Successful completion
    ERROR    Operation failed

NOTES:
    This routine was written to write the GCP results based on output from
    band-to-band and image-to-image.  It may be used for other GCP results
    files as long as the file format remains the same.
******************************************************************************/
#include <string.h>

#include "ias_miscellaneous.h" /* IAS miscellaneous library calls */
#include "ias_logging.h"      /* IAS logging library */
#include "tdist.h"

int put_residual_data 
(
    const char *out_gcp_residual_data_file, /* I: Name of GCP residual data
                                    output file */
    const char *report_name,  /* I: Name of the report */
    const char *l1g_file,     /* I: L1G filename for report header data */
    const char *ref_l0r,      /* I: Name of the reference L0R image */
    const char *ref_img,      /* I: Reference image name */
    const char *srch_img,     /* I: Search image name */
    int nref_bands,           /* I: Number of reference bands */
    const int *ref_bands,     /* I: Array of reference band numbers */
    int nsrch_bands,          /* I: Number of search bands */
    const int *srch_bands,    /* I: Array of search band numbers */
    int num_gcp_results,      /* I: Number of IAS_GCP_RESULTS structs in the
                                    array */
    int nsca,                 /* I: Number of SCAs */
    int nband_combos,         /* I: Total number of band combinations */
    int num_sca_bandcombo[IAS_MAX_NSCAS][NBANDCOMBO], /* I: Number of points in
                                    band combination for an SCA */
    const IAS_GCP_RESULTS *gcp_results, /* I: Array of IAS_GCP_RESULTS 
                                    structures */
    const char *work_order_id /* I: Work order ID */
)
{
    FILE *fptr;               /* File pointer for the GCP data file */
    int i;                    /* Loop variable */
    int j;                    /* Loop variable */
    int sca_index;            /* Loop variable */
    int band_combo;           /* Loop variable */
    int band_combos_per_sca;  /* Number of band combinations per SCA */
    IAS_REPORT_HEADER *header = NULL; /* Report header structure */
    int status;               /* Function return status */

    /* Open the GCP data output file for writing */
    fptr = fopen(out_gcp_residual_data_file, "w");
    if (fptr == NULL)
    {
        IAS_LOG_ERROR("Opening the GCP residual data output file %s.", 
                 out_gcp_residual_data_file);
        return ERROR;
    }

    /* Output the standard header to the file */
    header = ias_misc_initialize_gps_report_header(l1g_file, ref_l0r,
            work_order_id, report_name);
    if (!header)
    {
        IAS_LOG_ERROR("Writing to the GCP residual data output file.");
        fclose(fptr);
        return ERROR;
    }

    status = ias_misc_write_gps_report_header(fptr, header);
    ias_misc_free_report_header(header);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing to the GCP residual data output file.");
        fclose(fptr);
        return ERROR;
    }

    fprintf(fptr, "%d Number of records\n", num_gcp_results);

    if (nband_combos == 1)
        band_combos_per_sca = 1;
    else
        band_combos_per_sca = nband_combos/nsca;

    /* Output the label line for the GCP data */
    if (fprintf(fptr, "#Pt    RLine       RSamp             RLat         RLon"
               "         RElev      SLine       SSamp            DY          "
               "DX         Flag Coef          RBand SBand RSCA  SSCA    "
               "       SImage                                      "
               "RImage\n") < 0)
    {
        IAS_LOG_ERROR("Writing to the GCP data output file.");
        fclose(fptr);
        return ERROR;
    }

    /* Output the GCP data */
    for (sca_index = 0, i = 0; sca_index < nsca; sca_index++) 
    {
        for (band_combo = 0; band_combo < band_combos_per_sca; band_combo++)
        {
            if (num_sca_bandcombo[sca_index][band_combo] > 0)
            {
                /* output the band combination divider line for band combos
                   that are present in the output */
                fprintf(fptr, "Number of points for the band combination "
                        "reference = %d search = %d: %d points SCA %d\n",
                        gcp_results[i].reference_band,
                        gcp_results[i].search_band,
                        num_sca_bandcombo[sca_index][band_combo], 
                        gcp_results[i].reference_sca);
            }
            for (j = 0; j < num_sca_bandcombo[sca_index][band_combo]; j++, i++)
            {
                if (fprintf(fptr,
                           "%4s %12lf %12lf %12lf %12lf %12lf %12lf %12lf "
                           "%12lf %12lf %5d %12lf %5d %5d %5d %5d %s %s\n",
                           gcp_results[i].point_id, 
                           gcp_results[i].reference_line, 
                           gcp_results[i].reference_sample, 
                           gcp_results[i].latitude, 
                           gcp_results[i].longitude, 
                           gcp_results[i].elevation, 
                           gcp_results[i].predicted_search_line, 
                           gcp_results[i].predicted_search_sample, 
                           gcp_results[i].fit_line_offset,
                           gcp_results[i].fit_sample_offset, 
                           gcp_results[i].accept_flag, 
                           gcp_results[i].correlation_coefficient, 
                           gcp_results[i].reference_band, 
                           gcp_results[i].search_band,
                           gcp_results[i].reference_sca, 
                           gcp_results[i].search_sca,
                           gcp_results[i].reference_name,
                           gcp_results[i].search_name) < 0)
                {
                    IAS_LOG_ERROR("Writing to the GCP residual data output "
                                  "file.");
                    fclose(fptr);
                    return ERROR;
                }
            }
        }
    }

    status = fclose(fptr);
    if (status != 0)
    {
        IAS_LOG_ERROR("Closing residual report file.");
        return ERROR;
    }

    return SUCCESS;
}
