/******************************************************************************
NAME:  save_to_database.c

PURPOSE:  This routine sends results to the database for later analysis. 

RETURNS: SUCCESS or ERROR 
******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "ias_db_insert.h"     /* Database library */
#include "ias_db_insert_transaction_using_table.h" /* Database library */
#include "ias_db_get_connect_info.h" /* Database library */
#include "ias_miscellaneous.h" /* Used for ias_misc_get_char_id */
#include "ias_logging.h"       /* IAS logging library */
#include "save_to_database.h"  /* Header definitions for this file */
#include "tdist.h"             /* Header definitions for whole application */

#define MAX_DELETE_QUERY_LENGTH 85 /* Maximum size of delete query including
                                      ample space for table name and WO ID */

int save_to_database
(
    struct main_db_save_info* main_db_save, /* I: Top-level data to save */
    struct band_sca_db_save_info* band_sca_db_save, /* I/O: Band/SCA table data
                                                            to save */
    const char *char_type,     /* I: "I2I" or "B2B" to tell the routine if this 
                                     is image to image characterization or
                                     band to band characterization */
    int number_records         /* I: Number of records to submit */
)
{
    int status;                /* Function return status */
    int main_table_size;       /* Size of the main database table */ 
    int band_sca_table_size;   /* Size of the band/SCA database table */ 
    char *database_name = NULL; /* Database name */
    char *user_name = NULL;    /* Database user name */
    char *password = NULL;     /* Database password */
    char *host = NULL;         /* Database host */
    char main_database_table_name[35]; /* This table has records that apply to
                                          the whole application run */
    char band_sca_database_table_name[35]; /* This table has per-band/SCA 
                                              records.  For I2I the SCAs are 
                                              all -1 */
    char geometric_wo_common_id[IAS_CHAR_ID_SIZE];   
                               /* Primary key in GEOMETRIC_WORK_ORDER_COMMON
                                  to associate processing with the WO */
    struct ias_db_connection *db = NULL; /* Database connection */
    char query_string[MAX_DELETE_QUERY_LENGTH]; /* SQL statement string */
    struct ias_db_query *query = NULL; /* SQL query */
    char error_message[IAS_SQL_RTN_MSGLEN]; /* SQL query error message */

    /* This routine is called for I2I and B2B characterization */
    if (strcmp(char_type, "B2B") == 0)
    {
        strcpy(main_database_table_name, "band_characterization");
        strcpy(band_sca_database_table_name, "band_characterization_band_sca");
    }
    else if (strcmp(char_type, "I2I") == 0) 
    {
        strcpy(main_database_table_name, "image_characterization");
        strcpy(band_sca_database_table_name, "image_char_band_sca");
    }
    else
    {
        /* The char_type wasn't I2I or B2B */ 
        IAS_LOG_ERROR("Saving data to database.  Characterization type must "
            "be I2I or B2B");
        return ERROR;
    }

    /* Get the database connection info */
    if (ias_db_get_connect_info(&database_name, &user_name, &password, &host)
        != SUCCESS)
    {
        /* Failed getting the database connection info */
        IAS_LOG_ERROR("Getting database connection information");
        return ERROR;
    }

    /* Connect to the database for a transaction */
    db = ias_db_connect_with_transaction(database_name, user_name, password,
        host);

    /* Free the database connection information */
    free(database_name);
    free(user_name);
    free(password);
    free(host);
    database_name = NULL;
    user_name = NULL;
    password = NULL;
    host = NULL;

    if (db == NULL)
    {
        IAS_LOG_ERROR("Connecting to database");
        return ERROR;
    }

    /* Get the database ID corresponding to the current work order */
    status = ias_misc_get_char_id(db, 
        "GEOMETRIC_WORK_ORDER_COMMON", "geometric_wo_common_id", 
        main_db_save->wo_id, geometric_wo_common_id);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Retrieving database ID corresponding to current "
                      "geometric work order %s", main_db_save->wo_id);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Put the Geometric WO common ID in the main table */
    strncpy(main_db_save->geometric_wo_common_id, geometric_wo_common_id, 
        sizeof(main_db_save->geometric_wo_common_id));
    main_db_save->geometric_wo_common_id[sizeof(
        main_db_save->geometric_wo_common_id) - 1] = '\0';

    /* Build the main table of data to save */
    IAS_DB_INSERT_TABLE main_table[] =
    {
        IAS_DB_TABLE_STRUCT_STRING(main_db_save, geometric_wo_common_id),
        IAS_DB_TABLE_STRUCT_STRING(main_db_save, reference_image),
        IAS_DB_TABLE_STRUCT_DOUBLE(main_db_save, t_distribution_threshold),
    };

    main_table_size = sizeof(main_table) / sizeof(IAS_DB_INSERT_TABLE);

    /* Delete any previous records */
    status = snprintf(query_string, sizeof(query_string), "DELETE FROM %s "
        "WHERE geometric_wo_common_id=%s", main_database_table_name,
        geometric_wo_common_id);
    if (status < 0 || status >= sizeof(query_string))
    {
        IAS_LOG_ERROR("Preparing query string: %s ", query_string);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    query = ias_db_perform_query(db, query_string);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, error_message, sizeof(
            error_message));
        IAS_LOG_ERROR("Query of %s failed: %s", main_database_table_name,
            error_message);
        IAS_LOG_ERROR("Deleting records from %s where geometric_wo_common_id="
            "%s", main_database_table_name, geometric_wo_common_id);
        ias_db_query_close(query);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Close query */
    ias_db_query_close(query);
    query = NULL;

    /* Insert the data into the main database table */
    status = ias_db_insert_using_table(db, main_database_table_name, 
        main_table, main_table_size, 1);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Inserting to %s DB table", main_database_table_name);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Put the Geometric WO common ID in the band/SCA table */
    strncpy(band_sca_db_save->geometric_wo_common_id, geometric_wo_common_id, 
        sizeof(band_sca_db_save->geometric_wo_common_id));
    band_sca_db_save->geometric_wo_common_id[sizeof(
        band_sca_db_save->geometric_wo_common_id) - 1] = '\0';

    /* Build the band/SCA level table of data to save */
    IAS_DB_INSERT_TABLE band_sca_table[] =
    {
        IAS_DB_TABLE_STRUCT_STRING(band_sca_db_save, geometric_wo_common_id),
        IAS_DB_TABLE_STRUCT_INT_ARRAY(band_sca_db_save, reference_band_id),
        IAS_DB_TABLE_STRUCT_INT_ARRAY(band_sca_db_save, search_band_id),
        IAS_DB_TABLE_STRUCT_INT_ARRAY(band_sca_db_save, 
            sensor_chip_assembly_id),
        IAS_DB_TABLE_STRUCT_INT_ARRAY(band_sca_db_save, nbr_of_pnts_corr),
        IAS_DB_TABLE_STRUCT_INT_ARRAY(band_sca_db_save, 
            nbr_of_pnts_successfully_corr),
        IAS_DB_TABLE_STRUCT_INT_ARRAY(band_sca_db_save, nbr_of_vld_pnts),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save, rmse_line),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save, rmse_sample),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save, std_dev_line),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save, std_dev_sample),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save, min_line),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save, min_sample),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save, max_line),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save, max_sample),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save, med_line),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save, med_sample),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save, mean_line),
        IAS_DB_TABLE_STRUCT_DOUBLE_ARRAY(band_sca_db_save, mean_sample)
    };

    band_sca_table_size = sizeof(band_sca_table) / sizeof(IAS_DB_INSERT_TABLE);

    /* Delete any previous records */
    status = snprintf(query_string, sizeof(query_string), 
        "DELETE FROM %s WHERE "
        "geometric_wo_common_id=%s", band_sca_database_table_name,
        geometric_wo_common_id);
    if (status < 0 || status >= sizeof(query_string))
    {
        IAS_LOG_ERROR("Preparing query string: %s ", query_string);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    query = ias_db_perform_query(db, query_string);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, error_message, sizeof(
            error_message));
        IAS_LOG_ERROR("Query of %s failed: %s", band_sca_database_table_name,
            error_message);
        IAS_LOG_ERROR("Deleting records from %s where geometric_wo_common_id="
            "%s", band_sca_database_table_name, geometric_wo_common_id);
        ias_db_query_close(query);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* Close query */
    ias_db_query_close(query);
    query = NULL;

    /* Insert the data into the band/SCA database table */
    status = ias_db_insert_using_table(db, band_sca_database_table_name, 
        band_sca_table, band_sca_table_size, number_records);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Inserting to %s DB table", band_sca_database_table_name);
        ias_db_disconnect_with_transaction(db, IAS_DB_ROLLBACK);
        return ERROR;
    }

    /* The inserts worked, so commit the transaction. */ 
    status = ias_db_disconnect_with_transaction(db, IAS_DB_COMMIT);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Committing transaction and disconnecting from database");
    }

    return status;
}
