/*****************************************************************************
NAME:  tdist.c

PURPOSE:  This file runs the tdist routine for Band and Image Registration
          Accuracy Assessment.

RETURN VALUE:  integer value to indicate success or failure
    EXIT_SUCCESS:  The application completed successfully
    EXIT_FAILURE:  The application did not complete successfully

USAGE: tdist <parameter_filename>

    The input GCP results filename is the x2x_char.dat file output by the
    B2B  or I2I program.  The output GCP results filename is the output GCP
    file with the new outliers flagged by the t-dist.  This file is in the
    same form as the input GCP file.  The statistics filename is the name of
    the output statistics file from this routine.

NOTES:
    This routine was written to compute the outliers from the GCP results
    values which are output from band-to-band and image-to-image.  It may be
    used for other GCP results files as long as the file format remains the
    same.
*****************************************************************************/
#include <stdlib.h>
#include <string.h>

#include "ias_logging.h"              /* IAS logging library */
#include "tdist.h"
#include "save_to_database.h"         /* Used for database inserts */

#define TDIST_TYPE_SZ 4               /* Currently I2I or B2B */

int main (int argc, char *argv[])
{
    char gcp_results_file[PATH_MAX];  /* Name of the input GCP data file */
    char residual_file[PATH_MAX];     /* Name of the residual GCP data file */
    char report_name[REPORT_NAME_SIZE]; /* Name of the GCP data file report */
    char stats_file[PATH_MAX];        /* Name of the statistics file */
    char ref_l0r[PATH_MAX];           /* Name of the reference L0R image */
    char ref_img[PATH_MAX];           /* Reference image name */
    char srch_img[PATH_MAX];          /* Search image name */
    char l1g_file[PATH_MAX];          /* Name of the L1G image */
    char cpf_filename[PATH_MAX];      /* CPF file name */
    char char_type[TDIST_TYPE_SZ] = ""; /* I2I or B2B to tell the DB save 
                                         routine if this is Image to image or
                                         band to band characterization */
    char omf_file_name[PATH_MAX];     /* omf file name */
    char input_gcp_source[4];         /* input gcp source: B2B or I2I */
    int nref_bands;                   /* Number of reference bands */
    int ref_bands[IAS_MAX_NBANDS];    /* Array of reference band numbers */
    int nsrch_bands;                  /* Number of search bands */
    int srch_bands[IAS_MAX_NBANDS];   /* Array of search band numbers */
    int num_gcp_results;              /* Number of IAS_GCP_RESULTS structs in
                                         the array */
    int num_band_gcp_results;         /* Number of IAS_GCP_RESULTS structs for
                                         the current band */
    int curr_gcp;                     /* Current GCP data point */
    int nband_combos;                 /* Number of band combinations */
    int *corr_gcps;                   /* Array of the number of correlated 
                                         GCPs for each band combo */
    int band_start;                   /* Starting index for the current band */
    int number_db_save_records = 0;   /* Number of records to save to the DB */
    int status;                       /* Status returned from function call */
    double tolerance;                 /* Tolerance value for the t-dist */
    IAS_GCP_RESULTS *gcp_results;     /* Array of IAS_GCP_RESULTS structures */
    int num_sca_bandcombo[IAS_MAX_NSCAS][NBANDCOMBO]; /* Number of points in 
                                         band combination for an SCA */
    int statistics_flag;              /* Flag to create statistics file */
    int residual_flag;                /* Flag to create residuals file */
    int nsca = 0;                     /* Number of SCAs */
    int first_ref_band;               /* First reference band in GCP struct */
    int first_srch_band;              /* First search band in GCP structure */
    int first_sca_number;             /* First SCA in GCP structure */
    int combo_index;                  /* Band combination index */
    struct main_db_save_info main_db_save; /* Main DB save information */
    struct band_sca_db_save_info band_sca_db_save; /* Band/SCA level DB save 
                                         information */
    int db_save_flag = 0;             /* Flag to indicate if information should
                                         be written to the database */
    IAS_SATELLITE_ID satellite_id;    /* Satellite ID */

    /* Initialize the structures holding information to be saved in the DB */
    memset(&main_db_save, 0, sizeof(main_db_save));
    memset(&band_sca_db_save, 0, sizeof(band_sca_db_save));
    memset(num_sca_bandcombo, 0, sizeof(num_sca_bandcombo));

    /* Initialize the logging library */
    if (ias_log_initialize("tdist") != SUCCESS)
    {
        IAS_LOG_ERROR("Error initializing logging library");
        exit(EXIT_FAILURE);
    }

    /* Check the arguments and get the parameter file from the arguments */
    if (argc != 2)
    {
        IAS_LOG_ERROR("Usage: tdist <parameter file>");
        exit(EXIT_FAILURE);
    }

    /*  Get parameters from the ODL file */
    status = get_parameters(argv[1], gcp_results_file, cpf_filename, 
        &tolerance, stats_file, &statistics_flag, residual_file, 
        &residual_flag, ref_l0r, l1g_file, &main_db_save, &db_save_flag,
        &satellite_id, omf_file_name, input_gcp_source);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving parameters in get_parameters");
        exit(EXIT_FAILURE);
    }

    /* Initialize satellite attributes */
    status = ias_sat_attr_initialize(satellite_id);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attributes library");
        exit(EXIT_FAILURE);
    }

    IAS_LOG_INFO("TOLERANCE %f", tolerance);

    /* Read the GCP data file */
    if (get_gcp_data(gcp_results_file, report_name, ref_img, srch_img,
            &nref_bands, ref_bands, &nsrch_bands, srch_bands, &nband_combos,
            &num_gcp_results, &gcp_results, &corr_gcps) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading the GCP data file <%s>.", gcp_results_file);
        exit(EXIT_FAILURE);
    }

    /* Start looping through the band combinations.  Start looking at the 
       second GCP data point.  Loop until all GCP data points have been 
       checked for outliers */
    combo_index = 0;
    curr_gcp = 0;
    first_ref_band = gcp_results[curr_gcp].reference_band;
    first_srch_band = gcp_results[curr_gcp].search_band;
    first_sca_number = gcp_results[curr_gcp].reference_sca;
    while (curr_gcp < num_gcp_results)
    {
        /* While the current reference and search bands  and the next reference
           and search bands are the same, then we are still in the same band
           combination.  Keep track of the number of GCPs for this band
           combination.  Start comparing the second point in the band 
           combination with the first point in the band combination */
        band_start = curr_gcp;
        curr_gcp++;               /* Skip the first point */
        num_band_gcp_results = 1; /* Skip the first point */
        while (curr_gcp < num_gcp_results 
               && gcp_results[curr_gcp].reference_band == 
                  gcp_results[curr_gcp-1].reference_band 
               && gcp_results[curr_gcp].search_band == 
                  gcp_results[curr_gcp-1].search_band
               && gcp_results[curr_gcp].reference_sca == 
                  gcp_results[curr_gcp-1].reference_sca)
        {
            if (curr_gcp < num_gcp_results - 1)
            {
                if (gcp_results[curr_gcp].search_sca ==
                    gcp_results[curr_gcp-1].search_sca)
                {
                    curr_gcp++;
                    num_band_gcp_results++;
                }
                else
                    break;
            }
            else
            {
                curr_gcp++;
                num_band_gcp_results++;
            }
        }
        num_sca_bandcombo[nsca][combo_index] = num_band_gcp_results;

        if (curr_gcp < num_gcp_results)
        {
            if (first_ref_band == gcp_results[curr_gcp].reference_band &&
                first_srch_band == gcp_results[curr_gcp].search_band &&
                first_sca_number != gcp_results[curr_gcp].reference_sca)
            {
                nsca++;
                combo_index = 0;
            }
            else
            {
                combo_index++;
            }
        }

        /* Implement the t-dist outlier rejection for the current band
           combination */
        filter_outliers(tolerance, gcp_results[band_start].search_sca, 
            num_band_gcp_results, &gcp_results[band_start]);
    }
    nsca++; /* Only increments if next SCA is found, so add one to total */

    /* Define the characterization type to help report and DB operations */
    if (!strcmp(report_name, "I2I_CHAR"))
    {
        strcpy(char_type, "I2I");
    }
    else if (!strcmp(report_name, "B2B_CHAR"))
    {
        strcpy(char_type, "B2B");
    }

    /* Calculate and output the GCP statistics */
    if (statistics_flag)
    {
        if (put_gcp_stats(stats_file, report_name, l1g_file, ref_l0r, ref_img,
            srch_img, cpf_filename, corr_gcps, tolerance, nref_bands, ref_bands,
            nsrch_bands, srch_bands, num_gcp_results, gcp_results, char_type,
            &db_save_flag, &main_db_save, &band_sca_db_save, 
            &number_db_save_records) != SUCCESS)
        {
            IAS_LOG_ERROR("Writing to the GCP statistics file %s.", stats_file);
            free(gcp_results);
            free(corr_gcps);
            exit(EXIT_FAILURE);
        }
    }
    else
    {
        /* You can't save statistics to the database if you don't calculate
           the statistics */
        db_save_flag = 0;
    }

    /* Output the residual file */
    if (residual_flag)
    {
        if (put_residual_data(residual_file, report_name, l1g_file, ref_l0r,
            ref_img, srch_img, nref_bands, ref_bands, nsrch_bands, srch_bands,
            num_gcp_results, nsca, nband_combos, num_sca_bandcombo,
            gcp_results, main_db_save.wo_id) != SUCCESS) 
        {
            IAS_LOG_ERROR("Writing to the GCP residual file %s.", 
                residual_file);
            free(gcp_results);
            free(corr_gcps);
            exit(EXIT_FAILURE);
        }
    }

    /* Free the pointers */
    free(gcp_results);
    free(corr_gcps);

    /* If storing results in the database is requested, send the results to 
       the database */
    if (db_save_flag)
    {
        IAS_LOG_INFO("Writing results to database");
        if (save_to_database(&main_db_save, &band_sca_db_save, char_type,
            number_db_save_records) != SUCCESS)
        {
            IAS_LOG_ERROR("Storing results in database");
            exit(EXIT_FAILURE);
        }
    }

    if (update_omf(omf_file_name , residual_file, input_gcp_source) != SUCCESS)
    {
        IAS_LOG_ERROR("Updating omf file");
        exit(EXIT_FAILURE);
    }

    /* Write completion message */
    IAS_LOG_INFO("Successful completion");

    return EXIT_SUCCESS;
}
