#ifndef _TDIST_H_
#define _TDIST_H_

#include "ias_gcp.h"    /* GCP I/O library */
#include "ias_satellite_attributes.h" /* Satellite Attributes library */
#include "save_to_database.h" /* Set up database inserts */

#define NBANDCOMBO ((IAS_MAX_NBANDS * (IAS_MAX_NBANDS - 1)) / 2) 
                        /* Number of band combinations */
#define REPORT_NAME_SIZE 80 /* The name should easily fit within 1 line */

typedef struct GCPSTATS 
{
    double min[2];      /* Line, sample minimum values */
    double max[2];      /* Line, sample maximum values */
    double avg[2];      /* Line, sample average */
    double stddev[2];   /* Line, sample standard deviation */
    double rmse[2];     /* Line, sample RMSE */
    int ref_band;       /* Reference band number */
    int srch_band;      /* Search band number */
    int ttl_gcps;       /* Total number of GCPs */
    int num_outliers;   /* Number of outlier GCPs */
} GCPSTATS;

void filter_outliers 
(
    double tolerance,   /* I: Tolerance value for the t-dist */
    int sca,            /* I: SCA number */
    int num_gcp_results, /* I: Number of IAS_GCP_RESULTS structs in the 
                               array */
    IAS_GCP_RESULTS *gcp_results /* I/O: Array of IAS_GCP_RESULTS structures */
);

int get_gcp_data 
(
    const char *gcp_results_file, /* I: Name of the GCP results file */
    char report_name[REPORT_NAME_SIZE], /* O: Name of the report */
    char ref_img[PATH_MAX], /* O: Reference image name */
    char srch_img[PATH_MAX], /* O: Search image name */
    int *nref_bands,    /* O: Number of reference bands */
    int *ref_bands,     /* O: Array of reference band numbers */
    int *nsrch_bands,   /* O: Number of search bands */
    int *srch_bands,    /* O: Array of search band numbers */
    int *nband_combos,  /* O: Number of GCP band combinations */
    int *num_gcp_results, /* O: Number of IAS_GCP_RESULTS structs in the 
                                array */
    IAS_GCP_RESULTS **gcp_results, /* O: Address of an array of 
                                         IAS_GCP_RESULTS structures */
    int **corr_gcps     /* O: Array of the number of correlated GCPs */
);

int put_gcp_stats 
(
    const char *out_gcp_stats_file, /* I: Name of the GCP statistics output 
                                          file */
    const char *report_name,  /* I: Name of the report */
    const char *l1g_file,     /* I: L1G filename for report header data */
    const char *ref_L0R,      /* I: Name of the reference L0R image */
    const char *ref_img,      /* I: Reference image name */
    const char *srch_img,     /* I: Search image name */
    const char *cpf_filename, /* I: Name of the Calibration Parameter file */
    const int *ncorr_gcps,    /* I: Array of the number of correlated GCPs for
                                    each band combination */
    double thresh,            /* I: Threshold of the t-dist routine */
    int nref_bands,           /* I: Number of reference bands */
    const int *ref_bands,     /* I: Array of reference band numbers */
    int nsrch_bands,          /* I: Number of search bands */
    const int *srch_bands,    /* I: Array of search band numbers */
    int num_gcp_results,      /* I: Number of IAS_GCP_RESULTS structs in the
                                    array */
    const IAS_GCP_RESULTS *gcp_results, /* I: Array of IAS_GCP_RESULTS 
                                              structures */
    char *char_type,          /* I: "I2I" or "B2B" to tell the DB save routine 
                                    if this is image to image characterization
                                    or band to band characterization since
                                    tdist is used in both cases */
    int *db_save_flag,        /* I/O: Flag used to indicate whether results
                                      are saved to the database */
    struct main_db_save_info *main_db_save_info, /* I/O: info to save to DB */
    struct band_sca_db_save_info *band_sca_db_save_info, /* I/O: info to save 
                                                            to the DB */
    int *number_db_save_records /* O: The number of records that get added to
                                       structure array to be saved to the DB */
);

int put_residual_data 
(
    const char *out_gcp_residual_data_file, /* I: Name of the GCP residual 
                                    data output file */
    const char *report_name,  /* I: Name of the report */
    const char *l1g_file,     /* I: L1G filename for report header data */
    const char *ref_L0R,      /* I: Name of the reference L0R image */
    const char *ref_img,      /* I: Reference image name */
    const char *srch_img,     /* I: Search image name */
    int nref_bands,           /* I: Number of reference bands */
    const int *ref_bands,     /* I: Array of reference band numbers */
    int nsrch_bands,          /* I: Number of search bands */
    const int *srch_bands,    /* I: Array of search band numbers */
    int num_gcp_results,      /* I: Number of IAS_GCP_RESULTS structs in the
                                    array */
    int nsca,                 /* I: Number of SCAs */
    int nband_combos,         /* I: Total number of band combos */
    int num_sca_bandcombo[IAS_MAX_NSCAS][NBANDCOMBO], /* I: Number of points in
                                                            band combo for an 
                                                            SCA */
    const IAS_GCP_RESULTS *gcp_results, /* I: Array of IAS_GCP_RESULTS 
                                              structures */
    const char *work_order_id /* I: Work order ID */

);

int get_parameters
(
    const char parm_file[PATH_MAX],      /* I: WO parameter file */
    char gcp_results_filename[PATH_MAX], /* O: Input residuals file */
    char *cpf_filename,                  /* O: CPF file */
    double *tolerance,                   /* O: threshold tolerance */
    char statistics_filename[PATH_MAX],  /* O: Statistics file */
    int *statistics_flag,                /* O: Flag to create statistics file */
    char residual_filename[PATH_MAX],    /* O: Residuals file */
    int *residuals_flag,                 /* O: Flag to create residuals file */
    char l0r_filename[PATH_MAX],         /* O: L0R file */
    char l1g_filename[PATH_MAX],         /* O: L1G file */
    struct main_db_save_info *main_db_save, /* O: Structure containing
                                                  information to save to the 
                                                  DB */
    int *save_to_db_flag,                /* O: Flag to indicate if results are 
                                               saved to DB (1 = yes, 0 = no) */
    IAS_SATELLITE_ID *satellite_id,      /* O: Satellite ID */
    char omf_file_name[PATH_MAX],        /* O: omf file name */
    char input_gcp_source[4]             /* O input gcp source either B2B or
                                              I2I */
);

int update_omf
(
    char *omf_file_name,                 /* I: omf file name */
    char *residual_filename,             /* I: residual file name */
    char *input_gcp_source               /* I: gcp input source */
);

#endif
