/*****************************************************************************
Name: filter_to_sca_bounds

Purpose: Provides the routines to determine the bounding box for each of the
    SCAs and calculate the starting and ending sample number on any given
    line of the output image.  This is an optimization to avoid doing 
    calculations for pixels outside the current SCA.

Notes:
    - For each SCA, an bounding box is provided for each elevation plane
      in the grid.  When calculating the start and end samples on a given
      line, all the bounding boxes for the SCA are checked to find the min/max
      samples.  Another optimization would be to reduce it down to a single
      bounding box per SCA.  However, the current method already eliminates
      nearly 75% of the application run-time, so additional optimizations can
      wait for later.
*****************************************************************************/
#include <stdlib.h>
#include "ias_miscellaneous.h"
#include "ias_logging.h"
#include "terrain_occlusion.h"

/*****************************************************************************
Name: get_sca_bounds

Purpose: For each of the SCAs in the requested band, construct the bounding
    box (in terms of output line/sample) for each of the elevation planes
    in the grid.

Returns:
    Pointer to the 2D array of bounding boxes.  The first index in the array
    is the SCA and the second index is the bounding box at a specific elevation
    plane in the grid.

*****************************************************************************/
BOUNDING_BOX **get_sca_bounds
(
    const IAS_GRID_TYPE *grid_ptr,  /* I: Pointer to the grid */
    int band_index,                 /* I: Band index to use */
    int *boxes_per_sca              /* O: Bounding box count per sca */
)
{
    void **raw_sca_bounds;              /* Bounds of the SCAs as void** */
    BOUNDING_BOX **sca_bounding_boxes;  /* Bounds of the SCAs in the band */
    int total_scas;                     /* Cached count of SCAs in the grid */
    int zplanes;                        /* Cached count of grid zplanes */
    int sca;                            /* SCA loop counter */
    

    /* Cache some grid values.  The assumption is made that all the SCAs have
       the same number of elevation planes, which is a safe assumption. */
    total_scas = grid_ptr->scas_per_band[band_index];
    zplanes = grid_ptr->gridbands[band_index][0].nzplanes;

    /* Allocate a 2D array for the SCA bounds */
    raw_sca_bounds = ias_misc_allocate_2d_array(total_scas, zplanes,
            sizeof(BOUNDING_BOX));
    if (raw_sca_bounds == NULL)
    {
        IAS_LOG_ERROR("Error allocating SCA bounds array");
        return NULL;
    }
    /* typecast the void** to the type being used */
    sca_bounding_boxes = (BOUNDING_BOX **)raw_sca_bounds;

    /* Get the bounding boxes for each of the SCAs */
    for (sca = 0; sca < total_scas; sca++)
    {
        IAS_GRID_BAND_TYPE *grid_band_ptr
                = &grid_ptr->gridbands[band_index][sca];
        int grid_lines = grid_band_ptr->ngrid_lines;
        int grid_samps = grid_band_ptr->ngrid_samps;
        int zplane;
        int indices[4];
        int zplane_size = grid_lines * grid_samps;

        /* Set the indices for the 4 corners in the grid out_lines and
           out_samps arrays.  This is done to make the corner looping easier. */
        indices[0] = 0;
        indices[1] = grid_samps - 1;
        indices[2] = zplane_size - 1;
        indices[3] = grid_samps * (grid_lines - 1);

        /* Get the bounds for each elevation plane */
        for (zplane = 0; zplane < zplanes; zplane++)
        {
            int corner;
            double lines[5];
            double samps[5];
            BOUNDING_BOX *bounding_box = &sca_bounding_boxes[sca][zplane];

            /* Get all 4 corners of the bounding box */
            for (corner = 0; corner < 4; corner++)
            {
                int grid_index = indices[corner] + zplane * zplane_size;

                lines[corner] = grid_band_ptr->out_lines[grid_index];
                samps[corner] = grid_band_ptr->out_samps[grid_index];
            }
            /* Close the loop on the bounding box */
            lines[4] = lines[0];
            samps[4] = samps[0];

            /* Set the edges of the bounding box in the bounding box
               structure */
            for (corner = 0; corner < 4; corner++)
            {
                bounding_box->edges[corner].x1 = samps[corner];
                bounding_box->edges[corner].y1 = lines[corner];
                bounding_box->edges[corner].x2 = samps[corner + 1];
                bounding_box->edges[corner].y2 = lines[corner + 1];
            }
        }
    }

    /* Return the number of bounding boxes provided for each SCA.  This is to
       allow the rest of the code to not make an assumption that the number
       of bounding boxes is not the same as the number of elevation planes. */
    *boxes_per_sca = zplanes;

    return sca_bounding_boxes;
}

/*****************************************************************************
Name: free_sca_bounds

Purpose: Frees the 2D array that stores the SCA bounding boxes.

Returns: Nothing

*****************************************************************************/
void free_sca_bounds
(
    BOUNDING_BOX **sca_bounds   /* I: SCA bounds to free */
)
{
    ias_misc_free_2d_array((void **)sca_bounds);
}

/*****************************************************************************
Name: get_sample_range

Purpose: Given a set of bounding boxes and a line number, calculate the 
    starting and ending sample that need to be checked on this line.

Returns: Nothing

Notes:
    - If no intersections are found, the start sample and end sample will both
      be zero.

*****************************************************************************/
void get_sample_range
(
    const BOUNDING_BOX *bounds,     /* I: List of bounds to check */
    int bounds_count,               /* I: Number of bounding boxes in bounds
                                          (each one with 4 entries) */
    int line_number,                /* I: Line number to check */
    int max_samps,                  /* I: Maximum number of samples */
    int *start_sample,              /* O: Start sample found */
    int *end_sample                 /* O: End sample found */
)
{
    IAS_LINE_SEGMENT output_line;  /* Output line segment */
    int bounds_index;              /* Loop counter */
    double min_sample = 500000.0;  /* minimum sample number initialized to a 
                                      ridiculously large value */
    double max_sample = -500000.0; /* maximum sample number initialized to a 
                                      ridiculously small value */

    /* Initialize the output line segement to go well outside the bounds of
       the output image */
    output_line.x1 = -500000.0;
    output_line.x2 = 500000.0;

    /* Set the output line segment to the current line number */
    output_line.y1 = line_number;
    output_line.y2 = line_number;

    /* Check each line of the output image to see if and where it intersects
       the bounding box */
    for (bounds_index = 0; bounds_index < bounds_count; bounds_index++)
    {
        const IAS_LINE_SEGMENT *edge = &bounds[bounds_index].edges[0];
        double intersect_line[4];           /* line intersection points */
        double intersect_sample[4];           /* sample intersection points */
        int intersect_count = 0;
        int ii;

        /* Check the output line against each of the lines in the bounding
           box */
        for (ii = 0; ii < 4; ii++)
        {
            int status;                 /* Intersection status */

            /* Check for intersection of output line and the current edge
               of the bounding box */ 
            status = ias_math_find_line_segment_intersection(&output_line,
                    &edge[ii], &intersect_sample[intersect_count],
                    &intersect_line[intersect_count]);
            if (status == IAS_LINES_INTERSECT)
            {
                /* Intersection found */
                intersect_count++;
            }
        }
        /* If there are 2 or more intersections, find the minimum and maximum
           values since they are the ones that should be used for start/stop
           samples.  It should be pretty rare to get more than 2 intersections,
           but it can happen if one of the corners falls exactly on an output
           image line.  If there are zero or one intersections, the lookup 
           table entry can remain at the original zero values it is initialized
           to.  Note that one intersection could happen if the corner of the
           bounding box falls right on the output line. */
        if (intersect_count >= 2)
        {
            int index;

            for (index = 0; index < intersect_count; index++)
            {
                if (intersect_sample[index] < min_sample)
                    min_sample = intersect_sample[index];
                if (intersect_sample[index] > max_sample)
                    max_sample = intersect_sample[index];
            }
        }
    }

    /* Add a bit of "slop" to the start and end samples to make sure no points
       are skipped if there is some curvature across the SCA.  10 should be
       a sufficiently large value. */
    *start_sample = min_sample - 10;
    if (*start_sample < 0)
        *start_sample = 0;
    *end_sample = max_sample + 10;
    if (*end_sample > max_samps)
        *end_sample = max_samps;

    /* If the start sample is after the end sample, set both to zero since
       there weren't any intersections */
    if (*start_sample > *end_sample)
    {
        *start_sample = 0;
        *end_sample = 0;
    }
}
