/*****************************************************************************
NAME:        get_parms

PURPOSE:
Get the terrain occlusion parameters from the input parameter file and OMF
file.

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion 
ERROR    Operation failed

******************************************************************************/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "terrain_occlusion.h"

#include "ias_const.h"                    /* IAS defines, including new
                                             INVALID_ELEVATION constant */
#include "ias_parm.h"                     /* parameter library */
#include "ias_parm_standard_parameters.h" /* standard parameter definitions */
#include "ias_satellite_attributes.h"

int get_parms
(
    const char *parm_file,      /* I: Name of the Parameter File */
    char *grid_file,            /* O: Name of grid file */
    char *occlusion_file,       /* O: Name of the occlusion output file */
    char *geom_grid_file,       /* O: Name of DEM grid file */
    int *occlusion_band,        /* O: Number of band to process */
    char *wo_id,                /* O: Work Order Id */
    char *dem_file,             /* O: Name of unresampled DEM file */
    char *omf_file,             /* O: OMF file name */
    int *number_of_procs,       /* O: Number of threads to use */
    int *min_elevation,         /* O: Minimum elevation in DEM file */
    int *max_elevation,         /* O: Maximum elevation in DEM file */
    IAS_SATELLITE_ID *satellite_id   /* O: Satellite Id */
)
{
    int count;
    int status;

    /* Local variables for storing ODL values */
    char local_wo_id[IAS_WORKORDER_ID_SIZE];
    char local_grid_file[PATH_MAX];
    char local_geom_grid_file[PATH_MAX];
    char local_dem_file[PATH_MAX];
    char local_output_file[PATH_MAX];
    int local_occlusion_band;
    int local_num_of_procs;
    int local_min_elevation = 0;
    int local_max_elevation = 0;
    int local_satellite_number;
    int processing_pass;

    /* Initialize local strings to empty string */
    strcpy(local_wo_id, "");
    strcpy(local_grid_file, "");
    strcpy(local_dem_file, "");
    strcpy(local_output_file, "");
    strcpy(local_geom_grid_file, "");

    /* Start the parameter file table */    
    IAS_PARM_DECLARE_TABLE(parms, 4);

    /* Add the work order id */
    IAS_PARM_WORK_ORDER_ID(parms, local_wo_id, sizeof(local_wo_id), 0);

    /* Get the occlusion band */
    int default_occlusion_band[] = {6};
    IAS_PARM_ADD_INT(parms, OCCLUSION_BAND, "Band number used to calculate "
        "the terrain occlusion mask (0 = First try band 6, then the first "
        "available band)", IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        1, 0, IAS_MAX_NBANDS, 1, default_occlusion_band, &local_occlusion_band,
        sizeof(local_occlusion_band), 1);

    /* Read the output file name from parameters file */
    IAS_PARM_ADD_STRING(parms, OCCLUSION_FILENAME,
        "Output occlusion file name" , IAS_PARM_REQUIRED,
        0, NULL, 0, NULL, local_output_file,
        sizeof(local_output_file), 0);

    /* Get the processing pass */
    IAS_PARM_PROCESSING_PASS(parms, &processing_pass, 
        sizeof(processing_pass), 1);

    count = IAS_PARM_GET_TABLE_SIZE(parms);

    /* Check if help or template output was requested */
    status = ias_parm_provide_help(parm_file, parms, count,
                                   IAS_INPUT_PARAMETERS);

    if (status == 1)
    {
        /* --help or --template was requested.  Copy the requested option to 
           omf_file so the same operation will be performed for the OMF */
        strcpy(omf_file, parm_file);
    }
    else if (status == ERROR) /* Help failed. */
    {
        IAS_LOG_ERROR("Providing help for application parameter file %s",
            parm_file);
        exit(EXIT_FAILURE);
    }
    else
    {
        if (ias_parm_read(parm_file, parms, count) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading parameters from: %s", parm_file);
            return ERROR;
        }

        strcpy(occlusion_file, local_output_file);
        strcpy(wo_id, local_wo_id);
        *occlusion_band = local_occlusion_band;

        if (strcmp(wo_id, "") == 0)
        {
            IAS_LOG_ERROR("Missing work order id in the parameter file.");
            return ERROR;
        }

        if (strcmp(occlusion_file, "") == 0)
        {
            IAS_LOG_ERROR("Missing output filename in the parameter file.");
            return ERROR;
        }

        /* build the OMF file name */
        sprintf(omf_file, "%s.omf", wo_id);
    }

    /***** Read additional parameters from OMF *****/

    /* Start the OMF table */    
    IAS_PARM_DECLARE_TABLE(omf_parms, 7);
    
    /* Add the satellite number */
    IAS_PARM_SATELLITE(omf_parms, &local_satellite_number,
        sizeof(local_satellite_number), 1);

    /* Add the DEM file name */
    IAS_PARM_DEM_FILENAME(omf_parms, local_dem_file, 
        sizeof(local_dem_file), 1); 

    /* Add the grid DEM file name */
    IAS_PARM_GEOM_GRID_FILENAME_FROM_PASS(omf_parms, IAS_PARM_OPTIONAL,
        local_geom_grid_file, sizeof(local_geom_grid_file),
        1, processing_pass);

    /* Add the grid file name */
    IAS_PARM_GRID_FILENAME_FROM_PASS(omf_parms, IAS_PARM_OPTIONAL, 
        local_grid_file, sizeof(local_grid_file), 1, processing_pass);

    /* Add the maximum number of processor to use */
    IAS_PARM_MAX_PROCESSORS_TO_USE(omf_parms, IAS_PARM_OPTIONAL, 
        &local_num_of_procs, sizeof(local_num_of_procs), 1);

    /* Get minimum elevation. */
    int invalid_min_elevation[] = {INVALID_ELEVATION};
    IAS_PARM_MIN_ELEVATION(omf_parms, IAS_PARM_OPTIONAL, 1,
        invalid_min_elevation, &local_min_elevation,
        sizeof(local_min_elevation), 1);

    /* Get maximum elevation. */
    int invalid_max_elevation[] = {INVALID_ELEVATION};
    IAS_PARM_MAX_ELEVATION(omf_parms, IAS_PARM_OPTIONAL, 1,
        invalid_max_elevation, &local_max_elevation,
        sizeof(local_max_elevation), 1);

    count = IAS_PARM_GET_TABLE_SIZE(omf_parms);

    /* Check if help or template output is being requested */
    status = ias_parm_provide_help(omf_file, omf_parms, count, 
        IAS_OMF_PARAMETERS);

    if (status == 1)
    {
        /* the --help or --template was provided. can now exit success */
        exit(EXIT_SUCCESS);
    }
    else if (status == ERROR) /* Help failed */
    {
        IAS_LOG_ERROR("Providing help for OMF file: %s", omf_file);
        return ERROR;
    }

    /* Read the OMF file */
    if (ias_parm_read(omf_file, omf_parms, count) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from OMF file: %s", omf_file);
        return ERROR;
    }

    strcpy(dem_file, local_dem_file);
    strcpy(grid_file, local_grid_file);
    strcpy(geom_grid_file, local_geom_grid_file);

    *number_of_procs = local_num_of_procs;
    *min_elevation = local_min_elevation;
    *max_elevation = local_max_elevation;

    /* Convert the satellite number to the satellite id */
    *satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(
        local_satellite_number);
    if (*satellite_id == ERROR)
    {
        IAS_LOG_ERROR("Unable to determine satellite ID from satellite "
            "number %d", local_satellite_number);
        return ERROR;
    }

    /* A DEM file is required. Return an error if it was not provided */ 
    if (strcmp(dem_file, "") == 0)
    {
        IAS_LOG_ERROR("The DEM file is a required parameter");
        return ERROR;
    }

    /* A grid file is required. Return an error if it was not provided */
    if (strcmp(grid_file, "") == 0)
    {
        IAS_LOG_ERROR("The grid file is a required parameter");
        return ERROR;
    }

    /* A geom grid file is required. Return error if it was not provided */
    if (strcmp(geom_grid_file, "") == 0)
    {
        IAS_LOG_ERROR("The geometric grid file is a required parameter");
        return ERROR;
    }

    return SUCCESS; 
}
