/******************************************************************************
NAME:   occlusion_threading

PURPOSE: Contains all the terrain occlusion functions related to 
multi-threading the terrain_occlusion application.

******************************************************************************/

#include <stdlib.h>

#include "terrain_occlusion.h"

#include "ias_logging.h"
#include "ias_grid.h"
#include "ias_geom_grid.h"
#include "ias_threadpool.h"
#include "ias_miscellaneous.h"

/* Define a structure for the parameters provided to the thread function */
typedef struct params
{
    struct ias_threadpool *threadpool;   /* Threadpool being used */
    const IAS_GRID_TYPE *grid;           /* Pointer to grid */
    const IAS_GEOM_GRID_TYPE *geom_grid; /* Pointer to geom grid */
    const IAS_IMAGE *dem;                /* DEM image data */
    BOUNDING_BOX **sca_bounds;           /* Bounds of the SCAs in the band,
                                            indexed by sca, bounding box within
                                            the SCA */
    int sca_bounds_per_sca;              /* Bounding box count for each sca */
    double max_elev;                     /* Maximum elevation */
    int band_index;                      /* Band index */
    IAS_IMAGE *occ;                      /* Terrain occlusion mask */
    int *occluded_pixels;                /* To track occluded pixels per SCA */
    int thread_error;   /* Non-zero if any threads exits with an error */
} PARAMS;

/* Define a structure for passing information to the work queue */
typedef struct queue_message
{
    int *image_totals;      /* Pointer to array for overall totals */
    int *thread_totals;     /* Pointer to array for thread totals */
    int num_of_scas;        /* Number of SCAs in totals arrays */
} QUEUE_MESSAGE;

/* Mutex for serializing access to the work assignment operation */
static IAS_THREAD_MUTEX_TYPE work_assignment_mutex;

/******************************************************************************
Name: update_occluded_pixel_totals

Purpose: Add the occluded pixels found in a thread to the overall totals for
the image.  This function is called from the thread function and is run in
the main thread. 

Returns:  SUCCESS
******************************************************************************/
static int update_occluded_pixel_totals(void *message_in)
{
    QUEUE_MESSAGE *message = (QUEUE_MESSAGE *)message_in;
    int *totals = message->image_totals;
    int *additional = message->thread_totals;
    int scas = message->num_of_scas;
    int ii;

    /* Add the thread counts to the overall totals.  This function will be 
       running in the main thread so the calcuation doesn't need to be 
       protected by a mutex. */
    for (ii = 0; ii < scas; ii++)
    {
        totals[ii] += additional[ii];
    }

    /* The occluded pixel thread doesn't free its array of occluded pixel
       counts or the message added to the work queue. Free up the memory for
       both here. */
    free(message->thread_totals);
    free(message);

    return SUCCESS;
}

/******************************************************************************
Name: get_assigned_sca_line

Purpose: Returns the next SCA and line number to be processed. 

Returns:  SUCCESS or ERROR
******************************************************************************/
static int get_assigned_sca_line
(
    int total_scas,      /* I: Number SCA in the image */
    int max_line_number, /* I: Number of lines in image */
    int *assigned_sca,   /* O: SCA assigned */
    int *assigned_line,  /* O: Line number assigned */
    int *work_assigned   /* O: Flag to indicate if work was assigned */
)
{
    static int current_sca = 0;  /* initialize the SCA */
    static int current_line_number = 0;  /* next line to assign */

    /* Default to no work assigned */
    *work_assigned = 0;

    /* Lock the mutex before checking for new work */
    if (IAS_THREAD_LOCK_MUTEX(&work_assignment_mutex) != 0)
    {
        IAS_LOG_ERROR("Error locking work assignment mutex");
        return ERROR;
    }

    /* Check if there is more work to be assigned */
    if ((current_sca < total_scas) && (current_line_number < max_line_number))
    {
        /* Assign the current SCA number */
        *assigned_sca = current_sca;

        /* Assign the current line number */
        *assigned_line = current_line_number;

        /* Increment the next line to be assigned */
        current_line_number++;

        /* If the max lines has been exceeded, move to the next SCA */
        if (current_line_number >= max_line_number)
        {
            current_sca++;
            current_line_number = 0;
        }

        if (IAS_THREAD_UNLOCK_MUTEX(&work_assignment_mutex) != 0)
        {
            IAS_LOG_ERROR("Error unlocking work assignment mutex after "
                "work assigned for SCA/line %d/%d", current_sca, 
                current_line_number);
            return ERROR;
        }

        *work_assigned = 1;

        /* Print the start of processing for each SCA.  The first SCA will
           be at line number 1 at this point. */
        if (*assigned_line == 0)
        {
            IAS_LOG_INFO("Starting SCA %d", current_sca);
        }

        return SUCCESS;
    }

    if (IAS_THREAD_UNLOCK_MUTEX(&work_assignment_mutex) != 0)
    {
        IAS_LOG_ERROR("Error unlocking work assignment mutex");
        return ERROR;
    }

    /* No work assigned, but that is still successful */
    return SUCCESS;
}

/******************************************************************************
Name:  occlusion_thread

Purpose: Thread function that calculates the terrain occlusion mask.

Returns: SUCCESS or ERROR
******************************************************************************/
static int occlusion_thread
(
    void *params_in,  /* I: Pointer to input parameter structure */
    int thread_num    /* I: Thread number for thread running this function */
)
{
    PARAMS *params = (PARAMS *)params_in;
    const IAS_GRID_TYPE *grid = params->grid;
    const IAS_GEOM_GRID_TYPE *geom_grid = params->geom_grid;
    int band_index = params->band_index;
    int total_scas = grid->scas_per_band[band_index];
    int total_lines = grid->gridbands[band_index][0].lines;
    int sca;
    int line_number;
    int work_assigned;

    int occ_pixels_cnt;   /* occluded pixels found in each call to
                             terrain_occlusion_mask */
    int *occluded_pixels = NULL; /* store occluded pixels for current thread */

    /* allocate storage for the occluded pixel counts */
    occluded_pixels = (int*)calloc(total_scas, sizeof(int));
    if (occluded_pixels == NULL)
    {
        IAS_LOG_ERROR("Error allocate space for occluded pixel counts");
        params->thread_error = thread_num + 1;
        return ERROR;
    }

    /* Setup the structure for the calling the work queue */
    QUEUE_MESSAGE *message = (QUEUE_MESSAGE*)calloc(1, sizeof(QUEUE_MESSAGE));
    message->image_totals = params->occluded_pixels;
    message->thread_totals = occluded_pixels;
    message->num_of_scas = total_scas;

    while (1)
    {
        int start_samp;
        int end_samp;

        /* Get the next line to process */
        if (get_assigned_sca_line(total_scas, total_lines, &sca, &line_number,
                &work_assigned) != SUCCESS)
        {
            IAS_LOG_ERROR("Error getting assigned line");
            params->thread_error = thread_num + 1;
            return ERROR;
        }

        /* No work was assigned, exit loop */
        if (!work_assigned)
            break;

        /* If other threads have encountered errors, return ERROR */
        if (ias_threadpool_get_worker_return_status(params->threadpool)
                != SUCCESS)
        {
            IAS_LOG_WARNING("Exiting thread %d since another thread "
                "encountered an error", thread_num);
            params->thread_error = thread_num + 1;
            return ERROR;
        }

        /* Get the sample range for the current line by masking it with the
           bounding boxes for this SCA.  This is an optimization to allow
           only checking samples that might actually be covered by this SCA.
           This optimization shaves almost 90% of the run time off. */
        get_sample_range(params->sca_bounds[sca], params->sca_bounds_per_sca,
                line_number, grid->gridbands[band_index][sca].samps, 
                &start_samp, &end_samp);

        /* Calculate occlusion mask for the current line */
        if (calc_terrain_occlusion_mask(&(grid->gridbands[band_index][sca]),
                &(geom_grid->gridbands[0]), params->dem, 
                params->max_elev, line_number, start_samp, end_samp, 
                params->occ, &occ_pixels_cnt) != SUCCESS)
        {
            IAS_LOG_ERROR("Error calculating pixel mask for line %d",
                line_number);
            params->thread_error = thread_num + 1;
            return ERROR;
        }

        /* Add the occluded pixels for this SCA/line to the running total for
           the current thread */
        occluded_pixels[sca] += occ_pixels_cnt;
    }

    /* This thread has finished. Add the occlude pixels for this thread
       to the total occluded pixels */
    if (ias_threadpool_add_to_work_queue(params->threadpool,
            update_occluded_pixel_totals, message) != SUCCESS)
    {
        IAS_LOG_ERROR("Error adding call to the work queue in thread %d",
            thread_num);
        params->thread_error = thread_num + 1;
        return ERROR;
    }

    return SUCCESS;
}

/******************************************************************************
Name:  run_terrain_occlusion

Purpose: Sets up and runs terrain occlusion in multiple threads. 

Returns: SUCCESS or ERROR
******************************************************************************/
int run_terrain_occlusion
(
    const IAS_GRID_TYPE *grid_ptr,      /* I: Pointer to the grid */
    const IAS_IMAGE *dem,               /* I: DEM image data */
    const IAS_GEOM_GRID_TYPE *geom_ptr, /* I: Pointer to the geomgrid */
    double max_elev,                    /* I: Maximum elevation */
    int band_index,                     /* I: Band to be processed */
    int number_of_procs,                /* I: Number of processors to use */
    IAS_IMAGE *occlusion_mask           /* O: Terrain occlusion mask */
)
{
    PARAMS params;      /* parameters for the threads */
    int ii;
    int status;
    int total_scas = grid_ptr->scas_per_band[band_index];
    int occluded_pixels[total_scas];
    unsigned long total_occluded_pixels = 0;

    /* initialize occluded pixel counts to zero */
    for (ii = 0; ii < total_scas; ii++)
    {
        occluded_pixels[ii] = 0;
    }

    /* Get the bounding boxes for all the SCAs */
    params.sca_bounds = get_sca_bounds(grid_ptr, band_index,
            &params.sca_bounds_per_sca);
    if (params.sca_bounds == NULL)
    {
        IAS_LOG_ERROR("Error allocating SCA bounds array");
        return ERROR;
    }

    /* Setup the thread parameters */
    params.grid = grid_ptr;
    params.geom_grid = geom_ptr;
    params.dem = dem;
    params.max_elev = max_elev;
    params.band_index = band_index;
    params.occ = occlusion_mask;
    params.occluded_pixels = occluded_pixels;
    params.thread_error = 0;

    /* Create the work assignment mutex */
    if (IAS_THREAD_CREATE_MUTEX(&work_assignment_mutex) != 0)
    {
        IAS_LOG_ERROR("Error creating work assignment mutex");
        free_sca_bounds(params.sca_bounds);
        return ERROR;
    }

    /* Initialize the threadpool */
    params.threadpool = ias_threadpool_initialize(number_of_procs);
    if (!params.threadpool)
    {
        IAS_LOG_ERROR("Error initializing the thread pool");
        IAS_THREAD_DESTROY_MUTEX(&work_assignment_mutex);
        free_sca_bounds(params.sca_bounds);
        return ERROR;
    }

    /* Initialize the threadpool's work queue */
    if (ias_threadpool_enable_work_queue(params.threadpool) != SUCCESS)
    {
        IAS_LOG_ERROR("Error enabling the work queue");
        IAS_THREAD_DESTROY_MUTEX(&work_assignment_mutex);
        free_sca_bounds(params.sca_bounds);
        return ERROR;
    }

    /* Log the start of processing and the number of threads used */
    IAS_LOG_INFO("Starting terrain occlusion with %d threads", 
            number_of_procs);

    /* Run the build_grid_thread routine using the threadpool */
    status = ias_threadpool_run_function(params.threadpool, occlusion_thread,
            &params);
    if ((status == ERROR) || (params.thread_error > 0))
    {
        IAS_LOG_ERROR("Error encountered while calculating occlusion mask");
        IAS_THREAD_DESTROY_MUTEX(&work_assignment_mutex);
        ias_threadpool_destroy(params.threadpool);
        free_sca_bounds(params.sca_bounds);
        return ERROR;
    }

    /* Clean up the threadpool */
    ias_threadpool_destroy(params.threadpool);

    if (IAS_THREAD_DESTROY_MUTEX(&work_assignment_mutex) != 0)
    {
        IAS_LOG_ERROR("Error destroying work assignment mutex");
        free_sca_bounds(params.sca_bounds);
        return ERROR;
    }

    free_sca_bounds(params.sca_bounds);

    /* Print the occluded pixel counts for each SCA */
    for (ii = 0; ii < total_scas; ii++)
    {
        IAS_LOG_INFO("Occluded pixels in SCA %d: %d", ii, occluded_pixels[ii]);
        total_occluded_pixels += occluded_pixels[ii];
    }

    IAS_LOG_INFO("Total occluded pixels: %lu", total_occluded_pixels);

    return SUCCESS;
}
