/****************************************************************************
NAME:       occlusion

PURPOSE:
Calculate pixels that will be obstructed due to terrain within the image
acquisition extent.

RETURN VALUE:  Type = int
Value       Description
-----       -----------
EXIT_SUCCESS    Successful completion
EXIT_FAILURE    Operation failed

ALGORITHM REFERENCES:
*****************************************************************************/

#include <string.h>
#include <limits.h>
#include <stdlib.h>

#include "terrain_occlusion.h"

#include "ias_const.h"          /* Contains INVALID_ELEVATION definition */
#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "ias_geom_grid.h"
#include "ias_grid_io.h"
#include "ias_miscellaneous.h"

int main(int argc, char *argv[])
{
    IAS_IMAGE dem_image;         /* DEM data */
    int ii;                      /* Counter */
    int out_lines;               /* Output lines */
    int out_samps;               /* Output samples */
    int band_index;              /* Index into grid for band_number */
    IAS_IMAGE occ;               /* Occulation mask */
    int max_elev;              /* Maximum elevation in co-reg DEM */
    int min_elev;              /* Minimum elevation in co-reg DEM */

    char parm_file[PATH_MAX + 1];     /* Input Parameter file name */
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char omf_file[PATH_MAX + 1];      /* Name of the OMF file */
    char dem_file[PATH_MAX + 1];      /* DEM file name */
    char grid_file[PATH_MAX + 1];     /* Grid file name */
    char occ_file[PATH_MAX + 1];      /* Output occlusion file name */
    char geom_grid_file[PATH_MAX + 1]; /* Grid DEM file name */
    int occ_band;                     /* Band from parameters file */
    int number_of_procs;      /* Number of threads to use when processing */
    IAS_SATELLITE_ID satellite_id;    /* Satellite */
    IAS_DATA_TYPE dem_datatype;

    IAS_GEOM_GRID_TYPE geom_grid; /* geom grid structure */
    int geom_grid_band = 1;
    int num_bands = 1;
    int search_for_band = 0;     /* Flag to search for an available band */

    IAS_GRID_TYPE grid;          /* The grid structure */

    /* Initialize the logging library */
    if (ias_log_initialize("terrain_occlusion") != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing logging library");
        exit(EXIT_FAILURE);
    }

    /* Verify a parameter file was provided */
    if (argc != 2)
    {
        IAS_LOG_ERROR("USAGE: terrain_occlusion <parameter file name>");
        exit(EXIT_FAILURE);
    }

    strncpy(parm_file, argv[1], PATH_MAX);
    parm_file[PATH_MAX] = '\0';

    /* Get user's parameters */
    if (get_parms(parm_file, grid_file, occ_file, geom_grid_file, &occ_band,
            work_order_id, dem_file, omf_file, &number_of_procs, 
            &min_elev, &max_elev, &satellite_id) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from file: %s", parm_file);
        exit(EXIT_FAILURE);
    }

    IAS_LOG_INFO("Initiate ...");
    
    /* Initialize the satellite attributes */
    if (ias_sat_attr_initialize(satellite_id) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attributes library");
        exit(EXIT_FAILURE);
    }

    /* Setup the grid structure */
    if (ias_grid_initialize(&grid, 0, 0) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the grid structure");
        exit(EXIT_FAILURE);
    }

    if (ias_grid_header_read(grid_file, &grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading the grid from %s", grid_file);
        exit(EXIT_FAILURE);
    }

    if (ias_grid_band_header_read(grid_file, &grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading the grid band header from %s", grid_file);
        ias_grid_free(&grid);
        exit(EXIT_FAILURE);
    }

    /* If the occlusion band number is zero, first check for band 6.  If it 
       isn't available, look for the first available band */
    if (occ_band == 0)
    {
        search_for_band = 1;
        occ_band = 6;
    }

    /* Convert the band number to an index in the grid structure */
    band_index = ias_sat_attr_convert_band_number_to_index(occ_band);
    if (band_index == ERROR)
    {
        IAS_LOG_ERROR("Converting occlusion band %d to index", occ_band);
        ias_grid_free(&grid);
        exit(EXIT_FAILURE);
    }

    /* Verify the selected occlusion band is available in the grid */
    if (grid.bands_available[band_index] != 1)
    {
        if (!search_for_band)
        {
            /* Not searching for an available band, so this is an error */
            IAS_LOG_ERROR("Selected band %d is not available in grid %s",
                occ_band, grid_file);
            ias_grid_free(&grid);
            exit(EXIT_FAILURE);
        }

        /* Searching for a band, so look for the first available band index */
        for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
        {
            if (grid.bands_available[band_index] == 1)
            {
                /* Found the first available band in the grid */
                break;
            }
        }
        if (band_index >= IAS_MAX_NBANDS)
        {
            IAS_LOG_ERROR("No bands available in the grid %s", grid_file);
            ias_grid_free(&grid);
            exit(EXIT_FAILURE);
        }

        /* Found a valid band index, so get the band number */
        occ_band = ias_sat_attr_convert_band_index_to_number(band_index);
        if (occ_band == ERROR)
        {
            IAS_LOG_ERROR("Converting band index %d to number", band_index);
            ias_grid_free(&grid);
            exit(EXIT_FAILURE);
        }
    }
    if (search_for_band)
    {
        IAS_LOG_INFO("Using band number %d for occlusion calculation",
            occ_band);
    }

    /* Load the DEM */
    if (ias_misc_read_single_band_l1g(dem_file, 1, &dem_datatype, 
            &dem_image) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading L1G DEM image file %s", dem_file);
        return ERROR;
    }

    if (dem_datatype != IAS_I2)
    {
        IAS_LOG_ERROR("DEM does not contain short integer data");
        return ERROR;
    }

    /* Load the grid information for the current band */ 
    if (ias_grid_band_pointers_read(grid_file, &grid, 
                                    occ_band) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading the grid pointers from %s", grid_file);
        ias_grid_free(&grid);
        free(dem_image.data);
        exit(EXIT_FAILURE);
    }

    /* Load geometric grid */
    if (ias_geom_grid_initialize(IAS_MAX_NBANDS, &geom_grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing Geom Grid structure");
        ias_grid_free(&grid);
        free(dem_image.data);
        exit(EXIT_FAILURE);
    }

    /* Load the geometric grid header */
    if (ias_read_geom_grid_header(geom_grid_file, &geom_grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading the grid from %s", geom_grid_file);
        ias_geom_grid_free(&geom_grid_band, &num_bands, &geom_grid);
        ias_grid_free(&grid);
        free(dem_image.data);
        exit(EXIT_FAILURE);
    }

    if (ias_read_geom_grid_band_header(geom_grid_file, &geom_grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading the grid band header from %s", geom_grid_file);
        ias_geom_grid_free(&geom_grid_band, &num_bands, &geom_grid);
        ias_grid_free(&grid);
        free(dem_image.data);
        exit(EXIT_FAILURE);
    }

    if (ias_read_geom_grid_band_pointers(geom_grid_file, &geom_grid_band,
                &num_bands, &geom_grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading band %d from %s", occ_band, geom_grid_file);
        ias_geom_grid_free(&geom_grid_band, &num_bands, &geom_grid);
        ias_grid_free(&grid);
        free(dem_image.data);
        exit(EXIT_FAILURE);
    }

    out_lines = grid.gridbands[band_index][0].lines;
    out_samps = grid.gridbands[band_index][0].samps;
    
    /* Setup the occlusion mask structure */   
    occ.data_type = IAS_I2;
    occ.nl = out_lines;
    occ.ns = out_samps;
    occ.data = (short *)malloc(sizeof(short) * occ.nl * occ.ns);

    if (occ.data == NULL)
    {
        IAS_LOG_ERROR("Allocating occlusion buffer");
        ias_geom_grid_free(&geom_grid_band, &num_bands, &geom_grid);
        ias_grid_free(&grid);
        free(dem_image.data);
        exit(EXIT_FAILURE);
    }

    /* Initialize all pixels in the occlusion mask to not occluded */
    for (ii = 0; ii < occ.nl * occ.ns; ii++)
    {
        ((short*)occ.data)[ii] = NOT_OCCLUDED;
    }

    /* If the min/max elevation values were not in the OMF, calculate the 
       min/max elevation from the DEM file */
    if ((min_elev == INVALID_ELEVATION) || (max_elev == INVALID_ELEVATION))
    { 
        calc_dem_bounds(&dem_image, &min_elev, &max_elev);
    }
        
    IAS_LOG_INFO("Elevation range on unresampled DEM: %d to %d\n", min_elev,
        max_elev);

    /* Run terrain occlusion */
    if (run_terrain_occlusion(&grid, &dem_image, &geom_grid, max_elev, 
            band_index, number_of_procs, &occ) != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating occultation mask");
        ias_geom_grid_free(&geom_grid_band, &num_bands, &geom_grid);
        free(dem_image.data);
        free(occ.data);
        ias_grid_free(&grid);
        exit(EXIT_FAILURE);
    }

    /* Initialize the image data information with the corresponding
       grid information */
    occ.corners      = grid.corners;
    occ.pixel_size_x = grid.gridbands[band_index][0].pixsize;
    occ.pixel_size_y = grid.gridbands[band_index][0].pixsize;

    /* Free the geometric grid */
    ias_geom_grid_free(&geom_grid_band, &num_bands, &geom_grid);
    
    /* Free the DEM data */
    free(dem_image.data);

    /* Write the occlusion mask to the output file */ 
    if (write_l1g(occ_file, &occ, &grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing L1G output file %s", occ_file);
        free(occ.data);
        ias_grid_free(&grid);
        exit(EXIT_FAILURE);
    }

    /* Free grid data for the current band */
    if (ias_grid_free_band(&grid, occ_band) != SUCCESS)
    {
        IAS_LOG_ERROR("Freeing the grid data for band %d", occ_band);
        free(occ.data);
        ias_grid_free(&grid);
        exit(EXIT_FAILURE);
    }

    free(occ.data);
    
    /* Update the OMF with the occlusion filename */
    if (update_omf(omf_file, occ_file) != SUCCESS)
    {
        IAS_LOG_ERROR("Updating OMF file");
        ias_grid_free(&grid);
        exit(EXIT_FAILURE);
    }

    if (ias_grid_free(&grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Freeing the grid");
        exit(EXIT_FAILURE);
    }

    /* Write completion message */
    IAS_LOG_INFO("Successful completion.");

    return EXIT_SUCCESS;
}

