#ifndef _OCCLUSION_H
#define _OCCLUSION_H

#include "ias_types.h"
#include "ias_structures.h"
#include "ias_geom_grid.h"
#include "ias_grid.h"
#include "ias_math.h"

#define NOT_OCCLUDED  0  /* Value of non-occluded pixel in occlusion mask */
#define OCCLUDED      1  /* Value of occluded pixel in occlusion mask */


/* Define a type for tracking the SCA bounding boxes */
typedef struct bounding_box
{
    IAS_LINE_SEGMENT edges[4];  /* Four edge line segments define the bounding
                                   box */
} BOUNDING_BOX;

/** Prototypes for all the terrain occlusion functions **/

/* calc_dem_bounds.c */
void calc_dem_bounds
(
    const IAS_IMAGE *dem,    /* I: DEM image data */
    int *elev_min,           /* O: Minimum elevation in DEM */
    int *elev_max            /* O: Maximum elevation in DEM */
);

/* get_parms.h */
int get_parms
(
    const char *parm_file,         /* I: Name of the Parameter File */
    char *grid_file,               /* O: Name of grid file */
    char *occlusion_file,          /* O: Name of occlusion output file */
    char *geom_grid_file,          /* O: Name of grid DEM file */
    int *occlusion_band,           /* O: Band to process */
    char *wo_id,                   /* O: Work Order Id */
    char *dem_file,                /* O: Name of unresampled DEM file */
    char *omf_file,                /* O: OMF file name */
    int *number_of_procs,          /* O: Number of threads to use */
    int *min_elevation,            /* O: Minimum elevation in DEM file */
    int *max_elevation,            /* O: Maximum elevation in DEM file */
    IAS_SATELLITE_ID *satellite_id /* O: Satellite Id from OMF */
);

/* occlusion_threading.c */
int run_terrain_occlusion
(
    const IAS_GRID_TYPE *grid_ptr,      /* I: Pointer to grid band */
    const IAS_IMAGE *dem,               /* I: DEM image data */
    const IAS_GEOM_GRID_TYPE *geom_ptr, /* I: Pointer to the geomgrid */
    double max_elev,                    /* I: Maximum elevation */
    int band_index,                     /* I: Current SCA */
    int number_of_procs,                /* I: Number of processors to use */
    IAS_IMAGE *occlusion_mask           /* O: Terrain occlusion mask */
);

/* terrain_occlusion_mask.c */
int calc_terrain_occlusion_mask
(
    const IAS_GRID_BAND_TYPE *grid_band_ptr, /* I: pointer to grid band */
    const IAS_GEOM_GRID_BAND_TYPE *geom_band_ptr, /* I: pointer to geom band */
    const IAS_IMAGE *dem_ptr,  /* I: DEM image data */
    double max_elev,       /* I: Maximum elevation */
    int check_line,        /* I: Current line */
    int begin_samp,        /* I: Sample to start calculating the mask */
    int end_samp,          /* I: Sample to end calculating the mask */
    IAS_IMAGE *occ_mask,   /* O: Terrain occlusion mask */
    int *occ_pixels        /* O: Number of occluded pixels found */
);

/* write_l1g.c */
int write_l1g
(
    const char *occ_filename,    /* I: Terrain occlusion filename */
    const IAS_IMAGE *occ_mask,   /* I: Terrain occlusion mask data */
    const IAS_GRID_TYPE *grid    /* I: Grid used to generate occlusion mask */
);

/* update_omf.c */
int update_omf
(
    const char *omf_file_name,        /* I: OMF file name */
    const char *occlusion_file_name   /* I: Terrain occlusion file name */
);

/* filter_to_sca_bounds.c */
BOUNDING_BOX **get_sca_bounds
(
    const IAS_GRID_TYPE *grid_ptr,  /* I: Pointer to the grid */
    int band_index,                 /* I: Band index to use */
    int *boxes_per_sca              /* O: Bounding box count per sca */
);

void free_sca_bounds
(
    BOUNDING_BOX **sca_bounds   /* I: SCA bounds to free */
);

void get_sample_range
(
    const BOUNDING_BOX *bounds,     /* I: List of bounds to check */
    int bounds_count,               /* I: Number of bounding boxes in bounds
                                          (each one with 4 entries) */
    int line_number,                /* I: Line number to check */
    int max_samps,                  /* I: Maximum number of samples */
    int *start_sample,              /* O: Start sample found */
    int *end_sample                 /* O: End sample found */
);

#endif
