/******************************************************************************
NAME:   terrain_occlusion_mask

PURPOSE: Contains the main driver function for terrain occlusion and supporting
functions.

ALGORITHM REFERENCES:
    Terrain Occlusion ADD
******************************************************************************/

#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "terrain_occlusion.h"

#include "ias_structures.h"
#include "ias_logging.h"
#include "ias_grid.h"
#include "ias_grid_io.h"
#include "ias_geom_grid.h"
#include "ias_miscellaneous.h"

#define NO_INPUT_LS    -100 /* Status for no input line/sample found */
/* Limit to the number of warning messages output for being out of the source
   DEM since that can happen in some areas of the world, such as on the edge of
   Antarctica using the RAMP DEM. */
#define WARNING_MESSAGE_LIMIT 10

/******************************************************************************
Name: get_elevation_from_dem

Purpose: Map output pixel location to the input space pixel location and
    retreive the elevation from the unresampled DEM.

Returns: void
******************************************************************************/
static void get_elevation_from_dem
(
    const IAS_GEOM_GRID_BAND_TYPE *geom_band_ptr, /* I: Geom band pointer */
    const IAS_IMAGE *dem,                  /* I: Unresampled DEM pointer */
    int oline,                             /* I: Output line location */
    int osamp,                             /* I: Ouptut sample location */
    double scaling_factor,                 /* I: Resampling scaling factor */
    double *elevation                      /* O: Elevation from DEM */
)
{
    double dem_line;
    double dem_samp;
    int grid_cell_index;
    IAS_COEFFICIENTS *coeffs_ptr;
    int iline;
    int isamp;
    int loc;
    double temp_elev;
    double dl;
    double ds;
    static int message_count = 0;   /* Count the number of messages output */
    short *dem_data;

    /* Scale the output line/sample (which are in L1G band coordinates) to 
       match the pixel size of the resampled DEM */
    double doline = oline * scaling_factor;
    double dosamp = osamp * scaling_factor;

    int grid_cols = geom_band_ptr->grid_cols;
    int grid_row_index = (int)doline / geom_band_ptr->cell_lines;
    int grid_col_index = (int)dosamp / geom_band_ptr->cell_samps;

    /* Clamp the row index to fall in the grid */
    if (grid_row_index < 0)
        grid_row_index = 0;
    else if (grid_row_index >= geom_band_ptr->grid_rows)
        grid_row_index = geom_band_ptr->grid_rows - 1;

    /* Clamp the column index to fall in the grid */
    if (grid_col_index < 0)
        grid_col_index = 0;
    else if (grid_col_index >= grid_cols)
        grid_col_index = grid_cols - 1;

    /* Calculate the grid cell index */
    grid_cell_index = (grid_row_index * grid_cols) + grid_col_index;

    /* Map the output lines/sample back to the line/sample in the source DEM */
    coeffs_ptr = &geom_band_ptr->reverse_coeffs[grid_cell_index];
    dem_samp = coeffs_ptr->a[0] + coeffs_ptr->a[1] * dosamp +
               coeffs_ptr->a[2] * doline + coeffs_ptr->a[3] * dosamp * doline;
    dem_line = coeffs_ptr->b[0] + coeffs_ptr->b[1] * dosamp +
               coeffs_ptr->b[2] * doline + coeffs_ptr->b[3] * dosamp * doline;

    /* Make sure the calculated line will stay inside the image when the
       bilinear interpolation is done */
    if (dem_line < 0.0 || dem_line >= dem->nl - 1.0)
    {
        if (message_count < WARNING_MESSAGE_LIMIT)
        {
            /* Increment the message count.  This code runs in multiple threads
               and we're not locking access to this count, but that is OK since
               the number of messages output is not critical.  The goal here
               is to just prevent millions of messages being output in a few
               places in the world where the DEM coverage isn't complete (i.e.
               off the coast of Antarctica).  The count is incremented before
               the message is output to reduce the chance of the disabling 
               message being output multiple times. */
            message_count++;

            IAS_LOG_WARNING("Line value for output line/sample (%f, %f) "
                "outside DEM area (0, %d): %f", doline, dosamp, dem->nl - 1,
                dem_line);

            if (message_count >= WARNING_MESSAGE_LIMIT)
            {
                IAS_LOG_WARNING("Disabling further warnings about being "
                  "outside the DEM area (may see this message more than once)");
            }
        }

        if (dem_line < 0.0)
            dem_line = 0.0;
        else if (dem_line >= dem->nl - 1.0)
        {
            /* For the bottom edge of the image, make sure it is just slightly
               inside the image to make sure the bilinear interpolation doesn't
               go off the edge */
            dem_line = dem->nl - 1.0000001;
        }
    }

    /* Make sure the calculated sample will stay inside the image when the
       bilinear interpolation is done */
    if (dem_samp < 0.0 || dem_samp >= dem->ns - 1.0)
    {
        if (message_count < WARNING_MESSAGE_LIMIT)
        {
            /* Increment the message count.  This code runs in multiple threads
               and we're not locking access to this count, but that is OK since
               the number of messages output is not critical.  The goal here
               is to just prevent millions of messages being output in a few
               places in the world where the DEM coverage isn't complete (i.e.
               off the coast of Antarctica).  The count is incremented before
               the message is output to reduce the chance of the disabling 
               message being output multiple times. */
            message_count++;

            IAS_LOG_WARNING("Sample value for output line/sample (%f, %f) "
                "outside DEM area (0, %d): %f", doline, dosamp, dem->ns - 1,
                dem_samp);

            if (message_count >= WARNING_MESSAGE_LIMIT)
            {
                IAS_LOG_WARNING("Disabling further warnings about being "
                  "outside the DEM area (may see this message more than once)");
            }
        }

        if (dem_samp < 0.0)
            dem_samp = 0.0;
        else if (dem_samp >= dem->ns - 1.0)
        {
            /* For the right edge of the image, make sure it is just slightly
               inside the image to make sure the bilinear interpolation doesn't
               go off the edge */
            dem_samp = dem->ns - 1.0000001;
        }
    }

    /* Perform the bilinear interpolation of the elevation */
    iline = (int)dem_line;
    isamp = (int)dem_samp;
    dem_data = (short *)dem->data;

    dl = dem_line - iline;
    ds = dem_samp - isamp;

    loc = iline * dem->ns + isamp;
    temp_elev = (1.0 - ds) * (1.0 - dl) * dem_data[loc] +
                       ds  * (1.0 - dl) * dem_data[loc + 1] +
                (1.0 - ds) *      dl  * dem_data[loc + dem->ns] +
                       ds  *      dl  * dem_data[loc + dem->ns + 1];
    
    *elevation = temp_elev;
}

/******************************************************************************
Name: map_to_input_occlusion

Purpose: Maps output pixel location to input location.

Returns: SUCCESS, ERROR, or NO_INPUT_LS (if output line/sample maps to invalid
    input line/sample)
******************************************************************************/
static int map_to_input_occlusion
(
    const IAS_GRID_BAND_TYPE *grid_band_ptr, /* I: pointer to grid band */
    int num_rows,          /* I: Number of rows in grid */
    int num_cols,          /* I: Number of columns in grid */
    double elev,           /* I: Elevation of point */
    double oline,          /* I: Output line location */
    double osamp,          /* I: Output sample location */
    double *input_line,    /* O: Input line location */
    double *input_samp     /* O: Input sample location */
)
{
    int status;            /* Return status */
    int row, col;          /* Row and column index into grid */
    double d_iline;        /* Input line location */
    double d_isamp;        /* Input sample location */

    /* Map output point to input location and find grid cell number */
    status = ias_grid_3d_ols2ils(grid_band_ptr, oline, osamp,
                                 &d_iline, &d_isamp, elev);
    if (status == SUCCESS)
    {
        row = (int)(d_iline * grid_band_ptr->inv_cell_lines);
        col = (int)(d_isamp * grid_band_ptr->inv_cell_samps);

        if ((row < 0) || (row >= num_rows) || (col < 0) || (col >= num_cols))
        {
            /* Return status for out of bounds input line/sample found */
            return NO_INPUT_LS;
        }
    }
    else
    {
        if (status == IAS_GRID_ILS_NOT_CALCULATED)
        {
           return NO_INPUT_LS;
        }
        else
        {
            return ERROR;
        }
    }

    *input_line = d_iline;
    *input_samp = d_isamp;

    return SUCCESS;
}

/******************************************************************************
Name: calc_occ_line_eq

Purpose:
 Given two points (coline,cosamp) and (moline,mosamp) find
   the parametric equation for a line joining the two points.

   line = line0 + t * g
   samp = samp0 + t * h

   where the offsets (line0,samp0) and scales (g,h) are determined
   from the two endpoints.  t=0 and t=1 correspond to two
   endpoints. 

Returns: void
******************************************************************************/
static void calc_occ_line_eq
(
    double coline,  /* I: Current image line location */
    double cosamp,  /* I: Current image sample location */
    double moline,  /* I: Maximum elevation line location */
    double mosamp,  /* I: Maximum elevation line location */
    double *lscale, /* O: Line scale for parametric eq. */
    double *lconst, /* O: Line offset for parametric eq. */
    double *sscale, /* O: Sample scale for parametric eq. */
    double *sconst, /* O: Sample offset for parametric eq. */
    double *step_incr   /* O: Increment for walking line between the original 
                              output point and the max elevation point */
)
{
    double temp_dist_sq;
    double diff_line;
    double diff_samp;
    double max_diff;

    *lscale = moline - coline;
    *lconst = coline;

    *sscale = mosamp - cosamp;
    *sconst = cosamp;

    diff_line = fabs(coline - moline);
    diff_samp = fabs(cosamp - mosamp);
    
    temp_dist_sq = (diff_samp * diff_samp) + (diff_line * diff_line);
    if (temp_dist_sq < 1.0)
        temp_dist_sq = 1.0;

    /* Use the maximum of the difference in the line and sample directions.
       Also, if that turns out to be less than 1.0, use 1.0 instead */
    max_diff = diff_samp;
    if (max_diff < diff_line)
        max_diff = diff_line;
    if (max_diff < 1.0)
        max_diff = 1.0;

    *step_incr = max_diff / temp_dist_sq;
}

/******************************************************************************
Name: calc_terrain_occlusion_mask

Purpose: Main driver for determining terrain occlusion mask.

Returns: SUCCESS or ERROR
******************************************************************************/
int calc_terrain_occlusion_mask
(
    const IAS_GRID_BAND_TYPE *grid_band_ptr, /* I: pointer to grid band */
    const IAS_GEOM_GRID_BAND_TYPE *geom_band_ptr, /* I: pointer to geom band */
    const IAS_IMAGE *dem_ptr,  /* I: DEM image data */
    double max_elev,       /* I: Maximum elevation */
    int check_line,        /* I: Current line */
    int begin_samp,        /* I: Sample to start calculating the mask */
    int end_samp,          /* I: Sample to end calculating the mask */
    IAS_IMAGE *occ_mask,   /* O: Terrain occlusion mask */
    int *occ_pixels        /* O: Number of occluded pixels found */
)
{
    int ii;                /* Counter */
    int num_steps;         /* Number of steps */
    double input_line;     /* Input line location */
    double input_samp;     /* Input sample location */
    int cur_line;          /* Current output line location */
    int cur_samp;          /* Current output sample location */
    double moline;         /* Max elev output line location */
    double mosamp;         /* Max elev output sample location */
    double elev;           /* Current elevation */
    int num_cols;          /* Number of cols in grid */
    int num_rows;          /* Number of rows in grid */
    int inl;               /* Number of samples in input image */
    int ins;               /* Number of lines in input image */
    double lscale;         /* Scale connecting output line locs */
    double sscale;         /* Scale connecting output sample locs */
    double lconst;         /* Constant connecting output line locs */
    double sconst;         /* Constant connecting output sample locs */
    double elev_threshold; /* Elevation threshold for occluded pixel */
    double intersect_line; /* Check point line value when walking line */
    double intersect_samp; /* Check point samp value when walking line */
    double occ_count = 0;  /* To keep a count of occluded pixels */
    double cur_elev;       /* Elevation of point along line */
    double step_incr;      /* Increment for walking line between the original 
                              output point and the max elevation point */
    double cur_incr;       /* Currect increment size when walking line */
    int status;

    double scaling_factor = 
        grid_band_ptr->pixsize / geom_band_ptr->pixsize_line; 

    num_cols = grid_band_ptr->ngrid_samps - 1;
    num_rows = grid_band_ptr->ngrid_lines - 1;

    inl = grid_band_ptr->in_lines[grid_band_ptr->ngrid_lines - 1];
    ins = grid_band_ptr->in_samps[grid_band_ptr->ngrid_samps - 1];
    
    /* Loop on output line and samples */
    cur_line = check_line;

    for (cur_samp = begin_samp; cur_samp < end_samp; cur_samp++)
    {
        /* Get elevation from the unresampled DEM */
        get_elevation_from_dem(geom_band_ptr, dem_ptr, cur_line, cur_samp, 
            scaling_factor, &elev);
    
        /* Map the output to input pixel location */
        status = map_to_input_occlusion(grid_band_ptr, num_rows, num_cols,
                                    elev, cur_line, cur_samp, &input_line, 
                                    &input_samp);

        if (status == SUCCESS)
        {
            if (input_line > 0 && input_line < inl - 1 &&
                    input_samp > 0 && input_samp < ins - 1)
            {
                /* Map input line and sample to new output location
                   based on maximum elevation in DEM. */
                if (ias_grid_ils2ols_at_elevation(grid_band_ptr, max_elev,
                        input_line, input_samp, &moline, &mosamp) != SUCCESS)
                {
                    IAS_LOG_ERROR("Mapping point %f,%f elev %f for "
                        "occlusion mask", input_line, input_samp, max_elev);
                    return ERROR;
                }

                if (moline != cur_line && mosamp != cur_samp)
                {
                    calc_occ_line_eq(cur_line, cur_samp, moline, 
                                     mosamp, &lscale, &lconst, &sscale, 
                                     &sconst, &step_incr);

                    num_steps = (int)ceil(1.0 / step_incr);

                    for (ii = 0; ii <= num_steps; ii++)
                    {
                        cur_incr = ii * step_incr;

                        /* Calculate the point of intersection */
                        intersect_line = lconst + (cur_incr * lscale);
                        intersect_samp = sconst + (cur_incr * sscale);

                        /* Find occlusion elevation threshold */
                        if (intersect_line == cur_line 
                                && intersect_samp == cur_samp)
                        {
                            elev_threshold = max_elev;
                        }
                        else
                        {
                            elev_threshold = cur_incr * max_elev 
                                + (1.0 - cur_incr) * elev;
                        }

                        intersect_line = (int)intersect_line + 1;
                        intersect_samp = floor(intersect_samp + 0.5);

                        get_elevation_from_dem(geom_band_ptr, dem_ptr,
                            intersect_line, intersect_samp, scaling_factor,
                            &cur_elev);

                        /* Determine if pixel is masked */
                        if (cur_elev > elev_threshold)
                        {
                            ((short *)occ_mask->data)[cur_line * 
                                occ_mask->ns + cur_samp] = OCCLUDED;
                            occ_count++;
                            break;
                        }
                    }
                }
            }
        }
        else if (status == ERROR)
        {
            IAS_LOG_ERROR("Mapping point (%d, %d) at elevation %f back "
                "to input line/sample", cur_line, cur_samp, elev);
            return ERROR;
        }
        else if (status == NO_INPUT_LS)
        {
            /* Nothing to do in case of NO_INPUT_LS */
        }
    }

    /* Return the total occluded pixels for this line */
    *occ_pixels = occ_count;

    return SUCCESS;
}
