/******************************************************************************
NAME:    write_l1g

PURPOSE: Write the terrain occlusion mask to an L1G format file.

RETURN VALUE:      Type = int
Value    Description
-------  --------------
SUCCESS  L1G was successfully created
ERROR    Failed to create L1G

******************************************************************************/

/* System includes */
#include <string.h>

/* IAS includes */
#include "terrain_occlusion.h"

#include "ias_structures.h"
#include "ias_logging.h"
#include "ias_l1g.h"
#include "ias_geo.h"
#include "ias_grid.h"
#include "ias_miscellaneous.h"

int write_l1g
(
    const char *occ_filename,   /* I: Terrain occlusion filename */
    const IAS_IMAGE *occ_mask,  /* I: Terrain occlusion mask data */
    const IAS_GRID_TYPE *grid   /* I: Grid used to generate occlusion mask */
)
{
    int ii;
    int status;
    int unit_number;                  /* Numerical value corresponding to
                                         grid 'units' name */
    L1GIO *image_ptr = NULL;          /* L1G image data */
    L1G_BAND_IO *band_ptr = NULL;     /* L1G band data */
    IAS_L1G_FILE_METADATA file_meta;  /* L1G image metadata */
    IAS_L1G_BAND_METADATA band_meta;  /* L1G band metadata */
    IAS_PROJECTION proj_info;

    /* Create local variables to get rid of compiler warnings caused by 
       calling libary function with reference to const occ_mask structure */
    int data_type = occ_mask->data_type;
    int num_lines = occ_mask->nl;
    int num_samples = occ_mask->ns;

    /* Some variables to make function calls easier to read */
    int band_number = 1;   /* write to band one */
    int num_of_scas = 1;   /* write mask to one SCA */
    int sca_index = 0;     /* index when calling L1G library functions */
    int start_line = 0;   
    int start_sample = 0;

    /* Initialize the metadata structures */
    memset(&file_meta, 0, sizeof(IAS_L1G_FILE_METADATA));
    memset(&band_meta, 0, sizeof(IAS_L1G_BAND_METADATA));
    
    IAS_LOG_INFO("Writing occlusion file %s", occ_filename);

    /* Create the occlusion mask output file */
    image_ptr = ias_l1g_open_image(occ_filename, IAS_WRITE);

    /* Verify the image was created */
    if (image_ptr == NULL)
    {
        IAS_LOG_ERROR("Creating occlusion output file %s", occ_filename);
        return ERROR;
    }

    /* Open the occlusion mask band */
    band_ptr = ias_l1g_open_band(image_ptr, band_number, &data_type, 
        &num_of_scas, &num_lines, &num_samples);

    /* Verify the band was added to the image */
    if (band_ptr == NULL)
    {
        IAS_LOG_ERROR("Adding occlusion mask to output image %s", 
            occ_filename);
        ias_l1g_close_image(image_ptr);
        return ERROR;
    }

    /* Write the occlusion mask to the L1G band */
    status = ias_l1g_write_image(band_ptr, sca_index, start_line, start_sample,
                occ_mask->nl, occ_mask->ns, occ_mask->data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing band %d to occlusion mask file %s",
            band_number, occ_filename);
        ias_l1g_close_band(band_ptr);
        ias_l1g_close_image(image_ptr);
        return ERROR;
    }

    /* Close the image band */
    status = ias_l1g_close_band(band_ptr);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Closing band %d of occlusion mask file %s", band_number,
            occ_filename);
        ias_l1g_close_image(image_ptr);
        return ERROR;
    }

    /* Populate the projection information in the image level metadata using 
       information from the grid */
    file_meta.projection_code = grid->code;
    file_meta.zone_code = grid->zone;
    strcpy(file_meta.datum, grid->datum); 
    file_meta.spheroid_code = grid->spheroid;
    strcpy(file_meta.projection_units, grid->units);
    
    for (ii = 0; ii < IAS_PROJ_PARAM_SIZE; ii++)
    {
        file_meta.projection_parameters[ii] = grid->projprms[ii];
    }

    /* Populate the terrain occlusion band metadata */
    band_meta.band_number = band_number;

    strncpy(band_meta.band_name, "TERRAIN OCCLUSION", IAS_BAND_NAME_SIZE);
    band_meta.band_name[IAS_BAND_NAME_SIZE-1] = '\0';
    
    band_meta.upper_left_x = grid->corners.upleft.x;
    band_meta.upper_left_y = grid->corners.upleft.y;
    band_meta.upper_right_x = grid->corners.upright.x;
    band_meta.upper_right_y = grid->corners.upright.y;
    band_meta.lower_right_x = grid->corners.loright.x;
    band_meta.lower_right_y = grid->corners.loright.y;
    band_meta.lower_left_x = grid->corners.loleft.x;
    band_meta.lower_left_y = grid->corners.loleft.y;

    band_meta.pixel_range_valid = 1;
    band_meta.minimum_pixel_value = NOT_OCCLUDED;
    band_meta.maximum_pixel_value = OCCLUDED;

    /* Write the file metadata */
    status = ias_l1g_set_file_metadata(image_ptr, &file_meta);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing file metadata to occlusion file %s", 
            occ_filename);
        ias_l1g_close_image(image_ptr);
        return ERROR;
    }

    /* Write the band metadata */
    status = ias_l1g_set_band_metadata(image_ptr, &band_meta, band_number);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing band metadata to occlusion file %s",
            occ_filename);
        ias_l1g_close_image(image_ptr);
        return ERROR;
    }

    /* Close the occlusion file */
    status = ias_l1g_close_image(image_ptr);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Closing the occlusion file %s", occ_filename);
        return ERROR;
    }

    /* Write the ENVI header file  */
    status = ias_geo_get_units(grid->units, &unit_number);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Converting 'units' %s to numeric value",
            grid->units);
        return ERROR;
    }

    ias_geo_set_projection(grid->code, grid->zone, unit_number,
        grid->spheroid, grid->projprms, &proj_info);

    if (ias_misc_write_envi_header(occ_filename, &proj_info,
        "Terrain occlusion image file", occ_mask->nl, occ_mask->ns, 1,
        occ_mask->corners.upleft.x, occ_mask->corners.upleft.y,
        occ_mask->pixel_size_x, occ_mask->pixel_size_y,
        NULL, IAS_I2) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing ENVI header file");
        return ERROR;
    }

    return SUCCESS;
}
