/* Standard Library includes */
#include <math.h>

/* IAS Libray Includes */
#include "ias_const.h"
#include "ias_l1g.h"
#include "ias_logging.h"

/* Local Includes */
#include "convert_geotiff_to_l1g.h"

/* Local Defines */
#define CORNER_TOLERANCE .1 

/******************************************************************************
Name: check_metadata

Purpose: Performs a consistency check for all of the copied L1G meta data

Returns: Type: integer 
    SUCCESS / ERROR
******************************************************************************/
int check_metadata
(
    const IAS_L1G_FILE_METADATA *base_file_meta, /* I: Base file metadata */
    const IAS_L1G_BAND_METADATA *base_band_meta, /* I: Base band metadata */
    const IAS_L1G_FILE_METADATA *l1g_file_meta,  /* I: L1G file metadata */
    const IAS_L1G_BAND_METADATA *l1g_band_meta   /* I: L1G band metadata */
)
{
    /* Check the projection information */
    if (l1g_file_meta->projection_code != base_file_meta->projection_code)
    {
        IAS_LOG_ERROR("Projections do not match");
        return ERROR;
    }
   
    /* Check the zone codes */
    if (l1g_file_meta->zone_code != base_file_meta->zone_code)
    {
        IAS_LOG_ERROR("Zone codes do not match");
        return ERROR;
    }

    /* Check that the corners fall within the allowed tolerance */
    if (fabs(l1g_band_meta->upper_left_y - base_band_meta->upper_left_y) 
        > CORNER_TOLERANCE 
        || fabs(l1g_band_meta->upper_left_x - base_band_meta->upper_left_x)
        > CORNER_TOLERANCE
        || fabs(l1g_band_meta->upper_right_y - base_band_meta->upper_right_y) 
        > CORNER_TOLERANCE
        || fabs(l1g_band_meta->upper_right_x - base_band_meta->upper_right_x)
        > CORNER_TOLERANCE
        || fabs(l1g_band_meta->lower_left_y - base_band_meta->lower_left_y)
        > CORNER_TOLERANCE
        || fabs(l1g_band_meta->lower_left_x - base_band_meta->lower_left_x)
        > CORNER_TOLERANCE
        || fabs(l1g_band_meta->lower_right_y - base_band_meta->lower_right_y) 
        > CORNER_TOLERANCE
        || fabs(l1g_band_meta->lower_right_x - base_band_meta->lower_right_x) 
        > CORNER_TOLERANCE)
    {
        IAS_LOG_WARNING("Band image corners do not match within tolerance "
            "of %f", CORNER_TOLERANCE);
    }
    
    return SUCCESS;
}
