/*****************************************************************************
NAME:          convert_geotiff_to_l1g (main)

PURPOSE: Main application to convert a single-band GeoTIFF file to the IAS
         L1G file format

NOTES:
Uses a local implementation of a set of functions to do the GeoTIFF I/O, as
well as the IAS L1G I/O library. The local GeoTIFF I/O functions are setup in
the manner they are to lay the ground-work for them potentially becoming a
new IAS I/O library in the future, but that is a later implementation step.

RETURNS:
Type = int
Value         Description
-----         -----------
EXIT_SUCCESS  Successful completion
EXIT_FAILURE  Operation failed
******************************************************************************/
/* Standard Library Includes */
#include <stdlib.h>
#include <string.h>

/* IAS Libary Includes */
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_l1g.h"

/* Local Includes */
#include "geotiff_io.h"
#include "convert_geotiff_to_l1g.h"

/* Local Defines */
#define MAX_FILES 30

int main
(
    int argc, 
    char *argv[]
)
{
    L1GIO *l1g_file = NULL;                 /* L1G file handle */
    char *l1g_filename;                     /* Output L1G format filename */
    char *gtiff_filename[MAX_FILES];        /* Input GeoTIFF filenames */
    int total_bands;                        /* Total band files to process */
    int band_index;                         /* Loop variable */
    int band_number;                        /* Starting band number */
    IAS_L1G_FILE_METADATA base_file_meta;   /* L1G file metadata */
    IAS_L1G_BAND_METADATA base_band_meta;   /* L1G band metadata */

    /* Prepare log file */
    if (ias_log_initialize("convert_geotiff_to_l1g") != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the logging library");
        exit(EXIT_FAILURE);
    }

    /* Read the parameters from the command line */
    if (read_parameters(argc, argv, MAX_FILES, gtiff_filename, &l1g_filename, 
        &total_bands, &band_number) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters");
        exit(EXIT_FAILURE);
    }

    /* Create the L1G output file */
    l1g_file = ias_l1g_open_image(l1g_filename, IAS_WRITE);
    if (l1g_file == NULL)
    {
        IAS_LOG_ERROR("Creating L1G output file %s", l1g_filename);
        exit(EXIT_FAILURE);
    }

    /* Loop through all of the provided GeoTIFF files */
    IAS_LOG_INFO("GeoTIFF to L1G Converter started");
    for (band_index = 0; band_index < total_bands; band_index++, band_number++)
    {
        GEOTIFFIO *gtiff_file = NULL;          /* GeoTIFF file handle */
        GEOTIFF_BAND_IO *gtiff_band = NULL;    /* GeoTIFF band data handle */
        GEOTIFF_FILE_METADATA gtiff_file_meta; /* GeoTIFF file-level metadata */
        GEOTIFF_BAND_METADATA gtiff_band_meta; /* GeoTIFF band-level metadata */
        L1G_BAND_IO *l1g_band = NULL;          /* L1G band handle */
        IAS_L1G_FILE_METADATA l1g_file_meta;   /* L1G file metadata */
        IAS_L1G_BAND_METADATA l1g_band_meta;   /* L1G band metadata */
        IAS_DATA_TYPE dtype;                   /* Data type of image data */
        int num_lines;                         /* Number of lines */
        int num_samples;                       /* Number of samples per line */
        int index;                             /* Generic loop variable */
        int sca_count = 1; /* GeoTIFF format cannot handle separated SCAs */
        int sca_index = 0; /* so SCA count will always be 1 and band_index 0 */
        const char *cur_filename = gtiff_filename[band_index];

        IAS_LOG_INFO("Converting band number %d", band_number);

        /* Open the GeoTIFF file to convert */
        gtiff_file = geotiff_open_file(cur_filename, IAS_READ);
        if (gtiff_file == NULL)
        {
            IAS_LOG_ERROR("Opening source GeoTIFF file %s", cur_filename);
            ias_l1g_close_image(l1g_file);
            exit(EXIT_FAILURE);
        }

        /* Get the file level metadata from the GeoTIFF file */
        if (geotiff_get_file_metadata(gtiff_file, &gtiff_file_meta) != SUCCESS)
        {
            IAS_LOG_ERROR("Getting file metadata from %s", cur_filename);
            geotiff_close_file(gtiff_file);
            ias_l1g_close_image(l1g_file);
            exit(EXIT_FAILURE);
        }

        /* Some values are not able to be set in the file metadata structure in
           a way that we want embedded into the geotiff_io_tools since that is
           being prepared to be able to convert to a separate GeoTIFF I/O lib.
           Get those values using a routine local to this application. */
        if (get_conversion_file_metadata(gtiff_file, &gtiff_file_meta)
            != SUCCESS)
        {
            IAS_LOG_ERROR("Getting file metadata from %s", cur_filename);
            geotiff_close_file(gtiff_file);
            ias_l1g_close_image(l1g_file);
            exit(EXIT_FAILURE);
        }

        /* Debugging info */
        IAS_LOG_DEBUG("GeoTIFF file metadata contents:");
        IAS_LOG_DEBUG("<-------------------------");
        IAS_LOG_DEBUG("  projection code: %d", gtiff_file_meta.projection_code);
        IAS_LOG_DEBUG("  zone: %d", gtiff_file_meta.zone_code);
        IAS_LOG_DEBUG("  datum: %s", gtiff_file_meta.datum);
        IAS_LOG_DEBUG("  spheroid code: %d", gtiff_file_meta.spheroid_code);
        IAS_LOG_DEBUG("  projection units: %s", 
            gtiff_file_meta.projection_units);
        for (index = 0; index < IAS_PROJ_PARAM_SIZE; index++)
        {
            IAS_LOG_DEBUG("  projection param[%d]: %lf", index,
                gtiff_file_meta.projection_parameters[index]);
        }
        IAS_LOG_DEBUG("  WRS path: %d", gtiff_file_meta.wrs_path);
        IAS_LOG_DEBUG("  WRS row: %d", gtiff_file_meta.wrs_row);
        IAS_LOG_DEBUG("  spacecraft: %s", gtiff_file_meta.spacecraft);
        IAS_LOG_DEBUG("  capture direction: %s",
            gtiff_file_meta.capture_direction);
        IAS_LOG_DEBUG("  capture date: %s", gtiff_file_meta.capture_date);
        IAS_LOG_DEBUG("------------------------->");

        /* Open the band - this is another place that the code depends on the
           assumption that we are only supporting single-band GeoTIFF files.
           Would that need change, then major portions of this code would need
           to be wrapped in a loop to open and read the GeoTIFF band, get the
           GeoTIFF band-level metadata, open the L1G band and write the data
           and band-level metadata to it. */
        gtiff_band = geotiff_open_band(gtiff_file, 1, &dtype, &sca_count, 
            &num_lines, &num_samples);
        if (gtiff_band == NULL)
        {
            IAS_LOG_ERROR("Identifying band information in %s", cur_filename);
            geotiff_close_file(gtiff_file);
            ias_l1g_close_image(l1g_file);
            exit(EXIT_FAILURE);
        }
        gtiff_band->band_number = band_number;

        /* Get the band level metadata from the GeoTIFF file */
        if (geotiff_get_band_metadata(gtiff_band, &gtiff_band_meta) != SUCCESS)
        {
            IAS_LOG_ERROR("Getting band metadata from %s", cur_filename);
            geotiff_close_band(gtiff_band);
            geotiff_close_file(gtiff_file);
            ias_l1g_close_image(l1g_file);
            exit(EXIT_FAILURE);
        }

        /* Debugging info */
        IAS_LOG_DEBUG("GeoTIFF band metadata contents:");
        IAS_LOG_DEBUG("<-------------------------");
        IAS_LOG_DEBUG("  band number: %d", gtiff_band_meta.band_number);
        IAS_LOG_DEBUG("  band name: %s", gtiff_band_meta.band_name);
        IAS_LOG_DEBUG("  pixel size X: %lf",
            gtiff_band_meta.projection_distance_x);
        IAS_LOG_DEBUG("  pixel size Y: %lf",
            gtiff_band_meta.projection_distance_y);
        IAS_LOG_DEBUG("  UL X: %lf", gtiff_band_meta.upper_left_x);
        IAS_LOG_DEBUG("  UL Y: %lf", gtiff_band_meta.upper_left_y);
        IAS_LOG_DEBUG("  UR X: %lf", gtiff_band_meta.upper_right_x);
        IAS_LOG_DEBUG("  UR Y: %lf", gtiff_band_meta.upper_right_y);
        IAS_LOG_DEBUG("  LL X: %lf", gtiff_band_meta.lower_left_x);
        IAS_LOG_DEBUG("  LL Y: %lf", gtiff_band_meta.lower_left_y);
        IAS_LOG_DEBUG("  LR X: %lf", gtiff_band_meta.lower_right_x);
        IAS_LOG_DEBUG("  LR Y: %lf", gtiff_band_meta.lower_right_y);
        IAS_LOG_DEBUG("------------------------->");

        gtiff_band_meta.band_number = band_number;
        dtype = gtiff_band->data_type;
        sca_count = gtiff_band->sca_count;
        num_lines = gtiff_band->line_count;
        num_samples = gtiff_band->sample_count;

        /* Open the band in the L1G file to write to */
        l1g_band = ias_l1g_open_band(l1g_file, band_number, &dtype, &sca_count,
            &num_lines, &num_samples);
        if (l1g_band == NULL)
        {
            IAS_LOG_ERROR("Adding band number %d to output image %s",
                band_number, l1g_filename);
            ias_l1g_close_image(l1g_file);
            geotiff_close_band(gtiff_band);
            geotiff_close_file(gtiff_file);
            exit(EXIT_FAILURE);
        }

        /* Convert the GeoTIFF imagery to L1G format */
        if (convert_imagery(gtiff_band, l1g_filename, sca_index, l1g_band)
            != SUCCESS)
        {
            IAS_LOG_ERROR("Converting GeoTIFF imagery from %s to L1G imagery "
                "for %s", cur_filename, l1g_filename);
            ias_l1g_close_band(l1g_band);
            ias_l1g_close_image(l1g_file);
            geotiff_close_band(gtiff_band);
            geotiff_close_file(gtiff_file);
            exit(EXIT_FAILURE);
        }

        /* Done with the input GeoTIFF so close now */
        if (geotiff_close_band(gtiff_band) != SUCCESS)
        {
            IAS_LOG_ERROR("Closing GeoTIFF band number %d", band_number);
            ias_l1g_close_band(l1g_band);
            ias_l1g_close_image(l1g_file);
            geotiff_close_file(gtiff_file);
            exit(EXIT_FAILURE);
        }

        if (geotiff_close_file(gtiff_file) != SUCCESS)
        {
            IAS_LOG_ERROR("Closing GeoTIFF input file %s", cur_filename);
            ias_l1g_close_band(l1g_band);
            ias_l1g_close_image(l1g_file);
            exit(EXIT_FAILURE);
        }

        /* Put the GeoTIFF file and band metadata into the L1G structures */
        convert_metadata(&gtiff_file_meta, &gtiff_band_meta, &l1g_file_meta,
            &l1g_band_meta);
   
        /* Using the first band metadata as a baseline, then use it to make 
           consistency checks with the rest of the bands */
        if (band_index == 0)
        {
            base_file_meta = l1g_file_meta;
            base_band_meta = l1g_band_meta;
        }
        else if (check_metadata(&base_file_meta, &base_band_meta, 
            &l1g_file_meta, &l1g_band_meta) != SUCCESS)
        {
            IAS_LOG_ERROR("Metadata for file %s does not match", 
                cur_filename);
            ias_l1g_close_band(l1g_band);
            ias_l1g_close_image(l1g_file);
            exit(EXIT_FAILURE);
        }

        /* Write the file metadata */
        if (ias_l1g_set_file_metadata(l1g_file, &l1g_file_meta) != SUCCESS)
        {
            IAS_LOG_ERROR("Writing file metadata to L1G file %s", l1g_filename);
            ias_l1g_close_band(l1g_band);
            ias_l1g_close_image(l1g_file);
            exit(EXIT_FAILURE);
        }

        /* Write the band metadata */
        if (ias_l1g_set_band_metadata(l1g_file, &l1g_band_meta, 1) 
            != SUCCESS)
        {
            IAS_LOG_ERROR("Writing band number %d metadata to L1G file %s", 
                band_number, l1g_filename);
            ias_l1g_close_band(l1g_band);
            ias_l1g_close_image(l1g_file);
            exit(EXIT_FAILURE);
        }

        /* Close the image band */
        if (ias_l1g_close_band(l1g_band) != SUCCESS)
        {
            IAS_LOG_ERROR("Closing band number %d of L1G file %s", band_number,
                l1g_filename);
            ias_l1g_close_image(l1g_file);
            exit(EXIT_FAILURE);
        }
    }

    /* Close the L1G file */
    if (ias_l1g_close_image(l1g_file) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing the L1G file %s", l1g_filename);
        exit(EXIT_FAILURE);
    }

    IAS_LOG_INFO("Successful completion of GeoTIFF to L1G Conversion");

    return EXIT_SUCCESS;
}
