#include <stdlib.h>
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_l1g.h"
#include "geotiff_io.h"

/******************************************************************************
Name: convert_imagery

Purpose: Reads in the imagery from the GeoTIFF file and writes it out to the
         L1G file

Returns:
    SUCCESS
    ERROR
******************************************************************************/
int convert_imagery
(
    const GEOTIFF_BAND_IO *gtiff_band,      /* I: GEOTIFF_BAND_IO structure */
    const char *l1g_filename,               /* I: The name of the output file */
    int sca_index,                          /* I: For L1G file format */
    L1G_BAND_IO *l1g_band                   /* O: L1G band to write to */
)
{
    void *buff = NULL;      /* Data buffer */
    int word_nbytes;        /* Number of bytes in the dtype */
    int block_nlines;       /* Number of lines to read/write at a time */
    int cur_nlines;         /* Current number of lines reading/writing */
    int cur_line;           /* Current line marker for read/write in blocks */
    int data_lines_left_to_convert;

    /* Initialize data blocking variables */
    data_lines_left_to_convert = gtiff_band->line_count;
    block_nlines = 1000;
    cur_line = 0;

    /* Need to know the number of bytes in a word of data for the data
       type the GeoTIFF image data is in to size the buffer properly, then
       allocate the data buffer */
    if (ias_misc_get_sizeof_data_type(gtiff_band->data_type, &word_nbytes)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Getting size of IAS data type %d",
                gtiff_band->data_type);
        return ERROR;
    }

    /* Allocate data block buffer */
    buff = malloc(block_nlines * gtiff_band->sample_count * word_nbytes);
    if (!buff)
    {
        IAS_LOG_ERROR("Allocating memory for input data buffer");
        return ERROR;
    }

    while (data_lines_left_to_convert > 0)
    {
        /* Figure out size of the data block to process this time */
        if (data_lines_left_to_convert > block_nlines)
            cur_nlines = block_nlines;
        else
            cur_nlines = data_lines_left_to_convert;

        /* Read the image data for the open GeoTIFF band */
        if (geotiff_read_image(gtiff_band, sca_index, cur_line, 0, cur_nlines,
                    gtiff_band->sample_count, buff) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading band %d, line %d of %d total lines from "
                    "GeoTIFF file %s", gtiff_band->band_number, cur_line,
                    gtiff_band->line_count, gtiff_band->gtiff_file->filename);
            free(buff);
            return ERROR;
        }

        /* Write the image data from the GeoTIFF file data buffer to the
           L1G band */
        if (ias_l1g_write_image(l1g_band, sca_index, cur_line, 0, cur_nlines,
                    gtiff_band->sample_count, buff) != SUCCESS)
        {
            IAS_LOG_ERROR("Writing band %d to L1G file %s",
                    gtiff_band->band_number, l1g_filename);
            free(buff);
            return ERROR;
        }
        cur_line += block_nlines;
        data_lines_left_to_convert -= block_nlines;
    }

    /* Done writing the image data so free the data buffer */
    free(buff);

    return SUCCESS;
}
