#ifndef GEOTIFFF_IO_H
#define GEOTIFFF_IO_H

#include "geotiffio.h"
#include "xtiffio.h"
#include "ias_types.h"          /* IAS_DATA_TYPE */
#include "ias_const.h"          /* IAS-specific constants */
#include "ias_linked_list.h"

/****************************************************************************
Define forward references
TODO: for use with GeoTIFF I/O lib
****************************************************************************/
/* Create a forward reference to the GEOTIFFIO and GEOTIFF_BAND_IO structures
   that are defined locally in the library routine since external users should
   not need to know what is contained in the structure */
typedef struct GEOTIFFIO GEOTIFFIO;
typedef struct GEOTIFF_BAND_IO GEOTIFF_BAND_IO;


/****************************************************************************
Define macros
****************************************************************************/
#define GEOTIFF_SPACECRAFT_SIZE 32
#define GEOTIFF_CAPTURE_DIRECTION_SIZE 32
#define GEOTIFF_CAPTURE_DATE_SIZE 11


/****************************************************************************
Define the structures for holding file and band metadata
Define the structures for holding info from the TIFF tags and GEOTIFF keys
****************************************************************************/
typedef struct geotiff_file_metadata
{
    int projection_code;        /* GCTP projection code for the image */
    int zone_code;              /* Zone code for the UTM projection */
    char datum[IAS_DATUM_SIZE]; /* Datum (usually WGS84) */
    int spheroid_code;          /* GCTP spheroid code (usually 12 for WGS84) */
    char projection_units[IAS_UNITS_SIZE]; /* Projection units */
    double projection_parameters[IAS_PROJ_PARAM_SIZE];
    int wrs_path;
    int wrs_row;
    char spacecraft[GEOTIFF_SPACECRAFT_SIZE]; /* Spacecraft name */
    char capture_direction[GEOTIFF_CAPTURE_DIRECTION_SIZE];
                                /* ASCENDING or DESCENDING */
    char capture_date[GEOTIFF_CAPTURE_DATE_SIZE]; /* YYYY/MM/DD format */
} GEOTIFF_FILE_METADATA;

typedef struct geotiff_band_metadata
{
    int band_number;
    char band_name[IAS_BAND_NAME_SIZE];/* User readable band name. For 
                                          example, might be "RED" or "GREEN" */
    double projection_distance_y;   /* Pixel size in projection coordinates */
    double projection_distance_x;
    double upper_left_y;            /* Corner coordinates */
    double upper_left_x;
    double upper_right_y;
    double upper_right_x;
    double lower_left_y;
    double lower_left_x;
    double lower_right_y;
    double lower_right_x;
} GEOTIFF_BAND_METADATA;


/****************************************************************************
Define the private structures for holding file and band metadata
TODO: These would be in a local ias_geotiff_private.h header
      when GeoTIFF I/O lib setup
****************************************************************************/
/* This structure caches information about the band datasets in the image */
/* NOTE: The data type and line/sample count items don't differ from band
         to band in a multi-band GeoTIFF file and we only need to support
         single band files for now anyway, but keeping the storage of these
         items at this level for the following reasons:
           1) to be consistent with the design of the other IAS I/O libraries
              for when this gets moved to create a GeoTIFF I/O library and
           2) this could more easily potentially be extended to read multiple
              GeoTIFF files that have one band per file, but read each in for
              use together as the various bands of one image
*/
typedef struct band_dataset_cache_node
{
    IAS_LINKED_LIST_NODE node;  /* Linked list node */
    int band_number;            /* Band number from GeoTIFF for this record */
    IAS_DATA_TYPE data_type;    /* Data type of the band from the
                                   TIFFTAG_SAMPLEFORMAT &
                                   TIFFTAG_BITSPERSAMPLE */
    int sca_count;              /* Number of SCAs in the band (1 for SCA
                                   combined imagery) */
    int line_count;             /* Number of lines in the band from the
                                   TIFFTAG_IMAGELENGTH */
    int sample_count;           /* Number of samples in the band from the
                                   TIFFTAG_IMAGEWIDTH */
} BAND_DATASET_CACHE_NODE;

/* Macro to make it easy to delete the band dataset cache data */
#define BAND_DATASET_LINKED_LIST_DELETE(list) \
    ias_linked_list_delete((list), offsetof(BAND_DATASET_CACHE_NODE, node));

/* The main GEOTIFFIO structure. This structure is only visible internally.
   External interfaces use the forward reference to the GEOTIFFIO typedef to
   hide the contents of the structure from applications. */
struct GEOTIFFIO
{
    char *filename;             /* Name of the GeoTIFF file */
    IAS_ACCESS_MODE access_mode;/* Access mode for the file */
    TIFF *tif;                  /* TIFF-level descriptor */
    GTIF *gtif;                 /* GeoKey-level descriptor */
    int pixel_is_point;         /* Raster type - pixel is point flag */
    int zone_is_southern_hemisphere;  /* Southern hemisphere flag */
    IAS_LINKED_LIST_NODE band_datasets; /* Cache of the band dataset info */
    IAS_LINKED_LIST_NODE open_bands; /* Linked list of open bands */
};

/* Structure for tracking open bands */
struct GEOTIFF_BAND_IO
{
    int band_number;                /* Band number from GeoTIFF of currently
                                       open band */
    IAS_DATA_TYPE data_type;        /* Data type of the currently open band */
    int sca_count;                  /* Number of SCAs in the band (1 for SCA
                                       combined imagery) */
    int line_count;                 /* Lines in currently open band */
    int sample_count;               /* Samples in currently open band */
    struct GEOTIFFIO *gtiff_file;   /* Pointer to GeoTIFF file structure */
    IAS_LINKED_LIST_NODE node;      /* Linked list node */
};


/****************************************************************************
Function prototypes
****************************************************************************/

GEOTIFFIO *geotiff_open_file
(
    const char *gtiff_filename, /* I: Path & name of the GeoTIFF file */
    IAS_ACCESS_MODE access_mode /* I: Requested access mode of file */
);

int geotiff_close_file
(
    GEOTIFFIO *gtiff_file       /* I: GeoTIFF I/O structure to close */
);

GEOTIFF_BAND_IO *geotiff_open_band
(
    GEOTIFFIO *gtiff_file,      /* I: GeoTIFF I/O pointer for file info */
    int band_number,            /* I: Band number to open */
    IAS_DATA_TYPE *data_type,   /* I/O: Data type for the imagery */
    int *number_of_scas,        /* I/O: SCAs to include in the band (1 if an
                                        SCA-combined image) */
    int *line_count,            /* I/O: Lines in the band (input if writing a
                                        new band, output if opening existing) */
    int *sample_count           /* I/O: Samples in the band (input if writing a
                                        new band, output if opening existing) */
);

int geotiff_close_band
(
    GEOTIFF_BAND_IO *gtiff_band     /* I: GEOTIFF_BAND_IO structure */
);

int geotiff_get_band_number
(
    const GEOTIFFIO *gtiff_file,    /* I: GEOTIFF_BAND_IO structure */
    int *band_number                /* O: Band number */
);

int geotiff_read_image
(
    const GEOTIFF_BAND_IO *gtiff_band,  /* I: GEOTIFF_BAND_IO structure */
    int sca_index,                      /* I: SCA to read (0-rel) */
    int start_line,                     /* I: Line to start reading (0-rel) */
    int start_sample,                   /* I: Sample to start reading (0-rel) */
    int line_count,                     /* I: Number of lines to read */
    int sample_count,                   /* I: Number of samples to read */
    void *image_buff                    /* O: Data buffer */
);

int geotiff_get_file_metadata
(
    GEOTIFFIO *geotiff_file,            /* I/O: GEOTIFFIO structure */
    GEOTIFF_FILE_METADATA *metadata     /* O: output metadata */
);

int geotiff_get_band_metadata
(
    const GEOTIFF_BAND_IO *gtiff_band,  /* I: GEOTIFF_BAND_IO structure */
    GEOTIFF_BAND_METADATA *metadata     /* O: Output metadata */
);

int geotiff_get_band_count
(
    GEOTIFFIO *gtiff_file       /* I: GeoTIFF info structure for the image */
);

int geotiff_is_band_present
(
    const GEOTIFFIO *gtiff_file,    /* I: GeoTIFF info structure for the file */
    int band_number                 /* I: Band number to check */
);

int geotiff_get_band_size
(
    GEOTIFFIO *gtiff_file,      /* I: GeoTIFF info structure for the image */
    int band_number,            /* I: Band number to get information for */
    IAS_DATA_TYPE *data_type,   /* O: Data type for the imagery */
    int *number_of_scas,        /* O: SCAs in the band */
    int *line_count,            /* O: Lines in the band */
    int *sample_count           /* O: Samples in the band */
);

int geotiff_is_geotiff_file
(
    const char *gtiff_filename      /* I: GeoTIFF input file name */
);

#if 0
/* Commented out until deployed as a library */
int geotiff_write_image
(
    const GEOTIFF_BAND_IO *l1g_band,    /* I: GEOTIFF_BAND_IO structure */
    int start_line,                     /* I: Line to start writing (0-rel) */
    int start_sample,                   /* I: Sample to start writing (0-rel) */
    int line_count,                     /* I: Number of lines to write */
    int sample_count,                   /* I: Number of samples to write */
    void *data                          /* I: Data buffer */
);

/* Commented out until deployed as a library */
int geotiff_set_file_metadata
(
    GEOTIFFIO *geotiff_file,                /* I: GEOTIFFIO structure */
    const GEOTIFF_FILE_METADATA *metadata   /* I: metadata to set */
);

/* Commented out until deployed as a library */
int geotiff_set_band_metadata
(
    GEOTIFFIO *geotiff_file,                /* I: file to set metadata for */
    const GEOTIFF_BAND_METADATA *metadata,  /* I: array of metadata to set */
    int metadata_size                       /* I: number of entries in metadata
                                                  array */
);
#endif

#endif
