/* Standard Library Includes */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* COTS Library Includes */
#include "geotiffio.h"
#include "geo_normalize.h"
#include "xtiffio.h"
#include "gctp.h"

/* IAS Library Includes */
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_geo.h"

/* Local Library Includes */
#include "geotiff_io.h"

/* Local Defines */
#define FILE_SIGNATURE_BYTE_COUNT   4
#define GEO_KEY_INDEX 0                /* Retrieval index of the GeoKeys */
#define GEO_KEY_COUNT 1                /* Number of keys returned */

/******************************************************************************
NAME:   initialize_geotiff (static)
        geotiff_add_band_to_cache (static)
        geotiff_get_band_count
        find_band (static)
        geotiff_is_band_present
        geotiff_get_band_size
        get_gtiff_bit_depth (static)
        get_gtiff_sample_format (static)
        get_gtiff_data_type (static)
        get_ias_data_type (static)
        open_file_dataset (static)
        geotiff_open_file
        geotiff_is_geotiff_file

        geotiff_open_band

        geotiff_close_band
        geotiff_close_file

        geotiff_read_image

        geotiff_write_image

        geotiff_set_file_metadata
        geotiff_get_file_metadata

        geotiff_set_band_metadata
        geotiff_get_band_metadata

PURPOSE: Contains GeoTIFF formatted file I/O routines

NOTES:
A GeoTIFF file contains images. Only supporting single image files at this
time. Multiple images could be the multiple bands of data from a scene. The
terminology used by the TIFF library is that the .TIF is the file, which
contains images, but in our IAS world the file contains an individual band,
so band terminology is used for a file.

Even though the requirements of the conversion application are pretty simple
for now (per CalVal), the routines within this file were written to mimic
the functionality of the IAS L1G I/O library as close as possible to be
stood-up as its own GeoTIFF I/O library in the future. Developing it from the
start to support the multiple bands is best keeping that future as a library
in mind, even though the only application using it right now will only be
working with single band images.

Reading a single-image GeoTIFF file is all that has been implemented.
Writing a GeoTIFF file is not yet implemented.

Future enhancements could be to develop an interface to use the band caching
to develop a set of bands from multiple single image/band GeoTIFF files read
in to combine up the bands from the separate GeoTIFF files into a cohesive
set of data from a common image.
******************************************************************************/

/******************************************************************************
Name: initialize_geotiff

Purpose: Initializes the contents of a GEOTIFFIO structure

Returns:
    nothing
******************************************************************************/
static void initialize_geotiff
(
    GEOTIFFIO *gtiff_file       /* O: GeoTIFF I/O structure */
)
{
    gtiff_file->access_mode = -1;
    ias_linked_list_initialize_node(&gtiff_file->band_datasets);
    ias_linked_list_initialize_node(&gtiff_file->open_bands);
}

/******************************************************************************
Name: geotiff_add_band_to_cache

Purpose: Adds the size information for a band to the band dataset cache

Returns:
    SUCCESS     the band was successfully added to the cache
    ERROR       the band was not added to the cache
******************************************************************************/
static int geotiff_add_band_to_cache
(
    GEOTIFFIO *gtiff_file,      /* I: Pointer to GeoTIFF file structure */
    int band_number,            /* I: Band number being added */
    IAS_DATA_TYPE data_type,    /* I: Data type in the band */
    int sca_count,              /* I: Number of SCAs in the band */
    int line_count,             /* I: Number of lines in the band */
    int sample_count            /* I: Number of samples in the band */
)
{
    BAND_DATASET_CACHE_NODE *band;

    /* Verify the input parameter is valid */
    if (!gtiff_file)
    {
        IAS_LOG_ERROR("NULL pointer provided for the GeoTIFF file");
        return ERROR;
    }

    /* Allocate and initialize a band dataset cache node */
    band = malloc(sizeof(*band));
    if (!band)
    {
        IAS_LOG_ERROR("Allocating memory for band cache");
        return ERROR;
    }
    ias_linked_list_initialize_node(&band->node);

    /* Populate the cache node */
    band->band_number = band_number;
    band->data_type = data_type;
    band->sca_count = sca_count;
    band->line_count = line_count;
    band->sample_count = sample_count;

    /* Add the cache node to the end of the cache linked list */
    ias_linked_list_add_tail(&gtiff_file->band_datasets, &band->node);

    return SUCCESS;
}

/******************************************************************************
Name: geotiff_get_band_count

Purpose: Returns the number of bands present in the file

Returns:
    Number of bands present in the file
    ERROR
******************************************************************************/
int geotiff_get_band_count
(
    GEOTIFFIO *gtiff_file       /* I: GeoTIFF info structure for the file */
)
{
    /* Verify the input parameter is valid */
    if (!gtiff_file)
    {
        IAS_LOG_ERROR("NULL pointer provided for the GeoTIFF file");
        return ERROR;
    }

    return ias_linked_list_count_nodes(&gtiff_file->band_datasets);
}

/******************************************************************************
Name: find_band

Purpose: Searches the band dataset cache for the requested band and returns
         a pointer to the cache entry

Returns:
    Pointer to the band dataset cache node found
    NULL if node not found
******************************************************************************/
static BAND_DATASET_CACHE_NODE *find_band
(
    const GEOTIFFIO *gtiff_file,    /* I: GeoTIFF file to search for the band */
    int band_number                 /* I: Band number to find */
)
{
    BAND_DATASET_CACHE_NODE *current;

    /* Search the metadata cache for the correct band */
    GET_OBJECT_FOR_EACH_ENTRY(current, &gtiff_file->band_datasets,
            BAND_DATASET_CACHE_NODE, node)
    {
        if (band_number == current->band_number)
            return current;
    }

    /* The band wasn't found */
    return NULL;
}

/******************************************************************************
Name: geotiff_is_band_present

Purpose: Allows checking whether a band is present in the file

Returns:
    TRUE (1)        the band is present in the file
    FALSE (0)       the band is not present in the file

Notes:
    - This routine is included in this file to allow the find_band routine
      to remain static
******************************************************************************/
int geotiff_is_band_present
(
    const GEOTIFFIO *gtiff_file,    /* I: GeoTIFF info structure for the file */
    int band_number                 /* I: Band number to check */
)
{
    BAND_DATASET_CACHE_NODE *current;

    /* Verify the input parameter is valid */
    if (!gtiff_file)
    {
        IAS_LOG_ERROR("NULL pointer provided for the GeoTIFF file");
        return ERROR;
    }

    current = find_band(gtiff_file, band_number);
    if (current)
        return TRUE;
    else
        return FALSE;
}

/******************************************************************************
Name: geotiff_get_band_size

Purpose: For a given band number, return the size related information
         (data type, number of SCAs, lines, and samples)

Returns:
    SUCCESS     the band size information was successfully retrieved
    ERROR       the band size information could not be retrieved

Notes:
    - This routine is included in this file to allow the find_band routine
      to remain static
******************************************************************************/
int geotiff_get_band_size
(
    GEOTIFFIO *gtiff_file,      /* I: GeoTIFF info structure for the file */
    int band_number,            /* I: Band number to get information for */
    IAS_DATA_TYPE *data_type,   /* O: Data type for the imagery */
    int *sca_count,             /* O: SCAs in the band */
    int *line_count,            /* O: Lines in the band */
    int *sample_count           /* O: Samples in the band */
)
{
    BAND_DATASET_CACHE_NODE *current;

    /* Verify the input parameter is valid */
    if (!gtiff_file)
    {
        IAS_LOG_ERROR("NULL pointer provided for the GeoTIFF file");
        return ERROR;
    }

    /* Search the metadata for the correct band */
    current = find_band(gtiff_file, band_number);
    if (current)
    {
        /* The band was found, so return the proper data */
        *data_type = current->data_type;
        *sca_count = current->sca_count;
        *line_count = current->line_count;
        *sample_count = current->sample_count;

        return SUCCESS;
    }

    /* The band wasn't found */
    return ERROR;
}


/******************************************************************************
Name: get_ias_data_type

Purpose: For a GeoTIFF data type definition, returns the matching IAS_DATA_TYPE

Returns:
    IAS_DATA_TYPE matching the given GeoTIFF data type
    type_error if there is no match
******************************************************************************/
static IAS_DATA_TYPE get_ias_data_type
(
    unsigned short bits_per_sample, /* I: TIFF image bits per sample */
    unsigned short sample_format,   /* I: TIFF image sample format */
    const char *gtiff_filename      /* I: TIFF image filename */
)
{
    IAS_DATA_TYPE data_type;    /* IAS data type matching GeoTIFF data
                                   type definition */
    int data_type_bad = 0;      /* Flag for unsupported data type */

    /* The GeoTIFF data type is defined with a combination of two tags.
       Interpret the tag combination to determine the matching IAS_DATA_TYPE */
    if (sample_format == SAMPLEFORMAT_INT)
    {
        if (bits_per_sample == 16)
            data_type = IAS_I2;
        else
            data_type_bad++;
    }
    else if (sample_format == SAMPLEFORMAT_IEEEFP)
    {
        if (bits_per_sample == 32)
            data_type = IAS_R4;
        else
            data_type_bad++;

    }
    else if ((sample_format == SAMPLEFORMAT_COMPLEXINT)
            || (sample_format == SAMPLEFORMAT_COMPLEXIEEEFP))
    {
        data_type_bad++;
    }
    else
    {
        /* This captures the sample_format being SAMPLEFORMAT_UNIT,
           SAMPLEFORMAT_VOID, as well as if the sample format tag wasn't there
           or contained an invalid value. In all these cases, treat the data as
           if the field were not present (i.e. as unsigned integer data) */
        if (sample_format != SAMPLEFORMAT_UINT)
        {
            IAS_LOG_WARNING("Unrecognized sample format in file %s - assuming "
                    "image is unsigned integer data", gtiff_filename);
        }

        if (bits_per_sample == 8)
            data_type = IAS_BYTE;
        else if (bits_per_sample == 16)
            data_type = IAS_UI2;
        else
            data_type_bad++;
    }

    /* Error if the GeoTIFF data type definition didn't match our
       supported IAS data types */
    if (data_type_bad)
    {
        IAS_LOG_ERROR("Unsupported data type in %s", gtiff_filename);
        return type_error;
    }

    return data_type;
}

/******************************************************************************
Name: open_file_dataset

Purpose: Get information about the TIFF images in the file. These are the
         bands of the scene. Store the information about the bands in the
         band datasets list.

Returns:
    SUCCESS
    ERROR

Note: Single-image (one band) in a file is all that is supported at this time.
******************************************************************************/
static int open_file_dataset
(
    GEOTIFFIO *gtiff_file, /* I: GeoTIFF info structure for the file */
    int band_number        /* I: Band Number */
)
{
    BAND_DATASET_CACHE_NODE *current;
    int sca_count = 1;          /* Only supporting SCA-combined imagery */

    /* Types for reading from GeoTIFF file */
    tsample_t config = 0;       /* The image's planar configuration (BIP/BIL) */
    unsigned int width;                 /* Width of the TIFF image */
    unsigned int length;                /* Height of the TIFF image */
    unsigned short bits_per_sample;     /* TIFF image bits per sample */
    unsigned short sample_format;       /* TIFF image sample format */

    /* Types for internal IAS structures */
    int line_count;             /* Lines */
    int sample_count;           /* Samples */
    IAS_DATA_TYPE data_type;    /* IAS data type */

    /* Search the linked list to see if band_number appears;
       if the band is already open, issue an error */
    GET_OBJECT_FOR_EACH_ENTRY(current, &gtiff_file->band_datasets,
            BAND_DATASET_CACHE_NODE, node)
    {
        if (current->band_number == band_number)
        {
            IAS_LOG_ERROR("Duplicate data for band %d found in file %s",
                    band_number, gtiff_file->filename);
            return ERROR;
        }
    }

    /* Make sure data is BIP */
    if(!TIFFGetField(gtiff_file->tif, TIFFTAG_PLANARCONFIG, &config))
    {
        IAS_LOG_ERROR("Getting tag TIFFTAG_PLANARCONFIG");
        return ERROR;
    }
    if (config != PLANARCONFIG_CONTIG)
    {
        IAS_LOG_ERROR("Unsupported interleaving of image data (expected BIP)");
        return ERROR;
    }
    
    /* The following 4 fields are not always present and if they are not present
       try to fall back to the field default before returning with an error */
    /* Extract the image size */
    if(!TIFFGetField(gtiff_file->tif, TIFFTAG_IMAGELENGTH, &length))
    {
        if(!TIFFGetFieldDefaulted(gtiff_file->tif, TIFFTAG_IMAGELENGTH, 
            &length))
        {
           IAS_LOG_ERROR("Getting the image length field");
           return ERROR;
        }
    }
    if(!TIFFGetField(gtiff_file->tif, TIFFTAG_IMAGEWIDTH, &width))
    {
        if(!TIFFGetFieldDefaulted(gtiff_file->tif, TIFFTAG_IMAGEWIDTH, &width))
        {
           IAS_LOG_ERROR("Getting the image width field");
           return ERROR;
        }
    }

    /* Extract the bits per sample */
    if(!TIFFGetField(gtiff_file->tif, TIFFTAG_BITSPERSAMPLE, &bits_per_sample))
    {
        if(!TIFFGetFieldDefaulted(gtiff_file->tif, TIFFTAG_BITSPERSAMPLE,
            &bits_per_sample))
        {
           IAS_LOG_ERROR("Getting the image bits per sample field");
           return ERROR;
        }
    }
    
    /* Extact the sample format */
    if(!TIFFGetField(gtiff_file->tif, TIFFTAG_SAMPLEFORMAT, &sample_format))
    {
        if(!TIFFGetFieldDefaulted(gtiff_file->tif, TIFFTAG_SAMPLEFORMAT, 
            &sample_format))
        {
           IAS_LOG_ERROR("Getting the image sample format field");
           return ERROR;
        }
    }

    /* Convert the tag data to appropriate format for the band structure */
    line_count = (int)length;
    sample_count = (int)width;

    /* Sanity check the lines and samples. The other values are checked
       for invalid values within get_ias_data_type. */
    if ((line_count <= 0) || (sample_count <= 0))
    {
        IAS_LOG_ERROR("Invalid lines/samples image size: %u/%u", length, width);
        return ERROR;
    }

    /* Determine the IAS data type from the TIFF tags for this info */
    data_type = get_ias_data_type(bits_per_sample, sample_format,
            gtiff_file->filename);
    if (data_type == type_error)
    {
        IAS_LOG_ERROR("Getting IAS data type from GeoTIFF data type");
        return ERROR;
    }

    /* Add the band info to the band cache */
    if (geotiff_add_band_to_cache(gtiff_file, band_number, data_type,
                sca_count, line_count, sample_count) != SUCCESS)
    {
        IAS_LOG_ERROR("Adding the band data to the cache");
        return ERROR;
    }

    return SUCCESS;
}

/******************************************************************************
Name: geotiff_open_file

Purpose: Open the specified GeoTIFF image

Returns:
    GEOTIFFIO structure pointer used to access the image
    NULL if an error occurred

Notes:
    - For a future enhancement to support multi-band GeoTIFF data from
      multiple files, some code could be added at a level above this to
      open_files and call open_file for each file in a given set. This would
      then be called for each file and open_file_dataset ultimately called for
      all files to compile the band information all together from the
      separate files.
******************************************************************************/
GEOTIFFIO *geotiff_open_file
(
    const char *gtiff_filename, /* I: Path & name of the GeoTIFF file */
    IAS_ACCESS_MODE access_mode /* I: Requested access mode of file */
)
{
    GEOTIFFIO *gtiff_file = NULL;       /* Structure for the GeoTIFF file */
    int band_number = 1;                /* Only support single band files */
    unsigned short samples_per_pixel;
    geocode_t raster_type;
    int status;

    /* Check for currently unsupported access modes for quick exit */
    if (access_mode == IAS_WRITE)
    {
        /* Not supported at this time */
        IAS_LOG_ERROR("Writing GeoTIFF files not supported at this time");
        return NULL;
    }
    else if (access_mode == IAS_UPDATE)
    {
        /* Even though the base TIFF library supports appending to an
           existing TIFF file, the libtiff and libgeotiff libraries do
           not support update mode */
        IAS_LOG_ERROR("A GeoTIFF file cannot be opened for updating");
        return NULL;
    }

    /* Check that the file is a GeoTIFF file */
    status = geotiff_is_geotiff_file(gtiff_filename);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Determining file type of %s", gtiff_filename);
        return NULL;
    }
    else if (status != TRUE)
    {
        IAS_LOG_ERROR("Attempting to access non-GeoTIFF file: %s",
                gtiff_filename);
        return NULL;
    }

    /* Allocate space for the I/O data structure */
    gtiff_file = calloc(1, sizeof(*gtiff_file));
    if (!gtiff_file)
    {
        IAS_LOG_ERROR("Allocating memory for GeoTIFF file structure");
        return NULL;
    }

    /* Initialize the gtiff_file structure */
    initialize_geotiff(gtiff_file);

    /* Save the filename and access mode */
    gtiff_file->access_mode = access_mode;
    gtiff_file->filename = strdup(gtiff_filename);
    if (!gtiff_file->filename)
    {
        IAS_LOG_ERROR("Allocating memory for filename");
        free(gtiff_file);
        return NULL;
    }

    /* Opening an existing file in read mode (other modes were checked
       for earlier) */

    /* Open the TIFF descriptor */
    /*  ---------------------------------------------------------------
        The XTIFFOpen function differs TIFFOpen in that it extends the
        set of tags interpreted by libtiff. The XTIFFOpen implementation
        adds support for the GeoTIFF tag set and is required to support
        libgeotiff.
        --------------------------------------------------------------- */
    gtiff_file->tif = XTIFFOpen(gtiff_file->filename, "r");
    if (!gtiff_file->tif)
    {
        IAS_LOG_ERROR("Opening TIFF file %s for reading",
                gtiff_file->filename);
        free(gtiff_file->filename);
        free(gtiff_file);
        return NULL;
    }

    /* Open GTIF Key parser */
    gtiff_file->gtif = GTIFNew(gtiff_file->tif);
    if (!gtiff_file->gtif)
    {
        IAS_LOG_ERROR("Opening GTIF Key parser for %s",
                gtiff_file->filename);
        XTIFFClose(gtiff_file->tif);
        free(gtiff_file->filename);
        free(gtiff_file);
        return NULL;
    }

    /* Only the first directory is read, if the current directory is not the
       last then issue a warning that multiple GeoTiff directories exist */
    if(!TIFFLastDirectory(gtiff_file->tif))
    {
        IAS_LOG_WARNING("Multiple GeoTIFF directories present in %s",
            gtiff_file->filename);
    }

    /* Additionally, only supporting GeoTIFF files with a single band
       per image at this time. */
    if(!TIFFGetField(gtiff_file->tif, TIFFTAG_SAMPLESPERPIXEL,
                &samples_per_pixel))
    {
        IAS_LOG_ERROR("Getting tag TIFFTAG_SAMPLESPERPIXEL");
        GTIFFree(gtiff_file->gtif);
        XTIFFClose(gtiff_file->tif);
        free(gtiff_file->filename);
        free(gtiff_file);
        return NULL;
    }
    if (samples_per_pixel != 1)
    {
        /* A number other than 1 indicates there are multiple bands
           in the image, which we do not support at this time */
        IAS_LOG_ERROR("Multi-band GeoTIFF imagery not supported");
        GTIFFree(gtiff_file->gtif);
        XTIFFClose(gtiff_file->tif);
        free(gtiff_file->filename);
        free(gtiff_file);
        return NULL;
    }

    /* Read the file to get the band information and cache it */
    if (open_file_dataset(gtiff_file, band_number) != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving band dataset cache information");
        GTIFFree(gtiff_file->gtif);
        XTIFFClose(gtiff_file->tif);
        free(gtiff_file->filename);
        free(gtiff_file);
        return NULL;
    }

    /* Read the gtif raster type */
    if (!GTIFKeyGet(gtiff_file->gtif, GTRasterTypeGeoKey, 
        &raster_type, GEO_KEY_INDEX, GEO_KEY_COUNT))
    {
        IAS_LOG_ERROR("No Raster Type found");
        GTIFFree(gtiff_file->gtif);
        XTIFFClose(gtiff_file->tif);
        free(gtiff_file->filename);
        free(gtiff_file);
    }

    /* Issue info message if corners will need to be adjusted */
    if (raster_type != RasterPixelIsPoint)
    {
        gtiff_file->pixel_is_point = 0;
        IAS_LOG_INFO("Raster Type 'PixelIsArea' corners will need "
            "to be adjusted");
    }
    else
    {
        gtiff_file->pixel_is_point = 1;
        IAS_LOG_INFO("Raster Type 'PixelIsPoint' corners left unchanged");
    }

    return gtiff_file;
}

/*************************************************************************
NAME: geotiff_is_geotiff_file

PURPOSE: Determine weather the file is a GeoTIFF file 

RETURNS:
    TRUE (1)        is a GeoTIFF file
    FALSE (0)       is not a GeoTIFF file
    ERROR           an error occurred before file type could be determined
**************************************************************************/
int geotiff_is_geotiff_file
(
    const char *gtiff_filename      /* I: GeoTIFF input file name */
)
{
    /* The file signature of the first 4 bytes for the TIFF file type. Little
       endian (le_sig) for Intel, big endian (be_sig) for Motorola (not likely,
       but may as well check for).
       Byte patterns to look for are: */
    unsigned char le_sig[FILE_SIGNATURE_BYTE_COUNT] = {0x49, 0x49, 0x2a, 0x00};
    unsigned char be_sig[FILE_SIGNATURE_BYTE_COUNT] = {0x4d, 0x4d, 0x00, 0x2a};
    unsigned char buf[FILE_SIGNATURE_BYTE_COUNT];
    FILE *gtiff_fp;
    int size_read;
    int byte_match_count = 0;
    int sig_match = 0;              /* Flag for if the file signature matches */
    int i;

    /* Get the file signature header bytes from the file and compare against
       the TIFF file type signatures.
       http://www.fileformat.info/format/tiff/egff.htm */

    /* Open the file */
    gtiff_fp = fopen(gtiff_filename, "r");
    if (!gtiff_fp)
    {
        IAS_LOG_ERROR("Opening GeoTIFF %s", gtiff_filename);
        return ERROR;
    }

    /* Read the first group of bytes to get the unique file signature
       to determine if this is a TIFF file or not */
    size_read = fread(buf, sizeof(buf), FILE_SIGNATURE_BYTE_COUNT, gtiff_fp);
    if (size_read != FILE_SIGNATURE_BYTE_COUNT)
    {
        IAS_LOG_ERROR("Reading file signature header from %s", gtiff_filename);
        fclose(gtiff_fp);
        return ERROR;
    }

    /* Check for the little endian file signature */
    for (i = 0; i < FILE_SIGNATURE_BYTE_COUNT; i++)
    {
        if (buf[i] == le_sig[i])
            byte_match_count++;
    }
    if (byte_match_count == FILE_SIGNATURE_BYTE_COUNT)
        sig_match = 1;
    if (!sig_match)
    {
        /* Didn't find little endian file signature, so check for the
           big endian file signature */
        byte_match_count = 0;
        for (i = 0; i < FILE_SIGNATURE_BYTE_COUNT; i++)
        {
            if (buf[i] == be_sig[i])
                byte_match_count++;
        }
        if (byte_match_count == FILE_SIGNATURE_BYTE_COUNT)
            sig_match = 1;
    }

    /* If the file signature bytes don't match that of the known TIFF file
       signatures, it's pretty likely this is not a TIFF file */
    if (!sig_match)
    {
        IAS_LOG_DEBUG("Input not a GeoTIFF file: %s", gtiff_filename);
        fclose(gtiff_fp);
        return FALSE;
    }

    fclose(gtiff_fp);

    /* Made it past all the byte checking and all passed, so it's a real
       good assumption this is a GeoTIFF file */
    return TRUE;
}


/******************************************************************************
Name: geotiff_open_band

Purpose: Open a band in the file

Returns:
    Pointer to the GEOTIFF_BAND_IO band if the band was opened successfully
    NULL if the band was not opened successfully

Notes: Input band_number is not used, but included to setup the API similar
       to L1G
******************************************************************************/
GEOTIFF_BAND_IO *geotiff_open_band
(
    GEOTIFFIO *gtiff_file,      /* I: GeoTIFF I/O pointer for file info */
    int band_number,            /* I: Band number to open */
    IAS_DATA_TYPE *data_type,   /* I/O: Data type for the imagery */
    int *sca_count,             /* I/O: SCAs to include in the band (1 if an
                                        SCA-combined image) */
    int *line_count,            /* I/O: Lines in the band (input if writing a
                                        new band, output if opening existing) */
    int *sample_count           /* I/O: Samples in the band (input if writing a
                                        new band, output if opening existing) */
)
{
    GEOTIFF_BAND_IO *gtiff_band;
    BAND_DATASET_CACHE_NODE *gtiff_band_dataset = NULL;

    /* Verify the input parameter is valid */
    if (!gtiff_file)
    {
        IAS_LOG_ERROR("NULL pointer provided for the GeoTIFF file");
        return NULL;
    }

    /* Open read or write access to the band */
    if (gtiff_file->access_mode == IAS_WRITE)
    {
        /* Not supported at this time */
        IAS_LOG_ERROR("Writing GeoTIFF files not supported at this time");
        return NULL;
    }
    else if (gtiff_file->access_mode == IAS_READ)
    {
        TIFF *tif = gtiff_file->tif;
        BAND_DATASET_CACHE_NODE *current;
        /* Parameters of the image format to validate */
        unsigned int rows_per_strip;
        unsigned short compression;
        unsigned short photometric_scheme;

        /* Search the metadata cache for the correct band */
        GET_OBJECT_FOR_EACH_ENTRY(current, &gtiff_file->band_datasets,
                BAND_DATASET_CACHE_NODE, node)
        {
            if (band_number == current->band_number)
                gtiff_band_dataset = current;
        }
        if (gtiff_band_dataset == NULL)
        {
            IAS_LOG_ERROR("Locating cached band dataset for band %d",
                band_number);
            return NULL;
        }

        /* Populate the output parameters */
        *data_type = gtiff_band_dataset->data_type;
        *sca_count = gtiff_band_dataset->sca_count;
        *line_count = gtiff_band_dataset->line_count;
        *sample_count = gtiff_band_dataset->sample_count;

        /* Check that the image data is organized as expected. Checking the
           TIFFTAG_ROWSPERSTRIP ensures this data is organized in the file in
           strips, not tiles (this tag won't be used when the data is tiled)
           and also checks that there's just one strip, as expected.
           A return status of 0 indicates an internal error within TIFFGetField
           or that the field did not exist, meaning the image data organization
           is an unsupported format. The return status does not differentiate
           between the two. So, if the TIFFGetField call fails, futher
           investigation may be needed to exactly determine why. */
        if (!TIFFGetField(tif, TIFFTAG_ROWSPERSTRIP, &rows_per_strip))
        {
            IAS_LOG_ERROR("Getting tag TIFFTAG_ROWSPERSTRIP");
            return NULL;
        }
        if (rows_per_strip != 1)
        {
            IAS_LOG_ERROR("Unsupported image data organization");
            return NULL;
        }

        /* Not expecting to encounter compressed data, so not writing the
           code to support it */
        if (!TIFFGetField(tif, TIFFTAG_COMPRESSION, &compression))
        {
            IAS_LOG_ERROR("Getting tag TIFFTAG_COMPRESSION");
            return NULL;
        }
        if (compression != COMPRESSION_NONE)
        {
            IAS_LOG_ERROR("Compressed imagery not supported");
            return NULL;
        }

        /* Make sure the GeoTIFF was written from grayscale image data
           interpreted as 0 being black and maximum value being white 
           since that is the interpretation of the data in the IAS system */
        if (!TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &photometric_scheme))
        {
            IAS_LOG_ERROR("Getting tag TIFFTAG_PHOTOMETRIC");
            return NULL;
        }
        if (photometric_scheme != PHOTOMETRIC_MINISBLACK)
        {
            IAS_LOG_ERROR("Photometric format %hu not supported",
                    photometric_scheme);
            return NULL;
        }
    }
    else /* (gtiff_file->access_mode == IAS_UPDATE) */
    {
        /* Even though the base TIFF library supports appending to an
           existing TIFF file, the libtiff and libgeotiff libraries do
           not support update mode */
        IAS_LOG_ERROR("A GeoTIFF file cannot be opened for updating");
        return NULL;
    }

    /* Allocate space for the data structure */
    gtiff_band = malloc(sizeof(*gtiff_band));
    if (!gtiff_band)
    {
        IAS_LOG_ERROR("Allocating memory for band structure");
        return NULL;
    }

    /* Cache info about the current band being opened */
    gtiff_band->band_number = gtiff_band_dataset->band_number;
    gtiff_band->sca_count = gtiff_band_dataset->sca_count;
    gtiff_band->line_count = gtiff_band_dataset->line_count;
    gtiff_band->sample_count = gtiff_band_dataset->sample_count;
    gtiff_band->data_type = gtiff_band_dataset->data_type;
    gtiff_band->gtiff_file = gtiff_file;

    /* Add the band to the linked list of open bands */
    ias_linked_list_add_tail(&gtiff_file->open_bands, &gtiff_band->node);

    return gtiff_band;
}


/*******************************************************************************
Name: geotiff_close_band

Purpose: Close the currently open GeoTIFF band

Returns:
    SUCCESS
*******************************************************************************/
int geotiff_close_band
(
    GEOTIFF_BAND_IO *gtiff_band     /* I: GEOTIFF_BAND_IO structure */
)
{
    ias_linked_list_remove_node(&gtiff_band->node);
    free(gtiff_band);
    return SUCCESS;
}


/*******************************************************************************
Name: geotiff_close_file

Purpose: Close the specified GeoTIFF file

Returns:
    SUCCESS     the file was successfully closed
    ERROR       an error occured closing the file
*******************************************************************************/
int geotiff_close_file
(
    GEOTIFFIO *gtiff_file       /* I: GeoTIFF I/O structure to close */
)
{
    int status = SUCCESS;

    /* Verify the input parameter is valid */
    if (!gtiff_file)
    {
        IAS_LOG_ERROR("NULL pointer provided for the GeoTIFF file");
        status = ERROR;
    }

    /* If there are any bands open, close them and issue a warning since
       all bands should be closed before calling this routine */
    while ((status == SUCCESS)
            && (ias_linked_list_count_nodes(&gtiff_file->open_bands) > 0))
    {
        GEOTIFF_BAND_IO *gtiff_band;
        IAS_LINKED_LIST_NODE *node;

        IAS_LOG_WARNING("Closing image %s with a band still open. Closing "
                "band, but calling code should be fixed", gtiff_file->filename);

        node = ias_linked_list_get_head_node(&gtiff_file->open_bands);
        if (!node)
        {
            /* Error since this could only be null if no nodes were in the 
               linked list */
            IAS_LOG_ERROR("Invalid node returned from the open bands linked "
                    "list");
            status = ERROR;
        }
        if (status == SUCCESS)
        {
            gtiff_band = GET_OBJECT_OF(node, GEOTIFF_BAND_IO, node);
            geotiff_close_band(gtiff_band);
        }
    }

    /* Close the file */
    GTIFFree(gtiff_file->gtif);
    XTIFFClose(gtiff_file->tif);

    /* Free the linked lists */
    BAND_DATASET_LINKED_LIST_DELETE(&gtiff_file->band_datasets);

    /* Free the memory associated with the structure */
    free(gtiff_file->filename);
    free(gtiff_file);

    return status;
}


/******************************************************************************
Name: geotiff_read_image

Purpose: Read image data from the open band of a GeoTIFF file

Returns:
    SUCCESS
    ERROR
******************************************************************************/
int geotiff_read_image
(
    const GEOTIFF_BAND_IO *gtiff_band,  /* I: GEOTIFF_BAND_IO structure */
    int sca_index,                      /* I: SCA to read (0-rel) */
    int start_line,                     /* I: Line to start reading (0-rel) */
    int start_sample,                   /* I: Sample to start reading (0-rel) */
    int line_count,                     /* I: Number of lines to read */
    int sample_count,                   /* I: Number of samples to read */
    void *image_buff                    /* O: Data buffer */
)
{
    TIFF *tif=gtiff_band->gtiff_file->tif;   /* Local pointer for easy use */
    void *line_buff;        /* Single line data buffer */
    /* Typed pointers for working with data in the data buffers */
    IAS_BYTE_TYPE *byte_img_buf_ptr;
    IAS_BYTE_TYPE *byte_line_buf_ptr;
    int nbytes;             /* Number of bytes to read from the TIFF scanline */
    int word_nbytes;        /* Number of bytes in the data type */
    int line_offset;        /* Offset from first line read */
    int i;                  /* Loop counter */

    /* Verify the input parameter is valid */
    if (!gtiff_band)
    {
        IAS_LOG_ERROR("NULL band pointer provided");
        return ERROR;
    }

    /* Only supporting SCA-combined data right now, so if anything other
       than 0 is given for the SCA-index to read, error */
    if (sca_index != 0)
    {
        IAS_LOG_ERROR("SCA-separated imagery not supported; cannot read SCA %d",
                sca_index);
        return ERROR;
    }

    /* Verify the window of data being read actually falls within the band */
    if ((start_line < 0) || ((start_line + line_count) > gtiff_band->line_count)
            || (start_sample < 0)
            || ((start_sample + sample_count) > gtiff_band->sample_count))
    {
        IAS_LOG_ERROR("Invalid request to read imagery at line %d, sample %d "
                "for a window %d lines x %d samples when the band has "
                "%d lines, %d samples", start_line, start_sample, line_count,
                sample_count, gtiff_band->line_count, gtiff_band->sample_count);
        return ERROR;
    }

    /* Allocate memory for a local buffer for reading a line at a time */
    nbytes = TIFFScanlineSize(tif);
    line_buff = malloc(nbytes);
    if (!line_buff)
    {
        IAS_LOG_ERROR("Allocating line reading buffer");
        return ERROR;
    }

    /* Point the typed pointers at the data buffers */
    byte_img_buf_ptr = image_buff;
    byte_line_buf_ptr = line_buff;

    if (ias_misc_get_sizeof_data_type(gtiff_band->data_type, &word_nbytes)
            != SUCCESS)
    {
        IAS_LOG_ERROR("Getting size of IAS data type %d",
                gtiff_band->data_type);
        free(line_buff);
        return ERROR;
    }

    /* Read the data from the dataset */
    for (i = start_line, line_offset = 0; i < start_line + line_count;
            i++, line_offset++)
    {
        /* TIFF files are read one full line at a time */
        if(TIFFReadScanline(tif, line_buff, i, 0) == -1)
        {
            IAS_LOG_ERROR("Reading from file %s, band %d, line %d, number of "
                    "lines %d", gtiff_band->gtiff_file->filename,
                    gtiff_band->band_number, i, line_count);
            free(line_buff);
            return ERROR;
        }

        /* Add the samples wanted of the line read to the cumulative
           image data buffer */
        memcpy(&byte_img_buf_ptr[(line_offset * sample_count * word_nbytes)],
                &byte_line_buf_ptr[start_sample * word_nbytes],
                (sample_count * word_nbytes));
    }

    /* Free up the line by line reading buffer */
    free(line_buff);

    return SUCCESS;
}

/******************************************************************************
Name: geotiff_get_utm_zone

Purpose: Gets the UTM zone and the north/south extent using the projection
         type key.

Returns:
    SUCCESS
    ERROR
******************************************************************************/
int geotiff_get_utm_zone
(
   short projection_type,   /* I: Projection type key */
   int *north_south_extent, /* O: North or South Extent */
   int *utm_zone            /* O: UTM zone */
)
{

    /* Check that the projection type is in the correct range */
    if (projection_type < PCS_WGS84_UTM_zone_1N || (projection_type 
        > PCS_WGS84_UTM_zone_60N && projection_type < PCS_WGS84_UTM_zone_1S) 
        || projection_type > PCS_WGS84_UTM_zone_60S)
    {
        IAS_LOG_ERROR("Projection type %d is not in the valid range", 
            projection_type);
        return ERROR;
    }

    /* Determine the UTM zone by taking the projection type mod 100 since the
       zone is just the last 2 digits of the projection type */
    *utm_zone = projection_type % 100;
    if (*utm_zone < 1 || *utm_zone > 60)
    {
        IAS_LOG_ERROR("Invalid utm zone %d", *utm_zone);
        return ERROR;
    }
 
    /* If the projection type is greater than 32660 it is in the south extent */
    *north_south_extent = MapSys_UTM_North;
    if (projection_type > PCS_WGS84_UTM_zone_60N)
    {
        *north_south_extent = MapSys_UTM_South;
    }

    return SUCCESS;
}

/******************************************************************************
Name: geotiff_get_file_metadata

Purpose: Get the file-level metadata from the open GeoTIFF file and save
         the information in the metadata structure

Returns:
    SUCCESS
    ERROR
******************************************************************************/
int geotiff_get_file_metadata
(
    GEOTIFFIO *gtiff_file,              /* I: GEOTIFFIO structure */
    GEOTIFF_FILE_METADATA *metadata     /* O: Output metadata */
)
{
    GTIF *gtif = gtiff_file->gtif; /* GTIF pointer */
    geocode_t model;               /* Modelling system type */
    geocode_t angular_units;       /* Angular units */
    geocode_t linear_units;        /* Linear units */
    tagtype_t type;                /* Tag type */
    int citation_length;           /* Citation length */
    int size;                      /* Citation type size */
    int status;                    /* Routine return placeholder */
    int index;                     /* Generic loop index */

    /* Initialize the metadata structure */
    memset(metadata, 0, sizeof(*metadata));

    /* Initialize the projection parameters */
    for (index = 0; index < IAS_PROJ_PARAM_SIZE; index++)
    {
        metadata->projection_parameters[index] = 0.0;
    }

    /* Copy the fields that are always the same */
    /* Copy the datum to the metadata */
    status = snprintf(metadata->datum, IAS_DATUM_SIZE, "WGS84");
    if (status < 0 || status >= IAS_DATUM_SIZE)
    {
        IAS_LOG_ERROR("Copying DATUM to metadata");
        return ERROR;
    }

    /* Set the standard spheroid code */
    metadata->spheroid_code = WGS84_SPHEROID;

    /* Default the projection units to the meters (may change later based
       on the projection found) */
    status = snprintf(metadata->projection_units, IAS_UNITS_SIZE, "METERS");
    if (status < 0 || status >= IAS_UNITS_SIZE)
    {
        IAS_LOG_ERROR("Copying projection units to metadata");
        return ERROR;
    }

    /* Ensure the projection type is supported */
    if (!GTIFKeyGet(gtif, GTModelTypeGeoKey, &model, GEO_KEY_INDEX,
        GEO_KEY_COUNT))
    {
        IAS_LOG_ERROR("No Model Type found");
        return ERROR;
    }

    if (model != ModelTypeProjected && model != ModelTypeGeographic)
    {
        IAS_LOG_ERROR("Unsupported Model Type %hu", model);
        return ERROR;
    }

    /* Get the citation key length */
    citation_length = GTIFKeyInfo(gtif, GTCitationGeoKey, &size, &type);

    /* Handle the geographic projection */
    if (model == ModelTypeGeographic)
    {
        /* Set the projection code to GEO with no zone */
        metadata->projection_code = GEO;
        metadata->zone_code = 0;

        /* Copy the projection units to the metadata */
        status = snprintf(metadata->projection_units, IAS_UNITS_SIZE,
            "DEGREES");
        if (status < 0 || status >= IAS_UNITS_SIZE)
        {
            IAS_LOG_ERROR("Copying projection units to metadata");
            return ERROR;
        }
    }
    /* Use the citation string to quickly determine the projection since
       we're only supporting UTM and PS at this time */
    else if (citation_length > 0)
    {
        char *citation; /* Citation pointer */

        /* Allocate memory for the citation */
        citation = malloc(size * citation_length);
        if (!citation)
        {
            IAS_LOG_ERROR("Allocating citation string");
            return ERROR;
        }

        /* Retrieve the citation key */
        if (!GTIFKeyGet(gtif, GTCitationGeoKey, citation, GEO_KEY_INDEX,
            citation_length))
        {
            IAS_LOG_ERROR("Getting Projection Citation");
            free(citation);
            return ERROR;
        }

        /* Check the citation string for the projection information */
        if (strncmp(citation, "UTM", 3) == 0)
        {
            int zone; /* UTM zone */

            /* Get the UTM zone */
            if (sscanf(citation, "UTM Zone %d", &zone) != 1)
            {
                IAS_LOG_ERROR("Getting UTM Zone: %s", citation);
                free(citation);
                return ERROR;
            }
            metadata->projection_code = UTM;
            metadata->zone_code = zone;
        }
        else if (strncmp(citation, "PS", 2) == 0)
        {
            double central_meridian; /* Central meridian */
            double lat_of_origin;    /* Latitude of origin */
            double false_easting;    /* False easting */
            double false_northing;   /* False northing */

            /* Set the projection information for polar stereographic */
            metadata->projection_code = PS;
            metadata->zone_code = 0;

            /* Get the projection parameters important for PS */
            /* Central meridian */
            if (!GTIFKeyGet(gtif, ProjStraightVertPoleLongGeoKey, 
                &central_meridian, GEO_KEY_INDEX, GEO_KEY_COUNT))
            {
                IAS_LOG_ERROR("Getting Central Meridian");
                free(citation);
                return ERROR;
            }
            
            /* Convert central meridian to DMS */
            if (ias_geo_convert_deg2dms(central_meridian,
                &metadata->projection_parameters[4], "LON") != SUCCESS)
            {
                IAS_LOG_ERROR("Converting Central Meridian to DMS");
                free(citation);
                return ERROR;
            }

            /* Latitude of origin */
            if (!GTIFKeyGet(gtif, ProjNatOriginLatGeoKey, &lat_of_origin, 
                GEO_KEY_INDEX, GEO_KEY_COUNT))
            {
                IAS_LOG_ERROR("Getting Latitude of Origin");
                free(citation);
                return ERROR;
            }
 
            /* Convert latitude of origin to DMS */
            if (ias_geo_convert_deg2dms(lat_of_origin,
                &metadata->projection_parameters[5], "LAT") != SUCCESS)
            {
                IAS_LOG_ERROR("Converting Latitude of Origin to DMS");
                free(citation);
                return ERROR;
            }

            /* False easting */
            if (!GTIFKeyGet(gtif, ProjFalseEastingGeoKey, &false_easting, 
                GEO_KEY_INDEX, GEO_KEY_COUNT))
            {
                IAS_LOG_ERROR("Getting False Easting");
                free(citation);
                return ERROR;
            }
            metadata->projection_parameters[6] = false_easting;

            /* False northing */
            if (!GTIFKeyGet(gtif, ProjFalseNorthingGeoKey, &false_northing, 
                GEO_KEY_INDEX, GEO_KEY_COUNT))
            {
                IAS_LOG_ERROR("Getting False Easting");
                free(citation);
                return ERROR;
            }
            metadata->projection_parameters[7] = false_northing;
        }
        else
        {
            IAS_LOG_ERROR("Unsupported Projection: %s", citation);
            free(citation);
            return ERROR;
        }

        /* Check the citation string for the expected WGS84 datum */
        if (strstr(citation, "WGS84") == NULL)
        {
            IAS_LOG_ERROR("Unsupported Datum: %s", citation);
            free(citation);
            return ERROR;
        }

        /* Done with the citation string parsing, so free */
        free(citation);
        citation = NULL;
    }
    else
    {
        short projType;  /* Geo Key for zone/hemisphere */
        int utm_zone;    /* UTM zone of projection */
        int projection_hemisphere;

        IAS_LOG_INFO("Projection Citation not present using GeoKeys to get "
            "the projection information");

        /* Retrieve the projection type key */
        if(!GTIFKeyGet(gtif, ProjectedCSTypeGeoKey, &projType, GEO_KEY_INDEX,
            GEO_KEY_COUNT))
        {
            IAS_LOG_ERROR("Error getting ProjectedCSTypeGeoKey");
            return ERROR;
        }

        /* Retrieve the UTM zone and north south extent */
        if(geotiff_get_utm_zone(projType, &projection_hemisphere, &utm_zone) 
           != SUCCESS)
        {
            IAS_LOG_ERROR("ProjectedCSTypeGeoKey %d does not appear to be "
                "a supported UTM zone; only UTM projection is currently "
                "supported", projType);
            return ERROR;
        }

        /* Check if coordinates are in southern hemisphere */
        gtiff_file->zone_is_southern_hemisphere = 0;
        if (projection_hemisphere == MapSys_UTM_South)
        {
            gtiff_file->zone_is_southern_hemisphere = 1;
        }

        /* Set the zone information */
        metadata->projection_code = UTM;
        metadata->zone_code = utm_zone;
    }

    /* Get the projection units - expecting Meters only */
    if (!GTIFKeyGet(gtif, GeogLinearUnitsGeoKey, &linear_units, GEO_KEY_INDEX,
        GEO_KEY_COUNT))
    {
        if (!GTIFKeyGet(gtif, ProjLinearUnitsGeoKey, &linear_units, 
            GEO_KEY_INDEX, GEO_KEY_COUNT))
        {
            IAS_LOG_ERROR("Getting Linear Units");
            return ERROR;
        }
    }
    
    if (linear_units != Linear_Meter)
    {
        IAS_LOG_ERROR("Unsupported Linear Units %hu", linear_units);
        return ERROR;
    }

    /* Expecting angular values to be measured in angles, so validate
       the image's angular units */
    if (!GTIFKeyGet(gtif, GeogAngularUnitsGeoKey, &angular_units, GEO_KEY_INDEX,
        GEO_KEY_COUNT))
    {
        IAS_LOG_ERROR("Getting Angular Units");
        return ERROR;
    }

    if (angular_units != Angular_Degree)
    {
        IAS_LOG_ERROR("Unsupported Angular Units %hu", angular_units);
        return ERROR;
    }

    return SUCCESS;
}

/******************************************************************************
Name: geotiff_get_band_metadata

Purpose: Get the band-level metadata from the open GeoTIFF band and save
         the information in the metadata structure

Returns:
    SUCCESS
    ERROR
******************************************************************************/
int geotiff_get_band_metadata
(
    const GEOTIFF_BAND_IO *gtiff_band,  /* I: GEOTIFF_BAND_IO structure */
    GEOTIFF_BAND_METADATA *metadata     /* O: Output metadata */
)
{
    /* Setup local pointers to the TIFF-level descriptor and
       the Geo Key parser handle for easy use */
    TIFF *tif = gtiff_band->gtiff_file->tif;
    double *pixel_scale_xyz;    /* Projection distance X/Y/Z */
    double *UL_tie_point;
    double east_boundary_x;
    double south_boundary_y;
    short pixel_scale_size;
    short UL_tie_point_size;
    char *filename = gtiff_band->gtiff_file->filename;
    int status;

    /* Initialize the metadata structure */
    memset(metadata, 0, sizeof(*metadata));

    /* Setup the band number and band name metadata items */
    metadata->band_number = gtiff_band->band_number;

    status = snprintf(metadata->band_name, IAS_BAND_NAME_SIZE,
            "Converted Band %d", gtiff_band->band_number);
    if (status < 0 || status >= IAS_BAND_NAME_SIZE)
    {
        IAS_LOG_ERROR("Setting up band name for GeoTIFF file %s", filename);
        return ERROR;
    }
    metadata->band_name[IAS_BAND_NAME_SIZE - 1] = '\0';

    /* Get the projection distance (pixel size) */
    if (!TIFFGetField(tif, TIFFTAG_GEOPIXELSCALE, &pixel_scale_size,
                &pixel_scale_xyz))
    {
        IAS_LOG_ERROR("Getting Projection Distance from GeoTIFF file %s",
                filename);
        return ERROR;
    }
    metadata->projection_distance_x = pixel_scale_xyz[0];
    metadata->projection_distance_y = pixel_scale_xyz[1];

    /* Get the UL origin corner */
    if (!TIFFGetField(tif, TIFFTAG_GEOTIEPOINTS, &UL_tie_point_size,
                &UL_tie_point))
    {
        IAS_LOG_ERROR("Getting UL Origin Tiepoint from GeoTIFF file %s",
                filename);
        return ERROR;
    }
    metadata->upper_left_x = UL_tie_point[3];
    metadata->upper_left_y = UL_tie_point[4];

    /* Adjust the upper left corner to reference the center of the pixel 
       if the raster type is anything but RasterPixelIsPoint */
    if (!gtiff_band->gtiff_file->pixel_is_point)
    {
       metadata->upper_left_x += metadata->projection_distance_x / 2.0;
       metadata->upper_left_y -= metadata->projection_distance_y / 2.0;
    }

    /* Adjust the y coordinates if projection is in the southern hemisphere */
    if (gtiff_band->gtiff_file->zone_is_southern_hemisphere)
    {
        IAS_LOG_INFO("Adjusting points in southern hemisphere by 10 million");
        metadata->upper_left_y -= 10000000;
        metadata->upper_right_y -= 10000000;
        metadata->lower_left_y -= 10000000;
        metadata->lower_right_y -= 10000000;
    }

    /* Calculate the opposite boundaries to setup remaining corner points */
    east_boundary_x = metadata->upper_left_x
        + (metadata->projection_distance_x * (gtiff_band->sample_count - 1));
    south_boundary_y = metadata->upper_left_y
        - (metadata->projection_distance_y * (gtiff_band->line_count - 1));

    /* Finish defining the corner coordinates */
    metadata->upper_right_x = east_boundary_x;
    metadata->upper_right_y = metadata->upper_left_y;
    metadata->lower_left_x = metadata->upper_left_x;
    metadata->lower_left_y = south_boundary_y;
    metadata->lower_right_x = east_boundary_x;
    metadata->lower_right_y = south_boundary_y;

    return SUCCESS;
}

#if 0
/* Commented out until deployed as a library. This will likely be used by
   the writing aspects of the library which have not been implemented yet. */

/******************************************************************************
Name: get_gtiff_bit_depth

Purpose: Get the number of bits in data type from the IAS_DATA_TYPE

Returns:
    The number of the bit depth (8 or 16 supported at this time)
    ERROR if unsupported data type
******************************************************************************/
static int get_gtiff_bit_depth
(
    IAS_DATA_TYPE data_type     /* I: The IAS data type */
)
{
    /* Only support 8-bit or 16-bit data, otherwise signal error */
    if (data_type == IAS_BYTE)
        return 8;
    else if ((data_type == IAS_I2) || (data_type == IAS_UI2))
        return 16;

    return ERROR;
}

/******************************************************************************
Name: get_gtiff_sample_format

Purpose: Get the GeoTIFF data type correlating to the IAS_DATA_TYPE

Returns:
    The GeoTIFF data type for the SAMPLEFORMAT tag
******************************************************************************/
static int get_gtiff_sample_format
(
    IAS_DATA_TYPE data_type     /* I: The IAS data type */
)
{
    /* Due to error checking in get_gtiff_bit_depth, there are only a couple
       data types expected at this point */
    if (data_type == IAS_BYTE)
        return SAMPLEFORMAT_UINT;
    else if (data_type == IAS_I2)
        return SAMPLEFORMAT_INT;

    /* data_type == IAS_UI2; error checked by get_gtiff_bit_depth */
    return SAMPLEFORMAT_UINT;
}

/******************************************************************************
Name: get_gtiff_data_type

Purpose: For an IAS_DATA_TYPE, get the matching GeoTIFF data type definition

Returns:
    SUCCESS
    ERROR
******************************************************************************/
static int get_gtiff_data_type
(
    IAS_DATA_TYPE data_type,    /* I: IAS data type to convert */
    int *bits_per_sample,       /* O: GeoTIFF TIFFTAG_BITSPERSAMPLE value */
    int *sample_format          /* O: GeoTIFF TIFFTAG_SAMPLEFORMAT value */
)
{
    int return_val;

    /* Get the bits per data value */
    return_val = get_gtiff_bit_depth(data_type);
    if (return_val == ERROR)
    {
        IAS_LOG_ERROR("Unsupported IAS data type %d", data_type);
        return ERROR;
    }
    *bits_per_sample = return_val;

    /* Get the data format of the data values */
    *sample_format = get_gtiff_sample_format(data_type);

    return SUCCESS;
}

/* Prototypes for I/O library routines not yet implemented (only need read
   abilities at this time for the GeoTIFF converter). The following routines
   will need to be implemented when this code is deployed as a library.
   Look to the header of this file for file organization guidance (based on
   the L1G library files) for which routines should be grouped together and
   look to the L1G library for file naming guidance. */

/****************************************************************************
NOTE: Not implemented at this time, as the geotiff converter this is
      initially being implemented as part of does not need any writing
      capabilities at this time, but this will be fleshed out when these
      routines are moved out of the converter and implemented as a
      library.
****************************************************************************/
int geotiff_write_image
(
    const GEOTIFF_BAND_IO *gtiff_band,  /* I: GEOTIFF_BAND_IO structure */
    int start_line,                     /* I: Line to start writing (0-rel) */
    int start_sample,                   /* I: Sample to start writing (0-rel) */
    int line_count,                     /* I: Number of lines to write */
    int sample_count,                   /* I: Number of samples to write */
    void *data                          /* I: Data buffer */
);


/****************************************************************************
NOTE: Set metadata routines not implemented at this time, as the geotiff
      converter this is initially being implemented as part of does not
      need any writing capabilities at this time, but this will be
      fleshed out when these routines are moved out of the converter and
      implemented as a library.
****************************************************************************/
int geotiff_set_file_metadata
(
    GEOTIFFIO *gtiff_file,                  /* I: GEOTIFFIO structure */
    const GEOTIFF_FILE_METADATA *metadata   /* I: metadata to set */
);


/****************************************************************************
NOTE: Set metadata routines not implemented at this time, as the geotiff
      converter this is initially being implemented as part of does not
      need any writing capabilities at this time, but this will be
      fleshed out when these routines are moved out of the converter and
      implemented as a library.
****************************************************************************/
int geotiff_set_band_metadata
(
    GEOTIFFIO *gtiff_file,                  /* I: file to set metadata for */
    const GEOTIFF_BAND_METADATA *metadata,  /* I: array of metadata to set */
    int metadata_size                       /* I: number of entries in metadata
                                                  array */
);
#endif

