/******************************************************************************
Name: get_conversion_file_metadata
      get_conversion_band_metadata

Purpose: Get elements of the file-level and band-level metadata from the open
         GeoTIFF file that are done specifically for this conversion tool
         circumstances, then save the information in the metadata structure.
         These are aspects of metadata preparation for the L1G file that will
         be created that doesn't make sense to put into the set_metadata
         functions in geotiff_io_tools since those routines are eventually
         desired to be moved to a separate GeoTIFF I/O library, so we don't
         want to embed application-specific assumptions or dependencies into
         the code that is going to be a library in the future where we'd
         ultimate need to de-couple those aspects again.
******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "ias_logging.h"
#include "ias_math.h"
#include "geotiff_io.h"
#include "convert_geotiff_to_l1g.h"

/******************************************************************************
Name: parse_date_from_filename

Purpose: Parse the filename to determine the date of image acquisition

Returns:
    SUCCESS
    ERROR
******************************************************************************/
static int parse_date_from_filename
(
    const char *filename,   /* I: The GeoTIFF filename */
    int is_l7_file,         /* I: Flag indicating if this is an L7 file */
    int is_l7_ee_file,      /* I: Flag indicating if this is L7 file from EE */
    int is_l8_file,         /* I: Flag indicating if this is an L8 file */
    char *date              /* O: YYYY/MM/DD format */
)
{
    const char *token;          /* Pointer into the filename string */
    const char *filename_ptr;   /* Pointer at filename portion of string */

    /* Find the last '/', or if there is none assume this is a filename
       only with no path component. If '/' found, advance one more char to
       the beginning of the filename. */
    token = strrchr(filename, '/');
    if (token)
        token++;
    else
        token = filename;
    filename_ptr = token;

    /* The different satellites have different filename formats and
       dates are encoded in the filenames at different locations and
       using different date styles */
    if (is_l7_file)
    {
        int i;

        token = strstr(filename_ptr, "_B");
        if (token == NULL)
        {
            IAS_LOG_ERROR("Parsing date from L7 filename: %s", filename_ptr);
            return ERROR;
        }

        /* Back up the pointer 8 characters to position it at the
           start of the date portion of the filename */
        token -= 8;
        if (token < filename_ptr)
        {
            IAS_LOG_ERROR("Parsing date from L7 filename: %s", filename_ptr);
            return ERROR;
        }

        /* Copy the date to the output string pointer in proper format */
        for (i = 0; i < 4; i++)
            date[i] = *token++;
        date[4] = '/';
        date[5] = *token++;
        date[6] = *token++;
        date[7] = '/';
        date[8] = *token++;
        date[9] = *token++;
        date[10] = '\0';
    }
    else if (is_l8_file || is_l7_ee_file)
    {
        char year_string[5];
        char doy_string[4];
        int year;
        int doy;
        int mon;
        int day;
        int status;

        token = strstr(filename_ptr, "_B");
        if (token == NULL)
        {
            IAS_LOG_ERROR("Parsing date from filename: %s", filename_ptr);
            return ERROR;
        }

        /* Back up the pointer 12 characters to position it at the
           start of the date portion of the filename */
        token -= 12;
        if (token < filename_ptr)
        {
            IAS_LOG_ERROR("Parsing date from filename: %s", filename_ptr);
            return ERROR;
        }

        /* Now pointing at the start of the date portion of the filename.
           The 4-digit year is first. */
        strncpy(year_string, token, 4);
        year_string[4] = '\0';

        /* Move the token past the year to get the julian day of year */
        token += 4;
        strncpy(doy_string, token, 3);
        doy_string[3] = '\0';

        /* Convert the year and julian day of year strings to integers */
        if (sscanf(year_string, "%04d", &year) != 1)
        {
            IAS_LOG_ERROR("Converting year %s to integer", year_string);
            return ERROR;
        }
        if (sscanf(doy_string, "%03d", &doy) != 1)
        {
            IAS_LOG_ERROR("Converting day of year %s to integer", doy_string);
            return ERROR;
        }

        /* Convert the julian date to month/day of month */
        if (ias_math_convert_doy_to_month_day(doy, year, &mon, &day) != SUCCESS)
        {
            IAS_LOG_ERROR("Parsing date from filename: %s", filename_ptr);
            return ERROR;
        }

        /* Copy the date to the output string pointer in proper format */
        status = snprintf(date, GEOTIFF_CAPTURE_DATE_SIZE, "%04d/%02d/%02d",
                year, mon, day);
        if (status < 0 || status >= GEOTIFF_CAPTURE_DATE_SIZE)
        {
            IAS_LOG_ERROR("Compiling date string");
            return ERROR;
        }
    }

    return SUCCESS;
}

/******************************************************************************
Name: get_conversion_file_metadata

Purpose: Get the file-level metadata from the open GeoTIFF file and save
         the information in the metadata structure.

Returns:
    SUCCESS
    ERROR
******************************************************************************/
int get_conversion_file_metadata
(
    const GEOTIFFIO *gtiff_file,        /* I: GEOTIFFIO structure */
    GEOTIFF_FILE_METADATA *metadata     /* O: Output metadata */
)
{
    const char *base_filename;      /* Pointer into the filename string */
    char path_string[4];            /* Path string parsed out of filename */
    char row_string[4];             /* Row string parsed out of filename */
    int is_l7_file_flag = 0;        /* TRUE if this is a L7 file */
    int is_l7_ee_file_flag = 0;     /* TRUE if this is a L7 file from EE */
    int is_l8_file_flag = 0;        /* TRUE if this is a L8 file */

    /* Setup flags for the two most supported file types to process */
    if (is_l7_file(gtiff_file->filename))
        is_l7_file_flag = 1;
    if (is_l7_ee_file(gtiff_file->filename))
        is_l7_ee_file_flag = 1;
    if (is_l8_file(gtiff_file->filename))
        is_l8_file_flag = 1;

    /* Get the path/row information from the filename if the file matches
       the expected L7 or L8 file format. Set the capture direction based
       on the WRS row. */
    if (is_l7_file_flag || is_l7_ee_file_flag || is_l8_file_flag)
    {
        /* Find the last '/', or if there is none assume this is a filename
           only with no path component. If '/' found, advance one more char to
           the beginning of the filename. */
        base_filename = strrchr(gtiff_file->filename, '/');
        if (base_filename)
            base_filename++;
        else
            base_filename = gtiff_file->filename;

        /* Do a basic length check on the filename to know the following
           parsing won't be attempting to read past the end of the filename
           string */
        if (strlen(base_filename) < 9)
        {
            IAS_LOG_ERROR("Filename is not minimum length expected for"
                    "parsing out WRS path/row elements");
            return ERROR;
        }

        /* Grab the path and row characters out of the filename */
        strncpy(path_string, &base_filename[3], 3);
        path_string[3] = '\0';
        strncpy(row_string, &base_filename[6], 3);
        row_string[3] = '\0';
        base_filename = NULL;

        /* Convert to integer and save in the metadata struct */
        if (sscanf(path_string, "%03d", &metadata->wrs_path) != 1)
        {
            IAS_LOG_ERROR("Converting WRS path %s to integer", path_string);
            return ERROR;
        }
        if (sscanf(row_string, "%03d", &metadata->wrs_row) != 1)
        {
            IAS_LOG_ERROR("Converting WRS row %s to integer", row_string);
            return ERROR;
        }

        /* Set the capture direction based on the row */
        if ((metadata->wrs_row >= 123) && (metadata->wrs_row <= 246))
        {
            strncpy(metadata->capture_direction, "ASCENDING",
                    GEOTIFF_CAPTURE_DIRECTION_SIZE);
            metadata->capture_direction[GEOTIFF_CAPTURE_DIRECTION_SIZE - 1]
                = '\0';
        }
        else /* row is in 247,1-122 */
        {
            strncpy(metadata->capture_direction, "DESCENDING",
                    GEOTIFF_CAPTURE_DIRECTION_SIZE);
            metadata->capture_direction[GEOTIFF_CAPTURE_DIRECTION_SIZE - 1]
                = '\0';
        }
    }

    /* Set the spacecraft based on the file type determined from the filename */
    if (is_l7_file_flag || is_l7_ee_file_flag)
        strncpy(metadata->spacecraft, "L7", GEOTIFF_SPACECRAFT_SIZE);
    else if (is_l8_file_flag)
        strncpy(metadata->spacecraft, "L8", GEOTIFF_SPACECRAFT_SIZE);
    else
        metadata->spacecraft[0] = '\0';
    metadata->spacecraft[GEOTIFF_SPACECRAFT_SIZE - 1] = '\0';

    /* Get the acquisition date from the filename */
    if (parse_date_from_filename(gtiff_file->filename, is_l7_file_flag,
                is_l7_ee_file_flag, is_l8_file_flag, metadata->capture_date)
            != SUCCESS)
    {
        IAS_LOG_ERROR("Extracting acquisition date from the filename %s",
                gtiff_file->filename);
        return ERROR;
    }

    return SUCCESS;
}
