/*************************************************************************
NAME:    read_parameters

PURPOSE: Read the input and output file names from the command line

NOTES:   The file name strings returned are alloc'd within this routine
         and need to be free'd by the caller. Allocated memory is free'd
         within the routine upon error return status.

RETURNS:
    SUCCESS     parameters read and validated successfully
    ERROR       invalid parameters
**************************************************************************/
/* Standard Library Includes */
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

/* IAS Library Includes */
#include "ias_const.h"
#include "ias_logging.h"

/* Local Defines */
#define USAGE_MSG "\nUsage: convert_geotiff_to_l1g -i <input geotiff band " \
            "file> -s <starting band> <output L1G file>\n" \
            "--- Arguments:\n" \
            "       [-i] Required: must be included before each band and\n" \
            "            bands must be sequential.\n" \
            "       [-s] Optional: Indicates starting band number\n" \
            "            Default: 1\n" \
            "--- Examples:\n" \
            "       convert_geotiff_to_l1g -i band1.TIF -i band2.TIF l1g.h5\n" \
            "       convert_geotiff_to_l1g -i band9.TIF -i band10.TIF l1g.h5 " \
            " -s 9 l1g.h5\n" \
            "       convert_geotiff_to_l1g -i band5.TIF -s 5 l1g.h5\n"

int read_parameters
(
    int argc,                /* I: count of command line arguments */
    char *argv[],            /* I: command line arguments array */
    int max_files,           /* I: Max number of input GTIFF files */
    char *gtiff_file[],      /* O: GeoTIFF file name to convert */
    char **l1g_file,         /* O: Name of output L1G file to convert to */
    int *total_bands,        /* O: Total number of band file names read in */
    int *starting_band_number/* O: Starting band number in the list */
)
{
    int band_count;             /* Band files count */
    int band_number;            /* Local band number */

    /* Verify the needed input and output parameters were provided */
    if (argc < 3)
    {
        printf(USAGE_MSG);
        exit(EXIT_FAILURE);
    }

    /* Read the output L1G filename */
    *l1g_file = argv[argc - 1];
    if (!*l1g_file)
    {
        IAS_LOG_ERROR("Allocating memory for L1G filename");
        return ERROR;
    }
    argc--;

    /*Initialize the band variables */
    band_number = 1;
    band_count = 0;
     
    /* If only 2 arguments are provided then only GeoTIFF and output L1G were
       specified */
    if (argc == 2)
    {
        gtiff_file[band_count] = argv[1];
        if (!gtiff_file[band_count])
        {
            IAS_LOG_ERROR("Allocating memory for GeoTIFF filename");
            return ERROR;
        }

        /* Validate the input GeoTIFF file exists and that it is readable */
        if (access(gtiff_file[band_count], R_OK) != 0)
        {
            IAS_LOG_ERROR("Cannot access GeoTIFF source file: %s", 
                gtiff_file[band_count]);
            return ERROR;
        }
        band_count++;
    }
    else
    {
        int index; /* Parameter loop variable */

        /* Read in all of the band file names and starting band */
        for (index = 0; index < argc; index++)
        {
            /* Check if another band is included */
            if (strcmp(argv[index], "-i") == 0)
            {
                /* Check that there are still parameters left */
                if (index + 1 >= argc)
                {
                    IAS_LOG_ERROR("No GeoTIFF files specifed after the last "
                        "-i option");
                    return ERROR;
                }

                /* Check the file count */
                if (band_count >= max_files)
                {
                    IAS_LOG_ERROR("Too many input files were specified only "
                        "%d files supported", max_files);
                    return ERROR;
                }

                /* Read the input GeoTIFF filename */
                gtiff_file[band_count] = argv[++index];
                if (!gtiff_file[band_count])
                {
                    IAS_LOG_ERROR("Allocating memory for GeoTIFF filename");
                    return ERROR;
                }

                /* Validate the input GeoTIFF file exists and that it is 
                   readable */
                if (access(gtiff_file[band_count], R_OK) != 0)
                {
                    IAS_LOG_ERROR("Cannot access GeoTIFF source file: %s", 
                        gtiff_file[band_count]);
                    return ERROR;
                }
                band_count++;
            }
            else if (strcmp(argv[index], "-s") == 0)
            {     
                /* Check that there are still parameters left */
                if (index + 1 >= argc)
                {
                    IAS_LOG_ERROR("No band number specifed after the "
                        "-s option");
                    return ERROR;
                }

                /* Process the starting band number */
                band_number = atoi(argv[++index]);
                if (band_number < 1)
                {
                    IAS_LOG_ERROR("Invalid band number %d used as the "
                        "starting band number", band_number);
                    return ERROR;
                }
            }
        }
    }

    /* Make sure there were GeoTIFF band files found */
    if (band_count <= 0)
    {
        IAS_LOG_ERROR("No GeoTiff files were passed as arguments");
        return ERROR;
    }
    
    /* Output the total bands and starting band numbers */
    *total_bands = band_count;
    *starting_band_number = band_number;

    return SUCCESS;
}
