/*****************************************************************************
Name:           convert_geotiff_to_l1g_test

Purpose:
    The regression test for the GeoTIFF to L8 HDF conversion tool
*****************************************************************************/
#include <stdio.h>
#include <limits.h>
#include "cgreen/cgreen.h"

/* Local Defines */
#define NUMBER_OF_CONVERSION_TESTS 3

static char *unit_test_dir = NULL;

/* A basic test of the converter */
Ensure test_convert_geotiff_to_l1g()
{
    const char *test_data_dir = "ias_base/gps/tools/convert_geotiff_to_l1g";
    const char *test_image[NUMBER_OF_CONVERSION_TESTS];
    const char *l1g_hdf_file[NUMBER_OF_CONVERSION_TESTS];
    const char *l1g_verify_file[NUMBER_OF_CONVERSION_TESTS];
    const char *test_name[NUMBER_OF_CONVERSION_TESTS];
    int index;

    /* Initialize the test variables */
    test_image[0] = "L71037037_03720000521_B70.TIF";
    test_image[1] = "L71037037_03720000521_B4.TIF";
    test_image[2] = "L71037037_03720000521_B9.TIF";
    l1g_hdf_file[0] = "l1g_test_V2.h5"; 
    l1g_hdf_file[1] = "l1g_raster_V2.h5"; 
    l1g_hdf_file[2] = "l1g_citation_V2.h5"; 
    l1g_verify_file[0] = "l1g_test_V3.h5"; 
    l1g_verify_file[1] = "l1g_raster_V3.h5"; 
    l1g_verify_file[2] = "l1g_citation_V3.h5"; 
    test_name[0] = "Normal Conversion"; 
    test_name[1] = "Raster is Area and Default Sample size"; 
    test_name[2] = "Missing Citation with Multiple Directories"; 


    /* Loop through the conversion tests */
    for (index = 0; index < NUMBER_OF_CONVERSION_TESTS; index++)
    {
        char command[PATH_MAX * 2];
        int status;

        /* Print out the test that is currently running */
        printf("\n-- Running %s test \n", test_name[index]);

        /* Create the conversion command */
        status = snprintf(command, sizeof(command),
            "../convert_geotiff_to_l1g %s/%s/%s %s", unit_test_dir, 
            test_data_dir, test_image[index], l1g_hdf_file[index]);
        if (status < 0 || status >= sizeof(command))
        {
            assert_true_with_message(0, "Error preparing system command "
                "to execute the GeoTIFF to L1G HDF converter application");
            return;
        }
        
        /* Run the conversion command */
        status = system(command);
        if (status != 0)
        {
            assert_true_with_message(0, "GeoTIFF to L1G HDF Conversion failed. "
                "Command: %s", command);
            return;
        }

        /* Prepare the comparision command */
        status = snprintf(command, sizeof(command), "h5diff %s %s/%s/%s",
            l1g_hdf_file[index], unit_test_dir, test_data_dir, 
            l1g_verify_file[index]);
        if (status < 0 || status >= sizeof(command))
        {
            assert_true_with_message(0, "Error preparing system command "
                    "for validating the L1G HDF file");
            return;
        }
    
        /* Now compare the output file with one previously benchmarked */
        status = system(command);
        if (status != 0)
        {
            assert_true_with_message(0, "Differences found with GeoTIFF to L1G "
                "HDF Conversion output. Review %s and %s/%s/%s for details",
                l1g_hdf_file[index], unit_test_dir, test_data_dir,
                l1g_verify_file[index]);
            return;
        }
    }
}

Ensure test_convert_geotiff_to_l1g_multiple()
{
    const char *test_data_dir = "ias_base/gps/tools/convert_geotiff_to_l1g";
    const char *test_image[2];
    test_image[0] = "L71037037_03720000521_B70.TIF";
    test_image[1] = "L71037037_03720000521_B70.TIF";
    const char *l1g_hdf_file = "l1g_multiple.h5";
    char command[PATH_MAX * 2];
    int status;

    /* Print out the test that is currently running */
    printf("\n-- Running Multiple GeoTIFF file test \n");

    /* Run the converter on an image */
    status = snprintf(command, sizeof(command),
        "../convert_geotiff_to_l1g -s 5 -i %s/%s/%s -i %s/%s/%s %s",
        unit_test_dir, test_data_dir, test_image[0], unit_test_dir,
        test_data_dir, test_image[1], l1g_hdf_file);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
            "to execute the GeoTIFF to L1G HDF converter application");
        return;
    }

    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "GeoTIFF to L1G HDF Conversion failed. "
            "Command: %s", command);
        return;
    }

    /* Now compare the output file with one previously benchmarked */
    status = snprintf(command, sizeof(command), "h5diff %s %s/%s/%s",
        l1g_hdf_file, unit_test_dir, test_data_dir, l1g_hdf_file);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Error preparing system command "
            "for validating the L1G HDF file");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found with GeoTIFF to L1G "
            "HDF Conversion output. Review %s and %s/%s/%s for details",
            l1g_hdf_file, unit_test_dir, test_data_dir, l1g_hdf_file);
    }
}

int main(int argc, char *argv[])
{
    TestSuite *suite = create_test_suite();

    add_test(suite, test_convert_geotiff_to_l1g);
    add_test(suite, test_convert_geotiff_to_l1g_multiple);

    unit_test_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!unit_test_dir)
    {
        printf("The IAS_UNIT_TEST_DATA_DIR environment variable is not set\n");
        exit(EXIT_FAILURE);
    }

    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
