/* Standard Library Includes */
#include <stdlib.h>
#include <string.h>
#include <limits.h>

/* IAS Library Includes */
#include "ias_model_io.h"
#include "ias_grid_io.h"       
#include "ias_logging.h"        
#include "ias_angle_gen.h"        
#include "ias_miscellaneous.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"

/* Local defines */
#define USE_MESSAGE "Usage: create_angle_coeff_file <parameter filename>\n" 

/******************************************************************************
NAME: read_omf_parameters

PURPOSE: This function reads all the parameters from the OMF that are needed.

RETURN VALUE: Type = int
    Value    Description
    -----    -----------
    SUCCESS  The parameters were successfully read
    ERROR    An error occurred reading the parameters
******************************************************************************/
static int read_omf_parameters
(
    const char *omf_filename,             /* I: OMF filename */
    int processing_pass,                  /* I: Current processing pass */
    char *l0r_filename,                   /* O: L0R input filename */
    char *cpf_filename,                   /* O: CPF input filename */
    char *model_filename,                 /* O: Model input filename */
    char *grid_filename,                  /* O: Grid input filename */
    int *wrs_path,                        /* O: Target wrs path */
    int *wrs_row                          /* O: Target wrs row */
)
{
    int count;                /* Table parameter count */
    int help_requested;       /* Status placeholder */

    /* Start parameter table definition */
    IAS_PARM_DECLARE_TABLE(omf_table, 6);

    /* Retrieve the wrs path */
    IAS_PARM_TARGET_WRS_PATH(omf_table, IAS_PARM_REQUIRED, wrs_path,
        sizeof(int), 1);

    /* Retrieve the wrs row */
    IAS_PARM_TARGET_WRS_ROW(omf_table, IAS_PARM_REQUIRED, wrs_row,
        sizeof(int), 1);

    /* Retrieve the L0R filename */
    IAS_PARM_L0R_FILENAME(omf_table, IAS_PARM_REQUIRED, l0r_filename,
        sizeof(char) * PATH_MAX, 1);

    /* Retrieve the CPF filename */
    IAS_PARM_CPF(omf_table, IAS_PARM_REQUIRED, cpf_filename, 
        sizeof(char) * PATH_MAX, 1);

    /* Retrieve the CPF filename */
    IAS_PARM_MODEL_FILENAME_FROM_PASS(omf_table, IAS_PARM_REQUIRED, 
        model_filename, sizeof(char) * PATH_MAX, 1, processing_pass);

    /* Retrieve the Grid filename */
    IAS_PARM_GRID_FILENAME_FROM_PASS(omf_table, IAS_PARM_REQUIRED, 
        grid_filename, sizeof(char) * PATH_MAX, 1, processing_pass);

    /* Retrieve parameter count for OMF table */
    count = IAS_PARM_GET_TABLE_SIZE(omf_table);

    /* Call the ias_parm_provide_help routine to provide help for the
       parameters if requested; also verifies the OMF file exists */
    help_requested = ias_parm_provide_help(omf_filename, omf_table, count,
            IAS_OMF_PARAMETERS);
    if (help_requested)
    {
        /* Either help was provided or an error occurred*/
        if (help_requested == TRUE)
        {
            /* Successfully provided help, as requested, so just exit */
            exit(EXIT_SUCCESS);
        }
        
        IAS_LOG_ERROR("Error occurred providing OMF help");
        return ERROR;
    }

    /* Read the OMF file */
    if (ias_parm_read(omf_filename, omf_table, count) != SUCCESS)
    {
        IAS_LOG_ERROR("Error reading OMF parameter file %s", omf_filename);
        return ERROR;
    }

    return SUCCESS;
}

/******************************************************************************
NAME: read_parameters

PURPOSE: This function reads all the parameters from the parameter file.

RETURN VALUE: Type = int
    Value    Description
    -----    -----------
    SUCCESS  The parameters were successfully read
    ERROR    An error occurred reading the parameters

******************************************************************************/
static int read_parameters
(
    const char *parm_filename, /* I: parameter filename that contains resample 
                                    parameters */
    char *l0r_filename,                   /* O: L0R input filename */
    char *cpf_filename,                   /* O: CPF input filename */
    char *model_filename,                 /* O: Model input filename */
    char *grid_filename,                  /* O: Grid input filename */
    int *wrs_path,                        /* O: Target wrs path */
    int *wrs_row                          /* O: Target wrs row */  
)
{
    int count;                         /* loop counter */
    int help_requested;                /* Help request variable */
    int processing_pass = 0;           /* Processing pass requested */
    char local_omf_filename[PATH_MAX]; /* OMF filename */
    char local_wo_id[IAS_WORKORDER_ID_SIZE]; /* Work order ID */
    int status;

    /* start parameter table definition */
    IAS_PARM_DECLARE_TABLE(parms, 2);

    /* Get the work order ID */
    IAS_PARM_WORK_ORDER_ID(parms, local_wo_id, sizeof(local_wo_id), 1);

    /* Get processing pass. */
    IAS_PARM_PROCESSING_PASS(parms, &processing_pass, 
        sizeof(processing_pass), 1);

    /* calculate number of items in the parms list */
    count = IAS_PARM_GET_TABLE_SIZE(parms);

    /* call the ias_parm_provide_help routine to provide help for the 
       parameters if requested */
    help_requested = ias_parm_provide_help(parm_filename, parms, count,
        IAS_INPUT_PARAMETERS);
    if (help_requested)
    {
        if (help_requested == TRUE)
        {
            /* provide help for the OMF.  If it succeeds, it will exit.  If
               it returns, an error occurred, so return an error */
            read_omf_parameters(parm_filename, processing_pass, l0r_filename,
                cpf_filename, model_filename, grid_filename, wrs_path, wrs_row);
            return ERROR;
        }
        else
            return ERROR;
    }

    if (ias_parm_read(parm_filename, parms, count) != SUCCESS)
    {
        IAS_LOG_ERROR("Error reading parameter file %s", parm_filename);
        return ERROR;
    }

    /* Build the OMF filename. */
    status = snprintf(local_omf_filename, sizeof(local_omf_filename), "%s.omf",
        local_wo_id);
    if (status < 0 || status >= sizeof(local_omf_filename))
    {
        IAS_LOG_ERROR("Creating the omf filename from work order id %s",
            local_wo_id);
        return ERROR;
    }

    return read_omf_parameters(local_omf_filename, processing_pass,
        l0r_filename, cpf_filename, model_filename, grid_filename, wrs_path,
        wrs_row);
}


/******************************************************************************
NAME: read_l0r_metadata

PURPOSE: This function reads all the L0R metadata used to create the angle 
         coefficients

RETURN VALUE: Type = int
    Value    Description
    -----    -----------
    SUCCESS  The parameters were successfully read
    ERROR    An error occurred reading the parameters

******************************************************************************/
static int read_l0r_metadata
(
    char *l0r_filename,              /* I: L0R input filename */
    int wrs_path,                    /* I: Target wrs path */
    int wrs_row,                     /* I: Target wrs row */
    IAS_ANGLE_GEN_METADATA *metadata /* O: Metadata structure */
)
{
    L0RIO *l0r_img = NULL;                           /* L0R image */
    IAS_L0R_INTERVAL_METADATA l0r_interval_metadata; /* Interval metadata */
    int scene_count;                                 /* L0R scene count */

    /* Open the L0R */
    l0r_img = ias_l0r_open_from_filename(l0r_filename, 
        IAS_L0R_COMPRESSION_DEFAULT);
    if (!l0r_img)
    {
        IAS_LOG_ERROR("Reading the L0R Image %s", l0r_filename);
        return ERROR;
    }

    /* Open the L0R metadata file */
    if (ias_l0r_open_metadata(l0r_img, IAS_READ) != SUCCESS)
    {
        IAS_LOG_ERROR("Opening L0R metadata");
        ias_l0r_close(l0r_img);
        return ERROR;
    }

    /* Grab the interval metadata */
    if (ias_l0r_get_interval_metadata(l0r_img, &l0r_interval_metadata)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Reading L0R interval metadata");
        ias_l0r_close(l0r_img);
        return ERROR;
    }

    /* Copy the spacecraft id */
    strncpy(metadata->spacecraft_id, l0r_interval_metadata.spacecraft_id,
        sizeof(metadata->spacecraft_id));
    ias_misc_convert_to_uppercase(metadata->spacecraft_id);

    /* Get the count of scenes in the metadata */
    if (ias_l0r_get_scene_metadata_records_count(l0r_img, &scene_count) 
        != SUCCESS)
    {
        IAS_LOG_ERROR("Getting L0R scene metadata size");
        ias_l0r_close(l0r_img);
        return ERROR;
    }

    if (scene_count > 0)
    {
        int scene_index = 0;
        IAS_L0R_SCENE_METADATA *l0r_metadata_scenes = NULL;

        /* Allocate memory for the scene metadata */
        l0r_metadata_scenes = malloc(sizeof(IAS_L0R_SCENE_METADATA)
            * scene_count);
        if (!l0r_metadata_scenes)
        {
            IAS_LOG_ERROR("Allocating memory for L0R scene metadata");
            ias_l0r_close(l0r_img);
            return ERROR;
        }
        memset(l0r_metadata_scenes, 0, sizeof(IAS_L0R_SCENE_METADATA)
                * scene_count);

        /* Read all the scenes in */
        if (ias_l0r_get_scene_metadata(l0r_img, 0, scene_count,
            l0r_metadata_scenes) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading L0R scene metadata");
            free(l0r_metadata_scenes);
            ias_l0r_close(l0r_img);
            return ERROR;
        }

        /* Find the scene metadata for the path/row being processed */
        for (scene_index = 0; scene_index < scene_count; scene_index++)
        {
            if (wrs_path == l0r_metadata_scenes[scene_index].target_wrs_path
                && wrs_row == l0r_metadata_scenes[scene_index].target_wrs_row)
            {
                break;
            }
        }

        if (scene_index == scene_count) /* Make sure the scene was found */
        {
            IAS_LOG_ERROR("L0R scene metadata not found");
            free(l0r_metadata_scenes);
            ias_l0r_close(l0r_img);
            return ERROR;
        }

        /* Copy the landsat scene id */
        strncpy(metadata->landsat_scene_id, 
            l0r_metadata_scenes[scene_index].landsat_scene_id,
            sizeof(metadata->landsat_scene_id));

        free(l0r_metadata_scenes);
        l0r_metadata_scenes = NULL;
    }
    else
    {
        IAS_LOG_ERROR("No scenes found in L0R");
        return ERROR;
    }

    /* Close the L0R and metadata */
    if (ias_l0r_close(l0r_img) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing the L0R");
        return ERROR;
    }
    
    return SUCCESS;
}

/**************************************************************************
NAME: create_angle_coeff_file (main)

PURPOSE: Creates the ANG coefficient file using the required model, grid, 
         and CPF structures. Cycles through each band/SCA to generate L1T 
         to L1R rational polynomials, cycles through the bands to 
         generate satellite view and solar illumination vector rational 
         polynomials, and writes out the angle coefficient file.     

RETURNS: EXIT_SUCCESS or EXIT_FAILURE
***************************************************************************/
int main 
(
    int argc, 
    char *argv[] 
)
{
    char grid_filename[PATH_MAX];   /* Grid filename */
    char model_filename[PATH_MAX];  /* LOS model filename */
    char cpf_filename[PATH_MAX];    /* CPF filename */
    char ang_filename[PATH_MAX];    /* Ouput ANG filename */
    char l0r_filename[PATH_MAX];    /* L0R filename */
    int band_index;                 /* Grid band index */
    IAS_SATELLITE_ID sat_id;        /* Satellite ID */
    IAS_GRID_TYPE grid;             /* Grid structure */
    IAS_LOS_MODEL *model;           /* Model structure */
    IAS_CPF *cpf;                   /* CPF structure */
    const struct IAS_CPF_EARTH_CONSTANTS *cpf_earth; /* Earth constants */
    IAS_ANGLE_GEN_METADATA metadata;/* Metadata structure */
    int wrs_path;                   /* WRS path */
    int wrs_row;                    /* WRS row */

    /* If there are not enough arguments, generate a usage message */
    if (argc < 2)
    {
        printf(USE_MESSAGE);
        return EXIT_FAILURE;
    }

    /* Initialize the logging library */
    if (ias_log_initialize("Create ANG") != SUCCESS)
    {
        IAS_LOG_ERROR("Error initializing logging library");
        return EXIT_FAILURE;
    }

    /* Read in the omf parameters */
    if (read_parameters(argv[1], l0r_filename, cpf_filename, model_filename,
        grid_filename, &wrs_path, &wrs_row) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading the omf parameters in file %s", argv[1]);
        return EXIT_FAILURE;
    }

    /* Validate the model file */
    if (!ias_model_is_model_file(model_filename))
    {
        IAS_LOG_ERROR("Reading model from %s", model_filename);
        return EXIT_FAILURE;
    }

    /* Read the grid header to obtain the satellite number so that we can
       initialize the satellite attributes library. */
    if (ias_grid_header_read(grid_filename, &grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading grid header from %s", grid_filename);
        return EXIT_FAILURE;
    }

    /* Initialize satellite attributes library. */
    sat_id = ias_sat_attr_get_satellite_id_from_satellite_number(grid.satnum);
    if (sat_id == ERROR)
    {
        IAS_LOG_ERROR("Unsupported satellite number: %d", grid.satnum);
        return EXIT_FAILURE;
    }

    if (ias_sat_attr_initialize(sat_id) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attributes library");
        return EXIT_FAILURE;
    }

    /* Read the full grid structure */
    if (ias_grid_read(grid_filename, NULL, 0, &grid) != SUCCESS)
    {
        IAS_LOG_ERROR ("Reading the grid file %s", grid_filename);
        return EXIT_FAILURE;
    }

    /* Read the full model structure */
    model = ias_model_read(model_filename);
    if (!model)
    {
        IAS_LOG_ERROR ("Reading the model file %s", model_filename);
        ias_grid_free (&grid);
        return EXIT_FAILURE;
    }

    /* Read the CPF */
    cpf = ias_cpf_read(cpf_filename);
    if (!cpf)
    {
        IAS_LOG_ERROR("Cannot read CPF file %s", cpf_filename);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }

    /* Get the needed group from the CPF. */
    cpf_earth = ias_cpf_get_earth_const(cpf);
    if (!cpf_earth)
    {
        IAS_LOG_ERROR("Retrieving Earth constants group from CPF file %s",
            cpf_filename);
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }
    
    /* Initialize the leap seconds using the UTC ephemeris epoch time */
    if (ias_math_init_leap_seconds_from_UTC_time(
        model->spacecraft.ephemeris.utc_epoch_time, 
        &cpf_earth->leap_seconds_data) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing leap seconds");
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }

    /* Connect to NOVAS for sun angle computation */
    if (ias_geo_initialize_novas() != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the novas library");
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }

    /* Retrieve the information from the grid and model */
    if (ias_angle_gen_populate_metadata(cpf_earth->semi_major_axis, 
        cpf_earth->semi_minor_axis, &grid, model, &metadata) != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving the data from the model");
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        return EXIT_FAILURE;
    }
    
    /* Disconnect from NOVAS */
    if (ias_geo_shutdown_novas() != SUCCESS)
    { 
        IAS_LOG_ERROR("Shutting down the novas library");
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        ias_angle_gen_free(&metadata);
        return EXIT_FAILURE;
    }

    /* Free the CPF */
    ias_cpf_free(cpf);
    cpf = NULL;
    cpf_earth = NULL;

    /* Read the L0R metadata */
    if (read_l0r_metadata(l0r_filename, wrs_path, wrs_row, &metadata)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Reading the L0R metadata from %s", l0r_filename);
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        ias_angle_gen_free(&metadata);
        return EXIT_FAILURE;
    }
    
    /* Initialize the metadata projection transformation */
    if (ias_angle_gen_initialize_transformation(&metadata) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the projection transformation");
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        ias_angle_gen_free(&metadata);
        exit (EXIT_FAILURE);
    }

    /* Calculate the band coefficients */
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        int band_number;   /* Current Band number */

        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Failed to convert band index %d to band number",
                band_index);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            ias_angle_gen_free(&metadata);
            ias_angle_gen_free_transformation(&metadata);
            return EXIT_FAILURE;
        }

        /* Check if the band is present in the grid */
        if (!grid.bands_present[band_index])
        {
            IAS_LOG_WARNING("Band number %d not present in grid, Skipping", 
                band_number); 
            continue;
        }

        IAS_LOG_INFO("Calculating coefficients for band number %d",
            band_number);

        /* Retrieve the per band data */
        if (ias_angle_gen_get_band_data(band_index, &grid, model, &metadata) 
            != SUCCESS)
        {
            IAS_LOG_ERROR("Retrieving the band data for band index %d", 
                band_index);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            ias_angle_gen_free(&metadata);
            ias_angle_gen_free_transformation(&metadata);
            return EXIT_FAILURE;
        }
        
        /* Calculate the coefficients for the current band */
        if (ias_angle_gen_calculate_band_rpc(grid.gridbands[band_index], 
            band_index, &metadata) != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating the band coefficients for band "
                "number %d", band_number);
            ias_grid_free(&grid);
            ias_los_model_free(model);
            ias_angle_gen_free(&metadata);
            ias_angle_gen_free_transformation(&metadata);
            return EXIT_FAILURE;
        }
    }

    /* Free the model */
    ias_los_model_free(model);
    model = NULL;

    /* Free the grid */
    ias_grid_free(&grid);

    /* Free the transformation */
    ias_angle_gen_free_transformation(&metadata);

    /* Get the output metadata filename. */
    if (ias_angle_gen_create_ang_filename(metadata.landsat_scene_id, 
        sizeof(ang_filename), ang_filename) != SUCCESS)
    {
        IAS_LOG_ERROR("Creating the ANG filename");
        ias_angle_gen_free(&metadata);
        return EXIT_FAILURE;
    }

    IAS_LOG_INFO("Writing ANG metadata to file %s", ang_filename);

    /* Write the metdata */
    if (ias_angle_gen_write_ang(ang_filename, &metadata) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing output metadata file %s", 
            ang_filename);
        ias_angle_gen_free(&metadata);
        return EXIT_FAILURE;
    }

    /* Free the metadata */
    ias_angle_gen_free(&metadata);

    return EXIT_SUCCESS;
}
