/******************************************************************************
NAME: create_merged_l1g

PURPOSE:  This code is highly specific to a specific format/set of data
          from the calval team.  It was created to merge files during the
          TIRS b2b_char implementation.

******************************************************************************/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "ias_l1g.h"
#include "ias_types.h"
#include "ias_const.h"

int copy_l1g_band_image_data
(
    L1GIO *source_image,
    int source_band_nbr,
    L1GIO *dest_image,
    int dest_band_nbr
);

int main(int argc, char *argv[])
{
    int i;
    int status;
    int bands_processed = 0;

    int num_bands_oli;
    int num_bands_tirs;

    L1GIO *input_oli_image;
    L1GIO *input_tirs_image;

    L1GIO *output_image;

    char *input_oli_filename;
    char *input_tirs_filename;

    char *output_filename;

    IAS_L1G_BAND_METADATA bmd;
    IAS_L1G_FILE_METADATA fmd;

    if (argc < 4)
    {
        printf("Usage: create_merged_l1g <input_oli_file> <input_tirs_file> "
                      "<output_file>\n");
        exit(0);
    }
    
    input_oli_filename = argv[1];
    input_tirs_filename = argv[2];
    output_filename = argv[3];

    input_oli_image = ias_l1g_open_image(input_oli_filename, IAS_READ);
    if (input_oli_image == NULL)
    {
        printf("Can't open input file: %s\n", input_oli_filename);
        exit(0);
    }

    input_tirs_image = ias_l1g_open_image(input_tirs_filename, IAS_READ);
    if (input_tirs_image == NULL)
    {
        printf("Can't open input file: %s\n", input_tirs_filename);
        exit(0);
    }

    num_bands_oli = ias_l1g_get_band_count(input_oli_image);
    num_bands_tirs = ias_l1g_get_band_count(input_tirs_image);

    output_image = ias_l1g_open_image(output_filename, IAS_WRITE);
    if (output_image == NULL)
    {
        printf("Can't open output file: %s\n", output_filename);
        exit(0);
    }

    printf("Number of bands file 1: %d\n", num_bands_oli);
    printf("Number of bands file 2: %d\n", num_bands_tirs);

    /* process one band at at time */
    i = 1;
    while (bands_processed < num_bands_oli)
    {
        if (copy_l1g_band_image_data(input_oli_image, i, output_image, i) !=
            SUCCESS)
        {
            printf("Failed to copy band.  Band[%d]\n", i);
            return EXIT_FAILURE;
        }

        bands_processed++;
        i++;
    }

    /* copy the tirs bands and place them into the correct bands */
    if (copy_l1g_band_image_data(input_tirs_image, 1, output_image, 10) !=
        SUCCESS)
    {
        printf("Failed to copy band.  Band[%d]\n", 1);
        return EXIT_FAILURE;
    }

    if (copy_l1g_band_image_data(input_tirs_image, 2, output_image, 11) !=
        SUCCESS)
    {
        printf("Failed to copy band.  Band[%d]\n", 2);
        return EXIT_FAILURE;
    }

    printf("Processing OLI metadata bands.\n");
    /* write band metadata. There may be more metadata than bands. */

    i = 1;
    while (i <= num_bands_oli)
    {
        status = ias_l1g_get_band_metadata(input_oli_image, i, &bmd);
        if (status == ERROR)
        {
            i++;
            continue;
        }

        printf("Processing OLI metadata band: %d\n", i);

        bmd.maximum_pixel_value = 65535.0;
        status = ias_l1g_set_band_metadata(output_image, &bmd, 1);

        printf("status from ias_l1g_set_band_metadata: %d\n", status);

        i++;
    }

    status = ias_l1g_get_band_metadata(input_oli_image, 1, &bmd);
    if (status == ERROR)
    {
        printf("Failed to read TIRS band metadata: %d\n", 1);
        return EXIT_FAILURE;
    }

    printf("Processing TIRS metadata band: %d/%d\n", 1, 10);

    bmd.band_number = 10;
    snprintf(bmd.band_name, IAS_BAND_NAME_SIZE, "TIRS_THERMAL1");
    bmd.maximum_pixel_value = 65535.0;
    status = ias_l1g_set_band_metadata(output_image, &bmd, 1);

    printf("status from ias_l1g_set_band_metadata: %d\n", status);

    status = ias_l1g_get_band_metadata(input_oli_image, 1, &bmd);
    if (status == ERROR)
    {
        printf("Failed to read TIRS band metadata: %d\n", 2);
        return EXIT_FAILURE;
    }

    printf("Processing TIRS metadata band: %d/%d\n", 1, 11);

    bmd.band_number = 11;
    snprintf(bmd.band_name, IAS_BAND_NAME_SIZE, "TIRS_THERMAL2");
    bmd.maximum_pixel_value = 65535.0;
    status = ias_l1g_set_band_metadata(output_image, &bmd, 1);

    printf("status from ias_l1g_set_band_metadata: %d\n", status);

    status = ias_l1g_get_file_metadata(input_oli_image, &fmd);
    if(status == SUCCESS)
    {
        ias_l1g_set_file_metadata(output_image, &fmd);
    }

    return EXIT_SUCCESS;
}

/*****************************************************************************
    copy_l1g_band_image_data

    Copy the image data from a band and place it into another image.
    The ability to change the band number is available.

    Note: The data type is also converted to unsigned 16bit, and the input
          is assumed to be signed 16bit.  Nothing special is done for the
          conversion.
*****************************************************************************/
int copy_l1g_band_image_data
(
    L1GIO *source_image,
    int source_band_nbr,
    L1GIO *dest_image,
    int dest_band_nbr
)
{
    int status;

    int sca_count;
    int line_count;
    int sample_count;

    short int *band_data;

    IAS_DATA_TYPE data_type;

    L1G_BAND_IO *source_band;
    L1G_BAND_IO *dest_band;

    printf("Looking for band: %d\n", source_band_nbr);

    source_band = ias_l1g_open_band(source_image, source_band_nbr,
                                   &data_type, &sca_count,
                                   &line_count, &sample_count);

    if (source_band == NULL)
    {
        printf("Failed to find source band: %d\n", source_band_nbr);
        return ERROR;
    }

    printf("Processing Source band: %d\n", source_band_nbr);
    printf("    Data Type:       %d\n", data_type);
    printf("    SCA Count:       %d\n", sca_count);
    printf("    Line Count:      %d\n", line_count);
    printf("    Sample Count:    %d\n", sample_count);

    /* set the output data type to unsigned 16bit */
//    data_type = IAS_UI2;

    band_data = (short int *)malloc(
        sca_count * line_count * sample_count * sizeof(short int));
    if (band_data == NULL)
    {
        printf("Failed to allocate memory for band data.  Band[%d]\n",
               source_band_nbr);
        return ERROR;
    }

    status = ias_l1g_read_image(source_band, 0, 0, 0,
                                line_count, sample_count, band_data);
    if (status != SUCCESS)
    {
        printf("Failed to read source image data.  Band[%d]\n",
               source_band_nbr);
        return ERROR;
    }

    status = ias_l1g_close_band(source_band);
    if (status != SUCCESS)
    {
        printf("Failed to close source band.  Band[%d]\n",
               source_band_nbr);
        return EXIT_FAILURE;
    }

    printf("Processing Destination band: %d\n", dest_band_nbr);
    printf("    Data Type:       %d\n", data_type);
    printf("    SCA Count:       %d\n", sca_count);
    printf("    Line Count:      %d\n", line_count);
    printf("    Sample Count:    %d\n", sample_count);

    dest_band = ias_l1g_open_band(dest_image, dest_band_nbr, &data_type,
                                  &sca_count, &line_count, &sample_count);

    status = ias_l1g_write_image(dest_band, 0, 0, 0,
                                 line_count, sample_count, band_data);
    if (status != SUCCESS)
    {
        printf("Failed to write destination image data.  Band[%d]\n",
               dest_band_nbr);
        return ERROR;
    }

    status = ias_l1g_close_band(dest_band);
    if (status != SUCCESS)
    {
        printf("Failed to close destination band.  Band[%d]\n",
               dest_band_nbr);
        return ERROR;
    }

    free(band_data);

    return SUCCESS;
}

