#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "ias_logging.h"
#include "ias_const.h"
#include "ias_ancillary_io.h"

/******************************************************************************
NAME:           dump_adp

PURPOSE:
Dumps the ancillary data preprocessing file.

USAGE:  dump_adp <Preprocessed Ancillary filename> <Out filename>

RETURN VALUE:   Type = int
Value    Description
-----    -----------
EXIT_SUCCESS  Successful completion
EXIT_FAILURE  Operation failed

NOTES:

ALGORITHM REFERENCES:
******************************************************************************/

int main(int argc, char *argv[])
{
    const char *pre_file;
    const char *root_name;
    IAS_ANC_ATTITUDE_DATA *attitude_data;
    IAS_ANC_EPHEMERIS_DATA *ephemeris_data;
    char outfile[PATH_MAX];
    FILE *ofile;
    int field;
    int status;

    /* Initialize the logging library. */
    if (ias_log_initialize("dump_adp") != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing logging library");
        exit(EXIT_FAILURE);
    }

    /* Check the arguments and copy input/output file names */
    if ( argc != 3 )
    {
        printf("Usage: dump_adp <Preprocessed Ancillary File> <Out file>\n");
        exit (EXIT_FAILURE);
    }

    /* Assign the input name pointers */
    pre_file = argv[1];
    root_name = argv[2];

    /* Prepare to write the eci data to output file */
    status = snprintf(outfile, sizeof(outfile), "%s.eci.txt", root_name);
    if (status < 0 || status >= sizeof(outfile))
    {
        IAS_LOG_ERROR("Error constructing the output ECI file name ");
        exit(EXIT_FAILURE);
    }

    /* Retrieve the ancillary data from the input file to structure */
    if( ias_ancillary_read(pre_file, &attitude_data, &ephemeris_data ) != 
                SUCCESS )
    {
        IAS_LOG_ERROR("Error reading ancillary pre-processing file %s.",
            pre_file);
        exit(EXIT_FAILURE);
    }

    ofile = fopen(outfile,"w");
    if( ofile == NULL )
    {
        IAS_LOG_ERROR("Unable to open output file %s.",outfile);
        free(ephemeris_data);
        free(attitude_data);
        exit(EXIT_FAILURE);
    }

    /* Write the header for the ECI file */
    if (fprintf(ofile,"Entry_Number,Epoch_UTC_Year,Epoch_UTC_Day,"
        "Time_SOD,ECI_X,ECI_Y,ECI_Z,ECI_Vx,ECI_Vy,ECI_Vz\n") < 0)
    {
        IAS_LOG_ERROR("Unable to write to output file %s.",outfile);
        free(ephemeris_data);
        free(attitude_data);
        fclose(ofile);
        exit(EXIT_FAILURE);
    }
    
    /* Loop through the ECI data and write to the output file */
    for(field = 0; field < ephemeris_data->number_of_samples; field++)
    {
        /* Seconds of day are in microseconds */
        status = fprintf(ofile,"%d,%4.0f,%3.0f,%.6f,"
            "%e,%e,%e,"
            "%e,%e,%e\n", 
            field,
            ephemeris_data->utc_epoch_time[0],
            ephemeris_data->utc_epoch_time[1],
            ephemeris_data->records[field].seconds_from_epoch +
                ephemeris_data->utc_epoch_time[2],

            ephemeris_data->records[field].eci_position[0],
            ephemeris_data->records[field].eci_position[1],
            ephemeris_data->records[field].eci_position[2],

            ephemeris_data->records[field].eci_velocity[0],
            ephemeris_data->records[field].eci_velocity[1],
            ephemeris_data->records[field].eci_velocity[2]);

        if (status < 0)
        {
            IAS_LOG_ERROR("Unable to write to output file %s.",outfile);
            free(ephemeris_data);
            free(attitude_data);
            fclose(ofile);
            exit(EXIT_FAILURE);
        }

    }

    if (fclose(ofile) != 0)
    {
        IAS_LOG_ERROR("Unable to close output file %s.",outfile);
        free(ephemeris_data);
        free(attitude_data);
        exit(EXIT_FAILURE);
    }

    /* Prepare to write the ecef data to output file */
    status = snprintf(outfile, sizeof(outfile), "%s.ecef.txt", root_name);

    if (status < 0 || status >= sizeof(outfile))
    {
        IAS_LOG_ERROR("Error constructing output ECEF file name ");
        free(ephemeris_data);
        free(attitude_data);
        exit(EXIT_FAILURE);
    }

    ofile = fopen(outfile,"w");
    if( ofile == NULL )
    {
        IAS_LOG_ERROR ("Unable to open output file %s.",outfile);
        free(ephemeris_data);
        free(attitude_data);
        exit(EXIT_FAILURE);
    }

    /* Write the header for the ECEF file */
    if (fprintf(ofile,"Entry_Number,Epoch_UTC_Year,Epoch_UTC_Day,Time_SOD,"
        "ECEF_X,ECEF_Y,ECEF_Z,ECEF_Vx,ECEF_Vy,ECEF_Vz\n") < 0)
    {
        IAS_LOG_ERROR ("Unable to write to the output file %s.",outfile);
        free(ephemeris_data);
        free(attitude_data);
        exit(EXIT_FAILURE);
    }

    /* Loop through the ECEF data and write to the output file */
    for(field = 0; field < ephemeris_data->number_of_samples; field++ )
    {
        /* Seconds of day are in microseconds */
        status = fprintf(ofile,"%d,%4.0f,%3.0f,%.6f,"
            "%e,%e,%e,"
            "%e,%e,%e\n",
            field,
            ephemeris_data->utc_epoch_time[0],
            ephemeris_data->utc_epoch_time[1],
            ephemeris_data->records[field].seconds_from_epoch +
                ephemeris_data->utc_epoch_time[2],

            ephemeris_data->records[field].ecef_position[0],
            ephemeris_data->records[field].ecef_position[1],
            ephemeris_data->records[field].ecef_position[2],

            ephemeris_data->records[field].ecef_velocity[0],
            ephemeris_data->records[field].ecef_velocity[1],
            ephemeris_data->records[field].ecef_velocity[2]);

        if (status < 0)
        {
            IAS_LOG_ERROR ("Unable to write to the output file %s.",outfile);
            free(ephemeris_data);
            free(attitude_data);
            exit (EXIT_FAILURE);
        }
    }

    if (fclose(ofile) != 0)
    {
        IAS_LOG_ERROR("Unable to close output file %s.",outfile);
        free(ephemeris_data);
        free(attitude_data);
        exit(EXIT_FAILURE);
    }

    /* Clean up resources that are no longer needed */
    free(ephemeris_data);
    ephemeris_data = NULL;

    /* Prepare to write the attitude data to output file */
    status = snprintf(outfile, sizeof(outfile), "%s.att.txt", root_name);
    if (status < 0 || status >= sizeof(outfile))
    {
        IAS_LOG_ERROR("Error constructing output attitude file name ");
        free(attitude_data);
        exit (EXIT_FAILURE);
    }

    /* Open the output file for the attitude data */
    ofile = fopen(outfile,"w");
    if( ofile == NULL )
    {
        IAS_LOG_ERROR ("Unable to open output file %s.",outfile);
        free(attitude_data);
        exit (EXIT_FAILURE);
    }

    /* Write the header to the attitude output file */
    status = fprintf(ofile,"Entry_Number,Epoch_UTC_Year,Epoch_UTC_Day,Time_SOD,"
        "ECI_q1,ECI_q2,ECI_q3,ECI_q4,"
        "ECEF_q1,ECEF_q2,ECEF_q3,ECEF_q4,"
        "Roll,Pitch,Yaw,"
        "Roll_Rate,Pitch_Rate,Yaw_Rate\n");
    if (status < 0)
    {
        IAS_LOG_ERROR ("Unable to write to the output file %s.",outfile);
        free(attitude_data);
        exit (EXIT_FAILURE);
    }

    /* Loop through and write the data for the attitude data */
    for(field = 0; field < attitude_data->number_of_samples; field++)
    {
        /* Seconds of day are in microseconds */
        status = fprintf(ofile,"%d,%4.0f,%3.0f,%.6f,"
            "%e,%e,%e,%e,"
            "%e,%e,%e,%e,"
            "%e,%e,%e,"
            "%e,%e,%e\n",
            field,
            attitude_data->utc_epoch_time[0],
            attitude_data->utc_epoch_time[1],
            attitude_data->records[field].seconds_from_epoch +
                attitude_data->utc_epoch_time[2],

            attitude_data->records[field].eci_quaternion[0],
            attitude_data->records[field].eci_quaternion[1],
            attitude_data->records[field].eci_quaternion[2],
            attitude_data->records[field].eci_quaternion[3],

            attitude_data->records[field].ecef_quaternion[0],
            attitude_data->records[field].ecef_quaternion[1],
            attitude_data->records[field].ecef_quaternion[2],
            attitude_data->records[field].ecef_quaternion[3],

            attitude_data->records[field].roll,
            attitude_data->records[field].pitch,
            attitude_data->records[field].yaw,

            attitude_data->records[field].roll_rate,
            attitude_data->records[field].pitch_rate,
            attitude_data->records[field].yaw_rate);

        if (status < 0)
        {
            IAS_LOG_ERROR ("Unable to write to the output file %s.",outfile);
            free(attitude_data);
            exit(EXIT_FAILURE);
        }
    }

    free(attitude_data);
    attitude_data = NULL;

    if (fclose(ofile) != 0)
    {
        IAS_LOG_ERROR("Unable to close output file %s.",outfile);
        exit(EXIT_FAILURE);
    }

    exit(EXIT_SUCCESS);
}
