#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include "gctp.h"                 /* projection constants          */
#include "ias_grid_io.h"          /* Grid I/O library              */
#include "ias_logging.h"          /* IAS logging library           */
#include "ias_types.h"            /* IAS_FRAME_TYPE, IAS_CORRECTION_TYPE, 
                                     IAS_DATA_TYPE definitions */

#define BAND_NUMBER_STRING_LENGTH 2
#define SCA_NUMBER_STRING_LENGTH 2

/* This is the length of the basic switch: - and a letter.  It doesn't 
   include qualifiers like band or SCA number. */ 
#define SWITCH_PREFIX_LENGTH 2 

/******************************************************************************
NAME:           dump_grid

PURPOSE:
To read in the specified HDF grid file and write the grid to standard 
output.

USAGE:  dump_grid <grid HDF filename> 

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
FAILURE  Operation failed

NOTES:

ALGORITHM REFERENCES:
See ias_grid.h for a description of the grid structure.

******************************************************************************/

int main(int argc, char *argv[])
{
    IAS_GRID_TYPE grid;        /* Grid structure                           */
    char grid_filename[PATH_MAX]; /* Grid filename                         */
    int i, j, k, n;            /* Loop variables                           */
    int band_index;            /* Band index                               */
    int sca_index;             /* SCA index                                */
    int zplane;                /* Z plane number                           */
    int zoffset;               /* Offset into lookups due to elevation     */
    long count;                /* Count iterations through 2 nested loops  */
    IAS_GRID_BAND_TYPE ref;    /* Reference structure of current band      */
    int ncoeff;                /* Number of coefficients                   */
    char switch_prefix[SWITCH_PREFIX_LENGTH]; /* Parameter switch prefix   */
    int dump_coeff_flag;       /* Flag for dumping mapping coefficients    */
    int dump_header_flag;      /* Flag for only dumping the header         */
    int band_selected_flag;    /* Flag for dumping only 1 selected band    */
    int sca_selected_flag;     /* Flag for dumping only 1 selected SCA     */
    int selected_band_number = 0; /* Number of user-selected band          */
    char band_number_string[BAND_NUMBER_STRING_LENGTH];/* String version of 
                                  band number                              */
    int selected_band_index = 0;/* Index of user-selected band             */
    int selected_sca_number;   /* Number of user-selected SCA              */
    char sca_number_string[SCA_NUMBER_STRING_LENGTH]; /* String version of 
                                  SCA                                      */
    int selected_sca_index = 0;/* Index of user-selected SCA               */
    int option_length;         /* Length of the command line option string */
    IAS_SATELLITE_ID sat_id;   /* Satellite ID derived from satellite num  */

    /* Check arguments and get the grid filename from the arguments.       */
    if (argc < 2 || argc > 5)
    {
        printf ("Usage: dump_grid [ switches ] <grid HDF-file>\n");
        printf ("where [ switches ] can consist of:\n");
        printf ("\t-b# : dump only specified band\n");
        printf ("\t-s# : dump only specified SCA\n");
        printf ("\t-h  : dump only header information (small)\n");
        printf ("\t-m  : dump mapping coefficients (large)\n");
        printf ("\t    : No switch will not print mapping coefficients\n");
        printf ("\t      -m and -h cannot both be used\n");
        exit (EXIT_FAILURE);
    }

    /* Initialize parameter to their defaults. */
    dump_coeff_flag = FALSE;
    dump_header_flag = FALSE;
    band_selected_flag = FALSE;
    sca_selected_flag = FALSE;

    /* Skip the program name in argv; also skip the grid filename at the end
       of argv. */
    for (i = 1; i < argc - 1; i++)
    {
        /* Extract the first SWITCH_PREFIX_LENGTH characters from the 
           selection. */
        strncpy (switch_prefix, argv[i], SWITCH_PREFIX_LENGTH);

        /* Find the length of the command line option. */
        option_length = strlen (argv[i]);

        if (strncmp (switch_prefix, "-h", SWITCH_PREFIX_LENGTH) == 0)
        {
            dump_header_flag = TRUE;
            if (dump_coeff_flag)
            {
                IAS_LOG_ERROR ("-h and -m flags cannot both be used.");
                exit (EXIT_FAILURE);
            }
        }
        else if (strncmp (switch_prefix, "-m", SWITCH_PREFIX_LENGTH) == 0)
        {
            dump_coeff_flag = TRUE;
            if (dump_header_flag)
            {
                IAS_LOG_ERROR ("-h and -m flags cannot both be used.");
                exit (EXIT_FAILURE);
            }
        }
        else if (strncmp (switch_prefix, "-b", SWITCH_PREFIX_LENGTH) == 0)
        {
            band_selected_flag = TRUE;

            /* Make sure there is space for an actual band choice. */
            if (option_length <= BAND_NUMBER_STRING_LENGTH)
            {
                IAS_LOG_ERROR ("No specific band was selected.");
                exit (EXIT_FAILURE);
            }

            /* Extract the band number from the command line option. */
            strcpy (band_number_string, &argv[i][SWITCH_PREFIX_LENGTH]); 
            selected_band_number = atoi (band_number_string);

            /* Validate the band number. */
            if (selected_band_number < 1 || 
                selected_band_number > IAS_MAX_NBANDS)
            {
                IAS_LOG_ERROR ("The selected band is invalid.");
                exit (EXIT_FAILURE);
            }
        }
        else if (strncmp (switch_prefix, "-s", SWITCH_PREFIX_LENGTH) == 0)
        {
            sca_selected_flag = TRUE;

            /* Make sure there is space for an actual SCA choice. */
            if (option_length <= SCA_NUMBER_STRING_LENGTH)
            {
                IAS_LOG_ERROR ("No specific SCA was selected.");
                exit (EXIT_FAILURE);
            }

            /* Extract the SCA number from the command line option. */
            strcpy (sca_number_string, &argv[i][SWITCH_PREFIX_LENGTH]); 
            selected_sca_number = atoi (sca_number_string);

            /* Validate the SCA number. */
            if (selected_sca_number < 1 || selected_sca_number > IAS_MAX_NSCAS)
            {
                IAS_LOG_ERROR ("The selected SCA is invalid.");
                exit (EXIT_FAILURE);
            }

            /* Convert the SCA number to the SCA index. */
            selected_sca_index = selected_sca_number - 1;
        }
        else 
        {
            IAS_LOG_ERROR ("Error in input parameters.");
            exit (EXIT_FAILURE);
        }
    }

    /* Get the grid filename from the end of command line options. */
    strcpy (grid_filename, argv[i]);
    printf ("Grid filename is %s.\n", grid_filename);

    /* Read the grid header to obtain the satellite number so that we can
       initialize the satellite attributes library. */
    if (ias_grid_header_read(grid_filename, &grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading grid header from %s", grid_filename);
        exit(EXIT_FAILURE);
    }

    /* Initialize satellite attributes library. */
    sat_id = ias_sat_attr_get_satellite_id_from_satellite_number(grid.satnum);
    if (sat_id == ERROR)
    {
        IAS_LOG_ERROR("Unsupported satellite number: %d", grid.satnum);
        exit(EXIT_FAILURE);
    }
    if (ias_sat_attr_initialize(sat_id) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attributes library");
        exit(EXIT_FAILURE);
    }

    /* If user specified a band, convert the band number into a band index */
    if (band_selected_flag)
    {
        selected_band_index =
            ias_sat_attr_convert_band_number_to_index(selected_band_number);
        if (selected_band_index == ERROR)
        {
            IAS_LOG_ERROR("Invalid band number: %d", selected_band_number);
            exit(EXIT_FAILURE);
        }
    }

    /* Read the full grid structure */
    if (ias_grid_read (grid_filename, NULL, 0, &grid) != SUCCESS)
    {
        IAS_LOG_ERROR ("Error reading the grid file %s.", grid_filename);
        exit (EXIT_FAILURE);
    }

    /* Print out the non-band specific information. */
    printf ("General Grid Information\n");
    printf ("------------------------\n");

    printf ("Grid name: %s\n", grid.gridname);

    printf ("Maximum Number of SCAs: %d\n", grid.maximum_nscas);

    printf ("Number of bands: %d\n", grid.nbands);
    printf ("Bands:\n");
    for (band_index = 0; band_index < grid.nbands; band_index++)
    {
        if (grid.bands_available[band_index])
        {
            printf (" %d", ias_sat_attr_convert_band_index_to_number(
                band_index));
        }
        printf ("\n");
    }

    printf ("Data type: ");
    switch (grid.dtype)
    {
        case IAS_BYTE: printf ("BYTE"); break;
        case IAS_I2: printf ("I2"); break;
        case IAS_UI2: printf ("UI2"); break;
        case IAS_I4: printf ("I4"); break;
        case IAS_UI4: printf ("UI4"); break;
        case IAS_R4: printf ("R4"); break;
        case IAS_R8: printf ("R8"); break;
        case IAS_NDTYPES: printf ("Number of different types"); break;
        default: printf ("Unknown/invalid");
    }
    printf ("\n");

    printf ("Correction type: ");
    switch (grid.correction_type)
    {
        case IAS_SYSTEMATIC: printf ("Systematic"); break;
        case IAS_PRECISION: printf ("Precision corrected"); break;
        case IAS_TERRAIN: printf ("Terrain corrected"); break;
        default: printf ("Unknown/invalid");
    }
    printf ("\n");

    printf ("Satellite number: %d\n", grid.satnum);

    printf ("Framing method: ");
    switch (grid.frame_type)
    {
        case IAS_GEOBOX:
            printf ("GEOBOX (UL and LR lat/long coordinates)"); break;
        case IAS_PROJBOX:
            printf ("PROJBOX (LR projection coords + line/samp)"); break;
        case IAS_UL_SIZE:
            printf ("UL+SIZE (UL projection coords + output size)"); break;
        case IAS_MINBOX:
            printf ("MINBOX (Minbox framing)"); break;
        case IAS_MAXBOX:
            printf ("MAXBOX (Maxbox framing)"); break;
        case IAS_PATH_ORIENTED:
            printf ("PATH_ORIENTED (Path-oriented framing)"); break;
        case IAS_PATH_MINBOX:
            printf ("PATH_MINBOX (Path-oriented Minbox framing)"); break;
        case IAS_PATH_MAXBOX:
            printf ("PATH_MAXBOX (Path-oriented Maxbox framing)"); break;
        default:
            printf ("Unknown/invalid");
    }
    printf ("\n\n");

    printf ("Projection information\n");
    printf ("----------------------\n");
    printf ("Projection units: %s\n", grid.units);

    printf ("Projection code: ");
    switch (grid.code)
    {
        case UTM:
            printf ("UTM"); break;
        case LAMCC:
            printf ("LAMCC"); break;
        case PS:
            printf ("POLAR STEREOGRAPHIC"); break;
        case POLYC:
            printf ("POLYCONIC"); break;
        case TM:
            printf ("TRANSVERSE MERCATOR"); break;
        case HOM:
            printf ("OBLIQUE MERCATOR"); break;
        case SOM:
            printf ("SOM"); break;
        default:
            printf ("Unknown/invalid"); break;
    }
    printf ("\n");

    printf ("Datum code: %s\n", grid.datum);
    printf ("Spheroid code: %d\n", grid.spheroid);
    printf ("Zone code: %d\n", grid.zone);

    printf ("Projection parms:\n");
    for (i = 0; i < IAS_PROJ_PARAM_SIZE; i += 5)
    {
        printf ("  %f  %f  %f  %f  %f\n", 
            grid.projprms[i],
            grid.projprms[i+1], 
            grid.projprms[i+2], 
            grid.projprms[i+3],
            grid.projprms[i+4]);
    }

    printf ("Projection corners (lat or y, long or x):\n");
    printf (
        "  UL = (%f, %f)\n", grid.corners.upleft.y, grid.corners.upleft.x);
    printf (
        "  UR = (%f, %f)\n", grid.corners.upright.y, grid.corners.upright.x);
    printf (
        "  LL = (%f, %f)\n", grid.corners.loleft.y, grid.corners.loleft.x);
    printf (
        "  LR = (%f, %f)\n", grid.corners.loright.y, grid.corners.loright.x);
    printf ("\n");

    /* Print the active area corners. */
    for (band_index = 0; band_index < grid.nbands; band_index++)
    {
        /* Check if the user selected this band. */
        if (band_selected_flag && band_index != selected_band_index)
            continue;

        /* Check if the band is available in the grid. */
        if (!grid.bands_available[band_index])
            continue;

        printf ("Band %d active image area corners (lat, long):\n", 
            ias_sat_attr_convert_band_index_to_number(band_index));
        printf ("  UL = (%f, %f)\n", grid.active_corners[band_index].upleft.y, 
            grid.active_corners[band_index].upleft.x);
        printf ("  UR = (%f, %f)\n", grid.active_corners[band_index].upright.y,
            grid.active_corners[band_index].upright.x);
        printf ("  LL = (%f, %f)\n", grid.active_corners[band_index].loleft.y, 
            grid.active_corners[band_index].loleft.x);
        printf ("  LR = (%f, %f)\n", grid.active_corners[band_index].loright.y, 
            grid.active_corners[band_index].loright.x);
        printf ("\n");
    }

    printf ("Path-Oriented Information\n");
    printf ("-------------------------\n");
    printf ("WRS scene center lat/long: %f, %f\n", 
        grid.center_pt.lat, grid.center_pt.lng);
    printf ("Center point (projection): %f, %f\n",
        grid.proj_center_pt.x, grid.proj_center_pt.y);
    printf ("Rotation angle: %f\n", grid.rotate_angle);
    printf ("Orientation angle: %f\n", grid.orient_angle);
    printf ("\n\n\n");

    /* Print out the band-specific information. */
    for (band_index = 0; band_index < grid.nbands; band_index++)
    {
        /* Check if the user selected this band. */
        if (band_selected_flag && band_index != selected_band_index)
            continue;

        /* Check if the band is available in the grid. */
        if (!grid.bands_available[band_index])
            continue;

        printf ("*** Grid info for band ");
        printf ("%d", ias_sat_attr_convert_band_index_to_number (band_index));
        printf (" ***\n");

        printf("SCAs: %d\n", grid.scas_per_band[band_index]);

        for (sca_index = 0; sca_index < grid.scas_per_band[band_index]; 
             sca_index++)
        {
            /* Check if the user selected this SCA. */
            if (sca_selected_flag && sca_index != selected_sca_index)
                continue;

            /* Get the reference gridbands structure */
            ref = grid.gridbands[band_index][sca_index];  

            ncoeff = (ref.degree+1) * (ref.degree+1);

            printf ("\n");
  
            printf("SCA %d\n", sca_index + 1);

            printf (" Zspacing %f\n", ref.zspacing);
            printf (" Number of elevation planes present %d\n", ref.nzplanes);
            printf (" Location of zero elevation %d\n", ref.zeroplane);
            printf ("Grid cell size: %d lines x %d samples\n",
                ref.grid_cell_size.line, ref.grid_cell_size.samp);
            printf ("Pixel size: %f\n", ref.pixsize);
            printf ("Number of output lines: %d\n", ref.lines);
            printf ("Number of output samples: %d\n", ref.samps);
            printf ("Number of grid lines: %d\n", ref.ngrid_lines);
            printf ("Number of grid samples: %d\n", ref.ngrid_samps);
            printf ("\n");

            /* If only dumping the header values, skip the detailed dumps. */
            if (dump_header_flag)
            {
                continue;
            }
  
            for (zplane = 0; zplane < ref.nzplanes; zplane++)
            {
                printf ("Z plane %d\n",zplane);
                printf ("Grid Cell Points (line, sample)\n");
                printf ("Input space ---> Output space "
                         "(deltas line/sample odd/even)\n");
                printf ("-------------------------------\n");
                zoffset = zplane * ref.ngrid_lines * ref.ngrid_samps;
                for (j = 0, count = 0; j < ref.ngrid_lines; j++)
                {
                    for (k = 0; k < ref.ngrid_samps; k++, count++)
                    {
                        printf ("%4d %4d  (%4d, %4d) ---> (%4.4f, %4.4f)"
                            " (%4.4f %4.4f %4.4f %4.4f)\n",
                            j, k, ref.in_lines[j],
                            ref.in_samps[k], 
                            ref.out_lines[count+zoffset], 
                            ref.out_samps[count+zoffset], 
                            ref.delta_line_oe[2*count], 
                            ref.delta_line_oe[2*count+1],
                            ref.delta_samp_oe[2*count], 
                            ref.delta_samp_oe[2*count+1]);
                    }
                }

                printf ("\n");
    
                printf ("Degree of rough transformation polynomial: %d\n\n",
                    ref.degree);
                printf ("Reverse rough transformation coefficients (lines)\n");
                printf ("-------------------------------------------------\n");
                    
                for (j = 0; j < (ref.degree+1) * (ref.degree+1); j++)
                {
                    printf ("%f   ", ref.poly_lines[zplane*ncoeff+j]);
                }
                printf ("\n\n");
    
                printf (
                    "Reverse rough transformation coefficients (samples)\n");
                printf (
                    "---------------------------------------------------\n");
                for (j = 0; j < (ref.degree+1) * (ref.degree+1); j++)
                {
                    printf ("%f   ", ref.poly_samps[zplane*ncoeff+j]);
                }
                printf ("\n\n\n");

                if (dump_coeff_flag)
                {
                    printf (
                        "Grid cell mapping coefficients forward,reverse\n");
                    printf (
                       "---------------------------------------------------\n");
                    zoffset = zplane * (ref.ngrid_lines-1) * 
                        (ref.ngrid_samps-1);
                    for (j = 0, count = 0; j < (ref.ngrid_lines-1); j++)
                    { 
                        for (k = 0; k < (ref.ngrid_samps-1); k++, count++)
                        {
                            printf ("%5ld %4d %4d  Line ---> ", 
                                count + zoffset, j, k);
                            printf ("(");
                            for (n = 0; n < COEFS_SIZE; n++) 
                            {
                                printf ("%4.2e ",
                                    ref.sattoproj[count+zoffset].b[n]);
                            }
                            printf (") ");
                            printf ("(");

                            for (n = 0; n < COEFS_SIZE; n++)
                            {
                                printf ("%4.2e ", 
                                    ref.projtosat[count+zoffset].b[n]);
                            }
                            printf (")\n");
                            printf ("%5ld %4d %4d  Sample ---> ",
                                count+zoffset,j,k);
                            printf ("(");
                            for (n = 0; n < COEFS_SIZE; n++)
                            {
                                printf ("%4.2e ",
                                    ref.sattoproj[count+zoffset].a[n]);
                            }
                            printf (") ");
                            printf ("(");
                            for (n = 0; n < COEFS_SIZE; n++)
                            {
                                printf ("%4.2e ", 
                                    ref.projtosat[count+zoffset].a[n]);
                            }
                            printf (")\n");
                        }
                        printf ("\n");
                    } /* Loop on ref.ngrid_lines - 1*/
                } /* If (dump_coeff_flag */

                printf ("Line and sample sensitivity (x, y, z)\n");
                printf ("-------------------------------------\n");
                zoffset = zplane * ref.ngrid_lines * ref.ngrid_samps;
                for (j = 0, count = 0; j < ref.ngrid_lines; j++)
                { 
                    for (k = 0; k < ref.ngrid_samps; k++, count++)
                    {
                        printf ("%5ld %4d %4d  Line ---> ", 
                            count + zoffset, j, k);
                        printf ("(");
                        printf ("%4.2e ",
                            ref.line_sensitivity[count+zoffset].x);
                        printf ("%4.2e ",
                            ref.line_sensitivity[count+zoffset].y);
                        printf ("%4.2e ",
                            ref.line_sensitivity[count+zoffset].z);
                        printf (")\n");

                        printf ("%5ld %4d %4d  Sample ---> ",
                            count + zoffset, j, k);
                        printf ("(");
                        printf ("%4.2e ",
                            ref.samp_sensitivity[count+zoffset].x);
                        printf ("%4.2e ",
                            ref.samp_sensitivity[count+zoffset].y);
                        printf ("%4.2e ",
                            ref.samp_sensitivity[count+zoffset].z);
                        printf (")\n");
                    } /* Loop on ref.ngrid_samps */
                    printf ("\n");
                } /* Loop on ref.ngrid_lines */

                printf ("\n");
            } /* Loop on zplanes */
        } /* Loop on sca_index */
    } /* Loop on bands */

    /* Free the grid */
    if (ias_grid_free (&grid) != SUCCESS)
    {
        IAS_LOG_ERROR ("Error freeing the grid.");
        exit (EXIT_FAILURE);
    }

    exit (SUCCESS);
}
