#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "ias_logging.h"
#include "ias_const.h"
#include "ias_l1g.h"
#include "dump_l1g_meta.h"

/******************************************************************************
NAME:           dump_l1g_meta

PURPOSE:
Dumps the metadata from a LDCM formatted L1G image file.

USAGE:  dump_l1g_meta <HDF filename>  <output root name>

RETURN VALUE:   Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
FAILURE  Operation failed

******************************************************************************/
int main(int argc, char *argv[])
{
    const char *l1g_name;       /* L1g HDF5 file name */
    const char *root_name;      /* Root name of the output file */
    char outfile[PATH_MAX];
    IAS_L1G_BAND_METADATA bmd;  /* L1g band meta data */
    IAS_L1G_FILE_METADATA fmd;  /* L1g file meta data */
    L1GIO *img;                 /* L1G image pointer */
    FILE *ofile;                /* Output file pointer */
    int *band_number_list;      /* array of band numbers found in metadata */
    int band_count=0;    	/* number of band numbers found in metadata */
    int band_count2=0;          /* same as above but for a different call  */
    int band_index;             /* Band counter */
    int band_number;            /* Band number (user band) */
    int status;
    
    /* Check the arguments and input file name */
    if (argc != 3)
    {
        printf("Usage: dump_l1g_meta <input-file> <output-root>\n");
        exit(EXIT_FAILURE);
    }

    l1g_name = argv[1];
    root_name = argv[2];


    /* Open L1g file */
    img = ias_l1g_open_image(l1g_name, IAS_READ);
    if (img == NULL)
    {
        IAS_LOG_ERROR("Opening file %s\n", l1g_name);
        exit(EXIT_FAILURE);
    }

    /* Open the output file */
    status = snprintf(outfile, sizeof(outfile), "%s.meta.txt", root_name);
    if (status < 0 || status >= sizeof(outfile))
    {
        IAS_LOG_ERROR("Constructing output metadata file name ");
        ias_l1g_close_image(img);
        exit(EXIT_FAILURE);
    }

    ofile = fopen(outfile, "w");
    if (ofile == NULL)
    {
        IAS_LOG_ERROR("Unable to open output file %s.", outfile);
        ias_l1g_close_image(img);
        exit(EXIT_FAILURE);
    }

    /* read the file metadata */
    if (ias_l1g_get_file_metadata(img, &fmd) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading file metadata");
        fclose(ofile);
        ias_l1g_close_image(img);
        exit(EXIT_FAILURE);
    }

    if (print_file_meta(&fmd, ofile) != SUCCESS) 
    {
        IAS_LOG_ERROR("Printing file metadata");
        fclose(ofile);
        ias_l1g_close_image(img);
        exit(EXIT_FAILURE);
    }

    band_count = ias_l1g_get_band_count(img);

    if (band_count == ERROR)
    {
        ias_l1g_close_image(img);
        fclose(ofile);
        IAS_LOG_ERROR("Reading band count");
        exit(EXIT_FAILURE);
    }

    band_number_list = (int*)malloc( band_count*sizeof(int) );
    if( band_number_list == NULL )
    {
        ias_l1g_close_image(img);
        fclose(ofile);
        IAS_LOG_ERROR("Allocating band number list of size %d", band_count );
        exit(EXIT_FAILURE);
    }
   
    if (ias_l1g_get_band_list( img, band_number_list, band_count, &band_count2) 
        != SUCCESS)
    {
        free(band_number_list);
        ias_l1g_close_image(img);
        fclose(ofile);
        IAS_LOG_ERROR("Reading band number list");
        exit(EXIT_FAILURE);
    }

    if (band_count2 != band_count)
    {
        IAS_LOG_ERROR("Returned band number count %d does not match %d",
            band_count2, band_count );
        free(band_number_list);
        ias_l1g_close_image(img);
        fclose(ofile);
        IAS_LOG_ERROR("Reading band number list");
        exit(EXIT_FAILURE);
    }

    for (band_index = 0; band_index < band_count; band_index++)
    {
        band_number = band_number_list[band_index];

        /* Check to see if band is present */
        if (ias_l1g_is_band_present(img, band_number) == 1)
        {
            /* Read the band metadata */
            if (ias_l1g_get_band_metadata(img, band_number, &bmd) != SUCCESS)
            {
                free(band_number_list);
                ias_l1g_close_image(img);
                fclose(ofile);
                IAS_LOG_ERROR("Reading band metadata");
                exit(EXIT_FAILURE);
            }
            if (print_band_meta(&bmd, &fmd, ofile) != SUCCESS) 
            {
                free(band_number_list);
                ias_l1g_close_image(img);
                fclose(ofile);
                IAS_LOG_ERROR("Printing band metadata");
                exit(EXIT_FAILURE);
            }
        }
        else
        {
            IAS_LOG_WARNING("Image %s does not contain band %d", l1g_name,
                band_number);
        }
    }

    free(band_number_list);
    ias_l1g_close_image(img);

    if (fclose(ofile) != 0)
    {
        IAS_LOG_ERROR("Unable to close output file %s.", outfile);
        exit(EXIT_FAILURE);
    }

    exit (EXIT_SUCCESS);
}
