/******************************************************************************
NAME:           print_band_meta

PURPOSE:
Print the band specific metadata to a file.

Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gctp.h"
#include "ias_geo.h"
#include "ias_logging.h"
#include "ias_l1g.h"
#include "ias_const.h"
#include "ias_ancillary_io.h"
#include "dump_l1g_meta.h"

int print_band_meta
(
    const IAS_L1G_BAND_METADATA *bmd,   /* I: Pointer to the band metadata */
    const IAS_L1G_FILE_METADATA *fmd,   /* I: Pointer to the file metadata */
    FILE *ofile                         /* I/O: Output file pointer */
)
{
    /* Set/Define the input parameters for gcp */
    double geo_parm[] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 
                         0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
    int geo_units = DEGREE;     /* projection units             */
    double in_x,in_y;           /* Input projection coordinates */
    double upper_left_out_x;    /* Upper left output projection coordinates */
    double upper_left_out_y;    /* Upper left output projection coordinates */
    double upper_right_out_x;   /* Upper right output projection coordinates */
    double upper_right_out_y;   /* Upper right output projection coordinates */
    double lower_left_out_x;    /* Lower left output projection coordinates */
    double lower_left_out_y;    /* Lower left output projection coordinates */
    double lower_right_out_x;   /* Lower right output projection coordinates */
    double lower_right_out_y;   /* Lower right output projection coordinates */
    int iunits;                 /* Image projection units */
    int status;
    IAS_PROJECTION geo_proj;
    IAS_PROJECTION target_proj;
    IAS_GEO_PROJ_TRANSFORMATION *map2degrees = NULL;

    /* Determine input projection units */
    if (ias_geo_get_units(fmd->projection_units, &iunits) != SUCCESS)
    {
        IAS_LOG_ERROR("Unknown projection units %s\n",fmd->projection_units);
        return ERROR;
    }

    /* This is the new thread safe projection call */
    ias_geo_set_projection(GEO, NULLZONE, geo_units, fmd->spheroid_code,
        geo_parm, &geo_proj);
    ias_geo_set_projection(fmd->projection_code, fmd->zone_code,
        iunits, fmd->spheroid_code, fmd->projection_parameters, &target_proj);

    map2degrees = ias_geo_create_proj_transformation(&target_proj, &geo_proj);
    if (!map2degrees)
    {
        IAS_LOG_ERROR("Creating projection transformation");
        return ERROR;
    }

    /* Upper Left */
    in_x = bmd->upper_left_x;
    in_y = bmd->upper_left_y;

    /* This gets us the lat/long from the X/Y */
    status = ias_geo_transform_coordinate(map2degrees, in_x, in_y,
                &upper_left_out_x, &upper_left_out_y);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Converting upper left corner to the geographic "
                    "projection");
        ias_geo_destroy_proj_transformation(map2degrees);
        return ERROR;
    }

    /* Upper Right */
    in_x = bmd->upper_right_x;
    in_y = bmd->upper_right_y;

    status = ias_geo_transform_coordinate(map2degrees, in_x, in_y,
                &upper_right_out_x, &upper_right_out_y);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Converting upper right corner to the geographic "
                    "projection");
        ias_geo_destroy_proj_transformation(map2degrees);
        return ERROR;
    }

    /* Lower Right */
    in_x = bmd->lower_right_x;
    in_y = bmd->lower_right_y;

    status = ias_geo_transform_coordinate(map2degrees, in_x, in_y,
                &lower_right_out_x, &lower_right_out_y);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Converting lower right corner to the geographic "
                    "projection");
        ias_geo_destroy_proj_transformation(map2degrees);
        return ERROR;
    }

    /* Lower Left */
    in_x = bmd->lower_left_x;
    in_y = bmd->lower_left_y;

    status = ias_geo_transform_coordinate(map2degrees, in_x, in_y,
                &lower_left_out_x, &lower_left_out_y);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Converting lower right corner to the geographic "
                    "projection");
        ias_geo_destroy_proj_transformation(map2degrees);
        return ERROR;
    }
    ias_geo_destroy_proj_transformation(map2degrees);
    
    if (fprintf(ofile, "Band %d: %s\n" 
        "   Upper left\n\t(y/x) %f %f\t\t(lat/long) %.8f %.8f\n"
        "   Upper right\n\t(y/x) %f %f\t\t(lat/long) %.8f %.8f\n"
        "   Lower right\n\t(y/x) %f %f\t\t(lat/long) %.8f %.8f\n"
        "   Lower left\n\t(y/x) %f %f\t\t(lat/long) %.8f %.8f\n"
        "   Projection distance: %f,%f\n   Valid min/max pixel: ", 
        bmd->band_number, bmd->band_name,
        bmd->upper_left_y, bmd->upper_left_x,
        upper_left_out_y, upper_left_out_x,
        bmd->upper_right_y, bmd->upper_right_x,
        upper_right_out_y, upper_right_out_x,
        bmd->lower_right_y, bmd->lower_right_x,
        lower_right_out_y, lower_right_out_x,
        bmd->lower_left_y, bmd->lower_left_x,
        lower_left_out_y, lower_left_out_x,
        bmd->projection_distance_y, bmd->projection_distance_x) < 0)
    {
        IAS_LOG_ERROR("Unable to write band metadata for band number %d to "
            "output file", bmd->band_number);
        return ERROR;
    }

    if (!bmd->pixel_range_valid)
    {
        status = fprintf(ofile, "FALSE\n");
    }
    else /* Pixel range data is valid, so print it */
    {
        status = fprintf(ofile, "TRUE\n   Max/Min pixel: %f,%f\n",
            bmd->maximum_pixel_value, bmd->minimum_pixel_value);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Unable to write band metadata for band number %d to "
            "output file", bmd->band_number);
        return ERROR;
    }

    if (!bmd->radiance_valid)
    {
        status = fprintf(ofile, "   Valid radiance: FALSE\n");
    }
    else /* Radiance data is valid, so print it */
    {
        status = fprintf(ofile, "   Valid radiance: TRUE\n"
            "   Max/Min radiance: %f,%f\n"
            "   Spectral Radiance Scaling  (offset) %f, (gain) %f\n",
            bmd->maximum_radiance, bmd->minimum_radiance,
            bmd->spectral_radiance_scaling_offset,
            bmd->spectral_radiance_scaling_gain);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Unable to write band metadata for band number %d to "
            "output file", bmd->band_number);
        return ERROR;
    }

    if (!bmd->reflectance_valid)
    {
        status = fprintf(ofile, "   Valid reflectance: FALSE\n");
    }
    else /* Reflectance data is valid, so print it */
    {
        status = fprintf(ofile, "   Valid reflectance: TRUE\n"
            "   Scaling reflectance: %f,%f\n", 
            bmd->reflectance_scaling_offset, bmd->reflectance_scaling_gain);
    }
    if (status < 0)
    {
        IAS_LOG_ERROR("Unable to write band metadata for band number %d to "
            "output file", bmd->band_number);
        return ERROR;
    }

    if (fprintf(ofile, "   Source: \"%s\"\n",bmd->instrument_source) < 0)
    {
        IAS_LOG_ERROR("Unable to write band metadata for band number %d to "
            "output file", bmd->band_number);
        return ERROR;
    }

    return SUCCESS;
}
