/****************************************************************************
NAME:       dump_los_model

PURPOSE:    Dump a line of sight model file to ASCII output.

DEVELOPMENT HISTORY:

RETURN VALUE:
Type=int
Value           Description
-----           -----------
EXIT_SUCCESS    dump_los_model completed successfully
EXIT_FAILURE    Error occurred while running dump_los_model

*******************************************************************************
                        Property of the U.S. Government
                             USGS EROS Data Center
*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "ias_const.h"
#include "ias_satellite_attributes.h"
#include "ias_logging.h"
#include "ias_los_model.h"
#include "ias_model_io.h"
#include "dump_los_model.h"

int main(int argc, char *argv[])
{
    const char *filename;           /* OLI model file name */
    const char *root_name;          /* Base output name to add to */
    IAS_LOS_MODEL *model = 0;       /* OLI model structure */

    if (ias_log_initialize("dump_los_model") != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing log");
        exit(EXIT_FAILURE);
    }

    /* Make sure an input file exists */
    if (argc != 3)
    {
        IAS_LOG_ERROR("Usage: dump_los_model modelfile outputroot");
        exit(EXIT_FAILURE);
    }

    filename = argv[1];
    root_name = argv[2];

    /* Initiate the model */
    ias_sat_attr_initialize(IAS_L8);

    model = ias_model_read(filename);
    if (!model)
    {
        IAS_LOG_ERROR("Error reading input model %s", filename);
        exit(EXIT_FAILURE);
    }

    /* Call print attitude routine */
    if (print_att_model(model, root_name) != SUCCESS)
    {
        IAS_LOG_ERROR("Error dumping attitude information from %s", filename);
        ias_los_model_free(model);
        exit(EXIT_FAILURE);
    }

    /* Call print ephemeris routine */
    if (print_eph_model(model, root_name) != SUCCESS)
    {
        IAS_LOG_ERROR("Error dumping ephemeris information from %s", filename);
        ias_los_model_free(model);
        exit(EXIT_FAILURE);
    }

    /* Call print jitter routine */
    if (print_jitter_model(model, root_name) != SUCCESS)
    {
        IAS_LOG_ERROR("Error dumping jitter information from %s", filename);
        ias_los_model_free(model);
        exit(EXIT_FAILURE);
    }

    if (print_ssm_model(model, root_name) != SUCCESS)
    {
        IAS_LOG_ERROR("Error dumping ssm information from %s", filename);
        ias_los_model_free(model);
        exit(EXIT_FAILURE);
    }

    if (print_frame_times(model, root_name) != SUCCESS)
    {
        IAS_LOG_ERROR("Error dumping frame time information from %s", filename);
        ias_los_model_free(model);
        exit(EXIT_FAILURE);
    }

    /* Free the memory */
    ias_los_model_free(model);

    /* Write completion message */
    IAS_LOG_INFO("Successful completion");

    return(EXIT_SUCCESS);
}
