/*****************************************************************************
NAME: PRINT_ATT_MODEL

PURPOSE:
Write the attitude data from the LOS model file to an output file

RETURN VALUE:
Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    An error occurred during processing

ALGORITHM REFERENCES:
******************************************************************************
                        Property of the U.S. Government
                              USGS EROS Data Center
*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_los_model.h"
#include "ias_model_io.h"
#include "dump_los_model.h"

int print_att_model(const IAS_LOS_MODEL *mod, const char *root_name)
{
    char outfilename[PATH_MAX];
    int att_idx;
    FILE *ofile;

    int status;
    int max_name = PATH_MAX;

    status = snprintf(outfilename, sizeof(outfilename), "%s.att.txt", 
        root_name);
    if ((status < 0) || (status >= max_name))
    {
        IAS_LOG_ERROR("Unable to use output name as entered, size limit %d",
            max_name - 9);
        return ERROR;
    }

    ofile = fopen(outfilename,"w");
    if( ofile == NULL )
    {
        IAS_LOG_ERROR("Unable to open output file %s",outfilename);
        return ERROR;
    }

    if (fprintf(ofile,"Entry_Number,Epoch_UTC_Year,Epoch_UTC_Day,Epoch_UTC_SOD,"
        "SAT_X,SAT_Y,SAT_Z,PREC_X,PREC_Y,PREC_Z\n") < 0)
    {
        IAS_LOG_ERROR("Unable to write to output file %s",outfilename);
        fclose(ofile);
        return ERROR;
    } 

    for (att_idx = 0; att_idx < mod->spacecraft.attitude.sample_count;
            att_idx++)
    {
        if (fprintf(ofile,"%d,%4.0f,%3.0f,%.6f,%e,%e,%e,%e,%e,%e\n",att_idx,
            mod->spacecraft.attitude.utc_epoch_time[0],
            mod->spacecraft.attitude.utc_epoch_time[1],
            mod->spacecraft.attitude.utc_epoch_time[2] + 
            (mod->spacecraft.
                attitude.sample_records[att_idx].seconds_from_epoch),
            mod->spacecraft.attitude.sample_records[att_idx].
                satellite_attitude.x,
            mod->spacecraft.attitude.sample_records[att_idx].
                satellite_attitude.y,
            mod->spacecraft.attitude.sample_records[att_idx].
                satellite_attitude.z,
            mod->spacecraft.attitude.sample_records[att_idx].
                precision_attitude.x,
            mod->spacecraft.attitude.sample_records[att_idx].
                precision_attitude.y,
            mod->spacecraft.attitude.sample_records[att_idx].
                precision_attitude.z) < 0)
            {
                IAS_LOG_ERROR("Unable to write to output file %s",outfilename);
                fclose(ofile);
                return ERROR;
            }
    }

    /* Close output file */
    if (fclose(ofile) != 0)
    {
        IAS_LOG_ERROR("File %s was not closed properly",outfilename);
        return ERROR;
    }

    return SUCCESS;
}
