/*****************************************************************************
NAME: PRINT_EPH_MODEL

PURPOSE:
Write the ephemeris model data from the LOS model to an output file.

RETURN VALUE:
Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    An error occurred during processing

ALGORITHM REFERENCES:
******************************************************************************
                        Property of the U.S. Government
                              USGS EROS Data Center
*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_los_model.h"
#include "ias_model_io.h"
#include "dump_los_model.h"

int print_eph_model(const IAS_LOS_MODEL *mod, const char *root_name)
{
    char outfilename[PATH_MAX];
    int eph_idx;
    FILE *ofile;

    int status;
    int max_name = PATH_MAX;

    status = snprintf(outfilename, sizeof(outfilename), "%s.eph.txt", 
        root_name);
    if ((status < 0) || (status >= max_name))
    {
        IAS_LOG_ERROR("Unable to use output name as entered, size limit %d",
            max_name - 9);
        return ERROR;
    }
    

    ofile = fopen(outfilename,"w");
    if( ofile == NULL )
    {
        IAS_LOG_ERROR("Unable to open output file %s",outfilename);
        return ERROR;
    }

    if (fprintf(ofile,"Entry_Number,Epoch_UTC_Year,Epoch_UTC_Day,Epoch_UTC_SOD,"
        "ECI_X,ECI_Y,ECI_Z,ECI_VX,ECI_VY,ECI_VZ,ECEF_X,ECEF_Y,ECEF_Z,"
        "ECEF_VX,ECEF_VY,ECEF_VZ,"
        "PREC_ECI_X,PREC_ECI_Y,PREC_ECI_Z,"
        "PREC_ECI_VX,PREC_ECI_VY,PREC_ECI_VZ,"
        "PREC_ECEF_X,PREC_ECEF_Y,PREC_ECEF_Z,"
        "PREC_ECEF_VX,PREC_ECEF_VY,PREC_ECEF_VZ\n") < 0)
    {
        IAS_LOG_ERROR("Unable to write to output file %s",outfilename);
        fclose(ofile);
        return ERROR;
    }

    for (eph_idx = 0; eph_idx < mod->spacecraft.ephemeris.sample_count;
            eph_idx++)
    {
        if (fprintf(ofile,"%d,%4.0f,%3.0f,%.6f,%f,%f,%f,%f,%f,%f,"
            "%f,%f,%f,%f,%f,%f,"
            "%f,%f,%f,%f,%f,%f,"
            "%f,%f,%f,%f,%f,%f"
            "\n",eph_idx,
            mod->spacecraft.ephemeris.utc_epoch_time[0],
            mod->spacecraft.ephemeris.utc_epoch_time[1],
            mod->spacecraft.ephemeris.utc_epoch_time[2] + 
            (mod->spacecraft.
                ephemeris.sample_records[eph_idx].seconds_from_epoch),
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                eci_position.x,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                eci_position.y,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                eci_position.z,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                eci_velocity.x,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                eci_velocity.y,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                eci_velocity.z,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                ecef_position.x,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                ecef_position.y,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                ecef_position.z,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                ecef_velocity.x,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                ecef_velocity.y,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                ecef_velocity.z,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                precision_eci_position.x,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                precision_eci_position.y,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                precision_eci_position.z,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                precision_eci_velocity.x,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                precision_eci_velocity.y,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                precision_eci_velocity.z,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                precision_ecef_position.x,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                precision_ecef_position.y,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                precision_ecef_position.z,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                precision_ecef_velocity.x,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                precision_ecef_velocity.y,
            mod->spacecraft.ephemeris.sample_records[eph_idx].
                precision_ecef_velocity.z) < 0)
            {
                IAS_LOG_ERROR("Unable to write to output file %s",outfilename);
                fclose(ofile);
                return ERROR;
            }
    }

    /* Close file */
    if (fclose(ofile) != 0)
    {
        IAS_LOG_ERROR("File %s was not closed properly",outfilename);
        return ERROR;
    }

    return SUCCESS;
}
