/*******************************************************************************
NAME: print_frame_times

PURPOSE: Write just the frame times from the LOS model file to ASCII output.

RETURN VALUE:
Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    An error occurred during processing

*******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_los_model.h"
#include "ias_model_io.h"
#include "ias_miscellaneous.h"
#include "dump_los_model.h"

int print_frame_times(const IAS_LOS_MODEL *mod, const char *root_name)
{
    int sensor_index;
    const IAS_SENSOR_MODEL *sensor_model;

    sensor_model = &mod->sensor;

    for (sensor_index = 0; sensor_index < IAS_MAX_SENSORS; sensor_index++)
    {
        if (sensor_model->sensors[sensor_index].sensor_present)
        {
            char outfilename[PATH_MAX];
            FILE *ofile;
            int status;
            int band_index;
            int index;
            const IAS_SENSOR_BAND_MODEL *band_model;
            const char *sensor_name = ias_sat_attr_get_sensor_name(
                sensor_index);
            char lc_sensor_name[20];

            /* Get the lower case sensor name */
            strncpy(lc_sensor_name, sensor_name, sizeof(lc_sensor_name));
            lc_sensor_name[sizeof(lc_sensor_name) - 1] = '\0';
            ias_misc_convert_to_lowercase(lc_sensor_name);

            for (band_index = 0; band_index < sensor_model->band_count;
                 band_index++)
            {
                if (sensor_model->bands[band_index].band_present == 1
                    && sensor_model->bands[band_index].sensor->sensor_id
                        == sensor_index
                    )
                    break;
            }

            if (band_index == sensor_model->band_count)
            {
                IAS_LOG_ERROR("Unable to find a %s band", sensor_name);
                return ERROR;
            }

            band_model = &mod->sensor.bands[band_index];

            status = snprintf(outfilename, sizeof(outfilename),
                "%s.%s_frames.txt", root_name, lc_sensor_name);
            if ((status < 0) || (status >= sizeof(outfilename)))
            {
                IAS_LOG_ERROR("Unable to use output name as entered, "
                    "size limit %d", (int)sizeof(outfilename) - 9);
                return ERROR;
            }

            /* Open file */
            ofile = fopen(outfilename,"w");
            if (ofile == NULL)
            {
                IAS_LOG_ERROR("Unable to open output file %s", outfilename);
                return ERROR;
            }

            /* Begin printing to file */
            /* Write table header */
            if (fprintf(ofile, "Record_Index,Epoch_UTC_Year,Epoch_UTC_Day,"
                "Epoch_UTC_SOD,Frame_seconds_from_epoch\n") < 0)
            {
                IAS_LOG_ERROR("Unable to write to output file %s", outfilename);
                fclose(ofile);
                return ERROR;
            } 

            /* Write the data */
            for (index = 0; index < band_model->frame_count; index++)
            {
                if (fprintf(ofile,"%d,%f,%f,%f,%e\n", 
                    index,
                    band_model->utc_epoch_time[0],
                    band_model->utc_epoch_time[1],
                    band_model->utc_epoch_time[2],
                    band_model->frame_seconds_from_epoch[index]) < 0)
                {
                    IAS_LOG_ERROR("Unable to write to output file %s",
                        outfilename);
                    fclose(ofile);
                    return ERROR;
                }
            }

            /* Close output file */
            if (fclose(ofile) != 0)
            {
                IAS_LOG_ERROR("File %s was not closed properly", outfilename);
                return ERROR;
            }
        }
    }

    return SUCCESS;
}
