/*****************************************************************************
NAME: PRINT_JITTER_MODEL

PURPOSE:
Write the jitter data from the LOS model to an output file.

RETURN VALUE:
Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    An error occurred during processing

ALGORITHM REFERENCES:
******************************************************************************
                        Property of the U.S. Government
                              USGS EROS Data Center
*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include "ias_types.h"
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_los_model.h"
#include "ias_model_io.h"
#include "ias_sensor_model.h"
#include "dump_los_model.h"

int print_jitter_model(const IAS_LOS_MODEL *mod, const char *root_name)
{
    char outfilename[PATH_MAX]; /* Filename for Jitter output */
    const char *sen_name;           /* Name of this sensor */
    int jit_idx;                /* Jitter index */
    int sen_idx;                /* Sensor index */
    int bnd_idx = 0;            /* Band index */
    int frame_entries = 0;      /* Number of frame entries for the sensor */
    FILE *ofile;                /* Output file */
    double year, day, sod;      /* Included in each line */
    double time_from_start;     /* This value is added to sod */

    int status;
    int max_name = PATH_MAX;

    /* prepare Jitter output file */
    status = snprintf(outfilename, sizeof(outfilename), "%s.jitter.txt", 
        root_name);
    if ((status < 0) || (status >= max_name))
    {
        IAS_LOG_ERROR("Unable to use output name as entered, size limit %d",
            max_name - 12);
        return ERROR;
    }

    ofile = fopen(outfilename,"w");
    if( ofile == NULL )
    {
        IAS_LOG_ERROR("Unable to open output file %s",outfilename);
        return ERROR;
    }

    if (fprintf(ofile,"Sensor,Entry_Number,"
        "Epoch_UTC_Year,Epoch_UTC_Day,Epoch_UTC_SOD,"
        "Roll_Jitter,Pitch_Jitter,Yaw_Jitter\n") < 0)
    {
        IAS_LOG_ERROR("Unable to open output file %s",outfilename);
        return ERROR;
    }

    for (sen_idx = 0; sen_idx < IAS_MAX_SENSORS; sen_idx++)
    {
        int epoch_time_found;   /* Flag for epoch time found */

        /* Get the established name for this sensor index */
        sen_name = ias_sat_attr_get_sensor_name(sen_idx);

        if (!mod->sensor.sensors[sen_idx].sensor_present)
        {
            /* Sensor is not present in this image skip this sensor */
            IAS_LOG_INFO("Sensor %s not present",sen_name);
            continue;
        }

        /* Find the number of entries per frame for this sensor */
        frame_entries =  mod->sensor.sensors[sen_idx].
                        jitter_entries_per_frame;

        /* Find a band present where the lines per frame matches the jitter
           jitter entries per frame.  Use that band for the epoch time for all
           jitter values.  If no such band is found, send an informational 
           message indicating no bands were found for that sensor */
        epoch_time_found = FALSE;
        for (bnd_idx = 0; bnd_idx < mod->sensor.band_count; bnd_idx++)
        {
            if ((mod->sensor.bands[bnd_idx].band_present)
                && ((mod->sensor.bands[bnd_idx].sensor)->sensor_id == sen_idx)
                && (mod->sensor.bands[bnd_idx].sampling_char.lines_per_frame
                == frame_entries))
            {
                epoch_time_found = TRUE;

                /* Found it.  Break out of the band for loop */
                break;
            }
        }

        if (!epoch_time_found)
        {
            IAS_LOG_INFO("Could not find valid band information for sensor %s",
                sen_name);
            break;
        }

        year = mod->sensor.bands[bnd_idx].utc_epoch_time[0];
        day = mod->sensor.bands[bnd_idx].utc_epoch_time[1];
        sod = mod->sensor.bands[bnd_idx].utc_epoch_time[2];

        /* Process this sensor */
        for (jit_idx = 0; 
                jit_idx < mod->sensor.sensors[sen_idx].jitter_table_count;
                jit_idx++)
        {
            if (ias_sensor_find_time(jit_idx, 0.0, bnd_idx, 0, &(mod->sensor),
                IAS_NOMINAL_DETECTOR, &time_from_start) != SUCCESS)
            {
                IAS_LOG_ERROR("Unable to find start time for band index %d, "
                    "sensor %s, jitter %d",
                    bnd_idx, sen_name, jit_idx + 1);
                fclose(ofile);
                return ERROR;
            }

            /* Print to file */
            if (fprintf(ofile,"%s,%d,%4.0f,%3.0f,%.6f,"
                    "%16.15e,%17.16e,%16.15e\n",
                    sen_name,jit_idx, year, day, sod + time_from_start,
                    mod->sensor.sensors[sen_idx].jitter_table[jit_idx].x,
                    mod->sensor.sensors[sen_idx].jitter_table[jit_idx].y,
                    mod->sensor.sensors[sen_idx].jitter_table[jit_idx].z) < 0)
            {
                IAS_LOG_ERROR("Unable to write to output file %s",outfilename);
                fclose(ofile);
                return ERROR;
            }
        }   /* End jitter loop */
    }   /* End sensor loop */

    /* Close output file */
    if (fclose(ofile) != 0)
    {
        IAS_LOG_ERROR("File %s was not closed properly",outfilename);
        return ERROR;
    }

    return SUCCESS;
}
