
/*****************************************************************************
 NAME: extract_l1g_area.c

 PURPOSE: To extract a single band window from an L1G/T input file and create
          an L1G compliant output file.

 RETURN VALUE: Type = int
    Value         Description
    ------------  --------------------------
    EXIT_SUCCESS  Successful completion
    EXIT_FAILURE  Program terminated with errors

*****************************************************************************/

/* Standard C Includes */
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/* IAS Includes */
#include "gctp.h"
#include "ias_const.h"
#include "ias_types.h"
#include "ias_logging.h"
#include "ias_l1g.h"
#include "ias_gcp.h"
#include "ias_geo.h"
#include "ias_miscellaneous.h"

/* Local Includes */
#include "extract_l1g_area.h"

int main(int argc, char *argv[])
{
    int status; /* Capture function/routine statuses */
    int index;  /* Loop control */
    int data_count; /* The number of data samples */
    int data_type_size; /* Size of the data type */
    EXTRACT_AREA_PARAMETERS parameters; /* Holds the input parameters */

    /*** L1G Input/Output Data Variables ***/
    EXTRACT_AREA_L1G_DATA input_data;
    EXTRACT_AREA_L1G_DATA output_data;
    int upper_line;  /* The upper line of the extracted image data */
    int left_sample; /* The left sample of the extracted image data */

    IAS_R4_TYPE *input_r4_buffer = NULL;

    void *output_buffer = NULL;

    /* Initialize the logging */
    status = ias_log_initialize("extract_l1g_area");
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the logging library");
        return EXIT_FAILURE;
    }

    /* Verify that the parameter filename was provided */
    if (argc != 2)
    {
        IAS_LOG_ERROR("Usage: extract_l1g_area <parameter filename>");
        return EXIT_FAILURE;
    }

    /* Read the parameters from the parameter file */
    status = read_parameters(argv[1], &parameters);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading input parameters");
        return EXIT_FAILURE;
    }

    /* Notify the user that the application has started */
    IAS_LOG_INFO("Started");

    /* Set the amount of data we are going to extract */
    data_count = parameters.lines_samples[0] * parameters.lines_samples[1];

    /* Open the input image and load the metadata */
    status = open_l1g_data(&parameters, &input_data, &output_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Loading input image data %s", parameters.l1g_filename);
        return EXIT_FAILURE;
    }

    /* Test for invalid data types and figure out data type size */
    status = ias_misc_get_sizeof_data_type(input_data.data_type,
        &data_type_size);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Input image contains unsupported data type");
        close_l1g_data(&input_data, &output_data);
        return EXIT_FAILURE;
    }

    /* Further restrict the supported data types for this application */
    switch (input_data.data_type)
    {
        case IAS_R8:
        case IAS_CHAR:
        case IAS_UCHAR:
            IAS_LOG_ERROR("Input image contains unsupported data type");
            close_l1g_data(&input_data, &output_data);
            return EXIT_FAILURE;
            
        default:
            /* We have a valid type for this application */
            break;
    }

    /* Allocate memory for the input data */
    input_r4_buffer = malloc(sizeof(*input_r4_buffer) * data_count);
    if (input_r4_buffer == NULL)
    {
        IAS_LOG_ERROR("Allocating the image buffer");
        close_l1g_data(&input_data, &output_data);
        return EXIT_FAILURE;
    }

    /* Allocate memory for the output data */
    output_buffer = malloc(data_type_size * data_count);
    if (output_buffer == NULL)
    {
        IAS_LOG_ERROR("Allocating the image buffer");
        free(input_r4_buffer);
        close_l1g_data(&input_data, &output_data);
        return EXIT_FAILURE;
    }

    /* Extract the data from the input */
    status = ias_geo_extract_window(
        input_data.number_of_image_lines, input_data.number_of_image_samples,
        0, input_data.data_type,
        parameters.center_line_sample[0],
        parameters.center_line_sample[1],
        parameters.lines_samples, input_data.band_io_ptr, input_r4_buffer,
        &upper_line, &left_sample);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Failed to extract the image area");
        free(output_buffer);
        free(input_r4_buffer);
        close_l1g_data(&input_data, &output_data);
        return EXIT_FAILURE;
    }

    /* Copy the input buffer to type'd output buffer */
    switch (input_data.data_type)
    {
        case IAS_BYTE:
        {
            IAS_BYTE_TYPE *output_byte_buffer = output_buffer;
            for (index = 0; index < data_count; index++)
                output_byte_buffer[index] = input_r4_buffer[index];
            break;
        }
        case IAS_I2:
        {
            IAS_I2_TYPE *output_i2_buffer = output_buffer;
            for (index = 0; index < data_count; index++)
                output_i2_buffer[index] = input_r4_buffer[index];
            break;
        }
        case IAS_UI2:
        {
            IAS_UI2_TYPE *output_ui2_buffer = output_buffer;
            for (index = 0; index < data_count; index++)
                output_ui2_buffer[index] = input_r4_buffer[index];
            break;
        }
        case IAS_I4:
        {
            IAS_I4_TYPE *output_i4_buffer = output_buffer;
            for (index = 0; index < data_count; index++)
                output_i4_buffer[index] = input_r4_buffer[index];
            break;
        }
        case IAS_UI4:
        {
            IAS_UI4_TYPE *output_ui4_buffer = output_buffer;
            for (index = 0; index < data_count; index++)
                output_ui4_buffer[index] = input_r4_buffer[index];
            break;
        }
        case IAS_R4:
        {
            IAS_R4_TYPE *output_r4_buffer = output_buffer;
            for (index = 0; index < data_count; index++)
                output_r4_buffer[index] = input_r4_buffer[index];
            break;
        }
        default:
            /* Can't get here because of the other switch but just in case,
               return ERROR */
            return EXIT_FAILURE;
    }

    /* The input buffer is no longer needed */
    free(input_r4_buffer);
    input_r4_buffer = NULL;

    /* Write the new l1g output file */
    status = ias_l1g_write_image(output_data.band_io_ptr, 0,
        0, 0, output_data.number_of_image_lines,
        output_data.number_of_image_samples, output_buffer);

    /* Free this before we check the status because it is no longer needed */
    free(output_buffer);
    output_buffer = NULL;

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing output image data");
        close_l1g_data(&input_data, &output_data);
        return EXIT_FAILURE;
    }

    /* Fix the band metadata corner coordinates
       The math below assumes a north-up image */
    /* ULX */
    output_data.band_metadata.upper_left_x =
        input_data.band_metadata.upper_left_x
        + (left_sample * output_data.band_metadata.projection_distance_x);

    /* LLX */
    output_data.band_metadata.lower_left_x =
        output_data.band_metadata.upper_left_x;

    /* URX */
    output_data.band_metadata.upper_right_x =
        output_data.band_metadata.upper_left_x
        + ((parameters.lines_samples[1] - 1)
            * output_data.band_metadata.projection_distance_x);

    /* LRX */
    output_data.band_metadata.lower_right_x =
        output_data.band_metadata.upper_right_x;

    /* ULY */
    output_data.band_metadata.upper_left_y =
        input_data.band_metadata.upper_left_y
        - (upper_line * output_data.band_metadata.projection_distance_y);

    /* URY */
    output_data.band_metadata.upper_right_y =
        output_data.band_metadata.upper_left_y;

    /* LLY */
    output_data.band_metadata.lower_left_y =
        output_data.band_metadata.upper_left_y
        - ((parameters.lines_samples[0] - 1)
            * output_data.band_metadata.projection_distance_y);

    /* LRY */
    output_data.band_metadata.lower_right_y =
        output_data.band_metadata.lower_left_y;

    /* Update the band metadata */
    if (ias_l1g_set_band_metadata(output_data.io_ptr,
            &output_data.band_metadata, 1) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing band metadata for output band");
        close_l1g_data(&input_data, &output_data);
        return EXIT_FAILURE;
    }

    /* Close the images */
    status = close_l1g_data(&input_data, &output_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Failed to properly close bands or images");
        return EXIT_FAILURE;
    }

    IAS_LOG_INFO("Terminated");

    return EXIT_SUCCESS;
}

