
/* Standard C Includes */
#include <stdlib.h>

/* IAS Includes */
#include "ias_const.h"
#include "ias_l1g.h"
#include "ias_logging.h"

/* Local Includes */
#include "extract_l1g_area.h"

/*****************************************************************************
 NAME: open_l1g_data

 PURPOSE: Opens and loads the required information related to the specified
          input and output L1G data images.  Initializes and creates the
          initial output band.

 RETURN VALUE: Type = int
    Value    Description
    -------  --------------------------
    SUCCESS  Successful completion
    ERROR    Operation failed

*****************************************************************************/
int open_l1g_data
(
    const EXTRACT_AREA_PARAMETERS *parameters, /* I: Input parameters */
    EXTRACT_AREA_L1G_DATA *input_l1g_data, /* O: The file information to open
                                                 the input file for reading */
    EXTRACT_AREA_L1G_DATA *output_l1g_data /* O: The file information to open
                                                 the output file for writing */
)
{
    int status; /* capture function/routine statuses */

    const char *input_filename = parameters->l1g_filename;
    const char *output_filename = parameters->output_filename;
    int band_number = parameters->band_number;

    /* Open the input image */
    input_l1g_data->io_ptr = ias_l1g_open_image(input_filename, IAS_READ);
    if (input_l1g_data->io_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening input image %s", input_filename);
        return ERROR;
    }

    /* Open the output image */
    output_l1g_data->io_ptr = ias_l1g_open_image(output_filename, IAS_WRITE);
    if (output_l1g_data->io_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening output image %s", output_filename);
        ias_l1g_close_image(input_l1g_data->io_ptr);
        return ERROR;
    }

    /* Load the metadata */
    status = ias_l1g_get_file_metadata(input_l1g_data->io_ptr,
        &input_l1g_data->metadata);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving file metadata for %s", input_filename);
        ias_l1g_close_image(input_l1g_data->io_ptr);
        ias_l1g_close_image(output_l1g_data->io_ptr);
        return ERROR;
    }

    /* Copy the input image's metadata to the output image */
    if (ias_l1g_set_file_metadata(output_l1g_data->io_ptr,
            &input_l1g_data->metadata) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing output image's metadata %s", output_filename);
        ias_l1g_close_image(input_l1g_data->io_ptr);
        ias_l1g_close_image(output_l1g_data->io_ptr);
        return ERROR;
    }

    /* Verify the band is present */
    if (!ias_l1g_is_band_present(input_l1g_data->io_ptr, band_number))
    {
        IAS_LOG_ERROR("BAND_NUMBER %d not present in the input data",
            band_number);
        ias_l1g_close_image(input_l1g_data->io_ptr);
        ias_l1g_close_image(output_l1g_data->io_ptr);
        return ERROR;
    }

    /* Open the input band */
    input_l1g_data->band_io_ptr = ias_l1g_open_band(input_l1g_data->io_ptr,
        band_number, &input_l1g_data->data_type,
        &input_l1g_data->number_of_scas,
        &input_l1g_data->number_of_image_lines,
        &input_l1g_data->number_of_image_samples);
    if (input_l1g_data->band_io_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening L1G band %d of image %s", band_number,
            input_filename);
        close_l1g_data(input_l1g_data, output_l1g_data);
        return ERROR;
    }

    /* Verify that only 1 SCA is present in the image data */
    if (input_l1g_data->number_of_scas > SUPPORTED_SCAS)
    {
        IAS_LOG_ERROR("Only 1 SCA is supported, found %d in image %s",
            input_l1g_data->number_of_scas, input_filename);
        close_l1g_data(input_l1g_data, output_l1g_data);
        return ERROR;
    }

    /* Get the band metadata */
    status = ias_l1g_get_band_metadata(input_l1g_data->io_ptr,
        band_number, &input_l1g_data->band_metadata);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading band %d of metadata", band_number);
        close_l1g_data(input_l1g_data, output_l1g_data);
        return ERROR;
    }

    /* Copy some input to output information */
    output_l1g_data->data_type = input_l1g_data->data_type;
    output_l1g_data->number_of_scas = input_l1g_data->number_of_scas;
    output_l1g_data->number_of_image_lines = parameters->lines_samples[0];
    output_l1g_data->number_of_image_samples = parameters->lines_samples[1];

    /* Open the output band */
    output_l1g_data->band_io_ptr = ias_l1g_open_band(output_l1g_data->io_ptr,
        band_number,
        &input_l1g_data->data_type, &output_l1g_data->number_of_scas,
        &output_l1g_data->number_of_image_lines,
        &output_l1g_data->number_of_image_samples);
    if (input_l1g_data->band_io_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening L1G band %d of image %s", band_number,
            output_filename);
        close_l1g_data(input_l1g_data, output_l1g_data);
        return ERROR;
    }

    /* Copy the band metadata */
    output_l1g_data->band_metadata = input_l1g_data->band_metadata;
    output_l1g_data->band_metadata.band_number = band_number;

    /* Set the initial band metadata */
    if (ias_l1g_set_band_metadata(output_l1g_data->io_ptr,
            &output_l1g_data->band_metadata, 1) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing band metadata for output band");
        close_l1g_data(input_l1g_data, output_l1g_data);
        return ERROR;
    }

    return SUCCESS;
}

/*****************************************************************************
 NAME: close_l1g_data

 PURPOSE: Closes all the open information related to the specified input and
          output L1G data files.

 RETURN VALUE: Type = int
    Value    Description
    -------  --------------------------
    SUCCESS  Successful completion
    ERROR    Failed to close something

*****************************************************************************/
int close_l1g_data
(
    const EXTRACT_AREA_L1G_DATA *input_l1g_data, /* O: The file information
                                                       needed to close the
                                                       input file */
    const EXTRACT_AREA_L1G_DATA *output_l1g_data /* O: The file information
                                                       needed to close the
                                                       output file */
)
{
    int status = SUCCESS;
    int return_status = SUCCESS;

    if (input_l1g_data->band_io_ptr != NULL)
    {
        status = ias_l1g_close_band(input_l1g_data->band_io_ptr);
        if (status != SUCCESS)
            return_status = ERROR;
    }

    if (input_l1g_data->io_ptr != NULL)
    {
        status = ias_l1g_close_image(input_l1g_data->io_ptr);
        if (status != SUCCESS)
            return_status = ERROR;
    }

    if (output_l1g_data->band_io_ptr != NULL)
    {
        status = ias_l1g_close_band(output_l1g_data->band_io_ptr);
        if (status != SUCCESS)
            return_status = ERROR;
    }

    if (output_l1g_data->io_ptr != NULL)
    {
        status = ias_l1g_close_image(output_l1g_data->io_ptr);
        if (status != SUCCESS)
            return_status = ERROR;
    }

    return return_status;
}

