
/* Standard C Includes */
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <float.h>

/* IAS Includes */
#include "ias_const.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"

/* Local Includes */
#include "extract_l1g_area.h"

/*****************************************************************************
 NAME: read_parameters

 PURPOSE: To read parameters from the specified input ODL file.

 RETURN VALUE: Type = int
    Value    Description
    -------  --------------------------
    SUCCESS  Successful completion
    ERROR    Operation failed

*****************************************************************************/
int read_parameters
(
    const char *parameter_filename, /* I: The name of the parameter file */
    EXTRACT_AREA_PARAMETERS *parameters /* O: Parameters */
)
{
    int status; /* capture function/routine statuses */
    int count; /* count of elements in the parameter definition table */

    /* Zero the parameters */
    memset(parameters, 0, sizeof(EXTRACT_AREA_PARAMETERS));

    /* -------------------------------------------------------------------- */
    /* Common parameters table definition for the parameter file */

    IAS_PARM_DECLARE_TABLE(parameter_table, 5);

    IAS_PARM_L1G_FILENAME(parameter_table,
        parameters->l1g_filename, sizeof(parameters->l1g_filename), 0);

    IAS_PARM_ADD_INT(parameter_table, BAND_NUMBER,
        "Defines the band to extract data from",
        IAS_PARM_REQUIRED, IAS_PARM_NOT_ARRAY, 1, 1, 11, 0, NULL,
        &parameters->band_number, sizeof(parameters->band_number), 0);

    const char *default_output_filename[] = {"output.h5"};
    IAS_PARM_ADD_STRING(parameter_table, OUTPUT_FILENAME,
        "Name of the output l1g file",
        IAS_PARM_OPTIONAL, 0, NULL, 1, default_output_filename,
        parameters->output_filename, sizeof(parameters->output_filename), 0);

    IAS_PARM_ADD_INT(parameter_table, CENTER_LINE_SAMPLE,
        "Location within the input image for extraction",
        IAS_PARM_REQUIRED, IAS_PARM_ARRAY, 0, 0, 0, 0, NULL,
        parameters->center_line_sample,
        sizeof(parameters->center_line_sample), 2);

    IAS_PARM_ADD_INT(parameter_table, LINES_SAMPLES,
        "The number of lines and samples to extract",
        IAS_PARM_REQUIRED, IAS_PARM_ARRAY, 0, 0, 0, 0, NULL,
        parameters->lines_samples, sizeof(parameters->lines_samples), 2);

    /* -------------------------------------------------------------------- */
    /* Read the parameter file */

    /* Calculate the number of items in the parameter file */
    count = IAS_PARM_GET_TABLE_SIZE(parameter_table);

    /* Call the ias_parm_provide_help routine to provide help for the
       parameters if requested; also verifies the ODL file exists */
    status = ias_parm_provide_help(parameter_filename, parameter_table,
        count, IAS_INPUT_PARAMETERS);

    if (status)
    {
        /* Either help was provided or an error occurred */
        if (status == 1)
        {
              exit(EXIT_SUCCESS);
        }
        else
        {
            IAS_LOG_ERROR("Providing application parameter file help");
            return ERROR;
        }
    }
    else
    {
        /* Read the parameter file */
        status = ias_parm_read(parameter_filename, parameter_table, count);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading parameters from: %s", parameter_filename);
            return ERROR;
        }
    } /* END read parameter file */

    return SUCCESS;
}

