#!/usr/bin/perl -w
###############################################################################
#   NAME:   ias_assess_geo.pl
#
#   PURPOSE: Produce a report that helps CalVal Analysts assess the
#            geodetic or geometric accuracy over multiple scenes.
#
#   PARAMETERS: a - date range of image acquisition (dd-mmm-yyyy,dd-mmm-yyyy)
#               g - GCP source ([G]LS or [D]OQ)
#               r - off-nadir angle (absolute maximum offset from nadir
#                   in degrees)
#               w - geographic location of image extent (WRS Path/Row)
#                   (pathMin-pathMax,rowMin-rowMax)
#                   (max path range 1-233, max row range 1-248)
#               h - display this usage message
#               d - disable interactive prompting for missing or invalid
#                   parameter values
#
#   Another parameter that should be automatically sent in (and therefore
#   is not documented in the tool's usage text):
#
#               v - type of processing (geometric or geodetic)
#
#   OUTPUT:  Report file
###############################################################################

use strict;                 # enforce the rules
use Getopt::Long;           # parse command line options
use Logger;                 # to log errors, warnings, info msgs
use ias_const;              # for things like SUCCESS, FAILURE
use ias_db;                 # for database access
use ias_processing_system;  # ias_init_system and ias_validate_environment
use ias_assess_geo_and_x2x;
                    # local package with shared subroutines with ias_assess_x2x

use constant {
    GEOD    => 'Geodetic',
    GEOM    => 'Geometric'
};

# Define list of environment variables needed by ias_db.pm that we should
# check up-front to bail out sooner if our environment isn't setup as required
my @envVarsList = qw(
        IAS_DB_L8_OLITIRS
        ORACLE_SID
        );
my $envVarsListRef = \@envVarsList;

# Define database field names and stats headers needed for each application
# mode. These allow the code to be more generic of the processing mode and
# flexible to be extended to other uses as well.

#---------------------------------------------------------------
# GEODETIC mode supporting data
#---------------------------------------------------------------
my @geodDBFields = qw(
        L0RP_IMAGE_FILE
        WORK_ORDER_ID
        TARGET_WRS_PATH
        TARGET_WRS_ROW
        OFF_NADIR_ANGLE
        DATE_ACQUIRED
        NBR_OF_VLD_PNTS
        MEAN_LAT_OF_GCPS
        MEAN_LON_OF_GCPS
        PST_FIT_MEAN_ALNG_TRK
        PST_FIT_MEAN_ACRSS_TRK
        PST_FIT_STD_DEV_ALNG_TRK
        PST_FIT_STD_DEV_ACRSS_TRK
        PST_FIT_RMSE_ALNG_TRK
        PST_FIT_RMSE_ACRSS_TRK
        PRE_FIT_MEAN_ALNG_TRK
        PRE_FIT_MEAN_ACRSS_TRK
        PRE_FIT_STD_DEV_ALNG_TRK
        PRE_FIT_STD_DEV_ACRSS_TRK
        PRE_FIT_RMSE_ALNG_TRK
        PRE_FIT_RMSE_ACRSS_TRK
        GCP_SOURCE
        PRE_FIT_CORR_COEFF
        PST_FIT_CORR_COEFF
        );

my @geodNetStatsDBFields = qw(
        PRE_FIT_MEAN_ALNG_TRK
        PRE_FIT_MEAN_ACRSS_TRK
        PRE_FIT_RMSE_ALNG_TRK
        PRE_FIT_RMSE_ACRSS_TRK
        PST_FIT_MEAN_ALNG_TRK
        PST_FIT_MEAN_ACRSS_TRK
        PST_FIT_RMSE_ALNG_TRK
        PST_FIT_RMSE_ACRSS_TRK
        );
my $geodNetStatsDBFieldsRef = \@geodNetStatsDBFields;

my @geodNetStatsReportHeadings = (
        "Net Pre-fit Along-Track Mean",
        "Net Pre-fit Across-Track Mean",
        "Net Pre-fit Along-Track RMSE",
        "Net Pre-fit Across-Track RMSE",
        "Net Post-fit Along-Track Mean",
        "Net Post-fit Across-Track Mean",
        "Net Post-fit Along-Track RMSE",
        "Net Post-fit Across-Track RMSE"
        );
my $geodNetStatsReportHeadingsRef = \@geodNetStatsReportHeadings;

#---------------------------------------------------------------
# GEOMETRIC mode supporting data
#---------------------------------------------------------------
my @geomDBFields = qw(
        L0RP_IMAGE_FILE
        WORK_ORDER_ID
        TARGET_WRS_PATH
        TARGET_WRS_ROW
        OFF_NADIR_ANGLE
        DATE_ACQUIRED
        GCP_SOURCE
        PIXEL_SIZE
        NBR_OF_VLD_PNTS
        MEAN_LAT_OF_GCPS
        MEAN_LON_OF_GCPS
        MEAN_ALNG_TRK
        MEAN_ACRSS_TRK
        RMSE_ALNG_TRK
        RMSE_ACRSS_TRK
        STD_DEV_ALNG_TRK
        STD_DEV_ACRSS_TRK
        CORRELATION_COEFFICIENT
        );

my @geomNetStatsDBFields = qw(
        MEAN_ALNG_TRK
        MEAN_ACRSS_TRK
        RMSE_ALNG_TRK
        RMSE_ACRSS_TRK
        );
my $geomNetStatsDBFieldsRef = \@geomNetStatsDBFields;

my @geomNetStatsReportHeadings = (
        "Net Line Offset Mean",
        "Net Sample Offset Mean",
        "Net Line Offset RMSE",
        "Net Sample Offset RMSE"
        );
my $geomNetStatsReportHeadingsRef = \@geomNetStatsReportHeadings;

#--------------------------------------------------------------------
# Variable declarations
#--------------------------------------------------------------------
my $i;                                  # loop variable
my $status;                             # return status of sub-routines
my $initMsg;                            # program initialization message
my $errStr = "";                        # error string
my $dbTable;                            # the DB table to query
my $dbFieldsRef;                        # the DB fields to query
my $netStatsDBFieldsRef;                # the DB fields to query for net stats
my $netStatsReportHeadingsRef;          # report headings for net stats
my $sqlCommand;                         # SQL command string built in part
                                        # based on user selections
my $reportTitle;                        # report header title

my %resultsHofA = ();                   # results from DB query, keyed by DB
                                        # field name (Hash of Arrays)

my $acqDatesInput = 'NOT SELECTED';     # acquisition dates from user
my $gcpSourceInput = 'NOT SELECTED';    # GCP source from user 
my $rollAngleInput = 'NOT SELECTED';    # roll angle from user
my $wrsPathRowInput = 'NOT SELECTED';   # WRS Path and Row from user
my $toolModeInput = 'NOT SELECTED';     # Geometric or Geodetic from caller

my $acqDatesFlag = 1;                   # get acquisition dates?
my $gcpSourceFlag = 1;                  # get GCP source?
my $rollAngleFlag = 1;                  # get roll angle?
my $wrsPathRowFlag = 1;                 # get WRS Path and Row?
my $helpFlag = 0;                       # print a help/usage message?
my $noPromptFlag = 0;                   # disable input prompts?

my $readDateType;                       # string identifying the type of date
my $maxRollAngle = 16;                  # roll angle maximum degrees (+/-)
                                        # (15 degrees is maximum roll angle for
                                        # earth looking scenes)

# These are the more refined pieces of information extracted from the
# initial strings that came straight from the user.
my $acqDateStart;                       # start acquisition date
my $acqDateEnd;                         # end acquisition date
my $gcpSourceStr;                       # GCP source string: GLS or DOQ
my $rollAngle;                          # roll angle
my $wrsPathStart;                       # start WRS path
my $wrsPathEnd;                         # end WRS path
my $wrsRowStart;                        # start WRS row
my $wrsRowEnd;                          # end WRS row
my $toolMode;                           # use with GEOD/GEOM constants

#--------------------------------------------------------------------
# First step to get options from command line and set flags
#--------------------------------------------------------------------

$SIG{INT} = \&sigHdlr;

# Configure GetOptions to use the single-character syntax
Getopt::Long::Configure ('bundling');

GetOptions(
        'a:s'=>\$acqDatesInput,
        'g:s'=>\$gcpSourceInput,
        'r:s'=>\$rollAngleInput,
        'w:s'=>\$wrsPathRowInput,
        'v:s'=>\$toolModeInput,
        'h'=>\$helpFlag,
        'd'=>\$noPromptFlag
        );

# Unset the various processing flags, for those not indicated by user
$acqDatesFlag = 0 if ($acqDatesInput eq 'NOT SELECTED');
$gcpSourceFlag = 0 if ($gcpSourceInput eq 'NOT SELECTED');
$rollAngleFlag = 0 if ($rollAngleInput eq 'NOT SELECTED');
$wrsPathRowFlag = 0 if ($wrsPathRowInput eq 'NOT SELECTED');

#--------------------------------------------------------------------
# Gather the user's input data for options selected and
# setup program values associated with each option
#--------------------------------------------------------------------

# Set up some program variables to help us efficiently share code for
# either mode
$toolModeInput = uc($toolModeInput);

# Somewhat hidden from the user is the fact that the same code is used for
# both the Geometric and Geodetic Assessment tools. This is done through
# different aliases into the same script. This script needs to know which
# alias/mode was indicated.
if ($toolModeInput =~ /^GEOD/)
{
    $toolMode = GEOD;
    $dbTable = 'geodetic_characterization';
    $dbFieldsRef = \@geodDBFields;
    $netStatsDBFieldsRef = $geodNetStatsDBFieldsRef;
    $netStatsReportHeadingsRef = $geodNetStatsReportHeadingsRef;
    $initMsg = "Geodetic post-processing assessment was started with the ";
    $initMsg .= "following constraints:\n";
    $reportTitle = "Geodetic Assessment Report";
}
else
{
    if ($toolModeInput =~ /^GEOM/)
    {
        $toolMode = GEOM;
        $dbTable = 'geometric_accuracy';
        $dbFieldsRef = \@geomDBFields;
        $netStatsDBFieldsRef = $geomNetStatsDBFieldsRef;
        $netStatsReportHeadingsRef = $geomNetStatsReportHeadingsRef;
        $initMsg = "Geometric post-processing assessment was started with the ";
        $initMsg .= "following constraints:\n";
        $reportTitle = "Geometric Assessment Report";
    }
    else
    {
        &usageGeneric();
        exit FAILURE;
    }
}

# Give the user help if they requested and then quit
if ($helpFlag)
{
    #display usage and quit
    &usageGeodetic() if ($toolMode eq GEOD);
    &usageGeometric() if ($toolMode eq GEOM);
    exit SUCCESS;
}

#--------------------------------------------------------------------
# Initialize the IAS geometric/geodetic assessment application
#--------------------------------------------------------------------

our ($dbh, $errorLog) = &assessGeoInit($envVarsListRef);

#--------------------------------------------------------------------
# Print the report header
#--------------------------------------------------------------------
print_report_header($reportTitle);

#--------------------------------------------------------------------
# Build the program initialization message as options are read in
#--------------------------------------------------------------------
# Read the input value(s) for each query constraint the user provided
if ($acqDatesFlag)
{
    # readDateType is only needed with multiple types. Keep it for 
    # consistency with the x2x tool and future feature growth.
    $readDateType = 'Acquisition';
    ($acqDateStart, $acqDateEnd, $status) = &read_dates($errorLog,
            $acqDatesInput, $readDateType, $noPromptFlag);
    if ($status != SUCCESS)
    {
        # this would only occur due to prompting disabled and missing or
        # invalid input for this option that was provided.
        # error message already generated at lower level
        ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
        $errorLog->close_log();
        exit FAILURE;
    }
    $initMsg .= "  Acquisition Date Range: $acqDateStart to $acqDateEnd\n";
}

if ($gcpSourceFlag)
{
    ($gcpSourceStr, $status) = &readGcpSource($errorLog, $gcpSourceInput,
            $noPromptFlag);
    if ($status != SUCCESS)
    {
        # this would only occur due to prompting disabled and missing or
        # invalid input for this option that was indicated.
        # error message already generated at lower level
        ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
        $errorLog->close_log();
        exit FAILURE;
    }
    $initMsg .= "  GCP Source: $gcpSourceStr\n";
}

if ($rollAngleFlag)
{
    ($rollAngle, $status) = &readRollAngle($errorLog, $rollAngleInput,
            $maxRollAngle, $noPromptFlag);
    if ($status != SUCCESS)
    {
        # this would only occur due to prompting disabled and missing or
        # invalid input for this option that was indicated.
        # error message already generated at lower level
        ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
        $errorLog->close_log();
        exit FAILURE;
    }
    $initMsg .= "  Off-Nadir Angle Range: [-$rollAngle, $rollAngle] ";
    $initMsg .= "(degrees)\n";
}

if ($wrsPathRowFlag)
{
    ($wrsPathStart, $wrsPathEnd, $wrsRowStart, $wrsRowEnd, $status)
        = &read_wrs_path_row($errorLog, $wrsPathRowInput, $noPromptFlag);
    if ($status != SUCCESS)
    {
        # this would only occur due to prompting disabled and missing or
        # invalid input for this option that was provided.
        # error message already generated at lower level
        ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
        $errorLog->close_log();
        exit FAILURE;
    }
    $initMsg .= "  WRS Path Range: [$wrsPathStart, $wrsPathEnd]\n";
    $initMsg .= "  WRS Row Range: [$wrsRowStart, $wrsRowEnd]\n";
}

if (!$acqDatesFlag and !$gcpSourceFlag and !$rollAngleFlag and !$wrsPathRowFlag)
{
    $initMsg .= "  No user input constraints\n";
}

print "$initMsg\n";

#--------------------------------------------------------------------
# Query for the records from the database, with user constraints
#--------------------------------------------------------------------

# Form Oracle SQL command based on the user selections
my %sqlOptions = (
        acqDatesFlag    => $acqDatesFlag,
        gcpSourceFlag   => $gcpSourceFlag,
        rollAngleFlag   => $rollAngleFlag,
        wrsPathRowFlag  => $wrsPathRowFlag,
        acqDateRange    => [ $acqDateStart, $acqDateEnd ],
        gcpSource       => $gcpSourceStr,
        rollAngle       => $rollAngle,
        wrsPathRange    => [ $wrsPathStart, $wrsPathEnd ],
        wrsRowRange     => [ $wrsRowStart, $wrsRowEnd ]
        );
$sqlCommand = &buildSql($dbFieldsRef, $dbTable, \%sqlOptions);

# retrieve the requested records
$status = &retrieveRecords($dbh, $sqlCommand, \$errStr, \%resultsHofA);
if ($status != SUCCESS)
{
    ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
    $errorLog->fatal_msg(__FILE__, __LINE__, $errStr);
    $errorLog->close_log();
    exit FAILURE;
}

# close the database connection since we don't need anything more from it;
# should have only done inquiries of the DB, so disconnect with a rollback
$status = ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
if (!$status)
{
    $errorLog->fatal_msg(__FILE__, __LINE__, $ias_db_error);
    $errorLog->close_log();
    exit FAILURE;
}

#--------------------------------------------------------------------
# Print a report in CSV (comma separated values) format, containing first
# the set of records returned by the database query, built off the user's
# input constraints, then listing the net summary statistics across all
# the records returned by the query.
#--------------------------------------------------------------------

# Determine the number of records returned by the database query. Use a
# column from the query that should always have a value populated in it
# for every record that was entered to the DB.
my $arrayRef = $resultsHofA{WORK_ORDER_ID};
if (not defined($arrayRef))
{
    # Unexpected found either no rows for the query in the database, or
    # the rows returned didn't have the work order IDs set. Either way,
    # this would be an invalid situation to continue this script under.
    $errorLog->fatal_msg(__FILE__, __LINE__, 'No records found from query.');
    $errorLog->close_log();
    exit FAILURE;
}
my $numRecords = @$arrayRef;

#--------------------------------------------------------------------
# Print the database query field names as the column headers
#--------------------------------------------------------------------

# Print the report header containing the list of DB fields queried
my $lastArrayIndex = @$dbFieldsRef-1;
for ($i = 0; $i < $lastArrayIndex; $i++)
{
    print "$$dbFieldsRef[$i],";
}
# no comma after last field name; print the newline character
print "$$dbFieldsRef[$lastArrayIndex]\n";

#--------------------------------------------------------------------
# Print the database query results to the report
#--------------------------------------------------------------------

# Print the report contents
for ($i = 0; $i < $numRecords; $i++)
{
    for (my $j = 0; $j < $lastArrayIndex; $j++)
    {
        if (not defined($resultsHofA{$$dbFieldsRef[$j]}[$i]))
        {
            print ',';
            next;
        }
        print "$resultsHofA{$$dbFieldsRef[$j]}[$i],";
    }
    # no comma after last field value; print the newline character
    if (not defined($resultsHofA{$$dbFieldsRef[$lastArrayIndex]}[$i]))
    {
        print "\n";
        next;
    }
    print "$resultsHofA{$$dbFieldsRef[$lastArrayIndex]}[$i]\n";
}
print "\n";   # add some space after the DB report, before net summary stats

#--------------------------------------------------------------------
# Print the summary statistics column headers
#--------------------------------------------------------------------

# Print the summary results header containing the list of DB fields queried
$lastArrayIndex = @$netStatsReportHeadingsRef-1;
for ($i = 0; $i < $lastArrayIndex; $i++)
{
    print "$$netStatsReportHeadingsRef[$i],";
}
# no comma after last field name; print the newline character
print "$$netStatsReportHeadingsRef[$lastArrayIndex]\n";

#--------------------------------------------------------------------
# Print the summary statistics values
#--------------------------------------------------------------------

# Compute and print the summary results, checking each for error (value of 0)
my ($netStat, $earlyExitIndex);
$earlyExitIndex = -1;
# It works at this time to check for the words "MEAN" and "RMSE" in the DB
# field names and then do the correlating net calculation across all results
# for that field. However, may need extra attention if this code is expanded
# to support other assessment queries, depending on what net calculations need
# to be done on the DB query results and what those DB fields are named.
for ($i = 0; $i <= $lastArrayIndex; $i++)
{
    if ($$netStatsDBFieldsRef[$i] =~ /MEAN/)
    {
        ($netStat, $status) = &mean($errorLog,
                $resultsHofA{$$netStatsDBFieldsRef[$i]});
    }
    elsif ($$netStatsDBFieldsRef[$i] =~ /RMSE/)
    {
        ($netStat, $status) = &rmse($errorLog,
                $resultsHofA{$$netStatsDBFieldsRef[$i]});
    }
    # error check
    if ($status != SUCCESS)
    {
        # be sure we end the line of the reporting
        print "\n";
        $earlyExitIndex = $i;
        last;
    }
    # no error, so print calculated stat out to report
    if ($i < $lastArrayIndex)
    {
        print "$netStat,";
    }
    else
    {
        # On last loop, so print the newline character now
        print "$netStat\n";
    }
}
# if exited summary statistics reporting early; display field that had the error
if ($earlyExitIndex >= 0)
{
    # one of the stats calculations errored, so generate error
    $errStr = "Calculation of net statistics for ";
    $errStr .= "$$netStatsDBFieldsRef[$earlyExitIndex] failed\n";
    $errorLog->fatal_msg(__FILE__, __LINE__, $errStr);
    $errorLog->close_log();
    exit FAILURE;
}

# done
$errorLog->close_log();
exit SUCCESS;

###############################################################################
# Name: sigHdlr
#
# Purpose: Handle signals
###############################################################################
sub sigHdlr
{
    ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
    $errorLog->close_log();
    print "\nProgram interrupted; halting\n";
    exit FAILURE;
}

###############################################################################
# Name: usageOptions
#
# Purpose: Display command line options program accepts - use in conjunction
#          with tool mode usage subroutine.
###############################################################################
sub usageOptions
{
    print << 'USAGE';

Optional Parameters:
a - date range of image acquisition (dd-mmm-yyyy,dd-mmm-yyyy)
g - GCP source ([G]LS or [D]OQ)
r - off-nadir angle (absolute maximum offset from nadir in degrees)
w - geographic location of image extent (WRS Path/Row)
    (pathMin-pathMax,rowMin-rowMax) (max path range 1-233, max row range 1-248)
d - disable interactive prompting for missing or invalid parameter values
h - display this usage message

USAGE
}

###############################################################################
# Name: usageGeneric
#
# Purpose: Display command line options and general description of program
###############################################################################
sub usageGeneric
{
    print "ias_assess_geo.pl [-v -a -g -r -w -d -h]\n\n";
    print "v - variant (GEODETIC, GEOMETRIC -- required parameter)\n";

    &usageOptions();

    print "The GEODETIC variant helps analyze geodetic accuracy assessment ";
    print "accuracy.\n";
    print "The GEOMETRIC variant helps analyze geometric accuracy assessment ";
    print "accuracy.\n";
}

###############################################################################
# Name: usageGeodetic
#
# Purpose: Display command line options and general description of program
###############################################################################
sub usageGeodetic
{
    print "ias_assess_geo.pl [-a -g -r -w -d -h]\n";

    &usageOptions();

    print "ias_assess_geo.pl helps with the analysis of geodetic accuracy.\n";
}

###############################################################################
# Name: usageGeometric
#
# Purpose: Display command line options and general description of program
###############################################################################
sub usageGeometric
{
    print "ias_assess_geo.pl [-a -g -r -w -d -h]\n";

    &usageOptions();

    print "ias_assess_geo.pl helps with the analysis of geometric accuracy.\n";
}

###############################################################################
# Name: assessGeoInit
#
# Description: Initializes the IAS geometric/geodetic assessment application
#
# Arguments:
#   Array reference to the list of environment variables
#
# Returns:
#   An array of values:
#       handle to the database connection
#       reference to the error logging object
#   If an error occurs, the routine terminates the application with an error
###############################################################################
sub assessGeoInit
{
    my ($varsListRef) = @_;

    # Initialize the ias_processing_system module.
    if(ias_init_system() == ERROR)
    {
        print STDERR "ias_assess_geo.pl: $ias_error_msg\n";
        exit FAILURE;
    }

    if (ias_validate_environment($varsListRef) > 0)
    {
        print STDERR "ias_assess_geo.pl: $ias_error_msg\n";
        exit FAILURE;
    }

    my $Logger = Logger->new("STDOUT");
    if (!$Logger)
    {
        print STDERR "ias_assess_geo.pl: ERROR: Failed to establish ";
        print STDERR "logger object.\n";
        exit FAILURE;
    }

    set_default_log($Logger);
    $Logger->set_module_name('ias_assess_geo.pl');

    # get the username and password for the database
    my ($userId, $password) = ias_db_GetDBConnectionInfo("L8_OLITIRS");

    # connect to the database
    my $dbHandle = ias_db_ConnectToDB($userId, $password);
    if (!$dbHandle)
    {
        $Logger->fatal_msg(__FILE__, __LINE__, $ias_db_error);
        $Logger->close_log();
        exit FAILURE;
    }

    # return the important items
    return $dbHandle, $Logger;
}

###############################################################################
# NAME: buildSql
#
# PURPOSE: Form an SQL command based on the user selections.
#
# INPUTS: All user selections, including which filters they selected,
# and what particular options they selected within those filters.
#
# RETURNS: SQL command string that's ready to be sent to the database.
###############################################################################
sub buildSql
{
    my ($dbFieldsRef, $dbTable, $sqlOpts) = @_;

    # Build the comma separated list of field names to query for
    my $commaSepFields = join(', ', @$dbFieldsRef);

    # Start building the SQL command
    my $sql= qq{ SELECT $commaSepFields
        FROM $dbTable, geometric_work_order_common
            WHERE $dbTable.geometric_wo_common_id =
            geometric_work_order_common.geometric_wo_common_id
            AND geometric_work_order_common.valid = 'Y'
    };

    #------------------------------------------------------------------
    # Build additional parts of the SQL command if user specified any
    # query constraints. Add what was specified on input.
    #------------------------------------------------------------------

    # Handle the acquisition date range option
    if ($sqlOpts->{acqDatesFlag})
    {
        $sql .= " AND date_acquired >= '" . $sqlOpts->{acqDateRange}[0];
        $sql .= "' AND date_acquired <= '" . $sqlOpts->{acqDateRange}[1];
        $sql .= "'";
    }

    # Handle the GCP source option
    if ($sqlOpts->{gcpSourceFlag})
    {
        $sql .= " AND gcp_source = '" . $sqlOpts->{gcpSource} . "'";
    }

    # Handle the roll angle option
    if ($sqlOpts->{rollAngleFlag})
    {
        $sql .= ' AND off_nadir_angle BETWEEN -' . $sqlOpts->{rollAngle};
        $sql .= ' AND ' . $sqlOpts->{rollAngle};
    }

    # Handle the WRS path and row option
    if ($sqlOpts->{wrsPathRowFlag})
    {
        # path
        if ($sqlOpts->{wrsPathRange}[0] > $sqlOpts->{wrsPathRange}[1])
        {
            # if they searched for a path range like "230-5"
            $sql .= ' AND (target_wrs_path BETWEEN '; 
            $sql .= $sqlOpts->{wrsPathRange}[0];
            $sql .= ' AND 233 OR target_wrs_path BETWEEN 1';
            $sql .= ' AND ' . $sqlOpts->{wrsPathRange}[1] . ')';
        }
        else
        {
            # normal low-high range
            $sql .= ' AND target_wrs_path BETWEEN ';
            $sql .= $sqlOpts->{wrsPathRange}[0];
            $sql .= ' AND ' . $sqlOpts->{wrsPathRange}[1];
        }

        # row
        if ($sqlOpts->{wrsRowRange}[0] > $sqlOpts->{wrsRowRange}[1])
        {
            # if they searched for a row range like "245-3"
            $sql .= ' AND (target_wrs_row BETWEEN ';
            $sql .= $sqlOpts->{wrsRowRange}[0];
            $sql .= ' AND 248 OR target_wrs_path BETWEEN 1';
            $sql .= ' AND ' . $sqlOpts->{wrsRowRange}[1] . ')';
        }
        else
        {
            # normal low-high range
            $sql .= ' AND target_wrs_row BETWEEN ';
            $sql .= $sqlOpts->{wrsRowRange}[0];
            $sql .= ' AND ' . $sqlOpts->{wrsRowRange}[1];
        }
    }

    # sort the results; mostly to ensure regression testing can be done
    # in a reliable manner
    $sql .= ' ORDER BY WORK_ORDER_ID';

    return $sql;
}

###############################################################################
# NAME: readGcpSource
#
# PURPOSE: Retrieve GCP source from the command line.  The function
# does some rudimentary checks on the GCP source entered.
#
# INPUTS: Logger object, user-entered GCP source, and flag indicating if
# interactive prompting is disabled.
#
# RETURNS: Valid user-selected GCP source value.
###############################################################################
sub readGcpSource
{
    my ($logger, $gcpSourceInput, $noPromptFlag) = @_;

    my $validFlag;              # validity flag
    my $gcpSourceChar;          # GCP source selection; single character
    my $gcpSourceStr;           # GCP source selection; expanded string

    while ()
    {
        # Initialize flag to "true"
        $validFlag = 1;

        # If the filter was selected but no value for it was selected,
        # then we have to prompt for it and read it the first time in here.
        # Otherwise we only prompt if it was entered incorrectly.
        if (not defined($gcpSourceInput) or $gcpSourceInput eq '')
        {
            # If the user disabled prompting with the command line option
            # then just throw an error about missing needed input and quit
            if ($noPromptFlag)
            {
                $logger->fatal_msg(__FILE__, __LINE__,
                        "Missing or invalid GCP source; prompting disabled");
                return (undef, ERROR);
            }

            # Let the user know what data is needed
            print 'GCP Source ([G]LS or [D]OQ): ';

            # Read the date the user typed
            $gcpSourceInput = <STDIN>;
        }
        chomp($gcpSourceInput);

        $gcpSourceChar = $gcpSourceInput;

        # Clear the string so it's recognized as needing to be filled in
        # should the current version prove to be syntactically incorrect.
        $gcpSourceInput = '';

        # Make the GCP source type all upper case
        $gcpSourceChar = uc($gcpSourceChar);

        if (($gcpSourceChar !~ /^D/) and ($gcpSourceChar !~ /^G/))
        {
            $validFlag = 0;
            print "GCP source $gcpSourceChar invalid; valid values are ";
            print "[G]LS or [D]OQ.\n";
        }

        if ($validFlag)
        {
            # Expand input to a full string, as found in the DB, if not already
            if ($gcpSourceChar =~ /^D/)
            {
                $gcpSourceStr = 'DOQ';
            }
            else # it was a 'G'; error checked above
            {
                $gcpSourceStr = 'GLS';
            }
            return ($gcpSourceStr, SUCCESS);
        }
    }

    # safety net to ensure we don't fall out of this routine without
    # a controlled return value
    return (undef, ERROR);
}

###############################################################################
# NAME: readRollAngle
#
# PURPOSE: Retrieve roll angle from the command line.  The function
# does some rudimentary checks on the roll angle entered.
#
# INPUTS: Logger object, user-entered roll angle, maximum allowed roll angle,
# flag indicating if interactive prompting is disabled.
#
# RETURNS: Valid user-selected roll angle value.
###############################################################################
sub readRollAngle
{
    my ($logger, $rollAngleInput, $max, $noPromptFlag) = @_;

    my $validFlag;              # validity flag
    my $rollAngle;              # roll angle selection

    while ()
    {
        # Initialize flag to "true".
        $validFlag = 1;

        # If the filter was selected but no value for it was selected,
        # then we have to prompt for it and read it the first time in here.
        # Otherwise we only prompt if it was entered incorrectly.
        if ($rollAngleInput eq '')
        {
            # If the user disabled prompting with the command line option
            # then just throw an error about missing needed input and quit
            if ($noPromptFlag)
            {
                $logger->fatal_msg(__FILE__, __LINE__,
                        "Missing or invalid roll angle; prompting disabled");
                return (0, ERROR);
            }

            # Let the user know what data is needed.
            print 'Off-Nadir Angle Absolute Max: ';

            # Read the date the user typed.
            $rollAngleInput = readline(*STDIN);
        }
        chomp($rollAngleInput);

        $rollAngle = $rollAngleInput;

        # Clear the string so it's recognized as needing to be filled in
        # should the current version prove to be syntactically incorrect.
        $rollAngleInput = '';

        # Make sure it's all positive floats or integers in a format that SQL
        # will take.  Oracle's SQLPlus seems to take the format #.# where any
        # of the 3 parts are optional (eg: 100, 100.2, 100., .42) ... but of
        # course at least 1 digit is needed.
        $_ = $rollAngle;
        if (/^(\d*\.\d+|\d+\.\d*|\d+)$/)
        {
            # Make sure it falls in a valid range.
            if ($rollAngle <= $max)
            {
                # It's a valid element.
                $validFlag = 1;
            }
            else
            {
                $validFlag = 0;
                print "Roll angle $rollAngle invalid; ";
                print "valid range is between 0 and $max.\n";
            }
        }
        else
        {
            $validFlag = 0;
            print "Roll angle $rollAngle invalid; ";
            print "should be a non-negative integer or float.\n";
        }

        if ($validFlag)
        {
            return ($rollAngle, SUCCESS);
        }
    }

    # safety net to ensure we don't fall out of this routine without
    # a controlled return value
    return (0, ERROR);
}

###############################################################################
# NAME: retrieveRecords
#
# PURPOSE: retrieve and return selected geodetic or geometric accuracy
# assessment records from the database
#
# INPUTS: Pre-formatted SQL SELECT string to send to the database
#
# RETURNS: a hash reference to arrays of values per field, through a parameter
###############################################################################
sub retrieveRecords
{
    my ($dbh, $sql, $errStrRef, $resultsHofARef) = @_;

    # prepare SQL query
    my $sth = $dbh->prepare($sql);
    if (!$sth)
    {
        $$errStrRef = "Build of query returned an error: $DBI::errstr";
        return ERROR;
    }

    # execute SQL query
    if (!$sth->execute())
    {
        $$errStrRef = "Query returned an error: $DBI::errstr";
        return ERROR;
    }

    # fetch and save the results of the query
    while (my $dbRowRef = $sth->fetchrow_hashref())
    {
        foreach (keys %$dbRowRef)
        {
            push @{$resultsHofARef->{$_}}, $dbRowRef->{$_};
        }
    }

    $sth->finish();

    return SUCCESS;
}
