package ias_assess_geo_and_x2x;

###############################################################################
#
# This package contains supporting routines for the ias_assess_geo.pl and 
# ias_assess_x2x.pl utilities
#
###############################################################################

use strict;
use warnings;
use vars qw(@ISA @EXPORT);
use Exporter;
use Logger;
use POSIX;
@ISA = qw(Exporter);
@EXPORT = qw(
        read_wrs_path_row
        read_dates
        mean
        rmse
        print_report_header
        );

use ias_const;              # for SUCCESS, ERROR, FAILURE, etc.

###############################################################################
# NAME: read_dates
#
# PURPOSE: Wrapper for read_date_element subroutine, packaging start and end
# dates together.
#
# INPUTS: Logger object, user-entered date string, type of date (acquisition 
# date or processing date), and flag indicating if interactive prompting is
# disabled.
#
# RETURNS: A pair of date strings formatted to be ready to be used as
# part of a SQL command, error status.
###############################################################################
sub read_dates
{
    my ($logger, $date_input, $date_type_str, $no_prompt_flag) = @_;

    my @dates_input;             # holds begin and end date
    my $tmp_str;                 # temporary string for various messages
    my $acq_date_start;
    my $acq_date_end;
    my $status;

    # The 2 dates should be separated by a ","
    @dates_input = split /,/, $date_input;

    $tmp_str = 'Start Date For Image ' . $date_type_str;
    $tmp_str .= ' Date Range (dd-mmm-yyyy): ';

    ($acq_date_start, $status) = &read_date_element($logger, $tmp_str, 
            $dates_input[0], $no_prompt_flag);
    if ($status != SUCCESS)
    {
        return (undef, undef, ERROR);
    }

    $tmp_str = 'End Date For Image ' . $date_type_str;
    $tmp_str .= ' Date Range (dd-mmm-yyyy): ';

    ($acq_date_end, $status) = &read_date_element($logger, $tmp_str, 
            $dates_input[1], $no_prompt_flag);
    if ($status != SUCCESS)
    {
        return (undef, undef, ERROR);
    }

    return ($acq_date_start, $acq_date_end, SUCCESS);
}

###############################################################################
# NAME: read_date_element
#
# PURPOSE: Retrieve date from command line.  The date must be in dd-mmm-yyyy
# format, where mmm is made of alphabetic characters representing a month,
# and dd and yyyy are made of digits representing the day of the month and
# the year.  This is a SQL-ready format. The function does some rudimentary
# checks on the date format.
#
# INPUTS: Logger object, prompt to display, user-entered date string, and
# flag indicating if interactive prompting is disabled.
#
# RETURNS: A date string formatted to be ready to be used as part of a SQL
# command, error status.
###############################################################################
sub read_date_element
{
    my ($logger, $prompt_str, $date_input, $no_prompt_flag) = @_;

    # I'm not going to check leap years and such.  We should use a Perl
    # date/time package before coding that level of detail.
    my @months
    = ('JAN','FEB','MAR','APR','MAY','JUN','JUL','AUG','SEP','OCT','NOV','DEC');
    my %days_in_months = (
            JAN   => 31,
            FEB   => 29,
            MAR   => 31,
            APR   => 30,
            MAY   => 31,
            JUN   => 30,
            JUL   => 31,
            AUG   => 31,
            SEP   => 30,
            OCT   => 31,
            NOV   => 30,
            DEC   => 31
    );
    my $date_valid_flag;        # validity flag

    while ()
    {
        # Initialize date flag to "true"
        $date_valid_flag = 1;

        # If the filter was selected but no value for it was selected,
        # then we have to prompt for it and read it the first time in here.
        # Otherwise we only prompt if it was entered incorrectly.
        if (not defined($date_input) or $date_input eq '')
        {
            # If the user disabled prompting with the command line option
            # then just throw an error about missing needed input and quit
            if ($no_prompt_flag)
            {
                $logger->fatal_msg(__FILE__, __LINE__, 
                        "Missing or invalid date range; prompting disabled");
                return (undef, ERROR);
            }

            # Let the user know what data is needed
            print "$prompt_str";

            # Read the date the user typed
            $date_input = <STDIN>;
        }
        chomp($date_input);

        # Extract the day, month, and year from the entered string.
        # This restricts the format to digits dash letters dash digits,
        # with no extra stuff at the beginning or end. Additional checks
        # appear below.
        my ($day, $month, $year) = ($date_input =~ /^(\d+)-(\w+)-(\d+)$/);

        # Clear the date string so it's recognized as needing to be filled
        # in should the current version prove to be syntactically incorrect.
        $date_input = '';

        if (not defined($day) or not defined($month) or not defined($year))
        {
            print 'Could not find day, month, and year. Use format ';
            print "dd-mmm-yyyy.\n";
            $date_valid_flag = 0;
            next;
        }

        # Make the month all upper case
        $month = uc($month);

        # Check that the month is on the list of valid months. Also check
        # that the day of month is possible for that month.
        if (defined($days_in_months{$month}))
        {
            # Month is valid, so check the day
            if (($day < 1) or ($day > $days_in_months{$month}))
            {
                print "Day $day for month $month invalid; valid range is ";
                print "between 1 and $days_in_months{$month}.\n";

                # If the day isn't valid, the overall date can't be valid
                $date_valid_flag = 0;
            }
        }
        else
        {
            # If the month isn't found, the overall date can't be valid
            print "Month $month invalid;\n    valid values are @months\n";
            $date_valid_flag = 0;
        }

        # I don't think I'll have a range check on the year, since even
        # very small or large years are "valid" for range checks.

        # If it's valid move on
        if ($date_valid_flag)
        {
            return ("$day-$month-$year", SUCCESS);
        }
    }

    # safety net to ensure we don't fall out of this routine without
    # a controlled return value
    return (undef, ERROR);
}

###############################################################################
# NAME: read_wrs_path_row
#
# PURPOSE: Wrapper for read_wrs_element subroutine, packaging path and row
# together.
#
# INPUTS: Logger object, user-entered WRS path and row selection (could be 
# empty) The format is ###-###,###-### where the 4 numerical values are the
# path start, path end, row start, and row end values, respectively; and flag
# indicating if interactive prompting is disabled.
#
# RETURNS: End values for the user-selected WRS path and row ranges, error 
# status.
###############################################################################
sub read_wrs_path_row
{
    my ($logger, $wrs_input, $no_prompt_flag) = @_;

    my @wrs_input;           # holds begin and end paths and rows
    my $tmp_str;             # temporary string for various messages
    my $wrs_path_start;
    my $wrs_path_end;
    my $wrs_row_start;
    my $wrs_row_end;
    my $status;

    # The path and row ranges should be separated by a ","
    @wrs_input = split /,/, $wrs_input;

    $tmp_str = 'WRS Path Range (eg: 34-148): ';
    ($wrs_path_start, $wrs_path_end, $status) = &read_wrs_element($logger,
            $tmp_str, $wrs_input[0], 1, 233, $no_prompt_flag);
    if ($status != SUCCESS)
    {
        return (undef, undef, undef, undef, ERROR);
    }

    $tmp_str = 'WRS Row Range (eg: 34-148): ';
    ($wrs_row_start, $wrs_row_end, $status) = &read_wrs_element($logger,
            $tmp_str, $wrs_input[1], 1, 248, $no_prompt_flag);
    if ($status != SUCCESS)
    {
        return (undef, undef, undef, undef, ERROR);
    }

    return ($wrs_path_start, $wrs_path_end, $wrs_row_start, $wrs_row_end, 
            SUCCESS);
}

###############################################################################
# NAME: read_wrs_element
#
# PURPOSE: Retrieve WRS element (path or row) from the command line.  The
# WRS element must be in all digit format.  The function does some
# rudimentary checks on the WRS format.
#
# INPUTS: Logger object, prompt to display, user-entered WRS element range 
# ###-###, and valid range minimum and maximum for the particular WRS element
# type (WRS path or row); and flag indicating if interactive prompting is
# disabled.
#
# RETURNS: Valid user-selected WRS path or row user range values, error status.
###############################################################################
sub read_wrs_element
{
    my ($logger, $prompt_str, $range_input, $min, $max, $no_prompt_flag) = @_;

    my $element;                  # single element in list
    my $element_valid_flag;       # validity flag
    my $range_valid_flag;         # validity flag

    while ()
    {
        # Initialize flag to "true"
        $range_valid_flag = 1;

        # If the filter was selected but no value for it was selected,
        # then we have to prompt for it and read it the first time in here.
        # Otherwise we only prompt if it was entered incorrectly.
        if (not defined($range_input) or $range_input eq '')
        {
            # If the user disabled prompting with the command line option
            # then just throw an error about missing needed input and quit
            if ($no_prompt_flag)
            {
                $logger->fatal_msg(__FILE__, __LINE__, 
                        "Missing or invalid WRS path or row range; " .
                        "prompting disabled");
                return (0, 0, ERROR);
            }

            # Let the user know what data is needed
            print "$prompt_str";

            # Read the date the user typed
            $range_input = <STDIN>;
        }
        chomp($range_input);

        # Extract the start and end of the entered range string.  This
        # restricts the format to digits dash digits, with no extra
        # stuff at the beginning or end.  Additional checks appear below.
        my ($start, $end) = ($range_input =~ /^(\d+)-(\d+)$/);

        # Clear the range string so it's recognized as needing to be filled
        # in should the current version prove to be syntactically incorrect.
        $range_input = '';

        if (not defined($start) or not defined($end))
        {
            print "Could not find start or end. Use format ###-###.\n";
            $range_valid_flag = 0;
            next;
        }

        foreach $element (($start, $end))
        {
            # Validate the element
            $element_valid_flag = 0;
            $_ = $element;
            # Make sure it's all positive integers
            if (/^\d+$/)
            {
                # Make sure it falls in a valid range
                if (($element >= $min) and ($element <= $max))
                {
                    # It's a valid element
                    $element_valid_flag = 1;
                }
                else
                {
                    print "Element $element invalid; valid range is between ";
                    print "$min and $max.\n";
                }
            }
            else
            {
                print "Element $element invalid; should be a ";
                print "positive integer.\n";
            }

            # If the element isn't valid, the overall range can't be valid
            if (!$element_valid_flag)
            {
                $range_valid_flag = 0;
            }
        }

        # If it's valid move on
        if ($range_valid_flag)
        {
            return ($start, $end, SUCCESS);
        }
    }

    # safety net to ensure we don't fall out of this routine without
    # a controlled return value
    return (0, 0, ERROR);
}

###############################################################################
# NAME: mean 
#
# PURPOSE: Compute the mean of the values in an array.
#
# INPUTS: Logger object, an array reference for a list of numbers
#
# RETURNS: The mean of the array, error status
###############################################################################
sub mean 
{
    my ($logger, $listRef) = @_;

    my @list = @$listRef;
    my $result = 0;
    my $valid_count = 0;

    foreach (@list) 
    { 
        if (defined($_))
        {
            $result += $_;
            $valid_count++;
        }
    }

    if ($valid_count > 0)
    {
        return ($result / $valid_count, SUCCESS);
    }
    else
    {
        $logger->fatal_msg(__FILE__, __LINE__, 
                "No input passed to mean routine");
        return (0, ERROR); 
    }
}

###############################################################################
# NAME: rmse 
#
# PURPOSE: Compute the RMSE (root mean square error) of the values in an 
# array. The implementation is similar to ALIAS xxx_RMSE.
#
# INPUTS: Logger object, an array reference for a list of numbers
#
# RETURNS: The RMSE of the array, error status
###############################################################################
sub rmse 
{
    my ($logger, $listRef) = @_;

    my @list = @$listRef;
    my $sum_sqr = 0;
    my $valid_count = 0;

    foreach (@list) 
    { 
        if (defined($_))
        {
            $sum_sqr += $_ * $_;
            $valid_count++;
        }
    }

    if ($valid_count > 0)
    {
        return (sqrt($sum_sqr / $valid_count), SUCCESS);
    }
    else
    {
        $logger->fatal_msg(__FILE__, __LINE__, 
                "No input passed to rmse routine");
        return (0, ERROR); 
    }
}

##############################################################################
# NAME: print_report_header
#
# PURPOSE: Print a standard report header.
#
# INPUTS: Report name
##############################################################################
sub print_report_header
{
    my ($report_name) = @_;

    my $date_str = strftime("%a. %b. %d, %Y", localtime);
    my $time_str = strftime("%H:%M", localtime);

    my $rpt_padding = (70 - length($report_name)) / 2;

    print "$date_str                                   Time: $time_str\n";
    print "                           EROS Data Center\n";
    print " " x $rpt_padding . "$report_name\n";
    print "-" x 75 . "\n";
    if (defined $ENV{"IAS_DB_L8_OLITIRS"})
    {
        my @parts = split(/\//, $ENV{"IAS_DB_L8_OLITIRS"});
        print "Database schema:      $parts[0]\n";
    }
    print "\n";
}

1;
