#!/usr/bin/perl -w
############################################################################
#  NAME:   ias_assess_x2x.pl
#
#  PURPOSE: Produce a report that helps CalVal Analysts assess the
#           band or image registration quality over multiple scenes.
#
#  PARAMETERS: 
#           a - date range of image acquisition (dd-mmm-yyyy,dd-mmm-yyyy)
#           p - date range of processing date (dd-mmm-yyyy,dd-mmm-yyyy)
#           s - SCA list (Band registration only)
#           b - band list
#           t - acquisition type ([L]unar, [E]arth, e[N]gineering)
#           n - nadir ([N]adir, [O]ff-nadir)
#           w - geographic location of image extend (WRS Path/Row) (1-233,1-248)
#           h - display help
#           d - disable interactive prompting for missing or invalid 
#               parameter values
#           v - type of processing (B2B or I2I)
#
#  OUTPUT:  Report file
#
############################################################################

use strict;                 # enforce the rules.
use DBI;                    # include DBI package
use Getopt::Long;           # parse command line options
use Logger;                 # IAS logging module
use ias_processing_system;  # define ias_error_msg and ias_validate_environment
use ias_const;              # define SUCCESS and ERROR
use ias_db;                 # local package with DB subroutines
use ias_assess_geo_and_x2x; # shared routines with ias_assess_geo.pl

# declare some constants
use constant PROG => 'ias_assess_x2x.pl';

# Define list of environment variables needed by ias_db.pm that we should
# check up-front to bail out sooner if our environment isn't setup as required
my @envVarsList = qw(
        IAS_DB_L8_OLITIRS
        ORACLE_SID
        );
my $envVarsListRef = \@envVarsList;

#############################################################################
# Variable declarations
#############################################################################
my $dbh;                            # database handle
my $status;                         # subroutine return value

my $acquisition_date_range_str = 'NOT SELECTED'; # acquisition dates from user
my $processing_date_range_str = 'NOT SELECTED';  # processing dates from user
my $sca_list_str = 'NOT SELECTED';               # SCA list from user
my $band_list_str = 'NOT SELECTED';              # band list from user
my $type_of_acquisition_str = 'NOT SELECTED';    # acquisition type from user
my $nadir_str = 'NOT SELECTED';                  # nadir / off-nadir from user
my $wrs_path_row_str = 'NOT SELECTED';           # WRS Path and Row from user
my $tool_variant_str = 'NOT SELECTED';           # B2B or I2I from caller

my $acquisition_date_range_flag = 1; # get acquisition dates?
my $processing_date_range_flag = 1;  # get processing dates?
my $sca_list_flag = 1;               # get SCA list?
my $band_list_flag = 1;              # get band list?
my $type_of_acquisition_flag = 1;    # get acquisition type?
my $nadir_flag = 1;                  # get nadir type?
my $wrs_path_row_flag = 1;           # get WRS Path and Row?
my $help_flag = 0;                   # print a help/usage message?
my $no_prompt_flag = 0;              # disable input prompts?

my $b2b_flag = 0;                    # running as the B2B assessment tool?
my $i2i_flag = 0;                    # running as the I2I assessment tool?

# These are the more refined pieces of information extracted from the
# initial strings that came straight from the user.

my $start_acquisition_date;          # start acquisition date
my $end_acquisition_date;            # end acquisition date
my $start_processing_date;           # start processing date
my $end_processing_date;             # end processing date
my $sca_list;                        # list of SCA numbers
my $band_list;                       # list of band numbers
my $acquisition_type;                # acquisition type character
my $nadir;                           # nadir character
my $start_path;                      # start WRS path
my $end_path;                        # end WRS path
my $start_row;                       # start WRS row
my $end_row;                         # end WRS row

my $sql_command;                     # SQL command string built in part based
                                     # on user selections
my $user_name;                       # DB user name
my $password;                        # DB password

my $record_data_ref;                 # reference to record data returned from DB

# columns printed out
my @columns = qw(
    REFERENCE_BAND_ID  SEARCH_BAND_ID            SENSOR_CHIP_ASSEMBLY_ID
    TARGET_WRS_PATH    TARGET_WRS_ROW            COLLECTION_TYPE
    DATE_ACQUIRED      DATE_PROCESSED            GEOMETRIC_WO_COMMON_ID
    OFF_NADIR_ANGLE    NBR_OF_PNTS_CORR          NBR_OF_PNTS_SUCCESSFULLY_CORR
    NBR_OF_VLD_PNTS    T_DISTRIBUTION_THRESHOLD  STD_DEV_LINE
    STD_DEV_SAMPLE     MIN_LINE                  MIN_SAMPLE
    MAX_LINE           MAX_SAMPLE                MED_LINE
    MED_SAMPLE         RMSE_LINE                 RMSE_SAMPLE               
    MEAN_LINE          MEAN_SAMPLE);

# summary statistic arrays and net results
my @rmse_lines;
my @rmse_samples;
my @mean_lines;
my @mean_samples;
my $net_line_rmse;
my $net_sample_rmse;
my $net_line_mean;
my $net_sample_mean;

# Initialize the ias_processing_system module.
if(ias_init_system() == ERROR)
{
    print STDERR PROG . ": $ias_error_msg\n";
    exit FAILURE;
}

# Validate the required environment variables.
if (ias_validate_environment($envVarsListRef) > 0)
{
    print STDERR PROG . ": $ias_error_msg\n";
    exit FAILURE;
}

# Set up logger.
my $logger = Logger->new('STDOUT');
if (!$logger)
{
    print STDERR PROG . ": ERROR: Failed to establish logger object for" .
                       " STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name(PROG);
set_default_log($logger);

# Configure GetOptions to use the single-character syntax.
Getopt::Long::Configure ('bundling');

GetOptions('a:s'=>\$acquisition_date_range_str,
           'p:s'=>\$processing_date_range_str,
           's:s'=>\$sca_list_str,
           'b:s'=>\$band_list_str,
           't:s'=>\$type_of_acquisition_str,
           'n:s'=>\$nadir_str,
           'w:s'=>\$wrs_path_row_str,
           'v:s'=>\$tool_variant_str,
           'd'=>\$no_prompt_flag,
           'h'=>\$help_flag);

# Check to see if extra, unused command line parameters were passed in.
if ($ARGV[0])
{
    print 'Unknown option: ';
    foreach (@ARGV) 
    {
        print "$_ ";
    }
    print "\n";
}

# Set the various processing flags based on the input values.
if ($acquisition_date_range_str eq 'NOT SELECTED')
{
    $acquisition_date_range_flag = 0;
}

if ($processing_date_range_str eq 'NOT SELECTED')
{
    $processing_date_range_flag = 0;
}

if ($sca_list_str eq 'NOT SELECTED')
{
    $sca_list_flag = 0;
}

if ($band_list_str eq 'NOT SELECTED')
{
    $band_list_flag = 0;
}

if ($type_of_acquisition_str eq 'NOT SELECTED')
{
    $type_of_acquisition_flag = 0;
}

if ($nadir_str eq 'NOT SELECTED')
{
    $nadir_flag = 0;
}

if ($wrs_path_row_str eq 'NOT SELECTED')
{
    $wrs_path_row_flag = 0;
}

# Make the tool variant string all upper case.
$tool_variant_str =~ tr/a-z/A-Z/;

if (($tool_variant_str eq 'B2B') || ($tool_variant_str eq 'B'))
{
    print_report_header("Band Registration Accuracy Report");
    print "Assessing band registration accuracy:\n";
    $b2b_flag = 1;
}
elsif (($tool_variant_str eq 'I2I') || ($tool_variant_str eq 'I'))
{
    print_report_header("Image Registration Accuracy Report");
    print "Assessing image registration accuracy:\n";
    $i2i_flag = 1;
}
else
{
    # Default usage in case user did not specify the variant.
    # This way they can receive help on how to choose one.
    x2x_usage();
    $logger->close_log();
    exit SUCCESS;
}

if ($help_flag)
{
    # Display usage and quit.
    if ($b2b_flag)
    {
        b2b_usage();
        $logger->close_log();
    }
    else # $i2i_flag
    {
        i2i_usage();
        $logger->close_log();
    }
    exit SUCCESS;
}

if ($acquisition_date_range_flag)
{
    ($start_acquisition_date, $end_acquisition_date, $status) =
        read_dates($logger, $acquisition_date_range_str, 'acquisition', 
            $no_prompt_flag);
    if ($status != SUCCESS)
    {
        $logger->close_log();
        exit FAILURE;
    }

    print "  Acquisition date range: $start_acquisition_date to " 
        . "$end_acquisition_date\n";
}

if ($processing_date_range_flag)
{
    ($start_processing_date, $end_processing_date, $status) =
        read_dates($logger, $processing_date_range_str, 'processing', 
            $no_prompt_flag);
    if ($status != SUCCESS)
    {
        $logger->close_log();
        exit FAILURE;
    }

    print "  Processing date range: $start_processing_date to "
        . "$end_processing_date\n";
}

if ($sca_list_flag)
{
    # Only use SCA searching for B2B operation.
    if ($b2b_flag)
    {
        ($sca_list, $status) = read_number_list($logger, 
            'list of SCAs (eg: 1,3,4): ', $sca_list_str, 0, 14, 
            $no_prompt_flag);
        if ($status != SUCCESS)
        {
            $logger->fatal_msg(__FILE__, __LINE__,
                'When prompting is disabled and the SCA flag is ' .
                'present, the user must provide an integer or comma-' .
                'separated list of integers following the SCA flag.');
            $logger->close_log();
            exit FAILURE;
        }

        print "  SCA list: $sca_list\n";
    }
    else # Warn the user the SCA parameters will be ignored in I2I mode.
    {
        print "The SCA option was selected but will be ignored in I2I mode.\n";
    }
}

if ($band_list_flag)
{
    ($band_list, $status) = read_number_list($logger, 
        'list of bands (eg: 1,3,4): ', $band_list_str, 1, 11, $no_prompt_flag);
    if ($status != SUCCESS)
    {
        $logger->fatal_msg(__FILE__, __LINE__,
            'When prompting is disabled and the band flag is ' .
            'present, the user must provide an integer or comma-' .
            'separated list of integers following the band flag.');
        $logger->close_log();
        exit FAILURE;
    }

    print "  Band list: $band_list\n";
}

if ($type_of_acquisition_flag)
{
    ($acquisition_type, $status) = read_acquisition_type($logger, 
        $type_of_acquisition_str, $no_prompt_flag);
    if ($status != SUCCESS)
    {
        $logger->close_log();
        exit FAILURE;
    }

    print "  Acquisition type: $acquisition_type\n";
}

if ($nadir_flag)
{
    ($nadir, $status) = read_nadir($logger, 
        $nadir_str, $no_prompt_flag);
    if ($status != SUCCESS)
    {
        $logger->close_log();
        exit FAILURE;
    }

    print "  Nadir type: $nadir\n";
}

if ($wrs_path_row_flag)
{
    ($start_path, $end_path, $start_row, $end_row, $status) =
        read_wrs_path_row($logger, $wrs_path_row_str, $no_prompt_flag);
    if ($status != SUCCESS)
    {
        $logger->close_log();
        exit FAILURE;
    }

    print "  Path range: $start_path-$end_path, "
          . "Row range: $start_row-$end_row\n";
}
print "\n";

# Form Oracle SQL command based on the user selections.
($sql_command) = build_sql($acquisition_date_range_flag,
    $processing_date_range_flag, $sca_list_flag, $band_list_flag,
    $type_of_acquisition_flag, $wrs_path_row_flag, $b2b_flag, $i2i_flag,
    $start_acquisition_date, $end_acquisition_date,
    $start_processing_date, $end_processing_date,
    $sca_list, $band_list, $acquisition_type,
    $start_path, $end_path, $start_row, $end_row, $nadir);

# Get the username and password for the database.
($user_name, $password) = ias_db_GetDBConnectionInfo('L8_OLITIRS');

# Connect to the database.
$dbh = ias_db_ConnectToDB($user_name, $password);
if (!$dbh)
{
    $logger->fatal_msg(__FILE__, __LINE__, $ias_db_error);
    $logger->fatal_msg(__FILE__, __LINE__, 'Consider checking your ' .
                                           '$IAS_DB_L8_OLITIRS setting.');
    $logger->close_log();
    exit FAILURE;
}

# Retrieve data.
($record_data_ref, $status) = retrieve_records($logger, $sql_command);
if ($status != SUCCESS)
{
    ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
    $logger->close_log();
    exit FAILURE;
}

# Close the database connection.
$status = ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_ROLLBACK);
if (!$status)
{
    $logger->fatal_msg(__FILE__, __LINE__, $ias_db_error);
    $logger->close_log();
    exit FAILURE;
}

my $i = 0;                          # loop variable
my $j = 0;                          # loop variable
# Print the report contents.
foreach my $record (@{$record_data_ref})
{
    # Print the report header.
    if ($i == 0)
    {
        foreach my $column (@columns)
        {
            # Print a comma unless this is the first column.
            if ($j)
            {
                print ',';
            }
            print $column;
            $j++;
        }
        print "\n";
    }

    $j = 0;

    # Print the report contents.
    foreach my $column (@columns)
    {
        # Print a comma unless this is the first column.
        if ($j)
        {
            print ',';
        }
        if (defined($record->{$column}))
        {
            print $record->{$column};
        }
        $j++;
    }
    print "\n";

    # Save these for the summary statistics.
    push @rmse_lines, $record->{RMSE_LINE};
    push @rmse_samples, $record->{RMSE_SAMPLE};
    push @mean_lines, $record->{MEAN_LINE};
    push @mean_samples, $record->{MEAN_SAMPLE};
    $i++;
}

if ($i == 0)
{
    print "No records matched search criteria\n";
}
else
{
    # Print the summary results header.
    print "NET_LINE_RMSE,NET_SAMPLE_RMSE,NET_LINE_MEAN,NET_SAMPLE_MEAN\n";

    # Compute the summary results.
    ($net_line_rmse, $status) = rmse($logger, \@rmse_lines);
    if ($status != SUCCESS)
    {
        $logger->close_log();
        exit FAILURE;
    }
    ($net_sample_rmse, $status) = rmse($logger, \@rmse_samples);
    if ($status != SUCCESS)
    {
        $logger->close_log();
        exit FAILURE;
    }
    ($net_line_mean, $status) = mean($logger, \@mean_lines); 
    if ($status != SUCCESS)
    {
        $logger->close_log();
        exit FAILURE;
    }
    ($net_sample_mean, $status) = mean($logger, \@mean_samples);
    if ($status != SUCCESS)
    {
        $logger->close_log();
        exit FAILURE;
    }

    # Print the summary results.
    print "$net_line_rmse,$net_sample_rmse,$net_line_mean,$net_sample_mean\n";
}

$logger->close_log();
exit SUCCESS;



###########################################
### subroutines defined here
###########################################

##############################################################################
#
# Name: b2b_usage
#
# Purpose: Display command line options and general description of program.
#
##############################################################################

sub b2b_usage
{
    print << 'USAGE';
$0;
ias_x2x_assess.pl [-a -p -s -b -t -n -w -d -h]

a - date range of image acquisition (dd-mmm-yyyy,dd-mmm-yyyy)
p - date range of image processing (dd-mmm-yyyy,dd-mmm-yyyy)
s - SCA list (eg: 1,3,4)
b - band list (eg: 1,3,4)
t - acquisition type ([L]unar, [E]arth, e[N]gineering)
n - nadir ([N]adir, [O]ff-nadir)
w - geographic location of image extent by WRS Path/Row (1-233,1-248)
d - disable interactive prompting for missing or invalid parameter values
h - display this usage message

ias_b2b_assess.pl helps with the analysis of band registration accuracy
and prints a statistical report.

To use the evaluation database, reset your IAS_DB_L8_OLITIRS environment 
variable to point to the evaluation database.

USAGE
}

##############################################################################
#
# Name: i2i_usage
#
# Purpose: Display command line options and general description of program
#
##############################################################################

sub i2i_usage
{
    print << 'USAGE';
$0;
ias_x2x_assess.pl [-a -p -b -t -n -w -d -h]

a - date range of image acquisition (dd-mmm-yyyy,dd-mmm-yyyy)
p - date range of image processing (dd-mmm-yyyy,dd-mmm-yyyy)
b - band list (eg: 1,3,4)
t - acquisition type ([L]unar, [E]arth, e[N]gineering)
n - nadir ([N]adir, [O]ff-nadir)
w - geographic location of image extent by WRS Path/Row (1-233,1-248)
d - disable interactive prompting for missing or invalid parameter values
h - display this usage message

ias_i2i_assess.pl helps with the analysis of image registration accuracy
and prints a statistical report.

To use the evaluation database, reset your IAS_DB_L8_OLITIRS environment 
variable to point to the evaluation database.

USAGE
}

##############################################################################
#
# Name: x2x_usage
#
# Purpose: Display command line options and general description of program.
#
##############################################################################

sub x2x_usage
{   
    print << 'USAGE';
$0;
ias_x2x_assess.pl [-v -a -p -s -b -t -n -w -d -h]

v - variant (b2b, i2i -- required parameter)
a - date range of image acquisition (dd-mmm-yyyy,dd-mmm-yyyy)
p - date range of image processing (dd-mmm-yyyy,dd-mmm-yyyy)
s - SCA list (eg: 1,3,4)
b - band list (eg: 1,3,4)
t - acquisition type ([L]unar, [E]arth, e[N]gineering)
n - nadir ([N]adir, [O]ff-nadir)
w - geographic location of image extent by WRS Path/Row (1-233,1-248)
d - disable interactive prompting for missing or invalid parameter values
h - display this usage message
    
The b2b variant helps with the analysis of band registration accuracy.
The i2i variant helps with the analysis of image registration accuracy.
Both print a statistical report.

To use the evaluation database, reset your IAS_DB_L8_OLITIRS environment 
variable to point to the evaluation database.

USAGE
}

##################################################################
#
# NAME: retrieve_records
#
# PURPOSE: Retrieve and return selected band registration characterization
# records from the database.
#
# INPUTS: logger object, pre-formatted SQL SELECT string to send to the database
#
# RETURNS: reference to an array of hashes, error status
#
##################################################################
sub retrieve_records
{
    my $logger = $_[0];   # logger error log
    my $sql = $_[1];      # pre-built SQL query
    my $sth_fetch;        # statement handle
    my @record_data;      # array of hashes

    # Prepare SQL query.
    $sth_fetch = $dbh->prepare($sql);
    if (!$sth_fetch)
    {
        $logger->fatal_msg(__FILE__, __LINE__,
            "Build of query returned an error: $DBI::errstr.");
        return (undef, ERROR);
    }

    # Execute SQL query.
    if (!$sth_fetch->execute())
    {
        $logger->fatal_msg(__FILE__, __LINE__,
            "Query returned an error: $DBI::errstr.");
        return (undef, ERROR);
    }

    # Fetch and organize the results of the query.
    while (my $db_row_ref = $sth_fetch->fetchrow_hashref())
    {
        my $record = {};
        foreach my $column (@columns)
        {
            $record->{$column} = $db_row_ref->{$column};
        }
        push @record_data, $record;
    }

    return (\@record_data, SUCCESS);
}


##################################################################
#
# NAME: read_number_list
#
# PURPOSE: Retrieve a list of comma-separated numbers from the command line.
# This may be used for things like SCA lists or band lists.
#
# INPUTS: logger object, prompt to display, user-entered list of numbers, 
# and valid minimum and maximum for the particular type of number list (eg: 
# band or SCA list), no-prompt flag
#
# RETURNS: a comma-separated list of entries in string format ready to be used
# as part of a SQL command, error status
#
##################################################################
sub read_number_list
{
    my $logger = $_[0];           # logger error log
    my $prompt_string = $_[1];    # string to display to prompt for input
    my $list_str = $_[2];         # string read from the terminal
    my $min = $_[3];              # minimum valid value
    my $max = $_[4];              # maximum valid value
    my $no_prompt_flag = $_[5];   # flag to disable prompts
    my $final_list_str;           # string cleaned up for return
    my @list_array;               # entry array - no commas, just numbers
    my $element;                  # single number from the list
    my $parsed_list_element;      # single number from the parsed list
    my $valid_list;               # validity flag
    my $duplicate_entry;          # validity flag
    my $first_entry;              # flag for the first value in the list

    while ()
    {
        # Initialize flag to "true".
        $valid_list = 1;

        # If the filter was selected but no value for it was selected,
        # then we have to prompt for it and read it the first time in here.
        # Otherwise we only prompt if it was entered incorrectly.
        if ($list_str eq '')
        {
            # If the user disabled prompting with the command line option
            # then just throw an error about missing needed input and quit.
            if ($no_prompt_flag)
            {
                $logger->fatal_msg(__FILE__, __LINE__,
                    'Prompting is disabled and no input was provided.');
                return (undef, ERROR);
            }

            # Let the user know what data is needed.
            print $prompt_string;

            # Read the list the user typed.
            $list_str = <STDIN>;
        }
        chomp($list_str);

        # Make an array of numbers by splitting on the commas.
        @list_array = split /,/, $list_str;

        # Make sure user entered at least one item.
        if ((scalar @list_array) < 1)
        {
            $valid_list = 0;
        }

        # Clear the string so it's recognized as needing to be filled in
        # should the current version prove to be syntactically incorrect.
        $list_str = '';

        # Initialize the list of elements we've already validated.
        my @parsed_list_elements;

        # Verify each number.
        foreach $element (@list_array)
        {
            # Verify that the element is a number and that it's in range.
            $_ = $element;
            # Make sure it's all integers.  It's ok for the integers to have
            # whitespace around (but not within) them.
            if (/^\s*\d+\s*$/)
            {
                # Remove all whitespace around the number to ease later checks.
                $element =~ s/^\s+//;
                $element =~ s/\s+$//;

                # Make sure it falls in a valid range.
                if (($element >= $min) and ($element <= $max))
                {
                    # Make sure it's not a duplicate.
                    $duplicate_entry = 0;
                    foreach $parsed_list_element (@parsed_list_elements)
                    {
                        if ($element eq $parsed_list_element)
                        {
                            $duplicate_entry = 1;
                            $valid_list = 0;
                            print "Number $element is a duplicate.\n";
                            last;
                        }
                    }

                    if (!$duplicate_entry)
                    {
                        # Add the element to the parsed elements list.
                        push @parsed_list_elements, $element;
                    }
                }
                else
                {
                    print "Number $element should be between $min and $max.\n";
                    $valid_list = 0;
                }
            }
            else
            {
                print "Number $element should be an integer.\n";
                $valid_list = 0;
            }
        }

        # If it's valid move on.
        if ($valid_list)
        {
            # Initialize flag to true.
            $first_entry = 1;

            # Build the cleaned up (no white space) list to return.
            foreach $parsed_list_element (@parsed_list_elements)
            {
                # The elements that aren't the first get a comma in front.
                if ($first_entry)
                {
                    # Toggle it.
                    $first_entry = 0;
                }
                else
                {
                    $final_list_str .= ',';
                }
                $final_list_str .= $parsed_list_element;
            }

            return ($final_list_str, SUCCESS);
        }
    }
}

##################################################################
#
# NAME: read_acquisition_type
#
# PURPOSE: Retrieve acquisition type from the command line.  The function
# does some rudimentary checks on the acquisition type entered.
#
# INPUTS: logger object, user-entered acquisition type, no-prompt flag
#
# RETURNS: valid user-selected acquisition type value, error status
#
##################################################################
sub read_acquisition_type
{
    my $logger = $_[0];         # logger error log
    my $type_str = $_[1];       # acquisition type string read from the terminal
                                # We only want a character but a string might
                                # have been typed.
    my $no_prompt_flag = $_[2]; # flag to disable prompts

    my $valid_type;             # validity flag
    my $type;                   # type selection

    while ()
    {
        # Initialize flag to "true".
        $valid_type = 1;

        # If the filter was selected but no value for it was selected,
        # then we have to prompt for it and read it the first time in here.
        # Otherwise we only prompt if it was entered incorrectly.
        if ($type_str eq '')
        {
            # If the user disabled prompting with the command line option
            # then just throw an error about missing needed input and quit
            if ($no_prompt_flag)
            {
                $logger->fatal_msg(__FILE__, __LINE__,
                    'Prompting is disabled and no input was provided.');
                return (undef, ERROR);
            }

            # Let the user know what data is needed.
            print 'acquisition type ([L]unar, [E]arth, e[N]gineering): ';

            # Read the date the user typed.
            $type_str = <STDIN>;
        }
        chomp($type_str);

        $type = $type_str;

        # Clear the string so it's recognized as needing to be filled in
        # should the current version prove to be syntactically incorrect.
        $type_str = '';

        # Make the acquisition type all upper case.
        $type =~ tr/a-z/A-Z/;

        if (($type ne 'L') && ($type ne 'E') && ($type ne 'N'))
        {
            $valid_type = 0;
            print "The acquisition type $type should be L, E, or N.\n";
        }

        if ($valid_type)
        {
            return ($type, SUCCESS);
        }
    }
}

##################################################################
#
# NAME: read_nadir
#
# PURPOSE: Retrieve the nadir flag from the command line.  The function
# does some rudimentary checks on the nadir type entered.
#   
# INPUTS: logger object, user-entered nadir type, no-prompt flag
#
# RETURNS: valid user-selected nadir type value, error status
#
##################################################################
sub read_nadir
{
    my $logger = $_[0];         # logger error log
    my $type_str = $_[1];       # nadir type string read from the terminal
                                # We only want a character but a string might
                                # have been typed.
    my $no_prompt_flag = $_[2]; # flag to disable prompts

    my $valid_type;             # validity flag
    my $type;                   # type selection

    while ()
    {
        # Initialize flag to "true".
        $valid_type = 1;

        # If the filter was selected but no value for it was selected,
        # then we have to prompt for it and read it the first time in here.
        # Otherwise we only prompt if it was entered incorrectly.
        if ($type_str eq '')
        {
            # If the user disabled prompting with the command line option
            # then just throw an error about missing needed input and quit.
            if ($no_prompt_flag)
            {
                $logger->fatal_msg(__FILE__, __LINE__,
                    'Prompting is disabled and no input was provided.');
                return (undef, ERROR);
            }

            # Let the user know what data is needed.
            print 'nadir ([N]adir, [O]ff-nadir): ';

            # Read the date the user typed.
            $type_str = <STDIN>;
        }
        chomp($type_str);

        $type = $type_str;

        # Clear the string so it's recognized as needing to be filled in
        # should the current version prove to be syntactically incorrect.
        $type_str = '';

        # Make the nadir type all upper case.
        $type =~ tr/a-z/A-Z/;

        if (($type ne 'N') && ($type ne 'O'))
        {
            $valid_type = 0;
            print "The nadir type $type should be N or O.\n";
        }

        if ($valid_type)
        {
            return ($type, SUCCESS);
        }
    }
}

##################################################################
#
# NAME: build_sql
#
# PURPOSE: Form an SQL command based on the user selections.
#
# INPUTS: all user selections, including which filters they selected,
# and what particular options they selected within those filters
#
# RETURNS: SQL command string that's ready to be sent to the database
#
##################################################################
sub  build_sql
{
    my $acquisition_date_range_flag = $_[0];  # was filter selected?
    my $processing_date_range_flag = $_[1];   # was filter selected?
    my $sca_list_flag = $_[2];                # was filter selected?
    my $band_list_flag = $_[3];               # was filter selected?
    my $type_of_acquisition_flag = $_[4];     # was filter selected?
    my $wrs_path_row_flag = $_[5];            # was filter selected?
    my $b2b_flag = $_[6];                     # which tool is running?
    my $i2i_flag = $_[7];                     # which tool is running?
    my $start_acquisition_date = $_[8];       # user selection
    my $end_acquisition_date = $_[9];         # user selection
    my $start_processing_date = $_[10];       # user selection
    my $end_processing_date = $_[11];         # user selection
    my $sca_list = $_[12];                    # user selection
    my $band_list = $_[13];                   # user selection
    my $acquisition_type = $_[14];            # user selection
    my $start_path = $_[15];                  # user selection
    my $end_path = $_[16];                    # user selection
    my $start_row = $_[17];                   # user selection
    my $end_row = $_[18];                     # user selection
    my $nadir = $_[19];                       # user selection

    my $sql;                                  # SQL command string
    my $where_flag;                           # do we need a where clause?
    my $collection_type;                      # DB acquisition type 

    if (($acquisition_date_range_flag) ||
        ($processing_date_range_flag) ||
        ($sca_list_flag) ||
        ($band_list_flag) ||
        ($type_of_acquisition_flag) ||
        ($nadir_flag) ||
        ($wrs_path_row_flag))
    {
        $where_flag = 1;
    }
    else
    {
        $where_flag = 0;
    }

    # Start building the SQL command.

    $sql = qq{ SELECT 
        geometric_work_order_common.geometric_wo_common_id,
        off_nadir_angle,
        nbr_of_pnts_corr,
        nbr_of_pnts_successfully_corr, 
        nbr_of_vld_pnts, 
        t_distribution_threshold,
        rmse_line, 
        rmse_sample, 
        std_dev_line, 
        std_dev_sample,
        min_line, 
        min_sample, 
        max_line, 
        max_sample,
        med_line, 
        med_sample, 
        mean_line, 
        mean_sample, 
        reference_band_id, 
        search_band_id, 
        sensor_chip_assembly_id,
        target_wrs_path, 
        target_wrs_row,
        collection_type,
        date_acquired,
        date_processed
        };

    if ($b2b_flag)
    {
        $sql .= qq{
            FROM 
            band_characterization,
            band_characterization_band_sca,
            geometric_work_order_common
            WHERE 
            geometric_work_order_common.geometric_wo_common_id 
                = band_characterization.geometric_wo_common_id
            and geometric_work_order_common.geometric_wo_common_id 
                = band_characterization_band_sca.geometric_wo_common_id
            and geometric_work_order_common.valid = 'Y' 
            };
    }

    if ($i2i_flag)
    {
        $sql .= qq{
            FROM 
            image_char_band_sca,
            image_characterization,
            geometric_work_order_common
            WHERE 
            geometric_work_order_common.geometric_wo_common_id 
                = image_characterization.geometric_wo_common_id
            and geometric_work_order_common.geometric_wo_common_id 
                = image_char_band_sca.geometric_wo_common_id
            and geometric_work_order_common.valid = 'Y' 
            };
    }

    # Build additional parts of the SQL command based on user selections.
    if ($where_flag)
    {
        # Add all the customizations to the SQL where clause based on the
        # user selections.

        # Handle the acquisition date range option.
        if ($acquisition_date_range_flag)
        {
            $sql .= " AND date_acquired >= '";
            $sql .= $start_acquisition_date . "'";
            $sql .= " AND date_acquired <= '";
            $sql .= $end_acquisition_date . "'";
        }

        # Handle the processing date range option. 
        if ($processing_date_range_flag)
        {
            $sql .= " AND date_processed >= '";
            $sql .= $start_processing_date . "'";
            $sql .= " AND date_processed <= '";
            $sql .= $end_processing_date . "'";
        }

        # Handle the band option.
        if ($band_list_flag)
        {
            $sql .= ' AND ';
            $sql .= '(reference_band_id IN (' . $band_list . ') OR ';
            $sql .= 'search_band_id IN (' . $band_list . ')) ';
        }

        # Handle the SCA option.
        if ($b2b_flag)
        {
            if ($sca_list_flag)
            {
                $sql .= ' AND sensor_chip_assembly_id IN (' . $sca_list . ') ';
            }
        }

        # Handle the acquisition type option.
        if ($type_of_acquisition_flag)
        {
            if ($acquisition_type eq 'L') # Lunar
            {
                $collection_type = "'LUNAR'";
            }

            if ($acquisition_type eq 'E') # Earth
            {
                $collection_type = "'EARTH IMAGING'";
            }

            if ($acquisition_type eq 'N') # engineering
            {
                $collection_type = "'ENGINEERING'";
            }

            $sql .= ' AND collection_type = ' . $collection_type . ' '; 
        }

        # Handle the nadir type option.
        if ($nadir_flag)
        {
            if ($nadir eq 'N') # Nadir
            {
                $sql .= ' AND off_nadir_angle BETWEEN -0.5 and 0.5';
            }

            if ($nadir eq 'O') # Off-nadir
            {
                $sql .= ' AND off_nadir_angle NOT BETWEEN -0.5 and 0.5';
            }
        }

        # Handle the WRS path and row option.
        if ($wrs_path_row_flag)
        {
            $sql .= ' AND ';
            # In case they searched for a path range like "230-5"...
            if ($start_path > $end_path)
            {
                $sql .= '(target_wrs_path BETWEEN ' . $start_path . ' AND 233';
                $sql .= ' OR target_wrs_path BETWEEN 1 AND ' . $end_path . ')';
            }
            else
            {
                $sql .= 'target_wrs_path BETWEEN ' . $start_path;
                $sql .= ' AND ' . $end_path;
            }
            # In case they searched for a row range like "230-5"...
            $sql .= ' AND ';
            if ($start_row > $end_row)
            {
                $sql .= '(target_wrs_row BETWEEN ' . $start_row .' AND 248 ';
                $sql .= ' OR target_wrs_row BETWEEN 1 AND ' . $end_row . ')';
            }
            else
            {
                $sql .= 'target_wrs_row BETWEEN ' . $start_row;
                $sql .= ' AND ' . $end_row;
            }
        }

        # Sort the results so the ordering in the report file will be 
        # consistent each time the script runs
        $sql .= " ORDER BY geometric_work_order_common.geometric_wo_common_id,"
             .  "  reference_band_id, search_band_id, sensor_chip_assembly_id";
    }

    return $sql;
}
