/*=============================================================================
NAME: quality_band_extraction

PURPOSE:
    This application is to extract specific bits from quality band. The
    application band specifies which type of artifacts to be extracted. 

=============================================================================*/

/* System Includes */
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>

/* IAS Includes */
#include "ias_logging.h"
#include "ias_l1g.h"
#include "ias_geo.h"
#include "ias_const.h"
#include "ias_miscellaneous.h"
#include "config.h" /* This is for the endian define */

/* Number of lines to process in a block */
#define NUM_LINES 1000
/*------------------------------------------------------------------------------
NAME: usage

PURPOSE:
    This function print out the usage content 

------------------------------------------------------------------------------*/
static void usage()
{
    printf("Usage: quality_band_extraction <input_quality_band_filename>\n"
         "\t<output_mask_filename> <bit_mask> [bit_value]\nwhere:\n"
         "\t<input_quality_band_filename> = Quality band filename used as "
         "inputs\n"
         "\t<output_mask_filename> = Output mask file name\n"
         "\t<bit_mask> = Hexadecimal value for which bits should be extracted\n"
         "\t[bit_value] = Optional hexadecimal value with the exact pattern "
         "the\n\t\tmasked bits should match, if the bit_value parameter is "
         "left\n\t\toff, it should just use bits that are set\n"
         "-----------------------------------------------------------------"
         "-------------\n"
         "\tThe bits are defined below:\n"
         "\tNOTHING                                    0x0000 (No bits are selected)\n"
         "\tFILL                                       0x0001 (Bit 0)\n"
         "\tTERRAIN_OCCLUSION                          0x0002 (Bit 1)\n"
         "\tRADIOMETRIC_SAT_01 (1-2 bands saturated)   0x0004 (Bit 2)\n"
         "\tRADIOMETRIC_SAT_10 (3-4 bands saturated)   0x0008 (Bit 3)\n"
         "\tRADIOMETRIC_SAT_11 (more than 4 bands saturated) 0x000C (Bit 2 and 3)\n"
         "\tCLOUD SUMMARY                              0x0010 (Bit 4)\n"
         "\tCLOUD_CONFIDENCE_01 (0-35%% confidence)    0x0020 (Bit 5)\n"
         "\tCLOUD_CONFIDENCE_10 (36-64%% confidence)   0x0040 (Bit 6)\n"
         "\tCLOUD_CONFIDENCE_11 (65-100%% confidence)  0x0060 (Bit 5 and 6)\n"
         "\tCLOUD_SHADOW_01 (0-35%% confidence)        0x0080 (Bit 7)\n"
         "\tCLOUD_SHADOW_10 (36-64%% confidence)       0x0100 (Bit 8)\n"
         "\tCLOUD_SHADOW_11 (65-100%% confidence)      0x0180 (Bit 7 and 8)\n"
         "\tSNOW_ICE_01 (0-35%% confidence)            0x0200 (Bit 9)\n"
         "\tSNOW_ICE_10 (36-64%% confidence)           0x0400 (Bit 10)\n"
         "\tSNOW_ICE_11 (65-100%%) confidence)         0x0600 (Bit 9 and 10)\n"
         "\tCIRRUS (0-35%% confidence)                 0x0800 (Bit 11)\n"
         "\tCIRRUS (36-64%% confidence)                0x1000 (Bit 12)\n"
         "\tCIRRUS (65-100%% confidence)               0x1800 (Bit 11 and 12)\n"
         "\tRESERVED                                   0x2000 (Bit 13)\n"
         "\tRESERVED                                   0x4000 (Bit 14)\n" 
         "\tRESERVED                                   0x8000 (Bit 15)\n"); 
}

int main (int argc, char * argv[])
{
    L1GIO *qualband_ptr;
    L1G_BAND_IO *l1g_band;
    IAS_DATA_TYPE data_type;
    IAS_L1G_FILE_METADATA file_meta;
    IAS_L1G_BAND_METADATA band_meta;
    IAS_PROJECTION proj_info;
    const char *quality_band_filename;
    const char *bit_mask_string;
    unsigned int parm_value;
    unsigned short bit_mask;
    unsigned short bit_value = 0;
    int bit_value_provided = 0;
    unsigned short *data = NULL;
    unsigned char *output_mask = NULL;
    FILE * fd = NULL;
    int num_scas;
    int num_lines;
    int num_samples;
    int seg, line, sample;
    const char *output_filename;
    int counter = 0;
    int nsegments;
    int status;
    int band_number = 1;
    char description[PATH_MAX * 2];
    const char *temp_ptr;

    /* Initialize logging */
    if (ias_log_initialize("quality_band_extraction"))
    {
        IAS_LOG_ERROR("Initializing logging capabilities");
        exit(EXIT_FAILURE);
    }

    /* Process the command line arguments */
    if ((argc >= 2 && !strcmp(argv[1], "-help")) || (argc < 4) || (argc > 5))
    {
        usage();
        exit(EXIT_SUCCESS);
    }

    /* Assign the needed parameters based on the command line arguments */
    quality_band_filename = argv[1];
    output_filename = argv[2];
    bit_mask_string = argv[3];

    /* Parse the bit mask */
    if (sscanf(bit_mask_string, "%x", &parm_value) != 1)
    {
        IAS_LOG_ERROR("Error parsing BIT_MASK parameter '%s' as a hexadecimal "
            "number", bit_mask_string);
        exit(EXIT_FAILURE);
    }
    bit_mask = (unsigned short) parm_value;
    if (bit_mask == 0)
    {
        IAS_LOG_ERROR("Nothing to check (mask values all zero) when "
                      "bit_mask_string is 0x0000");
        exit(EXIT_FAILURE);
    }

    /* If the bit value is provided, parse it */
    if (argc == 5)
    {
        const char *bit_value_string;
        bit_value_string = argv[4];
        bit_value_provided = 1;
        if (sscanf(bit_value_string, "%x", &parm_value) != 1)
        {
            IAS_LOG_ERROR("Error parsing BIT_VALUES parameter '%s' as a "
                "hexadecimal number", bit_value_string);
            exit(EXIT_FAILURE);
        } 
        bit_value = (unsigned short) parm_value;
    }

    IAS_LOG_INFO("Processing started");

    /* Open the quality band image for imagery access */
    qualband_ptr = ias_l1g_open_image(quality_band_filename, IAS_READ);
    if (qualband_ptr == NULL)
    {
        IAS_LOG_ERROR("Opening image %s", quality_band_filename);
        exit(EXIT_FAILURE);
    }

    /* Read the file metadata and copy the needed proj info to the 
       IAS_PROJECTION structure */
    status = ias_l1g_get_file_metadata(qualband_ptr, &file_meta);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading file metadata from %s", quality_band_filename);
        ias_l1g_close_image(qualband_ptr);
        exit(EXIT_FAILURE);
    }
    proj_info.proj_code = file_meta.projection_code;
    proj_info.zone = file_meta.zone_code;

    /* Open the quality band */
    l1g_band = ias_l1g_open_band(qualband_ptr, band_number, &data_type, 
                     &num_scas, &num_lines, &num_samples);
    if (l1g_band == NULL)
    {
        IAS_LOG_ERROR("Opening Quality band of file %s", quality_band_filename);
        ias_l1g_close_image(qualband_ptr);
        exit(EXIT_FAILURE);
    }

    /* Quality band should have only one combined SCA */
    if (num_scas != 1)
    {
        IAS_LOG_ERROR("Quality band file %s does not only have one SCA", 
                      quality_band_filename);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(qualband_ptr);
        exit(EXIT_FAILURE);
    }

    /* Confirm the data is the expected type */
    if (data_type != IAS_UI2)
    {
        IAS_LOG_ERROR("Quality band file %s does not have the expected "
            "unsigned short int (UI*2) data type", quality_band_filename);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(qualband_ptr);
        exit(EXIT_FAILURE);
    }

    /* Read the band file metadata */
    status = ias_l1g_get_band_metadata(qualband_ptr, 1, &band_meta);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading band metadata from %s", quality_band_filename);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(qualband_ptr);
        exit(EXIT_FAILURE);
    }

    /* Get the number of segments for loop to read out the quality band data */
    nsegments = ceil((double)num_lines / (double)NUM_LINES);

    /* Allocate memory for the data buffer */
    data = malloc(NUM_LINES * num_samples * sizeof(*data));
    if (data == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for data buffer");
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(qualband_ptr);
        exit(EXIT_FAILURE);
    }

    /* Allocate memory for the mask buffer */
    output_mask = malloc(NUM_LINES * num_samples * sizeof(*output_mask));
    if (output_mask == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for mask buffer");
        free(data);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(qualband_ptr);
        exit(EXIT_FAILURE);
    }

    /* Open the mask file for writing */
    fd = fopen(output_filename, "w"); 
    if (fd == NULL)
    {
        IAS_LOG_ERROR("Opening report file: %s", output_filename);
        free(output_mask);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(qualband_ptr);
        exit(EXIT_FAILURE);
    }

    /* Based on mask parameters to fill the output mask accordingly */
    for (seg = 0; seg < nsegments; seg++)
    {
        int block_lines = NUM_LINES;
        
        /* If on the last segment, only read the lines remaining */
        if (seg == nsegments - 1)
            block_lines = num_lines - seg * NUM_LINES;

        /* Read the current block of lines */
        status = ias_l1g_read_image(l1g_band, 0, seg * NUM_LINES, 0, 
             block_lines, num_samples, data);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading quality band image file: %s",
                quality_band_filename);
            free(data);
            free(output_mask);
            fclose(fd);
            ias_l1g_close_band(l1g_band);
            ias_l1g_close_image(qualband_ptr);
            exit(EXIT_FAILURE);
        }

        /* Check each pixel in the block to build the output mask */
        for (line = 0; line < block_lines; line++)
        {
            /* cache pointers to the current line */
            unsigned short *data_line = &data[line * num_samples];
            unsigned char *mask_line = &output_mask[line * num_samples];

            if (bit_value_provided)
            {
                /* Loop over the samples in the line looking for the specific
                   bit value in the masked bits */
                for (sample = 0; sample < num_samples; sample++)
                {
                    unsigned char mask_value = 0;

                    if ((data_line[sample] & bit_mask) == bit_value)
                    {
                        mask_value = 1;
                        counter++;
                    }
                    mask_line[sample] = mask_value;
                }
            }
            else
            {
                /* Loop over the samples in the line looking for any bits set
                   in the masked value since a specific value wasn't
                   requested */
                for (sample = 0; sample < num_samples; sample++)
                {
                    unsigned char mask_value = 0;

                    if (data_line[sample] & bit_mask)
                    {
                        mask_value = 1;
                        counter++;
                    }
                    mask_line[sample] = mask_value;
                }
            }
        }

        /* Write out the mask file */
        status = fwrite(output_mask, sizeof(unsigned char), block_lines * 
                 num_samples, fd);
        if (status != block_lines * num_samples)
        {
            IAS_LOG_ERROR("Writing to %s", output_filename);
            free(output_mask);
            free(data);
            fclose(fd);
            ias_l1g_close_band(l1g_band);
            ias_l1g_close_image(qualband_ptr);
            exit(EXIT_FAILURE);
        }
    }

    /* free the data buffer */
    free(data);
    data = NULL;

    /* free the mask buffer */
    free(output_mask);
    output_mask = NULL;

    /* Close the mask file */
    status = fclose(fd);
    if ( status )
    {
        IAS_LOG_ERROR("Closing file %s", output_filename);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(qualband_ptr);
        exit(EXIT_FAILURE);
    }

    /* close the quality band */
    status = ias_l1g_close_band(l1g_band);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Closing quality band");
        ias_l1g_close_image(qualband_ptr);
        exit(EXIT_FAILURE);
    }

    /* close the quality band image */
    status = ias_l1g_close_image(qualband_ptr);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Closing quality band image");
        exit(EXIT_FAILURE);
    }

    /* Build the description for the envi header using just the name of the
       input file without the full path */
    temp_ptr = strrchr(quality_band_filename, '/');
    if (temp_ptr)
        temp_ptr++;
    else
        temp_ptr = quality_band_filename;
    if (bit_value_provided)
    {
        status = snprintf(description, sizeof(description), "Quality band mask "
            "extracted from %s with a mask of %#x and bit value of %#x.  "
            "%d pixels found", temp_ptr, bit_mask, bit_value, counter);
    }
    else
    {
        status = snprintf(description, sizeof(description), "Quality band mask "
            "extracted from %s with a mask of %#x", temp_ptr, bit_mask);
    }
    if (status < 0 || status >= sizeof(description))
    {
        IAS_LOG_ERROR("Buffer for envi header description not large enough");
        exit(EXIT_FAILURE);
    }

    /* Create an ENVI header file for the image */
    status = ias_misc_write_envi_header(output_filename, &proj_info, 
             description, num_lines, num_samples, 1, band_meta.upper_left_x, 
             band_meta.upper_left_y, band_meta.projection_distance_x, 
             band_meta.projection_distance_y, NULL, IAS_BYTE);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Creating ENVI header for image");
        exit(EXIT_FAILURE);
    }

    /* Provide the info about how many pixels were found */
    if (bit_value_provided)
    {
        IAS_LOG_INFO("Found %d pixels using a mask of %#x and bit value of %#x",
            counter, bit_mask, bit_value);
    }
    else
    {
        IAS_LOG_INFO("Found %d pixels using a mask of %#x", counter, bit_mask);
    }

    /* Log successful completion */
    IAS_LOG_INFO("Completed successfully");
    exit(EXIT_SUCCESS);

}
