#!/usr/bin/perl -w
#-------------------------------------------------------------------------------
# regression test for ias_assess_geo.pl
#
# > clears out records in the database tables that will be used
# > inserts a control set of records into the tables
# > runs the assessment tool on them with various query constraints
# > compares the generated output reports in each test case to control
# > fatal errors at point of failure if error occurs; error to STDOUT or STDERR
#-------------------------------------------------------------------------------

use strict;
use ias_db;
use ias_const;
use ias_processing_system;  # ias_init_system and ias_validate_environment
use Switch;

my $sth;
my $dbh;
my $sql;
my $userID;
my $password;
my @woIDs;
my $woIDs;
my $commonID;
my @commonIDs = ();
my $commonIDs;
my $dataDir = $ENV{IAS_UNIT_TEST_DATA_DIR};
$dataDir .= "/ias_base/gps/tools/ias_assess_geo/";

# make sure DB tests are enabled
if (not defined($ENV{ENABLE_DB_TESTS}))
{
    print "ias_assess_geo.pl testing not performed. Requires database.\n";
    exit SUCCESS;
}

# Ensure that the required environment varibles are defined
my @varList = qw( IAS_DB_L8_OLITIRS ORACLE_SID );
my $envVarList = \@varList;
if (ias_init_system() == ERROR)
{
    print "$ias_error_msg\n";
    exit FAILURE;
}
if (ias_validate_environment($envVarList) > 0)
{
    print "$ias_error_msg\n";
    exit FAILURE;
}

# geometric test cases
@woIDs = qw(assessGeo_a assessGeo_b assessGeo_c assessGeo_d assessGeo_e 
            assessGeo_f assessGeo_g); # f and g are marked 'invalid'

# get the username and password for the database
($userID, $password) = ias_db_GetDBConnectionInfo("L8_OLITIRS");

# connect to the database
$dbh = ias_db_ConnectToDB($userID, $password);
if (!$dbh)
{
    print "Error connecting to database: $ias_db_error\n";
    exit FAILURE;
}

# test status message
print "Database connection passed\n";

# prepare to delete existing records with same work_order_id as this test
# program will be using from the relevant tables
$woIDs = join "','", @woIDs;
$sql = qq{
    SELECT geometric_wo_common_id from geometric_work_order_common
        WHERE work_order_id in ('$woIDs')
};
$sth = &executeSql($sql);

# collect list of common IDs to get rid of records for, in case they
# already exist
push @commonIDs, $commonID while ($commonID = $sth->fetchrow_array());
if ($#commonIDs > -1)
{
    # delete existing records for work order IDs the test will be working with
    $commonIDs = join ',', @commonIDs;

    # clean up existing records from the geodetic accuracy assessment table
    $sql = qq{
        DELETE from geodetic_characterization
            WHERE geometric_wo_common_id in ($commonIDs)
    };
    my $rows = $dbh->do($sql);
    if (not defined $rows)
    {
        print "Error deleting geodetic_characterization table records: ",
              "$DBI::errstr\n";
        exit FAILURE;
    }
    if (!$dbh->commit())
    {
        print "Error committing geodetic_characterization table cleanup: ",
              "$DBI::errstr\n";
        exit FAILURE;
    }

    # clean up existing records from the geometric accuracy assessment table
    $sql = qq{
        DELETE from geometric_accuracy
            WHERE geometric_wo_common_id in ($commonIDs)
    };
    $rows = $dbh->do($sql);
    if (not defined $rows)
    {
        print "Error deleting geometric_accuracy table records: $DBI::errstr\n";
        exit FAILURE;
    }
    if (!$dbh->commit())
    {
        print "Error committing geometric_accuracy table cleanup: ",
              "$DBI::errstr\n";
        exit FAILURE;
    }

    # clean up existing records from the geometric work order common table
    $sql = qq{
        DELETE from geometric_work_order_common
            WHERE geometric_wo_common_id in ($commonIDs)
    };
    $rows = $dbh->do($sql);
    if (not defined $rows)
    {
        print "Error deleting geometric_work_order_common table records: ",
              "$DBI::errstr\n";
        exit FAILURE;
    }
    if (!$dbh->commit())
    {
        print "Error committing geometric_word_order_common table cleanup: ",
              "$DBI::errstr\n";
        exit FAILURE;
    }
}

# test status message
print "Database cleanup passed\n";

# insert the test data records we need, commit the data and close connection
&insertTestData($dbh, $dataDir, @woIDs);
$sth->finish();
if (!ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_COMMIT))
{
    print "Warning: Failed disconnect from database: $ias_db_error\n";
}

# test status message
print "Database loading passed\n";

# run a few different tests to run different queries with the different
# input constraints that can be used; provide test status messages
&testAssessment($dataDir, "WRS");
print "Geographic location constraints test passed\n";

&testAssessment($dataDir, "Date");
print "Date constraints test passed\n";

&testAssessment($dataDir, "WRS", "GCPSource");
print "Geographic location and GCP source constraints test passed\n";

&testAssessment($dataDir, "WRS", "RollAngle");
print "Geographic location and roll angle constraints test passed\n";

&testAssessment($dataDir, "Date", "GCPSource");
print "Date and GCP source constraints test passed\n";

&testAssessment($dataDir, "Date", "RollAngle");
print "Date and roll angle constraints test passed\n";

&testAssessment($dataDir, "WRS", "Date", "RollAngle");
print "Geographic location, date, and roll angle constraints test passed\n";

&testAssessment($dataDir, "WRS", "Date", "GCPSource");
print "Geographic location, date, and GCP source constraints test passed\n";

&testAssessment($dataDir, "WRS", "Date", "GCPSource", "RollAngle");
print "Geographic location, date, GCP source, and roll angle constraints test ",
      "passed\n";

# done with tests; remove intermediate output file, issue message and quit
unlink('output.txt');
print "test_ias_assess_geo.pl completed successfully\n";
exit SUCCESS;

#--------------------------------------------------------------------
# subroutines
#--------------------------------------------------------------------

# routine to run the tests of ias_assess_geo.pl
sub testAssessment
{
    my ($dataDir, @constraints) = @_;

    my $command;
    my $compareFile = $dataDir;
    
    # Run geodetic test with given constraints (query parameters)
    $command = "cd ../.. && ./ias_assess_geo.pl -v geodetic -d ";

    # routine will be passed the list of constraints to apply for this
    # specific test of the script; for each constraint indicated, tack
    # on a query-limiting parameter for the ias_assess_geo.pl script
    foreach (@constraints)
    {
        switch ($_)
        {
            $compareFile .= $_;
            case "Date" { $command .= "-a 01-JAN-2008,31-DEC-2008 "; }
            case "GCPSource" { $command .= "-g D "; }
            case "RollAngle" { $command .= "-r 14 "; }
            case "WRS" { $command .= "-w 223-228,52-57 "; }
            else
            {
                print "Error determining which assessment test to run.\n";
                exit FAILURE;
            }
        }
    }
    # wrap up the control file name that will be compared against
    $compareFile .= "Geod.txt";
    # finish building the command
    $command .= "> tests/ias_assess_geo/output.txt";

    # run the script for this geodetic test
    system($command);

    # check if the saved file of expected reporting output from the script
    # matches what was generated by running the script
    my $diff_cmd =
        "sed '/Geodetic post/,\$!'d output.txt | diff -q - $compareFile";
    if (system($diff_cmd) != 0)
    {
        print "Testing constraints (@constraints) failed: 'output.txt' is not ",
              "equal to '$compareFile'\n";
        exit FAILURE;
    }

    # Run geometric test with given constraints (query parameters)
    $compareFile = $dataDir; # reset the base of this file path/name var
    $command = "cd ../.. && ./ias_assess_geo.pl -v geometric -d ";

    # routine will be passed the list of constraints to apply for this
    # specific test of the script; for each constraint indicated, tack
    # on a query-limiting parameter for the ias_assess_geo.pl script
    foreach (@constraints)
    {
        switch ($_)
        {
            $compareFile .= $_;
            case "Date" { $command .= "-a 31-JUL-2005,31-DEC-2007 "; }
            case "GCPSource" { $command .= "-g G "; }
            case "RollAngle" { $command .= "-r 8 "; }
            case "WRS" { $command .= "-w 68-72,43-48 "; }
            else
            {
                print "Error determining which assessment test to run.\n";
                exit FAILURE;
            }
        }
    }
    # wrap up the control file name that will be compared against
    $compareFile .= "Geom.txt";
    # finish building the command
    $command .= "> tests/ias_assess_geo/output.txt";

    # run the script for this geometric test
    system($command);
    
    # check if the saved file of expected reporting output from the script
    # matches what was generated by running the script
    $diff_cmd =
        "sed '/Geometric post/,\$!'d output.txt | diff -q - $compareFile";
    if (system($diff_cmd) != 0)
    {
        print "Testing constraints (@constraints) failed: 'output.txt' is not ",
              "equal to '$compareFile'\n";
        exit FAILURE;
    }
}

# execute an SQL statement, returning the statement handle
sub executeSql
{
    my ($sql) = @_;

    $sth = $dbh->prepare($sql);
    if (!$sth)
    {
        print "Error preparing database query: $DBI::errstr\n";
        exit FAILURE;
    }

    if (!$sth->execute())
    {
        print "Query returned an error: $DBI::errstr\n";
        exit FAILURE;
    }
    
    return $sth;
}

# insert a set of records into the database tables that can be used
# to run ias_assess_geo.pl testing against
# NOTE: If any significant changes are going to be made to this area of the
# code, consider switching over to using Oracle::Sqlldr rather than this
# manual file structure and parsing that was implemented initially.
sub insertTestData
{
    my ($dbh, $dataDir, @woIDs) = @_;
    
    my $rowsInserted;

    # read the work order files and insert the data from each
    foreach (@woIDs)
    {
        my $file = $dataDir . $_;
        open(FILE, "<$file") or die "Unable to read $file\n";
        
        # get the associated geometric or geodetic file for database insert
        my $geomGeodTableFile = $dataDir . <FILE>;
        chop $geomGeodTableFile;

        # add the new work order record to the geometric common table
        $sql = "INSERT into geometric_work_order_common ";

        # parse the file, adding components of the insert statement to the
        # SQL string; field names are found first in the file, each on a new
        # line; VALUES keyword is next; then the values are at the end; format
        # of the file is specific and must not be changed without adjusting
        # this code for any format modifications (and vice versa)
        while (<FILE>)
        {
            chop;
            if (/VALUES/)
            {
                chop $sql; # remove last comma
                $sql .= " $_ ";
            }
            else
            {
                $sql .= "$_,";
            }
        }
        chop $sql; # drop trailing comma
        close FILE;

        $rowsInserted = $dbh->do($sql);
        if ($rowsInserted != 1)
        {
            print "Error inserting record: $sql\n";
            exit FAILURE;
        }
        
        # add the geometric or geodetic table record as well
        open(FILE, "<$geomGeodTableFile")
            or die "Unable to read $geomGeodTableFile\n";

        # the first two lines in the file will be the database table name
        # for the record associated with this work order (either the geodetic
        # or geometric assessment table) and then the work order ID
        my $table = <FILE>;
        chop $table;

        my $workOrderID = <FILE>;
        chop $workOrderID;

        # before being able to insert into the geometric or geodetic tables,
        # we need to look up the work order common ID from the work order ID
        $sql = qq{
            SELECT geometric_wo_common_id from geometric_work_order_common
                WHERE work_order_id='$workOrderID'
        };
        &executeSql($sql);

        my ($commonWorkOrderID) = $sth->fetchrow_array();
        if (not defined($commonWorkOrderID))
        {
            print "Error finding common work order ID for work order ",
                "ID $workOrderID\n";
            exit FAILURE;
        }

        # now insert the new row into the geometric or geodetic table from
        # values in the file
        $sql = "INSERT into $table (GEOMETRIC_WO_COMMON_ID,";

        # parse the file, adding components of the insert statement to the
        # SQL string; field names are found first in the file, each on a new
        # line; VALUES keyword is next; then the values are at the end; format
        # of the file is specific and must not be changed without adjusting
        # this code for any format modifications (and vice versa)
        while (<FILE>)
        {
            chop;
            if (/VALUES/)
            {
                chop $sql; # remove last comma
                $sql .= " $_ ($commonWorkOrderID,";
            }
            else
            {
                $sql .= "$_,";
            }
        }
        chop $sql; # drop trailing comma
        close FILE;

        $rowsInserted = $dbh->do($sql);
        if ($rowsInserted != 1)
        {
            print "Error inserting record: $sql\n";
            exit FAILURE;
        }
    }
}
