package ProductDeliveries;

###############################################################################
# Name: ProductDeliveries
#
# Description: This class is used to start new product deliveries, track current
#              deliveries, and process completed deliveries.
#
###############################################################################

use strict;
use warnings;
use ias_const;
use ias_StartProcess;
use ias_UpdateWOState;
use ias_put_timing_event;
use ias_processing_system;
use ias_db;
use ias_MessageIds;
use ias_wo_const;
use POSIX ":sys_wait_h";

use constant PROD_DELIVERY_UP   => "UP";
use constant PROG => "ProductDeliveries";

sub ProductDeliveries::new
{
    my $class = shift; # Class name

    # Add the dictionary of product deliveries
    my $self = {product_deliveries => {}};

    # Create the product deliveries object and return it
    bless $self, $class;
    return $self;
}

###############################################################################
# NAME: runSql
#
# DESCRIPTION: This private function executes an SQL statement and returns the
#              statement handle. Caller must do the commit.
#
# PARAMETERS:
#              database handle
#              an sql statement string
#              error message reference for returning error messages
#
# RETURNS:
#          the statement handle (defined or undefined)
#
###############################################################################
sub runSql($$$)
{
    my ($dbh, $sql, $errorRef) = @_;

    my $sth = $dbh->prepare($sql);
    if (not $sth)
    {
        $$errorRef = "LINE: " . __LINE__ . " SQL PREPARE ERROR : $DBI::errstr";
        undef($sth);
    }
    elsif (not $sth->execute())
    {
        $$errorRef = "LINE: " . __LINE__ . " SQL EXECUTE ERROR : $DBI::errstr";
        $sth->finish();
        undef($sth);
    }

    return $sth;
}

###############################################################################
# NAME: processReadyToShipWos
#
# DESCRIPTION: This method will get a list of all work orders that are
#              ready to be shipped and start a transfer_product script to
#              deliver the work order.
#
# PARAMETERS:
#              database handle
#              directory path to work orders directory
#              the state of product delivery
#              the limit of transfer_product processes to run concurrently
#              the host requesting the delivery
#              enabled instances array reference
#              reference to Logger object
#
# RETURNS:
#          SUCCESS
#          ERROR
#
###############################################################################
sub ProductDeliveries::processReadyToShipWos($$$$$$$)
{
    my ($self, $dbh, $woDirPath, $prodDeliveryState, $prodTxfrLimit,
        $host, $enabled_instances, $error_log) = @_;
    my $errorString = '';

    # Create the enabled instances query addition
    my $instances_list = "'" . join("','", @{$enabled_instances}) . "'";

    # Get completed work orders
    my $sql = "   SELECT work_order_id, customer_order_nbr, "
            . "          product_request_id, scene_id, processing_instance "
            . "    FROM work_orders "
            . "    WHERE state in ('" . DB_WO_READY_TO_SHIP . "') "
            . "    AND processing_instance in (" . $instances_list . ") "
            . "    AND wo_replaced = 0 and processing_node = '$host' "
            . "    ORDER BY priority ASC, date_entered ASC";

    # Execute the query
    my $sth = runSql($dbh, $sql, \$errorString);
    if (not defined $sth)
    {
        $error_log->fatal_msg("Retrieving ready to ship WO list " .
            ": $errorString");
        return ERROR;
    }

    # Start the delivery on products that haven't been started yet
    while (my $dbRowHref = $sth->fetchrow_hashref())
    {
        my %woData = (WoId => $dbRowHref->{WORK_ORDER_ID},
                      TramOrderNbr => $dbRowHref->{CUSTOMER_ORDER_NBR},
                      ProdReqId => $dbRowHref->{PRODUCT_REQUEST_ID},
                      ProcessingInstance => $dbRowHref->{PROCESSING_INSTANCE});

        # Make sure to not try to deliver a manual work order product
        if (defined $woData{TramOrderNbr})
        {
            # Start delivery only if delivery for this work order has not
            # already been started and under the transfers process limit
            if ((not exists $self->{product_deliveries}{$woData{WoId}})
                and (scalar keys %{$self->{product_deliveries}} <
                $prodTxfrLimit))
            {
                my $woPid = 0;

                # Start the delivery
                if ($self->startProductDelivery($dbh, $woDirPath,
                    $woData{WoId}, \$woPid, $prodDeliveryState,
                    $woData{ProcessingInstance}, \$errorString, $error_log)
                    != SUCCESS)
                {
                    $error_log->fatal_msg("Starting delivery : $errorString");
                }
                # Save the process info if product delivery was started
                # successfully
                elsif ($woPid > 0)
                {
                    # Create a hash for data we want to know when
                    # product transfer is done.
                    # Add this work order to the hash for tracking the
                    # transfer_product processes
                    my %prodTxfrWO = ();
                    $prodTxfrWO{pid} = $woPid;
                    $prodTxfrWO{prod_req_id} = $woData{ProdReqId};
                    $prodTxfrWO{scene_id} =$dbRowHref->{SCENE_ID};
                    $self->{product_deliveries}{$woData{WoId}} = \%prodTxfrWO;
                }
            }
        }
        else
        {
            # Delivery is not done for IAS orders or manual LPGS orders.
            # Advance the work order state to indicate the work order is
            # ready to be completed.
            my $return_msg = "";

            # Start a transaction
            $dbh->begin_work();

            if (!ias_UpdateWOState($dbh, $woData{WoId}, DB_WO_COMPLETE_WO,
                \$return_msg))
            {
                $error_log->fatal_msg("Updating the WO state for completed " .
                    "work order " . $woData{WoId});

                # rollback any database changes
                $dbh->rollback();
            }
            elsif (not $dbh->commit())
            {
                $error_log->warn_msg("Committing start update work order " .
                    "for work order $woData{WoId}: $DBI::errstr");
            }
        }
    }
    $sth->finish();

    return SUCCESS;
}

###############################################################################
# NAME: startProductDelivery
#
# DESCRIPTION: This method takes care of the work to be done for a work
#              order in READY_TO_SHIP state. It checks if the product delivery
#              state indicates to deliver products to the online cache or not
#              and calls transfer_product, if needed. Each product delivery
#              needs to be done on a new background process. Capture the PID
#              and store it in the transfer_product script hash.
#
# PARAMETERS:
#              database handle
#              work order directory data path
#              the work order id
#              string reference for returning the process id
#              the state of product delivery
#              processing instance name
#              error message reference for returning error messages
#              reference to Logger object
#
# RETURNS:
#          SUCCESS
#          ERROR
#
###############################################################################
sub ProductDeliveries::startProductDelivery($$$$$$$)
{
    my ($self, $dbh, $woDirPath, $woId, $pidRef, $prodDeliveryState,
        $processing_instance, $errorRef, $error_log) = @_;

    # Deliver the data to the online cache if the product delivery state is UP
    if ($prodDeliveryState eq PROD_DELIVERY_UP)
    {
        # Start a transaction
        $dbh->begin_work();

        if (ias_put_timing_event($dbh, $woId, "START", "TRANSFER_PRODUCT")
            != SUCCESS)
        {
            # Writing to the events table in the database failed
            $error_log->fatal_msg("Logging timing event for work " .
                "order ID $woId event START component TRANSFER_PRODUCT " .
                "$ias_db_error");

            # Rollback the database
            $dbh->rollback();
            return ERROR;
        }

        # Build the parameter for transfer_product - the work order path
        my $woPath = "$woDirPath$woId";
        my @args = ($processing_instance, "transfer_product", $woPath);
        my $woPid = ias_StartProcess($dbh, "run_in_instance.csh", @args);

        if ($woPid > 0)
        {
            if (not $dbh->commit())
            {
                # The commit failed so the timing event is left uncommitted.
                # Proceed anyway since ias_StartProcess worked
                $error_log->warn_msg("Committing start transfer_product " .
                    "for work order $woId: $DBI::errstr");
            }
            $$pidRef = $woPid;
        }
        else
        {
            # Rollback the database for LPGS
            $dbh->rollback();

            # Log an event to the events table
            if (ias_db_PutEvent($dbh, PROG, IAS_MSG_PWS_TRANSFER_FAIL, $woId,
                '', "Can't start transfer_product for work order $woId",
                IAS_DB_COMMITEVENT) != 1)
            {
                # Writing to the events table in the database failed
                $error_log->warn_msg("Updating the events table");
            }

            $$errorRef = "Can't start transfer_product for work order $woId";
            return ERROR;
        }
    }

    return SUCCESS;
}

###############################################################################
# NAME: isTransferProductPid
#
# DESCRIPTION: This method checks whether a process id is a transfer_product
#              script. This method will either return TRUE/FALSE or if requested
#              it will return the work order id (FALSE if not a transfer_product
#              script).
#
# PARAMETERS:
#              tranfer_product process id
#              return work order id flag
#
# RETURNS: Empty String/FALSE or WORK_ORDER_ID/TRUE
#
###############################################################################
sub ProductDeliveries::isTransferProductPid($$)
{
    my ($self, $pid) = @_;

    # Loop over all the deliveries
    foreach my $key (keys %{$self->{product_deliveries}})
    {
        if ($pid == $self->{product_deliveries}{$key}{pid})
        {
            # Return the WO ID
            return $key, TRUE;
        }
    }

    return "", FALSE;
}

###############################################################################
# NAME: numberOfProductTransfers
#
# DESCRIPTION: This method returns the number of product transfers
#
# PARAMETERS: N/A
#
# RETURNS: Current number of product transfers
#
###############################################################################
sub ProductDeliveries::numberOfProductTransfers($)
{
    my $self = shift;

    return scalar keys %{$self->{product_deliveries}};
}

###############################################################################
# NAME: updateCompletedTransfer
#
# DESCRIPTION: This method updates a single completed transfer. If the
#              transfer_product script exited normally then the work order
#              gets completed otherwise the work order is put into the
#              TRANSFER_FAILED state.
#
# PARAMETERS:
#              database handle
#              tranfer_product process id
#              tranfer_product return code
#              reference to Logger object
#
# RETURNS: False: If not a transfer product script pid (PID Not handled)
#          True: If pid is a transfer product script (PID Handled)
#
###############################################################################
sub ProductDeliveries::updateCompletedTransfer($$$$$)
{
    my ($self, $dbh, $pid, $return_code, $error_log) = @_;
    my $error = FALSE;
    my $db_error = FALSE;
    my $msg = "";
    my $status_flag = DB_WO_COMPLETE_WO;

    # Make sure the PID is a transfer_product script
    my ($wo_id, $is_trans_pid)  = $self->isTransferProductPid($pid);
    if (!$is_trans_pid)
    {
        return FALSE;
    }

    # Check for error situations
    if (WIFSIGNALED($return_code))
    {
        $msg = "Product delivery of work order $wo_id killed by signal "
            . WTERMSIG($return_code);
        $error = TRUE;
    }
    elsif (WIFSTOPPED($return_code))
    {
        $msg = "Product delivery of work order $wo_id stopped ("
            . WSTOPSIG($return_code) . ")";
        $error = TRUE;
    }
    elsif (WIFEXITED($return_code))
    {
        if (WEXITSTATUS($return_code) != 0)
        {
            $msg = "Product delivery of work order $wo_id exited ("
                . WEXITSTATUS($return_code) . ")";
            $error = TRUE;
        }
    }

    # Start a transaction
    $dbh->begin_work();

    if ($error)
    {
        # Log the error that occurred
        $error_log->fatal_msg($msg);

        # Add SUCCESS timing event for the transfer script
        if (ias_put_timing_event($dbh, $wo_id, "ABORTED", "TRANSFER_PRODUCT")
            != SUCCESS)
        {
            # Writing to the events table in the database failed
            $error_log->fatal_msg("Logging timing event for work " .
                "order ID $wo_id event ABORT component TRANSFER_PRODUCT " .
                "$ias_db_error");

            # Rollback the database
            $dbh->rollback();
            $db_error = TRUE;
        }

        # Log an event to the events table
        if (!$db_error)
        {
            if (ias_db_PutEvent($dbh, PROG, IAS_MSG_PWS_TRANSFER_FAIL, $wo_id,
                '', "transfer_product failed for work order $wo_id",
                IAS_DB_COMMITEVENT) != 1)
            {
                # Writing to the events table in the database failed
                $error_log->warn_msg("Updating the events table for failed " .
                    "transfer_product");
            }
        }

        # Advance the work order state to indicate the failure
        $status_flag = DB_WO_TRANSFER_FAILED;
    }
    else
    {
        # Add SUCCESS timing event for the transfer script
        if (ias_put_timing_event($dbh, $wo_id, "SUCCESS", "TRANSFER_PRODUCT")
            != SUCCESS)
        {
            # Writing to the events table in the database failed
            $error_log->fatal_msg("Logging timing event for work " .
                "order ID $wo_id event SUCCESS component TRANSFER_PRODUCT " .
                "$ias_db_error");

            # Rollback the database
            $dbh->rollback();
            $db_error = TRUE;
        }
    }

    # Update the WO state
    if (!$db_error)
    {
        $msg = "";
        if (!ias_UpdateWOState($dbh, $wo_id, $status_flag, \$msg))
        {
            # rollback any database changes
            $dbh->rollback();
            $error_log->fatal_msg("Updating WO state in DB::" . $msg);
        }
        # commit the values to the database
        elsif (!$dbh->commit())
        {
            $error_log->warn_msg($DBI::errstr);
        }
    }

    # Output a message on the status of the delivery
    if (!$error && !$db_error)
    {
        $error_log->info_msg("Product delivery completed for Work Order " .
            $wo_id . ", Product Request ID " .
            $self->{product_deliveries}{$wo_id}{prod_req_id} . ", Scene ID " .
            $self->{product_deliveries}{$wo_id}{scene_id});
    }
    else
    {
        $error_log->info_msg("Product delivery failed for Work Order " .
            $wo_id . ", Product Request ID " .
            $self->{product_deliveries}{$wo_id}{prod_req_id} . ", Scene ID " .
            $self->{product_deliveries}{$wo_id}{scene_id} . ". Delivery " .
            "will be reattempted next cycle");
    }

    # Remove the work order from the processes structure. If there was an error
    # then the transfer will be retried next cycle.
    delete $self->{product_deliveries}{$wo_id};

    return TRUE;
}

1;
