package pcs_lib;

##############################################################################
#
# This module is used to provide constants and shared routines for the PCS.
#
##############################################################################

use strict;
use ias_const;
use ias_wo_const;
use Sys::Hostname;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);
use Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(
                pcs_ResetScriptStatusForNode
                pcs_ResetWONode
            );

#############################################################################
# Name: pcs_ResetScriptStatusForNode
#
# Description: Reset last running script for all work orders on a given node.
#
# Parameters: 
#   open database handle
#   processing node hostname for LPGS or empty string for IAS
#   system ID
#   reference for returning an error message
#
# Return: 
#   SUCCESS or ERROR
#############################################################################
sub pcs_ResetScriptStatusForNode($$$$)
{
    my ($dbh, $host, $system_id, $err_ref) = @_;

    my $sql = "UPDATE wo_scripts "
            . "   SET state = '" . DB_WO_SCRIPT_PENDING . "' "
            . " WHERE state = '" . DB_WO_SCRIPT_RUNNING . "' " 
            . "   AND work_order_id IN "
            . "       (SELECT work_order_id "
            . "          FROM work_orders " 
            . "         WHERE state = '" . DB_WO_RUNNING . "' "
            . "           AND work_order_id LIKE '$system_id%' ";
    if ($host ne '')
    {
        # Only reset scripts for the current node if LPGS
        $sql .= "           AND processing_node = '$host')";
    }
    else
    {
        # If no host provided, this is IAS, and all scripts get reset
        $sql .= ")";
    }

    my $rows = $dbh->do($sql);
    if (not defined $rows)
    {
        $$err_ref = "Resetting scripts status failed: $DBI::errstr";
        return ERROR;
    }

    return SUCCESS;
}

##############################################################################
#
# Name: pcs_ResetWONode
#
# Description: Unset the processing_node for any WO in a active processing 
#   state or in a restartable state.
#
# Parameters: 
#   open database handle
#   processing node hostname for LPGS or empty string for IAS
#   system ID
#   reference for returning an error message
#
# Return: 
#   SUCCESS or ERROR
#
##############################################################################
sub pcs_ResetWONode($$$$)
{
    my ($dbh, $host, $system_id, $err_ref) = @_;

    my $sql;
    my $rows;

    # Reset the wo_scripts status before resetting the work_orders table
    my $result = pcs_ResetScriptStatusForNode($dbh, $host, $system_id,
            $err_ref);
    if ($result == ERROR)
    {
        return ERROR;
    }

    if ($host ne '')
    {
        # Start a transaction
        $dbh->begin_work();

        # Transition the READY_TO_SHIP orders from DOWN nodes to the current
        # host
        if (transition_ready_to_ship_orders($dbh, $host, $system_id, $err_ref)
            != SUCCESS)
        {
            $dbh->rollback();
            return ERROR;
        }

        # Commit the transaction
        if (!$dbh->commit())
        {
            $$err_ref = "Reseting ready to ship work orders: $DBI::errstr";
            return ERROR;
        }

        # Start a transaction
        $dbh->begin_work();

        # This is the LPGS environment, with multiple processing nodes,
        # so need to reset the assigned processing node
        $sql = "UPDATE work_orders "
            . "   SET processing_node = NULL "
            . " WHERE processing_node = '$host' "
            . "   AND work_order_id LIKE '$system_id%' "
            . "   AND state IN ('" . DB_WO_PENDING . "', "
            . "                 '" . DB_WO_RUNNING . "', "
            . "                 '" . DB_WO_HALTED . "', "
            . "                 '" . DB_WO_INACTIVE . "')";

        $rows = $dbh->do($sql);
        if (not defined $rows)
        {
            $$err_ref = "Resetting processing_node failed: $DBI::errstr";
            $dbh->rollback();
            return ERROR;
        }

        # Commit the transaction
        if (!$dbh->commit())
        {
            $$err_ref = "Commiting the processing node reset: $DBI::errstr";
            return ERROR;
        }
    }

    # Start a transaction
    $dbh->begin_work();

    # Reset any running work orders that were unassigned from this node
    # for LPGS, or all running work orders for IAS
    $sql = "UPDATE work_orders "
         . "   SET state = '" . DB_WO_PENDING . "', "
         . "       date_updated = SYSDATE "
         . " WHERE state = '" . DB_WO_RUNNING . "' "
         . "   AND work_order_id LIKE '$system_id%' ";
    if ($host ne '')
    {
         $sql .= "   AND processing_node IS NULL";
    }

    $rows = $dbh->do($sql);
    if (not defined $rows)
    {
        $$err_ref = "Resetting work order state failed: $DBI::errstr";
        $dbh->rollback();
        return ERROR;
    }

    # Commit the transaction
    if (!$dbh->commit())
    {
        $$err_ref = "Commiting the work order state reset: $DBI::errstr";
        return ERROR;
    }

    return SUCCESS;
}

##############################################################################
#
# Name: transition_ready_to_ship_orders
#
# Description: This function will transition any work orders in the
#              READY_TO_SHIP state that reside on DOWN nodes to the node
#              currently running the subsetter_interface script.
#
# Parameters: Database handler
#             Current processing node
#             system ID
#             Reference to return error code
#
# Returns: SUCCESS or ERROR
#
##############################################################################
sub transition_ready_to_ship_orders ($$$$)
{
    my ($dbh, $node, $system_id, $error_ref) = @_;
    my $hostname = hostname;
    my @work_order_ids; # Array of all work orders in the ready to ship state
                        # on down nodes

    # Update the processing node to the current node running PSI
    my $sql = "Update WORK_ORDERS set processing_node = '$hostname' where " .
        "processing_node = '$node' and work_order_id LIKE '$system_id%' and " .
        "state = '" . DB_WO_READY_TO_SHIP . "'";

    my $sth = $dbh->prepare($sql);
    if (not defined $sth)
    {
        $$error_ref = "Unable prepare update statement for " .
            "work orders $DBI::errstr";
        $dbh->rollback();
        return ERROR;
    }
    if(!$sth->execute())
    {
        $$error_ref = "Unable to update the processing nodes";
        $sth->finish();
        $dbh->rollback();
        return ERROR;
    }
    $sth->finish();

    return SUCCESS;
}

1;
