#!/usr/bin/perl
##############################################################################
#        NAME: shutdown_ias.pl
#
# DESCRIPTION: This application provides a command line tool for shutting down
#              the IAS/LPGS system.
#
#     RETURNS: SUCCESS or FAILURE
#
##############################################################################

# System Includes
use strict;
use warnings;
use Getopt::Long;
use DBI;
use File::Basename;

# IAS Includes
use Logger;
use ias_const;
use ias_db;

# Decvlare Constants
use constant PROG => basename($0);

# Setup to auto flush STDOUT
local $| = 1;

# =================== Begin main program =====================================
# Set up logger
my $logger = Logger->new("STDOUT");
if (!$logger)
{
    print PROG . ": ERROR: Failed to establish logger object.\n";
    exit FAILURE;
}
set_default_log($logger);
$logger->set_module_name(PROG);

# Command line argument variables
my $arg_immediately;
my $arg_gracefully;
my $arg_help;

# Retrieve the command line parameters
my %optctl = (
    immediately => \$arg_immediately,
    gracefully => \$arg_gracefully,
    help => \$arg_help
    );

my $opt_status = GetOptions (\%optctl,
     "immediately",
     "gracefully",
     "help"
     );

# Test for usage
if ($opt_status == 0 or $arg_help
    or (not $arg_immediately and not $arg_gracefully))
{
    usage();
    $logger->close_log();
    exit FAILURE;
}

# Execute the main routine
my $status = &_main;

if ($status != SUCCESS)
{
    $logger->close_log();
    exit FAILURE;
}

$logger->close_log();
exit SUCCESS;
# =================== End main program =======================================

##############################################################################
#        NAME: usage
#
# DESCRIPTION: Display the command line options and general description of the
#              program.
#
#     RETURNS: None
#
##############################################################################
sub usage
{
    print << "HELP";
  USAGE:
  $0 <directive> [-help]

  -immediately  - Directive to immediately shutdown the system
  -gracefully   - Directive to gracefully shutdown the system
  -help         - Display this usage message.  No processing will be
                  performed (optional)

  Example:
    shutdown_ias.pl -gracefully

    This command will place a directive into the system causing the IAS
    system to shutdown gracefully.
HELP
}

##############################################################################
#        NAME: _main
#
# DESCRIPTION: Inserts a shutdown command into the ipc_directives table in the
#              database to shutdown the IAS system.
#
#     RETURNS: SUCCESS or ERROR
#
##############################################################################
sub _main
{
    # Verify the USER and IAS_SYSTEM_USER_ID match
    my $system_user_id = $ENV{IAS_SYSTEM_USER_ID};
    my $session_user = $ENV{USER};
    if (not defined $system_user_id or not defined $session_user)
    {
        $logger->fatal_msg("Environment not setup");
        return ERROR;
    }
    if (not ($system_user_id eq $session_user))
    {
        $logger->fatal_msg("$session_user not authorized to shutdown system");
        return ERROR;
    }

    # Get the username and password for the database
    my ($user_id, $password) = ias_db_GetDBConnectionInfo();

    # Connect to the database
    my $dbh = ias_db_ConnectToDB($user_id, $password);
    if (not defined $dbh)
    {
        $logger->fatal_msg($ias_db_error);
        return ERROR;
    }

    # read the configuration for the IAS
    my $configReference = ias_db_GetIASConfig($dbh);
    if (not defined $configReference)
    {
        $logger->fatal_msg("Error in getting configuration information:: "
            . $ias_db_error);
        ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_COMMIT);
        return ERROR;
    }
    my %config = %{$configReference};

    # Check and see if we already have a shutdown directive present
    my $sql = "SELECT count(*)" .
              " FROM ipc_directives" .
              " WHERE directive LIKE 'shutdown%'" .
              " AND SYSTEM_ID = '" . $config{SystemId} . "'";

    my $count = $dbh->selectrow_array($sql);

    if ($count > 0)
    {
        $logger->warn_msg("A shutdown directive is already present");
        ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_COMMIT);
        return SUCCESS;
    }

    # Insert the correct directive
    $sql = "INSERT INTO ipc_directives (recipient, directive, system_id) ";
    if ($arg_immediately)
    {
        $sql .= "VALUES ('PSI', 'shutdown immediately', '";
    }
    else
    {
        $sql .= "VALUES ('PSI', 'shutdown gracefully', '";
    }
    $sql .= $config{SystemId} . "')";

    my $sth = ias_db_execute_sql($dbh, $sql);
    if (not defined $sth)
    {
        $logger->fatal_msg("Failed to insert directive:: " . $ias_db_error);
        ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_COMMIT);
        return ERROR;
    }
    $sth->finish();

    # Commit the values to the database
    if (!$dbh->commit())
    {
        $logger->fatal_msg($DBI::errstr);
        ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_COMMIT);
        return ERROR;
    }

    # Disconnect from the database
    my $status = ias_db_DisconnectFromDB($dbh, IAS_DB_DISCONNECT_COMMIT);
    if (!$status)
    {
        $logger->warn_msg($ias_db_error);
    }

    return SUCCESS;
}

