#!/usr/bin/perl -w
############################################################################
#   NAME:  PWGtest
#
#   PURPOSE: The PWG (PCS Work Order Generator) test will verify that a
#           work order can be submitted through the PWG call.
#
#   PARAMETERS: 1) Scene ID (Optional)
#               2) Product Type (Optional)
#               3) Delete Flag (Optional)
#
#   OUTPUT: The work order database updated with a work order
#           Error log messages
#           Process events messages
#
#   NOTE:   All parameters are optional.  The user should be able to run
#           this routine with or without parameters for it be successful.
#
############################################################################

use strict;
use Getopt::Long;
use File::Path;
use Logger;

# include global routines

use ias_const;
use ias_processing_system;
use ias_db;

# declare some constants
use constant PROG       => "PWGtest";
# debug flag is a method of having select information printed to standard out
# set to 1 to see debug output, set to 0 shut off debug output
use constant DEBUG      =>  0;      # Debug flag

#############################################################################
# Variable declarations
#############################################################################
my $sql;                # SQL statement
my $scene_id;           # Scene ID for the product

my $ias_system_id;      # IAS_SYSTEM_ID environment variable value
my $help;               # Validate parameters, print usage and exit
my $rtn;                # Return value from subroutines
my $delete_flag_str;    # String entered by user as a parameter
my $delete_flag;        # True/False to delete database entries
my $error_log;          # The logger object to use for log messages.

# initialize the ias_ProcSys module
if(ias_init_system() == ERROR)
{
    print STDERR PROG . "$ias_error_msg\n";

    exit FAILURE;
}

my $product_id;

# retrieve the optional command-line parameters
my %optctl = (
    scene_id => \$scene_id,
    product_id => \$product_id,
    delete_flag => \$delete_flag_str,
    help => \$help
    );

my $opt_status = GetOptions (\%optctl,
     "scene_id=s",
     "product_id=s",
     "delete_flag=s",
     "help",
     );

# initialize the logger object
$error_log = Logger->new($ias_log_filename);
if(!$error_log)
{
    print STDERR PROG . ": ERROR: Failed to establish logger object.\n";

    exit FAILURE;
}

set_default_log($error_log);
$error_log->set_module_name(PROG);

if ($opt_status == 0)
{
    usage();
    $error_log->fatal_msg(__FILE__, __LINE__,
                        "Unable to process arguments");
    $error_log->close_log();
    exit FAILURE;
}

if (DEBUG)
{
    print "SCENE_ID: $scene_id\n";
    print "PRODUCT ID: $product_id\n";
    print "DELETE FLAG: $delete_flag_str\n";
    print "HELP: $help\n";
}

if ($help)
{
    usage();
    $error_log->info_msg(__FILE__, __LINE__,
                        "Usage help completed");
    $error_log->close_log();
    exit SUCCESS;
}

# Check to see whether the IAS_SYSTEM_ID environment variable has been set.
# We can't assume that the IAS and/or LPGS setup scripts have been sourced
# prior to this point, and theoretically, this test should be runnable
# independent of whether they have been sourced.  If IAS_SYSTEM_ID isn't
# defined, just default to the IAS value for the duration of this test.
$ias_system_id = $ENV{'IAS_SYSTEM_ID'};
if (not defined($ias_system_id))
{
    $error_log->info_msg(__FILE__, __LINE__,
        "'IAS_SYSTEM_ID' environment variable not set, defaulting to IAS");
    print
        "'IAS_SYSTEM_ID' environment variable not set, defaulting to IAS\n";
    $ENV{'IAS_SYSTEM_ID'} = 'I';
}

# Testing PCS requires access to the database.  If the environment variable
# ENABLE_DB_TESTS is not defined, skip this test and return a successful status.
if (!defined $ENV{ENABLE_DB_TESTS})
{
    print "\n";
    print __FILE__ . ": Tests not performed, database tests are disabled.\n";
    print "\n";
    $error_log->info_msg(__FILE__, __LINE__,
       	"Tests not performed, database test are disabled.");
    $error_log->close_log();
    exit SUCCESS;
}

# define list of environment variables needed to run the application
my @vars_list = qw (
    ORACLE_SID
    TWO_TASK
    IAS_DB_COM
    );

# verify that the environment variables are defined
if (ias_validate_environment(\@vars_list))
{
    $error_log->fatal_msg(__FILE__, __LINE__,
                        "Unable to retrieve all environment variables.");
    $error_log->fatal_msg(__FILE__, __LINE__,
                        "PWGtest inserting work order has ended");
    $error_log->close_log();
    exit FAILURE;
}

my %args;

# connect to the database before creating the work order
# get the username and password for the sensor database
my ($user_id, $password) = ias_db_GetDBConnectionInfo();

# connect to the database
$args{'dbh'} = ias_db_ConnectToDB($user_id, $password);
if (!$args{'dbh'})
{
    $error_log->fatal_msg(__FILE__, __LINE__,
                            "Error connecting to LDCM database");
    $error_log->fatal_msg(__FILE__, __LINE__,
                            "PWGtest inserting work order has ended");
    $error_log->close_log();
    exit FAILURE;
}

# set up input parameters
if (!$scene_id)
{
   $scene_id = "LC8PWGtest";
}

# set delete flag according to input parameter
if (uc($delete_flag_str) eq "FALSE")
{
    $delete_flag = FALSE;
}
else
{
    $delete_flag = TRUE;
}

# make a unique comment by attaching a time stamp
my $wo_comment;
my ($sec, $min, $hour, $mday, $mon, $year, $wday, 
                $yday, $isdst) = localtime(time);
$year += 1900;
$wo_comment = "PWGTest_$year\_$yday\_$hour\_$min\_$sec";

# get a procedure at random.  This is assuming all procedures are
# able to be entered into the database using only defaults.
if (!$product_id)
{
    $product_id = get_wo_product($args{'dbh'});
    if (!defined($product_id))
    {
        $error_log->fatal_msg(__FILE__, __LINE__,
                  "End PWG test, error retrieving product ID from database");
        $args{'dbh'}->disconnect;
        $error_log->close_log();
        exit FAILURE;
    }
}

# call PWG 
$rtn = 
  system("../PWG -scene_id \"$scene_id\" -procedure_id \"$product_id\" -comment \"$wo_comment\"");
if ($rtn != SUCCESS)
{
    $error_log->fatal_msg(__FILE__, __LINE__,
                  "An Error returned from PWG call");
    $error_log->close_log();
    $args{'dbh'}->disconnect;
    exit FAILURE;
}


# delete the Work Order from the database
if ($delete_flag)
{
    $rtn = delete_work_order($args{'dbh'}, $wo_comment);
    if ($rtn == ERROR)
    {
        $error_log->fatal_msg(__FILE__, __LINE__,
                        "Error deleting the test work order");
        $error_log->close_log();
        exit FAILURE;
    }
}

$args{'dbh'}->disconnect;

print "PWGtest completed successfully\n";
$error_log->info_msg(__FILE__, __LINE__,
            "Test for PWG completed successfully");
exit SUCCESS;

##############################################################################
#               Subroutines defined here                                     #
##############################################################################

    #   #   #   #   #   #   #   #   #   #   #   #   #   #   #   #   #   #
##############################################################################
# Name: get_wo_product
#
# Purpose: Get a work order procedure to run PWG
#
##############################################################################
sub get_wo_product
{
    my ($dbh) = @_;

    my $wo_sth;

    # retrieve the database for the common schema
    my ($dbuser, $db_password) = ias_db_GetDBConnectionInfo();

    my $sql =
       qq{SELECT module_id FROM modules WHERE mod_type='PROCEDURE'};

    if (! ($wo_sth = $dbh->prepare($sql)))
    {
        $error_log->fatal_msg(__FILE__, __LINE__,
                    "Unable to prepare the select procedure name");
        $error_log->close_log();
        return undef;
    }

    # execute the query
    if (! ($wo_sth->execute()))
    {
        $error_log->fatal_msg(__FILE__, __LINE__,
                      "Unable to run select for procedure name");
        $error_log->close_log();
        $wo_sth->finish;
        return undef;
    }

    my $procedure_id = $wo_sth->fetchrow_array;

    return $procedure_id;
}

##############################################################################
# Name: delete_work_order
#
# Purpose: Delete the work order in this test
#
##############################################################################
sub delete_work_order
{
    my ($dbh, $wo_comment) = @_;
    my $wo_sth;

    # delete the test work order from the work_orders table
    my $sql = qq{DELETE from work_orders WHERE wo_comment='$wo_comment'};

    if (! ($wo_sth = $dbh->prepare($sql)))
    {
        $error_log->fatal_msg(__FILE__, __LINE__,
                    "Unable to prepare the work order delete statement");
        $error_log->close_log();
        $wo_sth->finish;
        $dbh->rollback;
        return ERROR;
    }

    # execute the work_orders delete
    if (! ($wo_sth->execute()))
    {
        $error_log->fatal_msg(__FILE__, __LINE__,
                      "Unable to delete the test work order");
        $error_log->close_log();
        $wo_sth->finish;
        $dbh->rollback;
        return ERROR;
    }

    my $num_rows_deleted = $wo_sth->rows;

    $wo_sth->finish;

    if ($num_rows_deleted < 1)
    {
        $error_log->fatal_msg(__FILE__, __LINE__,
            "Did not find the work order to delete with comment $wo_comment");
        $error_log->close_log();
        $dbh->rollback;
        return ERROR;
    }

    return SUCCESS;
}

##############################################################################
# Name: usage
#
# Purpose: Display command line options and general description of program
#
##############################################################################
sub usage
{
    print << "HELP";
    USAGE:
    $0 [-scene_id <scene ID> -product_id <product type>
            -delete_flag <TRUE/FALSE> -help]

    -scene_id       - The ID of the scene to be processed (optional)
    -product_type   - The name of the product ID to process (optional)
    -delete_flag    - Flag indicting that the work order should or should
                      not be deleted.  Valid values are TRUE, the work
                      order should be deleted or FALSE, do not delete
                      the work order.  Defaults to TRUE (optional)
    -help           - Display this usage message.  No processing will be
                      performed (optional)

    PWGtest.pl is a command line interface to the PWG test.  This routine
    ensures that the PWG can enter a work order in the database.  The 
    routine is designed to run with no arguments.  It will read the 
    database to retrieve a valid product type and enter a work order 
    using the default parameters.  The user may optionally enter a 
    scene ID, product type, and delete flag.

    This routine will clean up the work order upon completion if
    delete_flag is set to TRUE, which is the default.  Since PWG does 
    not return a work order ID, the comment is used to search for the new 
    work order in the database.  The comment will be assigned a unique
    number based on year/day of year/hour/min/sec.

    Example:
    PWGtest.pl -delete_flag False

    This command will enter a work order in the database and will
    not delete the work order upon completion given the delete flag
    is set to false.  Notice that the flag is not case sensitive.

HELP
}

