#!/usr/bin/perl -w
############################################################################
#   NAME:  createwotest
#
#   PURPOSE: The create work order test (createwotest) will verify that a
#           work order can be written to the work order database tables
#           with expected results by calling the ias_create_work_order
#           library routine.
#
#   PARAMETERS: None
#
#   OUTPUT: The work order database updated with a work order
#           Error log messages
#           Process events messages
#
#   NOTE:   All parameters are optional.  If the user enters no parameters,
#           the output files should be identical and this program will
#           return successful.  In order to give this routine flexibility,
#           The user is allowed to enter parameters.  The work order entered
#           will reflect the entered parameters but will likely cause 
#           an error upon return.
#
############################################################################

use strict;
use Getopt::Long;
use Logger;

# include global routines
use ias_create_work_order qw
{
    ias_create_work_order
};

use ias_const;
use ias_processing_system;
use ias_db;

# declare some constants
use constant PROG       => "createwotest";

# debug flag is a method of having select information printed to standard out
# set to 1 to see debug output, set to 0 shut off debug output
use constant DEBUG      =>  0;      # Debug flag

#############################################################################
# Variable declarations
#############################################################################
my $sql;                # SQL statement
my $wo_id;              # Subroutine return value
my $scene_id;           # Scene ID for the procedure
my $procedure_id;       # Procedure ID for work order

my $ias_system_id;      # IAS_SYSTEM_ID environment variable value
my $help;               # Validate parameters, print usage and exit
my $wo_comment = "";    # Comment to be attached to the work order
my $rtn;                # Return value from subroutines
my $delete_flag_str;    # String entered by user as a parameter
my $delete_flag;        # True/False to delete database entries
my $g_Logger;           # The logger object to use for log messages.

# retrieve the optional command-line parameters
my %optctl = (
    delete_flag => \$delete_flag_str,
    help => \$help
    );

my $opt_status = GetOptions (\%optctl,
     "delete_flag=s",
     "help",
     );

# initialize the ias_ProcSys module
if(ias_init_system() == ERROR)
{
    print STDERR PROG . "$ias_error_msg\n";
    exit FAILURE;
}

# initialize the logger object
$g_Logger = Logger->new($ias_log_filename);
if(!$g_Logger)
{
    print STDERR PROG . ": ERROR: Failed to establish logger object.\n";

    exit FAILURE;
}

set_default_log($g_Logger);
$g_Logger->set_module_name(PROG);

if ($opt_status == 0)
{
    usage();
    $g_Logger->fatal_msg(__FILE__, __LINE__, "Unable to process arguments");
    $g_Logger->close_log();
    exit FAILURE;
}

if (DEBUG)
{
    print "DELETE FLAG: $delete_flag_str\n";
    print "HELP: $help\n";
}

if ($help)
{
    usage();
    $g_Logger->info_msg(__FILE__, __LINE__,
                        "Usage help completed");
    $g_Logger->close_log();
    exit SUCCESS;
}

# Check to see whether the IAS_SYSTEM_ID environment variable has been set
# (it's set when 'ias/ias_setup' or 'lpgs/lpgs_setup' is sourced).
# We can't assume that the IAS and/or LPGS setup scripts have been sourced
# prior to this point, and theoretically, this test should be runnable
# independent of whether they have been sourced.  If IAS_SYSTEM_ID isn't
# defined, just default to the IAS value for the duration of this test.
$ias_system_id = $ENV{'IAS_SYSTEM_ID'};
if (not defined($ias_system_id))
{
    $g_Logger->info_msg(__FILE__, __LINE__,
        "'IAS_SYSTEM_ID' environment variable not set, defaulting to IAS");
    print
        "'IAS_SYSTEM_ID' environment variable not set, defaulting to IAS\n";
    $ENV{'IAS_SYSTEM_ID'} = 'I';
}

# Testing PCS requires access to the database.  If the environment variable
# ENABLE_DB_TESTS is not defined, skip this test and return a successful status.
if (!defined $ENV{ENABLE_DB_TESTS})
{
    print "\n";
    print __FILE__ . ": Tests not performed, database tests are disabled.\n";
    print "\n";
    $g_Logger->info_msg(__FILE__, __LINE__,
       	"Tests not performed, database test are disabled.");
    $g_Logger->close_log();
    exit SUCCESS;
}

# define list of environment variables needed to run the application
my @vars_list = qw (
    ORACLE_SID
    TWO_TASK
    );

# verify that the environment variables are defined
if (ias_validate_environment(\@vars_list))
{
    $g_Logger->fatal_msg(__FILE__, __LINE__,
                        "Unable to retrieve all environment variables.");
    $g_Logger->fatal_msg(__FILE__, __LINE__,
                        "createwotest has ended");
    $g_Logger->close_log();
    exit FAILURE;
}

my %args;

# connect to the database before creating the work order
# get the username and password for the sensor database
my ($user_id, $password) = ias_db_GetDBConnectionInfo();

# connect to the common database
my $dbhcommon = ias_db_ConnectToDB($user_id, $password);
if (!$dbhcommon)
{
    $g_Logger->fatal_msg(__FILE__, __LINE__,
                            "Error connecting to LDCM database");
    $g_Logger->fatal_msg(__FILE__, __LINE__,
                            "createwotest has ended");
    $g_Logger->close_log();
    exit FAILURE;
}

# Cleanup just incase anything was left behind
cleanup_db($dbhcommon);

# Load the testing procedures
my $sys_return = system("./load_test_schema.sh");
if ($sys_return != 0)
{
    $g_Logger->fatal_msg(__FILE__, __LINE__,
                        "Error loading the test schemas");
    cleanup_db($dbhcommon);
    $g_Logger->close_log();
    $dbhcommon->disconnect;
    exit FAILURE;
}

# enter the default required parameters
$scene_id = "LC8PWGtest";
$procedure_id = "PWGtestProcedure";

# enter the default arguments into the hash
$args{'order_id'} = "PWGordertest";
$args{'unit_number'} = 1;
$args{'l0r_data_path'} = "test_l0r_data_path";
my @parms = qw (DSW:CAL_PARM_FILENAME=testcpf
        DSW:BIAS_PARM_FILENAME_OLI=testbpfoli
        DSW:BIAS_PARM_FILENAME_TIRS=testbpftirs
        DSW:RLUT_FILENAME=testrlut);
$args{'parms'} = \@parms;
$args{'wo_priority'} = 6;
$args{'out_db_schema'} = "E";
$args{'wo_comment'} = "PWG test work order";
$args{'requester'} = "testrequester";
$args{'processing_instance'} = "DummyCollection";

# set delete flag according to input parameter
if (uc($delete_flag_str) eq "FALSE")
{
    $delete_flag = FALSE;
}
else
{
    $delete_flag = TRUE;
}

# call ias_create_work_order to enter the test work order into the db
$wo_id = ias_create_work_order($procedure_id, $scene_id, %args);

if (DEBUG)
{
    print "work order ID is:$wo_id\n";
}

if (!defined($wo_id))
{
    $g_Logger->fatal_msg(__FILE__, __LINE__,
                               "PWG inserting work order has failed");
    cleanup_db($dbhcommon);
    $g_Logger->close_log();
    $dbhcommon->disconnect;
    exit FAILURE;
}

# write the new work order results from the database table to a file
my $filename = "PWGtestfile.tex";
$rtn = dump_work_order($wo_id, $dbhcommon, $filename);
if ($rtn == ERROR)
{
    $g_Logger->fatal_msg(__FILE__, __LINE__,
                                "Error writing the work order information");
    cleanup_db($dbhcommon);
    $g_Logger->close_log();
    $dbhcommon->disconnect;
    exit FAILURE;
}

# test for clean up
if ($delete_flag)
{
    if (cleanup_db($dbhcommon, $wo_id) != SUCCESS)
    {
        $g_Logger->fatal_msg(__FILE__, __LINE__,
            "Cleanup failed. Manually cleanup needed.");
        $g_Logger->close_log();
        $dbhcommon->disconnect;
        exit FAILURE;
    }
}
$dbhcommon->disconnect;

# print the difference
$rtn = system("diff PWGtestKey.tex $filename");

# if the files are identical the return value will be 0, otherwise return error
if ($rtn == 0)
{
    print "createwotest completed successfully, work order is $wo_id\n";
    $g_Logger->info_msg(__FILE__, __LINE__,
        "Test for ias_create_work_order was successful, work order is $wo_id");
    $g_Logger->close_log();
    cleanup_db($dbhcommon, $wo_id);
    exit $rtn;
}
else
{
    print "Test for ias_create_work_order failed, work order is $wo_id\n";
    $g_Logger->info_msg(__FILE__, __LINE__,
        "Test for ias_create_work_order contained error, work order is $wo_id");
    $g_Logger->close_log();
    cleanup_db($dbhcommon, $wo_id);
    exit FAILURE;
}

##############################################################################
#               Subroutines defined here                                     #
##############################################################################

    #   #   #   #   #   #   #   #   #   #   #   #   #   #   #   #   #   #
##############################################################################
# Name: dump_work_order
#
# Purpose: Dump the results of the new work order to a file
#
##############################################################################
sub dump_work_order
{
    my ($wo_id, $dbhcommon, $filename) = @_;
    my $wo_sth;

    open (TESTFILE, ">$filename");

    # select the row for this work order in the work orders table
    my $sql = qq{SELECT procedure_id, scene_id, satellite, sensor,
        customer_order_nbr, customer_unit_nbr, cpf_name, wo_comment, priority,
        l0r_data_path, l0r_data_status, output_db_schema, requester
        FROM work_orders WHERE work_order_id = '$wo_id'};

    if (! ($wo_sth = $dbhcommon->prepare($sql)))
    {
        $g_Logger->fatal_msg(__FILE__, __LINE__,
                    "Unable to prepare the select work order statement");
        $wo_sth->finish;
        $dbhcommon->rollback;
        return ERROR;
    }

    # execute the query
    if (! ($wo_sth->execute()))
    {
        $g_Logger->fatal_msg(__FILE__, __LINE__,
                      "Unable to run select from work_orders");
        $wo_sth->finish;
        $dbhcommon->rollback;
        return ERROR;
    }

    my ($procedure_id, $scene_id, $satellite, $sensor, $customer_order_nbr, 
        $customer_unit_nbr, $cpf_name, $wo_comment, $priority, $l0r_data_path,
        $l0r_data_status, $output_db_schema, $requester) =
        $wo_sth->fetchrow_array;

    print TESTFILE "WORK_ORDERS printout\n";
    print TESTFILE "procedure id, scene_id, satellite, sensor, ";
    print TESTFILE "customer_order_nbr,\n";
    print TESTFILE "customer_unit_nbr, ";
    print TESTFILE "cpf_name, wo_comment, priority, ";
    print TESTFILE "l0r_data_path,\n";
    print TESTFILE "l0r_data_status, output_db_schema, requester\n";
    print TESTFILE "$procedure_id, $scene_id, $satellite, $sensor, ";
    print TESTFILE "$customer_order_nbr, \n";
    print TESTFILE "$customer_unit_nbr, ";

    if (!defined($cpf_name))
    {
        $cpf_name = '';
    }

    if (!defined($wo_comment))
    {
        $wo_comment = '';
    }

    if (!defined($priority))
    {
        $priority = '';
    }

    if (!defined($l0r_data_path))
    {
        $l0r_data_path = '';
    }

    if (!defined($l0r_data_status))
    {
        $l0r_data_status = '';
    }

    print TESTFILE "$cpf_name, $wo_comment, $priority, ";
    print TESTFILE "$l0r_data_path,\n";
    print TESTFILE "$l0r_data_status, $output_db_schema, $requester\n\n";

    # select the rows for this work order in the work order scripts table

    $sql = qq{SELECT script_id, seq_nbr, state
        FROM wo_scripts WHERE work_order_id = '$wo_id'
        ORDER BY seq_nbr};

    if (! ($wo_sth = $dbhcommon->prepare($sql)))
    {
        $g_Logger->fatal_msg(__FILE__, __LINE__,
                    "Unable to prepare the select wo_scripts statement");
        $dbhcommon->rollback;
        return ERROR;
    }

    # execute the query
    if (! ($wo_sth->execute()))
    {
        $g_Logger->fatal_msg(__FILE__, __LINE__,
                      "Unable to run select from wo_scripts");
        $wo_sth->finish;
        $dbhcommon->rollback;
        return ERROR;
    }

    print TESTFILE "WO_SCRIPTS printout\n";
    print TESTFILE "script_id, seq_nbr, state\n";
    my $script_id; 
    my $seq_nbr; 
    my $state;
    while (($script_id, $seq_nbr, $state) = $wo_sth->fetchrow_array)
    {
        print TESTFILE "$script_id, $seq_nbr, $state\n";
    }

    print TESTFILE "\n";

    # select the rows for this work order in the work order parms table
    $sql = qq{SELECT script_id, parm_name, parm_occurrence, parm_value
        FROM wo_parms WHERE work_order_id = '$wo_id'
        ORDER BY script_id, parm_name, parm_occurrence};

    if (! ($wo_sth = $dbhcommon->prepare($sql)))
    {
        $g_Logger->fatal_msg(__FILE__, __LINE__,
                    "Unable to prepare the select wo_parms statement");
        $dbhcommon->rollback;
        return ERROR;
    }

    # execute the query
    if (! ($wo_sth->execute()))
    {
        $g_Logger->fatal_msg(__FILE__, __LINE__,
                      "Unable to run select from wo_parms");
        $dbhcommon->rollback;
        return ERROR;
    }

    print TESTFILE "WO_PARMS printout\n";
    print TESTFILE "script_id, parm_name, parm_occurrence, parm_value\n";
    my $parm_name;
    my $parm_occurrence;
    my $parm_value;
    while (($script_id, $parm_name, $parm_occurrence, $parm_value) = 
                                    $wo_sth->fetchrow_array)
    {
        if (!defined($parm_value))
        {
            $parm_value = '';
        }

        print TESTFILE 
            "$script_id, $parm_name, $parm_occurrence, $parm_value\n";
    }

    print TESTFILE "\n";

    # select the rows for this work order in the work order common_parms table
    $sql = qq{SELECT parm_name, parm_occurrence, parm_value
        FROM wo_parms WHERE work_order_id = '$wo_id'
        ORDER BY parm_name, parm_occurrence};

    if (! ($wo_sth = $dbhcommon->prepare($sql)))
    {
        $g_Logger->fatal_msg(__FILE__, __LINE__,
                    "Unable to prepare the select wo_common_parms statement");
        $dbhcommon->rollback;
        return ERROR;
    }

    # execute the query
    if (! ($wo_sth->execute()))
    {
        $g_Logger->fatal_msg(__FILE__, __LINE__,
                      "Unable to run select from wo_parms");
        $dbhcommon->rollback;
        return ERROR;
    }

    print TESTFILE "WO_COMMON_PARMS printout\n";
    print TESTFILE "parm_name, parm_occurrence, parm_value\n";
    while (($parm_name, $parm_occurrence, $parm_value) = 
                                    $wo_sth->fetchrow_array)
    {
        if (!defined($parm_value))
        {
            $parm_value = '';
        }
        print TESTFILE "$parm_name, $parm_occurrence, $parm_value\n";
    }

    # close the output file
    close (TESTFILE);

    return SUCCESS;
}

##############################################################################
# Name: cleanup_db
#
# Purpose: cleanup the database
##############################################################################
sub cleanup_db
{
    my $dbhcommon = shift;
    my $wo_id = shift;
    my $ret_status = SUCCESS;
    my $sql;

    $sql = qq{DELETE from parm_list WHERE parm_name like 'PWGtest%'};
    if (execute_delete($dbhcommon, $sql, "PARM_LIST") != SUCCESS)
    {
        $g_Logger->fatal_msg("Error deleting the PARM_LIST");
        $ret_status = ERROR;
    }

    $sql = qq{DELETE from parm_edits WHERE parm_name like 'PWGtest%'};
    if (execute_delete($dbhcommon, $sql, "PARM_EDITS") != SUCCESS)
    {
        $g_Logger->fatal_msg("Error deleting the PARM_EDITS");
        $ret_status = ERROR;
    }

    $sql = qq{DELETE from def_parms WHERE module_id like 'PWGtest%'};
    if (execute_delete($dbhcommon, $sql, "DEF_PARMS") != SUCCESS)
    {
        $g_Logger->fatal_msg("Error deleting the DEF_PARMS");
        $ret_status = ERROR;
    }

    $sql = qq{DELETE from parameters WHERE } .
           qq{processing_instance='DummyCollection'};
    if (execute_delete($dbhcommon, $sql, "PARAMETERS") != SUCCESS)
    {
        $g_Logger->fatal_msg("Error deleting the PARAMETERS");
        $ret_status = ERROR;
    }

    $sql = qq{DELETE FROM sub_modules WHERE module_id like 'PWGtest%'};
    if (execute_delete($dbhcommon, $sql, "SUB_MODULES") != SUCCESS)
    {
        $g_Logger->fatal_msg("Error deleting the SUB_MODULES");
        $ret_status = ERROR;
    }

    $sql = qq{DELETE FROM modules WHERE } .
           qq{processing_instance='DummyCollection'};
    if (execute_delete($dbhcommon, $sql, "MODULES") != SUCCESS)
    {
        $g_Logger->fatal_msg("Error deleting the MODULES");
        $ret_status = ERROR;
    }

    $sql = qq{DELETE from work_orders WHERE } .
           qq{product_request_id='PWGordertest_00001'};
    if (execute_delete($dbhcommon, $sql, "WORK_ORDERS") != SUCCESS)
    {
        $g_Logger->fatal_msg("Error deleting the WORK_ORDERS");
        $ret_status = ERROR;
    }

    # Delete the processing instance
    $sql = qq{DELETE FROM PROCESSING_INSTANCES WHERE } .
           qq{instance_name like 'Dummy%'};
    if (execute_delete($dbhcommon, $sql, "Processing_Instances") != SUCCESS)
    {
        $g_Logger->fatal_msg("Error deleting the processing instance");
        $ret_status = ERROR;
    }

    return $ret_status;
}

##############################################################################
# Name: execute_delete
#
# Purpose: Delete the Dummy instance in the database processing instance table
#
##############################################################################
sub execute_delete
{
    my ($dbh, $sql, $table) = @_;
    my $wo_sth;

    # delete the test procedure from the modules table
    if (!($wo_sth = $dbh->prepare($sql)))
    {
        $g_Logger->fatal_msg("Preparing the $table delete statement:".
            "$DBI::errstr");
        return ERROR;
    }

    # execute the module delete
    if (!($wo_sth->execute()))
    {

        $g_Logger->fatal_msg("Unable to delete from $table table:$DBI::errstr");
        return ERROR;
    }
    my $num_rows_deleted = $wo_sth->rows;
    $wo_sth->finish;

    if ($num_rows_deleted == 0)
    {
        $g_Logger->info_msg("Did not delete any rows for table $table");
    }

    $dbh->commit();

    return SUCCESS;
}

##############################################################################
# Name: usage
#
# Purpose: Display command line options and general description of program
#
##############################################################################
sub usage
{
    print << "HELP";
    USAGE:
    $0 [-delete_flag <FALSE/TRUE> -help]

    -delete_flag    - Flag indicating that the work order and procedure ID
                      should or should not be deleted.  Valid values 
                      are TRUE, the work order should be deleted 
                      or FALSE, do not delete the work order.  Defaults 
                      to TRUE and is not case sensitive (optional)
    -help           - Display this usage message.  No processing will be
                      performed (optional)

    createwotest is a command line interface to test the ias_create_work_order
    perl library.  This routine will enter a test procedure into the
    database, setup a work order for that procedure, read the work order
    in the database and write the results to a file, compare the new
    file with a control test file, report the difference.  If differences
    occur, this routine will return FAILURE, if the files are identical, 
    SUCCESS will be returned.

    This routine is intended to run with no arguments.

    Examples:
    createwotest.pl

    Expected usage.  If the work order is entered correctly, this should
    result in a successful test.  The new test procedure and work order will
    be deleted.

    createwotest.pl -delete "False"

    This command will generate a work order with the default parameters.  
    The new work order will not be deleted from the database.

HELP
}

