/*==============================================================================
NAME: Algorithm Functions

PURPOSE:
    This file contains functions specific to the individual Cloud Cover
    Assessment (CCA) algorithms.

ALGORITHM REFERENCES:
    CCA Algorithm Description Document (ADD)

    The algorithms for run_algorithm_acca(), run_algorithm_atacca(),
    run_algorithm_cirrus(), and run_algorithm_see5() were copied from the
    prototype code that was supplied with the ADD and have been reformatted and
    optimized for performance. 
   
    The cfmask algorithm was implemented based on the L7 Cfmask code. The
    detail algorithm is explained in ADD. The code has been reformatted and
    optimized for performance.

    There are numerous constants in this file.  Any questions about these
    constants should reference the ADD and those algorithms.
==============================================================================*/

/* System Includes */
#include <stdlib.h>
#include <math.h>
#include <string.h>

/* IAS Includes */
#include "ias_l1g.h"
#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "ias_angle_gen.h"

/* Local Includes */
#include "assess_cloud_cover.h"
#include "get_names.h"
#include "algorithm.h"
#include "cfmask_support.h"
#include "calculate_map_azimuth_angle.h"
#include "get_parameters.h"

#define MAX_BANDS_PER_ALGORITHM 8


/*------------------------------------------------------------------------------
NAME: check_bands

PURPOSE:
    For a single algorithm, check to see if the required bands are present in
    the input image.  If all the required bands are present, the bandFlags array
    will be updated to mark those bands as ones the application should open and
    read.  If one or more bands are missing, a warning message will be displayed
    and the bandFlags array will not be updated.

RETURNS:
    This function will return SUCCESS if all of the required bands are present
    and the algorithm could be run, ERROR if one or more bands are not present.
------------------------------------------------------------------------------*/
static int check_bands
(
    int algorithm,                             /* I: Current algorithm */
    L1GIO *fdImageIn,                          /* I: Input image file desc */
    int bandType2Number[IAS_NUM_BAND_TYPES],   /* I: Map band type to number */
    int requiredBands[MAX_BANDS_PER_ALGORITHM],/* I: Type of bands required by
                                                  algorithm */
    int bandFlags[IAS_NUM_BAND_TYPES]          /* IO: Bands to be opened */
)
{
    int  bandNumber;
    int  bandType;
    int  i;

    /* Examine all of the required bands for this algorithm and, if any one is
       missing, set the allPresent flag to FALSE and add the missing band's name
       to the message that will be printed. */
    for (i = 0; i < MAX_BANDS_PER_ALGORITHM; i++)
    {
        bandType = requiredBands[i];

        if (bandType == IAS_UNKNOWN_BAND_TYPE)
            continue;

        bandNumber = bandType2Number[bandType];

        if (!ias_l1g_is_band_present(fdImageIn, bandNumber))
        {
            IAS_LOG_WARNING("Algorithm %s skipped due to missing band: %d",
                get_algorithm_name(algorithm), bandNumber);
            return ERROR;
        }
    }

    /* Since all of the bands were found, update the bandFlags array to indicate
       which bands the application should open for processing */
    for (i = 0; i < MAX_BANDS_PER_ALGORITHM; i++)
    {
        bandType = requiredBands[i];
        if (bandType != IAS_UNKNOWN_BAND_TYPE)
            bandFlags[bandType] = TRUE;
    }
    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: eval_cubist_therm

PURPOSE:
    Compute an artificial thermal band value if TIRS data is not present.

RETURNS:
    An artificial thermal band sample

ALGORITHM REFERENCES:
    CCA ADD.  This algorithm was replicated from the prototype software supplied
    with that document.
------------------------------------------------------------------------------*/
static float eval_cubist_therm
(
    float b2,           /* I: Sample value from band 2 - Blue */
    float b3,           /* I: Sample value from band 3 - Green */
    float b4,           /* I: Sample value from band 4 - Red */
    float b5,           /* I: Sample value from band 5 - NIR */
    float b6,           /* I: Sample value from band 6 - SWIR1 */
    float b7,           /* I: Sample value from band 7 - SWIR2 */
    float sineSolarElev /* I: Sine of solar elevation in radians */
)
{
    float value;

    /* Model identification information:
           - Cubist rules from file cubat_4_non.model ...
           - Rule 1 - 1706670 cases, error 5.65972 (224.642 / 288.334 / 327.228)
     */
    value = -92.7 * ((b4 - b7) / (b4 + b7))
        + 261.4 * ((b3 - b7) / (b3 + b7))
        -  48.8 * ((b3 - b6) / (b3 + b6))
        -  17.5 * ((b5 - b3) / (b5 + b3))
        - 146.9 * ((b2 - b7) / (b2 + b7))
        +  58.7 * ((b4 - b2) / (b4 + b2))
        - 117.0 * ((b3 - b2) / (b3 + b2))
        + 172.0 * sineSolarElev * b6
        +  76.0 * sineSolarElev * b5
        + 151.0 * sineSolarElev * b4
        - 951.0 * sineSolarElev * b3
        + 539.0 * sineSolarElev * b2
        +  28.0 * b7
        - 132.0 * b6
        - 106.2 * b5
        -  22.4 * b4
        + 633.1 * b3
        - 443.6 * b2
        + 302.0986;
    return value;
}

/*------------------------------------------------------------------------------
NAME: count_bits_set

PURPOSE: Counts the number of bits set in an unsigned character.

RETURNS: Number of bits set in the input byte
------------------------------------------------------------------------------*/
static unsigned char count_bits_set
(
    unsigned char value
)
{
    unsigned char count;

    /* Loop until all the bits in value are cleared */
    for (count = 0; value; count++)
    {
        /* clear the least significant bit set */
        value &= value - 1;
    }

    return count;
}

/*------------------------------------------------------------------------------
NAME: calculate_saturation_value

PURPOSE: Calculates the quality band saturation value

RETURNS: saturation value
------------------------------------------------------------------------------*/
static unsigned short calculate_saturation_value
(
    unsigned char mask_value
)
{
    unsigned char bits_set;

    /* Nothing to do if no bits are set */
    if (mask_value == 0)
        return 0;

    bits_set = count_bits_set(mask_value);

    if (bits_set <= 2)
        return CC_RAD_SAT_LOW;
    else if (bits_set <= 4)
        return CC_RAD_SAT_MED;
    else
        return CC_RAD_SAT_HIGH;
}

/*------------------------------------------------------------------------------
NAME: check_algorithm

PURPOSE:
    Examine the cloud cover algorithms that the user has requested to run on
    this image, the cloud cover control parameters, and the image bands
    available to determine which cloud cover algorithms should be run on the
    image.

    At least one band must be opened to determine which samples contain fill
    data and which samples are not.  If none of the algorithms will be run,
    locate the first available band and mark it to be opened to determine fill
    samples.

RETURNS:
    This function has no return status
------------------------------------------------------------------------------*/
void check_algorithm
(
    L1GIO *fdImageIn,               /* I: File descriptor to input image */
    int bandType2Number[IAS_NUM_BAND_TYPES],/* I: Map band type to band number*/
    const CLOUD_COVER_CTRL *pCtrl,  /* I: Control parameters */
    double sunElevation,            /* I: Sun elevation angle */
    CC_ALGORITHM algorithm,         /* I: Algorithm requested */
    int *use_thermal_for_cfmask,    /* I/O: Thermal band control for CFmask*/
    int *use_cirrus_for_cfmask,     /* I/O: Cirrus band control for CFmask */
    int *can_run,                   /* O: Flag indicating the algorithm
                                          can be run */
    int *can_run_cirrus,            /* O: Flag indicating the cirrus
                                          algorithm can be run */
    int bandFlags[IAS_NUM_BAND_TYPES] /* O: Bands needed for the algorithm */
)
{
    int  i;
    int requiredBands[MAX_BANDS_PER_ALGORITHM];

    /* Assume the algorithm can't be run */
    *can_run = 0;
    *can_run_cirrus = 0;

    /* Initialize all required band flags to false */
    for (i = 0; i < IAS_NUM_BAND_TYPES; i++)
        bandFlags[i] = FALSE;

    /* Do not run any cloud cover algorithm if the sun elevation angle is 
       less than the Minimum_Sun_Elevation in the CPF file */
    if (sunElevation < pCtrl->minimumSunElevationDegrees)
    {
        IAS_LOG_INFO("Skipping cloud cover assessment due to the low sun "
            "elevation");
    }
    else
    {
        /* Provide information on which algorithms will be run */
        IAS_LOG_INFO("Using %s algorithm", get_algorithm_name(algorithm));

        /* Check whether the requested bands are available */

        /* Initialize all of the required bands to the unknown band type */
        for (i = 0; i < MAX_BANDS_PER_ALGORITHM; i++)
            requiredBands[i] = IAS_UNKNOWN_BAND_TYPE;

        if (algorithm == ALGORITHM_CFMASK)
        {
            int alg = 0;
            int thermal_available = 0;

            /* Determine whether thermal is available */
            if (ias_l1g_is_band_present(fdImageIn,
                                    bandType2Number[IAS_THERMAL1_BAND]))
            {
                thermal_available = 1;
            }

            requiredBands[alg++] = IAS_BLUE_BAND;
            requiredBands[alg++] = IAS_GREEN_BAND;
            requiredBands[alg++] = IAS_RED_BAND;
            requiredBands[alg++] = IAS_NIR_BAND;
            requiredBands[alg++] = IAS_SWIR1_BAND;
            requiredBands[alg++] = IAS_SWIR2_BAND;

            /* Decide whether to require the thermal band */
            if (*use_thermal_for_cfmask == USE_BAND)
            {
                /* The thermal band was required by the parameter, so require
                   it here */
                requiredBands[alg++] = IAS_THERMAL1_BAND;
            }
            else if (*use_thermal_for_cfmask == USE_IF_AVAILABLE)
            {
                if (thermal_available)
                {
                    /* The thermal band should be used since it is available */
                    requiredBands[alg++] = IAS_THERMAL1_BAND;
                    *use_thermal_for_cfmask = USE_BAND;
                }
                else
                {
                    /* Update the use thermal indication to say it won't be
                       used since thermal is not available */
                    *use_thermal_for_cfmask = DO_NOT_USE_BAND;
                }
            }

            /* Decide whether to use the cirrus band */
            if (*use_cirrus_for_cfmask == USE_BAND)
            {
                requiredBands[alg++] = IAS_CIRRUS_BAND;
            }
            else if (*use_cirrus_for_cfmask == USE_IF_THERMAL_NOT_AVAILABLE)
            {
                if (!thermal_available)
                {
                    /* Thermal wasn't available, so use the Cirrus */
                    requiredBands[alg++] = IAS_CIRRUS_BAND;
                    *use_cirrus_for_cfmask = USE_BAND;
                }
                else
                {
                    /* Thermal was available, so not using Cirrus */
                    *use_cirrus_for_cfmask = DO_NOT_USE_BAND;
                }
            }
        }
        else if (algorithm == ALGORITHM_ACCA)
        {
            requiredBands[0] = IAS_GREEN_BAND;
            requiredBands[1] = IAS_RED_BAND;
            requiredBands[2] = IAS_NIR_BAND;
            requiredBands[3] = IAS_SWIR1_BAND;
            requiredBands[4] = IAS_THERMAL1_BAND;
        }
        else if (algorithm == ALGORITHM_ATACCA || algorithm == ALGORITHM_SEE5)
        {
            requiredBands[0] = IAS_BLUE_BAND;
            requiredBands[1] = IAS_GREEN_BAND;
            requiredBands[2] = IAS_RED_BAND;
            requiredBands[3] = IAS_NIR_BAND;
            requiredBands[4] = IAS_SWIR1_BAND;
            requiredBands[5] = IAS_SWIR2_BAND;
        }

        if (check_bands(algorithm, fdImageIn, bandType2Number, requiredBands,
                bandFlags) == SUCCESS)
        {
            /* Can run the algorithm */
            *can_run = 1;

            /* For CFmask, report whether the Thermal and Cirrus bands are
               being used */
            if (algorithm == ALGORITHM_CFMASK)
            {
                IAS_LOG_INFO("%s the thermal band for CFmask", 
                    *use_thermal_for_cfmask ? "Using" : "Not using");
                IAS_LOG_INFO("%s the cirrus band for CFmask", 
                    *use_cirrus_for_cfmask ? "Using" : "Not using");
            }
        }

        /* Check whether the Cirrus algorithm can be run */
        if (ias_l1g_is_band_present(fdImageIn,
                bandType2Number[IAS_CIRRUS_BAND]))
        {
            bandFlags[IAS_CIRRUS_BAND] = TRUE;
            IAS_LOG_INFO("Using %s algorithm", get_algorithm_name(
                ALGORITHM_CIRRUS));
            *can_run_cirrus = 1;
        }
        else
        {
            IAS_LOG_WARNING("Algorithm %s skipped due to missing band: %d",
                get_algorithm_name(ALGORITHM_CIRRUS),
                bandType2Number[IAS_CIRRUS_BAND]);
        }
    }

    /* If no algorithms will be run, locate the first available band and mark it
       as one that will be opened to determine fill samples. */
    if (!*can_run && !*can_run_cirrus)
    {
        for (i = 0; i < IAS_NUM_BAND_TYPES; i++)
        {
            if (ias_l1g_is_band_present(fdImageIn, bandType2Number[i]))
            {
                bandFlags[i] = TRUE;
                break;
            }
        }
    }
}

/*------------------------------------------------------------------------------
NAME: init_algorithm_masks

PURPOSE:
    Initialize the array of sample masks for each algorithm that is not to be
    run.  This will prevent each algorithm from having to repeatedly set an
    array of mask values to CC_MASK_EMPTY for every line that is processed.  The
    functions for those algorithms can simply return if they are not to run.

    This is the only function within this file that knows or cares about the
    number of lines that are being read from and written to our image files in
    block for I/O performance issues.
------------------------------------------------------------------------------*/
void init_algorithm_masks
(
    int nSamples,                 /* I: Samples per line*/
    int nLines2Read,              /* I: Lines in each algorithm's buffer */
    unsigned short * sampleMasks[NUM_BUFFERS] /* IO: Samples masks */
)
{
    int algorithm;
    int sample;
    int samplesInBuffer = (nLines2Read * nSamples);

    for (algorithm = 0; algorithm < NUM_BUFFERS; algorithm++)
    {
        for (sample = 0; sample < samplesInBuffer; sample++)
            sampleMasks[algorithm][sample] = CC_MASK_EMPTY;
    }
}

/*------------------------------------------------------------------------------
NAME: resolve_algorithm_masks

PURPOSE:
    Examine the sample masks from each of the algorithms and determine what the
    final mask should be for cloud cover assessment as a whole.

    The sampleMasks array contains a line of masks for each of the algorithms
    supported.  The last line of data in this array is for storing the final,
    resolved mask that will be assigned to the samples.
------------------------------------------------------------------------------*/
void resolve_algorithm_masks
(
    int can_run,                /* I: Flag that the primary algorithm can run */
    int can_run_cirrus,         /* I: Flag that the Cirrus algorithm ran */
    int nSamples,               /* I: Samples per line */
    unsigned short * const sampleImage[IAS_NUM_BAND_TYPES],/* I: Image DN 
                                                                values */
    const unsigned short *terrainMask,/* I: Terrain mask */
    unsigned short *sampleMasks[NUM_BUFFERS],/* IO: Sample masks */
    const unsigned char *landWaterMask,/* I: Land/water mask for current line
                                             (1 = land, 0 = water) */
    const unsigned char *l1r_saturation_mask, /* I: Saturated pixel mask for
                                                    the L1R image */
    const unsigned char *l1g_saturation_mask, /* I: Saturated pixel mask for
                                                    the L1G image */
    int *nFill,/* IO: Number of fill samples */
    int *nClouds,/* IO: Number of samples assigned high confidence 
                        cloud masks */
    int *nSnowIce,/* IO: Number of samples assigned high confidence
                         snow/ice masks */
    int *nCloudsLand,/* IO: Total cloud cover on land with high confidence
                            cloud masks */
    int *nTotalLand/* IO: Total land in an image */
)
{
    int sample;

    for (sample = 0; sample < nSamples; sample++)
    {
        int            bandType;
        int            isFill;
        unsigned short answer = 0;

        /* Assume the pixel isn't fill */
        isFill = FALSE;

        /* Check whether the sample mask indicates fill */
        if (sampleMasks[1][sample] == CC_MASK_FILL)
        {
            isFill = TRUE;
        }
        else
        {
            /* Does the original image data indicate this sample is fill?  We
               consider a sample to be fill if it has a value of 0 in any of
               the input bands. */
            for (bandType=0; bandType < IAS_NUM_BAND_TYPES; bandType++)
            {
                if (sampleImage[bandType] == NULL)
                    continue;
                /* Test for fill */
                if (sampleImage[bandType][sample] == 0)
                {
                    isFill = TRUE;
                    break;
                }
            }
        }

        if (isFill)
        {
            answer = CC_MASK_FILL;
            *nFill += 1;
        }
        else if (can_run || can_run_cirrus)
        {
            unsigned short maskTerrain = 0;
            unsigned short maskSaturation = 0;
            unsigned short cloud_summary = 0;

            /* Use the primary buffer as the source of the result and merge
               the cirrus results.  Only use the bits that we want since some
               algorithms fill them in with extra info (i.e. See5 identifies
               cirrus clouds that we don't want to use). */
            answer = (sampleMasks[CLOUD_BUFFER_INDEX][sample]
                        & (CC_CLOUD_HIGH | CC_CLOUD_SHADOW_HIGH | CC_SNOW_HIGH))
                      | sampleMasks[CIRRUS_BUFFER_INDEX][sample];

            /* The cloud cover assessment score is based on the number of
               samples that are classified with a high level of confidence as
               cloud.  Keep track of the number of samples classified as such */
            if ((answer & CC_CLOUD_HIGH) == CC_CLOUD_HIGH)
            {
                *nClouds += 1;
                cloud_summary = CC_CLOUD_SUMMARY;
            }

            /* The snow/ice score is based on the number of samples that are 
               classified with a high level of confidence as snow/ice.
               Keep track of the number of samples classified as such */
            if ((answer & CC_SNOW_HIGH) == CC_SNOW_HIGH)
                *nSnowIce += 1;

            /* The cloud cover land score is base on the number of samples
               that are classified with a high level of confidence as 
               cloud on land (not on water). Keep track of the number of
               samples classified as such */
            if (landWaterMask != NULL && landWaterMask[sample] == 1)
            {
                *nTotalLand += 1;
                if ((answer & CC_CLOUD_HIGH) == CC_CLOUD_HIGH)
                {
                    *nCloudsLand +=1; 
                }
            }
          
            /* If a terrain occlusion mask is being processed, then we'll need
               to include that mask into our final mask.  The terrain occlusion
               appplication generates a mask that is either 0 for non-occluded
               or 1 for occluded samples.  If the terrainMask is NULL, the
               default value from above will be a zero.  If it's not NULL, we'll
               use this application's bit mask for terrain occlusion. */
            if (terrainMask != NULL)
                maskTerrain = (terrainMask[sample] ? CC_MASK_TERRAIN : 0);

            if (l1r_saturation_mask)
            {
                unsigned char mask = l1r_saturation_mask[sample]
                                   | l1g_saturation_mask[sample];

                maskSaturation = calculate_saturation_value(mask);
            }

            /* Combine the resolved masks into a single mask for this sample */
            answer = answer | cloud_summary | maskTerrain | maskSaturation;
        }
        else if (terrainMask != NULL || l1r_saturation_mask != NULL)
        {
            unsigned short maskTerrain = 0;
            unsigned short maskSaturation = 0;

            if (terrainMask)
                maskTerrain = (terrainMask[sample] ? CC_MASK_TERRAIN : 0);

            if (l1r_saturation_mask)
            {
                unsigned char mask = l1r_saturation_mask[sample]
                                   | l1g_saturation_mask[sample];

                maskSaturation = calculate_saturation_value(mask);
            }

            answer = maskTerrain | maskSaturation;
        }

        sampleMasks[RESULT_BUFFER_INDEX][sample] = answer;
    }
}


/*------------------------------------------------------------------------------
NAME: run_algorithm_acca

PURPOSE:
    Implementation of the Automated Cloud Cover Assessment (ACCA) algorithm
    based on the LDCM ADD.  This function will examine every sample in an image
    line and determine a classification mask for those samples.
------------------------------------------------------------------------------*/
void run_algorithm_acca
(
    int nSamples,                           /* I: Samples per line */
    float *sampleScaled[IAS_NUM_BAND_TYPES],/* I: Array of pointers to scaled
                                               TOA values, 1/band */
    unsigned short sampleMasks[]            /* O: Line of masks computed
                                               by the algorithm */
)
{
    float b10;
    float b3;
    float b4;
    float b5;
    float b6;
    float tmp;
    int   i;

    /* The algorithm should be run, examine all samples */
    for (i = 0; i < nSamples; i++)
    {
        /* For ease of looking at the algorithm, extract the sample's scaled TOA
           reflectance for the OLI bands and the radiance value in the TIRS
           band.  Note that the sampleScaled array is indexed by band type to
           make this lookup efficient. */
        b3 = sampleScaled[IAS_GREEN_BAND][i];
        b4 = sampleScaled[IAS_RED_BAND][i];
        b5 = sampleScaled[IAS_NIR_BAND][i];
        b6 = sampleScaled[IAS_SWIR1_BAND][i];
        b10 = sampleScaled[IAS_THERMAL1_BAND][i];

        if (b4 < 0.07)
        {
            sampleMasks[i] = (CC_WATER_MED | CC_CLOUD_LOW);
            continue;
        }

        if (b4 <= 0.08)
        {
            sampleMasks[i] = CC_CLOUD_MED;
            continue;
        }

        tmp = (b3 - b6) / (b3 + b6);
        if ((tmp <= -0.25) || (tmp >= 0.7))
        {
            if (tmp > 0.8)
            {
                sampleMasks[i] = (CC_SNOW_HIGH | CC_CLOUD_LOW);
                continue;
            }
            sampleMasks[i] = CC_CLOUD_LOW;
            continue;
        }

        if (b10 >= 9.390745)
        {
            /* Thermal test */
            sampleMasks[i] = CC_CLOUD_LOW;
            continue;
        }

        tmp = 666.09 / (exp(5.70093 * (1.0 - b6)) - 1.0);
        if (b10 >= tmp)
        {
            if (b6 < 0.08)
            {
                sampleMasks[i] = CC_CLOUD_LOW;
                continue;
            }
            sampleMasks[i] = CC_CLOUD_MED;
            continue;
        }

        if (((b5 / b4) < 2.35) && ((b5 / b3) < 2.2) && ((b5 / b6) > 1.0))
        {
            sampleMasks[i] = CC_CLOUD_HIGH;
            continue;
        }

        sampleMasks[i] = CC_CLOUD_MED;
    }
}


/*------------------------------------------------------------------------------
NAME: run_algorithm_atacca

PURPOSE:
    Implementation of the Artificial Thermal Automated Cloud Cover Assessment
    (ATACCA) algorithm based on the LDCM ADD.  This function will examine every
    sample in an image line and determine a classification mask for those them.
------------------------------------------------------------------------------*/
void run_algorithm_atacca
(
    int nSamples,                           /* I: Samples per line */
    float sineSolarElev,                    /* I: Sine solar elevation */
    float *sampleScaled[IAS_NUM_BAND_TYPES],/* I: Array of pointers to scaled
                                               TOA values, 1/band */
    unsigned short sampleMasks[]             /* O: Line of masks computed by the
                                                algorithm */
)
{
    float b2;
    float b3;
    float b4;
    float b5;
    float b6;
    float b7;
    float bArtificialThermal;     /* Calculated artifical thermal band */
    float nFactor;                /* Normalization factor */
    float tmp;
    int   i;
    int   nHits              = 0;

    /* The algorithm should be run, examine all samples */
    for (i = 0; i < nSamples; i++)
    {
        /* For ease of looking at the algorithm, extract the sample's scaled TOA
           reflectance.  Note that the sampleScaled array is indexed by band
           type to make this lookup efficient. */
        b2 = sampleScaled[IAS_BLUE_BAND][i];
        b3 = sampleScaled[IAS_GREEN_BAND][i];
        b4 = sampleScaled[IAS_RED_BAND][i];
        b5 = sampleScaled[IAS_NIR_BAND][i];
        b6 = sampleScaled[IAS_SWIR1_BAND][i];
        b7 = sampleScaled[IAS_SWIR2_BAND][i];

        /* Calculate the AT band value */
        bArtificialThermal = eval_cubist_therm(b2, b3, b4, b5, b6, b7,
            sineSolarElev);

        if (b4 < 0.07)
        {
            sampleMasks[i] = (CC_WATER_MED | CC_CLOUD_LOW);
            continue;
        }

        if (b4 > 0.08)
        {
            tmp = (b3 - b6) / (b3 + b6);
            if ((tmp <= -0.25) || (tmp >= 0.7))
            {
                if (tmp > 0.8)
                {
                    sampleMasks[i] = (CC_SNOW_HIGH | CC_CLOUD_LOW);
                    continue;
                }
                sampleMasks[i] = CC_CLOUD_LOW;
                continue;
            }

            if (bArtificialThermal >= 300)
            {
                sampleMasks[i] = CC_CLOUD_LOW;
                continue;
            }

            tmp = (1.0 - b6) * bArtificialThermal;
            if (tmp >= 225.0)
            {
                if (b6 < 0.08)
                {
                    sampleMasks[i] = CC_CLOUD_LOW;
                    continue;
                }
            }
            else if (((b5 / b4) < 2.35) && ((b5 / b3) < 2.2) &&
                ((b5 / b6) > 1.0))
            {
                sampleMasks[i] = CC_CLOUD_HIGH;
                continue;
            }
        }

        /*----------------------------------------------------------------------
         * Run the gD02 disambiguation tests
         *--------------------------------------------------------------------*/
        nHits = 0;

        /* Calculate the normalization factor */
        nFactor = sqrt((b2 * b2) + (b3 * b3) + (b4 * b4) + (b5 * b5)
            + (b6 * b6) + (b7 * b7));

        /* Single band thresholds */
        if (b2 < 0.140)
            nHits++;
        if (b3 < 0.111)
            nHits++;
        if (b4 < 0.093)
            nHits++;

        tmp = b6 / nFactor;
        if ((tmp < 0.087) || (tmp > 0.481))
            nHits++;

        /* Two-band combinations */
        tmp = b4 / b2;
        if ((tmp < 0.640) || (tmp > 1.034))
            nHits++;

        tmp = ((sineSolarElev * b2) - b5) / ((sineSolarElev * b2) + b5);
        if ((tmp < -0.454) || (tmp > 0.262))
        {
            /* NDC25 */
            nHits++;
        }

        tmp = (b2 - b6) / (b2 + b6);
        if ((tmp < -0.138) || (tmp > 0.716))
        {
            /* ND26 */
            nHits++;
        }

        tmp = (sineSolarElev * b2) / b7;
        if ((tmp < 0.736) || (tmp > 3.914))
        {
            /* CB2/B7 */
            nHits++;
        }

        tmp = b4 / b3;
        if ((tmp < 0.810) || (tmp > 1.075))
        {
            /* B4/B3 */
            nHits++;
        }

        tmp = (b3 - b5) / (b3 + b5);
        if ((tmp < -0.404) || (tmp > 0.160))
        {
            /* ND35 */
            nHits++;
        }

        tmp = (b3 - b6) / (b3 + b6);
        if ((tmp < -0.186) || (tmp > 0.716))
        {
            /* ND36 */
            nHits++;
        }

        tmp = (b3 - b7) / (b3 + b7);
        if ((tmp < -0.018) || (tmp > 0.754))
        {
            /* ND37 */
            nHits++;
        }

        tmp = ((sineSolarElev * b4) - b5) / ((sineSolarElev * b4) + b5);
        if ((tmp < -0.566) || (tmp > -0.016))
        {
            /* NDC45 */
            nHits++;
        }

        tmp = (b4 - b6) / (b4 + b6);
        if ((tmp < -0.232) || (tmp > 0.692))
        {
            /* ND46 */
            nHits++;
        }

        tmp = (b4 - b7) / (b4 + b7);
        if ((tmp < -0.030) || (tmp > 0.738))
        {
            /* ND47 */
            nHits++;
        }

        tmp = (b6 - b7) / (b6 + b7);
        if ((tmp < -0.050) || (tmp > 0.300))
        {
            /* ND67 */
            nHits++;
        }

        if (nHits > 1)
        {
            /* 2 or more tests agree; no cloud */
            sampleMasks[i] = CC_CLOUD_LOW;
        }
        else if (nHits == 0)
        {
            /* No tests hit, likely a cloud. */
            sampleMasks[i] = CC_CLOUD_HIGH;
        }
        else
            sampleMasks[i] = CC_CLOUD_MED;
    }
}

/*------------------------------------------------------------------------------
NAME: create_cirrus_threshold_lookup_table

PURPOSE:
    Create the elevation threshold lookup table. Each index in this table
    is an elevation in meters, and it holds a threshold.
    Elevation can be negative, if the min_elev param is negative, so to find a 
    threshold in this table given an elevation, use this equation:
    table_index = elevation - min_elevation, 
        where 0 <= i < max_elevation + min_elevation + 1.

RETURNS:
    SUCCESS if table successfully created
    ERROR if memory could not be allocated
------------------------------------------------------------------------------*/
int create_cirrus_threshold_lookup_table
(
    const double *cirrus_values,    /* I: Array of 4 Cirrus threshold values */
    const int min_elev,             /* I: Minimum elevation */
    const int max_elev,             /* I: Maximum elevation
                                       table size = max - min */
    double **table                  /* O: Lookup table created */
)
{
    int i;          /* Loop counter */
    int elevation_range = max_elev - min_elev + 1;  

    /* Each index in the table represents an elevation to the nearest meter.
       max_elev-min_elev = elev range */
    *table = (double*) malloc(sizeof(double) * (elevation_range));
    if (*table == NULL)
    {
        IAS_LOG_ERROR("Allocating threshold lookup table"); 
        return ERROR;
    }
 
    for (i = 0; i < elevation_range; i++)
    {
        (*table)[i] = cirrus_values[0] * exp(cirrus_values[1] + (i + min_elev) /
            cirrus_values[2]) + cirrus_values[3];
    }

    return SUCCESS;
}

/*------------------------------------------------------------------------------
NAME: run_algorithm_cirrus

PURPOSE:
    Implementation of the Cirrus cloud cover assessment algorithm based on the
    LDCM ADD.  This function will examine every sample in an image line and
    determine a classification mask for them.
------------------------------------------------------------------------------*/
void run_algorithm_cirrus
(
    int nSamples,                           /* I: Samples per line */
    const float *scaled_samples,            /* I: Array of pointers to scaled
                                               TOA values, 1/band */
    const short *dem_line,                  /* I: Array of elevations */
    const double *threshold_lut,            /* I: Elevation to threshold LUT */
    int min_elev,                           /* I: Minimum elevation of DEM */
    int max_elev,                           /* I: Maximum elevation of DEM */
    unsigned short sampleMasks[]            /* O: Line of masks computed by the
                                               algorithm */
)
{
    int i;
    int max_index = max_elev - min_elev;

    /* The algorithm should be run, examine all samples. */
    for (i = 0; i < nSamples; i++)
    {
        int lookup_index = dem_line[i] - min_elev;

        if (lookup_index < 0)
            lookup_index = 0;
        else if (lookup_index > max_index)
            lookup_index = max_index;

        /* Compare the current sample to the threshold 
           stored at the appropriate elevation */
        if (scaled_samples[i] > threshold_lut[lookup_index])
            sampleMasks[i] = CC_CIRRUS_HIGH;
        else
            sampleMasks[i] = CC_CIRRUS_LOW;
    }
}

/*------------------------------------------------------------------------------
NAME: run_algorithm_see5

PURPOSE:
    Implementation of the See5 cloud cover assessment algorithm based on the
    LDCM ADD.  This algorithm was derived from the COTS package C5 by Rulequest
    Research.  The C5 software is a development tool and is not required to
    execute this See5 algorithm.

    This function will examine every sample in an image line and determine a
    classification mask for them.
------------------------------------------------------------------------------*/
void run_algorithm_see5
(
    int nSamples,                           /* I: Samples per line */
    float sineSolarElev,                    /* I: Sine solar elevation */
    float *sampleScaled[IAS_NUM_BAND_TYPES],/* I: Array of pointers to scaled
                                               TOA values, 1/band */
    unsigned short sampleMasks[]            /* O: Line of masks computed by the
                                               algorithm */
)
{
    float b2;
    float b3;
    float b4;
    float b5;
    float b6;
    float b7;
    int   i;

    /*==========================================================================
     * NOTE:
     * -----
     * Comments from original algorithm:
     *
     * See5 tree from tree file of see5_rE1non_n4_m2900.txt.  The 'Results'
     * comments are the binned classes of each leaf.  Classes are [Shadow,
     * ThickCloud, ThinCloud, Clear].  The Shadow bin is treated as Clear in
     * this implementation.  Cirrus is represented as ThinClouds.  All Clouds =
     * ThickCloud + ThinCloud.  The confidence levels are set by what percentage
     * of the total the class represents.  0-35% = Low, 36-65% = Mid, 66-100% =
     * Hi.
     *
     * Example:  Test 43 -- Results [3, 294, 2422, 1341]
     *     Cirrus% = ThinCloud/Total = 2422/4060 = 0.59 = Mid confidence.
     *     Cloud%  = (ThickCloud + ThinCloud)/Total = (294+2422)/4060 = 0.67 =
     *               High confidence.
     *
     * Note that this evaluation function makes no class distinctions other than
     * Cirrus and Cloud.  No water, snow etc.
     *========================================================================*/

    /*==========================================================================
     * NOTE:
     * -----
     * There are sections in the algorithm below that are ifdef'd out because
     * all paths of execution return the same mask value.  These blocks of code
     * have been left in the code for reference and for possible future updates.
     *========================================================================*/

    /* The algorithm should be run, examine all samples */
    for (i = 0; i < nSamples; i++)
    {
        /* For ease of looking at the algorithm, extract the sample's scaled TOA
           reflectance.  Note that the sampleScaled array is indexed by band
           type to make this lookup efficient. */
        b2 = sampleScaled[IAS_BLUE_BAND][i];
        b3 = sampleScaled[IAS_GREEN_BAND][i];
        b4 = sampleScaled[IAS_RED_BAND][i];
        b5 = sampleScaled[IAS_NIR_BAND][i];
        b6 = sampleScaled[IAS_SWIR1_BAND][i];
        b7 = sampleScaled[IAS_SWIR2_BAND][i];

        if (b2 <= 0.210943)
        {
            if (b2 <= 0.13663401)
            {
                if ((b3-b6)/(b3+b6) <= -0.242285)
                {
                    if ((b2-b7)/(b2+b7) <= -0.118556)
                    {
                        /* Test 1 -- Results [259,218,726,674785] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                        continue;
                    }
                    if (sineSolarElev * b2 <= 0.084307648)
                    {
                        if (b2 <= 0.116175)
                        {
                            /* Test 2 -- Results [1135,70,1827,515328] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                            continue;
                        }
                        if (sineSolarElev * b3 <= 0.059435301)
                        {
                            /* Test 3 -- Results [33,33,891,32484] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                            continue;
                        }
                        if ((b5/b6) <= 0.953466)
                        {
                            /* Test 4 -- Results [1,4,711,8729] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                            continue;
                        }
                        if (b2 <= 0.124386)
                        {
                            /* Test 5 -- Results [4,5,1346,4852] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                            continue;
                        }
                        /* Test 6 -- Results [3,10,2946,1765] */
                        sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                        continue;
                    }
                    /* Test 7 -- Results [1304,927,4083,399744] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                    continue;
                }
                if ((b5/b3) <= 0.61224502)
                {
                    /* Test 8 -- Results [989,151,7951,846625] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                    continue;
                }
                if (b7 <= 0.019161999)
                {
#if 0           /* All Paths Return (CC_CIRRUS_LOW | CC_CLOUD_LOW) */
                    if ((b5/b3) <= 0.74089003)
                    {
                        /* Test 9 -- Results [372,7,260,7242] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                        continue;
                    }
                    if (b6 <= 0.0406789)
                    {
                        if (sineSolarElev * b4 <= 0.039809663)
                        {
                            if (b4 <= 0.040796999)
                            {
                                if (b6 <= 0.0300411)
                                {
                                    /* Test 10 -- Results [15224,12,37,6716] */
                                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                                    continue;
                                }
                                if (sineSolarElev * b4 <= 0.029391319)
                                {
                                    if (sineSolarElev * b6 <= 0.025535824)
                                    {
                                        /* Test 11 -- Results [1688,0,3,1602] */
                                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                                        continue;
                                    }
                                    /* Test 12 -- Results [2015,1,7,3783] */
                                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                                    continue;
                                }
                                /* Test 13 -- Results [2076,9,4,1337] */
                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                                continue;
                            }
                            if (sineSolarElev * b3 <= 0.047896955)
                            {
                                /* Test 14 -- Results [603,6,27,2631] */
                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                                continue;
                            }
                            /* Test 15 -- Results [1994,8,6,1607] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                            continue;
                        }
                        /* Test 16 -- Results [483,7,12,3247] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                        continue;
                    }
#endif
                    /* Test 17 -- Results [3267,12,33,15891] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                    continue;
                }
                if (b2 <= 0.113181)
                {
#if 0           /* All Paths Return (CC_CIRRUS_LOW | CC_CLOUD_LOW) */
                    if ((b5/b3) <= 0.71711802)
                    {
                        /* Test 18 -- Results [67,4,680,19266] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                        continue;
                    }
                    if (sineSolarElev * b3 <= 0.042158656)
                    {
                        /* Test 19 -- Results [1535,10,211,34309] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                        continue;
                    }
                    if (b5 <= 0.0948378)
                    {
                        if (b2 <= 0.092718102)
                        {
                            if (b5 <= 0.080040701)
                            {
                                /* Test 20 -- Results [4624,11,9,2664] */
                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                                continue;
                            }
                            /* Test 21 -- Results [1583,17,19,2024] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                            continue;
                        }
                        if (sineSolarElev * b2 <= 0.070632555)
                        {
                            /* Test 22 -- Results [42,1,72,3354] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                            continue;
                        }
                        if ((b3-b6)/(b3+b6) <= 0.164114)
                        {
                            if (sineSolarElev * b2 <= 0.080494568)
                            {
                                /* Test 23 -- Results [1587,11,26,1478] */
                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                                continue;
                            }
                            /* Test 24 -- Results [2559,40,16,4356] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                            continue;
                        }
                        /* Test 25 -- Results [982,18,148,2156] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                        continue;
                    }
#endif
                    /* Test 26 -- Results [10844,345,2954,292715] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                    continue;
                }
                if (b2 <= 0.126233)
                {
                    /* Test 27 -- Results [3477,690,14102,154560] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                    continue;
                }
                if (sineSolarElev * b2 <= 0.1109539)
                {
                    if (sineSolarElev * b2 <= 0.078559302)
                    {
                        /* Test 28 -- Results [0,46,881,10193] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                        continue;
                    }
                    if (sineSolarElev * b5 <= 0.063801818)
                    {
                        if (sineSolarElev * b5 <= 0.047259722)
                        {
                            /* Test 29 -- Results [3,71,1630,3953] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                            continue;
                        }
                        if ((b3-b6)/(b3+b6) <= 0.346856)
                        {
                            /* Test 30 -- Results [1,283,2246,2296] */
                            sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                            continue;
                        }
                        /* Test 31 -- Results [4,32,2510,1052] */
                        sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                        continue;
                    }
                    /* Test 32 -- Results [584,433,7553,37061] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                    continue;
                }
                /* Test 33 -- Results [636,392,3507,61635] */
                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                continue;
            }
            if ((b2-b7)/(b2+b7) <= -0.047816999)
            {
                /* Test 34 -- Results [352,2458,9729,947822] */
                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                continue;
            }
            if ((b5/b3) <= 0.66206598)
            {
                if (sineSolarElev * b2 <= 0.081314541)
                {
                    if (sineSolarElev * b5 <= 0.02490015)
                    {
                        /* Test 35 -- Results [0,2,406,7866] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                        continue;
                    }
                    /* Test 36 -- Results [0,33,1807,1211] */
                    sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                    continue;
                }
                /* Test 37 -- Results [153,509,2290,60460] */
                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                continue;
            }
            if (b2 <= 0.171058)
            {
                if (sineSolarElev * b7 <= 0.10569602)
                {
                    if (sineSolarElev * b6 <= 0.17047222)
                    {
                        if ((b3-b6)/(b3+b6) <= 0.46690831)
                        {
                            if (b2 <= 0.15011001)
                            {
                                if (sineSolarElev * b2 <= 0.12550308)
                                {
                                    if (sineSolarElev * b3 <= 0.042622104)
                                    {
                                        /* Test 38 -- Results [0,0,779,5777] */
                                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                                        continue;
                                    }
                                    if (b4 <= 0.117747)
                                    {
                                        if (b7 <= 0.033315498)
                                        {
                                            /* Test 39 -- Results [41,143,474,2329] */
                                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                                            continue;
                                        }
                                        if ((b3-b6)/(b3+b6) <= 0.325333)
                                        {
                                            if (sineSolarElev * b2 <= 0.098944008)
                                            {
                                                if (sineSolarElev * b2 <= 0.085170656)
                                                {
                                                    /* Test 40 -- Results [0,108,1611,3852] */
                                                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                                                    continue;
                                                }
                                                /* Test 41 -- Results [0,69,5953,2701] */
                                                sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                                                continue;
                                            }
                                            if (b2 <= 0.142304)
                                            {
                                                /* Test 42 -- Results [211,691,7106,20650] */
                                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                                                continue;
                                            }
                                            if (b5 <= 0.098370299)
                                            {
                                                /* Test 43 -- Results [3,294,2422,1341] */
                                                sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                                continue;
                                            }
                                            if (sineSolarElev * b3 <= 0.089798197)
                                            {
                                                /* Test 44 -- Results [34,202,641,3178] */
                                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                                                continue;
                                            }
                                            if ((b5/b6) <= 1.52876)
                                            {
                                                /* Test 45 -- Results [39,114,1005,2912] */
                                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                                                continue;
                                            }
                                            if (sineSolarElev * b2 <= 0.12241966)
                                            {
#if 0                                       /* All Paths Return (CC_CIRRUS_MED | CC_CLOUD_MED) */
                                                if (b3 <= 0.122799)
                                                {
                                                    /* Test 46 -- Results [0,23,2612,1640] */
                                                    sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                                                    continue;
                                                }
#endif
                                                /* Test 47 -- Results [0,104,1984,2248] */
                                                sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                                                continue;
                                            }
                                            /* Test 48 -- Results [1,30,1408,2132] */
                                            sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                                            continue;
                                        }
                                        /* Test 49 -- Results [1,83,3430,1766] */
                                        sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                        continue;
                                    }
                                    /* Test 50 -- Results [22,117,981,7629] */
                                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                                    continue;
                                }
                                /* Test 51 -- Results [260,469,1524,21533] */
                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                                continue;
                            }
                            if (sineSolarElev * b3 <= 0.054698862)
                            {
                                if (sineSolarElev * b3 <= 0.045588575)
                                {
                                    /* Test 52 -- Results [0,11,1587,2024] */
                                    sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                                    continue;
                                }
                                /* Test 53 -- Results [0,2,5925,1042] */
                                sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                                continue;
                            }
                            if (sineSolarElev * b2 <= 0.14097953)
                            {
                                if (sineSolarElev * b7 <= 0.032078281)
                                {
                                    /* Test 54 -- Results [18,406,648,2322] */
                                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                                    continue;
                                }
                                if ((b3-b6)/(b3+b6) <= 0.30021599)
                                {
                                    if (sineSolarElev * b4 <= 0.10862193)
                                    {
                                        if (sineSolarElev * b5 <= 0.20268287)
                                        {
                                            if (sineSolarElev * b5 <= 0.061154772)
                                            {
                                                /* Test 55 -- Results [2,22,2880,793] */
                                                sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                                                continue;
                                            }
                                            if (sineSolarElev * b2 <= 0.10616802)
                                            {
                                                if (b5 <= 0.208418)
                                                {
                                                    /* Test 56 -- Results [0,211,2518,2860] */
                                                    sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                                                    continue;
                                                }
                                                /* Test 57 -- Results [0,204,2945,1142] */
                                                sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                                                continue;
                                            }
                                            if (sineSolarElev * b2 <= 0.12077776)
                                            {
                                                /* Test 58 -- Results [25,984,3435,11508] */
                                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                                                continue;
                                            }
                                            if ((b5/b6) <= 1.062915)
                                            {
                                                /* Test 59 -- Results [27,779,1821,1410] */
                                                sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                                                continue;
                                            }
                                            if (b5 <= 0.088948697)
                                            {
                                                /* Test 60 -- Results [78,147,794,2665] */
                                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                                                continue;
                                            }
                                            if (b2 <= 0.155118)
                                            {
                                                /* Test 61 -- Results [27,173,1681,3114] */
                                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
                                                continue;
                                            }
                                            if (sineSolarElev * b2 <= 0.12738113)
                                            {
                                                /* Test 62 -- Results [13,404,1227,2216] */
                                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
                                                continue;
                                            }
                                            if (b5 <= 0.216897)
                                            {
                                                if (sineSolarElev * b3 <= 0.1071302)
                                                {
                                                    /* Test 63 -- Results [3,135,1893,1179] */
                                                    sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                                                    continue;
                                                }
                                                /* Test 64 -- Results [9,113,1006,1772] */
                                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
                                                continue;
                                            }
                                            /* Test 65 -- Results [2,43,1818,1042] */
                                            sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                                            continue;
                                        }
                                        if (sineSolarElev * b2 <= 0.12340952)
                                        {
                                            /* Test 66 -- Results [0,390,1991,2036] */
                                            sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                                            continue;
                                        }
                                        if (b2 <= 0.160983)
                                        {
                                            if (sineSolarElev * b2 <= 0.13254747)
                                            {
                                                /* Test 67 -- Results [1,60,5900,3808] */
                                                sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                                                continue;
                                            }
                                            /* Test 68 -- Results [79,118,1169,2270] */
                                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
                                            continue;
                                        }
                                        /* Test 69 -- Results [0,125,5646,2206] */
                                        sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                                        continue;
                                    }
                                    /* Test 70 -- Results [4,175,1350,4028] */
                                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                                    continue;
                                }
                                /* Test 71 -- Results [15,239,5401,1385] */
                                sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                                continue;
                            }
                            /* Test 72 -- Results [221,908,2778,9331] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                            continue;
                        }
                        /* Test 73 -- Results [44,33,139,2733] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                        continue;
                    }
                    /* Test 74 -- Results [161,555,2412,14568] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                    continue;
                }
                if ((b5/b6) <= 1.358902)
                {
                    /* Test 75 -- Results [143,2090,5043,52554] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                    continue;
                }
                if (sineSolarElev * b6 <= 0.18711592)
                {
                    /* Test 76 -- Results [41,237,1640,1502] */
                    sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                    continue;
                }
                /* Test 77 -- Results [75,545,690,3582] */
                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                continue;
            }
            if (b4 <= 0.204326)
            {
                if (sineSolarElev * b3 <= 0.068438031)
                {
                    if (sineSolarElev * b3 <= 0.051834233)
                    {
                        /* Test 78 -- Results [0,558,1036,1496] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
                        continue;
                    }
                    /* Test 79 -- Results [0,952,12741,1128] */
                    sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                    continue;
                }
                if ((b5/b6) <= 1.38583)
                {
                    if (sineSolarElev * b5 <= 0.095773794)
                    {
                        if (sineSolarElev * b2 <= 0.14173992)
                        {
                            /* Test 80 -- Results [0,674,2267,10363] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                            continue;
                        }
                        /* Test 81 -- Results [127,474,1379,1103] */
                        sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                        continue;
                    }
                    if (sineSolarElev * b4 <= 0.093557164)
                    {
                        /* Test 82 -- Results [0,794,4486,1220] */
                        sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                        continue;
                    }
                    if (b2 <= 0.18686099)
                    {
                        if (b5 <= 0.147453)
                        {
                            /* Test 83 -- Results [29,813,2209,1083] */
                            sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                            continue;
                        }
                        if (sineSolarElev * b5 <= 0.24523753)
                        {
                            if ((b5/b6) <= 0.94214797)
                            {
                                /* Test 84 -- Results [14,880,697,2120] */
                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
                                continue;
                            }
                            if (sineSolarElev * b4 <= 0.1211436)
                            {
                                if (b2 <= 0.178883)
                                {
                                    /* Test 85 -- Results [9,624,2157,2523] */
                                    sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                                    continue;
                                }
                                /* Test 86 -- Results [2,878,1983,1186] */
                                sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                continue;
                            }
                            /* Test 87 -- Results [20,958,3949,8329] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
                            continue;
                        }
                        /* Test 88 -- Results [54,437,940,2878] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                        continue;
                    }
                    if (sineSolarElev * b5 <= 0.1048939)
                    {
                        /* Test 89 -- Results [25,704,665,2152] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
                        continue;
                    }
                    if (sineSolarElev * b4 <= 0.13176723)
                    {
                        if (b2 <= 0.19633)
                        {
                            /* Test 90 -- Results [11,2690,5464,2659] */
                            sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                            continue;
                        }
                        if ((b5/b3) <= 1.320477)
                        {
#if 0                   /* All Paths Return (CC_CIRRUS_MED | CC_CLOUD_HIGH) */
                            if (sineSolarElev * b2 <= 0.1644081)
                            {
                                /* Test 91 -- Results [0,1003,2058,1003] */
                                sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                continue;
                            }
#endif
                            /* Test 92 -- Results [4,1326,1162,409] */
                            sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                            continue;
                        }
                        /* Test 93 -- Results [0,2819,1753,643] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                        continue;
                    }
                    if ((b5/b6) <= 0.94209403)
                    {
                        /* Test 94 -- Results [2,1400,725,978] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                        continue;
                    }
                    if (sineSolarElev * b3 <= 0.16493233)
                    {
                        if (sineSolarElev * b6 <= 0.18003967)
                        {
                            /* Test 95 -- Results [7,1111,6549,2938] */
                            sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                            continue;
                        }
                        if (sineSolarElev * b6 <= 0.22373912)
                        {
                            if (b2 <= 0.198699)
                            {
                                /* Test 96 -- Results [20,701,1826,2123] */
                                sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                                continue;
                            }
                            /* Test 97 -- Results [10,1160,2466,1560] */
                            sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                            continue;
                        }
                        /* Test 98 -- Results [24,477,1162,2144] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
                        continue;
                    }
                    /* Test 99 -- Results [16,866,698,1350] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
                    continue;
                }
                if ((b2-b7)/(b2+b7) <= 0.55510998)
                {
                    if (sineSolarElev * b4 <= 0.14885563)
                    {
                        if ((b5/b3) <= 0.88910699)
                        {
                            /* Test 100 -- Results [1,541,9779,1840] */
                            sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                            continue;
                        }
                        if (sineSolarElev * b5 <= 0.20263328)
                        {
                            if (b6 <= 0.14352)
                            {
                                /* Test 101 -- Results [20,827,3582,4663] */
                                sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                                continue;
                            }
                            if (b4 <= 0.143301)
                            {
                                /* Test 102 -- Results [0,415,1221,1308] */
                                sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                                continue;
                            }
                            /* Test 103 -- Results [0,700,1526,983] */
                            sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                            continue;
                        }
                        if ((b3-b6)/(b3+b6) <= -0.15052)
                        {
                            if ((b2-b7)/(b2+b7) <= 0.14193401)
                            {
                                /* Test 104 -- Results [73,724,2923,2418] */
                                sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                                continue;
                            }
                            /* Test 105 -- Results [23,288,1053,1711] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
                            continue;
                        }
                        /* Test 106 -- Results [58,5690,34201,12031] */
                        sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                        continue;
                    }
                    /* Test 107 -- Results [9,668,1005,1403] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
                    continue;
                }
                /* Test 108 -- Results [46,1248,1138,3960] */
                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
                continue;
            }
            /* Test 109 -- Results [37,453,892,7073] */
            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
            continue;
        }
        if ((b5/b6) <= 4.1168199)
        {
            if ((b2-b7)/(b2+b7) <= -0.083967999)
            {
                /* Test 110 -- Results [281,907,2787,134037] */
                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                continue;
            }
            if (sineSolarElev * b7 <= 0.057414148)
            {
                if (b6 <= 0.076790303)
                {
                    /* Test 111 -- Results [21,275,929,5347] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                    continue;
                }
                if ((b3-b6)/(b3+b6) <= 0.071201801)
                {
                    /* Test 112 -- Results [0,583,3702,262] */
                    sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                    continue;
                }
                if ((b5/b6) <= 1.345597)
                {
                    /* Test 113 -- Results [0,2064,175,683] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                    continue;
                }
                if (sineSolarElev * b5 <= 0.087592557)
                {
                    /* Test 114 -- Results [1,455,1013,4733] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                    continue;
                }
                if (b3 <= 0.192458)
                {
                    /* Test 115 -- Results [2,1331,5867,1222] */
                    sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                    continue;
                }
                if (sineSolarElev * b5 <= 0.11743437)
                {
                    if (b2 <= 0.29436001)
                    {
                        /* Test 116 -- Results [0,1916,971,1057] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                        continue;
                    }
                    /* Test 117 -- Results [73,702,500,5300] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                    continue;
                }
                if (sineSolarElev * b6 <= 0.052905664)
                {
                    if (b2 <= 0.69653898)
                    {
                        if (b2 <= 0.458547)
                        {
                            /* Test 118 -- Results [0,331,984,1589] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
                            continue;
                        }
                        /* Test 119 -- Results [479,1130,4736,1052] */
                        sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                        continue;
                    }
                    /* Test 120 -- Results [302,548,310,2431] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                    continue;
                }
                if (b5 <= 0.30347499)
                {
                    if (b7 <= 0.073582202)
                    {
                        /* Test 121 -- Results [1,599,1180,1764] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
                        continue;
                    }
                    /* Test 122 -- Results [0,361,3012,1282] */
                    sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                    continue;
                }
                if (b5 <= 0.68293899)
                {
                    if (b6 <= 0.173669)
                    {
                        if (b7 <= 0.091056198)
                        {
                            if (b5 <= 0.340536)
                            {
                                /* Test 123 -- Results [0,610,1830,628] */
                                sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                continue;
                            }
                            /* Test 124 -- Results [30,1991,509,864] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                            continue;
                        }
                        /* Test 125 -- Results [12,694,2429,2573] */
                        sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                        continue;
                    }
                    if ((b5/b3) <= 0.98334301)
                    {
                        /* Test 126 -- Results [0,4874,1452,977] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                        continue;
                    }
                    if ((b5/b3) <= 1.115532)
                    {
                        /* Test 127 -- Results [0,1338,4323,1139] */
                        sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                        continue;
                    }
                    /* Test 128 -- Results [0,1580,1000,384] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                    continue;
                }
                /* Test 129 -- Results [0,770,401,1740] */
                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
                continue;
            }
            if (b2 <= 0.257061)
            {
                if (b4 <= 0.25563201)
                {
                    if (sineSolarElev * b6 <= 0.078351013)
                    {
                        /* Test 130 -- Results [0,6343,1882,795] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                        continue;
                    }
                    if ((b5/b6) <= 1.051919)
                    {
                        if (sineSolarElev * b5 <= 0.13801564)
                        {
                            /* Test 131 -- Results [0,8930,1010,1981] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                            continue;
                        }
                        if (sineSolarElev * b6 <= 0.18971382)
                        {
                            if (b2 <= 0.245692)
                            {
                                if (b4 <= 0.201566)
                                {
                                    if ((b2-b7)/(b2+b7) <= 0.15269201)
                                    {
                                        /* Test 132 -- Results [0,1948,1117,314] */
                                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                                        continue;
                                    }
                                    /* Test 133 -- Results [0,1441,1754,283] */
                                    sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                    continue;
                                }
                                /* Test 134 -- Results [0,886,1861,310] */
                                sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                continue;
                            }
                            /* Test 135 -- Results [0,2554,823,157] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                            continue;
                        }
                        if ((b5/b3) <= 1.285612)
                        {
                            /* Test 136 -- Results [7,3091,917,1226] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                            continue;
                        }
                        if ((b3-b6)/(b3+b6) <= -0.21574099)
                        {
                            /* Test 137 -- Results [7,834,1594,800] */
                            sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                            continue;
                        }
                        if (b2 <= 0.23133799)
                        {
                            /* Test 138 -- Results [0,1018,1221,1288] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
                            continue;
                        }
                        /* Test 139 -- Results [0,1586,936,731] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                        continue;
                    }
                    if ((b3-b6)/(b3+b6) <= -0.121033)
                    {
                        if (sineSolarElev * b5 <= 0.23169012)
                        {
                            /* Test 140 -- Results [0,2707,1129,246] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                            continue;
                        }
                        if (b2 <= 0.232825)
                        {
                            if (sineSolarElev * b2 <= 0.19279751)
                            {
                                if (sineSolarElev * b2 <= 0.1751333)
                                {
                                    /* Test 141 -- Results [0,1246,1162,557] */
                                    sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                    continue;
                                }
                                if (b6 <= 0.275996)
                                {
                                    /* Test 142 -- Results [1,522,1624,883] */
                                    sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                    continue;
                                }
                                /* Test 143 -- Results [6,829,997,1206] */
                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
                                continue;
                            }
                            /* Test 144 -- Results [10,1284,947,954] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                            continue;
                        }
                        /* Test 145 -- Results [1,4552,2942,1486] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                        continue;
                    }
                    if (sineSolarElev * b5 <= 0.11944039)
                    {
                        /* Test 146 -- Results [2,401,1972,802] */
                        sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                        continue;
                    }
                    if (sineSolarElev * b3 <= 0.12233101)
                    {
                        /* Test 147 -- Results [0,554,2353,164] */
                        sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                        continue;
                    }
                    if (b2 <= 0.240036)
                    {
                        if (sineSolarElev * b4 <= 0.13837875)
                        {
                            if (sineSolarElev * b2 <= 0.17140751)
                            {
                                /* Test 148 -- Results [3,4423,2434,1225] */
                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                                continue;
                            }
                            /* Test 149 -- Results [3,1161,2127,457] */
                            sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                            continue;
                        }
                        if (b4 <= 0.21204799)
                        {
                            if (sineSolarElev * b2 <= 0.19835448)
                            {
                                if (b2 <= 0.22565)
                                {
                                    /* Test 150 -- Results [14,5224,16488,3690] */
                                    sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                    continue;
                                }
                                if (sineSolarElev * b2 <= 0.18106489)
                                {
                                    /* Test 151 -- Results [0,2728,1820,385] */
                                    sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                    continue;
                                }
                                /* Test 152 -- Results [1,3362,9338,1484] */
                                sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                                continue;
                            }
#if 0                   /* All Paths Return (CC_CIRRUS_MED | CC_CLOUD_HIGH) */
                            if (b6 <= 0.20226599)
                            {
                                /* Test 153 -- Results [5,887,1395,624] */
                                sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                continue;
                            }
                            if (b4 <= 0.198002)
                            {
                                /* Test 154 -- Results [6,1541,1284,125] */
                                sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                continue;
                            }
#endif
                            /* Test 155 -- Results [5,1101,1451,358] */
                            sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                            continue;
                        }
                        /* Test 156 -- Results [0,1217,5258,1881] */
                        sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                        continue;
                    }
                    if ((b3-b6)/(b3+b6) <= 0.035213001)
                    {
                        if (sineSolarElev * b3 <= 0.17628644)
                        {
                            if (sineSolarElev * b5 <= 0.19981472)
                            {
                                /* Test 157 -- Results [0,1494,1636,68] */
                                sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                continue;
                            }
                            /* Test 158 -- Results [0,5478,1822,258] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                            continue;
                        }
#if 0               /* All Paths Return (CC_CIRRUS_MED | CC_CLOUD_HIGH) */
                        if (b4 <= 0.217106)
                        {
                            if (sineSolarElev * b2 <= 0.20839086)
                            {
                                /* Test 159 -- Results [0,1811,2558,403] */
                                sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                continue;
                            }
                            /* Test 160 -- Results [5,3967,2427,241] */
                            sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                            continue;
                        }
                        if (b6 <= 0.25634599)
                        {
                            /* Test 161 -- Results [0,621,1760,550] */
                            sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                            continue;
                        }
#endif
                        /* Test 162 -- Results [0,2313,1910,663] */
                        sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                        continue;
                    }
                    if (sineSolarElev * b4 <= 0.15283392)
                    {
                        /* Test 163 -- Results [0,1960,1558,525] */
                        sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                        continue;
                    }
                    /* Test 164 -- Results [1,3070,7773,1369] */
                    sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                    continue;
                }
                /* Test 165 -- Results [37,668,690,4229] */
                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                continue;
            }
            if ((b2-b7)/(b2+b7) <= 0.46417901)
            {
                if (sineSolarElev * b5 <= 0.10582455)
                {
                    /* Test 166 -- Results [0,4928,33,1140] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                    continue;
                }
                if (b6 <= 0.282325)
                {
                    if ((b5/b6) <= 2.1011641)
                    {
                        if ((b5/b3) <= 1.059024)
                        {
                            if ((b3-b6)/(b3+b6) <= 0.22738899)
                            {
                                if (b6 <= 0.201106)
                                {
                                    if (sineSolarElev * b4 <= 0.1717533)
                                    {
                                        /* Test 167 -- Results [1,963,3159,462] */
                                        sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                                        continue;
                                    }
                                    /* Test 168 -- Results [0,1846,1073,370] */
                                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                                    continue;
                                }
                                /* Test 169 -- Results [0,34533,4493,2343] */
                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                                continue;
                            }
                            /* Test 170 -- Results [0,41861,1653,1731] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                            continue;
                        }
                        if (b6 <= 0.225678)
                        {
                            if (sineSolarElev * b6 <= 0.10247797)
                            {
                                /* Test 171 -- Results [0,1513,1969,993] */
                                sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                continue;
                            }
                            if (sineSolarElev * b2 <= 0.19942892)
                            {
                                /* Test 172 -- Results [0,3400,1053,256] */
                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                                continue;
                            }
                            if (b5 <= 0.32815501)
                            {
                                if ((b2-b7)/(b2+b7) <= 0.374176)
                                {
                                    if (b6 <= 0.195005)
                                    {
                                        /* Test 173 -- Results [0,762,3171,300] */
                                        sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                                        continue;
                                    }
                                    if (sineSolarElev * b5 <= 0.24590094)
                                    {
                                        /* Test 174 -- Results [1,1698,1107,237] */
                                        sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                        continue;
                                    }
                                    /* Test 175 -- Results [1,1273,2094,499] */
                                    sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                    continue;
                                }
                                /* Test 176 -- Results [1,1485,1002,530] */
                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                                continue;
                            }
                            if (sineSolarElev * b5 <= 0.25459459)
                            {
                                /* Test 177 -- Results [0,2548,228,125] */
                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                                continue;
                            }
                            if (b5 <= 0.395439)
                            {
                                if ((b5/b3) <= 1.142877)
                                {
                                    /* Test 178 -- Results [0,1056,1724,121] */
                                    sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                    continue;
                                }
                                if (sineSolarElev * b4 <= 0.24533477)
                                {
#if 0                           /* All Paths Return (CC_CIRRUS_MED | CC_CLOUD_HIGH) */
                                    if (b5 <= 0.36769801)
                                    {
                                        if (sineSolarElev * b7 <= 0.1116424)
                                        {
                                            /* Test 179 -- Results [0,1232,1635,108] */
                                            sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                            continue;
                                        }
                                        /* Test 180 -- Results [0,3269,2296,266] */
                                        sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                        continue;
                                    }
#endif
                                    /* Test 181 -- Results [0,1199,1834,78] */
                                    sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                    continue;
                                }
                                /* Test 182 -- Results [0,2130,923,132] */
                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                                continue;
                            }
                            /* Test 183 -- Results [0,2222,871,62] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                            continue;
                        }
                        if (sineSolarElev * b6 <= 0.14142986)
                        {
                            if (sineSolarElev * b7 <= 0.081485748)
                            {
                                /* Test 184 -- Results [0,3433,1662,398] */
                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                                continue;
                            }
                            /* Test 185 -- Results [0,812,2131,1226] */
                            sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                            continue;
                        }
                        if (b2 <= 0.37162799)
                        {
                            if (sineSolarElev * b5 <= 0.25937405)
                            {
                                /* Test 186 -- Results [3,22889,3607,937] */
                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                                continue;
                            }
                            if ((b5/b3) <= 1.173331)
                            {
                                /* Test 187 -- Results [0,4365,629,719] */
                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                                continue;
                            }
                            if (b6 <= 0.25639001)
                            {
                                if (sineSolarElev * b7 <= 0.15463512)
                                {
                                    /* Test 188 -- Results [1,10332,5317,997] */
                                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                                    continue;
                                }
                                /* Test 189 -- Results [1,1663,2742,370] */
                                sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                continue;
                            }
                            /* Test 190 -- Results [2,19458,6082,1745] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                            continue;
                        }
                        /* Test 191 -- Results [0,8152,226,134] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                        continue;
                    }
                    /* Test 192 -- Results [0,2734,5503,767] */
                    sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                    continue;
                }
                if (sineSolarElev * b6 <= 0.10736789)
                {
                    if (sineSolarElev * b3 <= 0.19163743)
                    {
                        /* Test 193 -- Results [0,2545,2766,174] */
                        sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                        continue;
                    }
                    /* Test 194 -- Results [0,3287,305,33] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                    continue;
                }
                if ((b2-b7)/(b2+b7) <= -0.0029780001)
                {
                    /* Test 195 -- Results [28,9292,2336,2869] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                    continue;
                }
                if ((b2-b7)/(b2+b7) <= 0.4372125)
                {
                    if (b2 <= 0.303727)
                    {
                        /* Test 196 -- Results [15,41227,8820,4475] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                        continue;
                    }
                    if (b2 <= 0.781452)
                    {
                        if ((b3-b6)/(b3+b6) <= 0.156515)
                        {
                            /* Test 197 -- Results [4,198473,7959,7233] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                            continue;
                        }
                        if (sineSolarElev * b2 <= 0.24399243)
                        {
                            /* Test 198 -- Results [0,10992,685,852] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                            continue;
                        }
                        if (b2 <= 0.64625502)
                        {
                            /* Test 199 -- Results [0,2156,549,365] */
                            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                            continue;
                        }
                        /* Test 200 -- Results [0,867,80,1997] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                        continue;
                    }
                    /* Test 201 -- Results [0,36431,1855,177] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                    continue;
                }
                if (b6 <= 0.428361)
                {
                    /* Test 202 -- Results [0,17103,1521,573] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                    continue;
                }
                /* Test 203 -- Results [0,1025,2034,19] */
                sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                continue;
            }
            if (b2 <= 0.87853497)
            {
                if (sineSolarElev * b2 <= 0.33472255)
                {
                    if (sineSolarElev * b2 <= 0.27875862)
                    {
                        if (b2 <= 0.685642)
                        {
                            if ((b5/b6) <= 2.0256701)
                            {
                                /* Test 204 -- Results [3,7037,1613,1035] */
                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                                continue;
                            }
                            if (sineSolarElev * b5 <= 0.21889627)
                            {
                                /* Test 205 -- Results [0,2832,5300,1437] */
                                sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                continue;
                            }
                            if (b6 <= 0.15110099)
                            {
                                /* Test 206 -- Results [0,3802,1456,817] */
                                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                                continue;
                            }
                            /* Test 207 -- Results [0,1570,4901,1376] */
                            sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                            continue;
                        }
                        /* Test 208 -- Results [0,23304,1643,2342] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                        continue;
                    }
                    if (b2 <= 0.61873698)
                    {
                        /* Test 209 -- Results [2,6392,1942,2185] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                        continue;
                    }
                    if ((b5/b3) <= 0.98910898)
                    {
                        if (sineSolarElev * b7 <= 0.096774131)
                        {
                            if ((b5/b3) <= 0.97410101)
                            {
                                if (sineSolarElev * b7 <= 0.077995896)
                                {
                                    /* Test 210 -- Results [0,1175,5349,489] */
                                    sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                                    continue;
                                }
                                if (sineSolarElev * b2 <= 0.30728427)
                                {
                                    /* Test 211 -- Results [0,1874,4599,82] */
                                    sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                                    continue;
                                }
                                /* Test 212 -- Results [0,1551,1329,22] */
                                sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                                continue;
                            }
                            /* Test 213 -- Results [0,488,7080,109] */
                            sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                            continue;
                        }
                        if (sineSolarElev * b2 <= 0.30483893)
                        {
                            /* Test 214 -- Results [0,1496,1691,42] */
                            sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
                            continue;
                        }
                        /* Test 215 -- Results [0,8269,3190,13] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                        continue;
                    }
                    /* Test 216 -- Results [0,1588,23597,176] */
                    sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                    continue;
                }
                /* Test 217 -- Results [0,16127,952,870] */
                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                continue;
            }
            if ((b2-b7)/(b2+b7) <= 0.5822916)
            {
                /* Test 218 -- Results [0,6175,912,798] */
                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                continue;
            }
            /* Test 219 -- Results [0,1604,779,2539] */
            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
            continue;
        }
        if ((b5/b3) <= 0.97751498)
        {
            if ((b5/b3) <= 0.92588001)
            {
                if (b2 <= 0.39623201)
                {
                    /* Test 220 -- Results [13,42,45,5732] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                    continue;
                }
                if (sineSolarElev * b4 <= 0.11959475)
                {
                    if (b2 <= 0.50340301)
                    {
                        /* Test 221 -- Results [1138,860,96,828] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                        continue;
                    }
                    /* Test 222 -- Results [505,2312,39,92] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_HIGH);
                    continue;
                }
                if (b2 <= 0.573946)
                {
                    /* Test 223 -- Results [253,25,1083,3452] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                    continue;
                }
                if (b7 <= 0.036635)
                {
#if 0           /* All Paths Return (CC_CIRRUS_LOW | CC_CLOUD_LOW) */
                    if (sineSolarElev * b2 <= 0.37402633)
                    {
                        /* Test 224 -- Results [1538,356,317,1131] */
                        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                        continue;
                    }
#endif
                    /* Test 225 -- Results [977,1,34,1904] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                    continue;
                }
                if (sineSolarElev * b2 <= 0.23493494)
                {
                    /* Test 226 -- Results [1563,1192,678,697] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
                    continue;
                }
                if (sineSolarElev * b2 <= 0.32929409)
                {
                    /* Test 227 -- Results [66,147,1473,1824] */
                    sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                    continue;
                }
                /* Test 228 -- Results [782,1155,389,1088] */
                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
                continue;
            }
            if (sineSolarElev * b2 <= 0.2408447)
            {
                /* Test 229 -- Results [320,826,1961,6360] */
                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                continue;
            }
            if (b2 <= 0.73523998)
            {
                /* Test 230 -- Results [0,178,960,1766] */
                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
                continue;
            }
            if (b7 <= 0.099343203)
            {
                if (b4 <= 0.73772198)
                {
                    /* Test 231 -- Results [154,524,9404,1593] */
                    sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                    continue;
                }
                /* Test 232 -- Results [21,63,1880,6237] */
                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                continue;
            }
            /* Test 233 -- Results [10,1649,4703,886] */
            sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_HIGH);
            continue;
        }
        if (b6 <= 0.0941016)
        {
            /* Test 234 -- Results [37,316,6755,183398] */
            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
            continue;
        }
        if ((b5/b3) <= 0.98999)
        {
            if ((b5/b6) <= 7.2150021)
            {
                if (b5 <= 0.73649299)
                {
                    if (b2 <= 0.805439)
                    {
                        /* Test 235 -- Results [1,83,1152,1806] */
                        sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                        continue;
                    }
                    /* Test 236 -- Results [0,137,2735,714] */
                    sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                    continue;
                }
                /* Test 237 -- Results [0,312,1622,3435] */
                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_MED);
                continue;
            }
            /* Test 238 -- Results [0,12,950,4215] */
            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
            continue;
        }
        if (sineSolarElev * b2 <= 0.3876546)
        {
            /* Test 239 -- Results [0,1162,6375,48997] */
            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
            continue;
        }
        if ((b5/b6) <= 9.0230522)
        {
            if (b6 <= 0.118084)
            {
                if ((b5/b6) <= 8.6092997)
                {
                    /* Test 240 -- Results [0,0,5997,1996] */
                    sampleMasks[i] = (CC_CIRRUS_HIGH | CC_CLOUD_HIGH);
                    continue;
                }
                if (b3 <= 0.86038798)
                {
                    if ((b2-b7)/(b2+b7) <= 0.84930003)
                    {
                        /* Test 241 -- Results [0,0,2839,2065] */
                        sampleMasks[i] = (CC_CIRRUS_MED | CC_CLOUD_MED);
                        continue;
                    }
                    /* Test 242 -- Results [0,0,924,2590] */
                    sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                    continue;
                }
                /* Test 243 -- Results [0,0,540,3700] */
                sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
                continue;
            }
            /* Test 244 -- Results [0,69,250,4388] */
            sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
            continue;
        }
        /* Test 245 -- Results [0,0,287,7985] */
        sampleMasks[i] = (CC_CIRRUS_LOW | CC_CLOUD_LOW);
    }
}

/*------------------------------------------------------------------------------
NAME: cfmask_generate_cloud_mask

PURPOSE:
    Generate a cloud mask using the Cfmask routines.
 
RETURNS: SUCCESS
         ERROR
------------------------------------------------------------------------------*/
int cfmask_generate_cloud_mask
(
    int max_threads,              /* I: Maximum threads to use */
    int n_samples,                /* I: Number of samples in an image */
    int n_lines_in_image,         /* I: Number of lines in an image */
    float cfmask_cloud_prob_threshold, /* I: Cloud probability threshold
                                              for Cfmask */
    double solar_elevation,       /* I: Solar elevation angle */
    double solar_azimuth,         /* I: Solar azimuth angle */
    const IAS_L1G_FILE_METADATA *file_metadata, /* L1G file metadata to support
                                        calculating the map azimuth angle */
    IAS_L1G_BAND_METADATA metadata_in[IAS_NUM_BAND_TYPES], /* I: IAS L1G
                                                             band metadata */
    L1G_BAND_IO *fd_bands_in[IAS_NUM_BAND_TYPES], /* I: Opened bands */
    int band_type_2number[IAS_NUM_BAND_TYPES],    /* I: Map band type to 
                                                         band number */
    const CLOUD_COVER_CTRL *pCtrl,/* I: Control parameters */
    const char *angle_coeff_filename, /* I: Angle coefficient filename */
    int cfmask_dilate_flag,       /* I: Flag to turn on/off cloud dilation */
    int use_thermal,              /* I: Flag to use thermal band */
    int use_cirrus,               /* I: Flag to use cirrus band */
    unsigned char *cloud_mask,    /* O: Pointer for the cloud mask */
    unsigned char *conf_mask,     /* O: Pointer for the cloud confidence 
                                        level*/
    SATURATION_MASK *sat_mask     /* I/O: Saturated pixel tracking info */
)
{  
    unsigned short *cfmask_buf_image;  /* Buffer of DN values */
    short *buf_scaled_values[IAS_NUM_BAND_TYPES];/* Buffers of scaled samples */
    int band_type;  /* band type loop index */
    int total_samples = n_lines_in_image * n_samples;
    int index = 0;  /* Indexing for CFMASK_INPUTS buffer structure */
  
    /* Variables used in cfmask_potential_cloud_shadow_snow_mask function
       call  */
    CFMASK_INPUTS input;        /* Input structure */
    float clear_sky_pixel;      /* Percent of clear-sky pixels */
    float low_background_temp;  /* Percentile of low background temperature */
    float high_background_temp; /* Percentile of high background temperature */
    int band;
    IAS_ANGLE_GEN_METADATA angle_metadata; /* Angle metadata structure */
    double map_azimuth_angle_radians; /* Angle from straight-up in the image
                                         to true north at the image center */

    input.num_lines = n_lines_in_image;
    input.num_samps = n_samples;
    input.nband = NBAND_REFL_MAX;

    IAS_LOG_INFO("Starting cfmask_generate_cloud_mask");

    /* Allocate memory for cfmaskBufImage */
    cfmask_buf_image = malloc(total_samples * sizeof(unsigned short));
    if (cfmask_buf_image == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for lines=%d * samples=%d",
            n_lines_in_image, n_samples);
        return ERROR;
    }

    /* Read all the bands from the image */
    for (band_type = 0; band_type < IAS_NUM_BAND_TYPES; band_type++)
    {   
        int band_number = band_type_2number[band_type];

        if (fd_bands_in[band_type] == NULL || 
            band_type == IAS_COASTAL_AEROSOL_BAND || 
            band_type == IAS_PAN_BAND ||
            band_type == IAS_THERMAL2_BAND )
        {
            buf_scaled_values[band_type] = NULL;
            continue;
        }

        IAS_LOG_INFO("Reading and scaling band number %d", band_number);

        /* Read the full L1G image band */ 
        if (ias_l1g_read_image(fd_bands_in[band_type], 0, 0, 0, 
            n_lines_in_image, n_samples, cfmask_buf_image) 
            != SUCCESS)
        {
            IAS_LOG_ERROR("Reading %d lines from band %d",
                n_lines_in_image, band_number);
            free(cfmask_buf_image);
            return ERROR;
        }

        /* If identifying saturated pixels, do that now */
        if (get_enable_saturated_pixels_flag())
        {
            int band_index;

            IAS_LOG_DEBUG("Identifying saturated pixels for band number %d",
                band_number);

            band_index = ias_sat_attr_convert_band_number_to_index(band_number);
            if (band_index == ERROR)
            {
                IAS_LOG_ERROR("Failed converting band number %d to an index",
                    band_number);
               return ERROR;
            }

            identify_saturated_l1g_pixels(sat_mask, cfmask_buf_image, 0,
                n_lines_in_image, n_samples, band_number, band_index);
        }

        IAS_LOG_DEBUG("Scaling data");

        /* Allocate buffer for the short data type scaled values */
        buf_scaled_values[band_type] = malloc(total_samples * sizeof(short));
        if (buf_scaled_values[band_type] == NULL)
        {
            IAS_LOG_ERROR("Allocating memory for lines=%d x samples=%d",
                n_lines_in_image, n_samples);
            free(cfmask_buf_image);
            for (band = 0; band < band_type; band++)
                free(buf_scaled_values[band]);
            return ERROR;
        }

        /* Check for thermal1 band to convert DN to BT, TIRS */
        if (band_type == IAS_THERMAL1_BAND)
        {
           /* Convert DN to BT for thermal band, TIRS */
           dn_to_bt(cfmask_buf_image, n_samples, n_lines_in_image,
               max_threads, pCtrl->thermalConstantsTIRS1[0],
               pCtrl->thermalConstantsTIRS1[1],
               metadata_in[IAS_THERMAL1_BAND].spectral_radiance_scaling_offset,
               metadata_in[IAS_THERMAL1_BAND].spectral_radiance_scaling_gain,
               &input.therm_min, &input.therm_max,
               buf_scaled_values[band_type]);
        }
        else
        {
            /* Convert DN to TOA reflectance, OLI */
            dn_to_toa(cfmask_buf_image, n_samples, n_lines_in_image,
                max_threads, solar_elevation,
                metadata_in[band_type].reflectance_scaling_offset,
                metadata_in[band_type].reflectance_scaling_gain,
                &input.min[index], &input.max[index],
                buf_scaled_values[band_type]);
        }

        /* Check for the thermal1 band  */
        if (band_type == IAS_THERMAL1_BAND)
        {
            input.therm_buf = buf_scaled_values[band_type];
        }
        else
        {
            /* Assign the scaled values to the CFMASK_INPUTS buffers */
            input.buf[index] = buf_scaled_values[band_type];

            /*  Increment the indexing for CFMASK_INPUTS buffer */
            index++;
        }
    }  /* End of band_type for-loop */

    /* Free buffer space that was allocated for the DN values */
    free (cfmask_buf_image);
    cfmask_buf_image = NULL;

    /* Calling the cfmask_potential_cloud_shadow_snow_mask function
       Cloud probability threshold = 22.5 by default */
    if (cfmask_potential_cloud_shadow_snow_mask(&input,
        cfmask_cloud_prob_threshold, use_thermal, use_cirrus, &clear_sky_pixel,
        &low_background_temp, &high_background_temp, cloud_mask, conf_mask)
        != SUCCESS)
    {
        IAS_LOG_ERROR("cfmask_potential_cloud_shadow_mask function call");
        for (band = 0; band < IAS_NUM_BAND_TYPES; band++)
            free(buf_scaled_values[band]);
        return ERROR;
    }

    /* Assign NULL to input.buf[] for the reflective band since the buffers
       are going to be freed */
    for (index = 0; index < NBAND_REFL_MAX; index++)
        input.buf[index] = NULL;

    /* Free the buffers for reflective band for buf_scaled_values[] */
    for (band_type = 0; band_type < IAS_NUM_BAND_TYPES; band_type++)
    {
        if (band_type != IAS_THERMAL1_BAND)
        {
            free(buf_scaled_values[band_type]);
            buf_scaled_values[band_type] = NULL;
        }
    }

    /* Calculate the map azimuth angle */
    if (calculate_map_azimuth_angle(file_metadata,
        &metadata_in[IAS_BLUE_BAND], &map_azimuth_angle_radians) != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating the map azimuth angle");
        return ERROR;
    }

    /* Read the angle coefficient file */
    memset(&angle_metadata, 0, sizeof(angle_metadata));
    if (ias_angle_gen_read_ang(angle_coeff_filename, &angle_metadata)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Reading the angle coefficient file %s",
            angle_coeff_filename);
        return ERROR;
    }

    /* Call function cfmask_object_cloud_shadow_match
       Default buffer for cloud pixel dilate = 3
       Default buffer for cloud shadow pixel dilate = 3 */
    if (cfmask_object_cloud_shadow_match(&input, &angle_metadata,
        map_azimuth_angle_radians, file_metadata->roll_angle,
        metadata_in[IAS_BLUE_BAND].projection_distance_x, clear_sky_pixel,
        low_background_temp, high_background_temp, 3, 3, max_threads,
        cfmask_dilate_flag, cloud_mask, solar_elevation, solar_azimuth,
        use_thermal) != SUCCESS)
    {
        IAS_LOG_ERROR("cfmask_object_cloud_shadow function call");
        free(buf_scaled_values[IAS_THERMAL1_BAND]);
        ias_angle_gen_free_transformation(&angle_metadata);
        ias_angle_gen_free(&angle_metadata);
        return ERROR;
    }

    free(buf_scaled_values[IAS_THERMAL1_BAND]);
    buf_scaled_values[IAS_THERMAL1_BAND] = NULL;

    ias_angle_gen_free_transformation(&angle_metadata);
    ias_angle_gen_free(&angle_metadata);

    return SUCCESS;
}

/*------------------------------------------------------------------------------
NAME: run_algorithm_cfmask

PURPOSE:
    This function will examine every sample of cloud mask buffer and
    determine a classification mask for them.

RETURNS:
    SUCCESS
    ERROR

NOTE:
    Pixel classification  Value
    --------------------  -----
    CLEAR PIXEL           0
    WATER PIXEL           1
    CLOUD SHADOW PIXEL    2
    SNOW PIXEL            3
    CLOUD PIXEL           4
    FILL PIXEL            255
------------------------------------------------------------------------------*/
void run_algorithm_cfmask
(
    int nSamples,                  /* I: Samples per line */
    int nLine,                     /* I: Line of the image */
    const unsigned char *cloudMask,/* I: Cfmask cloud mask buffer */
    const unsigned char *conf_mask,/* I: Cfmask cloud confidence level buffer */
    unsigned short sampleMasks[]   /* O: Line of masks computed by the
                                      algorithm */
)
{
    int i;
    int mask_index = nLine * nSamples;
    unsigned char mask_value;

    /* Define a lookup table to convert the cfmask value to the confidence
       bits.  Since water is not included in the quality band, the water pixels
       translate to low confidence for the other classifications.  The cloud
       confidence bits come from the confidence mask, so the cloud value is
       ignored in the cloudMask.  The entries match the CF_*_PIXEL defines:
        CF_CLEAR_PIXEL => low confidence for all
        CF_WATER_PIXEL => high water confidence, low for rest or may be medium
                          cloud  (water not included)
        CF_CLOUD_SHADOW_PIXEL => high cloud shadow confidence, low for rest
                                 or may be medium cloud too
        CF_SNOW_PIXEL => high snow confidence, low for rest or may be medium
                         cloud too
        CF_CLOUD_PIXEL => ignored since the confidence mask supplies this */
    const static unsigned short maskBits[] =
        {CC_SNOW_LOW  | CC_CLOUD_SHADOW_LOW,   /* clear, so all low */
         CC_SNOW_LOW  | CC_CLOUD_SHADOW_LOW,   /* water, so all low */
         CC_SNOW_LOW  | CC_CLOUD_SHADOW_HIGH,  /* shadow */
         CC_SNOW_HIGH | CC_CLOUD_SHADOW_LOW,   /* snow */
         CC_SNOW_LOW  | CC_CLOUD_SHADOW_LOW};  /* cloud, so all low */

    for (i = 0; i < nSamples; i++, mask_index++)
    {
        mask_value = cloudMask[mask_index];

        if (mask_value == CF_FILL_PIXEL)
        {
            /* The pixel is fill */
            sampleMasks[i] = CC_MASK_FILL;
        }
        else if (mask_value < 5)
        {
            /* Translate the cfmask value to confidence values for the snow
               and shadow bits, and get the cloud confidence bits from the 
               confidence mask (those bits are defined in the correct position
               for the quality band, so no remapping needed) */
            sampleMasks[i] = maskBits[mask_value] | conf_mask[mask_index];
        }
        else
        {
            /* Exit here since the cloud mask isn't a legal value, which
               should never happen */
            IAS_LOG_ERROR("Classifying cfmask pixel");
            exit(EXIT_FAILURE);
        }
    }
}
