#ifndef ALGORITHM_H
#define ALGORITHM_H

#include "create_saturation_mask.h"

void check_algorithm
(
    L1GIO *fdImageIn,               /* I: File descriptor to input image */
    int bandType2Number[IAS_NUM_BAND_TYPES],/* I: Map band type to band number*/
    const CLOUD_COVER_CTRL *pCtrl,  /* I: Control parameters */
    double sun_elevation,           /* I: Sun elevation angle */
    CC_ALGORITHM algorithm,         /* I: Algorithm requested */
    int *use_thermal_for_cfmask,    /* I/O: Thermal band control for CFmask*/
    int *use_cirrus_for_cfmask,     /* I/O: Cirrus band control for CFmask */
    int *can_run,                   /* O: Flag indicating the algorithm
                                           can be run */
    int *can_run_cirrus,            /* O: Flag indicating the cirrus
                                          algorithm can be run */
    int bandFlags[IAS_NUM_BAND_TYPES]/* O: Bands to be opened */
);

void init_algorithm_masks
(
    int nSamples,/* I: Samples per line */
    int nLines2Read, /* I: Lines in each algorithm's buffer */
    unsigned short * sampleMasks[NUM_BUFFERS]/* IO: Algorithm sample masks */
);

void resolve_algorithm_masks
(
    int can_run,                /* I: Flag that the primary algorithm can run */
    int can_run_cirrus,         /* I: Flag that the Cirrus algorithm ran */
    int nSamples,               /* I: Samples per line */
    unsigned short * const sampleImage[IAS_NUM_BAND_TYPES],/*I: Image DN
                                                                values*/
    const unsigned short *terrainMask,/* I: Terrain mask */
    unsigned short *sampleMasks[NUM_BUFFERS],/* IO: Sample masks */
    const unsigned char *landWaterMask,/* I: Land/water mask for current line
                                             (1 = land, 0 = water) */
    const unsigned char *l1r_saturation_mask, /* I: Saturated pixel mask for
                                                    the L1R image */
    const unsigned char *l1g_saturation_mask, /* I: Saturated pixel mask for
                                                    the L1G image */
    int *nFill,/* O: Number of fill samples */
    int *nClouds,/* O: Number of high confidence cloud samples */
    int *nSnowIces,/* O: Number of high confidence snow/ice samples */
    int *nCloudsLand,/* O: Total cloud cover on land */
    int *nTotalLand/* O: Total land in an image */
);

void run_algorithm_acca
(
    int nSamples,/* I: Samples per line */
    float *sampleScaled[IAS_NUM_BAND_TYPES],/* I: Array of pointers to scaled
                                               TOA values, 1/band */
    unsigned short sampleMasks[]/* O: Line of sample masks computed */
);

void run_algorithm_atacca
(
    int nSamples,/* I: Samples per line */
    float sineSolarElev,/* I: Sine solar elevation */
    float *sampleScaled[IAS_NUM_BAND_TYPES],/* I: Array of pointers to scaled
                                               TOA values, 1/band */
    unsigned short sampleMasks[]/* O: Line of sample masks computed */
);

int create_cirrus_threshold_lookup_table
(
    const double *cirrus_values,    /* I: Array of 4 Cirrus threshold values */
    const int min_elev,             /* I: Minimum elevation */
    const int max_elev,             /* I: Maximum elevation
                                       table size = max - min */
    double **table                  /* O: Lookup table created */
);

void run_algorithm_cirrus
(
    int nSamples,                           /* I: Samples per line */
    const float *scaled_samples,            /* I: Array of pointers to scaled
                                               TOA values, 1/band */
    const short *dem_line,                  /* I: Array of elevations */
    const double *threshold_lut,            /* I: Elevation to threshold LUT */
    int min_elev,                           /* I: Minimum elevation of DEM */ 
    int max_elev,                           /* I: Maximum elevation of DEM */
    unsigned short sampleMasks[]            /* O: Line of masks computed by the
                                               algorithm */
);

void run_algorithm_see5
(
    int nSamples,/* I: Samples per line */
    float sineSolarElev,/* I: Sine solar elevation */
    float *sampleScaled[IAS_NUM_BAND_TYPES],/* I: Array of pointers to scaled
                                               TOA values, 1/band */
    unsigned short sampleMasks[]/* O: Line of sample masks computed */
);

int cfmask_generate_cloud_mask
(
    int max_threads,            /* I: Maximum threads to use */
    int n_samples,              /* I: Samples per line */
    int n_lines_in_image,       /* I: Total number of lines */
    float cfmask_cloud_prob_threshold, /* I: Cloud Probability threshold */
    double solar_elevation,     /* I: Solar elevation angle*/
    double solar_azimuth,       /* I: Sun azimuth angle */ 
    const IAS_L1G_FILE_METADATA *file_metadata, /* L1G file metadata to support
                                        calculating the map azimuth angle */
    IAS_L1G_BAND_METADATA metadata_in[IAS_NUM_BAND_TYPES],/* I: L1G Band 
                                                             Metadata */
    L1G_BAND_IO * fd_bands_in[IAS_NUM_BAND_TYPES],/* I: Input bands file
                                                     descriptors */ 
    int band_type_2number[],    /* I: Map band type to band number */
    const CLOUD_COVER_CTRL *pCtrl,/* I: Control parameters */ 
    const char *angle_coeff_filename, /* I: Angle coefficient filename */
    int cfmask_dilate_flag,       /* I: Flag to turn on/off cloud dilation */
    int use_thermal,              /* I: Flag to use thermal band */
    int use_cirrus,               /* I: Flag to use cirrus band */
    unsigned char *cloud_mask,    /* O: Buffer for cloudMask */
    unsigned char *conf_mask,     /* O: Pointer for the cloud confidence 
                                        level*/
    SATURATION_MASK *sat_mask     /* I/O: Saturated pixel tracking info */
);

void run_algorithm_cfmask
(
    int nSamples,               /* I: Samples per line */
    int nLine,                  /* I: Line of an image */
    const unsigned char *cloudMask,/* I: Cfmask cloud mask buffer */
    const unsigned char *conf_mask,/* I: Cfmask cloud confidence level buffer */
    unsigned short sampleMasks[]/* O: Line of masks computed by the
                                   algorithm */
);

#endif
