/*==============================================================================
NAME: assess_cloud_cover

PURPOSE:
    This application performs cloud cover assessment for L1G and L1T products.
    Five different algorithms are incorporated into this application. One
    of the following four (Cirrus always runs if the band is available)
    algorithms may be specified:

        1) Automated Cloud Cover Assessment (ACCA)
        2) Artificial Thermal Automated Cloud Cover Assessment (ATACCA)
        3) See 5
        4) Cfmask (the default)

    Each sample is assigned a low, medium, or high confidence level for each of
    4 classifications: clouds, cirrus, snow/ice, and cloud shadows.

    Once the cloud cover assessment algorithms have classified the samples, the
    terrain occlusion mask, if present, is combined into a final quality band
    mask.

ALGORITHM REFERENCES:
    Cloud Cover Assessment (CCA) Algorithm Description Document (ADD)
==============================================================================*/

/* System Includes */
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <math.h>

/* IAS Includes */
#include "ias_logging.h"
#include "ias_geo.h"
#include "ias_miscellaneous.h"
#include "ias_structures.h"
#include "ias_pixel_mask.h"
#include "ias_model_io.h"

/* Local Includes */
#include "assess_cloud_cover.h"
#include "ctrl_parms.h"
#include "algorithm.h"
#include "get_parameters.h"
#include "get_names.h"
#include "update_omf.h"
#include "cfmask_support.h"
#include "create_saturation_mask.h"

#define TARGET_MEMORY 134217728         /* 128 MB == 2^27 */

/*------------------------------------------------------------------------------
NAME: free_memory

PURPOSE:
    Free the memory used in the cloud cover assessment algorithms.
------------------------------------------------------------------------------*/
static void free_memory
(
    unsigned short *bufImage[IAS_NUM_BAND_TYPES],/* I: Buffer of DN values */
    float *bufScaled[IAS_NUM_BAND_TYPES],/* I: Buffer of scaled values */
    unsigned short *bufMasks[NUM_BUFFERS],/* I: Classification masks buffers */
    unsigned short **bufTerrain          /* I: Buffer of terrain masks */
)
{
    int bandType;
    int mask;

    for (bandType = 0; bandType < IAS_NUM_BAND_TYPES; bandType++)
    {
        free(bufImage[bandType]);
        bufImage[bandType] = NULL;

        free(bufScaled[bandType]);
        bufScaled[bandType] = NULL;
    }

    for (mask = 0; mask < NUM_BUFFERS; mask++)
    {
        free(bufMasks[mask]);
        bufMasks[mask] = NULL;
    }

    free(*bufTerrain);
    *bufTerrain = NULL;
}


/*------------------------------------------------------------------------------
NAME: alloc_memory

PURPOSE:
    Allocate memory to be used in the cloud cover assessment algorithms.  This
    application needs one buffer to hold the image data that is read from the
    L1G input image; a buffer to hold floating point, scaled image data; one
    buffer to hold the cloud cover masks that are assigned by the individual
    algorithms; one buffer to hold the terrain occlusion mask; and one buffer
    to hold the DEM file.

    Due to the fact that the algorithms are complicated and there are many
    references to the individual bands, these arrays are indexed by the
    IAS_BAND_TYPE to make it easier and more clear as to which bands are being
    accessed.  This feature will help keep the algorithms easy to read and
    accurate.

    The bufMasks buffer is indexed by the cloud cover algorithms, including one
    additional index that stores the final quality band mask that is resolved
    from the individual algorithms.  In addition, if a terrain occlusion image
    is supplied, the terrain occlusion mask is incorporated into this final
    quality band as well.

RETURNS:
    This function will return a status of SUCCESS if successful, ERROR if an
    error is encountered.
------------------------------------------------------------------------------*/
static int alloc_memory
(
    int nLines,                          /* I: Lines in input image */
    int nSamples,                        /* I: Samples per line */
    L1G_BAND_IO *
        fdBandsIn[IAS_NUM_BAND_TYPES],   /* I: Opened bands */
    L1G_BAND_IO *fdTerrain,              /* I: Terrain band */
    int targetMemorySize,                /* I: Memory usage for IO */
    unsigned short *
        bufImage[IAS_NUM_BAND_TYPES],    /* IO: Buffer of DN values */
    float *bufScaled[IAS_NUM_BAND_TYPES],/* IO: Buffer of scaled samples */
    unsigned short *bufMasks[NUM_BUFFERS],/* IO: Classification masks buffers */
    unsigned short **bufTerrain,         /* IO: Buffer terrain masks */
    int *nLines2Read                     /* O: Lines to read at a time */
)
{
    int bandType;
    int mask;
    int nInputBands     = 0;
    int samplesInBuffer = 0;
    int sizeInputLine   = 0;
    int sizeMaskLine    = 0;
    int sizeScaledLine  = 0;
    int sizeTerrainLine = 0;
    int sizeDemLine     = 0;
    int sizeTotal       = 0;

    /* How many bands are we processing from the input image? */
    for (bandType = 0; bandType < IAS_NUM_BAND_TYPES; bandType++)
        if (fdBandsIn[bandType] != NULL)
            nInputBands++;

    /* Compute the amount of memory required for a single line of data across
       all of the bands being processed from the input image.  Remember, the
       input data is read as a 2 byte unsigned integer and is then scaled into a
       floating point value. */
    sizeInputLine  = nSamples * nInputBands * sizeof(unsigned short);
    sizeScaledLine = nSamples * nInputBands * sizeof(float);

    /* Compute the amount of memory that is required to store the masks that are
       generated by each algorithm along with a final resovled mask for a single
       line of the input image. */
    sizeMaskLine   = nSamples * NUM_BUFFERS * sizeof(unsigned short);

    /* If a terrain occlusion mask has been provided how much memory is required
       for a single line of data? */
    if (fdTerrain != NULL)
        sizeTerrainLine = nSamples * sizeof(unsigned short);

    /* Compute the total memory requirement for a single line of data */
    sizeTotal = sizeInputLine + sizeScaledLine + sizeMaskLine + sizeTerrainLine
        + sizeDemLine;

    /* Now compute the number of lines that we can fit in a target memory
       footprint.  This will then be passed to the I/O routines to read the
       specified number of lines as a single block of data. */
    *nLines2Read = targetMemorySize / sizeTotal;
    if (*nLines2Read < 1)
        *nLines2Read = 1;
    else if (*nLines2Read > nLines)
        *nLines2Read = nLines;

    /* Number of samples in the buffer */
    samplesInBuffer = *nLines2Read * nSamples;

    if (IAS_LOG_DEBUG_ENABLED())
    {
        printf("=======================================================\n");
        printf("= I/O Memory Usage                                    =\n");
        printf("=======================================================\n");
        printf("    target memory:      %10d\n", targetMemorySize);
        printf("    nLines:             %10d\n", nLines);
        printf("    nSamples:           %10d\n", nSamples);
        printf("    nBands:             %10d\n", nInputBands);
        printf("\n");
        printf("    bytes/input line:   %10d\n", sizeInputLine);
        printf("    bytes/scaled line:  %10d\n", sizeScaledLine);
        printf("    bytes/mask line:    %10d\n", sizeMaskLine);
        printf("    bytes/terrain line: %10d\n", sizeTerrainLine);
        printf("                        ----------\n");
        printf("    total bytes/line:   %10d\n", sizeTotal);
        printf("\n");
        printf("    Lines/block:        %10d\n\n", *nLines2Read);
        printf("\n");
    }

    /* Allocate the memory for the input image */
    for (bandType = 0; bandType < IAS_NUM_BAND_TYPES; bandType++)
    {
        /* This application uses two different flags to indicate if a band has
           been successfully opened for processing.  If either the file
           descriptor or the buffer that contains the raw image data are NULL,
           that band was not opened for processing.  Use the file descriptor's
           NULL value to ensure the memory pointed to by bufImage is also NULL
           for these bands. */
        if (fdBandsIn[bandType] == NULL)
            continue;

        bufImage[bandType] = malloc(samplesInBuffer * sizeof(unsigned short));
        if (bufImage[bandType] == NULL)
        {
            IAS_LOG_ERROR("Allocating memory for nLines=%d x nSamples=%d",
               *nLines2Read, nSamples);
            free_memory(bufImage, bufScaled, bufMasks, bufTerrain);
            return ERROR;
        }

        bufScaled[bandType] = malloc(samplesInBuffer * sizeof(float));
        if (bufScaled[bandType] == NULL)
        {
            IAS_LOG_ERROR("Allocating memory for nLines=%d x nSamples=%d",
               nLines, nSamples);
            free_memory(bufImage, bufScaled, bufMasks, bufTerrain);
            return ERROR;
        }
    }

    /* Allocate the memory needed for each of the cloud cover algorithms */
    for (mask = 0; mask < NUM_BUFFERS; mask++)
    {
        bufMasks[mask] = malloc(samplesInBuffer * sizeof(unsigned short));
        if (bufMasks[mask] == NULL)
        {
            IAS_LOG_ERROR("Allocating memory for nLines=%d x nSamples=%d",
               nLines, nSamples);
            free_memory(bufImage, bufScaled, bufMasks, bufTerrain);
            return ERROR;
        }
    }

    /* Allocate the memory needed to read the terrain occlusion mask */
    if (fdTerrain != NULL)
    {
        *bufTerrain = malloc(samplesInBuffer * sizeof(unsigned short));
        if (*bufTerrain == NULL)
        {
            IAS_LOG_ERROR("Allocating memory for nLines=%d x nSamples=%d",
                nLines, nSamples);
            free_memory(bufImage, bufScaled, bufMasks, bufTerrain);
            return ERROR;
        }
    }

    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: set_projection

PURPOSE: Helper routine to setup a projection structure.
------------------------------------------------------------------------------*/
static int set_projection
(
    const IAS_L1G_FILE_METADATA *file_metadata,/* I: file metadata structure */
    IAS_PROJECTION *proj    /* O: Populated projection */
)
{
    int units;

    if (ias_geo_get_units(file_metadata->projection_units, &units) != SUCCESS)
    {
       IAS_LOG_ERROR("Projection unit conversion failed");
       return ERROR;
    }

    /* Set the information in the projection */
    ias_geo_set_projection(file_metadata->projection_code,
        file_metadata->zone_code, units, file_metadata->spheroid_code,
        file_metadata->projection_parameters, proj);

    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: scale_samples

PURPOSE:
    This function will scale L1G samples from their DN value coming out of the
    image into the appropriate value.  OLI bands will be converted to Top Of
    Atmosphere (TOA) reflectance and the TIRS band will be converted to TOA
    radiance value.
------------------------------------------------------------------------------*/
static void scale_samples
(
    unsigned short imageValues[], /* I: Line of DN values from input image */
    int nSamples,                 /* I: Samples per line */
    double scaleFactor,           /* I: Scaling factor */
    double scaleShift,            /* I: Scaling shift */
    double bandScaling,           /* I: Additional band scaling */
    float scaledValues[]          /* O: Line of scaled sample values */
)
{
    int i;

    for (i = 0; i < nSamples; i++)
    {
        scaledValues[i] = ((double)imageValues[i] * scaleFactor + scaleShift)
            * bandScaling;
    }
}


/*------------------------------------------------------------------------------
NAME: close_bands

PURPOSE:
    Close all of the L1G image bands that this application has opened.

RETURNS:
    This function returns a status of SUCCESS if the output quality band image
    is successfully closed, ERROR otherwise.
------------------------------------------------------------------------------*/
static int close_bands
(
    L1G_BAND_IO *fdIn[IAS_NUM_BAND_TYPES], /* IO: Input bands */
    L1G_BAND_IO **fdTerrain,               /* IO: Terrain occlusion band */
    L1G_BAND_IO **fdQualityOut             /* IO: Output band */
)
{
    int i;
    int isError = FALSE;

    for (i = 0; i < IAS_NUM_BAND_TYPES; i++)
    {
        if (fdIn[i] != NULL)
        {
            ias_l1g_close_band(fdIn[i]);
            fdIn[i] = NULL;
        }
    }

    if (fdTerrain && *fdTerrain != NULL)
    {
        ias_l1g_close_band(*fdTerrain);
        *fdTerrain = NULL;
    }

    if (fdQualityOut && *fdQualityOut != NULL)
    {
        if (ias_l1g_close_band(*fdQualityOut) != SUCCESS)
        {
            IAS_LOG_ERROR("Closing quality band");
            isError = TRUE;
        }
        *fdQualityOut = NULL;
    }

    if (isError)
        return ERROR;
    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: open_input_bands

PURPOSE:
    Open all of the input bands for reading.  This function assumes that all of
    the bands have been examined and are the same dimensions.

RETURNS:
    This function will return a status of SUCCESS if the bands are successfully
    opened, ERROR if an error is encountered.
------------------------------------------------------------------------------*/
static int open_input_bands
(
    L1GIO *fdImageIn,                       /* I: Input image file descriptor */
    int bandType2Number[IAS_NUM_BAND_TYPES],/* I: Map band type to band number*/
    int bandFlags[IAS_NUM_BAND_TYPES],      /* I: Bands to be opened */
    L1G_BAND_IO *fdBandsIn[IAS_NUM_BAND_TYPES],
                                            /* O: Input bands file descriptors*/
    IAS_L1G_BAND_METADATA metadataIn[IAS_NUM_BAND_TYPES]
                                            /* O: Metadata for input bands */
)
{
    int bandType;
    /* Open each of the required bands from the input image */
    for (bandType = 0; bandType < IAS_NUM_BAND_TYPES; bandType++)
    {
        int bandNumber;
        IAS_DATA_TYPE tmpDataType;
        int tmpLines;
        int tmpSamples;
        int tmpScas;

        /* Only open the bands that will be required to run the algorithms */
        if (!bandFlags[bandType])
            continue;

        bandNumber = bandType2Number[bandType];

        /* Retrieve band metadata */
        if (ias_l1g_get_band_metadata(fdImageIn, bandNumber,
                &metadataIn[bandType]) != SUCCESS)
        {
            IAS_LOG_ERROR("Retrieving metadata for band %d", bandNumber);
            close_bands(fdBandsIn, NULL, NULL);
            return ERROR;
        }

        /* Open the band for processing purposes */
        fdBandsIn[bandType] = ias_l1g_open_band(fdImageIn, bandNumber,
            &tmpDataType, &tmpScas, &tmpLines, &tmpSamples);
        if (fdBandsIn[bandType] == NULL)
        {
            IAS_LOG_ERROR("Opening band %d", bandNumber);
            close_bands(fdBandsIn, NULL, NULL);
            return ERROR;
        }

        /* Ensure that the band has valid radiance or reflectance scaling values
           based on the spectral type */
        if (ias_sat_attr_get_spectral_type_from_band_number(bandNumber)
            == IAS_SPECTRAL_THERMAL)
        {
            if (!metadataIn[bandType].radiance_valid)
            {
                IAS_LOG_ERROR(
                    "Band %d does not have valid radiance scaling factors",
                    bandNumber);
                close_bands(fdBandsIn, NULL, NULL);
                return ERROR;
            }
        }
        else if (!metadataIn[bandType].reflectance_valid)
        {
            IAS_LOG_ERROR(
                "Band %d does not have valid reflectance scaling factors",
                bandNumber);
            close_bands(fdBandsIn, NULL, NULL);
            return ERROR;
        }
    }

    return SUCCESS;
}

/*------------------------------------------------------------------------------
NAME: corners_match

PURPOSE:
    Check the corners of two IAS_L1G_BAND_METADATA's to ensure they are within
    a provided threshold.

RETURNS:
    1 for corners differ significantly from the provided value.
    0 for no significant difference.
------------------------------------------------------------------------------*/
int corners_match
(
    const IAS_L1G_BAND_METADATA *md1, /* I: Band 1 */
    const IAS_L1G_BAND_METADATA *md2, /* I: Band 2 */
    double threshold /* I: Corner tolerace threshold. Higher = greater
                        difference accepted. */
)
{
    double corner_diff = md1->projection_distance_x * threshold;

    if (   (fabs(md1->upper_left_y - md2->upper_left_y) > corner_diff)
        || (fabs(md1->upper_left_x - md2->upper_left_x) > corner_diff)
        || (fabs(md1->upper_right_y - md2->upper_right_y) > corner_diff)
        || (fabs(md1->upper_right_x - md2->upper_right_x) > corner_diff)
        || (fabs(md1->lower_left_y - md2->lower_left_y) > corner_diff)
        || (fabs(md1->lower_left_x - md2->lower_left_x) > corner_diff)
        || (fabs(md1->lower_right_y - md2->lower_right_y) > corner_diff)
        || (fabs(md1->lower_right_x - md2->lower_right_x) > corner_diff))
    {
        return 0;
    }
    return 1;
}

/*------------------------------------------------------------------------------
NAME: open_and_check_band

PURPOSE:
    Called by open_bands primarily to check the integrity of DEM and Terrain
    bands. It opens the band for processing, checks its data type, and checks
    its corners.

RETURNS:
    ERROR
    SUCCESS
------------------------------------------------------------------------------*/
int open_and_check_band 
(
    const char *file_type,      /* I: Type of file. Only used for output */
    L1GIO *file,                /* I: L1G file pointer */
    const IAS_L1G_BAND_METADATA *p_metadata, /* I: Comparison metadata */
    IAS_L1G_BAND_METADATA *band_metadata, /* I/O: Band metadata extracted from
                                             file */
    L1G_BAND_IO **fd_band       /* O: File descriptor */
)
{     
    IAS_DATA_TYPE data_type;
    int lines;
    int samples;
    int scas = 1;

    /* There should only be 1 band present */
    if (ias_l1g_get_band_metadata(file, 1, band_metadata) != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving metadata for the %s file", file_type);
        return ERROR;
    }

    /* Open the band for processing purposes */
    *fd_band = ias_l1g_open_band(file, 1, &data_type, &scas, &lines, &samples);
    if (*fd_band == NULL)
    {
        IAS_LOG_ERROR("Opening the %s band", file_type);
        return ERROR;
    }

    if (data_type != IAS_I2)
    {
        IAS_LOG_ERROR("Invalid data type found for the %s band", file_type);
        ias_l1g_close_band(*fd_band);
        *fd_band = NULL;
        return ERROR;
    }

    if (p_metadata != NULL)
    {
        /* Make sure the corners of the image are within 0.5 pixels */
        if (!corners_match(p_metadata, band_metadata, 0.5))
        {
            IAS_LOG_ERROR("Image corners do not match in the %s band", 
                file_type);
            ias_l1g_close_band(*fd_band);
            *fd_band = NULL;
            return ERROR;
        }
    }
    return SUCCESS;
}

/*------------------------------------------------------------------------------
NAME: open_bands

PURPOSE:
    Open all of the input bands for reading and the output quality band for
    writing.  This function assumes that all of the bands have been examined and
    are the same dimensions.

RETURNS:
    This function will return a status of SUCCESS if the bands are successfully
    opened, ERROR if an error is encountered.
------------------------------------------------------------------------------*/
static int open_bands
(
    L1GIO *fdImageIn,                       /* I: Input image file descriptor */
    L1GIO *fdTerrainIn,                     /* I: Terrain occlusion mask */
    L1GIO *fdQualityOut,                    /* I: Output image file descriptor*/
    int bandType2Number[IAS_NUM_BAND_TYPES],/* I: Map band type to band number*/
    int bandFlags[IAS_NUM_BAND_TYPES],      /* I: Bands to be opened */
    int nLines,                             /* I: Lines in image */
    int nSamples,                           /* I: Samples in image */
    L1G_BAND_IO *
        fdBandsIn[IAS_NUM_BAND_TYPES],      /* O: Input bands file descriptors*/
    IAS_L1G_BAND_METADATA
        metadataIn[IAS_NUM_BAND_TYPES],     /* O: Metadata for input bands */
    L1G_BAND_IO **fdBandTerrain,            /* O: Terrain band file descriptor*/
    L1G_BAND_IO **fdBandOut,                /* O: Output band file descriptor */
    IAS_L1G_BAND_METADATA *metadataOut      /* O: Metadata for output band */
)
{
    IAS_DATA_TYPE           tmpDataType;
    IAS_L1G_BAND_METADATA   terrainMetadata;
    IAS_L1G_BAND_METADATA * pMetadata       = NULL;
    int                     bandType;
    int                     tmpScas;

    /* Open each of the required bands from the input image */
    if (open_input_bands(fdImageIn, bandType2Number, bandFlags, fdBandsIn,
            metadataIn) != SUCCESS)
    {
        IAS_LOG_ERROR("Opening input image bands");
        return ERROR;
    }

    /* Find the first available band to use for the metadata */
    for (bandType = 0; bandType < IAS_NUM_BAND_TYPES; bandType++)
    {
        if (bandFlags[bandType])
        {
            pMetadata = &metadataIn[bandType];
            break;
        }
    }

    /* Open the terrain occlusion mask if one was specified */
    if (fdTerrainIn != NULL)
    {
        if (open_and_check_band("terrain occlusion", fdTerrainIn, 
            pMetadata, &terrainMetadata, fdBandTerrain) != 0)
        {
            IAS_LOG_ERROR("Error checking terrain occlusion band");
            close_bands(fdBandsIn, fdBandTerrain, fdBandOut);
            return ERROR;
        }

        if (pMetadata == NULL)
        {
            pMetadata = &terrainMetadata;
        }
    }

    /* Now we need to open the output image band for writing */
    tmpDataType = IAS_UI2;
    tmpScas = 1;
    *fdBandOut = ias_l1g_open_band(fdQualityOut, 1, &tmpDataType, &tmpScas,
        &nLines, &nSamples);
    if (*fdBandOut == NULL)
    {
        IAS_LOG_ERROR("Opening output band for writing");
        close_bands(fdBandsIn, fdBandTerrain, fdBandOut);
        return ERROR;
    }

    /* Copy the first available band's metadata to the output band's metadata.
       This pointer should not be NULL at this point but we'll perform a sanity
       check here just to be safe. */
    if (pMetadata == NULL)
    {
        IAS_LOG_ERROR("Unable to copy band metadata to quality band image");
        close_bands(fdBandsIn, fdBandTerrain, fdBandOut);
        return ERROR;
    }
    memcpy(metadataOut, pMetadata, sizeof(IAS_L1G_BAND_METADATA));
    strcpy(metadataOut->band_name, "Quality Band");
    metadataOut->band_number = 1;
    metadataOut->maximum_pixel_value = 0.0;
    metadataOut->minimum_pixel_value = 0.0;
    metadataOut->pixel_range_valid = 0;
    metadataOut->maximum_radiance = 0.0;
    metadataOut->minimum_radiance = 0.0;
    metadataOut->spectral_radiance_scaling_offset = 0.0;
    metadataOut->spectral_radiance_scaling_gain = 1.0;
    metadataOut->radiance_valid = 0;
    metadataOut->reflectance_scaling_offset = 0.0;
    metadataOut->reflectance_scaling_gain = 1.0;
    metadataOut->reflectance_valid = 0;
    strcpy(metadataOut->instrument_source, "calculated");
    if (ias_l1g_set_band_metadata(fdQualityOut, metadataOut, 1) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing band metadata for output quality band");
        close_bands(fdBandsIn, fdBandTerrain, fdBandOut);
        return ERROR;
    }
    return SUCCESS;
}

/*------------------------------------------------------------------------------
NAME: verify_band_size

PURPOSE:
    Receives target dimensions to verify a single band's size

RETURNS:
    SUCCESS
    ERROR
------------------------------------------------------------------------------*/
int verify_band_size
(
    const char *file_type,  /* I: Type of file. Used for output statements */
    L1GIO *file,            /* I: File who's size we're checking */
    int *targetLine,        /* IO: Target line */
    int *targetSample,      /* IO: Target sample */
    int *targetSca          /* IO: Target SCA */
)
{
    int tmpSca;
    int tmpLine;
    int tmpSample;
    IAS_DATA_TYPE tmpDataType;

    /* File should not be NULL */
    if (file == NULL)
    {
        IAS_LOG_ERROR("%s band cannot be NULL", file_type);
        return ERROR; 
    }

    /* The L1G file will only have 1 band */
    if (ias_l1g_get_band_size(file, 1, &tmpDataType, &tmpSca,
            &tmpLine, &tmpSample) != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving size information for %s band", file_type);
        return ERROR;
    }
 
    if (*targetLine < 0)
    {
        /* At this point it is possible that we won't be processing any
           bands from the input image.  This may occur if we are processing
           a TIRS only image or if there are bands missing from the input
           image that will prevent all algorithms from running.  This
           application must still run because we'll be translating the
           terrain occlusion mask from a 0/1 into a different bit position
           in the quality band. */
        *targetLine = tmpLine;
        *targetSample = tmpSample;
        *targetSca = tmpSca;
    }
    else if ((tmpLine != *targetLine) || (tmpSample != *targetSample)
        || (tmpSca != *targetSca))
    {
        IAS_LOG_ERROR("%s band does not have the same dimensions", file_type);
        return ERROR;
    }

    if ((tmpDataType != IAS_I2) && (tmpDataType != IAS_UI2))
    {
        IAS_LOG_ERROR("%s band data type is not 16 bit", file_type);
        return ERROR;
    }
    return SUCCESS;
}

/*------------------------------------------------------------------------------
NAME: verify_band_sizes

PURPOSE:
    This function will verify that all of the input bands are the same size.

RETURNS:
    The function will return a status of SUCCESS if the input bands are the same
    size, ERROR if any band sizes differ in size.
------------------------------------------------------------------------------*/
static int verify_band_sizes
(
    L1GIO * fdImageIn,                      /* I: Input L1G image */
    L1GIO * fdTerrainIn,                    /* I: Terrain occlusion mask */
    int bandType2Number[IAS_NUM_BAND_TYPES],/* I: Map band type to band number*/
    int bandFlags[IAS_NUM_BAND_TYPES],      /* I: Bands to be processed */
    int *nLines,                            /* O: Number of lines in image */
    int *nSamples                           /* O: Number of samples in image */
)
{
    IAS_DATA_TYPE targetDataType = IAS_UI2;
    IAS_DATA_TYPE tmpDataType;
    int           bandNumber;
    int           bandType;
    int           targetLine     = -1;
    int           targetSample   = -1;
    int           targetSca      = -1;
    int           tmpLine;
    int           tmpSample;
    int           tmpSca;

    *nLines = -1;
    *nSamples = -1;

    /* Examine the bands from the input image that we are going to process */
    for (bandType = 0; bandType < IAS_NUM_BAND_TYPES; bandType++)
    {
        /* Only examine bands that will be required to run the algorithms */
        if (!bandFlags[bandType])
            continue;
        bandNumber = bandType2Number[bandType];

        if (ias_l1g_get_band_size(fdImageIn, bandNumber, &tmpDataType, &tmpSca,
                &tmpLine, &tmpSample) != SUCCESS)
        {
            IAS_LOG_ERROR("Retrieving size information for band %d",
                bandNumber);
            return ERROR;
        }

        if (targetLine < 0)
        {
            /* Do not set targetDataType, use initialized value from above */
            targetLine = tmpLine;
            targetSample = tmpSample;
            targetSca = tmpSca;
        }
        else if (tmpDataType != targetDataType)
        {
            IAS_LOG_ERROR("Band %d data type is not IAS_UI2", bandNumber);
            return ERROR;
        }
        else if ((tmpLine != targetLine) || (tmpSample != targetSample)
            || (tmpSca != targetSca))
        {
            IAS_LOG_ERROR("Bands do not have the same dimensions");
            return ERROR;
        }
    }

    if (fdTerrainIn != NULL && verify_band_size("terrain occlusion", 
        fdTerrainIn, &targetLine, &targetSample, &targetSca) != SUCCESS)
    {
        IAS_LOG_ERROR("Verifying terrain occlusion band size");
        return ERROR;
    }

    if (targetLine < 0)
    {
        IAS_LOG_ERROR("No bands available for processing");
        return ERROR;
    }

    /* This application only supports L1G and L1T images that have had their
       SCA's combined by the resampler.  If the band has multiple SCA's issue
       and error and return. */
    if (targetSca != 1)
    {
        IAS_LOG_ERROR(
            "Found multiple SCAS (%d).  Only SCA combined images supported.",
            targetSca);
        return ERROR;
    }

    *nLines = targetLine;
    *nSamples = targetSample;
    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: generate_land_water_mask

PURPOSE:
    This function creates the land/water mask (land = 1) for the area covered
    by the scene.  If the output mask image name is provided, the mask image
    is written to a file with that name.

RETURNS:
    SUCCESS/ERROR
------------------------------------------------------------------------------*/
static int generate_land_water_mask
(
    const IAS_L1G_FILE_METADATA *file_metadata,/* I: file metadata structure */
    const IAS_L1G_BAND_METADATA *band_metadata,/* I: band metadata structure */
    const char pathWaterMaskPolygon[],  /* I: Path to water mask polygon file */
    const char pathOutLandWaterMaskImage[],/* I: Path to land water mask 
                                                 file */ 
    int nLinesInImage,                  /* I: Number of lines in image */
    int nSamples,                       /* I: Number of samples in image */
    unsigned char *landWaterMask        /* IO: Land water mask buffer */
)
{
    IAS_IMAGE mask_image;               /* Image data used to build mask */
    IAS_PROJECTION mask_projection;     /* Projection data */
    
    /* Set the l1g image band metadata to image structure */
    mask_image.corners = (struct IAS_CORNERS)
        {
            {band_metadata->upper_left_x, band_metadata->upper_left_y},
            {band_metadata->upper_right_x, band_metadata->upper_right_y},
            {band_metadata->lower_left_x, band_metadata->lower_left_y},
            {band_metadata->lower_right_x, band_metadata->lower_right_y}
        };
    mask_image.pixel_size_y = band_metadata->projection_distance_y;
    mask_image.pixel_size_x = band_metadata->projection_distance_x;
    mask_image.nl = nLinesInImage;
    mask_image.ns = nSamples;
    mask_image.band_number = band_metadata->band_number;

    /* Set the projection contents */
    if (set_projection(file_metadata, &mask_projection) != SUCCESS)
    {
        IAS_LOG_ERROR("Projection creation failed");
        return ERROR;
    }

    if (IAS_LOG_DEBUG_ENABLED())
    {
        printf("==================================================\n");
        printf("= Summary of land/water image and projection     =\n");
        printf("==================================================\n");
        printf("   upper_left_x = %lf\n", mask_image.corners.upleft.x);
        printf("   upper_left_y = %lf\n", mask_image.corners.upleft.y);
        printf("  upper_right_x = %lf\n", mask_image.corners.upright.x);
        printf("  upper_right_y = %lf\n", mask_image.corners.upright.y);
        printf("   lower_left_x = %lf\n", mask_image.corners.loleft.x);
        printf("   lower_left_y = %lf\n", mask_image.corners.loleft.y);
        printf("  lower_right_x = %lf\n", mask_image.corners.loright.x);
        printf("  lower_right_y = %lf\n", mask_image.corners.loright.y);
        printf("   pixel_size_y = %lf\n", mask_image.pixel_size_y);
        printf("   pixel_size_x = %lf\n", mask_image.pixel_size_x);
        printf("  nLinesInImage = %d\n", mask_image.nl);
        printf("       nSamples = %d\n", mask_image.ns);
        printf("    band_number = %d\n", mask_image.band_number);
        printf("projection code = %d\n", mask_projection.proj_code);
        printf("      zone code = %d\n", mask_projection.zone);
        printf("  spheroid code = %d\n", mask_projection.spheroid);
        printf("          units = %d\n", mask_projection.units);
    }

    if (ias_geo_shape_mask_projection(pathWaterMaskPolygon,
        &mask_image, &mask_projection, landWaterMask) 
        != SUCCESS)
    {
        IAS_LOG_ERROR("Creating land and water mask");
        return ERROR;
    }
       
    /* If the land water mask file name is input then dump the image 
       band file for debugging */
    if (strcmp(pathOutLandWaterMaskImage,"") != 0)
    { 
        if (ias_misc_write_envi_image(pathOutLandWaterMaskImage,
            &mask_projection, "Assess Cloud Cover Water Mask",
            nLinesInImage, nSamples, band_metadata->upper_left_x,
            band_metadata->upper_left_y,band_metadata->projection_distance_x,
            band_metadata->projection_distance_y, "Water Mask",
            IAS_BYTE, landWaterMask) != SUCCESS)
        {
            IAS_LOG_ERROR("Dumping land and water mask image");
            return ERROR;
        }
        IAS_LOG_INFO("Created land water mask image: %s",
            pathOutLandWaterMaskImage);
    }

    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: process_bands

PURPOSE:
    This function will read the image data from each of the necessary bands and
    make the appropriate calls to the individual CCA algorithms to classify the
    entire line of samples.

RETURNS:
    The function will return a status of SUCCESS if the image is successfully
    processed, ERROR if an error is encountered.
------------------------------------------------------------------------------*/
static int process_bands
(
    const char pathOmf[],          /* I: Path to the OMF */
    const char pathOut[],          /* I: Path to the output band */
    const char pathWaterMaskPolygon[],/* I: Path to water mask polygon file */
    const char pathOutLandWaterMaskImage[],/* I: Path to land water mask file */
    const char pathOutCfmask[],    /* I: Path to the cfmask band */
    const char pathDem[],          /* I: Path to the DEM */
    const char angle_coeff_filename[], /* I: Angle coefficient filename */
    float cfmaskCloudProbThreshold, /* I: Cloud probability threshold for
                                            cfmask */
    int cfmask_dilate_flag,        /* I: Flag to turn on/off cloud dilation */
    L1GIO * fdImageIn,             /* I: Input L1G image */
    L1GIO * fdTerrainIn,           /* I: Terrain occlusion mask */
    L1GIO * fdQualityOut,          /* I: Output L1G quality band */
    const CLOUD_COVER_CTRL *pCtrl, /* I: Cloud cover parameters */
    const IAS_L1G_FILE_METADATA *file_metadata,/* I: file metadata structure */
    double sineSolarElev,          /* I: Sine solar elevation */
    CC_ALGORITHM algorithm,        /* I: Algorithm to run */
    int targetMemorySize,          /* I: Target memory usage for image IO */
    int max_threads,               /* I: Maximum number of threads to use */
    int min_elev,                  /* I: Minimum elevation */
    int max_elev,                  /* I: Maximum elevation */
    const int is_band_present[IAS_MAX_NBANDS],/* I: Band present in L1G flags*/
    int use_thermal_for_cfmask,    /* I: Thermal band control for CFmask */
    int use_cirrus_for_cfmask,     /* I: Cirrus band control for CFmask */
    double *cloudScore,            /* O: Cloud score */
    double *snowIceScore,          /* O: Snow/ice score */
    double *cloudCoverLandScore,   /* O: Land cloud score */
    int band_has_saturation[IAS_MAX_NBANDS] /* O: Band Saturation counts */
)
{
    const int kScaIndex    = 0;
    const int kStartSample = 0;

    /* The following set of arrays are indexed using the band type.  This is
       done throughout the application to simplify the access to sample values
       in specific bands for the cloud cover assessment algorithms. */
    IAS_L1G_BAND_METADATA metadataIn[IAS_NUM_BAND_TYPES];
    L1G_BAND_IO *         fdBandsIn[IAS_NUM_BAND_TYPES];
    float *               bufScaled[IAS_NUM_BAND_TYPES];
    int                   bandFlags[IAS_NUM_BAND_TYPES];
    int                   bandType2Number[IAS_NUM_BAND_TYPES];
    int band_type_to_index[IAS_NUM_BAND_TYPES];
    unsigned short *      bufImage[IAS_NUM_BAND_TYPES];
    IAS_L1G_BAND_METADATA band_metadata;
    L1G_BAND_IO *         fdBandTerrain               = NULL;
    L1G_BAND_IO *         fdBandOut                   = NULL;
    float                 percent;
    int                   bandType;
    int                   bufferFirstLine             = -1;
    int                   bufferLastLine              = -1;
    int                   i;
    int                   line;
    int                   nDigits;
    int                   nLines2Read                 = 0;
    int                   nLinesInImage;
    int                   nSamples;
    int                   nSamplesClassified          = 0;
    int                   nSamplesFilled              = 0;
    int                   nSamplesHighConfidenceCloud = 0;
    int nSamplesHighConfidenceSnowIce = 0;
    int nSamplesHighConfidenceCloudLand = 0;
    int nSamplesTotalLand = 0;
    int                   sample;
    int                   status;
    int                   statusLines;
    int                   totalSamples                = 0;
    int                   totalTerrainSamples         = 0;
    unsigned short *      bufTerrain                  = NULL;
    short *               bufDem                      = NULL;
    unsigned char *landWaterMask                      = NULL;
    double *cirrus_lookup_table                       = NULL;
    IAS_PROJECTION l1g_projection;
    IAS_IMAGE dem;
    int can_run = 0;
    int can_run_cirrus = 0;
    SATURATION_MASK *sat_mask = NULL;  /* Tracks the saturated pixels */

    /* ClassificationCounts contains a counter that indicates the number of
       samples that fall into each of the classifications */
    int classificationCounts[NUM_CLASSES];

    /* The buffer that contains the masks for each of the algorithms is indexed
       by the enumeration type CC_ALGORITHM.  After each cloud cover algorithm
       has generated a mask for every sample in a line, a resolving algorithm is
       run to combine the individual masks into a single output mask. */
    unsigned short * bufMasks[NUM_BUFFERS];

    /* Buffer for cfmask band */
    unsigned char *cfmask_buf = NULL; 
    
    /* Buffer for cfmask cloud confidence level */
    unsigned char *conf_mask = NULL;

    /* Initialize the scene's cloud score, snow/ice score and land cloud
       cover score */
    *cloudScore = 0.0;
    *snowIceScore = 0.0;
    *cloudCoverLandScore = 0.0;

    /* Initialization of arrays */
    for (bandType = 0; bandType < IAS_NUM_BAND_TYPES; bandType++)
    {
        int band_number;
        fdBandsIn[bandType] = NULL;
        bufScaled[bandType] = NULL;
        bufImage[bandType] = NULL;

        band_number = ias_sat_attr_get_band_number_from_type(bandType);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Retrieving band number for band type %d", bandType);
            return ERROR;
        }
        bandType2Number[bandType] = band_number;

        band_type_to_index[bandType] = 
            ias_sat_attr_convert_band_number_to_index(band_number);
        if (band_type_to_index[bandType] == ERROR)
        {
            IAS_LOG_ERROR("Retrieving band index for band number %d",
                band_number);
            return ERROR;
        }
    }
    for (i = 0; i < NUM_BUFFERS; i++)
        bufMasks[i] = NULL;
    for (i = 0; i < NUM_CLASSES; i++)
        classificationCounts[i] = 0;

    /* Inialize the L1G projection */
    if (set_projection(file_metadata, &l1g_projection) != SUCCESS)
    {
        IAS_LOG_ERROR("Projection creation failed");
        return ERROR;
    }

    /* Determine whether the requested algorithm can be run on this image */
    check_algorithm(fdImageIn, bandType2Number, pCtrl,
        file_metadata->sun_elevation, algorithm, &use_thermal_for_cfmask,
        &use_cirrus_for_cfmask, &can_run, &can_run_cirrus, bandFlags);

    /* Verify the size of all input bands match */
    if (verify_band_sizes(fdImageIn, fdTerrainIn, bandType2Number, bandFlags,
            &nLinesInImage, &nSamples) != SUCCESS)
    {
        IAS_LOG_ERROR("Verifying input bands");
        return ERROR;
    }

    /* If including saturated pixels in the quality band, initialize the
       tracking structure */
    if (get_enable_saturated_pixels_flag())
    {
        /* Initialize the L1G saturation tracking */
        sat_mask = initialize_saturation_mask(nLinesInImage, nSamples);
        if (!sat_mask)
        {
            IAS_LOG_ERROR("Initializing the L1G saturation structure");
            return ERROR;
        }

        /* Check for saturation in bands that the cloud cover algorithms will
           not be loading.  To reduce IO, the rest of the bands are checked for
           saturation as they are read for cloud cover algorithms. */
        for (bandType = 0; bandType < IAS_NUM_BAND_TYPES; bandType++)
        {
            int band_index = band_type_to_index[bandType];

            if (!bandFlags[bandType] && want_l1g_saturation_for_band(sat_mask,
                    band_index))
            {
                /* The band will not be read for the cloud cover and it
                   needs to be checked for saturation, so do it now */
                int band_number = bandType2Number[bandType];

                if (read_and_check_l1g_band_saturation(sat_mask, fdImageIn,
                        band_number, band_index) != SUCCESS)
                {
                    IAS_LOG_ERROR("Checking saturation in band number %d",
                        band_number);
                    cleanup_saturation_mask(sat_mask);
                    return ERROR;
                }
            }
        }
    }

    /* If the cfmask algorithm needs to be run, run it */
    if (can_run && algorithm == ALGORITHM_CFMASK)
    {
        int save_cirrus_band_flag;

        /* Verify an angle coefficient filename is available */
        if (strcmp(angle_coeff_filename, "") == 0)
        {
            IAS_LOG_ERROR("CFmask required angle coefficient filename not "
                "provided");
            cleanup_saturation_mask(sat_mask);
            return ERROR;
        }

        /* Temporarily mask the Cirrus band from being read if it isn't being
           used for the cfmask algorithm */
        save_cirrus_band_flag = bandFlags[IAS_CIRRUS_BAND];
        if (!use_cirrus_for_cfmask)
            bandFlags[IAS_CIRRUS_BAND] = FALSE;

        /* Open all of the necessary input bands for cfmask */
        if (open_input_bands(fdImageIn, bandType2Number, bandFlags,
                fdBandsIn, metadataIn) != SUCCESS)
        {
            IAS_LOG_ERROR("Opening bands");
            cleanup_saturation_mask(sat_mask);
            return ERROR;
        }

        /* Allocate the memory for cfmask buffer */
        cfmask_buf = malloc(nLinesInImage * nSamples * sizeof(*cfmask_buf));

        if (cfmask_buf == NULL)
        {
            IAS_LOG_ERROR("Error in allocating the memory for cfmask");
            close_bands(fdBandsIn, NULL, NULL);
            cleanup_saturation_mask(sat_mask);
            return ERROR;
        }

        /* Allocate the memory for cfmask cloud confidence level */
        conf_mask = malloc(nLinesInImage * nSamples * sizeof(*conf_mask));
        
        if (conf_mask == NULL)
        {
            IAS_LOG_ERROR("Error in allocating memory for confidence level");
            free(cfmask_buf);
            close_bands(fdBandsIn, NULL, NULL);
            cleanup_saturation_mask(sat_mask);
            return ERROR;
        }

        /* Run cfmask algorithm */
        if (cfmask_generate_cloud_mask(max_threads, nSamples, nLinesInImage,
            cfmaskCloudProbThreshold, file_metadata->sun_elevation,
            file_metadata->sun_azimuth, file_metadata, metadataIn, fdBandsIn,
            bandType2Number, pCtrl, angle_coeff_filename, cfmask_dilate_flag,
            use_thermal_for_cfmask, use_cirrus_for_cfmask, cfmask_buf,
            conf_mask, sat_mask) != SUCCESS)
        {
            IAS_LOG_ERROR("Error in running cfmask");
            free(cfmask_buf);
            free(conf_mask);
            close_bands(fdBandsIn, NULL, NULL);
            cleanup_saturation_mask(sat_mask);
            return ERROR;
        }

        /* Close the bands that were opened for cfmask */
        if (close_bands(fdBandsIn, NULL, NULL) != SUCCESS)
        {
            IAS_LOG_ERROR("Closing quality band");
            free(cfmask_buf);
            free(conf_mask);
            cleanup_saturation_mask(sat_mask);
            return ERROR;
        }

        /* Write the cfmask buffer to an envi image file if requested */
        if (strcmp(pathOutCfmask,"") != 0) 
        {
            IAS_L1G_BAND_METADATA *bmd = &metadataIn[IAS_BLUE_BAND];

            /* Write the cfmask image */
            if (ias_misc_write_envi_image(pathOutCfmask, &l1g_projection,
                    "Cfmask algorithm image", nLinesInImage, nSamples,
                    bmd->upper_left_x, bmd->upper_left_y,
                    bmd->projection_distance_x, bmd->projection_distance_y,
                    "Cfmask", IAS_BYTE, cfmask_buf) != SUCCESS)
            {
                IAS_LOG_ERROR("Writing cfmask band file: %s", pathOutCfmask);
                free(cfmask_buf);
                free(conf_mask);
                cleanup_saturation_mask(sat_mask);
                return ERROR;
            }
            IAS_LOG_INFO("Created cfmask band file: %s", pathOutCfmask);
        }

        /* Since CFmask has run now, clear all the bands from the list of bands
           to read since only the Cirrus band might be needed now */
        for (i = 0; i < IAS_NUM_BAND_TYPES; i++)
            bandFlags[i] = FALSE;

        /* Restore the Cirrus band flag */
        bandFlags[IAS_CIRRUS_BAND] = save_cirrus_band_flag;
    }

    /* Open all of the necessary bands for cloud cover asessment */
    if (open_bands(fdImageIn, fdTerrainIn, fdQualityOut, bandType2Number,
            bandFlags, nLinesInImage, nSamples, fdBandsIn, metadataIn,
            &fdBandTerrain, &fdBandOut, &band_metadata) != SUCCESS)
    {
        IAS_LOG_ERROR("Opening bands");
        free(cfmask_buf);
        free(conf_mask);
        cleanup_saturation_mask(sat_mask);
        return ERROR;
    }

    totalSamples = nLinesInImage * nSamples;

    if (IAS_LOG_DEBUG_ENABLED())
        print_ctrl_parameters(pCtrl);

    /* Allocate buffers to hold the image data */
    if (alloc_memory(nLinesInImage, nSamples, fdBandsIn, fdBandTerrain,
        targetMemorySize, bufImage, bufScaled, bufMasks, &bufTerrain,
        &nLines2Read) != SUCCESS)
    {
        IAS_LOG_ERROR("Allocating memory");
        close_bands(fdBandsIn, &fdBandTerrain, &fdBandOut);
        free(cfmask_buf);
        free(conf_mask);
        cleanup_saturation_mask(sat_mask);
        return ERROR;
    }

    /* Read the DEM and create the saturation pixel and/or cirrus lookup
       table */
    if (get_enable_saturated_pixels_flag() || can_run_cirrus)
    {
        /* Setup the image corners to be verified */
        IAS_CORNERS image_corners =
        {
            {band_metadata.upper_left_x, band_metadata.upper_left_y},
            {band_metadata.upper_right_x, band_metadata.upper_right_y},
            {band_metadata.lower_left_x, band_metadata.lower_left_y},
            {band_metadata.lower_right_x, band_metadata.lower_right_y}
        };

        /* Read the DEM */
        if (ias_misc_read_coregistered_single_band_l1g(pathDem, &image_corners,
            nLinesInImage, nSamples, band_metadata.projection_distance_y,
            &dem, IAS_I2) != SUCCESS)
        {
            IAS_LOG_ERROR("Opening DEM image '%s'", pathDem);
            free_memory(bufImage, bufScaled, bufMasks, &bufTerrain);
            close_bands(fdBandsIn, &fdBandTerrain, &fdBandOut);
            free(cfmask_buf);
            free(conf_mask);
            cleanup_saturation_mask(sat_mask);
            return ERROR;
        }
        bufDem = (short *) dem.data;

        /* Populate the cirrus lookup table */
        if (can_run_cirrus)
        {
            /* Create the threshold lookup table */
            if (create_cirrus_threshold_lookup_table(pCtrl->cirrusThreshold,
                min_elev, max_elev, &cirrus_lookup_table) != SUCCESS)
            {
                IAS_LOG_ERROR("Creating cirrus threshold lookup table");
                free_memory(bufImage, bufScaled, bufMasks, &bufTerrain);
                close_bands(fdBandsIn, &fdBandTerrain, &fdBandOut);
                free(cfmask_buf);
                free(conf_mask);
                free(bufDem);
                cleanup_saturation_mask(sat_mask);
                return ERROR;
            }
        }

        /* Only create the saturated pixel mask if requested */
        if (get_enable_saturated_pixels_flag())
        {
            IAS_PIXEL_MASK_IO *pixel_mask_file = NULL;
            IAS_LOS_MODEL *model_ptr = NULL;

            /* Open the pixel mask */
            pixel_mask_file = ias_pm_open_pixel_mask(get_pixel_mask_filename(),
                IAS_READ);
            if (pixel_mask_file == NULL)
            {
                free_memory(bufImage, bufScaled, bufMasks, &bufTerrain);
                close_bands(fdBandsIn, &fdBandTerrain, &fdBandOut);
                free(cirrus_lookup_table);
                free(bufDem);
                free(cfmask_buf);
                free(conf_mask);
            }

            /* Read the model */
            model_ptr = ias_model_read(get_model_filename());
            if (!model_ptr)
            {
                IAS_LOG_ERROR("Reading LOS model from %s",
                    get_model_filename());
                free_memory(bufImage, bufScaled, bufMasks, &bufTerrain);
                close_bands(fdBandsIn, &fdBandTerrain, &fdBandOut);
                free(cirrus_lookup_table);
                free(bufDem);
                free(cfmask_buf);
                free(conf_mask);
                ias_pm_close_pixel_mask(pixel_mask_file);
                cleanup_saturation_mask(sat_mask);
                return ERROR;
            }

            /* Process the pixel mask for saturation */
            if (create_l1r_saturation_mask(sat_mask, pixel_mask_file, model_ptr,
                &l1g_projection, &image_corners, &band_metadata, min_elev,
                max_elev, &dem, get_output_saturation_mask_flag(),
                is_band_present, band_has_saturation) != SUCCESS)
            {
                IAS_LOG_ERROR("Unable to create the saturation mask");
                free_memory(bufImage, bufScaled, bufMasks, &bufTerrain);
                close_bands(fdBandsIn, &fdBandTerrain, &fdBandOut);
                free(cfmask_buf);
                free(conf_mask);
                free(cirrus_lookup_table);
                free(bufDem);
                ias_pm_close_pixel_mask(pixel_mask_file);
                ias_los_model_free(model_ptr);
                cleanup_saturation_mask(sat_mask);
                return ERROR;
            }

            /* Close the pixel mask and model */
            ias_pm_close_pixel_mask(pixel_mask_file);
            ias_los_model_free(model_ptr);
        }
    }

    /* To avoid repeatedly setting masks for an algorithm that is not being run,
       initialize the array of mask values */
    init_algorithm_masks(nSamples, nLines2Read, bufMasks);

    /* How many digits are needed for the status message's line numbers?
       Display roughly 10 status messages per image. */
    nDigits = log10(nLinesInImage) + 1;
    statusLines = nLinesInImage / 10;

    /* Generate land water mask if it will be needed */
    if (can_run && strcmp(pathWaterMaskPolygon, "") != 0)
    {
        /* Allocate the memory for landWaterMask buffer */ 
        landWaterMask = calloc((nLinesInImage * nSamples),
            sizeof(*landWaterMask));
        if (landWaterMask == NULL)
        {
            IAS_LOG_ERROR("Allocating memory for land cloud cover mask");
            free(cfmask_buf);
            free(conf_mask);
            free_memory(bufImage, bufScaled, bufMasks, &bufTerrain);
            close_bands(fdBandsIn, &fdBandTerrain, &fdBandOut);
            free(cirrus_lookup_table);
            free(bufDem);
            cleanup_saturation_mask(sat_mask);
            return ERROR;
        }

        if (generate_land_water_mask(file_metadata, &band_metadata,  
            pathWaterMaskPolygon, pathOutLandWaterMaskImage, 
            nLinesInImage, nSamples, landWaterMask) != SUCCESS)
        {
            IAS_LOG_ERROR("Generating land water mask");
            free(landWaterMask);
            free(cfmask_buf);
            free(conf_mask);
            free_memory(bufImage, bufScaled, bufMasks, &bufTerrain);
            close_bands(fdBandsIn, &fdBandTerrain, &fdBandOut);
            free(cirrus_lookup_table);
            free(bufDem);
            cleanup_saturation_mask(sat_mask);
            return ERROR;
        }
    }

    /* Process every line of the image */
    for (line = 0; line < nLinesInImage; line++)
    {
        float *          pLineScaled[IAS_NUM_BAND_TYPES];
        int              lineIndex;
        unsigned short * pLineImage[IAS_NUM_BAND_TYPES];
        unsigned short * pLineMasks[NUM_BUFFERS];
        unsigned short * pLineTerrain;
        const unsigned char *l1r_sat_mask = NULL;
        const unsigned char *l1g_sat_mask = NULL;
        short * pLineDem;
        unsigned char *land_mask_line = NULL;
                            /* Pointer to current line's land/water mask */

        /* Display a processing message */
        if ((line % statusLines) == 0)
        {
            percent = (float)line / (float)nLinesInImage * 100.0;
            IAS_LOG_INFO("Processing line %*d/%*d (%1.0lf%%)", nDigits, line,
                nDigits, nLinesInImage, percent);
        }

        /* Is this line in the block of lines loaded in memory?  If not, read
           another block of lines from the input images. */
        if ((line < bufferFirstLine) || (line > bufferLastLine))
        {
            /* If we are at the end of the image, adjust the number of lines
               that should be read. */
            if ((line + nLines2Read) >= nLinesInImage)
                nLines2Read = nLinesInImage - line;

            /* Read a block of lines from the input L1G/L1T image for all of the
               bands that we are processing */
            for (bandType = 0; bandType < IAS_NUM_BAND_TYPES; bandType++)
            {
                if (fdBandsIn[bandType] == NULL)
                    continue;
                if (ias_l1g_read_image(fdBandsIn[bandType], kScaIndex, line,
                    kStartSample, nLines2Read, nSamples, bufImage[bandType])
                    != SUCCESS)
                {
                    IAS_LOG_ERROR(
                        "Reading %d lines starting at line %d from band %d",
                        nLines2Read, line, bandType2Number[bandType]);
                    free(landWaterMask);
                    free(cfmask_buf);
                    free(conf_mask);
                    free(cirrus_lookup_table);
                    free_memory(bufImage, bufScaled, bufMasks, &bufTerrain);
                    close_bands(fdBandsIn, &fdBandTerrain, &fdBandOut);
                    free(bufDem);
                    cleanup_saturation_mask(sat_mask);
                    return ERROR;
                }
            }

            /* Read a block of lines from the terrain occlusion mask, if
               specified */
            if (fdBandTerrain != NULL)
            {
                if (ias_l1g_read_image(fdBandTerrain, kScaIndex, line,
                    kStartSample, nLines2Read, nSamples, bufTerrain) !=
                    SUCCESS)
                {

                    IAS_LOG_ERROR("Reading %d lines starting at line %d from "
                        "terrain mask", nLines2Read, line);
                    free(landWaterMask);
                    free(cfmask_buf);
                    free(conf_mask);
                    free(cirrus_lookup_table);
                    free_memory(bufImage, bufScaled, bufMasks, &bufTerrain);
                    close_bands(fdBandsIn, &fdBandTerrain, &fdBandOut);
                    free(bufDem);
                    cleanup_saturation_mask(sat_mask);
                    return ERROR;
                }
            }
            bufferFirstLine = line;
            bufferLastLine = line + nLines2Read - 1;
        }

        /* What is this line's index within the buffers? */
        lineIndex = line - bufferFirstLine;

        /*----------------------------------------------------------------------
         * To help keep the cloud cover algorithms simple, we use a temporary
         * array of pointers that provides access to a single line of the input
         * image in each band.  This hides the fact that our buffer of image
         * values and scaled values are actually buffers that contain blocks of
         * lines to reduce the amount of I/O that this application performs.
         * 
         * This keeps the rest of the code as simple as possible and keeps most
         * of the "blocks of lines" logic here.
         *
         * buffer[band_0] --> ppppppp line x        /------ line[band_0]
         *                    ppppppp line x+1 <---/  /---- line[band_1]
         *                    ppppppp line x+2       /         ...
         *                                          /   /-- line[band_N]
         * buffer[band_1] --> ppppppp line x       /   /
         *                    ppppppp line x+1 <--/   /
         *                    ppppppp line x+2       /
         *     ...                                  /
         * buffer[band_N] --> ppppppp line x       /
         *                    ppppppp line x+1 <--/
         *                    ppppppp line x+2
         *--------------------------------------------------------------------*/
        for (bandType = 0; bandType < IAS_NUM_BAND_TYPES; bandType++)
        {
            int band_index = band_type_to_index[bandType];

            if (fdBandsIn[bandType] == NULL)
            {
                pLineImage[bandType]  = NULL;
                pLineScaled[bandType] = NULL;
                continue;
            }

            pLineImage[bandType]  = bufImage[bandType] + lineIndex * nSamples;

            /* Identify the saturated pixels for this band if still needed */
            if (get_enable_saturated_pixels_flag())
            {
                identify_saturated_l1g_pixels(sat_mask, pLineImage[bandType],
                    line, 1, nSamples, bandType2Number[bandType], band_index);
            }

            /* If we're working with an image that doesn't have any cloud cover
               algorithms to be run, don't scale the input image into
               reflectance or radiance values.  These values are only used by
               the cloud cover algorithms. */
            if (!can_run && !can_run_cirrus)
            {
                pLineScaled[bandType] = NULL;
                continue;
            }

            /* While we are looping through the bands, scale the L1G DN values
               for each sample into Top Of Atmosphere reflectace for OLI bands
               and Top of Atmosphere radiance for TIRS. */
            pLineScaled[bandType] = bufScaled[bandType] + lineIndex * nSamples;
            if (bandType == IAS_THERMAL1_BAND)
            {
                scale_samples(pLineImage[bandType], nSamples,
                    metadataIn[bandType].spectral_radiance_scaling_gain,
                    metadataIn[bandType].spectral_radiance_scaling_offset,
                    pCtrl->bandScaling[bandType], pLineScaled[bandType]);
            }
            else 
            {
                scale_samples(pLineImage[bandType], nSamples,
                    metadataIn[bandType].reflectance_scaling_gain,
                    metadataIn[bandType].reflectance_scaling_offset,
                    pCtrl->bandScaling[bandType], pLineScaled[bandType]);
            }
        }
        if (bufTerrain == NULL)
            pLineTerrain = NULL;
        else
            pLineTerrain = bufTerrain + lineIndex * nSamples;

        /* Set the saturated pixel line and dem line */
        if (sat_mask)
        {
            l1r_sat_mask = get_l1r_saturation_mask(sat_mask, line);
            l1g_sat_mask = get_l1g_saturation_mask(sat_mask, line);
        }

        if (!bufDem)
            pLineDem = NULL;
        else
            pLineDem = bufDem + line * nSamples;

        for (i = 0; i < NUM_BUFFERS; i++)
            pLineMasks[i] = bufMasks[i] + lineIndex * nSamples;
        
        if (can_run_cirrus)
        {
            run_algorithm_cirrus(nSamples, pLineScaled[IAS_CIRRUS_BAND],
                pLineDem, cirrus_lookup_table, min_elev, max_elev,
                pLineMasks[CIRRUS_BUFFER_INDEX]);
        }

        /* Run the selected CCA algorithm on this line */
        if (can_run)
        {
            switch (algorithm)
            {
                case ALGORITHM_ACCA:
                    run_algorithm_acca(nSamples, pLineScaled,
                        pLineMasks[CLOUD_BUFFER_INDEX]);
                    break;

                case ALGORITHM_ATACCA:
                    run_algorithm_atacca(nSamples, sineSolarElev, pLineScaled,
                        pLineMasks[CLOUD_BUFFER_INDEX]);
                    break;

                case ALGORITHM_SEE5:
                    run_algorithm_see5(nSamples, sineSolarElev, pLineScaled,
                        pLineMasks[CLOUD_BUFFER_INDEX]);
                    break;

                case ALGORITHM_CFMASK:
                    run_algorithm_cfmask(nSamples, line, cfmask_buf, conf_mask,
                        pLineMasks[CLOUD_BUFFER_INDEX]);
                    break;

                /* Nothing to do for Cirrus */
                default:
                    break;
            }
        }

        /* Get the correct line from the land/water mask if it is available */
        if (landWaterMask)
            land_mask_line = &landWaterMask[line * nSamples];

        /* Combine the individual masks into a final output mask and
           incorporate the terrain occlusion mask. */
        resolve_algorithm_masks(can_run, can_run_cirrus, nSamples, pLineImage,
            pLineTerrain, pLineMasks, land_mask_line, l1r_sat_mask,
            l1g_sat_mask, &nSamplesFilled, &nSamplesHighConfidenceCloud,
            &nSamplesHighConfidenceSnowIce, &nSamplesHighConfidenceCloudLand,
            &nSamplesTotalLand);
     
        if (IAS_LOG_DEBUG_ENABLED())
        {
            /* Keep track of the number of samples falling into each
               classification.  Note: by &'ing with the classification's high
               confidence mask we'll get a non-zero value for low, medium, and
               high confidence levels. */
            for (sample = 0; sample < nSamples; sample++)
            {
                unsigned short value = pLineMasks[RESULT_BUFFER_INDEX][sample];
                if (value & CC_CLOUD_HIGH)
                    classificationCounts[CLASS_CLOUD]++;
                if (value & CC_CIRRUS_HIGH)
                    classificationCounts[CLASS_CIRRUS]++;
                if (value & CC_SNOW_HIGH)
                    classificationCounts[CLASS_SNOW_ICE]++;
                if (value & CC_CLOUD_SHADOW_HIGH)
                    classificationCounts[CLASS_CLOUD_SHADOW]++;
                if (value & CC_MASK_TERRAIN)
                    totalTerrainSamples++;
            }
        }

        /* If we are processing the last line in the buffer, then write all of
           the lines to the output quality band.  Note that we're writing out
           our buffer that contains blocks of lines here.  We're not using our
           temporary array that points to a single line of data. */
        if (line == bufferLastLine)
        {
            status = ias_l1g_write_image(fdBandOut, kScaIndex, bufferFirstLine,
                kStartSample, nLines2Read, nSamples,
                bufMasks[RESULT_BUFFER_INDEX]);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Writing %d lines starting at line %d to quality "
                    "band image" , nLines2Read, line);
                free(landWaterMask);
                free(cfmask_buf);
                free(conf_mask);
                free_memory(bufImage, bufScaled, bufMasks, &bufTerrain);
                close_bands(fdBandsIn, &fdBandTerrain, &fdBandOut);
                free(cirrus_lookup_table);
                free(bufDem);
                cleanup_saturation_mask(sat_mask);
                return ERROR;
            }
        }
    } /* end of for loop of nLinesInImage */

    /* The lookup table will not be used again */
    free(cirrus_lookup_table);
    cirrus_lookup_table = NULL;

    /* Perform the final saturated pixel updates */
    if (get_enable_saturated_pixels_flag())
    {
        /* Update the flags for which bands have saturation */
        update_bands_with_l1g_saturation(sat_mask, band_has_saturation);

        /* Output the saturation mask if requested */
        if (get_output_saturation_mask_flag())
        {
            /* Write the saturation mask to file */
            if (ias_misc_write_envi_image("l1g_saturation_mask.img",
                &l1g_projection, "L1G Saturation Mask", nLinesInImage, nSamples,
                band_metadata.upper_left_x, band_metadata.upper_left_y,
                band_metadata.projection_distance_x,
                band_metadata.projection_distance_y, "Saturation Mask",
                IAS_BYTE, get_l1g_saturation_mask(sat_mask, 0)) != SUCCESS)
            {
                IAS_LOG_ERROR("Creating the saturation mask image");
                free(landWaterMask);
                free(cfmask_buf);
                free(conf_mask);
                free_memory(bufImage, bufScaled, bufMasks, &bufTerrain);
                close_bands(fdBandsIn, &fdBandTerrain, &fdBandOut);
                free(bufDem);
                cleanup_saturation_mask(sat_mask);
                return ERROR;
            }
        }

        /* Clean up the L1G saturation tracking structure */
        cleanup_saturation_mask(sat_mask);
        sat_mask = NULL;
    }

    /* Ensure the processing message ends with 100% */
    IAS_LOG_INFO("Processing line %*d/%*d (100%%)", nDigits, line, nDigits,
        nLinesInImage);

    /* Calculate the cloud cover score for this image.  If none of the cloud
       cover algorithms were run on the image because we're processing a TIRS
       only image or because of missing bands, then the cloud score should be
       set to -1.0 to indicate a classification wasn't performed.  Some corner
       cases in short partial scenes can result in all the pixels being 
       classified as fill which leads to not calculating a cloud score. 

       The reason this application would still be run in these situations is to
       convert the terrain occlusion mask so that bit is moved from the least
       significant bit into the bit position used in the quality band. */
    nSamplesClassified = totalSamples - nSamplesFilled;
    if (!can_run || nSamplesClassified == 0)
    {
        *cloudScore = -1.0;
        *snowIceScore = -1.0;
        *cloudCoverLandScore = -1.0;
    }
    else
    {
        /* Compute the overall cloud score for this image.  By definition, this
           value is the number of samples that received a high confidence value
           for either the cloud or cirrus classifications divided by the total
           number of non-fill samples that have been classified. */
        *cloudScore = (double)nSamplesHighConfidenceCloud
            / (double)(nSamplesClassified) * 100.0;

        /* Round the score to the nearest 1/100th of a percent */
        *cloudScore = floor((*cloudScore * 100.0) + 0.5) / 100.0;

        /* Compute the overall snow/ice score for this image. By definition,
           this value is the number of samples that received a high confidence
           value for snow/ice classifications divided by the total number of
           non-fill samples that have been classified. */
        *snowIceScore = (double)nSamplesHighConfidenceSnowIce
            / (double)(nSamplesClassified) * 100.0;

        /* Round the score to the nearest 1/100th of a percent */
        *snowIceScore = floor((*snowIceScore * 100.0) + 0.5) / 100.0;

        /* Compute the overall cloud cover land  score for this image.  
           By definition, this value is the number of samples that received a 
           high confidence value for either the cloud or cirrus classifications
           divided by the total number of non-fill land (not water) samples 
           that have been classified. */
        if (nSamplesTotalLand == 0)
        {
            *cloudCoverLandScore = -1.0;
        }
        else
        {        
            *cloudCoverLandScore = (double)nSamplesHighConfidenceCloudLand
                / (double)(nSamplesTotalLand) * 100.0;

            /* Round the score to the nearest 1/100th of a percent */
            *cloudCoverLandScore = floor((*cloudCoverLandScore * 100.0) + 0.5) 
                / 100.0;
        }
    }

    if (IAS_LOG_DEBUG_ENABLED())
    {
        printf("=======================================================\n");
        printf("= Summary                                             =\n");
        printf("=======================================================\n");
        printf("    %10d nLines\n", nLinesInImage);
        printf("    %10d nSamples\n", nSamples);
        printf("    %10d total samples\n", totalSamples);
        printf("\n");
        printf("    %10d Cloud\n",    classificationCounts[CLASS_CLOUD]);
        printf("    %10d Cirrus\n",   classificationCounts[CLASS_CIRRUS]);
        printf("    %10d Snow/Ice\n", classificationCounts[CLASS_SNOW_ICE]);
        printf("    %10d Cloud Shadow\n", 
           classificationCounts[CLASS_CLOUD_SHADOW]);
        printf("    %10d Terrain occluded\n", totalTerrainSamples);
        printf("    %10d fill samples (%1.0f%%)\n", nSamplesFilled,
            ((double)nSamplesFilled / (double)totalSamples) * 100.0);
        printf("    %10d classified samples (%1.0f%%)\n", nSamplesClassified,
            ((double)nSamplesClassified / (double)totalSamples) * 100.0);
        printf("    %10d Total Land samples (%1.0f%%)\n", nSamplesTotalLand,
            ((double)nSamplesTotalLand / (double)totalSamples) * 100.0);
        printf("\n");
        printf("    %10d High Confidence Cloud Samples\n",
            nSamplesHighConfidenceCloud);
        printf("   %10.2lf%% Cloud Score\n", *cloudScore);
        printf("   %10.2lf%% Snow/ice Score\n", *snowIceScore);
        printf("   %10.2lf%% Cloud Cover Land Score\n", *cloudCoverLandScore);
        printf("\n");
    }

    /* Close the bands and free the dynamically allocated memory */
    free(bufDem);
    bufDem = NULL;
    free(cfmask_buf);
    cfmask_buf = NULL;
    free(conf_mask);
    conf_mask = NULL;
    free(landWaterMask);
    landWaterMask = NULL;
    free_memory(bufImage, bufScaled, bufMasks, &bufTerrain);

    if (close_bands(fdBandsIn, &fdBandTerrain, &fdBandOut) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing input bands");
        return ERROR;
    }

    return SUCCESS;
}

/*------------------------------------------------------------------------------
NAME: process_image

PURPOSE:
    This function will process a single L1Gt or L1T image and perform Cloud
    Cover Assessment on the image.

RETURNS:
    The function will return a status of SUCCESS if the image is successfully
    processed, ERROR if an error is encountered.
------------------------------------------------------------------------------*/
static int process_image
(
    const char pathCpf[],          /* I: Path of the CPF */
    const char pathIn[],           /* I: Path of input L1G or L1T image */
    const char pathOmf[],          /* I: Path of the OMF */
    char pathOut[],                /* I: Path of the output quality band */
    const char pathWaterMaskPolygon[], /* I: Path of water mask polygon file */
    const char pathOutLandWaterMaskImage[],/* I: Path of land water mask file */
    const char pathTerrain[],      /* I: Path of the terrain occlusion mask */
    const char pathDem[],          /* I: Path to DEM file */
    const char pathOutCfmask[],    /* I: Path of the cfmask band */
    const char angle_coeff_filename[],/* I: Angle coefficient filename */
    float cfmaskCloudProbThreshold,/* I: Cloud Probability Threshold for
                                         cfmask */
    int cfmask_dilate_flag,        /* I: Flag to turn on/off cloud dilation */
    CC_ALGORITHM algorithm,        /* I: Algorithm to run */
    int targetMemorySize,          /* I: Target memory usage for image IO */
    int max_threads,               /* I: Maximum number of threads to use */
    int min_elev,                  /* I: Minimum elevation */
    int max_elev,                  /* I: Maximum elevation */
    int use_thermal_for_cfmask,    /* I: Thermal band control for CFmask */
    int use_cirrus_for_cfmask      /* I: Cirrus band control for CFmask */
)
{
    CLOUD_COVER_CTRL      cloudCoverCtrl;
    IAS_L1G_FILE_METADATA metadata;
    L1GIO *               fdImageIn;
    L1GIO *               fdQualityOut;
    L1GIO *               fdTerrainIn              = NULL;
    double                cloudScore;
    double snowIceScore;
    double cloudCoverLandScore;
    double                sineSolarElev;
    int                   nBands;
    int                   sec;
    time_t                timeStart                = time(NULL);
    int band_has_saturation[IAS_MAX_NBANDS] = {0};
    int is_band_present[IAS_MAX_NBANDS] = {0};
    int index;

    /* Ensure this is an L1G image */
    if (!ias_l1g_is_l1g_file(pathIn))
    {
        IAS_LOG_ERROR("Not an L1G image: %s", pathIn);
        return ERROR;
    }

    /* Open the image for reading */
    fdImageIn = ias_l1g_open_image(pathIn, IAS_READ);
    if (fdImageIn == NULL)
    {
        IAS_LOG_ERROR("Opening image %s", pathIn);
        return ERROR;
    }

    /* Retrieve the image's metadata */
    if (ias_l1g_get_file_metadata(fdImageIn, &metadata) != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving L1G file metadata %s", pathIn);
        ias_l1g_close_image(fdImageIn);
        return ERROR;
    }

    /* Ensure the solar angles in the file metadata are valid.  These are
       imperative to have for cloud cover assessment */
    if (!metadata.sun_angles_valid)
    {
        IAS_LOG_ERROR("Sun elevation angle not present in L1G image metadata");
        ias_l1g_close_image(fdImageIn);
        return ERROR;
    }

    /* Compute the sine of the solar elevation angle.  This is a common factor
       used in radiance-reflectance rescaling as well as the algorithms to
       detect cloud cover.  The math library function sin() requires the angle
       to be in radians, not degrees.  The conversion from degrees to radians is
       performed by multiplying: degress * pi/180.  Use the IAS math library to
       compute the conversion factor. */
    sineSolarElev =
        sin(metadata.sun_elevation * ias_math_get_radians_per_degree());

    /* Initialize the cloud cover control parameters */
    if (get_ctrl_parameters_cpf(pathCpf, sineSolarElev, &cloudCoverCtrl)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving control parameters");
        ias_l1g_close_image(fdImageIn);
        return ERROR;
    }

    /* Retrieve the band count */
    nBands = ias_l1g_get_band_count(fdImageIn);

    /* Determine which bands are present */
    for (index = 0; index < IAS_MAX_NBANDS; index++)
    {
        int band_number;

        /* Convert the band index to number */
        band_number = ias_sat_attr_convert_band_index_to_number(index);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Converting band index %d to band number", index);
            ias_l1g_close_image(fdImageIn);
            return ERROR;
        }

        /* Check if the band number is present in the L1G */
        is_band_present[index] = ias_l1g_is_band_present(fdImageIn,
            band_number);
    }

    /* If a terrain occlusion mask has been computed, open that image so the
       terrain occlusion mask can be combined into a final quality mask along
       with the classification mask */
    if (strlen(pathTerrain) > 0)
    {
        int terrainBands;

        fdTerrainIn = ias_l1g_open_image(pathTerrain, IAS_READ);
        if (fdTerrainIn == NULL)
        {
            IAS_LOG_ERROR("Opening image %s", pathTerrain);
            ias_l1g_close_image(fdImageIn);
            return ERROR;
        }

        terrainBands = ias_l1g_get_band_count(fdTerrainIn);
        if (terrainBands != 1)
        {
            IAS_LOG_ERROR(
                "Expecting 1 band in terrain occlusion mask, found %d",
                terrainBands);
            ias_l1g_close_image(fdImageIn);
            ias_l1g_close_image(fdTerrainIn);
            return ERROR;
        }
    }

    if (IAS_LOG_DEBUG_ENABLED())
    {
        printf("=======================================================\n");
        printf("= Image Information                                   =\n");
        printf("=======================================================\n");
        printf("     input image = %s\n",      pathIn);
        printf("      spacecraft = %s\n",      metadata.spacecraft);
        printf("        path/row = %d/%d\n",   metadata.wrs_path,
            metadata.wrs_row);
        printf("      roll angle = %10.8lf\n", metadata.roll_angle);
        printf(" collection type = %s\n",      metadata.collection_type);
        printf("    capture date = %s\n",      metadata.capture_date);
        printf("    capture time = %s\n",      metadata.capture_time);
        printf(" solar elevation = %10.8lf\n", metadata.sun_elevation);
        printf(" sine solar elev = %10.8lf\n", sineSolarElev);
        printf("          nbands = %d\n",      nBands);
        if (fdTerrainIn != NULL)
            printf("    terrain mask = %s\n",  pathTerrain);
        printf("    output image = %s\n",      pathOut);
        printf("         pathCpf = %s\n", pathCpf);
        printf("         pathOmf = %s\n", pathOmf);
        printf("          pathIn = %s\n", pathIn);
        printf("         pathOut = %s\n", pathOut); 
        printf("   pathOutCfmask = %s\n", pathOutCfmask);
        printf("   solar azimuth = %10.8lf\n", metadata.sun_azimuth);
        printf("solar angle valid = %d\n", metadata.sun_angles_valid);
        printf("Cfmask Cloud Probability Threshold = %f\n",
            cfmaskCloudProbThreshold);
        printf("Pixel mask filename = %s\n", get_pixel_mask_filename());
        printf("Model filename = %s\n", get_model_filename());
    }

    /* Open an output L1G image for the quality band */
    fdQualityOut = ias_l1g_open_image(pathOut, IAS_WRITE);
    if (fdQualityOut == NULL)
    {
        IAS_LOG_ERROR("Opening image for writing %s", pathOut);
        ias_l1g_close_image(fdImageIn);
        if (fdTerrainIn != NULL)
            ias_l1g_close_image(fdTerrainIn);
        return ERROR;
    }

    /* Copy the input image's metadata to the output image */
    if (ias_l1g_set_file_metadata(fdQualityOut, &metadata) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing quality band image's metadata %s", pathOut);
        ias_l1g_close_image(fdImageIn);
        ias_l1g_close_image(fdQualityOut);
        if (fdTerrainIn != NULL)
            ias_l1g_close_image(fdTerrainIn);
        return ERROR;
    }

    /* Process the bands */
    if (process_bands(pathOmf, pathOut, pathWaterMaskPolygon,
        pathOutLandWaterMaskImage, pathOutCfmask, pathDem, angle_coeff_filename,
        cfmaskCloudProbThreshold, cfmask_dilate_flag, fdImageIn, fdTerrainIn,
        fdQualityOut, &cloudCoverCtrl, &metadata, sineSolarElev, algorithm,
        targetMemorySize, max_threads, min_elev, max_elev, is_band_present,
        use_thermal_for_cfmask, use_cirrus_for_cfmask, &cloudScore,
        &snowIceScore, &cloudCoverLandScore, band_has_saturation) != SUCCESS)
    {
        IAS_LOG_ERROR("Processing bands");
        ias_l1g_close_image(fdImageIn);
        ias_l1g_close_image(fdQualityOut);
        if (fdTerrainIn != NULL)
            ias_l1g_close_image(fdTerrainIn);
        return ERROR;
    }

    /* Close the image after processing is complete */
    ias_l1g_close_image(fdImageIn);
    if (fdTerrainIn != NULL)
        ias_l1g_close_image(fdTerrainIn);
    if (ias_l1g_close_image(fdQualityOut) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing quality band image: %s", pathOut);
        return ERROR;
    }

    /* Update the OMF file with the necessary parameters that describe which
       algorithms were run and the overall cloud score of the image */
    if (update_omf(pathOmf, pathOut, cloudScore, snowIceScore,
            cloudCoverLandScore, band_has_saturation) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing parameters to the OMF");
        return ERROR;
    }

    /* Compute the processing time */
    sec = (unsigned int)(time(NULL) - timeStart);

    IAS_LOG_INFO("Cloud Score %s: %4.2lf%%", pathIn, cloudScore);
    IAS_LOG_INFO("Snow/Ice Score %s: %4.2lf%%", pathIn, snowIceScore);
    IAS_LOG_INFO("Land Cloud Cover Score %s: %4.2lf%%", pathIn, 
        cloudCoverLandScore);
    IAS_LOG_INFO("Created quality band image: %s", pathOut);
    IAS_LOG_INFO("assess_cloud_cover took %d seconds", sec);
    return SUCCESS;
}

int main (int argc, char * argv[])
{
    IAS_SATELLITE_ID satId;
    char pathCpf[PATH_MAX] = "";
    char pathIn[PATH_MAX] = "";
    char pathOmf[PATH_MAX] = "";
    char pathDem[PATH_MAX] = "";
    char pathOut[PATH_MAX] = "";
    char pathWaterMaskPolygon[PATH_MAX]      = "";
    char pathOutLandWaterMaskImage[PATH_MAX] = "";
    char pathTerrain[PATH_MAX] = "";
    char pathOutCfmask[PATH_MAX] = "";
    char angle_coeff_filename[PATH_MAX] = "";
    double cfmaskCloudProbThreshold = 22.5;
    int cfmask_dilate_flag = 0;
    int satNum;
    int min_elev = 0;
    int max_elev = 0;
    int targetMemorySize = 0;
    int max_threads = 1; /* Maximum number of threads to use */
    CC_ALGORITHM algorithm;
    int use_thermal_for_cfmask = 0; /* Indicates when to use the thermal band
                                       for the CFmask algorithm */
    int use_cirrus_for_cfmask = 0;  /* Indicates when to use the cirrus band
                                       for the CFmask algorithm */


    /*--------------------------------------------------------------------------
     * This algorithm was originally developed to read a single line at a time,
     * run the cloud cover algorithms on that line, resolve the masks into an
     * output image line, and then write that line to the output file.  This ran
     * sufficiently fast on the development servers but when run on the
     * production LPGS servers the application was I/O bound.
     * 
     * To help alleviate this problem, we now are computing a number of image
     * lines that can be stored in a target amount of memory and then we are
     * allocating chunks of memory so that blocks of lines can be read at a
     * time.
     *------------------------------------------------------------------------*/
    targetMemorySize = TARGET_MEMORY;

    /* Initialize logging */
    if (ias_log_initialize("assess_cloud_cover"))
    {
        IAS_LOG_ERROR("Initializing logging capabilities");
        exit(EXIT_FAILURE);
    }

    /* Process the command line arguments */
    if (argc != 2)
    {
        IAS_LOG_ERROR("usage: assess_cloud_cover <parameter file>");
        exit(EXIT_FAILURE);
    }
   
    if (get_parameters(argv[1], &satNum, pathCpf, pathIn, pathOmf, pathDem,
        pathOut, pathWaterMaskPolygon, pathOutLandWaterMaskImage,
        pathTerrain, pathOutCfmask, angle_coeff_filename,
        &cfmaskCloudProbThreshold, &cfmask_dilate_flag, &max_threads,
        &algorithm, &min_elev, &max_elev, &use_thermal_for_cfmask,
        &use_cirrus_for_cfmask) != SUCCESS)
    {
        IAS_LOG_ERROR("Getting parameters");
        exit(EXIT_FAILURE);
    }

    /* Using the satellite number read from the OMF file, initialize the
       satellite attributes library */
    satId = ias_sat_attr_get_satellite_id_from_satellite_number(satNum);
    if (satId == ERROR)
    {
        IAS_LOG_ERROR("Unsupported satellite number: %d", satNum);
        exit(EXIT_FAILURE);
    }
    if (ias_sat_attr_initialize(satId) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attribute library");
        exit(EXIT_FAILURE);
    }

    /* Process the image specified */
    IAS_LOG_INFO("assess_cloud_cover %s", pathIn);
    if (process_image(pathCpf, pathIn, pathOmf, pathOut, 
        pathWaterMaskPolygon, pathOutLandWaterMaskImage, pathTerrain, pathDem,
        pathOutCfmask, angle_coeff_filename, cfmaskCloudProbThreshold,
        cfmask_dilate_flag, algorithm, targetMemorySize, max_threads,
        min_elev, max_elev, use_thermal_for_cfmask, use_cirrus_for_cfmask)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Processing image: %s", pathIn);
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}
