#ifndef ASSESS_CLOUD_COVER_H
#define ASSESS_CLOUD_COVER_H

#include "ias_cpf.h" /* For IAS_CPF_CIRRUS_THRESHOLD_SIZE */

#define NAME_STRLEN  256

/* The number and names of the algorithms that are used in cloud cover
   assessment */
typedef enum cc_algorithm
{
    ALGORITHM_ACCA=0,
    ALGORITHM_ATACCA,
    ALGORITHM_CIRRUS,
    ALGORITHM_SEE5,
    ALGORITHM_CFMASK,
    NUM_ALGORITHMS /* Must be last to represent the number of algorithms */
} CC_ALGORITHM;

/* This algorithm will create a set of masks for each line of the input image.
   A separate mask will be created for each algorithm run and the summary,
   which will be 3 - one for the cirrus algorithm and one for the primary 
   algorithm and one for the results. */
#define CIRRUS_BUFFER_INDEX 0
#define CLOUD_BUFFER_INDEX 1
#define RESULT_BUFFER_INDEX 2
#define NUM_BUFFERS 3

/* Define the possible values for the flags to use the thermal and cirrus
   bands for the CFmask algorithm */
#define DO_NOT_USE_BAND 0
#define USE_BAND 1
#define USE_IF_AVAILABLE 2
#define USE_IF_THERMAL_NOT_AVAILABLE 2

/* Classifications available in cloud cover assessment */
typedef enum cc_class
{
    CLASS_CLOUD=0,
    CLASS_CIRRUS,
    CLASS_SNOW_ICE,
    CLASS_VEG,
    CLASS_WATER,
    CLASS_CLOUD_SHADOW,
    NUM_CLASSES /* Must be last to represent the number of classifications */
} CC_CLASS;

/* This structure holds the cloud cover control parameters contained in the CPF.
   The Cirrus threshold is a value which, if exceeded by any sample in the
   Cirrus band, results in that sample begin flagged as cirrus clouds.

   The band scaling data member is used to store a scaling factor that
   will be applied to each of the bands used in the algorithm.  This provides a
   method to scale Landsat 8's psuedo radiance/reflectance into TOA
   radiance/reflectance.  This array is indexed by band type to match arrays
   throughout the application, which is done to simplify access to sample values
   in specific bands for the cloud cover algorithms. */
typedef struct cloud_cover_ctrl
{
    double         minimumSunElevationDegrees;
    double         cirrusThreshold[IAS_CPF_CIRRUS_THRESHOLD_SIZE];
    double         bandScaling[IAS_NUM_BAND_TYPES];
    double         thermalConstantsTIRS1[2];   /* Array to store the
                                                  thermal constants of
                                                  TIRS1 from CPF */
} CLOUD_COVER_CTRL;

/* Each sample is assigned a 16-bit mask indicating the status of that sample.
   Within the mask, this application will assign a 2-bit confidence value of
   none, low, medium, or high for each of the 5 classifications.  The bit
   positions of the mask are described below.  The representation of a 16-bit
   mask is shown on the right with a flag indicating where each class's
   confidence level is stored.
  
       Flag Description         Bits  Flag      RRRi issS SccC rrtf
       ----------------------   ----  ----      1111 1111 1111 1111
       Designated Fill            0    f
       Terrain Occlusion          1    t
       Radiometric Saturation   2-3    r
       Cloud                      4    C
       Cloud confidence         5-6    c
       Cloud Shadow             7-8    S
       Snow / Ice confidence   9-10    s
       Cirrus confidence      11-12    i
       <Reserved>             13-15    R

   The current implementation of this application is not setting the dropped
   frame and vegetation bits within this mask.
*/

/* Constants for well known, non-computed masks */
#define CC_MASK_EMPTY    0x0000
#define CC_MASK_FILL     0x0001
#define CC_MASK_TERRAIN  0x0002

/* Constant for Radiometric Saturation classification */
#define CC_RAD_SAT_LOW   0x0004
#define CC_RAD_SAT_MED   0x0008
#define CC_RAD_SAT_HIGH  0x000C

/* Constant for Cloud summary bit */
#define CC_CLOUD_SUMMARY 0x0010

/* Constant values for Cloud classification */
#define CC_CLOUD_LOW     0x0020
#define CC_CLOUD_MED     0x0040
#define CC_CLOUD_HIGH    0x0060

/* Constant values for Cloud Shadows classification */
#define CC_CLOUD_SHADOW_LOW    0x0080
#define CC_CLOUD_SHADOW_MED    0x0100
#define CC_CLOUD_SHADOW_HIGH   0x0180

/* Constant values for Snow & Ice classification */
#define CC_SNOW_LOW      0x0200
#define CC_SNOW_MED      0x0400
#define CC_SNOW_HIGH     0x0600

/* Constant values for Cirrus Cloud classification */
#define CC_CIRRUS_LOW    0x0800
#define CC_CIRRUS_MED    0x1000
#define CC_CIRRUS_HIGH   0x1800

/* Constant values for Water classification for ACCA and ATACCA.  Use the fill
   and terrain bits since the cloud cover algorithms do not set those. */
#define CC_WATER_LOW     0x0001
#define CC_WATER_MED     0x0002
#define CC_WATER_HIGH    0x0003

#endif
