/*****************************************************************************
Name: calculate_map_azimuth_angle

Purpose: Given the metadata for an L1G file, calculate the map azimuth angle.

Note: The angle returned is the angle from "straight-up" in the image and
    the direction of true north at the center of the image.  

Returns: SUCCESS/ERROR

*****************************************************************************/
#include <math.h>
#include "gctp.h"
#include "ias_geo.h"
#include "ias_logging.h"
#include "ias_angle_gen.h"
#include "calculate_map_azimuth_angle.h"

int calculate_map_azimuth_angle
(
    const IAS_L1G_FILE_METADATA *file_metadata, /* I: projection info */
    const IAS_L1G_BAND_METADATA *band_metadata, /* I: corner info */
    double *azimuth_angle_radians               /* O: azimuth angle */
)
{
    int index;                             /* Loop variable */
    int projection_units;                  /* Input units code */
    double geo_parameters[IAS_PROJ_PARAM_SIZE];/* Geo parameter array */
    IAS_PROJECTION image_projection;          /* Source projection */
    IAS_PROJECTION geo_projection;         /* Destination projection */
    IAS_GEO_PROJ_TRANSFORMATION *image2geo;
    IAS_GEO_PROJ_TRANSFORMATION *geo2image;
    double center_x;
    double center_y;
    double center_lat;
    double center_lng;
    double offset_x;
    double offset_y;
    double delta_x;
    double delta_y;

    /* Build the projection definition for the L1G file */
    if (ias_geo_get_units(file_metadata->projection_units, &projection_units)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Getting the projection units for %s",
            file_metadata->projection_units);
        return ERROR;
    }
    ias_geo_set_projection(file_metadata->projection_code,
        file_metadata->zone_code, projection_units,
        file_metadata->spheroid_code, file_metadata->projection_parameters,
        &image_projection);

    /* Build the projection definition for the geographic projection */
    for (index = 0; index < IAS_PROJ_PARAM_SIZE; index++)
    { 
        geo_parameters[index] = 0.0;
    }
    ias_geo_set_projection(GEO, NULLZONE, RADIAN, file_metadata->spheroid_code,
        geo_parameters, &geo_projection);

    /* Create projection transformations to go between the L1G projection
       and geographic */
    image2geo = ias_geo_create_proj_transformation(&image_projection,
        &geo_projection);
    if (!image2geo)
    {
        IAS_LOG_ERROR("Initializing the image to geo projection "
            "transformation");
        return ERROR;
    }
    geo2image = ias_geo_create_proj_transformation(&geo_projection,
        &image_projection);
    if (!image2geo)
    {
        ias_geo_destroy_proj_transformation(image2geo);
        IAS_LOG_ERROR("Initializing the geo to image projection "
            "transformation");
        return ERROR;
    }

    /* Estimate the scene center from the L1G corners */
    center_x = (band_metadata->upper_left_x + band_metadata->lower_right_x)
        / 2.0;
    center_y = (band_metadata->upper_left_y + band_metadata->lower_right_y)
        / 2.0;
    if (ias_geo_transform_coordinate(image2geo, center_x, center_y, &center_lng,
        &center_lat) != SUCCESS)
    {
        IAS_LOG_ERROR("Transforming scene center to lat/long");
        ias_geo_destroy_proj_transformation(image2geo);
        ias_geo_destroy_proj_transformation(geo2image);
        return ERROR;
    }

    /* Get the projection coordinates for a point slightly north of the center
       point */
    if (ias_geo_transform_coordinate(geo2image, center_lng,
            center_lat + 0.00005, &offset_x, &offset_y) != SUCCESS)
    {
        IAS_LOG_ERROR("Transforming offset location to projection "
            "coordinates");
        ias_geo_destroy_proj_transformation(image2geo);
        ias_geo_destroy_proj_transformation(geo2image);
        return ERROR;
    }

    /* Calculate the azimuth angle from the Y axis as the arctangent(dX/dY) */
    delta_y = offset_y - center_y;
    delta_x = offset_x - center_x;
    *azimuth_angle_radians = atan2(delta_x, delta_y);

    /* Free the projection transformations */
    ias_geo_destroy_proj_transformation(image2geo);
    ias_geo_destroy_proj_transformation(geo2image);

    return SUCCESS;
}
