/* System Includes */
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>

/* IAS Includes */
#include "ias_math.h"
#include "ias_logging.h"

/* Local Includes */
#include "cfmask_support.h"
#include "get_image_corners.h"

/* FIXME - Define this to use the same unusual mixed double/float precision
   used by the original LSRD version to make the results match. */
#define PRECISION

/* Number of cloud objects in an allocation block */
#define MIN_CLOUD_OBJ 9 

/* Define the nominal satellite height */
#define SATELLITE_HEIGHT_METERS 705000.0

/* Define a run-length encoded structure for storing a run of cloud pixels */
typedef struct rle_t
{
    int row;            /* Image row this record is on */
    int start_col;      /* Image column the pixel run starts on */
    int col_count;      /* Number of pixels in the run */
    int next_index;     /* Index to the next run */
} RLE_T;

/******************************************************************************
Name: create_cloud_runs

Purpose: Given a pixel mask, generate the set of run-length encoded line
    segments that cover the clouds.

Returns: SUCCESS/ERROR

******************************************************************************/
static int create_cloud_runs
(
    const unsigned char *pixel_mask, /* I: Cloud pixel mask */
    int nrows,                  /* I: Number of rows */
    int ncols,                  /* I: Number of columns */
    RLE_T **cloud_runs,         /* I/O: Array of cloud runs */
    int *out_run_count          /* O: Count of cloud runs found */
)
{
    int runs_allocated = 0;
    RLE_T *runs = NULL;
    RLE_T *temp_runs = NULL;
    int row;
    int run_count = 0;

    *out_run_count = 0;
    *cloud_runs = NULL;

    /* Loop over all the rows */
    for (row = 0; row < nrows; row++)
    {
        int col;
        const unsigned char *mask_row = &pixel_mask[row * ncols];

        /* Loop over the columns in this row */
        col = 0;
        while (col < ncols)
        {
            int run_len;
            RLE_T *run;

            /* Search for a starting cloud pixel */
            if (!(mask_row[col] & CF_CLOUD_BIT))
            {
                col++;
                continue;
            }

            /* Found a starting cloud pixel, so find the length of the cloud
               run */
            for (run_len = 1; run_len < ncols; run_len++)
            {
                if (!(mask_row[col + run_len] & CF_CLOUD_BIT))
                    break;
            }

            /* Grow the run array if needed */
            if (run_count >= runs_allocated)
            {
                /* Double the size of the array when more memory is needed */
                if (runs_allocated == 0)
                    runs_allocated = 10000;
                else
                    runs_allocated *= 2;

                temp_runs = realloc(runs, runs_allocated * sizeof(*runs));
                if (!temp_runs)
                {
                    IAS_LOG_ERROR("Allocating memory for identifying clouds");
                    free(runs);
                    return ERROR;
                }
                runs = temp_runs;
            }

            run = &runs[run_count];
            run_count++;

            run->row = row;
            run->start_col = col;
            run->col_count = run_len;
            run->next_index = -1;

            /* Resume checking at the next unchecked pixel */
            col += run_len + 1;
        }
    }

    /* Use realloc to shrink set of runs to the actual number found */
    temp_runs = realloc(runs, run_count * sizeof(*runs));
    if (!temp_runs)
    {
        IAS_LOG_ERROR("Failed shrinking cloud pixel run allocation");
        free(runs);
        return ERROR;
    }
    runs = temp_runs;

    *cloud_runs = runs;
    *out_run_count = run_count;

    return SUCCESS;
}

/******************************************************************************
Name: identify_clouds

Purpose: Given a pixel mask identifying individual pixels as clouds, group
    the individual pixels into cloud objects using run-length encoding.

Notes:
    - cloud number zero is reserved for the "no cloud" condition

Returns: SUCCESS/ERROR

******************************************************************************/
static int identify_clouds
(
    const unsigned char *pixel_mask, /* I: Cloud pixel mask */
    int nrows,                  /* I: Number of rows */
    int ncols,                  /* I: Number of columns */
    RLE_T **out_cloud_runs,     /* O: Array of cloud run-length encoded
                                      segments */
    int **out_cloud_lookup,     /* O: Array to map clouds to cloud runs */
    int **out_cloud_pixel_count,/* O: Cloud number */
    int *out_cloud_count,       /* O: Number of clouds in the cloud_lookup and
                                      cloud_pixel_count arrays */
    int *cloud_map              /* I/O: Image containing the cloud number for
                                        each pixel */
)
{
    RLE_T *runs = NULL;
    int run_count;
    int *cloud_lookup = NULL;  /* Array that points to the first RLE for each
                                  cloud number */
    int *temp_ptr;
    int next_cloud_number = 1;
    int run_index;
    int cloud_index;
    int cloud_count;
    int *cloud_pixel_count = NULL;

    *out_cloud_runs = NULL;
    *out_cloud_lookup = NULL;
    *out_cloud_pixel_count = NULL;
    *out_cloud_count = 0;

    /* Identify the set of "runs" of cloud pixels */
    if (create_cloud_runs(pixel_mask, nrows, ncols, &runs, &run_count)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Failed identifying clouds");
        return ERROR;
    }

    /* If no runs of clouds found, no need to do anything else */
    if (run_count == 0)
        return SUCCESS;

    /* Allocate the cloud lookup table sized to assume each run is a separate
       cloud.  Allocate one extra entry since the zero entry is reserved for
       indicating "no clouds". */
    cloud_lookup = malloc((run_count + 1) * sizeof(*cloud_lookup));
    if (!cloud_lookup)
    {
        IAS_LOG_ERROR("Failed allocating cloud lookup table");
        free(runs);
        return ERROR;
    }

    /* The first entry in the lookup is reserved for "no clouds" */
    cloud_lookup[0] = -1;

    /* Walk the cloud runs to group them into clouds */
    for (run_index = 0; run_index < run_count; run_index++)
    {
        RLE_T *run = &runs[run_index];
        int use_next = 1;
        int assigned_cloud_number = 0;
        int end_col = run->start_col + run->col_count;

        /* Check for overlap with clouds from the previous row if not the first
           row.  Note that the overlap includes cloud pixels on the diagonal,
           so the column range for the check includes an extra pixel on both
           ends. */
        if (run->row > 0)
        {
            int *prev_cloud_map_row = &cloud_map[(run->row - 1) * ncols];
            int col;
            int start = run->start_col - 1;
            if (start < 0)
                start = 0;

            for (col = start; col <= end_col; col++)
            {
                int cloud_number = prev_cloud_map_row[col];

                /* Check for a cloud in the previous row */
                if (cloud_number != 0)
                {
                    int fill_col;
                    int *cloud_map_row = &cloud_map[run->row * ncols];

                    /* Found a cloud, so assign this run to that cloud */
                    run->next_index = cloud_lookup[cloud_number];
                    cloud_lookup[cloud_number] = run_index;

                    /* Set this cloud number for the run in the cloud map */
                    for (fill_col = run->start_col; fill_col < end_col;
                        fill_col++)
                    {
                        cloud_map_row[fill_col] = cloud_number;
                    }

                    /* Assigned a cloud number, so no need to look */
                    assigned_cloud_number = cloud_number;
                    use_next = 0;
                    break;
                }
            }

            /* If an overlapping cloud from the previous row was found, look
               for additional overlapping clouds to merge with, continuing at
               the next column */
            for (col = col + 1; col <= end_col; col++)
            {
                int cloud_number = prev_cloud_map_row[col];

                /* Check for a cloud in the previous row */
                if (cloud_number != 0 && cloud_number != assigned_cloud_number)
                {
                    int fill_col;
                    int last_index;
                    int *cloud_map_row = &cloud_map[run->row * ncols];

                    /* Found another cloud to merge with, so renumber the newly
                       found cloud to match this number.  The cloud map update
                       is "lazy" since it just updates the cloud numbers in
                       the previous row and pixels to the left in the current
                       row.  A final pass will update the cloud_map to make
                       sure it has the correct cloud numbers for every pixel */
                    for (fill_col = 0; fill_col < ncols; fill_col++)
                    {
                        if (prev_cloud_map_row[fill_col] == cloud_number)
                        {
                            prev_cloud_map_row[fill_col]
                                = assigned_cloud_number;
                        }
                    }

                    /* Also fill the pixels on the current row that may belong
                       to the same cloud */
                    for (fill_col = 0; fill_col < col; fill_col++)
                    {
                        if (cloud_map_row[fill_col] == cloud_number)
                        {
                            cloud_map_row[fill_col] = assigned_cloud_number;
                        }
                    }

                    /* Find the last node in the cloud being merged */
                    last_index = cloud_lookup[cloud_number];
                    while (runs[last_index].next_index != -1)
                    {
                        last_index = runs[last_index].next_index;
                    }

                    /* Merge the newly found cloud list into the current
                       cloud's list */
                    runs[last_index].next_index
                        = cloud_lookup[assigned_cloud_number];
                    cloud_lookup[assigned_cloud_number]
                        = cloud_lookup[cloud_number];
                    cloud_lookup[cloud_number] = -1;
                }
            }
        }

        /* If no cloud number was assigned, use the next one */
        if (use_next)
        {
            int fill_col;
            int *cloud_map_row = &cloud_map[run->row * ncols];

            for (fill_col = run->start_col; fill_col < end_col; fill_col++)
            {
                cloud_map_row[fill_col] = next_cloud_number;
            }

            run->next_index = -1;
            cloud_lookup[next_cloud_number] = run_index;

            next_cloud_number++;

            /* Make sure we never overflow the capacity of the cloud counter
               since that would break this implementation.  This is incredibly
               unlikely since it would require much larger scenes than are
               currently processed. */
            if (next_cloud_number < 0)
            {
                IAS_LOG_ERROR("Too many clouds identified");
                free(cloud_lookup);
                free(runs);
                return ERROR;
            }
        }
    }

    /* Condense the cloud lookup table */
    cloud_count = next_cloud_number;
    next_cloud_number = 1;
    for (cloud_index = 1; cloud_index < cloud_count; cloud_index++)
    {
        if (cloud_lookup[cloud_index] != -1)
        {
            cloud_lookup[next_cloud_number] = cloud_lookup[cloud_index];
            next_cloud_number++;
        }
    }
    cloud_count = next_cloud_number;

    /* Shrink cloud_lookup array to only use what it needs */
    temp_ptr = realloc(cloud_lookup, sizeof(*cloud_lookup) * cloud_count);
    if (!temp_ptr)
    {
        IAS_LOG_ERROR("Failed shrinking cloud lookup array");
        free(cloud_lookup);
        free(runs);
        return ERROR;
    }
    cloud_lookup = temp_ptr;

    /* Allocate memory for the count of pixels in each cloud */
    cloud_pixel_count = malloc(cloud_count * sizeof(*cloud_pixel_count));
    if (!cloud_pixel_count)
    {
        IAS_LOG_ERROR("Failed allocating memory for the cloud pixel count "
            "array");
        free(cloud_lookup);
        free(runs);
        return ERROR;
    }

    /* Update the cloud map with the latest cloud numbers and to cover any
       cloud merges that occurred */
    for (cloud_index = 1; cloud_index < cloud_count; cloud_index++)
    {
        int run_index = cloud_lookup[cloud_index];
        int pixel_count = 0;

        while (run_index != -1)
        {
            RLE_T *run = &runs[run_index];
            int fill_col;
            int *cloud_map_row = &cloud_map[run->row * ncols];
            int end_col = run->start_col + run->col_count;

            for (fill_col = run->start_col; fill_col < end_col; fill_col++)
            {
                cloud_map_row[fill_col] = cloud_index;
            }
            pixel_count += run->col_count;

            run_index = run->next_index;
        }

        cloud_pixel_count[cloud_index] = pixel_count;
    }

    *out_cloud_runs = runs;
    *out_cloud_lookup = cloud_lookup;
    *out_cloud_pixel_count = cloud_pixel_count;
    *out_cloud_count = cloud_count;

    return SUCCESS;
}

/******************************************************************************
MODULE:  viewgeo

PURPOSE: Calculate the geometric parameters needed for the cloud/shadow match

RETURN: None
******************************************************************************/
static void viewgeo
(
    const int image_corner_lines[4],   /* I: Image corners (UL, UR, LR, LL) */
    const int image_corner_samples[4], /* I: Image corners (UL, UR, LR, LL) */
    double roll_angle_degrees,         /* I: Roll angle of the spacecraft */
    double pixel_size_meters,          /* I: Pixel size (in meters) */
    float *a,                          /* O: Coefficient */
    float *b,                          /* O: Coefficient */
    float *c,                          /* O: Coefficient */
    float *omiga_par,                  /* O: Angle of parallel line (radians) */
    float *omiga_per                   /* O: Angle of vertical line (radians) */
)
{
    int x_ul = image_corner_samples[0];
    int x_ur = image_corner_samples[1];
    int x_lr = image_corner_samples[2];
    int x_ll = image_corner_samples[3];
    int y_ul = image_corner_lines[0];
    int y_ur = image_corner_lines[1];
    int y_lr = image_corner_lines[2];
    int y_ll = image_corner_lines[3];
    float x_u;          /* Intermediate upper X center point */
    float x_l;          /* Intermediate lower X center point */
    float y_u;          /* Intermediate upper Y center point */
    float y_l;          /* Intermediate lower Y center point */
    float k_ulr = 1.0;  /* Intermediate slope value */
    float k_llr = 1.0;  /* Intermediate slope value */
    float k_aver;       /* Mean slope value */
    double across_track_shift_meters;
    double across_track_shift_pixels;
    double delta_x;
    double delta_y;

    x_u = (float)(x_ul + x_ur)/2.0;
    x_l = (float)(x_ll + x_lr)/2.0;
    y_u = (float)(y_ul + y_ur)/2.0;
    y_l = (float)(y_ll + y_lr)/2.0;

    if (x_ul != x_ur)
    {
        /* Get k of upper left and right points */
        k_ulr = (float)(y_ul - y_ur)/(float)(x_ul - x_ur);
    }
    else
        k_ulr = 0.0;
    if (x_ll != x_lr)
    {
        /* Get k of lower left and right points */
        k_llr = (float)(y_ll - y_lr)/(float)(x_ll - x_lr);
    }
    else
        k_llr = 0.0;

    k_aver = (k_ulr + k_llr)/2.0;

    /* Get the angle of parallel lines k (in radians) */
    *omiga_par = atan(k_aver);

    /* Factor in the roll angle to shift the nadir track to a better estimated
       location.  The roll angle is negated since it is defined as positive
       pointing to the left side image in the direction of satellite motion.
       So, a negative roll angle needs to move the nadir track to the right. */
    across_track_shift_meters = SATELLITE_HEIGHT_METERS
        * tan(-roll_angle_degrees * ias_math_get_radians_per_degree());
    across_track_shift_pixels = across_track_shift_meters / pixel_size_meters;
    delta_x = across_track_shift_pixels * cos(*omiga_par);
    delta_y = across_track_shift_pixels * sin(*omiga_par);

    y_u += delta_y;
    y_l += delta_y;
    x_u += delta_x;
    x_l += delta_x;

    /* At this time the line from (x_u, y_u) to (x_l, y_l) is an approximation
       of the nadir track in the image.  Calculate the A, B, and C coefficients
       to represent the line with the equation:
         AX(j) + BY(i) + C = 0
       Note that the Y axis is defined pointing down here, so the signs are
       a little different than expected. */
    *a = y_u - y_l;
    *b = x_l - x_u;
    *c = y_l * x_u - x_l * y_u;

    /* Get the angle which is perpendicular to the trace line */
    *omiga_per = atan((*b)/(*a));
}

/******************************************************************************
MODULE:  mat_truecloud

PURPOSE:  Modify the position of the cloud to account for the satellite
    viewing angle and estimated cloud height.

RETURN: None
******************************************************************************/
static void mat_truecloud
(
    const int *x,       /* I: Input pixel column */
    const int *y,       /* I: Input pixel row */
    int array_length,   /* I: Number of input array */
    int max_threads,    /* I: Maximum threads to use */
    const float *h,     /* I: Cloud pixel height */
    float a,            /* I: Coefficient */
    float b,            /* I: Coefficient */
    float c,            /* I: Coefficient */
    float omiga_par,    /* I: Angle of parallel line */
    float omiga_per,    /* I: Angle of perpendicular line */
    float *x_new,       /* O: Output pixel column */
    float *y_new        /* O: Output pixel row */
)
{
    float cos_omiga_par = cosf(omiga_par);
    float sin_omiga_par = sinf(omiga_par);
    float inv_cos_diff = 1.0 / cosf(omiga_per - omiga_par);
    float inv_coeff_dist = 1.0 / sqrtf(a * a + b * b);

    float inv_height = 1.0 / SATELLITE_HEIGHT_METERS;
    int i;

    /* Use OpenMP to parallelize this loop */
    #pragma omp parallel for private(i) num_threads(max_threads)
    for (i = 0; i < array_length; i++)
    {
        float dist;      /* Distance */
        float dist_par;  /* Distance in parallel direction */
        float dist_move; /* Distance moved */
        float delt_x;    /* Change in column */
        float delt_y;    /* Change in row */

        dist = (a * (float)x[i] + b * (float)y[i] + c) * inv_coeff_dist;
                 /* From the central perpendicular (unit: pixel) */
        dist_par = dist * inv_cos_diff;
        dist_move = (dist_par * h[i]) * inv_height; /* Cloud move distance (m)*/
        delt_x = dist_move * cos_omiga_par;
        delt_y = dist_move * sin_omiga_par;

        x_new[i] = x[i] + delt_x; /* New x, j */
        y_new[i] = y[i] + delt_y; /* New y, i */
    }
}

/******************************************************************************
NAME: count_masked_pixels

PURPOSE: Helper routine to create a cumulative count of the number of pixels
    that include the requested mask bit.

RETURN: None
******************************************************************************/
static void count_masked_pixels
(
    const unsigned char *in_mask,/* I: Mask to check for the masked bit */
    int samples,                /* I: Number of samples in the mask buffer */
    unsigned char bit_mask,     /* I: Bit mask to count in the buffer */
    int *count_buffer           /* O: Buffer holding the cumulative count */
)
{
    int sample;
    int count = 0;

    /* Loop through the sample in the input mask */
    for (sample = 0; sample < samples; sample++)
    {
        /* If the current sample has the bit set, add it to the count */
        if (in_mask[sample] & bit_mask)
            count++;

        /* Record the cumulative count in the count buffer */
        count_buffer[sample] = count;
    }
}

/******************************************************************************
NAME:  image_dilate

PURPOSE: Dilate the image with a n x n window

Note: The n x n window check is not a brute force check since that can be
    extremely slow if no pixels are present with the requested bit mask.
    Instead a rolling set of line buffers is created that keep the cumulative
    count of the number of pixels including the requested bit on each line.
    The cumulative count buffers can quickly be used to test each line by
    checking whether the cumulative count of bits set changes from the start to
    the end of the window.

RETURN: None
******************************************************************************/
static int image_dilate
(
    const unsigned char *in_mask, /* I: Mask to be dilated */
    int nlines,                   /* I: Number of lines in the mask */
    int nsamples,                 /* I: Number of samples in the mask */
    int idx,                      /* I: Pixel buffer 2 * idx + 1 */
    unsigned char bit_mask,       /* I: Type of image to dilute */
    unsigned char *out_mask       /* O: Mask after dilate */
)
{
    int buf_lines = 2 * idx + 1; /* Number of line buffers needed */
    int *count_line_buffers;     /* Memory for the number of lines in the
                                    window */
    int *count_line_map[buf_lines];/* Lookup map for each line buffer */
    int i;                         /* Loop counter */
    int line;

    IAS_LOG_INFO("Start dilate");

    /* Allocate a buffer for intermediate counts of pixels.  Each line allows
       for one extra entry to make handling the first sample on each line
       easier. */
    count_line_buffers = malloc(buf_lines * (nsamples + 1)
            * sizeof(*count_line_buffers));
    if (!count_line_buffers)
    {
        IAS_LOG_ERROR("Allocating memory for image dilation");
        return ERROR;
    }

    /* Initialize the line buffer pointers */
    for (i = 0; i < buf_lines; i++)
    {
        /* Set up the pointers to each line.  Each line leaves one entry 
           before it to make handling the start of the line easier. */
        count_line_map[i] = &count_line_buffers[i * (nsamples + 1) + 1];

        /* Use that extra entry to set the count off the left side of the 
           image to zero */
        count_line_map[i][-1] = 0;
    }

    /* Pre-populate the first lines, quitting one early since the main loop
       will populate each new line needed */
    for (line = 0; line < buf_lines - 1; line++)
    {
        count_masked_pixels(&in_mask[line * nsamples], nsamples, bit_mask, 
            count_line_map[line]);
    }

    /* Perform the dilation for each line of the mask image */
    for (line = idx; line < nlines - idx; line++)
    {
        int end_index;      /* Ending index for this line */
        int rotation_index; /* Index to use for the cumulative count line in
                               the rotating buffer */
        unsigned char *out_line; /* Current output line pointer */
        int sample;

        /* Calculate the last line index needed for the current window
           location */
        end_index = line + idx;

        /* Determine the next buffer to use in the rotating buffer */
        rotation_index = end_index % buf_lines;

        /* Populate the cumulative count buffer for this line */
        count_masked_pixels(&in_mask[end_index * nsamples], nsamples, bit_mask,
            count_line_map[rotation_index]);

        /* Set up the output line pointer */
        out_line = &out_mask[line * nsamples];

        /* Perform the dialation for each sample in this line */
        for (sample = idx; sample < nsamples - idx; sample++)
        {
            int mask = 0;

            /* Check each of the lines in the rotating line buffer.  The order
               they are checked in does not matter */
            for (i = 0; i < buf_lines; i++)
            {
                /* If the cumulative count changes between the sample to the
                   left of the current window and the last sample in the
                   window, the bit being checked for was set in the window */
                if (count_line_map[i][sample - idx - 1]
                    != count_line_map[i][sample + idx])
                {
                    mask = 1;
                    break;
                }
            }
            /* Set or clear the output mask pixel */
            if (mask)
                out_line[sample] |= bit_mask;
            else
                out_line[sample] &= ~bit_mask;
        }
    }

    free(count_line_buffers);
 
    IAS_LOG_INFO("End dilate");

    return SUCCESS;
}

/******************************************************************************
Name:  get_elevation_step

Purpose: Calculate the elevation step to use for the clouds when searching
    for the matching shadow.

Return: height step as an integer

******************************************************************************/
static int get_elevation_step
(
    double roll_angle_degrees,  /* I: Spacecraft roll angle */
    double sun_elevation_angle_degrees, /* I: Sun elevation angle */
    double pixel_size_meters    /* I: Pixel size used in the image */
)
{
    double largest_view_angle;  /* Worst case viewing angle */
    double max_step;            /* Largest step */

    /* Get the largest viewing angle in the scene.  A full scene is about 15
       degrees, so the largest angle will be half the full scene angle added
       to the roll angle */
    if (roll_angle_degrees >= 0.0)
        largest_view_angle = roll_angle_degrees + 7.5;
    else
        largest_view_angle = fabs(roll_angle_degrees - 7.5);

    /* Use the larger of the satellite viewing angle and the sun zenith
       angle to calculate the step.  The largest step from the two sources
       can't be used since if the sun is directly overhead the step would
       be infinity. */
    if (largest_view_angle > (90.0 - sun_elevation_angle_degrees))
    {
        /* Use the viewing angle to calculate the step.  Calculate the height
           step to move the cloud two pixels based on the largest viewing
           angle. */
        max_step = (2.0 * pixel_size_meters)
            / tan(largest_view_angle * ias_math_get_radians_per_degree());
    }
    else
    {
        /* Use the sun elevation angle to calculate the step.  Calculate the
           height step to move the shadow two pixels based on the sun
           elevation */
        max_step = 2.0 * pixel_size_meters * tan(sun_elevation_angle_degrees
            * ias_math_get_radians_per_degree());
    }

    /* Make sure the step is at least double the pixel size so the low sun
       elevations near the poles don't take too long to process */
    if (max_step < 2.0 * pixel_size_meters)
        max_step = 2 * pixel_size_meters;

    return rint(max_step);
}

/******************************************************************************
NAME:  cfmask_object_cloud_shadow_match

PURPOSE: Identify the final shadow pixels by doing a geometric cloud and shadow
         matching which ends in with the maximum cloud and shadow similarity

RETURN: SUCCESS
        FAILURE
******************************************************************************/
int cfmask_object_cloud_shadow_match
(
    const CFMASK_INPUTS *input, /* I: Input structure */
    const IAS_ANGLE_GEN_METADATA *angle_metadata, /* I: Angle metadata */
    double map_azimuth_angle_radians, /* I: Angle from image north to true 
                                            north */
    double roll_angle_degrees,  /* I: Satellite roll angle */
    double pixel_size_meters,   /* I: Pixel size in meters */
    float ptm,                  /* I: Percent of clear-sky pixels */
    float t_templ,              /* I: Percentile of low background temp */
    float t_temph,              /* I: Percentile of high background temp */
    int cldpix,                 /* I: Cloud buffer size */
    int sdpix,                  /* I: Shadow buffer size */
    int max_threads,            /* I: Maximum worker threads to use */
    int cfmask_dilate_flag,     /* I: Flag to turn on/off cloud dilation */
    unsigned char *pixel_mask,  /* I/O: Pixel mask */
    double sun_elevation_degrees,/* I: Solar elevation angle */
    double sun_azimuth_degrees, /* I: Solar azimuth angle */
    int use_thermal             /* I: Flag to use thermal band */
)
{
    int nrows = input->num_lines; /* Number of rows */
    int ncols = input->num_samps; /* Number of columns */
    int npixels = nrows * ncols;  /* Number of pixels in image */
    int col;                      /* Column index */
    float sun_from_east_radians;  /* Angle from the sun with east = 0 */
    int cloud_counter = 0;        /* Cloud pixel counter */
    int boundary_counter = 0;     /* Boundary pixel counter */
    float revised_ptm;            /* Revised percent of cloud */
    float t_similar;              /* Similarity threshold */
    float t_buffer;               /* Threshold for matching buffering */
    float max_similar = 0.95;     /* Max similarity threshold */
    int num_pix = 3;   /* Number of inward pixels (90m) for cloud base
                        temperature */
    float a, b, c, omiga_par, omiga_per; /* Variables used for viewgeo
                                            routine, see it for detail */
    int i_step;     /* iteration step */
    int *cloud_pixel_count = NULL; /* Array for the count of pixels
                                            in each of the identified clouds */
    int *cloud_lookup = NULL;  /* Array to point to the first run for each
                                  cloud */
    RLE_T *cloud_runs = NULL;  /* Array of cloud run-length encoded segments */
    unsigned int max_cloud_pixels = 0; /* max number of cloud pixels in a single
                                     cloud */
    int *cloud_map;            /* Image sized array with cloud numbers */
    int run_index;             /* Index into the cloud_runs */
    int cloud_type;            /* Cloud type iterator */
    float *cloud_pos_row_col;
    float *cloud_pos_row;
    float *cloud_pos_col;
    int *cloud_orig_row_col;   /* Array for original cloud row/col locations */
    int *cloud_orig_row;
    int *cloud_orig_col;
    int16_t *temp_obj = NULL;  /* Temperature for each cloud */
    int16_t temp_obj_max = 0;  /* Maximum temperature for each cloud */
    int16_t temp_obj_min = 0;  /* Minimum temperature for each cloud */
    int index;                 /* Loop index */
    float pct_obj;             /* Percent of edge pixels */
    float t_obj;               /* Cloud percentile value */
    float rate_elapse = 6.5;   /* Wet air lapse rate (degree K / km) */
    float rate_dlapse = 9.8;   /* Dry air lapse rate (degree K / km) */
    int max_cl_height;         /* Max cloud base height (m) */
    int min_cl_height;         /* Min cloud base height (m) */
    int max_height;            /* Refined maximum height (m) */
    int min_height;            /* Refined minimum height (m) */
    float record_thresh;       /* Record thresh value */
    int base_h;                /* Cloud base height */
    int out_all;               /* Total number of pixels outside image */
    int match_all;             /* Total number of matched pixels */
    int total_all;             /* Total number of pixels */
    float thresh_match;        /* Thresh match value */
    int i;
    int cloud_count = 0;       /* Cloud counter */
    int shadow_count = 0;      /* Shadow counter */
    int num_clouds;
    double radians_per_degree = ias_math_get_radians_per_degree();

    /* Define the X and Y (sample, line) components of the unit vector of the
       light direction from the sun */
#ifdef PRECISION
    double shadow_unit_vec_y;
    double shadow_unit_vec_x;
#else
    float shadow_unit_vec_y;
    float shadow_unit_vec_x;
#endif
    double tan_sun_elevation;

    /* Dynamic memory allocation */
    unsigned char *cal_mask = NULL;      /* Calibration pixel mask */

    /* Timing message */
    IAS_LOG_INFO("Starting cfmask object");

    cal_mask = calloc(npixels, sizeof(*cal_mask));
    if (cal_mask == NULL)
    {
        IAS_LOG_ERROR ("Allocating cal_mask memory");
        return ERROR;
    }

    IAS_LOG_DEBUG("map azimuth (radians) %f\n", map_azimuth_angle_radians);

    /* Correct the solar azimuth angle by adding the map azimuth angle to it
       to adjust for projections where north isn't straight up (like polar
       stereographic).  The sun azimuth angle is defined as zero being straight
       north with positive angles being clockwise. */
    sun_azimuth_degrees 
        += (map_azimuth_angle_radians * ias_math_get_degrees_per_radian());

    /* Reverse the solar azimuth angle since we're using it to cast the cloud
       shadow to the ground */
    sun_azimuth_degrees += 180;

    /* Make sure the resulting angle is in the -180 to 180 range. */
    while (sun_azimuth_degrees > 180.0)
        sun_azimuth_degrees -= 360.0;
    while (sun_azimuth_degrees < -180.0)
        sun_azimuth_degrees += 360.0;

    IAS_LOG_DEBUG("Angle from the sun (degrees from 'up'): %f\n",
        sun_azimuth_degrees);

    /* Solar elevation angle */
    tan_sun_elevation = tan(radians_per_degree * sun_elevation_degrees);

    /* Calculate the X (sample) and Y (line) components of the vector from the
       sun.  The angle is shifted by 90 degrees to make east be zero degrees. */
    sun_from_east_radians = radians_per_degree * (sun_azimuth_degrees - 90.0);
#ifdef PRECISION
    shadow_unit_vec_y = sin(sun_from_east_radians);
    shadow_unit_vec_x = cos(sun_from_east_radians);
#else
    shadow_unit_vec_y = sinf(sun_from_east_radians);
    shadow_unit_vec_x = cosf(sun_from_east_radians);
#endif

    for (index = 0; index < npixels; index++)
    {
        if (pixel_mask[index] & CF_CLOUD_BIT)
            cloud_counter++;

        /* Boundary layer includes both cloud_mask equals 0 and 1 */
        if (!(pixel_mask[index] & CF_FILL_BIT))
            boundary_counter++;
    }

    /* Revised percent of cloud on the scene after plcloud */
    if (boundary_counter != 0)
        revised_ptm = (float)cloud_counter / (float)boundary_counter;
    else
        revised_ptm = 0.0;

    /* If 0.1% or less of the scene is clear sky or the revised cloud fraction
       is greater than 90%, consider the non-cloud bits to be shadows */
    if (ptm <= 0.1 || revised_ptm >= 0.90)
    {
        for (index = 0; index < npixels; index++)
        {
            /* No Shadow Match due to too much cloud (>90 percent) 
               non-cloud pixels are just shadow pixels */
            if (!(pixel_mask[index] & CF_CLOUD_BIT) &&
                (!(pixel_mask[index] & CF_FILL_BIT)))
            {
                pixel_mask[index] |= CF_SHADOW_BIT;
            }
        }
    }
    else
    {
        short int *thermal_buf = input->therm_buf;
        int image_corner_lines[4];
        int image_corner_samples[4];

        /* Get the elevation step to use when iterating over the potential
           cloud heights */
        i_step = get_elevation_step(roll_angle_degrees, sun_elevation_degrees,
            pixel_size_meters);

        /* Get the image corners in satellite orientation (UL, UR, LR, LL) */
        if (get_image_corners(angle_metadata, image_corner_lines,
                image_corner_samples) != SUCCESS)
        {
            IAS_LOG_ERROR("Failed retrieving image corners");
            free(cal_mask);
            return ERROR;
        }

        /* Get view angle geometry */
        viewgeo(image_corner_lines, image_corner_samples, roll_angle_degrees,
                pixel_size_meters, &a, &b, &c, &omiga_par, &omiga_per);

        IAS_LOG_DEBUG("omiga_par %f\n", omiga_par);
        IAS_LOG_DEBUG("omiga_per %f\n", omiga_per);

        /* Allocate memory for array to map image pixels to cloud numbers */
        cloud_map = calloc(npixels, sizeof(*cloud_map));
        if (cloud_map == NULL)
        {
            IAS_LOG_ERROR("Allocating memory for cloud map");
            free(cal_mask);
            return ERROR;
        }

        /* Labeling the cloud pixels */
        IAS_LOG_INFO("Labeling clouds");

        if (identify_clouds(pixel_mask, nrows, ncols, &cloud_runs,
            &cloud_lookup, &cloud_pixel_count, &num_clouds, cloud_map)
            != SUCCESS)
        {
            IAS_LOG_ERROR("Failed labeling clouds");
            free(cloud_map);
            free(cal_mask);
            return ERROR;
        }

        /* If there are no clouds, set the maximum cloud pixels to 1 since
           leaving it at zero would cause problems later */
        if (num_clouds == 0)
            max_cloud_pixels = 1;

        /* The cloud pixels are not counted as cloud pixels if the total number
           of cloud pixels is less than MIN_CLOUD_OBJ within a cloud cluster */
        IAS_LOG_INFO("Filtering clouds");
        for (index = 1; index < num_clouds; index++)
        {
            if (cloud_pixel_count[index] <= MIN_CLOUD_OBJ)
            {
                cloud_pixel_count[index] = 0;
                cloud_lookup[index] = -1;
                continue;
            }

            if (max_cloud_pixels < cloud_pixel_count[index])
                max_cloud_pixels = cloud_pixel_count[index];
        } /* num_clouds loop */

        /* Build a copy of the pixel mask with only the cloud bits set and
           only where the cloud is larger than the MIN_CLOUD_OBJ size.  The
           cal_mask is already initialized to zero when allocated. */
        for (index = 0; index < npixels; index++)
        {
            unsigned char pixel_mask_value = pixel_mask[index];

            if ((pixel_mask_value & CF_CLOUD_BIT)
                && (!(pixel_mask_value & CF_FILL_BIT))
                && (cloud_pixel_count[cloud_map[index]] != 0))
            {
                cal_mask[index] = CF_CLOUD_BIT;
            }
        }

        IAS_LOG_INFO("Finding shadows");

        /* Allocate space for the row/col locations of the height adjusted
           cloud */
        cloud_pos_row_col = malloc(2 * max_cloud_pixels
            * sizeof(*cloud_pos_row_col));

        /* Allocate space for the row/col locations of the original cloud */
        cloud_orig_row_col = malloc(2 * max_cloud_pixels
            * sizeof(*cloud_orig_row_col));
        if (cloud_pos_row_col == NULL || cloud_orig_row_col == NULL)
        {
            IAS_LOG_ERROR ("Allocating cloud memory");
            free(cloud_orig_row_col);
            free(cloud_pos_row_col);
            free(cloud_map);
            free(cal_mask);
            return ERROR;
        }
        /* Set up pointers to the row/column info for the cloud row/col info */
        cloud_pos_row = cloud_pos_row_col;
        cloud_pos_col = &cloud_pos_row_col[max_cloud_pixels];
        cloud_orig_row = cloud_orig_row_col;
        cloud_orig_col = &cloud_orig_row_col[max_cloud_pixels];

        if (use_thermal)
        {
            /* Temperature of the cloud object */
            temp_obj = malloc(max_cloud_pixels * sizeof(int16_t));
            if (temp_obj == NULL)
            {
                IAS_LOG_ERROR ("Allocating temp_obj memory");
                free(cloud_orig_row_col);
                free(cloud_pos_row_col);
                free(cloud_map);
                free(cal_mask);
                return ERROR;
            }
        }

        /* Use iteration to get the optimal move distance, Calculate the
           moving cloud shadow */
        for (cloud_type = 1; cloud_type < num_clouds; cloud_type++)
        {
            float *cloud_height;              /* Cloud height */
            float *matched_height;            /* Best match height values */
            float cloud_radius;               /* Cloud radius */
            short int t_obj_int;              /* Integer object temperature */
            int cloud_pixels = cloud_pixel_count[cloud_type];

            if (cloud_pixels == 0)
                continue;

            /* For larger (> 10% scene area), use another set of t_similar
               and t_buffer to address some missing cloud shadow at the
               edge area. */
            if (cloud_pixels <= (int)(0.1 * boundary_counter))
            {
                t_similar = 0.3;
                t_buffer = 0.95;
            }
            else
            {
                t_similar = 0.1;
                t_buffer = 0.98;
            }

            /* Build the set of pixels for the current cloud and find the 
               min/max temperatures present in the cloud */
            temp_obj_max = SHRT_MIN;
            temp_obj_min = SHRT_MAX;
            index = 0;

            run_index = cloud_lookup[cloud_type];
            while (run_index != -1)
            {
                RLE_T *run = &cloud_runs[run_index];
                int end_col = run->start_col + run->col_count;

                for (col = run->start_col; col < end_col; col++)
                {
                    if (use_thermal)
                    {
                        temp_obj[index] = thermal_buf[run->row * ncols + col];
                        if (temp_obj[index] > temp_obj_max)
                            temp_obj_max = temp_obj[index];
                        if (temp_obj[index] < temp_obj_min)
                            temp_obj_min = temp_obj[index];
                    }
                    cloud_orig_col[index] = col;
                    cloud_orig_row[index] = run->row;
                    index++;
                }
                run_index = run->next_index;
            }

            /* Make sure the number of pixels counted is the same as the
               number expected */
            if (index != cloud_pixels)
            {
                IAS_LOG_ERROR("Inconsistent number of pixels found in a cloud"
                    "%d/%d - this is a bug", index, cloud_pixels);
                free(temp_obj);
                free(cloud_orig_row_col);
                free(cloud_pos_row_col);
                free(cloud_map);
                free(cal_mask);
                return ERROR;
            }

            /* Set the default min/max cloud height */
            min_cl_height = 200;
            max_cl_height = 12000;

            if (use_thermal)
            {
                /* The base temperature for cloud.  Assumes object is round with
                   cloud_radius being the radius of cloud */
                cloud_radius = sqrt(cloud_pixels / (2.0 * ias_math_get_pi()));

                /* Number of inward pixels for correct temperature */
                if (cloud_radius > num_pix)
                {
                    /* For small clouds, just use the minimum temperature */
                    pct_obj  = ((cloud_radius - num_pix) *
                        (cloud_radius - num_pix)) /
                        (cloud_radius * cloud_radius);
                    if (find_percentile_threshold_short(temp_obj, cloud_pixels,
                        temp_obj_min, temp_obj_max, 100.0 * pct_obj, &t_obj)
                        != SUCCESS)
                    {
                        IAS_LOG_ERROR(
                            "Error calling find_percentile_threshold_short");
                        free(temp_obj);
                        free(cloud_orig_row_col);
                        free(cloud_pos_row_col);
                        free(cloud_map);
                        free(cal_mask);
                        return ERROR;
                    }
                }
                else
                {
                    /* Use the minimum temperature instead */
                    t_obj = temp_obj_min;
                }
                t_obj_int = rint(t_obj);

                /* Refine the cloud height range in meters based on the
                   temperature difference in the clouds */
                min_height = (int)rint(1000.0 * (t_templ - t_obj)
                    / SCALING_FACTOR_DN_BT / rate_dlapse);
                max_height = (int)rint(1000.0 * (t_temph - t_obj)
                    / SCALING_FACTOR_DN_BT);

                /* Pick the smallest height range based */
                if (min_cl_height < min_height)
                    min_cl_height = min_height;
                if (max_cl_height > max_height)
                    max_cl_height = max_height;

                /* Put the edge of the cloud the same value as t_obj */
                for (index = 0; index < cloud_pixels; index++)
                {
                    if (temp_obj[index] > t_obj_int)
                        temp_obj[index] = t_obj_int;
                }
            }

            /* Allocate memory for cloud height and matched height */
            cloud_height = malloc(cloud_pixels * sizeof(float));
            matched_height = calloc(cloud_pixels, sizeof(float));
            if (cloud_height == NULL || matched_height == NULL)
            {
                IAS_LOG_ERROR ("Allocating cloud height memory");
                free(matched_height);
                free(cloud_height);
                free(temp_obj);
                free(cloud_orig_row_col);
                free(cloud_pos_row_col);
                free(cloud_map);
                free(cal_mask);
                return ERROR;
            }

            /* Initialize height and similarity info */
            record_thresh = 0.0;
            for (base_h = min_cl_height; base_h <= max_cl_height;
                 base_h += i_step)
            {
                float inv_shadow_step
                    = 1.0 / (pixel_size_meters * tan_sun_elevation);

                if (use_thermal)
                {
                    float temp_to_height = 1000.0 / SCALING_FACTOR_DN_BT
                        / rate_elapse;

                    for (index = 0; index < cloud_pixels; index++)
                    {
                        cloud_height[index] = temp_to_height
                            * (t_obj - (float)temp_obj[index]) + (float)base_h;
                    }
                }
                else
                {
                    for (index = 0; index  < cloud_pixels; index++)
                    {
                        cloud_height[index] = base_h;
                    }
                }

                /* Get the true position of the cloud
                   calculate cloud DEM with initial base height */
                mat_truecloud(cloud_orig_col, cloud_orig_row, cloud_pixels,
                    max_threads, cloud_height, a, b, c, omiga_par, omiga_per,
                    cloud_pos_col, cloud_pos_row);

                out_all = 0;
                match_all = 0;
                total_all = 0;

                /* Use OpenMP to parallelize this loop.  Reduce the three
                   counters at the end. */
                #pragma omp parallel for private(i) num_threads(max_threads) \
                 reduction(+:out_all,match_all,total_all)
                for (i = 0; i < cloud_pixels; i++)
                {
                    float i_xy;      /* Intermediate cloud height */
                    int row;
                    int col;

                    i_xy = cloud_height[i] * inv_shadow_step;

#ifdef PRECISION
                    col = rint(cloud_pos_col[i] + i_xy * shadow_unit_vec_x);
                    row = rint(cloud_pos_row[i] + i_xy * shadow_unit_vec_y);
#else
                    col = rintf(cloud_pos_col[i] + i_xy * shadow_unit_vec_x);
                    row = rintf(cloud_pos_row[i] + i_xy * shadow_unit_vec_y);
#endif

                    /* The id that is out of the image */
                    if (row < 0 || row >= nrows || col < 0 || col >= ncols)
                    {
                        out_all++;
                    }
                    else
                    {
                        int c_value;
                        unsigned char mask = pixel_mask[row * ncols + col];

                        c_value = cloud_map[row * ncols + col];

                        if ((mask & CF_FILL_BIT)
                             || ((c_value != cloud_type)
                                 && (mask & (CF_CLOUD_BIT | CF_SHADOW_BIT))))
                        {
                            match_all++;
                        }
                        if (c_value != cloud_type)
                        {
                            total_all++;
                        }
                    }
                }
                match_all += out_all;
                total_all += out_all;

                thresh_match = (float)match_all/(float)total_all;

                if (thresh_match >= t_buffer * record_thresh &&
                    base_h < max_cl_height - i_step &&
                    record_thresh < max_similar)
                {
                    if (thresh_match > record_thresh)
                    {
                        record_thresh = thresh_match;

                        /* Save the new heights */
                        memcpy(matched_height, cloud_height, 
                            sizeof(*matched_height) * cloud_pixels);
                    }
                }
                else if (record_thresh > t_similar)
                {
                    /* Match in the cloud shape, so set the shadow bits in
                       the mask */

                    /* Recalculate the cloud position using the height with
                       the best match */
                    mat_truecloud(cloud_orig_col, cloud_orig_row, cloud_pixels,
                        max_threads, matched_height, a, b, c, omiga_par,
                        omiga_per, cloud_pos_col, cloud_pos_row);

                    /* Use OpenMP to parallelize this loop */
                    #pragma omp parallel for private(i) num_threads(max_threads)
                    for (i = 0; i < cloud_pixels; i++)
                    {
                        int row;
                        int col;
                        float i_vir;

                        i_vir = matched_height[i] * inv_shadow_step;

#ifdef PRECISION
                        col = rint(cloud_pos_col[i]+ i_vir * shadow_unit_vec_x);
                        row = rint(cloud_pos_row[i]+ i_vir * shadow_unit_vec_y);
#else
                        col = rintf(cloud_pos_col[i]
                            + i_vir * shadow_unit_vec_x);
                        row = rintf(cloud_pos_row[i]
                            + i_vir * shadow_unit_vec_y);
#endif

                        /* Put data within range */
                        if (row < 0)
                            row = 0;
                        else if (row >= nrows)
                            row = nrows - 1;
                        if (col < 0)
                            col = 0;
                        else if (col >= ncols)
                            col = ncols - 1;

                        cal_mask[row * ncols + col] |= CF_SHADOW_BIT;
                    }

                    /* Done with this cloud */
                    break;
                }
                else
                {
                    record_thresh = 0.0;
                    continue;
                }
            }
            free(cloud_height);
            cloud_height = NULL;
            free(matched_height);
            matched_height = NULL;
        }

        /* Free all the memory */
        free(cloud_pos_row_col);
        cloud_pos_row_col = NULL;
        cloud_pos_row = NULL;
        cloud_pos_col = NULL;
        free(cloud_orig_row_col);
        cloud_orig_row_col = NULL;
        cloud_orig_row = NULL;
        cloud_orig_col = NULL;
        free(temp_obj);
        temp_obj = NULL;
            
        free(cloud_pixel_count);
        cloud_pixel_count = NULL;
        free(cloud_map);
        cloud_map = NULL;
        free(cloud_lookup);
        cloud_lookup = NULL;
        free(cloud_runs);
        cloud_runs = NULL;

        /* Check cloud dilate flag is turn on/off, 1 = on, 0 = off */
        if (cfmask_dilate_flag)
        {
            /* Dilate the cloud if requested.  If the cloud isn't dilated,
               the original clouds present in the pixel mask haven't been
               changed and therefore nothing needs to be done to update the
               pixel mask */
            if (image_dilate(cal_mask, nrows, ncols, cldpix, CF_CLOUD_BIT,
                pixel_mask) != SUCCESS)
            {
                IAS_LOG_ERROR("Failed dilating the cloud pixels");
                free(cal_mask);
                return ERROR;
            }
        }

        /* Unconditionaly dilate the shadow.  The main reason to do this is
           to fill in holes in shadow.  The holes exist in the shadow because
           the cloud is treated as a 2D height map in the calculations instead
           of a full volumetric cloud.  The result is a cloud shadow that is
           full of holes. */
        if (image_dilate(cal_mask, nrows, ncols, sdpix, CF_SHADOW_BIT,
                pixel_mask) != SUCCESS)
        {
            IAS_LOG_ERROR("Failed dilating the shadow pixels");
            free(cal_mask);
            return ERROR;
        }

    } /* ptm > 0.1 && revised_ptm < 0.90 */

    /* Release the memory */
    free(cal_mask);
    cal_mask = NULL;

    /* Convert the pixel mask from a bit mask to an integer value */
    for (index = 0; index < npixels; index++)
    {
        unsigned char bit_mask = pixel_mask[index];

        if (bit_mask & CF_FILL_BIT)
            pixel_mask[index] = CF_FILL_PIXEL;
        else if (bit_mask & CF_CLOUD_BIT)
        {
            pixel_mask[index] = CF_CLOUD_PIXEL;
            cloud_count++;
        }
        else if (bit_mask & CF_SHADOW_BIT)
        {
            pixel_mask[index] = CF_CLOUD_SHADOW_PIXEL;
            shadow_count++;
        }
        else if (bit_mask & CF_SNOW_BIT)
            pixel_mask[index] = CF_SNOW_PIXEL;
        else if (bit_mask & CF_WATER_BIT)
            pixel_mask[index] = CF_WATER_PIXEL;
        else
            pixel_mask[index] = CF_CLEAR_PIXEL;
    }

    /* Timing message */
    IAS_LOG_INFO("Completed cfmask object");

    return SUCCESS;
}
