/* System Includes */
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>

/* IAS Includes */
#include "ias_geo.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"

/* Local Includes */
#include "assess_cloud_cover.h"
#include "cfmask_support.h"

/* Define the amount to buffer temperatures (4 degrees) */
#define TEMPERATURE_BUFFER (4 * SCALING_FACTOR_DN_BT)

/* Define this to enable the saturation checks for sensors before L8 */
/*#define HANDLE_SATURATION 1 */


/*****************************************************************************
Name: phase 1

Purpose: Perform initial tests to classify pixels

Returns: nothing

*****************************************************************************/
static void phase1
(
    const CFMASK_INPUTS *input,       /* I: Input structure */
    int use_thermal,                  /* I: Flag to use thermal band */
    int use_cirrus,                   /* I: Flag to use cirrus band */
    unsigned char *pixel_mask_array,  /* O: Pixel mask */
    unsigned char *clear_mask_array,  /* O: Mask for clear pixels */
    float *clear_sky_threshold,       /* O: Percent of clear-sky pixels */
    float *clear_land_percent,        /* O: Percent of clear land pixels */
    float *clear_water_percent        /* O: Percent of clear water pixels */
)
{
    int npixels = input->num_lines * input->num_samps; /* Number of pixels */
    int index;
    int non_fill_count = 0;       /* Count of non-fill pixels */
    int clear_pixel_counter = 0;       /* Clear sky pixel counter */
    int clear_land_pixel_counter = 0;  /* Clear land pixel counter */
    int clear_water_pixel_counter = 0; /* Clear water pixel counter */

    short int *blue_buf = input->buf[CF_BLUE_BAND];
    short int *green_buf = input->buf[CF_GREEN_BAND];
    short int *red_buf = input->buf[CF_RED_BAND];
    short int *nir_buf = input->buf[CF_NIR_BAND];
    short int *swir1_buf = input->buf[CF_SWIR1_BAND];
    short int *swir2_buf = input->buf[CF_SWIR2_BAND];
    short int *cirrus_buf = input->buf[CF_CIRRUS_BAND];
    short int *thermal_buf = input->therm_buf;

    IAS_LOG_INFO("Starting phase1");

    for (index = 0; index < npixels; index++)
    {
        unsigned char pixel_mask = 0;
        unsigned char clear_mask = 0;
        short int blue = blue_buf[index];
        short int green = green_buf[index];
        short int red = red_buf[index];
        short int nir = nir_buf[index];
        short int swir1 = swir1_buf[index];
        short int swir2 = swir2_buf[index];
        short int thermal = 0;
        short int cirrus = 0;

        if (use_thermal)
            thermal = thermal_buf[index];
        if (use_cirrus)
            cirrus  = cirrus_buf[index];

        /* Process non-fill pixels only */
        if (blue == FILL_PIXEL || green == FILL_PIXEL ||
           red == FILL_PIXEL   || nir == FILL_PIXEL   ||
           swir1 == FILL_PIXEL || swir2 == FILL_PIXEL
           || (use_thermal && thermal == FILL_PIXEL)
           || (use_cirrus && cirrus == FILL_PIXEL))
        {
            /* Pixel is fill, so set the fill bit */
            pixel_mask = CF_FILL_BIT;
        }
        else
        {
            float ndvi;                 /* Vegetation index */
            float ndsi;                 /* Snow index */
            float whiteness = 0.0;      /* Whiteness value*/
            float hot;                  /* Hot value for hot test */
#ifdef HANDLE_SATURATION
            int satu_bv;         /* Sum of saturated bands 1, 2, 3 value */
#endif

            /* Count the non-fill pixels */
            non_fill_count++;

            if (red + nir != 0)
                ndvi = (float)(nir - red)/(nir + red);
            else
                ndvi = 0.01;

            if (green + swir1 != 0)
                ndsi = (float)(green - swir1)/(green + swir1);
            else
                ndsi = 0.01;

            /* Basic cloud test, equation 1 */
            if (ndsi < 0.8 && ndvi < 0.8 && swir2 > 300 
                && (!use_thermal || thermal < 2700))
            {
                float visi_mean;          /* Mean of visible bands */
                pixel_mask = CF_CLOUD_BIT;

                /* Visible bands flatness (sum(abs)/mean < 0.6 => bright
                   and dark cloud), equation 2 */
                visi_mean = (float)(blue + green + red)/3.0;
                if (visi_mean != 0)
                {
                    whiteness = ((fabs(blue - visi_mean)
                              + fabs(green - visi_mean)
                              + fabs(red - visi_mean))) / visi_mean;
                }
                else
                {
                    /* Just put a large value to remove them from cloud
                       pixel identification */
                    whiteness = 100.0;
                }

            }
            else
            {
                pixel_mask &= ~CF_CLOUD_BIT;
            }

#ifdef HANDLE_SATURATION
            /* Conditionally include this code for sensors that have
               saturation */
            /* Update cloud_mask,  if one visible band is saturated, 
               whiteness = 0, due to data type conversion, pixel value
               difference of 1 is possible */
            if (blue >= input->saturated_value_max[0] - 1 ||
                green >= input->saturated_value_max[1] - 1 ||
                red >= input->saturated_value_max[2] - 1)
            {
                whiteness = 0.0;
                satu_bv = 1;
            }
            else
            {
                satu_bv = 0;
            }
#endif

            if (whiteness >= 0.7)
                pixel_mask &= ~CF_CLOUD_BIT;

            /* Haze test, equation 3 */
            hot = blue - 0.5 * red - 800.0;

#ifdef HANDLE_SATURATION
            if ((hot <= 0.0) && (satu_bv != 1))
                pixel_mask &= ~CF_CLOUD_BIT;
#else
            if (hot <= 0.0)
                pixel_mask &= ~CF_CLOUD_BIT;
#endif

            /* Ratio 4/5 > 0.75 test, equation 4 */
            if (swir1 != 0 && (pixel_mask & CF_CLOUD_BIT))
            {
                if ((float)nir/(float)swir1 <= 0.75)
                    pixel_mask &= ~CF_CLOUD_BIT;
            }
            else
                pixel_mask &= ~CF_CLOUD_BIT;

            /* Cirrus cloud test */
            if (use_cirrus)
            {
                if ((pixel_mask & CF_CLOUD_BIT)
                    || (float)(cirrus / 400.0 - 0.25) > 0.0)
                {
                    pixel_mask |= CF_CLOUD_BIT;
                }
                else
                    pixel_mask &= ~CF_CLOUD_BIT;
            }

            /* It takes every snow pixels including snow pixel under thin
               clouds or icy clouds, equation 20 */
            if (ndsi > 0.15 &&  nir > 1100 && green > 1000
                && (!use_thermal || thermal < 1000))
            {
                pixel_mask |= CF_SNOW_BIT;
            }

            /* Water test (works over thin cloud), equation 5 */
            if ((ndvi < 0.01 && nir < 1100)
                || (ndvi < 0.1 && ndvi > 0.0 && nir < 500))
            {
                pixel_mask |= CF_WATER_BIT;
            }

            /* Test whether use thermal band or not */
            if (!(pixel_mask & CF_CLOUD_BIT))
            {
                clear_mask |= CF_CLEAR_BIT;
                clear_pixel_counter++;

                if (!(pixel_mask & CF_WATER_BIT))
                {
                    clear_mask |= CF_CLEAR_LAND_BIT;
                    clear_land_pixel_counter++;
                }
                else
                {
                    clear_mask |= CF_CLEAR_WATER_BIT;
                    clear_water_pixel_counter++;
                }
            }
        }

        pixel_mask_array[index] = pixel_mask;
        clear_mask_array[index] = clear_mask;

    } /* pixel loop */

    /* Calculate the percentage of clear sky and land pixels */
    *clear_sky_threshold = 100.0 * ((float)clear_pixel_counter
        / (float)non_fill_count);
    *clear_land_percent = 100.0 * ((float)clear_land_pixel_counter
        / (float)non_fill_count);
    *clear_water_percent = 100.0 * ((float)clear_water_pixel_counter
        / (float)non_fill_count);

    IAS_LOG_INFO("Completed phase1");
}

/*****************************************************************************
Name: find_temperature_thresholds

Purpose: Find the temperature thresholds over land and water

Returns: SUCCESS/ERROR

*****************************************************************************/
static int find_temperature_thresholds
(
    const CFMASK_INPUTS *input, /* I: Input structure */
    unsigned char clear_land_mask,  /*I: Mask for clear land */
    unsigned char clear_water_mask, /*I: Mask for clear water */
    float low_threshold,        /* I: Low temperature threshold */
    float high_threshold,       /* I: High temperature threshold */
    const unsigned char *clear_mask_array, /* I: Mask for clear pixels */
    float *low_temp_threshold,  /* O: Percentile of low background temp */
    float *high_temp_threshold, /* O: Percentile of high background temp */
    float *t_wtemp              /* O: High percentile water temperature */
)
{
    int npixels = input->num_lines * input->num_samps; /* Number of pixels */
    short int thermal_max = input->therm_max;
    short int thermal_min = input->therm_min;
    const short int *thermal_buf = input->therm_buf;

    IAS_LOG_INFO("Starting find_temperature_thresholds");

    /* Percentile background temperature (low) */
    if (find_masked_percentile_threshold_short(thermal_buf, npixels,
        clear_mask_array, clear_land_mask, thermal_min, thermal_max,
        100.0 * low_threshold, low_temp_threshold) != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating the low temperature threshold");
        return ERROR;
    }

    /* Percentile background temperature (high) */
    if (find_masked_percentile_threshold_short(thermal_buf, npixels,
        clear_mask_array, clear_land_mask, thermal_min, thermal_max,
        100.0 * high_threshold, high_temp_threshold) != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating the high temperature threshold");
        return ERROR;
    }

    /* Percentile water background temperature (high) */
    if (find_masked_percentile_threshold_short(thermal_buf, npixels,
        clear_mask_array, clear_water_mask, thermal_min, thermal_max,
        100.0 * high_threshold, t_wtemp) != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating the water high temperature threshold");
        return ERROR;
    }

    IAS_LOG_INFO("Completed find_temperature_thresholds");

    return SUCCESS;
}

/*****************************************************************************
Name: phase2

Purpose: Refine the pixels that are identified as clouds using dynamic
    thresholds over land and water

Returns: SUCCESS/ERROR

******************************************************************************/
static int phase2
(
    const CFMASK_INPUTS *input, /* I: Input structure */
    float t_wtemp,              /* I: High percentile water temperature */
    float cloud_prob_threshold, /* I: Cloud probability threshold */
    float low_temp_threshold,   /* I: Percentile of low background temp */
    float high_temp_threshold,  /* I: Percentile of high background temp */
    float high_threshold,
    unsigned char clear_land_mask,  /* I: Mask for clear land */
    unsigned char clear_water_mask, /* I: Mask for clear water */
    const unsigned char *clear_mask, /* I: Mask for clear pixels */
    int use_thermal,            /* I: Flag to use thermal band */
    int use_cirrus,             /* I: Flag to use cirrus band */
    unsigned char *pixel_mask,  /* I/O: Pixel mask */
    unsigned char *conf_mask    /* O: Pointer for cloud confidence level */
)
{
    int npixels = input->num_lines * input->num_samps; /* Number of pixels */
    int index;
    float *water_prob_array = NULL; /* Final water pixel probabilty value */
    float *land_prob_array = NULL;  /* Final land pixel probability value */
    float clr_mask_threshold = 0.0; /* Clear sky pixel threshold */
    float wclr_mask_threshold = 0.0;/* Water pixel threshold */
    float low_temp = low_temp_threshold + TEMPERATURE_BUFFER - 3500;
    float land_prob_max = -10000000.0;
    float land_prob_min = 10000000.0;
    float water_prob_max = -10000000.0;
    float water_prob_min = 10000000.0;
    float temperature_percent_diff = high_temp_threshold - low_temp_threshold;

    short int *blue_buf = input->buf[CF_BLUE_BAND];
    short int *green_buf = input->buf[CF_GREEN_BAND];
    short int *red_buf = input->buf[CF_RED_BAND];
    short int *nir_buf = input->buf[CF_NIR_BAND];
    short int *swir1_buf = input->buf[CF_SWIR1_BAND];
    short int *thermal_buf = input->therm_buf;
    short int *cirrus_buf = input->buf[CF_CIRRUS_BAND];

    IAS_LOG_INFO("Starting phase2");

    /* Fill conf_mask buffer with zero */
    memset(conf_mask, 0, npixels * sizeof(unsigned char));

    land_prob_array = malloc(npixels * sizeof(*land_prob_array));
    water_prob_array = malloc(npixels * sizeof(*water_prob_array));
    if (land_prob_array == NULL || water_prob_array == NULL)
    {
        IAS_LOG_ERROR("Error allocating buffers");
        free(land_prob_array);
        free(water_prob_array);
        return ERROR;
    }

    for (index = 0; index < npixels; index++)
    {
        short int swir1 = swir1_buf[index];
        float water_prob;
        float land_prob;


        if (pixel_mask[index] & CF_FILL_BIT)
        {
            land_prob = 0.0;
            water_prob = 0.0;
        }
        else if (pixel_mask[index] & CF_WATER_BIT)
        {
            const int t_bright = 1100;/* Brightness test value for water */
            float brightness_prob;    /* Brightness probability value */

            /* Brightness test (over water) */
            brightness_prob = (float)swir1/(float)t_bright;

            if (brightness_prob > 1.0)
                brightness_prob = 1.0;
            if (brightness_prob < 0.0)
                brightness_prob = 0.0;

            if (use_thermal)
            {
                float wtemp_prob; /* Water temperature probability value */

                /* Get cloud probability over water */
                /* Temperature test over water */
                wtemp_prob = (t_wtemp - (float)thermal_buf[index])/400.0;
                if (wtemp_prob < 0.0)
                    wtemp_prob = 0.0;
                brightness_prob *= wtemp_prob;
            }

            if (use_cirrus)
            {
                water_prob = 100.0 * (brightness_prob +
                    (float)cirrus_buf[index])/400.0;
            }
            else
            {
                water_prob = 100.0 * brightness_prob;
            }

            land_prob = 0.0;
        }
        else
        {
            float temp_prob;       /* Temperature probability */
            float ndvi;            /* Vegetation index */
            float ndsi;            /* Snow index */
            float visi_mean;       /* Visible band average */
            float whiteness;       /* Whiteness score */
            float max_value;       /* Maximum of ndvi, ndsi, whiteness */
            float vari_prob;
            short int blue = blue_buf[index];
            short int green = green_buf[index];
            short int red = red_buf[index];
            short int nir = nir_buf[index];

            if (red + nir != 0)
                ndvi = (float)(nir - red)/(nir + red);
            else
                ndvi = 0.01;

            if (green + swir1 != 0)
                ndsi = (float)(green - swir1)/(green + swir1);
            else
                ndsi = 0.01;

            /* NDVI and NDSI should not be negative */
            if (ndsi < 0.0)
                ndsi = 0.0;
            if (ndvi < 0.0)
                ndvi = 0.0;

            visi_mean = (blue + green + red)/3.0;
            if (visi_mean != 0)
            {
                whiteness = ((fabs((float)blue - visi_mean)
                          + fabs((float)green - visi_mean)
                          + fabs((float)red - visi_mean))) / visi_mean;
            }
            else
                whiteness = 0.0;

#ifdef HANDLE_SATURATION
            /* Conditionally include this code for sensors that have
               saturation */
            /* If one visible band is saturated, whiteness = 0 */
            if (blue >= input->saturated_value_max[0] - 1 ||
                green >= input->saturated_value_max[1] - 1 ||
                red >= input->saturated_value_max[2] - 1)
            {
                whiteness = 0.0;
            }
#endif

            /* Vari_prob=1-max(max(abs(NDSI),abs(NDVI)),whiteness); */
            if (ndsi > ndvi)
                max_value = ndsi;
            else
                max_value = ndvi;

            if (whiteness > max_value)
                max_value = whiteness;
            vari_prob = 1.0 - max_value;

            if (use_thermal)
            {
                temp_prob = (high_temp_threshold - (float)thermal_buf[index])
                    / temperature_percent_diff;

                /* Temperature can have prob > 1 */
                if (temp_prob < 0.0)
                    temp_prob = 0.0;

                vari_prob *= temp_prob;
            }

            /* Final probability mask (land) */
            if (use_cirrus)
            {
                land_prob = 100.0 * (vari_prob
                    + ((float)cirrus_buf[index]/400.0));
            }
            else
            {
                land_prob = 100.0 * vari_prob;
            }

            water_prob = 0.0;
        }

        if (land_prob > land_prob_max)
            land_prob_max = land_prob;
        if (land_prob < land_prob_min)
            land_prob_min = land_prob;
        if (water_prob > water_prob_max)
            water_prob_max = water_prob;
        if (water_prob < water_prob_min)
            water_prob_min = water_prob;

        land_prob_array[index] = land_prob;
        water_prob_array[index] = water_prob;

    } /* Pixel loop */

    /* Dynamic threshold for land */
    if (find_masked_percentile_threshold_float(land_prob_array,
        npixels, clear_mask, clear_land_mask, land_prob_min,
        land_prob_max, 100.0 * high_threshold, &clr_mask_threshold)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Error calculating the land threshold");
        free(land_prob_array);
        free(water_prob_array);
        return ERROR;
    }

    clr_mask_threshold += cloud_prob_threshold;

    /* Dynamic threshold for water */
    if (find_masked_percentile_threshold_float(water_prob_array,
        npixels, clear_mask, clear_water_mask, water_prob_min,
        water_prob_max, 100.0 * high_threshold, &wclr_mask_threshold)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Error calculating the water threshold");
        free(land_prob_array);
        free(water_prob_array);
        return ERROR;
    }

    wclr_mask_threshold += cloud_prob_threshold;

    /* Loop through each line in the image */
    for (index = 0; index < npixels; index++)
    {
        unsigned char pm = pixel_mask[index];

        if (use_thermal && thermal_buf[index] < low_temp)
        {
            /* The temperature of the cloud is lower than the dynamic
               cloud temperature */
            conf_mask[index] = CF_CLOUD_HIGH;

            /* Set the cloud bit */
            pixel_mask[index] = pm | CF_CLOUD_BIT;
        }
        else if (pm & CF_CLOUD_BIT)
        {
            if (((!(pm & CF_WATER_BIT)
                    && land_prob_array[index] > clr_mask_threshold)
                    || ((pm & CF_WATER_BIT)
                    && water_prob_array[index] > wclr_mask_threshold)))
            {
                /* The dynamic cloud threshold for land pixels exceeded the
                   land threshold, or the dynamic cloud threshold for
                   water pixels exceeded the water threshold, so the
                   confidence level is high */
                conf_mask[index] = CF_CLOUD_HIGH;

                /* The cloud bit is already set */
            }
            else if (((!(pm & CF_WATER_BIT)
                    && land_prob_array[index] > clr_mask_threshold - 10)
                    || ((pm & CF_WATER_BIT)
                    && water_prob_array[index] > wclr_mask_threshold - 10)))
            {
                /* The dynamic cloud threshold for land pixels exceeded the
                   (land threshold - 10), or the dynamic cloud threshold for
                   water pixel exceeded the (water threshold - 10), so the
                   confidence level is medium */
                conf_mask[index] = CF_CLOUD_MED;

                /* Unset the cloud bit */
                pixel_mask[index] = pm & ~CF_CLOUD_BIT;
            }
            else
            {
                /* The cloud confidence is in low range */
                conf_mask[index] = CF_CLOUD_LOW;

                /* Unset the cloud bit */
                pixel_mask[index] = pm & ~CF_CLOUD_BIT;
            }
        }
        else
        {
            /* Here the cloud confidence is in low range */
            conf_mask[index] = CF_CLOUD_LOW;

            /* The cloud bit is already unset */
        }
    }

    /* Free the memory */
    free(land_prob_array);
    free(water_prob_array);

    IAS_LOG_INFO("Completed phase2");

    return SUCCESS;
}

/*****************************************************************************
Name: identify_potential_cloud_shadows

Purpose: Identify the pixels that are potentially cloud shadows.

Returns: SUCCESS/ERROR

******************************************************************************/
static int identify_potential_cloud_shadows
(
    const CFMASK_INPUTS *input, /* I: Input structure */
    float low_threshold,        /* I: The low threshold */
    const unsigned char *clear_mask, /* I: Mask for clear pixels */
    unsigned char clear_land_mask,   /* I: Mask for clear land */
    unsigned char clear_water_mask,  /* I: Mask for clear water */
    unsigned char *pixel_mask   /* I/O: Pixel mask */
)
{
    int npixels = input->num_lines * input->num_samps; /* Number of pixels */
    int index;
    int nrows = input->num_lines; /* Number of rows */
    int ncols = input->num_samps; /* Number of columns */
    short int *filled_nir = NULL;  /* local minima filled NIR band */
    short int *filled_swir = NULL; /* local minima filled SWIR band */
    short int *nir_buf = input->buf[CF_NIR_BAND];
    short int nir_max = input->max[CF_NIR_BAND];
    short int nir_min = input->min[CF_NIR_BAND];
    short int *swir_buf = input->buf[CF_SWIR1_BAND];
    short int swir_max = input->max[CF_SWIR1_BAND];
    short int swir_min = input->min[CF_SWIR1_BAND];
    float nir_background;  /* NIR band background value */
    float swir1_background;/* SWIR1 band background value */
    int error = 0;         /* Error flag for the parallel section */

    IAS_LOG_INFO("Starting identify_potential_cloud_shadows");

    /* Estimate the background (land) NIR Ref */
    if (find_masked_percentile_threshold_short(nir_buf, npixels, clear_mask,
            clear_land_mask, nir_min, nir_max, 100.0 * low_threshold,
            &nir_background) != SUCCESS)
    {
        IAS_LOG_ERROR ("Error calculating NIR background threshold");
        return ERROR;
    }

    /* Estimate the background (land) SWIR Ref */
    if (find_masked_percentile_threshold_short(swir_buf, npixels, clear_mask,
            clear_land_mask, swir_min, swir_max, 100.0 * low_threshold,
            &swir1_background) != SUCCESS)
    {
        IAS_LOG_ERROR ("Error calculating SWIR1 background threshold");
        return ERROR;
    }

    /* Allocate memory for the local minima filled images */
    filled_nir = malloc(npixels * sizeof(*filled_nir));
    filled_swir = malloc(npixels * sizeof(*filled_swir));
    if (filled_nir == NULL || filled_swir == NULL)
    {
        IAS_LOG_ERROR("Allocating minima buffers");
        free(filled_nir);
        free(filled_swir);
        return ERROR;
    }

    /* Timing message */
    IAS_LOG_INFO("Starting minima fills");

    /* Perform the two local minima fills in parallel */
    #pragma omp parallel sections
    {
        {
            /* Create a version of band 4 with the local minima filled */
            if (fill_local_minima_in_image("NIR", nir_buf, nrows, ncols,
                nir_background, filled_nir) != SUCCESS)
            {
                IAS_LOG_ERROR ("Running fill_local_minima_in_image routine on "
                    "NIR band");
                error = 1;
            }
        }

        #pragma omp section
        {
            /* Create a version of band 5 with the local minima filled */
            if (fill_local_minima_in_image("SWIR1", swir_buf, nrows, ncols,
                swir1_background, filled_swir) != SUCCESS)
            {
                IAS_LOG_ERROR ("Running fill_local_minima_in_image routine on "
                    "SWIR1 band");
                error = 1;
            }
        }
    }

    /* If an error occurred finding the local minima, clean up and return an
       error */
    if (error)
    {
        free(filled_nir);
        free(filled_swir);
        return ERROR;
    }

    /* Timing message */
    IAS_LOG_INFO("Done with minima fills");

    /* Using the local minima NIR and SWIR images, identify the probable
       cloud shadows */
    for (index = 0; index < npixels; index++)
    {
        unsigned char pm = pixel_mask[index];

        /* Process non-fill pixels only */
        if (pm & CF_FILL_BIT)
        {
            /* Clear any other bits that are set */
            pm &= CF_FILL_BIT;
        }
        else
        {
            short int shadow_prob;   /* Shadow probability */
            short int nir_diff = filled_nir[index] - nir_buf[index];
            short int swir_diff = filled_swir[index] - swir_buf[index];

            if (nir_diff < swir_diff)
                shadow_prob = nir_diff;
            else
                shadow_prob = swir_diff;

            if (shadow_prob > 200)
                pm |= CF_SHADOW_BIT;

            /* Refine water mask (no confusion water/cloud) */
            if ((pm & CF_WATER_BIT) && (pm & CF_CLOUD_BIT))
                pm &= ~CF_WATER_BIT;
        }
        pixel_mask[index] = pm;
    }

    /* Release the memory */
    free(filled_nir);
    free(filled_swir);

    IAS_LOG_INFO("Completed identify_potential_cloud_shadows");

    return SUCCESS;
}

/*****************************************************************************
NAME:  cfmask_potential_cloud_shadow_snow_mask

PURPOSE: Identify the cloud pixels, snow pixels, water pixels, clear land 
         pixels, and potential shadow pixels

RETURN: SUCCESS/ERROR

*****************************************************************************/
int cfmask_potential_cloud_shadow_snow_mask
(
    const CFMASK_INPUTS *input,   /* I: Input structure */
    float cloud_prob_threshold,   /* I: Cloud probability threshold */
    int use_thermal,              /* I: Flag to use thermal band */
    int use_cirrus,               /* I: Flag to use cirrus band */
    float *clear_sky_threshold,   /* O: Percent of clear-sky pixels */
    float *low_temp_threshold,    /* O: Percentile of low background temp */
    float *high_temp_threshold,   /* O: Percentile of high background temp */
    unsigned char *pixel_mask,    /* I/O: Pixel mask */
    unsigned char *conf_mask      /* O: Pointer for cloud confidence level */
)
{
    int npixels = input->num_lines * input->num_samps; /* Number of pixels */
    float clear_land_percent;     /* Clear land pixel percentage */
    float clear_water_percent;    /* Clear water pixel percentage */
    float l_pt;                   /* Low percentile threshold */
    float h_pt;                   /* High percentile threshold */
    unsigned char *clear_mask = NULL;
    unsigned char clear_land_mask; /* Mask for clear land */
    unsigned char clear_water_mask;/* Mask for clear water */

    /* Log message to help understand where the application is spending its
       time */
    IAS_LOG_INFO("Starting cfmask potential");

    clear_mask = malloc(npixels * sizeof(*clear_mask));
    if (clear_mask == NULL)
    {
        IAS_LOG_ERROR("Allocating clear_mask memory");
        return ERROR;
    }

    /* Perform the first phase of cloud identification */
    phase1(input, use_thermal, use_cirrus, pixel_mask, clear_mask,
         clear_sky_threshold, &clear_land_percent, &clear_water_percent);

    /* If the scene is 99.9% or more clouds, just call the whole scene cloudy */
    if (*clear_sky_threshold <= 0.1)
    {
        int index;

        /* Log a message to indicate that the entire image is considered
           cloudy */
        IAS_LOG_INFO("Less than 0.1 percent clear sky, considering the entire "
            "image to be cloud or cloud shadow");

        /* Almost no clear sky, so consider entire image either clouds or
           cloud shadows */
        for (index = 0; index < npixels; index++)
        {
            /* All cloud or cloud shadows. If cloud, mark cloud confidence to
               high */
            if (pixel_mask[index] & CF_CLOUD_BIT)
            {
                conf_mask[index] = CF_CLOUD_HIGH;
            }
            else
            {
                pixel_mask[index] |= CF_SHADOW_BIT;
                conf_mask[index] = CF_CLOUD_LOW;
            }
        }

        if (use_thermal)
        {
            /* No thermal test is needed, all clouds */
            *low_temp_threshold = -1.0;
            *high_temp_threshold = -1.0;
        }
    }
    else
    {
        float t_wtemp;           /* High percentile water temperature */

        /* Temperature for snow test */
        l_pt = 0.175;
        h_pt = 1.0 - l_pt;

        /* Use the clear land pixels or clear water pixels if there are enough
           present (at least 0.1%), otherwise use any clear pixels */
        if (clear_land_percent >= 0.1)
            clear_land_mask = CF_CLEAR_LAND_BIT;
        else
        {
            clear_land_mask = CF_CLEAR_BIT;
            IAS_LOG_INFO("Clear land falling back to using all clear pixel");
        }
        if (clear_water_percent >= 0.1)
            clear_water_mask = CF_CLEAR_WATER_BIT;
        else
        {
            clear_water_mask = CF_CLEAR_BIT;
            IAS_LOG_INFO(
                "Clear water falling back to using all clear pixel");
        }

        /* Find the temperature thresholds */
        if (use_thermal)
        {
            if (find_temperature_thresholds(input, clear_land_mask,
                clear_water_mask, l_pt, h_pt, clear_mask, low_temp_threshold,
                high_temp_threshold, &t_wtemp) != SUCCESS)
            {
                IAS_LOG_ERROR("Failed identifying the temperature thresholds");
                free(clear_mask);
                return ERROR;
            }

            /* Expand temperature thresholds by the buffering amount */
            *low_temp_threshold -= TEMPERATURE_BUFFER;
            *high_temp_threshold += TEMPERATURE_BUFFER;
        }
        else
        {
            /* Make sure these are initialized in the non-thermal case */
            *low_temp_threshold = -1.0;
            *high_temp_threshold = -1.0;
        }

        if (phase2(input, t_wtemp, cloud_prob_threshold, *low_temp_threshold,
            *high_temp_threshold, h_pt, clear_land_mask, clear_water_mask,
            clear_mask, use_thermal, use_cirrus, pixel_mask, conf_mask)
            != SUCCESS)
        {
            IAS_LOG_ERROR("Error running phase 2");
            free(clear_mask);
            return ERROR;
        }

        /* Identify the pixels that are potentially shadows */
        if (identify_potential_cloud_shadows(input, l_pt, clear_mask,
            clear_land_mask, clear_water_mask, pixel_mask) != SUCCESS)
        {
            IAS_LOG_ERROR("Error identifying potential cloud shadows");
            free(clear_mask);
            return ERROR;
        }
    }

    free(clear_mask);

    /* Message to help understand the timing */
    IAS_LOG_INFO("Completed cfmask potential");

    return SUCCESS;
}
