/* System Includes */
#include <math.h>
#include <stdlib.h>
#include <limits.h>

/* IAS Library Includes */
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_math.h"

/* Local Includes */
#include "cfmask_support.h"

/* FIXME To get closer to matching the broken LSRD scaling define these.
   This should be removed removed before this is completed. */
/*#define USE_LSRD_SCALING_REFL 1*/
/*#define USE_LSRD_SCALING_TEMP 1*/

/******************************************************************************
NAME: find_percentile_threshold_short

PURPOSE: Calculate and return the data value where cumulative histogram
         exceeds the percentage threshold.

RETURN: SUCCESS/ERROR

******************************************************************************/
int find_percentile_threshold_short
(
    const short int *array,/* I: Input data pointer */
    int array_size,      /* I: Number of input data array */
    short int min,       /* I: Minimum value in the input data array */
    short int max,       /* I: Maximum value in the input data array  */
    float prct,          /* I: Percentage threshold */
    float *result        /* O: Percentile calculated */
)
{
    int *histogram;    /* Array to store data in an histogram */
    int i;             /* Loop variable */
    int bins;          /* Number of bins in the histogram */
    int sum = 0;       /* Cumulative sum for data store in array */
    int target_sum;    /* Target cumulative sum */

    /* Just return 0.0 if no input value */
    if (array_size == 0)
    {
        *result = 0.0;
        return SUCCESS;
    }

    /* Initialize the result in case the loop doesn't assign a value */
    *result = max;

    bins = max - min + 1;

    histogram = calloc(bins, sizeof(int));
    if (histogram == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for histogram array (%lu bytes)",
            bins * sizeof(int)); 
        return ERROR;
    }

    /* Accumulate the histogram */
    for (i = 0; i < array_size; i++)
    {
        histogram[array[i]-min]++;
    }

    target_sum = (prct / 100) * array_size;

    for (i = 0; i < bins; i++)
    {
         sum += histogram[i];
         if (sum > target_sum)
         {
             *result = (float)(min + i);
             break;
         }
    }
    free(histogram);

    return SUCCESS;
}

/******************************************************************************
NAME: find_percentile_threshold_float

PURPOSE: Calculate and return the float input data value where cumulative
         histogram exceeds the percentage threshold.

RETURN: SUCCESS/ERROR

******************************************************************************/
int find_percentile_threshold_float
(
    const float *array,/* I: Input data pointer */
    int array_size,   /* I: Number of input data array */
    float min,        /* I: Minimum value in the input data array */
    float max,        /* I: Maximum value in the input data array  */
    float prct,       /* I: Percentage threshold */
    float *result     /* O: Percentile calculated */
)
{
    int *histogram;   /* Array to store data in an histogram */
    int i;            /* Loop variables */
    int start, end;   /* Start/end variables */
    int bins;         /* Number of bins in the histogram */
    int sum = 0;      /* Cumulative sum for data store in array */
    int target_sum;   /* Target cumulative sum */

    /* Just return 0.0 if no input value */
    if (array_size == 0)
    {
        *result = 0.0;
        return SUCCESS;
    }

    /* Initialize the result in case the loop doesn't assign a value */
    *result = max;

    start = (int)rintf(min);
    end = (int)rintf(max);

    bins = end - start + 2;

    histogram = calloc(bins, sizeof(int));
    if (histogram == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for histogram array (%lu bytes)",
            bins * sizeof(int)); 
        return ERROR;
    }

    /* Accumulate the histogram */
    for (i = 0; i < array_size; i++)
    {
        histogram[(int) rintf(array[i])-start]++;
    }

    target_sum = (prct / 100) * array_size;

    for (i = 0; i < bins; i++)
    {
         sum += histogram[i];
         if (sum > target_sum)
         {
             *result = (float)(start + i);
             break;
         }
    }
    free(histogram);

    return SUCCESS;
}

/******************************************************************************
NAME: find_masked_percentile_threshold_short

PURPOSE: Calculate and return the data value where cumulative histogram
         exceeds the percentage threshold.  The input array values are included
         in the calculations when the mask array has the requested mask bit
         set.

RETURN: SUCCESS/ERROR

******************************************************************************/
int find_masked_percentile_threshold_short
(
    const short int *array, /* I: Input data pointer */
    int array_size,         /* I: Number of input data array */
    const unsigned char *mask_array, /* I: array of bit masks */
    unsigned char use_mask, /* I: Bit mask to include in the calculations */
    short int min,       /* I: Minimum value in the input data array */
    short int max,       /* I: Maximum value in the input data array  */
    float prct,          /* I: Percentage threshold */
    float *result        /* O: Percentile calculated */
)
{
    int *histogram;    /* Array to store data in a histogram */
    int i;             /* Loop variables */
    int bins;          /* Number of bins in the histogram */
    int total;
    int sum = 0;       /* Cumulative sum for data store in array */
    int target_sum;    /* Target cumulative sum */

    /* Just return 0.0 if no input value */
    if (array_size == 0)
    {
        *result = 0.0;
        return SUCCESS;
    }

    /* Initialize the result in case the loop doesn't assign a value */
    *result = max;

    bins = max - min + 1;

    histogram = calloc(bins, sizeof(int));
    if (histogram == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for histogram array (%lu bytes)",
            bins * sizeof(int)); 
        return ERROR;
    }

    /* Calculate the histogram of the values that match the mask */
    total = 0;
    for (i = 0; i < array_size; i++)
    {
        if (mask_array[i] & use_mask)
        {
            histogram[array[i] - min]++;
            total++;
        }
    }

    /* If no values were found, just return 0.0 */
    if (total == 0)
    {
        *result = 0.0;
        return SUCCESS;
    }
    
    target_sum = (prct / 100) * total;

    for (i = 0; i < bins; i++)
    {
         sum += histogram[i];
         if (sum > target_sum)
         {
             *result = (float)(min + i);
             break;
         }
    }
    free(histogram);

    return SUCCESS;
}
/******************************************************************************
NAME: find_masked_percentile_threshold_float

PURPOSE: Calculate and return the float input data value where cumulative
         histogram exceeds the percentage threshold.  The input array values
         are included in the calculations when the mask array has the requested
         mask bit set.

RETURN: SUCCESS/ERROR

******************************************************************************/
int find_masked_percentile_threshold_float
(
    const float *array, /* I: Input data pointer */
    int array_size,     /* I: Number of input data array */
    const unsigned char *mask_array, /* I: array of bit masks */
    unsigned char use_mask, /* I: Bit mask to include in the calculations */
    float min,          /* I: Minimum value in the input data array */
    float max,          /* I: Maximum value in the input data array  */
    float prct,         /* I: Percentage threshold */
    float *result       /* O: Percentile calculated */
)
{
    int *histogram;   /* Array to store data in an histogram */
    int i;            /* Loop variable */
    int start;        /* Starting value for the histogram */
    int end;          /* Ending value for the histogram */
    int bins;         /* Number of bins in the histogram */
    int total;
    int sum = 0;      /* Cumulative sum for data store in array */
    int target_sum;   /* Target cumulative sum */

    /* Just return 0 if no input value */
    if (array_size == 0)
    {
        *result = 0.0;
        return SUCCESS;
    }

    /* Initialize the result in case the loop doesn't assign a value */
    *result = max;

    start = (int)rintf(min);
    end = (int)rintf(max);

    bins = end - start + 2;

    histogram = calloc(bins, sizeof(int));
    if (histogram == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for histogram array (%lu bytes)",
            bins * sizeof(int)); 
        return ERROR;
    }

    total = 0;
    for (i = 0; i < array_size; i++)
    {
        if (mask_array[i] & use_mask)
        {
            histogram[(int) rintf(array[i])-start]++;
            total++;
        }
    }
    /* If no values were found, return 0 */
    if (total == 0)
    {
        *result = 0.0;
        return SUCCESS;
    }

    target_sum = (prct / 100) * total;

    for (i = 0; i < bins; i++)
    {
         sum += histogram[i];
         if (sum > target_sum)
         {
             *result = (float)(start + i);
             break;
         }
    }
    free(histogram);

    return SUCCESS;
}

/*****************************************************************************
NAME: dn_to_toa
 
PURPOSE: Convert the input DN values to TOA reflectance and scaled it 
         by 10000.0 for band 2 to 7.
          
RETURNS: Nothing
              
*****************************************************************************/
void dn_to_toa
(
    const unsigned short image_values[],/* I: DN values from the input image
                                              per-band */
    int nsamples,                 /* I: Samples in image */
    int nlines,                   /* I: Lines in image */
    int max_threads,              /* I: Maximum number of threads to use */
    double solar_elevation,       /* I: Solar elevation angle in degrees */ 
    double reflectance_scaling_offset, /* I: Offset to convert to reflectance */
    double reflectance_scaling_gain,   /* I: Gain to convert to reflectance */
    short *min_scaled_value,           /* O: Minimum scaled value */
    short *max_scaled_value,           /* O: Minimum scaled value */
    short scaled_values[]              /* O: Scaled TOA values for cfmask */
) 
{
    int pix;     /* Loop counter for total samples per band */
    int min_value = SHRT_MAX;
    int max_value = SHRT_MIN;

    int total_samples = nlines * nsamples;
    double solar_zenith = (90 - solar_elevation) * 
        ias_math_get_radians_per_degree();  
    double inv_cosine_solar_zenith = 1.0/cos(solar_zenith); 
    int message_output = 0;
    int range_exceeded_count = 0;

    /* Use OpenMP to parallelize the scaling.  Since the reduction operation
       in the available version of OpenMP doesn't support min/max reduction,
       the reduction is done manually. */
    #pragma omp parallel num_threads(max_threads)
    {
        /* Declare thread local min/max values */
        short lmin_value = SHRT_MAX;
        short lmax_value = SHRT_MIN;
        
        /* Perform the scaling in parallel, reducing the range_exceeded_count
           from all the threads at the end */
        #pragma omp for private(pix) nowait reduction(+:range_exceeded_count)
        for (pix = 0; pix < total_samples; pix++)
        {
            if (image_values[pix] == 0)
            {
                scaled_values[pix] = FILL_PIXEL;
            }
            else
            {
                int value;
                double temp; /* Temporary variable for calculation */

                temp = reflectance_scaling_gain * 
                    image_values[pix] + reflectance_scaling_offset;
                value = (int)(SCALING_FACTOR_DN_TOA * temp *
                    inv_cosine_solar_zenith
#ifndef USE_LSRD_SCALING_REFL
/* LSRD isn't rounding */
                    + 0.5
#endif
                    );

                if (value > SHRT_MAX || value < SHRT_MIN)
                {
                    /* Make sure other threads see this to limit the times
                       the message is output (still a minor race condition,
                       but it isn't likely to happen) */
                    #pragma omp flush(message_output)
                    if (!message_output)
                    {
                        IAS_LOG_WARNING("Image value (%d) scaled outside valid "
                            "range, limiting to valid range (only reporting"
                            " first occurrence", value);
                        message_output = 1;
                        #pragma omp flush(message_output)
                    }
                    if (value > SHRT_MAX)
                        value = SHRT_MAX;
                    else if (value < SHRT_MIN)
                        value = SHRT_MIN;
                    range_exceeded_count++;
                }    
                scaled_values[pix] = (short)value;

                if (value < lmin_value)
                    lmin_value = value;
                if (value > lmax_value)
                    lmax_value = value;
            }
        }
        #pragma omp critical
        {
            /* Reduce the min/max values */
            if (lmin_value < min_value)
                min_value = lmin_value;
            if (lmax_value > max_value)
                max_value = lmax_value;
        }
    }

    /* If any pixels exceeded the range, issue a message with the count */
    if (range_exceeded_count > 0)
    {
        IAS_LOG_WARNING("Total of %d pixels exceeded range",
            range_exceeded_count);
    }

    *min_scaled_value = (short)min_value;
    *max_scaled_value = (short)max_value;
}

/*****************************************************************************
NAME: dn_to_bt

PURPOSE: Convert the input DN values brightness temperature and scaled it
         by 100.0 for band 10.

RETURNS: Nothing

*****************************************************************************/
void dn_to_bt
(
    const unsigned short image_values[],/* I: DN values from the input image
                                              per-band */
    int nsamples,                  /* I: Samples in image */
    int nlines,                    /* I: Lines in image */
    int max_threads,               /* I: Maximum number of threads to use */
    double k1_constant,            /* I: TIRS thermal constant k1 */
    double k2_constant,            /* I: TIRS thermal constant k2 */
    double spectral_radiance_scaling_offset,/* I: Offset to convert to
                                                  radiance */
    double spectral_radiance_scaling_gain,  /* I: Gain to convert to radiance */
    short *min_scaled_value,        /* O: Minimum scaled value */
    short *max_scaled_value,        /* O: Minimum scaled value */
    short scaled_values[]           /* O: Scaled BT values for cfmask */
)
{
    int pix;
    int total_samples = nlines * nsamples;
    short min_value = SHRT_MAX;
    short max_value = SHRT_MIN;

    /* Use OpenMP to parallelize the scaling.  Since the reduction operation
       in the available version of OpenMP doesn't support min/max, the
       reduction is done manually. */
    #pragma omp parallel num_threads(max_threads)
    {
        /* Declare thread local min/max values */
        short lmin_value = SHRT_MAX;
        short lmax_value = SHRT_MIN;

        #pragma omp for private(pix) nowait
        for (pix = 0; pix < total_samples; pix++)
        {
            if (image_values[pix] == 0)
                scaled_values[pix] = FILL_PIXEL;
            else
            {
                short value;
                double temp;

                temp = spectral_radiance_scaling_gain *
                     image_values[pix] + spectral_radiance_scaling_offset;
                temp = k2_constant / (log ((k1_constant / temp) + 1.0));

#ifdef USE_LSRD_SCALING_TEMP
/* Scale like LSRD - It makes a TOA that is scaled by 10, but then their cfmask
   code reads it in, scales it back to float then by 100, making up precision
   that was lost (not that it is probably "real" precision). */
                temp = ((int)(temp * 10 + 0.5)) * 0.1;
#endif

                /* Convert the temperature from Kelvin to Celsius, then scale
                   to a fixed point representation stored in a short int */
                value = (short) (SCALING_FACTOR_DN_BT * (temp - 273.15) + 0.5);
                scaled_values[pix] = value;

                if (value < lmin_value)
                    lmin_value = value;
                if (value > lmax_value)
                    lmax_value = value;
            }
        }
        #pragma omp critical
        {
            /* Reduce the min/max values */
            if (lmin_value < min_value)
                min_value = lmin_value;
            if (lmax_value > max_value)
                max_value = lmax_value;
        }
    }

    *min_scaled_value = (short)min_value;
    *max_scaled_value = (short)max_value;
}
