#ifndef CFMASK_SUPPORT_H
#define CFMASK_SUPPORT_H

/* IAS Library Includes */
#include "ias_types.h"
#include "ias_angle_gen.h"

#define FILL_PIXEL -9999

/* Scaling factor for cfmask */
#define SCALING_FACTOR_DN_TOA 10000
#define SCALING_FACTOR_DN_BT 100

/* Classification for cfmask algorithm */
#define CF_WATER_BIT 0x01
#define CF_SHADOW_BIT 0x02
#define CF_SNOW_BIT 0x04
#define CF_CLOUD_BIT 0x08
#define CF_FILL_BIT 0x10

/* Clear mask bit definitions */
#define CF_CLEAR_BIT 0x01
#define CF_CLEAR_WATER_BIT 0x02
#define CF_CLEAR_LAND_BIT 0x04

/* Cfmask integer classification values */
#define CF_CLEAR_PIXEL 0
#define CF_WATER_PIXEL 1
#define CF_CLOUD_SHADOW_PIXEL 2
#define CF_SNOW_PIXEL 3
#define CF_CLOUD_PIXEL 4
#define CF_FILL_PIXEL 255

/* Cfmask cloud confidence level definition */
#define CF_CLOUD_LOW 0x20
#define CF_CLOUD_MED 0x40
#define CF_CLOUD_HIGH 0x60

/* Indices to the buffers in the input array */
#define CF_BLUE_BAND 0
#define CF_GREEN_BAND 1
#define CF_RED_BAND 2
#define CF_NIR_BAND 3
#define CF_SWIR1_BAND 4
#define CF_SWIR2_BAND 5
#define CF_CIRRUS_BAND 6
#define NBAND_REFL_MAX 7

/* Structure to hold information about the input bands for Cfmask */
typedef struct cfmask_inputs
{
    int nband;                  /* Number of input image bands */
    int num_lines;              /* Number of lines */
    int num_samps;              /* Number of samples */
    short *buf[NBAND_REFL_MAX]; /* Input data buffer for bands 2 to 7 */
    short min[NBAND_REFL_MAX];  /* Minimum non-fill values in each band */
    short max[NBAND_REFL_MAX];  /* Maximum non-fill values in each band */
    short *therm_buf;           /* Input data buffer for band 10 */
    short therm_min;            /* Minimum non-fill thermal value */
    short therm_max;            /* Maximum non-fill thermal value */
} CFMASK_INPUTS;

/* Cfmask to label cloud, snow, water and potential shadow pixel */
int cfmask_potential_cloud_shadow_snow_mask
(
    const CFMASK_INPUTS *input, /* I: Input structure */
    float cloud_prob_threshold, /* I: Cloud probability threshold */
    int use_thermal,            /* I: Flag to use thermal band */
    int use_cirrus,             /* I: Flag to use cirrus band */
    float *clear_sky_threshold, /* O: Percent of clear-sky pixels */
    float *low_temp_threshold,  /* O: Percentile of low background temp */
    float *high_temp_threshold, /* O: Percentile of high background temp */
    unsigned char *pixel_mask,  /* I/O: Pixel mask */
    unsigned char *conf_mask    /* O: Pointer for cloud confidence level */
);

/* Cfmask to segment the cloud pixels and identify the final shadow pixels */
int cfmask_object_cloud_shadow_match
(
    const CFMASK_INPUTS *input, /* I: Input structure */
    const IAS_ANGLE_GEN_METADATA *angle_metadata, /* I: Angle metadata */
    double map_azimuth_angle_radians, /* I: Angle from image north to true 
                                            north */
    double roll_angle_degrees,  /* I: Satellite roll angle */
    double pixel_size_meters,   /* I: Pixel size in meters */
    float ptm,                  /* I: Percent of clear-sky pixels */
    float t_templ,              /* I: Percentile of low background temp */
    float t_temph,              /* I: Percentile of high background temp */
    int cldpix,                 /* I: Cloud buffer size */
    int sdpix,                  /* I: Shadow buffer size */
    int max_threads,            /* I: Maximum worker threads to use */
    int cfmask_dilate_flag,     /* I: Flag to turn on/off cloud dilation */
    unsigned char *pixel_mask,  /* I/O: Pixel mask */
    double sun_elevation_degrees,/* I: Solar elevation angle */
    double sun_azimuth_degrees, /* I: Solar azimuth angle */
    int use_thermal             /* I: Flag to use thermal band */
);

/* Convert DN to TOA reflectance and scale it by 10000.0 for band 2 to 7 */
void dn_to_toa
(
    const unsigned short image_values[],/* I: DN values from the input image
                                              per-band */
    int nsamples,                      /* I: Samples in image */
    int nlines,                        /* I: Lines in image */
    int max_threads,                   /* I: Maximum number of threads to use */
    double solar_elevation,            /* I: Solar elevation angle in degrees */
    double reflectance_scaling_offset, /* I: Offset to convert to reflectance */
    double reflectance_scaling_gain,   /* I: Gain to convert to reflectance */
    short *min_scaled_value,           /* O: Minimum scaled value */
    short *max_scaled_value,           /* O: Minimum scaled value */
    short scaled_values[]              /* O: Scales TOA values for cfmask */
);

/* Convert DN to Brightness Temperature and scale it by 100.0 for band 10 */
void dn_to_bt
(
    const unsigned short image_values[],/* I: DN values from the input image
                                              per-band */
    int nsamples,                  /* I: Samples in image */
    int nlines,                    /* I: Lines in image */
    int max_threads,               /* I: Maximum number of threads to use */
    double k1_constant,            /* I: TIRS thermal constant k1 */
    double k2_constant,            /* I: TIRS thermal constant k2 */
    double spectral_radiance_scaling_offset,/* I: Offset to convert to 
                                                  radaince */
    double spectral_radiance_scaling_gain,  /* I: Gain to convert to radiance */
    short *min_scaled_value,        /* O: Minimum scaled value */
    short *max_scaled_value,        /* O: Minimum scaled value */
    short scaled_values[]           /* O: Scaled BT values for cfmask */
);

/* Calculate percentile of an integer array */
int find_percentile_threshold_short
(
    const short int *array, /* I: Input data pointer */
    int array_size,     /* I: Number of input data array */
    short int min,      /* I: Minimum value in the input data array */
    short int max,      /* I: Maximum value in the input data array  */
    float prct,         /* I: Percentage threshold */
    float *result       /* O: Percentile calculated */
);

/* Calculate percentile of a floating point array */
int find_percentile_threshold_float
(
    const float *array,/* I: Input data pointer */
    int array_size,   /* I: Number of input data array */
    float min,        /* I: Minimum value in the input data array */
    float max,        /* I: Maximum value in the input data array  */
    float prct,       /* I: Percentage threshold */
    float *result     /* O: Percentile calculated */
);

int find_masked_percentile_threshold_short
(
    const short int *array,/* I: Input data pointer */
    int array_size,        /* I: Number of input data array */
    const unsigned char *mask_array, /* I: array of bit masks */
    unsigned char use_mask, /* I: Bit mask to include in the calculations */
    short int min,       /* I: Minimum value in the input data array */
    short int max,       /* I: Maximum value in the input data array  */
    float prct,          /* I: Percentage threshold */
    float *result        /* O: Percentile calculated */
);

int find_masked_percentile_threshold_float
(
    const float *array, /* I: Input data pointer */
    int array_size,     /* I: Number of input data array */
    const unsigned char *mask_array, /* I: array of bit masks */
    unsigned char use_mask, /* I: Bit mask to include in the calculations */
    float min,          /* I: Minimum value in the input data array */
    float max,          /* I: Maximum value in the input data array  */
    float prct,         /* I: Percentage threshold */
    float *result       /* O: Percentile calculated */
);

int fill_local_minima_in_image
(
    const char *band_name,     /* I: Band name being filled */
    const short int *in_img,   /* I: Input image buffer */
    int num_rows,              /* I: Number of rows in the input image */
    int num_cols,              /* I: Number of columns in the input image */
    float boundary_val,        /* I: Background (land) percentage */ 
    short int *out_img         /* O: Output image buffer */
);

#endif
