/* System Includes */
#include <math.h>
#include <stdlib.h>
#include <string.h>

/* COTS Library Includes */
#include "gctp.h"

/* IAS Includes */
#include "ias_satellite_attributes.h"
#include "ias_logging.h"
#include "ias_pixel_mask.h"
#include "ias_radiometry.h"
#include "ias_geo.h"
#include "ias_math.h"

/* Local Includes */
#include "assess_cloud_cover.h"
#include "create_saturation_mask.h"
#include "get_parameters.h"

/* Local Defines */
#define ELEVATION_THRESHOLD .5
#define MINIMUM_ELEVATION_STEP 100.0
#define SATURATED_PIXEL_VALUE 0xFFFF

/* Define a structure to help track which bands have been checked for
   saturation and identify which pixels have saturation */
struct SATURATION_MASK
{
    unsigned char *l1g_mask;/* Bit mask to indicate which pixels in the L1G
                           have saturation.  Each bit in the mask is for one of
                           the bands. */
    unsigned char *l1r_mask;/* Bit mask for the L1R caused saturation */
    int image_lines;    /* Lines in the mask image */
    int image_samples;  /* Samples in the mask image */
    int check_band[IAS_MAX_NBANDS];
                        /* Flag to indicate whether each band index should be
                           checked for saturation (indexed by band index) */
    int l1g_band_done[IAS_MAX_NBANDS];
                        /* Flag to indicate the band has been added to the
                           L1G mask (indexed by band index). */
    int l1g_saturated_count[IAS_MAX_NBANDS];
                        /* Count of saturated pixels found per L1G band */
    unsigned char band_bit_mask[IAS_MAX_NBANDS];
                        /* Bit to use for each of the bands in the mask
                           (indexed by band index) */
};

/****************************************************************************
Name: create_projection_transformation
Purpose: Creates a projection transformation to convert coordinates using the
         input projection.

Returns: A pointer to the created transformation which is NULL if an error
         occurs.

****************************************************************************/
static IAS_GEO_PROJ_TRANSFORMATION *create_projection_transformation
(
    const IAS_PROJECTION *input_projection /* I: Input projection */
)
{
    IAS_PROJECTION geo;
    double proj_parms[IAS_PROJ_PARAM_SIZE];
    int index;

    /* Create the source geographic projection */
    for (index = 0; index < IAS_PROJ_PARAM_SIZE; index++)
        proj_parms[index] = 0.0;

    ias_geo_set_projection(GEO, -1, DEGREE, input_projection->spheroid,
        proj_parms, &geo);

    /* Create the transformation */
    return ias_geo_create_proj_transformation(&geo, input_projection);
}

/****************************************************************************
Name: calculate_pixel_percentage

Purpose: Calculates the target pixel overlap percentage and sets the pixel
         in the mask if percentage >= SATURATED_PIXEL_THRESHOLD parameter.

Returns: void

****************************************************************************/
static void calculate_pixel_percentage
(
    IAS_DBL_LS l1g_ls,              /* I: Input line and sample */
    int number_samples,             /* I: Total number of samples */
    int number_lines,               /* I: Total number of lines */
    IAS_BYTE_TYPE band_mask_value,  /* I: Band mask value */
    IAS_BYTE_TYPE *mask             /* IO: Saturation bit mask */
)
{
    IAS_LNG_LS output_space[4];
    double fill_percentage;
    int count = 0;
    int index;
    int int_line = (int)l1g_ls.line;
    double fract_line = l1g_ls.line - int_line;
    int int_samp = (int)l1g_ls.samp;
    double fract_samp = l1g_ls.samp - int_samp;
    double saturated_pixel_threshold = get_saturated_pixel_threshold();

    /* Check the upper left pixel */
    fill_percentage = (1 - fract_line) * (1 - fract_samp);
    if (fill_percentage >= saturated_pixel_threshold)
    {
        output_space[count].line = int_line;
        output_space[count++].samp = int_samp;
    }

    /* Check the upper right pixel */
    fill_percentage = (1 - fract_line)  * fract_samp;
    if (fill_percentage >= saturated_pixel_threshold)
    {
        output_space[count].line = int_line;
        output_space[count++].samp = int_samp + 1;
    }

    /* Check the lower left pixel */
    fill_percentage = fract_line * (1 - fract_samp);
    if (fill_percentage >= saturated_pixel_threshold)
    {
        output_space[count].line = int_line + 1;
        output_space[count++].samp = int_samp;
    }

    /* Check the lower right pixel */
    fill_percentage = fract_line * fract_samp;
    if (fill_percentage >= saturated_pixel_threshold)
    {
        output_space[count].line = int_line + 1;
        output_space[count++].samp = int_samp + 1;
    }

    /* Loop over each pixel that needs to be set with saturation */
    for (index = 0; index < count; index++)
    {
        int line = output_space[index].line;
        int samp = output_space[index].samp;

        if (line >= 0 && line < number_lines
            && samp >= 0 && samp < number_samples)
        {
            int pixel_index = number_samples * line + samp;

            /* If the bit for this band has already been set then it either has
               overlap or the L1R pixel maps to the same L1G pixel and shouldn't
               be set */
            if ((mask[pixel_index] & band_mask_value) == 0)
            {
                /* Set the mask value */
                mask[pixel_index] |= band_mask_value;
            }
        }
    }
}

/****************************************************************************
Name: convert_pixel_to_line_sample

Purpose: Converts the line/sample combination from L1R space to line/sample
         in the L1G output space.

Returns: SUCCESS / ERROR

****************************************************************************/
static int convert_pixel_to_line_sample
(
    int line,                   /* I: L1R line */
    int sample,                 /* I: L1r sample */
    int band_index,             /* I: Band index */
    int sca_index,              /* I: SCA index */
    double elevation,           /* I: Elevation of line/sample */
    int number_lines,           /* I: Total number of lines */
    int number_samps,           /* I: Total number of samples */
    const IAS_LOS_MODEL *model,       /* I: LOS model */
    const IAS_GEO_PROJ_TRANSFORMATION *geo2l1g, /* I: Geographic
                                                      transformation */
    double r2d,                 /* I: Radians to Degrees constant */
    const IAS_IMAGE *mask_image,      /* I: Saturation mask */
    int *outside_image_flag,    /* O: Flag indicating line/sample falls outside
                                      of the image */
    IAS_DBL_LS *l1g_ls          /* O: Output L1G line/sample */
)
{
    IAS_DBL_XY geo_coords;
    IAS_DBL_LAT_LONG coords;

    /* Set the FLAG */
    *outside_image_flag = FALSE;

    /* Calculate the latitude and longitude using the forward model */
    if (ias_los_model_input_line_samp_to_geodetic(line, sample, band_index,
        sca_index, elevation, model, IAS_EXACT_DETECTOR, NULL,
        &coords.lat, &coords.lng) != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating lat/long for elevation %f, line %d, "
            "sample %d", elevation, line, sample);
        return ERROR;
    }
    coords.lat *= r2d;
    coords.lng *= r2d;

    /* Convert the lat/lng to projection coordinates using a full
       transformation for every coordinate. */
    if (ias_geo_transform_coordinate(geo2l1g, coords.lng, coords.lat,
        &geo_coords.x, &geo_coords.y) != SUCCESS)
    {
        IAS_LOG_ERROR("Transforming latitude %f longitude %f to projection"
            "coordinates", coords.lat, coords.lng);
        return ERROR;
    }

    /* Convert the x/y location to an L1G line/sample */
    l1g_ls->line = (mask_image->corners.upleft.y - geo_coords.y)
        / mask_image->pixel_size_y;
    l1g_ls->samp = (geo_coords.x - mask_image->corners.upleft.x)
        / mask_image->pixel_size_x;

    /* Set line/samp to OUTSIDE_IMAGE if either falls outside the image */
    if (l1g_ls->line < 0 || l1g_ls->line >= number_lines
        || l1g_ls->samp < 0 || l1g_ls->samp >= number_samps)
    {
        *outside_image_flag = TRUE;
    }

    return SUCCESS;
}

/****************************************************************************
Name: update_mask_for_span

Purpose: Determines the L1G line/sample for the pixel span and updates the
         saturation bit mask and counts.

Returns: SUCCESS / ERROR

Note: There may be a case in a high elevation scene where the minimum elevation
      case kicks the pixel out of the image, terminating the analysis of a
      point, but the actual elevation is much higher and falls inside the image.
      This could also happen in the reverse with the maximum elevation for
      off-nadir scenes. These cases were determined as OK and not worth the
      effort to build into the algorithm.

****************************************************************************/
int update_mask_for_span
(
    int band_index,       /* I: Band Index */
    int sca_index,        /* I: SCA Index */
    int number_lines,     /* I: Number of lines in L1G */
    int number_samps,     /* I: Number of samps in L1G */
    double min_elevation, /* I: Minimum elevation in L1G */
    double max_elevation, /* I: Maximum elevation in L1G */
    const IAS_PIXEL_MASK_SPAN *pixel_mask_span, /* I; Pixel mask span*/
    const IAS_GEO_PROJ_TRANSFORMATION *geo2l1g,/* I; Geo to L1G
                                                     transformation */
    const IAS_IMAGE *dem,       /* I: Elevation image */
    const IAS_LOS_MODEL *model, /* I: LOS Model */
    IAS_BYTE_TYPE band_mask_value, /* I: Band mask value */
    double r2d,           /* I: Radians to degrees constant */
    int *min_max_elev_outside_image, /* O: Outside image counter */
    IAS_IMAGE *mask_image /* O: Saturation Mask */
)
{
    int pixel_index;
    int local_outside_image_counter = 0;

    /* Loop over every pixel in the span */
    for (pixel_index = 0; pixel_index < pixel_mask_span->length_of_span;
        pixel_index++)
    {
        IAS_DBL_LS max_l1g_ls;
        IAS_DBL_LS min_l1g_ls;
        IAS_DBL_LS l1g_ls;
        IAS_DBL_XY distance;
        double hypotenuse;
        int step;
        IAS_DBL_LS prev_l1g_ls;
        double prev_found_elev;
        double search_elev = max_elevation;
        int line = pixel_index + pixel_mask_span->starting_pixel_index;
        int sample = pixel_mask_span->detector_index;
        double prev_elev = max_elevation;
        double elev_diff = max_elevation - min_elevation;
        int loop_index;
        int min_outside_image_flag;
        int outside_image_flag;

        /* Find the point using the minimum elevation */
        if (convert_pixel_to_line_sample(line, sample, band_index, sca_index,
            min_elevation, number_lines, number_samps, model, geo2l1g, r2d,
            mask_image, &min_outside_image_flag, &min_l1g_ls) != SUCCESS)
        {
            IAS_LOG_ERROR("Converting L1R line %d and sample %d at elevation "
                "%f to L1G line/sample", line, sample, min_elevation);
            return ERROR;
        }

        /* Find the point using the maximum elevation */
        if (convert_pixel_to_line_sample(line, sample, band_index, sca_index,
            max_elevation, number_lines, number_samps, model, geo2l1g, r2d,
            mask_image, &outside_image_flag, &max_l1g_ls) != SUCCESS)
        {
            IAS_LOG_ERROR("Converting L1R line %d and sample %d at elevation "
                "%f to L1G line/sample", line, sample, max_elevation);
            return ERROR;
        }

        /* Move to the next pixel if either the max or min elevation was
           deemed to be outside of the image */
        if (outside_image_flag || min_outside_image_flag)
        {
            /* Only update the counter if one of the flags has been set (not
               both), due to the possibility of the minimum or maximum elevation
               being outside the image, but the pixel actually residing
               in the image. This can be accounted for in the algorithm
               but the effort is not beneficial at this point */
            if (!outside_image_flag || !min_outside_image_flag)
            {
                local_outside_image_counter++;
            }
            continue;
        }

        prev_l1g_ls = max_l1g_ls;

        /* Find the actual elevation at this point */
        if (ias_misc_read_elevation_at_line_sample(dem, max_l1g_ls.line,
            max_l1g_ls.samp, mask_image->pixel_size_y,
            mask_image->pixel_size_x, &prev_found_elev) != SUCCESS)
        {
            IAS_LOG_ERROR("Getting elevation using max elevation "
                "line %f and sample %f", max_l1g_ls.line, max_l1g_ls.samp);
            return ERROR;
        }

        /* Determine the elevation step */
        distance.y = fabs(max_l1g_ls.line - min_l1g_ls.line);
        distance.x = fabs(max_l1g_ls.samp - min_l1g_ls.samp);
        hypotenuse = sqrt((distance.x * distance.x)
            + (distance.y * distance.y));

        /* Ensure that the minimum/maximum step is followed */
        if (hypotenuse > 0.0)
        {
            step = elev_diff / hypotenuse;
            if (step < MINIMUM_ELEVATION_STEP)
            {
                step = MINIMUM_ELEVATION_STEP;
            }
        }
        else
        {
            step = MINIMUM_ELEVATION_STEP;
        }

        /* Check that the step is less than the elevation change */
        if (step > elev_diff)
        {
            step = elev_diff / 2;

            /* Ensure the step is not less than 1 */
            if (step < 1)
            {
                step = 1;
            }
        }

        /* Calculate the loop starting index and check that the step doesn't
           result in calculating the max/min line sample again */
        loop_index = elev_diff / step;
        if (loop_index <= 0)
        {
            l1g_ls = max_l1g_ls;
            loop_index = -1;
        }

        for (; loop_index >= 0; loop_index--)
        {
            double found_elevation;

            /* Adjust the search elevation */
            search_elev -= step;

            /* Set the min elevation on the last step */
            if (loop_index == 0)
                search_elev = min_elevation;

            /* Find the line/sample using the current elevation guess */
            if (convert_pixel_to_line_sample(line, sample, band_index,
                sca_index, search_elev, number_lines, number_samps, model,
                geo2l1g, r2d, mask_image, &outside_image_flag, &l1g_ls)
                != SUCCESS)
            {
                IAS_LOG_ERROR("Converting L1R line %d and sample %d at "
                    "elevation %f to L1G line/sample", line, sample,
                    search_elev);
                return ERROR;
            }

            /* Grab the actual elevation at the current line/sample */
            if (ias_misc_read_elevation_at_line_sample(dem, l1g_ls.line,
                l1g_ls.samp, mask_image->pixel_size_y,
                mask_image->pixel_size_x, &found_elevation) != SUCCESS)
            {
                IAS_LOG_ERROR("Getting elevation at line %f sample %f",
                    l1g_ls.line, l1g_ls.samp);
                return ERROR;
            }

            /* Using the distance between the last point found, current point,
               and the guessed elevations as point A1 and point A2. Use
               the same distance and the actual elevations as point B1 and B2
               then find the intersection between the 2 lines. */
            if (found_elevation >= search_elev)
            {
                double los_distance;
                double x;
                IAS_LINE_SEGMENT a = {0, prev_elev, 0, search_elev};
                IAS_LINE_SEGMENT b = {0, prev_found_elev, 0,
                                        found_elevation};

                /* Determine the distance between the pixels */
                distance.y = prev_l1g_ls.line - l1g_ls.line;
                distance.x = prev_l1g_ls.samp - l1g_ls.samp;
                los_distance = sqrt(distance.y * distance.y
                    + distance.x * distance.x);

                a.x2 = los_distance;
                b.x2 = los_distance;

                /* Determine their intersection points */
                if (ias_math_find_line_segment_intersection(&a, &b, &x,
                    &found_elevation) != IAS_LINES_INTERSECT)
                {
                    IAS_LOG_ERROR("Lines do not intersect at points"
                        "P1: (%f, %f), (%f, %f) P2: (%f, %f), (%f, %f)",
                        a.x1, a.y1, a.x2, a.y2, b.x1, b.y1, b.x2, b.y2);
                    return ERROR;
                }

                /* Use the elevation to find the final point */
                if (convert_pixel_to_line_sample(line, sample, band_index,
                    sca_index, found_elevation, number_lines, number_samps,
                    model, geo2l1g, r2d, mask_image, &outside_image_flag,
                    &l1g_ls) != SUCCESS)
                {
                    IAS_LOG_ERROR("Converting L1R line %d and sample %d at "
                        "elevation %f to L1G line/sample", line, sample,
                        found_elevation);
                    return ERROR;
                }
                break;
            }

            /* Track the previous elevations and point */
            prev_l1g_ls = l1g_ls;
            prev_elev = search_elev;
            prev_found_elev = found_elevation;
        }

        /* Calculate the pixels to be set in the area */
        calculate_pixel_percentage(l1g_ls, number_samps, number_lines,
            band_mask_value, (IAS_BYTE_TYPE *) mask_image->data);
    }

    /* Update the Min/Max elevation outside image counter */
    *min_max_elev_outside_image += local_outside_image_counter;

    return SUCCESS;
}

/*------------------------------------------------------------------------------
NAME: create_l1r_saturation_mask

PURPOSE: Creates the saturated pixel mask for pixels saturated in the L1R
         image.

RETURNS: SUCCESS / ERROR
------------------------------------------------------------------------------*/
int create_l1r_saturation_mask
(
    SATURATION_MASK *sat_mask,    /* I/O: Saturated pixel tracking structure */
    const IAS_PIXEL_MASK_IO *pixel_mask_file, /* I: Pixel mask image */
    const IAS_LOS_MODEL *model,               /* I: LOS Model */
    const IAS_PROJECTION *l1g_projection,     /* I: L1G Projection */
    IAS_CORNERS *corners,              /* I: L1G Image Corners */
    const IAS_L1G_BAND_METADATA *band_metadata, /* I: band metadata structure */
    double min_elevation,              /* I: Minimum elevation in image */
    double max_elevation,              /* I: Maximum elevation in image */
    const IAS_IMAGE *dem,              /* I: Elevation data */
    int output_saturation_mask_flag,   /* I: Flag to output saturation mask */
    const int is_band_present[IAS_MAX_NBANDS],/* I: Band present in L1G flags*/
    int band_has_saturation[IAS_MAX_NBANDS] /* O: Band saturation flags */
)
{
    double r2d = ias_math_get_degrees_per_radian();
    IAS_GEO_PROJ_TRANSFORMATION *geo2l1g = NULL;
    PIXEL_MASK_TYPE mask = IAS_RPS_ADCONV_SAT_HI | IAS_RPS_ADCONV_SAT_LO
        | IAS_RPS_ANALOG_SAT_HI | IAS_RPS_ANALOG_SAT_LO;
    IAS_IMAGE mask_image; /* Image data used to build mask */
    int band_index;
    int min_max_elev_outside_image = 0;
    int resolution_checked_flag = 0;

    IAS_LOG_INFO("Starting L1R saturated pixel identification");

    /* Set the l1g image band metadata to image structure */
    mask_image.corners = *corners;
    mask_image.pixel_size_y = band_metadata->projection_distance_y;
    mask_image.pixel_size_x = band_metadata->projection_distance_x;

    /* Allocate memory for the L1R saturated pixel mask */
    mask_image.data = calloc(sat_mask->image_lines * sat_mask->image_samples, 
        sizeof(unsigned char));
    if (!mask_image.data)
    {
        IAS_LOG_ERROR("Allocating memory for the L1R saturation mask");
        return ERROR;
    }

    /* Create the target projection transformation */
    geo2l1g = create_projection_transformation(l1g_projection);
    if (!geo2l1g)
    {
        IAS_LOG_ERROR("Creating projection transformation");
        free(mask_image.data);
        return ERROR;
    }

    /* Process each band/sca combination */
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        int band_number;
        int sca_index;
        int band_saturated = FALSE;
        int number_of_scas;

        /* Ensure the band is present before reading it */
        if (!is_band_present[band_index])
        {
            continue;
        }

        /* Continue if saturation should not be reported */
        if (!sat_mask->check_band[band_index])
        {
            continue;
        }

        /* Convert the band index to number */
        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Converting band index %d to band number",
                band_index);
            ias_geo_destroy_proj_transformation(geo2l1g);
            free(mask_image.data);
            return ERROR;
        }

        /* Check the resolution if not already checked */
        if (!resolution_checked_flag)
        {
            const IAS_BAND_ATTRIBUTES *band_attr;
            float x_res_diff;
            float y_res_diff;

            /* Retrieve the band attributes */
            band_attr = ias_sat_attr_get_band_attributes(band_number);
            if (!band_attr)
            {
                IAS_LOG_ERROR("Getting band attributes for band "
                    "number %d", band_number);
                ias_geo_destroy_proj_transformation(geo2l1g);
                free(mask_image.data);
                return ERROR;
            }

            x_res_diff = fabs(band_attr->pixel_resolution -
                mask_image.pixel_size_x);
            y_res_diff = fabs(band_attr->pixel_resolution -
                mask_image.pixel_size_y);

            /* Check that the resolution is within 1 meter */
            if (x_res_diff > 1.0 || y_res_diff > 1.0)
            {
                IAS_LOG_WARNING("The pixel resolution of the input and output "
                    "space is not close enough to be accurate, some pixels may "
                    "be excluded");
            }
            resolution_checked_flag = 1;
        }

        /* Get the number of scas for this band */
        number_of_scas = ias_sat_attr_get_scas_per_band(band_number);

        /* Loop over all scas */
        for (sca_index = 0; sca_index < number_of_scas; sca_index++)
        {
            IAS_PIXEL_MASK_ITERATOR *pixel_mask_iterator;
            IAS_PIXEL_MASK_SPAN pixel_mask_span;
            IAS_PIXEL_MASK *pixel_mask = NULL;
            int sca_number = sca_index + 1;

            /* Get the pixel mask for the specified band/SCA number. */
            pixel_mask = ias_pm_read_single_mask_from_file(pixel_mask_file,
                band_number, sca_number);
            if (pixel_mask == NULL)
            {
                IAS_LOG_ERROR("Reading band number %d SCA %d from pixel mask",
                    band_number, sca_number);
                ias_geo_destroy_proj_transformation(geo2l1g);
                free(mask_image.data);
                return ERROR;
            }

            /* Get an iterator for the mask specified on the command line */
            pixel_mask_iterator = ias_pm_get_iterator(pixel_mask,
                IAS_PM_INCLUDE, mask);
            if (pixel_mask_iterator == NULL)
            {
                IAS_LOG_ERROR("Getting the pixel mask iterator");
                ias_pm_destroy(pixel_mask);
                ias_geo_destroy_proj_transformation(geo2l1g);
                free(mask_image.data);
                return ERROR;
            }

            /* Process each span in the pixel mask for band/sca */
            while (ias_pm_get_next_span(pixel_mask_iterator, &pixel_mask_span))
            {
                if (update_mask_for_span(band_index, sca_index,
                    sat_mask->image_lines, sat_mask->image_samples,
                    min_elevation, max_elevation, &pixel_mask_span, geo2l1g,
                    dem, model, sat_mask->band_bit_mask[band_index], r2d,
                    &min_max_elev_outside_image, &mask_image) != SUCCESS)
                {
                    IAS_LOG_ERROR("Updating mask for band number %d SCA "
                        "number %d", band_number, sca_number);
                    ias_pm_destroy_iterator(pixel_mask_iterator);
                    ias_pm_destroy(pixel_mask);
                    ias_geo_destroy_proj_transformation(geo2l1g);
                    free(mask_image.data);
                    return ERROR;
                }

                /* Update the band as saturated */
                band_saturated = TRUE;
            }

            /* Set the band saturation */
            band_has_saturation[band_index] = band_saturated;

            /* Free up the mask and iterator */
            ias_pm_destroy_iterator(pixel_mask_iterator);
            ias_pm_destroy(pixel_mask);
        }
    }

    /* Output the number of times the min/max elevations fell outside the
       image */
    if (min_max_elev_outside_image > 0)
    {
        IAS_LOG_INFO("The Min/Max elevation fell outside the image %d times ",
            min_max_elev_outside_image);
    }

    /* Destroy the transformation */
    ias_geo_destroy_proj_transformation(geo2l1g);
    geo2l1g = NULL;

    /* Output the saturation mask if requested */
    if (output_saturation_mask_flag)
    {
        /* Write the saturation mask to file */
        if (ias_misc_write_envi_image("l1r_saturation_mask.img", l1g_projection,
            "L1R Saturation Mask", sat_mask->image_lines,
            sat_mask->image_samples, band_metadata->upper_left_x,
            band_metadata->upper_left_y, band_metadata->projection_distance_x,
            band_metadata->projection_distance_y, "Saturation Mask", IAS_BYTE,
            mask_image.data) != SUCCESS)
        {
            IAS_LOG_ERROR("Creating the saturation mask image");
            free(mask_image.data);
            return ERROR;
        }
    }

    /* Successfully created the mask, so store it in the tracking structure */
    sat_mask->l1r_mask = mask_image.data;

    IAS_LOG_INFO("Completed L1R saturated pixel identification");

    return SUCCESS;
}

/*------------------------------------------------------------------------------
NAME: get_l1r_saturation_mask

PURPOSE: Returns a pointer to the start of a specific line in the saturated L1R
         pixel mask buffer.

RETURNS: Pointer to the saturation buffer
------------------------------------------------------------------------------*/
const unsigned char *get_l1r_saturation_mask
(
    const SATURATION_MASK *mask,    /* I: Saturated pixel tracking structure */
    int line                        /* I: Line in L1G buffer wanted */
)
{
    if (!mask->l1r_mask)
        return NULL;

    if (line >= mask->image_lines)
    {
        IAS_LOG_WARNING("Invalid line requested, returning NULL");
        return NULL;
    }

    return &mask->l1r_mask[line * mask->image_samples];
}

/*------------------------------------------------------------------------------
NAME: initialize_saturation_mask

PURPOSE: Allocate and initialize the structure used to track the saturated
         pixels in the output image.

RETURNS: pointer to the structure allocated, or NULL if an error
------------------------------------------------------------------------------*/
SATURATION_MASK *initialize_saturation_mask
(
    int line_count,     /* I: Number of lines in the image */
    int sample_count    /* I: Number of samples in the image */
)
{
    int band_index;
    unsigned char current_bit_mask = 0x01;

    SATURATION_MASK *sat_mask;

    /* Allocate zeroed memory for the saturation tracking structure */
    sat_mask = calloc(1, sizeof(*sat_mask));
    if (!sat_mask)
    {
        IAS_LOG_ERROR("Failed allocating memory for the saturation tracking "
            "structure");
        return NULL;
    }
    
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        int band_number;
        int should_report_saturation_flag;
        const IAS_BAND_ATTRIBUTES *band_attr;

        sat_mask->l1g_saturated_count[band_index] = 0;

        /* Convert the band index to number */
        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Converting band index %d to band number",
                band_index);
            free(sat_mask);
            return NULL;
        }

        /* Retrieve the band attributes */
        band_attr = ias_sat_attr_get_band_attributes(band_number);
        if (!band_attr)
        {
            IAS_LOG_ERROR("Getting band attributes for band "
                "number %d", band_number);
            free(sat_mask);
            return NULL;
        }

        /* Check if this band should have saturation reported */
        if (ias_sat_attr_report_saturation_for_band(band_number,
            &should_report_saturation_flag) != SUCCESS)
        {
            IAS_LOG_ERROR("Getting band saturation report flag for band "
                "number %d", band_number);
            free(sat_mask);
            return NULL;
        }

        /* Flag whether the band should be checked for saturation */
        sat_mask->check_band[band_index] = should_report_saturation_flag
            && band_attr->can_saturate;

        /* Indicate the L1G band has not been checked yet */
        sat_mask->l1g_band_done[band_index] = 0;

        /* If the band is one that should be checked, assign it the next bit
           in the mask */
        if (sat_mask->check_band[band_index])
        {
            /* Make sure the bit mask hasn't exhausted the available bits */
            if (current_bit_mask == 0)
            {
                IAS_LOG_ERROR("Bands being checked for saturation exceed the "
                    "capacity of the current type");
                free(sat_mask);
                return NULL;
            }

            sat_mask->band_bit_mask[band_index] = current_bit_mask;
            current_bit_mask = current_bit_mask << 1;
        }
    }

    /* Allocate zeroed memory for the L1G mask image */
    sat_mask->image_lines = line_count;
    sat_mask->image_samples = sample_count;
    sat_mask->l1g_mask = calloc(line_count * sample_count,
        sizeof(unsigned char));
    if (!sat_mask->l1g_mask)
    {
        IAS_LOG_ERROR("Failed allocating %d bytes of memory for the "
            "L1G saturation mask", line_count * sample_count);
        free(sat_mask);
        return NULL;
    }

    /* Delaying the allocation for the L1R mask until it is created */
    sat_mask->l1r_mask = NULL;

    return sat_mask;
}

/*------------------------------------------------------------------------------
NAME: identify_saturated_l1g_pixels

PURPOSE: Identify the saturated L1G pixels in the provided buffer.

RETURNS: nothing
------------------------------------------------------------------------------*/
void identify_saturated_l1g_pixels
(
    SATURATION_MASK *mask,      /* I/O: Saturated pixel tracking structure */
    const unsigned short *image_values,/* I: DN values from the input image */
    int start_line,             /* I: Starting line in the image */
    int line_count,             /* I: Lines in image */
    int sample_count,           /* I: Samples in image */
    int band_number,            /* I: Band number being checked */
    int band_index              /* I: Band index being checked */
)
{
    int pixel_count = line_count * sample_count;
    int index;
    unsigned char bit_mask;
    unsigned char *mask_image;
    int saturated_count = 0;

    /* Skip bands that aren't being checked for saturation or that have already
       been checked */
    if (!mask->check_band[band_index] || mask->l1g_band_done[band_index])
    {
        return;
    }

    /* Get the bit mask to use for this band */
    bit_mask = mask->band_bit_mask[band_index];

    /* Cache a pointer to the saturated pixel mask with an offset to the
       starting line */
    mask_image = &mask->l1g_mask[start_line * sample_count];

    /* Check all the pixels in the band */
    for (index = 0; index < pixel_count; index++)
    {
        if (image_values[index] == SATURATED_PIXEL_VALUE)
        {
            /* The pixel is saturated, so flag it */
            mask_image[index] |= bit_mask;
            saturated_count++;
        }
    }

    /* Track the number of saturated pixels found in the band */
    mask->l1g_saturated_count[band_index] += saturated_count;

    /* Mark the band done if the last line was reached.  This is assuming the
       band is either processed all at once or from top to bottom. */
    if (start_line + line_count >= mask->image_lines)
    {
        mask->l1g_band_done[band_index] = 1;

        IAS_LOG_DEBUG("Found %d saturated L1G pixels in band number %d", 
            mask->l1g_saturated_count[band_index], band_number);
    }
}

/*------------------------------------------------------------------------------
NAME: want_l1g_saturation_for_band

PURPOSE: Returns an indication on whether the band needs to be checked for
         saturated L1G pixels.

RETURNS: non-zero if the band should be checked for saturation, zero otherwise
------------------------------------------------------------------------------*/
int want_l1g_saturation_for_band
(
    const SATURATION_MASK *mask,    /* I: Saturated pixel tracking structure */
    int band_index                  /* I: Band index to check */
)
{
    /* If an invalid band index was requested, issue a warning and return
       that the band does not need saturation checked.  This should never
       happen, so not calling it an error. */
    if (band_index < 0 || band_index >= IAS_MAX_NBANDS)
    {
        IAS_LOG_WARNING("Invalid band index %d requested", band_index);
        return 0;
    }
    return mask->check_band[band_index];
}

/*------------------------------------------------------------------------------
NAME: get_l1g_saturation_mask

PURPOSE: Returns a pointer to the start of a specific line in the saturated L1G
         pixel mask buffer.

RETURNS: Pointer to the saturation buffer
------------------------------------------------------------------------------*/
const unsigned char *get_l1g_saturation_mask
(
    const SATURATION_MASK *mask,    /* I: Saturated pixel tracking structure */
    int line                        /* I: Line in L1G buffer wanted */
)
{
    if (!mask->l1g_mask)
        return NULL;

    if (line >= mask->image_lines)
    {
        IAS_LOG_WARNING("Invalid line requested, returning NULL");
        return NULL;
    }

    return &mask->l1g_mask[line * mask->image_samples];
}

/*------------------------------------------------------------------------------
NAME: read_and_check_l1g_band_saturation

PURPOSE: Reads the indicated band from the L1G image and checks it for
         saturated pixels.

RETURNS: SUCCESS/ERROR
------------------------------------------------------------------------------*/
int read_and_check_l1g_band_saturation
(
    SATURATION_MASK *mask,   /* I/O: Saturated pixel tracking structure */
    L1GIO *l1g_file,         /* I: L1G image file containing the band to read */
    int band_number,         /* I: Band number to read */
    int band_index           /* I: Band index of the band number */
)
{
    unsigned short *image;
    L1G_BAND_IO *band;
    IAS_DATA_TYPE data_type;
    int lines;
    int samples;
    int scas = 1;

    /* Verify the band is in the image.  If not, skip it */
    if (!ias_l1g_is_band_present(l1g_file, band_number))
    {
        IAS_LOG_INFO("Band number %d not present, so no saturated pixels "
            "identified for it", band_number);
        return SUCCESS;
    }

    /* Allocate space for the band to read */
    image = malloc(mask->image_lines * mask->image_samples * sizeof(*image));
    if (!image)
    {
        IAS_LOG_ERROR("Failed to allocate memory for image");
        return ERROR;
    }

    /* Open the band and perform some consistency checks */
    band = ias_l1g_open_band(l1g_file, band_number, &data_type, &scas, &lines,
        &samples);
    if (!band)
    {
        IAS_LOG_ERROR("Opening band number %d", band_number);
        free(image);
        return ERROR;
    }
    if (data_type != IAS_UI2 || scas != 1 || lines != mask->image_lines
        || samples != mask->image_samples)
    {
        IAS_LOG_ERROR("Unexpected type or size for band number %d", 
            band_number);
        ias_l1g_close_band(band);
        free(image);
        return ERROR;
    }

    /* Read the full L1G band band */ 
    if (ias_l1g_read_image(band, 0, 0, 0, lines, samples, image) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading band number %d from the image", band_number);
        ias_l1g_close_band(band);
        free(image);
        return ERROR;
    }

    /* Close the band */
    ias_l1g_close_band(band);
    band = NULL;

    /* Identify the saturated pixels in the band */
    identify_saturated_l1g_pixels(mask, image, 0, lines, samples, band_number,
            band_index);

    /* Free the image */
    free(image);

    return SUCCESS;
}

/*------------------------------------------------------------------------------
NAME: update_bands_with_l1g_saturation

PURPOSE: Updates the provided array for bands that have L1G saturation.

RETURNS: nothing
------------------------------------------------------------------------------*/
void update_bands_with_l1g_saturation
(
    const SATURATION_MASK *mask,    /* I: Saturated pixel tracking structure */
    int band_has_saturation[IAS_MAX_NBANDS] /* I/O: Band saturation flags */
)
{
    int band_index;

    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        if (mask->l1g_saturated_count[band_index] > 0)
            band_has_saturation[band_index] = TRUE;
    }
}

/*------------------------------------------------------------------------------
NAME: cleanup_l1g_saturation

PURPOSE: Frees the provided saturation tracking structure.

RETURNS: nothing
------------------------------------------------------------------------------*/
void cleanup_saturation_mask
(
    SATURATION_MASK *mask       /* I: Saturated pixel tracking structure */
)
{
    if (mask)
    {
        free(mask->l1g_mask);
        mask->l1g_mask = NULL;
        free(mask->l1r_mask);
        mask->l1r_mask = NULL;
        free(mask);
    }
}
