/*==============================================================================
NAME: Control Parameter Functions

PURPOSE:
    This file contains functions that read, validate, and print the Cloud Cover
    Assessment (CCA) control parameters.
==============================================================================*/

/* System Includes */
#include <string.h>

/* IAS Library Includes */
#include "ias_logging.h"
#include "ias_geo.h"
#include "ias_cpf.h"

/* Local Includes */
#include "assess_cloud_cover.h"
#include "ctrl_parms.h"
#include "get_names.h"


/*------------------------------------------------------------------------------
NAME: ctrl_parameters_cpf

PURPOSE:
    This function will attempt to read the cloud cover control parameters from
    the CPF.

RETURNS:
    The function will return a status of SUCCESS if the values are successfully
    read from the CPF, ERROR if an error is encountered.
------------------------------------------------------------------------------*/
int get_ctrl_parameters_cpf
(
    const char pathCpf[],    /* I: Path to the CPF */
    double sineSolarElev,    /* I: Sine of solar elevation */
    CLOUD_COVER_CTRL *pCtrl  /* O: Control parameters */
)
{
    const IAS_CPF_CLOUD_COVER_ASSESSMENT * cpfCloud;
    /* TIRS Thermal constants */
    const struct IAS_CPF_TIRS_THERMAL_CONSTANTS *tirs_thermal_constants = NULL;

    IAS_CPF * pCpf;
    int       bandNumber;
    int       bandType;
    int       i;
    int       mapAlgorithm[NUM_ALGORITHMS];
    int       mapClass[NUM_CLASSES];

    /* Read the cloud cover assessment information from the CPF */
    pCpf = ias_cpf_read(pathCpf);
    if (pCpf == NULL)
    {
        IAS_LOG_ERROR("Reading CPF: %s", pathCpf);
        return ERROR;
    }

    cpfCloud = ias_cpf_get_cloud_cover_assessment(pCpf);
    if (cpfCloud == NULL)
    {
        IAS_LOG_ERROR("Reading cloud cover configuration from CPF: %s",
            pathCpf);
        ias_cpf_free(pCpf);
        return ERROR;
    }

    /* Get the thermal constants from CPF */
    tirs_thermal_constants = ias_cpf_get_tirs_thermal_constants(pCpf);
    if (tirs_thermal_constants == NULL)
    {
        IAS_LOG_ERROR("Retrieving CPF Thermal Constants");
        ias_cpf_free(pCpf);
        return ERROR;
    }

    pCtrl->thermalConstantsTIRS1[0] = tirs_thermal_constants->k1_constant[0];
    pCtrl->thermalConstantsTIRS1[1] = tirs_thermal_constants->k2_constant[0];
    

    /*==========================================================================
     * Now that the data has been successfully read in, we need to parse through
     * it.  This is a little complicated...
     *
     * The CPF is set up so that a user can edit the cloud cover aglorithm and
     * classification information and there is nothing keeping the user from
     * sorting these strings into an order that they like.
     *
     * The order that the algorithm and classification names are specified in,
     * dictates the order of the associated arrays. If "ACCA" is the first
     * algorithm, then it's value for Run_If_Thermal is the first entry in that
     * array.  Likewise with the classification names.  The order of these names
     * dictates the order of the values in the weights arrays.
     *
     * To ensure things match up properly, we have to create a map of our names
     * to their names.
     *========================================================================*/

    /* Initialize the maps so we can determine which items were found */
    for (i = 0; i < NUM_ALGORITHMS; i++)
        mapAlgorithm[i] = -1;
    for (i = 0; i < NUM_CLASSES; i++)
        mapClass[i] = -1;


    /* Copy the CPF's configuration info into our structure */
    for (i = 0; i < IAS_CPF_CIRRUS_THRESHOLD_SIZE; i++)
    {
        pCtrl->cirrusThreshold[i] = cpfCloud->cirrus_threshold[i];
    }
    pCtrl->minimumSunElevationDegrees = cpfCloud->minimum_sun_elevation_degrees;

    /* Free up the dynamically allocated memory in the IAS library */
    ias_cpf_free(pCpf);
    pCpf = NULL;

    /* Landsat 8 products use a psuedo-reflectance value that does not take into
       account the solar elevation angle.  All values are as if the solar
       elevation is at nadir.  The cloud cover assessment algorithms requires
       Top Of Atmosphere (TOA) reflectance values so we must divide the psuedo-
       reflectance by the sine of the solar elevation.
       
       The bandScaling factor is applied to the scaled sample values using
       multiplication so we must use the inverse of the sine value.
       
       The TIRS band is in radiance so it does not require a scaling value. */
    for (bandType = 0; bandType < IAS_NUM_BAND_TYPES; bandType++)
    {
        bandNumber = ias_sat_attr_get_band_number_from_type(bandType);
        if (bandNumber == ERROR)
        {
            IAS_LOG_ERROR("Retrieving band number for band type %d", bandType);
            return ERROR;
        }
        if (ias_sat_attr_get_spectral_type_from_band_number(bandNumber)
            == IAS_SPECTRAL_THERMAL)
            pCtrl->bandScaling[bandType] = 1.0;
        else
            pCtrl->bandScaling[bandType] = 1.0 / sineSolarElev;
    }

    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: ctrl_parameters_print

PURPOSE:
    This function prints the Cloud Cover control parameters and algorithm
    parameters to the screen for verification purposes.
------------------------------------------------------------------------------*/
void print_ctrl_parameters
(
    const CLOUD_COVER_CTRL *pCtrl     /* I: Control parameters */
)
{
    int  i;

    printf("=======================================================\n");
    printf("= Cloud Cover Control Parameters                      =\n");
    printf("=======================================================\n");
    printf("    cirrusThreshold:\n");
    for (i = 0; i < IAS_CPF_CIRRUS_THRESHOLD_SIZE; i++)
    {
        printf("%5.3lf\n", pCtrl->cirrusThreshold[i]);
    }
    printf("--------------------\n");
}
