/******************************************************************************
Name: get_image_corners

Purpose: Calculate the 4 corners of the actual imagery in the L1T.  The corners
    are returned in (UL, UR, LR, LL) order where UL is the upper-left corner
    of the image in satellite space.  This is to allow the correct satellite
    viewing angle model to be built.

Notes:
    There are various ways to calculate this.  The current implementation
    directly uses the values available in the angle coefficient file.  If that
    isn't available, some other ways to calculate it are:
    1) Load the grid headers from the grid file and call the
       ias_misc_calculate_trim_box routine (which is how the angle coefficient
       file values are calculated).  This method is not available before L8.
    2) Search the image (or pixel mask) for the image corners.  That will
       provide the corners, but they won't be properly oriented in satellite
       space.  To figure out the correct orientation in satellite space, the
       full grid would need to be read for one of the bands and the ols2ils
       routine used for all four corners to identify the proper orientation.

Returns: SUCCESS/ERROR

******************************************************************************/
#include <math.h>
#include "ias_satellite_attributes.h"
#include "ias_logging.h"
#include "get_image_corners.h"

int get_image_corners
(
    const IAS_ANGLE_GEN_METADATA *angle_metadata,/* I:angle coefficient info */
    int *corner_lines,          /* O: UL, UR, LR, LL corner lines */
    int *corner_samples         /* O: UL, UR, LR, LL corner samples */
)
{
    const IAS_ANGLE_GEN_BAND *angle_band;
    int band_number = IAS_BLUE_BAND; /* Use any band known to be used for the
                                        CFmask algorithm since all the corners
                                        are similar enough for this purpose */
    int band_index;
    int i;

    band_index = ias_sat_attr_convert_band_number_to_index(band_number);
    if (band_index == ERROR)
    {
        IAS_LOG_ERROR("Failed converting band number %d to an index", 
            band_number);
        return ERROR;
    }

    angle_band = &angle_metadata->band_metadata[band_index];

    for (i = 0; i < 4; i++)
    {
        corner_lines[i] = round(angle_band->active_l1t_corner_lines[i]);
        corner_samples[i] = round(angle_band->active_l1t_corner_samps[i]);
    }

    return SUCCESS;
}
