/*==============================================================================
NAME: Get Parameters

PURPOSE:
    This function retrieve the application's parameters from the specified
    Object Description Language (ODL) file.
==============================================================================*/

/* System Includes */
#include <stdlib.h>
#include <string.h>
#include <limits.h>

/* IAS Includes */
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "ias_satellite_attributes.h"

/* Local Includes */
#include "assess_cloud_cover.h"
#include "get_names.h"

/* Local variables */
static double saturated_pixel_threshold;
static char pixel_mask_filename[PATH_MAX];
static char model_filename[PATH_MAX];
static int enable_saturated_pixels_flag;
static int output_saturation_mask_flag;

/*------------------------------------------------------------------------------
NAME: get_parameters

PURPOSE:
    Read this application's command line arguments from an ODL file.

RETURNS:
    This function will return a status of SUCCESS if the arguments are
    successfully retrieved, ERROR if an error is encountered.
------------------------------------------------------------------------------*/
int get_parameters
(
    const char *arg,              /* I: Command line argument */
    int *satNum,                  /* O: Satellite number */
    char pathCpf[PATH_MAX],       /* O: Path of CPF */
    char pathIn[PATH_MAX],        /* O: Path of input L1G or L1T image */
    char pathOmf[PATH_MAX],       /* O: Path of OMF */
    char pathDem[PATH_MAX],       /* O: Path of DEM */
    char pathOut[PATH_MAX],       /* O: Path of quality band image */
    char pathWaterMaskPolygon[PATH_MAX],/* O: Path of water mask polygon 
                                              file */ 
    char pathOutLandWaterMaskImage[PATH_MAX],/* O: Path of mask file */
    char pathTerrain[PATH_MAX],   /* O: Path of terrain occlusion mask */
    char pathOutCfmask[PATH_MAX], /* O: Path of cfmask band mask */
    char angle_coeff_filename[PATH_MAX], /* O: Name of angle coefficient file */
    double *cfmask_cloud_prob_threshold, /* O: Cloud Probability Threshold
                                               for Cfmask */
    int *cfmask_dilate_flag,      /* O: Flag to turn on/off cloud dilation */
    int *max_threads,             /* O: Maximum number of threads to use */
    CC_ALGORITHM *requested_algorithm, /* O: Algorithm selected by the user */
    int *min_elev,                /* O: Minimum elevation */
    int *max_elev,                /* O: Maximum elevation */
    int *use_thermal_for_cfmask,  /* O: Thermal band control for CFmask */
    int *use_cirrus_for_cfmask    /* O: Cirrus band control for CFmask */
)
{
    char         workOrderId[IAS_WORKORDER_ID_SIZE];
    char         algorithmSpecified[NAME_STRLEN] = "";
    int          algorithm;
    int          parmSize;
    int          processingPass = 0;
    int          status;
    int          enable_land_cloud_score = 1; /* 1 for true, 0 for false */
    int          default_enable_land_cloud_score [] = {1};
    int default_min_elevation[] = {-500};
    int default_max_elevation[] = {10000};
    int default_enable_saturation_pixels_flag[] = {1};
    int default_output_saturation_mask_flag[] = {0};
    double default_saturated_pixel_threshold[] = {0.20};

    /* Initialize optional output paths */
    pathWaterMaskPolygon[0] = '\0';
    pathOutLandWaterMaskImage[0] = '\0';
    pathTerrain[0] = '\0';

    /*==========================================================================
     * Read this application's parameter file
     *========================================================================*/

    /* Set up the structure needed to read this application's parameters.  These
       macros actually define a hidden struct.  Using a name of xOdl so this
       name stands out. */
    IAS_PARM_DECLARE_TABLE(xOdl, 15);
    IAS_PARM_WORK_ORDER_ID(xOdl, workOrderId, sizeof(workOrderId), 1);
    IAS_PARM_PROCESSING_PASS(xOdl, &processingPass, sizeof(processingPass), 1);

    const char *valid_cc_alg[] = {"ACCA", "ATACCA", "SEE5", "CFMASK"};
    const char *default_cc_alg[] = {"CFMASK"};
    IAS_PARM_ADD_STRING(xOdl, CLOUD_COVER_ALGORITHM,
        "Specified cloud cover algorithm will be run instead of CFMask.",
        IAS_PARM_OPTIONAL, 4, valid_cc_alg, 1, default_cc_alg,
        algorithmSpecified, sizeof(algorithmSpecified), 1);

    int default_thermal[] = {USE_IF_AVAILABLE};
    IAS_PARM_ADD_INT(xOdl, CFMASK_USE_THERMAL,
        "Flag to control the use of the thermal band for the CFMASK algorithm."
        " (0 = do not use, 1 = use unconditionally, 2 = use if available)",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 0, 2, 1, default_thermal,
        use_thermal_for_cfmask, sizeof(*use_thermal_for_cfmask), 0);
    int default_cirrus[] = {USE_IF_THERMAL_NOT_AVAILABLE};
    IAS_PARM_ADD_INT(xOdl, CFMASK_USE_CIRRUS,
        "Flag to control the use of the cirrus band for the CFMASK algorithm."
        " (0 = do not use, 1 = use unconditionally, 2 = only use if the "
        "thermal band is not available)", IAS_PARM_OPTIONAL,
        IAS_PARM_NOT_ARRAY, 1, 0, 2, 1, default_cirrus, use_cirrus_for_cfmask,
        sizeof(*use_cirrus_for_cfmask), 0);
    IAS_PARM_ADD_DOUBLE(xOdl, CFMASK_CLOUD_PROB_THRESHOLD,
        "CFmask algorithm cloud probability threshold value",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 1, 100, 0, NULL,
        cfmask_cloud_prob_threshold, sizeof(*cfmask_cloud_prob_threshold), 1);
    IAS_PARM_ADD_STRING(xOdl, CFMASK_OUTPUT_FILENAME,
        "Name of the cfmask band mask file",
        IAS_PARM_OPTIONAL, 0, NULL, 0, NULL, pathOutCfmask, PATH_MAX, 1);
    int default_cfmask_dilate_flag[] = {0};
    IAS_PARM_ADD_BOOLEAN( xOdl, CFMASK_DILATE_FLAG,
        "Enables cloud dilation in the CFmask algorithm (1 = dilate, "
        "0 = don't dilate",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        default_cfmask_dilate_flag, cfmask_dilate_flag,
        sizeof(*cfmask_dilate_flag), 0 );
    IAS_PARM_ADD_STRING(xOdl, OUTPUT_FILENAME, "Name of the quality band image",
        IAS_PARM_REQUIRED, 0, NULL, 0, NULL, pathOut, PATH_MAX, 1);
    IAS_PARM_ADD_STRING(xOdl, WATER_MASK_FILENAME,
        "Polygon file defining the land masses in the world to override the "
        "default file used to create the water mask",
        IAS_PARM_OPTIONAL, 0, NULL, 0, NULL, pathWaterMaskPolygon, 
        PATH_MAX, 1);
    IAS_PARM_ADD_STRING(xOdl, LAND_WATER_MASK_OUTPUT_FILENAME,
        "Name of optional output land/water mask image filename",
        IAS_PARM_OPTIONAL, 0, NULL, 0, NULL, pathOutLandWaterMaskImage, 
        PATH_MAX, 1);
    IAS_PARM_ADD_BOOLEAN(xOdl, ENABLE_LAND_CLOUD_SCORE_FLAG,
        "Flag to enable the land cloud score. 1 = enable, 0 = disable",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        default_enable_land_cloud_score, &enable_land_cloud_score,
        sizeof(enable_land_cloud_score), 0);
    IAS_PARM_ADD_BOOLEAN(xOdl, ENABLE_SATURATED_PIXELS_FLAG,
        "Flag to enable the output of saturated pixels in the quality band",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        default_enable_saturation_pixels_flag, &enable_saturated_pixels_flag,
        sizeof(enable_saturated_pixels_flag), 0);
    IAS_PARM_ADD_BOOLEAN(xOdl, OUTPUT_SATURATION_MASK_FLAG,
        "Flag to write the saturation mask to a file",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        default_output_saturation_mask_flag, &output_saturation_mask_flag,
        sizeof(output_saturation_mask_flag), 0);
    IAS_PARM_ADD_DOUBLE(xOdl, SATURATED_PIXEL_THRESHOLD,
        "Threshold to include saturated pixels", IAS_PARM_OPTIONAL,
        IAS_PARM_NOT_ARRAY, 1, 0.0, 1.0, 1, default_saturated_pixel_threshold,
        &saturated_pixel_threshold, sizeof(saturated_pixel_threshold), 0);

    /* Read this application's parameters */
    parmSize = IAS_PARM_GET_TABLE_SIZE(xOdl);
    status = ias_parm_provide_help(arg, xOdl, parmSize, IAS_INPUT_PARAMETERS);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Providing help for application");
        return ERROR;
    }
    else if (status)
    {
        /* Since the user asked for help, copy the argument they specified into
           the OMF filename so they receive the same help message for that file
           */
        strcpy(pathOmf, arg);
    }
    else
    {
        if (ias_parm_read(arg, xOdl, parmSize) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading application parameters from file %s", arg);
            return ERROR;
        }

        *requested_algorithm = -1;
        for (algorithm = 0; algorithm < NUM_ALGORITHMS; algorithm++)
        {
            if (strcasecmp(algorithmSpecified,
                get_algorithm_name(algorithm)) == 0)
            {
                *requested_algorithm = algorithm;
                break;
            }
        }
        if (*requested_algorithm == -1)
        {
            IAS_LOG_ERROR("Algorithm name %s is not recognized",
                algorithmSpecified);
            return ERROR;
        }

        /* Create the OMF path.  We are assuming that this application is being
           run from within the current work order directory. */
        sprintf(pathOmf, "./%s.omf", workOrderId);
    }

    /* If the water mask should be used and the water mask filename was not
       provided, use the default path for water mask polygon file.
       Else set the value to empty string */
    if (enable_land_cloud_score && strcmp(pathWaterMaskPolygon, "") == 0)
    {
        const char *data_dir = getenv("IAS_DATA_DIR");
        if (!data_dir)
        {
            IAS_LOG_ERROR("IAS_DATA_DIR environment variable not set");
            return ERROR;
        }
        status = snprintf(pathWaterMaskPolygon, PATH_MAX, 
            "%s/sysdata/water_mask/cloud_score_land.ply", data_dir );
        if (status < 0 && status >= PATH_MAX)
        {
            IAS_LOG_ERROR("Creating the water polygon file name");
            return ERROR;
        }
    }
    else if (!enable_land_cloud_score)
    {
        IAS_LOG_INFO("Skipped calculation of land cloud score as requested");
        strcpy(pathWaterMaskPolygon, "");
    }

    /*==========================================================================
     * Read the Order Metadata File (OMF)
     *========================================================================*/

    /* Set up the structure needed to read this work order's Order Metadata File
       (OMF).  These macros actually define a hidden struct.  Using a name of
       xOmf so this name stands out. */
    IAS_PARM_DECLARE_TABLE(xOmf, 11);
    IAS_PARM_SATELLITE(xOmf, satNum, sizeof(*satNum), 1);
    IAS_PARM_CPF(xOmf, IAS_PARM_REQUIRED, pathCpf, PATH_MAX, 1);
    IAS_PARM_OCCLUSION_FILENAME(xOmf, pathTerrain, PATH_MAX);
    IAS_PARM_DEM_FILENAME_FROM_PASS(xOmf, IAS_PARM_OPTIONAL, pathDem, PATH_MAX,
        1, processingPass);
    IAS_PARM_L1G_FILENAME_FROM_PASS(xOmf, IAS_PARM_REQUIRED, pathIn, PATH_MAX,
        1, processingPass);
    IAS_PARM_MIN_ELEVATION(xOmf, IAS_PARM_OPTIONAL, 1,
        default_min_elevation, min_elev, sizeof(*min_elev), 1);
    IAS_PARM_MAX_ELEVATION(xOmf, IAS_PARM_OPTIONAL, 1,
        default_max_elevation, max_elev, sizeof(*max_elev), 1);

    /* Default to using a single thread */
    *max_threads = 1;
    IAS_PARM_MAX_PROCESSORS_TO_USE(xOmf, IAS_PARM_OPTIONAL, max_threads,
        sizeof(*max_threads), 1);

    strcpy(angle_coeff_filename, "");
    IAS_PARM_ANGLE_COEF_FILENAME_FROM_PASS(xOmf, IAS_PARM_OPTIONAL,
        angle_coeff_filename, PATH_MAX, 1, processingPass);

    strcpy(pixel_mask_filename, "");
    IAS_PARM_PIXEL_MASK_FILENAME(xOmf, IAS_PARM_OPTIONAL, pixel_mask_filename,
        PATH_MAX, 1);

    strcpy(model_filename, "");
    IAS_PARM_MODEL_FILENAME_FROM_PASS(xOmf, IAS_PARM_OPTIONAL,
        model_filename, PATH_MAX, 1, processingPass);

    /* Read this work order's parameters */
    parmSize = IAS_PARM_GET_TABLE_SIZE(xOmf);
    status = ias_parm_provide_help(pathOmf, xOmf, parmSize, IAS_OMF_PARAMETERS);
    if (status == 1)
        exit(EXIT_SUCCESS); /* Provided help to user, still exit successfully */
    if (status != SUCCESS)
        return ERROR;
    if (ias_parm_read(pathOmf, xOmf, parmSize) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading OMF parameters from file %s", pathOmf);
        return ERROR;
    }

    /* Throw an error if the pixel mask, dem, los model are missing
       when saturated pixels has been enabled */
    if (enable_saturated_pixels_flag && (strlen(pathDem) <= 0 ||
        strlen(pixel_mask_filename) <= 0 || strlen(model_filename) <= 0))
    {
        IAS_LOG_ERROR("The pixel mask, DEM, and LOS Model are needed "
            "to enable the saturated pixels");
        return ERROR;
    }

    return SUCCESS;
}

/*------------------------------------------------------------------------------
NAME: get_saturated_pixel_threshold

PURPOSE:
    Returns the read in saturated pixel threshold.

RETURNS:
    DOUBLE
------------------------------------------------------------------------------*/
float get_saturated_pixel_threshold(void)
{
    return saturated_pixel_threshold;
}

/*------------------------------------------------------------------------------
NAME: get_output_saturation_mask_flag

PURPOSE:
    Returns the saturation mask flag

RETURNS:
    int
------------------------------------------------------------------------------*/
int get_output_saturation_mask_flag(void)
{
    return output_saturation_mask_flag;
}

/*------------------------------------------------------------------------------
NAME: get_enable_saturated_pixels_flag

PURPOSE:
    Returns the enable saturated pixels flag

RETURNS:
    int
------------------------------------------------------------------------------*/
int get_enable_saturated_pixels_flag(void)
{
    return enable_saturated_pixels_flag;
}

/*------------------------------------------------------------------------------
NAME: get_model_filename

PURPOSE:
    Returns the model filename

RETURNS:
    CHAR *
------------------------------------------------------------------------------*/
const char * get_model_filename(void)
{
    return model_filename;
}

/*------------------------------------------------------------------------------
NAME: get_pixel_mask_filename

PURPOSE:
    Returns the pixe mask filename

RETURNS:
    CHAR *
------------------------------------------------------------------------------*/
const char * get_pixel_mask_filename(void)
{
    return pixel_mask_filename;
}
