/*******************************************************************************
NAME:
    test_cpf

PURPOSE:
    Test the reading of cloud cover assessment parameters from multiple
    Calibration Parameter Files (CPF). 

    The cloud cover assessment function get_ctrl_parameters_cpf() will read
    these CPF values using the IAS CPF library routines but will then have to
    map those values into a known structure.  This application is testing that
    dynamic mapping to ensure no errors are present.

*******************************************************************************/

/* System Includes */
#include <limits.h>
#include <math.h>

/* IAS Includes */
#include "ias_satellite_attributes.h"
#include "ias_logging.h"
#include "ias_math.h"
#include "cgreen/cgreen.h"

/* Local Includes */
#include "assess_cloud_cover.h"
#include "ctrl_parms.h"


/*******************************************************************************
NAME: test_cpf_default

PURPOSE:
    Read cpf_default.odl which contains the default cloud cover control
    parameters used by IAS.
*******************************************************************************/
Ensure test_cpf_default ()
{
    CLOUD_COVER_CTRL ctrl;
    char             cpfName[PATH_MAX] = "./cpf_default.odl";
    double           sineSolarElev;
    int              bandNumber;
    int              i;

    sineSolarElev = sin(45.0 * ias_math_get_radians_per_degree());
    if (get_ctrl_parameters_cpf(cpfName, sineSolarElev, &ctrl))
    {
        assert_true_with_message(0, "Error reading %s", cpfName);
        return;
    }
     
    assert_double_equal_with_message(ctrl.cirrusThreshold[0], 0,
        "Cirrus Threshold does not equal expected value");
    assert_double_equal_with_message(ctrl.cirrusThreshold[1], 0,
        "Cirrus Threshold does not equal expected value");
    assert_double_equal_with_message(ctrl.cirrusThreshold[2], 1000,
        "Cirrus Threshold does not equal expected value");
    assert_double_equal_with_message(ctrl.cirrusThreshold[3], 0.02,
        "Cirrus Threshold does not equal expected value");

    sineSolarElev = 1.0 / sineSolarElev;
    for (i = 0; i < IAS_NUM_BAND_TYPES; i++)
    {
        bandNumber = ias_sat_attr_get_band_number_from_type(i);
        switch (i)
        {
            case IAS_BLUE_BAND:
            case IAS_GREEN_BAND:
            case IAS_RED_BAND:
            case IAS_NIR_BAND:
            case IAS_SWIR1_BAND:
            case IAS_SWIR2_BAND:
            case IAS_CIRRUS_BAND:
                assert_double_equal_with_message(ctrl.bandScaling[i],
                    sineSolarElev, "Band %d scaling mismatch", bandNumber);
                break;
            case IAS_THERMAL1_BAND:
                assert_double_equal_with_message(ctrl.bandScaling[i], 1.0,
                    "Band %d scaling mismatch", bandNumber);
                break;
        }
    }
}


/*******************************************************************************
NAME: test_cpf01

PURPOSE:
    Read cpf01.odl which contains multiple values for the Cirrus_Threshold
    parameter.
*******************************************************************************/
Ensure test_cpf01 ()
{
    CLOUD_COVER_CTRL ctrl;

    char   cpfName[PATH_MAX] = "./cpf01.odl";
    double sineSolarElev;

    sineSolarElev = sin(45.0 * ias_math_get_radians_per_degree());
    if (get_ctrl_parameters_cpf(cpfName, sineSolarElev, &ctrl))
    {
        assert_true_with_message(0, "Error reading %s", cpfName);
        return;
    }
    
    assert_double_equal_with_message(ctrl.cirrusThreshold[0], 0.0005904,
        "Cirrus Threshold does not equal expected value");
    assert_double_equal_with_message(ctrl.cirrusThreshold[1], 0.0,
        "Cirrus Threshold does not equal expected value");
    assert_double_equal_with_message(ctrl.cirrusThreshold[2], 853.881,
        "Cirrus Threshold does not equal expected value");
    assert_double_equal_with_message(ctrl.cirrusThreshold[3], 0.02,
        "Cirrus Threshold does not equal expected value");
}

int main (int argc, char * argv[])
{
    TestSuite * suite  = create_test_suite();
    int         status;

    if (ias_sat_attr_initialize(IAS_L8))
    {
        assert_true_with_message(0,
            "Error initializing IAS satellite attributes");
    }

    add_test(suite, test_cpf_default);
    add_test(suite, test_cpf01);

    if (argc > 1)
        status = run_single_test(suite, argv[1], create_text_reporter());
    else
        status = run_test_suite(suite, create_text_reporter());
    if (status)
        return EXIT_FAILURE;
    return EXIT_SUCCESS;
}
