/*******************************************************************************
NAME:
    test_images

PURPOSE:
    This test will run the assess_cloud_cover application on the images in
    $IAS_UNIT_TEST_DATA_DIR/ias_base/qas/assess_cloud_cover.  These images are
    L1T and L1GT images.

*******************************************************************************/

/* System Includes */
#include <unistd.h>
#include <limits.h>
#include <string.h>
#include <sys/wait.h>

/* IAS Includes */
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "ias_satellite_attributes.h"
#include "cgreen/cgreen.h"

/* Local Includes */
#include "assess_cloud_cover.h"
#include "get_names.h"


/*******************************************************************************
NAME: read_omf

PURPOSE:
    Read the resulting parameters from the Order Metadata File (OMF) after a
    successful completion of assess_cloud_cover.

RETURNS:
    This function will return SUCCESS if the parameters are successfully read,
    ERROR otherwise.
*******************************************************************************/
int read_omf
(
    char     pathOmf[PATH_MAX],                      /* I: Path of the OMF */
    double * cloudScore,                             /* O: Cloud score */
    double * snowIceScore,                           /* O: Snow/ice score */
    double * cloudLandScore                          /* O: Cloud land score */
)
{
    short nArgs = 3;

    /* Set up the structure needed to update the OMF parameters */
    IAS_PARM_DECLARE_TABLE(xxxOmf, nArgs);

    IAS_PARM_CLOUD_COVER_SCORE(xxxOmf, cloudScore, sizeof(*cloudScore));
    IAS_PARM_CLOUD_COVER_LAND_SCORE(xxxOmf, cloudLandScore,
        sizeof(*cloudLandScore));
    IAS_PARM_SNOW_ICE_SCORE(xxxOmf, snowIceScore, sizeof(*snowIceScore));

    if (ias_parm_read(pathOmf, xxxOmf, nArgs) != SUCCESS)
    {
        assert_true_with_message(0, "Error reading OMF data from: %s", pathOmf);
        return ERROR;
    }

    /* Now that the parameters of interest have been read from the updated OMF
       file, we can remove it. */
    unlink(pathOmf);
    return SUCCESS;
}


/*******************************************************************************
NAME: run_cloud_cover

PURPOSE:
    Run the assess_cloud_cover application on a test image.

RETURNS:
    This function will return SUCCESS if the application is successfully run,
    ERROR otherwise.
*******************************************************************************/
int run_cloud_cover
(
    char * pathOdl      /* I: Path of the parameter ODL file */
)
{
    char cmd[PATH_MAX+64];
    int  status;

    status = snprintf(cmd, sizeof(cmd), "../assess_cloud_cover %s", pathOdl);
    if (status < 0 || status >= sizeof(cmd))
    {
        assert_true_with_message(0, "Insufficient command buffer size");
        return ERROR;
    }

    status = system(cmd);
    if (WIFSIGNALED(status) && (WTERMSIG(status) || WTERMSIG(status)))
    {
        assert_true_with_message(0, "Cloud cover assessment interrupted");
        return ERROR;
    }
    if (WEXITSTATUS(status) != 0)
    {
        assert_true_with_message(0, "Cloud cover assessment failed");
        return ERROR;
    }

    /* Toss the generated ODL file */
    unlink(pathOdl);
    return SUCCESS;
}


/*******************************************************************************
NAME: run_image_diff

PURPOSE:
    Run an H5Diff on the specified images.

RETURNS:
    This function will return SUCCESS if the image diff application is
    successfully run and the images are identical, ERROR if an error is
    encountered or if the images are different.
*******************************************************************************/
int run_image_diff
(
    char * path1,       /* I: Path to a quality band image */
    char * path2        /* I: Path to a quality band image */
)
{
    char cmd[PATH_MAX];
    int          status;

    status = snprintf(cmd, sizeof(cmd), "h5diff %s %s", path1, path2);
    if (status < 0 || status >= sizeof(cmd))
    {
        assert_true_with_message(0, "Insufficient command buffer size");
        return ERROR;
    }

    status = system(cmd);
    if (WIFSIGNALED(status) && (WTERMSIG(status) || WTERMSIG(status)))
    {
        assert_true_with_message(0, "h5diff file comparison interrupted");
        return ERROR;
    }
    if (WEXITSTATUS(status) != 0)
    {
        assert_true_with_message(0, "Quality band image %s differs from "
            "reference image %s", path1, path2);
        return ERROR;
    }
    return SUCCESS;
}
/*******************************************************************************
NAME: run_diff

PURPOSE:
    Run an diff on the specified file.

RETURNS:
    This function will return SUCCESS if the diff application is
    successfully run and the files are identical, ERROR if an error is
    encountered or if the files are different.
*******************************************************************************/
int run_diff
(
    const char * path1,       /* I: Path to a file */
    const char * path2,       /* I: Path to a file */
    int binary_compare        /* I: Flag indicating the compare is binary */
)
{
    char cmd[PATH_MAX];
    const char *diff_cmd = "diff";
    int status;

    if (binary_compare)
    {
        /* Don't report differences for binary files since it can result in a
           lot of garbage on the display */
        diff_cmd = "diff -q";
    }

    /* Create the command to be used to run an diff comparison */
    status = snprintf(cmd, sizeof(cmd), "%s %s %s", diff_cmd, path1, path2);
    if (status < 0 || status >= sizeof(cmd))
    {
        assert_true_with_message(0, "Insufficient command buffer size");
        return ERROR;
    }

    status = system(cmd);
    if (WIFSIGNALED(status) && (WTERMSIG(status) || WTERMSIG(status)))
    {
        assert_true_with_message(0, "diff file comparison interrupted");
        return ERROR;
    }
    if (WEXITSTATUS(status) != 0)
    {
        assert_true_with_message(0,
            "Cfmask band file differs from reference cfmask band file.");
        return ERROR;
    }
    return SUCCESS;
}


/*******************************************************************************
NAME: test_setup

PURPOSE:
    Set up the test for the specified image

RETURNS:
    This function will return SUCCESS if the test is successfully set up, ERROR
    otherwise.
*******************************************************************************/
int test_setup
(
    char * id,             /* I: ID of the image to work with */
    char * alg,            /* I: Algorithm to run */
    char * pathOdl,        /* O: Path to the ODL file */
    char * pathOmf,        /* O: Path to the OMF file */
    char * pathRefImage,   /* O: Path to reference image */
    char * pathRefQB,      /* O: Path to reference quality band image */
    char * pathNewQB,      /* O: Path to new quality band image */
    char * pathRefCfmaskBand, /* O: Path to cfmask band file */
    char * pathNewCfmaskBand  /* O: Path to new cfmask band file */
)
{
    /* Relative directory below the unit test directory for this application */
    const char kDirRelative[64] = "ias_base/qas/assess_cloud_cover";

    /* Length of the work order ID's used in this unit test */
    const int  kWorkOrderIdLen  = 6;

    char omf_id[40];
    FILE *       fd;
    char         cmd[PATH_MAX];
    char         sedCmd[PATH_MAX];
    const char * pEnv;
    int          processingPass;
    double       cfmaskCloudProbThreshold = 22.500000;

    /* Retrieve the directory specified by the environment variable
       IAS_UNIT_TEST_DATA_DIR.  If this environment variable doesn't exist, the
       test should fail. */
    if ((pEnv = getenv("IAS_UNIT_TEST_DATA_DIR")) == NULL)
    {
        assert_true_with_message(0,
            "Environment variable IAS_UNIT_TEST_DATA_DIR not defined");
        return ERROR;
    }

    if (strlen(id) >= sizeof(omf_id))
    {
        assert_true_with_message(0, "ID %s is too long", id);
        return ERROR;
    }
    strcpy(omf_id, id);

    /* The image being tested must be 1 of 7 */
    if ((strcmp(id, "LC81380162001164SGS00") == 0) ||
        (strcmp(id, "LC81380162001164SGS01") == 0) ||
        (strcmp(id, "LO81390452001251SGS00") == 0) ||
        (strcmp(id, "LC80160042013118LGN00") == 0))
    {
        snprintf(pathRefImage, PATH_MAX, "%s/%s/%s/l1t.h5", pEnv, kDirRelative,
            id);
        if (strcmp(id, "LC81380162001164SGS01") == 0)
        {
            /* quality band name for this interval contains the algorithm */
            snprintf(pathRefQB, PATH_MAX, "%s/%s/%s/quality_band_%s.h5", pEnv,
                kDirRelative, id, alg);
            snprintf(pathRefCfmaskBand, PATH_MAX,
                "%s/%s/%s/cfmask_band_V4.img", pEnv, kDirRelative, id);
        }
        else if (strcmp(id, "LC80160042013118LGN00") == 0)
        {
            snprintf(pathRefQB, PATH_MAX, "%s/%s/%s/quality_band_V4.h5", pEnv,
                kDirRelative, id);
            snprintf(pathRefCfmaskBand, PATH_MAX,
                "%s/%s/%s/cfmask_band_V2.img", pEnv, kDirRelative, id);
        }
        else if (strcmp(id, "LO81390452001251SGS00") == 0)
        {
            snprintf(pathRefQB, PATH_MAX, "%s/%s/%s/quality_band_cirrus.h5",
                pEnv, kDirRelative, id);
            snprintf(pathRefCfmaskBand, PATH_MAX,
                "%s/%s/%s/cfmask_band_V2.img", pEnv, kDirRelative, id);
        }
        else
        {
            snprintf(pathRefQB, PATH_MAX, "%s/%s/%s/quality_band_V3.h5", pEnv,
                kDirRelative, id);
            snprintf(pathRefCfmaskBand, PATH_MAX,
                "%s/%s/%s/cfmask_band_V2.img", pEnv, kDirRelative, id);
        }
        snprintf(pathNewQB, PATH_MAX, "./tmp%s-quality_band.h5", id);
        snprintf(pathNewCfmaskBand, PATH_MAX, "./tmp%s-cfmask_band.img",id);
        processingPass = 2;
    }
    else if (strcmp(id, "LT81510282000189SGS00") == 0)
    {
        snprintf(pathRefImage, PATH_MAX, "%s/%s/%s/sys_l1g.h5", pEnv,
            kDirRelative, id);
        snprintf(pathRefQB, PATH_MAX, "%s/%s/%s/quality_band_V2.h5", pEnv,
            kDirRelative, id);
        snprintf(pathRefCfmaskBand, PATH_MAX, "%s/%s/%s/cfmask_band.img", pEnv,
            kDirRelative, id);

        snprintf(pathNewQB, PATH_MAX, "./tmp%s-quality_band.h5", id);
        snprintf(pathNewCfmaskBand, PATH_MAX, "./tmp%s-cfmask_band.img",id);
        processingPass = 1;
    }
    else if (strcmp(id, "LC80160502014041LGN00") == 0)
    {
        strcpy(omf_id, "LC8999");
        snprintf(pathRefImage, PATH_MAX, "%s/%s/%s/l1t.h5", pEnv,
            kDirRelative, id);
        snprintf(pathRefQB, PATH_MAX, "%s/%s/%s/quality_band_%s", pEnv,
            kDirRelative, id, alg);
        snprintf(pathRefCfmaskBand, PATH_MAX, "%s/%s/%s/cfmask_band",
             pEnv, kDirRelative, id);

        snprintf(pathNewQB, PATH_MAX, "./tmp%s-quality_band_%s.h5", id,
            alg);
        snprintf(pathNewCfmaskBand, PATH_MAX, "./tmp%s-cfmask_band.img", id);
        processingPass = 2;
    }
    else if (strcmp(id, "LC81730212013114LGN01") == 0)
    {
        snprintf(pathRefImage, PATH_MAX, "%s/%s/%s/l1t.h5", pEnv,
            kDirRelative, id);
        snprintf(pathRefQB, PATH_MAX, "%s/%s/%s/quality_band.h5", pEnv,
            kDirRelative, id);
        snprintf(pathRefCfmaskBand, PATH_MAX, "%s/%s/%s/cfmask_band.img", pEnv,
            kDirRelative, id);

        snprintf(pathNewQB, PATH_MAX, "./tmp%s-quality_band.h5", id);
        snprintf(pathNewCfmaskBand, PATH_MAX, "./tmp%s-cfmask_band.img",id);
        processingPass = 3;
    }
    else
    {
        assert_true_with_message(0, "Logic error: unknown image %s", id);
        return ERROR;
    }

    if (strcmp(id, "LC80160502014041LGN00") == 0)
    {
        snprintf(pathOdl, PATH_MAX, "./tmp%*.*s.odl", kWorkOrderIdLen,
            kWorkOrderIdLen, "LC8999");
        snprintf(pathOmf, PATH_MAX, "./tmp%*.*s.omf", kWorkOrderIdLen,
            kWorkOrderIdLen, omf_id);
    }
    else
    {
        snprintf(pathOdl, PATH_MAX, "./tmp%*.*s.odl", kWorkOrderIdLen,
            kWorkOrderIdLen, id);
        snprintf(pathOmf, PATH_MAX, "./tmp%*.*s.omf", kWorkOrderIdLen,
            kWorkOrderIdLen, omf_id);
    }

    /* Create an ODL file that will be used for input to assess_cloud_cover.
       This file will be removed when the application completes successfully. */
    fd = fopen(pathOdl, "w");
    if (fd == NULL)
    {
        assert_true_with_message(0, "Error opening temporary ODL file: %s",
            pathOdl);
        return ERROR;
    }
    fprintf(fd, "OBJECT = PARAMETERS\n");
    fprintf(fd, "  WORK_ORDER_ID   = \"tmp%*.*s\"\n", kWorkOrderIdLen,
        kWorkOrderIdLen, omf_id);
    fprintf(fd, "  CFMASK_CLOUD_PROB_THRESHOLD = %f\n",
        cfmaskCloudProbThreshold);
    fprintf(fd, "  CFMASK_OUTPUT_FILENAME = \"%s\"\n", pathNewCfmaskBand);
    fprintf(fd, "  PROCESSING_PASS = %d\n", processingPass);
    fprintf(fd, "  OUTPUT_FILENAME = \"%s\"\n", pathNewQB);
    fprintf(fd, "  CLOUD_COVER_ALGORITHM = \"%s\"\n", alg);

    if (strcmp(id, "LC80160502014041LGN00") == 0)
    {
        fprintf(fd, "  CFMASK_USE_THERMAL = 0\n");
        fprintf(fd, "  CFMASK_USE_CIRRUS = 0\n");
    }

    if (strcmp(id, "LC80160042013118LGN00") == 0
        || strcmp(id, "LC81730212013114LGN01") == 0)
    {
        fprintf(fd, "  ENABLE_SATURATED_PIXELS_FLAG = 1\n");

        if (strcmp(id, "LC81730212013114LGN01") == 0)
            fprintf(fd, "  OUTPUT_SATURATION_MASK_FLAG = 1\n");
    }
    else
    {
        fprintf(fd, "  ENABLE_SATURATED_PIXELS_FLAG = 0\n");
    }

    fprintf(fd, "END_OBJECT = PARAMETERS\n");
    fprintf(fd, "END\n");
    fclose(fd);

    /* Copy the OMF file that is checked into the SVN repository so it can be
       modified by assess_cloud_cover with its output parameters.  This will
       prevent the modified files from getting checked into SVN by mistake.

       During this copy operation any occurence of the string
       IAS_UNIT_TEST_DATA_DIR must be replaced with the environment variable's
       current value.  This can be accomplished by cat'ing the original SVN
       controlled file through sed.  Since the environment variable contains '/'
       characters, we'll use the '@' character as our separation character in
       the sed command. */
    snprintf(sedCmd, sizeof(sedCmd),
        "s@IAS_UNIT_TEST_DATA_DIR@${IAS_UNIT_TEST_DATA_DIR}@g");
    snprintf(cmd, sizeof(cmd), "cat ./%*.*s.omf | sed -e \"%s\" - > %s",
        kWorkOrderIdLen, kWorkOrderIdLen, omf_id, sedCmd, pathOmf);
    system(cmd);

    return SUCCESS;
}


/*******************************************************************************
NAME: LO81390452001251SGS00

PURPOSE:
    Run the assess_cloud_cover application on the test image from Path 139.
    ACCA can not run without the thermal band, so the quality scores will
    be -1.
*******************************************************************************/
Ensure LO81390452001251SGS00()
{
    const double kExpectedScore = -1.00;
    const double kExpectedSnowIceScore = -1.00;
    const double kExpectedCloudLandScore = -1.00;

    char   pathNewQB[PATH_MAX];
    char   pathOdl[PATH_MAX];
    char   pathOmf[PATH_MAX];
    char   pathRefImage[PATH_MAX];
    char   pathRefQB[PATH_MAX];
    char   pathRefCfmaskBand[PATH_MAX];
    char   pathNewCfmaskBand[PATH_MAX];
    double cloudScore;
    double snowIceScore;
    double cloudLandScore;

    if (test_setup("LO81390452001251SGS00", "ACCA", pathOdl, pathOmf,
        pathRefImage, pathRefQB, pathNewQB, pathRefCfmaskBand,
        pathNewCfmaskBand) != SUCCESS)
    {
        return;
    }
    if (run_cloud_cover(pathOdl) != SUCCESS)
        return;
    if (read_omf(pathOmf, &cloudScore, &snowIceScore, &cloudLandScore)
        != SUCCESS)
    {
        return;
    }

    assert_double_equal_with_message(cloudScore, kExpectedScore,
        "Cloud score %3.2lf != %3.2lf", cloudScore, kExpectedScore);
    assert_double_equal_with_message(snowIceScore, kExpectedSnowIceScore,
        "Snow/Ice score %3.2lf != %3.2lf", snowIceScore, 
        kExpectedSnowIceScore);    
    assert_double_equal_with_message(cloudLandScore, kExpectedCloudLandScore,
        "Cloud Land score %3.2lf != %3.2lf", cloudLandScore, 
        kExpectedCloudLandScore);

    if (run_image_diff(pathRefQB, pathNewQB) != SUCCESS)
        return;
    unlink(pathNewQB);
}


/*******************************************************************************
NAME: LT81510282000189SGS00

PURPOSE:
    Run the assess_cloud_cover application on the test image from Path 151
    ATACCA can not run with only the thermal bands, so the quality scores will
    be -1.
*******************************************************************************/
Ensure LT81510282000189SGS00()
{
    const double kExpectedScore = -1.00;
    const double kExpectedSnowIceScore = -1.00;
    const double kExpectedCloudLandScore = -1.00;

    char   pathNewQB[PATH_MAX];
    char   pathOdl[PATH_MAX];
    char   pathOmf[PATH_MAX];
    char   pathRefImage[PATH_MAX];
    char   pathRefQB[PATH_MAX];
    char   pathRefCfmaskBand[PATH_MAX];
    char   pathNewCfmaskBand[PATH_MAX];
    double cloudScore;
    double snowIceScore;
    double cloudLandScore;

    if (test_setup("LT81510282000189SGS00", "ATACCA", pathOdl, pathOmf,
        pathRefImage, pathRefQB, pathNewQB, pathRefCfmaskBand,
        pathNewCfmaskBand) != SUCCESS)
    {
        return;
    }
    if (run_cloud_cover(pathOdl) != SUCCESS)
        return;
    if (read_omf(pathOmf, &cloudScore, &snowIceScore, &cloudLandScore)
        != SUCCESS)
    {
        return;
    }

    assert_double_equal_with_message(cloudScore, kExpectedScore,
        "Cloud score %3.2lf != %3.2lf", cloudScore, kExpectedScore);
    assert_double_equal_with_message(snowIceScore, kExpectedSnowIceScore,
        "Snow/Ice score %3.2lf != %3.2lf", snowIceScore, kExpectedSnowIceScore);
    assert_double_equal_with_message(cloudLandScore, kExpectedCloudLandScore,
        "Cloud land score %3.2lf != %3.2lf", cloudLandScore, 
        kExpectedCloudLandScore);

    if (run_image_diff(pathRefQB, pathNewQB) != SUCCESS)
        return;
    unlink(pathNewQB);
}


/*******************************************************************************
NAME: LC80160042013118LGN00 

PURPOSE:
    Run the assess_cloud_cover application on the test image from Path 016
    Added for Minimum_Sun_Elevation. The sun angle in this scene is
    25.76305695. The Minimum_Sun_Elevation is set to 50.00 for this scene.
    Since 25.76305695 < 50.00 none of the cloud cover algorith will run.
    Also the cloud cover score is -1.
*******************************************************************************/
Ensure LC80160042013118LGN00 ()
{
    const double kExpectedScore = -1;
    const double kExpectedSnowIceScore = -1;
    const double kExpectedCloudLandScore = -1;

    char   pathNewQB[PATH_MAX];
    char   pathOdl[PATH_MAX];
    char   pathOmf[PATH_MAX];
    char   pathRefImage[PATH_MAX];
    char   pathRefQB[PATH_MAX];
    char   pathRefCfmaskBand[PATH_MAX];
    char   pathNewCfmaskBand[PATH_MAX];
    double cloudScore;
    double snowIceScore;
    double cloudLandScore;

    if (test_setup("LC80160042013118LGN00", "ACCA", pathOdl, pathOmf,
        pathRefImage, pathRefQB ,pathNewQB, pathRefCfmaskBand,
        pathNewCfmaskBand) != SUCCESS)
    {
        return;
    }
    if (run_cloud_cover(pathOdl) != SUCCESS)
        return;
    if (read_omf(pathOmf, &cloudScore, &snowIceScore, &cloudLandScore)
        != SUCCESS)
    {
        return;
    }

    assert_double_equal_with_message(cloudScore, kExpectedScore,
        "Cloud score %3.2lf != %3.2lf", cloudScore, kExpectedScore);
    assert_double_equal_with_message(snowIceScore, kExpectedSnowIceScore,
        "Snow/Ice score %3.2lf != %3.2lf", snowIceScore, kExpectedSnowIceScore);
    assert_double_equal_with_message(cloudLandScore, kExpectedCloudLandScore,
        "Cloud Land Score %3.2lf != %3.2lf", cloudLandScore,
        kExpectedCloudLandScore);

    if (run_image_diff(pathRefQB, pathNewQB) != SUCCESS)
        return;
    unlink(pathNewQB);
}


/*******************************************************************************
NAME: LC80160502014041LGN00_Cfmask

PURPOSE:
    Run the assess_cloud_cover application on the test image from Path 016.
    This uses cfmask with all 4 combinations of thermal and cirrus flag.
    CFMASK_USE_THERMAL = 0  CFMASK_USE_CIRRUS = 0
    CFMASK_USE_THERMAL = 0  CFMASK_USE_CIRRUS = 1
    CFMASK_USE_THERMAL = 1  CFMASK_USE_CIRRUS = 0
    CFMASK_USE_THERMAL = 1  CFMASK_USE_CIRRUS = 1 
*******************************************************************************/
Ensure LC80160502014041LGN00_Cfmask()
{
    const double kExpectedScore[4] = {44.80, 45.64, 47.69, 47.74};
    const double kExpectedSnowIceScore[4] = {0.12, 0.04, 0.0, 0.0};
    const double kExpectedCloudLandScore[4] = {44.80, 45.64, 47.69, 47.74};

    char   pathNewQB[PATH_MAX];
    char   pathOdl[PATH_MAX];
    char   pathOmf[PATH_MAX];
    char   pathRefImage[PATH_MAX];
    char   pathRefQB[PATH_MAX];
    char   pathRefCfmaskBand[PATH_MAX];
    char   pathNewCfmaskBand[PATH_MAX];
    double cloudScore;
    double snowIceScore;
    double cloudLandScore;

    /* Initially both  flags are set to 0 */
    int cfmask_use_thermal = 0;
    int cfmask_use_cirrus = 0;
    int i;
    char command[200];
    char temp[PATH_MAX];

    /* Loop 4 times to get all the combinations of thermal and cirrus flag */
    for (i = 0; i < 4; i++)
    {
        if (test_setup("LC80160502014041LGN00", "CFMASK", pathOdl, pathOmf,
            pathRefImage, pathRefQB, pathNewQB, pathRefCfmaskBand,
            pathNewCfmaskBand) != SUCCESS)
        {
            return;
        }

        /* Set the values of thermal and cirrus flags to change in odl */
        if (i == 1)
        {
            cfmask_use_thermal = 0;
            cfmask_use_cirrus = 1;
        }
        if (i == 2)
        {
            cfmask_use_thermal = 1;
            cfmask_use_cirrus = 0;
        }
        if (i == 3)
        {
            cfmask_use_thermal = 1;
            cfmask_use_cirrus = 1;
        }

        /* Perform sed on the odl file to replace thermal and cirrus flags */
        snprintf(command, 200,
            "sed -i -e 's|CFMASK_USE_THERMAL = 0|CFMASK_USE_THERMAL = %d|' "
            "-e 's|CFMASK_USE_CIRRUS = 0|CFMASK_USE_CIRRUS = %d|' %s", 
            cfmask_use_thermal, cfmask_use_cirrus, pathOdl);
        system(command);

        /* Suffix the thermal and cirrus flag case */
        strcpy(temp, pathRefQB);
        snprintf(pathRefQB, PATH_MAX, "%s_T%dC%d.h5", temp, cfmask_use_thermal,
            cfmask_use_cirrus);
        strcpy(temp, pathRefCfmaskBand);
        snprintf(pathRefCfmaskBand, PATH_MAX, "%s_T%dC%d.img", temp,
            cfmask_use_thermal, cfmask_use_cirrus);

        /* Run the cloud cover test */
        if (run_cloud_cover(pathOdl) != SUCCESS)
            return;
        if (read_omf(pathOmf, &cloudScore, &snowIceScore, &cloudLandScore)
            != SUCCESS)
        {
            return;
        }

        assert_double_equal_with_message(cloudScore, kExpectedScore[i],
            "Cloud score %3.2lf != %3.2lf", cloudScore, kExpectedScore[i]);
        assert_double_equal_with_message(snowIceScore, kExpectedSnowIceScore[i],
            "Snow/Ice score %3.2lf != %3.2lf", snowIceScore,
             kExpectedSnowIceScore[i]);
        assert_double_equal_with_message(cloudLandScore,
             kExpectedCloudLandScore[i],
            "Cloud Land Score %3.2lf != %3.2lf", cloudLandScore,
            kExpectedCloudLandScore[i]);

       /* Run the diff for validation */
       if (run_image_diff(pathRefQB, pathNewQB) != SUCCESS)
            return;
        unlink(pathNewQB);
        if (run_diff(pathRefCfmaskBand, pathNewCfmaskBand, 1) != SUCCESS)
            return;
        unlink(pathNewCfmaskBand);
        strcat(pathRefCfmaskBand,".hdr");
        strcat(pathNewCfmaskBand,".hdr");
        if (run_diff(pathRefCfmaskBand, pathNewCfmaskBand, 0) != SUCCESS)
            return;
        unlink(pathNewCfmaskBand);
    }
}

/*******************************************************************************
NAME: LC81380162001164SGS01_ACCA

PURPOSE:
    Run the assess_cloud_cover application on the test image from Path 138 using
    ACCA.
*******************************************************************************/
Ensure LC81380162001164SGS01_ACCA()
{
    double kExpectedScore = 23.63;
    double kExpectedSnowIceScore = 0.00;
    double kExpectedCloudLandScore = 23.63;

    char   pathNewQB[PATH_MAX];
    char   pathOdl[PATH_MAX];
    char   pathOmf[PATH_MAX];
    char   pathRefImage[PATH_MAX];
    char   pathRefQB[PATH_MAX];
    char   pathRefCfmaskBand[PATH_MAX];
    char   pathNewCfmaskBand[PATH_MAX];
    double cloudScore;
    double snowIceScore;
    double cloudLandScore;

    if (test_setup("LC81380162001164SGS01", "ACCA", pathOdl, pathOmf,
        pathRefImage, pathRefQB, pathNewQB, pathRefCfmaskBand,
        pathNewCfmaskBand) != SUCCESS)
    {
        return;
    }
    if (run_cloud_cover(pathOdl) != SUCCESS)
        return;
    if (read_omf(pathOmf, &cloudScore, &snowIceScore, &cloudLandScore)
        != SUCCESS)
    {
        return;
    }

    assert_double_equal_with_message(cloudScore, kExpectedScore,
        "Cloud score %3.2lf != %3.2lf", cloudScore, kExpectedScore);
    assert_double_equal_with_message(snowIceScore, kExpectedSnowIceScore,
        "Snow/Ice score %3.2lf != %3.2lf", snowIceScore, kExpectedSnowIceScore);
    assert_double_equal_with_message(cloudLandScore, kExpectedCloudLandScore,
        "Cloud Land score %3.2lf != %3.2lf", cloudLandScore, 
        kExpectedCloudLandScore);

    if (run_image_diff(pathRefQB, pathNewQB) != SUCCESS)
        return;
    unlink(pathNewQB);
}

/*******************************************************************************
NAME: LC81380162001164SGS01_ATACCA

PURPOSE:
    Run the assess_cloud_cover application on the test image from Path 138 using
    ATACCA.
*******************************************************************************/
Ensure LC81380162001164SGS01_ATACCA()
{
    const double kExpectedScore = 24.97;
    const double kExpectedSnowIceScore = 0.10;
    const double kExpectedCloudLandScore = 24.97;

    char   pathNewQB[PATH_MAX];
    char   pathOdl[PATH_MAX];
    char   pathOmf[PATH_MAX];
    char   pathRefImage[PATH_MAX];
    char   pathRefQB[PATH_MAX];
    char   pathRefCfmaskBand[PATH_MAX];
    char   pathNewCfmaskBand[PATH_MAX];
    double cloudScore;
    double snowIceScore;
    double cloudLandScore;

    if (test_setup("LC81380162001164SGS01", "ATACCA", pathOdl, pathOmf,
        pathRefImage, pathRefQB, pathNewQB, pathRefCfmaskBand,
        pathNewCfmaskBand) != SUCCESS)
    {
        return;
    }
    if (run_cloud_cover(pathOdl) != SUCCESS)
        return;
    if (read_omf(pathOmf, &cloudScore, &snowIceScore, &cloudLandScore)
        != SUCCESS)
    {
        return;
    }

    assert_double_equal_with_message(cloudScore, kExpectedScore,
        "Cloud score %3.2lf != %3.2lf", cloudScore, kExpectedScore);
    assert_double_equal_with_message(snowIceScore, kExpectedSnowIceScore,
        "Snow/Ice score %3.2lf != %3.2lf", snowIceScore, kExpectedSnowIceScore);
    assert_double_equal_with_message(cloudLandScore, kExpectedCloudLandScore,
        "Cloud Land score %3.2lf != %3.2lf", cloudLandScore, 
        kExpectedCloudLandScore);

    if (run_image_diff(pathRefQB, pathNewQB) != SUCCESS)
        return;
    unlink(pathNewQB);
}

/*******************************************************************************
NAME: LC81380162001164SGS01_See5

PURPOSE:
    Run the assess_cloud_cover application on the test image from Path 138 using
    See5.
*******************************************************************************/
Ensure LC81380162001164SGS01_See5()
{
    const double kExpectedScore = 19.02;
    const double kExpectedSnowIceScore = 0.00;
    const double kExpectedCloudLandScore = 19.02;

    char   pathNewQB[PATH_MAX];
    char   pathOdl[PATH_MAX];
    char   pathOmf[PATH_MAX];
    char   pathRefImage[PATH_MAX];
    char   pathRefQB[PATH_MAX];
    char   pathRefCfmaskBand[PATH_MAX];
    char   pathNewCfmaskBand[PATH_MAX];
    double cloudScore;
    double snowIceScore;
    double cloudLandScore;

    if (test_setup("LC81380162001164SGS01", "SEE5", pathOdl, pathOmf,
        pathRefImage, pathRefQB, pathNewQB, pathRefCfmaskBand,
        pathNewCfmaskBand) != SUCCESS)
    {
        return;
    }
    if (run_cloud_cover(pathOdl) != SUCCESS)
        return;
    if (read_omf(pathOmf, &cloudScore, &snowIceScore, &cloudLandScore)
        != SUCCESS)
    {
        return;
    }

    assert_double_equal_with_message(cloudScore, kExpectedScore,
        "Cloud score %3.2lf != %3.2lf", cloudScore, kExpectedScore);
    assert_double_equal_with_message(snowIceScore, kExpectedSnowIceScore,
        "Snow/Ice score %3.2lf != %3.2lf", snowIceScore, kExpectedSnowIceScore);
    assert_double_equal_with_message(cloudLandScore, kExpectedCloudLandScore,
        "Cloud Land Score %3.2lf != %3.2lf", cloudLandScore,
        kExpectedCloudLandScore);

    if (run_image_diff(pathRefQB, pathNewQB) != SUCCESS)
        return;
    unlink(pathNewQB);
}

/*******************************************************************************
NAME: LC81380162001164SGS01_Cfmask

PURPOSE:
    Run the assess_cloud_cover application on the test image from Path 138 using
    Cfmask with CFMASK_USE_THERMAL(=2) and CFMASK_USE_CIRRUS(=2) as
    default value.
*******************************************************************************/

Ensure LC81380162001164SGS01_Cfmask()
{
    double kExpectedScore = 23.17;
    double kExpectedSnowIceScore = 0.09;
    double kExpectedCloudLandScore = 23.17;

    char   pathNewQB[PATH_MAX];
    char   pathOdl[PATH_MAX];
    char   pathOmf[PATH_MAX];
    char   pathRefImage[PATH_MAX];
    char   pathRefQB[PATH_MAX];
    char   pathRefCfmaskBand[PATH_MAX];
    char   pathNewCfmaskBand[PATH_MAX];
    double cloudScore;
    double snowIceScore;
    double cloudLandScore;

    if (test_setup("LC81380162001164SGS01", "CFMASK", pathOdl, pathOmf,
        pathRefImage, pathRefQB, pathNewQB, pathRefCfmaskBand,
        pathNewCfmaskBand) != SUCCESS)
    {
        return;
    }
    if (run_cloud_cover(pathOdl) != SUCCESS)
        return;
    if (read_omf(pathOmf, &cloudScore, &snowIceScore, &cloudLandScore)
        != SUCCESS)
    {
        return;
    }

    assert_double_equal_with_message(cloudScore, kExpectedScore,
        "Cloud score %3.2lf != %3.2lf", cloudScore, kExpectedScore);
    assert_double_equal_with_message(snowIceScore, kExpectedSnowIceScore,
        "Snow/Ice score %3.2lf != %3.2lf", snowIceScore, kExpectedSnowIceScore);
    assert_double_equal_with_message(cloudLandScore, kExpectedCloudLandScore,
        "Cloud Land score %3.2lf != %3.2lf", cloudLandScore,
        kExpectedCloudLandScore);

    if (run_image_diff(pathRefQB, pathNewQB) != SUCCESS)
        return;
    unlink(pathNewQB);
}

/*******************************************************************************
NAME: LC81730212013114LGN01

PURPOSE:
    Run the assess_cloud_cover application on the test image that has both
    L1R saturation and L1T saturation and verify the masks and quality band
    are correct.
*******************************************************************************/

Ensure LC81730212013114LGN01()
{
    double kExpectedScore = 0.04;
    double kExpectedSnowIceScore = 0.06;
    double kExpectedCloudLandScore = 0.04;

    char pathNewQB[PATH_MAX];
    char pathOdl[PATH_MAX];
    char pathOmf[PATH_MAX];
    char pathRefImage[PATH_MAX];
    char pathRefQB[PATH_MAX];
    char pathRefCfmaskBand[PATH_MAX];
    char pathNewCfmaskBand[PATH_MAX];
    char ref_dir[PATH_MAX];
    char *temp_ptr;
    char ref_l1r_sat[PATH_MAX];
    char ref_l1g_sat[PATH_MAX];
    double cloudScore;
    double snowIceScore;
    double cloudLandScore;

    printf("\nRunning LC81730212013114LGN01 with the CFmask algorithm to "
           "verify the saturated pixel identification works as expected\n\n");

    /* Test using the CFMASK algorithm since it is the primary algorithm.
       All the algorithms produce the same result for the saturation masks */
    if (test_setup("LC81730212013114LGN01", "CFMASK", pathOdl, pathOmf,
        pathRefImage, pathRefQB, pathNewQB, pathRefCfmaskBand,
        pathNewCfmaskBand) != SUCCESS)
    {
        return;
    }
    if (run_cloud_cover(pathOdl) != SUCCESS)
        return;
    if (read_omf(pathOmf, &cloudScore, &snowIceScore, &cloudLandScore)
        != SUCCESS)
    {
        return;
    }

    assert_double_equal_with_message(cloudScore, kExpectedScore,
        "Cloud score %3.2lf != %3.2lf", cloudScore, kExpectedScore);
    assert_double_equal_with_message(snowIceScore, kExpectedSnowIceScore,
        "Snow/Ice score %3.2lf != %3.2lf", snowIceScore, kExpectedSnowIceScore);
    assert_double_equal_with_message(cloudLandScore, kExpectedCloudLandScore,
        "Cloud Land score %3.2lf != %3.2lf", cloudLandScore,
        kExpectedCloudLandScore);

    /* Get the reference directory to allow checking the saturation masks */
    strcpy(ref_dir, pathRefQB);
    temp_ptr = strrchr(ref_dir, '/');
    if (!temp_ptr)
    {
        assert_true_with_message(0, "Unexpected path format");
        return;
    }
    *temp_ptr = '\0';

    /* Build the saturation mask reference locations  */
    snprintf(ref_l1r_sat, sizeof(ref_l1r_sat), "%s/l1r_saturation_mask.img",
        ref_dir);
    snprintf(ref_l1g_sat, sizeof(ref_l1g_sat), "%s/l1g_saturation_mask.img",
        ref_dir);

    /* Compare the saturation masks */
    if (run_diff("l1r_saturation_mask.img", ref_l1r_sat, 1) != SUCCESS)
        return;
    if (run_diff("l1g_saturation_mask.img", ref_l1g_sat, 1) != SUCCESS)
        return;

    /* Make sure the quality band is as expected to confirm merging the
       saturation bits into the quality band works as expected */
    if (run_image_diff(pathRefQB, pathNewQB) != SUCCESS)
        return;
    unlink(pathNewQB);
    unlink("l1r_saturation_mask.img");
    unlink("l1g_saturation_mask.img");

    /* Now process the same scene again with the SEE5 algorithm and compare
       the saturation masks to make sure the different algorithm gets the same
       results since the buffering works a little differently for different
       algorithms.  Note: don't really care about the cloud scores or the
       quality band for this phase of testing since the SEE5 algorithm is
       covered by other tests. */
    printf("\nRunning LC81730212013114LGN01 with the SEE5 algorithm to "
           "verify the saturated pixel identification works as expected\n\n");

    if (test_setup("LC81730212013114LGN01", "SEE5", pathOdl, pathOmf,
        pathRefImage, pathRefQB, pathNewQB, pathRefCfmaskBand,
        pathNewCfmaskBand) != SUCCESS)
    {
        return;
    }
    if (run_cloud_cover(pathOdl) != SUCCESS)
        return;

    /* Compare the saturation masks */
    if (run_diff("l1r_saturation_mask.img", ref_l1r_sat, 1) != SUCCESS)
        return;
    if (run_diff("l1g_saturation_mask.img", ref_l1g_sat, 1) != SUCCESS)
        return;
    unlink(pathNewQB);
    unlink("l1r_saturation_mask.img");
    unlink("l1g_saturation_mask.img");
}

int main (int argc, char * argv[])
{
    TestSuite * suite  = create_test_suite();
    int         status;

    add_test(suite, LC80160502014041LGN00_Cfmask);
    add_test(suite, LO81390452001251SGS00);
    add_test(suite, LT81510282000189SGS00);
    add_test(suite, LC80160042013118LGN00);
    add_test(suite, LC81380162001164SGS01_ACCA);
    add_test(suite, LC81380162001164SGS01_ATACCA);
    add_test(suite, LC81380162001164SGS01_See5);
    add_test(suite, LC81380162001164SGS01_Cfmask);
    add_test(suite, LC81730212013114LGN01);
    if (argc > 1)
        status = run_single_test(suite, argv[1], create_text_reporter());
    else
        status = run_test_suite(suite, create_text_reporter());
    if (status)
        return EXIT_FAILURE;
    return EXIT_SUCCESS;
}
