/*==============================================================================
NAME: Update Order Metadata File (OMF)

PURPOSE:
    This function updates the work order's OMF file with the application's
    results.
==============================================================================*/

/* System Includes */
#include <string.h>
#include <limits.h>

/* IAS Includes */
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "ias_satellite_attributes.h"

/* Local Includes */
#include "assess_cloud_cover.h"
#include "get_names.h"
#include "get_parameters.h"

/*------------------------------------------------------------------------------
NAME: update_omf

PURPOSE:
    Write this application's algorithm information, final cloud cover score
    and final snow/ice score for the image to the work order's OMF file.

RETURNS:
    This function returns SUCCESS if the OMF file is successfully updated, ERROR
    if an error is encountered.
------------------------------------------------------------------------------*/
int update_omf
(
    const char pathOmf[],          /* I: Path to the OMF */
    char pathOut[],                /* I: Path quality band image */
    double cloudScore,             /* I: Cloud score for entire image */
    double snowIceScore,           /* I: Snow/Ice score for entire image */
    double cloudCoverLandScore,    /* I: Land cloud cover score */
    int band_saturation_count[IAS_MAX_NBANDS] /* I: Band saturation counts */
)
{
    short  i;
    short  nArgs                                  = 5;
    int saturated_bands[IAS_MAX_NBANDS];
    int total_bands_saturated = 0;
    int write_count;

    /* Set up the structure needed to update the OMF parameters */
    IAS_PARM_DECLARE_TABLE(xxxOmf, nArgs);

    IAS_PARM_CLOUD_COVER_SCORE(xxxOmf, &cloudScore, sizeof(cloudScore));
    IAS_PARM_SET_COUNT_WRITE(xxxOmf, CLOUD_COVER_SCORE, 1);

    IAS_PARM_CLOUD_COVER_LAND_SCORE(xxxOmf, &cloudCoverLandScore,
        sizeof(cloudCoverLandScore));
    IAS_PARM_SET_COUNT_WRITE(xxxOmf, CLOUD_COVER_LAND_SCORE, 1);
    
    IAS_PARM_SNOW_ICE_SCORE(xxxOmf, &snowIceScore, sizeof(snowIceScore));
    IAS_PARM_SET_COUNT_WRITE(xxxOmf, SNOW_ICE_SCORE, 1);

    IAS_PARM_QUALITY_BAND_FILENAME(xxxOmf, IAS_PARM_REQUIRED, pathOut,
        PATH_MAX);
    IAS_PARM_SET_COUNT_WRITE(xxxOmf, QUALITY_BAND_FILENAME, 1);

    /* Convert the band list count to a list of band numbers with saturation */
    for (i = 0; i < IAS_MAX_NBANDS; i++)
    {
        if (band_saturation_count[i] > 0)
        {
            int band_number = ias_sat_attr_convert_band_index_to_number(i);
            if (band_number == ERROR)
            {
                IAS_LOG_ERROR("Unable to convert band index %d to band number",
                    i);
                return ERROR;
            }
            saturated_bands[total_bands_saturated++] = band_number;
        }
    }

    /* Set the first band to -1 to indicate no saturation */
    if (total_bands_saturated < 1)
    {
        saturated_bands[total_bands_saturated++] = -1;
    }

    IAS_PARM_SATURATED_BANDS_LIST(xxxOmf, IAS_PARM_OPTIONAL, saturated_bands,
        sizeof(saturated_bands), 1);
    IAS_PARM_SET_COUNT_WRITE(xxxOmf, SATURATED_BANDS_LIST,
        total_bands_saturated);

    /* Don't output the bands if saturation is turned off */
    if (!get_enable_saturated_pixels_flag())
    {
        IAS_PARM_POP_LAST_PARAMETER(xxxOmf);
    }

    /* Get the number of parameters to write */
    write_count = IAS_PARM_GET_TABLE_COUNT(xxxOmf);

    /* Write our parameters to the OMF */
    if (ias_parm_update(pathOmf, xxxOmf, write_count) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing parameters to %s", pathOmf);
        return ERROR;
    }
    return SUCCESS;
}
