/*****************************************************************************
 NAME: apply_sun_angle_correction (main)

 PURPOSE: Applies the sun angle correction to the input L1G image or
          reverses it. Optionally converts the TIRS bands to temperature.

 RETURN VALUE: Type = int
    EXIT_SUCCESS / EXIT_FAILURE
*****************************************************************************/
/* Standard C Includes */
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>

/* IAS Library Includes */
#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "ias_math.h"
#include "ias_l1g.h"
#include "ias_miscellaneous.h"
#include "ias_angle_gen.h"

/* Local Includes */
#include "apply_sun_angle_correction.h"

int main
(
    int argc,
    char *argv[]
)
{
    SUN_ANGLE_CORRECTION_PARAMETERS parameters; /* Holds the input parameters */
    L1GIO *input_l1g = NULL;     /* Input L1G image I/O */
    L1GIO *output_l1g = NULL;    /* Ouput L1G image I/O */
    L1GIO *sun_angle_l1g = NULL; /* Sun angle correction image I/O */
    IAS_L1G_FILE_METADATA file_metadata;/* L1G File metadata */
    short *angle_band = NULL;       /* Sun angle correction data buffer */
    int band_index;                 /* Band loop variable */
    int processing_error = FALSE;   /* Processing error flag */
    double angle_band_scaling_factor = 0; /* Angle band scaling factor */

    /* Initialize the logging */
    if (ias_log_initialize("Apply Sun Angle Correction") != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the logging library");
        return EXIT_FAILURE;
    }

    /* Verify that the parameter filename was provided */
    if (argc != 2)
    {
        IAS_LOG_ERROR("Usage: apply_sun_angle_correction <parameter filename>");
        return EXIT_FAILURE;
    }

    /* Read the parameters from the parameter file */
    if (read_parameters(argv[1], &parameters) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading input parameters from %s", argv[1]);
        return EXIT_FAILURE;
    }

    /* Notify the user that the sun angle correction is being reversed */
    if (parameters.reverse_sun_correction_flag)
    {
        IAS_LOG_INFO("Reversing the sun angle correction for file %s", 
            parameters.l1g_filename);
    }

    /* Open the input L1G image */
    input_l1g = ias_l1g_open_image(parameters.l1g_filename, IAS_READ);
    if (!input_l1g)
    {
        IAS_LOG_ERROR("Opening input L1G image %s", parameters.l1g_filename);
        return EXIT_FAILURE;
    }

    /* Retrieve the L1G file metadata */
    if (ias_l1g_get_file_metadata(input_l1g, &file_metadata) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading the file metadata from %s", 
            parameters.l1g_filename);
        ias_l1g_close_image(input_l1g);
        return EXIT_FAILURE;
    }

    /* If the scene center correction is being performed then validate the
       scene center sun elevation angle */
    if (parameters.sun_angle_correction == SAC_SCENE_CENTER && 
        file_metadata.sun_elevation 
        < parameters.minimum_sun_elevation_angle_threshold)
    {
        IAS_LOG_ERROR("Scene center sun elevation angle of %.2f falls below "
            "the minimum threshold of %.2f", file_metadata.sun_elevation,
            parameters.minimum_sun_elevation_angle_threshold);
        ias_l1g_close_image(input_l1g);
        return EXIT_FAILURE;
    }

    /* Open the output L1G image */
    output_l1g = ias_l1g_open_image(parameters.l1g_output_filename,
        IAS_WRITE);
    if (!output_l1g)
    {
        IAS_LOG_ERROR("Opening output L1G imagery %s",
            parameters.l1g_output_filename);
        ias_l1g_close_image(input_l1g);
        return EXIT_FAILURE;
    }

    /* Open the angle band L1G if per pixel correction will be performed */
    if (parameters.sun_angle_correction == SAC_PER_PIXEL)
    {
        sun_angle_l1g = ias_l1g_open_image(
            parameters.sun_angle_correction_filename, IAS_READ);
        if (!sun_angle_l1g)
        {
            IAS_LOG_ERROR("Opening the angle band image %s", 
                parameters.sun_angle_correction_filename);
            ias_l1g_close_image(input_l1g);
            ias_l1g_close_image(output_l1g);
            return EXIT_FAILURE;
        }
    }

    /* Loop through all the bands that need to be processed */
    for(band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        float *output_buffer = NULL; /* L1G output buffer */
        IAS_DATA_TYPE data_type;     /* Input L1G data type */ 
        int number_of_scas;          /* Number of SCAs read */
        int line_count;              /* Number of lines read */
        int sample_count;            /* Number of samples read */
        const IAS_BAND_ATTRIBUTES *band_attributes; /* Band attributes */
        IAS_L1G_BAND_METADATA band_metadata; /* Band metadata */
        int band_number;             /* Current band number */
        int buffer_size;             /* L1G buffer size */

        /* Check if band should be processed */
        if (!parameters.process_band[band_index])
            continue;

        /* Convert the band index to band number */
        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Converting band index %d to band number",
                band_index);
            processing_error = TRUE;
            break;
        }

        /* Get the band attributes for the current band */
        band_attributes = ias_sat_attr_get_band_attributes(band_number);
        if (!band_attributes)
        {
            IAS_LOG_ERROR("Getting band attributes for band number %d",
                band_number);
            processing_error = TRUE;
            break;
        }
        
        /* Get the current band metadata */
        if (ias_l1g_get_band_metadata(input_l1g, band_number, &band_metadata)
            != SUCCESS)
        {
            IAS_LOG_ERROR("Reading metadata for band number %d", band_number);
            processing_error = TRUE;
            break;
        }

        /* Get the data for the current band from the input L1G */
        output_buffer = read_l1g_band(input_l1g, CHECK_DATA_TYPE, 
            band_number, 0, CONVERT_TO_FLOAT, &number_of_scas, &line_count, 
            &sample_count, &data_type);
        if (!output_buffer)
        {
            IAS_LOG_ERROR("Reading input image for band number %d", 
                band_number);
            processing_error = TRUE;
            break;
        }
        buffer_size = line_count * sample_count;
        
        /* Apply the sun angle correction to the OLI bands and if requested
           convert the TIRS bands to temperature */
        if (band_attributes->sensor_id == IAS_OLI)
        {
            double gain = band_metadata.reflectance_scaling_gain;
            double bias = band_metadata.reflectance_scaling_offset;

            IAS_LOG_INFO("Performing sun angle correction for band number %d",
                band_number);

            /* If a single band has been specified only read in the angle 
               band once for per-pixel calculation otherwise read in the 
               corresponding angle band for each band */
            if (!angle_band && parameters.sun_angle_correction == SAC_PER_PIXEL)
            {   
                IAS_L1G_BAND_METADATA angle_band_metadata; /* Band metadata */
                int band_to_read = band_number;

                /* Check if single angle band was specified */
                if (parameters.sun_angle_band != 0)
                {
                    band_to_read = parameters.sun_angle_band;
                }
                
                /* Read the angle band */
                angle_band = read_angle_band(sun_angle_l1g, band_to_read,
                    line_count, sample_count);
                if (!angle_band)
                {
                    IAS_LOG_ERROR("Reading angle band number %d", band_to_read);
                    free(output_buffer);
                    processing_error = TRUE;
                    break;
                }
                
                /* Get the current band metadata */
                if (ias_l1g_get_band_metadata(sun_angle_l1g, band_number, 
                    &angle_band_metadata) != SUCCESS)
                {
                    IAS_LOG_ERROR("Reading metadata for band number %d",
                        band_number);
                    processing_error = TRUE;
                    break;
                }

                /* Set the angle band scaling factor */
                angle_band_scaling_factor 
                    = angle_band_metadata.spectral_radiance_scaling_gain;
            }

            /* Scale the image to reflectance before applying the sun 
               angle correction */
            if (data_type != IAS_R4 || !parameters.reverse_sun_correction_flag)
            {
                scale_to_reflectance(gain, bias, DN_TO_REFLECTANCE_SCALING, 
                    buffer_size, output_buffer);
            }

            if (perform_sun_angle_correction(
                parameters.reverse_sun_correction_flag, 
                parameters.sun_angle_correction, file_metadata.sun_elevation, 
                angle_band_scaling_factor, 
                parameters.minimum_sun_elevation_angle_threshold, buffer_size,
                angle_band, output_buffer) != SUCCESS)
            {
                IAS_LOG_ERROR("Performing sun angle correction for band number "
                    "%d", band_number);  
                free(output_buffer);
                processing_error = TRUE;
                break;             
            }   
   
            /* Check if scaling factors should be overridden */
            if (parameters.override_scaling_factors_flag)
            {
                gain = parameters.reflectance_scaling_gain_override[band_index];
                bias = parameters.reflectance_scaling_bias_override[band_index];
                band_metadata.reflectance_scaling_gain = gain;
                band_metadata.reflectance_scaling_offset = bias;
            }
            IAS_LOG_INFO("Scaling using gain %lf and offset %lf", gain, bias);

            /* Scale back to DN if output data type is not float */
            if (parameters.output_data_type != IAS_R4)
            {
                scale_to_reflectance(gain, bias, REFLECTANCE_TO_DN_SCALING,
                    buffer_size, output_buffer);

                /* If reversing the sun angle correction with a floating
                   point reflectance input. Update the radiance rescaling 
                   parameters back to be relative to DN */
                if (data_type == IAS_R4
                    && parameters.reverse_sun_correction_flag)
                {
                    band_metadata.spectral_radiance_scaling_gain *= 
                        band_metadata.reflectance_scaling_gain;

                    band_metadata.spectral_radiance_scaling_offset += 
                        ((band_metadata.reflectance_scaling_offset 
                        * band_metadata.spectral_radiance_scaling_gain) 
                        / band_metadata.reflectance_scaling_gain);
                }
            }
            else
            {
                /* The output will be in floating point reflectance so
                   the radiance rescaling factors will need to be updated
                   to be relative to reflectance */
                band_metadata.spectral_radiance_scaling_offset = 
                    band_metadata.spectral_radiance_scaling_offset
                    - ((band_metadata.reflectance_scaling_offset 
                    * band_metadata.spectral_radiance_scaling_gain) 
                    / band_metadata.reflectance_scaling_gain);

                band_metadata.spectral_radiance_scaling_gain = 
                    band_metadata.spectral_radiance_scaling_gain 
                    / band_metadata.reflectance_scaling_gain;
            }

            /* If a single angle band was specified to be applied to 
               all bands then don't free up the angle band. Otherwise
               free up the angle band so the next angle band can be read */
            if (parameters.sun_angle_correction == SAC_PER_PIXEL 
                && parameters.sun_angle_band == 0)
            {
                free(angle_band);
                angle_band = NULL;
            }        
        }
        else if (parameters.convert_tirs_to_temperature_flag)
        {
            double gain = band_metadata.spectral_radiance_scaling_gain;
            double bias = band_metadata.spectral_radiance_scaling_offset;
            int tirs_index; /* TIRS thermal constants index */
                
            /* Retrieve the TIRS thermal constant index */
            tirs_index = 
                ias_sat_attr_convert_band_number_to_sensor_band_index(
                band_attributes->sensor_id, band_number);
            if (tirs_index == ERROR)
            {
                IAS_LOG_ERROR("Converting band number %d to sensor band index",
                    band_number);
                free(output_buffer);
                processing_error = TRUE;
                break;  
            }
            
            IAS_LOG_INFO("Converting band number %d to temperature",
                band_number);

            /* Convert the TIRS bands to temperature */
            convert_to_temperature(gain, bias, 
                parameters.tirs_constants.k1_constant[tirs_index], 
                parameters.tirs_constants.k2_constant[tirs_index],
                parameters.reverse_sun_correction_flag, buffer_size, 
                output_buffer);
        }

        /* Write the band to file */
        if (write_l1g_band(output_l1g, band_number, 
            parameters.output_data_type, number_of_scas, line_count,
            sample_count, output_buffer) != SUCCESS)  
        {
            IAS_LOG_ERROR("Writing band number %d", band_number);
            free(output_buffer);
            processing_error = TRUE;
            break;  
        }
        free(output_buffer);
        output_buffer = NULL;

        /* Output the band metadata */
        if (ias_l1g_set_band_metadata(output_l1g, &band_metadata, 1) 
            != SUCCESS)
        {
            IAS_LOG_ERROR("Setting the band metadata for band number %d",
                band_number);
            processing_error = TRUE;
            break; 
        }
    }
    
    /* Check for processing errors */
    if (processing_error)
    {
        ias_l1g_close_image(input_l1g);
        ias_l1g_close_image(output_l1g);
        if (sun_angle_l1g)
            ias_l1g_close_image(sun_angle_l1g);
        free(angle_band);
        return EXIT_FAILURE;
    }

    /* Close the angle band image if it was used */
    if (parameters.sun_angle_correction == SAC_PER_PIXEL)
    {
        /* Free up the angle band */
        free(angle_band);
        angle_band = NULL;

        if (ias_l1g_close_image(sun_angle_l1g) != SUCCESS)
        {
            IAS_LOG_ERROR("Closing input image");
            ias_l1g_close_image(output_l1g);
            ias_l1g_close_image(input_l1g);
            return EXIT_FAILURE;
        }
    }

    /* Close the input L1G */
    if (ias_l1g_close_image(input_l1g) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing input image");
        ias_l1g_close_image(output_l1g);
        return EXIT_FAILURE;
    }

    /* Update the sun angle correction metadata field */
    if (parameters.reverse_sun_correction_flag)
    {
        strncpy(file_metadata.sun_angle_correction, "NONE",
            sizeof(file_metadata.sun_angle_correction));
    }
    else if (parameters.sun_angle_correction == SAC_PER_PIXEL)
    {
        strncpy(file_metadata.sun_angle_correction, "PER_PIXEL",
            sizeof(file_metadata.sun_angle_correction));
    }
    else
    {
        strncpy(file_metadata.sun_angle_correction, "SCENE_CENTER", 
            sizeof(file_metadata.sun_angle_correction));
    }
    file_metadata.sun_angle_correction[
        sizeof(file_metadata.sun_angle_correction) - 1] = '\0';
 
    /* Set the output file metadata */
    if (ias_l1g_set_file_metadata(output_l1g, &file_metadata) != SUCCESS)
    {
        IAS_LOG_ERROR("Setting the L1G file metadata");
        ias_l1g_close_image(output_l1g);
        return EXIT_FAILURE;
    }
        
    /* Close the output image */
    if (ias_l1g_close_image(output_l1g) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing output image");
        return EXIT_FAILURE;
    }

    /* Update the OMF */
    if (update_omf(&parameters) != SUCCESS)
    {
        IAS_LOG_ERROR("Updating the OMF");
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}
