#ifndef APPLY_SUN_ANGLE_CORRECTION_H
#define APPLY_SUN_ANGLE_CORRECTION_H

/* Standard Library Includes */
#include <limits.h>

/* IAS Library Includes */
#include "ias_types.h"
#include "ias_satellite_attributes.h"
#include "ias_l1g.h"
#include "ias_cpf.h"

/* Local Defines */
#define NO_CHECK 0
#define NO_CONVERT 0
#define REFLECTANCE_TO_DN_SCALING 0
#define DN_TO_REFLECTANCE_SCALING 1
#define CONVERT_TO_FLOAT 1
#define CHECK_DATA_TYPE 1
#define SCALED_FILL_PIXEL -55000
#define CHECK_SCALED_FILL_PIXEL SCALED_FILL_PIXEL + 1

/* Correction method enumeration */
typedef enum sun_angle_correction_method
{
    SAC_UNKNOWN = 0,
    SAC_PER_PIXEL, 
    SAC_SCENE_CENTER
} SUN_ANGLE_CORRECTION_METHOD;


typedef struct sun_angle_correction_parameters
{
    int override_scaling_factors_flag;
    double reflectance_scaling_gain_override[IAS_MAX_NBANDS];
    double reflectance_scaling_bias_override[IAS_MAX_NBANDS];
    IAS_DATA_TYPE output_data_type;   /* Output data type */
    int process_band[IAS_MAX_NBANDS]; /* Array of flags to indicate if the
                                         band index should be processed */
    int convert_tirs_to_temperature_flag; /* Flag to indicate if the TIRS bands 
                                             should be converted to temp */
    int reverse_sun_correction_flag; /* Flag to indicate that the sun angle 
                                        correction should be reversed */
    int sun_angle_band; /* Band to use during per-pixel correction */
    char sun_angle_correction_filename[PATH_MAX]; /* Sun angle filename */
    char l1g_output_filename[PATH_MAX]; /* Filename of sun corrected L1G */
    char l1g_filename[PATH_MAX]; /* Input L1G filename */
    char cpf_filename[PATH_MAX]; /* CPF filename */
    int output_processing_pass; /* Output processing pass */
    SUN_ANGLE_CORRECTION_METHOD sun_angle_correction; /* Correction method */
    IAS_CPF_TIRS_THERMAL_CONSTANTS tirs_constants; /* Thermal constants */
    char omf_filename[PATH_MAX];   /* OMF filename */
    int processing_pass;           /* Parameter file processing pass */
    IAS_SATELLITE_ID satellite_id; /* Satellite ID */
    double minimum_sun_elevation_angle_threshold; /* Minimum sun elevation angle
                                                     threshold */
} SUN_ANGLE_CORRECTION_PARAMETERS;

/*************************** PROTOTYPES START *********************************/
void *convert_data_buffer
(
    int buffer_size,                /* I: Number of elements in buffers */
    IAS_DATA_TYPE input_data_type,  /* I: Input data type */
    IAS_DATA_TYPE output_data_type, /* I: Conversion data type */
    const void *data_to_convert,    /* I: Data buffer to convert */
    int *pixels_clamped             /* O: Number of pixels clamped */
);

void convert_to_temperature
(
    double gain,           /* I: Radiance rescale gain */
    double bias,           /* I: Radiance rescale bias */
    double k1_constant,    /* I: Thermal conversion constant K1 */
    double k2_constant,    /* I: Thermal conversion constant K2 */
    int reverse_conversion,/* I: Flag to indicate reverse the conversion */
    int buffer_size,       /* I: Number of elements in buffer */
    float *image_buffer    /* I/O: Image buffer */
);

int perform_sun_angle_correction
(
    int reverse_correction,        /* I: Correction reversal flag */
    SUN_ANGLE_CORRECTION_METHOD correction_method, /* I: Correction method */
    double scene_center_sun_angle, /* I: Scene center sun angle (degrees) */
    double angle_scaling_parameter,/* I: Angle band scaling factor */
    double min_sun_elevation_angle,/* I: Minimum sun elevation angle (degrees)*/
    int buffer_size,               /* I: Size of the input and output buffers */
    const short *angle_buffer,     /* I: Sun zenith angle correction buffer */
    float *output_buffer           /* I/O: Output image buffer */
);

short *read_angle_band
(
    L1GIO *sun_angle_l1g_io, /* I: Sun L1G image I/O */
    int band_to_read,        /* I: Angle band to read */
    int number_of_lines,     /* I: Verification number of lines */
    int number_of_samps      /* I: Verification number of samps */
);

void *read_l1g_band
(
    L1GIO *l1g_io,            /* I: L1G image I/O */
    int check_datatype_flag,  /* I: Flag to verify valid datatype */
    int band_number,          /* I: Band number to read */
    int sca_number,           /* I: SCA number to read */
    int convert_to_float_flag,/* I: Flag to convert output to float */
    int *number_of_scas,      /* O: Number of SCAs read from band */
    int *number_of_lines,     /* O: Number of lines read from band */
    int *number_of_samps,     /* O: Number of samps read from band */
    IAS_DATA_TYPE *data_type  /* O: Data type read from band */
);

int read_parameters
(
    const char *parameter_filename,             /* I: Parameter file name */
    SUN_ANGLE_CORRECTION_PARAMETERS *parameters /* O: Application parameters */
);

void scale_to_reflectance
(
    double gain,        /* I: Reflectance rescale gain */
    double bias,        /* I: Reflectance rescale bias */
    int reverse_scaling,/* I: Flag to indicate reverse the scaling */
    int buffer_size,    /* I: Number of elements in buffer */
    float *image_buffer /* I/O: Buffer to scale */
);

int update_omf
(
    SUN_ANGLE_CORRECTION_PARAMETERS *parameters /* I: Sun angle parameters */
);

int write_l1g_band
(
    L1GIO *l1g_io,          /* I: Output L1G I/O */
    int band_number,        /* I: Band number to output */
    IAS_DATA_TYPE data_type,/* I: Output data type */
    int number_of_scas,     /* I: Number of output SCAs */
    int number_of_lines,    /* I: Number of output lines */
    int number_of_samps,    /* I: Number of output samps */
    float *output_buffer    /* I: Output data buffer */
);

#endif /* APPLY_SUN_ANGLE_CORRECTION_H */

