/*****************************************************************************
NAME: convert_data_buffer

PURPOSE: Converts the input data buffer to the specified output data type and
         returns a pointer to the converted buffer

NOTE: The caller is responsible for freeing the returned data buffer.

RETURN VALUE: Type = int
    On Success: Returns void pointer to converted data buffer
    On Error:   Returns NULL
*****************************************************************************/
/* Standard C Includes */
#include <stdlib.h>
#include <math.h>

/* IAS Library Includes */
#include "ias_logging.h"
#include "ias_miscellaneous.h"

/* Local Includes */
#include "apply_sun_angle_correction.h"

void *convert_data_buffer
(
    int buffer_size,                /* I: Number of elements in data buffer */
    IAS_DATA_TYPE input_data_type,  /* I: Input data type */
    IAS_DATA_TYPE output_data_type, /* I: Conversion data type */
    const void *data_to_convert,    /* I: Data buffer to convert */
    int *pixels_clamped             /* O: Number of pixels clamped */
)
{
    int index;                    /* Generic loop variable */
    int data_type_size;           /* Converted data size */
    void *converted_data;         /* Converted data */
    int local_pixels_clamped = 0; /* Number of pixels clamped */

    /* Check if no conversion is needed */
    if (input_data_type == output_data_type)
    {
        IAS_LOG_ERROR("Datatypes are of the same type no conversion needed");
        return NULL;
    }

    /* Get the datatype size */
    if (ias_misc_get_sizeof_data_type(output_data_type, &data_type_size)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Getting data size for the conversion data type");
        return NULL;
    }

    /* Allocate the conversion buffer */
    converted_data = (void *)malloc(data_type_size * buffer_size);
    if (!converted_data)
    {
        IAS_LOG_ERROR("Allocating memory for conversion buffer");
        return NULL;
    }
  
    /* Convert the buffer from the input to output data type */
    if (input_data_type == IAS_R4 && output_data_type == IAS_UI2)
    {
        unsigned short *output = (unsigned short *)converted_data;
        float *input = (float *)data_to_convert;

        for (index = 0; index < buffer_size; index++)
        {
            if (input[index] < 0)
            {
                output[index] = 0;
                if (input[index] >= CHECK_SCALED_FILL_PIXEL)
                    local_pixels_clamped++;
            }
            else if (input[index] > USHRT_MAX)
            {
                output[index] = USHRT_MAX;
                local_pixels_clamped++;
            }
            else
            {
                output[index] = round(input[index]);
            }
        }
    }
    else if (input_data_type == IAS_UI2 && output_data_type == IAS_R4)
    {
        float *output = (float *)converted_data;
        unsigned short *input = (unsigned short *)data_to_convert;

        for (index = 0; index < buffer_size; index++)
        {
            output[index] = input[index];
        }
    }
    else
    {
        IAS_LOG_ERROR("Unsupported conversion from input data type %d to "
            "output data type: %d", input_data_type, output_data_type);
        free(converted_data);
        return NULL;     
    }

    /* Output the number of clamped pixels */
    *pixels_clamped = local_pixels_clamped;

    return converted_data;
}
