/*****************************************************************************
NAME: convert_to_temperature

PURPOSE: Converts TIRS bands from spectral radiance to brightness temperature
         using the thermal constants.

RETURN VALUE: Type = int
    SUCCESS / ERROR
*****************************************************************************/
/* Standard C Includes */
#include <math.h>

/* Local Includes */
#include "apply_sun_angle_correction.h"

void convert_to_temperature
(
    double gain,           /* I: Radiance rescale gain */
    double bias,           /* I: Radiance rescale bias */
    double k1_constant,    /* I: Thermal conversion constant K1 */
    double k2_constant,    /* I: Thermal conversion constant K2 */
    int reverse_conversion,/* I: Flag to indicate reverse the conversion */
    int buffer_size,       /* I: Number of elements in buffer */
    float *image_buffer    /* I/O: Image buffer */
)
{
    int pixel; /* Pixel loop index */


    /* Convert the pixel to temperature or reverse it if requested */
    if (reverse_conversion)
    {
        double reverse_gain = 1.0 / gain;

        /* Apply the conversion to each pixel in the image */
        for (pixel = 0; pixel < buffer_size; pixel++)
        {
            float cur_pixel = image_buffer[pixel];

            /* Don't process the fill pixels when converting back to DN. 
               All fill pixels will be SCALED_FILL_PIXEL when converting
               back to DN */
            if (cur_pixel < CHECK_SCALED_FILL_PIXEL)
            {
                image_buffer[pixel] = 0.0;
                continue;
            }
            cur_pixel = k1_constant / (exp(k2_constant / cur_pixel) - 1);
            image_buffer[pixel]  = (cur_pixel - bias) * reverse_gain;
        }
    }
    else
    {
        /* Apply the conversion to each pixel in the image */
        for (pixel = 0; pixel < buffer_size; pixel++)
        {
            float cur_pixel = image_buffer[pixel];

            /* Dont process the fill pixels when converting to temperature
               All fill pixels will be 0.0 when converting to temperature. */
            if (cur_pixel == 0.0)
            {
                image_buffer[pixel] = SCALED_FILL_PIXEL;
                continue;
            }
            cur_pixel = (cur_pixel * gain) + bias;

            /* There is a chance that by adjusting for a paticular gain and bias
               the current pixel could be equal to exactly zero. This would 
               result in a bad temperature calculation. */
            if (cur_pixel == 0.0)
            {
                cur_pixel = 1;
            }
            else
            {
                cur_pixel = k2_constant / log((k1_constant / cur_pixel) + 1);
            }

            image_buffer[pixel] = cur_pixel;
        }
    }

}
