/*****************************************************************************
NAME: perform_sun_angle_correction

PURPOSE: Performs the "Per-Pixel" or "Scene Center" sun angle correction.

NOTE: The input is expected to be scaled to reflectance

RETURN VALUE: Type = int
    SUCCESS / ERROR
*****************************************************************************/
/* Standard C Includes */
#include <math.h>

/* IAS Library Includes */
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_math.h"

/* Local Includes */
#include "apply_sun_angle_correction.h"

int perform_sun_angle_correction
(
    int reverse_correction,        /* I: Correction reversal flag */
    SUN_ANGLE_CORRECTION_METHOD correction_method, /* I: Correction method */
    double scene_center_sun_angle, /* I: Scene center sun angle (degrees) */
    double angle_scaling_parameter,/* I: Angle band scaling factor */
    double min_sun_elevation_angle,/* I: Minimum sun elevation angle (degrees)*/
    int buffer_size,               /* I: Size of the input and output buffers */
    const short *angle_buffer,     /* I: Sun zenith angle correction buffer */
    float *output_buffer           /* I/O: Output image buffer */
)
{
    double radians_per_degree = ias_math_get_radians_per_degree();
    double sun_correction;  /* Sun angle correction variable */
    int pixel;              /* Generic loop variable */ 

    /* Set the inverted scaling factor if per pixel correction */
    if (correction_method == SAC_PER_PIXEL)
    {
        double inverted_scaling_factor;
        double max_sun_zenith_angle = 90.0 - min_sun_elevation_angle;
        int zenith_angle_warning_issued = FALSE;

        if (angle_scaling_parameter == 0.0)
        {
            IAS_LOG_ERROR("Angle band scaling factor results in 0 divide");
            return ERROR;
        }
        inverted_scaling_factor = 1.0 / angle_scaling_parameter;

        /* Apply the correction to each pixel in the image */
        for (pixel = 0; pixel < buffer_size; pixel++)
        {
            double sun_zenith_angle; /* sun zenith angle in degrees */

            /* Don't process the fill pixels */
            if (output_buffer[pixel] < CHECK_SCALED_FILL_PIXEL)        
                continue;

            /* Calculate the solar zenith angle */
            sun_zenith_angle = angle_buffer[pixel] * inverted_scaling_factor;
            if (sun_zenith_angle > max_sun_zenith_angle)
            {
                if (!zenith_angle_warning_issued)
                {
                    zenith_angle_warning_issued = TRUE;
                    IAS_LOG_INFO("Per-pixel angle band contains sun elevation "
                        "angles below the minimum allowable threshold of %.2f",
                        min_sun_elevation_angle);
                }
                continue;
            }
            else if(sun_zenith_angle >= 90.0)
            {
                /* Do not correct angles that are in the dark */
                continue;
            }

            /* Calculate the sun_correction for the current pixel. If the 
               correction is equal to 0.0 then skip the correction otherwise
               perform the correction */
            sun_correction = cos(sun_zenith_angle * radians_per_degree);
            if (sun_correction == 0.0)
                continue;

            /* Check if the sun angle correction should be reversed and if not
               turn the division into multiplication */
            if (!reverse_correction)
            {
                sun_correction = 1.0 / sun_correction;
            }
            
            /* Adjust for the sun angle */
            output_buffer[pixel] *= sun_correction;
        }
    }
    else if (correction_method == SAC_SCENE_CENTER)
    {
        /* Calculate the sun_correction for the scene center elevation angle. 
           If the correction is equal to 0.0 then no correction will be made 
           on this band. Otherwise perform the correction */
        sun_correction = sin(scene_center_sun_angle * radians_per_degree);
        if (sun_correction == 0.0)
            return SUCCESS;

        /* Check if the sun angle correction should be reversed if not
           turn the division into multiplication */
        if (!reverse_correction)
        {
            sun_correction = 1.0 / sun_correction;
        }

        /* Apply the correction to each pixel in the image */
        for (pixel = 0; pixel < buffer_size; pixel++)
        {
            /* Don't process the fill pixels */
            if (output_buffer[pixel] < CHECK_SCALED_FILL_PIXEL)        
                continue;

            /* Adjust for the sun angle */
            output_buffer[pixel] *= sun_correction;
        }
    }

    return SUCCESS;
}
