/*****************************************************************************
NAME: read_angle_band

PURPOSE:  Reads a single band from angle band L1G and returns a pointer 
          to the data. 

RETURN VALUE: Type = short *
    On Success: Returns void * to input data
    On Error  : Returns NULL
*****************************************************************************/
/* IAS Library Includes */
#include "ias_logging.h"
#include "ias_types.h"
#include "ias_l1g.h"

/* Local Includes */
#include "apply_sun_angle_correction.h"

short *read_angle_band
(
    L1GIO *sun_angle_l1g_io, /* I: Sun L1G image I/O */
    int band_to_read,        /* I: Angle band to read */
    int number_of_lines,     /* I: Verification number of lines */
    int number_of_samps      /* I: Verification number of samps */
)
{
    IAS_DATA_TYPE angle_data_type; /* Data type read from band */
    int num_angle_scas;            /* Number of SCAs read from band */
    int num_angle_lines;           /* Number of lines read from band */
    int num_angle_samps;           /* Number of samps read from band */
    void *angle_band;              /* Local angle band */

    /* Read in the solar zenith angle band */
    angle_band = read_l1g_band(sun_angle_l1g_io, NO_CHECK, 
        band_to_read, 1, NO_CONVERT, &num_angle_scas, 
        &num_angle_lines, &num_angle_samps, &angle_data_type);
    if (!angle_band)
    {
        IAS_LOG_ERROR("Reading angle band for band number %d", band_to_read);
        return NULL;
    }

    /* Make consistency checks with input data */
    if (num_angle_lines != number_of_lines)
    {
        IAS_LOG_ERROR("Number of lines in angle band (%d) do not match input "
            "image (%d)", num_angle_lines, number_of_lines);
        return NULL;
    }
    else if (num_angle_samps != number_of_samps)
    {
        IAS_LOG_ERROR("Number of samples in angle band (%d) do not match input "
            "image (%d)", num_angle_samps, number_of_samps);
        return NULL;
    }
    else if (angle_data_type != IAS_I2)
    {
        IAS_LOG_ERROR("Invalid datatype found for angle band number %d "
            "expected IAS_I2", band_to_read);
        return NULL;
    }

    return angle_band;
}
