/*****************************************************************************
NAME: read_l1g_band

PURPOSE:  Reads a single band from L1G image and returns the pointer to the
          data. Optionally converts data to float and checks to make sure
          the input data type is valid.

RETURN VALUE: Type = void *
    On Success: Returns void * to input data
    On Error  : Returns NULL
*****************************************************************************/
/* Standard C Includes */
#include <stdlib.h>

/* IAS Library Includes */
#include "ias_logging.h"
#include "ias_types.h"
#include "ias_miscellaneous.h"
#include "ias_l1g.h"

/* Local Includes */
#include "apply_sun_angle_correction.h"

void *read_l1g_band
(
    L1GIO *l1g_io,            /* I: L1G image I/O */
    int check_datatype_flag,  /* I: Flag to verify valid datatype */
    int band_number,          /* I: Band number to read */
    int sca_number,           /* I: SCA number to read */
    int convert_to_float_flag,/* I: Flag to convert output to float */
    int *number_of_scas,      /* O: Number of SCAs read from band */
    int *number_of_lines,     /* O: Number of lines read from band */
    int *number_of_samps,     /* O: Number of samps read from band */
    IAS_DATA_TYPE *data_type  /* O: Data type read from band */
)
{
    L1G_BAND_IO *l1g_band_io; /* Current band I/O */
    void *output_buffer;      /* Return data buffer */
    int buffer_size;          /* Size of buffer to allocate */
    int data_type_size;       /* Size of data type in bytes */

    /* Open the band for reading */
    l1g_band_io = ias_l1g_open_band(l1g_io, band_number, data_type,
        number_of_scas, number_of_lines, number_of_samps);
    if (!l1g_band_io)
    {
        IAS_LOG_ERROR("Opening band number %d", band_number);
        return NULL;
    }
    buffer_size = (*number_of_lines) * (*number_of_samps);

    /* Check the expected data type */
    if (check_datatype_flag && *data_type != IAS_UI2 && *data_type != IAS_R4)
    {
        IAS_LOG_ERROR("Unexpected datatype for band number %d", band_number);
        ias_l1g_close_band(l1g_band_io);
        return NULL;
    }

    /* Get the datatype size and calculate the buffer size */
    if (ias_misc_get_sizeof_data_type(*data_type, &data_type_size)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Getting data size for band number %d", band_number);
        ias_l1g_close_band(l1g_band_io); 
        return NULL;
    }

    /* Allocate the space needed to read the band */
    output_buffer = (void *)malloc(data_type_size * buffer_size);
    if (!output_buffer)
    {
        IAS_LOG_ERROR("Allocating memory for band number %d", band_number);
        ias_l1g_close_band(l1g_band_io);
        return NULL;
    }

    /* Read the current band into memory */
    if (ias_l1g_read_image(l1g_band_io, sca_number, 0, 0, *number_of_lines,
        *number_of_samps, output_buffer) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading band number %d", band_number);
        ias_l1g_close_band(l1g_band_io);
        free(output_buffer);
        return NULL;
    }
    
    /* Close the band I/O */
    if (ias_l1g_close_band(l1g_band_io) != SUCCESS)    
    {
        IAS_LOG_ERROR("Closing band number %d", band_number);
        free(output_buffer);
        return NULL;
    }

    /* Check if data should be converted to float */
    if (convert_to_float_flag && *data_type != IAS_R4)
    {
        int pixels_clamped; /* Number of pixels clamped */
        float *converted_data = convert_data_buffer(buffer_size, *data_type, 
            IAS_R4, output_buffer, &pixels_clamped);
        if (!converted_data)
        {
            IAS_LOG_ERROR("Converting data to float for band %d", band_number);
            free(output_buffer);
            return NULL;
        }

        /* Free up the output buffer after the conversion */
        free(output_buffer);
        output_buffer = NULL;

        /* Set the new input data pointer */
        output_buffer = converted_data;
    }

    return output_buffer;
}
