/* Standard C Includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

/* IAS Includes */
#include "ias_const.h"
#include "ias_parm.h"
#include "ias_cpf.h"
#include "ias_satellite_attributes.h"
#include "ias_parm_standard_parameters.h"
#include "ias_miscellaneous.h"

/* Local Includes */
#include "apply_sun_angle_correction.h"

static char local_sun_angle_correction[13]; /* Sun angle correction string */
static char local_output_data_type[5];      /* Output data type string */
static int local_band_list[IAS_MAX_NBANDS]; /* Parameter file band list */ 
static int local_band_count; /* Number of bands present in local band list */

/******************************************************************************
NAME: process_parameters

PURPOSE: This function performs special processing on parameters which need
         to be translated after being read from the parameter and/or OMF files.

RETURN VALUE: Type = int
    Value     Description
    -----     -----------
    SUCCESS   The parameters were successfully processed
    ERROR     An error occurred processing the parameters
******************************************************************************/
int process_parameters
(
    SUN_ANGLE_CORRECTION_PARAMETERS *parameters /* I/O: Application parameters*/
)
{
    int band_already_in_list[IAS_MAX_NBANDS];
    int count;

    /* Initialize the satellite attributes library */
    if (ias_sat_attr_initialize(parameters->satellite_id) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the satellite attributes library");
        return ERROR;
    }

    /* Convert the string data type */
    if (ias_misc_convert_string_to_data_type(local_output_data_type, 
        &parameters->output_data_type) != SUCCESS)
    {
        IAS_LOG_ERROR("Converting datatype %s", local_output_data_type);
        return ERROR;
    }
    
    /* Check that the CPF filename was provided if converting TIRS bands 
       to temperature */
    if (parameters->convert_tirs_to_temperature_flag 
        && strlen(parameters->cpf_filename) <= 0)
    {
        IAS_LOG_ERROR("CAL_PARM_FILENAME is required when converting TIRS "
            "bands to temperature");
        return ERROR;
    }

    /* Get the CPF thermal constants if converting TIRS bands to temperature */
    if (parameters->convert_tirs_to_temperature_flag)
    {
        IAS_CPF *cpf; /* Input CPF */
        const IAS_CPF_TIRS_THERMAL_CONSTANTS *local_constants;

        /* Read the CPF */
        cpf = ias_cpf_read(parameters->cpf_filename);
        if (!cpf)
        {
            IAS_LOG_ERROR("Reading CPF %s", parameters->cpf_filename);
            return ERROR;
        }
        
        /* Read the thermal constants */
        local_constants = ias_cpf_get_tirs_thermal_constants(cpf);
        if (!local_constants) 
        {
            IAS_LOG_ERROR("Loading the TIRS thermal constants from the CPF");
            ias_cpf_free(cpf);
            return ERROR;
        }
        parameters->tirs_constants = *local_constants;

        /* Free the CPF */
        ias_cpf_free(cpf);
    }

    /* Store the correction type */
    if (strcmp("SCENE_CENTER", local_sun_angle_correction) == 0)
    {
        parameters->sun_angle_correction = SAC_SCENE_CENTER;
    }
    else
    {
        parameters->sun_angle_correction = SAC_PER_PIXEL;

        /* Verify the sun angle correction filename was provided in the OMF */
        if (strlen(parameters->sun_angle_correction_filename) <= 0)
        {
            IAS_LOG_ERROR("Sun angle correction filename not provided in OMF");
            return ERROR;
        }
    }

    /* Initialize the band present list */
    for (count = 0; count < IAS_MAX_NBANDS; count++)
    {
        parameters->process_band[count] = 0;
        band_already_in_list[count] = 0;
    }

    /* Determine which bands to process */
    for (count = 0; count < local_band_count; count++)
    {
        /* Convert the band number entered into a band index */
        int temp_band_index = ias_sat_attr_convert_band_number_to_index(    
            local_band_list[count]);
        if (temp_band_index == ERROR)
        {
            IAS_LOG_ERROR("Illegal band number %d", local_band_list[count]);
            return ERROR;
        }
        /* Make sure the band isn't already in the list */
        else if (!band_already_in_list[temp_band_index])
        {
            /* Put band in list and flag that it has been added 
               to the list */
            parameters->process_band[temp_band_index] = 1;
            band_already_in_list[temp_band_index] = 1;
        }
        else if (band_already_in_list[temp_band_index] == 1)
        {
             /* Only print the warning message once for each band */
             band_already_in_list[temp_band_index] += 1;
             IAS_LOG_WARNING("Warning: Band %d appears multiple times in the "
                "BAND_LIST, only the first one was used", 
                local_band_list[count]);
        }
    }

    return SUCCESS;
}

/******************************************************************************
NAME: read_omf_parameters

PURPOSE: This function reads all the parameters from the OMF file

RETURN VALUE: Type = int
    Value    Description
    -----    -----------
    SUCCESS  The parameters were successfully read
    ERROR    An error occurred reading the parameters
******************************************************************************/
int read_omf_parameters
(
    const char *omf_filename,                   /* I: OMF filename */
    SUN_ANGLE_CORRECTION_PARAMETERS *parameters /* O: Application parameters */
)
{
    int count;            /* Table parameter count */
    int help_requested;   /* Status placeholder */
    int satellite_number; /* OMF satellite number */

    /* Start parameter table definition */
    IAS_PARM_DECLARE_TABLE(omf_table, 4);

    IAS_PARM_L1G_FILENAME_FROM_PASS(omf_table, IAS_PARM_REQUIRED,
        parameters->l1g_filename, sizeof(parameters->l1g_filename), 1, 
        parameters->processing_pass);

    IAS_PARM_SUN_ANGLE_FILENAME_FROM_PASS(omf_table, IAS_PARM_OPTIONAL, 
        parameters->sun_angle_correction_filename,
        sizeof(parameters->sun_angle_correction_filename), 1, 
        parameters->processing_pass);

    IAS_PARM_CPF(omf_table, IAS_PARM_OPTIONAL, parameters->cpf_filename,
        sizeof(parameters->cpf_filename), 1);
 
    IAS_PARM_SATELLITE(omf_table, &satellite_number, 
        sizeof(satellite_number), 1);

    /* Retrieve parameter count for OMF table */
    count = IAS_PARM_GET_TABLE_SIZE(omf_table);

    /* Call the ias_parm_provide_help routine to provide help for the
       parameters if requested; also verifies the OMF file exists */
    help_requested = ias_parm_provide_help(omf_filename, omf_table, count,
        IAS_OMF_PARAMETERS);
    if (help_requested)
    {
        /* Either help was provided or an error occurred */
        if (help_requested == TRUE)
        {
            /* Successfully provided help, as requested, so just exit */
            exit(EXIT_SUCCESS);
        }
        
        IAS_LOG_ERROR("Error occurred providing OMF help");
        return ERROR;
    }

    /* Read the OMF file */
    if (ias_parm_read(omf_filename, omf_table, count) != SUCCESS)
    {
        IAS_LOG_ERROR("Error reading OMF parameter file %s", omf_filename);
        return ERROR;
    }
    
    /* Convert the satellite number into a satellite id */
    parameters->satellite_id 
        = ias_sat_attr_get_satellite_id_from_satellite_number(satellite_number);
    if (parameters->satellite_id == ERROR)
    {
        IAS_LOG_ERROR( "Unable to determine satellite ID "
            "from satellite number %d", satellite_number);
        return ERROR;
    }

    return SUCCESS;
}

/*****************************************************************************
 NAME: read_parameters

 PURPOSE: Read parameters from the specified input ODL file and OMF file.

 RETURN VALUE: Type = int
    Value    Description
    -------  --------------------------
    SUCCESS  Successful completion
    ERROR    Operation failed

*****************************************************************************/
int read_parameters
(
    const char *parameter_filename,             /* I: Parameter file name */
    SUN_ANGLE_CORRECTION_PARAMETERS *parameters /* O: Application parameters */
)
{
    int help_requested;                     /* Help request variable */
    int count;                              /* Table parameter count */
    int status;                             /* Status variable */
    char local_wo_id[IAS_WORKORDER_ID_SIZE];/* Work order ID */

    /* Zero the parameters */
    memset(parameters, 0, sizeof(SUN_ANGLE_CORRECTION_PARAMETERS));

    /* Parameters table definition */
    IAS_PARM_DECLARE_TABLE(parameter_table, 13);

    /* Get the work order ID */
    IAS_PARM_WORK_ORDER_ID(parameter_table, local_wo_id, 
        sizeof(local_wo_id), 1);

    IAS_PARM_PROCESSING_PASS(parameter_table, &parameters->processing_pass,
        sizeof(parameters->processing_pass), 1);

    IAS_PARM_BAND_LIST(parameter_table, IAS_PARM_REQUIRED, local_band_list, 
        sizeof(local_band_list), 1);

    IAS_PARM_SUN_ANGLE_CORRECTED_L1G_FILENAME(parameter_table, 
        IAS_PARM_REQUIRED, parameters->l1g_output_filename, 
        sizeof(parameters->l1g_output_filename), 1);

    const char *valid_correction_methods[] = {"SCENE_CENTER", "PER_PIXEL"};
    IAS_PARM_ADD_STRING(parameter_table, SUN_ANGLE_CORRECTION,
        "Type of sun angle correction to apply to the input image",
        IAS_PARM_REQUIRED, 2, valid_correction_methods, 0, NULL,
        local_sun_angle_correction, sizeof(local_sun_angle_correction), 1);

    int default_sun_angle_band[] = {0};
    IAS_PARM_ADD_INT(parameter_table, SUN_ANGLE_BAND, "Band to use for "
        "per-pixel sun angle correction. 0 indicates all bands will be "
        "corrected with their corresponding sun angle band. 1-9 indicates "
        "the specified band will be used to correct all bands", 
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 0, 9, 1, 
        default_sun_angle_band, &parameters->sun_angle_band,
        sizeof(parameters->sun_angle_band), 0);

    int default_reverse_correction_flag[] = {0};
    IAS_PARM_ADD_BOOLEAN(parameter_table, REVERSE_SUN_CORRECTION_FLAG, 
        "Flag to indicate that the sun correction will be reversed {0 = "
        "Perform correction; 1 = Reverse correction}", IAS_PARM_OPTIONAL,
        IAS_PARM_NOT_ARRAY, default_reverse_correction_flag, 
        &parameters->reverse_sun_correction_flag, 
        sizeof(parameters->reverse_sun_correction_flag), 0);

    int default_temperature_flag[] = {1};
    IAS_PARM_ADD_BOOLEAN(parameter_table, CONVERT_TIRS_TO_TEMP_FLAG, 
        "Flag to indicate that that the TIRS bands will be converted to "
        "temperature {0 = Do not convert; 1 = Convert to temperature}", 
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, default_temperature_flag, 
        &parameters->convert_tirs_to_temperature_flag,
        sizeof(parameters->convert_tirs_to_temperature_flag), 0);
  
    double default_angle_threshold[] = {0.0};
    IAS_PARM_ADD_DOUBLE(parameter_table, MINIMUM_SUN_ANGLE_THRESHOLD,
        "The minimum allowed sun elevation (in degrees) allowed to be "
        "corrected in the scene.  A sun elevation lower than this in the "
        "scene will cause a processing failure.", IAS_PARM_OPTIONAL,
        IAS_PARM_NOT_ARRAY, 0, 0, 0, 1, default_angle_threshold,
        &parameters->minimum_sun_elevation_angle_threshold,
        sizeof(parameters->minimum_sun_elevation_angle_threshold), 0);

    /* The following 3 parameters are added so that Cal/Val can play 
       with the gain and bias parameters for scenes that are around the
       poles and have higher solar elevation angles which results in
       clamped pixels. */ 
    int default_override_scaling_factors_flag[] = {0};
    IAS_PARM_ADD_BOOLEAN(parameter_table, OVERRIDE_SCALING_FACTORS_FLAG, 
        "Flag to indicate that that the multiplicative and additive scaling "
        "factors will be overridden {0 = Do not override scaling factors; "
        "1 = Override scaling factors}", IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        default_override_scaling_factors_flag, 
        &parameters->override_scaling_factors_flag,
        sizeof(parameters->override_scaling_factors_flag), 0);

    IAS_PARM_ADD_DOUBLE(parameter_table, REFLECTANCE_MULTIPLICATION_FACTORS, 
        "Reflectance multiplication factors for bands 1 - 9 to override the "
        "factors pulled from the input L1G", IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
        0, 0, 0, 0, NULL, parameters->reflectance_scaling_gain_override,
        sizeof(parameters->reflectance_scaling_gain_override), 0);

    IAS_PARM_ADD_DOUBLE(parameter_table, REFLECTANCE_ADDITIVE_FACTORS, 
        "Reflectance additive factors for bands 1 - 9 to override the "
        "factors pulled from the input L1G", IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
        0, 0, 0, 0, NULL, parameters->reflectance_scaling_bias_override,
        sizeof(parameters->reflectance_scaling_bias_override), 0);

    const char *valid_output_types[] = {"R*4", "UI*2"};
    const char *default_output_type[] = {"UI*2"};
    IAS_PARM_ADD_STRING(parameter_table, CORRECTION_OUTPUT_DATA_TYPE,
        "Output L1G data type", IAS_PARM_REQUIRED, 2, valid_output_types, 1,
        default_output_type, local_output_data_type,
        sizeof(local_output_data_type), 1);

    /* -------------------------------------------------------------------- */
    /* Read the parameter file */

    /* Calculate the number of items in the parameter file */
    count = IAS_PARM_GET_TABLE_SIZE(parameter_table);

    /* Call the ias_parm_provide_help routine to provide help for the
       parameters if requested; also verifies the ODL file exists */
    help_requested = ias_parm_provide_help(parameter_filename, parameter_table,
        count, IAS_INPUT_PARAMETERS);
    if (help_requested)
    {
        /* Either help was provided or an error occurred */
        if (help_requested == TRUE)
        {
            /* Provide help for the OMF.  If it succeeds, it will exit.  If
               it returns, an error occurred, so return an error */
            read_omf_parameters(parameter_filename, parameters);
            IAS_LOG_ERROR("Providing application parameter file help");
            return ERROR;
        }
        
        IAS_LOG_ERROR("Error occurred providing ODL help");
        return ERROR;
    }
  
    /* Read the parameter file */
    if (ias_parm_read(parameter_filename, parameter_table, count) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from: %s", parameter_filename);
        return ERROR;
    }
    
    /* Make sure override scaling factors are present if requested */
    if (parameters->override_scaling_factors_flag && (IAS_PARM_GET_COUNT_READ(
        parameter_table, REFLECTANCE_MULTIPLICATION_FACTORS) != 9 ||
        IAS_PARM_GET_COUNT_READ(parameter_table, REFLECTANCE_ADDITIVE_FACTORS) 
        != 9))
    {
        IAS_LOG_ERROR("Reading the scaling factors from parameter file %s",
            parameter_filename);
        return ERROR;
    }

    /* Build the OMF filename */
    status = snprintf(parameters->omf_filename,
        sizeof(parameters->omf_filename), "%s.omf", local_wo_id);
    if (status < 0 || status >= sizeof(parameters->omf_filename))
    {
        IAS_LOG_ERROR("Creating the OMF filename");
        return ERROR;
    }

    /* Read in the OMF parameters */
    if (read_omf_parameters(parameters->omf_filename, parameters) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from %s", parameters->omf_filename);
        return ERROR;
    }
    
    /* Save the band count for later */
    local_band_count = IAS_PARM_GET_COUNT_READ(parameter_table, BAND_LIST);

    /* Process all of the parameters */
    if (process_parameters(parameters) != SUCCESS)
    {
        IAS_LOG_ERROR("Processing the parameters");
        return ERROR;
    }

    return SUCCESS;
}
