/*****************************************************************************
NAME: scale_to_reflectance
 
PURPOSE: Scales the imagery to reflectance, if the scaling_type has been set
         to REFLECTANCE_TO_DN_SCALING the image will be scaled back.

RETURN VALUE: Type = int
    SUCCESS / ERROR
*****************************************************************************/
/* IAS Library Includes */
#include "ias_math.h"

/* Local Includes */
#include "apply_sun_angle_correction.h"

void scale_to_reflectance
(
    double gain,        /* I: Reflectance rescale gain */
    double bias,        /* I: Reflectance rescale bias */
    int scaling_type,   /* I: Flag to indicate scaling type */
    int buffer_size,    /* I: Number of elements in buffer */
    float *image_buffer /* I/O: Buffer to scale */
)
{
    int pixel;                        /* Pixel loop index */

    /* Rescale the pixel */
    if (scaling_type == DN_TO_REFLECTANCE_SCALING)
    {
        /* Scale each image pixel from DN to reflectance */
        for (pixel = 0; pixel < buffer_size; pixel++)
        {
            float cur_pixel = image_buffer[pixel];

            /* Dont process the fill pixels when converting to reflectance.
               All fill pixels will be 0.0 when scaling to reflectance. */
            if (cur_pixel == 0.0)
            {
                image_buffer[pixel] = SCALED_FILL_PIXEL;
                continue;
            }
            image_buffer[pixel] = (cur_pixel * gain) + bias;
        }
    }
    else
    {
        double reverse_gain = 1.0 / gain; /* Invert the gain so division turns
                                             into multiplication */

        /* Scale each image pixel from reflectance to DN */
        for (pixel = 0; pixel < buffer_size; pixel++)
        {
            float cur_pixel = image_buffer[pixel];

            /* Don't process the fill pixels when converting back to DN. 
               All fill pixels will be SCALED_FILL_PIXEL when scaled
               to reflectance */
            if (cur_pixel < CHECK_SCALED_FILL_PIXEL)
            {
                image_buffer[pixel] = 0.0;
                continue;
            }
            image_buffer[pixel] = (cur_pixel - bias) * reverse_gain;
        }
    }
}
