/*****************************************************************************
NAME: write_l1g_band
 
PURPOSE: Write out the band to the L1G image. If the output data type is not
         float then the band will be converted.

RETURN VALUE: Type = int
    SUCCESS / ERROR
*****************************************************************************/
/* Standard C Includes */
#include <stdlib.h>

/* IAS Library Includes */
#include "ias_logging.h"
#include "ias_types.h"
#include "ias_l1g.h"

/* Local Includes */
#include "apply_sun_angle_correction.h"

int write_l1g_band
(
    L1GIO *l1g_io,          /* I: Output L1G I/O */
    int band_number,        /* I: Band number to output */
    IAS_DATA_TYPE data_type,/* I: Output data type */
    int number_of_scas,     /* I: Number of output SCAs */
    int number_of_lines,    /* I: Number of output lines */
    int number_of_samps,    /* I: Number of output samps */
    float *data_buffer      /* I: Data buffer */
)
{
    L1G_BAND_IO *l1g_band_io;         /* Output L1G band I/O */
    void *data_ptr = data_buffer;     /* Output data pointer */
    int data_converted = 0;           /* Flag to indicated data was converted */
    int pixels_clamped = 0;           /* Number of pixels clamped */
    int buffer_size = number_of_lines * number_of_samps;

    /* Open the band for writing */
    l1g_band_io = ias_l1g_open_band(l1g_io, band_number, &data_type, 
        &number_of_scas, &number_of_lines, &number_of_samps);
    if (!l1g_band_io)
    {
        IAS_LOG_ERROR("Opening band_number %d", band_number);
        return ERROR;
    }

    /* Check if data should be converted to unsigned int because the 
       sun angle correction data is in float */
    if (data_type != IAS_R4)
    {
        /* Allocate the conversion buffer */
        data_ptr = convert_data_buffer(buffer_size, IAS_R4, data_type, 
            data_buffer, &pixels_clamped); 
        if (!data_ptr)
        {
            IAS_LOG_ERROR("Converting output buffer for band number "
                "%d", band_number);
            ias_l1g_close_band(l1g_band_io);
            return ERROR;
        }

        /* Set the conversion flag */
        data_converted = 1;
    }
   
    /* Output how many of the pixels were clamped */
    IAS_LOG_INFO("Number of pixels clamped for band number %d = %d "
        "which is %.2f%% of the image", band_number, pixels_clamped,
        ((double)pixels_clamped / (double)buffer_size) * 100.0);

    /* Write out the band data */
    if (ias_l1g_write_image(l1g_band_io, 0, 0, 0, number_of_lines, 
        number_of_samps, data_ptr) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing band number %d", band_number);
        ias_l1g_close_band(l1g_band_io);
        if (data_converted)
            free(data_ptr);
        return ERROR;
    }
    
    /* Check if the output was converted so the converted data can be freed */
    if (data_converted)
    {
        free(data_ptr);
        data_ptr = NULL;
    }

    /* Close the L1G band */
    if (ias_l1g_close_band(l1g_band_io) != SUCCESS)    
    {
        IAS_LOG_ERROR("Closing band number %d", band_number);
        return ERROR;
    }

    return SUCCESS;
}
