#!/usr/bin/env python3
""" Characterize SCA Overlap

    Using the input from the resampler SCA overlap characterization binary file
    this application computes the statistical information about each OLI
    and TIRS overlap zone.
"""
import sys
import multiprocessing
import time
from ias.log import logger as log
from ias.log.logger import log_exceptions
from ias.db import database as db
from ias.io.cpf import cpf as ias_cpf
from ias.io.parameter_file import parameter_file as parm_file
from ias.misc.const import SUCCESS, ERROR
from ias.misc import satellite_attributes as sa
import numpy as np
import sca_overlap_statistics as sca_overlap_stats
import resampler_char_file


@log_exceptions
def main():
    # Initialize the logger
    logger = log.Logger("SCA Overlap Characterization")

    # Parse in input arguments
    parms = read_parameters()

    # Skip processing if a report is not requested or the chacterization has
    # not been requested to be stored in the database
    if not parms.gen_report and not parms.store_in_db:
        logger.info("Skipping statistics generation")
        sys.exit(SUCCESS)

    # Check if the characterization file is available
    if parms.overlap_filename is None:
        logger.info("No SCA Overlap file available, skipping characterization")
        sys.exit(SUCCESS)

    # Create the characterization file reader
    sat_attr = sa.set_default(sa.get_satellite_id_from_number(
        parms.satellite_number))
    char_file = resampler_char_file.ResamplerCharFile(parms.overlap_filename)

    # Get the cpf and combine the rescaling factors for OLI and TIRS
    cpf = ias_cpf.CPF(filename=parms.cpf_name)
    rescale = cpf.get_oli_radiance_rescale()
    tirs_rescale = cpf.get_tirs_radiance_rescale()
    rescale.reflectance_additive_factor = (np.hstack((
        rescale.reflectance_additive_factor,
        tirs_rescale.radiance_additive_factor)))
    rescale.reflectance_multiplicative_factor = (np.hstack((
        rescale.reflectance_multiplicative_factor,
        tirs_rescale.radiance_multiplicative_factor)))

    # Create the processing pool and stats queue
    pool = multiprocessing.Pool(processes=parms.max_processors)
    queue = multiprocessing.Manager().list()
    async_results = list()

    # Open the binary char file for reading
    char_file.open()

    # Create the band range, process the largest bands first because it is the
    # bottleneck for all other bands putting it first reduces that bottleneck
    band_count = sat_attr.get_band_count()
    band_range = np.arange(band_count)
    band_range = sorted(band_range, key=char_file.get_record_count,
                        reverse=True)

    # Loop over each band processing the statistics
    logger.info("Characterizing the SCA Overlap zones")
    for band_index in band_range:
        band_number = sat_attr.convert_band_index_to_number(band_index)

        # Convert the sensor ID to the correct smoothing window index
        sensor_id = sat_attr.get_sensor_id(band_number)
        if sensor_id == sa.SensorId.OLI:
            sensor_index = 0
        else:
            sensor_index = 1

        # Only process requested bands
        if (band_number not in parms.band_list or
                not char_file.band_present[band_index]):
            continue

        # Read the records for the current band
        records = char_file.read_band(band_index)

        # Get the number of sca overlap regions
        regions = sat_attr.get_scas_per_band(band_number) - 1

        # Determine which smooth window to use
        # Asynchronously process each band
        result = pool.apply_async(process_band, [
            band_index, band_number, records,
            rescale.reflectance_additive_factor,
            rescale.reflectance_multiplicative_factor,
            parms.smooth_window[sensor_index], parms.filter_results, regions,
            queue])
        result.band_number = band_number
        async_results.append(result)

    # Close the characterization file
    char_file.close()

    # Close the pool to no longer accept new tasks
    pool.close()

    # Wait until the results to be completed from the submitted tasks.  If any
    # tasks return an exception, set the error detected flag.  All the children
    # need to complete before we can safely terminate.
    error_detected = False
    for result in async_results:
        try:
            result.get()
        except Exception as err:
            logger.warn("Exception received for band number {0}: {1}".format(
                    result.band_number, str(err)))
            error_detected = True

    # Wait for the processes to exit
    pool.join()

    # If an error was detected, issue an error message and terminate with an
    # error.  Now that all the children processes have terminated, it is safe
    # to exit the parent.
    if error_detected:
        logger.error("Error calculating results")
        exit(1)

    # Sort the results in band order
    queue = sorted(queue, key=get_key)

    # Report any overlap regions where it failed to calculate a valid
    # correlation coefficient.  The data will still go into the database
    # if requested, but with a correlation coefficient of 0.0 which cal/val
    # has agreed they can filter out.
    for result in queue:
        for stats in result.stats:
            if (stats.ratio.valid_correlation is False) or (
                    stats.delta.valid_correlation is False):
                logger.warn("Invalid correlation coefficient in band number "
                            "{0} SCA overlap region {1}".format(
                                result.band_number, stats.sca_number))

    # Create the database connection
    if parms.store_in_db:
        mydb = db.DatabaseConnection()
        mydb.set_login_using_satellite_sensor(
            sat_attr.get_satellite_name(),
            sat_attr.get_satellite_sensor_name())
        mydb.connect()

        rad_table = "RADIOMETRIC_WORK_ORDER_COMMON"
        sca_table = "SCA_OVERLAP"
        if parms.unit_test:
            rad_table = "DEV_RAD_WORK_ORDER_COMMON_SCA"
            sca_table = "DEV_SCA_OVERLAP"

        # Retrieve the radiometric wo com id
        sql = ("Select RADIOMETRIC_WO_COMMON_ID from {0} Where "
               "WORK_ORDER_ID = '{1}'".format(rad_table, parms.wo_id))
        cursor = mydb.execute(sql)
        rad_id = cursor.fetchall()[0][0]

        # Delete any current statistics for the radiometric wo id
        sql = "Delete from {0} Where RADIOMETRIC_WO_COMMON_ID = '{1}'".format(
            sca_table, str(rad_id))
        cursor = mydb.execute(sql)

    # Write the report header
    if parms.gen_report:
        report_file = open(parms.report_filename, 'w')
        write_report_header(report_file)

    # Issue a message about creating the final results, giving priority to
    # the database message
    if parms.store_in_db:
        logger.info("Storing the SCA Overlap statistics into the database")
    elif parms.gen_report:
        logger.info("Writing SCA Overlap statistics report")

    for result in queue:
        for stats in result.stats:
            if parms.store_in_db:
                stats.insert_db_row(
                    mydb, rad_id, result.band_number, stats.sca_number,
                    parms.unit_test)
            if parms.gen_report:
                write_report_record(report_file, result.band_number, stats)

    # Close the report file
    if parms.gen_report:
        report_file.close()

    # Disconnect from the database
    if parms.store_in_db:
        mydb.close(db.DISCONNECT_COMMIT)

    logger.info("Successfully completed SCA Overlap statistics generation")


def get_key(result):
    """ Returns the key for sorting the return data """
    return result.band_number


def process_band(band_index, band_number, records, additive, multiplicative,
                 smooth_window, filter_results, sca_overlap_regions, queue):
    """ Function used to process each band in the multiprocessing format

        Parameters:
            Band index - Input band index
            Band number - Input band number
            Records - Input characterization records
            Additive - Additive factors
            Multiplicative - Multiplicative factors
            Smooth_Window - Smoothing window requested
            Filter results - Flag to indicate results should be filtered
            Sca Overlap Regions - Number of SCA overlap regions
            Queue - Used to store the results
    """
    # Filter the records to only keep the ones that should be used in the fit
    records = resampler_char_file.filter_records(records)

    # Sort the remaining records by the SCA overlap zone followed by the left
    # L1R line number, right L1R line number, left L1R value, and right L1R
    # value.  The values are included in case the first three fields are
    # identical.  If they aren't specified, the FULL versus LIMITED options
    # can sort differently and get different fits.
    records = np.sort(records, order=('sca_zone', 'l1r_line_left',
                                      'l1r_line_right', 'l1r_value_left',
                                      'l1r_value_right'))

    overlap_stats = sca_overlap_stats.OverlapStatistics(
        band_number, band_index, sca_overlap_regions)

    # Get the char records for the current band
    q1, q2, num_rec_per_sca = resampler_char_file.count_sca_records(
        records, band_number)

    # Convert the records from DN to reflectance for OLI bands and to radiance
    # in the TIRS bands
    q1 = ((q1.astype(dtype=np.float64) + additive[band_index]) *
          multiplicative[band_index])
    q2 = ((q2.astype(dtype=np.float64) + additive[band_index]) *
          multiplicative[band_index])

    # Prepare the start-end indexes for the scas.  Since the end index is used
    # in a slice, the end index is not included in the results.  So, end is
    # the index to the first index of the next overlap area.
    start = 0
    end = num_rec_per_sca[0]

    # Calculate the statistics for each SCA
    for sca in range(sca_overlap_regions):

        # Pull the values from the combined SCA array
        sca_q1 = q1[start:end]
        sca_q2 = q2[start:end]
        stats = overlap_stats.stats[sca]

        # If there are no records for the overlap area, consider it an error
        # to avoid runtime warnings
        if len(sca_q1) < 1:
            raise Exception(
                "No data in band number {0} overlap region number {1}".format(
                    band_number, sca + 1))

        # In case the sca_q2 value has zeroes in it, turn off divide by zero
        # warning messages.  Any 'inf' values will be filtered out in the
        # stats calculation.
        old_err_settings = np.seterr(divide='ignore')

        # Calculate the radiance ratios
        stats.ratio.add_data(np.divide(sca_q1, sca_q2))

        # Restore the divide by zero error setting
        np.seterr(divide=old_err_settings['divide'])

        # Calculate the radiance deltas
        stats.delta.add_data(np.subtract(sca_q1, sca_q2))

        # Create an array that contains which indicies are valid, where valid
        # means the ratio is not zero, infinite, or NAN
        valid_indices = np.where((stats.ratio.data != 0.0) &
                                 (np.isfinite(stats.ratio.data)))

        # Finish the statistics calculations
        stats.ratio.calculate_stats(smooth_window, valid_indices, sca + 1)
        stats.delta.calculate_stats(smooth_window, valid_indices, sca + 1)

        # Advance the indexing
        if sca + 1 < sca_overlap_regions:
            start = end
            end = start + num_rec_per_sca[sca+1]
    # Add the statistics to the shared memory queue
    queue.append(overlap_stats)


def read_parameters():
    """ Reads the Parameter and OMF files using the requested parameters """
    parser = parm_file.ParameterParser()
    parser.add_standard_parameter('WORK_ORDER_ID', 'wo_id')
    parser.add_standard_parameter('STORE_CHARACTERIZATION_IN_DB',
                                  'store_in_db')
    parser.add_standard_parameter('BAND_LIST', 'band_list', is_list=True,
                                  required=False)
    parser.add_standard_parameter('GENERATE_REPORT', 'gen_report',
                                  required=False)
    parser.add_standard_parameter('REPORT_FILENAME', 'report_filename',
                                  required=False,
                                  default="SCA_OVERLAP_CHAR_REPORT.txt")
    parser.add_parameter('SMOOTHING_WINDOW', 'smooth_window', is_list=True,
                         required=False, default=(250, 250), data_type=int,
                         help=('Smoothing window (OLI, TIRS) to use on each '
                               'side of the smoothed data'))
    parser.add_parameter('FILTER_RESULTS_FLAG', 'filter_results',
                         required=False, default=True, data_type=int,
                         help='Filter SCA zone results flag')
    parser.add_parameter('UNIT_TEST', 'unit_test', required=False,
                         default=False, help='Unit testing flag')

    # OMF parameters
    omf = parm_file.ParameterParser()
    omf.add_parameter("SCA_OVERLAP_CHAR_FILENAME", 'overlap_filename',
                      required=False, data_type=str,
                      help=("Name of the SCA overlap characterization "
                            "data file"))
    omf.add_standard_parameter("MAX_PROCESSORS_TO_USE", 'max_processors')
    omf.add_standard_parameter('CAL_PARM_FILENAME', 'cpf_name')
    omf.add_standard_parameter('SATELLITE', 'satellite_number', required=True)

    # Parse the parameter file if it exists
    if len(sys.argv) >= 2:
        # Provide help if requested
        if parser.provide_help(sys.argv[1], parm_file.PARAMETERS):
            omf.provide_help(sys.argv[1], parm_file.OMF)
            sys.exit(SUCCESS)

        parms = parser.parse_parameters(sys.argv[1])
        parms = omf.parse_parameters("{0}.omf".format(parms.wo_id),
                                     target=parms)
    else:
        log.Logger.get_default_logger().error("No parameter file provided")
        sys.exit(ERROR)

    return parms


def write_report_record(file, band_number, stats):
    """ Writes out the per SCA zone report record"""
    file.write("{:>5d} {:>12d} {:>15d} {:>15.8e} {:>15.8e} {:>15.8e} "
               "{:>15.8e} {:>15.8e} {:>15.8e} {:>15.8e} {:>15.8e} {:>15.8e} "
               "{:>15.8e} {:>15d} {:>15.8e} {:>15.8e} {:>15.8e} {:>15.8e} "
               "{:>15.8e} {:>15.8e} {:>15.8e} {:>15.8e} {:>15.8e} {:>15.8e} "
               "{:>15d}\n".format(
                   band_number, stats.sca_number, stats.ratio.smooth_pnts,
                   stats.ratio.mean, stats.ratio.stdev, stats.ratio.coeff[0],
                   stats.ratio.coeff[1], stats.ratio.coeff[2],
                   stats.ratio.coeff[3], stats.ratio.max_error,
                   stats.ratio.chi_squared, stats.ratio.correlation,
                   stats.ratio.rmse_residual, stats.ratio.sigma_points,
                   stats.delta.mean, stats.delta.stdev, stats.delta.coeff[0],
                   stats.delta.coeff[1], stats.delta.coeff[2],
                   stats.delta.coeff[3], stats.delta.max_error,
                   stats.delta.chi_squared, stats.delta.correlation,
                   stats.delta.rmse_residual, stats.delta.sigma_points))


def write_report_header(file):
    """ Write out the report header """
    file.write("{:>5s} {:>12s} {:>15s} {:>15s} {:>15s} {:>15s} "
               "{:>15s} {:>15s} {:>15s} {:>15s} {:>15s} {:>15s} "
               "{:>15s} {:>15s} {:>15s} {:>15s} {:>15s} {:>15s} "
               "{:>15s} {:>15s} {:>15s} {:>15s} {:>15s} {:>15s} "
               "{:>15s}\n".format(
                   "BAND", "SCA_REGION", "OVERLAP_POINTS", "RATIO_MEAN",
                   "RATIO_STDEV", "RATIO_COEFF_1", "RATIO_COEFF_2",
                   "RATIO_COEFF_3", "RATIO_COEFF_4", "RATIO_MAX_ERR",
                   "RATIO_CHI_SQ", "RATIO_CORR",
                   "RATIO_RMSE_ERR", "RATIO_SIG_PNTS",
                   "DELTA_MEAN", "DELTA_STDEV", "DELTA_COEFF_1",
                   "DELTA_COEFF_2", "DELTA_COEFF_3", "DELTA_COEFF_4",
                   "DELTA_MAX_ERR", "DELTA_CHI_SQ",
                   "DELTA_CORR", "DELTA_RMSE_ERR",
                   "DELTA_SIG_PNTS"))

# Runs main if called from command line
if __name__ == "__main__":
    main()
