#!/usr/bin/env python3
""" Resampler Char File

    Module used to load the binary file produced by the resampler for the SCA
    overlap characterization
"""
import os
import argparse
import struct
import numpy as np
from ias.misc import satellite_attributes as sa
from ias.log import logger as lg
from ias.log.logger import log_exceptions

# Private tags signifing the start of a section
_FILE_TYPE_TAG = 0x216677FA
_BAND_START_TAG = 0x5A5AA5A5
_FILE_END_TAG = 0x2352CE03
_LIMITED_TAG = 1
_FULL_TAG = 2


class ResamplerCharFile():
    """ Class used to read the resampler binary overlap file """

    def __init__(self, filename):
        """ Initializes the ResampleCharFile object

            Filename: Required filename to load

            Assumptions: The filename refers to a binary file that is
                         structured correctly
        """
        self.filename = filename
        self.fp = None
        self.sat_attr = sa.get_default()
        band_count = self.sat_attr.get_band_count()
        self.band_start = np.zeros(band_count, dtype=int)
        self.band_present = np.zeros(band_count, dtype=int)
        self.band_records = np.zeros(band_count, dtype=int)
        self.record_dtype = None
        self.file_type = _FILE_TYPE_TAG
        self.data_type = _LIMITED_TAG

    def __del__(self):
        """ ResampleCharFile destructor """
        if self.fp is not None:
            self.fp.close()
            self.fp = None

    def close(self):
        """ Close the file pointer to the binary file """
        if self.fp is not None:
            self.fp.close()
            self.fp = None

    def open(self):
        """ Opens the binary file and gets all present bands and the byte
            locations per band to make reading quicker """

        # Open the binary file
        self.fp = open(self.filename, 'rb')

        # Read the file type tag
        byte_count = self.fp.read(8)
        self.file_type, self.data_type = struct.unpack('ii', byte_count)
        if self.file_type != _FILE_TYPE_TAG:
            raise Exception("Invalid file type")

        # Determine if this is a limited or full characterization file
        # A data_type of 1 is limited and 2 is full
        # Build the record data type
        self.record_dtype = np.dtype([
            ('l1t_line', 'i4'), ('l1t_samp', 'i4'), ('sca_zone', 'i4'),
            ('l1r_value_left', 'f4'), ('l1r_value_right', 'f4'),
            ('l1r_line_left', 'f4'), ('l1r_line_right', 'f4'),
            ('l1r_samp_left', 'f4'), ('l1r_samp_right', 'f4')])
        bytes_to_read = 36
        if self.data_type == _LIMITED_TAG:
            bytes_to_read = 20
            self.record_dtype = np.dtype([
                ('sca_zone', 'i4'), ('l1r_value_left', 'f4'),
                ('l1r_value_right', 'f4'),
                ('l1r_line_left', 'f4'), ('l1r_line_right', 'f4')])

        # Read the first band tag
        byte_count = self.fp.read(4)
        band_start_tag = struct.unpack('i', byte_count)
        if band_start_tag[0] != _BAND_START_TAG:
            raise Exception("Invalid band start tag")

        # Loop over all the possible bands creating a lookup table to be used
        # when reading a band. Also mark the band as present and save the
        # record count
        band_count = self.sat_attr.get_band_count()
        for ignored_value in range(band_count):
            # Save the band starting byte
            band_start = self.fp.tell()

            # Read the band number and record count
            byte_count = self.fp.read(8)
            (band_number, record_count) = struct.unpack('ii', byte_count)
            band_index = self.sat_attr.convert_band_number_to_index(
                band_number)
            self.band_present[band_index] = True
            self.band_start[band_index] = band_start
            self.band_records[band_index] = record_count

            # Go to the next band start tag
            self.fp.seek(self.fp.tell() + (record_count * bytes_to_read))
            byte_count = self.fp.read(4)
            next_tag = struct.unpack('i', byte_count)
            if next_tag[0] == _FILE_END_TAG:
                break
            elif next_tag[0] != _BAND_START_TAG:
                raise Exception("File formatted incorrectly for band number "
                                "{0}".format(band_number))

    def read_band(self, band_index):
        """ Reads the band specified into the 2d structured array

            Returns:
                2D SCA characterization Array
        """
        # Only read the present bands
        if not self.band_present[band_index]:
            raise Exception("Band not present in file")

        # Go the the begining of the band data
        self.fp.seek(self.band_start[band_index])

        # Read the band number and record count
        byte_count = self.fp.read(8)
        (band_number, record_count) = struct.unpack('ii', byte_count)

        # Verify the band index corresponds to the correct band number
        verify_band_number = self.sat_attr.convert_band_index_to_number(
            band_index)
        if verify_band_number != band_number:
            raise Exception("Band number {0} does not match the band number "
                            "({1}) in the characterization file".format(
                                verify_band_number, band_number))

        # Create the records structure
        records = np.fromfile(self.fp, dtype=self.record_dtype,
                              count=record_count)

        # Read the next tag and verify its a new band or end of file
        byte_count = self.fp.read(4)
        next_tag = struct.unpack('i', byte_count)
        if next_tag[0] == _FILE_END_TAG:
            pass
        elif next_tag[0] != _BAND_START_TAG:
            raise Exception("File formmatted incorrectly "
                            "{0}".format(next_tag[0]))
        return records

    def get_record_count(self, band_index):
        """ Get the record count for the requested band index
        """
        return self.band_records[band_index]


def filter_records(records):
    """ Filter the records to only keep the ones that are near integer line
        locations

        Returns:
            The filtered set of records
    """
    left_sca_lines = records[:]["l1r_line_left"]
    right_sca_lines = records[:]["l1r_line_right"]

    # Determine the fractional value of the absolute value and rounded values
    # of the left and right sca line locations
    left_sca_fractional = np.absolute(np.absolute(left_sca_lines) -
                                      np.round(left_sca_lines))
    right_sca_fractional = np.absolute(np.absolute(right_sca_lines) -
                                       np.round(right_sca_lines))

    # Find bad records in the left and right scas.  Check the locations in the
    # left and right sca where they are not close enough to integer locations
    left_sca_bad = np.where(left_sca_fractional >= .1)[0]
    right_sca_bad = np.where(right_sca_fractional >= .1)[0]

    # Intersect the bad locations because we only want to remove locations
    # where both the left and right scas are not close to integer locations
    removal_locations = np.intersect1d(left_sca_bad, right_sca_bad)

    # Remove any bad records if necessary
    if len(removal_locations) > 0:
        records = np.delete(records, removal_locations)
    return records


def count_sca_records(records, band_number):
    """ Helper function used to filter the records that are close to
        integer locations and to count the number of records per sca

        Returns:
            Left SCA values (Array)
            Right SCA values (Array)
            Per SCA record count (Array)
    """

    # Get the satellite attributes
    sat_attr = sa.get_default()

    record_count_per_sca = np.zeros(sat_attr.get_scas_per_band(band_number),
                                    dtype=int)

    # Count the number of records for each SCA
    last_sca = records[0]['sca_zone']
    count = 0
    for record in records:
        # Check if the next SCA is present and save off the record count
        # and reset the count and save off the new SCA zone
        if record['sca_zone'] != last_sca:
            record_count_per_sca[last_sca - 1] = count
            last_sca = record['sca_zone']
            count = 0
        count += 1
    record_count_per_sca[last_sca - 1] = count

    # Return only the values
    return (records[:]["l1r_value_left"], records[:]["l1r_value_right"],
            record_count_per_sca)


@log_exceptions
def main():
    """ Main routine to dump the binary characterization file to ascii
        or to rewrite the binary file as sorted """

    logger = lg.Logger("Dump_SCA_Overlap_Char")
    argparser = argparse.ArgumentParser(description='DUMP_SCA_OVERLAP_CHAR')
    argparser.add_argument(
        dest='in_file', help='Input characterization filename')
    argparser.add_argument(
        dest='out_file', help='Output ascii characterization filename')
    argparser.add_argument('--band_number', dest='band_number', required=False,
                           help='Band number to dump if only a single band '
                           'is wanted', type=int)
    argparser.add_argument(
        '--filter', dest='filter_output', required=False,
        help='Filter the records to only include the ones used in the fit',
        action='store_true')
    argparser.add_argument(
        '--binary', dest='binary_output', required=False,
        help='Output as a binary file', action='store_true')
    args = argparser.parse_args()

    # Get the list of bands to include if it was provided
    if args.band_number:
        included_band_numbers = [args.band_number]
    else:
        included_band_numbers = None

    # Open the input file, assuming L8
    sat_attr = sa.set_default(sa.get_satellite_id_from_number(8))

    ifile = ResamplerCharFile(args.in_file)
    ifile.open()

    # Split the filename extension off the name
    out_file_base, out_file_ext = os.path.splitext(args.out_file)

    # Open the output file if output to binary is requested
    ofile = None
    if args.binary_output:
        ofile = open(args.out_file, 'wb')
        ofile.write(struct.pack('i', ifile.file_type))
        ofile.write(struct.pack('i', ifile.data_type))

    # Loop over available bands
    band_count = sat_attr.get_band_count()
    for band_index in range(band_count):
        if not ifile.band_present[band_index]:
            continue
        band_num = sat_attr.convert_band_index_to_number(band_index)

        # If a single band was requested, skip this band if it is not the
        # requested band
        if ((included_band_numbers is not None) and
                (band_num not in included_band_numbers)):
            continue

        logger.info("Processing band {0}".format(band_num))

        # Read all the records for the current band
        records = ifile.read_band(band_index)

        # Filter the records if requested
        if args.filter_output:
            records = filter_records(records)

        # Sort the records
        records = np.sort(records, order=('sca_zone', 'l1r_line_left',
                                          'l1r_line_right', 'l1r_value_left',
                                          'l1r_value_right'))

        # Output to the requested file type
        if args.binary_output:
            ofile.write(struct.pack('i', _BAND_START_TAG))
            ofile.write(struct.pack('i', band_num))
            ofile.write(struct.pack('i', len(records)))
            records.tofile(ofile)
        else:
            ofile = open("{0}_B{1:02d}{2}".format(out_file_base, band_num,
                                                  out_file_ext), 'w')
            if ifile.data_type == _LIMITED_TAG:
                ofile.write(
                    "{:>10s} {:>10s} {:>15s} {:>15s} {:>15s} "
                    "{:>15s}\n".format(
                        "Band", "SCA ZONE", "(L) L1R Value",
                        "(R) L1R Value", "(L) L1R Line", "(R) L1R Line"))
            else:
                ofile.write(
                    "{:>10s} {:>10s} {:>10s} {:>10s} {:>15s} {:>15s} "
                    "{:>15s} {:>15s} {:>15s} {:>15s}\n".format(
                        "Band", "L1T Line", "L1T Samp", "SCA ZONE",
                        "(L) L1R Value", "(R) L1R Value", "(L) L1R Line",
                        "(R) L1R Line", "(L) L1R Samp", "(R) L1R Samp"))

            for rec in records:
                if ifile.data_type == _LIMITED_TAG:
                    ofile.write(
                        "{:>10d} {:>10d} {:>15.5f} {:>15.5f} {:>15.5f} "
                        "{:>15.5f}\n".format(band_num, *rec))
                else:
                    ofile.write(
                        "{:>10d} {:>10d} {:>10d} {:>10d} {:>15.5f} "
                        "{:>15.5f} {:>15.5f} {:>15.5f} {:>15.5f} "
                        "{:>15.5f}\n".format(band_num, *rec))
            ofile.close()

    # Close the opened files
    if args.binary_output:
        ofile.write(struct.pack('i', _FILE_END_TAG))
        ofile.close()
    ifile.close()

# Runs main if called from command line
if __name__ == "__main__":
    main()
