#!/usr/bin/env python3
""" SCA Overlap Statistics

    Module used to calculate the SCA overlap characterization statistics
"""
import numpy as np
from scipy.ndimage.filters import uniform_filter
from scipy.stats import pearsonr


class FitData():
    """ Class used to calculate the ratio and delta statistics for the
        SCA overlap
    """

    def __init__(self):
        self.coeff = np.zeros(4, dtype=np.float64)
        self.chi_squared = None
        self.correlation = None
        self.valid_correlation = False
        self.max_error = None
        self.sigma_points = None
        self.rmse_residual = None
        self.status = 0
        self.mean = None
        self.stdev = None
        self.data = None
        self.smooth_pnts = None

    def add_data(self, data):
        """ Stores the data and gets the mean and standard deviation """
        self.data = data
        self.mean = np.mean(data)
        self.stdev = np.std(data)

    def smooth_data(self, smooth_window, valid_indices, region_number):
        """ Smooths the data using the generic smoothing function
            with the smooth window as a cut off
        """
        if smooth_window < 1 or smooth_window >= self.data.size / 2:
            raise Exception(
                "Smoothing window size of {0} is not valid "
                "with data size of {1} for overlap region number {2}".format(
                    smooth_window, self.data.size, region_number))

        # Construct the data to smooth from the valid entries
        smooth = self.data[valid_indices]

        smooth = uniform_filter(smooth, smooth_window * 2)
        return smooth[smooth_window:-smooth_window]

    def calculate_stats(self, smooth_window, valid_indices, region_number):
        """ Calculates all the SCA overlap statistics """

        # Smooth the data
        smooth = self.smooth_data(smooth_window, valid_indices, region_number)
        self.smooth_pnts = smooth.size

        # Fit the smoothed data using a polynomial fit
        coeffs = np.polyfit(np.arange(smooth.size), smooth, 3, full=False)

        # Calculate the Y fit
        yfit = np.polyval(coeffs, np.arange(smooth.size))

        # The numpy polyfit and polyval routine expect the coefficients in high
        # to low order, but we want them in low to high order.  So, reverse the
        # coeffs using a slice since we're done operating on them with numpy.
        self.coeff = coeffs[::-1]

        # Calculate the chi square, correlation, max error, and rmse
        # The chi square is calculated with weights of 1 and degrees of
        # freedom of 1.  Only calculate the correlation if the standard dev
        # is greater than zero since values of zero will cause a runtime
        # error and should only happen if the results aren't correct.
        self.chi_squared = np.sum((smooth - yfit) ** 2)
        if self.stdev > 0.0:
            self.correlation = pearsonr(smooth, yfit)[0]
            self.valid_correlation = True
        else:
            # Set to zero for no correlation
            self.correlation = 0.0

        self.max_error = np.amax(np.absolute(smooth - yfit))
        self.rmse_residual = np.std(smooth - yfit)

        # Calculate the number of points larger than 3 * Sigma
        sigma_points = np.abs(smooth - yfit)
        three_sigma = self.stdev / np.sqrt(smooth_window * 2) * 3
        large_sigma_point_count = [pnt for pnt in sigma_points
                                   if pnt > three_sigma]
        self.sigma_points = len(large_sigma_point_count)


class ScaOverlapStatistics():
    """ Overlap Statistics class used to keep track of the ratio and delta
        statistics
    """
    def __init__(self, sca_number):
        self.ratio = FitData()
        self.delta = FitData()
        self.sca_number = sca_number

    def insert_db_row(self, db, rad_id, band_number, sca_overlap, unit_test):
        """ Method used to insert the per sca statistics into the database"""

        # Check if the insert will happen in the unit test
        insert_table = "SCA_OVERLAP"
        if unit_test:
            insert_table = "DEV_" + insert_table

        insert_sql = """Insert into {0}
            (RADIOMETRIC_WO_COMMON_ID, BAND_ID, SCA_OVERLAP, OVERLAP_POINTS,
             RATIO_MEAN, RATIO_STDEV, RATIO_COEFF_1, RATIO_COEFF_2,
             RATIO_COEFF_3, RATIO_COEFF_4, RATIO_MAX_ERROR, RATIO_CHI_SQUARE,
             RATIO_CORRELATION, RATIO_RMSE_ERROR, RATIO_THREE_SIGMA_POINTS,
             DELTA_MEAN, DELTA_STDEV, DELTA_COEFF_1, DELTA_COEFF_2,
             DELTA_COEFF_3, DELTA_COEFF_4, DELTA_MAX_ERROR,
             DELTA_CHI_SQUARE, DELTA_CORRELATION, DELTA_RMSE_ERROR,
             DELTA_THREE_SIGMA_POINTS)
             VALUES
             (:RADIOMETRIC_WO_COMMON_ID, :BAND, :SCA_OVERLAP, :OVERLAP_POINTS,
             :RATIO_MEAN, :RATIO_STDEV, :RATIO_COEFF_1, :RATIO_COEFF_2,
             :RATIO_COEFF_3, :RATIO_COEFF_4, :RATIO_MAX_ERROR,
             :RATIO_CHI_SQUARE, :RATIO_CORRELATION, :RATIO_RMSE_ERROR,
             :RATIO_THREE_SIGMA_POINTS, :DELTA_MEAN, :DELTA_STDEV,
             :DELTA_COEFF_1, :DELTA_COEFF_2, :DELTA_COEFF_3, :DELTA_COEFF_4,
             :DELTA_MAX_ERROR, :DELTA_CHI_SQUARE, :DELTA_CORRELATION,
             :DELTA_RMSE_ERROR, :DELTA_THREE_SIGMA_POINTS)""".format(
                 insert_table)

        parameters = {"RADIOMETRIC_WO_COMMON_ID": rad_id,
                      "BAND": band_number,
                      "SCA_OVERLAP": sca_overlap,
                      "OVERLAP_POINTS": self.ratio.smooth_pnts,
                      "RATIO_MEAN": self.ratio.mean,
                      "RATIO_STDEV": self.ratio.stdev,
                      "RATIO_COEFF_1": self.ratio.coeff[0],
                      "RATIO_COEFF_2": self.ratio.coeff[1],
                      "RATIO_COEFF_3": self.ratio.coeff[2],
                      "RATIO_COEFF_4": self.ratio.coeff[3],
                      "RATIO_MAX_ERROR": self.ratio.max_error,
                      "RATIO_CHI_SQUARE": self.ratio.chi_squared,
                      "RATIO_CORRELATION": self.ratio.correlation,
                      "RATIO_RMSE_ERROR": self.ratio.rmse_residual,
                      "RATIO_THREE_SIGMA_POINTS": self.ratio.sigma_points,
                      "DELTA_MEAN": self.delta.mean,
                      "DELTA_STDEV": self.delta.stdev,
                      "DELTA_COEFF_1": self.delta.coeff[0],
                      "DELTA_COEFF_2": self.delta.coeff[1],
                      "DELTA_COEFF_3": self.delta.coeff[2],
                      "DELTA_COEFF_4": self.delta.coeff[3],
                      "DELTA_MAX_ERROR": self.delta.max_error,
                      "DELTA_CHI_SQUARE": self.delta.chi_squared,
                      "DELTA_CORRELATION": self.delta.correlation,
                      "DELTA_RMSE_ERROR": self.delta.rmse_residual,
                      "DELTA_THREE_SIGMA_POINTS": self.delta.sigma_points}

        cursor = db.execute(insert_sql, parameters)
        if cursor.rowcount != 1:
            cursor.close()
            raise Exception("Insert row failed")
        cursor.close()


class OverlapStatistics():
    """ Class used to combine the SCA statistics into one object """
    def __init__(self, band_number, band_index, sca_overlap_regions):
        self.band_number = band_number
        self.band_index = band_index

        # Create a new statistics object for each SCA
        self.stats = list()
        for sca in range(sca_overlap_regions):
            self.stats.append(ScaOverlapStatistics(sca + 1))
