#!/usr/bin/env python3
""" Test SCA Overlap

    Testing module used to test the SCA Overlap characterization
"""

import unittest
from subprocess import call
from ias.misc import satellite_attributes as sa
from ias.db import database as db
import os
import filecmp
import imp
import numpy as np
import sca_overlap_statistics as sos
import resampler_char_file as rcf


class TestSCAOverlapCharacterization(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """ Setup method ran before all unit tests"""
        db_cred = os.getenv("IAS_DB_L8_OLITIRS")
        if not db_cred:
            raise Exception("IAS_DB_L8_OLITIRS environment variable not set")

        # Create the development tables
        status = call(["sqlplus", db_cred, "@create_tables.sql"])
        if status != 0:
            raise Exception("Creating db tables")

        # Delete the "Fake" record from the radiometric wo common table
        # incase it wasn't removed during a previous execution
        mydb = db.DatabaseConnection("L8_OLITIRS")
        mydb.connect()
        sql = ("Delete from RADIOMETRIC_WORK_ORDER_COMMON where "
               "RADIOMETRIC_WO_COMMON_ID = '1000031000' and "
               "WORK_ORDER_ID = 'SCA_OVER112'")
        cursor = mydb.execute(sql)
        mydb.close(db.DISCONNECT_COMMIT)

        # Remove the data link as well
        try:
            os.unlink("sca_overlap.dat")
        except:
            pass

        # Load single rad com record
        status = call(["./load_data.csh"])
        if status != 0:
            raise Exception("Loading data into db tables")

    @staticmethod
    def create_sql_dump_file(filename, results_file, rad_id):
        with open(filename, "w") as f:
            f.write(
                ("""SPOOL "{0}"\nSET LINESIZE 10000\nSET TRIMSPOOL ON\n"""
                 """SET TRIMOUT ON\nSET PAGESIZE 50000\nSET TERMOUT OFF\n"""
                 """Select * from DEV_SCA_OVERLAP where """
                 """radiometric_wo_common_id = '{1}' order by\n"""
                 """radiometric_wo_common_id,band_id,sca_overlap;\n"""
                 """exit;""").format(results_file, rad_id))

    def test_limited_sca_characterization(self):
        """ Test the SCA Overlap characterization application as a whole"""
        char_file = ("{0}/ias_base/rps/characterize_sca_overlap/"
                     "sca_overlap_limited.dat")

        # Get the unit test directory
        ut_dir = os.getenv("IAS_UNIT_TEST_DATA_DIR")
        if not ut_dir:
            raise Exception("IAS_UNIT_TEST_DATA_DIR environment variable not "
                            "set")

        # Cleanup the sca overlap table
        mydb = db.DatabaseConnection("L8_OLITIRS")
        mydb.connect()
        sql = ("Delete from DEV_SCA_OVERLAP")
        cursor = mydb.execute(sql)
        mydb.close(db.DISCONNECT_COMMIT)

        # Open the output omf file
        out_file = open("SCA_OVER112.omf", "w")

        # Create the OMF file from the template
        status = call(["sed", "-e",
                       "s|$IAS_UNIT_TEST_DATA_DIR|{0}|".format(ut_dir),
                       "omf"], stdout=out_file)
        out_file.close()
        if status != 0:
            raise Exception("Running SED to create the OMF")

        # Remove the data results from a previous run
        try:
            os.remove("limited_results.db")
        except:
            pass

        # Remove the data results from a previous run
        try:
            os.remove("SCA_OVERLAP_CHAR_REPORT.txt")
        except:
            pass
        try:
            os.unlink("sca_overlap.dat")
        except:
            pass

        # Link the overlap data
        os.symlink(char_file.format(ut_dir), "sca_overlap.dat")

        # Characterize the sca overlap data
        status = call(["../characterize_sca_overlap.py", "odl"])
        if status != 0:
            raise Exception("Running characterize sca overlap")

        # Unlink the data file
        os.unlink("sca_overlap.dat")

        # Create the dump sql file
        TestSCAOverlapCharacterization.create_sql_dump_file(
            "limited_dump.sql", "limited_results.db", "1000031000")

        db_cred = os.getenv("IAS_DB_L8_OLITIRS")
        if not db_cred:
            raise Exception("IAS_DB_L8_OLITIRS environment variable not set")

        # Dump the records
        status = call(["sqlplus", db_cred, "@limited_dump.sql"])
        if status != 0:
            raise Exception("Creating db tables")

        # Check the results
        self.assertEqual(filecmp.cmp("limited_results.db",
                                     "results.verify"), True)
        self.assertEqual(filecmp.cmp("SCA_OVERLAP_CHAR_REPORT.txt",
                                     "SCA_OVERLAP_CHAR_REPORT.verify"), True)

    def test_full_sca_characterization(self):
        char_file = ("{0}/ias_base/rps/characterize_sca_overlap/"
                     "sca_overlap_full.dat")

        # Get the unit test directoy
        ut_dir = os.getenv("IAS_UNIT_TEST_DATA_DIR")
        if not ut_dir:
            raise Exception("IAS_UNIT_TEST_DATA_DIR environment variable not "
                            "set")

        # Cleanup the sca overlap table
        mydb = db.DatabaseConnection("L8_OLITIRS")
        mydb.connect()
        sql = ("Delete from DEV_SCA_OVERLAP")
        cursor = mydb.execute(sql)
        mydb.close(db.DISCONNECT_COMMIT)

        # Open the output omf file
        out_file = open("SCA_OVER112.omf", "w")

        # Create the OMF file from the template
        status = call(["sed", "-e",
                       "s|$IAS_UNIT_TEST_DATA_DIR|{0}|".format(ut_dir),
                       "omf"], stdout=out_file)
        out_file.close()
        if status != 0:
            raise Exception("Running SED to create the OMF")

        # Remove and existing results
        try:
            os.remove("full_results.db")
        except:
            pass

        # Remove the data results from a previous run
        try:
            os.remove("SCA_OVERLAP_CHAR_REPORT.txt")
        except:
            pass
        try:
            os.unlink("sca_overlap.dat")
        except:
            pass

        # Link the characterization data
        os.symlink(char_file.format(ut_dir), "sca_overlap.dat")

        status = call(["../characterize_sca_overlap.py", "odl"])
        if status != 0:
            raise Exception("Running characterize sca overlap")

        # Unlink the characterization data
        os.unlink("sca_overlap.dat")

        # Create the sql dump file
        TestSCAOverlapCharacterization.create_sql_dump_file(
            "full_dump.sql", "full_results.db", "1000031000")

        db_cred = os.getenv("IAS_DB_L8_OLITIRS")
        if not db_cred:
            raise Exception("IAS_DB_L8_OLITIRS environment variable not set")

        # Dump the new results
        status = call(["sqlplus", db_cred, "@full_dump.sql"])
        if status != 0:
            raise Exception("Creating db tables")

        # Check the results
        self.assertEqual(filecmp.cmp("full_results.db",
                                     "results.verify"), True)
        self.assertEqual(filecmp.cmp("SCA_OVERLAP_CHAR_REPORT.txt",
                                     "SCA_OVERLAP_CHAR_REPORT.verify"), True)

    def test_insert_data(self):
        """ Tests the statistics insert method"""
        stats = sos.ScaOverlapStatistics(1)
        stats.ratio.data = np.arange(100)
        stats.delta.data = np.arange(100)

        # Check that an exception is thrown if the window is bigger than
        # that data
        with self.assertRaises(Exception):
            stats.ratio.smooth_data(50)
        with self.assertRaises(Exception):
            stats.ratio.smooth_data(0)
        stats.ratio.smooth_pnts = 100

        # Verify that the stats were inserted correctly
        stats.ratio.mean = 1.1
        stats.ratio.stdev = 2.2
        stats.ratio.coeff[0] = 3.3
        stats.ratio.coeff[1] = 10.10
        stats.ratio.coeff[2] = 4.4
        stats.ratio.coeff[3] = 5.4
        stats.ratio.chi_squared = 9.9
        stats.ratio.correlation = 7.7
        stats.ratio.max_error = 6.6
        stats.ratio.rmse_residual = 12.12
        stats.ratio.sigma_points = 9

        stats.delta.mean = 21.1
        stats.delta.stdev = 22.2
        stats.delta.coeff[0] = 23.3
        stats.delta.coeff[1] = 210.10
        stats.delta.coeff[2] = 24.4
        stats.delta.coeff[3] = 25.4
        stats.delta.chi_squared = 29.9
        stats.delta.correlation = 27.7
        stats.delta.max_error = 26.6
        stats.delta.rmse_residual = 212.12
        stats.delta.sigma_points = 28

        # Insert the row then rerieve it for verificiation
        mydb = db.DatabaseConnection("L8_OLITIRS")
        mydb.connect()
        stats.insert_db_row(mydb, "99999", 11, 1, True)
        sql = ("""Select RADIOMETRIC_WO_COMMON_ID, BAND_ID, SCA_OVERLAP,
                  OVERLAP_POINTS, RATIO_MEAN, RATIO_STDEV, RATIO_COEFF_1,
                  RATIO_COEFF_2, RATIO_COEFF_3, RATIO_COEFF_4,
                  RATIO_MAX_ERROR, RATIO_CHI_SQUARE, RATIO_CORRELATION,
                  RATIO_RMSE_ERROR, RATIO_THREE_SIGMA_POINTS, DELTA_MEAN,
                  DELTA_STDEV, DELTA_COEFF_1, DELTA_COEFF_2, DELTA_COEFF_3,
                  DELTA_COEFF_4, DELTA_MAX_ERROR, DELTA_CHI_SQUARE,
                  DELTA_CORRELATION, DELTA_RMSE_ERROR, DELTA_THREE_SIGMA_POINTS
                  from DEV_SCA_OVERLAP where
                  RADIOMETRIC_WO_COMMON_ID = '99999'""")
        cursor = mydb.execute(sql)
        rows = cursor.fetchall()
        self.assertEqual(cursor.rowcount, 1)
        mydb.close(db.DISCONNECT_ROLLBACK)

        # Verify the rad id, band, sca zone, and number of points
        self.assertEqual(99999, rows[0][0])
        self.assertEqual(11, rows[0][1])
        self.assertEqual(1, rows[0][2])
        self.assertEqual(100, rows[0][3])

        # Verify the ratio stats
        self.assertAlmostEqual(stats.ratio.mean, rows[0][4])
        self.assertAlmostEqual(stats.ratio.stdev, rows[0][5])
        self.assertAlmostEqual(stats.ratio.coeff[0], rows[0][6])
        self.assertAlmostEqual(stats.ratio.coeff[1], rows[0][7])
        self.assertAlmostEqual(stats.ratio.coeff[2], rows[0][8])
        self.assertAlmostEqual(stats.ratio.coeff[3], rows[0][9])
        self.assertAlmostEqual(stats.ratio.chi_squared, rows[0][11])
        self.assertAlmostEqual(stats.ratio.correlation, rows[0][12])
        self.assertAlmostEqual(stats.ratio.max_error, rows[0][10])
        self.assertAlmostEqual(stats.ratio.rmse_residual, rows[0][13])
        self.assertAlmostEqual(stats.ratio.sigma_points, rows[0][14])

        # Verify the delta stats
        self.assertAlmostEqual(stats.delta.mean, rows[0][15])
        self.assertAlmostEqual(stats.delta.stdev, rows[0][16])
        self.assertAlmostEqual(stats.delta.coeff[0], rows[0][17])
        self.assertAlmostEqual(stats.delta.coeff[1], rows[0][18])
        self.assertAlmostEqual(stats.delta.coeff[2], rows[0][19])
        self.assertAlmostEqual(stats.delta.coeff[3], rows[0][20])
        self.assertAlmostEqual(stats.delta.chi_squared, rows[0][22])
        self.assertAlmostEqual(stats.delta.correlation, rows[0][23])
        self.assertAlmostEqual(stats.delta.max_error, rows[0][21])
        self.assertAlmostEqual(stats.delta.rmse_residual, rows[0][24])
        self.assertAlmostEqual(stats.delta.sigma_points, rows[0][25])

    def test_char_file_io(self):
        """ Test the resample characterization file module"""
        char_file = ("{0}/ias_base/rps/characterize_sca_overlap/"
                     "sca_overlap_full.dat")

        # Get the unit test directory
        ut_dir = os.getenv("IAS_UNIT_TEST_DATA_DIR")
        if not ut_dir:
            raise Exception("IAS_UNIT_TEST_DATA_DIR environment variable not "
                            "set")
        zeros = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        present = [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
        full_start = [12, 30371352, 60743628, 91141392, 121533360, 151928136,
                      182439876, 212942976, 411703740, 442185564, 461318604]
        sat_attr = sa.set_default(sa.get_satellite_id_from_number(8))
        band_count = sat_attr.get_band_count()

        # Test the Full read
        input_file = rcf.ResamplerCharFile(char_file.format(ut_dir))
        self.assertEqual(char_file.format(ut_dir), input_file.filename)
        self.assertEqual(None, input_file.fp)
        np.testing.assert_array_almost_equal(zeros, input_file.band_start)
        np.testing.assert_array_almost_equal(zeros, input_file.band_present)
        input_file.open()

        # Verify the data type, present bands, and band starts
        self.assertEqual(input_file.data_type, 2)
        self.assertNotEqual(None, input_file.fp)
        np.testing.assert_array_almost_equal(present, input_file.band_present)
        np.testing.assert_array_almost_equal(full_start, input_file.band_start)

        # Calculate and compare the band starts using the number of
        # records read
        for band_index in range(band_count):
            records = input_file.read_band(band_index)
            if band_index < 10:
                self.assertEqual(
                    full_start[band_index + 1] -
                    full_start[band_index], (len(records) * 36) + 12)
        input_file.close()

        # Test the limited read
        full_start = [12, 6267344, 12646136, 18642628, 24669960, 30876532,
                      36737284, 42733876, 82251768, 88426760, 91946132]
        char_file = ("{0}/ias_base/rps/characterize_sca_overlap/"
                     "sca_overlap_limited.dat")

        input_file = rcf.ResamplerCharFile(char_file.format(ut_dir))
        self.assertEqual(char_file.format(ut_dir), input_file.filename)
        self.assertEqual(None, input_file.fp)
        np.testing.assert_array_almost_equal(zeros, input_file.band_start)
        np.testing.assert_array_almost_equal(zeros, input_file.band_present)
        input_file.open()

        # Verify the data type, present bands, and band starts
        self.assertEqual(input_file.data_type, 1)
        self.assertNotEqual(None, input_file.fp)
        np.testing.assert_array_almost_equal(present, input_file.band_present)
        np.testing.assert_array_almost_equal(full_start, input_file.band_start)

        # Calculate and compare the band starts using the number of
        # records read
        for band_index in range(band_count):
            records = input_file.read_band(band_index)
            if band_index < 10:
                self.assertEqual(
                    full_start[band_index + 1] -
                    full_start[band_index], (len(records) * 20) + 12)
        input_file.close()

        # Test the dump functionality
        char_file = ("{0}/ias_base/rps/characterize_sca_overlap/"
                     "sca_overlap_dump.dat")
        status = call(["../resampler_char_file.py", char_file.format(ut_dir),
                       "sca_dump.dat", "--binary"])
        if status != 0:
            raise Exception("Dumping SCA statistics")

        # Check the results
        self.assertEqual(filecmp.cmp("sca_dump.dat",
                                     char_file.format(ut_dir)), True)
        status = call(["../resampler_char_file.py", char_file.format(ut_dir),
                       "sca_dump.txt"])
        if status != 0:
            raise Exception("Dumping SCA statistics")

        # Check the results
        band_verify = ("{0}/ias_base/rps/characterize_sca_overlap/"
                       "B01_sca_dump.txt")
        self.assertEqual(filecmp.cmp("sca_dump_B01.txt",
                                     band_verify.format(ut_dir)), True)

        # Test loading a band file
        char_file = ("{0}/ias_base/rps/characterize_sca_overlap/"
                     "sca_overlap_bad.dat")
        input_file = rcf.ResamplerCharFile(char_file.format(ut_dir))
        with self.assertRaises(Exception):
            input_file.open()

    @classmethod
    def tearDownClass(cls):
        """ Tear down method ran after all unit tests"""

        # Unlink the data just incase an error occurred
        try:
            os.unlink("sca_overlap.dat")
        except:
            pass

        db_cred = os.getenv("IAS_DB_L8_OLITIRS")
        if not db_cred:
            raise Exception("IAS_DB_L8_OLITIRS environment variable not set")

        # Cleanup the database tables
        status = call(["sqlplus", db_cred, "@cleanup_tables.sql"])
        if status != 0:
            raise Exception("Cleaning up db tables")

        # Delete the radiometric wo common record
        mydb = db.DatabaseConnection("L8_OLITIRS")
        mydb.connect()
        sql = ("Delete from RADIOMETRIC_WORK_ORDER_COMMON where "
               "RADIOMETRIC_WO_COMMON_ID = '1000031000' and "
               "WORK_ORDER_ID = 'SCA_OVER112'")
        cursor = mydb.execute(sql)
        mydb.close(db.DISCONNECT_COMMIT)


if __name__ == '__main__':
    unittest.main()
