/*****************************************************************************
Name: blur_l1g_image

Purpose: Blurs an image by averaging all the non-fill pixels in the
    -kernel_size/2 to kernel_size/2 window around the original pixel.

Returns: SUCCESS / ERROR

*****************************************************************************/
/* Standard Library Includes */
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* IAS Library Includes */
#include "ias_logging.h"

/* Local Includes */
#include "blur_l1g_image.h"

/* Local Defines */
#define CLAMP_EDGE(edge, low, high) \
    ((edge < low) ? low : ((edge >= high) ? high - 1 : edge))

int blur_l1g_image
(
    const L1G_INFO *in_l1g, /* I: Original L1G image to blur */
    int kernel_size,        /* I: Blurring kernel size in one dimension
                                  (kernel is kernel_size x kernel_size) */
    L1G_INFO *out_l1g       /* O: Blurred L1G image */
)
{
    int line;
    double *per_sample_sums;
    int *per_sample_count;
    int minimum_valid_pixels;

    /* Initialize the output structure */
    *out_l1g = *in_l1g;

    /* The minimum number of valid pixels in the blur kernel is half the kernel
       size (somewhat arbitrary).  If that is very small, grow it to the
       whole kernel size */
    minimum_valid_pixels = (kernel_size < 6) ? kernel_size - 1: kernel_size / 2;

    /* Allocate memory for the blurred image */
    out_l1g->radiance = malloc(sizeof(*out_l1g->radiance) * out_l1g->lines
        * out_l1g->samples);
    if (!out_l1g->radiance)
    {
        IAS_LOG_ERROR("Allocating memory for the blurred image with "
            "%d lines and %d samples", out_l1g->lines, out_l1g->samples);
        out_l1g->radiance = NULL;
        return ERROR;
    }

    /* Allocate memory for the per sample sums */
    per_sample_sums = malloc(sizeof(*per_sample_sums) * out_l1g->samples);
    if (!per_sample_sums)
    {
        IAS_LOG_ERROR("Allocating memory for the individual sums");
        free(out_l1g->radiance);
        out_l1g->radiance = NULL;
        return ERROR;
    }

    /* Allocate memory for the per sample counts */
    per_sample_count = malloc(sizeof(*per_sample_count) * out_l1g->samples);
    if (!per_sample_count)
    {
        IAS_LOG_ERROR("Allocating memory for the individual counts");
        free(per_sample_sums);
        free(out_l1g->radiance);
        out_l1g->radiance = NULL;
        return ERROR;
    }

    /* Perform the blurring */
    for (line = 0; line < out_l1g->lines; line++)
    {
        int sample;
        int top_edge;
        int low_edge;
        double sum = 0.0;
        int count = 0;

        /* Calculate the range of lines to sample in the input image */
        top_edge = line - kernel_size / 2;
        low_edge = line + kernel_size / 2;

        /* Clamp the range to the input image */
        top_edge = CLAMP_EDGE(top_edge, 0, in_l1g->lines);
        low_edge = CLAMP_EDGE(low_edge, 0, in_l1g->lines);

        /* Calculate the sums for each sample on the current line from the
           blur window */
        for (sample = 0; sample < out_l1g->samples; sample++)
        {
            int cur_line;
            int samp_index = top_edge * in_l1g->samples + sample;

            for (cur_line = top_edge; cur_line < low_edge; cur_line++,
                samp_index += in_l1g->samples)
            {
                float pixel = in_l1g->radiance[samp_index];
                if (pixel != 0.0)
                {
                    sum += pixel;
                    count++;
                }
            }

            /* Keep track of the running total across the line */
            per_sample_count[sample] = count;
            per_sample_sums[sample] = sum;
        }

        /* Calculate the blurred output for each sample using the
           pre-calculated sums for the line */
        for (sample = 0; sample < out_l1g->samples; sample++)
        {
            int left_edge;
            int right_edge;
            double sum = 0.0;
            int count = 0;

            /* Calculate the range of lines to sample in the input image */
            left_edge = sample - kernel_size / 2;
            right_edge = sample + kernel_size / 2;

            /* Clamp the range to the input image */
            left_edge = CLAMP_EDGE(left_edge, 0, in_l1g->samples);
            right_edge = CLAMP_EDGE(right_edge, 0, in_l1g->samples);

            /* Get the running total at the end of the window */
            sum = per_sample_sums[right_edge - 1];
            count = per_sample_count[right_edge - 1];

            /* Subtract the running total before start of the window */
            if (left_edge > 0)
            {
                sum -= per_sample_sums[left_edge - 1];
                count -= per_sample_count[left_edge - 1];
            }

            /* Only use the blurred pixel if more than the minimum number of
               pixels contributed to it.  This is to avoid a unblurred edge. */
            if (count > minimum_valid_pixels)
            {
                sum /= count;
            }
            else
            {
                sum = 0.0;
            }

            out_l1g->radiance[line * out_l1g->samples + sample] = sum;
        }
    }

    free(per_sample_sums);
    free(per_sample_count);

    return SUCCESS;
}
