/***************************************************************************
Name: calculate_average_land_water_radiance

Purpose: Calculates the average land and water radiance for the input
         radiance buffer

Returns: Void
****************************************************************************/
/* IAS Library Includes */
#include "ias_types.h"
#include "ias_logging.h"

/* Local Includes */
#include "calculate_average_land_water_radiance.h"

void calculate_average_land_water_radiance
(
    int lines,                      /* I: Number of lines in radiance buffer */
    int samples,                    /* I: Number of samps in radiance buffer */
    int mask_samples,               /* I: Number of samps in water mask */
    int mask_line_offset,           /* I: Mask to l1g line offset */
    int mask_samp_offset,           /* I: Mask to l1g sample offset */
    const float *radiance_buffer,   /* I: Image radiance buffer */
    const IAS_BYTE_TYPE *water_mask,/* I: Land/Water mask buffer */
    float *average_water_radiance,  /* O: Average calculated water radiance */
    float *average_land_radiance    /* O: Average calculated land radiance */
)
{
    int line;                   /* Line loop variable */
    double water_radiance = 0;  /* Local water radiance summation */
    double land_radiance = 0;   /* Local land radiance summation */
    int total_water_pixels = 0; /* Total water pixels count */
    int total_land_pixels = 0;  /* Total land pixels count */

    /* Initialize the output radiances */
    *average_water_radiance = 0;
    *average_land_radiance = 0;

    /* Loop through each line and sample getting the radiance summation */
    for (line = 0; line < lines; line++)
    {
        int samp; /* Sample loop variable */
        int line_index = line * samples;
        for (samp = 0; samp < samples; samp++)
        {
            int pixel = line_index + samp;

            /* Only process non-fill pixels */
            if (radiance_buffer[pixel] > 0.0)
            {
                int mask_pixel = (line + mask_line_offset) * mask_samples
                    + samp + mask_samp_offset;
                if (water_mask[mask_pixel] == LAND_PIXEL)
                {
                    land_radiance += radiance_buffer[pixel];
                    total_land_pixels++;
                }
                else
                {
                    water_radiance += radiance_buffer[pixel];
                    total_water_pixels++;
                }
            }
        }
    }

    /* Calculate the average radiance for land and water */
    if (total_water_pixels != 0)
    {
        *average_water_radiance = water_radiance / total_water_pixels;
    }
    else
    {
        IAS_LOG_INFO("No water pixels detected, using average land radiance "
            "for the average water radiance");
        *average_water_radiance = land_radiance / total_land_pixels;
    }

    if (total_land_pixels != 0)
    {
        *average_land_radiance = land_radiance / total_land_pixels;
    }
    else
    {
        IAS_LOG_INFO("No land pixels detected, using average water radiance "
            "for the average land radiance");
        *average_land_radiance = *average_water_radiance;
    }

    /*----------------------------DEBUG INFO----------------------------------*/
    IAS_LOG_DEBUG("Number of land pixels = %d", total_land_pixels);
    IAS_LOG_DEBUG("Average land radiance = %lf", *average_land_radiance);
    IAS_LOG_DEBUG("Number of water pixels = %d", total_water_pixels);
    IAS_LOG_DEBUG("Average water radiance = %lf", *average_water_radiance);
    /*-------------------------END DEBUG INFO---------------------------------*/
}
