/*******************************************************************************
Name: convert_stray_light_to_lat_long

Purpose: Calculate the geodetic coordinates for the L0Rp line and sample
    location. Calculations are done based on sensor, scene, and target specific
    information stored in the line-of-sight model.

Returns:
    SUCCESS or ERROR
*******************************************************************************/

#include <math.h>
#include "ias_los_model.h"
#include "ias_logging.h"
#include "convert_stray_light_to_lat_long.h"

/* The corrections within this routine dependent on these macro definitions
   and they are not optional. The macros are defined for CalVal for ease of
   testing. */
#define CENTER_OF_MASS 0
#define VELOCITY_ABERR 0
#define LIGHT_TRAVEL 0

static void build_los_matrix
(
    double sensor2acs[3][3],/* I: Input 3 X 3 Matrix */
    MODEL_CACHE *cache      /* O: Output matrix */
)
{
    double temp[3][3];

    ias_math_multiply_3x3_matrix(cache->orb2ecf, cache->attpert, temp);
    ias_math_multiply_3x3_matrix(temp, sensor2acs, cache->los2sc);
}

int prepare_cache_for_line
(
    int line,                           /* I: Input line number */
    int band_index,                     /* I: Input band index */
    int sca_index,                      /* I: SCA index */
    const IAS_LOS_MODEL *model,         /* I: LOS model information */
    const double major_axis,            /* I: Semi-major axis */
    const double minor_axis,            /* I: Semi-minor axis */
    MODEL_CACHE *cache                  /* O: Output matrix */
)
{
    const IAS_SENSOR_BAND_MODEL *band = &model->sensor.bands[band_index];
    IAS_VECTOR satpos;          /* Spacecraft position */
    double image_time;          /* Time in imagery */
    double delta_att_time;      /* Delta time for attitude of image time */
    double delta_eph_time;      /* Delta time for ephemeris of image time */
    IAS_VECTOR satvel;          /* Spacecraft velocity */
    double roll;                /* Spacecraft roll */
    double pitch;               /* Spacecraft pitch */
    double yaw;                 /* Spacecraft yaw */
    IAS_SENSOR_DETECTOR_TYPE dettype = IAS_NOMINAL_DETECTOR;
    int i;

    /* Moved to a new line so calculate the cached information for the
       new line */
    if (model->acquisition_type != IAS_EARTH)
    {
        IAS_LOG_ERROR("Only Earth looking imagery supported");
        return ERROR;
    }

    /* Get the time for the first detector in the band/SCA.  It should work
       for all of them. */
    if (ias_sensor_find_time(line, 0, band_index, sca_index, &model->sensor,
        dettype, &image_time) != SUCCESS)
    {
        IAS_LOG_ERROR("Finding time of SCA %d, Band index %d, L1R line %d",
            sca_index, band_index, line);
        return ERROR;
    }

    /* Get the delta time relative to the attitude epoch */
    if (ias_math_get_time_difference(band->utc_epoch_time,
        model->spacecraft.attitude.utc_epoch_time, &delta_att_time) != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating time difference between the attitude "
            "and image epoch times");
        return ERROR;
    }
    delta_att_time += image_time;

    if (ias_sc_model_find_attitude_at_time(&model->spacecraft.attitude,
        delta_att_time, &roll, &pitch, &yaw) != SUCCESS)
    {
        IAS_LOG_ERROR("Finding attitude at time %f", delta_att_time);
        return ERROR;
    }

    /* Get the delta time relative to the ephemeris epoch */
    if (ias_math_get_time_difference(band->utc_epoch_time,
        model->spacecraft.ephemeris.utc_epoch_time, &delta_eph_time) != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating time difference between the ephemeris "
            "and image epoch times");
        return ERROR;
    }
    delta_eph_time += image_time;

    ias_sc_model_get_position_and_velocity_at_time(
        &model->spacecraft.ephemeris, model->acquisition_type,
        delta_eph_time, &satpos, &satvel);

    /* Calculate the orbital-to-ECF and body-to-orbital orientation
       matrices */
    if (ias_geo_compute_orientation_matrices(&satpos, &satvel, roll,
            pitch, yaw, cache->orb2ecf, cache->attpert) != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating orientation matrices");
        return ERROR;
    }

    /* Calculate the LOS to ECEF matrix */
    build_los_matrix(band->sensor->sensor2acs, cache);

    /* Update the cache values */
    cache->line = line;
    cache->image_time = image_time;
    cache->satpos = satpos;
    cache->satvel = satvel;
    cache->roll = roll;
    cache->pitch = pitch;
    cache->yaw = yaw;
    cache->band = band;
    cache->model = model;
    cache->satpos_arr[0] = satpos.x;
    cache->satpos_arr[1] = satpos.y;
    cache->satpos_arr[2] = satpos.z;
    cache->ellipsoid[0] = major_axis;
    cache->ellipsoid[1] = major_axis;
    cache->ellipsoid[2] = minor_axis;
    cache->inv_ellipsoid[0] = 1.0 / major_axis;
    cache->inv_ellipsoid[1] = cache->inv_ellipsoid[0];
    cache->inv_ellipsoid[2] = 1.0 / minor_axis;

    /* Initialize the P vector and determine the magnitude */
    for (i = 0; i < 3; i++)
    {
        cache->P[i] = cache->satpos_arr[i] * cache->inv_ellipsoid[i];
    }
    cache->MagP = COMPUTE_VECTOR_LENGTH(cache->P);

    return SUCCESS;
}

#if 0
int convert_stray_light_to_lat_long_cached
(
    MODEL_CACHE *cache,
    const IAS_VECTOR *stray_light_vector,/* I: stray light vector in sensor
                                               coordinate system */
    double *target_latd,                /* O: Target latitude */
    double *target_long                 /* O: Target longitude */
)
{
    IAS_VECTOR new_los;         /* Earth or stellar LOS */
    IAS_VECTOR vel_aberr_los;   /* LOS vector adjusted for velocity aberration*/
    double target_latc;         /* Target geocentric latitude (radians)*/
    double target_earth_radius; /* Radius of the earth at the target  */
    double target_height;       /* Height of the target in geodetic coords */
    IAS_VECTOR target_vec;      /* Target vector */
    IAS_VECTOR satpos;          /* Spacecraft position */
    double target_elev = 0.0;   /* Target elevation (at line/sample) */

    satpos = cache->satpos;

    /* Note: The SSM is currently being ignored. In the future if deemed
             necessary it can be added back here. This code was removed per
             Cal/Val's recommendation since the image is being blurred it will
             not affect the results. */

    if (CENTER_OF_MASS)
    {
        IAS_VECTOR offpos;

        ias_geo_correct_for_center_of_mass(
            cache->band->sensor->center_mass2sensor_offset,
            cache->orb2ecf, cache->attpert, &satpos, &offpos);
        satpos.x = offpos.x;
        satpos.y = offpos.y;
        satpos.z = offpos.z;
    }

    /* Convert the LOS vector to ECEF */
    ias_math_transform_3dvec(stray_light_vector, cache->los2sc, &new_los);

    if (VELOCITY_ABERR)
    {
        if (ias_geo_correct_for_velocity_aberration(&satpos, &cache->satvel,
                cache->model->acquisition_type, &cache->model->earth,
                &new_los, &vel_aberr_los) != SUCCESS)
        {
            IAS_LOG_ERROR("Adjusting for vel. aberr.");
            return ERROR;
        }
    }
    else
    {
        vel_aberr_los = new_los;
    }

    if (LIGHT_TRAVEL)
    {
        IAS_VECTOR ltarvec;     /* Light travel time target vector */

        if(ias_geo_correct_for_light_travel_time(&satpos, &cache->model->earth,
                &target_vec, &ltarvec, &target_latc, target_long,
                &target_earth_radius) != SUCCESS)
        {
            IAS_LOG_ERROR("Correcting light travel time");
            return ERROR;
        }
    }
    else
    {
        if (ias_geo_find_target_position(&satpos, &vel_aberr_los,
                &cache->model->earth, target_elev, &target_vec, &target_latc,
                target_long, &target_earth_radius) != SUCCESS)
        {
            IAS_LOG_ERROR("Targeting Earth");
            return ERROR;
        }
    }

    /* Calculate the geodetic latitude and height of the target */
    if (ias_geo_convert_geocentric_height_to_geodetic(target_latc,
                target_earth_radius, &cache->model->earth, target_latd,
                &target_height) != SUCCESS)
    {
        IAS_LOG_ERROR("Converting geocentric lat/height to geodetic");
        return ERROR;
    }

    return SUCCESS;
}
#endif
