/***************************************************************************
Name: create_land_water_mask

Purpose: Creates the land water mask using the source L1G image and the
         stray light vectors max bounding box. Returns a pointer to the
         water mask and also the line and sample offset for the mask to
         L1G image.

Returns:
    On success: IAS_BYTE_TYPE *
    On error: NULL
****************************************************************************/
/* Standard Library Includes */
#include <stdlib.h>
#include <math.h>
#include <limits.h>

/* IAS Library Includes */
#include "ias_const.h"
#include "ias_types.h"
#include "ias_logging.h"
#include "ias_math.h"
#include "ias_miscellaneous.h"

/* Local Includes */
#include "create_tirs_ghost_image.h"
#include "create_land_water_mask.h"

IAS_BYTE_TYPE *create_land_water_mask
(
    int band_number,            /* I: Band number to process */
    int output_mask_to_envi,    /* I: Flag indicating output mask to envi */
    const char *mask_filename,  /* I: Water mask filename */
    double minimum_latitude,    /* I: Bounding box minimum latitude */
    double maximum_latitude,    /* I: Bounding box maximum latitude */
    double minimum_longitude,   /* I: Bounding box minimum longitude */
    double maximum_longitude,   /* I: Bounding box maximum longitude */
    const IAS_GEO_PROJ_TRANSFORMATION *geo2proj,/* I: Geographic to projection*/
    const L1G_INFO *l1g_info,   /* I: Source L1G image information */
    int *mask_samples,          /* O: Number of samples in mask */
    int *mask_line_offset,      /* O: Mask line offset */
    int *mask_samp_offset       /* O: Mask sample offset */
)
{
    IAS_IMAGE input_image;             /* Image data used to build mask */
    IAS_BYTE_TYPE *water_mask = NULL;  /* Water mask */
    IAS_DBL_XY projection[8];          /* Projection coordinates */
    IAS_DBL_XY max_proj;               /* Max projection coordinates */
    IAS_DBL_XY min_proj;               /* Min projection coordinates */
    int index;                         /* Generic loop index */

    /* Transform the bounding box corners */
    if (ias_geo_transform_coordinate(geo2proj, minimum_longitude,
        maximum_latitude, &projection[0].x, &projection[0].y) != SUCCESS)
    {
        IAS_LOG_ERROR("Transforming upper left corner");
        return NULL;
    }
    if (ias_geo_transform_coordinate(geo2proj, maximum_longitude,
        maximum_latitude, &projection[1].x, &projection[1].y) != SUCCESS)
    {
        IAS_LOG_ERROR("Transforming upper right corner");
        return NULL;
    }
    if (ias_geo_transform_coordinate(geo2proj, minimum_longitude,
        minimum_latitude, &projection[2].x, &projection[2].y) != SUCCESS)
    {
        IAS_LOG_ERROR("Transforming lower left corner");
        return NULL;
    }
    if (ias_geo_transform_coordinate(geo2proj, maximum_longitude,
        minimum_latitude, &projection[3].x, &projection[3].y) != SUCCESS)
    {
        IAS_LOG_ERROR("Transforming lower right corner");
        return NULL;
    }

    /* Initialize the Min/Max projection corners */
    max_proj = projection[0];
    min_proj = projection[0];

    /* Add the L1G corners to the coordinates to be processed */
    projection[4] = l1g_info->corners.upleft;
    projection[5] = l1g_info->corners.upright;
    projection[6] = l1g_info->corners.loleft;
    projection[7] = l1g_info->corners.loright;

    /* Find the Min/Max projection coordinates */
    for (index = 1; index < 8; index++)
    {
        if (max_proj.x < projection[index].x)
            max_proj.x = projection[index].x;
        if (min_proj.x > projection[index].x)
            min_proj.x = projection[index].x;
        if (max_proj.y < projection[index].y)
            max_proj.y = projection[index].y;
        if (min_proj.y > projection[index].y)
            min_proj.y = projection[index].y;
    }
    input_image.corners.upleft.x = min_proj.x;
    input_image.corners.upleft.y = max_proj.y;
    input_image.corners.loright.x = max_proj.x;
    input_image.corners.loright.y = min_proj.y;

    /* Calculate the output image corners, padding the coordinates to an
       even multiple of the output pixel size */
    input_image.corners.upleft.x = (int)floor(input_image.corners.upleft.x
        / l1g_info->x_pixel_size);
    input_image.corners.upleft.x *= l1g_info->x_pixel_size;
    input_image.corners.upleft.y = (int)ceil(input_image.corners.upleft.y
        / l1g_info->y_pixel_size);
    input_image.corners.upleft.y *= l1g_info->y_pixel_size;
    input_image.corners.loright.x = (int)ceil(input_image.corners.loright.x
        / l1g_info->x_pixel_size);
    input_image.corners.loright.x *= l1g_info->x_pixel_size;
    input_image.corners.loright.y = (int)floor(input_image.corners.loright.y
        / l1g_info->y_pixel_size);
    input_image.corners.loright.y *= l1g_info->y_pixel_size;

    /* Set the other 2 corners using the upper left and lower right corners */
    input_image.corners.upright.x = input_image.corners.loright.x;
    input_image.corners.upright.y = input_image.corners.upleft.y;
    input_image.corners.loleft.x = input_image.corners.upleft.x;
    input_image.corners.loleft.y = input_image.corners.loright.y;

    /* Copy to image data to image struct */
    input_image.data_type = IAS_BYTE;
    input_image.pixel_size_y = l1g_info->x_pixel_size;
    input_image.pixel_size_x = l1g_info->y_pixel_size;
    input_image.nl = round((input_image.corners.upleft.y
        - input_image.corners.loright.y) / l1g_info->y_pixel_size);
    input_image.ns = round((input_image.corners.loright.x
        - input_image.corners.upleft.x) / l1g_info->x_pixel_size);
    input_image.band_number = band_number;
    *mask_line_offset = round((input_image.corners.upleft.y
        - l1g_info->corners.upleft.y) / l1g_info->y_pixel_size);
    *mask_samp_offset = round((l1g_info->corners.upleft.x
        - input_image.corners.upleft.x) / l1g_info->x_pixel_size);
    *mask_samples = input_image.ns;

    /*----------------------------DEBUG INFO----------------------------------*/
    IAS_LOG_DEBUG("Water Mask Corners:(%f,%f), (%f,%f), (%f,%f), (%f,%f)",
        input_image.corners.upleft.x, input_image.corners.upleft.y,
        input_image.corners.upright.x, input_image.corners.upright.y,
        input_image.corners.loleft.x, input_image.corners.loleft.y,
        input_image.corners.loright.x, input_image.corners.loright.y);
    IAS_LOG_DEBUG("Water mask lines = %d", input_image.nl);
    IAS_LOG_DEBUG("Water mask samples = %d", input_image.ns);
    IAS_LOG_DEBUG("Water mask lines offset = %d", *mask_line_offset);
    IAS_LOG_DEBUG("Water mask samples offset = %d", *mask_samp_offset);
    /*-------------------------END DEBUG INFO---------------------------------*/

    /* Allocate memory for the water mask */
    water_mask = calloc(input_image.nl * input_image.ns, sizeof(*water_mask));
    if (!water_mask)
    {
        IAS_LOG_ERROR("Allocating memory for water mask");
        return NULL;
    }

    /* Build mask and polygon list */
    if (ias_geo_shape_mask_projection(mask_filename, &input_image,
        &l1g_info->proj, water_mask) != SUCCESS)
    {
        IAS_LOG_ERROR("Creating the water mask");
        free(water_mask);
        return NULL;
    }

    /* Output water mask to envi if requested */
    if (output_mask_to_envi)
    {
        char output_filename[PATH_MAX]; /* Water mask output filename */
        int status;                     /* Generic function return status */

        /* Create the water mask output filename */
        status = snprintf(output_filename, sizeof(output_filename),
            "%s/water_mask_B%d.img", TIRS_STRAY_LIGHT_DIRECTORY, band_number);
        if (status < 0 || status >= sizeof(output_filename))
        {
            IAS_LOG_ERROR("Output filename buffer is too small - needs %d"
                " bytes", status);
            free(water_mask);
            return NULL;
        }

        /* Dump the water_mask to file */
        if (ias_misc_write_envi_image(output_filename, &l1g_info->proj,
            "Water mask image", input_image.nl, input_image.ns,
            input_image.corners.upleft.x, input_image.corners.upleft.y,
            input_image.pixel_size_x, input_image.pixel_size_y, "Water Mask",
            input_image.data_type, water_mask) != SUCCESS)
        {
            IAS_LOG_ERROR("Writing water mask %s to file", output_filename);
            free(water_mask);
            return NULL;
        }
    }

    return water_mask;
}
