/******************************************************************************
Name: create_tirs_ghost_image

Purpose: Create the TIRS ghost image caused by stray light contamination.

Notes: Based on prototype code from M. Montanaro.

******************************************************************************/
/* Standard Library Includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>

/* IAS Library Includes */
#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "ias_model_io.h"
#include "ias_threadpool.h"
#include "ias_l1r.h"
#include "ias_miscellaneous.h"

/* Local Includes */
#include "get_l1g.h"
#include "grid.h"
#include "find_image_corners.h"
#include "read_l0r_metadata.h"
#include "create_tirs_ghost_image.h"
#include "blur_l1g_image.h"
#include "read_stray_light_vectors.h"
#include "read_radiance_coefficients.h"
#include "sample_stray_light.h"
#include "read_envi_image.h"
#include "create_land_water_mask.h"
#include "calculate_average_land_water_radiance.h"
#include "write_l1r_metadata.h"
#include "create_projection_transformation.h"

#define USAGE "USAGE: create_tirs_ghost_image <ParameterFile>"

/* Define this to write a flat file for each band/sca with an ENVI header */
#define WRITE_ENVI_FILE

/*****************************************************************************
Name: main (create_tirs_ghost_image)

Purpose: This is the main routine to create the TIRS ghost image.

Returns: EXIT_SUCCESS or EXIT_FAILURE

*****************************************************************************/
int main(int argc, char *argv[])
{
    L1G_INFO in_l1g_info; /* L1G imagery and metadata */
    L1G_INFO blurred_l1g; /* Downsampled L1G */
    IAS_GEO_PROJ_TRANSFORMATION *geo2l1g = NULL; /* Geographic to L1G trans */
    GHOST_IMAGE_PARAMETERS parameters;
    IAS_LOS_MODEL *model_ptr;
    L0R_METADATA l0r_metadata;
    STRAY_LIGHT_VECTORS stray_light_vectors;
    RADIANCE_COEFFICIENTS rad_coeffs;
    SAMPLE_PARAMS ghost_params;
    L1RIO *l1r_image = NULL;
    L1R_BAND_IO *l1r_band = NULL;
    GRID grid;
    int l1r_frames;
    struct ias_threadpool *threadpool_ptr;
    int is_envi_source_file = 0;
    int len;
    int band_index;
    float *ghost_image = NULL;
    int error_occurred = FALSE;

    /* Initialize the logging library */
    if (ias_log_initialize("Create TIRS Ghost Image") != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing logging library");
        exit(EXIT_FAILURE);
    }

    /* Make sure enough arguments were provided */
    if (argc < 2)
    {
        IAS_LOG_ERROR(USAGE);
        exit(EXIT_FAILURE);
    }

    /* Read the parameters */
    if (read_parameters(argv[1], &parameters) != SUCCESS)
    {
        IAS_LOG_ERROR("Failed to read input parameters from %s", argv[1]);
        exit(EXIT_FAILURE);
    }

    /* If the source filename ends with .img, assume it is an ENVI file */
    len = strlen(parameters.source_image_filename);
    if (len > 4
        && strcmp(&parameters.source_image_filename[len - 4], ".img") == 0)
    {
        is_envi_source_file = 1;
    }

    /* Issue message if using the average radiances */
    if (parameters.use_land_water_radiance_average_flag)
    {
        IAS_LOG_INFO("Using land and water radiance averages");
    }

    /* Read the L0R metadata */
    if (read_l0r_metadata(parameters.l0r_filename,
        parameters.landsat_scene_id, &l0r_metadata) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading L0R metadata from %s for scene ID %s",
            parameters.l0r_filename, parameters.landsat_scene_id);
        exit(EXIT_FAILURE);
    }
    IAS_LOG_DEBUG("L0R frames: %d %d %d %d %d %d",
        l0r_metadata.subinterval_frames, l0r_metadata.subinterval_start_frame,
        l0r_metadata.subinterval_end_frame, l0r_metadata.target_scene_frames,
        l0r_metadata.target_scene_start_frame,
        l0r_metadata.target_scene_end_frame);

    l1r_frames = l0r_metadata.target_scene_end_frame -
                l0r_metadata.target_scene_start_frame + 1;

    /* If the sampling location option has been selected, adjust the frame
       requested to be relative to the sub-interval start instead of the scene
       start since the sampling code is looking relative to the start of the
       sub-interval */
    parameters.sampling_location.frame_index
        += l0r_metadata.target_scene_start_frame;

    /* Read the a/b coefficients */
    if (read_radiance_coefficients(parameters.radiance_coefficient_filename,
        &rad_coeffs) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading 'a' and 'b' coefficients from %s",
            parameters.radiance_coefficient_filename);
        exit(EXIT_FAILURE);
    }

    /* Read the stray light optics vector data */
    IAS_LOG_INFO("Reading stray light optics vector data");
    if (read_stray_light_vectors(parameters.stray_light_vector_filename,
        &stray_light_vectors) != SUCCESS)
    {
        IAS_LOG_ERROR("Failed reading stray light vectors from %s",
            parameters.stray_light_vector_filename);
        exit(EXIT_FAILURE);
    }

    /* Read the LOS model file */
    IAS_LOG_INFO("Reading LOS model");
    model_ptr = ias_model_read(parameters.model_filename);
    if (!model_ptr)
    {
        IAS_LOG_ERROR("Reading LOS model from %s",
            parameters.model_filename);
        free_stray_light_vectors(&stray_light_vectors);
        exit(EXIT_FAILURE);
    }

    /* Create the sampling mutex */
    if (IAS_THREAD_CREATE_MUTEX(&ghost_params.frame_mutex) != 0)
    {
        IAS_LOG_ERROR("Creating frame mutex");
        free_stray_light_vectors(&stray_light_vectors);
        ias_los_model_free(model_ptr);
        exit(EXIT_FAILURE);
    }

    /* Open the output L1G ghost image */
    l1r_image = ias_l1r_open_image(parameters.stray_light_correction_filename,
        IAS_WRITE);
    if (!l1r_image)
    {
        IAS_LOG_ERROR("Opening output L1R ghost image %s",
            parameters.stray_light_correction_filename);
        free_stray_light_vectors(&stray_light_vectors);
        ias_los_model_free(model_ptr);
        IAS_THREAD_DESTROY_MUTEX(&ghost_params.frame_mutex);
        exit(EXIT_FAILURE);
    }

    /* Initialize the threadpool */
    threadpool_ptr = ias_threadpool_initialize(parameters.max_processors);
    if (!threadpool_ptr)
    {
        IAS_LOG_ERROR("Initializing the thread pool");
        free_stray_light_vectors(&stray_light_vectors);
        ias_los_model_free(model_ptr);
        IAS_THREAD_DESTROY_MUTEX(&ghost_params.frame_mutex);
        ias_l1r_close_image(l1r_image);
        exit(EXIT_FAILURE);
    }

    /* Loop over the bands in the TIRS image */
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        const IAS_BAND_ATTRIBUTES *band_attr;
        char output_filename[PATH_MAX];
        int band_number;
        int sca_index;
        int sca_count;
        int detector_count;
        int corner_lines[4];
        int corner_samples[4];
        IAS_DATA_TYPE output_type = IAS_R4;
        IAS_DBL_LAT_LONG max_coords =  {-91.0, -361.0};
        IAS_DBL_LAT_LONG min_coords =  {91.0, 361.0};
        int valid_grid_flag;
        int source_band_number;
        int status;

        /* Retrieve the band number using the band index */
        band_number = ias_sat_attr_convert_band_index_to_number(
            band_index);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Converting band index %d to band number",
                band_index);
            error_occurred = TRUE;
            break;
        }
        source_band_number = band_number;

        band_attr = ias_sat_attr_get_band_attributes(band_number);
        if (!band_attr)
        {
            IAS_LOG_ERROR("Retrieving band attributes for band number %d",
                band_number);
            error_occurred = TRUE;
            break;
        }

        /* We are only concerned with the TIRS bands */
        /* Check if this band should be processed */
        if (!parameters.process_band[band_index]
            || band_attr->sensor_id != IAS_TIRS)
        {
            IAS_LOG_INFO("Skipping band number %d", band_number);
            continue;
        }

        sca_count = band_attr->scas;
        detector_count = band_attr->detectors_per_sca;
        ghost_params.detector_count = detector_count;

        /* The ghost image contains l1r_frames of data */
        ghost_image = malloc(sizeof(*ghost_image) * l1r_frames
            * detector_count);
        if (!ghost_image)
        {
            IAS_LOG_ERROR("Allocating memory for the ghost image");
            error_occurred = TRUE;
            break;
        }

        /* If the source data is not TIRS, assume a single band */
        if (strcmp(parameters.source_image_sensor, "TIRS") != 0)
            source_band_number = 1;

        IAS_LOG_INFO("Reading radiance source image for band number %d",
            source_band_number);

        if (is_envi_source_file)
        {
            if (read_envi_image(parameters.source_image_filename,
                source_band_number, parameters.spheroid, &in_l1g_info)
                != SUCCESS)
            {
                IAS_LOG_ERROR("Reading radiance source image from %s",
                    parameters.source_image_filename);
                free(ghost_image);
                error_occurred = TRUE;
                break;
            }
        }
        else
        {
            /* Read the L1G image */
            if (get_l1g(parameters.source_image_filename, source_band_number,
                &in_l1g_info) != SUCCESS)
            {
                IAS_LOG_ERROR("Reading radiance source image from %s",
                    parameters.source_image_filename);
                error_occurred = TRUE;
                break;
            }
        }

        /* Blur the input image the requested amount.  Blurring was chosen
           over downsampling since it left artifacts in the ghost image. */
        IAS_LOG_INFO("Blurring the input image using a kernel size of %d",
            parameters.blur_kernel_size);
        if (blur_l1g_image(&in_l1g_info, parameters.blur_kernel_size,
            &blurred_l1g) != SUCCESS)
        {
            IAS_LOG_ERROR("Failed Blurring the L1G image");
            error_occurred = TRUE;
            break;
        }

        if (parameters.output_files_to_envi_flag)
        {
            status = snprintf(output_filename, sizeof(output_filename),
                "%s/blurred_B%d.img", TIRS_STRAY_LIGHT_DIRECTORY, band_number);
            if (status < 0 || status >= sizeof(output_filename))
            {
                IAS_LOG_ERROR("Output filename buffer is too small - needs %d"
                    " bytes", status);
                error_occurred = TRUE;
                break;
            }

            /* Write blurred file to envi image */
            if (ias_misc_write_envi_image(output_filename, &blurred_l1g.proj,
                "Blurred image with sampled & weighted radiances",
                blurred_l1g.lines, blurred_l1g.samples,
                blurred_l1g.corners.upleft.x, blurred_l1g.corners.upleft.y,
                blurred_l1g.x_pixel_size, blurred_l1g.y_pixel_size,
                "Blurred L1G", IAS_R4, blurred_l1g.radiance) != SUCCESS)
            {
                IAS_LOG_ERROR("Writing blurred image %s", output_filename);
                error_occurred = TRUE;
                break;
            }
        }

        /* Find the image corners using the original input image.  This allows
           the sampling process to avoid the edges of the blurred image which
           is larger than the original input image. */
        if (find_image_corners(in_l1g_info.radiance, in_l1g_info.lines,
            in_l1g_info.samples, corner_lines, corner_samples) != SUCCESS)
        {
            IAS_LOG_ERROR("Finding input L1G image corners");
            error_occurred = TRUE;
            break;
        }

        /* Create the projection transformation */
        geo2l1g = create_geographic_to_target_transformation(&blurred_l1g.proj);
        if (!geo2l1g)
        {
            IAS_LOG_ERROR("Creating projection transformation");
            error_occurred = TRUE;
            break;
        }

        IAS_LOG_INFO("Finding bounding box");

        /* Find the bounding box of the stray light vectors for this band */
        for (sca_index = 0; sca_index < sca_count; sca_index++)
        {
            IAS_DBL_LAT_LONG sca_max;
            IAS_DBL_LAT_LONG sca_min;

            /* Find the bounding box using the model */
            if (find_bounding_box_using_model(model_ptr, &stray_light_vectors,
                band_index, sca_index, l0r_metadata.subinterval_frames,
                parameters.semi_major_axis, parameters.semi_minor_axis,
                parameters.number_bounding_box_frame_sections,
                &sca_min.lat, &sca_max.lat, &sca_min.lng, &sca_max.lng)
                != SUCCESS)
            {
                IAS_LOG_ERROR("Finding bounding box with model");
                error_occurred = TRUE;
                break;
            }

            if (sca_min.lat < min_coords.lat)
                min_coords.lat = sca_min.lat;
            if (sca_max.lat > max_coords.lat)
                max_coords.lat = sca_max.lat;
            if (sca_min.lng < min_coords.lng)
                min_coords.lng = sca_min.lng;
            if (sca_max.lng > max_coords.lng)
                max_coords.lng = sca_max.lng;
        }

        /* Check for SCA processing error */
        if (error_occurred)
            break;

        IAS_LOG_DEBUG("lat min/max %f, %f   lng min/max %f, %f", min_coords.lat,
            max_coords.lat, min_coords.lng, max_coords.lng);

        /* Create a grid for converting lat/long to projection */
        IAS_LOG_INFO("Creating grid");
        if (create_grid(geo2l1g, min_coords.lng, max_coords.lat, max_coords.lng,
            min_coords.lat, in_l1g_info.x_pixel_size, in_l1g_info.y_pixel_size,
            &grid, &valid_grid_flag) != SUCCESS)
        {
            IAS_LOG_ERROR("Creating grid");
            error_occurred = TRUE;
            break;
        }

        /* Set up the parameters for the sampling routine */
        ghost_params.model = model_ptr;
        ghost_params.grid = &grid;
        ghost_params.valid_grid_flag = valid_grid_flag;
        ghost_params.l1g_info = &blurred_l1g;
        ghost_params.start_frame = l0r_metadata.target_scene_start_frame;
        ghost_params.end_frame = l0r_metadata.target_scene_end_frame;
        ghost_params.l1r_frames = l1r_frames;
        ghost_params.band_index = band_index;
        ghost_params.band_number = band_number;
        ghost_params.sample_location_requested_flag
            = parameters.sample_location_requested_flag;
        ghost_params.sampling_location = &parameters.sampling_location;
        ghost_params.apply_radiance_coefficients
            = parameters.apply_radiance_coefficients;
        ghost_params.corner_lines = corner_lines;
        ghost_params.corner_samples = corner_samples;
        ghost_params.stray_light_vectors = &stray_light_vectors;
        ghost_params.rad_coeffs = &rad_coeffs;
        ghost_params.ghost = ghost_image;
        ghost_params.use_land_water_radiance_average_flag
            = parameters.use_land_water_radiance_average_flag;
        ghost_params.water_mask = NULL;
        ghost_params.semi_major_axis = parameters.semi_major_axis;
        ghost_params.semi_minor_axis = parameters.semi_minor_axis;

        /* Retrieve the water mask and calculate the average land and
           water radiances using the orginal L1G image if requested */
        if (parameters.use_land_water_radiance_average_flag)
        {
            /* Create the water mask */
            IAS_LOG_INFO("Creating the land/water mask");
            ghost_params.water_mask = create_land_water_mask(band_number,
                parameters.output_files_to_envi_flag,
                parameters.water_mask_filename, min_coords.lat, max_coords.lat,
                min_coords.lng, max_coords.lng, geo2l1g, &in_l1g_info,
                &ghost_params.mask_samples, &ghost_params.mask_line_offset,
                &ghost_params.mask_samp_offset);
            if (!ghost_params.water_mask)
            {
                IAS_LOG_ERROR("Creating the land/water mask for band number %d",
                    band_number);
                error_occurred = TRUE;
                break;
            }

            /* Calculate the land and water radiance averages */
            IAS_LOG_INFO("Calculating the land/water radiance averages");
            calculate_average_land_water_radiance(in_l1g_info.lines,
                in_l1g_info.samples, ghost_params.mask_samples,
                ghost_params.mask_line_offset, ghost_params.mask_samp_offset,
                in_l1g_info.radiance, ghost_params.water_mask,
                &blurred_l1g.average_water_radiance,
                &blurred_l1g.average_land_radiance);
        }

        /* Open the output band */
        l1r_band = ias_l1r_open_band(l1r_image, band_number, &output_type,
            &sca_count, &l1r_frames, &detector_count);
        if (!l1r_band)
        {
            IAS_LOG_ERROR("Opening band number %d of %s", band_number,
                parameters.stray_light_correction_filename);
            error_occurred = TRUE;
            break;
        }

        /* Loop over the SCAs in the current band */
        for (sca_index = 0; sca_index < sca_count; sca_index++)
        {

            IAS_LOG_INFO("Sampling band number %d, SCA number %d",
                band_number, sca_index + 1);

            /* Zero out the ghost image.  At least one detector does not have
               any stray light vectors, so we need to ensure the ghost image
               is set to zero for that. */
            memset(ghost_image, 0, sizeof(*ghost_image) * l1r_frames
                * detector_count);

            /* Set the correct SCA index for the sampling routine */
            ghost_params.sca_index = sca_index;
            ghost_params.next_frame = ghost_params.start_frame;
            ghost_params.outside_grid_flag = 0;

            /* Sample radiances from the input image at the stray light vector
               locations */
            status = ias_threadpool_run_function(threadpool_ptr,
                sample_stray_light, &ghost_params);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Creating ghost image for band number %d,"
                    " SCA %d", band_number, sca_index + 1);
                error_occurred = TRUE;
                break;
            }

            /* If any points were outside the grid, issue a warning.  If it
               happens and someone is interested, debug can be turned on to
               show how often it happens */
            if (ghost_params.outside_grid_flag)
            {
                IAS_LOG_WARNING("Some points were outside the grid, probably "
                    "due to unusual conditions near the poles and/or "
                    "off-nadir");
            }

            /* Write the image to the L1R formatted image */
            if (ias_l1r_write_image(l1r_band, sca_index, 0, 0, l1r_frames,
                detector_count, ghost_image) != SUCCESS)
            {
                IAS_LOG_ERROR("Failed writing band number %d, SCA number %d "
                    "to %s", band_number, sca_index + 1,
                    parameters.stray_light_correction_filename);
                error_occurred = TRUE;
                break;
            }

            if (parameters.output_files_to_envi_flag)
            {
                /* Build the target band/SCA filename */
                status = snprintf(output_filename, sizeof(output_filename),
                    "%s/%s_B%d_SCA%d", TIRS_STRAY_LIGHT_DIRECTORY,
                    parameters.stray_light_correction_filename,
                    band_number, sca_index+1);
                if (status < 0 || status >= sizeof(output_filename))
                {
                    IAS_LOG_ERROR("Output filename buffer is too small - "
                        "needs %d bytes", status);
                    error_occurred = TRUE;
                    break;
                }

                /* Write output file to envi image */
                if (ias_misc_write_envi_image(output_filename, NULL,
                    "Stray light correction ghost image",
                    l0r_metadata.target_scene_frames, detector_count,
                    0.0, 0.0, 0.0, 0.0, NULL, IAS_R4, ghost_image) != SUCCESS)
                {
                    IAS_LOG_ERROR("Writing output image %s",
                        output_filename);
                    error_occurred = TRUE;
                    break;
                }
            }
        } /* End SCA loop */

        /* Check for SCA processing error */
        if (error_occurred)
            break;

        /* Close the output band */
        if (ias_l1r_close_band(l1r_band) != SUCCESS)
        {
            IAS_LOG_ERROR("Closing band number %d of %s", band_number,
                parameters.stray_light_correction_filename);
            error_occurred = TRUE;
            l1r_band = NULL;
            break;
        }
        l1r_band = NULL;

        /* Free the water mask */
        free(ghost_params.water_mask);
        ghost_params.water_mask = NULL;

        /* Free the grid */
        free_grid(&grid);

        /* Destroy the projection transformation */
        ias_geo_destroy_proj_transformation(geo2l1g);
        geo2l1g = NULL;

        /* Free the blurred radiance */
        free(blurred_l1g.radiance);
        blurred_l1g.radiance = NULL;

        /* Free the L1G radiance */
        free(in_l1g_info.radiance);
        in_l1g_info.radiance = NULL;

        /* Free the ghost image */
        free(ghost_image);
        ghost_image = NULL;

    } /* End band loop */

    if (IAS_THREAD_DESTROY_MUTEX(&ghost_params.frame_mutex) != 0)
    {
        IAS_LOG_ERROR("Failed destroying the frame mutex");
        error_occurred = TRUE;
    }

    /* Destroy the main thread pool */
    ias_threadpool_destroy(threadpool_ptr);
    threadpool_ptr = NULL;

    ias_los_model_free(model_ptr);
    model_ptr = NULL;

    /* Free all the memory */
    free_stray_light_vectors(&stray_light_vectors);

    /* Check for band processing error */
    if (error_occurred)
    {
        /* Free the water mask */
        free(ghost_params.water_mask);
        ghost_params.water_mask = NULL;

        /* Free the grid */
        free_grid(&grid);

        /* Destroy the projection transformation */
        ias_geo_destroy_proj_transformation(geo2l1g);
        geo2l1g = NULL;

        /* Free the blurred radiance */
        free(blurred_l1g.radiance);
        blurred_l1g.radiance = NULL;

        /* Free the L1G radiance */
        free(in_l1g_info.radiance);
        in_l1g_info.radiance = NULL;

        /* Free the ghost image */
        free(ghost_image);
        ghost_image = NULL;

        /* Close the output band and image */
        if (l1r_band)
            ias_l1r_close_band(l1r_band);
        ias_l1r_close_image(l1r_image);

        /* All memory has been freed exit with an error */
        exit(EXIT_FAILURE);
    }

    /* Write out the l1r metadata */
    if (write_l1r_metadata(parameters.source_image_filename,
        parameters.source_image_sensor,
        parameters.use_land_water_radiance_average_flag, l1r_image) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing metadata to output L1R ghost image");
        exit(EXIT_FAILURE);
    }

    /* Close the output ghost image */
    if (ias_l1r_close_image(l1r_image) != SUCCESS)
    {
        IAS_LOG_ERROR("Failed closing the output L1R ghost image");
        exit(EXIT_FAILURE);
    }

    if (update_omf(parameters.omf_filename,
            parameters.stray_light_correction_filename) != SUCCESS)
    {
        IAS_LOG_ERROR("Updating the OMF file");
        exit(EXIT_FAILURE);
    }

    IAS_LOG_INFO("Finished");
    exit(EXIT_SUCCESS);
}
