/******************************************************************************
Name: find_pixel_nearest_polygon

Purpose: Finds the closest valid point in the provided corners to the
         input line/sample

Notes: Only supports a polygon with 4 corners for now.  If we ever use
   an L1G with crenelation on the top/bottom, this will need to change.

******************************************************************************/
/* Standard Library Includes */
#include <stdlib.h>
#include <math.h>

/* IAS Library Includes */
#include "ias_const.h"
#include "ias_logging.h"

/* Local Includes */
#include "find_pixel_nearest_polygon.h"

/* Local Defines */
#define dot(v1, v2) ((v1).x * (v2).x + (v1).y * (v2).y)
#define cross_z(v1, v2) ((v1).x * (v2).y - (v1).y * (v2).x)

int find_pixel_nearest_polygon
(
    double in_line,            /* I: Line coordinate */
    double in_sample,          /* I: Sample coordinate */
    const int *corner_lines,   /* I: Corner lines array */
    const int *corner_samples, /* I: Corner samples array */
    const L1G_INFO *l1g_info,  /* I: Input L1G data */
    int *return_line,          /* O: Nearest pixel line */
    int *return_sample,        /* O: Nearest pixel sample */
    float *radiance            /* O: Nearest pixel radiance */
)
{
    int cur_corner;
    double found_line;
    double found_sample;
    int line;
    int sample;
    float pixel;
    IAS_DBL_XY v;
    IAS_DBL_XY perp_v;
    double mag;

    /* Loop through the corners */
    for (cur_corner = 0; cur_corner < 4; cur_corner++)
    {
        IAS_DBL_XY w;
        int next_corner = (cur_corner + 1) % 4;
        double w_proj;
        double v_proj;
        double ratio;

        /* Create a vector from this line segment */
        v.x = corner_samples[next_corner] - corner_samples[cur_corner];
        v.y = corner_lines[next_corner] - corner_lines[cur_corner];

        /* Create the vector from the point and the first vertex of the line
           segment */
        w.x = in_sample - corner_samples[cur_corner];
        w.y = in_line - corner_lines[cur_corner];

        /* If Z component of the cross product is positive, the point is inside
           the polygon relative to this point, so check the next edge.  Note the
           sign is reversed from what is obvious because the Y axis is
           inverted in line/sample space. */
        if (cross_z(v, w) > 0.0)
        {
            continue;
        }

        /* Project w onto the line segment */
        w_proj = dot(w, v);

        /* If the projection of w onto v is less than or equal to zero, the
           point is off the end of the line before the line segment starting
           point, so that starting point is the closest. */
        if (w_proj <= 0.0)
        {
            found_line = corner_lines[cur_corner];
            found_sample = corner_samples[cur_corner];
            break;
        }

        /* Project v onto itself to get the length squared of the line
           segment */
        v_proj = dot(v, v);

        /* If the w projection onto the line is larger than the length of the
           line segment, the point is past the end of the line segment, so the
           ending point is the closest */
        if (w_proj >= v_proj)
        {
            found_line = corner_lines[next_corner];
            found_sample = corner_samples[next_corner];
            break;
        }

        /* The ratio of the projections can be used to find the closest location
           on the line */
        ratio = w_proj / v_proj;
        found_line = corner_lines[cur_corner] + ratio * v.y;
        found_sample = corner_samples[cur_corner] + ratio * v.x;

        /* Found the closest point on the polygon, so break out of the loop */
        break;
    }
    if (cur_corner == 4)
    {
        /* This can happen with the curvature at the edge of the image.
           A pixel can fall inside the bounding lines but still be fill.  Try
           to eliminate this which is largely copied from above.  The best
           way to trigger this is to use the direct LOS vectors to create
           the L1R from the L1T.  Note that this is largely eliminated by
           the image blurring since that expands the image enough that it
           is very unlikely to have fill inside the edge lines. */
        double min_dist_sq = 10000000000.0;
        double closest_line = 0.0;
        double closest_sample = 0.0;
        IAS_DBL_XY closest_v = {0, 0};
        int found = 0;

        for (cur_corner = 0; cur_corner < 4; cur_corner++)
        {
            IAS_DBL_XY w;
            int next_corner = (cur_corner + 1) % 4;
            double w_proj;
            double v_proj;
            double ratio;

            /* Create a vector from this line segment */
            v.x = corner_samples[next_corner] - corner_samples[cur_corner];
            v.y = corner_lines[next_corner] - corner_lines[cur_corner];

            /* Create the vector from the point and the first vertex of the line
               segment */
            w.x = in_sample - corner_samples[cur_corner];
            w.y = in_line - corner_lines[cur_corner];

            /* Project w onto the line segment */
            w_proj = dot(w, v);

            /* If the projection of w onto v is less than or equal to zero, the
               point is off the end of the line before the line segment starting
               point, so that starting point is the closest. */
            if (w_proj <= 0.0)
            {
                found_line = corner_lines[cur_corner];
                found_sample = corner_samples[cur_corner];
            }
            else
            {
                /* Project v onto itself to get the length squared of the line
                   segment */
                v_proj = dot(v, v);

                /* If the w projection onto the line is larger than the length
                   of the line segment, the point is past the end of the line
                   segment, so the ending point is the closest */
                if (w_proj >= v_proj)
                {
                    found_line = corner_lines[next_corner];
                    found_sample = corner_samples[next_corner];
                }
                else
                {
                    /* The ratio of the projections can be used to find the
                       closest location on the line */
                    ratio = w_proj / v_proj;
                    found_line = corner_lines[cur_corner] + ratio * v.y;
                    found_sample = corner_samples[cur_corner] + ratio * v.x;
                }
            }
            double s_diff = in_sample - found_sample;
            double l_diff = in_line - found_line;
            double dist_sq = s_diff * s_diff + l_diff * l_diff;
            if (dist_sq < min_dist_sq)
            {
                min_dist_sq = dist_sq;
                closest_line = found_line;
                closest_sample = found_sample;
                closest_v = v;
                found = 1;
            }
        }
        if (!found)
        {
            IAS_LOG_ERROR("Failure to find closest pixel");
            return ERROR;
        }
        found_line = closest_line;
        found_sample = closest_sample;
        v = closest_v;
    }

    /* Round the location to the nearest line/sample */
    line = round(found_line);
    sample = round(found_sample);

    /* If the L1G image value at that location is non-zero, we're done */
    pixel = l1g_info->radiance[line * l1g_info->samples + sample];
    if (pixel != 0.0)
    {
        *radiance = pixel;
        *return_line = line;
        *return_sample = sample;
        return SUCCESS;
    }

    /* The edge line is still in fill, so look inward to find the nearest
       valid pixel.  Find the unit vector perpendicular to the edge that looks
       inward.  Since the corners are in clockwise order, the perpendicular
       direction is found by swapping the vector coordinates and negating the
       Y term to go in the clockwise direction. */
    perp_v.x = -v.y;
    perp_v.y = v.x;

    /* Make a unit vector */
    mag = sqrt(dot(perp_v, perp_v));
    perp_v.x /= mag;
    perp_v.y /= mag;

    /* Walk the unit vector until a non-fill pixel is found.  If we have to
       walk more than 10 unit vectors, something must be wrong */
    for (cur_corner = 1; cur_corner < 10; cur_corner++)
    {
        int next_l = round(line + cur_corner * perp_v.y);
        int next_s = round(sample + cur_corner * perp_v.x);

        pixel = l1g_info->radiance[next_l * l1g_info->samples + next_s];
        if (pixel != 0.0)
        {
            *radiance = pixel;
            *return_line = line;
            *return_sample = sample;
            return SUCCESS;
        }
    }

    /* If we get here, no valid radiance found... return an error */
    IAS_LOG_ERROR("Failed to find a valid radiance");
    return ERROR;
}
