/******************************************************************************
Name: get_l1g

Purpose: Reads the provided L1G image band

Note: Caller is responsible for freeing the l1g_info->radiance memory on
      a successful function call

Returns: Type = Integer
         SUCCESS / ERROR
******************************************************************************/
/* Standard Library Includes */
#include <stdlib.h>

/* IAS Library Includes */
#include "ias_l1g.h"
#include "ias_miscellaneous.h"
#include "ias_logging.h"

/* Local Includes */
#include "get_l1g.h"

int get_l1g
(
    const char *filename,   /* I: Name of the L1G image to read */
    int band_number,        /* I: Band number to read */
    L1G_INFO *l1g_info      /* O: Structure to return info in */
)
{
    L1GIO *l1g_file;
    L1G_BAND_IO *l1g_band;
    IAS_L1G_FILE_METADATA fmd;
    IAS_L1G_BAND_METADATA bmd;
    IAS_DATA_TYPE data_type;
    int scas;
    int lines;
    int samples;
    int size;
    void *orig_imagery;
    float *radiance;
    int units;

    /* Open the L1G file */
    l1g_file = ias_l1g_open_image(filename, IAS_READ);
    if (!l1g_file)
    {
        IAS_LOG_ERROR("Opening L1G image %s", filename);
        return ERROR;
    }

    /* Verify the band is present */
    if (!ias_l1g_is_band_present(l1g_file, band_number))
    {
        IAS_LOG_ERROR("Band number %d is not available in %s", band_number,
            filename);
        ias_l1g_close_image(l1g_file);
        return ERROR;
    }

    /* Get the file metadata */
    if (ias_l1g_get_file_metadata(l1g_file, &fmd) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading file metadata from %s", filename);
        ias_l1g_close_image(l1g_file);
        return ERROR;
    }

    /* Get the band metadata */
    if (ias_l1g_get_band_metadata(l1g_file, band_number, &bmd) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading band metadata from %s for band number %d",
            filename, band_number);
        ias_l1g_close_image(l1g_file);
        return ERROR;
    }

    /* Open the band */
    l1g_band = ias_l1g_open_band(l1g_file, band_number, &data_type, &scas,
            &lines, &samples);
    if (!l1g_band)
    {
        IAS_LOG_ERROR("Reading band size info from %s for band number %d",
            filename, band_number);
        ias_l1g_close_image(l1g_file);
        return ERROR;
    }

    /* Make sure there is a single SCA in the image */
    if (scas != 1)
    {
        IAS_LOG_ERROR("Found %d SCAs in the image %s instead of the expected 1",
            scas, filename);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(l1g_file);
        return ERROR;
    }

    /* Make sure the data type is either float or unsigned short int */
    if (data_type != IAS_R4 && data_type != IAS_UI2)
    {
        IAS_LOG_ERROR("Unexpected imagery data type found in %s", filename);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(l1g_file);
        return ERROR;
    }

    /* If the data type is short int, make sure the scaling factors to get to
       radiance are valid */
    if (data_type == IAS_UI2 && !bmd.radiance_valid)
    {
        IAS_LOG_ERROR("Radiance scaling parameters are not valid for band "
            "number %d in %s", band_number, filename);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(l1g_file);
        return ERROR;
    }

    /* Allocate memory for the imagery */
    if (ias_misc_get_sizeof_data_type(data_type, &size) != SUCCESS)
    {
        IAS_LOG_ERROR("Getting size of data type %d", data_type);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(l1g_file);
        return ERROR;
    }
    size *= lines * samples;

    orig_imagery = malloc(size);
    if (!orig_imagery)
    {
        IAS_LOG_ERROR("Allocating %d bytes of memory for the imagery",
            size);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(l1g_file);
        return ERROR;
    }

    /* Read the whole band of imagery */
    if (ias_l1g_read_image(l1g_band, 0, 0, 0, lines, samples, orig_imagery)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Reading band number %d from %s", band_number,
            filename);
        free(orig_imagery);
        ias_l1g_close_band(l1g_band);
        ias_l1g_close_image(l1g_file);
        return ERROR;
    }

    /* Close the band and image (no need to make sure it closed successfully) */
    ias_l1g_close_band(l1g_band);
    l1g_band = NULL;
    ias_l1g_close_image(l1g_file);
    l1g_file = NULL;

    /* Convert the imagery to radiance (if needed) */
    if (data_type == IAS_UI2)
    {
        int index;
        int elements = lines * samples;
        float gain = bmd.spectral_radiance_scaling_gain;
        float offset = bmd.spectral_radiance_scaling_offset;
        unsigned short *dn = orig_imagery;

        /* Allocate memory for the converted data */
        radiance = malloc(elements * sizeof(float));
        if (!radiance)
        {
            IAS_LOG_ERROR("Allocating memory for the radiance buffer");
            free(orig_imagery);
            return ERROR;
        }

        for (index = 0; index < elements; index++)
        {
            /* Only convert if not fill */
            if (dn[index] > 0)
            {
                double rad = dn[index] * gain + offset;
                radiance[index] = rad;
            }
            else
                radiance[index] = 0.0;
        }
        free(orig_imagery);
        orig_imagery = NULL;
    }
    else
    {
        radiance = orig_imagery;
    }

    /* Populate the L1G info structure */
    if (ias_geo_get_units(fmd.projection_units, &units) != SUCCESS)
    {
        IAS_LOG_ERROR("Converting unit string '%s' to unit number",
            fmd.projection_units);
        free(radiance);
        return ERROR;
    }
    ias_geo_set_projection(fmd.projection_code, fmd.zone_code, units,
        fmd.spheroid_code, fmd.projection_parameters, &l1g_info->proj);
    l1g_info->lines = lines;
    l1g_info->samples = samples;
    l1g_info->x_pixel_size = bmd.projection_distance_x;
    l1g_info->y_pixel_size = bmd.projection_distance_y;
    l1g_info->radiance = radiance;

    /* Copy the corners */
    l1g_info->corners.upleft.x = bmd.upper_left_x;
    l1g_info->corners.upleft.y = bmd.upper_left_y;
    l1g_info->corners.upright.x = bmd.upper_right_x;
    l1g_info->corners.upright.y = bmd.upper_right_y;
    l1g_info->corners.loleft.x = bmd.lower_left_x;
    l1g_info->corners.loleft.y = bmd.lower_left_y;
    l1g_info->corners.loright.x = bmd.lower_right_x;
    l1g_info->corners.loright.y = bmd.lower_right_y;

    return SUCCESS;
}
