/* Standary Library Includes */
#include <stdlib.h>
#include <math.h>

/* COTS Library Includes */
#include "gctp.h"           /* DEGREE definition */

/* IAS Library Includes */
#include "ias_logging.h"

/* Local Includes */
#include "grid.h"

/* The target size of a grid cell in the target projection (in pixels) */
#define TARGET_PROJ_PIXELS 10

/******************************************************************************
Name: find_grid_dimensions

Purpose:  Find the number of rows and columns in the input grid

Note: Estimate the number of grid cells needed to make a grid
      TARGET_PROJ_PIXELS x TARGET_PROJ_PIXELS in the target projection

Returns: Type = Integer
         SUCCESS / ERROR
******************************************************************************/
static int find_grid_dimensions
(
    IAS_GEO_PROJ_TRANSFORMATION *proj_trans, /* I: Target projection */
    double in_ul_x,      /* I: Upper left x coordinate */
    double in_ul_y,      /* I: Upper left y coordinate */
    double in_lr_x,      /* I: Lower right x coordinate */
    double in_lr_y,      /* I: Lower right y coordinate */
    double x_pixel_size, /* I: Pixel size x dimension */
    double y_pixel_size, /* I: Pixel size y dimension */
    int *rows,           /* O: Calculated rows */
    int *columns         /* O: Calculated columns */
)
{
    double target_x_ul;
    double target_y_ul;
    double target_x_lr;
    double target_y_lr;
    double x_size;
    double y_size;

    if (ias_geo_transform_coordinate(proj_trans, in_ul_x, in_ul_y ,
        &target_x_ul, &target_y_ul) != SUCCESS)
    {
        IAS_LOG_ERROR("Transforming upper left coordinate");
        return ERROR;
    }

    if (ias_geo_transform_coordinate(proj_trans, in_lr_x, in_lr_y,
        &target_x_lr, &target_y_lr) != SUCCESS)
    {
        IAS_LOG_ERROR("Transforming lower right coordinate");
        return ERROR;
    }

    y_size = abs(target_y_ul - target_y_lr);
    x_size = abs(target_x_lr - target_x_ul);

    *rows = round(y_size / (y_pixel_size * TARGET_PROJ_PIXELS) + 1);
    *columns = round(x_size / (x_pixel_size * TARGET_PROJ_PIXELS) + 1);

    return SUCCESS;
}

/******************************************************************************
Name: create_grid

Purpose:  Create grid using the provided projection and image corners

Note: Estimate the number of grid cells needed to make a grid
      TARGET_PROJ_PIXELS x TARGET_PROJ_PIXELS in the target projection

Returns: Type = Integer
         SUCCESS / ERROR
******************************************************************************/
int create_grid
(
    IAS_GEO_PROJ_TRANSFORMATION *proj_trans, /* I: Target projection */
    double in_ul_x,           /* I: Upper left x coordinate */
    double in_ul_y,           /* I: Upper left y coordinate */
    double in_lr_x,           /* I: Lower right x coordinate */
    double in_lr_y,           /* I: Lower right y coordinate */
    double x_pixel_size,      /* I: X pixel size of the target image */
    double y_pixel_size,      /* I: Y pixel size of the target image */
    GRID *grid,               /* O: Grid to create */
    int *valid_grid_created_flag /* O: Indicates a valid grid was created
                                       (1 = valid, 0 = invalid).  This can be
                                       set to invalid if the input bounding box
                                       crosses 180 degrees */
)
{
    double cell_size;
    double x_cell_size;
    double y_cell_size;
    double x_dist = in_lr_x - in_ul_x;
    double y_dist = in_ul_y - in_lr_y;
    int rows;
    int columns;
    int row;
    int col;
    int line_count;
    int sample_count;
    double corner_longitudes[4];

    /* Initialize the grid pointers */
    grid->in_y = NULL;
    grid->in_x = NULL;
    grid->out_y = NULL;
    grid->out_x = NULL;
    grid->coeffs = NULL;
    *valid_grid_created_flag = 0;

    /* Test for crossing the 180 degree boundary, fall back to using the
       projection transformations if it does */
    corner_longitudes[0] = in_ul_x;
    corner_longitudes[1] = in_lr_x;
    corner_longitudes[2] = in_lr_x;
    corner_longitudes[3] = in_ul_x;
    if (ias_geo_does_cross_180(DEGREE, corner_longitudes))
    {
        IAS_LOG_WARNING("The input window crosses the 180 degree boundary - "
            "falling back to directly using projection transformations which "
            "will be slow");
        return SUCCESS;
    }

    if (x_dist < 0.0 || y_dist < 0.0)
    {
        IAS_LOG_ERROR("Input corners are invalid");
        return ERROR;
    }

    /* Estimate the number of grid cells needed */
    if (find_grid_dimensions(proj_trans, in_ul_x, in_ul_y, in_lr_x, in_lr_y,
            x_pixel_size, y_pixel_size, &rows, &columns) != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating the grid dimensions");
        return ERROR;
    }

    /* Settle on the smallest cell size in the two directions */
    x_cell_size = x_dist / columns;
    y_cell_size = y_dist / rows;
    if (x_cell_size < y_cell_size)
        cell_size = x_cell_size;
    else
        cell_size = y_cell_size;

    rows = y_dist / cell_size + 1;
    columns = x_dist / cell_size + 1;

    line_count = rows + 1;
    sample_count = columns + 1;

    /* Allocate the grid x/y fields */
    grid->in_y = malloc(line_count * sizeof(double));
    grid->in_x = malloc(sample_count * sizeof(double));
    grid->out_y = malloc(line_count * sample_count * sizeof(double));
    grid->out_x = malloc(line_count * sample_count * sizeof(double));
    grid->coeffs = malloc(rows * columns * sizeof(IAS_COEFFICIENTS));
    if (!grid->in_y || !grid->in_x || !grid->out_y || !grid->out_x
        || !grid->coeffs)
    {
        IAS_LOG_ERROR("Allocating grid memory");
        free_grid(grid);
        return ERROR;
    }

    /* Populate the output x/y locations */
    for (row = 0; row < line_count; row++)
    {
        grid->in_y[row] = in_ul_y - row * cell_size;

        for (col = 0; col < sample_count; col++)
        {
            int index = row * sample_count + col;

            /* Populate the input x values in the grid */
            if (row == 0)
            {
                grid->in_x[col] = in_ul_x + col * cell_size;
            }

            /* Convert the stray light lat/lng to projection coordinates */
            if (ias_geo_transform_coordinate(proj_trans, grid->in_x[col],
                grid->in_y[row], &grid->out_x[index], &grid->out_y[index])
                != SUCCESS)
            {
                IAS_LOG_ERROR("Transforming coordinates");
                free_grid(grid);
                return ERROR;
            }
        }
    }

    /* Calculate the coefficients */
    for (row = 0; row < rows; row++)
    {
        double in_y[4];

        in_y[0] = in_y[1] = grid->in_y[row];
        in_y[2] = in_y[3] = grid->in_y[row + 1];

        for (col = 0; col < columns; col++)
        {
            int index = row * sample_count + col;
            double in_x[4];
            double out_x[4];
            double out_y[4];
            int cell_index = row * columns + col;

            in_x[0] = in_x[2] = grid->in_x[col];
            in_x[1] = in_x[3] = grid->in_x[col + 1];

            out_x[0] = grid->out_x[index];
            out_x[1] = grid->out_x[index + 1];
            out_x[2] = grid->out_x[index + sample_count];
            out_x[3] = grid->out_x[index + sample_count + 1];

            out_y[0] = grid->out_y[index];
            out_y[1] = grid->out_y[index + 1];
            out_y[2] = grid->out_y[index + sample_count];
            out_y[3] = grid->out_y[index + sample_count + 1];

            if (ias_geo_compute_proj2proj_poly(in_x, in_y, out_x, out_y, 1,
                    grid->coeffs[cell_index].a, grid->coeffs[cell_index].b)
                    != SUCCESS)
            {
                IAS_LOG_ERROR("Calculating coefficients");
                free_grid(grid);
                return ERROR;
            }
        }
    }

    /* Populate the grid */
    grid->rows = rows;
    grid->columns = columns;
    grid->in_y_size = line_count;
    grid->in_x_size = sample_count;
    grid->ul_y = in_ul_y;
    grid->ul_x = in_ul_x;
    grid->cell_size = cell_size;

    *valid_grid_created_flag = 1;

    return SUCCESS;
}

/******************************************************************************
Name: free_grid

Purpose:  Frees the allocated grid memory

Returns: Void

******************************************************************************/
void free_grid
(
    GRID *grid /* I: Grid to free */
)
{
    if (!grid)
        return;

    /* Free the input x/y coordinates */
    free(grid->in_y);
    grid->in_y = NULL;
    free(grid->in_x);
    grid->in_x = NULL;

    /* Free the output x/y coordinates */
    free(grid->out_y);
    grid->out_y = NULL;
    free(grid->out_x);
    grid->out_x = NULL;

    /* Free the coefficients */
    free(grid->coeffs);
    grid->coeffs = NULL;
}
