/* Standary Library Includes */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>

/* COTS Library Includes */
#include "gctp.h"

/* IAS Library Includes */
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_geo.h"

/* Local Includes */
#include "../../config.h"   /* HAVE_LITTLE_ENDIAN definition */
#include "get_l1g.h"
#include "read_envi_image.h"

/*****************************************************************************
Name: read_envi_header

Purpose: Read the lines, samples, corners, and projection information from
    an ENVI header.

Note: This is very stripped down and not very robust.  Some day it should be
    made more robust and turned into a library routine.

Returns: SUCCESS / ERROR
*****************************************************************************/
static int read_envi_header
(
    const char *filename,   /* I: Name of the L1G image to read */
    int spheroid,           /* I: Geographic spheroid */
    int *band_count,        /* O: Number of bands in the file */
    L1G_INFO *l1g_info      /* O: Structure to return info in */
)
{
    char header_name[PATH_MAX];
    int status;
    FILE *file;
    char line[1000];
    int got_lines = 0;
    int got_samples = 0;
    int got_proj = 0;
    int data_type = 0;
    int is_bsq = 1;         /* Assume it is BSQ unless found otherwise */
    int bands = 0;

    /* Build the header filename */
    status = snprintf(header_name, sizeof(header_name), "%s.hdr", filename);
    if (status < 0 || status >= sizeof(header_name))
    {
        IAS_LOG_ERROR("Header filename buffer is not large enough");
        return ERROR;
    }

    /* Open the header file */
    file = fopen(header_name, "r");
    if (!file)
    {
        IAS_LOG_ERROR("Failed opening envi header file %s", header_name);
        return ERROR;
    }

    /* Loop through the file */
    while (fgets(line, sizeof(line), file) != NULL)
    {
        char *temp_ptr;

        if (strncmp(line, "lines", 5) == 0)
        {
            temp_ptr = strchr(line, '=');
            if (!temp_ptr || sscanf(&temp_ptr[1], "%d", &l1g_info->lines) != 1)
            {
                IAS_LOG_ERROR("Failed parsing \"lines\" field from: %s", line);
                fclose(file);
                return ERROR;
            }
            got_lines = 1;
        }
        else if (strncmp(line, "samples", 7) == 0)
        {
            temp_ptr = strchr(line, '=');
            if (!temp_ptr || sscanf(&temp_ptr[1], "%d", &l1g_info->samples)
                != 1)
            {
                IAS_LOG_ERROR("Failed parsing \"samples\" field from: %s",
                    line);
                fclose(file);
                return ERROR;
            }
            got_samples = 1;
        }
        else if (strncmp(line, "bands", 5) == 0)
        {
            /* This is optional, if not found, will assume there is one band */
            temp_ptr = strchr(line, '=');
            if (!temp_ptr || sscanf(&temp_ptr[1], "%d", &bands) != 1)
            {
                IAS_LOG_ERROR("Failed parsing \"bands\" field from: %s", line);
                fclose(file);
                return ERROR;
            }
            if (bands < 1)
            {
                IAS_LOG_ERROR("Invalid number of bands: %d", bands);
                fclose(file);
                return ERROR;
            }
        }
        else if (strncmp(line, "interleave", 10) == 0)
        {
            if (!strstr(line, "bsq"))
                is_bsq = 0;

        }
        else if (strncmp(line, "data type", 9) == 0)
        {
            temp_ptr = strchr(line, '=');
            if (!temp_ptr || sscanf(&temp_ptr[1], "%d", &data_type) != 1)
            {
                IAS_LOG_ERROR("Failed parsing \"data type\" field from: %s",
                    line);
                fclose(file);
                return ERROR;
            }
        }
        else if (strncmp(line, "byte order", 10) == 0)
        {
            int byte_order;
            int native_byte_order;

            /* This is optional, if not found, will assume there is one band */
            temp_ptr = strchr(line, '=');
            if (!temp_ptr || sscanf(&temp_ptr[1], "%d", &byte_order) != 1)
            {
                IAS_LOG_ERROR("Failed parsing \"byte order\" field from: %s",
                    line);
                fclose(file);
                return ERROR;
            }

            /* HAVE_LITTLE_ENDIAN is defined in config.h */
            if (HAVE_LITTLE_ENDIAN)
                native_byte_order = 0;
            else
                native_byte_order = 1;

            if (byte_order != native_byte_order)
            {
                IAS_LOG_ERROR("Only native byte ordering is supported");
                fclose(file);
                return ERROR;
            }
        }
        else if (strncmp(line, "map info", 8) == 0)
        {
            double proj_parms[IAS_PROJ_PARAM_SIZE];
            int index;
            double x_offset;
            double y_offset;
            double ulx;
            double uly;

            for (index = 0; index < IAS_PROJ_PARAM_SIZE; index++)
                proj_parms[index] = 0.0;

            temp_ptr = strchr(line, '=');
            if (temp_ptr)
            {
                temp_ptr = strchr(temp_ptr, '{');
            }
            if (!temp_ptr)
            {
                IAS_LOG_ERROR("Failed parsing \"map info\" field from: %s",
                    line);
                fclose(file);
                return ERROR;
            }
            temp_ptr++;

            if (strncmp(temp_ptr, "Geographic Lat/Lon", 18) == 0)
            {
                /* Handle a geographic projection (ignores datum and units
                   fields at the end of the line) */
                if (sscanf(temp_ptr,
                    "Geographic Lat/Lon, %lf, %lf, %lf, %lf, %lf, %lf",
                    &x_offset, &y_offset, &ulx, &uly, &l1g_info->x_pixel_size,
                    &l1g_info->y_pixel_size) != 6)
                {
                    IAS_LOG_ERROR("Failed parsing Geographic projection info "
                        "from: %s", line);
                    fclose(file);
                    return ERROR;
                }

                ias_geo_set_projection(GEO, -1, DEGREE, spheroid, proj_parms,
                    &l1g_info->proj);
            }
            else if (strncmp(temp_ptr, "UTM", 3) == 0)
            {
                int zone;
                char hemisphere[20];

                /* Handle a UTM projection (ignores datum and units fields at
                   the end of the line) */
                if (sscanf(temp_ptr,
                    "UTM, %lf, %lf, %lf, %lf, %lf, %lf, %d, %19s",
                    &x_offset, &y_offset, &ulx, &uly, &l1g_info->x_pixel_size,
                    &l1g_info->y_pixel_size, &zone, hemisphere) != 8)
                {
                    IAS_LOG_ERROR("Failed parsing UTM projection info "
                        "from: %s", line);
                    fclose(file);
                    return ERROR;
                }
                if (strncmp(hemisphere, "North", 5) != 0)
                {
                    /* Not the northern hemisphere, so adjust for the southern
                       hemisphere if needed */
                    if (uly > 0)
                        uly -= 10000000.0;
                }

                ias_geo_set_projection(UTM, zone, METER, WGS84_SPHEROID,
                    proj_parms, &l1g_info->proj);
            }
            else
            {
                IAS_LOG_ERROR("Unsupported projection specified in \"map "
                    "info\": %s", line);
                fclose(file);
                return ERROR;
            }

            /* Verify the offset is as expected */
            if (x_offset != 1.0 || y_offset != 1.0)
            {
                IAS_LOG_ERROR("The projection offset values are not 1");
                fclose(file);
                return ERROR;
            }

            /* Adjust the upper left corner by a half pixel since envi headers
               specify the upper left corner of the pixel instead of the
               center */
            l1g_info->corners.upleft.y = uly - l1g_info->y_pixel_size / 2.0;
            l1g_info->corners.upleft.x = ulx + l1g_info->x_pixel_size / 2.0;

            got_proj = 1;
        }
    }

    fclose(file);

    /* Verify the required fields were read */
    if (!got_lines || !got_samples || !got_proj || !data_type)
    {
        IAS_LOG_ERROR("Incomplete envi header file in %s", header_name);
        return ERROR;
    }

    /* If more than one band was found, make sure it is BSQ since that is all
       that is currently supported */
    if (bands > 1 && !is_bsq)
    {
        IAS_LOG_ERROR("Multiple bands only supported for BSQ interleave");
        return ERROR;
    }

    /* Verify the data type is 4 (floating point) */
    if (data_type != 4)
    {
        IAS_LOG_ERROR("Input file is not floating point");
        return ERROR;
    }

    /* Populate the other 3 corners now that we know the lines and samples
       are available */
    l1g_info->corners.upright.y = l1g_info->corners.upleft.y;
    l1g_info->corners.upright.x = l1g_info->corners.upleft.x
        + (l1g_info->samples - 1) * l1g_info->x_pixel_size;
    l1g_info->corners.loleft.y = l1g_info->corners.upleft.y
        - (l1g_info->lines - 1) * l1g_info->y_pixel_size;
    l1g_info->corners.loleft.x = l1g_info->corners.upleft.x;
    l1g_info->corners.loright.y = l1g_info->corners.loleft.y;
    l1g_info->corners.loright.x = l1g_info->corners.upright.x;

    *band_count = bands;

    return SUCCESS;
}

/*****************************************************************************
Name: read_envi_image

Purpose: Reads a flat image file with a matching ENVI header file.

Returns: SUCCESS / ERROR
*****************************************************************************/
int read_envi_image
(
    const char *filename,   /* I: Name of the L1G image to read */
    int band_number,        /* I: Band number to read */
    int spheroid,           /* I: Geographic spheroid */
    L1G_INFO *l1g_info      /* O: Structure to return info in */
)
{
    FILE *file;
    int pixels;
    int band_count;
    int bytes;

    /* Read the envi header file */
    if (read_envi_header(filename, spheroid, &band_count, l1g_info) != SUCCESS)
    {
        IAS_LOG_ERROR("Failed reading envi header for %s", filename);
        return ERROR;
    }

    /* Make sure the requested band is available */
    if (band_number < 1 || band_number > band_count)
    {
        IAS_LOG_ERROR("Band number %d is not available in %s", band_number,
            filename);
        return ERROR;
    }

    /* Calculate the number of pixels to read */
    pixels = l1g_info->lines * l1g_info->samples;
    bytes = pixels * sizeof(float);

    /* Read the imagery */
    file = fopen(filename, "rb");
    if (!file)
    {
        IAS_LOG_ERROR("Failed opening envi image file %s", filename);
        return ERROR;
    }

    /* Seek to the correct band in the file */
    if (fseek(file, (band_number - 1) * bytes, SEEK_SET) != 0)
    {
        IAS_LOG_ERROR("Failed seeking band number %d in %s", band_number,
            filename);
        fclose(file);
        return ERROR;
    }

    /* Allocate memory for the image */
    l1g_info->radiance = malloc(pixels * sizeof(*l1g_info->radiance));
    if (!l1g_info->radiance)
    {
        IAS_LOG_ERROR("Failed allocating memory for %d radiance samples",
            pixels);
        fclose(file);
        return ERROR;
    }

    /* Read the imagery */
    if (fread(l1g_info->radiance, sizeof(*l1g_info->radiance), pixels, file)
        != pixels)
    {
        IAS_LOG_ERROR("Failed reading imagery from %s", filename);
        free(l1g_info->radiance);
        l1g_info->radiance = NULL;
        fclose(file);
        return ERROR;
    }

    fclose(file);

    return SUCCESS;
}
