/*****************************************************************************
Name: read_l0r_metadata

Purpose: Reads the L0R metadata needed to size the extents of the image
    being processed.

Returns: SUCCESS/ERROR

*****************************************************************************/
/* Standary Library Includes */
#include <stdlib.h>
#include <string.h>

/* IAS Library Includes */
#include "ias_logging.h"
#include "ias_l0r.h"

/* Local Includes */
#include "read_l0r_metadata.h"

int read_l0r_metadata
(
    const char *l0r_filename,   /* I: L0R filename */
    const char *scene_id,       /* I: Scene ID of the target scene */
    L0R_METADATA *l0r_metadata  /* O: Metadata read from the file */
)
{
    L0RIO *l0r_ptr; /* L0R file access pointer */
    int record_count;
    IAS_L0R_SCENE_METADATA *scene_metadata;
    int index;
    int start_frame;

    /* Open the L0R file */
    l0r_ptr = ias_l0r_open_from_filename(l0r_filename,
        IAS_L0R_COMPRESSION_DEFAULT);
    if (!l0r_ptr)
    {
        IAS_LOG_ERROR("Opening L0R file %s", l0r_filename);
        return ERROR;
    }

    IAS_LOG_INFO("Reading L0R metadata from %s", l0r_filename);

    /* Open read access to the L0R metadata */
    if (ias_l0r_open_metadata(l0r_ptr, IAS_READ) != SUCCESS)
    {
        IAS_LOG_ERROR("Opening read access to L0R metadata");
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Retrieve the L0R scene record count */
    if (ias_l0r_get_scene_metadata_records_count(l0r_ptr, &record_count)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Reading L0R scene count");
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* If no records available, that is a problem */
    if (record_count < 1)
    {
        IAS_LOG_ERROR("No scene metadata available");
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }
    if (record_count > 3)
    {
        IAS_LOG_WARNING("Found %d scenes when a maximum of 3 expected",
            record_count);
    }

    /* Allocate memory for the scene metadata */
    scene_metadata = malloc(record_count * sizeof(*scene_metadata));
    if (!scene_metadata)
    {
        IAS_LOG_ERROR("Failed allocating memory for %d records of scene "
            "metadata", record_count);
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Read the scene metadata */
    if (ias_l0r_get_scene_metadata(l0r_ptr, 0, record_count, scene_metadata)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Reading L0R scene metadata");
        free(scene_metadata);
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    /* Close the L0R file */
    if (ias_l0r_close(l0r_ptr) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing L0R");
        free(scene_metadata);
        return ERROR;
    }

    /* Find the target scene */
    for (index = 0; index < record_count; index++)
    {
        if (strcmp(scene_id, scene_metadata[index].landsat_scene_id) == 0)
            break;
    }
    if (index >= record_count)
    {
        IAS_LOG_ERROR("Scene ID %s not found in the L0R", scene_id);
        free(scene_metadata);
        return ERROR;
    }

    /* Extract the needed metadata from the scene metadata */
    l0r_metadata->subinterval_start_frame
        = scene_metadata[0].scene_start_frame_tirs;
    l0r_metadata->subinterval_end_frame
        = scene_metadata[record_count - 1].scene_stop_frame_tirs;
    l0r_metadata->subinterval_frames = l0r_metadata->subinterval_end_frame
        - l0r_metadata->subinterval_start_frame + 1;

    l0r_metadata->target_scene_start_frame
        = scene_metadata[index].scene_start_frame_tirs;
    l0r_metadata->target_scene_end_frame
        = scene_metadata[index].scene_stop_frame_tirs;
    l0r_metadata->target_scene_frames = l0r_metadata->target_scene_end_frame
        - l0r_metadata->target_scene_start_frame + 1;

    /* Make all the line numbers relative to the first one */
    start_frame = l0r_metadata->subinterval_start_frame;
    l0r_metadata->subinterval_start_frame -= start_frame;
    l0r_metadata->subinterval_end_frame -= start_frame;
    l0r_metadata->target_scene_start_frame -= start_frame;
    l0r_metadata->target_scene_end_frame -= start_frame;

    /* Free the scene metadata */
    free(scene_metadata);

    return SUCCESS;
}
