/* Standard C Includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* IAS Includes */
#include "ias_const.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "ias_satellite_attributes.h"
#include "ias_miscellaneous.h"

/* Local Includes */
#include "create_tirs_ghost_image.h"
#include "local_defines.h"

/* Local Defines */
#define DEFAULT_BLUR_KERNEL_SIZE 60
#define DEFAULT_GOES_BLUR_KERNEL_SIZE 3

/* Parameter counts struct */
typedef struct parameter_counts
{
    int water_mask;
    int report_sampling_location;
    int radiance_coefficient_filename;
    int stray_light_vector_filename;
    int blur_kernel_size;
} PARAMETER_COUNTS;

/******************************************************************************
Name: read_omf_parameters

Purpose: Reads the parameters from the OMF file that create tirs ghost image
         needs.

Returns: Type = int
    Value    Description
    -----    -----------
    SUCCESS  The parameters were successfully read
    ERROR    An error occurred reading the parameters
******************************************************************************/
int read_omf_parameters
(
    const char *omf_filename,          /* I: OMF filename */
    int local_processing_pass,         /* I: Processing pass */
    GHOST_IMAGE_PARAMETERS *parameters,/* O: Generation parameters */
    char *cpf_filename                 /* O: CPF filename */
)
{
    int count;                /* Table parameter count */
    int help_requested;       /* Status placeholder */

    /* Start parameter table definition */
    IAS_PARM_DECLARE_TABLE(omf_table, 6);

    IAS_PARM_L1G_FILENAME_FROM_PASS(omf_table, IAS_PARM_OPTIONAL,
        parameters->l1g_filename, sizeof(parameters->l1g_filename), 1,
        local_processing_pass);

    IAS_PARM_MODEL_FILENAME_FROM_PASS(omf_table, IAS_PARM_REQUIRED,
        parameters->model_filename, sizeof(parameters->model_filename), 1,
        local_processing_pass);

    IAS_PARM_L0R_FILENAME(omf_table, IAS_PARM_REQUIRED,
        parameters->l0r_filename, sizeof(parameters->l0r_filename), 1);

    strcpy(parameters->landsat_scene_id, "");
    IAS_PARM_LANDSAT_SCENE_ID(omf_table, IAS_PARM_REQUIRED,
        parameters->landsat_scene_id, sizeof(parameters->landsat_scene_id), 1);

    IAS_PARM_MAX_PROCESSORS_TO_USE(omf_table, IAS_PARM_OPTIONAL,
        &parameters->max_processors, sizeof(parameters->max_processors), 1);

    IAS_PARM_CPF(omf_table, IAS_PARM_REQUIRED, cpf_filename, PATH_MAX, 1);

    /* Retrieve parameter count for OMF table */
    count = IAS_PARM_GET_TABLE_SIZE(omf_table);

    /* Call the ias_parm_provide_help routine to provide help for the
       parameters if requested; also verifies the OMF file exists */
    help_requested = ias_parm_provide_help(omf_filename, omf_table, count,
        IAS_OMF_PARAMETERS);
    if (help_requested)
    {
        /* Either help was provided or an error occurred*/
        if (help_requested == TRUE)
        {
            /* Successfully provided help, as requested, so just exit */
            exit(EXIT_SUCCESS);
        }

        IAS_LOG_ERROR("Error occurred providing OMF help");
        return ERROR;
    }

    /* Read the OMF file */
    if (ias_parm_read(omf_filename, omf_table, count) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading OMF parameter file %s", omf_filename);
        return ERROR;
    }

    return SUCCESS;
}

/******************************************************************************
Name: process_parameters

Purpose: This function performs special processing on parameters which need
         to be translated after being read from the parameter and/or OMF files.

Returns: Type = Integer
    Value     Description
    -----     -----------
    SUCCESS   The parameters were successfully processed
    ERROR     An error occurred processing the parameters
******************************************************************************/
int process_parameters
(
    const PARAMETER_COUNTS *parameter_counts, /* I: Parameter counts */
    const char *cpf_filename,                 /* I: CPF filename */
    const int *sampling_location,             /* I: Sampling location */
    GHOST_IMAGE_PARAMETERS *parameters        /* O: Processing parameters */
)
{
    IAS_CPF *cpf = NULL;
    const IAS_CPF_EARTH_CONSTANTS *cpf_earth = NULL;
    int status; /* Status placeholder variable */
    int have_source_image = (strlen(parameters->source_image_filename) > 0);

    /* Force the sensor name to be uppercase */
    ias_misc_convert_to_uppercase(parameters->source_image_sensor);

    /* Set the blur kernel size if not set by the user */
    if (parameter_counts->blur_kernel_size == 0)
    {
        parameters->blur_kernel_size = DEFAULT_BLUR_KERNEL_SIZE;
    }

    /* Check if the GOES sensor is being requested */
    if (strcmp(parameters->source_image_sensor, "GOES") == 0)
    {
        if (parameter_counts->blur_kernel_size == 0)
        {
            parameters->blur_kernel_size = DEFAULT_GOES_BLUR_KERNEL_SIZE;
            IAS_LOG_INFO("Defaulting kernel size to the GOES blur kernel "
                "default (%d)", DEFAULT_GOES_BLUR_KERNEL_SIZE);
        }

        if (!have_source_image)
        {
            IAS_LOG_ERROR("The source image filename must be specified "
                "when creating a GOES ghost image");
            return ERROR;
        }
    }
    /* Verify that a source image filename is available (either from the
       parameter file or from the OMF L1G name) */
    else if (!have_source_image)
    {
        if (strlen(parameters->l1g_filename) == 0)
        {
            IAS_LOG_ERROR("Either an L1G_FILENAME is needed in the OMF or the"
                " SOURCE_IMAGE_FILENAME needs to be provided");
            return ERROR;
        }

        /* Copy the L1G filename to the source image filename so there is one
           place to get the definitive filename */
        strncpy(parameters->source_image_filename, parameters->l1g_filename,
            sizeof(parameters->source_image_filename));
        parameters->source_image_filename[
            sizeof(parameters->source_image_filename) - 1] = '\0';
    }
    IAS_LOG_INFO("Using source radiance image %s",
        parameters->source_image_filename);

    /* If the vector filename was not provided, use the default */
    if (parameter_counts->stray_light_vector_filename == 0)
    {
        const char *data_dir = getenv("IAS_DATA_DIR");
        if (!data_dir)
        {
            IAS_LOG_ERROR("IAS_DATA_DIR environment variable not set");
            return ERROR;
        }

        status = snprintf(parameters->stray_light_vector_filename,
            sizeof(parameters->stray_light_vector_filename), "%s/%s",
            data_dir,
            "sysdata/stray_light_vectors/tirs_stray_light_vectors.csv");
        if (status < 0 || status >= sizeof(
            parameters->stray_light_vector_filename))
        {
            IAS_LOG_ERROR("Creating stray light vector filename");
            return ERROR;
        }
    }
    IAS_LOG_INFO("Using TIRS stray light vectors from %s",
        parameters->stray_light_vector_filename);

    /* If the radiance coefficient filename was not provided, use the default */
    if (parameter_counts->radiance_coefficient_filename == 0)
    {
        const char *data_dir = getenv("IAS_DATA_DIR");
        if (!data_dir)
        {
            IAS_LOG_ERROR("IAS_DATA_DIR environment variable not set");
            return ERROR;
        }

        status = snprintf(parameters->radiance_coefficient_filename,
            sizeof(parameters->radiance_coefficient_filename), "%s/%s/%s_%s",
            data_dir, "sysdata/stray_light_vectors",
            parameters->source_image_sensor, "radiance_coefficients.csv");
        if (status < 0 || status >= sizeof(
            parameters->radiance_coefficient_filename))
        {
            IAS_LOG_ERROR("Creating radiance coefficient filename");
            return ERROR;
        }
    }
    IAS_LOG_INFO("Using %s radiance coefficients from %s",
        parameters->source_image_sensor,
        parameters->radiance_coefficient_filename);

    /* If the sampling location was provided, utilize it */
    parameters->sample_location_requested_flag = 0;
    if (parameter_counts->report_sampling_location == 4)
    {
        int band_index = ias_sat_attr_convert_band_number_to_index(
            sampling_location[0]);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Invalid band number (%d) provided in the "
                "REPORT_SAMPLING_LOCATION parameter", sampling_location[0]);
            return ERROR;
        }

        parameters->sampling_location.band_index = band_index;
        parameters->sampling_location.sca_index = sampling_location[1] - 1;
        parameters->sampling_location.detector_index = sampling_location[2] - 1;
        parameters->sampling_location.frame_index = sampling_location[3] - 1;
        parameters->sample_location_requested_flag = 1;

        IAS_LOG_INFO("Will report sampling location for band number %d, SCA "
            "number %d, detector number %d, frame number %d",
            sampling_location[0], sampling_location[1], sampling_location[2],
            sampling_location[3]);
    }

    /* If the water mask should be used and the water mask filename was
       not provided, use the default */
    if (parameters->use_land_water_radiance_average_flag == 1 &&
        parameter_counts->water_mask == 0)
    {
        const char *data_dir = getenv("IAS_DATA_DIR");
        if (!data_dir)
        {
            IAS_LOG_ERROR("IAS_DATA_DIR environment variable not set");
            return ERROR;
        }

        status = snprintf(parameters->water_mask_filename,
            sizeof(parameters->water_mask_filename), "%s%s",
            data_dir, "/sysdata/water_mask/land_nobuf.ply");
        if (status < 0 || status >= sizeof(parameters->water_mask_filename))
        {
            IAS_LOG_ERROR("Creating the water mask filename");
            return ERROR;
        }
    }

    /* Create output directory for envi files if requested */
    if (parameters->output_files_to_envi_flag)
    {
        if (initialize_directory(TIRS_STRAY_LIGHT_DIRECTORY) != SUCCESS)
        {
            IAS_LOG_ERROR("Creating %s directory", TIRS_STRAY_LIGHT_DIRECTORY);
            return ERROR;
        }
    }

    /* Read the cpf */
    cpf = ias_cpf_read(cpf_filename);
    if (!cpf)
    {
        IAS_LOG_ERROR("Opening the CPF %s", cpf_filename);
        return ERROR;
    }

    /* Retrieve the cpf earth constants */
    cpf_earth = ias_cpf_get_earth_const(cpf);
    if (!cpf_earth)
    {
        IAS_LOG_ERROR("Retrieving Earth constants group from the CPF");
        ias_cpf_free(cpf);
        return ERROR;
    }

    parameters->semi_major_axis = cpf_earth->semi_major_axis;
    parameters->semi_minor_axis = cpf_earth->semi_minor_axis;
    parameters->spheroid = cpf_earth->spheroid_code;

    /* Free the cpf */
    ias_cpf_free(cpf);
    cpf = NULL;

    return SUCCESS;
}

/*****************************************************************************
Name: verify_local_defines

Purpose: Verify the local defines with the satelite attributes

Returns: Type = Integer
    Value    Description
    -------  --------------------------
    SUCCESS  Successful completion
    ERROR    Operation failed

*****************************************************************************/
int verify_local_defines
(
    GHOST_IMAGE_PARAMETERS *parameters      /* I: Processing parameters */
)
{
    int band_index = 0;
    int bands_being_processed = 0;

    /* Loop over the bands */
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        const IAS_BAND_ATTRIBUTES *band_attr;
        int band_number;

        /* Retrieve the band number using the band index */
        band_number = ias_sat_attr_convert_band_index_to_number(
            band_index);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Converting band index %d to band number",
                band_index);
            return ERROR;
        }

        /* Read in the band attributes */
        band_attr = ias_sat_attr_get_band_attributes(band_number);
        if (!band_attr)
        {
            IAS_LOG_ERROR("Retrieving band attributes for band number %d",
                band_number);
            return ERROR;
        }

        /* We are only concerned with the TIRS bands */
        /* Check if this band should be processed */
        if (!parameters->process_band[band_index]
            || band_attr->sensor_id != IAS_TIRS)
        {
            continue;
        }
        bands_being_processed++;

        /* Check the number of scas and detectors match the local defines */
        if (band_attr->scas != NUMBER_OF_TIRS_SCAS)
        {
            IAS_LOG_ERROR("The number of SCAs (%d) for band number %d does not "
                "match the expected number of %d", band_attr->scas, band_number,
                NUMBER_OF_TIRS_SCAS);
            return ERROR;
        }
        else if (band_attr->detectors_per_sca != NUMBER_OF_TIRS_DETECTORS)
        {
            IAS_LOG_ERROR("The number of detectors (%d) for band number %d "
                "does not match the expected number of %d",
                band_attr->detectors_per_sca, band_number,
                NUMBER_OF_TIRS_DETECTORS);
            return ERROR;
        }
    }

    /* Check that the number of bands being processed is not greater than the
       local define */
    if (bands_being_processed > NUMBER_OF_TIRS_BANDS)
    {
        IAS_LOG_ERROR("The number of bands (%d) found is greater than the "
            "expected number of bands (%d)", bands_being_processed,
            NUMBER_OF_TIRS_BANDS);
        return ERROR;
    }

    return SUCCESS;
}

/*****************************************************************************
Name: read_parameters

Purpose: Read parameters from the specified input parameter file and OMF file.

Returns: Type = Integer
    Value    Description
    -------  --------------------------
    SUCCESS  Successful completion
    ERROR    Operation failed

*****************************************************************************/
int read_parameters
(
    const char *parameter_filename,         /* I: Parameter file name */
    GHOST_IMAGE_PARAMETERS *parameters      /* O: Processing parameters */
)
{
    char cpf_filename[PATH_MAX];
    PARAMETER_COUNTS counts;                /* Parameter counts struct */
    int help_requested;                     /* Help request variable */
    int count;                              /* Table parameter count */
    int status;                             /* Status variable */
    char local_wo_id[IAS_WORKORDER_ID_SIZE];/* Work order ID */
    int list_of_bands[IAS_MAX_NBANDS];
    int band_index = 0;
    int number_of_bands = 0; /* Number of bands in band list */
    int sampling_location[4]; /* Band number, SCA number, detector number,
                                 line number to report sampling points */
    int local_processing_pass = 0;

    /* Zero the parameters */
    memset(parameters, 0, sizeof(GHOST_IMAGE_PARAMETERS));

    /* Initialize the satellite attributes library */
    if (ias_sat_attr_initialize(IAS_L8) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the satellite attributes library");
        return ERROR;
    }

    /* Parameters table definition */
    IAS_PARM_DECLARE_TABLE(parameter_table, 15);

    /* Common parameters table definition for the parameter file */
    IAS_PARM_WORK_ORDER_ID(parameter_table, local_wo_id,
        sizeof(local_wo_id), 1);

    int default_band_list[] = {10, 11};
    IAS_PARM_ADD_INT(parameter_table, BAND_LIST, "Defines the bands to process",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY, 1, 10, 11, 2, default_band_list,
        list_of_bands, sizeof(list_of_bands), 1);

    IAS_PARM_PROCESSING_PASS(parameter_table, &local_processing_pass,
        sizeof(local_processing_pass), 0);

    IAS_PARM_STRAY_LIGHT_CORRECTION_FILENAME(parameter_table,
        IAS_PARM_REQUIRED, parameters->stray_light_correction_filename,
        sizeof(parameters->stray_light_correction_filename), 1);

    IAS_PARM_ADD_STRING(parameter_table, STRAY_LIGHT_VECTOR_FILENAME,
        "Vector file defining the line of sight (LOS) vectors to override "
        "the default system installed file", IAS_PARM_OPTIONAL, 0, NULL, 0,
        NULL, parameters->stray_light_vector_filename,
        sizeof(parameters->stray_light_vector_filename), 1);

    IAS_PARM_ADD_STRING(parameter_table, RADIANCE_COEFFICIENT_FILENAME,
        "File defining the 'a' and 'b' coefficients to override the default "
        "system installed file", IAS_PARM_OPTIONAL, 0, NULL, 0, NULL,
        parameters->radiance_coefficient_filename,
        sizeof(parameters->radiance_coefficient_filename), 1);

    int default_apply_radiance_coefficients[] = {1};
    IAS_PARM_ADD_INT(parameter_table, APPLY_RADIANCE_COEFFICIENTS,
        "Flag to indicate the 'a' and 'b' coefficients should be applied to "
        "convert the ghost image to radiance units",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has a limited range of 0-1*/
        1, 0, 1,
        /* has one default value */
        1, default_apply_radiance_coefficients,
        &parameters->apply_radiance_coefficients,
        sizeof(parameters->apply_radiance_coefficients), 0);

    IAS_PARM_ADD_INT(parameter_table, BLUR_KERNEL_SIZE,
        "Size of the kernel to use when blurring the input L1G image. The "
        "blur function calculates the average pixel value in the window "
        "from -size/2 to size/2. DEFAULTS are (TIRS = 60, GOES = 3)",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has a limited range of 0-100*/
        1, 1, 100,
        /* has one default value */
        0, NULL, &parameters->blur_kernel_size,
        sizeof(parameters->blur_kernel_size), 0);

    IAS_PARM_ADD_INT(parameter_table, REPORT_SAMPLING_LOCATION,
        "This debugging parameter requests a file to be output with the "
        "line/sample locations in the L1G image that were sampled to create "
        "the (band number, SCA number, detector, frame) in the ghost image",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
        /* has no range limit */
        0, 0, 0,
        /* no default values */
        0, NULL, sampling_location, sizeof(sampling_location), 4);

    /* Get the name of the source image.  This is optional since when TIRS is
       the source, the name is available in the OMF file. */
    IAS_PARM_ADD_STRING(parameter_table, SOURCE_IMAGE_FILENAME,
        "Image filename to use as the radiance source for the sampling. "
        "Overrides the L1G_FILENAME from the OMF for the non-TIRS case.",
        IAS_PARM_OPTIONAL, 0, NULL, 0, NULL,
        parameters->source_image_filename,
        sizeof(parameters->source_image_filename), 1);

    /* Allow the source sensor name to be provided to allow running with
       various source temperature data */
    const char *default_sensor[] = {"TIRS"};
    IAS_PARM_ADD_STRING(parameter_table, SOURCE_IMAGE_SENSOR,
        "Name of the sensor that provides the source image",
        IAS_PARM_OPTIONAL, 0, NULL, 1, default_sensor,
        parameters->source_image_sensor,
        sizeof(parameters->source_image_sensor), 1);

    int default_land_water_radiance_flag[] = {0};
    IAS_PARM_ADD_INT(parameter_table, USE_LAND_WATER_RADIANCE_AVERAGE_FLAG,
        "Flag to indicate that the land and water radiance averages will be "
        "used instead of the nearest pixel radiance {0 = use nearest pixel "
        "radiance; 1 = use land and water average radiance}", IAS_PARM_OPTIONAL,
        IAS_PARM_NOT_ARRAY, 1, 0, 1, 1, default_land_water_radiance_flag,
        &parameters->use_land_water_radiance_average_flag,
        sizeof(parameters->use_land_water_radiance_average_flag), 1);

    /* Get the water mask filename.  Note that the name is a bit misleading
       since the polygons in the file are polygons defining the land masses.
       It would be clearer to name it the LAND_MASS_FILENAME, but the precedent
       has been set already by L1-7. */
    IAS_PARM_ADD_STRING(parameter_table, WATER_MASK_FILENAME,
        "Polygon file defining the land masses in the world to override the "
        "default file used to create the water mask", IAS_PARM_OPTIONAL,
        0, NULL, 0, NULL, parameters->water_mask_filename,
        sizeof(parameters->water_mask_filename), 0);

    int default_output_envi_files[] = {0};
    IAS_PARM_ADD_INT(parameter_table, OUTPUT_FILES_TO_ENVI,
        "Flag to indicate that the optional envi files will be output to "
        "the tirs_stray_light directory {0 - Do not output envi files, "
        "1 - Output envi files}", IAS_PARM_OPTIONAL,
        IAS_PARM_NOT_ARRAY, 1, 0, 1, 1, default_output_envi_files,
        &parameters->output_files_to_envi_flag,
        sizeof(parameters->output_files_to_envi_flag), 1);

    int default_number_bounding_box_frame_sections[] = {7};
    IAS_PARM_ADD_INT(parameter_table, NUMBER_BOUNDING_BOX_FRAME_SECTIONS,
        "The number of frame sections that will be sampled to determine "
        "the bounding box", IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 0, 1000,
        1, default_number_bounding_box_frame_sections,
        &parameters->number_bounding_box_frame_sections,
        sizeof(parameters->number_bounding_box_frame_sections), 1);

    /* Calculate the number of items in the parameter file */
    count = IAS_PARM_GET_TABLE_SIZE(parameter_table);

    /* Call the ias_parm_provide_help routine to provide help for the
       parameters if requested; also verifies the ODL file exists */
    help_requested = ias_parm_provide_help(parameter_filename, parameter_table,
        count, IAS_INPUT_PARAMETERS);
    if (help_requested)
    {
        /* Either help was provided or an error occurred */
        if (help_requested == TRUE)
        {
            /* Provide help for the OMF.  If it succeeds, it will exit.  If
               it returns, an error occurred, so return an error */
            read_omf_parameters(parameter_filename, local_processing_pass,
                parameters, cpf_filename);
            IAS_LOG_ERROR("Providing application parameter file help");
            return ERROR;
        }

        IAS_LOG_ERROR("Error occurred providing ODL help");
        return ERROR;
    }

    /* Read the parameter file */
    if (ias_parm_read(parameter_filename, parameter_table, count) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from: %s", parameter_filename);
        return ERROR;
    }

    number_of_bands = IAS_PARM_GET_COUNT_READ(parameter_table,
        BAND_LIST);
    if (number_of_bands == 0)
    {
        IAS_LOG_ERROR("No bands found in list");
        return ERROR;
    }

    /* Initialize the band list */
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
        parameters->process_band[band_index] = FALSE;

    for (band_index = 0; band_index < number_of_bands; band_index++)
    {
        int sat_band_index = ias_sat_attr_convert_band_number_to_index(
                list_of_bands[band_index]);
        if (sat_band_index == ERROR)
        {
            IAS_LOG_ERROR("Converting band number %d to corresponding index",
                list_of_bands[band_index]);
            return ERROR;
        }
        else if (parameters->process_band[sat_band_index])
        {
            IAS_LOG_ERROR("Duplicate band %d found in list",
                list_of_bands[band_index]);
            return ERROR;
        }
        else
        {
            /* First occurrence of band in list, flag it as 'found' */
            parameters->process_band[sat_band_index] = TRUE;
        }
    }

    /* Build the OMF filename */
    status = snprintf(parameters->omf_filename,
        sizeof(parameters->omf_filename), "%s.omf", local_wo_id);
    if (status < 0 || status >= sizeof(parameters->omf_filename))
    {
        IAS_LOG_ERROR("Creating the OMF filename");
        return ERROR;
    }

    /* Read in the OMF parameters */
    if (read_omf_parameters(parameters->omf_filename, local_processing_pass,
        parameters, cpf_filename) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading parameters from %s", parameters->omf_filename);
        return ERROR;
    }

    counts.water_mask = IAS_PARM_GET_COUNT_READ(parameter_table,
        WATER_MASK_FILENAME);
    counts.report_sampling_location = IAS_PARM_GET_COUNT_READ(parameter_table,
        REPORT_SAMPLING_LOCATION);
    counts.radiance_coefficient_filename = IAS_PARM_GET_COUNT_READ(
        parameter_table, RADIANCE_COEFFICIENT_FILENAME);
    counts.stray_light_vector_filename = IAS_PARM_GET_COUNT_READ(
        parameter_table, STRAY_LIGHT_VECTOR_FILENAME);
    counts.blur_kernel_size = IAS_PARM_GET_COUNT_READ(
        parameter_table, BLUR_KERNEL_SIZE);

    /* Process the input parameters */
    if (process_parameters(&counts, cpf_filename, sampling_location,
        parameters) != SUCCESS)
    {
        IAS_LOG_ERROR("Processing input parameters");
        return ERROR;
    }

    /* Verify the BAND/SCA/Detector defines */
    if (verify_local_defines(parameters) != SUCCESS)
    {
        IAS_LOG_ERROR("Verifying local defines");
        return ERROR;
    }

    return SUCCESS;
}
