/******************************************************************************
Name: read_radiance_coefficients

Purpose: Reads the 'a' and 'b' coefficients for all the detectors.

Returns: SUCCESS/ERROR

******************************************************************************/
/* Standard Library Includes */
#include <stdio.h>
#include <stdlib.h>

/* IAS Library Includes */
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_satellite_attributes.h"

/* Local Includes */
#include "read_radiance_coefficients.h"

int read_radiance_coefficients
(
    const char *filename,               /* I: Name of coefficients file */
    RADIANCE_COEFFICIENTS *rad_coeffs   /* O: Coefficients structure */
)
{
    int band_index;
    FILE *file;
    int line = 0;

    /* Open the file */
    file = fopen(filename, "r");
    if (!file)
    {
        IAS_LOG_ERROR("Opening file %s", filename);
        return ERROR;
    }

    /* Loop over the bands */
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        const IAS_BAND_ATTRIBUTES *band_attr;
        int sca_index;
        int band_number;
        int tirs_index;

        /* Retrieve the band number using the band index */
        band_number = ias_sat_attr_convert_band_index_to_number(
            band_index);
        if (band_number == ERROR)
        {
            IAS_LOG_ERROR("Converting band index %d to band number",
                band_index);
            fclose(file);
            return ERROR;
        }

        /* Read the band attributes */
        band_attr = ias_sat_attr_get_band_attributes(band_number);
        if (!band_attr)
        {
            IAS_LOG_ERROR("Retrieving band attributes for band number %d",
                band_number);
            fclose(file);
            return ERROR;
        }

        /* We are only concerned with the TIRS bands */
        if (band_attr->sensor_id != IAS_TIRS)
            continue;

        /* Retrieve the TIRS sensor index */
        tirs_index = ias_sat_attr_convert_band_number_to_sensor_band_index(
            IAS_TIRS, band_number);
        if (tirs_index == ERROR)
        {
            IAS_LOG_ERROR("Converting band number %d to sensor band index",
                band_number);
            fclose(file);
            return ERROR;
        }

        for (sca_index = 0; sca_index < band_attr->scas; sca_index++)
        {
            int det_index;
            int input_band_number;
            int sca_number;
            int det_number;

            for (det_index = 0; det_index < band_attr->detectors_per_sca;
                 det_index++)
            {
                line++;

                /* Parse the line */
                if (fscanf(file, "%d,%d,%d,%lf,%lf", &input_band_number,
                    &sca_number, &det_number,
                    &rad_coeffs->coeffs[tirs_index][sca_index][det_index].a,
                    &rad_coeffs->coeffs[tirs_index][sca_index][det_index].b)
                    != 5)
                {
                    IAS_LOG_ERROR("Parsing line %d of %s", line, filename);
                    fclose(file);
                    return ERROR;
                }

                /* Verify the line is the expected band/sca/detector */
                if ((input_band_number != band_number)
                    || (sca_number != sca_index + 1)
                    || (det_number != det_index + 1))
                {
                    IAS_LOG_ERROR("Parsing line %d of %s - not the "
                        "expected band number %d, SCA number %d, detector "
                        "number %d", line, filename, band_number,
                        sca_index + 1, det_index + 1);
                    fclose(file);
                    return ERROR;
                }
            }
        }
    }

    fclose(file);

    return SUCCESS;
}
