/******************************************************************************
Name: read_stray_light_vectors

Purpose: Reads the TIRS stray light vectors and weights from a comma delimited
    file containing one line per vector.  Each line contains the following
    information:
      x-comp, y-comp, z-comp, sca number, band number, detector number, weight

Returns: SUCCESS / ERROR

******************************************************************************/
/* Standard Library Includes*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* IAS Library Includes */
#include "ias_logging.h"
#include "ias_satellite_attributes.h"

/* Local Includes */
#include "read_stray_light_vectors.h"

int read_stray_light_vectors
(
    const char *filename,            /* I: Stray light vector filename */
    STRAY_LIGHT_VECTORS *stray_light /* O: Stray light vectors */
)
{
    int line_count;
    STRAY_LIGHT_RECORD *records;
    int index;
    FILE *file;
    int band_index;
    int previous_band = 0;
    int previous_sca = 0;
    int previous_detector = 0;
    int band_number_list[NUMBER_OF_TIRS_BANDS];
    int number_tirs_bands = 0;

    /* Initialize the output structure */
    memset(stray_light, 0, sizeof(*stray_light));

    /* Open the file */
    file = fopen(filename, "r");
    if (!file)
    {
        IAS_LOG_ERROR("Opening file %s", filename);
        return ERROR;
    }

    /* Get the count for the number of rows in the file */
    if (fscanf(file, "%d", &line_count) != 1)
    {
        IAS_LOG_ERROR("Parsing line for the number of records");
        fclose(file);
        return ERROR;
    }
    /* Verify the count is greater than 0 */
    if (line_count <= 0)
    {
        IAS_LOG_ERROR("The number of vectors must be greater than 0");
        fclose(file);
        return ERROR;
    }

    /* Allocate space for all the records */
    records = malloc(line_count * sizeof(*records));
    if (!records)
    {
        IAS_LOG_ERROR("Failed allocating memory for %d stray light vector "
            "records", line_count);
        fclose(file);
        return ERROR;
    }

    /* Get the list of TIRS bands */
    if (ias_sat_attr_get_sensor_band_numbers(IAS_TIRS, IAS_NORMAL_BAND,
        IAS_BLIND_BAND | IAS_SECONDARY_BAND, band_number_list,
        NUMBER_OF_TIRS_BANDS, &number_tirs_bands) != SUCCESS)
    {
        IAS_LOG_ERROR("Getting the TIRS band list");
        fclose(file);
        return ERROR;
    }

    /* Read all the records */
    for (index = 0; ; index++)
    {
        const IAS_BAND_ATTRIBUTES *band_attr;
        STRAY_LIGHT_RECORD *rec = &records[index];
        int sca_number;
        int band_number;
        int detector_number;
        int sca_index;
        int detector_index;
        double stray_light_vector_x;
        double stray_light_vector_y;
        double stray_light_vector_z;
        double weight;

        /* Check that the allocated memory hasn't been overrun */
        if (index > line_count)
        {
            IAS_LOG_ERROR("More records present (%d) than specified (%d)",
                index + 1, line_count);
            free(records);
            fclose(file);
            return ERROR;
        }

        /* Parse the line */
        if (fscanf(file, "%lf,%lf,%lf,%d,%d,%d,%lf",
            &stray_light_vector_x, &stray_light_vector_y, &stray_light_vector_z,
            &sca_number, &band_number, &detector_number, &weight) != 7)
        {
            /* Check if all lines specified were read */
            if (line_count == index)
                break;
            IAS_LOG_ERROR("Parsing line %d of %s", index + 1, filename);
            free(records);
            fclose(file);
            return ERROR;
        }
        rec->stray_light_vector.x = stray_light_vector_x;
        rec->stray_light_vector.y = stray_light_vector_y;
        rec->stray_light_vector.z = stray_light_vector_z;
        rec->weight = weight;

        /* Verify that the file is in the correct order */
        if (index == 0)
        {
            previous_band = band_number;
            previous_sca = sca_number;
            previous_detector = detector_number;
        }
        else
        {
            if (band_number < previous_band)
            {
                IAS_LOG_ERROR("Line %d of %s skips backwards from band %d "
                    "to band %d", index + 1, filename, previous_band,
                    band_number);
                free(records);
                fclose(file);
                return ERROR;
            }
            else if (band_number == previous_band)
            {
                if (sca_number < previous_sca)
                {
                    IAS_LOG_ERROR("Line %d of %s skips backwards from SCA "
                        "number %d to SCA number %d", index + 1, filename,
                        previous_sca, sca_number);
                    free(records);
                    fclose(file);
                    return ERROR;
                }
                if (sca_number == previous_sca
                    && detector_number < previous_detector)
                {
                    IAS_LOG_ERROR("Line %d of %s skips backwards from detector "
                        "number %d to detector number %d", index + 1, filename,
                        previous_detector, detector_number);
                    free(records);
                    fclose(file);
                    return ERROR;
                }
            }

            /* Reset the previous verification to the current */
            previous_band = band_number;
            previous_sca = sca_number;
            previous_detector = detector_number;
        }

        /* Retrieve the current band attributes */
        band_attr = ias_sat_attr_get_band_attributes(band_number);
        if (!band_attr)
        {
            IAS_LOG_ERROR("Retrieving band attributes for band number %d",
                band_number);
            free(records);
            fclose(file);
            return ERROR;
        }

        /* Verify read band number, SCA number, and detector number are valid */
        if (band_attr->sensor_id != IAS_TIRS
            || band_attr->band_classification != IAS_NORMAL_BAND
            || sca_number < 1 || sca_number > band_attr->scas
            || detector_number < 1
            || detector_number > band_attr->detectors_per_sca)
        {
            IAS_LOG_ERROR("Invalid band/SCA/detector combination for line %d "
                "of %s", index + 1, filename);
            free(records);
            fclose(file);
            return ERROR;
        }

        /* Build the index to detector records */
        /* Retrieve the TIRS sensor index */
        band_index = ias_sat_attr_convert_band_number_to_sensor_band_index(
            IAS_TIRS, band_number);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Converting band number %d to sensor band index",
                band_number);
            free(records);
            fclose(file);
            return ERROR;
        }
        sca_index = sca_number - 1;
        detector_index = detector_number - 1;

        /* If this is the vector found for this entry, set the index value */
        if (stray_light->count[band_index][sca_index][detector_index] == 0)
        {
            stray_light->indexed[band_index][sca_index][detector_index] = rec;
        }
        stray_light->count[band_index][sca_index][detector_index]++;
    }
    fclose(file);

    /* Verify all the bands/scas/detectors have stray light vectors available */
    for (band_index = 0; band_index < number_tirs_bands; band_index++)
    {
        const IAS_BAND_ATTRIBUTES *band_attr;
        int band_number;
        int sca_index;
        int tirs_band_index;

        /* Retrieve the band number using the band index */
        band_number = band_number_list[band_index];

        band_attr = ias_sat_attr_get_band_attributes(band_number);
        if (!band_attr)
        {
            IAS_LOG_ERROR("Retrieving band attributes for band number %d",
                band_number);
            free(records);
            return ERROR;
        }

        /* Retrieve the TIRS sensor index */
        tirs_band_index =
            ias_sat_attr_convert_band_number_to_sensor_band_index(
            IAS_TIRS, band_number);
        if (tirs_band_index == ERROR)
        {
            IAS_LOG_ERROR("Converting band number %d to sensor "
                "band index", band_number);
            free(records);
            return ERROR;
        }

        for (sca_index = 0; sca_index < band_attr->scas; sca_index++)
        {
            int detector_index;

            for (detector_index = 0; detector_index
                 < band_attr->detectors_per_sca; detector_index++)
            {
                /* Band 10, SCA 3, detector 640 does not have any stray light
                   vectors, so exclude it from the check.  */
                if (tirs_band_index == 0 && sca_index == 2
                    && detector_index == 639)
                    continue;

                if (stray_light->count[tirs_band_index][sca_index]
                        [detector_index] == 0)
                {
                    IAS_LOG_ERROR("No stray light vector for Band number %d, "
                        "SCA number %d, detector number %d", band_number,
                        sca_index + 1, detector_index + 1);
                    free(records);
                    return ERROR;
                }
            }
        }
    }
    stray_light->records = records;

    return SUCCESS;
}

/******************************************************************************
Name: free_stray_light_vectors

Purpose: Frees the memory allocated in a STRAY_LIGHT_VECTORS structure.

******************************************************************************/
void free_stray_light_vectors
(
    STRAY_LIGHT_VECTORS *stray_light /* I: Stray light vectors */
)
{
    free(stray_light->records);
    stray_light->records = NULL;
}
