/******************************************************************************
Name: sample_stray_light

Purpose: Creates a ghost image by sampling the stray light contributions as
    defined by the stray light vectors and weights.

******************************************************************************/
/* Standard Library Includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>

/* IAS Library Includes */
#include "ias_logging.h"
#include "ias_model_io.h"
#include "ias_threadpool.h"
#include "ias_satellite_attributes.h"

/* Local Includes */
#include "grid.h"
#include "convert_stray_light_to_lat_long.h"
#include "find_pixel_nearest_polygon.h"
#include "read_stray_light_vectors.h"
#include "read_radiance_coefficients.h"
#include "sample_stray_light.h"
#include "calculate_average_land_water_radiance.h"
#include "create_projection_transformation.h"
#include "local_defines.h"

/* Define this to use the LOS model for the lat/long calculations.  The
   model isn't used at this time since it is significantly slower. */
/*#define USE_LOS_MODEL */
#ifndef USE_LOS_MODEL
#define stray_light_to_lat_lng(cache, los, lat, lng) \
    find_lat_long(cache, los, lat, lng)
#else
#define stray_light_to_lat_lng(cache, los, lat, lng) \
    convert_stray_light_to_lat_long_cached(cache, los, lat, lng)
#endif

/****************************************************************************
Name: find_lat_long

Purpose: Calculate the lat/long of the intersection of a line-of-sight vector
   with the Earth ellipsoid given a position vector of the observer.

Returns: SUCCESS / ERROR

****************************************************************************/
static int find_lat_long
(
    MODEL_CACHE *cache,     /* I: Cached line of sight model info */
    const IAS_VECTOR *los,  /* I: Stray light vector */
    double *lat,            /* O: Calculated Lat/Long in radians */
    double *lng             /* O: Calculated Lat/Long in radians */
)
{
    double B[3];
    double D[3];
    double PdotB;
    double MagC;
    double inv_MagB;
    double mag_C_div_B;
    int k;

    /* Check that the stray light vectors have been initialized */
    if (!los)
    {
        IAS_LOG_ERROR("Stray light vector has not been initialized");
        return ERROR;
    }

    double los_ecef[3];

    for (k = 0; k < 3; k++)
    {
        /* Transform stray light vector from TIRS coord. to ECEF coord */
        los_ecef[k] = cache->los2sc[k][0] * los->x
            + cache->los2sc[k][1] * los->y
            + cache->los2sc[k][2] * los->z;

        B[k] = los_ecef[k] * cache->inv_ellipsoid[k];
    }

    inv_MagB = 1.0 / COMPUTE_VECTOR_LENGTH(B);
    PdotB = -1.0 * (cache->P[0] * B[0] + cache->P[1] * B[1] + cache->P[2]
        * B[2]) * inv_MagB;
    MagC = (PdotB - sqrt(PdotB * PdotB + 1 - cache->MagP * cache->MagP));
    mag_C_div_B = MagC * inv_MagB;
    for (k = 0; k < 3; k++)
    {
        double C = mag_C_div_B * B[k];
        D[k] = (cache->P[k] + C) * cache->ellipsoid[k];
    }

    *lat = atan(D[2]/(0.99330562 * sqrt(D[0] * D[0] + D[1] * D[1])));
    *lng =  atan2(D[1], D[0]);

    return SUCCESS;
}

/****************************************************************************
Name: cleanup_sample_stray_light

Purpose: Cleanup routine for sample_stray_light to create cleaner code.

Returns: void

****************************************************************************/
static void cleanup_sample_stray_light
(
    FILE **point_file,                    /* I: File pointer */
    IAS_GEO_PROJ_TRANSFORMATION **geo2l1g,/* I: Geographic transformation */
    IAS_GEO_PROJ_TRANSFORMATION **l1g2geo /* I: L1G transformation */
)
{
    /* Cleanup */
    if (point_file && *point_file)
        fclose(*point_file);
    if (geo2l1g)
    {
        ias_geo_destroy_proj_transformation(*geo2l1g);
        *geo2l1g = NULL;
    }
    if (l1g2geo)
    {
        ias_geo_destroy_proj_transformation(*l1g2geo);
        *l1g2geo = NULL;
    }
}

/****************************************************************************
Name: report_sampling_location

Purpose: Reports the location of the input line and sample after converting
         it to geographic projection.

Return: Type = Integer
        SUCESS / ERROR
****************************************************************************/
static int report_sampling_location
(
    IAS_GEO_PROJ_TRANSFORMATION **l1g2geo, /* I/O: L1G to GEO Transformation */
    const L1G_INFO *l1g_info,              /* I: L1G info */
    FILE *point_file,                      /* I: Points file ptr */
    int line,                              /* I: Location line */
    int sample,                            /* I: Location sample */
    int output_nearest_flag                /* I: Flag to indicate output nearest
                                                 or output coordinate */
)
{
    IAS_DBL_LAT_LONG geo_coords;
    IAS_DBL_XY proj_coords;
    char output_type = 'O';

    /* Create the geographic transformation if it has not
       yet been created */
    if (l1g2geo && !*l1g2geo)
    {
        *l1g2geo = create_source_to_geographic_transformation(&l1g_info->proj);
        if (!l1g2geo)
        {
            IAS_LOG_ERROR("Creating geographic transformation");
            return ERROR;
        }
    }

    /* Determine the output format */
    if (output_nearest_flag)
        output_type = 'N';

    /* Convert line/sample to x/y location */
    proj_coords.y = l1g_info->corners.upleft.y - line * l1g_info->y_pixel_size;
    proj_coords.x = sample * l1g_info->x_pixel_size
        + l1g_info->corners.upleft.x;

    /* Convert the x/y combination to geographic latitude and longitude */
    if (ias_geo_transform_coordinate(*l1g2geo, proj_coords.x, proj_coords.y,
        &geo_coords.lng, &geo_coords.lat) != SUCCESS)
    {
        IAS_LOG_ERROR("Transforming line/sample to geographic coordinates");
        return ERROR;
    }

    /* Output the coordinates */
    if (fprintf(point_file, "%c %d %d %f %f\n", output_type, sample + 1,
        line + 1, geo_coords.lng, geo_coords.lat) < 1)
    {
        IAS_LOG_ERROR("Writing to point file");
        return ERROR;
    }

    return SUCCESS;
}

/****************************************************************************
Name: sample_stray_light

Purpose: For each stray light vector lat/long, find the nearest lat/long in the
         external sensor data and take the radiance value associated with that
         nearest lat/long.

Return: Type = Integer
        SUCESS / ERROR
****************************************************************************/
int sample_stray_light
(
    void *params,   /* I: Sample parameters struct */
    int thread_num  /* I: Thread number */
)
{
    SAMPLE_PARAMS *parms = (SAMPLE_PARAMS *)params;
    const IAS_LOS_MODEL *model = parms->model;
    const GRID *grid = parms->grid;
    int valid_grid_flag = parms->valid_grid_flag;
    L1G_INFO *l1g_info = parms->l1g_info;
    int sample_location_flag = parms->sample_location_requested_flag;
    SAMPLING_LOCATION sampling_location = *parms->sampling_location;
    int start_frame = parms->start_frame;
    int end_frame = parms->end_frame;
    int l1r_frames = parms->l1r_frames;
    int band_index = parms->band_index;
    int band_number = parms->band_number;
    int sca_index = parms->sca_index;
    int detector_count = parms->detector_count;
    int apply_radiance_coefficients = parms->apply_radiance_coefficients;
    const int *corner_lines = parms->corner_lines;
    const int *corner_samples = parms->corner_samples;
    const STRAY_LIGHT_VECTORS *stray_light_vectors = parms->stray_light_vectors;
    const RADIANCE_COEFFICIENTS *rad_coeffs = parms->rad_coeffs;
    float *ghost = parms->ghost;
    double major_axis = parms->semi_major_axis;
    double minor_axis = parms->semi_minor_axis;
    double r2d = ias_math_get_degrees_per_radian();
    FILE *point_file = NULL;
    IAS_GEO_PROJ_TRANSFORMATION *l1g2geo = NULL;
    IAS_GEO_PROJ_TRANSFORMATION *geo2l1g = NULL;
    int l1r_frame;
    MODEL_CACHE cache;
    int outside_grid_count = 0;

    /* Retrieve the TIRS sensor index */
    int tirs_band_index = ias_sat_attr_convert_band_number_to_sensor_band_index(
        IAS_TIRS, band_number);
    if (tirs_band_index == ERROR)
    {
        IAS_LOG_ERROR("Converting band number %d to sensor band index",
            band_number);
        return ERROR;
    }

    const RAD_COEFF *coeffs_per_detector = rad_coeffs->coeffs
        [tirs_band_index][sca_index];
    const int *sl_vector_count = stray_light_vectors->count[tirs_band_index]
        [sca_index];
    STRAY_LIGHT_RECORD *const *sl_vectors = stray_light_vectors->indexed
        [tirs_band_index][sca_index];

    /* Open the sample location points file if it has been requested and it
       is the correct band and SCA */
    if (sample_location_flag && band_index == sampling_location.band_index
        && sca_index == sampling_location.sca_index)
    {
        char filename[200];
        int status;

        /* Open a per-thread points file that will be merged using an external
           script */
        status = snprintf(filename, sizeof(filename), "points_t%d.txt",
            thread_num);
        if (status < 0 || status >= sizeof(filename))
        {
            IAS_LOG_ERROR("Filename buffer is not large enough");
            return ERROR;
        }

        point_file = fopen(filename,"w");
        if (!point_file)
        {
            IAS_LOG_ERROR("Opening point file %s", filename);
            return ERROR;
        }
    }
    else
    {
        /* Didn't pass the band/sca filter, so not sampling this time */
        sample_location_flag = 0;
    }

    memset(&cache, 0, sizeof(cache));
    cache.line = -1;

    /* Walk through all the stray light vectors */
    while(1)
    {
        int ghost_frame;
        int det_index;

        /* Lock the frame mutex */
        if (IAS_THREAD_LOCK_MUTEX(&parms->frame_mutex) != 0)
        {
            IAS_LOG_ERROR("Locking the tile mutex");
            cleanup_sample_stray_light(&point_file, NULL, NULL);
            return ERROR;
        }

        l1r_frame = parms->next_frame;
        parms->next_frame++;

        /* Unlock the frame mutex */
        if (IAS_THREAD_UNLOCK_MUTEX(&parms->frame_mutex) != 0)
        {
            IAS_LOG_ERROR("Unlocking the tile mutex");
            cleanup_sample_stray_light(&point_file, NULL, NULL);
            return ERROR;
        }

        /* Done after the ending frame has been assigned */
        if (l1r_frame > end_frame)
            break;

        ghost_frame = l1r_frame - start_frame;

        /* Issue a period message to let us know progress is being made */
        if (ghost_frame % 1000 == 0)
        {
            /* Print timestamp of every 1000th line for sanity check */
            IAS_LOG_INFO("Creating output image line %d of %d", ghost_frame,
                l1r_frames);
        }

        /* Cache the model information for the current frame */
        if (prepare_cache_for_line(l1r_frame, band_index, sca_index, model,
            major_axis, minor_axis, &cache) != SUCCESS)
        {
            IAS_LOG_ERROR("Updating model cache");
            cleanup_sample_stray_light(&point_file, NULL, NULL);
            return ERROR;
        }

        /* For the first frame of ghost image, print out the timestamp and
           satellite position to help cal/val line up the data with the
           prototype */
        if (ghost_frame == 0)
        {
            IAS_LOG_DEBUG("Ghost image first frame timestamp %f",
                cache.image_time);
            IAS_LOG_DEBUG("Ghost image first frame satellite position: x: %f, "
                "y %f, z %f", cache.satpos.x, cache.satpos.y, cache.satpos.z);
        }

        /* Walk through the stray light vectors for each detector */
        for (det_index = 0; det_index < detector_count; det_index++)
        {
            int count = sl_vector_count[det_index];
            const STRAY_LIGHT_RECORD *sl = sl_vectors[det_index];
            int sl_index;
            double ghost_pixel = 0.0;
            const RAD_COEFF *coeffs = &coeffs_per_detector[det_index];

            for (sl_index = 0; sl_index < count; sl_index++)
            {
                IAS_DBL_XY coords;
                IAS_LNG_LS l1g_ls;
                float pixel;
                int is_fill;
                const IAS_VECTOR *los = &sl[sl_index].stray_light_vector;
                double lat;
                double lng;
                int output_point = 0;
                int outside_grid_flag = 0;

                /* Calculate the lat/long */
                if (stray_light_to_lat_lng(&cache, los, &lat, &lng) != SUCCESS)
                {
                    IAS_LOG_ERROR("Converting stray light to lat/lng");
                    cleanup_sample_stray_light(&point_file, NULL, NULL);
                    return ERROR;
                }

                lat *= r2d;
                lng *= r2d;

                if (valid_grid_flag)
                {
                    /* A valid grid is available, so use it to convert lat/long
                       to the source L1G projection */
                    int cell_row = (int)((grid->ul_y - lat) / grid->cell_size);
                    int cell_col = (int)((lng - grid->ul_x) / grid->cell_size);
                    int cell_index;
                    IAS_COEFFICIENTS *coeff_ptr;

                    if (cell_row < 0 || cell_row >= grid->rows
                        || cell_col < 0 || cell_col >= grid->columns)
                    {
                        /* If outside the grid, set the flag to let it fall
                           back to using the full transformation.  Also count
                           the number of points that fall outside the grid to
                           allow a message to be output. */
                        outside_grid_flag = 1;
                        outside_grid_count++;

                        IAS_LOG_DEBUG("Cell outside grid (%d, %d) of (%d, %d)",
                            cell_row, cell_col, grid->rows, grid->columns);
                    }
                    else
                    {
                        cell_index = cell_row * grid->columns + cell_col;
                        coeff_ptr = &grid->coeffs[cell_index];

                        coords.x = coeff_ptr->a[0] + coeff_ptr->a[1] * lng +
                            coeff_ptr->a[2] * lat + coeff_ptr->a[3] * lng * lat;
                        coords.y = coeff_ptr->b[0] + coeff_ptr->b[1] * lng +
                            coeff_ptr->b[2] * lat + coeff_ptr->b[3] * lng * lat;
                    }
                }

                /* If a valid grid isn't present or the current location falls
                   off the grid edges, fall back to using the full projection
                   transformation */
                if (!valid_grid_flag || outside_grid_flag)
                {
                    /* Create the target projection transformation if it has not
                       yet been created */
                    if (!geo2l1g)
                    {
                        geo2l1g = create_geographic_to_target_transformation(
                            &l1g_info->proj);
                        if (!geo2l1g)
                        {
                            IAS_LOG_ERROR("Creating projection transformation");
                            cleanup_sample_stray_light(&point_file, NULL, NULL);
                            return ERROR;
                        }
                    }

                    /* Convert the stray light lat/lng to projection
                       coordinates using a full transformation for every
                       coordinate.  This is slow and is only used if a valid
                       grid couldn't be made (i.e. because the longitudes cross
                       180 degrees. */
                    if (ias_geo_transform_coordinate(geo2l1g, lng, lat ,
                        &coords.x, &coords.y) != SUCCESS)
                    {
                        IAS_LOG_ERROR("Transforming coordinates\n");
                        cleanup_sample_stray_light(&point_file, &geo2l1g, NULL);
                        return ERROR;
                    }
                }

                /* Convert the x/y location to an L1G line/sample */
                l1g_ls.line = round((l1g_info->corners.upleft.y - coords.y)
                                    / l1g_info->y_pixel_size);
                l1g_ls.samp = round((coords.x - l1g_info->corners.upleft.x)
                                    / l1g_info->x_pixel_size);

                /* Report the sampling location if it matches the current
                   frame and detector */
                if (sample_location_flag
                    && l1r_frame == sampling_location.frame_index
                    && det_index == sampling_location.detector_index)
                {
                    if (report_sampling_location(&l1g2geo, l1g_info, point_file,
                        l1g_ls.line, l1g_ls.samp, 0) != SUCCESS)
                    {
                        IAS_LOG_ERROR("Reporting output locations");
                        cleanup_sample_stray_light(&point_file, &geo2l1g, NULL);
                        return ERROR;
                    }
                    output_point = 1;
                }

                is_fill = 1;
                if (l1g_ls.line >= 0 && l1g_ls.line < l1g_info->lines
                    && l1g_ls.samp >= 0 && l1g_ls.samp < l1g_info->samples)
                {
                    pixel = l1g_info->radiance[l1g_ls.line * l1g_info->samples +
                        l1g_ls.samp];
                    if (pixel > 0.0)
                        is_fill = 0;
                }
                if (is_fill)
                {
                    if (parms->use_land_water_radiance_average_flag)
                    {
                        int mask_index = (l1g_ls.line + parms->mask_line_offset)
                            * parms->mask_samples + l1g_ls.samp
                            + parms->mask_samp_offset;

                        /* Set the pixel using the land water mask */
                        if (parms->water_mask[mask_index] == LAND_PIXEL)
                        {
                            pixel = l1g_info->average_land_radiance;
                        }
                        else
                        {
                            pixel = l1g_info->average_water_radiance;
                        }
                    }
                    else
                    {
                        /* Find the nearest non-fill pixel to use */
                        int nearest_line;
                        int nearest_sample;
                        int status;

                        /* Get the radiance from the nearest pixel.  Note that
                           rounded l1g_line/sample locations are good enough
                           since the input image is blurred input image. */
                        status = find_pixel_nearest_polygon(l1g_ls.line,
                            l1g_ls.samp, corner_lines, corner_samples, l1g_info,
                            &nearest_line, &nearest_sample, &pixel);
                        if (status != SUCCESS)
                        {
                            IAS_LOG_ERROR("Unexpected error finding nearest "
                                "pixel");
                            cleanup_sample_stray_light(&point_file, &geo2l1g,
                                &l1g2geo);
                            return ERROR;
                        }

                        if (output_point)
                        {
                            if (report_sampling_location(&l1g2geo, l1g_info,
                                point_file, nearest_line, nearest_sample, 1)
                                != SUCCESS)
                            {
                                IAS_LOG_ERROR("Reporting nearest locations");
                                cleanup_sample_stray_light(&point_file,
                                    &geo2l1g, NULL);
                                return ERROR;
                            }
                        }
                    }
                }

                ghost_pixel += pixel * sl[sl_index].weight;
            }

            /* Apply the 'a' and 'b' coefficients if requested.  The
               coefficients will normally be applied, but there are times
               cal/val may want to not apply them. */
            if (apply_radiance_coefficients)
            {
                ghost_pixel = ghost_pixel * coeffs->a + coeffs->b;
            }

            ghost[ghost_frame * NUMBER_OF_TIRS_DETECTORS + det_index]
                = ghost_pixel;
        }
    }

    if (point_file)
    {
        if (fclose(point_file) != 0)
        {
            IAS_LOG_ERROR("Closing point location file");
            return ERROR;
        }
    }

    /* Update the outside grid flag in the parameters to let the caller issue
       one message instead of having multiple threads provide the same
       message */
    if (outside_grid_count > 0)
        parms->outside_grid_flag = 1;

    /* Destroy the transformations */
    cleanup_sample_stray_light(NULL, &geo2l1g, &l1g2geo);

    return SUCCESS;
}

/*****************************************************************************
Name: find_bound_box_using_model

Purpose: Find the min/max lat/long bounding box using the model.


Note: This should really be doing a walk along the edges somehow, but for
      now, checking the top and bottom is providing the same value as doing the
      entire image
*****************************************************************************/
int find_bounding_box_using_model
(
    const IAS_LOS_MODEL *model, /* I: LOS Model */
    const STRAY_LIGHT_VECTORS *stray_light_vectors, /* I: Stray light vectors */
    int band_index,         /* I: Band index */
    int sca_index,          /* I: SCA index */
    int frame_count,        /* I: Number of frames in imagery */
    double major_axis,      /* I: Semi-major axis */
    double minor_axis,      /* I: Semi-minor axis */
    int frame_sections,     /* I: Number of frame sections to sample */
    double *out_min_lat,    /* O: Minimum latitude in degrees */
    double *out_max_lat,    /* O: Maximum latitude in degrees */
    double *out_min_lng,    /* O: Minimum longitude in degrees */
    double *out_max_lng     /* O: Maximum longitude in degrees */
)
{
    MODEL_CACHE cache;
    double min_lat = 91.0;
    double max_lat = -91.0;
    double min_lng = 361.0;
    double max_lng = -361.0;
    int frame;
    int band_number;
    int tirs_band_index;
    double r2d = ias_math_get_degrees_per_radian();
    int frame_increment;
    int last_frame = frame_count - 1;

    /* Retrieve the band number using the band index */
    band_number = ias_sat_attr_convert_band_index_to_number(band_index);
    if (band_number == ERROR)
    {
        IAS_LOG_ERROR("Converting band index %d to band number", band_index);
        return ERROR;
    }

    /* Retrieve the TIRS sensor index */
    tirs_band_index = ias_sat_attr_convert_band_number_to_sensor_band_index(
        IAS_TIRS, band_number);
    if (tirs_band_index == ERROR)
    {
        IAS_LOG_ERROR("Converting band number %d to sensor band index",
            band_number);
        return ERROR;
    }

    /* Initialize the cache */
    memset(&cache, 0, sizeof(cache));
    cache.line = -1;

    /* Calculate the frame increment, which is used to walk along
       the edge of the image incase curvature is present */
    frame_increment = frame_count / frame_sections;

    /* loop over the start and end frames */
    for (frame = 0; frame < frame_count; frame += frame_increment)
    {
        int det_index;

        /* Cache the LOS model info for this frame */
        if (prepare_cache_for_line(frame, band_index, sca_index, model,
            major_axis, minor_axis, &cache) != SUCCESS)
        {
            IAS_LOG_ERROR("Updating model cache");
            return ERROR;
        }

        /* Loop through the detectors of this band/SCA */
        for (det_index = 0; det_index < NUMBER_OF_TIRS_DETECTORS;
             det_index++)
        {
            int count = stray_light_vectors->count[tirs_band_index][sca_index]
                [det_index];
            STRAY_LIGHT_RECORD *sl = stray_light_vectors->indexed
                [tirs_band_index][sca_index][det_index];
            int index;

            /* Loop through the stray light vectors for this detector */
            for (index = 0; index < count; index++)
            {
                /* Use the shortcut method of calculating the lat/long.
                   Not as accurate as the full LOS model, but quite a bit
                   faster at this time. */
                IAS_VECTOR *los = &sl[index].stray_light_vector;
                double lat;
                double lng;

                /* get the lat/long */
                if (stray_light_to_lat_lng(&cache, los, &lat, &lng) != SUCCESS)
                {
                    IAS_LOG_ERROR("Converting stray light to lat/lng");
                    return ERROR;
                }

                /* Find the min/max lat/long */
                if (lat < min_lat)
                    min_lat = lat;
                if (lat > max_lat)
                    max_lat = lat;
                if (lng < min_lng)
                    min_lng = lng;
                if (lng > max_lng)
                    max_lng = lng;
            }
        }

        /* Ensure that the last frame is sampled */
        if (frame != last_frame && frame + frame_increment > last_frame)
        {
            frame_increment = last_frame - frame;
        }
    }

    *out_min_lat = min_lat * r2d;
    *out_max_lat = max_lat * r2d;
    *out_min_lng = min_lng * r2d;
    *out_max_lng = max_lng * r2d;

    return SUCCESS;
}
