/* Standard Library Includes */
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

/* IAS Library Includse */
#include "ias_logging.h"
#include "ias_satellite_attributes.h"

/* COTS Includes */
#include "cgreen/cgreen.h"

/* Local Defines */
#define NUMBER_OF_TESTS 3

/*******************************************************************************
 * Name: test_create_tirs_ghost_image
 *
 * Purpose: This test uses 2 normal runs and 1 using the land/water radiance 
 *          averages to test the tirs ghost image creation.
 *          
 * Notes: Adding to this test do the following:
 *        1. Increment NUMBER_OF_TESTS
 *        2. Add directory name to tests
 *        3. Place all the needed files under: 
 *             - $IAS_UNIT_TEST_DATA_DIR/ias_base/rps/create_tirs_ghost_image/
 *        4. Rename the odl and omf files to odl/omf
 *        5. Update the paths in the omf
 *        6. Update the vector and coefficient paths in the odl and update
 *           the work order id to be the same as the directory name.
 *
 * ****************************************************************************/
Ensure test_create_tirs_ghost_image
(
)
{
    int test_index; /* Test loop variable */

    /* Tests directories */
    const char *tests[NUMBER_OF_TESTS] = { "test_01", "test_02", "test_03"};
    
    /* OMF Sed command */
    const char *omf_sed_cmd = "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s/ias_base/"
        "rps/create_tirs_ghost_image/%s|' %s/ias_base/rps/"
        "create_tirs_ghost_image/%s/omf > %s.omf";

    /* ODL Sed command */
    const char *odl_sed_cmd = "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s/ias_base/"
        "rps/create_tirs_ghost_image/coefficients|' %s/ias_base/rps/"
        "create_tirs_ghost_image/%s/odl > %s.odl";

    /* Create TIRS ghost image command */
    const char *create_cmd = "export IAS_LOG_LEVEL=ERROR; "
        "../create_tirs_ghost_image %s.odl";

    /* H5Diff Command for ghost image files */
    const char *h5diff_cmd = "h5diff -q -n 1 %s/ias_base/rps/"
        "create_tirs_ghost_image/%s/stray_light_correction.h5 "
        "stray_light_correction.h5";

    /* Test Cleanup command */
    const char *cleanup_cmd = "rm stray_light_correction.h5 %s.omf %s.odl";

    /* Unit test directory */
    char *unit_test_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!unit_test_dir)
    {
        assert_true_with_message(0, "Error: Getting unit test directory");
        return;
    }

    /* Set the version override flag so the h5 ias versions are the same */
    if (setenv("IAS_VERSION_OVERRIDE", "Y", 1) != 0)
    {
        assert_true_with_message(0, "Error: Setting ias version override");
        return;
    }

    /* Run through the tests */
    for (test_index = 0; test_index < NUMBER_OF_TESTS; test_index++)
    {
        char system_command[PATH_MAX];
        const char *cur_test = tests[test_index];
        int status;

        printf("\nRunning test (%s)\n", cur_test);
 
        /* Create the omf file */
        status = snprintf(system_command, sizeof(system_command), omf_sed_cmd,
            unit_test_dir, cur_test, unit_test_dir, cur_test, cur_test);
        if (status < 0 || status >= sizeof(system_command))
        {
            assert_true_with_message(0, "Error: Creating system sed command "
                "for the create_tirs_ghost_image omf file (%s)", cur_test);
            return;
        }
        else if (system(system_command) != 0)
        {
            assert_true_with_message(0, "Error: Running command %s", 
                system_command);
            return;
        }

        /* Create the odl file */
        status = snprintf(system_command, sizeof(system_command), odl_sed_cmd,
            unit_test_dir, unit_test_dir, cur_test, cur_test);
        if (status < 0 || status >= sizeof(system_command))
        {
            assert_true_with_message(0, "Error: Creating system sed command "
                "for the create_tirs_ghost_image odl file (%s)", cur_test);
            return;
        }
        else if (system(system_command) != 0)
        {
            assert_true_with_message(0, "Error: Runing command %s", 
                system_command);
            return;
        }

        /* Run the application */
        status = snprintf(system_command, sizeof(system_command), create_cmd, 
            cur_test);
        if (status < 0 || status >= sizeof(system_command))
        {
            assert_true_with_message(0, "Error: Creating the call command "
                "for create_tirs_ghost_image for test (%s)", cur_test);
            return;
        }
        else if (system(system_command) != 0)
        {
            assert_true_with_message(0, "Error: Runing command %s", 
                system_command);
            return;
        }

        /* Run the H5Diff on the stray light correction files */
        status = snprintf(system_command, sizeof(system_command), h5diff_cmd, 
            unit_test_dir, cur_test);
        if (status < 0 || status >= sizeof(system_command))
        {
            assert_true_with_message(0, "Error: Creating the call command "
                "for create_tirs_ghost_image for test (%s)", cur_test);
            return;
        }
        else if (system(system_command) != 0)
        {
            assert_true_with_message(0, "Error: Runing command %s", 
                system_command);
            return;
        }

        /* Cleanup the output files */
        status = snprintf(system_command, sizeof(system_command), cleanup_cmd, 
            cur_test, cur_test);
        if (status < 0 || status >= sizeof(system_command))
        {
            assert_true_with_message(0, "Error: Creating the cleanup command "
                "for create_tirs_ghost_image for test (%s)", cur_test);
            return;
        }
        else if (system(system_command) != 0)
        {
            assert_true_with_message(0, "Error: Runing command %s", 
                system_command);
            return;
        }
        
        assert_true(1);
    }
}

int main(int argc, char **argv)
{
    TestSuite *suite = create_test_suite();

    /* Tests Start */
    add_test(suite, test_create_tirs_ghost_image);
    /* Tests End */

    if (argc > 1)
        return run_single_test(suite, argv[1], create_text_reporter());
    return run_test_suite(suite, create_text_reporter());
}

