/*******************************************************************************
Name: write_l1r_metadata

Purpose: Write out the l1r file and band metadata

Returns:
     Type = integer
     SUCCESS / ERROR
 ******************************************************************************/
/* Standard Library Includes */
#include <stdlib.h>
#include <string.h>

/* IAS Library Includes */
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_l1r.h"

/* Local Includes */
#include "write_l1r_metadata.h"
#include "local_defines.h"

int write_l1r_metadata
(
    const char *source_image_filename, /* I: Stray light image source */
    const char *source_image_sensor,   /* I: Stray light image sensor */
    int used_land_mask_flag,/* I: Flag to indicate the land mask was used during
                                  processing */
    L1RIO *l1rio_ptr        /* I/O: L1R pointer */
)
{
    IAS_L1R_FILE_METADATA fmd;    /* File metadata */
    IAS_L1R_BAND_METADATA bmd[NUMBER_OF_TIRS_BANDS]; /* Band metadata */
    int band_number_list[NUMBER_OF_TIRS_BANDS];
    const char *ias_version = NULL;
    int index;
    int number_tirs_bands = 0;

    /* Zero out the metadata */
    memset(&bmd, 0, sizeof(bmd));
    memset(&fmd, 0, sizeof(fmd));

    /* Set the image source. This will make it to the eventually make it to the
       MTL file. The flow it takes is the following:
       create_tirs_ghost_image -> driver -> OMF -> format_and_package -> MTL */
    if (strcmp("TIRS", source_image_sensor) == 0)
    {
        strncpy(fmd.collect_type, "TIRS", sizeof(fmd.collect_type));
    }
    else
    {
        const char *base_file = strrchr(source_image_filename, '/');
        if (base_file)
            base_file++;
        else
            base_file = source_image_filename;
        strncpy(fmd.collect_type, base_file, sizeof(fmd.collect_type));
    }
    fmd.collect_type[sizeof(fmd.collect_type) - 1] = '\0';

    /* Grab the IAS software version */
    ias_version = ias_misc_get_software_version();
    strncpy(fmd.software_version, ias_version, sizeof(fmd.software_version));
    fmd.software_version[sizeof(fmd.software_version) - 1] = '\0';

    /* Set the custom radiometric steps */
    strncpy(fmd.custom_rad_processing_steps_applied, "NONE",
        sizeof(fmd.custom_rad_processing_steps_applied));
    fmd.custom_rad_processing_steps_applied[sizeof(
        fmd.custom_rad_processing_steps_applied) - 1] = '\0';

    /* Get the list of TIRS bands */
    if (ias_sat_attr_get_sensor_band_numbers(IAS_TIRS, IAS_NORMAL_BAND,
        IAS_BLIND_BAND | IAS_SECONDARY_BAND, band_number_list,
        NUMBER_OF_TIRS_BANDS, &number_tirs_bands) != SUCCESS)
    {
        IAS_LOG_ERROR("Getting the TIRS band list");
        return ERROR;
    }

    /* Loop through each of the stray light correction bands */
    for (index = 0; index < number_tirs_bands; index++)
    {
        char band_name[IAS_BAND_NAME_SIZE];
        int status;

        /* Set the band number */
        bmd[index].band_number = band_number_list[index];

        status = snprintf(band_name, sizeof(band_name),
            "B%d stray light correction", bmd[index].band_number);
        if (status < 0 || status >= sizeof(band_name))
        {
            IAS_LOG_ERROR("Creating band name for band number %d",
                bmd[index].band_number);
            return ERROR;
        }

        /* Set the band name */
        strncpy(bmd[index].band_name, band_name, sizeof(bmd[index].band_name));
        bmd[index].band_name[sizeof(bmd[index].band_name) - 1] = '\0';

        /* Set the instrument source */
        strncpy(bmd[index].instrument_source, source_image_sensor,
            sizeof(bmd[index].instrument_source));
        bmd[index].instrument_source[sizeof(bmd[index].instrument_source) - 1]
            = '\0';

        /* Set the discontinuity factor source */
        strncpy(bmd[index].discontinuity_factor_source, "NONE",
            sizeof(bmd[index].discontinuity_factor_source));
        bmd[index].discontinuity_factor_source[sizeof(
            bmd[index].discontinuity_factor_source) - 1] = '\0';

        /* Set the bias source */
        strncpy(bmd[index].bias_source, "NONE", sizeof(bmd[index].bias_source));
        bmd[index].bias_source[sizeof(bmd[index].bias_source) - 1] = '\0';

        /* Set the linearization */
        strncpy(bmd[index].linearization, "NONE",
            sizeof(bmd[index].linearization));
        bmd[index].linearization[sizeof(bmd[index].linearization) - 1] = '\0';
    }

    /* Write the metadata */
    if ((ias_l1r_set_file_metadata(l1rio_ptr, &fmd) != SUCCESS) ||
        (ias_l1r_set_band_metadata(l1rio_ptr, bmd, NUMBER_OF_TIRS_BANDS)
        != SUCCESS))
    {
        IAS_LOG_ERROR("Updating metadata for L1G image");
        return ERROR;
    }

    return SUCCESS;
}
