/*----------------------------------------------------------------------------
NAME:    char_lamp_detector_response

PURPOSE: Process the post warmup lamp data to get the radiance statistics 
         for each detector, each SCA, and each band. Store the statistics
         results together with lamp telemetry and lamp response data into 
         database and optionally output them into the report file.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include "ias_const.h"
#include "ias_math.h"
#include "ias_logging.h"
#include "ias_pixel_mask.h"
#include "ias_radiometry.h"
#include "ias_db_insert.h"
#include "ias_satellite_attributes.h"
#include "rps_local.h"

#define STRLEN 64

/* Prototypes for report generation */
static FILE *create_detector_report_file(int, int, const char *, IAS_REPORT *);
static FILE *create_sca_report_file(int, const char *, IAS_REPORT *);
static FILE *create_band_report_file(const char *, IAS_REPORT *);
static FILE *create_report_file(IAS_REPORT *, REPORT_FILE_TYPE);

static int write_lamp_response_data
(
    FILE *report_file,           /* I: Report file pointer */
    int lamp_pair_id,            /* I: Lamp pair id */
    float lamp_on_time,          /* I: Lamp power on time */
    float post_warmup_start_time,/* I: Start time for post warmup period */
    float post_warmup_stop_time, /* I: Stop time for post warmup period */
    int post_warmup_start_frame, /* I: Start frame number for post warmup 
                                       period */ 
    int post_warmup_stop_frame   /* I: Stop frame number for post warmup 
                                       period */ 
);

static int write_lamp_response_detector_report_file
( 
    FILE *report_file,           /* I: Report file pointer */
    DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS *db_detector_results,
                                 /* I: Pointer to results of the 
                                       lamp response detector results */
    int lamp_pair_id,            /* I: Lamp pair id */
    float lamp_on_time,          /* I: Lamp power on time */
    float post_warmup_start_time,/* I: Start time for post warmup period */
    float post_warmup_stop_time, /* I: Stop time for post warmup period */
    int post_warmup_start_frame, /* I: Start frame number for post warmup 
                                       period */ 
    int post_warmup_stop_frame   /* I: Stop frame number for post warmup 
                                       period */ 
);

static int write_lamp_response_sca_report_file
(
    FILE *report_file,           /* I: Report file pointer */
    DB_CHAR_LAMP_RESPONSE_SCA_RESULTS *db_sca_results,
                                 /* I: Pointer to results of the 
                                       lamp response SCA results */
    int lamp_pair_id,            /* I: Lamp pair id */
    float lamp_on_time,          /* I: Lamp power on time */
    float post_warmup_start_time,/* I: Start time for post warmup period */
    float post_warmup_stop_time, /* I: Stop time for post warmup period */
    int post_warmup_start_frame, /* I: Start frame number for post warmup 
                                       period */ 
    int post_warmup_stop_frame   /* I: Stop frame number for post warmup 
                                       period */ 
);

static int write_lamp_response_band_report_file
(
    FILE *report_file,           /* I: Report file pointer */
    DB_CHAR_LAMP_RESPONSE_BAND_RESULTS *db_band_results,
                                 /* I: Pointer to results of the 
                                       lamp telemetry results */
    int lamp_pair_id,            /* I: Lamp pair id */
    float lamp_on_time,          /* I: Lamp power on time */
    float post_warmup_start_time,/* I: Start time for post warmup period */
    float post_warmup_stop_time, /* I: Stop time for post warmup period */
    int post_warmup_start_frame, /* I: Start frame number for post warmup 
                                       period */ 
    int post_warmup_stop_frame   /* I: Stop frame number for post warmup 
                                       period */ 
);

static int write_lamp_response_report_file
(
    FILE *report_file,          /* I: Report file pointer */
    DB_CHAR_LAMP_RESPONSE_RESULTS *db_response_results
                                /* I: Pointer to the lamp response 
                                      structure results */
);

static int write_lamp_telemetry_report_file
(
    FILE *report_file,           /* I: Report file pointer */
    DB_CHAR_LAMP_TELEMETRY_RESULTS *db_telemetry_results,
                                 /* I: Pointer to results of the 
                                       lamp telemetry results */
    int lamp_pair_id,            /* I: Lamp pair id */
    float lamp_on_time,          /* I: Lamp power on time */
    float post_warmup_start_time,/* I: Start time for post warmup period */
    float post_warmup_stop_time, /* I: Stop time for post warmup period */
    int post_warmup_start_frame, /* I: Start frame number for post warmup 
                                       period */ 
    int post_warmup_stop_frame   /* I: Stop frame number for post warmup 
                                       period */ 
);

/* Detector response characterization for lamp */
int char_lamp_detector_response
(
    int band_number,             /* I: Band number, 1-based */
    int sca_number,              /* I: SCA number, 1-based */
    int num_detectors,           /* I: Number of detectors in the SCA */
    int num_lines,               /* I: Number of lines in the data */
    int lamp_pair_id,            /* I: Lamp pair ID */
    float lamp_on_time,          /* I: Lamp power on time */
    float post_warmup_start_time,/* I: Start time for post warmup period */
    float post_warmup_stop_time, /* I: Stop time for post warmup period */
    int post_warmup_start_frame, /* I: Start frame number for post warmup 
                                       period */ 
    int post_warmup_stop_frame,  /* I: Stop frame number for post warmup 
                                       period */ 
    IAS_PIXEL_MASK *pixel_mask,  /* I: Contains inop detector and other
                                       ignore pixel locations */
    IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *statistics_results,
                                 /* I: Statistics results from the Histogram
                                       Statistics calculation */  
    IAS_REPORT *report_info,     /* I: General information for the test
                                       report.  If no report is desired pass
                                       NULL. */
    DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS *db_detector_results,
                                 /* I/O: Pointer to results of the 
                                       lamp response detector results */
    DB_CHAR_LAMP_RESPONSE_ONE_SCA_RESULTS *db_one_sca_results
                                 /* O: Pointer to results of the 
                                       lamp response one SCA results */
)
{
    int det;                     /* Loop counter for the detectors */
    int det_count;               /* counter used for statistics calculation */

    PIXEL_MASK_TYPE *temp_mask;  /* Line of pixel masks from image */
    int pan_band;                /* Flag to indicate this is pan band */
    const IAS_BAND_ATTRIBUTES *band_attributes;
    FILE *detector_report_file = NULL;    /* File handle for the report */
    int generate_report = 1;
    int status;

    /* Set the flags false if NULL was passed in the report structures. */
    if (report_info == NULL)
        generate_report = 0;

    /* Determine if we're operating on the pan band */
    pan_band = 0;
    band_attributes = ias_sat_attr_get_band_attributes(band_number);
    if (!band_attributes)
    {
        IAS_LOG_ERROR("Getting band attributes");
        return ERROR;
    }
    if (band_attributes->band_type == IAS_PAN_BAND)
        pan_band = 1;

    /* Allocate the temporary lines of pixel masks */
    temp_mask = malloc(sizeof(PIXEL_MASK_TYPE) * num_lines);
    if (!temp_mask)
    {
        IAS_LOG_ERROR("Allocating pixel mask memory");
        return ERROR;
    }

    /* Initialize the db_one_sca_results mean and stdev values */
    db_one_sca_results->db_mean = 0;
    db_one_sca_results->db_stdev = 0;
    db_one_sca_results->db_corrected = 0;
    det_count = 0;

    /* Loop through each detector of the input data */
    for (det = 0; det < num_detectors; det++)
    {
        /* Grab the pixel mask information for all frames of data for
           the current detector */
        status = ias_pm_get_image_subset(pixel_mask, det, 0, 1,
             num_lines, temp_mask);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Retrieving pixel mask subset for detector %d",
                          (det + 1));
            free(temp_mask);
            return ERROR;
        }

        /* At this point the iterator is pointing to the first span of 
           the next detector or the span is NULL (in the case of a single
           detector being passed in).  That is key for the loop. */

        /* Now loop through the first line of the input image, set the 
           statistics values be zero if the first line is flagged as 
           inoperable 
           Note: db_corrected is set to be the same as db_mean for now 
                 in case the correction value will be available and 
                 needed to be added in later */
        if (temp_mask[0] & IAS_RPS_INOP_DETECTOR)
        {
            db_detector_results->detector[det] = det + 1;
            db_detector_results->db_mean[det] = 0.0;
            db_detector_results->db_stdev[det] = 0.0;
            db_detector_results->db_skewness[det] = 0.0;
            db_detector_results->db_kurtosis[det] = 0.0;
            db_detector_results->db_corrected[det] = 
                    db_detector_results->db_mean[det] * 1.0;
        }
        else
        {
            /* Assign detector number for database outputs and populate the
               output detector level db structure */
            /* If pan band, then odd and even line statistics need to be 
               averaged */
            if (pan_band)
            {
                db_detector_results->detector[det] = det + 1;
                db_detector_results->db_mean[det] = 
                     (statistics_results[det].mean[0] + 
                      statistics_results[det].mean[1])/2;
                db_detector_results->db_stdev[det] = 
                     (statistics_results[det].standard_deviation[0] +
                      statistics_results[det].standard_deviation[1])/2;
                db_detector_results->db_skewness[det] = 
                     (statistics_results[det].skewness[0] +
                      statistics_results[det].skewness[1])/2;
                db_detector_results->db_kurtosis[det] = 
                     (statistics_results[det].kurtosis[0] +
                      statistics_results[det].kurtosis[1])/2;
            }
            else
            {
                db_detector_results->detector[det] = det + 1;
                db_detector_results->db_mean[det] = 
                        statistics_results[det].mean[0];
                db_detector_results->db_stdev[det] = 
                        statistics_results[det].standard_deviation[0];
                db_detector_results->db_skewness[det] = 
                        statistics_results[det].skewness[0];
                db_detector_results->db_kurtosis[det] = 
                        statistics_results[det].kurtosis[0];
            }

            /* Do lamp effective radiance correction if needed */
            db_detector_results->db_corrected[det] 
                = db_detector_results->db_mean[det] * 1.0;

            /* Calculate the statistics for the db_one_sca_results structure, 
               only counts the det_mean which has non-zero value, or excludes
               the dead detector */
            db_one_sca_results->db_mean 
                      += db_detector_results->db_mean[det];
            db_one_sca_results->db_stdev 
                      += db_detector_results->db_stdev[det];
            db_one_sca_results->db_corrected += 
                      db_detector_results->db_corrected[det];
            det_count ++;
        }
    }  /* END for det loop */

    /* Free the temp_mask */
    free(temp_mask);

    /* Calculate the statistics for the db_sca_results structure */
    db_one_sca_results->db_mean = db_one_sca_results->db_mean / det_count;
    db_one_sca_results->db_stdev = db_one_sca_results->db_stdev / det_count;
    db_one_sca_results->db_corrected = db_one_sca_results->db_corrected 
                                       / det_count;        

    /* Output the report file */
    if (generate_report)
    {
        /* Create detector report file */
        detector_report_file = create_detector_report_file(band_number, 
                               sca_number, 
                               db_detector_results->processing_level, 
                               report_info);
        if (!detector_report_file)
        {
            IAS_LOG_ERROR("Creating detector report file");
            return ERROR;
        }

        /* Write detector report file */
        if (write_lamp_response_detector_report_file(detector_report_file, 
            db_detector_results, lamp_pair_id, lamp_on_time, 
            post_warmup_start_time, post_warmup_stop_time, 
            post_warmup_start_frame, post_warmup_stop_frame) != SUCCESS)
        {
            IAS_LOG_ERROR("Writing detector report file");
            return ERROR;
        }
            
        /* Close detector report file */
        fclose(detector_report_file);
    }

    return SUCCESS;
}  

/*----------------------------------------------------------------------------
NAME: rps_output_lamp_sca_results 

PURPOSE: Write out SCA level lamp statistics results 

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int rps_output_lamp_sca_results
(
    int band_number,             /* I: 1-based band number */
    DB_CHAR_LAMP_RESPONSE_SCA_RESULTS *db_sca_results,
                                 /* I: Pointer to lamp SCA results */
    IAS_REPORT *report_info,     /* I: General information for the test
                                       report.  If no report is desired pass
                                       NULL. */
    int lamp_pair_id,            /* I: Lamp pair id */
    float lamp_on_time,          /* I: Lamp power on time */
    float post_warmup_start_time,/* I: Start time for post warmup period */
    float post_warmup_stop_time, /* I: Stop time for post warmup period */
    int post_warmup_start_frame, /* I: Start frame number for post warmup 
                                       period */ 
    int post_warmup_stop_frame   /* I: Stop frame number for post warmup 
                                       period */ 
)
{
    FILE *sca_report_file = NULL;

    /* Create SCA report file */
    sca_report_file = create_sca_report_file(band_number, 
                                             db_sca_results->processing_level, 
                                             report_info);
    if (!sca_report_file)
    {
        IAS_LOG_ERROR("Creating SCA report file");
        return ERROR;
    }

    /* Write SCA report file */
    if (write_lamp_response_sca_report_file(sca_report_file, 
        db_sca_results, lamp_pair_id, lamp_on_time, post_warmup_start_time, 
        post_warmup_stop_time, post_warmup_start_frame, post_warmup_stop_frame) 
        != SUCCESS)
    {
        IAS_LOG_ERROR("Writing SCA report file");
        return ERROR;
    }
            
    /* Close SCA report file */
    fclose(sca_report_file); 

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME: rps_output_lamp_band_results 

PURPOSE: Write out band level lamp statistics results 

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int rps_output_lamp_band_results
(
    DB_CHAR_LAMP_RESPONSE_BAND_RESULTS *db_band_results,
                                 /* I: Pointer to lamp band results */
    IAS_REPORT *report_info,     /* I: General information for the test
                                        report.  If no report is desired pass
                                        NULL. */
    int lamp_pair_id,            /* I: Lamp pair id */
    float lamp_on_time,          /* I: Lamp power on time */
    float post_warmup_start_time,/* I: Start time for post warmup period */
    float post_warmup_stop_time, /* I: Stop time for post warmup period */
    int post_warmup_start_frame, /* I: Start frame number for post warmup 
                                       period */ 
    int post_warmup_stop_frame   /* I: Stop frame number for post warmup 
                                       period */ 
)
{
    FILE *band_report_file = NULL;

    /* Create band report file */
    band_report_file = create_band_report_file(
                       db_band_results->processing_level, report_info);
    if (!band_report_file)
    {
        IAS_LOG_ERROR("Creating band report file");
        return ERROR;
    }

    /* Write band report file */
    if (write_lamp_response_band_report_file(band_report_file, 
        db_band_results, lamp_pair_id, lamp_on_time, post_warmup_start_time, 
        post_warmup_stop_time, post_warmup_start_frame, post_warmup_stop_frame) 
        != SUCCESS)
    {
        IAS_LOG_ERROR("Writing band report file");
        return ERROR;
    }
            
    /* Close band report file */
    fclose(band_report_file);

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME: rps_output_lamp_response_results 

PURPOSE: Write out lamp response data statistics results 

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int rps_output_lamp_response_results
(
    DB_CHAR_LAMP_RESPONSE_RESULTS *db_response_results,
                                /* I: Pointer to lamp response results */
    IAS_REPORT *report_info     /* I: General information for the test
                                       report.  If no report is desired pass
                                       NULL. */
)
{
    FILE *response_report_file = NULL;

    /* Create lamp response data report file */
    response_report_file = create_report_file(report_info,
                                              REPORT_FILE_RESPONSE);
    if (!response_report_file)
    {
        IAS_LOG_ERROR("Creating lamp response report file");
        return ERROR;
    }

    /* Write lamp response report file */
    if (write_lamp_response_report_file(response_report_file, 
                    db_response_results) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing lamp response report file");
        return ERROR;
    }
            
    /* Close lamp report file */
    fclose(response_report_file);

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME: rps_output_lamp_telemetry_results 

PURPOSE: Write out lamp telemetry data statistics results 

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int rps_output_lamp_telemetry_results
(
    DB_CHAR_LAMP_TELEMETRY_RESULTS *db_telemetry_results,
                                /* I: Pointer to lamp telemetry results */
    IAS_REPORT *report_info,    /* I: General information for the test
                                       report.  If no report is desired pass
                                       NULL. */
    int lamp_pair_id,            /* I: Lamp pair id */
    float lamp_on_time,          /* I: Lamp power on time */
    float post_warmup_start_time,/* I: Start time for post warmup period */
    float post_warmup_stop_time, /* I: Stop time for post warmup period */
    int post_warmup_start_frame, /* I: Start frame number for post warmup 
                                       period */ 
    int post_warmup_stop_frame   /* I: Stop frame number for post warmup 
                                       period */ 
)
{
    FILE *telemetry_report_file = NULL;

    /* Create lamp telemetry report file */
    telemetry_report_file = create_report_file(report_info,
                                               REPORT_FILE_TELEMETRY);
    if (!telemetry_report_file)
    {
        IAS_LOG_ERROR("Creating lamp telemetry report file");
        return ERROR;
    }

    /* Write lamp telemetry report file */
    if (write_lamp_telemetry_report_file(telemetry_report_file, 
        db_telemetry_results, lamp_pair_id, lamp_on_time, 
        post_warmup_start_time, post_warmup_stop_time, post_warmup_start_frame, 
        post_warmup_stop_frame) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing lamp telemetry report file");
        return ERROR;
    }
            
    /* Close lamp telemetry report file */
    fclose(telemetry_report_file);

    return SUCCESS;
}

/*-------------------------------------------------------------------------
   Database Section
-------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------
NAME:  rps_db_allocate_lamp_detector_results  

PURPOSE: Allocate memory for lamp detector db structure

RETURNS: pointer to the DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS structure
-----------------------------------------------------------------------------*/
DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS 
                 *rps_db_allocate_lamp_detector_results
(
    int number_of_detectors /* I: Number of detectors to allocate */
)
{
    DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS *db_detector_results = NULL;

    db_detector_results = malloc(
        sizeof(DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS));
    if ( !db_detector_results )
    {
        rps_db_free_lamp_detector_results(db_detector_results);
        IAS_LOG_ERROR("Allocating memory for DB results structure");
        return NULL;
    }

    db_detector_results->detector = malloc(number_of_detectors * sizeof (int));
    db_detector_results->db_mean = 
                      malloc(number_of_detectors * sizeof (float));
    db_detector_results->db_stdev = 
                      malloc(number_of_detectors * sizeof (float));
    db_detector_results->db_skewness = 
                      malloc(number_of_detectors * sizeof (float));
    db_detector_results->db_kurtosis = 
                      malloc(number_of_detectors * sizeof (float));
    db_detector_results->db_corrected = 
                      malloc(number_of_detectors * sizeof (float));
    if (!db_detector_results->detector || !db_detector_results->db_mean
        || !db_detector_results->db_stdev || !db_detector_results->db_skewness 
        || !db_detector_results->db_kurtosis 
        || !db_detector_results->db_corrected)
    {
        rps_db_free_lamp_detector_results(db_detector_results);
        IAS_LOG_ERROR("Allocating memory for DB results elements");
        return NULL;
    }

    db_detector_results->processing_level = NULL;
 
    return db_detector_results;
}

/*----------------------------------------------------------------------------
NAME:  rps_store_lamp_detector_in_database

PURPOSE: Storing the lamp detector level statistics into db table, 
         the message needs to be memory allocated first by calling 
         ias_rps_db_allocate_lamp_detector_results routine

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int rps_store_lamp_detector_in_database
(
    void *message /* I: Pointer to DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS
                        structure.  A void pointer is used since it is
                        called from the threading routines and a void
                        pointer is used to support passing messages to
                        different routines */
)
{
    DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS *db_detector_results =
        (DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS *)message;
    int status = SUCCESS;    /* function return status */
    int number_db_columns;   /* Number of database columns being inserted */
    
    /* Create the lamp_response_detector table */
    IAS_DB_INSERT_TABLE table[]=
    {
        IAS_DB_TABLE_STRING("radiometric_wo_common_id",
                db_detector_results->rad_common_id),
        IAS_DB_TABLE_INT("band_id", db_detector_results->band),
        IAS_DB_TABLE_INT("sensor_chip_assembly_id", 
                           db_detector_results->sca),
        IAS_DB_TABLE_INT_ARRAY("detector_id",
                db_detector_results->detector),
        IAS_DB_TABLE_FLOAT_ARRAY("mean", 
                           db_detector_results->db_mean),
        IAS_DB_TABLE_FLOAT_ARRAY("stdev", 
                           db_detector_results->db_stdev),
        IAS_DB_TABLE_FLOAT_ARRAY("skewness", 
                           db_detector_results->db_skewness),
        IAS_DB_TABLE_FLOAT_ARRAY("kurtosis", 
                           db_detector_results->db_kurtosis),
        IAS_DB_TABLE_FLOAT_ARRAY("corrected", 
                           db_detector_results->db_corrected),
        IAS_DB_TABLE_STRING("processing_level", 
                           db_detector_results->processing_level)
    };

    /* Get the number of columns inserted into the table and insert the
       data into the database */
    number_db_columns = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);
    status = ias_db_insert_using_table(db_detector_results->dbase,
        "LAMP_RESPONSE_DETECTOR", table, number_db_columns,
        db_detector_results->num_detectors);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Inserting data into lamp_response_detector table "
            "for band %d, sca %d", db_detector_results->band, 
            db_detector_results->sca);
    }

    /* Free the db_detector_results */
    rps_db_free_lamp_detector_results(db_detector_results);

    return status;
}

/*----------------------------------------------------------------------------
NAME: rps_db_free_lamp_detector_results 

PURPOSE: Free the allocated lamp detector db structure

RETURNS: None
-----------------------------------------------------------------------------*/
void rps_db_free_lamp_detector_results
(
    DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS *db_detector_results 
                                               /* I: structure to free */
)
{
    /* Check if memory is allocated */
    if (!db_detector_results)
    {
        return;
    }

    free(db_detector_results->detector);
    free(db_detector_results->db_mean);
    free(db_detector_results->db_stdev);
    free(db_detector_results->db_skewness);
    free(db_detector_results->db_kurtosis);
    free(db_detector_results->db_corrected);
    free(db_detector_results->processing_level);
    free(db_detector_results); 
}

/*----------------------------------------------------------------------------
NAME:  rps_db_allocate_lamp_sca_results  

PURPOSE: Allocate memory for lamp SCA db structure

RETURNS: pointer to the DB_CHAR_LAMP_RESPONSE_SCA_RESULTS structure
-----------------------------------------------------------------------------*/
DB_CHAR_LAMP_RESPONSE_SCA_RESULTS 
                 *rps_db_allocate_lamp_sca_results
(
    int number_of_scas /* I: Number of SCAs to allocate */
)
{
    DB_CHAR_LAMP_RESPONSE_SCA_RESULTS *db_sca_results = NULL;

    db_sca_results = malloc(
        sizeof(DB_CHAR_LAMP_RESPONSE_SCA_RESULTS));
    if (!db_sca_results)
    {
        rps_db_free_lamp_sca_results(db_sca_results);
        IAS_LOG_ERROR("Allocating memory for DB results structure");
        return NULL;
    }

    db_sca_results->sca = malloc(number_of_scas * sizeof (int));
    db_sca_results->db_mean = 
                      malloc(number_of_scas * sizeof (float));
    db_sca_results->db_stdev = 
                      malloc(number_of_scas * sizeof (float));
    db_sca_results->db_corrected = 
                      malloc(number_of_scas * sizeof (float));
    if (!db_sca_results->sca || !db_sca_results->db_mean
        || !db_sca_results->db_stdev || !db_sca_results->db_corrected)
    {
        rps_db_free_lamp_sca_results(db_sca_results);
        IAS_LOG_ERROR("Allocating memory for DB results elements");
        return NULL;
    }
    db_sca_results->processing_level = NULL;

    return db_sca_results;
}

/*----------------------------------------------------------------------------
NAME:  rps_store_lamp_sca_in_database

PURPOSE: Storing the lamp SCA level statistics into db table, 
         the message needs to be memory allocated first by calling 
         ias_rps_db_allocate_lamp_sca_results routine

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int rps_store_lamp_sca_in_database
(
    DB_CHAR_LAMP_RESPONSE_SCA_RESULTS *db_sca_results 
                  /* I: Pointer to DB_CHAR_LAMP_RESPONSE_SCA_RESULTS
                        structure.  A void pointer is used since it is
                        called from the threading routines and a void
                        pointer is used to support passing messages to
                        different routines */
)
{
    int status = SUCCESS;    /* function return status */
    int number_db_columns;   /* Number of database columns being inserted */
    
    /* Create the lamp_response_sca table */
    IAS_DB_INSERT_TABLE table[]=
    {
        IAS_DB_TABLE_STRING("radiometric_wo_common_id",
                           db_sca_results->rad_common_id),
        IAS_DB_TABLE_INT("band_id", db_sca_results->band),
        IAS_DB_TABLE_INT_ARRAY("sensor_chip_assembly_id", 
                           db_sca_results->sca),
        IAS_DB_TABLE_FLOAT_ARRAY("mean", 
                           db_sca_results->db_mean),
        IAS_DB_TABLE_FLOAT_ARRAY("stdev", 
                           db_sca_results->db_stdev),
        IAS_DB_TABLE_FLOAT_ARRAY("corrected", 
                           db_sca_results->db_corrected),
        IAS_DB_TABLE_STRING("processing_level", 
                           db_sca_results->processing_level)
    };

    /* Get the number of columns inserted into the table and insert the
       data into the database */
    number_db_columns = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);

    status = ias_db_insert_using_table(db_sca_results->dbase,
        "LAMP_RESPONSE_SCA", table, number_db_columns, 
        db_sca_results->num_scas);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Inserting data into lamp_response_sca table for band %d",
             db_sca_results->band);
    }

    return status;
}

/*----------------------------------------------------------------------------
NAME: rps_db_free_lamp_sca_results 

PURPOSE: Free the allocated lamp SCA db structure

RETURNS: None
-----------------------------------------------------------------------------*/
void rps_db_free_lamp_sca_results
(
    DB_CHAR_LAMP_RESPONSE_SCA_RESULTS *db_sca_results 
                                               /* I: structure to free */
)
{
    free(db_sca_results->sca);
    free(db_sca_results->db_mean);
    free(db_sca_results->db_stdev);
    free(db_sca_results->db_corrected);
    free(db_sca_results->processing_level);
    free(db_sca_results);
}

/*----------------------------------------------------------------------------
NAME:  rps_db_allocate_lamp_band_results  

PURPOSE: Allocate memory for lamp band db structure

RETURNS: pointer to the DB_CHAR_LAMP_RESPONSE_BAND_RESULTS structure
-----------------------------------------------------------------------------*/
DB_CHAR_LAMP_RESPONSE_BAND_RESULTS 
                 *rps_db_allocate_lamp_band_results
(
    int number_of_bands /* I: Number of bands to allocate */
)
{
    DB_CHAR_LAMP_RESPONSE_BAND_RESULTS *db_band_results = NULL;

    db_band_results = malloc(
        sizeof(DB_CHAR_LAMP_RESPONSE_BAND_RESULTS));
    if ( !db_band_results )
    {
        rps_db_free_lamp_band_results(db_band_results);
        IAS_LOG_ERROR("Allocating memory for DB results structure");
        return NULL;
    }

    db_band_results->band = malloc(number_of_bands * sizeof (int));
    db_band_results->db_mean = 
                      malloc(number_of_bands * sizeof (float));
    db_band_results->db_stdev = 
                      malloc(number_of_bands * sizeof (float));
    db_band_results->db_corrected = 
                      malloc(number_of_bands * sizeof (float));
    if (!db_band_results->band || !db_band_results->db_mean
        || !db_band_results->db_stdev || !db_band_results->db_corrected)
    {
        rps_db_free_lamp_band_results(db_band_results);
        IAS_LOG_ERROR("Allocating memory for DB results elements");
        return NULL;
    }
    db_band_results->processing_level = NULL;

    return db_band_results;
}

/*----------------------------------------------------------------------------
NAME:  rps_store_lamp_band_in_database

PURPOSE: Storing the lamp band level statistics into db table, 
         the message needs to be memory allocated first by calling 
         ias_rps_db_allocate_lamp_band_results routine

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int rps_store_lamp_band_in_database
(
    DB_CHAR_LAMP_RESPONSE_BAND_RESULTS *db_band_results
                  /* I: Pointer to DB_CHAR_LAMP_RESPONSE_BAND_RESULTS
                        structure.  A void pointer is used since it is
                        called from the threading routines and a void
                        pointer is used to support passing messages to
                        different routines */
)
{
    int status = SUCCESS;    /* function return status */
    int number_db_columns;   /* Number of database columns being inserted */
    
    /* Create the lamp_response_band table */
    IAS_DB_INSERT_TABLE table[]=
    {
        IAS_DB_TABLE_STRING("radiometric_wo_common_id",
                           db_band_results->rad_common_id),
        IAS_DB_TABLE_INT_ARRAY("band_id", db_band_results->band),
        IAS_DB_TABLE_FLOAT_ARRAY("mean", 
                           db_band_results->db_mean),
        IAS_DB_TABLE_FLOAT_ARRAY("stdev", 
                           db_band_results->db_stdev),
        IAS_DB_TABLE_FLOAT_ARRAY("corrected", 
                           db_band_results->db_corrected),
        IAS_DB_TABLE_STRING("processing_level", 
                           db_band_results->processing_level)
    };

    /* Get the number of columns inserted into the table and insert the
       data into the database */
    number_db_columns = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);

    status = ias_db_insert_using_table(db_band_results->dbase,
        "LAMP_RESPONSE_BAND", table, number_db_columns, 
        db_band_results->num_bands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Inserting data into lamp_response_band table");
    }

    return status;
}

/*----------------------------------------------------------------------------
NAME: rps_db_free_lamp_band_results 

PURPOSE: Free the allocated lamp band db structure

RETURNS: None
-----------------------------------------------------------------------------*/
void rps_db_free_lamp_band_results
(
    DB_CHAR_LAMP_RESPONSE_BAND_RESULTS *db_band_results 
                                               /* I: structure to free */
)
{
    free(db_band_results->band);
    free(db_band_results->db_mean);
    free(db_band_results->db_stdev);
    free(db_band_results->db_corrected);
    free(db_band_results->processing_level);
    free(db_band_results);
}

/*----------------------------------------------------------------------------
NAME:  rps_store_lamp_response_in_database

PURPOSE: Storing the lamp response into db table

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int rps_store_lamp_response_in_database
(
    DB_CHAR_LAMP_RESPONSE_RESULTS *db_response_results
                  /* I: Pointer to DB_CHAR_LAMP_RESPONSE_RESULTS
                        structure.  A void pointer is used since it is
                        called from the threading routines and a void
                        pointer is used to support passing messages to
                        different routines */
)
{
    int status = SUCCESS;    /* function return status */
    int number_db_columns;   /* Number of database columns being inserted */
    int number_records = 1;    

    /* Create the lamp_response table */
    IAS_DB_INSERT_TABLE table[]=
    {
        IAS_DB_TABLE_STRING("radiometric_wo_common_id",
                          db_response_results->rad_common_id),
        IAS_DB_TABLE_INT("lamp_pair_id", db_response_results->lamp_pair_id),
        IAS_DB_TABLE_FLOAT("lamp_on_time", 
                          db_response_results->lamp_on_time),
        IAS_DB_TABLE_FLOAT("interval_start_time", 
                          db_response_results->interval_start_time),
        IAS_DB_TABLE_FLOAT("interval_stop_time", 
                          db_response_results->interval_stop_time),
        IAS_DB_TABLE_INT("interval_start_line", 
                          db_response_results->interval_start_line),
        IAS_DB_TABLE_INT("interval_stop_line", 
                          db_response_results->interval_stop_line),
        IAS_DB_TABLE_FLOAT("lamp_warmup_time", 
                          db_response_results->lamp_warmup_time),
        IAS_DB_TABLE_INT("lamp_warmup_start_line", 
                          db_response_results->lamp_warmup_start_line),
        IAS_DB_TABLE_INT("lamp_warmup_stop_line", 
                          db_response_results->lamp_warmup_stop_line)
    };

    /* Get the number of columns inserted into the table and insert the
       data into the database */
    number_db_columns = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);

    status = ias_db_insert_using_table( db_response_results->dbase,
                 "LAMP_RESPONSE", table, number_db_columns, number_records );
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Inserting data into lamp_response table " );
    }

    return status;
}

/*----------------------------------------------------------------------------
NAME:  rps_store_lamp_telemetry_in_database

PURPOSE: Storing the lamp telemetry into db table

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int rps_store_lamp_telemetry_in_database
(
    DB_CHAR_LAMP_TELEMETRY_RESULTS *db_telemetry_results
                  /* I: Pointer to DB_CHAR_LAMP_TELEMETRY_RESULTS
                        structure.  A void pointer is used since it is
                        called from the threading routines and a void
                        pointer is used to support passing messages to
                        different routines */
)
{
    int status = SUCCESS;    /* function return status */
    int number_db_columns;   /* Number of database columns being inserted */
    int number_records = 1;
    
    /* Create the lamp_response_sca table */
    IAS_DB_INSERT_TABLE table[]=
    {
        IAS_DB_TABLE_STRING("radiometric_wo_common_id",
                            db_telemetry_results->rad_common_id),
        IAS_DB_TABLE_FLOAT("lamp_current_a_mean", 
                            db_telemetry_results->lamp_current_a_mean),
        IAS_DB_TABLE_FLOAT("lamp_current_a_stdev", 
                            db_telemetry_results->lamp_current_a_stdev),
        IAS_DB_TABLE_FLOAT("lamp_current_b_mean", 
                            db_telemetry_results->lamp_current_b_mean),
        IAS_DB_TABLE_FLOAT("lamp_current_b_stdev", 
                            db_telemetry_results->lamp_current_b_stdev),
        IAS_DB_TABLE_FLOAT("lamp_voltage_a_mean", 
                            db_telemetry_results->lamp_voltage_a_mean),
        IAS_DB_TABLE_FLOAT("lamp_voltage_a_stdev", 
                            db_telemetry_results->lamp_voltage_a_stdev),
        IAS_DB_TABLE_FLOAT("lamp_voltage_b_mean", 
                            db_telemetry_results->lamp_voltage_b_mean),
        IAS_DB_TABLE_FLOAT("lamp_voltage_b_stdev", 
                            db_telemetry_results->lamp_voltage_b_stdev),
        IAS_DB_TABLE_FLOAT("photodiode_current_a_mean", 
                            db_telemetry_results->photodiode_micro_amp_a_mean),
        IAS_DB_TABLE_FLOAT("photodiode_current_a_stdev", 
                            db_telemetry_results->photodiode_micro_amp_a_stdev),
        IAS_DB_TABLE_FLOAT("photodiode_current_b_mean", 
                            db_telemetry_results->photodiode_micro_amp_b_mean),
        IAS_DB_TABLE_FLOAT("photodiode_current_b_stdev", 
                            db_telemetry_results->photodiode_micro_amp_b_stdev),
        IAS_DB_TABLE_FLOAT("lamp_housing_temp_a_mean", 
                            db_telemetry_results->lamp_housing_temp_a_mean),
        IAS_DB_TABLE_FLOAT("lamp_housing_temp_a_stdev", 
                            db_telemetry_results->lamp_housing_temp_a_stdev),
        IAS_DB_TABLE_FLOAT("lamp_housing_temp_b_mean", 
                            db_telemetry_results->lamp_housing_temp_b_mean),
        IAS_DB_TABLE_FLOAT("lamp_housing_temp_b_stdev", 
                            db_telemetry_results->lamp_housing_temp_b_stdev),
        IAS_DB_TABLE_FLOAT("monitor_diode_temp_a_mean", 
                            db_telemetry_results->monitor_diode_temp_a_mean),
        IAS_DB_TABLE_FLOAT("monitor_diode_temp_a_stdev", 
                            db_telemetry_results->monitor_diode_temp_a_stdev),
        IAS_DB_TABLE_FLOAT("monitor_diode_temp_b_mean", 
                            db_telemetry_results->monitor_diode_temp_b_mean),
        IAS_DB_TABLE_FLOAT("monitor_diode_temp_b_stdev", 
                            db_telemetry_results->monitor_diode_temp_b_stdev),
        IAS_DB_TABLE_FLOAT("fp_temp_7_mean", 
                            db_telemetry_results->fp_temp_7_mean),
        IAS_DB_TABLE_FLOAT("fp_temp_7_stdev", 
                            db_telemetry_results->fp_temp_7_stdev),
        IAS_DB_TABLE_FLOAT("fp_temp_14_mean", 
                            db_telemetry_results->fp_temp_14_mean),
        IAS_DB_TABLE_FLOAT("fp_temp_14_stdev", 
                            db_telemetry_results->fp_temp_14_stdev)
    };

    /* Get the number of columns inserted into the table and insert the
       data into the database */
    number_db_columns = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);

    status = ias_db_insert_using_table(db_telemetry_results->dbase,
        "LAMP_TELEMETRY", table, number_db_columns, number_records);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Inserting data into lamp_telemetry table " );
    }

    return status;
}

/*----------------------------------------------------------------------------
NAME: create_detector_report_file   

PURPOSE: Open the detector level report file 

RETURNS: File pointer
-----------------------------------------------------------------------------*/
/* Open the report file and allocate temporary storage variables */
static FILE *create_detector_report_file
(
    int band_number,             /* I: 1-based band number */
    int sca_number,              /* I: 1-based SCA number */
    const char *processing_level,/* I: processing label */
    IAS_REPORT *report_info      /* I: IAS report structure pointer */
)
{
    FILE *report_file = NULL;
    char file_base_base_name[] = "char_lamp_detector";
    char *file_name = NULL;
    IAS_REPORT_HEADER *copy_of_report_header;
    char *file_base_name;
    int length;

    /* Build the base report file name.  Add one to the length for the NULL
       terminator and one for "_" between file_base_base_name and 
       processing_level */
    length = strlen(file_base_base_name) + strlen(processing_level) + 2;
    file_base_name = malloc(sizeof(char) * length);
    if (!file_base_name)
    {
        IAS_LOG_ERROR("Allocating memory for file_base_name");
        return NULL;        
    }
    else
    {
        sprintf(file_base_name, "%s_%s", file_base_base_name,
                     processing_level);
    }

    /* Get output report file name */
    file_name = ias_rps_get_report_file_name(report_info, band_number, 
                sca_number, file_base_name);
    if (!file_name)
    {
        IAS_LOG_ERROR("Getting report filename %s", file_name);
        free(file_base_name);
        return NULL;        
    }

    /* Free the file_base_name memory */
    free(file_base_name);
    file_base_name = NULL;

    /* Check the report file, if it hasn't opened, then open it,
       otherwise, appending it */
    report_file = fopen(file_name, "w");
    if (!report_file)
    {
        IAS_LOG_ERROR("Opening report file %s", file_name);
        free(file_name);
        return NULL;
    }

    /* Get the local copy of the report header since the "standard" header
       includes non-standard things such as band, SCA and algorithm name.
       In a threaded environment the input header cannot be manipulated. */
    copy_of_report_header = ias_misc_copy_report_header(report_info->header);
    if ( !copy_of_report_header )
    {
        IAS_LOG_ERROR("Creating local copy of report header");
        fclose(report_file);
        free(file_name);
        return NULL;
    }

    /* Set the band number and SCA number.  Also, set the algorithm name in
       the report header */
    ias_misc_set_header_band_number(copy_of_report_header, band_number);
    ias_misc_set_header_sca_number(copy_of_report_header, sca_number);
    ias_misc_set_header_algorithm_name(copy_of_report_header,
        "Detector Response Lamp Characterization");

    if (ias_misc_write_report_header(report_file, copy_of_report_header))
    {
        ias_misc_free_report_header(copy_of_report_header);
        fclose(report_file);
        free(file_name);
        return NULL;
    }

    ias_misc_free_report_header(copy_of_report_header);
    free(file_name);
    return report_file;
}

/*----------------------------------------------------------------------------
NAME: create_sca_report_file   

PURPOSE: Open the SCA level report file 

RETURNS: File pointer
-----------------------------------------------------------------------------*/
/* Open the report file and allocate temporary storage variables */
static FILE *create_sca_report_file
(
    int band_number,             /* I: 1-based band number */
    const char *processing_level,/* I: processing label */
    IAS_REPORT *report_info      /* I: IAS report structure pointer */
)
{
    FILE *report_file = NULL;
    char file_base_base_name[] = "char_lamp_sca";
    char *file_name = NULL;
    IAS_REPORT_HEADER *copy_of_report_header;
    char *file_base_name;
    int length;

    /* Build the base report file name.  Add one to the length for the NULL
       terminator and one for "_" between file_base_base_name and 
       processing_level */
    length = strlen(file_base_base_name) + strlen(processing_level) + 2;
    file_base_name = malloc(sizeof(char) * length);
    if (!file_base_name)
    {
        IAS_LOG_ERROR("Allocating memory for file_base_name");
        return NULL;        
    }
    else
    {
        sprintf(file_base_name, "%s_%s", file_base_base_name,
                     processing_level);
    }

    /* Get output report file name */
    file_name = ias_rps_get_report_file_name(report_info, band_number, 
                -1, file_base_name);
    if (!file_name)
    {
        IAS_LOG_ERROR("Getting report filename %s", file_name);
        free(file_base_name);
        return NULL;        
    }

    /* Free the file_base_name memory */
    free(file_base_name);
    file_base_name = NULL;

    /* Check the report file, if it hasn't opened, then open it,
       otherwise, appending it */
    report_file = fopen(file_name, "w");
    if (!report_file)
    {
        IAS_LOG_ERROR("Opening report file %s", file_name);
        free(file_name);
        return NULL;
    }

    /* Get the local copy of the report header since the "standard" header
       includes non-standard things such as band, SCA and algorithm name.
       In a threaded environment the input header cannot be manipulated. */
    copy_of_report_header = ias_misc_copy_report_header(report_info->header);
    if (!copy_of_report_header)
    {
        IAS_LOG_ERROR("Creating local copy of report header");
        fclose(report_file);
        free(file_name);
        return NULL;
    }

    /* Set the band number and SCA number to 0 in the report header--
       they won't be written out.  Also, set the algorithm name in
       the report header */
    ias_misc_set_header_band_number(copy_of_report_header, band_number);
    ias_misc_set_header_sca_number(copy_of_report_header, 0);
    ias_misc_set_header_algorithm_name(copy_of_report_header,
        "Detector Response Lamp Characterization");

    if (ias_misc_write_report_header(report_file, copy_of_report_header))
    {
        ias_misc_free_report_header(copy_of_report_header);
        fclose(report_file);
        free(file_name);
        return NULL;
    }

    ias_misc_free_report_header(copy_of_report_header);
    free(file_name);
    return report_file;
}

/*----------------------------------------------------------------------------
NAME: create_band_report_file   

PURPOSE: Open the band level report file 

RETURNS: File pointer
-----------------------------------------------------------------------------*/
/* Open the report file and allocate temporary storage variables */
static FILE *create_band_report_file
(
    const char *processing_level,/* I: RPS processing label */
    IAS_REPORT *report_info      /* I: IAS report structure pointer */
)
{
    FILE *report_file = NULL;
    char file_base_base_name[] = "char_lamp_band";
    char *file_name = NULL;
    IAS_REPORT_HEADER *copy_of_report_header;
    char *file_base_name;
    int length;

    /* Build the base report file name.  Add one to the length for the NULL
       terminator and one for "_" between file_base_base_name and 
       processing_level */
    length = strlen(file_base_base_name) + strlen(processing_level) + 2;
    file_base_name = malloc(sizeof(char) * length);
    if (!file_base_name)
    {
        IAS_LOG_ERROR("Allocating memory for file_base_name");
        return NULL;        
    }
    else
    {
        sprintf(file_base_name, "%s_%s", file_base_base_name,
                     processing_level);
    }

    /* Get output report file name */
    file_name = ias_rps_get_report_file_name(report_info, -1, 
                -1, file_base_name);
    if (!file_name)
    {
        IAS_LOG_ERROR("Getting report filename %s", file_name);
        free(file_base_name);
        return NULL;        
    }

    /* Free the file_base_name memory */
    free(file_base_name);
    file_base_name = NULL;

    /* Check the report file, if it hasn't opened, then open it,
       otherwise, appending it */
    report_file = fopen(file_name, "w");
    if (!report_file)
    {
        IAS_LOG_ERROR("Opening report file %s", file_name);
        free(file_name);
        return NULL;
    }

    /* Get the local copy of the report header since the "standard" header
       includes non-standard things such as band, SCA and algorithm name.
       In a threaded environment the input header cannot be manipulated. */
    copy_of_report_header = ias_misc_copy_report_header(report_info->header);
    if (!copy_of_report_header)
    {
        IAS_LOG_ERROR("Creating local copy of report header");
        fclose(report_file);
        free(file_name);
        return NULL;
    }

    /* Set the band number and SCA number to 0 in the report header--
       they won't be written out.  Also, set the algorithm name in
       the report header */
    ias_misc_set_header_band_number(copy_of_report_header, 0);
    ias_misc_set_header_sca_number(copy_of_report_header, 0);
    ias_misc_set_header_algorithm_name(copy_of_report_header,
        "Detector Response Lamp Characterization");

    if (ias_misc_write_report_header(report_file, copy_of_report_header))
    {
        ias_misc_free_report_header(copy_of_report_header);
        fclose(report_file);
        free(file_name);
        return NULL;
    }

    ias_misc_free_report_header(copy_of_report_header);
    free(file_name);
    return report_file;
}

/*----------------------------------------------------------------------------
NAME: create_report_file   

PURPOSE: Open the report file for lamp response data or lamp telemetry data 

RETURNS: File pointer
-----------------------------------------------------------------------------*/
static FILE *create_report_file
(
    IAS_REPORT *report_info,       /* I: IAS report structure pointer */
    REPORT_FILE_TYPE file_type
)
{
    FILE *report_file = NULL;
    char file_base_name[STRLEN];
    char *file_name = NULL;
    IAS_REPORT_HEADER *copy_of_report_header;

    switch(file_type)
    {
        case REPORT_FILE_RESPONSE:
            strcpy(file_base_name, "char_lamp_response");
            break;
        case REPORT_FILE_TELEMETRY:
            strcpy(file_base_name, "char_lamp_telemetry");
            break;
        default:
            IAS_LOG_ERROR("Unsupported report file type: %d", file_type);
            return NULL;
    }

    /* Get output report file name */
    file_name = ias_rps_get_report_file_name(report_info, -1, 
                                             -1, file_base_name);
    if (!file_name)
    {
        IAS_LOG_ERROR("Getting report filename %s", file_name);
        return NULL;        
    }

    /* Check the report file, if it hasn't opened, then open it,
       otherwise, appending it */
    report_file = fopen(file_name, "w");
    if (!report_file)
    {
        IAS_LOG_ERROR("Opening report file %s", file_name);
        free(file_name);
        return NULL;
    }

    /* Get the local copy of the report header since the "standard" header
       includes non-standard things such as band, SCA and algorithm name.
       In a threaded environment the input header cannot be manipulated. */
    copy_of_report_header = ias_misc_copy_report_header(report_info->header);
    if ( !copy_of_report_header )
    {
        IAS_LOG_ERROR("Creating local copy of report header");
        fclose(report_file);
        free(file_name);
        return NULL;
    }

    /* Set the band number and SCA number to 0 in the report header--
       they won't be written out.  Also, set the algorithm name in
       the report header */
    ias_misc_set_header_band_number(copy_of_report_header, 0);
    ias_misc_set_header_sca_number(copy_of_report_header, 0);
    ias_misc_set_header_algorithm_name(copy_of_report_header,
        "Detector Response Lamp Characterization");

    if (ias_misc_write_report_header(report_file, copy_of_report_header))
    {
        ias_misc_free_report_header(copy_of_report_header);
        fclose(report_file);
        free(file_name);
        return report_file;
    }

    ias_misc_free_report_header(copy_of_report_header);
    free(file_name);
    return report_file;
}

/*----------------------------------------------------------------------------
NAME: write_lamp_response_data

PURPOSE: Write the lamp response data (lamp pair id, lamp on time, post warmup
         start time, post warmup stop time, post warmup start line, post
         warmup stop line) to the report file

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int write_lamp_response_data
(
    FILE *report_file,           /* I: Report file pointer */
    int lamp_pair_id,            /* I: Lamp pair id */
    float lamp_on_time,          /* I: Lamp power on time */
    float post_warmup_start_time,/* I: Start time for post warmup period */
    float post_warmup_stop_time, /* I: Stop time for post warmup period */
    int post_warmup_start_frame, /* I: Start frame number for post warmup 
                                       period */ 
    int post_warmup_stop_frame   /* I: Stop frame number for post warmup 
                                       period */ 
)
{
    int status;
    char lamp_name[10];
   
    switch(lamp_pair_id)
    {
        case LAMP_WORKING_PAIR:
            strcpy(lamp_name, "WORKING");
            break;
        case LAMP_REFERENCE_PAIR:
            strcpy(lamp_name, "REFERENCE");
            break;
        case LAMP_PRISTINE_PAIR:
            strcpy(lamp_name, "PRISTINE");
            break;
        default:
            IAS_LOG_ERROR("Unsupported lamp pair id: %d", lamp_pair_id);
            return ERROR;
    }

    /* Print out the stim lamp pair used */ 
    status = fprintf(report_file, "STIM Lamp Pair used: %s\n", 
                     lamp_name);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp name for lamp detector data");
        return ERROR;
    }

    /* Print out the lamp power on time */ 
    status = fprintf(report_file, "Lamp power on time: %f\n", 
                     lamp_on_time);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp power on time");
        return ERROR;
    }

    /* Print out the lamp post warmup start time */ 
    status = fprintf(report_file, "Lamp post warmup start time: %f\n", 
                     post_warmup_start_time);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp post warmup start time");
        return ERROR;
    }

    /* Print out the lamp post warmup stop time */ 
    status = fprintf(report_file, "Lamp post warmup stop time: %f\n", 
                     post_warmup_stop_time);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp post warmup stop time");
        return ERROR;
    }

    /* Print out the lamp post warmup start frame number */ 
    status = fprintf(report_file, "Lamp post warmup start frame number: %d\n", 
                     post_warmup_start_frame);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp post warmup start frame number");
        return ERROR;
    }

    /* Print out the lamp post warmup stop frame number */ 
    status = fprintf(report_file, "Lamp post warmup stop frame number: %d\n\n",
                     post_warmup_stop_frame);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp post warmup stop frame number");
        return ERROR;
    }

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME: write_lamp_response_detector_report_file  

PURPOSE: Write the detector level statistics to the report file

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int write_lamp_response_detector_report_file
(
    FILE *report_file,           /* I: Report file pointer */
    DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS *db_detector_results,
                                 /* I: Pointer to results of the 
                                       lamp response detector results */
    int lamp_pair_id,            /* I: Lamp pair id */
    float lamp_on_time,          /* I: Lamp power on time */
    float post_warmup_start_time,/* I: Start time for post warmup period */
    float post_warmup_stop_time, /* I: Stop time for post warmup period */
    int post_warmup_start_frame, /* I: Start frame number for post warmup 
                                       period */ 
    int post_warmup_stop_frame   /* I: Stop frame number for post warmup 
                                       period */ 
)
{
    int status;
    int det;

    /* Write out the stim lamp response data to the report file */
    status = write_lamp_response_data(report_file, lamp_pair_id, lamp_on_time,
             post_warmup_start_time, post_warmup_stop_time, 
             post_warmup_start_frame, post_warmup_stop_frame);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing stim lamp response data");
        return ERROR;
    }

    /* Print out the detector level detector response lamp statistics */ 
    status = fprintf(report_file, "Detector level lamp statistics "
                     "for band %d and SCA %d and processing level %s\n", 
                     db_detector_results->band, db_detector_results->sca,
                     db_detector_results->processing_level);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing column title for lamp detector data");
        return ERROR;
    }

    status = fprintf(report_file, "det_number,   mean,       stdev,   "
                     " skewness,   kurtosis,   corrected \n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing column title for lamp detector data");
        return ERROR;
    }

    for (det = 0; det < db_detector_results->num_detectors; det++)
    {
        status = fprintf(report_file, "%7d, %10.3lf, %10.3lf, %10.3lf, "
                 "%10.3lf, %10.3lf\n", db_detector_results->detector[det], 
                 db_detector_results->db_mean[det], 
                 db_detector_results->db_stdev[det],
                 db_detector_results->db_skewness[det], 
                 db_detector_results->db_kurtosis[det],
                 db_detector_results->db_corrected[det]);
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing lamp detector data");
            return ERROR;
        }
    }

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME: write_lamp_response_sca_report_file  

PURPOSE: Write the SCA level statistics to the report file

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int write_lamp_response_sca_report_file
(
    FILE *report_file,           /* I: Report file pointer */
    DB_CHAR_LAMP_RESPONSE_SCA_RESULTS *db_sca_results,
                                 /* I: Pointer to results of the 
                                       lamp response SCA results */
    int lamp_pair_id,            /* I: Lamp pair id */
    float lamp_on_time,          /* I: Lamp power on time */
    float post_warmup_start_time,/* I: Start time for post warmup period */
    float post_warmup_stop_time, /* I: Stop time for post warmup period */
    int post_warmup_start_frame, /* I: Start frame number for post warmup 
                                       period */ 
    int post_warmup_stop_frame   /* I: Stop frame number for post warmup 
                                       period */ 
)
{
    int status;
    int sca;

    /* Write out the stim lamp response data to the report file */
    status = write_lamp_response_data(report_file, lamp_pair_id, lamp_on_time,
             post_warmup_start_time, post_warmup_stop_time, 
             post_warmup_start_frame, post_warmup_stop_frame);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing stim lamp response data");
        return ERROR;
    }

    /* Print out the sca level detector response lamp statistics */ 
    status = fprintf(report_file, "Processing level %s lamp statistics for "
                    "each SCA in Band %d\n ", db_sca_results->processing_level,
                    db_sca_results->band);

    status = fprintf(report_file, "SCA_number,  SCA_mean,      SCA_stdev," 
                                  " SCA_corrected \n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing column title for lamp SCA data");
        return ERROR;
    }

    for (sca = 0; sca < db_sca_results->num_scas; sca++)
    {
        status = fprintf(report_file, "     %d,     %10.3lf,   %10.3lf, "
                   "%10.3lf\n", db_sca_results->sca[sca],
                   db_sca_results->db_mean[sca], db_sca_results->db_stdev[sca],
                   db_sca_results->db_corrected[sca]);
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing lamp SCA data");
            return ERROR;
        }
    }

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME: write_lamp_response_band_report_file  

PURPOSE: Write the band level statistics to the report file

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int write_lamp_response_band_report_file
(
    FILE *report_file,           /* I: Report file pointer */
    DB_CHAR_LAMP_RESPONSE_BAND_RESULTS *db_band_results,
                                 /* I: Pointer to results of the 
                                       lamp telemetry results */
    int lamp_pair_id,            /* I: Lamp pair id */
    float lamp_on_time,          /* I: Lamp power on time */
    float post_warmup_start_time,/* I: Start time for post warmup period */
    float post_warmup_stop_time, /* I: Stop time for post warmup period */
    int post_warmup_start_frame, /* I: Start frame number for post warmup 
                                       period */ 
    int post_warmup_stop_frame   /* I: Stop frame number for post warmup 
                                       period */ 
)
{
    int band;
    int status;

    /* Write out the stim lamp response data to the report file */
    status = write_lamp_response_data(report_file, lamp_pair_id, lamp_on_time,
             post_warmup_start_time, post_warmup_stop_time, 
             post_warmup_start_frame, post_warmup_stop_frame);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing stim lamp response data");
        return ERROR;
    }

    /* Print out the detector response characterization lamp telemetry data */ 
    status = fprintf(report_file, "Processing level %s lamp statistics for "
                     "each band\n ", db_band_results->processing_level);

    if (status < 0)
    {
        IAS_LOG_ERROR("Writing processing level lamp band data");
        return ERROR;
    }

    status = fprintf(report_file, "band_number, band_mean, band_stdev, "
                                  "band_corrected \n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing column title for lamp band data");
        return ERROR;
    }

    for (band = 0; band < db_band_results->num_bands; band++)
    {
        status = fprintf(report_file, "%7d, %14.3lf, %10.3lf, %10.3lf\n",
                         db_band_results->band[band], 
                         db_band_results->db_mean[band], 
                         db_band_results->db_stdev[band],
                         db_band_results->db_corrected[band]);
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing lamp band data");
            return ERROR;
        }
    }


    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME: write_lamp_response_report_file  

PURPOSE: Write the lamp response data to the report file

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int write_lamp_response_report_file
(
    FILE *report_file,          /* I: Report file pointer */
    DB_CHAR_LAMP_RESPONSE_RESULTS *db_response_results
                                /* I: Pointer to results of the 
                                      lamp response results */
)
{
    int status;   

    /* Write out the stim lamp response data to the report file */
    status = write_lamp_response_data(report_file, 
             db_response_results->lamp_pair_id,
             db_response_results->lamp_on_time,
             db_response_results->interval_start_time, 
             db_response_results->interval_stop_time, 
             db_response_results->interval_start_line, 
             db_response_results->interval_stop_line);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing stim lamp response data");
        return ERROR;
    }

    /* Print out the detector response characterization lamp response */ 
    status = fprintf(report_file, "Detector response characterization"
                     " lamp response data\n\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp response report data");
        return ERROR;
    }

    status = fprintf(report_file, "lamp_pair_id, lamp_on_time," 
              " interval_start_time\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp response report data");
        return ERROR;
    }

    status = fprintf(report_file, "%7d, %14.3lf, %16.3lf\n",
                     db_response_results->lamp_pair_id, 
                     db_response_results->lamp_on_time,
                     db_response_results->interval_start_time);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp response report data");
        return ERROR;
    }

    status = fprintf(report_file, "interval_stop_time, interval_start_line,"
              " interval_stop_line\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp response report data");
        return ERROR;
    }

    status = fprintf(report_file, "%14.3lf, %15d, %20d\n",
                     db_response_results->interval_stop_time,
                     db_response_results->interval_start_line,
                     db_response_results->interval_stop_line);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp response report data");
        return ERROR;
    }

    status = fprintf(report_file, "lamp_warmup_time, lamp_warmup_start_line,"
              " lamp_warmup_stop_line\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp response report data");
        return ERROR;
    }

    status = fprintf(report_file, "%12.3lf, %16d, %20d\n",
                     db_response_results->lamp_warmup_time,
                     db_response_results->lamp_warmup_start_line,
                     db_response_results->lamp_warmup_stop_line);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp response report data");
        return ERROR;
    }

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME: write_lamp_telemetry_report_file  

PURPOSE: Write the lamp telemetry data to the report file

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int write_lamp_telemetry_report_file
(
    FILE *report_file,           /* I: Report file pointer */
    DB_CHAR_LAMP_TELEMETRY_RESULTS *db_telemetry_results,
                                 /* I: Pointer to results of the 
                                       lamp telemetry results */
    int lamp_pair_id,            /* I: Lamp pair id */
    float lamp_on_time,          /* I: Lamp power on time */
    float post_warmup_start_time,/* I: Start time for post warmup period */
    float post_warmup_stop_time, /* I: Stop time for post warmup period */
    int post_warmup_start_frame, /* I: Start frame number for post warmup 
                                       period */ 
    int post_warmup_stop_frame   /* I: Stop frame number for post warmup 
                                       period */ 
)
{
    int status;

    /* Write out the stim lamp response data to the report file */
    status = write_lamp_response_data(report_file, lamp_pair_id, lamp_on_time,
             post_warmup_start_time, post_warmup_stop_time, 
             post_warmup_start_frame, post_warmup_stop_frame);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing stim lamp response data");
        return ERROR;
    }

    /* Print out the detector response characterization lamp telemetry data */ 
    status = fprintf(report_file, "Detector response characterization"
                     " lamp telemetry data\n\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing column title for lamp telemetry data");
        return ERROR;
    }

    status = fprintf(report_file, " lamp_current_a_mean, lamp_current_a_stdev"
                     "\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp current A data title line1");
        return ERROR;
    }

    status = fprintf(report_file, "%16.3lf, %16.3lf\n ",
            db_telemetry_results->lamp_current_a_mean, 
            db_telemetry_results->lamp_current_a_stdev); 
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp current A data");
        return ERROR;
    }

    status = fprintf(report_file, "lamp_current_b_mean, lamp_current_b_stdev"
                     "\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp current B data title line");
        return ERROR;
    }

    status = fprintf(report_file, "%16.3lf, %16.3lf\n ",
            db_telemetry_results->lamp_current_b_mean, 
            db_telemetry_results->lamp_current_b_stdev); 
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp current B data");
        return ERROR;
    }

    status = fprintf(report_file, "lamp_voltage_a_mean, lamp_voltage_a_stdev"
                     "\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp voltage A data title line");
        return ERROR;
    }

    status = fprintf(report_file, "%16.3lf, %16.3lf\n ",
            db_telemetry_results->lamp_voltage_a_mean, 
            db_telemetry_results->lamp_voltage_a_stdev); 
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp voltage A data");
        return ERROR;
    }

    status = fprintf(report_file, "lamp_voltage_b_mean, lamp_voltage_b_stdev "
                     "\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp voltage B data title line");
        return ERROR;
    }

    status = fprintf(report_file, "%16.3lf, %16.3lf\n ",
            db_telemetry_results->lamp_voltage_b_mean, 
                     db_telemetry_results->lamp_voltage_b_stdev); 
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp vlotage B data");
        return ERROR;
    }

    status = fprintf(report_file, "photodiode_micro_amp_a_mean, "
                     "photodiode_micro_amp_a_stdev\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing photodiode current A data title line");
        return ERROR;
    }

    status = fprintf(report_file, "%16.3lf, %16.3lf\n ",
            db_telemetry_results->photodiode_micro_amp_a_mean, 
            db_telemetry_results->photodiode_micro_amp_a_stdev); 
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing photodiode current A data");
        return ERROR;
    }

    status = fprintf(report_file, "photodiode_micro_amp_b_mean, "
                                  "photodiode_micro_amp_b_stdev\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing photodiode current B data title line");
        return ERROR;
    }

    status = fprintf(report_file, "%16.3lf, %16.3lf\n ",
            db_telemetry_results->photodiode_micro_amp_b_mean, 
                     db_telemetry_results->photodiode_micro_amp_b_stdev); 
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing photodiode current B data");
        return ERROR;
    }

    status = fprintf(report_file, "lamp_housing_temp_a_mean, "
                     "lamp_housing_temp_a_stdev\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp housing temp A data title line");
        return ERROR;
    }

    status = fprintf(report_file, "%20.3lf, %20.3lf\n ",
            db_telemetry_results->lamp_housing_temp_a_mean, 
            db_telemetry_results->lamp_housing_temp_a_stdev); 
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp housing temp A data");
        return ERROR;
    }

    status = fprintf(report_file, "lamp_housing_temp_b_mean, "
                                  "lamp_housing_temp_b_stdev\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp housing temp B data title line");
        return ERROR;
    }

    status = fprintf(report_file, "%20.3lf, %20.3lf\n ",
            db_telemetry_results->lamp_housing_temp_b_mean, 
                     db_telemetry_results->lamp_housing_temp_b_stdev); 
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp housing temp B data");
        return ERROR;
    }

    status = fprintf(report_file, "monitor_diode_temp_a_mean, "
                     "monitor_diode_temp_a_stdev\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing monitor diode temp A data title line");
        return ERROR;
    }

    status = fprintf(report_file, "%20.3lf, %20.3lf\n ",
            db_telemetry_results->monitor_diode_temp_a_mean, 
            db_telemetry_results->monitor_diode_temp_a_stdev); 
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing monitor diode temp A data");
        return ERROR;
    }

    status = fprintf(report_file, "monitor_diode_temp_b_mean, "
                                  "monitor_diode_temp_b_stdev\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing monitor diode temp B data title line");
        return ERROR;
    }

    status = fprintf(report_file, "%20.3lf, %20.3lf\n ",
            db_telemetry_results->monitor_diode_temp_b_mean, 
                     db_telemetry_results->monitor_diode_temp_b_stdev); 
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing monitor diode temp B data");
        return ERROR;
    }

    status = fprintf(report_file,"fp_temp_7_mean, fp_temp_7_stdev\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp focal plane temp 7 title line");
        return ERROR;
    }

    status = fprintf(report_file, " %12.3lf, %12.3lf\n",
                     db_telemetry_results->fp_temp_7_mean, 
                     db_telemetry_results->fp_temp_7_stdev);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp lamp focal plane temp 7 data");
        return ERROR;
    }

    status = fprintf(report_file," fp_temp_14_mean, fp_temp_14_stdev\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp focal plane temp 14 title line");
        return ERROR;
    }

    status = fprintf(report_file, " %12.3lf, %12.3lf\n",
                     db_telemetry_results->fp_temp_14_mean, 
                     db_telemetry_results->fp_temp_14_stdev);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing lamp lamp focal plane temp 14 data");
        return ERROR;
    }

    return SUCCESS;
}
