/*----------------------------------------------------------------------------
NAME:    gain_application

PURPOSE: Apply the gain and temperature sensitivity corrections as needed.

RETURNS: None
-----------------------------------------------------------------------------*/
#include "ias_radiometry.h"

void gain_application
(
    float *l1r_data,             /* I/O: One SCA of l1r data to apply gain */
    int band,                    /* I: Band number, 1-based */
    int sca,                     /* I: SCA number, 1-based */
    int num_frames,              /* I: Number of frames in the data */
    int num_detectors,           /* I: Number of detectors in the SCA */
    double absolute_gain,        /* I: SCA level absolute gain from CPF */
    double *relative_gain,       /* I: detector level relative gain from CPF */
    const double *temp_correction_factors,/* I: detector level temperature
                                             correction factors */
    int perform_abs_gain,        /* I: Flag for the absolute gain correction */
    int perform_relative_gain,   /* I: Flag for the relative gain correction */
    int perform_temp_sens        /* I: Flag for the temperature sensitivity 
                                       correction */
)
{
    int frame;                   /* Loop counter for the frames */
    int detector;                /* Loop counter for the detectors */

    /* Loop through each detector of the input data */
    for (detector = 0; detector < num_detectors; detector++)
    {
        double overall_gain = 1; /* Overall detector level correction factor */
        float *detector_ptr;     /* pointer pointing to detector */

        detector_ptr = &l1r_data[detector];

        /* Calculate the overall detector level correction factor */ 
        /* Apply absolute gain if needed */
        if (perform_abs_gain)
            overall_gain *= absolute_gain;

        /* Apply relative gain if needed */
        if (perform_relative_gain)
            overall_gain *= relative_gain[detector];

        
        /* The gains are applied by division so get the inverse.  In the
           interest of efficiency no division by zero check is made since
           gains will never be zero. */
        overall_gain = 1.0 / overall_gain;

        /* Apply temperature sensitivity correction if needed */
        if (perform_temp_sens)
            overall_gain *= temp_correction_factors[detector];

        /* Now loop through the frames of the input image.   */
        for (frame = 0; frame < num_frames; frame++)
        {
            /* Update the l1r image data with gain applied */
            detector_ptr[frame * num_detectors] *= overall_gain;

        }  /* END for frame */
    }  /* END for det loop */
}  /* END ias_rad_gain_application */

