/*************************************************************************
 NAME:     inoperable_detectors_fill

 PURPOSE:  Module implementing routine(s) to perform correction of image
           data from inoperable detectors

 ROUTINES: -replace_image_data                    (internal)
           - classify_inoperable_detectors        (internal)
           -inoperable_detectors_fill
**************************************************************************/
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "ias_logging.h"
#include "ias_pixel_mask.h"
#include "rps_local.h"
#include "rps_algorithm.h"


/* Enumeration for the various detector fill cases */
typedef enum inop_detector_fill_type
{
    INOP_FILL_ZERO,
    INOP_FILL_EDGE_REPLACEMENT,       /* An inoperable detector at the left
                                         or right edge of the focal plane (i.e.
                                         detector 1 and possibly 2 on the
                                         left end, detector n and possibly
                                         n-1 at the right end) */
    INOP_FILL_NEIGHBOR_AVERAGE        /* One or two consecutive inoperable
                                         detectors between two operable
                                         neighbors  */
}   INOP_DETECTOR_FILL_TYPE;


/* Data structure used for inoperable detector status "book-keeping" */
typedef struct inop_detector_status
{
    int det_index;                      /* Inoperable det. index */
    int good_left_detector_index;       /* Index of left replacement det. */
    int good_right_detector_index;      /* Index of right replacement det. */
    INOP_DETECTOR_FILL_TYPE fill_type;  /* Type of correction the
                                           inoperable detector needs */
}   INOP_DETECTOR_STATUS;



/*-------------------------------------------------------------------------
 NAME:     classify_inoperable_detectors

 PURPOSE:  Determines the type of data replacement needed for an
           inoperable detector(s) flagged in the pixel mask.  Also
           determines the replacement data source detector(s) for each
           inoperable detector.

 RETURNS:  Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
static int classify_inoperable_detectors
(
    IAS_PIXEL_MASK *pmask,                   /* I: Pixel mask for the
                                                current band/SCA */
    int   number_detectors,                  /* I: Number of detectors
                                                for the current band/SCA */
    int   *inop_detector_count,              /* O: Count of inoperable
                                                detectors that were found */
    INOP_DETECTOR_STATUS **inop_status_info  /* I/O: Inoperable detector
                                                status information */
)
{
    IAS_PIXEL_MASK_SPAN span;

    /* Tracking pointer to inoperable detector status data block */
    INOP_DETECTOR_STATUS *det_status_ptr = NULL;

    /* Local array to record inoperable detector indices as they're found
       while checking the pixel mask */
    int inoperable_detector_array[number_detectors];

    /* Each inoperable detector is considered to be in a 'group'.  The 'group
       count' is just the number of consecutive inoperable detectors in
       that group.

       Example:

       For a focal plane array with 16 detectors, 8 of which are found to
       be inoperable (detectors 1, 2, 6, 10, 11, 14, 15, 16):
       inoperable detector group #1:  {1, 2}         => group count = 2
       inoperable detector group #2:  {6}            => group count = 1
       inoperable detector group #3   {10, 11}       => group count = 2
       inoperable detector group #4   {14, 15, 16}   => group count = 3

       The group count array in this case would be as follows:
       group_count[0] = 2;                           => detector #1
       group_count[1] = 2;                           => detector #2
       group_count[2] = 1;                           => detector #6
       group_count[3] = 2;                           => detector #10
       group_count[4] = 2;                           => detector #11
       group_count[5] = 3;                           => detector #14
       group_count[6] = 3;                           => detector #15
       group_count[7] = 3;                           => detector #16  */
    int *group_counts;      /* Local group count array */

    int icounter;       /* Loop counter through inoperable detector list */
    int local_inop_count = 0;
    int curr_det_index;
    int num_consecutive_detectors = 1;
    int start_det_index = 0;    /* Starting detector index for a group of
                                   consecutive inoperable detectors */


    /* Initialize an iterator for the pixel mask that includes all spans
       flagged with inoperable detectors. */
    IAS_PIXEL_MASK_ITERATOR *inop_iterator = ias_pm_get_iterator(pmask,
        IAS_PM_INCLUDE, IAS_RPS_INOP_DETECTOR);
    if (inop_iterator == NULL)
    {
        IAS_LOG_ERROR("Getting inoperable detector iterator from pixel "
            "mask");
        return ERROR;
    }

    /* Check for any spans involving an inoperable detector.  It's assumed
       that all spans for a given detector are flagged as inoperable.
       The while loop will quit if either i) there aren't any spans at
       all; or ii) after the last span has been processed.  There can be
       multiple spans for a given detector: an image with one dropped frame
       will have three spans for the detector.  Keep track of the current
       detector so that it doesn't get counted multiple times. */
    curr_det_index = -1; /* An illegal detector index */
    while (ias_pm_get_next_span(inop_iterator, &span))
    {
        /* Only count the first span that shows up for this detector */
        if ( span.detector_index != curr_det_index )
        {
            /* Found an inoperable detector.  Save its index first, then count
               it */
            inoperable_detector_array[local_inop_count] =
                span.detector_index;
            local_inop_count++;

            /* Update the current detector index so it doesn't get counted
               again */
            curr_det_index = span.detector_index;
        }
    }

    /* Destroy the iterator and set it to NULL */
    ias_pm_destroy_iterator(inop_iterator);
    inop_iterator = NULL;

    /* Set the returned inoperable detector count value */
    *inop_detector_count = local_inop_count;

    /* If there are no inoperable detectors, it's not a failure.  Just
       return with a SUCCESS status */
    if (local_inop_count == 0)
        return SUCCESS;

    /* Allocate the status information data block.  If we've successfully
       allocated the data block, set the local access pointer to the base
       address of the block */
    *inop_status_info =
        malloc(local_inop_count * sizeof(INOP_DETECTOR_STATUS));
    if (*inop_status_info == NULL)
    {
        IAS_LOG_ERROR("Allocating inoperable detector status information "
            "data block");
        return ERROR;
    }
    det_status_ptr = *inop_status_info;

    /* Create a working array to contain the group count.  Initialize each
       detector's group count to 1 to start with.  While we're at it,
       initialize the detector status information data block with the
       inoperable detector indices */
    group_counts = malloc(local_inop_count * sizeof(*group_counts));
    if (group_counts == NULL)
    {
        IAS_LOG_ERROR("Allocating temporary group count array buffer");
        free(*inop_status_info);
        *inop_status_info = NULL;
        return ERROR;
    }
    for (icounter = 0; icounter < local_inop_count; icounter++)
    {
        det_status_ptr[icounter].det_index =
            inoperable_detector_array[icounter];
        group_counts[icounter] = 1;
    }

    /* Take another pass through the inoperable detector list to determine
       the actual group count for each inoperable detector.  Since we're
       checking the next array element to the right, stop the loop at
       (local_inop_count - 2) rather than (local_inop_count - 1) to avoid
       reading past the end of the data block */
    for (icounter = 0; icounter < (local_inop_count - 1); icounter++)
    {
        curr_det_index = det_status_ptr[icounter].det_index;
        int next_det_index = det_status_ptr[icounter + 1].det_index;
        int next_curr_diff = next_det_index - curr_det_index;

        /* If the next detector is also inoperable, update the group
           count for all detectors in the current 'group'.  If we have
           two consecutive inoperable detectors in a group, record the
           first inoperable detector's index so we can log its location*/
        if (next_curr_diff == 1)
        {
            int length_count;    /* consecutive detector loop counter */
            num_consecutive_detectors++;
            if (num_consecutive_detectors == 2)
                start_det_index = curr_det_index;

            /* This loop does the group count update, beginning with
               the current detector and going 'backwards' to the starting
               detector in the group */
            for (length_count = 0;
                 length_count < num_consecutive_detectors;
                 length_count++)
            {
                group_counts[icounter + 1 - length_count] =
                    num_consecutive_detectors;
            }
        }
        else
        {
            /* Issue a warning if 3 or more consecutive detectors are
               inoperable.  NOTE:  no warning will be issued here for
               a group of 3 or more consecutive inoperable detectors on
               the far right edge of the focal plane since the main
               processing loop stops when icounter = local_inop_count - 1.
               The last detector is checked in the last iteration of the
               loop (when icounter = local_inop_count -  2) and its status
               is updated as would be expected */
            if (num_consecutive_detectors > 2)
            {
                IAS_LOG_WARNING("%d consecutive inoperable detectors found "
                    "starting with detector %d.  Current algorithm "
                    "description indicates three or more consecutive "
                    "inoperable detectors are to be zero-filled.",
                    num_consecutive_detectors, (start_det_index + 1));
            }

            /* No more detectors in the current 'group'.  Reset the
               counter for the number of consecutive detectors to 1 */
            num_consecutive_detectors = 1;
        }
    }

    /* There could be a group of 3 or more inoperable detectors at the
       far right edge of the focal plane.  Check the count for the
       number of consecutive detectors and issue the warning at this point
       if needed */
    if (num_consecutive_detectors > 2)
    {
        IAS_LOG_WARNING("%d consecutive inoperable detectors found "
            "starting with detector %d.  Current algorithm description "
            "indicates three or more consecutive inoperable detectors "
            "are to be zero-filled.", num_consecutive_detectors,
            (start_det_index + 1));
    }

    /* Take a final pass through the inoperable detector list to determine
       the required fill types and replacement data detector(s) */
    for (icounter = 0; icounter < local_inop_count; icounter++)
    {
        /* Get the index for the current inoperable detector */
        curr_det_index = det_status_ptr[icounter].det_index;

        /* Set the fill type to indicate zero fill if the detector is part of
           a group of 3 or more consecutive inoperable detectors.  The '2'
           compared to here is the current maximum number of consecutive
           inoperable detectors that won't be zero-filled.  A later release
           should consider parameterization of this value if an algorithm
           to correct 3 or more consecutive detectors is developed.  */
        if (group_counts[icounter] > 2)
        {
            det_status_ptr[icounter].good_left_detector_index = -1;
            det_status_ptr[icounter].good_right_detector_index = -1;
            det_status_ptr[icounter].fill_type = INOP_FILL_ZERO;
        }
        else
        {
            /* Inoperable detectors are on the left edge of the SCA.
               Detector 1 (and possibly detector 2) */
            if (curr_det_index <= 1)
            {
                det_status_ptr[icounter].good_left_detector_index = -1;
                det_status_ptr[icounter].good_right_detector_index =
                    group_counts[icounter];
                det_status_ptr[icounter].fill_type =
                    INOP_FILL_EDGE_REPLACEMENT;
            }
            /* Inoperable detectors are on the right edge of the SCA.
               Detector n (and possibly detector n-1) */
            else if (curr_det_index >= (number_detectors - 2))
            {
                det_status_ptr[icounter].good_left_detector_index =
                    number_detectors - group_counts[icounter] - 1;
                det_status_ptr[icounter].good_right_detector_index = -1;
                det_status_ptr[icounter].fill_type =
                    INOP_FILL_EDGE_REPLACEMENT;
            }
            else
            {
                /*  One inoperable detector in the middle of
                    the SCA, surrounded by its two nearest neighbors */
                if (group_counts[icounter] == 1)
                {
                    det_status_ptr[icounter].good_left_detector_index =
                        curr_det_index - 1;
                    det_status_ptr[icounter].good_right_detector_index =
                        curr_det_index + 1;
                    det_status_ptr[icounter].fill_type =
                        INOP_FILL_NEIGHBOR_AVERAGE;
                }
                else
                {
                    /* Two consecutive inoperable detectors in the middle
                       of the SCA, surrounded by their nearest neighbors.
                       Look at the group count for the previous inoperable
                       detector to determine whether the current detector is
                       the leftmost or the rightmost detector in the sequence.
                       For the rightmost detector, use the same
                       replacement detectors as for the leftmost.

                       NOTE:  the '2' compared to here is also the current
                              maximum allowed number of consecutive
                              inoperable detectors that won't be zero-filled.
                              It's probably obvious, but I thought it
                              should be explicitly noted for greater
                              clarity */
                    if (group_counts[icounter - 1] != 2)
                    /* 'left' detector */
                    {
                        det_status_ptr[icounter].good_left_detector_index =
                            curr_det_index - 1;
                        det_status_ptr[icounter].good_right_detector_index =
                            curr_det_index + 2;
                        det_status_ptr[icounter].fill_type =
                            INOP_FILL_NEIGHBOR_AVERAGE;
                    }
                    else
                    /* 'right' detector */
                    {
                        det_status_ptr[icounter].good_left_detector_index =
                            det_status_ptr[icounter - 1]
                               .good_left_detector_index;
                        det_status_ptr[icounter].good_right_detector_index =
                            det_status_ptr[icounter - 1]
                            .good_right_detector_index;
                        det_status_ptr[icounter].fill_type =
                            INOP_FILL_NEIGHBOR_AVERAGE;
                    }
                }
            }
        }
    }

    /* Clean up */
    free(group_counts);

    return SUCCESS;
}   /* END -- classify_inoperable_detectors */



/*-------------------------------------------------------------------------
 NAME:     replace_image_data

 PURPOSE:  Performs an in-place, pixel-by-pixel replacement of image data
           for each inoperable detector

 RETURNS:  Integer status code of SUCCESS or ERROR

 NOTES:    For the current implementation, it is assumed that the
           replacement pixels used to correct the current line of
           data in the inoperable detector are valid.  It is possible,
           however, they are not (even though the detector itself
           is considered operable).  Should this prove to be an
           issue, the code will need to be updated to obtain and use
           pixel masks for the replacement detector(s).

           In the variable names, 'offset' refers to the detector spatial 
           offset, as determined from the CPF detector delays by the RPS
           library function 'ias_rps_convert_detector_delays2spatial_offsets'
---------------------------------------------------------------------------*/
static int replace_image_data
(
    float *l1r_buffer,                /* I/O: Image data buffer */
    int *det_offsets,                 /* I: Normalized detector offsets
                                         derived from CPF detector
                                         delays */
    INOP_DETECTOR_STATUS *inop_status_info,
                                      /* I: Inoperable detector status
                                         information */
    int   number_inop_detectors,      /* I: Total number of inoperable
                                         detectors */
    int   number_detectors,           /* I: Total number of detectors in
                                         the current band/SCA */
    int   number_lines                /* I: Total number of image data
                                          lines in the current band/SCA */
)
{
    int inop_count;            /* Loop counter for inoperable detectors */

    for (inop_count = 0; inop_count < number_inop_detectors; inop_count++)
    {
        /* Give the replacement data scale factors an initial value of 1.0.
           This will account for edge replacements using only one
           detector */
        double scale_factor_left = 1.0;
        double scale_factor_right = 1.0;

        int center_line;
        int left_det_offset;
        int right_det_offset;
        int line;

        /* Initialize the starting line indices to an obviously invalid value,
           equal to the total number of lines of image data.  The checks
           of the line indices later on are intended to ensure that
           pixel values are obtained only from the replacement detector(s)
           that are present, AND that the current line from which a
           pixel value is obtained is within the "valid" memory area
           of the image buffer.  These will be recalculated later as
           needed */
        int left_line = number_lines;
        int right_line = number_lines;

        /* Get the index and offset for the 'center' inoperable detector */
        int center_det_index = inop_status_info[inop_count].det_index;
        int center_det_offset = det_offsets[center_det_index];

        /* Get the index for each replacement detector.  If it isn't -1,
           get the corresponding offset, and calculate the starting
           line index.  NOTE:  The starting line index can be negative--
           in effect, there's no valid image data at that pixel
           'location' */
        int left_det_index =
            inop_status_info[inop_count].good_left_detector_index;
        int right_det_index =
            inop_status_info[inop_count].good_right_detector_index;

        if (left_det_index != -1)
        {
            left_det_offset = det_offsets[left_det_index];
            left_line = left_det_offset - center_det_offset;
        }
        if (right_det_index != -1)
        {
            right_det_offset = det_offsets[right_det_index];
            right_line = right_det_offset - center_det_offset;
        }

        /* For neighbor averaging fill, determine both left and right
           scaling factors */
        if (inop_status_info[inop_count].fill_type
            == INOP_FILL_NEIGHBOR_AVERAGE)
        {
            scale_factor_left = (double)(right_det_index - center_det_index)
                / (right_det_index - left_det_index);
            scale_factor_right = (double)(center_det_index - left_det_index)
                / (right_det_index - left_det_index);
        }

        /* Set the 'center' inoperable detector starting line index to 0
           to ensure replacement always starts on the first line */
        center_line = 0;

        /* Process each line of image data */
        for (line = 0; line < number_lines; line++)
        {
            float left_pixel_value = 0.0;
            float right_pixel_value = 0.0;
            int center_curr_pixel = center_det_index + (center_line
                * number_detectors);
            int got_left = 0;         /* Flag indicating we have valid
                                         data from the left replacement
                                         detector */
            int got_right = 0;        /* Flag indicating we have valid
                                         data from the right replacement
                                         detector */

            /* If the current replacement detector line indices are within
               the image data buffer, get the current pixel value(s) from the
               replacement detector(s), and set the flags indicating that we
               have data from one or both.  */
            if ((left_line >= 0) && (left_line < number_lines))
            {
                left_pixel_value = l1r_buffer[left_det_index
                    + left_line * number_detectors];
                got_left = 1;
            }

            if ((right_line >= 0) && (right_line < number_lines))
            {
                right_pixel_value = l1r_buffer[right_det_index
                    + right_line * number_detectors];
                got_right = 1;
            }

            /* For the replacement detector data we have, scale the
               input pixel value(s) (summing the scaled values if data
               from both replacement detectors are available), and write
               the results to the image buffer.  */
            if (got_left)
            {
                if (got_right)
                {
                    l1r_buffer[center_curr_pixel] =
                        (float)(scale_factor_left) * left_pixel_value
                        + (float)(scale_factor_right) * right_pixel_value;
                }
                else
                {
                    l1r_buffer[center_curr_pixel] = (float)(scale_factor_left)
                        * left_pixel_value;
                }
            }
            else if (got_right)
            {
                l1r_buffer[center_curr_pixel] = (float)(scale_factor_right)
                    * right_pixel_value;
            }
            else
            {
                /* This accounts for the case of 3 or more consecutive
                   inoperable detectors.  It also accounts for the
                   possibility that there are no valid replacement pixels
                   at the current line indices in both replacement
                   detectors. */
                l1r_buffer[center_curr_pixel] = 0.0;
            }

            /* Increment the line indices for the center 'inoperable'
               detector, as well as the replacement detectors.  For
               an edge replacement using one replacement detector, the other
               detector's starting line index is already invalid; it will
               still be invalid when incremented... */
            left_line++;
            center_line++;
            right_line++;
        }
    }

    return SUCCESS;
}   /* END  -- replace_image_data */



/*------------------------------------------------------------------------
 NAME:     inoperable_detectors_fill

 PURPOSE:  Main routine to perform pixel correction for inoperable
           detectors

 RETURNS:  Integer status code of SUCCESS or ERROR
-------------------------------------------------------------------------*/
int inoperable_detectors_fill
(
    float *l1r_buffer,           /* I/O:  On input--original pixel values
                                    On output--adjusted pixel values */
    int number_detectors,        /* I: Number of detectors in the current
                                    band/SCA */
    int number_lines,            /* I: Number of lines in the current SCA */
    int *det_offsets,            /* I: Normalized CPF detector offsets
                                    for the current band/SCA */
    IAS_PIXEL_MASK *pmask        /* I: Pixel mask for the current band/SCA */
)
{
    INOP_DETECTOR_STATUS *inop_status_info = NULL;
                                             /* Inoperable status
                                                information for all
                                                detectors in the current
                                                band/SCA */
    int number_inoperable_detectors = 0;     /* Total number of inoperable
                                                detectors that were found */
    int status;


    /* Classify any inoperable detectors identified in the pixel mask */
    status = classify_inoperable_detectors(pmask, number_detectors,
        &number_inoperable_detectors, &inop_status_info);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Inoperable detector classification failed.");
        if (inop_status_info)
            free(inop_status_info);
        return ERROR;
    }
    IAS_LOG_DEBUG("Found %d detectors", number_inoperable_detectors );

    /* Fix the data in any inoperable detector(s) */
    if (number_inoperable_detectors != 0)
    {
        status = replace_image_data(l1r_buffer, det_offsets,
            inop_status_info, number_inoperable_detectors,
            number_detectors, number_lines);
        free(inop_status_info);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Inoperable detector data replacement failed.");
        }
        else
        {
            IAS_LOG_DEBUG("Inoperable detector data replacement completed "
                "successfully.");
        }
    }

    return status;
}   /* END -- inoperable detectors fill */
