/*----------------------------------------------------------------------------
NAME:    oli bias determination

PURPOSE: Estimate the bias which needs to be removed for each detector. There 
         are five sources the detector-level bias can come from, the 
         pre_shutter, post_shutter from BPF, the average of the two, bias from
         cpf, or the bias estimated with a1/C1 coefficients and the VRP band 
         mean value. There are two options to extend the detector-level bias to
         frame-level bias The first option is simply to extend the per detector
         bias to per frame level. The second option is to use a linear model 
         with a coefficient a0 from BPF and the VRP band data. 

RETURNS: SUCCESS or ERROR

ALGORITHM REFERENCES: See the Bias_Determination_V3.1f.doc Algorithm 
                      Description Document for details of this algorithm.

Note:    1. Per discussion with Cal/Val, they will not care about per-frame
         bias value if it's within the dropped_frame line because it doesn't 
         have any meaning.
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "ias_logging.h"
#include "ias_pixel_mask.h"
#include "ias_const.h"
#include "ias_radiometry.h"
#include "rps_local.h"

/* Prototypes for report generation */
int generate_report_file
(
    int band_number,              /* I: 1-based band number */
    int sca_number,               /* I: 1-based SCA number */
    int num_detectors,            /* I: Number of detectors in the SCA */
    int num_lines,                /* I: Number of lines in the data */
    IAS_REPORT *report_info,      /* I: IAS report structure pointer */
    double *perframe_bias         /* O: Pointer to the results of bias
                                        value per detector, per frame */
);

int compute_vrp_mean
(
    uint16_t *l0r_vrp_data,      /* I: One SCA of VRP data in DN */
    int vrp_detectors,           /* I: Number of detectors in the VRP band */
    int vrp_lines,               /* I: Number of lines in the VRP band */
    BIAS_SPECTRAL_TYPE bias_spectral_type,    
        /* Flag to indicate the spectral type for bias application, VNIR =0,
           SWIR = 1, PAN_ODD =2, PAN_EVEN = 3 */
    IAS_PIXEL_MASK *pixel_mask,  /* I: Contains inop detector and other
                                       ignore pixel locations */
    double *line_mean,           /* O: Array for each line vrp data mean */
    double *vrp_mean             /* O: VRP band mean value for the whole SCA */
);

/* OLI Bias Determination Algorithm */
int oli_bias_determination
(
    uint16_t *l0r_vrp_data,     /* I: One SCA of VRP data in DN */
    int band_number,            /* I: Band number, 1-based */
    int sca_number,             /* I: SCA number, 1-based */
    int num_detectors,          /* I: Number of detectors in the SCA */
    int num_lines,              /* I: Number of lines in the data */
    int vrp_detectors,          /* I: Number of detectors in the VRP band */
    int vrp_lines,              /* I: Number of lines in the VRP band */
    int perframe_bias_flag,     /* Flag to indicate the selected bias type, 
                                   0 = PER_DETECTOR_BIAS, 1 = PER_FRAME_BIAS */
    int a1_c1_estimate_flag,    /* Flag to indicate the detector level bias are
                                   from a1/C1 estimation option, YES/NO = 1/0 */
    BIAS_SPECTRAL_TYPE bias_spectral_type,    
        /* Flag to indicate the spectral type for bias application, NON_PAN = 0,
           PAN_ODD = 1, PAN_EVEN = 2 */
    double *bias_from_file,     /* I: Per detector bias from BPF/CPF */
    const double *a1_parameter, /* I: Per detector a1_parameter from BPF */
    const double *c1_parameter, /* I: Per detector c1_parameter from BPF */
    const double a0_parameter,  /* I: a0_parameter in model calculation */
    IAS_PIXEL_MASK *pixel_mask, /* I: Contains inop detector and other
                                      ignore pixel locations */
    IAS_REPORT *report_info,    /* I: General information for the test report. 
                                      If no report is desired pass NULL. */
    double *perframe_bias       /* O: Pointer to the results of bias
                                      value per detector, per frame */
)
{
    int det = 0;                 /* loop counter for the detectors */
    int line = 0;                /* loop counter for the frames */
    double *line_mean = NULL;    /* pointer for each line vrp data mean */
    double vrp_mean;             /* vrp band mean value for the whole SCA */
    int generate_report = 1;     /* Report generation flag */
    int start_line;              /* Starting line for the loop */
    int step_size;               /* Step size for the loop */
    int status;                  /* Return status */
    double weighted_vrp_mean;    /* The whole SCA VRP mean value times A0 */
    int odd_start_line;          /* Old start line number for PAN band */
    int even_start_line;         /* Even start line number for PAN band */
    double *perdetector_bias = NULL;/* Pointer to hold detector-level bias */

    /* The per-frame bias calculation requires vrp_lines >= num_lines and 
       normally the vrp_lines should match num_lines, otherwise there might
       be a mis-alignment somewhere, or difficult to align the two */
    if (vrp_lines != num_lines)
    {
        IAS_LOG_ERROR("Number of lines in VRP band doesn't match "
                      "number of lines in the corresponding image band");
        return ERROR;
    }

    /* Set the flags false if NULL was passed in the report structures. */
    if (report_info == NULL)
        generate_report = 0;

    /* Determine what the starting even and odd lines are based on the MACRO
       defined in the library */
    if ( IAS_IS_EVEN_LINE(0) )
    {
        even_start_line = 0;
        odd_start_line = 1;
    }
    else
    {
        odd_start_line = 0;
        even_start_line = 1;
    }

    /* Set the loop variables for different spectral types */
    switch(bias_spectral_type)
    {
        case PAN_ODD:
            start_line = odd_start_line;
            step_size = 2;
            break;

        case PAN_EVEN:
            start_line = even_start_line;
            step_size = 2;
            break;

        case NON_PAN:
            start_line = 0;
            step_size = 1;
            break;

        default:
            IAS_LOG_ERROR("Unsupported bias_spectral_type");
            return ERROR;
    }

    /* The VRP band data is only needed when per-frame output bias type is
       selected or per-detector output bias type is selected and the detector 
       level bias are estimated with a1 & C1 coefficients */
    if (perframe_bias_flag || a1_c1_estimate_flag)
    {
        /* Allocate memory for the lines of VRP frame mean */
        line_mean = malloc(vrp_lines * sizeof(double));
        if (!line_mean)    
        {
            IAS_LOG_ERROR("Allocating memory for line_mean");
            return ERROR;    
        }

        /* Call the compute_vrp_mean routine */
        status = compute_vrp_mean(l0r_vrp_data, vrp_detectors, vrp_lines, 
                 bias_spectral_type, pixel_mask, line_mean, &vrp_mean);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Call compute_vrp_mean at band_number: %d, "
                          "SCA_number: %d", band_number, sca_number);
            free(line_mean);
            return ERROR; 
        }
    }

    /* Populate perdetector_bias properly */
    if (a1_c1_estimate_flag) 
    {
        /* Allocate memory for perdetector_bias  */
        perdetector_bias = malloc(num_detectors * sizeof(double));
        if (!perdetector_bias)
        {
            IAS_LOG_ERROR("Allocating memory for bias_estimate");
            free(line_mean);
            return ERROR;    
        }

        /* Estimate detector level bias with a1/C1 estimation */
        for (det = 0; det < num_detectors; det++)
            perdetector_bias[det] = a1_parameter[det] * vrp_mean + 
                                    c1_parameter[det];
    }
    else
    {
        /* Link the perdetector_bias to bias_from_file */
        perdetector_bias = bias_from_file;
    }

    if (!perframe_bias_flag)  /* Extend the detector level bias to frame 
                                 level */ 
    {
        for (line = start_line; line < num_lines; line += step_size)
        {
            double *bias_ptr;  /* Local bias pointer for each line */
            bias_ptr = &perframe_bias[line * num_detectors];
            for (det = 0; det < num_detectors; det++)
                bias_ptr[det] = perdetector_bias[det];
        }
    }
    else /* Calculated frame level bias with a0 coefficient and VRP data */
    {
        /* Calculate a0_parameter * vrp_mean to reduce calculation in the 
           loop */
        weighted_vrp_mean = a0_parameter * vrp_mean;

        /* Calculate the bias based on model calculation */
        for (line = start_line; line < num_lines; line += step_size)
        {
            double *bias_ptr = NULL;   /* Local bias pointer for each line */
            double temp_line_mean;
            bias_ptr = &perframe_bias[line * num_detectors];
            /* Calculate a0_parameter * line_mean to reduce calculation in the
               loop */
            temp_line_mean = a0_parameter * line_mean[line];
            for (det = 0; det < num_detectors; det++)
                bias_ptr[det] = temp_line_mean - weighted_vrp_mean 
                                + perdetector_bias[det];
        }
    }

    /* free the allocated memory */
    if (perframe_bias_flag || a1_c1_estimate_flag)
        free(line_mean);
    if (a1_c1_estimate_flag)
        free(perdetector_bias);

    /* Output the report file */
    if (generate_report)
    {
        /* There is only one pan band report and it's outputted after both
           pan_odd and pan_even bands are processed */
        if (bias_spectral_type != PAN_ODD)
        {
            /* Generate detector report file for one band, one SCA */
            status = generate_report_file(band_number, sca_number, 
                     num_detectors, num_lines, report_info, perframe_bias);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Generating report file");
                return ERROR;
            }
        }
    }

    return SUCCESS;
} 

/*----------------------------------------------------------------------------
NAME: compute_vrp_mean

PURPOSE: Compute VRP data per frame mean and the whole SCA mean values

RETURNS: SUCCESS or ERROR

Note:    Per discussion with Cal/Val, they will not care about per-frame
         bias value if it's within the dropped_frame line because it doesn't 
         have any meaning, therefore, the value of line_mean value for a 
         dropped frame doesn't matter. The setting here is just for unit test
         comparison purpose with val/val prototype code results.
-----------------------------------------------------------------------------*/
int compute_vrp_mean
(
    uint16_t *l0r_vrp_data,      /* I: One SCA of VRP data in DN */
    int vrp_detectors,           /* I: Number of detectors in the VRP band */
    int vrp_lines,               /* I: Number of lines in the VRP band */
    BIAS_SPECTRAL_TYPE bias_spectral_type,    
        /* Flag to indicate the spectral type for bias application, VNIR =0,
           SWIR = 1, PAN_ODD =2, PAN_EVEN = 3 */
    IAS_PIXEL_MASK *pixel_mask,  /* I: Contains inop detector and other
                                       ignore pixel locations */
    double *line_mean,           /* O: Array for each line vrp data mean */
    double *vrp_mean             /* O: VRP band mean value for the whole SCA */
)
{
    int det = 0;                 /* loop counter for the detectors */
    int valid_pixel_count;       /* counter used for statistics calculation */
    double valid_pixel_sum;      /* summary used for statistics calculation */
    int line = 0;                /* loop counter for the frames */
    double valid_line_sum;       /* summary used for statistics calculation */
    int valid_line_pixel_count;  /* valid pixel counter within one line */
    PIXEL_MASK_TYPE artifacts_to_ignore;/* Mask values of pixels to ignore */
    PIXEL_MASK_TYPE *temp_mask;  /* Pixel mask for the VRP image */
    int return_status = SUCCESS; /* Status to return to caller */
    int start_line;              /* Starting line for the loop */
    int step_size;               /* Step size for the loop */
    int status;                  /* Return status */
    int odd_start_line;          /* Old start line number for PAN band */
    int even_start_line;         /* Even start line number for PAN band */

    /* Create a convenience mask of artifacts to ignore */
    artifacts_to_ignore = IAS_RPS_INOP_DETECTOR|IAS_RPS_IMPNOISE|
                          IAS_RPS_ADCONV_SAT_LO|IAS_RPS_ADCONV_SAT_HI|
                          IAS_RPS_ANALOG_SAT_LO|IAS_RPS_ANALOG_SAT_HI;

    /* Determine what the starting even and odd lines are based on the MACRO
       defined in the library */
    if ( IAS_IS_EVEN_LINE(0) )
    {
        even_start_line = 0;
        odd_start_line = 1;
    }
    else
    {
        odd_start_line = 0;
        even_start_line = 1;
    }

    /* Set the loop variables for different spectral types */
    switch(bias_spectral_type)
    {
        case PAN_ODD:
            start_line = odd_start_line;
            step_size = 2;
            break;

        case PAN_EVEN:
            start_line = even_start_line;
            step_size = 2;
            break;

        case NON_PAN:
            start_line = 0;
            step_size = 1;
            break;

        default:
            IAS_LOG_ERROR("Unsupported bias_spectral_type");
            return ERROR;
    }

    /* Allocate memory for the temporary pixel masks */
    temp_mask = malloc(sizeof(PIXEL_MASK_TYPE) * vrp_lines 
                       * vrp_detectors);
    if (!temp_mask)
    {
        IAS_LOG_ERROR("Allocating pixel mask memory");
        return ERROR;
    }

    /* Grab the pixel mask information for all frames of data 
       for the current detector */
    status = ias_pm_get_image_subset(pixel_mask, 0, 0, vrp_detectors,
                                     vrp_lines, temp_mask);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving pixel mask subset for "
                      "the whole SCA");
        free(temp_mask);
        return ERROR;
    }

    /* Calculate line_mean and SCA mean VRP data */
    valid_pixel_count = 0;
    valid_pixel_sum = 0.0;
    /* Note: Pixel mask is transpose compared with image data */
    for (line = start_line; line < vrp_lines; line += step_size)
    {
        uint16_t *l0r_vrp_ptr;
        l0r_vrp_ptr = &l0r_vrp_data[line * vrp_detectors];
        /* Now ignore the the whole line of detectors if drop frame */
        if (temp_mask[line] & IAS_RPS_DROPLINE)
        {
            line_mean[line] = -1.0;
            continue;
        }

        valid_line_pixel_count = 0;
        valid_line_sum = 0.0;
        for (det = 0; det < vrp_detectors; det++)
        {
            PIXEL_MASK_TYPE *line_mask;  /* Line of pixel masks from image */
            line_mask = temp_mask + det * vrp_lines;
            /* See if the current pixel is an inop_detector, 
               fill frame, or saturated noise and skip it */
            if (line_mask[line] & artifacts_to_ignore)
            {
                continue;
            }
            valid_line_sum += (double)l0r_vrp_ptr[det];
            valid_line_pixel_count++;
        }
        valid_pixel_count += valid_line_pixel_count;
        valid_pixel_sum += valid_line_sum;

        /* Calculate frame vrp mean values */
        if (valid_line_pixel_count != 0)
            line_mean[line] = valid_line_sum / valid_line_pixel_count;
        else
            line_mean[line] = 0;
    }

    /* Calculate vrp mean values */
    if (valid_pixel_count != 0)
        *vrp_mean = valid_pixel_sum / valid_pixel_count;
    else
        *vrp_mean = 0.0;

    /* free the allocated memory */
    free(temp_mask);

    return return_status;
} 


/*----------------------------------------------------------------------------
NAME: generate_report_file   

PURPOSE: Generate the report file in binary format

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int generate_report_file
(
    int band_number,              /* I: 1-based band number */
    int sca_number,               /* I: 1-based SCA number */
    int num_detectors,            /* I: Number of detectors in the SCA */
    int num_lines,                /* I: Number of lines in the data */
    IAS_REPORT *report_info,      /* I: IAS report structure pointer */
    double *perframe_bias         /* O: Pointer to the results of bias
                                        value per detector, per frame */
)
{
    FILE *report_file = NULL;
    char file_base_name[] = "binary_oli_bias_determination";
    char *file_name;
    int num_values;             /* Number of values for write */
    int n;                      /* Number of values write_to_file */

    /* Get output report file name */
    file_name = ias_rps_get_report_file_name(report_info, band_number, 
                                             sca_number, file_base_name);
    if (file_name == NULL)
    {
        IAS_LOG_ERROR("Getting report filename %s", file_name);
        return ERROR;        
    }

    /* Open the report file in binary format */
    report_file = fopen(file_name, "w");
    if (!report_file)
    {
        IAS_LOG_ERROR("Opening report file %s", file_name);
        free(file_name);
        return ERROR;
    }

    /* Write the frame level bias to the report file */
    num_values = num_detectors * num_lines;
    n = fwrite(&perframe_bias[0], sizeof(double), num_values, report_file);
    if (n != num_values)
    {
        IAS_LOG_ERROR("Writing binary report file %s", file_name);
        free(file_name);
        fclose(report_file);
        return ERROR;
    }

    if (fclose(report_file) != SUCCESS)
    {
        IAS_LOG_ERROR("Closing binary report file %s", file_name);
        free(file_name);
        return ERROR;
    }
    free(file_name);

    return SUCCESS;
}

