/******************************************************************************
 NAME:    oli_bias_removal

 PURPOSE: Subtracts per-line bias from OLI L1R image data
 
 RETURNS: Nothing

 NOTES:   Assumes image, bias, and temperature correction data buffers
          have been initialized through a memory management call in the
          API routine.  In addition, assumes the bias determination
          algorithm produces per-line detector bias for all OLI bands

******************************************************************************/
#include "ias_logging.h"
#include "rps_local.h"


int oli_bias_removal
(
    float *l1r_image_buffer,         /* I/O: Image data for the current
                                        band/SCA */
    const double *cpf_bias_buffer1,  /* I: CPF bias data buffer #1 (pass in
                                        NULL if per-line bias is used).  For
                                        PAN band, "odd" state detector
                                        biases.  For MS bands, detector
                                        biases */
    const double *cpf_bias_buffer2,  /* I: CPF bias data buffer #2 (pass in
                                        NULL if per-line bias is used).  For
                                        PAN band, "even" state detector
                                        biases.  For MS bands, pass in NULL*/
    const double *per_line_bias_buffer,
                                     /* I: Per-line bias data for the current
                                        band/SCA (pass in NULL if CPF current
                                        bias is used) */
    BIAS_SOURCE bias_source,         /* I: Source of bias correction to
                                        apply */
    int perframe_bias_flag,          /* Flag to indicate the selected bias type
                                        0 = PER_DETECTOR_BIAS, 
                                        1 = PER_FRAME_BIAS */
    int pan_flag,                    /* I: Flag indicating whether PAN band
                                        is being processed (0=no, 1=yes) */
    int number_detectors,            /* I: # of detectors per SCA in the
                                        current band */
    int number_lines,                /* I: # of lines of image data in the
                                        current band/SCA */
    double *temp_correction_factors
                                     /* I: Temperature sensitivity
                                       correction factors for the current
                                       band/SCA (pass in NULL if no
                                       temperature correction is to be
                                       performed--default case */
)
{
    const double *odd_cpf_bias_ptr = NULL;  /* Local pointer to start of "odd"
                                         line CPF bias data buffer */
    const double *even_cpf_bias_ptr = NULL; /* Local pointer to start of "even"
                                         line CPF bias data buffer */
    const double *bias_ptr = NULL;/* Generic pointer to bias.  For CPF
                                    (detector-specific) correction, this
                                    points to the start of the selected
                                    (odd/even) CPF bias data buffer.
                                    For per-line correction, this
                                    points to the start of the current
                                    line in the per-line bias data buffer.
                                    In both cases, data pointed to are
                                    accessed as they would be in an
                                    array (i.e. bias_ptr[det], where 'det'
                                    is the index of the current detector) */
    float *image_line = NULL;    /* Local pointer to the current line
                                    of image data.  Data pointed to
                                    are accessed as they would be in
                                    an array */
    int line_start;              /* Index in the input image data buffer
                                    to the first pixel (detector 1) of
                                    the current line */
    int line;                    /* Line loop counter */
    int det;                     /* Detector loop counter */
    int use_cpf_bias = 0;        /* Local flag indicating whether CPF
                                    bias values are being used */
    int apply_temp_correction = 0; /* Local flag indicating whether temp.
                                      correction is applied */


    /* Make sure we have some kind of bias data available by checking
       for NULL pointers.  If we're using CPF bias data, set up local
       pointers to the individual "odd" and "even" bias data buffers */
    if (bias_source == CPF_BIAS && perframe_bias_flag == 0)
    {
        if ((cpf_bias_buffer1 == NULL)
            || (pan_flag && (cpf_bias_buffer2 == NULL)))
        {
            IAS_LOG_ERROR("CPF bias data missing");
            return ERROR;
        }

        /* Indicate we're using CPF bias.  Use a local flag so we don't
           have to do both comparisons again for each line */
        use_cpf_bias = 1;

        /* Set up local pointers to the CPF bias data buffers */
        odd_cpf_bias_ptr = cpf_bias_buffer1;
        if (pan_flag)
            even_cpf_bias_ptr = cpf_bias_buffer2;
        else
            even_cpf_bias_ptr = cpf_bias_buffer1;
    }
    else
    {
        if (per_line_bias_buffer == NULL)
        {
            IAS_LOG_ERROR("Per-line bias correction specified, but no "
                "per-line bias data exists");
            return ERROR;
        }
    }

    /* Check whether we have temperature correction factors and if we do,
       set the flag to indicate temperature correction should be applied */
    if (temp_correction_factors != NULL)
        apply_temp_correction = 1;

    /* Loop through all lines of image data */
    for (line = 0; line < number_lines; line++)
    {
        /* Determine the location representing the start of the
           current image line, and set the image line pointer to that
           location */
        line_start = line * number_detectors;
        image_line = &l1r_image_buffer[line_start];

        /* Set the local bias pointer to the "correct" set of biases */
        if (use_cpf_bias)
        {
            if (IAS_IS_EVEN_LINE(line))
            {
                bias_ptr = even_cpf_bias_ptr;
            }
            else
            {
                bias_ptr = odd_cpf_bias_ptr;
            }
        }
        else
        {
            bias_ptr = &per_line_bias_buffer[line_start];
        }

        /* Remove the bias */
        if (apply_temp_correction)
        {
            for (det = 0; det < number_detectors; det++)
            {
                image_line[det]
                    -= (temp_correction_factors[det] *  bias_ptr[det]);
            }
        }
        else
        {
            for (det = 0; det < number_detectors; det++)
            {
                image_line[det] -= bias_ptr[det];
            }
        }
    }

    return SUCCESS;
}   /* END oli_bias_removal */
