/*----------------------------------------------------------------------------
NAME:    reflectance_conversion

PURPOSE: Apply the radiance reflectance conversion as needed.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include "ias_logging.h"
#include "rps_local.h"

int reflectance_conversion
(
    float *l1r_data,             /* I/O: One SCA of l1r data to apply gain */
    int num_detectors,           /* I: Number of detectors in the SCA */
    int num_frames,              /* I: Number of frames in the data */
    double earth_sun_distance,   /* I: earth sun distance in astronomy unit
                                       from NOVAS libray calculation */
    double rad_to_reflect_conv_coeff,       
                                 /* I: band level radiance to reflectance
                                       conversion coefficient */
    double *reflect_to_rad_conv_coeff /* O: band level reflectance to radiance
                                            conversion coefficient */   
)
{
    long pixel;                   /* Loop counter for the pixels */
    double inverse_reflect_to_rad_conv_coeff;
    float *pixel_ptr;
    long num_pixels;

    /* Calculate the inverse reflectance to radiance conversion coeff. */
    inverse_reflect_to_rad_conv_coeff =  (earth_sun_distance 
        * earth_sun_distance) * rad_to_reflect_conv_coeff;

    /* Calculate band level reflectance to radiance conversion coefficient */
    if (inverse_reflect_to_rad_conv_coeff != 0)
    {
        *reflect_to_rad_conv_coeff = 1.0 / inverse_reflect_to_rad_conv_coeff;
    }
    else
    {
        IAS_LOG_ERROR("Either earth_sun_distance or rad_to_reflect_conv_coeff "
                      "is zero");
        return ERROR;  
    }

    /* Calculate number of pixels */
    num_pixels = num_frames * num_detectors;

    pixel_ptr = l1r_data;

    /* Loop through the pixel across frames and detectors */
    for (pixel = 0; pixel < num_pixels; pixel++)
    {
        pixel_ptr[pixel] *= inverse_reflect_to_rad_conv_coeff; 
    }

    return SUCCESS;
}  /* END ias_rad_reflectance_conversion */

