/*----------------------------------------------------------------------------
NAME:   saturated_pixel_replacement 

PURPOSE: Apply saturated pixel replacement correction to the L1R image

RETURNS: SUCCESS or ERROR

ALGORITHM REFERENCES: See the Sat_Pixel_Replacement_V3.1.doc Algorithm
                      Description Document for details of this algorithm.

-----------------------------------------------------------------------------*/
#include <stdlib.h>

#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_pixel_mask.h"
#include "rps_local.h"

int saturated_pixel_replacement
(
    float *data_buffer,         /* I/O: Input and output L1R data buffer */
    int num_detectors,          /* I: Number of detectors in the SCA */
    int num_lines,              /* I: Number of lines in the data */
    IAS_PIXEL_MASK *pixel_mask, /* I: Contains pixel artifacts */
    float high_replace,         /* I: High replacement value */
    float low_replace,          /* I: Low replacement value */
    int digital_high_flag,      /* I: Digitally high pixels are replaced flag*/
    int digital_low_flag,       /* I: Digitally low pixels are replaced flag */
    int analog_high_flag,       /* I: Analog high pixels are replaced flag */
    int analog_low_flag         /* I: Analog high pixels are replaced flag */
)
{
    PIXEL_MASK_TYPE artifacts_to_replace_low = 0;/* replace low pixels */
    PIXEL_MASK_TYPE artifacts_to_replace_high = 0;/* replace high pixels */
    IAS_PIXEL_MASK_ITERATOR *low_it;
    IAS_PIXEL_MASK_ITERATOR *high_it;
    IAS_PIXEL_MASK_SPAN high_span;
    IAS_PIXEL_MASK_SPAN low_span;
    int span_idx;                   /* Span of pixel max loop control */
    int replace_high_flag = FALSE;  /* Flag that high pixels are replaced */
    int replace_low_flag = FALSE;   /* Flag that low pixels are replaced */
    float *data_ptr;

    /* Create a convenience mask of artifacts to replace high pixels */
    if (digital_high_flag)
    {
        artifacts_to_replace_high |= IAS_RPS_ADCONV_SAT_HI;
        replace_high_flag = TRUE;
    }

    if (analog_high_flag)
    {
        artifacts_to_replace_high |= IAS_RPS_ANALOG_SAT_HI;
        replace_high_flag = TRUE;
    }

    /* Create a convenience mask of artifacts to replace low pixels */
    if (digital_low_flag)
    {
        artifacts_to_replace_low |= IAS_RPS_ADCONV_SAT_LO;
        replace_low_flag = TRUE;
    }

    if (analog_low_flag)
    {
        artifacts_to_replace_low |= IAS_RPS_ANALOG_SAT_LO;
        replace_low_flag = TRUE;
    }

    /* Replace the high pixels if the flag indicates */
    if (replace_high_flag)
    {
        /* Get an iterator for high saturated pixels */
        high_it = ias_pm_get_iterator(pixel_mask, IAS_PM_INCLUDE,
            artifacts_to_replace_high);
        if (!high_it)
        {
            IAS_LOG_ERROR("Could not get the high saturated pixel iterator");
            return ERROR;
        }

        while (ias_pm_get_next_span(high_it, &high_span))
        {
            /* Find the pointer to the value to change */
            data_ptr = data_buffer + 
                        ((high_span.starting_pixel_index * num_detectors) + 
                        high_span.detector_index);

            /* Step through the span and assign the replacement value */
            for (span_idx = 0; span_idx < high_span.length_of_span; span_idx++)
            {
                /* Assign the data pointer */
                *data_ptr = high_replace;
                data_ptr += num_detectors;
            }
        }
        ias_pm_destroy_iterator(high_it);
        high_it = NULL;
    }

    /* Replace the low pixels if the flag indicates */
    if (replace_low_flag)
    {
        /* Get an iterator for low saturated pixels */
        low_it = ias_pm_get_iterator(pixel_mask, IAS_PM_INCLUDE,
            artifacts_to_replace_low);
        if (!low_it)
        {
            IAS_LOG_ERROR("Could not get the low saturated pixel iterator");
            return ERROR;
        }

        /* Retrieve the next span */
        while (ias_pm_get_next_span(low_it, &low_span))
        {
            /* Find the pointer to the value to change */
            data_ptr = data_buffer + 
                        ((low_span.starting_pixel_index * num_detectors) + 
                        low_span.detector_index);

            /* Step through the span and assign the replacement value */
            for (span_idx = 0; span_idx < low_span.length_of_span; span_idx++)
            {
                /* Assign the data pointer */
                *data_ptr = low_replace;
                data_ptr += num_detectors;
            }
        }
        ias_pm_destroy_iterator(low_it);
        low_it = NULL;
    }

    return SUCCESS;
}
