/*----------------------------------------------------------------------------
NAME:   sca_discontinuity_correction 

PURPOSE: Apply discontinuity correction to the L1R image

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_pixel_mask.h"
#include "rps_local.h"

int sca_discontinuity_correction
(
    float *data_buffer,         /* I/O: Input and output L1R data buffer */
    int num_detectors,          /* I: Number of detectors in the SCA */
    int num_lines,              /* I: Number of lines in the data */
    IAS_PIXEL_MASK *pixel_mask, /* I: Contains pixel artifacts */
    double correction_factor    /* I: Normalized corection factor */
)
{
    PIXEL_MASK_TYPE artifacts_to_ignore; /* Mask values of pixels to ignore */
    PIXEL_MASK_TYPE *temp_mask;          /* Line of pixel masks from image */
    int det_index;       /* detector loop index */
    int line_index;      /* Line loop index */
    int status;

    /* Create a convenience mask of artifacts to ignore which includes
       dropped frame, impulse noise, or saturated pixels */
    artifacts_to_ignore = IAS_RPS_DROPLINE|IAS_RPS_IMPNOISE|
                          IAS_RPS_ADCONV_SAT_LO|IAS_RPS_ADCONV_SAT_HI|
                          IAS_RPS_ANALOG_SAT_LO|IAS_RPS_ANALOG_SAT_HI;

    /* Allocate the temporary lines of pixel masks */
    temp_mask = malloc(sizeof(PIXEL_MASK_TYPE) * num_lines);
    if (!temp_mask)
    {
        IAS_LOG_ERROR("Allocating pixel mask memory");
        return ERROR;
    }

    /* Loop through the detectors */
    for (det_index = 0; det_index < num_detectors; det_index++)
    {
        float *data_buffer_ptr = &data_buffer[det_index];
        /* Grab the pixel mask information for all frames of data for
           the current detector */
        status = ias_pm_get_image_subset(pixel_mask, det_index, 0, 1,
             num_lines, temp_mask);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Retrieving pixel mask subset for detector %d",
                          (det_index + 1));
            free(temp_mask);
            return ERROR;
        }

        /* Skip the whole detector if the first line pixel is flagged as 
           inoperable */
        if (temp_mask[0] & IAS_RPS_INOP_DETECTOR)
        {
            continue;
        }

        /* Loop through the lines of the input image.   */
        for (line_index = 0; line_index < num_lines; line_index++)
        {
            /* See if the current pixel is to skip */
            if (!(temp_mask[line_index] & artifacts_to_ignore))
            {
                /* Update the image data with SCA discontinuity correction 
                  applied */
                data_buffer_ptr[0] *= correction_factor;
            }

            /* Move the pointer to the next line */
            data_buffer_ptr += num_detectors;
        }
    }

    free(temp_mask);
    return SUCCESS;
} 

