/*----------------------------------------------------------------------------
NAME:    striping_characterization

         Contains:
             striping_characterization
             calculate_overall_striping_metric
             generate_overall_striping_metric_report
             db_store_overall_striping_metric_in_database
             db_allocate_overall_striping_metric_results
             db_free_overall_striping_metric_results
             generate_reports (local)
             compute_filters_and_diff (local)
             compare_doubles (local)
             decreasing_compare_array_map (local)

PURPOSE: Determines a quantitative metric for the amount of striping present
         in an image through calculation of spatial-domain statistics for
         each detector.  The detector statistics are further processed to
         obtain an overall striping metric.  A striping correction matrix,
         the same size as the input image, is also produced that may be used
         to reduce striping in the image.

         NOTE: The main algorithm operates on complete, nominally aligned
         image data.  Complete lines are those in which all samples, with
         the exception of left and right edge samples, have neighbors.

ALGORITHM REFERENCES: See the Striping_Char_v3.0.docx Algorithm Description
         document for details of this algorithm.
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "ias_logging.h"
#include "ias_db_insert.h"
#include "rps_local.h"

#define MEDIAN_FILTER_WIDTH 75
#define MEAN_FILTER_WIDTH   15
#define NUMBER_PEAKS        15

/* This macro computes the relative start and end line index of a span
   taking into account the detector offset and the index of the last full
   line of the image */
#define GET_SPAN_RELATIVE_LINES( span_start, span_length, det_offset, \
        last_line_index, rel_start, rel_end ) \
            (rel_start) = (span_start) - (det_offset); \
            (rel_end) = (rel_start) + (span_length) - 1; \
            if ( (rel_end) < 0 || (rel_start) > (last_line_index)) \
                (rel_end) = -1; \
            if ( (rel_start) < 0 ) \
                (rel_start) = 0; \
            if ( (rel_end) > (last_line_index) ) \
                (rel_end) = (last_line_index);

/* This structure is used for sorting the mean-subtracted detector striping
   metric while retaining the index of the original locations */
typedef struct array_map
{
    double value;
    int index;
} ARRAY_MAP;

/* Local routine prototypes */
static void compute_filters_and_diff
(
    int complete_lines,             /* I: Number of complete lines in the
                                          image */
    int number_detectors,           /* I: */
    float *left,                    /* I: */
    float *curr,                    /* I: */
    float *right,                   /* I: */
    double *cthomptr,               /* O: Across-track homogeneity filter
                                          for the current detector */
    double *athomptr_right,         /* O: Along-track homogeneity filter
                                          for the detector to the right of
                                          the current detector */
    double *ctdiff                  /* O: Across track difference between
                                          the detectors to the right and
                                          left of the current detector */
);

static int generate_reports
(
    const IAS_REPORT *report_info, /* I: General information for the report */
    int band_number,            /* I: */
    int sca_number,             /* I: */
    int number_detectors,       /* I: */
    int number_lines,           /* I: */
    double *ssm,                /* I: Full image scene striping metric */
    double *scm,                /* I: Full image striping correction matrix
                                      or NULL if output isn't desired */
    double *dsm                 /* I: Detector striping metric, number
                                      detectors in length */
);


/*-------------------------------------------------------------------------
   Two local routines for use as comparison routines for qsort.

   This comparison function is used by qsort for sorting double values
   in increasing order */
static int compare_doubles
(
    const void *one,
    const void *two
)
{
    const double *a = (const double *)one;
    const double *b = (const double *)two;
    if ( *a > *b )
        return 1;
    else if ( *a < *b )
        return -1;
    else
        return 0;
}


/* This comparison function is used by qsort for sorting ARRAY_MAP
   structures by the double value in decreasing order */
static int decreasing_compare_array_map
(
    const void *one,
    const void *two
)
{
    const ARRAY_MAP *a = (const ARRAY_MAP *)one;
    const ARRAY_MAP *b = (const ARRAY_MAP *)two;
    if ( a->value < b->value )
        return 1;
    else if ( a->value > b->value )
        return -1;
    else
        return 0;
}


/*----------------------------------------------------------------------------
NAME:    calculate_overall_striping_metric

PURPOSE: Determines the overall striping metric from the detector striping
         metric.  The result is calculated in radiance and DN.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int calculate_overall_striping_metric
(
    double *dsm,          /* I: Detector striping metric number_detectors
                                in length */
    int number_detectors, /* I: Number of detectors */
    double absolute_gain, /* I: CPF absolute gain value */
    double *osm_radiance, /* O: Overall striping metric in radiance */
    double *osm_dn        /* O: Overall striping metric in DN */
)
{
    int i;
    int status;
    int detector_index;
    int half_filter = MEDIAN_FILTER_WIDTH / 2;
    int partial_filter_width;/* Used to adjust the filter width on the edges */
    int number_peaks;    /* Number of peaks found */
    double mean_of_dsm;  /* Mean of the detector striping metric */
    double mean_of_peaks;/* Mean of the NUMBER_PEAKS peaks */
    size_t num_bytes;    /* Used to pre-compute the number of bytes for
                            memcpy */
    double dsm_subset[MEDIAN_FILTER_WIDTH]; /* Subset of the detector
                                               striping metric that is being
                                               smoothed */
    double *filter_start;/* Pointer to the starting sample of the filter */
    double *median_filter_buffer;   /* Contains the result of the median
                                       filtered DSM */
    double *mean_filter_buffer;     /* Contains the result of the smoothed
                                       median filtered results */
    ARRAY_MAP *mean_subtracted_dsm; /* This array contains the mean subtracted
                                       values and the original index of the
                                       sample.  The original indices are
                                       needed to maintain knowledge of its
                                       neighbors after it is sorted. */
    double maximum_peaks[NUMBER_PEAKS]; /* The maximum peak values found */
    int peak_indices[NUMBER_PEAKS];     /* Indices of the peaks */
    double overall_striping_metric; /* Metric in radiance */

    /* Allocate memory for two buffers for smoothing the DSM */
    median_filter_buffer = malloc( sizeof(double) * number_detectors );
    mean_filter_buffer = malloc( sizeof(double) * number_detectors );
    if ( !median_filter_buffer || !mean_filter_buffer )
    {
        free( mean_filter_buffer );
        free( median_filter_buffer );
        IAS_LOG_ERROR("Allocating smoothed detector striping metric");
        return ERROR;
    }

    /* Get the mean of the input vector, DSM.  This is used in the final
       computation of the overall metric. */
    status = ias_math_compute_mean( dsm, number_detectors, &mean_of_dsm );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Computing mean of detector striping metric");
        free( mean_filter_buffer );
        free( median_filter_buffer );
        return ERROR;
    }

    /*---------------------------------------------------------------------
       Run a median filter across dsm.  The filter is adjusted at the edges.
       On the left side there are no values to be used in the mean so the
       first value is replaced with the median of the first
       MEDIAN_FILTER_WIDTH / 2 + 1 values.  The next one adds in another
       value and so on.

       Note: The prototype code used the Matlab median for the first and
       last MEDIAN_FILTER_WIDTH / 2 + 1 values which uses the definition of
       median that averages the two middle values for the median.  The rest
       of the values were smoothed with the medfilt1 routine which uses the
       smaller of the two middle values.
    ---------------------------------------------------------------------*/
    detector_index = 0;
    filter_start = dsm;
    partial_filter_width = half_filter + 1;
    /* Smooth the first half_filter values */
    for ( i = 0; i < half_filter; i++ )
    {
        int median_index;
        double median;

        memcpy( &dsm_subset[0], filter_start,
            sizeof(double) * partial_filter_width );
        qsort( dsm_subset, partial_filter_width, sizeof(double),
            compare_doubles );

        median_index = partial_filter_width / 2;
        if ( partial_filter_width % 2 == 0 )
        {
            median = ( dsm_subset[median_index - 1] +
                dsm_subset[median_index] ) / 2.0;
        }
        else
        {
            median = dsm_subset[median_index];
        }
        /* Make the adjustment and update the pointers and filter width.  The
           filter start doesn't change in this loop, it is always pointing
           to the first element. */
        median_filter_buffer[detector_index++] = median;
        partial_filter_width++;
    }

    /* Smooth the middle.  The detector index and pointer to start of filter 
       don't need to be updated since they are at the proper locations from
       the previous loop. */
    num_bytes = sizeof(double) * MEDIAN_FILTER_WIDTH;
    for ( i = 0; i < number_detectors - MEDIAN_FILTER_WIDTH + 1; i++ )
    {
        memcpy( &dsm_subset[0], filter_start, num_bytes );
        qsort( dsm_subset, MEDIAN_FILTER_WIDTH, sizeof(double),
            compare_doubles );
        median_filter_buffer[detector_index++] = dsm_subset[half_filter];
        filter_start++;
    }

    /* As with the left, the right side has non-existent values beyond the 
       last detector.  The right side is filtered the same way as the left
       side.  Again the detector index and pointer to the start don't need
       updating. */
    partial_filter_width = MEDIAN_FILTER_WIDTH - 1;
    for ( i = 0; i < half_filter; i++ )
    {
        int median_index;
        double median;

        memcpy( &dsm_subset[0], filter_start,
            sizeof(double) * partial_filter_width );
        qsort( dsm_subset, partial_filter_width, sizeof(double),
            compare_doubles );

        median_index = partial_filter_width / 2;
        if ( partial_filter_width % 2 == 0 )
        {
            median = ( dsm_subset[median_index - 1] +
                dsm_subset[median_index] ) / 2.0;
        }
        else
        {
            median = dsm_subset[median_index];
        }
        median_filter_buffer[detector_index++] = median;
        partial_filter_width--;
        filter_start++;
    }

    /*---------------------------------------------------------------------
       Smooth the median filter buffer just created with an averaging filter.
       The filter length is MEAN_FILTER_WIDTH.  As with the median filter,
       the filter width is adjusted near the edges.
    ---------------------------------------------------------------------*/
    /* As defined in the ADD the first value isn't adjusted so just copy
       that value over */
    mean_filter_buffer[0] = median_filter_buffer[0];

    /* Set up and filter the left side */
    partial_filter_width = 3;
    filter_start = median_filter_buffer;
    detector_index = 1;
    for ( i = 0; i < MEAN_FILTER_WIDTH / 2 - 1; i++ )
    {
        int status = ias_math_compute_mean( filter_start, partial_filter_width,
            &mean_filter_buffer[detector_index] );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Computing mean for detector index %d",
                detector_index );
            free( median_filter_buffer );
            free( mean_filter_buffer );
            return ERROR;
        }
        /* Adjust the index and filter.  The start of the filter for this
           loop is always the first detector. */
        detector_index++;
        partial_filter_width += 2;
    }

    /* Filter the middle.  The detector index and filter start don't need
       updating - they are at the right place already. */
    for ( i = 0; i <= number_detectors - MEAN_FILTER_WIDTH; i++ )
    {
        int status = ias_math_compute_mean( filter_start, MEAN_FILTER_WIDTH,
            &mean_filter_buffer[detector_index] );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Computing mean for detector index %d",
                detector_index );
            free( median_filter_buffer );
            free( mean_filter_buffer );
            return ERROR;
        }
        detector_index++;
        filter_start++;
    }

    /* Filter the right.  The detector index doesn't need updating.  However
       the filter length and start must be updated. */
    partial_filter_width = MEAN_FILTER_WIDTH - 2;
    filter_start++;
    for ( i = 0; i < MEAN_FILTER_WIDTH / 2 - 1; i++ )
    {
        int status = ias_math_compute_mean( filter_start, partial_filter_width,
            &mean_filter_buffer[detector_index] );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Computing mean for detector index %d",
                detector_index );
            free( median_filter_buffer );
            free( mean_filter_buffer );
            return ERROR;
        }
        detector_index++;
        filter_start += 2;
        partial_filter_width -= 2;
    }

    /* As with the first value the last value is just copied */
    mean_filter_buffer[detector_index] = median_filter_buffer[detector_index];

    /* The median smoothed buffer is no longer needed */
    free( median_filter_buffer );
    median_filter_buffer = NULL;

    /*----------------------------------------------------------------------
       Subtract the final smoothed vector (combination of median and mean)
       from the input detector striping metric.  Since this is sorted later
       and the detector index is needed for finding peaks, use the ARRAY_MAP
       for this operation.  
    ----------------------------------------------------------------------*/
    /* Allocate the array */
    mean_subtracted_dsm = malloc( sizeof(ARRAY_MAP) * number_detectors );
    if ( !mean_subtracted_dsm )
    {
        IAS_LOG_ERROR("Allocating the mean subtracted metric array");
        free( mean_filter_buffer );
        return ERROR;
    }

    /* Subtract the smoothed buffer and assigned the result and the detector
       index to the array */
    for ( detector_index = 0; detector_index < number_detectors;
        detector_index++ )
    {
        mean_subtracted_dsm[detector_index].value = dsm[detector_index] -
            mean_filter_buffer[detector_index];
        mean_subtracted_dsm[detector_index].index = detector_index;
    }
    /* The mean filter buffer is no longer needed */
    free( mean_filter_buffer );
    mean_filter_buffer = NULL;

    /* Sort the subtracted result in reverse order */
    qsort( mean_subtracted_dsm, number_detectors, sizeof(ARRAY_MAP),
        decreasing_compare_array_map );

    /*--------------------------------------------------------------------
       Use the sorted array to find the top NUMBER_PEAKS peaks.  If a high
       value has a higher neighbor, it isn't a peak.  The first value in
       the array is the highest and can have no higher neighbor.
    --------------------------------------------------------------------*/
    /* Set the highest peak */
    number_peaks = 1;
    maximum_peaks[0] = mean_subtracted_dsm[0].value;
    peak_indices[0] = mean_subtracted_dsm[0].index;

    /* Loop through the rest of the array looking for the remaining peaks */
    for ( i = 1; i < number_detectors; i++ )
    {
        int j;
        int is_peak = 1;
        /* See if this is a true peak.  A true peak cannot be a neighbor of
           a previously identified peak.  Search the previous peaks for
           neighbors, if one is found set clear the flag. */
        for ( j = 0; j < number_peaks; j++ )
        {
            if ( peak_indices[j] - 1 == mean_subtracted_dsm[i].index
                || peak_indices[j] + 1 == mean_subtracted_dsm[i].index )
            {
                is_peak = 0;
                break;
            }
        }

        /* If it's a peak, identify it and keep its index */
        if ( is_peak )
        {
            peak_indices[number_peaks] = mean_subtracted_dsm[i].index;
            maximum_peaks[number_peaks++] = mean_subtracted_dsm[i].value;
        }

        /* Break out once we've identified all the peaks needed */
        if ( number_peaks == NUMBER_PEAKS )
            break;
    }
    free( mean_subtracted_dsm );
    mean_subtracted_dsm = NULL;

    /*-------------------------------------------------------------------
       Final computation is the Overall Striping Metric (OSM) and is the
       cube root of the product of the mean of DSM, maximum peak, and mean
       of all the peaks.  The OSM is also converted to DN by multiplying
       the band absolute gain.
    -------------------------------------------------------------------*/
    /* Compute mean of the peaks */
    status = ias_math_compute_mean( maximum_peaks, NUMBER_PEAKS,
        &mean_of_peaks );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Computing mean of %d peaks", NUMBER_PEAKS );
        return ERROR;
    }

    /* Compute OSM in radiance */
    overall_striping_metric =
        pow( mean_of_dsm * maximum_peaks[0] * mean_of_peaks, 1.0 / 3.0 );
    *osm_radiance = overall_striping_metric;
    /* Convert to DN */
    *osm_dn = overall_striping_metric * absolute_gain;

    return SUCCESS;
}  /* END calculate_overall_striping_metric */


/*----------------------------------------------------------------------------
NAME:    striping_characterization

PURPOSE: Computes the Scene Striping Correction Matrix (SSCM), Scene
         Striping Metric (SSM), and Detector Striping Metric (DSM).  The
         SSM is optionally written to disk if reporting is turned on.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int striping_characterization
(
    float *image_buffer,        /* I: One SCA of image data */
    int band_number,            /* I: */
    int sca_number,             /* I: */
    int number_detectors,       /* I: */
    int number_lines,           /* I: */
    int *detector_offsets,      /* I: Integer detector offsets with most
                                      negative offset shifted to zero */
    int max_offset,             /* I: */
    IAS_PIXEL_MASK *pixel_mask, /* I: */
    double striping_metric_cutoff, /* I: The CPF cutoff value */
    int dump_matrix_flag,       /* I: If flag is true (non-zero) the striping
                                      correction matrix will be dumped to a
                                      file.  If the flag is false (0) data
                                      will not be dumped. */
    const IAS_REPORT *report_info,/* I: General information for the report.
                                       If no report is desired a NULL pointer
                                       will be passed. */
    double *detector_striping_metric, /* O: Metric is number detectors in
                                            size */
    double *striping_correction_matrix/* O: Correction matrix sized the same
                                            as the image buffer */
)
{
    int status;
    int det;            /* Detector loop counter */
    int line;           /* Line loop counter */
    int complete_lines; /* Number of complete image lines.  Due to the
                           detector offsets the image must be spatially
                           aligned (nominally).  This will be set to the
                           number of complete image lines. */
    int last_line_index;/* This is one less than complete_lines and is pre-
                           computed in anticipation that it will be used
                           often */
    size_t buffer_size;
    float *left;        /* These are pointers into the input image buffer */
    float *curr;        /*   Pointer curr is for the current detector and */
    float *right;       /*   the others for the detector on either side.  */

    /* The following buffers contain data for the current detector being
       processed.  The detector is the current (center) detector of the
       three being utilized. */
    double *ctdiff;     /* The cross-track differences */
    double *cthom;      /* The cross-track homogeneity filter */
    double *acthom;     /* Noise-reduced cross-track homogeneity filter */
    double *aathom;     /* Noise-reduced along-track homogeneity filter */
    double *homfilt;    /* Final homogeneity filter, a combination of the
                           noise-reduced along- and cross-track filters */

    /* The following buffers and pointers are for the along-track
       homogeneity filters.  Each buffer contains the filter for one
       detector and will be rotated in FIFO fashion using the pointers.
       The buffers are number_lines in length. */
    double *athom0;
    double *athom1;
    double *athom2;
    double *athomptr_left;
    double *athomptr_curr;
    double *athomptr_right;

    /* Iterators and stuff for determining artifacts */
    char inoperable_detectors[number_detectors];
    IAS_PIXEL_MASK_ITERATOR *inop_it;
    IAS_PIXEL_MASK_ITERATOR *left_it;
    IAS_PIXEL_MASK_ITERATOR *curr_it;
    IAS_PIXEL_MASK_ITERATOR *right_it;
    IAS_PIXEL_MASK_SPAN inop_span;
    IAS_PIXEL_MASK_SPAN left_span;
    IAS_PIXEL_MASK_SPAN curr_span;
    IAS_PIXEL_MASK_SPAN right_span;
    int relative_span_start_line;
    int relative_span_end_line;
    int i;

    double ssm_one_det_sum;
    double *scene_striping_ptr;
    double *ssm_one_det;/* One detector of the scene striping metric */
    double *ssm = NULL; /* This is one full SCA of data. It is:
                           abs( ctdiff * homfilt ) and is only used if the
                           reporting is turned on. */
    double *report_scm; /* Pointer to the striping correction matrix
                           if its output is desired */

    int additional_lines_to_clear; /* Additional lines to clear due to the
                                      along-track filter size */
    int number_to_clear; /* Number of values in the blocks to clear */
    int lines_to_skip;   /* Number of lines to skip while clearing edges */
    int post_image_index;/* Index of the last block of image data to clear */

    /* Set the pointer for the striping correction matrix output */
    if ( dump_matrix_flag )
        report_scm = striping_correction_matrix;
    else
        report_scm = NULL;

    /* Allocate internal buffers:
         double number_lines:
           cthom, athom0, athom1, athom2
           acthom, aathom, homfilt, ctdiff 
         double number_lines * number_detectors:
           ssm
    ------------------------------------------------------------------*/
    buffer_size = number_lines * sizeof(double);
    cthom = malloc( buffer_size );
    if ( !cthom )
    {
        IAS_LOG_ERROR("Allocating the cross-track filter buffer");
        return ERROR;
    }

    athom0 = malloc( buffer_size );
    athom1 = malloc( buffer_size );
    athom2 = malloc( buffer_size );
    if ( !athom0 || !athom1 || !athom2 )
    {
        IAS_LOG_ERROR("Allocating the along-track filter buffers");
        free( cthom );
        free( athom0 );
        free( athom1 );
        free( athom2 );
        return ERROR;
    }

    acthom = malloc( buffer_size );
    aathom = malloc( buffer_size );
    homfilt = malloc( buffer_size );
    if ( !acthom || !aathom || !homfilt )
    {
        IAS_LOG_ERROR("Allocating the noise-reduced filter buffers");
        free( cthom );
        free( athom0 );
        free( athom1 );
        free( athom2 );
        free( acthom );
        free( aathom );
        free( homfilt );
        return ERROR;
    }

    ctdiff = malloc( buffer_size ); 
    ssm_one_det = malloc( buffer_size );
    if ( !ctdiff || !ssm_one_det )
    {
        IAS_LOG_ERROR("Allocating the cross-track difference buffer");
        free( cthom );
        free( athom0 );
        free( athom1 );
        free( athom2 );
        free( acthom );
        free( aathom );
        free( homfilt );
        free( ctdiff );
        free( ssm_one_det );
        return ERROR;
    }

    if ( report_info )
    {
        ssm = malloc( number_lines * number_detectors * sizeof(double));
        if ( !ssm )
        {
            IAS_LOG_ERROR("Allocating the scene striping metric buffer");
            free( cthom );
            free( athom0 );
            free( athom1 );
            free( athom2 );
            free( acthom );
            free( aathom );
            free( homfilt );
            free( ctdiff );
            free( ssm_one_det );
            return ERROR;
        }
    }

    /* Get an iterator for inoperable detectors and fill the inoperable
     * detector array */
    inop_it = ias_pm_get_iterator( pixel_mask, IAS_PM_INCLUDE,
        IAS_RPS_INOP_DETECTOR );
    if ( !inop_it )
    {
        IAS_LOG_ERROR("Getting an iterator for inoperable detectors");
        free( cthom );
        free( athom0 );
        free( athom1 );
        free( athom2 );
        free( acthom );
        free( aathom );
        free( homfilt );
        free( ctdiff );
        free( ssm_one_det );
        free( ssm );
        return ERROR;
    }
    memset( inoperable_detectors, 0, number_detectors );
    while ( ias_pm_get_next_span( inop_it, &inop_span ))
        inoperable_detectors[inop_span.detector_index] = 1;
    ias_pm_destroy_iterator( inop_it );
    inop_it = NULL;

    /* Compute a few useful values for processing */
    last_line_index = number_lines - max_offset - 1;
    complete_lines = last_line_index + 1;
    buffer_size = number_lines * sizeof(double);

    /* Pre-compute the along-track filter for detector 0 so it is ready for
       use on the first iteration of the main loop.  Since detector 0 
       across-track values aren't needed for the algorithm, pass in image
       pointers to detector 0 for all columns.  Neither across-track values
       are valid here.  Vector athom0 will contain valid along-track values
       for detector 0 */
    left = image_buffer + ( detector_offsets[0] * number_detectors );
    curr = right = left;
    athomptr_curr = athom0;
    /* If the detector is inoperable set the entire along-track vector to
       zero */
    if ( inoperable_detectors[0] )
    {
        memset( athomptr_curr, 0, buffer_size );
    }
    else
    {
        compute_filters_and_diff( complete_lines, number_detectors, left,
            curr, right, cthom, athomptr_curr, ctdiff );
    }
    /* The along-track filter for detector 1 also needs to be pre-computed.
       Since across-track values for detector 1 will be computed on the
       first iteration of the main loop, pass in image pointers to detector 1
       for all columns. Neither across-track values are valid here.  Vector
       athom1 will contain valid along-track values for detector 1 */
    left = image_buffer + 1 + ( detector_offsets[1] * number_detectors );
    curr = right = left;
    athomptr_curr = athom1;
    /* If the detector is inoperable set the filter to zero otherwise
       calculate it */
    if ( inoperable_detectors[1] )
    {
        memset( athomptr_curr, 0, buffer_size );
    }
    else
    {
        compute_filters_and_diff( complete_lines, number_detectors, left,
            curr, right, cthom, athomptr_curr, ctdiff );
    }

    /* At this point we have valid along-track filters for detector 0 and 1.
       If there are any artifacts for these detectors, other than being
       inoperable (already handled), they have to be taken care of before
       the main loop.  Since three iterators are needed go ahead and create
       all of them here.  */
    curr_it = ias_pm_get_iterator( pixel_mask, IAS_PM_EXCLUDE,
        IAS_RPS_INOP_DETECTOR );
    right_it = ias_pm_get_iterator( pixel_mask, IAS_PM_EXCLUDE,
        IAS_RPS_INOP_DETECTOR );
    left_it = ias_pm_get_iterator( pixel_mask, IAS_PM_EXCLUDE,
        IAS_RPS_INOP_DETECTOR );
    if ( !left_it || !curr_it || !right_it )
    {
        IAS_LOG_ERROR("Getting an iterator for inoperable detectors");
        free( cthom );
        free( athom0 );
        free( athom1 );
        free( athom2 );
        free( acthom );
        free( aathom );
        free( homfilt );
        free( ctdiff );
        free( ssm_one_det );
        free( ssm );
        ias_pm_destroy_iterator( left_it );
        ias_pm_destroy_iterator( curr_it );
        ias_pm_destroy_iterator( right_it );
        return ERROR;
    }

    /* Get the first span for the left iterator and the first span less than
       or equal to detector one  */
    if ( !ias_pm_get_next_span( left_it, &left_span ))
        left_span.detector_index = number_detectors;
    if ( !ias_pm_get_next_span( curr_it, &curr_span ))
        curr_span.detector_index = number_detectors;
    if ( !ias_pm_get_next_span( right_it, &right_span ))
        right_span.detector_index = number_detectors;

    /* Move the current and right spans to detector 1 or beyond */
    while ( curr_span.detector_index <= 1 )
    {
        if ( !ias_pm_get_next_span( curr_it, &curr_span ))
            curr_span.detector_index = number_detectors;
        if ( !ias_pm_get_next_span( right_it, &right_span ))
            right_span.detector_index = number_detectors;
    }

    /* All image pixels that don't have spatially aligned neighbors will
       have their output values set to zero.  To simplify the logic at the
       top and bottom of the image, pre-fill blocks of data at the beginning
       and end of the image and on the edges.
       ---------------------------------------------------------------------*/
    detector_striping_metric[0] =
        detector_striping_metric[number_detectors - 1] = 0.0;

    /* Clear two additional lines to account for the shortened along track 
       homogeneity filter */
    additional_lines_to_clear = 2;
    /* This is the number of values in the blocks to clear */
    number_to_clear = ( max_offset + additional_lines_to_clear ) *
        number_detectors;
    /* Compute the number of lines to skip less one.  There is one less line
       so the scene_striping_ptr[1] won't be accessed in the last iteration
       of the clearing loops below. */
    /* This is the number of lines in a block */
    lines_to_skip = max_offset + additional_lines_to_clear;
    /* There are two blocks */
    lines_to_skip = 2 * lines_to_skip;
    /* Skip one more line so the scene_striping_ptr[1] won't be accessed in
       the last iteration of the clearing loops below. */
    lines_to_skip++;

    /* Compute the index of the last block of data to clear.  In the
       following calculation, the last line is the last full line of the
       image.  However the algorithm doesn't fill this line so it must be
       cleared and that is why post image index is one less. */
    post_image_index = ( last_line_index - 1 ) * number_detectors;

    /* Clear the first and last blocks */
    memset( striping_correction_matrix, 0,
        sizeof(*striping_correction_matrix) * number_to_clear );
    memset( striping_correction_matrix + post_image_index, 0,
        sizeof(*striping_correction_matrix) * number_to_clear );

    /* Clear the first detector following the cleared block at the beginning */
    striping_correction_matrix[number_to_clear] = 0.0;
    /* Set the last detector and next line first detector to zero */
    scene_striping_ptr =
        &striping_correction_matrix[( max_offset + 2 ) * number_detectors - 1] +
        ( max_offset * number_detectors );
    scene_striping_ptr = striping_correction_matrix + number_to_clear +
        number_detectors - 1;
    for ( i = 0; i < number_lines - lines_to_skip; i++ )
    {
        /* The following is why lines_to_skip has one additional line added 
           above */
        scene_striping_ptr[0] = scene_striping_ptr[1] = 0.0;
        scene_striping_ptr += number_detectors;
    }
    /* Clear the last detector prior to the cleared block at the end */
    scene_striping_ptr[0] = 0.0;

    /* If the scene striping metric has been allocated, clear it in the same
       way as the striping correction matrix above. */
    if ( ssm )
    {
        /* Clear the first and last blocks */
        memset( ssm, 0, sizeof(*ssm) * number_to_clear );
        memset( ssm + post_image_index, 0, sizeof(*ssm) * number_to_clear );
        /* Clear the first detector following the cleared block at the
           beginning */
        ssm[number_to_clear] = 0.0;
        /* Set the last detector and next line first detector to zero */
        scene_striping_ptr = ssm + number_to_clear + number_detectors - 1;
        for ( i = 0; i < number_lines - lines_to_skip; i++ )
        {
            /* The following is why lines_to_skip has one additional line
               added above */
            scene_striping_ptr[0] = scene_striping_ptr[1] = 0.0;
            scene_striping_ptr += number_detectors;
        }
        /* Clear the last detector prior to the cleared block at the end */
        scene_striping_ptr[0] = 0.0;
    }

    /* Final set up: assign pointers to all the vectors so that when the
       pointers are rotated in the loop they are pointing to the correct
       data.  Also assign the image pointers. */
    athomptr_left = athom2;
    athomptr_curr = athom0;
    athomptr_right = athom1;
    curr = image_buffer + ( detector_offsets[0] * number_detectors );
    right = image_buffer + 1 + ( detector_offsets[1] * number_detectors );
    left = image_buffer + 2 + ( detector_offsets[2] * number_detectors );


    /*--------------------- main loop for filter ---------------------*/
    buffer_size = sizeof(double) * number_lines;
    for ( det = 1; det < number_detectors - 1; det++ )
    {
        /* Adjust the pointers */
        double *temp = athomptr_left;
        athomptr_left = athomptr_curr;
        athomptr_curr = athomptr_right;
        athomptr_right = temp;
        left = curr;
        curr = right;
        /* The right detector pointer is current detector plus 1 */
        right = image_buffer + det + 1 + ( detector_offsets[det + 1] *
            number_detectors );


        /* Compute:
             cross-track filter and difference for curr detector = det
             along-track filter for right detector = det+1
        -----------------------------------------------------------------*/
        compute_filters_and_diff( complete_lines, number_detectors,
            left, curr, right, cthom, athomptr_right, ctdiff );


        /* Smooth the along- and across-track filters.  The along-track
           filter is smoothed across-track with a width 3 mean filter.  The
           across-track filter is smoothed along-track with a width 5 mean
           filter except at the edges.  Because the along-track smoothing
           is greater than width 3, there are special handling for the top
           and bottom two lines to ensure we don't access beyond the array
           bounds.
        -----------------------------------------------------------------*/
        /* Smooth the first two lines */
        line = 0;
        acthom[line] = ( cthom[line] + cthom[line + 1] + cthom[line + 2] ) /
            3.0;
        aathom[line] = ( athomptr_left[line] + athomptr_curr[line] +
            athomptr_right[line] ) / 3.0;
        line = 1;
        acthom[line] = ( cthom[line - 1] + cthom[line] + cthom[line + 1] +
            cthom[line + 2] ) / 4.0;
        aathom[line] = ( athomptr_left[line] + athomptr_curr[line] +
            athomptr_right[line] ) / 3.0;

        /* Smooth the middle */
        for ( line = 2; line <= last_line_index - 2; line++ )
        {
            aathom[line] = ( athomptr_left[line] + athomptr_curr[line] +
                athomptr_right[line]) / 3.0;
            acthom[line] = ( cthom[line - 2] + cthom[line - 1] + cthom[line] +
                cthom[line + 1] + cthom[line + 2] ) / 5.0;
        }

        /* Smooth the last two lines */
        line = last_line_index - 1;
        acthom[line] = ( cthom[line - 2] + cthom[line - 1] + cthom[line] +
            cthom[line + 1] ) / 4.0;
        aathom[line] = ( athomptr_left[line] + athomptr_curr[line] +
            athomptr_right[line] ) / 3.0;
        line = last_line_index;
        acthom[line] = ( cthom[line - 2] + cthom[line - 1] + cthom[line] ) /
            3.0;
        aathom[line] = ( athomptr_left[line] + athomptr_curr[line] +
            athomptr_right[line] ) / 3.0;


        /* Compute the homogeneity filter.  As with the smoothing, the top
           and bottom lines have special handling -- in this case set them
           to zero.  The filter is the mean of the smoothed along- and
           across-track filters divided by the CPF striping metric cutoff
           raised to the 4th power; one is added and the inverse taken.
           Also take care of the image artifacts.  Artifacts affect their
           neighbors so they need to be taken into account.  If left,
           current or right detectors are inoperable, there is no need to
           calculate the filter.
        -----------------------------------------------------------------*/
        if ( inoperable_detectors[det - 1] || inoperable_detectors[det]
            || inoperable_detectors[det + 1] )
        {
            memset( homfilt, 0, buffer_size );
        }
        else
        {
            homfilt[0] = 0.0;
            double cutoff = 0.5 / striping_metric_cutoff;
            for ( line = 1; line < last_line_index; line++ )
            {
                double temp = fabs( acthom[line] + aathom[line] ) * cutoff;
                homfilt[line] = 1.0 / ( 1.0 + temp * temp * temp * temp );
            }
            homfilt[last_line_index] = 0.0;
        }

        /* Take care of any other artifacts.  NOTE: The pixel mask line
           numbers are zero-based and relative to the entire image.  The
           filters are nominally aligned and are relative to the first full
           line of adjacent pixels in the image.  So the start and end lines
           here are relative to the image subset that are relative to the
           full image lines.  Full in this context means a line where all
           detectors have spatially aligned neighbors.

           Handle the left side */
        while ( left_span.detector_index <= det - 1 )
        {
            if ( left_span.detector_index == det - 1 )
            {
                GET_SPAN_RELATIVE_LINES( left_span.starting_pixel_index,
                    left_span.length_of_span, detector_offsets[det - 1],
                    last_line_index, relative_span_start_line,
                    relative_span_end_line );

                for ( i = relative_span_start_line;
                    i <= relative_span_end_line; i++ )
                {
                    homfilt[i] = 0.0;
                }
            }
            if ( !ias_pm_get_next_span( left_it, &left_span ))
                left_span.detector_index = number_detectors;
        }
        /* The current detector */
        while ( curr_span.detector_index <= det )
        {
            if ( curr_span.detector_index == det )
            {
                /* Artifacts for the current detector also affect the
                   two along-track neighbors.  Set up the span for the start
                   of the detector, the macro handles the end.  If the index
                   is zero simply add to the length of the span.  If it's
                   greater than 0, back up the starting index one and add
                   two to the length. */
                if ( curr_span.starting_pixel_index > 0 )
                {
                    curr_span.starting_pixel_index -= 1;
                    curr_span.length_of_span += 2;
                }
                else if ( curr_span.starting_pixel_index == 0 )
                {
                    curr_span.length_of_span += 1;
                }

                GET_SPAN_RELATIVE_LINES( curr_span.starting_pixel_index,
                    curr_span.length_of_span, detector_offsets[det],
                    last_line_index, relative_span_start_line,
                    relative_span_end_line );

                for ( i = relative_span_start_line;
                    i <= relative_span_end_line; i++ )
                {
                    homfilt[i] = 0.0;
                }
            }
            if ( !ias_pm_get_next_span( curr_it, &curr_span ))
                curr_span.detector_index = number_detectors;
        }
        /* and the right detector */
        while ( right_span.detector_index <= det + 1 )
        {
            if ( right_span.detector_index == det + 1 )
            {
                GET_SPAN_RELATIVE_LINES( right_span.starting_pixel_index,
                    right_span.length_of_span, detector_offsets[det + 1],
                    last_line_index, relative_span_start_line,
                    relative_span_end_line );

                for ( i = relative_span_start_line;
                    i <= relative_span_end_line; i++ )
                {
                    homfilt[i] = 0.0;
                }
            }
            if ( !ias_pm_get_next_span( right_it, &right_span ))
                right_span.detector_index = number_detectors;
        }


        /* For the current detector compute:
             Striping correction matrix (SCM), this is one of the outputs of
                 the algorithm and is copied to the output here.  The SCM is
                 the mean of across-track difference and homogeneity filter.
             Scene striping metric (SSM), if the report generation is desired
                 this will be copied to a full image.  The SSM is the absolute
                 value of the product of the across-track difference and
                 homogeneity filter.
             Detector striping metric (DSM), this is the other output of the
                 algorithm and is copied to the output here.  The DSM is the
                 mean of the SSM for the current detector.
        -------------------------------------------------------------------*/
        scene_striping_ptr = striping_correction_matrix + det + 
            ( detector_offsets[det] * number_detectors );
        ssm_one_det_sum = 0.0;
        for ( line = 1; line < last_line_index; line++ )
        {
            double temp = ctdiff[line] * homfilt[line];
            scene_striping_ptr[0] = temp * 0.5;
            ssm_one_det[line] = fabs( temp );
            ssm_one_det_sum += ssm_one_det[line];
            scene_striping_ptr += number_detectors;
        }

        /* The detector striping metric is the mean of the detector scene
           striping metric */
        detector_striping_metric[det] = ssm_one_det_sum / 
            ( complete_lines - 2 );

        /* If the report was specified copy the scene striping metric to
           this detector to the output array */
        if ( ssm != NULL )
        {
            scene_striping_ptr = ssm + det + ( detector_offsets[det] *
                number_detectors );
            for ( line = 1; line < last_line_index; line++ )
            {
                scene_striping_ptr[0] = ssm_one_det[line];
                scene_striping_ptr += number_detectors;
            }
        }
    }  /* END for detector loop */

    /* Output the reports if needed */
    status = SUCCESS;
    if ( report_info != NULL )
    {
        status = generate_reports( report_info, band_number, sca_number,
            number_detectors, number_lines, ssm, report_scm,
            detector_striping_metric );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Generating reports for band %d SCA %d",
                band_number, sca_number );
        }
    }

    free( cthom );
    free( athom0 );
    free( athom1 );
    free( athom2 );
    free( acthom );
    free( aathom );
    free( homfilt );
    free( ctdiff );
    free( ssm_one_det );
    free( ssm );
    ias_pm_destroy_iterator( left_it );
    ias_pm_destroy_iterator( curr_it );
    ias_pm_destroy_iterator( right_it );

    return status;
}  /* END striping_characterization */


/*----------------------------------------------------------------------------
  This internal routine calculates the cross-track difference and filter for
  the current detector.  It also calculates the along-track filter for the
  right detector.  It was set up this way so that there is one for each line
  loop that makes these calculations.  It was also pulled out of the main
  routine so that it could be used to pre-calculate the along-track filter
  for detectors 0 and 1 prior to the main filter creation loop.
-----------------------------------------------------------------------------*/
static void compute_filters_and_diff
(
    int complete_lines,             /* I: */
    int number_detectors,           /* I: */
    float *left,                    /* I: */
    float *curr,                    /* I: */
    float *right,                   /* I: */
    double *cthomptr,               /* O: */
    double *athomptr_right,         /* O: */
    double *ctdiff                  /* O: */
)
{
    int line;
    float *left_ptr = left;
    float *curr_ptr = curr;
    float *right_ptr = right;
    float *previous_right_line;
    float *next_right_line;

    /* Get the cross-track values for the first line */
    line = 0;
    ctdiff[line] = *curr_ptr - ( *left_ptr + *right_ptr ) * 0.5;
    cthomptr[line] = fabs( *left_ptr - *right_ptr );

    /* Loop through all but the last line */
    for ( line = 1; line < complete_lines - 1; line++ )
    {
        /* Update the pointers into the image lines */
        left_ptr += number_detectors;
        curr_ptr += number_detectors;
        /* These are updated for the along-track values */
        previous_right_line = right_ptr;
        right_ptr += number_detectors;
        next_right_line = right_ptr + number_detectors;

        /* Compute the cross-track values */
        ctdiff[line] = *curr_ptr - ( *left_ptr + *right_ptr ) * 0.5;
        cthomptr[line] = fabs( *left_ptr - *right_ptr );

        /* Compute the along-track value */
        athomptr_right[line] = fabs( *right_ptr - ( *previous_right_line +
            *next_right_line ) * 0.5 );
    }

    /* Get the cross-track values for the last line */
    line = complete_lines - 1;
    left_ptr += number_detectors;
    curr_ptr += number_detectors;
    right_ptr += number_detectors;
    next_right_line = right_ptr + number_detectors;
    ctdiff[line] = *curr_ptr - ( *left_ptr + *right_ptr ) * 0.5;
    cthomptr[line] = fabs( *left_ptr - *right_ptr );
}  /* END compute_filters_and_diff */


/*----------------------------------------------------------------------------
  Database section
----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------
NAME:    db_store_striping_metric

PURPOSE: Routine to store the striping metric in the database

RETURNS: SUCCESS or ERROR
----------------------------------------------------------------------------*/
int db_store_overall_striping_metric_in_database
(
    DB_OVERALL_STRIPING_METRIC_RESULTS* db_results /* I: Pointer to the
                                                         results */
)
{
    int number_db_columns;   /* Number of database columns being inserted */
    int status = SUCCESS;    /* Return status of database insert */

    /* Create the table */
    IAS_DB_INSERT_TABLE table[]=
    {
        IAS_DB_TABLE_STRING( "radiometric_wo_common_id",
            db_results->rad_common_id ),
        IAS_DB_TABLE_INT_ARRAY("band_id", db_results->band_numbers ),
        IAS_DB_TABLE_DOUBLE_ARRAY("striping_factor_radiance",
            db_results->osm_radiance ),
        IAS_DB_TABLE_DOUBLE_ARRAY("striping_factor_dn", db_results->osm_dn )
    };

    /* Get the number of columns inserted into the table and insert the
       data into the database */
    number_db_columns = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);
    status = ias_db_insert_using_table( db_results->dbase, "striping", table,
        number_db_columns, db_results->number_bands );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Inserting overall striping metric");
    }

    return status;
}  /* END db_store_striping_metric */


/*----------------------------------------------------------------------------
NAME:    db_allocate_overall_striping_metric_results

PURPOSE: Allocate a structure to hold the overall striping metric results

RETURNS: Pointer to the DB_OVERALL_STRIPING_METRIC_RESULTS structure or
         NULL on error.
----------------------------------------------------------------------------*/
DB_OVERALL_STRIPING_METRIC_RESULTS *db_allocate_overall_striping_metric_results
(
    int number_of_bands /* I: Number of bands to allocate */
)
{
    DB_OVERALL_STRIPING_METRIC_RESULTS *db_results = malloc(
        sizeof(DB_OVERALL_STRIPING_METRIC_RESULTS));
    if ( !db_results )
    {
        IAS_LOG_ERROR("Allocating overall striping metric results structure");
        return NULL;
    }

    db_results->number_bands = number_of_bands;
    db_results->band_numbers = malloc( sizeof(int) * number_of_bands );
    db_results->osm_radiance = malloc( sizeof(double) * number_of_bands );
    db_results->osm_dn = malloc( sizeof(double) * number_of_bands );
    if ( !db_results->band_numbers || !db_results->osm_radiance
        || !db_results->osm_dn )
    {
        IAS_LOG_ERROR("Allocating overall striping metric results");
        db_free_overall_striping_metric_results( db_results );
        return NULL;
    }

    return db_results;
}  /* END db_allocate_overall_striping_metric_results */


/*----------------------------------------------------------------------------
NAME:    db_free_overall_striping_metric_results

PURPOSE: Free the structure allocated by
         db_allocate_overall_striping_metric_results

RETURNS: none
----------------------------------------------------------------------------*/
void db_free_overall_striping_metric_results
(
    DB_OVERALL_STRIPING_METRIC_RESULTS *db_results /* I: Structure to free */
)
{
    if ( db_results )
    {
        free( db_results->band_numbers );
        free( db_results->osm_radiance );
        free( db_results->osm_dn );
        free( db_results );
    }
}  /* END db_free_overall_striping_metric_results */


/*----------------------------------------------------------------------------
  Report section
----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------
NAME:    generate_reports

PURPOSE: Local routine to write out the report

RETURNS: SUCCESS or ERROR
----------------------------------------------------------------------------*/
static int generate_reports
(
    const IAS_REPORT *report_info,/* I: General information for the report */
    int band_number,            /* I: */
    int sca_number,             /* I: */
    int number_detectors,       /* I: */
    int number_lines,           /* I: */
    double *ssm,                /* I: Full image scene striping metric */
    double *scm,                /* I: Full image striping correction matrix
                                      or NULL if output isn't desired */
    double *dsm                 /* I: Detector striping metric, number
                                      detectors in length */
)
{
    FILE *report_file;
    char dsm_file_base_name[] = "striping_characterization_dsm";
    char ssm_file_base_name[] = "binary_striping_characterization_ssm";
    char scm_file_base_name[] = "binary_striping_characterization_scm";
    char *file_name;
    int status;
    IAS_REPORT_HEADER *copy_of_report_header;
    int detector_index;

    /* Get the standard report file name */
    file_name = ias_rps_get_report_file_name( report_info, band_number,
        sca_number, dsm_file_base_name );
    if ( !file_name )
    {
        IAS_LOG_ERROR("Getting the standard report file name");
        return ERROR;
    }

    /* Open the report file */
    report_file = fopen( file_name, "w" );
    if ( !report_file )
    {
        IAS_LOG_ERROR("Opening report file: %s", file_name );
        free( file_name );
        return ERROR;
    }

    /* Get a local copy of the report header since the "standard" header
       includes non-standard things such as band, SCA and algorithm name.
       In a threaded environment the input header cannot be manipulated. */
    copy_of_report_header = ias_misc_copy_report_header( report_info->header );
    if ( !copy_of_report_header )
    {
        IAS_LOG_ERROR("Creating local copy of report header");
        fclose( report_file );
        free( file_name );
        return ERROR;
    }

    /* Initialize band number, SCA number, and algorithm name members
       of the IAS report header and write the header. */
    ias_misc_set_header_band_number( copy_of_report_header, band_number );
    ias_misc_set_header_sca_number( copy_of_report_header, sca_number );
    ias_misc_set_header_algorithm_name( copy_of_report_header,
        "Striping Characterization -- Detector Striping Metric" );
    status = ias_misc_write_report_header( report_file,
        copy_of_report_header );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Writing out the standard header");
        ias_misc_free_report_header( copy_of_report_header );
        fclose( report_file );
        free( file_name );
        return ERROR;
    }
    ias_misc_free_report_header( copy_of_report_header );

    /* Write out a header */
    status = fprintf( report_file, "\nDetector      Value\n" );
    if ( status < 0 )
    {
        IAS_LOG_ERROR("Writing to report file %s", file_name );
        fclose( report_file );
        free( file_name );
        return ERROR;
    }

    /* Write the detector striping metric */
    for ( detector_index = 0; detector_index < number_detectors;
        detector_index++ )
    {
        status = fprintf( report_file, " %4d     %e\n", detector_index + 1,
            dsm[detector_index] );
        if ( status < 0 )
        {
            IAS_LOG_ERROR("Writing data to report file %s", file_name );
            fclose( report_file );
            free( file_name );
            return ERROR;
        }
    }
    status = fclose( report_file );
    free( file_name );
    if ( status != 0 )
    {
        IAS_LOG_ERROR("Closing file %s", file_name );
        return ERROR;
    }

    /* Get the standard report file name */
    file_name = ias_rps_get_report_file_name( report_info, band_number,
        sca_number, ssm_file_base_name );
    if ( !report_file )
    {
        IAS_LOG_ERROR("Getting the standard report file name");
        return ERROR;
    }

    /* Open the report file */
    report_file = fopen( file_name, "w" );
    if ( !report_file )
    {
        IAS_LOG_ERROR("Opening report file: %s", file_name );
        free( file_name );
        return ERROR;
    }

    /* Write the scene striping metric.  This is a full image so it's
       written in binary. */
    status = fwrite( ssm, sizeof(double), number_detectors * number_lines,
        report_file );
    if ( status != number_detectors * number_lines )
        IAS_LOG_ERROR("Writing to file: %s", file_name );

    status = fclose( report_file );
    free( file_name );
    if ( status != 0 )
    {
        IAS_LOG_ERROR("Closing file %s", file_name );
        return ERROR;
    }

    /* Get the standard report file name */
    file_name = ias_rps_get_report_file_name( report_info, band_number,
        sca_number, scm_file_base_name );
    if ( !report_file )
    {
        IAS_LOG_ERROR("Getting the standard report file name");
        return ERROR;
    }

    /* Only output the scene striping metric if its pointer isn't NULL */
    if ( scm != NULL )
    {
        /* Open the report file */
        report_file = fopen( file_name, "w" );
        if ( !report_file )
        {
            IAS_LOG_ERROR("Opening report file: %s", file_name );
            free( file_name );
            return ERROR;
        }

        /* Write the scene striping metric.  This is a full image so it's
           written in binary. */
        status = fwrite( scm, sizeof(double), number_detectors * number_lines,
            report_file );
        if ( status != number_detectors * number_lines )
            IAS_LOG_ERROR("Writing to file: %s", file_name );

        status = fclose( report_file );
        free( file_name );
        if ( status != 0 )
        {
            IAS_LOG_ERROR("Closing file %s", file_name );
            return ERROR;
        }
    }

    return SUCCESS;
}  /* END generate_reports */


/*----------------------------------------------------------------------------
NAME:    generate_overall_striping_metric_report

PURPOSE: Public routine to write out the overall striping metric report.
         This is public since it is called once in the post-process for all
         bands.

RETURNS: SUCCESS or ERROR
----------------------------------------------------------------------------*/
int generate_overall_striping_metric_report
(
    const IAS_REPORT *report_info,  /* I: General information for the report */
    DB_OVERALL_STRIPING_METRIC_RESULTS *results /* I: Pointer to the results */
)
{
    FILE *report_file;
    char osm_file_base_name[] = "striping_characterization_osm";
    char *file_name;
    int status;
    IAS_REPORT_HEADER *copy_of_report_header;
    int index;

    /* Get the standard report file name */
    file_name = ias_rps_get_report_file_name( report_info, -1, -1,
        osm_file_base_name );
    if ( !file_name )
    {
        IAS_LOG_ERROR("Getting the standard report file name");
        return ERROR;
    }

    /* Open the report file */
    report_file = fopen( file_name, "w" );
    if ( !report_file )
    {
        IAS_LOG_ERROR("Opening report file: %s", file_name );
        free( file_name );
        return ERROR;
    }

    /* Get a local copy of the report header since the "standard" header
       includes non-standard things such as band, SCA and algorithm name.
       In a threaded environment the input header cannot be manipulated. */
    copy_of_report_header = ias_misc_copy_report_header( report_info->header );
    if ( !copy_of_report_header )
    {
        IAS_LOG_ERROR("Creating local copy of report header");
        fclose( report_file );
        free( file_name );
        return ERROR;
    }

    /* Initialize band number, SCA number, and algorithm name members
       of the IAS report header and write the header. */
    ias_misc_set_header_algorithm_name( copy_of_report_header,
        "Striping Characterization - Overall Striping Metric" );
    status = ias_misc_write_report_header( report_file,
        copy_of_report_header );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Writing out the standard header");
        ias_misc_free_report_header( copy_of_report_header );
        fclose( report_file );
        free( file_name );
        return ERROR;
    }
    ias_misc_free_report_header( copy_of_report_header );

    /* Write out a header */
    status = fprintf( report_file, "\n"
        "               Overall Metric\n"
        "Band      Radiance           DN\n" );
    if ( status < 0 )
    {
        IAS_LOG_ERROR("Writing to report file %s", file_name );
        fclose( report_file );
        free( file_name );
        return ERROR;
    }

    /* Write the detector striping metric */
    for ( index = 0; index < results->number_bands; index++ )
    {
        status = fprintf( report_file, " %2d     %e     %e\n",
            results->band_numbers[index], results->osm_radiance[index],
            results->osm_dn[index] );
        if ( status < 0 )
        {
            IAS_LOG_ERROR("Writing data to report file %s", file_name );
            fclose( report_file );
            free( file_name );
            return ERROR;
        }
    }
    status = fclose( report_file );
    free( file_name );
    if ( status != 0 )
    {
        IAS_LOG_ERROR("Closing file %s", file_name );
        return ERROR;
    }

    return SUCCESS;
}  /* END generate_overall_striping_metric_report */
