/***************************************************************************
 NAME:      api_3q_response_linearization

 PURPOSE:   Routines needed to gather data for the OLI/TIRS 3q response
            linearization algorithm routine and call it

 ROUTINES:  decrement_counters            (internal routine)
            api_oli_3q_response_linearization
            api_tirs_3q_response_linearization
            api_3q_response_linearization    (internal routine)
            response_linearization_3q_preprocess
****************************************************************************/
#include <stdlib.h>
#include <string.h>
#include "ias_logging.h"
#include "ias_rlut.h"
#include "ias_satellite_attributes.h"
#include "rps_local.h"
#include "ias_rps_structures.h"  /* RLUT_VERSION_LENGTH */

/* Define a list of memory buffers needed by the algorithm and the overall
   size of this data structure */
static MEMORY_BUFFER_TYPE memory_buffers[] =
{
    IMAGE_L1R,
    MM_3Q_RESPONSE_LINEARIZATION_PARAMS
};

static int number_of_memory_buffers =
    sizeof(memory_buffers) / sizeof(MEMORY_BUFFER_TYPE);


/*----------------------------------------------------------------------------
 NAME:    decrement_counters

 PURPOSE: Internal routine to decrement the consumer counts of all required
          memory buffers

 RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int decrement_counters
(
    int band_index,     /* I: */
    int sca_index       /* I: */
)
{
    int i;                                  /* Local loop counter */
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int vrp_index = band_attributes->vrp_band_index;
    int blind_index = band_attributes->blind_band_index;
    int status;
    if ( ! band_attributes->process_blind_band_flag )
        blind_index = -1;

    /* Decrement the consumer counts for the memory buffers */
    for (i = 0; i < number_of_memory_buffers; i++)
    {
        status = decrement_consumer_count(memory_buffers[i], band_index,
            sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
        if ( vrp_index >= 0 )
        {
            status = decrement_consumer_count(memory_buffers[i], vrp_index,
                sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
        }
        if ( blind_index >= 0 )
        {
            status = decrement_consumer_count(memory_buffers[i], blind_index,
                sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
        }
    }
    return SUCCESS;
}   /* END decrement_counters */


/**************************************************************************
 NAME:     api_3q_response_linearization

 PURPOSE:  Main API call

 RETURNS:  Integer status code of SUCCESS or ERROR
***************************************************************************/
static int api_3q_response_linearization
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int band_index,     /* I: */
    int sca_index       /* I: */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes(band_index);
    int band_number = band_attributes->band_number;
    int number_lines = band_attributes->number_lines;
    int number_detectors = band_attributes->number_detectors;
    const IAS_RLUT_LINEARIZATION_PARAMS *params = NULL;
                                   /* Array of linearization parameter
                                      data structures for all detectors
                                      in the current band/SCA */
    float *data_buffer = NULL;     /* Level 1 data buffer */
    int status;

    /* Note: It isn't necessary to decrement the usage of the memory buffer
       on error conditions from this point on.  The main thread will clean
       up allocated buffers when error conditions dictate.  It doesn't hurt
       but it isn't necessary. */

    /* Add the ID for response linearization to the RPS processing level */
    if (set_rps_level(band_index, sca_index, 
            IAS_RPS_LEVEL_RESPONSE_LINEAR) == ERROR)
    {
        IAS_LOG_ERROR("Updating the RPS processing level");
        return ERROR;
    }

    /* This algorithm doesn't process OLI blind or VRP bands therefore the
        processing level *must* be set to skip */
    if ( band_attributes->vrp_band_index >= 0 )
    {
        if (set_rps_level(band_attributes->vrp_band_index, sca_index,
            IAS_RPS_LEVEL_SKIP) == ERROR)
        {
            IAS_LOG_ERROR("Updating RPS processing level");
            return ERROR;
        }
    }
    if ( band_attributes->blind_band_index >= 0 &&
         band_attributes->process_blind_band_flag )
    {
        if (set_rps_level(band_attributes->blind_band_index, sca_index,
            IAS_RPS_LEVEL_SKIP) == ERROR)
        {
            IAS_LOG_ERROR("Updating RPS processing level");
            return ERROR;
        }
    }

    /* Get the image data buffer */
    data_buffer = (float *)get_memory_buffer(IMAGE_L1R, CONSUMER, band_index,
        sca_index);
    if (data_buffer == NULL)
    {
        IAS_LOG_ERROR("Getting image memory buffer for band %d SCA %d",
            band_number, sca_index + 1);
        return ERROR;
    }

    /* Get the correct set of linearization parameters */
    params = (const IAS_RLUT_LINEARIZATION_PARAMS *)get_memory_buffer(
        MM_3Q_RESPONSE_LINEARIZATION_PARAMS, CONSUMER, band_index, sca_index);
    if (params == NULL)
    {
        IAS_LOG_ERROR("Getting linearization parameters for band %d SCA %d",
            band_number, sca_index + 1);
        return ERROR;
    }

    /* Do the linearization correction */
    status = ias_rps_3q_response_linearization(data_buffer, params,
        number_detectors, number_lines);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("In call from ias_rps_3q_response_linearization for "
            "band %d SCA %d", band_number, sca_index + 1);
        return ERROR;
    }

    /* Set the band metadata */
    if ( sca_index == 0 )
    {
        strcpy( blob->l1r_band_metadata[band_index]->linearization,
            "Standard");
    }

    /* Decrement the memory management counters */
    status = decrement_counters(band_index, sca_index);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }

    return SUCCESS;
}   /* END api_3q_response_linearization */



/**************************************************************************
 NAME:     api_oli_3q_response_linearization

 PURPOSE:  "Wrapper" routine around main API call.  Ensures that the
           current band/SCA is an OLI band before calling the API
           routine

 RETURNS:  Integer status code of SUCCESS or ERROR
***************************************************************************/
int api_oli_3q_response_linearization
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int band_index,     /* I: */
    int sca_index,      /* I: */
    int algo_index      /* I: Algorithm index */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int status;

    /* Only process OLI bands */
    if (band_attributes->spectral_type == IAS_SPECTRAL_THERMAL)
    {
        status = decrement_counters(band_index, sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
        return SUCCESS;
    }

    if (band_attributes->band_classification != IAS_NORMAL_BAND)
    {
        status = decrement_counters(band_index, sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
        return SUCCESS;
    }

    return api_3q_response_linearization(blob, band_index, sca_index);
}   /* END api_oli_3q_response_linearization */



/**************************************************************************
 NAME:     api_tirs_3q_response_linearization

 PURPOSE:  "Wrapper" routine around main API call.  Ensures that the
           current band/SCA is a TIRS band before calling the API routine

 RETURNS:  Integer status code of SUCCESS or ERROR
***************************************************************************/
int api_tirs_3q_response_linearization
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int band_index,     /* I: */
    int sca_index,      /* I: */
    int algo_index      /* I: Algorithm index */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int status;

    /* Only process TIRS bands */
    if (band_attributes->spectral_type != IAS_SPECTRAL_THERMAL)
    {
        status = decrement_counters(band_index, sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
        return SUCCESS;
    }

    if (band_attributes->band_classification == IAS_SECONDARY_BAND)
    {
        status = decrement_counters(band_index, sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
        return SUCCESS;
    }

    return api_3q_response_linearization(blob, band_index, sca_index);
}



/***************************************************************************
 NAME:    response_linearization_3q_preprocess

 PURPOSE: Do preprocessing for the algorithm by incrementing the memory
          buffer consumer counts.

 RETURNS: SUCCESS or ERROR
****************************************************************************/
int response_linearization_3q_preprocess
(
    BLOB *blob                 /* I: Miscellaneous parameters */
)
{
    int i;                     /* Local loop counter */
    int status;                /* Status flag */
    IAS_RLUT_FILE_ATTRIBUTES rlut_attr;  /* RLUT file attributes structure */
    
    /* Open the RLUT file */
    if ( blob->rlut == NULL )
    {
        blob->rlut = ias_rlut_open_file( blob->rlut_filename, IAS_READ );
        if ( !blob->rlut )
        {
            IAS_LOG_ERROR("Opening RLUT file %s to read", blob->rlut_filename );
            return ERROR;
        }
    }

    /* Allocate memory in the report structure for the RLUT version */
    if (blob->report_information.rlut_version == NULL)
    {
        char *rlut_version = malloc(RLUT_VERSION_LENGTH * sizeof(char));
        blob->report_information.rlut_version = rlut_version;
        if ( rlut_version == NULL )
        {
            IAS_LOG_ERROR("Allocating memory for RLUT version in report info");
            return ERROR;
        }

        /* Read the RLUT file attributes */
        status = ias_rlut_read_file_attributes(blob->rlut, &rlut_attr);
        if ( status == ERROR )
        {
            IAS_LOG_ERROR("Reading RLUT file %s attributes",
                blob->rlut_filename);
            return ERROR;
        }

        /* Copy the RLUT version to the report information structure */
        status = snprintf( rlut_version, RLUT_VERSION_LENGTH, 
            "R%d", rlut_attr.version );
        if ( status < 0 || status >= RLUT_VERSION_LENGTH )
        {
            IAS_LOG_ERROR("RLUT version length exceeds allocate memory");
            return ERROR;
        }
    }

    /* Increment the memory buffer consumer counts */
    for (i = 0; i < number_of_memory_buffers; i++)
    {
        status = increment_consumer_count(memory_buffers[i]);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string(memory_buffers[i]));
            return ERROR;
        }
    }

    return SUCCESS;
}   /* END response_linearization_3q_preprocess */
